;========================================================================================
;  fputest.s
;  Copyright (C) 2003-2019 Makoto Kamada
;
;  This file is part of the XEiJ (X68000 Emulator in Java).
;  You can use, modify and redistribute the XEiJ if the conditions are met.
;  Read the XEiJ License for more details.
;  https://stdkmd.net/xeij/
;========================================================================================

	.include	doscall.mac
	.include	iocscall.mac

	.cpu	68000


;--------------------------------------------------------------------------------
;	定数
;--------------------------------------------------------------------------------

MI	equ	1<<27
ZE	equ	1<<26
IN	equ	1<<25
NA	equ	1<<24
BS	equ	1<<15
SN	equ	1<<14
OE	equ	1<<13
OF	equ	1<<12
UF	equ	1<<11
DZ	equ	1<<10
X2	equ	1<<9
X1	equ	1<<8
AV	equ	1<<7
AO	equ	1<<6
AU	equ	1<<5
AZ	equ	1<<4
AX	equ	1<<3


;--------------------------------------------------------------------------------
;	マクロ
;--------------------------------------------------------------------------------

leamsg	.macro	p0,an
	.data
@msg:
	.dc.b	p0,0
	.text
	lea.l	@msg,an
	.endm

peamsg	.macro	p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,q0,q1,q2,q3,q4,q5,q6,q7,q8,q9
	.data
@msg:
	.dc.b	p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,q0,q1,q2,q3,q4,q5,q6,q7,q8,q9,0
	.text
	pea.l	@msg
	.endm

putmsg	.macro	p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,q0,q1,q2,q3,q4,q5,q6,q7,q8,q9
	peamsg	p0,p1,p2,p3,p4,p5,p6,p7,p8,p9,q0,q1,q2,q3,q4,q5,q6,q7,q8,q9
	jbsr	printstr
	addq.l	#4,sp
	.endm

putstr	.macro	p0
	move.l	p0,-(sp)
	jbsr	printstr
	addq.l	#4,sp
	.endm

putchr	.macro	p0
	move.b	p0,-(sp)
	jbsr	printchr
	addq.l	#2,sp
	.endm

putlong	.macro	p0
	move.l	p0,-(sp)
	jbsr	printlong
	addq.l	#4,sp
	.endm

putdec	.macro	p0
	move.l	p0,-(sp)
	jbsr	printdec
	addq.l	#4,sp
	.endm

putdecz2	.macro	p0
	pea.l	2.w
	move.l	p0,-(sp)
	jbsr	printdecz
	addq.l	#8,sp
	.endm

putdecz4	.macro	p0
	pea.l	4.w
	move.l	p0,-(sp)
	jbsr	printdecz
	addq.l	#8,sp
	.endm

putfix	.macro	p0,p1
	move.b	p1,-(sp)
	move.l	p0,-(sp)
	jbsr	printfix
	addq.l	#6,sp
	.endm

puthex2	.macro	p0
	move.b	p0,-(sp)
	jbsr	printhex2
	addq.l	#2,sp
	.endm

puthex4	.macro	p0
	move.w	p0,-(sp)
	jbsr	printhex4
	addq.l	#2,sp
	.endm

puthex8	.macro	p0
	move.l	p0,-(sp)
	jbsr	printhex8
	addq.l	#4,sp
	.endm

puthex16	.macro	p0,p1
	puthex8	p0
	putchr	#','
	puthex8	p1
	.endm

puthex24	.macro	p0,p1,p2
	puthex8	p0
	putchr	#','
	puthex8	p1
	putchr	#','
	puthex8	p2
	.endm

putcrlf	.macro
	jbsr	printcrlf
	.endm

putdate	.macro
	jbsr	printdate
	.endm


;--------------------------------------------------------------------------------
;	メイン
;--------------------------------------------------------------------------------
	.text
	.even
main::
;------------------------------------------------
;bssが確保されているか確認する
	lea.l	(16,a0),a0
	suba.l	a0,a1
	movem.l	a0-a1,-(sp)
	DOS	_SETBLOCK
	addq.l	#8,sp
	tst.l	d0
	bmi	exit
;------------------------------------------------
;スタックエリアを設定する
	lea.l	stack_area_end,sp	;スタックエリアの末尾
;------------------------------------------------
;FPUの種類を確認する
	jbsr	fpu_check
	beq	exit
;------------------------------------------------
;コマンドラインをコピーする
	pea.l	1(a2)
	jbsr	option_start
	addq.l	#4,sp
	beq	exit
;------------------------------------------------
;アボートの準備をする
	move.l	sp,abort_sp
	move.w	#_CTRLVC,-(sp)
	DOS	_INTVCG
	addq.l	#2,sp
	move.l	d0,abort_ctrlvc		;元の_CTRLVC
	move.w	#_ERRJVC,-(sp)
	DOS	_INTVCG
	addq.l	#2,sp
	move.l	d0,abort_errjvc		;元の_ERRJVC
	pea.l	abort
	move.w	#_CTRLVC,-(sp)
	DOS	_INTVCS
	addq.l	#6,sp
	pea.l	abort
	move.w	#_ERRJVC,-(sp)
	DOS	_INTVCS
	addq.l	#6,sp
;------------------------------------------------
;ベクタを変更する
	pea.l	trapv_routine		;TRAPV/TRAPcc/FTRAPccルーチン
	move.w	#7,-(sp)
	DOS	_INTVCS
	addq.l	#6,sp
	move.l	d0,trapv_vector		;元のTRAPV/TRAPcc/FTRAPccベクタ
;------------------------------------------------
;ロギングを開始する
	jbsr	logging_start
	beq	exit
;------------------------------------------------
;統計を開始する
	jbsr	statistics_start
;------------------------------------------------
;開始メッセージ
	putmsg	'begin: '
	putdate
	putcrlf
;------------------------------------------------
;圧縮された間接データを展開する
	jbsr	indirect_start
;------------------------------------------------
;全体ループ開始
loop::
;------------------------------------------------
;引数を確認する
	jbsr	option_loop
	beq	loopend
;------------------------------------------------
;テストを実行する
	jbsr	fabs_test
	jbsr	facos_test
	jbsr	fadd_test
	jbsr	fasin_test
	jbsr	fatan_test
	jbsr	fatanh_test
	jbsr	fbccl060_test
	jbsr	fbccl88x_test
	jbsr	fbccw060_test
	jbsr	fbccw88x_test
	jbsr	fcmp_test
	jbsr	fcos_test
	jbsr	fcosh_test
	jbsr	fdabs_test
	jbsr	fdadd_test
	jbsr	fdbcc060_test
	jbsr	fdbcc88x_test
	jbsr	fddiv_test
	jbsr	fdiv_test
	jbsr	fdmove_test
	jbsr	fdmul_test
	jbsr	fdneg_test
	jbsr	fdsqrt_test
	jbsr	fdsub_test
	jbsr	fetox_test
	jbsr	fetoxm1_test
	jbsr	fgetexp_test
	jbsr	fgetman_test
	jbsr	fint_test
	jbsr	fintrz_test
	jbsr	flog10_test
	jbsr	flog2_test
	jbsr	flogn_test
	jbsr	flognp1_test
	jbsr	fmod_test
	jbsr	fmovebregto_test
	jbsr	fmovebtoreg_test
	jbsr	fmovedregto_test
	jbsr	fmovedtoreg_test
	jbsr	fmovelregto_test
	jbsr	fmoveltoreg_test
	jbsr	fmovepregto_test
	jbsr	fmoveptoreg_test
	jbsr	fmovesregto_test
	jbsr	fmovestoreg_test
	jbsr	fmovewregto_test
	jbsr	fmovewtoreg_test
	jbsr	fmovexregto_test
	jbsr	fmovextoreg_test
	jbsr	fmovecr881_test
	jbsr	fmovecr882_test
	jbsr	fmul_test
	jbsr	fneg_test
	jbsr	frem_test
	jbsr	fsabs_test
	jbsr	fsadd_test
	jbsr	fscale_test
	jbsr	fscc060_test
	jbsr	fscc88x_test
	jbsr	fsdiv_test
	jbsr	fsgldiv060_test
	jbsr	fsgldiv88x_test
	jbsr	fsglmul060_test
	jbsr	fsglmul88x_test
	jbsr	fsin_test
	jbsr	fsincos_test
	jbsr	fsinh_test
	jbsr	fsmove_test
	jbsr	fsmul_test
	jbsr	fsneg_test
	jbsr	fsqrt_test
	jbsr	fssqrt_test
	jbsr	fssub_test
	jbsr	fsub_test
	jbsr	ftan_test
	jbsr	ftanh_test
	jbsr	ftentox_test
	jbsr	ftrapcc060_test
	jbsr	ftrapcc88x_test
	jbsr	ftrapccl060_test
	jbsr	ftrapccl88x_test
	jbsr	ftrapccw060_test
	jbsr	ftrapccw88x_test
	jbsr	ftst_test
	jbsr	ftwotox_test

;------------------------------------------------
;次の引数に進む
	jbsr	option_next
	jbra	loop

;------------------------------------------------
;全体ループ終了
loopend::
;------------------------------------------------
;アボート処理
abort::
	movea.l	abort_sp,sp
	move.l	abort_ctrlvc,-(sp)	;元の_CTRLVC
	move.w	#_CTRLVC,-(sp)
	DOS	_INTVCS
	addq.l	#6,sp
	move.l	abort_errjvc,-(sp)	;元の_ERRJVC
	move.w	#_ERRJVC,-(sp)
	DOS	_INTVCS
	addq.l	#6,sp
;------------------------------------------------
;ベクタを復元する
	move.l	trapv_vector,d0		;元のTRAPV/TRAPcc/FTRAPccベクタ
	beq	@f
	move.l	d0,-(sp)
	move.w	#7,-(sp)
	DOS	_INTVCS
	addq.l	#6,sp
@@:
;------------------------------------------------
;統計を終了する
	jbsr	statistics_end
;------------------------------------------------
;終了メッセージ
	putmsg	'end: '
	putdate
	putcrlf
	putcrlf
;------------------------------------------------
;ロギングを終了する
	jbsr	logging_end
;------------------------------------------------
;プログラムを終了する
exit::
	DOS	_EXIT

;------------------------------------------------
;アボート処理
	.data
	.align	4
abort_sp::		.dc.l	0	;アボートするときのスタックポインタ
abort_ctrlvc::		.dc.l	0	;元の_CTRLVC
abort_errjvc::		.dc.l	0	;元の_ERRJVC

;------------------------------------------------
;スタックエリア
	.bss
	.align	4
stack_area::	.ds.b	1024*64		;スタックエリア
stack_area_end::			;スタックエリアの末尾


;--------------------------------------------------------------------------------
;	TRAPV/TRAPcc/FTRAPcc
;--------------------------------------------------------------------------------

;TRAPV/TRAPcc/FTRAPccルーチン
	.text
	.even
trapv_routine::
	addq.l	#1,trapv_occurred	;TRAPV/TRAPcc/FTRAPccが発生した
	rte

	.data
	.align	4
trapv_vector::		.dc.l	0	;元のTRAPV/TRAPcc/FTRAPccベクタ
trapv_occurred::	.dc.l	0	;0=TRAPV/TRAPcc/FTRAPccは発生していない,1=TRAPV/TRAPcc/FTRAPccが発生した


;--------------------------------------------------------------------------------
;	FPUの種類
;--------------------------------------------------------------------------------

MC68881		equ	1<<0
MC68882		equ	1<<1
MC68040		equ	1<<2
FPSP040		equ	1<<3
MC68060		equ	1<<4
FPSP060		equ	1<<5

;----------------------------------------------------------------
;successful=fpu_check()
;	FPUの種類を確認する
;>d0.l:0=failed,1=successful
;------------------------------------------------
	.cpu	68030
	.text
	.even
fpu_check::
	movem.l	d1-d2/a1,-(sp)
	lea.l	$0CBD.w,a1		;FPUの有無
	IOCS	_B_BPEEK
	tst.b	d0
	beq	80f			;FPUなし
	lea.l	$0CBC.w,a1		;MPUの種類
	IOCS	_B_BPEEK
	move.l	#MC68060,d2
	cmp.b	#6,d0
	beq	@f
	move.l	#MC68040,d2
	cmp.b	#4,d0
	beq	@f
	move.l	#MC68882,d2
	fmove.l	#0,fpcr
	fmovecr.x	#1,fp0
	fmove.x	fp0,-(sp)
	move.l	(sp)+,d0
	or.l	(sp)+,d0
	or.l	(sp)+,d0
	bne	@f
	move.l	#MC68881,d2
@@:
	move.l	d2,fpu_type		;FPUの種類
	move.l	#-1,fpu_last		;前回のFPUの種類
	moveq.l	#1,d0
99:
	movem.l	(sp)+,d1-d2/a1
	rts

80:
	putmsg	'no floating point unit',13,10
	moveq.l	#0,d0
	bra	99b
	.cpu	68000

	.bss
	.align	4
fpu_type::		.ds.l	1	;FPUの種類
fpu_last::		.ds.l	1	;前回のFPUの種類


;--------------------------------------------------------------------------------
;	動作モード
;--------------------------------------------------------------------------------

OPTION_MARGIN_DEFAULT	equ	1	;超越関数の許容誤差
OPTION_MARGIN_LIMIT	equ	11
OPTION_MARGIN_MASK	equ	$F
OPTION_MAXIMUM_DEFAULT	equ	100	;テスト毎の出力される結果の最大数
OPTION_MAXIMUM_LIMIT	equ	1000000
OPTION_MAXIMUM_MASK	equ	$FFFFF
OPTION_MAXIMUM_SHIFT	equ	4
OPTION_DESTINATION	equ	1<<24	;デスティネーションオペランドを調べる
OPTION_FAILED		equ	1<<25	;失敗したテストの結果を出力する
OPTION_NANS		equ	1<<26	;NaNの仮数部を調べる
OPTION_FPSP		equ	1<<27	;浮動小数点ソフトウェアパッケージで処理される命令をテストする
OPTION_STATUS		equ	1<<28	;ステータスレジスタを調べる
OPTION_SUCCESSFUL	equ	1<<29	;成功したテストの結果を出力する
OPTION_DEFAULT		equ	OPTION_STATUS+OPTION_FAILED+OPTION_DESTINATION+(OPTION_MAXIMUM_DEFAULT<<OPTION_MAXIMUM_SHIFT)+OPTION_MARGIN_DEFAULT

;----------------------------------------------------------------
;successful=option_start(cmdl)
;	全体ループの開始前に呼び出す
;	コマンドラインをコピーする
;	引数がひとつもないとき使用法を表示する
;	全体ループの準備をする
;<4(sp).l:コマンドライン
;>d0.l:0=failed,1=successful
;------------------------------------------------
	.offsym	0,_a6
_size:
regs	reg	d1-d2/a0-a2
_regs:	.ds.l	.sizeof.(regs)
_a6:	.ds.l	1
_pc:	.ds.l	1
_cmdl:	.ds.l	1
;------------------------------------------------
	.text
	.even
option_start::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
;------------------------------------------------
;コマンドラインをコピーする
	movea.l	(_cmdl,a6),a0		;コマンドライン
	lea.l	option_args,a1		;単語列バッファ
	lea.l	option_argsend,a2	;単語列バッファの末尾
	bra	2f			;単語の間の空白を読み飛ばす

;単語が終わった
1:
	cmpa.l	a2,a1
	bcc	11f			;バッファが溢れた
	clr.b	(a1)+			;単語の末尾の0
;単語の間の空白を読み飛ばす
2:
	move.b	(a0)+,d0
	beq	10f			;単語が始まる前にコマンドラインが終わった
	cmp.b	#' ',d0
	bls	2b			;単語の間の空白を読み飛ばす
	cmp.b	#$22,d0
	beq	6f			;$22～$22の中の次の文字を読み出す
	cmp.b	#$27,d0
	beq	8f			;'～'の中の次の文字を読み出す
	cmp.b	#'a',d0
	blo	3f			;文字をバッファに書き込む
	cmp.b	#'z',d0
	bhi	3f			;文字をバッファに書き込む
	and.b	#$DF,d0			;大文字にする
;文字をバッファに書き込む
3:
	cmpa.l	a2,a1
	bcc	11f			;バッファが溢れた
	move.b	d0,(a1)+
;次の文字を読み出す
4:
	move.b	(a0)+,d0
	beq	9f			;単語の途中でコマンドラインが終わった
	cmp.b	#' ',d0
	bls	1b			;単語が終わった
	cmp.b	#$22,d0
	beq	6f			;$22～$22の中の次の文字を読み出す
	cmp.b	#$27,d0
	beq	8f			;'～'の中の次の文字を読み出す
	cmp.b	#'a',d0
	blo	3b			;文字をバッファに書き込む
	cmp.b	#'z',d0
	bhi	3b			;文字をバッファに書き込む
	and.b	#$DF,d0			;大文字にする
	bra	3b			;文字をバッファに書き込む

;$22～$22の中の文字をバッファに書き込む
5:
	cmpa.l	a2,a1
	bcc	11f			;バッファが溢れた
	move.b	d0,(a1)+
;$22～$22の中の次の文字を読み出す
6:
	move.b	(a0)+,d0
	beq	9f			;単語の途中でコマンドラインが終わった
	cmp.b	#$22,d0
	beq	4b			;次の文字を読み出す
	cmp.b	#'a',d0
	blo	5b			;$22～$22の中の文字をバッファに書き込む
	cmp.b	#'z',d0
	bhi	5b			;$22～$22の中の文字をバッファに書き込む
	and.b	#$DF,d0			;大文字にする
	bra	5b			;$22～$22の中の文字をバッファに書き込む

;'～'の中の文字をバッファに書き込む
7:
	cmpa.l	a2,a1
	bcc	11f			;バッファが溢れた
	move.b	d0,(a1)+
;'～'の中の次の文字を読み出す
8:
	move.b	(a0)+,d0
	beq	9f			;単語の途中でコマンドラインが終わった
	cmp.b	#$27,d0
	beq	4b			;次の文字を読み出す
	cmp.b	#'a',d0
	blo	7b			;'～'の中の文字をバッファに書き込む
	cmp.b	#'z',d0
	bhi	7b			;'～'の中の文字をバッファに書き込む
	and.b	#$DF,d0			;大文字にする
	bra	7b			;'～'の中の文字をバッファに書き込む

;単語の途中でコマンドラインが終わった
9:
	cmpa.l	a2,a1
	bcc	11f			;バッファが溢れた
	clr.b	(a1)+			;単語の末尾の0
;単語が始まる前にコマンドラインが終わった
10:
	cmpa.l	a2,a1
	bcc	11f			;バッファが溢れた
	clr.b	(a1)			;バッファの末尾の0
	bra	12f

;バッファが溢れた
11:
	putmsg	'too long command line',13,10
	moveq.l	#0,d0
	bra	99f

12:
;------------------------------------------------
;引数がひとつもないとき使用法を表示する
	tst.b	option_args
	bne	@f
	putstr	#option_usage
	moveq.l	#0,d0
	bra	99f

@@:
;------------------------------------------------
;全体ループの準備をする
	lea.l	option_args,a1		;単語列バッファ
	move.l	a1,option_word		;次の単語
	move.l	#OPTION_DEFAULT,option_mode	;動作モード
	move.l	#-1,option_lastmode	;前回の動作モード
	lea.l	option_donestart,a1	;実行済みのテストのIDと動作モードのリストの先頭
	move.l	a1,option_donepointer	;実行済みのテストのIDと動作モードのリストの末尾
;------------------------------------------------
98:
	moveq.l	#1,d0
99:
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

;------------------------------------------------
;使用法
	.text
option_usage::
	.dc.b	'usage: fputest.x <mnemonic or option> ...',13,10
	.dc.b	'  FABS FACOS FADD ... -- Run tests selected by the first few letters of a mnemonic.',13,10
	.dc.b	'  all -- Run all tests.',13,10
	.dc.b	'  destination=0..[1] -- Check the destination operand.',13,10
	.dc.b	'  easy -- It',39,'s the same as margin=1 nans=0.',13,10
	.dc.b	'  failed=0..[1] -- Output the results of failed tests.',13,10
	.dc.b	'  fpsp=[0]..1 -- Test the instructions processed by the software package.',13,10
	.dc.b	'  hard -- It',39,'s the same as margin=0 nans=1.',13,10
	.dc.b	'  logging=0..[1] -- Output the results to ',34,'fputest.log',34,'.',13,10
	.dc.b	'  margin=0..[1]..11 -- The acceptable margin of error in transcendental functions.',13,10
	.dc.b	'  maximum=0..[100]..1000000 -- The maximum number of results outputted for each test.',13,10
	.dc.b	'  nans=[0]..1 -- Check the mantissa of NaNs.',13,10
	.dc.b	'  status=0..[1] -- Check the status register.',13,10
	.dc.b	'  stdout=0..[1] -- Output the results to the standard output.',13,10
	.dc.b	'  successful=[0]..1 -- Output the results of successful tests.',13,10
	.dc.b	0

;----------------------------------------------------------------
;option_loop()
;	全体ループの先頭で呼び出す
;	次の単語がなければ全体ループを終了する
;	次の単語によって動作モードを変更する
	.text
	.even
option_loop::
	movea.l	option_word,a2		;次の単語
	move.l	option_mode,d2		;動作モード
	bra	20f

;次の単語に進む
10:
	tst.b	(a2)+
	bne	10b
;------------------------------------------------
;次の単語がなければ全体ループを終了する
20:
	tst.b	(a2)
	beq	80f			;全体ループを終了する
;------------------------------------------------
;次の単語によって動作モードを更新する
;--------------------------------
;destination=0..1
	leamsg	'DESTINATION',a0
	movea.l	a2,a1
	bsr	option_stringstartwith
	bne	2f
	lea.l	(.sizeof.('DESTINATION'),a2),a0
	moveq.l	#1,d0
	jbsr	option_equal_n
	bmi	2f
	and.l	#.not.(OPTION_DESTINATION),d2
	neg.l	d0
	and.l	#OPTION_DESTINATION,d0
	or.l	d0,d2			;destination
	bra	10b			;次の単語に進む

2:
;--------------------------------
;easy=1
	leamsg	'EASY',a0
	movea.l	a2,a1
	bsr	option_stringequals
	bne	2f
	and.l	#.not.(OPTION_NANS+OPTION_MARGIN_MASK),d2
	or.l	#1,d2			;margin=1 nans=0
	bra	10b			;次の単語に進む

2:
;--------------------------------
;failed=0..1
	leamsg	'FAILED',a0
	movea.l	a2,a1
	bsr	option_stringstartwith
	bne	2f
	lea.l	(.sizeof.('FAILED'),a2),a0
	moveq.l	#1,d0
	jbsr	option_equal_n
	bmi	2f
	and.l	#.not.(OPTION_FAILED),d2
	neg.l	d0
	and.l	#OPTION_FAILED,d0
	or.l	d0,d2			;failed
	bra	10b			;次の単語に進む

2:
;--------------------------------
;fpsp=0..1
	leamsg	'FPSP',a0
	movea.l	a2,a1
	bsr	option_stringstartwith
	bne	2f
	lea.l	(.sizeof.('FPSP'),a2),a0
	moveq.l	#1,d0
	jbsr	option_equal_n
	bmi	2f
	bne	1f
	and.l	#.not.(OPTION_FPSP),d2	;fpsp=0
	cmpi.l	#FPSP040,fpu_type
	bne	@f
	move.l	#MC68040,fpu_type
	bra	10b			;次の単語に進む

@@:
	cmpi.l	#FPSP060,fpu_type
	bne	10b			;次の単語に進む
	move.l	#MC68060,fpu_type
	bra	10b			;次の単語に進む

1:
	or.l	#OPTION_FPSP,d2		;fpsp=1
	cmpi.l	#MC68040,fpu_type
	bne	@f
	move.l	#FPSP040,fpu_type
	bra	10b			;次の単語に進む

@@:
	cmpi.l	#MC68060,fpu_type
	bne	10b			;次の単語に進む
	move.l	#FPSP060,fpu_type
	bra	10b			;次の単語に進む

2:
;--------------------------------
;hard=1
	leamsg	'HARD',a0
	movea.l	a2,a1
	bsr	option_stringequals
	bne	2f
	and.l	#.not.(OPTION_NANS+OPTION_MARGIN_MASK),d2
	or.l	#OPTION_NANS+0,d2	;margin=0 nans=1
	bra	10b			;次の単語に進む

2:
;--------------------------------
;logging=0..1
	leamsg	'LOGGING',a0
	movea.l	a2,a1
	bsr	option_stringstartwith
	bne	2f
	lea.l	(.sizeof.('LOGGING'),a2),a0
	moveq.l	#1,d0
	jbsr	option_equal_n
	bmi	2f
	sne.b	logging_logging
	bra	10b			;次の単語に進む

2:
;--------------------------------
;margin=0..OPTION_MARGIN_LIMIT
	leamsg	'MARGIN=',a0
	movea.l	a2,a1
	bsr	option_stringstartwith
	bne	2f
	lea.l	(.sizeof.('MARGIN'),a2),a0
	moveq.l	#OPTION_MARGIN_LIMIT,d0
	jbsr	option_equal_n
	bmi	2f
	and.l	#.not.(OPTION_MARGIN_MASK),d2
	or.l	d0,d2			;margin
	bra	10b			;次の単語に進む

2:
;--------------------------------
;maxmum=0..OPTION_MAXIMUM_LIMIT
	leamsg	'MAXIMUM=',a0
	movea.l	a2,a1
	bsr	option_stringstartwith
	bne	2f
	lea.l	(.sizeof.('MAXIMUM'),a2),a0
	move.l	#OPTION_MAXIMUM_LIMIT,d0
	jbsr	option_equal_n
	bmi	2f
	and.l	#.not.(OPTION_MAXIMUM_MASK<<OPTION_MAXIMUM_SHIFT),d2
	lsl.l	#OPTION_MAXIMUM_SHIFT,d0
	or.l	d0,d2			;maximum
	bra	10b			;次の単語に進む

2:
;--------------------------------
;nans=0..1
	leamsg	'NANS',a0
	movea.l	a2,a1
	bsr	option_stringstartwith
	bne	2f
	lea.l	(.sizeof.('NANS'),a2),a0
	moveq.l	#1,d0
	jbsr	option_equal_n
	bmi	2f
	and.l	#.not.(OPTION_NANS),d2
	neg.l	d0
	and.l	#OPTION_NANS,d0
	or.l	d0,d2			;nans
	bra	10b			;次の単語に進む

2:
;--------------------------------
;successful=0..1
	leamsg	'SUCCESSFUL',a0
	movea.l	a2,a1
	bsr	option_stringstartwith
	bne	2f
	lea.l	(.sizeof.('SUCCESSFUL'),a2),a0
	moveq.l	#1,d0
	jbsr	option_equal_n
	bmi	2f
	and.l	#.not.(OPTION_SUCCESSFUL),d2
	neg.l	d0
	and.l	#OPTION_SUCCESSFUL,d0
	or.l	d0,d2			;successful
	bra	10b			;次の単語に進む

2:
;--------------------------------
;status=0..1
	leamsg	'STATUS',a0
	movea.l	a2,a1
	bsr	option_stringstartwith
	bne	2f
	lea.l	(.sizeof.('STATUS'),a2),a0
	moveq.l	#1,d0
	jbsr	option_equal_n
	bmi	2f
	and.l	#.not.(OPTION_STATUS),d2
	neg.l	d0
	and.l	#OPTION_STATUS,d0
	or.l	d0,d2			;status
	bra	10b			;次の単語に進む

2:
;--------------------------------
;stdout=0..1
	leamsg	'STDOUT',a0
	movea.l	a2,a1
	bsr	option_stringstartwith
	bne	2f
	lea.l	(.sizeof.('STDOUT'),a2),a0
	moveq.l	#1,d0
	jbsr	option_equal_n
	bmi	2f
	sne.b	logging_stdout
	bra	10b			;次の単語に進む

2:
;--------------------------------
	move.l	d2,option_mode		;動作モード
	move.l	a2,option_word		;次の単語
;
	bra	98f

;全体ループを終了する
80:
	moveq.l	#0,d0
	bra	99f

98:
	moveq.l	#1,d0
99:
	rts

;----------------------------------------------------------------
;mnemonic_start(mnemonic,fpu)
;	個々のテストの開始前に呼び出す
;	ニモニックを確認する
;	FPUを確認する
;	実行済みかどうか確認する
;	ニモニックのカウンタをクリアする
;	初回または動作モードが変更されたとき動作モードを表示する
;	バックグラウンドスレッドを回す
;<(sp).l:復帰アドレス。テストのIDとして使う
;<4(sp).l:テストのニモニックの条件
;<8(sp).l:テストのFPUの条件
;>d0.l:0=失敗,1=成功
;>eq=失敗,ne=成功
;------------------------------------------------
	.offsym	0,_a6
_size:
regs	reg	d1-d2/a0-a2
_regs:	.ds.l	.sizeof.(regs)
_a6:	.ds.l	1
_pc:	.ds.l	1
_mnem:	.ds.l	1
_fpu:	.ds.l	1
;------------------------------------------------
	.text
	.even
mnemonic_start::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
;------------------------------------------------
;ニモニックを確認する
	leamsg	'ALL',a0
	movea.l	option_word,a1		;次の単語
	bsr	option_stringequals
	beq	@f			;allはすべてのニモニックと一致する
	movea.l	option_word,a0		;次の単語
	movea.l	(_mnem,a6),a1		;テストのニモニックの条件
	bsr	option_stringstartwith
	bne	97f			;ニモニックが一致しなかった。失敗
@@:
;------------------------------------------------
;FPUを確認する
	move.l	(_fpu,a6),d0		;テストのFPUの条件
	and.l	fpu_type,d0	;FPUの種類
	beq	97f			;FPUが一致しなかった。失敗
;------------------------------------------------
;実行済みかどうか確認する
	move.l	(_pc,a6),d0		;テストのID
	move.l	option_mode,d1		;動作モード
	lea.l	option_donestart,a0	;実行済みのテストのIDと動作モードのリストの先頭
	movea.l	option_donepointer,a1	;実行済みのテストのIDと動作モードのリストの末尾
	bra	3f

1:
	cmp.l	(a0),d0			;テストのID
	bne	2f
	cmp.l	4(a0),d1		;動作モード
	beq	97f			;実行済みだった。失敗
2:
	addq.l	#8,a0
3:
	cmpa.l	a1,a0
	blo	1b
;実行済みではなかった
	move.l	d0,(a1)+		;実行済みのテストのID
	move.l	d1,(a1)+		;実行済みの動作モード
	move.l	a1,option_donepointer
;------------------------------------------------
;ニモニックのカウンタをクリアする
	clr.l	mnemonic_counter
	clr.l	statistics_tested
	clr.l	statistics_failed
;------------------------------------------------
;初回はFPUの種類を表示する
	move.l	fpu_type,d2	;FPUの種類
	cmp.l	fpu_last,d2	;前回のFPUの種類
	beq	3f
	move.l	d2,fpu_last
	putmsg	'fpu: '
;
	move.l	#MC68881,d0
	and.l	d2,d0
	beq	1f
	leamsg	'MC68881',a0
	bra	2f

1:
	move.l	#MC68882,d0
	and.l	d2,d0
	beq	1f
	leamsg	'MC68882',a0
	bra	2f

1:
	move.l	#MC68040,d0
	and.l	d2,d0
	beq	1f
	leamsg	'MC68040',a0
	bra	2f

1:
	move.l	#FPSP040,d0
	and.l	d2,d0
	beq	1f
	leamsg	'040FPSP',a0
	bra	2f

1:
	move.l	#MC68060,d0
	and.l	d2,d0
	beq	1f
	leamsg	'MC68060',a0
	bra	2f

1:
	move.l	#FPSP060,d0
	and.l	d2,d0
	beq	1f
	leamsg	'060FPSP',a0
	bra	2f

1:
	leamsg	'???',a0
2:
	putstr	a0
	putcrlf
3:
;------------------------------------------------
;初回または動作モードが変更されたとき動作モードを表示する
	move.l	option_mode,d2		;動作モード
	cmp.l	option_lastmode,d2	;前回の動作モード
	beq	3f
	move.l	d2,option_lastmode
	putmsg	'option:'
;--------------------------------
	putmsg	' destination'
	move.l	d2,d0
	and.l	#OPTION_DESTINATION,d0
	bne	@f
	putmsg	'=0'
@@:
;--------------------------------
	move.l	d2,d0
	and.l	#OPTION_NANS+OPTION_MARGIN_MASK,d0
	cmp.l	#1,d0			;margin=1 nans=0
	bne	@f
	putmsg	' easy'
@@:
;--------------------------------
	putmsg	' failed'
	move.l	d2,d0
	and.l	#OPTION_FAILED,d0
	bne	@f
	putmsg	'=0'
@@:
;--------------------------------
	putmsg	' fpsp'
	move.l	d2,d0
	and.l	#OPTION_FPSP,d0
	bne	@f
	putmsg	'=0'
@@:
;--------------------------------
	move.l	d2,d0
	and.l	#OPTION_NANS+OPTION_MARGIN_MASK,d0
	cmp.l	#OPTION_NANS+0,d0	;margin=0 nans=1
	bne	@f
	putmsg	' hard'
@@:
;--------------------------------
	putmsg	' margin='
	move.l	d2,d0
	and.l	#OPTION_MARGIN_MASK,d0
	putdec	d0
;--------------------------------
	putmsg	' maximum='
	move.l	d2,d0
	lsr.l	#OPTION_MAXIMUM_SHIFT,d0
	and.l	#OPTION_MAXIMUM_MASK,d0
	putdec	d0
;--------------------------------
	putmsg	' nans'
	move.l	d2,d0
	and.l	#OPTION_NANS,d0
	bne	@f
	putmsg	'=0'
@@:
;--------------------------------
	putmsg	' status'
	move.l	d2,d0
	and.l	#OPTION_STATUS,d0
	bne	@f
	putmsg	'=0'
@@:
;--------------------------------
	putmsg	' successful'
	move.l	d2,d0
	and.l	#OPTION_SUCCESSFUL,d0
	bne	@f
	putmsg	'=0'
@@:
;------------------------------------------------
	putcrlf
3:
;------------------------------------------------
;バックグラウンドスレッドを回す
	move.l	fpu_type,d0	;FPUの種類
	and.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,d0
	beq	@f
	.cpu	68030
	fnop
	.cpu	68000
@@:
	DOS	_CHANGE_PR
;------------------------------------------------
;成功
98:
	moveq.l	#1,d0
99:
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

;失敗
97:
	moveq.l	#0,d0
	bra	99b

	.data
	.align	4
mnemonic_counter::	.dc.l	1	;ニモニックのカウンタ

;----------------------------------------------------------------
;mnemonic_end()
;	ニモニックの成績を表示する
;	ニモニックの成績を総合成績に加える
	.text
	.even
mnemonic_end::
	movem.l	d0-d1,-(sp)
	move.l	statistics_tested,d0
	move.l	statistics_failed,d1
	putmsg	'score: '
	movem.l	d0-d1,-(sp)
	jbsr	statistics_output
	addq.l	#8,sp
	add.l	d0,statistics_ttl_tested
	add.l	d1,statistics_ttl_failed
	movem.l	(sp)+,d0-d1
	rts

;----------------------------------------------------------------
;option_next()
;	全体ループの末尾で呼び出す
;	次の単語に進む
	.text
	.even
option_next::
	move.l	a0,-(sp)
;------------------------------------------------
;次の単語に進む
	movea.l	option_word,a0		;次の単語
@@:
	tst.b	(a0)+
	bne	@b
	move.l	a0,option_word		;次の単語
;------------------------------------------------
	movea.l	(sp)+,a0
	rts

;----------------------------------------------------------------
;	文字列比較
;<a0.l:文字列
;<a1.l:文字列
;>eq=一致,ne=不一致
	.text
	.even
option_stringequals::
1:
	tst.b	(a0)
	beq	2f
	cmpm.b	(a1)+,(a0)+
	beq	1b
	rts

2:
	tst.b	(a1)
	rts

;----------------------------------------------------------------
;	先頭文字列比較
;<a0.l:先頭文字列
;<a1.l:文字列
;>eq=一致,ne=不一致
	.text
	.even
option_stringstartwith::
1:
	tst.b	(a0)
	beq	2f
	cmpm.b	(a1)+,(a0)+
	beq	1b
2:
	rts

;----------------------------------------------------------------
;	'=N'を読み取る
;<d0.l:上限
;<a0.l:'='の位置
;>d0.l:N。'='がないとき1。エラーのとき-1
	.text
	.even
option_equal_n::
	movem.l	d1-d2/a0,-(sp)
	move.l	d0,d2
	moveq.l	#1,d0
	tst.b	(a0)
	beq	9f			;最初に文字がない
	cmpi.b	#'=',(a0)+
	bne	8f			;最初の文字が'='でない
	move.b	(a0)+,d0
	beq	8f			;'='の後に文字がない
	sub.b	#'0',d0
	blo	8f			;'='の後の文字が数字でない
	cmp.b	#9,d0
	bhi	8f			;'='の後の文字が数字でない
1:
	moveq.l	#0,d1
	move.b	(a0)+,d1
	beq	2f			;次の文字がない
	sub.b	#'0',d1
	blo	8f			;次の文字が数字でない
	cmp.b	#9,d1
	bhi	8f			;次の文字が数字でない
	add.l	d0,d0
	add.l	d0,d1
	lsl.l	#2,d0
	add.l	d1,d0
	bra	1b

2:
	cmp.l	d2,d0
	bls	9f
8:
	moveq.l	#-1,d0
9:
	movem.l	(sp)+,d1-d2/a0
	tst.l	d0
	rts

;----------------------------------------------------------------

	.bss
option_args::		.ds.b	4096	;単語列バッファ。単語,0,単語,0,…,単語,0,0
option_argsend::

	.bss
	.align	4
option_word::		.ds.l	1	;次の単語
option_mode::		.ds.l	1	;動作モード
option_lastmode::	.ds.l	1	;動作モード

	.bss
	.align	4
option_donepointer::	.ds.l	1	;実行済みのテストのIDと動作モードの書き込み位置
option_donestart::	.ds.l	2*1024*32	;実行済みのテストのIDと動作モードのリスト
option_doneend::


;--------------------------------------------------------------------------------
;	logging
;--------------------------------------------------------------------------------

;------------------------------------------------
;logging_start()
;>d0.l:0=failed,1=successful
	.text
	.even
logging_start::
	move.l	d1,-(sp)
;
	st.b	logging_logging
	st.b	logging_stdout
;open log file
	move.w	#2,-(sp)		;read and write
	peamsg	'fputest.log'		;log file name
	DOS	_OPEN
	addq.l	#6,sp
	tst.l	d0
	bpl	@f
	move.w	#$0020,-(sp)		;file
	peamsg	'fputest.log'		;log file name
	DOS	_CREATE
	addq.l	#6,sp
	tst.l	d0
	bpl	@f
	putmsg	'cannot open fputest.log to write'
	moveq.l	#0,d0
	bra	99f

@@:
	move.w	d0,d1			;log file handle
	move.w	d1,logging_file_handle
;check device type
	move.w	d1,-(sp)		;log file handle
	clr.w	-(sp)			;device information
	DOS	_IOCTRL
	addq.l	#4,sp
	tst.b	d0
	bmi	1f			;character device
;block device or remote device
;seek end
	move.w	#2,-(sp)		;SEEK_END
	clr.l	-(sp)
	move.w	d1,-(sp)		;log file handle
	DOS	_SEEK
	addq.l	#8,sp
	bra	2f

;character device
1:
;check output status
	move.w	d1,-(sp)		;log file handle
	move.w	#7,-(sp)		;output status
	DOS	_IOCTRL
	addq.l	#4,sp
	tst.l	d0
	bne	@f
	putmsg	'cannot output to fputest.log'
	moveq.l	#0,d0
	bra	99f

@@:
2:
	move.l	#logging_cache_start,logging_cache_pointer
;
98:
	moveq.l	#1,d0
99:
	move.l	(sp)+,d1
	tst.l	d0
	rts

;------------------------------------------------
;logging_write(buffer,length)
;<4(sp).l:buffer
;<8(sp).l:length
regs		reg	d0-d3/a0-a3
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_a6:	.ds.l	1
_pc:	.ds.l	1
_buff:	.ds.l	1
_leng:	.ds.l	1
	.text
	.even
logging_write::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
;ログファイルに出力する
	tst.b	logging_logging
	beq	19f
	move.w	logging_file_handle,d1	;log file handle
	bmi	19f
;出力データをキャッシュにコピーする
	move.l	(_leng,a6),d2		;出力データの残りのバイト数
	movea.l	(_buff,a6),a0		;出力データの読み出し位置
	movea.l	logging_cache_pointer,a1	;キャッシュの書き込み位置
	lea.l	logging_cache_end,a3	;キャッシュの終了アドレス。固定
	bra	17f

10:
	move.l	a3,d3			;キャッシュの終了アドレス
	sub.l	a1,d3			;キャッシュの残りのバイト数。キャッシュの終了アドレス-キャッシュの書き込み位置
	cmp.l	d2,d3			;キャッシュの残りのバイト数<=>出力データの残りのバイト数
	blo	11f
	move.l	d2,d3			;出力データの残りのバイト数
11:
;<d3.l:出力データをキャッシュにコピーするバイト数。出力データの残りのバイト数とキャッシュの残りのバイト数の少ない方。0ではない
	move.l	d3,d0			;出力データをキャッシュにコピーするバイト数
	subq.l	#1,d0
	swap.w	d0
12:
	swap.w	d0
13:
	move.b	(a0)+,(a1)+		;出力データの読み出し位置からキャッシュの書き込み位置へ
	dbra	d0,13b
	swap.w	d0
	dbra	d0,12b
	cmpa.l	a3,a1			;キャッシュの書き込み位置<=>キャッシュの終了アドレス
	blo	14f
;キャッシュが一杯になった
	lea.l	logging_cache_start,a1	;キャッシュの開始アドレス
	movea.l	a3,a2			;キャッシュの終了アドレス
	suba.l	a1,a2			;キャッシュの全体のバイト数
	movem.l	a1-a2,-(sp)
	move.w	d1,-(sp)		;log file handle
	DOS	_WRITE
	lea.l	10(sp),sp
14:
	move.l	a1,logging_cache_pointer	;キャッシュの書き込み位置
	sub.l	d3,d2			;出力データの残りのバイト数
17:
	tst.l	d2			;出力データの残りのバイト数
	bne	10b
19:
;標準出力に出力する
	tst.b	logging_stdout
	beq	29f
	move.l	(_leng,a6),-(sp)	;length
	move.l	(_buff,a6),-(sp)	;buffer
	move.w	#1,-(sp)		;stdout
	DOS	_WRITE
	lea.l	10(sp),sp
29:
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

;------------------------------------------------
;logging_end()
logging_end::
	movem.l	d0-d1/a1-a2,-(sp)
;close log file
	move.w	logging_file_handle,d1	;log file handle
	bmi	19f
	lea.l	logging_cache_start,a1	;キャッシュの開始アドレス
	movea.l	logging_cache_pointer,a2	;キャッシュの書き込み位置
	suba.l	a1,a2			;キャッシュの現在のバイト数
	move.l	a2,d0
	beq	11f
	movem.l	a1-a2,-(sp)
	move.w	d1,-(sp)		;log file handle
	DOS	_WRITE
	lea.l	10(sp),sp
11:
	move.l	a1,logging_cache_pointer	;キャッシュの書き込み位置
	move.w	d1,-(sp)		;log file handle
	DOS	_CLOSE
	addq.l	#2,sp
	move.w	#-1,logging_file_handle
19:
;
	sf.b	logging_logging
	st.b	logging_stdout
;
	movem.l	(sp)+,d0-d1/a1-a2
	rts

;------------------------------------------------

	.data
logging_logging::	.dc.b	0	;ファイルに出力する
logging_stdout::	.dc.b	-1	;標準出力に出力する
	.even
logging_file_handle::	.dc.w	-1	;log file handle

	.bss
	.align	4
logging_cache_pointer::	.ds.l	1	;キャッシュの書き込み位置
logging_cache_start::	.ds.b	1024*64	;キャッシュ
logging_cache_end::


;--------------------------------------------------------------------------------
;	statistics
;--------------------------------------------------------------------------------

;--------------------------------------------------------------------------------
	.text
	.even
statistics_start::
	clr.l	statistics_ttl_tested
	clr.l	statistics_ttl_failed
	rts

;--------------------------------------------------------------------------------
;statistics_update()
;<(4,sp).l:0=failed,1=successful
;>d0.l:0=not output,1=output
	.text
	.even
statistics_update::
	addq.l	#1,statistics_tested
	tst.l	(4,sp)
	bne	1f			;successful
;failed
	addq.l	#1,statistics_failed
	move.l	#OPTION_FAILED,d0
	bra	2f

;successful
1:
	move.l	#OPTION_SUCCESSFUL,d0
2:
	and.l	option_mode,d0
	beq	8f			;not output
	addq.l	#1,mnemonic_counter
	move.l	option_mode,d0
	lsr.l	#OPTION_MAXIMUM_SHIFT,d0
	and.l	#OPTION_MAXIMUM_MASK,d0
	cmp.l	mnemonic_counter,d0
	blo	8f			;not output
;output
	move.l	mnemonic_counter,d0
	putdec	d0
	putmsg	': ',
	moveq.l	#1,d0
9:
	rts

;not output
8:
	moveq.l	#0,d0
	bra	9b

;--------------------------------------------------------------------------------
	.text
	.even
statistics_end::
	putmsg	'total: '
	move.l	statistics_ttl_failed,-(sp)
	move.l	statistics_ttl_tested,-(sp)
	jbsr	statistics_output
	addq.l	#8,sp
	rts

;--------------------------------------------------------------------------------
;<(4,sp).l:tested
;<(8,sp).l:failed
	.text
	.even
statistics_output::
	movem.l	d0-d6,-(sp)
	move.l	(4*7+4,sp),d6		;d6=tested
	move.l	(4*7+8,sp),d4		;d4=failed
	putmsg	'tested='
	move.l	d6,d5			;tested
	sub.l	d4,d5			;d5=successful=tested-failed
	putdec	d6			;tested
	tst.l	d6			;tested
	beq	8f			;no tests were performed
	move.l	#10000,d0
	move.l	d5,d1			;successful
	jbsr	mull			;d0:d1=10000*successful
	moveq.l	#0,d2
	move.l	d6,d3			;d2:d3=tested
	jbsr	divq			;d1=10000*successful/tested
	move.l	#10000,d0
	sub.l	d1,d0			;d0=10000-10000*successful/tested
	putmsg	' failed='
	putdec	d4			;failed
	putchr	#'('
	putfix	d0,#2			;10000-10000*successful/tested
	putmsg	'%) successful='
	putdec	d5			;successful
	putchr	#'('
	putfix	d1,#2			;10000*successful/tested
	putmsg	'%)'
8:
	putcrlf
	movem.l	(sp)+,d0-d6
	rts

	.bss

	.align	4
statistics_tested::	.ds.l	1	;number of tests
statistics_failed::	.ds.l	1	;number of failed tests
statistics_ttl_tested::	.ds.l	1	;number of total tests
statistics_ttl_failed::	.ds.l	1	;number of total failed tests


;--------------------------------------------------------------------------------
;	結果を比較する
;--------------------------------------------------------------------------------

;--------------------------------------------------------------------------------
;successful=test_status()
;	statusを比較する
;<(4,sp).l:actual status
;<(8,sp).l:expected status
;>d0.l:0=failed,1=successful
	.cpu	68030
	.offsym	0,_a6
_size:
_a6:	.ds.l	1
_pc:	.ds.l	1
_asta:	.ds.l	1	;actual status
_esta:	.ds.l	1	;expected status
	.text
	.even
test_status::
	link.w	a6,#_size
;------------------------------------------------
;status
	move.l	option_mode,d0
	and.l	#OPTION_STATUS,d0
	beq	19f			;statusをテストしない。statusが一致
;statusをテストする
	move.l	(_asta,a6),d0		;actual status
	cmp.l	(_esta,a6),d0		;expected status
	bne	97f			;statusが一致しない。失敗
;statusが一致
19:
;------------------------------------------------
;成功
98:
	moveq.l	#1,d0
99:
	unlk	a6
	rts

;失敗
97:
	moveq.l	#0,d0
	bra	99b
	.cpu	68000

;--------------------------------------------------------------------------------
;successful=test_single()
;	singleの結果を比較する。誤差は許容しない
;<(4,sp).s:actual result
;<(8,sp).l:actual status
;<(12,sp).s:expected result
;<(16,sp).l:expected status
;>d0.l:0=failed,1=successful
	.cpu	68030
	.offsym	0,_a6
_size:
_a6:	.ds.l	1
_pc:	.ds.l	1
_ares:	.ds.s	1	;actual result
_asta:	.ds.l	1	;actual status
_eres:	.ds.s	1	;expected result
_esta:	.ds.l	1	;expected status
	.text
	.even
test_single::
	link.w	a6,#_size
;------------------------------------------------
;status
	move.l	option_mode,d0
	and.l	#OPTION_STATUS,d0
	beq	19f			;statusをテストしない。statusが一致
;statusをテストする
	move.l	(_asta,a6),d0		;actual status
	cmp.l	(_esta,a6),d0		;expected status
	bne	97f			;statusが一致しない。失敗
;statusが一致
19:
;------------------------------------------------
;result
	move.l	option_mode,d0
	and.l	#OPTION_DESTINATION,d0
	beq	39f			;resultをテストしない。resultが一致
;resultをテストする
	move.l	(_ares,a6),d0		;actual result
	cmp.l	(_eres,a6),d0		;expected result
	bne	97f			;resultが一致しない。失敗
;resultが一致
39:
;------------------------------------------------
;成功
98:
	moveq.l	#1,d0
99:
	unlk	a6
	rts

;失敗
97:
	moveq.l	#0,d0
	bra	99b
	.cpu	68000

;--------------------------------------------------------------------------------
;successful=test_double()
;	doubleの結果を比較する。誤差は許容しない
;<(4,sp).d:actual result
;<(12,sp).l:actual status
;<(16,sp).d:expected result
;<(24,sp).l:expected status
;>d0.l:0=failed,1=successful
	.cpu	68030
	.offsym	0,_a6
_size:
_a6:	.ds.l	1
_pc:	.ds.l	1
_ares:	.ds.d	1	;actual result
_asta:	.ds.l	1	;actual status
_eres:	.ds.d	1	;expected result
_esta:	.ds.l	1	;expected status
	.text
	.even
test_double::
	link.w	a6,#_size
;------------------------------------------------
;status
	move.l	option_mode,d0
	and.l	#OPTION_STATUS,d0
	beq	19f			;statusをテストしない。statusが一致
;statusをテストする
	move.l	(_asta,a6),d0		;actual status
	cmp.l	(_esta,a6),d0		;expected status
	bne	97f			;statusが一致しない。失敗
;statusが一致
19:
;------------------------------------------------
;result
	move.l	option_mode,d0
	and.l	#OPTION_DESTINATION,d0
	beq	39f			;resultをテストしない。resultが一致
;resultをテストする
	move.l	(_ares,a6),d0		;actual result 1st
	cmp.l	(_eres,a6),d0		;expected result 1st
	bne	97f			;resultが一致しない。失敗
	move.l	(4+_ares,a6),d0		;actual result 2nd
	cmp.l	(4+_eres,a6),d0		;expected result 2nd
	bne	97f			;resultが一致しない。失敗
;resultが一致
39:
;------------------------------------------------
;成功
98:
	moveq.l	#1,d0
99:
	unlk	a6
	rts

;失敗
97:
	moveq.l	#0,d0
	bra	99b
	.cpu	68000

;--------------------------------------------------------------------------------
;successful=test_packed()
;	packedの結果を比較する。誤差は許容しない
;<(4,sp).p:actual result
;<(12,sp).l:actual status
;<(16,sp).p:expected result
;<(24,sp).l:expected status
;>d0.l:0=failed,1=successful
	.cpu	68030
	.offsym	0,_a6
_size:
_a6:	.ds.l	1
_pc:	.ds.l	1
_ares:	.ds.p	1	;actual result
_asta:	.ds.l	1	;actual status
_eres:	.ds.p	1	;expected result
_esta:	.ds.l	1	;expected status
	.text
	.even
test_packed::
	link.w	a6,#_size
;------------------------------------------------
;status
	move.l	option_mode,d0
	and.l	#OPTION_STATUS,d0
	beq	19f			;statusをテストしない。statusが一致
;statusをテストする
	move.l	(_asta,a6),d0		;actual status
	cmp.l	(_esta,a6),d0		;expected status
	bne	97f			;statusが一致しない。失敗
;statusが一致
19:
;------------------------------------------------
;result
	move.l	option_mode,d0
	and.l	#OPTION_DESTINATION,d0
	beq	39f			;resultをテストしない。resultが一致
;resultをテストする
	move.l	(_ares,a6),d0		;actual result 1st
	cmp.l	(_eres,a6),d0		;expected result 1st
	bne	97f			;resultが一致しない。失敗
	move.l	(4+_ares,a6),d0		;actual result 2nd
	cmp.l	(4+_eres,a6),d0		;expected result 2nd
	bne	97f			;resultが一致しない。失敗
	move.l	(8+_ares,a6),d0		;actual result 3rd
	cmp.l	(8+_eres,a6),d0		;expected result 3rd
	bne	97f			;resultが一致しない。失敗
;resultが一致
39:
;------------------------------------------------
;成功
98:
	moveq.l	#1,d0
99:
	unlk	a6
	rts

;失敗
97:
	moveq.l	#0,d0
	bra	99b
	.cpu	68000

;--------------------------------------------------------------------------------
;successful=test_extended()
;	extendedの結果を比較する。誤差を許容する
;<(4,sp).x:actual result
;<(16,sp).l:actual status
;<(20,sp).x:expected result
;<(32,sp).l:expected status
;<(36,sp).l:fpcr(rp<<6,-1=strict)
;>d0.l:0=failed,1=successful
	.cpu	68030
regs		reg	d1-d6
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp2
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
_ares:	.ds.x	1	;actual result
_asta:	.ds.l	1	;actual status
_eres:	.ds.x	1	;expected result
_esta:	.ds.l	1	;expected status
_fpcr:	.ds.l	1	;fpcr(rp<<6,-1=strict)
	.text
	.even
test_extended::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;----------------------------------------------------------------
;status
	move.l	option_mode,d0
	and.l	#OPTION_STATUS,d0
	beq	19f			;statusをテストしない。statusが一致
;statusをテストする
	move.l	(_asta,a6),d0		;actual status
	cmp.l	(_esta,a6),d0		;expected status
	bne	97f			;statusが一致しない。失敗
;statusが一致
19:
;----------------------------------------------------------------
;result
	move.l	option_mode,d0
	and.l	#OPTION_DESTINATION,d0
	beq	39f			;resultをテストしない。resultが一致
;resultをテストする
	fmove.l	#0,fpcr
;------------------------------------------------
;expected resultが正規化数または非正規化数か調べる
	move.l	(_eres,a6),d0		;expected result 1st
	and.l	#$7FFF0000,d0
	cmp.l	#$7FFF0000,d0
	beq	30f			;expected resultが±InfまたはNaN。完全に一致しなければならない
	or.l	(4+_eres,a6),d0		;expected result 2nd
	or.l	(8+_eres,a6),d0		;expected result 3rd
	beq	30f			;expected resultが±0。完全に一致しなければならない
;expected resultは正規化数または非正規化数
;------------------------------------------------
;1ulpを求める
	move.l	(_fpcr,a6),d0		;fpcr(rp<<6,-1=strict)
	cmp.l	#-1,d0
	beq	30f			;strict。完全に一致しなければならない
	move.l	#$00000000,d4		;extendedの1ulp
	move.l	#$00000001,d5
	lsr.b	#6,d0			;rounding precision
	beq	49f			;extended
;singleまたはdouble
	subq.b	#2,d0
	bcc	43f			;double
;--------------------------------
;single
	move.l	#$00000100,d4		;singleの1ulp
	move.l	#$00000000,d5
	movem.l	(_ares,a6),d1-d3	;actual result
	and.l	#$7FFF0000,d1
	cmp.l	#$7FFF0000,d1
	beq	49f			;actual resultが±InfまたはNaN
;actual resultが±0または非正規化数または正規化数
	move.l	d2,d0
	or.l	d3,d0
	beq	49f			;actual resultが±0
;actual resultが非正規化数または正規化数
;actual resultがsingleに丸められていることを確認する
	tst.l	d2
	bmi	42f			;actual resultが正規化数
;actual resultが非正規化数
41:
	add.l	d3,d3
	addx.l	d2,d2
	bpl	41b			;actual resultが非正規化数
;actual resultが正規化数
42:
	and.l	#$000000FF,d2
	or.l	d3,d2
	bne	97f			;actual resultがsingleに丸められていない。失敗
	bra	49f

;--------------------------------
;double
43:
	move.l	#$00000000,d4		;doubleの1ulp
	move.l	#$00000800,d5
	movem.l	(_ares,a6),d1-d3	;actual result
	and.l	#$7FFF0000,d1
	cmp.l	#$7FFF0000,d1
	beq	49f			;actual resultが±InfまたはNaN
;actual resultが±0または非正規化数または正規化数
	move.l	d2,d0
	or.l	d3,d0
	beq	49f			;actual resultが±0
;actual resultが非正規化数または正規化数
;actual resultがdoubleに丸められていることを確認する
	tst.l	d2
	bmi	45f			;actual resultが正規化数
;actual resultが非正規化数
44:
	add.l	d3,d3
	addx.l	d2,d2
	bpl	44b			;actual resultが非正規化数
;actual resultが正規化数
45:
	and.l	#$000007FF,d3
	bne	97f			;actual resultがdoubleに丸められていない。失敗
49:
;<d4-d5:1ulp
;------------------------------------------------
;マージンを求める
	move.l	option_mode,d0
	and.l	#OPTION_MARGIN_MASK,d0	;margin
	bne	1f
	moveq.l	#0,d4
	moveq.l	#0,d5
	bra	2f

1:
	subq.w	#1,d0
	lsl.l	d0,d4			;extended,single,doubleのいずれも下位ワードが1bit立っているだけなのでこれだけでよい
	lsl.l	d0,d5
2:
;<d4-d5:マージン
;------------------------------------------------
;許容範囲の絶対値の下限を求める
	movem.l	(_eres,a6),d1-d3	;expected result
	tst.l	d2
	bpl	22f			;非正規化数
;正規化数
	sub.l	d5,d3
	subx.l	d4,d2
	bmi	23f			;指数部が変わらない
	move.l	d1,d0
	and.l	#$7FFF0000,d0
	beq	23f			;指数部が0のときは正規化数が非正規化数になる
	sub.l	#$00010000,d1		;指数部を1減らす
	add.l	d3,d3			;仮数部を左にずらす
	addx.l	d2,d2
	bra	23f

;非正規化数
22:
	sub.l	d5,d3
	subx.l	d4,d2
	bpl	23f
	and.l	#$80000000,d1		;引けなかったとき許容範囲の絶対値の下限は0
	move.l	#$00000000,d2
	move.l	#$00000000,d3
23:
	movem.l	d1-d3,-(sp)
	fmove.x	(sp)+,fp1		;許容範囲の絶対値の下限
;<fp1.x:許容範囲の絶対値の下限
;------------------------------------------------
;許容範囲の絶対値の上限を求める
	movem.l	(_eres,a6),d1-d3	;expected result
	tst.l	d2
	bpl	25f			;非正規化数
;正規化数
	add.l	d5,d3
	addx.l	d4,d2
	bmi	26f			;指数部が変わらない
	move.l	d1,d0
	and.l	#$7FFF0000,d0
	cmp.l	#$7FFE0000,d0
	beq	24f			;指数部が$7FFEのときは±Infになる
	add.l	#$00010000,d1		;指数部を1増やす
	roxr.l	#1,d2			;仮数部を右にずらす
	roxr.l	#1,d3
	or.l	#$80000000,d2
	bra	26f

24:
	or.l	#$7FFF0000,d1
	moveq.l	#0,d2
	moveq.l	#0,d3
	bra	26f

;非正規化数
25:
	add.l	d5,d3			;仮数部にマージンを加える。非正規化数が正規化数になる場合がある
	addx.l	d4,d2
26:
	movem.l	d1-d3,-(sp)
	fmove.x	(sp)+,fp2		;許容範囲の絶対値の上限
;<fp2.x:許容範囲の絶対値の上限
;------------------------------------------------
;下限<=上限にする。マージンが0のときは下限==上限になる。下限と上限は範囲に含まれる
	fcmp.x	fp1,fp2
	fbge	27f
	fmove.x	fp2,fp0
	fmove.x	fp1,fp2
	fmove.x	fp0,fp1
27:
;<fp1.x:許容範囲の下限
;<fp2.x:許容範囲の上限
;------------------------------------------------
;比較する
	fmove.x	(_ares,a6),fp0		;actual result
	fcmp.x	fp1,fp0
	fbult	97f			;actual resultがNaNまたは小さすぎる。失敗
	fcmp.x	fp2,fp0
	fbugt	97f			;actual resultがNaNまたは大きすぎる。失敗
	bra	98f			;成功

;------------------------------------------------
;完全に一致しなければならない
30:
	movem.l	(_eres,a6),d1-d3	;expected result
	movem.l	(_ares,a6),d4-d6	;actual result
	move.l	#OPTION_NANS,d0
	and.l	option_mode,d0
	bne	35f			;NaNの仮数部を比較する
;NaNの仮数部を無視する
;------------------------------------------------
;expected resultがNaNのとき仮数をすべて1にする
	move.l	d1,d0
	and.l	#$7FFF0000,d0
	cmp.l	#$7FFF0000,d0
	bne	32f			;±InfまたはNaNではない
	tst.l	d2
	bne	31f			;NaN
	tst.l	d3
	beq	32f			;±Inf
;NaN
31:
	move.l	#$FFFFFFFF,d2		;NaNの仮数部をすべて1にする
	move.l	#$FFFFFFFF,d3
32:
;------------------------------------------------
;actual resultがNaNのとき仮数をすべて1にする
	move.l	d4,d0
	and.l	#$7FFF0000,d0
	cmp.l	#$7FFF0000,d0
	bne	34f			;±InfまたはNaNではない
	tst.l	d5
	bne	33f			;NaN
	tst.l	d6
	beq	34f			;±Inf
;NaN
33:
	move.l	#$FFFFFFFF,d5		;NaNの仮数部をすべて1にする
	move.l	#$FFFFFFFF,d6
34:
;------------------------------------------------
;比較する
35:
	cmp.l	d1,d4
	bne	97f			;resultが一致しない。失敗
	cmp.l	d2,d5
	bne	97f			;resultが一致しない。失敗
	cmp.l	d3,d6
	bne	97f			;resultが一致しない。失敗
;resultが一致
39:
;------------------------------------------------
;成功
98:
	moveq.l	#1,d0
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

;失敗
97:
	moveq.l	#0,d0
	bra	99b
	.cpu	68000


;--------------------------------------------------------------------------------
;	結果を出力する
;--------------------------------------------------------------------------------

;--------------------------------------------------------------------------------
;output_status()
;	output status
;<(4,sp).l:actual status
;<(8,sp).l:expected status
;<(12,sp).l:0=failed,1=successful
	.cpu	68030
regs		reg	d3-d4/a3
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_a6:	.ds.l	1
_pc:	.ds.l	1
_asta:	.ds.l	1	;actual status
_esta:	.ds.l	1	;expected status
_succ:	.ds.l	1	;0=failed,1=successful
	.text
	.even
output_status::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	movea.l	(_esta,a6),a3		;expected status
	move.l	(_asta,a6),d3		;actual status
;successfulかどうかに関わらず完全に一致しているときだけexpectedを省略する
	moveq.l	#1,d4			;0=完全に一致してはいない,1=完全に一致している
	cmp.l	a3,d3
	beq	2f
	moveq.l	#0,d4			;0=完全に一致してはいない,1=完全に一致している
;expected
	putchr	#9
	move.l	a3,-(sp)
	jbsr	printfpsr
	addq.l	#4,sp
	putmsg	9,';expected',13,10
2:
;actual
	putchr	#9
	move.l	d3,-(sp)
	jbsr	printfpsr
	addq.l	#4,sp
	tst.l	d4
	beq	3f			;完全に一致してはいない
;完全に一致している
	putmsg	9,';'
	bra	4f

3:
;完全に一致してはいない
	putmsg	9,';actual ... '
4:
	tst.l	(_succ,a6)
	beq	5f			;failed
;successful
	putmsg	'OK',13,10
	bra	6f

5:
;failed
	putmsg	'ERROR',13,10
6:
;
99:
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

;--------------------------------------------------------------------------------
;output_single()
;	output result and status
;<(4,sp).s:actual result
;<(8,sp).l:actual status
;<(12,sp).s:expected result
;<(16,sp).l:expected status
;<(20,sp).l:0=failed,1=successful
	.cpu	68030
regs		reg	d0/d3-d4/a0/a3
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_a6:	.ds.l	1
_pc:	.ds.l	1
_ares:	.ds.s	1	;actual result
_asta:	.ds.l	1	;actual status
_eres:	.ds.s	1	;expected result
_esta:	.ds.l	1	;expected status
_succ:	.ds.l	1	;0=failed,1=successful
	.text
	.even
output_single::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	movea.l	(_eres,a6),a0		;expected result
	movea.l	(_esta,a6),a3		;expected status
	move.l	(_ares,a6),d0		;actual result
	move.l	(_asta,a6),d3		;actual status
;successfulかどうかに関わらず完全に一致しているときだけexpectedを省略する
	moveq.l	#1,d4			;0=完全に一致してはいない,1=完全に一致している
	cmp.l	a0,d0
	bne	1f
	cmp.l	a3,d3
	beq	2f
1:
	moveq.l	#0,d4			;0=完全に一致してはいない,1=完全に一致している
;expected
	putchr	#9
	puthex8	a0
	putchr	#','
	move.l	a3,-(sp)
	jbsr	printfpsr
	addq.l	#4,sp
	putmsg	9,';expected',13,10
2:
;actual
	putchr	#9
	puthex8	d0
	putchr	#','
	move.l	d3,-(sp)
	jbsr	printfpsr
	addq.l	#4,sp
	tst.l	d4
	beq	3f			;完全に一致してはいない
;完全に一致している
	putmsg	9,';'
	bra	4f

3:
;完全に一致してはいない
	putmsg	9,';actual ... '
4:
	tst.l	(_succ,a6)
	beq	5f			;failed
;successful
	putmsg	'OK',13,10
	bra	6f

5:
;failed
	putmsg	'ERROR',13,10
6:
;
99:
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

;--------------------------------------------------------------------------------
;output_double()
;	output result and status
;<(4,sp).d:actual result
;<(12,sp).l:actual status
;<(16,sp).d:expected result
;<(24,sp).l:expected status
;<(28,sp).l:0=failed,1=successful
	.cpu	68030
regs		reg	d0-d1/d3-d4/a0-a1/a3
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_a6:	.ds.l	1
_pc:	.ds.l	1
_ares:	.ds.d	1	;actual result
_asta:	.ds.l	1	;actual status
_eres:	.ds.d	1	;expected result
_esta:	.ds.l	1	;expected status
_succ:	.ds.l	1	;0=failed,1=successful
	.text
	.even
output_double::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	movem.l	(_eres,a6),a0-a1	;expected result
	movea.l	(_esta,a6),a3		;expected status
	movem.l	(_ares,a6),d0-d1	;actual result
	move.l	(_asta,a6),d3		;actual status
;successfulかどうかに関わらず完全に一致しているときだけexpectedを省略する
	moveq.l	#1,d4			;0=完全に一致してはいない,1=完全に一致している
	cmp.l	a0,d0
	bne	1f
	cmp.l	a1,d1
	bne	1f
	cmp.l	a3,d3
	beq	2f
1:
	moveq.l	#0,d4			;0=完全に一致してはいない,1=完全に一致している
;expected
	putchr	#9
	puthex16	a0,a1
	putchr	#','
	move.l	a3,-(sp)
	jbsr	printfpsr
	addq.l	#4,sp
	putmsg	9,';expected',13,10
2:
;actual
	putchr	#9
	puthex16	d0,d1
	putchr	#','
	move.l	d3,-(sp)
	jbsr	printfpsr
	addq.l	#4,sp
	tst.l	d4
	beq	3f			;完全に一致してはいない
;完全に一致している
	putmsg	9,';'
	bra	4f

3:
;完全に一致してはいない
	putmsg	9,';actual ... '
4:
	tst.l	(_succ,a6)
	beq	5f			;failed
;successful
	putmsg	'OK',13,10
	bra	6f

5:
;failed
	putmsg	'ERROR',13,10
6:
;
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

;--------------------------------------------------------------------------------
;output_packed()
;	output result and status
;<(4,sp).p:actual result
;<(16,sp).l:actual status
;<(20,sp).p:expected result
;<(32,sp).l:expected status
;<(36,sp).l:0=failed,1=successful
	.cpu	68030
regs		reg	d0-d4/a0-a3
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_a6:	.ds.l	1
_pc:	.ds.l	1
_ares:	.ds.p	1	;actual result
_asta:	.ds.l	1	;actual status
_eres:	.ds.p	1	;expected result
_esta:	.ds.l	1	;expected status
_succ:	.ds.l	1	;0=failed,1=successful
	.text
	.even
output_packed::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	movem.l	(_eres,a6),a0-a2	;expected result
	movea.l	(_esta,a6),a3		;expected status
	movem.l	(_ares,a6),d0-d2	;actual result
	move.l	(_asta,a6),d3		;actual status
;successfulかどうかに関わらず完全に一致しているときだけexpectedを省略する
	moveq.l	#1,d4			;0=完全に一致してはいない,1=完全に一致している
	cmp.l	a0,d0
	bne	1f
	cmp.l	a1,d1
	bne	1f
	cmp.l	a2,d2
	bne	1f
	cmp.l	a3,d3
	beq	2f
1:
	moveq.l	#0,d4			;0=完全に一致してはいない,1=完全に一致している
;expected
	putchr	#9
	puthex24	a0,a1,a2
	putchr	#','
	move.l	a3,-(sp)
	jbsr	printfpsr
	addq.l	#4,sp
	putmsg	9,';expected',13,10
2:
;actual
	putchr	#9
	puthex24	d0,d1,d2
	putchr	#','
	move.l	d3,-(sp)
	jbsr	printfpsr
	addq.l	#4,sp
	tst.l	d4
	beq	3f			;完全に一致してはいない
;完全に一致している
	putmsg	9,';'
	bra	4f

3:
;完全に一致してはいない
	putmsg	9,';actual ... '
4:
	tst.l	(_succ,a6)
	beq	5f			;failed
;successful
	putmsg	'OK',13,10
	bra	6f

5:
;failed
	putmsg	'ERROR',13,10
6:
;
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

;--------------------------------------------------------------------------------
;output_extended()
;	output result and status
;<(4,sp).x:actual result
;<(16,sp).l:actual status
;<(20,sp).x:expected result
;<(32,sp).l:expected status
;<(36,sp).l:0=failed,1=successful
	.cpu	68030
regs		reg	d0-d4/a0-a3
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_a6:	.ds.l	1
_pc:	.ds.l	1
_ares:	.ds.x	1	;actual result
_asta:	.ds.l	1	;actual status
_eres:	.ds.x	1	;expected result
_esta:	.ds.l	1	;expected status
_succ:	.ds.l	1	;0=failed,1=successful
	.text
	.even
output_extended::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	movem.l	(_eres,a6),a0-a2	;expected result
	movea.l	(_esta,a6),a3		;expected status
	movem.l	(_ares,a6),d0-d2	;actual result
	move.l	(_asta,a6),d3		;actual status
;successfulかどうかに関わらず完全に一致しているときだけexpectedを省略する
	moveq.l	#1,d4			;0=完全に一致してはいない,1=完全に一致している
	cmp.l	a0,d0
	bne	1f
	cmp.l	a1,d1
	bne	1f
	cmp.l	a2,d2
	bne	1f
	cmp.l	a3,d3
	beq	2f
1:
	moveq.l	#0,d4			;0=完全に一致してはいない,1=完全に一致している
;expected
	putchr	#9
	puthex24	a0,a1,a2
	putchr	#','
	move.l	a3,-(sp)
	jbsr	printfpsr
	addq.l	#4,sp
	putmsg	9,';expected',13,10
2:
;actual
	putchr	#9
	puthex24	d0,d1,d2
	putchr	#','
	move.l	d3,-(sp)
	jbsr	printfpsr
	addq.l	#4,sp
	tst.l	d4
	beq	3f			;完全に一致してはいない
;完全に一致している
	putmsg	9,';'
	bra	4f

3:
;完全に一致してはいない
	putmsg	9,';actual ... '
4:
	tst.l	(_succ,a6)
	beq	5f			;failed
;successful
	putmsg	'OK',13,10
	bra	6f

5:
;failed
	putmsg	'ERROR',13,10
6:
;
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000


;--------------------------------------------------------------------------------
;	出力サブルーチン
;--------------------------------------------------------------------------------

;----------------------------------------------------------------
;printfpcrrprm(fpcr)
;<(4,sp).l:fpcr
	.text
	.even
printfpcrrprm::
	movem.l	d0/a0,-(sp)
	move.l	(4*2+4,sp),d0		;fpcr
	and.w	#$00F0,d0		;rp,rm
	lsr.w	#4-2,d0
	lea.l	(10f,pc,d0.w),a0
	putstr	a0
	movem.l	(sp)+,d0/a0
	rts

	.align	4
10:
	.dc.l	'XRN'<<8|0
	.dc.l	'XRZ'<<8|0
	.dc.l	'XRM'<<8|0
	.dc.l	'XRP'<<8|0
	.dc.l	'SRN'<<8|0
	.dc.l	'SRZ'<<8|0
	.dc.l	'SRM'<<8|0
	.dc.l	'SRP'<<8|0
	.dc.l	'DRN'<<8|0
	.dc.l	'DRZ'<<8|0
	.dc.l	'DRM'<<8|0
	.dc.l	'DRP'<<8|0
	.dc.l	'DRN'<<8|0
	.dc.l	'DRZ'<<8|0
	.dc.l	'DRM'<<8|0
	.dc.l	'DRP'<<8|0

;----------------------------------------------------------------
;printfpsr
;<4(sp).l:fpsr
	.text
	.even
printfpsr::
	movem.l	d0/a0,-(sp)
	move.l	(4*2+4,sp),d0		;fpsr
	lea.l	(-128,sp),sp
	movea.l	sp,a0
	jbsr	fpsrstr
	putstr	sp
	lea.l	(128,sp),sp
	movem.l	(sp)+,d0/a0
	rts

;----------------------------------------------------------------
;fpsrstr
;<d0.l:fpsr
;<a0.l:buffer
;>a0.l:buffer
	.text
	.even
fpsrstr::
	movem.l	d0-d5/a1,-(sp)
	move.l	d0,d5			;fpsr
	sf.b	d4
;condition code byte
	lea.l	(10f,pc),a1
	moveq.l	#27,d1
1:
	move.b	(a1)+,d2
	move.b	(a1)+,d3
	btst.l	d1,d5
	beq	2f
	move.b	d2,(a0)+
	move.b	d3,(a0)+
	move.b	#'+',(a0)+
	st.b	d4
2:
	subq.w	#1,d1
	cmp.w	#24,d1
	bhs	1b
;quotient byte
	btst.l	#23,d5
	beq	1f
	move.b	#'(',(a0)+
	move.b	#'1',(a0)+
	move.b	#'<',(a0)+
	move.b	#'<',(a0)+
	move.b	#'2',(a0)+
	move.b	#'3',(a0)+
	move.b	#')',(a0)+
	move.b	#'+',(a0)+
	st.b	d4
1:
	move.l	d5,d0
	and.l	#$007F0000,d0
	beq	2f
	swap.w	d0
	move.b	#'(',(a0)+
	jbsr	decstr
	move.b	#'<',(a0)+
	move.b	#'<',(a0)+
	move.b	#'1',(a0)+
	move.b	#'6',(a0)+
	move.b	#')',(a0)+
	move.b	#'+',(a0)+
	st.b	d4
2:
;exception byte, accrued exception byte
	lea.l	(11f,pc),a1
	moveq.l	#15,d1
1:
	move.b	(a1)+,d2
	move.b	(a1)+,d3
	btst.l	d1,d5
	beq	2f
	move.b	d2,(a0)+
	move.b	d3,(a0)+
	move.b	#'+',(a0)+
	st.b	d4
2:
	subq.w	#1,d1
	cmp.w	#3,d1
	bhs	1b
	tst.b	d4
	beq	3f
	clr.b	-(a0)			;remove unnecessary '+'
	bra	8f

3:
	move.b	#'0',(a0)+
	clr.b	(a0)
8:
	movem.l	(sp)+,d0-d5/a1
	rts

10:
;		 27   26   25   24
	.dc.b	'MI','ZE','IN','NA'
11:
;		 15   14   13   12   11   10    9    8    7    6    5    4    3
	.dc.b	'BS','SN','OE','OF','UF','DZ','X2','X1','AV','AO','AU','AZ','AX'

;----------------------------------------------------------------
;printdate
;	yyyy-mm-ddThh:mm:ss+09:00
	.text
	.even
printdate::
	move.l	a0,-(sp)
	lea.l	(-28,sp),sp
	movea.l	sp,a0
	jbsr	datestr
	putstr	sp
	lea.l	(28,sp),sp
	movea.l	(sp)+,a0
	rts

;----------------------------------------------------------------
;datestr
;	yyyy-mm-ddThh:mm:ss+09:00
	.text
	.even
datestr::
	movem.l	d0-d3,-(sp)
;get date and time
	DOS	_GETDATE
	move.l	d0,d2
;<d2.l:date((dayofweek(0=sunday)<<16)+((year-1980)<<9)+(month<<5)+dayofmonth)
	DOS	_GETTIM2
	move.l	d0,d3
;<d3.l:time((hour<<16)+(minute<<8)+second)
;year
					;........ .....www yyyyyyym mmmddddd
	rol.w	#7,d2			;........ .....www mmmmdddd dyyyyyyy
	moveq.l	#$7F,d0			;________ ________ ________ _1111111
	and.w	d2,d0			;________ ________ ________ _yyyyyyy
	add.w	#1980,d0
	move.l	#4,d1
	jbsr	deczstr
	move.b	#'-',(a0)+
;month
	rol.w	#4,d2			;........ .....www dddddyyy yyyymmmm
	moveq.l	#$0F,d0			;________ ________ ________ ____1111
	and.w	d2,d0			;________ ________ ________ ____mmmm
	move.l	#2,d1
	jbsr	deczstr
	move.b	#'-',(a0)+
;dayofmonth
	rol.w	#5,d2			;........ .....www yyyyyyym mmmddddd
	moveq.l	#$1F,d0			;________ ________ ________ ___11111
	and.w	d2,d0			;________ ________ ________ ___ddddd
;	move.l	#2,d1
	jbsr	deczstr
	move.b	#'T',(a0)+
;hour
					;........ ...hhhhh ..mmmmmm ..ssssss
	swap.w	d3			;..mmmmmm ..ssssss ........ ...hhhhh
	moveq.l	#$1F,d0			;________ ________ ________ ___11111
	and.w	d3,d0			;________ ________ ________ ___hhhhh
;	move.l	#2,d1
	jbsr	deczstr
	move.b	#':',(a0)+
;minute
	rol.l	#8,d3			;..ssssss ........ ...hhhhh ..mmmmmm
	moveq.l	#$3F,d0			;________ ________ ________ __111111
	and.w	d3,d0			;________ ________ ________ __mmmmmm
;	move.l	#2,d1
	jbsr	deczstr
	move.b	#':',(a0)+
;second
	rol.l	#8,d3			;........ ...hhhhh ..mmmmmm ..ssssss
	moveq.l	#$3F,d0			;________ ________ ________ __111111
	and.w	d3,d0			;________ ________ ________ __ssssss
;	move.l	#2,d1
	jbsr	deczstr
	move.b	#'+',(a0)+
	move.b	#'0',(a0)+
	move.b	#'9',(a0)+
	move.b	#':',(a0)+
	move.b	#'0',(a0)+
	move.b	#'0',(a0)+
	clr.b	(a0)
	movem.l	(sp)+,d0-d3
	rts

;----------------------------------------------------------------
;printlong(number)
;	%d
;<(4,sp).l:number
	.text
	.even
printlong::
	move.l	(4,sp),-(sp)
	bpl	@f
	putchr	#'-'
	neg.l	(sp)
@@:
	jbsr	printdec
	addq.l	#4,sp
	rts

;----------------------------------------------------------------
;printdec(number)
;	%u
;<(4,sp).l:number
	.text
	.even
printdec::
	movem.l	d0/a0,-(sp)
	move.l	(4*2+4,sp),d0		;number
	lea.l	(-12,sp),sp
	movea.l	sp,a0
	jbsr	decstr
	putstr	sp
	lea.l	(12,sp),sp
	movem.l	(sp)+,d0/a0
	rts

;----------------------------------------------------------------
;printdecs(number,digits)
;	%*u
;<(4,sp).l:number
;<(8,sp).l:digits
	.text
	.even
printdecs::
	movem.l	d0-d2/a0,-(sp)
	movem.l	(4*4+4,sp),d0-d1	;number,digits
	moveq.l	#4,d2
	add.l	d1,d2
	and.w	#-4,d2
	suba.l	d2,sp
	movea.l	sp,a0
	jbsr	decsstr
	putstr	sp
	adda.l	d2,sp
	movem.l	(sp)+,d0-d2
	rts

;----------------------------------------------------------------
;printdecz(number,digits)
;	%0*u
;<(4,sp).l:number
;<(8,sp).l:digits
	.text
	.even
printdecz::
	movem.l	d0-d2/a0,-(sp)
	movem.l	(4*4+4,sp),d0-d1	;number,digits
	moveq.l	#4,d2
	add.l	d1,d2
	and.w	#-4,d2
	suba.l	d2,sp
	movea.l	sp,a0
	jbsr	deczstr
	putstr	sp
	adda.l	d2,sp
	movem.l	(sp)+,d0-d2
	rts

;----------------------------------------------------------------
;decstr
;	%u
;<d0.l:number
;<a0.l:buffer
;>a0.l:buffer
	.text
	.even
decstr::
	movem.l	d0-d2/a1,-(sp)
	tst.l	d0
	bne	1f
	move.b	#'0',(a0)+
	bra	5f

1:
	lea.l	baseten,a1
2:
	move.l	(a1)+,d1
	cmp.l	d1,d0
	blo	2b
3:
	moveq.l	#'0'-1,d2
4:
	addq.b	#1,d2
	sub.l	d1,d0
	bhs	4b
	add.l	d1,d0
	move.b	d2,(a0)+
	move.l	(a1)+,d1
	bne	3b
5:
	clr.b	(a0)
	movem.l	(sp)+,d0-d2/a1
	rts

;----------------------------------------------------------------
;decsstr
;	%*u
;<d0.l:number
;<d1.l:digits
;<a0.l:buffer
;>a0.l:buffer
	.text
	.even
decsstr::
	movem.l	d0-d2/a1-a2,-(sp)
	lea.l	baseten+4*10,a1
	movea.l	a1,a2
;remove zero digits
	tst.l	d1
	bne	@f
	moveq.l	#1,d1
@@:
;remove 11 or more digits
	cmp.l	#11,d1
	blo	3f
	sub.l	#11,d1
	swap.w	d1
1:
	swap.w	d1
2:
	move.b	#' ',(a0)+
	dbra	d1,2b
	swap.w	d1
	dbra	d1,1b
	moveq.l	#10,d1
3:
	lsl.w	#2,d1
	suba.w	d1,a2			;specified start point
;calculate minimum digits
	tst.l	d0
	beq	2f
	lea.l	(-4*10,a1),a1
1:
	cmp.l	(a1)+,d0
	blo	1b
2:
	subq.l	#4,a1			;required start point
;fill
	cmpa.l	a1,a2
	bhs	2f
1:
	move.b	#' ',(a0)+
	addq.l	#4,a2
	cmpa.l	a1,a2
	blo	1b
2:
;print decimal number
	move.l	(a1)+,d1
1:
	moveq.l	#'0'-1,d2
2:
	addq.b	#1,d2
	sub.l	d1,d0
	bhs	2b
	add.l	d1,d0
	move.b	d2,(a0)+
	move.l	(a1)+,d1
	bne	1b
	move.b	d1,(a0)			;0
	movem.l	(sp)+,d0-d2/a1-a2
	rts

;----------------------------------------------------------------
;deczstr
;	%0*u
;<d0.l:number
;<d1.l:digits
;<a0.l:buffer
;>a0.l:buffer
	.text
	.even
deczstr::
	movem.l	d0-d2/a1-a2,-(sp)
	lea.l	baseten+4*10,a1
	movea.l	a1,a2
;remove zero digits
	tst.l	d1
	bne	@f
	moveq.l	#1,d1
@@:
;remove 11 or more digits
	cmp.l	#11,d1
	blo	3f
	sub.l	#11,d1
	swap.w	d1
1:
	swap.w	d1
2:
	move.b	#'0',(a0)+
	dbra	d1,2b
	swap.w	d1
	dbra	d1,1b
	moveq.l	#10,d1
3:
	lsl.w	#2,d1
	suba.w	d1,a2			;specified start point
;calculate minimum digits
	tst.l	d0
	beq	2f
	lea.l	(-4*10,a1),a1
1:
	cmp.l	(a1)+,d0
	blo	1b
2:
	subq.l	#4,a1			;required start point
;fill
	cmpa.l	a1,a2
	bhs	2f
1:
	move.b	#'0',(a0)+
	addq.l	#4,a2
	cmpa.l	a1,a2
	blo	1b
2:
;print decimal number
	move.l	(a1)+,d1
1:
	moveq.l	#'0'-1,d2
2:
	addq.b	#1,d2
	sub.l	d1,d0
	bhs	2b
	add.l	d1,d0
	move.b	d2,(a0)+
	move.l	(a1)+,d1
	bne	1b
	move.b	d1,(a0)			;0
	movem.l	(sp)+,d0-d2/a1-a2
	rts

;----------------------------------------------------------------
;printfix(number,digits)
;	print fixed point decimal number
;<4(sp).l:number. fixed point decimal number * 10^(number of digits after decimal point)
;<8(sp).b:digits. number of digits after decimal point
regs		reg	d0-d4/a0-a1
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_a6:	.ds.l	1
_pc:	.ds.l	1
_numb:	.ds.l	1
_digi:	.ds.w	1
	.text
	.even
printfix::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	moveq.l	#0,d3
	move.b	(_digi,a6),d3
;<d3.l:number of digits after decimal point
	move.l	d3,d4
	addq.w	#3,d4
	and.w	#-4,d4			;round up to multiples of four
	add.w	#12,d4			;12 bytes for integer part and decimal point
;<d4.l:buffer size
	suba.l	d4,sp
;<sp.l:buffer
	movea.l	sp,a0
	move.l	(_numb,a6),d0		;fixed point decimal number * 10^(number of digits after decimal point)
	bne	20f			;non-zero
;zero
	move.b	#'0',(a0)+
	move.w	d3,d2			;number of digits after decimal point
	beq	13f			;omit decimal point
	move.b	#'.',(a0)+
	bra	12f

11:
	move.b	#'0',(a0)+
12:
	dbra	d2,11b
13:
	bra	80f			;print

20:
;non-zero
;<d0.l:fixed point decimal number * 10^(number of digits after decimal point)
	lea.l	baseten,a1
;zero suppression
21:
	move.l	(a1)+,d1
	cmp.l	d1,d0
	blo	21b
;convert to decimal number
22:
	moveq.l	#'0'-1,d2
23:
	addq.b	#1,d2
	sub.l	d1,d0
	bcc	23b
	add.l	d1,d0
	move.b	d2,(a0)+
	move.l	(a1)+,d1
	bne	22b
;
	move.l	a0,d2
	sub.l	sp,d2
;<d2.l:actual number of digits
	cmp.w	d3,d2
	bls	40f
;actual number of digits > number of digits after decimal point
;integer part exists
;insert '.'
	move.w	d3,d2			;number of digits after decimal point
	beq	33f			;omit decimal point
	movea.l	a0,a1
	addq.l	#1,a0
	bra	32f

31:
	move.b	-(a1),1(a1)
32:
	dbra	d2,31b
	move.b	#'.',(a1)
33:
	bra	80f			;print

40:
;actual number of digits <= number of digits after decimal point
;no integer part exists
;insert '0.00...'
	move.w	d3,d0			;number of digits after decimal point
	sub.w	d2,d0			;number of zeros after decimal point
	movea.l	a0,a1
	lea.l	2(a0,d0.w),a0
	bra	42f

41:
	move.b	-(a1),2(a1,d0.w)
42:
	dbra	d2,41b
	movea.l	sp,a1
	move.b	#'0',(a1)+
	move.b	#'.',(a1)+
	bra	44f

43:
	move.b	#'0',(a1)+
44:
	dbra	d0,43b

;print
80:
	suba.l	sp,a0
	move.l	a0,-(sp)		;length
	pea.l	4(sp)			;buffer
	jbsr	logging_write
;	addq.l	#8,sp
;	adda.l	d4,sp
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

;----------------------------------------------------------------
	.text
	.align	4
baseten::
	.dc.l	1000000000
	.dc.l	100000000
	.dc.l	10000000
	.dc.l	1000000
	.dc.l	100000
	.dc.l	10000
	.dc.l	1000
	.dc.l	100
	.dc.l	10
	.dc.l	1
	.dc.l	0

;----------------------------------------------------------------
;printhex2(number)
;	print hexadecimal number $XX
;<4(sp).b:number
	.text
	.even
printhex2::
	movem.l	d0-d2/a0,-(sp)
	move.b	4*4+4(sp),d0		;number
	subq.l	#4,sp			;buffer
	movea.l	sp,a0
	move.b	#'$',(a0)+
	moveq.l	#2-1,d2
2:
	rol.b	#4,d0
	moveq.l	#15,d1
	and.w	d0,d1
	move.b	10f(pc,d1.w),(a0)+
	dbra	d2,2b
	pea.l	1+2.w			;length
	pea.l	4(sp)			;buffer
	jbsr	logging_write
	lea.l	8+4(sp),sp
	movem.l	(sp)+,d0-d2/a0
	rts

10:
	.dc.b	'0123456789ABCDEF'

;----------------------------------------------------------------
;printhex4(number)
;	print hexadecimal number $XXXX
;<4(sp).w:number
	.text
	.even
printhex4::
	movem.l	d0-d2/a0,-(sp)
	move.w	4*4+4(sp),d0		;number
	subq.l	#6,sp			;buffer
	movea.l	sp,a0
	move.b	#'$',(a0)+
	moveq.l	#4-1,d2
2:
	rol.w	#4,d0
	moveq.l	#15,d1
	and.w	d0,d1
	move.b	10f(pc,d1.w),(a0)+
	dbra	d2,2b
	pea.l	1+4.w			;length
	pea.l	4(sp)			;buffer
	jbsr	logging_write
	lea.l	8+6(sp),sp
	movem.l	(sp)+,d0-d2/a0
	rts

10:
	.dc.b	'0123456789ABCDEF'

;----------------------------------------------------------------
;printhex8(number)
;	print hexadecimal number $XXXXXXXX
;<4(sp).l:number
	.text
	.even
printhex8::
	movem.l	d0-d2/a0,-(sp)
	move.l	4*4+4(sp),d0		;number
	lea.l	-10(sp),sp		;buffer
	movea.l	sp,a0
	move.b	#'$',(a0)+
	moveq.l	#8-1,d2
2:
	rol.l	#4,d0
	moveq.l	#15,d1
	and.w	d0,d1
	move.b	10f(pc,d1.w),(a0)+
	dbra	d2,2b
	pea.l	1+8.w			;length
	pea.l	4(sp)			;buffer
	jbsr	logging_write
	lea.l	8+10(sp),sp
	movem.l	(sp)+,d0-d2/a0
	rts

10:
	.dc.b	'0123456789ABCDEF'

;----------------------------------------------------------------
;printcrlf()
;	print CR and LF
	.text
	.even
printcrlf::
	move.l	d0,-(sp)
	pea.l	2.w			;length
	pea.l	10f(pc)			;buffer
	jbsr	logging_write
	addq.l	#8,sp
	move.l	(sp)+,d0
	rts

10:
	.dc.b	13,10

;----------------------------------------------------------------
;printchr(character)
;	print character
;<4(sp).b:character
	.text
	.even
printchr::
	move.l	d0,-(sp)
	pea.l	1.w			;length
	pea.l	4+4+4(sp)		;character
	jbsr	logging_write
	addq.l	#8,sp
	move.l	(sp)+,d0
	rts

;----------------------------------------------------------------
;printstr(string)
;	print string
;<4(sp).l:string
	.text
	.even
printstr::
	movem.l	d0/a0-a1,-(sp)
	movea.l	4*3+4(sp),a0		;string
	movea.l	a0,a1
@@:
	tst.b	(a1)+
	bne	@b
	subq.l	#1,a1
	suba.l	a0,a1			;length
	movem.l	a0-a1,-(sp)		;length, string
	jbsr	logging_write
	addq.l	#8,sp
	movem.l	(sp)+,d0/a0-a1
	rts


;--------------------------------------------------------------------------------
;	計算サブルーチン
;--------------------------------------------------------------------------------

;----------------------------------------------------------------
;mull
;	unsigned multiplication. long*long
;<d0.l:multiplicand
;<d1.l:multiplier
;>d0:d1.q:product
	.text
	.even
mull::
	cmp.l	#$0000FFFF,d0
	bhi	10f
	cmp.l	#$0000FFFF,d1
	bhi	10f
	mulu.w	d0,d1
	moveq.l	#0,d0
	rts

10:
	movem.l	d2-d4,-(sp)
					;    d0      d1      d2      d3      d4
					;   A   B   C   D   .   .   .   .   .   .
	move.l	d1,d2			;                   C   D
	move.l	d1,d3			;                           C   D
	swap.w	d2			;                   D   C
	move.l	d2,d4			;                                   D   C
	mulu.w	d0,d1			;            B*D
	mulu.w	d0,d4			;                                    B*C
	swap.w	d0			;   B   A
	mulu.w	d0,d3			;                            A*D
	mulu.w	d2,d0			;    A*C
					;  ACh ACl BDh BDl
	move.w	d1,d2			;                  --- BDl
	move.w	d0,d1			;      ---     ACl
	swap.w	d1			;          ACl BDh
	swap.w	d0			;  --- ACh
	add.l	d3,d1			;            +AD
	clr.w	d3
	addx.w	d3,d0
	add.l	d4,d1			;            +BC
	addx.w	d3,d0
	swap.w	d0			;  ACh ---
	swap.w	d1			;          BDh ACl
	move.w	d1,d0			;      ACl     ---
	move.w	d2,d1			;              BDl
	movem.l	(sp)+,d2-d4
	rts

;----------------------------------------------------------------
;divq
;	unsigned division. quad/quad
;<d0:d1.q:dividend
;<d2:d3.q:divisor
;>d0:d1.q:quotient
;>d2:d3.q:remainder
;>x:0
;>n:1=quotient is negative
;>z:1=quotient is zero
;>v:0
;>c:1=divide by zero. d0:d1 and d2:d3 are not changed. z=0,n=0
	.text
	.even
divq::
	tst.l	d2
	bne	20f			;$FFFFFFFF<divisor
;divisor<=$FFFFFFFF
	tst.l	d3
	beq	40f			;divisor==0
	tst.l	d0
	bne	50f			;$FFFFFFFF<dividend && divisor<=$FFFFFFFF
;dividend<=$FFFFFFFF && divisor<=$FFFFFFFF
	cmp.l	d3,d1
	bls	60f			;dividend<=divisor
;divisor<dividend
10:
	movem.l	d5-d6,-(sp)
	move.l	d3,d5
	moveq.l	#0,d3
	moveq.l	#31,d6
1:
	add.l	d1,d1
	addx.l	d3,d3
	cmp.l	d5,d3
	blo	2f
	addq.b	#1,d1
	sub.l	d5,d3
2:
	dbra	d6,1b
	subq.w	#1,d6			;$0000FFFF->$0000FFFE. x=0
	or.l	d0,d6			;n=*,z=0,v=0,c=0
	movem.l	(sp)+,d5-d6
	rts

;$FFFFFFFF<divisor
20:
	cmp.l	d2,d0
	bhi	30f			;divisor<dividend
	blo	70f			;dividend<divisor
	cmp.l	d3,d1
	bls	60f			;dividend<=divisor
;divisor<dividend
30:
	movem.l	d4-d6,-(sp)
	move.l	d2,d4
	move.l	d3,d5
	moveq.l	#0,d2
	move.l	d0,d3
	move.l	d1,d0
	moveq.l	#0,d1
	moveq.l	#31,d6
1:
	add.l	d1,d1
	addx.l	d0,d0
	addx.l	d3,d3
	addx.l	d2,d2
  .if 1
	cmp.l	d4,d2
	blo	3f
	bhi	2f
	cmp.l	d5,d3
	blo	3f
2:
	sub.l	d5,d3
	subx.l	d4,d2
	addq.b	#1,d1
3:
  .else
	addq.b	#1,d1
	sub.l	d5,d3
	subx.l	d4,d2
	bhs	2f
	subq.b	#1,d1
	add.l	d5,d3
	addx.l	d4,d2
2:
  .endif
	dbra	d6,1b
	subq.w	#1,d6			;$0000FFFF->$0000FFFE. x=0
	or.l	d0,d6			;n=*,z=0,v=0,c=0
	movem.l	(sp)+,d4-d6
	rts

;divisor==0
40:
	move.w	#%00001,ccr		;x=0,n=0,z=0,v=0,c=1
	rts

;$FFFFFFFF<dividend && divisor<=$FFFFFFFF
50:
	movem.l	d5-d6,-(sp)
	move.l	d3,d5
	moveq.l	#0,d3
	moveq.l	#63,d6
1:
	add.l	d1,d1
	addx.l	d0,d0
	addx.l	d3,d3
	cmp.l	d5,d3
	blo	2f
	addq.b	#1,d1
	sub.l	d5,d3
2:
	dbra	d6,1b
	subq.w	#1,d6			;$0000FFFF->$0000FFFE. x=0
	or.l	d0,d6			;n=*,z=0,v=0,c=0
	movem.l	(sp)+,d5-d6
	rts

;dividend<=divisor
60:
	beq	80f			;dividend==divisor
;dividend<divisor
70:
	move.l	d0,d2			;remainder=dividend
	move.l	d1,d3
	sub.l	d0,d0			;quotient=0. x=0
	moveq.l	#0,d1			;n=0,z=1,v=0,c=0
	rts

;dividend==divisor
80:
	moveq.l	#0,d2			;remainder=0
	moveq.l	#0,d3
	sub.l	d0,d0			;quotient=1. x=0
	moveq.l	#1,d1			;n=0,z=0,v=0,c=0
	rts

  .if 0
	movem.l	d4-d6,-(sp)
	move.l	d2,d4
	move.l	d3,d5
	moveq.l	#0,d2
	moveq.l	#0,d3
	moveq.l	#63,d6
1:
	add.l	d1,d1
	addx.l	d0,d0
	addx.l	d3,d3
	addx.l	d2,d2
	addq.b	#1,d1
	sub.l	d5,d3
	subx.l	d4,d2
	bcc	2f
	subq.b	#1,d1
	add.l	d5,d3
	addx.l	d4,d2
2:
	dbra	d6,1b
	movem.l	(sp)+,d4-d6
	rts
  .endif


;--------------------------------------------------------------------------------
;	解凍
;--------------------------------------------------------------------------------

DICTIONARY_BITS	equ	9
DICTIONARY_SIZE	equ	512
COMPRESS_PAGE	equ	1		;1=単語辞書のページ番号を圧縮する
COMPRESS_CHAR	equ	1		;1=文字を圧縮する

;----------------------------------------------------------------
;brcode n
;	n bit取り出す
;	n	取り出すbit数。0..25bit
;<d1.l:入力プール。右寄せ
;<d2.w:入力プールの残りbit数
;<a0.l:入力アドレス
;>d0.l:n bitのデータ。0拡張
;>d1.l:入力プール。右寄せ
;>d2.w:入力プールの残りbit数
;>a0.l:入力アドレス
brcode	.macro	n
	sub.w	n,d2
	bpl	@skip
@loop:
	lsl.l	#8,d1
	move.b	(a0)+,d1
	addq.w	#8,d2
	bmi	@loop
@skip:
	move.l	d1,d0
	lsr.l	d2,d0
	lsl.l	d2,d0
	eor.l	d0,d1
	lsr.l	d2,d0
	.endm

;----------------------------------------------------------------
;decompress(in,out)
;	解凍する
;	汎用ではない。エラーチェックを行っていないのでデータが壊れているとクラッシュする
regs		reg	d0-d7/a0-a5
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_a6:	.ds.l	1
_pc:	.ds.l	1
_in:	.ds.l	1			;入力アドレス
_out:	.ds.l	1			;出力アドレス
	.text
	.even
decompress::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
;単語辞書を初期化する
	lea.l	decompress_wdic,a3	;単語辞書
	movea.l	a3,a0
	move.w	#DICTIONARY_SIZE-1,d0
@@:
	clr.l	(a0)+
	clr.l	(a0)+
	dbra	d0,@b
;文字辞書を初期化する
	lea.l	decompress_cdic,a5	;文字辞書
	movea.l	a5,a0
	move.l	#$00010203,d3
	move.l	#$04040404,d4
	moveq.l	#256/4-1,d0
@@:
	move.l	d3,(a0)+
	add.l	d4,d3
	dbra	d0,@b
;
	moveq.l	#0,d1			;入力プール。右寄せ
	moveq.l	#0,d2			;入力プールの残りbit数
	moveq.l	#0,d5			;単語辞書のアドレスのゲタ
	movea.l	(_in,a6),a0		;入力アドレス
	movea.l	(_out,a6),a1		;出力アドレス
	movea.l	a1,a2
	adda.l	(a0)+,a2		;出力バッファの末尾
;解凍ループ
	cmpa.l	a2,a1
	bhs	199f
100:
	brcode	#1			;1bit取り出す
	bne	20f			;単語辞書にある
;単語辞書にない
	moveq.l	#1,d3			;length
	subq.w	#8,d5			;単語辞書のアドレスのゲタをずらす
	and.w	#(DICTIONARY_SIZE-1)<<3,d5
	movem.l	d3/a1,(a3,d5.l)		;単語辞書の先頭に登録する。length,address
	bra	30f

;単語辞書にある
20:
  .if COMPRESS_PAGE
;単語辞書のページ番号が圧縮されている
;単語辞書のページ番号を求める
	moveq.l	#0,d4			;wid。単語辞書のページ番号に2を加えた値の先頭の1を除いたbit数
@@:
	addq.w	#1,d4
	brcode	#1			;brcode(1)
	bne	@b
	brcode	d4			;brcode(wid)
	bset.l	d4,d0			;(1<<wid)+brcode(wid)
	subq.w	#2,d0			;(1<<wid)+brcode(wid)-2。単語辞書のページ番号
  .else
;単語辞書のページ番号が圧縮されていない
	brcode	#DICTIONARY_BITS	;brcode(DICTIONARY_BITS)。単語辞書のページ番号
  .endif
;単語辞書から取り出す
	lsl.l	#3,d0			;ページオフセット
	add.w	d5,d0			;ページオフセットに単語辞書のアドレスのゲタを加える
	and.w	#(DICTIONARY_SIZE-1)<<3,d0
	movem.l	(a3,d0.l),d3/a4		;length,address
;単語辞書のアドレスのゲタをずらす
	subq.l	#8,d5
	and.l	#(DICTIONARY_SIZE-1)<<3,d5
;1文字伸ばす
	addq.l	#1,d3
;単語辞書の先頭に登録する
	movem.l	d3/a1,(a3,d5.l)		;length,address
;単語辞書から取り出した文字を出力する
	subq.l	#1+1,d3
@@:
	move.b	(a4)+,(a1)+
	dbra	d3,@b
30:
  .if COMPRESS_CHAR
;文字が圧縮されている
;文字辞書のページ番号を求める
	moveq.l	#0,d4			;wid。文字辞書のページ番号に2を加えた値の先頭の1を除いたbit数
@@:
	addq.w	#1,d4
	brcode	#1			;brcode(1)
	bne	@b
	brcode	d4			;brcode(wid)
	bset.l	d4,d0			;(1<<wid)+brcode(wid)
	subq.w	#2,d0			;(1<<wid)+brcode(wid)-2。文字辞書のページ番号
;文字辞書から取り出す
	lea.l	(a5,d0.w),a4		;文字辞書のページアドレス
	move.b	(a4),d3
;文字辞書から取り出した文字を出力する
	move.b	d3,(a1)+
;取り出した文字を文字辞書の先頭に移動させる
	bra	2f

1:
	move.b	-(a4),1(a4)
2:
	dbra	d0,1b
	move.b	d3,(a5)
  .else
;文字が圧縮されていない
;入力された文字を出力する
	move.b	(a0)+,(a1)+
  .endif
	cmpa.l	a2,a1
	blo	100b
199:
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.bss
	.align	4
decompress_wdic::	.ds.l	2*DICTIONARY_SIZE	;単語辞書。length,address,...
decompress_cdic::	.ds.b	256			;文字辞書。0..255を最近使われた順に並べる

;--------------------------------------------------------------------------------
;	FABS.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fabs_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#63,-(sp)
	peamsg	'FABS.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FABS.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fabs_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fabs.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FABS.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fabs.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FABS.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fabs_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$E6,$3A,$29,$F7,$3F,$80,$9C,$C6,$1C,$4E,$E7,$51
	.dc.b	$E6,$9D,$3F,$88,$9D,$3C,$23,$2F,$96,$78,$CE,$A3,$84,$E9,$F0,$CE
	.dc.b	$A3,$11,$DE,$38,$E1,$E2,$8F,$8E,$DA,$34,$C5,$7A,$FE,$BB,$FC,$DF
	.dc.b	$F0,$5F,$AE,$6E,$7F,$D0,$1F,$F3,$9D,$FA,$9B,$5C,$7D,$2F,$D1,$E3
	.dc.b	$FD,$6A,$7F,$CF,$31,$FC,$B6,$11,$C3,$2D,$4F,$D7,$9F,$83,$87,$9C
	.dc.b	$8F,$BE,$76,$FF,$07,$8F,$24,$FF,$0D,$6C,$7C,$63,$BC,$7C,$D6,$9F
	.dc.b	$DB,$9F,$25,$79,$BD,$6E,$33,$F2,$CD,$9F,$90,$DB,$E5,$11,$FA,$EF
	.dc.b	$CA,$C3,$FB,$CF,$C3,$7C,$89,$B8,$79,$5A,$79,$63,$E1,$3F,$DA,$63
	.dc.b	$E4,$B4,$F3,$5F,$F7,$09,$D9,$F4,$B8,$3E,$E1,$27,$F1,$2F,$8D,$F4
	.dc.b	$EB,$F6,$EF,$C5,$1E,$CD,$7E,$D4,$7C,$5C,$FE,$14,$FC,$E1,$1F,$18
	.dc.b	$D1,$FB,$07,$F2,$FC,$3F,$8D,$23,$F5,$CF,$0E,$BF,$33,$0F,$14,$7D
	.dc.b	$5B,$F8,$27,$81,$F9,$F7,$DD,$D3,$FC,$0F,$8B,$F3,$44,$8F,$D3,$BE
	.dc.b	$F9,$FA,$11,$E0,$3D,$58,$FB,$9E,$67,$CC,$3D,$D5,$DD,$FC,$04,$FE
	.dc.b	$7B,$3F,$D4,$12,$3F,$6A,$F7,$9F,$97,$3E,$03,$D6,$FC,$69,$EC,$7B
	.dc.b	$34,$FC,$6D,$F8,$E3,$F8,$DF,$C3,$FE,$D4,$8F,$F0,$27,$B3,$F9,$F3
	.dc.b	$EA,$65,$F2,$9F,$3D,$F7,$7E,$BD,$FE,$6A,$3D,$97,$DF,$C6,$72,$FF
	.dc.b	$05,$6D,$FF,$18,$6B,$FC,$7D,$AF,$F1,$2E,$0F,$0E,$DF,$6C,$F5,$F1
	.dc.b	$7F,$01,$9F,$9A,$78,$8F,$9B,$96,$AB,$F1,$96,$E6,$F1,$12,$BC,$4F
	.dc.b	$C0,$7F,$C0,$D8,$77,$7E,$DE,$1D,$DF,$82,$7C,$ED,$BF,$14,$7D,$27
	.dc.b	$FA,$EF,$3F,$BC,$8F,$BD,$ED,$B7,$CB,$3E,$66,$BF,$54,$FA,$DB,$7E
	.dc.b	$14,$F9,$78,$78,$73,$74,$7C,$B2,$63,$F2,$DE,$B9,$FA,$63,$F6,$15
	.dc.b	$E9,$C3,$AE,$DF,$5C,$F8,$D3,$FC,$43,$07,$4F,$D7,$24,$7D,$77,$AB
	.dc.b	$F5,$0F,$8B,$BC,$7F,$20,$7D,$23,$1B,$7E,$F1,$FF,$00,$F5,$7E,$0C
	.dc.b	$8F,$E0,$DE,$AF,$EA,$CF,$4C,$F2,$D7,$F8,$BD,$7D,$07,$F2,$58,$3A
	.dc.b	$FF,$1C,$4B,$FC,$73,$D2,$BF,$51,$5F,$99,$CF,$E1,$D7,$9C,$EE,$7C
	.dc.b	$3E,$CF,$CC,$11,$FC,$C3,$CE,$7F,$66,$73,$D7,$F0,$CF,$D9,$D7,$E6
	.dc.b	$E1,$F0,$6D,$93,$B7,$F4,$84,$8F,$D2,$3D,$5F,$D9,$1E,$49,$FC,$73
	.dc.b	$F9,$F3,$E7,$99,$BF,$F1,$23,$FA,$DE,$EF,$D8,$11,$FE,$4D,$3F,$09
	.dc.b	$EA,$F4,$CF,$E1,$C7,$48,$EF,$83,$BF,$F7,$44,$9F,$DD,$3E,$77,$F5
	.dc.b	$27,$D7,$3F,$6B,$87,$C3,$70,$7C,$2F,$03,$F8,$21,$1F,$E0,$8F,$3E
	.dc.b	$4F,$D2,$EB,$F9,$93,$E6,$EB,$F9,$9A,$E7,$8B,$C1,$FE,$1C,$48,$FE
	.dc.b	$1C,$F3,$E4,$FA,$79,$7F,$19,$3C,$93,$F2,$4F,$F2,$AF,$85,$FC,$5C
	.dc.b	$8F,$F1,$77,$93,$EA,$DB,$E3,$ED,$F9,$38,$FE,$27,$8B,$C3,$FE,$40
	.dc.b	$48,$FE,$40,$F2,$7C,$07,$F0,$C3,$EE,$4F,$37,$12,$F1,$3F,$95,$12
	.dc.b	$3F,$95,$3C,$E7,$ED,$6B,$F1,$A7,$F3,$91,$F6,$BC,$5F,$E6,$44,$8F
	.dc.b	$E6,$4F,$1C,$FF,$35,$CD,$C1,$FD,$79,$FA,$EF,$1B,$F9,$D1,$2B,$F9
	.dc.b	$D3,$C5,$FE,$67,$A7,$C4,$9E,$31,$F1,$3E,$17,$FA,$11,$1F,$E8,$4F
	.dc.b	$17,$E1,$4F,$17,$F9,$74,$FE,$FB,$E2,$FF,$A5,$12,$3F,$A5,$3C,$67
	.dc.b	$F9,$56,$7F,$06,$7F,$35,$5F,$C1,$BE,$3B,$FA,$91,$23,$FA,$93,$C6
	.dc.b	$7F,$1D,$97,$C1,$9F,$09,$F9,$AF,$2B,$FC,$BD,$66,$92,$3F,$5B,$1F
	.dc.b	$C1,$1F,$D6,$BF,$31,$F1,$9D,$0F,$C0,$2F,$2F,$A9,$23,$FC,$FB,$19
	.dc.b	$7A,$9D,$08,$7C,$8B,$F8,$1B,$7E,$27,$2F,$E9,$7E,$67,$A0,$95,$D1
	.dc.b	$F8,$19,$7E,$D3,$0F,$D1,$DF,$F0,$E5,$E6,$FC,$02,$5F,$91,$E2,$FE
	.dc.b	$6A,$38,$9F,$AD,$F3,$BE,$39,$23,$D6,$F0,$AE,$DD,$1F,$BB,$9E,$5E
	.dc.b	$7F,$CE,$23,$F0,$9E,$17,$FB,$99,$E1,$5E,$7F,$43,$EA,$92,$FE,$4B
	.dc.b	$C1,$FE,$27,$3C,$2B,$F8,$17,$A3,$F7,$09,$3F,$41,$E0,$FD,$69,$E0
	.dc.b	$FF,$1C,$F4,$BF,$06,$47,$EB,$BC,$36,$FC,$D4,$7E,$F2,$FF,$88,$7A
	.dc.b	$7F,$8A,$24,$7D,$E7,$87,$07,$DA,$9F,$D4,$CF,$E1,$BD,$4F,$C8,$91
	.dc.b	$FC,$2B,$C3,$83,$F3,$11,$DC,$FE,$63,$EA,$FE,$60,$8F,$E3,$1E,$0F
	.dc.b	$E5,$0E,$07,$E5,$3D,$6F,$CF,$91,$FC,$9B,$C3,$7F,$E2,$D8,$7F,$85
	.dc.b	$AF,$FC,$BB,$EB,$FE,$98,$97,$F9,$A7,$84,$FB,$0E,$11,$FE,$72,$F6
	.dc.b	$3F,$5A,$47,$F4,$2F,$07,$F5,$11,$C2,$FF,$D7,$EF,$ED,$09,$1F,$A8
	.dc.b	$78,$3E,$C7,$80,$7F,$D6,$9E,$CF,$EF,$08,$FE,$C1,$E3,$1F,$9C,$8F
	.dc.b	$EB,$7E,$D7,$F0,$22,$3F,$B6,$77,$C3,$F8,$68,$FD,$EF,$B7,$FC,$20
	.dc.b	$8F,$EF,$1F,$C6,$D7,$FB,$5F,$37,$5F,$FA,$D1,$B5,$FE,$7C,$6B,$F9
	.dc.b	$33,$F8,$2D,$FF,$14,$23,$FC,$18,$FE,$E6,$7F,$8A,$F2,$FD,$7B,$C0
	.dc.b	$87,$C3,$3F,$D8,$65,$3F,$D6,$7E,$E7,$F2,$02,$3F,$C5,$1F,$49,$87
	.dc.b	$81,$C0,$D1,$FC,$68,$FE,$52,$43,$07,$FA,$E7,$D5,$FF,$55,$93,$1F
	.dc.b	$F7,$53,$FE,$FC,$8E,$07,$68,$EE,$FA,$5A,$FD,$DF,$21,$EE,$E1,$B1
	.dc.b	$C1,$AF,$A1,$AF,$FA,$44,$F9,$13,$FC,$ED,$FE,$ED,$1A,$7E,$43,$5F
	.dc.b	$04,$F7,$3C,$1E,$F4,$8F,$03,$5F,$8F,$CF,$FE,$B5,$AF,$F4,$27,$93
	.dc.b	$FE,$16,$9F,$55,$6D,$FA,$92,$8C,$47,$8B,$DF,$F7,$11,$F7,$3F,$C4
	.dc.b	$FE,$1F,$EF,$6F,$F8,$AC,$5F,$0F,$8E,$1F,$C1,$46,$30,$FD,$35,$6E
	.dc.b	$F5,$1F,$37,$E0,$BE,$61,$1F,$98,$F7,$78,$DF,$4C,$38,$C3,$0E,$CB
	.dc.b	$0E,$CF,$03,$ED,$7C,$1F,$D9,$23,$F6,$5F,$E5,$DF,$0F,$D2,$8D,$31
	.dc.b	$F0,$FE,$0D,$9F,$0D,$F8,$32,$61,$F8,$37,$9B,$E1,$3C,$D5,$CD,$C5
	.dc.b	$AF,$A7,$C3,$FE,$34,$93,$F8,$D7,$FA,$11,$E5,$B9,$1C,$A3,$C9,$3F
	.dc.b	$95,$24,$7E,$55,$E2,$F8,$27,$8B,$F1,$F4,$FA,$F5,$ED,$C3,$F3,$C4
	.dc.b	$9F,$CF,$3F,$D4,$F2,$F8,$DA,$C9,$F5,$C7,$C5,$D7,$F4,$C4,$8F,$D3
	.dc.b	$3C,$5F,$83,$3C,$5F,$11,$FC,$AF,$35,$7C,$FD,$BF,$60,$49,$FD,$83
	.dc.b	$FD,$97,$7F,$B0,$48,$F9,$BC,$1F,$5F,$E1,$FF,$70,$48,$FD,$C3,$C5
	.dc.b	$FA,$73,$C5,$F1,$9F,$8D,$9F,$C0,$78,$3F,$C0,$49,$1F,$C0,$5F,$E0
	.dc.b	$9C,$7F,$85,$24,$7D,$E8,$FF,$80,$1C,$47,$8C,$7F,$07,$6B,$FB,$4F
	.dc.b	$86,$FE,$1C,$47,$F8,$73,$D9,$FC,$64,$AB,$BC,$7A,$34,$FE,$05,$18
	.dc.b	$E4,$FE,$7F,$5B,$AE,$38,$70,$7F,$81,$D7,$57,$F8,$D3,$5C,$9F,$C7
	.dc.b	$DC,$3F,$C6,$1A,$FF,$B3,$8D,$9F,$C8,$FE,$1B,$F9,$41,$1F,$E5,$0D
	.dc.b	$74,$60,$86,$08,$7F,$18,$78,$3F,$C6,$BE,$1B,$F9,$A1,$1F,$E6,$8D
	.dc.b	$7F,$9A,$35,$E0,$68,$86,$8A,$7F,$26,$F8,$6F,$E8,$04,$7F,$A0,$35
	.dc.b	$FE,$80,$D7,$81,$A2,$1A,$29,$FC,$CB,$E1,$BF,$A6,$11,$FE,$98,$D7
	.dc.b	$FA,$63,$5E,$06,$88,$68,$A7,$F3,$EE,$BF,$EB,$04,$7F,$AC,$3F,$D7
	.dc.b	$B4,$FE,$B3,$A7,$F8,$13,$EE,$FF,$BC,$36,$BF,$92,$75,$FF,$30,$97
	.dc.b	$6D,$EF,$95,$FC,$7E,$72,$FE,$5B,$E1,$FF,$B4,$11,$FE,$D0,$F8,$6F
	.dc.b	$A4,$43,$F9,$2B,$DE,$FA,$C7,$ED,$2B,$FC,$3D,$D9,$FE,$0F,$7B,$C7
	.dc.b	$EE,$3E,$25,$F1,$1F,$DE,$9B,$1F,$DF,$9A,$FF,$83,$9A,$FF,$85,$D7
	.dc.b	$C4,$7F,$4E,$6C,$7F,$3B,$6B,$E3,$F1,$35,$32,$34,$FC,$7F,$17,$11
	.dc.b	$DE,$3C,$84,$7C,$8D,$74,$60,$86,$08,$79,$DF,$77,$C4,$48,$F7,$35
	.dc.b	$E1,$BF,$DD,$F1,$FE,$D1,$E4,$7F,$C8,$78,$FB,$1F,$C1,$78,$DF,$7C
	.dc.b	$D6,$BF,$E3,$D9,$7C,$47,$CD,$26,$1F,$35,$F5,$5F,$78,$FC,$26,$1C
	.dc.b	$9F,$F3,$45,$77,$BF,$B5,$87,$79,$F9,$27,$CF,$F8,$8F,$B4,$47,$ED
	.dc.b	$3F,$77,$E1,$FE,$B1,$23,$F2,$5E,$5D,$7F,$09,$17,$2F,$CF,$79,$9F
	.dc.b	$77,$E2,$3F,$0C,$4A,$FC,$33,$D5,$FC,$05,$71,$CB,$F2,$18,$59,$F5
	.dc.b	$CD,$5C,$6C,$3C,$51,$F8,$D9,$E8,$FE,$45,$AE,$4F,$CB,$38,$7E,$83
	.dc.b	$5E,$E6,$8F,$CE,$7C,$47,$E7,$C8,$E0,$D1,$FA,$9F,$3F,$F7,$47,$9F
	.dc.b	$65,$7E,$F7,$D0,$95,$F0,$0F,$D0,$7C,$47,$EA,$C8,$FE,$AD,$F5,$ED
	.dc.b	$FC,$27,$0E,$B3,$C4,$FA,$27,$F0,$5A,$EA,$E0,$FD,$C7,$C4,$7E,$DC
	.dc.b	$8F,$ED,$DE,$BB,$7E,$FE,$BA,$CF,$1B,$FC,$CE,$3D,$1F,$E3,$58,$7B
	.dc.b	$7E,$23,$F8,$11,$2B,$F8,$13,$5F,$E0,$4D,$7F,$81,$35,$FD,$03,$5F
	.dc.b	$CC,$BF,$C3,$7E,$22,$47,$F0,$D6,$BF,$C3,$5A,$FF,$0D,$6B,$FA,$B6
	.dc.b	$BF,$A4,$6A,$D1,$F1,$1F,$C6,$C9,$1F,$C6,$DA,$FF,$1B,$6B,$FC,$6D
	.dc.b	$AF,$EE,$5A,$FE,$C9,$FE,$4D,$F1,$12,$3F,$93,$35,$FE,$4C,$D7,$F9
	.dc.b	$33,$5F,$E0,$8D,$7F,$7A,$D7,$D5,$F1,$1F,$CC,$C8,$FF,$33,$6B,$FC
	.dc.b	$CD,$AF,$F3,$36,$BF,$C4,$5A,$FF,$0A,$7F,$9F,$7C,$3C,$8F,$E7,$CD
	.dc.b	$7F,$9F,$35,$FE,$7C,$D7,$F8,$E3,$5F,$E2,$AD,$7E,$97,$C3,$FF,$4B
	.dc.b	$23,$FD,$2D,$AF,$F4,$B6,$BF,$D2,$DA,$FF,$29,$6B,$FC,$89,$FE,$AD
	.dc.b	$F0,$F2,$3F,$AB,$35,$FE,$AC,$D7,$FA,$B3,$5F,$E6,$8D,$7F,$96,$B5
	.dc.b	$FC,$57,$C3,$FF,$63,$23,$FD,$8D,$AF,$F6,$36,$BF,$D8,$DA,$FF,$41
	.dc.b	$6B,$FC,$E9,$FE,$DD,$F0,$F2,$3F,$B7,$35,$FF,$7D,$FC,$53,$7F,$F5
	.dc.b	$CB,$63,$FD,$4C,$D7,$FB,$F7,$BD,$FE,$0F,$F7,$F7,$52,$7F,$C1,$D1
	.dc.b	$FD,$D7,$97,$FB,$F1,$23,$FB,$F3,$FE,$FA,$BE,$9C,$68,$4C,$9C,$C6
	.dc.b	$21,$18,$77,$D4,$C7,$C5,$1F,$E8,$69,$D3,$C3,$1F,$43,$EC,$F8,$8F
	.dc.b	$11,$1E,$C7,$79,$47,$9E,$7A,$E3,$E7,$AF,$34,$7F,$8C,$6F,$38,$F1
	.dc.b	$8B,$3E,$6C,$BE,$6F,$C4,$7C,$12,$47,$80,$F8,$0F,$77,$E8,$CF,$27
	.dc.b	$FA,$CD,$DC,$BD,$A7,$BB,$36,$1F,$6B,$E2,$3E,$71,$2F,$B1,$F2,$9E
	.dc.b	$CF,$C7,$C3,$B5,$7C,$B8,$F0,$90,$D1,$F7,$8F,$C2,$FC,$47,$DB,$23
	.dc.b	$DB,$04,$76,$D3,$E4,$4F,$D3,$3A,$3F,$E5,$DC,$BD,$98,$78,$B2,$FC
	.dc.b	$9F,$C4,$7E,$1C,$97,$D5,$E9,$AF,$CF,$8E,$6F,$F9,$46,$F9,$C6,$0F
	.dc.b	$A4,$7E,$73,$E2,$3F,$24,$47,$D0,$F4,$7E,$E9,$C4,$FA,$25,$CF,$E8
	.dc.b	$E3,$F8,$C8,$FC,$87,$C4,$7E,$74,$8F,$C3,$7A,$3F,$04,$F9,$67,$EB
	.dc.b	$0E,$91,$F0,$5F,$D8,$FC,$47,$E9,$C8,$FC,$D7,$9D,$78,$A3,$FC,$FD
	.dc.b	$87,$C0,$7A,$E9,$FE,$89,$D3,$F7,$BF,$11,$FB,$12,$4F,$D7,$79,$E3
	.dc.b	$F8,$2B,$E5,$7F,$63,$0F,$C7,$61,$FC,$18,$FD,$D1,$30,$FC,$03,$C9
	.dc.b	$F9,$1B,$FE,$2D,$F1,$EB,$F5,$DA,$7D,$43,$F7,$18,$7F,$05,$23,$F8
	.dc.b	$B7,$A3,$C0,$FA,$65,$BF,$84,$FF,$A4,$EF,$F8,$1B,$FB,$C8,$FE,$34
	.dc.b	$DC,$7F,$42,$D3,$F3,$0E,$1F,$31,$93,$F9,$3B,$97,$F0,$A7,$F2,$AE
	.dc.b	$FF,$E3,$84,$7F,$44,$FE,$03,$16,$7F,$EB,$73,$E6,$E5,$FC,$77,$A1
	.dc.b	$EE,$FF,$1F,$D7,$F4,$F1,$8C,$7F,$21,$E2,$7F,$29,$23,$FA,$F7,$B6
	.dc.b	$1D,$63,$F1,$67,$F2,$CA,$E7,$7D,$70,$F0,$3F,$C2,$FC,$5F,$E6,$84
	.dc.b	$AF,$DD,$3D,$1F,$F6,$55,$FA,$1F,$E6,$F8,$F9,$56,$1D,$9F,$D3,$79
	.dc.b	$1F,$CF,$C8,$FF,$05,$7A,$3F,$EC,$9A,$E8,$AB,$A5,$73,$8F,$BD,$E4
	.dc.b	$FF,$49,$23,$FC,$3D,$E7,$1F,$EE,$7A,$E7,$5C,$B2,$F7,$DF,$EE,$7C
	.dc.b	$DF,$EA,$44,$7F,$8C,$3C,$A3,$FD,$B5,$5C,$B3,$F9,$15,$F0,$1F,$E6
	.dc.b	$5E,$77,$F5,$B2,$47,$F2,$17,$93,$FD,$5A,$7F,$9D,$61,$E3,$8E,$47
	.dc.b	$D5,$BF,$2B,$F9,$BF,$4F,$FB,$31,$1F,$E5,$8F,$47,$FD,$37,$5D,$30
	.dc.b	$F2,$C7,$C9,$6D,$F7,$8F,$E2,$19,$7E,$EF,$D1,$FE,$E4,$48,$FE,$70
	.dc.b	$F5,$E0,$FC,$BC,$7F,$21,$AE,$44,$BF,$EA,$51,$FD,$EC,$87,$F4,$13
	.dc.b	$96,$1E,$57,$DC,$7C,$03,$F0,$F9,$7C,$0C,$7F,$87,$71,$7A,$3D,$2F
	.dc.b	$EA,$F5,$F8,$36,$E2,$69,$DA,$3C,$D8,$FC,$02,$38,$3F,$B0,$CF,$E5
	.dc.b	$DF,$EF,$ED,$7E,$89,$FB,$FB,$FC,$53,$FA,$03,$91,$FD,$DE,$38,$3E
	.dc.b	$B3,$9C,$F9,$BD,$E9,$1E,$06,$BF,$C2,$8E,$87,$E8,$4E,$B5,$FD,$9B
	.dc.b	$37,$47,$DC,$6C,$FC,$A3,$EF,$17,$07,$C1,$7C,$62,$47,$C0,$7E,$4F
	.dc.b	$B5,$F8,$42,$47,$3A,$75,$FE,$23,$E0,$A4,$F5,$7C,$5D,$B8,$8F,$85
	.dc.b	$1F,$50,$8F,$CE,$BE,$AF,$DE,$3F,$83,$BE,$BA,$E5,$1F,$4A,$78,$BD
	.dc.b	$0F,$AF,$F0,$DF,$80,$23,$83,$B1,$F0,$6B,$BE,$9F,$26,$FC,$CF,$F2
	.dc.b	$A8,$CF,$0F,$F2,$FC,$3F,$4B,$F0,$DF,$8C,$26,$1F,$86,$7E,$3B,$D4
	.dc.b	$87,$AC,$67,$F2,$1F,$8C,$6B,$0F,$03,$83,$86,$AF,$87,$FC,$D9,$23
	.dc.b	$F2,$AF,$81,$DE,$5E,$87,$FC,$BF,$9F,$B3,$3F,$E9,$13,$99,$FE,$51
	.dc.b	$9F,$F0,$0F,$87,$FD,$39,$23,$F4,$2F,$5C,$BC,$B1,$D8,$FC,$16,$1D
	.dc.b	$5F,$AE,$7A,$F0,$F9,$0F,$E3,$FE,$1F,$F6,$64,$7F,$5A,$F5,$8F,$2C
	.dc.b	$71,$3D,$27,$C3,$AF,$09,$23,$AC,$7C,$6F,$87,$FD,$F9,$1F,$DC,$BD
	.dc.b	$5F,$29,$E9,$3E,$A6,$BF,$4D,$C5,$75,$76,$FF,$6C,$CB,$F9,$2F,$C3
	.dc.b	$FF,$0C,$24,$FF,$06,$7B,$3F,$D4,$F4,$F9,$D7,$D6,$FA,$4E,$4F,$71
	.dc.b	$F4,$8F,$DA,$7C,$2F,$F1,$82,$3F,$C4,$9E,$BB,$AF,$AE,$BF,$6E,$31
	.dc.b	$5D,$16,$5F,$CF,$3E,$1F,$F9,$19,$27,$F8,$EB,$D2,$F9,$C7,$17,$A4
	.dc.b	$FF,$44,$F8,$6F,$E5,$64,$9F,$E4,$AF,$17,$D7,$9F,$BF,$6F,$B7,$1F
	.dc.b	$31,$BA,$7C,$B8,$FE,$25,$F0,$DF,$CD,$88,$FF,$30,$7A,$3C,$4F,$C1
	.dc.b	$94,$F9,$45,$CF,$DE,$FE,$C0,$FE,$6F,$97,$E9,$5A,$7F,$02,$3F,$9A
	.dc.b	$39,$7C,$E6,$8C,$9B,$30,$7F,$42,$F8,$6F,$E9,$A4,$7F,$A3,$BF,$82
	.dc.b	$C5,$E5,$7F,$33,$3E,$B5,$7F,$70,$F8,$B7,$C1,$53,$ED,$7C,$3F,$F5
	.dc.b	$82,$63,$FD,$49,$E9,$7C,$E3,$EB,$1F,$CF,$6B,$9B,$6B,$F0,$9F,$0F
	.dc.b	$FD,$88,$91,$FD,$71,$E6,$FF,$3B,$AF,$2B,$C4,$87,$CD,$F8,$6F,$ED
	.dc.b	$64,$7F,$B2,$BC,$9F,$D5,$C7,$F0,$DB,$F1,$C7,$F2,$16,$ED,$FE,$BC
	.dc.b	$6E,$FF,$84,$3F,$DF,$0B,$E1,$BF,$BD,$B6,$BF,$DA,$6D,$7F,$DC,$4D
	.dc.b	$7E,$C5,$7F,$63,$BF,$8D,$1F,$87,$9E,$E8,$38,$6E,$5E,$93,$C2,$47
	.dc.b	$C2,$FB,$EB,$B4,$7E,$EA,$7D,$EF,$53,$B3,$87,$DB,$3C,$DF,$4F,$C3
	.dc.b	$7B,$88,$FB,$9E,$D8,$F3,$DB,$F1,$73,$DE,$BF,$15,$F0,$72,$3E,$BE
	.dc.b	$3F,$37,$E1,$BE,$49,$23,$E4,$BD,$1F,$85,$7D,$1F,$8D,$7C,$DF,$B1
	.dc.b	$7F,$95,$7F,$1D,$F0,$DF,$50,$8F,$D4,$7A,$3F,$D7,$B4,$FC,$14,$F3
	.dc.b	$C7,$F0,$0F,$39,$FC,$27,$C3,$7D,$E2,$3F,$79,$E7,$8F,$96,$BF,$94
	.dc.b	$EA,$BE,$9A,$7E,$21,$E9,$3F,$8D,$F8,$4F,$C4,$93,$1F,$C4,$BD,$32
	.dc.b	$F2,$C7,$EA,$6B,$9B,$F6,$34,$F8,$4F,$DB,$3F,$0F,$1F,$94,$23,$FC
	.dc.b	$2E,$FF,$06,$FE,$88,$FC,$24,$69,$FC,$BB,$5E,$98,$BE,$1B,$F3,$E4
	.dc.b	$8F,$CF,$BE,$77,$99,$FE,$31,$F6,$3F,$C7,$C7,$A1,$FC,$DF,$C3,$7E
	.dc.b	$A0,$8F,$EA,$1F,$F3,$2E,$AF,$6B,$FC,$8C,$49,$FF,$1E,$7B,$7F,$E4
	.dc.b	$9F,$72,$47,$F8,$FD,$FF,$43,$F6,$B8,$7F,$8D,$7C,$37,$ED,$89,$5F
	.dc.b	$B6,$7B,$E1,$D6,$3F,$B7,$BF,$E5,$5C,$3A,$C7,$F7,$2C,$2E,$FD,$63
	.dc.b	$F6,$D8,$D7,$F8,$57,$C3,$7F,$04,$23,$FC,$11,$AF,$F0,$46,$BF,$39
	.dc.b	$AF,$F0,$E6,$BF,$B6,$7F,$87,$FC,$2C,$8F,$E1,$ED,$7F,$87,$B5,$FB
	.dc.b	$2D,$7F,$8C,$B5,$FE,$02,$D7,$F8,$7F,$C2,$FF,$1C,$23,$FC,$71,$AF
	.dc.b	$F1,$C6,$BF,$86,$6B,$FC,$99,$AF,$F0,$E7,$F9,$3F,$C2,$C8,$FE,$4E
	.dc.b	$D7,$F9,$3B,$5F,$C8,$B5,$FE,$62,$D7,$F8,$CB,$5F,$E4,$FE,$0F,$F3
	.dc.b	$42,$3F,$CD,$1A,$FF,$34,$6B,$F9,$E6,$BF,$CF,$9A,$FF,$26,$7F,$9F
	.dc.b	$FC,$2C,$8F,$E7,$ED,$7F,$9F,$B5,$FD,$4B,$5F,$E9,$2D,$7F,$98,$B5
	.dc.b	$FE,$7F,$F0,$BF,$D3,$08,$FF,$4C,$6B,$FD,$31,$AF,$ED,$9A,$FF,$56
	.dc.b	$6B,$FC,$F9,$FE,$AF,$F0,$B2,$3F,$AB,$B5,$FE,$AE,$D7,$F8,$0B,$5F
	.dc.b	$EC,$2D,$7F,$A4,$B5,$FE,$AF,$F0,$BF,$D9,$08,$FF,$64,$6B,$FD,$91
	.dc.b	$AF,$F0,$E6,$BF,$DB,$9A,$FF,$56,$7F,$B7,$F1,$48,$FE,$DE,$D7,$FB
	.dc.b	$7B,$5F,$E3,$2D,$7F,$BC,$B5,$FE,$C2,$D5,$F1,$8E,$0F,$F7,$BB,$63
	.dc.b	$F9,$71,$FD,$D5,$C3,$86,$BF,$CC,$DA,$FF,$6F,$6B,$C6,$3A,$91,$EA
	.dc.b	$78,$9C,$3D,$5F,$1F,$0C,$8D,$CA,$31,$1E,$36,$BE,$36,$BE,$36,$BE
	.dc.b	$B6,$BF,$0B,$8B,$E0,$91,$EE,$D7,$E0,$35,$F8,$0D,$7E,$03,$5F,$90
	.dc.b	$FC,$CE,$69,$1E,$A6,$BF,$2D,$AF,$CB,$6B,$F2,$DA,$FD,$36,$BF,$6B
	.dc.b	$AF,$EC,$91,$F8,$ED,$7E,$C3,$5F,$B0,$D7,$EC,$35,$FC,$03,$F8,$4F
	.dc.b	$4E,$47,$D2,$6B,$F8,$36,$BF,$83,$6B,$F8,$36,$BF,$8B,$6B,$F9,$2F
	.dc.b	$2F,$F2,$24,$7E,$FB,$5F,$C8,$35,$FC,$83,$5F,$C8,$35,$FC,$C3,$F9
	.dc.b	$CF,$56,$47,$E2,$9A,$FE,$6D,$AF,$E6,$DA,$FE,$6D,$AF,$E8,$DA,$FE
	.dc.b	$AB,$DA,$FD,$49,$1F,$CB,$B5,$FD,$43,$5F,$D4,$35,$FD,$43,$5F,$D8
	.dc.b	$3F,$B4,$F8,$59,$1F,$A2,$6B,$FB,$6F,$8D,$7B,$7F,$E5,$FE,$BF,$EE
	.dc.b	$D1,$89,$FF,$37,$46,$27,$FC,$13,$C5,$BE,$2F,$DC,$7C,$3D,$BF,$E0
	.dc.b	$0D,$7F,$83,$BF,$E1,$0E,$4C,$71,$A3,$F8,$49,$27,$F6,$EF,$EF,$96
	.dc.b	$5E,$35,$97,$8D,$F2,$E1,$FE,$77,$9D,$4F,$F8,$3F,$B1,$97,$83,$0E
	.dc.b	$D8,$7F,$13,$6B,$FC,$4F,$DD,$FE,$36,$47,$F8,$4B,$E3,$C3,$AE,$3D
	.dc.b	$F4,$F7,$46,$2B,$FC,$3F,$87,$77,$CA,$7F,$88,$B9,$FF,$C6,$26,$5F
	.dc.b	$5D,$FF,$21,$F8,$6F,$E5,$64,$8F,$E3,$2F,$87,$0E,$E4,$8E,$A3,$3F
	.dc.b	$7B,$8D,$5F,$0D,$F4,$09,$1C,$25,$D5,$88,$FE,$6D,$3F,$CF,$08,$FF
	.dc.b	$30,$D3,$DC,$FD,$4C,$3C,$4F,$A8,$F8,$27,$86,$78,$BD,$0F,$E6,$6F
	.dc.b	$F4,$8F,$86,$91,$FC,$D9,$AF,$E5,$67,$AF,$07,$DF,$8E,$C7,$BF,$6F
	.dc.b	$44,$7D,$FE,$CF,$F2,$E9,$8D,$BF,$A6,$7C,$37,$F5,$62,$5F,$F4,$47
	.dc.b	$E6,$3F,$2B,$3F,$69,$EE,$D7,$E5,$C7,$83,$67,$C3,$7E,$68,$91,$F2
	.dc.b	$FB,$EB,$FD,$7E,$3F,$B2,$91,$F9,$5A,$FB,$32,$FC,$54,$7D,$3D,$78
	.dc.b	$BF,$D9,$6B,$83,$CC,$FE,$B7,$F0,$DF,$DC,$08,$E0,$EA,$7A,$EF,$B3
	.dc.b	$F5,$0F,$54,$FF,$A0,$BC,$19,$1F,$87,$8F,$DA,$FC,$37,$F7,$C2,$47
	.dc.b	$F6,$A7,$E3,$BF,$4B,$5F,$8F,$7F,$C1,$23,$F2,$1F,$0D,$FB,$E2,$3F
	.dc.b	$8F,$F1,$2B,$FE,$03,$8F,$F0,$49,$1F,$49,$D6,$7D,$2F,$DF,$3A,$9F
	.dc.b	$DF,$23,$83,$CC,$FE,$FF,$F0,$DF,$E1,$42,$38,$3A,$F1,$FE,$2C,$9B
	.dc.b	$7F,$BF,$F8,$7E,$FC,$7F,$A5,$F8,$58,$7F,$AA,$3C,$69,$87,$88,$FC
	.dc.b	$21,$FE,$22,$21,$FE,$10,$3E,$3D,$76,$47,$EA,$4F,$C0,$1F,$5C,$87
	.dc.b	$C1,$3F,$6D,$E3,$FE,$B9,$FF,$6D,$6B,$FC,$8D,$BA,$7D,$46,$9F,$28
	.dc.b	$A7,$EA,$CC,$1F,$90,$30,$43,$AF,$C3,$7F,$1F,$23,$FC,$05,$F5,$E9
	.dc.b	$F1,$FC,$AF,$F6,$59,$F5,$AF,$F5,$1E,$5F,$FB,$38,$C7,$C3,$7F,$29
	.dc.b	$26,$1F,$C2,$DE,$6F,$FB,$1A,$B9,$BF,$0B,$3F,$EA,$7F,$0D,$FC,$B8
	.dc.b	$91,$FC,$67,$CF,$AF,$F5,$B8,$FE,$6C,$47,$EF,$E5,$E7,$7F,$90,$5F
	.dc.b	$9E,$75,$1C,$4F,$E3,$CF,$F3,$FF,$76,$47,$F2,$16,$BF,$23,$0F,$E9
	.dc.b	$D5,$E8,$8F,$B0,$7E,$96,$BF,$DD,$FE,$94,$8F,$E1,$D1,$FD,$23,$E1
	.dc.b	$BF,$A6,$12,$3F,$97,$3F,$11,$FE,$79,$5D,$6F,$F1,$51,$FD,$93,$E1
	.dc.b	$BF,$A8,$91,$FE,$79,$EA,$D7,$FB,$F4,$7F,$59,$23,$F1,$0F,$A0,$7C
	.dc.b	$43,$E8,$67,$FD,$AE,$37,$1C,$8F,$E8,$4F,$F6,$3F,$86,$91,$FD,$15
	.dc.b	$AF,$C5,$9F,$A0,$FF,$38,$E7,$63,$F8,$7B,$FF,$03,$43,$D7,$A7,$A7
	.dc.b	$2F,$F0,$9F,$C3,$7F,$71,$26,$1F,$D4,$DF,$8D,$5D,$23,$D4,$7D,$29
	.dc.b	$FF,$0D,$FC,$37,$F7,$62,$47,$F6,$1F,$6A,$BF,$E2,$38,$FE,$FC,$47
	.dc.b	$E3,$1F,$48,$EB,$3F,$4B,$95,$3B,$9E,$47,$F6,$77,$FC,$11,$F0,$D2
	.dc.b	$3F,$B5,$35,$F8,$C7,$D2,$3E,$36,$9F,$AB,$9F,$8B,$EE,$49,$FE,$AD
	.dc.b	$1F,$E1,$0F,$86,$FF,$0A,$12,$3F,$BB,$3F,$12,$BF,$11,$BF,$EF,$F2
	.dc.b	$FF,$1E,$FC,$37,$F8,$60,$97,$FE,$03,$F7,$AB,$FE,$45,$8F,$F1,$01
	.dc.b	$1F,$87,$B7,$CF,$7C,$E7,$E2,$0F,$E3,$BA,$A7,$84,$73,$3F,$C1,$4F
	.dc.b	$F8,$AF,$E1,$A4,$7F,$83,$5A,$FC,$43,$E8,$1D,$8F,$B9,$8F,$F8,$FE
	.dc.b	$7F,$13,$F0,$12,$7F,$BD,$47,$F8,$D3,$0F,$F1,$C1,$23,$FC,$34,$FF
	.dc.b	$93,$7E,$19,$D3,$FE,$57,$6C,$FF,$8B,$1A,$FF,$84,$3E,$F1,$F0,$7F
	.dc.b	$E4,$09,$93,$FE,$09,$7F,$CC,$12,$F8,$4C,$CF,$F9,$BE,$7F,$C9,$44
	.dc.b	$7F,$DB,$6F,$6D,$DA,$7F,$46,$94,$E6,$30,$D8,$FE,$97,$7E,$47,$C4
	.dc.b	$7F,$8D,$E3,$B3,$E1,$3F,$C7,$5F,$0D,$FE,$5B,$23,$83,$C7,$97,$C3
	.dc.b	$23,$FE,$4E,$AE,$CF,$F8,$06,$BD,$FF,$0D,$23,$C2,$7F,$81,$7E,$1F
	.dc.b	$FC,$D0,$47,$FC,$94,$FD,$1B,$EB,$1F,$E3,$CA,$F6,$2B,$EF,$97,$FA
	.dc.b	$53,$0F,$F3,$79,$27,$C9,$85,$7F,$D3,$91,$FE,$77,$23,$EC,$3E,$E6
	.dc.b	$1E,$0A,$F9,$F1,$F2,$B4,$E4,$FF,$6E,$AE,$2F,$43,$FC,$C5,$A7,$FA
	.dc.b	$0C,$8E,$0E,$C7,$AE,$FB,$69,$F6,$F2,$F5,$BF,$E9,$7B,$FC,$3E,$92
	.dc.b	$FF,$CA,$51,$FE,$8C,$C7,$FD,$20,$48,$FF,$39,$3F,$25,$F5,$90,$FC
	.dc.b	$B3,$F9,$BD,$E4,$7E,$6D,$FC,$AE,$B8,$8F,$C7,$1F,$E1,$45,$1F,$A5
	.dc.b	$24,$7B,$34,$F8,$EF,$E5,$33,$F6,$BF,$24,$EC,$7E,$72,$78,$BD,$8F
	.dc.b	$CD,$BF,$B1,$E2,$91,$FA,$26,$BB,$BF,$C1,$92,$BF,$5B,$87,$A7,$0F
	.dc.b	$1E,$9F,$EA,$1B,$FC,$B6,$F9,$D7,$F4,$FC,$DF,$BF,$24,$FE,$C1,$F9
	.dc.b	$4F,$D7,$3E,$55,$F4,$C3,$F8,$27,$44,$9F,$E0,$8F,$EF,$B8,$F1,$1E
	.dc.b	$78,$FE,$18,$47,$E4,$61,$FB,$8C,$FE,$3C,$7E,$76,$F8,$BF,$4B,$45
	.dc.b	$F1,$8E,$87,$F0,$6E,$AF,$E3,$04,$70,$76,$3E,$D5,$F6,$D7,$F1,$F1
	.dc.b	$F7,$6B,$F8,$AF,$24,$8F,$59,$FC,$7F,$B7,$F9,$21,$1F,$E2,$AF,$C6
	.dc.b	$D7,$F0,$39,$7D,$39,$FE,$09,$9F,$B2,$31,$1F,$C9,$FA,$3F,$7A,$FB
	.dc.b	$73,$F9,$ED,$3F,$4E,$53,$F0,$0E,$9F,$90,$68,$A6,$8F,$E5,$3E,$0F
	.dc.b	$F3,$B2,$38,$3E,$21,$FC,$2F,$AF,$F9,$A9,$E7,$9F,$E6,$FD,$72,$7F
	.dc.b	$03,$C1,$FC,$BB,$C3,$FE,$8C,$48,$FE,$6E,$FC,$38,$FD,$E5,$F4,$9E
	.dc.b	$27,$EA,$FB,$3F,$A6,$11,$FE,$55,$DB,$5F,$CF,$47,$F5,$32,$3E,$8E
	.dc.b	$25,$75,$8F,$CB,$1F,$C9,$E7,$6F,$23,$FA,$5B,$FD,$77,$C6,$91,$FD
	.dc.b	$39,$AF,$C9,$3F,$AE,$4F,$5E,$0F,$BB,$A7,$95,$FD,$6F,$81,$FD,$70
	.dc.b	$C6,$7F,$D4,$BC,$7F,$ED,$44,$C3,$FA,$DB,$F1,$5F,$EA,$B8,$75,$9F
	.dc.b	$B3,$7F,$8A,$F2,$FF,$B8,$12,$BF,$A7,$78,$55,$FE,$05,$1F,$DD,$C8
	.dc.b	$FC,$5B,$E8,$FE,$80,$FE,$F1,$1A,$78,$9F,$DB,$9F,$F0,$07,$9D,$23
	.dc.b	$FB,$8B,$5F,$89,$7F,$C3,$2B,$AC,$7F,$1F,$3A,$9F,$D8,$FC,$4F,$EF
	.dc.b	$E7,$9E,$7F,$4B,$E8,$FF,$83,$49,$3F,$DF,$9F,$88,$FE,$4C,$FE,$D9
	.dc.b	$5F,$A9,$F4,$FF,$C2,$44,$8F,$EE,$1E,$45,$7F,$92,$47,$F8,$64,$8F
	.dc.b	$C3,$39,$9F,$92,$3F,$C1,$2E,$9E,$27,$F8,$41,$FF,$11,$7A,$D2,$3F
	.dc.b	$C2,$6D,$7E,$1E,$BF,$9E,$7A,$9F,$92,$C7,$F8,$1D,$7F,$88,$3C,$99
	.dc.b	$3F,$E2,$98,$FF,$16,$7B,$1F,$E2,$F2,$47,$F8,$81,$F8,$7C,$BF,$9E
	.dc.b	$8E,$B7,$F9,$27,$FC,$65,$ED,$FF,$8D,$48,$FF,$84,$FC,$CA,$FF,$47
	.dc.b	$8F,$F1,$E1,$1F,$88,$7E,$78,$F8,$9D,$A8,$D4,$71,$3F,$C6,$6F,$F9
	.dc.b	$17,$D5,$91,$FE,$36,$6B,$F1,$0F,$CE,$9E,$8D,$3F,$3A,$7F,$8D,$A7
	.dc.b	$FC,$87,$E7,$49,$F5,$9F,$E4,$D9,$FF,$29,$11,$FF,$21,$B5,$FE,$C4
	.dc.b	$D7,$FA,$33,$5F,$F2,$1B,$5F,$EF,$4F,$F6,$F4,$3E,$3D,$F0,$D9,$8C
	.dc.b	$36,$33,$1C,$5A,$FF,$79,$6B,$FE,$08,$6B,$FE,$57,$6B,$E0,$6B,$D8
	.dc.b	$E9,$1D,$C8,$F7,$F8,$6F,$11,$1C,$10,$F1,$39,$69,$93,$C8,$E5,$A7
	.dc.b	$AF,$E1,$BD,$C4,$70,$43,$DC,$E5,$A6,$4F,$80,$E5,$A7,$C7,$F8,$6F
	.dc.b	$94,$47,$04,$3E,$53,$96,$99,$3E,$63,$96,$9F,$4F,$E1,$BE,$B1,$1C
	.dc.b	$10,$FA,$CE,$5A,$64,$FB,$0E,$5A,$7D,$FF,$86,$FC,$11,$1C,$1F,$8B
	.dc.b	$E1,$7A,$DF,$E9,$5E,$2F,$F2,$1E,$8F,$6B,$1B,$7E,$0D,$FF,$51,$D6
	.dc.b	$30,$FC,$27,$C3,$57,$F2,$8D,$B3,$9F,$BD,$FF,$2A,$E3,$DA,$3F,$2C
	.dc.b	$4A,$FC,$B3,$F9,$9C,$39,$12,$3F,$D3,$18,$75,$9E,$11,$F9,$B6,$BF
	.dc.b	$9B,$F8,$7F,$D1,$11,$FD,$13,$D5,$FD,$5D,$FC,$0B,$ED,$1F,$EC,$5B
	.dc.b	$EC,$EB,$2F,$F5,$D6,$2F,$77,$D8,$6A,$58,$7C,$AF,$87,$FD,$91,$23
	.dc.b	$F6,$4F,$85,$DD,$FC,$D7,$B1,$FA,$FC,$7E,$7B,$83,$F8,$07,$C3,$FC
	.dc.b	$D2,$39,$C5,$EF,$62,$3F,$7D,$3F,$C0,$C8,$FF,$03,$AF,$7E,$3F,$06
	.dc.b	$3E,$69,$DF,$65,$70,$8E,$67,$F0,$B7,$F8,$77,$C3,$C8,$FE,$1C,$D7
	.dc.b	$F8,$75,$F6,$CB,$DD,$1F,$34,$ED,$5F,$85,$99,$18,$3C,$5F,$0F,$FC
	.dc.b	$6C,$8F,$F1,$B7,$E5,$63,$E9,$8F,$96,$7C,$93,$D6,$60,$FE,$51,$F0
	.dc.b	$FF,$95,$23,$F2,$7E,$12,$BF,$C9,$63,$F9,$51,$1F,$90,$7C,$5B,$F6
	.dc.b	$3F,$EF,$8E,$0E,$2F,$F2,$B9,$E0,$F3,$3F,$99,$7C,$3F,$F3,$52,$38
	.dc.b	$3A,$9E,$7D,$7F,$64,$F7,$3D,$78,$FE,$DA,$BF,$59,$F0,$B2,$7E,$C4
	.dc.b	$7E,$C7,$E1,$FF,$A1,$12,$3F,$A1,$3F,$23,$6F,$A7,$7F,$24,$FF,$58
	.dc.b	$E5,$FB,$9F,$87,$97,$F8,$37,$F7,$46,$1F,$E9,$4F,$F4,$F2,$3E,$A9
	.dc.b	$E9,$5D,$8F,$F7,$45,$71,$9F,$EA,$15,$C1,$E6,$7F,$57,$C7,$FA,$D1
	.dc.b	$1C,$1D,$6B,$A3,$F7,$CF,$C6,$1F,$26,$7F,$88,$61,$23,$D4,$7E,$77
	.dc.b	$5F,$EC,$C4,$7F,$B3,$3F,$15,$E9,$8F,$A9,$FB,$C7,$D3,$3E,$B6,$BF
	.dc.b	$82,$5C,$4D,$BF,$85,$C7,$F6,$D9,$FE,$00,$D3,$EF,$94,$FD,$39,$83
	.dc.b	$F4,$E6,$48,$60,$FE,$ED,$C5,$FD,$FC,$8F,$F7,$F7,$D4,$BE,$45,$EA
	.dc.b	$7F,$37,$6C,$7A,$F1,$F7,$3E,$AE,$0F,$E7,$07,$B7,$87,$F9,$04,$7F
	.dc.b	$82,$F9,$BF,$C1,$C4,$7E,$A1,$FE,$16,$BE,$AF,$EC,$5B,$7D,$AB,$FA
	.dc.b	$1F,$07,$F8,$3E,$BF,$A4,$1F,$D1,$F8,$EB,$FE,$19,$3F,$C3,$84,$3B
	.dc.b	$9E,$DC,$BD,$87,$9D,$D1,$FE,$1F,$29,$FE,$2A,$3C,$DC,$7F,$E2,$9C
	.dc.b	$7F,$B1,$C7,$F8,$BD,$AE,$CD,$9E,$07,$D7,$7C,$DB,$1F,$88,$8F,$63
	.dc.b	$FC,$D7,$97,$FB,$11,$EC,$7E,$27,$81,$E1,$23,$E1,$7E,$CD,$74,$8F
	.dc.b	$71,$F5,$EB,$C1,$E2,$7D,$B2,$47,$F4,$3E,$7A,$FA,$4F,$C1,$47,$BC
	.dc.b	$8F,$AF,$2F,$92,$F5,$3E,$D4,$69,$E5,$68,$C4,$7C,$58,$78,$FF,$2C
	.dc.b	$91,$F2,$DF,$9D,$5D,$A3,$F0,$8F,$EA,$67,$A4,$7F,$5B,$E9,$FF,$04
	.dc.b	$1E,$AD,$D9,$7C,$FF,$27,$EC,$12,$BE,$C3,$D9,$FE,$D1,$7F,$24,$F9
	.dc.b	$5C,$1F,$CB,$23,$F2,$9E,$67,$E8,$C8,$FF,$6B,$EC,$AF,$C0,$8F,$C3
	.dc.b	$11,$FC,$36,$7E,$BC,$BC,$B1,$FA,$18,$F6,$4E,$E3,$91,$F8,$D7,$F2
	.dc.b	$1E,$7C,$8F,$C8,$35,$F9,$F3,$D7,$3F,$C7,$3D,$23,$FC,$01,$DD,$FE
	.dc.b	$2D,$3C,$F8,$78,$A3,$F8,$07,$A3,$F9,$D2,$3F,$9D,$7E,$33,$F4,$8F
	.dc.b	$8F,$5F,$4C,$F0,$3E,$B3,$FC,$1D,$E0,$FD,$97,$F0,$11,$C9,$A7,$C4
	.dc.b	$21,$EA,$29,$A3,$04,$30,$7E,$2F,$D6,$FD,$A1,$1F,$DA,$3E,$8B,$F7
	.dc.b	$76,$BC,$3F,$F2,$71,$E8,$AF,$D3,$9F,$E0,$18,$78,$9F,$E5,$33,$5C
	.dc.b	$3F,$ED,$79,$FC,$EF,$9D,$FC,$0C,$8F,$F0,$3F,$63,$F9,$41,$1F,$F1
	.dc.b	$87,$8F,$5F,$E3,$D1,$72,$F2,$F9,$19,$9F,$D3,$4F,$F0,$F2,$3D,$48
	.dc.b	$7E,$B9,$CB,$4C,$9C,$8A,$7F,$11,$F6,$FF,$8D,$11,$FE,$34,$F2,$AF
	.dc.b	$C2,$7C,$93,$E0,$3F,$CF,$0D,$AF,$D3,$DF,$A7,$93,$FC,$BF,$E6,$7F
	.dc.b	$9E,$4C,$4F,$F3,$DF,$80,$FE,$50,$47,$F9,$43,$DE,$7F,$97,$5F,$59
	.dc.b	$FF,$19,$37,$1F,$CB,$6C,$C7,$FA,$A7,$C0,$CB,$FD,$CB,$FE,$8D,$FB
	.dc.b	$EF,$CE,$5F,$08,$FE,$70,$49,$FE,$70,$F8,$5F,$F7,$3E,$1D,$F1,$F1
	.dc.b	$E9,$EA,$9F,$EC,$95,$C5,$FE,$3A,$D7,$F8,$E7,$A5,$FD,$14,$8F,$F4
	.dc.b	$57,$B2,$BE,$CF,$82,$BB,$66,$BF,$57,$A5,$2B,$DA,$7F,$16,$F8,$6F
	.dc.b	$EA,$44,$7F,$A9,$3D,$B0,$EB,$1D,$F0,$EB,$F1,$1F,$DE,$C9,$1F,$E8
	.dc.b	$7D,$5E,$A6,$23,$F9,$CC,$FF,$5F,$23,$FD,$14,$E8,$7F,$A3,$EF,$B9
	.dc.b	$E0,$CF,$EB,$47,$F2,$C9,$E2,$F4,$3F,$B4,$3F,$DA,$FE,$1E,$47,$F6
	.dc.b	$B6,$BF,$D4,$EB,$AE,$3F,$92,$3D,$71,$E9,$F5,$BF,$00,$7A,$78,$3F
	.dc.b	$0F,$F1,$1F,$DD,$C9,$3F,$DD,$DF,$53,$EA,$3F,$1F,$5D,$5F,$CE,$9F
	.dc.b	$D1,$3E,$0F,$FC,$44,$47,$E4,$7B,$55,$FE,$BF,$1F,$E0,$52,$3F,$1F
	.dc.b	$1F,$4B,$F1,$EB,$EF,$CE,$A3,$89,$FE,$0B,$7C,$B2,$F8,$8A,$49,$E8
	.dc.b	$EA,$BD,$11,$EA,$3D,$27,$F8,$AE,$BF,$49,$ED,$C8,$F6,$1F,$67,$E2
	.dc.b	$3B,$91,$F4,$3D,$5F,$41,$0F,$E3,$25,$CF,$F8,$3D,$F8,$9F,$E9,$9C
	.dc.b	$2F,$77,$CB,$5F,$E2,$5C,$FF,$7C,$DB,$EC,$FD,$F6,$B9,$3E,$53,$87
	.dc.b	$F6,$A6,$08,$7E,$1F,$E1,$FE,$49,$1F,$A4,$F6,$C3,$F8,$EF,$C2,$FF
	.dc.b	$1C,$3D,$1D,$3D,$2B,$FD,$85,$93,$E0,$3F,$86,$1A,$8F,$E0,$FF,$11
	.dc.b	$F5,$88,$FD,$F7,$B5,$7F,$1C,$BE,$D3,$F1,$E3,$F8,$FF,$C4,$7F,$2C
	.dc.b	$24,$7E,$CF,$E0,$EB,$FA,$18,$FC,$19,$1F,$7E,$5E,$A7,$F6,$57,$E5
	.dc.b	$7F,$87,$4E,$DE,$47,$E4,$DF,$C6,$FC,$2C,$8F,$CB,$35,$FD,$84,$FA
	.dc.b	$1F,$E6,$6F,$5D,$95,$FC,$BF,$E1,$65,$7F,$03,$8F,$E8,$FF,$11,$F9
	.dc.b	$A2,$47,$E8,$DF,$CE,$DF,$5C,$7D,$53,$F9,$93,$B9,$EC,$21,$F5,$3E
	.dc.b	$1F,$C8,$FE,$2E,$79,$BF,$AF,$6B,$EA,$30,$43,$04,$39,$19,$3F,$2D
	.dc.b	$F1,$1F,$B2,$23,$FB,$B7,$AD,$7F,$66,$E3,$FE,$CA,$7A,$F6,$FC,$05
	.dc.b	$7F,$2D,$AF,$EA,$86,$5F,$E9,$F8,$7E,$F8,$8F,$F0,$77,$AC,$FF,$64
	.dc.b	$AE,$B3,$F0,$AF,$FB,$AE,$7F,$DD,$09,$1F,$CE,$F1,$AF,$F2,$98,$FE
	.dc.b	$16,$47,$F8,$A5,$74,$7F,$B7,$BE,$57,$FA,$B4,$6D,$E4,$7F,$1B,$7F
	.dc.b	$8A,$ED,$23,$F8,$F3,$5F,$E7,$13,$D6,$FE,$34,$7A,$70,$FE,$FF,$7F
	.dc.b	$DD,$F3,$93,$FD,$2E,$3F,$C1,$FC,$1F,$C8,$89,$1F,$CA,$DF,$E4,$B8
	.dc.b	$75,$BF,$8D,$3D,$4F,$EF,$71,$96,$9F,$D7,$F7,$F8,$0F,$F1,$38,$E6
	.dc.b	$FC,$42,$1E,$A2,$9B,$29,$E0,$34,$43,$F8,$F7,$2F,$F3,$A2,$3F,$D0
	.dc.b	$DE,$EF,$F0,$ED,$FF,$C3,$07,$9D,$FF,$03,$4F,$F8,$6F,$86,$47,$F6
	.dc.b	$F8,$FF,$19,$F4,$7F,$46,$24,$7F,$4F,$6B,$FD,$A5,$AF,$F3,$16,$BF
	.dc.b	$E3,$06,$BF,$E2,$27,$FA,$8F,$54,$8F,$EB,$4D,$7F,$BA,$35,$FE,$78
	.dc.b	$D7,$FC,$76,$D7,$FC,$5A,$D7,$FC,$45,$DD,$FD,$78,$8F,$F6,$76,$BF
	.dc.b	$E0,$16,$BF,$D2,$5A,$FF,$92,$1A,$FF,$8E,$9F,$ED,$3E,$04,$8F,$EE
	.dc.b	$4D,$7F,$C1,$8D,$7F,$AA,$35,$FF,$29,$B5,$FF,$22,$B5,$FF,$1D,$69
	.dc.b	$FD,$D8,$8F,$F7,$F7,$FC,$1B,$CC,$E7,$FF,$2A,$74,$4A,$FF,$18,$47
	.dc.b	$F9,$5F,$A6,$47,$F8,$CD,$FF,$2D,$75,$A2,$72,$FF,$AD,$FE,$1D,$E3
	.dc.b	$7F,$86,$9B,$3F,$E3,$56,$BF,$EF,$B6,$BC,$CE,$99,$79,$5F,$F2,$16
	.dc.b	$5E,$57,$FB,$0E,$5E,$47,$FB,$1E,$57,$1F,$EF,$B9,$73,$7C,$9E,$47
	.dc.b	$8C,$95,$E3,$6B,$E3,$6B,$E3,$19,$23,$31,$E7,$8C,$3C,$C8,$7B,$7C
	.dc.b	$9F,$84,$47,$E1,$35,$D1,$82,$18,$21,$E3,$84,$71,$8C,$79,$7F,$3C
	.dc.b	$8F,$CF,$7E,$97,$9B,$F0,$F7,$F4,$36,$7E,$D5,$FF,$B4,$5F,$B7,$3F
	.dc.b	$4F,$91,$E0,$EE,$7E,$1D,$DB,$E1,$0C,$FF,$05,$3B,$B7,$85,$B8,$F1
	.dc.b	$D7,$E0,$DA,$FE,$0D,$AF,$BD,$AF,$A0,$E0,$43,$81,$F8,$DF,$43,$F2
	.dc.b	$04,$70,$6E,$D9,$7B,$7C,$4F,$31,$E5,$9F,$82,$7C,$4B,$F5,$78,$B2
	.dc.b	$32,$7C,$22,$E6,$F4,$BF,$40,$48,$FD,$03,$FA,$3F,$57,$E4,$12,$39
	.dc.b	$EA,$F2,$71,$1F,$2F,$3F,$C7,$56,$34,$F4,$CF,$EC,$6D,$E5,$F0,$8F
	.dc.b	$C3,$57,$77,$F6,$0D,$7F,$60,$D7,$66,$88,$68,$A7,$E4,$7D,$8F,$DE
	.dc.b	$11,$FD,$E3,$D6,$FC,$B1,$FE,$6D,$F3,$3F,$04,$7A,$71,$F5,$C7,$C0
	.dc.b	$F3,$BF,$0C,$65,$FC,$9F,$B7,$FC,$24,$8F,$F0,$97,$F8,$67,$BB,$F9
	.dc.b	$12,$47,$CA,$F4,$6B,$F2,$4F,$CE,$C6,$31,$F2,$EB,$FE,$C2,$9D,$1F
	.dc.b	$BE,$9E,$AF,$F1,$76,$BF,$C5,$DA,$FE,$CD,$AF,$F0,$36,$BE,$03,$91
	.dc.b	$FC,$87,$DF,$FE,$48,$47,$06,$8F,$1C,$BD,$3F,$D5,$1E,$9C,$FB,$4F
	.dc.b	$E2,$3D,$5F,$D7,$19,$7F,$77,$F0,$3F,$CC,$48,$FF,$31,$7F,$9A,$7C
	.dc.b	$27,$EE,$49,$1F,$92,$F6,$2B,$F1,$39,$93,$8D,$7E,$4F,$B3,$AA,$FE
	.dc.b	$29,$3D,$1F,$E7,$CD,$7F,$9F,$35,$D9,$A2,$1A,$29,$FC,$7F,$E1,$BF
	.dc.b	$A5,$11,$FE,$94,$F5,$AF,$B7,$ED,$FF,$0F,$3C,$DA,$7F,$35,$BF,$E2
	.dc.b	$5E,$E4,$C3,$F6,$F1,$FC,$67,$E1,$FF,$AB,$12,$3F,$AB,$3F,$D6,$73
	.dc.b	$FE,$3C,$48,$FD,$BF,$BF,$5F,$8D,$17,$2F,$C6,$FC,$06,$67,$F9,$C4
	.dc.b	$F3,$7F,$B1,$35,$FE,$C4,$D7,$66,$88,$68,$A7,$F4,$3F,$87,$FE,$DC
	.dc.b	$47,$FB,$73,$D6,$BF,$3F,$F0,$5F,$CC,$CF,$36,$9F,$D6,$6F,$F9,$B7
	.dc.b	$C1,$CC,$3F,$8D,$C7,$F3,$DF,$87,$FE,$F4,$48,$FE,$F4,$FF,$7D,$F8
	.dc.b	$7F,$E8,$44,$8F,$E3,$7F,$0D,$5F,$C5,$45,$CB,$F8,$2F,$DF,$FF,$6D
	.dc.b	$7F,$C1,$11,$CE,$7F,$C0,$CD,$7F,$C0,$CD,$7F,$B1,$35,$FE,$E4,$D7
	.dc.b	$FA,$DB,$5F,$EF,$3F,$11,$FE,$13,$23,$FE,$13,$7B,$4F,$91,$FE,$BD
	.dc.b	$47,$D1,$9F,$6A,$FE,$83,$2B,$D9,$A8,$FE,$D6,$7F,$88,$48,$7F,$88
	.dc.b	$4F,$F0,$61,$FE,$1F,$F9,$56,$3F,$DB,$5B,$1F,$D3,$9A,$FF,$16,$A6
	.dc.b	$5B,$A9,$3B,$7D,$3B,$62,$3B,$DF,$27,$FC,$38,$D7,$C5,$3E,$47,$FC
	.dc.b	$79,$19,$6C,$65,$46,$5B,$19,$8C,$11,$C1,$9D,$64,$F7,$94,$62,$3C
	.dc.b	$F3,$F4,$B5,$91,$ED,$21,$F0,$F8,$7E,$B9,$EF,$DB,$FA,$8E,$0D,$FC
	.dc.b	$84,$AF,$03,$E7,$E0,$FA,$04,$8F,$BB,$CB,$5F,$E8,$B4,$9C,$5F,$E6
	.dc.b	$B8,$F5,$5F,$32,$7A,$3E,$D6,$BF,$49,$AE,$CD,$10,$D1,$4F,$AF,$C7
	.dc.b	$F3,$48,$FC,$97,$AD,$7E,$B3,$A3,$F3,$27,$9B,$8F,$F1,$D7,$F9,$CE
	.dc.b	$C9,$87,$E3,$63,$F2,$1C,$DF,$64,$91,$F5,$5F,$B9,$CF,$F9,$52,$47
	.dc.b	$E8,$BB,$6B,$F1,$A2,$E5,$FB,$EE,$FC,$CF,$E4,$27,$9B,$F8,$16,$BF
	.dc.b	$96,$6B,$B3,$44,$34,$53,$F3,$7D,$5F,$91,$23,$F8,$D7,$AD,$7F,$10
	.dc.b	$F0,$BF,$7A,$79,$B4,$FD,$95,$FF,$00,$F0,$E6,$1F,$B1,$8F,$DA,$76
	.dc.b	$FE,$74,$91,$F9,$97,$F4,$1D,$DF,$B9,$24,$7F,$06,$F1,$AB,$F8,$A8
	.dc.b	$B9,$7F,$22,$F2,$B3,$3F,$B4,$9E,$6F,$E9,$5A,$FE,$E9,$AE,$CD,$10
	.dc.b	$D1,$4F,$DF,$F8,$3F,$B5,$23,$FB,$17,$AD,$7F,$34,$F2,$7F,$90,$9E
	.dc.b	$6D,$3F,$8B,$5F,$F2,$4F,$36,$61,$FC,$56,$3F,$8C,$7C,$3F,$F0,$12
	.dc.b	$47,$EF,$5F,$E0,$9E,$37,$F1,$D2,$47,$F2,$AF,$3E,$BF,$AE,$8B,$97
	.dc.b	$F4,$6F,$43,$33,$FC,$62,$79,$BF,$C2,$5A,$FF,$1E,$6B,$B3,$44,$34
	.dc.b	$53,$F9,$1F,$93,$FC,$64,$8F,$F1,$57,$AD,$7F,$58,$F5,$BF,$A2,$9E
	.dc.b	$6D,$3F,$9D,$5F,$FA,$97,$F0,$0F,$F3,$0D,$7F,$98,$9B,$3A,$C3,$F9
	.dc.b	$E7,$99,$FC,$9C,$8F,$F2,$37,$F9,$57,$9B,$FD,$0C,$91,$FE,$90,$A7
	.dc.b	$B1,$88,$FE,$2B,$37,$0F,$07,$B3,$99,$FE,$7F,$3D,$1F,$E6,$0D,$7F
	.dc.b	$A2,$B5,$D9,$A2,$1A,$29,$FD,$2B,$D0,$FE,$80,$47,$F9,$E3,$D6,$BE
	.dc.b	$17,$B5,$F1,$4F,$36,$7F,$D8,$6F,$E4,$7B,$73,$0F,$EC,$11,$FD,$8F
	.dc.b	$D3,$FE,$98,$48,$FE,$90,$FF,$4F,$F5,$3F,$B4,$12,$3E,$37,$BD,$5F
	.dc.b	$E7,$31,$72,$FA,$9E,$AE,$67,$FB,$1C,$F3,$7F,$AA,$35,$FE,$D2,$D7
	.dc.b	$66,$88,$68,$A7,$F6,$EF,$5B,$FB,$21,$1F,$EC,$0F,$5A,$FC,$37,$C0
	.dc.b	$7E,$24,$F3,$69,$FE,$05,$BF,$C6,$7C,$1C,$C3,$FC,$09,$1F,$E0,$7F
	.dc.b	$67,$FB,$81,$23,$FB,$63,$FD,$DF,$E5,$DE,$DF,$FB,$E5,$B3,$FE,$DA
	.dc.b	$6B,$FE,$A2,$E5,$7C,$1E,$EA,$4E,$DF,$CF,$7C,$36,$23,$FC,$37,$7C
	.dc.b	$91,$03,$97,$0D,$76,$D7,$73,$E9,$23,$FE,$15,$8D,$12,$33,$1E,$2F
	.dc.b	$7B,$E0,$11,$E0,$EB,$6F,$DA,$7C,$3F,$ED,$CF,$53,$FE,$D8,$BF,$86
	.dc.b	$FE,$E7,$E1,$FF,$78,$6B,$7F,$4F,$C1,$7C,$D2,$63,$E5,$7E,$87,$C1
	.dc.b	$7C,$F2,$47,$EE,$7E,$1E,$BF,$ED,$0C,$53,$8C,$7F,$84,$7C,$3E,$AB
	.dc.b	$E5,$1F,$2A,$7A,$BF,$0D,$AF,$D2,$6B,$C2,$BA,$BB,$21,$B2,$9B,$F8
	.dc.b	$6F,$C4,$11,$FA,$0F,$6B,$FE,$39,$F0,$FF,$C7,$CF,$46,$7F,$73,$0F
	.dc.b	$E4,$55,$2F,$F1,$D1,$F9,$1A,$FC,$A1,$23,$EE,$3F,$97,$C7,$F2,$C4
	.dc.b	$8F,$E4,$13,$5F,$91,$17,$0F,$E5,$D9,$E6,$7E,$04,$F3,$7F,$10,$D7
	.dc.b	$F2,$ED,$76,$68,$86,$8A,$7C,$BC,$FF,$58,$47,$F2,$8F,$5A,$FE,$81
	.dc.b	$B7,$F4,$33,$CD,$9F,$E8,$2F,$FA,$36,$F3,$0F,$D9,$47,$ED,$76,$FD
	.dc.b	$C1,$23,$F4,$0F,$EE,$F7,$FD,$D1,$23,$F5,$F8,$B0,$FD,$87,$E0,$2E
	.dc.b	$4D,$57,$E3,$EE,$DE,$E3,$F6,$F4,$E8,$DC,$7E,$26,$BC,$0F,$EC,$5A
	.dc.b	$FF,$00,$6B,$B3,$44,$34,$53,$F2,$9C,$7F,$C5,$48,$FE,$F5,$EB,$7E
	.dc.b	$58,$5B,$9B,$F8,$64,$E6,$31,$3F,$C6,$EF,$DB,$97,$C2,$EB,$DB,$67
	.dc.b	$47,$F2,$1E,$6F,$E4,$C4,$7F,$87,$3F,$CA,$7A,$3F,$94,$12,$3E,$7C
	.dc.b	$FC,$EE,$FC,$47,$EE,$E6,$E5,$E5,$ED,$CC,$FE,$CE,$7A,$3F,$C6,$DA
	.dc.b	$FF,$2A,$6B,$B3,$44,$34,$53,$F7,$DD,$7F,$CF,$C8,$FF,$27,$7A,$D7
	.dc.b	$C7,$F0,$BE,$1C,$FE,$01,$F3,$E3,$FC,$D2,$F3,$3F,$43,$C5,$BB,$1D
	.dc.b	$C7,$F3,$8E,$EF,$E9,$C4,$7F,$9B,$35,$FE,$8A,$D7,$F8,$A3,$5F,$E8
	.dc.b	$6D,$7A,$1F,$CF,$FB,$FF,$AD,$11,$C1,$0F,$EA,$2E,$5A,$64,$FE,$44
	.dc.b	$E5,$A7,$F4,$BF,$03,$FB,$31,$1C,$10,$FE,$BA,$E5,$A6,$4F,$E5,$CE
	.dc.b	$5A,$7F,$57,$F0,$7F,$B9,$11,$C1,$0F,$ED,$2E,$5A,$64,$FE,$74,$E5
	.dc.b	$A7,$F6,$3F,$0B,$FB,$F1,$1C,$10,$FE,$EA,$E5,$A6,$4F,$E8,$CE,$5A
	.dc.b	$7F,$6F,$F0,$FF,$C1,$44,$70,$43,$FC,$02,$E5,$A6,$4F,$EA,$4E,$5A
	.dc.b	$7F,$7B,$D3,$FC,$2C,$47,$07,$F6,$03,$FC,$17,$F3,$2F,$1B,$F2,$AD
	.dc.b	$8F,$D9,$9A,$8F,$CB,$F8,$DA,$64,$68,$4F,$15,$C3,$FC,$2F,$C9,$DC
	.dc.b	$78,$67,$4F,$F7,$26,$BF,$E1,$D6,$BE,$13,$D0,$76,$AA,$D3,$31,$C1
	.dc.b	$C1,$DE,$37,$19,$27,$91,$82,$4F,$27,$73,$FB,$6F,$36,$4F,$BE,$BD
	.dc.b	$D8,$7F,$2B,$6E,$5F,$D1,$BC,$BF,$DE,$9F,$02,$75,$E5,$F8,$89,$7E
	.dc.b	$87,$CB,$E6,$7D,$22,$47,$CD,$E5,$7A,$18,$8F,$9B,$8F,$C7,$AC,$61
	.dc.b	$FC,$E7,$D4,$D5,$7C,$D9,$EA,$FC,$66,$BF,$4D,$AE,$CD,$10,$D1,$4F
	.dc.b	$B1,$E7,$FC,$C2,$3F,$49,$EB,$5F,$D2,$FD,$2F,$E3,$E7,$9B,$0F,$45
	.dc.b	$7B,$7D,$7F,$E4,$66,$5F,$CA,$FA,$5F,$60,$8F,$DE,$7E,$DF,$A7,$F9
	.dc.b	$82,$47,$C7,$F6,$EB,$F1,$8F,$CF,$C6,$31,$FE,$BD,$EC,$EA,$7F,$25
	.dc.b	$3D,$1F,$C5,$B5,$FC,$C3,$5D,$9A,$21,$A2,$9F,$9D,$F5,$BF,$22,$47
	.dc.b	$F2,$EF,$5A,$FE,$E1,$EF,$FF,$44,$3C,$D9,$FE,$8E,$FF,$A3,$FC,$1C
	.dc.b	$C3,$F2,$51,$FB,$8F,$67,$F3,$A4,$8F,$D1,$BF,$A0,$F6,$BF,$7A,$48
	.dc.b	$FC,$67,$C1,$D7,$E3,$1F,$BA,$8C,$65,$FE,$07,$F8,$1F,$F3,$3B,$F9
	.dc.b	$F8,$E9,$3F,$B0,$6B,$FB,$D6,$BF,$BD,$6B,$FB,$D6,$BF,$C4,$9A,$FF
	.dc.b	$0A,$F7,$7F,$6E,$47,$F7,$AF,$69,$F2,$7C,$47,$F6,$CB,$3E,$8D,$3B
	.dc.b	$5F,$F1,$59,$7C,$46,$CD,$47,$F2,$2F,$80,$FE,$0A,$47,$F8,$5B,$FC
	.dc.b	$23,$E0,$7F,$92,$92,$3A,$7C,$3D,$7E,$4F,$7A,$71,$9F,$C9,$F8,$7D
	.dc.b	$57,$F1,$F9,$E8,$FF,$18,$6B,$FC,$95,$AE,$CD,$10,$D1,$4F,$E5,$5F
	.dc.b	$09,$FC,$78,$8F,$F2,$47,$AD,$7D,$BF,$87,$F9,$07,$9B,$5F,$E6,$17
	.dc.b	$F2,$FE,$1E,$61,$FC,$7E,$3F,$A2,$ED,$FC,$A8,$91,$FC,$C1,$FE,$5B
	.dc.b	$3F,$D2,$89,$1F,$46,$2B,$F1,$A2,$E1,$F8,$DC,$73,$3F,$D0,$E7,$9B
	.dc.b	$FC,$F1,$AF,$F4,$96,$BB,$34,$43,$45,$3F,$A7,$63,$FD,$08,$8F,$F4
	.dc.b	$87,$AD,$7E,$7F,$0F,$C6,$1E,$6C,$7F,$AA,$5F,$E3,$F5,$98,$7F,$43
	.dc.b	$8F,$ED,$39,$FF,$4E,$24,$7F,$54,$7F,$A8,$EB,$FD,$B8,$91,$F9,$5E
	.dc.b	$0A,$FE,$2A,$2E,$5F,$B1,$E1,$CC,$FF,$67,$9E,$6F,$F6,$06,$BF,$DB
	.dc.b	$5A,$EC,$D1,$0D,$14,$FE,$E5,$C1,$FD,$98,$8F,$F6,$C7,$AD,$7F,$03
	.dc.b	$E3,$FD,$81,$E6,$D3,$FB,$C5,$FE,$CF,$9E,$61,$FD,$9E,$3F,$C1,$BC
	.dc.b	$7F,$DC,$89,$1F,$DE,$1F,$EE,$AF,$F8,$48,$87,$EE,$79,$6B,$FA,$E2
	.dc.b	$E3,$FC,$57,$AB,$31,$FE,$0B,$39,$9F,$E0,$42,$9F,$E1,$12,$9B,$34
	.dc.b	$43,$45,$04,$1E,$63,$9F,$FC,$26,$D8,$FF,$0B,$B5,$FF,$18,$35,$F0
	.dc.b	$1F,$CC,$39,$9D,$DF,$C6,$CF,$6E,$BF,$E2,$39,$FE,$3D,$DF,$1F,$F0
	.dc.b	$91,$FE,$8C,$B7,$57,$88,$91,$E2,$7C,$BD,$7B,$24,$7F,$25,$F0,$EB
	.dc.b	$FD,$0A,$93,$8D,$7F,$A1,$78,$FA,$AF,$84,$7C,$29,$EA,$FC,$06,$BF
	.dc.b	$01,$AF,$0A,$EA,$EC,$86,$CA,$6F,$BB,$E6,$91,$F9,$AF,$6B,$FE,$AD
	.dc.b	$E3,$FF,$43,$3D,$1A,$7D,$7B,$EF,$7F,$D7,$BF,$02,$FF,$71,$5F,$FB
	.dc.b	$8D,$B5,$BC,$7E,$67,$83,$F7,$09,$7F,$71,$FB,$FE,$17,$D1,$24,$7F
	.dc.b	$98,$ED,$E5,$E2,$3E,$64,$DC,$7C,$1E,$66,$67,$E2,$4F,$47,$F1,$2D
	.dc.b	$7F,$12,$D7,$66,$88,$68,$A7,$CF,$F1,$7F,$2A,$47,$F2,$AF,$5A,$F8
	.dc.b	$5E,$6F,$FB,$5B,$13,$E7,$D3,$E5,$61,$9D,$66,$3F,$E8,$15,$1F,$93
	.dc.b	$F1,$FF,$46,$49,$FD,$1B,$FA,$6F,$23,$F3,$04,$8F,$91,$E7,$57,$F2
	.dc.b	$51,$73,$FA,$DE,$7E,$67,$F1,$53,$CD,$FD,$7B,$5F,$D7,$B5,$D9,$A2
	.dc.b	$1A,$29,$F9,$5F,$33,$F7,$64,$7F,$76,$F5,$AF,$C4,$7A,$7F,$1C,$F3
	.dc.b	$69,$F8,$FB,$F9,$5E,$94,$C3,$E3,$47,$ED,$FC,$FF,$E0,$E4,$8F,$E0
	.dc.b	$EF,$F0,$AF,$43,$F7,$84,$8F,$C6,$7A,$D5,$FD,$B4,$5C,$BF,$35,$ED
	.dc.b	$66,$7F,$5D,$3C,$DF,$E2,$6D,$7F,$89,$B5,$D9,$A2,$1A,$29,$FB,$9F
	.dc.b	$4F,$F8,$F9,$1F,$E3,$EF,$5A,$FD,$67,$B3,$F8,$B3,$CD,$A7,$EC,$EF
	.dc.b	$F1,$DE,$EC,$C3,$F1,$51,$FC,$6F,$D6,$FE,$56,$48,$FE,$56,$FF,$2E
	.dc.b	$F5,$FF,$90,$12,$3F,$61,$F0,$35,$FE,$35,$17,$2F,$DF,$7B,$F9,$9F
	.dc.b	$E2,$53,$CD,$FE,$6E,$D7,$F9,$BB,$5D,$9A,$21,$A2,$9F,$C7,$7D,$AF
	.dc.b	$E8,$64,$7F,$A1,$BD,$6B,$F8,$87,$C2,$7E,$BC,$F3,$69,$FC,$5E,$FF
	.dc.b	$65,$F0,$B3,$0F,$D7,$47,$F3,$FF,$77,$FA,$79,$23,$FA,$7B,$FD,$4B
	.dc.b	$DE,$FE,$88,$48,$FE,$29,$F0,$B5,$FE,$7D,$17,$2F,$E4,$5F,$0B,$99
	.dc.b	$FE,$6D,$3C,$DF,$EB,$6D,$7F,$AD,$B5,$D9,$A2,$1A,$29,$FD,$07,$E0
	.dc.b	$7F,$B3,$91,$FE,$CE,$F5,$AF,$E6,$9F,$0B,$FC,$4C,$F3,$69,$FC,$EE
	.dc.b	$FF,$8B,$7C,$2C,$C3,$F8,$94,$7F,$63,$F8,$4F,$EE,$64,$8F,$EE,$6F
	.dc.b	$F7,$6F,$85,$FE,$D0,$48,$FE,$FB,$8B,$0F,$EF,$DF,$81,$AB,$55,$FD
	.dc.b	$95,$B5,$B8,$FF,$02,$CA,$F7,$1F,$D7,$EB,$C0,$FF,$81,$1A,$FF,$81
	.dc.b	$1A,$EC,$D1,$0D,$14,$FE,$E3,$F3,$8C,$BF,$B9,$C7,$F8,$56,$30,$E0
	.dc.b	$38,$76,$D7,$C0,$7F,$B0,$AD,$A7,$F8,$7F,$1F,$74,$A7,$2F,$F4,$1C
	.dc.b	$3E,$1E,$5F,$23,$3E,$0D,$9D,$1F,$39,$F3,$69,$E5,$8F,$2B,$72,$34
	.dc.b	$C6,$BD,$27,$FC,$67,$28,$C4,$7B,$9B,$8B,$7C,$D6,$EB,$C4,$72,$AC
	.dc.b	$9C,$38,$EC,$7C,$99,$E0,$FC,$89,$F1,$B9,$6B,$93,$04,$30,$43,$EC
	.dc.b	$BE,$FE,$09,$1D,$DA,$F0,$D7,$E8,$E2,$E1,$F6,$9E,$69,$FF,$7C,$E1
	.dc.b	$E6,$AD,$F3,$7C,$23,$4F,$DA,$87,$1F,$CF,$25,$7B,$DF,$A5,$C9,$F2
	.dc.b	$09,$1F,$20,$FC,$17,$56,$1F,$0C,$5B,$F3,$F4,$66,$7F,$2F,$3D,$1F
	.dc.b	$98,$D7,$F2,$6D,$76,$68,$86,$8A,$7E,$8B,$9F,$F1,$04,$7E,$C3,$D6
	.dc.b	$BE,$4F,$6F,$E1,$0F,$37,$1F,$E9,$EF,$F0,$FE,$0C,$C3,$DD,$A2,$7F
	.dc.b	$13,$D5,$F9,$52,$4F,$E1,$5F,$CC,$75,$FE,$38,$91,$F1,$F8,$5D,$F8
	.dc.b	$8F,$AF,$37,$4F,$BB,$E2,$66,$7F,$7B,$3D,$1F,$C9,$35,$FD,$CB,$5D
	.dc.b	$9A,$21,$A2,$9F,$C2,$3B,$BF,$5C,$47,$F3,$CF,$5A,$FC,$7F,$93,$FA
	.dc.b	$A3,$CD,$B7,$F0,$DB,$FD,$6F,$8D,$30,$F8,$F1,$FA,$EF,$07,$F7,$44
	.dc.b	$8F,$D5,$35,$FE,$22,$D7,$F8,$DB,$5F,$E3,$4D,$7F,$7E,$FF,$05,$F0
	.dc.b	$A4,$7E,$D5,$AF,$F1,$C6,$BF,$C9,$9A,$FF,$25,$6B,$FC,$29,$AF,$F2
	.dc.b	$6F,$0F,$F8,$91,$1F,$E0,$4D,$7F,$94,$B5,$FE,$66,$D7,$F9,$93,$5F
	.dc.b	$E2,$EF,$F1,$DF,$12,$47,$F0,$D6,$BF,$CD,$3C,$5F,$E2,$E4,$7F,$15
	.dc.b	$E6,$57,$F7,$F1,$75,$FE,$0B,$E7,$E6,$7F,$A2,$CF,$F2,$B2,$3F,$C6
	.dc.b	$DA,$F7,$30,$43,$04,$39,$9F,$C8,$FC,$8F,$E6,$C4,$70,$71,$AF,$E2
	.dc.b	$FE,$77,$F3,$23,$CD,$BF,$F5,$5B,$FE,$6B,$E9,$4B,$FD,$7C,$7F,$36
	.dc.b	$F2,$FF,$A1,$12,$3F,$99,$3F,$D1,$F9,$BF,$DD,$98,$FF,$54,$FC,$13
	.dc.b	$FD,$A6,$DA,$FE,$83,$EB,$FF,$44,$97,$B5,$49,$3F,$C6,$E1,$EF,$7F
	.dc.b	$46,$F4,$3F,$A9,$12,$7F,$A1,$3E,$1B,$E9,$10,$FD,$F3,$D1,$61,$E5
	.dc.b	$C3,$0F,$9E,$FF,$A9,$FA,$5F,$D7,$C9,$5F,$D3,$DF,$EC,$98,$BD,$CF
	.dc.b	$51,$27,$D1,$F0,$3E,$DF,$85,$F7,$7C,$1C,$9F,$64,$7F,$60,$F5,$FF
	.dc.b	$B6,$12,$3F,$AE,$3D,$AF,$A4,$7A,$38,$CC,$C7,$C4,$F8,$0E,$A7,$C9
	.dc.b	$C3,$C0,$F1,$38,$61,$FD,$C2,$BF,$BE,$12,$3F,$B6,$3F,$E0,$1F,$9E
	.dc.b	$7B,$3F,$DE,$9B,$3F,$D3,$DA,$F5,$18,$3D,$EE,$15,$27,$83,$F8,$1F
	.dc.b	$87,$C4,$7F,$89,$AF,$9B,$EE,$3D,$EE,$1C,$35,$CC,$F9,$C8,$FF,$8B
	.dc.b	$63,$49,$A4,$8D,$C7,$87,$DC,$F8,$04,$79,$3A,$E1,$FC,$E7,$C3,$FE
	.dc.b	$64,$F4,$BF,$E3,$2B,$F8,$6F,$E8,$7E,$1F,$F3,$C6,$B4,$F4,$FB,$FF
	.dc.b	$34,$98,$F9,$DF,$C6,$FC,$3F,$E9,$7E,$1E,$4F,$E8,$E3,$F5,$1F,$0F
	.dc.b	$23,$F4,$CF,$EA,$FE,$1F,$EB,$7C,$27,$D9,$24,$FC,$37,$AB,$FA,$DB
	.dc.b	$EA,$FD,$9E,$26,$1D,$63,$18,$5D,$3F,$79,$8F,$E0,$FE,$17,$F0,$C4
	.dc.b	$AF,$A0,$FE,$9B,$E1,$FF,$7C,$48,$FD,$E7,$C3,$FF,$01,$9F,$E0,$5A
	.dc.b	$49,$FE,$01,$1F,$8F,$9F,$C9,$12,$3E,$E3,$D5,$FE,$0D,$5D,$5F,$54
	.dc.b	$75,$7F,$85,$57,$57,$DD,$5F,$9B,$AF,$CF,$12,$3F,$10,$FE,$86,$FF
	.dc.b	$40,$48,$FE,$29,$AD,$7F,$41,$C0,$9C,$65,$FC,$87,$8B,$55,$F9,$B3
	.dc.b	$F2,$53,$D5,$FC,$B3,$5F,$D1,$35,$E1,$5D,$5D,$90,$D9,$4D,$E9,$FB
	.dc.b	$92,$3F,$A3,$7B,$5F,$F3,$5E,$3F,$E6,$47,$A3,$87,$F5,$B8,$7F,$9E
	.dc.b	$3D,$BF,$F4,$6D,$7F,$A3,$9B,$3A,$C3,$F8,$17,$17,$F0,$72,$3F,$B1
	.dc.b	$7F,$D1,$52,$EA,$E9,$D9,$2B,$FC,$B1,$39,$6F,$81,$E1,$23,$AF,$0B
	.dc.b	$F8,$97,$2F,$F1,$52,$47,$EF,$DE,$F8,$75,$8F,$E2,$96,$C3,$D9,$3D
	.dc.b	$27,$38,$5C,$FD,$D8,$FF,$1F,$E7,$FE,$44,$4A,$FE,$18,$FB,$3C,$6F
	.dc.b	$8C,$48,$F5,$F8,$9F,$27,$C9,$F9,$5E,$54,$9F,$7C,$7F,$2B,$EC,$FE
	.dc.b	$5C,$48,$FE,$30,$F5,$7E,$2D,$F5,$7D,$77,$D5,$F9,$37,$D5,$F7,$DF
	.dc.b	$F3,$7E,$DF,$E7,$44,$8F,$E4,$8F,$F3,$DE,$EF,$E7,$84,$8F,$A7,$E5
	.dc.b	$D7,$F3,$B1,$75,$FC,$17,$99,$99,$FC,$AC,$F3,$7F,$98,$35,$FE,$78
	.dc.b	$D7,$66,$88,$68,$A7,$E8,$3C,$2F,$EA,$44,$7F,$9E,$3D,$6B,$F2,$9E
	.dc.b	$87,$E5,$CF,$37,$3F,$F4,$6B,$FC,$D7,$9D,$30,$FE,$B1,$1F,$D6,$FC
	.dc.b	$4F,$EB,$C4,$8F,$E9,$0F,$E6,$3D,$3F,$D1,$7B,$12,$7F,$2F,$1F,$A5
	.dc.b	$F5,$64,$7E,$69,$FD,$47,$B7,$FD,$9F,$CA,$FE,$D4,$49,$FE,$A8,$F5
	.dc.b	$7F,$43,$7D,$5F,$CB,$DF,$47,$F3,$37,$75,$FD,$96,$1F,$DC,$FC,$AF
	.dc.b	$EE,$C4,$AF,$EC,$0F,$EC,$3D,$DF,$DD,$12,$3F,$5F,$EF,$7E,$F7,$E0
	.dc.b	$7F,$7D,$F0,$B2,$7F,$67,$1F,$DF,$FC,$FF,$F0,$11,$23,$FB,$63,$D5
	.dc.b	$FD,$CD,$F5,$7F,$5F,$7D,$5F,$DE,$DF,$57,$F6,$77,$FE,$0B,$F4,$3F
	.dc.b	$C1,$C4,$8F,$EF,$0F,$F8,$45,$FF,$08,$10,$FF,$6F,$7E,$0A,$F8,$5C
	.dc.b	$47,$F3,$22,$EA,$E5,$FF,$05,$53,$E1,$F7,$5F,$CB,$0E,$C7,$F8,$1C
	.dc.b	$A7,$F8,$4C,$A6,$CD,$10,$D1,$4F,$E7,$DF,$40,$F5,$7F,$A0,$47,$F8
	.dc.b	$6E,$30,$E0,$38,$76,$D7,$C0,$7F,$8B,$69,$F1,$1F,$E2,$5C,$7D,$D3
	.dc.b	$FE,$44,$9C,$BF,$E3,$53,$E1,$E3,$F0,$BE,$23,$83,$63,$47,$F9,$32
	.dc.b	$B9,$BE,$4F,$67,$CC,$D3,$D1,$8A,$32,$D8,$F6,$60,$F8,$8E,$F1,$E5
	.dc.b	$F8,$8A,$A5,$18,$78,$47,$DE,$F8,$8C,$11,$FC,$0F,$9D,$CB,$07,$BB
	.dc.b	$E2,$24,$F9,$9F,$4D,$F4,$8F,$95,$87,$73,$C1,$8F,$58,$F9,$D8,$DC
	.dc.b	$FE,$1D,$B2,$F0,$7B,$DF,$00,$97,$F1,$1E,$FF,$11,$EB,$24,$7E,$63
	.dc.b	$E2,$3D,$DF,$11,$EF,$F8,$89,$3F,$9C,$8F,$67,$C1,$FC,$C2,$47,$D0
	.dc.b	$7A,$BF,$A2,$BE,$AF,$E1,$6F,$AB,$FA,$6B,$EA,$FE,$26,$FE,$47,$C2
	.dc.b	$7D,$82,$47,$DC,$7E,$DF,$C2,$FC,$D2,$47,$EC,$3E,$1E,$BF,$E7,$FE
	.dc.b	$34,$E3,$7F,$8F,$F0,$FA,$AF,$D1,$CF,$47,$F1,$2D,$7F,$3E,$D7,$66
	.dc.b	$88,$68,$A7,$EB,$A3,$F2,$24,$7F,$2A,$F5,$9F,$B3,$3F,$8D,$3C,$DC
	.dc.b	$DF,$B3,$7F,$21,$8C,$BF,$D1,$4E,$5A,$7F,$AC,$30,$63,$FA,$02,$47
	.dc.b	$E9,$1F,$C7,$ED,$F9,$BD,$E4,$FF,$18,$8F,$CF,$72,$C8,$FE,$36,$FE
	.dc.b	$87,$9B,$F3,$9B,$FE,$B0,$93,$FB,$07,$AB,$FC,$8A,$FA,$BF,$C1,$6F
	.dc.b	$A3,$FC,$1E,$EE,$BF,$9A,$C3,$F5,$1B,$FE,$E0,$95,$FB,$C7,$F4,$FD
	.dc.b	$3F,$B3,$24,$7C,$21,$F8,$37,$FB,$F3,$6F,$F5,$B3,$FE,$FA,$8D,$3F
	.dc.b	$EF,$86,$F8,$5F,$C0,$23,$51,$F8,$9E,$3F,$E1,$24,$7F,$86,$BE,$27
	.dc.b	$F9,$FD,$F8,$9F,$F6,$74,$F8,$5F,$F7,$1B,$D3,$2F,$E3,$2F,$86,$37
	.dc.b	$7E,$1E,$EF,$E1,$1C,$BF,$C6,$09,$3F,$C7,$1F,$E3,$7C,$DF,$C3,$49
	.dc.b	$1E,$CF,$0E,$BF,$A1,$8B,$8F,$83,$C8,$CC,$FF,$42,$9E,$6F,$F2,$86
	.dc.b	$BF,$CF,$9A,$EC,$D1,$0D,$14,$FE,$9B,$D5,$FC,$C0,$8F,$F3,$47,$AD
	.dc.b	$7C,$2F,$27,$E8,$1E,$6D,$FF,$A9,$5F,$D3,$F3,$66,$1F,$77,$A3,$F8
	.dc.b	$F3,$7B,$BF,$9E,$91,$FE,$82,$FF,$44,$E0,$FF,$7F,$47,$E0,$3D,$0F
	.dc.b	$BC,$7D,$FF,$33,$F0,$1E,$AC,$9F,$C2,$BF,$83,$F5,$BF,$9D,$78,$5F
	.dc.b	$D2,$89,$3F,$D3,$9E,$D5,$D2,$3B,$1F,$8E,$7A,$3F,$84,$AB,$BF,$E0
	.dc.b	$AF,$FA,$37,$87,$FD,$58,$95,$FD,$69,$F5,$7B,$7F,$95,$24,$7E,$6B
	.dc.b	$DC,$FC,$C7,$BF,$F9,$9F,$7A,$4F,$E7,$A3,$FA,$97,$91,$FD,$88,$91
	.dc.b	$FD,$99,$EA,$FE,$92,$FA,$BF,$9A,$BE,$AF,$EA,$2F,$AB,$F9,$EB,$FE
	.dc.b	$BD,$E4,$FF,$6E,$24,$7F,$72,$7F,$B8,$F9,$5F,$D8,$C9,$1F,$B2,$F8
	.dc.b	$3A,$FF,$37,$8B,$AF,$E9,$BE,$1B,$33,$FE,$18,$9E,$6F,$F7,$E6,$BF
	.dc.b	$E1,$46,$BB,$34,$43,$45,$3F,$DC,$1C,$2F,$3B,$FC,$16,$D8,$FF,$11
	.dc.b	$35,$FF,$13,$B5,$EC,$7F,$00,$F8,$5F,$E1,$47,$AB,$AB,$FC,$59,$7F
	.dc.b	$C3,$7E,$1A,$57,$E6,$E3,$FC,$1F,$6F,$43,$C2,$49,$F0,$BF,$9B,$F8
	.dc.b	$6F,$E2,$DF,$0D,$27,$F8,$DC,$7F,$19,$F8,$59,$1F,$C7,$9F,$F0,$DF
	.dc.b	$E0,$FF,$EE,$48,$1F,$E2,$1F,$57,$DE,$47,$DE,$F7,$7F,$93,$DF,$77
	.dc.b	$F8,$ED,$F6,$7F,$8F,$DD,$E0,$FF,$1A,$E1,$DE,$7F,$C5,$DE,$B7,$CC
	.dc.b	$23,$F3,$1F,$F1,$CC,$7F,$BB,$61,$3C,$BE,$17,$D2,$7A,$BE,$1B,$D7
	.dc.b	$F0,$DE,$CF,$86,$97,$FD,$07,$04,$FC,$CF,$6F,$ED,$92,$7E,$DB,$E0
	.dc.b	$C3,$AC,$71,$7B,$3C,$E7,$AB,$E4,$E3,$F9,$2A,$D8,$79,$23,$FA,$97
	.dc.b	$0B,$DC,$FC,$59,$23,$F1,$6F,$E3,$BD,$EF,$C0,$92,$3D,$5F,$0D,$5F
	.dc.b	$01,$F6,$29,$39,$DB,$E7,$FC,$36,$EB,$F3,$73,$D5,$FC,$CB,$5F,$CC
	.dc.b	$B5,$D9,$A2,$1A,$29,$FA,$6F,$80,$FD,$29,$1F,$D2,$BD,$6B,$F0,$3F
	.dc.b	$0B,$F9,$03,$CD,$AF,$C8,$C7,$D3,$7F,$5F,$E1,$BF,$2A,$69,$F7,$7C
	.dc.b	$1F,$EC,$88,$FE,$C9,$FC,$3F,$C3,$7E,$70,$91,$F8,$1F,$86,$FC,$FF
	.dc.b	$C3,$7E,$82,$A4,$FE,$16,$3F,$5E,$7E,$F8,$8F,$EF,$9E,$AF,$E2,$EB
	.dc.b	$AB,$F4,$8F,$C0,$57,$67,$F2,$15,$D9,$D7,$12,$FF,$7B,$5F,$C3,$09
	.dc.b	$3F,$C3,$1F,$D0,$EB,$FB,$6D,$A4,$FE,$6E,$3F,$73,$C3,$23,$F3,$CF
	.dc.b	$EF,$39,$7F,$84,$E7,$FC,$60,$93,$FC,$61,$EA,$FE,$96,$FA,$BE,$6B
	.dc.b	$E8,$FA,$2E,$F0,$7E,$EB,$0F,$E2,$BA,$FF,$24,$25,$7F,$24,$7F,$93
	.dc.b	$F0,$7F,$1A,$24,$7E,$CF,$9E,$BF,$27,$81,$38,$C3,$F8,$3F,$5E,$AB
	.dc.b	$F9,$9C,$F4,$7F,$98,$B5,$FE,$62,$D7,$66,$88,$68,$A7,$F3,$EE,$3F
	.dc.b	$E7,$A4,$7F,$9E,$BD,$6B,$F8,$CF,$67,$F2,$83,$CD,$C9,$F9,$9B,$FE
	.dc.b	$57,$E0,$4C,$3F,$7F,$1F,$CE,$B9,$BF,$A4,$92,$3F,$A4,$BF,$ED,$68
	.dc.b	$77,$7F,$B3,$9B,$3F,$E5,$96,$BF,$8F,$5E,$1E,$FC,$8E,$1E,$4C,$AE
	.dc.b	$11,$FD,$2B,$AF,$FA,$C9,$23,$FA,$CB,$E1,$C3,$C0,$48,$F0,$5F,$57
	.dc.b	$C5,$7D,$5F,$15,$FF,$56,$EC,$FE,$CA,$48,$FE,$CA,$FB,$78,$7D,$9E
	.dc.b	$7C,$9F,$64,$7B,$FD,$49,$1E,$F7,$E1,$7A,$7F,$D8,$BC,$1F,$EE,$24
	.dc.b	$9F,$EE,$2F,$57,$E3,$5F,$57,$E3,$5F,$47,$E4,$5D,$E3,$71,$7F,$38
	.dc.b	$E0,$FF,$24,$D7,$E4,$FC,$1F,$EF,$E4,$7F,$BF,$BF,$E0,$29,$FE,$EA
	.dc.b	$48,$FA,$DE,$BD,$7F,$39,$16,$FB,$C7,$F6,$EB,$7B,$7B,$8F,$F0,$7D
	.dc.b	$75,$7F,$C1,$AD,$7F,$C1,$AD,$76,$68,$86,$8A,$7F,$54,$1F,$46,$F1
	.dc.b	$71,$1F,$D6,$23,$0E,$0D,$C5,$69,$DB,$0F,$5B,$FB,$8F,$77,$F3,$07
	.dc.b	$BE,$FF,$9E,$1F,$9B,$F8,$29,$3F,$D8,$A3,$FB,$3D,$72,$C1,$E4,$F7
	.dc.b	$6D,$78,$DA,$FA,$1A,$FB,$73,$FF,$6E,$BF,$A7,$F8,$3F,$D5,$1F,$AB
	.dc.b	$F8,$6F,$D6,$7C,$37,$EB,$7E,$1A,$57,$EB,$63,$C7,$9F,$C3,$F3,$7C
	.dc.b	$44,$9F,$13,$FB,$7C,$3A,$C7,$73,$F6,$D8,$76,$7F,$71,$87,$67,$F7
	.dc.b	$18,$76,$7F,$C5,$F9,$3C,$EF,$80,$48,$F8,$0F,$F0,$5F,$85,$FE,$09
	.dc.b	$F0,$D2,$7F,$82,$47,$F0,$7F,$86,$91,$FC,$1D,$FE,$15,$F0,$DD,$E3
	.dc.b	$F0,$3E,$9F,$CD,$23,$F3,$5E,$CF,$F0,$FB,$EC,$FF,$0F,$BE,$AF,$F1
	.dc.b	$1B,$BC,$5F,$82,$E1,$63,$DF,$37,$A9,$F6,$C9,$87,$DB,$7E,$F7,$AB
	.dc.b	$B2,$47,$F1,$FF,$86,$AF,$F6,$DA,$4E,$31,$FC,$8F,$E1,$1F,$EE,$BA
	.dc.b	$FC,$EC,$F4,$7F,$10,$D7,$F1,$0D,$7F,$49,$5D,$1D,$90,$D9,$4D,$C7
	.dc.b	$E5,$7D,$99,$1F,$95,$6B,$C0,$FF,$02,$57,$FB,$F7,$9B,$FC,$8A,$D8
	.dc.b	$FC,$0D,$7A,$E7,$3F,$0F,$F5,$E8,$E6,$BF,$0F,$ED,$7E,$94,$8F,$E9
	.dc.b	$5F,$DB,$4B,$E1,$B9,$12,$7F,$CF,$B9,$FC,$8F,$86,$ED,$F0,$DD,$FE
	.dc.b	$1A,$56,$B3,$F3,$CF,$DB,$F7,$FF,$68,$49,$FD,$A3,$E0,$C7,$B4,$7C
	.dc.b	$07,$D5,$87,$68,$E8,$7F,$A3,$B2,$63,$E0,$8F,$8F,$87,$E2,$7E,$03
	.dc.b	$F8,$11,$23,$F8,$13,$EE,$F8,$3F,$8D,$F0,$92,$7A,$8C,$3F,$DF,$AF
	.dc.b	$D2,$F8,$4F,$98,$47,$70,$F8,$4F,$CC,$1F,$C4,$49,$1F,$C4,$5F,$03
	.dc.b	$C2,$BC,$0F,$DE,$BE,$8F,$73,$AD,$FD,$BA,$FD,$23,$FC,$74,$91,$FC
	.dc.b	$75,$FE,$41,$5F,$A9,$24,$79,$B0,$AF,$E8,$E2,$EF,$FA,$2D,$B3,$3F
	.dc.b	$9C,$9E,$6F,$F2,$96,$BF,$CA,$5A,$EC,$D1,$0D,$14,$FD,$3E,$3F,$CD
	.dc.b	$48,$FF,$35,$7A,$D7,$ED,$38,$3F,$34,$79,$B8,$3F,$A7,$5F,$E7,$78
	.dc.b	$E6,$1F,$63,$91,$3F,$C2,$78,$3F,$A1,$12,$7F,$A1,$3F,$9E,$E4,$FD
	.dc.b	$31,$23,$ED,$F3,$7E,$A7,$B3,$F5,$5D,$B2,$7E,$FC,$7F,$15,$E4,$FE
	.dc.b	$9C,$48,$FE,$9C,$F5,$7F,$0B,$7D,$5F,$DD,$DF,$57,$F1,$37,$D5,$FD
	.dc.b	$FD,$FF,$21,$E5,$FE,$B4,$48,$FE,$B4,$FE,$DB,$C2,$FD,$F7,$87,$27
	.dc.b	$F2,$71,$FC,$07,$C6,$91,$F9,$67,$F9,$D7,$E1,$3F,$EE,$9D,$FF,$7D
	.dc.b	$3F,$D9,$7A,$BF,$B4,$11,$FE,$F5,$D4,$C3,$B4,$7F,$14,$C7,$F8,$74
	.dc.b	$76,$8F,$E2,$F8,$DE,$5F,$EF,$B9,$77,$EB,$FE,$E8,$4A,$FE,$E8,$FF
	.dc.b	$77,$F8,$8F,$E8,$04,$8F,$D9,$CF,$FB,$E6,$BF,$BF,$4E,$67,$F9,$E3
	.dc.b	$EC,$F3,$F1,$1F,$CD,$23,$A3,$FE,$01,$6B,$FE,$01,$6B,$B3,$44,$34
	.dc.b	$53,$F9,$F9,$FE,$0D,$21,$FE,$0D,$3A,$CF,$C0,$F4,$7E,$79,$E7,$7D
	.dc.b	$31,$EC,$F5,$BE,$99,$9E,$9F,$F8,$03,$F4,$CF,$07,$FC,$50,$DA,$FF
	.dc.b	$16,$35,$FF,$18,$B5,$FF,$1B,$66,$F6,$3E,$F9,$23,$E5,$FB,$7F,$82
	.dc.b	$F1,$7F,$07,$EF,$49,$F9,$F1,$FD,$FF,$0F,$EE,$71,$E0,$F1,$BC,$64
	.dc.b	$7F,$C7,$D8,$F7,$7E,$BE,$3D,$DF,$F6,$E4,$31,$F0,$47,$DC,$C7,$C0
	.dc.b	$EB,$1F,$07,$8F,$F0,$09,$1F,$01,$FF,$25,$5B,$E0,$FF,$37,$F0,$92
	.dc.b	$BF,$0F,$1F,$9E,$F8,$59,$1F,$8A,$7F,$43,$F0,$FE,$DF,$37,$E6,$92
	.dc.b	$7E,$6B,$DA,$BA,$47,$AA,$FF,$53,$1D,$23,$DB,$77,$8F,$F3,$F8,$7C
	.dc.b	$9F,$37,$EC,$92,$BE,$CB,$F7,$3C,$FF,$9C,$48,$FC,$FF,$C2,$D5,$7F
	.dc.b	$E6,$6E,$35,$66,$3F,$5B,$F0,$DB,$AF,$F3,$B5,$69,$FC,$3D,$77,$69
	.dc.b	$0F,$C4,$B9,$69,$93,$9B,$97,$AF,$A3,$F9,$42,$3F,$94,$79,$5F,$F0
	.dc.b	$0F,$86,$FE,$1E,$79,$FA,$3F,$3B,$7F,$C4,$BE,$1A,$61,$FD,$C3,$1F
	.dc.b	$43,$F0,$2B,$F4,$7E,$A7,$E8,$89,$1F,$A2,$7F,$07,$F0,$DF,$C7,$09
	.dc.b	$1F,$C2,$3E,$1B,$F8,$FF,$C3,$7F,$20,$F8,$69,$3F,$C3,$23,$F3,$DE
	.dc.b	$CF,$EB,$89,$1F,$AE,$7A,$BF,$C4,$AF,$AB,$F9,$CB,$EA,$FF,$16,$BE
	.dc.b	$AF,$E8,$2F,$F5,$5E,$D7,$EE,$89,$1F,$BA,$7F,$D4,$9F,$85,$7F,$BD
	.dc.b	$38,$3F,$D4,$6F,$FB,$D2,$7F,$C2,$6D,$7E,$20,$0F,$1D,$31,$18,$F8
	.dc.b	$4F,$F3,$C4,$9A,$63,$E1,$7F,$7D,$EF,$7F,$0E,$25,$7F,$0E,$7C,$BF
	.dc.b	$0F,$CB,$E1,$E4,$F9,$23,$9B,$E6,$BA,$F9,$8F,$2D,$FF,$0B,$F8,$1F
	.dc.b	$E3,$64,$8F,$E3,$6F,$F1,$EF,$82,$FE,$20,$48,$F8,$3F,$0B,$5E,$9C
	.dc.b	$A9,$C6,$DF,$23,$E1,$B5,$5F,$A2,$9E,$8F,$F2,$86,$BF,$CA,$1A,$EC
	.dc.b	$D1,$0D,$14,$FD,$6F,$C3,$FF,$34,$23,$FC,$D1,$EB,$5F,$6B,$E1,$FE
	.dc.b	$89,$E6,$E1,$FE,$7B,$7F,$52,$A5,$FE,$86,$3F,$98,$57,$F4,$02,$47
	.dc.b	$F4,$07,$F0,$F1,$F7,$09,$1F,$86,$CF,$EF,$F0,$7E,$03,$8A,$4F,$E2
	.dc.b	$A3,$F9,$E6,$7F,$D3,$09,$1F,$D3,$1E,$AF,$E4,$2F,$AB,$F6,$F8,$98
	.dc.b	$76,$8F,$CA,$E1,$D9,$D6,$1F,$D2,$76,$FE,$B2,$48,$FE,$B2,$FE,$83
	.dc.b	$9F,$F2,$BD,$12,$7F,$3F,$1F,$98,$EC,$91,$FA,$27,$F3,$7D,$7F,$D5
	.dc.b	$78,$FF,$B2,$92,$7F,$B2,$BD,$5F,$D4,$DF,$57,$D5,$7D,$1F,$65,$DE
	.dc.b	$2F,$D8,$E1,$FD,$87,$93,$FB,$89,$2B,$FB,$8B,$FD,$D3,$9B,$FB,$31
	.dc.b	$23,$F6,$FD,$F5,$FC,$BC,$5B,$FD,$EF,$89,$99,$FE,$6F,$3C,$DF,$EF
	.dc.b	$AD,$7F,$BE,$B5,$D9,$A2,$1A,$29,$FD,$0B,$A3,$FC,$12,$47,$FC,$12
	.dc.b	$F5,$AF,$E1,$FE,$37,$F0,$63,$CD,$D1,$FE,$12,$BF,$E1,$3E,$5C,$C3
	.dc.b	$F9,$B4,$7F,$0E,$B5,$7F,$10,$FA,$87,$6F,$F1,$2B,$91,$87,$07,$F1
	.dc.b	$97,$0E,$1A,$F1,$E7,$FF,$0F,$5F,$F2,$4F,$37,$F8,$D9,$E1,$C3,$F8
	.dc.b	$FF,$C0,$7F,$BE,$4F,$F7,$25,$FF,$A0,$BD,$3D,$3F,$C9,$E3,$FD,$C6
	.dc.b	$FF,$AF,$9D,$E2,$F0,$3C,$B3,$E5,$6A,$8D,$3A,$6B,$EC,$CB,$CF,$87
	.dc.b	$F3,$26,$C7,$F3,$76,$BF,$E9,$63,$CA,$7F,$A6,$3D,$2B,$87,$9A,$3F
	.dc.b	$95,$37,$3F,$9F,$1E,$B9,$E3,$83,$C2,$F3,$B6,$7C,$D1,$83,$D7,$7F
	.dc.b	$E8,$08,$EB,$ED,$78,$0F,$0F,$B3,$E2,$F7,$E4,$FC,$77,$CB,$F0,$1E
	.dc.b	$0F,$1F,$B9,$27,$E2,$3D,$AB,$A4,$7C,$FB,$EA,$F3,$21,$EE,$BE,$AF
	.dc.b	$FB,$CB,$27,$91,$EC,$24,$7D,$17,$DF,$E4,$FC,$22,$47,$DD,$F8,$3A
	.dc.b	$FF,$53,$A4,$E3,$7F,$97,$3F,$EC,$8D,$76,$FE,$62,$7A,$BF,$7D,$AF
	.dc.b	$E5,$DA,$EC,$D1,$0D,$14,$FD,$1F,$97,$F5,$88,$FE,$2D,$EB,$5E,$4A
	.dc.b	$7C,$37,$E1,$0F,$47,$1F,$EB,$F0,$FC,$3F,$C3,$CC,$7E,$44,$7E,$13
	.dc.b	$CE,$FC,$19,$23,$F3,$0F,$E0,$BE,$1F,$F1,$E4,$8F,$D2,$7C,$3F,$E4
	.dc.b	$BE,$1F,$F2,$7F,$0F,$27,$F5,$11,$F9,$0F,$4F,$F1,$E4,$8F,$D2,$3D
	.dc.b	$5F,$D7,$5F,$57,$F3,$77,$D5,$FD,$95,$F5,$7F,$3F,$7F,$9C,$F5,$3F
	.dc.b	$36,$48,$FD,$83,$F9,$AF,$87,$FD,$3F,$C3,$C9,$FD,$E4,$7E,$AB,$E1
	.dc.b	$A4,$7E,$FD,$FD,$6F,$C3,$FE,$A3,$DA,$FD,$39,$27,$F7,$8F,$57,$F8
	.dc.b	$35,$F5,$7F,$67,$7D,$1F,$DA,$DD,$E4,$FB,$18,$7E,$D3,$DA,$FD,$99
	.dc.b	$2B,$F8,$43,$FB,$6F,$77,$F7,$24,$8F,$E2,$9F,$09,$5F,$CC,$45,$BF
	.dc.b	$C2,$7C,$2E,$67,$F9,$6C,$F3,$7F,$8A,$35,$FE,$58,$D7,$66,$88,$68
	.dc.b	$A7,$F3,$8F,$7F,$F8,$59,$1F,$E4,$0F,$5A,$FC,$B7,$C2,$FF,$18,$3C
	.dc.b	$DD,$1F,$D0,$EF,$F8,$EF,$E1,$7F,$AE,$AF,$D7,$B6,$75,$87,$EF,$FE
	.dc.b	$0B,$F8,$C9,$1F,$E5,$CF,$F3,$29,$7C,$4F,$42,$4F,$F9,$8E,$FA,$FC
	.dc.b	$3F,$83,$E1,$FC,$3F,$0F,$2B,$59,$7F,$99,$B1,$57,$F2,$AF,$87,$FE
	.dc.b	$52,$4A,$FE,$7C,$F8,$72,$EF,$1F,$E5,$DA,$F1,$BD,$0F,$F3,$A2,$C7
	.dc.b	$C5,$19,$10,$C7,$C9,$1C,$F0,$FE,$77,$1F,$CE,$89,$1F,$D3,$DF,$85
	.dc.b	$87,$CC,$CE,$4F,$5B,$FF,$79,$3E,$5D,$BE,$89,$1D,$6B,$FD,$23,$1F
	.dc.b	$E8,$E4,$8F,$EB,$8F,$67,$A5,$F6,$7E,$06,$1C,$DE,$C7,$86,$FE,$FD
	.dc.b	$FF,$54,$CB,$FA,$99,$27,$FB,$43,$FD,$5B,$3F,$EB,$24,$8F,$57,$0D
	.dc.b	$7F,$9C,$45,$DB,$F1,$3C,$D9,$9F,$F0,$6C,$F3,$7F,$BA,$35,$FF,$06
	.dc.b	$35,$D9,$A2,$1A,$29,$FE,$18,$DF,$FB,$59,$1F,$F0,$03,$D6,$BF,$37
	.dc.b	$CF,$F9,$E3,$CD,$1F,$E2,$3A,$FD,$0F,$64,$BF,$EB,$91,$FE,$00,$F1
	.dc.b	$DC,$7F,$E0,$E6,$CF,$F8,$49,$AF,$F8,$A1,$AF,$F8,$B7,$89,$DD,$FA
	.dc.b	$E2,$47,$E0,$FC,$3F,$D8,$F8,$1F,$B2,$F1,$A4,$FE,$1E,$3C,$5D,$1E
	.dc.b	$12,$47,$85,$FD,$85,$F4,$8F,$D0,$3F,$BC,$BE,$B1,$F9,$1B,$EA,$FE
	.dc.b	$96,$FD,$DD,$3E,$D2,$47,$B5,$FD,$E7,$93,$FC,$1B,$CB,$93,$F9,$B8
	.dc.b	$FE,$13,$E7,$48,$FC,$F3,$FC,$33,$D1,$F9,$5D,$DF,$24,$93,$F2,$5E
	.dc.b	$AF,$E9,$6F,$AB,$FB,$9B,$E8,$FE,$EE,$EF,$2F,$F1,$6C,$3E,$B7,$7F
	.dc.b	$D5,$25,$7D,$57,$EC,$78,$3F,$6C,$91,$FB,$FF,$C3,$3F,$DE,$8F,$FB
	.dc.b	$DF,$8D,$3A,$8F,$F7,$34,$61,$E1,$E9,$E5,$FF,$35,$46,$5F,$C1,$4F
	.dc.b	$81,$FC,$33,$5C,$98,$21,$82,$1D,$1F,$C8,$78,$72,$3F,$20,$D7,$86
	.dc.b	$5F,$E2,$6C,$1E,$CF,$F9,$61,$B3,$FC,$77,$0F,$35,$7F,$8B,$1B,$D4
	.dc.b	$99,$4F,$7B,$91,$F0,$EF,$87,$8F,$FA,$32,$57,$E8,$DF,$F3,$9D,$3D
	.dc.b	$FF,$11,$27,$E8,$7C,$27,$97,$E1,$7C,$DF,$0D,$27,$EA,$47,$EC,$7C
	.dc.b	$BF,$D8,$12,$3F,$60,$F6,$BE,$90,$CB,$E1,$47,$C1,$C3,$B4,$FD,$FC
	.dc.b	$3B,$3B,$D3,$E0,$2C,$99,$7F,$01,$F3,$3F,$80,$12,$7F,$80,$3E,$EF
	.dc.b	$86,$F9,$DF,$0D,$27,$F1,$91,$F4,$7E,$16,$47,$E3,$DF,$A9,$F0,$DF
	.dc.b	$C3,$7D,$1F,$E1,$84,$9F,$E1,$8F,$57,$F2,$D7,$D5,$F4,$1F,$5A,$FA
	.dc.b	$B8,$3D,$57,$8D,$FF,$09,$87,$F1,$BF,$4B,$F8,$D1,$2B,$F8,$D3,$FC
	.dc.b	$77,$D3,$FE,$40,$48,$FD,$37,$C2,$D7,$F4,$71,$6F,$F2,$1F,$0D,$99
	.dc.b	$FD,$1C,$F3,$7F,$93,$35,$FE,$4C,$D7,$66,$88,$68,$A7,$EB,$BD,$6F
	.dc.b	$E6,$44,$7F,$99,$3D,$6B,$F4,$5F,$0D,$FA,$13,$CD,$C5,$F7,$BA,$7F
	.dc.b	$CE,$F7,$FC,$07,$E1,$BF,$4C,$69,$FE,$05,$ED,$7F,$3F,$23,$FC,$FD
	.dc.b	$FC,$EF,$C3,$7E,$B4,$91,$FC,$2F,$E1,$BF,$61,$F0,$BF,$B1,$F8,$59
	.dc.b	$3F,$C3,$E3,$FA,$67,$C0,$FF,$4B,$24,$7F,$4B,$7A,$BF,$C5,$6F,$AB
	.dc.b	$F7,$7A,$98,$76,$8F,$E3,$58,$76,$75,$87,$F5,$9F,$80,$FE,$B0,$48
	.dc.b	$FE,$B0,$FE,$DB,$E1,$7F,$DB,$BF,$86,$FE,$8A,$7F,$14,$DE,$1F,$E1
	.dc.b	$93,$AF,$84,$95,$FE,$27,$AF,$F7,$D9,$FE,$E1,$E1,$F9,$DC,$3F,$D9
	.dc.b	$C9,$3F,$D9,$DF,$03,$C4,$F0,$9F,$C5,$EF,$AB,$E2,$BB,$87,$FB,$0B
	.dc.b	$0E,$D3,$FC,$DF,$E1,$3F,$BA,$11,$FE,$E8,$FF,$77,$F8,$6F,$EF,$44
	.dc.b	$8F,$44,$57,$F9,$EC,$5A,$ED,$8E,$67,$F9,$EC,$F3,$7F,$C0,$0D,$7F
	.dc.b	$C0,$0D,$76,$68,$86,$8A,$7F,$48,$C3,$FC,$18,$47,$FC,$18,$F5,$AF
	.dc.b	$81,$9F,$D6,$3C,$D9,$FE,$6E,$FE,$CF,$0C,$C3,$F9,$F4,$7F,$86,$5F
	.dc.b	$F0,$C1,$0F,$F0,$C1,$F7,$CF,$E4,$39,$B8,$BF,$06,$48,$FB,$DC,$BF
	.dc.b	$86,$E6,$FC,$3F,$4C,$9F,$C1,$47,$F2,$AD,$3F,$DE,$FF,$58,$E1,$ED
	.dc.b	$5F,$D3,$63,$0E,$1A,$F8,$DA,$FE,$4F,$2F,$23,$F8,$FC,$BC,$8F,$E5
	.dc.b	$B2,$F2,$3F,$93,$CB,$C8,$FE,$6B,$2F,$26,$BF,$CF,$B1,$71,$F9,$29
	.dc.b	$18,$7C,$91,$98,$C4,$7C,$7C,$9D,$FF,$A8,$F0,$24,$FE,$8E,$3F,$57
	.dc.b	$E2,$C8,$FD,$33,$FA,$EF,$1B,$C7,$1F,$0F,$9F,$B1,$1F,$88,$F9,$70
	.dc.b	$EF,$1F,$AF,$C3,$BB,$FB,$6C,$3B,$3F,$B8,$C2,$F0,$FE,$9F,$1E,$F3
	.dc.b	$EE,$E8,$F6,$11,$FA,$4F,$BF,$AB,$7A,$1A,$7F,$06,$F2,$6D,$7F,$A9
	.dc.b	$E7,$4D,$9F,$E1,$7E,$56,$A3,$EE,$47,$57,$F0,$0D,$72,$60,$86,$08
	.dc.b	$7F,$15,$9E,$EE,$0F,$B9,$DB,$F5,$C8,$FE,$35,$ED,$97,$F0,$9F,$33
	.dc.b	$F8,$F9,$E5,$7F,$2F,$7F,$C8,$BD,$19,$7E,$B3,$DD,$E0,$7E,$0C,$8F
	.dc.b	$E6,$5F,$AD,$E9,$7F,$2A,$24,$7F,$25,$F5,$FF,$96,$FB,$3F,$4B,$C8
	.dc.b	$FF,$76,$4F,$F8,$BD,$AB,$97,$FD,$17,$9F,$EA,$BC,$5F,$C9,$12,$7F
	.dc.b	$50,$F8,$1F,$E6,$58,$F8,$1F,$E7,$38,$F8,$1F,$F6,$E6,$1D,$8E,$F8
	.dc.b	$76,$E9,$FD,$27,$8D,$F9,$F2,$57,$ED,$5F,$D3,$57,$FB,$AA,$3F,$C1
	.dc.b	$CF,$FB,$A5,$EB,$EF,$FB,$7D,$BF,$77,$C2,$49,$F0,$BF,$07,$E1,$7F
	.dc.b	$1B,$E5,$FE,$B4,$93,$FC,$0D,$EF,$5F,$1E,$E4,$FA,$A3,$A3,$E2,$BB
	.dc.b	$BB,$3F,$F1,$DE,$5F,$EF,$9A,$FC,$B7,$99,$FB,$B2,$3F,$C4,$5F,$DF
	.dc.b	$79,$BF,$A0,$24,$7C,$8F,$84,$AF,$F8,$B7,$A5,$38,$C7,$C1,$3F,$6F
	.dc.b	$E1,$76,$FE,$5E,$7A,$BF,$C7,$DA,$FF,$1F,$6B,$B3,$44,$34,$53,$F4
	.dc.b	$58,$FF,$15,$23,$FC,$AD,$EB,$5E,$48,$43,$E1,$7F,$24,$7A,$73,$FE
	.dc.b	$49,$7F,$95,$F8,$59,$7F,$BD,$8F,$E0,$7E,$9F,$F2,$32,$47,$F3,$97
	.dc.b	$F0,$BF,$0B,$F9,$DF,$85,$93,$F9,$08,$FD,$07,$C2,$C8,$FC,$9B,$FA
	.dc.b	$3F,$85,$FE,$1F,$EB,$FF,$2F,$24,$FF,$45,$7A,$BF,$9A,$BE,$AF,$E4
	.dc.b	$AF,$A3,$F9,$4B,$BB,$7C,$1D,$98,$FF,$1C,$F6,$3F,$9E,$12,$FF,$A9
	.dc.b	$3F,$9E,$F8,$4F,$DB,$12,$3F,$53,$F0,$9F,$B9,$F8,$3F,$DD,$7C,$24
	.dc.b	$9F,$D6,$C7,$F2,$8F,$77,$FA,$41,$23,$FA,$F3,$D5,$FD,$9D,$F5,$7F
	.dc.b	$57,$7D,$5F,$DB,$DF,$57,$F5,$F7,$FC,$D3,$DE,$FE,$A8,$48,$FE,$D4
	.dc.b	$FF,$57,$F7,$FF,$9C,$92,$3F,$80,$FC,$1D,$7F,$35,$17,$6F,$8D,$F0
	.dc.b	$99,$9F,$E5,$D3,$CD,$FE,$EC,$D7,$FB,$B3,$5D,$9A,$21,$A2,$9F,$CE
	.dc.b	$7E,$0B,$FB,$61,$1F,$F0,$13,$D6,$BE,$C7,$C2,$7F,$27,$3C,$DD,$3F
	.dc.b	$DE,$6F,$F9,$57,$C2,$4C,$3F,$A9,$47,$F5,$8F,$88,$FE,$F0,$48,$FF
	.dc.b	$07,$3F,$DF,$FF,$0E,$FF,$7F,$C1,$65,$FD,$99,$B1,$FE,$E3,$CB,$F5
	.dc.b	$51,$C0,$FF,$75,$C7,$FA,$BD,$B1,$AB,$E5,$AF,$F7,$0F,$AE,$65,$FE
	.dc.b	$0F,$6E,$1F,$E2,$46,$BF,$E3,$36,$BF,$E7,$3C,$3C,$18,$F9,$4F,$36
	.dc.b	$5E,$7C,$65,$79,$F3,$FF,$6D,$DF,$91,$B9,$78,$F4,$F1,$12,$BC,$4F
	.dc.b	$C5,$E3,$F8,$44,$8E,$80,$F9,$1C,$9F,$1F,$AB,$E4,$75,$C9,$D9,$F3
	.dc.b	$63,$E1,$70,$FC,$12,$47,$C1,$7C,$0F,$0C,$D8,$78,$63,$E7,$61,$DD
	.dc.b	$D6,$1D,$DF,$A5,$87,$CF,$E2,$F9,$C4,$8F,$9C,$FD,$1E,$4F,$A8,$48
	.dc.b	$F0,$74,$57,$FD,$D5,$C4,$9C,$72,$7E,$0B,$B3,$55,$FE,$A7,$AD,$3F
	.dc.b	$6E,$BB,$34,$87,$DD,$72,$D3,$27,$37,$2F,$5E,$6F,$C4,$11,$FC,$43
	.dc.b	$CA,$FF,$2B,$E1,$7E,$64,$F3,$F1,$FD,$7A,$61,$E6,$7F,$31,$E3,$7E
	.dc.b	$7C,$DB,$F7,$61,$D3,$F9,$72,$47,$E5,$DF,$D5,$79,$1F,$A7,$F3,$24
	.dc.b	$FD,$98,$FD,$57,$9D,$23,$EE,$3F,$AD,$F4,$BF,$49,$DD,$FA,$32,$4F
	.dc.b	$E8,$DE,$AF,$E0,$EF,$AB,$FA,$FB,$E8,$FE,$C6,$EE,$BF,$B6,$C3,$F6
	.dc.b	$1D,$FF,$AF,$25,$7E,$BD,$FE,$05,$E8,$FE,$FC,$91,$F9,$1F,$57,$F8
	.dc.b	$17,$B5,$FC,$0F,$DB,$93,$F9,$58,$FD,$E7,$89,$FB,$B2,$47,$EE,$DE
	.dc.b	$AF,$E6,$EF,$AB,$FC,$1E,$FA,$BF,$9F,$BE,$AF,$F0,$BB,$FE,$11,$E3
	.dc.b	$7F,$07,$24,$7F,$07,$7F,$85,$78,$DF,$C3,$49,$1F,$A9,$F8,$0A,$FE
	.dc.b	$6A,$2E,$BF,$E0,$3F,$C3,$FF,$DF,$9C,$DF,$85,$E5,$57,$4B,$FE,$28
	.dc.b	$D7,$F8,$A3,$5F,$D0,$35,$E4,$D7,$F1,$4D,$7F,$23,$E4,$FF,$21,$23
	.dc.b	$FC,$85,$ED,$87,$97,$E1,$7F,$D2,$F8,$9F,$4E,$3D,$AF,$F8,$6C,$BE
	.dc.b	$0F,$66,$A3,$F3,$5E,$67,$F2,$F2,$3F,$CB,$DF,$F6,$04,$BE,$1F,$BF
	.dc.b	$C3,$CA,$EB,$1E,$2F,$87,$91,$E0,$7C,$BF,$0F,$FC,$F7,$D1,$FE,$78
	.dc.b	$49,$FE,$78,$F6,$AE,$90,$D7,$E9,$61,$EE,$9E,$B3,$9D,$3F,$71,$93
	.dc.b	$2C,$E1,$F6,$72,$FE,$99,$E9,$FF,$4B,$26,$1F,$D2,$DF,$6F,$C3,$FC
	.dc.b	$A2,$47,$C8,$F8,$7F,$9B,$F0,$DF,$3B,$E1,$E4,$FC,$C8,$FE,$B1,$ED
	.dc.b	$FF,$57,$24,$7F,$57,$7A,$BF,$4A,$FA,$BE,$93,$FA,$36,$8C,$3B,$C7
	.dc.b	$D8,$C3,$BB,$B1,$2C,$3F,$B4,$7B,$3F,$D9,$C9,$3F,$D9,$DF,$ED,$5E
	.dc.b	$DF,$F6,$D2,$47,$E1,$3E,$1A,$BF,$CE,$A2,$ED,$FA,$1F,$87,$CC,$FF
	.dc.b	$39,$9E,$6F,$F7,$36,$BF,$DC,$DA,$EC,$D1,$0D,$14,$FE,$89,$EE,$FF
	.dc.b	$7F,$23,$FD,$FD,$EB,$5F,$B3,$F8,$7F,$CF,$9E,$6E,$3F,$DA,$5F,$E8
	.dc.b	$BE,$1E,$61,$FD,$B3,$B7,$FD,$60,$FF,$56,$AF,$F0,$61,$1F,$F0,$63
	.dc.b	$F9,$EF,$87,$FD,$5F,$C3,$C9,$FD,$4C,$7E,$BB,$E1,$E4,$7E,$B1,$FD
	.dc.b	$8F,$C3,$FF,$49,$D3,$FD,$15,$F6,$0F,$88,$FE,$99,$72,$30,$E0,$FF
	.dc.b	$03,$38,$70,$D7,$8F,$23,$1F,$24,$7E,$0E,$19,$7F,$02,$9F,$2C,$E7
	.dc.b	$2B,$8F,$F1,$BC,$BC,$D9,$FF,$5F,$C9,$F0,$FE,$3C,$3C,$6D,$51,$A7
	.dc.b	$4D,$7D,$59,$BE,$23,$F8,$99,$23,$F8,$8D,$7F,$16,$D7,$C3,$1F,$E2
	.dc.b	$5F,$C4,$3F,$DF,$67,$FB,$F2,$FF,$DD,$D1,$F1,$2F,$C5,$87,$58,$F0
	.dc.b	$C6,$0F,$E2,$A7,$77,$E2,$3F,$C9,$30,$F0,$47,$F8,$BA,$FB,$E3,$88
	.dc.b	$FD,$65,$B2,$EE,$48,$F7,$BE,$2D,$3D,$31,$F1,$1B,$27,$37,$FF,$00
	.dc.b	$74,$FC,$1C,$68,$95,$E3,$E1,$AF,$E0,$23,$B3,$F3,$9A,$E4,$C1,$0C
	.dc.b	$10,$FC,$73,$C1,$FC,$86,$FF,$28,$8F,$DA,$7A,$D7,$FB,$86,$DC,$DE
	.dc.b	$A3,$CB,$1E,$AA,$F9,$5D,$7E,$E3,$3A,$7E,$2F,$87,$EB,$92,$7F,$0E
	.dc.b	$FE,$2F,$07,$67,$C6,$F0,$65,$7D,$58,$F9,$3E,$14,$8F,$B0,$FC,$CF
	.dc.b	$17,$E7,$F3,$FE,$10,$93,$F9,$47,$B5,$74,$8F,$E5,$18,$30,$FA,$B3
	.dc.b	$D6,$73,$85,$D3,$E1,$63,$F7,$3A,$3F,$22,$4A,$FD,$0B,$EB,$F1,$BE
	.dc.b	$F9,$23,$F1,$DE,$47,$E0,$BC,$DF,$C1,$F9,$D2,$7F,$25,$1F,$88,$EE
	.dc.b	$FC,$E9,$23,$F5,$AF,$57,$F3,$17,$D5,$F5,$DF,$57,$F3,$97,$D5,$F7
	.dc.b	$DF,$E5,$3B,$FF,$52,$48,$FD,$CB,$FA,$CF,$07,$F3,$24,$8F,$D3,$7A
	.dc.b	$75,$FD,$2C,$5D,$7F,$0D,$EA,$66,$7F,$8E,$4F,$37,$F8,$2B,$5F,$E3
	.dc.b	$6D,$76,$68,$86,$8A,$7F,$27,$F0,$FF,$80,$91,$FE,$22,$F5,$AF,$CC
	.dc.b	$7A,$FF,$B0,$3C,$DC,$FF,$C7,$6F,$F6,$7E,$C4,$C3,$FD,$5F,$DC,$9F
	.dc.b	$DC,$F8,$DF,$C3,$89,$3F,$C7,$9F,$CC,$7B,$7F,$BC,$F8,$19,$3F,$C2
	.dc.b	$E3,$F7,$FF,$05,$23,$F8,$73,$FC,$0B,$E1,$FF,$82,$F9,$7F,$C6,$89
	.dc.b	$3F,$CA,$9E,$AF,$F1,$5B,$EA,$FE,$62,$FA,$3F,$9A,$BB,$B7,$ED,$F0
	.dc.b	$FE,$23,$E6,$7F,$26,$25,$7F,$36,$7F,$9D,$75,$7F,$BE,$89,$F0,$DF
	.dc.b	$C6,$9F,$F7,$CC,$FF,$83,$5F,$F7,$13,$FE,$D9,$5F,$0B,$B3,$87,$C3
	.dc.b	$7F,$20,$F4,$3F,$9A,$92,$BF,$A2,$BE,$0B,$EC,$48,$FF,$6F,$DF,$83
	.dc.b	$27,$1F,$FA,$C6,$3C,$32,$7C,$2F,$0C,$7C,$2F,$DA,$F4,$BF,$A2,$91
	.dc.b	$FE,$AA,$FF,$48,$F4,$BF,$9A,$92,$3C,$3F,$0B,$5F,$E8,$31,$71,$F1
	.dc.b	$C7,$C1,$F8,$5D,$3F,$DE,$67,$A3,$FD,$89,$AF,$F7,$86,$BB,$34,$43
	.dc.b	$45,$3F,$C0,$9E,$AF,$F5,$E2,$3F,$DB,$9E,$B5,$E4,$8F,$89,$F0,$9F
	.dc.b	$64,$F4,$69,$FD,$EE,$FE,$E7,$C2,$CA,$FC,$FC,$7F,$56,$F6,$3F,$B5
	.dc.b	$92,$3F,$BD,$BF,$86,$F8,$5F,$C2,$7C,$2C,$9F,$AF,$1F,$87,$F8,$59
	.dc.b	$1F,$69,$FC,$57,$C2,$FF,$62,$8F,$EE,$E4,$3F,$C0,$E7,$53,$F0,$35
	.dc.b	$D5,$FC,$85,$74,$7F,$25,$57,$4F,$81,$C4,$C3,$FB,$8F,$A4,$F7,$7F
	.dc.b	$C2,$CD,$C3,$FC,$34,$D7,$FC,$5C,$D7,$FC,$67,$4F,$84,$FD,$11,$23
	.dc.b	$F2,$BF,$0B,$FA,$5F,$85,$FD,$37,$C2,$C9,$FC,$CC,$78,$BE,$0F,$C2
	.dc.b	$48,$F0,$BF,$B0,$BE,$91,$FA,$D7,$F6,$17,$D6,$3F,$47,$7D,$5F,$D9
	.dc.b	$5F,$BB,$E0,$FD,$A4,$8F,$6B,$F0,$3E,$13,$E1,$92,$3F,$5B,$F0,$B5
	.dc.b	$FF,$44,$68,$9C,$6F,$F3,$1D,$4F,$F9,$5A,$74,$FC,$A8,$EC,$D2,$1F
	.dc.b	$31,$CB,$4C,$9C,$DC,$BD,$6B,$EB,$91,$FA,$EF,$2A,$FB,$F7,$FC,$54
	.dc.b	$F3,$F3,$2A,$61,$E6,$7F,$8C,$6B,$FC,$6C,$DB,$AC,$FF,$0A,$47,$F0
	.dc.b	$AF,$F2,$4D,$FF,$91,$12,$3B,$63,$FA,$FF,$C4,$7F,$E0,$0C,$78,$5F
	.dc.b	$F7,$73,$AC,$7C,$8A,$3F,$DD,$CE,$4E,$1D,$3F,$E4,$88,$E3,$1F,$53
	.dc.b	$6F,$CC,$91,$FC,$CB,$E5,$7F,$93,$E1,$E5,$7F,$D8,$53,$E4,$7C,$1C
	.dc.b	$DD,$6E,$47,$84,$C9,$74,$70,$7E,$A0,$91,$FA,$87,$FD,$01,$7F,$EF
	.dc.b	$F8,$E4,$BB,$3D,$47,$AF,$A3,$D9,$DF,$2B,$6F,$BF,$C1,$FD,$BF,$2F
	.dc.b	$ED,$89,$3F,$B6,$7B,$D7,$48,$EF,$7D,$5E,$64,$3E,$0D,$F5,$7D,$9C
	.dc.b	$FF,$C0,$88,$FF,$02,$7F,$82,$F4,$7F,$08,$24,$7B,$3C,$6A,$FE,$9A
	.dc.b	$2F,$1F,$E0,$7C,$7C,$CF,$E9,$67,$9B,$FC,$39,$AF,$F0,$E6,$BB,$34
	.dc.b	$43,$45,$3F,$61,$D9,$FC,$68,$8F,$F1,$A7,$AD,$7E,$4F,$CC,$FC,$51
	.dc.b	$E6,$E6,$FC,$8F,$13,$0F,$C7,$79,$33,$1F,$C0,$47,$F2,$8E,$EF,$E4
	.dc.b	$E4,$8F,$E4,$EF,$E0,$3C,$FF,$CB,$12,$3F,$03,$EA,$7E,$67,$D5,$FC
	.dc.b	$D7,$B5,$27,$F0,$B1,$FC,$D3,$C3,$FE,$66,$48,$FE,$66,$F5,$7F,$17
	.dc.b	$7D,$5F,$CD,$DF,$57,$F1,$F7,$D5,$FC,$FD,$FF,$40,$F1,$3F,$9F,$92
	.dc.b	$3F,$9F,$BF,$98,$F6,$FF,$5D,$EE,$C9,$FC,$CC,$7E,$C7,$DE,$91,$F9
	.dc.b	$C7,$F6,$9F,$01,$FD,$33,$C9,$FE,$96,$49,$FE,$96,$F5,$7F,$47,$7D
	.dc.b	$5F,$D9,$DF,$47,$F6,$B7,$79,$3F,$85,$61,$FD,$63,$CA,$FE,$AE,$4A
	.dc.b	$FE,$AE,$FF,$5A,$F2,$FF,$AE,$92,$3F,$63,$F0,$55,$FE,$77,$16,$FF
	.dc.b	$8B,$7C,$16,$67,$F9,$D4,$F3,$7F,$B1,$B5,$FE,$C6,$D7,$66,$88,$68
	.dc.b	$A7,$F4,$5F,$3F,$FB,$79,$1F,$ED,$EF,$5A,$FE,$51,$F0,$9F,$C7,$CF
	.dc.b	$37,$4F,$E5,$AF,$FD,$D3,$F8,$97,$F8,$6E,$BF,$C3,$8D,$9D,$61,$FC
	.dc.b	$5F,$E2,$7F,$BE,$91,$FE,$FA,$FF,$30,$64,$FF,$BB,$E7,$F9,$9B,$63
	.dc.b	$FA,$43,$52,$F8,$5F,$1F,$67,$93,$E1,$65,$7F,$29,$AF,$F7,$1B,$FC
	.dc.b	$A7,$D4,$FF,$07,$91,$FF,$07,$BE,$37,$C3,$F1,$1E,$E2,$3C,$F0,$EE
	.dc.b	$F5,$BE,$EB,$3F,$87,$8F,$F4,$C7,$FC,$81,$F6,$8F,$5F,$B5,$7F,$63
	.dc.b	$8C,$38,$6B,$E3,$6B,$F1,$BE,$17,$E8,$FC,$34,$8F,$74,$7D,$4F,$86
	.dc.b	$91,$F0,$5F,$AF,$ED,$78,$F8,$3F,$AF,$E0,$F6,$FC,$76,$8C,$3E,$38
	.dc.b	$CC,$62,$3E,$37,$0B,$1F,$14,$7C,$1B,$FB,$39,$7E,$1A,$7C,$73,$93
	.dc.b	$ED,$E5,$8D,$3F,$13,$9F,$96,$7E,$27,$BB,$E0,$23,$F1,$5F,$1F,$BB
	.dc.b	$B8,$F4,$37,$43,$F9,$6F,$86,$C3,$27,$FB,$64,$E9,$FC,$37,$C3,$65
	.dc.b	$FA,$13,$D9,$FA,$8D,$72,$60,$86,$08,$7E,$76,$7B,$38,$3E,$8F,$C0
	.dc.b	$FC,$C2,$3F,$81,$7B,$65,$FA,$CF,$85,$FD,$79,$E5,$7F,$17,$7F,$B2
	.dc.b	$F8,$69,$7E,$C8,$F8,$5F,$09,$F6,$09,$1F,$8D,$7E,$FF,$13,$E1,$BF
	.dc.b	$76,$49,$FD,$47,$C3,$7E,$FA,$BF,$7F,$9C,$9F,$D6,$47,$CF,$9F,$C2
	.dc.b	$92,$3F,$34,$F6,$AE,$91,$F8,$13,$F7,$2F,$F0,$A8,$ED,$1F,$B7,$AE
	.dc.b	$CE,$ED,$7F,$76,$FF,$26,$49,$FD,$43,$EC,$C3,$F8,$AE,$F2,$7F,$81
	.dc.b	$47,$F1,$8E,$59,$1F,$C1,$5F,$E3,$7C,$DF,$89,$D3,$F3,$E4,$9F,$DA
	.dc.b	$3D,$5F,$E1,$97,$D5,$F3,$5F,$47,$D1,$77,$5F,$E4,$B8,$7E,$57,$6F
	.dc.b	$D5,$92,$BF,$80,$3F,$AE,$E0,$FC,$D1,$23,$FC,$85,$F8,$9B,$B3,$11
	.dc.b	$F2,$F8,$95,$8C,$7F,$D9,$D7,$FE,$F5,$E8,$FD,$FC,$F5,$BF,$E1,$CD
	.dc.b	$7F,$87,$35,$FC,$C3,$5E,$27,$02,$1C,$0F,$E3,$1C,$9F,$C1,$88,$E0
	.dc.b	$DD,$F9,$3C,$0F,$DC,$60,$7D,$18,$76,$AF,$C3,$4B,$BF,$66,$A3,$F9
	.dc.b	$47,$3F,$F1,$42,$3F,$CA,$5F,$E3,$72,$F0,$FB,$F9,$12,$BE,$5C,$78
	.dc.b	$BC,$99,$1F,$39,$F2,$F9,$DF,$C2,$FB,$3F,$90,$92,$7F,$9B,$3D,$AB
	.dc.b	$A4,$7A,$2F,$DB,$1D,$23,$D7,$77,$5F,$AB,$87,$F1,$7E,$EF,$E5,$A4
	.dc.b	$AF,$E8,$4F,$AB,$CF,$F8,$E4,$8F,$C1,$FA,$7F,$2B,$D1,$F9,$7E,$B4
	.dc.b	$9F,$C3,$C7,$F2,$3F,$0B,$F9,$C9,$23,$FA,$73,$D5,$FC,$6D,$F5,$7E
	.dc.b	$8D,$F5,$7F,$23,$7D,$5F,$AB,$7F,$CB,$FC,$4F,$E8,$A4,$8F,$EB,$4F
	.dc.b	$F4,$8F,$13,$F9,$A1,$23,$F3,$7E,$D5,$7F,$35,$17,$5F,$CC,$FB,$B9
	.dc.b	$9F,$E5,$53,$CD,$FE,$CC,$D7,$FB,$33,$5D,$9A,$21,$A2,$9F,$CD,$7C
	.dc.b	$7F,$EB,$A4,$7F,$B9,$3D,$6B,$F5,$7F,$03,$F9,$83,$CD,$D1,$FD,$AA
	.dc.b	$FF,$37,$F0,$53,$0F,$E9,$B1,$FD,$4F,$CA,$FE,$D2,$48,$FE,$FC,$FE
	.dc.b	$5B,$E1,$7F,$47,$F0,$D2,$7F,$7F,$1F,$A6,$F8,$69,$1F,$C0,$9F,$D4
	.dc.b	$FC,$37,$F5,$FC,$BF,$BA,$92,$7F,$C1,$4F,$57,$F8,$4D,$F5,$7F,$5B
	.dc.b	$7D,$1F,$D7,$DD,$D3,$F8,$4E,$1F,$DB,$7E,$D9,$E8,$FF,$81,$9B,$7F
	.dc.b	$E1,$A6,$BF,$E2,$96,$BF,$E3,$DE,$E7,$C3,$FF,$00,$24,$7F,$B7,$7D
	.dc.b	$EF,$F7,$11,$FE,$EC,$C5,$AF,$FB,$D2,$3F,$DE,$73,$FD,$25,$B3,$A8
	.dc.b	$08,$E5,$3F,$EF,$CC,$1E,$A7,$A4,$95,$E9,$7F,$83,$E1,$E4,$8F,$E2
	.dc.b	$2F,$F8,$36,$FC,$BF,$0F,$7E,$1E,$57,$A6,$3E,$37,$AF,$F1,$49,$1F
	.dc.b	$15,$F9,$1E,$C7,$CB,$24,$7C,$1F,$85,$AF,$F6,$5C,$D3,$8D,$7D,$D3
	.dc.b	$ED,$F8,$5D,$BF,$E6,$9A,$DB,$F4,$EB,$BB,$48,$7D,$57,$2D,$32,$73
	.dc.b	$72,$F5,$F6,$FF,$00,$47,$F0,$0F,$2B,$F2,$E8,$F8,$5F,$90,$7A,$78
	.dc.b	$3E,$96,$1F,$2F,$E1,$A6,$3F,$63,$0F,$6E,$5E,$A8,$F8,$5E,$FF,$E3
	.dc.b	$C8,$FE,$3D,$FF,$2B,$E4,$F8,$6F,$AC,$49,$FC,$67,$C3,$7D,$99,$FB
	.dc.b	$5F,$0D,$27,$F2,$11,$F9,$DF,$82,$FC,$E1,$23,$F3,$8F,$6B,$E9,$1F
	.dc.b	$91,$7F,$09,$7D,$63,$F3,$57,$D5,$FC,$B5,$FE,$A7,$E1,$3F,$50,$48
	.dc.b	$FD,$43,$EA,$F8,$6F,$C7,$7C,$34,$9F,$D2,$47,$E4,$7E,$1A,$47,$E9
	.dc.b	$DF,$CA,$57,$ED,$5F,$DA,$12,$3F,$68,$F5,$7F,$5D,$5D,$5F,$D5,$57
	.dc.b	$47,$F5,$93,$76,$FB,$77,$FC,$06,$FF,$80,$12,$BF,$80,$3F,$C0,$EF
	.dc.b	$F8,$31,$23,$F7,$9A,$57,$F3,$91,$6F,$F0,$FA,$E6,$7F,$35,$3C,$DF
	.dc.b	$E1,$8D,$7F,$86,$35,$D9,$A2,$1A,$29,$FA,$5C,$FF,$8C,$11,$FE,$30
	.dc.b	$F5,$AF,$CC,$F1,$FE,$FC,$F3,$71,$7F,$09,$BF,$E0,$5C,$53,$0F,$CC
	.dc.b	$4E,$5A,$7F,$12,$E1,$FE,$24,$FF,$25,$DF,$F9,$41,$1F,$F6,$0A,$E6
	.dc.b	$FE,$1D,$D5,$27,$F9,$24,$7F,$11,$EC,$91,$FC,$9D,$FF,$50,$7E,$29
	.dc.b	$FE,$EF,$8F,$07,$27,$F3,$52,$3F,$CD,$5F,$D3,$D7,$58,$FE,$55,$7F
	.dc.b	$C7,$63,$AC,$7F,$2D,$BB,$B7,$F1,$0C,$3B,$4B,$2F,$D8,$3F,$D0,$79
	.dc.b	$7F,$A2,$11,$C1,$0F,$E8,$8E,$5A,$64,$FE,$34,$E5,$FE,$39,$CD,$FD
	.dc.b	$3C,$8F,$F4,$F7,$FD,$AF,$4F,$0B,$E1,$F8,$D2,$BE,$98,$9D,$38,$5E
	.dc.b	$27,$CA,$24,$67,$C9,$FE,$C1,$D7,$FD,$7C,$91,$FD,$7D,$F0,$E1,$DE
	.dc.b	$3F,$D3,$96,$C3,$EA,$CF,$49,$CE,$17,$4F,$99,$8F,$F6,$DE,$CF,$ED
	.dc.b	$84,$AF,$ED,$8D,$7F,$B6,$35,$FE,$80,$D7,$FA,$F3,$5F,$C9,$3F,$DD
	.dc.b	$FB,$64,$7F,$77,$6B,$FD,$DD,$AF,$F4,$B6,$BF,$DA,$5A,$FE,$75,$AF
	.dc.b	$E0,$FB,$FF,$C0,$84,$7F,$C0,$8D,$7F,$C0,$8D,$7F,$AC,$35,$FE,$EC
	.dc.b	$D7,$F5,$4F,$F8,$3F,$BE,$47,$F8,$3D,$AF,$F8,$3D,$AF,$F6,$36,$BF
	.dc.b	$E0,$16,$BF,$B5,$6B,$F9,$B3,$FC,$38,$43,$FC,$38,$53,$FC,$38,$53
	.dc.b	$FB,$81,$4F,$F0,$71,$4F,$E0,$47,$F8,$50,$87,$F8,$5B,$EE,$3F,$C4
	.dc.b	$6E,$0C,$78,$35,$A5,$8D,$BB,$6B,$D1,$AF,$76,$BC,$8E,$11,$D5,$46
	.dc.b	$9B,$19,$8C,$0F,$0F,$04,$9F,$23,$5F,$23,$5F,$23,$5F,$23,$5F,$84
	.dc.b	$F7,$F1,$24,$7B,$DA,$FB,$DA,$FB,$DA,$FB,$DA,$FC,$D6,$BE,$EF,$1B
	.dc.b	$D8,$47,$E6,$35,$F9,$8D,$7E,$63,$5F,$98,$D7,$ED,$3F,$1F,$C6,$91
	.dc.b	$F5,$DA,$FD,$76,$BF,$5D,$AF,$D7,$6B,$F8,$56,$BF,$5B,$C7,$FA,$84
	.dc.b	$7F,$08,$D7,$F0,$8D,$7F,$08,$D7,$F0,$8D,$7F,$24,$FD,$FF,$22,$47
	.dc.b	$E3,$DA,$FE,$3D,$AF,$E3,$DA,$FE,$3D,$AF,$E7,$5A,$FE,$3B,$C9,$FC
	.dc.b	$61,$1F,$CE,$35,$FC,$E3,$5F,$CE,$35,$FC,$E3,$5F,$D5,$3F,$97,$F2
	.dc.b	$A4,$7E,$9D,$AF,$E9,$DA,$FE,$9D,$AF,$E9,$DA,$FE,$D5,$AF,$E9,$BC
	.dc.b	$CF,$D2,$11,$FD,$A3,$5F,$DA,$35,$FD,$A3,$5F,$DA,$35,$FE,$04,$FE
	.dc.b	$BF,$CC,$91,$FB,$F6,$BF,$BF,$6B,$FB,$F6,$BF,$BF,$6B,$FC,$35,$AF
	.dc.b	$EF,$BC,$DF,$DE,$11,$FE,$18,$D7,$F8,$63,$5F,$E1,$8D,$7F,$86,$35
	.dc.b	$FE,$34,$FF,$07,$F3,$A4,$7F,$17,$6B,$FC,$5D,$AF,$F1,$76,$BF,$C5
	.dc.b	$DA,$FF,$25,$6B,$FC,$5B,$D1,$FE,$28,$47,$F9,$23,$5F,$E4,$8D,$7F
	.dc.b	$92,$35,$FE,$48,$D7,$F9,$93,$FC,$7F,$D2,$91,$FC,$BD,$AF,$F2,$F6
	.dc.b	$BF,$CB,$DA,$FF,$2F,$6B,$FC,$F5,$AF,$F2,$EF,$53,$F9,$61,$1F,$E7
	.dc.b	$8D,$7F,$9E,$35,$FE,$78,$D7,$F9,$E3,$5F,$E9,$4F,$F3,$7F,$52,$47
	.dc.b	$F4,$76,$BF,$D1,$DA,$FF,$47,$6B,$FD,$1D,$AF,$F5,$56,$BF,$D1,$BD
	.dc.b	$7F,$E8,$84,$7F,$AA,$35,$FE,$A8,$D7,$FA,$A3,$5F,$EA,$8D,$7F,$B1
	.dc.b	$3F,$D3,$FD,$89,$1F,$D7,$DA,$FF,$5F,$6B,$FD,$7D,$AF,$F5,$F6,$BF
	.dc.b	$DB,$5A,$FF,$5E,$F8,$8F,$EB,$84,$7F,$B6,$35,$FE,$D8,$D7,$FB,$63
	.dc.b	$5F,$ED,$8D,$7F,$BD,$3F,$D9,$FD,$D9,$1F,$DD,$DA,$FF,$77,$6B,$FD
	.dc.b	$DD,$AF,$F7,$76,$BF,$E0,$56,$AA,$7B,$FF,$DE,$5B,$1F,$E0,$76,$BF
	.dc.b	$E0,$B6,$BF,$E1,$06,$BF,$E1,$36,$A3,$B4,$75,$23,$D5,$AF,$57,$C5
	.dc.b	$F0,$12,$3C,$4D,$7C,$4D,$7C,$4D,$7C,$4D,$7C,$52,$8C,$47,$C1,$F8
	.dc.b	$2F,$80,$47,$E0,$35,$F8,$0D,$7E,$03,$5F,$80,$D7,$BB,$F2,$FE,$0A
	.dc.b	$47,$CB,$6B,$F2,$DA,$FC,$B6,$BF,$2D,$AF,$A9,$AF,$D9,$F8,$7F,$B0
	.dc.b	$47,$EC,$35,$FB,$0D,$7E,$C3,$5F,$B0,$D7,$E3,$BF,$83,$F8,$69,$1F
	.dc.b	$83,$6B,$F8,$36,$BF,$83,$6B,$F8,$36,$BF,$49,$AF,$E4,$7E,$23,$F2
	.dc.b	$04,$7F,$20,$D7,$F2,$0D,$7F,$20,$D7,$F2,$0D,$7E,$FB,$F9,$BF,$87
	.dc.b	$91,$F9,$B6,$BF,$9B,$6B,$F9,$B6,$BF,$9B,$6B,$F8,$A6,$BF,$A9,$F8
	.dc.b	$7F,$D4,$11,$FD,$43,$5F,$D4,$35,$FD,$43,$5F,$D4,$35,$FC,$BB,$FB
	.dc.b	$3F,$87,$91,$FB,$36,$BF,$B3,$6B,$FB,$36,$BF,$B3,$6B,$FA,$26,$BF
	.dc.b	$C0,$7E,$1B,$F8,$01,$1F,$E0,$0D,$7F,$80,$35,$FE,$00,$D7,$F8,$03
	.dc.b	$5F,$D7,$BF,$C2,$FE,$1A,$47,$F0,$B6,$BF,$C2,$DA,$FF,$0B,$6B,$FC
	.dc.b	$2D,$AF,$EE,$9A,$FF,$19,$F8,$6F,$E3,$04,$7F,$8C,$35,$FE,$30,$D7
	.dc.b	$F8,$C3,$5F,$E3,$0D,$7F,$83,$BF,$C8,$FE,$1A,$47,$F2,$36,$BF,$C8
	.dc.b	$DA,$FF,$23,$6B,$FC,$8D,$AF,$F1,$26,$BF,$CC,$7E,$1B,$F9,$81,$1F
	.dc.b	$E6,$0D,$7F,$98,$35,$FE,$60,$D7,$F9,$83,$5F,$E3,$EF,$F3,$BF,$86
	.dc.b	$91,$FC,$ED,$AF,$F3,$B6,$BF,$CE,$DA,$FF,$3B,$6B,$FC,$A9,$AF,$F4
	.dc.b	$9F,$37,$FA,$41,$1F,$E9,$0D,$7F,$A4,$35,$FE,$90,$D7,$FA,$43,$5F
	.dc.b	$E6,$EF,$F5,$3F,$3E,$47,$F5,$36,$BF,$D4,$DA,$FF,$53,$6B,$FD,$4D
	.dc.b	$AF,$F4,$26,$BF,$D8,$7D,$1F,$EC,$04,$7F,$B0,$35,$FE,$C0,$D7,$FB
	.dc.b	$03,$5F,$EC,$0D,$7F,$A7,$BF,$DA,$FD,$19,$1F,$DA,$DA,$FF,$6B,$6B
	.dc.b	$FD,$AD,$AF,$F6,$B6,$BF,$D6,$9A,$FF,$79,$F7,$3F,$BC,$11,$FE,$F0
	.dc.b	$D7,$FB,$C3,$5F,$EF,$0D,$7F,$BC,$35,$FE,$CE,$FF,$BE,$FE,$E5,$EC
	.dc.b	$D2,$4F,$F7,$D2,$08,$DB,$83,$FC,$04,$E1,$C3,$5E,$2D,$7A,$35,$E3
	.dc.b	$1E,$07,$8B,$5E,$3E,$DF,$9C,$8F,$84,$F2,$38,$7C,$2D,$78,$9B,$94
	.dc.b	$66,$3C,$6D,$7D,$7E,$DF,$C5,$23,$C5,$AF,$B9,$AF,$B9,$AF,$B9,$AF
	.dc.b	$43,$C5,$EF,$7D,$12,$38,$21,$F2,$9C,$B4,$C9,$F3,$1C,$B4,$F7,$7C
	.dc.b	$07,$DD,$23,$82,$1F,$59,$CB,$4C,$9F,$61,$CB,$4F,$95,$F0,$5F,$89
	.dc.b	$23,$82,$1F,$82,$72,$D3,$27,$E1,$1C,$B4,$FA,$DF,$07,$F9,$52,$38
	.dc.b	$21,$F8,$E7,$2D,$32,$7E,$41,$CB,$4F,$C1,$7C,$2F,$E8,$48,$E0,$87
	.dc.b	$E6,$9C,$B4,$C9,$F9,$C7,$2D,$3F,$1D,$F0,$DF,$AD,$23,$82,$1F,$A6
	.dc.b	$72,$D3,$27,$EA,$1C,$B4,$FC,$D7,$C3,$7E,$E4,$8E,$08,$7E,$C9,$CB
	.dc.b	$4C,$9F,$B4,$72,$D3,$F4,$DF,$0B,$FC,$14,$8E,$08,$7E,$F9,$CB,$4C
	.dc.b	$9F,$C0,$1C,$B4,$FD,$97,$C3,$7F,$11,$23,$82,$1F,$C2,$9C,$B4,$C9
	.dc.b	$FC,$31,$CB,$4F,$DF,$7C,$37,$F1,$D2,$38,$21,$FC,$59,$CB,$4C,$9F
	.dc.b	$C6,$1C,$B4,$FE,$15,$F0,$DF,$CA,$48,$E0,$87,$F2,$27,$2D,$32,$7F
	.dc.b	$24,$72,$D3,$F8,$B7,$C2,$FF,$35,$23,$82,$1F,$CB,$9C,$B4,$C9,$FC
	.dc.b	$C1,$CB,$4F,$E4,$5F,$0B,$FD,$04,$8E,$08,$7F,$3A,$72,$D3,$27,$F3
	.dc.b	$C7,$2D,$3F,$97,$7C,$2F,$F4,$D2,$38,$21,$FD,$19,$CB,$4C,$9F,$D2
	.dc.b	$1C,$B4,$FE,$75,$F0,$7F,$D6,$48,$E0,$87,$F5,$27,$2D,$32,$7F,$54
	.dc.b	$72,$D3,$FA,$37,$C1,$FF,$65,$23,$82,$1F,$D7,$9C,$B4,$C9,$FD,$81
	.dc.b	$CB,$4F,$EA,$5F,$07,$FD,$C4,$8E,$0E,$BF,$87,$7F,$BE,$5F,$F7,$77
	.dc.b	$DD,$BE,$0B,$FB,$A4,$E6,$53,$97,$FD,$E9,$19,$E9,$61,$C3,$CF,$FE
	.dc.b	$E2,$FF,$7D,$F8,$67,$6F,$F7,$4F,$86,$C3,$71,$FF,$7F,$4E,$9F,$EF
	.dc.b	$F5,$A7,$FC,$03,$5F,$E0,$B2,$3F,$DB,$DF,$8B,$8F,$9F,$2F,$34,$A3
	.dc.b	$19,$F9,$63,$A9,$EA,$CB,$C4,$FF,$83,$2F,$AB,$4F,$EF,$87,$F8,$6C
	.dc.b	$86,$0F,$F0,$35,$F8,$2F,$BC,$75,$3E,$6D,$F8,$5F,$F0,$F6,$17,$3F
	.dc.b	$F0,$74,$7F,$87,$8E,$A7,$F8,$78,$CD,$BE,$1A,$92,$3C,$2E,$AF,$B4
	.dc.b	$F8,$F2,$FF,$17,$36,$BE,$09,$F1,$63,$33,$F3,$E3,$EF,$B6,$7E,$5E
	.dc.b	$7F,$56,$7F,$C6,$C6,$67,$C0,$FC,$26,$BF,$E3,$99,$FA,$7A,$7F,$8F
	.dc.b	$9B,$1E,$00,$75,$3C,$26,$23,$84,$F4,$8A,$77,$3E,$8F,$C2,$FC,$62
	.dc.b	$38,$21,$D6,$91,$FE,$51,$AF,$23,$67,$E3,$B5,$FD,$2B,$D1,$F6,$35
	.dc.b	$F6,$7C,$2F,$D6,$23,$F8,$46,$BE,$17,$A3,$FA,$C6,$D1,$A7,$08,$EA
	.dc.b	$73,$3F,$1F,$F0,$BF,$84,$23,$82,$1E,$F7,$A9,$1F,$98,$7E,$11,$E2
	.dc.b	$FD,$76,$BF,$5F,$E1,$7F,$22,$47,$F3,$ED,$7E,$63,$EA,$23,$D8,$FE
	.dc.b	$1A,$ED,$E8,$72,$3F,$53,$F0,$BF,$9E,$23,$82,$1F,$65,$F8,$E4,$7B
	.dc.b	$1F,$C5,$9E,$2F,$E4,$5A,$FE,$47,$E1,$7F,$56,$47,$F7,$2D,$7F,$0C
	.dc.b	$FD,$42,$3D,$B0,$46,$E3,$A1,$C8,$FE,$07,$F0,$BF,$B8,$23,$83,$47
	.dc.b	$EC,$4F,$D0,$FD,$E3,$CB,$7B,$9F,$C2,$65,$39,$8C,$61,$E2,$9B,$87
	.dc.b	$ED,$FE,$1A,$47,$F2,$CD,$FF,$80,$CE,$23,$F6,$93,$FC,$30,$8F,$93
	.dc.b	$9B,$FA,$06,$3D,$E3,$CC,$78,$B2,$F2,$B8,$C7,$C4,$DB,$E6,$F6,$3F
	.dc.b	$92,$BF,$C6,$BE,$1A,$47,$F2,$86,$BF,$EB,$FB,$F0,$61,$E2,$C3,$BB
	.dc.b	$73,$F7,$D7,$C2,$BE,$2F,$F0,$D6,$BF,$C8,$3E,$1B,$F9,$49,$1F,$E6
	.dc.b	$EF,$F3,$0C,$3B,$F2,$FF,$B4,$DB,$3F,$53,$1F,$0C,$FC,$A8,$FA,$35
	.dc.b	$CD,$FE,$3C,$D7,$F8,$F7,$C3,$7F,$38,$23,$FD,$19,$F0,$57,$48,$ED
	.dc.b	$8F,$DF,$C3,$C3,$1F,$DC,$35,$FE,$77,$7C,$A3,$F9,$63,$5F,$E5,$9F
	.dc.b	$0D,$FD,$18,$8F,$F5,$47,$BD,$F5,$C3,$C1,$3F,$6E,$3E,$5B,$F6,$E7
	.dc.b	$8B,$FC,$E5,$AF,$F3,$9F,$86,$FE,$A6,$47,$FB,$0B,$D9,$FC,$3E,$3E
	.dc.b	$A3,$D7,$BF,$E1,$B6,$FE,$ED,$5C,$6B,$FA,$33,$5F,$E8,$DF,$0D,$FD
	.dc.b	$80,$8F,$F6,$E7,$B4,$FE,$0E,$3B,$E5,$EC,$BE,$DB,$7F,$60,$C3,$FB
	.dc.b	$1B,$2B,$F5,$31,$FD,$9B,$E1,$BF,$B6,$92,$3F,$BD,$BD,$A7,$F0,$97
	.dc.b	$F2,$32,$F5,$ED,$CE,$FF,$45,$57,$2F,$6C,$7E,$DA,$3F,$BC,$91,$FF
	.dc.b	$03,$BD,$5F,$C0,$9E,$AD,$7E,$9C,$FF,$86,$E7,$FA,$29,$27,$F8,$93
	.dc.b	$FD,$23,$E1,$BF,$C0,$A4,$8F,$F0,$9B,$D5,$FB,$C7,$53,$F7,$17,$FE
	.dc.b	$06,$8E,$51,$FC,$75,$AF,$F6,$7F,$86,$FF,$08,$91,$FF,$0F,$BD,$71
	.dc.b	$FB,$D9,$7D,$1B,$FF,$08,$CF,$F5,$C2,$57,$F2,$97,$FA,$FE,$5F,$E1
	.dc.b	$C2,$47,$F8,$A9,$E9,$5F,$3C,$E9,$1F,$C8,$23,$8B,$FC,$CD,$AF,$F8
	.dc.b	$26,$DF,$0D,$FB,$D2,$47,$F8,$38,$87,$F2,$C7,$BC,$FE,$94,$FA,$39
	.dc.b	$7F,$03,$8F,$EE,$F5,$D2,$BF,$A7,$B5,$F0,$46,$BE,$1B,$F8,$61,$23
	.dc.b	$B9,$FC,$F9,$EE,$FD,$3D,$3F,$CB,$19,$FF,$91,$6D,$5C,$EB,$BB,$5F
	.dc.b	$77,$C3,$7F,$19,$23,$D8,$FE,$8E,$F6,$BF,$D3,$1D,$A3,$F8,$74,$7B
	.dc.b	$08,$FB,$1F,$7F,$C3,$7F,$25,$24,$75,$B4,$75,$CB,$F4,$B9,$FC,$8A
	.dc.b	$FE,$B7,$7E,$BD,$3E,$D4,$F8,$27,$F8,$F7,$C3,$7F,$31,$23,$D5,$E8
	.dc.b	$FE,$93,$2F,$43,$F1,$F0,$EE,$D3,$B9,$FC,$A7,$E1,$BF,$9E,$11,$F4
	.dc.b	$3D,$1F,$D1,$CF,$43,$E2,$CF,$46,$C7,$43,$F8,$DF,$C3,$7F,$47,$23
	.dc.b	$F0,$DE,$9A,$7F,$04,$8F,$C5,$1E,$53,$BD,$7F,$4D,$CF,$EF,$7C,$37
	.dc.b	$F5,$12,$4F,$CD,$79,$E9,$FA,$08,$E5,$7F,$AE,$D3,$E5,$47,$F2,$EF
	.dc.b	$86,$FE,$B4,$47,$EB,$3C,$A3,$EA,$47,$88,$F9,$98,$FB,$CF,$23,$FD
	.dc.b	$47,$E1,$BF,$3E,$47,$F3,$EF,$E8,$B0,$FB,$71,$ED,$E7,$FE,$5E,$F4
	.dc.b	$21,$DC,$FC,$99,$FD,$77,$E1,$BF,$56,$47,$F5,$6F,$5C,$3E,$4E,$DF
	.dc.b	$4A,$7F,$B1,$18,$C7,$D0,$F6,$CD,$AF,$EA,$7E,$1B,$F6,$E4,$9F,$DB
	.dc.b	$BD,$5F,$97,$9F,$CD,$C3,$B5,$FC,$D9,$FE,$B3,$7F,$B3,$F8,$6F,$E0
	.dc.b	$24,$8F,$E0,$2F,$3A,$E3,$7C,$F1,$F2,$D7,$C1,$C3,$FB,$16,$1F,$BD
	.dc.b	$F8,$6F,$E1,$64,$9F,$E1,$6F,$3A,$E3,$5F,$8B,$9F,$3E,$7F,$05,$FC
	.dc.b	$A6,$3F,$E3,$6F,$86,$FE,$2A,$48,$FE,$2A,$F3,$7E,$36,$5E,$56,$D7
	.dc.b	$37,$EA,$9F,$DE,$3E,$1B,$F8,$F9,$1F,$E3,$EF,$37,$D9,$3C,$DB,$3C
	.dc.b	$F0,$FA,$71,$FA,$23,$F9,$7B,$FE,$4B,$6C,$FF,$92,$9B,$3F,$C3,$63
	.dc.b	$11,$FB,$E3,$27,$F9,$15,$CB,$FD,$A0,$C9,$0E,$04,$34,$53,$FB,$19
	.dc.b	$B3,$24,$32,$6E,$7D,$0D,$8F,$0B,$FE,$74,$AF,$8C,$FF,$51,$6F,$89
	.dc.b	$E8,$C7,$F0,$B8,$7E,$CB,$0F,$C7,$47,$E4,$34,$FE,$81,$97,$D9,$C7
	.dc.b	$FA,$D3,$70,$E5,$9F,$E1,$33,$FA,$18,$F2,$8F,$C3,$E1,$F8,$DC,$3F
	.dc.b	$16,$43,$FB,$26,$9F,$66,$B8,$47,$FB,$C3,$83,$F7,$2F,$DA,$3F,$87
	.dc.b	$D7,$E8,$5F,$CF,$90,$FE,$DD,$B7,$3C,$7F,$81,$E9,$FC,$AF,$2F,$36
	.dc.b	$BF,$83,$C3,$F6,$98,$7E,$C8,$87,$F7,$7D,$F9,$4F,$E4,$F2,$F1,$E7
	.dc.b	$E6,$AF,$D5,$5F,$F6,$5E,$0F,$E1,$11,$FD,$FD,$BB,$7C,$68,$FC,$E9
	.dc.b	$FE,$1E,$3F,$29,$A7,$9B,$3F,$E3,$F1,$FE,$56,$E1,$FE,$3C,$FF,$82
	.dc.b	$5B,$5C,$E3,$E2,$6B,$ED,$7A,$BF,$65,$FF,$3C,$71,$7F,$31,$7F,$D3
	.dc.b	$4D,$79,$67,$EB,$D3,$DB,$97,$8E,$7F,$8D,$69,$FD,$11,$FE,$84,$43
	.dc.b	$FD,$49,$C7,$FC,$ED,$F6,$1F,$C1,$0F,$AE,$7F,$A2,$79,$3F,$AA,$BF
	.dc.b	$EA,$B6,$BE,$2A,$F0,$3F,$CB,$0E,$59,$7A,$DF,$F3,$CF,$2F,$F6,$97
	.dc.b	$FD,$68,$D7,$97,$07,$D7,$E9,$FE,$FC,$FC,$E3,$FA,$AD,$7F,$B2,$B9
	.dc.b	$FF,$BE,$D7,$FA,$F5,$AF,$23,$FA,$55,$72,$7C,$7C,$3F,$5A,$BF,$DA
	.dc.b	$7D,$1F,$E1,$29,$FF,$63,$35,$E5,$1F,$D1,$F6,$FC,$4E,$DC,$E3,$D8
	.dc.b	$7F,$B7,$1F,$F1,$3C,$FF,$B3,$DA,$F2,$BF,$E7,$35,$CA,$3F,$C4,$84
	.dc.b	$3F,$C6,$27,$F8,$D0,$FF,$12,$74,$FF,$8F,$DF,$F6,$D3,$F5,$EF,$F1
	.dc.b	$11,$FE,$FA,$8F,$F1,$D7,$17,$F2,$46,$C7,$F9,$3C,$FF,$17,$75,$7F
	.dc.b	$96,$1F,$F7,$23,$CF,$0F,$F6,$94,$7F,$AB,$A3,$FC,$AA,$47,$FC,$AA
	.dc.b	$FF,$99,$1F,$F2,$FE,$1F,$C6,$13,$4C,$71,$7E,$BC,$FE,$A1,$8A,$FF
	.dc.b	$CD,$13,$FC,$DF,$1E,$7B,$7F,$9C,$1B,$87,$F9,$F4,$FF,$42,$90,$FF
	.dc.b	$30,$CB,$AF,$D5,$1F,$C4,$76,$FF,$43,$BC,$4F,$F4,$81,$6B,$D3,$D9
	.dc.b	$FE,$9D,$8E,$6D,$7E,$C1,$FE,$10,$BF,$E3,$62,$3F,$D3,$73,$CA,$7F
	.dc.b	$D3,$C7,$FA,$A8,$FF,$59,$90,$F3,$E3,$E6,$7F,$4D,$97,$FA,$CD,$E0
	.dc.b	$27,$E6,$CF,$FA,$D6,$30,$7C,$8E,$DF,$F6,$63,$F0,$9A,$FA,$0F,$F1
	.dc.b	$36,$1E,$89,$FF,$65,$1F,$40,$FF,$4C,$5F,$FB,$32,$4F,$D5,$7F,$9D
	.dc.b	$77,$FF,$B9,$9F,$9E,$D7,$B6,$BF,$72,$BF,$DC,$91,$FE,$D5,$21,$D1
	.dc.b	$CB,$4F,$E8,$DD,$FF,$EF,$57,$27,$E3,$1F,$8D,$87,$D0,$75,$1C,$0F
	.dc.b	$F4,$A4,$23,$FD,$2D,$1F,$92,$F0,$2C,$9A,$48,$FF,$02,$6F,$FC,$35
	.dc.b	$FE,$43,$97,$C1,$79,$1D,$A7,$83,$F9,$07,$AF,$83,$23,$C6,$D7,$FC
	.dc.b	$19,$5D,$6B,$F6,$F6,$AD,$57,$EA,$9A,$FE,$93,$C2,$F4,$91,$F7,$BF
	.dc.b	$13,$83,$EC,$E7,$B9,$C9,$0F,$D2,$CF,$C2,$F0,$E4,$7C,$86,$BE,$63
	.dc.b	$F9,$29,$FA,$90,$7D,$73,$F7,$AF,$07,$F8,$0B,$5F,$DD,$F8,$9F,$40
	.dc.b	$8F,$D6,$7E,$9D,$FF,$29,$BF,$E4,$6F,$AA,$7F,$C1,$13,$C1,$FD,$EB
	.dc.b	$5F,$E1,$FE,$2F,$DC,$23,$F8,$27,$AB,$F3,$EB,$F8,$DE,$7F,$EB,$EF
	.dc.b	$33,$F5,$33,$FC,$5A,$FF,$DF,$11,$8D,$11,$8E,$3F,$E4,$73,$9C,$BA
	.dc.b	$DF,$F1,$06,$BF,$C8,$9F,$8E,$F6,$3F,$B1,$4E,$5F,$96,$D7,$56,$9E
	.dc.b	$71,$FC,$C8,$D1,$FC,$BB,$C7,$FC,$D9,$1F,$D0,$B5,$E6,$F8,$AB,$84
	.dc.b	$78,$0F,$E5,$F1,$96,$9C,$88,$7F,$2C,$7F,$53,$E4,$48,$FD,$7B,$5F
	.dc.b	$E6,$2F,$37,$FC,$80,$D7,$FA,$1B,$6C,$D3,$94,$7F,$4A,$6B,$FD,$13
	.dc.b	$C9,$FD,$C9,$1F,$DF,$B5,$F9,$2F,$52,$3E,$F3,$FC,$A4,$ED,$FE,$AE
	.dc.b	$D7,$FA,$8F,$95,$FC,$18,$8F,$F0,$C6,$BF,$59,$F5,$11,$EC,$7F,$60
	.dc.b	$76,$FF,$64,$6B,$FD,$7B,$CB,$FE,$26,$47,$F8,$CB,$5F,$C1,$BF,$1C
	.dc.b	$8F,$6B,$46,$E3,$FB,$8B,$5F,$ED,$7E,$67,$F2,$02,$3F,$C9,$9A,$FE
	.dc.b	$41,$FA,$84,$7B,$3A,$7F,$BE,$B5,$FE,$EF,$E6,$FF,$2C,$23,$FC,$C9
	.dc.b	$EB,$BF,$F2,$9B,$57,$EF,$6B,$FC,$0A,$DF,$39,$E7,$E6,$B7,$5F,$C4
	.dc.b	$38,$3F,$69,$7F,$B3,$8F,$F4,$63,$55,$FE,$C1,$4F,$EF,$A3,$FA,$1C
	.dc.b	$68,$FE,$8A,$E0,$CD,$78,$23,$F8,$F3,$70,$F5,$5F,$9F,$0F,$45,$72
	.dc.b	$21,$E1,$3F,$93,$DF,$5F,$43,$B9,$23,$47,$A1,$F1,$CF,$C1,$A9,$3F
	.dc.b	$CC,$2F,$FA,$95,$FC,$68,$C7,$1F,$F0,$E8,$61,$D6,$7C,$51,$87,$BB
	.dc.b	$5C,$FA,$3F,$08,$8E,$9F,$1B,$D8,$F9,$E5,$3E,$C9,$63,$C8,$FF,$45
	.dc.b	$6E,$7F,$46,$BE,$B4,$78,$63,$EA,$65,$83,$BD,$7A,$1A,$FD,$B9,$EA
	.dc.b	$D8,$CB,$F0,$5A,$FD,$06,$BF,$84,$7E,$FF,$A5,$23,$E1,$B5,$F8,$ED
	.dc.b	$7C,$0D,$7E,$73,$5F,$BA,$D7,$F2,$3E,$9F,$E3,$08,$FD,$06,$BF,$51
	.dc.b	$AF,$B5,$AF,$DB,$6B,$F8,$A7,$F2,$FE,$A4,$8F,$B6,$D7,$EF,$B5,$F9
	.dc.b	$0D,$7F,$0C,$D7,$F2,$AD,$7F,$53,$EA,$FE,$90,$8F,$E2,$1A,$FE,$31
	.dc.b	$AF,$D5,$6B,$F9,$36,$BF,$A2,$7F,$5F,$EB,$48,$FC,$9B,$5F,$CB,$B5
	.dc.b	$FC,$03,$5F,$CF,$35,$FD,$6B,$5F,$E0,$3E,$BF,$EF,$08,$FE,$81,$AF
	.dc.b	$E9,$1A,$FE,$35,$AF,$EA,$DA,$FE,$E9,$FE,$0F,$EC,$48,$FD,$5B,$5F
	.dc.b	$D7,$B5,$FC,$C3,$5F,$DB,$35,$FE,$0A,$D7,$F8,$CF,$B3,$FC,$50,$8F
	.dc.b	$EE,$1A,$FE,$F1,$AF,$E9,$5A,$FF,$03,$6B,$FC,$49,$FE,$3F,$ED,$48
	.dc.b	$FE,$06,$D7,$F8,$3B,$5F,$D8,$35,$FE,$1C,$D7,$F8,$EB,$5F,$E6,$3E
	.dc.b	$DF,$F2,$C2,$3F,$C4,$1A,$FF,$14,$6B,$FB,$D6,$BF,$C6,$DA,$FF,$2A
	.dc.b	$7F,$9B,$FB,$92,$3F,$8D,$B5,$FE,$3E,$D7,$F8,$43,$5F,$E4,$CD,$7F
	.dc.b	$9A,$B5,$FE,$93,$EE,$FF,$44,$23,$FC,$A1,$AF,$F2,$C6,$BF,$C5,$5A
	.dc.b	$FF,$33,$6B,$FD,$09,$FE,$9F,$EF,$48,$FE,$66,$D7,$F9,$BB,$5F,$E4
	.dc.b	$0D,$7F,$9F,$35,$FE,$9A,$D7,$FB,$0F,$BF,$FD,$70,$8F,$F4,$06,$BF
	.dc.b	$D1,$1A,$FF,$2D,$6B,$FD,$2D,$AF,$F5,$A7,$FB,$3F,$C0,$48,$FE,$96
	.dc.b	$D7,$FA,$7B,$5F,$E7,$0D,$7F,$AB,$35,$FE,$CA,$D7,$FB,$CF,$C0,$FF
	.dc.b	$74,$23,$FD,$61,$AF,$F5,$C6,$BF,$D1,$5A,$FF,$63,$6B,$FD,$C9,$FE
	.dc.b	$FF,$F0,$52,$3F,$B1,$B5,$FE,$CE,$D7,$FA,$83,$5F,$ED,$CD,$7F,$BE
	.dc.b	$B5,$FF,$08,$FC,$1F,$F8,$30,$8F,$F7,$06,$BF,$DD,$1A,$FF,$5D,$6B
	.dc.b	$FD,$ED,$AF,$F8,$29,$FF,$6D,$F5,$7F,$A2,$E3,$FD,$1A,$D7,$FD,$6A
	.dc.b	$D4,$D2,$78,$9C,$89,$1F,$E1,$28,$CB,$63,$2F,$F8,$6E,$1F,$09,$E2
	.dc.b	$24,$71,$6B,$B6,$BF,$E3,$38,$E2,$EC,$86,$CA,$6F,$E1,$7D,$E4,$7C
	.dc.b	$CD,$7C,$8D,$78,$1A,$21,$A2,$9E,$3F,$86,$F9,$64,$7E,$13,$5F,$80
	.dc.b	$D7,$81,$A2,$1A,$29,$EF,$F7,$3E,$B9,$1F,$9C,$D7,$E6,$35,$E0,$68
	.dc.b	$86,$8A,$7C,$BF,$86,$FC,$19,$1F,$B4,$D7,$EC,$35,$E0,$68,$86,$8A
	.dc.b	$7D,$7F,$86,$FC,$79,$1F,$C3,$35,$FC,$23,$5E,$06,$88,$68,$A7,$E0
	.dc.b	$FE,$1B,$F3,$64,$7F,$24,$D7,$F2,$0D,$78,$1A,$21,$A2,$9F,$8F,$F8
	.dc.b	$6F,$D3,$91,$FC,$F3,$5F,$CE,$35,$E0,$68,$86,$8A,$7E,$6F,$E1,$BF
	.dc.b	$66,$47,$F5,$4D,$7F,$50,$D7,$81,$A2,$1A,$29,$FA,$7F,$86,$FD,$F9
	.dc.b	$1F,$DB,$35,$FD,$A3,$5E,$06,$88,$68,$A7,$EC,$FE,$1B,$F8,$59,$1F
	.dc.b	$E0,$4D,$7F,$80,$35,$E0,$68,$86,$8A,$7E,$FF,$E1,$BF,$8B,$91,$FE
	.dc.b	$1C,$D7,$F8,$63,$5E,$06,$88,$68,$A7,$F0,$BF,$86,$FE,$46,$47,$F8
	.dc.b	$D3,$5F,$E3,$0D,$78,$1A,$21,$A2,$9F,$C5,$FE,$1B,$F9,$79,$1F,$E4
	.dc.b	$CD,$7F,$92,$35,$E0,$68,$86,$8A,$7F,$23,$F8,$6F,$E7,$64,$7F,$99
	.dc.b	$35,$FE,$60,$D7,$81,$A2,$1A,$29,$FC,$BF,$E1,$BF,$A3,$91,$FE,$7C
	.dc.b	$D7,$F9,$E3,$5E,$06,$88,$68,$A7,$F3,$BF,$86,$FE,$A6,$47,$FA,$53
	.dc.b	$5F,$E9,$0D,$78,$1A,$21,$A2,$9F,$D1,$FE,$1B,$FA,$F9,$1F,$EA,$CD
	.dc.b	$7F,$AA,$35,$E0,$68,$86,$8A,$7F,$53,$F8,$6F,$ED,$64,$7F,$B1,$35
	.dc.b	$FE,$C0,$D7,$81,$A2,$1A,$29,$FD,$7F,$E1,$BF,$BB,$91,$FE,$DC,$D7
	.dc.b	$FB,$63,$5E,$06,$88,$68,$A0,$83,$E1,$DF,$01,$88,$91,$87,$07,$17
	.dc.b	$0E,$1A,$F0,$8D,$BB,$6B,$DD,$B1,$C2,$3C,$6D,$78,$1C,$9C,$3C,$4C
	.dc.b	$7C,$3F,$22,$38,$21,$E5,$72,$D3,$27,$B9,$CB,$4F,$07,$C3,$F9,$88
	.dc.b	$E0,$87,$C1,$72,$D3,$27,$CA,$72,$D3,$D9,$D5,$F0,$88,$E0,$87,$CD
	.dc.b	$72,$D3,$27,$D6,$72,$D3,$E4,$7C,$3F,$CE,$23,$82,$1F,$65,$CB,$4C
	.dc.b	$9F,$82,$72,$D3,$EA,$7C,$3F,$DA,$23,$82,$1F,$85,$72,$D3,$27,$E3
	.dc.b	$9C,$B4,$FC,$07,$C3,$7E,$18,$8E,$08,$7E,$45,$CB,$4C,$9F,$9A,$72
	.dc.b	$D3,$F1,$9F,$0D,$F9,$22,$38,$21,$F9,$D7,$2D,$32,$7E,$99,$CB,$4F
	.dc.b	$CC,$7C,$37,$E7,$88,$E0,$87,$EA,$5C,$B4,$C9,$FB,$27,$2D,$3F,$49
	.dc.b	$F0,$DF,$AA,$23,$82,$1F,$B5,$72,$D3,$27,$EF,$9C,$B4,$FD,$87,$C3
	.dc.b	$7E,$D8,$8E,$08,$7F,$01,$72,$D3,$27,$F0,$A7,$2D,$3F,$79,$F0,$BF
	.dc.b	$C0,$88,$E0,$87,$F0,$D7,$2D,$32,$7F,$16,$72,$D3,$F8,$47,$C3,$7F
	.dc.b	$0E,$23,$82,$1F,$C6,$5C,$B4,$C9,$FC,$89,$CB,$4F,$E2,$9F,$0D,$FC
	.dc.b	$68,$8E,$08,$7F,$25,$72,$D3,$27,$F2,$E7,$2D,$3F,$90,$7C,$37,$F2
	.dc.b	$62,$38,$21,$FC,$C5,$CB,$4C,$9F,$CE,$9C,$B4,$FE,$59,$F0,$DF,$CC
	.dc.b	$88,$E0,$87,$F3,$D7,$2D,$32,$7F,$46,$72,$D3,$F9,$C7,$C3,$7F,$3E
	.dc.b	$23,$82,$1F,$D2,$5C,$B4,$C9,$FD,$49,$CB,$4F,$E8,$9F,$0D,$FD,$28
	.dc.b	$8E,$08,$7F,$55,$72,$D3,$27,$F5,$E7,$2D,$3F,$A8,$7C,$37,$F5,$62
	.dc.b	$38,$21,$FD,$85,$CB,$4C,$9F,$DA,$9C,$B4,$FE,$B9,$F0,$9F,$D8,$88
	.dc.b	$E0,$87,$F6,$D7,$2D,$32,$7F,$76,$72,$D3,$FB,$47,$C3,$7F,$6E,$23
	.dc.b	$82,$1F,$DE,$5C,$B4,$C9,$FE,$02,$72,$D3,$FB,$EF,$C4,$FF,$7C,$4D
	.dc.b	$23,$FE,$00,$6B,$C0,$C1,$0C,$10,$FF,$07,$BC,$FE,$13,$91,$23,$C2
	.dc.b	$D7,$6D,$7F,$C3,$50,$8C,$47,$F8,$82,$30,$F2,$21,$E0,$F8,$5F,$41
	.dc.b	$1F,$7B,$5D,$18,$21,$82,$1E,$27,$83,$E4,$F8,$3F,$88,$47,$E5,$B5
	.dc.b	$F8,$2D,$78,$1A,$21,$A2,$9F,$03,$E0,$FE,$81,$1F,$AE,$D7,$E6,$B5
	.dc.b	$E0,$68,$86,$8A,$7C,$CF,$83,$FB,$84,$7F,$06,$FF,$8F,$BE,$25,$E9
	.dc.b	$FF,$93,$BE,$19,$EA,$FF,$94,$25,$EA,$E6,$30,$DC,$3F,$CA,$D3,$C9
	.dc.b	$FF,$2C,$57,$27,$3E,$AE,$D4,$A3,$0D,$8E,$37,$E1,$F5,$78,$6C,$F8
	.dc.b	$5F,$CB,$12,$BF,$3A,$FA,$F3,$E7,$9C,$9E,$67,$9A,$3B,$E5,$23,$27
	.dc.b	$91,$F1,$E1,$E0,$F8,$5F,$D2,$12,$7F,$56,$D7,$F4,$2D,$7F,$36,$D7
	.dc.b	$F3,$4D,$7F,$CF,$98,$A3,$F6,$1F,$0B,$27,$F6,$ED,$7F,$CF,$7F,$0A
	.dc.b	$F7,$3E,$52,$F7,$29,$88,$CF,$B9,$2B,$DB,$48,$F9,$FB,$FC,$C8,$FE
	.dc.b	$00,$49,$FE,$0E,$FF,$A0,$8F,$03,$CA,$FB,$BD,$2F,$B2,$E1,$FC,$02
	.dc.b	$AF,$03,$E2,$C3,$C2,$FD,$0C,$DF,$0B,$FC,$44,$91,$FC,$61,$FE,$29
	.dc.b	$F0,$BF,$03,$E0,$24,$F1,$8F,$C2,$9F,$13,$E0,$24,$64,$87,$F1,$68
	.dc.b	$E6,$FF,$24,$6B,$93,$04,$30,$43,$F8,$BB,$C1,$FE,$33,$F0,$BF,$CA
	.dc.b	$48,$FF,$30,$7A,$E7,$F1,$31,$57,$A2,$BE,$AF,$C1,$D7,$E5,$6F,$FC
	.dc.b	$86,$71,$7F,$33,$57,$C1,$FE,$93,$E1,$7F,$9C,$13,$0F,$E7,$EF,$D9
	.dc.b	$BF,$C3,$E7,$27,$F0,$71,$F8,$A9,$91,$F8,$67,$F1,$BF,$0B,$FB,$03
	.dc.b	$FA,$21,$23,$FA,$5B,$D5,$FC,$6D,$75,$7D,$27,$FB,$67,$0F,$7B,$D4
	.dc.b	$86,$8F,$C3,$57,$81,$FC,$26,$7F,$D4,$C8,$FF,$5A,$7F,$29,$C9,$FA
	.dc.b	$22,$47,$E7,$B9,$3F,$4B,$7F,$A6,$E1,$93,$FA,$28,$FE,$17,$A7,$F5
	.dc.b	$F2,$47,$F6,$67,$AB,$FA,$8B,$EA,$F9,$E5,$7D,$A3,$F5,$B7,$D9,$D4
	.dc.b	$FF,$18,$E1,$FE,$D8,$48,$FE,$E6,$FF,$6F,$E1,$FD,$DF,$5C,$9F,$D2
	.dc.b	$47,$F2,$03,$F7,$FD,$52,$32,$43,$FB,$8C,$73,$7F,$BF,$B5,$C9,$82
	.dc.b	$18,$21,$FD,$C9,$E0,$FF,$74,$E5,$FF,$02,$11,$FF,$05,$BD,$78,$BF
	.dc.b	$39,$3E,$79,$FE,$1F,$DB,$5F,$BD,$CA,$9C,$57,$E6,$FB,$7F,$9E,$F4
	.dc.b	$7F,$84,$49,$7F,$E1,$84,$C2,$0F,$8A,$78,$3F,$D3,$5B,$1F,$CF,$1A
	.dc.b	$FD,$93,$F9,$4F,$83,$FC,$AB,$C1,$91,$FC,$9A,$3F,$A7,$F6,$7F,$88
	.dc.b	$C9,$1F,$E2,$A7,$C2,$F8,$EB,$AE,$5F,$C8,$1B,$D9,$9C,$7F,$9B,$D7
	.dc.b	$80,$FE,$67,$8F,$81,$FC,$66,$3F,$D9,$7E,$2F,$F9,$4C,$7B,$9A,$FC
	.dc.b	$16,$BF,$11,$AF,$C3,$F1,$BE,$17,$8D,$23,$E1,$47,$C5,$F1,$A4,$7C
	.dc.b	$57,$E4,$78,$DE,$1F,$13,$DB,$3D,$5A,$B0,$FE,$FC,$FC,$26,$BF,$53
	.dc.b	$56,$1E,$18,$FD,$F4,$31,$FA,$D3,$E2,$9C,$E3,$76,$F9,$59,$7C,$3F
	.dc.b	$17,$C3,$97,$F8,$4A,$FC,$AD,$7C,$51,$AF,$17,$F0,$5E,$44,$9F,$EF
	.dc.b	$B5,$FE,$0D,$7F,$88,$37,$0F,$B1,$E4,$7E,$24,$FA,$57,$2B,$BB,$D5
	.dc.b	$A4,$3E,$13,$96,$99,$3E,$D3,$96,$9E,$AF,$1F,$DA,$47,$07,$5A,$FE
	.dc.b	$D7,$7E,$7E,$3F,$C1,$5F,$E7,$FC,$AF,$35,$FB,$22,$DF,$E4,$FC,$AF
	.dc.b	$C6,$7C,$3F,$CA,$25,$7C,$A7,$EE,$F1,$3C,$FF,$D5,$12,$7F,$55,$E7
	.dc.b	$FE,$B7,$CF,$FD,$77,$9F,$27,$F5,$D1,$F9,$9F,$87,$FA,$E4,$8F,$AE
	.dc.b	$F6,$BE,$91,$F8,$47,$F7,$17,$D6,$3F,$71,$7D,$5F,$C4,$DF,$E9,$7E
	.dc.b	$1F,$F0,$64,$8F,$C1,$BE,$AF,$57,$F8,$17,$AB,$27,$F8,$14,$7F,$05
	.dc.b	$F5,$64,$7F,$05,$7F,$84,$7A,$BF,$B1,$F8,$7F,$C7,$92,$7F,$1E,$F5
	.dc.b	$7F,$86,$5F,$57,$F3,$B7,$D1,$FC,$FD,$DE,$1F,$E0,$98,$7E,$F7,$E1
	.dc.b	$FF,$36,$4A,$FC,$DB,$F9,$EF,$87,$FE,$35,$ED,$49,$FC,$CC,$7F,$05
	.dc.b	$3F,$8F,$7B,$52,$32,$43,$F6,$91,$CD,$FD,$3B,$5C,$98,$21,$82,$1F
	.dc.b	$C4,$5E,$0F,$F1,$3F,$87,$FD,$99,$1F,$D9,$BD,$71,$FC,$BC,$F9,$E7
	.dc.b	$F9,$97,$BD,$5F,$D8,$C5,$9F,$E5,$1E,$EF,$F2,$1F,$87,$FD,$F9,$2B
	.dc.b	$F7,$EF,$F1,$2F,$8C,$7C,$27,$FA,$2D,$7C,$25,$31,$19,$F8,$49,$7F
	.dc.b	$E7,$EA,$47,$1F,$84,$FE,$61,$F1,$1F,$C3,$89,$5F,$C3,$9F,$05,$F5
	.dc.b	$8E,$58,$76,$7A,$5F,$67,$FC,$39,$87,$67,$C1,$87,$F3,$DF,$88,$FE
	.dc.b	$36,$48,$FE,$36,$FB,$71,$F6,$12,$3D,$11,$EF,$F8,$8F,$81,$F1,$12
	.dc.b	$7D,$91,$FD,$24,$FE,$4E,$47,$F9,$3B,$D5,$F8,$55,$D5,$F3,$D7,$57
	.dc.b	$E3,$57,$57,$D7,$5F,$D5,$63,$F9,$99,$23,$F9,$9B,$FC,$DB,$3F,$A9
	.dc.b	$C3,$27,$F9,$8C,$7F,$5D,$3E,$C6,$B2,$32,$43,$FA,$64,$73,$7F,$9F
	.dc.b	$B5,$C9,$82,$18,$21,$FD,$A5,$E0,$FF,$6B,$D7,$FA,$59,$1F,$E9,$6F
	.dc.b	$5D,$FF,$97,$CF,$9E,$7F,$11,$C9,$5F,$E9,$31,$67,$FD,$3F,$B3,$87
	.dc.b	$FB,$D7,$07,$F5,$82,$5F,$F5,$87,$F3,$5E,$0F,$E6,$7C,$09,$3F,$95
	.dc.b	$8F,$CE,$76,$48,$FC,$C3,$F9,$FE,$CF,$F0,$37,$27,$F6,$42,$4F,$F6
	.dc.b	$47,$AB,$F9,$FB,$EA,$FE,$52,$FA,$3F,$96,$BB,$9F,$AB,$0F,$F7,$B7
	.dc.b	$86,$E7,$FF,$00,$B6,$FF,$C0,$AD,$7F,$C4,$6D,$7F,$C6,$BA,$FF,$81
	.dc.b	$AF,$E2,$F9,$FF,$B7,$3F,$71,$E5,$FE,$E6,$FF,$75,$E4,$CC,$3F,$6D
	.dc.b	$1E,$2E,$AF,$09,$23,$C2,$FF,$03,$BE,$91,$FB,$D7,$AB,$C8,$FE,$01
	.dc.b	$5D,$5F,$DA,$D7,$BB,$B3,$DA,$48,$F6,$BF,$03,$B3,$FC,$A1,$97,$FB
	.dc.b	$73,$E3,$7F,$DB,$75,$FE,$99,$6C,$7F,$B8,$2D,$5F,$28,$FE,$2D,$E4
	.dc.b	$CB,$FF,$2D,$D6,$E7,$51,$F3,$1F,$03,$48,$7C,$E7,$2D,$32,$7F,$99
	.dc.b	$2B,$C2,$E0,$FF,$34,$38,$7E,$BF,$83,$F7,$08,$FD,$C7,$AD,$79,$63
	.dc.b	$CF,$7C,$AF,$E1,$BF,$E7,$3C,$1E,$D7,$AA,$BF,$D7,$98,$F7,$C3,$94
	.dc.b	$7C,$7F,$5B,$F1,$BE,$27,$E3,$09,$1F,$8C,$7F,$C9,$F8,$3E,$0F,$BC
	.dc.b	$FF,$A0,$A3,$FD,$C2,$F9,$0C,$7C,$1E,$94,$71,$74,$63,$DF,$D4,$1A
	.dc.b	$63,$DB,$FD,$0F,$91,$FA,$02,$57,$E8,$1F,$2E,$1C,$B0,$93,$DB,$0F
	.dc.b	$03,$F8,$5C,$3A,$B9,$7B,$3E,$4D,$1E,$47,$EB,$89,$1F,$AE,$7E,$1F
	.dc.b	$C3,$FA,$09,$1D,$FE,$1B,$D7,$F0,$DE,$CF,$87,$93,$E3,$8F,$DD,$FA
	.dc.b	$1F,$BA,$24,$7E,$E9,$EA,$FA,$6F,$AB,$F9,$DB,$EA,$FB,$6F,$AB,$FA
	.dc.b	$1B,$FE,$0F,$E8,$7F,$06,$24,$7F,$06,$7F,$84,$F9,$FF,$43,$E1,$24
	.dc.b	$FC,$18,$FE,$1E,$7D,$4F,$85,$91,$92,$1F,$C3,$A3,$9B,$FC,$49,$AE
	.dc.b	$4C,$10,$C1,$0F,$D3,$3C,$1F,$D4,$7A,$5F,$C7,$88,$FF,$1E,$7A,$F3
	.dc.b	$FE,$F6,$7C,$F3,$F7,$FE,$1E,$BF,$A0,$8B,$3F,$C1,$FE,$1F,$F9,$5F
	.dc.b	$AF,$FC,$A8,$95,$FC,$A9,$FC,$FF,$C2,$FE,$53,$E1,$24,$FE,$2E,$3F
	.dc.b	$2F,$F0,$F2,$3F,$1C,$FE,$6B,$E1,$FF,$9B,$FB,$7F,$CD,$89,$3F,$CD
	.dc.b	$9E,$AF,$E5,$6F,$AB,$FC,$66,$FA,$3F,$C6,$EE,$E5,$FC,$7F,$0F,$E8
	.dc.b	$7E,$F7,$F4,$22,$57,$F4,$27,$FD,$8B,$E6,$7F,$B8,$23,$FD,$EE,$D4
	.dc.b	$DE,$BF,$F7,$CD,$FF,$2B,$F8,$3F,$D9,$FC,$24,$BF,$D5,$47,$F5,$2F
	.dc.b	$80,$FE,$A2,$48,$FE,$A2,$F8,$3E,$17,$F8,$06,$3D,$CE,$D3,$FB,$EA
	.dc.b	$E8,$FF,$37,$BC,$3F,$B1,$F8,$4F,$EB,$A4,$7F,$AE,$BF,$D8,$3E,$17
	.dc.b	$AF,$C2,$C9,$FD,$C4,$7F,$66,$3C,$3F,$0B,$23,$24,$3F,$B2,$C7,$37
	.dc.b	$FB,$4B,$5C,$98,$21,$82,$1F,$CF,$5E,$0F,$F3,$F9,$FE,$EA,$47,$FB
	.dc.b	$AB,$D7,$4F,$EA,$93,$E7,$9F,$E4,$79,$7F,$BD,$9F,$34,$E9,$BE,$59
	.dc.b	$FB,$75,$FE,$04,$C3,$FC,$06,$48,$FF,$01,$BF,$82,$E5,$FA,$64,$8E
	.dc.b	$FC,$BF,$5B,$87,$EB,$F0,$49,$F1,$C7,$F8,$43,$83,$FC,$1E,$48,$FF
	.dc.b	$07,$BD,$5F,$4D,$F5,$59,$B3,$F4,$47,$E4,$31,$F0,$3F,$03,$1F,$03
	.dc.b	$C2,$77,$A3,$2F,$E9,$5F,$1D,$CA,$A0,$6B,$D9,$AF,$C7,$6B,$FA,$0F
	.dc.b	$0B,$F2,$BD,$D2,$3E,$8C,$7E,$63,$AA,$47,$D4,$7F,$37,$D9,$E2,$E5
	.dc.b	$EB,$81,$A4,$AF,$1E,$5F,$EF,$97,$EE,$E3,$DD,$F7,$1F,$B4,$C7,$BB
	.dc.b	$83,$F6,$B8,$E3,$6F,$95,$1F,$BA,$CB,$E1,$F2,$F8,$89,$1F,$0B,$34
	.dc.b	$79,$BA,$7F,$67,$E4,$4A,$FE,$E3,$49,$F9,$A7,$EE,$3C,$49,$3A,$8C
	.dc.b	$BD,$5E,$8D,$21,$F5,$27,$A3,$83,$EC,$38,$75,$1C,$9A,$6C,$FB,$5D
	.dc.b	$9F,$38,$8F,$B5,$E1,$97,$C4,$8F,$C1,$4F,$99,$F9,$5B,$3C,$EC,$56
	.dc.b	$F0,$F1,$36,$FF,$8B,$F9,$7F,$8E,$EE,$FB,$A4,$C7,$E6,$3F,$CB,$3D
	.dc.b	$5F,$E3,$A4,$8F,$DC,$FA,$7F,$C8,$3D,$1F,$E4,$3E,$8C,$9F,$DE,$C7
	.dc.b	$E6,$BC,$2F,$C4,$92,$3E,$C3,$D5,$FE,$07,$7D,$5F,$5D,$F5,$7F,$83
	.dc.b	$DF,$57,$DF,$7F,$A6,$F1,$3F,$2A,$48,$FC,$23,$FA,$4F,$8F,$FF,$7B
	.dc.b	$C7,$F8,$C9,$FF,$7B,$B7,$DF,$FF,$1C,$46,$A3,$F2,$8F,$FB,$A5,$BE
	.dc.b	$E6,$8D,$FB,$9F,$B6,$F2,$3F,$4A,$49,$FC,$A3,$E1,$F7,$BA,$E1,$27
	.dc.b	$C5,$1C,$DF,$25,$D7,$C3,$7F,$C0,$BC,$BF,$D8,$92,$3F,$40,$FE,$D3
	.dc.b	$CB,$F4,$FC,$1C,$9F,$CF,$C7,$F0,$A3,$DB,$F0,$72,$32,$43,$F3,$51
	.dc.b	$CD,$FD,$63,$5C,$98,$21,$82,$1F,$A0,$78,$3F,$A2,$F4,$7F,$84,$91
	.dc.b	$FD,$C3,$D7,$9F,$F2,$D3,$E7,$9F,$AD,$F0,$B5,$FE,$0D,$D2,$9C,$57
	.dc.b	$E8,$F8,$1F,$0B,$FC,$A3,$D3,$FE,$2E,$4C,$3F,$83,$3F,$83,$F8,$3F
	.dc.b	$BE,$48,$FC,$27,$C1,$FE,$0B,$E1,$7F,$07,$F0,$B2,$7F,$11,$1F,$CD
	.dc.b	$3D,$6F,$E4,$64,$8F,$E2,$4F,$57,$F1,$D7,$D5,$FC,$5D,$F5,$7F,$25
	.dc.b	$7D,$5F,$C7,$DF,$F4,$0F,$5F,$F9,$79,$23,$F8,$F3,$F9,$BF,$84,$FC
	.dc.b	$BF,$C2,$C9,$FC,$E4,$7E,$6B,$E1,$64,$7E,$7D,$FC,$EF,$8D,$FD,$33
	.dc.b	$DE,$FE,$76,$49,$FE,$54,$F5,$7F,$4D,$7D,$5F,$D1,$DF,$47,$F4,$B7
	.dc.b	$78,$7E,$96,$1F,$D6,$3D,$EF,$E8,$E4,$AF,$E6,$CF,$F4,$AF,$7B,$F5
	.dc.b	$FE,$F4,$9F,$E6,$B1,$FD,$80,$FD,$9F,$C1,$48,$C9,$0F,$E6,$11,$CD
	.dc.b	$FE,$84,$D7,$26,$08,$60,$87,$F3,$67,$83,$FC,$E3,$E0,$3F,$AF,$91
	.dc.b	$FE,$9C,$F5,$C7,$F9,$64,$F9,$E7,$F8,$67,$C1,$D7,$F2,$F1,$67,$F2
	.dc.b	$DF,$09,$FD,$F3,$E1,$3F,$B5,$92,$BF,$AD,$3F,$DE,$BE,$43,$FD,$B7
	.dc.b	$1F,$EC,$86,$BF,$D9,$DB,$5F,$D0,$D7,$C2,$FF,$B4,$57,$C4,$FF,$1C
	.dc.b	$F7,$65,$F3,$8F,$F0,$6D,$7F,$7C,$24,$7F,$6B,$7C,$58,$73,$24,$78
	.dc.b	$E3,$B4,$F4,$8F,$E4,$F5,$87,$FB,$7C,$7F,$82,$08,$FF,$77,$7D,$DC
	.dc.b	$9E,$7E,$09,$3E,$F8,$F5,$69,$23,$E1,$3E,$DD,$BF,$DD,$DF,$23,$FD
	.dc.b	$8E,$7F,$C4,$CD,$7F,$C6,$2D,$7F,$C9,$4D,$5A,$7F,$B9,$B5,$F9,$58
	.dc.b	$F8,$A7,$E7,$63,$E1,$7E,$86,$36,$BE,$86,$3E,$2D,$FF,$C6,$39,$1A
	.dc.b	$4B,$FF,$1B,$C3,$7F,$AD,$D7,$27,$FC,$BD,$87,$F7,$DA,$D4,$63,$6F
	.dc.b	$5D,$FD,$DE,$79,$39,$3D,$B1,$F1,$0F,$19,$1F,$0B,$5F,$F2,$5D,$76
	.dc.b	$6C,$65,$F9,$6D,$7C,$27,$23,$D7,$C7,$F0,$08,$E0,$D1,$FE,$3D,$9E
	.dc.b	$D8,$7B,$6B,$F2,$DE,$17,$9E,$BD,$42,$B7,$8F,$82,$3F,$41,$DB,$F8
	.dc.b	$0E,$5F,$9E,$48,$F9,$AF,$EA,$7C,$BF,$D1,$12,$3F,$55,$E5,$FE,$97
	.dc.b	$C7,$FD,$37,$87,$27,$F5,$D1,$F8,$CE,$CF,$B6,$48,$FB,$2F,$57,$F6
	.dc.b	$97,$D5,$FD,$A5,$F5,$7F,$71,$7D,$5F,$DC,$5F,$E6,$3B,$7F,$0E,$48
	.dc.b	$FC,$2B,$FC,$07,$D3,$FD,$D7,$AB,$27,$F8,$14,$7E,$F7,$D3,$91,$FC
	.dc.b	$15,$FE,$01,$E9,$FE,$93,$C1,$FC,$99,$27,$F2,$2F,$57,$F8,$65,$F5
	.dc.b	$7F,$86,$5F,$47,$F8,$75,$DE,$4F,$E1,$D8,$7E,$C3,$C1,$FC,$F9,$2B
	.dc.b	$F3,$AF,$E8,$BC,$5F,$E2,$5E,$EC,$9F,$BB,$CA,$9F,$DC,$9F,$C5,$FD
	.dc.b	$99,$3A,$8C,$BF,$A8,$7A,$34,$87,$E6,$9C,$B4,$C9,$FC,$11,$CB,$4F
	.dc.b	$D9,$F8,$FF,$B8,$23,$83,$AE,$5F,$B3,$9F,$1C,$7F,$A5,$FE,$49,$F0
	.dc.b	$78,$AD,$4F,$A4,$FE,$70,$7F,$16,$F2,$3F,$82,$91,$FE,$06,$FF,$AE
	.dc.b	$2B,$FD,$C5,$1D,$7E,$1F,$FD,$93,$5F,$E7,$F9,$CB,$7E,$1B,$3D,$9F
	.dc.b	$EC,$E9,$7C,$3E,$30,$FF,$74,$43,$E1,$F1,$08,$C3,$FB,$07,$8B,$F5
	.dc.b	$BC,$DF,$E3,$64,$7F,$8C,$BE,$7C,$B9,$CF,$93,$0F,$14,$63,$1C,$E3
	.dc.b	$AC,$66,$3D,$E3,$F9,$CF,$A3,$FC,$A0,$91,$FC,$99,$F8,$9F,$0F,$DB
	.dc.b	$E1,$E4,$F9,$E3,$C3,$F0,$D2,$3D,$4F,$93,$E1,$FF,$A2,$FA,$7F,$CD
	.dc.b	$09,$3F,$CC,$9E,$AF,$C1,$BE,$AF,$DD,$BE,$8F,$DF,$BB,$C1,$F4,$F0
	.dc.b	$FE,$A3,$EB,$FF,$40,$25,$7F,$3E,$7F,$A1,$FA,$FF,$1B,$E1,$64,$FE
	.dc.b	$8A,$3F,$AC,$9F,$2B,$E1,$64,$64,$87,$F3,$68,$E6,$FF,$4A,$6B,$93
	.dc.b	$04,$30,$43,$FB,$2B,$C1,$FE,$CF,$EE,$7F,$58,$23,$FD,$59,$EB,$8F
	.dc.b	$F5,$39,$F3,$CF,$E1,$FE,$1A,$BF,$CF,$62,$CF,$E5,$FE,$1F,$FB,$AF
	.dc.b	$BF,$FD,$90,$95,$FD,$89,$FD,$07,$C2,$FE,$28,$91,$F9,$3F,$85,$FC
	.dc.b	$6F,$C3,$FE,$3B,$E1,$E4,$FE,$5E,$3F,$C0,$3F,$07,$FD,$C0,$91,$FD
	.dc.b	$B9,$EA,$FE,$76,$FA,$BF,$B9,$BE,$AF,$E8,$6F,$AB,$FB,$DB,$FF,$06
	.dc.b	$FC,$1F,$F7,$C2,$47,$F7,$A7,$F7,$1F,$0F,$FA,$2F,$87,$93,$FA,$B8
	.dc.b	$FD,$2F,$C3,$48,$FD,$73,$FA,$8F,$87,$FD,$9F,$C9,$A9,$50,$82,$34
	.dc.b	$E0,$DB,$87,$0D,$78,$F8,$59,$E1,$FE,$15,$8F,$9A,$7F,$8D,$5F,$95
	.dc.b	$FE,$39,$8D,$AF,$E2,$18,$F9,$AF,$C1,$5E,$06,$BE,$56,$E2,$6D,$FF
	.dc.b	$10,$E3,$FF,$75,$60,$F0,$BD,$D1,$87,$E0,$5A,$BE,$29,$C7,$49,$8F
	.dc.b	$F8,$1A,$74,$D9,$EE,$F9,$A3,$2F,$C4,$8C,$B4,$D3,$87,$E7,$C7,$06
	.dc.b	$C6,$5F,$2B,$EE,$CA,$47,$76,$BD,$46,$3E,$F7,$EB,$CF,$B1,$FF,$0F
	.dc.b	$E2,$E5,$C4,$70,$CB,$BB,$ED,$37,$C7,$F8,$8E,$2F,$9E,$48,$F8,$0F
	.dc.b	$D4,$B7,$77,$D1,$24,$F8,$3B,$BE,$A7,$57,$D5,$E8,$93,$E4,$8F,$CA
	.dc.b	$F3,$7D,$C2,$47,$CD,$7B,$5F,$48,$F4,$1F,$9A,$C3,$C2,$F7,$3A,$1F
	.dc.b	$02,$FB,$BC,$04,$2F,$F4,$BC,$BF,$8C,$24,$FD,$D7,$DD,$E3,$7E,$37
	.dc.b	$C7,$93,$F3,$23,$F2,$1E,$34,$8F,$9E,$FE,$4F,$C6,$FD,$8F,$57,$E6
	.dc.b	$09,$3F,$89,$7A,$BF,$5A,$FA,$BE,$7C,$BF,$C3,$4F,$42,$19,$BE,$CF
	.dc.b	$E5,$E3,$F4,$A4,$7F,$2C,$FE,$A3,$B3,$F4,$DE,$84,$9F,$CF,$47,$F0
	.dc.b	$63,$F5,$5E,$64,$8C,$90,$FE,$03,$1C,$DF,$D1,$35,$C9,$82,$18,$21
	.dc.b	$FA,$27,$83,$FA,$4F,$03,$F7,$A4,$7F,$5C,$F5,$E8,$FC,$F4,$7F,$17
	.dc.b	$9F,$43,$F9,$EF,$5A,$BF,$7F,$AD,$38,$AE,$39,$3D,$2F,$E4,$BE,$17
	.dc.b	$F0,$C2,$61,$FB,$D7,$F3,$DE,$F7,$F0,$DF,$6E,$4F,$EB,$23,$F8,$87
	.dc.b	$B9,$23,$F5,$EF,$F1,$3F,$6B,$F9,$8F,$93,$FC,$60,$93,$FC,$25,$EA
	.dc.b	$FE,$DA,$FA,$BF,$81,$BE,$8F,$E0,$EE,$E7,$EA,$C3,$F9,$EF,$95,$FC
	.dc.b	$90,$95,$FC,$55,$FE,$53,$F2,$9F,$EF,$AF,$07,$FA,$03,$67,$FA,$1B
	.dc.b	$53,$FE,$FE,$7F,$D4,$EF,$FB,$F9,$FE,$17,$1F,$CB,$CF,$F1,$FC,$66
	.dc.b	$3E,$87,$97,$FC,$D8,$8F,$F2,$77,$C7,$F0,$FC,$09,$1D,$BE,$1F,$9F
	.dc.b	$C3,$F5,$8E,$E7,$89,$FE,$B7,$E7,$7F,$42,$24,$7F,$33,$7F,$A2,$FA
	.dc.b	$5E,$4F,$84,$93,$FC,$EE,$3F,$B1,$9E,$8F,$84,$91,$92,$1F,$D7,$A3
	.dc.b	$9B,$FC,$FD,$AE,$4C,$10,$C1,$0F,$E7,$EF,$07,$FA,$0F,$A9,$FD,$68
	.dc.b	$8F,$F4,$B7,$AF,$07,$E7,$BA,$7F,$CC,$13,$FD,$CA,$BE,$7F,$C2,$FA
	.dc.b	$A3,$CF,$5B,$9F,$A7,$F0,$DF,$E0,$0F,$57,$FB,$39,$2B,$FA,$C3,$F8
	.dc.b	$4F,$84,$FB,$3F,$09,$27,$EF,$C7,$DC,$F8,$69,$1F,$82,$7E,$FF,$C3
	.dc.b	$7F,$83,$3D,$AF,$EE,$64,$9F,$EC,$8F,$57,$F1,$37,$D5,$FE,$65,$CC
	.dc.b	$C3,$AC,$63,$0B,$A7,$EF,$A9,$8F,$F8,$6B,$DB,$FF,$00,$92,$FF,$B9
	.dc.b	$3F,$9E,$F8,$6F,$C9,$92,$3F,$33,$F0,$DF,$96,$F8,$5F,$CB,$FC,$34
	.dc.b	$9F,$CE,$C7,$FA,$63,$E1,$BF,$C1,$F0,$69,$1E,$C7,$E9,$2F,$B3,$EC
	.dc.b	$BE,$CF,$EA,$2F,$B3,$F0,$2F,$B7,$C1,$FF,$87,$30,$34,$93,$FE,$22
	.dc.b	$87,$C1,$7E,$BB,$E1,$64,$FF,$8E,$B6,$FE,$47,$5A,$8C,$78,$1F,$E3
	.dc.b	$D8,$C4,$7D,$0F,$85,$FD,$B9,$EF,$C2,$47,$77,$C8,$47,$04,$3F,$C8
	.dc.b	$93,$D1,$FF,$24,$B5,$F0,$47,$17,$6D,$77,$F0,$DF,$04,$8F,$BD,$E1
	.dc.b	$87,$B2,$7C,$B3,$F8,$89,$B3,$98,$F7,$9F,$E5,$CA,$F7,$C7,$E2,$5F
	.dc.b	$B8,$7C,$F2,$47,$CD,$7F,$C9,$BF,$2B,$FE,$EE,$8F,$E3,$E7,$F1,$C3
	.dc.b	$FD,$B6,$E5,$FC,$CE,$7F,$C8,$08,$FF,$26,$C3,$F1,$59,$FD,$D2,$47
	.dc.b	$DB,$7B,$F0,$75,$8E,$E7,$A8,$5F,$7C,$7F,$45,$87,$69,$F0,$9D,$CE
	.dc.b	$18,$B3,$FC,$61,$23,$F1,$4F,$BB,$A7,$D6,$48,$F0,$60,$EC,$F7,$F4
	.dc.b	$FC,$0E,$99,$5A,$8F,$D3,$72,$7E,$64,$91,$F9,$77,$B3,$D3,$0E,$CF
	.dc.b	$F8,$C7,$0E,$CF,$33,$C3,$7D,$9D,$5F,$EC,$F9,$3F,$4C,$48,$FD,$23
	.dc.b	$FA,$9E,$3F,$AB,$E1,$C9,$F8,$71,$FB,$B3,$EC,$F8,$12,$32,$43,$F3
	.dc.b	$71,$CD,$FD,$83,$5C,$98,$21,$82,$1F,$A2,$78,$3F,$A4,$E6,$FD,$F1
	.dc.b	$1F,$DE,$3D,$78,$7F,$73,$3E,$79,$FA,$7E,$4D,$7F,$CB,$1C,$29,$C5
	.dc.b	$7F,$88,$B5,$78,$BF,$C6,$7B,$3F,$86,$13,$0F,$E1,$4F,$E7,$BC,$CF
	.dc.b	$CD,$F9,$B2,$7F,$03,$1F,$9E,$F3,$E4,$7E,$11,$FD,$0F,$9B,$FC,$97
	.dc.b	$C1,$FE,$30,$49,$FE,$2C,$F5,$7F,$17,$7D,$5F,$C0,$5F,$47,$F0,$57
	.dc.b	$73,$F5,$61,$FC,$C7,$C1,$FE,$48,$4A,$FE,$44,$FE,$DB,$DD,$FD,$99
	.dc.b	$23,$F3,$3E,$E7,$ED,$BD,$9F,$DB,$FA,$F2,$7F,$3B,$1F,$CF,$7C,$4F
	.dc.b	$E6,$04,$8F,$E5,$CF,$57,$F4,$77,$D5,$FC,$C5,$F5,$7F,$4F,$7D,$5F
	.dc.b	$CE,$5F,$F4,$9F,$1F,$F9,$E1,$23,$F9,$D3,$FC,$FF,$C7,$FE,$17,$F0
	.dc.b	$12,$7F,$73,$1F,$D4,$4F,$E1,$FF,$03,$23,$24,$3F,$98,$C7,$37,$FA
	.dc.b	$33,$5C,$98,$21,$82,$1F,$CE,$1E,$0F,$F3,$AF,$33,$FA,$A1,$1F,$EA
	.dc.b	$4F,$5D,$FF,$A7,$CF,$9E,$7F,$84,$FC,$25,$7F,$2F,$16,$7F,$2D,$F0
	.dc.b	$9F,$DB,$7C,$EF,$EC,$04,$AF,$EB,$CF,$FB,$FB,$E5,$BF,$DA,$D1,$FC
	.dc.b	$8D,$7C,$1A,$30,$D7,$FD,$79,$1C,$1B,$F0,$F9,$C3,$FD,$0F,$3F,$F0
	.dc.b	$06,$79,$35,$2F,$86,$FF,$01,$7A,$BF,$DD,$49,$3F,$DC,$DF,$37,$C3
	.dc.b	$75,$F4,$BC,$57,$E6,$AF,$29,$E9,$9D,$3E,$0C,$6B,$E0,$6C,$F8,$CF
	.dc.b	$14,$7F,$3C,$F5,$BF,$C0,$C4,$7F,$C0,$8F,$C7,$F8,$5E,$24,$8F,$91
	.dc.b	$F0,$DD,$3E,$17,$AF,$C3,$49,$F9,$91,$FE,$1F,$F6,$BF,$C2,$44,$8F
	.dc.b	$F0,$83,$D5,$FA,$57,$D5,$F9,$F7,$D5,$FA,$D7,$D5,$FA,$77,$EE,$F9
	.dc.b	$7F,$F5,$BC,$7D,$70,$E1,$FB,$4D,$7E,$F3,$5D,$FB,$DF,$1B,$E1,$24
	.dc.b	$7F,$67,$94,$F7,$3E,$5F,$C3,$49,$D4,$65,$F1,$1D,$9A,$D2,$CF,$46
	.dc.b	$C6,$9F,$19,$82,$1D,$23,$83,$63,$84,$62,$3E,$13,$5F,$83,$F0,$3A
	.dc.b	$23,$E2,$78,$CF,$A6,$1C,$9F,$EE,$2A,$EE,$FE,$7B,$E1,$F4,$E6,$AF
	.dc.b	$1F,$E5,$BE,$1F,$EA,$7C,$1F,$90,$98,$FC,$07,$F6,$5F,$0B,$F9,$4F
	.dc.b	$85,$93,$FB,$38,$FC,$BF,$C3,$C8,$FD,$B3,$F9,$AF,$87,$FC,$07,$C3
	.dc.b	$FC,$02,$4F,$CC,$7A,$BF,$BD,$BE,$AF,$EE,$AF,$A3,$FB,$CB,$BA,$7E
	.dc.b	$EB,$0F,$C6,$57,$CC,$25,$7D,$87,$F8,$56,$9F,$AC,$24,$7F,$0B,$DF
	.dc.b	$F5,$F9,$FE,$C3,$39,$3F,$C3,$E3,$F3,$19,$FD,$82,$47,$E1,$1E,$AF
	.dc.b	$F1,$5B,$EA,$FF,$12,$BE,$AF,$F1,$9B,$EA,$FF,$16,$BF,$D2,$67,$F8
	.dc.b	$42,$47,$E4,$1F,$C3,$E5,$FC,$1B,$DA,$FF,$76,$68,$DB,$F9,$04,$61
	.dc.b	$FD,$4C,$FE,$BC,$FE,$16,$CB,$FD,$34,$65,$FC,$EB,$D9,$A4,$3F,$34
	.dc.b	$E5,$A6,$4F,$CC,$B9,$69,$FA,$8D,$FF,$3E,$47,$07,$5B,$FC,$C4,$F8
	.dc.b	$E3,$F3,$75,$EC,$EC,$C5,$6A,$7D,$27,$F9,$97,$07,$4F,$F0,$DE,$1F
	.dc.b	$D6,$92,$7F,$72,$FE,$DE,$5E,$37,$CB,$F0,$A5,$73,$8F,$9D,$E2,$48
	.dc.b	$EC,$FD,$1F,$0B,$F8,$D7,$47,$EE,$89,$3F,$C1,$9E,$D5,$D2,$3D,$67
	.dc.b	$F5,$7C,$18,$7D,$B9,$ED,$3A,$12,$C3,$39,$FD,$F9,$E5,$8F,$F2,$BE
	.dc.b	$8F,$E1,$64,$8F,$E2,$EF,$BF,$D2,$FC,$41,$23,$E4,$7A,$5F,$8B,$F3
	.dc.b	$BF,$19,$E7,$49,$F9,$91,$FC,$DF,$C0,$FE,$2E,$48,$FE,$46,$F5,$7E
	.dc.b	$95,$F5,$7D,$5A,$30,$ED,$1F,$5F,$0E,$CE,$B0,$FE,$89,$E0,$7F,$24
	.dc.b	$24,$7F,$30,$7F,$93,$F7,$FE,$87,$D5,$93,$FC,$FE,$3F,$A6,$1F,$A5
	.dc.b	$F6,$24,$64,$87,$F3,$38,$E6,$FF,$3C,$6B,$93,$04,$30,$43,$F9,$BB
	.dc.b	$C1,$FE,$73,$E1,$7F,$3C,$23,$FD,$21,$EB,$C1,$FC,$D6,$7C,$F3,$F9
	.dc.b	$9F,$7A,$BF,$81,$F0,$D3,$8A,$FC,$04,$FE,$13,$6F,$ED,$5E,$3F,$F4
	.dc.b	$A2,$47,$F5,$67,$F3,$DF,$0B,$FC,$1F,$D9,$93,$FA,$78,$FE,$15,$F0
	.dc.b	$92,$3F,$54,$FF,$0D,$F8,$2F,$EE,$DE,$6F,$F5,$62,$4F,$F6,$27,$AB
	.dc.b	$FB,$1B,$EA,$FE,$02,$FA,$3F,$82,$BB,$97,$AB,$2F,$F6,$A4,$FF,$77
	.dc.b	$F3,$3F,$B1,$91,$FE,$DE,$FF,$6D,$F9,$8F,$F7,$07,$85,$FE,$3D,$6C
	.dc.b	$FF,$90,$5A,$FE,$A0,$FE,$4D,$F0,$9F,$C9,$FE,$1A,$47,$F0,$78,$FF
	.dc.b	$09,$7A,$1F,$DC,$89,$1F,$DF,$9F,$07,$C3,$7F,$8D,$2F,$A9,$DB,$2F
	.dc.b	$F0,$C6,$7E,$BC,$3A,$E1,$FA,$0C,$7C,$2F,$FB,$E7,$E6,$5E,$A7,$F8
	.dc.b	$1D,$B1,$FE,$0E,$6B,$FE,$12,$6B,$FE,$25,$5E,$A7,$8F,$E1,$24,$F5
	.dc.b	$8F,$F2,$4C,$EA,$3F,$CC,$98,$FE,$23,$37,$C2,$FA,$CD,$DC,$AF,$13
	.dc.b	$E5,$23,$82,$1E,$57,$2D,$32,$79,$DC,$BE,$9F,$63,$E0,$11,$F8,$0F
	.dc.b	$2B,$FC,$EE,$1E,$99,$FE,$44,$D8,$F5,$7C,$37,$9E,$73,$76,$7D,$DF
	.dc.b	$0F,$F0,$BD,$AF,$9C,$4A,$F9,$CF,$E2,$FE,$17,$EE,$7C,$24,$9F,$8F
	.dc.b	$CB,$F8,$D9,$FC,$07,$C3,$C9,$C9,$F8,$F7,$F0,$9F,$0F,$F4,$3E,$0B
	.dc.b	$EE,$12,$7E,$E3,$DD,$E2,$38,$7F,$DE,$AE,$1A,$FF,$41,$57,$E1,$8C
	.dc.b	$5D,$D3,$EE,$61,$F8,$0F,$82,$FC,$61,$2B,$F1,$8F,$EA,$7E,$17,$F3
	.dc.b	$44,$8F,$07,$C3,$7E,$77,$E1,$BF,$3D,$F0,$D2,$7C,$91,$F8,$CF,$85
	.dc.b	$FC,$C1,$23,$F3,$0F,$57,$D5,$7D,$5F,$55,$F5,$7D,$D7,$D5,$F7,$5F
	.dc.b	$E6,$3E,$1B,$F4,$84,$8F,$D2,$3F,$A7,$BF,$D9,$61,$27,$F5,$91,$FA
	.dc.b	$03,$F6,$D3,$23,$24,$3F,$59,$1C,$DF,$D8,$35,$C9,$82,$18,$21,$FB
	.dc.b	$37,$83,$FB,$5C,$3F,$78,$47,$F7,$8F,$5C,$FF,$3B,$3E,$79,$FC,$06
	.dc.b	$F5,$FC,$FC,$59,$FD,$C6,$7F,$BC,$CB,$F8,$41,$2B,$F8,$43,$FC,$97
	.dc.b	$AB,$F8,$B7,$2C,$9F,$C6,$47,$F1,$9E,$89,$1F,$8F,$7F,$8E,$72,$FF
	.dc.b	$08,$DF,$F8,$A1,$27,$F8,$A3,$D5,$FC,$B5,$F5,$7F,$2D,$7D,$1F,$CC
	.dc.b	$5D,$CB,$F8,$86,$1F,$C5,$38,$BF,$90,$12,$BF,$90,$3F,$EA,$9F,$9A
	.dc.b	$FF,$7B,$57,$F5,$E6,$F2,$FF,$BE,$DB,$5F,$EF,$F7,$FD,$D8,$FF,$89
	.dc.b	$DF,$F7,$C1,$FE,$E6,$71,$E0,$7F,$BB,$23,$FD,$F9,$D5,$FA,$E9,$FE
	.dc.b	$59,$D1,$FC,$CC,$8F,$FB,$02,$7B,$A6,$93,$C5,$E9,$E1,$75,$C2,$57
	.dc.b	$EF,$AB,$BF,$4F,$F3,$F2,$47,$F3,$F7,$FA,$17,$4F,$97,$C8,$93,$FD
	.dc.b	$1E,$3F,$9D,$9E,$9F,$2E,$46,$48,$7F,$47,$8E,$6F,$F4,$B6,$B9,$30
	.dc.b	$43,$04,$3F,$A7,$3C,$1F,$EA,$1D,$DF,$D5,$C8,$FF,$57,$7A,$EF,$FC
	.dc.b	$E2,$7C,$F3,$F0,$7D,$3A,$FF,$3A,$8B,$3F,$77,$CF,$FE,$AF,$E0,$7F
	.dc.b	$63,$25,$7F,$63,$7F,$09,$EF,$7D,$9F,$6E,$4F,$CD,$8F,$B9,$ED,$C8
	.dc.b	$FA,$0F,$DF,$F5,$FF,$B1,$F8,$DF,$DB,$C9,$3F,$DB,$DE,$AF,$D7,$BE
	.dc.b	$AF,$E2,$EF,$A3,$F8,$DB,$B9,$7E,$57,$0F,$ED,$FE,$37,$F7,$B2,$57
	.dc.b	$F7,$B7,$F3,$9F,$0B,$F9,$12,$47,$E1,$7E,$17,$F2,$9F,$03,$F9,$5F
	.dc.b	$81,$93,$F8,$98,$FE,$F7,$E7,$7F,$81,$C9,$1F,$E0,$77,$AB,$F8,$FB
	.dc.b	$EA,$FE,$8E,$FA,$BF,$93,$BE,$AF,$E9,$EC,$FF,$54,$F9,$CA,$C2,$B5
	.dc.b	$C3,$5C,$94,$2F,$3B,$F6,$1F,$07,$27,$B4,$76,$D7,$FC,$B7,$5B,$9F
	.dc.b	$D4,$7C,$27,$ED,$CD,$78,$3E,$2B,$B3,$E8,$AE,$0F,$95,$AF,$9A,$3C
	.dc.b	$91,$88,$E4,$7A,$5C,$3C,$E3,$64,$74,$D7,$B3,$D3,$D0,$91,$CD,$AE
	.dc.b	$8F,$79,$0F,$E2,$DF,$0D,$EA,$7E,$3C,$DA,$F5,$E3,$F1,$23,$F8,$C7
	.dc.b	$C3,$67,$D5,$F5,$92,$FD,$4F,$D1,$F9,$C7,$C3,$F5,$7F,$DC,$73,$FE
	.dc.b	$10,$6B,$FE,$CD,$76,$E7,$E2,$38,$11,$FE,$35,$AF,$97,$D9,$F9,$64
	.dc.b	$8F,$94,$F8,$6F,$C1,$F1,$12,$7C,$11,$CD,$F0,$5D,$D3,$F9,$F6,$1F
	.dc.b	$07,$DA,$FA,$E4,$AF,$AC,$FB,$7E,$1F,$D8,$48,$F6,$7C,$47,$BF,$E2
	.dc.b	$3E,$07,$C4,$49,$F8,$11,$F3,$7D,$FF,$C1,$92,$3F,$04,$F5,$7E,$35
	.dc.b	$F5,$7E,$35,$F5,$7E,$55,$F5,$7E,$55,$FD,$9F,$82,$FC,$79,$23,$F1
	.dc.b	$CF,$E4,$BE,$03,$EA,$7C,$2C,$9F,$D8,$47,$E0,$4F,$B1,$F0,$D2,$32
	.dc.b	$43,$F3,$D1,$CD,$FC,$D3,$5C,$98,$21,$82,$1F,$BA,$78,$3F,$BC,$F8
	.dc.b	$6F,$D3,$91,$FD,$33,$D7,$7F,$DB,$CF,$9E,$7F,$19,$F1,$15,$4E,$9F
	.dc.b	$43,$FE,$D8,$E5,$7C,$47,$E7,$FE,$23,$F6,$A4,$AF,$DA,$3F,$9B,$D3
	.dc.b	$F3,$58,$C9,$FC,$D4,$7E,$77,$29,$1F,$9D,$7F,$41,$5F,$AB,$CB,$F8
	.dc.b	$09,$27,$F8,$03,$D5,$FD,$25,$F5,$7F,$49,$7D,$1F,$D3,$5D,$CB,$D5
	.dc.b	$87,$ED,$F1,$FE,$1A,$4A,$FE,$18,$FE,$CF,$A3,$F6,$44,$8F,$D9,$74
	.dc.b	$7E,$D7,$93,$F6,$DC,$92,$7F,$6D,$1F,$C0,$F8,$3F,$8C,$92,$3F,$8C
	.dc.b	$3D,$5F,$DE,$5F,$57,$F7,$97,$D5,$FE,$01,$7D,$5F,$E0,$17,$FC,$3F
	.dc.b	$83,$F9,$29,$23,$F9,$23,$FC,$A3,$93,$F8,$57,$54,$9F,$E8,$D1,$FC
	.dc.b	$5C,$FE,$1D,$DB,$23,$24,$3F,$9A,$47,$37,$F9,$83,$5C,$98,$21,$82
	.dc.b	$1F,$D4,$1E,$0F,$F5,$2E,$6F,$E7,$A4,$7F,$9E,$3D,$77,$FE,$9B,$3E
	.dc.b	$79,$FF,$58,$F8,$0F,$13,$11,$F9,$8E,$9F,$EB,$F5,$8B,$FC,$C4,$FF
	.dc.b	$35,$E8,$FE,$96,$47,$FA,$53,$FD,$87,$07,$9F,$FC,$F5,$79,$94,$C4
	.dc.b	$67,$CC,$97,$E4,$78,$79,$1F,$D0,$FB,$7F,$AC,$92,$7F,$AC,$3D,$EB
	.dc.b	$A4,$7B,$2F,$AB,$CC,$87,$B4,$FE,$FF,$83,$0E,$F1,$FE,$01,$EF,$FE
	.dc.b	$CE,$47,$FB,$33,$ED,$F6,$7D,$84,$8F,$97,$EC,$7B,$FD,$4F,$81,$E8
	.dc.b	$C9,$F9,$F1,$FD,$77,$C4,$FE,$E6,$48,$FE,$E4,$F5,$7E,$AD,$F5,$7D
	.dc.b	$03,$2F,$F7,$2F,$03,$8B,$FC,$2F,$0C,$3C,$15,$C2,$18,$F9,$63,$FC
	.dc.b	$3D,$E2,$7F,$81,$08,$FF,$80,$DF,$F0,$D7,$C1,$75,$8F,$C3,$FC,$07
	.dc.b	$DB,$3F,$BC,$46,$63,$CA,$DC,$B1,$83,$E0,$3F,$04,$70,$CE,$5F,$F8
	.dc.b	$A6,$F9,$3F,$E2,$F6,$BF,$E3,$26,$BF,$E3,$66,$A6,$92,$3D,$24,$8E
	.dc.b	$CD,$7F,$C7,$50,$F2,$BD,$E4,$8F,$23,$BC,$3D,$D3,$ED,$AF,$21,$F9
	.dc.b	$DF,$82,$F5,$DF,$C2,$8B,$5F,$0B,$E0,$FE,$3F,$97,$F3,$09,$5F,$05
	.dc.b	$FD,$46,$1F,$A7,$F8,$39,$3F,$B6,$8F,$D5,$7C,$2C,$8F,$DC,$BF,$AD
	.dc.b	$F8,$5F,$A7,$E8,$FD,$82,$4F,$CD,$7A,$BF,$C0,$2F,$AB,$F8,$5B,$E8
	.dc.b	$FE,$1E,$EE,$5F,$6B,$0F,$BF,$E8,$FE,$10,$95,$F6,$5F,$E0,$1F,$09
	.dc.b	$FB,$F2,$47,$F0,$EF,$85,$FE,$05,$F0,$BF,$C0,$FE,$16,$4F,$F1,$28
	.dc.b	$FC,$5F,$B1,$F9,$02,$47,$E1,$5E,$AF,$F1,$8B,$EA,$FE,$76,$FA,$BF
	.dc.b	$C7,$2F,$AB,$FA,$1B,$FC,$BF,$B1,$F9,$C2,$47,$E4,$5F,$CF,$F3,$7F
	.dc.b	$94,$FE,$75,$F0,$94,$FC,$FC,$4A,$FF,$43,$ED,$FA,$1A,$FF,$77,$37
	.dc.b	$E1,$35,$3F,$A4,$AE,$AF,$E7,$DA,$FE,$49,$AF,$E8,$5A,$FE,$85,$AF
	.dc.b	$E8,$5A,$FE,$87,$DB,$FD,$B9,$1F,$D6,$3D,$AB,$F3,$33,$E4,$7D,$1F
	.dc.b	$11,$7A,$3B,$5F,$B1,$AF,$F1,$0C,$5F,$0D,$FC,$03,$E0,$3F,$82,$92
	.dc.b	$7F,$74,$FF,$A8,$65,$F0,$7F,$2F,$E0,$E5,$75,$3F,$D5,$F3,$A7,$01
	.dc.b	$7C,$2F,$D3,$24,$6A,$1F,$0B,$FC,$4B,$E1,$7F,$8B,$92,$7F,$86,$3E
	.dc.b	$3C,$3C,$31,$EB,$C3,$EF,$47,$48,$F7,$61,$75,$FB,$98,$7F,$1E,$F8
	.dc.b	$5F,$E4,$64,$AF,$E3,$0F,$BF,$3F,$C4,$12,$3C,$D9,$7E,$2E,$7F,$19
	.dc.b	$F0,$B2,$7D,$51,$FC,$AA,$3F,$97,$92,$3F,$92,$3D,$5F,$81,$5D,$5F
	.dc.b	$A9,$5D,$5F,$89,$3D,$5F,$B1,$5F,$CD,$B2,$FE,$76,$48,$FE,$60,$FF
	.dc.b	$3E,$CF,$F4,$1C,$12,$7E,$9F,$6A,$7F,$9F,$9F,$A5,$E2,$93,$A8,$CB
	.dc.b	$FC,$ED,$E8,$D2,$1F,$CC,$1C,$B4,$C9,$FD,$05,$CB,$4F,$E8,$DB,$7F
	.dc.b	$54,$23,$83,$AE,$DF,$CD,$A7,$C7,$1F,$99,$EA,$B3,$98,$F4,$1F,$8B
	.dc.b	$E6,$FE,$B7,$C1,$FD,$80,$91,$FD,$55,$FC,$E7,$87,$FC,$14,$91,$F9
	.dc.b	$3F,$07,$F8,$47,$83,$FC,$27,$BA,$4F,$E5,$E3,$FB,$3F,$47,$F6,$C2
	.dc.b	$47,$F6,$17,$AB,$F9,$DB,$EA,$FE,$C6,$FA,$BF,$A1,$BE,$AF,$ED,$6F
	.dc.b	$FB,$9F,$3F,$F7,$82,$47,$F6,$D4,$C2,$0F,$9E,$FF,$70,$BF,$E0,$E7
	.dc.b	$FD,$CB,$A7,$F7,$D6,$F4,$F1,$77,$19,$F1,$FF,$DF,$50,$69,$FC,$AE
	.dc.b	$7F,$C0,$FD,$9F,$E0,$C2,$5F,$F8,$05,$F1,$BD,$7C,$DE,$7E,$64,$8F
	.dc.b	$E0,$B8,$78,$2F,$C5,$87,$69,$CB,$DD,$FC,$CF,$77,$F8,$68,$8F,$F8
	.dc.b	$3D,$EE,$FC,$4F,$9F,$78,$7E,$AF,$52,$4F,$F1,$49,$CB,$67,$DF,$E9
	.dc.b	$C8,$C1,$0F,$5C,$F0,$6C,$F8,$14,$65,$FA,$6D,$7E,$B3,$5F,$34,$71
	.dc.b	$8C,$47,$28,$C1,$1F,$6C,$68,$78,$BC,$49,$1A,$78,$3F,$E4,$E8,$F4
	.dc.b	$6F,$FE,$3D,$6E,$7F,$81,$F5,$7F,$15,$39,$C2,$CF,$A1,$B7,$E8,$3E
	.dc.b	$CF,$AF,$87,$E0,$F8,$DE,$B2,$3E,$87,$E3,$6E,$F8,$5F,$C6,$92,$7F
	.dc.b	$07,$F0,$3F,$90,$F8,$1F,$C8,$FB,$B2,$7F,$0F,$1D,$F6,$79,$7F,$24
	.dc.b	$93,$F1,$9E,$F8,$75,$8F,$D1,$3F,$9C,$C3,$B4,$7E,$4B,$0E,$CF,$E9
	.dc.b	$F0,$ED,$E2,$7D,$52,$47,$D2,$7D,$7F,$09,$FA,$6F,$85,$93,$F9,$C8
	.dc.b	$FD,$4F,$C2,$C8,$FC,$FB,$FA,$CF,$85,$F5,$7A,$5F,$81,$24,$FD,$E7
	.dc.b	$AB,$FA,$6B,$EA,$FE,$EE,$FA,$3F,$BD,$BB,$C9,$FB,$DC,$3E,$37,$A5
	.dc.b	$F8,$D2,$57,$E2,$9F,$C8,$7A,$5F,$BE,$F8,$29,$3F,$A5,$8F,$9C,$7F
	.dc.b	$02,$F8,$39,$19,$21,$FA,$48,$E6,$FE,$59,$AE,$4C,$10,$C1,$0F,$D3
	.dc.b	$BC,$1F,$D4,$FA,$DF,$A5,$23,$FA,$27,$AE,$3F,$9B,$9F,$3C,$FF,$0C
	.dc.b	$F8,$5A,$FE,$86,$2C,$FF,$21,$F8,$5F,$C5,$7B,$1F,$B1,$25,$7E,$B9
	.dc.b	$FD,$E7,$D0,$7F,$BA,$E3,$FC,$36,$D7,$FD,$72,$DA,$FE,$59,$1F,$CA
	.dc.b	$3E,$1B,$F9,$4F,$C3,$48,$FE,$3F,$1F,$99,$F7,$7F,$80,$12,$3F,$7A
	.dc.b	$F8,$3E,$1B,$F5,$98,$75,$F0,$7F,$9F,$36,$BF,$A0,$F7,$7F,$40,$CB
	.dc.b	$F9,$E6,$3D,$E7,$FA,$06,$5E,$37,$F9,$07,$BD,$FC,$3C,$8F,$F0,$C7
	.dc.b	$DF,$F0,$BE,$0F,$85,$93,$F0,$23,$C7,$F0,$B2,$3E,$1B,$E6,$F8,$6F
	.dc.b	$DA,$7C,$2F,$F1,$B2,$4F,$F1,$87,$AB,$F2,$AF,$AB,$F2,$AF,$A3,$F3
	.dc.b	$2E,$E7,$FE,$D3,$D5,$8F,$F0,$1F,$86,$FE,$50,$4B,$FE,$4A,$FF,$2B
	.dc.b	$F8,$6F,$92,$48,$FE,$80,$FF,$09,$3E,$6F,$C3,$48,$C9,$0F,$E7,$F1
	.dc.b	$CD,$FE,$62,$D7,$26,$08,$60,$87,$F4,$27,$83,$FD,$12,$3F,$A0,$11
	.dc.b	$FE,$7A,$F5,$C3,$F9,$B4,$F9,$E7,$F2,$3A,$57,$F9,$E4,$59,$F8,$D7
	.dc.b	$FC,$97,$3F,$E9,$84,$AF,$E9,$2F,$E7,$3A,$3F,$1A,$48,$FC,$EF,$2F
	.dc.b	$E4,$39,$7F,$23,$C1,$27,$F4,$31,$FC,$C7,$83,$FA,$C1,$23,$FA,$AB
	.dc.b	$D5,$FD,$3D,$F5,$7F,$4F,$7D,$5F,$D5,$DF,$57,$F5,$77,$FC,$F7,$83
	.dc.b	$FB,$21,$23,$FB,$0B,$FB,$4F,$07,$F4,$BE,$0C,$9F,$DA,$C7,$EA,$3B
	.dc.b	$64,$7E,$E1,$FD,$5F,$57,$F4,$9E,$3F,$EE,$04,$9F,$ED,$AF,$57,$F7
	.dc.b	$F7,$D5,$FD,$FD,$F4,$7F,$80,$DD,$E0,$FC,$B6,$1F,$D5,$7A,$7F,$BE
	.dc.b	$12,$BF,$BC,$BF,$E0,$7F,$A1,$FD,$C7,$57,$EF,$3C,$9F,$DF,$9F,$DA
	.dc.b	$E3,$35,$FE,$3E,$6A,$E8,$FE,$C2,$FF,$14,$F2,$38,$30,$FF,$11,$CF
	.dc.b	$F8,$55,$34,$8F,$8C,$87,$02,$9F,$E3,$A7,$6E,$4E,$BD,$7E,$B2,$38
	.dc.b	$34,$33,$63,$FE,$57,$AF,$59,$EC,$C3,$FC,$27,$5E,$F3,$D4,$71,$9E
	.dc.b	$0F,$F2,$4F,$37,$E3,$76,$7C,$B2,$47,$C0,$7F,$C6,$9A,$BD,$AF,$F0
	.dc.b	$B3,$67,$FD,$AA,$D7,$B1,$FE,$6C,$7F,$DC,$AE,$BD,$5E,$3E,$94,$8C
	.dc.b	$1F,$ED,$EC,$DE,$17,$DD,$24,$7D,$27,$C7,$87,$88,$91,$E2,$F5,$BC
	.dc.b	$B8,$75,$8F,$18,$C5,$8E,$A3,$C1,$E1,$FE,$2C,$8F,$E0,$1F,$83,$F0
	.dc.b	$5F,$03,$DF,$93,$F0,$23,$E1,$FC,$04,$8F,$86,$FC,$6F,$7B,$F2,$5E
	.dc.b	$4F,$E5,$C9,$3F,$8C,$7A,$BF,$2A,$FA,$BE,$8C,$98,$75,$8C,$61,$79
	.dc.b	$3F,$07,$1F,$3F,$1F,$D0,$79,$3F,$A5,$24,$7E,$69,$FD,$47,$99,$F6
	.dc.b	$BE,$0E,$4F,$D1,$8F,$D4,$1F,$7B,$E1,$24,$64,$87,$E6,$E3,$9B,$FA
	.dc.b	$66,$B9,$30,$43,$04,$3F,$D5,$FC,$88,$E3,$18,$F3,$FF,$7C,$47,$F6
	.dc.b	$6F,$6C,$BE,$0C,$7E,$D2,$7D,$2F,$E5,$3E,$1A,$BF,$82,$BE,$CF,$AA
	.dc.b	$B8,$EC,$F8,$6F,$E1,$1E,$8F,$F0,$D2,$57,$F0,$27,$F4,$7F,$09,$FA
	.dc.b	$22,$47,$E8,$BE,$13,$F4,$BF,$0D,$FA,$6F,$86,$93,$FA,$68,$FE,$29
	.dc.b	$EA,$FF,$19,$24,$7F,$0E,$7A,$BF,$AC,$BE,$AF,$E0,$EF,$AB,$FB,$0B
	.dc.b	$EA,$FE,$1E,$FF,$90,$7B,$3F,$C9,$49,$1F,$C6,$9F,$DD,$FC,$37,$EE
	.dc.b	$BE,$1A,$4F,$EE,$A3,$F7,$BF,$0B,$23,$F7,$AF,$F0,$0F,$86,$FE,$59
	.dc.b	$EE,$FF,$31,$24,$FF,$26,$7A,$BF,$C1,$2F,$AB,$F9,$BB,$E8,$FE,$76
	.dc.b	$EF,$0F,$D2,$C3,$F9,$C7,$B9,$FC,$F4,$95,$FC,$C9,$FE,$81,$EE,$FF
	.dc.b	$A9,$FE,$8B,$FD,$D9,$3F,$EF,$AB,$FC,$1C,$62,$3F,$01,$3F,$D0,$CF
	.dc.b	$E3,$3E,$4C,$8F,$E2,$11,$97,$F9,$FB,$D9,$A4,$3F,$9D,$39,$69,$93
	.dc.b	$F3,$6E,$5A,$7F,$4A,$F8,$2F,$EB,$04,$70,$75,$BF,$CD,$F3,$AB,$C9
	.dc.b	$3F,$C6,$B0,$FF,$7F,$3F,$D6,$E3,$35,$C1,$F8,$9F,$0D,$FD,$87,$E0
	.dc.b	$BF,$B3,$12,$BF,$AD,$BF,$EE,$FB,$63,$F2,$C9,$3F,$2F,$23,$25,$5F
	.dc.b	$47,$E1,$7E,$97,$C2,$CA,$E0,$6E,$3F,$B9,$1F,$DD,$C8,$FF,$6C,$7C
	.dc.b	$58,$78,$23,$DC,$FD,$DA,$ED,$1D,$05,$AF,$C1,$1F,$16,$FF,$C0,$11
	.dc.b	$FE,$05,$24,$7F,$7A,$7E,$07,$1F,$E2,$78,$E4,$F5,$BF,$F7,$63,$77
	.dc.b	$FC,$71,$1D,$6B,$FE,$0D,$F8,$7F,$F0,$91,$23,$FC,$0E,$F6,$7A,$5F
	.dc.b	$67,$ED,$61,$CD,$FB,$97,$77,$FC,$CE,$1F,$E1,$9D,$FF,$C3,$C4,$AF
	.dc.b	$F0,$9B,$FE,$21,$F8,$7F,$D1,$75,$C9,$FE,$D5,$1F,$E2,$13,$F4,$DD
	.dc.b	$72,$32,$43,$FC,$29,$1C,$DF,$F0,$FB,$5C,$98,$21,$82,$1F,$CE,$9E
	.dc.b	$0F,$E2,$0F,$C4,$FD,$27,$F3,$67,$01,$C3,$B6,$BC,$38,$FC,$59,$7E
	.dc.b	$92,$BD,$93,$F5,$FC,$0A,$FF,$3D,$8B,$3F,$C0,$BC,$1F,$13,$FD,$1B
	.dc.b	$07,$4F,$8A,$FC,$0D,$7C,$ED,$7C,$CF,$E9,$FC,$FF,$E2,$04,$8F,$C4
	.dc.b	$79,$7F,$C4,$FC,$BF,$E2,$9E,$3C,$9F,$C6,$47,$8B,$5F,$77,$77,$5C
	.dc.b	$8D,$25,$78,$F9,$FF,$C4,$6F,$E5,$72,$F0,$3F,$C0,$F2,$F0,$3F,$99
	.dc.b	$CB,$C0,$FF,$07,$CB,$C0,$FD,$2E,$EF,$01,$1F,$81,$92,$3E,$07,$D2
	.dc.b	$7F,$B9,$67,$F5,$E7,$F2,$F3,$98,$75,$93,$D4,$FE,$68,$49,$FD,$67
	.dc.b	$A5,$C3,$07,$83,$F0,$49,$3E,$27,$C3,$EB,$F7,$8F,$DA,$1E,$0C,$7B
	.dc.b	$E1,$D6,$3F,$92,$3F,$D0,$F0,$ED,$E3,$7C,$D2,$4F,$B9,$FA,$1E,$37
	.dc.b	$3F,$72,$4F,$F0,$DC,$D3,$E7,$3C,$1E,$E4,$9D,$46,$5F,$92,$F4,$69
	.dc.b	$0F,$C7,$4F,$47,$07,$E8,$9C,$3A,$8E,$4D,$36,$7E,$97,$C7,$FC,$41
	.dc.b	$1F,$B2,$F0,$E7,$FC,$F4,$F9,$67,$FC,$C3,$A3,$E0,$B1,$5A,$9F,$59
	.dc.b	$FC,$B6,$5F,$05,$F5,$FC,$AF,$CB,$12,$7F,$0E,$D7,$DA,$D7,$D8,$D7
	.dc.b	$F7,$8D,$7F,$81,$BF,$A1,$F3,$64,$7E,$49,$AF,$C8,$6B,$F1,$DA,$FF
	.dc.b	$07,$6B,$FC,$39,$AF,$F0,$DF,$33,$F5,$C4,$7F,$3E,$D7,$EA,$B5,$FA
	.dc.b	$8D,$7F,$8A,$35,$FE,$36,$FE,$E7,$CF,$91,$FA,$A6,$BF,$80,$F3,$FF
	.dc.b	$31,$F0,$92,$3F,$1D,$1F,$A3,$3F,$39,$F0,$D2,$32,$43,$91,$FB,$68
	.dc.b	$FE,$10,$47,$04,$36,$43,$F1,$CE,$DC,$1F,$CB,$FD,$4F,$E2,$44,$7F
	.dc.b	$81,$BB,$D7,$F9,$44,$F9,$E7,$F0,$DF,$0D,$5F,$F5,$6F,$12,$71,$5F
	.dc.b	$E7,$4D,$DF,$0D,$FC,$07,$D7,$FE,$40,$4C,$3F,$88,$BF,$C8,$FD,$7F
	.dc.b	$DF,$7C,$1C,$9F,$DA,$47,$F0,$93,$F8,$17,$C2,$48,$C9,$0F,$D7,$C7
	.dc.b	$37,$F8,$EB,$5C,$98,$21,$82,$1F,$B5,$78,$3F,$B7,$F6,$7F,$9C,$11
	.dc.b	$FE,$52,$F5,$CB,$FA,$74,$F9,$E7,$F5,$9F,$0D,$5F,$8F,$16,$7E,$37
	.dc.b	$C3,$7F,$25,$F7,$FF,$A2,$12,$BF,$9A,$BF,$D1,$FD,$FF,$E4,$5F,$09
	.dc.b	$27,$F8,$C4,$7F,$2D,$3F,$93,$7B,$D2,$32,$43,$F8,$9C,$73,$7F,$A0
	.dc.b	$B5,$C9,$82,$18,$21,$FC,$65,$E0,$FF,$1B,$F8,$0F,$EB,$84,$7F,$A6
	.dc.b	$BD,$71,$FE,$E5,$3E,$79,$FE,$21,$F0,$B5,$FC,$5C,$59,$FC,$57,$C3
	.dc.b	$7F,$49,$F8,$2F,$ED,$04,$AF,$EB,$2F,$F7,$7F,$25,$F1,$3F,$D1,$FE
	.dc.b	$1E,$57,$F9,$26,$72,$D9,$FE,$99,$F0,$F2,$30,$43,$FC,$0D,$3C,$1B
	.dc.b	$3D,$C8,$FF,$82,$63,$4F,$F8,$79,$AF,$F8,$8D,$AF,$F9,$B1,$AF,$F6
	.dc.b	$48,$E6,$F6,$3A,$7C,$37,$B4,$8E,$0F,$04,$7F,$74,$9F,$36,$5F,$EE
	.dc.b	$46,$FA,$BF,$E3,$3E,$AB,$7E,$1B,$E0,$47,$FA,$16,$1C,$F2,$F9,$84
	.dc.b	$8F,$73,$F3,$F2,$FF,$01,$FD,$2B,$82,$9E,$38,$95,$FE,$79,$96,$BC
	.dc.b	$0C,$C3,$5F,$45,$75,$8F,$98,$D7,$E9,$B5,$F9,$CD,$7E,$73,$5F,$9C
	.dc.b	$D7,$FC,$23,$C1,$F8,$72,$3F,$65,$ED,$3E,$28,$FC,$44,$FB,$B8,$FC
	.dc.b	$F3,$99,$F4,$6D,$E0,$7D,$DC,$3F,$4F,$87,$F2,$84,$8F,$C3,$3F,$97
	.dc.b	$E3,$F9,$3C,$B2,$7F,$2D,$1F,$6C,$F9,$BC,$F2,$32,$43,$F2,$B1,$CD
	.dc.b	$FC,$93,$5C,$98,$21,$82,$1F,$97,$78,$3F,$99,$E5,$FD,$61,$1F,$CF
	.dc.b	$3D,$72,$FC,$74,$F9,$E7,$EF,$77,$D7,$E2,$9F,$9F,$8C,$4B,$3F,$F4
	.dc.b	$37,$03,$AF,$F3,$5D,$3F,$BB,$26,$1F,$AD,$7F,$7D,$D3,$F9,$0F,$0A
	.dc.b	$4F,$EF,$63,$F4,$47,$E5,$3B,$A4,$64,$87,$EF,$23,$9B,$FB,$96,$B9
	.dc.b	$30,$43,$04,$3F,$7C,$F0,$7F,$80,$76,$FF,$13,$23,$FC,$15,$EB,$9F
	.dc.b	$ED,$67,$CF,$3F,$A5,$F1,$2B,$F2,$0F,$F3,$3C,$62,$7E,$44,$7F,$BD
	.dc.b	$33,$78,$9F,$C0,$7B,$FF,$91,$12,$7F,$8A,$3F,$C9,$7C,$2F,$DB,$F9
	.dc.b	$32,$7F,$92,$47,$F0,$93,$F7,$7E,$34,$8C,$90,$FE,$47,$1C,$DF,$E4
	.dc.b	$0D,$72,$60,$86,$08,$7F,$25,$78,$3F,$C9,$FC,$3F,$E7,$44,$7F,$96
	.dc.b	$3D,$72,$FE,$39,$3E,$79,$FE,$19,$E6,$57,$E4,$78,$89,$C5,$7C,$89
	.dc.b	$69,$FE,$70,$8F,$E4,$DE,$3F,$F4,$92,$4F,$F3,$B7,$FA,$67,$8F,$FC
	.dc.b	$7B,$D1,$93,$FD,$2E,$3F,$98,$1F,$C8,$B1,$91,$92,$1F,$D2,$A3,$9B
	.dc.b	$FD,$1D,$AE,$4C,$10,$C1,$0F,$E9,$8F,$07,$FA,$77,$95,$FD,$84,$8F
	.dc.b	$F5,$37,$AE,$1F,$D0,$E7,$CF,$3F,$CC,$FD,$7A,$FC,$88,$B3,$F1,$F2
	.dc.b	$FF,$6A,$CF,$F6,$9F,$2F,$FB,$71,$1F,$EC,$0F,$F7,$1F,$3B,$FD,$8B
	.dc.b	$ED,$3D,$AA,$7F,$55,$89,$5F,$D0,$FB,$7F,$A4,$98,$7E,$CC,$73,$7F
	.dc.b	$B6,$35,$FE,$E6,$D7,$FB,$73,$5F,$ED,$CD,$7F,$B7,$35,$FF,$68,$FA
	.dc.b	$1F,$E0,$72,$3F,$DE,$5E,$D9,$7F,$69,$9F,$23,$F8,$9C,$1F,$05,$8C
	.dc.b	$3B,$E1,$FE,$0D,$BF,$83,$EF,$7F,$73,$F5,$7F,$C2,$84,$9F,$F0,$32
	.dc.b	$61,$07,$D3,$5C,$BC,$1F,$0B,$F3,$0F,$F7,$8C,$67,$1E,$2D,$AE,$99
	.dc.b	$BE,$17,$E9,$9C,$22,$1D,$CF,$14,$D7,$99,$4F,$19,$5C,$38,$38,$38
	.dc.b	$23,$B2,$1A,$39,$7B,$11,$E2,$D7,$C9,$1E,$B7,$B1,$FC,$8D,$C5,$3E
	.dc.b	$1F,$D7,$39,$2E,$5F,$5F,$E1,$FE,$3F,$B5,$E9,$24,$FA,$1F,$67,$BF
	.dc.b	$F9,$8F,$85,$93,$F5,$6F,$F8,$46,$4A,$FA,$27,$E7,$BE,$16,$56,$F6
	.dc.b	$FA,$73,$A8,$F8,$AF,$66,$90,$F7,$B9,$69,$93,$F0,$95,$E1,$70,$76
	.dc.b	$3E,$8F,$C0,$FD,$52,$38,$3B,$61,$EA,$AF,$24,$7F,$4D,$D5,$F0,$D8
	.dc.b	$AD,$4F,$AC,$FC,$2D,$7E,$F3,$E1,$FF,$23,$F0,$5F,$83,$24,$FE,$09
	.dc.b	$FC,$37,$C1,$FF,$02,$F8,$79,$3C,$CF,$CE,$C7,$E6,$8F,$E0,$FF,$0F
	.dc.b	$23,$5A,$A7,$51,$F8,$E7,$AB,$48,$7E,$21,$CB,$4C,$9F,$29,$CB,$4F
	.dc.b	$CC,$E9,$F9,$D2,$38,$3A,$E7,$F8,$29,$F1,$C7,$C9,$9B,$39,$8F,$4C
	.dc.b	$FF,$22,$F8,$7F,$DB,$9F,$A9,$24,$7E,$A1,$FD,$66,$3F,$CA,$32,$93
	.dc.b	$F0,$A3,$F7,$E7,$F2,$CA,$91,$92,$1F,$A8,$8E,$6F,$ED,$1A,$E4,$C1
	.dc.b	$0C,$10,$FC,$7B,$C1,$FC,$8E,$5F,$C0,$48,$FF,$00,$7A,$E1,$FA,$69
	.dc.b	$F3,$CF,$DF,$C7,$F3,$0F,$EE,$63,$2D,$35,$B7,$F1,$CD,$BF,$87,$12
	.dc.b	$3F,$86,$BF,$C4,$75,$FE,$9B,$CB,$27,$F0,$F1,$FC,$90,$FE,$A3,$C7
	.dc.b	$23,$24,$3F,$86,$C7,$37,$F8,$CB,$5C,$98,$21,$82,$1F,$B4,$78,$3F
	.dc.b	$B6,$E3,$FE,$4C,$47,$F9,$2B,$D7,$3F,$E1,$73,$E7,$9F,$D2,$7D,$3B
	.dc.b	$A7,$11,$FC,$7E,$6D,$7F,$51,$8F,$E8,$3C,$BF,$CC,$C8,$FF,$32,$7F
	.dc.b	$9B,$73,$FF,$4C,$9F,$F1,$06,$0E,$BC,$73,$FF,$50,$8C,$61,$FB,$0B
	.dc.b	$FE,$9A,$69,$FE,$AD,$0E,$DC,$42,$30,$F1,$21,$FC,$EA,$7C,$6F,$F4
	.dc.b	$66,$B9,$30,$43,$04,$3F,$90,$BC,$1F,$E4,$7D,$3F,$D4,$C8,$FF,$52
	.dc.b	$7A,$E3,$FC,$E2,$7D,$73,$F3,$BC,$3A,$FF,$4B,$8B,$3F,$1B,$C0,$FE
	.dc.b	$DB,$DB,$FD,$7C,$95,$FD,$79,$FE,$C5,$DB,$F0,$B2,$78,$D4,$FE,$EB
	.dc.b	$12,$BE,$EF,$87,$B3,$0F,$F5,$B8,$E6,$FF,$6A,$6B,$FD,$89,$AF,$F4
	.dc.b	$26,$BF,$E0,$06,$BF,$DE,$9A,$FF,$82,$BC,$1F,$EF,$04,$7F,$BB,$BD
	.dc.b	$B2,$FE,$BB,$3E,$47,$F0,$FE,$3D,$E0,$ED,$7F,$CD,$2B,$E0,$79,$3F
	.dc.b	$E1,$2F,$0F,$FC,$08,$49,$FF,$01,$BF,$E0,$7F,$13,$E3,$7A,$12,$7F
	.dc.b	$9E,$C7,$F8,$68,$F9,$5E,$6C,$8C,$90,$FF,$01,$C7,$37,$FC,$1E,$D7
	.dc.b	$26,$08,$60,$87,$F6,$A7,$83,$FE,$2F,$FA,$8F,$F4,$23,$72,$FF,$10
	.dc.b	$46,$23,$FC,$6A,$D7,$FC,$4F,$3E,$97,$CD,$A7,$EB,$BD,$4C,$47,$F8
	.dc.b	$5E,$6C,$FE,$CB,$D1,$FD,$4D,$BC,$BE,$E4,$BF,$F1,$94,$65,$B1,$97
	.dc.b	$FC,$71,$19,$6C,$65,$F6,$35,$F4,$BE,$CF,$89,$91,$D9,$AF,$46,$BF
	.dc.b	$15,$AF,$C2,$6B,$FE,$59,$9F,$F4,$2B,$C4,$F4,$F9,$DF,$28,$8E,$08
	.dc.b	$7A,$9C,$B4,$C9,$F5,$5C,$B4,$F8,$BE,$77,$D6,$23,$82,$1F,$19,$CB
	.dc.b	$4C,$9F,$81,$72,$D3,$E8,$F9,$FF,$82,$23,$82,$1F,$49,$CB,$4C,$9F
	.dc.b	$8D,$72,$D3,$EE,$FC,$47,$E3,$88,$E0,$87,$DE,$72,$D3,$27,$E6,$5C
	.dc.b	$B4,$FC,$4F,$A5,$F9,$A2,$38,$3F,$1B,$E9,$7F,$19,$F6,$24,$7F,$8D
	.dc.b	$72,$4F,$E5,$8F,$E3,$DE,$D4,$9D,$46,$5E,$64,$3F,$2A,$FE,$A8,$8E
	.dc.b	$4F,$D8,$38,$76,$64,$FD,$7B,$FB,$2F,$5A,$47,$E9,$1A,$F0,$1C,$1F
	.dc.b	$EF,$DB,$F4,$C7,$98,$FE,$DB,$F0,$3E,$77,$7C,$9D,$9D,$18,$3F,$AF
	.dc.b	$FB,$DF,$B6,$F6,$3F,$81,$92,$3F,$6A,$FF,$06,$F6,$BF,$A1,$7C,$24
	.dc.b	$9F,$95,$1F,$BE,$3F,$A3,$7C,$14,$8C,$90,$FD,$DC,$73,$7F,$80,$B5
	.dc.b	$C9,$82,$18,$21,$FC,$6D,$E0,$FF,$1D,$F7,$3F,$8D,$91,$FE,$1A,$F5
	.dc.b	$D3,$E4,$4F,$9E,$7F,$C2,$7F,$0D,$5F,$88,$35,$4E,$6B,$FC,$1B,$F0
	.dc.b	$DF,$C7,$3E,$03,$F9,$49,$2F,$F8,$DB,$FC,$B3,$E0,$3F,$C1,$3F,$52
	.dc.b	$F8,$7A,$7F,$24,$89,$5F,$E8,$79,$7C,$3F,$03,$31,$E4,$CD,$5D,$63
	.dc.b	$F9,$4B,$5F,$E4,$ED,$7F,$A0,$35,$FE,$7A,$D7,$F8,$E3,$5F,$81,$F0
	.dc.b	$BF,$D1,$08,$FF,$36,$7B,$67,$F9,$19,$F2,$AE,$3F,$F7,$EE,$2F,$87
	.dc.b	$CC,$78,$30,$FD,$85,$FC,$4F,$87,$FE,$8B,$F0,$FF,$D4,$89,$3F,$D1
	.dc.b	$1F,$EA,$B7,$F2,$B2,$93,$FB,$98,$FE,$9A,$7C,$E9,$91,$92,$1F,$D2
	.dc.b	$A3,$9B,$FD,$41,$AE,$4C,$10,$C1,$0F,$ED,$4F,$07,$FB,$65,$7F,$6A
	.dc.b	$23,$FD,$71,$EB,$97,$ED,$E7,$CF,$3F,$7F,$6A,$FD,$D8,$B3,$F2,$E0
	.dc.b	$33,$FC,$0B,$FC,$67,$6F,$EF,$24,$7F,$B5,$BF,$DF,$36,$FC,$8F,$04
	.dc.b	$9F,$E4,$11,$FD,$D0,$FC,$AF,$14,$8C,$90,$FE,$E3,$1C,$DF,$EE,$ED
	.dc.b	$72,$60,$86,$08,$7F,$84,$5E,$0F,$B0,$F6,$FD,$53,$8F,$11,$23,$0E
	.dc.b	$0D,$C5,$69,$E0,$CB,$F9,$25,$7B,$63,$F5,$9D,$35,$FD,$54,$59,$F9
	.dc.b	$D1,$F3,$79,$FF,$CF,$F8,$39,$7B,$4A,$30,$F8,$5A,$F9,$5A,$FA,$79
	.dc.b	$FF,$80,$F6,$C8,$FB,$78,$78,$29,$5E,$03,$F8,$3F,$5C,$AD,$F3,$FC
	.dc.b	$7A,$DC,$60,$F9,$35,$E6,$8C,$C7,$8E,$30,$47,$D0,$D7,$46,$08,$60
	.dc.b	$87,$D3,$E9,$EC,$47,$E1,$3C,$F0,$F6,$4F,$AE,$7F,$8E,$77,$D7,$F9
	.dc.b	$6D,$F9,$AB,$71,$FC,$87,$BF,$E8,$F6,$FA,$C9,$1F,$3D,$F7,$76,$FF
	.dc.b	$2C,$F0,$64,$FC,$33,$EB,$47,$DA,$3F,$98,$F8,$32,$35,$8A,$75,$1F
	.dc.b	$6D,$EA,$D2,$1F,$75,$CB,$4C,$9F,$97,$72,$D3,$F0,$DE,$0F,$D5,$23
	.dc.b	$83,$AE,$5F,$83,$9F,$1C,$7F,$52,$E9,$FF,$30,$4E,$63,$D3,$3A,$8F
	.dc.b	$CC,$CF,$F9,$51,$FC,$67,$83,$F8,$32,$57,$E5,$DF,$C3,$78,$DF,$ED
	.dc.b	$CF,$AB,$78,$F4,$FD,$0C,$4A,$FE,$A9,$C7,$FD,$68,$C3,$F6,$A3,$9B
	.dc.b	$FA,$36,$BF,$A3,$6B,$FB,$B6,$BF,$9B,$6B,$FC,$09,$AF,$EB,$FC,$7F
	.dc.b	$CE,$11,$FD,$83,$DB,$1F,$D5,$4F,$91,$FD,$DE,$0F,$3B,$18,$77,$C3
	.dc.b	$F0,$37,$F0,$7C,$BF,$DF,$F9,$7F,$A9,$24,$FE,$F5,$FD,$67,$97,$F1
	.dc.b	$FD,$29,$3F,$8A,$8F,$E0,$E7,$CB,$F3,$A4,$64,$87,$F0,$58,$E6,$FF
	.dc.b	$09,$6B,$93,$04,$30,$43,$F9,$23,$C1,$FE,$4D,$E7,$FF,$01,$23,$FC
	.dc.b	$55,$EB,$AF,$F0,$F9,$F3,$CF,$E9,$F5,$7A,$98,$8F,$DA,$CD,$AF,$97
	.dc.b	$EB,$7F,$24,$F4,$7F,$87,$12,$BF,$91,$3F,$C4,$7D,$3F,$C6,$7B,$92
	.dc.b	$7F,$5F,$1F,$CB,$0F,$C8,$7B,$52,$32,$43,$F9,$54,$73,$7F,$97,$35
	.dc.b	$C9,$82,$18,$21,$FD,$25,$E0,$FF,$4B,$F5,$7F,$93,$11,$FE,$74,$F5
	.dc.b	$C7,$F9,$A4,$F9,$E7,$E3,$7B,$F5,$FE,$33,$16,$7F,$1F,$EE,$7F,$48
	.dc.b	$F6,$BF,$99,$12,$BF,$A3,$3F,$CD,$7D,$AF,$D8,$7C,$0C,$9F,$E2,$71
	.dc.b	$FD,$40,$FD,$A7,$C1,$C8,$C9,$0F,$E9,$D1,$CD,$FE,$A4,$D7,$26,$08
	.dc.b	$60,$87,$F6,$D7,$83,$FD,$BF,$DE,$FE,$94,$47,$FA,$F3,$D7,$1F,$EB
	.dc.b	$13,$E7,$9F,$C5,$7C,$1D,$7F,$9E,$C5,$9F,$D9,$FC,$37,$F6,$CF,$80
	.dc.b	$FE,$AC,$4A,$FE,$D4,$FF,$59,$F8,$1F,$E2,$9F,$0D,$27,$F9,$BC,$7F
	.dc.b	$74,$3F,$8C,$7C,$34,$8C,$90,$FE,$E5,$1C,$DF,$EE,$CD,$72,$60,$86
	.dc.b	$08,$7F,$84,$5E,$0F,$F8,$52,$7F,$C0,$69,$A4,$7B,$61,$FD,$F6,$7D
	.dc.b	$11,$FA,$FF,$88,$AF,$F6,$28,$B3,$FC,$63,$E1,$FF,$C4,$8F,$F8,$52
	.dc.b	$DF,$0F,$FE,$12,$6D,$FF,$88,$5A,$FF,$89,$5A,$F9,$FE,$1F,$F9,$F4
	.dc.b	$C8,$FE,$CD,$29,$EF,$E2,$7F,$8C,$E3,$13,$FC,$16,$7F,$A3,$1A,$7F
	.dc.b	$CA,$71,$A7,$FD,$4E,$D7,$C0,$79,$63,$D0,$47,$04,$3D,$0E,$5A,$64
	.dc.b	$E6,$7B,$72,$F8,$44,$7E,$13,$C7,$0F,$91,$3E,$79,$FA,$CD,$F4,$19
	.dc.b	$E7,$0F,$55,$F0,$BF,$8B,$7F,$D9,$5D,$67,$F4,$48,$FD,$17,$EA,$57
	.dc.b	$5E,$19,$3F,$3A,$3B,$1E,$1D,$A4,$64,$87,$D9,$8E,$6F,$DD,$6B,$93
	.dc.b	$04,$30,$43,$F2,$CF,$07,$F3,$1B,$7E,$24,$8F,$E2,$5E,$B8,$FE,$3E
	.dc.b	$7C,$F3,$FE,$2B,$CD,$C9,$88,$FB,$F3,$6B,$FC,$D3,$4E,$1F,$99,$C5
	.dc.b	$F9,$72,$5F,$E5,$DF,$CD,$71,$7D,$3E,$69,$3F,$94,$8F,$A8,$7D,$7E
	.dc.b	$89,$19,$21,$F9,$F8,$E6,$FE,$8D,$AE,$4C,$10,$C1,$0F,$DE,$3C,$1F
	.dc.b	$DF,$73,$7E,$BC,$8F,$EB,$DE,$B9,$7E,$D6,$7C,$F3,$F1,$FC,$0A,$FE
	.dc.b	$92,$2C,$FC,$6E,$CF,$C8,$74,$7E,$EC,$95,$FB,$B7,$F7,$DD,$5F,$97
	.dc.b	$F0,$E4,$FE,$E2,$3F,$30,$7E,$6F,$BE,$46,$48,$7F,$03,$8E,$6F,$F0
	.dc.b	$76,$B9,$30,$43,$04,$3F,$90,$3C,$1F,$E4,$5D,$9F,$C4,$C8,$FF,$13
	.dc.b	$7A,$E3,$FC,$66,$7C,$F3,$F8,$BF,$16,$BF,$C2,$22,$CF,$E2,$BC,$7F
	.dc.b	$DA,$78,$1F,$C7,$C9,$5F,$C7,$DF,$E4,$5E,$07,$EE,$FC,$C9,$3F,$C7
	.dc.b	$23,$F7,$87,$EF,$FC,$A9,$19,$21,$FC,$9E,$39,$BF,$CA,$DA,$E4,$C1
	.dc.b	$0C,$10,$FE,$88,$F0,$7F,$A3,$78,$9F,$CD,$C8,$FF,$37,$7A,$E3,$FC
	.dc.b	$F6,$7C,$F3,$FA,$FF,$42,$BF,$CB,$22,$CF,$EB,$BC,$DF,$E3,$1E,$37
	.dc.b	$F4,$32,$57,$F4,$37,$FA,$37,$93,$FC,$7F,$D2,$93,$FD,$02,$3F,$90
	.dc.b	$1F,$C8,$FD,$39,$19,$21,$FD,$2E,$39,$BF,$D3,$DA,$E4,$C1,$0C,$10
	.dc.b	$FE,$D0,$F0,$7F,$B5,$79,$7F,$D6,$C8,$FF,$5B,$7A,$E3,$FD,$86,$7C
	.dc.b	$F3,$FC,$4F,$DA,$AF,$F5,$08,$B3,$FC,$4B,$D7,$FE,$79,$E7,$FF,$67
	.dc.b	$25,$7F,$67,$6B,$FD,$9D,$AF,$F7,$D6,$BF,$D8,$DA,$FF,$60,$7F,$B9
	.dc.b	$79,$F2,$3F,$B9,$35,$FE,$E4,$D7,$FC,$10,$D7,$FB,$73,$5F,$ED,$6D
	.dc.b	$7F,$C1,$9E,$87,$F7,$F2,$3F,$DF,$D3,$08,$3E,$B1,$EF,$FF,$61,$5F
	.dc.b	$03,$4C,$46,$7E,$02,$57,$FB,$A6,$91,$C7,$DD,$FE,$D5,$F5,$AF,$53
	.dc.b	$BD,$F8,$63,$0E,$03,$87,$6D,$7C,$F7,$E6,$7C,$18,$F9,$5F,$16,$3E
	.dc.b	$45,$B7,$3A,$F2,$C7,$9F,$2F,$33,$FF,$00,$70,$7C,$47,$8E,$3C,$6D
	.dc.b	$8C,$C7,$9D,$AF,$9A,$34,$47,$4D,$74,$60,$87,$BB,$E2,$3B,$11,$EC
	.dc.b	$D7,$46,$08,$60,$87,$DC,$94,$71,$8C,$7B,$9E,$B2,$3E,$B6,$BF,$41
	.dc.b	$AF,$03,$44,$34,$53,$C1,$EE,$7C,$72,$3F,$1D,$AF,$DC,$6B,$C0,$D1
	.dc.b	$0D,$14,$F6,$7B,$BF,$4C,$8F,$D3,$7E,$B7,$A1,$F9,$9F,$86,$93,$F7
	.dc.b	$63,$F2,$67,$E7,$7E,$1A,$46,$48,$7E,$42,$39,$BF,$7D,$AE,$4C,$10
	.dc.b	$C1,$0F,$A4,$F0,$7E,$A7,$C0,$FE,$2C,$8F,$E2,$DE,$BB,$FE,$87,$45
	.dc.b	$7A,$2B,$F2,$DF,$09,$5F,$CE,$F1,$FE,$AE,$71,$7F,$9F,$F8,$7F,$D8
	.dc.b	$FC,$1F,$E6,$49,$7F,$99,$7B,$DF,$18,$F8,$35,$E1,$C3,$F7,$3F,$0B
	.dc.b	$FC,$0F,$E1,$E5,$7E,$9E,$32,$FE,$8D,$FD,$31,$1F,$89,$5E,$BC,$BC
	.dc.b	$B8,$7F,$01,$C3,$8D,$7E,$FA,$F8,$34,$FD,$77,$C3,$7E,$C4,$8E,$0F
	.dc.b	$D9,$4F,$47,$F8,$EE,$9F,$21,$F8,$E7,$9C,$87,$C8,$37,$5F,$C7,$E7
	.dc.b	$F8,$BF,$C3,$FE,$FC,$93,$FB,$F7,$F8,$14,$77,$8F,$E2,$F8,$FF,$1E
	.dc.b	$89,$1F,$C8,$E7,$F9,$51,$87,$C0,$7F,$1E,$8E,$6F,$A8,$87,$F2,$07
	.dc.b	$2D,$32,$7E,$F5,$CB,$4C,$9F,$CB,$73,$FE,$30,$47,$F8,$C3,$D3,$3F
	.dc.b	$BD,$1F,$18,$F2,$E3,$FD,$01,$C4,$69,$CD,$7F,$45,$CB,$F9,$EE,$BF
	.dc.b	$C9,$49,$3F,$C9,$5E,$DA,$7D,$F8,$E6,$FF,$41,$E9,$FE,$8B,$C3,$FD
	.dc.b	$3F,$69,$5F,$CC,$E3,$2F,$F2,$F7,$F9,$91,$1F,$E6,$53,$F3,$72,$F3
	.dc.b	$1F,$A9,$78,$C7,$F4,$A9,$E0,$D3,$F9,$D7,$0F,$F3,$D2,$38,$3A,$9E
	.dc.b	$83,$F1,$B9,$FF,$5C,$AE,$B7,$FB,$7B,$B8,$FF,$6D,$E8,$70,$FF,$62
	.dc.b	$E9,$FE,$94,$4C,$BF,$A5,$3F,$D3,$71,$F5,$5F,$C7,$9F,$A7,$1D,$DF
	.dc.b	$ED,$DC,$7F,$DD,$4C,$3A,$3C,$E7,$F6,$88,$E8,$E0,$FE,$88,$E1,$D1
	.dc.b	$93,$EA,$14,$FD,$5B,$FD,$87,$A6,$47,$F6,$16,$BC,$8F,$11,$33,$FE
	.dc.b	$F5,$F5,$CF,$37,$CF,$1F,$56,$7F,$B8,$5F,$F8,$3F,$9B,$FC,$0F,$D9
	.dc.b	$FD,$BC,$93,$FD,$BD,$F4,$57,$18,$ED,$9F,$F8,$6F,$1E,$D1,$FE,$1E
	.dc.b	$9F,$0F,$85,$FE,$1B,$E9,$93,$9B,$E1,$3F,$EF,$FB,$3F,$E0,$16,$CF
	.dc.b	$F8,$55,$AF,$F8,$69,$AF,$CC,$3D,$64,$3F,$7E,$F9,$1B,$1D,$48,$F5
	.dc.b	$7C,$1E,$0C,$8F,$03,$5F,$9D,$87,$C1,$25,$FC,$1D,$3D,$57,$2B,$DC
	.dc.b	$7F,$90,$30,$F4,$DF,$F8,$8F,$AF,$E0,$78,$BE,$F2,$57,$BD,$F9,$F8
	.dc.b	$FE,$12,$3E,$79,$CF,$1F,$91,$E2,$FF,$91,$FB,$A4,$F8,$06,$49,$DC
	.dc.b	$7C,$B7,$E7,$11,$FC,$CE,$5E,$C7,$F2,$27,$39,$F9,$51,$F8,$C9,$E1
	.dc.b	$3F,$4D,$AF,$F9,$72,$5E,$4F,$DB,$24,$7D,$B7,$EF,$61,$F9,$4C,$7D
	.dc.b	$D8,$7E,$62,$3C,$F5,$F6,$B0,$B3,$FE,$4B,$F0,$3F,$11,$E4,$7E,$1C
	.dc.b	$95,$F8,$77,$AB,$F8,$39,$F6,$3F,$12,$FF,$21,$E6,$7F,$9B,$FC,$39
	.dc.b	$3E,$F8,$CB,$F9,$17,$F2,$84,$7F,$29,$3E,$97,$E3,$9F,$90,$78,$BF
	.dc.b	$97,$8E,$0D,$3F,$39,$E6,$7E,$7C,$8E,$0E,$A7,$A7,$3F,$8A,$7E,$3D
	.dc.b	$F3,$10,$FC,$FC,$FF,$A2,$63,$F5,$9E,$7F,$EA,$C9,$3F,$AB,$7E,$2B
	.dc.b	$F8,$C3,$FD,$03,$CF,$FB,$3C,$24,$7E,$2F,$EB,$DE,$3F,$FA,$67,$4F
	.dc.b	$F4,$34,$62,$7F,$0D,$5C,$1F,$DB,$3D,$9F,$C2,$35,$F5,$1A,$3D,$E5
	.dc.b	$21,$C0,$C9,$F8,$DF,$4F,$F8,$31,$1F,$E0,$CF,$B2,$BF,$15,$1E,$43
	.dc.b	$D5,$E5,$79,$63,$31,$F5,$CF,$77,$95,$FC,$4F,$D3,$FE,$24,$48,$FE
	.dc.b	$24,$FF,$15,$F5,$FE,$16,$DF,$EB,$BC,$DE,$6E,$2F,$FD,$7F,$18,$7F
	.dc.b	$55,$5F,$C7,$CD,$28,$79,$D8,$6C,$70,$21,$FC,$86,$7C,$4F,$F2,$56
	.dc.b	$B9,$30,$43,$04,$3F,$6C,$F0,$7F,$71,$EB,$7F,$31,$23,$FC,$C5,$EB
	.dc.b	$9F,$F4,$C9,$F5,$4F,$F0,$FF,$B0,$7C,$36,$23,$F9,$8C,$DA,$FC,$3F
	.dc.b	$A5,$FC,$FF,$DC,$FE,$7C,$4A,$FE,$7C,$F6,$D7,$FA,$65,$F1,$7F,$DD
	.dc.b	$18,$F6,$AF,$E2,$9E,$E7,$C7,$F5,$A4,$60,$87,$F4,$78,$FE,$98,$47
	.dc.b	$E2,$65,$EB,$7E,$6B,$FE,$E1,$AE,$D1,$FD,$2A,$78,$3C,$CF,$EA,$8F
	.dc.b	$F5,$7F,$80,$91,$FD,$5D,$AF,$D0,$AE,$CF,$C4,$3E,$46,$9F,$D8,$6A
	.dc.b	$47,$A2,$7F,$5B,$EB,$FF,$65,$F7,$FF,$B2,$12,$7F,$B2,$3F,$1D,$FE
	.dc.b	$9D,$A7,$C7,$43,$2F,$5D,$FF,$81,$3E,$0F,$F1,$DE,$D4,$AF,$E7,$D1
	.dc.b	$FD,$CC,$FE,$E4,$47,$FB,$93,$FE,$F5,$AE,$7B,$7D,$69,$F9,$2F,$D6
	.dc.b	$D3,$94,$70,$21,$FD,$E6,$BF,$BF,$7C,$1C,$8F,$EF,$CD,$7D,$27,$33
	.dc.b	$A9,$EE,$3E,$49,$E8,$21,$FE,$00,$8F,$D5,$7B,$BF,$E0,$DF,$86,$FF
	.dc.b	$06,$12,$7F,$C1,$8F,$AE,$FE,$44,$7F,$68,$D3,$F6,$B9,$7B,$A3,$E5
	.dc.b	$BE,$47,$FD,$17,$CA,$F7,$BF,$70,$70,$88,$6C,$E2,$7C,$92,$19,$3E
	.dc.b	$B5,$F8,$BC,$86,$1E,$28,$E7,$3C,$F0,$E0,$F4,$39,$38,$7F,$01,$F0
	.dc.b	$FF,$5C,$8E,$9F,$80,$FC,$C3,$E0,$1F,$30,$FF,$53,$C7,$BF,$83,$F4
	.dc.b	$91,$E2,$7F,$A0,$FC,$07,$C6,$F8,$7F,$C2,$12,$7C,$6F,$67,$F4,$04
	.dc.b	$9F,$D7,$61,$ED,$8F,$E9,$DF,$07,$FC,$84,$C3,$E2,$8E,$4F,$D1,$C3
	.dc.b	$F6,$AD,$8C,$BF,$87,$3E,$33,$87,$EC,$B5,$F9,$E7,$C8,$7F,$39,$F0
	.dc.b	$F2,$3E,$5B,$5E,$67,$C5,$BF,$4B,$FE,$2F,$FB,$17,$07,$A2,$73,$3E
	.dc.b	$93,$FB,$57,$C2,$FE,$33,$E1,$FF,$54,$48,$FB,$2F,$C7,$9E,$24,$E6
	.dc.b	$FD,$1E,$1D,$AB,$F9,$87,$C4,$7F,$47,$F8,$89,$18,$21,$F8,$78,$FD
	.dc.b	$C1,$1F,$80,$7A,$EB,$E4,$59,$EF,$97,$E8,$AB,$85,$73,$3F,$1E,$FF
	.dc.b	$03,$B9,$1F,$92,$6B,$F2,$0F,$84,$4B,$EC,$FF,$16,$C3,$D8,$FE,$42
	.dc.b	$BC,$0F,$AA,$7F,$C1,$F1,$FB,$4C,$BF,$89,$12,$7F,$42,$FC,$9C,$7D
	.dc.b	$64,$8E,$D8,$FE,$F3,$6F,$ED,$D9,$C9,$FE,$B5,$19,$7F,$56,$FF,$1F
	.dc.b	$23,$F2,$0F,$55,$75,$59,$7A,$B2,$FE,$95,$5C,$27,$F7,$0D,$7F,$35
	.dc.b	$C1,$FC,$AC,$8F,$EE,$9F,$53,$F9,$83,$EB,$1E,$A3,$CE,$43,$F6,$F1
	.dc.b	$FE,$09,$D7,$F8,$CF,$27,$F3,$72,$4F,$F0,$67,$AB,$FA,$23,$D6,$75
	.dc.b	$3F,$80,$F3,$7F,$84,$F8,$24,$7F,$7F,$8F,$E4,$A7,$F4,$32,$3F,$C4
	.dc.b	$9E,$BC,$3F,$CA,$63,$B6,$5F,$80,$7E,$3E,$DC,$9E,$04,$3F,$8F,$57
	.dc.b	$F4,$FE,$99,$1F,$C7,$9A,$FE,$36,$BF,$4C,$FC,$70,$F7,$D7,$EA,$6D
	.dc.b	$FE,$0A,$BF,$05,$7B,$8F,$F2,$F7,$1F,$F4,$0E,$BF,$EB,$C4,$8F,$E5
	.dc.b	$AF,$B3,$4F,$DF,$5F,$E9,$A7,$FC,$45,$DD,$FE,$39,$E6,$91,$82,$1F
	.dc.b	$CE,$63,$FB,$51,$1F,$C6,$67,$EA,$7E,$3D,$75,$7F,$99,$CE,$DE,$47
	.dc.b	$F4,$07,$FB,$A7,$81,$23,$FA,$1B,$5F,$EA,$5A,$7C,$27,$E3,$5F,$78
	.dc.b	$C4,$7F,$81,$EB,$FB,$09,$1F,$F4,$11,$FD,$93,$C3,$FF,$01,$11,$FE
	.dc.b	$A2,$D7,$FA,$B3,$5F,$E7,$0D,$7F,$C7,$6D,$7F,$B6,$BF,$E0,$DF,$0E
	.dc.b	$47,$F5,$C6,$BF,$D8,$5A,$FF,$43,$6B,$FE,$46,$6B,$FD,$E1,$AF,$F8
	.dc.b	$8F,$C6,$FF,$0D,$11,$FE,$D2,$D7,$FB,$73,$5F,$EA,$0D,$7F,$CA,$6D
	.dc.b	$7F,$C0,$AF,$F8,$97,$C6,$91,$FD,$D1,$AF,$F7,$96,$BF,$D6,$DA,$FF
	.dc.b	$97,$9A,$FF,$84,$1A,$FF,$8E,$FC,$9F,$F1,$91,$1F,$F0,$0B,$5F,$F0
	.dc.b	$33,$5F,$ED,$0D,$7F,$CD,$6D,$7F,$C3,$AF,$F8,$33,$CF,$FD,$B4,$7F
	.dc.b	$87,$DA,$FF,$8D,$1A,$FF,$93,$9A,$9A,$4D,$7A,$12,$3F,$C4,$11,$96
	.dc.b	$C7,$87,$CF,$91,$A6,$BF,$E2,$B9,$7C,$27,$FA,$AB,$57,$97,$4F,$F5
	.dc.b	$6C,$EA,$BF,$CC,$7B,$FF,$11,$87,$A3,$C7,$C6,$95,$AC,$FD,$D3,$C8
	.dc.b	$F8,$84,$8F,$4B,$F1,$F2,$F0,$47,$F8,$0F,$1F,$0C,$71,$8D,$90,$F5
	.dc.b	$4F,$CF,$F5,$24,$7C,$46,$BE,$F9,$EA,$FF,$1C,$BF,$F0,$1D,$F6,$8F
	.dc.b	$41,$F4,$EF,$C0,$EC,$C1,$FE,$17,$F1,$FE,$2F,$AD,$F7,$C9,$1F,$51
	.dc.b	$F6,$3E,$B2,$1C,$CF,$BD,$87,$CE,$F6,$7E,$97,$95,$27,$C6,$7E,$0E
	.dc.b	$34,$FE,$01,$FC,$71,$1F,$C1,$CF,$BB,$6F,$E7,$19,$7B,$67,$C1,$B2
	.dc.b	$F8,$47,$E0,$DA,$FE,$5F,$DB,$FC,$D1,$1F,$C7,$BE,$C7,$D5,$C1,$F8
	.dc.b	$87,$DB,$87,$E1,$71,$F3,$C7,$82,$EC,$FF,$4B,$F3,$7F,$1B,$EE,$FE
	.dc.b	$9C,$95,$F9,$C7,$B4,$FA,$A3,$AF,$07,$EA,$70,$FC,$AF,$BF,$F9,$9F
	.dc.b	$3E,$57,$C8,$D5,$3A,$8F,$D3,$BF,$B5,$23,$F2,$0E,$A7,$F8,$AF,$6F
	.dc.b	$91,$8F,$E7,$E7,$73,$FA,$96,$BF,$BD,$F8,$4F,$E0,$04,$7F,$6A,$FA
	.dc.b	$B2,$FC,$7C,$76,$3F,$02,$79,$E7,$CC,$FD,$67,$F6,$9E,$8F,$EC,$BE
	.dc.b	$17,$F8,$61,$27,$F8,$0B,$D5,$F4,$1D,$72,$FD,$C3,$F8,$0F,$88,$FD
	.dc.b	$F7,$A9,$23,$E3,$47,$EF,$8F,$E3,$04,$7F,$86,$BD,$5F,$F1,$AE,$9F
	.dc.b	$8D,$CB,$E9,$47,$23,$64,$3F,$87,$4F,$F2,$3F,$85,$91,$FC,$61,$AF
	.dc.b	$ED,$4F,$89,$C7,$F8,$B7,$F2,$FC,$1E,$57,$B1,$EB,$3F,$D1,$1E,$9F
	.dc.b	$F1,$6F,$88,$FE,$60,$48,$FE,$4A,$FA,$6B,$A7,$0F,$E2,$E7,$A9,$F8
	.dc.b	$7B,$FE,$45,$EB,$C8,$C1,$0F,$31,$0F,$E7,$91,$FA,$36,$9F,$8B,$34
	.dc.b	$7C,$62,$9F,$CF,$DE,$0F,$F4,$3F,$86,$FE,$90,$47,$F9,$EB,$F4,$77
	.dc.b	$FB,$77,$FA,$D9,$FF,$54,$7C,$0F,$E5,$B6,$ED,$39,$9F,$DA,$FB,$5F
	.dc.b	$CE,$BE,$1F,$FA,$A1,$23,$FA,$4B,$D7,$2E,$51,$FE,$8B,$C3,$FA,$0F
	.dc.b	$C3,$7F,$45,$F7,$A4,$FE,$B6,$32,$FF,$52,$7F,$B0,$11,$F8,$58,$F8
	.dc.b	$C9,$97,$FA,$CA,$F8,$CF,$F5,$3A,$E0,$D3,$FA,$EF,$BD,$FD,$AC,$8E
	.dc.b	$0E,$B5,$F0,$08,$FA,$4F,$31,$0F,$E4,$33,$FD,$7F,$E0,$3F,$AE,$FC
	.dc.b	$37,$F7,$72,$4F,$F6,$C7,$E2,$3F,$81,$C7,$FC,$C5,$8F,$F6,$4F,$86
	.dc.b	$FE,$D1,$F0,$52,$BF,$88,$46,$5F,$EE,$AF,$F8,$0C,$8F,$A0,$FC,$11
	.dc.b	$D7,$2F,$88,$F2,$38,$90,$FE,$F1,$06,$10,$7D,$92,$F1,$1A,$8C,$38
	.dc.b	$37,$15,$A7,$94,$FE,$2A,$7E,$98,$FE,$7E,$FB,$5F,$E9,$73,$70,$FF
	.dc.b	$43,$70,$F9,$1F,$F4,$B5,$BE,$1F,$B3,$6F,$C4,$D7,$CE,$D7,$F5,$67
	.dc.b	$86,$7F,$12,$FD,$D3,$FB,$54,$3E,$1F,$FC,$35,$F0,$F2,$7F,$B4,$CE
	.dc.b	$5B,$3E,$22,$3E,$27,$EF,$BD,$F3,$FA,$6F,$68,$DB,$E6,$6B,$F4,$65
	.dc.b	$F0,$FE,$E2,$47,$B9,$EB,$87,$AA,$3F,$DB,$FA,$7C,$7C,$3E,$B6,$16
	.dc.b	$7E,$F6,$1E,$97,$C4,$7F,$81,$3A,$3B,$BE,$EC,$BE,$69,$1F,$9A,$F8
	.dc.b	$2F,$F4,$B1,$E0,$3E,$2E,$9F,$E4,$1C,$24,$60,$87,$D4,$8F,$B2,$47
	.dc.b	$F0,$CF,$F1,$49,$EB,$7F,$8E,$9F,$AB,$3B,$79,$1F,$80,$7F,$07,$B4
	.dc.b	$8F,$C1,$B5,$FE,$5F,$9F,$C7,$9E,$D1,$F8,$18,$FC,$6C,$7C,$78,$FD
	.dc.b	$0B,$72,$FF,$35,$47,$C7,$D7,$1C,$1F,$92,$24,$7E,$49,$FC,$AF,$B9
	.dc.b	$CF,$17,$6D,$3A,$47,$5E,$5D,$71,$4C,$3F,$1F,$19,$7F,$3A,$FE,$80
	.dc.b	$8F,$E8,$27,$A3,$FE,$3E,$BE,$EF,$E6,$23,$93,$FA,$49,$E2,$D3,$F5
	.dc.b	$3D,$1F,$AC,$23,$83,$B1,$D7,$0E,$EF,$EA,$2B,$AE,$BF,$B6,$BB,$BF
	.dc.b	$F0,$CC,$5C,$9F,$2F,$BB,$F7,$24,$C7,$F7,$2F,$C8,$7B,$10,$EC,$7E
	.dc.b	$CB,$AF,$E6,$74,$48,$C1,$0F,$E0,$31,$FC,$18,$8F,$A8,$FF,$34,$5F
	.dc.b	$C6,$6E,$9F,$51,$E4,$7E,$CE,$78,$BD,$0F,$E1,$FD,$BF,$C4,$88,$E0
	.dc.b	$ED,$A3,$3F,$85,$7F,$22,$3F,$8D,$EB,$F8,$DC,$25,$FA,$67,$83,$FC
	.dc.b	$D7,$17,$4F,$E4,$FB,$FF,$90,$92,$7F,$90,$BF,$25,$F5,$9F,$26,$BB
	.dc.b	$BF,$C6,$BC,$1F,$CA,$76,$C8,$C1,$0F,$E5,$31,$FC,$B8,$8F,$B3,$4F
	.dc.b	$5B,$D5,$FE,$89,$9F,$E9,$E7,$73,$C8,$FE,$6A,$FF,$38,$F1,$24,$7F
	.dc.b	$38,$6B,$FC,$E2,$7E,$35,$76,$8F,$C0,$9F,$CD,$AB,$B3,$E7,$C3,$E3
	.dc.b	$AC,$DD,$FF,$B9,$F1,$BF,$A3,$92,$7F,$A3,$BF,$1E,$B8,$D7,$E0,$DB
	.dc.b	$BF,$E0,$EB,$FA,$2F,$93,$FB,$BF,$06,$47,$F4,$48,$FE,$0C,$7F,$54
	.dc.b	$23,$FD,$51,$ED,$9F,$AE,$FA,$CF,$ED,$4E,$46,$C8,$7F,$5A,$9F,$EB
	.dc.b	$FE,$5C,$8F,$EB,$ED,$7E,$41,$F4,$CF,$C4,$61,$E9,$AF,$ED,$35,$E6
	.dc.b	$9E,$E7,$E8,$BC,$3F,$E3,$DE,$77,$F6,$C2,$47,$F6,$C7,$D2,$F4,$3B
	.dc.b	$1F,$E0,$6D,$FB,$65,$FB,$FF,$17,$F9,$01,$89,$FE,$A3,$3F,$95,$7F
	.dc.b	$AF,$35,$FC,$49,$FB,$53,$25,$32,$7B,$88,$68,$FF,$0A,$FD,$9F,$F8
	.dc.b	$45,$34,$8F,$A8,$F4,$1F,$44,$FE,$07,$E8,$79,$DC,$9F,$90,$CF,$F0
	.dc.b	$BE,$47,$F4,$5A,$7B,$1F,$E1,$F6,$D7,$F8,$8D,$AF,$92,$3C,$1E,$1B
	.dc.b	$1E,$DC,$4C,$BA,$7B,$7F,$D1,$FC,$C9,$8F,$FB,$E3,$24,$EA,$3B,$3E
	.dc.b	$12,$3F,$B5,$3E,$79,$ED,$D3,$F3,$B7,$C6,$3E,$3C,$F0,$69,$E9,$F6
	.dc.b	$BD,$84,$70,$75,$3E,$2D,$7F,$0C,$7E,$49,$E7,$9E,$D1,$FB,$4B,$F4
	.dc.b	$CF,$F5,$5F,$37,$E5,$7C,$07,$C9,$24,$7C,$97,$E3,$E9,$F1,$B2,$FA
	.dc.b	$1F,$07,$FD,$AF,$D0,$97,$F1,$63,$2F,$D1,$7E,$A1,$1F,$46,$FF,$45
	.dc.b	$F8,$9C,$7F,$D5,$6F,$8C,$7E,$2A,$B8,$34,$FB,$9F,$0B,$F7,$C8,$E0
	.dc.b	$EB,$8F,$9D,$F2,$10,$6B,$F0,$EB,$FC,$09,$E9,$FE,$2F,$E1,$BF,$14
	.dc.b	$4A,$FC,$53,$F0,$9F,$39,$FD,$0F,$1F,$C9,$7C,$37,$F8,$37,$D5,$93
	.dc.b	$F8,$78,$CB,$F9,$37,$F2,$C4,$7C,$E7,$43,$F9,$44,$F0,$7F,$57,$1B
	.dc.b	$69,$F9,$DF,$85,$FD,$01,$1C,$1D,$0F,$D4,$6B,$F3,$5F,$21,$0F,$86
	.dc.b	$FF,$A4,$38,$95,$FE,$27,$F5,$FF,$5D,$F0,$DF,$AD,$25,$FE,$B5,$F8
	.dc.b	$8F,$63,$E2,$1F,$A7,$F8,$6F,$F1,$87,$B7,$23,$04,$3F,$83,$C7,$EE
	.dc.b	$48,$FA,$78,$3F,$38,$7E,$A1,$E0,$7F,$A9,$71,$4F,$08,$E6,$7F,$01
	.dc.b	$F8,$4F,$E0,$84,$70,$75,$CF,$F8,$DD,$F5,$8F,$86,$7F,$AC,$EF,$AB
	.dc.b	$E7,$8F,$D5,$5F,$F9,$2F,$DC,$FE,$25,$F0,$BF,$C4,$49,$5F,$C4,$5A
	.dc.b	$FF,$24,$6B,$E9,$6B,$F4,$5A,$F8,$5F,$E3,$9F,$0B,$23,$F8,$E3,$5F
	.dc.b	$E5,$ED,$7E,$23,$5F,$B8,$D7,$D8,$D7,$F9,$AF,$C2,$7F,$29,$23,$FC
	.dc.b	$A5,$AF,$F3,$C6,$BF,$45,$AF,$E2,$5A,$FC,$97,$F9,$A7,$C2,$C8,$FE
	.dc.b	$68,$D7,$FA,$3B,$5F,$B8,$D7,$F2,$8D,$7E,$A3,$5F,$E9,$BF,$0B,$FD
	.dc.b	$04,$8F,$F4,$16,$BF,$D5,$1A,$FE,$25,$AF,$E8,$5A,$FE,$05,$FE,$99
	.dc.b	$F0,$B2,$3F,$A6,$35,$FE,$BE,$D7,$F2,$8D,$7F,$58,$D7,$F1,$8D,$7F
	.dc.b	$B2,$FC,$2F,$F5,$92,$3F,$D6,$5A,$FF,$6C,$6B,$FA,$16,$BF,$B9,$6B
	.dc.b	$F9,$97,$FB,$27,$C2,$C8,$FE,$C8,$D7,$FB,$BB,$5F,$D6,$35,$FE,$08
	.dc.b	$D7,$F4,$8D,$7F,$BE,$FC,$2F,$F7,$12,$3F,$DC,$5A,$FF,$81,$1A,$FE
	.dc.b	$E5,$AF,$F1,$16,$BF,$B1,$4C,$20,$FB,$5F,$F2,$7B,$86,$BB,$6B,$C9
	.dc.b	$AF,$1E,$1E,$6E,$DA,$ED,$AE,$CC,$C7,$22,$46,$9C,$74,$52,$4F,$06
	.dc.b	$BE,$08,$FF,$08,$39,$21,$98,$DB,$ED,$6B,$ED,$DF,$C0,$47,$CA,$D7
	.dc.b	$B3,$5F,$8C,$D7,$91,$C0,$87,$97,$93,$D8,$47,$E0,$B5,$D1,$82,$18
	.dc.b	$21,$F5,$5E,$0F,$D7,$E6,$F9,$04,$7E,$6B,$5F,$8C,$D7,$81,$A2,$1A
	.dc.b	$29,$F8,$3E,$9F,$A8,$47,$EC,$B5,$FA,$4D,$78,$1A,$21,$A2,$9F,$8F
	.dc.b	$EB,$FC,$01,$1F,$C2,$B5,$FB,$CD,$78,$1A,$21,$A2,$9F,$9B,$F0,$3F
	.dc.b	$18,$47,$F2,$2D,$7F,$14,$D7,$81,$A2,$1A,$29,$FA,$7F,$0B,$F3,$04
	.dc.b	$7F,$3A,$D7,$F2,$CD,$78,$1A,$21,$A2,$9F,$B3,$F1,$7F,$48,$47,$F5
	.dc.b	$2D,$7F,$44,$D7,$81,$A2,$1A,$29,$FB,$FF,$23,$F6,$04,$7F,$6A,$FE
	.dc.b	$E3,$ED,$5C,$5F,$C2,$7A,$BF,$DA,$B1,$8A,$FE,$1B,$18,$9D,$70,$FF
	.dc.b	$0F,$A7,$1E,$2F,$F8,$9C,$38,$F1,$08,$C3,$D2,$B9,$4E,$BC,$8F,$F7
	.dc.b	$2E,$3F,$EE,$6E,$47,$9D,$FC,$24,$98,$7F,$0E,$7D,$59,$F6,$CB,$D1
	.dc.b	$3F,$E1,$63,$C1,$19,$CF,$59,$49,$F1,$1E,$87,$C5,$9F,$A3,$17,$97
	.dc.b	$FC,$64,$95,$FC,$79,$FE,$43,$C4,$F8,$5F,$17,$4C,$AE,$A2,$74,$E2
	.dc.b	$DE,$77,$A7,$AA,$4E,$63,$C4,$7F,$29,$23,$FC,$B9,$F0,$E1,$DE,$3E
	.dc.b	$06,$CC,$7E,$8E,$3E,$09,$F8,$98,$F8,$1E,$44,$3F,$98,$61,$FC,$DF
	.dc.b	$D1,$91,$FC,$F1,$AF,$F4,$4A,$E9,$C3,$F8,$8C,$3C,$4B,$5F,$EA,$93
	.dc.b	$A7,$FA,$86,$1E,$CB,$F7,$1F,$0F,$DD,$FA,$1E,$A7,$F4,$72,$47,$F4
	.dc.b	$C7,$DE,$FC,$CD,$3D,$CD,$CB,$DF,$AF,$58,$F7,$E1,$73,$FC,$66,$3F
	.dc.b	$77,$D7,$FE,$A8,$4A,$FE,$B2,$F6,$7E,$5E,$1E,$CC,$FD,$78,$F3,$7D
	.dc.b	$73,$EA,$3E,$6D,$FE,$27,$D9,$FE,$C0,$48,$FE,$CA,$F5,$C7,$E4,$47
	.dc.b	$A4,$FC,$3E,$BF,$1D,$FE,$E9,$5D,$DF,$EE,$D9,$FE,$56,$3F,$2D,$EE
	.dc.b	$7F,$6D,$24,$FF,$72,$7B,$63,$FB,$7C,$FD,$53,$FE,$08,$C3,$B3,$FE
	.dc.b	$07,$BE,$CF,$F8,$27,$0E,$CF,$D5,$3F,$BE,$C7,$F7,$B2,$3F,$E0,$49
	.dc.b	$E7,$AF,$D4,$8F,$92,$7B,$34,$E6,$FF,$7D,$AE,$2D,$3F,$C1,$3F,$01
	.dc.b	$FE,$07,$23,$83,$B4,$FA,$9F,$E1,$98,$FE,$01,$F3,$1F,$48,$BA,$7E
	.dc.b	$4D,$F8,$13,$FA,$4F,$86,$FD,$C9,$23,$D4,$FC,$87,$F8,$DD,$77,$CB
	.dc.b	$E3,$F1,$32,$F5,$E3,$F8,$CB,$F7,$1F,$6B,$E0,$FF,$82,$91,$F8,$CF
	.dc.b	$57,$D2,$7E,$04,$FE,$31,$7F,$50,$EF,$5D,$B6,$FC,$8C,$FE,$1B,$E1
	.dc.b	$BF,$88,$91,$FA,$4F,$5D,$BE,$96,$3F,$8E,$9E,$BA,$FC,$77,$E3,$12
	.dc.b	$3E,$33,$FC,$5E,$3F,$84,$FC,$3F,$F1,$E2,$47,$DF,$7B,$6B,$F1,$67
	.dc.b	$A5,$7C,$3A,$E9,$8F,$A1,$F0,$9F,$68,$87,$F0,$B9,$FD,$83,$FE,$5A
	.dc.b	$CF,$F8,$C3,$63,$2F,$33,$07,$22,$9F,$C5,$7E,$1B,$F9,$A9,$1F,$CB
	.dc.b	$3F,$46,$7E,$74,$FF,$49,$F2,$3D,$37,$E7,$C3,$E1,$D7,$D3,$F7,$7F
	.dc.b	$95,$F7,$FF,$40,$26,$5F,$A1,$7A,$63,$CA,$3F,$B1,$E3,$5F,$98,$7E
	.dc.b	$18,$FB,$F5,$FC,$DB,$E1,$BF,$A5,$92,$3F,$58,$F4,$78,$63,$FD,$DA
	.dc.b	$BA,$63,$EA,$9F,$E1,$37,$FD,$03,$E1,$7F,$AA,$92,$3F,$6C,$F3,$7E
	.dc.b	$99,$FD,$22,$B9,$E3,$D2,$BF,$A4,$FC,$37,$F5,$E2,$57,$EF,$DE,$58
	.dc.b	$FC,$6A,$F9,$53,$E6,$D3,$E5,$11,$FE,$21,$7F,$D5,$3E,$1B,$FB,$49
	.dc.b	$23,$F8,$53,$D2,$78,$47,$F8,$27,$1F,$2B,$D4,$E6,$7F,$28,$F8,$5F
	.dc.b	$EE,$64,$7F,$8A,$3C,$DF,$E0,$27,$F4,$98,$FE,$CB,$7D,$23,$E9,$9F
	.dc.b	$CC,$BE,$1B,$FB,$E9,$1F,$E3,$CF,$37,$ED,$9F,$24,$FA,$27,$94,$CE
	.dc.b	$5F,$8A,$9F,$EE,$7F,$0D,$FE,$08,$24,$FF,$29,$7A,$6D,$FC,$AB,$0F
	.dc.b	$2C,$F3,$8F,$C9,$3F,$D1,$BE,$17,$FC,$22,$47,$F9,$93,$CB,$5F,$A7
	.dc.b	$1E,$C3,$EF,$E3,$F5,$DF,$39,$FC,$43,$E1,$BF,$C3,$64,$7F,$9E,$3C
	.dc.b	$DF,$90,$70,$8E,$4F,$4D,$7F,$BF,$E7,$E8,$8F,$F0,$33,$FE,$26,$23
	.dc.b	$FC,$02,$78,$92,$7F,$C9,$99,$71,$9F,$E0,$55,$C1,$E6,$7F,$4F,$F8
	.dc.b	$5F,$F1,$89,$1C,$1D,$6B,$F1,$39,$FC,$B8,$F4,$47,$99,$A7,$F9,$63
	.dc.b	$D3,$FF,$13,$FC,$37,$F8,$F0,$91,$FD,$75,$FF,$6F,$FD,$B3,$DB,$FF
	.dc.b	$28,$7B,$12,$BF,$C7,$F3,$97,$FB,$DF,$C3,$FF,$95,$7D,$99,$18,$21
	.dc.b	$FE,$2B,$8F,$F2,$59,$1F,$F7,$7C,$F3,$7F,$C7,$B5,$DD,$FF,$2F,$46
	.dc.b	$77,$FE,$2F,$5C,$67,$A1,$FD,$D5,$FF,$2B,$FC,$0C,$8F,$EF,$0D,$7F
	.dc.b	$DF,$F5,$CF,$4F,$F3,$F6,$1D,$F0,$FE,$8B,$39,$AF,$F0,$56,$1F,$E4
	.dc.b	$E2,$3F,$0B,$2F,$F4,$0C,$A3,$E4,$90,$F6,$9F,$32,$3B,$61,$E0,$8E
	.dc.b	$47,$C6,$BF,$6B,$63,$A2,$AF,$23,$7E,$1F,$EB,$91,$F9,$6F,$77,$FC
	.dc.b	$95,$8F,$81,$F3,$4F,$2D,$99,$7B,$27,$FB,$AB,$EC,$CB,$DB,$F0,$DF
	.dc.b	$84,$24,$FD,$87,$B3,$F6,$0E,$E7,$B0,$E6,$7A,$CF,$96,$67,$3F,$9F
	.dc.b	$F0,$FF,$90,$24,$7E,$11,$EB,$1F,$EB,$8C,$BD,$53,$F0,$EF,$D1,$3F
	.dc.b	$22,$3B,$9F,$88,$8F,$AB,$F0,$FF,$9C,$24,$7E,$41,$EB,$AF,$D3,$8F
	.dc.b	$51,$C4,$F3,$3F,$E7,$EC,$BB,$BF,$53,$4F,$A6,$3E,$17,$F5,$24,$AF
	.dc.b	$CE,$BD,$9F,$F6,$4E,$5E,$A9,$E3,$8F,$49,$F8,$A7,$84,$FD,$0F,$C3
	.dc.b	$7E,$D0,$8F,$EA,$1E,$8F,$FB,$46,$BA,$4F,$D9,$8F,$A0,$7F,$AD,$71
	.dc.b	$C3,$FA,$CF,$86,$FD,$F9,$1F,$D9,$BD,$3C,$65,$F4,$9F,$86,$79,$B5
	.dc.b	$FA,$07,$EA,$30,$FD,$1F,$C2,$FF,$0A,$24,$FE,$F9,$E9,$5C,$E3,$DD
	.dc.b	$5C,$B5,$F9,$D1,$D8,$FD,$07,$C3,$7F,$14,$23,$FC,$21,$E6,$FE,$40
	.dc.b	$E7,$1F,$57,$0F,$99,$1F,$04,$E9,$AF,$F0,$F7,$F5,$A7,$E2,$CF,$E2
	.dc.b	$71,$F9,$C7,$31,$B7,$A1,$A2,$1F,$C3,$23,$83,$A3,$07,$F1,$EF,$84
	.dc.b	$FE,$5A,$47,$07,$2C,$3E,$53,$F0,$73,$4F,$D8,$9C,$CF,$B8,$FE,$21
	.dc.b	$2F,$86,$FE,$26,$7F,$3A,$24,$7F,$2E,$7F,$0F,$87,$E4,$23,$FD,$9D
	.dc.b	$7D,$71,$F4,$61,$D8,$FF,$33,$E1,$D6,$BF,$A0,$E7,$FD,$18,$8F,$F3
	.dc.b	$A7,$AE,$5C,$E3,$CB,$85,$7E,$71,$E9,$15,$D5,$FE,$7F,$CF,$FD,$48
	.dc.b	$8F,$F4,$67,$AB,$F1,$67,$A6,$5E,$B8,$63,$E8,$CF,$F9,$AF,$67,$F5
	.dc.b	$C2,$4F,$F5,$07,$9B,$E3,$C7,$E7,$BF,$86,$CB,$D1,$1E,$5B,$FE,$8B
	.dc.b	$D5,$FD,$98,$95,$FD,$69,$E7,$1F,$27,$5F,$E7,$35,$D2,$BE,$06,$5E
	.dc.b	$7A,$F3,$47,$F6,$CF,$27,$FB,$89,$1F,$EC,$AF,$4B,$FA,$A7,$4D,$3E
	.dc.b	$0E,$5E,$6D,$3F,$A9,$FA,$FF,$DE,$89,$87,$F6,$E7,$93,$FD,$0A,$7F
	.dc.b	$A6,$E3,$F7,$AF,$9C,$FE,$46,$FF,$AE,$FB,$7F,$E0,$42,$47,$F7,$87
	.dc.b	$9B,$F7,$2B,$93,$F2,$4E,$47,$F8,$23,$DB,$FF,$06,$91,$FF,$00,$BC
	.dc.b	$B4,$FE,$BB,$AF,$C7,$AF,$95,$5E,$73,$99,$FE,$04,$F6,$FF,$C2,$E4
	.dc.b	$7F,$C1,$6F,$3C,$FD,$BA,$FC,$79,$F2,$1F,$7F,$3C,$BF,$DE,$7E,$DF
	.dc.b	$FD,$8A,$FF,$04,$6B,$FD,$C9,$AE,$E7,$B3,$FA,$A3,$AE,$1E,$87,$F3
	.dc.b	$B5,$D9,$53,$E1,$BF,$88,$C7,$56,$BE,$36,$BC,$1E,$EF,$E7,$AF,$D8
	.dc.b	$D3,$F8,$1D,$7D,$A7,$07,$88,$FF,$0B,$5F,$FB,$8A,$DF,$0D,$D9,$B3
	.dc.b	$E2,$6B,$E6,$6B,$F1,$63,$81,$1E,$0D,$78,$35,$E0,$75,$F8,$6F,$01
	.dc.b	$1C,$10,$F0,$39,$69,$93,$EB,$39,$69,$E9,$F8,$6F,$61,$1C,$10,$F6
	.dc.b	$39,$69,$93,$F0,$4E,$5A,$7C,$5F,$07,$E4,$11,$C1,$0F,$90,$E5,$A6
	.dc.b	$4F,$C7,$39,$69,$F4,$7E,$1B,$EA,$11,$C1,$0F,$A8,$E5,$A6,$4F,$CD
	.dc.b	$39,$69,$F7,$7E,$1B,$F0,$04,$70,$43,$F0,$0E,$5A,$64,$FD,$33,$96
	.dc.b	$9F,$89,$F8,$6F,$C6,$11,$C1,$0F,$C6,$39,$69,$93,$F6,$4E,$5A,$7E
	.dc.b	$57,$E1,$BF,$30,$47,$04,$3F,$30,$E5,$A6,$4F,$DF,$39,$69,$FA,$1F
	.dc.b	$86,$FD,$21,$1C,$10,$FD,$23,$96,$99,$3F,$85,$39,$69,$FA,$DF,$86
	.dc.b	$FD,$81,$1C,$1F,$B1,$F8,$5F,$E1,$1C,$AF,$02,$9F,$C2,$F8,$13,$4D
	.dc.b	$F2,$CB,$FD,$FF,$27,$F1,$99,$E1,$F0,$DF,$BF,$24,$FE,$FD,$FE,$05
	.dc.b	$1D,$A3,$89,$FE,$0E,$C3,$B3,$83,$F8,$DC,$77,$7F,$87,$E3,$FE,$96
	.dc.b	$C7,$CB,$1B,$5F,$05,$FC,$3C,$91,$FC,$3D,$F0,$BD,$67,$E0,$36,$FC
	.dc.b	$2F,$F1,$CA,$EA,$F8,$32,$F1,$72,$FF,$1D,$C7,$73,$FC,$ED,$EA,$FF
	.dc.b	$1B,$6B,$93,$E8,$10,$D1,$4D,$18,$3F,$92,$7C,$37,$F2,$92,$3F,$CA
	.dc.b	$5E,$F3,$E3,$8F,$85,$3F,$73,$C5,$F5,$CE,$67,$D9,$3F,$80,$F1,$7E
	.dc.b	$F7,$C3,$7F,$35,$24,$FF,$35,$7A,$EF,$F7,$63,$A9,$F2,$70,$EB,$19
	.dc.b	$8E,$A7,$CD,$AF,$C5,$7C,$37,$F4,$12,$47,$F4,$17,$FA,$27,$C0,$FE
	.dc.b	$4B,$CB,$93,$F6,$E3,$F2,$47,$E5,$BC,$B9,$19,$3F,$01,$F0,$DF,$D3
	.dc.b	$48,$FF,$4D,$7A,$BF,$7E,$BA,$BF,$76,$BA,$BF,$83,$AE,$AE,$4E,$C7
	.dc.b	$F5,$58,$E6,$D3,$FA,$D3,$87,$FB,$E3,$5F,$EF,$8D,$7F,$C0,$4D,$7F
	.dc.b	$BB,$FC,$37,$F6,$52,$3F,$D9,$5E,$FB,$7F,$6B,$9F,$3C,$FE,$8B,$CF
	.dc.b	$B3,$98,$F3,$BF,$E6,$CF,$3F,$FC,$09,$F7,$2F,$87,$FE,$FE,$DA,$FF
	.dc.b	$01,$B5,$FF,$03,$B5,$FF,$4C,$FD,$CB,$D5,$FE,$0F,$10,$FD,$B4,$65
	.dc.b	$FD,$67,$AB,$FC,$33,$D5,$91,$83,$F7,$96,$F8,$7F,$19,$23,$C6,$FF
	.dc.b	$85,$2F,$A4,$7F,$00,$3C,$07,$F8,$41,$CE,$7F,$E8,$AC,$3B,$CF,$F0
	.dc.b	$6C,$3F,$8E,$C7,$C2,$F8,$7F,$82,$49,$F8,$2D,$7E,$0B,$5F,$F5,$CE
	.dc.b	$28,$C4,$09,$19,$79,$90,$F9,$7F,$0F,$F4,$08,$FD,$06,$BA,$30,$43
	.dc.b	$04,$3C,$6F,$07,$CB,$EE,$FD,$C2,$3F,$71,$AF,$DC,$6B,$C0,$D1,$0D
	.dc.b	$14,$F8,$3F,$0F,$F8,$82,$3F,$88,$6B,$F8,$86,$BC,$0D,$10,$D1,$4F
	.dc.b	$9B,$F0,$FF,$94,$23,$F9,$46,$BF,$94,$6B,$C0,$D1,$0D,$14,$FB,$3F
	.dc.b	$0F,$FA,$02,$3F,$A0,$7A,$E8,$EA,$FB,$92,$C9,$7F,$AA,$3F,$CD,$DB
	.dc.b	$3E,$1F,$FD,$87,$4F,$BA,$74,$72,$C3,$8C,$FE,$83,$0E,$4F,$E8,$AF
	.dc.b	$F6,$A7,$ED,$08,$FE,$D1,$F2,$E3,$CE,$78,$CF,$E8,$67,$C3,$87,$82
	.dc.b	$3F,$27,$1C,$9F,$DD,$B5,$FC,$DF,$C4,$7F,$02,$23,$FC,$09,$FD,$84
	.dc.b	$74,$C7,$C1,$3F,$9F,$AE,$CF,$C5,$BF,$8F,$18,$CB,$E2,$67,$BB,$F9
	.dc.b	$70,$F8,$0F,$92,$4C,$3F,$84,$FC,$07,$F1,$6F,$88,$FE,$2A,$48,$FE
	.dc.b	$2A,$F9,$30,$E9,$1F,$B0,$CB,$C3,$8F,$9A,$3E,$93,$E0,$CF,$D5,$1F
	.dc.b	$B7,$CB,$C4,$FF,$18,$CB,$C4,$FF,$0B,$CB,$F9,$3F,$C4,$7F,$26,$24
	.dc.b	$7F,$26,$7C,$4F,$63,$C4,$79,$0F,$88,$5A,$F8,$6E,$C2,$F8,$4F,$C5
	.dc.b	$FC,$24,$9F,$E5,$71,$FC,$E3,$E2,$3F,$9B,$92,$3F,$9B,$BE,$2D,$FE
	.dc.b	$9F,$17,$E3,$A7,$B3,$63,$EA,$69,$F0,$A7,$C3,$7D,$DF,$E5,$58,$7E
	.dc.b	$72,$3F,$A3,$FC,$47,$F4,$62,$4F,$F4,$67,$C1,$B7,$CF,$C3,$C3,$BF
	.dc.b	$D2,$AF,$9D,$8F,$7D,$1F,$11,$FA,$8F,$88,$95,$FD,$22,$3F,$AA,$9F
	.dc.b	$D5,$08,$FF,$54,$7B,$BF,$38,$F6,$DF,$77,$E5,$DF,$79,$EB,$1F,$D1
	.dc.b	$6B,$F6,$F1,$FD,$89,$FE,$C2,$48,$FE,$C2,$F6,$7E,$C9,$D8,$FB,$53
	.dc.b	$EB,$DB,$F8,$09,$23,$FB,$0E,$7F,$DB,$B4,$FE,$DA,$48,$FE,$DA,$F5
	.dc.b	$DB,$EA,$6B,$F2,$4F,$CB,$71,$7E,$1F,$0E,$97,$FD,$75,$FE,$1F,$87
	.dc.b	$F7,$AD,$FF,$BC,$92,$7F,$BC,$BD,$75,$F8,$98,$F6,$9F,$A7,$87,$E4
	.dc.b	$F8,$FF,$8C,$F0,$C9,$FE,$F3,$1F,$E0,$6E,$3F,$F0,$29,$23,$FC,$0A
	.dc.b	$F5,$7F,$3D,$7D,$5F,$A3,$7D,$5F,$F0,$1D,$F5,$7F,$BB,$5F,$F8,$4B
	.dc.b	$8F,$FC,$22,$48,$FF,$08,$BD,$5F,$A0,$7E,$9F,$8B,$F5,$3C,$7F,$9E
	.dc.b	$9E,$13,$FE,$16,$6B,$FD,$DB,$BB,$FC,$3A,$47,$FC,$3A,$F5,$9F,$CB
	.dc.b	$C7,$DC,$C7,$D7,$3B,$8F,$F1,$0B,$5F,$E8,$67,$F2,$02,$94,$CF,$06
	.dc.b	$1F,$E9,$2D,$7E,$89,$FE,$1F,$7F,$8B,$77,$78,$A3,$06,$9F,$E5,$B1
	.dc.b	$C6,$3B,$1D,$0E,$8E,$47,$6E,$09,$1E,$38,$CB,$E5,$CB,$E5,$C7,$E4
	.dc.b	$63,$F5,$98,$FF,$06,$C7,$F8,$D4,$F0,$9F,$7B,$5F,$7D,$F6,$73,$DD
	.dc.b	$E4,$23,$E1,$7E,$84,$F7,$8F,$C3,$9F,$C8,$D4,$70,$8E,$67,$13,$D5
	.dc.b	$E0,$FC,$02,$38,$3D,$9C,$DF,$C7,$78,$3F,$28,$7E,$57,$3F,$7C,$FF
	.dc.b	$8D,$F0,$EE,$FF,$B3,$5A,$B2,$FF,$01,$3B,$C7,$FB,$FE,$1F,$CA,$69
	.dc.b	$E0,$FD,$22,$57,$CF,$7D,$EF,$CC,$8F,$DE,$BF,$9B,$9E,$55,$FC,$66
	.dc.b	$7C,$3A,$7D,$1A,$F8,$9E,$27,$DE,$23,$F6,$DE,$AF,$F1,$68,$E0,$49
	.dc.b	$FC,$8D,$75,$3F,$93,$65,$F0,$AB,$FC,$2B,$3F,$8A,$23,$F8,$77,$AE
	.dc.b	$7F,$27,$1F,$E7,$47,$A2,$7F,$0B,$7D,$23,$E2,$79,$1F,$94,$24,$FE
	.dc.b	$45,$E7,$CB,$BD,$7F,$61,$3F,$81,$BF,$33,$8C,$FF,$43,$E4,$7E,$78
	.dc.b	$91,$F9,$B7,$9B,$BC,$FE,$6C,$FE,$A2,$B9,$CF,$CF,$BF,$CD,$F9,$DF
	.dc.b	$A8,$24,$7E,$95,$E6,$EC,$95,$CD,$F7,$9D,$CF,$C5,$79,$BF,$B2,$23
	.dc.b	$FA,$F7,$9C,$7D,$38,$FE,$02,$F3,$7E,$01,$CC,$FF,$23,$FC,$17,$EF
	.dc.b	$08,$FE,$E5,$E7,$8F,$B8,$91,$F9,$53,$CA,$7A,$1F,$E2,$3E,$EF,$F0
	.dc.b	$62,$47,$F0,$37,$9B,$E4,$8E,$6F,$F9,$63,$2F,$93,$7E,$98,$FF,$2D
	.dc.b	$FC,$47,$F1,$02,$3F,$C3,$5E,$6F,$EF,$23,$85,$7E,$70,$FC,$9C,$7B
	.dc.b	$D7,$C4,$7F,$1A,$24,$7F,$17,$79,$F1,$7E,$5C,$FC,$C4,$F3,$9F,$80
	.dc.b	$7B,$CF,$DE,$7C,$47,$F2,$42,$3F,$C8,$5E,$7C,$5F,$9A,$DF,$E3,$E3
	.dc.b	$F4,$CF,$3F,$C4,$7F,$2C,$25,$FF,$29,$78,$CF,$A7,$4F,$8B,$8F,$C8
	.dc.b	$1F,$11,$FC,$D0,$97,$FC,$C5,$E2,$F9,$CF,$9C,$78,$A3,$97,$C4,$7F
	.dc.b	$3C,$24,$7F,$39,$78,$E3,$F3,$2F,$C3,$3F,$64,$FE,$39,$F1,$1F,$D1
	.dc.b	$08,$FF,$41,$78,$DF,$9E,$78,$CF,$D7,$8F,$EA,$FF,$11,$FD,$30,$91
	.dc.b	$FD,$25,$E2,$F7,$3D,$27,$E6,$8F,$64,$3E,$1F,$FA,$A9,$23,$FA,$93
	.dc.b	$C9,$F0,$1F,$6B,$0F,$A5,$3F,$92,$FB,$BF,$E6,$33,$FD,$B9,$AF,$F7
	.dc.b	$26,$BF,$EB,$66,$BE,$5E,$AF,$E8,$3D,$3F,$C3,$E3,$CB,$97,$C0,$E2
	.dc.b	$FE,$45,$6F,$88,$FE,$EC,$DA,$FE,$F4,$D7,$C5,$1E,$99,$FC,$9D,$73
	.dc.b	$E2,$FF,$36,$5B,$E2,$39,$93,$0E,$6F,$19,$FE,$55,$1F,$27,$4F,$D7
	.dc.b	$53,$E2,$3C,$64,$9F,$1B,$C7,$8F,$F1,$11,$F2,$76,$F9,$1F,$11,$EA
	.dc.b	$24,$7A,$9E,$0F,$78,$FD,$9B,$F9,$3B,$7C,$47,$C2,$24,$7C,$27,$8E
	.dc.b	$3F,$5A,$3C,$07,$E0,$E5,$F1,$1F,$28,$91,$F2,$9E,$37,$E0,$AE,$53
	.dc.b	$F8,$21,$4F,$88,$FA,$64,$8F,$A6,$F2,$7C,$27,$E4,$A7,$E4,$FC,$47
	.dc.b	$DA,$24,$7D,$A7,$83,$BB,$F1,$1F,$8F,$F8,$8F,$C0,$92,$3F,$02,$F0
	.dc.b	$8F,$E2,$11,$C4,$F8,$1F,$11,$F8,$82,$3F,$88,$78,$70,$FE,$17,$6F
	.dc.b	$F6,$26,$6F,$88,$FC,$79,$2F,$F1,$EF,$07,$EE,$9F,$16,$1F,$11,$F9
	.dc.b	$62,$47,$E5,$9E,$0F,$DD,$3E,$A7,$C4,$7E,$70,$8F,$E7,$1D,$EB,$F4
	.dc.b	$B9,$7F,$47,$F0,$BF,$A2,$25,$7E,$89,$DE,$BF,$22,$3F,$01,$F0,$DF
	.dc.b	$A8,$23,$FA,$87,$6F,$D6,$AF,$D7,$FC,$37,$EB,$89,$1F,$AE,$76,$FC
	.dc.b	$D3,$E0,$7C,$37,$ED,$08,$FE,$D1,$DD,$FD,$4A,$FD,$DF,$C3,$7E,$E8
	.dc.b	$95,$FB,$A7,$7C,$1F,$C0,$F3,$5F,$F0,$9F,$85,$FE,$02,$4B,$FE,$02
	.dc.b	$F0,$7F,$BB,$6E,$F8,$6F,$E0,$E4,$8F,$E0,$EE,$DF,$E7,$D9,$3E,$1B
	.dc.b	$FD,$04,$D8,$FF,$39,$46,$0E,$18,$FF,$6F,$97,$A9,$C0,$95,$FE,$79
	.dc.b	$77,$97,$F7,$89,$7C,$37,$62,$57,$F9,$FD,$DB,$FC,$E6,$1F,$0D,$E4
	.dc.b	$24,$7F,$A1,$5D,$BF,$B5,$87,$C3,$7A,$89,$1F,$E8,$B7,$71,$FD,$DF
	.dc.b	$E1,$BD,$E4,$9F,$F4,$83,$F0,$BC,$1F,$F3,$74,$7F,$A8,$0F,$F7,$7E
	.dc.b	$5F,$EA,$33,$FC,$C4,$4C,$BF,$8B,$BF,$E2,$93,$10,$16,$F1,$3E,$79
	.dc.b	$23,$FD,$40,$F9,$AF,$C3,$A7,$E0,$A7,$95,$3E,$1B,$EB,$92,$3F,$D5
	.dc.b	$0F,$1E,$BF,$F2,$CC,$7F,$B1,$25,$7E,$C8,$C3,$5F,$F4,$09,$89,$FF
	.dc.b	$4B,$91,$FF,$65,$CF,$28,$C4,$6C,$CC,$F3,$7F,$D9,$67,$87,$E1,$5F
	.dc.b	$8A,$7F,$AF,$5B,$1D,$1A,$EE,$34,$47,$4D,$74,$60,$86,$0F,$A1,$F7
	.dc.b	$A2,$00

;--------------------------------------------------------------------------------
;	FACOS.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
facos_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FACOS.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FACOS.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	facos_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1687+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	facos.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FACOS.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1687+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	facos.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FACOS.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
facos_data_compressed::
	.dc.b	$00,$02,$93,$04,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$FA,$77,$E2,$7F,$C2
	.dc.b	$CF,$C1,$F0,$9F,$C1,$BF,$D7,$5F,$EA,$2D,$F3,$38,$F9,$9C,$BC,$CE
	.dc.b	$7E,$67,$4F,$33,$AF,$99,$DB,$CC,$EF,$E6,$78,$3C,$CF,$0F,$99,$E2
	.dc.b	$F3,$3C,$7E,$67,$F6,$1F,$C5,$1F,$2F,$7E,$2B,$E7,$FA,$7F,$D9,$7E
	.dc.b	$6D,$F2,$9B,$F5,$71,$F0,$F9,$F9,$5D,$7C,$9E,$FE,$3B,$87,$C5,$F1
	.dc.b	$F7,$F9,$7B,$7C,$FD,$BE,$9D,$9D,$7B,$3B,$76,$77,$EC,$F0,$76,$78
	.dc.b	$7B,$3C,$5D,$9E,$3E,$CF,$27,$67,$97,$B3,$CD,$D9,$E7,$EC,$FE,$F9
	.dc.b	$C9,$EC,$ED,$FF,$04,$3F,$EC,$4E,$DF,$F1,$AB,$FE,$7F,$2F,$C5,$F0
	.dc.b	$F8,$8E,$3F,$01,$CB,$DD,$E7,$EE,$F4,$F7,$7A,$FB,$BD,$BD,$DE,$FE
	.dc.b	$EF,$83,$DD,$F0,$FB,$BE,$2F,$77,$C7,$EE,$F9,$3D,$DF,$2F,$BB,$F4
	.dc.b	$78,$3C,$FE,$EF,$A3,$DD,$FF,$0A,$7C,$E7,$F6,$D1,$F1,$38,$F8,$9C
	.dc.b	$FC,$4E,$BE,$27,$7F,$13,$C3,$E2,$78,$FC,$37,$2F,$84,$E7,$F0,$9D
	.dc.b	$3E,$13,$AF,$C2,$76,$F8,$4E,$FF,$09,$E0,$F8,$4F,$0F,$C2,$78,$BE
	.dc.b	$13,$C7,$F0,$9E,$4F,$84,$F2,$FC,$27,$9B,$E1,$3C,$FF,$03,$FE,$2B
	.dc.b	$F9,$DF,$C3,$F9,$9F,$E3,$75,$EF,$E7,$CC,$D7,$BF,$BF,$7F,$87,$BF
	.dc.b	$C7,$DF,$E5,$EF,$F3,$F7,$FF,$2D,$E9,$75,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$47,$BF,$FE,$43,$F9,$EF,$C4,$74,$BE,$03,$F9,$0A,$87
	.dc.b	$C2,$6F,$E0,$B8,$7C,$9F,$1F,$5B,$97,$2F,$3F,$83,$E9,$F0,$7D,$7E
	.dc.b	$0F,$B7,$C1,$F7,$F8,$3F,$07,$C1,$F8,$7E,$0F,$C5,$F0,$7E,$3F,$83
	.dc.b	$F2,$7C,$1F,$97,$E0,$FC,$DF,$07,$E7,$F8,$3F,$DD,$6B,$FB,$BD,$7F
	.dc.b	$79,$AF,$EF,$75,$FD,$F6,$BE,$DF,$83,$F7,$7C,$1F,$F9,$87,$E0,$3F
	.dc.b	$C7,$47,$E3,$BA,$3F,$CD,$CB,$E0,$74,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$FF,$0D,$E8,$ED,$F0,$3D,$FE,$07,$C1
	.dc.b	$F0,$3F,$C4,$7A,$FF,$89,$75,$FF,$13,$EB,$FE,$29,$D7,$FC,$57,$AF
	.dc.b	$F8,$B7,$5F,$F1,$7E,$6F,$E3,$1C,$DF,$E7,$9F,$C5,$7F,$CA,$47,$A7
	.dc.b	$9C,$FA,$5F,$F4,$43,$FE,$E4,$3F,$D3,$C7,$F8,$38,$FE,$8C,$7F,$45
	.dc.b	$78,$F0,$F5,$E1,$ED,$C1,$DF,$83,$E4,$7C,$17,$C9,$F8,$2F,$95,$F0
	.dc.b	$5F,$2F,$E0,$BE,$67,$C1,$7C,$DF,$82,$F9,$DF,$05,$F3,$FE,$0B,$E8
	.dc.b	$7C,$17,$FA,$47,$E8,$3F,$4F,$C0,$7F,$4E,$FF,$A7,$D7,$C0,$EF,$E0
	.dc.b	$78,$7C,$0F,$1F,$81,$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF
	.dc.b	$03,$DF,$E0,$7C,$1F,$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0
	.dc.b	$3F,$37,$C9,$F9,$DD,$9F,$3F,$B3,$E8,$76,$7D,$1E,$CF,$A5,$D9,$F4
	.dc.b	$FB,$3E,$A7,$67,$D5,$EC,$FA,$DD,$9F,$EA,$FF,$4F,$DE,$7B,$BB,$4F
	.dc.b	$B9,$FF,$5F,$2E,$BE,$1D,$7C,$7A,$F9,$75,$F3,$EB,$E9,$D7,$D7,$AF
	.dc.b	$E6,$7C,$0F,$CD,$F8,$1F,$9D,$F0,$3E,$1E,$EF,$A1,$F0,$3F,$EC,$1F
	.dc.b	$A1,$FD,$F9,$FB,$CE,$0F,$F6,$72,$F7,$F5,$EF,$EF,$DF,$E1,$EF,$F1
	.dc.b	$F7,$F9,$7B,$FC,$FD,$FE,$9C,$FD,$7D,$FF,$81,$E4,$7C,$1F,$23,$E1
	.dc.b	$79,$1F,$0F,$C8,$F8,$9E,$47,$8F,$DF,$F2,$7B,$3E,$5F,$67,$CD,$EC
	.dc.b	$F9,$FD,$9F,$47,$B3,$E9,$F6,$7D,$5E,$CF,$AF,$D9,$F6,$7B,$3F,$ED
	.dc.b	$DF,$A2,$FF,$5B,$1F,$A5,$3F,$49,$D8,$5E,$FE,$FD,$FE,$1E,$FF,$1F
	.dc.b	$7F,$97,$BF,$F6,$7D,$8E,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B
	.dc.b	$9E,$1F,$73,$C5,$EE,$78,$FD,$CF,$27,$B9,$E5,$F7,$3C,$DE,$E7,$9F
	.dc.b	$DC,$F4,$7B,$9F,$87,$F5,$3D,$5E,$E7,$FB,$DF,$CF,$FD,$39,$F8,$93
	.dc.b	$F1,$1D,$27,$F1,$08,$C7,$B7,$C7,$DB,$E5,$ED,$F3,$F6,$FA,$7B,$7D
	.dc.b	$7D,$BE,$DE,$CF,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC
	.dc.b	$9D,$FE,$5E,$2F,$37,$B7,$E7,$F6,$FD,$1E,$DF,$8F,$8D,$F4,$7F,$D3
	.dc.b	$8F,$C8,$38,$F6,$F5,$1E,$B8,$F5,$3F,$EB,$D3,$FC,$EA,$7F,$54,$3F
	.dc.b	$CE,$6E,$BD,$7E,$3E,$BF,$2F,$5F,$F6,$19,$74,$F5,$FA,$FA,$FD,$BD
	.dc.b	$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F
	.dc.b	$AF,$EB,$FA,$4F,$67,$09,$FE,$64,$7E,$38,$FC,$6B,$F4,$D7,$AD,$C7
	.dc.b	$D6,$E5,$EB,$7F,$05,$E8,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07
	.dc.b	$AD,$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$79
	.dc.b	$FD,$6F,$47,$AD,$E9,$F5,$BD,$5E,$B7,$D5,$FA,$5F,$E6,$67,$F1,$53
	.dc.b	$F1,$A7,$E3,$39,$CB,$D6,$E1,$EB,$71,$F5,$BF,$8F,$79,$7C,$FD,$6E
	.dc.b	$9E,$B7,$E0,$7F,$16,$FF,$06,$9F,$80,$3F,$87,$65,$F9,$77,$F0,$09
	.dc.b	$FF,$7E,$1F,$EB,$43,$FC,$26,$7E,$01,$D7,$99,$D7,$CC,$ED,$E6,$77
	.dc.b	$F2,$FC,$1E,$5F,$87,$CB,$F1,$79,$7E,$3F,$2F,$C9,$E5,$F9,$7C,$BF
	.dc.b	$85,$EB,$FC,$3F,$5F,$E2,$7A,$FF,$17,$D7,$F8,$DE,$BF,$C7,$F5,$FE
	.dc.b	$47,$AF,$F2,$7D,$7F,$95,$EB,$FC,$BF,$57,$E6,$79,$DF,$37,$AF,$F4
	.dc.b	$5F,$01,$FE,$2E,$3F,$42,$7C,$2E,$9F,$E8,$1E,$CE,$BD,$FD,$FB,$FC
	.dc.b	$3D,$FE,$3E,$FF,$2E,$EE,$7D,$DD,$3B,$BA,$F7,$76,$EE,$EF,$DD,$E0
	.dc.b	$EE,$F0,$F7,$78,$BB,$BC,$7D,$DE,$4F,$6F,$CB,$ED,$F9,$BD,$BF,$DC
	.dc.b	$7D,$3F,$E7,$8F,$F0,$71,$FC,$A8,$FC,$EF,$29,$FC,$EB,$FC,$45,$7B
	.dc.b	$5C,$BD,$AE,$7E,$D7,$4F,$6B,$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF
	.dc.b	$0F,$B5,$E2,$F6,$BC,$7E,$D7,$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED
	.dc.b	$7A,$3D,$AF,$4F,$B5,$FC,$47,$EA,$3F,$C1,$E7,$F7,$13,$F9,$49,$F9
	.dc.b	$D3,$F3,$9C,$BF,$CA,$57,$B3,$C7,$D9,$E5,$EC,$F3,$F6,$7A,$7B,$1D
	.dc.b	$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C
	.dc.b	$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$F9,$47,$D4,$FF,$9D,$0F,$E4
	.dc.b	$E7,$E1,$CF,$D3,$8F,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB
	.dc.b	$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F
	.dc.b	$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6,$F5,$FF,$9F,$7B,$BF,$C9,$0F,$F2
	.dc.b	$A1,$FD,$C0,$FC,$39,$F8,$6F,$1B,$F9,$1F,$AF,$C3,$D7,$E3,$EB,$F2
	.dc.b	$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5,$FC,$3E
	.dc.b	$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F,$AF,$FD
	.dc.b	$57,$EA,$BF,$CA,$67,$F8,$D4,$FE,$DA,$7F,$21,$3F,$0B,$C8,$7F,$0A
	.dc.b	$FF,$6F,$5E,$B7,$3F,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E
	.dc.b	$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7
	.dc.b	$F5,$BD,$1E,$B7,$F6,$DF,$AB,$FF,$5C,$9F,$DB,$0F,$E1,$87,$F2,$C3
	.dc.b	$F0,$DC,$7F,$E0,$55,$EA,$F1,$F5,$79,$7A,$BC,$FD,$5E,$9E,$AF,$5F
	.dc.b	$57,$B7,$AB,$DF,$D5,$F0,$7A,$BE,$1F,$57,$C5,$EA,$F8,$FD,$5F,$27
	.dc.b	$AB,$E5,$F5,$7C,$DE,$AF,$9F,$D5,$FF,$02,$7D,$67,$F6,$A3,$FD,$34
	.dc.b	$7F,$8D,$8F,$E1,$A7,$E1,$CF,$E1,$7C,$9F,$E1,$D5,$EA,$72,$F5,$39
	.dc.b	$FA,$9D,$3D,$4E,$BE,$A7,$6F,$53,$BF,$A9,$E0,$F5,$3C,$3E,$A7,$8B
	.dc.b	$D4,$F1,$FA,$9E,$4F,$53,$CB,$EA,$79,$BD,$4F,$3F,$A9,$E8,$F5,$3F
	.dc.b	$C3,$9E,$BF,$FA,$68,$FF,$3E,$1F,$E3,$43,$FB,$41,$F8,$83,$F0,$FC
	.dc.b	$9F,$B7,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7,$4F,$53,$AF,$A9,$DB,$D4
	.dc.b	$EF,$EA,$78,$3D,$4F,$0F,$A9,$E2,$F5,$3C,$7E,$A7,$93,$D4,$F2,$FA
	.dc.b	$9E,$6F,$53,$FC,$65,$F5,$BF,$EF,$E3,$FC,$62,$7F,$51,$3F,$B8,$9F
	.dc.b	$86,$E2,$3F,$86,$7F,$C9,$8B,$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD
	.dc.b	$3E,$FE,$9F,$83,$D3,$F0,$FA,$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F
	.dc.b	$A7,$E6,$F4,$FC,$FE,$9F,$A3,$D3,$FF,$25,$7D,$77,$F8,$C0,$FF,$70
	.dc.b	$1F,$E7,$C3,$FA,$99,$FC,$34,$FC,$3F,$19,$FC,$3B,$FE,$65,$7F,$DC
	.dc.b	$03,$D1,$EB,$E8,$F6,$F4,$7B,$FA,$3E,$0F,$47,$C3,$E8,$F8,$BD,$1F
	.dc.b	$1F,$A3,$E4,$F4,$7C,$BE,$8F,$9B,$D1,$F3,$FA,$3E,$8F,$47,$D3,$E8
	.dc.b	$FF,$98,$FE,$BF,$FD,$C2,$7F,$B0,$CF,$F3,$D9,$FE,$2F,$3F,$87,$9F
	.dc.b	$88,$3F,$86,$F1,$7F,$9F,$57,$A1,$CF,$D0,$E9,$E8,$75,$F4,$3B,$7A
	.dc.b	$1D,$FD,$0F,$07,$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F
	.dc.b	$43,$CD,$E8,$79,$FD,$0F,$F3,$D7,$C1,$3E,$5E,$FC,$7E,$3E,$23,$3F
	.dc.b	$07,$AF,$5F,$7E,$DF,$0F,$5B,$8F,$AD,$CB,$D6,$E7,$EB,$74,$F5,$BA
	.dc.b	$FA,$DD,$BD,$6E,$FE,$B7,$83,$D6,$F0,$FA,$DE,$2F,$5B,$C7,$EB,$79
	.dc.b	$3D,$6F,$2F,$AD,$E6,$F5,$BF,$D2,$1F,$61,$FE,$76,$1F,$2B,$8F,$91
	.dc.b	$CF,$C6,$6B,$E0,$F7,$F0,$7C,$3E,$03,$8F,$C0,$72,$F7,$79,$FB,$BD
	.dc.b	$3D,$DE,$BE,$EF,$6F,$77,$BF,$BB,$E0,$F7,$7C,$3E,$EF,$8B,$DD,$F1
	.dc.b	$FB,$BE,$4F,$77,$CB,$EE,$F9,$BD,$DF,$3F,$BB,$FE,$A7,$FB,$17,$CC
	.dc.b	$5F,$96,$C7,$C9,$67,$E3,$35,$E9,$6F,$E2,$B8,$7C,$37,$1F,$82,$E5
	.dc.b	$F0,$5C,$FE,$0B,$A7,$C1,$75,$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05
	.dc.b	$E1,$F8,$2F,$17,$C1,$78,$FE,$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17
	.dc.b	$FA,$F7,$EC,$BF,$09,$F0,$BF,$ED,$17,$FD,$7A,$7F,$97,$4F,$F3,$E9
	.dc.b	$FC,$34,$FE,$18,$9F,$E1,$8F,$1F,$77,$A7,$BB,$D7,$DD,$ED,$EE,$F7
	.dc.b	$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE,$3F,$77,$C9,$EE,$F9,$7D,$DF
	.dc.b	$37,$BB,$E7,$F7,$7D,$1E,$EF,$FB,$57,$D8,$FF,$5D,$9F,$86,$E1,$FF
	.dc.b	$75,$2F,$6F,$5E,$DE,$FD,$BE,$13,$C7,$DB,$E5,$ED,$F3,$F6,$FA,$7B
	.dc.b	$7D,$7D,$BE,$DE,$DF,$7F,$6F,$C1,$ED,$F8,$7D,$BF,$17,$B7,$E3,$F6
	.dc.b	$FC,$9E,$DF,$97,$DB,$F3,$7B,$7E,$7F,$6F,$D1,$ED,$FF,$BB,$7E,$CF
	.dc.b	$F0,$FC,$FF,$F0,$05,$7B,$59,$F6,$B5,$ED,$6F,$DA,$E1,$ED,$71,$F6
	.dc.b	$B9,$7B,$5C,$FD,$AE,$9E,$D7,$5F,$6B,$B7,$B5,$DF,$DA,$F0,$7B,$5E
	.dc.b	$1F,$6B,$C5,$ED,$78,$FD,$AF,$27,$B5,$E5,$F6,$BC,$DE,$D7,$9F,$DA
	.dc.b	$73,$CF,$B4,$FC,$4A,$5F,$0B,$A8,$C7,$C2,$D9,$FF,$63,$9F,$D5,$CF
	.dc.b	$C5,$9F,$8A,$8E,$5E,$E6,$FD,$CE,$1E,$E7,$1F,$73,$97,$B9,$CF,$DC
	.dc.b	$E9,$EE,$75,$F7,$3B,$7B,$9D,$FD,$CF,$07,$B9,$E1,$F7,$3C,$5E,$E7
	.dc.b	$8F,$DC,$F3,$7D,$AF,$9F,$84,$FE,$2D,$BE,$EE,$BD,$DD,$FB,$BC,$3D
	.dc.b	$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B,$FB,$BE
	.dc.b	$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF,$9B,$DD
	.dc.b	$F3,$FB,$BE,$8F,$77,$E7,$7C,$27,$E2,$B9,$FF,$C8,$BE,$EE,$3D,$DC
	.dc.b	$FB,$BA,$F7,$77,$EE,$F0,$F7,$78,$FB,$BC,$BD,$DE,$7E,$EF,$4F,$77
	.dc.b	$AF,$BB,$DB,$DD,$EF,$EE,$F8,$3D,$DF,$0F,$BB,$E2,$F7,$7C,$7E,$EF
	.dc.b	$93,$DD,$F2,$FB,$BE,$6F,$77,$F0,$9F,$6D,$F8,$83,$F0,$FC,$FF,$8C
	.dc.b	$7F,$DD,$83,$E1,$F7,$F0,$FC,$3A,$B8,$FC,$3F,$2F,$7B,$9F,$BD,$D3
	.dc.b	$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$7B,$C3,$EF,$78,$BD,$EF,$1F
	.dc.b	$BD,$E4,$F7,$BC,$BE,$F7,$9B,$DE,$F3,$FB,$DF,$9B,$FB,$7F,$C3,$1F
	.dc.b	$85,$3F,$87,$73,$7E,$91,$FF,$76,$9F,$EA,$73,$FC,$36,$7F,$86,$92
	.dc.b	$F6,$7A,$7B,$3D,$7D,$9E,$DE,$CF,$7F,$67,$C1,$EC,$F8,$7D,$9F,$17
	.dc.b	$B3,$E3,$F6,$7C,$9E,$CF,$97,$D9,$F3,$7B,$3E,$7F,$67,$D1,$EC,$FA
	.dc.b	$7D,$9F,$DA,$7D,$C7,$E1,$CF,$E1,$FC,$7F,$BC,$5E,$CE,$BD,$9D,$FB
	.dc.b	$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B
	.dc.b	$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF
	.dc.b	$9B,$D9,$F3,$FB,$3F,$C2,$F8,$7F,$87,$9F,$88,$E6,$FF,$7F,$FC,$16
	.dc.b	$7D,$9D,$7B,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3
	.dc.b	$D7,$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67
	.dc.b	$C9,$EC,$F9,$7D,$9F,$37,$B3,$FC,$87,$EE,$7F,$86,$1F,$C2,$CF,$C3
	.dc.b	$F3,$17,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE,$FF,$39,$FB,$AF
	.dc.b	$E1,$87,$E1,$B9,$CF,$E1,$9F,$E8,$6B,$DC,$E1,$EE,$71,$F3,$79,$7B
	.dc.b	$9C,$FD,$CE,$9E,$E7,$5F,$73,$B7,$B9,$DF,$DC,$F0,$7B,$9E,$1F,$73
	.dc.b	$C5,$EE,$78,$FD,$CF,$27,$B9,$E5,$F7,$3C,$DE,$E7,$9F,$DC,$F4,$7B
	.dc.b	$9F,$D4,$7E,$EF,$FA,$A9,$F8,$71,$EE,$E7,$DD,$D7,$BB,$BF,$77,$87
	.dc.b	$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B,$BD,$7D,$DE,$DE,$EF,$7F,$77
	.dc.b	$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7,$7C,$9E,$EF,$97,$DD,$F3,$7B
	.dc.b	$BF,$D9,$FD,$9F,$EA,$47,$F5,$13,$F0,$E7,$E1,$BC,$AF,$EE,$2B,$DB
	.dc.b	$E1,$ED,$F1,$8E,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$37
	.dc.b	$B7,$E7,$F6,$FD,$1E,$DF,$F7,$FF,$BC,$FE,$A4,$7E,$1C,$FC,$37,$29
	.dc.b	$FC,$33,$FE,$0A,$5E,$D7,$1F,$6B,$97,$B5,$CF,$DA,$E9,$ED,$75,$F6
	.dc.b	$BB,$7B,$5D,$FD,$AF,$07,$B5,$E1,$F6,$BC,$5E,$D7,$8F,$DA,$F2,$7B
	.dc.b	$5E,$5F,$6B,$CD,$ED,$79,$FD,$AF,$47,$B5,$FE,$17,$FB,$DF,$F0,$D9
	.dc.b	$FC,$34,$FC,$3F,$2F,$F8,$81,$FE,$A8,$DF,$63,$87,$B1,$C7,$D8,$E5
	.dc.b	$EC,$73,$F6,$3A,$7B,$0F,$E0,$18,$5C,$7A,$43,$EF,$BF,$C5,$BF,$04
	.dc.b	$E7,$E3,$18,$6D,$7F,$36,$6C,$EE,$39,$CE,$DA,$6C,$B1,$B5,$AA,$FC
	.dc.b	$51,$B3,$2E,$27,$84,$EE,$BB,$56,$9B,$19,$8E,$4D,$9E,$35,$CE,$FD
	.dc.b	$24,$AF,$2B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD,$F8,$A7,$73,$9A,$E9
	.dc.b	$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57,$72,$47,$9D,$B7,$E4
	.dc.b	$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C,$73,$B6,$9B,$2C,$6D
	.dc.b	$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E,$EB,$C1,$5A,$6C,$66
	.dc.b	$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C,$A3,$83,$A2,$1A,$9E
	.dc.b	$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F,$46,$BC,$4C,$46,$C9
	.dc.b	$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C,$F0,$6C,$ED,$B3,$B8
	.dc.b	$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD,$71,$C4,$D3,$89,$E5
	.dc.b	$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF,$D6,$4A,$F3,$B5,$E8
	.dc.b	$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE,$97,$E3,$6C,$EA,$31
	.dc.b	$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E,$48,$C1,$1C,$12,$FC
	.dc.b	$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6,$DF,$51,$2F,$DA,$4B
	.dc.b	$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6,$63,$93,$67,$8D,$73
	.dc.b	$BF,$59,$2B,$CE,$D7,$A1,$CA,$38,$3A,$21,$A9,$EB,$7E,$29,$DC,$E6
	.dc.b	$BA,$5F,$8D,$B3,$A8,$C4,$F4,$6B,$C4,$C4,$6C,$95,$DC,$91,$E7,$6D
	.dc.b	$F9,$23,$04,$70,$4B,$F1,$CF,$06,$CE,$DB,$3B,$8F,$1C,$ED,$A6,$CB
	.dc.b	$1B,$7D,$44,$BF,$69,$2F,$D7,$1C,$4D,$38,$9E,$53,$BA,$F0,$56,$9B
	.dc.b	$19,$8E,$4D,$9E,$35,$CE,$FD,$64,$AF,$3B,$5E,$87,$28,$E0,$E8,$86
	.dc.b	$A7,$AD,$F8,$A7,$73,$9A,$E9,$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11
	.dc.b	$B2,$57,$72,$47,$9D,$B7,$E4,$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C
	.dc.b	$EE,$3C,$73,$B6,$9B,$2C,$6D,$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2
	.dc.b	$79,$4E,$EB,$C1,$5A,$6C,$66,$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED
	.dc.b	$7A,$1C,$A3,$83,$A2,$1A,$9E,$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A
	.dc.b	$8C,$4F,$46,$BC,$4C,$46,$C9,$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04
	.dc.b	$BF,$1C,$F0,$6C,$ED,$B3,$B8,$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6
	.dc.b	$92,$FD,$71,$C4,$D3,$89,$E5,$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3
	.dc.b	$5C,$EF,$D6,$4A,$F3,$B5,$E8,$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77
	.dc.b	$39,$AE,$97,$E3,$6C,$EA,$31,$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79
	.dc.b	$DB,$7E,$48,$C1,$1C,$12,$FC,$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69
	.dc.b	$B2,$C6,$DF,$51,$2F,$DA,$4B,$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15
	.dc.b	$A6,$C6,$63,$93,$67,$8D,$73,$BF,$59,$2B,$CE,$D7,$A1,$CA,$38,$3A
	.dc.b	$21,$A9,$EB,$7E,$29,$DC,$E6,$BA,$5F,$8D,$B3,$A8,$C4,$F4,$6B,$C4
	.dc.b	$C4,$6C,$95,$DC,$91,$E7,$6D,$F9,$23,$04,$70,$4B,$F1,$CF,$06,$CE
	.dc.b	$DB,$3B,$8F,$1C,$ED,$A6,$CB,$1B,$7D,$44,$BF,$69,$2F,$D7,$1C,$4D
	.dc.b	$38,$9E,$53,$BA,$F0,$56,$9B,$19,$8E,$4D,$9E,$35,$CE,$FD,$64,$AF
	.dc.b	$3B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD,$F8,$A7,$73,$9A,$E9,$7E,$36
	.dc.b	$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57,$72,$47,$9D,$B7,$E4,$8C,$11
	.dc.b	$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C,$73,$B6,$9B,$2C,$6D,$F5,$12
	.dc.b	$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E,$EB,$C1,$5A,$6C,$66,$39,$36
	.dc.b	$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C,$A3,$83,$A2,$1A,$9E,$B7,$E2
	.dc.b	$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F,$46,$BC,$4C,$46,$C9,$5D,$C9
	.dc.b	$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C,$F0,$6C,$ED,$B3,$B8,$F1,$CE
	.dc.b	$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD,$71,$C4,$D3,$89,$E5,$3B,$AF
	.dc.b	$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF,$D6,$4A,$F3,$B5,$E8,$72,$8E
	.dc.b	$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE,$97,$E3,$6C,$EA,$31,$3D,$1A
	.dc.b	$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E,$48,$C1,$1C,$12,$FC,$73,$C1
	.dc.b	$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6,$DF,$51,$2F,$DA,$4B,$F5,$C7
	.dc.b	$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6,$63,$93,$67,$8D,$73,$BF,$59
	.dc.b	$2B,$CE,$D7,$A1,$CA,$38,$3A,$21,$A9,$EB,$7E,$29,$DC,$E6,$BA,$5F
	.dc.b	$8D,$B3,$A8,$C4,$F4,$6B,$C4,$C4,$6C,$95,$DC,$91,$E7,$6D,$F9,$23
	.dc.b	$04,$70,$4B,$F1,$CF,$06,$CE,$DB,$3B,$8F,$1C,$ED,$A6,$CB,$1B,$7D
	.dc.b	$44,$BF,$69,$2F,$D7,$1C,$4D,$38,$9E,$53,$BA,$F0,$56,$9B,$19,$8E
	.dc.b	$4D,$9E,$35,$CE,$FD,$64,$AF,$3B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD
	.dc.b	$F8,$A7,$73,$9A,$E9,$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57
	.dc.b	$72,$47,$9D,$B7,$E4,$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C
	.dc.b	$73,$B6,$9B,$2C,$6D,$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E
	.dc.b	$EB,$C1,$5A,$6C,$66,$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C
	.dc.b	$A3,$83,$A2,$1A,$9E,$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F
	.dc.b	$46,$BC,$4C,$46,$C9,$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C
	.dc.b	$F0,$6C,$ED,$B3,$B8,$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD
	.dc.b	$71,$C4,$D3,$89,$E5,$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF
	.dc.b	$D6,$4A,$F3,$B5,$E8,$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE
	.dc.b	$97,$E3,$6C,$EA,$31,$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E
	.dc.b	$48,$C1,$1C,$12,$FC,$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6
	.dc.b	$DF,$51,$2F,$DA,$4B,$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6
	.dc.b	$63,$93,$67,$8D,$73,$BF,$59,$28,$29,$2B,$CD,$AF,$77,$BA,$D2,$9D
	.dc.b	$3D,$3E,$6D,$EA,$70,$FB,$E7,$C7,$F2,$97,$AD,$98,$C3,$6F,$95,$D5
	.dc.b	$4F,$5B,$1F,$8C,$79,$25,$8F,$67,$DB,$19,$8C,$47,$29,$E3,$F2,$52
	.dc.b	$3E,$1C,$EA,$93,$DB,$E3,$3B,$FC,$3F,$83,$DA,$98,$70,$D7,$E2,$E0
	.dc.b	$AF,$46,$7E,$7A,$D1,$E4,$3D,$26,$B3,$F5,$FB,$52,$36,$2D,$1F,$03
	.dc.b	$DA,$F8,$3E,$D7,$C2,$F6,$A5,$F0,$B4,$F5,$8C,$1E,$03,$AE,$3E,$57
	.dc.b	$E4,$E7,$EA,$F6,$BE,$69,$23,$A7,$B5,$F4,$3D,$AF,$A3,$ED,$49,$F0
	.dc.b	$47,$77,$99,$EF,$3C,$27,$83,$6F,$B3,$ED,$48,$F4,$3F,$73,$DA,$FB
	.dc.b	$BE,$D4,$9F,$63,$F8,$0F,$6B,$B7,$3F,$D7,$9F,$61,$E0,$3E,$F1,$F0
	.dc.b	$FD,$AF,$C3,$FB,$52,$30,$7C,$7F,$6B,$F1,$64,$7E,$57,$B5,$DB,$87
	.dc.b	$EE,$C7,$6C,$3F,$17,$C5,$F2,$5F,$C4,$7B,$5F,$94,$24,$FD,$AF,$C6
	.dc.b	$7F,$23,$3F,$92,$5F,$1B,$F9,$D7,$F5,$53,$BF,$07,$13,$DF,$1F,$0C
	.dc.b	$FC,$A3,$DD,$8F,$86,$3F,$6B,$6F,$88,$E9,$EE,$4B,$F1,$47,$7F,$72
	.dc.b	$47,$95,$F1,$7B,$9D,$E3,$F8,$2D,$BD,$CF,$31,$98,$43,$DC,$F5,$7B
	.dc.b	$92,$7E,$04,$7B,$7C,$A9,$1F,$0D,$F8,$1E,$E7,$48,$F8,$DE,$E7,$C3
	.dc.b	$30,$FD,$BC,$1E,$E7,$C7,$F7,$24,$FC,$D8,$F9,$5E,$E7,$CB,$F7,$24
	.dc.b	$FD,$18,$F9,$DE,$E4,$8F,$A8,$FD,$1F,$73,$E9,$7B,$9F,$4F,$DC,$95
	.dc.b	$F6,$63,$F8,$97,$03,$E5,$6F,$C8,$CA,$FD,$BC,$E5,$BF,$13,$C3,$E2
	.dc.b	$64,$60,$DF,$C0,$74,$23,$CB,$E0,$3B,$FC,$07,$83,$E0,$24,$F6,$8F
	.dc.b	$1F,$C0,$79,$3D,$D9,$3E,$28,$F3,$FC,$4C,$8F,$2B,$EA,$F8,$9F,$5F
	.dc.b	$C4,$FB,$3E,$26,$57,$AA,$3D,$FF,$13,$F0,$3E,$26,$4F,$BA,$3E,$1F
	.dc.b	$C4,$FC,$4F,$89,$93,$F0,$A3,$E3,$FC,$27,$C8,$F8,$59,$3F,$1A,$3E
	.dc.b	$5F,$B1,$FC,$F7,$F1,$AF,$E6,$DE,$67,$F4,$E6,$D7,$F5,$6E,$8F,$CF
	.dc.b	$3F,$D6,$23,$F9,$AB,$FC,$99,$AF,$F0,$16,$FA,$1F,$D0,$17,$C2,$F7
	.dc.b	$F8,$49,$5C,$5F,$17,$C2,$F8,$FE,$16,$4F,$57,$CD,$F0,$BE,$7F,$83
	.dc.b	$F4,$7C,$2C,$AF,$14,$7A,$FE,$13,$D9,$F0,$92,$7C,$D1,$EF,$F8,$39
	.dc.b	$1E,$97,$E1,$7C,$27,$C3,$F8,$4F,$89,$F0,$92,$BD,$D1,$F1,$FE,$12
	.dc.b	$47,$C1,$7E,$57,$C2,$7C,$BF,$84,$F9,$9F,$09,$2B,$E3,$47,$CF,$F8
	.dc.b	$4F,$A1,$F0,$92,$7E,$54,$7D,$3F,$84,$91,$F3,$5F,$EE,$5F,$8D,$FF
	.dc.b	$1A,$EB,$FF,$01,$BF,$C6,$A7,$FA,$BB,$5F,$E6,$4D,$EA,$FE,$E2,$FF
	.dc.b	$1E,$8C,$FC,$0F,$42,$3C,$BE,$0B,$BF,$C1,$78,$3E,$0A,$4F,$68,$F1
	.dc.b	$FC,$17,$93,$E0,$A4,$F8,$A3,$CF,$E0,$C8,$F2,$BE,$AF,$82,$F5,$FC
	.dc.b	$17,$B3,$E0,$A5,$7A,$A3,$DF,$F0,$12,$3D,$AF,$C2,$F8,$1F,$87,$F0
	.dc.b	$1F,$13,$E0,$65,$7C,$28,$F8,$FF,$03,$F2,$3E,$06,$4F,$C6,$8F,$97
	.dc.b	$F0,$12,$3E,$4B,$F3,$BE,$07,$FC,$41,$F8,$E7,$F0,$B9,$FE,$B8,$BE
	.dc.b	$07,$26,$BE,$0B,$AC,$FD,$0C,$6F,$4F,$D0,$CB,$B4,$FD,$4C,$AC,$FD
	.dc.b	$4C,$B9,$4F,$F8,$43,$2E,$CD,$CB,$2F,$8B,$E0,$BC,$BF,$05,$23,$84
	.dc.b	$7A,$3E,$0B,$D3,$F0,$52,$7A,$47,$B3,$E0,$7D,$BF,$05,$27,$C1,$1F
	.dc.b	$03,$E0,$7E,$0F,$C1,$49,$F2,$47,$C4,$F8,$29,$1E,$77,$E3,$FC,$17
	.dc.b	$C8,$F8,$1F,$93,$F0,$52,$BD,$91,$CD,$F9,$78,$5E,$2F,$F2,$2F,$0B
	.dc.b	$2E,$B5,$F4,$32,$B5,$AC,$BB,$4F,$C7,$CB,$B3,$F5,$B2,$FB,$3F,$03
	.dc.b	$F6,$BE,$0A,$4F,$CD,$8F,$BB,$F0,$32,$3E,$83,$F8,$0F,$82,$FC,$0F
	.dc.b	$C0,$C9,$FA,$8F,$E1,$3E,$0B,$AC,$7D,$7C,$3A,$BF,$86,$C3,$A3,$F8
	.dc.b	$8C,$2E,$9E,$EC,$7A,$4F,$E3,$31,$B3,$F0,$B1,$E6,$FC,$4C,$6F,$99
	.dc.b	$FE,$82,$E0,$67,$D2,$31,$9D,$CB,$FD,$1B,$4C,$FA,$D7,$1C,$ED,$6B
	.dc.b	$3F,$CF,$FC,$17,$F9,$FF,$DA,$FE,$83,$26,$E3,$FD,$0A,$31,$F1,$1B
	.dc.b	$F8,$29,$1F,$E6,$8A,$FE,$83,$97,$FA,$61,$FF,$3A,$46,$7E,$07,$B1
	.dc.b	$DF,$E0,$BC,$31,$C4,$F1,$18,$3F,$C9,$93,$D1,$FF,$54,$BF,$D3,$1F
	.dc.b	$05,$FA,$70,$E9,$1D,$C8,$72,$AF,$77,$C1,$48,$F4,$3F,$07,$E0,$BE
	.dc.b	$17,$C0,$C9,$F6,$3F,$17,$E0,$BE,$37,$C1,$7C,$7C,$E5,$7C,$18,$E8
	.dc.b	$FA,$3E,$0B,$E6,$18,$E7,$FF,$51,$D7,$FA,$1A,$7A,$90,$F7,$DF,$67
	.dc.b	$E6,$47,$67,$E1,$DF,$D7,$F8,$29,$1F,$45,$FB,$5F,$03,$F6,$FE,$0B
	.dc.b	$EE,$7C,$14,$AF,$AD,$1F,$7F,$E0,$BF,$01,$F0,$32,$7E,$D4,$74,$7E
	.dc.b	$8E,$17,$4F,$4E,$3D,$67,$F0,$18,$F5,$7E,$BE,$3D,$5F,$C2,$63,$D5
	.dc.b	$FB,$78,$F3,$E5,$FC,$36,$56,$7E,$EE,$3D,$5F,$C0,$E3,$D7,$2F,$93
	.dc.b	$97,$59,$FC,$2E,$5D,$67,$E6,$E5,$F9,$BF,$82,$95,$F9,$57,$F3,$DF
	.dc.b	$03,$F9,$FF,$82,$93,$F9,$97,$F4,$5F,$05,$FE,$E1,$FC,$77,$FB,$84
	.dc.b	$FF,$BE,$1F,$ED,$9B,$FF,$BD,$1B,$19,$8E,$EF,$E6,$30,$B9,$7E,$03
	.dc.b	$1E,$F3,$F9,$CC,$6C,$FE,$13,$1F,$06,$1E,$1F,$87,$97,$FE,$D4,$CB
	.dc.b	$C6,$9F,$EF,$97,$FE,$EE,$6D,$79,$7E,$23,$D5,$F0,$9E,$BF,$84,$98
	.dc.b	$73,$7D,$DF,$07,$EF,$F8,$4F,$81,$F0,$72,$BC,$11,$F0,$F0,$91,$E3
	.dc.b	$7A,$BC,$C8,$7F,$B0,$F2,$61,$DA,$39,$9E,$BB,$EC,$EE,$FE,$87,$C2
	.dc.b	$7D,$1F,$84,$93,$F0,$63,$EA,$7C,$24,$8F,$88,$FD,$7F,$83,$FB,$1F
	.dc.b	$09,$F6,$7E,$12,$57,$C9,$8E,$6F,$97,$0B,$C3,$F2,$71,$E9,$3E,$9C
	.dc.b	$6C,$FC,$DC,$7A,$CF,$D3,$C7,$AB,$EF,$C7,$9F,$0F,$E1,$72,$B3,$F8
	.dc.b	$9C,$B9,$CF,$E2,$72,$B3,$F8,$DC,$7A,$BF,$27,$1E,$B9,$7D,$DC,$BF
	.dc.b	$29,$F0,$92,$BF,$08,$FE,$5F,$E0,$FF,$31,$F0,$9F,$99,$F8,$29,$5F
	.dc.b	$89,$8F,$CE,$7C,$1C,$8F,$C6,$3F,$9F,$F8,$3E,$91,$F6,$7E,$0F,$F4
	.dc.b	$46,$33,$FC,$75,$F4,$9F,$BD,$87,$81,$E6,$43,$F0,$58,$75,$42,$9B
	.dc.b	$E2,$1F,$8F,$7F,$8C,$67,$31,$C2,$3F,$C4,$AF,$19,$E1,$18,$9D,$CF
	.dc.b	$2F,$85,$E9,$18,$97,$C1,$E2,$F8,$FC,$25,$C3,$B3,$5F,$1C,$DA,$7C
	.dc.b	$1D,$24,$FB,$30,$ED,$F0,$9B,$F8,$3E,$1F,$09,$2B,$59,$FA,$27,$A7
	.dc.b	$C1,$F5,$F8,$49,$5A,$3E,$04,$78,$7E,$13,$C5,$F0,$92,$74,$7C,$88
	.dc.b	$F3,$7C,$1C,$8C,$EC,$8F,$57,$C2,$7A,$FE,$0F,$D9,$F0,$92,$F7,$1E
	.dc.b	$FF,$84,$91,$C9,$F8,$5F,$09,$D6,$3E,$91,$F5,$70,$C7,$17,$CF,$C7
	.dc.b	$B4,$F8,$B1,$EC,$F0,$3E,$FE,$3D,$9C,$64,$CB,$18,$7D,$9C,$BE,$97
	.dc.b	$C2,$7D,$3F,$84,$98,$7B,$A3,$EB,$7C,$27,$D9,$FC,$7F,$FC,$36,$7F
	.dc.b	$87,$32,$FF,$0D,$3F,$E0,$BA,$3F,$E0,$B7,$C8,$DC,$7F,$16,$EC,$FF
	.dc.b	$0D,$C6,$BE,$07,$B1,$DF,$E1,$3C,$1F,$05,$27,$EA,$D2,$3C,$9F,$05
	.dc.b	$E5,$F8,$2F,$37,$C0,$CB,$DC,$7A,$7E,$03,$D5,$F0,$32,$79,$C7,$B7
	.dc.b	$E0,$BD,$DF,$05,$27,$BC,$7C,$1F,$82,$91,$E2,$7E,$27,$C0,$FC,$5F
	.dc.b	$82,$93,$E6,$7E,$47,$C0,$FC,$9F,$82,$93,$EA,$7E,$67,$C1,$7C,$DF
	.dc.b	$82,$93,$EE,$7E,$87,$C1,$7D,$1F,$82,$93,$F0,$9F,$D6,$78,$5F,$D6
	.dc.b	$8F,$ED,$30,$7F,$AC,$BF,$C1,$9A,$E7,$E0,$B8,$FC,$14,$8F,$D0,$CA
	.dc.b	$3A,$FB,$FD,$BE,$0A,$56,$9F,$0F,$BF,$E2,$F8,$0F,$1F,$BF,$2B,$9C
	.dc.b	$79,$BE,$02,$47,$67,$D3,$F0,$1E,$AF,$80,$F5,$FB,$F2,$BC,$71,$EE
	.dc.b	$F8,$0F,$7F,$C0,$49,$F3,$C7,$C2,$F8,$09,$1E,$A7,$E2,$FB,$DF,$1B
	.dc.b	$DE,$F8,$FE,$F4,$AF,$7C,$7C,$AF,$80,$F9,$7F,$01,$27,$E1,$C7,$F1
	.dc.b	$9F,$C8,$3F,$95,$7B,$7F,$CB,$5F,$E4,$CD,$F8,$7D,$7C,$16,$FE,$07
	.dc.b	$87,$C1,$4C,$7F,$86,$E1,$FC,$A1,$BF,$05,$D7,$E0,$A4,$60,$DF,$C0
	.dc.b	$F8,$7E,$0A,$46,$0E,$7F,$03,$E5,$23,$DB,$E0,$BD,$1E,$0F,$A7,$E0
	.dc.b	$A4,$F8,$A3,$D9,$F0,$5E,$DF,$82,$93,$E6,$8F,$81,$F0,$32,$3D,$2F
	.dc.b	$C3,$F8,$2F,$89,$F0,$5F,$17,$E0,$A5,$7B,$A3,$E4,$7C,$0F,$C9,$F8
	.dc.b	$29,$3F,$0A,$3E,$67,$C1,$48,$F8,$AF,$F4,$7F,$C8,$7F,$D5,$27,$F0
	.dc.b	$D3,$F8,$57,$57,$F5,$D6,$CF,$F6,$27,$5F,$03,$C7,$E0,$64,$FF,$3C
	.dc.b	$9F,$E1,$FF,$05,$FD,$71,$FE,$15,$19,$F8,$1F,$01,$E1,$F8,$1F,$17
	.dc.b	$C0,$78,$FE,$06,$57,$38,$F3,$7C,$07,$9F,$E0,$64,$F7,$8F,$57,$C0
	.dc.b	$48,$F1,$3E,$DF,$81,$F7,$7C,$0F,$BF,$E0,$65,$79,$E3,$E1,$7C,$0F
	.dc.b	$C3,$F8,$19,$3E,$B8,$F8,$DF,$03,$23,$DC,$FC,$9F,$81,$F9,$5F,$03
	.dc.b	$F2,$FE,$06,$57,$C3,$8F,$9D,$F0,$32,$3E,$33,$FD,$DF,$F2,$2F,$F3
	.dc.b	$29,$FC,$30,$FE,$12,$4F,$82,$D7,$C1,$4A,$FE,$D3,$3F,$C2,$5B,$F0
	.dc.b	$5C,$FE,$0A,$46,$9E,$DF,$05,$DF,$E0,$64,$F2,$7C,$5F,$05,$E3,$F8
	.dc.b	$2F,$27,$C1,$4A,$EF,$1E,$7F,$81,$91,$E2,$7D,$5F,$05,$EB,$F8,$19
	.dc.b	$3E,$67,$DD,$F0,$5E,$FF,$82,$F8,$1F,$05,$2B,$D7,$1F,$0F,$E0,$64
	.dc.b	$7B,$9F,$8D,$F0,$5F,$1F,$E0,$7E,$47,$C1,$4A,$F8,$71,$FE,$21,$DB
	.dc.b	$F8,$4C,$7F,$8E,$1F,$E1,$2D,$F8,$2D,$7C,$16,$FE,$0A,$4F,$F8,$52
	.dc.b	$BF,$84,$3F,$DF,$D7,$C1,$75,$F8,$1E,$DF,$03,$2B,$84,$78,$7E,$07
	.dc.b	$C5,$F0,$32,$7A,$47,$97,$E0,$64,$77,$7D,$1F,$03,$E9,$F8,$0F,$57
	.dc.b	$C0,$CA,$F2,$47,$37,$C9,$F0,$1E,$F3,$1E,$B7,$F9,$2B,$67,$C0,$FC
	.dc.b	$32,$57,$B7,$E0,$7E,$37,$C0,$FC,$7F,$80,$93,$F0,$63,$A3,$F0,$F1
	.dc.b	$AF,$C3,$DF,$E7,$7C,$0C,$8F,$90,$FD,$1F,$80,$FA,$5F,$03,$F4,$FE
	.dc.b	$06,$57,$CD,$8E,$7B,$7F,$98,$3F,$23,$FF,$35,$4F,$F8,$B5,$B1,$EB
	.dc.b	$CB,$FC,$B9,$1F,$E2,$12,$7C,$46,$3E,$27,$3F,$07,$AF,$83,$98,$FF
	.dc.b	$93,$AF,$F8,$9F,$C2,$7F,$9D,$F9,$3F,$C9,$93,$93,$A9,$E4,$8F,$F3
	.dc.b	$E4,$FF,$18,$8F,$27,$C1,$F8,$8F,$1F,$C1,$F9,$3E,$0B,$CB,$F0,$72
	.dc.b	$BC,$11,$CD,$FF,$44,$DF,$B3,$0E,$B1,$91,$0F,$82,$F7,$92,$3D,$5F
	.dc.b	$07,$F0,$BE,$0F,$E1,$FC,$1C,$9F,$74,$73,$BE,$D1,$E0,$31,$F0,$5F
	.dc.b	$28,$8F,$C5,$F8,$3F,$9B,$F0,$5F,$3B,$E0,$E4,$FC,$98,$E8,$FC,$63
	.dc.b	$F0,$87,$CD,$F8,$3F,$AA,$47,$E8,$7C,$1F,$D8,$F8,$2F,$B3,$F0,$72
	.dc.b	$7E,$A4,$74,$7E,$91,$D0,$FC,$17,$C1,$FE,$00,$8F,$FA,$23,$F2,$4F
	.dc.b	$F5,$3F,$6F,$FA,$A1,$FF,$19,$9F,$E3,$02,$7C,$27,$59,$FC,$61,$F3
	.dc.b	$E3,$F9,$43,$FE,$7B,$6C,$F3,$F8,$4F,$F5,$54,$3E,$0E,$FC,$1E,$3E
	.dc.b	$0E,$61,$E3,$8E,$D8,$7E,$6E,$3B,$7C,$1F,$22,$3E,$9F,$83,$EB,$F0
	.dc.b	$7D,$BC,$09,$3E,$D8,$E7,$D9,$F8,$68,$F8,$98,$7B,$FE,$0F,$CA,$48
	.dc.b	$F8,$9F,$07,$E8,$F8,$3F,$4F,$C1,$C9,$F9,$11,$D1,$FC,$21,$D3,$0F
	.dc.b	$F6,$D7,$03,$E0,$BE,$09,$27,$E8,$7C,$1F,$C4,$F8,$3F,$8B,$F0,$72
	.dc.b	$7E,$A4,$75,$7F,$36,$7C,$FF,$82,$F9,$64,$7E,$D7,$C1,$FF,$B5,$FF
	.dc.b	$25,$FE,$6F,$07,$F9,$BB,$7E,$13,$5F,$09,$D7,$C5,$FC,$EC,$B9,$7F
	.dc.b	$76,$FF,$AD,$6F,$F9,$CF,$27,$FB,$B1,$BC,$1F,$EF,$76,$CF,$FB,$F1
	.dc.b	$FE,$7A,$FF,$B5,$1F,$F7,$3B,$7E,$0B,$CC,$79,$FE,$0B,$C5,$1F,$B9
	.dc.b	$97,$C1,$7A,$C9,$1C,$FD,$AF,$77,$C1,$7B,$FE,$0A,$4F,$78,$E6,$FD
	.dc.b	$33,$C9,$F0,$3F,$14,$8F,$97,$E0,$BE,$47,$B5,$F2,$7E,$0A,$4F,$A6
	.dc.b	$39,$BF,$D1,$8F,$67,$C1,$7C,$F2,$3E,$FF,$81,$FA,$5F,$05,$F4,$FE
	.dc.b	$0A,$4F,$C3,$8E,$7C,$7F,$CC,$E3,$E3,$7C,$17,$D9,$23,$F2,$7E,$07
	.dc.b	$EE,$7C,$12,$3F,$DF,$FE,$BF,$D8,$3E,$CF,$22,$7A,$F3,$FF,$35,$DB
	.dc.b	$85,$FF,$BB,$EF,$FD,$C0,$4F,$82,$CE,$6B,$DC,$FF,$95,$A8,$FF,$40
	.dc.b	$7C,$4B,$E0,$38,$CF,$06,$CE,$BE,$07,$C6,$FF,$82,$71,$CB,$D3,$E0
	.dc.b	$6E,$1F,$D1,$71,$F1,$4F,$69,$CF,$C0,$EB,$E0,$37,$F0,$32,$BE,$1C
	.dc.b	$76,$7F,$35,$8F,$4F,$81,$91,$F2,$1E,$FF,$03,$E0,$CF,$C3,$F0,$32
	.dc.b	$BE,$6C,$73,$F0,$BF,$47,$AF,$92,$7E,$1F,$C0,$FA,$08,$FD,$4F,$81
	.dc.b	$F5,$FC,$07,$B3,$E0,$64,$FD,$88,$E9,$AF,$C0,$F8,$19,$3F,$71,$F8
	.dc.b	$7F,$03,$F1,$3E,$03,$E2,$FC,$0C,$AF,$C0,$C7,$27,$F0,$7F,$03,$F2
	.dc.b	$88,$FE,$1F,$5F,$47,$E4,$DF,$CF,$BA,$BF,$06,$FF,$BC,$0F,$F7,$69
	.dc.b	$3E,$0B,$A5,$2B,$8F,$C1,$7D,$8A,$FE,$7B,$7C,$A3,$1F,$05,$9F,$81
	.dc.b	$D7,$C1,$4C,$7C,$57,$E0,$9E,$B8,$F9,$3E,$0B,$A1,$2B,$67,$E7,$F2
	.dc.b	$7C,$17,$87,$E0,$BC,$5D,$12,$B6,$29,$3D,$B2,$F7,$FC,$17,$A4,$95
	.dc.b	$C2,$9F,$05,$ED,$F8,$1F,$77,$C1,$4A,$D4,$73,$7C,$9F,$03,$F0,$C8
	.dc.b	$F3,$F8,$1F,$8D,$F0,$1F,$1F,$E0,$64,$F7,$8E,$4F,$9F,$E0,$7E,$61
	.dc.b	$1F,$27,$C0,$FE,$9B,$F2,$7F,$E7,$67,$F3,$A2,$7C,$0F,$27,$F9,$8C
	.dc.b	$9A,$63,$65,$F3,$F8,$5E,$9F,$01,$2F,$E4,$DF,$F3,$87,$F8,$F1,$FD
	.dc.b	$F8,$FF,$05,$98,$E7,$FD,$14,$76,$F8,$6F,$31,$E9,$9F,$C4,$72,$7F
	.dc.b	$10,$97,$C1,$7A,$FE,$0A,$57,$82,$3D,$57,$EF,$F8,$29,$3E,$48,$F8
	.dc.b	$5F,$03,$F0,$FE,$0A,$4F,$A2,$38,$BE,$AA,$E2,$F8,$8F,$E5,$18,$FC
	.dc.b	$4F,$82,$F9,$9F,$05,$27,$E0,$C7,$CB,$AE,$3A,$FB,$9F,$6C,$FC,$6F
	.dc.b	$2B,$EA,$7C,$14,$8C,$1F,$D3,$CF,$EE,$E7,$C8,$AF,$E7,$CF,$8C,$FA
	.dc.b	$D5,$DB,$7F,$8C,$F6,$8D,$64,$F8,$2F,$C1,$7C,$14,$9F,$AF,$1E,$77
	.dc.b	$E0,$10,$E3,$BF,$E2,$7E,$07,$F1,$5F,$05,$27,$EF,$C7,$17,$BE,$3E
	.dc.b	$37,$E3,$D7,$E4,$7E,$0B,$F2,$5F,$05,$27,$F0,$F1,$C7,$5F,$8D,$1E
	.dc.b	$17,$F0,$B5,$F9,$9F,$81,$FE,$7F,$F9,$47,$F8,$38,$FF,$07,$ED,$FD
	.dc.b	$5E,$FE,$47,$C2,$7F,$84,$8F,$F0,$9F,$91,$FC,$C3,$87,$E2,$CF,$56
	.dc.b	$BE,$E3,$F8,$0D,$BE,$0F,$BF,$C3,$C9,$CC,$79,$EB,$8E,$9E,$89,$E4
	.dc.b	$7D,$5B,$FE,$BC,$9F,$EB,$CF,$6F,$84,$F5,$FC,$27,$38,$FC,$6F,$0F
	.dc.b	$C9,$7F,$65,$BF,$C0,$AF,$83,$F0,$92,$79,$1F,$00,$FF,$41,$1F,$AD
	.dc.b	$CB,$CA,$47,$A9,$FA,$33,$DF,$1C,$DF,$43,$DD,$E8,$7E,$46,$3F,$C0
	.dc.b	$4F,$C9,$39,$9F,$80,$3A,$1C,$8F,$97,$F0,$9F,$60,$ED,$9F,$E6,$8F
	.dc.b	$F3,$A4,$75,$3F,$BE,$1D,$4F,$CE,$97,$83,$F8,$AE,$9F,$DF,$6F,$FC
	.dc.b	$26,$AF,$B5,$7B,$0F,$49,$FC,$44,$F8,$87,$A2,$FE,$C1,$2B,$F8,$76
	.dc.b	$5E,$67,$AD,$74,$AF,$61,$C8,$FD,$A9,$D6,$BD,$93,$F3,$E3,$D2,$7C
	.dc.b	$53,$F0,$B1,$FE,$01,$BF,$54,$7F,$8C,$DF,$53,$F2,$0F,$51,$DF,$83
	.dc.b	$F3,$8F,$FB,$57,$16,$9F,$87,$6C,$7D,$2B,$ED,$9F,$A5,$F0,$65,$ED
	.dc.b	$9F,$9E,$7A,$72,$FF,$4E,$3E,$DB,$E6,$FA,$6B,$E3,$3F,$BC,$3C,$AF
	.dc.b	$47,$D0,$7C,$67,$E1,$BF,$E9,$C3,$B1,$E0,$79,$BF,$19,$B1,$F7,$AF
	.dc.b	$D1,$5F,$4A,$BB,$BF,$92,$79,$9F,$4A,$BB,$BC,$E3,$E9,$57,$AA,$7F
	.dc.b	$85,$CF,$23,$E8,$5F,$67,$E9,$5F,$EC,$DE,$47,$E1,$EF,$F0,$AF,$13
	.dc.b	$F7,$93,$C5,$FD,$F4,$F1,$BF,$05,$78,$B3,$FA,$15,$C0,$F8,$F8,$7D
	.dc.b	$67,$F2,$EF,$E9,$1F,$A5,$8F,$DD,$3F,$9E,$46,$E3,$F8,$76,$3E,$1C
	.dc.b	$FF,$17,$1F,$E5,$83,$F5,$87,$C5,$C3,$E4,$57,$CE,$BF,$C4,$61,$E4
	.dc.b	$BF,$C8,$DF,$43,$DD,$C1,$FC,$A5,$FB,$57,$E5,$3E,$7E,$5E,$E8,$F8
	.dc.b	$C4,$3F,$8C,$9F,$1E,$BA,$CF,$E1,$E3,$F0,$D8,$F9,$0F,$C8,$C7,$57
	.dc.b	$11,$CE,$BB,$3F,$50,$EA,$72,$3E,$D9,$F3,$23,$F8,$2E,$7F,$4F,$2F
	.dc.b	$7C,$F8,$F3,$F9,$13,$98,$FE,$1D,$9F,$C3,$9C,$C7,$D4,$BF,$29,$E4
	.dc.b	$7C,$58,$79,$09,$5C,$63,$C4,$7E,$30,$E4,$7C,$A9,$E5,$3C,$70,$F2
	.dc.b	$CF,$AC,$FC,$31,$F4,$CC,$67,$F4,$DF,$B8,$73,$3F,$59,$5F,$A0,$9E
	.dc.b	$79,$7E,$1C,$FA,$04,$8F,$8B,$3E,$43,$99,$F8,$93,$F8,$03,$CD,$FE
	.dc.b	$18,$F0,$7C,$91,$CD,$F7,$1F,$C5,$CF,$EA,$6F,$E1,$A7,$F3,$51,$C8
	.dc.b	$DC,$71,$7D,$EF,$9E,$7E,$76,$9F,$C8,$27,$FC,$49,$87,$C7,$7E,$B9
	.dc.b	$F6,$CF,$9E,$7E,$C8,$F6,$47,$F8,$0B,$A7,$C3,$1E,$C3,$F8,$D1,$F3
	.dc.b	$F5,$FC,$AD,$FF,$0F,$8F,$9F,$3E,$A8,$FC,$0C,$7C,$E3,$F2,$98,$FC
	.dc.b	$E8,$F8,$98,$FA,$EF,$9D,$7B,$A7,$D6,$F3,$39,$1F,$8A,$AE,$4E,$CF
	.dc.b	$CD,$9F,$3A,$3A,$9C,$9F,$DE,$E5,$F8,$3C,$BE,$24,$7B,$4F,$D2,$E1
	.dc.b	$F2,$DE,$6F,$97,$5F,$D0,$46,$E7,$F9,$83,$F2,$B1,$F4,$BF,$2C,$FB
	.dc.b	$87,$83,$2F,$97,$7F,$CE,$34,$FD,$04,$F9,$CF,$51,$F8,$B8,$FD,$E3
	.dc.b	$F9,$73,$19,$7E,$35,$F1,$4F,$29,$F3,$47,$DA,$3F,$B0,$67,$E0,$AF
	.dc.b	$E6,$71,$F6,$75,$FD,$BB,$E6,$3F,$A2,$E9,$F1,$6B,$84,$7E,$02,$FE
	.dc.b	$23,$F9,$0B,$FC,$04,$62,$BF,$5D,$AF,$E4,$6B,$F9,$99,$FE,$13,$BF
	.dc.b	$E6,$77,$F5,$23,$E6,$BC,$8C,$3E,$E3,$E0,$4F,$E4,$DE,$47,$E6,$8F
	.dc.b	$51,$F4,$75,$FA,$B8,$7C,$D8,$FC,$F1,$F3,$4F,$A8,$FF,$60,$9F,$33
	.dc.b	$FA,$F8,$EA,$7F,$84,$2B,$C7,$A7,$DC,$25,$F6,$8F,$EA,$D3,$DA,$BE
	.dc.b	$8B,$F2,$DF,$B2,$FA,$CF,$CC,$1E,$97,$E5,$D7,$5C,$BF,$AF,$E5,$F2
	.dc.b	$AB,$F4,$07,$1D,$7F,$2F,$7E,$D8,$F8,$27,$B6,$FF,$AD,$D7,$18,$FC
	.dc.b	$81,$EC,$9E,$07,$82,$FF,$02,$FC,$38,$EF,$3F,$81,$78,$61,$F8,$89
	.dc.b	$E1,$1F,$3C,$FC,$71,$FC,$3B,$3F,$C3,$BF,$AA,$9F,$93,$7D,$9F,$D9
	.dc.b	$47,$E3,$B5,$DE,$5F,$D0,$A3,$E5,$4F,$E0,$DF,$D1,$E3,$EB,$78,$1E
	.dc.b	$43,$FA,$F6,$BF,$A1,$9E,$11,$F7,$9E,$F9,$7D,$2A,$F6,$4F,$EE,$72
	.dc.b	$FC,$9B,$FB,$0C,$FE,$93,$FD,$EB,$2F,$A3,$AF,$E4,$63,$81,$F6,$71
	.dc.b	$F6,$BF,$0C,$FC,$03,$FC,$81,$FB,$17,$EE,$CB,$CD,$1F,$A0,$3D,$B7
	.dc.b	$D9,$FD,$B9,$ED,$3E,$51,$E5,$D3,$E5,$3F,$C8,$B0,$FC,$CE,$1F,$D9
	.dc.b	$30,$ED,$3F,$28,$EC,$7C,$33,$81,$F7,$AF,$E8,$C7,$C2,$8D,$1F,$BD
	.dc.b	$38,$E3,$F8,$C9,$F3,$47,$DF,$3C,$DA,$7F,$22,$7F,$25,$8F,$E6,$A7
	.dc.b	$71,$FC,$08,$F9,$27,$EB,$4F,$61,$F2,$74,$FC,$93,$F9,$83,$6F,$E6
	.dc.b	$31,$F9,$F1,$F9,$FD,$7F,$64,$FE,$08,$F8,$07,$A4,$F5,$EB,$F8,$9A
	.dc.b	$FC,$15,$FC,$09,$FE,$D8,$7C,$0C,$BF,$02,$FF,$5A,$3F,$C8,$7A,$7E
	.dc.b	$06,$BF,$C8,$51,$F8,$13,$D2,$78,$CF,$D1,$9E,$F7,$6F,$BC,$ED,$97
	.dc.b	$DA,$7F,$82,$4F,$DA,$7F,$AE,$E7,$FB,$EC,$FF,$86,$C7,$EF,$AF,$F8
	.dc.b	$69,$F1,$9F,$D1,$9F,$1A,$37,$1F,$B6,$8F,$21,$F9,$43,$FA,$56,$5F
	.dc.b	$A2,$8F,$29,$FE,$7E,$8E,$D3,$FA,$AC,$7F,$01,$1F,$08,$FC,$A9,$DC
	.dc.b	$FE,$89,$1F,$52,$7C,$AF,$EC,$0F,$D1,$57,$77,$FA,$C4,$77,$3C,$C7
	.dc.b	$09,$FC,$A9,$F8,$17,$CB,$1E,$C3,$B1,$E5,$CF,$F0,$33,$FE,$4C,$D7
	.dc.b	$F5,$AF,$E0,$70,$F4,$BF,$EA,$0C,$BE,$D5,$F8,$8E,$91,$F6,$B0,$F1
	.dc.b	$3F,$50,$FD,$31,$F9,$ED,$7F,$4D,$7F,$E8,$83,$F0,$F3,$FD,$87,$5F
	.dc.b	$CF,$BF,$D2,$EB,$E3,$1F,$BA,$3F,$02,$7E,$7E,$FE,$33,$F5,$27,$F2
	.dc.b	$8F,$E1,$EB,$B6,$BF,$87,$D3,$F3,$57,$F3,$2F,$78,$7C,$B8,$F8,$B8
	.dc.b	$7E,$E3,$4F,$8B,$AF,$F2,$D8,$FB,$A6,$4F,$CB,$D7,$9C,$F9,$EF,$DC
	.dc.b	$3E,$25,$7C,$F7,$6F,$F0,$F7,$F8,$0E,$BF,$43,$3F,$EC,$B1,$A3,$F3
	.dc.b	$7A,$7F,$14,$7C,$27,$D5,$C7,$CB,$1F,$31,$FF,$5A,$69,$FC,$26,$7F
	.dc.b	$20,$7F,$0F,$3E,$5E,$5F,$C3,$F2,$FF,$24,$4E,$E7,$EF,$1F,$96,$8F
	.dc.b	$5B,$F4,$70,$F9,$6F,$D1,$3F,$1E,$7E,$AA,$37,$5F,$1D,$FC,$AE,$5F
	.dc.b	$06,$7F,$54,$7F,$A0,$34,$FC,$E3,$FE,$5C,$3F,$85,$DF,$67,$F6,$35
	.dc.b	$FB,$73,$F8,$8B,$F8,$B3,$78,$FC,$07,$FA,$81,$F0,$0F,$E1,$57,$F9
	.dc.b	$77,$EA,$1F,$8A,$3A,$BF,$8A,$7C,$9A,$7E,$12,$3F,$BC,$1F,$84,$8F
	.dc.b	$A0,$7D,$53,$FA,$A9,$F6,$8F,$E5,$37,$F6,$9F,$7E,$BF,$A8,$7D,$C7
	.dc.b	$F9,$40,$DE,$7F,$47,$1F,$F0,$FC,$EE,$7F,$C3,$E6,$CF,$C0,$D7,$59
	.dc.b	$F8,$97,$F9,$B9,$F8,$87,$E8,$71,$F8,$8F,$F3,$33,$EE,$9F,$83,$9F
	.dc.b	$AF,$AF,$E4,$71,$F1,$5F,$F7,$1C,$7C,$5A,$7F,$2B,$C3,$FA,$31,$E2
	.dc.b	$AD,$BF,$E4,$E3,$F1,$15,$F3,$DF,$C4,$63,$E2,$C7,$F7,$D1,$F5,$5F
	.dc.b	$D4,$4E,$EF,$F2,$8F,$F2,$9B,$EB,$5F,$EC,$6C,$3C,$7A,$FE,$8F,$0F
	.dc.b	$C6,$CF,$EC,$CF,$9E,$78,$8F,$9E,$FE,$F5,$FD,$99,$B7,$E4,$BF,$C2
	.dc.b	$4F,$92,$ED,$FC,$33,$FC,$73,$5F,$E1,$31,$B3,$F5,$73,$FB,$17,$FD
	.dc.b	$35,$7E,$8B,$FE,$A9,$1F,$86,$3F,$63,$9F,$CF,$7F,$2E,$7D,$43,$F2
	.dc.b	$E7,$A3,$1F,$C8,$BE,$83,$F2,$27,$B8,$FB,$4F,$C9,$8E,$91,$F7,$F1
	.dc.b	$FC,$5B,$F0,$CF,$C5,$BE,$DD,$3F,$5C,$FB,$75,$FE,$63,$7F,$C5,$6B
	.dc.b	$CE,$7E,$58,$F3,$CF,$47,$C2,$79,$CF,$D6,$D7,$49,$FE,$01,$3F,$3A
	.dc.b	$7F,$5A,$7C,$E8,$F6,$BF,$39,$D9,$FB,$3C,$3E,$FC,$7E,$B4,$FC,$81
	.dc.b	$E2,$8F,$CB,$3E,$DD,$3F,$2C,$FE,$40,$FA,$66,$CF,$CB,$1E,$83,$E1
	.dc.b	$C7,$8A,$FD,$11,$F8,$58,$F1,$3D,$73,$FF,$73,$BE,$73,$47,$F2,$2C
	.dc.b	$3F,$8A,$BF,$84,$3F,$8A,$9E,$23,$EA,$4F,$A1,$F1,$5F,$A0,$FD,$0B
	.dc.b	$FE,$48,$CF,$F1,$4F,$F6,$A3,$A1,$FB,$D3,$A5,$7E,$B1,$FC,$53,$FA
	.dc.b	$C7,$EA,$1F,$0C,$FF,$26,$9F,$0C,$FD,$61,$ED,$3F,$C6,$87,$B4,$FC
	.dc.b	$7B,$F6,$71,$FA,$71,$F8,$43,$E9,$9E,$C3,$E6,$9F,$9A,$3E,$69,$F8
	.dc.b	$4D,$7F,$81,$61,$FE,$D6,$D7,$F5,$73,$F8,$43,$DB,$7F,$95,$AD,$C7
	.dc.b	$F9,$F8,$FE,$11,$5F,$95,$9F,$E3,$31,$F9,$53,$C4,$7B,$0F,$9B,$97
	.dc.b	$D8,$CF,$EE,$C7,$B2,$BE,$EB,$A3,$E0,$C6,$8F,$E5,$D1,$A8,$F6,$67
	.dc.b	$F5,$F5,$FD,$7E,$1F,$44,$FD,$24,$EB,$0F,$06,$3E,$A8,$FF,$35,$EB
	.dc.b	$FA,$E7,$FB,$E9,$F9,$DA,$FE,$36,$F9,$4F,$C6,$63,$FC,$F5,$F8,$AF
	.dc.b	$E5,$EB,$A6,$5F,$E3,$6A,$F0,$9F,$81,$AF,$A2,$7E,$05,$FE,$FE,$FE
	.dc.b	$07,$4F,$E2,$BA,$7F,$72,$8D,$E9,$F9,$27,$FC,$85,$AF,$E4,$9F,$F2
	.dc.b	$66,$5F,$CD,$F4,$FE,$41,$1F,$8C,$3E,$8E,$1D,$34,$FE,$D5,$1F,$81
	.dc.b	$3F,$2F,$5F,$81,$8F,$85,$1F,$C9,$72,$FB,$6F,$E3,$1F,$99,$3F,$A7
	.dc.b	$8F,$98,$7D,$6D,$3E,$09,$FE,$22,$7F,$5E,$78,$E7,$8B,$F9,$47,$E2
	.dc.b	$65,$F6,$63,$F1,$11,$A7,$F1,$07,$63,$FB,$83,$F5,$34,$FC,$1C,$7D
	.dc.b	$43,$FA,$14,$7F,$88,$32,$F8,$6F,$CE,$8F,$86,$7E,$E5,$D3,$E6,$74
	.dc.b	$78,$72,$F9,$11,$FC,$94,$F9,$07,$AF,$1F,$90,$FE,$88,$FC,$49,$FB
	.dc.b	$39,$F3,$67,$F5,$E3,$FC,$69,$3F,$5D,$F0,$3F,$DB,$5F,$C1,$1E,$57
	.dc.b	$4F,$F0,$D3,$C0,$7E,$C4,$FE,$1A,$FE,$C4,$FE,$A0,$FC,$A7,$CA,$E8
	.dc.b	$F2,$CF,$DC,$3C,$AF,$80,$FC,$59,$FC,$5B,$1F,$F2,$44,$FF,$8B,$F1
	.dc.b	$F9,$F3,$FE,$8B,$36,$7E,$5B,$1F,$9F,$3E,$B8,$F9,$E6,$AB,$F9,$64
	.dc.b	$EA,$3F,$96,$1E,$13,$E2,$9E,$17,$4F,$E8,$0D,$5F,$F3,$C3,$EC,$BF
	.dc.b	$15,$FB,$35,$F0,$1F,$D8,$BF,$16,$3D,$9A,$FF,$06,$8D,$1F,$9B,$9F
	.dc.b	$C9,$D7,$F0,$6A,$FD,$29,$F3,$E3,$9B,$F2,$A7,$E7,$BF,$A5,$79,$BF
	.dc.b	$9D,$7C,$0F,$EE,$AB,$C0,$F3,$8F,$01,$AB,$FD,$14,$73,$34,$73,$3C
	.dc.b	$07,$A8,$FB,$C6,$8F,$D3,$D7,$A9,$FE,$53,$8F,$ED,$1F,$F3,$36,$BF
	.dc.b	$C3,$6B,$BD,$2F,$FD,$CD,$F0,$6B,$93,$97,$19,$A7,$50,$9D,$29,$63
	.dc.b	$FC,$0E,$3F,$D8,$55,$C2,$B7,$4F,$84,$C6,$1D,$23,$17,$E2,$8C,$5F
	.dc.b	$3B,$FF,$14,$7C,$27,$0A,$F2,$FC,$25,$7C,$4D,$C1,$3C,$70,$FF,$07
	.dc.b	$7C,$1E,$7E,$12,$61,$DC,$F7,$C6,$DF,$F1,$5F,$C2,$73,$F8,$49,$1B
	.dc.b	$3E,$04,$6D,$FF,$5B,$7C,$1F,$87,$E1,$24,$6F,$34,$EF,$85,$C7,$E7
	.dc.b	$F8,$4F,$47,$C2,$4C,$38,$46,$E7,$D9,$F0,$7E,$DF,$84,$95,$D6,$34
	.dc.b	$F7,$F8,$4F,$89,$E5,$7D,$6E,$5F,$AE,$DE,$3F,$A1,$3C,$2B,$EE,$47
	.dc.b	$DE,$97,$C2,$72,$F8,$E9,$5F,$06,$72,$FE,$3B,$BF,$FA,$7A,$DF,$DA
	.dc.b	$EF,$E1,$7C,$7F,$0B,$23,$07,$13,$E6,$4B,$BF,$D3,$F0,$92,$73,$1E
	.dc.b	$CF,$85,$F6,$FC,$2C,$9F,$24,$7E,$0F,$C2,$FF,$0E,$F8,$7F,$EB,$BD
	.dc.b	$58,$F3,$BF,$79,$FA,$18,$CB,$D8,$F7,$7C,$2F,$CA,$F8,$59,$18,$3D
	.dc.b	$07,$21,$3C,$9C,$37,$93,$E8,$DF,$E5,$F5,$F7,$57,$CB,$CB,$EB,$FC
	.dc.b	$2C,$8C,$1D,$AF,$EC,$FC,$2F,$DC,$24,$7D,$3C,$3E,$FF,$C2,$7E,$03
	.dc.b	$E1,$64,$FD,$78,$F8,$9C,$9F,$E0,$0C,$8F,$E1,$BE,$16,$57,$DD,$7F
	.dc.b	$13,$F0,$BF,$8A,$F8,$4F,$C5,$FC,$2C,$AF,$C1,$47,$47,$FB,$20,$C3
	.dc.b	$AB,$F8,$7F,$82,$FC,$99,$1F,$C5,$7C,$2F,$EE,$7F,$29,$7C,$35,$F8
	.dc.b	$89,$5F,$AF,$9C,$BF,$0F,$C0,$FE,$27,$1F,$8E,$8F,$A4,$DE,$3F,$E2
	.dc.b	$AD,$9D,$FC,$37,$7F,$86,$F0,$7C,$2C,$C3,$B4,$7A,$DF,$ED,$47,$AF
	.dc.b	$8B,$CD,$F0,$D2,$3C,$8F,$A7,$C0,$F5,$7C,$2F,$AF,$E1,$A5,$7A,$63
	.dc.b	$A6,$BF,$D3,$23,$A1,$ED,$F8,$6F,$84,$47,$E0,$77,$7C,$5C,$BE,$37
	.dc.b	$C3,$49,$F8,$91,$D3,$7F,$E9,$B1,$FC,$EF,$97,$E8,$BF,$13,$E1,$7E
	.dc.b	$71,$1F,$99,$F0,$DF,$47,$E1,$7E,$97,$C3,$49,$FA,$11,$D5,$FE,$A8
	.dc.b	$7A,$0F,$9F,$F0,$BF,$64,$8F,$D7,$F8,$6F,$B9,$F0,$DF,$CD,$7F,$2A
	.dc.b	$FE,$93,$3F,$A7,$6F,$0F,$D6,$9E,$D3,$FD,$73,$5F,$E9,$FA,$FA,$AB
	.dc.b	$FA,$7D,$BE,$1B,$A1,$2F,$F9,$A4,$3E,$1F,$C1,$F0,$FE,$1F,$87,$95
	.dc.b	$A8,$EF,$E7,$7F,$90,$23,$C9,$F0,$FE,$72,$3D,$3E,$1F,$D5,$F0,$DE
	.dc.b	$BF,$87,$93,$E0,$8E,$6F,$F8,$EC,$FE,$D9,$9F,$F5,$27,$CB,$F0,$DF
	.dc.b	$0C,$F8,$9F,$0F,$F1,$7E,$1B,$E3,$7C,$3C,$AF,$5C,$74,$7F,$C6,$36
	.dc.b	$F8,$6F,$98,$48,$F8,$3F,$0F,$F3,$FE,$17,$E8,$7C,$3C,$9F,$8B,$1C
	.dc.b	$DF,$F1,$88,$B7,$C2,$FD,$72,$47,$CC,$F8,$6F,$B5,$F0,$9F,$6F,$E1
	.dc.b	$64,$FD,$08,$E9,$D3,$FE,$37,$EB,$FF,$03,$E2,$F8,$5F,$C1,$12,$FF
	.dc.b	$B5,$7E,$57,$FD,$4D,$FF,$09,$3F,$D3,$1F,$E5,$8B,$E1,$7F,$BF,$36
	.dc.b	$38,$FC,$37,$83,$0F,$F5,$B6,$3F,$E1,$4A,$7C,$37,$02,$5F,$83,$E1
	.dc.b	$B9,$FB,$FD,$3E,$16,$4F,$92,$3A,$3F,$E8,$EC,$7F,$C4,$38,$3E,$1B
	.dc.b	$C6,$49,$F5,$FC,$37,$9B,$E1,$BC,$FE,$34,$9F,$7C,$74,$7F,$DE,$1A
	.dc.b	$F9,$FE,$17,$DA,$48,$F8,$BF,$0B,$F0,$3E,$13,$E0,$FC,$2C,$9F,$93
	.dc.b	$1C,$FC,$6F,$D5,$ED,$FA,$8F,$4B,$E3,$FC,$2C,$BF,$9E,$FC,$AF,$85
	.dc.b	$F9,$7F,$09,$F3,$3E,$16,$57,$D4,$8E,$8F,$ED,$38,$DF,$0B,$F4,$89
	.dc.b	$1F,$6B,$E1,$7F,$C6,$5C,$1F,$CD,$A3,$FB,$22,$F8,$4C,$9A,$F8,$5E
	.dc.b	$B3,$FE,$2B,$E5,$FD,$F6,$FF,$E4,$EB,$FE,$71,$3F,$E5,$56,$A6,$8F
	.dc.b	$84,$F0,$FC,$1F,$8B,$E1,$25,$7F,$8D,$E9,$3E,$08,$FE,$F1,$87,$81
	.dc.b	$ED,$F0,$9E,$A2,$3C,$7E,$13,$DB,$F0,$7E,$EF,$84,$93,$D6,$3A,$75
	.dc.b	$FF,$99,$30,$E9,$3E,$7F,$84,$F8,$A4,$7C,$9F,$09,$F2,$3A,$7E,$4F
	.dc.b	$C2,$49,$F4,$47,$4E,$1F,$B9,$1E,$BF,$84,$F9,$E4,$7D,$DF,$09,$F4
	.dc.b	$BE,$0F,$E9,$FC,$24,$9F,$85,$1C,$DF,$E8,$D8,$7D,$8F,$84,$91,$F1
	.dc.b	$DF,$B7,$F0,$9F,$73,$E1,$3F,$CD,$DF,$96,$7E,$3A,$BF,$D1,$0D,$7E
	.dc.b	$A4,$75,$7F,$A3,$70,$FF,$A1,$A7,$F9,$C4,$7F,$9A,$5A,$FF,$8A,$5B
	.dc.b	$1B,$F8,$4E,$FF,$07,$E0,$F8,$49,$7C,$63,$B7,$27,$CE,$C1,$F0,$9E
	.dc.b	$62,$57,$7F,$84,$F4,$FC,$1F,$AB,$E1,$24,$F8,$E3,$FD,$15,$E5,$7A
	.dc.b	$CF,$E8,$9A,$FA,$7E,$13,$E0,$92,$BD,$5F,$09,$F1,$3E,$0F,$E2,$FC
	.dc.b	$24,$9F,$74,$74,$7F,$C6,$1C,$BF,$2B,$E1,$24,$7C,$37,$E6,$FC,$27
	.dc.b	$CE,$F8,$3F,$9F,$F0,$92,$BE,$44,$73,$7E,$57,$C2,$7D,$43,$0F,$CB
	.dc.b	$F8,$4F,$AE,$47,$E8,$7C,$1F,$DA,$F8,$4F,$B7,$F0,$72,$7E,$A4,$73
	.dc.b	$7F,$D3,$DF,$96,$FF,$35,$4D,$E8,$FE,$8F,$1A,$F8,$3F,$F5,$94,$FF
	.dc.b	$39,$E4,$FF,$62,$B7,$C5,$FF,$64,$BF,$CE,$6B,$FC,$D0,$FF,$36,$7B
	.dc.b	$FC,$17,$23,$9F,$C1,$F4,$F8,$2E,$BF,$07,$2B,$CB,$1E,$0F,$83,$91
	.dc.b	$E8,$7C,$7F,$07,$E4,$F8,$39,$3E,$C7,$CF,$E1,$7A,$3D,$5F,$4F,$C1
	.dc.b	$CA,$F8,$31,$EC,$F8,$29,$1F,$11,$F7,$FC,$1F,$C0,$F8,$39,$3F,$21
	.dc.b	$F8,$7F,$07,$F1,$3E,$0B,$E2,$FC,$1C,$AF,$9B,$1F,$23,$E0,$A4,$7D
	.dc.b	$07,$E5,$FC,$17,$FB,$77,$F2,$EF,$E5,$9D,$1F,$EF,$07,$F9,$65,$7F
	.dc.b	$B4,$DA,$FF,$A4,$DD,$F4,$7F,$BE,$1F,$E5,$24,$F8,$AE,$9F,$05,$D7
	.dc.b	$E0,$65,$F3,$7C,$1F,$05,$E1,$F8,$29,$3D,$DF,$27,$C1,$79,$7E,$07
	.dc.b	$CD,$F0,$52,$BC,$91,$E9,$F8,$29,$1E,$77,$D9,$F0,$5E,$DF,$3A,$4F
	.dc.b	$AD,$F8,$1F,$05,$F0,$7E,$0B,$E1,$7C,$14,$AF,$81,$1F,$17,$E0,$7E
	.dc.b	$37,$C1,$49,$F8,$91,$F2,$7E,$0A,$47,$C7,$7E,$67,$C1,$7C,$DF,$82
	.dc.b	$4F,$FB,$F7,$D8,$FE,$38,$7F,$1D,$E7,$4E,$A7,$FD,$81,$3F,$C6,$3A
	.dc.b	$30,$FF,$94,$63,$97,$C1,$6F,$A9,$5A,$F8,$8C,$3C,$3E,$0A,$DF,$8B
	.dc.b	$04,$62,$11,$87,$CD,$F0,$79,$8F,$E4,$31,$E9,$F8,$2C,$19,$F8,$3D
	.dc.b	$7C,$16,$FE,$0E,$57,$BE,$39,$7C,$14,$8F,$84,$F5,$F8,$2E,$DF,$05
	.dc.b	$DF,$29,$5F,$1E,$3C,$5F,$05,$E3,$F8,$29,$3F,$2E,$3C,$DF,$03,$23
	.dc.b	$E7,$3E,$9F,$82,$F5,$7C,$17,$AF,$E0,$A5,$7D,$38,$F7,$7C,$17,$BF
	.dc.b	$E0,$64,$FD,$78,$F8,$5F,$05,$F1,$FF,$2F,$FE,$30,$7F,$19,$DB,$FC
	.dc.b	$A5,$5F,$7B,$A7,$E9,$BF,$80,$9F,$E2,$CD,$F8,$8E,$47,$3F,$83,$E9
	.dc.b	$F0,$3D,$7E,$0A,$57,$3E,$CF,$C2,$61,$FC,$6E,$18,$F6,$50,$F8,$1F
	.dc.b	$37,$C1,$49,$E4,$78,$E3,$D5,$F0,$5E,$BF,$82,$93,$A3,$9C,$7B,$FE
	.dc.b	$0B,$E0,$7C,$14,$9D,$62,$9F,$89,$F0,$32,$72,$FC,$7F,$82,$F9,$1F
	.dc.b	$05,$F2,$7E,$0A,$57,$28,$F9,$9F,$05,$23,$AB,$F3,$FE,$0B,$E8,$7C
	.dc.b	$0C,$9F,$0B,$F4,$FE,$0B,$EA,$7C,$17,$D5,$F8,$29,$5E,$68,$FC,$F7
	.dc.b	$E6,$1F,$C6,$23,$F6,$6D,$4D,$3F,$8F,$F2,$FC,$D7,$FC,$8A,$4F,$83
	.dc.b	$E7,$F0,$7D,$3E,$0F,$AF,$C1,$CC,$3E,$55,$FF,$91,$1B,$F0,$7E,$22
	.dc.b	$3A,$F8,$3F,$2F,$C1,$48,$E2,$FA,$3E,$0F,$D3,$F0,$7E,$AF,$83,$95
	.dc.b	$DA,$3D,$BF,$07,$23,$C2,$FC,$0F,$83,$F8,$3F,$05,$F0,$BE,$0E,$57
	.dc.b	$9A,$3E,$2F,$C1,$48,$F4,$BF,$23,$E0,$FE,$4F,$C1,$C9,$F6,$BF,$33
	.dc.b	$E0,$FE,$6F,$C1,$49,$F8,$2F,$F0,$DF,$CC,$7F,$C7,$47,$F1,$B8,$FE
	.dc.b	$46,$FF,$1B,$7F,$37,$E5,$7F,$21,$8C,$74,$FE,$75,$E1,$27,$F8,$B3
	.dc.b	$C0,$E7,$F0,$3D,$CF,$07,$C0,$F8,$7E,$03,$C5,$F0,$32,$BF,$3D,$7F
	.dc.b	$C7,$1E,$47,$5F,$80,$F4,$7C,$0C,$8D,$C7,$AF,$E0,$7D,$9F,$03,$27
	.dc.b	$9C,$7B,$FE,$06,$47,$67,$E1,$7B,$1F,$0F,$E0,$24,$F8,$9F,$8D,$F0
	.dc.b	$3F,$1F,$E0,$64,$F9,$9F,$95,$F0,$3D,$23,$CF,$AB,$1C,$71,$7C,$7C
	.dc.b	$BA,$D6,$B2,$B3,$F2,$F2,$EB,$3C,$B2,$B3,$F3,$F2,$FB,$1F,$01,$F6
	.dc.b	$7E,$06,$5F,$C8,$8F,$B9,$F0,$12,$3E,$5B,$F7,$FE,$07,$F0,$1F,$03
	.dc.b	$27,$E7,$BF,$D3,$FC,$1F,$F2,$01,$FC,$92,$0F,$F8,$E7,$2F,$E3,$8D
	.dc.b	$E9,$FE,$C8,$EC,$FE,$45,$1A,$F7,$BA,$1D,$7D,$FE,$DE,$F7,$7F,$7E
	.dc.b	$63,$FC,$77,$0F,$E4,$CD,$F7,$BC,$9E,$FC,$8C,$1B,$F7,$7D,$04,$79
	.dc.b	$7B,$FE,$BF,$7A,$47,$57,$DD,$EF,$73,$8F,$C6,$E2,$C3,$1B,$7B,$B2
	.dc.b	$E9,$5A,$CA,$CF,$C2,$CB,$AC,$F2,$CA,$CF,$C6,$CB,$E6,$7B,$B2,$BD
	.dc.b	$CF,$CF,$F7,$BE,$87,$BD,$27,$E1,$3F,$4F,$DE,$FA,$9E,$F7,$D5,$F7
	.dc.b	$A5,$7C,$78,$E8,$FA,$B0,$B9,$7D,$6C,$7A,$4F,$BB,$1B,$3F,$6B,$1E
	.dc.b	$B3,$F4,$71,$EA,$FC,$4C,$79,$F2,$FD,$EF,$77,$F0,$C6,$27,$FC,$53
	.dc.b	$C4,$CB,$AD,$7E,$11,$F5,$4F,$9C,$EC,$43,$B1,$F8,$1F,$7B,$FC,$41
	.dc.b	$F9,$97,$F4,$13,$FA,$15,$A3,$3F,$15,$2B,$FB,$25,$7F,$44,$C3,$FC
	.dc.b	$6C,$BE,$03,$9F,$C0,$74,$F8,$09,$7C,$23,$BF,$C5,$48,$E6,$FE,$8F
	.dc.b	$2E,$D1,$FC,$4B,$0E,$CF,$F9,$21,$A9,$F0,$3D,$4F,$2D,$F8,$1E,$57
	.dc.b	$EE,$F8,$19,$1E,$A7,$E0,$FC,$07,$C2,$F8,$09,$3E,$E7,$E2,$FC,$07
	.dc.b	$C6,$F8,$09,$3F,$09,$F9,$3E,$FF,$48,$F4,$7C,$07,$CC,$31,$D1,$FE
	.dc.b	$15,$D5,$8F,$6A,$F7,$3D,$9E,$84,$3E,$16,$1D,$9F,$A3,$C7,$F6,$08
	.dc.b	$FD,$4F,$80,$FB,$7E,$FF,$DC,$F8,$09,$3F,$62,$3E,$FF,$C0,$7E,$03
	.dc.b	$E0,$24,$FD,$C8,$E6,$FD,$1C,$2E,$9E,$9C,$7A,$CF,$E0,$B1,$EA,$FD
	.dc.b	$7C,$7A,$BF,$86,$C7,$AB,$F6,$F1,$E7,$CF,$F8,$8F,$80,$FC,$89,$89
	.dc.b	$FB,$D7,$D3,$2F,$91,$8F,$82,$79,$1D,$48,$60,$FC,$47,$C0,$7E,$73
	.dc.b	$E0,$24,$7E,$62,$3F,$CF,$DF,$99,$FF,$A3,$63,$FA,$FB,$6F,$FC,$94
	.dc.b	$DC,$BF,$D2,$4F,$F7,$03,$FB,$9B,$8F,$81,$E6,$74,$F8,$7E,$BF,$05
	.dc.b	$30,$FF,$25,$5A,$3C,$71,$E0,$3C,$64,$37,$F0,$FC,$DE,$37,$D9,$E7
	.dc.b	$7E,$AF,$83,$F5,$FC,$1C,$9E,$F1,$EE,$F8,$2F,$7F,$C1,$C9,$F1,$C7
	.dc.b	$C2,$F8,$29,$1E,$67,$E2,$FC,$1F,$38,$F8,$BF,$05,$F2,$0C,$70,$7C
	.dc.b	$6B,$E9,$3F,$2B,$0F,$03,$CC,$87,$CE,$C3,$AB,$F1,$3E,$0B,$E9,$91
	.dc.b	$F8,$FF,$05,$F5,$BE,$0B,$EB,$FC,$14,$9F,$97,$1F,$6B,$E0,$BE,$DF
	.dc.b	$C0,$C9,$F9,$F1,$F7,$BE,$0B,$EF,$FC,$14,$9F,$A7,$1D,$5F,$AB,$87
	.dc.b	$57,$F0,$78,$75,$7E,$CE,$1D,$5F,$C3,$E1,$F8,$AF,$82,$91,$F7,$9E
	.dc.b	$CF,$23,$A9,$0C,$1C,$8E,$A4,$30,$7E,$53,$E0,$BF,$2E,$47,$FD,$29
	.dc.b	$F9,$A7,$FA,$7F,$87,$FD,$40,$FF,$82,$4F,$F0,$34,$7F,$BB,$9A,$9D
	.dc.b	$7C,$17,$FB,$7D,$FF,$04,$F6,$7F,$BB,$DB,$3C,$7E,$0F,$5F,$07,$23
	.dc.b	$C2,$F1,$F8,$3E,$5F,$0F,$CF,$E2,$25,$79,$A3,$B6,$52,$3D,$2F,$87
	.dc.b	$0F,$17,$C2,$49,$F6,$BE,$5F,$85,$F3,$7C,$2F,$9F,$E1,$65,$7C,$28
	.dc.b	$F5,$7C,$2C,$8F,$8A,$FB,$7E,$17,$DD,$F0,$B2,$7E,$4B,$F0,$7E,$17
	.dc.b	$E1,$7C,$0C,$9F,$9A,$FC,$5F,$85,$F8,$DF,$09,$F1,$FE,$16,$57,$D2
	.dc.b	$8F,$95,$F0,$B2,$3E,$AB,$F3,$7E,$0B,$A4,$7F,$9B,$39,$FF,$BC,$3D
	.dc.b	$8F,$B7,$87,$67,$EE,$E1,$D9,$E1,$81,$FF,$6A,$69,$FE,$B7,$27,$C1
	.dc.b	$E6,$3F,$CF,$31,$FD,$98,$FE,$CA,$E3,$E1,$35,$D7,$FE,$E7,$9C,$B7
	.dc.b	$E0,$F3,$D9,$CE,$7F,$D7,$17,$D3,$E1,$32,$F8,$E3,$37,$CF,$E1,$33
	.dc.b	$29,$7C,$26,$1B,$1E,$A8,$C7,$C0,$67,$E1,$35,$F0,$32,$BD,$F1,$C7
	.dc.b	$E1,$39,$7C,$24,$9F,$87,$1D,$7E,$0E,$47,$C6,$7C,$1F,$07,$E1,$F8
	.dc.b	$09,$3F,$29,$F2,$7C,$07,$97,$E0,$BC,$DF,$07,$2B,$E7,$C7,$A7,$E0
	.dc.b	$E4,$7D,$27,$D9,$F0,$5E,$DF,$83,$F8,$3F,$9A,$FF,$49,$AF,$B0,$D7
	.dc.b	$E2,$46,$BE,$0A,$47,$DD,$E6,$74,$7D,$BB,$FF,$5F,$9F,$D4,$27,$9F
	.dc.b	$C1,$76,$23,$C6,$78,$7C,$1F,$8B,$E0,$7C,$7F,$05,$2B,$47,$92,$3C
	.dc.b	$FF,$01,$E8,$F8,$09,$3A,$19,$27,$DB,$F0,$1E,$EF,$82,$95,$BA,$4F
	.dc.b	$C2,$F8,$29,$39,$7E,$2F,$C1,$7C,$6F,$82,$F8,$FE,$FC,$AE,$51,$F2
	.dc.b	$BE,$03,$E5,$FB,$F2,$7B,$47,$CE,$F8,$0F,$9F,$F0,$32,$7C,$51,$F4
	.dc.b	$BE,$0B,$E9,$FC,$14,$9F,$34,$7D,$6F,$82,$91,$E9,$7F,$33,$F9,$B7
	.dc.b	$F2,$DE,$7F,$D6,$3F,$EB,$52,$7C,$06,$BE,$07,$7F,$03,$C3,$E0,$66
	.dc.b	$3F,$26,$FF,$95,$A7,$FD,$5E,$FF,$BB,$97,$C1,$78,$0F,$0F,$C1,$F8
	.dc.b	$BE,$0E,$57,$37,$CB,$F0,$7E,$6F,$82,$F3,$FC,$1C,$AF,$04,$7A,$BE
	.dc.b	$0F,$D7,$EA,$C9,$F2,$47,$BB,$E0,$FD,$FF,$07,$27,$D1,$1F,$0B,$E0
	.dc.b	$FE,$1F,$C0,$49,$F6,$47,$C6,$F7,$FE,$3F,$C0,$49,$F8,$11,$F2,$BE
	.dc.b	$0E,$47,$C3,$7E,$6F,$C1,$FF,$06,$FC,$DD,$F0,$77,$E2,$25,$FE,$46
	.dc.b	$72,$DF,$84,$E0,$47,$5F,$09,$CF,$E1,$3A,$7C,$24,$9E,$51,$DF,$E1
	.dc.b	$24,$75,$7C,$5F,$09,$E3,$F8,$39,$3E,$17,$CD,$F0,$9E,$7F,$83,$93
	.dc.b	$E5,$7D,$5F,$09,$EB,$F8,$49,$3E,$97,$DD,$F0,$9E,$FF,$84,$93,$ED
	.dc.b	$7E,$17,$C2,$7C,$3F,$83,$F8,$9F,$09,$2B,$E1,$47,$C7,$F8,$49,$1F
	.dc.b	$15,$F9,$5F,$09,$F2,$FE,$13,$F9,$8F,$AD,$FC,$3C,$FE,$21,$CE,$8C
	.dc.b	$FC,$26,$BE,$1A,$63,$FB,$FB,$4F,$2F,$84,$E7,$F0,$92,$B5,$1D,$BE
	.dc.b	$0F,$BF,$C2,$49,$E5,$1E,$2F,$83,$F1,$FC,$24,$9E,$D1,$E6,$F8,$2F
	.dc.b	$3F,$C2,$49,$F1,$47,$AB,$E0,$E4,$79,$5F,$6F,$C1,$FB,$B6,$F7,$FC
	.dc.b	$1C,$AF,$54,$7C,$2F,$82,$91,$ED,$7E,$2F,$C1,$FF,$56,$F1,$5E,$55
	.dc.b	$25,$7C,$3F,$82,$AB,$E1,$7F,$9A,$BF,$DC,$EB,$8B,$FC,$9D,$AE,$FE
	.dc.b	$03,$FB,$A6,$4F,$84,$E7,$7C,$A9,$7C,$23,$0D,$BE,$3F,$07,$F7,$3E
	.dc.b	$12,$5F,$D7,$7E,$FF,$C2,$7F,$70,$FC,$E3,$F8,$C4,$FF,$81,$DA,$9A
	.dc.b	$3E,$17,$87,$C2,$F1,$F8,$59,$5F,$D0,$6B,$FC,$93,$F8,$1B,$E1,$BF
	.dc.b	$67,$3F,$EA,$98,$C7,$AD,$FE,$AB,$8F,$F0,$B4,$61,$C3,$5E,$13,$C9
	.dc.b	$FF,$0C,$46,$5B,$19,$79,$B5,$46,$DC,$CF,$85,$CC,$EE,$75,$0F,$A7
	.dc.b	$7A,$0F,$A7,$FF,$16,$66,$F8,$3F,$F1,$55,$6A,$73,$18,$AE,$16,$BE
	.dc.b	$0F,$F8,$D6,$F8,$39,$AE,$17,$CA,$35,$7C,$AF,$8C,$79,$63,$15,$C2
	.dc.b	$39,$1C,$2B,$93,$B6,$DF,$82,$F9,$B6,$B9,$46,$2F,$AD,$F3,$3A,$9C
	.dc.b	$CB,$5C,$5D,$4F,$1B,$EF,$5C,$A7,$84,$E5,$E3,$19,$AE,$51,$AB,$F1
	.dc.b	$91,$EA,$D7,$46,$08,$60,$91,$D0,$93,$DC,$97,$E6,$24,$F9,$1B,$7E
	.dc.b	$72,$5F,$86,$F9,$CE,$A3,$17,$D2,$F9,$12,$F8,$BE,$6B,$E2,$E6,$B8
	.dc.b	$DF,$38,$D5,$F3,$BE,$31,$EC,$8C,$57,$08,$E4,$70,$AE,$4E,$DB,$7E
	.dc.b	$0B,$E6,$DA,$E5,$18,$BE,$B7,$CC,$EA,$73,$2D,$71,$75,$3C,$6F,$BD
	.dc.b	$72,$9E,$13,$97,$8C,$66,$B9,$46,$AF,$C6,$47,$AB,$5D,$18,$21,$82
	.dc.b	$47,$42,$4F,$72,$5F,$98,$93,$E4,$6D,$F9,$C9,$7E,$1B,$E7,$3A,$8C
	.dc.b	$5F,$4B,$E4,$4B,$E2,$F9,$AF,$8B,$9A,$E3,$7C,$E3,$57,$CE,$F8,$C7
	.dc.b	$B2,$31,$5C,$23,$91,$C2,$B9,$3B,$6D,$F8,$2F,$9B,$6B,$94,$62,$FA
	.dc.b	$DF,$33,$A9,$CC,$B5,$C5,$D4,$F1,$BE,$F5,$CA,$78,$4E,$5E,$31,$9A
	.dc.b	$E5,$1A,$BF,$19,$1E,$AD,$74,$60,$86,$09,$1D,$09,$3D,$C9,$7E,$62
	.dc.b	$4F,$91,$B7,$E7,$25,$F8,$6F,$9C,$EA,$31,$7D,$2F,$91,$2F,$8B,$E6
	.dc.b	$BE,$2E,$6B,$8D,$F3,$8D,$5F,$3B,$E3,$1E,$C8,$C5,$70,$8E,$47,$0A
	.dc.b	$E4,$ED,$B7,$E0,$BE,$6D,$AE,$51,$8B,$EB,$7C,$CE,$A7,$32,$D7,$17
	.dc.b	$53,$C6,$FB,$D7,$29,$E1,$39,$78,$C6,$6B,$94,$6A,$FC,$64,$7A,$B5
	.dc.b	$D1,$82,$18,$24,$74,$24,$F7,$25,$F9,$89,$3E,$46,$DF,$9C,$97,$E1
	.dc.b	$BE,$73,$A8,$C5,$F4,$BE,$44,$BE,$2F,$9A,$F8,$B9,$AE,$37,$CE,$35
	.dc.b	$7C,$EF,$8C,$7B,$23,$15,$C2,$39,$1C,$2B,$93,$B6,$DF,$82,$F9,$B6
	.dc.b	$B9,$46,$2F,$AD,$F3,$3A,$9C,$CB,$5C,$5D,$4F,$1B,$EF,$5C,$A7,$84
	.dc.b	$E5,$E3,$19,$AE,$51,$AB,$F1,$91,$EA,$D7,$46,$08,$60,$91,$D0,$93
	.dc.b	$DC,$97,$E6,$24,$F9,$1B,$7E,$72,$5F,$86,$F9,$CE,$A3,$17,$D2,$F9
	.dc.b	$12,$F8,$BE,$6B,$E2,$E6,$B8,$DF,$38,$D5,$F3,$BE,$31,$EC,$8C,$57
	.dc.b	$08,$E4,$70,$AE,$4E,$DB,$7E,$0B,$E6,$DA,$E5,$18,$BE,$B7,$CC,$EA
	.dc.b	$73,$2D,$71,$75,$3C,$6F,$BD,$72,$9E,$13,$97,$8C,$66,$B9,$46,$AF
	.dc.b	$C6,$47,$AB,$5D,$18,$21,$82,$47,$42,$4F,$72,$5F,$98,$93,$E4,$6D
	.dc.b	$F9,$C9,$7E,$1B,$E7,$3A,$8C,$5F,$4B,$E4,$4B,$E2,$F9,$AF,$8B,$9A
	.dc.b	$E3,$7C,$E3,$57,$CE,$F8,$C7,$B2,$31,$5C,$23,$91,$C2,$B9,$3B,$6D
	.dc.b	$F8,$2F,$9B,$6B,$94,$62,$FA,$DF,$33,$A9,$CC,$B5,$C5,$D4,$F1,$BE
	.dc.b	$F5,$CA,$78,$4E,$5E,$31,$9A,$E5,$1A,$BF,$19,$1E,$AD,$74,$60,$86
	.dc.b	$09,$1D,$09,$3D,$C9,$7E,$62,$4F,$91,$B7,$E7,$25,$F8,$6F,$9C,$EA
	.dc.b	$31,$7D,$2F,$91,$2F,$8B,$E6,$BE,$2E,$6B,$8D,$F3,$8D,$5F,$3B,$E3
	.dc.b	$1E,$C8,$C5,$70,$8E,$47,$0A,$E4,$ED,$B7,$E0,$BE,$6D,$AE,$51,$8B
	.dc.b	$EB,$7C,$CE,$A7,$32,$D7,$17,$53,$C6,$FB,$D7,$29,$E1,$39,$78,$C6
	.dc.b	$6B,$94,$6A,$FC,$64,$7A,$B5,$D1,$82,$18,$24,$74,$24,$F7,$25,$F9
	.dc.b	$89,$3E,$46,$DF,$9C,$97,$E1,$BE,$73,$A8,$C5,$F4,$BE,$44,$BE,$2F
	.dc.b	$9A,$F8,$B9,$AE,$37,$CE,$35,$7C,$EF,$8C,$7B,$23,$15,$C2,$39,$1C
	.dc.b	$2B,$93,$B6,$DF,$82,$F9,$B6,$B9,$46,$2F,$AD,$F3,$3A,$9C,$CB,$5C
	.dc.b	$5D,$4F,$1B,$EF,$5C,$A7,$84,$E5,$E3,$19,$AE,$51,$AB,$F1,$91,$EA
	.dc.b	$D7,$46,$08,$60,$91,$D0,$93,$DC,$97,$E6,$24,$F9,$1B,$7E,$72,$5F
	.dc.b	$86,$F9,$CE,$A3,$17,$D2,$F9,$12,$F8,$BE,$6B,$E2,$E6,$B8,$DF,$38
	.dc.b	$D5,$F3,$BE,$31,$EC,$8C,$57,$08,$E4,$70,$AE,$4E,$DB,$7E,$0B,$E6
	.dc.b	$DA,$E5,$18,$BE,$B7,$CC,$EA,$73,$2D,$71,$75,$3C,$6F,$BD,$72,$9E
	.dc.b	$13,$97,$8C,$66,$B9,$46,$AF,$C6,$47,$AB,$5D,$18,$21,$82,$47,$42
	.dc.b	$4F,$72,$5F,$98,$93,$E4,$6D,$F9,$C9,$7E,$1B,$E7,$3A,$8C,$5F,$4B
	.dc.b	$E4,$4B,$E2,$F9,$AF,$8B,$9A,$E3,$7C,$E3,$57,$CE,$F8,$C7,$B2,$31
	.dc.b	$5C,$23,$91,$C2,$B9,$3B,$6D,$F8,$2F,$9B,$6B,$94,$62,$FA,$DF,$33
	.dc.b	$A9,$CC,$B5,$C5,$D4,$F1,$BE,$F5,$CA,$78,$4E,$5E,$31,$9A,$E5,$1A
	.dc.b	$BF,$19,$1E,$AD,$74,$60,$86,$09,$1D,$09,$3D,$C9,$7E,$62,$4F,$91
	.dc.b	$B7,$E7,$25,$F8,$6F,$9C,$EA,$31,$7D,$2F,$91,$2F,$8B,$E6,$BE,$2E
	.dc.b	$6B,$8D,$F3,$8D,$5F,$3B,$E3,$1E,$C8,$C5,$70,$8E,$47,$0A,$E4,$ED
	.dc.b	$B7,$E0,$BE,$6D,$AE,$51,$8B,$EB,$7C,$CE,$A7,$32,$D7,$17,$53,$C6
	.dc.b	$FB,$D7,$29,$E1,$39,$78,$C6,$6B,$94,$6A,$FC,$64,$7A,$B5,$D1,$82
	.dc.b	$18,$24,$74,$24,$F7,$25,$F9,$89,$3E,$46,$DF,$9C,$97,$E1,$BE,$73
	.dc.b	$A8,$C5,$F4,$BE,$44,$BE,$2F,$9A,$F8,$B9,$AE,$37,$CE,$35,$7C,$EF
	.dc.b	$8C,$7B,$23,$15,$C2,$39,$1C,$2B,$93,$B6,$DF,$82,$F9,$B6,$B9,$46
	.dc.b	$2F,$AD,$F3,$3A,$9C,$CB,$5C,$5D,$4F,$1B,$EF,$5C,$A7,$84,$E5,$E3
	.dc.b	$19,$AE,$51,$AB,$F1,$91,$EA,$D7,$46,$08,$60,$91,$D0,$93,$DC,$97
	.dc.b	$E6,$24,$F9,$1B,$7E,$72,$5F,$86,$F9,$CE,$A3,$17,$D2,$F9,$12,$F8
	.dc.b	$BE,$6B,$E2,$E6,$B8,$DF,$38,$D5,$F3,$BE,$31,$EC,$8C,$57,$08,$E4
	.dc.b	$70,$AE,$4E,$DB,$7E,$0B,$E6,$DA,$E5,$18,$BE,$B7,$CC,$EA,$73,$2D
	.dc.b	$71,$75,$3C,$6F,$BD,$72,$9E,$13,$97,$8C,$66,$B9,$46,$AF,$C6,$47
	.dc.b	$AB,$5D,$18,$21,$82,$47,$42,$4F,$72,$5F,$98,$93,$E4,$6D,$F9,$C9
	.dc.b	$7E,$1B,$E7,$3A,$8C,$5F,$4B,$E4,$4B,$E2,$F9,$AF,$8B,$9A,$E3,$7C
	.dc.b	$E3,$57,$CE,$F8,$C7,$B2,$31,$5C,$23,$91,$C2,$B9,$3B,$6D,$F8,$2F
	.dc.b	$9B,$6B,$94,$62,$FA,$DF,$33,$A9,$CC,$B5,$C5,$D4,$F1,$BE,$F5,$CA
	.dc.b	$78,$4E,$5E,$31,$9A,$E5,$1A,$BF,$19,$1E,$AD,$74,$60,$86,$09,$1D
	.dc.b	$09,$3D,$C9,$7E,$62,$4F,$91,$B7,$E7,$25,$F8,$6F,$9C,$EA,$31,$7D
	.dc.b	$2F,$91,$2F,$8B,$E6,$BE,$2E,$6B,$8D,$F3,$8D,$5F,$3B,$E3,$1E,$C8
	.dc.b	$C5,$70,$8E,$47,$0A,$E4,$ED,$B7,$E0,$BE,$6D,$AE,$51,$8B,$EB,$7C
	.dc.b	$CE,$A7,$32,$D2,$16,$95,$2F,$10,$BE,$15,$BA,$D4,$F1,$9C,$BC,$0C
	.dc.b	$5F,$67,$07,$37,$0E,$E3,$8B,$63,$2F,$8A,$32,$D8,$CB,$DA,$BA,$BD
	.dc.b	$5B,$3C,$23,$47,$89,$E7,$F5,$02,$00

;--------------------------------------------------------------------------------
;	FADD.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fadd_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#63,-(sp)
	peamsg	'FADD.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FADD.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fadd_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fadd.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FADD.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fadd.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FADD.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fadd_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3F,$D9,$BE,$53,$FB,$3B,$4A,$7F,$6C,$6B,$4A,$FF
	.dc.b	$70,$F9,$EF,$4C,$7F,$46,$F8,$1C,$67,$D2,$AC,$E4,$57,$07,$FA,$A9
	.dc.b	$89,$FE,$F3,$F2,$9F,$A7,$9D,$4F,$85,$B1,$99,$ED,$3C,$4E,$D3,$C1
	.dc.b	$57,$89,$FF,$02,$5D,$7F,$BE,$BE,$EB,$FC,$04,$78,$A3,$07,$C2,$F8
	.dc.b	$CC,$C7,$F8,$3D,$AF,$F8,$49,$AF,$F3,$98,$F9,$7F,$9B,$3E,$0B,$6D
	.dc.b	$8C,$C6,$0C,$F6,$C8,$E2,$DA,$FA,$F5,$FB,$29,$EB,$E3,$F4,$23,$E0
	.dc.b	$F8,$2E,$E4,$70,$79,$3E,$0B,$C6,$47,$07,$A3,$D4,$F3,$91,$C1,$F8
	.dc.b	$6F,$9B,$FE,$46,$DF,$73,$E1,$C6,$32,$FC,$6B,$5F,$F1,$AB,$FA,$F9
	.dc.b	$E5,$DF,$C0,$8F,$5F,$8A,$E6,$47,$07,$CC,$F5,$FE,$59,$1C,$1F,$9A
	.dc.b	$F9,$1F,$F1,$03,$77,$FA,$F1,$8C,$7F,$42,$D7,$FC,$93,$BB,$D8,$FA
	.dc.b	$E4,$BF,$B6,$FD,$AF,$0A,$47,$DE,$6D,$74,$EF,$EE,$49,$F2,$3F,$82
	.dc.b	$F2,$24,$7E,$19,$B3,$F7,$BC,$3F,$C3,$92,$7F,$16,$FE,$43,$E6,$5F
	.dc.b	$0D,$49,$3F,$BA,$8C,$BF,$92,$F1,$7F,$22,$47,$07,$E5,$A7,$F2,$A4
	.dc.b	$70,$7F,$04,$3F,$CA,$05,$EF,$FD,$0B,$8C,$3F,$85,$B5,$F9,$7F,$0D
	.dc.b	$FA,$02,$47,$E9,$1F,$D1,$F8,$D2,$3F,$4E,$D9,$FE,$15,$AF,$EB,$BB
	.dc.b	$BF,$99,$57,$F1,$66,$CE,$A3,$F5,$FB,$7E,$B8,$8E,$0F,$D9,$FB,$7F
	.dc.b	$B2,$23,$83,$F9,$0F,$CC,$7F,$A0,$DB,$E3,$7E,$FA,$31,$7F,$CA,$1A
	.dc.b	$FF,$A4,$3F,$37,$FF,$21,$57,$F2,$A6,$C7,$F2,$E6,$BB,$E5,$FE,$13
	.dc.b	$F5,$0F,$36,$92,$FF,$99,$C6,$5B,$E9,$70,$23,$83,$F9,$CF,$D4,$FF
	.dc.b	$A9,$1B,$87,$F1,$68,$C3,$FD,$01,$AF,$FA,$A1,$FF,$0F,$C7,$2F,$87
	.dc.b	$E0,$47,$AF,$81,$CC,$8E,$0F,$0F,$C2,$77,$23,$83,$CB,$F0,$9E,$32
	.dc.b	$38,$3F,$A8,$FC,$0F,$F2,$5C,$7F,$AA,$B5,$FF,$5E,$B5,$FF,$60,$35
	.dc.b	$FF,$61,$BF,$BB,$73,$DF,$FC,$DA,$31,$1A,$8C,$78,$F2,$BC,$2D,$9F
	.dc.b	$0F,$6F,$12,$4F,$99,$E9,$D9,$23,$D2,$D9,$FE,$DB,$F3,$5F,$E3,$39
	.dc.b	$FE,$DA,$D8,$CE,$3F,$D2,$A3,$11,$FE,$DA,$7F,$CB,$D3,$C3,$E0,$B4
	.dc.b	$47,$9F,$CD,$71,$23,$83,$BF,$BD,$D4,$8E,$0F,$1F,$C1,$F8,$48,$E0
	.dc.b	$FF,$01,$FC,$D7,$E1,$B1,$FF,$5C,$9F,$D8,$1C,$3F,$EB,$93,$1E,$84
	.dc.b	$9E,$2D,$8F,$F0,$7C,$FF,$43,$7F,$D7,$93,$DB,$8B,$A1,$D7,$CA,$91
	.dc.b	$E2,$6C,$F2,$E2,$F1,$12,$7C,$EF,$97,$E1,$E4,$7A,$9B,$3E,$2F,$92
	.dc.b	$F5,$12,$7D,$EF,$F7,$CF,$9A,$FE,$DD,$3F,$E2,$66,$A3,$FD,$93,$3F
	.dc.b	$E2,$88,$CC,$7F,$B2,$E3,$1E,$8F,$32,$3B,$E3,$EC,$47,$07,$3F,$23
	.dc.b	$C4,$47,$07,$F8,$DB,$E6,$BF,$0F,$7F,$ED,$0E,$5F,$F0,$9C,$62,$3C
	.dc.b	$CD,$5E,$07,$F8,$CA,$BF,$8B,$7A,$BF,$E1,$A8,$C4,$FF,$17,$8C,$5C
	.dc.b	$AF,$E3,$8D,$9F,$E3,$9C,$DC,$49,$3F,$C8,$5F,$1F,$E9,$92,$3D,$0D
	.dc.b	$9F,$E4,$9F,$01,$E1,$24,$FF,$29,$7D,$7E,$04,$8F,$80,$D9,$FE,$59
	.dc.b	$E2,$FA,$49,$3F,$CC,$5F,$67,$C8,$48,$FE,$68,$D9,$FE,$69,$EE,$FC
	.dc.b	$12,$4F,$F3,$97,$FC,$7B,$F3,$5F,$E1,$79,$FF,$33,$B5,$FC,$11,$F8
	.dc.b	$37,$0F,$FB,$EB,$9F,$E6,$7B,$B2,$FF,$A2,$36,$7F,$A2,$7A,$1F,$44
	.dc.b	$93,$FD,$25,$FA,$9F,$37,$23,$FA,$63,$67,$FA,$67,$81,$F6,$49,$3F
	.dc.b	$D4,$5F,$4F,$C8,$48,$FF,$3E,$E4,$8C,$57,$B3,$E4,$3D,$E4,$AD,$BF
	.dc.b	$85,$F0,$A4,$7F,$5D,$6C,$FF,$5D,$F0,$BF,$14,$49,$FE,$C4,$FE,$37
	.dc.b	$F4,$39,$1F,$D9,$5B,$3F,$33,$C8,$F9,$E4,$9F,$3B,$F4,$BC,$0F,$F7
	.dc.b	$54,$7D,$09,$66,$9C,$C7,$D4,$8C,$4F,$F1,$96,$C7,$F9,$FB,$A9,$CF
	.dc.b	$C2,$3F,$79,$18,$AF,$E4,$0D,$8E,$13,$B5,$F2,$EF,$97,$C3,$6F,$F9
	.dc.b	$43,$67,$FD,$1A,$D8,$F5,$2F,$92,$78,$FF,$EB,$28,$C3,$63,$35,$FE
	.dc.b	$B8,$9C,$B7,$E0,$5E,$8F,$16,$C6,$63,$07,$5F,$33,$F3,$38,$F8,$9B
	.dc.b	$3F,$9C,$BF,$F7,$FB,$EA,$8C,$BF,$1B,$E2,$6B,$FC,$31,$BA,$9A,$6A
	.dc.b	$9E,$F5,$25,$7C,$67,$5E,$C7,$FB,$42,$7F,$DE,$13,$F0,$E3,$07,$FB
	.dc.b	$52,$DF,$25,$49,$3F,$E3,$17,$5F,$25,$23,$FC,$68,$D7,$FC,$69,$EB
	.dc.b	$73,$23,$FE,$39,$7C,$5E,$54,$8F,$A8,$D7,$E9,$79,$5E,$72,$3F,$65
	.dc.b	$F3,$78,$B2,$3F,$C8,$8D,$7F,$DE,$3F,$9D,$7F,$BB,$93,$48,$FF,$BC
	.dc.b	$9F,$C0,$BC,$39,$30,$47,$FD,$F3,$0F,$86,$E0,$48,$CB,$F3,$7D,$B9
	.dc.b	$1F,$8F,$6B,$E4,$E9,$F0,$11,$EA,$F8,$FE,$76,$47,$81,$AF,$B3,$E7
	.dc.b	$7D,$04,$7C,$AF,$EE,$5F,$E5,$3F,$32,$F1,$B1,$1F,$DB,$63,$0E,$1A
	.dc.b	$F1,$6B,$FC,$0F,$2F,$F8,$00,$FF,$75,$6B,$F3,$3E,$2B,$E5,$1D,$94
	.dc.b	$BC,$1C,$36,$3C,$11,$83,$E9,$7E,$09,$23,$E7,$B5,$FB,$7E,$E7,$D7
	.dc.b	$23,$F5,$1F,$B5,$EC,$48,$FA,$ED,$7E,$27,$CD,$7A,$09,$8F,$9D,$F5
	.dc.b	$FC,$D7,$F3,$99,$F8,$ED,$7E,$63,$5F,$E4,$CD,$7F,$AB,$AF,$27,$E1
	.dc.b	$36,$39,$C6,$0F,$8D,$F3,$4F,$06,$92,$7E,$BC,$65,$BE,$2F,$02,$38
	.dc.b	$37,$F1,$BD,$08,$E0,$E7,$E0,$F8,$08,$E0,$EF,$F1,$5E,$42,$38,$3C
	.dc.b	$7C,$FE,$82,$38,$3C,$FF,$0B,$EC,$23,$83,$F2,$3E,$43,$D3,$FC,$33
	.dc.b	$63,$31,$83,$3E,$94,$8E,$2D,$7F,$3D,$3F,$C8,$DE,$BE,$3F,$42,$3E
	.dc.b	$0F,$13,$B9,$1C,$1E,$4F,$13,$C6,$47,$07,$EA,$3E,$C5,$DB,$F9,$D6
	.dc.b	$C6,$63,$06,$79,$A4,$71,$6B,$FB,$59,$FC,$23,$79,$3A,$11,$F0,$7C
	.dc.b	$C7,$72,$38,$3F,$77,$F6,$5F,$D9,$5B,$C9,$FB,$18,$C4,$7F,$03,$6B
	.dc.b	$FD,$F5,$7C,$EE,$C9,$1D,$1E,$5B,$48,$EE,$D7,$A7,$1F,$72,$3E,$47
	.dc.b	$C5,$E5,$48,$F3,$B5,$F2,$71,$F9,$C8,$FB,$1F,$E1,$5F,$67,$FD,$BA
	.dc.b	$3F,$8E,$35,$FF,$0B,$35,$FF,$23,$B5,$FF,$45,$2E,$3E,$64,$8F,$E4
	.dc.b	$71,$96,$F8,$5E,$02,$38,$37,$EB,$F9,$08,$E0,$FE,$5B,$F8,$03,$D4
	.dc.b	$FE,$44,$D8,$CC,$60,$FE,$4D,$F3,$BF,$E2,$E8,$FE,$6A,$D7,$F9,$D3
	.dc.b	$5F,$F1,$A3,$5F,$F3,$2F,$E0,$0F,$33,$F9,$83,$63,$F9,$FB,$5F,$E8
	.dc.b	$8D,$7F,$9E,$7B,$7F,$D8,$48,$FF,$5E,$7C,$B7,$23,$CA,$D7,$FB,$6F
	.dc.b	$D1,$FF,$69,$23,$FD,$99,$FE,$D9,$F2,$B2,$3F,$B4,$B5,$FE,$AD,$F3
	.dc.b	$FE,$37,$FC,$FE,$D7,$FD,$84,$D4,$DF,$33,$F9,$84,$7F,$53,$F7,$7F
	.dc.b	$BD,$11,$FE,$EF,$D5,$FD,$F4,$8E,$0F,$EC,$7F,$3F,$C9,$34,$93,$FE
	.dc.b	$05,$EB,$93,$FE,$02,$6B,$FD,$BF,$F0,$0F,$DD,$BF,$E6,$96,$BF,$E9
	.dc.b	$A6,$BF,$ED,$D6,$AF,$9F,$6D,$FD,$A5,$B1,$FD,$E1,$AF,$F7,$D6,$BD
	.dc.b	$F4,$FE,$E1,$46,$92,$3F,$C3,$3F,$49,$FE,$1F,$23,$83,$FC,$0B,$F4
	.dc.b	$1F,$C1,$1B,$DB,$FD,$EE,$31,$1F,$E0,$D6,$BF,$E0,$5E,$1F,$F1,$51
	.dc.b	$1F,$F1,$3B,$C7,$E6,$24,$76,$6B,$CF,$C8,$EC,$47,$C6,$F8,$7B,$24
	.dc.b	$79,$9A,$FF,$8F,$BE,$3F,$FC,$74,$47,$FC,$6E,$FF,$8F,$7E,$0A,$47
	.dc.b	$F8,$E9,$AF,$F9,$1B,$E3,$FF,$C8,$44,$7F,$C7,$EF,$F9,$17,$E9,$24
	.dc.b	$7F,$90,$9A,$FF,$8B,$FE,$17,$FB,$B3,$7D,$1F,$F1,$14,$62,$3F,$D5
	.dc.b	$8D,$7F,$DB,$AF,$F3,$46,$F8,$FC,$08,$FF,$8E,$3E,$81,$F1,$3F,$E2
	.dc.b	$D6,$C6,$63,$06,$7D,$09,$1C,$5A,$FF,$91,$67,$F9,$27,$D4,$A1,$F1
	.dc.b	$EF,$3B,$FD,$E3,$5F,$DD,$67,$31,$8B,$FF,$7A,$D6,$E7,$95,$74,$76
	.dc.b	$63,$C9,$61,$DE,$B8,$4A,$BB,$46,$1B,$87,$8A,$F8,$D7,$43,$83,$8A
	.dc.b	$E1,$87,$92,$BC,$D5,$DD,$E5,$85,$8F,$23,$F3,$6F,$AC,$7F,$B8,$70
	.dc.b	$F5,$36,$32,$FC,$4F,$21,$F5,$5F,$55,$B1,$F3,$5A,$F3,$6B,$FE,$75
	.dc.b	$F9,$BF,$F4,$A1,$1F,$F3,$43,$5F,$F3,$BB,$5E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$C1,$0F,$F7,$F3,$D5,$A7,$22,$9C,$4A,$71,$3F,$10,$1E,$4E,$1A
	.dc.b	$FE,$35,$AE,$CE,$C6,$CA,$6C,$A6,$CF,$11,$B2,$9B,$29,$B3,$CC,$64
	.dc.b	$A6,$4F,$59,$0E,$86,$9F,$39,$1F,$1B,$5E,$ED,$78,$1A,$21,$A2,$9A
	.dc.b	$30,$43,$04,$3C,$87,$A4,$A7,$A4,$F3,$98,$7C,$4D,$3B,$94,$EC,$53
	.dc.b	$B1,$C8,$87,$72,$9D,$CC,$10,$C1,$1F,$0B,$C1,$AE,$DA,$EC,$F2,$9B
	.dc.b	$29,$B2,$9B,$3D,$26,$CA,$6C,$A6,$CF,$69,$92,$99,$3E,$11,$0E,$86
	.dc.b	$9F,$39,$1F,$1B,$5E,$ED,$78,$1A,$21,$A2,$9A,$30,$43,$04,$3C,$87
	.dc.b	$A4,$A7,$A4,$F3,$98,$7C,$4D,$3B,$94,$EC,$53,$B1,$C8,$87,$72,$9D
	.dc.b	$CC,$10,$C1,$1F,$0B,$C1,$AE,$DA,$EC,$F2,$9B,$29,$B2,$9B,$3D,$26
	.dc.b	$CA,$6C,$A6,$CF,$69,$92,$99,$3E,$11,$0E,$86,$9F,$39,$1F,$1B,$5E
	.dc.b	$ED,$78,$1A,$21,$A2,$9A,$30,$43,$04,$3C,$87,$A4,$A7,$A4,$F3,$98
	.dc.b	$7C,$4D,$3B,$94,$EC,$53,$B1,$C8,$87,$72,$9D,$CC,$10,$C1,$1F,$0B
	.dc.b	$C1,$AE,$DA,$EC,$F2,$9B,$29,$B2,$9B,$3D,$26,$CA,$6C,$A6,$CF,$69
	.dc.b	$92,$99,$3E,$11,$0E,$86,$9F,$39,$1F,$1B,$5E,$ED,$78,$1A,$21,$A2
	.dc.b	$9A,$30,$43,$04,$3C,$87,$A4,$FC,$B5,$A7,$FC,$1F,$D2,$F5,$7F,$C2
	.dc.b	$98,$7F,$8E,$E3,$17,$FE,$34,$AD,$D6,$A3,$FE,$00,$B5,$FF,$0B,$57
	.dc.b	$F8,$6C,$8F,$AF,$5F,$F1,$BA,$F8,$27,$17,$F8,$AE,$3F,$BD,$4E,$A7
	.dc.b	$31,$FE,$37,$3F,$BE,$7D,$76,$6B,$FC,$A0,$F4,$8F,$F0,$23,$94,$E7
	.dc.b	$4D,$EF,$FE,$10,$9F,$15,$61,$FF,$0C,$D7,$13,$83,$CA,$78,$5F,$92
	.dc.b	$72,$6E,$3F,$C5,$B8,$78,$4E,$25,$78,$DF,$F9,$9F,$7F,$F3,$E5,$FB
	.dc.b	$9F,$F3,$E6,$3E,$3C,$7F,$CD,$93,$9A,$E3,$1F,$E6,$26,$D7,$08,$F3
	.dc.b	$1F,$E2,$C9,$E0,$D9,$F0,$1C,$0F,$F2,$1E,$3F,$1E,$BC,$71,$A9,$E0
	.dc.b	$77,$3E,$2E,$1C,$5C,$CF,$07,$47,$F8,$CE,$BA,$56,$E3,$D6,$F5,$75
	.dc.b	$59,$7E,$99,$C8,$B5,$ED,$AE,$71,$E0,$AE,$0F,$E0,$2B,$95,$FC,$08
	.dc.b	$EE,$73,$34,$71,$21,$AB,$ED,$3C,$AB,$CD,$3E,$27,$E3,$C6,$63,$CD
	.dc.b	$5D,$DF,$CA,$C6,$6B,$E2,$1D,$23,$B9,$CC,$A7,$90,$F1,$57,$13,$BB
	.dc.b	$C2,$BB,$1D,$DE,$07,$33,$89,$D3,$0F,$C5,$39,$24,$71,$8C,$BD,$27
	.dc.b	$DE,$E6,$7F,$16,$74,$21,$CC,$F0,$BC,$5F,$87,$1F,$21,$CB,$D0,$8F
	.dc.b	$D3,$9F,$0C,$62,$BE,$3B,$F0,$27,$BD,$75,$9F,$95,$18,$9E,$06,$DF
	.dc.b	$6B,$87,$F2,$E7,$C9,$70,$FE,$40,$C1,$E8,$3C,$87,$C0,$39,$99,$8F
	.dc.b	$0B,$87,$CE,$F6,$3C,$86,$E3,$C9,$3E,$97,$F2,$F5,$F0,$27,$47,$D7
	.dc.b	$3C,$13,$B7,$18,$7E,$81,$EB,$7E,$29,$D4,$7D,$43,$E7,$9D,$0C,$57
	.dc.b	$79,$F3,$C6,$0E,$F7,$F1,$1F,$B9,$3B,$8F,$C9,$47,$B4,$F8,$0F,$A9
	.dc.b	$E2,$77,$38,$98,$9E,$EE,$2F,$D4,$F2,$3A,$5F,$A9,$D6,$1E,$07,$CD
	.dc.b	$3C,$27,$51,$EE,$AE,$B1,$F9,$79,$D5,$FA,$EB,$4F,$37,$F6,$A7,$E0
	.dc.b	$2B,$CA,$E8,$C4,$7D,$83,$E2,$57,$5B,$E4,$D9,$D9,$F5,$A3,$33,$FA
	.dc.b	$1C,$3E,$54,$E8,$FC,$81,$F6,$A3,$E5,$61,$E8,$8F,$94,$78,$CD,$1F
	.dc.b	$1A,$3C,$67,$57,$E7,$3E,$D7,$81,$F3,$4E,$E5,$3E,$69,$E3,$29,$D8
	.dc.b	$F9,$F5,$EA,$7F,$47,$3E,$5C,$3E,$3D,$FB,$27,$81,$F3,$0B,$87,$C5
	.dc.b	$6E,$1E,$7B,$E4,$FC,$0C,$3B,$61,$F9,$B8,$F6,$D7,$82,$73,$1F,$4E
	.dc.b	$3F,$1D,$7D,$5F,$63,$4F,$C5,$16,$39,$D7,$59,$FD,$F1,$FE,$01,$3F
	.dc.b	$A7,$96,$7F,$C1,$ED,$9D,$3F,$D2,$8C,$18,$8D,$BF,$DD,$0C,$CE,$3E
	.dc.b	$0F,$F3,$84,$AF,$D7,$36,$7F,$50,$D9,$FD,$3B,$67,$FB,$23,$67,$FA
	.dc.b	$4B,$5F,$E6,$6D,$7F,$9F,$36,$BF,$AA,$36,$7F,$AD,$B6,$7F,$B7,$36
	.dc.b	$BD,$13,$E6,$24,$F9,$9B,$1E,$66,$9E,$62,$C7,$99,$AF,$99,$AF,$99
	.dc.b	$B5,$E6,$69,$E6,$2C,$79,$9B,$3E,$68,$F2,$11,$F2,$36,$3C,$8D,$3C
	.dc.b	$85,$3C,$85,$3C,$85,$9F,$23,$63,$C8,$D3,$C8,$59,$F2,$1E,$22,$47
	.dc.b	$89,$B1,$E2,$6B,$E2,$6B,$E2,$6C,$F8,$9B,$3E,$26,$CF,$89,$A7,$8A
	.dc.b	$3C,$04,$8F,$03,$5F,$03,$5F,$03,$67,$C0,$D3,$C0,$58,$F0,$36,$7C
	.dc.b	$11,$D8,$87,$62,$9D,$8B,$3D,$9B,$3D,$9A,$76,$2C,$F6,$7A,$10,$E8
	.dc.b	$59,$E8,$D9,$E8,$D3,$A1,$67,$A3,$C8,$93,$C9,$B3,$C9,$A7,$22,$CF
	.dc.b	$29,$E0,$49,$E0,$D9,$E0,$D9,$E1,$3A,$24,$E9,$B1,$A8,$C1,$1C,$12
	.dc.b	$3F,$D3,$D5,$FB,$D6,$9F,$AA,$2C,$FE,$65,$B3,$F8,$A6,$9F,$74,$B1
	.dc.b	$F4,$9A,$87,$C1,$FF,$24,$70,$62,$3F,$95,$B4,$FE,$5B,$1B,$79,$7C
	.dc.b	$1D,$4D,$24,$F2,$F5,$F4,$47,$07,$F3,$8E,$77,$17,$56,$C6,$63,$07
	.dc.b	$87,$E2,$1F,$16,$F8,$BF,$E8,$31,$86,$FB,$1F,$ED,$9C,$FF,$A5,$4E
	.dc.b	$5F,$E9,$9C,$EE,$86,$FF,$D2,$A3,$0D,$8C,$CF,$06,$BF,$EE,$A3,$9F
	.dc.b	$43,$9F,$FD,$D7,$19,$3F,$AA,$F3,$7C,$C4,$D2,$3F,$41,$FF,$28,$73
	.dc.b	$7F,$B6,$39,$BF,$D9,$6F,$FB,$59,$FF,$66,$F3,$DA,$FF,$60,$B5,$FF
	.dc.b	$55,$B5,$FF,$5B,$BC,$DF,$F6,$3C,$D7,$FD,$62,$D7,$FD,$7A,$DE,$7F
	.dc.b	$F7,$55,$7E,$2A,$5C,$FF,$D6,$CF,$D2,$71,$E5,$B1,$F8,$C7,$F6,$86
	.dc.b	$E3,$F4,$A6,$CB,$1B,$7F,$71,$1B,$8F,$D4,$46,$DA,$6C,$FD,$E4,$6F
	.dc.b	$5F,$D6,$46,$5F,$E0,$F1,$99,$F5,$5F,$91,$F0,$D7,$E9,$67,$0F,$F4
	.dc.b	$F7,$D6,$E6,$3C,$A7,$F6,$68,$CC,$7C,$48,$CB,$C5,$EA,$E4,$F3,$BC
	.dc.b	$4C,$C7,$82,$BF,$A6,$BD,$0C,$47,$F4,$17,$0F,$68,$DB,$C8,$C1,$B3
	.dc.b	$13,$E0,$3C,$87,$CA,$3D,$86,$63,$9B,$93,$D0,$FC,$A8,$F4,$46,$67
	.dc.b	$CA,$FF,$48,$7C,$AF,$38,$F1,$3B,$7C,$AE,$DF,$2C,$F5,$8F,$2B,$F6
	.dc.b	$9F,$2B,$93,$CA,$65,$F2,$C7,$E1,$63,$2F,$DE,$3B,$C7,$27,$C7,$1C
	.dc.b	$67,$D3,$3F,$D5,$5F,$4C,$77,$9D,$9C,$E7,$D8,$7E,$35,$CB,$4F,$61
	.dc.b	$EF,$36,$75,$7D,$E7,$9C,$E2,$F9,$4E,$31,$E5,$8F,$4C,$7F,$5F,$9F
	.dc.b	$2C,$71,$8D,$BE,$59,$F4,$47,$8C,$F0,$1E,$37,$2E,$1E,$CF,$98,$ED
	.dc.b	$1E,$67,$91,$F3,$8E,$E6,$2B,$E4,$B9,$3A,$47,$8C,$F3,$3E,$D3,$27
	.dc.b	$17,$DB,$5F,$BF,$8F,$F4,$03,$FD,$34,$91,$FD,$51,$B1,$FE,$80,$6B
	.dc.b	$C2,$3F,$AF,$57,$FA,$1B,$91,$C3,$FE,$0E,$6C,$FF,$83,$9B,$5F,$E0
	.dc.b	$E6,$D7,$F8,$65,$B3,$FD,$A9,$B3,$F3,$5A,$FB,$5A,$FC,$36,$DA,$7C
	.dc.b	$A7,$F7,$08,$CB,$89,$FE,$ED,$18,$7F,$AA,$C6,$EF,$D4,$47,$D4,$D8
	.dc.b	$F5,$36,$FD,$4D,$9F,$53,$67,$D4,$D7,$D4,$D7,$D4,$DB,$F5,$47,$90
	.dc.b	$F3,$C6,$67,$D5,$0B,$F5,$5F,$D3,$AD,$4F,$13,$4E,$23,$55,$EC,$24
	.dc.b	$FB,$1B,$1E,$C6,$9E,$C2,$C7,$B1,$AF,$B1,$AF,$B1,$B7,$EC,$9F,$0C
	.dc.b	$F7,$23,$DD,$B3,$DD,$A7,$72,$9D,$CA,$77,$2C,$F7,$8F,$B8,$78,$8F
	.dc.b	$93,$7F,$21,$CC,$F0,$73,$7F,$1D,$CD,$F9,$48,$F9,$5B,$3E,$56,$BE
	.dc.b	$56,$BE,$56,$DF,$96,$3E,$EC,$F9,$5F,$B1,$7E,$57,$27,$94,$CD,$F9
	.dc.b	$5C,$DF,$4A,$F4,$9C,$DF,$9E,$66,$38,$C6,$4F,$69,$EB,$23,$EB,$6B
	.dc.b	$EB,$6B,$EB,$6D,$FA,$E3,$F1,$F1,$D5,$F2,$BA,$3C,$AF,$AE,$7C,$B1
	.dc.b	$98,$F2,$99,$7C,$AF,$8C,$87,$8C,$A7,$8C,$B3,$E3,$BF,$95,$1E,$43
	.dc.b	$A1,$E4,$37,$3E,$4A,$F8,$C7,$92,$32,$79,$1C,$C7,$92,$3C,$44,$3C
	.dc.b	$45,$AF,$15,$FD,$78,$F1,$57,$27,$C4,$6E,$BC,$47,$C3,$8F,$14,$64
	.dc.b	$F1,$46,$4F,$13,$E0,$25,$78,$23,$E9,$47,$81,$E0,$F8,$07,$2F,$FA
	.dc.b	$25,$FF,$4B,$9F,$E9,$87,$0E,$30,$FF,$42,$CF,$FA,$37,$93,$FD,$18
	.dc.b	$47,$FD,$21,$F3,$3F,$E8,$F2,$38,$3F,$D1,$71,$FE,$8D,$AF,$F4,$4B
	.dc.b	$5F,$0F,$27,$F4,$CD,$37,$CB,$FE,$9B,$F8,$CF,$F5,$4E,$46,$92,$FF
	.dc.b	$D1,$9A,$C7,$FD,$18,$D7,$FD,$10,$7F,$AD,$39,$9C,$BF,$E8,$D8,$FF
	.dc.b	$47,$F2,$FF,$A3,$9F,$86,$7F,$A3,$0F,$F4,$51,$0F,$F4,$89,$8E,$37
	.dc.b	$1F,$FA,$30,$E3,$CB,$FE,$90,$6C,$B9,$76,$FF,$A3,$A3,$31,$88,$FF
	.dc.b	$47,$C6,$0F,$F4,$A9,$5E,$07,$D8,$9F,$F6,$95,$71,$86,$DF,$B5,$7F
	.dc.b	$DA,$D3,$AA,$FB,$F3,$93,$FD,$A0,$4E,$1F,$F6,$BB,$FE,$D3,$E9,$E4
	.dc.b	$7E,$B7,$E5,$6C,$FF,$54,$C3,$FD,$B9,$B7,$F5,$A7,$FD,$C9,$5B,$9D
	.dc.b	$3C,$23,$FA,$B1,$C2,$3F,$BC,$56,$6B,$94,$E4,$FF,$75,$D7,$37,$8B
	.dc.b	$BA,$FC,$CB,$99,$E2,$E6,$78,$CF,$F4,$2A,$C4,$79,$A3,$CA,$64,$AF
	.dc.b	$66,$D7,$F8,$86,$78,$CF,$94,$E4,$7F,$3A,$73,$5D,$B8,$11,$8B,$E6
	.dc.b	$7C,$02,$D7,$98,$F0,$1C,$DC,$BD,$AF,$DB,$1E,$78,$DC,$61,$B1,$D5
	.dc.b	$A7,$58,$D1,$EB,$8E,$B1,$EB,$3D,$95,$E8,$8C,$DF,$AE,$FD,$55,$D0
	.dc.b	$FF,$0A,$38,$BE,$B5,$E1,$6C,$F8,$63,$EB,$9E,$98,$CC,$7B,$63,$2F
	.dc.b	$C1,$3B,$1E,$5A,$FB,$D1,$A9,$E3,$3E,$8A,$F1,$4E,$4F,$11,$63,$C4
	.dc.b	$43,$D2,$FB,$8F,$9F,$7F,$06,$3C,$37,$F0,$67,$51,$F0,$67,$57,$E2
	.dc.b	$AD,$4F,$0A,$E7,$7E,$48,$F0,$5F,$09,$F2,$47,$C1,$33,$3B,$70,$79
	.dc.b	$8E,$87,$85,$D3,$F0,$63,$CB,$7F,$32,$BB,$4F,$BC,$F9,$53,$F4,$4F
	.dc.b	$0B,$83,$E8,$4F,$AE,$3D,$27,$43,$CF,$3D,$27,$CA,$71,$9F,$C6,$3E
	.dc.b	$67,$C5,$5E,$6B,$FB,$33,$E3,$7E,$3C,$F7,$8F,$65,$F7,$7E,$1B,$5F
	.dc.b	$87,$7E,$17,$C1,$1E,$4B,$EC,$FB,$63,$B3,$EA,$7C,$8E,$CE,$8F,$C9
	.dc.b	$32,$78,$0A,$78,$23,$AB,$F1,$A3,$E4,$DF,$A9,$A7,$7B,$F6,$1F,$09
	.dc.b	$EF,$1F,$0A,$7E,$4B,$C5,$F3,$9F,$08,$F0,$9E,$59,$E7,$3E,$98,$CC
	.dc.b	$F3,$7E,$19,$E9,$9D,$B6,$7C,$AF,$58,$CB,$ED,$8F,$29,$F2,$5F,$2B
	.dc.b	$EB,$38,$C7,$AE,$3C,$31,$B7,$C2,$73,$32,$7C,$73,$DB,$1C,$4F,$0B
	.dc.b	$ED,$8E,$A7,$B6,$33,$3E,$D3,$F4,$97,$F8,$17,$FA,$65,$7F,$78,$7F
	.dc.b	$AC,$57,$F7,$A8,$FE,$6A,$65,$C2,$8D,$BB,$6D,$7F,$52,$EA,$FE,$7E
	.dc.b	$4A,$FE,$4E,$DA,$FE,$2A,$D9,$FE,$26,$D9,$FD,$6B,$67,$F6,$4D,$7F
	.dc.b	$52,$D7,$F9,$93,$6B,$FB,$D3,$67,$FB,$83,$67,$FB,$5B,$6B,$D1,$3E
	.dc.b	$62,$4F,$99,$B1,$E6,$69,$E6,$2C,$79,$9A,$F9,$9A,$F9,$9B,$5E,$66
	.dc.b	$9E,$62,$C7,$99,$B3,$E6,$8F,$21,$1F,$23,$63,$C8,$D3,$C8,$53,$C8
	.dc.b	$53,$C8,$59,$F2,$36,$3C,$8D,$3C,$85,$9F,$21,$E2,$24,$78,$9B,$1E
	.dc.b	$26,$BE,$26,$BE,$26,$CF,$89,$B3,$E2,$6C,$F8,$9A,$78,$A3,$C0,$48
	.dc.b	$F0,$35,$F0,$35,$F0,$36,$7C,$0D,$3C,$05,$8F,$03,$67,$C1,$1D,$88
	.dc.b	$76,$29,$D8,$B3,$D9,$B3,$D9,$A7,$62,$CF,$67,$A1,$0E,$85,$9E,$8D
	.dc.b	$9E,$8D,$3A,$16,$7A,$3C,$89,$3C,$9B,$3C,$9A,$72,$2C,$F2,$9E,$04
	.dc.b	$9E,$0D,$9E,$0D,$9E,$13,$A2,$4E,$9B,$1A,$8C,$11,$C1,$23,$FD,$6B
	.dc.b	$5F,$BD,$69,$FA,$A2,$CF,$E6,$5B,$3F,$8A,$69,$F7,$4B,$1F,$49,$AA
	.dc.b	$DD,$5F,$E7,$28,$19,$70,$D9,$FE,$5D,$3C,$1B,$F4,$5A,$83,$49,$1F
	.dc.b	$EB,$E9,$FF,$61,$F2,$7F,$37,$E5,$FF,$61,$46,$0F,$F5,$9D,$BE,$0B
	.dc.b	$FD,$86,$48,$F0,$BF,$CF,$39,$5F,$29,$49,$3F,$D0,$E3,$2F,$FB,$14
	.dc.b	$FF,$62,$76,$B5,$C4,$7C,$09,$4E,$63,$FD,$8E,$43,$FD,$9A,$6F,$9F
	.dc.b	$FD,$98,$FF,$B1,$88,$7F,$B0,$8A,$7F,$AE,$47,$2F,$FA,$E7,$8D,$B7
	.dc.b	$FA,$DE,$75,$39,$8C,$47,$D4,$3F,$DA,$D1,$97,$9F,$27,$FB,$42,$30
	.dc.b	$7F,$B1,$63,$71,$FE,$C2,$9F,$F5,$D4,$60,$E9,$5F,$EC,$58,$FF,$61
	.dc.b	$92,$3F,$B6,$47,$E0,$97,$67,$FB,$3C,$E0,$FF,$AF,$9F,$D8,$71,$FF
	.dc.b	$5D,$3F,$4D,$5F,$90,$94,$61,$F6,$3C,$23,$FA,$7B,$98,$FC,$9B,$97
	.dc.b	$FA,$03,$97,$9C,$7E,$F6,$3C,$0F,$EC,$0F,$03,$B3,$96,$9F,$CE,$1C
	.dc.b	$47,$13,$C2,$71,$9F,$2B,$97,$C2,$E5,$F2,$B9,$9E,$8E,$63,$CA,$7F
	.dc.b	$0F,$9F,$2B,$FC,$06,$3C,$A6,$E3,$CA,$FF,$3D,$7C,$B1,$C6,$3C,$AF
	.dc.b	$13,$CA,$64,$F2,$99,$3C,$A6,$67,$CA,$E4,$F2,$BF,$CA,$A7,$CA,$FF
	.dc.b	$13,$8E,$AE,$67,$C2,$E6,$3A,$B9,$7C,$2E,$5E,$B1,$EE,$72,$FE,$0F
	.dc.b	$93,$DF,$1D,$A3,$E0,$1F,$0C,$F9,$86,$1F,$45,$70,$7F,$A5,$3F,$1A
	.dc.b	$BC,$47,$B8,$F1,$3B,$8F,$E8,$CE,$E7,$C9,$3B,$7C,$87,$82,$7C,$87
	.dc.b	$F6,$29,$E0,$E0,$F8,$31,$E5,$9E,$EF,$18,$F2,$BE,$B7,$E9,$3B,$3B
	.dc.b	$B8,$3A,$1E,$67,$84,$7B,$1E,$11,$AA,$F6,$57,$D8,$9E,$75,$BA,$D4
	.dc.b	$78,$DF,$A6,$7A,$DE,$A7,$C2,$3A,$BE,$C3,$47,$D0,$79,$57,$81,$FC
	.dc.b	$19,$E0,$39,$47,$80,$F6,$1C,$CF,$86,$73,$32,$7B,$CC,$4F,$A5,$DB
	.dc.b	$F3,$1F,$F1,$C9,$FC,$8C,$FC,$E9,$4F,$D7,$96,$BF,$8C,$12,$7F,$D0
	.dc.b	$6D,$8F,$F4,$43,$FC,$C9,$FE,$89,$1D,$7B,$3F,$C1,$44,$7F,$C1,$4D
	.dc.b	$8F,$F0,$53,$6B,$FC,$26,$D9,$FE,$40,$D9,$FB,$AD,$7E,$20,$70,$FD
	.dc.b	$63,$FB,$9B,$8A,$F3,$57,$9E,$73,$5E,$42,$3E,$46,$C7,$91,$A7,$90
	.dc.b	$B1,$E4,$6B,$E4,$AF,$04,$7F,$81,$27,$C1,$3E,$78,$DB,$E7,$30,$78
	.dc.b	$49,$1E,$16,$C7,$85,$A7,$84,$A7,$85,$EB,$5F,$01,$FA,$86,$0F,$21
	.dc.b	$88,$E2,$78,$EB,$D2,$FD,$C3,$FB,$53,$8A,$F5,$B6,$79,$47,$9C,$8F
	.dc.b	$9D,$B3,$E7,$6B,$E7,$3C,$35,$E7,$7B,$CF,$9D,$F1,$1F,$60,$DB,$EA
	.dc.b	$76,$60,$D9,$F3,$27,$BB,$ED,$8E,$91,$EA,$69,$EA,$3D,$04,$8F,$43
	.dc.b	$5F,$44,$74,$9F,$44,$79,$63,$13,$E5,$8C,$3C,$5F,$11,$E7,$8F,$13
	.dc.b	$63,$C4,$F8,$08,$78,$1E,$53,$E0,$8F,$74,$60,$F0,$38,$38,$9E,$09
	.dc.b	$F2,$BF,$1A,$BE,$D3,$F2,$A3,$27,$90,$B3,$E4,$8F,$11,$D2,$BC,$55
	.dc.b	$CA,$3D,$E6,$67,$C7,$19,$8F,$21,$C8,$87,$96,$3D,$4F,$95,$FA,$AF
	.dc.b	$95,$CD,$79,$5A,$79,$63,$91,$E1,$39,$1F,$34,$E4,$66,$39,$36,$39
	.dc.b	$57,$B1,$F0,$C7,$B0,$F3,$C6,$5E,$86,$4F,$09,$A8,$F8,$27,$27,$E0
	.dc.b	$9A,$24,$73,$9F,$9F,$1E,$63,$91,$EA,$72,$F7,$8E,$13,$FE,$5F,$E1
	.dc.b	$FF,$45,$BF,$E8,$E2,$4F,$FA,$1B,$93,$FD,$2B,$C6,$EB,$72,$FF,$A3
	.dc.b	$EB,$FA,$34,$FF,$A3,$9B,$3A,$7F,$D1,$FD,$DF,$D2,$5F,$F5,$0E,$28
	.dc.b	$C4,$7F,$A9,$DB,$19,$8F,$F5,$1F,$BD,$82,$38,$3F,$D2,$5F,$29,$FE
	.dc.b	$97,$23,$83,$FD,$22,$39,$BF,$D2,$3B,$61,$B1,$FE,$91,$3F,$D0,$4B
	.dc.b	$97,$FD,$23,$18,$6B,$FE,$95,$ED,$F4,$AE,$CA,$47,$DA,$DE,$DF,$F4
	.dc.b	$61,$FE,$89,$7F,$D8,$6E,$39,$FF,$D1,$C7,$FA,$2E,$7F,$D2,$27,$FA
	.dc.b	$61,$C7,$47,$FA,$56,$3F,$BA,$74,$7F,$A2,$49,$C9,$F3,$3A,$3F,$D1
	.dc.b	$4F,$EF,$A2,$CF,$FA,$2E,$34,$47,$EB,$4F,$FA,$4F,$9F,$F0,$F1,$FC
	.dc.b	$11,$AF,$EC,$9A,$FE,$55,$BC,$F8,$6D,$7E,$72,$7F,$28,$FE,$A0,$C6
	.dc.b	$FF,$AC,$8F,$CB,$46,$0F,$DB,$4F,$F7,$83,$9C,$72,$3F,$40,$7E,$9A
	.dc.b	$BC,$13,$C6,$3D,$0D,$7D,$0D,$7D,$0D,$8E,$31,$D4,$D9,$4D,$95,$DD
	.dc.b	$7A,$63,$2D,$8C,$CF,$BE,$79,$57,$AA,$B8,$36,$7D,$E7,$8C,$F3,$39
	.dc.b	$9F,$7B,$5F,$E1,$4D,$7F,$83,$B4,$F0,$BF,$08,$F4,$47,$43,$B3,$E4
	.dc.b	$6C,$F9,$1A,$79,$23,$89,$B2,$C6,$DB,$1B,$8F,$84,$F7,$8C,$9E,$A7
	.dc.b	$AB,$4C,$9E,$C7,$2D,$32,$7B,$9C,$B4,$CC,$FC,$07,$EC,$CF,$89,$F4
	.dc.b	$34,$F4,$16,$7D,$07,$D7,$72,$D3,$31,$F9,$0A,$EC,$FE,$46,$3D,$D3
	.dc.b	$CE,$3D,$8D,$7D,$4D,$7D,$0D,$AE,$75,$D1,$FD,$05,$7B,$9C,$BD,$8F
	.dc.b	$25,$76,$6B,$D9,$AF,$66,$C7,$28,$F8,$71,$83,$A3,$F8,$7A,$F2,$46
	.dc.b	$23,$E5,$BB,$78,$9B,$71,$1F,$22,$30,$7C,$78,$E4,$FC,$73,$89,$FC
	.dc.b	$00,$A6,$63,$93,$96,$B9,$8F,$A6,$E5,$FC,$B3,$FE,$97,$9F,$E3,$51
	.dc.b	$FD,$39,$B1,$FD,$4D,$AE,$E3,$47,$F8,$CF,$8B,$FB,$03,$FC,$25,$B1
	.dc.b	$FC,$15,$B5,$FC,$65,$B3,$FE,$21,$6C,$FF,$75,$6B,$FE,$05,$6B,$FE
	.dc.b	$1E,$6D,$7F,$91,$9B,$3E,$3E,$4F,$E6,$31,$8A,$FF,$2F,$B6,$FF,$D3
	.dc.b	$0D,$AF,$EF,$D5,$C0,$8F,$AD,$B5,$EB,$6D,$7A,$DB,$3E,$B6,$CF,$AD
	.dc.b	$AF,$AD,$AF,$AD,$B7,$EB,$6C,$FA,$CF,$11,$27,$C4,$DA,$F1,$34,$F1
	.dc.b	$16,$3C,$4D,$7C,$4D,$7C,$4D,$AF,$13,$4F,$14,$78,$09,$3E,$06,$C7
	.dc.b	$81,$A7,$80,$A7,$80,$A7,$80,$B3,$E0,$6C,$78,$27,$B1,$1E,$CD,$8E
	.dc.b	$CD,$7B,$35,$EC,$D9,$EC,$D9,$EC,$74,$24,$F4,$6B,$D1,$AF,$46,$CF
	.dc.b	$46,$9D,$23,$91,$0E,$45,$39,$16,$39,$36,$39,$3C,$08,$70,$2C,$70
	.dc.b	$6C,$70,$74,$48,$D3,$63,$51,$82,$38,$24,$7E,$AA,$BF,$2A,$D8,$FC
	.dc.b	$33,$4F,$B2,$5A,$F9,$CD,$8F,$8A,$D3,$DC,$53,$D2,$53,$CC,$5A,$F2
	.dc.b	$B6,$3C,$B3,$FB,$CA,$F2,$B6,$3C,$AD,$9F,$2B,$4F,$29,$63,$CA,$D8
	.dc.b	$F2,$B5,$F2,$B5,$F2,$B6,$BC,$AD,$9F,$29,$FC,$38,$F1,$96,$7F,$4B
	.dc.b	$CB,$FE,$89,$E6,$FF,$44,$40,$91,$FE,$89,$70,$1E,$0F,$FA,$28,$FF
	.dc.b	$45,$63,$FE,$85,$70,$43,$88,$5D,$DF,$E8,$C2,$47,$27,$FD,$23,$C8
	.dc.b	$EB,$A4,$9F,$F4,$C4,$65,$BF,$29,$C0,$8E,$0F,$F4,$67,$63,$4A,$72
	.dc.b	$E7,$FF,$44,$CF,$FA,$4D,$B7,$FE,$8F,$E7,$D1,$FE,$89,$9F,$F5,$47
	.dc.b	$3D,$7D,$A7,$FA,$3C,$FF,$56,$72,$7F,$BF,$4F,$F7,$F3,$9E,$57,$37
	.dc.b	$FA,$2F,$27,$17,$C6,$E6,$FF,$41,$C6,$2F,$E8,$E1,$8A,$F9,$7D,$99
	.dc.b	$AD,$57,$D9,$8F,$EA,$DD,$F9,$AD,$9F,$76,$B7,$4E,$8D,$3F,$79,$AF
	.dc.b	$39,$E0,$7D,$A8,$FF,$0B,$4F,$FA,$17,$E8,$7F,$B8,$D7,$DE,$6B,$F8
	.dc.b	$86,$BF,$8D,$6A,$5E,$0F,$81,$B7,$E0,$9E,$1D,$59,$AF,$2D,$B9,$29
	.dc.b	$FD,$D7,$E0,$6E,$1C,$70,$FC,$04,$E1,$FE,$EB,$36,$BE,$49,$CE,$B8
	.dc.b	$36,$77,$18,$AE,$0F,$E9,$27,$47,$F7,$EA,$CD,$7C,$FA,$AF,$E8,$62
	.dc.b	$CF,$0A,$E5,$5E,$97,$B1,$2B,$F3,$0E,$8F,$F1,$15,$72,$9E,$BD,$08
	.dc.b	$C4,$F1,$BF,$23,$4E,$57,$D0,$F3,$3D,$0D,$38,$34,$7C,$A7,$4E,$23
	.dc.b	$D4,$ED,$F9,$77,$EB,$8D,$3E,$47,$E5,$98,$9E,$EF,$53,$33,$CA,$32
	.dc.b	$F6,$8F,$95,$18,$9F,$83,$5D,$6B,$85,$72,$AF,$7C,$60,$8F,$F1,$CB
	.dc.b	$F9,$0F,$33,$C6,$E2,$7C,$C4,$9E,$2F,$F2,$78,$D4,$62,$3D,$4D,$8C
	.dc.b	$CF,$88,$93,$F1,$DF,$15,$7A,$5C,$C7,$B8,$F0,$38,$AF,$85,$5D,$27
	.dc.b	$91,$0F,$0C,$F4,$3D,$8F,$23,$D1,$5D,$DE,$13,$CD,$E2,$73,$AE,$F3
	.dc.b	$B8,$C9,$E5,$8F,$80,$7D,$53,$75,$E0,$7E,$14,$F1,$73,$3E,$33,$9B
	.dc.b	$83,$9D,$FD,$CA,$E4,$7C,$4B,$F3,$4F,$79,$D5,$73,$8C,$57,$E5,$9E
	.dc.b	$6F,$DC,$BF,$15,$7E,$56,$73,$5E,$3B,$EB,$3E,$F6,$DF,$8A,$7E,$4B
	.dc.b	$5F,$87,$1A,$7C,$53,$D6,$7E,$C3,$E6,$AE,$E7,$C4,$8E,$4F,$C1,$76
	.dc.b	$79,$CC,$C7,$39,$FA,$8F,$38,$DC,$7A,$6F,$DB,$1F,$04,$CD,$77,$9F
	.dc.b	$2C,$75,$69,$C5,$EE,$71,$7B,$BB,$8F,$4B,$F4,$E7,$BC,$FE,$BC,$D9
	.dc.b	$CC,$EB,$3E,$13,$8B,$F9,$08,$E3,$1F,$C5,$A3,$73,$D4,$F1,$BD,$4F
	.dc.b	$C3,$D7,$C1,$7B,$CF,$59,$F7,$9C,$63,$99,$DD,$E7,$1D,$CF,$1C,$7D
	.dc.b	$AA,$F1,$CF,$EA,$63,$89,$97,$89,$99,$E2,$7D,$98,$F1,$C7,$86,$33
	.dc.b	$5F,$66,$7D,$33,$FE,$08,$7F,$90,$1F,$E2,$69,$FE,$A0,$D9,$FE,$EE
	.dc.b	$FF,$33,$31,$3C,$67,$71,$FC,$F0,$C9,$FC,$D7,$C5,$FE,$9E,$49,$FE
	.dc.b	$56,$D8,$FD,$43,$67,$F7,$6D,$9F,$E2,$0D,$9F,$E0,$8D,$7F,$72,$D9
	.dc.b	$FE,$04,$DA,$FE,$EE,$DA,$FE,$DA,$DA,$F3,$47,$90,$8F,$91,$B5,$E4
	.dc.b	$69,$E4,$2C,$79,$1A,$F9,$1B,$1E,$46,$D7,$91,$A7,$90,$B3,$E4,$3C
	.dc.b	$44,$9F,$13,$63,$C4,$D3,$C4,$53,$C4,$58,$F1,$36,$7C,$4D,$9F,$13
	.dc.b	$4F,$14,$F8,$09,$1E,$06,$C7,$81,$AF,$81,$A7,$80,$B3,$E0,$6C,$F8
	.dc.b	$1B,$3E,$08,$EC,$47,$B3,$5E,$CD,$8E,$CD,$9E,$CD,$3B,$16,$7B,$1D
	.dc.b	$08,$74,$2C,$F4,$6C,$F4,$6C,$F4,$69,$D0,$E4,$49,$E4,$D9,$E4,$D9
	.dc.b	$E4,$D3,$94,$F0,$24,$F0,$6C,$F0,$6C,$F0,$9D,$12,$74,$D8,$D4,$60
	.dc.b	$8E,$09,$3F,$E5,$6A,$FD,$43,$67,$F2,$ED,$9F,$C4,$34,$FB,$65,$8F
	.dc.b	$A0,$D8,$F8,$ED,$7E,$03,$4F,$59,$6B,$D0,$D9,$F3,$B6,$7C,$F1,$FE
	.dc.b	$65,$AF,$3B,$4F,$39,$63,$CE,$D9,$F3,$B4,$FC,$76,$0F,$A1,$FE,$47
	.dc.b	$5F,$EC,$B6,$C6,$67,$FD,$A7,$1A,$EB,$FF,$6B,$D7,$FB,$49,$AF,$FB
	.dc.b	$71,$FF,$6F,$3F,$EE,$0D,$1D,$9F,$EC,$EE,$BF,$F3,$F6,$1F,$E4,$7F
	.dc.b	$02,$DF,$F9,$05,$AB,$AB,$FD,$A8,$FF,$9A,$17,$81,$A7,$3A,$FF,$AE
	.dc.b	$1B,$7F,$EE,$E7,$75,$FE,$99,$7F,$D3,$CD,$AE,$47,$3D,$FA,$6F,$E2
	.dc.b	$DD,$F3,$94,$CD,$F3,$F1,$EB,$E5,$AF,$24,$79,$7B,$FF,$DC,$D3,$FE
	.dc.b	$DC,$C9,$18,$8F,$F7,$DC,$63,$4F,$F7,$9C,$E6,$7D,$5D,$FF,$EE,$A2
	.dc.b	$77,$FF,$BA,$A3,$FB,$01,$FE,$E7,$3F,$DD,$66,$3B,$FF,$DD,$8F,$FB
	.dc.b	$83,$7F,$F5,$D2,$EC,$FF,$75,$CA,$1D,$FF,$EE,$C7,$81,$FE,$98,$1D
	.dc.b	$DF,$EE,$97,$39,$FF,$B8,$A3,$FD,$BF,$0E,$EA,$7F,$BA,$B1,$F0,$BA
	.dc.b	$ED,$91,$83,$CA,$FF,$B9,$1F,$24,$F9,$7C,$1F,$F6,$C4,$7F,$B4,$FB
	.dc.b	$BF,$DB,$B3,$FE,$AC,$5D,$5A,$8C,$77,$7E,$7A,$77,$1F,$EE,$D8,$C1
	.dc.b	$FA,$2C,$DC,$9F,$B4,$C3,$F5,$15,$FE,$FF,$8C,$36,$3F,$8B,$47,$F7
	.dc.b	$73,$30,$9E,$73,$A3,$FB,$4C,$EA,$33,$3F,$C9,$23,$94,$FF,$02,$3C
	.dc.b	$31,$CE,$38,$1F,$C9,$CD,$BF,$C1,$8F,$21,$CE,$3B,$9C,$DC,$CF,$F1
	.dc.b	$39,$E2,$FF,$0F,$31,$3B,$9F,$03,$E9,$7F,$B8,$4F,$8A,$31,$3F,$D9
	.dc.b	$23,$13,$E1,$37,$3C,$E7,$DE,$76,$8C,$9E,$93,$B9,$C5,$E9,$1C,$67
	.dc.b	$E0,$3E,$C9,$E1,$3D,$23,$E3,$B6,$3F,$AF,$CF,$92,$3C,$07,$48,$E3
	.dc.b	$1D,$1F,$1C,$7C,$39,$E7,$3C,$5A,$78,$63,$53,$F3,$23,$2F,$27,$C7
	.dc.b	$CF,$D6,$77,$3E,$E7,$C7,$3F,$DF,$EB,$B4,$F3,$21,$FD,$BD,$CB,$C6
	.dc.b	$3A,$9F,$48,$AF,$C9,$6C,$F6,$8E,$6F,$F7,$19,$E6,$FC,$78,$E8,$7C
	.dc.b	$89,$FA,$07,$57,$07,$C8,$70,$F9,$5D,$BE,$13,$71,$F0,$0E,$C4,$9F
	.dc.b	$C8,$CF,$98,$F9,$31,$B8,$F9,$11,$F1,$4E,$B1,$EB,$9F,$01,$E3,$3C
	.dc.b	$06,$1E,$51,$E6,$7E,$03,$C8,$EE,$75,$8E,$51,$D4,$DB,$88,$DB,$EC
	.dc.b	$7D,$0F,$CC,$9F,$96,$48,$ED,$1F,$B4,$3F,$D2,$91,$FC,$3C,$91,$FE
	.dc.b	$91,$6C,$7F,$20,$38,$1F,$DF,$4E,$31,$FC,$DE,$4F,$F2,$6E,$77,$C9
	.dc.b	$7F,$8C,$9B,$1F,$E3,$26,$D7,$F8,$ED,$B5,$FE,$17,$6C,$FF,$10,$6C
	.dc.b	$FE,$49,$AF,$E1,$83,$89,$FC,$53,$6F,$CB,$3F,$C8,$EB,$CD,$7E,$52
	.dc.b	$3E,$56,$C7,$95,$B7,$E5,$6C,$F9,$5B,$3E,$56,$BE,$5B,$F1,$4F,$C7
	.dc.b	$3C,$27,$72,$47,$76,$D7,$76,$9D,$CB,$1D,$DA,$F7,$AF,$0C,$FC,$26
	.dc.b	$DF,$C2,$8F,$87,$3F,$22,$73,$5E,$58,$FE,$E5,$7E,$87,$E4,$18,$9D
	.dc.b	$B8,$39,$4F,$8A,$BC,$0D,$BF,$03,$F0,$27,$DC,$49,$F7,$36,$3D,$CD
	.dc.b	$3D,$C5,$3D,$CF,$53,$D0,$F8,$8D,$47,$89,$B3,$E2,$7D,$53,$C5,$F5
	.dc.b	$5F,$BE,$BE,$2B,$67,$D1,$1E,$62,$3E,$66,$CF,$99,$AF,$98,$F0,$C7
	.dc.b	$9A,$7A,$47,$88,$C9,$E2,$BF,$9B,$5E,$26,$D7,$88,$F0,$12,$7C,$0D
	.dc.b	$7C,$13,$E2,$3E,$11,$6F,$E1,$47,$C7,$8C,$4F,$96,$BB,$CF,$96,$72
	.dc.b	$F9,$67,$E4,$47,$95,$A7,$96,$3C,$64,$3C,$6F,$96,$3C,$6D,$AF,$1D
	.dc.b	$FC,$4A,$F2,$4E,$4F,$21,$D9,$F2,$39,$7C,$87,$CA,$AF,$23,$63,$C8
	.dc.b	$F8,$8F,$24,$78,$9B,$5E,$2B,$F8,$31,$E2,$33,$5E,$27,$A1,$E2,$32
	.dc.b	$78,$AB,$E3,$9E,$22,$C7,$89,$F0,$47,$8C,$8E,$09,$5E,$39,$F8,$13
	.dc.b	$E3,$37,$19,$7C,$66,$67,$8C,$F9,$DF,$0C,$79,$CF,$07,$6F,$FA,$A1
	.dc.b	$FF,$53,$76,$FF,$3F,$7F,$CE,$51,$FE,$AA,$4E,$DF,$E5,$FF,$25,$FE
	.dc.b	$80,$C8,$D2,$47,$FA,$BC,$FD,$07,$6F,$FA,$6C,$FF,$54,$77,$7F,$A8
	.dc.b	$23,$FD,$48,$E1,$FF,$52,$1F,$EA,$7D,$74,$FF,$AC,$7B,$BF,$D4,$CF
	.dc.b	$FA,$AF,$5F,$F5,$23,$FA,$DE,$EF,$F5,$1E,$BF,$EB,$3E,$EB,$BF,$FA
	.dc.b	$CE,$FF,$D4,$B7,$FE,$B3,$ED,$FF,$4F,$3F,$E9,$1E,$EF,$F5,$13,$FE
	.dc.b	$A7,$8C,$BF,$EC,$36,$C3,$BB,$FD,$47,$3F,$EA,$D6,$BF,$EA,$22,$1F
	.dc.b	$37,$B3,$FD,$46,$FF,$A7,$FB,$7F,$D2,$E7,$FA,$89,$FF,$50,$F8,$5F
	.dc.b	$DC,$63,$FD,$3D,$1F,$EA,$5E,$DD,$E7,$FE,$AA,$7F,$D3,$AF,$FA,$66
	.dc.b	$3F,$8D,$F8,$58,$7F,$D5,$1D,$BF,$E9,$E7,$FD,$55,$DD,$78,$3F,$A1
	.dc.b	$C3,$BB,$FD,$52,$67,$0F,$F0,$13,$FC,$22,$5C,$94,$FD,$CF,$85,$FE
	.dc.b	$2A,$BF,$F0,$24,$6F,$8B,$F8,$7F,$73,$C3,$FF,$51,$13,$2F,$C8,$5A
	.dc.b	$75,$18,$5A,$F2,$39,$D5,$9F,$E5,$9E,$8F,$AA,$FF,$AC,$B5,$FF,$02
	.dc.b	$B7,$B0,$D3,$F9,$1F,$D1,$7F,$B4,$EB,$99,$FD,$31,$C5,$7F,$46,$9D
	.dc.b	$F1,$7F,$18,$6D,$70,$9F,$EC,$95,$C3,$6F,$E9,$75,$A9,$CD,$D9,$CD
	.dc.b	$F7,$8C,$3F,$ED,$0B,$E4,$FF,$8B,$6F,$37,$FE,$E7,$9C,$B6,$BC,$95
	.dc.b	$CC,$91,$E0,$7C,$D7,$FE,$02,$AE,$32,$69,$8C,$3E,$3D,$72,$AE,$17
	.dc.b	$E2,$BF,$9B,$39,$3D,$11,$9A,$E3,$3E,$12,$57,$2A,$F1,$9C,$0E,$51
	.dc.b	$EB,$72,$D3,$8B,$B7,$EA,$3C,$1F,$55,$78,$70,$F3,$61,$EC,$8F,$8B
	.dc.b	$3C,$8E,$97,$E3,$79,$99,$39,$DF,$5C,$3D,$73,$B2,$47,$8D,$FB,$73
	.dc.b	$C6,$FC,$13,$EA,$8E,$66,$6F,$C7,$5C,$A3,$C7,$3C,$CF,$C0,$1E,$DA
	.dc.b	$E7,$1E,$69,$E0,$F4,$6D,$F9,$AB,$94,$F4,$21,$F6,$0F,$13,$ED,$C3
	.dc.b	$E3,$E1,$F2,$63,$C0,$D9,$D9,$F1,$6B,$B3,$D3,$0F,$2E,$1F,$08,$F2
	.dc.b	$B8,$BF,$5C,$F2,$9F,$3E,$1E,$09,$F1,$1E,$02,$CE,$DF,$5B,$E9,$9E
	.dc.b	$B8,$79,$5B,$3C,$1B,$3E,$7A,$EC,$D9,$EC,$D3,$B1,$F1,$B0,$F6,$BF
	.dc.b	$36,$3C,$38,$7E,$96,$B9,$DF,$AA,$79,$B6,$B9,$9C,$4F,$E2,$B1,$A3
	.dc.b	$D7,$7D,$AF,$E3,$4F,$86,$FA,$D6,$9B,$19,$8E,$4D,$3B,$CF,$96,$33
	.dc.b	$1E,$13,$E4,$3E,$AC,$3E,$43,$4F,$90,$58,$F8,$98,$7C,$A3,$D8,$58
	.dc.b	$F2,$1D,$AB,$C1,$3E,$2A,$FC,$EC,$F3,$AE,$2D,$38,$CE,$DF,$D3,$57
	.dc.b	$C2,$8F,$46,$1F,$16,$FE,$0C,$7A,$8F,$C8,$3E,$A6,$D7,$A9,$A7,$A9
	.dc.b	$F0,$3F,$1E,$FD,$4D,$8C,$BE,$0C,$3B,$E1,$EC,$AD,$CE,$6B,$E0,$5F
	.dc.b	$BA,$3B,$CF,$B9,$B3,$EE,$8F,$C9,$C7,$FB,$B6,$BF,$A6,$B6,$BF,$AC
	.dc.b	$46,$23,$F8,$F1,$9A,$FE,$CF,$19,$8F,$E4,$86,$4F,$EA,$3C,$5F,$C8
	.dc.b	$09,$3F,$C2,$5B,$1F,$C5,$5B,$7F,$B6,$6C,$FE,$79,$B3,$FB,$D6,$BF
	.dc.b	$8B,$A4,$61,$FE,$92,$DB,$FE,$DA,$DA,$FE,$D8,$DA,$F3,$C7,$94,$8F
	.dc.b	$95,$B7,$E5,$69,$E5,$2C,$79,$5A,$F9,$67,$B4,$FF,$5F,$6C,$FF,$6D
	.dc.b	$6C,$FF,$6B,$6B,$FD,$8D,$B7,$EE,$9E,$AF,$B9,$B7,$EE,$69,$EE,$2D
	.dc.b	$7B,$8F,$61,$2F,$D8,$D8,$F6,$34,$F6,$14,$F6,$3D,$23,$D8,$D3,$D8
	.dc.b	$58,$F6,$35,$F6,$36,$BD,$91,$D5,$F6,$36,$BD,$8D,$9F,$63,$4F,$65
	.dc.b	$7A,$89,$1E,$A6,$C7,$A9,$AF,$A8,$E4,$FA,$9B,$1E,$A6,$9E,$A2,$9E
	.dc.b	$A2,$D7,$AA,$7A,$BE,$A6,$D7,$A9,$B5,$EA,$6D,$7A,$A3,$D0,$47,$D0
	.dc.b	$D7,$D1,$1C,$0F,$41,$63,$D0,$D8,$F4,$35,$F4,$36,$BD,$13,$D5,$F4
	.dc.b	$36,$BD,$0D,$3D,$05,$AF,$41,$E6,$21,$E6,$75,$5E,$66,$9E,$62,$C7
	.dc.b	$99,$AF,$99,$B5,$E6,$8E,$AF,$99,$B7,$E6,$6C,$F9,$9A,$79,$8F,$21
	.dc.b	$F6,$CF,$E4,$51,$FC,$80,$91,$FC,$7B,$D9,$FE,$46,$47,$07,$FA,$23
	.dc.b	$BF,$FD,$A8,$4F,$31,$96,$7C,$47,$85,$FE,$8E,$E8,$FE,$2F,$98,$6A
	.dc.b	$CB,$10,$DD,$F4,$38,$AD,$4F,$19,$CB,$CE,$32,$AF,$BB,$DA,$FA,$C3
	.dc.b	$A3,$A4,$63,$7B,$5E,$5A,$E3,$7D,$1E,$11,$FE,$E8,$3F,$DD,$A4,$3B
	.dc.b	$1F,$ED,$BF,$07,$FD,$D3,$3E,$4E,$D4,$7F,$BB,$E7,$FD,$D8,$7F,$BB
	.dc.b	$CF,$F7,$89,$83,$98,$8F,$F6,$0B,$FE,$F1,$3F,$DE,$04,$36,$B9,$FF
	.dc.b	$DD,$F1,$87,$FD,$E6,$FF,$BB,$FB,$FD,$EF,$FB,$B6,$FF,$D8,$75,$FE
	.dc.b	$F2,$8F,$A6,$64,$FF,$7A,$46,$BB,$BF,$DD,$F1,$F6,$27,$FD,$E9,$E3
	.dc.b	$E4,$FF,$77,$C7,$DF,$EF,$A9,$FF,$68,$BF,$EA,$22,$19,$3F,$DE,$67
	.dc.b	$E1,$38,$FF,$DE,$6F,$E8,$CF,$F7,$D6,$DF,$EF,$33,$FD,$E5,$8F,$FB
	.dc.b	$B3,$7F,$C3,$BF,$84,$E0,$FF,$6D,$CE,$8F,$6F,$7F,$FB,$BE,$FF,$23
	.dc.b	$75,$FC,$B7,$7F,$FB,$D2,$BF,$1C,$70,$2C,$7E,$62,$3F,$DE,$91,$88
	.dc.b	$FC,$EF,$77,$FB,$DA,$31,$8F,$FB,$E9,$BC,$1F,$EF,$77,$33,$6B,$F0
	.dc.b	$3D,$DE,$86,$F7,$7F,$BC,$AB,$B4,$E7,$0F,$D0,$BB,$AF,$B2,$D7,$F5
	.dc.b	$AD,$7F,$64,$D7,$FD,$4A,$7F,$86,$C8,$7F,$88,$8F,$E2,$7C,$BF,$DD
	.dc.b	$0F,$E2,$0F,$F1,$73,$F8,$7C,$FF,$1F,$6D,$7F,$1F,$69,$FC,$7C,$A7
	.dc.b	$13,$F8,$A9,$7B,$3E,$F1,$FC,$7D,$C5,$7F,$87,$CF,$0C,$64,$DB,$99
	.dc.b	$FE,$45,$3D,$27,$31,$FE,$0A,$9E,$3B,$F2,$6C,$70,$9D,$D7,$F7,$59
	.dc.b	$CB,$FC,$AE,$33,$7F,$D8,$27,$2F,$C1,$9F,$3D,$77,$3A,$5F,$76,$9D
	.dc.b	$EB,$8B,$F1,$E7,$89,$E1,$76,$F8,$63,$6F,$CB,$37,$5E,$39,$DB,$E3
	.dc.b	$BF,$3C,$7F,$5F,$AE,$35,$C2,$3C,$13,$F2,$E7,$2F,$09,$EC,$E6,$7A
	.dc.b	$9C,$27,$67,$B6,$7D,$07,$C7,$BC,$5F,$8D,$B5,$E0,$6C,$F6,$6B,$F8
	.dc.b	$36,$DF,$E1,$1B,$7F,$7E,$7B,$57,$CB,$3C,$AF,$A2,$3C,$B1,$98,$F2
	.dc.b	$DF,$8E,$30,$D8,$F9,$D1,$C9,$F8,$C7,$95,$D9,$E9,$2B,$93,$89,$E4
	.dc.b	$2C,$75,$3E,$A4,$F9,$23,$33,$F5,$4F,$05,$FD,$18,$F8,$51,$D9,$A7
	.dc.b	$03,$C1,$5D,$23,$8D,$7C,$8B,$F8,$F1,$F9,$3A,$F2,$56,$DB,$1F,$6C
	.dc.b	$E2,$77,$BF,$C0,$46,$8F,$99,$1A,$7C,$47,$EB,$EF,$D9,$5D,$4F,$CA
	.dc.b	$47,$13,$F3,$A7,$F8,$4A,$7F,$D5,$B3,$FE,$12,$8F,$EA,$27,$F9,$3C
	.dc.b	$D1,$8A,$FF,$34,$B8,$33,$5F,$CF,$FD,$4F,$E2,$BC,$06,$93,$0F,$E0
	.dc.b	$B1,$96,$C6,$6B,$FC,$51,$39,$6C,$66,$BF,$7D,$39,$6C,$66,$BF,$75
	.dc.b	$39,$6C,$66,$BF,$71,$39,$6C,$65,$FE,$49,$19,$6C,$66,$FF,$B1,$4E
	.dc.b	$5B,$19,$C3,$FB,$0D,$FF,$8B,$2B,$51,$8B,$FF,$15,$56,$EB,$FA,$C5
	.dc.b	$69,$B1,$99,$FF,$09,$CE,$5B,$19,$C3,$E8,$57,$CA,$24,$FC,$86,$C7
	.dc.b	$C6,$69,$F1,$0B,$1F,$09,$AF,$C0,$6E,$1E,$E6,$E1,$EB,$9F,$F2,$45
	.dc.b	$F8,$E7,$E0,$36,$3D,$F1,$C0,$F6,$95,$F5,$B6,$FD,$71,$E9,$23,$E9
	.dc.b	$6C,$7A,$5A,$7A,$4A,$7A,$4B,$7E,$96,$DF,$A7,$0F,$99,$7E,$03,$B1
	.dc.b	$2F,$B3,$6B,$B3,$5E,$CD,$BE,$CD,$BE,$D8,$7C,$3A,$F8,$F7,$EF,$3E
	.dc.b	$39,$5F,$8E,$79,$9F,$8E,$D3,$E3,$96,$BD,$11,$E6,$24,$79,$9A,$F9
	.dc.b	$9B,$7E,$66,$DF,$99,$F1,$4F,$98,$EC,$F9,$9A,$79,$AB,$DD,$1E,$66
	.dc.b	$C7,$99,$A7,$99,$F2,$10,$F2,$16,$BC,$8D,$AF,$23,$E0,$3C,$93,$D2
	.dc.b	$BC,$8D,$8F,$24,$7B,$0F,$21,$5F,$23,$63,$C9,$1E,$22,$57,$89,$B5
	.dc.b	$E2,$9E,$D1,$E2,$AF,$09,$1F,$66,$1F,$C2,$4F,$C6,$C6,$63,$E6,$1F
	.dc.b	$E7,$E8,$FE,$3C,$DC,$3F,$D5,$6F,$CF,$3F,$D0,$7D,$9F,$EC,$0E,$BF
	.dc.b	$F5,$EC,$7F,$36,$87,$67,$FB,$0A,$31,$3A,$7F,$D1,$7E,$B7,$F3,$37
	.dc.b	$F9,$E8,$70,$FF,$40,$6B,$C2,$FF,$D0,$FD,$7F,$EB,$4C,$BF,$8E,$76
	.dc.b	$7F,$A6,$B3,$FF,$61,$DF,$FA,$A2,$B4,$FF,$AE,$F9,$7F,$D3,$F1,$88
	.dc.b	$FF,$63,$6D,$89,$FF,$5A,$6D,$FE,$87,$F1,$7A,$18,$CB,$FD,$3D,$E2
	.dc.b	$DC,$3F,$D7,$C7,$FA,$E2,$BF,$D7,$DC,$BF,$EC,$0A,$DC,$FF,$AF,$4F
	.dc.b	$F5,$91,$FE,$B7,$9F,$F4,$A7,$8B,$FE,$B4,$8F,$F5,$F7,$75,$ED,$FF
	.dc.b	$5D,$47,$FA,$2D,$47,$FA,$E1,$FF,$5B,$47,$FA,$C7,$C5,$FF,$75,$BF
	.dc.b	$EB,$62,$1F,$EB,$A8,$FF,$5A,$BF,$DB,$1A,$E7,$BF,$FD,$76,$FF,$B0
	.dc.b	$3B,$BF,$D6,$E7,$FA,$EB,$A7,$F9,$74,$7F,$AD,$A3,$FD,$F5,$E4,$FF
	.dc.b	$BB,$E3,$FD,$6D,$A7,$FA,$CD,$FF,$6C,$9F,$EB,$73,$FD,$7D,$C3,$CB
	.dc.b	$86,$4F,$F5,$0E,$CB,$5C,$3B,$FF,$D6,$4F,$FB,$DE,$7F,$D5,$8A,$3F
	.dc.b	$D6,$8F,$6E,$FD,$9F,$EB,$27,$FD,$D3,$C3,$FE,$AF,$7F,$AC,$F8,$D9
	.dc.b	$85,$A6,$47,$FB,$A6,$3F,$D6,$A7,$FA,$D0,$B1,$FD,$8B,$C5,$FD,$2B
	.dc.b	$C7,$A1,$18,$84,$62,$7F,$3E,$D8,$FF,$0D,$6F,$BE,$F3,$49,$5F,$A6
	.dc.b	$F0,$1E,$6F,$56,$C6,$63,$04,$7F,$D8,$BE,$77,$F6,$59,$DE,$1F,$EC
	.dc.b	$C8,$FE,$A1,$EA,$CA,$FF,$6C,$7E,$76,$F2,$31,$7F,$E0,$0A,$F6,$68
	.dc.b	$69,$2F,$DA,$FF,$73,$B7,$9F,$CB,$7F,$9F,$87,$CF,$9C,$DF,$83,$E6
	.dc.b	$F8,$1C,$7C,$AB,$19,$3F,$83,$CA,$7E,$73,$63,$31,$83,$C5,$E0,$72
	.dc.b	$B6,$1E,$25,$69,$F4,$78,$13,$0F,$87,$E4,$DB,$E3,$1F,$87,$78,$CF
	.dc.b	$88,$91,$E6,$6C,$70,$3E,$07,$81,$CF,$1F,$16,$3F,$BC,$BF,$A3,$E0
	.dc.b	$7C,$93,$E5,$77,$D8,$C9,$F0,$A3,$E0,$91,$C1,$F6,$BE,$23,$86,$3E
	.dc.b	$13,$DE,$FD,$5F,$9B,$93,$F8,$0F,$9B,$B5,$BC,$55,$F3,$89,$5F,$49
	.dc.b	$B3,$BE,$3E,$39,$FB,$1F,$C1,$7C,$BC,$BF,$C6,$FC,$DD,$BE,$D3,$F6
	.dc.b	$89,$3F,$79,$B1,$E3,$F9,$2E,$18,$7C,$87,$F1,$DF,$25,$2F,$F3,$3F
	.dc.b	$25,$6F,$D5,$3F,$86,$24,$FE,$29,$B1,$EF,$F9,$AE,$18,$7D,$47,$FC
	.dc.b	$99,$DE,$F9,$7A,$7E,$77,$E3,$EE,$3F,$C2,$0F,$C2,$4E,$DF,$EB,$AD
	.dc.b	$C3,$FD,$31,$5F,$BF,$24,$F1,$8F,$E2,$52,$F5,$BA,$C7,$D9,$8C,$4E
	.dc.b	$E1,$1D,$71,$FC,$E3,$A8,$E1,$7E,$AB,$E7,$5A,$6D,$FA,$1B,$5E,$26
	.dc.b	$BD,$9B,$7D,$6B,$BD,$F5,$BF,$24,$F8,$DE,$86,$23,$57,$E4,$8D,$DF
	.dc.b	$42,$3B,$32,$58,$F8,$51,$A7,$91,$88,$D5,$FB,$5A,$78,$A3,$71,$96
	.dc.b	$C6,$63,$D6,$F8,$23,$27,$C3,$79,$B4,$F0,$12,$38,$B6,$3A,$11,$FB
	.dc.b	$B5,$F0,$A7,$C4,$49,$FC,$4C,$65,$E1,$3F,$45,$F0,$B4,$EC,$78,$E7
	.dc.b	$C2,$73,$71,$1C,$AF,$EA,$46,$08,$73,$9E,$33,$BB,$F1,$4F,$A2,$3E
	.dc.b	$24,$F5,$8E,$11,$E8,$32,$71,$3C,$F5,$F8,$77,$11,$F3,$0E,$27,$8A
	.dc.b	$34,$F7,$3D,$A6,$88,$FA,$9B,$5E,$67,$F4,$77,$67,$9C,$70,$8C,$4E
	.dc.b	$5B,$5E,$89,$F0,$5F,$B6,$F8,$92,$B7,$3E,$48,$F1,$1A,$BE,$97,$CE
	.dc.b	$7C,$AF,$C2,$8E,$37,$F3,$A7,$E0,$47,$07,$A3,$63,$C4,$D9,$F9,$31
	.dc.b	$E5,$BE,$71,$F7,$5B,$5F,$2D,$AF,$C1,$8F,$25,$F9,$DF,$21,$8B,$FD
	.dc.b	$65,$F8,$23,$35,$E4,$6C,$66,$3E,$D5,$73,$7E,$25,$F8,$23,$E4,$57
	.dc.b	$53,$C3,$1F,$2A,$7F,$B5,$1F,$9B,$2C,$FF,$19,$6D,$FF,$89,$49,$5F
	.dc.b	$CB,$1B,$1C,$8F,$C2,$FC,$77,$E0,$49,$3F,$01,$B1,$F6,$5B,$87,$DF
	.dc.b	$6C,$FE,$9D,$B3,$F9,$56,$BD,$A3,$A1,$1E,$8D,$9E,$8D,$3A,$16,$3A
	.dc.b	$35,$E8,$72,$24,$72,$6C,$72,$69,$C8,$A7,$28,$E0,$48,$E0,$D8,$E0
	.dc.b	$D7,$84,$68,$8E,$9A,$E8,$C1,$0C,$12,$FF,$53,$5F,$21,$B3,$EF,$6C
	.dc.b	$FA,$1B,$5E,$36,$CF,$81,$57,$F7,$43,$A1,$63,$A3,$6B,$A3,$6B,$A3
	.dc.b	$67,$A3,$67,$D5,$5D,$A7,$FB,$CC,$FF,$9D,$48,$E0,$FE,$BF,$F0,$DC
	.dc.b	$6B,$FB,$9B,$68,$D2,$4F,$F6,$6A,$FF,$06,$C6,$2B,$FB,$F4,$62,$74
	.dc.b	$76,$2C,$7F,$80,$E7,$85,$FF,$80,$2B,$84,$E6,$B8,$3D,$1B,$3F,$E0
	.dc.b	$6A,$DB,$C6,$32,$7F,$86,$A3,$2F,$F8,$72,$AD,$7F,$85,$A6,$BF,$E1
	.dc.b	$28,$B7,$F0,$2B,$9C,$FF,$87,$27,$35,$A6,$CE,$A7,$93,$6B,$94,$7E
	.dc.b	$F5,$AF,$F0,$6A,$FF,$54,$12,$7F,$C4,$9F,$9E,$7F,$AE,$DE,$8D,$9F
	.dc.b	$F3,$13,$FE,$AE,$8F,$E1,$10,$8C,$3F,$C2,$DA,$FF,$16,$6B,$FE,$81
	.dc.b	$5E,$67,$FA,$FD,$B3,$D1,$B3,$CE,$75,$E6,$4A,$D3,$67,$C1,$E5,$F3
	.dc.b	$24,$F3,$7B,$79,$72,$3B,$36,$7D,$1E,$97,$8C,$93,$E3,$7C,$DF,$35
	.dc.b	$23,$CC,$D9,$F8,$1F,$35,$EB,$24,$FA,$DF,$77,$0C,$8F,$73,$67,$E4
	.dc.b	$70,$FC,$32,$4F,$C3,$7E,$37,$04,$8F,$8C,$D9,$FA,$1B,$FC,$B2,$4F
	.dc.b	$CB,$7E,$77,$C6,$C8,$F9,$CD,$9F,$B1,$F1,$FF,$4C,$93,$F4,$DF,$AD
	.dc.b	$F1,$D2,$3E,$B3,$67,$F0,$1F,$19,$F6,$C9,$3F,$6D,$FB,$DF,$19,$23
	.dc.b	$EF,$36,$7F,$11,$F1,$9F,$83,$24,$FE,$0D,$FF,$7F,$61,$23,$FD,$0A
	.dc.b	$D9,$FC,$F6,$1F,$ED,$57,$FD,$16,$D7,$FD,$0A,$D7,$FD,$02,$DE,$66
	.dc.b	$14,$97,$FE,$92,$78,$7C,$FB,$E3,$69,$27,$F4,$71,$97,$F4,$BF,$04
	.dc.b	$F8,$17,$A5,$86,$E7,$FD,$C6,$5F,$04,$F5,$F5,$7F,$A3,$CB,$F5,$F7
	.dc.b	$FA,$F6,$F6,$F8,$E3,$47,$ED,$3E,$99,$E8,$70,$6C,$66,$30,$74,$EB
	.dc.b	$FD,$9C,$7F,$A9,$7C,$87,$C0,$FF,$BD,$FC,$2F,$F8,$01,$18,$9F,$F5
	.dc.b	$55,$70,$7F,$DC,$CB,$C1,$E1,$C3,$FE,$FF,$8D,$5A,$31,$E1,$6A,$C2
	.dc.b	$B9,$3F,$EB,$7F,$7D,$EA,$EE,$19,$3E,$07,$0D,$8E,$13,$BA,$7A,$B6
	.dc.b	$24,$F5,$8D,$7A,$BB,$64,$78,$63,$96,$FD,$BE,$06,$4F,$96,$3B,$7A
	.dc.b	$DA,$8C,$1E,$97,$C5,$EA,$B8,$3F,$94,$E6,$8C,$57,$F8,$CE,$31,$3A
	.dc.b	$9C,$CF,$F9,$02,$5E,$B3,$E1,$39,$B6,$79,$35,$E0,$D7,$FC,$81,$F0
	.dc.b	$5F,$8C,$4D,$23,$FC,$FB,$AB,$EC,$3E,$9E,$4F,$F2,$8C,$63,$2F,$D7
	.dc.b	$57,$DA,$F8,$2F,$B5,$7F,$70,$FE,$8B,$3F,$B5,$E5,$E1,$87,$F9,$69
	.dc.b	$B3,$BC,$2B,$F8,$48,$AF,$E3,$67,$FD,$F4,$D3,$C8,$7E,$2B,$8D,$F0
	.dc.b	$5C,$39,$F8,$F9,$F2,$F8,$C5,$BE,$27,$E4,$BC,$FE,$E6,$73,$FE,$18
	.dc.b	$F9,$17,$2D,$3C,$F9,$3E,$0B,$18,$FE,$72,$BE,$4E,$1F,$81,$E5,$E4
	.dc.b	$69,$F0,$C7,$EE,$27,$F6,$5F,$05,$F0,$BC,$6F,$93,$1F,$2F,$14,$E7
	.dc.b	$82,$4F,$73,$E8,$57,$F6,$AB,$F0,$BC,$B0,$FD,$7B,$99,$FA,$DF,$02
	.dc.b	$F8,$2F,$A2,$BE,$0A,$C6,$19,$3F,$E0,$29,$D3,$63,$F9,$74,$62,$FF
	.dc.b	$D1,$D9,$BE,$0B,$F9,$93,$6F,$FC,$16,$D9,$DC,$E8,$FC,$1F,$BC,$F8
	.dc.b	$2F,$9F,$18,$65,$FF,$85,$27,$4D,$8F,$E8,$11,$8A,$FF,$50,$2F,$82
	.dc.b	$C9,$2B,$93,$67,$F0,$B3,$FC,$2A,$7E,$74,$7F,$18,$8E,$51,$88,$46
	.dc.b	$DE,$3E,$07,$88,$93,$E7,$7C,$BC,$72,$3D,$4D,$9F,$3B,$F0,$BB,$FE
	.dc.b	$1B,$67,$C1,$69,$F7,$92,$3E,$23,$F0,$BB,$E4,$7C,$76,$DF,$C4,$7E
	.dc.b	$7F,$C1,$7D,$06,$CF,$68,$F9,$64,$7E,$83,$F3,$BE,$06,$47,$D3,$6D
	.dc.b	$7D,$07,$ED,$ED,$F7,$1B,$3E,$A8,$FA,$E4,$7E,$E3,$F6,$BE,$02,$47
	.dc.b	$DF,$6D,$7D,$C7,$F0,$FF,$01,$F8,$86,$CF,$C6,$8F,$C1,$91,$FC,$43
	.dc.b	$F8,$6F,$7E,$47,$E2,$DB,$5F,$88,$7F,$27,$EF,$FE,$51,$B3,$F4,$A3
	.dc.b	$F1,$E4,$7F,$28,$FF,$83,$FD,$CF,$DE,$7B,$9F,$BD,$7F,$E0,$03,$67
	.dc.b	$FE,$00,$5F,$F1,$4E,$84,$E6,$30,$F2,$9F,$C8,$4E,$9B,$5D,$1A,$74
	.dc.b	$E9,$FE,$07,$5C,$DB,$19,$3F,$84,$CF,$1B,$F1,$36,$7F,$85,$B6,$76
	.dc.b	$71,$9F,$E1,$B5,$D6,$3F,$89,$BE,$78,$F4,$C7,$91,$A4,$AF,$1B,$E0
	.dc.b	$8E,$8A,$FD,$CE,$4E,$11,$D8,$F8,$75,$E1,$8F,$1C,$60,$FF,$4C,$CF
	.dc.b	$28,$FE,$AC,$75,$9F,$F2,$21,$27,$FD,$93,$5F,$EE,$D8,$FF,$6B,$12
	.dc.b	$32,$7F,$BB,$DC,$4F,$FB,$24,$FF,$7B,$38,$9F,$F6,$44,$62,$78,$BF
	.dc.b	$EE,$AF,$73,$51,$FE,$C9,$33,$1F,$EB,$E3,$FD,$EA,$E2,$FF,$DF,$F1
	.dc.b	$FE,$C7,$8C,$5F,$FB,$AE,$B9,$CE,$4F,$25,$F8,$8F,$F7,$ED,$70,$3A
	.dc.b	$4F,$09,$F0,$4E,$E7,$C9,$1E,$23,$9B,$F4,$5A,$77,$8E,$0F,$B6,$3B
	.dc.b	$3A,$3A,$BD,$23,$C7,$19,$BF,$6B,$A3,$A4,$F7,$9E,$B1,$92,$4F,$01
	.dc.b	$5E,$B6,$D7,$B8,$FC,$24,$F7,$9D,$1F,$26,$33,$5D,$AB,$8D,$79,$6F
	.dc.b	$4D,$AE,$8D,$3A,$3F,$02,$B9,$5F,$68,$DC,$7C,$03,$B5,$74,$6C,$68
	.dc.b	$F8,$C7,$C1,$2D,$7D,$D3,$F3,$77,$E6,$6D,$FF,$A7,$C9,$3F,$A2,$6C
	.dc.b	$7E,$88,$FC,$71,$C2,$3F,$86,$FB,$9F,$E1,$39,$FF,$0A,$36,$32,$7E
	.dc.b	$4D,$C4,$FF,$19,$6D,$FF,$84,$9B,$3F,$E1,$66,$CF,$F8,$51,$AF,$EF
	.dc.b	$DB,$66,$92,$7C,$04,$9F,$C4,$46,$5B,$19,$AF,$B5,$39,$6C,$64,$F9
	.dc.b	$73,$E1,$70,$FE,$40,$F3,$38,$7F,$1F,$3C,$E3,$8D,$79,$DD,$3F,$88
	.dc.b	$AF,$85,$18,$BF,$96,$D3,$E5,$DF,$A0,$93,$E6,$6C,$76,$BF,$04,$79
	.dc.b	$4F,$41,$A3,$17,$E3,$70,$7A,$27,$C0,$FD,$76,$D7,$9A,$37,$3E,$99
	.dc.b	$FB,$06,$8B,$1A,$BF,$5B,$63,$D7,$5C,$C8,$FA,$4C,$9C,$9F,$71,$A8
	.dc.b	$F9,$E7,$8A,$FA,$C6,$A7,$81,$27,$DF,$3A,$8C,$47,$BC,$F7,$12,$79
	.dc.b	$3D,$8F,$D4,$CF,$F9,$E1,$F6,$98,$9F,$E9,$31,$FE,$96,$F4,$3F,$DB
	.dc.b	$31,$EA,$9F,$F6,$DC,$61,$ED,$1F,$EA,$38,$C5,$7B,$65,$18,$9F,$3B
	.dc.b	$63,$E0,$CF,$FA,$2A,$31,$1F,$7A,$30,$FB,$CF,$09,$4E,$A5,$8E,$38
	.dc.b	$7E,$0E,$78,$D7,$A6,$B6,$DA,$DD,$FB,$E6,$4F,$F9,$7E,$72,$D8,$C9
	.dc.b	$FE,$AC,$8C,$B4,$CD,$78,$1F,$8A,$7A,$27,$91,$1C,$9E,$43,$93,$98
	.dc.b	$F9,$53,$F1,$09,$3E,$F3,$C6,$7F,$7E,$7E,$A3,$4F,$BC,$78,$4F,$E4
	.dc.b	$4F,$3F,$8E,$F3,$D7,$D9,$7F,$01,$E3,$70,$AF,$B2,$DA,$34,$C7,$89
	.dc.b	$2F,$F9,$75,$7E,$02,$30,$F8,$3E,$3B,$91,$27,$76,$8E,$DE,$2C,$9C
	.dc.b	$B6,$BD,$3E,$87,$8C,$93,$C9,$F4,$7E,$7A,$8A,$43,$CA,$E5,$F0,$FC
	.dc.b	$07,$B4,$8E,$0F,$2F,$17,$C1,$23,$83,$9F,$C7,$F8,$08,$E0,$F6,$F1
	.dc.b	$7C,$92,$38,$3E,$0F,$0F,$CD,$23,$83,$CF,$F1,$FE,$C2,$38,$3E,$4F
	.dc.b	$BF,$F5,$48,$E0,$F9,$BE,$FF,$D9,$23,$83,$E1,$FB,$7F,$20,$8E,$0F
	.dc.b	$AB,$EF,$FE,$04,$8E,$0F,$B3,$F0,$1F,$85,$23,$83,$E7,$FB,$FF,$50
	.dc.b	$8E,$0F,$F2,$BD,$FE,$3F,$A4,$D2,$47,$F9,$72,$F4,$47,$07,$DC,$F8
	.dc.b	$DF,$C0,$91,$C1,$F9,$4F,$80,$60,$F8,$0C,$37,$97,$FC,$0D,$2D,$1E
	.dc.b	$86,$AF,$F3,$BF,$36,$F9,$0F,$CF,$E0,$8C,$3A,$8C,$C6,$23,$F6,$3F
	.dc.b	$04,$7F,$87,$35,$FE,$18,$DF,$03,$F3,$51,$BF,$82,$7A,$D9,$DB,$84
	.dc.b	$62,$7F,$DE,$CD,$55,$C9,$46,$63,$0F,$FB,$B6,$3F,$DF,$DD,$9F,$BE
	.dc.b	$7F,$E0,$07,$90,$F8,$2F,$04,$7A,$63,$0F,$0F,$0F,$2A,$9E,$1E,$1E
	.dc.b	$91,$E4,$CB,$E0,$46,$32,$F3,$3C,$27,$E3,$47,$A7,$DF,$7C,$0F,$6C
	.dc.b	$9F,$05,$4F,$0C,$EA,$F8,$E3,$F3,$61,$E9,$EB,$09,$5A,$3E,$A4,$E9
	.dc.b	$C0,$5E,$9F,$6A,$93,$A9,$CB,$78,$3C,$DF,$05,$23,$06,$FD,$3F,$3C
	.dc.b	$C8,$C1,$CF,$D3,$FB,$2F,$F9,$66,$D1,$88,$FF,$5E,$C6,$36,$FE,$62
	.dc.b	$D8,$E1,$F0,$2F,$82,$CB,$63,$31,$83,$F3,$6B,$E0,$A9,$1F,$CF,$BF
	.dc.b	$A0,$EB,$FC,$AE,$1F,$E0,$B6,$F0,$7F,$9B,$E7,$F3,$DF,$05,$F8,$3B
	.dc.b	$FC,$F5,$69,$FD,$07,$35,$7F,$48,$D7,$F6,$37,$F4,$9F,$EA,$2D,$E6
	.dc.b	$D1,$C7,$95,$F0,$58,$9E,$1F,$05,$CC,$C5,$FE,$44,$FD,$83,$DF,$CF
	.dc.b	$93,$C0,$F0,$3E,$5D,$7F,$03,$1F,$84,$6D,$73,$18,$3E,$0B,$87,$3F
	.dc.b	$1E,$49,$87,$18,$B7,$C6,$DF,$05,$F0,$FC,$8F,$83,$39,$8C,$70,$4C
	.dc.b	$3F,$2B,$5F,$12,$BC,$4F,$0C,$FF,$80,$39,$7F,$81,$60,$F7,$FE,$41
	.dc.b	$F4,$7D,$DF,$E4,$98,$7F,$00,$8C,$BF,$DF,$27,$83,$63,$FC,$0D,$18
	.dc.b	$C3,$FD,$5D,$83,$D1,$FE,$78,$4B,$FF,$02,$C6,$5B,$EB,$FB,$CF,$81
	.dc.b	$36,$3F,$DB,$35,$FE,$C0,$9F,$E8,$C4,$8F,$EF,$7F,$1B,$FE,$00,$23
	.dc.b	$83,$FB,$FF,$1F,$F8,$10,$8E,$0F,$F0,$1F,$C1,$FF,$82,$08,$E0,$FF
	.dc.b	$03,$FC,$1F,$F8,$30,$8E,$0F,$F0,$5F,$91,$FE,$10,$23,$83,$FC,$1F
	.dc.b	$E1,$FF,$85,$08,$E0,$FF,$09,$F2,$FF,$86,$08,$E0,$FF,$0B,$FC,$87
	.dc.b	$F8,$70,$8E,$0F,$F0,$DF,$85,$FE,$20,$23,$83,$FC,$3F,$F2,$1F,$E2
	.dc.b	$42,$38,$3F,$C4,$7F,$21,$FE,$28,$23,$83,$FC,$4F,$CF,$FE,$2C,$23
	.dc.b	$83,$FC,$57,$E4,$7F,$8C,$08,$E0,$FF,$17,$FC,$7F,$F8,$D0,$8E,$0F
	.dc.b	$F1,$9F,$37,$F8,$E0,$8E,$0F,$F2,$5E,$5F,$94,$E3,$FF,$7D,$35,$FE
	.dc.b	$D0,$1C,$3F,$DA,$DA,$FF,$6F,$5F,$0D,$FE,$5D,$6C,$F6,$8C,$19,$F8
	.dc.b	$69,$1E,$26,$D7,$8B,$CA,$FE,$31,$7F,$D5,$2B,$FA,$3B,$FD,$39,$B1
	.dc.b	$A3,$C0,$47,$DA,$D9,$FD,$CF,$C5,$6E,$7F,$7A,$D8,$CC,$74,$32,$57
	.dc.b	$27,$F1,$3A,$F2,$38,$EC,$FF,$01,$46,$1F,$E0,$B5,$D5,$B5,$CD,$B5
	.dc.b	$CE,$FF,$8A,$56,$EB,$FC,$31,$D4,$8C,$57,$F8,$76,$31,$3E,$68,$F1
	.dc.b	$3F,$56,$B9,$B9,$BE,$E7,$9A,$33,$3C,$0E,$97,$D9,$5F,$67,$46,$23
	.dc.b	$51,$9A,$EB,$7E,$89,$C9,$DA,$33,$5F,$87,$BC,$56,$CF,$31,$E9,$2C
	.dc.b	$79,$1F,$E3,$B5,$FE,$5B,$23,$83,$F9,$BF,$B1,$D2,$7E,$33,$5F,$5B
	.dc.b	$70,$34,$93,$E0,$9D,$B6,$76,$D9,$DC,$FC,$49,$CB,$63,$38,$7C,$08
	.dc.b	$E6,$7F,$4A,$8F,$EA,$46,$AB,$EF,$B8,$7B,$1E,$C2,$47,$56,$CF,$F5
	.dc.b	$29,$E5,$3F,$02,$72,$7C,$D7,$C0,$E1,$D1,$CC,$C9,$27,$D1,$5E,$C7
	.dc.b	$B4,$F1,$9F,$C1,$36,$74,$FA,$8F,$4C,$73,$70,$F7,$9E,$6E,$4F,$BC
	.dc.b	$53,$E3,$BA,$3C,$75,$F1,$63,$55,$F9,$1A,$EB,$5B,$69,$F0,$5C,$4F
	.dc.b	$C0,$6B,$EA,$39,$D7,$58,$EF,$3E,$D8,$F9,$B3,$E4,$8E,$AF,$A2,$79
	.dc.b	$D7,$C2,$8E,$D3,$E1,$36,$F1,$32,$71,$73,$1C,$63,$47,$96,$7A,$D7
	.dc.b	$96,$BA,$4F,$23,$75,$C4,$F8,$8E,$5C,$4F,$A9,$B3,$F5,$AB,$AB,$4F
	.dc.b	$1C,$EC,$F0,$57,$9C,$E8,$E1,$F3,$47,$03,$FC,$27,$1F,$ED,$12,$38
	.dc.b	$24,$7F,$87,$23,$2D,$8F,$F1,$8F,$AF,$D6,$FF,$DB,$31,$F8,$F8,$C1
	.dc.b	$1C,$12,$7F,$0B,$19,$6C,$66,$FE,$BC,$E5,$B1,$93,$E9,$C6,$5A,$66
	.dc.b	$7E,$64,$65,$B1,$97,$E9,$C6,$5B,$19,$AF,$AD,$39,$6C,$66,$FF,$CA
	.dc.b	$D3,$96,$C6,$4F,$8A,$5A,$FF,$2A,$CE,$E3,$31,$F0,$09,$7E,$E6,$C7
	.dc.b	$B1,$A7,$A8,$A7,$A0,$B7,$E6,$6D,$F9,$1B,$3E,$0A,$EC,$47,$B3,$FE
	.dc.b	$68,$BE,$0E,$67,$FA,$E4,$62,$76,$7F,$62,$71,$1E,$59,$E5,$3D,$CD
	.dc.b	$C7,$27,$0E,$DE,$46,$1F,$64,$72,$7E,$4C,$7D,$59,$FF,$7E,$47,$03
	.dc.b	$FC,$E2,$E1,$EC,$62,$3F,$D0,$FF,$15,$E6,$8F,$ED,$2D,$C0,$D2,$4F
	.dc.b	$8E,$32,$F2,$24,$F9,$0F,$F4,$8C,$FF,$A6,$49,$1C,$5B,$1A,$8C,$1C
	.dc.b	$9F,$F1,$DC,$F6,$7F,$D1,$73,$EB,$7C,$F3,$FE,$8A,$8F,$F5,$37,$B1
	.dc.b	$D3,$0F,$F4,$FB,$67,$FC,$2A,$DA,$FF,$0A,$35,$FF,$04,$B6,$CF,$49
	.dc.b	$FF,$56,$B9,$31,$3F,$EA,$E7,$11,$A2,$47,$B4,$D3,$87,$DB,$1E,$B9
	.dc.b	$EF,$3F,$EA,$48,$C1,$FE,$C1,$F6,$7B,$DF,$FA,$E9,$FF,$27,$F8,$7F
	.dc.b	$E5,$92,$57,$F9,$99,$FF,$2E,$78,$92,$3F,$CD,$2D,$C7,$FC,$D3,$E7
	.dc.b	$FF,$98,$49,$3F,$E6,$E7,$FC,$C9,$E7,$C8,$FF,$38,$B6,$7F,$CE,$3E
	.dc.b	$77,$F9,$A4,$93,$FE,$76,$7F,$CD,$9F,$11,$23,$FC,$F2,$D9,$FF,$3C
	.dc.b	$FC,$47,$F9,$C4,$93,$FE,$7E,$7F,$CE,$9F,$1F,$23,$FD,$02,$D9,$FF
	.dc.b	$40,$F9,$1F,$E7,$92,$4F,$FA,$19,$FF,$3E,$7C,$4C,$8F,$F4,$4B,$67
	.dc.b	$FD,$13,$F1,$FF,$E8,$12,$4F,$FA,$39,$FF,$42,$7C,$54,$8F,$F4,$8B
	.dc.b	$67,$FD,$23,$F1,$5F,$E8,$92,$4F,$FA,$59,$FF,$46,$7C,$7C,$8F,$F4
	.dc.b	$CB,$67,$FD,$33,$F1,$7F,$E9,$12,$4F,$FA,$79,$FF,$4A,$7C,$7C,$8F
	.dc.b	$F5,$0B,$67,$FD,$3E,$3F,$3E,$3F,$A5,$6C,$EA,$34,$64,$D7,$A5,$23
	.dc.b	$52,$8C,$4F,$6F,$A3,$6F,$FB,$16,$DF,$56,$CE,$E7,$F6,$DF,$20,$F8
	.dc.b	$0A,$4B,$F2,$46,$5F,$37,$C2,$3E,$15,$F0,$98,$6E,$3F,$02,$5F,$0B
	.dc.b	$E7,$8C,$4E,$9F,$F6,$1E,$AF,$77,$E1,$47,$C7,$6C,$66,$3E,$0C,$6B
	.dc.b	$E4,$9E,$95,$24,$FC,$B8,$CB,$F3,$7E,$3F,$DC,$FD,$2F,$09,$AF,$D4
	.dc.b	$84,$E9,$4E,$9F,$B3,$19,$8C,$43,$E1,$9E,$0F,$39,$CC,$62,$7C,$17
	.dc.b	$DE,$31,$1D,$A3,$15,$C5,$D7,$83,$67,$F1,$3F,$0D,$F0,$DE,$07,$E3
	.dc.b	$3B,$9E,$8E,$E1,$18,$7C,$27,$2D,$B5,$4F,$81,$B1,$27,$62,$74,$E3
	.dc.b	$E1,$5E,$77,$7C,$1E,$C5,$3C,$53,$AA,$E1,$F0,$D8,$B9,$18,$3A,$79
	.dc.b	$3C,$26,$46,$0F,$06,$FD,$FD,$99,$18,$3C,$9B,$F8,$26,$46,$0F,$47
	.dc.b	$95,$F7,$1F,$DA,$E4,$8C,$47,$E0,$C4,$67,$8F,$F6,$CD,$8E,$3E,$53
	.dc.b	$E1,$B2,$D8,$CC,$60,$FC,$EA,$F8,$6A,$47,$F4,$2F,$E8,$BC,$3F,$C8
	.dc.b	$61,$FE,$08,$6F,$37,$F0,$C9,$FE,$17,$F0,$DF,$56,$FF,$35,$5A,$7F
	.dc.b	$39,$CB,$5F,$E2,$2D,$7F,$85,$5F,$DE,$7F,$64,$DE,$6D,$1C,$79,$5F
	.dc.b	$01,$89,$E1,$F0,$DC,$CC,$5F,$E5,$0F,$E3,$EF,$7F,$86,$93,$C0,$F0
	.dc.b	$3E,$5D,$5C,$D4,$F0,$77,$5A,$FC,$9C,$FE,$B9,$BC,$99,$C7,$C1,$39
	.dc.b	$7C,$59,$BD,$DF,$85,$EA,$7C,$09,$CC,$63,$83,$3C,$13,$2F,$D0,$DE
	.dc.b	$F1,$FD,$FB,$F1,$B0,$EB,$EE,$7F,$87,$F2,$34,$97,$FC,$F6,$32,$FE
	.dc.b	$B3,$E1,$9E,$EF,$CD,$AF,$AD,$18,$9F,$0F,$C2,$FA,$0F,$4F,$C3,$58
	.dc.b	$C9,$FD,$27,$24,$FF,$85,$C9,$19,$F8,$BF,$F0,$D1,$1C,$1F,$D4,$7E
	.dc.b	$3B,$EE,$B7,$9B,$FC,$75,$18,$D7,$FA,$DB,$5F,$D3,$2F,$87,$D9,$2B
	.dc.b	$A3,$FD,$7F,$A2,$47,$F5,$A6,$D7,$4F,$86,$EE,$49,$F2,$3E,$2F,$12
	.dc.b	$47,$9D,$B3,$FD,$BB,$CA,$75,$FF,$6D,$25,$7F,$68,$7D,$7F,$05,$23
	.dc.b	$E0,$36,$BD,$BE,$27,$C0,$24,$FC,$57,$FB,$C7,$6C,$8F,$EE,$6D,$9F
	.dc.b	$8B,$F0,$BF,$20,$93,$F3,$5F,$97,$D7,$23,$E8,$36,$7D,$5E,$47,$F8
	.dc.b	$14,$93,$FE,$00,$7E,$9F,$C2,$C8,$FB,$0D,$9F,$AB,$D1,$F6,$09,$3F
	.dc.b	$75,$FF,$08,$7C,$94,$8F,$F0,$5B,$67,$FC,$2F,$A7,$E4,$5B,$D5,$FE
	.dc.b	$7B,$8C,$47,$F2,$56,$BF,$DA,$57,$C3,$6C,$97,$FE,$1D,$8C,$BF,$E1
	.dc.b	$AF,$8F,$FF,$11,$11,$C1,$F9,$BF,$87,$FD,$B9,$FC,$66,$7F,$A7,$C7
	.dc.b	$ED,$23,$2E,$C9,$7F,$E3,$0E,$B4,$FF,$9E,$4F,$F8,$00,$49,$FD,$EF
	.dc.b	$C3,$F2,$AF,$E3,$14,$8C,$78,$1F,$C0,$1B,$5F,$E0,$79,$D3,$E3,$39
	.dc.b	$96,$B9,$CF,$07,$FC,$83,$3C,$1E,$F7,$E3,$77,$5C,$6F,$C0,$D3,$C1
	.dc.b	$7C,$E5,$7C,$EB,$71,$C8,$EB,$5E,$47,$11,$A3,$F8,$DD,$FF,$24,$B9
	.dc.b	$5B,$9C,$B4,$F0,$96,$BD,$75,$C1,$EB,$1A,$7C,$26,$0F,$83,$5E,$FA
	.dc.b	$F1,$4E,$E7,$FD,$97,$3F,$EA,$F2,$47,$F1,$D7,$FD,$C0,$47,$FD,$BC
	.dc.b	$D8,$FF,$6E,$4F,$FC,$01,$7F,$DE,$73,$FE,$FA,$69,$FE,$F9,$21,$83
	.dc.b	$E5,$FC,$3F,$39,$FF,$7D,$12,$7F,$DF,$0F,$5C,$13,$B9,$C9,$2B,$FD
	.dc.b	$F7,$5B,$9D,$46,$27,$4F,$81,$F9,$8D,$3D,$6E,$AB,$EF,$36,$79,$CE
	.dc.b	$EB,$CC,$F5,$3D,$B5,$E6,$AF,$8E,$7F,$5C,$AF,$EC,$67,$28,$D4,$F8
	.dc.b	$0E,$47,$C5,$73,$5D,$A3,$13,$C4,$E7,$5D,$A3,$91,$E5,$7D,$57,$A8
	.dc.b	$CD,$F4,$9F,$AC,$D8,$F0,$BF,$01,$B5,$DE,$7D,$47,$7B,$F2,$57,$1A
	.dc.b	$F0,$1F,$0A,$B8,$3E,$07,$F3,$13,$C1,$CB,$6B,$E9,$D7,$32,$4F,$08
	.dc.b	$FA,$47,$84,$B5,$EC,$BF,$75,$F3,$8F,$D1,$1E,$42,$5F,$59,$DD,$7F
	.dc.b	$0D,$BF,$F4,$6D,$58,$FE,$FD,$18,$8F,$E1,$6D,$7F,$10,$7F,$96,$6B
	.dc.b	$ED,$B6,$35,$18,$8F,$C6,$7C,$3F,$81,$FE,$2A,$D7,$F8,$6B,$60,$D2
	.dc.b	$57,$D2,$9F,$E5,$31,$87,$EE,$B7,$0F,$AA,$D9,$F0,$9C,$63,$FC,$5B
	.dc.b	$19,$9F,$C0,$B5,$F8,$ED,$AF,$A8,$DB,$FF,$1E,$B6,$FF,$C7,$2D,$AF
	.dc.b	$2C,$77,$3F,$9D,$46,$6F,$C3,$19,$9F,$86,$D7,$E1,$B6,$FE,$1B,$6B
	.dc.b	$93,$D0,$C1,$27,$E1,$B6,$BE,$1B,$4F,$87,$1A,$24,$69,$B1,$A8,$C1
	.dc.b	$1C,$12,$3E,$4B,$F0,$88,$F2,$3D,$A7,$C4,$30,$7B,$8C,$47,$57,$67
	.dc.b	$E3,$67,$F1,$64,$9F,$F2,$93,$FE,$6E,$F8,$7F,$A7,$1F,$E6,$A6,$CF
	.dc.b	$F6,$E6,$D7,$F9,$B1,$B3,$FD,$B5,$B3,$FD,$A1,$E9,$1C,$CF,$5C,$F1
	.dc.b	$71,$3D,$5C,$9C,$CC,$C6,$E3,$DB,$19,$78,$1F,$E0,$99,$FD,$39,$23
	.dc.b	$FC,$DC,$FF,$A2,$BE,$1F,$C7,$3F,$E0,$86,$D7,$FA,$19,$B5,$FE,$07
	.dc.b	$6C,$FF,$80,$9A,$FF,$7A,$7A,$47,$FA,$41,$CE,$07,$33,$DE,$7D,$67
	.dc.b	$53,$81,$1F,$5D,$78,$E3,$11,$FC,$16,$7F,$81,$92,$3F,$D1,$4F,$FB
	.dc.b	$1F,$B9,$F1,$9E,$0B,$52,$E5,$6A,$3C,$13,$A9,$CD,$71,$9E,$10,$AE
	.dc.b	$0F,$AC,$FF,$64,$38,$AF,$44,$62,$BC,$84,$AE,$E7,$4A,$E0,$D8,$D4
	.dc.b	$F5,$AF,$23,$E1,$8C,$C6,$2B,$8C,$FD,$5A,$F7,$4E,$6B,$9C,$FF,$A8
	.dc.b	$1F,$F5,$29,$27,$FD,$54,$FF,$BB,$FA,$DE,$1F,$69,$F8,$31,$87,$A9
	.dc.b	$FE,$69,$F2,$3F,$CD,$C4,$7F,$CD,$0F,$F9,$C7,$D1,$91,$FE,$6B,$6E
	.dc.b	$1F,$EA,$1F,$17,$FA,$15,$FF,$B0,$AB,$E7,$CF,$82,$72,$7F,$9E,$FC
	.dc.b	$F9,$3F,$E7,$56,$CF,$FA,$0B,$E2,$3F,$D0,$04,$9F,$F3,$D3,$DB,$C7
	.dc.b	$91,$DA,$7F,$09,$82,$73,$1F,$E8,$CF,$7F,$FD,$14,$4A,$FF,$42,$3F
	.dc.b	$E8,$DF,$2A,$47,$FA,$1D,$B3,$FE,$BD,$F8,$EF,$69,$27,$C1,$0C,$13
	.dc.b	$FE,$97,$F8,$79,$5F,$E9,$16,$D7,$FA,$7B,$E1,$FF,$D3,$84,$9F,$F4
	.dc.b	$B3,$F2,$BA,$A4,$76,$C3,$FD,$16,$D8,$FF,$52,$7C,$34,$8F,$F4,$F3
	.dc.b	$6F,$FD,$57,$F0,$DF,$EA,$92,$4F,$FA,$8D,$FA,$FD,$92,$3B,$36,$7F
	.dc.b	$DB,$87,$DC,$24,$78,$DF,$BF,$E8,$C8,$F3,$36,$3F,$DC,$9D,$BF,$84
	.dc.b	$24,$FA,$DF,$F0,$47,$C1,$BE,$12,$92,$7E,$6C,$65,$F9,$FF,$06,$F8
	.dc.b	$27,$C2,$61,$B8,$FE,$0E,$5F,$08,$CF,$57,$F6,$3E,$35,$E9,$FD,$AC
	.dc.b	$11,$87,$51,$98,$C4,$7E,$3F,$E1,$5E,$D7,$16,$C7,$E1,$1A,$EE,$35
	.dc.b	$E9,$C8,$D3,$5F,$CD,$7C,$23,$C1,$FE,$2F,$E2,$7F,$08,$F8,$5F,$8D
	.dc.b	$5F,$B8,$3F,$91,$38,$51,$FC,$3A,$B9,$46,$C3,$87,$7C,$3F,$C9,$23
	.dc.b	$13,$C6,$FC,$33,$9F,$0E,$DF,$F3,$1F,$49,$EE,$62,$1E,$03,$E1,$70
	.dc.b	$D8,$F1,$57,$06,$FB,$5C,$63,$11,$E4,$3B,$5E,$67,$B7,$C1,$3E,$17
	.dc.b	$AF,$B9,$2F,$DD,$39,$6F,$C2,$F0,$A9,$18,$37,$F0,$3D,$26,$46,$0E
	.dc.b	$7B,$FA,$B8,$E4,$60,$EF,$BF,$92,$64,$60,$F1,$FC,$2F,$D4,$7F,$8B
	.dc.b	$64,$8C,$47,$F2,$68,$C7,$17,$F0,$C6,$C7,$0F,$31,$F0,$B9,$6C,$66
	.dc.b	$30,$7E,$6D,$7C,$2D,$23,$F9,$F7,$F4,$1D,$BF,$D9,$F0,$FF,$3F,$B7
	.dc.b	$9B,$F9,$64,$FF,$6F,$F8,$57,$2F,$F3,$CA,$D6,$1F,$DB,$39,$6B,$FD
	.dc.b	$D5,$AF,$EC,$70,$FC,$93,$F8,$86,$F2,$E8,$E3,$CC,$F8,$4C,$4F,$0F
	.dc.b	$83,$E6,$62,$FF,$0C,$7F,$80,$DE,$FF,$07,$27,$81,$E0,$7C,$BB,$39
	.dc.b	$E9,$E0,$ED,$B5,$FA,$79,$FD,$93,$79,$73,$8F,$82,$72,$F8,$BD,$8F
	.dc.b	$73,$77,$F5,$FC,$17,$C4,$9C,$D7,$F8,$09,$F8,$55,$C7,$83,$57,$B7
	.dc.b	$E1,$FC,$0F,$CC,$F8,$57,$C0,$E3,$4F,$DE,$46,$1B,$1F,$EB,$38,$C6
	.dc.b	$1F,$D9,$5A,$FF,$74,$6B,$FE,$24,$C7,$FB,$C3,$FD,$AD,$AF,$FA,$EF
	.dc.b	$D8,$FF,$5C,$9F,$EB,$AF,$5E,$61,$FD,$8A,$BF,$B8,$36,$3F,$D7,$FE
	.dc.b	$B4,$8D,$36,$FF,$BC,$7B,$6F,$85,$FF,$65,$36,$BF,$B6,$B6,$77,$3F
	.dc.b	$E0,$EE,$89,$5F,$D9,$5B,$3F,$E0,$3F,$5B,$FD,$4A,$DF,$7F,$FC,$2A
	.dc.b	$4A,$EB,$E3,$78,$08,$E0,$FE,$E1,$D3,$FE,$1C,$23,$83,$FB,$A7,$BF
	.dc.b	$FE,$20,$23,$83,$D3,$E2,$FB,$08,$E0,$FE,$F9,$D7,$FE,$28,$23,$83
	.dc.b	$FC,$01,$EF,$FF,$8B,$08,$E0,$F8,$BD,$3F,$20,$8E,$0F,$F0,$47,$7F
	.dc.b	$F8,$D0,$8E,$0F,$F0,$67,$A9,$FE,$38,$23,$83,$E8,$F4,$7D,$42,$38
	.dc.b	$3F,$C2,$9F,$07,$FE,$40,$23,$83,$FC,$31,$F2,$3F,$E4,$42,$38,$3F
	.dc.b	$C5,$3F,$04,$F8,$4F,$F3,$2F,$17,$F9,$DA,$31,$3A,$9C,$CF,$FA,$06
	.dc.b	$72,$FD,$FF,$84,$FF,$28,$92,$BF,$C5,$0F,$F9,$53,$E4,$64,$7F,$8A
	.dc.b	$DB,$3F,$9F,$F7,$3F,$6C,$7E,$DF,$DF,$B5,$FD,$A7,$B5,$3E,$2A,$46
	.dc.b	$27,$F4,$53,$98,$C4,$F5,$9F,$F7,$C4,$FF,$03,$F7,$7A,$CF,$F0,$96
	.dc.b	$F8,$9F,$C1,$5B,$5F,$B5,$6C,$FF,$A8,$6D,$18,$8F,$1B,$63,$33,$E9
	.dc.b	$BF,$03,$4F,$19,$2B,$C4,$FB,$EB,$71,$FC,$76,$6D,$F9,$6B,$83,$67
	.dc.b	$89,$1F,$E3,$D3,$C2,$7D,$47,$02,$CE,$DB,$3C,$1F,$11,$23,$47,$C1
	.dc.b	$9E,$8F,$C3,$AA,$FD,$48,$B7,$F0,$A3,$F8,$E4,$FD,$52,$57,$F1,$97
	.dc.b	$FD,$FF,$1F,$2A,$3F,$DF,$2F,$56,$B9,$3F,$DF,$26,$4A,$66,$FF,$9F
	.dc.b	$74,$F4,$BF,$E2,$71,$F1,$CC,$D7,$48,$CB,$F0,$8F,$E6,$93,$C4,$F5
	.dc.b	$9F,$D2,$CF,$C7,$11,$FE,$64,$DA,$FF,$78,$C7,$23,$FD,$EE,$FE,$26
	.dc.b	$3F,$DF,$A4,$8C,$C7,$E7,$23,$FE,$00,$F3,$E6,$39,$9F,$91,$71,$7F
	.dc.b	$D6,$DD,$D7,$AD,$B3,$F1,$2D,$18,$AF,$83,$3D,$0F,$D5,$C7,$4A,$FE
	.dc.b	$A8,$ED,$F5,$46,$EB,$9C,$7A,$1F,$93,$1E,$58,$C4,$FD,$13,$94,$7C
	.dc.b	$18,$C9,$C6,$F3,$1E,$28,$F5,$1A,$7D,$53,$AA,$EF,$1B,$3C,$91,$DE
	.dc.b	$78,$5F,$B8,$F8,$CE,$0F,$5B,$89,$E3,$1E,$06,$9E,$07,$83,$E9,$33
	.dc.b	$1C,$4E,$D1,$F2,$0D,$C7,$DB,$6D,$7C,$88,$E8,$72,$8F,$19,$C5,$C1
	.dc.b	$DE,$7B,$46,$5B,$19,$8C,$1C,$5F,$F8,$03,$3D,$DB,$5F,$C9,$49,$3F
	.dc.b	$E0,$96,$C7,$F8,$45,$B5,$FE,$27,$E5,$FF,$15,$47,$F8,$AD,$B5,$FC
	.dc.b	$D5,$B5,$FE,$33,$6D,$71,$9F,$C3,$46,$1F,$F1,$03,$67,$F9,$63,$5F
	.dc.b	$E7,$AD,$B3,$49,$3D,$09,$3F,$95,$8C,$B6,$33,$5F,$22,$3A,$C6,$0F
	.dc.b	$92,$F8,$9C,$3F,$9B,$3D,$8E,$1F,$C8,$47,$B5,$EC,$47,$B5,$7A,$27
	.dc.b	$83,$F8,$18,$E0,$6A,$BC,$ED,$9D,$DF,$CF,$69,$F3,$CB,$7E,$43,$C0
	.dc.b	$FA,$9F,$69,$83,$D5,$3E,$08,$D5,$F9,$67,$47,$B2,$3E,$A9,$DA,$78
	.dc.b	$47,$28,$CB,$C2,$7D,$91,$89,$F5,$4F,$95,$E2,$F5,$3E,$71,$98,$E0
	.dc.b	$79,$2F,$E0,$46,$4F,$01,$C2,$7E,$83,$F9,$E8,$FF,$00,$12,$7F,$CC
	.dc.b	$8F,$FA,$93,$B5,$F0,$BD,$2D,$4B,$95,$A8,$F9,$D1,$FE,$9E,$8C,$3E
	.dc.b	$30,$E2,$7D,$AD,$9F,$5C,$F4,$AF,$11,$23,$AD,$78,$A7,$51,$E3,$8C
	.dc.b	$47,$69,$E0,$F8,$6B,$B3,$C0,$F3,$47,$B6,$BD,$91,$C5,$E8,$62,$7D
	.dc.b	$71,$A3,$47,$FA,$12,$7F,$D0,$64,$9E,$31,$F1,$FB,$1F,$0B,$E4,$C3
	.dc.b	$CF,$18,$AD,$D7,$49,$CD,$6E,$7C,$B5,$BA,$F4,$C7,$08,$C5,$7C,$2C
	.dc.b	$3D,$F4,$9C,$D7,$2B,$FF,$70,$C6,$27,$E2,$5F,$82,$BA,$56,$8F,$14
	.dc.b	$FA,$DC,$46,$8F,$4D,$F8,$2F,$A5,$78,$6B,$FC,$A5,$1E,$98,$F9,$4D
	.dc.b	$7C,$95,$F3,$5E,$CF,$C4,$FC,$FF,$F2,$2F,$78,$FF,$75,$C6,$1F,$F7
	.dc.b	$6C,$F8,$23,$18,$7A,$E7,$75,$E8,$6B,$CE,$3F,$13,$3C,$23,$F0,$6D
	.dc.b	$9F,$D0,$35,$F6,$47,$94,$FC,$A4,$F2,$69,$C8,$F0,$39,$78,$12,$3D
	.dc.b	$37,$C1,$EF,$15,$57,$BD,$FC,$93,$5E,$D1,$F2,$5D,$BF,$E8,$18,$F4
	.dc.b	$0F,$92,$E8,$E1,$ED,$C8,$F5,$3F,$D1,$0D,$AF,$F5,$0C,$60,$FB,$7E
	.dc.b	$2C,$8F,$06,$7F,$EF,$2C,$7F,$2F,$3E,$1F,$92,$95,$E5,$C9,$18,$AF
	.dc.b	$C0,$62,$F8,$8F,$09,$2F,$FD,$62,$FE,$2B,$C7,$91,$E0,$6D,$7E,$27
	.dc.b	$07,$C0,$7C,$12,$57,$85,$F6,$FA,$32,$3F,$D7,$AD,$AF,$F7,$A7,$93
	.dc.b	$F9,$82,$4F,$B9,$F9,$9F,$05,$23,$D8,$D9,$F5,$FC,$37,$CA,$24,$FF
	.dc.b	$B2,$9F,$D1,$F6,$48,$F9,$2D,$9F,$57,$37,$D9,$24,$FC,$97,$EA,$FC
	.dc.b	$2C,$8F,$F6,$AB,$67,$F5,$5C,$5F,$EB,$C7,$F5,$FD,$BF,$B2,$25,$7D
	.dc.b	$3F,$92,$FC,$31,$1C,$1F,$73,$D1,$FD,$C9,$1C,$1F,$80,$EF,$FD,$E9
	.dc.b	$1C,$1F,$7F,$D5,$FC,$91,$1C,$1F,$97,$F8,$56,$EF,$85,$C3,$73,$FE
	.dc.b	$09,$3F,$73,$1C,$D7,$E7,$FE,$11,$F0,$BF,$A1,$B4,$61,$D4,$66,$31
	.dc.b	$1F,$C3,$F9,$5F,$1D,$C5,$B1,$FA,$B6,$BB,$8E,$DF,$0C,$F4,$69,$27
	.dc.b	$F6,$51,$96,$FC,$8F,$02,$38,$3F,$71,$F0,$CF,$6F,$FA,$9F,$8B,$FD
	.dc.b	$73,$E1,$DE,$96,$38,$FF,$AB,$C6,$30,$FE,$93,$87,$16,$F8,$BB,$7F
	.dc.b	$BC,$46,$56,$1F,$DA,$23,$0D,$9E,$12,$F4,$DE,$E6,$3C,$17,$C3,$D3
	.dc.b	$73,$AA,$FE,$B6,$BD,$AC,$DC,$9C,$C7,$6F,$84,$7C,$3F,$5F,$72,$57
	.dc.b	$F6,$49,$CB,$7E,$1F,$85,$48,$C1,$BF,$82,$E9,$32,$30,$73,$E3,$F5
	.dc.b	$73,$48,$C1,$DF,$7F,$24,$C8,$C1,$E3,$F8,$7F,$92,$FF,$57,$C9,$18
	.dc.b	$8F,$F0,$34,$63,$87,$F9,$A3,$63,$87,$C3,$BE,$1F,$2D,$8C,$C6,$0F
	.dc.b	$CE,$2F,$87,$A4,$7F,$40,$FE,$87,$C0,$FE,$91,$06,$BC,$47,$C3,$FD
	.dc.b	$D7,$FC,$43,$1C,$38,$E7,$4F,$F8,$52,$79,$36,$3A,$F2,$48,$FF,$0C
	.dc.b	$F3,$BC,$4C,$5F,$E3,$A7,$FC,$51,$F0,$7E,$43,$3B,$FF,$8E,$25,$1E
	.dc.b	$8F,$83,$95,$E1,$8B,$3C,$CF,$F2,$07,$37,$F6,$5B,$3F,$D9,$F1,$7C
	.dc.b	$3E,$63,$5C,$93,$1F,$F1,$7D,$F0,$6C,$7F,$6F,$F6,$FD,$F8,$F7,$0E
	.dc.b	$2B,$FC,$9F,$8B,$E1,$38,$70,$71,$E0,$99,$7F,$8F,$30,$E0,$FF,$93
	.dc.b	$9F,$D5,$BF,$E4,$0F,$86,$7C,$17,$F7,$F6,$C6,$7B,$31,$32,$3F,$BF
	.dc.b	$4E,$B4,$FF,$14,$35,$FF,$2B,$35,$FF,$2D,$79,$BF,$3D,$FF,$3C,$B5
	.dc.b	$FF,$3A,$B7,$DD,$37,$92,$5F,$F8,$36,$7E,$F3,$63,$87,$C3,$48,$D3
	.dc.b	$6F,$FC,$79,$F2,$7F,$CD,$A7,$FC,$2C,$D8,$CF,$99,$E3,$8C,$47,$F8
	.dc.b	$7A,$BF,$C3,$44,$AF,$F0,$53,$E9,$F3,$7F,$CE,$B3,$FE,$7A,$6B,$FE
	.dc.b	$97,$6B,$FD,$37,$D4,$FF,$12,$7C,$9C,$BF,$F0,$A3,$67,$FC,$15,$8F
	.dc.b	$F8,$A4,$93,$FE,$19,$7F,$C5,$9E,$2C,$8F,$F0,$E3,$67,$FC,$25,$D3
	.dc.b	$FE,$31,$24,$FF,$88,$5F,$F1,$A7,$B9,$23,$FC,$48,$D9,$FF,$0D,$78
	.dc.b	$BF,$E3,$92,$4F,$F8,$A5,$FF,$1E,$79,$92,$3F,$C5,$8D,$9F,$F1,$17
	.dc.b	$C7,$FF,$90,$49,$3F,$E3,$17,$FC,$89,$D3,$23,$FC,$68,$D9,$FF,$15
	.dc.b	$79,$5F,$E4,$92,$4F,$F8,$E5,$FF,$26,$7C,$7C,$8F,$F1,$E3,$67,$FC
	.dc.b	$65,$F1,$5F,$E5,$12,$4F,$F9,$05,$FC,$87,$E8,$1F,$DF,$E3,$FC,$B0
	.dc.b	$D9,$FF,$41,$35,$FF,$58,$35,$FF,$5B,$AF,$8C,$E6,$49,$E6,$F6,$EA
	.dc.b	$91,$D9,$B5,$DB,$AF,$FC,$DD,$7F,$D4,$AB,$FD,$66,$BC,$DC,$CC,$9F
	.dc.b	$3B,$EA,$23,$83,$8F,$87,$BB,$FE,$15,$7F,$C3,$DB,$E1,$F8,$0F,$0F
	.dc.b	$9B,$6F,$C3,$16,$78,$CD,$F1,$3F,$DB,$F7,$FD,$42,$FC,$FE,$5F,$AB
	.dc.b	$CD,$F5,$B6,$32,$7B,$23,$2F,$BB,$E0,$7E,$09,$F0,$BE,$52,$C6,$4F
	.dc.b	$AB,$1F,$60,$8E,$0F,$8F,$F2,$3C,$33,$F7,$BF,$33,$C9,$93,$F3,$3E
	.dc.b	$52,$D7,$C4,$9B,$A7,$B2,$6D,$7D,$EF,$92,$FA,$BF,$05,$F5,$9B,$1F
	.dc.b	$59,$B8,$7C,$E6,$DA,$7F,$DF,$BE,$2F,$F4,$47,$FD,$DE,$DA,$FA,$BE
	.dc.b	$CF,$FA,$BE,$31,$5C,$64,$23,$83,$83,$2E,$1E,$FE,$07,$F1,$B2,$5F
	.dc.b	$E3,$3E,$4B,$B4,$7E,$41,$FC,$8F,$C1,$49,$FC,$8F,$CA,$5B,$FC,$4C
	.dc.b	$DC,$7F,$03,$36,$BC,$5F,$05,$E6,$7C,$F3,$E9,$B4,$FF,$52,$7D,$6E
	.dc.b	$DF,$EB,$CE,$5D,$9F,$ED,$17,$06,$E3,$A3,$8F,$43,$46,$FD,$4B,$7E
	.dc.b	$23,$CB,$1F,$D3,$E7,$FB,$0D,$FF,$BF,$DA,$FC,$B3,$CF,$1B,$3C,$5E
	.dc.b	$9F,$48,$F3,$BE,$5F,$52,$5F,$79,$F6,$FB,$38,$7D,$B3,$72,$F4,$CD
	.dc.b	$AF,$93,$F2,$5E,$FF,$91,$F8,$0D,$8E,$98,$7C,$A8,$C3,$F2,$1B,$7F
	.dc.b	$3F,$D0,$FE,$CB,$3F,$E0,$96,$C6,$7D,$4F,$F7,$E4,$62,$FF,$AC,$7A
	.dc.b	$78,$33,$36,$3C,$33,$BB,$FE,$B9,$5A,$9C,$BB,$8F,$03,$96,$C6,$88
	.dc.b	$F9,$6B,$85,$FA,$6A,$D7,$F6,$B9,$B5,$E0,$8E,$35,$EB,$69,$EA,$2C
	.dc.b	$7F,$71,$9E,$2F,$FA,$F6,$AD,$FF,$83,$AB,$73,$96,$C6,$63,$D2,$4B
	.dc.b	$F3,$4F,$C4,$BE,$77,$D1,$B1,$C9,$B5,$C6,$3D,$87,$69,$E0,$D3,$81
	.dc.b	$63,$81,$EB,$71,$5E,$28,$F9,$51,$83,$E1,$38,$38,$9E,$D2,$DF,$B6
	.dc.b	$79,$12,$78,$BF,$36,$77,$5E,$66,$CF,$99,$A7,$4B,$F7,$BF,$AB,$BF
	.dc.b	$C9,$B6,$BF,$C5,$04,$AF,$D8,$47,$DD,$2C,$F0,$7D,$67,$18,$FC,$9F
	.dc.b	$A7,$F6,$88,$FD,$86,$C7,$B9,$B8,$7B,$63,$FC,$69,$18,$9F,$79,$FE
	.dc.b	$39,$71,$1E,$39,$FD,$64,$62,$3E,$54,$62,$7C,$04,$8F,$03,$63,$C1
	.dc.b	$0A,$57,$37,$F5,$91,$98,$C5,$FE,$D9,$B7,$F0,$0E,$87,$A2,$7D,$6E
	.dc.b	$1F,$43,$EB,$30,$FA,$1C,$3E,$62,$5F,$9A,$FC,$B3,$C1,$F2,$98,$9D
	.dc.b	$C7,$85,$F4,$3B,$BF,$43,$4F,$44,$70,$3C,$A7,$A0,$C1,$E5,$31,$5E
	.dc.b	$3B,$FA,$1B,$BE,$37,$FA,$21,$FD,$17,$E3,$2E,$1F,$C9,$A7,$4D,$9D
	.dc.b	$34,$D7,$C5,$FF,$4D,$F8,$AF,$E9,$CD,$8C,$9C,$E3,$2D,$3F,$91,$FA
	.dc.b	$7F,$CA,$8F,$E5,$7D,$36,$32,$7E,$73,$4F,$C0,$4F,$FA,$1C,$91,$9D
	.dc.b	$91,$FC,$D3,$D3,$E5,$5F,$CE,$5F,$EC,$5F,$15,$2F,$DD,$A5,$AF,$ED
	.dc.b	$33,$77,$FE,$CB,$36,$BF,$BA,$FC,$4F,$F6,$BF,$90,$FE,$D8,$D8,$F9
	.dc.b	$0D,$C3,$FB,$9B,$6F,$FA,$67,$A5,$FD,$20,$FE,$93,$E9,$DB,$FE,$8B
	.dc.b	$1F,$EA,$32,$3E,$D6,$DF,$F5,$3F,$4F,$83,$67,$EC,$7C,$87,$F7,$FF
	.dc.b	$90,$FF,$00,$36,$32,$7D,$D8,$CB,$4F,$C0,$7C,$3F,$F8,$1C,$FF,$04
	.dc.b	$7C,$95,$8C,$92,$FF,$C1,$F1,$97,$FD,$DD,$F6,$3F,$E6,$F9,$FE,$DE
	.dc.b	$D8,$CF,$AB,$FE,$33,$D7,$FC,$39,$39,$6F,$AD,$FE,$DB,$8E,$15,$FE
	.dc.b	$0B,$6B,$FE,$37,$6B,$FE,$3C,$79,$5E,$7D,$6E,$CE,$EA,$D6,$E7,$FC
	.dc.b	$49,$F2,$78,$87,$E8,$3F,$E4,$FB,$FD,$14,$E6,$31,$3F,$E2,$E8,$C4
	.dc.b	$F2,$3F,$C1,$B3,$AC,$7F,$C7,$0B,$E5,$39,$1C,$FE,$52,$D7,$08,$B8
	.dc.b	$2F,$8D,$61,$F8,$FE,$A7,$9F,$43,$7D,$7C,$33,$1F,$D4,$4E,$9B,$1F
	.dc.b	$34,$8F,$D6,$F5,$B8,$63,$F6,$9F,$81,$F1,$FF,$E5,$18,$FF,$33,$68
	.dc.b	$8C,$47,$F9,$B2,$31,$5F,$B7,$6C,$7F,$54,$5F,$25,$C2,$3E,$7B,$63
	.dc.b	$33,$C6,$31,$F0,$B2,$3A,$36,$BA,$79,$1C,$49,$3E,$17,$A7,$91,$23
	.dc.b	$C8,$D9,$F2,$79,$1E,$12,$4F,$A5,$F2,$7C,$2C,$8F,$63,$67,$D9,$F0
	.dc.b	$BE,$92,$4F,$C1,$7D,$9F,$0B,$23,$E2,$36,$7E,$27,$8F,$F0,$49,$3F
	.dc.b	$25,$F8,$9E,$34,$8F,$98,$D9,$F9,$9E,$37,$C9,$24,$FD,$17,$E6,$7C
	.dc.b	$2C,$8F,$A8,$D9,$FF,$03,$F2,$3D,$CF,$D8,$12,$BF,$12,$FD,$7F,$B2
	.dc.b	$7A,$74,$93,$FC,$D6,$32,$FF,$37,$F9,$77,$BF,$F7,$DB,$19,$8C,$1F
	.dc.b	$83,$F9,$77,$89,$49,$3F,$D0,$63,$2F,$F4,$3F,$95,$FC,$5F,$9F,$7A
	.dc.b	$7F,$C0,$9B,$7F,$88,$A7,$F2,$DF,$1B,$9A,$FC,$A7,$C5,$7F,$4B,$AF
	.dc.b	$DF,$B7,$B0,$D3,$1F,$15,$27,$F3,$8D,$7F,$39,$E7,$71,$23,$FA,$17
	.dc.b	$A7,$9D,$23,$F4,$8D,$7F,$49,$E7,$78,$48,$FE,$A5,$F3,$7C,$93,$E2
	.dc.b	$7F,$DA,$DE,$3F,$FB,$0F,$E4,$FE,$3D,$FF,$A7,$1A,$FF,$09,$E1,$FF
	.dc.b	$6E,$5F,$18,$FF,$50,$AF,$8A,$E7,$3A,$C2,$D7,$FB,$9A,$B8,$3C,$CC
	.dc.b	$56,$97,$BA,$F7,$F1,$D6,$DA,$9B,$9D,$57,$FA,$C9,$7C,$06,$6E,$4E
	.dc.b	$63,$B7,$C9,$BD,$3E,$BE,$24,$AF,$F5,$DC,$E5,$BE,$9F,$0A,$91,$83
	.dc.b	$7E,$9F,$49,$91,$83,$9E,$FE,$AF,$12,$46,$0E,$FB,$F9,$26,$46,$0F
	.dc.b	$1F,$C5,$7F,$39,$7F,$CE,$79,$23,$11,$FE,$E5,$8C,$72,$7F,$87,$5B
	.dc.b	$1C,$3D,$17,$C9,$E5,$B1,$98,$C1,$F9,$B5,$F2,$74,$8F,$E7,$DF,$D0
	.dc.b	$77,$7F,$5F,$83,$5E,$23,$E4,$FE,$EA,$24,$71,$E3,$9D,$1F,$EE,$C9
	.dc.b	$E6,$D8,$ED,$C9,$23,$FD,$E3,$CF,$FE,$76,$AF,$C0,$C7,$1F,$1F,$C8
	.dc.b	$6E,$3F,$20,$78,$5F,$47,$C2,$C9,$E0,$78,$5F,$EF,$5B,$B9,$69,$FD
	.dc.b	$F7,$07,$99,$8B,$F6,$51,$FD,$6F,$1F,$23,$59,$78,$A7,$F6,$53,$C0
	.dc.b	$F8,$90,$F9,$3F,$82,$FA,$1B,$3D,$63,$E5,$60,$F0,$3F,$C2,$15,$AB
	.dc.b	$E6,$FE,$AD,$ED,$5F,$46,$30,$F5,$DF,$8E,$FF,$E1,$9F,$93,$7C,$8F
	.dc.b	$F6,$D6,$E1,$FE,$69,$6D,$FF,$B1,$9A,$FA,$83,$AC,$FD,$11,$8E,$AE
	.dc.b	$D7,$DD,$FF,$14,$7C,$9F,$FB,$36,$7F,$0C,$D4,$D2,$7A,$7F,$E0,$17
	.dc.b	$FD,$45,$1E,$16,$BF,$31,$AF,$E4,$7E,$5B,$F0,$EF,$FB,$79,$AF,$A5
	.dc.b	$AF,$D4,$6B,$F9,$9F,$97,$E6,$9A,$4A,$FF,$0A,$79,$2F,$62,$93,$5F
	.dc.b	$F4,$73,$FE,$19,$FD,$0B,$FD,$A5,$1F,$E9,$56,$E1,$FE,$F2,$6B,$F3
	.dc.b	$5A,$FE,$23,$07,$AD,$FE,$23,$6D,$7F,$A7,$DB,$3F,$EF,$C6,$BF,$5D
	.dc.b	$78,$9C,$49,$3C,$5F,$37,$CB,$48,$F3,$36,$BA,$7A,$FE,$12,$4F,$85
	.dc.b	$F2,$78,$B2,$3C,$8D,$9F,$77,$0F,$C3,$24,$FC,$37,$D9,$E6,$C8,$F6
	.dc.b	$36,$7D,$9D,$1F,$04,$93,$F0,$5F,$9D,$E3,$C8,$F9,$CD,$9F,$89,$E6
	.dc.b	$FC,$92,$4F,$C9,$7E,$67,$9B,$23,$E6,$36,$7E,$B7,$37,$DB,$24,$FD
	.dc.b	$B7,$FC,$AD,$F2,$CD,$A9,$27,$FD,$A9,$19,$6E,$9C,$08,$E0,$FC,$37
	.dc.b	$81,$F8,$92,$38,$3F,$DB,$DF,$25,$FC,$59,$FE,$03,$8A,$30,$FF,$24
	.dc.b	$B7,$C9,$FF,$A6,$67,$3D,$7F,$B8,$77,$5F,$BD,$6B,$FC,$61,$AF,$F3
	.dc.b	$D6,$BD,$FC,$2F,$F5,$25,$BE,$4F,$B4,$61,$CF,$C9,$CC,$BC,$F4,$F9
	.dc.b	$3C,$61,$E7,$AB,$97,$F0,$CA,$D5,$F0,$9F,$F5,$97,$C6,$7F,$4F,$8F
	.dc.b	$82,$FF,$25,$BF,$EA,$4E,$5F,$E8,$AE,$53,$9B,$FE,$32,$E4,$C4,$FF
	.dc.b	$38,$71,$E1,$78,$0F,$0F,$93,$67,$C3,$1F,$E9,$79,$FE,$63,$7F,$C9
	.dc.b	$DA,$F3,$8D,$1E,$AF,$27,$94,$79,$9F,$6F,$93,$2F,$DB,$E8,$DA,$F5
	.dc.b	$CD,$CB,$D3,$36,$BD,$1F,$25,$F1,$BE,$4F,$E3,$B6,$3E,$3B,$70,$F8
	.dc.b	$AD,$BF,$E7,$7E,$23,$E0,$BF,$DC,$0D,$BF,$C7,$AF,$83,$CC,$C9,$FE
	.dc.b	$8D,$87,$F8,$0E,$BF,$9F,$12,$7F,$94,$9D,$3C,$7E,$13,$FD,$EA,$51
	.dc.b	$F7,$BE,$4A,$5F,$DE,$F9,$3B,$7F,$72,$6E,$5F,$6A,$6D,$7D,$9F,$93
	.dc.b	$FC,$3F,$C9,$FE,$21,$B1,$F8,$86,$E1,$F8,$66,$DF,$C2,$F1,$BE,$09
	.dc.b	$F0,$BA,$AD,$FC,$28,$FE,$86,$47,$E0,$DB,$44,$E6,$3D,$DE,$37,$26
	.dc.b	$D7,$E6,$BE,$4F,$F3,$A7,$E7,$B9,$EC,$64,$93,$F9,$C8,$CB,$F9,$EF
	.dc.b	$92,$FD,$2F,$C9,$FE,$99,$B5,$FA,$66,$CE,$E3,$F5,$33,$F7,$BC,$69
	.dc.b	$1F,$7B,$DD,$B5,$F5,$66,$ED,$F8,$69,$B5,$FE,$1E,$D5,$58,$BF,$C1
	.dc.b	$55,$7E,$24,$59,$FF,$41,$57,$E3,$23,$07,$E4,$2A,$C6,$4F,$21,$6B
	.dc.b	$C8,$DB,$F2,$34,$F2,$57,$7A,$E7,$5C,$0F,$C8,$38,$AF,$D0,$CF,$86
	.dc.b	$B9,$57,$E6,$E3,$94,$62,$3D,$6D,$BF,$5C,$75,$AE,$73,$9B,$EC,$F9
	.dc.b	$0F,$5C,$ED,$C3,$4F,$92,$F5,$BF,$0C,$64,$F0,$96,$FC,$37,$E0,$8C
	.dc.b	$4F,$AD,$B5,$E3,$3A,$15,$C9,$CE,$FC,$EF,$C1,$8F,$3D,$75,$23,$E8
	.dc.b	$7E,$8C,$72,$72,$FA,$8E,$A7,$F9,$2E,$FF,$0F,$3F,$A8,$6C,$7F,$0A
	.dc.b	$6B,$BB,$FF,$18,$92,$78,$36,$3F,$85,$78,$FF,$8D,$24,$FD,$56,$C7
	.dc.b	$C7,$6D,$F9,$A3,$F7,$11,$89,$F6,$1F,$9D,$71,$1F,$0C,$EF,$1D,$48
	.dc.b	$F5,$6C,$F5,$B5,$72,$AF,$B3,$3F,$CC,$23,$13,$F0,$A3,$15,$C8,$F1
	.dc.b	$D7,$F2,$DA,$DD,$FF,$36,$6C,$FF,$36,$6C,$F9,$8F,$21,$2B,$C9,$1A
	.dc.b	$7B,$C6,$67,$BB,$F1,$1F,$13,$F0,$63,$67,$28,$CC,$7B,$1F,$90,$62
	.dc.b	$FD,$8E,$23,$B1,$CD,$C4,$F3,$8C,$9C,$CF,$81,$1F,$E0,$0F,$93,$FF
	.dc.b	$01,$FC,$9F,$F8,$11,$B1,$93,$FC,$0B,$19,$69,$FE,$08,$F8,$6F,$F0
	.dc.b	$59,$FE,$0C,$F9,$3B,$19,$25,$FF,$82,$A3,$2F,$FB,$43,$E2,$DE,$67
	.dc.b	$FA,$E7,$2F,$F5,$1C,$62,$FF,$C6,$AB,$CD,$D4,$7F,$B1,$23,$57,$CE
	.dc.b	$FF,$D5,$6E,$5A,$FF,$BB,$63,$6F,$FA,$CD,$AE,$E7,$BF,$97,$DA,$38
	.dc.b	$55,$55,$FE,$E1,$AF,$F1,$4F,$C9,$E7,$E1,$BF,$DB,$D1,$83,$FC,$9A
	.dc.b	$E2,$BF,$C5,$73,$B9,$FF,$19,$B5,$DF,$C9,$F3,$3A,$7C,$9D,$8C,$92
	.dc.b	$BF,$DF,$36,$9D,$47,$CD,$F9,$3F,$46,$26,$FC,$26,$19,$87,$F7,$09
	.dc.b	$D3,$63,$E4,$D7,$C0,$6C,$EA,$3E,$D7,$C9,$F2,$6D,$7F,$9B,$3D,$AF
	.dc.b	$F7,$DC,$38,$3F,$DB,$F0,$F8,$8F,$98,$EA,$33,$18,$F8,$3C,$B6,$FA
	.dc.b	$3C,$3E,$22,$47,$7D,$2E,$3D,$E0,$CC,$55,$AC,$7C,$77,$6F,$76,$BF
	.dc.b	$75,$7A,$34,$F3,$E7,$E8,$65,$FF,$B5,$5F,$F6,$C1,$1C,$1F,$0F,$E3
	.dc.b	$B7,$87,$94,$F2,$3F,$0B,$E1,$7E,$1F,$CB,$5A,$F9,$4D,$AE,$07,$95
	.dc.b	$C6,$3F,$31,$F9,$7F,$2B,$2B,$E9,$7C,$B5,$BE,$3A,$2B,$18,$7D,$4A
	.dc.b	$B7,$FE,$99,$F1,$BF,$D4,$07,$D6,$F6,$EC,$7F,$AA,$3D,$BF,$F1,$AB
	.dc.b	$FE,$AC,$9F,$C3,$C6,$32,$FF,$2D,$B5,$FF,$53,$47,$E0,$3D,$BE,$71
	.dc.b	$E4,$7F,$55,$F2,$1F,$90,$E8,$34,$98,$7F,$91,$A1,$F0,$9B,$24,$65
	.dc.b	$FC,$CE,$CF,$6E,$92,$7F,$C9,$F1,$97,$A7,$A7,$E1,$23,$83,$C1,$E9
	.dc.b	$F9,$48,$E0,$E7,$EE,$F8,$08,$E0,$F4,$7C,$27,$B4,$8E,$0F,$67,$C1
	.dc.b	$FC,$12,$38,$3F,$CC,$BF,$A1,$FF,$AD,$77,$FC,$B8,$70,$FE,$69,$AF
	.dc.b	$E8,$9A,$B5,$FA,$8F,$E8,$1A,$FE,$A1,$AE,$F8,$3E,$A7,$A5,$2F,$E9
	.dc.b	$36,$7F,$D3,$36,$F8,$9F,$E1,$2D,$AF,$4C,$60,$FE,$17,$D3,$23,$F8
	.dc.b	$93,$6B,$FD,$47,$F2,$8F,$7F,$F8,$9B,$6B,$FD,$0E,$D9,$DC,$EB,$D2
	.dc.b	$95,$A6,$CF,$FA,$BB,$E4,$DE,$6F,$F1,$E6,$D7,$FA,$49,$B3,$B9,$FE
	.dc.b	$47,$F2,$7F,$94,$F4,$7F,$D2,$ED,$9F,$F5,$CA,$F9,$7F,$85,$18,$EA
	.dc.b	$FF,$4E,$3F,$EB,$EF,$99,$72,$FF,$A8,$21,$18,$7F,$D4,$6D,$78,$46
	.dc.b	$F9,$1F,$13,$49,$3F,$A5,$75,$EB,$C8,$FD,$43,$5F,$AF,$F0,$BE,$1E
	.dc.b	$53,$49,$1F,$AF,$F5,$FC,$04,$70,$7E,$53,$E1,$B9,$11,$C1,$FE,$BC
	.dc.b	$F9,$47,$9F,$FC,$BB,$C7,$72,$FF,$30,$F9,$5F,$A7,$7F,$CE,$27,$2F
	.dc.b	$F3,$A8,$CB,$FC,$8A,$F9,$BB,$AF,$E7,$31,$FD,$36,$30,$7F,$42,$76
	.dc.b	$F1,$BF,$14,$F0,$F4,$2C,$FF,$48,$F8,$7F,$86,$F2,$1D,$8D,$F7,$18
	.dc.b	$6C,$F0,$8F,$E6,$EB,$4C,$ED,$27,$A4,$E5,$ED,$F2,$BF,$E1,$F7,$C1
	.dc.b	$83,$C1,$EE,$D8,$DF,$B3,$E2,$AF,$F1,$5C,$60,$87,$F4,$87,$2D,$F6
	.dc.b	$78,$7B,$32,$3C,$31,$CF,$7E,$FE,$1C,$9F,$2C,$77,$DF,$51,$83,$D2
	.dc.b	$F8,$FD,$5F,$F0,$04,$7F,$16,$18,$FF,$72,$E4,$FF,$7F,$B6,$34,$D7
	.dc.b	$F8,$EA,$F5,$5F,$2B,$C1,$B3,$C1,$AF,$06,$BE,$1F,$95,$FD,$72,$69
	.dc.b	$1F,$31,$FC,$5C,$FE,$DE,$5B,$FF,$25,$C7,$F8,$33,$E5,$5C,$5F,$CC
	.dc.b	$BC,$8F,$F2,$35,$FE,$AB,$9F,$F1,$0D,$39,$C6,$B9,$77,$CD,$2B,$EF
	.dc.b	$F9,$96,$FC,$3E,$47,$43,$1A,$7F,$97,$9E,$FF,$19,$2F,$BC,$5A,$FF
	.dc.b	$03,$D7,$F9,$5D,$BC,$BF,$E5,$BE,$87,$C2,$FF,$97,$6B,$47,$EC,$2B
	.dc.b	$FC,$71,$CD,$93,$57,$71,$FD,$27,$CA,$FC,$57,$FC,$52,$DA,$FF,$18
	.dc.b	$E3,$FC,$5A,$77,$BF,$F9,$AE,$BE,$26,$3D,$DB,$3C,$A7,$5B,$FF,$9C
	.dc.b	$7E,$55,$F2,$17,$AF,$0D,$BF,$DB,$4F,$F9,$F3,$3F,$F2,$0B,$53,$4C
	.dc.b	$7C,$AB,$D2,$FF,$2F,$B6,$BF,$06,$DA,$FC,$D3,$5F,$EE,$3F,$2B,$FB
	.dc.b	$D9,$FC,$4B,$63,$F4,$0D,$77,$CB,$FE,$6C,$87,$A1,$49,$5C,$9F,$F3
	.dc.b	$8F,$9D,$FE,$1E,$8F,$F7,$F7,$AD,$82,$3F,$E1,$A8,$FF,$0F,$91,$C1
	.dc.b	$F9,$CF,$63,$F8,$53,$7C,$4F,$F1,$B9,$1F,$D0,$FC,$AE,$B3,$FD,$C3
	.dc.b	$5F,$F0,$7B,$5F,$F3,$3B,$5F,$F3,$77,$CA,$7F,$8C,$09,$3F,$E2,$E7
	.dc.b	$FC,$07,$D5,$23,$FC,$36,$D9,$EF,$E2,$FF,$A4,$2C,$D2,$57,$F8,$E3
	.dc.b	$8B,$FC,$7A,$47,$07,$F8,$9B,$B3,$FC,$22,$47,$07,$BB,$CC,$EC,$47
	.dc.b	$07,$F9,$13,$C7,$FF,$24,$91,$C1,$FE,$32,$F0,$3F,$C3,$A4,$70,$7C
	.dc.b	$AF,$2F,$D4,$47,$07,$F9,$43,$A3,$FC,$AA,$47,$07,$F8,$FB,$E4,$BF
	.dc.b	$C5,$24,$70,$7F,$16,$F9,$4F,$D1,$B7,$3F,$F5,$BC,$63,$6F,$E6,$0D
	.dc.b	$7F,$C8,$FE,$D7,$F0,$36,$F8,$3F,$E6,$62,$5F,$F9,$3B,$E5,$3F,$C7
	.dc.b	$24,$70,$7F,$26,$E6,$78,$BF,$EE,$56,$C7,$F6,$75,$F2,$99,$99,$3C
	.dc.b	$5F,$CE,$91,$C1,$C7,$BB,$78,$FF,$A8,$2F,$F6,$5E,$1F,$F1,$5F,$8F
	.dc.b	$7C,$A7,$FB,$D5,$B8,$7F,$77,$5F,$29,$99,$93,$97,$84,$63,$0F,$F5
	.dc.b	$6D,$3E,$4F,$A7,$CA,$75,$6E,$1D,$5B,$5C,$1B,$8F,$F2,$9C,$BF,$D4
	.dc.b	$4D,$F8,$0F,$8D,$E6,$7C,$76,$61,$F2,$DF,$DF,$91,$C1,$F2,$FB,$37
	.dc.b	$8F,$C3,$7E,$07,$CA,$49,$F8,$1A,$5A,$F5,$CD,$C7,$E4,$4D,$AF,$F6
	.dc.b	$FA,$F9,$4F,$95,$E2,$FC,$B6,$CF,$CB,$6E,$1F,$19,$B7,$EF,$F9,$4F
	.dc.b	$C1,$1F,$83,$F5,$ED,$FE,$0E,$3F,$8D,$91,$FC,$53,$6F,$EF,$76,$F0
	.dc.b	$6C,$FD,$BF,$94,$FB,$DF,$29,$F7,$DB,$19,$3F,$01,$19,$69,$F8,$3F
	.dc.b	$83,$FC,$31,$F8,$7F,$94,$B1,$92,$5F,$E0,$A3,$2F,$FB,$C3,$E4,$5E
	.dc.b	$17,$F9,$53,$5F,$EC,$D1,$8B,$FF,$23,$2F,$13,$50,$B9,$3B,$7A,$5F
	.dc.b	$F6,$F7,$42,$3F,$AC,$12,$35,$1C,$9E,$AF,$4E,$FA,$F1,$8B,$3F,$DF
	.dc.b	$2B,$EF,$FC,$A6,$BE,$07,$FC,$E3,$1F,$A7,$8C,$3F,$AB,$6B,$93,$FC
	.dc.b	$C5,$1C,$2D,$97,$F8,$1E,$1F,$29,$D4,$ED,$F2,$96,$F8,$C5,$BE,$31
	.dc.b	$FE,$D1,$9C,$5F,$CC,$F9,$3F,$49,$EA,$F7,$6D,$7C,$E8,$B7,$F5,$66
	.dc.b	$D7,$1A,$D1,$F5,$23,$2F,$CE,$CD,$78,$BF,$AB,$55,$FA,$51,$67,$FC
	.dc.b	$A5,$5D,$27,$F0,$93,$9A,$D3,$67,$57,$E3,$AD,$C7,$8C,$C9,$E3,$9E
	.dc.b	$11,$A2,$4F,$13,$F1,$F5,$DA,$B8,$CF,$A1,$AF,$A1,$B7,$E8,$8E,$71
	.dc.b	$DC,$CC,$F3,$8C,$9C,$CA,$73,$2C,$F3,$AF,$25,$73,$9C,$D7,$39,$CC
	.dc.b	$73,$6B,$CD,$B5,$CE,$B9,$46,$2F,$F0,$11,$EA,$7E,$2C,$66,$3D,$26
	.dc.b	$4B,$5E,$69,$DC,$74,$24,$7E,$CA,$3A,$BB,$8F,$F4,$0D,$7E,$11,$B3
	.dc.b	$FB,$09,$F8,$47,$ED,$8A,$7C,$13,$13,$C5,$FF,$50,$35,$CC,$7E,$D3
	.dc.b	$E5,$3F,$18,$4A,$FA,$ED,$9F,$9E,$DB,$F8,$ED,$9F,$5C,$FF,$07,$8C
	.dc.b	$3E,$D8,$FE,$27,$18,$8F,$A1,$1F,$A1,$8C,$5F,$9E,$BF,$8D,$C6,$2F
	.dc.b	$CD,$5E,$42,$5F,$91,$B5,$E4,$69,$E4,$B5,$F3,$8F,$19,$AA,$F3,$1F
	.dc.b	$CB,$63,$4D,$8D,$3E,$83,$F9,$4C,$6A,$BC,$F3,$AA,$F5,$3E,$D3,$17
	.dc.b	$EA,$BF,$41,$2B,$D0,$D9,$F4,$5F,$57,$94,$64,$F3,$CF,$81,$FE,$32
	.dc.b	$7F,$3D,$F1,$DE,$E7,$F3,$25,$F0,$56,$30,$DB,$FE,$89,$7F,$EA,$FA
	.dc.b	$DC,$60,$FE,$86,$F4,$71,$7D,$EB,$B7,$C1,$49,$CC,$AB,$F5,$8F,$96
	.dc.b	$74,$FA,$0D,$3E,$7D,$33,$87,$8F,$3F,$51,$EB,$CE,$E3,$DB,$1F,$CE
	.dc.b	$63,$E6,$AF,$D3,$46,$23,$E2,$46,$0F,$89,$F2,$9C,$A3,$C5,$28,$F9
	.dc.b	$1E,$FF,$C9,$F9,$4B,$87,$76,$DF,$06,$E5,$F2,$5F,$A1,$F2,$92,$BC
	.dc.b	$DF,$29,$6F,$BC,$DB,$FA,$13,$6B,$FB,$8F,$B5,$FE,$38,$6F,$8D,$F6
	.dc.b	$FE,$53,$FB,$6C,$E7,$1F,$EF,$8A,$7F,$83,$36,$BF,$D6,$11,$82,$3F
	.dc.b	$27,$E5,$39,$C7,$E1,$DF,$EF,$3F,$17,$FE,$5A,$BF,$AC,$FF,$96,$AB
	.dc.b	$EE,$B9,$CB,$FD,$EC,$E6,$77,$E7,$F2,$F3,$F9,$B6,$F8,$F9,$DD,$8E
	.dc.b	$FE,$7D,$8C,$92,$FC,$31,$97,$C7,$E6,$79,$8F,$3F,$99,$6B,$CB,$1F
	.dc.b	$C9,$F2,$34,$91,$FC,$A6,$BF,$C4,$93,$A8,$F6,$F9,$7C,$E7,$E0,$BF
	.dc.b	$13,$CC,$F8,$BE,$4D,$AF,$86,$DA,$F8,$ED,$DB,$E2,$BF,$33,$C9,$95
	.dc.b	$F2,$BC,$9B,$7F,$3A,$6D,$FC,$C9,$B5,$FE,$34,$C9,$E4,$FD,$63,$EB
	.dc.b	$F8,$D6,$7E,$AC,$7C,$12,$3E,$FC,$55,$CA,$30,$D8,$FD,$A7,$4F,$F8
	.dc.b	$B7,$1F,$BB,$7F,$E3,$C8,$C5,$FF,$91,$BC,$67,$BD,$FD,$55,$B1,$DD
	.dc.b	$B8,$7C,$F6,$BF,$EC,$2F,$94,$FD,$B9,$27,$CD,$DE,$8F,$EC,$18,$7F
	.dc.b	$8F,$27,$D4,$DA,$FB,$4D,$7F,$B2,$35,$F3,$7A,$5E,$52,$4F,$95,$FE
	.dc.b	$0F,$F2,$92,$3C,$2D,$9E,$BE,$4F,$88,$93,$E2,$7E,$07,$C0,$48,$F8
	.dc.b	$0D,$9F,$C7,$6F,$FE,$71,$9F,$9F,$1F,$95,$23,$EC,$F9,$1F,$79,$1C
	.dc.b	$1F,$2F,$D3,$F9,$C4,$70,$7D,$AF,$87,$FD,$01,$1C,$1F,$23,$E4,$7E
	.dc.b	$59,$1C,$1F,$85,$F9,$47,$C1,$FF,$83,$DB,$19,$8C,$19,$EA,$91,$C5
	.dc.b	$B9,$7D,$4E,$4F,$B2,$49,$FB,$2F,$F8,$6F,$E4,$D8,$BD,$AF,$CB,$46
	.dc.b	$1B,$AB,$4C,$D6,$F1,$FD,$D7,$C9,$BE,$4F,$F3,$D0,$8C,$3F,$A1,$6B
	.dc.b	$C2,$37,$6F,$26,$92,$7F,$51,$19,$7F,$39,$ED,$F6,$EA,$34,$91,$FA
	.dc.b	$CF,$8C,$F1,$F2,$1A,$48,$FC,$DF,$BF,$E8,$EA,$34,$91,$FB,$6F,$8A
	.dc.b	$E6,$47,$07,$E9,$3E,$03,$91,$1C,$1F,$C0,$7E,$4B,$FC,$0A,$FF,$76
	.dc.b	$F9,$37,$89,$FE,$C0,$F0,$DD,$1F,$EC,$29,$CD,$7F,$7C,$BE,$2F,$F7
	.dc.b	$97,$FC,$1A,$E7,$DA,$E5,$5F,$EC,$A9,$DD,$F4,$AF,$F6,$94,$70,$7F
	.dc.b	$C0,$5F,$0F,$F1,$5F,$F0,$4F,$AB,$F3,$1B,$C3,$D1,$FF,$6D,$4E,$DE
	.dc.b	$3A,$E5,$46,$70,$F0,$46,$38,$35,$2A,$E1,$F2,$7F,$EE,$7A,$F0,$F8
	.dc.b	$3C,$F2,$3A,$F8,$1D,$FC,$05,$AF,$F0,$FD,$F1,$7F,$C3,$8F,$FB,$D9
	.dc.b	$CF,$C0,$F0,$77,$3E,$08,$F0,$F0,$77,$F1,$25,$74,$8F,$2F,$BF,$A8
	.dc.b	$C1,$E0,$7D,$3E,$FF,$FB,$8A,$3F,$AE,$62,$8C,$47,$FA,$1E,$31,$C1
	.dc.b	$FD,$0D,$B1,$C3,$DF,$7C,$9E,$5B,$19,$8C,$1F,$AE,$5F,$27,$48,$FE
	.dc.b	$C9,$FD,$9F,$87,$F8,$1C,$3E,$9C,$FE,$EF,$9F,$FC,$8E,$8E,$63,$E4
	.dc.b	$FE,$DB,$93,$81,$FE,$49,$3F,$59,$1C,$B9,$BB,$3F,$EA,$5B,$B3,$C6
	.dc.b	$6B,$DF,$96,$BF,$83,$AF,$C8,$F5,$E2,$73,$E4,$C9,$F3,$C5,$AF,$17
	.dc.b	$C6,$F2,$31,$9F,$B9,$EE,$B9,$29,$E2,$E8,$7C,$46,$31,$FF,$34,$D7
	.dc.b	$F9,$BF,$0F,$F3,$33,$4F,$CD,$9C,$E3,$7C,$BE,$CF,$93,$F8,$B3,$FA
	.dc.b	$06,$CF,$C9,$C1,$F1,$DB,$E0,$E1,$C1,$2B,$FC,$F3,$7C,$32,$FA,$8F
	.dc.b	$CE,$7F,$CF,$1F,$27,$E1,$6E,$BF,$32,$3C,$D5,$F5,$DD,$67,$FB,$27
	.dc.b	$85,$7F,$B6,$5A,$EE,$BF,$DB,$D1,$88,$FC,$27,$C9,$BE,$63,$FD,$C4
	.dc.b	$DC,$3F,$C3,$8D,$9F,$F1,$E3,$5F,$F7,$33,$FD,$16,$37,$E5,$72,$23
	.dc.b	$CB,$BF,$A9,$1C,$1F,$E2,$9F,$92,$E9,$7F,$93,$6B,$F9,$96,$A6,$9E
	.dc.b	$17,$FC,$A0,$D7,$FD,$F4,$D7,$A5,$7F,$BF,$E3,$F8,$8C,$F8,$6B,$04
	.dc.b	$7F,$C7,$9F,$0B,$F8,$A6,$F1,$FF,$A0,$88,$FF,$90,$7C,$F7,$47,$EC
	.dc.b	$B2,$FD,$84,$62,$75,$39,$9F,$EE,$8D,$97,$C7,$52,$57,$57,$FC,$6F
	.dc.b	$EB,$C8,$FF,$1B,$B6,$FF,$C6,$FB,$7F,$8F,$09,$3F,$E3,$C7,$FC,$7F
	.dc.b	$AC,$8F,$F1,$FB,$67,$FC,$7F,$A7,$F9,$10,$93,$FE,$44,$7F,$C8,$F9
	.dc.b	$C8,$FF,$23,$B6,$7F,$C8,$FF,$0D,$FE,$4C,$24,$FF,$93,$1F,$F2,$7F
	.dc.b	$87,$23,$FC,$9E,$D9,$FF,$27,$FC,$C7,$F9,$50,$93,$FE,$54,$7F,$CA
	.dc.b	$FF,$2F,$23,$FC,$AE,$D9,$FF,$2B,$F4,$FF,$97,$09,$3F,$E5,$C7,$FC
	.dc.b	$BF,$E4,$C8,$FF,$2F,$B6,$7F,$CB,$FB,$FF,$99,$09,$3F,$E6,$47,$FC
	.dc.b	$CF,$F2,$1F,$E7,$73,$F8,$FF,$E8,$9F,$E2,$D9,$FF,$7A,$FA,$09,$D4
	.dc.b	$7F,$9B,$1E,$2D,$25,$FF,$BD,$E9,$F2,$DC,$23,$0F,$6F,$95,$EF,$E6
	.dc.b	$DB,$FE,$79,$DB,$FD,$CB,$04,$E7,$0E,$8D,$9D,$3E,$8F,$36,$5F,$23
	.dc.b	$FA,$15,$FF,$0C,$6C,$6A,$FC,$CF,$C0,$EE,$C6,$1E,$1F,$0F,$E1,$9F
	.dc.b	$13,$CD,$B8,$75,$C3,$FA,$96,$3F,$CC,$5F,$EA,$01,$C3,$E2,$AF,$D3
	.dc.b	$46,$23,$8C,$7D,$0F,$37,$AC,$7C,$97,$E9,$F9,$9F,$53,$E4,$2E,$1E
	.dc.b	$1A,$ED,$18,$7E,$83,$72,$F9,$4F,$DC,$F9,$19,$5C,$72,$7C,$8E,$30
	.dc.b	$FB,$15,$70,$FA,$53,$6F,$FD,$75,$D1,$FD,$93,$D7,$3F,$88,$F2,$B0
	.dc.b	$C9,$FF,$5E,$3F,$EB,$F2,$38,$3C,$5C,$DB,$CB,$F1,$0F,$E4,$BE,$47
	.dc.b	$F2,$7F,$23,$6B,$D0,$DA,$FC,$73,$73,$FC,$43,$F9,$AF,$6E,$57,$BF
	.dc.b	$E4,$6D,$FE,$56,$6D,$FE,$3E,$6D,$7E,$23,$E1,$7F,$46,$7E,$93,$E1
	.dc.b	$6C,$7C,$98,$FB,$64,$7F,$DD,$1C,$3F,$ED,$36,$C7,$F8,$2B,$F4,$5F
	.dc.b	$F5,$E6,$5F,$A5,$8C,$39,$F4,$75,$E8,$DB,$FB,$D7,$FD,$AF,$3F,$F2
	.dc.b	$AB,$FD,$BE,$CE,$67,$FC,$40,$E4,$C4,$7F,$89,$DF,$F0,$C2,$9E,$7E
	.dc.b	$9F,$97,$D4,$F3,$32,$FC,$2E,$5A,$7E,$47,$E4,$DC,$9F,$E2,$66,$C6
	.dc.b	$79,$B1,$32,$38,$3B,$8C,$63,$E3,$9F,$F1,$C5,$72,$8C,$1F,$E3,$F7
	.dc.b	$18,$FF,$9D,$9F,$CC,$CF,$42,$4F,$01,$1C,$23,$AD,$78,$2B,$95,$D9
	.dc.b	$C9,$CA,$32,$D3,$37,$D6,$FA,$3E,$5A,$E4,$F9,$EE,$BE,$28,$B3,$E8
	.dc.b	$8E,$51,$83,$E2,$57,$78,$E8,$D7,$D7,$5B,$8C,$57,$3B,$F3,$D7,$49
	.dc.b	$CD,$6E,$30,$7A,$5C,$4F,$03,$E6,$38,$9F,$A7,$7E,$4A,$E3,$5B,$9C
	.dc.b	$9F,$7C,$F1,$CF,$BA,$B9,$57,$A8,$F0,$4E,$8F,$29,$5E,$AF,$CE,$AE
	.dc.b	$D1,$C5,$AF,$C2,$8E,$11,$8B,$F8,$51,$E4,$9E,$67,$17,$A1,$99,$E1
	.dc.b	$7E,$D6,$C7,$F3,$A9,$F4,$1F,$CF,$2B,$F2,$8D,$AF,$F1,$04,$62,$3C
	.dc.b	$06,$67,$93,$F8,$16,$C7,$E2,$23,$F2,$BC,$9F,$84,$24,$FD,$C6,$C7
	.dc.b	$A8,$DC,$64,$8E,$65,$3C,$2F,$E2,$36,$BE,$01,$FA,$E7,$15,$F1,$4F
	.dc.b	$D7,$38,$7D,$07,$E7,$DC,$57,$A4,$EF,$3E,$2A,$EE,$D3,$BD,$74,$70
	.dc.b	$F7,$3D,$67,$A6,$BA,$DF,$F0,$B6,$CF,$F0,$F6,$CF,$8E,$39,$BE,$6A
	.dc.b	$FE,$09,$3A,$6C,$EA,$FD,$13,$98,$F8,$46,$6F,$C0,$D3,$C0,$5B,$F0
	.dc.b	$1E,$77,$C3,$9B,$D5,$FE,$7D,$ED,$7F,$3F,$6C,$FF,$3A,$6D,$7F,$15
	.dc.b	$AE,$13,$FC,$FA,$BF,$A3,$7C,$BC,$8F,$E8,$3F,$2D,$6F,$84,$DD,$3F
	.dc.b	$A1,$CD,$AF,$E5,$3C,$5F,$CC,$CF,$E6,$9C,$B6,$3F,$9C,$47,$E6,$08
	.dc.b	$FE,$51,$B8,$7F,$1D,$E5,$E0,$D9,$FE,$A7,$F2,$9F,$D6,$FE,$57,$FA
	.dc.b	$E3,$63,$27,$F5,$88,$CB,$4F,$EB,$7F,$2B,$FD,$8C,$FE,$C9,$F2,$B6
	.dc.b	$32,$4B,$F8,$D1,$97,$F9,$E7,$C6,$39,$FF,$DE,$AD,$AF,$F1,$23,$FD
	.dc.b	$7B,$64,$2A,$4E,$DE,$75,$FE,$99,$8C,$47,$F8,$E9,$AF,$10,$E1,$E7
	.dc.b	$1E,$2E,$2F,$1B,$6B,$FB,$CF,$CA,$7F,$80,$4F,$F0,$17,$CA,$D8,$C9
	.dc.b	$2F,$F0,$91,$97,$F0,$DF,$2B,$FE,$09,$F8,$FF,$F0,$53,$6B,$FC,$0A
	.dc.b	$D9,$DC,$7F,$82,$27,$E2,$ED,$23,$E2,$EF,$6B,$E5,$CF,$C1,$24,$FF
	.dc.b	$18,$D1,$18,$9F,$A3,$BF,$1D,$BE,$6C,$7C,$C8,$FF,$0F,$7C,$AC,$C3
	.dc.b	$FC,$33,$F2,$B6,$BF,$41,$36,$FF,$C3,$73,$6B,$FC,$33,$F1,$5F,$E2
	.dc.b	$7F,$94,$FF,$14,$36,$3F,$C4,$6D,$C3,$F5,$4D,$BF,$1C,$7D,$6D,$7F
	.dc.b	$0A,$7E,$1B,$6B,$7F,$86,$8F,$C0,$11,$F8,$0D,$9F,$C7,$E5,$C1,$B3
	.dc.b	$FE,$34,$F9,$5F,$F7,$07,$59,$A4,$8F,$E8,$3F,$A3,$7E,$49,$A7,$FB
	.dc.b	$9D,$C5,$7F,$4F,$6B,$FE,$8A,$5F,$2F,$B2,$4F,$47,$97,$CB,$48,$EE
	.dc.b	$DA,$E9,$F2,$5D,$C9,$3E,$47,$C5,$F2,$92,$3C,$ED,$9F,$27,$CB,$79
	.dc.b	$C9,$3E,$C7,$D5,$F2,$72,$3D,$ED,$9F,$67,$C4,$7B,$C9,$3F,$11,$F8
	.dc.b	$5F,$0F,$23,$E3,$B6,$7E,$27,$C3,$FC,$72,$4F,$CC,$7E,$57,$C3,$C8
	.dc.b	$F9,$ED,$9F,$EF,$5F,$37,$FE,$2F,$7F,$D6,$AD,$7F,$78,$D7,$F8,$33
	.dc.b	$57,$CD,$B9,$FF,$1B,$1F,$C0,$E3,$17,$FD,$F9,$B1,$FE,$09,$6B,$FE
	.dc.b	$C4,$F9,$86,$CF,$8B,$FF,$02,$C6,$1B,$C0,$F4,$73,$5B,$BF,$C3,$7C
	.dc.b	$5F,$F8,$31,$72,$A3,$0D,$8F,$C8,$C6,$FE,$CD,$F1,$74,$93,$FE,$16
	.dc.b	$8C,$BF,$E1,$8E,$57,$C1,$77,$6C,$66,$30,$78,$3E,$32,$47,$E8,$1A
	.dc.b	$FE,$76,$FC,$A4,$7F,$4A,$FA,$3E,$32,$47,$EA,$1A,$FE,$97,$DC,$F6
	.dc.b	$91,$FD,$6B,$F0,$3E,$32,$47,$EC,$1A,$FF,$8E,$FE,$59,$F0,$9F,$CE
	.dc.b	$FC,$4F,$E0,$FF,$2F,$F7,$6B,$FD,$D8,$D7,$FD,$15,$5C,$1F,$F7,$82
	.dc.b	$F8,$6E,$5D,$1F,$D1,$63,$53,$C3,$0C,$BF,$D4,$63,$83,$FE,$F9,$F5
	.dc.b	$DE,$EE,$A3,$FD,$57,$F1,$B7,$DA,$39,$AD,$CF,$1F,$7B,$17,$23,$07
	.dc.b	$F5,$FD,$7F,$91,$FC,$BB,$E2,$BA,$74,$4A,$D6,$DF,$D7,$6B,$51,$83
	.dc.b	$FB,$43,$F8,$07,$FD,$58,$E9,$B5,$E9,$8D,$5F,$83,$E2,$B1,$CF,$2F
	.dc.b	$A0,$A4,$F1,$E0,$F8,$1C,$B2,$B7,$5F,$33,$0F,$5F,$16,$E7,$37,$E1
	.dc.b	$8F,$17,$C5,$7F,$B7,$67,$FC,$F5,$68,$C4,$7F,$7B,$3C,$4E,$1F,$84
	.dc.b	$D7,$E0,$FC,$53,$E5,$FF,$CF,$74,$8C,$4F,$66,$BD,$63,$53,$FB,$78
	.dc.b	$7C,$BD,$23,$FB,$B7,$F7,$9E,$07,$F6,$CC,$FF,$87,$B7,$83,$FD,$29
	.dc.b	$3F,$E1,$BF,$97,$72,$FA,$EB,$58,$7F,$85,$F9,$2B,$FE,$21,$6B,$FC
	.dc.b	$37,$0F,$A6,$FE,$49,$BC,$BA,$38,$F3,$BC,$BC,$4F,$0F,$73,$99,$8B
	.dc.b	$FF,$78,$1F,$E2,$DD,$7F,$93,$FB,$BE,$03,$75,$E0,$7C,$B1,$F8,$28
	.dc.b	$FC,$2B,$67,$FD,$F5,$83,$C0,$DF,$3F,$0E,$69,$87,$08,$B7,$FF,$00
	.dc.b	$7E,$5F,$DC,$DD,$FD,$7F,$27,$F1,$27,$39,$FF,$82,$1F,$85,$5C,$78
	.dc.b	$35,$7B,$7E,$5F,$CB,$F9,$67,$F1,$4D,$7F,$C8,$CD,$7F,$CA,$0D,$4F
	.dc.b	$B6,$FF,$C2,$EF,$F0,$BC,$BF,$4A,$E5,$E7,$D3,$F6,$BE,$5D,$E7,$FF
	.dc.b	$52,$C3,$FC,$B9,$18,$AF,$F6,$7B,$63,$FD,$B4,$D7,$F4,$0F,$F4,$78
	.dc.b	$DF,$7F,$39,$4E,$5E,$31,$9F,$9C,$FF,$30,$41,$A4,$E4,$E5,$F3,$1D
	.dc.b	$E7,$FC,$D0,$D7,$FC,$E4,$D4,$D3,$C0,$FE,$D5,$AF,$2A,$EA,$79,$FE
	.dc.b	$6D,$F2,$D4,$99,$7F,$BC,$23,$2F,$FB,$CB,$CF,$FC,$D3,$7A,$7E,$04
	.dc.b	$63,$0F,$F8,$00,$D7,$F8,$32,$F9,$8D,$92,$BA,$3F,$27,$E3,$9F,$2D
	.dc.b	$49,$3F,$F0,$01,$F9,$DF,$1E,$F6,$29,$27,$F8,$A6,$3F,$E9,$79,$D4
	.dc.b	$62,$3F,$DF,$BC,$5C,$89,$19,$9F,$99,$E9,$3A,$BF,$C9,$C4,$9F,$F2
	.dc.b	$9B,$E2,$ED,$91,$C5,$B8,$78,$AF,$FC,$0F,$3D,$7D,$0F,$49,$1F,$F2
	.dc.b	$EF,$99,$FE,$5B,$23,$83,$C9,$E8,$7C,$02,$38,$3D,$1E,$87,$C4,$23
	.dc.b	$83,$FC,$D3,$C1,$FE,$67,$23,$83,$E0,$57,$CC,$23,$83,$E2,$47,$D0
	.dc.b	$23,$83,$FC,$EB,$F3,$1F,$E7,$32,$38,$3E,$87,$CB,$30,$FE,$56,$D8
	.dc.b	$FE,$B7,$F2,$6F,$17,$35,$27,$8B,$FA,$DA,$E1,$AF,$DE,$6B,$FD,$4D
	.dc.b	$CF,$81,$FE,$9A,$C1,$B7,$3A,$4E,$63,$F9,$EC,$BE,$5A,$FC,$B6,$1B
	.dc.b	$97,$99,$B8,$72,$8F,$F6,$E5,$71,$7D,$75,$D3,$E5,$A4,$7B,$BE,$4E
	.dc.b	$DF,$13,$FD,$D9,$6A,$CE,$5E,$9A,$B7,$DB,$E4,$FC,$E7,$A3,$E4,$EC
	.dc.b	$7C,$A8,$FE,$25,$81,$A4,$8F,$D1,$46,$5F,$F0,$5F,$C3,$FF,$BF,$5E
	.dc.b	$46,$32,$FE,$C3,$1F,$CF,$A3,$FD,$F4,$FF,$3E,$C7,$FD,$8E,$E6,$41
	.dc.b	$D4,$F0,$F2,$79,$F9,$3D,$1B,$1E,$2F,$3B,$B9,$E0,$F1,$AC,$64,$98
	.dc.b	$7C,$93,$B4,$69,$C7,$81,$F8,$B9,$FF,$13,$B6,$33,$F2,$DF,$DF,$23
	.dc.b	$19,$FF,$B8,$1D,$4F,$F8,$75,$AF,$B0,$38,$7D,$ED,$73,$CD,$D7,$93
	.dc.b	$B3,$70,$F9,$FE,$7F,$CC,$F4,$3E,$6B,$63,$27,$D4,$8C,$B4,$FA,$FF
	.dc.b	$2D,$F5,$0F,$AB,$F2,$D6,$32,$4C,$7E,$37,$0A,$75,$1F,$1B,$8F,$E1
	.dc.b	$1F,$0F,$96,$DF,$C8,$8F,$E6,$84,$7F,$95,$6B,$F8,$9A,$FE,$5E,$DD
	.dc.b	$BE,$57,$2F,$29,$FA,$6F,$E2,$BE,$4F,$F1,$7F,$25,$70,$FC,$8B,$6B
	.dc.b	$CE,$DB,$FC,$6B,$F9,$2F,$92,$95,$F9,$8F,$96,$B7,$EE,$9B,$7F,$94
	.dc.b	$9B,$5F,$67,$E5,$3F,$C5,$8D,$E6,$FF,$1E,$C3,$9E,$C4,$9F,$F3,$7C
	.dc.b	$F0,$C7,$F3,$AE,$E3,$33,$B9,$FC,$A4,$ED,$3B,$73,$1B,$C1,$5E,$33
	.dc.b	$BD,$E5,$C5,$70,$72,$78,$08,$E4,$F0,$C7,$17,$CF,$55,$F1,$C5,$BF
	.dc.b	$31,$1C,$4E,$5B,$5D,$AF,$CF,$7F,$0A,$72,$F8,$63,$2F,$69,$E0,$7A
	.dc.b	$A3,$37,$DE,$33,$3E,$66,$BC,$4E,$8F,$12,$3E,$17,$71,$EA,$BF,$B7
	.dc.b	$7E,$4B,$EF,$18,$AD,$BE,$D6,$9C,$A7,$83,$DA,$FC,$55,$CA,$37,$1F
	.dc.b	$39,$B1,$98,$C1,$E8,$3D,$85,$39,$C7,$5B,$F4,$3F,$72,$35,$18,$79
	.dc.b	$CF,$53,$B3,$FE,$B1,$8F,$F2,$59,$2F,$FC,$A4,$DA,$FF,$10,$36,$7F
	.dc.b	$C4,$AF,$E0,$39,$3F,$04,$49,$F9,$6D,$9F,$1B,$6F,$C5,$3F,$AB,$8C
	.dc.b	$3E,$B6,$CF,$C0,$9F,$BF,$18,$8F,$D1,$DF,$EC,$23,$17,$E3,$BF,$09
	.dc.b	$2B,$C2,$DB,$F0,$C7,$EC,$AB,$8D,$F3,$AD,$1E,$32,$CF,$8D,$F4,$18
	.dc.b	$7C,$6F,$EF,$A7,$9B,$4D,$5F,$96,$7C,$64,$BF,$1B,$FB,$EA,$E2,$FF
	.dc.b	$02,$6C,$7C,$2A,$E7,$7E,$7A,$E5,$1E,$76,$9E,$73,$DC,$F0,$3C,$8F
	.dc.b	$F3,$CF,$96,$FE,$EA,$7F,$76,$F9,$6B,$3F,$DF,$22,$CF,$F4,$29,$B7
	.dc.b	$FD,$E3,$E5,$3F,$BE,$FC,$A7,$F7,$E6,$E1,$FE,$04,$6D,$FF,$4A,$6D
	.dc.b	$7E,$EB,$7F,$E5,$27,$F2,$AE,$0B,$7F,$CA,$23,$F2,$64,$7F,$26,$DB
	.dc.b	$FE,$4F,$C1,$C1,$B3,$FE,$14,$F9,$4F,$F0,$89,$FE,$12,$F9,$6B,$19
	.dc.b	$24,$FF,$5E,$8C,$BF,$D8,$3E,$3F,$FC,$33,$F2,$DF,$E1,$A6,$D7,$F8
	.dc.b	$81,$B3,$B8,$FF,$11,$CF,$F4,$4D,$64,$7F,$42,$D6,$D7,$F3,$E9,$FD
	.dc.b	$79,$27,$F5,$ED,$8F,$E8,$78,$F0,$D7,$F9,$0B,$FC,$A9,$FE,$4F,$F0
	.dc.b	$CF,$53,$18,$7F,$83,$E3,$0F,$F8,$B2,$77,$EB,$EB,$0F,$F1,$7D,$69
	.dc.b	$FF,$12,$36,$7A,$DE,$E5,$3F,$EB,$C7,$71,$98,$7C,$C5,$F8,$9C,$37
	.dc.b	$1F,$11,$E5,$71,$7E,$06,$E1,$EC,$F8,$9E,$87,$5F,$82,$B7,$CA,$FD
	.dc.b	$33,$FE,$E9,$C4,$D2,$4F,$FA,$E6,$32,$F1,$9F,$5F,$C1,$F4,$7E,$0B
	.dc.b	$EB,$F8,$19,$3B,$A7,$C4,$E2,$FE,$3D,$5C,$BE,$9C,$DB,$FA,$5F,$13
	.dc.b	$F1,$7D,$DF,$8C,$D8,$E8,$DC,$3E,$8B,$6F,$FB,$1F,$94,$F8,$CF,$A4
	.dc.b	$DB,$F2,$7C,$C6,$26,$47,$FB,$4D,$FF,$6F,$91,$C1,$EB,$F3,$77,$8F
	.dc.b	$E0,$1F,$BD,$EE,$FF,$83,$27,$FD,$C0,$D9,$FF,$72,$B5,$FF,$2E,$07
	.dc.b	$0F,$F9,$7D,$7C,$BF,$F7,$C8,$C3,$5F,$F3,$FC,$6A,$7B,$CF,$DB,$F3
	.dc.b	$E4,$7D,$96,$D7,$8B,$E4,$79,$3E,$16,$D9,$A6,$3E,$46,$5F,$B1,$B5
	.dc.b	$FF,$00,$38,$BF,$D7,$F3,$F8,$7D,$7F,$18,$47,$E1,$FC,$B7,$62,$38
	.dc.b	$3E,$3F,$C2,$78,$88,$E0,$FC,$97,$B3,$F9,$72,$38,$3E,$7F,$C2,$7A
	.dc.b	$88,$E0,$FA,$7F,$09,$EE,$23,$83,$F3,$DE,$DF,$E8,$C8,$E0,$FA,$F9
	.dc.b	$FF,$AB,$BE,$51,$E2,$FC,$A6,$CE,$A5,$18,$78,$B5,$FC,$0F,$CA,$37
	.dc.b	$FE,$25,$1D,$A3,$1E,$0F,$37,$EA,$7B,$56,$B7,$5F,$8A,$F9,$B6,$9E
	.dc.b	$18,$46,$1F,$23,$5F,$5B,$5F,$C9,$FC,$AB,$E1,$B9,$36,$3D,$8D,$7E
	.dc.b	$1B,$5F,$CD,$F5,$7F,$0A,$7E,$13,$5F,$92,$D7,$7E,$17,$9E,$1C,$14
	.dc.b	$93,$C9,$F6,$74,$77,$8F,$D3,$61,$82,$3F,$B0,$E7,$D9,$1C,$1F,$C6
	.dc.b	$FE,$03,$E3,$47,$FC,$01,$8C,$10,$FD,$BF,$C3,$78,$08,$FE,$F1,$E3
	.dc.b	$ED,$C8,$FE,$4C,$D7,$F0,$FF,$28,$F1,$BF,$DB,$5E,$17,$F9,$DB,$E5
	.dc.b	$7F,$13,$5F,$CA,$DA,$FF,$BB,$6B,$83,$FC,$BD,$78,$FC,$BA,$7F,$DC
	.dc.b	$91,$A9,$E1,$86,$50,$8E,$2F,$F3,$9F,$81,$7C,$16,$E1,$E1,$BE,$23
	.dc.b	$0D,$8E,$13,$BA,$FF,$5C,$46,$08,$34,$FF,$49,$C6,$A3,$19,$7F,$BE
	.dc.b	$FE,$55,$E7,$76,$F8,$D9,$5F,$D3,$67,$2D,$E8,$F3,$46,$23,$99,$EE
	.dc.b	$71,$8F,$FA,$89,$FC,$83,$7C,$EE,$35,$2B,$B4,$74,$DF,$DF,$C9,$27
	.dc.b	$C5,$1E,$0E,$1D,$46,$0F,$33,$E4,$F3,$BE,$04,$7F,$06,$CB,$FA,$FE
	.dc.b	$5F,$C0,$DB,$19,$9D,$BF,$D0,$A1,$E6,$BE,$57,$83,$67,$A3,$5E,$0D
	.dc.b	$7F,$8A,$FC,$AF,$F0,$14,$D2,$3F,$C5,$7C,$2F,$F0,$A2,$6B,$C0,$7C
	.dc.b	$AB,$8F,$FA,$BC,$F0,$E4,$9C,$9F,$E0,$3B,$E4,$D8,$EB,$F0,$32,$3F
	.dc.b	$C1,$1C,$CF,$2F,$17,$F6,$8F,$CB,$FB,$FE,$43,$3C,$3F,$D8,$5F,$3F
	.dc.b	$C0,$4C,$3C,$11,$6B,$95,$39,$BF,$C6,$0D,$AF,$0F,$C7,$62,$73,$CD
	.dc.b	$27,$FC,$39,$5B,$BE,$73,$FE,$39,$F9,$5F,$6C,$79,$DB,$3F,$EF,$0D
	.dc.b	$1F,$21,$BE,$0E,$1C,$13,$2F,$F1,$46,$1C,$1F,$DA,$E2,$8F,$D9,$C7
	.dc.b	$F8,$B7,$E5,$7C,$59,$7F,$BF,$1F,$E0,$31,$9D,$3F,$C9,$91,$8B,$F9
	.dc.b	$4D,$7F,$C1,$AD,$7B,$B5,$FE,$0A,$D5,$F2,$AE,$DF,$F0,$2B,$6F,$FA
	.dc.b	$E3,$67,$FB,$1B,$5F,$F0,$DF,$CA,$B8,$7F,$C1,$AD,$AF,$EC,$8D,$9F
	.dc.b	$ED,$6D,$78,$FB,$BC,$09,$3C,$1F,$F0,$BF,$CA,$F7,$9F,$92,$D7,$F8
	.dc.b	$FB,$53,$7E,$57,$FC,$E1,$1F,$E3,$06,$BF,$49,$AF,$F2,$86,$AF,$9B
	.dc.b	$F3,$3F,$61,$AF,$F2,$E6,$BB,$C7,$FC,$5B,$E9,$BE,$5A,$93,$2F,$F0
	.dc.b	$34,$65,$FF,$04,$7C,$73,$E0,$3F,$C6,$AD,$8C,$C6,$0F,$F1,$CF,$CD
	.dc.b	$FF,$77,$8F,$F0,$83,$71,$FF,$0B,$B5,$FF,$27,$B5,$FC,$8F,$BB,$FE
	.dc.b	$08,$9F,$F0,$C3,$63,$FC,$3E,$D7,$FC,$AE,$D7,$FC,$AD,$D1,$FE,$48
	.dc.b	$24,$FF,$98,$1F,$F4,$67,$C8,$48,$FF,$45,$36,$7A,$FA,$3F,$E5,$2B
	.dc.b	$F2,$47,$F5,$86,$A6,$FA,$5F,$D5,$6F,$FC,$DB,$E6,$FF,$96,$08,$FF
	.dc.b	$A5,$FE,$43,$FD,$38,$47,$07,$BB,$D2,$F0,$11,$C1,$FE,$78,$E0,$FF
	.dc.b	$31,$91,$C1,$FE,$3D,$FD,$1F,$FB,$6B,$72,$FF,$35,$C6,$1F,$F2,$3B
	.dc.b	$5F,$93,$3E,$F2,$4F,$D0,$7F,$CE,$3F,$31,$23,$FD,$12,$D9,$E7,$E0
	.dc.b	$76,$24,$F8,$DF,$C3,$F8,$3F,$EC,$FE,$3F,$EF,$73,$FE,$B7,$5F,$09
	.dc.b	$99,$93,$F9,$07,$8B,$49,$7F,$CD,$6B,$FC,$FA,$D9,$ED,$BF,$FB,$77
	.dc.b	$E5,$1E,$35,$F2,$70,$DC,$7F,$D9,$13,$A6,$CE,$9A,$6B,$E5,$7A,$1D
	.dc.b	$7E,$56,$C6,$49,$97,$F4,$9B,$4E,$A3,$D1,$CD,$F0,$CF,$89,$EC,$5B
	.dc.b	$F8,$91,$FA,$E2,$3F,$B2,$6C,$FC,$1E,$4E,$0D,$9F,$5F,$C8,$FC,$1F
	.dc.b	$91,$F8,$4D,$8C,$9F,$02,$32,$D3,$E1,$FC,$8F,$C9,$3E,$57,$CA,$D8
	.dc.b	$C9,$2F,$DD,$19,$7F,$B4,$7C,$AF,$F2,$BF,$2D,$5F,$D3,$23,$0D,$F4
	.dc.b	$73,$5C,$3C,$3B,$78,$85,$FF,$1A,$24,$FF,$B1,$9B,$7F,$F0,$05,$FB
	.dc.b	$53,$CE,$31,$C7,$F8,$3F,$95,$91,$F8,$0F,$95,$B5,$F5,$E6,$E3,$E1
	.dc.b	$E3,$56,$2F,$FB,$2C,$3E,$4F,$F1,$DF,$27,$F8,$F6,$CF,$E3,$1B,$8F
	.dc.b	$E0,$DB,$87,$B7,$37,$07,$C0,$BF,$8C,$DB,$FF,$33,$CB,$87,$11,$F1
	.dc.b	$26,$CF,$2A,$FF,$5F,$36,$7F,$D8,$AD,$7F,$DF,$AD,$7B,$C7,$C9,$E1
	.dc.b	$E7,$5D,$1F,$D5,$79,$72,$7F,$4F,$F2,$B6,$FF,$3D,$37,$2F,$91,$36
	.dc.b	$BE,$3F,$C6,$7E,$CF,$E5,$7F,$68,$D8,$FD,$8B,$70,$FD,$43,$6F,$E3
	.dc.b	$6D,$EE,$3D,$FB,$5B,$FC,$04,$5B,$FC,$8C,$DA,$E3,$5A,$3F,$0D,$19
	.dc.b	$7F,$29,$56,$BE,$7C,$D7,$F2,$31,$75,$FC,$05,$73,$9F,$CF,$CE,$6B
	.dc.b	$4D,$9D,$5F,$8A,$B7,$1E,$23,$27,$88,$AF,$89,$AF,$89,$B7,$E2,$8E
	.dc.b	$64,$9E,$C7,$E3,$6F,$C7,$5C,$1B,$3B,$8F,$DA,$46,$27,$8C,$60,$EF
	.dc.b	$7E,$BA,$E0,$7E,$51,$C4,$F0,$8C,$4F,$73,$89,$5E,$27,$AC,$B7,$EB
	.dc.b	$9E,$51,$96,$C6,$67,$E3,$4F,$0A,$FC,$1C,$76,$8E,$77,$EB,$8D,$DF
	.dc.b	$C3,$74,$F4,$8F,$87,$3F,$A4,$AF,$E1,$84,$AF,$D5,$B6,$3F,$56,$D7
	.dc.b	$84,$7F,$11,$CF,$F0,$C4,$7E,$BB,$63,$DE,$DB,$F6,$B6,$7C,$E7,$F2
	.dc.b	$77,$13,$F3,$0F,$E5,$AE,$1F,$8D,$1B,$3B,$91,$EE,$D9,$EE,$D3,$BC
	.dc.b	$7F,$0E,$9E,$98,$2F,$F8,$6C,$61,$B5,$C6,$BC,$97,$FC,$6E,$B5,$3F
	.dc.b	$0C,$D4,$F9,$8E,$CE,$2F,$B3,$B8,$F3,$92,$BC,$ED,$8F,$3B,$FD,$12
	.dc.b	$BB,$BF,$12,$3A,$5F,$9A,$38,$1E,$68,$F8,$CE,$0F,$21,$CD,$E2,$78
	.dc.b	$EB,$FA,$97,$C0,$3C,$CF,$E6,$2D,$95,$7F,$B1,$E1,$F2,$BA,$A9,$5A
	.dc.b	$BE,$71,$CA,$73,$7C,$7E,$57,$AF,$C2,$76,$6E,$1C,$4F,$F5,$AE,$28
	.dc.b	$CD,$71,$9D,$3E,$49,$57,$A3,$E1,$25,$6A,$95,$FE,$BC,$8E,$B3,$A9
	.dc.b	$FF,$61,$9C,$DF,$EE,$8E,$A3,$2E,$5A,$F4,$8F,$8F,$ED,$F3,$CF,$E5
	.dc.b	$BF,$2B,$E1,$25,$F8,$F0,$7C,$26,$31,$F9,$95,$70,$FA,$73,$6F,$E1
	.dc.b	$FC,$17,$D4,$F5,$3E,$AB,$63,$A3,$70,$FA,$CD,$BF,$F0,$77,$8E,$F8
	.dc.b	$9F,$ED,$AD,$BF,$27,$CA,$E2,$64,$7F,$A5,$1F,$F4,$31,$1C,$1E,$BF
	.dc.b	$33,$78,$FE,$15,$FE,$F3,$E8,$7F,$BB,$21,$8F,$F8,$7E,$30,$DF,$13
	.dc.b	$7F,$01,$30,$FB,$15,$A7,$FD,$EE,$A7,$57,$D2,$73,$E2,$F8,$3C,$5F
	.dc.b	$0B,$63,$9B,$67,$81,$F9,$CB,$FF,$1B,$B6,$BD,$3E,$64,$BF,$1F,$99
	.dc.b	$6B,$88,$B5,$7F,$A4,$89,$5F,$E9,$76,$CF,$09,$FF,$33,$CF,$BA,$7E
	.dc.b	$3F,$99,$F2,$3C,$59,$3F,$0B,$C5,$B8,$71,$9B,$B7,$CB,$9B,$5F,$1B
	.dc.b	$C5,$FA,$3E,$4F,$D2,$6C,$7C,$C6,$E1,$E3,$6D,$FD,$5F,$27,$EC,$9F
	.dc.b	$6B,$C9,$B7,$F5,$23,$FD,$5E,$47,$FD,$68,$DB,$F7,$78,$BC,$1B,$3F
	.dc.b	$B3,$F5,$1F,$2B,$FA,$E9,$FE,$1B,$18,$9F,$D6,$B6,$3F,$62,$D7,$F6
	.dc.b	$FD,$DF,$EA,$E7,$F6,$AD,$73,$E7,$F5,$7F,$80,$BF,$EC,$6F,$1A,$5F
	.dc.b	$FB,$2D,$B3,$DB,$E3,$F8,$60,$69,$2B,$F8,$0E,$1F,$EB,$4A,$F1,$F6
	.dc.b	$F7,$8C,$47,$F1,$A8,$FF,$6A,$F9,$12,$BF,$DB,$4D,$9F,$6F,$C2,$78
	.dc.b	$09,$3F,$09,$E9,$F2,$72,$3C,$2D,$9F,$F6,$FF,$9D,$FE,$E2,$24,$FF
	.dc.b	$B9,$DF,$5F,$C8,$48,$F9,$AD,$9F,$93,$F1,$DE,$92,$4F,$B9,$FF,$76
	.dc.b	$F9,$F2,$3F,$DE,$4D,$9F,$E4,$BF,$23,$F1,$FC,$FF,$E4,$AD,$8C,$FC
	.dc.b	$AB,$D8,$CF,$0F,$F4,$F9,$FE,$5B,$F2,$4D,$3F,$96,$C2,$30,$FF,$30
	.dc.b	$6B,$C2,$37,$6F,$91,$A4,$9F,$E6,$F1,$97,$F9,$CF,$92,$F4,$3B,$B6
	.dc.b	$33,$18,$33,$BC,$8E,$2D,$7F,$A2,$FC,$FF,$F8,$2D,$FE,$8A,$D7,$3D
	.dc.b	$1E,$C8,$C4,$7F,$50,$E8,$FE,$AA,$FE,$83,$A3,$E0,$C6,$23,$F4,$31
	.dc.b	$F1,$0F,$53,$FA,$0C,$F0,$47,$FA,$AF,$8A,$F5,$BE,$5B,$63,$31,$83
	.dc.b	$A7,$B5,$23,$F3,$AD,$7F,$B0,$CB,$E7,$79,$92,$3B,$BF,$63,$E4,$9E
	.dc.b	$AF,$F8,$22,$7F,$93,$FC,$9B,$A7,$1E,$2F,$FB,$86,$31,$3F,$DD,$EF
	.dc.b	$8B,$7D,$6D,$F6,$73,$9C,$CF,$F3,$26,$DF,$FB,$AD,$57,$09,$FF,$0C
	.dc.b	$F9,$8F,$47,$10,$F1,$1F,$05,$86,$C7,$F8,$26,$B8,$37,$D2,$E3,$18
	.dc.b	$8F,$E8,$6D,$CB,$FD,$F8,$FF,$57,$8F,$F0,$97,$C9,$FF,$85,$9F,$F0
	.dc.b	$A5,$BE,$33,$C0,$79,$27,$53,$FD,$39,$FC,$A3,$7D,$AD,$1B,$EC,$E1
	.dc.b	$F1,$D6,$7B,$C6,$1F,$0F,$1F,$B7,$A6,$46,$0F,$2F,$07,$86,$64,$60
	.dc.b	$F4,$FB,$7F,$DE,$9F,$F1,$1E,$5F,$E1,$AD,$BF,$C4,$6D,$8C,$CE,$DF
	.dc.b	$F1,$6C,$3D,$A7,$C9,$F0,$6C,$F4,$6B,$C1,$AF,$F8,$CB,$E4,$FF,$82
	.dc.b	$A6,$91,$FF,$1A,$F8,$EE,$7F,$F2,$14,$7F,$11,$6B,$C4,$7C,$9F,$F7
	.dc.b	$A7,$F3,$13,$C5,$FE,$FE,$7F,$84,$CF,$CE,$73,$E7,$4F,$F2,$3C,$DB
	.dc.b	$FF,$25,$CD,$AF,$D1,$73,$57,$FC,$D6,$DE,$57,$9B,$99,$FF,$0F,$5F
	.dc.b	$A3,$DB,$C1,$9B,$B1,$EC,$F6,$F8,$18,$AF,$3B,$DF,$3F,$F3,$04,$7F
	.dc.b	$9D,$79,$BC,$1F,$27,$97,$5C,$93,$0F,$F3,$45,$F0,$BF,$F1,$CC,$FA
	.dc.b	$7E,$4F,$E5,$47,$F8,$89,$B3,$F1,$E3,$F5,$9F,$19,$BE,$0E,$1C,$13
	.dc.b	$1F,$F3,$A6,$1C,$1F,$E0,$CF,$CD,$B7,$C9,$FF,$99,$A7,$FC,$F4,$D8
	.dc.b	$FF,$18,$BF,$8B,$8C,$F3,$FC,$D8,$D4,$FF,$9F,$1A,$FF,$13,$6B,$FE
	.dc.b	$9C,$6A,$F9,$36,$DF,$E0,$96,$DF,$FA,$01,$B3,$FE,$84,$6B,$FE,$8A
	.dc.b	$5F,$1D,$C4,$93,$C5,$E9,$D7,$23,$A3,$6B,$FD,$1F,$F2,$5D,$1F,$E5
	.dc.b	$4D,$7F,$D6,$6D,$4D,$3C,$33,$FE,$96,$6C,$74,$AF,$F1,$24,$62,$3F
	.dc.b	$D4,$2F,$FA,$9A,$78,$56,$88,$FF,$A8,$3D,$87,$CE,$7F,$8B,$5B,$19
	.dc.b	$8C,$19,$F9,$79,$1C,$5B,$87,$FA,$B2,$BF,$C0,$13,$D7,$D3,$E8,$47
	.dc.b	$FD,$5F,$F2,$5E,$5A,$FF,$59,$B5,$FF,$5C,$B5,$FE,$A0,$D7,$FD,$B3
	.dc.b	$FA,$47,$F0,$39,$FF,$5C,$B6,$3F,$D7,$AD,$7F,$D8,$6D,$7F,$AD,$7D
	.dc.b	$A3,$E5,$3F,$C9,$8D,$AF,$F6,$1B,$67,$FD,$8E,$D7,$9F,$BF,$FE,$53
	.dc.b	$BF,$F2,$D4,$7F,$65,$6B,$E6,$8D,$7C,$FC,$BF,$3B,$67,$CF,$EC,$F3
	.dc.b	$24,$FB,$1E,$DE,$DC,$8F,$7B,$67,$DF,$ED,$78,$C9,$3F,$11,$F3,$7C
	.dc.b	$DC,$8F,$8E,$D9,$F8,$FE,$1F,$AC,$93,$F3,$1F,$77,$14,$8F,$9E,$D9
	.dc.b	$F9,$FC,$5F,$0C,$93,$F5,$1F,$8D,$C3,$FC,$9F,$6F,$C2,$D7,$FA,$C1
	.dc.b	$FE,$D1,$5F,$EF,$D7,$3D,$DF,$C4,$5D,$CF,$DA,$AE,$75,$F8,$06,$C7
	.dc.b	$3E,$8E,$D1,$DD,$F0,$73,$F8,$79,$AD,$F7,$6D,$F8,$DB,$8F,$85,$F4
	.dc.b	$7C,$F4,$AF,$37,$2D,$BF,$54,$DB,$F4,$4D,$AF,$C2,$77,$FC,$13,$E1
	.dc.b	$7A,$16,$3E,$04,$7E,$2C,$8F,$E7,$1B,$87,$C7,$F9,$7E,$0D,$9F,$91
	.dc.b	$F2,$FF,$D4,$FF,$49,$7C,$17,$CF,$8C,$36,$55,$FE,$E0,$AD,$C7,$23
	.dc.b	$39,$FF,$41,$8C,$CE,$FE,$4F,$B1,$DF,$E4,$ED,$F0,$3F,$C8,$D4,$9C
	.dc.b	$D7,$7A,$B7,$FD,$37,$BB,$F5,$CD,$DB,$D5,$EA,$7F,$6D,$9C,$E5,$FC
	.dc.b	$0A,$7F,$D8,$51,$FB,$3C,$31,$3F,$C3,$27,$27,$C3,$F5,$39,$36,$3C
	.dc.b	$D3,$F0,$B3,$FE,$C8,$7F,$97,$A3,$30,$8C,$F9,$3F,$37,$CA,$F9,$CD
	.dc.b	$AD,$CF,$1A,$F9,$9E,$A7,$D3,$3E,$A7,$91,$6B,$71,$67,$E9,$4D,$D7
	.dc.b	$E8,$7A,$5F,$6C,$FB,$9F,$27,$70,$EF,$1F,$EE,$3C,$7F,$B4,$61,$FE
	.dc.b	$75,$3D,$91,$92,$4F,$9B,$E4,$F9,$4F,$E0,$5F,$C4,$7C,$9F,$E2,$7D
	.dc.b	$3B,$5E,$C6,$D7,$E1,$DB,$97,$E1,$5F,$C8,$79,$52,$BE,$1F,$C4,$DB
	.dc.b	$FC,$74,$DB,$FC,$64,$DA,$FE,$67,$D7,$FE,$2B,$F2,$FF,$CF,$31,$F4
	.dc.b	$97,$66,$66,$4E,$5E,$11,$8C,$58,$7F,$86,$6F,$9C,$62,$FB,$9D,$6B
	.dc.b	$85,$6E,$EC,$FF,$46,$9B,$5F,$A8,$BE,$77,$BA,$E4,$78,$C9,$3D,$CE
	.dc.b	$17,$E6,$8C,$CF,$19,$D0,$AE,$B3,$E1,$36,$7A,$23,$71,$A7,$C9,$7E
	.dc.b	$67,$95,$D9,$F9,$35,$CA,$7C,$6D,$7C,$15,$BA,$E0,$D9,$D5,$72,$AF
	.dc.b	$0D,$65,$F1,$D5,$BF,$61,$F4,$5C,$5F,$A6,$30,$F4,$AE,$ED,$3B,$CF
	.dc.b	$0A,$F3,$9D,$A3,$4F,$89,$FA,$2D,$25,$78,$A3,$E5,$46,$1F,$11,$1F
	.dc.b	$92,$F4,$8E,$11,$83,$81,$E2,$72,$FC,$88,$E0,$73,$70,$70,$3E,$9B
	.dc.b	$FE,$0C,$AF,$F0,$79,$1C,$12,$3F,$C1,$F1,$96,$C7,$E8,$FB,$3E,$E1
	.dc.b	$1F,$AA,$D8,$F0,$60,$8C,$5F,$A2,$31,$5C,$09,$5C,$2B,$71,$EB,$9D
	.dc.b	$CE,$63,$75,$CE,$B7,$3F,$16,$7F,$C5,$B1,$88,$F8,$27,$E6,$1C,$57
	.dc.b	$E6,$63,$F4,$11,$8B,$FE,$04,$DA,$F8,$07,$80,$F2,$39,$9E,$91,$97
	.dc.b	$C4,$7E,$DE,$35,$7F,$C0,$DB,$3E,$63,$84,$F9,$A7,$51,$8B,$F3,$1F
	.dc.b	$0A,$3E,$04,$66,$FC,$ED,$3C,$F5,$F1,$C9,$3C,$1F,$EE,$1F,$0D,$FD
	.dc.b	$DF,$E0,$FF,$BC,$36,$BF,$A5,$B6,$77,$1F,$D3,$A7,$FB,$F7,$C1,$48
	.dc.b	$FE,$A3,$F2,$F6,$BF,$BE,$4D,$C3,$FB,$D4,$DA,$FF,$00,$D7,$FB,$AF
	.dc.b	$B3,$FD,$B5,$06,$F6,$E1,$91,$FD,$36,$7F,$C1,$2E,$CC,$3C,$DA,$65
	.dc.b	$E0,$DB,$FF,$66,$2E,$CE,$B3,$FC,$8A,$30,$7F,$6C,$F9,$7F,$F7,$77
	.dc.b	$C6,$F8,$27,$FC,$31,$18,$7F,$C4,$0D,$9E,$07,$F7,$A9,$E1,$7F,$ED
	.dc.b	$B7,$FD,$AD,$5F,$E2,$26,$A7,$FD,$B0,$60,$9E,$BF,$8B,$D7,$F1,$BF
	.dc.b	$23,$A6,$61,$FE,$08,$3F,$DB,$98,$7B,$A3,$38,$7C,$27,$FD,$D8,$FD
	.dc.b	$29,$FE,$F2,$FF,$79,$6B,$98,$FA,$BD,$7C,$AB,$E4,$BF,$13,$D9,$97
	.dc.b	$EF,$F6,$AE,$1F,$D3,$33,$56,$30,$F8,$55,$6F,$F0,$74,$F6,$BE,$7F
	.dc.b	$C0,$7D,$06,$CF,$CB,$6E,$3D,$5B,$87,$DC,$EA,$FC,$59,$F8,$CE,$BB
	.dc.b	$7F,$16,$3E,$0D,$FE,$29,$AA,$11,$86,$9F,$8C,$7F,$2F,$CF,$D2,$7F
	.dc.b	$10,$FE,$1B,$E5,$65,$7E,$07,$E5,$6D,$7C,$39,$B8,$FE,$0E,$6D,$7C
	.dc.b	$2F,$95,$FC,$7F,$CA,$FE,$41,$B1,$F8,$B6,$E1,$F3,$5B,$7E,$98,$FF
	.dc.b	$36,$FD,$D7,$E4,$2B,$F5,$ED,$8F,$D4,$F3,$6E,$B8,$42,$74,$D3,$E2
	.dc.b	$BF,$A1,$7F,$C7,$CD,$7A,$9A,$3A,$57,$83,$96,$C7,$90,$A7,$77,$F5
	.dc.b	$BF,$2B,$F4,$B8,$8D,$25,$7E,$FA,$1F,$2D,$B2,$46,$5E,$5F,$2B,$23
	.dc.b	$83,$70,$FE,$15,$4F,$96,$F0,$12,$BB,$3E,$38,$91,$E1,$6C,$F2,$F9
	.dc.b	$2F,$41,$27,$CC,$FA,$FE,$5A,$47,$A5,$B3,$FC,$6F,$4F,$D2,$CF,$F1
	.dc.b	$A6,$C6,$7A,$7E,$89,$F4,$B0,$95,$B6,$CF,$F2,$5F,$A1,$79,$3F,$9C
	.dc.b	$6D,$7F,$26,$6C,$EE,$7F,$43,$F3,$EF,$90,$A4,$BF,$E5,$B1,$97,$F9
	.dc.b	$7F,$CB,$BE,$0F,$FD,$7F,$87,$F3,$58,$C7,$AF,$AF,$84,$72,$F0,$BD
	.dc.b	$DF,$FA,$1F,$57,$CC,$7F,$3A,$8F,$E4,$4D,$8C,$C7,$E0,$63,$1A,$48
	.dc.b	$FC,$23,$5F,$BB,$F2,$BC,$48,$FE,$25,$E9,$F2,$92,$3F,$18,$D7,$FA
	.dc.b	$77,$E9,$5F,$C1,$13,$49,$B7,$9D,$7C,$55,$24,$FF,$54,$8C,$B7,$E2
	.dc.b	$B8,$11,$C1,$BD,$BA,$11,$C1,$CF,$5F,$01,$1C,$1D,$FE,$2B,$C8,$47
	.dc.b	$07,$8F,$E2,$BD,$04,$70,$7F,$6A,$F9,$47,$BB,$FD,$DB,$C0,$FE,$0B
	.dc.b	$F2,$BF,$85,$AF,$F6,$AB,$5F,$F4,$4D,$70,$7F,$DB,$2B,$DE,$E5,$BF
	.dc.b	$F8,$02,$35,$3C,$30,$CA,$11,$C5,$FF,$71,$F9,$8F,$43,$70,$F6,$DE
	.dc.b	$C6,$1B,$1C,$27,$75,$FD,$4A,$30,$43,$FD,$42,$E5,$FF,$51,$FC,$A7
	.dc.b	$F8,$4D,$FF,$79,$D3,$E5,$7B,$1E,$29,$D4,$F0,$F5,$71,$EB,$48,$FF
	.dc.b	$7D,$CF,$F8,$3D,$BF,$03,$C6,$A4,$60,$DE,$BE,$BF,$87,$91,$83,$9F
	.dc.b	$07,$86,$64,$60,$EF,$F0,$5F,$ED,$27,$FC,$51,$8A,$31,$1F,$EF,$78
	.dc.b	$C7,$1F,$F8,$A9,$B1,$C3,$D9,$7C,$AE,$5B,$19,$8C,$1F,$B9,$5F,$2B
	.dc.b	$48,$FE,$F5,$FD,$F7,$87,$FD,$0A,$3F,$C4,$27,$E0,$63,$F0,$5F,$2A
	.dc.b	$E3,$FF,$23,$CE,$B9,$38,$1C,$6E,$F3,$FE,$27,$4F,$F0,$EF,$95,$D4
	.dc.b	$FE,$9F,$CC,$F9,$1C,$56,$CE,$BE,$E7,$8C,$CE,$5F,$EE,$40,$F9,$FD
	.dc.b	$C9,$3C,$67,$FC,$B7,$38,$7F,$DD,$78,$B9,$FF,$AF,$B6,$BF,$17,$F1
	.dc.b	$B8,$9C,$F2,$C9,$E9,$6B,$E0,$F4,$9F,$ED,$3F,$2B,$FE,$48,$3F,$C9
	.dc.b	$3A,$7F,$B9,$6B,$FB,$DC,$67,$7D,$9C,$37,$B8,$F6,$F9,$5F,$C2,$3F
	.dc.b	$E6,$A6,$E5,$FE,$6F,$6B,$F4,$9A,$FF,$9E,$9F,$F2,$7C,$5F,$80,$FE
	.dc.b	$FB,$18,$9F,$DE,$28,$FF,$00,$FC,$AB,$E4,$3F,$C6,$2D,$BF,$F3,$DB
	.dc.b	$6F,$FD,$00,$D7,$F0,$6B,$B3,$89,$27,$8B,$FE,$3C,$F9,$FF,$F2,$64
	.dc.b	$7F,$A1,$9B,$5F,$E8,$B6,$BF,$8E,$6B,$FE,$96,$F9,$6F,$03,$70,$FF
	.dc.b	$23,$C6,$23,$F8,$C3,$53,$E5,$AF,$F4,$AB,$63,$FD,$34,$D7,$F3,$AD
	.dc.b	$7F,$D4,$BF,$3E,$F8,$6F,$F2,$AB,$6B,$FD,$3E,$D9,$FF,$52,$35,$FD
	.dc.b	$3F,$C3,$3E,$5B,$FC,$BA,$DA,$FF,$53,$B6,$7F,$D5,$8D,$78,$FC,$AF
	.dc.b	$02,$4F,$07,$FC,$CF,$F3,$2F,$8F,$A4,$9F,$F5,$9C,$65,$FF,$5A,$FC
	.dc.b	$BF,$F9,$B5,$A7,$F9,$C1,$C5,$FF,$AF,$5A,$FF,$00,$F8,$F7,$C9,$FF
	.dc.b	$9D,$9B,$5F,$EC,$06,$CF,$FB,$15,$AF,$F0,$8F,$53,$F8,$04,$FF,$73
	.dc.b	$F0,$3F,$CF,$F1,$89,$FE,$FF,$1E,$3F,$52,$57,$A1,$B3,$E5,$F5,$3D
	.dc.b	$04,$9F,$6B,$D3,$E6,$24,$7F,$80,$5B,$3E,$DF,$53,$E0,$12,$7E,$2B
	.dc.b	$F0,$FD,$49,$1F,$21,$B3,$FE,$14,$E2,$F4,$92,$7F,$C1,$CF,$CB,$B9
	.dc.b	$1F,$41,$B3,$F3,$67,$E8,$12,$7E,$AB,$F1,$38,$64,$7F,$86,$5B,$3F
	.dc.b	$EF,$1F,$95,$F9,$FE,$07,$FA,$21,$B5,$FE,$FC,$F9,$5F,$F7,$73,$FE
	.dc.b	$FE,$8C,$D7,$F7,$D7,$EA,$B9,$E1,$FF,$58,$C6,$23,$51,$83,$FD,$0F
	.dc.b	$5F,$EF,$48,$FE,$49,$5C,$4F,$E4,$F1,$9B,$FF,$4D,$BF,$CA,$1B,$E1
	.dc.b	$F2,$F1,$39,$B6,$B8,$B6,$FA,$9F,$E1,$3B,$FF,$11,$CF,$6F,$43,$C6
	.dc.b	$79,$3D,$0B,$19,$25,$72,$15,$FC,$8D,$B5,$EC,$F3,$E5,$7A,$BC,$2B
	.dc.b	$5C,$6F,$FD,$E6,$F8,$EB,$FD,$AF,$18,$8E,$B1,$83,$E3,$FC,$AF,$2D
	.dc.b	$3E,$1B,$F2,$BE,$0B,$E5,$F9,$37,$2F,$92,$DC,$3C,$0D,$BF,$96,$FD
	.dc.b	$2E,$E9,$5F,$43,$E4,$6D,$F9,$E6,$DF,$D2,$9B,$5F,$E3,$DA,$76,$7F
	.dc.b	$8E,$23,$FA,$FB,$63,$E4,$3F,$D6,$E3,$83,$9A,$FF,$6B,$BB,$9F,$E0
	.dc.b	$DE,$BF,$F0,$52,$57,$F0,$57,$FD,$9F,$2F,$95,$B1,$87,$F1,$BF,$25
	.dc.b	$F8,$EF,$95,$B7,$F8,$C6,$E1,$F4,$DB,$8F,$E3,$9F,$CA,$FC,$AC,$AF
	.dc.b	$C9,$FC,$AD,$BF,$B5,$36,$FF,$2B,$36,$BE,$21,$FA,$8A,$FF,$31,$35
	.dc.b	$3F,$CF,$E7,$FD,$EE,$E4,$7B,$7F,$EF,$A9,$E6,$D7,$9B,$5E,$6D,$8E
	.dc.b	$5E,$DF,$FB,$D1,$A7,$EB,$3E,$47,$FD,$EF,$F0,$DE,$4B,$FC,$A4,$61
	.dc.b	$FD,$AB,$67,$81,$FE,$1A,$9E,$18,$7A,$AB,$FC,$FD,$1F,$EC,$49,$D7
	.dc.b	$A5,$DB,$D3,$EE,$C8,$C9,$CE,$31,$5F,$AD,$87,$3F,$A4,$F5,$75,$D8
	.dc.b	$FF,$15,$9A,$2C,$69,$A4,$9E,$91,$96,$C6,$71,$FA,$33,$96,$C6,$4F
	.dc.b	$B7,$19,$69,$99,$FB,$B1,$96,$C6,$5F,$BF,$19,$6C,$64,$F5,$3F,$06
	.dc.b	$BD,$CF,$98,$95,$F8,$68,$CB,$E6,$8C,$CF,$E4,$27,$2F,$9A,$32,$7E
	.dc.b	$4A,$32,$F9,$8C,$9F,$94,$72,$F9,$8C,$C7,$F1,$F3,$CF,$3E,$12,$3E
	.dc.b	$13,$D9,$5C,$9E,$0D,$3E,$13,$C8,$F3,$47,$29,$F3,$47,$23,$E3,$57
	.dc.b	$16,$99,$8F,$2C,$65,$B1,$99,$E9,$1E,$D8,$C1,$E0,$3D,$0F,$BA,$BD
	.dc.b	$15,$CE,$3F,$D2,$F3,$FC,$D4,$93,$FA,$F2,$3F,$E9,$C8,$F9,$87,$08
	.dc.b	$CB,$F9,$DE,$AF,$88,$49,$FC,$13,$63,$E7,$B6,$FD,$33,$FE,$6F,$36
	.dc.b	$49,$DB,$63,$73,$C6,$3D,$D3,$FC,$4A,$30,$FE,$06,$7F,$88,$C6,$23
	.dc.b	$F0,$11,$89,$E8,$72,$8F,$E3,$33,$B8,$FB,$51,$D5,$DB,$D5,$FE,$47
	.dc.b	$1B,$AF,$E2,$8D,$8F,$DD,$B6,$7E,$34,$72,$7C,$6F,$23,$07,$81,$F1
	.dc.b	$B8,$9E,$AD,$3A,$96,$3A,$9F,$20,$95,$F2,$23,$4F,$C1,$8C,$C7,$C1
	.dc.b	$3E,$93,$FE,$03,$C9,$EC,$7F,$2C,$3F,$96,$FA,$F6,$BF,$94,$45,$AE
	.dc.b	$13,$6F,$F9,$54,$2F,$FD,$37,$18,$7F,$DD,$17,$FC,$C7,$D8,$FE,$73
	.dc.b	$EC,$FF,$3A,$6C,$64,$FE,$6D,$1F,$D0,$7A,$64,$7F,$28,$EA,$B8,$FF
	.dc.b	$2F,$9F,$E9,$04,$9F,$CD,$47,$F5,$68,$C3,$FC,$D7,$9F,$8E,$5F,$D6
	.dc.b	$1F,$E9,$7F,$2B,$2F,$FA,$37,$CA,$DB,$F8,$D3,$6F,$D9,$36,$BF,$C1
	.dc.b	$5A,$3E,$53,$FA,$AF,$CA,$FF,$56,$6C,$FF,$51,$6E,$3F,$45,$B8,$7C
	.dc.b	$AF,$8F,$FE,$AB,$5F,$E1,$56,$D7,$F7,$2E,$7F,$EB,$15,$B8,$B5,$B7
	.dc.b	$FB,$65,$7F,$65,$7F,$C3,$C1,$C3,$B8,$F8,$F1,$DA,$35,$CF,$D1,$1C
	.dc.b	$4C,$1A,$9C,$D7,$E0,$BD,$17,$C1,$E9,$B3,$A3,$FC,$60,$BD,$FD,$4C
	.dc.b	$9D,$5F,$82,$B8,$BF,$E5,$97,$84,$6A,$11,$AB,$F1,$7B,$5E,$1F,$73
	.dc.b	$C4,$DB,$E1,$5F,$EF,$BA,$F4,$56,$A7,$2F,$A7,$E6,$24,$70,$F3,$6E
	.dc.b	$3E,$39,$FF,$5C,$92,$7F,$23,$82,$31,$3E,$9F,$67,$8E,$3F,$1D,$F8
	.dc.b	$DE,$34,$C3,$CB,$E3,$5B,$F8,$13,$6F,$E7,$4D,$AF,$97,$F3,$1F,$53
	.dc.b	$F4,$B7,$C7,$FD,$18,$C3,$67,$E4,$4E,$9F,$87,$3A,$7E,$67,$A5,$FE
	.dc.b	$5F,$CB,$FD,$4D,$9F,$5F,$99,$FC,$0C,$E6,$4F,$FB,$B5,$FF,$5A,$4F
	.dc.b	$FA,$A8,$95,$FB,$49,$E1,$87,$F9,$26,$3F,$D6,$93,$E9,$F7,$FF,$C7
	.dc.b	$BF,$31,$FE,$45,$EB,$34,$95,$FB,$E8,$7C,$C6,$C9,$19,$7F,$C9,$D8
	.dc.b	$7F,$95,$23,$F8,$2B,$72,$FE,$C0,$D7,$F4,$ED,$7F,$7B,$5E,$32,$4F
	.dc.b	$85,$F3,$7C,$BC,$8F,$23,$67,$AF,$97,$E2,$24,$F8,$9F,$DB,$F8,$F2
	.dc.b	$3F,$66,$D9,$FE,$E9,$D9,$FC,$B6,$7F,$6F,$87,$EF,$88,$FF,$20,$F8
	.dc.b	$97,$BB,$FE,$6D,$6C,$66,$30,$7F,$9C,$7E,$55,$F2,$74,$93,$FC,$AA
	.dc.b	$32,$FF,$2C,$F9,$57,$C8,$FE,$3B,$3F,$F0,$44,$63,$9F,$5F,$04,$F0
	.dc.b	$F8,$5E,$EF,$F8,$9F,$CE,$3E,$03,$F9,$BC,$23,$0F,$F3,$A6,$BF,$E1
	.dc.b	$56,$BF,$C7,$37,$F4,$BF,$83,$D7,$9C,$62,$3F,$C3,$8D,$78,$7C,$AE
	.dc.b	$88,$FE,$2D,$F1,$F2,$BE,$06,$92,$7F,$12,$FA,$3C,$2F,$F4,$D4,$7F
	.dc.b	$12,$E0,$C1,$1F,$E0,$7F,$29,$B2,$38,$3F,$27,$F0,$5E,$02,$38,$3F
	.dc.b	$85,$F0,$77,$23,$83,$F8,$7E,$FE,$32,$38,$3F,$3F,$F0,$5E,$C2,$38
	.dc.b	$3F,$8B,$FC,$2F,$AC,$8E,$0F,$D3,$FB,$3F,$10,$8E,$0F,$ED,$FF,$26
	.dc.b	$C7,$FD,$D9,$E3,$7F,$81,$3E,$53,$E9,$37,$8F,$3B,$7F,$BC,$6F,$86
	.dc.b	$1F,$CE,$D6,$7A,$8F,$EA,$31,$A5,$7F,$E1,$08,$C3,$67,$84,$BD,$77
	.dc.b	$B9,$8F,$82,$79,$74,$DC,$EA,$BF,$A6,$3F,$E2,$46,$D6,$8D,$F8,$3D
	.dc.b	$BE,$53,$FD,$BF,$3F,$D4,$F0,$FE,$C2,$DF,$47,$5E,$FE,$FC,$5B,$3F
	.dc.b	$D6,$6F,$8B,$EC,$DF,$FB,$7A,$84,$F7,$E9,$E5,$3C,$63,$0F,$3E,$3F
	.dc.b	$77,$A3,$23,$07,$7E,$0F,$44,$C8,$C1,$E3,$E5,$F9,$AF,$FA,$0F,$14
	.dc.b	$62,$3F,$C1,$71,$8E,$2F,$F1,$7B,$63,$87,$BC,$F9,$4C,$B6,$33,$18
	.dc.b	$3F,$80,$AF,$1E,$91,$FE,$0A,$FF,$06,$F0,$FF,$C2,$31,$FE,$F7,$3E
	.dc.b	$41,$FE,$1E,$8F,$F0,$BF,$CA,$7E,$06,$3E,$5C,$F0,$E5,$9C,$3F,$E0
	.dc.b	$99,$E5,$9F,$F9,$69,$FC,$3C,$67,$97,$B9,$FE,$20,$E5,$7C,$66,$27
	.dc.b	$97,$C0,$F9,$0C,$5F,$90,$FF,$14,$BE,$8F,$81,$93,$C0,$F1,$3F,$E2
	.dc.b	$CC,$7F,$88,$47,$F8,$75,$B5,$CF,$0F,$F1,$0B,$D6,$32,$7F,$9C,$8F
	.dc.b	$D9,$4F,$1E,$2E,$9C,$DD,$5F,$F1,$CF,$CA,$7C,$09,$FC,$7B,$6F,$FC
	.dc.b	$7F,$87,$F7,$D8,$C4,$1F,$8A,$FB,$5B,$87,$FA,$09,$D9,$8A,$FC,$DF
	.dc.b	$CA,$7E,$29,$FF,$1C,$3F,$E9,$29,$FF,$44,$B9,$AF,$90,$E6,$78,$FC
	.dc.b	$27,$D5,$9F,$F2,$9C,$FF,$A6,$9F,$F2,$EB,$5D,$CF,$FA,$0B,$E5,$1D
	.dc.b	$3C,$65,$18,$9D,$4E,$63,$13,$FE,$E3,$F9,$46,$DC,$5B,$7F,$E8,$E6
	.dc.b	$CE,$E7,$5F,$2D,$2B,$4D,$9F,$F7,$77,$CA,$F6,$7F,$CE,$AD,$7F,$80
	.dc.b	$B5,$FF,$38,$B5,$7C,$AB,$D8,$F6,$36,$FF,$D4,$2D,$9F,$F7,$E3,$5F
	.dc.b	$F4,$1F,$D0,$7A,$1B,$8F,$C5,$8C,$47,$F1,$26,$BF,$E8,$2F,$8A,$FE
	.dc.b	$F1,$7F,$EA,$E6,$C7,$FA,$38,$38,$7F,$D2,$2D,$7F,$8E,$FC,$B7,$F9
	.dc.b	$1D,$FF,$4A,$B5,$CF,$B9,$F5,$E3,$11,$FE,$92,$DD,$F1,$9F,$6D,$B8
	.dc.b	$7F,$B0,$1B,$3E,$36,$BF,$EA,$2F,$8F,$7C,$27,$E0,$DB,$5F,$EC,$86
	.dc.b	$CF,$AD,$AF,$FA,$AB,$D2,$FE,$03,$3F,$E7,$EE,$CF,$C6,$46,$27,$E1
	.dc.b	$35,$FF,$58,$FC,$07,$FA,$0E,$BF,$CC,$5F,$09,$F9,$38,$C4,$FC,$96
	.dc.b	$BF,$EB,$8F,$96,$7A,$9F,$99,$6D,$FF,$B7,$9B,$3F,$45,$AF,$9F,$E5
	.dc.b	$7C,$A4,$9F,$F4,$AB,$FE,$64,$F2,$24,$7F,$99,$1B,$3C,$FD,$4F,$01
	.dc.b	$27,$C0,$FC,$0F,$95,$91,$FE,$A0,$6C,$FF,$9B,$3B,$BF,$CE,$24,$9F
	.dc.b	$F3,$8B,$EB,$99,$1E,$B6,$CF,$CA,$F7,$3E,$69,$27,$FD,$5A,$FF,$9F
	.dc.b	$3E,$52,$47,$F9,$F1,$B3,$F6,$7D,$4F,$F6,$BF,$17,$F8,$02,$BF,$D9
	.dc.b	$0F,$FA,$64,$C4,$FF,$37,$6B,$F8,$79,$FC,$31,$27,$F0,$CF,$4F,$88
	.dc.b	$EA,$D8,$FF,$6A,$7C,$AB,$E4,$7F,$DC,$EB,$E5,$2C,$61,$B3,$C1,$B8
	.dc.b	$7F,$49,$AE,$13,$D3,$E5,$39,$9D,$3A,$AC,$64,$95,$C6,$32,$FF,$53
	.dc.b	$8F,$F5,$07,$8E,$7F,$C1,$E6,$0B,$E1,$7F,$BF,$9C,$48,$E1,$EB,$FF
	.dc.b	$2B,$8F,$F5,$34,$65,$AE,$9F,$E0,$8D,$3C,$5E,$BF,$59,$EC,$FC,$6F
	.dc.b	$8F,$F2,$4F,$F6,$9F,$82,$E8,$FC,$38,$CE,$3A,$84,$EA,$31,$A2,$B7
	.dc.b	$8F,$F8,$B2,$B4,$FF,$8B,$A7,$4F,$B6,$BA,$4E,$AF,$E0,$3F,$D8,$89
	.dc.b	$E9,$74,$8C,$FA,$75,$F1,$B6,$77,$7D,$4F,$14,$7C,$DE,$99,$19,$AF
	.dc.b	$95,$EC,$61,$FB,$13,$F3,$C9,$3F,$30,$38,$8F,$BD,$D3,$C7,$2F,$8D
	.dc.b	$5F,$DA,$E7,$D7,$EA,$79,$4F,$37,$A9,$67,$C5,$3F,$32,$70,$FD,$39
	.dc.b	$B5,$F1,$BC,$AF,$77,$B7,$EF,$6E,$1C,$B5,$FE,$CB,$5F,$32,$B5,$39
	.dc.b	$7F,$1D,$F2,$92,$3C,$27,$F4,$A2,$B9,$C7,$FA,$74,$23,$F2,$44,$7E
	.dc.b	$23,$F9,$9E,$8E,$17,$F0,$9F,$A9,$F2,$92,$BC,$DF,$29,$70,$FC,$84
	.dc.b	$DB,$FC,$1C,$DA,$FC,$57,$C9,$FD,$DF,$94,$FB,$CD,$8F,$80,$DC,$3F
	.dc.b	$34,$DB,$FC,$C7,$C6,$FE,$36,$4D,$AF,$C9,$C7,$F9,$26,$34,$47,$F6
	.dc.b	$9C,$FF,$8F,$8C,$4F,$D8,$79,$46,$66,$4E,$5E,$2D,$23,$DC,$C9,$5C
	.dc.b	$D7,$A2,$32,$D8,$C9,$FE,$98,$AF,$0B,$89,$FE,$11,$1E,$67,$F8,$95
	.dc.b	$79,$1E,$C4,$AF,$CB,$C6,$5E,$D1,$9A,$FE,$2F,$39,$7C,$11,$4F,$E2
	.dc.b	$51,$5F,$9A,$58,$F9,$B3,$D0,$8F,$23,$F2,$33,$C6,$BD,$13,$E4,$9F
	.dc.b	$65,$74,$69,$D0,$E4,$F9,$CF,$82,$7C,$87,$31,$E0,$8C,$C7,$80,$EB
	.dc.b	$3E,$07,$DE,$62,$3D,$55,$E5,$6C,$64,$F1,$C6,$5A,$66,$3C,$87,$F4
	.dc.b	$18,$D3,$88,$F5,$3E,$43,$EA,$9F,$ED,$43,$E0,$1F,$E3,$F9,$FE,$6A
	.dc.b	$E5,$AF,$C2,$30,$7F,$9C,$5C,$BD,$4C,$C7,$E6,$79,$7E,$81,$1F,$86
	.dc.b	$D8,$F4,$CF,$FA,$50,$FD,$2C,$66,$FE,$64,$FE,$B2,$31,$5F,$40,$FC
	.dc.b	$B3,$88,$F7,$4F,$ED,$23,$11,$EE,$8C,$4F,$88,$91,$E2,$9D,$D7,$8A
	.dc.b	$73,$5E,$28,$F3,$57,$ED,$EB,$6D,$AD,$9E,$57,$F4,$F1,$A8,$F8,$86
	.dc.b	$A3,$CE,$FA,$8C,$3E,$77,$0F,$95,$CC,$F9,$63,$04,$AF,$35,$77,$8F
	.dc.b	$E2,$0D,$8F,$83,$1B,$AF,$33,$F3,$5E,$AF,$AC,$D1,$FE,$07,$F2,$9F
	.dc.b	$03,$FD,$95,$B3,$FD,$39,$FF,$6D,$D7,$F7,$B7,$35,$FE,$E9,$6C,$66
	.dc.b	$B9,$7C,$A7,$37,$A7,$CA,$75,$F4,$2E,$3D,$5B,$8F,$36,$E1,$DD,$F2
	.dc.b	$79,$D2,$BC,$9E,$85,$BF,$14,$FF,$9F,$89,$3F,$E1,$CC,$FF,$CE,$F3
	.dc.b	$FE,$25,$0F,$F9,$F3,$0F,$67,$A5,$CD,$B5,$F0,$7C,$6F,$88,$7C,$5F
	.dc.b	$1E,$C6,$49,$87,$C3,$8C,$BF,$17,$E5,$3F,$C2,$1F,$14,$F8,$DF,$9D
	.dc.b	$18,$6D,$FF,$7C,$9D,$3F,$46,$74,$FF,$80,$3D,$36,$7F,$E1,$B6,$C7
	.dc.b	$1F,$94,$C4,$C8,$FF,$1E,$CF,$FA,$DB,$13,$49,$3E,$6F,$27,$84,$B5
	.dc.b	$FF,$1D,$65,$FE,$D7,$6F,$CA,$FB,$0F,$6F,$C0,$DC,$BD,$51,$6F,$94
	.dc.b	$7D,$1B,$56,$6F,$AF,$B3,$F1,$BC,$0F,$8E,$DB,$F8,$8D,$C3,$99,$FE
	.dc.b	$5F,$C3,$FC,$24,$E0,$3F,$E1,$19,$F9,$9E,$07,$D2,$3E,$9F,$9F,$63
	.dc.b	$27,$F9,$8A,$7F,$D0,$8D,$9F,$F2,$D5,$7A,$5A,$F8,$5F,$BB,$E7,$F3
	.dc.b	$8F,$15,$23,$F0,$3E,$7C,$BF,$BD,$DF,$6F,$73,$74,$F9,$F3,$6B,$F3
	.dc.b	$1F,$A6,$7F,$84,$C7,$7F,$F9,$6D,$C3,$70,$C3,$67,$FC,$8F,$3C,$1B
	.dc.b	$E6,$F1,$39,$79,$56,$32,$7F,$A1,$25,$3F,$02,$D1,$89,$D4,$E6,$31
	.dc.b	$3F,$4D,$79,$5D,$63,$0F,$F9,$C7,$E5,$7A,$79,$46,$93,$0F,$A9,$E8
	.dc.b	$F8,$78,$3F,$D3,$24,$7C,$1E,$64,$9F,$43,$71,$FE,$C8,$7F,$A4,$BE
	.dc.b	$57,$B9,$27,$FD,$43,$8A,$3B,$7A,$52,$7F,$08,$DA,$E7,$A3,$CE,$F8
	.dc.b	$04,$AF,$90,$FF,$6B,$C2,$47,$FA,$BB,$CA,$FF,$51,$B6,$3E,$67,$C4
	.dc.b	$3E,$56,$92,$7F,$D6,$B1,$97,$EA,$7A,$1F,$3C,$8E,$0F,$F5,$DF,$CA
	.dc.b	$3E,$47,$F8,$86,$9F,$0A,$31,$E3,$6B,$DE,$FC,$5C,$E6,$B6,$FF,$82
	.dc.b	$FE,$51,$ED,$FF,$B1,$A1,$18,$7F,$D9,$2D,$7E,$7B,$5F,$F0,$A5,$7F
	.dc.b	$BB,$5F,$F6,$7B,$5F,$AA,$D7,$7E,$07,$7D,$9F,$29,$49,$3F,$61,$F2
	.dc.b	$73,$BE,$32,$92,$7F,$78,$FA,$79,$1E,$BD,$24,$FE,$7D,$F7,$79,$EF
	.dc.b	$94,$A4,$9F,$E9,$8F,$83,$E3,$64,$7F,$09,$6B,$F8,$6F,$63,$C0,$47
	.dc.b	$F5,$8F,$6E,$19,$1F,$D5,$DA,$FE,$63,$37,$C8,$FB,$89,$1F,$C5,$DF
	.dc.b	$67,$B5,$23,$F7,$0D,$79,$7B,$BF,$18,$8F,$F1,$F7,$E2,$7C,$E4,$8F
	.dc.b	$E0,$0D,$7F,$53,$F2,$6C,$FF,$B5,$79,$3F,$EF,$DF,$95,$7C,$AE,$3B
	.dc.b	$3F,$B7,$46,$30,$FF,$6B,$E1,$C5,$BA,$6D,$08,$D3,$FC,$ED,$B8,$2F
	.dc.b	$FB,$D4,$7C,$49,$FF,$71,$F8,$3F,$C5,$9F,$F7,$2F,$A2,$F6,$7B,$47
	.dc.b	$F5,$06,$CE,$E3,$FA,$42,$F2,$F3,$E5,$C9,$EB,$1C,$BE,$55,$E5,$74
	.dc.b	$C7,$FC,$53,$5E,$0F,$7F,$53,$BF,$26,$CF,$FB,$E2,$F8,$B6,$3A,$76
	.dc.b	$EE,$30,$6D,$F0,$71,$F7,$F7,$A4,$F3,$8F,$27,$16,$A3,$07,$77,$D1
	.dc.b	$CD,$FD,$EE,$3F,$91,$E2,$8C,$47,$FA,$26,$31,$C7,$FC,$A2,$11,$87
	.dc.b	$8F,$C1,$3E,$57,$52,$8C,$4E,$A7,$31,$88,$FE,$36,$7F,$0B,$F2,$BF
	.dc.b	$86,$26,$9F,$C3,$FD,$1F,$C0,$47,$F7,$D3,$FA,$D1,$4E,$03,$E5,$7F
	.dc.b	$9A,$3F,$E3,$F8,$E3,$8F,$F7,$AE,$4B,$D1,$FE,$A7,$6D,$6E,$3F,$27
	.dc.b	$3B,$69,$FE,$AD,$E5,$DC,$FE,$61,$BC,$AF,$8C,$CC,$FE,$2E,$B8,$BF
	.dc.b	$E5,$1F,$4F,$26,$AE,$CE,$A7,$F4,$7E,$97,$23,$2F,$C1,$7C,$38,$FF
	.dc.b	$1F,$9E,$0D,$9F,$F1,$16,$3F,$E7,$89,$FD,$BF,$2F,$0E,$49,$1C,$22
	.dc.b	$D7,$F8,$A7,$E5,$7E,$6E,$1F,$42,$11,$8A,$FF,$35,$61,$FE,$50,$9E
	.dc.b	$1B,$7C,$4A,$FF,$00,$DF,$16,$DF,$F9,$C5,$E5,$5F,$C0,$1C,$BF,$C0
	.dc.b	$BE,$55,$F1,$5F,$39,$B5,$F1,$57,$BB,$99,$93,$F9,$79,$DD,$7F,$80
	.dc.b	$1A,$FF,$B9,$5A,$FF,$A0,$FE,$57,$FD,$0A,$FF,$12,$6A,$69,$3E,$1B
	.dc.b	$FC,$44,$9A,$4A,$FD,$44,$7F,$9B,$C9,$F0,$3C,$0F,$F1,$57,$C9,$F6
	.dc.b	$27,$C3,$E0,$CD,$CB,$FD,$94,$F7,$6C,$72,$A9,$1A,$6E,$1F,$E2,$0F
	.dc.b	$81,$FF,$5B,$BF,$EA,$10,$E1,$FF,$52,$35,$FE,$66,$D7,$7F,$13,$FE
	.dc.b	$41,$83,$49,$5F,$C1,$E1,$F1,$3B,$24,$65,$FF,$25,$7C,$87,$F6,$D8
	.dc.b	$FE,$1C,$DB,$FF,$1B,$B5,$F4,$35,$FF,$5D,$2F,$98,$E6,$49,$E6,$F6
	.dc.b	$F6,$24,$76,$6D,$7F,$91,$3E,$23,$E6,$37,$4F,$F2,$F4,$62,$3E,$43
	.dc.b	$5F,$F6,$57,$C7,$7F,$4D,$BF,$E4,$CD,$8F,$F2,$7B,$5F,$A0,$D7,$FD
	.dc.b	$A5,$1F,$C9,$CF,$E5,$C5,$7F,$CA,$ED,$7E,$C3,$5F,$F4,$1F,$CC,$7F
	.dc.b	$9B,$C9,$1F,$E7,$77,$FC,$E5,$F0,$92,$3F,$CF,$4D,$9F,$F2,$CF,$CE
	.dc.b	$7F,$9D,$C9,$3F,$E7,$F7,$FC,$F5,$EE,$48,$FF,$41,$36,$7F,$CC,$3F
	.dc.b	$0B,$FE,$7F,$24,$FF,$A1,$DF,$F4,$17,$CE,$48,$FF,$45,$36,$7F,$CD
	.dc.b	$3F,$31,$FE,$99,$3F,$D3,$5E,$AD,$AF,$F4,$BC,$7F,$A2,$C8,$FF,$A4
	.dc.b	$DB,$3F,$E8,$DF,$7F,$83,$67,$FD,$3F,$F3,$7F,$EA,$53,$F9,$BF,$C6
	.dc.b	$FF,$82,$63,$24,$9F,$F4,$B4,$EA,$38,$1F,$CC,$E7,$F2,$9E,$07,$F2
	.dc.b	$37,$F2,$11,$9F,$7B,$AF,$2F,$66,$C6,$4E,$51,$E2,$E5,$91,$D7,$DD
	.dc.b	$B9,$7F,$B6,$70,$57,$FA,$BD,$B5,$FD,$CD,$B3,$FE,$9C,$6B,$FB,$D7
	.dc.b	$38,$F9,$FE,$23,$AB,$6B,$D9,$F2,$BF,$10,$F8,$BE,$65,$8C,$93,$0F
	.dc.b	$24,$65,$F3,$7C,$B7,$CC,$F9,$6F,$9A,$DA,$F9,$0D,$9D,$C7,$CB,$9F
	.dc.b	$ED,$3C,$AE,$4A,$7D,$6F,$88,$B5,$FD,$8E,$74,$FF,$AD,$A7,$FD,$8E
	.dc.b	$47,$07,$D7,$F6,$F8,$E1,$F2,$DF,$C0,$F8,$32,$7E,$E7,$81,$6B,$E8
	.dc.b	$4D,$DB,$EF,$4D,$AF,$F2,$AD,$38,$1E,$CF,$E3,$3C,$5F,$C6,$B6,$BF
	.dc.b	$10,$DA,$FB,$8D,$C3,$F1,$5E,$0F,$E5,$0F,$CA,$F8,$96,$FF,$21,$1F
	.dc.b	$81,$23,$F8,$16,$DF,$E1,$FD,$DE,$0D,$9F,$CC,$7C,$A7,$E8,$0F,$D0
	.dc.b	$FC,$A5,$8C,$92,$7F,$21,$19,$58,$FF,$80,$7F,$4D,$7C,$67,$F8,$81
	.dc.b	$7C,$AD,$8C,$37,$0F,$F0,$D4,$E9,$B3,$A7,$FC,$3D,$F2,$1F,$E2,$5C
	.dc.b	$3B,$C7,$CE,$8C,$61,$FE,$11,$7F,$C0,$B2,$73,$3B,$F5,$FF,$C6,$71
	.dc.b	$89,$FF,$45,$BC,$0C,$3C,$63,$67,$0F,$66,$D7,$92,$2C,$F9,$0E,$73
	.dc.b	$8B,$F2,$4D,$AF,$44,$DA,$F0,$CD,$7D,$11,$6F,$C6,$47,$13,$96,$D7
	.dc.b	$8C,$F8,$EE,$2B,$C8,$7A,$E3,$26,$DC,$D7,$6A,$F7,$4E,$6F,$D2,$D3
	.dc.b	$D2,$5A,$F4,$B5,$F4,$B6,$BA,$3E,$08,$EF,$5E,$5A,$FB,$17,$6B,$EB
	.dc.b	$1B,$8F,$0D,$70,$8E,$93,$B8,$F0,$9E,$A7,$0F,$BC,$C4,$78,$EB,$E4
	.dc.b	$34,$E6,$58,$EC,$70,$2D,$78,$89,$1F,$0A,$32,$FA,$1A,$7E,$09,$E8
	.dc.b	$FF,$8A,$EF,$F3,$4D,$AF,$F3,$43,$C4,$C4,$FF,$AA,$E7,$71,$92,$1F
	.dc.b	$02,$BF,$41,$EC,$FE,$1C,$95,$F4,$5B,$3E,$43,$F4,$8E,$27,$D7,$1F
	.dc.b	$A2,$8C,$4F,$AE,$7A,$BC,$C9,$3C,$E9,$5C,$23,$C9,$3F,$B5,$8C,$47
	.dc.b	$D3,$AE,$6F,$18,$CD,$7E,$79,$B7,$FB,$66,$DF,$F0,$26,$D7,$E6,$1B
	.dc.b	$7D,$0F,$5C,$7B,$A7,$CF,$18,$BF,$3C,$60,$F7,$1E,$73,$13,$EA,$3E
	.dc.b	$23,$87,$D4,$ED,$F5,$39,$8F,$53,$6F,$D4,$D3,$D4,$59,$F5,$36,$BD
	.dc.b	$4F,$F4,$AC,$9F,$2B,$FC,$AC,$FE,$59,$F0,$76,$7F,$94,$C5,$BF,$DF
	.dc.b	$D2,$B1,$7F,$CD,$7E,$0F,$F9,$A7,$B5,$FC,$D5,$B8,$FF,$32,$6E,$1C
	.dc.b	$9B,$5F,$CB,$7E,$2B,$F9,$49,$FC,$AB,$E3,$2D,$FF,$36,$8F,$CD,$DF
	.dc.b	$F2,$C6,$BF,$CC,$DB,$A9,$A7,$F3,$4F,$8A,$E7,$3F,$C5,$1F,$E9,$5E
	.dc.b	$F4,$BF,$E8,$FF,$0D,$6B,$DF,$37,$2F,$EA,$53,$6B,$E5,$7C,$37,$F5
	.dc.b	$3F,$84,$FE,$A8,$D8,$FE,$A2,$DC,$3E,$4B,$6F,$FA,$67,$87,$FD,$20
	.dc.b	$FE,$93,$E9,$DB,$FE,$AB,$1E,$C2,$3F,$B9,$6D,$FF,$2E,$F4,$F8,$36
	.dc.b	$7F,$B0,$FC,$B7,$F6,$7F,$7B,$FB,$43,$63,$27,$F6,$68,$CB,$4F,$ED
	.dc.b	$3E,$FF,$F6,$F3,$FB,$87,$C1,$58,$C9,$2F,$F0,$B1,$97,$FA,$0F,$A9
	.dc.b	$FD,$90,$FE,$CB,$EA,$DA,$FE,$E3,$1F,$60,$8F,$F1,$56,$CF,$F4,$EF
	.dc.b	$85,$FF,$34,$3D,$24,$E5,$FF,$02,$7B,$DF,$E0,$5F,$26,$C7,$F8,$09
	.dc.b	$B3,$F9,$A6,$F0,$7F,$83,$5F,$F0,$67,$CB,$4A,$FF,$04,$FC,$85,$BF
	.dc.b	$D0,$CD,$BF,$F0,$A4,$DA,$FF,$1A,$E2,$F5,$51,$E6,$AB,$87,$A2,$B5
	.dc.b	$39,$38,$FA,$FA,$AD,$D2,$B6,$FC,$3B,$FD,$0C,$62,$3F,$DD,$8D,$7C
	.dc.b	$71,$A3,$A5,$F9,$BD,$6B,$3E,$39,$AF,$78,$FF,$17,$7C,$B7,$EB,$BC
	.dc.b	$23,$49,$7F,$EF,$BA,$FA,$72,$F9,$6E,$04,$9D,$4A,$75,$18,$5F,$2D
	.dc.b	$DC,$91,$97,$F7,$FF,$A7,$7E,$02,$39,$B6,$BC,$4D,$7F,$CC,$E1,$C3
	.dc.b	$FE,$6A,$5F,$13,$D0,$93,$D1,$EF,$E5,$48,$EE,$DA,$F8,$1F,$3C,$E8
	.dc.b	$FE,$26,$DA,$F7,$B6,$77,$3F,$C6,$3E,$79,$F0,$54,$97,$F1,$A3,$2F
	.dc.b	$C8,$F9,$87,$27,$F9,$8B,$0F,$A1,$18,$F3,$F5,$F0,$1E,$C9,$CD,$6D
	.dc.b	$F8,$7B,$3C,$FF,$A7,$1F,$15,$B1,$98,$F9,$B1,$AF,$35,$F0,$34,$93
	.dc.b	$F7,$23,$2D,$F6,$38,$11,$C1,$F6,$3D,$2E,$E4,$70,$7D,$CF,$8D,$F1
	.dc.b	$91,$C1,$F8,$0F,$91,$F3,$91,$C1,$F8,$4F,$99,$F5,$91,$C1,$F8,$8F
	.dc.b	$4B,$DE,$47,$07,$E3,$3D,$1F,$86,$47,$07,$E4,$3D,$1F,$8E,$47,$07
	.dc.b	$E5,$3E,$57,$E5,$91,$C1,$F9,$8F,$89,$F9,$E4,$70,$7E,$73,$E2,$7E
	.dc.b	$99,$1C,$1F,$A0,$F4,$3E,$B9,$1C,$1F,$AE,$F9,$47,$AB,$FE,$C8,$F1
	.dc.b	$FF,$85,$FC,$AF,$E4,$2B,$F8,$81,$FC,$A9,$C3,$FE,$7C,$AE,$2E,$82
	.dc.b	$F5,$BA,$76,$7F,$B5,$23,$73,$C7,$0C,$BF,$CD,$A3,$83,$CF,$85,$E0
	.dc.b	$EA,$1E,$83,$DE,$C3,$63,$84,$EE,$BF,$2D,$1F,$E9,$27,$9B,$A3,$B7
	.dc.b	$57,$5F,$95,$FF,$0C,$47,$AE,$DF,$0B,$D8,$D7,$B9,$BF,$72,$D7,$C1
	.dc.b	$BE,$2F,$FA,$81,$47,$5E,$DE,$13,$B8,$C3,$C7,$6F,$5F,$6C,$8C,$1D
	.dc.b	$78,$7C,$D3,$23,$07,$87,$DA,$FF,$71,$BF,$CB,$F1,$46,$23,$F8,$FC
	.dc.b	$63,$93,$F9,$7B,$63,$87,$B6,$F9,$5C,$B6,$33,$18,$3F,$82,$BF,$D4
	.dc.b	$4F,$E0,$FE,$96,$08,$FF,$3C,$F3,$3F,$A4,$BF,$CB,$39,$FF,$10,$FF
	.dc.b	$5F,$1F,$2A,$E5,$FE,$6B,$5C,$39,$A5,$F0,$8B,$7F,$EC,$E3,$FD,$D2
	.dc.b	$7F,$BA,$B9,$5F,$19,$E0,$E7,$F0,$F8,$93,$0F,$0C,$5B,$E5,$E2,$79
	.dc.b	$8C,$70,$FF,$A9,$9F,$EC,$2B,$9E,$9F,$D9,$32,$7C,$46,$31,$FD,$25
	.dc.b	$7E,$B9,$BC,$B9,$C7,$FD,$CB,$39,$7F,$DC,$FF,$2B,$F1,$5B,$B7,$5F
	.dc.b	$87,$FE,$E3,$39,$AF,$F0,$0B,$FA,$6C,$3F,$64,$DB,$E3,$18,$E0,$FE
	.dc.b	$F3,$F2,$AF,$7F,$FC,$CA,$D9,$FE,$E2,$DA,$FE,$F0,$D7,$FB,$3B,$9C
	.dc.b	$3F,$C2,$2F,$ED,$27,$F8,$13,$97,$FC,$63,$28,$D2,$9F,$F0,$CF,$CA
	.dc.b	$BE,$27,$FC,$EE,$DB,$FF,$01,$B6,$BF,$C1,$4D,$7F,$BC,$AC,$F8,$92
	.dc.b	$78,$BF,$E8,$2F,$9E,$FC,$64,$7F,$83,$9B,$5F,$E1,$56,$BF,$E0,$46
	.dc.b	$BE,$AF,$96,$F0,$37,$0F,$F4,$6C,$62,$3F,$C9,$ED,$4F,$96,$BF,$C3
	.dc.b	$AD,$8F,$F1,$23,$5F,$F0,$9B,$5F,$93,$F3,$DF,$EA,$B9,$FF,$12,$B6
	.dc.b	$3F,$C5,$8D,$7F,$C3,$6D,$7E,$8F,$C1,$7E,$52,$7F,$C5,$AD,$8F,$F1
	.dc.b	$A3,$5F,$F1,$1B,$5E,$3F,$1D,$FE,$AA,$83,$49,$3F,$E3,$8F,$9D,$FE
	.dc.b	$68,$DF,$1F,$FD,$63,$18,$BF,$F2,$13,$5F,$F1,$8F,$CC,$7F,$82,$2B
	.dc.b	$FC,$82,$D8,$FF,$24,$35,$DF,$A7,$FE,$BC,$87,$AD,$49,$5C,$9F,$F6
	.dc.b	$0F,$CD,$3E,$56,$92,$7F,$C9,$F1,$97,$FC,$A3,$F3,$1E,$5C,$7F,$32
	.dc.b	$D7,$FD,$16,$D4,$D8,$78,$DF,$ED,$06,$9F,$E5,$C2,$CF,$F9,$8D,$AF
	.dc.b	$F9,$4B,$E3,$BF,$CA,$2F,$FA,$61,$A9,$A6,$3E,$6D,$E9,$FF,$B7,$9B
	.dc.b	$5F,$E6,$A6,$D7,$F9,$C5,$AF,$F9,$81,$7A,$7C,$49,$3C,$5F,$F3,$5F
	.dc.b	$C7,$C8,$FF,$35,$B6,$BA,$7C,$7F,$84,$93,$E1,$7C,$9F,$2B,$23,$C8
	.dc.b	$D9,$FF,$39,$CF,$F9,$E0,$93,$FE,$78,$7F,$DF,$34,$AF,$F4,$8B,$67
	.dc.b	$FC,$21,$F1,$AF,$96,$CD,$49,$FF,$45,$D6,$F0,$FF,$49,$35,$FF,$40
	.dc.b	$B5,$FE,$2E,$EF,$BB,$FD,$5B,$87,$F8,$0A,$BF,$C3,$D3,$97,$F8,$EC
	.dc.b	$66,$BF,$8E,$46,$3C,$5F,$F5,$3C,$FE,$BE,$F9,$F4,$62,$65,$7A,$9D
	.dc.b	$C6,$2F,$99,$FE,$DA,$B7,$A7,$D3,$E5,$BA,$B7,$0F,$86,$DA,$E2,$2B
	.dc.b	$8C,$FC,$8A,$F2,$F7,$48,$F9,$7E,$05,$BE,$37,$FC,$6C,$95,$FE,$D0
	.dc.b	$8F,$82,$D7,$CC,$FC,$0E,$CE,$5A,$7C,$A7,$E1,$FC,$B4,$C7,$EC,$FC
	.dc.b	$B5,$C3,$BC,$DB,$FA,$33,$6B,$FA,$5F,$C5,$BE,$17,$FD,$A6,$BE,$5A
	.dc.b	$C6,$1B,$3F,$E5,$59,$D3,$67,$4F,$F9,$67,$A7,$FD,$40,$DF,$43,$FD
	.dc.b	$BF,$1F,$DA,$F1,$FF,$77,$BA,$CF,$E6,$B5,$FD,$D4,$FC,$37,$F7,$0D
	.dc.b	$8D,$47,$F6,$47,$94,$F4,$F4,$7B,$D7,$81,$F6,$FC,$B4,$C3,$D5,$EE
	.dc.b	$5C,$7E,$EE,$6A,$C6,$2B,$F8,$F7,$87,$E4,$FB,$9F,$23,$D3,$F9,$2D
	.dc.b	$8F,$8C,$DC,$7A,$B6,$FE,$0F,$C1,$FC,$43,$E2,$FC,$25,$BF,$95,$1E
	.dc.b	$F2,$3F,$EF,$CD,$91,$8C,$3E,$4F,$C2,$71,$6D,$7D,$7F,$57,$EE,$9F
	.dc.b	$7B,$DF,$B1,$92,$4F,$C2,$8C,$BF,$13,$DF,$FC,$2F,$8D,$F8,$66,$D7
	.dc.b	$E0,$DB,$3B,$8F,$C3,$4F,$E0,$3C,$09,$1F,$85,$F1,$2D,$7E,$02,$6E
	.dc.b	$DF,$84,$9B,$5F,$76,$6D,$7F,$3B,$9A,$FF,$3B,$8B,$AF,$F3,$2A,$FC
	.dc.b	$94,$60,$FC,$8D,$58,$C9,$E3,$2D,$78,$DB,$7E,$36,$9E,$32,$C7,$8D
	.dc.b	$AF,$8D,$B7,$E3,$8E,$84,$9F,$D0,$DF,$92,$B6,$FA,$AF,$CD,$5B,$9F
	.dc.b	$CF,$46,$2B,$8C,$7A,$67,$47,$F5,$FA,$E5,$1D,$0F,$1D,$71,$7D,$53
	.dc.b	$FB,$08,$C3,$C6,$3A,$57,$68,$EA,$7B,$4E,$11,$D5,$CD,$FB,$E7,$81
	.dc.b	$27,$BB,$F2,$E7,$77,$E6,$6C,$F9,$8E,$93,$C6,$3E,$71,$A3,$BD,$FC
	.dc.b	$67,$FC,$01,$5F,$C1,$49,$5F,$A7,$6C,$7F,$9A,$8F,$71,$4F,$61,$CA
	.dc.b	$3F,$8D,$78,$9F,$8C,$23,$F7,$1B,$1F,$01,$B7,$E7,$8F,$F0,$74,$62
	.dc.b	$7E,$09,$FC,$5D,$C4,$78,$A7,$F9,$1C,$62,$3E,$74,$62,$7C,$04,$8F
	.dc.b	$03,$63,$C1,$82,$B9,$2B,$FE,$63,$18,$6D,$71,$3C,$B7,$FE,$27,$9D
	.dc.b	$47,$C2,$35,$1E,$77,$D4,$61,$F3,$B8,$7C,$A4,$BF,$2C,$F4,$8F,$E7
	.dc.b	$ED,$8F,$74,$6E,$7C,$6F,$CA,$7B,$5F,$AD,$F2,$46,$CF,$21,$E9,$77
	.dc.b	$7E,$E8,$D1,$E6,$34,$7F,$40,$F9,$6F,$EC,$3E,$AF,$F6,$26,$D7,$F5
	.dc.b	$F6,$CE,$E3,$FB,$14,$FF,$68,$F8,$C9,$1F,$D9,$7D,$AB,$5F,$D3,$26
	.dc.b	$E5,$FD,$2E,$03,$A3,$FD,$77,$19,$8C,$4F,$F6,$8F,$95,$98,$7F,$6D
	.dc.b	$F9,$6B,$5F,$DA,$27,$FB,$21,$27,$FA,$33,$63,$FC,$33,$F0,$3F,$D6
	.dc.b	$E7,$FB,$B4,$F8,$F0,$F0,$4E,$6B,$FB,$1F,$AB,$FE,$04,$F1,$7F,$C0
	.dc.b	$AD,$C3,$FC,$04,$D9,$DC,$7F,$81,$67,$FC,$17,$F2,$D2,$3F,$C1,$1F
	.dc.b	$2D,$6B,$FB,$7C,$DC,$7E,$06,$AA,$C5,$FF,$B8,$32,$7C,$B7,$9E,$30
	.dc.b	$DC,$7E,$5C,$BE,$3B,$10,$8E,$35,$F4,$31,$F7,$DF,$F8,$36,$73,$5F
	.dc.b	$C0,$DA,$FF,$07,$6B,$E0,$AE,$47,$97,$E5,$BB,$CF,$2B,$F9,$71,$E2
	.dc.b	$F8,$FF,$F2,$F5,$FE,$01,$7C,$CE,$7D,$2D,$37,$1E,$B5,$B8,$C3,$FE
	.dc.b	$52,$AE,$33,$A8,$43,$E4,$FC,$07,$87,$8E,$D6,$E7,$4D,$9D,$34,$D7
	.dc.b	$1F,$A4,$F9,$7F,$A7,$FE,$17,$4F,$CB,$5A,$75,$5F,$CA,$F1,$F1,$C7
	.dc.b	$F8,$E6,$75,$39,$6C,$FF,$31,$6D,$7E,$86,$DF,$35,$D2,$BF,$CE,$31
	.dc.b	$8B,$F8,$0F,$FB,$76,$B8,$7B,$9E,$6F,$7B,$CE,$DA,$DB,$6F,$9B,$EC
	.dc.b	$E2,$93,$D3,$8A,$D7,$82,$6E,$1E,$59,$B5,$FB,$AA,$7B,$2F,$93,$FC
	.dc.b	$8B,$67,$CF,$C7,$89,$91,$FE,$CD,$7F,$DA,$04,$70,$7B,$BE,$2B,$77
	.dc.b	$F0,$9F,$F3,$BF,$CC,$7F,$99,$7A,$0D,$25,$7D,$38,$7C,$C6,$C9,$19
	.dc.b	$7F,$CD,$DC,$1F,$EF,$9C,$3F,$B2,$35,$FE,$D0,$D7,$E0,$35,$FC,$E7
	.dc.b	$C5,$FD,$F2,$57,$DF,$7F,$05,$F1,$52,$3F,$04,$D9,$EF,$F3,$5E,$22
	.dc.b	$4F,$E2,$ED,$1F,$E8,$3F,$98,$FC,$CD,$7F,$79,$6B,$FD,$F5,$AF,$D5
	.dc.b	$6B,$FC,$C3,$E6,$3F,$83,$7A,$FF,$97,$6C,$7F,$80,$D7,$2B,$3D,$72
	.dc.b	$FF,$A9,$EB,$FC,$11,$F3,$BF,$EE,$77,$F4,$0D,$73,$F0,$7F,$A3,$51
	.dc.b	$97,$FC,$25,$EA,$BA,$74,$D8,$FD,$53,$5F,$F0,$B3,$5E,$3F,$03,$C0
	.dc.b	$8F,$07,$C1,$F5,$A9,$A4,$9F,$DB,$46,$5F,$DA,$7C,$8F,$9F,$B8,$D2
	.dc.b	$47,$EE,$7E,$4B,$44,$70,$7D,$7C,$FE,$07,$1F,$BE,$3F,$0F,$18,$3B
	.dc.b	$7B,$12,$3F,$82,$35,$FE,$09,$EC,$78,$C8,$FF,$09,$7A,$FC,$B4,$8F
	.dc.b	$C3,$35,$FE,$19,$F3,$5E,$B2,$3F,$C4,$5F,$77,$CD,$48,$FE,$28,$D7
	.dc.b	$F4,$3F,$13,$EA,$23,$F9,$37,$E3,$7A,$F2,$3F,$8E,$35,$FD,$4F,$C4
	.dc.b	$FC,$22,$3F,$9B,$7F,$09,$F3,$2F,$77,$F8,$57,$91,$FE,$01,$F9,$AF
	.dc.b	$E7,$57,$FC,$F5,$AA,$3F,$C0,$D5,$C6,$3F,$A1,$2F,$7B,$9F,$83,$FC
	.dc.b	$56,$35,$3C,$63,$18,$6A,$3F,$C2,$73,$C5,$FE,$99,$C6,$F0,$B5,$0F
	.dc.b	$5D,$F1,$18,$6C,$70,$9D,$D3,$E1,$6C,$49,$EB,$1D,$3E,$69,$F0,$9C
	.dc.b	$F9,$65,$7F,$5D,$9C,$B7,$E1,$38,$54,$8C,$1B,$EE,$E9,$32,$30,$73
	.dc.b	$DB,$D1,$DF,$23,$07,$7E,$1F,$24,$C8,$C1,$E3,$F8,$3F,$A4,$FF,$A2
	.dc.b	$72,$46,$23,$FD,$1D,$18,$E3,$FE,$C8,$D8,$E1,$F0,$6F,$9A,$CB,$63
	.dc.b	$31,$83,$F8,$53,$FE,$1B,$3F,$86,$7B,$58,$23,$FD,$C7,$C7,$FF,$10
	.dc.b	$BF,$C6,$79,$FF,$C0,$CF,$F8,$98,$7C,$D3,$97,$FD,$49,$5C,$39,$65
	.dc.b	$F0,$8B,$7F,$89,$3F,$B0,$72,$F6,$3F,$C5,$FC,$CF,$47,$17,$E0,$F8
	.dc.b	$8F,$19,$8B,$E4,$7F,$98,$9F,$3F,$8B,$27,$81,$E8,$7F,$C7,$7C,$3F
	.dc.b	$C6,$E3,$FC,$40,$DA,$E7,$83,$BF,$7D,$1C,$39,$66,$1C,$22,$DF,$8B
	.dc.b	$E6,$BE,$33,$76,$F5,$FC,$77,$F9,$2A,$73,$9F,$F9,$D5,$FD,$26,$1F
	.dc.b	$E3,$16,$DF,$18,$C7,$07,$F9,$4B,$E6,$9F,$0B,$76,$C3,$6B,$FC,$75
	.dc.b	$3F,$AD,$7F,$C5,$4D,$BF,$F6,$B3,$5F,$F6,$D3,$5F,$F0,$97,$CD,$33
	.dc.b	$FF,$45,$46,$1A,$FF,$A4,$23,$55,$FE,$37,$6C,$7F,$B8,$FE,$6B,$F0
	.dc.b	$93,$FE,$3B,$6C,$67,$97,$9A,$8C,$BF,$EE,$D7,$FB,$7B,$9F,$7F,$71
	.dc.b	$83,$3F,$35,$E0,$9F,$F7,$A3,$5F,$F1,$63,$5F,$F5,$3B,$57,$CD,$7F
	.dc.b	$08,$C3,$FC,$A0,$D8,$FF,$7E,$35,$FF,$80,$0D,$7F,$C7,$1F,$41,$E8
	.dc.b	$6E,$3E,$C8,$C4,$7F,$AD,$9A,$BE,$25,$E4,$FC,$26,$E1,$FE,$61,$6C
	.dc.b	$FF,$91,$4F,$F5,$FB,$87,$1F,$37,$FC,$AA,$7F,$CD,$0D,$8E,$01,$C3
	.dc.b	$C3,$C8,$FA,$3E,$2F,$F8,$F2,$BF,$CE,$0D,$9E,$2D,$7C,$4D,$7F,$CB
	.dc.b	$1F,$37,$FE,$1E,$9F,$F3,$C3,$63,$C2,$D7,$D4,$D7,$FC,$C1,$F3,$5F
	.dc.b	$E4,$14,$D2,$7C,$DF,$E1,$BE,$7D,$E3,$52,$61,$FE,$87,$8C,$BF,$E8
	.dc.b	$92,$DF,$C2,$6B,$F2,$DA,$FF,$9C,$5A,$FF,$BB,$3E,$4B,$FD,$DB,$1F
	.dc.b	$E9,$36,$C7,$CC,$6B,$F5,$5A,$FF,$9E,$FE,$7F,$F1,$C4,$E7,$FC,$FB
	.dc.b	$F9,$1F,$87,$7A,$5F,$A2,$6E,$1F,$EA,$16,$DF,$DA,$6B,$F8,$3F,$65
	.dc.b	$F2,$7F,$EA,$52,$57,$FA,$95,$FF,$39,$F0,$48,$FF,$31,$36,$BA,$7C
	.dc.b	$77,$84,$93,$E1,$7F,$D6,$1F,$29,$23,$FD,$60,$D9,$FF,$3E,$FC,$DF
	.dc.b	$FA,$00,$93,$FE,$6F,$7F,$7B,$B7,$F7,$98,$FF,$67,$36,$7F,$9C,$BF
	.dc.b	$C8,$3C,$44,$FF,$A8,$E3,$51,$89,$FA,$9E,$67,$42,$57,$D5,$6C,$74
	.dc.b	$F2,$FF,$DF,$37,$CA,$BA,$47,$FB,$F3,$E2,$1F,$2F,$4C,$7A,$D6,$BF
	.dc.b	$DC,$93,$A6,$CE,$9A,$6B,$D6,$E9,$E6,$75,$6C,$64,$E7,$19,$69,$F1
	.dc.b	$3D,$CF,$39,$E8,$F7,$6C,$64,$FC,$94,$7C,$C2,$38,$3E,$6F,$CB,$F0
	.dc.b	$D7,$E6,$BE,$FF,$42,$4F,$B3,$D9,$B5,$FD,$AB,$4F,$F2,$54,$FC,$0A
	.dc.b	$F8,$D5,$AE,$1F,$A5,$6B,$FF,$0B,$DF,$C6,$ED,$F9,$9E,$CF,$CD,$6C
	.dc.b	$64,$F9,$71,$97,$F0,$7D,$1F,$58,$FA,$FD,$76,$32,$7E,$C6,$3F,$16
	.dc.b	$47,$07,$E3,$39,$F8,$63,$F8,$C7,$EF,$7C,$BF,$DF,$F3,$ED,$7D,$C6
	.dc.b	$D7,$C1,$6E,$9E,$ED,$3F,$B7,$D7,$E0,$BE,$5F,$F1,$07,$E2,$7D,$1B
	.dc.b	$19,$26,$1F,$2E,$32,$FE,$7B,$E4,$BF,$26,$7E,$53,$E4,$ED,$7E,$92
	.dc.b	$3F,$4E,$47,$F8,$83,$67,$FC,$3F,$ED,$FE,$9A,$78,$08,$CB,$F9,$DF
	.dc.b	$2F,$F3,$DF,$2F,$63,$F3,$6D,$9F,$C0,$37,$4F,$80,$FE,$97,$E5,$E5
	.dc.b	$7E,$8B,$E5,$ED,$FE,$1E,$6D,$FC,$79,$B5,$E3,$A6,$2F,$95,$FD,$D5
	.dc.b	$FE,$2A,$31,$5E,$98,$C7,$C6,$6A,$39,$4E,$AF,$E0,$DE,$5F,$81,$19
	.dc.b	$BE,$35,$CC,$E3,$5B,$AF,$1B,$6B,$84,$F6,$BC,$4C,$8E,$55,$DA,$38
	.dc.b	$4E,$63,$C7,$3A,$7A,$B6,$93,$BB,$F8,$17,$6B,$C0,$6D,$C5,$F7,$8E
	.dc.b	$55,$88,$EA,$78,$0F,$C1,$B9,$BF,$5C,$71,$33,$3D,$48,$F1,$BF,$8E
	.dc.b	$D8,$F8,$E6,$A3,$27,$A1,$C1,$A8,$EA,$71,$3C,$11,$B7,$2D,$24,$71
	.dc.b	$7A,$B4,$91,$C9,$F8,$AF,$53,$E0,$C7,$27,$F0,$35,$F0,$6F,$E1,$36
	.dc.b	$BC,$F3,$A6,$BC,$9E,$EF,$E5,$A3,$83,$DE,$FE,$CC,$F0,$7E,$6C,$71
	.dc.b	$3A,$1B,$AF,$0C,$ED,$B1,$98,$E5,$5D,$A3,$FC,$79,$5F,$9D,$6D,$7F
	.dc.b	$99,$23,$11,$F3,$CC,$D7,$F8,$F1,$FC,$19,$83,$FC,$7F,$1E,$61,$F2
	.dc.b	$BF,$85,$24,$FA,$DB,$3E,$F6,$E1,$F3,$63,$F7,$31,$89,$F8,$F5,$D1
	.dc.b	$E4,$49,$E4,$DA,$E5,$3B,$9C,$92,$33,$5D,$63,$F0,$07,$EF,$1C,$3F
	.dc.b	$3C,$F1,$D7,$53,$C7,$1A,$AF,$80,$DB,$FD,$F3,$67,$F6,$8D,$AF,$E5
	.dc.b	$6D,$BF,$E2,$8D,$9F,$E2,$2D,$9F,$24,$78,$5F,$39,$E4,$3D,$8F,$90
	.dc.b	$D3,$E9,$BF,$85,$5F,$02,$73,$1E,$87,$7E,$43,$C7,$FE,$30,$BE,$5E
	.dc.b	$C6,$1B,$3F,$C4,$67,$4D,$9D,$39,$F6,$79,$9D,$38,$AC,$64,$99,$7F
	.dc.b	$BB,$2D,$3A,$8C,$71,$79,$0F,$E5,$9F,$17,$FD,$4E,$33,$7E,$39,$FF
	.dc.b	$5E,$D9,$A4,$8F,$CD,$46,$5F,$E5,$54,$F9,$7E,$D1,$87,$BF,$C7,$78
	.dc.b	$3E,$42,$CF,$06,$E2,$DF,$FC,$2F,$87,$FA,$4D,$B7,$C6,$FB,$F1,$7A
	.dc.b	$8F,$5F,$15,$8C,$93,$0E,$8F,$F5,$DC,$3F,$81,$9F,$C1,$A3,$27,$F7
	.dc.b	$3F,$3D,$F2,$FF,$D4,$5B,$5E,$EE,$4C,$4C,$8F,$F4,$03,$FE,$83,$23
	.dc.b	$83,$E2,$FC,$36,$EF,$B8,$C1,$1F,$53,$E6,$FE,$AF,$A9,$6F,$E6,$B6
	.dc.b	$F8,$DF,$FB,$E3,$6F,$8F,$7C,$2F,$E9,$FB,$3F,$78,$FB,$FF,$37,$63
	.dc.b	$24,$C3,$9C,$65,$EB,$F3,$7F,$86,$3F,$B9,$7A,$AF,$96,$C5,$FE,$02
	.dc.b	$7F,$C9,$B5,$F3,$5A,$FF,$87,$73,$46,$23,$E0,$36,$33,$3F,$DC,$ED
	.dc.b	$F3,$7D,$AF,$E3,$C6,$27,$5E,$9F,$93,$D4,$F2,$B6,$32,$73,$8C,$B4
	.dc.b	$ED,$EA,$7B,$0F,$6F,$A9,$63,$24,$C3,$C3,$19,$7C,$7E,$8F,$C4,$3E
	.dc.b	$2F,$A3,$6B,$D5,$1E,$FC,$BF,$CE,$B5,$FE,$FC,$6B,$FE,$2F,$0E,$1F
	.dc.b	$F1,$93,$C3,$8B,$FC,$F2,$FF,$9E,$23,$C3,$5F,$1F,$D0,$FF,$52,$7C
	.dc.b	$D7,$CD,$F3,$3F,$DF,$D5,$FE,$4D,$6B,$FE,$81,$8F,$A1,$EE,$4B,$FE
	.dc.b	$A9,$C3,$FE,$FB,$3F,$DF,$98,$FF,$BB,$2F,$CB,$E9,$FA,$C9,$5F,$C0
	.dc.b	$BE,$23,$F8,$39,$1C,$1D,$3E,$6B,$EF,$11,$C1,$FD,$8E,$DE,$97,$C7
	.dc.b	$24,$65,$FC,$47,$CC,$FF,$8D,$3A,$3F,$B5,$37,$2F,$94,$17,$88,$F8
	.dc.b	$CD,$F8,$9F,$99,$AD,$FC,$CB,$E6,$3F,$BA,$42,$30,$FF,$77,$6B,$C2
	.dc.b	$37,$6F,$42,$92,$7F,$BF,$46,$5F,$F0,$07,$4B,$E0,$3B,$B6,$33,$18
	.dc.b	$3C,$7F,$32,$9A,$49,$FF,$05,$46,$5F,$95,$EC,$7A,$B8,$4D,$24,$7F
	.dc.b	$38,$F8,$9F,$7F,$9C,$69,$23,$F0,$7A,$7C,$4E,$7F,$83,$1F,$97,$6B
	.dc.b	$FC,$AE,$3C,$1E,$CC,$8F,$B8,$D7,$EA,$DB,$DE,$F0,$92,$3F,$A5,$BD
	.dc.b	$FE,$56,$47,$E5,$1A,$FE,$A7,$DB,$F7,$11,$FC,$53,$EC,$F9,$89,$1F
	.dc.b	$D5,$9A,$FE,$1B,$27,$C9,$7B,$49,$1F,$A2,$7E,$57,$BB,$23,$F2,$CD
	.dc.b	$79,$7C,$5F,$C5,$23,$FA,$A7,$E9,$7B,$D2,$3F,$3C,$D7,$F2,$FF,$2E
	.dc.b	$F1,$FF,$D6,$1E,$3F,$FB,$DB,$E6,$1E,$E6,$3C,$2F,$F5,$A4,$63,$0F
	.dc.b	$F4,$C6,$1C,$5B,$E5,$6D,$FF,$66,$46,$5F,$E9,$ED,$C1,$7F,$EB,$F8
	.dc.b	$57,$1A,$FC,$27,$8B,$78,$5E,$8D,$35,$D0,$9B,$F1,$18,$65,$FF,$AB
	.dc.b	$65,$3D,$7E,$61,$E7,$F5,$E7,$97,$FE,$B3,$9C,$B7,$CF,$E1,$52,$30
	.dc.b	$6F,$C0,$E9,$32,$30,$73,$E3,$F5,$78,$32,$30,$77,$DF,$C9,$32,$30
	.dc.b	$78,$FC,$FF,$F0,$A3,$FC,$3B,$24,$62,$3F,$9D,$46,$38,$FF,$CF,$CD
	.dc.b	$8E,$1E,$63,$E6,$32,$D8,$CC,$60,$FE,$26,$FF,$BC,$8F,$E2,$BF,$05
	.dc.b	$82,$3F,$E8,$BF,$2B,$FD,$CC,$FF,$1F,$E7,$FF,$6E,$3F,$EE,$E1,$F3
	.dc.b	$1F,$BF,$7F,$DF,$D3,$C3,$96,$57,$08,$B3,$FE,$16,$3E,$CC,$74,$E4
	.dc.b	$FF,$7A,$73,$FF,$86,$23,$07,$83,$E2,$BC,$67,$92,$6C,$79,$3E,$63
	.dc.b	$CE,$62,$FB,$3F,$EF,$FD,$FF,$90,$4F,$FB,$D5,$B3,$FE,$28,$C3,$F7
	.dc.b	$33,$9E,$5E,$1C,$52,$30,$71,$7F,$3F,$5F,$E3,$0F,$98,$F8,$CD,$D3
	.dc.b	$D7,$ED,$EB,$2D,$CA,$B7,$18,$7F,$41,$82,$7A,$6D,$C6,$31,$1C,$70
	.dc.b	$C3,$BA,$E2,$FF,$90,$BE,$61,$E1,$6A,$11,$8B,$ED,$8F,$92,$1E,$0E
	.dc.b	$DD,$57,$3B,$FF,$17,$46,$23,$FC,$1C,$D7,$CA,$D7,$D5,$F3,$1F,$BE
	.dc.b	$C3,$FD,$AC,$D8,$FE,$DA,$D7,$FC,$30,$D7,$DE,$BD,$CF,$F1,$F3,$67
	.dc.b	$9C,$60,$CF,$B9,$23,$BB,$6B,$FB,$DF,$CB,$F5,$7E,$73,$53,$49,$5F
	.dc.b	$E4,$D7,$C1,$1C,$A3,$06,$66,$57,$FB,$CE,$BF,$CA,$35,$E0,$9F,$F2
	.dc.b	$B6,$3F,$DD,$EB,$FD,$F0,$D9,$FF,$07,$B5,$FF,$1D,$35,$FC,$0B,$FE
	.dc.b	$3C,$8E,$3E,$7F,$42,$3D,$3D,$EE,$E4,$70,$7F,$87,$29,$F2,$9F,$E9
	.dc.b	$77,$FC,$90,$D7,$F1,$ED,$7F,$28,$D7,$8F,$C2,$7F,$9C,$60,$D2,$4F
	.dc.b	$78,$FF,$80,$13,$DB,$E0,$F8,$12,$7C,$9F,$08,$F3,$FF,$CF,$8D,$8C
	.dc.b	$C6,$0C,$FC,$E4,$8E,$2D,$BF,$F1,$CD,$3E,$3F,$99,$2B,$BB,$FE,$89
	.dc.b	$F9,$3F,$C1,$13,$CC,$C1,$FE,$6F,$F8,$E9,$5F,$E4,$D6,$E1,$FE,$46
	.dc.b	$F0,$FF,$84,$B7,$D2,$FF,$4A,$46,$23,$FC,$E8,$D7,$FC,$E7,$ED,$FF
	.dc.b	$9E,$C9,$7F,$E5,$97,$FC,$FD,$F3,$F2,$3F,$CB,$8D,$9F,$F4,$07,$CE
	.dc.b	$FF,$A0,$C9,$3F,$E6,$17,$FD,$0D,$F0,$BF,$EA,$E3,$F8,$17,$EA,$1F
	.dc.b	$C3,$67,$FD,$49,$19,$7F,$CD,$0F,$06,$92,$BF,$D9,$B3,$FC,$62,$3F
	.dc.b	$D1,$73,$D3,$C0,$EB,$E0,$F6,$F1,$ED,$7F,$0E,$3F,$8C,$DF,$F4,$0C
	.dc.b	$39,$D6,$E7,$51,$FC,$86,$7F,$8D,$F8,$3F,$CA,$1E,$0D,$F2,$3D,$67
	.dc.b	$B3,$C8,$B7,$CA,$D6,$AC,$C7,$AA,$AD,$F5,$F4,$FE,$29,$F1,$BE,$2E
	.dc.b	$E3,$CF,$1F,$F2,$E3,$FA,$96,$E1,$FE,$86,$6B,$FE,$B1,$6B,$F0,$2B
	.dc.b	$E8,$FC,$AF,$48,$F9,$6F,$D4,$F6,$FE,$AF,$CA,$DC,$3C,$0D,$AF,$A2
	.dc.b	$DC,$BE,$23,$F7,$3E,$76,$57,$9F,$B6,$DF,$D8,$9B,$7F,$2E,$6D,$7F
	.dc.b	$4C,$EA,$FE,$2F,$3F,$D5,$9B,$1F,$63,$B7,$FA,$54,$63,$0F,$EA,$CF
	.dc.b	$F1,$E7,$FD,$A0,$D7,$A7,$89,$FA,$D2,$57,$CB,$ED,$E7,$1F,$94,$7F
	.dc.b	$2B,$F2,$DF,$96,$F9,$6B,$5F,$45,$B7,$F9,$26,$E5,$F8,$57,$F3,$BF
	.dc.b	$2B,$2B,$EC,$7C,$B5,$BF,$CC,$CD,$BF,$C6,$4D,$AF,$CA,$FB,$FF,$A6
	.dc.b	$3F,$4F,$F2,$D6,$3F,$05,$1F,$DE,$B0,$DC,$7E,$0E,$31,$1F,$99,$6B
	.dc.b	$F9,$27,$F6,$3F,$2D,$CE,$3F,$56,$FF,$75,$FD,$47,$FC,$D3,$5F,$99
	.dc.b	$5F,$2F,$98,$D3,$5D,$36,$99,$7F,$83,$AB,$8D,$73,$F9,$7E,$C7,$7F
	.dc.b	$8C,$B1,$92,$5F,$27,$FC,$1F,$5F,$EE,$07,$37,$FE,$1A,$F1,$7F,$AC
	.dc.b	$37,$E0,$3F,$C1,$D0,$F8,$6B,$12,$BF,$69,$3C,$1F,$F0,$DB,$B7,$31
	.dc.b	$BC,$3C,$47,$F8,$8E,$78,$36,$B5,$7E,$1B,$E2,$76,$8C,$BD,$AE,$D7
	.dc.b	$18,$DB,$D8,$5F,$66,$CF,$B2,$B8,$C6,$2B,$8D,$FC,$39,$CD,$F1,$23
	.dc.b	$8B,$CB,$6B,$89,$F0,$9C,$5F,$8E,$B8,$CE,$A3,$E1,$57,$28,$C4,$78
	.dc.b	$CE,$11,$E7,$8C,$CF,$8E,$31,$3A,$9C,$C6,$2F,$E5,$4E,$48,$FB,$67
	.dc.b	$67,$4B,$F3,$C7,$03,$53,$C2,$7E,$81,$B3,$8B,$93,$E9,$14,$F5,$47
	.dc.b	$38,$F8,$87,$07,$11,$C1,$AF,$08,$EE,$FC,$99,$F2,$C6,$27,$CA,$70
	.dc.b	$BF,$F1,$65,$7C,$C9,$FF,$12,$46,$23,$E6,$1F,$DE,$2F,$FB,$D9,$27
	.dc.b	$6D,$8F,$B1,$E8,$7E,$08,$93,$F4,$9B,$1E,$FA,$FF,$1D,$36,$33,$7E
	.dc.b	$56,$D7,$A4,$FD,$3B,$89,$F4,$1E,$08,$EC,$47,$A3,$67,$A3,$4E,$91
	.dc.b	$FA,$D9,$E8,$FC,$B3,$11,$B7,$BC,$FF,$01,$9D,$D7,$E7,$DB,$7F,$C0
	.dc.b	$9B,$7F,$C0,$1B,$7E,$73,$CA,$4B,$F2,$B6,$7C,$AF,$ED,$27,$8C,$F8
	.dc.b	$67,$C9,$3E,$73,$4E,$2F,$C8,$DB,$F2,$34,$F2,$1C,$B1,$7C,$77,$F3
	.dc.b	$3F,$43,$F9,$A3,$67,$F9,$A3,$70,$FE,$11,$5C,$27,$F9,$C5,$7F,$3B
	.dc.b	$F3,$E4,$7F,$3B,$F9,$7B,$7C,$26,$E7,$FC,$DE,$6D,$7F,$24,$E9,$FE
	.dc.b	$56,$7F,$2C,$F5,$6C,$7F,$2A,$8F,$F2,$69,$1F,$CC,$B7,$0F,$E3,$9F
	.dc.b	$7C,$F9,$2E,$31,$87,$FA,$87,$CA,$FF,$51,$F9,$5B,$87,$F5,$16,$D7
	.dc.b	$B9,$B9,$FF,$4D,$7F,$AC,$7C,$84,$AF,$EB,$1F,$17,$6F,$E2,$CD,$BF
	.dc.b	$EA,$93,$6B,$FD,$B7,$97,$FB,$CA,$7F,$DC,$AD,$8C,$FC,$57,$FB,$66
	.dc.b	$31,$87,$FB,$95,$FF,$7C,$BF,$EE,$40,$E1,$EB,$F0,$7F,$DC,$A3,$11
	.dc.b	$C6,$3A,$7C,$1F,$56,$DF,$F7,$1F,$5F,$FB,$A1,$FD,$D7,$A6,$C6,$49
	.dc.b	$7F,$82,$8C,$BF,$84,$E7,$FE,$F9,$D3,$FD,$F5,$B5,$FD,$F5,$B3,$B8
	.dc.b	$FF,$00,$4F,$BF,$E4,$64,$7F,$96,$F3,$7C,$AE,$2F,$EE,$D7,$B4,$95
	.dc.b	$FE,$BD,$A4,$62,$7F,$07,$F2,$5C,$77,$F9,$EF,$F8,$47,$DF,$98,$7F
	.dc.b	$84,$7E,$1A,$DF,$E7,$A6,$DF,$F8,$36,$6D,$7F,$81,$BD,$EF,$F0,$D7
	.dc.b	$4F,$F8,$6D,$B1,$FE,$1B,$6E,$1F,$A8,$6D,$FE,$03,$DD,$FB,$C7,$DF
	.dc.b	$F7,$AD,$FC,$18,$FB,$44,$7D,$CD,$BF,$CF,$FB,$DC,$1B,$3F,$E2,$BE
	.dc.b	$CF,$F1,$9F,$81,$FE,$55,$90,$70,$D8,$FC,$47,$BB,$2B,$E6,$36,$7F
	.dc.b	$20,$3D,$D7,$A7,$FE,$40,$6D,$7F,$BB,$DB,$3C,$1A,$FE,$67,$E8,$9C
	.dc.b	$5F,$E4,$86,$D7,$FB,$DD,$B3,$E0,$6B,$FA,$1F,$A1,$79,$9F,$E5,$06
	.dc.b	$D7,$FB,$FD,$B3,$E8,$6B,$FA,$9F,$97,$FE,$43,$E2,$7B,$21,$18,$7D
	.dc.b	$EB,$CA,$F8,$11,$8F,$27,$F8,$73,$F1,$3F,$52,$AD,$BE,$8A,$32,$FE
	.dc.b	$E5,$AF,$D4,$35,$F3,$7C,$20,$D2,$47,$9E,$1F,$37,$B2,$46,$5E,$5F
	.dc.b	$29,$23,$83,$5F,$47,$8D,$DC,$8F,$57,$C5,$E2,$C8,$F0,$35,$F8,$1E
	.dc.b	$37,$9C,$8F,$95,$F5,$7C,$A4,$8F,$43,$5F,$91,$F2,$9E,$F2,$3E,$D7
	.dc.b	$E1,$7C,$9C,$8F,$80,$D7,$E8,$78,$BF,$1C,$8F,$C5,$7E,$57,$89,$23
	.dc.b	$E4,$35,$FB,$1C,$FF,$3C,$8F,$CD,$7E,$97,$C8,$C8,$FA,$0D,$7F,$01
	.dc.b	$F1,$9F,$5C,$8F,$D5,$7E,$D7,$C8,$48,$FB,$0D,$7F,$11,$E1,$7D,$F2
	.dc.b	$3F,$75,$FC,$4F,$CA,$B6,$FC,$37,$88,$EB,$FC,$47,$CB,$7E,$8B,$0F
	.dc.b	$DE,$CE,$5F,$DF,$C6,$5F,$F4,$05,$F3,$77,$5F,$91,$87,$0F,$68,$C6
	.dc.b	$16,$BF,$84,$D7,$17,$9D,$65,$7C,$33,$E2,$F1,$B3,$C8,$A6,$E7,$55
	.dc.b	$E2,$3F,$8A,$AF,$8C,$D5,$C9,$D1,$DA,$3C,$1F,$2C,$E9,$EF,$E9,$CA
	.dc.b	$DC,$E5,$BD,$5C,$2A,$46,$0D,$F5,$74,$99,$18,$39,$EF,$EC,$F4,$64
	.dc.b	$60,$EF,$BF,$92,$64,$60,$F1,$F5,$FF,$B6,$1F,$E5,$79,$23,$11,$FD
	.dc.b	$F6,$31,$C9,$FC,$55,$B1,$C3,$AD,$F2,$D9,$6C,$66,$30,$7F,$12,$7F
	.dc.b	$D7,$E7,$F1,$4F,$6F,$04,$7F,$9C,$73,$7C,$E7,$F9,$F3,$5E,$3F,$2D
	.dc.b	$F7,$9F,$E3,$1E,$4F,$E3,$F2,$FC,$84,$7F,$4E,$E4,$C5,$FF,$85,$DB
	.dc.b	$1C,$23,$B4,$62,$3F,$AA,$F2,$57,$F0,$53,$F8,$0E,$77,$B5,$9A,$D7
	.dc.b	$C7,$CA,$F4,$45,$BF,$F6,$F0,$F8,$FE,$86,$72,$F3,$C6,$6B,$F1,$B1
	.dc.b	$FE,$30,$83,$71,$7B,$3E,$78,$D1,$FB,$5A,$E3,$C7,$9E,$7D,$3E,$DF
	.dc.b	$96,$F9,$78,$FF,$BC,$5B,$7F,$23,$17,$AD,$BE,$0E,$1C,$13,$3F,$D0
	.dc.b	$E1,$C1,$FE,$90,$FE,$BD,$FD,$2F,$CB,$3C,$3F,$6E,$7F,$E0,$08,$C4
	.dc.b	$FF,$5F,$6C,$7F,$7C,$6B,$FE,$4D,$73,$9F,$DC,$5E,$17,$D7,$73,$5F
	.dc.b	$D6,$5C,$BF,$B6,$F9,$67,$B1,$E2,$C1,$18,$9F,$EE,$0D,$8F,$F0,$6B
	.dc.b	$5F,$F2,$F3,$FE,$16,$27,$AB,$CD,$FF,$0C,$FD,$13,$E4,$FB,$36,$FF
	.dc.b	$BE,$36,$FF,$C3,$2D,$7F,$CD,$5F,$2F,$DD,$FE,$E4,$D7,$FC,$4E,$D4
	.dc.b	$D3,$FC,$52,$4C,$3D,$CF,$F8,$B3,$E8,$9E,$1F,$C1,$6E,$1F,$E0,$F6
	.dc.b	$DF,$F8,$B1,$AF,$F9,$F7,$7F,$F7,$3B,$7D,$FF,$95,$18,$8F,$F0,$43
	.dc.b	$5F,$F1,$EA,$F7,$B6,$4B,$FF,$0F,$46,$5F,$F1,$07,$D0,$FB,$49,$BF
	.dc.b	$D7,$7E,$07,$CC,$BE,$1F,$ED,$36,$BF,$C5,$0D,$C7,$FC,$8A,$D7,$FD
	.dc.b	$2C,$BE,$1F,$89,$27,$8B,$F8,$5F,$A1,$7C,$3D,$24,$FF,$8D,$63,$2F
	.dc.b	$F8,$DF,$E6,$9F,$29,$F8,$D6,$C6,$63,$06,$7E,$4E,$47,$16,$E5,$FE
	.dc.b	$5C,$F9,$C7,$CA,$7E,$5D,$B5,$FE,$45,$6C,$EE,$7F,$39,$C5,$FA,$08
	.dc.b	$FF,$32,$B5,$FF,$5B,$35,$FF,$1F,$35,$FF,$39,$F9,$5F,$C3,$2F,$FC
	.dc.b	$A4,$D8,$FF,$37,$B5,$FF,$60,$35,$FF,$24,$7B,$8F,$89,$FD,$73,$6B
	.dc.b	$FC,$8D,$18,$3C,$FF,$13,$23,$FC,$C1,$82,$31,$7E,$AF,$9B,$F6,$92
	.dc.b	$B6,$F6,$99,$1F,$E5,$36,$CF,$FA,$1B,$E5,$9D,$9F,$CF,$5B,$5F,$DE
	.dc.b	$BE,$6F,$EB,$D7,$F2,$E8,$FF,$31,$4E,$A3,$07,$FA,$61,$FF,$4B,$92
	.dc.b	$BF,$86,$36,$7A,$6F,$DE,$7F,$A6,$5A,$2E,$D8,$F9,$7B,$7E,$36,$E3
	.dc.b	$E0,$16,$8C,$F8,$9F,$D7,$A3,$15,$EB,$F9,$7E,$E7,$83,$E5,$EC,$64
	.dc.b	$98,$75,$A4,$EA,$31,$F1,$7E,$83,$D3,$F2,$F6,$FE,$44,$7F,$2B,$BF
	.dc.b	$F4,$DD,$7F,$A2,$5A,$FF,$AF,$9A,$FA,$A3,$E1,$FC,$BF,$38,$F1,$3F
	.dc.b	$DC,$7E,$47,$FC,$99,$7F,$E2,$E5,$F2,$F9,$8D,$35,$D3,$6B,$E4,$4F
	.dc.b	$09,$E5,$F2,$FD,$4E,$DF,$1D,$63,$24,$AE,$31,$97,$E9,$74,$BE,$03
	.dc.b	$FC,$0C,$DA,$FF,$21,$AF,$95,$CC,$C9,$FD,$5B,$FD,$46,$78,$57,$EB
	.dc.b	$23,$07,$4E,$9E,$11,$E0,$7E,$27,$C7,$FC,$5F,$12,$E1,$F1,$5B,$5F
	.dc.b	$01,$BC,$9F,$E2,$2D,$7F,$92,$37,$C3,$F9,$C7,$CF,$F7,$2E,$1F,$3E
	.dc.b	$2E,$1F,$2A,$6D,$7F,$8B,$34,$FD,$23,$7A,$BE,$1F,$8F,$F1,$19,$8F
	.dc.b	$F0,$A7,$E0,$11,$C1,$FE,$6A,$F9,$57,$C8,$70,$CB,$F4,$11,$89,$FC
	.dc.b	$17,$CB,$FE,$17,$C8,$FC,$33,$63,$27,$E1,$E3,$2D,$3F,$15,$F2,$FF
	.dc.b	$8D,$3F,$1D,$F1,$16,$32,$4C,$3F,$13,$19,$7C,$7E,$EF,$F9,$17,$1F
	.dc.b	$F1,$A1,$FD,$A3,$5F,$F2,$7B,$88,$D1,$BF,$46,$E1,$EF,$8B,$7E,$E9
	.dc.b	$B7,$CA,$6D,$7F,$BC,$EF,$B3,$63,$36,$13,$E4,$BC,$C6,$26,$CF,$28
	.dc.b	$E0,$F9,$EB,$C8,$7D,$3A,$DD,$70,$6C,$6A,$BD,$0D,$BF,$43,$4F,$44
	.dc.b	$F0,$9F,$93,$1D,$4D,$3E,$2A,$ED,$18,$8F,$43,$6F,$D1,$18,$39,$92
	.dc.b	$FE,$8D,$F7,$7C,$71,$E6,$35,$5F,$24,$E9,$3B,$71,$1B,$9E,$24,$8C
	.dc.b	$BD,$9F,$25,$FA,$27,$A9,$EB,$8E,$A6,$8A,$E8,$F5,$4F,$57,$E9,$D7
	.dc.b	$8D,$E4,$78,$E7,$C2,$47,$33,$E3,$7A,$5A,$FF,$CB,$CD,$AF,$C3,$4C
	.dc.b	$3E,$51,$63,$4E,$4C,$9F,$10,$D4,$F4,$24,$FE,$E7,$DD,$FC,$69,$1F
	.dc.b	$B0,$D9,$F2,$B6,$FE,$04,$7E,$62,$31,$3E,$D8,$C3,$C8,$8F,$26,$CF
	.dc.b	$2B,$FF,$50,$4E,$49,$19,$AE,$B1,$F2,$2B,$F5,$91,$88,$FA,$71,$83
	.dc.b	$A9,$E3,$AE,$B5,$CE,$33,$3C,$CF,$D5,$D7,$3A,$FE,$0A,$DB,$FE,$10
	.dc.b	$DA,$F9,$07,$2A,$F0,$9C,$9C,$57,$26,$D7,$26,$9C,$A7,$DE,$F9,$EB
	.dc.b	$B4,$7F,$2A,$6D,$FF,$16,$7E,$07,$1F,$F8,$F6,$3F,$A3,$7B,$DF,$DC
	.dc.b	$7A,$7F,$B9,$33,$0F,$EE,$4E,$5A,$7F,$75,$E9,$FE,$F1,$D3,$FD,$E5
	.dc.b	$B1,$93,$FB,$D4,$65,$A7,$F4,$1F,$8F,$FE,$58,$7F,$2D,$F9,$0B,$19
	.dc.b	$3F,$29,$A7,$F1,$E9,$FC,$E9,$23,$3B,$23,$F9,$B7,$C8,$72,$AF,$E9
	.dc.b	$6F,$F8,$37,$D5,$97,$FE,$0D,$F8,$0B,$5F,$E0,$89,$BB,$7F,$69,$9B
	.dc.b	$5F,$AA,$DD,$F1,$BF,$E1,$7E,$AF,$F0,$C3,$67,$FC,$30,$DC,$7F,$C2
	.dc.b	$6D,$C3,$FA,$3F,$9F,$FD,$2C,$FE,$99,$EF,$5B,$FE,$C9,$1F,$BA,$23
	.dc.b	$EE,$6D,$FF,$66,$F3,$F8,$36,$7F,$A2,$59,$FF,$3A,$FA,$16,$7F,$C9
	.dc.b	$9E,$8A,$31,$59,$94,$BE,$5F,$37,$C8,$EA,$59,$E7,$7A,$7F,$AA,$57
	.dc.b	$19,$DC,$62,$74,$FF,$96,$CB,$E7,$BE,$0A,$C6,$1B,$5C,$A3,$FC,$C5
	.dc.b	$18,$7F,$CC,$D3,$C6,$74,$BE,$0B,$A9,$DB,$E1,$6C,$EA,$3E,$05,$7F
	.dc.b	$26,$AE,$13,$F9,$06,$C7,$F2,$56,$BF,$CA,$DC,$E7,$F2,$57,$C3,$76
	.dc.b	$9F,$2B,$EE,$F8,$69,$7E,$6F,$8E,$B5,$EA,$9B,$87,$39,$B5,$E5,$F8
	.dc.b	$3F,$8F,$F3,$7F,$21,$B1,$F0,$1B,$87,$C4,$6D,$FD,$DF,$BB,$FF,$16
	.dc.b	$4F,$FA,$35,$B1,$E9,$F9,$CF,$AB,$18,$BF,$E9,$2F,$0A,$FE,$94,$D7
	.dc.b	$3E,$57,$CA,$7F,$C2,$91,$96,$C7,$DE,$F9,$CE,$AD,$AF,$F2,$17,$27
	.dc.b	$F5,$82,$3F,$BC,$7A,$57,$E1,$7E,$6E,$F1,$9A,$4A,$FE,$0F,$F8,$1F
	.dc.b	$FA,$06,$BF,$AE,$35,$FE,$C4,$D7,$FB,$2B,$5F,$F1,$67,$CD,$FF,$A6
	.dc.b	$09,$EC,$F8,$9F,$F4,$DF,$CD,$BD,$6F,$F2,$CD,$7F,$69,$8C,$69,$FC
	.dc.b	$61,$7A,$FF,$9D,$9D,$57,$07,$FB,$A7,$D0,$3D,$1F,$E3,$F0,$8C,$3F
	.dc.b	$C8,$9A,$FF,$77,$6B,$FD,$F7,$B5,$F3,$9C,$9B,$1F,$CA,$9A,$FF,$7F
	.dc.b	$6B,$C7,$E7,$38,$11,$E0,$F9,$3B,$3E,$84,$7E,$97,$87,$04,$7B,$7A
	.dc.b	$5E,$12,$38,$3C,$5E,$97,$94,$8E,$0F,$D8,$7C,$F7,$72,$38,$3D,$5F
	.dc.b	$33,$ED,$23,$83,$DD,$F3,$3F,$04,$8E,$0F,$DE,$7C,$4F,$AC,$8E,$0F
	.dc.b	$8D,$E8,$FC,$92,$38,$3E,$57,$A1,$F3,$48,$E0,$FE,$11,$F0,$FF,$1C
	.dc.b	$8E,$0F,$A5,$F2,$BF,$54,$8E,$0F,$AD,$F1,$FF,$64,$8E,$0F,$E2,$9F
	.dc.b	$13,$F4,$C8,$E0,$FB,$DE,$77,$E0,$48,$E0,$FE,$39,$C3,$F6,$C8,$E0
	.dc.b	$FE,$DF,$F3,$6F,$6F,$FC,$93,$E3,$7F,$45,$F9,$C7,$81,$8F,$0B,$FC
	.dc.b	$9B,$18,$C3,$F9,$B6,$1C,$5B,$C7,$B7,$FD,$31,$19,$58,$7F,$51,$8C
	.dc.b	$36,$78,$4B,$CA,$FE,$78,$FF,$43,$DF,$FB,$9B,$7D,$2D,$CE,$A3,$AE
	.dc.b	$5F,$E6,$28,$79,$FA,$F3,$E5,$7F,$4D,$9C,$BF,$D3,$FE,$6D,$F3,$9E
	.dc.b	$2C,$9E,$FD,$3C,$B3,$AA,$E1,$DB,$8B,$91,$83,$A7,$6F,$09,$91,$83
	.dc.b	$C1,$C7,$DF,$DD,$91,$83,$C9,$C1,$E0,$99,$18,$3D,$1D,$BF,$31,$FF
	.dc.b	$3E,$64,$8C,$47,$FA,$0E,$31,$C5,$FE,$2E,$6C,$70,$F9,$C7,$CE,$65
	.dc.b	$B1,$98,$C1,$FC,$61,$7C,$2D,$23,$FC,$71,$FE,$3B,$E7,$7F,$80,$A3
	.dc.b	$FB,$C1,$F5,$8F,$F4,$B4,$7F,$B7,$7E,$73,$F0,$11,$FE,$9A,$9E,$1C
	.dc.b	$93,$9B,$FC,$09,$3C,$B3,$FE,$F7,$3D,$79,$A5,$7E,$1F,$E1,$ED,$F7
	.dc.b	$3F,$DD,$5E,$C7,$8C,$CE,$1F,$E0,$47,$CD,$EC,$4B,$EF,$16,$B9,$57
	.dc.b	$FA,$A8,$FF,$15,$16,$7F,$D6,$3D,$1F,$E6,$F7,$F5,$7C,$DF,$E2,$29
	.dc.b	$FF,$46,$C7,$EB,$B8,$F1,$39,$7F,$C4,$FF,$39,$ED,$9F,$F5,$DB,$70
	.dc.b	$0C,$1E,$FF,$0D,$F8,$EF,$30,$FF,$15,$61,$C7,$03,$FE,$1A,$8F,$F6
	.dc.b	$3F,$CE,$3C,$4F,$F1,$BA,$F0,$EC,$61,$93,$FE,$76,$9D,$36,$3F,$BC
	.dc.b	$C6,$2F,$FD,$9D,$F3,$8F,$8A,$FE,$FA,$DB,$FF,$3F,$B6,$77,$3F,$E0
	.dc.b	$2F,$9C,$7B,$B4,$97,$FE,$86,$8C,$B7,$E4,$B8,$11,$C1,$FE,$8C,$F9
	.dc.b	$CE,$D8,$7F,$B9,$5A,$FF,$BA,$1A,$FC,$E6,$AF,$9C,$78,$5F,$E1,$66
	.dc.b	$D7,$FA,$61,$B3,$FE,$ED,$6B,$FE,$F3,$FA,$4F,$3B,$FE,$F1,$6B,$9C
	.dc.b	$BF,$C4,$31,$88,$FF,$7D,$EE,$F8,$8F,$F1,$2B,$70,$FF,$53,$B6,$7F
	.dc.b	$DF,$8D,$7F,$0E,$7E,$23,$E7,$5F,$11,$FE,$2E,$6D,$7F,$AC,$1B,$3C
	.dc.b	$0E,$13,$FE,$34,$FA,$7F,$C8,$D7,$FA,$D9,$B3,$B0,$E1,$E4,$D7,$C7
	.dc.b	$F3,$CF,$93,$FF,$1F,$B6,$BF,$D8,$2D,$9E,$8D,$7C,$9F,$3C,$F9,$3F
	.dc.b	$F2,$3B,$6B,$FD,$92,$D9,$F2,$35,$F6,$7D,$2B,$D8,$FF,$27,$B6,$BF
	.dc.b	$DA,$2D,$9F,$63,$5F,$89,$F3,$EF,$63,$FC,$AE,$DA,$FF,$6C,$B6,$7E
	.dc.b	$23,$5F,$99,$F3,$EF,$67,$FC,$BE,$DA,$FF,$70,$B6,$7E,$63,$5F,$A9
	.dc.b	$F4,$4F,$4B,$FC,$CE,$DA,$FF,$74,$B6,$7E,$A3,$5F,$E5,$FF,$2D,$FE
	.dc.b	$6E,$F3,$8D,$25,$7F,$92,$7E,$4B,$FC,$E7,$F0,$46,$92,$3F,$DE,$FF
	.dc.b	$41,$F8,$8C,$3F,$10,$D7,$F2,$2D,$7F,$8C,$B5,$7C,$BF,$FB,$56,$7F
	.dc.b	$E0,$0B,$63,$F1,$8D,$7F,$28,$D7,$F3,$B6,$F9,$4F,$39,$2B,$FC,$C4
	.dc.b	$FF,$A3,$3E,$01,$F4,$14,$93,$FE,$C1,$7E,$26,$DF,$ED,$8D,$7E,$57
	.dc.b	$A1,$FE,$E7,$5F,$33,$99,$97,$FA,$1C,$FF,$B1,$C7,$28,$C5,$7D,$19
	.dc.b	$FD,$9B,$D2,$2E,$9C,$A7,$FA,$B4,$FC,$49,$FF,$71,$7C,$BF,$FA,$45
	.dc.b	$BF,$37,$FE,$EC,$8C,$57,$F5,$D5,$19,$F8,$DD,$36,$B8,$9D,$3B,$79
	.dc.b	$1C,$FD,$7B,$19,$26,$1D,$63,$2F,$9B,$CB,$F8,$C7,$C7,$F0,$6D,$7F
	.dc.b	$BE,$73,$FE,$52,$49,$FC,$81,$23,$FD,$FD,$F3,$5F,$ED,$1B,$FE,$BB
	.dc.b	$5F,$D0,$F8,$15,$FC,$6A,$B5,$4D,$5F,$37,$F0,$FC,$1F,$88,$DB,$F9
	.dc.b	$2D,$AF,$8C,$DC,$3A,$C7,$F6,$CB,$FF,$5A,$57,$D0,$F8,$FF,$9D,$F0
	.dc.b	$7F,$3D,$91,$93,$EA,$C6,$2B,$FC,$1D,$97,$F8,$0D,$FE,$EE,$D4,$FF
	.dc.b	$46,$C3,$C3,$18,$F6,$7C,$71,$89,$47,$FB,$91,$B7,$F8,$F6,$D7,$F3
	.dc.b	$16,$BF,$CE,$5F,$01,$FE,$10,$7C,$73,$DA,$7E,$67,$B3,$F8,$BE,$D9
	.dc.b	$1F,$90,$E7,$B7,$F8,$D9,$B8,$FA,$34,$FD,$1D,$FE,$5A,$7F,$29,$CF
	.dc.b	$F9,$5E,$7B,$3F,$9A,$6E,$1F,$97,$6E,$1C,$DF,$24,$7C,$8F,$89,$AF
	.dc.b	$C8,$6E,$BF,$CD,$6B,$E3,$46,$0F,$8C,$4B,$F8,$CD,$AF,$AB,$F3,$7B
	.dc.b	$9F,$C8,$BF,$AB,$F3,$24,$7E,$C3,$DA,$B5,$FA,$D9,$B8,$7C,$19,$B5
	.dc.b	$FD,$BB,$57,$C3,$FE,$D7,$A3,$F6,$CD,$9F,$DD,$B7,$1F,$DC,$37,$0F
	.dc.b	$C0,$78,$5E,$E3,$DF,$EC,$5B,$FD,$3C,$5B,$FC,$7C,$DA,$E3,$5A,$3F
	.dc.b	$5B,$19,$7F,$55,$56,$BF,$2B,$35,$FC,$AC,$5D,$BF,$C7,$57,$D2,$7F
	.dc.b	$3B,$39,$AD,$36,$FC,$13,$A6,$D7,$E6,$EB,$71,$83,$F7,$8E,$27,$81
	.dc.b	$EC,$8D,$3F,$AA,$AE,$CE,$AB,$C8,$49,$E6,$70,$8F,$5B,$6F,$D6,$FC
	.dc.b	$23,$15,$F3,$9A,$6E,$FC,$97,$D6,$FE,$3C,$E9,$F1,$3E,$06,$9E,$77
	.dc.b	$91,$B8,$E5,$1F,$C0,$A7,$AC,$F8,$E3,$E0,$1C,$8F,$49,$6F,$D2,$F5
	.dc.b	$31,$7E,$96,$CF,$81,$A7,$1A,$F9,$CF,$F0,$59,$FE,$18,$49,$FE,$18
	.dc.b	$D8,$FE,$18,$D7,$84,$7F,$38,$F0,$7F,$0C,$47,$E7,$36,$3D,$0D,$C3
	.dc.b	$E0,$E1,$FD,$DE,$B6,$4A,$DB,$67,$75,$C6,$3E,$34,$7F,$25,$8C,$4F
	.dc.b	$C3,$8C,$4F,$13,$67,$F4,$99,$DC,$FC,$C9,$DB,$E0,$23,$E0,$9E,$51
	.dc.b	$D5,$E4,$61,$E4,$E1,$E2,$F8,$1F,$E8,$33,$E0,$AF,$F1,$54,$FC,$93
	.dc.b	$93,$89,$FE,$93,$3B,$BF,$E3,$ED,$AE,$13,$D8,$F2,$1D,$9E,$73,$C1
	.dc.b	$A7,$0B,$F6,$4F,$13,$D2,$71,$31,$E1,$7F,$8C,$63,$FA,$AB,$67,$FC
	.dc.b	$5F,$F2,$DF,$CD,$23,$15,$FE,$D4,$6A,$68,$8E,$1F,$37,$FC,$F3,$C4
	.dc.b	$FF,$44,$E1,$FD,$89,$FF,$6B,$18,$C3,$FD,$32,$E3,$0E,$1E,$EF,$16
	.dc.b	$CF,$FB,$92,$B9,$7B,$BD,$4E,$DF,$03,$67,$47,$48,$FF,$22,$11,$FF
	.dc.b	$1D,$60,$8C,$5A,$FC,$DF,$01,$CE,$31,$7E,$68,$F6,$7C,$0C,$8E,$F3
	.dc.b	$FE,$F8,$C1,$F1,$39,$8E,$26,$6B,$18,$7C,$39,$B7,$F0,$BD,$9F,$91
	.dc.b	$F3,$7F,$25,$B1,$DB,$06,$3F,$EC,$EC,$3E,$65,$EE,$B5,$28,$FE,$DB
	.dc.b	$F3,$5F,$E2,$C3,$FC,$5B,$0F,$9B,$CF,$CC,$4B,$FF,$2F,$52,$3F,$C3
	.dc.b	$38,$FF,$B3,$A3,$13,$A9,$CC,$84,$F6,$F9,$BE,$91,$F0,$8F,$4B,$F8
	.dc.b	$5F,$A0,$97,$E8,$F9,$2B,$5B,$C9,$58,$D3,$F1,$55,$6F,$F1,$3F,$09
	.dc.b	$F8,$FF,$9F,$FC,$83,$63,$E1,$B7,$1E,$8D,$BF,$F1,$AF,$EA,$7F,$E5
	.dc.b	$2A,$F9,$CD,$AF,$42,$FA,$0F,$F0,$BB,$9A,$E3,$5F,$0A,$93,$C5,$FF
	.dc.b	$77,$35,$FF,$7B,$B5,$F9,$0F,$38,$CF,$D0,$76,$7B,$BE,$0F,$9D,$91
	.dc.b	$E6,$F5,$ED,$F9,$A6,$E1,$EC,$9B,$5E,$0F,$5F,$D3,$EB,$FA,$9B,$1F
	.dc.b	$01,$B8,$7C,$06,$DF,$C6,$F5,$FE,$09,$F0,$BD,$4B,$7F,$22,$3E,$16
	.dc.b	$86,$92,$3F,$D5,$51,$97,$E7,$FA,$BC,$B1,$F9,$EF,$F9,$13,$BD,$F3
	.dc.b	$5F,$35,$B5,$FE,$B3,$6D,$7E,$44,$8E,$1C,$0F,$A0,$FB,$53,$FE,$BB
	.dc.b	$6C,$68,$D7,$0F,$DB,$8C,$C7,$DA,$F5,$FD,$A4,$AF,$5B,$F8,$1F,$9F
	.dc.b	$70,$FE,$A6,$BA,$46,$3A,$FF,$D9,$4B,$8D,$9E,$2F,$85,$FA,$3E,$7D
	.dc.b	$8F,$FB,$46,$11,$87,$FD,$A8,$D7,$D4,$D7,$E0,$DF,$FB,$F5,$FF,$6D
	.dc.b	$35,$F8,$0D,$77,$E4,$77,$F2,$1F,$3D,$49,$3F,$EE,$38,$CB,$FC,$B3
	.dc.b	$83,$CD,$DB,$F6,$23,$04,$3E,$77,$1B,$DA,$F6,$36,$3F,$0B,$18,$3C
	.dc.b	$5E,$D4,$8F,$03,$5F,$B1,$4F,$9E,$F0,$92,$3F,$9F,$BD,$BD,$99,$1F
	.dc.b	$92,$6B,$F8,$1C,$9F,$33,$F0,$49,$1E,$F7,$DB,$F3,$D2,$3F,$A5,$B5
	.dc.b	$F6,$FB,$9E,$C2,$3F,$A0,$7E,$67,$B5,$23,$E5,$35,$F5,$FC,$37,$CA
	.dc.b	$23,$FD,$61,$F8,$FE,$EC,$8F,$D5,$B5,$F5,$7C,$6F,$D9,$23,$F5,$DF
	.dc.b	$AB,$F0,$D2,$3F,$B1,$B5,$FA,$BE,$FF,$D4,$23,$FB,$87,$EE,$EF,$23
	.dc.b	$FB,$5B,$5F,$BB,$EF,$7D,$C2,$3F,$C0,$1F,$CD,$7C,$EB,$E0,$7F,$71
	.dc.b	$E3,$7F,$B9,$BE,$79,$E6,$E3,$C3,$FD,$E4,$63,$1F,$EC,$58,$71,$6F
	.dc.b	$CE,$ED,$FE,$A5,$19,$7F,$A8,$37,$0F,$F7,$9A,$9E,$13,$FD,$E3,$D8
	.dc.b	$BF,$33,$FE,$BE,$8F,$EE,$0D,$7F,$AD,$7A,$FF,$E0,$97,$FB,$A2,$F6
	.dc.b	$B9,$B9,$8D,$D6,$27,$BF,$CE,$BE,$7B,$BF,$7C,$BF,$EF,$D3,$96,$FC
	.dc.b	$B7,$0A,$91,$83,$7E,$07,$49,$91,$83,$9F,$17,$B3,$A2,$46,$0E,$FB
	.dc.b	$79,$26,$46,$0F,$1F,$6F,$F8,$95,$FE,$4D,$97,$E9,$B8,$7F,$94,$42
	.dc.b	$30,$EA,$78,$3F,$CC,$A1,$F3,$CF,$9E,$E3,$18,$6C,$66,$7B,$B5,$EF
	.dc.b	$F1,$5F,$C8,$D3,$48,$FF,$45,$E6,$FF,$14,$3F,$CB,$9A,$F0,$1F,$3D
	.dc.b	$FE,$38,$7F,$A4,$C7,$0E,$33,$4E,$7E,$6E,$27,$FC,$38,$7F,$92,$E3
	.dc.b	$BF,$26,$75,$FF,$14,$B6,$3F,$AB,$73,$7F,$61,$AE,$33,$97,$FA,$D7
	.dc.b	$C0,$D7,$FC,$58,$D7,$F1,$D5,$EC,$8C,$7C,$0C,$8C,$1F,$01,$E5,$83
	.dc.b	$96,$9D,$71,$78,$D8,$BF,$F2,$75,$7E,$D5,$BC,$79,$CB,$FC,$75,$39
	.dc.b	$7F,$C7,$9F,$3D,$F3,$5B,$AF,$5F,$8B,$F5,$4E,$6B,$E4,$52,$3E,$43
	.dc.b	$CB,$7D,$DF,$07,$E1,$7C,$F7,$FA,$A6,$F8,$B7,$2F,$F0,$03,$FE,$CF
	.dc.b	$BF,$F2,$73,$97,$FC,$A8,$FF,$AC,$5F,$F0,$1B,$5E,$3E,$77,$F8,$E6
	.dc.b	$31,$1F,$E5,$BF,$23,$FC,$BE,$FF,$81,$5A,$FF,$83,$1A,$FF,$84,$5A
	.dc.b	$F1,$F9,$FF,$F2,$24,$1A,$4B,$FD,$4C,$7F,$6E,$6F,$9F,$C0,$8F,$EB
	.dc.b	$BE,$C3,$AD,$9A,$4F,$5B,$FC,$A4,$F7,$9E,$18,$60,$8F,$3B,$D9,$1C
	.dc.b	$1F,$BC,$F2,$9C,$1F,$E5,$D6,$C6,$63,$06,$7E,$32,$47,$16,$E5,$FE
	.dc.b	$2D,$A7,$C6,$F3,$25,$77,$7F,$CD,$5F,$33,$FA,$5A,$FF,$1C,$B5,$FF
	.dc.b	$1F,$35,$FF,$48,$35,$2F,$7B,$99,$2B,$F8,$BC,$65,$BE,$EF,$80,$8E
	.dc.b	$0F,$E3,$BF,$2C,$F4,$3F,$CF,$CD,$8C,$C6,$0C,$F9,$F2,$38,$B7,$1F
	.dc.b	$F2,$75,$3E,$4F,$99,$2B,$BB,$FE,$8B,$F4,$7F,$27,$46,$93,$9B,$5F
	.dc.b	$1D,$2F,$F9,$9C,$FE,$8E,$BF,$CC,$2D,$7F,$CC,$CB,$E3,$7B,$12,$78
	.dc.b	$BF,$E9,$BF,$94,$7C,$A5,$24,$FF,$40,$8C,$BF,$CD,$3C,$DF,$F5,$17
	.dc.b	$90,$69,$23,$FA,$3F,$C8,$FF,$13,$6F,$8B,$FE,$A9,$8C,$67,$FE,$73
	.dc.b	$6B,$FE,$AA,$E8,$FF,$58,$9F,$EB,$2E,$8B,$7F,$EB,$88,$FF,$48,$91
	.dc.b	$FF,$47,$B6,$7F,$D7,$7C,$FC,$1B,$3F,$EC,$0D,$3F,$D7,$A7,$F3,$1F
	.dc.b	$D5,$3F,$80,$F8,$1F,$EC,$68,$CC,$EC,$FF,$7A,$4E,$9F,$F5,$EC,$71
	.dc.b	$EC,$E5,$D9,$67,$6D,$C3,$FD,$C9,$7F,$C0,$5B,$5C,$2B,$AF,$A3,$E3
	.dc.b	$3C,$9E,$8D,$8C,$9F,$DC,$AF,$FC,$18,$D9,$FF,$6A,$46,$2F,$FD,$3C
	.dc.b	$DC,$BF,$99,$BF,$41,$B3,$F0,$77,$F8,$5E,$34,$C3,$DF,$E8,$DC,$7D
	.dc.b	$33,$70,$E7,$36,$BF,$D6,$98,$FF,$9B,$1F,$8F,$E7,$7C,$DF,$47,$E7
	.dc.b	$36,$32,$7C,$C8,$CA,$C5,$87,$F4,$AF,$94,$FC,$8C,$62,$7F,$D8,$51
	.dc.b	$89,$F6,$FC,$CF,$F6,$B8,$C4,$EB,$87,$FD,$F1,$1C,$AF,$FD,$8E,$D7
	.dc.b	$53,$FE,$1E,$8E,$31,$E3,$C9,$ED,$F5,$BF,$C4,$3F,$89,$F9,$69,$87
	.dc.b	$E1,$BC,$FB,$87,$E0,$26,$E7,$F4,$A6,$D7,$C8,$F1,$FF,$25,$F4,$9F
	.dc.b	$93,$6C,$7E,$41,$B8,$7E,$29,$B7,$F7,$3E,$5F,$F3,$47,$E6,$FC,$2B
	.dc.b	$7F,$98,$8F,$C2,$91,$FC,$53,$6F,$F2,$BE,$17,$06,$CF,$E8,$3C,$1F
	.dc.b	$D3,$1F,$A7,$F9,$7B,$19,$24,$FE,$7A,$32,$FD,$73,$EE,$7E,$AB,$F7
	.dc.b	$67,$EF,$65,$FE,$06,$8F,$F0,$44,$7F,$BF,$5D,$4F,$10,$E9,$E6,$7F
	.dc.b	$8A,$5C,$3A,$31,$E8,$FF,$8B,$7C,$16,$14,$C6,$37,$0F,$19,$F9,$1B
	.dc.b	$57,$08,$CC,$7E,$56,$30,$F9,$5F,$E5,$D1,$A8,$E3,$5D,$5D,$3E,$08
	.dc.b	$D4,$F2,$C7,$FC,$4E,$D8,$F1,$36,$BC,$17,$CE,$B7,$76,$F7,$36,$BC
	.dc.b	$D3,$5F,$5C,$5B,$E9,$1E,$6A,$F1,$CE,$6B,$4D,$8F,$3C,$7A,$67,$D7
	.dc.b	$18,$23,$F0,$E7,$83,$A3,$11,$AB,$F4,$B4,$F4,$96,$3D,$2D,$7D,$2D
	.dc.b	$BF,$49,$2B,$BD,$74,$AE,$E7,$85,$C1,$2B,$BC,$65,$F9,$F7,$F0,$1E
	.dc.b	$CD,$8D,$1E,$42,$9E,$42,$DF,$92,$3E,$35,$77,$9E,$0D,$78,$36,$78
	.dc.b	$57,$D2,$3C,$27,$BE,$FD,$AF,$98,$D5,$78,$2F,$FB,$61,$23,$FB,$8B
	.dc.b	$63,$FB,$73,$5F,$CF,$1F,$81,$33,$1F,$99,$C3,$F6,$38,$79,$DB,$1E
	.dc.b	$19,$FD,$A4,$61,$F6,$C7,$EA,$63,$11,$D6,$7F,$D0,$71,$8C,$3F,$22
	.dc.b	$DB,$FF,$51,$36,$FC,$57,$F9,$98,$C3,$E0,$30,$48,$FC,$34,$6A,$FB
	.dc.b	$CE,$4F,$AF,$19,$9F,$34,$66,$7D,$CD,$8F,$74,$72,$7A,$C6,$4F,$79
	.dc.b	$FD,$5E,$7A,$34,$D1,$F0,$63,$B9,$C8,$8F,$90,$DD,$FC,$46,$DF,$C4
	.dc.b	$6C,$FC,$43,$8D,$FA,$1A,$7A,$0E,$F1,$D5,$CB,$E8,$17,$E8,$69,$D8
	.dc.b	$F9,$D1,$C1,$F5,$1E,$D7,$F8,$F6,$3F,$EB,$CA,$FE,$75,$E8,$FF,$33
	.dc.b	$F5,$BF,$9A,$32,$B6,$E5,$A7,$2F,$5B,$F9,$D7,$AD,$FC,$ED,$B1,$93
	.dc.b	$BC,$65,$A7,$E5,$DF,$C4,$9F,$8A,$8B,$19,$3F,$3F,$8F,$D7,$9F,$F6
	.dc.b	$AB,$F9,$76,$B9,$8D,$BF,$1A,$3F,$2B,$ED,$F4,$8F,$D0,$BF,$D4,$49
	.dc.b	$3F,$13,$E6,$2C,$FF,$59,$CD,$58,$CF,$FB,$05,$5B,$FE,$C1,$F3,$1F
	.dc.b	$D6,$FE,$63,$FA,$E3,$63,$E7,$B7,$1E,$8D,$BF,$D6,$FC,$47,$EA,$8F
	.dc.b	$D5,$FC,$BD,$BF,$DE,$47,$BA,$FF,$02,$DE,$03,$49,$3F,$C0,$FE,$23
	.dc.b	$95,$7F,$00,$7F,$B8,$5F,$F7,$1A,$B5,$F8,$46,$D7,$C1,$6E,$1F,$DF
	.dc.b	$DF,$EF,$17,$2B,$F1,$7F,$0F,$6F,$E4,$4D,$BF,$F0,$34,$DA,$FE,$1D
	.dc.b	$E9,$7F,$08,$3F,$84,$FA,$96,$3F,$8D,$47,$B4,$8F,$DC,$CF,$FD,$68
	.dc.b	$D8,$FE,$23,$F2,$FC,$72,$FE,$3A,$FF,$83,$7E,$23,$FC,$1D,$EB,$DB
	.dc.b	$FD,$0B,$6B,$EF,$37,$3F,$F0,$E3,$FE,$15,$CA,$57,$EA,$31,$B7,$F8
	.dc.b	$59,$B7,$FE,$23,$9B,$5F,$E1,$3F,$90,$7C,$8A,$3F,$C1,$D1,$86,$99
	.dc.b	$C3,$FC,$51,$19,$6C,$FE,$06,$31,$1F,$D0,$9A,$F7,$8D,$1C,$EF,$A4
	.dc.b	$25,$F2,$B9,$BE,$07,$91,$C4,$FF,$9E,$FE,$99,$E2,$FE,$45,$B5,$FA
	.dc.b	$E6,$CF,$ED,$9A,$FF,$56,$FA,$57,$C6,$7E,$65,$B5,$FB,$66,$CF,$EF
	.dc.b	$9A,$FF,$5E,$F9,$47,$CA,$3E,$13,$F7,$B1,$86,$FC,$97,$F2,$2A,$E1
	.dc.b	$5B,$7F,$84,$F7,$39,$7F,$82,$C7,$F9,$51,$B1,$98,$FE,$12,$D7,$F8
	.dc.b	$8F,$E0,$CF,$98,$E2,$D8,$FE,$1E,$D7,$71,$AF,$7A,$46,$9A,$F9,$36
	.dc.b	$F0,$91,$F0,$BE,$4F,$99,$91,$E4,$6B,$EC,$F9,$AF,$49,$1F,$4B,$EE
	.dc.b	$F3,$3F,$84,$41,$A4,$E1,$D7,$3C,$9F,$E5,$33,$F9,$86,$C7,$3E,$59
	.dc.b	$1A,$6B,$FC,$DA,$5C,$1E,$12,$47,$47,$C9,$C1,$23,$BB,$5E,$5F,$35
	.dc.b	$E9,$23,$E4,$7D,$9B,$48,$F3,$B5,$F1,$7C,$D7,$C1,$23,$EC,$7E,$27
	.dc.b	$C5,$48,$F7,$B5,$F5,$7C,$4F,$C9,$23,$F1,$1F,$99,$F3,$52,$3E,$3B
	.dc.b	$5F,$85,$F2,$9F,$44,$8F,$CC,$7E,$A7,$C8,$C8,$F9,$ED,$7E,$57,$C7
	.dc.b	$FD,$92,$3F,$51,$FC,$CF,$CA,$BE,$0B,$FD,$41,$E0,$FE,$FB,$E5,$BF
	.dc.b	$5F,$7F,$E9,$B6,$BF,$EB,$3A,$E0,$FF,$A7,$D7,$C1,$F2,$E6,$FF,$55
	.dc.b	$46,$A7,$86,$19,$7F,$DF,$D1,$C1,$FF,$54,$7A,$0F,$4F,$50,$F7,$1D
	.dc.b	$B8,$6C,$70,$9D,$D3,$D5,$B1,$27,$AC,$74,$F9,$67,$C3,$73,$E2,$95
	.dc.b	$FE,$B9,$9C,$B7,$E1,$F8,$54,$8C,$1B,$F8,$7E,$93,$23,$07,$3D,$BD
	.dc.b	$1F,$19,$23,$07,$7E,$1F,$24,$C8,$C1,$E3,$F8,$77,$27,$F8,$53,$34
	.dc.b	$62,$7F,$C3,$B1,$89,$D4,$E6,$7F,$CF,$F2,$F8,$87,$CB,$73,$6C,$F2
	.dc.b	$6B,$C1,$AF,$F8,$AB,$E3,$BF,$90,$A6,$91,$FE,$DD,$CD,$FE,$E3,$7F
	.dc.b	$DE,$A0,$E0,$3E,$5B,$FC,$4A,$FF,$A2,$63,$87,$1F,$F4,$A8,$E0,$D7
	.dc.b	$9B,$78,$E4,$FF,$BC,$79,$5D,$D8,$BF,$93,$3F,$EA,$0F,$7F,$91,$9D
	.dc.b	$FF,$C8,$6F,$5F,$7E,$4F,$A6,$2D,$72,$3D,$54,$E5,$F3,$36,$BC,$5D
	.dc.b	$B8,$9C,$F1,$C9,$E2,$7E,$C2,$31,$7D,$47,$CB,$7F,$9D,$A7,$71,$F2
	.dc.b	$63,$0E,$F4,$7C,$87,$C7,$9C,$E5,$FD,$E1,$FF,$3B,$5F,$D1,$BE,$BB
	.dc.b	$EF,$DA,$E0,$FC,$FF,$96,$7C,$3F,$F9,$1D,$B7,$FE,$64,$6D,$FF,$9B
	.dc.b	$1A,$FF,$B3,$1C,$E3,$FE,$08,$5F,$0D,$D6,$FB,$3F,$77,$C3,$7C,$7F
	.dc.b	$F9,$55,$B8,$FF,$9C,$9B,$8F,$F9,$E9,$AF,$FB,$69,$7C,$87,$12,$4F
	.dc.b	$17,$FC,$C1,$F4,$AF,$85,$A4,$9F,$F3,$FC,$65,$FF,$40,$FC,$BF,$6C
	.dc.b	$3E,$43,$5F,$CD,$35,$34,$F1,$C7,$2B,$FF,$38,$46,$27,$FD,$26,$D7
	.dc.b	$FD,$E7,$F4,$9F,$89,$7F,$D2,$ED,$73,$C9,$FE,$79,$8C,$47,$DB,$F8
	.dc.b	$DF,$E4,$B8,$7F,$A6,$1B,$1F,$EA,$06,$BF,$F0,$01,$AF,$1F,$94,$FF
	.dc.b	$42,$41,$A4,$9F,$F5,$1F,$D1,$BE,$4B,$FD,$14,$D8,$CC,$60,$FF,$47
	.dc.b	$7C,$CB,$D1,$A4,$9F,$F5,$6C,$65,$BE,$8F,$02,$38,$3F,$D6,$7F,$44
	.dc.b	$F4,$BF,$D3,$4D,$8C,$C6,$0F,$F4,$F7,$CD,$3E,$22,$92,$7F,$D7,$B1
	.dc.b	$96,$FC,$47,$02,$38,$3F,$D8,$7F,$43,$BE,$0F,$F6,$5B,$5F,$D6,$07
	.dc.b	$0F,$EB,$DA,$FF,$21,$F9,$B7,$CA,$FF,$AC,$DB,$5F,$ED,$06,$CF,$FB
	.dc.b	$61,$AF,$67,$F5,$71,$BE,$BE,$44,$7F,$DB,$3F,$3B,$C2,$BF,$80,$35
	.dc.b	$FE,$64,$D4,$DF,$31,$E9,$7F,$B1,$DB,$5F,$EE,$46,$D7,$F4,$39,$FF
	.dc.b	$66,$7C,$B7,$E5,$23,$F9,$F3,$53,$49,$E0,$70,$25,$FF,$BC,$23,$F8
	.dc.b	$5C,$FF,$17,$F9,$AF,$C7,$3F,$C9,$5B,$E1,$FF,$A5,$3E,$69,$F3,$7F
	.dc.b	$D1,$E7,$38,$71,$E9,$FF,$70,$47,$2C,$3E,$BB,$5D,$4F,$FC,$00,$8E
	.dc.b	$31,$FD,$5E,$DF,$37,$D6,$FF,$AA,$46,$2F,$F9,$AD,$23,$FD,$37,$8F
	.dc.b	$97,$E4,$37,$E6,$70,$65,$F0,$AD,$BF,$CD,$5F,$F1,$03,$98,$EF,$DF
	.dc.b	$E0,$F9,$BB,$3C,$5B,$8F,$F4,$9B,$46,$31,$F8,$CF,$A3,$E1,$65,$F8
	.dc.b	$78,$2E,$1C,$67,$FD,$7E,$F8,$DA,$FC,$36,$BE,$F6,$E1,$FE,$85,$AF
	.dc.b	$67,$37,$4C,$7E,$B3,$F2,$39,$64,$FC,$1F,$9B,$B8,$FA,$66,$DF,$DD
	.dc.b	$9B,$5F,$D6,$7E,$37,$FC,$14,$FD,$07,$FA,$FB,$9B,$FF,$04,$46,$1F
	.dc.b	$F0,$83,$C1,$FE,$DC,$E5,$39,$AF,$F6,$FB,$9E,$45,$FF,$84,$DB,$7F
	.dc.b	$80,$7F,$67,$7F,$DE,$5C,$B6,$7F,$4A,$49,$FC,$F3,$63,$FC,$01,$E7
	.dc.b	$78,$67,$FB,$B5,$72,$C3,$FD,$E3,$39,$AF,$5F,$9D,$EE,$3D,$FE,$B5
	.dc.b	$C3,$E2,$45,$9F,$7C,$DB,$F6,$FA,$9F,$1B,$D5,$F8,$ED,$C3,$E6,$36
	.dc.b	$FE,$3B,$6B,$C7,$5F,$E3,$5C,$4D,$9C,$32,$FF,$92,$DF,$F9,$06,$B7
	.dc.b	$18,$84,$61,$DB,$5F,$E5,$AE,$6F,$E1,$FC,$07,$59,$F9,$2F,$E0,$09
	.dc.b	$5F,$85,$F9,$BB,$3F,$80,$9B,$5F,$7A,$6D,$7E,$37,$D0,$FC,$4F,$CD
	.dc.b	$FE,$29,$B1,$F8,$F6,$E1,$F8,$A6,$DF,$D6,$F9,$AF,$88,$7C,$5F,$9B
	.dc.b	$B7,$FC,$1B,$34,$E3,$0F,$C0,$D5,$BE,$57,$BA,$E4,$D2,$57,$E2,$23
	.dc.b	$2D,$8C,$CF,$28,$EB,$18,$3F,$C8,$57,$E2,$71,$1F,$56,$FC,$51,$AB
	.dc.b	$B3,$98,$DB,$F6,$23,$2F,$DA,$9B,$5D,$63,$C8,$78,$8B,$7E,$52,$5F
	.dc.b	$92,$B8,$DF,$26,$C7,$82,$D1,$8B,$E7,$3C,$4D,$57,$16,$9C,$67,$85
	.dc.b	$F8,$2B,$D8,$72,$3F,$2D,$5C,$5C,$1E,$FA,$E4,$74,$73,$7E,$48,$C5
	.dc.b	$75,$9F,$89,$18,$3E,$61,$F8,$3B,$F1,$B8,$8F,$8D,$3C,$6B,$E5,$36
	.dc.b	$B7,$3E,$6A,$E4,$78,$DF,$D3,$DF,$F3,$6B,$FA,$51,$89,$FD,$84,$62
	.dc.b	$3C,$06,$4F,$D8,$CF,$13,$F7,$DF,$37,$F7,$09,$5F,$4D,$B3,$F0,$9B
	.dc.b	$87,$A5,$B3,$ED,$8F,$E3,$71,$89,$F5,$1F,$C8,$9C,$3F,$3C,$C4,$F7
	.dc.b	$23,$DD,$B3,$DD,$A7,$73,$F3,$F3,$D2,$FE,$5C,$60,$DD,$78,$AB,$CB
	.dc.b	$18,$8F,$14,$61,$F0,$11,$F0,$36,$3C,$0F,$F1,$FA,$F0,$5F,$CA,$9E
	.dc.b	$97,$E2,$AD,$46,$27,$51,$CC,$F2,$47,$9D,$C1,$E4,$71,$7E,$22,$3E
	.dc.b	$28,$FE,$5B,$7E,$28,$FE,$24,$FC,$EE,$3F,$CC,$4F,$F6,$4B,$FE,$D3
	.dc.b	$8F,$F6,$A6,$DF,$F6,$F6,$CE,$E3,$FB,$95,$FF,$70,$3F,$B8,$FC,$DD
	.dc.b	$8C,$92,$7F,$B8,$C6,$5F,$EC,$3E,$3F,$F4,$B3,$FA,$67,$93,$6B,$F9
	.dc.b	$44,$7E,$6C,$8F,$E6,$9B,$3F,$DA,$BE,$53,$83,$67,$FC,$0D,$F3,$7F
	.dc.b	$E0,$4F,$9B,$FF,$02,$B6,$32,$7F,$82,$E3,$2D,$3F,$C1,$D8,$7F,$83
	.dc.b	$0F,$F0,$6E,$16,$32,$4B,$FF,$06,$C6,$5F,$F7,$E7,$CD,$BE,$4D,$5F
	.dc.b	$E4,$88,$C3,$4C,$CF,$09,$BC,$18,$85,$70,$7F,$B6,$B5,$FF,$49,$B5
	.dc.b	$FF,$80,$0D,$47,$29,$EB,$5A,$F9,$3B,$3A,$AE,$D7,$E1,$53,$A9,$78
	.dc.b	$1F,$E8,$4A,$F3,$2F,$37,$4E,$70,$F0,$D7,$09,$CF,$97,$A6,$CF,$F9
	.dc.b	$B2,$B9,$37,$E7,$79,$9D,$3E,$06,$C6,$49,$83,$1F,$1C,$EA,$31,$3E
	.dc.b	$2B,$FF,$73,$D7,$A3,$E3,$BC,$FE,$CF,$A1,$95,$FE,$17,$AF,$F7,$4D
	.dc.b	$FC,$A6,$BF,$0A,$7F,$CC,$B1,$D2,$7B,$79,$9C,$AB,$C2,$71,$7E,$57
	.dc.b	$EA,$CF,$9A,$BF,$05,$86,$DF,$FA,$3A,$74,$D9,$D3,$4D,$7C,$07,$43
	.dc.b	$AF,$C7,$D8,$C9,$31,$F1,$8B,$4E,$E3,$3F,$3F,$FE,$EF,$6F,$B5,$E8
	.dc.b	$F9,$0F,$C1,$4E,$78,$3F,$2C,$EE,$BF,$DC,$AD,$7F,$02,$D7,$F2,$6D
	.dc.b	$4F,$3C,$7F,$0F,$8B,$E0,$BB,$E3,$F2,$1F,$C9,$7C,$EB,$E6,$FF,$D2
	.dc.b	$AD,$BF,$F2,$73,$71,$FF,$29,$B5,$FF,$7C,$FC,$EB,$E6,$7E,$94,$7F
	.dc.b	$96,$23,$1C,$3F,$E5,$75,$F3,$5F,$E5,$89,$D5,$70,$7F,$3A,$3E,$81
	.dc.b	$A7,$F9,$7E,$30,$D8,$CC,$7F,$98,$DA,$FF,$9A,$3E,$79,$F3,$7C,$5B
	.dc.b	$1F,$E6,$96,$BB,$8D,$74,$48,$D3,$5F,$F3,$9F,$E0,$EF,$95,$F2,$36
	.dc.b	$3F,$CE,$8D,$77,$1E,$9F,$31,$F3,$F4,$93,$FE,$7B,$8C,$B7,$E7,$F8
	.dc.b	$11,$C1,$FE,$81,$E5,$7C,$A7,$C5,$6C,$66,$30,$7C,$9E,$EF,$E5,$51
	.dc.b	$FB,$3F,$1B,$04,$7F,$6D,$E2,$EC,$8E,$0F,$07,$CD,$F7,$23,$83,$F7
	.dc.b	$DC,$7D,$C8,$E0,$FE,$05,$F3,$FE,$32,$38,$3D,$9F,$39,$EB,$23,$83
	.dc.b	$F8,$57,$CC,$FA,$C8,$E0,$FE,$1D,$F2,$3E,$F2,$38,$3E,$47,$0F,$C7
	.dc.b	$23,$83,$F8,$B7,$3F,$C7,$23,$83,$F8,$D7,$CE,$FC,$B2,$38,$3E,$A7
	.dc.b	$C5,$FD,$32,$38,$3F,$91,$7C,$BF,$D3,$23,$83,$EE,$7C,$5F,$DB,$23
	.dc.b	$83,$FD,$81,$F3,$DF,$79,$FE,$23,$F3,$FF,$B9,$7F,$AA,$F8,$DF,$D7
	.dc.b	$E3,$11,$FC,$5A,$78,$4F,$F6,$35,$C3,$FD,$32,$31,$CB,$7C,$3F,$E9
	.dc.b	$F5,$C1,$B5,$9B,$FE,$C7,$EF,$BE,$0F,$10,$F8,$D7,$7E,$1B,$1E,$6C
	.dc.b	$3F,$AD,$DF,$18,$C1,$D3,$E1,$39,$C6,$63,$F9,$7B,$74,$FE,$F8,$A7
	.dc.b	$C3,$F3,$FF,$CC,$AB,$99,$E3,$E2,$F1,$1A,$F3,$77,$E6,$D9,$DD,$71
	.dc.b	$7F,$9E,$A8,$EB,$E6,$70,$9D,$C6,$1E,$3C,$3E,$FF,$56,$46,$0E,$BC
	.dc.b	$3E,$69,$91,$83,$C3,$D7,$FD,$C5,$FF,$80,$38,$A3,$11,$FC,$9A,$31
	.dc.b	$B7,$F2,$88,$FF,$0A,$C6,$1E,$3D,$2F,$9F,$D4,$27,$F8,$D4,$61,$B1
	.dc.b	$C2,$3F,$98,$2F,$91,$A4,$7F,$9A,$3F,$CD,$7E,$0B,$FC,$39,$06,$BC
	.dc.b	$47,$CF,$B8,$FF,$9F,$CF,$0E,$79,$CB,$F6,$2F,$93,$63,$AF,$2C,$8F
	.dc.b	$BB,$CC,$EF,$C5,$FF,$74,$1F,$11,$E4,$33,$C1,$FE,$43,$8C,$DF,$7F
	.dc.b	$88,$F4,$9B,$70,$7E,$2A,$7F,$B5,$37,$9F,$FC,$75,$8B,$E7,$FF,$C7
	.dc.b	$95,$AC,$3F,$C0,$AF,$F3,$6B,$E5,$CB,$AC,$36,$FF,$90,$FE,$7F,$DF
	.dc.b	$7F,$E0,$C6,$DF,$F9,$67,$17,$B7,$BE,$0E,$1C,$13,$3F,$CF,$E1,$C1
	.dc.b	$FF,$23,$BF,$B5,$7F,$47,$F3,$EF,$91,$FF,$29,$2F,$8F,$B1,$86,$4F
	.dc.b	$E2,$27,$4D,$8F,$95,$A7,$F9,$92,$72,$FE,$4F,$C4,$7B,$7A,$84,$62
	.dc.b	$FF,$26,$D8,$E1,$3B,$F9,$F7,$C6,$52,$5F,$E6,$A3,$2D,$F4,$78,$11
	.dc.b	$C1,$FA,$0F,$9F,$ED,$87,$F8,$39,$AF,$F9,$B1,$AF,$F9,$F9,$AB,$E7
	.dc.b	$FE,$E3,$FE,$16,$6B,$FE,$74,$6B,$FE,$86,$6A,$FA,$6F,$3B,$FB,$76
	.dc.b	$BF,$E2,$06,$BF,$E7,$E6,$BF,$E9,$0F,$97,$7C,$27,$CB,$6E,$1F,$BC
	.dc.b	$6C,$FF,$03,$6B,$FE,$2C,$FA,$07,$C2,$FD,$36,$D7,$F0,$46,$CF,$F0
	.dc.b	$B6,$BF,$E3,$4F,$A6,$FC,$AB,$79,$BE,$E4,$62,$3F,$D2,$AD,$7F,$D4
	.dc.b	$FF,$42,$FA,$1F,$C1,$37,$0F,$E2,$6D,$9F,$E3,$4D,$7F,$C8,$7F,$43
	.dc.b	$FE,$5D,$27,$AB,$F8,$B7,$FC,$C1,$F4,$CF,$57,$F1,$CD,$C3,$F8,$FB
	.dc.b	$6F,$F9,$33,$5F,$F2,$7F,$D1,$3E,$3F,$F2,$CD,$AF,$E4,$ED,$9F,$E5
	.dc.b	$CD,$7F,$CA,$FF,$44,$F9,$1F,$CF,$36,$BF,$97,$B6,$7F,$9B,$35,$FF
	.dc.b	$2F,$FD,$37,$FA,$D9,$FE,$74,$D7,$3F,$23,$FA,$98,$C4,$7F,$B0,$A3
	.dc.b	$F5,$EF,$F9,$A1,$AE,$7B,$FF,$61,$18,$8F,$F6,$CF,$D1,$BE,$7F,$F6
	.dc.b	$8D,$C7,$FA,$2B,$67,$FA,$63,$5F,$F3,$AE,$5F,$05,$FF,$69,$B5,$FF
	.dc.b	$72,$B5,$F8,$9A,$7F,$01,$F9,$EF,$D1,$93,$CA,$C1,$9F,$12,$63,$FD
	.dc.b	$5A,$7F,$8B,$36,$3F,$85,$7D,$37,$FA,$EB,$0F,$EC,$0D,$7F,$D1,$2D
	.dc.b	$7F,$DC,$ED,$7F,$DF,$34,$FA,$0F,$F7,$93,$6F,$FC,$0E,$FF,$1A,$F2
	.dc.b	$D1,$FE,$F0,$8D,$57,$FB,$01,$AF,$FA,$61,$AE,$EB,$E0,$12,$7F,$C2
	.dc.b	$7D,$49,$F0,$CF,$67,$FD,$97,$19,$9F,$F0,$CF,$C4,$FF,$39,$9F,$F6
	.dc.b	$7B,$70,$E1,$08,$FF,$11,$4F,$0F,$2B,$A5,$6A,$78,$4F,$86,$B8,$56
	.dc.b	$A3,$13,$CA,$7B,$CE,$5F,$44,$7F,$68,$7C,$7E,$7E,$5F,$17,$A1,$65
	.dc.b	$36,$30,$DB,$F6,$BC,$71,$DC,$7C,$68,$C4,$E5,$91,$FE,$1A,$9D,$DF
	.dc.b	$F8,$85,$AF,$F8,$32,$7F,$A8,$46,$B1,$F8,$EF,$F8,$46,$BB,$1F,$53
	.dc.b	$D3,$F4,$11,$FA,$3F,$41,$70,$FA,$33,$6F,$E6,$4D,$AF,$93,$E5,$7C
	.dc.b	$1F,$A0,$F8,$4D,$8F,$B6,$DC,$3E,$DB,$6F,$CB,$F3,$FF,$0C,$F8,$9F
	.dc.b	$41,$6F,$F1,$31,$EC,$23,$F0,$DB,$7F,$8A,$FA,$0E,$0D,$9F,$C5,$D7
	.dc.b	$D5,$AF,$AC,$D8,$C9,$F9,$08,$CB,$4F,$C9,$D7,$DD,$3E,$F7,$D0,$58
	.dc.b	$C9,$2F,$F2,$F1,$97,$F3,$BC,$FF,$7C,$FC,$07,$45,$AF,$D2,$C7,$D8
	.dc.b	$23,$F7,$DB,$3F,$A6,$FA,$0E,$0D,$9F,$D3,$FD,$07,$E4,$7E,$83,$F2
	.dc.b	$4D,$8C,$9F,$AC,$8C,$B4,$FD,$7E,$1F,$99,$3F,$35,$85,$8C,$92,$FF
	.dc.b	$67,$19,$5A,$FE,$D3,$E8,$1F,$23,$FA,$56,$DF,$FA,$51,$FE,$1A,$D8
	.dc.b	$D1,$1D,$16,$BA,$BF,$B9,$9E,$71,$83,$F8,$63,$8A,$F1,$4F,$17,$C9
	.dc.b	$56,$39,$16,$B8,$C5,$55,$E6,$7B,$18,$8D,$47,$79,$E6,$7A,$C9,$3C
	.dc.b	$23,$4F,$9A,$38,$CF,$26,$9C,$2B,$D2,$D3,$E0,$46,$DC,$4F,$A1,$AF
	.dc.b	$A1,$E6,$F8,$63,$E9,$46,$27,$E1,$BF,$38,$E3,$5E,$6A,$F7,$46,$5E
	.dc.b	$4E,$AB,$9B,$4E,$27,$53,$CE,$71,$AE,$F3,$C8,$8F,$B9,$B1,$C1,$A7
	.dc.b	$C7,$7B,$38,$3D,$CE,$23,$A9,$CC,$AF,$36,$C7,$33,$95,$7B,$C9,$3E
	.dc.b	$98,$F9,$87,$D9,$7E,$5B,$FD,$24,$F8,$A7,$FB,$46,$3F,$26,$E5,$AF
	.dc.b	$AC,$C5,$7F,$A0,$E3,$31,$D8,$CC,$FE,$97,$E4,$7F,$1A,$47,$EA,$B6
	.dc.b	$3D,$2D,$BF,$6C,$7E,$62,$31,$3F,$04,$FD,$43,$88,$F6,$CF,$ED,$E3
	.dc.b	$11,$EC,$8C,$4F,$80,$91,$E0,$6C,$78,$29,$5C,$9F,$98,$DB,$FE,$04
	.dc.b	$DA,$F7,$1C,$8F,$35,$FE,$EA,$34,$FE,$A5,$B7,$E4,$8F,$11,$1F,$14
	.dc.b	$6A,$7D,$E7,$59,$F7,$B8,$78,$3E,$07,$E4,$3B,$7E,$13,$B9,$EA,$E6
	.dc.b	$FD,$AD,$BF,$6B,$C7,$E4,$9F,$11,$FE,$05,$7F,$CB,$D7,$FE,$18,$73
	.dc.b	$3F,$E2,$38,$C7,$87,$BA,$E0,$D9,$D1,$FE,$19,$AE,$6D,$FA,$0E,$C7
	.dc.b	$7F,$76,$C7,$0C,$FF,$C6,$14,$AC,$C7,$F1,$C8,$FF,$76,$CE,$E1,$1B
	.dc.b	$8C,$2F,$8D,$F6,$1E,$DF,$56,$CE,$A3,$FD,$C1,$3B,$8F,$D4,$57,$8A
	.dc.b	$72,$48,$EB,$EA,$72,$9F,$23,$FE,$2D,$F9,$4F,$F3,$06,$BF,$E4,$05
	.dc.b	$F4,$19,$F9,$DD,$37,$0F,$54,$E9,$B3,$A7,$3E,$F7,$53,$B7,$C7,$D8
	.dc.b	$C9,$32,$FE,$83,$69,$D4,$7F,$44,$FA,$07,$AD,$FE,$43,$6D,$F8,$BE
	.dc.b	$43,$13,$23,$E1,$D2,$B6,$FC,$66,$BF,$C3,$DA,$FF,$96,$5D,$E9,$F0
	.dc.b	$FC,$EE,$93,$E6,$7E,$47,$D1,$FC,$9F,$86,$B8,$7C,$46,$D7,$B1,$B7
	.dc.b	$F3,$5F,$A1,$EF,$4A,$F9,$7F,$47,$6F,$E1,$CD,$BF,$A9,$36,$BF,$CD
	.dc.b	$DB,$7E,$29,$FA,$7C,$9F,$6F,$D2,$FB,$8C,$8F,$87,$3B,$C7,$F0,$ED
	.dc.b	$7F,$98,$35,$EB,$3F,$85,$F5,$39,$47,$D6,$7F,$11,$EA,$FE,$27,$D2
	.dc.b	$B8,$7E,$11,$B5,$F7,$1B,$9F,$E3,$5F,$F3,$E7,$EA,$FF,$E9,$39,$FF
	.dc.b	$00,$B6,$7F,$D2,$2B,$CF,$DC,$CB,$FD,$54,$ED,$FF,$48,$3B,$73,$1D
	.dc.b	$BD,$39,$1C,$3E,$32,$CF,$82,$7F,$D3,$EF,$26,$BF,$EA,$BA,$FF,$53
	.dc.b	$3F,$D5,$5A,$EE,$BC,$7E,$9F,$4C,$7D,$2F,$FA,$67,$E8,$5F,$36,$EC
	.dc.b	$FC,$AC,$61,$BF,$43,$EF,$C7,$F7,$73,$97,$F7,$BF,$43,$FA,$27,$F3
	.dc.b	$AD,$73,$D7,$FA,$05,$19,$7F,$B6,$53,$97,$4D,$8F,$D3,$B5,$FE,$16
	.dc.b	$D7,$FB,$A7,$D0,$FF,$15,$7F,$5A,$D7,$F8,$8B,$5D,$F8,$5E,$3F,$A0
	.dc.b	$7C,$BD,$24,$FE,$DA,$32,$FE,$E3,$EF,$5F,$49,$EB,$6C,$66,$30,$7B
	.dc.b	$FD,$D7,$C7,$52,$4F,$F0,$28,$CB,$FA,$2F,$93,$F8,$DD,$C6,$92,$3F
	.dc.b	$84,$F3,$FF,$B3,$9B,$EF,$FC,$D8,$C4,$7F,$30,$6B,$FE,$19,$EF,$FE
	.dc.b	$EE,$FF,$12,$6B,$FC,$D9,$AE,$FD,$2F,$1F,$CD,$C8,$FD,$BB,$5E,$DF
	.dc.b	$01,$E1,$23,$E6,$7F,$01,$4F,$A1,$A4,$9F,$13,$F0,$3E,$72,$47,$F0
	.dc.b	$46,$BE,$BF,$82,$F7,$11,$F8,$6F,$6F,$A1,$91,$ED,$6B,$FC,$33,$93
	.dc.b	$E6,$91,$FE,$22,$FC,$9F,$82,$91,$F3,$9A,$FC,$5F,$9E,$F5,$91,$F9
	.dc.b	$4F,$D8,$FA,$19,$1F,$C7,$1A,$FD,$3F,$85,$FA,$C4,$7E,$DB,$F8,$0F
	.dc.b	$8D,$91,$FC,$91,$AF,$DB,$F8,$6F,$BC,$47,$F0,$6F,$EB,$7E,$83,$F1
	.dc.b	$11,$FD,$EF,$E8,$7F,$2A,$FF,$A6,$9B,$E3,$BC,$84,$CA,$E2,$FF,$A8
	.dc.b	$17,$26,$A3,$1D,$16,$F9,$47,$18,$C5,$EE,$F8,$CF,$F8,$37,$95,$E4
	.dc.b	$EA,$3C,$3E,$8F,$ED,$F1,$E7,$08,$C6,$1E,$1B,$E3,$3C,$FC,$BE,$4A
	.dc.b	$32,$FF,$AE,$1B,$7D,$14,$F0,$FA,$1F,$EF,$95,$FE,$21,$C1,$C5,$D1
	.dc.b	$B1,$FE,$25,$9D,$BF,$EC,$45,$EB,$73,$F5,$E4,$EE,$3B,$7A,$DD,$D9
	.dc.b	$1C,$E3,$C5,$BF,$7F,$7E,$4F,$78,$F3,$70,$EA,$30,$78,$DF,$57,$5F
	.dc.b	$FB,$CA,$3F,$C5,$F9,$23,$11,$FE,$99,$8C,$6F,$FE,$12,$6C,$70,$E9
	.dc.b	$7D,$0E,$5B,$19,$8C,$1F,$CE,$D7,$D0,$D2,$3F,$CF,$DF,$E8,$1E,$9F
	.dc.b	$E1,$B0,$FE,$B6,$DE,$4F,$F2,$2C,$FE,$4B,$E8,$7E,$D5,$FE,$3E,$B4
	.dc.b	$FE,$47,$9A,$BF,$98,$6B,$FD,$B4,$FF,$7B,$9F,$8E,$BF,$F2,$E3,$F8
	.dc.b	$06,$F2,$F2,$E4,$7D,$0F,$28,$E7,$EA,$CA,$D9,$C9,$EF,$E3,$48,$C1
	.dc.b	$D9,$F2,$EA,$E5,$A7,$7F,$A1,$B5,$FE,$6F,$9F,$E7,$CD,$E5,$CE,$89
	.dc.b	$FD,$19,$A7,$13,$FB,$3F,$A1,$F8,$13,$FA,$86,$C7,$ED,$F5,$FF,$3E
	.dc.b	$D7,$03,$F7,$F3,$A6,$ED,$F3,$23,$E4,$46,$C3,$B6,$C6,$FC,$09,$7F
	.dc.b	$E8,$9F,$A1,$7B,$38,$BE,$58,$2F,$B7,$83,$9C,$FD,$0F,$DC,$FA,$1F
	.dc.b	$F3,$76,$5F,$E9,$36,$E1,$E3,$7F,$5F,$19,$EE,$F4,$7C,$7C,$8C,$1F
	.dc.b	$E9,$BC,$91,$ED,$F8,$E9,$39,$8B,$87,$BF,$E8,$3B,$E7,$FC,$49,$B5
	.dc.b	$FC,$85,$F0,$CE,$F1,$FC,$9F,$D4,$38,$E9,$F9,$6A,$FF,$0E,$5F,$FA
	.dc.b	$B6,$74,$FF,$AB,$FD,$DF,$C8,$B7,$A7,$F3,$B0,$FA,$4B,$1F,$A0,$F8
	.dc.b	$37,$C4,$53,$F4,$7F,$6D,$FE,$A6,$C3,$F9,$B4,$E9,$FF,$5E,$7D,$BF
	.dc.b	$FB,$23,$6F,$E7,$ED,$55,$FC,$BE,$77,$F1,$DF,$AE,$FB,$6C,$57,$E0
	.dc.b	$DB,$C3,$FC,$D9,$7C,$4E,$FB,$3F,$6B,$F4,$1F,$E9,$6C,$7F,$A4,$37
	.dc.b	$0F,$E9,$EF,$F2,$E8,$FF,$68,$39,$87,$0F,$FB,$45,$FC,$93,$71,$FE
	.dc.b	$8B,$F6,$7F,$52,$7F,$AC,$CE,$6F,$FA,$B7,$D9,$FF,$B4,$67,$F3,$4D
	.dc.b	$BF,$E9,$AF,$E9,$1B,$F0,$5F,$C3,$3E,$81,$F3,$34,$C7,$C1,$5C,$7F
	.dc.b	$DC,$DF,$32,$F9,$8F,$E2,$6D,$8F,$F7,$57,$C8,$7A,$E3,$1A,$7E,$AF
	.dc.b	$EC,$FF,$89,$37,$CC,$FE,$37,$1F,$B0,$D7,$FB,$0B,$A7,$FB,$1F,$2F
	.dc.b	$EB,$31,$FF,$7B,$36,$FF,$6D,$F6,$3F,$E8,$F7,$FB,$DC,$66,$BF,$B5
	.dc.b	$F0,$FE,$0E,$7E,$5F,$57,$F2,$AB,$FD,$2E,$1F,$95,$8F,$E5,$BE,$83
	.dc.b	$D6,$A6,$3C,$6B,$5F,$E0,$8C,$BF,$BB,$FC,$9F,$CB,$8F,$F6,$B7,$CC
	.dc.b	$7F,$B8,$F9,$BF,$DC,$8F,$F3,$8F,$AE,$F1,$5F,$F7,$E6,$D7,$F8,$37
	.dc.b	$EB,$9F,$03,$BD,$7F,$9F,$FD,$07,$AE,$BF,$C2,$4D,$E0,$FF,$0C,$AF
	.dc.b	$96,$DF,$A1,$FD,$1B,$F0,$7C,$5F,$FB,$C1,$B7,$FE,$1B,$3F,$C4,$2E
	.dc.b	$1C,$35,$FF,$10,$3D,$D3,$4E,$7A,$7F,$86,$1C,$57,$84,$F1,$5D,$BF
	.dc.b	$90,$7F,$8A,$FF,$07,$FB,$CD,$E8,$FF,$68,$FD,$17,$F5,$E9,$CE,$5F
	.dc.b	$E3,$67,$F1,$2F,$A5,$AF,$1A,$F6,$3B,$AF,$F6,$3A,$F8,$D7,$D1,$61
	.dc.b	$B5,$FE,$8C,$8C,$11,$F3,$CF,$87,$E2,$7F,$DC,$B3,$E5,$7F,$DC,$FE
	.dc.b	$97,$9E,$2D,$70,$A6,$3D,$9F,$E9,$8D,$7D,$53,$E6,$8C,$47,$C5,$8C
	.dc.b	$57,$C3,$3B,$57,$5A,$FC,$CD,$74,$F0,$3E,$41,$FE,$FA,$F4,$ED,$6E
	.dc.b	$3B,$11,$F8,$F3,$E6,$8F,$89,$E0,$49,$FF,$57,$B6,$7F,$D5,$8F,$F8
	.dc.b	$4B,$E7,$3F,$CD,$53,$FF,$00,$23,$0F,$F8,$5B,$47,$D1,$6F,$E7,$30
	.dc.b	$D9,$F8,$D7,$FE,$F2,$7C,$6D,$C3,$ED,$B5,$FC,$C3,$5F,$F3,$A3,$BE
	.dc.b	$4F,$F3,$4B,$FE,$AF,$9E,$B7,$E7,$F8,$FF,$47,$D4,$C9,$FA,$D9,$7F
	.dc.b	$16,$C3,$8C,$E7,$1F,$F0,$16,$28,$F6,$69,$9A,$FF,$3B,$DF,$F2,$8C
	.dc.b	$3D,$7E,$07,$C1,$F2,$3E,$33,$63,$27,$82,$32,$FC,$2E,$FF,$98,$7C
	.dc.b	$9F,$22,$C6,$4F,$8B,$1F,$9C,$CF,$F9,$F6,$9F,$12,$30,$7D,$5F,$A0
	.dc.b	$E3,$1E,$A7,$FC,$93,$F6,$2A,$A7,$D7,$FB,$0B,$87,$E6,$67,$4D,$9D
	.dc.b	$34,$F9,$3F,$31,$D3,$E3,$3F,$00,$D8,$C9,$CE,$32,$D3,$B7,$C9,$79
	.dc.b	$0F,$F2,$EF,$EB,$1F,$8C,$CF,$FC,$65,$8A,$75,$5C,$0F,$D7,$56,$F8
	.dc.b	$7F,$D5,$35,$B9,$D5,$F2,$BE,$B1,$CA,$30,$EA,$BF,$D6,$37,$CE,$3F
	.dc.b	$D8,$17,$CE,$B6,$4A,$DC,$F3,$7C,$37,$63,$FD,$91,$7C,$EB,$FD,$9F
	.dc.b	$7C,$EF,$75,$E1,$36,$71,$9D,$DD,$57,$CD,$5F,$3A,$F3,$D6,$EB,$53
	.dc.b	$9A,$E7,$5D,$DF,$F7,$0C,$EA,$BB,$56,$E7,$BD,$62,$7A,$34,$E5,$1A
	.dc.b	$6C,$F1,$39,$57,$58,$CD,$7D,$26,$9C,$A3,$55,$F3,$63,$D6,$7D,$E3
	.dc.b	$C0,$E4,$F4,$B8,$BF,$97,$1F,$E7,$99,$E7,$18,$24,$FC,$E3,$C3,$3C
	.dc.b	$CF,$51,$6F,$D5,$7E,$78,$C3,$F0,$A7,$8D,$74,$9F,$1C,$74,$7F,$B7
	.dc.b	$57,$E2,$5B,$5F,$D5,$23,$11,$EE,$3F,$AC,$B8,$7E,$B9,$88,$E0,$7E
	.dc.b	$EB,$D5,$FA,$E4,$AF,$A2,$D9,$F5,$36,$FC,$D1,$F8,$88,$C4,$F8,$67
	.dc.b	$F5,$31,$89,$F0,$C6,$27,$A9,$23,$AB,$63,$AD,$FF,$99,$23,$07,$53
	.dc.b	$F5,$75,$D6,$BF,$33,$7E,$B9,$DC,$78,$1D,$98,$76,$FF,$15,$9E,$55
	.dc.b	$F8,$F6,$D7,$3A,$F2,$92,$FC,$B3,$AB,$E4,$75,$BE,$4D,$BE,$55,$C2
	.dc.b	$30,$70,$BE,$8F,$B5,$FD,$ED,$F5,$7F,$8C,$B6,$7E,$54,$70,$3F,$63
	.dc.b	$F4,$BF,$C8,$3E,$63,$F9,$0B,$6F,$F8,$EB,$67,$F3,$8D,$AF,$E2,$FF
	.dc.b	$3D,$FC,$A3,$A7,$F9,$4B,$6F,$F9,$2C,$5B,$FD,$24,$DA,$FE,$3F,$D3
	.dc.b	$FC,$C0,$FE,$63,$EF,$5C,$3F,$96,$C7,$F8,$AB,$6F,$ED,$98,$7F,$8B
	.dc.b	$9A,$FF,$92,$DA,$FF,$20,$84,$FF,$88,$FE,$C9,$F1,$9D,$6B,$FD,$DD
	.dc.b	$18,$9F,$E7,$7F,$01,$DC,$8F,$F8,$5B,$0F,$F6,$3D,$78,$7E,$63,$FA
	.dc.b	$39,$FD,$22,$AD,$7F,$44,$EE,$FE,$97,$F4,$5F,$D3,$1B,$1F,$D2,$23
	.dc.b	$D8,$47,$B3,$EF,$F9,$6F,$EA,$67,$F5,$4F,$86,$B3,$FD,$42,$34,$F9
	.dc.b	$5B,$8F,$F9,$4B,$3F,$F0,$23,$7B,$BF,$CA,$71,$88,$FF,$39,$35,$FE
	.dc.b	$8F,$EA,$FC,$F2,$5F,$C1,$7E,$97,$81,$23,$E2,$36,$7E,$A7,$B3,$F5
	.dc.b	$C9,$3F,$25,$FB,$5F,$4D,$23,$E6,$36,$7E,$E7,$AD,$F7,$C9,$3F,$45
	.dc.b	$FC,$17,$85,$23,$EA,$36,$7F,$09,$E1,$7E,$1C,$93,$F6,$5F,$C5,$7A
	.dc.b	$72,$3E,$E3,$67,$FC,$F3,$83,$E1,$BF,$CF,$2D,$AF,$F2,$BB,$67,$73
	.dc.b	$AF,$A6,$95,$A6,$CF,$4F,$13,$99,$27,$9B,$DB,$E2,$64,$76,$6C,$FF
	.dc.b	$A3,$3F,$13,$67,$FE,$8C,$6D,$7F,$9A,$9B,$3B,$9F,$F4,$8F,$D3,$3C
	.dc.b	$0F,$64,$FF,$A5,$A3,$19,$FF,$9C,$97,$80,$F9,$4C,$5F,$0B,$FF,$5C
	.dc.b	$6C,$F9,$A4,$7F,$A3,$E1,$F2,$98,$87,$87,$FB,$27,$73,$A7,$85,$6E
	.dc.b	$75,$0F,$A1,$75,$67,$83,$A6,$0F,$95,$C5,$F4,$6B,$C6,$AF,$D2,$3C
	.dc.b	$6B,$E3,$4A,$FF,$53,$4E,$5F,$57,$99,$CB,$17,$CE,$53,$11,$DB,$E8
	.dc.b	$DE,$9E,$FE,$72,$61,$FE,$B1,$9C,$BF,$EB,$3F,$A4,$79,$FE,$67,$FC
	.dc.b	$99,$19,$F4,$31,$3E,$1F,$83,$F1,$44,$9F,$9D,$1E,$EF,$AC,$7B,$FE
	.dc.b	$7B,$95,$FE,$C1,$9C,$BF,$EC,$3E,$87,$A7,$F1,$97,$A7,$4C,$47,$C1
	.dc.b	$F9,$1F,$85,$F2,$B2,$FE,$F4,$7C,$EF,$BF,$FF,$6D,$C7,$E0,$7E,$3F
	.dc.b	$E4,$19,$F8,$2F,$A7,$F6,$FF,$2A,$7F,$DA,$8B,$97,$26,$B9,$BE,$87
	.dc.b	$D3,$7D,$1F,$A8,$97,$F8,$D8,$FB,$BF,$64,$F5,$7E,$B7,$D5,$CA,$FF
	.dc.b	$70,$4E,$5F,$F7,$17,$DE,$3D,$6F,$C2,$AF,$62,$98,$8F,$C0,$7C,$5F
	.dc.b	$E0,$7E,$32,$5F,$E7,$63,$F1,$BF,$74,$F8,$9F,$C7,$FE,$B3,$F9,$FA
	.dc.b	$FC,$FF,$95,$FA,$18,$C7,$99,$F9,$CB,$E1,$5F,$EF,$5F,$21,$8F,$E6
	.dc.b	$16,$F4,$C4,$76,$F3,$FB,$FD,$24,$BF,$F7,$F6,$5F,$B7,$7F,$E0,$07
	.dc.b	$DC,$27,$CB,$E1,$C9,$CC,$7E,$9B,$EF,$BE,$14,$7E,$EB,$5F,$D8,$C7
	.dc.b	$C9,$8F,$59,$FB,$E7,$11,$AE,$2F,$83,$F0,$DF,$0B,$E9,$E5,$F9,$23
	.dc.b	$B7,$BF,$A8,$C1,$DF,$8B,$F8,$36,$29,$FD,$CF,$D6,$7E,$F6,$B4,$7F
	.dc.b	$0E,$FA,$D7,$C6,$FE,$F9,$B1,$A1,$3C,$5F,$49,$8E,$FE,$BE,$AC,$8E
	.dc.b	$13,$F9,$37,$97,$D6,$FC,$97,$A7,$8C,$F8,$3F,$E1,$AD,$8D,$F1,$7F
	.dc.b	$10,$AF,$D7,$C7,$81,$E4,$D8,$D7,$1F,$F1,$6F,$AD,$7C,$0F,$4F,$0A
	.dc.b	$5F,$AE,$72,$DF,$6F,$85,$48,$C1,$BF,$3B,$A4,$C8,$C1,$CF,$6F,$4F
	.dc.b	$C3,$48,$C1,$DF,$83,$C9,$32,$30,$78,$FC,$8F,$E7,$6F,$E7,$C6,$5F
	.dc.b	$C9,$BA,$B1,$18,$9D,$BF,$A5,$9D,$3C,$57,$AC,$FA,$DE,$31,$86,$C6
	.dc.b	$67,$83,$5E,$FF,$3B,$E0,$7F,$A9,$B5,$34,$D7,$81,$FD,$4F,$1F,$E9
	.dc.b	$F3,$FD,$7F,$AF,$F1,$4F,$21,$F5,$AE,$5F,$DF,$5F,$1A,$FE,$01,$CB
	.dc.b	$6F,$FB,$03,$5F,$EC,$B3,$5F,$DA,$4F,$6E,$5C,$1B,$E6,$FF,$26,$E5
	.dc.b	$E4,$8E,$1D,$B2,$B8,$1E,$67,$A7,$74,$8C,$1E,$A7,$C5,$5F,$DD,$23
	.dc.b	$F4,$CD,$C3,$F4,$B8,$7F,$23,$9C,$F3,$70,$E3,$91,$83,$8B,$ED,$FA
	.dc.b	$AF,$91,$7F,$AE,$6E,$1F,$DE,$B0,$FD,$E5,$FE,$D6,$72,$FF,$81,$5F
	.dc.b	$87,$5C,$75,$E9,$AF,$57,$B7,$D6,$7D,$69,$FF,$07,$D2,$31,$08,$C5
	.dc.b	$FE,$49,$B1,$F9,$75,$F5,$9E,$93,$AC,$F9,$5F,$B1,$E5,$FE,$D2,$3F
	.dc.b	$36,$DA,$FD,$03,$5F,$E8,$0D,$7D,$9F,$35,$E4,$B3,$49,$5E,$5F,$2F
	.dc.b	$44,$70,$7E,$AB,$EB,$3A,$5F,$F8,$B9,$A9,$A6,$3E,$93,$ED,$47,$ED
	.dc.b	$5A,$FF,$55,$6B,$FE,$35,$6B,$FE,$3A,$F8,$9F,$F4,$EB,$FD,$6D,$AF
	.dc.b	$F8,$ED,$AE,$FE,$4F,$E8,$FB,$2F,$8F,$A4,$C7,$F8,$1C,$65,$FE,$0B
	.dc.b	$F5,$8F,$91,$FB,$2D,$8C,$C6,$0F,$BB,$F5,$6E,$5A,$49,$FE,$1F,$19
	.dc.b	$7F,$88,$FD,$5B,$E5,$7F,$0A,$D8,$CC,$60,$FC,$4E,$CF,$89,$A4,$9F
	.dc.b	$E3,$71,$97,$F8,$EF,$D5,$7F,$43,$6F,$99,$F9,$28,$C7,$07,$F2,$56
	.dc.b	$BF,$E0,$5F,$AA,$FF,$57,$3F,$E6,$86,$A6,$98,$FA,$0F,$F2,$6D,$FF
	.dc.b	$2C,$6C,$FF,$31,$6B,$FE,$11,$6B,$FE,$71,$FA,$AF,$F2,$2C,$FF,$34
	.dc.b	$6C,$7F,$39,$6B,$FE,$19,$6B,$FE,$79,$DD,$EE,$7E,$B9,$B5,$FC,$F9
	.dc.b	$B3,$FD,$0D,$AF,$E3,$F8,$FE,$69,$27,$E5,$BF,$B8,$F1,$DE,$6D,$24
	.dc.b	$FF,$3C,$7F,$7B,$F5,$3F,$C2,$AF,$FA,$83,$5F,$F1,$83,$5F,$F4,$83
	.dc.b	$5F,$F4,$AF,$D4,$FE,$45,$FE,$B0,$D7,$FC,$70,$D7,$FD,$30,$D7,$FD
	.dc.b	$3B,$F5,$1E,$C7,$FD,$38,$D7,$39,$FF,$12,$8C,$43,$EA,$3E,$2F,$8B
	.dc.b	$FE,$9C,$6E,$5F,$EA,$57,$FC,$82,$FF,$86,$5C,$CF,$FA,$B1,$E6,$4B
	.dc.b	$FE,$3E,$D9,$FF,$10,$7D,$37,$FA,$C9,$E5,$3F,$D4,$63,$15,$FE,$22
	.dc.b	$7F,$03,$7F,$EA,$E5,$E5,$6F,$CB,$E0,$CC,$3F,$BD,$CE,$9B,$1F,$CC
	.dc.b	$23,$17,$FE,$03,$5F,$4F,$92,$57,$27,$87,$8F,$E5,$3C,$CD,$8F,$5F
	.dc.b	$89,$E9,$3D,$5F,$11,$63,$27,$92,$3C,$44,$70,$7C,$4F,$17,$86,$3F
	.dc.b	$05,$F8,$9F,$27,$F1,$7E,$5A,$D7,$CA,$70,$FC,$07,$DC,$47,$FC,$3B
	.dc.b	$F3,$A7,$FA,$8B,$5C,$98,$33,$F1,$FF,$E0,$38,$FF,$57,$C7,$F9,$5D
	.dc.b	$E1,$87,$F9,$59,$CB,$7E,$1B,$81,$C7,$E6,$6D,$71,$8F,$09,$1F,$2B
	.dc.b	$67,$A7,$CC,$70,$6C,$F8,$7E,$83,$C9,$E2,$F9,$5B,$1E,$58,$B3,$E5
	.dc.b	$9B,$CB,$FE,$F6,$CB,$F7,$37,$EB,$EE,$F7,$76,$7B,$D9,$19,$3E,$04
	.dc.b	$63,$0F,$81,$F3,$BF,$14,$F8,$DE,$25,$AF,$8D,$1F,$38,$8F,$D2,$6C
	.dc.b	$FC,$9F,$03,$83,$67,$FD,$09,$F4,$7F,$DE,$73,$FB,$12,$11,$99,$C4
	.dc.b	$63,$3F,$F1,$D4,$E6,$38,$35,$E3,$F0,$1A,$9D,$BC,$3D,$FE,$3E,$FD
	.dc.b	$C3,$C6,$DB,$FF,$4A,$E0,$8C,$63,$E6,$9F,$2C,$78,$BE,$5A,$5F,$B3
	.dc.b	$E6,$2D,$72,$9B,$7C,$70,$56,$2F,$DB,$0A,$C5,$FB,$2A,$BF,$12,$2C
	.dc.b	$F2,$AF,$97,$18,$3E,$2D,$58,$C9,$E5,$2D,$79,$5B,$7E,$59,$E1,$39
	.dc.b	$3E,$1B,$88,$F5,$57,$6A,$E2,$7A,$9F,$2C,$EE,$30,$78,$6B,$A5,$72
	.dc.b	$8E,$31,$EF,$6B,$EF,$6D,$FB,$E3,$8C,$74,$69,$C4,$A7,$12,$C7,$19
	.dc.b	$F0,$46,$2F,$D4,$DA,$EF,$7E,$98,$FC,$6C,$EA,$B8,$92,$BC,$4F,$C9
	.dc.b	$7C,$F1,$F0,$9A,$74,$2C,$F9,$2B,$B3,$E7,$AE,$4F,$8C,$D4,$F0,$7F
	.dc.b	$46,$7B,$4B,$7F,$89,$8F,$F5,$FB,$63,$FD,$88,$D7,$77,$FE,$76,$8C
	.dc.b	$47,$CB,$33,$3F,$B6,$F3,$FF,$1E,$47,$E9,$B6,$3E,$6B,$6F,$E2,$47
	.dc.b	$ED,$23,$13,$F1,$8F,$CF,$B8,$8F,$9B,$1D,$CE,$A4,$7A,$B6,$7A,$C7
	.dc.b	$F0,$69,$E4,$FC,$99,$EA,$6C,$F0,$3E,$37,$0F,$80,$FD,$FC,$72,$AF
	.dc.b	$91,$5F,$B7,$8C,$5F,$F1,$06,$DF,$9E,$FC,$A4,$AF,$2D,$FF,$98,$63
	.dc.b	$13,$E2,$7D,$0F,$89,$F4,$1E,$53,$67,$96,$7C,$51,$B9,$F2,$BE,$E9
	.dc.b	$D9,$C8,$EA,$62,$36,$F5,$9F,$2D,$7A,$A3,$E0,$D7,$5F,$91,$FE,$5B
	.dc.b	$F2,$5F,$CB,$9B,$1F,$CE,$1B,$3F,$C9,$9B,$7F,$C9,$7E,$87,$FC,$3A
	.dc.b	$DF,$73,$FD,$E1,$18,$BF,$E7,$6F,$FB,$42,$FF,$A0,$35,$34,$E0,$EA
	.dc.b	$79,$FB,$F2,$BF,$A4,$5F,$87,$C5,$C5,$A3,$C9,$5F,$ED,$4C,$0D,$25
	.dc.b	$FE,$B2,$BC,$71,$E8,$F8,$0E,$77,$ED,$7D,$7E,$2F,$32,$5F,$1B,$F4
	.dc.b	$C7,$7F,$77,$E0,$9C,$27,$CE,$7C,$27,$E2,$F8,$7F,$1B,$D0,$B3,$AB
	.dc.b	$F8,$B1,$98,$ED,$9A,$3D,$84,$8C,$FC,$4F,$D0,$3E,$8F,$89,$63,$A5
	.dc.b	$FC,$E9,$F8,$84,$8E,$27,$AF,$F3,$67,$87,$FE,$FC,$6C,$66,$30,$7F
	.dc.b	$C0,$0F,$C2,$FF,$C4,$F7,$FC,$59,$AF,$F7,$A6,$BF,$DF,$DA,$BE,$41
	.dc.b	$BF,$F6,$F8,$FF,$00,$C6,$2F,$E7,$36,$3E,$C3,$5F,$E4,$AB,$EC,$39
	.dc.b	$47,$F8,$36,$31,$3C,$A7,$33,$E1,$7D,$5E,$F4,$9F,$B9,$87,$FB,$BA
	.dc.b	$8D,$FA,$67,$85,$C9,$B5,$F8,$36,$D7,$E3,$5A,$FC,$5F,$8C,$F9,$04
	.dc.b	$9E,$EF,$93,$2F,$F3,$CC,$7C,$0F,$3F,$04,$7E,$1F,$2E,$C8,$E0,$F8
	.dc.b	$FC,$BC,$C8,$E0,$F9,$7F,$57,$DC,$8E,$0F,$9F,$E8,$78,$C8,$E0,$FD
	.dc.b	$07,$D2,$FF,$AD,$73,$34,$9E,$3F,$CE,$FA,$57,$9E,$F8,$4F,$D4,$46
	.dc.b	$1B,$F4,$AF,$07,$35,$BC,$7F,$6D,$F8,$4B,$6F,$D8,$42,$30,$FE,$CD
	.dc.b	$AF,$08,$DF,$C1,$3E,$46,$92,$7F,$75,$19,$7F,$79,$F4,$CE,$FE,$ED
	.dc.b	$8C,$C6,$0F,$1F,$D3,$3D,$6A,$49,$FE,$0D,$19,$7B,$F9,$FE,$42,$38
	.dc.b	$3F,$86,$7D,$33,$D9,$F7,$B6,$33,$18,$3E,$1F,$D3,$7F,$92,$E3,$F8
	.dc.b	$A3,$5F,$E3,$CD,$7F,$C1,$6D,$7F,$CD,$F9,$BE,$9B,$E3,$92,$3E,$2B
	.dc.b	$F4,$3E,$99,$EF,$D2,$4F,$F2,$38,$CB,$FC,$97,$E4,$DF,$4D,$F6,$1B
	.dc.b	$19,$8C,$1F,$5B,$D3,$91,$F5,$1A,$FF,$37,$FA,$67,$77,$E0,$9B,$1F
	.dc.b	$CC,$DA,$EE,$3F,$0F,$F4,$BF,$E7,$D8,$FE,$70,$D7,$FA,$13,$5F,$F1
	.dc.b	$7B,$5F,$8B,$EA,$7E,$28,$8F,$E2,$1F,$C9,$FD,$2B,$C6,$A4,$9F,$E8
	.dc.b	$D1,$97,$FA,$47,$D2,$BE,$5B,$F3,$6D,$8C,$C6,$0F,$CC,$FD,$1C,$8F
	.dc.b	$CB,$B5,$FE,$B5,$F4,$AF,$5B,$F4,$AD,$8F,$EA,$CD,$77,$1F,$AA,$F4
	.dc.b	$95,$49,$3F,$D7,$63,$2F,$F3,$2F,$0B,$F6,$3E,$67,$ED,$E3,$04,$3F
	.dc.b	$B6,$7D,$23,$E6,$BF,$6E,$D8,$FE,$D0,$D7,$71,$FB,$CF,$A3,$7D,$25
	.dc.b	$24,$FF,$6F,$8C,$BF,$D1,$3D,$DF,$19,$1C,$1B,$FA,$4E,$84,$70,$7F
	.dc.b	$79,$FA,$37,$CF,$FF,$B4,$3D,$FF,$F6,$07,$D2,$3D,$EC,$7A,$9F,$ED
	.dc.b	$48,$C6,$1F,$DB,$30,$E2,$DF,$1F,$6F,$FB,$76,$32,$FF,$85,$6B,$FD
	.dc.b	$BB,$18,$C3,$FD,$9C,$AB,$8D,$7F,$64,$DA,$F7,$3E,$26,$9A,$F7,$BE
	.dc.b	$56,$3F,$ED,$77,$F6,$37,$DC,$FF,$77,$3A,$CB,$FD,$DA,$D8,$CF,$89
	.dc.b	$E0,$FA,$47,$C4,$F8,$3B,$E5,$7F,$82,$A7,$2D,$F4,$38,$54,$8C,$1B
	.dc.b	$F1,$BA,$4C,$8C,$1C,$F8,$FD,$BF,$17,$23,$07,$7E,$1F,$24,$C8,$C1
	.dc.b	$E3,$F8,$AF,$F2,$03,$FE,$3B,$CB,$FD,$F5,$BF,$F8,$D5,$B1,$99,$DB
	.dc.b	$FE,$AC,$87,$D1,$BE,$93,$83,$67,$A3,$5E,$0D,$7F,$C9,$5F,$33,$FD
	.dc.b	$51,$34,$8F,$F9,$1F,$BD,$C7,$FE,$34,$8F,$F3,$10,$38,$8F,$A4,$FF
	.dc.b	$12,$B9,$37,$CF,$C9,$FB,$C7,$17,$1A,$7F,$8A,$EE,$DF,$FB,$2A,$6D
	.dc.b	$73,$E4,$AF,$85,$BC,$EF,$85,$CC,$FF,$40,$BF,$47,$6E,$0C,$DD,$8F
	.dc.b	$67,$C5,$70,$31,$5E,$63,$FC,$E7,$1F,$E5,$4C,$DC,$DE,$26,$C7,$F9
	.dc.b	$63,$2F,$E0,$6E,$E7,$5C,$72,$30,$73,$7D,$7F,$49,$FE,$8D,$3F,$D1
	.dc.b	$D4,$EF,$CE,$FA,$DE,$61,$FE,$66,$BE,$11,$CB,$E0,$F9,$FB,$76,$B4
	.dc.b	$E1,$A7,$18,$C6,$BF,$53,$CA,$FD,$5C,$F5,$F9,$5F,$9F,$18,$9F,$B3
	.dc.b	$8A,$7D,$27,$AB,$CA,$B7,$E5,$70,$F9,$FE,$2E,$DF,$27,$ED,$FD,$27
	.dc.b	$5E,$2F,$C3,$4A,$3E,$6B,$F8,$87,$55,$F8,$D7,$33,$C3,$E8,$3E,$8D
	.dc.b	$6F,$E9,$7F,$0F,$18,$9E,$B4,$9F,$C5,$FC,$27,$E2,$70,$FF,$48,$AF
	.dc.b	$A8,$CF,$56,$9F,$C8,$FD,$33,$E2,$3F,$27,$F6,$0F,$A9,$CC,$61,$FC
	.dc.b	$BF,$D4,$3E,$6B,$77,$86,$E7,$FE,$A0,$FA,$8F,$F7,$A6,$FF,$09,$57
	.dc.b	$E8,$1C,$CF,$EB,$5C,$BF,$EA,$8F,$A8,$7C,$F7,$E9,$9F,$81,$1F,$B3
	.dc.b	$72,$EF,$D5,$F8,$BF,$09,$FA,$CF,$B0,$FF,$28,$4F,$FA,$CB,$95,$39
	.dc.b	$F3,$FF,$63,$F5,$3F,$CF,$6B,$FD,$6E,$FF,$A2,$B5,$FB,$2E,$7A,$3F
	.dc.b	$7E,$E6,$7F,$82,$7D,$4F,$F9,$5D,$FF,$5F,$AF,$AC,$FD,$EB,$9A,$FE
	.dc.b	$0E,$E6,$7F,$D8,$BD,$7F,$E4,$F7,$FD,$8F,$F5,$7F,$E1,$B7,$3F,$17
	.dc.b	$FC,$1A,$31,$7F,$C4,$3E,$AF,$F8,$23,$7C,$DF,$E1,$71,$8D,$3F,$22
	.dc.b	$A3,$F8,$7F,$33,$E9,$FF,$88,$FD,$87,$FA,$07,$3F,$F6,$BC,$E5,$FF
	.dc.b	$6C,$C6,$71,$FE,$3F,$C5,$F8,$79,$CF,$85,$F9,$4E,$CE,$6F,$F1,$CF
	.dc.b	$45,$EB,$DC,$B0,$DB,$FF,$72,$FD,$67,$F0,$56,$FC,$EF,$F2,$48,$C6
	.dc.b	$7F,$CB,$9F,$A0,$FF,$37,$FA,$CF,$13,$FC,$CB,$EC,$1F,$4F,$FC,$B2
	.dc.b	$73,$C7,$FB,$29,$CE,$3F,$EF,$4F,$AD,$F7,$61,$FC,$EB,$D1,$FE,$85
	.dc.b	$39,$E3,$FE,$6B,$18,$9F,$E7,$FF,$5B,$C3,$1F,$DE,$B6,$FF,$A3,$3F
	.dc.b	$E3,$C7,$FE,$00,$B9,$9F,$E9,$CD,$5F,$5B,$D6,$B9,$E1,$FD,$12,$31
	.dc.b	$7F,$D4,$5F,$15,$7F,$53,$38,$9F,$EE,$C7,$2E,$CC,$73,$49,$E1,$1F
	.dc.b	$D7,$E7,$F0,$91,$88,$F1,$CF,$FB,$8F,$CE,$FE,$AB,$F1,$3F,$C9,$6B
	.dc.b	$FB,$36,$3F,$C7,$30,$E3,$87,$F6,$2A,$D3,$FD,$92,$1E,$97,$AC,$F6
	.dc.b	$7D,$15,$C3,$99,$ED,$8F,$51,$1F,$24,$F2,$6E,$3E,$F2,$38,$25,$F8
	.dc.b	$79,$FE,$54,$FC,$B6,$4F,$47,$E4,$11,$EF,$F3,$7F,$2C,$8E,$0F,$8D
	.dc.b	$97,$F9,$FD,$F4,$DF,$F8,$26,$30,$FC,$19,$F9,$35,$FD,$FB,$EA,$DF
	.dc.b	$5B,$4C,$61,$4E,$76,$9D,$C7,$D5,$6C,$7D,$A6,$BF,$83,$6B,$FE,$17
	.dc.b	$5F,$1D,$FD,$4A,$30,$D7,$FC,$41,$1A,$9E,$D3,$99,$F1,$3E,$9F,$9D
	.dc.b	$F5,$36,$7C,$F5,$FE,$F1,$6C,$E9,$A6,$BC,$CE,$93,$FC,$E2,$FD,$51
	.dc.b	$82,$4F,$C2,$CD,$E2,$7C,$99,$FC,$9B,$F2,$FB,$65,$78,$BB,$6E,$1F
	.dc.b	$E3,$4C,$7F,$AB,$4F,$C9,$AF,$09,$1F,$99,$DB,$F5,$7B,$FE,$B3,$63
	.dc.b	$DA,$DB,$E8,$DE,$5F,$EF,$B8,$7F,$0E,$AF,$A1,$7C,$9F,$8A,$FF,$8F
	.dc.b	$BE,$AF,$FB,$FE,$1F,$87,$FB,$0C,$7C,$4F,$EB,$9B,$7F,$05,$B3,$FA
	.dc.b	$D7,$97,$C4,$49,$FC,$3D,$FF,$A6,$5B,$19,$AF,$F2,$AC,$A3,$C5,$F1
	.dc.b	$39,$CF,$FC,$2B,$3F,$DC,$AB,$FC,$C4,$D4,$BE,$B7,$D5,$F0,$53,$1F
	.dc.b	$0B,$ED,$8C,$57,$82,$2D,$F6,$9B,$5F,$E6,$BB,$EC,$48,$CD,$F8,$AE
	.dc.b	$BE,$28,$B3,$C2,$7C,$15,$E2,$25,$78,$5F,$8B,$5C,$27,$8B,$63,$77
	.dc.b	$E5,$69,$E5,$9D,$B8,$AF,$23,$5F,$23,$6F,$C8,$F8,$0C,$4F,$29,$F2
	.dc.b	$12,$BE,$8C,$7A,$A3,$13,$E1,$3A,$5F,$9E,$32,$7D,$67,$15,$CA,$7C
	.dc.b	$13,$A3,$EC,$D7,$86,$75,$19,$BF,$A1,$5E,$63,$B9,$63,$57,$DA,$3C
	.dc.b	$91,$F8,$C8,$C3,$F1,$0C,$47,$1A,$EE,$7C,$D8,$F0,$5F,$F0,$DB,$34
	.dc.b	$95,$FD,$4E,$32,$D8,$CC,$FF,$5A,$9C,$B6,$3F,$AB,$F9,$7F,$88,$E4
	.dc.b	$FF,$16,$4F,$C3,$6E,$1E,$4C,$3F,$D7,$91,$87,$DB,$1F,$E7,$88,$C4
	.dc.b	$FB,$63,$15,$D4,$95,$F8,$3A,$DD,$7D,$08,$FE,$35,$3A,$8F,$13,$6B
	.dc.b	$C5,$1F,$D2,$63,$13,$E2,$3C,$AE,$23,$C5,$18,$3C,$11,$98,$F0,$1F
	.dc.b	$DC,$DC,$61,$FB,$BB,$F2,$B6,$33,$87,$C8,$BE,$64,$BF,$3B,$70,$F3
	.dc.b	$C3,$0E,$75,$C6,$36,$49,$DC,$EA,$BE,$19,$CC,$F7,$4F,$C1,$70,$FB
	.dc.b	$8C,$1C,$DD,$BE,$1B,$F4,$C7,$48,$D1,$C4,$D3,$D1,$F8,$B5,$F6,$A3
	.dc.b	$F9,$4F,$D6,$FF,$87,$BA,$BF,$9B,$36,$32,$7F,$53,$8C,$B4,$FE,$AD
	.dc.b	$F2,$3F,$E2,$63,$FD,$F7,$F5,$BF,$E6,$7B,$FC,$54,$66,$76,$7F,$7E
	.dc.b	$9D,$3F,$92,$8E,$3F,$1F,$23,$5F,$13,$72,$61,$FE,$CC,$C1,$7F,$DD
	.dc.b	$DB,$7F,$E3,$B6,$D7,$F9,$01,$AF,$F9,$27,$4F,$F4,$93,$FE,$39,$F4
	.dc.b	$BB,$61,$FA,$17,$D7,$F0,$BD,$49,$87,$56,$D7,$F9,$07,$EB,$F7,$3F
	.dc.b	$DE,$53,$FE,$1C,$FD,$6A,$B4,$F8,$FD,$BA,$AF,$6C,$19,$3C,$1C,$98
	.dc.b	$8F,$F0,$B4,$66,$3F,$CC,$8D,$79,$07,$0F,$5C,$FE,$0B,$FE,$6B,$6B
	.dc.b	$C1,$AE,$F1,$FF,$12,$7D,$7F,$F8,$9E,$3F,$CD,$6D,$7F,$CE,$8D,$7C
	.dc.b	$2D,$7D,$7E,$17,$A0,$98,$7F,$9C,$70,$FF,$1C,$CF,$F9,$2F,$6F,$F1
	.dc.b	$96,$DF,$62,$3F,$8B,$C6,$0F,$83,$F4,$12,$F9,$36,$7F,$CF,$FF,$5C
	.dc.b	$F8,$5F,$F1,$E3,$6B,$FD,$00,$D9,$DC,$FC,$CF,$AA,$95,$E5,$6C,$F4
	.dc.b	$F9,$5E,$64,$9E,$6F,$A3,$CC,$91,$FE,$68,$6C,$FF,$A4,$B4,$7D,$57
	.dc.b	$DA,$25,$7C,$37,$E0,$F2,$48,$FF,$38,$B6,$B9,$79,$3F,$82,$24,$FC
	.dc.b	$B7,$E4,$FD,$14,$8F,$F3,$CB,$67,$C5,$E5,$7E,$28,$93,$F4,$DF,$F3
	.dc.b	$17,$D5,$F9,$BE,$2F,$FD,$4C,$DA,$FF,$55,$AD,$9E,$A6,$B4,$5F,$FA
	.dc.b	$B7,$EA,$FF,$D4,$78,$5F,$4B,$FD,$60,$A3,$2F,$FA,$D9,$AF,$E7,$38
	.dc.b	$B7,$C1,$F6,$E3,$F1,$11,$83,$9F,$C1,$BE,$0F,$F4,$3D,$FF,$AD,$A4
	.dc.b	$E5,$FF,$47,$B5,$FD,$69,$FA,$7F,$8C,$E3,$C3,$FA,$AF,$AB,$78,$98
	.dc.b	$AF,$DB,$46,$1F,$F4,$CB,$5E,$79,$BE,$C3,$84,$3E,$B1,$CB,$86,$C7
	.dc.b	$FA,$89,$AF,$91,$F8,$0F,$9B,$37,$99,$D1,$B1,$EC,$8E,$F8,$3D,$5F
	.dc.b	$0F,$D6,$3C,$BA,$4B,$FE,$2D,$DD,$E6,$FA,$CF,$F3,$04,$7F,$AD,$1A
	.dc.b	$FC,$66,$F1,$7C,$D7,$FC,$37,$1F,$33,$EB,$3C,$84,$7F,$91,$7C,$67
	.dc.b	$C2,$FA,$C7,$CE,$52,$4F,$F2,$7F,$98,$F8,$FF,$38,$F1,$E9,$27,$FD
	.dc.b	$8F,$1F,$91,$7E,$DF,$A7,$F0,$C8,$FF,$31,$F7,$7E,$8F,$D6,$3C,$6A
	.dc.b	$49,$FE,$6D,$EF,$7D,$6F,$AC,$FF,$44,$47,$FB,$51,$AF,$E1,$9B,$C3
	.dc.b	$F8,$D5,$F4,$1F,$4F,$EB,$25,$7E,$39,$FA,$7D,$5F,$82,$FA,$C7,$35
	.dc.b	$25,$7F,$47,$FA,$0F,$C3,$FD,$63,$8E,$92,$7F,$DC,$F0,$E0,$FC,$17
	.dc.b	$D5,$C9,$FC,$EB,$F8,$2F,$A7,$FC,$8F,$D6,$3E,$22,$92,$BF,$AB,$7C
	.dc.b	$47,$E5,$BD,$36,$14,$93,$FE,$F6,$87,$C6,$7E,$73,$D9,$FE,$03,$3F
	.dc.b	$ED,$9F,$4F,$06,$7B,$7F,$43,$F5,$8F,$A5,$A4,$AF,$F8,$03,$F5,$9F
	.dc.b	$11,$BE,$17,$EA,$20,$D3,$1C,$7E,$1F,$8F,$95,$FE,$E9,$78,$72,$C8
	.dc.b	$EA,$D7,$F7,$F6,$E3,$FD,$AF,$D5,$BE,$A3,$F6,$FF,$58,$F8,$8A,$6A
	.dc.b	$37,$E7,$5C,$7F,$BC,$78,$3F,$C3,$F0,$E5,$18,$F5,$3F,$B9,$C7,$ED
	.dc.b	$A3,$F8,$7F,$C0,$FC,$37,$47,$F8,$07,$C7,$F2,$38,$3F,$D0,$FA,$7F
	.dc.b	$0A,$9E,$11,$CB,$93,$1C,$72,$38,$4F,$EB,$9E,$5F,$58,$F8,$FF,$E2
	.dc.b	$78,$FC,$F9,$FE,$42,$DD,$3F,$94,$56,$E7,$8F,$B3,$FC,$6A,$31,$1F
	.dc.b	$CA,$1B,$3C,$27,$5E,$5E,$D9,$5E,$28,$E5,$C1,$E2,$F0,$64,$F9,$A3
	.dc.b	$B7,$0E,$A3,$07,$A9,$F1,$7C,$77,$F9,$12,$3F,$C3,$E7,$F8,$A7,$2F
	.dc.b	$F8,$01,$96,$23,$13,$B7,$73,$A7,$4F,$EE,$BE,$B3,$84,$23,$0E,$BC
	.dc.b	$94,$EA,$73,$3C,$3E,$6F,$FA,$EA,$69,$1E,$BD,$DF,$E4,$1C,$7F,$AB
	.dc.b	$47,$F9,$90,$F5,$CF,$F9,$0F,$EB,$3F,$CB,$F1,$FE,$3C,$AE,$1C,$FF
	.dc.b	$03,$92,$CF,$F9,$01,$AF,$F9,$2E,$6B,$FB,$19,$FC,$63,$FC,$C9,$BC
	.dc.b	$BC,$38,$FF,$C4,$99,$79,$23,$8F,$64,$AD,$9C,$9E,$BF,$1D,$23,$07
	.dc.b	$67,$C7,$4E,$4A,$77,$EF,$B5,$F9,$F9,$FD,$93,$79,$33,$9F,$F7,$69
	.dc.b	$CB,$FD,$E3,$EB,$3E,$42,$6E,$8E,$FF,$87,$1A,$3E,$14,$F1,$DF,$3B
	.dc.b	$E9,$DF,$C1,$70,$F8,$7B,$A7,$EC,$F0,$46,$63,$0E,$B0,$F2,$3F,$4B
	.dc.b	$C2,$FD,$84,$F1,$F9,$6C,$76,$F9,$19,$1D,$7C,$AB,$87,$B5,$F4,$9E
	.dc.b	$AF,$8F,$B3,$E3,$7E,$FF,$D6,$7C,$4E,$1F,$F1,$03,$67,$FD,$06,$FF
	.dc.b	$82,$1F,$E1,$8E,$6F,$FD,$20,$FF,$40,$8C,$F9,$7D,$3C,$AE,$AF,$E3
	.dc.b	$31,$6F,$4C,$6D,$6F,$F8,$C7,$D5,$BE,$63,$F2,$6F,$C2,$CF,$FC,$5E
	.dc.b	$E6,$77,$BE,$BE,$8F,$7E,$AD,$9E,$6F,$E7,$3F,$02,$76,$5F,$23,$0D
	.dc.b	$AF,$E5,$1F,$51,$FD,$A5,$BE,$47,$E9,$23,$1A,$7F,$90,$1F,$ED,$53
	.dc.b	$FC,$BF,$EA,$3F,$2C,$DE,$EF,$D5,$C6,$2B,$FC,$8E,$FC,$94,$B3,$DF
	.dc.b	$CC,$70,$6D,$FF,$3B,$FA,$77,$CC,$FE,$D9,$B1,$FC,$FF,$7F,$F6,$07
	.dc.b	$4F,$F9,$66,$73,$97,$F9,$5A,$7F,$B1,$D6,$FC,$2F,$DF,$C7,$B2,$FF
	.dc.b	$CC,$0E,$AB,$FC,$BB,$F8,$0F,$CF,$6F,$2F,$F0,$68,$FA,$95,$FE,$67
	.dc.b	$74,$FF,$9A,$7E,$9F,$F4,$ED,$F1,$FF,$86,$C6,$32,$FF,$67,$BF,$C2
	.dc.b	$F0,$FF,$35,$AE,$ED,$FC,$F7,$06,$D7,$F5,$DF,$A4,$FF,$51,$B7,$D9
	.dc.b	$FE,$33,$18,$C3,$FC,$EA,$FF,$B4,$72,$FF,$3E,$7D,$33,$E8,$7F,$8F
	.dc.b	$B7,$0F,$ED,$1F,$4B,$FE,$92,$7F,$B5,$46,$70,$FF,$3F,$CB,$E8,$BF
	.dc.b	$93,$B6,$B8,$F9,$98,$9F,$E5,$7F,$27,$EF,$BF,$F4,$0F,$97,$8E,$1C
	.dc.b	$BF,$CC,$3E,$8F,$E1,$6B,$FE,$91,$6E,$5F,$EF,$4F,$B3,$7D,$8E,$F0
	.dc.b	$93,$FD,$FE,$9E,$96,$32,$FF,$4A,$C6,$27,$53,$FD,$22,$42,$74,$FF
	.dc.b	$61,$6B,$FD,$99,$AF,$03,$87,$CF,$F3,$6D,$7F,$62,$94,$E5,$B1,$9F
	.dc.b	$27,$E0,$1F,$06,$6C,$7A,$FC,$EF,$88,$7C,$5F,$8B,$B1,$93,$B4,$75
	.dc.b	$23,$E1,$F9,$1F,$01,$1C,$13,$1F,$8B,$F3,$1F,$46,$7E,$93,$27,$D4
	.dc.b	$FA,$48,$FF,$7E,$ED,$3F,$05,$AE,$4C,$19,$FA,$1F,$ED,$B0,$6C,$61
	.dc.b	$CF,$54,$8F,$F1,$D4,$FE,$6F,$8B,$FD,$A8,$FE,$6E,$3E,$92,$9F,$11
	.dc.b	$23,$CC,$F9,$7E,$66,$47,$A5,$B9,$7F,$87,$EB,$FD,$A1,$3E,$0F,$96
	.dc.b	$F3,$D7,$A1,$91,$93,$CB,$18,$AF,$F7,$1D,$7F,$24,$7A,$79,$5F,$14
	.dc.b	$95,$F2,$5F,$87,$F2,$1F,$11,$B1,$F0,$3E,$2B,$E3,$9F,$23,$E1,$2C
	.dc.b	$64,$FA,$51,$F4,$48,$FF,$8C,$BE,$97,$F8,$9B,$F9,$77,$F8,$BB,$71
	.dc.b	$FF,$0F,$C6,$31,$FD,$1B,$C1,$BE,$96,$A7,$6C,$AF,$F3,$75,$71,$A3
	.dc.b	$C5,$B3,$DB,$E1,$3B,$FC,$25,$BE,$2D,$A7,$07,$E3,$1F,$1C,$66,$31
	.dc.b	$1D,$A3,$D3,$DF,$2B,$C7,$F5,$16,$B9,$C7,$C1,$9C,$5F,$2C,$97,$8A
	.dc.b	$F8,$DE,$77,$C4,$3F,$C8,$7F,$65,$FE,$EB,$8F,$7C,$65,$A7,$FA,$3E
	.dc.b	$35,$8F,$80,$F9,$73,$B7,$07,$C6,$7A,$FC,$AE,$71,$F2,$45,$BF,$E1
	.dc.b	$D7,$D2,$73,$1E,$0B,$E7,$8F,$F6,$98,$F1,$61,$C6,$7C,$31,$BA,$F9
	.dc.b	$37,$CE,$3D,$E6,$4B,$3E,$68,$D3,$C8,$EF,$87,$55,$1A,$7D,$93,$CD
	.dc.b	$E1,$5E,$0B,$E1,$87,$9E,$79,$D7,$A0,$E0,$F2,$32,$78,$E7,$8E,$1E
	.dc.b	$1C,$3C,$17,$CF,$0F,$8B,$7C,$1F,$06,$1D,$27,$4F,$BA,$33,$3B,$8C
	.dc.b	$11,$FE,$8B,$5D,$AF,$C3,$3E,$79,$E0,$7B,$9E,$0F,$C4,$C3,$6E,$8C
	.dc.b	$47,$8E,$7D,$53,$AB,$F4,$9E,$42,$CF,$4C,$3D,$37,$E4,$AE,$13,$E0
	.dc.b	$C3,$CC,$F5,$3C,$D7,$CA,$BC,$C4,$3C,$67,$96,$FE,$A3,$FE,$C4,$6B
	.dc.b	$F8,$43,$18,$7F,$5F,$8C,$C7,$E1,$4C,$D7,$F1,$19,$D1,$FC,$3B,$D1
	.dc.b	$FC,$09,$23,$EA,$36,$7C,$4D,$BF,$44,$7F,$0C,$8C,$4F,$C3,$6C,$7A
	.dc.b	$ED,$18,$8F,$C5,$57,$EF,$E3,$13,$F2,$A7,$A5,$78,$88,$F8,$9B,$5E
	.dc.b	$29,$EB,$18,$3C,$45,$9F,$15,$76,$BF,$81,$39,$9F,$1B,$FC,$CE,$74
	.dc.b	$FF,$2A,$6C,$7E,$F1,$B7,$E9,$9E,$57,$E9,$3C,$75,$E7,$24,$F9,$EB
	.dc.b	$8D,$77,$38,$DF,$92,$B7,$1E,$A6,$9E,$A7,$FC,$4B,$5F,$CC,$3E,$9F
	.dc.b	$FC,$4F,$18,$3F,$C4,$4F,$FB,$28,$CC,$FF,$1E,$70,$FF,$BB,$1B,$1C
	.dc.b	$63,$FC,$63,$1C,$7D,$6D,$1B,$F0,$6D,$6E,$95,$FE,$9A,$8C,$36,$3C
	.dc.b	$13,$C1,$F1,$78,$1E,$13,$F9,$DF,$CF,$5B,$E3,$1F,$E3,$9B,$FF,$5A
	.dc.b	$C7,$F8,$F6,$BD,$B3,$93,$5E,$DF,$AC,$E5,$1E,$F7,$DD,$EE,$73,$25
	.dc.b	$73,$9F,$7C,$61,$F8,$DE,$7F,$09,$F5,$BF,$1F,$E3,$24,$7B,$BC,$1B
	.dc.b	$5F,$36,$7D,$04,$9E,$D1,$F1,$A3,$FC,$A9,$F3,$9F,$E7,$98,$F6,$35
	.dc.b	$FF,$2D,$35,$FF,$31,$B5,$7D,$2F,$F8,$DE,$BA,$B6,$3D,$8D,$7E,$43
	.dc.b	$5F,$F3,$37,$C9,$3D,$EF,$F3,$33,$6B,$CE,$D9,$F8,$CD,$7E,$92,$FB
	.dc.b	$3E,$24,$9E,$2F,$F9,$C6,$DF,$43,$49,$3E,$87,$FC,$ED,$D0,$F0,$29
	.dc.b	$27,$E4,$C6,$5F,$7F,$CB,$F5,$23,$83,$E7,$FC,$CF,$FA,$9F,$3F,$BC
	.dc.b	$D7,$F1,$4D,$7F,$D1,$0D,$7F,$D2,$39,$FE,$41,$F8,$3C,$7F,$18,$95
	.dc.b	$F1,$B8,$FE,$49,$1C,$1F,$2B,$D0,$F9,$A4,$70,$7C,$EF,$37,$E8,$91
	.dc.b	$C1,$F8,$8F,$B3,$FC,$01,$3D,$8F,$F4,$E3,$F8,$3F,$A7,$7D,$33,$E9
	.dc.b	$BF,$1F,$18,$6F,$D3,$BE,$3F,$35,$BD,$BF,$45,$FA,$DD,$6D,$FE,$61
	.dc.b	$46,$5F,$D9,$35,$FF,$58,$2F,$B3,$E0,$D8,$FD,$0B,$5E,$F1,$AF,$36
	.dc.b	$46,$9A,$F9,$3C,$DF,$F5,$E7,$8B,$FE,$4E,$B4,$63,$7F,$0B,$6B,$FD
	.dc.b	$93,$2F,$99,$C4,$7A,$BE,$52,$C9,$A6,$3E,$2F,$3E,$84,$AF,$E3,$D4
	.dc.b	$8E,$3E,$87,$2F,$99,$E7,$F4,$92,$F7,$1D,$BD,$1E,$FE,$94,$9E,$71
	.dc.b	$E2,$F8,$2F,$1F,$D9,$49,$EF,$1E,$6F,$8A,$91,$E2,$7D,$3E,$97,$AB
	.dc.b	$D3,$93,$E6,$7D,$BF,$69,$EE,$F9,$39,$3E,$A7,$E0,$FC,$A7,$C4,$F8
	.dc.b	$0F,$E5,$9D,$1F,$E7,$A7,$F0,$D5,$FC,$F5,$AF,$F9,$99,$DF,$47,$FB
	.dc.b	$BD,$7C,$37,$3F,$85,$95,$C6,$3B,$7B,$1D,$FD,$99,$3D,$63,$C5,$E9
	.dc.b	$78,$FD,$29,$3E,$18,$F3,$7C,$27,$9F,$CC,$93,$E5,$8F,$57,$99,$EB
	.dc.b	$F3,$24,$FA,$63,$DD,$E6,$48,$F6,$3F,$07,$E1,$3E,$17,$B3,$27,$E0
	.dc.b	$3F,$17,$D9,$F8,$DF,$17,$27,$E2,$3F,$27,$8B,$F1,$BE,$8F,$C6,$E5
	.dc.b	$FF,$51,$B7,$97,$F9,$7F,$9B,$AE,$6F,$C9,$FD,$07,$C8,$CB,$FB,$53
	.dc.b	$5F,$F4,$9B,$70,$FF,$56,$FD,$03,$D5,$FC,$DA,$F9,$0A,$4B,$FE,$E9
	.dc.b	$0F,$98,$F2,$F4,$C9,$FE,$F1,$AA,$39,$FC,$D4,$9E,$2D,$7F,$D4,$B6
	.dc.b	$F8,$4F,$D5,$FC,$FB,$CE,$FD,$6F,$D0,$3E,$06,$9A,$8D,$F9,$F7,$1F
	.dc.b	$F0,$45,$F1,$6C,$7E,$DB,$BF,$E0,$72,$A6,$79,$3E,$17,$FC,$20,$D7
	.dc.b	$FD,$73,$3C,$1E,$0B,$97,$A7,$87,$27,$FC,$31,$39,$7F,$C3,$5F,$3F
	.dc.b	$FC,$31,$EB,$CB,$FC,$23,$0F,$A1,$1A,$F5,$3F,$85,$D7,$F0,$D8,$C1
	.dc.b	$0F,$F6,$5B,$96,$FA,$7C,$3D,$49,$1E,$08,$E7,$BF,$7F,$88,$93,$E4
	.dc.b	$8E,$FB,$EA,$30,$7A,$1F,$1F,$A9,$FE,$F6,$8F,$F0,$E6,$28,$C4,$7F
	.dc.b	$8A,$63,$1D,$7F,$E1,$46,$C7,$0F,$4D,$F4,$19,$6C,$66,$30,$7F,$45
	.dc.b	$5F,$41,$48,$FF,$49,$7F,$A5,$78,$0E,$5F,$F2,$44,$FF,$53,$6B,$C8
	.dc.b	$7D,$07,$F8,$55,$FC,$11,$B3,$FD,$EE,$7F,$10,$8E,$7C,$BD,$63,$35
	.dc.b	$73,$C4,$7E,$2E,$71,$5F,$E4,$BE,$6B,$3D,$9A,$B9,$3F,$C3,$33,$E8
	.dc.b	$9C,$CF,$EA,$63,$D7,$DB,$89,$CB,$AF,$36,$4E,$8F,$82,$F6,$C7,$FA
	.dc.b	$94,$7F,$9A,$DB,$67,$D3,$87,$EF,$E7,$5C,$9C,$78,$E4,$60,$E4,$FA
	.dc.b	$BE,$83,$E4,$5F,$C2,$6C,$FC,$CB,$77,$EF,$6E,$1B,$4A,$FF,$3F,$5F
	.dc.b	$0C,$FF,$68,$FD,$17,$CD,$1F,$5A,$3F,$C5,$18,$F7,$F2,$BC,$5E,$6F
	.dc.b	$8D,$95,$D1,$CB,$4F,$F4,$67,$87,$FD,$F9,$F0,$F9,$5F,$52,$31,$8F
	.dc.b	$80,$F1,$C7,$BC,$F8,$1E,$55,$AF,$4B,$87,$D7,$E5,$5A,$E4,$33,$47
	.dc.b	$E0,$BE,$83,$F8,$F5,$7F,$A7,$DB,$5C,$1F,$F0,$24,$67,$87,$87,$A3
	.dc.b	$23,$87,$A3,$6F,$BE,$5F,$1A,$7F,$3A,$D8,$FE,$40,$B5,$FC,$9B,$98
	.dc.b	$7D,$03,$DC,$FC,$BB,$F1,$38,$7F,$3E,$E7,$0F,$F5,$B3,$97,$F5,$6B
	.dc.b	$D7,$E3,$EB,$F2,$6C,$FF,$AF,$7E,$83,$FD,$75,$7F,$AA,$7F,$D0,$8D
	.dc.b	$EA,$FD,$3C,$62,$7F,$6F,$F4,$1F,$D4,$5F,$F6,$32,$F1,$7F,$5A,$E7
	.dc.b	$4F,$A2,$E6,$7F,$D9,$9F,$41,$FA,$2C,$7F,$72,$FD,$06,$F6,$FE,$DA
	.dc.b	$31,$3F,$C1,$96,$99,$F6,$B4,$D9,$FF,$6C,$7C,$FB,$E8,$3F,$7E,$D8
	.dc.b	$FF,$6E,$6F,$FC,$5D,$DC,$FF,$04,$8F,$D9,$69,$F8,$86,$E7,$FC,$39
	.dc.b	$F6,$37,$D4,$FE,$17,$F4,$1F,$4E,$BF,$1E,$D9,$FE,$2A,$FD,$58,$FE
	.dc.b	$C0,$E7,$0F,$E4,$3F,$41,$FA,$FA,$FE,$36,$FF,$5D,$8F,$F7,$93,$97
	.dc.b	$7E,$2F,$F1,$A8,$FC,$DC,$6B,$D9,$C7,$6E,$5F,$E3,$FF,$41,$FE,$8A
	.dc.b	$9F,$F7,$E2,$F9,$AC,$FC,$46,$9F,$E4,$BF,$40,$F8,$DA,$7F,$28,$FA
	.dc.b	$0F,$F5,$86,$BF,$CC,$27,$58,$FF,$7B,$8F,$E6,$72,$F8,$7E,$07,$1F
	.dc.b	$32,$D6,$CF,$E8,$3F,$41,$ED,$E2,$FF,$02,$35,$3F,$0D,$FD,$DB,$9F
	.dc.b	$9F,$C4,$7F,$4A,$0E,$1F,$E9,$AD,$7F,$C2,$0D,$77,$E7,$FF,$4A,$76
	.dc.b	$D7,$6D,$77,$5F,$EE,$AF,$2B,$FA,$5F,$C0,$1F,$E9,$B7,$FD,$26,$BF
	.dc.b	$88,$CE,$6B,$E0,$47,$8B,$CD,$D9,$C3,$E1,$6C,$EA,$7D,$73,$E1,$BF
	.dc.b	$F6,$C4,$7B,$DA,$FB,$A3,$58,$4A,$F8,$93,$F0,$23,$0E,$2B,$C8,$FF
	.dc.b	$26,$5E,$3F,$AA,$7D,$6F,$B3,$E0,$FD,$BF,$2F,$70,$F3,$5A,$33,$1E
	.dc.b	$0C,$3E,$9C,$7A,$49,$19,$CF,$BB,$FE,$58,$C3,$8D,$7C,$87,$FB,$BF
	.dc.b	$92,$F9,$FA,$62,$A9,$F1,$E3,$4F,$85,$B5,$F0,$9A,$FE,$1D,$AF,$F8
	.dc.b	$49,$FD,$24,$71,$E6,$FC,$04,$62,$79,$46,$0F,$2F,$D5,$79,$9B,$5E
	.dc.b	$3F,$9F,$F4,$FD,$07,$A9,$B1,$E7,$8E,$A4,$7D,$0D,$BF,$C4,$F6,$F0
	.dc.b	$6C,$FB,$FE,$83,$E2,$9F,$1B,$CB,$B1,$92,$4F,$F6,$2C,$FF,$99,$36
	.dc.b	$BE,$6F,$CC,$FC,$EE,$DB,$7F,$2D,$B5,$C5,$BD,$9F,$E3,$0C,$3F,$80
	.dc.b	$CF,$F6,$EF,$9F,$FE,$3C,$9A,$7F,$1F,$7F,$D0,$0E,$3E,$0F,$F5,$8F
	.dc.b	$F8,$02,$3F,$74,$D7,$FC,$8A,$D7,$8C,$7E,$C3,$C2,$FF,$1D,$CF,$83
	.dc.b	$91,$18,$F8,$19,$9F,$EA,$EB,$FB,$DE,$1E,$79,$CB,$67,$5F,$05,$8C
	.dc.b	$FF,$C9,$B7,$FE,$47,$9E,$5D,$FE,$0F,$03,$C2,$C8,$C9,$DE,$31,$87
	.dc.b	$C5,$F3,$7C,$E7,$A3,$CD,$B5,$E5,$8B,$3F,$2E,$6E,$3E,$29,$B5,$F1
	.dc.b	$26,$D7,$CC,$9A,$FD,$18,$B8,$A6,$C6,$1B,$5E,$12,$57,$9D,$F9,$13
	.dc.b	$B7,$37,$F1,$23,$13,$B3,$D1,$3A,$3E,$CD,$76,$8D,$57,$AE,$75,$1F
	.dc.b	$84,$AE,$73,$98,$E6,$70,$9E,$67,$07,$DC,$DB,$F7,$36,$B9,$BC,$27
	.dc.b	$9D,$75,$24,$65,$F6,$3C,$EF,$D1,$1A,$9F,$9D,$3D,$4D,$15,$F3,$1A
	.dc.b	$29,$E6,$2D,$F9,$9B,$7E,$67,$85,$78,$9D,$DF,$8A,$FC,$C4,$8D,$BF
	.dc.b	$83,$3C,$6F,$F4,$0B,$EC,$FE,$B2,$7F,$5E,$49,$FE,$6D,$3A,$9F,$8C
	.dc.b	$6E,$33,$3F,$D7,$BA,$FF,$1B,$B7,$F3,$F9,$FA,$4D,$AF,$7B,$70,$F9
	.dc.b	$10,$8C,$4F,$A6,$31,$5D,$09,$1D,$1B,$1D,$1B,$5D,$2B,$9C,$FC,$20
	.dc.b	$74,$AE,$44,$79,$36,$79,$57,$48,$F2,$57,$4A,$FE,$35,$7F,$06,$78
	.dc.b	$1A,$AE,$13,$E1,$3C,$07,$62,$5F,$69,$E3,$1B,$BE,$D8,$7E,$0E,$BF
	.dc.b	$DB,$71,$8C,$3C,$15,$CA,$FC,$07,$27,$CA,$FF,$81,$E7,$83,$70,$F2
	.dc.b	$CE,$8F,$F0,$9B,$8C,$3C,$13,$FE,$F2,$AE,$35,$A8,$ED,$3A,$8F,$9B
	.dc.b	$3A,$C3,$CF,$7C,$F0,$F3,$C7,$FB,$D3,$E7,$BF,$D9,$97,$F8,$1F,$88
	.dc.b	$FF,$78,$C6,$30,$46,$3D,$EF,$F1,$5B,$73,$F6,$3C,$3D,$5E,$3F,$0F
	.dc.b	$67,$FC,$4D,$87,$FA,$57,$1F,$C0,$B5,$34,$FF,$4F,$C7,$E1,$A3,$33
	.dc.b	$E4,$F6,$24,$F4,$F1,$6D,$7F,$B2,$AF,$FD,$71,$3F,$E1,$D6,$C7,$F8
	.dc.b	$B9,$AF,$F9,$4D,$AF,$77,$7A,$7D,$EF,$4B,$AB,$6B,$D7,$DB,$FE,$4F
	.dc.b	$AF,$F3,$ED,$F5,$8C,$1F,$E5,$3F,$9E,$FF,$4D,$47,$FB,$79,$B7,$F4
	.dc.b	$65,$7F,$E4,$B9,$CB,$E3,$FA,$DE,$C4,$7D,$8F,$F9,$7F,$EB,$7F,$C4
	.dc.b	$18,$7F,$9B,$5A,$FC,$96,$BF,$B0,$6A,$FA,$07,$D1,$7F,$9A,$5B,$7F
	.dc.b	$E4,$56,$CF,$F9,$59,$AF,$F9,$EF,$8D,$C1,$FE,$71,$6D,$7F,$92,$63
	.dc.b	$07,$C3,$B9,$1F,$E5,$1C,$B9,$DF,$DF,$6B,$FC,$1B,$07,$D0,$F9,$49
	.dc.b	$5E,$57,$E7,$7C,$EC,$8E,$8D,$AF,$47,$C7,$7B,$49,$3E,$D7,$D1,$E7
	.dc.b	$C8,$FF,$30,$36,$7F,$CE,$BA,$7F,$9E,$9F,$D6,$F1,$FB,$C9,$5F,$E6
	.dc.b	$7E,$4F,$84,$47,$07,$BF,$C9,$FC,$21,$1C,$1F,$E6,$FF,$94,$F9,$44
	.dc.b	$70,$7C,$7F,$2F,$F1,$84,$70,$7F,$9E,$FE,$79,$EB,$3E,$7B,$0D,$CF
	.dc.b	$FD,$1F,$2D,$DD,$BA,$BF,$F4,$27,$D6,$B6,$FF,$43,$60,$8C,$3A,$8C
	.dc.b	$C6,$23,$FD,$37,$F3,$CF,$9F,$E2,$D8,$FF,$48,$35,$DC,$76,$F5,$7F
	.dc.b	$D2,$D1,$FE,$95,$6B,$FE,$A6,$6B,$FE,$CB,$6B,$FC,$1D,$FE,$2E,$F1
	.dc.b	$F0,$7A,$11,$F0,$F9,$7F,$C3,$7B,$FF,$63,$1F,$D5,$30,$4E,$78,$3F
	.dc.b	$89,$46,$27,$FD,$80,$BE,$53,$51,$C7,$E1,$2A,$94,$61,$B1,$CB,$E2
	.dc.b	$B1,$E6,$48,$EA,$EF,$CD,$E1,$F2,$7C,$7E,$7E,$57,$8A,$3A,$7C,$FF
	.dc.b	$5F,$3E,$4F,$9A,$3C,$1E,$87,$87,$E7,$E4,$FA,$A3,$C9,$F0,$F2,$3D
	.dc.b	$AF,$9F,$D2,$F4,$7A,$12,$7E,$0B,$EB,$F9,$FF,$67,$C7,$C9,$F8,$AF
	.dc.b	$BF,$E3,$FE,$1F,$4F,$FC,$01,$E8,$F9,$F4,$8F,$C8,$BF,$EE,$B6,$A8
	.dc.b	$DC,$71,$E7,$FB,$DF,$27,$D3,$E2,$65,$F1,$8E,$FF,$09,$E0,$F5,$A4
	.dc.b	$F5,$8F,$1F,$A3,$E4,$F9,$F9,$3E,$18,$F3,$F9,$BE,$8F,$0E,$4F,$96
	.dc.b	$3D,$7E,$67,$B3,$E7,$E4,$FA,$63,$DF,$E6,$48,$F6,$3F,$0B,$E1,$BE
	.dc.b	$1F,$C2,$C9,$F8,$0F,$C6,$F6,$3E,$3F,$BB,$27,$E2,$3F,$2B,$8F,$F1
	.dc.b	$BE,$97,$C6,$E5,$FF,$09,$F9,$B8,$33,$CB,$F9,$2F,$9F,$73,$52,$65
	.dc.b	$FE,$32,$BF,$A7,$F3,$EF,$57,$F3,$2B,$E3,$E9,$2B,$87,$57,$F7,$59
	.dc.b	$7C,$BF,$93,$AA,$57,$F8,$7D,$E7,$F3,$12,$38,$9F,$E4,$2A,$46,$63
	.dc.b	$F0,$17,$D7,$E7,$FF,$57,$F3,$CF,$37,$F5,$BF,$3F,$FB,$EB,$D9,$FE
	.dc.b	$4C,$71,$B7,$E3,$AB,$84,$F8,$02,$F3,$BA,$DC,$9D,$4E,$5E,$1D,$FF
	.dc.b	$29,$E5,$E4,$32,$FD,$FB,$63,$7C,$5F,$C0,$A9,$C9,$62,$4F,$48,$FE
	.dc.b	$0F,$F3,$FF,$C3,$63,$D6,$2F,$EA,$3E,$0E,$4D,$FA,$7C,$3D,$3B,$3E
	.dc.b	$DB,$E5,$BF,$EB,$9F,$E2,$D1,$9F,$4F,$83,$BA,$F0,$C7,$8B,$7F,$5F
	.dc.b	$C5,$CA,$E9,$1E,$6E,$1D,$46,$0F,$03,$EA,$F4,$FF,$71,$1F,$E3,$93
	.dc.b	$FC,$FF,$8A,$33,$19,$8C,$65,$B8,$DC,$E9,$FF,$21,$4B,$D2,$7C,$FE
	.dc.b	$E5,$39,$8C,$36,$3A,$35,$E1,$F3,$FF,$D7,$13,$48,$F5,$EE,$FE,$8B
	.dc.b	$8F,$F1,$88,$FF,$33,$9F,$E9,$A9,$FF,$4D,$FC,$FF,$F8,$8A,$3F,$CB
	.dc.b	$15,$C3,$A7,$E1,$72,$D9,$FF,$2E,$35,$FF,$51,$CD,$7F,$6F,$3F,$97
	.dc.b	$7F,$12,$DE,$5E,$1C,$BF,$E2,$48,$C1,$BE,$CE,$47,$39,$B1,$CF,$CC
	.dc.b	$EC,$62,$FF,$1E,$7F,$9C,$0F,$E0,$9C,$9E,$5D,$7F,$AC,$D7,$E8,$3B
	.dc.b	$F1,$39,$E4,$93,$FA,$DA,$DD,$FE,$69,$FD,$C4,$7E,$C7,$E7,$FF,$D1
	.dc.b	$E7,$FA,$43,$4F,$E0,$31,$9D,$7E,$24,$FC,$2B,$FD,$1B,$6F,$FA,$4B
	.dc.b	$CB,$6D,$F9,$9C,$3C,$AB,$8F,$77,$0D,$3C,$51,$8C,$3F,$5F,$E1,$7F
	.dc.b	$80,$27,$CD,$E4,$FD,$58,$C4,$F8,$A3,$CA,$79,$BC,$9B,$5E,$D7,$0F
	.dc.b	$BF,$C9,$B5,$EA,$7F,$01,$F3,$FF,$A3,$9F,$E2,$8F,$E7,$5B,$D7,$AE
	.dc.b	$BF,$C3,$61,$FE,$E8,$9F,$E3,$4F,$FA,$E0,$C6,$1F,$C2,$1C,$79,$9C
	.dc.b	$3C,$CE,$2D,$AF,$E4,$5F,$3F,$FE,$19,$6F,$83,$F9,$38,$FB,$3C,$7F
	.dc.b	$DD,$9D,$63,$FC,$51,$FF,$67,$26,$8F,$4B,$97,$A7,$F9,$DF,$9F,$7A
	.dc.b	$54,$C7,$A5,$73,$FE,$69,$F3,$EF,$93,$FD,$23,$63,$3F,$43,$89,$FD
	.dc.b	$47,$D0,$7F,$39,$9F,$E7,$8D,$D7,$F9,$4B,$F3,$5B,$E4,$F0,$D6,$57
	.dc.b	$0F,$62,$D7,$F3,$0F,$69,$F3,$1F,$B5,$7F,$D4,$3C,$5F,$E1,$87,$37
	.dc.b	$FD,$2D,$CB,$FE,$D7,$5E,$6F,$1D,$39,$36,$7F,$A9,$7C,$FF,$CC,$BF
	.dc.b	$F1,$1B,$F4,$23,$FD,$C0,$E5,$BE,$37,$F0,$7F,$9F,$79,$74,$C7,$95
	.dc.b	$6B,$FA,$F4,$FF,$8B,$5B,$ED,$71,$F9,$4E,$4F,$F1,$1F,$9F,$7C,$65
	.dc.b	$31,$F1,$D6,$FF,$B4,$7C,$F7,$F9,$ED,$BF,$31,$FC,$66,$31,$C9,$FD
	.dc.b	$61,$46,$7E,$13,$5C,$96,$F9,$BF,$C8,$7E,$7B,$D7,$1F,$DD,$5F,$E2
	.dc.b	$58,$1C,$BF,$D9,$1C,$FD,$23,$C5,$FE,$69,$21,$19,$AF,$E8,$CD,$8F
	.dc.b	$E9,$CD,$7F,$AB,$7C,$FB,$C0,$E7,$08,$C5,$7F,$4F,$6C,$7F,$57,$6B
	.dc.b	$CB,$CE,$FE,$79,$1F,$EA,$99,$FF,$08,$3C,$3A,$3F,$CC,$2E,$6B,$FB
	.dc.b	$FA,$F3,$78,$9C,$B2,$B7,$CA,$3C,$B3,$B8,$F2,$46,$0C,$F1,$C8,$F4
	.dc.b	$36,$FD,$0F,$9B,$93,$CE,$D9,$FF,$13,$68,$F3,$7F,$AB,$70,$A7,$FA
	.dc.b	$C4,$E5,$FF,$16,$46,$5A,$7C,$A7,$18,$7F,$7D,$6B,$FE,$05,$6B,$F6
	.dc.b	$4F,$C4,$F6,$F8,$0F,$0C,$D3,$BF,$C9,$F9,$0F,$2F,$C9,$D8,$C9,$FE
	.dc.b	$05,$B4,$F8,$C9,$19,$F9,$FF,$31,$1C,$13,$0F,$F1,$9D,$3E,$7F,$E1
	.dc.b	$D7,$C4,$7E,$1E,$BF,$1B,$C0,$B8,$7C,$47,$0F,$8D,$F8,$84,$70,$7C
	.dc.b	$AF,$07,$78,$F8,$1F,$A3,$F3,$F2,$7E,$7F,$C1,$DA,$F2,$CD,$E5,$FF
	.dc.b	$56,$E9,$FE,$55,$BF,$AB,$3F,$E0,$FF,$9E,$FD,$AC,$7F,$98,$DF,$EB
	.dc.b	$B5,$FE,$B4,$73,$7F,$E5,$27,$33,$F5,$5C,$CE,$FE,$0B,$A6,$1D,$5E
	.dc.b	$DE,$44,$BF,$C5,$FC,$F5,$C7,$A4,$FE,$BB,$1F,$E7,$11,$FE,$47,$6B
	.dc.b	$DE,$34,$7A,$7D,$5E,$59,$79,$DF,$67,$C8,$4C,$7D,$9F,$21,$70,$F5
	.dc.b	$4D,$BF,$74,$DA,$F6,$7C,$87,$C5,$3F,$C6,$9F,$47,$FC,$F3,$0F,$E1
	.dc.b	$99,$A7,$55,$F2,$5D,$FC,$E6,$2B,$FC,$F3,$F3,$FF,$DD,$32,$FF,$6F
	.dc.b	$B6,$FF,$3D,$8B,$A7,$FC,$85,$39,$BF,$E3,$4D,$4D,$9E,$15,$FE,$9B
	.dc.b	$8D,$12,$7D,$35,$D1,$93,$B7,$2D,$23,$E8,$32,$57,$37,$EA,$8C,$B6
	.dc.b	$32,$7E,$E6,$BC,$D5,$E0,$AE,$44,$AF,$F8,$03,$7E,$A8,$C5,$F1,$78
	.dc.b	$4E,$AA,$CE,$4E,$F3,$FC,$FE,$B9,$57,$8E,$39,$9F,$28,$F3,$CF,$17
	.dc.b	$27,$CC,$32,$6D,$EA,$E5,$FF,$64,$4F,$67,$94,$F7,$23,$F7,$2B,$CF
	.dc.b	$5C,$CE,$F5,$C5,$C5,$70,$6D,$71,$3C,$37,$F6,$DF,$0D,$F4,$9F,$21
	.dc.b	$D1,$C4,$EE,$3F,$D9,$37,$DD,$DD,$F7,$6D,$F7,$7E,$1D,$71,$9F,$A1
	.dc.b	$3E,$29,$F0,$57,$3A,$D4,$FC,$23,$95,$75,$8C,$D7,$D5,$BF,$2B,$FD
	.dc.b	$B2,$3F,$CA,$EF,$FB,$15,$AF,$D5,$30,$6C,$95,$FE,$BF,$8D,$C7,$C8
	.dc.b	$3F,$4F,$F0,$3F,$08,$93,$F2,$9B,$1E,$E6,$DF,$9D,$B3,$EB,$9E,$6F
	.dc.b	$12,$3C,$5B,$3C,$5A,$71,$8F,$DB,$CF,$47,$CC,$62,$36,$E2,$93,$B8
	.dc.b	$FA,$A7,$88,$F0,$12,$BC,$0D,$AF,$05,$72,$7A,$C7,$23,$C0,$6C,$F0
	.dc.b	$57,$79,$FE,$12,$D8,$FE,$26,$D9,$FE,$1C,$D8,$FE,$3A,$DA,$FE,$3A
	.dc.b	$D9,$FE,$04,$D9,$F5,$4F,$A0,$93,$FC,$1F,$CD,$FE,$43,$E6,$FF,$22
	.dc.b	$6C,$7F,$1D,$6D,$7F,$2C,$6D,$7F,$0A,$FA,$4F,$E5,$3E,$2F,$F3,$86
	.dc.b	$DF,$F2,$58,$B7,$FA,$49,$B5,$FC,$D3,$DC,$FE,$62,$7F,$40,$ED,$B8
	.dc.b	$7F,$2D,$8F,$CC,$11,$FC,$93,$6F,$F9,$FF,$85,$C1,$B3,$FC,$D7,$BB
	.dc.b	$F4,$04,$7F,$34,$FE,$8F,$E5,$FF,$A0,$9F,$D0,$9B,$1F,$E5,$BC,$7F
	.dc.b	$C6,$15,$FE,$5D,$F9,$E7,$D0,$FF,$95,$9B,$5F,$E2,$99,$FF,$15,$46
	.dc.b	$1C,$07,$0E,$FE,$7F,$FD,$36,$E9,$AE,$7C,$5F,$F3,$1C,$62,$3F,$C6
	.dc.b	$DE,$3B,$E8,$7F,$CC,$ED,$C3,$BC,$A3,$0F,$89,$AF,$9F,$E5,$FF,$D8
	.dc.b	$B3,$C1,$B1,$E8,$6B,$BF,$77,$C7,$E9,$CA,$EF,$8A,$31,$5F,$03,$A9
	.dc.b	$E2,$FF,$9F,$1B,$7E,$66,$CE,$E7,$DF,$F2,$32,$BA,$B6,$7E,$5F,$BA
	.dc.b	$F9,$EF,$F4,$3B,$6B,$E0,$B6,$77,$3F,$E8,$CD,$3F,$D0,$11,$F0,$7A
	.dc.b	$70,$4B,$E3,$E4,$74,$23,$83,$E3,$7C,$27,$32,$38,$3E,$57,$A5,$DC
	.dc.b	$8E,$0F,$AB,$F4,$9F,$82,$C3,$FC,$E0,$D7,$FD,$54,$D4,$DF,$A1,$FF
	.dc.b	$0C,$67,$F7,$9B,$3F,$8E,$5F,$43,$FC,$72,$31,$E2,$7C,$77,$F2,$7F
	.dc.b	$3F,$5B,$7F,$88,$51,$97,$F3,$8D,$7F,$D0,$CB,$E8,$38,$36,$3F,$22
	.dc.b	$D7,$BC,$75,$E7,$79,$D4,$93,$F9,$88,$CB,$F9,$AF,$D7,$3F,$D8,$69
	.dc.b	$A4,$D7,$D0,$BA,$E9,$27,$93,$AF,$27,$FD,$3D,$DF,$FB,$C9,$E9,$B6
	.dc.b	$30,$FF,$51,$61,$F2,$A7,$F8,$14,$62,$F5,$1F,$ED,$F9,$FF,$25,$3A
	.dc.b	$31,$EA,$E9,$F4,$36,$37,$0C,$3C,$50,$F7,$B5,$D9,$2B,$A1,$F0,$63
	.dc.b	$97,$8F,$23,$34,$8E,$DE,$B7,$7F,$22,$56,$9F,$17,$6F,$8F,$E4,$E4
	.dc.b	$F2,$7C,$DF,$1D,$E7,$F6,$3D,$1E,$C4,$AE,$F1,$EB,$F0,$7D,$9F,$39
	.dc.b	$27,$C7,$1E,$FF,$67,$E0,$78,$92,$7C,$F1,$F0,$FC,$A9,$1E,$A7,$E4
	.dc.b	$7E,$BB,$F9,$5E,$4F,$AC,$FE,$2D,$BD,$9A,$37,$E2,$70,$F6,$A6,$1F
	.dc.b	$22,$D1,$D3,$E3,$BA,$FC,$34,$AD,$3E,$0F,$95,$F0,$FB,$B2,$79,$3E
	.dc.b	$4F,$77,$CB,$F2,$12,$7B,$3E,$8F,$A2,$F4,$FB,$DE,$AF,$7E,$57,$8E
	.dc.b	$3D,$BF,$2D,$EE,$F8,$79,$3E,$78,$F8,$3F,$0F,$F0,$BE,$06,$4F,$AE
	.dc.b	$3F,$19,$EA,$3E,$5E,$FC,$0C,$AF,$C3,$CE,$5F,$C4,$FC,$FB,$C2,$FC
	.dc.b	$A7,$D0,$3E,$62,$9F,$97,$9D,$57,$0F,$0F,$1E,$24,$8C,$1D,$BE,$0F
	.dc.b	$C6,$47,$BF,$87,$CC,$8E,$0F,$1F,$83,$FA,$5F,$9F,$75,$7E,$9F,$E8
	.dc.b	$1F,$07,$4D,$46,$FB,$2E,$1F,$A4,$BE,$2D,$8F,$D8,$7D,$07,$B6,$3F
	.dc.b	$53,$CE,$CB,$F6,$8D,$8F,$D6,$CE,$DF,$DA,$7A,$46,$47,$CB,$72,$63
	.dc.b	$87,$B9,$E0,$E6,$FE,$01,$F3,$FF,$90,$9F,$0F,$2F,$80,$CF,$85,$AF
	.dc.b	$0A,$CF,$EF,$AB,$83,$D8,$28,$EB,$E1,$71,$9E,$11,$87,$97,$07,$BB
	.dc.b	$CA,$91,$83,$B7,$07,$9A,$64,$60,$F1,$78,$7F,$C5,$5F,$F0,$CE,$28
	.dc.b	$C4,$7F,$87,$E3,$1E,$3F,$F8,$41,$B1,$C3,$C4,$7D,$06,$5B,$19,$8C
	.dc.b	$1F,$D4,$57,$D0,$52,$3F,$D5,$5F,$EA,$DE,$07,$F8,$D3,$0F,$E7,$53
	.dc.b	$FE,$71,$3E,$41,$F6,$67,$ED,$7D,$03,$93,$FC,$6D,$7C,$79,$7F,$C7
	.dc.b	$1C,$B7,$0F,$C0,$1F,$A5,$8F,$E6,$2F,$5C,$7B,$CE,$E3,$F0,$DC,$B5
	.dc.b	$FD,$A5,$7E,$43,$D4,$C4,$E7,$B6,$4F,$48,$B5,$E8,$F8,$BE,$46,$31
	.dc.b	$F6,$BD,$D7,$35,$3C,$5D,$2E,$FC,$63,$FD,$22,$BF,$5A,$DE,$5C,$E3
	.dc.b	$F8,$D9,$CB,$F8,$FF,$A0,$F7,$57,$C2,$6C,$7E,$8F,$07,$81,$F0,$67
	.dc.b	$32,$7E,$43,$CF,$7D,$70,$6D,$E1,$F0,$9C,$7E,$82,$E5,$FD,$7B,$04
	.dc.b	$66,$30,$EB,$0F,$33,$F4,$7C,$3F,$E7,$B3,$C7,$E4,$71,$DF,$E5,$64
	.dc.b	$75,$F8,$EB,$87,$BD,$F5,$1E,$BF,$8E,$B3,$E3,$7E,$F7,$D0,$7E,$86
	.dc.b	$3F,$B9,$AE,$DC,$F7,$E9,$FC,$25,$FF,$09,$AF,$EF,$2F,$C6,$31,$87
	.dc.b	$EF,$5C,$7C,$4F,$0F,$89,$E2,$DA,$FF,$00,$FD,$07,$F0,$16,$F8,$BF
	.dc.b	$91,$8F,$C1,$C6,$0D,$FD,$06,$4D,$7C,$ED,$8F,$F0,$5F,$D0,$7F,$9F
	.dc.b	$BA,$BF,$C4,$4D,$8F,$D0,$AF,$98,$FC,$F3,$9D,$7F,$92,$7D,$03,$DF
	.dc.b	$FD,$13,$6F,$FC,$2E,$BD,$FC,$CF,$E9,$FE,$81,$E6,$53,$1C,$97,$0F
	.dc.b	$F0,$FC,$B5,$E0,$71,$FA,$0B,$3F,$E2,$4F,$5F,$FD,$53,$BF,$F8,$ED
	.dc.b	$7C,$47,$ED,$5C,$CF,$EF,$1C,$BC,$7D,$0D,$69,$B6,$C7,$F8,$C3,$E8
	.dc.b	$3E,$87,$07,$F3,$37,$E7,$B7,$97,$F8,$14,$63,$2F,$F2,$67,$D0,$3C
	.dc.b	$CF,$E0,$CD,$C3,$FC,$76,$BC,$BC,$CE,$BE,$36,$56,$BE,$66,$E7,$FD
	.dc.b	$0F,$E8,$3F,$C8,$D3,$FE,$44,$6C,$67,$D0,$FE,$29,$18,$CF,$FC,$BD
	.dc.b	$F4,$1F,$E7,$7B,$FF,$25,$AF,$85,$FE,$34,$E7,$0F,$EA,$0E,$67,$7F
	.dc.b	$03,$AE,$3D,$B6,$3F,$CA,$BF,$41,$ED,$BF,$EB,$0F,$B6,$33,$B7,$F2
	.dc.b	$78,$C4,$FF,$5D,$F9,$F7,$87,$FC,$AF,$E8,$3F,$21,$87,$F9,$82,$72
	.dc.b	$FF,$98,$A3,$35,$FD,$95,$79,$DC,$4E,$5E,$15,$AE,$11,$FD,$1A,$BF
	.dc.b	$BE,$57,$FB,$09,$FE,$7E,$D7,$31,$B5,$0F,$2B,$0D,$AF,$F7,$7B,$67
	.dc.b	$FD,$FE,$D7,$FA,$59,$CA,$7D,$DE,$3F,$BD,$B5,$ED,$E5,$F8,$5C,$BF
	.dc.b	$0D,$B1,$F0,$63,$C2,$47,$C2,$F9,$F8,$91,$48,$78,$B0,$FF,$0E,$46
	.dc.b	$A3,$11,$F8,$3F,$93,$FE,$B9,$3F,$A8,$8D,$CE,$26,$3C,$27,$FD,$60
	.dc.b	$D8,$F1,$1E,$3F,$92,$91,$B6,$D7,$8E,$91,$E0,$F9,$2F,$0B,$6B,$FD
	.dc.b	$43,$2F,$90,$F2,$FC,$87,$99,$B3,$A8,$F8,$44,$7C,$CF,$C5,$EF,$91
	.dc.b	$E9,$6E,$3E,$3F,$03,$E0,$9F,$0A,$2C,$F8,$A2,$CF,$F8,$D7,$05,$63
	.dc.b	$AB,$FD,$99,$0F,$90,$FA,$84,$C3,$E4,$36,$F9,$7C,$87,$D0,$8F,$A2
	.dc.b	$FF,$83,$7E,$83,$FA,$EC,$FF,$90,$DB,$87,$FB,$65,$FD,$B4,$67,$D3
	.dc.b	$E0,$7E,$66,$38,$C7,$E4,$9A,$FF,$87,$9A,$FF,$97,$DF,$E6,$B8,$7F
	.dc.b	$B4,$1F,$EC,$11,$E1,$F2,$BC,$9E,$54,$AD,$3E,$78,$C4,$F9,$E2,$E1
	.dc.b	$E7,$9B,$5E,$1F,$8E,$F7,$7B,$7E,$F6,$C7,$85,$B8,$7C,$16,$DF,$C1
	.dc.b	$F4,$3E,$31,$F1,$FE,$36,$DF,$A6,$2D,$FC,$99,$B5,$FF,$00,$7E,$83
	.dc.b	$FC,$32,$49,$FF,$1E,$38,$F7,$29,$88,$B8,$7F,$BD,$27,$4D,$8C,$9F
	.dc.b	$E7,$98,$CB,$4C,$D7,$86,$AD,$7F,$94,$2B,$FC,$0B,$39,$8E,$D3,$AA
	.dc.b	$FF,$4A,$56,$A7,$84,$E6,$53,$EB,$9E,$24,$8F,$24,$FA,$E3,$15,$DE
	.dc.b	$73,$1E,$B3,$27,$62,$BE,$23,$C1,$5E,$E9,$F9,$91,$8A,$CB,$6B,$8B
	.dc.b	$E6,$30,$78,$0F,$8C,$EA,$BA,$3F,$16,$37,$5E,$73,$BC,$6E,$BB,$B4
	.dc.b	$E5,$3C,$1F,$53,$D8,$F0,$4F,$38,$CB,$D2,$33,$18,$23,$F0,$5E,$CE
	.dc.b	$8C,$14,$FA,$73,$E4,$71,$5E,$C7,$75,$E1,$7E,$EC,$FF,$51,$9F,$F4
	.dc.b	$E4,$7F,$4A,$6B,$EA,$31,$1B,$25,$7F,$3C,$9D,$CF,$48,$FF,$6E,$57
	.dc.b	$F4,$68,$FF,$6F,$7C,$77,$FB,$C2,$7E,$AB,$63,$D4,$D9,$F3,$42,$31
	.dc.b	$3F,$12,$31,$F1,$7F,$93,$9C,$C7,$84,$FF,$6E,$38,$7E,$2C,$F9,$0F
	.dc.b	$F5,$8C,$66,$FF,$D1,$AD,$C3,$FD,$60,$D9,$E1,$5E,$AB,$F0,$12,$3D
	.dc.b	$6D,$8F,$5B,$70,$F5,$D7,$39,$F2,$CE,$6B,$D9,$39,$AF,$91,$5E,$E9
	.dc.b	$EA,$47,$AB,$63,$AD,$73,$9F,$85,$1C,$AB,$A5,$7C,$0A,$D5,$7C,$43
	.dc.b	$C7,$1C,$CF,$1C,$6B,$0F,$8A,$DC,$3E,$2B,$67,$E2,$D7,$9C,$F2,$93
	.dc.b	$0F,$2C,$75,$8F,$2B,$E3,$30,$F8,$E3,$D1,$3E,$1F,$2F,$F9,$3F,$C2
	.dc.b	$7F,$28,$6D,$FF,$25,$6D,$FF,$23,$6D,$FF,$31,$F7,$7F,$97,$9F,$CC
	.dc.b	$3C,$3B,$7F,$CB,$62,$DF,$F2,$B9,$B5,$FC,$E7,$DC,$FE,$6E,$7F,$38
	.dc.b	$F0,$EE,$1F,$CD,$63,$F5,$44,$7F,$48,$FF,$8F,$7E,$83,$FC,$45,$82
	.dc.b	$FE,$EC,$61,$ED,$5F,$2A,$30,$FF,$37,$D3,$F6,$C4,$7F,$60,$FE,$E7
	.dc.b	$E8,$24,$7E,$CD,$B7,$FE,$33,$F9,$FF,$F4,$9B,$7D,$5F,$F2,$84,$62
	.dc.b	$3F,$C9,$2D,$7F,$4B,$BF,$F9,$57,$88,$D2,$61,$FE,$3C,$AF,$F4,$2D
	.dc.b	$72,$F0,$FA,$92,$75,$E9,$78,$08,$E0,$FF,$22,$E8,$F9,$8F,$90,$48
	.dc.b	$CB,$E8,$F8,$D9,$1E,$16,$E1,$FE,$51,$CF,$FC,$F2,$FF,$99,$1A,$FF
	.dc.b	$00,$F3,$3D,$87,$B7,$8E,$5F,$F9,$5A,$BF,$CE,$D5,$FC,$46,$7E,$17
	.dc.b	$24,$AD,$36,$7C,$1E,$FF,$C7,$24,$F3,$7E,$57,$9F,$23,$B3,$67,$FC
	.dc.b	$D1,$A7,$CA,$44,$70,$DD,$BF,$18,$7F,$A2,$7E,$75,$F3,$AF,$9D,$FF
	.dc.b	$37,$C6,$1B,$C0,$EB,$CD,$6F,$0F,$F3,$C7,$D1,$FF,$A7,$9F,$F3,$BB
	.dc.b	$5C,$EF,$FE,$7A,$51,$97,$FD,$23,$F3,$CF,$9B,$D3,$63,$FD,$02,$D7
	.dc.b	$FD,$0C,$D7,$FD,$33,$E9,$3D,$2E,$CD,$8F,$F4,$4B,$5F,$F4,$73,$5F
	.dc.b	$F5,$0F,$CE,$FF,$AA,$D3,$49,$C5,$E9,$F9,$C7,$8B,$49,$5F,$E9,$68
	.dc.b	$CB,$7E,$0B,$81,$1C,$1F,$E9,$CB,$77,$7D,$88,$FD,$F6,$16,$7F,$39
	.dc.b	$1F,$C9,$9A,$E8,$FB,$EF,$27,$5B,$F3,$F0,$7F,$A1,$CF,$FA,$6A,$79
	.dc.b	$7A,$99,$BE,$D7,$67,$9F,$35,$7E,$1F,$75,$C0,$D9,$C7,$CE,$CA,$F3
	.dc.b	$3B,$F5,$78,$7A,$B2,$7D,$4F,$3F,$07,$A7,$C8,$49,$F7,$3D,$FE,$33
	.dc.b	$C1,$EB,$78,$7D,$79,$5F,$0E,$3C,$9F,$0F,$E5,$F8,$F9,$3F,$1E,$3D
	.dc.b	$1F,$2F,$E9,$F6,$64,$FC,$B8,$F6,$7B,$32,$3E,$73,$F0,$7A,$7F,$3D
	.dc.b	$1F,$3C,$3F,$93,$E7,$FE,$DE,$DE,$6F,$B4,$D9,$FD,$9A,$F6,$BA,$7B
	.dc.b	$52,$7F,$B0,$CA,$3C,$1F,$0F,$E1,$F9,$29,$5A,$7C,$9F,$25,$E5,$F6
	.dc.b	$E4,$F2,$7D,$1E,$E7,$A7,$E3,$E4,$F6,$7D,$9F,$3B,$ED,$F9,$DF,$77
	.dc.b	$BD,$2B,$C7,$1F,$07,$DF,$F8,$5F,$0F,$27,$CF,$1F,$17,$E2,$3E,$37
	.dc.b	$C0,$C9,$F5,$C7,$E2,$7D,$87,$CA,$DF,$4A,$57,$F8,$02,$72,$FF,$80
	.dc.b	$BE,$71,$E1,$7E,$47,$E7,$5F,$2F,$4F,$CA,$4E,$AB,$87,$87,$8F,$12
	.dc.b	$46,$0E,$DF,$19,$E3,$23,$DF,$C3,$E6,$47,$07,$8F,$C1,$FD,$17,$CE
	.dc.b	$3A,$BF,$49,$F3,$AF,$84,$A6,$A3,$7D,$97,$0F,$F0,$ED,$F1,$6C,$7E
	.dc.b	$B7,$E0,$FD,$B1,$FE,$22,$E7,$65,$FB,$16,$C7,$F8,$9A,$76,$FE,$4D
	.dc.b	$F9,$2D,$E0,$E8,$75,$E5,$FD,$D7,$CE,$FE,$36,$7B,$72,$75,$33,$E0
	.dc.b	$EB,$C1,$B3,$FE,$34,$AE,$0F,$E8,$54,$74,$F0,$78,$4E,$E3,$0F,$1D
	.dc.b	$FD,$5E,$4C,$8C,$1D,$77,$F2,$CC,$8C,$1E,$1F,$0B,$FC,$40,$FF,$25
	.dc.b	$3F,$AE,$E2,$8C,$C6,$63,$1E,$36,$E3,$73,$A7,$F9,$74,$BC,$37,$CE
	.dc.b	$EE,$53,$98,$C3,$63,$A3,$5E,$1F,$3B,$FD,$75,$34,$8F,$5E,$FF,$F4
	.dc.b	$E6,$3F,$E7,$68,$FE,$D9,$E4,$7F,$98,$9F,$EF,$35,$FD,$EB,$E7,$7F
	.dc.b	$C5,$31,$FD,$16,$F8,$D7,$C3,$E4,$B3,$FD,$25,$AF,$F8,$0A,$6B,$FA
	.dc.b	$A9,$ED,$C9,$83,$7C,$DF,$CF,$23,$27,$8F,$B3,$89,$B7,$07,$9F,$E2
	.dc.b	$FA,$9A,$70,$7F,$9F,$34,$72,$F9,$1B,$1F,$9F,$EE,$C4,$E7,$8E,$4F
	.dc.b	$18,$B9,$FF,$31,$F9,$DF,$8E,$DD,$7D,$3D,$FF,$02,$73,$97,$FA,$89
	.dc.b	$F8,$15,$C7,$7D,$5E,$DE,$1F,$01,$C7,$E1,$EE,$7F,$E9,$4C,$11,$98
	.dc.b	$C3,$AC,$3C,$8F,$D0,$F0,$BF,$9D,$4F,$1F,$90,$C7,$6F,$95,$91,$D7
	.dc.b	$E3,$6E,$1E,$D7,$D4,$7A,$FE,$36,$CF,$8D,$FB,$BF,$3A,$F3,$E9,$8F
	.dc.b	$3A,$D7,$FA,$B2,$FE,$0F,$0F,$FB,$11,$B1,$FE,$0F,$7F,$9A,$BB,$E0
	.dc.b	$FC,$4F,$C3,$B6,$A6,$39,$AD,$7F,$AE,$3E,$77,$F9,$23,$79,$FF,$23
	.dc.b	$0F,$0A,$C6,$3E,$6A,$57,$1F,$A3,$BC,$5F,$DB,$7E,$D5,$5F,$9A,$6D
	.dc.b	$7F,$B1,$97,$C2,$66,$7F,$3F,$F3,$CE,$CA,$63,$CC,$B8,$7F,$B3,$BE
	.dc.b	$7B,$F8,$03,$7C,$8F,$D3,$C7,$CB,$8C,$1F,$ED,$43,$ED,$E7,$FE,$01
	.dc.b	$6F,$2F,$FB,$B9,$FE,$F8,$6F,$AF,$F6,$3E,$A7,$FA,$4E,$3F,$DB,$AD
	.dc.b	$AF,$F0,$53,$FE,$70,$8C,$FC,$6F,$0F,$76,$47,$0D,$2D,$FF,$84,$7E
	.dc.b	$D1,$E1,$7E,$F9,$B5,$FE,$EA,$5F,$27,$99,$FE,$07,$F3,$DF,$EB,$2A
	.dc.b	$FF,$77,$B6,$FF,$C3,$8B,$E9,$37,$EA,$70,$F5,$A5,$F0,$F9,$7B,$87
	.dc.b	$F8,$8F,$D0,$7C,$0F,$F1,$06,$D7,$FB,$F1,$7C,$16,$67,$F8,$AF,$CF
	.dc.b	$7C,$AA,$FF,$30,$36,$FF,$D3,$2F,$F4,$E8,$FF,$4D,$F2,$9D,$F2,$F2
	.dc.b	$F7,$E4,$68,$E7,$F1,$D8,$C3,$FC,$71,$E7,$BD,$1F,$E4,$AD,$AE,$13
	.dc.b	$FC,$9F,$D2,$CD,$7F,$2A,$F9,$D7,$A1,$FC,$B3,$E7,$5F,$31,$98,$C3
	.dc.b	$96,$63,$D3,$15,$70,$8C,$3D,$1B,$3F,$D0,$E7,$FD,$73,$39,$7F,$AD
	.dc.b	$35,$FE,$C4,$D7,$FD,$10,$BE,$4B,$94,$78,$1B,$19,$AF,$EB,$CD,$9E
	.dc.b	$B3,$ED,$F9,$3F,$73,$6B,$D5,$8B,$D4,$FE,$9B,$F1,$3F,$4E,$73,$18
	.dc.b	$61,$FD,$AE,$35,$87,$F7,$36,$BF,$DE,$DA,$FF,$A8,$5F,$D2,$3B,$F9
	.dc.b	$1F,$5C,$62,$38,$46,$3E,$4B,$C6,$79,$1B,$5F,$46,$93,$B8,$F2,$36
	.dc.b	$BC,$9F,$1D,$D8,$93,$E9,$6C,$7F,$A0,$69,$F1,$DF,$0A,$7E,$1B,$F8
	.dc.b	$7D,$3E,$2F,$C1,$DC,$3C,$4E,$1F,$88,$FB,$08,$FC,$7F,$2B,$DE,$47
	.dc.b	$07,$CB,$EB,$FA,$27,$D2,$8B,$1E,$F8,$B9,$79,$E6,$F4,$FE,$A6,$03
	.dc.b	$E7,$3F,$8C,$BF,$E2,$76,$BF,$E2,$F6,$BF,$ED,$56,$BC,$9E,$31,$D6
	.dc.b	$3A,$75,$7F,$84,$A7,$38,$71,$8C,$7A,$52,$3F,$1B,$F4,$56,$FF,$D5
	.dc.b	$D7,$FD,$26,$7F,$C6,$8D,$8F,$35,$7E,$7A,$31,$1E,$7F,$6B,$C1,$F1
	.dc.b	$DE,$16,$DF,$EC,$E9,$38,$AE,$F5,$72,$F5,$56,$AF,$85,$F3,$69,$FB
	.dc.b	$DF,$A4,$FF,$7C,$BD,$7B,$FF,$C6,$31,$8C,$7D,$71,$8F,$96,$F8,$8D
	.dc.b	$AE,$07,$F0,$7F,$9E,$FA,$AD,$AE,$7C,$FF,$E3,$C9,$CC,$7F,$B4,$DB
	.dc.b	$9F,$D1,$6D,$7D,$E9,$63,$FA,$CA,$D4,$73,$EF,$95,$EB,$E8,$B3,$D6
	.dc.b	$6D,$7D,$DB,$FD,$5D,$F9,$E7,$2F,$7C,$11,$F0,$B0,$CE,$1F,$90,$9C
	.dc.b	$D7,$13,$C0,$58,$F0,$37,$0F,$05,$72,$95,$72,$76,$7F,$BD,$70,$F2
	.dc.b	$B8,$AF,$89,$87,$19,$CC,$63,$0E,$C7,$A6,$7D,$0E,$2B,$D1,$1C,$70
	.dc.b	$F1,$61,$E1,$BE,$55,$E1,$9C,$D6,$EF,$94,$FA,$23,$0F,$96,$3C,$06
	.dc.b	$B0,$F2,$B8,$8D,$1E,$27,$E3,$B8,$3A,$3C,$6B,$C2,$D3,$C3,$87,$9D
	.dc.b	$E9,$87,$CC,$73,$87,$92,$3D,$64,$8D,$9F,$8B,$C3,$CB,$87,$8C,$F1
	.dc.b	$5F,$57,$F8,$6C,$69,$F2,$9A,$BF,$9A,$DA,$F3,$34,$F6,$4F,$1A,$E9
	.dc.b	$1C,$AF,$C8,$F6,$C3,$DE,$FF,$AF,$E7,$F0,$8F,$FB,$1D,$AF,$CB,$31
	.dc.b	$87,$F6,$68,$CC,$7A,$CC,$E1,$FE,$C3,$8F,$09,$FE,$07,$8F,$7F,$7F
	.dc.b	$E0,$89,$3F,$39,$B1,$A8,$C1,$1C,$13,$0F,$5B,$F9,$C9,$E3,$18,$AF
	.dc.b	$03,$63,$35,$F5,$2B,$F5,$91,$87,$EA,$4F,$F0,$38,$C4,$7C,$99,$FD
	.dc.b	$B4,$62,$BD,$AD,$BF,$E1,$CD,$9F,$E2,$8D,$AF,$E2,$0D,$BE,$D3,$ED
	.dc.b	$7C,$CD,$3C,$D3,$ED,$70,$F9,$9F,$19,$87,$CC,$7B,$1C,$5F,$99,$B7
	.dc.b	$E6,$69,$E6,$2C,$F9,$9E,$98,$BE,$53,$F9,$57,$97,$FC,$D9,$B3,$FC
	.dc.b	$ED,$B5,$FC,$2D,$B8,$FF,$2D,$F7,$FF,$99,$1F,$CF,$BA,$AD,$FF,$39
	.dc.b	$8B,$7F,$C4,$E6,$D7,$F3,$5E,$AF,$E7,$47,$F4,$6F,$8C,$B8,$7F,$41
	.dc.b	$8F,$CA,$11,$FC,$93,$F9,$7F,$2B,$FA,$11,$C3,$0F,$E8,$4F,$F5,$0E
	.dc.b	$CF,$F2,$0E,$A6,$92,$FF,$A2,$57,$F4,$5F,$9C,$FF,$3B,$B8,$0E,$1D
	.dc.b	$B5,$E1,$E4,$FF,$93,$BC,$37,$97,$49,$7C,$EB,$FC,$A9,$1F,$E5,$4F
	.dc.b	$92,$FF,$47,$4E,$5B,$3E,$16,$BE,$66,$BF,$D6,$FE,$53,$FD,$86,$F9
	.dc.b	$DA,$FB,$1A,$EF,$DA,$F8,$7E,$1C,$BF,$26,$BF,$E9,$46,$C7,$F9,$A3
	.dc.b	$E7,$9C,$34,$93,$EB,$8C,$BE,$DF,$59,$F0,$7F,$E7,$06,$C6,$63,$07
	.dc.b	$C1,$E8,$91,$F0,$5B,$9F,$D1,$AF,$F1,$6A,$69,$3C,$8F,$95,$F3,$D2
	.dc.b	$FE,$53,$6B,$ED,$7D,$27,$B5,$BA,$7F,$A1,$23,$11,$F8,$26,$BF,$E0
	.dc.b	$77,$EF,$3F,$E4,$66,$A6,$98,$F9,$F7,$83,$F0,$A7,$F1,$51,$8D,$3F
	.dc.b	$04,$BD,$B7,$A5,$8B,$E1,$7F,$94,$F9,$DA,$DC,$3F,$14,$A3,$2F,$F8
	.dc.b	$75,$AF,$F9,$71,$7C,$F7,$06,$C7,$E4,$DA,$FE,$7D,$AF,$E9,$3A,$1E
	.dc.b	$67,$81,$B1,$F9,$B6,$BF,$A7,$6B,$FA,$CF,$9E,$FF,$58,$A6,$93,$83
	.dc.b	$D7,$D0,$F8,$5A,$4A,$FD,$44,$65,$FD,$57,$EB,$DF,$09,$34,$8F,$C6
	.dc.b	$7E,$6B,$C3,$6C,$11,$E7,$F1,$7F,$E4,$5F,$0F,$FC,$F1,$4F,$92,$C6
	.dc.b	$5F,$E4,$98,$C5,$FF,$B6,$D5,$EA,$3F,$D1,$7F,$93,$3E,$7B,$11,$E6
	.dc.b	$F2,$AD,$FF,$91,$23,$FD,$C9,$5B,$7F,$DC,$DF,$66,$F8,$EE,$2F,$EC
	.dc.b	$A3,$3E,$06,$27,$9F,$9D,$D3,$E0,$65,$79,$63,$BF,$C1,$78,$3E,$12
	.dc.b	$4F,$A6,$3C,$BF,$64,$F8,$CB,$F0,$72,$BF,$DF,$13,$97,$E0,$78,$DE
	.dc.b	$AF,$93,$91,$83,$E2,$78,$BE,$EF,$84,$91,$83,$FD,$5F,$D1,$FE,$A5
	.dc.b	$F9,$5F,$EA,$51,$8E,$8F,$F3,$B4,$3E,$8B,$E4,$B6,$74,$7F,$AE,$1E
	.dc.b	$2A,$BE,$CF,$C3,$7C,$EF,$5E,$4F,$D7,$8F,$A3,$F0,$9F,$4B,$EA,$64
	.dc.b	$FD,$B8,$FA,$FF,$60,$FA,$3B,$F3,$52,$BC,$22,$74,$FE,$0F,$C6,$FB
	.dc.b	$BE,$DC,$8C,$1C,$3E,$B9,$1F,$82,$F7,$24,$E6,$3F,$0F,$F5,$EF,$4E
	.dc.b	$FC,$4C,$AF,$0C,$E5,$FC,$97,$B7,$F8,$CF,$3A,$46,$0F,$3F,$67,$FB
	.dc.b	$4F,$93,$FD,$A8,$BE,$A3,$F2,$AD,$9D,$F4,$7E,$63,$F0,$8F,$CB,$47
	.dc.b	$CD,$8F,$F5,$54,$7E,$8A,$30,$7F,$BA,$9C,$5F,$FB,$85,$E2,$FF,$B8
	.dc.b	$FD,$DC,$D7,$FA,$CD,$FD,$AD,$72,$0E,$27,$87,$9B,$FA,$AF,$A6,$FD
	.dc.b	$15,$6F,$E9,$FF,$62,$FD,$46,$BC,$E3,$75,$FE,$F3,$5F,$3B,$CF,$C9
	.dc.b	$95,$F7,$27,$2F,$DE,$C7,$D3,$97,$9F,$CE,$FD,$DC,$62,$42,$9F,$2D
	.dc.b	$E7,$24,$75,$7D,$5C,$7F,$C1,$3E,$99,$F3,$BF,$C1,$BE,$9D,$D3,$4D
	.dc.b	$46,$FE,$46,$E3,$F9,$2B,$E2,$D8,$FE,$1F,$E7,$FC,$2E,$16,$FF,$EE
	.dc.b	$1E,$F6,$9F,$C5,$1B,$87,$E7,$2F,$85,$7F,$B9,$DF,$AF,$1E,$E7,$2D
	.dc.b	$F5,$7B,$F2,$FF,$1E,$FA,$47,$C9,$F5,$F4,$E5,$FE,$A2,$72,$DF,$A7
	.dc.b	$E1,$52,$30,$6F,$CA,$E9,$32,$30,$73,$E1,$F5,$7C,$5C,$8C,$1D,$F8
	.dc.b	$7C,$93,$23,$07,$8F,$E9,$FF,$84,$3F,$E3,$FC,$91,$88,$FF,$22,$C6
	.dc.b	$38,$BF,$C7,$4D,$8E,$1F,$3E,$FA,$7C,$B6,$33,$18,$3F,$AC,$7C,$E4
	.dc.b	$8F,$EF,$6D,$7F,$CA,$1E,$F7,$D8,$7F,$48,$7F,$81,$B9,$3F,$88,$BF
	.dc.b	$90,$9F,$C8,$FD,$3B,$97,$FC,$B7,$7C,$73,$E5,$7F,$96,$E5,$C6,$3F
	.dc.b	$E6,$06,$C7,$E6,$8E,$C7,$81,$CB,$F9,$EE,$5A,$FF,$65,$9F,$E4,$7C
	.dc.b	$DF,$CD,$DF,$2C,$EB,$BA,$46,$0F,$4B,$CB,$DB,$91,$83,$DA,$F8,$31
	.dc.b	$73,$53,$C7,$87,$F6,$5A,$FE,$67,$3A,$6F,$1F,$13,$83,$83,$F9,$CF
	.dc.b	$D3,$FB,$A7,$E2,$B6,$3F,$4D,$DD,$F0,$23,$1A,$7E,$F9,$FE,$FF,$3C
	.dc.b	$B5,$D6,$DB,$78,$7B,$1C,$7D,$8B,$5E,$07,$0D,$3C,$71,$8C,$3F,$91
	.dc.b	$60,$F5,$3C,$47,$D2,$F1,$7F,$5B,$7E,$A8,$CB,$EC,$F5,$6D,$F2,$7D
	.dc.b	$5E,$A4,$9D,$1F,$09,$FB,$BF,$4E,$FA,$7F,$C3,$DF,$E0,$E3,$15,$FE
	.dc.b	$99,$6C,$7F,$A9,$5A,$FF,$AA,$AF,$E2,$BF,$EA,$96,$B9,$D7,$AB,$FC
	.dc.b	$A2,$32,$FE,$47,$E2,$3F,$9A,$DF,$FA,$9D,$B1,$FE,$B3,$6B,$FE,$B8
	.dc.b	$6B,$F9,$9F,$A7,$FE,$57,$3F,$EA,$F6,$C7,$FA,$ED,$AF,$FB,$01,$AF
	.dc.b	$1F,$92,$F2,$A8,$FF,$54,$BF,$CF,$9A,$E6,$37,$F4,$F8,$23,$83,$9F
	.dc.b	$D3,$FF,$44,$BF,$F6,$13,$6B,$FD,$9C,$D7,$FD,$A6,$D7,$F6,$9F,$4E
	.dc.b	$F0,$7C,$AD,$AF,$F6,$63,$67,$FD,$B0,$D7,$FD,$BB,$1F,$88,$7F,$DB
	.dc.b	$8D,$73,$E3,$FC,$08,$C4,$7F,$03,$FA,$7F,$F4,$B4,$7F,$B6,$5B,$1F
	.dc.b	$EE,$56,$BF,$EE,$A6,$BC,$7E,$53,$E5,$C7,$FA,$E6,$79,$46,$0F,$9F
	.dc.b	$F4,$EE,$7A,$49,$FF,$74,$46,$5F,$F7,$57,$D3,$FF,$1F,$6F,$87,$F6
	.dc.b	$23,$18,$FF,$BE,$5A,$FF,$BF,$9F,$ED,$B3,$CB,$CF,$E0,$47,$FD,$ED
	.dc.b	$E3,$7F,$AF,$89,$F3,$3F,$84,$7F,$D8,$3F,$4F,$F1,$B0,$FF,$00,$B5
	.dc.b	$34,$9F,$41,$F8,$C4,$D2,$5F,$F3,$19,$FE,$65,$E0,$3C,$6F,$C9,$B6
	.dc.b	$75,$3F,$E1,$08,$C3,$FD,$22,$7F,$33,$F4,$EF,$87,$A4,$BE,$34,$9D
	.dc.b	$46,$1F,$E2,$0D,$F0,$39,$91,$C1,$FE,$62,$87,$67,$F4,$67,$8D,$7F
	.dc.b	$8E,$DC,$CE,$FE,$0F,$5F,$05,$B6,$C7,$38,$F4,$91,$F2,$3F,$B4,$F9
	.dc.b	$AF,$F0,$DC,$7A,$9B,$7F,$11,$AF,$CB,$6B,$FD,$75,$FF,$30,$47,$F9
	.dc.b	$E9,$B1,$FE,$51,$79,$F9,$1F,$DA,$1C,$B7,$E8,$7F,$B1,$C1,$BF,$31
	.dc.b	$86,$5F,$95,$F4,$11,$C1,$E1,$FA,$8D,$E1,$E1,$7A,$FD,$3F,$6F,$A1
	.dc.b	$B5,$E7,$70,$FC,$17,$E2,$11,$F8,$9E,$C7,$C7,$23,$83,$E4,$7C,$B7
	.dc.b	$A8,$F5,$C5,$8F,$89,$F1,$3E,$E3,$DF,$EE,$D8,$C9,$F4,$63,$EA,$11
	.dc.b	$C1,$F3,$7A,$B8,$67,$F3,$5F,$F0,$5F,$D3,$37,$BE,$1E,$1B,$7F,$EA
	.dc.b	$E9,$FE,$31,$C3,$FE,$1A,$6C,$7F,$85,$DE,$4E,$EB,$AF,$B9,$DB,$DC
	.dc.b	$B3,$CD,$C3,$4E,$D1,$8C,$BF,$D2,$56,$FA,$6F,$39,$E8,$FA,$7B,$7E
	.dc.b	$58,$FE,$8F,$3B,$8F,$D6,$B6,$BF,$6E,$D7,$F7,$CE,$72,$F6,$7B,$1D
	.dc.b	$27,$C2,$FC,$6E,$FF,$8F,$F1,$17,$0F,$88,$E1,$A7,$C7,$8C,$57,$FA
	.dc.b	$AF,$E9,$BF,$C3,$29,$A7,$FA,$62,$7F,$DB,$6E,$3C,$1F,$F1,$FB,$A9
	.dc.b	$B1,$8F,$06,$57,$0F,$72,$D7,$79,$B9,$FF,$94,$30,$56,$2F,$FC,$9D
	.dc.b	$1F,$C6,$27,$55,$E0,$AD,$5F,$FA,$B2,$B5,$7C,$2B,$4F,$A1,$DC,$FB
	.dc.b	$5C,$BE,$38,$D1,$1F,$89,$5C,$2B,$AC,$77,$AE,$37,$E4,$BE,$0D,$9E
	.dc.b	$31,$83,$C8,$57,$D1,$7D,$6B,$B3,$67,$53,$F4,$E3,$0F,$9C,$EA,$5A
	.dc.b	$F0,$12,$7E,$4C,$F0,$8C,$4F,$82,$7A,$DF,$B2,$3E,$24,$7D,$73,$37
	.dc.b	$E4,$69,$C2,$FA,$3E,$C9,$F1,$DF,$67,$99,$E5,$8F,$19,$C9,$C1,$E3
	.dc.b	$29,$C2,$39,$5F,$E1,$27,$84,$62,$BC,$35,$F4,$E7,$D0,$F2,$AF,$61
	.dc.b	$27,$E4,$C7,$F6,$98,$FE,$C6,$DB,$FF,$1D,$C6,$23,$D8,$66,$BF,$B1
	.dc.b	$C6,$63,$C0,$7F,$6E,$7B,$F8,$FF,$87,$24,$FC,$E6,$CF,$95,$B8,$7B
	.dc.b	$9B,$3F,$2E,$3F,$49,$18,$9E,$F1,$89,$E8,$48,$E8,$D8,$E8,$D3,$A1
	.dc.b	$FA,$C9,$E9,$3F,$3E,$30,$6D,$EE,$E1,$EA,$49,$EA,$D8,$EA,$FE,$CE
	.dc.b	$7A,$D7,$C4,$9E,$93,$DC,$D3,$88,$D4,$73,$3C,$26,$8C,$1C,$4F,$E2
	.dc.b	$D3,$DA,$FF,$89,$B6,$BF,$8B,$B6,$7F,$8F,$36,$FE,$54,$74,$BF,$0C
	.dc.b	$7F,$13,$F2,$7F,$8E,$FB,$3F,$C7,$9B,$87,$F2,$26,$DF,$F2,$96,$D7
	.dc.b	$F1,$BF,$0F,$F9,$2F,$AF,$FC,$99,$B7,$FC,$AA,$2D,$FF,$31,$9B,$5F
	.dc.b	$C8,$FC,$9F,$E5,$A7,$F2,$EF,$0E,$E1,$FC,$CA,$3F,$2E,$47,$F2,$AF
	.dc.b	$E6,$BE,$32,$47,$E6,$1B,$87,$F9,$13,$E8,$FF,$C3,$6F,$F9,$2D,$AF
	.dc.b	$FB,$F5,$A8,$38,$6F,$D3,$FF,$A8,$A0,$D2,$57,$F9,$2A,$1E,$26,$C9
	.dc.b	$19,$79,$7C,$A4,$8E,$0D,$BF,$47,$89,$DC,$93,$D5,$F1,$7C,$34,$8F
	.dc.b	$03,$67,$FC,$C7,$F1,$4F,$99,$FF,$5C,$B6,$BF,$CB,$ED,$9D,$CE,$B8
	.dc.b	$A5,$69,$B3,$D3,$8F,$99,$27,$9B,$DB,$E5,$A4,$76,$6C,$F9,$3C,$FF
	.dc.b	$19,$27,$C6,$FF,$B3,$BE,$91,$E3,$52,$4F,$F9,$DA,$32,$FF,$9E,$3E
	.dc.b	$91,$E7,$3D,$EC,$37,$1F,$F4,$2C,$BE,$91,$E0,$6A,$FF,$D0,$3F,$53
	.dc.b	$FE,$AF,$BB,$BF,$FA,$11,$46,$5F,$F4,$7B,$5F,$F4,$B7,$D2,$3E,$97
	.dc.b	$83,$63,$FD,$1A,$D7,$FD,$2E,$D7,$FD,$3D,$EA,$3B,$7C,$0D,$8F,$F4
	.dc.b	$AB,$5F,$F4,$FB,$5F,$F5,$37,$D2,$FF,$BD,$93,$49,$BF,$AF,$E9,$1E
	.dc.b	$45,$25,$7F,$A8,$A3,$2F,$FA,$93,$E9,$3E,$12,$69,$1F,$8D,$F8,$7F
	.dc.b	$8D,$FD,$50,$70,$FE,$B5,$AE,$72,$F9,$D7,$FA,$C8,$FF,$59,$B5,$FF
	.dc.b	$5E,$B5,$FF,$62,$35,$EF,$4F,$87,$FB,$0D,$8F,$54,$60,$FB,$9F,$4A
	.dc.b	$F6,$29,$27,$FD,$87,$19,$7F,$D8,$BF,$4A,$F9,$9F,$C2,$36,$33,$18
	.dc.b	$3C,$3E,$6C,$8F,$94,$D7,$FD,$AF,$8B,$E9,$7C,$E4,$8F,$A0,$FE,$47
	.dc.b	$E9,$1F,$4B,$49,$3F,$ED,$88,$CB,$F5,$F9,$FE,$09,$1C,$1F,$6F,$C0
	.dc.b	$F8,$A4,$70,$7F,$B8,$BE,$51,$F5,$3F,$A1,$6C,$66,$30,$7C,$EF,$83
	.dc.b	$91,$F8,$76,$BE,$EF,$47,$E9,$91,$FC,$63,$FA,$DF,$A4,$7D,$2D,$24
	.dc.b	$FF,$BD,$23,$2F,$E4,$FA,$FE,$E9,$1C,$1F,$EF,$AF,$A5,$FC,$69,$3A
	.dc.b	$3F,$74,$FE,$43,$E9,$5F,$05,$FB,$D6,$D7,$E6,$0F,$CC,$B8,$70,$1C
	.dc.b	$3B,$E3,$7A,$7F,$90,$24,$7E,$A5,$FC,$EF,$74,$8F,$CE,$B5,$EB,$F4
	.dc.b	$7F,$E8,$47,$B8,$70,$EB,$DA,$FE,$29,$18,$8F,$47,$9E,$F9,$BF,$E3
	.dc.b	$0D,$8F,$E7,$F1,$83,$F8,$E7,$D1,$3D,$2A,$49,$F0,$C6,$5B,$F3,$3C
	.dc.b	$08,$E0,$DF,$CD,$FF,$00,$F6,$DA,$7F,$02,$CB,$D3,$18,$AF,$ED,$D1
	.dc.b	$89,$FE,$0F,$82,$F9,$46,$1B,$5F,$DD,$67,$4F,$F7,$7B,$E8,$A3,$2A
	.dc.b	$33,$0F,$88,$E7,$2F,$A1,$7C,$EE,$3E,$89,$EF,$D3,$73,$AA,$F0,$67
	.dc.b	$FC,$7A,$1F,$11,$AC,$25,$68,$FF,$06,$4E,$9C,$0F,$83,$FE,$2C,$F9
	.dc.b	$EF,$F9,$4F,$C2,$F9,$E3,$6D,$74,$7F,$85,$E7,$8E,$BF,$09,$FC,$04
	.dc.b	$68,$ED,$E6,$7C,$08,$DD,$FF,$36,$77,$5F,$2B,$E8,$5F,$17,$F1,$B1
	.dc.b	$FC,$16,$3C,$DB,$5B,$9F,$97,$4F,$A2,$B1,$2F,$DF,$39,$6F,$8B,$C2
	.dc.b	$A4,$60,$DE,$BE,$1F,$8E,$91,$83,$9E,$FE,$09,$91,$83,$BF,$C6,$FE
	.dc.b	$05,$FD,$CE,$28,$C4,$7E,$FE,$31,$9F,$EC,$1B,$1C,$3E,$65,$F4,$59
	.dc.b	$6C,$66,$30,$7F,$66,$F9,$59,$1F,$E0,$96,$BF,$C2,$BD,$DF,$D1,$61
	.dc.b	$F9,$88,$FF,$09,$F1,$7E,$11,$FF,$29,$57,$F9,$4F,$E8,$BF,$0D,$1F
	.dc.b	$DB,$2F,$8D,$7F,$17,$E4,$B3,$FD,$C1,$AF,$F9,$6E,$6B,$FD,$96,$7B
	.dc.b	$77,$60,$DF,$23,$E2,$31,$3E,$3F,$07,$89,$8B,$FC,$61,$FE,$D7,$2F
	.dc.b	$A7,$D8,$DC,$FA,$DF,$1E,$AE,$6A,$78,$3D,$4B,$5F,$A4,$9F,$D9,$B7
	.dc.b	$93,$38,$7E,$6E,$72,$FE,$77,$E8,$BE,$4B,$76,$F5,$F8,$1F,$0A,$73
	.dc.b	$5F,$C1,$9F,$85,$5C,$78,$35,$7B,$78,$7A,$5C,$7E,$1E,$E7,$FB,$8C
	.dc.b	$11,$98,$C3,$AC,$3C,$8F,$87,$D1,$93,$FC,$03,$C5,$FD,$6B,$CE,$3C
	.dc.b	$DF,$1B,$23,$07,$77,$D5,$E9,$48,$C1,$E3,$7E,$FF,$D0,$BE,$8A,$FC
	.dc.b	$9E,$1B,$5F,$C4,$2F,$E0,$F0,$7F,$28,$7E,$D3,$77,$FC,$4C,$62,$7F
	.dc.b	$C7,$7F,$0A,$D3,$F1,$AB,$C2,$B1,$87,$F2,$3F,$45,$F5,$B1,$FE,$6A
	.dc.b	$FE,$C6,$FF,$90,$B9,$7F,$C8,$EE,$78,$3F,$8C,$2F,$A1,$FC,$DF,$D7
	.dc.b	$7F,$AD,$6B,$8C,$62,$BF,$3F,$F4,$7B,$9F,$E5,$AB,$EA,$33,$5A,$7F
	.dc.b	$4B,$F4,$9F,$D3,$EB,$F9,$9B,$FE,$04,$6F,$4E,$BD,$7F,$D5,$FD,$23
	.dc.b	$C7,$BF,$37,$86,$DF,$F3,$C3,$EE,$E9,$FC,$C5,$FE,$48,$67,$CD,$FD
	.dc.b	$AB,$88,$FE,$6D,$F4,$9F,$E9,$58,$FE,$8A,$BE,$7F,$F7,$6E,$67,$71
	.dc.b	$8F,$A4,$CF,$AB,$6F,$8B,$FC,$0B,$E9,$1E,$15,$F0,$70,$DA,$FE,$A7
	.dc.b	$F4,$9F,$D4,$DB,$F3,$3F,$C2,$A3,$19,$FF,$4B,$7F,$D9,$1D,$3F,$DC
	.dc.b	$1F,$EC,$F5,$CB,$E9,$F8,$7B,$1C,$5F,$E2,$7E,$8F,$FA,$E3,$0F,$EA
	.dc.b	$9F,$6F,$FE,$B7,$BF,$EC,$73,$9C,$7F,$BD,$CE,$6B,$FA,$D7,$D3,$7C
	.dc.b	$C9,$FE,$FE,$FF,$0B,$BF,$ED,$6E,$5F,$F4,$8B,$99,$E3,$F4,$5F,$C9
	.dc.b	$23,$FC,$D7,$3C,$23,$1F,$19,$FC,$A3,$CC,$7A,$17,$DB,$C3,$6F,$FB
	.dc.b	$CF,$D3,$3C,$EF,$E5,$ED,$8C,$FC,$F6,$26,$47,$0F,$2A,$EB,$FD,$D6
	.dc.b	$D1,$C8,$FE,$89,$01,$1A,$9F,$F2,$2B,$63,$FC,$AA,$D7,$FC,$B4,$FF
	.dc.b	$B2,$E3,$7F,$11,$D2,$79,$B6,$75,$1F,$E5,$EF,$82,$E9,$C1,$FE,$2C
	.dc.b	$6D,$FF,$AF,$2B,$73,$9C,$3F,$A7,$FD,$2B,$EA,$29,$8B,$A7,$F8,$82
	.dc.b	$34,$FF,$97,$9B,$8F,$F9,$B9,$AF,$F9,$CD,$AF,$FB,$31,$FD,$24,$71
	.dc.b	$EE,$F6,$46,$27,$94,$60,$F2,$FD,$17,$99,$B5,$E3,$8E,$24,$7C,$8F
	.dc.b	$4F,$A2,$91,$E7,$6D,$7C,$7F,$57,$FB,$7E,$A7,$E1,$CE,$1F,$8B,$F2
	.dc.b	$D2,$FE,$17,$A1,$70,$FF,$1B,$67,$FE,$D3,$9F,$89,$5E,$C2,$3F,$1F
	.dc.b	$C9,$F7,$91,$C1,$2B,$9F,$D1,$7D,$49,$FA,$AC,$9F,$A4,$E5,$A7,$F9
	.dc.b	$53,$E9,$BF,$8C,$D7,$FA,$8D,$AF,$FB,$F1,$AF,$E5,$5A,$9F,$D9,$3F
	.dc.b	$E7,$EF,$A9,$FF,$09,$C6,$21,$AA,$73,$F0,$B2,$FC,$5F,$0D,$6F,$FC
	.dc.b	$B5,$87,$F4,$57,$C9,$5F,$9B,$8C,$47,$96,$31,$ED,$78,$3E,$0B,$C2
	.dc.b	$DC,$3D,$D1,$C4,$8F,$FA,$C2,$D1,$8B,$FF,$35,$64,$F9,$1E,$51,$87
	.dc.b	$DB,$F3,$5F,$82,$EA,$B8,$FF,$9F,$AF,$F4,$EF,$C9,$9D,$3F,$3D,$B7
	.dc.b	$DB,$C6,$FC,$59,$FE,$3A,$FA,$7F,$F0,$CD,$7F,$AA,$E3,$2F,$57,$5E
	.dc.b	$0E,$2B,$FD,$01,$5F,$CE,$3B,$3F,$D4,$AD,$BF,$E1,$92,$8F,$E3,$F1
	.dc.b	$FE,$94,$75,$1F,$C4,$5A,$EF,$C1,$F2,$78,$3E,$56,$57,$17,$2D,$24
	.dc.b	$F9,$23,$2D,$8C,$DF,$9A,$72,$D8,$C9,$E8,$BF,$19,$E8,$AE,$44,$9E
	.dc.b	$0F,$C2,$BF,$2B,$9A,$E5,$7D,$9A,$76,$8F,$81,$19,$6C,$65,$F8,$4F
	.dc.b	$C7,$9E,$B5,$E0,$24,$79,$E3,$2F,$69,$F1,$1F,$36,$32,$F6,$32,$7D
	.dc.b	$57,$2F,$80,$CD,$FC,$1A,$EF,$1C,$5A,$F1,$AF,$44,$FC,$8B,$F6,$4F
	.dc.b	$3A,$DC,$FC,$CB,$EE,$EE,$FD,$C7,$2B,$CC,$E9,$B1,$93,$BC,$65,$A6
	.dc.b	$6F,$D3,$19,$6C,$65,$FC,$A9,$EF,$2D,$7F,$84,$9B,$5F,$D6,$E3,$11
	.dc.b	$F3,$4C,$CF,$27,$E6,$18,$3F,$77,$F2,$1F,$33,$43,$49,$3F,$7E,$32
	.dc.b	$D8,$CC,$FD,$69,$CB,$63,$38,$FD,$29,$CB,$63,$35,$F8,$38,$FE,$DD
	.dc.b	$18,$AF,$B5,$3F,$B6,$AE,$2E,$CC,$4E,$E7,$ED,$46,$0F,$31,$1F,$23
	.dc.b	$6B,$C4,$D3,$A4,$7F,$83,$EB,$BB,$E2,$31,$5D,$EF,$F8,$1C,$EA,$7C
	.dc.b	$26,$AB,$A3,$C4,$FE,$7F,$5C,$70,$FE,$CE,$D9,$FE,$4C,$D9,$FA,$B1
	.dc.b	$C9,$EA,$72,$71,$7E,$29,$E1,$87,$CA,$9C,$DF,$66,$9D,$8F,$E3,$DE
	.dc.b	$67,$F7,$8F,$27,$F9,$73,$63,$27,$F3,$98,$CB,$4F,$F1,$A5,$1F,$E5
	.dc.b	$7F,$57,$5C,$98,$3F,$9D,$76,$FF,$BD,$23,$59,$AB,$93,$FE,$3F,$8C
	.dc.b	$C6,$21,$18,$AF,$E4,$55,$CD,$FF,$7F,$FD,$87,$F8,$86,$F8,$36,$3F
	.dc.b	$94,$67,$AF,$06,$47,$E9,$71,$57,$F8,$A5,$FD,$8B,$5F,$E8,$ED,$7F
	.dc.b	$CA,$9C,$5F,$E9,$37,$F6,$ED,$7F,$A6,$B5,$DF,$D2,$7F,$94,$78,$3F
	.dc.b	$6F,$1F,$8F,$DF,$04,$C3,$C9,$6F,$A6,$FF,$2C,$36,$33,$18,$33,$94
	.dc.b	$8E,$2D,$BF,$E0,$5B,$78,$DF,$97,$97,$F9,$8E,$30,$67,$07,$87,$E0
	.dc.b	$26,$1E,$47,$D5,$E7,$C8,$F9,$5A,$23,$18,$F8,$33,$EC,$4A,$FA,$18
	.dc.b	$FA,$E7,$FC,$E1,$C5,$F1,$49,$3D,$1F,$36,$92,$38,$B6,$7F,$CF,$1E
	.dc.b	$4F,$CD,$24,$F9,$1F,$76,$72,$3C,$2D,$9F,$E4,$DF,$4D,$F3,$FE,$0B
	.dc.b	$F0,$8D,$8C,$EE,$E8,$CE,$AB,$FC,$57,$D3,$7F,$B3,$AE,$FA,$7F,$8D
	.dc.b	$51,$97,$F9,$A3,$5F,$E7,$EB,$6E,$0D,$8E,$B1,$83,$A7,$6B,$B6,$92
	.dc.b	$7F,$39,$19,$7F,$3D,$F4,$AF,$23,$C8,$D8,$CC,$60,$F4,$7D,$23,$9E
	.dc.b	$92,$7F,$51,$19,$7F,$55,$F4,$BF,$EA,$16,$F9,$BF,$06,$31,$1F,$D4
	.dc.b	$DA,$FF,$5E,$FA,$5F,$33,$FD,$5D,$AE,$73,$F9,$51,$88,$FF,$1C,$FC
	.dc.b	$DF,$D3,$7F,$74,$D7,$FB,$03,$5F,$ED,$6D,$7F,$6B,$F2,$3F,$57,$23
	.dc.b	$49,$1D,$3E,$2B,$FC,$8F,$E0,$FF,$13,$B4,$63,$A7,$C6,$DA,$FE,$0B
	.dc.b	$F4,$15,$7D,$2B,$D1,$C3,$63,$F9,$39,$FE,$DF,$70,$FF,$25,$BF,$E1
	.dc.b	$7F,$55,$CF,$E9,$9F,$27,$49,$5C,$84,$3E,$3F,$C5,$A7,$FB,$F2,$74
	.dc.b	$D7,$B1,$FE,$EF,$71,$AF,$F3,$1D,$DE,$BF,$2F,$A6,$97,$C9,$F7,$4B
	.dc.b	$E7,$FE,$07,$D3,$3D,$EA,$4B,$F3,$43,$E3,$FC,$BE,$04,$9F,$2B,$D3
	.dc.b	$7F,$4F,$D3,$49,$F4,$BE,$0D,$FE,$77,$CC,$39,$A9,$2B,$E2,$43,$8F
	.dc.b	$E1,$FC,$0C,$9F,$86,$FA,$BD,$FF,$8F,$E9,$C9,$F8,$EF,$BB,$A7,$EE
	.dc.b	$7D,$2B,$E9,$A9,$2B,$E9,$43,$E3,$3E,$8F,$8D,$27,$E8,$BF,$23,$E3
	.dc.b	$3F,$0D,$F4,$DF,$A2,$EB,$FE,$C0,$DE,$EF,$F4,$BB,$6F,$FA,$8F,$C9
	.dc.b	$7E,$37,$E9,$9F,$0D,$49,$3F,$83,$87,$C3,$7E,$0B,$DC,$93,$F8,$27
	.dc.b	$EC,$FC,$7F,$DA,$CA,$4F,$E2,$FC,$0F,$EE,$5A,$7E,$0F,$2F,$CE,$7D
	.dc.b	$33,$E7,$A9,$2F,$F2,$51,$F8,$AC,$3F,$C7,$5F,$3F,$FA,$3F,$78,$D3
	.dc.b	$F4,$DF,$57,$FA,$CB,$FC,$D7,$D3,$FE,$2D,$BE,$37,$EA,$E3,$11,$F9
	.dc.b	$80,$FE,$91,$E3,$E0,$6C,$8F,$4E,$CE,$5C,$9E,$3D,$64,$F8,$23,$B7
	.dc.b	$0E,$A3,$07,$91,$FD,$E7,$D3,$3E,$8B,$F7,$DF,$4E,$F8,$9A,$6A,$37
	.dc.b	$F1,$D7,$0F,$DB,$5F,$16,$C7,$F0,$7F,$86,$F8,$51,$ED,$F8,$B6,$5F
	.dc.b	$C2,$DB,$1F,$BF,$9D,$F3,$FF,$89,$DF,$9B,$19,$F4,$FA,$1D,$79,$3F
	.dc.b	$8A,$7D,$33,$E6,$7A,$72,$CB,$FE,$15,$39,$6F,$D3,$F0,$A9,$18,$37
	.dc.b	$E5,$74,$99,$18,$39,$F0,$7A,$79,$64,$60,$EF,$C1,$E4,$99,$18,$3C
	.dc.b	$7F,$23,$FE,$47,$7F,$D3,$99,$23,$11,$FE,$A8,$8C,$6F,$FE,$3D,$6C
	.dc.b	$70,$F9,$B7,$D3,$E5,$B1,$98,$C1,$FD,$E5,$7C,$D5,$23,$FD,$F5,$FE
	.dc.b	$FD,$F0,$FF,$E5,$AC,$3F,$CB,$13,$FE,$07,$EF,$FE,$68,$8E,$63,$E9
	.dc.b	$DC,$BF,$EC,$0B,$E5,$5F,$EC,$2E,$5B,$7C,$4F,$EA,$31,$D2,$73,$1D
	.dc.b	$66,$BF,$AE,$9F,$0F,$2E,$B9,$BF,$AF,$C6,$0F,$2F,$67,$03,$8C,$D8
	.dc.b	$F4,$FC,$9F,$43,$17,$FB,$13,$FD,$E5,$1F,$D9,$B9,$BC,$7A,$9F,$07
	.dc.b	$A9,$87,$3C,$92,$BF,$AC,$56,$DB,$1E,$CF,$A7,$F9,$58,$7E,$FD,$B5
	.dc.b	$FE,$85,$D1,$DF,$BD,$F8,$6F,$32,$FE,$C9,$87,$07,$E5,$3F,$2B,$1E
	.dc.b	$9E,$B7,$7F,$87,$F0,$36,$B8,$C5,$9D,$C6,$0E,$BE,$A7,$9C,$FA,$7E
	.dc.b	$1F,$EB,$63,$27,$8B,$E6,$3D,$86,$DC,$1E,$6F,$57,$E0,$1A,$70,$7F
	.dc.b	$7D,$FA,$76,$FF,$80,$6C,$67,$D5,$C4,$FE,$12,$FE,$AE,$5F,$91,$6A
	.dc.b	$7F,$23,$5F,$E0,$97,$23,$2E,$47,$36,$C7,$F8,$3F,$E9,$9D,$BF,$92
	.dc.b	$5F,$4F,$63,$0F,$E5,$BE,$9F,$E9,$63,$FA,$37,$E9,$72,$FF,$AF,$DC
	.dc.b	$C1,$CE,$1F,$A1,$5C,$3C,$F8,$7A,$36,$7F,$C4,$5F,$4B,$FD,$81,$FF
	.dc.b	$12,$7D,$2F,$F9,$DD,$CF,$A9,$FA,$88,$C6,$3F,$D2,$D7,$CE,$E7,$87
	.dc.b	$4D,$BE,$7F,$4D,$FB,$0E,$5F,$DE,$46,$0F,$F1,$93,$8C,$3F,$AB,$7B
	.dc.b	$7F,$EA,$C9,$FF,$1B,$36,$33,$F1,$3F,$BA,$8C,$70,$7F,$B8,$5F,$EF
	.dc.b	$B5,$C3,$D6,$D6,$BB,$7F,$80,$FD,$87,$F5,$39,$FF,$21,$37,$1F,$EC
	.dc.b	$CB,$CB,$DF,$81,$FC,$23,$E9,$7E,$5D,$FF,$92,$5F,$F4,$93,$78,$F5
	.dc.b	$C9,$FC,$3B,$EC,$3F,$DE,$59,$7F,$1B,$5F,$21,$9B,$D3,$BF,$84,$E1
	.dc.b	$F0,$96,$7F,$CA,$D5,$FE,$33,$7F,$CB,$1F,$4B,$E7,$E2,$FF,$80,$0D
	.dc.b	$8F,$E4,$0F,$A5,$DE,$BF,$C8,$23,$FC,$95,$1C,$BE,$1F,$1C,$D9,$7F
	.dc.b	$92,$FD,$87,$E7,$B5,$FF,$04,$37,$15,$FF,$2D,$87,$C5,$F0,$F3,$FF
	.dc.b	$96,$FD,$2B,$E1,$E9,$FC,$C3,$EC,$3F,$55,$87,$F9,$C6,$74,$D9,$E3
	.dc.b	$F2,$7C,$23,$4E,$1F,$E7,$99,$FF,$3D,$AF,$F5,$E4,$62,$3F,$A4,$46
	.dc.b	$1C,$35,$E3,$F2,$DE,$A9,$E1,$1F,$EB,$F7,$FA,$9B,$FD,$46,$34,$F5
	.dc.b	$6B,$CB,$93,$FA,$5C,$7D,$C9,$FF,$46,$3F,$80,$9F,$F6,$54,$EB,$E7
	.dc.b	$E5,$6B,$E6,$ED,$7F,$8D,$6D,$5E,$28,$FF,$69,$36,$33,$5E,$86,$CF
	.dc.b	$97,$97,$44,$9D,$36,$3B,$7C,$97,$A2,$3D,$2C,$9E,$8F,$72,$3D,$FD
	.dc.b	$DF,$11,$1C,$1E,$3F,$8F,$F8,$67,$C4,$8B,$1F,$13,$BF,$E3,$9F,$23
	.dc.b	$E0,$AC,$64,$F6,$47,$BC,$8F,$D6,$FA,$5F,$AD,$97,$DF,$6B,$F8,$46
	.dc.b	$BF,$E0,$16,$BF,$E0,$B7,$8F,$37,$E6,$A3,$1F,$05,$FE,$04,$9C,$D7
	.dc.b	$18,$C1,$1F,$F6,$44,$6A,$78,$C6,$3C,$1F,$B1,$3F,$B1,$9F,$34,$60
	.dc.b	$93,$CB,$E0,$7C,$11,$E1,$7C,$5E,$44,$9E,$FF,$55,$6F,$FC,$F5,$87
	.dc.b	$F3,$09,$F8,$95,$E0,$23,$F9,$CF,$B0,$7D,$87,$B7,$EA,$BF,$C4,$31
	.dc.b	$89,$F5,$C6,$5F,$9C,$FA,$88,$E0,$F8,$3F,$1F,$BC,$7C,$6F,$F8,$B3
	.dc.b	$EA,$7F,$9D,$6F,$F3,$65,$1F,$D5,$8C,$D7,$F8,$D5,$C4,$71,$F7,$79
	.dc.b	$36,$7F,$DE,$D3,$FD,$C8,$9F,$07,$DF,$E0,$A4,$E9,$F1,$46,$2B,$FD
	.dc.b	$41,$5F,$CE,$1B,$E5,$FC,$2F,$2B,$D0,$CC,$3B,$B9,$69,$27,$8B,$FE
	.dc.b	$55,$CB,$DD,$AF,$FB,$AE,$35,$1F,$07,$0F,$03,$6F,$A5,$F2,$24,$66
	.dc.b	$BA,$1F,$E5,$9B,$E9,$3A,$8F,$91,$5E,$78,$FF,$5E,$4E,$AB,$94,$6C
	.dc.b	$FF,$35,$5F,$8A,$FA,$61,$C6,$B6,$48,$C9,$EA,$3B,$96,$BF,$CE,$97
	.dc.b	$DE,$33,$1D,$CF,$BF,$87,$BA,$BA,$47,$8A,$3C,$13,$DA,$38,$61,$E5
	.dc.b	$75,$1F,$09,$EF,$87,$07,$27,$7C,$3C,$4E,$30,$EE,$F8,$AB,$92,$9D
	.dc.b	$B9,$3C,$38,$79,$E7,$D7,$5C,$B0,$F2,$57,$1C,$3C,$4F,$19,$EF,$87
	.dc.b	$4A,$D9,$30,$E9,$5E,$5C,$3A,$91,$F6,$BF,$35,$E8,$E5,$F0,$39,$9E
	.dc.b	$8F,$B2,$BC,$15,$E4,$9F,$3C,$62,$78,$4F,$8A,$BA,$9E,$32,$1D,$8F
	.dc.b	$87,$1F,$D9,$2B,$F5,$11,$FD,$C1,$AF,$E0,$0C,$4F,$F5,$78,$CC,$7D
	.dc.b	$63,$33,$CE,$7F,$83,$79,$7E,$72,$47,$B5,$B3,$F8,$46,$E1,$F4,$DB
	.dc.b	$3F,$16,$FF,$C6,$35,$B8,$E4,$47,$04,$AE,$51,$96,$C6,$63,$A5,$76
	.dc.b	$8F,$8B,$5D,$AB,$64,$AD,$9C,$CB,$3C,$CE,$35,$CD,$FA,$27,$F2,$47
	.dc.b	$15,$F8,$86,$DF,$C6,$AE,$95,$C9,$AF,$2A,$F2,$47,$82,$FC,$93,$B3
	.dc.b	$C4,$6E,$FA,$CF,$F2,$C6,$C7,$F2,$67,$F9,$D7,$B3,$FC,$EF,$D2,$B5
	.dc.b	$FC,$C9,$B5,$FC,$B1,$B7,$FC,$7D,$FE,$85,$F1,$72,$BF,$9C,$F8,$96
	.dc.b	$FF,$99,$CD,$BF,$E4,$D3,$6B,$FD,$F1,$F6,$1F,$E2,$BB,$57,$F4,$B8
	.dc.b	$C3,$FD,$37,$E9,$FF,$D3,$13,$AA,$E3,$E2,$FF,$BF,$AF,$94,$FF,$92
	.dc.b	$1A,$EA,$7F,$C8,$31,$88,$07,$FB,$33,$FE,$4D,$6B,$CF,$9B,$FC,$6D
	.dc.b	$18,$8E,$23,$E4,$7F,$CC,$B5,$FE,$4E,$6C,$7F,$95,$DA,$FF,$97,$1A
	.dc.b	$FC,$2F,$A6,$EA,$49,$F3,$BE,$0F,$0E,$47,$A9,$B3,$E7,$F8,$EF,$29
	.dc.b	$27,$DE,$FF,$93,$7A,$7E,$4C,$FF,$9B,$1A,$FA,$9A,$FC,$46,$BE,$DE
	.dc.b	$4F,$F2,$B6,$3F,$C0,$23,$17,$EF,$9C,$F1,$C8,$F8,$4D,$9F,$85,$F3
	.dc.b	$3C,$89,$3F,$1D,$EB,$F4,$F2,$3E,$53,$67,$FD,$01,$F6,$1E,$42,$7B
	.dc.b	$3F,$E6,$97,$CF,$F5,$0F,$35,$F4,$FF,$E8,$08,$C3,$7E,$A1,$E0,$E6
	.dc.b	$B7,$8F,$FA,$37,$E9,$EB,$6F,$FD,$12,$A3,$2F,$FA,$59,$AF,$E1,$17
	.dc.b	$D8,$70,$6C,$7F,$A4,$9A,$F7,$8E,$BE,$9B,$B2,$92,$7F,$D3,$31,$97
	.dc.b	$FD,$37,$F4,$FF,$E8,$94,$D2,$70,$7A,$3E,$A1,$CD,$49,$5F,$EA,$48
	.dc.b	$CB,$FE,$A6,$FA,$5F,$7A,$69,$1F,$87,$F5,$1E,$37,$FD,$68,$D7,$FD
	.dc.b	$72,$D7,$F4,$ED,$7F,$63,$FA,$FF,$C8,$6E,$DF,$3A,$31,$06,$FD,$F7
	.dc.b	$BC,$99,$7D,$47,$E1,$47,$CC,$9F,$F6,$4B,$5F,$DB,$B5,$F0,$ED,$DF
	.dc.b	$D7,$FD,$DF,$87,$FE,$5B,$A4,$63,$0E,$2D,$BF,$4F,$A5,$5F,$37,$CE
	.dc.b	$5B,$34,$C7,$CB,$67,$E0,$E5,$7F,$86,$29,$1C,$7D,$5E,$5E,$AF,$3F
	.dc.b	$8E,$97,$B8,$ED,$F1,$DD,$FD,$79,$3C,$E3,$C5,$EB,$78,$FE,$2E,$4F
	.dc.b	$78,$F3,$7C,$8C,$8F,$13,$E9,$F8,$BF,$57,$B3,$27,$CC,$FB,$7A,$FD
	.dc.b	$DF,$51,$27,$D4,$FC,$1F,$87,$F8,$5F,$11,$F1,$7F,$60,$FF,$7D,$F4
	.dc.b	$FF,$A1,$DB,$D3,$F5,$27,$5E,$2E,$FE,$5E,$65,$FE,$3C,$BF,$DB,$B7
	.dc.b	$E6,$BA,$7B,$D2,$30,$6F,$DF,$F0,$7C,$4C,$8C,$1C,$FE,$2B,$C9,$F5
	.dc.b	$B2,$30,$77,$F5,$3D,$1E,$A4,$8C,$1E,$3F,$53,$D8,$47,$CD,$EA,$7B
	.dc.b	$FE,$5A,$47,$A5,$F8,$5F,$01,$F0,$FE,$02,$4F,$B5,$F8,$DF,$61,$F8
	.dc.b	$BF,$65,$C3,$49,$5F,$E5,$EF,$A4,$72,$7E,$47,$E9,$1E,$E5,$25,$7F
	.dc.b	$99,$21,$F3,$DB,$F9,$C9,$39,$79,$7C,$67,$3F,$9E,$F1,$74,$CA,$E5
	.dc.b	$1D,$FE,$37,$51,$83,$B3,$FA,$3F,$A3,$FC,$DC,$7F,$9D,$3E,$91,$F1
	.dc.b	$5F,$A7,$6C,$7F,$9E,$27,$6F,$F7,$D3,$FB,$AA,$F3,$FA,$7A,$12,$78
	.dc.b	$47,$EC,$7B,$BE,$1C,$7F,$A0,$BC,$C6,$5F,$B5,$6C,$7F,$A1,$A7,$6F
	.dc.b	$70,$FD,$07,$5C,$5D,$4E,$DC,$9F,$BF,$FA,$47,$AB,$D7,$96,$5F,$FA
	.dc.b	$4A,$72,$DF,$57,$85,$48,$C1,$BF,$57,$A4,$C8,$C1,$CF,$6F,$57,$24
	.dc.b	$8C,$1D,$F7,$F2,$4C,$8C,$1E,$3F,$4F,$FC,$5A,$FF,$86,$0F,$F1,$0E
	.dc.b	$48,$CC,$66,$31,$E5,$6E,$37,$3A,$7F,$C4,$32,$F5,$1F,$49,$B9,$4E
	.dc.b	$63,$0D,$8E,$8D,$78,$7D,$27,$F5,$02,$3F,$DC,$1F,$EE,$3E,$5B,$CD
	.dc.b	$FF,$5F,$CF,$F7,$A6,$A9,$FF,$1E,$FD,$27,$DC,$8F,$F1,$BD,$71,$7F
	.dc.b	$00,$7E,$AA,$B9,$F2,$67,$4D,$4F,$F9,$1A,$B1,$5F,$26,$AC,$FF,$90
	.dc.b	$B9,$2B,$FE,$4C,$6F,$33,$E7,$B3,$3F,$BC,$BF,$4F,$6E,$0C,$DD,$8F
	.dc.b	$6F,$9B,$C0,$C5,$7A,$1E,$F9,$B9,$A9,$E0,$F0,$2D,$FF,$B9,$27,$FA
	.dc.b	$6B,$79,$33,$87,$E8,$67,$2F,$E8,$FE,$93,$E6,$37,$7F,$5F,$81,$EE
	.dc.b	$9C,$D7,$C6,$7E,$55,$71,$E0,$D5,$ED,$E1,$E6,$71,$F3,$2E,$7F,$EF
	.dc.b	$8C,$3F,$D3,$B8,$7F,$9D,$1A,$9A,$7F,$BD,$8B,$E5,$78,$CF,$A3,$E2
	.dc.b	$7E,$D2,$BA,$47,$9F,$CA,$91,$83,$C0,$FA,$FC,$A9,$18,$3C,$8F,$DC
	.dc.b	$FA,$47,$3D,$31,$EB,$DA,$FF,$45,$F0,$7E,$23,$07,$2F,$E1,$1B,$3A
	.dc.b	$3F,$D1,$2F,$F1,$07,$1A,$FF,$A5,$23,$6F,$F1,$55,$9F,$33,$A3,$67
	.dc.b	$8C,$BE,$93,$F6,$91,$85,$ED,$7E,$59,$CC,$FF,$21,$73,$3C,$BE,$93
	.dc.b	$EE,$BF,$E9,$E7,$EE,$E5,$FA,$47,$35,$FE,$AA,$73,$3B,$E2,$FD,$1C
	.dc.b	$7D,$A0,$E3,$8B,$F4,$FF,$49,$FE,$BC,$C3,$F5,$FF,$49,$FE,$78,$C3
	.dc.b	$D7,$39,$7F,$D5,$F1,$9C,$7F,$9C,$2F,$77,$F6,$2F,$D6,$9F,$86,$E5
	.dc.b	$BC,$3A,$F6,$3F,$6F,$C9,$FE,$C2,$9D,$C6,$30,$FD,$DF,$BA,$F9,$1A
	.dc.b	$63,$C1,$B7,$F4,$25,$F3,$5C,$0E,$3F,$23,$67,$EA,$FD,$27,$EB,$B8
	.dc.b	$3F,$86,$B6,$3F,$D9,$6F,$F5,$F7,$7E,$07,$F0,$CF,$A4,$F9,$31,$F7
	.dc.b	$9F,$D2,$98,$D3,$FA,$D3,$FA,$8F,$A4,$FE,$4D,$3F,$84,$7F,$DD,$6F
	.dc.b	$FB,$61,$CC,$FF,$61,$72,$FF,$B9,$17,$CE,$7F,$1C,$5F,$3F,$63,$0F
	.dc.b	$F2,$0F,$A4,$F5,$57,$FB,$B1,$A9,$F6,$3F,$DB,$DC,$8F,$83,$E4,$73
	.dc.b	$E2,$B8,$FE,$73,$E9,$3F,$4D,$7F,$DD,$DB,$B7,$F3,$67,$FD,$DE,$DF
	.dc.b	$5F,$F9,$9F,$D1,$BE,$4F,$F9,$DD,$FF,$BE,$23,$18,$7F,$57,$6C,$7F
	.dc.b	$5F,$6B,$FD,$9D,$FF,$5A,$46,$FE,$53,$9C,$84,$E9,$E5,$1F,$EE,$8F
	.dc.b	$0B,$FA,$2F,$C0,$9F,$2C,$61,$FE,$93,$F1,$BF,$4A,$BF,$D2,$3F,$3B
	.dc.b	$8C,$72,$C7,$FB,$74,$A3,$C7,$3C,$27,$31,$89,$7D,$1D,$24,$F2,$6D
	.dc.b	$7F,$88,$E7,$F4,$13,$FC,$31,$E2,$DE,$CF,$3F,$1F,$A3,$E7,$2C,$ED
	.dc.b	$C3,$E6,$7C,$44,$7F,$C0,$DA,$3B,$BE,$29,$C3,$1F,$F1,$A4,$66,$7C
	.dc.b	$5F,$47,$F1,$8F,$8F,$EE,$D8,$C9,$F1,$23,$E0,$11,$F8,$FD,$DF,$0C
	.dc.b	$8E,$09,$7F,$07,$BB,$E9,$CF,$D4,$7F,$BE,$FD,$1F,$FB,$12,$BF,$97
	.dc.b	$BF,$D3,$EB,$FC,$94,$E5,$BF,$01,$C0,$FA,$B1,$B8,$FF,$19,$35,$FC
	.dc.b	$50,$70,$FE,$35,$ED,$CF,$FE,$14,$CD,$F1,$1D,$70,$FE,$83,$18,$7C
	.dc.b	$BE,$FF,$A0,$F4,$FC,$05,$8C,$93,$1F,$4C,$65,$F5,$F9,$1F,$03,$C7
	.dc.b	$F8,$2D,$AF,$83,$1E,$C2,$3F,$E4,$1C,$FF,$99,$57,$F9,$25,$7C,$7F
	.dc.b	$CA,$3E,$5F,$C8,$5A,$F9,$71,$F1,$88,$F3,$6D,$7E,$13,$AF,$83,$67
	.dc.b	$FB,$0F,$D1,$7F,$54,$86,$BF,$E9,$09,$CF,$B3,$FB,$2F,$89,$91,$F2
	.dc.b	$1F,$F1,$CD,$F1,$AE,$17,$FE,$85,$8F,$E1,$F5,$F5,$9B,$1F,$C0,$5F
	.dc.b	$EE,$FC,$3F,$E4,$87,$30,$69,$3D,$5E,$FE,$17,$83,$D0,$98,$7A,$BD
	.dc.b	$1B,$87,$86,$6E,$3D,$66,$D7,$8E,$6D,$78,$E6,$BE,$18,$B8,$FF,$47
	.dc.b	$BF,$34,$60,$F8,$35,$6F,$E5,$C6,$1A,$4B,$F2,$3E,$73,$15,$D1,$A7
	.dc.b	$C7,$AE,$93,$C0,$F9,$8E,$27,$84,$62,$78,$BF,$36,$BD,$35,$B9,$F8
	.dc.b	$07,$86,$7A,$57,$22,$47,$84,$E7,$1F,$0D,$B7,$F0,$DE,$71,$C9,$B3
	.dc.b	$F3,$DD,$DF,$89,$A7,$2B,$F2,$57,$9E,$3C,$95,$E2,$8C,$1D,$4F,$AD
	.dc.b	$1C,$9C,$1F,$87,$9E,$6F,$9C,$F8,$86,$AF,$FC,$CA,$48,$DB,$67,$A5
	.dc.b	$78,$6B,$A4,$77,$7F,$D8,$52,$BF,$E9,$2D,$9F,$F5,$D3,$67,$FD,$84
	.dc.b	$D7,$8C,$F0,$24,$F0,$7F,$AF,$FD,$87,$E4,$B3,$34,$93,$F5,$23,$2D
	.dc.b	$8C,$D7,$C4,$9C,$B6,$33,$87,$B6,$72,$D8,$CD,$7D,$A3,$F9,$6C,$E9
	.dc.b	$C5,$7F,$79,$8C,$3F,$5A,$BF,$97,$57,$28,$C5,$72,$8C,$47,$D5,$9D
	.dc.b	$CF,$A4,$91,$E7,$6C,$79,$5A,$78,$A3,$F8,$BC,$F4,$C3,$FD,$43,$7E
	.dc.b	$49,$ED,$5E,$13,$8B,$E4,$3D,$55,$C7,$0F,$A3,$3A,$9E,$26,$9C,$5F
	.dc.b	$6A,$E1,$1F,$D2,$9B,$3F,$DC,$9B,$87,$BA,$3D,$84,$7D,$8D,$9F,$E4
	.dc.b	$DD,$9F,$EE,$4F,$A3,$FF,$27,$D7,$F2,$58,$C4,$FF,$30,$6B,$FC,$E9
	.dc.b	$B5,$FC,$C7,$E9,$FF,$2A,$49,$FC,$AB,$FE,$20,$FB,$0F,$F0,$AC,$AF
	.dc.b	$FD,$F3,$18,$7F,$9A,$4F,$FB,$DA,$31,$5F,$84,$DB,$FC,$6A,$7F,$BE
	.dc.b	$BE,$91,$F4,$B8,$9F,$ED,$B3,$FE,$2E,$DF,$FC,$A7,$97,$F7,$83,$FA
	.dc.b	$7C,$64,$FF,$19,$F2,$BE,$8A,$92,$F8,$57,$F9,$26,$75,$18,$86,$2F
	.dc.b	$5B,$FC,$7A,$D9,$D3,$67,$9B,$5F,$25,$7C,$46,$F9,$1E,$42,$57,$79
	.dc.b	$FF,$1F,$CF,$91,$AF,$2F,$1F,$D6,$7C,$1F,$95,$95,$FE,$3A,$6C,$F8
	.dc.b	$BE,$AF,$CE,$49,$F3,$BF,$27,$9A,$47,$F9,$09,$B3,$F1,$39,$3D,$0F
	.dc.b	$F8,$91,$AF,$F9,$BD,$A9,$BB,$BE,$57,$EB,$12,$FF,$C9,$EF,$F9,$9A
	.dc.b	$7D,$B5,$D7,$2C,$12,$BE,$97,$D2,$FD,$9F,$9F,$BC,$27,$3F,$4A,$EE
	.dc.b	$CC,$7D,$9F,$B0,$FF,$6F,$E5,$7B,$7E,$E2,$8C,$BF,$82,$6B,$F8,$BF
	.dc.b	$A6,$7D,$0F,$06,$C7,$E1,$5A,$FE,$29,$AF,$E4,$FD,$57,$87,$E0,$6C
	.dc.b	$7E,$35,$AF,$E4,$9A,$F8,$7D,$2F,$01,$1F,$03,$EA,$FA,$5F,$F1,$DC
	.dc.b	$7E,$5D,$AF,$E7,$1A,$FE,$95,$AF,$F9,$7B,$E9,$BF,$D3,$0F,$E7,$DA
	.dc.b	$FE,$91,$AF,$EB,$5A,$FF,$99,$BE,$97,$CE,$DC,$3E,$6C,$62,$3F,$D6
	.dc.b	$8D,$5F,$4D,$F5,$DF,$D7,$35,$FD,$9B,$5F,$DE,$35,$FF,$3A,$FD,$37
	.dc.b	$FA,$B9,$34,$9C,$9F,$7B,$E8,$FF,$DA,$13,$FB,$B6,$BF,$C0,$1A,$FF
	.dc.b	$09,$6B,$FE,$82,$7D,$04,$C3,$A1,$F8,$AF,$A6,$69,$FE,$88,$F1,$98
	.dc.b	$7F,$A2,$A7,$38,$7F,$B5,$67,$2F,$FB,$5E,$32,$FF,$27,$9E,$1E,$9F
	.dc.b	$37,$97,$CE,$59,$E0,$D5,$48,$C2,$F9,$5B,$F0,$D2,$BC,$57,$E5,$8D
	.dc.b	$FA,$9C,$3D,$5E,$3F,$39,$2F,$67,$9E,$3A,$FC,$77,$6F,$A6,$93,$A3
	.dc.b	$C7,$1E,$2F,$63,$C7,$EC,$49,$D6,$A9,$F3,$FC,$8C,$9C,$BE,$AF,$8C
	.dc.b	$F5,$FB,$32,$78,$BE,$EF,$37,$DF,$F4,$D2,$7A,$BF,$0B,$E2,$3E,$1F
	.dc.b	$C4,$7C,$7E,$AF,$E8,$DD,$1F,$4A,$D1,$8E,$5F,$B1,$5B,$F1,$B8,$7C
	.dc.b	$B4,$BF,$75,$FE,$5F,$0E,$4B,$E6,$7B,$7C,$D4,$AD,$C7,$87,$DE,$F1
	.dc.b	$7B,$12,$79,$C7,$97,$E9,$BC,$DE,$B4,$9E,$F1,$E9,$F3,$3D,$5E,$AC
	.dc.b	$9F,$1C,$7B,$7E,$9A,$47,$99,$F8,$1E,$AF,$C1,$F9,$79,$3E,$A7,$E2
	.dc.b	$7C,$67,$C5,$F8,$09,$3E,$E7,$E4,$7C,$2F,$E3,$FD,$AF,$74,$FF,$B7
	.dc.b	$B8,$30,$7E,$4F,$E9,$9C,$BF,$95,$FA,$67,$D1,$D2,$5F,$E1,$E1,$F3
	.dc.b	$FB,$F9,$D9,$39,$79,$7C,$77,$3F,$9F,$F1,$74,$CA,$E5,$1D,$FE,$93
	.dc.b	$51,$83,$B3,$FA,$7F,$A5,$79,$FF,$A9,$FA,$67,$C7,$53,$51,$BF,$46
	.dc.b	$E1,$F9,$DB,$E2,$D8,$FD,$97,$7F,$C0,$8F,$D1,$79,$AC,$BF,$6C,$D8
	.dc.b	$FD,$34,$ED,$FE,$00,$03,$F3,$9C,$98,$E2,$EE,$78,$39,$3F,$81,$7D
	.dc.b	$2B,$D5,$EF,$CB,$2F,$F6,$93,$96,$FA,$BC,$2A,$46,$0D,$FA,$BD,$26
	.dc.b	$46,$0E,$7B,$FB,$39,$24,$60,$EF,$BF,$92,$64,$60,$F1,$FA,$9F,$E4
	.dc.b	$57,$FC,$63,$92,$31,$1F,$E3,$B8,$C7,$8F,$FE,$33,$6C,$70,$F5,$5F
	.dc.b	$4D,$96,$C6,$63,$07,$F8,$01,$7D,$35,$23,$FE,$04,$7F,$C0,$BE,$5F
	.dc.b	$F3,$6C,$3F,$83,$4F,$F8,$37,$CD,$FE,$3C,$FD,$7A,$FB,$1F,$4D,$F6
	.dc.b	$23,$ED,$1B,$3E,$C9,$FA,$49,$E3,$C9,$D5,$FF,$2C,$61,$67,$8C,$D7
	.dc.b	$F0,$9C,$95,$FE,$1D,$5F,$8E,$F8,$DC,$4E,$7B,$64,$F9,$A2,$D7,$8B
	.dc.b	$CD,$E4,$63,$4F,$6B,$DD,$FE,$91,$5F,$CE,$DB,$3F,$92,$E9,$FD,$BC
	.dc.b	$E7,$9B,$87,$1C,$8C,$1C,$5F,$57,$D3,$7C,$3C,$3F,$98,$36,$7E,$3D
	.dc.b	$BB,$F7,$BF,$0D,$E5,$7F,$4D,$BE,$19,$FE,$9D,$F9,$8F,$E9,$97,$97
	.dc.b	$E0,$F3,$3C,$2D,$AE,$51,$6B,$84,$DC,$FC,$BE,$57,$A0,$FA,$5E,$1F
	.dc.b	$EB,$70,$F1,$47,$B3,$CA,$91,$83,$CC,$FC,$0F,$2A,$46,$0F,$53,$F7
	.dc.b	$7E,$99,$EB,$5F,$5B,$0D,$AF,$ED,$96,$F4,$3F,$08,$BD,$1B,$18,$73
	.dc.b	$8E,$BC,$FB,$7C,$9F,$C6,$7D,$33,$DC,$A6,$3D,$AB,$5F,$DE,$FE,$9B
	.dc.b	$F1,$FC,$7F,$C4,$9B,$C5,$FD,$A5,$77,$7E,$5D,$CD,$7E,$89,$7A,$B9
	.dc.b	$35,$EA,$D9,$FF,$04,$7D,$2F,$F9,$B9,$BF,$11,$FA,$18,$FF,$16,$C6
	.dc.b	$0D,$FB,$B9,$35,$C9,$63,$FC,$25,$C3,$F8,$6D,$4D,$3F,$0F,$AF,$F8
	.dc.b	$05,$D5,$3E,$8B,$F6,$0B,$E7,$AC,$61,$FD,$A7,$D2,$FF,$A4,$71,$FF
	.dc.b	$10,$3E,$9C,$7F,$9B,$B9,$6E,$9C,$3C,$DE,$3E,$65,$BE,$2F,$EF,$FE
	.dc.b	$97,$E3,$C7,$F8,$AD,$B7,$FD,$09,$FF,$56,$46,$78,$7F,$83,$7D,$2F
	.dc.b	$FB,$9A,$3F,$C6,$4D,$AF,$E9,$2B,$E4,$37,$F3,$1C,$3E,$86,$5F,$0F
	.dc.b	$5E,$E1,$FD,$43,$E9,$7F,$9A,$4F,$F8,$F5,$69,$FC,$55,$CC,$EE,$3F
	.dc.b	$8C,$7D,$2F,$F5,$BA,$FF,$22,$3F,$4C,$C4,$FF,$5C,$71,$DF,$C3,$C6
	.dc.b	$E2,$DA,$FF,$26,$7D,$4F,$DA,$7F,$DE,$8D,$BF,$F7,$CB,$5F,$E6,$21
	.dc.b	$C3,$FC,$D3,$E2,$FF,$CE,$B3,$FE,$FD,$6C,$70,$3F,$1F,$19,$3F,$A0
	.dc.b	$BF,$DD,$61,$F4,$BF,$CD,$5B,$3F,$E6,$17,$9F,$57,$F9,$35,$CD,$7F
	.dc.b	$7A,$5F,$39,$C4,$E5,$F3,$16,$F9,$47,$82,$BF,$D1,$AD,$AF,$F7,$79
	.dc.b	$3A,$70,$4A,$F6,$36,$BC,$1C,$3E,$78,$F4,$3F,$D3,$73,$FF,$6A,$D7
	.dc.b	$A7,$DC,$C7,$C7,$FB,$59,$1F,$1D,$F1,$11,$F9,$5F,$4B,$F0,$CD,$E3
	.dc.b	$EF,$7E,$37,$D2,$C9,$F8,$DF,$07,$6B,$E2,$4F,$B4,$93,$F5,$5F,$85
	.dc.b	$F4,$9F,$C0,$A7,$EB,$35,$FB,$ED,$7F,$B8,$B5,$FF,$16,$3C,$78,$3F
	.dc.b	$D0,$0F,$F3,$B8,$FF,$16,$B9,$AE,$3F,$07,$F1,$7E,$8F,$FB,$94,$BE
	.dc.b	$0F,$15,$FE,$9F,$8C,$36,$3C,$E4,$7F,$19,$69,$DB,$EA,$9E,$BE,$07
	.dc.b	$66,$D7,$5F,$2F,$C2,$78,$BC,$CB,$19,$25,$FA,$A3,$2F,$F9,$23,$EA
	.dc.b	$7F,$81,$69,$F0,$3D,$2F,$F0,$64,$62,$FE,$3C,$7B,$8F,$9E,$47,$D3
	.dc.b	$9A,$3F,$29,$F4,$BF,$BA,$BF,$F5,$53,$6B,$FC,$43,$6E,$9F,$F0,$D4
	.dc.b	$E7,$1F,$F2,$F3,$FC,$09,$FD,$BB,$5F,$F1,$4B,$5D,$9E,$03,$9F,$B5
	.dc.b	$D1,$B7,$FE,$D1,$9F,$EA,$6D,$F7,$FC,$3E,$8F,$89,$91,$93,$84,$62
	.dc.b	$BF,$CD,$F5,$FE,$0A,$9E,$9D,$FF,$23,$CB,$F5,$B2,$32,$7F,$B8,$73
	.dc.b	$FE,$94,$E2,$F9,$FA,$BF,$08,$FA,$3D,$96,$7D,$11,$67,$EC,$4D,$DB
	.dc.b	$E6,$CD,$AF,$9B,$36,$BE,$5C,$D5,$69,$C6,$4A,$D4,$E6,$FC,$24,$BF
	.dc.b	$3B,$EB,$BE,$35,$C1,$B1,$BC,$3F,$DD,$57,$CE,$F8,$92,$3D,$33,$CC
	.dc.b	$E0,$7D,$EB,$E8,$E2,$7D,$6D,$7D,$77,$E0,$7A,$47,$C2,$35,$87,$B5
	.dc.b	$E1,$87,$8F,$0F,$0D,$70,$3A,$9F,$3A,$78,$E1,$E2,$8D,$0B,$E8,$E2
	.dc.b	$FA,$57,$A4,$E2,$7B,$63,$8C,$F0,$BF,$09,$F1,$B0,$F5,$38,$AF,$55
	.dc.b	$70,$C3,$C5,$39,$8F,$3B,$5F,$87,$3E,$07,$C9,$5D,$23,$C9,$1E,$D3
	.dc.b	$A3,$F1,$23,$E5,$47,$EC,$2B,$F7,$64,$9F,$E2,$F7,$F0,$27,$51,$89
	.dc.b	$D4,$7F,$67,$9F,$74,$FE,$E3,$CE,$FC,$79,$1F,$AA,$D9,$F8,$AD,$C3
	.dc.b	$D9,$5F,$DE,$DB,$19,$AF,$5C,$7F,$0E,$8C,$57,$C4,$9F,$01,$D8,$8F
	.dc.b	$66,$CF,$46,$9D,$0E,$31,$B2,$46,$DB,$1B,$7F,$8C,$CF,$89,$FC,$35
	.dc.b	$FF,$1E,$8C,$5F,$C4,$9D,$D7,$81,$F4,$18,$3A,$98,$3F,$21,$3C,$AB
	.dc.b	$A4,$F1,$9E,$EF,$F1,$F9,$D3,$4D,$57,$86,$79,$C6,$2B,$9C,$78,$0E
	.dc.b	$67,$F2,$0A,$E6,$D9,$E7,$D9,$FC,$EB,$E2,$7F,$9D,$B6,$3F,$98,$36
	.dc.b	$7F,$94,$B7,$14,$FF,$B9,$3E,$85,$F4,$DF,$E1,$D6,$C7,$F6,$66,$DF
	.dc.b	$F8,$CD,$AF,$F8,$F3,$E9,$3F,$C5,$77,$D2,$02,$33,$5D,$EF,$FD,$D5
	.dc.b	$18,$9D,$FC,$8F,$FA,$AE,$B8,$DF,$39,$4E,$54,$67,$E7,$9F,$4D,$FE
	.dc.b	$F5,$AD,$4F,$17,$85,$7F,$94,$9A,$F9,$8E,$BE,$E7,$F8,$F6,$33,$1E
	.dc.b	$AC,$5F,$49,$FE,$41,$6E,$5F,$E0,$A6,$CF,$F9,$5D,$AF,$F9,$83,$8D
	.dc.b	$E1,$7F,$92,$5B,$5F,$E1,$26,$CF,$F9,$7D,$AF,$CD,$E4,$F0,$12,$7C
	.dc.b	$0F,$F9,$4F,$E6,$7E,$9C,$FA,$DA,$FD,$26,$A6,$E5,$FD,$AE,$3F,$CD
	.dc.b	$CD,$7E,$37,$37,$C3,$3F,$CC,$1F,$2C,$E4,$A4,$C7,$FC,$57,$19,$7F
	.dc.b	$C5,$B7,$FE,$A6,$BF,$A1,$EA,$FC,$D2,$4E,$BD,$CF,$A0,$47,$07,$F8
	.dc.b	$DB,$EA,$7D,$24,$E5,$FF,$39,$3E,$CF,$A5,$FF,$0E,$7D,$2F,$F8,$F1
	.dc.b	$B8,$FF,$A2,$5F,$F5,$77,$81,$9F,$05,$5F,$E4,$2F,$A4,$B7,$63,$FC
	.dc.b	$8A,$A3,$2F,$FA,$4D,$AF,$FA,$71,$7B,$9C,$1B,$1D,$63,$07,$4E,$B7
	.dc.b	$95,$49,$3F,$E5,$48,$CB,$FE,$56,$FA,$3F,$F4,$82,$69,$36,$F3,$FD
	.dc.b	$27,$DD,$9F,$F2,$EB,$5F,$F5,$53,$5F,$F5,$9B,$5F,$D1,$7D,$1F,$C0
	.dc.b	$4D,$23,$F1,$3E,$91,$E1,$52,$47,$F9,$B2,$32,$FF,$9B,$BE,$8B,$E3
	.dc.b	$BF,$AF,$6B,$FC,$19,$A9,$BF,$49,$F0,$09,$8F,$D4,$7E,$1F,$47,$F0
	.dc.b	$0A,$FC,$8F,$87,$F6,$23,$11,$FE,$C9,$6B,$FE,$D2,$7E,$6B,$9D,$B8
	.dc.b	$11,$FF,$42,$FE,$C3,$F4,$09,$87,$E1,$1C,$7B,$5F,$C2,$7C,$5F,$E2
	.dc.b	$D0,$F8,$0C,$57,$F0,$D8,$C5,$FF,$1C,$8C,$5E,$A3,$F9,$3C,$BE,$07
	.dc.b	$10,$8F,$F3,$C3,$C4,$DF,$3E,$A3,$84,$64,$E1,$EE,$67,$C7,$91,$99
	.dc.b	$FF,$6F,$B7,$E5,$B9,$11,$D7,$9D,$D7,$DC,$91,$C5,$F0,$78,$FE,$1F
	.dc.b	$82,$93,$D5,$F2,$7C,$FF,$97,$E3,$FC,$DE,$F4,$AF,$14,$7A,$7D,$EF
	.dc.b	$57,$D3,$C9,$F3,$47,$B7,$E5,$3D,$DF,$01,$27,$D5,$1F,$07,$D1,$91
	.dc.b	$ED,$7E,$27,$95,$F2,$A7,$F7,$3D,$3F,$D2,$DB,$D1,$FD,$05,$B7,$FD
	.dc.b	$61,$FD,$CB,$FD,$3A,$35,$F0,$5C,$C9,$1F,$3E,$5F,$03,$DF,$E7,$64
	.dc.b	$E5,$F1,$7C,$C7,$8F,$E6,$24,$F1,$7C,$DF,$07,$E7,$F8,$49,$3D,$5F
	.dc.b	$57,$C4,$7A,$FE,$9B,$D9,$F0,$B2,$BC,$51,$EF,$F8,$6F,$81,$F3,$32
	.dc.b	$7C,$D1,$F0,$FE,$43,$E2,$7C,$84,$9F,$54,$7E,$43,$E8,$DF,$35,$F9
	.dc.b	$2F,$75,$F3,$74,$D4,$7E,$5B,$E8,$DE,$05,$F0,$66,$5F,$93,$9C,$B7
	.dc.b	$C1,$E1,$E0,$C8,$C1,$BE,$9F,$07,$C5,$48,$C1,$CF,$C0,$F0,$4C,$8C
	.dc.b	$1F,$A1,$FA,$27,$4F,$EA,$BE,$8F,$FA,$C4,$FE,$95,$AF,$DD,$3F,$C2
	.dc.b	$F3,$C1,$E5,$D5,$FB,$28,$C4,$6C,$38,$F0,$3F,$C2,$CA,$7F,$73,$F4
	.dc.b	$7F,$2A,$BF,$69,$D0,$CF,$F7,$AD,$8F,$DC,$4E,$DE,$CF,$D0,$73,$C5
	.dc.b	$D0,$EB,$C7,$FC,$1B,$E8,$FF,$23,$3D,$B8,$FA,$99,$F0,$35,$E0,$59
	.dc.b	$FE,$0D,$5C,$1F,$62,$8E,$9E,$07,$09,$DC,$61,$E3,$C3,$EA,$E5,$91
	.dc.b	$83,$AE,$FE,$59,$91,$83,$C3,$E0,$7F,$8E,$9F,$F1,$91,$FE,$48,$C5
	.dc.b	$19,$8C,$C6,$38,$B7,$1B,$9D,$3F,$E3,$B9,$78,$0F,$A3,$DC,$A7,$31
	.dc.b	$86,$C7,$46,$BC,$3E,$8F,$FB,$21,$1F,$F0,$2B,$FE,$06,$F8,$2F,$F3
	.dc.b	$06,$3F,$E5,$79,$46,$3C,$5F,$3D,$7F,$99,$3E,$8D,$CB,$F0,$2B,$57
	.dc.b	$FE,$53,$E5,$AF,$F9,$A9,$AF,$4C,$3F,$9F,$3F,$B6,$6F,$2E,$8E,$3C
	.dc.b	$CF,$3B,$13,$C3,$B3,$99,$8B,$FC,$E0,$7B,$FC,$94,$9E,$07,$81,$F2
	.dc.b	$EC,$E7,$A7,$83,$C0,$B5,$FD,$4E,$7F,$66,$DE,$5C,$E3,$FA,$B9,$CB
	.dc.b	$FA,$DF,$A3,$F7,$37,$7F,$5F,$7F,$C4,$9C,$D7,$FA,$15,$F8,$55,$C7
	.dc.b	$83,$57,$B7,$87,$C5,$F1,$FA,$3B,$9F,$F6,$7C,$11,$98,$C3,$AC,$3C
	.dc.b	$8F,$87,$E8,$E4,$FF,$6F,$F0,$FF,$A0,$BC,$E3,$CD,$F2,$92,$30,$77
	.dc.b	$7D,$5F,$29,$23,$07,$8D,$FC,$0F,$D1,$BC,$BB,$F2,$F8,$6D,$7F,$7F
	.dc.b	$BF,$F4,$FB,$FE,$9B,$5F,$35,$F8,$A7,$3A,$7F,$AD,$1C,$CF,$1C,$75
	.dc.b	$F4,$7B,$6C,$7F,$82,$FE,$8D,$E9,$7E,$51,$B1,$9F,$47,$13,$F9,$8F
	.dc.b	$A3,$FE,$8F,$3F,$E1,$36,$E5,$FC,$A5,$FC,$74,$67,$BF,$87,$CA,$48
	.dc.b	$E1,$F2,$96,$FF,$98,$7D,$1F,$F9,$BA,$7F,$C3,$CD,$8C,$FA,$FF,$A8
	.dc.b	$8C,$75,$FF,$B2,$57,$AD,$93,$5C,$77,$0F,$F1,$47,$07,$E3,$1F,$F6
	.dc.b	$23,$71,$FF,$6D,$BF,$8E,$C3,$FD,$A8,$E6,$3E,$B7,$BD,$FB,$8B,$7B
	.dc.b	$D6,$30,$FE,$F3,$E8,$FF,$D2,$38,$FF,$8D,$DF,$3B,$FE,$D3,$73,$7F
	.dc.b	$D3,$DC,$BF,$EE,$C7,$F9,$2C,$71,$F3,$7A,$79,$DD,$5F,$E1,$3F,$46
	.dc.b	$E1,$A6,$38,$6D,$FF,$91,$FE,$8F,$FC,$B0,$DF,$88,$FE,$23,$1F,$E5
	.dc.b	$38,$C1,$BF,$82,$C9,$AF,$5E,$C7,$F9,$47,$E8,$FC,$BB,$7F,$67,$7F
	.dc.b	$99,$C6,$75,$FE,$3B,$18,$9F,$ED,$9F,$46,$7F,$90,$BF,$34,$C5,$7F
	.dc.b	$97,$5C,$77,$EB,$C6,$DB,$67,$FC,$C5,$F4,$6E,$6F,$E5,$7D,$49,$FE
	.dc.b	$5B,$39,$7F,$CD,$11,$96,$FC,$5F,$03,$8B,$4C,$9F,$D3,$21,$1F,$D0
	.dc.b	$72,$F5,$46,$27,$53,$99,$4B,$E2,$70,$DA,$E9,$18,$23,$E6,$F8,$9F
	.dc.b	$5C,$7B,$1F,$6F,$47,$BB,$86,$D7,$B1,$C3,$E7,$7B,$91,$F5,$7B,$7E
	.dc.b	$22,$38,$3D,$DF,$0F,$F2,$8F,$97,$16,$3E,$4C,$7A,$88,$F0,$3E,$37
	.dc.b	$C3,$FB,$88,$E0,$99,$FF,$8B,$73,$7C,$3F,$D8,$AF,$B2,$FF,$64,$FA
	.dc.b	$3F,$93,$7F,$E9,$F7,$F9,$8D,$FF,$8D,$5C,$B7,$E0,$78,$1F,$46,$37
	.dc.b	$1F,$E1,$06,$BF,$91,$0E,$1F,$CA,$3F,$E3,$D9,$FF,$05,$B6,$3F,$C2
	.dc.b	$6D,$7A,$7C,$07,$7E,$7F,$41,$E9,$6D,$7A,$62,$CF,$A6,$6F,$85,$F9
	.dc.b	$78,$F0,$F9,$1E,$92,$5F,$89,$B1,$F0,$7D,$BF,$8D,$1F,$1D,$F9,$1F
	.dc.b	$3D,$2B,$E4,$79,$D6,$FE,$44,$FC,$52,$4F,$B9,$FC,$C7,$D1,$FE,$CA
	.dc.b	$7F,$CC,$4F,$F2,$13,$18,$F9,$E1,$19,$F5,$B8,$92,$FF,$C6,$71,$FD
	.dc.b	$FA,$7F,$C8,$2F,$F2,$BF,$88,$7C,$FF,$F8,$8A,$30,$DF,$77,$1F,$01
	.dc.b	$70,$FF,$70,$57,$F8,$B1,$B3,$A6,$9A,$9E,$9E,$C4,$8E,$3E,$0D,$CB
	.dc.b	$DB,$37,$4F,$66,$2A,$C5,$FE,$1E,$3F,$85,$5F,$C4,$F0,$3D,$1E,$AF
	.dc.b	$A5,$95,$E6,$72,$D2,$4F,$17,$D7,$3C,$23,$04,$C3,$C7,$19,$7C,$B5
	.dc.b	$6B,$C9,$36,$BF,$96,$5F,$9A,$F7,$5D,$8F,$0B,$89,$ED,$1C,$8C,$9E
	.dc.b	$FB,$E9,$1F,$7A,$ED,$F5,$AD,$1D,$67,$99,$E2,$AD,$D7,$06,$CF,$69
	.dc.b	$D3,$5E,$CD,$AE,$D1,$A3,$F2,$75,$CE,$74,$D8,$F0,$9C,$8B,$3C,$9F
	.dc.b	$CB,$D7,$D4,$72,$F9,$1B,$5D,$0F,$14,$66,$FE,$41,$23,$E0,$CF,$91
	.dc.b	$D9,$C5,$E6,$78,$AF,$CD,$3C,$63,$EB,$C6,$27,$9C,$FE,$01,$E7,$3F
	.dc.b	$EA,$2B,$FC,$8B,$FA,$AB,$FD,$B9,$27,$F5,$ED,$8E,$0D,$AF,$DC,$7D
	.dc.b	$1F,$F6,$19,$F8,$ED,$8F,$24,$23,$15,$E4,$3F,$94,$38,$8F,$A9,$3D
	.dc.b	$27,$F0,$F3,$99,$F9,$F1,$FD,$FE,$74,$FD,$F9,$D5,$FD,$56,$DF,$FA
	.dc.b	$91,$B3,$C4,$F0,$9D,$88,$73,$2D,$FB,$1B,$7E,$C9,$E7,$3F,$EF,$D6
	.dc.b	$C7,$92,$38,$4F,$32,$38,$24,$73,$8E,$A7,$C5,$8F,$F0,$E4,$F4,$8F
	.dc.b	$39,$CA,$78,$34,$E1,$3F,$3D,$E6,$FC,$C7,$8B,$E0,$77,$3D,$63,$BC
	.dc.b	$70,$7E,$79,$ED,$71,$7F,$35,$B7,$F3,$5A,$7C,$D9,$E6,$F9,$5F,$11
	.dc.b	$83,$FA,$1E,$BF,$CD,$FE,$4F,$F9,$C3,$67,$F9,$AC,$59,$DC,$60,$EC
	.dc.b	$BE,$8B,$FC,$A6,$FF,$BF,$9A,$FF,$75,$35,$E5,$7F,$8A,$E3,$31,$FD
	.dc.b	$E7,$87,$FA,$12,$D3,$FA,$A4,$61,$FE,$AD,$3D,$9C,$F0,$7F,$BC,$23
	.dc.b	$FD,$D1,$87,$91,$B7,$CE,$3F,$C0,$D0,$F0,$B8,$7C,$1C,$9F,$35,$7F
	.dc.b	$96,$EB,$8B,$E9,$6B,$CA,$7F,$C8,$9C,$BF,$E4,$18,$E2,$D7,$E0,$35
	.dc.b	$FF,$0B,$35,$ED,$E5,$7C,$32,$5F,$8A,$3E,$64,$76,$C2,$4E,$6B,$D4
	.dc.b	$D8,$F9,$9F,$1F,$23,$96,$28,$C6,$1E,$ED,$3F,$DD,$35,$DD,$B1,$9E
	.dc.b	$5F,$F3,$04,$62,$3E,$D7,$C0,$3A,$7F,$CC,$8D,$BF,$43,$67,$E5,$B5
	.dc.b	$FC,$15,$3C,$EE,$A4,$AF,$13,$FE,$6F,$E6,$F6,$D7,$D4,$6B,$F8,$B6
	.dc.b	$BF,$C6,$B4,$FF,$3B,$7D,$03,$E4,$9F,$41,$F2,$E3,$0D,$F9,$47,$76
	.dc.b	$6B,$78,$7E,$0B,$E9,$FF,$C1,$0F,$D3,$6B,$9E,$EF,$AC,$A3,$2F,$E7
	.dc.b	$FE,$85,$F4,$3A,$6C,$7D,$D6,$BF,$90,$6B,$FA,$7F,$A1,$7C,$F7,$66
	.dc.b	$C7,$E1,$5A,$FE,$61,$AF,$EB,$FE,$87,$FD,$30,$9A,$4E,$2F,$4F,$D0
	.dc.b	$BA,$A9,$2B,$F2,$51,$97,$F2,$95,$FE,$A8,$6F,$A9,$F0,$A3,$11,$FA
	.dc.b	$A6,$BF,$C0,$7E,$83,$CA,$FF,$9D,$DA,$FF,$B2,$5A,$9B,$F4,$3F,$5A
	.dc.b	$3F,$48,$D7,$F6,$ED,$7F,$87,$35,$FF,$42,$7D,$07,$C2,$26,$5F,$61
	.dc.b	$F8,$BF,$40,$D3,$ED,$B6,$BF,$64,$D7,$F8,$2B,$5F,$E3,$8B,$E8,$38
	.dc.b	$92,$38,$BD,$FA,$24,$7E,$F6,$FF,$89,$36,$3F,$17,$D9,$F3,$E3,$F5
	.dc.b	$3B,$E0,$8F,$F0,$5F,$A0,$F9,$FE,$2F,$F9,$E5,$78,$DF,$EA,$38,$C3
	.dc.b	$75,$FF,$53,$4F,$06,$F5,$66,$FF,$DF,$53,$FA,$97,$F9,$EB,$5F,$E5
	.dc.b	$46,$A3,$82,$8F,$F7,$23,$C3,$2C,$C7,$BA,$38,$9B,$F6,$B3,$E3,$48
	.dc.b	$EE,$F0,$F8,$4E,$3E,$DC,$9F,$1B,$D3,$DB,$EB,$E4,$C9,$F3,$BE,$0F
	.dc.b	$93,$F0,$FC,$87,$8B,$DD,$95,$EC,$8F,$2F,$BB,$E6,$F3,$24,$FC,$08
	.dc.b	$F4,$F2,$FA,$BE,$6E,$4F,$C4,$8F,$6F,$BB,$23,$E3,$BF,$03,$DD,$F8
	.dc.b	$B3,$FC,$32,$7E,$0B,$79,$FF,$D6,$0D,$E6,$FF,$6D,$3F,$C1,$23,$3E
	.dc.b	$F7,$22,$57,$F7,$09,$7B,$DD,$BE,$6E,$4E,$5F,$0F,$CB,$78,$BE,$5A
	.dc.b	$4F,$17,$CB,$EF,$F9,$BE,$02,$4F,$57,$D3,$F1,$3E,$AF,$A0,$F5,$FC
	.dc.b	$8C,$AF,$14,$7B,$BE,$0F,$DF,$F0,$72,$7C,$D1,$F0,$BE,$3B,$E1,$FC
	.dc.b	$74,$9F,$54,$7E,$2B,$E8,$1F,$31,$F8,$CF,$75,$F2,$54,$D4,$7E,$43
	.dc.b	$E8,$1E,$05,$F0,$66,$5F,$E1,$49,$CB,$7C,$1E,$1E,$0C,$8C,$1B,$E9
	.dc.b	$F0,$7C,$D4,$8C,$1C,$FC,$0F,$04,$C8,$C1,$FE,$22,$F9,$F7,$4F,$E9
	.dc.b	$3E,$83,$FD,$C7,$3F,$E2,$66,$BF,$A2,$9D,$BC,$7A,$BF,$57,$18,$8D
	.dc.b	$56,$27,$F5,$FF,$0D,$F1,$A7,$FC,$63,$CE,$CB,$F6,$6D,$8F,$F1,$AC
	.dc.b	$ED,$FD,$8B,$F3,$1C,$F0,$F4,$3A,$F1,$7E,$F3,$E7,$FF,$1B,$3D,$B8
	.dc.b	$BA,$99,$EF,$D7,$7D,$9F,$F2,$1D,$70,$7F,$7E,$A3,$A7,$7F,$09,$DC
	.dc.b	$61,$E3,$B7,$AB,$92,$46,$0E,$BB,$F9,$66,$46,$0F,$0F,$7F,$F9,$25
	.dc.b	$FF,$1D,$E2,$8C,$47,$F9,$4E,$31,$E0,$FF,$8B,$DB,$1C,$3C,$07,$D0
	.dc.b	$65,$B1,$98,$C1,$FE,$02,$5F,$41,$48,$FF,$81,$9F,$F0,$3F,$C1,$3C
	.dc.b	$5F,$F3,$4C,$FF,$84,$9A,$F2,$1F,$40,$E5,$FF,$35,$57,$19,$FE,$56
	.dc.b	$7E,$A2,$B9,$F2,$E7,$4D,$48,$EC,$70,$8D,$4F,$F9,$D7,$96,$CE,$CF
	.dc.b	$D8,$C7,$E3,$BC,$EC,$C7,$39,$AB,$B2,$98,$8B,$5E,$EF,$90,$D9,$8A
	.dc.b	$F4,$BD,$31,$73,$53,$BF,$4F,$EF,$AF,$FD,$1F,$3A,$6F,$27,$13,$83
	.dc.b	$83,$FD,$2B,$F4,$1F,$02,$7E,$1B,$63,$DD,$4E,$EF,$6C,$E7,$2F,$90
	.dc.b	$FC,$D7,$FA,$CB,$6F,$9E,$FC,$37,$E3,$F0,$9C,$BE,$2E,$E7,$E2,$70
	.dc.b	$D3,$CB,$18,$C7,$AF,$D0,$78,$8F,$9F,$E1,$7F,$3E,$AF,$5C,$79,$BE
	.dc.b	$4A,$46,$0F,$7B,$EA,$F9,$19,$18,$3E,$1B,$F6,$BE,$F3,$E6,$61,$FD
	.dc.b	$D1,$BA,$FE,$23,$E8,$7F,$C8,$97,$FE,$BE,$9C,$BF,$85,$F8,$CF,$F5
	.dc.b	$6D,$7E,$39,$B1,$9F,$6F,$F0,$F1,$8C,$3F,$24,$BD,$4C,$9A,$F1,$ED
	.dc.b	$FF,$B3,$3E,$F9,$EB,$7E,$41,$B1,$FE,$D0,$FA,$17,$46,$6B,$F2,$BF
	.dc.b	$6D,$8A,$FF,$6A,$B7,$3F,$F0,$8A,$F5,$37,$7F,$9C,$FC,$33,$F3,$57
	.dc.b	$FE,$DC,$6D,$FF,$86,$1F,$D1,$B7,$C0,$F9,$5E,$D4,$AF,$F7,$1F,$D1
	.dc.b	$7D,$7A,$F9,$93,$FA,$8F,$BF,$FE,$6D,$5F,$EE,$86,$CF,$F8,$91,$FC
	.dc.b	$94,$67,$7F,$D7,$AF,$26,$98,$F2,$EC,$7F,$BC,$BA,$3F,$CB,$4D,$F8
	.dc.b	$8F,$DB,$FB,$2F,$7B,$F7,$35,$AC,$FF,$C6,$7F,$71,$EC,$EA,$FE,$0E
	.dc.b	$DC,$7F,$81,$3F,$9F,$6F,$93,$FC,$0A,$17,$4C,$77,$5A,$FC,$47,$BF
	.dc.b	$F8,$F3,$F8,$4F,$D1,$BC,$1C,$4F,$F1,$5C,$BF,$8B,$F9,$EF,$4F,$F8
	.dc.b	$7B,$67,$47,$F1,$55,$E2,$6A,$7F,$8A,$FD,$C7,$F0,$5D,$FF,$C9,$ED
	.dc.b	$C3,$F9,$1B,$FD,$4F,$0E,$73,$FC,$96,$35,$1F,$C8,$7E,$8D,$A7,$F2
	.dc.b	$06,$C7,$0A,$7D,$1B,$E0,$75,$7E,$2F,$AC,$97,$BC,$5F,$5B,$8C,$7F
	.dc.b	$CC,$7E,$73,$E9,$BF,$96,$B6,$FB,$3F,$EF,$4D,$7F,$9D,$B9,$AF,$F3
	.dc.b	$5F,$D2,$3E,$A7,$F9,$AF,$C4,$7F,$B8,$70,$F3,$4E,$5F,$44,$66,$BF
	.dc.b	$CE,$6B,$E2,$F8,$9C,$BE,$72,$D7,$08,$FE,$97,$5F,$D1,$6B,$84,$FF
	.dc.b	$47,$8C,$4E,$9F,$EA,$13,$A7,$41,$7C,$5E,$E3,$0D,$8C,$D7,$09,$CB
	.dc.b	$63,$31,$EA,$ED,$F8,$35,$F0,$9F,$EA,$F8,$FF,$BA,$AF,$E0,$F2,$E3
	.dc.b	$E6,$7E,$3B,$23,$A3,$E1,$23,$DF,$C3,$F2,$11,$C1,$E3,$F3,$FE,$89
	.dc.b	$F4,$A2,$C7,$D0,$8F,$61,$1E,$07,$B7,$E9,$7E,$23,$74,$F8,$91,$8D
	.dc.b	$3E,$4B,$5F,$A0,$E7,$87,$FB,$FB,$FC,$45,$B3,$FD,$E2,$31,$5F,$DF
	.dc.b	$73,$FF,$6C,$3F,$9B,$87,$C1,$7C,$8F,$94,$E0,$C9,$F3,$3E,$42,$3E
	.dc.b	$9F,$8E,$F3,$91,$C1,$ED,$F3,$FC,$47,$8E,$2C,$78,$62,$E7,$FE,$C6
	.dc.b	$C1,$58,$CF,$FC,$37,$F5,$9F,$ED,$EC,$3F,$08,$DB,$FC,$6B,$5F,$CA
	.dc.b	$35,$FF,$0E,$3B,$9F,$8F,$F4,$5F,$E1,$C8,$C5,$7F,$88,$70,$4F,$F8
	.dc.b	$83,$E9,$BF,$C5,$53,$FC,$05,$B8,$7F,$B7,$2F,$F5,$98,$7F,$8C,$27
	.dc.b	$2D,$9F,$CF,$47,$07,$F5,$0D,$7F,$C6,$0D,$7F,$C7,$13,$FA,$99,$E9
	.dc.b	$E8,$79,$F9,$BC,$8C,$C3,$F4,$D8,$7F,$58,$C3,$AC,$F8,$09,$3E,$9F
	.dc.b	$82,$E1,$3F,$EF,$4A,$47,$C3,$F9,$2F,$81,$EE,$DB,$F8,$0D,$BE,$0D
	.dc.b	$CB,$E0,$3F,$1F,$C9,$95,$FC,$9F,$4F,$EA,$4F,$C9,$8C,$DF,$85,$FA
	.dc.b	$5D,$38,$C3,$E5,$74,$FD,$13,$EB,$7C,$5D,$C3,$E9,$45,$C3,$E9,$4D
	.dc.b	$AE,$35,$A3,$C5,$19,$7E,$7D,$5A,$F9,$F3,$56,$BF,$C2,$5F,$8D,$3E
	.dc.b	$2A,$E9,$18,$6D,$FE,$5A,$F9,$5F,$18,$F1,$D7,$9A,$B6,$78,$8A,$F8
	.dc.b	$9A,$F8,$AF,$C9,$93,$0E,$B1,$86,$D7,$09,$D3,$CB,$0F,$13,$DE,$FC
	.dc.b	$47,$A4,$98,$79,$8F,$54,$71,$C3,$B3,$4E,$D8,$78,$2B,$8C,$7E,$BA
	.dc.b	$31,$5C,$E3,$07,$19,$D9,$EF,$29,$EF,$7A,$9B,$8E,$51,$B6,$BB,$38
	.dc.b	$E1,$F0,$1E,$78,$7C,$07,$FD,$44,$D5,$1B,$9E,$73,$93,$51,$9A,$F9
	.dc.b	$13,$AC,$3F,$05,$7D,$9F,$DE,$92,$7F,$66,$D8,$FE,$C3,$7F,$8B,$7F
	.dc.b	$80,$C6,$5B,$3F,$C3,$7E,$9B,$F1,$24,$7E,$D3,$63,$DA,$DB,$F4,$B6
	.dc.b	$7C,$E7,$F0,$C7,$13,$F0,$AB,$F8,$E4,$62,$3E,$55,$7F,$19,$8C,$5F
	.dc.b	$C3,$6D,$70,$BF,$19,$2F,$C6,$DB,$F1,$B4,$F1,$C7,$D2,$AE,$71,$EF
	.dc.b	$35,$1E,$57,$0F,$42,$4F,$46,$C7,$47,$F6,$F3,$D1,$FE,$38,$D9,$F6
	.dc.b	$4F,$69,$F7,$BF,$14,$C5,$FB,$DB,$7E,$F9,$EE,$79,$4E,$B1,$A3,$CF
	.dc.b	$5F,$D5,$BA,$FF,$DC,$FF,$47,$FE,$48,$AF,$E8,$D1,$8F,$16,$5F,$F9
	.dc.b	$2E,$B5,$7C,$63,$13,$FE,$F0,$F4,$5F,$4D,$FE,$27,$6C,$FF,$89,$5B
	.dc.b	$8F,$F8,$B9,$AF,$F8,$D7,$E4,$DF,$5B,$FE,$FA,$6D,$7F,$96,$EB,$8E
	.dc.b	$28,$5C,$AD,$BF,$D7,$FC,$0F,$ED,$06,$D5,$55,$F4,$3F,$E6,$0B,$FE
	.dc.b	$85,$5C,$1F,$EF,$2F,$F9,$07,$87,$FD,$89,$18,$3F,$C8,$1F,$4F,$FD
	.dc.b	$22,$FF,$C9,$2D,$7F,$CA,$0D,$7C,$E1,$C3,$EA,$F6,$9F,$05,$FE,$4F
	.dc.b	$6D,$7F,$93,$5B,$3F,$E5,$66,$BF,$E5,$DF,$9C,$FB,$4D,$E0,$FF,$2C
	.dc.b	$46,$23,$C2,$D7,$E4,$FC,$F3,$9B,$FC,$BC,$DC,$3F,$CB,$8D,$9F,$F3
	.dc.b	$2B,$5F,$B1,$E6,$7F,$99,$78,$3F,$D0,$11,$8A,$FF,$32,$46,$23,$E0
	.dc.b	$FC,$5B,$EB,$BF,$CD,$8D,$AF,$F3,$53,$67,$FC,$E6,$D7,$FC,$F5,$F5
	.dc.b	$DE,$94,$D2,$69,$FE,$77,$FA,$87,$9D,$F3,$2B,$FC,$FB,$18,$E0,$FF
	.dc.b	$3C,$2F,$9F,$7A,$38,$BE,$17,$FE,$89,$FA,$6A,$DC,$3F,$D0,$0A,$32
	.dc.b	$FD,$E6,$BF,$96,$5F,$4F,$C1,$B1,$FE,$89,$6B,$FE,$92,$6B,$FE,$99
	.dc.b	$EA,$79,$DE,$06,$C7,$EB,$63,$07,$93,$E9,$BF,$D3,$10,$69,$38,$3D
	.dc.b	$3D,$6F,$5E,$92,$BF,$D3,$91,$97,$FD,$3D,$F4,$BF,$01,$34,$8F,$C4
	.dc.b	$FA,$67,$89,$49,$1F,$EA,$78,$CB,$FE,$A9,$FA,$EF,$8E,$9A,$4D,$FE
	.dc.b	$7F,$D3,$FB,$09,$9E,$0F,$A7,$D7,$FC,$4A,$FF,$33,$E2,$60,$8F,$E7
	.dc.b	$37,$D9,$1C,$1F,$92,$F6,$BE,$EF,$57,$FB,$92,$30,$43,$FD,$8F,$FA
	.dc.b	$0B,$C0,$FC,$1B,$63,$FD,$88,$D7,$71,$EA,$F2,$A4,$7F,$B2,$74,$FF
	.dc.b	$14,$36,$3F,$1B,$F5,$BF,$1A,$3F,$8C,$EB,$FC,$2B,$C3,$7C,$CF,$F0
	.dc.b	$CF,$23,$F8,$E4,$62,$BF,$92,$C7,$F1,$FB,$E4,$FF,$21,$FB,$87,$CA
	.dc.b	$F1,$C7,$F8,$AC,$7F,$BB,$EF,$8C,$63,$D5,$D4,$7F,$30,$9E,$2E,$8C
	.dc.b	$7B,$BA,$50,$E0,$C3,$63,$81,$A9,$E1,$F7,$EE,$EE,$5A,$BE,$3A,$9A
	.dc.b	$8E,$DF,$72,$F7,$6F,$BB,$32,$EB,$39,$7B,$FE,$20,$F8,$EF,$32,$F9
	.dc.b	$CA,$62,$3D,$5F,$80,$6E,$BC,$FA,$73,$7E,$99,$FC,$E7,$2D,$D7,$C1
	.dc.b	$F9,$F7,$D7,$5F,$AF,$97,$F0,$27,$2D,$F6,$F8,$7C,$74,$8C,$1F,$1B
	.dc.b	$F0,$07,$D9,$7C,$4F,$96,$93,$98,$FA,$1F,$78,$F9,$8B,$F1,$72,$BE
	.dc.b	$7C,$E5,$BE,$87,$0F,$1E,$46,$0F,$AB,$F8,$07,$E5,$1F,$78,$E3,$FE
	.dc.b	$74,$FC,$4F,$6F,$F0,$13,$BF,$0F,$FB,$77,$DE,$38,$3F,$08,$BE,$C6
	.dc.b	$98,$8F,$C4,$78,$6F,$AD,$BF,$5D,$33,$FC,$3C,$E5,$FC,$4F,$DE,$BE
	.dc.b	$2F,$F2,$1F,$05,$FD,$BA,$7C,$ED,$D3,$FC,$01,$5B,$9E,$7C,$1E,$0D
	.dc.b	$E4,$60,$FC,$AF,$DE,$BE,$AB,$F3,$8B,$E3,$A9,$88,$FD,$07,$DA,$7E
	.dc.b	$2E,$BF,$3B,$9F,$E2,$5F,$1F,$81,$AF,$0F,$F4,$DF,$69,$F0,$E7,$F4
	.dc.b	$6F,$C9,$F3,$BF,$C0,$ED,$8F,$F0,$FB,$B7,$E9,$AF,$03,$9F,$2C,$9E
	.dc.b	$71,$DB,$CE,$F8,$BF,$27,$27,$BC,$78,$BC,$ED,$46,$0F,$1B,$FB,$BF
	.dc.b	$B3,$FE,$19,$1F,$82,$FB,$4F,$F0,$43,$FB,$A6,$BF,$86,$8D,$F3,$7F
	.dc.b	$8B,$D7,$83,$CF,$D4,$95,$FC,$0A,$72,$FF,$04,$FB,$1F,$94,$FE,$43
	.dc.b	$EA,$59,$7F,$0D,$6C,$6F,$87,$F8,$85,$7C,$FC,$BF,$C7,$CF,$26,$C6
	.dc.b	$B8,$BF,$8B,$7D,$A3,$BB,$A7,$B1,$31,$FE,$2B,$39,$6F,$81,$C2,$A4
	.dc.b	$60,$DF,$97,$D2,$64,$60,$E7,$BF,$A7,$92,$46,$0E,$FC,$3E,$49,$91
	.dc.b	$83,$C7,$E4,$7F,$96,$9F,$CF,$0C,$BF,$80,$6D,$88,$C4,$ED,$FD,$24
	.dc.b	$E9,$E2,$FF,$2D,$FB,$4E,$0D,$8C,$FA,$A9,$CB,$63,$FC,$43,$F4,$9F
	.dc.b	$E0,$32,$38,$3C,$5F,$17,$FE,$71,$C7,$FC,$E1,$1F,$E2,$EE,$CF,$E8
	.dc.b	$2F,$F9,$EA,$BF,$CF,$7F,$68,$E5,$FD,$D6,$1C,$6F,$F7,$9C,$B7,$0F
	.dc.b	$F3,$5B,$5F,$F4,$1C,$D7,$F9,$E4,$F6,$E5,$C1,$BE,$67,$DA,$62,$7C
	.dc.b	$7D,$9C,$4C,$5F,$F2,$00,$F5,$F9,$C9,$3C,$0F,$5B,$E3,$D9,$CF,$4F
	.dc.b	$07,$95,$6B,$FB,$0C,$FF,$7A,$6F,$2E,$71,$FE,$5F,$39,$7F,$98,$FD
	.dc.b	$A7,$C9,$6E,$DE,$BE,$FF,$85,$39,$AF,$F4,$F3,$F0,$AB,$8F,$06,$AF
	.dc.b	$6F,$0F,$12,$65,$FD,$D2,$FF,$8F,$36,$33,$7F,$90,$9F,$E4,$93,$F8
	.dc.b	$F6,$C6,$67,$EF,$BF,$EB,$68,$7D,$46,$21,$18,$AE,$53,$98,$C4,$F7
	.dc.b	$F2,$F0,$4A,$F0,$3B,$FA,$A9,$1E,$36,$CF,$E9,$BE,$CF,$FA,$84,$FE
	.dc.b	$95,$B1,$9D,$FC,$91,$88,$FE,$95,$F1,$3F,$86,$AF,$D6,$36,$3F,$62
	.dc.b	$D7,$7C,$9E,$D8,$FF,$63,$C6,$BE,$6F,$04,$AF,$DB,$7D,$9B,$EC,$BE
	.dc.b	$23,$63,$31,$83,$E4,$7D,$93,$A6,$92,$7F,$80,$46,$5F,$E0,$5F,$65
	.dc.b	$F9,$86,$F5,$7D,$18,$C6,$5F,$C2,$9A,$FF,$69,$D7,$FB,$F5,$7C,$7F
	.dc.b	$5F,$E2,$91,$FE,$1D,$F6,$3F,$CD,$1B,$E5,$7D,$D8,$C5,$FF,$16,$6B
	.dc.b	$FD,$D6,$7F,$C7,$87,$F1,$72,$BF,$C7,$1A,$EF,$C2,$FC,$47,$1F,$F9
	.dc.b	$16,$7F,$8F,$36,$7F,$91,$B5,$FF,$01,$35,$FF,$78,$FD,$8F,$F9,$06
	.dc.b	$7F,$93,$36,$3F,$95,$B5,$FF,$05,$35,$E3,$3F,$99,$83,$49,$3F,$8D
	.dc.b	$F9,$CF,$C5,$91,$C1,$FC,$D7,$C4,$FE,$F0,$DF,$27,$F4,$91,$8C,$3F
	.dc.b	$9E,$B5,$FF,$0E,$7D,$6F,$F0,$CA,$FE,$7C,$D8,$FE,$86,$D7,$7E,$17
	.dc.b	$EB,$FE,$C3,$FD,$D0,$4C,$70,$7E,$CE,$BF,$D8,$57,$D3,$DD,$C1,$1F
	.dc.b	$E7,$7F,$41,$FB,$9F,$0B,$FD,$F9,$97,$F3,$08,$C1,$FB,$DF,$AC,$FF
	.dc.b	$78,$D7,$F5,$36,$CF,$F5,$86,$BF,$E3,$76,$A2,$2F,$77,$A1,$27,$A3
	.dc.b	$67,$FC,$AB,$3F,$E9,$C9,$FF,$46,$BC,$5B,$F1,$DF,$D1,$E1,$F5,$B7
	.dc.b	$EA,$70,$DA,$E3,$1E,$B2,$3E,$B7,$DD,$C3,$23,$DC,$DC,$3D,$DD,$BD
	.dc.b	$CF,$04,$59,$F2,$F7,$F8,$CF,$27,$45,$8C,$9F,$26,$3E,$61,$1C,$1E
	.dc.b	$EE,$1E,$18,$7C,$27,$FB,$27,$D6,$34,$BE,$AE,$1B,$7F,$EB,$29,$FF
	.dc.b	$11,$3F,$E0,$96,$DF,$F8,$39,$AF,$F9,$B5,$AF,$D7,$0F,$F8,$FA,$39
	.dc.b	$7A,$9F,$CF,$E3,$13,$CE,$31,$EA,$79,$8F,$3B,$6B,$C9,$F1,$5E,$A3
	.dc.b	$D7,$F1,$D6,$32,$79,$E3,$B1,$1C,$1E,$DF,$13,$86,$1F,$E9,$AC,$7F
	.dc.b	$93,$37,$EB,$3C,$C4,$AF,$84,$FA,$7E,$57,$E5,$9F,$31,$B1,$FE,$DA
	.dc.b	$FA,$BF,$CC,$37,$A7,$FC,$29,$18,$BF,$F5,$4B,$FD,$D6,$7E,$43,$FD
	.dc.b	$EE,$3F,$D3,$0D,$73,$3F,$D9,$E3,$11,$FA,$55,$EA,$76,$3B,$F8,$F6
	.dc.b	$FC,$91,$C4,$8F,$F9,$16,$78,$B6,$3C,$FE,$B7,$1B,$F3,$BE,$AF,$82
	.dc.b	$F5,$FA,$56,$FE,$03,$86,$9F,$0E,$31,$5F,$03,$C4,$F8,$67,$C4,$F4
	.dc.b	$2D,$7C,$98,$B3,$F1,$A6,$FA,$8A,$F8,$FF,$4F,$F3,$0F,$9B,$F3,$D6
	.dc.b	$3E,$94,$5B,$F9,$F3,$6B,$94,$7C,$99,$C5,$7D,$2A,$B3,$98,$E6,$7D
	.dc.b	$D8,$CC,$35,$56,$14,$D8,$C3,$6B,$9D,$79,$1B,$7E,$48,$E9,$5C,$27
	.dc.b	$27,$C0,$2C,$7E,$06,$BA,$3F,$8C,$9F,$87,$3A,$8E,$6D,$BF,$2B,$67
	.dc.b	$B4,$F1,$8C,$1E,$F2,$B9,$3D,$05,$BF,$45,$7C,$29,$ED,$7E,$49,$CD
	.dc.b	$F9,$2F,$85,$7A,$DA,$7A,$DE,$F3,$C1,$F1,$46,$8F,$59,$6F,$D6,$DA
	.dc.b	$F5,$D7,$18,$C5,$72,$8C,$11,$F1,$3E,$AB,$F4,$46,$0F,$C2,$1E,$43
	.dc.b	$13,$E7,$3A,$9F,$EB,$2B,$F8,$4D,$BF,$DC,$12,$7F,$60,$D8,$D4,$60
	.dc.b	$FE,$15,$F3,$7F,$8B,$23,$F6,$5B,$1F,$05,$B8,$7A,$E3,$F8,$24,$62
	.dc.b	$7C,$F1,$F9,$78,$C3,$F9,$D3,$F8,$9B,$87,$E6,$CF,$DC,$8C,$5F,$B2
	.dc.b	$31,$5E,$32,$57,$8D,$B1,$E3,$8F,$04,$60,$F1,$CA,$BA,$DF,$F2,$66
	.dc.b	$CF,$C2,$38,$57,$9E,$FC,$2F,$72,$3D,$EF,$A4,$F0,$9F,$5B,$FC,$BA
	.dc.b	$BA,$C7,$C0,$8D,$4F,$B5,$D1,$8B,$F2,$34,$F2,$4F,$81,$F5,$D7,$4B
	.dc.b	$F2,$9E,$37,$E6,$7D,$43,$E4,$7F,$DA,$CB,$E4,$AC,$61,$97,$FE,$21
	.dc.b	$BF,$E0,$95,$B8,$C1,$FD,$BD,$E8,$E3,$0F,$F1,$44,$6D,$4F,$2F,$17
	.dc.b	$FD,$61,$3F,$E2,$59,$FE,$F0,$70,$8C,$9A,$F5,$BC,$E7,$A1,$B7,$DE
	.dc.b	$54,$F8,$2F,$69,$EE,$F1,$AD,$6E,$74,$F8,$2B,$CF,$5E,$42,$46,$61
	.dc.b	$D3,$E7,$7C,$ED,$7F,$C2,$0D,$7F,$C9,$0D,$5F,$0F,$FE,$DA,$7C,$0D
	.dc.b	$7E,$0B,$5F,$F0,$D3,$5F,$97,$E9,$FC,$82,$57,$97,$0F,$F5,$2C,$FA
	.dc.b	$FD,$D7,$67,$FB,$75,$B5,$EB,$6C,$EE,$7F,$DC,$58,$3E,$42,$92,$FE
	.dc.b	$14,$65,$F8,$9F,$8A,$BD,$BF,$F7,$53,$63,$31,$83,$FD,$DD,$F5,$0F
	.dc.b	$A7,$E1,$97,$D1,$8C,$4F,$CF,$5F,$50,$F0,$B1,$7C,$2F,$F0,$FF,$70
	.dc.b	$EF,$FA,$D0,$8C,$3F,$65,$AF,$DF,$6B,$F9,$0F,$A8,$7D,$4F,$26,$C7
	.dc.b	$E0,$5A,$FE,$1D,$AF,$E6,$3E,$77,$FB,$7B,$F8,$76,$BF,$8D,$6B,$BF
	.dc.b	$03,$CF,$F5,$3F,$DB,$E3,$F4,$2D,$7F,$CB,$4D,$7F,$D2,$4D,$5F,$53
	.dc.b	$F9,$47,$F2,$CD,$7F,$38,$D7,$F5,$AD,$7F,$CC,$DF,$53,$F2,$13,$48
	.dc.b	$FC,$CF,$A9,$F4,$BF,$E6,$D6,$BF,$EA,$56,$A6,$C7,$D9,$8F,$D5,$35
	.dc.b	$FD,$83,$5F,$E0,$2D,$7F,$CF,$5F,$4B,$EF,$27,$1F,$DE,$7E,$17,$D4
	.dc.b	$33,$FC,$0B,$6B,$F7,$0D,$7F,$7C,$D7,$F8,$7A,$FA,$8E,$24,$8E,$2F
	.dc.b	$7F,$02,$47,$F0,$3B,$46,$23,$F8,$57,$D4,$7C,$D7,$FD,$90,$D4,$D3
	.dc.b	$12,$F0,$7D,$84,$9E,$CF,$E7,$3B,$5E,$FD,$24,$FE,$ED,$FD,$0F,$D4
	.dc.b	$7D,$22,$71,$60,$FD,$2F,$D4,$7A,$2B,$F8,$E3,$5F,$E4,$4D,$7F,$99
	.dc.b	$B5,$FF,$56,$2F,$73,$99,$23,$9B,$FB,$2F,$A8,$F9,$D1,$FE,$B3,$6B
	.dc.b	$FE,$EB,$6A,$6F,$D4,$3D,$7F,$DD,$B6,$7F,$99,$35,$FE,$70,$D7,$FA
	.dc.b	$4F,$D4,$3E,$57,$F8,$1B,$63,$F9,$D3,$5F,$E8,$0D,$7F,$A8,$FD,$45
	.dc.b	$7F,$D9,$4D,$41,$C3,$67,$F8,$8F,$C4,$FF,$6B,$8F,$E9,$0D,$7F,$A7
	.dc.b	$35,$FE,$BE,$D7,$FD,$B0,$BE,$0F,$99,$23,$9B,$FC,$7B,$EA,$3F,$9B
	.dc.b	$47,$F5,$56,$BF,$D6,$DA,$FF,$6C,$6B,$FE,$E5,$FA,$77,$C0,$FF,$28
	.dc.b	$6C,$7F,$5F,$6B,$FD,$95,$AF,$F7,$67,$F2,$6E,$F8,$B9,$11,$FE,$D1
	.dc.b	$F4,$FF,$DC,$DB,$E8,$FF,$36,$8C,$47,$F7,$16,$BF,$E0,$2F,$A7,$7D
	.dc.b	$0F,$F3,$C6,$C7,$F7,$36,$BF,$DE,$5A,$FF,$82,$BE,$2D,$F3,$FF,$D1
	.dc.b	$1B,$1F,$DE,$DA,$FF,$80,$5A,$FF,$84,$BE,$9D,$EF,$7F,$4C,$6C,$7F
	.dc.b	$80,$DA,$FF,$82,$5A,$F8,$7E,$9F,$C0,$47,$C0,$FF,$56,$FA,$67,$C1
	.dc.b	$52,$4F,$F8,$3E,$32,$FF,$84,$7E,$99,$F3,$FF,$D7,$9B,$19,$8C,$1F
	.dc.b	$D8,$BE,$99,$F3,$D4,$93,$FE,$1B,$8C,$B7,$9B,$81,$1C,$1B,$F0,$BA
	.dc.b	$11,$C1,$CF,$BF,$C0,$47,$07,$F8,$A7,$E9,$5F,$45,$FA,$A8,$FF,$21
	.dc.b	$F9,$97,$E9,$BF,$DB,$31,$FE,$3B,$6B,$A3,$F6,$35,$C9,$FF,$20,$2F
	.dc.b	$90,$E9,$F0,$9F,$ED,$98,$D4,$F1,$0E,$23,$53,$C2,$FC,$9F,$17,$F3
	.dc.b	$2F,$87,$C7,$FF,$1F,$7F,$C9,$CD,$7A,$46,$EF,$FD,$D0,$FF,$0D,$6F
	.dc.b	$07,$59,$FE,$13,$3B,$BE,$FE,$16,$65,$F4,$CF,$93,$C6,$5F,$A6,$CB
	.dc.b	$FD,$E8,$DC,$3E,$55,$6E,$FF,$CC,$4F,$E7,$2B,$FD,$F4,$E6,$3E,$13
	.dc.b	$63,$3D,$7E,$18,$78,$96,$25,$74,$8D,$70,$79,$38,$E4,$F8,$23,$96
	.dc.b	$FA,$8C,$1E,$47,$B7,$87,$FE,$82,$8F,$F3,$86,$3F,$C4,$B2,$FF,$37
	.dc.b	$36,$33,$3B,$7F,$D0,$B0,$F9,$D7,$D3,$70,$6C,$F4,$6B,$C1,$AF,$F9
	.dc.b	$FF,$E7,$7F,$C4,$04,$7F,$C9,$2F,$F9,$2B,$E0,$1B,$FF,$A3,$67,$FC
	.dc.b	$A4,$D5,$3F,$D1,$3E,$99,$C9,$FE,$9A,$BE,$35,$F9,$13,$F7,$37,$CF
	.dc.b	$97,$3A,$FF,$90,$6A,$CF,$FA,$8A,$6D,$7E,$9F,$96,$BF,$D4,$9B,$C8
	.dc.b	$F8,$CC,$CF,$F6,$1B,$F4,$7C,$A6,$0C,$DD,$8F,$67,$CA,$70,$31,$5E
	.dc.b	$77,$BE,$7F,$84,$9F,$F5,$73,$67,$F4,$BD,$DA,$E7,$DF,$34,$BD,$C5
	.dc.b	$AF,$D9,$7D,$2F,$CA,$6E,$FE,$AE,$EF,$6C,$E7,$2F,$93,$48,$F9,$2F
	.dc.b	$2E,$0D,$DF,$07,$8F,$85,$32,$FF,$62,$5F,$FA,$BF,$0E,$6F,$F3,$08
	.dc.b	$CC,$F8,$3E,$77,$86,$3E,$67,$C7,$F4,$DE,$4F,$1E,$DF,$47,$0F,$6F
	.dc.b	$97,$B7,$EE,$7D,$67,$B3,$EC,$2C,$F9,$1F,$B7,$F4,$9F,$CD,$A3,$FD
	.dc.b	$B5,$F6,$CF,$A9,$CF,$27,$E0,$23,$13,$FE,$DF,$F8,$FF,$C0,$37,$EC
	.dc.b	$3F,$0B,$18,$CF,$F9,$02,$8C,$FD,$0E,$BE,$86,$DF,$37,$F1,$BF,$54
	.dc.b	$FA,$3F,$C7,$FD,$B3,$ED,$F3,$18,$7F,$27,$F5,$8F,$0A,$C6,$1B,$8F
	.dc.b	$FB,$D7,$EB,$1E,$17,$E6,$97,$83,$63,$0F,$C4,$F9,$EF,$CF,$FD,$BF
	.dc.b	$F7,$AB,$FF,$7F,$73,$7F,$B7,$BB,$BF,$0B,$39,$E4,$FE,$7D,$F5,$BF
	.dc.b	$E5,$5B,$FF,$1B,$D7,$E2,$A1,$C7,$FA,$A7,$55,$C2,$3F,$5B,$5F,$C3
	.dc.b	$4E,$27,$F5,$71,$F6,$BA,$F2,$F6,$FE,$D3,$B3,$FD,$25,$3C,$2F,$FC
	.dc.b	$8F,$F5,$EF,$80,$D7,$C9,$7E,$EA,$31,$5B,$A7,$D7,$FE,$E5,$FE,$B0
	.dc.b	$FC,$88,$D3,$9D,$7F,$4A,$E6,$77,$8F,$F0,$68,$78,$16,$30,$FE,$43
	.dc.b	$AF,$F8,$5F,$DB,$FF,$87,$B1,$F9,$B1,$8F,$0B,$F8,$7F,$A1,$FE,$53
	.dc.b	$9F,$33,$F8,$56,$FA,$BA,$ED,$FE,$2B,$F5,$FF,$C1,$76,$FD,$CB,$FE
	.dc.b	$00,$AF,$73,$97,$FC,$04,$E7,$6F,$EE,$DF,$60,$CF,$F8,$FF,$DB,$BE
	.dc.b	$BF,$F9,$0C,$E7,$0F,$91,$39,$7A,$FC,$47,$F2,$6F,$3F,$FC,$EF,$18
	.dc.b	$3E,$73,$8C,$7F,$C0,$8B,$CC,$E1,$F4,$DC,$5B,$5F,$57,$EC,$1F,$47
	.dc.b	$FC,$C9,$7C,$45,$8C,$32,$78,$7C,$CD,$CB,$FC,$25,$7F,$EC,$89,$C9
	.dc.b	$23,$35,$D6,$BB,$7C,$BC,$AF,$17,$D8,$DB,$E7,$3F,$D1,$3E,$A1,$D7
	.dc.b	$9E,$5F,$E5,$31,$8C,$3D,$5C,$FE,$9E,$7F,$53,$63,$DD,$17,$1F,$3C
	.dc.b	$DB,$FE,$67,$AF,$EA,$AF,$E2,$78,$5F,$0F,$E1,$FE,$23,$23,$E4,$74
	.dc.b	$5C,$7E,$0B,$F2,$49,$1C,$8D,$9F,$A3,$FB,$1F,$F1,$13,$7E,$1B,$FB
	.dc.b	$0C,$31,$B1,$27,$F5,$5A,$FE,$81,$E0,$63,$BF,$8F,$C2,$F2,$F8,$5B
	.dc.b	$3D,$DC,$34,$F1,$46,$33,$FF,$25,$65,$FE,$D5,$9F,$37,$CB,$79,$3B
	.dc.b	$BC,$AC,$8F,$57,$A1,$6F,$8B,$EA,$3F,$AE,$BF,$E6,$3B,$FE,$55,$3F
	.dc.b	$E6,$58,$C4,$E9,$FF,$68,$CE,$9D,$05,$F5,$9F,$18,$F8,$FF,$57,$6F
	.dc.b	$E6,$47,$3A,$E0,$DA,$E3,$3D,$27,$2D,$FB,$1F,$EF,$11,$E5,$6B,$FE
	.dc.b	$7B,$6B,$FE,$E6,$6A,$D9,$F6,$3F,$97,$3F,$C3,$73,$FD,$FE,$BF,$D0
	.dc.b	$93,$A7,$F9,$24,$7F,$46,$7C,$5E,$97,$A6,$31,$7F,$09,$AF,$96,$3A
	.dc.b	$FC,$3F,$66,$D7,$48,$B3,$FE,$85,$AE,$7C,$FF,$E1,$07,$FC,$77,$5E
	.dc.b	$2F,$80,$F3,$9E,$8F,$AA,$B1,$93,$E3,$47,$94,$8F,$C9,$EC,$FF,$1C
	.dc.b	$FD,$B7,$F9,$32,$3D,$B1,$87,$E1,$B5,$C9,$FD,$16,$38,$63,$E7,$FB
	.dc.b	$2F,$92,$7C,$AD,$AD,$7C,$88,$B3,$ED,$9B,$9F,$B6,$6D,$7D,$99,$B5
	.dc.b	$F6,$66,$AC,$FE,$A3,$E1,$AE,$C4,$7B,$36,$FB,$34,$EC,$58,$EC,$D7
	.dc.b	$B5,$F8,$29,$76,$30,$DA,$E1,$3A,$6C,$E9,$A6,$A7,$85,$7A,$49,$7E
	.dc.b	$67,$D2,$62,$7F,$39,$3C,$27,$44,$8C,$9F,$96,$AF,$05,$71,$3F,$44
	.dc.b	$E2,$78,$46,$27,$C8,$79,$9E,$07,$CF,$AF,$04,$76,$8C,$DF,$BD,$F3
	.dc.b	$D7,$47,$84,$F3,$8C,$4E,$DB,$7E,$39,$E9,$1F,$9F,$8C,$57,$D4,$3A
	.dc.b	$CE,$8F,$13,$FC,$06,$FF,$64,$48,$FE,$0C,$D8,$FD,$A3,$5E,$11,$FE
	.dc.b	$1F,$FB,$CF,$ED,$2F,$E0,$5B,$1F,$09,$B7,$ED,$6C,$F8,$E4,$23,$2F
	.dc.b	$A6,$7F,$D1,$31,$88,$F8,$A2,$33,$E1,$FE,$7A,$75,$8F,$C7,$BE,$B7
	.dc.b	$D6,$B4,$48,$F4,$36,$3D,$0D,$BF,$43,$67,$D1,$3C,$DE,$24,$78,$B6
	.dc.b	$78,$B4,$E3,$7E,$6C,$3F,$D8,$6D,$8F,$99,$7D,$EF,$E1,$47,$AA,$75
	.dc.b	$1F,$06,$31,$3B,$8F,$8B,$7F,$12,$7C,$11,$C8,$F9,$F5,$CA,$75,$1E
	.dc.b	$82,$3E,$86,$CF,$A2,$BB,$61,$F3,$67,$9C,$63,$0F,$5B,$6B,$D6,$78
	.dc.b	$A7,$CA,$FA,$0F,$EF,$FB,$7F,$88,$4F,$F7,$2F,$D8,$BE,$CB,$13,$A2
	.dc.b	$57,$C1,$3F,$13,$3B,$8C,$E5,$FE,$D0,$8F,$F5,$54,$FF,$73,$6B,$F8
	.dc.b	$46,$BC,$BE,$37,$FB,$64,$7F,$AC,$67,$FD,$B3,$CB,$F9,$18,$C5,$FE
	.dc.b	$32,$31,$F0,$BC,$09,$1F,$DC,$67,$FD,$89,$87,$F7,$EA,$FF,$24,$37
	.dc.b	$C7,$FE,$F5,$18,$8F,$F0,$43,$5F,$ED,$7F,$75,$FD,$FF,$9C,$D2,$61
	.dc.b	$DF,$DD,$F4,$91,$C1,$FE,$03,$9F,$F7,$45,$7F,$84,$1A,$FF,$85,$DA
	.dc.b	$FF,$7A,$E6,$F0,$79,$92,$B9,$36,$7F,$C3,$3F,$B1,$7F,$9C,$DF,$F1
	.dc.b	$03,$5F,$D8,$B5,$FD,$FB,$57,$D9,$FF,$9D,$3E,$1F,$FC,$30,$D8,$FF
	.dc.b	$12,$AF,$B3,$7A,$3A,$F4,$17,$FE,$1F,$FB,$3F,$8B,$85,$DF,$FC,$44
	.dc.b	$A3,$2F,$F8,$C1,$AF,$A2,$F7,$97,$68,$FF,$15,$35,$F7,$C6,$BE,$3A
	.dc.b	$46,$9A,$FC,$4F,$92,$E6,$47,$9B,$E4,$FA,$F7,$C7,$D2,$4F,$F8,$E6
	.dc.b	$32,$F8,$BE,$1F,$CA,$47,$07,$9B,$C5,$F4,$91,$C1,$CF,$C2,$F0,$11
	.dc.b	$C1,$EE,$EE,$F8,$24,$70,$7F,$92,$FE,$AD,$F6,$BF,$31,$B1,$98,$C1
	.dc.b	$F2,$BE,$E6,$47,$CA,$6B,$FE,$60,$9F,$F5,$8B,$D7,$C2,$E8,$47,$E8
	.dc.b	$F1,$7D,$42,$38,$3C,$9E,$CF,$8C,$8E,$0F,$47,$B3,$FD,$0B,$C2,$75
	.dc.b	$FF,$44,$C5,$39,$8C,$36,$BF,$AA,$CF,$FC,$01,$7F,$DF,$DF,$28,$F7
	.dc.b	$35,$08,$C3,$A8,$CF,$3B,$CC,$DF,$D6,$BE,$B6,$92,$FF,$C8,$DA,$F4
	.dc.b	$FA,$D7,$8D,$49,$3F,$DA,$37,$F5,$BF,$DA,$A5,$E4,$F1,$F7,$24,$EF
	.dc.b	$87,$FB,$1D,$27,$B7,$B9,$EA,$FA,$D7,$C4,$D2,$61,$FE,$59,$F8,$AF
	.dc.b	$7F,$D6,$BD,$9A,$49,$EC,$7F,$7C,$87,$B3,$F1,$BE,$B5,$E2,$52,$57
	.dc.b	$01,$0F,$0B,$E6,$FD,$6F,$09,$D3,$5F,$E8,$41,$C7,$17,$C4,$87,$B0
	.dc.b	$FA,$AF,$AF,$F3,$BF,$D3,$67,$C0,$D7,$9B,$6B,$93,$FC,$46,$37,$F0
	.dc.b	$BC,$88,$F2,$F8,$6F,$C1,$FD,$6B,$D3,$A4,$9F,$F4,$14,$FE,$23,$EB
	.dc.b	$5F,$09,$49,$3F,$0E,$1F,$27,$BF,$0E,$4E,$5F,$C8,$FE,$23,$FE,$CF
	.dc.b	$D7,$E0,$7D,$77,$E5,$EB,$E2,$B6,$33,$D1,$F9,$8E,$1F,$CC,$57,$D1
	.dc.b	$FA,$EF,$8C,$DE,$2F,$CE,$C6,$23,$E7,$AE,$1F,$D0,$F9,$6F,$A7,$A4
	.dc.b	$AF,$B9,$0E,$4F,$D3,$FD,$73,$E5,$29,$2B,$FD,$4D,$E0,$74,$FA,$E9
	.dc.b	$1C,$9E,$FC,$5F,$B0,$FA,$E7,$BF,$49,$5F,$8B,$87,$C9,$7E,$DB,$EB
	.dc.b	$9E,$DD,$25,$7E,$47,$EB,$BE,$37,$95,$FE,$48,$5F,$0F,$FB,$E6,$CE
	.dc.b	$E3,$1C,$B9,$E4,$93,$CD,$E1,$DD,$C7,$D3,$93,$DD,$E9,$E5,$C8,$F1
	.dc.b	$35,$FC,$EE,$2E,$2F,$E1,$FF,$5A,$FA,$AF,$E2,$3F,$5D,$FE,$F4,$BF
	.dc.b	$D5,$35,$FD,$2C,$ED,$FD,$72,$F9,$3E,$75,$27,$31,$FC,$73,$DD,$F8
	.dc.b	$F1,$FB,$5F,$33,$F8,$43,$73,$D7,$16,$FC,$CB,$3F,$BB,$BE,$2F,$ED
	.dc.b	$9F,$DA,$F1,$FF,$9D,$3E,$B9,$F2,$DF,$CA,$B2,$FA,$37,$FE,$84,$6A
	.dc.b	$9D,$BF,$C1,$DF,$C8,$36,$7F,$99,$C6,$A3,$F8,$2B,$A6,$F7,$6B,$C0
	.dc.b	$91,$CE,$38,$ED,$E3,$E5,$93,$DE,$3A,$EF,$A8,$C1,$E3,$7C,$3E,$4F
	.dc.b	$FA,$56,$3F,$D2,$27,$FA,$7B,$1F,$DF,$F8,$B8,$8C,$4E,$DD,$CE,$9D
	.dc.b	$3F,$C0,$BE,$BB,$84,$7C,$3F,$25,$18,$9D,$4F,$09,$CB,$FE,$22,$F9
	.dc.b	$DF,$F1,$1A,$69,$1E,$BF,$0D,$FD,$33,$1F,$E5,$51,$FE,$30,$E1,$FE
	.dc.b	$6C,$FF,$AE,$2B,$FD,$73,$F5,$CE,$3F,$7E,$1C,$6F,$E0,$F2,$DC,$3F
	.dc.b	$D6,$6D,$7F,$D8,$33,$5F,$DA,$CF,$6E,$5C,$1B,$E5,$FF,$38,$CF,$FA
	.dc.b	$F2,$38,$7A,$32,$B8,$1E,$67,$A7,$C2,$C8,$C1,$EA,$7C,$5A,$FF,$15
	.dc.b	$8F,$EA,$4D,$BF,$E9,$B8,$7F,$9E,$A3,$FA,$61,$92,$F1,$F2,$39,$F2
	.dc.b	$DB,$FE,$B9,$F5,$DF,$27,$0F,$EA,$ED,$BF,$EA,$D8,$7F,$82,$6B,$7A
	.dc.b	$7C,$5A,$F8,$77,$C3,$5C,$4E,$5D,$78,$B2,$FB,$FB,$F6,$FB,$CF,$3F
	.dc.b	$17,$19,$FF,$65,$CB,$F9,$2E,$1E,$6F,$88,$F0,$79,$5E,$16,$C7,$AA
	.dc.b	$3C,$9F,$2D,$8C,$BD,$6E,$27,$DF,$F3,$56,$7D,$EF,$DF,$FA,$D7,$9B
	.dc.b	$4C,$79,$B6,$FF,$C0,$58,$FF,$89,$DB,$BF,$E1,$E3,$FD,$33,$AF,$F7
	.dc.b	$D7,$59,$7F,$80,$9F,$F0,$CD,$72,$F8,$2E,$1E,$EF,$17,$F2,$1F,$56
	.dc.b	$FA,$4B,$F3,$98,$6D,$FF,$84,$FE,$F9,$D7,$F9,$77,$F3,$D1,$98,$C7
	.dc.b	$76,$79,$24,$F1,$D2,$EB,$FE,$16,$FA,$8F,$C4,$4A,$24,$7F,$3B,$8F
	.dc.b	$F1,$13,$A8,$CF,$6F,$E9,$E1,$A5,$F9,$8C,$36,$FF,$C5,$5C,$FF,$E2
	.dc.b	$E6,$F8,$1F,$AF,$8F,$F1,$64,$60,$FF,$18,$6F,$F9,$D7,$C9,$97,$C5
	.dc.b	$F4,$FF,$6D,$3F,$E2,$99,$DC,$7E,$E7,$EF,$9F,$43,$4C,$79,$B6,$BF
	.dc.b	$C7,$92,$FA,$8E,$1F,$45,$C5,$B3,$FE,$41,$FA,$76,$9F,$C1,$57,$7D
	.dc.b	$8C,$3F,$C2,$7E,$99,$F3,$57,$DE,$C3,$70,$FF,$26,$4F,$B7,$83,$E8
	.dc.b	$35,$3E,$D7,$FC,$96,$E7,$E9,$BF,$CB,$D3,$FE,$55,$6C,$FF,$94,$1F
	.dc.b	$F2,$EC,$67,$E2,$FF,$8E,$7C,$8F,$EE,$3C,$1F,$F2,$93,$67,$E6,$BF
	.dc.b	$56,$FF,$CB,$2E,$5F,$C2,$3F,$9E,$8D,$F8,$FD,$3C,$8E,$AF,$F2,$9E
	.dc.b	$47,$C8,$D3,$F9,$67,$DE,$FE,$8B,$1F,$F3,$2C,$EB,$0F,$F3,$73,$63
	.dc.b	$3F,$2B,$C6,$72,$F5,$3F,$A0,$C7,$E5,$E7,$FD,$4F,$3F,$D4,$9B,$1F
	.dc.b	$D6,$DA,$FF,$60,$7F,$C8,$D1,$BF,$63,$9C,$FE,$21,$B3,$A8,$E5,$F0
	.dc.b	$1E,$D3,$DD,$36,$BD,$3E,$F7,$C1,$3E,$17,$B9,$63,$27,$96,$3A,$91
	.dc.b	$FE,$D3,$F4,$DF,$98,$6E,$3E,$48,$C6,$3F,$DD,$5A,$FF,$7A,$73,$C1
	.dc.b	$FE,$8D,$7F,$C3,$35,$FE,$99,$72,$FE,$A9,$CC,$FF,$A5,$9A,$F3,$EC
	.dc.b	$F0,$12,$7C,$BE,$E7,$8C,$8E,$0F,$4F,$B9,$FD,$A6,$FF,$D1,$31,$E1
	.dc.b	$9C,$4E,$79,$24,$F9,$7B,$6E,$3E,$39,$F8,$04,$9F,$8A,$FD,$6F,$BD
	.dc.b	$7D,$25,$24,$FF,$84,$A3,$2F,$CC,$F1,$3E,$51,$1C,$1F,$43,$C5,$F4
	.dc.b	$1E,$98,$B1,$FE,$C1,$FA,$6F,$E4,$8F,$F8,$79,$B8,$FF,$89,$DA,$FF
	.dc.b	$8B,$1A,$FF,$18,$77,$BF,$FB,$05,$FE,$DD,$39,$9F,$F0,$84,$61,$FE
	.dc.b	$64,$FF,$3E,$8F,$17,$99,$AF,$6B,$6C,$8F,$23,$F9,$A9,$E1,$5E,$58
	.dc.b	$F9,$1F,$63,$DB,$D3,$FF,$0F,$68,$8C,$61,$EF,$9E,$44,$7E,$03,$6F
	.dc.b	$FD,$BB,$87,$F0,$37,$CF,$1E,$AE,$DF,$5F,$9B,$EC,$F1,$6E,$1F,$2D
	.dc.b	$B7,$C9,$BA,$FC,$57,$E1,$FC,$F4,$AF,$A5,$E9,$DB,$F0,$CD,$BF,$99
	.dc.b	$36,$BF,$A3,$FD,$33,$E7,$FF,$C7,$8D,$8F,$F7,$E7,$C7,$3E,$A7,$35
	.dc.b	$27,$FE,$00,$56,$F2,$FF,$7D,$DE,$E7,$35,$B9,$FF,$7F,$D6,$F1,$FF
	.dc.b	$7F,$5E,$E7,$35,$B9,$5F,$9A,$3F,$CA,$15,$B9,$CD,$7F,$73,$BE,$97
	.dc.b	$B3,$C2,$7A,$B0,$E8,$E6,$7B,$B9,$36,$E0,$95,$E6,$9E,$31,$DF,$0A
	.dc.b	$B0,$F2,$5F,$5B,$F2,$CB,$0F,$2B,$C6,$7C,$0F,$67,$81,$1F,$F0,$75
	.dc.b	$74,$78,$1B,$2D,$79,$67,$D2,$F8,$2B,$BC,$F8,$6F,$E4,$E1,$E8,$9C
	.dc.b	$9A,$C3,$B5,$6F,$0E,$67,$8D,$F7,$57,$1B,$F8,$35,$C2,$F9,$C6,$27
	.dc.b	$57,$CA,$FD,$71,$87,$D4,$74,$8D,$3E,$13,$37,$D6,$B8,$61,$EF,$9E
	.dc.b	$38,$78,$2F,$E5,$CF,$C1,$AD,$C7,$F8,$92,$F9,$3D,$AB,$E0,$61,$AC
	.dc.b	$3E,$5C,$EE,$3C,$D1,$FE,$C2,$9F,$CF,$36,$FF,$B1,$C6,$23,$F1,$86
	.dc.b	$6B,$FD,$7F,$19,$8F,$B2,$64,$FD,$1F,$37,$D3,$24,$F9,$5B,$1E,$E6
	.dc.b	$E3,$ED,$9F,$D8,$C6,$1F,$AD,$5F,$CF,$2B,$84,$73,$23,$82,$57,$37
	.dc.b	$C7,$1A,$8C,$1F,$8B,$AE,$D3,$F1,$8F,$D8,$38,$7E,$C4,$6E,$79,$1E
	.dc.b	$49,$D5,$7E,$41,$B7,$FC,$41,$B3,$EC,$38,$DF,$73,$F5,$51,$A7,$CE
	.dc.b	$7A,$A3,$B1,$C1,$F1,$BF,$C7,$A3,$85,$F9,$67,$81,$D6,$FE,$44,$7C
	.dc.b	$0A,$E0,$73,$8F,$43,$6F,$D0,$F7,$FA,$67,$D0,$FF,$2A,$F9,$EF,$F7
	.dc.b	$34,$62,$7F,$93,$C6,$5A,$7F,$A2,$F0,$FE,$0F,$3B,$C7,$F9,$3D,$7F
	.dc.b	$34,$F1,$7F,$1E,$4B,$FC,$6B,$F9,$2F,$93,$FE,$72,$6E,$3F,$9A,$BF
	.dc.b	$EF,$7F,$9F,$72,$7F,$8E,$9B,$3F,$E8,$8C,$3F,$D1,$31,$87,$01,$C3
	.dc.b	$BC,$7F,$D7,$2D,$E5,$FF,$22,$C6,$23,$FD,$26,$D7,$F9,$B7,$D3,$FE
	.dc.b	$A0,$97,$FA,$67,$8F,$97,$23,$C3,$2A,$FF,$6F,$BE,$46,$BE,$9F,$90
	.dc.b	$FF,$56,$CF,$06,$C7,$A9,$AE,$FB,$BF,$CB,$7F,$76,$FA,$6A,$4B,$F0
	.dc.b	$C6,$5F,$1F,$D4,$7F,$A6,$26,$FA,$7C,$70,$F8,$F2,$FA,$87,$C4,$EA
	.dc.b	$FD,$7C,$0D,$BD,$B8,$7C,$D6,$C6,$63,$71,$AF,$BA,$7D,$DD,$24,$FC
	.dc.b	$68,$CB,$F2,$3E,$19,$DF,$D9,$B1,$98,$C1,$E2,$FB,$BF,$F3,$EC,$1A
	.dc.b	$4D,$FC,$FF,$76,$F2,$A9,$2B,$EB,$C6,$5F,$B3,$F7,$7E,$E4,$D2,$3F
	.dc.b	$0B,$EB,$3C,$AF,$E4,$1A,$FE,$55,$AF,$FB,$79,$AF,$FB,$CF,$EF,$3E
	.dc.b	$93,$F9,$86,$BF,$9D,$6B,$FE,$E6,$6B,$FE,$FB,$FA,$CF,$82,$4C,$7E
	.dc.b	$AB,$F1,$3E,$F1,$AF,$D8,$6D,$7E,$59,$AF,$EA,$1A,$FE,$B5,$7D,$E7
	.dc.b	$12,$47,$17,$BF,$83,$23,$F4,$36,$8C,$47,$ED,$3E,$F3,$E7,$37,$6F
	.dc.b	$C5,$C6,$23,$F7,$8D,$47,$BF,$C2,$F6,$91,$EE,$FE,$4B,$EA,$24,$7E
	.dc.b	$61,$AF,$F0,$4C,$1E,$E7,$C6,$24,$7A,$1F,$CD,$FC,$6C,$8F,$D0,$B5
	.dc.b	$FE,$1B,$CA,$F9,$7F,$2C,$3C,$7F,$F1,$4C,$E6,$3F,$93,$70,$EA,$31
	.dc.b	$1B,$8C,$7C,$77,$2A,$E3,$ED,$59,$E0,$DA,$F8,$76,$8C,$42,$1E,$D6
	.dc.b	$3B,$5E,$9D,$F4,$E5,$F9,$0F,$9B,$3A,$70,$73,$F6,$BA,$E0,$F6,$A9
	.dc.b	$88,$CF,$CB,$EB,$EE,$65,$F2,$B4,$F8,$FE,$26,$4F,$6C,$51,$E7,$EC
	.dc.b	$EF,$D3,$2B,$93,$E2,$FB,$CF,$1F,$CB,$49,$EC,$F9,$BE,$5B,$E0,$78
	.dc.b	$BF,$07,$DB,$95,$DE,$3E,$3E,$2F,$99,$BF,$27,$2B,$E0,$CE,$5B,$F0
	.dc.b	$5C,$3E,$0A,$46,$0F,$57,$DA,$7C,$F2,$3E,$DF,$17,$BF,$D9,$F8,$3D
	.dc.b	$99,$3D,$A3,$C7,$ED,$79,$3D,$09,$3E,$28,$FB,$BF,$B1,$FF,$9D,$3C
	.dc.b	$1F,$F3,$AB,$7C,$1F,$C8,$B6,$BF,$4E,$FF,$A9,$5F,$5F,$7F,$C0,$23
	.dc.b	$EE,$EF,$F8,$7D,$32,$3E,$0B,$F1,$B8,$7C,$3F,$01,$E2,$F5,$65,$7E
	.dc.b	$1B,$1F,$F7,$BB,$F8,$8F,$A7,$F9,$0F,$E7,$4F,$DC,$B8,$D3,$FD,$1A
	.dc.b	$DD,$3F,$53,$F4,$CF,$7F,$F3,$6B,$E5,$E9,$30,$FC,$AC,$3E,$9F,$7E
	.dc.b	$64,$9C,$BC,$BE,$3F,$9F,$C0,$49,$E2,$F6,$F8,$F9,$1D,$1A,$FA,$46
	.dc.b	$0F,$8F,$FD,$7F,$D3,$3C,$FF,$D8,$FD,$3B,$C4,$A6,$A3,$7E,$85,$C7
	.dc.b	$F5,$B7,$C5,$B1,$FB,$AF,$9D,$F8,$91,$FC,$83,$CC,$69,$FB,$E6,$C7
	.dc.b	$ED,$A7,$69,$78,$BC,$FC,$69,$3B,$8F,$E0,$FF,$4A,$F9,$2E,$9C,$D2
	.dc.b	$BF,$80,$CE,$5B,$F4,$FC,$2A,$46,$0D,$FA,$BD,$26,$46,$0E,$7B,$FA
	.dc.b	$79,$A4,$60,$EF,$BF,$92,$64,$60,$F1,$FA,$BF,$CB,$5F,$F1,$EE,$5F
	.dc.b	$C7,$BB,$3F,$C5,$ED,$8C,$CE,$DF,$F2,$34,$3E,$99,$F4,$FC,$1B,$3D
	.dc.b	$1A,$F0,$6B,$FE,$4C,$FA,$7F,$F0,$21,$1F,$F1,$1B,$FE,$24,$F8,$7F
	.dc.b	$F7,$EE,$3F,$EE,$A9,$FF,$15,$F7,$FF,$2C,$7F,$AD,$D7,$F5,$CF,$A7
	.dc.b	$FB,$11,$FD,$2A,$F8,$D7,$F9,$77,$92,$CF,$F4,$E6,$BF,$D9,$26,$BF
	.dc.b	$C6,$E7,$B7,$26,$0D,$F3,$3E,$D3,$13,$E3,$F4,$78,$98,$BF,$D0,$07
	.dc.b	$AF,$A3,$27,$81,$EB,$7C,$7A,$FF,$63,$8F,$E9,$2D,$AF,$D5,$61,$FB
	.dc.b	$B9,$CF,$37,$0E,$39,$18,$38,$BE,$EF,$A8,$F9,$37,$FB,$36,$E0,$AF
	.dc.b	$EE,$38,$3C,$1E,$1B,$71,$DE,$57,$F5,$CC,$38,$BF,$DE,$1F,$91,$8F
	.dc.b	$5F,$37,$C0,$78,$7D,$0B,$5C,$63,$C9,$E7,$E2,$BC,$8F,$9F,$DD,$F4
	.dc.b	$7A,$16,$FC,$0E,$1F,$17,$A1,$6F,$D8,$FB,$CF,$81,$E8,$59,$F4,$3F
	.dc.b	$80,$FA,$87,$AF,$4C,$7A,$F6,$BF,$BC,$E2,$D7,$F0,$ED,$8C,$EB,$89
	.dc.b	$CF,$A7,$AF,$5E,$D7,$27,$F1,$FF,$50,$FA,$6B,$F2,$38,$6D,$7F,$82
	.dc.b	$3E,$A1,$D9,$F9,$67,$F8,$A4,$66,$31,$E5,$E7,$92,$4F,$1D,$AF,$27
	.dc.b	$F3,$1F,$A8,$FC,$57,$67,$DF,$8F,$F2,$1C,$7F,$31,$AF,$F0,$BB,$A7
	.dc.b	$FC,$66,$E6,$39,$69,$FA,$88,$FF,$41,$4F,$08,$C7,$BD,$FA,$CE,$7F
	.dc.b	$E5,$33,$FE,$22,$6D,$FF,$43,$7F,$49,$19,$F0,$BF,$67,$BF,$E7,$6B
	.dc.b	$FA,$43,$72,$F4,$47,$C9,$F5,$77,$EB,$7E,$EB,$EA,$1F,$41,$4C,$7A
	.dc.b	$17,$3F,$F1,$9C,$BE,$A3,$87,$C2,$F1,$6C,$FF,$8E,$3E,$A1,$BF,$F0
	.dc.b	$67,$FC,$5B,$19,$8C,$78,$1F,$C2,$BE,$9D,$F2,$F7,$EA,$30,$DB,$FF
	.dc.b	$21,$CF,$AD,$FF,$2F,$B7,$83,$FC,$EA,$FF,$73,$8D,$EF,$FC,$5B,$EA
	.dc.b	$3F,$C9,$11,$FE,$4C,$6D,$7F,$6D,$7F,$C8,$B1,$9F,$8B,$FE,$3B,$F5
	.dc.b	$1F,$B4,$E4,$F9,$2D,$9F,$F3,$7B,$F5,$23,$3E,$4F,$0F,$13,$8F,$91
	.dc.b	$6F,$FC,$B7,$7F,$EF,$2B,$FA,$7C,$CF,$A8,$FE,$56,$DA,$FF,$30,$FD
	.dc.b	$47,$FA,$A1,$FF,$31,$C6,$75,$FF,$02,$2F,$91,$E2,$72,$BB,$5C,$23
	.dc.b	$F9,$BF,$CF,$7D,$48,$C1,$FE,$6E,$71,$7D,$3D,$CF,$1F,$C2,$F9,$1B
	.dc.b	$5E,$28,$B3,$B8,$C1,$E6,$F8,$4F,$59,$EC,$F8,$1B,$19,$3D,$7F,$47
	.dc.b	$8A,$F5,$BF,$D3,$3E,$A3,$F2,$73,$FE,$82,$7F,$C0,$CD,$D7,$5E,$04
	.dc.b	$8D,$3C,$63,$17,$FE,$29,$AE,$2F,$F8,$AF,$4F,$C9,$B9,$AF,$8B,$D7
	.dc.b	$E1,$F8,$1F,$17,$C0,$DC,$BC,$4E,$1A,$79,$63,$17,$DF,$A3,$D4,$7A
	.dc.b	$FC,$1B,$5E,$B8,$F7,$7B,$18,$AF,$33,$FD,$B3,$EA,$1F,$19,$F0,$FC
	.dc.b	$AC,$37,$0F,$89,$E5,$7C,$73,$18,$7B,$DC,$47,$FB,$12,$3F,$D8,$BC
	.dc.b	$1F,$C4,$1C,$07,$0E,$DA,$FF,$B3,$3E,$9F,$FD,$C3,$3F,$EB,$47,$FA
	.dc.b	$CC,$7F,$B9,$1C,$BF,$E5,$77,$29,$FE,$D2,$F8,$DA,$F5,$F6,$BF,$C2
	.dc.b	$CF,$36,$BC,$7D,$1B,$3F,$CF,$2B,$D3,$39,$6C,$66,$3D,$AD,$ED,$FE
	.dc.b	$42,$FF,$42,$8F,$35,$7F,$86,$A3,$13,$AF,$5B,$B9,$2F,$AB,$E2,$F0
	.dc.b	$78,$1C,$5B,$1F,$16,$2E,$1F,$0E,$6D,$78,$BC,$8F,$01,$E1,$F4,$2E
	.dc.b	$1F,$36,$3D,$E4,$7D,$AF,$C2,$F8,$8F,$47,$C8,$7A,$5B,$3F,$55,$B8
	.dc.b	$FD,$36,$E1,$FD,$0B,$E9,$BF,$1B,$5F,$EE,$E6,$D7,$FA,$3D,$7D,$37
	.dc.b	$F8,$F9,$CD,$71,$E2,$C1,$99,$B7,$CE,$2D,$F3,$9B,$5C,$6B,$47,$86
	.dc.b	$32,$A6,$5F,$82,$31,$08,$C3,$A9,$E2,$F0,$BA,$EE,$2C,$F0,$8E,$53
	.dc.b	$E9,$AF,$1C,$7C,$18,$DD,$70,$6C,$EA,$FD,$55,$B8,$F5,$19,$3C,$F7
	.dc.b	$DC,$97,$CE,$35,$3E,$96,$BE,$46,$CF,$D2,$6D,$F8,$A7,$C0,$4B,$F1
	.dc.b	$9F,$76,$FB,$46,$6B,$C7,$18,$9D,$C6,$0E,$A7,$82,$34,$FA,$AB,$E3
	.dc.b	$57,$33,$E7,$46,$CE,$2E,$4F,$A0,$E2,$BB,$C7,$BC,$D5,$FC,$5A,$EB
	.dc.b	$3A,$9F,$84,$6A,$FD,$6F,$F1,$BA,$FC,$BB,$6B,$F8,$3C,$62,$3D,$17
	.dc.b	$FC,$1A,$31,$1E,$83,$F9,$0B,$87,$BD,$FF,$89,$BB,$3F,$81,$D7,$29
	.dc.b	$FE,$91,$EB,$7E,$32,$73,$1C,$EB,$84,$E4,$95,$F7,$2B,$75,$DA,$3F
	.dc.b	$37,$18,$BF,$CA,$37,$0F,$EE,$F7,$CA,$FE,$64,$E6,$FF,$BC,$4E,$70
	.dc.b	$F9,$13,$9A,$F3,$D7,$86,$B4,$FA,$EB,$CE,$E4,$93,$DA,$76,$FB,$63
	.dc.b	$CF,$18,$8F,$51,$FB,$7A,$F0,$57,$48,$E5,$5D,$23,$CB,$5C,$70,$F7
	.dc.b	$B7,$0F,$7D,$6C,$F3,$1E,$03,$17,$E3,$3C,$13,$C6,$3E,$19,$A8,$F2
	.dc.b	$8C,$3C,$B1,$CE,$3C,$A4,$C3,$CB,$1C,$70,$F2,$BF,$C9,$7E,$0F,$F9
	.dc.b	$11,$FC,$8F,$E6,$AC,$64,$97,$FC,$A6,$32,$FF,$2B,$F8,$BF,$CA,$92
	.dc.b	$7F,$28,$FE,$63,$CE,$91,$F9,$76,$CF,$E5,$BC,$6F,$CE,$92,$7F,$38
	.dc.b	$FE,$83,$E9,$64,$7E,$7D,$B3,$FD,$FB,$47,$2F,$F6,$C6,$D7,$F6,$A6
	.dc.b	$CE,$E7,$FB,$8F,$C1,$BB,$29,$2F,$FB,$8C,$65,$E5,$F3,$5D,$48,$E0
	.dc.b	$FE,$EF,$F4,$9F,$E7,$EF,$A4,$BE,$A9,$CF,$CA,$BB,$F3,$1F,$DF,$BE
	.dc.b	$D5,$F2,$FF,$E0,$0D,$91,$87,$51,$98,$C4,$7F,$87,$BE,$6D,$F4,$BC
	.dc.b	$5B,$1F,$E0,$A6,$BB,$8E,$DF,$38,$F6,$69,$27,$FC,$23,$19,$6F,$C1
	.dc.b	$70,$23,$83,$FC,$2F,$F4,$4F,$85,$F5,$36,$33,$18,$3C,$3F,$21,$23
	.dc.b	$BB,$5F,$F1,$CD,$FF,$B7,$9F,$1F,$93,$E8,$23,$E0,$F1,$3B,$91,$C1
	.dc.b	$FE,$2A,$F9,$BF,$F5,$8B,$78,$FE,$74,$62,$3F,$C8,$ED,$7F,$C9,$9F
	.dc.b	$3F,$FE,$06,$7F,$C6,$6D,$7F,$C9,$AD,$77,$E3,$7C,$CF,$A7,$91,$F2
	.dc.b	$9A,$F0,$F8,$3F,$BD,$46,$92,$3E,$8F,$27,$D6,$23,$83,$C7,$EF,$72
	.dc.b	$23,$83,$CF,$EF,$7F,$83,$3C,$8F,$F7,$2E,$1F,$EA,$B9,$FF,$09,$46
	.dc.b	$27,$5D,$9F,$E1,$2A,$78,$B8,$BF,$F0,$C7,$0B,$BB,$10,$8C,$3D,$2F
	.dc.b	$FC,$37,$18,$86,$3E,$2F,$7B,$8E,$2F,$16,$C4,$AE,$47,$F8,$9E,$74
	.dc.b	$E0,$70,$FC,$97,$18,$7E,$3D,$EF,$8F,$8B,$9B,$CF,$5B,$9C,$9F,$E3
	.dc.b	$3E,$B7,$C9,$FB,$1B,$1D,$DA,$F0,$0B,$E4,$F8,$7B,$32,$78,$3C,$FD
	.dc.b	$AF,$8F,$9B,$D2,$BE,$94,$BF,$4C,$78,$7D,$AF,$17,$B5,$27,$C5,$1E
	.dc.b	$5F,$6B,$CD,$F4,$12,$7C,$D1,$DB,$DC,$91,$F1,$1F,$0F,$C8,$7B,$7D
	.dc.b	$C9,$3E,$D7,$E0,$7D,$07,$C1,$F9,$39,$3F,$05,$F8,$9F,$27,$EA,$F9
	.dc.b	$3F,$5F,$8D,$2B,$E8,$C7,$E1,$31,$FE,$11,$1F,$1D,$79,$59,$35,$E5
	.dc.b	$EF,$CA,$E1,$E5,$4B,$FB,$D6,$9F,$93,$EE,$C9,$FB,$EF,$CC,$FA,$0F
	.dc.b	$07,$91,$E1,$F7,$E5,$72,$8F,$27,$AF,$E5,$F8,$A9,$3D,$A3,$F0,$BE
	.dc.b	$1C,$8F,$C7,$78,$48,$FC,$4F,$9D,$ED,$EC,$95,$E6,$7E,$07,$6F,$C1
	.dc.b	$E9,$93,$EA,$7E,$27,$07,$E4,$BE,$03,$F2,$7E,$B4,$AE,$F1,$FA,$CF
	.dc.b	$A0,$F9,$3E,$5F,$86,$7F,$D1,$11,$8D,$3F,$D0,$ED,$8F,$F4,$67,$D0
	.dc.b	$7E,$E1,$FD,$32,$F7,$BF,$6E,$D8,$CC,$63,$E4,$B3,$E6,$C9,$E2,$F0
	.dc.b	$F3,$B8,$FB,$92,$7A,$BD,$3E,$2A,$47,$81,$AF,$E9,$F1,$7C,$5F,$F0
	.dc.b	$8F,$9F,$FD,$44,$FE,$EF,$E8,$1E,$1F,$F0,$C6,$C7,$EF,$E7,$6F,$EC
	.dc.b	$57,$9D,$CF,$CF,$93,$B8,$FE,$2B,$F0,$FF,$16,$1E,$5B,$4F,$E3,$2D
	.dc.b	$8F,$E1,$93,$B3,$F2,$EE,$3C,$5E,$67,$4F,$1B,$F9,$0F,$CF,$BE,$33
	.dc.b	$9F,$B1,$2F,$F8,$B4,$E5,$BF,$41,$C2,$A4,$60,$DF,$AD,$D2,$64,$60
	.dc.b	$E7,$B7,$A3,$9A,$46,$0E,$FB,$F9,$26,$46,$0F,$1F,$AB,$FC,$C9,$FF
	.dc.b	$67,$9F,$ED,$3C,$91,$98,$CC,$63,$B3,$71,$B9,$D3,$FE,$D8,$97,$CF
	.dc.b	$BE,$83,$72,$9C,$C6,$1B,$1D,$1A,$F0,$FA,$0F,$F1,$12,$69,$1E,$BF
	.dc.b	$0D,$FA,$5C,$7F,$4D,$1F,$EE,$D8,$C7,$81,$E8,$AF,$F7,$87,$D0,$39
	.dc.b	$7E,$0D,$6A,$FF,$DE,$3C,$B5,$FF,$7B,$35,$FF,$1C,$61,$FD,$45,$FF
	.dc.b	$14,$37,$97,$47,$1E,$67,$C6,$E2,$78,$7A,$3C,$CC,$5F,$F4,$B3,$F8
	.dc.b	$F4,$75,$F4,$BC,$06,$DC,$1F,$D8,$76,$FE,$AE,$DE,$7F,$07,$87,$E9
	.dc.b	$9C,$DF,$F5,$77,$FA,$B5,$71,$E5,$D5,$ED,$F8,$1F,$41,$EE,$BF,$EB
	.dc.b	$CD,$BF,$9B,$A3,$C2,$DF,$07,$0E,$09,$9F,$F7,$5C,$38,$2C,$64,$7C
	.dc.b	$87,$AF,$9D,$E0,$3C,$3E,$8D,$AE,$51,$E4,$F4,$31,$5E,$47,$CF,$EF
	.dc.b	$7A,$3C,$FB,$7E,$17,$0F,$8F,$CF,$B7,$EC,$7D,$FE,$7C,$9D,$1E,$97
	.dc.b	$F0,$55,$E9,$8F,$F0,$83,$EC,$AF,$63,$97,$FC,$08,$E5,$FE,$FD,$87
	.dc.b	$2F,$80,$E6,$DA,$E7,$E4,$76,$3B,$FA,$B6,$3F,$C3,$DF,$3E,$F6,$FF
	.dc.b	$28,$D8,$CF,$B7,$89,$FC,$C7,$CF,$BE,$1A,$98,$F6,$6D,$FF,$8A,$65
	.dc.b	$C1,$C3,$83,$8B,$67,$FC,$5D,$F3,$FF,$CB,$9B,$E4,$FE,$96,$3F,$83
	.dc.b	$C6,$0D,$E7,$93,$5E,$F5,$8F,$F1,$C5,$BC,$0F,$D7,$3F,$B3,$8C,$C6
	.dc.b	$3C,$4F,$D9,$7C,$FB,$CC,$BE,$AE,$1B,$7F,$E4,$3F,$9F,$79,$FF,$B9
	.dc.b	$7F,$43,$E1,$FF,$04,$73,$3B,$F1,$75,$F0,$72,$39,$78,$B7,$B7,$FC
	.dc.b	$7D,$F3,$CF,$77,$F8,$23,$6B,$FC,$A4,$BE,$07,$33,$FC,$27,$E7,$9F
	.dc.b	$2B,$4C,$79,$D7,$0F,$F2,$D4,$FB,$5B,$C3,$C7,$87,$93,$FC,$4F,$E7
	.dc.b	$9F,$0D,$79,$30,$DB,$FF,$32,$7C,$F7,$F0,$AD,$FF,$CA,$CF,$CF,$8C
	.dc.b	$FA,$3F,$C7,$63,$13,$AF,$1F,$06,$7C,$9B,$7F,$E6,$FF,$9E,$F7,$3F
	.dc.b	$E7,$1F,$9D,$F4,$5F,$F9,$8E,$72,$FF,$99,$63,$3D,$3F,$CD,$1B,$5F
	.dc.b	$C0,$E5,$1F,$E8,$18,$E9,$BF,$F3,$18,$C4,$F3,$DF,$06,$66,$D7,$8F
	.dc.b	$E0,$38,$1C,$7D,$EB,$19,$3D,$1F,$3D,$8C,$3C,$CF,$F3,$FF,$9D,$F1
	.dc.b	$CF,$FA,$35,$FE,$F8,$DC,$B5,$EF,$C8,$D3,$C6,$31,$7F,$E8,$5A,$FD
	.dc.b	$43,$7C,$6F,$E9,$9B,$3E,$2B,$C1,$13,$1F,$07,$C0,$5B,$E4,$F9,$49
	.dc.b	$1E,$59,$B3,$DF,$BF,$D4,$7A,$FE,$32,$D7,$AE,$3D,$DE,$06,$2B,$CC
	.dc.b	$FF,$5F,$F9,$E7,$C4,$FC,$3F,$9A,$C3,$70,$F8,$9E,$8F,$C7,$31,$87
	.dc.b	$BD,$FE,$D1,$F3,$DF,$C8,$FA,$3F,$B9,$B6,$BF,$01,$8B,$E5,$B7,$84
	.dc.b	$9F,$F5,$E5,$7F,$29,$38,$47,$F4,$97,$1E,$EF,$F7,$E8,$FC,$54,$7F
	.dc.b	$57,$7F,$C3,$6D,$73,$1B,$79,$79,$9E,$63,$CF,$DF,$6F,$C3,$1B,$95
	.dc.b	$FE,$99,$EB,$19,$8C,$47,$8B,$DB,$C1,$2B,$FC,$43,$28,$E1,$ED,$FC
	.dc.b	$63,$E3,$B6,$7E,$0F,$B7,$D4,$8F,$26,$DF,$FB,$46,$FF,$9C,$CF,$03
	.dc.b	$E4,$FA,$1F,$48,$8E,$67,$84,$7D,$6F,$6A,$4F,$D0,$F3,$2D,$75,$9F
	.dc.b	$71,$27,$D6,$FC,$5F,$9E,$7C,$7F,$F8,$9D,$FC,$4C,$FF,$BD,$5C,$B7
	.dc.b	$93,$53,$B6,$C6,$4E,$1E,$6F,$33,$FC,$65,$F3,$BF,$D6,$E3,$24,$C7
	.dc.b	$FD,$E5,$87,$F2,$E8,$FF,$07,$72,$FF,$90,$AB,$85,$6A,$FA,$57,$1A
	.dc.b	$D7,$B7,$8A,$D4,$E4,$FE,$2D,$7C,$AB,$FC,$9B,$87,$7A,$D9,$23,$9D
	.dc.b	$71,$8F,$F1,$04,$EE,$78,$E1,$EC,$C3,$C1,$18,$BE,$16,$C3,$C1,$1F
	.dc.b	$C9,$F0,$E4,$DC,$38,$BE,$6C,$2D,$F8,$26,$BC,$E2,$D7,$28,$D5,$7C
	.dc.b	$83,$91,$E2,$8D,$41,$B7,$E4,$8E,$27,$C1,$9F,$21,$CD,$FF,$20,$46
	.dc.b	$DE,$57,$E2,$3E,$14,$F7,$7C,$F3,$BA,$F0,$92,$37,$5D,$CF,$0E,$1F
	.dc.b	$25,$3F,$49,$D1,$E0,$BF,$9B,$7D,$6B,$C5,$1E,$79,$F8,$CF,$8C,$DE
	.dc.b	$1A,$8E,$75,$C6,$75,$1E,$07,$E4,$BE,$0C,$3D,$11,$D6,$78,$4E,$6F
	.dc.b	$D7,$87,$A4,$95,$F2,$A7,$A1,$F2,$B0,$F1,$3F,$EB,$8C,$3F,$24,$DB
	.dc.b	$FD,$A1,$2B,$FD,$63,$3A,$9E,$86,$E3,$33,$FB,$9F,$3F,$F0,$A4,$8F
	.dc.b	$C1,$36,$7D,$8D,$C3,$CF,$1F,$9E,$8C,$4F,$A6,$7F,$4F,$18,$9F,$AF
	.dc.b	$3F,$9B,$8C,$47,$D5,$3F,$76,$E2,$FF,$02,$DB,$F3,$57,$90,$95,$E4
	.dc.b	$6D,$F9,$23,$F3,$15,$C6,$7F,$8A,$36,$3B,$C7,$03,$D0,$79,$4C,$3E
	.dc.b	$83,$DC,$E1,$F4,$4F,$BA,$31,$7E,$86,$DF,$A2,$BC,$C4,$AF,$35,$7F
	.dc.b	$B1,$AB,$67,$98,$B3,$E6,$38,$3E,$67,$C6,$60,$F3,$1F,$01,$FE,$65
	.dc.b	$F1,$5F,$CD,$4F,$E6,$DF,$2D,$63,$24,$9F,$E5,$71,$97,$F2,$27,$F1
	.dc.b	$3E,$3F,$EC,$9A,$7F,$5B,$AF,$F7,$EB,$67,$FC,$78,$7F,$00,$9F,$E8
	.dc.b	$DE,$54,$AF,$C9,$B6,$76,$3E,$4B,$F2,$CF,$FB,$1D,$AE,$74,$FE,$E9
	.dc.b	$18,$F8,$47,$77,$F7,$76,$E1,$D2,$FF,$DA,$D3,$93,$D3,$E7,$49,$F4
	.dc.b	$B6,$7C,$7F,$4E,$F9,$DF,$F0,$13,$6B,$A3,$67,$73,$FE,$07,$F9,$EF
	.dc.b	$F0,$E3,$E1,$6C,$7B,$97,$CA,$BC,$0D,$78,$3D,$EB,$E1,$FC,$C3,$E7
	.dc.b	$7D,$50,$8C,$3E,$D6,$BC,$23,$7F,$32,$F8,$BA,$49,$F8,$91,$97,$E3
	.dc.b	$7D,$3F,$F8,$7D,$34,$9B,$78,$BE,$79,$EC,$52,$57,$CF,$8C,$BD,$BC
	.dc.b	$8F,$19,$1C,$1E,$2F,$23,$CE,$47,$07,$3F,$0F,$C0,$47,$07,$DD,$E7
	.dc.b	$FF,$25,$37,$7F,$8F,$18,$8F,$C6,$B5,$FC,$AF,$CE,$7F,$86,$1F,$C3
	.dc.b	$35,$FC,$9B,$5D,$F8,$9D,$3C,$29,$1E,$16,$BC,$3E,$7B,$EB,$D1,$A4
	.dc.b	$8F,$A3,$F1,$7F,$58,$8E,$0F,$1F,$CF,$72,$23,$83,$CF,$E1,$7F,$AA
	.dc.b	$7C,$57,$17,$FA,$AF,$14,$E6,$30,$DA,$09,$78,$99,$87,$2B,$BB,$0D
	.dc.b	$8F,$EC,$CD,$7F,$82,$D7,$FB,$D5,$FE,$D4,$BC,$2E,$8E,$63,$7E,$E6
	.dc.b	$27,$3E,$E7,$87,$89,$F5,$34,$CD,$F4,$9A,$F8,$1E,$BE,$E7,$1E,$09
	.dc.b	$3F,$DE,$ED,$1D,$76,$F7,$75,$FE,$17,$87,$E4,$DF,$C4,$8F,$B0,$FF
	.dc.b	$80,$DA,$FF,$22,$78,$70,$FF,$B6,$5F,$F7,$EC,$6F,$93,$A9,$DB,$C6
	.dc.b	$F6,$71,$C9,$F2,$3E,$FF,$93,$F1,$E5,$27,$C0,$F9,$A7,$CF,$32,$7C
	.dc.b	$8F,$AB,$E7,$7E,$41,$F2,$78,$A4,$FC,$18,$F7,$FD,$0F,$C0,$F2,$64
	.dc.b	$FB,$63,$E1,$FC,$07,$C4,$E3,$93,$F0,$63,$EA,$72,$C8,$F9,$8F,$D7
	.dc.b	$AF,$C4,$7D,$47,$F1,$2E,$2F,$97,$48,$C7,$17,$FB,$FD,$B1,$F9,$27
	.dc.b	$FE,$00,$BF,$E3,$15,$F0,$3D,$09,$19,$F8,$BF,$C3,$69,$F8,$7C,$E4
	.dc.b	$FD,$F8,$F1,$FB,$B2,$3A,$3E,$6F,$7B,$CF,$F3,$92,$7C,$0F,$AB,$E8
	.dc.b	$FF,$25,$C1,$F9,$3E,$B9,$5F,$8B,$8F,$7F,$BD,$F0,$3E,$1A,$4F,$A6
	.dc.b	$3E,$1F,$C3,$FC,$4F,$52,$4F,$B6,$3F,$5D,$FB,$27,$F3,$F8,$FE,$DA
	.dc.b	$DE,$4F,$92,$DE,$4F,$DB,$CA,$3E,$4A,$FA,$1F,$DC,$BF,$E5,$9F,$A2
	.dc.b	$FE,$78,$FF,$78,$6D,$FD,$49,$DC,$FF,$7C,$5D,$3F,$C0,$57,$5D,$31
	.dc.b	$19,$EB,$95,$FE,$67,$AF,$DE,$5F,$5E,$BE,$5C,$D2,$B4,$F5,$E7,$91
	.dc.b	$C5,$AF,$8F,$8B,$F8,$97,$D0,$BE,$73,$F8,$A7,$D1,$3C,$EA,$6A,$37
	.dc.b	$D7,$70,$FF,$3D,$5F,$16,$C7,$F1,$DF,$94,$F8,$7C,$9F,$33,$9D,$A7
	.dc.b	$F2,$16,$DF,$FA,$0E,$B6,$FF,$28,$5F,$3D,$CF,$D5,$93,$B8,$FE,$55
	.dc.b	$F4,$3F,$9A,$8F,$F4,$65,$FD,$0E,$2F,$D2,$B7,$2F,$BB,$5B,$AF,$F1
	.dc.b	$7A,$F0,$39,$F8,$32,$7F,$D2,$D3,$96,$FC,$AF,$82,$A4,$60,$DF,$07
	.dc.b	$B7,$9E,$46,$0E,$7C,$1E,$89,$91,$83,$BF,$85,$FE,$AC,$7F,$C6,$F8
	.dc.b	$A3,$11,$FE,$40,$8C,$6B,$FE,$31,$6C,$70,$F0,$9F,$45,$96,$C6,$63
	.dc.b	$07,$F8,$81,$7D,$15,$23,$FE,$24,$7F,$C4,$BE,$E7,$F3,$AC,$3F,$88
	.dc.b	$CF,$F8,$B7,$83,$FD,$78,$FF,$60,$AF,$EC,$3F,$44,$E5,$FF,$2B,$E1
	.dc.b	$C6,$FF,$CB,$3C,$B7,$0F,$D8,$B5,$FE,$D3,$35,$FF,$17,$4F,$6E,$5C
	.dc.b	$1B,$E5,$FF,$3E,$4F,$F9,$A8,$FC,$4C,$F9,$3D,$9E,$47,$07,$07,$A3
	.dc.b	$D9,$EC,$69,$C1,$FE,$D7,$D5,$CD,$E5,$6C,$7E,$43,$C1,$C4,$E7,$92
	.dc.b	$4F,$18,$B8,$FF,$9F,$BE,$8B,$E4,$37,$5F,$57,$7F,$C1,$9C,$E5,$FC
	.dc.b	$F9,$F8,$35,$C7,$7D,$5E,$DE,$1F,$19,$32,$FF,$77,$5F,$FA,$1F,$0E
	.dc.b	$53,$9C,$3C,$5F,$45,$67,$B3,$E1,$F7,$BC,$5F,$1B,$71,$E6,$E1,$EB
	.dc.b	$F1,$97,$0F,$63,$E9,$F9,$F9,$3A,$3C,$6F,$DD,$AF,$F1,$2C,$7F,$89
	.dc.b	$B1,$FE,$B7,$1F,$02,$3F,$99,$BA,$BF,$F5,$2B,$97,$FC,$5A,$B1,$E9
	.dc.b	$EE,$75,$6D,$75,$39,$F8,$9E,$13,$C5,$F4,$56,$35,$87,$F8,$E3,$E8
	.dc.b	$9F,$25,$F9,$36,$CE,$A5,$0F,$A0,$D5,$7E,$67,$E8,$9F,$21,$4C,$7C
	.dc.b	$1D,$C7,$94,$B8,$78,$70,$F1,$6C,$F7,$FA,$27,$8F,$FA,$56,$C6,$7C
	.dc.b	$7C,$4E,$74,$93,$C7,$E8,$AE,$FF,$EC,$EC,$5D,$FF,$AF,$7F,$6D,$5E
	.dc.b	$D7,$2D,$F0,$FF,$69,$EA,$BC,$AB,$E8,$E1,$B5,$F1,$7E,$8B,$F9,$63
	.dc.b	$7C,$2F,$DE,$46,$33,$FF,$6F,$BF,$AA,$6F,$CC,$68,$DF,$8F,$6F,$E8
	.dc.b	$7D,$13,$E8,$BF,$83,$36,$33,$F3,$18,$9F,$E1,$5F,$44,$F8,$4A,$63
	.dc.b	$E1,$6D,$FD,$B9,$F6,$37,$7E,$3B,$F2,$7F,$8A,$7D,$13,$E0,$2F,$16
	.dc.b	$1B,$7F,$85,$FA,$27,$A1,$FC,$6D,$F9,$7C,$5F,$EF,$E7,$33,$BE,$7D
	.dc.b	$78,$D2,$39,$79,$36,$DE,$4F,$F9,$03,$E8,$1C,$9F,$C9,$FE,$8B,$D3
	.dc.b	$5A,$C3,$FC,$37,$5A,$7F,$31,$3A,$BF,$F2,$4B,$FE,$89,$8C,$F0,$73
	.dc.b	$E1,$E8,$C8,$FF,$7F,$4E,$F0,$FF,$13,$9F,$E9,$E7,$0E,$1A,$FF,$8B
	.dc.b	$17,$BD,$FD,$22,$38,$35,$46,$A7,$BC,$E6,$31,$3E,$D3,$C1,$F0,$9F
	.dc.b	$06,$7E,$13,$FD,$27,$E8,$B9,$DF,$ED,$DF,$F2,$FB,$6F,$5E,$EC,$8F
	.dc.b	$5E,$1F,$EB,$8A,$DB,$86,$BF,$01,$AF,$F9,$21,$FD,$9B,$BF,$0F,$D9
	.dc.b	$18,$8E,$11,$8F,$13,$CC,$79,$DB,$5E,$78,$D3,$E4,$95,$7F,$80,$E3
	.dc.b	$0D,$8F,$37,$77,$80,$93,$A7,$C7,$DD,$23,$8B,$67,$DD,$F0,$BE,$82
	.dc.b	$4F,$67,$D7,$E5,$C8,$F0,$B6,$7E,$4F,$CF,$FF,$37,$7F,$0A,$D7,$FC
	.dc.b	$DA,$D7,$FD,$CA,$D4,$FF,$64,$78,$FA,$BF,$85,$8C,$47,$28,$FC,$47
	.dc.b	$CE,$FF,$95,$EF,$F9,$A3,$67,$F0,$C7,$49,$FF,$06,$46,$79,$3F,$D5
	.dc.b	$37,$FB,$4A,$F3,$F4,$F7,$26,$1F,$46,$7F,$73,$5C,$3C,$6F,$F0,$B4
	.dc.b	$62,$7D,$B1,$8F,$52,$47,$F4,$79,$FE,$E4,$FC,$28,$CE,$1F,$21,$F7
	.dc.b	$12,$3C,$1E,$A7,$6F,$27,$BB,$63,$A4,$7C,$52,$3E,$76,$D7,$C2,$FA
	.dc.b	$A4,$F1,$8C,$3E,$AF,$53,$D7,$EA,$58,$F4,$38,$7E,$13,$F4,$C8,$FE
	.dc.b	$57,$E7,$DF,$41,$F3,$B5,$FF,$1D,$46,$27,$EA,$C6,$5A,$7F,$6B,$8D
	.dc.b	$61,$F6,$38,$3F,$A9,$BF,$6A,$3F,$C7,$BB,$FF,$B8,$1B,$5F,$EB,$87
	.dc.b	$F9,$F3,$CF,$C9,$F0,$91,$ED,$E5,$5B,$E9,$36,$FA,$4D,$AE,$93,$68
	.dc.b	$2F,$BB,$CA,$7C,$04,$7D,$15,$C5,$E1,$56,$B8,$DF,$AE,$30,$7C,$8A
	.dc.b	$B7,$CE,$7C,$AD,$3C,$A5,$BF,$2D,$F4,$8C,$C7,$94,$93,$C9,$B1,$A8
	.dc.b	$F1,$3F,$18,$E9,$1E,$86,$CF,$58,$EF,$7E,$67,$DF,$3A,$8F,$75,$E6
	.dc.b	$FA,$D6,$8F,$09,$CA,$BA,$CE,$DF,$21,$F7,$5C,$46,$CE,$53,$D0,$93
	.dc.b	$C2,$3E,$1C,$F0,$7D,$47,$08,$D3,$89,$E2,$F0,$8F,$51,$E4,$2B,$E8
	.dc.b	$3F,$05,$3D,$6F,$F2,$71,$C4,$D3,$FE,$94,$7F,$B3,$B5,$FA,$A6,$2F
	.dc.b	$FB,$14,$66,$3D,$C6,$6B,$F5,$D3,$A3,$FD,$4D,$C8,$F9,$6F,$C7,$36
	.dc.b	$7E,$CB,$6B,$D0,$DC,$7D,$F1,$FE,$5E,$AE,$24,$AE,$2D,$8E,$2D,$9E
	.dc.b	$35,$CE,$BE,$5B,$6B,$EB,$4F,$F1,$07,$F5,$31,$97,$E8,$4F,$EB,$1F
	.dc.b	$13,$E0,$23,$E0,$6C,$F8,$2B,$C5,$1E,$06,$CF,$82,$3A,$57,$7C,$3E
	.dc.b	$AC,$F6,$3A,$12,$FA,$57,$63,$A1,$63,$A1,$EA,$BE,$EF,$A8,$C4,$6D
	.dc.b	$C1,$DE,$7A,$BD,$E3,$AB,$4E,$A6,$DC,$C7,$02,$38,$24,$70,$BF,$76
	.dc.b	$1F,$84,$BF,$E6,$3E,$EF,$E3,$09,$7F,$CD,$DF,$F1,$85,$7E,$EE,$FF
	.dc.b	$D3,$F9,$70,$AF,$F4,$F4,$E7,$27,$C7,$7E,$58,$93,$FD,$05,$FF,$1D
	.dc.b	$7C,$67,$F8,$C3,$0F,$F1,$5B,$5F,$F2,$3B,$5F,$C0,$35,$FE,$5B,$BB
	.dc.b	$E7,$BF,$C8,$8D,$BF,$F1,$13,$67,$FC,$6E,$D7,$FC,$A7,$BB,$C9,$FF
	.dc.b	$02,$92,$BF,$C5,$51,$97,$FC,$59,$F2,$FF,$E1,$12,$57,$F9,$4D,$FF
	.dc.b	$0A,$7C,$F3,$E6,$BC,$93,$FE,$46,$8C,$65,$FE,$35,$5F,$3C,$EE,$C5
	.dc.b	$F0,$BF,$F3,$3F,$D2,$BD,$EF,$F1,$E4,$23,$0F,$F8,$FD,$AF,$F9,$51
	.dc.b	$AF,$F9,$C3,$27,$C1,$F2,$6C,$7B,$23,$07,$6F,$5F,$FD,$0F,$1F,$E4
	.dc.b	$B6,$BF,$E6,$06,$BF,$E7,$C6,$BF,$B5,$F9,$D7,$A5,$E8,$6C,$7F,$95
	.dc.b	$5A,$FF,$9A,$9A,$FF,$A1,$BE,$57,$FD,$D6,$FF,$96,$DA,$FF,$9C,$1A
	.dc.b	$EF,$9F,$E2,$FC,$C3,$E7,$69,$27,$FC,$C9,$19,$7C,$5E,$AF,$94,$8E
	.dc.b	$0F,$F3,$57,$D5,$7D,$56,$F0,$FD,$28,$C4,$7F,$A0,$9A,$FF,$A7,$BE
	.dc.b	$7B,$FD,$2C,$9A,$4E,$6F,$B7,$F5,$3F,$ED,$59,$FF,$3C,$35,$FF,$46
	.dc.b	$B5,$FF,$54,$B5,$FE,$44,$BE,$9B,$99,$23,$9B,$F8,$8A,$FF,$6E,$47
	.dc.b	$FA,$15,$AF,$FA,$69,$AF,$A7,$BF,$F1,$FF,$80,$7A,$89,$86,$0E,$9F
	.dc.b	$3D,$2B,$A3,$5E,$9F,$3D,$E1,$23,$E1,$7C,$9E,$DC,$8F,$23,$5F,$27
	.dc.b	$B7,$E9,$23,$E9,$7D,$9E,$DC,$8F,$63,$5F,$67,$B7,$F0,$48,$FC,$17
	.dc.b	$F5,$7F,$40,$C6,$92,$7F,$D5,$51,$97,$F1,$3D,$DF,$D4,$7C,$BF,$EA
	.dc.b	$D0,$E5,$FE,$D8,$FF,$90,$23,$31,$88,$FF,$46,$BF,$C1,$5D,$6B,$FD
	.dc.b	$6E,$3E,$9C,$F0,$F2,$FB,$4E,$5B,$5F,$E4,$CA,$DC,$FF,$63,$C1,$B6
	.dc.b	$A9,$18,$7B,$9F,$E5,$47,$0F,$FA,$81,$79,$5C,$6F,$E0,$4F,$A3,$3C
	.dc.b	$19,$BD,$54,$9E,$A7,$C3,$79,$7C,$EF,$8F,$03,$78,$25,$70,$9C,$BC
	.dc.b	$B1,$FE,$92,$DF,$91,$F2,$1E,$5F,$2F,$CD,$F3,$5E,$7E,$29,$5C,$63
	.dc.b	$C1,$C7,$23,$B3,$E3,$AF,$8D,$DC,$FA,$2A,$4A,$F5,$C3,$E0,$37,$F0
	.dc.b	$12,$72,$F2,$F9,$BF,$76,$9E,$FC,$E5,$7B,$63,$BF,$C0,$C8,$EA,$F8
	.dc.b	$BE,$07,$C7,$F3,$52,$7C,$2F,$9B,$E8,$3E,$57,$07,$CB,$EC,$95,$F2
	.dc.b	$63,$D7,$F0,$3E,$CF,$87,$93,$EA,$8F,$7F,$C4,$7C,$0F,$83,$93,$EE
	.dc.b	$8F,$C4,$7B,$1F,$D4,$7C,$0F,$CE,$B7,$BF,$FB,$8F,$AB,$AE,$7F,$C7
	.dc.b	$7C,$FF,$E2,$AF,$F4,$7F,$41,$FD,$33,$2F,$F7,$5B,$63,$F4,$F3,$BC
	.dc.b	$3F,$C7,$4F,$E0,$2B,$FD,$E2,$FE,$49,$FC,$5B,$5C,$F7,$F8,$3B,$E4
	.dc.b	$69,$F1,$F6,$F9,$39,$E4,$F2,$7C,$FD,$12,$3A,$B5,$F0,$F1,$FE,$A7
	.dc.b	$E7,$DF,$35,$FA,$BF,$A0,$7C,$15,$35,$1B,$EC,$B8,$7E,$8A,$F8,$B6
	.dc.b	$3F,$69,$F0,$3F,$0F,$83,$FC,$B5,$D0,$D3,$F7,$0D,$BF,$D5,$D6,$DF
	.dc.b	$F6,$5A,$F9,$CE,$7E,$C4,$9D,$C7,$F0,$1F,$A0,$78,$5D,$3C,$99,$5F
	.dc.b	$B6,$9C,$B7,$C3,$E1,$52,$30,$6F,$E5,$3A,$4C,$8C,$1C,$F8,$3D,$3C
	.dc.b	$D2,$30,$77,$E0,$F2,$4C,$8C,$1E,$3F,$11,$D5,$FE,$7D,$8F,$F3,$F6
	.dc.b	$69,$CC,$64,$46,$67,$11,$89,$DB,$CE,$5E,$23,$E8,$37,$18,$6C,$66
	.dc.b	$7B,$4E,$5F,$F1,$8F,$D0,$7F,$8C,$93,$48,$F8,$BD,$6F,$F4,$D6,$3F
	.dc.b	$D1,$A3,$FC,$7D,$CD,$FC,$99,$1C,$87,$D0,$39,$7E,$15,$F2,$AF,$89
	.dc.b	$CB,$6F,$FD,$44,$D7,$9C,$D7,$FC,$7F,$3D,$F9,$70,$6F,$9B,$FD,$3B
	.dc.b	$3F,$8B,$9F,$27,$C9,$F1,$36,$E0,$F4,$7C,$7F,$53,$4E,$0F,$E8,$3A
	.dc.b	$39,$BC,$8D,$8F,$E8,$B8,$7F,$4E,$77,$3A,$E3,$91,$83,$9B,$ED,$FA
	.dc.b	$0F,$93,$97,$E1,$A7,$FD,$85,$18,$C3,$FB,$9E,$0E,$FE,$1A,$F1,$DA
	.dc.b	$61,$FD,$5B,$0E,$0F,$28,$ED,$F1,$F2,$35,$F4,$B7,$3F,$01,$F6,$2D
	.dc.b	$5E,$5F,$A1,$D6,$9F,$E0,$49,$47,$A7,$E0,$BD,$5F,$23,$72,$F1,$38
	.dc.b	$7C,$9F,$23,$72,$EE,$30,$47,$C3,$F9,$E9,$5C,$0F,$63,$F8,$3B,$FC
	.dc.b	$0D,$F2,$6E,$5F,$E1,$87,$F0,$11,$9D,$7F,$15,$E6,$7F,$76,$C3,$FB
	.dc.b	$DB,$67,$C6,$FD,$96,$F9,$3F,$91,$FB,$E7,$85,$4C,$78,$77,$0F,$F0
	.dc.b	$2F,$D1,$3E,$7B,$F3,$0D,$8C,$FD,$4E,$27,$A1,$23,$FC,$1B,$83,$E8
	.dc.b	$31,$9F,$C6,$FB,$EF,$A7,$5F,$E1,$26,$C6,$79,$FF,$4D,$18,$E1,$FF
	.dc.b	$1C,$FD,$17,$F0,$4B,$FF,$0C,$B6,$FE,$8A,$F6,$7F,$5C,$E6,$BF,$C8
	.dc.b	$35,$FC,$B8,$FF,$10,$96,$BE,$CB,$FE,$DD,$8D,$ED,$FB,$78,$7A,$D4
	.dc.b	$C7,$B1,$63,$FC,$57,$1F,$EE,$A6,$FB,$9F,$BF,$8F,$CF,$46,$0F,$F1
	.dc.b	$8F,$DE,$35,$FE,$08,$D8,$CF,$C8,$E2,$73,$F1,$12,$78,$FC,$CD,$D7
	.dc.b	$F2,$3F,$45,$FD,$82,$7F,$C7,$AD,$8C,$CF,$F1,$18,$C6,$DF,$E6,$4F
	.dc.b	$31,$EF,$FF,$15,$6E,$1F,$E4,$65,$EC,$E6,$7F,$8D,$7D,$E3,$EB,$29
	.dc.b	$8E,$5B,$87,$F9,$3B,$E8,$DF,$05,$FC,$85,$B1,$9F,$92,$C4,$FF,$25
	.dc.b	$F2,$9E,$7D,$31,$E8,$5B,$FF,$2C,$CB,$E5,$38,$1C,$7C,$8B,$3F,$E5
	.dc.b	$EF,$85,$FE,$F0,$DF,$77,$F9,$97,$D1,$7C,$28,$FE,$7B,$CF,$FE,$49
	.dc.b	$9F,$E6,$D5,$F0,$27,$97,$1E,$27,$2C,$8E,$DC,$76,$FF,$81,$DF,$CA
	.dc.b	$7F,$82,$FD,$17,$E3,$6B,$BB,$F9,$1E,$6F,$F4,$AB,$99,$FE,$9E,$E6
	.dc.b	$77,$E0,$7F,$49,$8C,$4E,$AF,$13,$9F,$76,$5F,$F9,$FB,$4F,$F6,$83
	.dc.b	$FE,$80,$C1,$3A,$C3,$CB,$EE,$DC,$BC,$6F,$7E,$7F,$07,$D1,$5C,$79
	.dc.b	$38,$69,$D6,$31,$7F,$02,$7D,$1E,$B6,$30,$F7,$BF,$D8,$FE,$B1,$F3
	.dc.b	$FE,$DE,$8C,$37,$2F,$47,$4F,$C1,$31,$97,$CC,$7F,$B6,$7D,$1F,$F7
	.dc.b	$69,$FF,$4E,$B6,$7F,$9A,$E0,$F9,$2D,$C4,$AD,$F4,$D8,$FE,$77,$7F
	.dc.b	$D4,$1F,$F0,$0D,$BE,$8B,$BF,$5F,$A5,$B8,$78,$3B,$3C,$66,$2F,$9B
	.dc.b	$88,$FF,$10,$E1,$F9,$E6,$C7,$F8,$46,$7F,$C1,$F3,$97,$FC,$55,$83
	.dc.b	$EA,$BD,$FE,$2F,$C9,$6E,$1F,$02,$39,$52,$FF,$86,$3F,$E3,$36,$BF
	.dc.b	$E3,$96,$BF,$A2,$5E,$B7,$26,$D7,$A2,$30,$47,$E7,$FA,$FF,$4A,$3E
	.dc.b	$9B,$F5,$3E,$6B,$91,$27,$E0,$BD,$7E,$27,$EC,$9B,$8F,$B2,$E1,$A7
	.dc.b	$DC,$8C,$61,$FD,$C3,$E8,$DF,$39,$FE,$2D,$6D,$7F,$B7,$BE,$93,$F6
	.dc.b	$5D,$08,$FD,$FC,$EA,$F9,$5F,$C8,$8C,$47,$F9,$75,$AF,$08,$F9,$BF
	.dc.b	$46,$FA,$9F,$07,$CE,$7F,$8F,$63,$18,$7F,$BB,$B0,$4E,$6B,$73,$FE
	.dc.b	$F2,$AD,$E1,$EB,$AF,$57,$CD,$4A,$FF,$7B,$F1,$FF,$13,$AF,$81,$39
	.dc.b	$6C,$EB,$C6,$C6,$9E,$F9,$FF,$2C,$E6,$BC,$DF,$F2,$0C,$3A,$5E,$ED
	.dc.b	$7B,$AF,$55,$F0,$9F,$D1,$E1,$5D,$C5,$C3,$9D,$7C,$E8,$D5,$F3,$AF
	.dc.b	$D9,$DE,$27,$83,$6B,$B9,$EB,$2C,$FB,$89,$87,$9F,$0E,$F1,$EE,$BE
	.dc.b	$4F,$6A,$E3,$5E,$DC,$3C,$4D,$AF,$E2,$58,$70,$8C,$57,$86,$FB,$BE
	.dc.b	$B3,$99,$F2,$67,$83,$98,$F0,$1C,$0F,$23,$C1,$F9,$2D,$7E,$4B,$70
	.dc.b	$F5,$C6,$0E,$45,$AF,$81,$5E,$08,$EA,$78,$23,$D2,$F0,$C3,$DF,$7E
	.dc.b	$73,$C0,$E2,$38,$CF,$49,$CE,$1F,$51,$A7,$96,$FC,$15,$E3,$8E,$8F
	.dc.b	$79,$E0,$D8,$FF,$5F,$DF,$E0,$63,$F5,$93,$FA,$F2,$4E,$9B,$1F,$EC
	.dc.b	$99,$E0,$FC,$DB,$FB,$5F,$1F,$F8,$C2,$5F,$B1,$B3,$F0,$5B,$3F,$10
	.dc.b	$E3,$3B,$24,$ED,$B1,$B7,$F8,$3C,$F2,$7C,$46,$0D,$98,$9F,$D7,$C6
	.dc.b	$DF,$9B,$1B,$7B,$12,$7B,$3C,$23,$9B,$C0,$C1,$B7,$B3,$FB,$69,$EC
	.dc.b	$FD,$33,$73,$D1,$D4,$F4,$74,$60,$E8,$77,$31,$1C,$69,$5D,$9F,$C4
	.dc.b	$4E,$AB,$C7,$1A,$7B,$3A,$31,$1D,$8F,$0B,$8A,$ED,$5D,$6F,$F9,$D3
	.dc.b	$6F,$F9,$53,$6F,$FA,$03,$FE,$EB,$FA,$1F,$F4,$4C,$7F,$8E,$83,$87
	.dc.b	$FC,$7A,$D7,$FD,$F6,$D5,$A7,$F6,$6C,$3F,$91,$36,$F8,$FC,$37,$F3
	.dc.b	$DE,$6F,$E7,$DF,$0D,$FD,$08,$93,$E1,$C0,$74,$8F,$EE,$7F,$35,$FD
	.dc.b	$0A,$75,$28,$C3,$C5,$AF,$AD,$AF,$F9,$5B,$E6,$3F,$D3,$0F,$BD,$AE
	.dc.b	$7C,$AF,$F0,$0C,$62,$3C,$7F,$58,$F8,$6F,$F0,$2B,$70,$F1,$B6,$7D
	.dc.b	$4D,$7E,$67,$D2,$FF,$8A,$6B,$D2,$D8,$F8,$0B,$E9,$5E,$1E,$BC,$4F
	.dc.b	$05,$7C,$5E,$2F,$F7,$7B,$F0,$9A,$E7,$6F,$8A,$FD,$A6,$FD,$63,$E7
	.dc.b	$29,$27,$E5,$C6,$5F,$9B,$F3,$DF,$E2,$24,$D2,$6D,$DB,$E9,$5F,$1F
	.dc.b	$49,$5F,$5A,$32,$FD,$8F,$A5,$F2,$A6,$91,$F4,$E7,$FE,$9A,$7D,$DE
	.dc.b	$46,$08,$F5,$F1,$3C,$44,$70,$7E,$1B,$9B,$FC,$B0,$DD,$FE,$2C,$62
	.dc.b	$3F,$11,$1F,$23,$E9,$3F,$C4,$71,$F8,$E6,$BF,$95,$6B,$FA,$76,$BF
	.dc.b	$EA,$29,$7D,$1F,$62,$47,$8D,$ED,$E5,$C8,$EC,$D7,$F4,$5D,$0F,$8F
	.dc.b	$FB,$6D,$8F,$D0,$B5,$DC,$7A,$7E,$B2,$47,$A5,$AF,$A7,$C1,$FF,$5B
	.dc.b	$F8,$1F,$E1,$7C,$FF,$C3,$93,$FB,$08,$C4,$EB,$8F,$FD,$7B,$5F,$C6
	.dc.b	$AB,$51,$FC,$13,$EA,$35,$1F,$C1,$1A,$FF,$0A,$6C,$BE,$7A,$FD,$0E
	.dc.b	$19,$5D,$0F,$E0,$32,$9E,$3E,$37,$4E,$0F,$F3,$D5,$6C,$FE,$05,$E8
	.dc.b	$69,$C5,$FF,$BC,$1B,$59,$5A,$7C,$1F,$77,$CF,$C0,$F9,$9A,$4A,$ED
	.dc.b	$E0,$3D,$EF,$6A,$F1,$6C,$49,$DC,$74,$F9,$9D,$F2,$C9,$F4,$62,$9F
	.dc.b	$8F,$F8,$1F,$BE,$7D,$9D,$AF,$5B,$E5,$B6,$3E,$04,$ED,$F5,$F3,$7F
	.dc.b	$BC,$BE,$F1,$F4,$FF,$49,$FB,$13,$A3,$F9,$6B,$8E,$1F,$F2,$C3,$C2
	.dc.b	$3E,$67,$DD,$FB,$5B,$BF,$DC,$8F,$85,$1A,$DF,$CB,$E9,$F9,$BE,$3A
	.dc.b	$4F,$AA,$38,$7C,$AF,$1F,$72,$4F,$68,$FC,$3F,$DD,$BD,$1B,$84,$AF
	.dc.b	$BD,$39,$7E,$2F,$9B,$F1,$3C,$09,$18,$3F,$09,$D3,$F8,$67,$E0,$1F
	.dc.b	$77,$C4,$FC,$AC,$7E,$DA,$37,$DD,$8E,$CC,$F6,$C9,$E7,$1F,$4B,$DB
	.dc.b	$FA,$7F,$11,$27,$EB,$C7,$E8,$3E,$E5,$F2,$B7,$E0,$E5,$7E,$5E,$72
	.dc.b	$FE,$67,$EE,$7F,$8F,$3E,$67,$EF,$37,$BF,$F5,$71,$AE,$BF,$F4,$77
	.dc.b	$DB,$BE,$07,$F0,$BE,$74,$C3,$F4,$93,$97,$F4,$DF,$71,$FA,$37,$E1
	.dc.b	$AF,$8D,$FD,$BB,$63,$7C,$3F,$BA,$F3,$DF,$21,$7E,$9E,$57,$EC,$67
	.dc.b	$2F,$EC,$FB,$BF,$D4,$5A,$FF,$A8,$DF,$E5,$2D,$F8,$4F,$E0,$B1,$AF
	.dc.b	$1B,$FB,$FD,$3E,$E3,$E0,$DF,$FA,$AD,$7D,$2F,$F0,$C6,$CF,$F0,$17
	.dc.b	$75,$C9,$74,$6F,$A2,$56,$E3,$F8,$AF,$D6,$7E,$86,$FC,$3F,$5A,$FA
	.dc.b	$5F,$E3,$2D,$BF,$E1,$D5,$B7,$FC,$2A,$BE,$97,$9F,$99,$27,$71,$DB
	.dc.b	$D6,$91,$C9,$F0,$FA,$DE,$2F,$A7,$93,$D9,$F2,$F4,$48,$F0,$B5,$FC
	.dc.b	$DE,$3F,$6F,$87,$FD,$A1,$F5,$9F,$C2,$E7,$F9,$27,$D6,$BB,$3F,$99
	.dc.b	$B6,$3F,$94,$4E,$DF,$D2,$AF,$96,$FE,$72,$E6,$FA,$FA,$B8,$9F,$E7
	.dc.b	$BF,$55,$F1,$6F,$F5,$BE,$27,$E2,$5B,$A6,$B8,$F7,$D1,$67,$F9,$C5
	.dc.b	$F1,$CB,$FD,$D0,$FD,$C8,$FD,$BF,$D6,$7F,$11,$7A,$FC,$6F,$F4,$CC
	.dc.b	$BE,$94,$27,$57,$FE,$F1,$7F,$B2,$37,$D3,$FE,$A7,$1A,$E8,$FC,$7B
	.dc.b	$A6,$F9,$7A,$F6,$A6,$1F,$D3,$6B,$FA,$23,$78,$3C,$5C,$F2,$30,$6F
	.dc.b	$87,$BC,$C8,$C1,$CF,$E6,$3F,$A8,$BF,$E5,$FB,$46,$23,$FC,$E5,$18
	.dc.b	$D3,$FC,$BE,$D8,$E1,$E1,$3E,$B7,$2D,$8C,$C6,$0F,$F1,$7F,$CF,$C8
	.dc.b	$FF,$22,$B5,$FF,$38,$7C,$5B,$7F,$ED,$91,$FE,$52,$6A,$9F,$F3,$AF
	.dc.b	$D6,$B9,$3F,$D1,$17,$C6,$BF,$97,$9F,$AB,$BE,$7C,$D9,$D3,$53,$FE
	.dc.b	$7E,$AC,$57,$FA,$46,$AC,$FF,$04,$E6,$AF,$FA,$11,$BC,$BF,$E2,$EA
	.dc.b	$F3,$C6,$A7,$FB,$44,$7A,$BC,$DC,$4E,$5D,$7C,$34,$9D,$1F,$01,$EF
	.dc.b	$97,$F8,$0E,$3F,$C1,$CD,$BF,$CF,$F8,$3A,$E7,$DF,$24,$BD,$C5,$AF
	.dc.b	$EA,$3F,$5B,$F2,$DB,$B7,$AB,$C0,$F6,$CE,$72,$F8,$AF,$C9,$AE,$3B
	.dc.b	$EA,$F6,$F0,$F9,$19,$97,$F8,$8E,$FF,$D6,$17,$FE,$AA,$53,$A8,$F1
	.dc.b	$FA,$F7,$1E,$EF,$8B,$E2,$FC,$7F,$2F,$71,$E8,$F9,$BC,$99,$18,$3C
	.dc.b	$0F,$AB,$C9,$FB,$B8,$FF,$3C,$CF,$F9,$9D,$7F,$AE,$1B,$1F,$EB,$F6
	.dc.b	$BF,$ED,$1B,$F8,$0D,$D3,$9B,$FE,$C4,$8C,$BF,$D2,$1A,$FF,$B2,$3E
	.dc.b	$47,$F9,$A5,$FF,$B1,$9B,$1F,$EC,$D6,$BF,$ED,$F6,$BF,$D4,$BE,$AB
	.dc.b	$FC,$E1,$3F,$EC,$E6,$C7,$FB,$55,$AF,$FB,$9D,$AF,$F5,$AF,$AD,$7D
	.dc.b	$67,$A2,$11,$8A,$FF,$6C,$B6,$3F,$DC,$0D,$7A,$4F,$1A,$C4,$79,$63
	.dc.b	$07,$5F,$AC,$FF,$79,$13,$87,$07,$87,$EB,$1E,$05,$26,$5F,$EE,$A8
	.dc.b	$CB,$FE,$EC,$3F,$DC,$18,$7F,$03,$F1,$BD,$51,$82,$7D,$5E,$08,$F1
	.dc.b	$7E,$00,$FA,$CA,$48,$FF,$7C,$C6,$5F,$F7,$DF,$D6,$7F,$90,$5B,$F0
	.dc.b	$BF,$26,$31,$87,$F8,$18,$07,$2E,$57,$D6,$71,$25,$76,$7E,$A7,$D5
	.dc.b	$FF,$65,$8E,$95,$D2,$30,$E1,$AF,$85,$AF,$F8,$63,$C6,$7C,$17,$DF
	.dc.b	$6D,$74,$9F,$E5,$13,$93,$F0,$9F,$56,$FA,$7A,$4A,$E1,$19,$79,$7D
	.dc.b	$5B,$E9,$FF,$18,$D8,$CC,$60,$FC,$87,$82,$F1,$A9,$27,$C9,$19,$6F
	.dc.b	$D5,$F0,$23,$83,$D5,$F2,$1F,$E4,$76,$FA,$BF,$9D,$8C,$6B,$F2,$9A
	.dc.b	$FD,$75,$F0,$5B,$25,$74,$6C,$FF,$9D,$A7,$E4,$B6,$3F,$A0,$C6,$3C
	.dc.b	$6F,$E9,$3F,$55,$FE,$B3,$BF,$F6,$13,$FE,$B3,$6C,$FF,$48,$8C,$5F
	.dc.b	$FA,$09,$FC,$E9,$3A,$3F,$6C,$E3,$D4,$A4,$BF,$B5,$1F,$E2,$56,$FA
	.dc.b	$9D,$4E,$D3,$63,$C7,$F5,$5E,$13,$C5,$C7,$63,$27,$48,$F0,$91,$EF
	.dc.b	$E8,$F9,$08,$E0,$F1,$F1,$7A,$08,$E0,$F3,$FC,$0F,$B0,$8E,$0F,$CA
	.dc.b	$7D,$4F,$EA,$9B,$AF,$F1,$78,$C3,$FA,$46,$BF,$B1,$7F,$AD,$37,$8F
	.dc.b	$81,$27,$AF,$A5,$CC,$8E,$0F,$0F,$2F,$72,$38,$25,$7F,$AE,$2F,$FC
	.dc.b	$1D,$1F,$EF,$F7,$F3,$11,$9F,$53,$FC,$0B,$1F,$C0,$A3,$FD,$16,$D7
	.dc.b	$F8,$3B,$53,$4F,$EC,$B3,$D7,$C3,$E7,$DF,$D1,$EB,$E4,$F6,$F3,$6D
	.dc.b	$78,$DC,$3F,$31,$F9,$44,$7E,$7F,$BB,$E6,$3C,$FF,$11,$63,$D9,$1F
	.dc.b	$44,$8F,$02,$65,$F0,$FE,$B3,$83,$63,$F9,$1F,$57,$F8,$B3,$E9,$DF
	.dc.b	$0D,$FE,$28,$5F,$57,$63,$0C,$AF,$E2,$D3,$A6,$C7,$F5,$98,$C5,$FF
	.dc.b	$26,$5F,$55,$D7,$E4,$7B,$36,$FC,$67,$FA,$F2,$7F,$C6,$D5,$FE,$F2
	.dc.b	$6E,$3E,$7A,$E1,$39,$F8,$2F,$41,$E9,$F9,$FB,$87,$49,$FE,$83,$82
	.dc.b	$B3,$1D,$6A,$E1,$F0,$E6,$DF,$8E,$6D,$72,$9A,$AD,$38,$7B,$57,$86
	.dc.b	$75,$5E,$58,$C4,$EA,$7D,$51,$89,$D4,$60,$97,$E5,$9E,$37,$DE,$BC
	.dc.b	$0D,$8E,$46,$8B,$5C,$9F,$9D,$3D,$AB,$A4,$E4,$F7,$4F,$17,$07,$A9
	.dc.b	$E7,$1F,$89,$AF,$04,$EA,$FE,$1D,$68,$EE,$58,$FA,$73,$C2,$BC,$27
	.dc.b	$39,$D4,$F7,$3A,$4E,$E3,$9C,$7D,$97,$75,$F2,$EB,$9C,$7D,$93,$37
	.dc.b	$F1,$9A,$70,$BE,$97,$F0,$0E,$87,$AE,$78,$36,$78,$1C,$9F,$2B,$5F
	.dc.b	$F0,$85,$FD,$D6,$CF,$E9,$49,$3F,$AA,$6C,$7E,$9D,$B7,$FB,$5F,$77
	.dc.b	$F1,$84,$AF,$7B,$6B,$E8,$36,$7E,$0C,$7E,$82,$31,$3F,$22,$7A,$3C
	.dc.b	$89,$3C,$9B,$1C,$A3,$F3,$D3,$C9,$F8,$D1,$C9,$E0,$47,$85,$A7,$84
	.dc.b	$FD,$B9,$E0,$6A,$33,$3F,$66,$BF,$81,$46,$2F,$F8,$A3,$63,$F5,$2D
	.dc.b	$BF,$E0,$CD,$BE,$B3,$EF,$39,$9E,$C3,$13,$F1,$5C,$3E,$C7,$6F,$AD
	.dc.b	$DB,$F0,$9D,$BE,$B7,$71,$EB,$73,$7E,$B9,$EC,$73,$3B,$18,$9E,$0F
	.dc.b	$F1,$FF,$87,$FE,$54,$4A,$FE,$5E,$FF,$2D,$F3,$FF,$9B,$1F,$CD,$DB
	.dc.b	$1F,$C3,$FD,$CF,$F1,$F5,$7F,$95,$9A,$FF,$98,$DA,$FF,$99,$03,$87
	.dc.b	$FC,$CF,$B3,$D7,$FF,$4A,$36,$BF,$CA,$CD,$9F,$F3,$1B,$5F,$F3,$87
	.dc.b	$CE,$7F,$88,$5F,$F3,$3B,$5C,$F8,$7F,$E9,$F8,$C4,$78,$FE,$9D,$F4
	.dc.b	$5E,$18,$FF,$37,$C6,$31,$FF,$33,$2F,$9F,$78,$38,$BE,$17,$FE,$81
	.dc.b	$FB,$17,$CD,$7F,$9B,$61,$18,$7F,$CE,$0D,$7F,$CF,$AD,$7F,$D1,$5F
	.dc.b	$4E,$FA,$8E,$4D,$8F,$F3,$C3,$5F,$F4,$2B,$5F,$F4,$97,$D3,$3E,$07
	.dc.b	$C4,$D8,$FF,$40,$35,$FF,$46,$B5,$FF,$4D,$7D,$3B,$E1,$7D,$4D,$8F
	.dc.b	$F4,$43,$5F,$F4,$AB,$5F,$F5,$17,$D7,$BC,$BF,$84,$D8,$FF,$48,$35
	.dc.b	$FF,$4E,$B5,$F4,$F9,$5E,$82,$3E,$87,$C7,$E2,$48,$F1,$B5,$FF,$53
	.dc.b	$FD,$2F,$D6,$7F,$D3,$CD,$73,$C3,$F5,$63,$11,$FE,$B9,$FA,$6F,$F7
	.dc.b	$22,$69,$3A,$7E,$E7,$D8,$BD,$9A,$4A,$FF,$56,$46,$5B,$EF,$F0,$23
	.dc.b	$83,$FD,$65,$7F,$E1,$86,$F8,$BF,$89,$8C,$47,$FB,$09,$AF,$FB,$3F
	.dc.b	$ED,$FE,$43,$FA,$D6,$BF,$ED,$26,$A6,$C7,$F8,$2A,$3C,$3E,$37,$72
	.dc.b	$3E,$4F,$1B,$C4,$47,$07,$A3,$E4,$BC,$C4,$70,$7B,$3C,$3F,$51,$1C
	.dc.b	$1F,$03,$DB,$F7,$11,$C1,$F1,$3D,$FF,$84,$47,$07,$FB,$67,$BD,$EE
	.dc.b	$7E,$BD,$B1,$98,$C1,$9F,$72,$47,$16,$BF,$C2,$3E,$3B,$F7,$1E,$29
	.dc.b	$A4,$8F,$F7,$47,$D4,$33,$FD,$F3,$63,$31,$83,$F8,$17,$D4,$7F,$55
	.dc.b	$8F,$F7,$83,$5F,$F7,$EB,$5F,$E4,$C1,$C3,$FC,$A7,$EA,$18,$FF,$0D
	.dc.b	$6C,$7F,$BE,$DA,$F2,$0E,$1E,$BD,$0F,$A8,$FD,$11,$23,$F4,$EF,$5F
	.dc.b	$A8,$91,$D5,$AF,$5F,$A8,$7B,$5F,$C7,$DB,$1E,$10,$E1,$E1,$1F,$B1
	.dc.b	$F9,$39,$1F,$B6,$6B,$E9,$F9,$47,$D3,$FF,$2B,$6C,$75,$6B,$B8,$D7
	.dc.b	$24,$8D,$35,$FE,$0B,$ED,$FF,$01,$23,$FC,$19,$FE,$71,$F4,$CF,$A7
	.dc.b	$A4,$9F,$64,$65,$F7,$7D,$3B,$C6,$FE,$80,$D8,$CC,$60,$FE,$89,$F4
	.dc.b	$EF,$1E,$92,$7E,$44,$65,$F9,$5F,$4E,$ED,$FE,$98,$D8,$CC,$60,$FE
	.dc.b	$A1,$F4,$CF,$A7,$A4,$9F,$A9,$19,$7E,$B7,$99,$FC,$80,$9E,$57,$F5
	.dc.b	$97,$F9,$1F,$D3,$BE,$23,$FA,$E3,$6B,$EF,$35,$FC,$83,$5F,$CD,$AF
	.dc.b	$9D,$E2,$48,$E2,$FF,$67,$FA,$5F,$F5,$44,$7E,$25,$AF,$E6,$DA,$FE
	.dc.b	$99,$AF,$F8,$35,$7D,$37,$32,$47,$37,$B7,$8F,$FE,$13,$F8,$16,$7C
	.dc.b	$63,$0D,$AF,$F5,$65,$FE,$F5,$57,$04,$23,$FC,$3A,$F0,$F8,$0E,$6E
	.dc.b	$A3,$2F,$2A,$F4,$7C,$96,$A3,$97,$05,$A4,$D3,$11,$BF,$A5,$FE,$2D
	.dc.b	$5F,$EC,$1F,$A6,$7C,$57,$36,$C7,$FB,$16,$76,$F8,$4F,$4A,$F8,$3E
	.dc.b	$9E,$5C,$9E,$11,$E5,$F8,$EF,$F0,$46,$DF,$06,$5F,$21,$E6,$7F,$DA
	.dc.b	$4D,$7B,$09,$E3,$B7,$F9,$0D,$7D,$17,$5F,$7E,$57,$FB,$6E,$72,$FF
	.dc.b	$B7,$7E,$97,$F8,$EE,$3E,$1C,$7F,$0A,$DF,$53,$E5,$4E,$B2,$FF,$28
	.dc.b	$2F,$A6,$CF,$9F,$31,$E9,$1C,$3C,$8E,$2C,$8F,$04,$74,$DF,$EB,$66
	.dc.b	$E7,$A6,$6B,$CD,$C1,$5F,$83,$97,$D0,$D7,$E4,$3F,$3E,$3D,$9F,$2D
	.dc.b	$F8,$C7,$F9,$C4,$23,$11,$FD,$12,$31,$3F,$CB,$5B,$1C,$3E,$7D,$F4
	.dc.b	$D9,$6C,$66,$30,$7F,$95,$57,$CE,$52,$3F,$E5,$97,$FC,$B5,$F0,$3F
	.dc.b	$9F,$C3,$F0,$13,$FE,$62,$E1,$FD,$0C,$62,$39,$8F,$A6,$FC,$03,$FE
	.dc.b	$81,$AE,$55,$FD,$7B,$96,$CF,$FA,$15,$AF,$EB,$26,$BF,$E5,$B9,$EE
	.dc.b	$7F,$66,$E6,$C9,$C3,$81,$F2,$58,$9D,$D7,$FA,$47,$C2,$E6,$65,$FC
	.dc.b	$A9,$FC,$76,$36,$FE,$F3,$C9,$F0,$9C,$23,$31,$FC,$FD,$DE,$8E,$5F
	.dc.b	$4B,$63,$F4,$7F,$25,$89,$CF,$34,$9E,$31,$6F,$F4,$FF,$4D,$F3,$5B
	.dc.b	$B7,$B7,$C2,$F8,$B3,$9C,$BF,$64,$FC,$5A,$E3,$C1,$AB,$DB,$C3,$E5
	.dc.b	$E6,$5E,$EB,$FE,$F2,$D8,$EB,$E9,$C8,$C1,$B7,$C3,$F2,$F2,$39,$7C
	.dc.b	$55,$CB,$D3,$E9,$F9,$8C,$69,$E6,$7D,$3E,$4F,$AB,$E2,$2E,$1E,$37
	.dc.b	$F0,$38,$FA,$EF,$F8,$73,$F5,$70,$FE,$16,$E6,$BF,$D9,$8E,$67,$F8
	.dc.b	$A7,$C9,$7F,$42,$9F,$B6,$FE,$DD,$BE,$5F,$E3,$A3,$17,$FC,$5B,$E9
	.dc.b	$7F,$83,$57,$E0,$9F,$B5,$19,$F0,$FF,$2B,$18,$C3,$F9,$17,$D3,$3E
	.dc.b	$9B,$F3,$3F,$6D,$FD,$13,$1F,$C5,$CE,$5F,$C6,$C6,$71,$F0,$57,$4F
	.dc.b	$B2,$EA,$DA,$FC,$9F,$D4,$39,$3F,$4A,$B9,$2C,$61,$FD,$4F,$D4,$3C
	.dc.b	$5B,$F4,$58,$6E,$1F,$9E,$3F,$C9,$65,$E0,$FD,$8B,$8D,$3F,$9C,$BF
	.dc.b	$11,$BF,$51,$AF,$A2,$DB,$67,$F5,$2B,$E7,$BF,$76,$FF,$5A,$8C,$C6
	.dc.b	$3D,$5F,$DF,$FD,$47,$E4,$B0,$FE,$8E,$FB,$EB,$F6,$6E,$79,$3F,$A8
	.dc.b	$39,$9D,$FC,$A7,$F0,$8F,$2F,$FB,$DD,$F0,$8C,$3F,$C3,$3E,$A3,$FB
	.dc.b	$2B,$FB,$E5,$19,$F5,$F4,$F3,$F8,$4F,$E2,$7F,$50,$EC,$C6,$1F,$C1
	.dc.b	$A3,$0D,$E3,$DF,$D3,$70,$6C,$7F,$0C,$FA,$87,$1F,$F1,$E5,$F1,$76
	.dc.b	$30,$FF,$22,$F2,$3F,$87,$70,$7F,$6A,$7F,$B5,$5F,$F1,$67,$3C,$5F
	.dc.b	$DC,$DC,$CE,$FC,$3E,$5F,$0B,$CD,$B7,$FC,$7B,$E4,$FE,$A5,$7F,$76
	.dc.b	$7F,$C9,$D1,$9E,$6F,$E6,$71,$89,$D7,$BF,$8E,$2C,$B6,$FA,$45,$9D
	.dc.b	$C6,$0F,$E5,$BF,$51,$F8,$DD,$3F,$C1,$4D,$8F,$F1,$5C,$A3,$FD,$7B
	.dc.b	$1C,$3D,$CF,$E8,$B1,$B9,$FF,$18,$FD,$7F,$F9,$36,$B4,$7F,$3A,$71
	.dc.b	$1F,$E1,$29,$FF,$0F,$4F,$8B,$DB,$E7,$EE,$74,$64,$64,$F3,$C6,$2B
	.dc.b	$97,$2F,$88,$F1,$F2,$DA,$F6,$C7,$9B,$D8,$C5,$78,$5C,$4F,$97,$86
	.dc.b	$CF,$F8,$9B,$34,$7B,$B9,$7D,$FF,$09,$6F,$E4,$B8,$7D,$DC,$97,$0E
	.dc.b	$4F,$F6,$AF,$A9,$FE,$1D,$C3,$FE,$37,$7F,$A4,$57,$CE,$72,$DE,$6E
	.dc.b	$07,$F7,$2C,$FF,$4F,$87,$F8,$EA,$B8,$BF,$43,$B3,$B7,$B3,$DF,$D9
	.dc.b	$B8,$FF,$6A,$BF,$CA,$BD,$E3,$C9,$F0,$3B,$8D,$38,$69,$C6,$31,$87
	.dc.b	$87,$E3,$BD,$9E,$57,$B5,$B5,$DE,$2C,$EE,$30,$78,$FD,$4F,$88,$7C
	.dc.b	$5F,$4E,$C6,$4F,$F0,$0F,$D8,$FF,$0A,$AF,$81,$5F,$E1,$48,$C7,$07
	.dc.b	$CB,$DB,$E7,$1F,$3F,$86,$E1,$F0,$63,$E9,$7D,$56,$31,$F9,$4E,$27
	.dc.b	$FD,$CD,$BA,$3F,$C5,$52,$8C,$4E,$A7,$3F,$52,$FA,$9A,$49,$FF,$78
	.dc.b	$46,$5F,$F7,$97,$1F,$F0,$4C,$FF,$CF,$0D,$8F,$D7,$CE,$F1,$FC,$97
	.dc.b	$17,$F8,$CA,$3F,$C5,$73,$FE,$24,$6C,$FF,$9F,$17,$CF,$EF,$E0,$65
	.dc.b	$78,$9F,$39,$1F,$2F,$C0,$F5,$3B,$7C,$3D,$8E,$D1,$70,$ED,$37,$3F
	.dc.b	$F4,$86,$2A,$C5,$FA,$AA,$DF,$FA,$D2,$B5,$3D,$0F,$54,$EA,$53,$A7
	.dc.b	$D9,$5E,$4A,$E1,$18,$6D,$F8,$AF,$95,$F1,$8F,$2D,$F1,$BE,$04,$8C
	.dc.b	$DF,$88,$E6,$7D,$29,$E7,$7D,$1B,$1A,$AF,$64,$61,$E9,$1C,$4D,$3E
	.dc.b	$C9,$E5,$18,$B4,$62,$78,$9F,$1D,$C5,$F9,$A3,$E8,$D7,$1B,$F0,$4E
	.dc.b	$49,$1E,$19,$D9,$D0,$F9,$B3,$C2,$35,$3C,$27,$D7,$1F,$0D,$E1,$7F
	.dc.b	$2E,$36,$7D,$D3,$C4,$7A,$EF,$CE,$EA,$7D,$4E,$9E,$71,$A7,$B9,$E6
	.dc.b	$AF,$13,$E6,$9F,$CA,$91,$FE,$A6,$70,$2D,$7E,$1D,$B7,$FB,$42,$4F
	.dc.b	$1A,$FF,$73,$74,$FE,$9A,$3E,$AB,$6B,$E3,$36,$7C,$6D,$9F,$64,$A3
	.dc.b	$0F,$AA,$7A,$E5,$F9,$B9,$CC,$FA,$CF,$F7,$8B,$8C,$3E,$64,$0E,$93
	.dc.b	$C4,$97,$E6,$6C,$79,$9B,$7E,$66,$CF,$9A,$FB,$61,$FE,$BA,$6C,$7F
	.dc.b	$AB,$1B,$7E,$4A,$E2,$FA,$E7,$FC,$03,$3A,$BF,$6C,$78,$C9,$3E,$36
	.dc.b	$C7,$8D,$A7,$8E,$BA,$CF,$C3,$9E,$6F,$12,$3C,$5B,$3C,$4E,$6F,$94
	.dc.b	$F9,$15,$E5,$8F,$3D,$76,$C3,$E6,$D7,$CC,$9E,$2F,$85,$D9,$88,$DD
	.dc.b	$73,$9E,$B5,$FC,$C7,$CF,$FE,$7E,$4B,$FE,$7A,$D9,$60,$9F,$F1,$4F
	.dc.b	$CD,$FF,$8A,$27,$F9,$B5,$7F,$BB,$23,$15,$B9,$FC,$16,$FF,$E5,$A7
	.dc.b	$F3,$E7,$F4,$DF,$33,$FA,$9C,$66,$3F,$46,$D7,$FC,$6F,$C0,$FA,$CF
	.dc.b	$EB,$2D,$CB,$FA,$8C,$60,$FE,$BB,$F5,$4F,$AA,$FF,$36,$5F,$82,$31
	.dc.b	$8F,$8A,$10,$F9,$F7,$81,$98,$DC,$FA,$35,$78,$FC,$61,$18,$7A,$35
	.dc.b	$E1,$1B,$8F,$F5,$44,$6B,$D7,$C1,$1F,$27,$D2,$BE,$AB,$B3,$63,$31
	.dc.b	$83,$A7,$CA,$48,$F0,$B5,$F9,$F2,$F0,$79,$92,$3B,$BE,$5F,$2A,$47
	.dc.b	$A9,$AF,$2F,$23,$C6,$47,$CE,$F9,$BC,$1F,$F3,$6F,$37,$E0,$65,$82
	.dc.b	$73,$C9,$FE,$72,$8C,$4F,$F0,$7A,$E1,$0F,$07,$73,$FC,$2F,$E7,$7E
	.dc.b	$AB,$BB,$FF,$3D,$C6,$23,$F2,$F1,$8E,$2E,$30,$FA,$77,$83,$8F,$A4
	.dc.b	$F3,$67,$E7,$6E,$1F,$E8,$66,$DF,$F1,$B5,$F5,$7D,$3E,$9F,$FC,$D7
	.dc.b	$3E,$33,$F9,$03,$87,$C6,$63,$7F,$1E,$4F,$A1,$E5,$ED,$CA,$E4,$2F
	.dc.b	$FD,$92,$E3,$17,$CD,$7A,$7C,$29,$39,$7A,$F3,$FC,$0F,$A7,$FE,$79
	.dc.b	$3D,$1A,$F9,$03,$8D,$FE,$17,$D7,$7C,$5F,$A8,$95,$E1,$7E,$47,$93
	.dc.b	$23,$C8,$D7,$A7,$C9,$3E,$2B,$E7,$7C,$CC,$9F,$53,$F4,$7E,$9E,$47
	.dc.b	$B5,$AF,$97,$8D,$F2,$5F,$5F,$DF,$93,$F0,$DF,$B5,$EF,$C8,$F8,$CD
	.dc.b	$7D,$DF,$5C,$DB,$F0,$7E,$43,$E0,$69,$2B,$E8,$C3,$E0,$76,$48,$CB
	.dc.b	$CB,$E2,$BF,$1B,$E4,$BE,$5E,$92,$BE,$DC,$3E,$07,$F2,$7F,$52,$E1
	.dc.b	$A4,$AF,$EB,$5A,$7E,$63,$EA,$7F,$CA,$51,$F8,$46,$BF,$65,$BD,$1F
	.dc.b	$AF,$5F,$2B,$FA,$0F,$A9,$F5,$57,$E3,$1A,$FE,$05,$B5,$FB,$75,$AF
	.dc.b	$EA,$3C,$8F,$CB,$57,$E5,$1A,$FE,$25,$B5,$FE,$12,$EA,$F0,$7C,$4C
	.dc.b	$AF,$03,$E3,$F9,$3F,$D9,$FD,$4B,$E1,$29,$2B,$F4,$30,$F8,$4F,$17
	.dc.b	$C9,$C9,$F1,$3F,$BC,$F7,$BD,$F1,$FA,$8F,$9B,$7D,$4F,$F0,$06,$C6
	.dc.b	$63,$1C,$79,$ED,$93,$C5,$FE,$0F,$F5,$1F,$D2,$23,$DD,$F1,$1F,$0C
	.dc.b	$8F,$F8,$3F,$9F,$7F,$53,$23,$F7,$34,$8E,$7F,$2D,$FC,$4B,$EA,$5F
	.dc.b	$19,$49,$7F,$C0,$23,$F8,$AB,$FC,$0B,$EA,$7F,$92,$BF,$B7,$6F,$3F
	.dc.b	$F2,$D6,$FA,$DF,$E1,$CF,$2F,$F8,$FF,$D4,$BC,$8A,$4C,$3F,$86,$C3
	.dc.b	$83,$F9,$2F,$A2,$F7,$E9,$2B,$F8,$9C,$3D,$EF,$A1,$94,$9F,$F1,$83
	.dc.b	$F4,$F1,$FE,$5F,$F5,$2F,$90,$A4,$AF,$E3,$B1,$F0,$9B,$F4,$1C,$08
	.dc.b	$EB,$DB,$FE,$73,$F5,$3F,$E8,$C8,$FE,$4A,$D7,$F8,$D3,$78,$7F,$A8
	.dc.b	$3F,$A4,$CB,$FC,$A5,$28,$FC,$3B,$C3,$C7,$EA,$76,$F4,$FB,$F4,$D9
	.dc.b	$EF,$18,$7C,$3F,$2F,$E4,$E4,$91,$83,$CB,$BF,$A2,$64,$60,$FE,$7B
	.dc.b	$F5,$1F,$C1,$DF,$E7,$FF,$53,$FD,$7D,$BF,$07,$FD,$62,$75,$1F,$CD
	.dc.b	$D7,$CC,$67,$D2,$93,$D2,$3F,$AF,$FC,$1B,$E9,$7F,$B0,$FC,$CB,$3A
	.dc.b	$6A,$37,$F4,$D7,$0F,$EA,$37,$C5,$B1,$FD,$AB,$EA,$7F,$83,$F2,$7C
	.dc.b	$CC,$7E,$83,$7D,$0D,$7B,$1B,$F4,$EE,$3F,$D7,$30,$E2,$FF,$A1,$5F
	.dc.b	$E1,$EE,$7C,$BE,$0E,$E7,$C1,$1E,$1D,$FD,$7C,$B2,$BA,$47,$97,$83
	.dc.b	$51,$83,$C0,$FA,$7E,$37,$FC,$B7,$1F,$E5,$2C,$7F,$C7,$7D,$5F,$E4
	.dc.b	$36,$C6,$67,$6F,$FA,$DA,$1F,$1A,$FA,$9E,$0D,$9E,$8D,$78,$35,$FF
	.dc.b	$30,$FD,$0F,$F9,$25,$34,$8F,$F9,$57,$E3,$3F,$CD,$58,$FF,$8A,$E3
	.dc.b	$FC,$AB,$C3,$FE,$06,$7F,$D5,$D5,$FE,$AF,$FA,$9F,$E1,$31,$FE,$B1
	.dc.b	$BE,$35,$FE,$CE,$E6,$B3,$FA,$46,$BB,$8E,$91,$88,$72,$D3,$7C,$AF
	.dc.b	$9A,$C5,$F8,$FC,$4E,$06,$2F,$FC,$1C,$7F,$03,$8F,$47,$99,$D4,$DB
	.dc.b	$83,$FC,$43,$AB,$97,$CA,$D8,$FF,$12,$E3,$FE,$A2,$8C,$5F,$F8,$6D
	.dc.b	$FE,$E5,$3C,$B9,$38,$DF,$27,$DF,$F5,$3F,$2A,$FF,$C4,$4D,$BF,$F2
	.dc.b	$C6,$0F,$03,$7B,$F0,$DE,$67,$FE,$36,$C3,$83,$FC,$39,$FD,$83,$9F
	.dc.b	$5B,$B9,$E0,$F6,$2C,$71,$F5,$FC,$66,$34,$F1,$BE,$6F,$92,$93,$DF
	.dc.b	$E1,$6D,$7A,$3D,$6F,$59,$8D,$FF,$CE,$19,$A3,$DF,$E5,$FC,$0F,$2E
	.dc.b	$E3,$E9,$7F,$03,$97,$F5,$48,$FF,$28,$B7,$1F,$F2,$33,$FE,$68,$8C
	.dc.b	$F2,$70,$CE,$47,$09,$B5,$C3,$E3,$BF,$1B,$F8,$07,$FB,$26,$BA,$C6
	.dc.b	$1F,$C8,$FD,$47,$CA,$A0,$7E,$5B,$FE,$57,$74,$E7,$8B,$8F,$0F,$2E
	.dc.b	$2B,$7F,$E6,$AB,$FF,$05,$37,$5E,$E7,$83,$E5,$2C,$7F,$9C,$3E,$9F
	.dc.b	$FC,$6C,$DE,$7F,$D2,$C6,$33,$FF,$34,$28,$CF,$C6,$7E,$AB,$8F,$D3
	.dc.b	$7C,$E3,$17,$9F,$93,$93,$FE,$7E,$AF,$F5,$F3,$63,$F6,$7E,$DF,$F7
	.dc.b	$98,$FF,$41,$B6,$BF,$CF,$0F,$F7,$98,$CF,$A1,$FB,$AF,$B7,$FD,$0D
	.dc.b	$7F,$9F,$5B,$D3,$FE,$B3,$7F,$A4,$D7,$FA,$39,$CB,$FD,$99,$FF,$0C
	.dc.b	$39,$BE,$9E,$EF,$57,$F8,$3F,$DB,$FD,$BC,$3F,$D1,$4F,$FA,$92,$BF
	.dc.b	$D3,$2E,$70,$FF,$61,$39,$9D,$FC,$0F,$2F,$81,$E6,$DB,$E3,$F3,$DD
	.dc.b	$8E,$FE,$0D,$8F,$F5,$2F,$D3,$3E,$1B,$F8,$CB,$63,$3C,$58,$9F,$E3
	.dc.b	$BF,$6F,$F9,$9C,$FF,$D4,$0D,$D7,$FD,$AC,$FF,$AC,$EF,$FD,$62,$E5
	.dc.b	$FF,$04,$BF,$EE,$97,$3C,$7D,$3B,$BA,$BF,$CA,$BE,$9F,$FD,$19,$3F
	.dc.b	$EB,$A6,$E1,$FE,$AF,$FB,$67,$D2,$EF,$E4,$F8,$7A,$F3,$0E,$11,$6F
	.dc.b	$FD,$6D,$87,$FA,$4E,$7F,$D8,$BF,$44,$F1,$F5,$B7,$FB,$BE,$31,$3E
	.dc.b	$1F,$5B,$C9,$EB,$79,$5B,$1E,$58,$B8,$F6,$9B,$7F,$EF,$9C,$7F,$40
	.dc.b	$DF,$92,$F7,$7C,$27,$BD,$98,$FB,$FC,$0B,$8F,$A1,$FE,$9B,$F6,$CF
	.dc.b	$A4,$F8,$DF,$53,$86,$DF,$C7,$8F,$95,$F1,$F8,$C3,$E0,$B8,$9F,$83
	.dc.b	$E3,$59,$FF,$6D,$70,$23,$FA,$DF,$D4,$7E,$CA,$7F,$DB,$CF,$F0,$83
	.dc.b	$1B,$2F,$FC,$8B,$19,$F4,$FF,$B1,$E0,$F8,$4E,$71,$87,$AF,$AF,$DB
	.dc.b	$D3,$B8,$7F,$BE,$30,$FE,$51,$8F,$5C,$BF,$8D,$C6,$27,$BF,$C1,$71
	.dc.b	$78,$3F,$1B,$C4,$F4,$FA,$16,$7A,$38,$7B,$7A,$17,$0F,$6B,$F0,$09
	.dc.b	$1C,$0D,$9E,$5F,$03,$E2,$F8,$1F,$19,$B1,$E9,$8B,$5F,$22,$6E,$7E
	.dc.b	$B8,$67,$F1,$6F,$E0,$6D,$F4,$38,$7E,$8B,$23,$E2,$77,$DB,$FA,$6F
	.dc.b	$F8,$43,$E9,$FF,$82,$63,$D5,$B5,$FE,$78,$FA,$9F,$E3,$D6,$AD,$7D
	.dc.b	$1F,$13,$97,$45,$BF,$E9,$33,$CE,$33,$1F,$E8,$2A,$FF,$41,$E1,$F9
	.dc.b	$38,$C4,$7F,$50,$8C,$38,$6B,$C5,$7D,$0F,$F8,$CE,$38,$35,$46,$A7
	.dc.b	$AC,$62,$75,$3F,$E2,$99,$F3,$9E,$4F,$77,$E2,$4F,$C5,$7E,$37,$C3
	.dc.b	$49,$DD,$F9,$FE,$7F,$0F,$C5,$8F,$6C,$F6,$23,$E2,$C1,$1E,$2F,$8B
	.dc.b	$FA,$47,$E8,$39,$6D,$76,$BF,$5C,$E1,$FD,$7C,$DD,$3E,$C4,$DA,$FF
	.dc.b	$57,$F5,$FF,$BF,$5E,$58,$A7,$58,$7C,$F7,$C1,$3A,$73,$1E,$08,$FE
	.dc.b	$3F,$86,$2B,$94,$6C,$C9,$67,$E1,$1D,$5C,$61,$FC,$96,$B9,$5F,$6C
	.dc.b	$3C,$D7,$C0,$F0,$17,$0E,$ED,$9F,$3B,$70,$EA,$7E,$32,$33,$87,$C4
	.dc.b	$8C,$CF,$8C,$F9,$2E,$1E,$F1,$B7,$27,$52,$39,$3D,$51,$C6,$BC,$B1
	.dc.b	$98,$FA,$38,$7B,$9E,$D8,$7B,$A7,$57,$E3,$6A,$8D,$4F,$AC,$ED,$87
	.dc.b	$B1,$C4,$F9,$1B,$5E,$48,$C1,$D6,$32,$79,$0E,$67,$92,$FC,$D1,$C0
	.dc.b	$F1,$38,$9F,$2D,$F9,$27,$A1,$0F,$93,$5D,$5E,$58,$7C,$48,$F0,$C7
	.dc.b	$F9,$32,$79,$3C,$0F,$A8,$7E,$F7,$0F,$D2,$4F,$E9,$DB,$1F,$C0,$DA
	.dc.b	$EE,$34,$7E,$7B,$CE,$FC,$B1,$1F,$AC,$D8,$F2,$B7,$0F,$A3,$82,$31
	.dc.b	$5E,$8A,$FE,$17,$18,$7E,$04,$EE,$7A,$92,$3A,$B6,$BA,$D7,$19,$F8
	.dc.b	$07,$49,$D1,$DE,$BA,$47,$22,$57,$29,$DC,$78,$AB,$A5,$7C,$F8,$FE
	.dc.b	$55,$18,$BF,$E7,$CD,$AF,$8C,$72,$9F,$1D,$78,$2F,$B4,$78,$27,$C9
	.dc.b	$5C,$63,$F6,$8D,$9F,$E9,$0D,$9F,$E7,$AD,$9F,$1C,$77,$AF,$34,$7C
	.dc.b	$13,$B1,$FE,$90,$F3,$9D,$FF,$CE,$DB,$5F,$CE,$9B,$3F,$D2,$1A,$FF
	.dc.b	$45,$F9,$1F,$E7,$A4,$9F,$E6,$0F,$F4,$5F,$9C,$FF,$38,$CF,$F5,$06
	.dc.b	$BF,$D5,$5A,$FF,$A8,$83,$87,$FD,$4B,$F4,$AF,$A7,$7D,$3F,$F4,$E8
	.dc.b	$C3,$7E,$81,$E2,$66,$B7,$7F,$D8,$3E,$C5,$F6,$5F,$D5,$A1,$18,$7F
	.dc.b	$AC,$B5,$E1,$1B,$FA,$97,$D3,$D2,$4F,$F6,$08,$CB,$FD,$8B,$47,$CD
	.dc.b	$77,$6C,$66,$30,$67,$CC,$91,$C5,$AF,$F7,$8A,$78,$9E,$32,$47,$85
	.dc.b	$F3,$7C,$14,$8F,$23,$5E,$5E,$27,$CF,$E8,$FF,$0F,$5F,$F9,$2E,$7F
	.dc.b	$B1,$35,$FE,$8C,$FF,$91,$5F,$EC,$DF,$51,$FE,$25,$9F,$F1,$37,$D3
	.dc.b	$7D,$59,$FE,$D4,$D7,$FA,$74,$EF,$9B,$EF,$36,$3F,$DC,$2D,$4D,$1E
	.dc.b	$1F,$9E,$7F,$0D,$F5,$2C,$F1,$7F,$EE,$A8,$C3,$FD,$ED,$AF,$F5,$FF
	.dc.b	$A8,$7C,$2F,$08,$4B,$E2,$70,$D8,$E5,$1B,$C9,$F3,$D7,$C7,$95,$F1
	.dc.b	$1E,$98,$FE,$0A,$3F,$B7,$7D,$4B,$D4,$F0,$36,$3F,$C2,$0D,$7E,$3F
	.dc.b	$C1,$F7,$F2,$64,$7D,$07,$C5,$E4,$F6,$FA,$A9,$3F,$E1,$9C,$51,$E7
	.dc.b	$F9,$99,$3F,$61,$AF,$D4,$F2,$7D,$84,$7E,$EB,$EA,$F8,$FF,$81,$F1
	.dc.b	$72,$7F,$02,$FC,$3F,$1E,$47,$E1,$1A,$FF,$83,$7A,$1A,$FC,$FF,$15
	.dc.b	$F0,$34,$95,$FE,$31,$87,$C1,$6C,$91,$97,$97,$CD,$7D,$9F,$0D,$F1
	.dc.b	$74,$95,$CF,$E0,$BC,$04,$70,$77,$F8,$3E,$5F,$17,$23,$FC,$85,$68
	.dc.b	$ED,$E8,$7E,$1F,$EE,$59,$52,$5F,$F9,$26,$1F,$2B,$BF,$95,$93,$97
	.dc.b	$96,$7E,$9E,$CF,$57,$D3,$CA,$F0,$C7,$7F,$33,$C1,$F0,$F2,$7B,$47
	.dc.b	$8F,$E1,$FC,$9E,$74,$9F,$14,$7C,$5F,$8C,$91,$EC,$7E,$47,$C6,$FA
	.dc.b	$F8,$24,$FA,$5F,$76,$5E,$FC,$E4,$FB,$5F,$85,$7F,$A9,$F3,$FF,$D1
	.dc.b	$DE,$77,$FA,$9D,$FE,$63,$5F,$E6,$B6,$BF,$E4,$F7,$7E,$6F,$EE,$BD
	.dc.b	$5F,$D7,$FC,$9C,$AF,$DC,$47,$EC,$BE,$5F,$F6,$7F,$15,$27,$F7,$91
	.dc.b	$E1,$A9,$1D,$DF,$24,$7E,$EB,$D0,$FD,$DF,$AB,$2B,$F8,$24,$7E,$FB
	.dc.b	$DF,$FD,$FF,$C3,$C9,$FE,$11,$1E,$DC,$64,$7A,$DF,$81,$F1,$5F,$C1
	.dc.b	$BC,$89,$3F,$C3,$DF,$E1,$3E,$C7,$F0,$AF,$88,$93,$FC,$4D,$FE,$1B
	.dc.b	$F1,$BF,$27,$E2,$FE,$56,$92,$BE,$44,$7F,$15,$FA,$1F,$F4,$CC,$7F
	.dc.b	$A5,$5A,$FF,$9E,$5B,$C9,$FE,$D7,$5F,$59,$C7,$E8,$A5,$71,$7A,$7A
	.dc.b	$3D,$7D,$1B,$3D,$63,$0F,$7F,$A0,$F1,$7C,$2C,$8C,$1E,$3F,$4F,$CD
	.dc.b	$32,$30,$7F,$AA,$FE,$85,$E3,$FF,$2F,$FA,$2F,$E9,$53,$FE,$91,$6E
	.dc.b	$DF,$EE,$FA,$DC,$F1,$F2,$3F,$9B,$46,$23,$FD,$30,$D9,$E1,$3F,$CE
	.dc.b	$F8,$1D,$9F,$CF,$7C,$66,$74,$D4,$6F,$B6,$E5,$FE,$C1,$BE,$2D,$8F
	.dc.b	$E8,$DF,$45,$FC,$52,$3F,$D8,$D8,$7D,$0F,$03,$FC,$30,$D8,$09,$E0
	.dc.b	$FF,$AB,$DF,$E1,$2F,$FB,$3D,$CE,$6A,$FF,$11,$46,$1B,$E6,$78,$BC
	.dc.b	$BF,$1B,$2F,$94,$79,$B8,$3E,$07,$1C,$9E,$D1,$EA,$DF,$51,$83,$C4
	.dc.b	$FB,$BC,$F6,$7F,$EF,$AC,$D1,$8A,$FF,$1B,$1E,$E8,$CC,$EE,$75,$3A
	.dc.b	$97,$D0,$BE,$8B,$A3,$67,$99,$F1,$DC,$3C,$5A,$EB,$E8,$BF,$C7,$09
	.dc.b	$A4,$79,$9F,$E4,$2D,$BF,$DF,$8D,$3F,$DC,$07,$D6,$8F,$AF,$F4,$4E
	.dc.b	$2F,$6F,$8F,$F4,$EB,$7C,$97,$5C,$47,$DE,$8C,$63,$C0,$EE,$E2,$3F
	.dc.b	$CB,$9C,$95,$FC,$2B,$79,$1F,$21,$99,$4D,$F1,$66,$1E,$A8,$B3,$EA
	.dc.b	$F1,$78,$18,$CB,$CB,$08,$F0,$47,$F5,$4A,$FC,$AB,$FE,$39,$E6,$F1
	.dc.b	$77,$E9,$DF,$34,$AD,$9F,$E4,$78,$C6,$1F,$86,$8F,$F3,$C7,$D0,$FC
	.dc.b	$DC,$3E,$2B,$63,$F4,$BD,$DF,$06,$31,$5F,$A6,$7F,$6F,$3D,$A3,$1B
	.dc.b	$EF,$83,$83,$C7,$C8,$95,$BA,$5F,$FB,$78,$4F,$7F,$A2,$95,$B9,$CB
	.dc.b	$63,$C9,$E3,$C8,$D7,$91,$70,$F7,$1E,$7F,$23,$D4,$67,$5F,$7B,$ED
	.dc.b	$F2,$3D,$DF,$11,$70,$F0,$3F,$81,$C7,$EA,$47,$A5,$B8,$76,$3F,$CF
	.dc.b	$BD,$1B,$E4,$E1,$9C,$BE,$17,$6F,$83,$FE,$59,$8F,$89,$F4,$5F,$C2
	.dc.b	$DE,$DF,$11,$F9,$18,$C5,$77,$FA,$2F,$A6,$9F,$A6,$E7,$8F,$F3,$11
	.dc.b	$8D,$FF,$42,$BC,$DC,$F1,$E9,$B9,$7D,$59,$FF,$3B,$37,$7E,$7E,$7F
	.dc.b	$47,$F4,$9F,$44,$F5,$AF,$7E,$1B,$7F,$7E,$5E,$B7,$EA,$F9,$BF,$97
	.dc.b	$CE,$63,$1B,$E7,$D1,$93,$F8,$7A,$FF,$5C,$B6,$3F,$69,$ED,$3C,$FA
	.dc.b	$63,$E0,$AD,$7E,$3F,$E8,$BF,$AD,$B7,$C1,$FD,$E4,$7F,$B7,$B9,$3F
	.dc.b	$82,$BA,$CF,$EE,$3F,$E5,$6A,$E5,$8F,$0F,$7B,$8B,$FC,$17,$E8,$BE
	.dc.b	$AE,$1F,$C3,$9F,$F5,$0D,$7E,$7D,$CB,$FE,$E7,$73,$87,$E2,$57,$CE
	.dc.b	$F3,$F8,$4E,$8D,$AE,$5F,$2F,$DC,$F0,$78,$B6,$3F,$59,$F4,$4F,$73
	.dc.b	$F8,$C3,$63,$3C,$98,$9F,$E3,$9F,$45,$FC,$A6,$7F,$68,$DD,$3F,$38
	.dc.b	$FF,$AC,$23,$3E,$27,$0F,$0B,$8F,$3D,$9E,$2F,$F2,$7F,$A1,$FF,$43
	.dc.b	$C7,$EF,$DB,$87,$E9,$FE,$89,$F3,$FB,$F8,$AE,$1F,$13,$30,$E1,$16
	.dc.b	$FF,$5D,$87,$FA,$3A,$7F,$85,$7C,$EB,$CB,$D7,$3F,$F3,$E8,$C4,$F8
	.dc.b	$7D,$AF,$27,$B5,$E5,$6C,$79,$62,$E3,$DA,$6D,$FF,$48,$C7,$F3,$D7
	.dc.b	$EB,$F0,$BD,$DF,$11,$EF,$64,$7B,$FC,$4B,$8F,$A1,$FE,$95,$F4,$4F
	.dc.b	$9F,$F8,$DF,$13,$86,$DF,$C7,$F8,$9F,$94,$63,$0F,$82,$E2,$3E,$0F
	.dc.b	$95,$67,$F8,$97,$02,$3F,$AC,$7D,$17,$EB,$A7,$F8,$B3,$FC,$20,$C6
	.dc.b	$FF,$D9,$1F,$E1,$74,$FA,$2F,$EC,$0D,$C3,$F9,$C4,$EA,$7A,$FA,$92
	.dc.b	$73,$ED,$DC,$7B,$CF,$8B,$B3,$19,$77,$7E,$1F,$87,$E6,$F4,$2E,$3D
	.dc.b	$9F,$4F,$A3,$23,$07,$89,$C6,$3F,$E1,$2D,$3F,$A2,$B6,$3F,$B5,$BF
	.dc.b	$E3,$C8,$CB,$FE,$22,$5D,$FF,$1B,$BF,$E3,$B6,$FD,$F1,$CA,$94,$8C
	.dc.b	$36,$7F,$C3,$2D,$9F,$26,$DA,$24,$E9,$E3,$DD,$23,$8B,$67,$FC,$5B
	.dc.b	$F4,$3F,$C1,$9F,$F1,$CB,$5F,$F2,$B8,$70,$FF,$96,$9A,$E7,$93,$D5
	.dc.b	$1F,$91,$AF,$F1,$7B,$67,$BC,$FB,$F8,$BF,$81,$70,$7E,$AF,$83,$FC
	.dc.b	$45,$7B,$8C,$61,$FD,$EE,$3F,$C4,$6D,$F7,$77,$F2,$FC,$19,$1E,$07
	.dc.b	$C6,$47,$C7,$F2,$BD,$BD,$EE,$ED,$8E,$D1,$EA,$23,$EA,$7D,$BE,$57
	.dc.b	$98,$F3,$FC,$55,$9F,$34,$5D,$BF,$BC,$64,$AC,$67,$FE,$46,$85,$62
	.dc.b	$FE,$7C,$6D,$E6,$4B,$E7,$3E,$08,$C5,$73,$8E,$53,$F5,$6E,$BF,$66
	.dc.b	$2D,$F5,$8E,$77,$C8,$F0,$CF,$58,$C9,$2F,$B4,$FD,$3B,$E7,$5D,$1B
	.dc.b	$1C,$EB,$C4,$D7,$C4,$DB,$F1,$5F,$49,$ED,$3E,$23,$B1,$BB,$EC,$D3
	.dc.b	$95,$F8,$27,$89,$E0,$29,$E0,$2D,$F8,$27,$B4,$62,$BE,$05,$F8,$C9
	.dc.b	$3A,$7E,$C5,$F7,$73,$5D,$DA,$F7,$6D,$F7,$8F,$D4,$5F,$1B,$FB,$F7
	.dc.b	$67,$31,$F0,$1F,$23,$EE,$3E,$03,$EA,$8F,$80,$FE,$E2,$7F,$2B,$3F
	.dc.b	$B0,$6C,$7E,$F9,$AE,$E3,$47,$F9,$9F,$E4,$7F,$B4,$E1,$EA,$6C,$7C
	.dc.b	$13,$F8,$1B,$89,$F3,$9F,$E5,$57,$0F,$C1,$8D,$F8,$BF,$92,$9C,$E3
	.dc.b	$FA,$09,$F0,$4E,$48,$FE,$1E,$77,$87,$FA,$95,$B1,$C0,$D4,$F0,$BF
	.dc.b	$BF,$3A,$7C,$91,$9C,$3D,$8D,$AF,$64,$FF,$BD,$2B,$85,$7F,$AE,$9B
	.dc.b	$3E,$A3,$B9,$27,$BD,$73,$7E,$1D,$71,$8C,$47,$C3,$3B,$38,$C3,$E1
	.dc.b	$47,$9A,$BA,$57,$99,$A7,$9A,$17,$E4,$79,$C6,$9F,$0B,$E7,$7C,$AE
	.dc.b	$5F,$11,$CA,$FC,$58,$7C,$D6,$CF,$CD,$9E,$24,$BF,$23,$FE,$93,$E0
	.dc.b	$7D,$07,$F5,$96,$D7,$FB,$B9,$B3,$FE,$FA,$6B,$FD,$30,$FF,$05,$7C
	.dc.b	$7F,$F6,$1E,$DF,$F3,$8C,$7F,$C0,$06,$BB,$6B,$FD,$53,$E8,$5F,$43
	.dc.b	$E2,$8E,$82,$33,$8F,$FC,$00,$5F,$2A,$F7,$B1,$7C,$6F,$C9,$BB,$CA
	.dc.b	$E7,$1F,$85,$6B,$F4,$9B,$1B,$8D,$FC,$FB,$4D,$B6,$33,$18,$39,$FC
	.dc.b	$CB,$E1,$A9,$27,$C3,$19,$7C,$7F,$01,$FE,$31,$6E,$FE,$48,$C4,$7C
	.dc.b	$26,$BF,$3F,$07,$C2,$79,$09,$1E,$87,$DB,$F6,$6F,$26,$92,$7E,$1C
	.dc.b	$65,$F7,$E1,$EE,$23,$83,$E1,$F8,$3F,$08,$8E,$0F,$9B,$F3,$DF,$E2
	.dc.b	$14,$D2,$73,$7D,$05,$F4,$14,$93,$C9,$F9,$7F,$07,$23,$E4,$B5,$FC
	.dc.b	$2C,$BE,$7F,$A1,$23,$C6,$FD,$4E,$19,$1F,$49,$AF,$2F,$77,$C8,$47
	.dc.b	$D6,$F9,$FE,$7B,$FC,$53,$D8,$CF,$FC,$57,$8A,$73,$18,$6D,$8F,$87
	.dc.b	$E1,$61,$72,$3E,$7B,$0D,$8E,$18,$F7,$79,$46,$54,$65,$F9,$FF,$29
	.dc.b	$D7,$2F,$F0,$B4,$72,$CB,$CF,$39,$35,$E1,$C9,$D1,$E7,$AE,$2F,$6A
	.dc.b	$F7,$CF,$9B,$97,$B7,$C3,$CA,$E2,$7E,$01,$F1,$7C,$3F,$BF,$E7,$9C
	.dc.b	$18,$AD,$E4,$AD,$4F,$C5,$F0,$1E,$ED,$F1,$66,$1D,$A7,$2D,$9E,$04
	.dc.b	$75,$F2,$DF,$3F,$C0,$7C,$3D,$24,$F3,$F7,$BC,$1E,$1F,$85,$91,$DE
	.dc.b	$39,$7B,$3C,$FE,$1E,$4F,$F6,$F8,$FB,$9F,$3C,$CA,$92,$7E,$2C,$3E
	.dc.b	$47,$7F,$17,$27,$2F,$2C,$BD,$1E,$2F,$A7,$E7,$65,$7F,$7D,$8E,$FF
	.dc.b	$3D,$E0,$F8,$39,$3D,$A3,$C7,$F0,$7E,$4F,$83,$93,$E2,$8F,$89,$F1
	.dc.b	$12,$3F,$C1,$0F,$C7,$F8,$9F,$5F,$04,$9F,$4B,$EE,$CB,$DF,$9C,$9F
	.dc.b	$6B,$F0,$AF,$F3,$DE,$5F,$F4,$5F,$0B,$F2,$CF,$F3,$7A,$FC,$3B,$5F
	.dc.b	$C7,$1F,$A7,$78,$78,$3F,$E9,$8F,$A0,$D5,$7E,$7D,$FE,$55,$5C,$03
	.dc.b	$88,$D7,$A5,$FA,$F8,$7C,$AD,$F8,$69,$7F,$98,$9C,$BE,$4F,$4F,$D2
	.dc.b	$47,$CF,$5C,$FC,$EE,$9E,$6C,$9E,$51,$DF,$E1,$3C,$1F,$09,$27,$B4
	.dc.b	$7C,$4C,$A4,$7C,$27,$E3,$FC,$37,$9F,$CA,$93,$E5,$7D,$5F,$0B,$EB
	.dc.b	$F8,$59,$3E,$97,$DD,$F1,$3F,$43,$E0,$FE,$8E,$92,$BE,$7C,$7F,$11
	.dc.b	$F9,$F7,$A9,$49,$3F,$BA,$87,$C1,$6F,$E7,$E4,$E5,$E5,$EA,$73,$F5
	.dc.b	$2C,$F1,$8C,$3C,$FE,$77,$C1,$F2,$12,$30,$77,$F5,$BC,$93,$23,$07
	.dc.b	$F0,$CF,$9E,$79,$3F,$CA,$7E,$7D,$EF,$53,$F9,$64,$EA,$B8,$79,$78
	.dc.b	$B9,$18,$3F,$8B,$F0,$FE,$1F,$C1,$FF,$72,$F9,$2C,$FF,$9B,$36,$B7
	.dc.b	$DF,$FC,$E6,$FF,$0F,$1F,$91,$79,$3F,$CF,$DA,$9D,$F8,$5F,$CD,$FE
	.dc.b	$7F,$F8,$CC,$FE,$66,$FE,$9B,$FC,$A9,$AF,$E7,$63,$73,$C7,$CF,$FE
	.dc.b	$95,$18,$CB,$FD,$FA,$DA,$FE,$A0,$FF,$13,$7F,$9A,$BF,$C4,$5B,$3C
	.dc.b	$8E,$7E,$77,$4D,$3E,$0F,$1C,$AE,$31,$E0,$DB,$51,$83,$AB,$E4,$F4
	.dc.b	$7F,$AA,$C7,$F8,$EB,$1F,$F1,$96,$7F,$E3,$26,$C6,$67,$6F,$F9,$2E
	.dc.b	$1F,$3C,$F9,$FE,$0D,$9E,$8D,$78,$35,$FF,$24,$FC,$FF,$F8,$F9,$34
	.dc.b	$8F,$F9,$13,$C9,$FE,$DD,$8F,$F6,$B8,$FF,$26,$6F,$FD,$71,$FE,$F5
	.dc.b	$5F,$DE,$FE,$7F,$EE,$47,$F7,$D3,$67,$F1,$33,$FC,$85,$3C,$78,$FA
	.dc.b	$BF,$E6,$7C,$2C,$F1,$9A,$FF,$81,$FC,$9A,$FF,$18,$AF,$C2,$7C,$6E
	.dc.b	$27,$3E,$3C,$9F,$34,$5A,$F1,$78,$FC,$8C,$69,$ED,$7B,$BF,$D4,$EB
	.dc.b	$FB,$CB,$67,$F2,$9E,$06,$B9,$77,$C9,$2F,$71,$6B,$F9,$97,$CF,$FC
	.dc.b	$2C,$3F,$C2,$8D,$AF,$D6,$F7,$FB,$A3,$15,$FE,$25,$7E,$3B,$CB,$6D
	.dc.b	$6F,$B7,$87,$95,$30,$FF,$07,$DB,$CA,$C6,$1E,$2F,$9F,$EC,$63,$3F
	.dc.b	$1B,$E1,$F2,$A6,$1C,$BC,$BB,$7E,$AF,$2F,$CC,$63,$0F,$5B,$E9,$F8
	.dc.b	$99,$7E,$2F,$32,$DF,$F5,$3C,$5E,$67,$DC,$7F,$D4,$78,$FF,$8D,$DC
	.dc.b	$DF,$F8,$A1,$CB,$F8,$B5,$8F,$1C,$39,$36,$7F,$C5,$FF,$3D,$FE,$6A
	.dc.b	$6F,$C4,$7E,$2A,$1F,$1D,$63,$F1,$BF,$3D,$F8,$BD,$3F,$B4,$37,$3F
	.dc.b	$F2,$52,$EB,$DF,$5F,$0E,$19,$9F,$F8,$FA,$BF,$CB,$0D,$8E,$BE,$57
	.dc.b	$6F,$94,$B3,$B7,$F3,$BF,$3D,$FE,$32,$8F,$F2,$3B,$73,$FE,$F4,$BD
	.dc.b	$FD,$FC,$07,$E9,$38,$7F,$0F,$5F,$13,$AF,$1D,$79,$75,$D3,$BE,$BB
	.dc.b	$9F,$F9,$53,$04,$7E,$BB,$E7,$BF,$DB,$7C,$7F,$E6,$95,$F4,$39,$F9
	.dc.b	$3D,$3B,$F2,$B8,$79,$17,$2F,$F2,$FD,$FF,$16,$6F,$BD,$D4,$ED,$EF
	.dc.b	$58,$FF,$32,$FC,$FF,$D4,$6F,$67,$F0,$08,$FF,$6D,$C6,$0D,$FC,$7E
	.dc.b	$4D,$7B,$76,$39,$FA,$BC,$7D,$8E,$4D,$8F,$F3,$9F,$CF,$7F,$A4,$5B
	.dc.b	$CB,$FC,$3A,$31,$CF,$FE,$29,$7E,$69,$F9,$27,$FD,$30,$FE,$B6,$79
	.dc.b	$FB,$BC,$79,$F9,$78,$92,$7B,$F8,$F7,$0F,$F1,$A7,$CF,$3E,$83,$F8
	.dc.b	$EF,$CF,$BD,$7F,$E3,$F3,$9C,$3F,$D1,$33,$96,$CF,$0E,$8E,$2D,$8D
	.dc.b	$C5,$C3,$FC,$85,$5C,$5B,$1E,$0F,$86,$95,$DB,$E2,$2D,$70,$F7,$3C
	.dc.b	$A6,$3B,$BF,$D5,$98,$A3,$D3,$D5,$30,$FF,$4F,$FC,$FF,$DB,$C3,$C3
	.dc.b	$3E,$EE,$B9,$5E,$C8,$B3,$FE,$57,$F9,$EF,$A3,$97,$FA,$E5,$B5,$FD
	.dc.b	$2F,$07,$CC,$FF,$44,$9C,$DF,$FB,$2D,$FD,$19,$1D,$73,$FF,$4A,$F9
	.dc.b	$EF,$F1,$2E,$1F,$EB,$46,$CF,$86,$75,$E5,$75,$F6,$65,$FF,$AE,$B0
	.dc.b	$FE,$04,$D8,$F1,$7C,$0C,$8C,$1B,$7C,$DD,$32,$3E,$2F,$CF,$DC,$BE
	.dc.b	$07,$8F,$EB,$31,$BF,$C1,$7D,$DE,$A4,$BF,$9B,$E1,$5B,$F9,$1E,$4F
	.dc.b	$C3,$31,$87,$C9,$71,$5F,$ED,$2F,$A3,$FF,$7B,$3E,$B8,$FE,$DB,$18
	.dc.b	$AF,$A5,$BF,$CD,$3E,$77,$C5,$5B,$F8,$51,$F4,$7D,$0C,$57,$D9,$7F
	.dc.b	$BB,$FC,$FF,$F3,$18,$FF,$49,$3F,$83,$CB,$FD,$E8,$E5,$BE,$4F,$03
	.dc.b	$EB,$70,$DC,$7E,$B3,$F6,$7D,$8E,$BF,$41,$67,$FD,$DB,$82,$33,$18
	.dc.b	$F8,$0B,$97,$89,$F2,$92,$38,$CD,$9F,$3F,$97,$EA,$3D,$7C,$36,$BB
	.dc.b	$C7,$BB,$E0,$31,$5E,$E7,$E0,$FC,$4F,$F8,$9E,$03,$24,$66,$30,$D8
	.dc.b	$CD,$FF,$8A,$67,$FC,$55,$0F,$A0,$F9,$A7,$69,$F8,$2F,$D0,$F9,$89
	.dc.b	$1E,$0E,$03,$FE,$33,$F8,$17,$83,$AC,$7E,$AD,$DB,$DC,$7A,$2B,$77
	.dc.b	$F3,$AB,$4F,$91,$FF,$1A,$C6,$A7,$9C,$EC,$8F,$FB,$4E,$35,$5F,$82
	.dc.b	$AE,$73,$F1,$B1,$47,$4C,$33,$3F,$26,$2D,$F2,$C3,$C4,$76,$C3,$A3
	.dc.b	$F0,$4F,$81,$7C,$B0,$EB,$1B,$32,$5C,$3C,$58,$76,$AE,$0D,$9D,$4F
	.dc.b	$8D,$B8,$78,$E3,$8C,$6A,$FB,$C7,$C1,$BE,$73,$C1,$B1,$F2,$0D,$47
	.dc.b	$1B,$F0,$46,$70,$F6,$B6,$7D,$2D,$3C,$25,$18,$79,$5F,$41,$9C,$3C
	.dc.b	$B5,$F8,$08,$DE,$1E,$48,$EE,$66,$B9,$57,$0B,$FA,$37,$DE,$73,$87
	.dc.b	$19,$F2,$DF,$27,$6D,$AF,$C7,$61,$B7,$D3,$1D,$E3,$11,$E8,$8C,$4F
	.dc.b	$C5,$3F,$4D,$1F,$00,$FD,$46,$1F,$41,$B7,$FA,$D2,$4F,$08,$F8,$66
	.dc.b	$9C,$4F,$B6,$3F,$73,$C5,$FE,$24,$9F,$49,$FE,$9D,$71,$1E,$88,$C7
	.dc.b	$AB,$F9,$89,$CC,$FA,$67,$27,$E2,$A3,$35,$DE,$7F,$CA,$B5,$C4,$95
	.dc.b	$B6,$C7,$DD,$AE,$37,$FE,$A6,$6D,$7E,$86,$7D,$35,$C3,$0F,$A1,$1D
	.dc.b	$A7,$A1,$1E,$93,$B8,$F7,$C7,$27,$53,$CA,$BB,$CA,$3F,$15,$18,$6D
	.dc.b	$73,$9F,$8C,$73,$3B,$D7,$43,$BB,$E5,$32,$49,$F0,$4F,$38,$F9,$87
	.dc.b	$38,$CC,$F3,$8E,$E7,$38,$CB,$CE,$3B,$D7,$9D,$FA,$B1,$93,$8C,$65
	.dc.b	$ED,$1C,$2B,$C5,$87,$D9,$AF,$BF,$CD,$FE,$45,$BF,$CA,$B5,$FE,$78
	.dc.b	$D7,$FD,$CC,$D7,$FA,$3F,$D0,$7F,$87,$AF,$F2,$CD,$8F,$CF,$AF,$97
	.dc.b	$78,$3A,$F0,$D7,$F9,$DC,$3F,$CF,$B8,$5E,$EF,$D0,$28,$CB,$E6,$8C
	.dc.b	$FD,$07,$FB,$26,$3F,$4C,$D7,$F5,$ED,$7F,$AB,$35,$FF,$80,$0B,$E0
	.dc.b	$FF,$B0,$C7,$E1,$76,$BF,$2F,$FE,$49,$8F,$CB,$07,$0E,$CF,$ED,$35
	.dc.b	$CD,$DB,$5D,$07,$1C,$1F,$C6,$DB,$5E,$4F,$07,$C7,$3F,$C7,$FE,$81
	.dc.b	$E4,$E2,$BC,$F1,$8C,$3C,$AD,$7B,$E2,$F3,$38,$47,$F7,$EF,$9E,$79
	.dc.b	$18,$84,$61,$F7,$B5,$F4,$E8,$F4,$F8,$3F,$56,$3D,$FC,$18,$33,$E2
	.dc.b	$F5,$F9,$DF,$E7,$1B,$FF,$39,$6E,$DF,$41,$B7,$FE,$12,$5F,$0D,$C4
	.dc.b	$91,$F4,$A3,$2F,$D4,$F9,$5F,$23,$7D,$6F,$64,$7F,$A6,$63,$07,$1F
	.dc.b	$A0,$F2,$11,$C1,$D7,$C6,$F8,$9B,$BD,$9A,$49,$FC,$1F,$DC,$7A,$16
	.dc.b	$4F,$9F,$F8,$04,$8F,$47,$93,$F0,$FC,$9F,$88,$C8,$F6,$47,$C7,$F9
	.dc.b	$3F,$0F,$D0,$C9,$F8,$11,$F2,$FC,$AD,$46,$0F,$88,$FC,$FF,$80,$FB
	.dc.b	$9F,$42,$F5,$29,$9A,$F9,$5E,$55,$7F,$2B,$AA,$FE,$C4,$F2,$D6,$D7
	.dc.b	$03,$F4,$78,$7B,$DF,$D2,$F8,$EF,$A0,$FC,$62,$FA,$1A,$62,$33,$F3
	.dc.b	$16,$B8,$C6,$1E,$7F,$01,$F9,$5F,$0D,$F1,$54,$EB,$1D,$BE,$06,$D7
	.dc.b	$03,$C6,$F8,$BE,$07,$F3,$DE,$13,$E3,$69,$9A,$F4,$FC,$0D,$7F,$6F
	.dc.b	$B2,$FD,$73,$CB,$9F,$9F,$99,$2F,$F8,$2E,$29,$FD,$67,$5F,$F2,$19
	.dc.b	$FC,$3F,$9F,$83,$39,$6B,$C0,$DF,$81,$2B,$F1,$B1,$E8,$F3,$64,$76
	.dc.b	$7D,$7E,$EF,$6E,$89,$3F,$94,$7C,$3E,$07,$8B,$C0,$93,$F9,$87,$CB
	.dc.b	$EC,$7C,$4F,$84,$F8,$BE,$64,$AF,$5C,$7A,$BC,$2F,$5F,$B3,$27,$F4
	.dc.b	$31,$EE,$F6,$7D,$FF,$0D,$27,$F4,$B1,$F4,$39,$64,$7C,$67,$E9,$F9
	.dc.b	$DF,$C4,$67,$FA,$1C,$FF,$2E,$6B,$FE,$73,$F5,$75,$F1,$5B,$F0,$E4
	.dc.b	$ED,$E5,$E1,$7D,$EF,$7F,$EF,$F9,$D2,$BE,$AC,$77,$F9,$49,$1D,$DF
	.dc.b	$17,$87,$E3,$F0,$E4,$F8,$DF,$37,$C5,$FF,$27,$FD,$97,$F8,$01,$FC
	.dc.b	$4A,$7F,$9E,$B7,$D4,$FF,$2B,$3F,$C1,$77,$FC,$54,$FB,$7C,$4F,$80
	.dc.b	$7C,$1F,$17,$E1,$78,$FF,$0F,$E3,$A5,$7C,$38,$F0,$7B,$D2,$3F,$A9
	.dc.b	$E2,$8F,$27,$89,$FC,$EF,$E7,$5D,$D7,$BE,$61,$FD,$6A,$72,$DF,$4B
	.dc.b	$87,$CE,$C8,$C1,$BE,$EE,$93,$23,$07,$3E,$EF,$07,$0C,$8C,$1D,$FD
	.dc.b	$1F,$24,$C8,$C1,$FD,$B3,$E7,$1D,$1F,$D4,$FE,$75,$E0,$53,$FA,$B4
	.dc.b	$EA,$B8,$75,$62,$E4,$60,$FE,$EF,$F0,$EF,$92,$FE,$BD,$D2,$CE,$9F
	.dc.b	$D8,$67,$55,$C3,$E5,$71,$72,$30,$7F,$80,$BE,$77,$F8,$F3,$FE,$04
	.dc.b	$C1,$C9,$D8,$FE,$D9,$3A,$9F,$F0,$0A,$F5,$73,$E1,$49,$E9,$1C,$3E
	.dc.b	$6B,$8B,$23,$C1,$1D,$35,$F5,$F2,$49,$F2,$47,$83,$7D,$46,$0F,$43
	.dc.b	$E4,$F0,$7F,$CC,$51,$FE,$94,$C9,$18,$8F,$F7,$E4,$63,$C0,$FF,$49
	.dc.b	$36,$38,$78,$2F,$9D,$CB,$63,$31,$83,$FC,$76,$BE,$76,$91,$FF,$1F
	.dc.b	$BF,$E4,$0F,$88,$FF,$7E,$E1,$FE,$BF,$9F,$F2,$4F,$C1,$7F,$8C,$91
	.dc.b	$CC,$7C,$EB,$9B,$FC,$77,$1C,$0F,$F1,$89,$F8,$39,$E5,$CB,$D9,$DE
	.dc.b	$16,$7F,$C8,$73,$5E,$FC,$D5,$39,$E6,$7C,$6E,$63,$5F,$37,$31,$F4
	.dc.b	$45,$BF,$1F,$C3,$F3,$31,$AF,$BD,$F0,$2E,$6A,$77,$F0,$EE,$1F,$E5
	.dc.b	$49,$FD,$4B,$79,$B3,$87,$F9,$42,$72,$FF,$94,$BE,$77,$E2,$47,$FB
	.dc.b	$AD,$B5,$FE,$62,$D7,$FD,$E6,$FF,$98,$23,$35,$F4,$1F,$94,$F3,$DB
	.dc.b	$E3,$46,$A0,$EA,$FB,$7C,$66,$8D,$FC,$95,$C3,$B7,$C0,$72,$31,$87
	.dc.b	$9D,$EB,$ED,$49,$F2,$7C,$05,$AF,$67,$B7,$E2,$31,$97,$B5,$F2,$FB
	.dc.b	$32,$FD,$9E,$CD,$BF,$F3,$AE,$2F,$91,$FC,$03,$FE,$41,$AF,$F4,$0B
	.dc.b	$96,$FC,$3E,$B1,$DE,$16,$76,$FE,$27,$E7,$7F,$CB,$71,$FE,$8A,$6E
	.dc.b	$3F,$E8,$47,$F9,$D4,$67,$E2,$FF,$21,$F3,$BF,$8E,$8F,$F4,$8B,$6B
	.dc.b	$FD,$16,$BA,$B7,$D5,$C3,$86,$5F,$0F,$85,$B8,$70,$F3,$3A,$FB,$3D
	.dc.b	$9B,$5F,$E9,$EF,$9D,$FF,$17,$B7,$D8,$FD,$14,$3D,$9B,$1F,$A5,$E1
	.dc.b	$FC,$01,$3A,$F1,$E1,$67,$AB,$5D,$77,$2F,$F5,$5D,$A3,$F5,$DF,$3B
	.dc.b	$FE,$C4,$9F,$F5,$7A,$F7,$B3,$EF,$E9,$DF,$99,$C3,$CB,$B7,$C1,$E7
	.dc.b	$F2,$B2,$39,$FC,$9D,$E4,$FF,$59,$FC,$EF,$D0,$6F,$BD,$FB,$F8,$FF
	.dc.b	$6C,$57,$FB,$01,$D7,$86,$8F,$F6,$3C,$3E,$17,$81,$C7,$D6,$B3,$C7
	.dc.b	$D7,$E9,$ED,$75,$6C,$7F,$B3,$3E,$77,$F8,$DB,$79,$BF,$88,$46,$34
	.dc.b	$FF,$64,$BF,$AD,$3F,$28,$E3,$9B,$7E,$2F,$07,$8F,$8B,$30,$ED,$E4
	.dc.b	$5B,$FF,$69,$FC,$EB,$E1,$3F,$8F,$7C,$EB,$D9,$FE,$41,$39,$C3,$FD
	.dc.b	$C3,$39,$6C,$F0,$E8,$E2,$D8,$DC,$5C,$3F,$DC,$35,$C5,$B1,$E0,$F8
	.dc.b	$59,$5D,$BE,$5A,$D7,$0F,$47,$CA,$63,$9F,$E9,$47,$F4,$2C,$53,$EA
	.dc.b	$EA,$98,$7F,$BE,$3E,$77,$EC,$BF,$EF,$EA,$4F,$C0,$EB,$93,$EE,$8B
	.dc.b	$87,$FB,$E3,$E7,$3E,$96,$5E,$08,$46,$2B,$FA,$9C,$31,$7C,$7F,$F4
	.dc.b	$9A,$D4,$FF,$58,$7F,$93,$11,$D7,$3F,$F4,$FF,$9D,$FF,$10,$E1,$FD
	.dc.b	$7B,$1F,$04,$62,$BC,$95,$BF,$27,$B7,$B3,$30,$E0,$79,$31,$FE,$0C
	.dc.b	$E2,$FC,$53,$E5,$F6,$E4,$E6,$7D,$B3,$87,$E1,$74,$FA,$CF,$EC,$9A
	.dc.b	$FF,$B3,$6B,$86,$09,$F8,$1E,$34,$9C,$C5,$BF,$99,$EA,$7C,$53,$9F
	.dc.b	$83,$6B,$9F,$91,$F2,$4C,$57,$D3,$71,$1E,$7F,$9D,$FD,$A3,$C0,$C1
	.dc.b	$F4,$76,$8F,$9B,$E3,$6D,$FD,$D9,$FA,$DE,$7E,$38,$FE,$F6,$C8,$FF
	.dc.b	$01,$7C,$EF,$F3,$89,$FC,$13,$F8,$CB,$FF,$09,$B9,$C7,$F0,$6E,$61
	.dc.b	$C9,$FE,$15,$9F,$9C,$D8,$57,$4F,$23,$F0,$5C,$9F,$83,$F6,$3C,$3F
	.dc.b	$25,$71,$FA,$D9,$23,$CB,$EF,$C9,$CC,$5C,$BA,$C7,$92,$7F,$C5,$91
	.dc.b	$E8,$6A,$AF,$F0,$DC,$FF,$8A,$A3,$4D,$75,$E5,$70,$9F,$F3,$0C,$E5
	.dc.b	$B3,$C5,$AF,$F8,$D1,$FF,$4A,$C6,$FD,$EE,$44,$79,$7C,$87,$D0,$3E
	.dc.b	$8F,$C5,$D8,$F8,$31,$E1,$23,$C0,$F3,$6B,$FD,$2F,$0F,$82,$7F,$92
	.dc.b	$1C,$6D,$F6,$BE,$5B,$FC,$7D,$18,$AF,$DD,$B6,$B8,$7A,$5F,$82,$AD
	.dc.b	$B2,$BF,$51,$5B,$AF,$95,$59,$7C,$27,$FA,$DE,$B1,$7E,$1B,$E3,$18
	.dc.b	$7C,$2F,$77,$24,$9E,$77,$DF,$36,$1C,$67,$CF,$5B,$C3,$B4,$70,$C3
	.dc.b	$4F,$8E,$7F,$17,$58,$C3,$D4,$75,$71,$5C,$30,$EF,$5D,$4F,$39,$6F
	.dc.b	$F4,$75,$C2,$31,$1F,$11,$B7,$DA,$FD,$D3,$E5,$9C,$C6,$30,$FE,$61
	.dc.b	$5B,$9E,$78,$78,$AF,$81,$2B,$77,$C6,$BC,$B3,$96,$CF,$C4,$36,$F8
	.dc.b	$0F,$86,$58,$E0,$7B,$1C,$47,$08,$FB,$B5,$E3,$8C,$12,$7B,$9F,$C1
	.dc.b	$30,$F8,$95,$D0,$F8,$85,$C3,$DC,$F4,$8D,$D7,$38,$F1,$1B,$C3,$F7
	.dc.b	$35,$F9,$F8,$FF,$50,$B5,$F8,$26,$23,$64,$BF,$D3,$B6,$3F,$2F,$CB
	.dc.b	$FE,$0E,$BF,$04,$23,$15,$EC,$8F,$F5,$9C,$61,$FC,$89,$CB,$CC,$FC
	.dc.b	$7C,$66,$3F,$20,$66,$7F,$0B,$19,$C3,$F5,$4D,$C3,$FD,$64,$D9,$FE
	.dc.b	$AC,$DA,$E4,$7C,$B7,$18,$7F,$77,$6C,$FF,$76,$6C,$FA,$30,$F2,$92
	.dc.b	$7C,$6D,$8E,$8D,$AF,$87,$5C,$A7,$E2,$4E,$6B,$E2,$46,$1E,$64,$79
	.dc.b	$B6,$B9,$CF,$19,$F9,$31,$E9,$8C,$47,$C8,$8C,$4F,$13,$BD,$78,$C9
	.dc.b	$5C,$30,$F9,$6D,$C3,$E5,$B6,$BE,$5B,$70,$F9,$78,$79,$27,$B4,$7F
	.dc.b	$BC,$3E,$77,$FC,$3B,$7F,$D0,$DB,$1F,$D1,$D7,$C8,$3C,$1D,$78,$4B
	.dc.b	$FE,$95,$F2,$3F,$E2,$6C,$2C,$FF,$4D,$51,$97,$FA,$93,$5F,$F7,$EF
	.dc.b	$CE,$34,$E0,$D8,$FE,$AC,$D7,$FA,$D3,$5F,$F1,$B8,$F9,$B7,$B3,$E1
	.dc.b	$6C,$7F,$5F,$6B,$FD,$8D,$AF,$8B,$E1,$3C,$24,$7C,$2F,$93,$C3,$91
	.dc.b	$E4,$6B,$FD,$B7,$E6,$1E,$6F,$C1,$6C,$7F,$6F,$6B,$B8,$F7,$FA,$72
	.dc.b	$3D,$ED,$7D,$FC,$BF,$E5,$4E,$4F,$F2,$C6,$9F,$B9,$9F,$F1,$F3,$5F
	.dc.b	$EA,$AB,$C3,$E1,$CB,$FE,$5C,$F9,$AF,$F3,$05,$FF,$98,$5B,$19,$37
	.dc.b	$61,$0F,$85,$CC,$3E,$71,$9E,$2B,$53,$FE,$51,$6B,$FD,$A2,$76,$FF
	.dc.b	$95,$5F,$F4,$C3,$7D,$6E,$B1,$88,$FE,$DF,$2F,$99,$ED,$E1,$F7,$64
	.dc.b	$F7,$8F,$2F,$CD,$7B,$23,$D5,$C9,$F8,$9D,$7F,$D0,$0D,$D7,$FC,$FF
	.dc.b	$5B,$AF,$7F,$99,$EB,$F9,$A9,$1E,$B8,$F7,$78,$BE,$F6,$47,$BE,$3E
	.dc.b	$17,$C9,$7C,$6C,$5F,$0D,$4C,$D7,$C5,$F1,$AB,$EE,$AF,$95,$1F,$2F
	.dc.b	$DF,$F9,$FF,$36,$F9,$CA,$66,$FE,$77,$8F,$5F,$8F,$5F,$4E,$38,$EB
	.dc.b	$67,$67,$FA,$16,$D1,$F7,$3C,$57,$CB,$5F,$9C,$97,$FE,$8B,$9C,$B7
	.dc.b	$E5,$F8,$54,$8C,$1B,$F8,$0F,$C3,$F8,$2F,$89,$A7,$78,$F0,$7C,$0D
	.dc.b	$AE,$07,$77,$C9,$F0,$3F,$90,$F0,$DF,$19,$4C,$D7,$97,$E0,$AB,$FE
	.dc.b	$36,$D5,$7E,$D9,$E5,$CD,$CF,$C8,$97,$FD,$4A,$3F,$13,$7F,$8A,$F3
	.dc.b	$24,$FF,$AA,$FB,$3F,$91,$B7,$C0,$FC,$7F,$81,$23,$07,$F5,$CF,$27
	.dc.b	$D0,$47,$FA,$FF,$B7,$F9,$6E,$79,$1D,$5F,$CC,$F8,$1F,$9A,$F0,$24
	.dc.b	$F8,$5F,$CE,$FA,$9F,$0F,$E1,$3E,$27,$91,$2B,$FB,$5C,$7E,$8B,$C1
	.dc.b	$FD,$1F,$AB,$27,$D5,$1F,$A6,$F9,$AF,$D3,$FC,$3C,$9F,$74,$7C,$FE
	.dc.b	$99,$1F,$DD,$9F,$A5,$E6,$FF,$02,$9F,$E8,$53,$FE,$CF,$6B,$FE,$66
	.dc.b	$6F,$93,$FE,$EF,$5F,$15,$C7,$C3,$95,$C5,$E9,$E1,$7D,$FF,$77,$F0
	.dc.b	$1E,$74,$AF,$F0,$3C,$78,$7E,$4E,$47,$85,$F2,$78,$7E,$5F,$0E,$4F
	.dc.b	$95,$F4,$7C,$5F,$F1,$AF,$5B,$F8,$01,$FC,$5B,$67,$9F,$FC,$86,$FF
	.dc.b	$AC,$FA,$5B,$8F,$81,$E3,$7C,$1F,$1A,$57,$C1,$8F,$89,$E4,$7C,$5F
	.dc.b	$8D,$93,$F1,$63,$C1,$ED,$C8,$FF,$7B,$66,$8F,$27,$8F,$FC,$BB,$E7
	.dc.b	$1D,$F7,$C0,$98,$7F,$BF,$67,$2D,$F4,$B8,$7C,$E4,$8C,$1B,$EF,$E9
	.dc.b	$32,$30,$73,$EF,$F0,$76,$C8,$C1,$DF,$D2,$F2,$4C,$8C,$1F,$D1,$F9
	.dc.b	$7E,$4B,$FD,$27,$E6,$DD,$5F,$D2,$3E,$73,$FA,$A4,$EF,$9F,$E7,$46
	.dc.b	$23,$FD,$5B,$5C,$1D,$1C,$FA,$FF,$A8,$C6,$63,$77,$89,$FE,$AB,$F0
	.dc.b	$7F,$DF,$B1,$F1,$D3,$A5,$9F,$F5,$A6,$C7,$61,$5C,$73,$FC,$12,$F9
	.dc.b	$3E,$BE,$EC,$AE,$13,$97,$97,$CE,$7F,$26,$7F,$D8,$F8,$39,$7C,$07
	.dc.b	$F6,$A9,$D6,$7E,$65,$EB,$67,$C3,$98,$74,$8E,$1F,$33,$C5,$91,$E0
	.dc.b	$8E,$9C,$1E,$DE,$59,$3E,$48,$F0,$70,$EA,$30,$7A,$1F,$27,$84,$CF
	.dc.b	$FC,$5B,$3F,$E0,$3C,$D3,$98,$CC,$62,$77,$08,$C3,$AF,$09,$F3,$9A
	.dc.b	$94,$62,$75,$39,$8C,$47,$4F,$9C,$FF,$26,$26,$91,$F0,$FC,$27,$E4
	.dc.b	$B0,$FC,$0C,$7F,$96,$79,$7E,$F3,$FE,$1D,$AF,$F0,$F7,$CE,$7E,$02
	.dc.b	$3F,$80,$5F,$1A,$FF,$26,$F2,$D9,$FE,$08,$D7,$FC,$4D,$35,$FF,$28
	.dc.b	$CF,$6E,$6C,$1B,$E5,$7C,$46,$27,$C7,$F3,$3C,$4C,$5F,$D1,$8F,$F2
	.dc.b	$EC,$75,$F9,$19,$1C,$0F,$5B,$E3,$D5,$CB,$4F,$07,$85,$6B,$F3,$D3
	.dc.b	$FB,$46,$F2,$E7,$1F,$01,$FB,$B7,$DD,$F3,$9F,$26,$32,$1C,$57,$EB
	.dc.b	$B4,$78,$1C,$37,$E3,$C1,$32,$FD,$76,$1C,$1F,$E5,$CF,$C8,$C7,$AA
	.dc.b	$F9,$1F,$09,$E2,$F9,$1B,$5C,$BE,$03,$CA,$62,$F8,$BE,$8F,$6A,$4F
	.dc.b	$83,$E0,$2D,$76,$F6,$FD,$A6,$37,$FE,$0F,$9A,$3E,$0F,$B3,$30,$F4
	.dc.b	$FB,$37,$0F,$87,$F1,$7F,$72,$7F,$85,$36,$33,$EE,$FE,$1E,$31,$97
	.dc.b	$F1,$45,$E5,$64,$D7,$91,$70,$FE,$29,$F6,$AF,$A2,$FC,$83,$63,$3E
	.dc.b	$0E,$27,$3D,$12,$78,$FA,$B7,$5F,$B3,$F6,$8E,$FF,$CD,$36,$BF,$91
	.dc.b	$3F,$C7,$DB,$1A,$E0,$91,$AE,$2B,$7F,$84,$FC,$03,$F9,$6C,$FF,$2B
	.dc.b	$6C,$67,$B3,$F5,$11,$8E,$2F,$E6,$B5,$FC,$08,$FE,$64,$5A,$FC,$7B
	.dc.b	$FC,$40,$CE,$DF,$B1,$FB,$5F,$F4,$B4,$7F,$39,$6C,$F0,$FA,$5D,$3B
	.dc.b	$F1,$25,$6F,$BA,$CF,$E7,$2F,$F0,$13,$CF,$4E,$DD,$7D,$DF,$E0,$1E
	.dc.b	$37,$FA,$4E,$7F,$A4,$36,$FF,$4A,$BE,$1B,$7F,$0B,$C3,$CD,$97,$C3
	.dc.b	$D5,$B8,$7E,$B3,$ED,$3F,$73,$3F,$D4,$DB,$19,$F4,$3F,$88,$46,$38
	.dc.b	$BF,$AE,$AF,$9B,$C9,$AF,$9B,$B8,$73,$E3,$E2,$72,$FA,$5B,$1F,$D8
	.dc.b	$7C,$9F,$F3,$6B,$7C,$7F,$E3,$D1,$FB,$38,$C1,$FD,$9F,$F0,$AF,$F6
	.dc.b	$23,$7E,$07,$F9,$24,$7F,$8A,$A3,$07,$F6,$DF,$9E,$FF,$72,$B7,$E5
	.dc.b	$BF,$95,$7C,$F6,$F8,$7F,$86,$D6,$9B,$E1,$7F,$2F,$8F,$F3,$DD,$F1
	.dc.b	$3A,$46,$7C,$39,$18,$3C,$0F,$1D,$E4,$7F,$7F,$C3,$FC,$60,$F9,$63
	.dc.b	$B7,$04,$8C,$1E,$97,$FA,$07,$CE,$FE,$3B,$8F,$FC,$1A,$D9,$FF,$24
	.dc.b	$5F,$E0,$5F,$F0,$64,$67,$3F,$F7,$63,$FC,$85,$BB,$F4,$7A,$BD,$BC
	.dc.b	$39,$5E,$7E,$8B,$5F,$CC,$AD,$E9,$F9,$0C,$E5,$E4,$7C,$FE,$E4,$C3
	.dc.b	$C7,$16,$F9,$7A,$1E,$C3,$E9,$FC,$A5,$AF,$4F,$A5,$F0,$0C,$57,$85
	.dc.b	$F8,$7E,$0C,$9F,$99,$F1,$D6,$BC,$DE,$17,$C8,$31,$A7,$C8,$7E,$5F
	.dc.b	$C6,$CB,$FA,$9E,$FD,$BF,$77,$C6,$FD,$03,$18,$7D,$07,$E9,$FC,$54
	.dc.b	$BF,$9F,$16,$BF,$AB,$FC,$DF,$E5,$9B,$DD,$FD,$E7,$E7,$9F,$2D,$FD
	.dc.b	$EE,$B5,$97,$F9,$21,$FC,$2D,$F2,$F8,$AD,$61,$B7,$87,$B1,$33,$FF
	.dc.b	$25,$7C,$F3,$E7,$B1,$87,$8B,$E0,$7B,$18,$CF,$B3,$E1,$E3,$98,$7E
	.dc.b	$57,$DA,$B7,$EA,$E2,$F3,$18,$C3,$CC,$FA,$7E,$02,$5F,$F9,$6F,$7F
	.dc.b	$F2,$83,$63,$DD,$F1,$52,$3C,$FD,$57,$1F,$8F,$F2,$7F,$0C,$C6,$5F
	.dc.b	$0D,$F8,$DC,$72,$FF,$D8,$1E,$25,$BF,$F0,$2F,$1B,$DE,$FF,$13,$36
	.dc.b	$BC,$97,$89,$FF,$15,$7D,$DB,$E4,$FF,$C6,$F8,$04,$66,$FF,$C9,$AD
	.dc.b	$8F,$F2,$AB,$5F,$F5,$C1,$C2,$7F,$C6,$DF,$85,$FF,$B4,$2B,$FC,$FD
	.dc.b	$5C,$74,$FF,$42,$CE,$6F,$FC,$3B,$2F,$8C,$F2,$9F,$E3,$FF,$91,$FE
	.dc.b	$7D,$7F,$E8,$C9,$D6,$1D,$2B,$D3,$7E,$86,$CE,$A3,$CF,$18,$F5,$25
	.dc.b	$7B,$1B,$3E,$C7,$C1,$D5,$F1,$9B,$3E,$AF,$88,$F9,$27,$97,$A6,$C6
	.dc.b	$49,$7F,$E3,$8C,$FF,$C2,$31,$FE,$DD,$CB,$FA,$53,$6F,$8D,$6E,$EE
	.dc.b	$7E,$F9,$B5,$FE,$E5,$C3,$9D,$FD,$7C,$39,$B6,$B6,$FD,$B9,$DB,$F7
	.dc.b	$23,$2D,$AF,$F7,$1E,$18,$BE,$73,$93,$C9,$3E,$CC,$3C,$8D,$3C,$95
	.dc.b	$C6,$FA,$E1,$E0,$C3,$99,$23,$E4,$57,$1C,$3D,$0D,$55,$C6,$7A,$4F
	.dc.b	$98,$D5,$73,$71,$5F,$8A,$9D,$BE,$83,$11,$C2,$3C,$31,$82,$61,$EC
	.dc.b	$78,$34,$93,$F5,$CE,$EE,$23,$D2,$72,$8F,$86,$77,$3F,$53,$3F,$94
	.dc.b	$23,$82,$47,$EB,$23,$2D,$8F,$D4,$FC,$47,$D5,$23,$F0,$9B,$1E,$13
	.dc.b	$F0,$8E,$2B,$CD,$1F,$BC,$8C,$4F,$AE,$3F,$63,$18,$7D,$73,$FA,$F8
	.dc.b	$C4,$7E,$0E,$31,$3E,$12,$47,$84,$EC,$E2,$3C,$37,$FE,$A1,$9D,$36
	.dc.b	$33,$5D,$A3,$F0,$ED,$BF,$E2,$ED,$BF,$86,$75,$3D,$55,$E4,$BE,$B3
	.dc.b	$B7,$DB,$3B,$7D,$AF,$EC,$E3,$53,$C4,$FE,$1D,$F0,$7F,$C9,$8F,$E4
	.dc.b	$FE,$FD,$AF,$E3,$71,$FC,$38,$8F,$F2,$67,$F8,$8F,$24,$8F,$E5,$2D
	.dc.b	$C3,$FD,$FB,$9B,$C7,$FE,$32,$DA,$FF,$7D,$B6,$77,$3A,$F9,$69,$5A
	.dc.b	$6C,$FF,$32,$1F,$41,$FE,$63,$E5,$C4,$63,$DD,$FE,$8C,$EE,$5F,$26
	.dc.b	$F4,$77,$1C,$67,$A7,$CF,$BE,$83,$FA,$34,$7F,$B5,$3C,$BF,$F6,$43
	.dc.b	$FD,$27,$89,$AE,$A7,$33,$FE,$DA,$BE,$4E,$E7,$9C,$3B,$FA,$DF,$2A
	.dc.b	$FE,$A9,$3C,$1E,$58,$6A,$F8,$DD,$9D,$FC,$6D,$5F,$38,$F6,$30,$D8
	.dc.b	$F1,$1C,$0A,$31,$7A,$1C,$D5,$D3,$11,$9F,$8D,$F0,$7C,$F7,$F9,$1A
	.dc.b	$F9,$B5,$F8,$51,$CF,$6E,$11,$E0,$C1,$39,$8C,$7A,$15,$F9,$55,$E3
	.dc.b	$8F,$77,$CF,$FC,$79,$70,$5F,$9F,$FF,$52,$47,$B1,$AE,$B8,$3E,$DD
	.dc.b	$71,$AE,$4B,$D1,$EB,$32,$BE,$2C,$E5,$BE,$CF,$CF,$FA,$07,$CE,$D3
	.dc.b	$CB,$1E,$6F,$1A,$D7,$03,$9B,$EA,$EC,$FB,$1F,$59,$FE,$AD,$9F,$A6
	.dc.b	$D7,$D0,$7F,$85,$F1,$4F,$C2,$F4,$F9,$4F,$9A,$30,$FA,$3D,$2A,$F2
	.dc.b	$14,$AF,$97,$3F,$33,$E2,$7E,$DF,$D0,$CB,$F8,$31,$F4,$3D,$2D,$46
	.dc.b	$0F,$8A,$FD,$4F,$57,$F2,$5C,$7F,$BF,$9F,$A5,$E2,$65,$F9,$7E,$8D
	.dc.b	$7C,$9A,$2F,$EE,$CF,$E6,$FC,$3F,$F1,$C5,$7E,$5B,$B9,$E5,$FE,$70
	.dc.b	$D7,$A3,$BF,$42,$D7,$E7,$2F,$8B,$F0,$5F,$F1,$83,$9F,$43,$83,$B9
	.dc.b	$F0,$47,$87,$E6,$BF,$53,$F3,$D2,$BA,$47,$97,$E6,$F5,$18,$3C,$0F
	.dc.b	$A7,$E1,$7F,$69,$E0,$BE,$26,$99,$AF,$37,$91,$5F,$35,$7C,$18,$F8
	.dc.b	$5E,$87,$EF,$7D,$0F,$B1,$5F,$B8,$6B,$ED,$8F,$91,$F0,$DC,$23,$E1
	.dc.b	$46,$1F,$89,$F5,$55,$F8,$93,$F3,$E3,$F8,$57,$DB,$3C,$1B,$E1,$4B
	.dc.b	$FE,$0F,$39,$6F,$CC,$F0,$F9,$69,$18,$3F,$86,$FD,$B3,$D8,$FE,$2C
	.dc.b	$B8,$A9,$88,$CF,$06,$BE,$36,$5F,$28,$FE,$3D,$F6,$AF,$97,$BF,$2F
	.dc.b	$2B,$F8,$DC,$E5,$BF,$13,$C3,$D5,$91,$83,$F9,$0F,$7B,$E0,$BF,$95
	.dc.b	$3F,$BF,$8F,$D4,$37,$D2,$FE,$20,$EE,$38,$F9,$3C,$3E,$56,$46,$0F
	.dc.b	$E5,$BF,$6B,$F9,$86,$F8,$3F,$CD,$E0,$DB,$C7,$16,$7C,$09,$7C,$A3
	.dc.b	$87,$CA,$F1,$F9,$59,$3D,$A3,$FA,$17,$DA,$FE,$3E,$3F,$9E,$BF,$D5
	.dc.b	$35,$FE,$50,$DF,$5B,$FA,$EB,$B9,$FE,$0D,$F6,$8D,$BF,$A6,$2F,$8A
	.dc.b	$A6,$23,$FA,$87,$77,$AA,$BF,$A5,$BF,$AE,$7F,$88,$37,$1F,$E7,$0E
	.dc.b	$E3,$3E,$17,$2F,$A6,$91,$CA,$3A,$EB,$DB,$C2,$93,$DA,$3F,$B0,$7D
	.dc.b	$A7,$AE,$3F,$D7,$8F,$F6,$D7,$FA,$EB,$5F,$E4,$4E,$D7,$D3,$FF,$43
	.dc.b	$DF,$FB,$FC,$FF,$3C,$9E,$3E,$D7,$F6,$DA,$D4,$7F,$B2,$5F,$88,$FF
	.dc.b	$6A,$5E,$BE,$CE,$1B,$F1,$E8,$E5,$C1,$2F,$84,$75,$F5,$B5,$18,$3A
	.dc.b	$3E,$1F,$17,$C5,$E2,$C9,$F0,$47,$97,$D3,$D4,$60,$F2,$3F,$E0,$5F
	.dc.b	$A6,$7C,$57,$F8,$1F,$E9,$DF,$01,$4D,$46,$FB,$6E,$1F,$E0,$6B,$E2
	.dc.b	$D8,$FF,$08,$7C,$BB,$C6,$FF,$09,$7A,$0D,$29,$A8,$DF,$91,$70,$FF
	.dc.b	$08,$DF,$16,$C7,$F8,$6F,$E9,$FF,$11,$D1,$FE,$32,$C7,$E6,$D7,$F8
	.dc.b	$61,$AF,$F5,$D9,$DB,$FE,$FD,$7F,$1D,$DD,$FE,$07,$73,$06,$9A,$AF
	.dc.b	$F1,$FB,$FB,$78,$E9,$E2,$78,$CF,$27,$89,$E5,$DF,$DF,$CB,$2B,$FC
	.dc.b	$57,$7F,$19,$F2,$6F,$B7,$35,$ED,$8F,$77,$89,$FE,$1A,$9F,$F2,$E5
	.dc.b	$7F,$25,$CF,$FC,$B6,$D8,$CC,$ED,$FF,$2A,$9E,$BF,$31,$F4,$FC,$1B
	.dc.b	$1D,$1A,$F0,$AF,$E6,$93,$FE,$68,$80,$FA,$0C,$34,$D3,$83,$87,$C6
	.dc.b	$7F,$96,$71,$FF,$1A,$4F,$F9,$B7,$7F,$F2,$6B,$D2,$BA,$FD,$3B,$8F
	.dc.b	$D3,$87,$1B,$F5,$F2,$5C,$3F,$CE,$2D,$7C,$B3,$5F,$C8,$4F,$6E,$4C
	.dc.b	$1B,$E4,$7C,$96,$27,$C7,$F3,$BC,$4C,$5F,$F8,$94,$3D,$7E,$72,$4F
	.dc.b	$03,$D6,$F8,$F6,$FF,$09,$C7,$F9,$A9,$B5,$FE,$2F,$C3,$F9,$44,$E7
	.dc.b	$9B,$87,$2C,$8C,$1C,$5F,$77,$D3,$FC,$9B,$FF,$1C,$37,$0F,$F3,$06
	.dc.b	$0E,$ED,$ED,$C3,$79,$97,$F9,$F3,$0E,$0F,$FA,$69,$F8,$F8,$F4,$F9
	.dc.b	$DE,$E7,$83,$E4,$2D,$71,$F7,$3C,$66,$2B,$C6,$F9,$BD,$39,$3D,$FE
	.dc.b	$42,$D7,$A3,$D5,$F5,$98,$DF,$FC,$F1,$9A,$3D,$FF,$1D,$30,$F4,$7D
	.dc.b	$2D,$C3,$FC,$B1,$F2,$BF,$6A,$7F,$D2,$ED,$8C,$FB,$DF,$85,$8C,$65
	.dc.b	$FE,$85,$7F,$C9,$95,$C3,$C8,$D7,$8B,$B7,$F1,$9F,$4E,$F9,$7A,$63
	.dc.b	$C3,$B7,$FE,$A6,$9F,$F5,$63,$7A,$38,$FB,$BC,$9F,$CB,$7D,$3F,$E2
	.dc.b	$A7,$FD,$5A,$DD,$BF,$CE,$4F,$F8,$66,$33,$DD,$C3,$82,$47,$0E,$2B
	.dc.b	$7F,$E7,$9F,$A7,$FE,$59,$3F,$EB,$86,$C6,$7B,$3F,$4D,$18,$E2,$FF
	.dc.b	$53,$57,$F8,$C4,$FF,$5F,$96,$BF,$D0,$8F,$F8,$CC,$CE,$DF,$AF,$FA
	.dc.b	$7F,$E0,$71,$FE,$C6,$6C,$F0,$F7,$74,$EF,$C3,$95,$BE,$EB,$3F,$E9
	.dc.b	$1B,$FC,$04,$F3,$D3,$B7,$5F,$77,$F7,$DE,$33,$DB,$A6,$3E,$6A,$DF
	.dc.b	$FB,$62,$5E,$6F,$03,$8F,$A7,$67,$FD,$BB,$F4,$FF,$B4,$E1,$FF,$64
	.dc.b	$36,$3F,$DF,$AF,$F8,$C9,$DF,$8F,$FC,$3E,$1F,$0D,$4C,$7B,$B6,$BA
	.dc.b	$F1,$EC,$E1,$EE,$58,$FF,$76,$F9,$3F,$E8,$16,$F8,$FF,$C6,$E1,$E4
	.dc.b	$D8,$FE,$3D,$F4,$FF,$ED,$3A,$FF,$7A,$B7,$9B,$FD,$74,$A3,$7F,$31
	.dc.b	$FC,$97,$E9,$DF,$0F,$4F,$E5,$1F,$4F,$BA,$FF,$7F,$CE,$9B,$E2,$71
	.dc.b	$38,$38,$37,$E2,$75,$34,$E0,$E7,$C1,$E1,$3B,$FC,$3D,$8E,$FC,$5E
	.dc.b	$53,$1A,$FF,$B4,$F1,$F9,$B3,$FD,$43,$97,$FD,$F9,$F4,$FF,$8E,$6D
	.dc.b	$7A,$BA,$BF,$A1,$4E,$6B,$FA,$BC,$A3,$F8,$FB,$FD,$69,$AE,$4E,$3C
	.dc.b	$7C,$BC,$59,$5E,$13,$9E,$FF,$91,$71,$5C,$E7,$C3,$E9,$49,$CC,$F8
	.dc.b	$67,$0F,$FB,$BF,$0F,$EB,$D3,$A6,$C6,$6B,$D7,$EA,$7B,$3D,$29,$7D
	.dc.b	$31,$7C,$AE,$30,$E9,$D1,$F0,$8C,$6D,$F2,$1F,$8B,$E1,$4C,$39,$7C
	.dc.b	$C5,$BF,$27,$87,$F2,$8C,$61,$F4,$1F,$9B,$F1,$F2,$FC,$5F,$07,$6F
	.dc.b	$D9,$F0,$1F,$48,$C6,$1F,$61,$FA,$BF,$4F,$2F,$E0,$45,$A5,$7E,$65
	.dc.b	$FF,$07,$FD,$27,$E7,$63,$10,$8C,$57,$DE,$FA,$77,$CC,$7F,$80,$E7
	.dc.b	$35,$CB,$AF,$06,$7C,$0B,$87,$D6,$AF,$F6,$E3,$7D,$8F,$F0,$67,$D3
	.dc.b	$BE,$9F,$A4,$75,$F7,$E5,$6C,$E4,$F8,$78,$E4,$7D,$7F,$66,$E9,$EE
	.dc.b	$8F,$CA,$E5,$FE,$84,$7C,$5C,$5A,$E0,$F7,$CE,$6B,$81,$F1,$3E,$27
	.dc.b	$DC,$7B,$EA,$DF,$AF,$E2,$7E,$11,$EA,$EB,$B1,$EF,$FA,$7F,$8C,$63
	.dc.b	$0E,$A7,$CD,$7E,$57,$1C,$9F,$CD,$FB,$0F,$1B,$17,$C3,$95,$EF,$7F
	.dc.b	$8C,$DB,$7F,$37,$0C,$4F,$F8,$DF,$E9,$9F,$37,$FE,$39,$FA,$77,$D4
	.dc.b	$66,$30,$CC,$3F,$CB,$79,$7F,$94,$AB,$71,$88,$46,$1D,$B5,$FF,$31
	.dc.b	$DB,$E7,$3F,$CA,$71,$FE,$5C,$8C,$57,$36,$C7,$81,$AF,$29,$FF,$28
	.dc.b	$7C,$DF,$D5,$E5,$FC,$AD,$70,$7F,$2F,$19,$BF,$E2,$73,$FD,$2E,$FE
	.dc.b	$7C,$FE,$41,$FA,$31,$9A,$E3,$D3,$D8,$97,$FC,$32,$74,$FF,$1E,$BE
	.dc.b	$55,$E2,$F8,$8F,$9C,$7C,$FF,$5E,$D7,$18,$B8,$F8,$26,$D7,$1A,$D1
	.dc.b	$DE,$32,$FA,$AA,$D7,$C9,$9A,$B0,$FF,$65,$D7,$28,$EC,$DB,$ED,$1F
	.dc.b	$02,$FC,$D3,$93,$B1,$63,$B3,$5E,$D1,$E0,$AF,$EF,$11,$F6,$E3,$07
	.dc.b	$47,$B4,$F1,$9F,$2C,$79,$0C,$92,$BC,$A7,$DC,$BE,$75,$C5,$B3,$B8
	.dc.b	$FB,$B1,$89,$E3,$18,$3C,$17,$E7,$9D,$47,$BA,$BC,$55,$C1,$F5,$47
	.dc.b	$73,$89,$5F,$01,$A3,$F2,$D5,$F9,$C7,$E2,$C7,$E6,$E3,$17,$F0,$27
	.dc.b	$F4,$71,$82,$4E,$CF,$EB,$38,$7F,$A5,$A7,$B8,$71,$7E,$78,$C7,$3F
	.dc.b	$E0,$A7,$35,$ED,$AF,$D1,$CE,$49,$1F,$6E,$77,$5E,$BA,$DC,$E6,$B7
	.dc.b	$3E,$2A,$DB,$EE,$3F,$57,$1A,$AF,$24,$EB,$0F,$BF,$1D,$2B,$95,$6E
	.dc.b	$FC,$C7,$EA,$DC,$47,$8E,$30,$78,$C8,$F2,$6C,$F8,$27,$8C,$F9,$63
	.dc.b	$C0,$66,$76,$78,$23,$FD,$DB,$7E,$7C,$3F,$86,$DF,$2C,$3D,$37,$E2
	.dc.b	$BF,$01,$30,$FE,$1F,$EC,$7F,$14,$3F,$97,$7C,$25,$AF,$B7,$1F,$85
	.dc.b	$C8,$D2,$47,$F2,$B8,$FF,$06,$B7,$C8,$E0,$47,$5E,$D7,$32,$38,$3F
	.dc.b	$9A,$7D,$43,$D1,$FE,$8B,$9F,$F3,$A8,$C7,$2E,$BE,$79,$E9,$70,$BD
	.dc.b	$DF,$F8,$E8,$FC,$87,$77,$F3,$F6,$9E,$77,$1A,$1D,$7D,$33,$EA,$3F
	.dc.b	$C8,$7C,$3F,$E0,$AF,$33,$FC,$DB,$87,$F8,$D9,$AF,$F6,$2A,$E0,$FF
	.dc.b	$8E,$97,$83,$CB,$7F,$F2,$6C,$6A,$78,$61,$94,$23,$8B,$CF,$E3,$34
	.dc.b	$BE,$85,$EC,$E1,$B1,$B3,$C5,$E8,$F0,$B5,$D3,$11,$9F,$8C,$EB,$F5
	.dc.b	$0F,$5E,$99,$C3,$D9,$B5,$7E,$05,$75,$8E,$DE,$95,$9D,$9F,$0D,$F4
	.dc.b	$FD,$1F,$F8,$D3,$87,$FC,$6B,$F4,$9F,$7E,$BF,$CC,$2D,$7F,$C1,$B3
	.dc.b	$BE,$1F,$F3,$AA,$F4,$F9,$C4,$9F,$F3,$4C,$E5,$BF,$0D,$F2,$3E,$A1
	.dc.b	$F3,$74,$F1,$C7,$93,$C6,$B5,$C0,$E6,$FA,$3E,$9F,$E8,$FD,$47,$FA
	.dc.b	$EA,$7F,$CE,$ED,$7C,$F1,$EE,$F5,$78,$47,$92,$30,$F9,$BD,$3A,$FB
	.dc.b	$A7,$E2,$C7,$C6,$F0,$FE,$9F,$D4,$4A,$F6,$C7,$CA,$EE,$D4,$60,$F8
	.dc.b	$2F,$CE,$F5,$BF,$0F,$C5,$FE,$80,$9F,$9B,$E3,$65,$F8,$FE,$8D,$7E
	.dc.b	$7D,$7D,$68,$FC,$7F,$8B,$FE,$30,$9F,$F4,$AF,$7B,$CD,$FC,$81,$AF
	.dc.b	$47,$7E,$7D,$AF,$F4,$DD,$F1,$7F,$C8,$EF,$F8,$AD,$CF,$CD,$F0,$77
	.dc.b	$3E,$08,$F0,$FC,$DF,$E7,$7E,$A2,$57,$48,$F2,$FC,$E6,$A3,$07,$81
	.dc.b	$F4,$FC,$47,$EA,$3C,$17,$C5,$D3,$35,$E6,$F2,$6B,$E6,$AF,$83,$1F
	.dc.b	$0B,$D9,$FD,$8F,$9A,$EA,$A6,$6F,$DF,$F1,$35,$F7,$D7,$CA,$8F,$97
	.dc.b	$F4,$36,$76,$7C,$67,$F7,$BF,$50,$F0,$2F,$83,$2B,$FD,$8B,$39,$6F
	.dc.b	$CD,$70,$F7,$A4,$60,$FF,$65,$FD,$43,$E5,$FF,$84,$AE,$3A,$62,$33
	.dc.b	$C3,$AF,$8F,$97,$CA,$3F,$88,$FD,$43,$E6,$2F,$CC,$4A,$FF,$6C,$CE
	.dc.b	$5B,$F1,$9C,$3D,$59,$18,$3F,$DB,$FE,$03,$E1,$7F,$8E,$BF,$B9,$8F
	.dc.b	$F4,$6B,$7D,$2F,$F5,$DB,$B8,$E3,$E4,$F0,$F7,$E4,$60,$FF,$75,$FD
	.dc.b	$47,$E5,$DB,$E1,$7F,$2A,$83,$6F,$1C,$59,$F0,$65,$F2,$8E,$1F,$2D
	.dc.b	$C7,$E5,$A4,$F6,$8F,$E6,$BF,$51,$F8,$E8,$FF,$7D,$BF,$D1,$B5,$FF
	.dc.b	$69,$B7,$D6,$FE,$7F,$28,$E0,$FF,$AB,$3E,$A3,$FB,$7B,$FD,$12,$BF
	.dc.b	$DB,$70,$F8,$7F,$E8,$6D,$9D,$5F,$48,$E3,$1F,$EB,$6E,$EF,$7B,$A3
	.dc.b	$FA,$7B,$FB,$37,$5E,$4F,$F4,$E8,$DC,$71,$F0,$F1,$F5,$12,$39,$09
	.dc.b	$FF,$5F,$38,$5A,$F3,$F0,$E4,$E6,$3F,$AD,$FD,$47,$C3,$8F,$EC,$2F
	.dc.b	$F6,$07,$FD,$9A,$DD,$7F,$DF,$6E,$E1,$F5,$1F,$D1,$5E,$3E,$CF,$43
	.dc.b	$FB,$3C,$66,$37,$BE,$3D,$79,$1E,$A9,$F9,$4F,$FB,$71,$74,$F2,$E9
	.dc.b	$93,$B8,$EB,$EB,$76,$64,$73,$8F,$0F,$8D,$E2,$F1,$A4,$F7,$8F,$2F
	.dc.b	$A9,$A8,$C1,$E3,$7F,$BF,$FD,$3B,$E3,$3F,$C0,$3F,$50,$F8,$0A,$6A
	.dc.b	$37,$DD,$70,$FA,$B7,$C5,$B1,$FE,$0A,$F9,$97,$8F,$FE,$0C,$F4,$9A
	.dc.b	$53,$51,$BF,$26,$E1,$F8,$0B,$E2,$D8,$FF,$0A,$FD,$47,$E2,$FA,$1E
	.dc.b	$07,$E7,$B2,$73,$77,$37,$DF,$C3,$C5,$B9,$7E,$33,$1E,$58,$7E,$99
	.dc.b	$FC,$9E,$3D,$27,$FC,$4F,$1F,$BD,$51,$95,$1C,$27,$95,$79,$BC,$5F
	.dc.b	$3F,$0F,$C1,$E7,$98,$78,$63,$D7,$C3,$A8,$C1,$E5,$7D,$FE,$2F,$82
	.dc.b	$3F,$CA,$07,$81,$C3,$FE,$53,$CB,$F9,$86,$7A,$8D,$36,$3F,$9B,$7D
	.dc.b	$45,$F3,$8D,$25,$7F,$96,$FE,$87,$FC,$B8,$9A,$47,$B7,$C6,$FC,$7C
	.dc.b	$3E,$24,$7F,$9A,$37,$FD,$CB,$FE,$50,$AF,$F2,$8F,$D4,$39,$7F,$CD
	.dc.b	$58,$71,$BF,$F3,$67,$35,$C3,$F8,$CB,$5F,$F2,$CC,$D7,$F1,$B3,$DB
	.dc.b	$9B,$06,$F9,$9F,$2F,$89,$F1,$FC,$FF,$13,$17,$F7,$23,$FC,$F9,$1D
	.dc.b	$7E,$2A,$47,$03,$D6,$F8,$F6,$FF,$7E,$D7,$E1,$39,$BC,$1E,$1E,$5D
	.dc.b	$73,$4C,$3F,$8D,$DF,$03,$FD,$53,$7F,$E7,$4F,$A8,$F9,$0F,$8C,$C6
	.dc.b	$DE,$BF,$07,$E1,$4E,$AB,$F2,$4F,$C2,$AE,$01,$C4,$6A,$B8,$F0,$72
	.dc.b	$F9,$E9,$8F,$F2,$EB,$FF,$45,$B6,$3B,$FB,$B2,$30,$6D,$F1,$FA,$72
	.dc.b	$39,$7C,$75,$CF,$C7,$83,$D5,$F4,$99,$D3,$D2,$FB,$3E,$36,$61,$E3
	.dc.b	$FA,$6B,$87,$B3,$E5,$3E,$DC,$FF,$44,$6C,$67,$DE,$FC,$34,$63,$1F
	.dc.b	$DB,$2F,$2B,$26,$BC,$7B,$87,$F4,$EF,$A8,$7C,$BF,$E3,$DB,$19,$F0
	.dc.b	$B1,$39,$ED,$93,$C7,$DD,$BA,$FD,$2F,$A8,$77,$7E,$65,$B5,$FD,$71
	.dc.b	$FF,$0F,$36,$35,$C1,$23,$5C,$56,$FE,$EF,$D4,$7F,$2A,$9F,$EC,$AD
	.dc.b	$8C,$F5,$FE,$9E,$31,$C5,$FC,$5E,$BF,$C6,$C7,$F6,$C2,$D7,$E2,$1F
	.dc.b	$F1,$C1,$9D,$BF,$61,$F5,$1F,$EC,$88,$FE,$E6,$D9,$E1,$EE,$E9,$DF
	.dc.b	$87,$2B,$7D,$B6,$7F,$2B,$7F,$83,$9E,$7A,$76,$EB,$EE,$FE,$FF,$C6
	.dc.b	$FF,$00,$4F,$F8,$09,$B7,$F9,$E7,$FD,$F5,$19,$F6,$78,$79,$92,$38
	.dc.b	$7A,$96,$FF,$4B,$F5,$1F,$B8,$27,$9F,$FC,$35,$C7,$C2,$D3,$1E,$ED
	.dc.b	$C3,$FC,$27,$2E,$1E,$07,$1F,$6E,$CF,$F8,$63,$C7,$FF,$41,$B7,$C6
	.dc.b	$FE,$35,$0F,$1E,$C7,$F1,$DF,$A8,$FF,$68,$57,$F8,$81,$BA,$FE,$F1
	.dc.b	$FE,$BF,$19,$F9,$6F,$E4,$9F,$50,$F8,$6A,$7F,$27,$FA,$7D,$E1,$FC
	.dc.b	$12,$75,$7F,$E2,$C5,$E1,$72,$38,$C6,$5E,$BE,$1C,$8C,$1C,$5F,$0F
	.dc.b	$0C,$8F,$F1,$BE,$1F,$E6,$F7,$B4,$79,$B8,$A4,$60,$F1,$3F,$CF,$7E
	.dc.b	$A3,$F1,$9D,$FF,$DB,$5B,$3F,$E5,$AB,$FB,$F9,$FF,$1E,$9C,$A0,$FF
	.dc.b	$25,$9F,$F3,$0B,$63,$4F,$07,$7C,$3E,$1F,$0E,$57,$AB,$A6,$D7,$7F
	.dc.b	$47,$CC,$63,$2F,$33,$E9,$F4,$E5,$FF,$96,$32,$4E,$2F,$D1,$F3,$BE
	.dc.b	$F3,$EB,$7C,$A5,$BE,$1C,$FF,$0C,$C5,$7B,$DF,$8D,$E0,$49,$F9,$FF
	.dc.b	$1B,$6B,$E2,$78,$3F,$2C,$C6,$BF,$2D,$F9,$DF,$1F,$2F,$EB,$FB,$D6
	.dc.b	$FE,$67,$C0,$7D,$33,$18,$7D,$37,$EB,$7D,$44,$BF,$74,$5A,$FE,$A3
	.dc.b	$F4,$BF,$98,$BF,$F1,$33,$6B,$FD,$27,$F5,$0F,$97,$FE,$F9,$39,$BF
	.dc.b	$B4,$FE,$25,$E7,$F1,$FA,$F0,$36,$F0,$F5,$E6,$3F,$E8,$6F,$A8,$7D
	.dc.b	$46,$30,$F1,$FC,$07,$63,$1A,$F9,$1F,$0F,$1C,$C3,$F3,$3E,$CD,$BF
	.dc.b	$5F,$17,$98,$C6,$1E,$C7,$D3,$F0,$12,$FF,$D2,$FB,$FF,$94,$9B,$1E
	.dc.b	$EF,$89,$91,$E7,$EA,$B8,$FC,$8F,$A8,$F8,$66,$32,$F9,$2F,$C6,$E3
	.dc.b	$97,$FE,$A4,$F5,$DE,$2E,$30,$FF,$00,$F2,$3D,$DF,$F1,$43,$6F,$CB
	.dc.b	$86,$27,$FC,$59,$F4,$CF,$89,$A7,$F8,$BF,$E5,$FF,$97,$EB,$FE,$5F
	.dc.b	$9D,$61,$FE,$B4,$6E,$5F,$E5,$46,$BF,$ED,$00,$E1,$FF,$69,$B5,$FD
	.dc.b	$C7,$D3,$BE,$6F,$FC,$93,$18,$6A,$8D,$57,$F9,$75,$B1,$FE,$64,$7C
	.dc.b	$77,$FE,$75,$6C,$FF,$B8,$9D,$C6,$71,$F8,$3F,$0F,$2B,$FD,$9F,$97
	.dc.b	$F8,$25,$F3,$C6,$70,$FF,$16,$E5,$FE,$40,$AD,$47,$F1,$18,$FF,$21
	.dc.b	$9F,$DB,$30,$F9,$1E,$EF,$CB,$E1,$F9,$EC,$C7,$E5,$39,$69,$2F,$BC
	.dc.b	$65,$B1,$9A,$EA,$FA,$2B,$A4,$E8,$93,$FE,$97,$C1,$3A,$8E,$33,$AA
	.dc.b	$FE,$54,$AB,$71,$DC,$FB,$37,$C5,$DC,$6A,$79,$1F,$0E,$74,$FF,$35
	.dc.b	$8D,$57,$80,$95,$D0,$F5,$1D,$E7,$93,$5F,$25,$F4,$7F,$1B,$76,$FA
	.dc.b	$D7,$39,$F1,$9A,$7A,$B8,$C3,$EB,$46,$F0,$F8,$D8,$75,$9C,$D7,$88
	.dc.b	$FF,$36,$DF,$67,$15,$EA,$6D,$FC,$4B,$E5,$87,$95,$B1,$EB,$BE,$55
	.dc.b	$E5,$3C,$95,$F7,$EF,$A4,$64,$FD,$66,$1F,$E3,$86,$CF,$E9,$A3,$11
	.dc.b	$F0,$4C,$DF,$FB,$0E,$33,$1E,$B3,$27,$E7,$3E,$07,$E0,$12,$7E,$84
	.dc.b	$FF,$98,$A7,$E4,$4E,$5B,$3A,$8C,$5F,$A2,$B8,$36,$33,$3F,$3C,$FE
	.dc.b	$06,$E2,$BE,$49,$FA,$B7,$0F,$C5,$9F,$E2,$51,$8A,$FA,$B3,$FC,$3A
	.dc.b	$31,$7F,$C7,$1B,$7F,$C5,$DB,$3E,$F8,$F4,$11,$F5,$92,$FD,$11,$96
	.dc.b	$C6,$63,$D4,$F9,$DD,$18,$7C,$8F,$F0,$CE,$2F,$E5,$87,$F2,$DF,$83
	.dc.b	$B1,$93,$FD,$F1,$92,$7F,$90,$B6,$33,$18,$F9,$F9,$1C,$1B,$7F,$CC
	.dc.b	$47,$D3,$BE,$9F,$FA,$24,$66,$31,$C7,$B8,$DF,$CF,$3D,$1C,$56,$A7
	.dc.b	$A6,$7F,$EA,$B7,$CF,$DD,$C5,$FE,$10,$D8,$D4,$70,$E7,$C1,$1F,$7B
	.dc.b	$C7,$E9,$9F,$4F,$FC,$4B,$85,$BF,$F1,$4F,$2B,$FC,$D1,$87,$F8,$FE
	.dc.b	$72,$FF,$90,$63,$2F,$FB,$8A,$F9,$BB,$AF,$E3,$B0,$ED,$ED,$18,$C2
	.dc.b	$D7,$F9,$2E,$B8,$BC,$EB,$2A,$1F,$19,$8F,$A1,$7B,$14,$DC,$EA,$BC
	.dc.b	$47,$5F,$3F,$37,$23,$27,$58,$E3,$F1,$9E,$0F,$A7,$FF,$25,$CE,$9A
	.dc.b	$FB,$E3,$BE,$BC,$23,$9E,$09,$CC,$63,$CF,$AF,$C7,$AF,$3C,$7B,$BE
	.dc.b	$9F,$E6,$CB,$82,$FD,$27,$E2,$E3,$CE,$D7,$5C,$3F,$D2,$AB,$8D,$72
	.dc.b	$5E,$87,$59,$95,$F0,$27,$2D,$F6,$7E,$7F,$D3,$BE,$6E,$9E,$58,$F3
	.dc.b	$7A,$76,$B8,$1C,$DF,$57,$7F,$D8,$FA,$77,$BF,$4C,$D7,$87,$D2,$AF
	.dc.b	$A8,$FF,$46,$66,$BF,$89,$3F,$17,$D2,$B5,$C4,$F5,$3F,$27,$E2,$BE
	.dc.b	$C7,$D3,$C9,$F7,$47,$CD,$F4,$B5,$18,$3E,$13,$F4,$7C,$1F,$C7,$F2
	.dc.b	$3D,$CA,$66,$BE,$47,$83,$5F,$40,$D7,$F6,$3C,$DF,$46,$7F,$33,$E2
	.dc.b	$FF,$8B,$EB,$F1,$FE,$03,$CE,$FC,$D1,$AF,$47,$7E,$85,$AF,$CA,$DF
	.dc.b	$17,$BB,$FE,$49,$73,$F1,$5C,$1D,$CF,$82,$3C,$3F,$35,$FA,$7F,$A7
	.dc.b	$95,$D2,$3C,$BF,$37,$A8,$C1,$E0,$7D,$3F,$0D,$FB,$2F,$09,$F1,$54
	.dc.b	$CD,$79,$BC,$9A,$F9,$AB,$E0,$C7,$C2,$F6,$7F,$77,$E6,$BB,$29,$9B
	.dc.b	$F7,$F7,$57,$DF,$5F,$2A,$3E,$5F,$D0,$59,$D9,$F1,$9F,$E0,$FF,$4E
	.dc.b	$F0,$AF,$87,$2B,$F7,$F3,$96,$FA,$BC,$3D,$59,$18,$3F,$82,$FD,$3B
	.dc.b	$D8,$FE,$26,$B8,$E9,$88,$CF,$0E,$BE,$32,$5F,$28,$FE,$37,$F4,$EF
	.dc.b	$96,$BF,$2D,$2B,$F8,$9C,$E5,$BF,$11,$C3,$D5,$91,$83,$F8,$CF,$80
	.dc.b	$F8,$1F,$E4,$EF,$EF,$23,$F2,$AD,$F1,$3F,$86,$BB,$8E,$3E,$57,$0F
	.dc.b	$94,$91,$83,$F9,$2F,$D3,$FE,$5D,$BE,$17,$F3,$48,$36,$F1,$C5,$9F
	.dc.b	$06,$5F,$28,$E1,$F2,$9C,$7E,$52,$4F,$68,$FE,$7F,$F4,$FF,$8E,$8F
	.dc.b	$E6,$AF,$F4,$ED,$7F,$92,$B7,$D6,$FF,$34,$BB,$9F,$D9,$FD,$3B,$6F
	.dc.b	$E9,$2B,$E2,$29,$88,$FE,$9B,$DD,$EA,$AF,$E8,$6F,$EB,$5F,$DE,$B7
	.dc.b	$1F,$E6,$AE,$E3,$3E,$17,$2F,$A7,$91,$CA,$3A,$EB,$DB,$C2,$93,$DA
	.dc.b	$3F,$AE,$FD,$3F,$AE,$01,$FE,$D0,$FF,$56,$6B,$FC,$41,$E0,$BE,$9F
	.dc.b	$FA,$0E,$FF,$E8,$C9,$FE,$75,$3C,$7D,$AF,$ED,$75,$A8,$EE,$FC,$57
	.dc.b	$FB,$1A,$F6,$36,$70,$DF,$8F,$47,$2E,$09,$7C,$23,$AF,$AF,$A8,$C1
	.dc.b	$D1,$F0,$F8,$BE,$2F,$16,$4F,$82,$3C,$BE,$BE,$A3,$07,$91,$FF,$01
	.dc.b	$FD,$33,$E8,$FF,$C0,$BF,$4E,$F8,$0A,$6A,$37,$DB,$70,$FE,$FF,$7C
	.dc.b	$5B,$1F,$E0,$EF,$98,$78,$DF,$E1,$0F,$45,$A5,$35,$1B,$F2,$2E,$1F
	.dc.b	$E0,$AB,$E2,$D8,$FF,$0C,$FD,$3F,$E2,$7A,$3F,$D7,$F8,$FC,$DA,$FF
	.dc.b	$08,$B5,$FE,$97,$3B,$7F,$06,$FE,$3F,$BB,$FC,$0A,$E6,$0D,$35,$5F
	.dc.b	$EC,$C7,$F7,$11,$D3,$C4,$F1,$9E,$4F,$13,$CB,$BF,$BF,$96,$57,$F8
	.dc.b	$92,$FE,$33,$E4,$DF,$6E,$6B,$DB,$1E,$EF,$13,$FC,$31,$3F,$E6,$3A
	.dc.b	$FE,$4D,$9F,$F9,$7D,$B1,$99,$DB,$FE,$E2,$3D,$7E,$63,$E9,$F8,$36
	.dc.b	$3A,$35,$E1,$5F,$CD,$67,$FC,$C9,$0F,$A0,$A4,$7F,$CD,$0F,$F9,$A7
	.dc.b	$E3,$3F,$CB,$78,$FF,$8C,$27,$FC,$E5,$BF,$F9,$11,$FD,$9D,$7E,$D3
	.dc.b	$E9,$DC,$7E,$7C,$38,$DF,$A7,$92,$E1,$FE,$FC,$6B,$FB,$C9,$AF,$E3
	.dc.b	$A7,$B7,$26,$0D,$F2,$3E,$57,$13,$E3,$F9,$DE,$26,$2F,$FC,$22,$1E
	.dc.b	$BF,$39,$27,$81,$EB,$7C,$7B,$7F,$84,$63,$FC,$C2,$DA,$FF,$0D,$E1
	.dc.b	$FC,$9E,$73,$CD,$C3,$96,$46,$0E,$2F,$BB,$E9,$FE,$4D,$FF,$89,$1B
	.dc.b	$87,$F9,$73,$0F,$F5,$64,$E7,$5E,$1B,$4C,$7F,$CE,$78,$70,$7F,$91
	.dc.b	$D2,$3E,$43,$D7,$E7,$7C,$07,$87,$E4,$2D,$72,$F7,$7C,$86,$2B,$C8
	.dc.b	$F9,$FD,$39,$3E,$0F,$90,$B5,$E9,$F5,$7D,$86,$36,$FF,$3A,$E6,$8F
	.dc.b	$81,$F1,$D3,$0F,$4F,$D2,$DC,$3F,$C8,$BF,$29,$F6,$E7,$FD,$1E,$D8
	.dc.b	$CF,$BD,$F8,$68,$C6,$5F,$E8,$25,$E3,$E4,$D7,$89,$70,$FF,$4C,$FD
	.dc.b	$3F,$FC,$00,$7E,$3B,$71,$FE,$96,$FF,$BE,$0C,$27,$FD,$53,$1C,$B9
	.dc.b	$7F,$2F,$0F,$72,$C7,$E6,$BE,$9F,$F1,$D5,$FE,$A8,$6D,$FF,$98,$9F
	.dc.b	$F0,$E4,$67,$AF,$86,$D2,$38,$70,$5B,$FF,$34,$FD,$3F,$F2,$E9,$FF
	.dc.b	$59,$B6,$33,$E6,$7E,$AA,$31,$C1,$FE,$A6,$AF,$F1,$A1,$FE,$BA,$2D
	.dc.b	$7F,$9D,$1F,$F1,$B1,$9F,$17,$F6,$7F,$4F,$FD,$BA,$3F,$D8,$2D,$9E
	.dc.b	$1E,$E6,$9D,$F8,$52,$B7,$DB,$67,$FD,$03,$7F,$83,$9E,$7A,$76,$EA
	.dc.b	$EE,$FF,$02,$F1,$5E,$D5,$31,$F3,$36,$FF,$DA,$72,$F3,$38,$1C,$7D
	.dc.b	$2B,$3F,$ED,$8F,$A7,$FD,$C7,$0F,$FB,$21,$B1,$FA,$87,$FC,$6A,$EF
	.dc.b	$C6,$FE,$27,$1F,$EB,$48,$E5,$F0,$B8,$F7,$32,$EB,$8E,$4F,$7F,$6E
	.dc.b	$F1,$7F,$A8,$3C,$8F,$F4,$1C,$FF,$BA,$9B,$19,$F2,$3F,$90,$C6,$34
	.dc.b	$FF,$6E,$FD,$3F,$FB,$52,$FF,$DE,$2D,$BF,$F5,$5A,$8F,$E5,$2E,$67
	.dc.b	$FD,$CB,$F4,$EF,$85,$FE,$5A,$DA,$FF,$7D,$FD,$3F,$17,$FD,$FB,$19
	.dc.b	$C3,$FD,$6E,$FE,$76,$33,$E1,$72,$BE,$6F,$4F,$0A,$4E,$8E,$6F,$83
	.dc.b	$82,$47,$83,$E1,$6E,$3D,$B8,$7C,$A6,$34,$F4,$BF,$D1,$3E,$9F,$F2
	.dc.b	$11,$FF,$00,$1B,$7F,$C8,$35,$74,$6F,$0E,$1C,$12,$BF,$C8,$F8,$71
	.dc.b	$9C,$1F,$ED,$4C,$3F,$6B,$1D,$7C,$3F,$EA,$78,$FE,$4A,$F8,$9F,$CB
	.dc.b	$AB,$6D,$F4,$3C,$C7,$9E,$AC,$70,$C3,$FC,$AF,$5F,$D9,$23,$10,$8C
	.dc.b	$3E,$3A,$F8,$1E,$97,$C1,$F4,$25,$FD,$8F,$94,$B7,$E9,$E6,$F8,$C6
	.dc.b	$34,$F8,$CF,$C9,$F0,$25,$FE,$03,$E2,$ED,$FC,$1F,$07,$E7,$18,$C3
	.dc.b	$E7,$3F,$47,$E3,$65,$FE,$23,$E0,$6D,$FC,$9F,$7B,$EB,$18,$C3,$EB
	.dc.b	$3F,$67,$E9,$E5,$FC,$0E,$1F,$F2,$93,$63,$FC,$03,$F4,$9F,$99,$8F
	.dc.b	$EB,$6D,$C3,$FC,$FF,$F4,$EF,$96,$DF,$CE,$F0,$EB,$98,$7C,$CC,$38
	.dc.b	$60,$7F,$D8,$11,$9F,$5B,$B9,$FE,$10,$FA,$77,$D3,$E2,$79,$7B,$DE
	.dc.b	$43,$18,$38,$3F,$60,$FF,$A4,$B1,$FE,$A7,$5E,$8E,$4F,$59,$F8,$AF
	.dc.b	$62,$C7,$97,$8F,$DE,$62,$FA,$47,$C0,$8F,$87,$F1,$12,$38,$1E,$D7
	.dc.b	$E3,$FC,$44,$8F,$C5,$EA,$EB,$C6,$1D,$0F,$37,$D3,$7C,$E3,$3A,$FA
	.dc.b	$A3,$E3,$47,$D2,$E3,$98,$7E,$5B,$D7,$78,$B8,$BE,$87,$FB,$AF,$DC
	.dc.b	$FF,$1C,$F2,$9D,$65,$D6,$31,$87,$E8,$3E,$99,$F3,$1F,$E4,$BC,$3F
	.dc.b	$D7,$71,$89,$FF,$2F,$46,$27,$4A,$74,$E9,$B7,$EB,$F9,$7F,$F5,$A5
	.dc.b	$F7,$BF,$F2,$6C,$62,$FD,$3F,$4F,$C2,$BF,$C5,$2D,$8F,$F6,$6A,$F9
	.dc.b	$BF,$F2,$AB,$98,$3E,$27,$FC,$E8,$D7,$9D,$FA,$23,$11,$EE,$F5,$3D
	.dc.b	$79,$7F,$92,$E3,$FC,$F8,$D7,$FC,$76,$D7,$86,$1F,$27,$E1,$BF,$CC
	.dc.b	$F3,$FD,$6A,$7E,$4C,$63,$E9,$E5,$7C,$FF,$7A,$D7,$D8,$BF,$F0,$E9
	.dc.b	$23,$3A,$FE,$0E,$33,$7D,$A7,$37,$E4,$8B,$7F,$62,$6B,$F5,$62,$D7
	.dc.b	$F8,$3A,$FA,$46,$0E,$97,$F8,$37,$0F,$9C,$C1,$2B,$D1,$1C,$1B,$19
	.dc.b	$9E,$AF,$18,$C9,$23,$2F,$C1,$BE,$EE,$67,$E7,$C6,$27,$67,$C3,$71
	.dc.b	$1F,$04,$ED,$1A,$71,$3C,$CF,$88,$F9,$4C,$C6,$E3,$26,$CF,$91,$1B
	.dc.b	$73,$1B,$72,$6E,$FE,$64,$F4,$24,$7E,$AA,$FE,$35,$71,$A4,$7E,$40
	.dc.b	$95,$FA,$EA,$EA,$7E,$4A,$32,$FE,$B5,$AE,$63,$FA,$CE,$AF,$63,$F0
	.dc.b	$2D,$8F,$54,$A3,$0F,$BD,$B8,$7B,$5B,$3E,$A9,$E1,$5D,$09,$1D,$27
	.dc.b	$73,$EF,$8F,$1C,$EA,$3B,$36,$BB,$36,$7B,$57,$2A,$F9,$B5,$FC,$26
	.dc.b	$B5,$1D,$4F,$E3,$F3,$BC,$3F,$D0,$AD,$C3,$F8,$6D,$70,$AE,$78,$7F
	.dc.b	$AD,$2B,$4F,$98,$F5,$4E,$EF,$9B,$6F,$9C,$E8,$97,$EC,$6E,$1E,$C6
	.dc.b	$9E,$CA,$FE,$0B,$E9,$C8,$FA,$FC,$76,$FF,$81,$4F,$F1,$22,$4F,$E0
	.dc.b	$F1,$FF,$21,$4F,$FB,$30,$7D,$3F,$F8,$BB,$8F,$F1,$0D,$8F,$E6,$EB
	.dc.b	$E7,$5E,$8E,$BD,$2F,$E8,$95,$FC,$F3,$77,$67,$E4,$63,$F0,$CD,$8C
	.dc.b	$C7,$E5,$5A,$F1,$E9,$D1,$1D,$3C,$FF,$19,$7A,$B4,$93,$F9,$F8,$CB
	.dc.b	$FA,$1F,$A6,$7D,$47,$FB,$AB,$93,$FB,$27,$9C,$F4,$71,$C7,$FE,$EE
	.dc.b	$8C,$61,$FE,$3A,$C3,$8B,$7C,$1D,$A1,$1A,$7F,$B4,$B7,$0F,$EE,$4F
	.dc.b	$F7,$98,$E9,$F1,$7C,$97,$D0,$FA,$1F,$EE,$2D,$AE,$33,$BA,$7A,$76
	.dc.b	$25,$F9,$E3,$5F,$17,$D7,$EA,$1E,$C5,$33,$5E,$DD,$2B,$E8,$AE,$D1
	.dc.b	$DF,$D5,$B3,$B3,$E2,$3E,$9F,$A3,$FF,$07,$69,$FE,$0F,$FA,$4F,$F3
	.dc.b	$BD,$7F,$98,$9A,$FF,$83,$27,$7C,$1F,$26,$10,$F5,$3A,$4C,$9F,$F3
	.dc.b	$5C,$E5,$BF,$0D,$F2,$7E,$A1,$F3,$74,$F2,$47,$97,$D7,$B5,$C0,$E6
	.dc.b	$FA,$7E,$9F,$E9,$7D,$43,$E0,$A9,$9A,$F0,$FA,$F5,$F5,$57,$C1,$8F
	.dc.b	$85,$EB,$59,$D9,$E8,$7E,$37,$C4,$FD,$3F,$A8,$93,$EC,$8F,$95,$E1
	.dc.b	$EA,$30,$7C,$07,$E7,$7B,$3F,$87,$E3,$7C,$0D,$33,$5F,$1B,$C4,$AF
	.dc.b	$CF,$AF,$AD,$1F,$8F,$F2,$3F,$C4,$13,$FE,$96,$F0,$1E,$97,$E4,$DB
	.dc.b	$1F,$E9,$A9,$DB,$FE,$44,$7F,$C7,$AE,$7D,$5E,$87,$5F,$57,$B7,$CD
	.dc.b	$D9,$E5,$18,$7A,$7C,$DF,$7F,$A8,$91,$83,$C1,$F3,$9E,$79,$91,$83
	.dc.b	$C9,$F1,$3F,$A9,$F0,$9F,$33,$4F,$74,$7B,$FC,$EB,$5C,$0F,$63,$F0
	.dc.b	$FD,$DF,$D9,$7A,$0E,$AA,$66,$BE,$17,$65,$7E,$15,$7C,$B8,$F9,$9F
	.dc.b	$43,$67,$67,$C9,$7F,$7D,$F5,$0F,$02,$F8,$32,$BF,$D8,$F3,$96,$FC
	.dc.b	$D7,$0F,$6A,$46,$0F,$F6,$6F,$D4,$3D,$EF,$E1,$4B,$92,$98,$8C,$F0
	.dc.b	$EB,$DC,$97,$CA,$3F,$89,$7D,$43,$E6,$2F,$CC,$4A,$FF,$6D,$CE,$5B
	.dc.b	$F1,$DC,$3D,$59,$18,$3F,$DC,$3E,$0B,$E1,$7F,$8F,$3F,$B9,$8F,$F4
	.dc.b	$6B,$7D,$2F,$F4,$B3,$B8,$E3,$E4,$F0,$F8,$29,$18,$3F,$DD,$BF,$51
	.dc.b	$F9,$67,$FD,$2E,$BC,$3F,$E5,$8D,$8D,$DE,$38,$B3,$E0,$C9,$E7,$1C
	.dc.b	$3E,$5B,$8F,$CB,$49,$EF,$1F,$CD,$FE,$A3,$F1,$F1,$FE,$FE,$7F,$A3
	.dc.b	$BF,$EA,$86,$EB,$FE,$B1,$77,$1F,$EA,$CF,$A8,$FF,$16,$3F,$D2,$67
	.dc.b	$FD,$6D,$0F,$8D,$FE,$88,$D9,$D7,$B1,$FC,$9E,$38,$C7,$FA,$DB,$BB
	.dc.b	$DC,$DF,$27,$FA,$64,$7E,$DA,$35,$C1,$8F,$0F,$3F,$51,$27,$A9,$FD
	.dc.b	$6E,$BF,$D8,$6E,$02,$DB,$AF,$87,$27,$53,$97,$87,$D4,$7C,$2D,$3F
	.dc.b	$9F,$3F,$D8,$63,$FD,$9A,$DD,$BF,$DA,$CE,$E3,$FB,$67,$D4,$7F,$46
	.dc.b	$7F,$B7,$47,$F3,$E7,$AF,$B3,$FD,$A6,$75,$07,$E4,$3F,$ED,$A5,$EB
	.dc.b	$EC,$E1,$C3,$C7,$A7,$97,$4C,$BF,$81,$69,$ED,$EB,$6E,$73,$18,$77
	.dc.b	$E3,$78,$FC,$69,$18,$39,$FA,$9E,$79,$91,$83,$E7,$7D,$3B,$E3,$BF
	.dc.b	$C0,$9F,$50,$F8,$0A,$7F,$81,$E7,$55,$C3,$BF,$17,$23,$07,$D9,$F9
	.dc.b	$97,$8F,$FE,$0F,$F4,$99,$D3,$FC,$25,$3A,$AE,$1E,$4E,$2E,$46,$0F
	.dc.b	$C2,$7D,$47,$E3,$7A,$1D,$3F,$DE,$F2,$F9,$F5,$F8,$96,$A9,$DB,$FE
	.dc.b	$29,$AE,$55,$FE,$03,$7F,$37,$3A,$94,$69,$E1,$E0,$7F,$8A,$63,$75
	.dc.b	$C7,$C5,$C7,$83,$2F,$CF,$1B,$DB,$E1,$72,$C9,$F5,$C7,$3E,$0D,$46
	.dc.b	$0F,$7B,$DF,$C5,$F5,$47,$F9,$A4,$F2,$B8,$7F,$CD,$99,$7F,$33,$D3
	.dc.b	$FD,$64,$D8,$E0,$FF,$39,$FA,$8C,$F9,$C6,$92,$7F,$CD,$5F,$43,$FE
	.dc.b	$6B,$4D,$23,$FE,$BF,$F8,$FF,$9B,$87,$C9,$8F,$F3,$BE,$FF,$BB,$7F
	.dc.b	$CB,$15,$FE,$59,$FA,$87,$2F,$FA,$0F,$0E,$37,$FE,$85,$E4,$B8,$7F
	.dc.b	$8D,$DA,$FF,$98,$66,$BF,$8E,$9E,$DC,$B8,$37,$CA,$F9,$9C,$4F,$8F
	.dc.b	$E7,$B8,$98,$BF,$C3,$47,$FA,$56,$3A,$FC,$54,$8E,$07,$AD,$F1,$ED
	.dc.b	$FE,$1A,$8F,$E3,$0D,$AE,$67,$F9,$DF,$0F,$E5,$D3,$AE,$6E,$3C,$B2
	.dc.b	$30,$72,$7D,$FF,$51,$F2,$AF,$98,$71,$87,$E5,$F0,$77,$F0,$DB,$8E
	.dc.b	$F3,$2F,$E5,$98,$70,$7F,$CC,$4F,$C9,$C7,$AB,$FE,$95,$8D,$FC,$87
	.dc.b	$8B,$E4,$3C,$6F,$93,$DE,$93,$A3,$AB,$E8,$F5,$24,$78,$3E,$3E,$E3
	.dc.b	$DB,$D6,$F6,$98,$DF,$F6,$19,$A3,$E0,$FC,$74,$C3,$D3,$F4,$D7,$0F
	.dc.b	$87,$F2,$DF,$76,$7F,$A4,$36,$33,$EF,$FE,$22,$31,$97,$EF,$57,$93
	.dc.b	$93,$5E,$35,$C3,$FA,$97,$D4,$3E,$67,$F2,$2D,$8C,$F8,$58,$9C,$F4
	.dc.b	$49,$E3,$EF,$5D,$7E,$CF,$D4,$3B,$BF,$36,$DA,$FE,$C0,$FF,$88,$DB
	.dc.b	$1A,$E0,$91,$AE,$2B,$7F,$84,$FA,$8F,$E5,$B3,$FD,$A5,$B1,$9E,$CF
	.dc.b	$D4,$C6,$38,$BF,$8F,$57,$F8,$EC,$FE,$E0,$5A,$FC,$7B,$FE,$3D,$33
	.dc.b	$B7,$EC,$BE,$A3,$FD,$A5,$1F,$DD,$DB,$3C,$3D,$ED,$3B,$F0,$E5,$6F
	.dc.b	$BA,$CF,$E7,$2F,$F0,$13,$CF,$4E,$DD,$7D,$DF,$E0,$3E,$37,$F8,$12
	.dc.b	$7F,$C0,$CD,$BF,$D2,$AF,$9A,$DF,$B3,$C3,$CD,$97,$C3,$D5,$B8,$7E
	.dc.b	$B3,$EA,$3F,$71,$C3,$FD,$05,$B5,$FE,$C7,$7F,$C7,$CF,$F8,$55,$CC
	.dc.b	$FF,$BC,$DF,$F5,$D4,$67,$E1,$FF,$8B,$C3,$DF,$B1,$8E,$29,$5C,$FD
	.dc.b	$EB,$87,$EF,$BC,$8F,$F4,$6C,$FF,$88,$DB,$19,$F1,$FF,$91,$C6,$33
	.dc.b	$FE,$B1,$F5,$1F,$ED,$BB,$FF,$15,$36,$FF,$84,$BF,$DA,$23,$3F,$31
	.dc.b	$FC,$B3,$EA,$1F,$11,$4F,$E5,$FF,$51,$BC,$3F,$88,$CE,$AF,$FC,$6C
	.dc.b	$BC,$1E,$47,$18,$CB,$D7,$C2,$91,$83,$8B,$E1,$E1,$91,$FE,$41,$C3
	.dc.b	$FC,$F4,$F6,$8F,$37,$0C,$8C,$1E,$27,$FA,$2F,$D4,$7E,$4B,$6F,$EF
	.dc.b	$4D,$9F,$F8,$01,$7F,$85,$7F,$C9,$D1,$9C,$C1,$FE,$4C,$E1,$AE,$4E
	.dc.b	$FB,$F8,$3C,$29,$5E,$9E,$7B,$5F,$CE,$2D,$E8,$F9,$8C,$E5,$E6,$7D
	.dc.b	$3E,$9C,$C3,$CB,$16,$F9,$7A,$1E,$E3,$EA,$FC,$AD,$AF,$5F,$37,$C2
	.dc.b	$31,$5E,$17,$E2,$F7,$C9,$F9,$DF,$1D,$6B,$CD,$E0,$7C,$A3,$1A,$7C
	.dc.b	$A7,$E6,$FC,$84,$BF,$AD,$F0,$96,$FD,$DF,$01,$F4,$8C,$61,$F4,$9F
	.dc.b	$AB,$F5,$12,$FE,$8C,$5A,$FE,$BB,$F4,$DF,$97,$BF,$F1,$93,$6B,$EA
	.dc.b	$CB,$EA,$1F,$31,$FE,$04,$AD,$4F,$DA,$7F,$12,$F4,$F9,$0D,$78,$1B
	.dc.b	$78,$7A,$D3,$1F,$F4,$87,$D3,$BE,$A3,$18,$79,$3D,$FE,$C6,$35,$F2
	.dc.b	$BE,$1E,$39,$87,$E6,$7D,$8B,$7E,$CE,$2F,$31,$8C,$3D,$AF,$A7,$E2
	.dc.b	$E5,$FF,$A8,$77,$FF,$44,$B6,$3D,$DF,$15,$23,$CF,$D5,$71,$F9,$3F
	.dc.b	$51,$F0,$CC,$65,$F2,$9F,$8D,$C7,$2F,$FD,$5D,$EB,$7F,$15,$7C,$11
	.dc.b	$FE,$2E,$E3,$7B,$94,$E1,$8F,$F8,$4A,$33,$5F,$EB,$7F,$A7,$FE,$4B
	.dc.b	$3F,$E5,$56,$BF,$60,$38,$7E,$DB,$5F,$D8,$BC,$B2,$FF,$39,$7C,$C3
	.dc.b	$9B,$FC,$83,$39,$BF,$C1,$CE,$5F,$17,$C2,$68,$FF,$23,$C7,$F8,$02
	.dc.b	$3B,$C6,$5F,$F3,$56,$1E,$48,$FF,$2C,$3E,$26,$B9,$8D,$D3,$E4,$B2
	.dc.b	$D8,$CC,$60,$8F,$FB,$6F,$EA,$3D,$15,$EE,$9F,$F2,$DC,$62,$3E,$2F
	.dc.b	$97,$EE,$3F,$CB,$FF,$B3,$7C,$E8,$FF,$20,$E1,$F5,$DB,$3B,$C3,$FD
	.dc.b	$D1,$5B,$D3,$FD,$35,$7B,$6D,$6E,$FF,$C9,$75,$BB,$FA,$73,$9A,$DC
	.dc.b	$74,$DB,$0E,$8F,$C0,$4F,$9A,$33,$7F,$6A,$2A,$BE,$CE,$6B,$89,$1F
	.dc.b	$1D,$70,$C3,$D1,$3D,$E7,$2E,$F0,$F2,$C7,$C2,$8F,$31,$93,$84,$68
	.dc.b	$FF,$5E,$5F,$8A,$7E,$7C,$ED,$F2,$1D,$8E,$84,$9E,$51,$FC,$4A,$79
	.dc.b	$57,$9A,$B8,$D6,$9F,$E9,$73,$AC,$3D,$87,$C7,$BF,$04,$F8,$8E,$CF
	.dc.b	$5C,$36,$EA,$B9,$3C,$30,$F0,$57,$CD,$AD,$BC,$8F,$0B,$87,$F6,$51
	.dc.b	$F8,$96,$DF,$E6,$89,$5F,$9E,$6C,$7E,$ED,$B8,$7E,$33,$CC,$FA,$C4
	.dc.b	$AF,$23,$6B,$C4,$7F,$00,$71,$3F,$16,$31,$3C,$08,$F0,$BF,$F1,$5B
	.dc.b	$6B,$8C,$FD,$C3,$F8,$4B,$88,$FA,$31,$83,$9D,$77,$AE,$6D,$AE,$15
	.dc.b	$F4,$63,$D9,$5C,$0E,$D1,$FC,$52,$77,$5F,$40,$E7,$3A,$9F,$59,$E7
	.dc.b	$79,$3E,$43,$93,$87,$C0,$F9,$DE,$04,$AF,$34,$6E,$7C,$15,$FC,$13
	.dc.b	$EA,$BF,$8A,$E4,$69,$2B,$F8,$EF,$D5,$7F,$98,$F3,$BE,$CF,$F3,$CC
	.dc.b	$FF,$9A,$09,$7C,$FB,$CF,$DD,$FF,$26,$FA,$A7,$CC,$7F,$28,$C1,$18
	.dc.b	$75,$19,$8C,$47,$6D,$1F,$41,$C5,$B1,$FC,$C9,$AE,$E3,$5F,$29,$23
	.dc.b	$4D,$7D,$34,$F6,$BC,$04,$8F,$43,$E7,$FA,$97,$D5,$7F,$82,$B9,$1C
	.dc.b	$3F,$E0,$CF,$35,$E8,$66,$30,$DC,$7F,$C8,$B8,$72,$7F,$C8,$F7,$C9
	.dc.b	$CF,$83,$B5,$3B,$7F,$DF,$2D,$9E,$11,$A9,$E1,$2F,$8E,$E5,$F4,$6F
	.dc.b	$6A,$9A,$8D,$FA,$77,$0F,$1E,$7F,$E2,$6C,$39,$46,$0F,$4C,$F6,$F8
	.dc.b	$EF,$37,$D5,$7F,$95,$27,$41,$C3,$E8,$8F,$2E,$DC,$63,$76,$9C,$C6
	.dc.b	$3D,$3A,$FC,$0A,$F6,$C7,$C5,$FA,$AF,$83,$BF,$C2,$FA,$6F,$F5,$C5
	.dc.b	$F8,$9A,$A3,$E3,$CF,$0D,$3F,$C8,$6B,$D4,$E9,$12,$BD,$73,$96,$FB
	.dc.b	$9F,$57,$EA,$9F,$3F,$4F,$24,$79,$7C,$AB,$5C,$0E,$6F,$A7,$EA,$7E
	.dc.b	$F7,$D5,$3C,$6A,$66,$BC,$3E,$B5,$7D,$5A,$7A,$2F,$E1,$4F,$C3,$F5
	.dc.b	$6D,$70,$3D,$2F,$C7,$F8,$AF,$AF,$F5,$52,$7D,$B1,$F2,$FD,$5D,$46
	.dc.b	$0F,$82,$FC,$FF,$63,$F2,$9C,$8F,$7E,$99,$AF,$8F,$EA,$D7,$CE,$7F
	.dc.b	$A9,$39,$17,$F6,$67,$F3,$DE,$5F,$EE,$39,$FF,$37,$E0,$3D,$3F,$CC
	.dc.b	$9A,$F2,$37,$E4,$5C,$FF,$1F,$87,$1E,$4F,$F3,$EA,$9E,$BF,$3F,$C2
	.dc.b	$B7,$18,$78,$FC,$FF,$E9,$BE,$AA,$46,$0E,$BF,$41,$E6,$99,$18,$3C
	.dc.b	$3F,$13,$FB,$7F,$09,$F3,$94,$F6,$C7,$BB,$C9,$B5,$C0,$F4,$BF,$0B
	.dc.b	$D7,$FE,$01,$ED,$3B,$69,$9A,$F7,$F7,$D7,$E6,$9F,$CC,$2B,$E5,$C7
	.dc.b	$CC,$F4,$AC,$F0,$3E,$3B,$FC,$37,$EA,$BF,$CC,$91,$FB,$85,$E2,$64
	.dc.b	$D7,$89,$BF,$9C,$E1,$EF,$4B,$E1,$1F,$C5,$FE,$A9,$F3,$17,$92,$57
	.dc.b	$F0,$49,$CB,$78,$78,$7C,$84,$8C,$1F,$C2,$BE,$A9,$F3,$3F,$C8,$D7
	.dc.b	$CC,$D3,$11,$9F,$8D,$D7,$AB,$2F,$94,$7F,$2B,$F0,$5F,$0B,$7D,$29
	.dc.b	$5F,$C6,$27,$2D,$F3,$38,$7C,$BC,$8C,$1F,$C7,$BE,$AB,$F2,$CD,$F0
	.dc.b	$FF,$9C,$41,$B7,$8E,$2C,$F8,$32,$F9,$47,$0F,$97,$E3,$EF,$49,$ED
	.dc.b	$1F,$D0,$FE,$AB,$F1,$B1,$FC,$B9,$FE,$9E,$DF,$7B,$5E,$27,$F4,$9F
	.dc.b	$AA,$6D,$7E,$32,$5F,$F3,$69,$CB,$FC,$E3,$BF,$D0,$DF,$27,$FA,$94
	.dc.b	$7E,$AE,$35,$C1,$8F,$13,$3F,$17,$27,$94,$70,$DF,$8F,$89,$27,$B4
	.dc.b	$7F,$5D,$FA,$AF,$4C,$7E,$69,$7C,$16,$4D,$70,$7F,$63,$FA,$AF,$E7
	.dc.b	$14,$15,$FC,$C9,$FE,$A2,$D7,$F4,$B1,$C1,$2D,$FA,$7B,$52,$7F,$AC
	.dc.b	$CE,$5B,$D3,$E1,$E2,$91,$83,$7E,$CF,$96,$64,$60,$E7,$E3,$7A,$7C
	.dc.b	$79,$18,$3B,$FA,$DE,$D9,$91,$83,$FB,$5F,$D4,$BE,$9B,$FC,$09,$F5
	.dc.b	$4F,$82,$A7,$F8,$1E,$75,$5C,$3B,$F1,$72,$30,$7F,$76,$F9,$9F,$D5
	.dc.b	$BF,$DE,$3D,$36,$5F,$E1,$16,$C6,$FC,$9F,$F0,$A5,$7E,$D2,$30,$43
	.dc.b	$F8,$9B,$97,$F8,$AF,$D5,$7E,$25,$FE,$2F,$4E,$3E,$C7,$F8,$7E,$75
	.dc.b	$ED,$FF,$87,$1F,$C9,$D7,$2E,$FD,$1B,$EE,$E1,$E2,$D9,$FF,$08,$D7
	.dc.b	$0B,$E5,$3D,$75,$F6,$72,$4B,$DC,$78,$77,$D4,$60,$E6,$F9,$7C,$5F
	.dc.b	$F0,$54,$7F,$98,$31,$46,$23,$F2,$87,$E5,$63,$38,$FF,$99,$A1,$18
	.dc.b	$79,$7A,$0F,$AA,$D4,$FF,$2A,$6C,$66,$77,$1F,$E6,$95,$F4,$54,$8F
	.dc.b	$F9,$B5,$FF,$37,$7C,$7F,$F9,$37,$0F,$F2,$3C,$FF,$9D,$7D,$CF,$F1
	.dc.b	$BB,$FA,$BA,$FD,$67,$D5,$39,$3D,$73,$B3,$FA,$79,$FC,$CA,$B8,$F2
	.dc.b	$F5,$77,$8D,$9F,$87,$35,$FD,$C7,$A3,$5F,$E7,$B5,$F9,$1F,$98,$C4
	.dc.b	$E7,$E7,$E4,$F9,$A2,$D7,$8B,$E7,$B9,$18,$D7,$DA,$F7,$7F,$C1,$55
	.dc.b	$FE,$54,$6C,$FE,$4B,$C3,$D7,$36,$F9,$65,$EE,$2D,$7F,$6E,$FA,$AF
	.dc.b	$85,$87,$F9,$C1,$B5,$F8,$0E,$B7,$83,$EF,$9C,$DF,$F1,$77,$E4,$3C
	.dc.b	$F7,$D7,$06,$DE,$1E,$34,$C7,$FC,$D7,$7F,$C9,$1B,$1D,$7C,$E9,$18
	.dc.b	$36,$F8,$7D,$59,$1C,$BE,$42,$E5,$EB,$F5,$BC,$C6,$36,$F6,$3E,$9F
	.dc.b	$8E,$97,$E2,$FA,$7B,$7F,$E1,$0F,$95,$74,$7D,$D6,$D7,$FA,$09,$7B
	.dc.b	$D9,$9D,$79,$52,$B5,$E3,$DC,$3F,$C3,$7F,$54,$FA,$AF,$C4,$36,$BF
	.dc.b	$D1,$AF,$F8,$39,$B1,$AE,$79,$1A,$F7,$ED,$FF,$89,$BE,$A9,$DF,$F9
	.dc.b	$36,$D7,$FA,$61,$FC,$9C,$EE,$75,$BC,$AD,$70,$DA,$FF,$18,$FD,$57
	.dc.b	$F2,$89,$FF,$50,$B6,$33,$D7,$FA,$18,$C7,$27,$FA,$E2,$BD,$55,$FE
	.dc.b	$3A,$6C,$67,$C7,$FD,$44,$60,$FE,$1B,$F5,$5F,$D8,$5F,$F5,$7B,$6B
	.dc.b	$97,$BB,$FA,$F8,$C5,$FF,$B1,$9F,$6B,$7C,$3D,$77,$6D,$E1,$A4,$C7
	.dc.b	$AF,$5D,$CB,$FC,$9F,$E3,$7F,$91,$A7,$FD,$7E,$D8,$CF,$CE,$7E,$FE
	.dc.b	$31,$97,$FB,$59,$79,$99,$35,$EA,$5C,$3F,$D9,$3F,$55,$FB,$6B,$FF
	.dc.b	$6E,$B6,$3F,$96,$BF,$E2,$D7,$7E,$2F,$F0,$D8,$FF,$55,$47,$2F,$87
	.dc.b	$C7,$B9,$97,$5C,$52,$7B,$FB,$77,$6F,$F3,$5F,$91,$FE,$7A,$9F,$F6
	.dc.b	$DB,$63,$3E,$3F,$F1,$B8,$C6,$9F,$EF,$2F,$AA,$FF,$65,$5F,$FB,$89
	.dc.b	$B7,$FE,$76,$51,$FC,$8D,$CC,$FF,$BE,$BE,$A9,$F0,$FF,$C9,$DB,$5F
	.dc.b	$EE,$CF,$AA,$E2,$FF,$BB,$A3,$38,$7F,$A0,$97,$81,$C4,$E5,$76,$B8
	.dc.b	$78,$3D,$4C,$5F,$17,$C1,$C1,$27,$C1,$F0,$F6,$BB,$70,$F9,$4C,$69
	.dc.b	$E9,$7F,$9F,$7D,$57,$E4,$2F,$FB,$1B,$6B,$FC,$71,$AB,$A3,$71,$C3
	.dc.b	$82,$5F,$F8,$F7,$1F,$8F,$3F,$E3,$FC,$39,$27,$F9,$6B,$7C,$2F,$09
	.dc.b	$FD,$3F,$1F,$C8,$DF,$8A,$3C,$BE,$8C,$8E,$C6,$4B,$5D,$E3,$D7,$E9
	.dc.b	$49,$CC,$F8,$67,$0F,$FA,$CB,$0F,$E6,$CD,$F8,$1F,$17,$CA,$FC,$47
	.dc.b	$E2,$F3,$4C,$39,$DF,$0D,$3D,$F8,$7B,$6B,$E5,$F7,$4B,$FC,$17,$C6
	.dc.b	$DA,$F0,$78,$1F,$40,$C6,$1C,$B1,$47,$D4,$F8,$F9,$87,$E2,$FD,$EB
	.dc.b	$87,$A7,$E0,$3E,$C9,$8C,$79,$3F,$73,$EA,$A5,$FA,$22,$D7,$FB,$4B
	.dc.b	$E9,$FF,$37,$3F,$E2,$26,$C7,$F9,$DF,$EA,$9F,$2F,$FE,$03,$9C,$C9
	.dc.b	$FF,$5A,$D7,$CE,$6C,$75,$F8,$FE,$1D,$7C,$7D,$69,$7F,$53,$EA,$9F
	.dc.b	$55,$8C,$3C,$9E,$C7,$73,$1B,$FC,$EC,$FF,$AB,$CF,$A7,$93,$C8,$7E
	.dc.b	$27,$D8,$B9,$78,$F8,$BD,$06,$2B,$E0,$3E,$BF,$7E,$4E,$CF,$53,$EF
	.dc.b	$F8,$99,$1E,$7E,$AB,$7F,$27,$EA,$BE,$21,$8D,$3C,$CF,$C7,$E3,$97
	.dc.b	$F9,$2F,$59,$E2,$E3,$06,$1F,$8A,$E4,$7B,$9F,$E3,$16,$E1,$F3,$71
	.dc.b	$C4,$FF,$8D,$7E,$A7,$F9,$66,$1C,$AB,$FC,$B7,$F2,$FF,$DE,$9F,$D0
	.dc.b	$CE,$B1,$FC,$BC,$E6,$B8,$1F,$EB,$57,$11,$AE,$7E,$DF,$13,$FE,$43
	.dc.b	$9C,$E1,$FE,$68,$C3,$F9,$F6,$1F,$EB,$D6,$A9,$FA,$F8,$7E,$9A,$72
	.dc.b	$4F,$84,$F5,$7A,$7E,$B7,$D9,$F2,$1F,$E5,$1F,$AA,$F4,$65,$DE,$31
	.dc.b	$9C,$8F,$F3,$B4,$FC,$2C,$FF,$CC,$2D,$7F,$CF,$AD,$7C,$DE,$6F,$C9
	.dc.b	$3E,$57,$0D,$AF,$8F,$16,$7E,$28,$CF,$EC,$CE,$89,$1A,$8F,$EB,$6D
	.dc.b	$AE,$5A,$7C,$1B,$B7,$F2,$66,$D7,$92,$39,$D7,$C7,$9C,$CF,$86,$AD
	.dc.b	$F8,$A3,$C1,$18,$3D,$77,$C8,$FA,$CF,$0A,$D4,$66,$7E,$C4,$61,$E7
	.dc.b	$1F,$88,$8C,$5F,$A6,$30,$71,$9F,$23,$6F,$C9,$3B,$9E,$64,$7D,$93
	.dc.b	$BB,$F2,$34,$E5,$7E,$07,$D2,$78,$2B,$93,$86,$9E,$A7,$93,$83,$EC
	.dc.b	$47,$B5,$FF,$52,$DF,$D6,$6D,$7F,$AB,$9F,$A4,$7F,$AA,$8F,$69,$63
	.dc.b	$83,$E1,$38,$D7,$FB,$1E,$3F,$03,$7F,$D1,$3E,$A3,$F3,$35,$D2,$31
	.dc.b	$97,$DC,$9C,$CF,$69,$CD,$FC,$D9,$CD,$F4,$8F,$F6,$BC,$63,$0F,$F5
	.dc.b	$B3,$63,$FA,$FB,$6F,$F3,$2D,$BE,$B1,$F1,$63,$13,$D3,$0F,$21,$1F
	.dc.b	$13,$67,$83,$6B,$EF,$1D,$0B,$3C,$CE,$24,$9E,$2D,$8D,$C0,$57,$91
	.dc.b	$C3,$5C,$D7,$86,$BF,$05,$5E,$16,$9E,$19,$EE,$48,$EE,$F2,$8F,$11
	.dc.b	$C1,$EF,$3D,$5F,$27,$D4,$FF,$88,$78,$7F,$09,$08,$C3,$F8,$75,$F3
	.dc.b	$8F,$3F,$71,$A9,$FF,$7A,$E4,$E1,$E7,$1F,$01,$B1,$98,$EE,$D7,$F2
	.dc.b	$9F,$52,$F9,$5F,$91,$1F,$C5,$A3,$0F,$F8,$81,$AF,$E0,$BE,$7B,$7B
	.dc.b	$7E,$55,$BB,$7E,$31,$B3,$F4,$1A,$FD,$4E,$4B,$8B,$E4,$A9,$27,$C1
	.dc.b	$ED,$70,$C8,$D2,$47,$97,$DB,$EB,$F4,$EF,$A9,$EF,$E7,$3D,$2A,$6A
	.dc.b	$37,$E1,$5C,$3F,$C7,$B7,$C5,$B1,$D3,$E3,$7D,$3F,$42,$F6,$E9,$9A
	.dc.b	$E3,$EA,$D7,$F9,$FE,$0B,$F2,$CF,$9B,$E3,$BE,$1F,$D4,$BD,$9A,$67
	.dc.b	$0F,$1E,$D5,$E7,$5E,$E8,$F7,$FA,$B6,$76,$7A,$9F,$9F,$F4,$7F,$E4
	.dc.b	$CE,$9F,$F2,$6F,$D2,$7F,$9C,$6B,$FC,$B6,$D7,$FA,$AC,$EF,$AB,$F0
	.dc.b	$6B,$D6,$E7,$12,$7F,$CC,$73,$96,$FC,$EF,$DE,$FA,$97,$D4,$D3,$C7
	.dc.b	$1E,$4F,$32,$D7,$03,$9B,$E8,$FA,$8F,$C3,$FD,$4B,$C9,$A6,$6B,$C3
	.dc.b	$EB,$D7,$D3,$5F,$02,$3E,$0F,$AD,$67,$67,$A1,$F8,$BF,$17,$F4,$BE
	.dc.b	$A6,$4F,$B2,$3E,$4F,$AF,$A8,$C1,$F0,$1F,$9B,$EC,$FE,$73,$8D,$F0
	.dc.b	$54,$CD,$7C,$6F,$5A,$BF,$3A,$BE,$AC,$7E,$93,$CF,$79,$7F,$8E,$F3
	.dc.b	$65,$FF,$A4,$27,$2D,$F2,$B8,$54,$8C,$1B,$F4,$7A,$4C,$8C,$1C,$FE
	.dc.b	$7B,$F3,$7F,$53,$23,$07,$7F,$93,$F2,$4C,$8C,$1E,$3F,$8B,$FD,$6F
	.dc.b	$CD,$48,$C1,$E7,$F2,$7D,$93,$23,$07,$AF,$D8,$FE,$07,$ED,$BA,$29
	.dc.b	$F0,$E3,$E2,$75,$5A,$E0,$7C,$37,$E4,$7A,$BA,$8C,$1F,$1D,$FE,$1F
	.dc.b	$F5,$2E,$EB,$DF,$2B,$FD,$77,$39,$6F,$CD,$F0,$F7,$E4,$60,$FF,$60
	.dc.b	$FD,$4B,$E8,$3F,$8D,$AE,$3A,$62,$33,$C1,$AF,$90,$97,$CA,$3F,$91
	.dc.b	$FD,$4B,$E5,$EF,$CC,$4A,$FF,$67,$CE,$5B,$F1,$DC,$3D,$49,$18,$3F
	.dc.b	$DA,$BD,$EF,$86,$FE,$5E,$FE,$D6,$3F,$CA,$2D,$F4,$3F,$D9,$EE,$E3
	.dc.b	$8F,$8F,$C3,$E5,$E4,$60,$FF,$70,$FD,$4F,$F4,$9B,$FF,$6B,$AF,$0B
	.dc.b	$F9,$EB,$67,$78,$63,$8B,$3E,$04,$9E,$71,$C3,$DE,$E3,$EF,$C9,$EF
	.dc.b	$1F,$D2,$3E,$A5,$E1,$DF,$80,$95,$FE,$F4,$9C,$BF,$EF,$6F,$A9,$6F
	.dc.b	$FD,$41,$7C,$65,$31,$1F,$D5,$3B,$FD,$15,$FE,$76,$7F,$AA,$3F,$F0
	.dc.b	$01,$AE,$78,$38,$78,$7C,$7E,$2E,$4E,$E3,$A6,$DD,$7C,$39,$3C,$E3
	.dc.b	$FB,$1F,$D4,$FF,$55,$8F,$F0,$C7,$2B,$D9,$FE,$E7,$2F,$83,$D9,$C3
	.dc.b	$87,$FB,$5F,$D4,$FF,$37,$C3,$FC,$41,$5F,$CC,$1E,$77,$CB,$D9,$DB
	.dc.b	$C3,$87,$8F,$AF,$2B,$67,$F8,$A7,$0F,$8E,$E0,$2E,$BF,$0F,$1C,$9D
	.dc.b	$4E,$5B,$EB,$79,$AA,$46,$0D,$F9,$1E,$AF,$22,$46,$0E,$7E,$A7,$BA
	.dc.b	$64,$60,$F2,$7D,$43,$E2,$7F,$C1,$BF,$52,$F8,$0A,$7F,$84,$27,$55
	.dc.b	$C3,$C1,$C5,$C8,$C1,$EF,$F9,$87,$93,$FE,$18,$F5,$19,$D3,$FC,$37
	.dc.b	$3A,$AE,$1E,$5E,$2E,$46,$0F,$95,$F5,$3F,$8A,$EB,$FB,$D1,$FE,$57
	.dc.b	$C5,$CB,$DC,$FF,$13,$D6,$F9,$FF,$C6,$AF,$E5,$AF,$8D,$A3,$4F,$F9
	.dc.b	$85,$AE,$7C,$0E,$7E,$3F,$46,$47,$92,$3B,$ED,$EF,$E5,$93,$E8,$8F
	.dc.b	$1F,$06,$A3,$07,$B1,$F3,$F8,$EC,$BF,$CB,$B9,$7F,$25,$6C,$66,$BF
	.dc.b	$D4,$93,$96,$F9,$EF,$A9,$E2,$D8,$CC,$60,$FF,$2F,$AF,$A1,$A4,$7F
	.dc.b	$CC,$6F,$F9,$93,$E4,$3E,$7E,$3F,$22,$7F,$CD,$9B,$FE,$75,$FF,$44
	.dc.b	$57,$FA,$27,$EA,$5C,$BF,$E6,$49,$D9,$FE,$FA,$3F,$0B,$5C,$79,$3A
	.dc.b	$BB,$C6,$CF,$F9,$B2,$6B,$FE,$95,$E5,$AF,$F4,$1A,$FD,$77,$2B,$E5
	.dc.b	$F3,$5A,$F9,$D9,$5E,$78,$B7,$E3,$F8,$BE,$66,$36,$F7,$3E,$07,$FD
	.dc.b	$0D,$5F,$BB,$6C,$FD,$78,$FF,$40,$E3,$FC,$99,$D7,$37,$1E,$59,$18
	.dc.b	$39,$3E,$CF,$A9,$F8,$B8,$7E,$51,$B3,$D3,$BF,$5B,$EF,$79,$3F,$C3
	.dc.b	$6B,$79,$FF,$AE,$5F,$9A,$F9,$4C,$78,$DD,$CF,$07,$8F,$6B,$97,$9F
	.dc.b	$E3,$31,$5C,$45,$FF,$4A,$9F,$07,$A7,$E9,$3D,$5F,$1D,$63,$C9,$EA
	.dc.b	$FB,$4C,$5F,$3B,$47,$C1,$F8,$D9,$5E,$BF,$A6,$B7,$E1,$F9,$6F,$B9
	.dc.b	$3F,$CA,$DB,$19,$F7,$BF,$0B,$18,$E2,$FD,$EA,$F2,$72,$6B,$C6,$B8
	.dc.b	$7F,$35,$FA,$97,$D4,$FE,$39,$B1,$9F,$0B,$13,$9E,$A9,$3C,$7D,$DB
	.dc.b	$AF,$C7,$FA,$97,$7F,$E6,$1B,$5F,$D0,$DF,$CC,$4E,$E7,$5C,$12,$B5
	.dc.b	$C5,$6B,$E9,$7D,$4F,$F2,$A9,$FE,$98,$D8,$CF,$67,$E9,$A3,$1C,$5F
	.dc.b	$C7,$AB,$FC,$60,$7F,$53,$2D,$7D,$D7,$FC,$64,$67,$6F,$D7,$FD,$4A
	.dc.b	$69,$FB,$29,$7B,$B8,$7F,$AF,$4F,$B2,$FF,$96,$B6,$3F,$DC,$AF,$E5
	.dc.b	$23,$3E,$17,$6C,$E5,$72,$EA,$B8,$FE,$47,$C6,$7B,$7F,$C0,$9B,$5F
	.dc.b	$DB,$57,$D3,$66,$75,$E6,$CA,$D7,$AB,$70,$FC,$E7,$D4,$FE,$EB,$5F
	.dc.b	$E8,$6D,$AF,$F7,$DB,$FE,$35,$7F,$BD,$39,$9F,$88,$FF,$AB,$63,$3F
	.dc.b	$0D,$FC,$5A,$1E,$ED,$8C,$70,$CA,$E7,$EE,$5C,$3F,$5D,$E4,$7E,$B5
	.dc.b	$BE,$2F,$F2,$08,$FF,$79,$57,$F8,$2D,$D6,$7F,$D6,$BE,$A7,$FD,$9D
	.dc.b	$3F,$E0,$F6,$DF,$EE,$5F,$F1,$24,$67,$E5,$BF,$95,$FD,$4B,$E1,$E9
	.dc.b	$FC,$BB,$EA,$77,$87,$F0,$19,$D5,$FF,$86,$D7,$81,$C8,$E3,$19,$7A
	.dc.b	$F8,$32,$30,$71,$7C,$3C,$32,$3F,$C4,$D8,$7F,$9C,$5E,$D1,$E6,$E1
	.dc.b	$91,$83,$C4,$FF,$44,$FA,$9F,$CB,$47,$F8,$BD,$B9,$7F,$19,$BF,$C2
	.dc.b	$CE,$63,$86,$F2,$30,$71,$7A,$78,$32,$3C,$BC,$F6,$BF,$92,$60,$F4
	.dc.b	$7C,$46,$78,$BF,$C1,$B1,$9B,$F2,$7A,$7E,$73,$6E,$0F,$47,$9D,$EB
	.dc.b	$3E,$97,$C9,$D8,$F6,$7A,$1E,$F3,$19,$78,$9F,$85,$DF,$27,$E5,$FC
	.dc.b	$6D,$AF,$3F,$83,$F1,$CC,$69,$E6,$7E,$57,$CD,$4B,$FA,$7E,$ED,$BF
	.dc.b	$7F,$BF,$F3,$CC,$61,$EE,$7E,$97,$D4,$CB,$FA,$51,$6B,$FA,$57,$D3
	.dc.b	$7E,$52,$FF,$C6,$2D,$AF,$E1,$D2,$FA,$97,$CB,$FF,$80,$2B,$53,$FC
	.dc.b	$4D,$FC,$33,$D3,$E3,$F5,$E0,$6D,$E1,$EB,$4C,$7F,$CE,$5F,$52,$FA
	.dc.b	$9C,$61,$E4,$F7,$FB,$18,$D7,$CA,$F8,$78,$E6,$1F,$96,$F6,$2D,$FB
	.dc.b	$38,$BC,$C6,$30,$F6,$BE,$9F,$8A,$97,$FE,$83,$DF,$FC,$AE,$D8,$F7
	.dc.b	$7C,$4C,$8F,$3F,$55,$C7,$E4,$FD,$4F,$C3,$31,$97,$CA,$7E,$37,$1C
	.dc.b	$BF,$F4,$87,$AD,$FC,$49,$F0,$47,$F8,$AB,$8D,$EE,$53,$86,$3F,$DF
	.dc.b	$E3,$35,$FE,$99,$FA,$8F,$E4,$12,$6F,$CB,$FF,$6F,$9F,$B3,$1A,$3F
	.dc.b	$D1,$8F,$1F,$83,$FF,$1C,$C6,$5B,$EC,$C9,$FF,$53,$53,$4C,$61,$FE
	.dc.b	$11,$C3,$FB,$9B,$7D,$DC,$ED,$FE,$7A,$9C,$B7,$CB,$EE,$7F,$92,$3E
	.dc.b	$9F,$D7,$7C,$E3,$C7,$E6,$4C,$FD,$D1,$67,$C1,$EA,$7F,$97,$DF,$EC
	.dc.b	$58,$FF,$60,$AF,$F2,$E4,$62,$74,$FC,$56,$CE,$E3,$FC,$55,$F5,$3F
	.dc.b	$16,$7F,$CB,$6F,$CC,$F6,$FF,$8B,$E7,$FE,$94,$6E,$1F,$A4,$CD,$F1
	.dc.b	$7F,$E6,$29,$CE,$3F,$E6,$33,$C2,$FF,$20,$7F,$CE,$0D,$7F,$5A,$D7
	.dc.b	$FC,$D5,$E7,$F0,$25,$FB,$21,$3F,$E7,$87,$9C,$FC,$BF,$8B,$F1,$36
	.dc.b	$BF,$DB,$D5,$FD,$F5,$EF,$F0,$FF,$47,$8F,$D0,$C8,$DB,$96,$93,$0F
	.dc.b	$F2,$85,$73,$7F,$C2,$59,$7F,$86,$33,$F8,$D7,$CE,$76,$F3,$6D,$F3
	.dc.b	$9D,$BF,$EF,$3A,$E4,$F9,$8D,$16,$7F,$CC,$11,$A7,$FD,$21,$5D,$EB
	.dc.b	$9C,$6B,$0C,$5F,$19,$C9,$FE,$CE,$BE,$55,$F3,$2F,$6F,$4A,$F2,$5E
	.dc.b	$23,$51,$E4,$7E,$0C,$67,$0F,$89,$1A,$3C,$53,$B7,$18,$78,$A3,$F1
	.dc.b	$55,$C6,$FF,$0E,$48,$FC,$05,$FF,$A2,$2F,$F1,$33,$97,$F3,$F1,$97
	.dc.b	$E3,$1A,$9F,$C0,$C7,$94,$8F,$8D,$FF,$21,$47,$BC,$FB,$AE,$9F,$AE
	.dc.b	$62,$38,$46,$6F,$D5,$3F,$EC,$18,$C4,$FC,$F3,$F3,$AE,$23,$B4,$FE
	.dc.b	$AE,$31,$1F,$12,$76,$76,$23,$E2,$9E,$11,$87,$C3,$1F,$A4,$9E,$67
	.dc.b	$D2,$3D,$AF,$53,$CD,$3F,$A4,$F3,$FF,$70,$7E,$E7,$D1,$B1,$93,$E8
	.dc.b	$C7,$EE,$88,$FD,$4F,$63,$F7,$A4,$70,$7E,$DF,$3F,$E4,$B7,$FE,$EF
	.dc.b	$6B,$FC,$27,$93,$F8,$AC,$62,$3F,$57,$EB,$E0,$95,$CD,$FE,$37,$F5
	.dc.b	$3F,$E6,$CE,$BF,$E0,$8D,$AF,$E4,$2B,$E7,$9C,$3A,$F7,$FF,$A6,$D7
	.dc.b	$F2,$6F,$A9,$FF,$23,$BF,$B1,$F6,$BF,$BF,$F1,$7E,$95,$BA,$7F,$80
	.dc.b	$AB,$75,$FE,$D1,$5E,$D6,$A3,$FD,$29,$1A,$B5,$71,$BD,$52,$B8,$BF
	.dc.b	$E9,$BA,$DB,$CB,$DD,$DE,$EF,$9E,$A7,$2B,$4D,$EB,$C3,$2F,$A6,$5E
	.dc.b	$99,$F0,$FD,$E3,$A3,$AE,$B2,$F6,$7F,$AB,$67,$4E,$03,$FC,$C9,$C3
	.dc.b	$A7,$FD,$E8,$DD,$3D,$71,$C3,$A3,$E0,$FD,$43,$EA,$BE,$1F,$AE,$F6
	.dc.b	$A9,$A8,$DF,$95,$70,$F1,$5F,$16,$C7,$4F,$8E,$F9,$BF,$55,$F3,$25
	.dc.b	$9F,$8A,$30,$F9,$2D,$5E,$4F,$73,$95,$F6,$8C,$3E,$0F,$8E,$FB,$3F
	.dc.b	$55,$FE,$6D,$8F,$88,$D7,$BF,$1F,$F9,$23,$05,$7C,$0D,$BE,$0B,$2B
	.dc.b	$D7,$1F,$13,$DC,$D4,$60,$F7,$BF,$88,$FA,$AF,$97,$A7,$CC,$FA,$6F
	.dc.b	$F6,$6D,$7D,$56,$BE,$23,$FD,$CD,$3C,$32,$FF,$2D,$AF,$77,$A4,$49
	.dc.b	$FB,$D3,$96,$FC,$27,$E6,$3E,$A9,$F3,$94,$F2,$47,$97,$E0,$2D,$70
	.dc.b	$39,$BE,$9F,$A9,$FD,$0F,$D5,$3D,$3A,$66,$BC,$3E,$FD,$7D,$43,$35
	.dc.b	$FC,$39,$F8,$9E,$F5,$AE,$27,$A9,$F9,$1F,$1D,$F6,$3E,$AA,$4F,$BA
	.dc.b	$3E,$67,$BD,$A8,$C1,$F0,$9F,$A1,$F0,$3F,$B8,$E3,$7C,$45,$33,$5F
	.dc.b	$23,$DF,$AF,$A3,$55,$FD,$99,$FE,$03,$E7,$3E,$AB,$F2,$FF,$33,$30
	.dc.b	$FD,$74,$E5,$BE,$67,$0A,$91,$83,$7F,$3B,$D2,$64,$60,$E7,$ED,$FE
	.dc.b	$93,$EA,$A4,$60,$EF,$EE,$79,$26,$46,$0F,$1F,$A5,$FB,$6F,$66,$46
	.dc.b	$0F,$3F,$9D,$EC,$99,$18,$3D,$7E,$FF,$F2,$0F,$81,$74,$D3,$E1,$C7
	.dc.b	$C4,$EB,$B5,$C0,$F8,$6F,$C8,$FA,$3D,$46,$0F,$8E,$FF,$2C,$FA,$A7
	.dc.b	$7D,$F0,$25,$7F,$19,$9C,$B7,$E7,$38,$7C,$3C,$8C,$1F,$C7,$FE,$A9
	.dc.b	$F0,$7F,$CE,$17,$1D,$31,$19,$E0,$D7,$C0,$4B,$E5,$1F,$D0,$3E,$AB
	.dc.b	$F9,$44,$7E,$71,$FE,$4D,$E4,$FF,$B3,$9B,$00,$F0,$F2,$FF,$95,$AF
	.dc.b	$91,$E9,$E9,$CA,$FE,$6D,$39,$7F,$9C,$77,$BE,$2B,$FA,$8B,$FB,$D8
	.dc.b	$FD,$6B,$70,$FE,$6E,$EE,$38,$F8,$FC,$3E,$62,$46,$0F,$E8,$9F,$54
	.dc.b	$AF,$EB,$AB,$C1,$A6,$23,$3C,$3A,$EF,$97,$CA,$38,$FB,$DC,$BD,$F9
	.dc.b	$3D,$A3,$FB,$4F,$D5,$3C,$2B,$EE,$CA,$FE,$AD,$39,$7F,$AC,$7D,$53
	.dc.b	$6F,$EE,$2B,$E3,$E9,$88,$FE,$EB,$DD,$E9,$AF,$EC,$0F,$F7,$16,$F8
	.dc.b	$FA,$E0,$DF,$87,$C3,$EA,$A5,$70,$8E,$7B,$F4,$F0,$E4,$F4,$8F,$F0
	.dc.b	$1F,$D5,$7F,$61,$8F,$ED,$CB,$E0,$32,$6B,$83,$FC,$13,$F5,$5F,$CE
	.dc.b	$EB,$F0,$D5,$FC,$D1,$E3,$EC,$6B,$BF,$7C,$3C,$3D,$69,$5F,$DE,$F0
	.dc.b	$F8,$4D,$EB,$EB,$C7,$23,$06,$FD,$5F,$0C,$C8,$C1,$CF,$C7,$F2,$F8
	.dc.b	$F2,$30,$77,$F4,$BD,$33,$23,$07,$F8,$37,$EA,$5F,$27,$FE,$26,$FA
	.dc.b	$A7,$BD,$4F,$F1,$4C,$EA,$B8,$78,$38,$B9,$18,$3F,$C3,$1F,$37,$FA
	.dc.b	$C7,$FC,$6F,$8F,$F8,$CB,$D3,$3F,$E3,$49,$D1,$FA,$FE,$3F,$D9,$52
	.dc.b	$33,$F1,$3B,$F8,$8F,$F1,$DF,$D5,$7E,$16,$BF,$AB,$E2,$E5,$EC,$6B
	.dc.b	$C0,$DF,$81,$6B,$FC,$59,$7C,$7C,$9F,$F1,$3B,$F9,$18,$FE,$C2,$FE
	.dc.b	$B5,$DC,$F2,$78,$F9,$1E,$2D,$FD,$9C,$D2,$7A,$C7,$9B,$87,$51,$83
	.dc.b	$C2,$FA,$BC,$8F,$F1,$3C,$7C,$03,$F8,$96,$3F,$C8,$32,$FF,$79,$36
	.dc.b	$34,$1C,$3C,$5F,$E5,$3F,$55,$C3,$D1,$34,$93,$FE,$FD,$1F,$47,$5A
	.dc.b	$65,$C1,$BF,$90,$FF,$33,$E1,$FE,$4E,$9F,$F3,$FF,$2F,$F9,$5D,$FE
	.dc.b	$55,$5F,$CA,$FE,$AB,$ED,$47,$DC,$BE,$35,$EF,$E5,$B3,$F8,$06,$BF
	.dc.b	$CC,$E6,$BF,$ED,$E3,$FD,$5A,$7F,$AB,$B9,$1F,$37,$AE,$6D,$FC,$BC
	.dc.b	$C7,$C9,$16,$FF,$C3,$A3,$E5,$FA,$99,$D7,$D8,$F8,$9F,$E5,$D5,$FE
	.dc.b	$73,$6C,$FE,$43,$C3,$D7,$36,$F9,$65,$EE,$2D,$7F,$8B,$3E,$AB,$E3
	.dc.b	$B7,$6F,$7E,$CF,$07,$E0,$D6,$B3,$FF,$3D,$BF,$EF,$E2,$56,$B8,$37
	.dc.b	$E3,$CC,$FF,$D1,$15,$FC,$6D,$B1,$D3,$D0,$91,$83,$6F,$83,$D5,$91
	.dc.b	$CB,$E3,$EE,$BE,$9F,$5F,$CA,$63,$4F,$2B,$E8,$F8,$D9,$7E,$2F,$A8
	.dc.b	$B7,$FE,$4C,$F9,$8F,$AF,$3F,$E9,$D6,$C6,$7D,$FF,$C0,$C6,$31,$FF
	.dc.b	$4D,$BF,$E3,$CA,$E1,$E5,$6B,$C7,$DB,$F8,$8F,$AA,$7C,$D5,$31,$E0
	.dc.b	$DB,$FF,$56,$4F,$FB,$29,$BC,$FC,$7D,$EE,$4F,$E4,$BE,$A9,$DF,$4C
	.dc.b	$77,$DB,$FF,$5B,$4B,$87,$81,$C7,$8E,$CF,$FA,$EF,$EA,$BF,$90,$B7
	.dc.b	$7F,$CF,$C7,$E0,$63,$07,$FB,$04,$FF,$11,$97,$5F,$D2,$BF,$E2,$43
	.dc.b	$07,$FB,$1F,$EA,$BF,$C3,$4D,$E4,$DF,$BD,$FA,$C9,$CC,$7F,$9F,$96
	.dc.b	$99,$35,$D9,$6F,$A6,$7C,$4E,$5D,$36,$3F,$DA,$DE,$2B,$DC,$FD,$CB
	.dc.b	$63,$3F,$39,$89,$CF,$99,$27,$8F,$AB,$75,$FF,$48,$FD,$57,$EB,$BB
	.dc.b	$3F,$DB,$6D,$AF,$E8,$AF,$F8,$7D,$FF,$73,$39,$9F,$F1,$D3,$FE,$9F
	.dc.b	$8C,$FC,$3F,$F0,$C8,$7B,$D6,$31,$C5,$2B,$9F,$BB,$70,$FF,$51,$F8
	.dc.b	$DF,$E2,$19,$FF,$7A,$36,$33,$E2,$FF,$18,$8C,$67,$FE,$F2,$FA,$AF
	.dc.b	$F6,$05,$FF,$BE,$DB,$7F,$EA,$E5,$1F,$C8,$1C,$CF,$FB,$EB,$EA,$9F
	.dc.b	$0F,$FC,$91,$B5,$FE,$57,$9F,$F7,$FF,$D5,$72,$8C,$1F,$DA,$A7,$78
	.dc.b	$7F,$AF,$17,$83,$CC,$E9,$85,$AE,$35,$FE,$63,$AF,$D3,$3C,$7C,$1D
	.dc.b	$36,$76,$A7,$E6,$C6,$1B,$1F,$EC,$CC,$3F,$CC,$93,$9E,$0F,$67,$04
	.dc.b	$8C,$1C,$5F,$E8,$7F,$55,$F9,$8D,$3E,$3E,$1F,$E7,$78,$C5,$7F,$82
	.dc.b	$30,$73,$70,$AE,$3B,$CA,$F3,$DF,$09,$E5,$3D,$BC,$29,$5E,$CC,$3F
	.dc.b	$35,$5F,$EE,$4B,$FE,$07,$3C,$BD,$0F,$29,$C2,$74,$FA,$3D,$A9,$3B
	.dc.b	$3C,$2F,$B3,$E0,$A4,$77,$F9,$3B,$7D,$F9,$7E,$09,$8D,$BE,$0C,$32
	.dc.b	$4F,$C6,$ED,$98,$7A,$7E,$36,$DF,$A7,$BF,$E5,$98,$C7,$9E,$3F,$E9
	.dc.b	$DB,$F8,$BF,$35,$F4,$0F,$EE,$7B,$FF,$8B,$63,$27,$CA,$F8,$0F,$AA
	.dc.b	$6D,$C1,$F3,$BE,$AB,$EC,$9A,$70,$7E,$23,$E9,$FF,$3E,$FF,$90,$1B
	.dc.b	$9F,$FA,$CF,$EA,$9F,$2D,$FE,$04,$9C,$C9,$FC,$5A,$D3,$E2,$C6,$1B
	.dc.b	$F3,$BC,$7E,$43,$97,$A9,$32,$FC,$CF,$D5,$3E,$AB,$19,$73,$9F,$27
	.dc.b	$BF,$E1,$33,$97,$67,$C9,$C9,$27,$F1,$1E,$B5,$BE,$47,$FB,$37,$5F
	.dc.b	$94,$E0,$F5,$7C,$3E,$F1,$F4,$BE,$FF,$80,$98,$7C,$78,$B5,$C6,$DF
	.dc.b	$15,$F1,$4F,$F1,$4E,$BF,$E0,$19,$D4,$7C,$9F,$AA,$98,$7D,$18,$B3
	.dc.b	$CF,$8B,$E7,$1F,$E3,$2F,$80,$FD,$CE,$9F,$32,$3F,$C6,$BC,$4F,$72
	.dc.b	$9C,$31,$FE,$07,$19,$7F,$82,$FD,$57,$F3,$5C,$4D,$F9,$8F,$EF,$D3
	.dc.b	$FE,$80,$8D,$1F,$E9,$47,$8F,$C2,$7F,$91,$23,$2D,$F6,$64,$FF,$12
	.dc.b	$A6,$D8,$C3,$E7,$61,$FE,$04,$AF,$1F,$BF,$9D,$7F,$57,$39,$AD,$F9
	.dc.b	$7D,$CF,$F2,$9F,$D4,$7A,$E3,$38,$7B,$3C,$CF,$21,$B7,$07,$C0,$F4
	.dc.b	$FF,$CB,$7F,$55,$EC,$6F,$C3,$49,$FF,$3D,$DA,$BF,$C9,$AF,$FA,$0D
	.dc.b	$AF,$FA,$91,$A8,$FE,$F2,$7F,$9C,$CC,$7B,$9F,$DF,$9F,$F4,$63,$5F
	.dc.b	$F5,$6B,$5D,$9E,$EF,$9B,$FF,$38,$C7,$F2,$58,$64,$9C,$FA,$1F,$13
	.dc.b	$E2,$EE,$3F,$D0,$AB,$FC,$29,$7F,$9D,$6E,$FF,$C6,$1B,$3E,$2B,$E4
	.dc.b	$FF,$4A,$C7,$FC,$26,$D7,$C7,$EF,$79,$19,$19,$38,$46,$30,$FD,$64
	.dc.b	$F2,$6B,$CA,$78,$46,$1B,$87,$F2,$C9,$57,$38,$C4,$59,$F9,$33,$6B
	.dc.b	$E7,$CD,$7F,$C5,$71,$63,$C5,$5C,$AB,$CE,$78,$0F,$A7,$1B,$9E,$0D
	.dc.b	$8D,$57,$66,$BD,$9B,$1D,$9E,$67,$68,$F7,$12,$BC,$4F,$C6,$AE,$F3
	.dc.b	$F9,$2A,$F9,$13,$F8,$87,$F1,$8F,$F9,$19,$A6,$9E,$E7,$FB,$51,$C3
	.dc.b	$E7,$30,$FE,$AB,$E5,$7E,$36,$26,$92,$BD,$B1,$96,$C6,$6B,$D5,$1F
	.dc.b	$A1,$7F,$2B,$19,$C3,$DA,$7E,$3D,$C4,$F9,$A3,$0F,$42,$4F,$43,$F5
	.dc.b	$95,$E1,$70,$76,$8D,$CF,$67,$C5,$3B,$8C,$4F,$2A,$F0,$BF,$0E,$3C
	.dc.b	$2F,$49,$E1,$5E,$53,$9D,$70,$9F,$D8,$F0,$7F,$12,$3F,$6B,$E9,$5A
	.dc.b	$FD,$84,$7D,$6B,$FF,$17,$D7,$C5,$38,$CF,$F0,$E8,$CC,$7A,$38,$F8
	.dc.b	$12,$74,$F3,$E2,$91,$C5,$B3,$D4,$7D,$57,$F8,$C3,$AB,$F8,$DB,$63
	.dc.b	$4B,$E7,$9C,$1A,$F5,$BF,$A5,$57,$3C,$DF,$2F,$FC,$92,$3F,$88,$B6
	.dc.b	$33,$1E,$B8,$D7,$D5,$7F,$97,$23,$F8,$97,$B9,$F8,$5E,$2F,$E2,$0D
	.dc.b	$DB,$F1,$15,$BA,$FE,$E4,$BD,$BD,$47,$F2,$D8,$D5,$FF,$8C,$1B,$5C
	.dc.b	$AB,$74,$97,$BD,$8C,$9F,$27,$4D,$CE,$AB,$FC,$D2,$BB,$33,$72,$73
	.dc.b	$1D,$BE,$AB,$F8,$AC,$7F,$2F,$C7,$B1,$9E,$9D,$74,$D9,$FF,$3A,$D7
	.dc.b	$07,$FC,$20,$FF,$48,$7F,$C2,$5F,$4E,$FA,$AF,$5F,$AF,$FE,$4C,$9F
	.dc.b	$F3,$EB,$5D,$C7,$C0,$97,$BB,$62,$4F,$FA,$12,$72,$DF,$8F,$F8,$FF
	.dc.b	$55,$F3,$E3,$FD,$12,$D7,$C7,$1D,$7D,$DE,$CC,$8E,$91,$E1,$F9,$0F
	.dc.b	$A3,$F5,$4F,$06,$99,$AF,$16,$95,$F7,$D7,$A6,$3D,$5E,$E5,$9D,$9E
	.dc.b	$97,$EE,$FD,$2F,$F9,$C3,$A7,$FC,$E3,$F4,$DF,$E8,$EA,$FF,$50,$B5
	.dc.b	$FF,$1E,$4E,$FA,$BF,$8B,$2F,$73,$9C,$49,$FF,$54,$CE,$5B,$F1,$9F
	.dc.b	$8C,$FA,$A7,$CF,$D3,$C7,$1E,$4F,$7E,$D7,$03,$9B,$E8,$FA,$9F,$CA
	.dc.b	$FD,$53,$C9,$A6,$6B,$C3,$EF,$D7,$D3,$5F,$02,$3E,$0F,$BB,$67,$67
	.dc.b	$A1,$F8,$BF,$21,$F4,$BE,$AA,$4F,$B2,$3E,$4F,$BB,$A8,$C1,$F0,$1F
	.dc.b	$9B,$F0,$3F,$AA,$E3,$7C,$55,$33,$5F,$1B,$DE,$AF,$CE,$AF,$AB,$1F
	.dc.b	$B2,$F3,$DF,$55,$F8,$EF,$66,$5F,$FB,$5A,$72,$DF,$33,$85,$48,$C1
	.dc.b	$BF,$A0,$E9,$32,$30,$73,$FA,$0F,$CD,$FB,$52,$30,$77,$F6,$BC,$93
	.dc.b	$23,$07,$8F,$D2,$FD,$7F,$CE,$48,$C1,$E7,$F3,$FD,$93,$23,$07,$AF
	.dc.b	$DA,$FE,$23,$EF,$BA,$29,$F0,$E3,$E2,$75,$5A,$E0,$7C,$37,$E4,$7A
	.dc.b	$DA,$8C,$1F,$1D,$FE,$3B,$F5,$4E,$EB,$DF,$2B,$FE,$00,$CE,$5B,$F3
	.dc.b	$5C,$3E,$16,$46,$0F,$F2,$C7,$8F,$FE,$DA,$FA,$A7,$C2,$7F,$2A,$5C
	.dc.b	$94,$C4,$67,$87,$5F,$2B,$30,$E7,$E4,$FF,$80,$F0,$FE,$6B,$1F,$E6
	.dc.b	$5F,$AA,$FE,$3D,$97,$FB,$A1,$FE,$44,$DF,$67,$F9,$CC,$6A,$FF,$C1
	.dc.b	$8B,$E4,$B3,$E9,$CC,$3B,$9D,$63,$FA,$0F,$80,$F8,$9B,$E7,$CA,$D8
	.dc.b	$9D,$38,$5E,$47,$2F,$99,$93,$98,$FE,$9B,$F5,$5C,$63,$9A,$F0,$F2
	.dc.b	$6A,$F7,$C9,$C3,$C1,$97,$C2,$39,$FB,$DD,$3D,$F9,$3D,$23,$FA,$DF
	.dc.b	$D5,$3C,$4B,$F2,$D2,$BD,$73,$97,$DB,$F5,$4D,$FF,$B1,$AF,$91,$A6
	.dc.b	$23,$FB,$3F,$81,$E9,$AF,$8A,$FF,$66,$6F,$93,$AE,$0D,$F8,$BC,$3E
	.dc.b	$2E,$57,$08,$E7,$BF,$4F,$16,$4F,$48,$FE,$EB,$F5,$5F,$D6,$7B,$9D
	.dc.b	$BF,$E4,$98,$FE,$AB,$5F,$EF,$36,$CE,$A3,$FD,$5D,$1C,$67,$FC,$A1
	.dc.b	$F5,$5F,$CF,$67,$3E,$C7,$63,$FC,$07,$19,$83,$F0,$DF,$C1,$AF,$5B
	.dc.b	$67,$0E,$1E,$3D,$7C,$B8,$A5,$F0,$8E,$BE,$AE,$A3,$07,$47,$C3,$E4
	.dc.b	$78,$BC,$89,$3E,$08,$F2,$FA,$7A,$8C,$1E,$47,$FC,$37,$F5,$2F,$92
	.dc.b	$FF,$0E,$FD,$53,$DE,$A6,$A3,$7E,$0D,$C3,$F3,$F7,$C5,$B1,$FE,$26
	.dc.b	$F9,$CF,$D8,$F1,$FF,$9E,$7D,$46,$9F,$E2,$A6,$DF,$EA,$2B,$6F,$FB
	.dc.b	$7D,$FD,$B3,$7C,$BE,$87,$5F,$8A,$FF,$19,$FD,$57,$E2,$E5,$D3,$F6
	.dc.b	$DF,$F1,$F7,$37,$73,$7D,$FC,$3B,$EC,$FE,$E2,$F9,$5F,$17,$F3,$11
	.dc.b	$D3,$5F,$F4,$94,$7E,$D6,$30,$7F,$BD,$5D,$CF,$1A,$F2,$F8,$FE,$6E
	.dc.b	$1F,$81,$CF,$30,$F0,$47,$AB,$87,$51,$83,$C8,$FB,$BC,$7F,$6C,$7F
	.dc.b	$9C,$F2,$FE,$53,$97,$F9,$D6,$11,$87,$53,$C1,$FF,$43,$47,$F3,$2F
	.dc.b	$AA,$E0,$D8,$E9,$E9,$27,$2D,$8F,$F3,$B7,$D1,$FF,$9D,$D3,$48,$F8
	.dc.b	$FE,$47,$F5,$98,$FD,$38,$FF,$41,$EF,$FC,$99,$1C,$87,$D5,$39,$7F
	.dc.b	$D3,$97,$CA,$BF,$D3,$DC,$B6,$FF,$D8,$8D,$79,$CD,$7F,$DF,$B3,$DF
	.dc.b	$97,$06,$F9,$9F,$37,$89,$F2,$7D,$0F,$13,$17,$F7,$8F,$3B,$D7,$E2
	.dc.b	$A4,$F0,$3D,$8F,$8F,$5F,$E8,$51,$FD,$11,$B5,$CF,$0F,$E5,$13,$9E
	.dc.b	$6E,$1C,$B2,$30,$71,$7D,$DF,$55,$F2,$AF,$C8,$DC,$3F,$41,$83,$C0
	.dc.b	$DE,$DC,$37,$99,$7F,$56,$C3,$83,$FE,$ED,$7E,$3E,$3D,$3C,$EE,$E7
	.dc.b	$83,$E4,$6D,$71,$F4,$3C,$66,$2B,$C6,$F9,$BD,$69,$3D,$FE,$3A,$D7
	.dc.b	$A3,$D7,$F5,$98,$DF,$F7,$19,$FF,$73,$BF,$37,$C6,$FC,$03,$E0,$FD
	.dc.b	$45,$8F,$ED,$FF,$31,$F6,$9B,$CF,$F8,$38,$FF,$0D,$46,$0D,$F9,$39
	.dc.b	$35,$E3,$58,$FE,$F3,$F5,$4F,$9A,$FC,$63,$63,$3E,$06,$27,$3D,$12
	.dc.b	$78,$FC,$05,$E1,$FC,$47,$D5,$3B,$BF,$2C,$DA,$FF,$03,$AE,$FC,$CE
	.dc.b	$B8,$25,$6B,$92,$E1,$F9,$3F,$AA,$FE,$57,$3F,$E1,$06,$C6,$7B,$7F
	.dc.b	$49,$18,$E6,$FE,$49,$5F,$E3,$23,$FC,$2E,$5A,$FC,$F3,$FE,$34,$33
	.dc.b	$BF,$EB,$7E,$AB,$F6,$F1,$FE,$1E,$6C,$F0,$F7,$74,$EF,$59,$5B,$EE
	.dc.b	$B3,$FA,$BB,$FB,$CD,$D3,$B7,$5F,$77,$F7,$9E,$2B,$DB,$A6,$3E,$72
	.dc.b	$E5,$FE,$30,$9F,$F1,$6B,$7C,$CE,$3E,$A7,$27,$F8,$37,$D5,$7E,$CB
	.dc.b	$87,$FA,$33,$6B,$F8,$9E,$88,$FF,$18,$47,$F8,$F1,$D4,$7F,$18,$7F
	.dc.b	$D5,$D1,$9F,$88,$EB,$EE,$F6,$7B,$F1,$49,$E5,$EE,$5C,$3F,$86,$79
	.dc.b	$1F,$E3,$19,$FF,$24,$B6,$33,$E4,$7F,$1D,$8C,$67,$FD,$77,$EA,$BF
	.dc.b	$D9,$D7,$FE,$50,$6D,$FF,$17,$7F,$C4,$31,$9F,$83,$FE,$4D,$F5,$4F
	.dc.b	$88,$A7,$F2,$9F,$A9,$DE,$1F,$C7,$E7,$57,$FE,$5C,$5E,$0F,$23,$8C
	.dc.b	$65,$EB,$E1,$48,$C1,$C5,$F0,$F0,$C8,$FF,$33,$E1,$FE,$6B,$7B,$47
	.dc.b	$9B,$86,$46,$0F,$13,$FC,$FF,$EA,$BF,$23,$B7,$F7,$F6,$CF,$DA,$B7
	.dc.b	$3E,$F4,$E1,$C1,$2B,$FC,$EF,$7C,$27,$FA,$9B,$FC,$91,$BE,$17,$73
	.dc.b	$E3,$74,$DA,$E3,$E8,$F8,$CC,$57,$F4,$7B,$47,$9F,$D2,$95,$C0,$F0
	.dc.b	$BE,$BF,$42,$47,$C4,$F9,$3B,$87,$3E,$7F,$80,$63,$7F,$80,$FC,$3E
	.dc.b	$F9,$7F,$33,$E3,$AD,$F8,$FC,$0F,$90,$63,$0F,$90,$FC,$BF,$9A,$97
	.dc.b	$F5,$3D,$DB,$7E,$BF,$7F,$E8,$18,$C3,$E8,$3F,$4F,$EA,$A5,$FC,$C8
	.dc.b	$B5,$FD,$A7,$EA,$3F,$2B,$7F,$E3,$56,$D7,$EC,$3E,$A9,$F2,$FF,$DE
	.dc.b	$E7,$37,$FE,$06,$7F,$0C,$F3,$F8,$FD,$78,$3B,$78,$7C,$9C,$C7,$FD
	.dc.b	$67,$F5,$2F,$AA,$C6,$1E,$3F,$67,$B1,$8D,$7C,$8F,$87,$8E,$61,$F9
	.dc.b	$6F,$62,$DF,$AF,$8B,$CC,$63,$0F,$63,$E9,$F8,$A9,$7F,$EC,$4D,$FF
	.dc.b	$D0,$4D,$8F,$77,$C5,$48,$F3,$F5,$5C,$7E,$47,$D5,$7C,$33,$19,$7C
	.dc.b	$97,$E3,$71,$CB,$FF,$68,$FA,$CF,$1B,$18,$7F,$86,$79,$1E,$E7,$F8
	.dc.b	$9D,$B7,$E5,$C3,$13,$FE,$2B,$FA,$AF,$E3,$F6,$6F,$CC,$7F,$6F,$8F
	.dc.b	$F1,$24,$68,$FF,$45,$5F,$FB,$89,$B1,$AF,$84,$E9,$F0,$9D,$7D,$A9
	.dc.b	$7F,$EE,$8C,$1B,$63,$0E,$BE,$F7,$8C,$C7,$07,$F9,$F7,$3F,$EE,$93
	.dc.b	$E2,$DF,$D0,$7F,$90,$FE,$A3,$DB,$19,$CB,$C9,$E7,$7B,$4D,$B8,$3F
	.dc.b	$DF,$3E,$E7,$F0,$39,$FE,$E8,$73,$F8,$AF,$F2,$74,$64,$FE,$C3,$1F
	.dc.b	$D2,$0D,$CF,$F9,$CE,$33,$1F,$D3,$1A,$FF,$98,$3E,$AB,$E2,$BC,$5A
	.dc.b	$FF,$52,$6B,$BE,$0F,$F3,$53,$C6,$FF,$CC,$12,$6C,$E1,$F3,$FB,$D3
	.dc.b	$0F,$F3,$7E,$BF,$EA,$5C,$FF,$C0,$0F,$FA,$9F,$2F,$F2,$0B,$6B,$86
	.dc.b	$5F,$E6,$1A,$DD,$FA,$E7,$E1,$F9,$F2,$3A,$9F,$DA,$A5,$EC,$66,$FF
	.dc.b	$CD,$58,$FF,$88,$5B,$7F,$5E,$75,$8F,$8E,$2E,$1C,$E6,$D7,$FA,$22
	.dc.b	$F9,$5E,$EB,$C3,$3F,$E8,$CC,$3C,$31,$F7,$30,$E3,$3F,$E2,$F7,$C1
	.dc.b	$19,$24,$79,$1E,$6A,$79,$B9,$25,$78,$5F,$8D,$87,$6C,$3C,$B7,$F8
	.dc.b	$1C,$3B,$D7,$08,$C1,$E0,$BF,$A9,$7C,$6B,$CF,$87,$82,$7B,$12,$B8
	.dc.b	$DF,$81,$E4,$62,$35,$1F,$0E,$FA,$4F,$E5,$6F,$F1,$73,$F0,$E3,$04
	.dc.b	$7F,$38,$64,$B8,$7D,$97,$F4,$0D,$7C,$11,$F9,$FF,$3B,$E1,$11,$F2
	.dc.b	$C7,$D9,$8C,$3F,$0C,$C4,$69,$FF,$32,$46,$B0,$F5,$C6,$27,$55,$FD
	.dc.b	$B9,$B5,$B7,$C7,$3D,$0E,$C7,$12,$CF,$16,$CF,$17,$F1,$D7,$FC,$02
	.dc.b	$31,$1F,$0E,$FF,$51,$18,$AF,$BF,$3D,$E7,$CF,$1D,$5B,$3F,$C5,$7D
	.dc.b	$99,$87,$E3,$BF,$67,$FE,$21,$9A,$73,$8F,$D2,$8C,$57,$FB,$FD,$B1
	.dc.b	$99,$FB,$71,$FE,$38,$8E,$1E,$CE,$09,$5F,$72,$1F,$57,$FE,$26,$EA
	.dc.b	$C3,$63,$F0,$2B,$E7,$DE,$B6,$BD,$7F,$E9,$75,$F8,$6D,$32,$F5,$6B
	.dc.b	$99,$F0,$28,$CB,$FC,$A5,$71,$69,$B1,$CE,$30,$72,$FA,$BF,$F2,$9C
	.dc.b	$7E,$DB,$5A,$FF,$17,$F6,$FF,$CF,$7C,$2D,$BF,$CF,$B7,$A8,$C5,$04
	.dc.b	$3D,$BE,$35,$C2,$ED,$6E,$B9,$4F,$4A,$CC,$F2,$E5,$B8,$BE,$46,$9A
	.dc.b	$C3,$F9,$2D,$7F,$0E,$7F,$00,$DA,$D1,$BE,$3E,$BF,$57,$F1,$27,$F9
	.dc.b	$A6,$3D,$8C,$EB,$AE,$6B,$3F,$C6,$AB,$86,$DF,$02,$11,$FD,$79,$46
	.dc.b	$7E,$A9,$F5,$7E,$CF,$5D,$ED,$53,$DF,$3A,$BF,$E4,$EB,$CA,$CD,$C9
	.dc.b	$CC,$70,$F8,$EF,$8F,$F5,$7F,$3E,$7F,$97,$B5,$F1,$F2,$7F,$AD,$30
	.dc.b	$57,$83,$DC,$E5,$5E,$08,$C3,$E2,$F8,$EF,$A7,$F5,$7F,$E7,$28,$FE
	.dc.b	$76,$D7,$A9,$F8,$1B,$4F,$B3,$5F,$6B,$27,$DB,$1F,$03,$DC,$D4,$60
	.dc.b	$F8,$2F,$E0,$BE,$AF,$E7,$70,$7C,$FF,$A7,$FF,$49,$D7,$F4,$C6,$BE
	.dc.b	$21,$3C,$35,$FF,$69,$AF,$77,$A4,$49,$FE,$AB,$39,$6F,$C2,$7E,$4B
	.dc.b	$EA,$DF,$3B,$4F,$24,$79,$7E,$02,$D7,$03,$9B,$E9,$FA,$AF,$CD,$FD
	.dc.b	$5B,$CF,$A6,$6B,$C3,$EF,$D7,$D5,$9A,$FE,$14,$FC,$3F,$7A,$D7,$03
	.dc.b	$D2,$FC,$7F,$8D,$FA,$FF,$57,$27,$DB,$1F,$2F,$DE,$D4,$60,$F8,$2F
	.dc.b	$CF,$F8,$1F,$D8,$F1,$BE,$22,$99,$AF,$8F,$EF,$57,$CF,$5F,$5E,$3F
	.dc.b	$73,$E7,$BE,$AF,$F2,$BE,$FC,$BF,$F0,$1C,$E5,$BE,$67,$0A,$91,$83
	.dc.b	$7F,$3D,$D2,$64,$60,$E7,$EC,$7E,$8B,$EA,$E4,$60,$EF,$F3,$FE,$49
	.dc.b	$91,$83,$C7,$E9,$7E,$D7,$E7,$24,$60,$F3,$F9,$FE,$C9,$91,$83,$D7
	.dc.b	$EF,$7F,$18,$F8,$27,$45,$3E,$1C,$7C,$4E,$BB,$5C,$0F,$86,$FC,$8F
	.dc.b	$A4,$D4,$60,$F8,$EF,$F2,$4F,$AB,$77,$DF,$02,$57,$F8,$BA,$72,$DF
	.dc.b	$9D,$E1,$F1,$12,$30,$7F,$8D,$3E,$AD,$F1,$1F,$CC,$17,$1D,$31,$19
	.dc.b	$E0,$D7,$C0,$CB,$E5,$1F,$CE,$3E,$AF,$F9,$17,$97,$FC,$95,$FE,$4F
	.dc.b	$E5,$7F,$45,$A4,$61,$FE,$90,$F0,$8D,$2F,$92,$E9,$E9,$CA,$FF,$25
	.dc.b	$CE,$5F,$F2,$6F,$7B,$E2,$FF,$A4,$BF,$BB,$8F,$E0,$8D,$C7,$F9,$C3
	.dc.b	$B8,$E3,$E3,$70,$F9,$99,$18,$3F,$CB,$3F,$56,$AF,$EA,$AB,$C1,$A6
	.dc.b	$23,$3C,$3A,$EF,$97,$CA,$38,$FB,$DC,$BD,$F9,$3D,$A3,$FB,$0F,$D5
	.dc.b	$BC,$2B,$EE,$CA,$FF,$35,$CE,$5F,$F3,$6F,$D5,$B6,$FE,$D2,$BE,$42
	.dc.b	$98,$8F,$ED,$BD,$DE,$9A,$FF,$3A,$BF,$DB,$5B,$E3,$EB,$83,$7E,$1F
	.dc.b	$0F,$AB,$95,$C2,$39,$EF,$D3,$C3,$93,$D2,$3F,$BD,$FD,$5F,$F6,$08
	.dc.b	$FF,$41,$BF,$D7,$5B,$F0,$1A,$E0,$FF,$00,$FD,$5F,$F3,$99,$FF,$06
	.dc.b	$57,$F3,$27,$8F,$B1,$AE,$FD,$F0,$F0,$F5,$65,$7F,$A4,$30,$F8,$4D
	.dc.b	$EB,$EB,$C7,$23,$06,$FD,$4F,$0C,$C8,$C1,$CF,$C7,$F2,$F8,$F2,$30
	.dc.b	$77,$F4,$BD,$33,$23,$07,$FA,$8B,$EA,$9F,$21,$FE,$1E,$FA,$B7,$BD
	.dc.b	$4F,$F1,$0C,$EA,$B8,$78,$38,$B9,$18,$3F,$D5,$BF,$39,$FA,$B7,$FC
	.dc.b	$61,$8F,$F8,$AB,$D3,$3F,$E2,$C9,$D1,$FA,$F9,$FF,$1C,$38,$F8,$9D
	.dc.b	$7C,$4F,$F8,$CB,$EA,$FF,$09,$5F,$DB,$70,$72,$75,$35,$E0,$6F,$C0
	.dc.b	$B5,$FE,$C0,$BE,$3E,$4F,$F8,$9D,$FC,$84,$7F,$77,$7F,$58,$EE,$79
	.dc.b	$3C,$7C,$8F,$16,$DE,$BE,$59,$3D,$63,$CD,$C1,$A8,$C1,$E1,$7D,$5E
	.dc.b	$47,$F8,$D2,$3F,$D6,$78,$FF,$1E,$CB,$FC,$F4,$D8,$CC,$ED,$FE,$30
	.dc.b	$23,$F9,$3F,$D5,$F0,$F4,$4D,$24,$FF,$9D,$BE,$93,$FC,$EE,$9A,$47
	.dc.b	$FD,$15,$F2,$1F,$E6,$AC,$7F,$C9,$91,$FE,$86,$EF,$FF,$74,$BF,$E6
	.dc.b	$DA,$FF,$37,$7D,$5B,$9B,$CB,$87,$1B,$F3,$F3,$5C,$3F,$98,$35,$FF
	.dc.b	$3B,$4D,$7F,$D0,$F3,$DB,$9B,$06,$F9,$9F,$39,$89,$F1,$FC,$BF,$13
	.dc.b	$17,$FE,$31,$3F,$D7,$27,$FA,$0F,$E5,$FB,$1B,$9F,$5B,$E3,$DB,$F9
	.dc.b	$64,$7F,$A2,$5F,$F1,$FF,$37,$87,$C3,$D3,$BE,$69,$5B,$3F,$D2,$B1
	.dc.b	$8C,$3F,$C8,$5F,$57,$F2,$5B,$B7,$AF,$C1,$F8,$53,$9C,$BF,$CF,$4F
	.dc.b	$C2,$AE,$3C,$1A,$BD,$BC,$3C,$79,$97,$86,$BF,$D4,$17,$FE,$77,$8C
	.dc.b	$42,$3F,$CF,$F1,$87,$F9,$E3,$7D,$0E,$73,$E8,$7C,$BE,$AC,$9E,$1F
	.dc.b	$1F,$6B,$DB,$EB,$FA,$8C,$69,$EA,$7D,$BF,$1B,$2F,$C1,$F5,$36,$FF
	.dc.b	$CC,$DF,$33,$F7,$27,$CE,$D8,$CF,$BF,$F8,$78,$C6,$3F,$E9,$F5,$E5
	.dc.b	$64,$D7,$8F,$70,$F8,$3F,$56,$F9,$DF,$C8,$36,$33,$E1,$62,$73,$DB
	.dc.b	$27,$8F,$BD,$75,$FF,$3E,$7D,$5F,$E3,$09,$E0,$7E,$6D,$C7,$05,$31
	.dc.b	$C5,$70,$FA,$7F,$57,$FC,$89,$BB,$7E,$8E,$1D,$76,$3F,$4D,$3F,$E2
	.dc.b	$43,$ED,$97,$2F,$F4,$73,$FE,$26,$33,$B7,$EB,$7E,$AF,$FD,$C5,$1F
	.dc.b	$82,$6C,$F0,$F7,$B4,$EF,$59,$5B,$EC,$B3,$FE,$9B,$BF,$F1,$14,$F3
	.dc.b	$D3,$B7,$4F,$77,$F7,$9E,$33,$DC,$A6,$3E,$76,$DF,$E5,$25,$E6,$F0
	.dc.b	$38,$FA,$B6,$7F,$33,$F5,$7F,$B4,$D3,$FD,$58,$DE,$CF,$F7,$03,$FE
	.dc.b	$21,$9D,$F8,$9F,$C3,$63,$FD,$3D,$1C,$BE,$1F,$1E,$F6,$5D,$71,$C9
	.dc.b	$EF,$EE,$DC,$BF,$D7,$5E,$3F,$E9,$9B,$E2,$FF,$19,$8F,$D5,$C6,$2B
	.dc.b	$F5,$FF,$57,$FE,$BD,$6F,$CB,$7F,$1F,$84,$58,$FE,$45,$F5,$6F,$88
	.dc.b	$A7,$F2,$5F,$AB,$DD,$7E,$EA,$74,$DF,$0B,$89,$C1,$C1,$BF,$0F,$A9
	.dc.b	$A7,$07,$3E,$0F,$09,$DF,$E2,$2C,$77,$E1,$F2,$98,$D7,$FD,$B9,$8F
	.dc.b	$CC,$7F,$DB,$DF,$57,$F8,$CE,$5F,$6C,$04,$66,$FF,$CE,$1A,$3A,$7F
	.dc.b	$A1,$4E,$6B,$83,$FC,$8E,$BF,$CE,$CD,$8C,$9C,$38,$78,$FA,$52,$BF
	.dc.b	$8F,$E1,$F9,$17,$F9,$0D,$FF,$42,$1E,$97,$80,$E1,$87,$A9,$F1,$FA
	.dc.b	$92,$B8,$C5,$9F,$7F,$C2,$7A,$0F,$AD,$F2,$76,$BB,$FA,$7E,$C3,$15
	.dc.b	$F1,$9F,$7F,$85,$27,$CF,$F1,$D6,$BE,$5F,$87,$F1,$0C,$70,$FC,$C8
	.dc.b	$F9,$71,$F2,$3E,$6A,$5F,$C0,$F8,$0B,$5F,$53,$E0,$BE,$69,$8C,$39
	.dc.b	$70,$7F,$8F,$AF,$E2,$FD,$5F,$D2,$36,$E0,$FE,$93,$F5,$1F,$9B,$C1
	.dc.b	$7E,$38,$C3,$E5,$FA,$B7,$CB,$FF,$7D,$9C,$DF,$FA,$BD,$FF,$52,$3D
	.dc.b	$3A,$F5,$E0,$6D,$E1,$EC,$49,$FE,$B3,$F5,$6F,$AB,$C6,$9E,$29,$E9
	.dc.b	$F0,$1D,$CC,$EB,$E5,$7C,$5C,$92,$7F,$0B,$ED,$5B,$E4,$7E,$5F,$5F
	.dc.b	$F2,$5C,$FC,$0E,$3D,$63,$EF,$7D,$9F,$01,$27,$47,$C6,$7E,$07,$C5
	.dc.b	$48,$F4,$F5,$DC,$7B,$8C,$5F,$57,$F1,$CD,$65,$F3,$5F,$95,$C7,$30
	.dc.b	$FE,$F7,$EC,$3C,$6C,$63,$D3,$91,$EF,$7F,$8B,$5B,$7F,$6B,$0C,$4F
	.dc.b	$F8,$C7,$EA,$FF,$95,$4F,$FB,$8D,$B7,$FB,$9F,$98,$FE,$DF,$39,$F4
	.dc.b	$B8,$7C,$24,$8F,$F0,$6D,$F0,$B3,$FE,$8C,$8C,$FB,$7D,$CF,$F2,$06
	.dc.b	$1F,$02,$32,$72,$F8,$0F,$21,$B7,$07,$6D,$7D,$07,$F9,$27,$EA,$3D
	.dc.b	$B8,$FC,$CC,$7E,$25,$79,$3C,$EF,$71,$C1,$C1,$FE,$24,$F7,$7F,$B1
	.dc.b	$D7,$F9,$B1,$AF,$F1,$90,$E1,$FE,$38,$D7,$27,$8F,$E6,$BF,$CB,$91
	.dc.b	$FC,$6E,$18,$27,$FC,$C1,$F5,$6F,$8C,$A6,$37,$B8,$FF,$8E,$27,$4F
	.dc.b	$FA,$03,$0F,$E5,$91,$87,$0D,$7F,$D0,$AD,$7D,$58,$FF,$92,$DF,$77
	.dc.b	$C0,$7F,$9F,$1F,$5B,$5F,$7B,$63,$FC,$A5,$CF,$FC,$45,$F8,$77,$F2
	.dc.b	$67,$DF,$1F,$03,$CD,$93,$F3,$7D,$0B,$87,$E8,$2B,$F9,$8C,$FD,$1A
	.dc.b	$F8,$F5,$AC,$3B,$45,$9F,$4C,$DB,$FA,$33,$5F,$AB,$16,$FA,$C7,$79
	.dc.b	$EB,$5D,$C9,$19,$7C,$C6,$2F,$C2,$D3,$C2,$58,$F0,$B5,$F0,$B6,$BC
	.dc.b	$33,$CA,$7E,$05,$F9,$6F,$CC,$DB,$E0,$D9,$F2,$1C,$23,$A4,$FD,$32
	.dc.b	$3F,$83,$8F,$CA,$C6,$1F,$C3,$5F,$C2,$7F,$39,$19,$6C,$FF,$A1,$35
	.dc.b	$FD,$3C,$78,$30,$FE,$9D,$18,$7D,$46,$38,$3E,$84,$66,$BF,$1B,$7F
	.dc.b	$B3,$9C,$92,$3E,$34,$EE,$FE,$49,$FC,$89,$C3,$EA,$8F,$24,$F1,$8F
	.dc.b	$24,$7E,$CE,$77,$87,$CE,$8C,$56,$F0,$F3,$C6,$0F,$0C,$65,$F3,$9D
	.dc.b	$8C,$92,$FB,$9F,$E1,$9B,$EF,$87,$B6,$75,$7D,$AB,$57,$FC,$03,$CF
	.dc.b	$95,$F3,$7D,$AB,$5F,$A1,$9F,$D6,$12,$7E,$D6,$5F,$EE,$97,$EE,$7D
	.dc.b	$5F,$F8,$97,$A2,$FA,$9F,$CF,$30,$FE,$05,$2F,$9E,$7A,$BA,$BF,$C2
	.dc.b	$6E,$F9,$6F,$C3,$61,$FC,$49,$B1,$98,$DC,$6B,$EA,$DC,$74,$93,$F8
	.dc.b	$F8,$CB,$E6,$F8,$5E,$B9,$9A,$48,$FC,$B7,$D5,$BD,$CF,$E5,$9C,$9F
	.dc.b	$8E,$6F,$07,$DD,$6D,$7F,$92,$6B,$83,$FE,$DA,$5E,$E6,$A1,$1A,$7F
	.dc.b	$B6,$B6,$FF,$C5,$0F,$08,$E5,$F0,$1C,$76,$7C,$9D,$3A,$52,$6F,$36
	.dc.b	$19,$7E,$38,$F1,$7D,$5F,$C4,$8F,$CE,$65,$D4,$CF,$06,$B9,$AC,$FF
	.dc.b	$BD,$2B,$86,$9F,$E3,$D7,$FA,$5C,$7F,$81,$BE,$A1,$F5,$7E,$FF,$6D
	.dc.b	$EF,$53,$E1,$4E,$AB,$FD,$FE,$BC,$EC,$DC,$9C,$C7,$0F,$90,$F9,$5F
	.dc.b	$57,$F1,$67,$FC,$2A,$DC,$FF,$CA,$18,$2B,$BF,$C0,$71,$BE,$F1,$87
	.dc.b	$C3,$F2,$3F,$57,$EA,$FF,$CD,$F1,$FE,$83,$C3,$FC,$B5,$18,$8F,$74
	.dc.b	$FA,$B5,$F5,$B2,$3D,$71,$EE,$F8,$1D,$46,$0F,$7B,$F8,$3F,$A9,$FF
	.dc.b	$33,$61,$FE,$67,$FA,$7F,$F4,$75,$78,$7A,$FF,$BE,$46,$23,$FC,$67
	.dc.b	$5C,$27,$FC,$00,$BE,$03,$A4,$49,$E4,$7F,$A7,$67,$4E,$03,$FE,$E1
	.dc.b	$71,$F5,$6F,$9F,$FC,$C3,$63,$84,$ED,$F1,$FC,$17,$AF,$E4,$64,$6A
	.dc.b	$3D,$DF,$55,$FA,$1F,$AB,$7A,$74,$CD,$74,$F8,$2A,$FB,$AB,$E3,$47
	.dc.b	$C7,$F8,$1B,$3B,$3C,$6F,$CB,$F9,$2F,$B5,$F5,$72,$7C,$F1,$F3,$FE
	.dc.b	$0B,$51,$83,$D6,$FD,$3F,$84,$FD,$B7,$1B,$E3,$29,$9A,$F8,$3F,$01
	.dc.b	$5F,$A7,$5F,$6E,$3F,$7D,$E9,$3E,$AF,$F2,$FF,$03,$2F,$E8,$4E,$5B
	.dc.b	$E6,$70,$A9,$18,$37,$F4,$1D,$26,$46,$0E,$7E,$EF,$E9,$3D,$E9,$18
	.dc.b	$3B,$FD,$07,$92,$64,$60,$F1,$FA,$5F,$B7,$F9,$49,$18,$3C,$FE,$7F
	.dc.b	$B2,$64,$60,$F5,$FB,$9F,$C7,$7E,$0D,$D1,$4F,$87,$1F,$13,$AA,$D7
	.dc.b	$03,$E1,$BF,$23,$D9,$D4,$60,$F8,$EF,$F2,$9F,$AB,$77,$5E,$F9,$5F
	.dc.b	$9B,$9C,$B7,$D8,$E1,$F0,$F2,$30,$7E,$87,$EA,$DF,$0F,$FC,$D5,$71
	.dc.b	$D3,$11,$9E,$0D,$7C,$AC,$BE,$51,$FC,$F7,$EA,$FF,$91,$F9,$7F,$EE
	.dc.b	$17,$F9,$3F,$95,$FE,$75,$6D,$03,$C2,$35,$F2,$9C,$FD,$29,$5F,$B7
	.dc.b	$A4,$FF,$4B,$EF,$7C,$5D,$F3,$A5,$FE,$F6,$72,$DF,$1F,$87,$CC,$C8
	.dc.b	$C1,$FC,$0F,$EA,$D5,$FD,$5D,$78,$54,$C4,$67,$8B,$5E,$04,$BE,$51
	.dc.b	$C7,$DE,$E5,$EF,$C9,$ED,$1F,$D8,$FE,$AD,$E1,$DF,$76,$57,$F1,$59
	.dc.b	$CB,$FC,$5F,$EA,$DB,$7F,$6B,$5F,$25,$4C,$47,$F6,$FE,$FF,$4D,$7F
	.dc.b	$1F,$7F,$B4,$B7,$C8,$D7,$06,$FC,$4E,$1F,$57,$2B,$84,$73,$DF,$A7
	.dc.b	$89,$27,$A4,$7F,$7D,$FA,$BF,$EB,$F1,$FC,$B5,$FE,$B8,$DF,$80,$D7
	.dc.b	$07,$F8,$0F,$EA,$FF,$9D,$4F,$E0,$EB,$F9,$73,$FE,$F0,$6F,$7B,$BF
	.dc.b	$FD,$7D,$5C,$28,$FC,$27,$F9,$E2,$F5,$7B,$1D,$F8,$BC,$1D,$7E,$1E
	.dc.b	$39,$7C,$23,$C9,$EA,$6A,$30,$74,$7D,$1E,$3F,$A7,$C7,$93,$E0,$8F
	.dc.b	$67,$A5,$A8,$C1,$E4,$7F,$C4,$5F,$54,$F9,$2F,$F1,$27,$D5,$BD,$EA
	.dc.b	$6A,$37,$E0,$DC,$3F,$AC,$5F,$16,$C7,$F8,$B7,$E4,$3F,$5B,$1F,$E3
	.dc.b	$0C,$BF,$C6,$1E,$9F,$F8,$FE,$3F,$B0,$4E,$8F,$D8,$3F,$B3,$71,$F1
	.dc.b	$3C,$FE,$27,$FC,$71,$F5,$7F,$88,$A7,$3F,$D3,$7F,$75,$87,$CF,$7F
	.dc.b	$B6,$35,$DF,$3F,$FB,$D6,$B9,$57,$19,$47,$E5,$9D,$9F,$C1,$5D,$CF
	.dc.b	$1F,$03,$C7,$E0,$79,$3C,$7B,$7F,$DE,$EF,$96,$1D,$5F,$DD,$C6,$78
	.dc.b	$3E,$1F,$34,$8C,$1C,$78,$3D,$F3,$23,$07,$5F,$1F,$E0,$BF,$EF,$DB
	.dc.b	$FE,$4F,$87,$FA,$F5,$B1,$99,$DB,$FC,$7C,$F6,$7A,$0F,$AB,$E0,$D8
	.dc.b	$E8,$D7,$81,$F1,$9C,$38,$1F,$49,$FE,$88,$69,$FE,$8A,$70,$7F,$A3
	.dc.b	$3E,$4B,$EC,$63,$F4,$67,$FD,$27,$C1,$FE,$23,$7F,$9A,$57,$F3,$5F
	.dc.b	$AB,$72,$F8,$B0,$E3,$7E,$4E,$5B,$87,$B5,$AF,$F3,$D9,$AF,$FA,$5A
	.dc.b	$7B,$72,$E0,$DF,$2B,$E7,$31,$3E,$3F,$97,$E2,$62,$FF,$0A,$7F,$49
	.dc.b	$7A,$FC,$BC,$9E,$07,$AD,$F1,$ED,$FC,$F2,$3F,$C8,$0D,$AE,$78,$7F
	.dc.b	$2A,$9C,$F3,$70,$E4,$91,$83,$8B,$EE,$FA,$BF,$93,$7E,$46,$E1,$F9
	.dc.b	$DC,$1D,$FB,$DB,$86,$F3,$2F,$F2,$96,$1C,$1F,$C3,$3F,$1F,$1E,$9E
	.dc.b	$37,$73,$C1,$E3,$DA,$E3,$E7,$F8,$CC,$57,$8D,$F3,$7A,$D2,$7B,$FC
	.dc.b	$75,$AF,$47,$AD,$EB,$31,$BF,$ED,$73,$FE,$DF,$7E,$6F,$8D,$F8,$07
	.dc.b	$C1,$FA,$9B,$1F,$E6,$BF,$98,$FB,$4D,$E7,$FC,$1C,$7F,$86,$23,$06
	.dc.b	$FC,$8C,$9A,$F1,$6C,$7F,$9D,$7E,$AD,$F3,$BF,$8C,$6C,$67,$C0,$C4
	.dc.b	$E7,$A2,$4F,$1F,$76,$F0,$FE,$23,$EA,$DD,$DF,$96,$6D,$7F,$A0,$D7
	.dc.b	$7E,$67,$5C,$12,$B5,$C9,$70,$FC,$9F,$D5,$FF,$BA,$E7,$FD,$18,$D8
	.dc.b	$CF,$6F,$E9,$23,$1C,$DF,$C8,$6B,$FC,$62,$7F,$A4,$CB,$5F,$9E,$7F
	.dc.b	$C6,$66,$77,$FD,$6F,$D5,$FF,$8A,$63,$FD,$34,$D9,$E1,$EE,$69,$DE
	.dc.b	$B2,$B7,$DD,$67,$F5,$77,$F7,$9B,$A7,$6E,$BE,$EF,$EF,$3C,$57,$B5
	.dc.b	$4C,$7C,$F5,$CB,$FD,$59,$2F,$33,$81,$C7,$D3,$B3,$FE,$B1,$FA,$BF
	.dc.b	$D9,$70,$FF,$43,$6C,$7F,$27,$7F,$C6,$2E,$FC,$7F,$E1,$B1,$FE,$AA
	.dc.b	$8E,$5F,$0F,$8F,$73,$2E,$B8,$E4,$F7,$F6,$EF,$0F,$F0,$AF,$27,$FC
	.dc.b	$63,$3F,$EC,$36,$C6,$7C,$9F,$E3,$71,$8D,$3F,$AC,$7D,$5F,$FB,$3E
	.dc.b	$FF,$D9,$4D,$BF,$E2,$AF,$F8,$5E,$33,$F0,$5F,$C9,$3E,$AD,$F0,$D4
	.dc.b	$FE,$4F,$F5,$7B,$C3,$F8,$EC,$EA,$FF,$DA,$CB,$C2,$E4,$71,$8C,$BD
	.dc.b	$7C,$39,$18,$38,$BE,$1E,$19,$1F,$EE,$1C,$3F,$CD,$2F,$68,$F3,$70
	.dc.b	$C8,$C1,$E2,$7F,$9E,$FD,$5F,$E3,$F6,$FE,$F4,$D9,$FE,$BD,$6E,$8D
	.dc.b	$E9,$C3,$82,$57,$FB,$C6,$F8,$4F,$FA,$0D,$FE,$44,$DF,$0F,$B9,$F1
	.dc.b	$BA,$AD,$71,$F4,$7C,$66,$2B,$FA,$2D,$A3,$CF,$E9,$4A,$E0,$78,$5F
	.dc.b	$5F,$A3,$23,$E2,$7C,$95,$C3,$9F,$3F,$C0,$31,$BF,$C0,$7E,$1F,$81
	.dc.b	$2F,$E6,$7C,$6D,$BF,$1F,$83,$F2,$0C,$61,$F2,$1F,$97,$F3,$52,$FE
	.dc.b	$A7,$C1,$5B,$F5,$FB,$FF,$40,$C6,$1F,$41,$FA,$7F,$57,$2F,$E6,$45
	.dc.b	$AF,$EC,$FF,$51,$F9,$5B,$FF,$18,$36,$BF,$C2,$9F,$56,$F9,$6F,$EF
	.dc.b	$33,$9B,$FF,$5B,$3F,$86,$79,$FC,$76,$BC,$1D,$BC,$3E,$4A,$63,$FE
	.dc.b	$BD,$CB,$FC,$7D,$F5,$6F,$AB,$CE,$3E,$4F,$6B,$B9,$8A,$F2,$BE,$2E
	.dc.b	$39,$87,$E5,$FD,$9B,$7E,$CE,$2F,$39,$8C,$3D,$AF,$AB,$E2,$A5,$F8
	.dc.b	$CF,$F1,$96,$FF,$E8,$57,$17,$EE,$9F,$83,$F1,$52,$73,$2D,$5D,$59
	.dc.b	$8F,$9B,$F5,$7F,$20,$C6,$7F,$39,$F9,$7C,$73,$0E,$7E,$BB,$C6,$C6
	.dc.b	$1F,$E1,$CE,$47,$BB,$FE,$2A,$6D,$F8,$B7,$FA,$CF,$92,$33,$5F,$E2
	.dc.b	$4F,$AB,$FE,$52,$FF,$BB,$5A,$BE,$5F,$FB,$AB,$BF,$4B,$FC,$6F,$3F
	.dc.b	$E9,$3B,$F7,$B6,$75,$18,$F8,$4D,$FB,$52,$3E,$2E,$0D,$F1,$7E,$3F
	.dc.b	$7B,$A9,$8E,$1F,$F4,$06,$5F,$DE,$A7,$D1,$BF,$88,$FF,$24,$FD,$47
	.dc.b	$B2,$B7,$1E,$6F,$3A,$67,$E6,$8B,$7F,$E4,$4F,$73,$FC,$66,$FF,$9A
	.dc.b	$1A,$FF,$9C,$5A,$8F,$EF,$FE,$6F,$F9,$8D,$C4,$7F,$12,$58,$FF,$72
	.dc.b	$7B,$4F,$9B,$E2,$3F,$CB,$F5,$FE,$37,$CB,$FD,$38,$DC,$7F,$D0,$18
	.dc.b	$7F,$7C,$9C,$FB,$BC,$0F,$06,$3F,$E1,$06,$DF,$F9,$06,$3D,$51,$8F
	.dc.b	$26,$61,$FE,$81,$9F,$F0,$BC,$61,$FF,$42,$52,$3F,$CE,$7F,$57,$E9
	.dc.b	$C3,$CB,$18,$CB,$D3,$F0,$F3,$1F,$CF,$FD,$5F,$F8,$89,$FD,$0E,$3F
	.dc.b	$D8,$5F,$95,$3B,$CB,$FC,$E7,$5B,$7E,$94,$67,$1F,$86,$FF,$A1,$27
	.dc.b	$51,$8A,$EB,$3B,$7F,$D4,$D3,$A7,$B9,$A2,$57,$82,$F8,$3F,$62,$AA
	.dc.b	$AE,$4E,$6B,$8D,$F8,$63,$94,$60,$EE,$FA,$9F,$87,$7D,$0F,$85,$1C
	.dc.b	$CE,$26,$C9,$3D,$4F,$F6,$F5,$78,$A3,$FC,$9B,$7C,$CF,$89,$56,$FC
	.dc.b	$53,$F9,$5A,$E1,$3E,$08,$E3,$3F,$81,$8F,$F0,$BC,$7E,$7E,$BF,$12
	.dc.b	$49,$D3,$63,$F0,$AD,$BF,$CB,$FC,$FF,$B4,$95,$E2,$AF,$F3,$1C,$61
	.dc.b	$F6,$C7,$EA,$E3,$13,$EF,$8C,$1C,$6D,$3A,$8F,$9B,$3E,$07,$51,$9A
	.dc.b	$F9,$93,$97,$B9,$FB,$48,$D3,$F0,$8F,$E0,$0E,$2B,$E3,$1E,$0A,$EA
	.dc.b	$49,$E6,$7B,$1C,$57,$A6,$3B,$47,$E8,$BD,$3F,$DA,$E3,$FE,$86,$9F
	.dc.b	$F7,$83,$5F,$F7,$AB,$5F,$F7,$F7,$D5,$FF,$88,$7B,$3F,$6C,$D8,$FF
	.dc.b	$7C,$2F,$9F,$71,$6B,$D6,$FE,$83,$5F,$EF,$DF,$AB,$7D,$07,$EF,$E1
	.dc.b	$18,$7F,$81,$35,$E1,$1B,$CB,$FC,$45,$1F,$C1,$DA,$FA,$BD,$6D,$1B
	.dc.b	$F5,$A4,$6D,$AF,$8F,$E4,$FC,$58,$1A,$48,$F1,$FB,$9F,$CB,$5F,$F7
	.dc.b	$BF,$4F,$F9,$65,$FE,$D6,$D7,$FD,$F7,$57,$89,$C5,$C6,$1E,$E7,$39
	.dc.b	$CD,$7F,$39,$9E,$15,$BC,$34,$F1,$BB,$5B,$F8,$0A,$FF,$3C,$CA,$E5
	.dc.b	$EB,$8F,$EF,$F8,$AB,$6B,$E0,$6C,$49,$DC,$74,$FA,$B7,$6F,$4C,$A5
	.dc.b	$7F,$82,$A7,$2D,$E8,$E1,$52,$30,$7F,$84,$3E,$A5,$F5,$7E,$8F,$6F
	.dc.b	$F9,$E4,$FF,$85,$1A,$B8,$7E,$2A,$AE,$2F,$3F,$3B,$E0,$C6,$23,$87
	.dc.b	$0F,$D7,$8C,$57,$27,$FC,$3F,$3E,$3F,$91,$F9,$3F,$57,$F5,$23,$FC
	.dc.b	$4A,$D7,$99,$FD,$B6,$53,$E0,$F7,$FC,$2C,$9F,$74,$79,$3E,$47,$EA
	.dc.b	$FD,$5F,$E5,$B1,$F5,$B7,$2F,$34,$FA,$B4,$E1,$87,$C7,$8C,$3F,$27
	.dc.b	$DF,$AF,$90,$F8,$93,$F0,$A3,$F0,$7F,$53,$F2,$31,$F9,$3F,$4F,$FE
	.dc.b	$99,$BF,$F2,$0B,$5E,$53,$BC,$7F,$3C,$BE,$03,$9C,$49,$FF,$24,$CE
	.dc.b	$5B,$F0,$DF,$92,$FA,$A7,$D5,$D3,$C7,$1E,$4F,$83,$B5,$C0,$E6,$FA
	.dc.b	$3E,$AB,$F3,$7F,$56,$F4,$A9,$9A,$F0,$FC,$1D,$7D,$35,$F0,$23,$E0
	.dc.b	$FC,$15,$9D,$9E,$87,$E2,$FC,$8F,$D4,$F4,$24,$FB,$23,$E4,$FC,$16
	.dc.b	$A3,$07,$C0,$7E,$6F,$C2,$7E,$C3,$8D,$F1,$54,$CD,$7C,$6F,$83,$AF
	.dc.b	$CE,$AF,$AB,$1F,$B8,$F4,$7F,$8F,$73,$FE,$7B,$BD,$F0,$1F,$94,$35
	.dc.b	$E8,$6F,$CC,$B9,$7F,$9E,$B0,$E3,$D5,$FC,$55,$FE,$01,$1F,$85,$5F
	.dc.b	$3D,$C6,$38,$57,$86,$3C,$5F,$3D,$FA,$2F,$AB,$95,$D6,$3C,$DE,$D6
	.dc.b	$A3,$07,$85,$F5,$7A,$3F,$BB,$F6,$A4,$F9,$63,$DD,$E7,$6A,$30,$7A
	.dc.b	$5F,$85,$EE,$FF,$1A,$F8,$17,$5D,$33,$5E,$FE,$DA,$FB,$6B,$E5,$47
	.dc.b	$CB,$FA,$4B,$3B,$3E,$33,$FC,$9B,$EA,$DE,$0D,$F0,$A5,$7F,$A9,$27
	.dc.b	$2D,$F9,$DE,$1F,$13,$23,$07,$FA,$AB,$EA,$DF,$0F,$FC,$C9,$71,$53
	.dc.b	$11,$9E,$1D,$7C,$AC,$BE,$51,$FC,$EB,$EA,$DE,$BD,$F5,$E5,$7F,$AE
	.dc.b	$27,$2D,$F9,$3E,$1E,$A4,$8C,$1F,$EB,$EF,$01,$F1,$9F,$D1,$9F,$DC
	.dc.b	$47,$F0,$96,$F8,$7F,$CB,$9D,$C7,$1F,$27,$87,$CC,$C8,$C1,$FE,$CA
	.dc.b	$FA,$BF,$CB,$B7,$C1,$FE,$A7,$1F,$90,$8D,$5E,$38,$B3,$E0,$49,$E5
	.dc.b	$1C,$3D,$DE,$3E,$F4,$9E,$D1,$FD,$7B,$EA,$DE,$1D,$F8,$A9,$5F,$ED
	.dc.b	$C9,$CB,$FE,$DE,$FA,$B6,$DF,$D9,$97,$C9,$D3,$11,$FD,$AB,$BF,$D1
	.dc.b	$5F,$CB,$5F,$D5,$F1,$7F,$52,$6F,$95,$FE,$03,$77,$19,$F1,$39,$7D
	.dc.b	$5C,$BF,$F7,$7D,$7F,$B9,$5B,$BF,$7F,$12,$46,$0F,$F7,$AF,$D5,$FF
	.dc.b	$62,$C7,$FB,$0B,$FD,$62,$3F,$A0,$36,$33,$BF,$F8,$0B,$EA,$FF,$9D
	.dc.b	$57,$FB,$0E,$BF,$99,$BD,$3D,$9D,$6B,$BD,$F8,$7A,$F2,$BF,$D9,$59
	.dc.b	$7F,$70,$C7,$E3,$C6,$2F,$FC,$45,$0E,$AF,$07,$0C,$AD,$1F,$ED,$39
	.dc.b	$D3,$80,$BD,$6F,$45,$49,$D4,$E5,$BE,$3F,$B7,$C7,$91,$83,$7E,$A7
	.dc.b	$C1,$99,$18,$3C,$1F,$54,$F9,$5F,$F1,$3F,$D5,$BE,$02,$9F,$E2,$A9
	.dc.b	$D5,$70,$EF,$C5,$C8,$C1,$EB,$F8,$BF,$F1,$0B,$FD,$CF,$D3,$65,$FE
	.dc.b	$34,$6C,$6F,$C9,$FF,$1B,$D7,$EE,$A3,$FB,$D3,$CB,$8F,$FC,$3E,$DE
	.dc.b	$3F,$F2,$83,$B8,$FF,$11,$FD,$5F,$E4,$AB,$FE,$00,$60,$E5,$F0,$1F
	.dc.b	$E4,$49,$D4,$9F,$CB,$E1,$FE,$5B,$72,$4F,$1B,$87,$8B,$C5,$95,$F6
	.dc.b	$67,$F7,$EF,$F8,$45,$F8,$EF,$C1,$E5,$F8,$07,$87,$6F,$16,$F6,$79
	.dc.b	$46,$1E,$9E,$2F,$F8,$FD,$FF,$37,$03,$D0,$E5,$CD,$FF,$2E,$D7,$71
	.dc.b	$B8,$C3,$87,$F9,$AF,$D5,$EA,$3E,$17,$9E,$8C,$4E,$A7,$84,$E5,$FF
	.dc.b	$4B,$7D,$27,$FA,$5D,$34,$8F,$5F,$92,$FF,$3D,$E1,$FE,$58,$8F,$F5
	.dc.b	$1E,$FF,$A1,$7D,$95,$ED,$FA,$B7,$1F,$C0,$C3,$8D,$FC,$2E,$4B,$87
	.dc.b	$FA,$45,$AF,$C5,$9A,$FF,$9D,$A7,$B7,$A1,$83,$7C,$AF,$9D,$C4,$F8
	.dc.b	$FE,$5F,$89,$8B,$FF,$26,$1F,$EC,$5D,$7D,$5F,$27,$D8,$DD,$7A,$DF
	.dc.b	$1C,$7F,$86,$23,$F7,$CD,$9F,$D2,$78,$5A,$E8,$DF,$2C,$BD,$C5,$AF
	.dc.b	$F2,$E7,$D5,$FC,$86,$ED,$EA,$F0,$3E,$0C,$E7,$2F,$F4,$2B,$F0,$6B
	.dc.b	$8F,$06,$AF,$6F,$0F,$1A,$65,$FC,$4E,$DF,$21,$8C,$3C,$3E,$7F,$63
	.dc.b	$19,$76,$7C,$3E,$9C,$C3,$97,$C7,$DB,$F4,$FA,$BE,$63,$18,$79,$9F
	.dc.b	$4F,$C7,$4B,$F1,$7D,$4D,$BF,$F3,$E7,$CC,$FD,$79,$FE,$50,$D8,$CF
	.dc.b	$BD,$F8,$28,$C6,$3F,$EA,$65,$E5,$64,$D7,$8F,$70,$FE,$65,$F5,$6F
	.dc.b	$9D,$FC,$5B,$63,$3E,$16,$27,$3D,$D2,$78,$FB,$B7,$5F,$F4,$97,$D5
	.dc.b	$FE,$1E,$7F,$DC,$CD,$8F,$DD,$3F,$E1,$C8,$FE,$84,$E6,$0B,$7F,$CE
	.dc.b	$B6,$77,$C3,$89,$FD,$0F,$D5,$FF,$2E,$9F,$E9,$8D,$BF,$F5,$22,$EB
	.dc.b	$DE,$DF,$A8,$AF,$F1,$B9,$FD,$4C,$B5,$FE,$AB,$7F,$C7,$26,$75,$FD
	.dc.b	$7F,$D5,$FF,$13,$8F,$EB,$4D,$9E,$1E,$EE,$9D,$F8,$92,$B7,$DB,$67
	.dc.b	$FD,$71,$6D,$7A,$9D,$BA,$6D,$FF,$67,$F1,$DE,$CF,$F0,$06,$C6,$7E
	.dc.b	$77,$13,$9F,$3A,$4F,$1F,$5A,$E7,$FE,$C7,$FA,$BF,$DA,$76,$7F,$BB
	.dc.b	$1B,$5F,$CB,$E5,$1F,$E8,$88,$FE,$F0,$EA,$3F,$9A,$3F,$EB,$58,$CF
	.dc.b	$C3,$7F,$14,$87,$BD,$63,$1C,$52,$B9,$FB,$B7,$0F,$F6,$D7,$8F,$FE
	.dc.b	$3C,$9F,$F0,$33,$63,$3E,$3F,$F2,$08,$C6,$6A,$F4,$7D,$5F,$FB,$5D
	.dc.b	$FF,$07,$36,$FF,$DC,$CE,$E3,$3F,$2F,$FC,$AB,$EA,$DF,$0F,$4F,$E5
	.dc.b	$BF,$57,$BC,$3F,$DD,$D3,$AB,$FF,$0D,$2F,$07,$91,$C6,$32,$F5,$F0
	.dc.b	$A4,$60,$E2,$F8,$78,$64,$7F,$89,$70,$FF,$3B,$BD,$A3,$CD,$C3,$23
	.dc.b	$07,$89,$FE,$87,$F5,$7F,$93,$DB,$E2,$67,$FD,$A6,$31,$5F,$77,$07
	.dc.b	$3F,$0A,$E3,$C1,$2B,$FC,$6D,$7C,$27,$94,$F6,$F4,$A5,$79,$FA,$2D
	.dc.b	$32,$F0,$AC,$1E,$97,$98,$DC,$EA,$33,$7E,$0F,$53,$D6,$6D,$C1,$E4
	.dc.b	$F3,$FD,$E7,$D6,$F9,$4B,$1E,$8E,$7F,$86,$63,$2F,$24,$7E,$77,$04
	.dc.b	$FC,$8E,$E9,$3F,$47,$E3,$6D,$F0,$3E,$57,$81,$F3,$8C,$E9,$E9,$7E
	.dc.b	$8F,$CD,$CB,$FB,$5E,$F5,$BE,$47,$C3,$F8,$0F,$AE,$67,$0F,$84,$FD
	.dc.b	$AF,$AB,$97,$F4,$A2,$D7,$2F,$A8,$FC,$E5,$FC,$E6,$D7,$EE,$7E,$AD
	.dc.b	$F3,$1F,$E0,$69,$CD,$FF,$8A,$DF,$C6,$BC,$FE,$3F,$5E,$0E,$DE,$1E
	.dc.b	$B4,$C7,$FC,$F1,$F5,$4F,$AB,$C6,$1E,$3F,$80,$EC,$63,$5F,$23,$E1
	.dc.b	$E3,$98,$7E,$77,$D8,$B7,$EB,$E2,$F3,$18,$C3,$D8,$FA,$7E,$02,$5F
	.dc.b	$FA,$27,$7F,$F2,$F3,$63,$DD,$F1,$52,$3C,$FD,$77,$1F,$91,$F5,$7F
	.dc.b	$0C,$C6,$5F,$25,$F8,$DC,$72,$FF,$D2,$DE,$B7,$F1,$A7,$C1,$1F,$E3
	.dc.b	$0E,$37,$B9,$4E,$18,$FD,$E8,$CD,$7F,$A7,$FE,$AF,$F9,$44,$9B,$F2
	.dc.b	$FF,$DE,$27,$F9,$84,$68,$FF,$4B,$3C,$7E,$13,$FC,$7F,$19,$6F,$B3
	.dc.b	$27,$FD,$59,$4D,$31,$87,$E4,$30,$FE,$F8,$DF,$7B,$3B,$7E,$DA,$72
	.dc.b	$DF,$33,$B9,$FE,$4E,$FA,$8F,$5D,$F3,$8F,$1F,$9B,$33,$F7,$45,$9F
	.dc.b	$CF,$FB,$7F,$E3,$59,$FE,$99,$7F,$DB,$27,$3E,$67,$5F,$89,$95,$FE
	.dc.b	$C5,$AD,$65,$FC,$2B,$0E,$58,$7F,$A0,$E7,$35,$FD,$5E,$7F,$C8,$15
	.dc.b	$BF,$73,$D1,$C1,$FE,$66,$9C,$D6,$E7,$11,$96,$C7,$FB,$56,$3F,$82
	.dc.b	$BF,$ED,$7F,$AB,$EF,$8F,$FA,$4E,$72,$FE,$F2,$32,$9F,$73,$FE,$9A
	.dc.b	$6B,$FD,$B5,$AF,$FA,$75,$AA,$1F,$41,$E9,$3D,$5F,$0D,$6F,$CD,$16
	.dc.b	$BE,$7C,$DC,$BE,$44,$DA,$F8,$D3,$6B,$D5,$35,$F0,$45,$BF,$0D,$76
	.dc.b	$3E,$24,$66,$B4,$D8,$D5,$78,$9B,$7E,$26,$9E,$22,$C7,$89,$AF,$89
	.dc.b	$B7,$E2,$8F,$C5,$DF,$C2,$9C,$BE,$26,$9C,$2F,$C9,$3F,$EC,$4A,$E4
	.dc.b	$FF,$B3,$1A,$FE,$26,$BF,$D6,$CF,$8C,$E0,$E1,$FA,$8D,$8E,$31,$BB
	.dc.b	$FE,$BF,$F5,$5F,$C9,$70,$F5,$B6,$7D,$31,$FE,$DB,$9D,$F8,$5F,$46
	.dc.b	$B5,$1F,$1E,$BF,$A3,$C6,$2B,$CC,$71,$3F,$C6,$CE,$63,$C5,$18,$AE
	.dc.b	$84,$7C,$6D,$8F,$19,$D9,$C5,$78,$4F,$2C,$F0,$24,$F0,$8C,$B6,$3C
	.dc.b	$06,$A3,$93,$88,$D1,$FE,$FB,$AF,$43,$F3,$23,$D8,$E6,$BC,$71,$F0
	.dc.b	$23,$CC,$E9,$7D,$5F,$F8,$6B,$A7,$FD,$F8,$D8,$FF,$2D,$03,$7F,$3E
	.dc.b	$E2,$DF,$AB,$FC,$EA,$76,$32,$7D,$4E,$23,$D0,$D7,$4D,$8D,$C6,$FE
	.dc.b	$AD,$F4,$5F,$E6,$BD,$BF,$C4,$71,$89,$F5,$35,$FB,$FE,$B6,$F6,$FB
	.dc.b	$AD,$8F,$90,$D7,$CC,$DC,$3F,$DB,$31,$F1,$57,$AD,$92,$47,$83,$E8
	.dc.b	$B9,$60,$69,$23,$CB,$F2,$9A,$F7,$7C,$1D,$5F,$E5,$F9,$FE,$6D,$19
	.dc.b	$7D,$7E,$ED,$7F,$C3,$F8,$6D,$B1,$E0,$F8,$1E,$78,$C9,$F8,$71,$E4
	.dc.b	$F8,$2D,$46,$0F,$8E,$FC,$1F,$AB,$76,$F9,$30,$95,$F8,$59,$CB,$7A
	.dc.b	$B8,$54,$8C,$1F,$8B,$FA,$BF,$B2,$FF,$8D,$BD,$9F,$F2,$2B,$F8,$F6
	.dc.b	$BF,$E3,$A8,$DF,$27,$F9,$79,$FA,$31,$9E,$6F,$AD,$1A,$E5,$FF,$56
	.dc.b	$3A,$AE,$1F,$21,$F6,$FE,$95,$AD,$38,$C7,$2F,$76,$E5,$E7,$8C,$3E
	.dc.b	$9F,$91,$FC,$27,$D5,$FF,$9B,$A3,$F4,$CD,$7D,$11,$E4,$D7,$CA,$C8
	.dc.b	$F8,$51,$E8,$F7,$75,$18,$3E,$33,$F9,$1F,$AB,$FF,$3E,$E9,$FE,$7E
	.dc.b	$FA,$8F,$F4,$5D,$7E,$D5,$AF,$F9,$82,$77,$BF,$FB,$35,$7B,$DC,$E2
	.dc.b	$4F,$EF,$67,$2D,$F8,$5F,$D0,$7D,$5B,$E7,$E9,$E3,$8F,$27,$C0,$DA
	.dc.b	$E0,$73,$7D,$1F,$55,$FA,$9F,$AB,$7A,$54,$CD,$78,$7E,$06,$BE,$9A
	.dc.b	$F8,$11,$F0,$7E,$02,$CE,$CF,$43,$F1,$7E,$47,$E9,$FA,$12,$7D,$91
	.dc.b	$F2,$7D,$FD,$46,$0F,$80,$FC,$DF,$82,$FD,$F7,$1B,$E2,$A9,$9A,$F8
	.dc.b	$DE,$FD,$7E,$75,$7D,$58,$FE,$0D,$E8,$FF,$0D,$E7,$FF,$59,$F7,$BE
	.dc.b	$6B,$F2,$06,$BD,$0D,$F9,$37,$2F,$E5,$D8,$71,$EA,$FE,$55,$08,$FE
	.dc.b	$10,$EB,$CD,$E2,$F0,$AF,$0C,$78,$BE,$7B,$F3,$DF,$57,$2B,$AC,$79
	.dc.b	$BD,$BD,$46,$0F,$0B,$EA,$F3,$FF,$90,$78,$4F,$66,$99,$AF,$3F,$91
	.dc.b	$5F,$F5,$7E,$9F,$C4,$78,$7F,$DA,$6F,$EB,$A3,$DB,$EE,$7F,$2B,$F8
	.dc.b	$07,$65,$3E,$4C,$7C,$AE,$EB,$5C,$0F,$8A,$FC,$EF,$A4,$D4,$60,$F9
	.dc.b	$2F,$F3,$7F,$AB,$78,$57,$C3,$95,$FD,$6E,$72,$DF,$99,$E1,$F0,$D2
	.dc.b	$30,$7F,$61,$FA,$B7,$C1,$7F,$43,$5C,$B4,$C4,$67,$97,$5F,$2B,$2F
	.dc.b	$94,$7F,$4B,$FA,$BF,$F2,$84,$7F,$6D,$7F,$94,$BF,$ED,$86,$F2,$7F
	.dc.b	$BF,$1D,$C6,$7E,$47,$97,$A7,$23,$94,$7F,$57,$F0,$1F,$15,$7D,$09
	.dc.b	$5F,$DE,$E7,$2D,$F2,$78,$7C,$CC,$8C,$1F,$E0,$1F,$AB,$FC,$D3,$7C
	.dc.b	$2F,$EC,$90,$6D,$E3,$87,$3E,$0C,$BE,$51,$C3,$DC,$E3,$EE,$C9,$ED
	.dc.b	$1F,$DB,$FE,$AF,$F2,$11,$FE,$11,$7F,$A1,$F1,$3C,$7F,$D2,$C6,$1B
	.dc.b	$E2,$7F,$77,$FA,$B6,$F7,$E4,$26,$3F,$E1,$B9,$CB,$FE,$1D,$EE,$F5
	.dc.b	$37,$C9,$FF,$00,$47,$EC,$21,$9F,$85,$FF,$0F,$47,$07,$39,$7E,$E2
	.dc.b	$1E,$1F,$2F,$AB,$95,$DE,$3A,$EF,$DB,$C3,$93,$E3,$8F,$F0,$87,$D5
	.dc.b	$FB,$63,$FC,$58,$FF,$5B,$C7,$F8,$2B,$63,$3F,$31,$FE,$17,$FA,$BF
	.dc.b	$E8,$31,$C3,$D9,$E6,$66,$F5,$C1,$BF,$5E,$57,$F8,$FE,$D3,$CF,$AF
	.dc.b	$A7,$1C,$AD,$47,$7F,$5B,$51,$83,$93,$E3,$F1,$FC,$9E,$3C,$9E,$D1
	.dc.b	$E7,$F5,$35,$18,$3C,$4F,$F8,$C7,$EA,$9F,$29,$FE,$33,$FA,$B7,$BF
	.dc.b	$4D,$46,$FC,$0B,$87,$F9,$66,$F8,$B6,$3F,$C7,$9F,$15,$FA,$EE,$2F
	.dc.b	$E5,$FE,$A3,$4F,$F2,$03,$6F,$FC,$C5,$5B,$7F,$CB,$CF,$ED,$3B,$3F
	.dc.b	$09,$1F,$E4,$2F,$86,$EA,$76,$F8,$7F,$F2,$57,$D5,$7E,$2E,$B7,$83
	.dc.b	$9F,$B9,$AF,$03,$7E,$3D,$CB,$FC,$E5,$87,$1D,$7F,$A2,$BF,$91,$C7
	.dc.b	$C2,$FE,$DA,$7F,$26,$7F,$9C,$DD,$4F,$0A,$F1,$F8,$FE,$4E,$1F,$7F
	.dc.b	$3C,$C3,$BC,$7A,$38,$75,$18,$3C,$6F,$B3,$C7,$FD,$2C,$7F,$9F,$32
	.dc.b	$FE,$4B,$9F,$F9,$ED,$B1,$99,$DB,$FE,$88,$8F,$E5,$BF,$57,$BF,$44
	.dc.b	$D2,$4F,$FA,$0F,$E9,$3F,$D0,$89,$A4,$7F,$D1,$5F,$23,$FB,$7C,$7E
	.dc.b	$AC,$7F,$A4,$F7,$FF,$4E,$3F,$E9,$7A,$FF,$4C,$7D,$5F,$F9,$EA,$3F
	.dc.b	$B8,$DF,$1A,$FF,$4E,$F2,$D9,$FE,$EA,$D7,$FD,$41,$35,$FF,$33,$CF
	.dc.b	$6E,$5C,$1B,$E5,$7C,$E6,$27,$C7,$F4,$3C,$4C,$5F,$DA,$3F,$D5,$8F
	.dc.b	$5F,$93,$93,$C0,$F5,$BE,$3D,$7F,$C2,$11,$FE,$B9,$6D,$73,$C3,$F7
	.dc.b	$B3,$9E,$5E,$1C,$92,$30,$71,$7D,$DF,$57,$F2,$6F,$C8,$DC,$3F,$37
	.dc.b	$E0,$6B,$6D,$EF,$31,$FF,$65,$5E,$DF,$F0,$EB,$F1,$B0,$E7,$E3,$76
	.dc.b	$3B,$FC,$ED,$AE,$3E,$7F,$88,$C5,$78,$9F,$2F,$AD,$27,$BF,$C7,$DA
	.dc.b	$F3,$FA,$FE,$A3,$1B,$7E,$E3,$34,$7B,$BE,$3A,$61,$E8,$FA,$9B,$87
	.dc.b	$D3,$F9,$9F,$B3,$3F,$EE,$56,$C6,$7E,$03,$F0,$91,$8C,$BF,$82,$AF
	.dc.b	$2B,$26,$BC,$7B,$87,$FB,$BF,$EA,$DF,$3D,$F8,$D6,$C6,$7C,$1C,$4E
	.dc.b	$7A,$A4,$F1,$F7,$EE,$BF,$89,$FA,$B7,$77,$E5,$DB,$5F,$EF,$D5,$E0
	.dc.b	$66,$75,$C3,$2B,$5C,$77,$0F,$CA,$7D,$5F,$F2,$79,$FF,$80,$1C,$5F
	.dc.b	$C8,$A3,$13,$AE,$DF,$D3,$46,$27,$F9,$3D,$7F,$8B,$0E,$27,$F9,$80
	.dc.b	$C5,$7E,$85,$FF,$17,$9A,$47,$F8,$0F,$7F,$D9,$7D,$5E,$27,$52,$8C
	.dc.b	$3C,$FD,$ED,$BC,$35,$95,$C2,$BF,$29,$87,$F3,$C7,$E2,$C6,$73,$EE
	.dc.b	$78,$3A,$AC,$78,$3C,$57,$B9,$FC,$15,$B1,$9F,$9D,$C4,$E7,$CC,$93
	.dc.b	$C7,$D5,$BA,$FF,$01,$FA,$BF,$DC,$EB,$FD,$39,$B5,$FE,$86,$7F,$C6
	.dc.b	$4F,$C0,$73,$3F,$19,$FF,$58,$C6,$7E,$23,$F8,$C4,$7F,$85,$A3,$07
	.dc.b	$1E,$0C,$9A,$F7,$2C,$7C,$CF,$1F,$FD,$08,$DF,$13,$F9,$14,$3C,$6B
	.dc.b	$1F,$C9,$7E,$AF,$FD,$A3,$5F,$51,$B9,$FF,$1D,$7F,$C7,$91,$9F,$84
	.dc.b	$FE,$59,$F5,$6F,$88,$A7,$F2,$FF,$AB,$DE,$1F,$C9,$67,$57,$F8,$05
	.dc.b	$E0,$72,$38,$C6,$5E,$BE,$0C,$8C,$1C,$5F,$0F,$C1,$C8,$FC,$5E,$1F
	.dc.b	$E6,$C7,$B4,$79,$B8,$A4,$60,$F1,$3F,$D1,$7E,$AF,$F2,$3C,$9F,$E0
	.dc.b	$66,$CF,$FA,$FA,$DD,$1B,$D3,$87,$14,$AF,$CD,$DF,$09,$FD,$0B,$FC
	.dc.b	$AD,$BE,$17,$73,$E3,$74,$DA,$E3,$E8,$F8,$CC,$57,$F4,$DB,$47,$9F
	.dc.b	$DB,$95,$C0,$F0,$BE,$BF,$3E,$47,$C4,$F9,$5B,$87,$3E,$8F,$80,$63
	.dc.b	$7F,$80,$FC,$3E,$F9,$7F,$33,$E3,$AD,$F8,$FC,$1F,$90,$63,$0F,$90
	.dc.b	$FC,$BF,$91,$97,$F5,$3D,$EB,$7E,$BF,$80,$FA,$06,$30,$FA,$0F,$D3
	.dc.b	$FA,$B9,$7F,$32,$2D,$7F,$6F,$FA,$9F,$CA,$DF,$F8,$F1,$B5,$FE,$F4
	.dc.b	$FA,$B7,$CC,$7F,$80,$67,$37,$FC,$51,$FC,$33,$CF,$E4,$35,$E0,$ED
	.dc.b	$E1,$EB,$4C,$7F,$8B,$FD,$5B,$EA,$31,$87,$8F,$DF,$EC,$63,$5F,$23
	.dc.b	$E1,$E3,$98,$7E,$5B,$D8,$B7,$EB,$E2,$F3,$18,$C3,$D8,$FA,$7E,$2A
	.dc.b	$5F,$F2,$8D,$FF,$CA,$8D,$8F,$77,$C5,$48,$F3,$F5,$5C,$7E,$47,$D5
	.dc.b	$FC,$33,$19,$7C,$97,$E3,$71,$CB,$FE,$6B,$EB,$7F,$11,$BF,$A1,$3F
	.dc.b	$E2,$BE,$37,$B9,$4E,$13,$FE,$23,$8C,$DF,$F3,$EF,$AB,$FE,$3F,$26
	.dc.b	$FC,$C7,$F7,$19,$FE,$8F,$39,$F8,$4F,$F1,$C5,$7F,$A3,$27,$97,$C2
	.dc.b	$62,$72,$EB,$DA,$93,$FD,$33,$06,$B8,$C7,$FC,$6D,$8F,$F7,$66,$FB
	.dc.b	$D9,$DF,$FD,$0B,$39,$6F,$9B,$E2,$3F,$C9,$5F,$51,$EC,$BE,$71,$E6
	.dc.b	$F3,$A6,$7E,$F8,$B3,$FE,$45,$F7,$3F,$AB,$CF,$B9,$B1,$FD,$9A,$FF
	.dc.b	$B5,$E1,$FE,$96,$9C,$CF,$F8,$01,$FF,$2D,$1F,$D7,$9E,$9F,$0B,$88
	.dc.b	$CF,$9B,$30,$FE,$DB,$83,$D7,$C6,$1E,$3F,$33,$99,$8C,$BC,$0F,$6F
	.dc.b	$5E,$61,$FD,$DB,$EA,$7C,$15,$E5,$9F,$1F,$D0,$4A,$FE,$F7,$A7,$F5
	.dc.b	$0C,$7E,$04,$FF,$9C,$7E,$AF,$FC,$0F,$1F,$78,$38,$7F,$02,$D7,$FC
	.dc.b	$12,$D7,$7C,$5C,$CE,$9C,$B7,$0F,$F0,$5D,$7D,$97,$9C,$E9,$CB,$FE
	.dc.b	$83,$CF,$FB,$14,$7B,$72,$C7,$9D,$96,$47,$9F,$1F,$F0,$DD,$7B,$5F
	.dc.b	$53,$5F,$CC,$B5,$DE,$5D,$CE,$6F,$FA,$1E,$11,$99,$ED,$29,$CC,$61
	.dc.b	$B1,$E3,$6B,$E2,$F8,$8F,$67,$9F,$ED,$6D,$7D,$98,$FF,$56,$CF,$A9
	.dc.b	$AF,$76,$D1,$BE,$7F,$C7,$39,$CF,$A5,$F9,$5F,$07,$C0,$97,$E7,$6C
	.dc.b	$FE,$87,$0D,$CF,$E1,$9F,$CD,$9E,$2D,$CD,$24,$7C,$48,$CB,$63,$33
	.dc.b	$F5,$67,$2D,$C3,$F4,$E7,$6A,$FC,$AB,$87,$A3,$E1,$23,$F3,$FE,$1F
	.dc.b	$E1,$91,$C1,$F4,$FD,$1F,$8E,$47,$07,$D7,$F4,$3E,$59,$1C,$1F,$6F
	.dc.b	$D1,$F9,$E4,$70,$7D,$FF,$93,$FA,$64,$70,$7E,$1B,$8D,$F3,$DF,$69
	.dc.b	$B1,$98,$C1,$F3,$76,$91,$F2,$DB,$A7,$E7,$7E,$5B,$FC,$6F,$3F,$B5
	.dc.b	$6C,$7F,$34,$6B,$FE,$7C,$6A,$FA,$BF,$F2,$3C,$FE,$F5,$B1,$FC,$F1
	.dc.b	$AF,$FA,$11,$AF,$1F,$6F,$FA,$5B,$FA,$F8,$FF,$17,$F8,$FF,$A4,$5F
	.dc.b	$3E,$F7,$31,$7C,$2F,$C1,$E5,$FF,$4A,$ED,$65,$E9,$F5,$F0,$DA,$FF
	.dc.b	$48,$E4,$BC,$C7,$97,$E3,$2B,$ED,$8B,$3B,$8E,$33,$AF,$AB,$7C,$AF
	.dc.b	$0E,$C7,$BB,$4D,$46,$FE,$56,$E1,$FE,$A2,$BE,$2D,$8E,$9E,$7F,$8B
	.dc.b	$8D,$E8,$53,$35,$C7,$D0,$AB,$87,$E0,$BF,$EB,$2C,$58,$79,$EB,$D1
	.dc.b	$F0,$7E,$EF,$2B,$FC,$E7,$7F,$D2,$23,$2F,$97,$E0,$EB,$D7,$8F,$FB
	.dc.b	$BE,$3C,$23,$07,$79,$F8,$9E,$5F,$5D,$E4,$FB,$E3,$E4,$7C,$3E,$A3
	.dc.b	$07,$C3,$7E,$8F,$D5,$BC,$FF,$36,$12,$BF,$D9,$33,$96,$F8,$DC,$2A
	.dc.b	$46,$0F,$F6,$7F,$D5,$3E,$AF,$EE,$F8,$DF,$EA,$39,$FF,$69,$B5,$F5
	.dc.b	$89,$E0,$F2,$F6,$3F,$0B,$18,$8D,$B7,$93,$E8,$C7,$19,$DF,$95,$F8
	.dc.b	$CF,$A5,$6B,$4C,$D7,$9B,$CB,$AF,$4A,$EF,$1E,$0F,$2F,$F2,$BF,$57
	.dc.b	$FE,$84,$AF,$F7,$5B,$5F,$1C,$79,$B4,$E1,$1F,$0A,$30,$FC,$4F,$32
	.dc.b	$BE,$A9,$F6,$C7,$E8,$BE,$AF,$D7,$AF,$B3,$EA,$3F,$D4,$77,$FE,$FA
	.dc.b	$6B,$F2,$27,$7D,$1F,$E3,$C5,$E2,$F3,$89,$23,$84,$77,$F8,$AF,$D8
	.dc.b	$FD,$5B,$CF,$A6,$6B,$86,$5F,$3B,$17,$C4,$66,$7C,$58,$7A,$63,$D5
	.dc.b	$F5,$5F,$BC,$FA,$B7,$BB,$4C,$DF,$4D,$7D,$1F,$0F,$88,$F6,$5F,$C3
	.dc.b	$8F,$89,$F0,$D6,$78,$1E,$BC,$D3,$A8,$C8,$87,$CA,$7D,$AF,$66,$4E
	.dc.b	$A7,$2D,$F8,$5F,$A5,$52,$30,$6F,$E2,$3F,$8B,$F2,$3E,$36,$9F,$66
	.dc.b	$3E,$D7,$C3,$DA,$E0,$77,$7F,$8F,$FA,$EF,$AB,$FC,$B7,$9D,$2B,$D3
	.dc.b	$39,$6F,$95,$C2,$A4,$60,$DF,$A3,$D2,$64,$60,$E7,$F4,$1F,$A3,$F7
	.dc.b	$24,$60,$EF,$F4,$1E,$49,$91,$83,$C7,$E8,$7E,$EB,$E0,$64,$60,$F3
	.dc.b	$F8,$FE,$C9,$91,$83,$D7,$F0,$FF,$CF,$BE,$2B,$EB,$C7,$FA,$27,$A3
	.dc.b	$E1,$CF,$C4,$EC,$B3,$C0,$F8,$6F,$C8,$F5,$F5,$18,$3E,$3B,$FD,$2B
	.dc.b	$EA,$FF,$2B,$1F,$82,$5E,$0E,$4D,$78,$3B,$F6,$F8,$7B,$B2,$F8,$47
	.dc.b	$F5,$4F,$AB,$7C,$45,$E3,$95,$F8,$E9,$CB,$78,$38,$7C,$B4,$8C,$1F
	.dc.b	$94,$FA,$B7,$AF,$FD,$81,$FE,$59,$1F,$C8,$9B,$E4,$FF,$1C,$77,$1C
	.dc.b	$7E,$23,$87,$A5,$23,$07,$E8,$3B,$FF,$AC,$B7,$CE,$FE,$DB,$1F,$C1
	.dc.b	$23,$5E,$FE,$3C,$6C,$FC,$D4,$9E,$51,$FD,$D3,$EA,$FF,$AD,$C7,$EB
	.dc.b	$17,$85,$93,$58,$6F,$8B,$87,$81,$2F,$84,$73,$F7,$7A,$7B,$D2,$7A
	.dc.b	$47,$F8,$0B,$EA,$DE,$1D,$F8,$C9,$5F,$BC,$9C,$BF,$BE,$FA,$B6,$FF
	.dc.b	$E0,$A5,$F2,$74,$C4,$7F,$83,$BB,$FD,$35,$FC,$19,$FF,$01,$BF,$D4
	.dc.b	$9B,$CF,$FD,$31,$DC,$67,$C3,$E5,$F5,$72,$39,$47,$5D,$BB,$78,$72
	.dc.b	$7B,$47,$F8,$7B,$EA,$FF,$B0,$63,$FD,$61,$7C,$06,$4D,$6F,$FE,$24
	.dc.b	$FA,$BF,$E7,$B8,$7F,$39,$AF,$E6,$CF,$F1,$E6,$BF,$DA,$A3,$69,$F8
	.dc.b	$6F,$F2,$45,$EB,$75,$3B,$6F,$DF,$A7,$C1,$C1,$2F,$84,$78,$FD,$4D
	.dc.b	$46,$0E,$8F,$9F,$C5,$F4,$78,$D2,$7C,$11,$EB,$F4,$B5,$18,$3C,$8F
	.dc.b	$F9,$0B,$EA,$9F,$13,$FE,$44,$FA,$B7,$BF,$4D,$46,$FB,$AE,$1F,$D0
	.dc.b	$2F,$8B,$63,$FC,$9B,$F1,$7F,$AC,$F7,$7F,$B1,$4F,$F8,$13,$3F,$F2
	.dc.b	$8F,$A6,$77,$F1,$1C,$3C,$BB,$5F,$D3,$30,$E4,$FF,$6B,$7F,$6A,$FF
	.dc.b	$86,$BE,$AF,$F1,$38,$F6,$C7,$E7,$CF,$F5,$46,$BB,$F5,$FF,$CC,$35
	.dc.b	$F8,$EE,$BF,$F0,$DB,$CA,$7F,$BC,$35,$3B,$BF,$5B,$FB,$68,$F1,$78
	.dc.b	$FD,$0E,$BE,$3F,$6E,$0F,$7F,$34,$AF,$EC,$F7,$F2,$1F,$57,$06,$DC
	.dc.b	$D7,$9A,$3C,$FE,$3F,$F9,$86,$7F,$DF,$14,$8C,$47,$CE,$3C,$EE,$1F
	.dc.b	$3B,$5F,$6F,$A2,$FA,$BF,$F7,$AD,$7F,$30,$6C,$66,$7B,$B5,$FF,$7D
	.dc.b	$FD,$2F,$FB,$F1,$FF,$0F,$35,$34,$D7,$C9,$7F,$9E,$37,$FF,$24,$CF
	.dc.b	$FA,$7F,$7F,$F0,$4B,$C8,$7D,$5F,$FA,$39,$FC,$4D,$71,$9F,$77,$2D
	.dc.b	$9F,$C6,$B5,$FF,$18,$4D,$7F,$CD,$D3,$DB,$97,$06,$F9,$1F,$3D,$89
	.dc.b	$F1,$FC,$DF,$13,$17,$FE,$68,$0F,$5F,$9B,$93,$C0,$F5,$BE,$3D,$3F
	.dc.b	$D2,$B1,$FE,$26,$6D,$7F,$9C,$B0,$FE,$53,$39,$E5,$E1,$C9,$23,$07
	.dc.b	$17,$DD,$F5,$7F,$26,$FF,$CF,$6D,$C3,$FC,$EB,$87,$FB,$06,$FF,$C6
	.dc.b	$B3,$97,$F6,$6F,$C4,$AE,$3B,$74,$DB,$AB,$DB,$C6,$97,$CB,$C7,$B7
	.dc.b	$E0,$F3,$FC,$66,$33,$FF,$42,$E5,$FD,$32,$FC,$3E,$A7,$A0,$F4,$FC
	.dc.b	$75,$8F,$2F,$AD,$EC,$31,$8F,$B1,$F7,$FC,$6C,$9F,$5F,$CC,$DA,$FF
	.dc.b	$4D,$FC,$CF,$DA,$9F,$F2,$9B,$63,$3E,$F7,$E1,$63,$1A,$7F,$A6,$97
	.dc.b	$91,$93,$5E,$2D,$C3,$FC,$BB,$F5,$6F,$9D,$FC,$73,$63,$3D,$F8,$9C
	.dc.b	$F3,$C9,$E3,$EF,$DD,$7F,$D6,$3F,$57,$F8,$A9,$FF,$34,$B6,$33,$DF
	.dc.b	$F9,$B8,$C6,$DF,$EB,$05,$C1,$93,$5C,$57,$0F,$F3,$9F,$D5,$FF,$2D
	.dc.b	$6E,$FF,$A5,$8F,$C2,$65,$FE,$C2,$74,$FF,$B1,$27,$FC,$62,$7F,$9F
	.dc.b	$0A,$E7,$C5,$FD,$74,$63,$2F,$F6,$1F,$D5,$FF,$A7,$E3,$FD,$06,$DA
	.dc.b	$E9,$EE,$FE,$D6,$31,$5C,$35,$C1,$9E,$DB,$87,$2D,$38,$1C,$7A,$6C
	.dc.b	$7F,$A4,$3C,$5F,$F2,$7B,$7D,$BF,$E0,$50,$F9,$EB,$18,$F2,$E5,$71
	.dc.b	$F5,$2E,$FF,$EE,$4F,$AB,$FD,$AE,$DF,$ED,$E6,$D7,$F5,$A7,$FC,$66
	.dc.b	$FF,$A8,$1C,$CF,$FA,$FD,$FF,$58,$C6,$7E,$1F,$F8,$A4,$3D,$DB,$18
	.dc.b	$E2,$95,$CF,$DC,$B8,$7F,$BD,$BC,$7F,$D5,$37,$C5,$FE,$3D,$0F,$1E
	.dc.b	$C7,$F2,$1F,$AB,$FF,$69,$E1,$FE,$B4,$6D,$FF,$C0,$05,$1B,$F8,$4F
	.dc.b	$E5,$1F,$56,$F8,$7A,$7F,$2B,$FA,$BD,$D7,$FA,$FA,$74,$DF,$07,$89
	.dc.b	$C1,$C1,$BF,$0B,$A9,$A7,$07,$3E,$1F,$09,$DF,$E1,$EC,$77,$E2,$F2
	.dc.b	$98,$D5,$C9,$EA,$7F,$DD,$D9,$7D,$07,$35,$FE,$10,$FA,$BF,$C7,$B6
	.dc.b	$3D,$9D,$3F,$D1,$E7,$31,$FE,$F7,$7F,$93,$37,$8F,$4E,$DE,$1E,$94
	.dc.b	$BF,$F7,$0E,$FF,$92,$AF,$09,$CE,$7B,$7A,$52,$BF,$DC,$F8,$7E,$F1
	.dc.b	$F3,$47,$8F,$DA,$91,$A8,$B7,$D0,$57,$F9,$6A,$79,$FA,$3F,$5F,$E4
	.dc.b	$BD,$8F,$B7,$A2,$4E,$8F,$0B,$F0,$7C,$09,$1E,$AF,$8D,$B8,$F7,$C5
	.dc.b	$E1,$7C,$73,$3C,$5F,$46,$7C,$71,$F2,$FE,$6A,$61,$F1,$3E,$12,$D7
	.dc.b	$3F,$80,$FA,$06,$30,$E5,$8F,$F9,$42,$FE,$2F,$D2,$FD,$53,$6E,$05
	.dc.b	$FC,$3F,$A8,$FC,$ED,$F8,$DB,$3F,$E5,$0F,$AB,$7C,$B7,$F8,$16,$73
	.dc.b	$7C,$97,$5E,$4D,$78,$17,$0E,$F8,$F5,$F5,$F9,$1F,$E0,$FF,$AB,$7D
	.dc.b	$5E,$2B,$67,$F9,$83,$1F,$E9,$ED,$F7,$F5,$9F,$A5,$F2,$72,$49,$FC
	.dc.b	$5F,$B3,$6B,$DD,$C7,$E9,$31,$A7,$BD,$F6,$7C,$54,$BF,$0D,$7B,$67
	.dc.b	$0F,$C7,$F8,$AF,$84,$7A,$7A,$AD,$70,$96,$8F,$AB,$F9,$06,$A3,$E8
	.dc.b	$3F,$2F,$8E,$57,$1F,$8B,$FE,$3C,$F3,$BF,$A2,$F5,$E4,$7B,$DF,$E3
	.dc.b	$76,$C6,$EF,$13,$FE,$3B,$FA,$BF,$E5,$B9,$1B,$F2,$FF,$32,$3F,$07
	.dc.b	$1A,$3F,$D3,$77,$EB,$6C,$7E,$A5,$D7,$C2,$75,$F8,$09,$5F,$06,$FE
	.dc.b	$03,$C6,$3C,$5F,$01,$23,$07,$57,$CD,$E6,$48,$F9,$5F,$51,$EC,$CB
	.dc.b	$C1,$3E,$BF,$36,$57,$96,$2C,$FE,$F7,$DB,$FF,$1A,$73,$FE,$B5,$B5
	.dc.b	$FE,$AA,$C3,$FB,$83,$FE,$AC,$8C,$D7,$E9,$11,$FE,$63,$2C,$7D,$D7
	.dc.b	$71,$A7,$5E,$54,$9F,$4F,$A9,$6B,$C9,$E5,$73,$31,$8F,$95,$ED,$EA
	.dc.b	$4B,$FC,$57,$D4,$F8,$2F,$D9,$3E,$3F,$9F,$95,$E3,$8B,$3F,$C7,$7E
	.dc.b	$AF,$FC,$58,$DE,$0F,$F3,$E4,$7C,$08,$FE,$02,$EB,$2F,$E2,$4F,$08
	.dc.b	$FF,$51,$35,$FF,$68,$B5,$E3,$3F,$E9,$26,$DF,$E8,$67,$FA,$EC,$78
	.dc.b	$63,$1E,$B7,$FA,$37,$26,$54,$C6,$56,$BE,$0C,$75,$B5,$70,$9F,$F5
	.dc.b	$B4,$7F,$1A,$8C,$38,$6B,$FE,$BB,$73,$97,$F2,$FF,$94,$7C,$F7,$7A
	.dc.b	$FE,$89,$18,$BF,$DD,$36,$3B,$B5,$FF,$63,$35,$FF,$7A,$B5,$F2,$79
	.dc.b	$DE,$F3,$E0,$79,$F6,$BE,$EC,$7B,$71,$FE,$8B,$1E,$CA,$F7,$CE,$49
	.dc.b	$1E,$4F,$57,$94,$FA,$1F,$CB,$7C,$4F,$C7,$E4,$34,$95,$D6,$18,$FE
	.dc.b	$74,$E3,$1F,$9D,$70,$ED,$E8,$47,$04,$BF,$A7,$3D,$23,$11,$FE,$EB
	.dc.b	$8F,$6B,$AC,$3C,$C4,$9F,$81,$C0,$F6,$7F,$00,$D8,$CC,$60,$F9,$1E
	.dc.b	$8C,$8F,$91,$A7,$F6,$1C,$3F,$05,$3F,$89,$D7,$FC,$19,$1F,$63,$DF
	.dc.b	$C1,$2F,$E8,$4B,$CE,$E0,$48,$CB,$E4,$F9,$D9,$1E,$86,$DF,$C2,$F8
	.dc.b	$7F,$A4,$49,$FA,$0F,$D5,$F8,$A9,$1F,$4D,$B3,$F2,$BD,$6F,$F1,$A4
	.dc.b	$FC,$96,$C7,$F1,$76,$BF,$C5,$DA,$BE,$AF,$FB,$9C,$FF,$A2,$63,$FB
	.dc.b	$FC,$61,$FF,$08,$1B,$F1,$F8,$C6,$67,$FC,$0B,$3C,$39,$7F,$A7,$BF
	.dc.b	$B0,$9F,$F4,$27,$8D,$F9,$77,$F5,$7E,$E5,$AD,$D7,$23,$C9,$AF,$4C
	.dc.b	$31,$7F,$3E,$7C,$C6,$8A,$E5,$F3,$AF,$2B,$CB,$D5,$FB,$BF,$56,$DF
	.dc.b	$EC,$9B,$7F,$02,$AB,$F1,$7E,$26,$E1,$CE,$3F,$BC,$35,$FF,$23,$28
	.dc.b	$E3,$F5,$6F,$94,$E7,$CE,$F7,$A9,$A8,$DF,$CA,$5C,$7F,$C0,$97,$C5
	.dc.b	$B1,$D3,$E4,$3C,$BD,$1F,$E2,$A9,$FE,$23,$C9,$97,$8F,$9D,$5F,$F2
	.dc.b	$D6,$0B,$F2,$CF,$9B,$B3,$DF,$E5,$7E,$AE,$FF,$B5,$C6,$5F,$1F,$C1
	.dc.b	$D7,$9C,$7B,$A3,$DF,$E5,$72,$D6,$57,$AA,$3E,$1F,$C2,$EA,$30,$7B
	.dc.b	$9F,$9F,$F5,$6F,$8B,$F1,$E1,$2B,$FC,$4D,$39,$6F,$8B,$C2,$A4,$60
	.dc.b	$FF,$16,$7D,$53,$EA,$FE,$DF,$81,$FE,$B9,$9F,$F1,$83,$5F,$F3,$F4
	.dc.b	$ED,$E3,$EC,$FE,$0A,$31,$1A,$AC,$4E,$7C,$AF,$C3,$FD,$5B,$4A,$66
	.dc.b	$FC,$7E,$5D,$7B,$D7,$58,$ED,$E2,$FE,$43,$EA,$DE,$95,$33,$7E,$CD
	.dc.b	$6B,$E8,$AF,$34,$79,$FC,$BB,$3B,$3E,$1B,$F9,$CF,$A7,$FF,$45,$78
	.dc.b	$BF,$E8,$BF,$A7,$FF,$7C,$57,$F9,$4D,$AF,$FA,$86,$77,$E3,$7E,$F9
	.dc.b	$78,$DC,$E2,$4F,$F9,$6E,$72,$DF,$37,$F5,$7F,$56,$FA,$BA,$78,$E3
	.dc.b	$C9,$E5,$5A,$E0,$73,$7D,$1F,$55,$FB,$4F,$AB,$7B,$74,$CD,$78,$7E
	.dc.b	$1A,$BE,$9A,$F8,$11,$F0,$7E,$12,$CE,$CF,$43,$F1,$7E,$53,$E9,$7B
	.dc.b	$32,$7D,$91,$F2,$7E,$17,$51,$83,$E0,$3F,$37,$E1,$FF,$84,$F1,$BE
	.dc.b	$3E,$99,$AF,$8D,$F0,$B5,$F9,$D5,$F5,$63,$F8,$8F,$BA,$FA,$BF,$C6
	.dc.b	$F9,$D2,$FF,$D1,$13,$96,$F9,$7C,$2A,$46,$0D,$FA,$5D,$26,$46,$0E
	.dc.b	$7F,$41,$F9,$AF,$76,$46,$0E,$FF,$33,$E4,$99,$18,$3C,$7E,$87,$EB
	.dc.b	$BE,$06,$46,$0F,$3F,$91,$EC,$99,$18,$3D,$7F,$0B,$FC,$C3,$E1,$DD
	.dc.b	$14,$F8,$71,$F1,$3A,$AD,$70,$3E,$1B,$F2,$3D,$8D,$46,$0F,$8E,$FF
	.dc.b	$3C,$FA,$BF,$E8,$DE,$2F,$FB,$AD,$78,$19,$35,$E0,$6F,$DC,$E1,$EF
	.dc.b	$4C,$7F,$D7,$17,$FC,$FD,$FF,$5D,$7D,$5B,$E2,$7F,$A4,$AE,$4A,$62
	.dc.b	$33,$C3,$AF,$87,$97,$CA,$3F,$A8,$FD,$5F,$FA,$32,$3F,$D8,$CF,$F2
	.dc.b	$15,$B3,$CA,$FE,$E3,$39,$8C,$4F,$FB,$E2,$39,$3C,$27,$7F,$11,$DB
	.dc.b	$D2,$93,$DA,$3F,$B0,$F7,$FF,$4D,$8E,$00,$FE,$95,$E6,$E8,$DF,$C0
	.dc.b	$70,$F1,$F8,$FC,$D4,$AF,$F6,$F5,$A7,$FB,$6F,$D5,$FF,$53,$9F,$F7
	.dc.b	$12,$F0,$B2,$6A,$B7,$C5,$C3,$C0,$97,$C2,$39,$FB,$DD,$3D,$F9,$3D
	.dc.b	$23,$FB,$DF,$D5,$BC,$3B,$F3,$12,$BF,$DE,$53,$97,$FD,$E9,$F5,$6D
	.dc.b	$FF,$C0,$6B,$E4,$E9,$88,$FF,$03,$F8,$1E,$9A,$FF,$7E,$3F,$DE,$5F
	.dc.b	$92,$7E,$3B,$C8,$DF,$07,$0F,$13,$8F,$D5,$CA,$E3,$1D,$37,$EB,$E2
	.dc.b	$49,$EB,$1F,$E1,$6F,$AB,$FE,$B5,$DB,$F8,$09,$43,$E0,$34,$6F,$87
	.dc.b	$FC,$3B,$F5,$7F,$CF,$2B,$EE,$57,$F3,$57,$EF,$5F,$38,$C4,$7A,$A7
	.dc.b	$82,$7E,$23,$C3,$BB,$FC,$87,$0F,$57,$B9,$E0,$E2,$F0,$F5,$F8,$B8
	.dc.b	$E6,$1C,$4F,$C7,$5F,$E1,$DC,$1D,$FD,$4E,$2E,$23,$4F,$96,$3D,$BE
	.dc.b	$47,$BB,$C8,$93,$B8,$F8,$3E,$96,$A3,$07,$37,$FC,$83,$F5,$4F,$89
	.dc.b	$FF,$21,$FD,$5B,$DE,$A6,$A3,$7E,$0D,$C3,$CF,$7C,$5B,$1F,$E4,$CF
	.dc.b	$8C,$FD,$B7,$17,$F9,$A3,$D3,$FE,$15,$5F,$86,$6F,$2F,$DB,$AD,$CF
	.dc.b	$EC,$5F,$DD,$C6,$7E,$23,$A1,$D7,$E2,$7F,$CB,$3F,$57,$F9,$19,$FC
	.dc.b	$26,$1F,$4F,$0F,$96,$D7,$F2,$B3,$BC,$7F,$CF,$6F,$E6,$E3,$3E,$07
	.dc.b	$43,$AF,$81,$DB,$C7,$B3,$F5,$AB,$86,$3F,$93,$7F,$7D,$1D,$B7,$FF
	.dc.b	$36,$E5,$F6,$2B,$CB,$D1,$E8,$9F,$4E,$F6,$FA,$46,$1E,$DE,$3F,$F2
	.dc.b	$27,$FD,$FB,$8F,$F2,$FC,$3F,$DC,$8D,$8C,$CE,$DF,$F4,$74,$7F,$37
	.dc.b	$FA,$BD,$FA,$26,$92,$7F,$DD,$C3,$E9,$6B,$4C,$B8,$37,$F2,$7F,$CC
	.dc.b	$B1,$FF,$54,$B7,$87,$FD,$3D,$5C,$7E,$AF,$FD,$08,$FF,$AA,$A7,$4F
	.dc.b	$FA,$B3,$92,$BF,$CC,$5A,$A2,$C6,$39,$25,$FD,$EF,$9D,$B7,$FC,$48
	.dc.b	$7C,$D7,$23,$39,$7F,$AF,$9E,$BF,$17,$2F,$F6,$35,$C1,$B1,$E4,$AF
	.dc.b	$F4,$64,$7E,$D9,$B3,$F9,$BF,$0B,$5C,$DB,$E3,$97,$B8,$B5,$FC,$9B
	.dc.b	$EA,$FD,$93,$FC,$01,$B1,$FA,$5F,$03,$E1,$C6,$31,$FE,$40,$FF,$A2
	.dc.b	$6B,$86,$DA,$DF,$6F,$0F,$1A,$4F,$5F,$1E,$E5,$E1,$F3,$FA,$98,$CB
	.dc.b	$AB,$E0,$F5,$65,$F9,$7E,$3A,$DF,$A7,$D6,$F2,$98,$C3,$CA,$FA,$3E
	.dc.b	$36,$5F,$B7,$E6,$6D,$FF,$46,$F9,$97,$3F,$E0,$1B,$5F,$C8,$97,$BF
	.dc.b	$99,$D7,$93,$2B,$5E,$55,$C3,$FA,$8F,$D5,$BE,$73,$F1,$6D,$AF,$E5
	.dc.b	$AF,$F8,$65,$B1,$AF,$1A,$46,$BD,$EB,$7F,$D7,$3E,$AF,$F0,$8D,$EF
	.dc.b	$FC,$BC,$3C,$0B,$18,$E0,$98,$7F,$3C,$A7,$16,$30,$FE,$CF,$F5,$7F
	.dc.b	$EE,$BA,$FE,$82,$D8,$CF,$6F,$E9,$23,$1C,$BF,$D8,$6B,$D5,$5F,$DC
	.dc.b	$1B,$19,$F1,$FF,$57,$18,$3F,$C3,$92,$FA,$BF,$F0,$F4,$7F,$50,$6C
	.dc.b	$73,$F7,$7F,$67,$18,$BF,$EE,$0F,$BE,$B8,$6D,$AE,$ED,$BC,$34,$93
	.dc.b	$D7,$AE,$E5,$FE,$06,$F1,$BF,$CA,$73,$FD,$81,$B1,$9F,$9E,$FE,$07
	.dc.b	$18,$CF,$FC,$00,$FF,$8B,$6B,$87,$99,$AF,$57,$6F,$F0,$AF,$AB,$FD
	.dc.b	$AD,$FF,$81,$DB,$5F,$09,$FF,$19,$3F,$DC,$1C,$E3,$FE,$41,$7F,$D6
	.dc.b	$31,$9F,$87,$E9,$EE,$75,$7B,$70,$C9,$E5,$ED,$DC,$3F,$C4,$7E,$3F
	.dc.b	$F8,$CE,$7F,$BE,$B6,$33,$E3,$FF,$20,$8C,$67,$FE,$1D,$FA,$BF,$F6
	.dc.b	$8D,$FF,$81,$1B,$7F,$E2,$E7,$FC,$C9,$19,$F8,$4F,$E5,$3F,$56,$F8
	.dc.b	$7A,$7F,$2C,$FA,$BD,$E1,$FE,$36,$9D,$5F,$F8,$41,$78,$1C,$8E,$31
	.dc.b	$97,$AF,$83,$23,$07,$17,$C3,$C3,$23,$FC,$37,$83,$E1,$F1,$8F,$87
	.dc.b	$8B,$CE,$63,$6F,$F1,$E6,$7F,$42,$7F,$C9,$7F,$57,$F9,$1C,$7F,$32
	.dc.b	$D8,$FF,$3B,$62,$E8,$FE,$93,$39,$93,$FC,$A6,$B8,$F0,$EA,$76,$F0
	.dc.b	$F0,$E4,$FA,$3E,$16,$D7,$8B,$D4,$EA,$63,$4F,$F2,$96,$28,$F0,$FB
	.dc.b	$93,$0F,$F1,$DC,$E5,$B1,$E6,$F4,$64,$7C,$6F,$92,$B8,$FC,$0E,$8F
	.dc.b	$59,$8C,$BB,$BE,$EE,$F9,$7F,$3B,$E3,$6D,$FC,$8F,$07,$E1,$98,$C3
	.dc.b	$CE,$FC,$6F,$9A,$97,$F5,$BE,$16,$DF,$D0,$F8,$2F,$96,$63,$0F,$7B
	.dc.b	$F3,$BE,$AE,$5F,$BA,$2D,$7F,$A0,$BE,$9F,$F2,$F7,$FE,$77,$6D,$7F
	.dc.b	$12,$FA,$B7,$CB,$7F,$80,$27,$37,$FE,$B1,$7F,$11,$5F,$E6,$46,$C7
	.dc.b	$5F,$07,$87,$6F,$1F,$5A,$5F,$F9,$AB,$EA,$DF,$57,$8C,$3C,$9E,$C7
	.dc.b	$73,$1A,$F9,$5F,$17,$1C,$C3,$F3,$3E,$C5,$BF,$67,$17,$9C,$C6,$1E
	.dc.b	$D7,$D5,$F0,$12,$FD,$11,$6B,$E2,$7C,$57,$BC,$FF,$0F,$E9,$FE,$99
	.dc.b	$AF,$74,$7C,$3F,$AB,$91,$83,$E1,$3F,$1F,$8A,$47,$FA,$23,$D6,$78
	.dc.b	$F8,$C7,$FD,$67,$C4,$F7,$3F,$C5,$AD,$BF,$99,$86,$27,$FC,$63,$F5
	.dc.b	$7F,$C8,$AC,$DF,$97,$FE,$EB,$1F,$EB,$E8,$D1,$FE,$89,$BF,$F4,$CB
	.dc.b	$63,$5F,$07,$D3,$E1,$3A,$FC,$3C,$BF,$F4,$FE,$0D,$B1,$87,$5F,$7F
	.dc.b	$C6,$63,$93,$FD,$81,$9F,$F7,$B9,$F1,$6F,$E8,$3F,$C9,$3F,$4F,$ED
	.dc.b	$8C,$E5,$E4,$F3,$FD,$A6,$DC,$1F,$EB,$0F,$6B,$FC,$59,$3F,$ED,$A6
	.dc.b	$E3,$FE,$CF,$BF,$ED,$52,$3F,$C0,$1E,$DF,$F9,$62,$BF,$B2,$47,$17
	.dc.b	$11,$97,$5E,$6C,$C3,$FD,$7F,$6F,$5B,$18,$78,$7C,$DE,$86,$32,$FF
	.dc.b	$70,$63,$FC,$2E,$7C,$DF,$0B,$E1,$3F,$CD,$3F,$53,$E2,$AD,$C7,$97
	.dc.b	$E7,$A6,$7E,$F8,$B7,$FE,$F8,$FA,$BF,$99,$7F,$EE,$D6,$D7,$F8,$21
	.dc.b	$F8,$32,$72,$FC,$36,$B9,$F6,$7F,$CF,$71,$83,$FA,$7C,$77,$8F,$F7
	.dc.b	$D3,$D5,$AF,$36,$BC,$E7,$FD,$16,$B1,$63,$A6,$CE,$9B,$3F,$E9,$26
	.dc.b	$BF,$E2,$23,$94,$FF,$A4,$3E,$51,$F3,$DF,$2A,$30,$CB,$F0,$CE,$9B
	.dc.b	$1E,$68,$C5,$7F,$A8,$1F,$D3,$E4,$AF,$F1,$8C,$61,$B5,$E1,$F1,$7E
	.dc.b	$CF,$9D,$EB,$24,$FA,$DB,$83,$0E,$8F,$E4,$F0,$FE,$C5,$5F,$95,$9C
	.dc.b	$CF,$C7,$AF,$77,$AF,$2B,$D6,$FA,$48,$FC,$ED,$DE,$87,$E7,$4F,$CF
	.dc.b	$7C,$55,$9F,$D4,$47,$D6,$9D,$C7,$18,$C1,$9B,$91,$D5,$B8,$7F,$B1
	.dc.b	$AB,$FC,$CB,$39,$24,$66,$7F,$07,$C1,$FE,$16,$9F,$C1,$B6,$7F,$D9
	.dc.b	$ED,$7E,$F3,$5F,$A3,$E9,$FD,$72,$4F,$DC,$7E,$D7,$AF,$23,$EF,$B6
	.dc.b	$7E,$E7,$9B,$F7,$C9,$3F,$84,$7F,$05,$F4,$12,$3F,$0E,$D9,$F7,$7C
	.dc.b	$47,$C7,$24,$FC,$F7,$E5,$7C,$5C,$8F,$A4,$D9,$F8,$DE,$B7,$E0,$27
	.dc.b	$F3,$5A,$A3,$13,$FE,$7E,$8C,$4F,$F9,$F2,$3F,$8C,$46,$1F,$D8,$AF
	.dc.b	$57,$8C,$69,$AF,$FA,$75,$B1,$B9,$E9,$2F,$AB,$FF,$19,$79,$3F,$B6
	.dc.b	$6C,$7E,$BC,$D7,$3F,$FB,$ED,$C4,$69,$FD,$A2,$DB,$FD,$2F,$78,$6C
	.dc.b	$FF,$A6,$A7,$4F,$89,$AF,$F0,$CE,$67,$A5,$C2,$3F,$4C,$D5,$1A,$7D
	.dc.b	$6D,$7F,$8A,$CB,$CD,$FF,$54,$F4,$7E,$B2,$9E,$A6,$32,$FD,$EB,$70
	.dc.b	$FE,$8D,$87,$17,$F6,$93,$E9,$F8,$CD,$47,$C9,$C6,$D6,$E3,$F7,$D5
	.dc.b	$B7,$89,$FD,$47,$EA,$DF,$29,$CB,$BD,$EE,$D3,$AC,$EA,$B8,$85,$F2
	.dc.b	$9A,$B9,$3A,$9C,$B7,$D1,$F3,$ED,$FE,$7B,$8E,$6D,$7C,$E7,$96,$3C
	.dc.b	$3E,$8F,$89,$91,$D6,$3C,$BF,$09,$F0,$BC,$9F,$F4,$8C,$FF,$64,$8C
	.dc.b	$BE,$3F,$84,$AF,$63,$D5,$1E,$E8,$F7,$F9,$5C,$F3,$95,$EB,$8F,$87
	.dc.b	$F0,$DA,$8C,$1E,$F7,$E9,$FD,$5B,$CE,$F2,$61,$2B,$E4,$CE,$5B,$D9
	.dc.b	$C2,$A4,$60,$F9,$FF,$56,$FA,$BF,$BF,$E0,$3E,$0E,$9F,$82,$9D,$57
	.dc.b	$0F,$84,$C5,$C8,$C1,$D3,$CB,$FC,$4F,$D5,$B5,$A7,$38,$E9,$E6,$5A
	.dc.b	$E0,$79,$1F,$07,$8D,$F9,$2F,$AB,$FF,$4D,$4F,$E1,$9A,$FC,$FC,$D3
	.dc.b	$E7,$D7,$8C,$FB,$A3,$0F,$C0,$F1,$EB,$C6,$7D,$D1,$FA,$0F,$AB,$FE
	.dc.b	$AB,$E2,$7F,$56,$FA,$8F,$F5,$D5,$FE,$59,$AF,$DF,$9D,$F8,$3F,$E3
	.dc.b	$18,$47,$EC,$9F,$CE,$BA,$7F,$C6,$CD,$77,$E2,$F8,$7C,$7F,$D8,$FD
	.dc.b	$5B,$EA,$E9,$9A,$E7,$E7,$57,$C5,$5E,$98,$F5,$79,$DF,$BA,$FA,$B7
	.dc.b	$B5,$4C,$DF,$93,$CD,$AF,$AE,$BE,$14,$7C,$3F,$85,$B3,$B3,$D6,$FC
	.dc.b	$7F,$92,$FA,$DE,$C4,$9F,$7C,$7C,$BF,$85,$D4,$60,$F8,$6F,$CF,$F8
	.dc.b	$8F,$E1,$FC,$6F,$5A,$99,$AF,$93,$F0,$B5,$FA,$15,$F5,$E3,$F8,$BF
	.dc.b	$AA,$FA,$BF,$C8,$F9,$D2,$FF,$86,$CE,$5B,$E5,$F0,$A9,$18,$37,$E8
	.dc.b	$F4,$99,$18,$39,$FC,$FF,$E7,$BE,$AE,$46,$0E,$FF,$33,$E4,$99,$18
	.dc.b	$3C,$7E,$87,$EC,$3E,$06,$46,$0F,$3F,$91,$EC,$99,$18,$3D,$7F,$09
	.dc.b	$FC,$DB,$E2,$3F,$A2,$78,$7F,$E2,$C6,$F8,$7F,$93,$E0,$FF,$58,$DF
	.dc.b	$C2,$EC,$F8,$F8,$48,$C1,$F1,$BD,$6F,$97,$32,$30,$7F,$36,$FA,$B7
	.dc.b	$81,$FD,$19,$FE,$BB,$9F,$F8,$F1,$B9,$7E,$8D,$DD,$71,$F7,$78,$7B
	.dc.b	$B2,$30,$7F,$42,$FA,$B7,$C5,$FF,$52,$5C,$94,$C4,$67,$87,$5F,$0D
	.dc.b	$2F,$94,$7F,$5A,$FA,$B7,$B1,$7D,$89,$5F,$D4,$27,$2D,$F8,$5E,$1E
	.dc.b	$A4,$8C,$1F,$D5,$BB,$FF,$A6,$71,$7F,$96,$9F,$DF,$47,$F4,$16,$C6
	.dc.b	$7E,$37,$FB,$5C,$3C,$6B,$F3,$52,$FA,$47,$F7,$0F,$AB,$55,$F0,$A5
	.dc.b	$7F,$66,$9C,$B7,$8B,$87,$81,$23,$06,$FD,$EE,$9E,$FC,$8C,$1F,$DC
	.dc.b	$3E,$AD,$E1,$FF,$80,$1F,$E8,$51,$9F,$7F,$13,$FE,$04,$FA,$B6,$F7
	.dc.b	$E4,$E5,$FF,$7A,$9C,$BF,$DF,$3D,$4F,$EE,$EF,$F6,$57,$FC,$00,$DF
	.dc.b	$27,$FC,$1F,$1A,$E4,$FF,$42,$BF,$98,$9E,$1F,$57,$A3,$7E,$37,$0D
	.dc.b	$B8,$F8,$D2,$BC,$11,$FE,$1A,$FA,$BF,$5C,$7F,$75,$5F,$07,$93,$5C
	.dc.b	$1F,$E2,$0F,$AB,$FE,$69,$3F,$CC,$6B,$FC,$2A,$D7,$FB,$F4,$7F,$39
	.dc.b	$1B,$F1,$F5,$64,$FF,$87,$27,$2D,$EA,$ED,$C5,$23,$06,$FD,$4F,$14
	.dc.b	$C8,$C1,$CF,$C7,$F3,$78,$F2,$30,$77,$F4,$BD,$53,$23,$07,$F8,$BB
	.dc.b	$EA,$9F,$11,$FE,$3E,$FA,$BF,$F3,$74,$FF,$8C,$9A,$FF,$86,$E7,$6F
	.dc.b	$1E,$DF,$F2,$2C,$62,$35,$58,$9F,$F2,$4F,$C5,$FF,$84,$E7,$FC,$49
	.dc.b	$E9,$B1,$FF,$26,$B6,$3F,$C8,$33,$BE,$AF,$F6,$2B,$FB,$48,$FF,$16
	.dc.b	$39,$F3,$FF,$B7,$36,$07,$F9,$3B,$CA,$FF,$2C,$FD,$5F,$E3,$AF,$FD
	.dc.b	$9F,$83,$93,$C2,$6B,$BB,$7E,$2D,$C7,$FC,$A7,$87,$1C,$FF,$BE,$3F
	.dc.b	$95,$6F,$8D,$C2,$77,$5E,$08,$F0,$EF,$EE,$E4,$95,$D2,$3C,$BC,$1A
	.dc.b	$8C,$1E,$07,$D3,$E2,$FF,$BA,$E3,$FD,$2D,$97,$F2,$2D,$BF,$D3,$10
	.dc.b	$8C,$3A,$9E,$0F,$F8,$4C,$D7,$9C,$FA,$BE,$35,$FC,$B9,$B1,$99,$EE
	.dc.b	$D7,$BF,$D2,$F8,$1F,$9E,$D4,$D3,$5F,$42,$DF,$FC,$FB,$28,$C3,$FE
	.dc.b	$67,$39,$8F,$AB,$FF,$43,$BF,$E2,$B8,$DE,$7C,$23,$EF,$FA,$18,$BF
	.dc.b	$73,$63,$84,$78,$23,$11,$FE,$CE,$8F,$1F,$2E,$39,$5F,$2F,$AA,$FE
	.dc.b	$8F,$5E,$8F,$99,$C1,$9B,$B1,$EC,$F9,$9E,$06,$2B,$D8,$F8,$32,$FE
	.dc.b	$77,$3F,$E9,$E6,$CF,$E9,$3C,$2D,$73,$6F,$96,$5E,$E2,$D7,$FA,$37
	.dc.b	$EA,$FE,$54,$FF,$AA,$1B,$1F,$AE,$F0,$3D,$F1,$8C,$7E,$43,$FE,$89
	.dc.b	$AE,$1B,$EB,$83,$6F,$0F,$16,$4F,$5F,$1A,$E5,$E1,$F3,$7A,$98,$CB
	.dc.b	$AB,$E0,$F4,$A5,$F9,$7E,$3E,$DF,$A7,$D4,$F2,$98,$C3,$CA,$FA,$3E
	.dc.b	$36,$5F,$B7,$E6,$2D,$FF,$AB,$7E,$63,$EB,$CF,$FB,$21,$B1,$9F,$77
	.dc.b	$EF,$C6,$31,$FC,$42,$F2,$B2,$6B,$C8,$B8,$7F,$B4,$BE,$AD,$F3,$BF
	.dc.b	$88,$6C,$67,$C1,$C4,$E7,$BE,$4F,$1F,$72,$EB,$FE,$C3,$FA,$B7,$81
	.dc.b	$F9,$26,$D7,$FB,$81,$FC,$7C,$EE,$75,$C1,$2B,$5C,$56,$BF,$D9,$DF
	.dc.b	$57,$FC,$AE,$7F,$DD,$2D,$8C,$F6,$7E,$82,$31,$C5,$FA,$9A,$FF,$16
	.dc.b	$9F,$EF,$02,$D7,$FB,$65,$FF,$17,$99,$DB,$F5,$3F,$57,$FC,$BA,$3F
	.dc.b	$DE,$ED,$9E,$1E,$E6,$9D,$F8,$72,$B7,$DB,$67,$FD,$CD,$6D,$3A,$9D
	.dc.b	$BA,$AD,$FF,$3C,$CF,$F8,$07,$8C,$F6,$BF,$76,$D9,$D1,$FE,$B1,$5F
	.dc.b	$3F,$A9,$DF,$9B,$2B,$79,$FF,$1A,$F5,$70,$FF,$BE,$7E,$AF,$F6,$B9
	.dc.b	$7F,$0F,$6D,$7F,$AF,$57,$A1,$FC,$31,$CD,$7F,$52,$7F,$D5,$F3,$DA
	.dc.b	$11,$87,$B7,$B7,$C7,$E1,$79,$71,$CB,$E1,$6F,$6F,$18,$32,$FE,$45
	.dc.b	$E4,$7F,$8E,$2F,$B3,$67,$51,$FC,$A7,$C8,$FE,$43,$39,$AF,$E5,$5F
	.dc.b	$56,$F9,$4F,$E4,$AD,$BF,$43,$FC,$E5,$B1,$FC,$AB,$EA,$DF,$0D,$4F
	.dc.b	$E5,$BF,$57,$BC,$FC,$27,$F7,$49,$DD,$FC,$35,$E1,$73,$39,$46,$5E
	.dc.b	$DE,$1C,$8C,$1C,$5F,$17,$14,$8F,$9D,$83,$E1,$B1,$8F,$88,$63,$F4
	.dc.b	$75,$FE,$94,$DB,$FF,$7C,$2C,$3D,$DC,$7F,$D2,$3E,$AF,$F3,$D7,$FD
	.dc.b	$45,$B7,$FE,$FF,$CB,$F1,$EF,$DF,$8C,$E3,$FE,$14,$7F,$99,$47,$F8
	.dc.b	$5A,$11,$87,$47,$7E,$0F,$07,$89,$2B,$DB,$D3,$6B,$D9,$6F,$53,$CC
	.dc.b	$67,$2F,$33,$E9,$F7,$66,$1E,$58,$B7,$CB,$D1,$F7,$1F,$6B,$E4,$2D
	.dc.b	$7A,$F9,$FE,$11,$8A,$F0,$BF,$17,$C0,$93,$F4,$BE,$4E,$D7,$9B,$C2
	.dc.b	$F9,$46,$34,$F9,$4F,$CD,$F9,$A9,$7F,$6B,$E0,$2D,$FB,$BE,$6B,$E9
	.dc.b	$18,$C3,$E9,$3F,$57,$EA,$E5,$FD,$18,$B5,$F8,$1F,$AA,$FC,$E3,$7B
	.dc.b	$BF,$C0,$7F,$56,$F9,$6F,$F0,$2D,$6B,$2F,$F0,$73,$F8,$BC,$7F,$74
	.dc.b	$DC,$39,$46,$3B,$78,$FA,$F2,$3F,$80,$7D,$53,$EA,$F1,$7E,$3F,$81
	.dc.b	$EE,$63,$2E,$EF,$8B,$8E,$61,$F9,$CF,$66,$DF,$AF,$8B,$CE,$63,$0F
	.dc.b	$3B,$EA,$F8,$19,$7E,$88,$B5,$F0,$FE,$2B,$DE,$7F,$89,$34,$FF,$51
	.dc.b	$D7,$BA,$3E,$1F,$D5,$C8,$C1,$F0,$9F,$8F,$C3,$23,$F8,$F7,$AF,$FA
	.dc.b	$97,$E3,$C7,$F8,$C3,$85,$EE,$D3,$86,$5F,$B5,$8C,$D7,$F2,$7F,$AB
	.dc.b	$FE,$57,$26,$FC,$C7,$F7,$B9,$FD,$54,$68,$FF,$4D,$BC,$7E,$17,$FC
	.dc.b	$7F,$19,$6F,$B5,$27,$F9,$A5,$34,$C6,$1F,$C1,$F0,$FE,$FE,$DF,$7B
	.dc.b	$3C,$5F,$E5,$19,$CB,$7C,$BE,$E7,$F9,$3B,$E9,$FD,$77,$CE,$3C,$7E
	.dc.b	$64,$CF,$DD,$16,$7F,$8B,$7B,$9F,$E3,$56,$FB,$9C,$F6,$FF,$2C,$CE
	.dc.b	$72,$FF,$31,$3F,$E0,$86,$BA,$9D,$BC,$3C,$C9,$8F,$F5,$3C,$FF,$A6
	.dc.b	$3F,$D5,$67,$FC,$27,$8F,$8B,$C9,$EE,$70,$71,$3B,$F8,$4F,$19,$FE
	.dc.b	$6C,$FA,$9F,$14,$7F,$1D,$BF,$E9,$CD,$F9,$EF,$49,$C1,$F0,$3C,$7C
	.dc.b	$AE,$57,$67,$FB,$45,$F2,$C3,$FD,$7A,$AF,$C0,$70,$9D,$B8,$7F,$D4
	.dc.b	$B1,$B6,$C7,$FA,$26,$7F,$99,$4E,$5F,$F5,$30,$C1,$F0,$DB,$8F,$F6
	.dc.b	$6C,$62,$FF,$D6,$2D,$8E,$2D,$7F,$D6,$0B,$CB,$FB,$27,$DA,$AB,$5F
	.dc.b	$46,$3B,$57,$85,$B5,$C6,$7B,$C6,$3C,$B9,$1E,$26,$D7,$AB,$07,$97
	.dc.b	$D1,$B5,$F1,$23,$06,$7C,$A9,$1F,$21,$B5,$F1,$3E,$4F,$FC,$5C,$DF
	.dc.b	$47,$CF,$18,$8F,$F6,$73,$5F,$9B,$F4,$3E,$12,$5F,$D4,$7D,$FF,$56
	.dc.b	$F9,$EA,$49,$FF,$6E,$C6,$5F,$9F,$F0,$DF,$10,$8E,$09,$7F,$4A,$32
	.dc.b	$D8,$FA,$FF,$15,$F3,$08,$E0,$FB,$7A,$7F,$B5,$1F,$B2,$D9,$FC,$E3
	.dc.b	$5F,$C1,$B5,$64,$F2,$30,$DA,$FF,$33,$48,$DC,$E8,$E5,$C2,$F4,$A9
	.dc.b	$2B,$8D,$7E,$56,$37,$AC,$9C,$D2,$31,$5F,$9F,$C3,$FC,$69,$5F,$58
	.dc.b	$F2,$90,$EB,$F5,$7F,$D9,$9F,$F0,$B4,$A3,$0F,$F8,$65,$AF,$AD,$AE
	.dc.b	$FC,$5F,$89,$1F,$D9,$E7,$9B,$67,$94,$EB,$E1,$7F,$A2,$C7,$EC,$A7
	.dc.b	$FC,$F3,$E3,$FE,$D2,$3C,$5F,$3E,$F6,$B1,$5A,$9E,$1E,$87,$F8,$B7
	.dc.b	$2E,$ED,$8F,$43,$5F,$43,$5F,$76,$F6,$BE,$D4,$7C,$48,$C1,$9F,$85
	.dc.b	$F2,$47,$C5,$6C,$9A,$63,$CD,$FF,$20,$78,$8C,$7F,$C8,$5B,$FF,$9C
	.dc.b	$70,$FF,$43,$35,$FF,$2C,$60,$BE,$35,$F4,$7E,$32,$D7,$D1,$9A,$FD
	.dc.b	$76,$D2,$73,$F5,$6F,$95,$DF,$6B,$DD,$A6,$A3,$7F,$11,$72,$FF,$4D
	.dc.b	$5F,$16,$C7,$4F,$90,$F0,$EB,$FE,$71,$9F,$5F,$36,$5E,$3E,$95,$7F
	.dc.b	$CE,$98,$2F,$CB,$3E,$6F,$84,$F5,$F9,$5F,$E8,$6B,$FE,$B1,$19,$7C
	.dc.b	$7F,$09,$5E,$71,$EE,$8F,$7F,$C3,$F2,$D6,$57,$AA,$3E,$1F,$C4,$EA
	.dc.b	$30,$7B,$9F,$97,$F5,$6F,$43,$C7,$84,$AF,$F5,$E4,$E5,$BE,$07,$0A
	.dc.b	$91,$83,$FD,$85,$F5,$0F,$AB,$FA,$FE,$23,$D0,$A7,$DA,$9D,$57,$0F
	.dc.b	$86,$C5,$C8,$C1,$D3,$CF,$FC,$0F,$D5,$B5,$A7,$38,$E9,$E8,$5A,$E0
	.dc.b	$79,$1F,$07,$95,$F8,$AF,$AB,$FF,$46,$4F,$FB,$65,$AF,$FA,$AF,$34
	.dc.b	$F9,$F5,$E3,$3E,$E8,$C3,$F0,$3D,$0A,$F1,$9F,$74,$7E,$63,$EA,$FF
	.dc.b	$49,$3F,$EE,$7F,$A8,$FF,$7C,$B7,$CC,$D7,$8B,$BF,$89,$B3,$FE,$ED
	.dc.b	$BE,$2F,$FA,$FD,$FD,$57,$8F,$85,$E9,$FE,$9B,$EA,$FF,$93,$57,$FB
	.dc.b	$D5,$AF,$18,$F2,$7C,$47,$95,$91,$DE,$3D,$1F,$55,$FB,$0F,$AB,$7B
	.dc.b	$F4,$CD,$79,$7D,$0A,$F9,$6B,$E0,$47,$C1,$F4,$2C,$EC,$F6,$3F,$17
	.dc.b	$E5,$3E,$A7,$B7,$27,$E0,$47,$C9,$F8,$6D,$46,$0F,$88,$FC,$DF,$89
	.dc.b	$FE,$0B,$C8,$F8,$EA,$66,$BE,$57,$C4,$57,$E5,$57,$D5,$8F,$E1,$BE
	.dc.b	$C3,$EA,$FF,$1F,$E7,$4B,$FE,$B9,$CD,$F4,$AB,$51,$8E,$8F,$F7,$04
	.dc.b	$7F,$09,$8F,$F7,$A3,$A6,$DF,$6F,$2F,$BF,$A5,$67,$91,$FD,$A6,$B9
	.dc.b	$61,$DE,$7C,$BF,$41,$FA,$2F,$80,$97,$C0,$5A,$7D,$7F,$41,$C2,$75
	.dc.b	$39,$6C,$7C,$0F,$43,$F6,$FF,$07,$27,$71,$F1,$3C,$8D,$46,$0E,$6F
	.dc.b	$C8,$F8,$7F,$E6,$BF,$16,$EA,$A6,$6B,$C3,$D9,$5F,$8B,$5F,$42,$3E
	.dc.b	$8F,$B1,$67,$67,$A1,$FE,$83,$F5,$6F,$02,$F8,$32,$BE,$14,$E5,$BE
	.dc.b	$E7,$0F,$7A,$46,$0F,$91,$F5,$6F,$8A,$FE,$9A,$B8,$A9,$88,$CF,$0E
	.dc.b	$BE,$5E,$5F,$28,$FE,$AB,$F5,$7F,$E6,$D8,$FA,$8F,$F2,$95,$D3,$FB
	.dc.b	$A8,$C3,$7D,$8E,$3F,$15,$CB,$D3,$97,$CA,$3F,$B0,$F8,$1F,$D4,$63
	.dc.b	$9F,$81,$FE,$2F,$7F,$A7,$79,$BA,$37,$F0,$1C,$3C,$8E,$3F,$09,$2B
	.dc.b	$F1,$18,$27,$FB,$6F,$D5,$FF,$57,$9F,$C6,$2F,$0F,$26,$AB,$7C,$7C
	.dc.b	$3C,$29,$7C,$23,$9F,$BD,$D3,$DF,$93,$D2,$3F,$BD,$FD,$5B,$C4,$BE
	.dc.b	$F4,$AF,$CE,$4E,$5F,$CF,$7D,$5F,$F8,$45,$BF,$29,$FE,$04,$83,$78
	.dc.b	$3F,$C0,$FE,$07,$A6,$BF,$4C,$FE,$B3,$93,$E4,$36,$3F,$CB,$0E,$DF
	.dc.b	$9A,$63,$C4,$E7,$F5,$72,$39,$C7,$6D,$BB,$F8,$92,$7B,$C7,$F8,$63
	.dc.b	$EA,$FF,$B1,$47,$ED,$D7,$C1,$64,$D6,$FF,$E1,$EF,$AB,$FE,$7F,$DD
	.dc.b	$FD,$02,$BF,$9C,$B7,$E3,$B5,$B6,$F8,$38,$7A,$F3,$3F,$E0,$98,$FC
	.dc.b	$46,$F5,$F5,$E2,$91,$83,$7E,$B7,$86,$64,$60,$E7,$E4,$79,$7C,$89
	.dc.b	$18,$3B,$FA,$9E,$99,$91,$83,$F8,$B7,$D5,$3E,$27,$FC,$77,$F5,$6F
	.dc.b	$80,$A7,$F8,$FA,$75,$5C,$3C,$0C,$5C,$8C,$1F,$C8,$7E,$33,$F5,$7C
	.dc.b	$5F,$A3,$F5,$3F,$C8,$93,$FC,$95,$AF,$DC,$3F,$C8,$9F,$13,$FE,$4C
	.dc.b	$AF,$DC,$46,$5E,$6F,$FA,$6D,$AF,$F9,$2F,$CB,$FF,$29,$FD,$5F,$E2
	.dc.b	$63,$FC,$29,$CB,$FE,$56,$C3,$FC,$DB,$1D,$E7,$47,$E3,$7A,$3F,$BD
	.dc.b	$B8,$F6,$79,$F9,$1D,$3C,$8B,$7F,$CF,$6B,$86,$BF,$BF,$7F,$6D,$1D
	.dc.b	$76,$F0,$73,$48,$E1,$1E,$6E,$0F,$3B,$23,$A4,$7A,$BC,$8F,$E2,$B1
	.dc.b	$FE,$A4,$17,$FC,$8F,$4F,$F4,$B3,$63,$4D,$7F,$D5,$6F,$F2,$BF,$AB
	.dc.b	$DF,$A4,$69,$27,$FD,$1B,$F4,$BF,$E8,$E4,$D2,$3F,$EA,$5F,$92,$FF
	.dc.b	$14,$E1,$FE,$22,$8F,$F5,$36,$BF,$D8,$5F,$F6,$85,$7F,$B4,$7E,$AF
	.dc.b	$FD,$01,$1F,$C9,$6F,$8D,$7F,$B0,$39,$AC,$FF,$29,$6B,$FE,$D9,$9A
	.dc.b	$FF,$99,$67,$B7,$2E,$0D,$F2,$3E,$7B,$13,$E3,$F9,$AE,$26,$2F,$F2
	.dc.b	$01,$EB,$F3,$52,$78,$1E,$B7,$C7,$AF,$FA,$26,$3F,$BF,$36,$BF,$35
	.dc.b	$87,$F2,$09,$CF,$2F,$0E,$49,$18,$38,$BE,$EF,$AB,$F9,$37,$FA,$36
	.dc.b	$E1,$FE,$4E,$C3,$FD,$69,$7F,$E0,$D9,$CB,$FD,$3D,$F8,$95,$C7,$6E
	.dc.b	$9B,$75,$7B,$78,$F2,$F9,$79,$16,$FC,$1E,$87,$8C,$C6,$7F,$E5,$EC
	.dc.b	$BF,$A2,$5F,$87,$D6,$F4,$1E,$9F,$90,$B1,$E5,$F5,$FD,$86,$31,$F6
	.dc.b	$3E,$FF,$8E,$93,$EB,$F9,$AB,$5F,$C0,$FE,$67,$ED,$4F,$F8,$AD,$B1
	.dc.b	$9F,$80,$FC,$2C,$63,$4F,$F3,$9A,$F2,$72,$6B,$C9,$B8,$7F,$8D,$7E
	.dc.b	$AD,$F3,$DF,$8E,$6C,$67,$C0,$C4,$E7,$9E,$4F,$1F,$7E,$EB,$FC,$67
	.dc.b	$EA,$FF,$15,$3F,$E4,$16,$C6,$7B,$FF,$37,$18,$DB,$FD,$12,$B8,$32
	.dc.b	$6B,$8A,$E1,$FE,$4B,$FA,$BF,$F7,$63,$77,$FD,$2C,$7E,$13,$2F,$E5
	.dc.b	$4E,$9F,$E5,$93,$FE,$2B,$3F,$CA,$85,$73,$E2,$FE,$BA,$31,$97,$FA
	.dc.b	$73,$EA,$D5,$FB,$26,$CF,$2F,$7B,$13,$9D,$64,$FF,$98,$6B,$F2,$77
	.dc.b	$CE,$79,$67,$2B,$5D,$17,$0F,$E8,$5E,$2F,$FA,$A2,$7F,$CD,$4D,$8C
	.dc.b	$FC,$A7,$F0,$48,$C6,$3F,$EB,$27,$FC,$51,$5C,$3C,$AD,$7A,$9B,$7F
	.dc.b	$85,$FD,$5F,$ED,$EF,$FD,$72,$DA,$FF,$1C,$3F,$E2,$D7,$FC,$F4,$E7
	.dc.b	$1F,$86,$FF,$AC,$23,$3F,$11,$D3,$DD,$EA,$F6,$E0,$93,$CB,$DC,$B8
	.dc.b	$7F,$5F,$F1,$BF,$56,$DF,$0F,$F8,$FC,$7F,$BD,$6B,$FD,$14,$EB,$3F
	.dc.b	$F6,$7F,$D5,$BE,$17,$F9,$2B,$6B,$FD,$20,$9C,$C7,$F2,$AF,$AB,$7C
	.dc.b	$45,$3F,$96,$FD,$5E,$EB,$FD,$31,$3A,$6F,$77,$13,$83,$83,$7D,$FD
	.dc.b	$4D,$38,$39,$F0,$F8,$4E,$FF,$11,$63,$BF,$17,$94,$C7,$07,$F8,$0F
	.dc.b	$1F,$9C,$FF,$81,$7E,$AF,$F2,$2D,$3D,$3D,$1F,$D1,$A3,$3C,$1F,$EF
	.dc.b	$47,$F9,$2C,$FE,$69,$A9,$DE,$1F,$E7,$77,$27,$ED,$5C,$7A,$3F,$1B
	.dc.b	$E1,$BB,$3D,$FD,$29,$3F,$EB,$FA,$E1,$7D,$E7,$C9,$ED,$CA,$D4,$59
	.dc.b	$F6,$79,$FE,$93,$E9,$FC,$9D,$AE,$9E,$8F,$B4,$C5,$7C,$37,$E0,$78
	.dc.b	$12,$7F,$DA,$1C,$BF,$EB,$F7,$C9,$1F,$17,$C1,$91,$83,$D0,$FC,$9F
	.dc.b	$9A,$91,$CB,$DD,$B8,$FD,$1F,$7F,$E7,$18,$E2,$F9,$CF,$D1,$FA,$49
	.dc.b	$7F,$0E,$2D,$7F,$8F,$7E,$AB,$F2,$CD,$ED,$FE,$FB,$F5,$6F,$97,$FE
	.dc.b	$FF,$5A,$CB,$F3,$57,$FC,$4A,$1D,$7A,$37,$85,$CB,$FD,$E1,$3F,$ED
	.dc.b	$36,$FA,$BF,$E0,$AF,$AA,$7D,$5F,$68,$EF,$EF,$4A,$D9,$C9,$F1,$FC
	.dc.b	$AC,$8F,$9B,$EC,$5D,$3D,$9C,$5E,$83,$1B,$7A,$21,$A2,$7D,$BF,$17
	.dc.b	$30,$F3,$C5,$AF,$8D,$F1,$5F,$08,$F4,$75,$DA,$F6,$FD,$5F,$C6,$31
	.dc.b	$5F,$35,$F9,$3C,$72,$7F,$9C,$E7,$E3,$F8,$B7,$8D,$9C,$3F,$CE,$1C
	.dc.b	$AF,$77,$FC,$5E,$DC,$3E,$5E,$38,$9F,$F1,$9F,$D5,$FF,$23,$B3,$7E
	.dc.b	$63,$FB,$95,$FF,$B8,$63,$47,$FA,$35,$F1,$99,$F8,$4E,$7F,$0B,$D1
	.dc.b	$EB,$F0,$12,$77,$87,$F5,$2C,$1B,$E6,$3F,$D1,$18,$FF,$78,$6F,$C0
	.dc.b	$67,$83,$E9,$4E,$5B,$BF,$A8,$FF,$26,$FD,$3F,$BE,$F9,$C7,$BB,$CE
	.dc.b	$99,$F8,$A7,$DF,$38,$7F,$D3,$1E,$DF,$F5,$E9,$FF,$80,$0D,$8F,$ED
	.dc.b	$37,$FD,$B2,$73,$F0,$DF,$E5,$CA,$14,$C3,$63,$C1,$1F,$C3,$2B,$7E
	.dc.b	$6F,$A3,$D7,$E4,$F3,$F3,$25,$F0,$8B,$5E,$6F,$5F,$C0,$7F,$9B,$3E
	.dc.b	$A7,$C3,$5D,$A3,$C9,$F3,$F2,$30,$78,$9E,$11,$FE,$7F,$AF,$F0,$2D
	.dc.b	$7F,$5A,$7F,$C0,$E6,$23,$4A,$34,$E9,$4E,$9B,$5F,$E9,$F9,$E1,$39
	.dc.b	$24,$7D,$3F,$94,$7C,$E7,$AE,$7F,$D0,$F1,$89,$F9,$3E,$77,$CF,$3E
	.dc.b	$87,$C4,$D8,$C9,$DF,$64,$FA,$1B,$19,$8C,$79,$F2,$38,$37,$2F,$F5
	.dc.b	$B7,$D5,$BD,$AF,$82,$DA,$F6,$36,$77,$3F,$07,$CF,$95,$F0,$76,$FF
	.dc.b	$79,$36,$3E,$5E,$9F,$E2,$3B,$F9,$E1,$C3,$F4,$9A,$FF,$8E,$9A,$EF
	.dc.b	$CB,$E6,$4A,$FA,$13,$FC,$A6,$7D,$7F,$3D,$EA,$24,$FA,$9F,$8D,$F1
	.dc.b	$12,$3E,$53,$67,$FD,$B1,$B3,$E2,$FE,$61,$2B,$E8,$36,$76,$D7,$DB
	.dc.b	$3C,$67,$EB,$7A,$9F,$D1,$1C,$B6,$7F,$62,$D7,$FC,$B3,$28,$C4,$7E
	.dc.b	$C3,$3B,$3F,$3A,$3F,$DB,$4D,$AF,$0C,$EB,$D4,$91,$FE,$12,$AE,$B1
	.dc.b	$87,$8F,$CB,$74,$24,$EE,$3F,$7F,$1E,$2F,$AB,$FF,$18,$47,$FB,$A5
	.dc.b	$B3,$F7,$1B,$1F,$AC,$6B,$FE,$74,$7F,$B5,$C6,$FD,$5E,$84,$7C,$11
	.dc.b	$FD,$A1,$BE,$57,$84,$8E,$BD,$DF,$E8,$57,$F8,$C1,$48,$CC,$79,$7A
	.dc.b	$5F,$05,$FB,$7A,$D7,$97,$7D,$EF,$E0,$78,$FA,$DB,$87,$AD,$AF,$AD
	.dc.b	$AF,$83,$D1,$B7,$FD,$A6,$3D,$71,$83,$3E,$97,$F0,$78,$F3,$B6,$4D
	.dc.b	$35,$D4,$F1,$A9,$2B,$F6,$AF,$1F,$43,$F8,$DF,$91,$FD,$EF,$AB,$FC
	.dc.b	$E1,$5F,$C4,$78,$B3,$AF,$D5,$BB,$3F,$57,$E2,$EB,$F2,$A2,$E8,$6C
	.dc.b	$EB,$EA,$DF,$2B,$C3,$BD,$F0,$14,$D4,$6F,$E5,$6E,$5F,$D5,$6F,$8B
	.dc.b	$63,$A7,$A9,$E2,$DB,$FC,$53,$3F,$CA,$7A,$72,$F1,$F5,$2A,$E1,$FF
	.dc.b	$0D,$62,$C3,$CD,$5E,$7F,$13,$DB,$E7,$7F,$1F,$C3,$FB,$33,$5E,$5C
	.dc.b	$5F,$D1,$30,$C6,$3D,$3E,$13,$95,$7A,$70,$F8,$51,$F0,$FC,$EE,$BB
	.dc.b	$CA,$F7,$47,$C7,$F8,$7D,$46,$0F,$84,$FD,$0F,$AB,$7A,$3E,$6C,$25
	.dc.b	$7F,$7B,$9C,$B7,$C5,$E1,$52,$30,$7F,$80,$7E,$A1,$F5,$7F,$73,$C8
	.dc.b	$7C,$2D,$3E,$FC,$EA,$B8,$7B,$98,$B9,$18,$3A,$7A,$3F,$86,$FA,$B6
	.dc.b	$B4,$E7,$1D,$3D,$2B,$5C,$0F,$23,$E0,$F3,$BF,$1F,$F5,$6F,$5E,$99
	.dc.b	$AF,$56,$D5,$FA,$A7,$FB,$FB,$5F,$E6,$CE,$8F,$81,$E8,$7B,$26,$46
	.dc.b	$0F,$F1,$27,$D5,$FF,$97,$FA,$3F,$CC,$1F,$51,$FC,$1E,$7F,$C5,$0D
	.dc.b	$7B,$89,$E1,$D7,$FE,$08,$7F,$5B,$1A,$72,$8E,$11,$FA,$C8,$DD,$7F
	.dc.b	$9B,$D7,$99,$FB,$0F,$AB,$7A,$54,$E1,$1C,$7D,$6B,$8F,$AA,$30,$FB
	.dc.b	$3E,$AB,$F7,$3F,$57,$FC,$1E,$3F,$C8,$4D,$7C,$D1,$E3,$F8,$7F,$23
	.dc.b	$23,$E2,$C7,$9F,$D2,$D4,$60,$F9,$2F,$AF,$E4,$FE,$CF,$B9,$27,$E6
	.dc.b	$C7,$BF,$E1,$B5,$18,$3E,$8B,$F0,$FE,$23,$F8,$7F,$13,$E3,$A9,$9A
	.dc.b	$FA,$FF,$11,$5F,$9D,$5F,$2E,$3F,$8B,$FB,$6F,$AB,$FC,$A7,$9D,$2F
	.dc.b	$FC,$C5,$39,$6F,$95,$C2,$A4,$60,$DF,$A3,$D2,$64,$60,$E7,$F4,$1F
	.dc.b	$A1,$F8,$19,$18,$3B,$FC,$CF,$92,$64,$60,$F1,$F9,$FF,$B3,$F8,$49
	.dc.b	$18,$3C,$FE,$3F,$B2,$64,$60,$F5,$FC,$37,$F3,$6F,$8A,$74,$53,$E1
	.dc.b	$C7,$C4,$F7,$AD,$70,$3E,$1B,$F2,$3D,$8D,$46,$0F,$8E,$FF,$42,$FA
	.dc.b	$B7,$7D,$EF,$95,$FE,$8D,$9C,$B7,$DC,$E1,$EE,$C8,$C1,$FE,$93,$FA
	.dc.b	$B7,$C6,$FF,$4E,$5C,$74,$C4,$67,$83,$5F,$2D,$2F,$94,$7F,$56,$FA
	.dc.b	$BF,$F4,$9C,$7F,$A7,$DF,$E4,$AF,$EE,$1B,$E4,$FF,$B7,$5D,$C6,$7E
	.dc.b	$2F,$97,$A5,$23,$94,$7F,$62,$EE,$FE,$A3,$1F,$C0,$9F,$DC,$F9,$9F
	.dc.b	$E8,$86,$C7,$EF,$DD,$E1,$FE,$EB,$5E,$2F,$3F,$9A,$95,$FE,$B5,$9C
	.dc.b	$BF,$EB,$7F,$AB,$FE,$B0,$DF,$03,$FB,$9C,$1B,$38,$E0,$CF,$74,$BE
	.dc.b	$51,$C3,$DD,$E3,$EF,$49,$ED,$1F,$DF,$BE,$AD,$E0,$DF,$72,$57,$FB
	.dc.b	$26,$72,$FF,$B2,$FE,$AD,$B7,$F8,$19,$FF,$0B,$C6,$7E,$2B,$13,$FE
	.dc.b	$0A,$ED,$F4,$4F,$FB,$4D,$FE,$FE,$FF,$2C,$6F,$13,$8F,$FD,$67,$1C
	.dc.b	$27,$5E,$0F,$3F,$AB,$91,$CE,$3B,$6F,$DF,$C1,$93,$DE,$3F,$C3,$7F
	.dc.b	$57,$FD,$82,$3F,$DC,$8B,$E0,$32,$6B,$83,$FC,$43,$F5,$7F,$CE,$A9
	.dc.b	$5F,$CD,$1E,$3E,$B6,$B5,$DF,$07,$0F,$52,$57,$FB,$C7,$0F,$8A,$FF
	.dc.b	$4A,$7E,$FB,$78,$7B,$9E,$0E,$BF,$0F,$A7,$67,$94,$61,$E9,$E2,$F9
	.dc.b	$BC,$59,$18,$3C,$1E,$8F,$AA,$64,$60,$FE,$A9,$CD,$F3,$DF,$EA,$DF
	.dc.b	$54,$F8,$8F,$F2,$0F,$D5,$BD,$DA,$7F,$91,$27,$55,$CB,$87,$FD,$D3
	.dc.b	$0F,$03,$58,$4A,$D1,$DA,$3F,$C9,$BF,$17,$FB,$4B,$FF,$78,$FA,$2D
	.dc.b	$3F,$CA,$2D,$BE,$27,$F9,$33,$E1,$B8,$FC,$3F,$26,$46,$B0,$FD,$EE
	.dc.b	$DE,$D8,$7D,$5F,$E3,$B2,$EF,$CD,$D8,$FF,$2F,$C6,$67,$FB,$C3,$F9
	.dc.b	$1A,$E4,$E6,$39,$FA,$DC,$3C,$8E,$3E,$45,$9F,$3D,$71,$C7,$FC,$05
	.dc.b	$1E,$29,$EF,$C1,$EE,$E6,$93,$C2,$3C,$7C,$3A,$8C,$1D,$1F,$3F,$91
	.dc.b	$FE,$5F,$8F,$F5,$06,$3F,$CA,$B4,$FF,$51,$47,$C0,$9D,$3F,$EF,$09
	.dc.b	$D3,$C3,$CF,$7D,$5F,$09,$FE,$6A,$D8,$CC,$EE,$3F,$D3,$0B,$E9,$69
	.dc.b	$1F,$F4,$E3,$FE,$9D,$F9,$4F,$C7,$63,$F4,$E7,$FD,$63,$B7,$DC,$7F
	.dc.b	$C4,$D5,$FE,$27,$FA,$BF,$F4,$4C,$7C,$8B,$E3,$5F,$2A,$3F,$CC,$AF
	.dc.b	$1E,$4C,$C6,$E3,$07,$5E,$5C,$C7,$FB,$B2,$3F,$17,$CC,$F9,$EC,$CA
	.dc.b	$AF,$CD,$4B,$E4,$79,$DD,$FC,$4C,$8C,$1E,$B7,$BE,$BF,$E9,$18,$FC
	.dc.b	$CB,$6F,$FC,$97,$8F,$FA,$FA,$73,$CB,$C3,$92,$46,$0E,$2F,$AB,$EA
	.dc.b	$FE,$4D,$FF,$95,$1B,$87,$E7,$F0,$FF,$5B,$5F,$EA,$27,$2F,$E4,$9F
	.dc.b	$93,$5C,$75,$E9,$B7,$57,$B7,$8F,$2F,$97,$91,$6F,$C1,$E8,$78,$CC
	.dc.b	$67,$FA,$FC,$BF,$A5,$DF,$87,$D4,$F4,$1E,$9F,$8D,$B1,$E5,$F5,$7D
	.dc.b	$86,$31,$F6,$3E,$FF,$8D,$93,$EB,$F9,$BB,$5F,$E7,$4F,$99,$FB,$53
	.dc.b	$FC,$19,$B1,$9F,$7B,$F0,$31,$8D,$3F,$82,$2F,$27,$26,$BC,$6B,$87
	.dc.b	$F1,$0F,$AB,$7C,$F7,$E2,$9B,$19,$F0,$31,$39,$E7,$93,$C7,$DD,$BA
	.dc.b	$FF,$A2,$BE,$AD,$DB,$F9,$46,$D7,$F1,$E7,$F2,$4D,$8D,$70,$48,$D7
	.dc.b	$15,$BF,$F4,$AF,$D5,$FF,$29,$9F,$E4,$ED,$8C,$F9,$FF,$A2,$8C,$71
	.dc.b	$7F,$25,$AF,$F1,$99,$FC,$B8,$B5,$FE,$A0,$7F,$C6,$A6,$76,$FD,$5F
	.dc.b	$D5,$FF,$A7,$63,$F9,$AB,$67,$87,$BD,$A7,$7E,$14,$AD,$F7,$59,$FF
	.dc.b	$56,$DB,$4E,$A7,$6E,$AB,$7F,$D0,$BC,$5F,$F5,$3B,$7D,$CF,$DE,$C3
	.dc.b	$E7,$EC,$63,$CC,$95,$C7,$D5,$B9,$FF,$AF,$3E,$AF,$F6,$9C,$3F,$D1
	.dc.b	$DB,$5F,$CE,$1F,$F1,$83,$FD,$49,$CC,$FF,$AF,$57,$C3,$7F,$0E,$6C
	.dc.b	$EF,$DE,$C4,$E7,$8A,$4F,$3F,$76,$E1,$FE,$CE,$F2,$3F,$52,$DF,$1B
	.dc.b	$F8,$CC,$7F,$BE,$2B,$FB,$1B,$AC,$FF,$AF,$7D,$5B,$E8,$3F,$8F,$B6
	.dc.b	$BF,$B4,$BF,$E2,$D6,$C7,$F2,$4F,$AB,$7C,$3D,$3F,$93,$FD,$56,$F3
	.dc.b	$FF,$70,$CE,$AF,$FB,$9A,$F0,$79,$1C,$63,$2F,$5F,$0A,$46,$0E,$2F
	.dc.b	$87,$86,$47,$F7,$EC,$1F,$41,$8C,$7C,$3C,$5E,$73,$1B,$7F,$7D,$CF
	.dc.b	$E7,$CF,$FB,$DB,$EA,$FF,$25,$3F,$E0,$16,$E3,$FD,$CB,$1F,$C3,$3F
	.dc.b	$E0,$C8,$CD,$7F,$BD,$DF,$E4,$ED,$E0,$FE,$93,$1A,$83,$A6,$F8,$7A
	.dc.b	$3E,$4F,$45,$BE,$DE,$97,$23,$15,$E7,$7A,$FB,$92,$76,$79,$5F,$0F
	.dc.b	$A1,$23,$E3,$7C,$9D,$BF,$81,$E9,$79,$8C,$6D,$F0,$5F,$4F,$81,$2F
	.dc.b	$E7,$7C,$75,$BF,$91,$E1,$7B,$8C,$61,$F2,$5F,$83,$F3,$52,$FE,$B7
	.dc.b	$BF,$6F,$E8,$7C,$0F,$C6,$31,$87,$D1,$7E,$4F,$D2,$CB,$FA,$11,$69
	.dc.b	$5F,$E3,$7F,$A8,$FC,$C4,$62,$BF,$C5,$31,$89,$FF,$1E,$7D,$5B,$E5
	.dc.b	$FF,$BE,$CE,$6B,$7D,$78,$33,$E0,$58,$FF,$25,$CF,$FB,$71,$BE,$BF
	.dc.b	$F8,$1B,$EA,$DF,$57,$D2,$3A,$FB,$F2,$B6,$72,$7C,$3C,$72,$3F,$35
	.dc.b	$EC,$DD,$FD,$40,$64,$E2,$F4,$9B,$E0,$F8,$4F,$B3,$E2,$E6,$1E,$88
	.dc.b	$B7,$D0,$E7,$F1,$5F,$08,$F5,$75,$5A,$F7,$FD,$5F,$C6,$31,$5C,$72
	.dc.b	$47,$CA,$E3,$95,$FE,$70,$F5,$DE,$36,$30,$F0,$72,$BD,$DF,$F1,$6B
	.dc.b	$6F,$E6,$E1,$89,$FF,$18,$FD,$5F,$F2,$8B,$37,$E6,$3F,$BA,$5F,$F9
	.dc.b	$9A,$34,$7F,$A4,$DF,$F4,$0B,$67,$E0,$3A,$F8,$3E,$BE,$D4,$AF,$F4
	.dc.b	$3D,$FB,$1E,$31,$E2,$F7,$A4,$60,$EA,$F9,$B4,$91,$FE,$90,$FA,$8F
	.dc.b	$66,$5E,$09,$F5,$F9,$B2,$BC,$B1,$67,$EA,$7B,$9F,$E2,$EE,$7F,$F4
	.dc.b	$93,$6B,$FD,$07,$87,$F6,$E9,$CF,$B9,$FE,$56,$AF,$F0,$0D,$F2,$71
	.dc.b	$19,$75,$E5,$CC,$3C,$DE,$AD,$BF,$0F,$97,$CC,$C6,$3E,$27,$B7,$AB
	.dc.b	$2F,$FD,$5B,$F5,$3E,$0A,$F3,$CF,$8F,$E7,$E5,$7F,$AC,$B3,$FE,$A0
	.dc.b	$FF,$A7,$14,$7F,$9B,$AE,$CF,$48,$C5,$65,$91,$FE,$BC,$F9,$47,$CF
	.dc.b	$63,$3F,$CE,$46,$27,$53,$FE,$89,$9F,$D5,$4E,$5F,$F4,$20,$70,$FF
	.dc.b	$A1,$DA,$F9,$FC,$DE,$4F,$F9,$FD,$B9,$1A,$63,$CD,$98,$7F,$AB,$27
	.dc.b	$FC,$2C,$D8,$E7,$F5,$7F,$D8,$E3,$FD,$62,$DB,$F2,$B5,$FF,$4D,$35
	.dc.b	$FF,$62,$E0,$F4,$3C,$AD,$AF,$F5,$DB,$67,$DE,$D7,$DD,$E9,$70,$24
	.dc.b	$F0,$79,$F9,$32,$39,$B6,$79,$FD,$0F,$80,$93,$E0,$7E,$2F,$C2,$48
	.dc.b	$F8,$4D,$9F,$A3,$A3,$E2,$3E,$59,$2B,$E4,$3F,$3B,$D3,$47,$F9,$1E
	.dc.b	$BF,$92,$52,$73,$1F,$EC,$78,$C4,$EE,$3F,$D6,$D3,$A8,$FF,$62,$4E
	.dc.b	$A7,$84,$E5,$FC,$B3,$5E,$ED,$7F,$D8,$1E,$95,$93,$49,$5E,$1D,$3F
	.dc.b	$D7,$15,$FE,$DB,$6C,$7E,$3F,$CB,$E9,$18,$9F,$D3,$7C,$B7,$52,$57
	.dc.b	$AF,$04,$78,$FE,$AF,$FC,$61,$3F,$ED,$F6,$D7,$F4,$90,$E1,$FE,$98
	.dc.b	$D7,$F4,$5F,$05,$E2,$24,$FA,$5E,$DE,$44,$8F,$99,$5F,$E3,$06,$C7
	.dc.b	$8F,$E1,$3F,$9F,$5F,$EB,$AB,$FC,$DD,$E4,$7E,$C2,$3C,$7D,$0F,$6F
	.dc.b	$15,$A9,$E1,$CB,$E9,$CB,$D8,$D8,$F6,$4A,$31,$1E,$C6,$BF,$03,$96
	.dc.b	$C9,$A4,$8F,$5F,$07,$95,$F8,$ED,$93,$4C,$6B,$27,$D2,$D7,$F5,$FF
	.dc.b	$09,$CF,$84,$D2,$47,$C0,$F3,$3E,$B7,$7F,$E6,$A3,$F8,$A7,$AF,$8D
	.dc.b	$7F,$7E,$D8,$FB,$35,$5F,$BB,$F1,$57,$44,$D3,$11,$BF,$AB,$7C,$AF
	.dc.b	$1F,$0D,$EF,$53,$51,$BF,$95,$B9,$7F,$84,$F5,$F3,$61,$CA,$31,$AF
	.dc.b	$E3,$E1,$5D,$FE,$43,$CD,$C1,$FE,$76,$BF,$E7,$5D,$39,$78,$9F,$E1
	.dc.b	$FB,$7A,$39,$8F,$F0,$C4,$2F,$D5,$3E,$BF,$85,$F8,$5E,$6F,$FA,$36
	.dc.b	$FF,$B9,$46,$5E,$A2,$DF,$0B,$98,$EB,$3F,$16,$3E,$37,$99,$D3,$79
	.dc.b	$5C,$AD,$3F,$2F,$E2,$37,$39,$8C,$3E,$0F,$AB,$7A,$1E,$6C,$64,$E6
	.dc.b	$3E,$D4,$BC,$3B,$12,$79,$47,$DF,$FA,$87,$D5,$FE,$07,$C7,$7C,$35
	.dc.b	$35,$1B,$F7,$6E,$1F,$0A,$F8,$B6,$3A,$7C,$97,$E2,$FE,$AF,$E2,$CF
	.dc.b	$C8,$6B,$FE,$49,$A4,$F8,$FD,$2E,$33,$D6,$30,$F7,$F3,$7F,$2D,$F5
	.dc.b	$6F,$5A,$9E,$B8,$F6,$6B,$6B,$81,$E7,$7E,$07,$9F,$A8,$C1,$EB,$7F
	.dc.b	$45,$F5,$7F,$A9,$F0,$3F,$55,$F5,$3F,$EF,$DA,$FC,$0B,$5F,$F3,$14
	.dc.b	$EF,$C1,$FE,$4C,$FF,$B8,$E3,$F0,$EE,$5F,$F3,$93,$53,$B7,$9F,$99
	.dc.b	$FB,$3F,$AB,$FE,$6B,$1F,$90,$6B,$A3,$AF,$9B,$E8,$99,$18,$3C,$3E
	.dc.b	$A7,$EF,$7E,$AD,$EE,$D3,$DF,$1F,$03,$E2,$2D,$70,$3D,$2F,$C4,$F8
	.dc.b	$7D,$46,$0F,$6B,$F2,$3D,$0F,$AB,$EE,$49,$F8,$31,$F3,$3E,$1F,$51
	.dc.b	$83,$E2,$BF,$43,$E2,$3F,$89,$70,$FF,$84,$E7,$E7,$FB,$99,$7E,$5F
	.dc.b	$C3,$57,$E8,$D7,$D8,$8F,$E3,$5E,$CB,$EA,$FF,$25,$E7,$4B,$FD,$CC
	.dc.b	$E5,$BE,$5F,$0A,$91,$83,$7E,$8F,$49,$91,$83,$9F,$CF,$FE,$7F,$EA
	.dc.b	$E4,$60,$EF,$F4,$1E,$49,$91,$83,$C7,$E8,$7E,$C7,$E1,$24,$60,$F3
	.dc.b	$F9,$DE,$C9,$91,$83,$D7,$F0,$DF,$CE,$7E,$25,$D1,$4F,$87,$1F,$13
	.dc.b	$AA,$D7,$03,$E1,$BF,$23,$D7,$D4,$60,$F8,$EF,$F4,$5F,$AB,$FE,$C1
	.dc.b	$1F,$C8,$9F,$ED,$0D,$EE,$D7,$7E,$FD,$BE,$1E,$E4,$AE,$11,$FD,$3F
	.dc.b	$EA,$DF,$17,$78,$E5,$7F,$2E,$9C,$B7,$83,$87,$CB,$C8,$C1,$FC,$D3
	.dc.b	$EA,$FF,$90,$37,$D7,$FE,$B9,$1F,$CA,$23,$5E,$BE,$3E,$2F,$3E,$9C
	.dc.b	$9E,$51,$FD,$8F,$BF,$FA,$74,$7F,$B7,$DF,$DB,$79,$BF,$DB,$5B,$E6
	.dc.b	$FF,$89,$5D,$C6,$7C,$7E,$5F,$09,$2F,$FA,$4D,$7F,$7A,$7F,$A5,$FD
	.dc.b	$5F,$F5,$36,$F8,$5F,$DC,$E3,$FB,$0C,$6A,$71,$C5,$9F,$02,$4F,$28
	.dc.b	$E1,$EE,$F1,$F7,$64,$F6,$8F,$EF,$DF,$56,$F0,$EF,$BB,$2B,$FA,$EC
	.dc.b	$E5,$FE,$BF,$F5,$6D,$BF,$C0,$CF,$F8,$6A,$33,$F1,$18,$9F,$F0,$57
	.dc.b	$7F,$A2,$5C,$5F,$E1,$38,$FE,$FD,$18,$E3,$FF,$26,$46,$27,$5C,$5C
	.dc.b	$BC,$3E,$7F,$57,$30,$FE,$E3,$69,$EF,$C1,$E0,$F0,$E5,$6A,$3F,$C3
	.dc.b	$9F,$57,$FD,$72,$3F,$BC,$AF,$80,$C9,$AE,$1F,$F1,$17,$D5,$FF,$35
	.dc.b	$9F,$E5,$FA,$FF,$8A,$DB,$B7,$F9,$8E,$7F,$9D,$1F,$0F,$0F,$60,$0B
	.dc.b	$D5,$EA,$76,$E1,$EF,$D5,$E0,$E2,$97,$FE,$0E,$BF,$C3,$3D,$0E,$7E
	.dc.b	$A7,$07,$55,$E8,$8F,$4F,$8B,$EA,$F1,$A5,$73,$8F,$6F,$A5,$A8,$C1
	.dc.b	$DD,$FF,$21,$7D,$57,$F8,$A2,$3F,$C3,$FF,$57,$FE,$94,$6F,$B9,$AF
	.dc.b	$88,$DF,$81,$67,$FC,$4B,$7C,$5F,$42,$8F,$F2,$77,$C5,$7E,$CE,$7D
	.dc.b	$9E,$93,$3F,$F2,$93,$63,$FC,$5F,$3B,$E5,$FF,$2E,$3F,$BE,$8F,$88
	.dc.b	$E7,$93,$FD,$2C,$D8,$FA,$4E,$E7,$FC,$C9,$F5,$7F,$8F,$AF,$F4,$E5
	.dc.b	$FD,$5B,$FF,$1E,$B5,$FA,$13,$B7,$EE,$3F,$9D,$AE,$7E,$AF,$F9,$A6
	.dc.b	$35,$27,$57,$FE,$A9,$7F,$7D,$1D,$3B,$F8,$1C,$7C,$5E,$5B,$7C,$2E
	.dc.b	$49,$5F,$E5,$0B,$F9,$99,$7A,$B6,$DC,$E6,$FC,$51,$E3,$F1,$7F,$63
	.dc.b	$3F,$EB,$93,$D2,$E1,$C5,$A3,$11,$E7,$6B,$F0,$3C,$F7,$D5,$FF,$AC
	.dc.b	$6B,$F9,$CB,$63,$33,$DC,$F8,$4E,$1F,$F5,$C7,$D2,$FF,$A9,$53,$48
	.dc.b	$FF,$AF,$7E,$47,$FD,$E7,$9F,$FA,$6E,$3F,$D7,$DB,$FF,$9C,$9F,$D7
	.dc.b	$57,$EB,$FE,$AF,$FD,$23,$1F,$ED,$7B,$E3,$5F,$07,$D0,$B3,$FE,$DB
	.dc.b	$6B,$FB,$79,$AF,$F9,$BA,$7B,$79,$F8,$37,$C8,$F9,$DC,$4F,$8F,$E6
	.dc.b	$78,$98,$BF,$DC,$07,$AF,$C4,$49,$E0,$7A,$DF,$1E,$BF,$E9,$58,$FF
	.dc.b	$47,$B6,$BF,$4F,$E1,$6B,$9B,$7C,$B2,$F7,$16,$BF,$84,$7D,$5F,$C8
	.dc.b	$6E,$DE,$AF,$03,$E0,$CE,$72,$FF,$64,$3F,$06,$B8,$F0,$6A,$F6,$F0
	.dc.b	$F1,$A6,$5F,$EA,$4B,$FE,$58,$D8,$EB,$E7,$C8,$C1,$B7,$C3,$E9,$C8
	.dc.b	$E5,$F1,$77,$2F,$4F,$A9,$E6,$31,$A7,$99,$F4,$FC,$54,$BF,$17,$CC
	.dc.b	$DB,$FE,$51,$F2,$FF,$5E,$7F,$D7,$2D,$8C,$FB,$9F,$82,$8C,$63,$FE
	.dc.b	$E9,$5E,$4E,$4D,$78,$D7,$0F,$F6,$1F,$D5,$BE,$73,$F1,$6D,$8C,$F8
	.dc.b	$18,$9C,$F6,$C9,$E3,$EE,$DD,$7F,$9F,$7D,$5F,$E1,$E7,$FD,$9E,$D8
	.dc.b	$CF,$81,$F9,$88,$C6,$DF,$EF,$F5,$C1,$93,$5C,$57,$0F,$F6,$CF,$D5
	.dc.b	$FF,$2F,$6E,$FF,$A2,$8F,$F7,$FC,$60,$FF,$6F,$9F,$E3,$62,$ED,$FA
	.dc.b	$87,$F8,$41,$83,$FD,$CB,$F5,$7F,$D8,$5B,$A6,$FD,$DF,$D7,$CE,$63
	.dc.b	$FA,$E3,$FE,$D7,$9E,$3A,$EB,$AF,$6F,$0C,$E4,$F6,$E8,$B8,$7F,$65
	.dc.b	$F1,$5E,$D7,$EF,$1B,$5F,$EF,$85,$F3,$B9,$9D,$79,$92,$B5,$EA,$DC
	.dc.b	$3F,$B8,$7D,$5F,$EC,$79,$55,$FD,$F6,$53,$98,$C2,$F3,$FF,$86,$B9
	.dc.b	$9F,$F0,$23,$FE,$B2,$9F,$F0,$35,$74,$9C,$FC,$2F,$0F,$77,$8B,$CB
	.dc.b	$8E,$5F,$03,$BF,$B9,$8B,$FF,$03,$78,$DF,$E3,$D9,$D4,$7F,$1D,$F1
	.dc.b	$3F,$8F,$CF,$FB,$FA,$73,$28,$E0,$FF,$83,$FE,$AD,$F3,$FF,$C9,$9B
	.dc.b	$5C,$1F,$86,$D8,$FE,$57,$F5,$6F,$86,$A7,$F2,$EF,$AB,$DE,$1F,$E1
	.dc.b	$B9,$D5,$F8,$97,$83,$C8,$E3,$19,$7A,$F8,$52,$30,$71,$7C,$3C,$12
	.dc.b	$3D,$B8,$3E,$7F,$18,$F8,$78,$7C,$E6,$38,$BE,$94,$33,$FA,$CF,$F8
	.dc.b	$C7,$EA,$FF,$31,$3A,$6E,$3F,$83,$C7,$F1,$AF,$CC,$8C,$D7,$F9,$0D
	.dc.b	$FE,$5D,$1F,$89,$6C,$64,$FE,$A6,$E1,$3F,$BE,$8E,$1E,$9F,$CB,$E9
	.dc.b	$DB,$C3,$D3,$93,$E1,$8B,$7E,$9F,$6F,$A9,$8C,$BD,$4F,$83,$E0,$E5
	.dc.b	$FC,$8F,$92,$B7,$F0,$7D,$1F,$29,$8C,$3E,$13,$E8,$F0,$65,$FD,$0F
	.dc.b	$8D,$B7,$F2,$7C,$3F,$69,$8C,$3E,$53,$F0,$3E,$66,$5F,$D8,$F8,$5B
	.dc.b	$7F,$47,$E0,$BE,$29,$8C,$3E,$93,$F2,$3E,$AE,$5F,$D1,$8B,$5F,$E6
	.dc.b	$9F,$A8,$FC,$CB,$7B,$7F,$C0,$3F,$56,$F9,$6F,$F0,$1D,$6B,$2F,$C1
	.dc.b	$BF,$89,$C7,$F4,$ED,$C3,$94,$63,$AF,$8F,$AF,$23,$F5,$BF,$56,$FA
	.dc.b	$BC,$5F,$8F,$E0,$7B,$98,$CB,$BB,$E2,$E3,$98,$7E,$6B,$D9,$B7,$EB
	.dc.b	$E2,$F3,$98,$C3,$CE,$FA,$BE,$06,$5F,$A2,$2D,$7C,$3F,$8A,$F7,$9F
	.dc.b	$E2,$2D,$3F,$BC,$C6,$6B,$DF,$F5,$7F,$10,$DB,$83,$E1,$F0,$FC,$83
	.dc.b	$FC,$57,$F0,$3F,$EB,$58,$C9,$FC,$3B,$85,$EF,$7F,$8C,$1B,$1A,$C3
	.dc.b	$FD,$3D,$18,$7F,$D4,$1F,$57,$FC,$A5,$37,$E5,$FF,$BB,$D7,$F3,$38
	.dc.b	$D0,$F8,$5F,$F1,$EB,$63,$F8,$FB,$A8,$CF,$B7,$27,$F9,$1D,$34,$C6
	.dc.b	$1E,$2F,$80,$E6,$63,$8B,$F8,$3E,$5F,$DF,$E7,$CF,$E5,$F8,$0F,$F2
	.dc.b	$6F,$D3,$FA,$EB,$71,$E4,$F3,$26,$7E,$48,$B7,$FE,$BF,$F6,$FF,$C6
	.dc.b	$57,$FC,$51,$B5,$FD,$47,$0F,$EE,$95,$BF,$73,$FC,$B7,$5F,$DA,$E3
	.dc.b	$93,$88,$CB,$AF,$26,$5F,$F4,$2B,$7A,$B8,$C3,$C5,$E5,$F4,$31,$97
	.dc.b	$8D,$EF,$EA,$CC,$3F,$A5,$FD,$47,$86,$BD,$13,$E4,$F9,$E9,$5F,$D4
	.dc.b	$34,$FE,$9F,$5E,$D9,$FF,$39,$D9,$FF,$40,$63,$FE,$EB,$8C,$5F,$FA
	.dc.b	$36,$11,$87,$FD,$20,$D7,$FD,$41,$F2,$6F,$9D,$E7,$08,$C5,$73,$6C
	.dc.b	$7F,$A6,$5A,$F2,$F0,$BF,$D0,$F1,$F9,$08,$FF,$7F,$3C,$31,$FE,$92
	.dc.b	$E6,$19,$3C,$BF,$0B,$70,$F5,$36,$7F,$D5,$0D,$7F,$D7,$1F,$57,$FD
	.dc.b	$92,$7E,$03,$63,$FD,$5C,$D7,$7E,$BF,$BB,$07,$A1,$49,$7E,$F7,$E1
	.dc.b	$DB,$D7,$A4,$9E,$EF,$3F,$26,$47,$C7,$6E,$3F,$1F,$E8,$3C,$04,$9F
	.dc.b	$98,$F8,$FC,$59,$1F,$3D,$B3,$F0,$3E,$27,$E2,$92,$7E,$2B,$F2,$3D
	.dc.b	$1F,$F1,$34,$7F,$B6,$F6,$46,$2B,$FC,$69,$18,$9F,$F2,$74,$0F,$A7
	.dc.b	$3B,$71,$2F,$43,$11,$F6,$1B,$19,$AE,$73,$96,$C6,$63,$FD,$AC,$D8
	.dc.b	$F3,$4F,$23,$AE,$CF,$2A,$92,$7A,$C6,$5B,$E0,$F0,$23,$83,$C7,$F5
	.dc.b	$5F,$E2,$C7,$FD,$FC,$DC,$7F,$CA,$CD,$7E,$13,$5F,$D1,$7C,$0F,$90
	.dc.b	$93,$E2,$7C,$FE,$4C,$8F,$2B,$67,$FD,$E9,$5F,$E6,$CF,$27,$F6,$8D
	.dc.b	$8F,$F3,$31,$E0,$E9,$FF,$27,$38,$11,$B7,$D3,$EC,$7F,$9B,$70,$B1
	.dc.b	$EE,$8C,$3E,$C6,$BE,$CE,$3A,$F4,$6D,$1A,$6B,$47,$9F,$49,$5F,$49
	.dc.b	$E3,$A7,$F0,$F8,$F3,$B6,$CD,$35,$CC,$F1,$69,$2B,$F7,$2F,$3C,$1F
	.dc.b	$0F,$49,$3F,$C4,$DE,$DB,$FF,$A3,$FB,$3F,$86,$FA,$97,$7F,$F4,$96
	.dc.b	$1F,$E7,$6C,$3F,$CF,$7A,$30,$D4,$7E,$03,$1A,$FD,$08,$B3,$B8,$E3
	.dc.b	$3A,$FA,$A7,$CA,$70,$F0,$5E,$DD,$35,$1B,$F8,$4B,$87,$FA,$32,$F8
	.dc.b	$B6,$3A,$7C,$7F,$8B,$4F,$F3,$A4,$FF,$A4,$9A,$B7,$FF,$39,$DA,$BC
	.dc.b	$9E,$8F,$2A,$ED,$18,$7C,$1E,$1F,$BB,$CA,$FF,$45,$C7,$F5,$D8,$F6
	.dc.b	$1E,$DF,$0E,$C7,$03,$D0,$FC,$1F,$2B,$96,$B2,$7D,$91,$F1,$7E,$13
	.dc.b	$51,$83,$E0,$3F,$3B,$EA,$9E,$77,$8F,$09,$5F,$EB,$09,$CB,$7C,$3E
	.dc.b	$15,$23,$07,$FA,$DB,$EA,$1F,$55,$F6,$BC,$67,$C0,$53,$EE,$CE,$AB
	.dc.b	$87,$C0,$E2,$E4,$60,$E9,$F2,$3F,$84,$FA,$AF,$E8,$51,$FE,$C4,$6B
	.dc.b	$F4,$FB,$BF,$DC,$18,$2B,$BF,$C1,$F8,$19,$5E,$78,$F1,$F9,$9F,$90
	.dc.b	$FA,$A7,$AB,$4C,$D7,$B7,$6A,$F7,$3F,$9C,$62,$BF,$6C,$FB,$BC,$EB
	.dc.b	$5C,$4F,$8C,$FE,$7B,$EA,$BF,$D7,$7D,$BF,$EB,$CF,$A8,$FF,$80,$35
	.dc.b	$FE,$E0,$6B,$DC,$4F,$0D,$BF,$DF,$CF,$EB,$63,$4E,$5F,$E9,$8D,$78
	.dc.b	$F8,$DE,$1F,$33,$F5,$FF,$54,$FA,$BA,$66,$BF,$DE,$37,$FC,$55,$B5
	.dc.b	$97,$BF,$9D,$EB,$F5,$3F,$73,$F5,$7F,$C3,$67,$FD,$F2,$D7,$CB,$1F
	.dc.b	$07,$D0,$E1,$1D,$A3,$0F,$83,$E1,$2B,$EC,$9F,$91,$1F,$27,$E4,$7E
	.dc.b	$C7,$B7,$2B,$CF,$1F,$37,$D8,$D4,$60,$F5,$BF,$47,$E1,$BF,$87,$F1
	.dc.b	$7F,$84,$A7,$E8,$7B,$59,$7E,$0F,$C2,$D7,$E6,$D7,$D9,$8F,$E2,$FE
	.dc.b	$C3,$EA,$FF,$2D,$E6,$CB,$FE,$F9,$D1,$FE,$F8,$AD,$46,$3A,$3F,$C0
	.dc.b	$0F,$F9,$07,$CB,$E7,$84,$8D,$1F,$E0,$5A,$FE,$2E,$FE,$0D,$7A,$3E
	.dc.b	$2A,$93,$C0,$FF,$26,$47,$9B,$E7,$FF,$55,$F5,$72,$75,$69,$F5,$FD
	.dc.b	$06,$E7,$31,$87,$7E,$87,$EF,$7E,$0A,$46,$0E,$7E,$7F,$C4,$99,$18
	.dc.b	$3B,$FC,$27,$F3,$EF,$87,$FE,$BB,$1F,$E9,$CE,$AF,$97,$3F,$33,$DD
	.dc.b	$B3,$C0,$F3,$BF,$43,$D8,$D4,$60,$F5,$BF,$D2,$BE,$AF,$F3,$11,$FA
	.dc.b	$95,$E0,$E4,$D7,$83,$BF,$73,$87,$BD,$2F,$E5,$5F,$F6,$27,$E6,$7D
	.dc.b	$5B,$E2,$BF,$AB,$AE,$3A,$62,$33,$C1,$AF,$94,$97,$CA,$3F,$AF,$FD
	.dc.b	$5F,$F9,$CA,$3E,$C3,$FC,$C1,$FE,$00,$DE,$C7,$67,$E2,$23,$84,$EB
	.dc.b	$E2,$39,$FA,$72,$39,$C7,$F6,$CE,$FF,$EB,$38,$72,$9F,$F2,$84,$7F
	.dc.b	$09,$8C,$47,$F5,$BF,$37,$87,$C6,$F1,$F1,$F9,$7C,$34,$AF,$C7,$60
	.dc.b	$9F,$EF,$3F,$57,$FD,$8E,$7F,$24,$FE,$59,$BE,$16,$AB,$7C,$5C,$3C
	.dc.b	$09,$5C,$23,$9F,$BF,$D3,$E0,$24,$F4,$8F,$F0,$3F,$D5,$BC,$3B,$EF
	.dc.b	$4A,$FD,$14,$E5,$FD,$27,$D5,$FF,$89,$1B,$F1,$9F,$E1,$08,$FF,$30
	.dc.b	$46,$B8,$3F,$C2,$7E,$07,$A6,$3F,$99,$3F,$B2,$E4,$F8,$ED,$E8,$FD
	.dc.b	$DB,$B8,$CF,$87,$CB,$EA,$E5,$FE,$D6,$93,$DB,$7E,$FE,$1C,$AD,$47
	.dc.b	$F8,$93,$EA,$FF,$B5,$C7,$EF,$57,$C1,$64,$D7,$07,$F8,$AB,$EA,$FF
	.dc.b	$A0,$4F,$F3,$DA,$E1,$EC,$E7,$6D,$70,$EF,$D6,$95,$FC,$27,$0F,$86
	.dc.b	$FF,$51,$7E,$F3,$9E,$1E,$C7,$7E,$BF,$07,$AB,$67,$94,$61,$E9,$E4
	.dc.b	$79,$7C,$89,$18,$3C,$1E,$9F,$A6,$64,$60,$FE,$39,$F5,$4F,$89,$FF
	.dc.b	$23,$FD,$5F,$FA,$96,$7F,$90,$35,$FE,$D1,$3B,$78,$F7,$FF,$93,$A3
	.dc.b	$11,$AA,$C4,$FF,$94,$BE,$33,$F6,$1C,$3F,$CD,$BD,$2F,$E0,$57,$FD
	.dc.b	$CD,$BD,$9F,$A8,$AD,$CF,$FA,$ED,$FD,$AC,$67,$E2,$3A,$1D,$7E,$27
	.dc.b	$FC,$BF,$F5,$7F,$89,$BF,$F6,$16,$0E,$5F,$F3,$1B,$63,$F9,$E4,$ED
	.dc.b	$3F,$8F,$CB,$FA,$5B,$92,$78,$DD,$BC,$6E,$EC,$AE,$71,$E2,$DB,$BF
	.dc.b	$24,$9E,$F1,$E6,$DF,$51,$83,$C6,$FA,$BC,$6F,$F3,$24,$7F,$AA,$71
	.dc.b	$FE,$49,$BF,$FA,$AE,$3C,$F3,$A7,$FD,$63,$3A,$78,$79,$EF,$AB,$E1
	.dc.b	$3F,$CB,$DB,$19,$9D,$C7,$FA,$85,$7D,$25,$23,$FE,$A5,$7F,$D4,$DF
	.dc.b	$45,$FC,$87,$1F,$E3,$F2,$8C,$6F,$FE,$C1,$AF,$F7,$A7,$D5,$FF,$A3
	.dc.b	$2B,$FB,$BD,$69,$FE,$F3,$C9,$5F,$F7,$D3,$5E,$97,$FD,$CD,$FF,$37
	.dc.b	$B7,$93,$47,$1E,$37,$CB,$E2,$78,$7C,$C7,$33,$17,$FE,$43,$0F,$7F
	.dc.b	$8C,$93,$C0,$F0,$3E,$5D,$7F,$D2,$D1,$FE,$05,$6D,$7E,$6F,$C1,$D7
	.dc.b	$36,$F9,$65,$EE,$2D,$7F,$94,$FE,$AF,$DA,$DD,$FD,$5D,$FF,$0E,$73
	.dc.b	$97,$F4,$C7,$E0,$D7,$1E,$0D,$5E,$DE,$1E,$2C,$CB,$FC,$35,$7F,$C7
	.dc.b	$9B,$1D,$7C,$E9,$18,$36,$F8,$7D,$49,$1C,$BE,$3E,$E5,$E9,$F5,$7C
	.dc.b	$C6,$34,$F3,$3E,$9F,$8D,$97,$E2,$FA,$9B,$7F,$E7,$3F,$99,$FA,$F3
	.dc.b	$FE,$2F,$6C,$67,$DE,$FC,$24,$63,$1F,$ED,$AB,$C8,$C9,$AF,$22,$E1
	.dc.b	$FE,$39,$FA,$B7,$CE,$FE,$35,$B1,$9F,$07,$13,$9E,$D9,$3C,$7D,$DB
	.dc.b	$AF,$FA,$27,$EA,$FF,$0F,$3F,$E4,$56,$C6,$7C,$0F,$CD,$46,$36,$FF
	.dc.b	$01,$AE,$0C,$9A,$E2,$B8,$7F,$93,$FE,$AD,$BF,$E8,$DB,$19,$EB,$C4
	.dc.b	$FE,$9E,$3F,$C5,$47,$F9,$5C,$B9,$FF,$A7,$DF,$F1,$61,$9D,$FF,$5D
	.dc.b	$F5,$69,$A7,$3F,$72,$D7,$F9,$86,$5B,$70,$38,$F6,$59,$D6,$9D,$0E
	.dc.b	$BD,$36,$3F,$CD,$5E,$2F,$F9,$21,$BE,$CF,$EF,$E1,$F5,$76,$31,$E6
	.dc.b	$4A,$E3,$E9,$DD,$BF,$D7,$5F,$57,$FB,$0E,$6F,$F1,$7B,$6B,$FD,$06
	.dc.b	$FF,$9F,$1F,$F3,$D3,$99,$F8,$0B,$E1,$BF,$88,$36,$77,$EE,$62,$73
	.dc.b	$C5,$27,$9F,$B7,$70,$FF,$66,$78,$DF,$E2,$99,$FF,$44,$36,$33,$E2
	.dc.b	$FF,$1D,$8C,$67,$FE,$46,$FA,$B7,$C1,$FF,$21,$6E,$1F,$E9,$17,$FC
	.dc.b	$AE,$D8,$FE,$4D,$F5,$6F,$86,$A7,$F2,$9F,$AB,$DE,$7F,$ED,$F9,$D5
	.dc.b	$FF,$A6,$D7,$83,$C8,$E3,$19,$7A,$F8,$52,$30,$71,$7C,$3C,$32,$3F
	.dc.b	$D4,$D8,$3E,$1B,$18,$F8,$78,$BC,$E6,$36,$FF,$32,$E7,$F3,$E7,$FD
	.dc.b	$E9,$F5,$7F,$93,$C7,$FD,$60,$D8,$FF,$7C,$63,$F8,$69,$CC,$7F,$47
	.dc.b	$9F,$E4,$B1,$FE,$C2,$6C,$7E,$75,$E5,$7F,$E7,$47,$30,$7F,$6D,$3C
	.dc.b	$7D,$2F,$95,$D1,$D5,$ED,$E9,$49,$F2,$C5,$BF,$4F,$B7,$E3,$31,$8F
	.dc.b	$77,$CD,$F0,$52,$FE,$3F,$C9,$5B,$F8,$3C,$FE,$B3,$18,$79,$DF,$77
	.dc.b	$7C,$BF,$9F,$F1,$96,$FE,$4F,$83,$F0,$CC,$61,$EF,$7E,$37,$CD,$4B
	.dc.b	$FA,$FE,$FD,$BF,$A3,$F3,$5F,$2C,$C6,$1F,$1D,$F9,$DF,$57,$2F,$EA
	.dc.b	$C5,$A5,$7F,$A8,$3E,$AB,$F3,$31,$8A,$FF,$7A,$C6,$27,$FD,$4B,$F5
	.dc.b	$6F,$97,$FF,$00,$4E,$6B,$7D,$58,$33,$DF,$63,$FD,$E9,$2F,$5B,$89
	.dc.b	$FE,$09,$FA,$A7,$D5,$E2,$B7,$F0,$1D,$8C,$6B,$E5,$3F,$91,$61,$FD
	.dc.b	$36,$7C,$FC,$7E,$43,$F5,$9E,$C5,$8F,$0F,$17,$A0,$C5,$F3,$17,$FE
	.dc.b	$51,$9F,$3F,$C0,$7B,$8E,$0E,$0F,$5F,$C4,$FC,$23,$D5,$D5,$63,$DF
	.dc.b	$F5,$5F,$18,$C6,$1D,$F0,$47,$CA,$E2,$95,$FC,$F7,$93,$FD,$A7,$EB
	.dc.b	$7F,$A9,$A3,$E4,$CF,$F8,$BF,$91,$EE,$53,$8E,$5E,$B8,$CD,$73,$3F
	.dc.b	$A4,$43,$EA,$FF,$96,$37,$E0,$BF,$C7,$5F,$2F,$F2,$E3,$FA,$74,$E8
	.dc.b	$FF,$48,$57,$38,$CF,$C2,$7F,$90,$67,$4D,$F6,$A5,$6E,$97,$F0,$23
	.dc.b	$17,$FD,$83,$5F,$F7,$8B,$FD,$FA,$BE,$A3,$63,$37,$DF,$DF,$F0,$79
	.dc.b	$93,$0E,$DF,$4F,$F0,$31,$E3,$3E,$5F,$36,$57,$82,$2C,$FE,$07,$DB
	.dc.b	$FF,$1A,$E4,$AF,$8F,$18,$7E,$4E,$1E,$77,$3E,$DF,$F9,$86,$7F,$B7
	.dc.b	$C7,$BD,$B7,$D6,$75,$1B,$F2,$A5,$7A,$3D,$5B,$7E,$3F,$33,$A1,$8C
	.dc.b	$BC,$91,$FE,$01,$C1,$3E,$1F,$84,$98,$7C,$FF,$A9,$F0,$E1,$C3,$0F
	.dc.b	$F0,$1D,$79,$FE,$7E,$5F,$9E,$2C,$FE,$8E,$DF,$0D,$FE,$7D,$5E,$6D
	.dc.b	$8C,$39,$F3,$66,$1E,$18,$B7,$CE,$7F,$D1,$EF,$F8,$66,$D1,$88,$46
	.dc.b	$2B,$FD,$50,$D8,$FC,$C9,$C7,$CD,$E1,$06,$92,$78,$F9,$9A,$23,$83
	.dc.b	$FD,$69,$F2,$7F,$E1,$DB,$EA,$1C,$3E,$06,$BF,$AA,$6B,$FE,$33,$7F
	.dc.b	$8E,$47,$4F,$2F,$D9,$18,$9E,$AD,$7D,$7F,$56,$F9,$DF,$84,$DB,$FF
	.dc.b	$64,$36,$7C,$AD,$7E,$1F,$17,$F9,$56,$7F,$D9,$CD,$8F,$5B,$5D,$FC
	.dc.b	$37,$CF,$87,$C4,$D2,$57,$27,$D3,$E5,$FE,$42,$3E,$13,$6F,$E1,$35
	.dc.b	$F8,$4D,$7E,$57,$97,$73,$FE,$59,$1F,$ED,$06,$D7,$CD,$9E,$18,$3D
	.dc.b	$2A,$49,$FF,$06,$5F,$E8,$63,$0F,$FB,$6A,$78,$4E,$A3,$FC,$C3,$3A
	.dc.b	$8F,$8F,$F2,$9D,$88,$EE,$91,$E9,$FA,$BF,$F1,$9C,$FE,$41,$B7,$FC
	.dc.b	$A9,$AF,$F9,$CD,$A9,$FE,$D7,$1C,$7D,$3E,$84,$7C,$D1,$FE,$31,$6F
	.dc.b	$8F,$E1,$23,$AF,$85,$FE,$7F,$7F,$96,$AF,$F3,$8F,$93,$F9,$88,$F0
	.dc.b	$7C,$F7,$FB,$ED,$D5,$70,$7F,$35,$E5,$FF,$42,$C2,$C7,$C7,$DF,$FB
	.dc.b	$E3,$67,$4F,$F4,$8F,$DA,$39,$57,$EC,$9B,$19,$C7,$6F,$FA,$56,$32
	.dc.b	$FC,$5F,$53,$2F,$C5,$6E,$7F,$F0,$02,$3E,$1C,$63,$D9,$91,$F1,$9A
	.dc.b	$FF,$1B,$14,$FA,$DF,$85,$18,$6C,$66,$3E,$33,$5F,$DF,$AF,$A2,$D9
	.dc.b	$23,$6F,$2F,$A2,$91,$C9,$AF,$D3,$F4,$BF,$B6,$F8,$BF,$CB,$F7,$FF
	.dc.b	$42,$CF,$F6,$16,$BF,$E0,$FD,$15,$C3,$6F,$C6,$7C,$6D,$AF,$C6,$4D
	.dc.b	$7F,$18,$DA,$FF,$67,$C7,$F6,$DF,$AD,$65,$BF,$09,$EF,$D3,$94,$EA
	.dc.b	$B8,$7C,$C6,$2E,$46,$0E,$9E,$B7,$87,$5F,$F4,$34,$7A,$FA,$39,$CF
	.dc.b	$4F,$5A,$CF,$03,$C8,$F8,$3E,$1B,$D5,$E8,$FF,$A6,$A7,$FB,$BC,$65
	.dc.b	$F5,$7C,$35,$7F,$C6,$9A,$6D,$B1,$EA,$F4,$39,$E9,$27,$E1,$47,$BB
	.dc.b	$E2,$B5,$18,$3E,$33,$F2,$FE,$B5,$EA,$79,$30,$95,$FE,$17,$9C,$B7
	.dc.b	$C3,$E1,$52,$30,$7F,$87,$7E,$A9,$F5,$BF,$5F,$CF,$7C,$35,$3E,$D4
	.dc.b	$EA,$B8,$7C,$06,$2E,$46,$0E,$9E,$A7,$E0,$7E,$B5,$AD,$39,$C7,$4F
	.dc.b	$56,$D7,$03,$C8,$F8,$3C,$FF,$C5,$7D,$6F,$F8,$22,$7F,$C6,$8D,$7F
	.dc.b	$CC,$59,$FF,$44,$7D,$9A,$F1,$7D,$D5,$EA,$8F,$5F,$AB,$67,$67,$C3
	.dc.b	$7F,$31,$F5,$9F,$E1,$1F,$13,$FC,$25,$F5,$4F,$37,$F3,$8D,$BF,$F2
	.dc.b	$35,$6F,$C6,$FF,$2B,$BF,$A8,$BF,$F0,$D3,$99,$07,$0D,$F3,$FC,$3F
	.dc.b	$17,$FA,$9F,$AD,$FE,$7B,$3F,$E5,$36,$BF,$E8,$1A,$4F,$A3,$D9,$E3
	.dc.b	$3E,$48,$C3,$E6,$FA,$CF,$DA,$7D,$6B,$E0,$A9,$F0,$23,$E0,$FC,$5D
	.dc.b	$AE,$07,$B9,$F8,$BE,$CE,$A3,$07,$C2,$7E,$4F,$CB,$FD,$8F,$81,$93
	.dc.b	$F1,$A3,$E6,$FC,$5E,$A3,$07,$CA,$7E,$8F,$BB,$F3,$BD,$79,$3F,$3A
	.dc.b	$3E,$AF,$B5,$A8,$C1,$F4,$9F,$E2,$1E,$DF,$F2,$58,$FF,$3D,$78,$0F
	.dc.b	$3B,$F2,$86,$BD,$7D,$F9,$76,$BF,$D0,$17,$C5,$FF,$57,$3F,$C4,$5B
	.dc.b	$E9,$70,$77,$3E,$08,$F0,$FD,$17,$E8,$7E,$06,$57,$48,$F2,$FD,$16
	.dc.b	$A3,$07,$81,$F4,$FA,$3F,$B3,$F8,$49,$3E,$48,$F6,$F9,$3A,$8C,$1E
	.dc.b	$87,$E0,$FC,$3F,$F3,$1F,$8B,$74,$53,$35,$EE,$EA,$AF,$BA,$BE,$4C
	.dc.b	$7C,$AF,$62,$CE,$CF,$8A,$FF,$3D,$FA,$DF,$F7,$24,$7F,$A9,$DF,$C9
	.dc.b	$B7,$BB,$5D,$FB,$F7,$B8,$7B,$D2,$B8,$47,$F4,$7F,$AD,$7C,$65,$E2
	.dc.b	$95,$FE,$B3,$9C,$B7,$87,$87,$CB,$C8,$C1,$FE,$B9,$FA,$D7,$B1,$FD
	.dc.b	$4D,$FE,$4B,$1F,$EE,$96,$F9,$5F,$E8,$F7,$71,$C7,$E2,$F8,$7A,$92
	.dc.b	$30,$7F,$B1,$7B,$FF,$A6,$BF,$EF,$37,$F8,$07,$9B,$FE,$D8,$6C,$7F
	.dc.b	$00,$77,$87,$FA,$75,$78,$DF,$D9,$D7,$CE,$53,$11,$FD,$AF,$EB,$7F
	.dc.b	$35,$7F,$ED,$55,$E0,$64,$D5,$EF,$87,$87,$7C,$BE,$11,$CF,$DD,$E9
	.dc.b	$EF,$49,$E9,$1F,$DE,$7E,$B5,$E1,$5F,$76,$57,$FB,$96,$72,$FF,$B9
	.dc.b	$FE,$B7,$D2,$DF,$96,$FF,$01,$41,$BC,$1F,$E0,$5E,$EF,$4D,$7F,$BB
	.dc.b	$DF,$D7,$71,$7F,$B0,$5B,$C7,$FC,$A2,$7F,$8F,$F0,$F1,$F0,$B9,$7D
	.dc.b	$6C,$9E,$51,$D7,$83,$B7,$85,$27,$B4,$7F,$85,$7E,$B7,$FA,$FC,$7F
	.dc.b	$BF,$D7,$C0,$E4,$D7,$0F,$F8,$6F,$EB,$7F,$9B,$CF,$F3,$2D,$BF,$DA
	.dc.b	$CD,$D7,$CD,$3F,$CF,$0F,$81,$87,$F4,$0A,$FF,$6E,$43,$D4,$E8,$75
	.dc.b	$E1,$ED,$D5,$DF,$8A,$61,$C7,$1E,$B3,$E3,$F4,$F7,$39,$3F,$A6,$E3
	.dc.b	$F8,$C7,$01,$47,$AF,$C4,$F6,$78,$B2,$B7,$5F,$E3,$65,$A7,$A3,$D2
	.dc.b	$E1,$1A,$BF,$89,$1F,$E3,$FF,$AC,$7C,$47,$F9,$07,$EB,$5E,$FD,$35
	.dc.b	$1B,$F0,$2E,$3E,$6B,$E2,$D8,$FF,$25,$7C,$5F,$ED,$B8,$3F,$05,$E8
	.dc.b	$F9,$75,$46,$35,$DF,$C3,$F0,$F2,$6E,$5F,$0B,$1E,$4F,$F7,$27,$F7
	.dc.b	$EF,$BA,$FF,$1D,$F5,$BF,$90,$8F,$EE,$F9,$7D,$77,$E5,$B5,$E0,$7F
	.dc.b	$7C,$8E,$09,$FC,$EB,$A3,$1E,$AF,$F9,$96,$36,$6B,$3F,$5C,$F3,$F1
	.dc.b	$F5,$E0,$C8,$FB,$75,$FC,$09,$E8,$FC,$87,$B6,$1F,$81,$6F,$3F,$7E
	.dc.b	$1F,$07,$05,$9E,$71,$87,$AF,$8F,$F8,$07,$F6,$07,$B5,$C3,$8C,$11
	.dc.b	$88,$F9,$2D,$7E,$27,$9E,$FA,$DF,$F6,$05,$7F,$3E,$6C,$66,$7B,$9F
	.dc.b	$25,$C3,$FE,$C7,$FA,$7F,$F5,$3A,$69,$1F,$F6,$67,$D3,$7C,$AD,$FF
	.dc.b	$D6,$0D,$E1,$EB,$5F,$E3,$2F,$AD,$FF,$47,$DF,$F0,$DA,$D3,$FC,$3F
	.dc.b	$96,$BF,$E3,$86,$A8,$B1,$8E,$59,$87,$E1,$BE,$56,$DF,$E0,$47,$CE
	.dc.b	$72,$33,$9F,$FB,$15,$EB,$F3,$92,$FF,$73,$5C,$1B,$1E,$4A,$FF,$0E
	.dc.b	$47,$EF,$9B,$3F,$82,$F0,$B5,$CF,$BE,$39,$7B,$8B,$5F,$95,$FA,$DF
	.dc.b	$64,$FF,$08,$6C,$7E,$33,$C1,$F8,$71,$8C,$7F,$2F,$28,$FF,$4A,$D7
	.dc.b	$1D,$B7,$BF,$07,$8F,$8D,$27,$B7,$8F,$70,$F1,$79,$FD,$8C,$65,$DB
	.dc.b	$2F,$EA,$77,$E5,$F4,$BC,$67,$93,$E3,$EC,$7A,$7D,$4F,$39,$8C,$7D
	.dc.b	$CF,$AB,$E3,$64,$FB,$FE,$72,$D7,$ED,$FE,$69,$CF,$F8,$26,$D7,$F2
	.dc.b	$C5,$EE,$66,$75,$E5,$4A,$D7,$8F,$70,$FE,$05,$F5,$AF,$9F,$FC,$6B
	.dc.b	$6B,$F9,$BB,$FE,$22,$6C,$6B,$C6,$91,$AF,$76,$DF,$F0,$DF,$AD,$FC
	.dc.b	$33,$7B,$FF,$33,$1F,$E2,$2A,$FE,$86,$EB,$97,$EC,$AE,$0D,$9C,$38
	.dc.b	$AD,$7F,$4A,$FA,$D6,$DF,$A4,$6C,$67,$B3,$13,$FA,$88,$F5,$67,$FC
	.dc.b	$85,$B9,$7E,$21,$FE,$12,$7F,$56,$32,$7F,$7F,$FA,$DF,$F5,$3B,$FD
	.dc.b	$69,$B3,$CF,$DD,$FD,$AC,$62,$BF,$26,$FB,$E7,$F9,$83,$63,$3D,$BC
	.dc.b	$3C,$3E,$3A,$4C,$7B,$75,$5C,$7F,$9C,$78,$BF,$EA,$E9,$FE,$D6,$D8
	.dc.b	$CF,$C9,$FF,$06,$8C,$65,$FA,$47,$FC,$7B,$5C,$3C,$CD,$7A,$BB,$7F
	.dc.b	$86,$FD,$6F,$EE,$2F,$F5,$6D,$AF,$F1,$03,$FE,$3E,$7F,$BD,$B9,$C7
	.dc.b	$FD,$96,$FF,$AE,$E3,$3F,$0B,$D3,$DC,$EA,$F6,$E1,$93,$CB,$DB,$B8
	.dc.b	$7F,$55,$F1,$FF,$5A,$DF,$17,$F9,$0C,$3C,$7B,$1F,$C9,$3E,$B5,$F2
	.dc.b	$94,$C6,$34,$FF,$08,$FD,$6B,$E1,$BF,$95,$B6,$3F,$C2,$BF,$5B,$FD
	.dc.b	$E9,$DC,$EB,$C1,$91,$83,$9B,$CB,$C2,$91,$83,$BB,$DB,$86,$47,$6F
	.dc.b	$9F,$BB,$FF,$73,$CB,$E4,$37,$8B,$3C,$9F,$C7,$27,$2F,$F1,$EF,$AD
	.dc.b	$FC,$76,$1F,$E5,$36,$DF,$FB,$F3,$57,$4F,$F4,$89,$CC,$9F,$E5,$15
	.dc.b	$C7,$8F,$53,$B7,$87,$89,$27,$E2,$FC,$2D,$AF,$F0,$55,$BD,$3E,$C6
	.dc.b	$73,$F3,$3E,$1F,$72,$61,$FE,$42,$AE,$0D,$8F,$37,$A3,$23,$E9,$FC
	.dc.b	$9D,$C7,$BF,$4F,$AC,$C6,$5F,$0D,$F7,$78,$12,$FE,$DF,$C6,$DB,$F3
	.dc.b	$F8,$5F,$0C,$C6,$1F,$2D,$F8,$DF,$35,$2F,$F0,$7F,$0B,$6F,$DF,$F0
	.dc.b	$5F,$2C,$C6,$1F,$4D,$F9,$DF,$4D,$2F,$DD,$16,$BF,$C6,$1F,$55,$F9
	.dc.b	$69,$FF,$31,$B6,$3F,$C6,$9F,$5A,$F9,$7F,$EF,$F3,$9C,$7F,$AE,$AE
	.dc.b	$CC,$9A,$F0,$6E,$3D,$3D,$6E,$27,$F8,$23,$EB,$5F,$5B,$89,$F0,$FB
	.dc.b	$1D,$8C,$6B,$E6,$7C,$3C,$72,$7F,$4B,$EC,$5A,$F5,$F1,$79,$8C,$6D
	.dc.b	$E6,$7D,$3E,$FC,$BF,$7C,$5A,$F8,$7F,$17,$EE,$3C,$DD,$56,$BE,$2F
	.dc.b	$D6,$FC,$23,$15,$F2,$9F,$8B,$C7,$27,$FD,$1B,$EB,$3C,$7C,$5F,$F9
	.dc.b	$7F,$91,$EE,$7F,$8A,$9B,$7F,$4B,$0C,$4F,$F8,$BB,$EB,$7F,$8E,$59
	.dc.b	$BF,$2F,$FD,$C2,$3F,$CD,$51,$A3,$FD,$1F,$7F,$E9,$E6,$C6,$BE,$0B
	.dc.b	$A7,$C1,$F5,$F8,$59,$7F,$EA,$4C,$1B,$63,$0E,$BF,$01,$E3,$31,$C9
	.dc.b	$FE,$1E,$CF,$FB,$B4,$F8,$B7,$F4,$1F,$E4,$6F,$A7,$F6,$C6,$72,$F2
	.dc.b	$79,$FE,$D3,$6E,$0F,$F5,$AF,$B7,$FE,$28,$9F,$F4,$33,$71,$FF,$16
	.dc.b	$DF,$F6,$49,$09,$D5,$7F,$B9,$5F,$EB,$CD,$7F,$CB,$26,$A3,$8B,$A6
	.dc.b	$F9,$9A,$3D,$5E,$B5,$AF,$2F,$99,$C8,$C5,$79,$9E,$BE,$B4,$9F,$F6
	.dc.b	$7F,$D4,$75,$AF,$34,$F8,$BE,$7A,$57,$C0,$8B,$3F,$EA,$5A,$79,$7F
	.dc.b	$E4,$C8,$FF,$80,$11,$8A,$FF,$42,$C2,$30,$FF,$A2,$1A,$FF,$A6,$17
	.dc.b	$97,$CE,$3E,$4C,$62,$79,$B6,$39,$4E,$DE,$CF,$FA,$0B,$E4,$FF,$9E
	.dc.b	$D7,$BA,$31,$E9,$FF,$A1,$AF,$FD,$BF,$3D,$3C,$BE,$51,$8C,$3C,$D1
	.dc.b	$D7,$0F,$F0,$54,$7F,$AB,$9A,$FD,$E0,$E1,$FC,$0B,$5C,$F9,$1C,$C9
	.dc.b	$7F,$1A,$7F,$CA,$AD,$F9,$FF,$01,$1C,$1F,$33,$87,$FC,$84,$DF,$84
	.dc.b	$F8,$51,$8C,$3F,$D6,$AD,$7F,$D8,$AB,$E1,$F6,$4A,$E8,$F2,$F4,$3F
	.dc.b	$DB,$B1,$C1,$B5,$F9,$96,$BF,$96,$6B,$F0,$BC,$6B,$91,$A4,$AF,$F2
	.dc.b	$2C,$FE,$7B,$D0,$E3,$89,$A4,$9D,$CF,$EA,$FE,$B3,$64,$8C,$D2,$31
	.dc.b	$3F,$95,$8C,$4E,$A7,$2F,$FB,$A3,$EA,$BF,$C3,$95,$F9,$B6,$C6,$4F
	.dc.b	$6B,$88,$FC,$06,$2F,$82,$F7,$92,$BC,$F1,$97,$D3,$5F,$D5,$5B,$E3
	.dc.b	$F4,$23,$AE,$3F,$E6,$78,$7F,$BB,$AB,$FC,$2F,$E3,$7E,$B1,$7C,$FF
	.dc.b	$FB,$76,$75,$5C,$1F,$67,$B3,$F9,$5C,$2C,$78,$A3,$0F,$E6,$6D,$18
	.dc.b	$8F,$E3,$3F,$57,$CE,$BF,$3D,$08,$C3,$FA,$16,$BC,$31,$E4,$AF,$FD
	.dc.b	$6A,$FC,$2F,$47,$31,$88,$FA,$11,$8F,$62,$47,$C8,$6B,$F5,$79,$B8
	.dc.b	$91,$F9,$AF,$83,$07,$D0,$FF,$12,$9F,$43,$63,$31,$FC,$59,$AF,$B8
	.dc.b	$F7,$7C,$A6,$C8,$ED,$E5,$E6,$C8,$E4,$D7,$EC,$79,$FF,$C1,$7C,$6F
	.dc.b	$F7,$8F,$1F,$F9,$CA,$7F,$A2,$C6,$3D,$7D,$70,$7D,$3B,$B5,$FD,$2B
	.dc.b	$7F,$F5,$DB,$C2,$74,$D3,$53,$C1,$FA,$1F,$15,$7E,$AD,$F2,$D9,$F0
	.dc.b	$9E,$F5,$35,$1B,$DE,$E1,$D4,$5F,$27,$0A,$3B,$6F,$E1,$DF,$FC,$65
	.dc.b	$5C,$5A,$B7,$6B,$FC,$A7,$05,$F9,$FD,$3E,$77,$DA,$30,$F8,$3E,$13
	.dc.b	$DD,$E6,$9A,$7B,$CF,$81,$F0,$96,$78,$1E,$87,$E2,$79,$BC,$B6,$93
	.dc.b	$EC,$8F,$91,$F0,$FA,$8C,$1F,$01,$F9,$DF,$56,$F8,$CF,$1E,$12,$BE
	.dc.b	$44,$E5,$BE,$27,$0A,$91,$83,$E7,$7D,$4B,$EA,$FE,$D7,$93,$FE,$BB
	.dc.b	$9F,$A4,$D7,$E5,$71,$7F,$51,$AE,$11,$CB,$DC,$FC,$14,$62,$36,$7F
	.dc.b	$85,$1C,$72,$7D,$28,$C4,$F0,$F9,$3F,$C5,$7D,$33,$6A,$66,$BC,$FE
	.dc.b	$95,$7A,$8C,$57,$E2,$9F,$1F,$9D,$F9,$5F,$AB,$FF,$52,$5F,$E3,$5A
	.dc.b	$F2,$BF,$EA,$CF,$C4,$D7,$8B,$F1,$AB,$DB,$1E,$EF,$4A,$CE,$CF,$96
	.dc.b	$FE,$8F,$EA,$7D,$D1,$E6,$FA,$BF,$F6,$0B,$F9,$D6,$BE,$A8,$DE,$FC
	.dc.b	$1F,$DC,$46,$78,$3A,$1D,$7C,$4E,$DE,$5F,$EC,$7E,$AF,$FA,$95,$7E
	.dc.b	$A5,$AF,$58,$F2,$7A,$BC,$23,$F5,$D8,$7F,$32,$6C,$7A,$7C,$AF,$DE
	.dc.b	$7D,$5B,$DC,$A6,$6B,$8F,$C3,$D7,$D5,$5F,$06,$3E,$17,$AF,$67,$67
	.dc.b	$81,$F8,$DF,$29,$F5,$7D,$C9,$3E,$48,$F9,$5F,$0D,$A8,$C1,$E8,$7E
	.dc.b	$77,$C4,$7F,$11,$E3,$7A,$74,$CD,$7B,$BE,$1E,$BF,$3E,$BE,$B4,$7F
	.dc.b	$19,$F6,$3F,$99,$F3,$7D,$8E,$FF,$E0,$B7,$FC,$4D,$AF,$E4,$27,$7C
	.dc.b	$FF,$E3,$37,$F8,$B4,$67,$D5,$E8,$75,$F2,$3B,$79,$F6,$7F,$8F,$57
	.dc.b	$0C,$7F,$0E,$FF,$1D,$8C,$FC,$FF,$E9,$3E,$AE,$46,$0E,$3F,$41,$E9
	.dc.b	$99,$18,$3A,$F9,$BF,$BC,$F8,$29,$18,$3C,$3E,$3F,$C1,$99,$18,$3C
	.dc.b	$BF,$0B,$FC,$F7,$E2,$5D,$54,$F9,$11,$F2,$7B,$2D,$70,$3D,$AF,$CD
	.dc.b	$F5,$75,$18,$3E,$0B,$FD,$27,$EA,$DE,$05,$F0,$A5,$7F,$43,$9C,$B7
	.dc.b	$D9,$E1,$ED,$48,$C1,$FD,$27,$EA,$DF,$19,$FD,$55,$71,$D3,$11,$9E
	.dc.b	$0D,$7C,$BC,$BE,$51,$FD,$77,$EA,$FF,$D5,$91,$FD,$55,$FE,$5C,$FF
	.dc.b	$12,$6F,$47,$F0,$07,$71,$9F,$89,$E5,$E8,$C8,$E5,$1F,$DA,$7D,$BF
	.dc.b	$EA,$F7,$FC,$21,$FE,$02,$DF,$37,$5F,$0B,$BF,$23,$87,$C3,$4C,$7F
	.dc.b	$B4,$60,$9F,$EE,$9F,$57,$F9,$8C,$7F,$89,$BF,$8D,$AF,$E5,$2D,$8C
	.dc.b	$D7,$0E,$0E,$3D,$D2,$FF,$BA,$5F,$CA,$6F,$B9,$DB,$DD,$91,$83,$FB
	.dc.b	$D7,$D5,$BC,$1F,$F0,$33,$FD,$2F,$3F,$E4,$AD,$F1,$FF,$D3,$8E,$EB
	.dc.b	$FC,$05,$F5,$7F,$E5,$F6,$FC,$97,$F8,$3E,$0D,$E0,$FF,$09,$76,$FA
	.dc.b	$F0,$FF,$53,$3F,$AD,$6F,$91,$AF,$23,$7E,$1F,$0F,$AB,$99,$FF,$84
	.dc.b	$70,$4F,$4D,$FA,$F8,$72,$B5,$1F,$E2,$1F,$AB,$F5,$47,$FB,$C9,$7B
	.dc.b	$D9,$35,$F3,$3F,$E2,$7F,$AB,$FE,$7F,$B7,$FA,$EE,$BF,$9B,$F1,$FF
	.dc.b	$48,$6C,$7F,$AF,$E7,$78,$7F,$BF,$9F,$82,$FF,$89,$DF,$F5,$03,$7D
	.dc.b	$6E,$DC,$5D,$FA,$FC,$1C,$72,$B8,$47,$8F,$D5,$D4,$60,$E8,$F9,$FC
	.dc.b	$7F,$47,$8F,$27,$C1,$1E,$BF,$4F,$51,$83,$C8,$FF,$91,$7E,$A9,$F0
	.dc.b	$FF,$E4,$7F,$AB,$7B,$D4,$D4,$6F,$C1,$B8,$7F,$90,$EF,$8B,$63,$FC
	.dc.b	$A1,$F1,$5F,$B0,$E2,$FF,$72,$7A,$7F,$E5,$8A,$FF,$02,$B7,$97,$FB
	.dc.b	$9D,$6E,$47,$F9,$53,$E1,$F9,$FC,$3C,$8F,$F2,$94,$FE,$EA,$BF,$BD
	.dc.b	$FD,$5F,$E2,$E7,$8E,$1F,$41,$EF,$CD,$FE,$63,$9D,$41,$FC,$A3,$E0
	.dc.b	$72,$4F,$03,$87,$91,$C5,$93,$FE,$63,$9F,$DD,$69,$FF,$00,$5F,$87
	.dc.b	$1E,$EE,$6F,$F3,$96,$3F,$C0,$63,$87,$07,$8F,$7F,$23,$23,$A4,$79
	.dc.b	$FC,$8F,$F3,$34,$7F,$A4,$F0,$FE,$4F,$AF,$FA,$56,$3D,$53,$A7,$FD
	.dc.b	$6B,$3A,$78,$7A,$0F,$AB,$E1,$3F,$CC,$9B,$19,$9D,$C7,$FA,$9D,$7D
	.dc.b	$2D,$23,$FE,$AB,$7F,$D5,$9F,$43,$FE,$49,$C7,$FC,$5B,$28,$C6,$DF
	.dc.b	$EA,$BA,$FB,$BF,$57,$FE,$90,$AF,$BE,$6E,$7F,$D1,$CF,$F9,$AD,$BC
	.dc.b	$9C,$A3,$77,$5F,$C3,$C5,$5C,$B4,$D7,$03,$E4,$F1,$87,$6F,$9A,$E0
	.dc.b	$62,$FF,$CB,$21,$E9,$F3,$52,$78,$1E,$67,$C5,$AF,$FA,$62,$3F,$D4
	.dc.b	$CD,$AF,$CF,$F8,$3A,$E7,$DF,$2C,$BD,$C5,$AF,$F3,$4F,$D5,$FC,$36
	.dc.b	$EF,$EA,$F0,$3E,$04,$E7,$2F,$F2,$DB,$F1,$EB,$8F,$06,$AF,$6F,$0F
	.dc.b	$16,$65,$FE,$BB,$BF,$E4,$6D,$8E,$BE,$7C,$8C,$1B,$7C,$3E,$AC,$8E
	.dc.b	$5F,$1F,$72,$F4,$FA,$DE,$63,$1A,$79,$9F,$4F,$C6,$CB,$F1,$7C,$DD
	.dc.b	$BF,$F4,$4F,$CC,$7D,$79,$FF,$68,$36,$33,$EF,$FE,$06,$31,$8F,$F9
	.dc.b	$F1,$79,$19,$35,$E4,$DC,$3F,$DB,$5F,$56,$F9,$EF,$C5,$36,$33,$E0
	.dc.b	$E2,$73,$DB,$27,$8F,$BD,$75,$FF,$4F,$FD,$5F,$E1,$E7,$FD,$CC,$D8
	.dc.b	$CF,$81,$F9,$78,$C6,$DF,$E9,$25,$C1,$93,$5C,$57,$0F,$F7,$87,$D5
	.dc.b	$B7,$FD,$03,$63,$3D,$78,$9F,$D2,$47,$F8,$B8,$FF,$7C,$17,$3F,$F5
	.dc.b	$AB,$FE,$30,$33,$BF,$EA,$FE,$AF,$F9,$8C,$7F,$BF,$DB,$3C,$3D,$DD
	.dc.b	$3B,$DA,$56,$FB,$2C,$FF,$B0,$AD,$A7,$53,$B7,$4D,$BF,$E7,$FC,$1F
	.dc.b	$EB,$4F,$15,$EC,$FE,$F5,$B3,$A3,$F8,$3A,$F9,$ED,$4E,$FC,$C9,$5B
	.dc.b	$CF,$FA,$5D,$BD,$4C,$C7,$FB,$5B,$EA,$FF,$6D,$97,$FB,$11,$B7,$FC
	.dc.b	$51,$79,$FF,$C3,$DC,$D7,$F5,$67,$FD,$65,$3D,$E1,$18,$7B,$7B,$7C
	.dc.b	$7E,$1F,$97,$1C,$BE,$16,$F6,$F1,$87,$FB,$AB,$C7,$FF,$1B,$57,$56
	.dc.b	$C6,$7C,$6F,$E4,$51,$8D,$3F,$DB,$BF,$56,$F8,$4F,$E4,$CD,$C3,$CC
	.dc.b	$9C,$C7,$F2,$BF,$AB,$7C,$3D,$3F,$97,$7D,$5E,$EB,$DD,$3A,$6F,$85
	.dc.b	$C4,$E0,$E0,$DF,$87,$D4,$D3,$83,$9F,$17,$84,$EF,$F0,$F6,$3B,$F1
	.dc.b	$F9,$4C,$70,$36,$FF,$7D,$E5,$F3,$DF,$F7,$EF,$D5,$FE,$61,$B3,$EA
	.dc.b	$EA,$FE,$91,$39,$C3,$FA,$8B,$FC,$AD,$BC,$3A,$AD,$BC,$3D,$39,$3F
	.dc.b	$7F,$7F,$CA,$D7,$8E,$39,$CF,$6F,$4E,$47,$E1,$70,$FE,$02,$F8,$23
	.dc.b	$C7,$ED,$C8,$D4,$5B,$E8,$7F,$89,$2B,$F9,$BB,$80,$E3,$3F,$B9,$1F
	.dc.b	$E2,$AC,$7F,$DE,$37,$E1,$F8,$2F,$83,$CD,$23,$07,$95,$F8,$BD,$F2
	.dc.b	$3D,$DF,$19,$70,$F0,$65,$FB,$C9,$F6,$78,$39,$CF,$E8,$BF,$3B,$E6
	.dc.b	$A4,$FC,$7F,$7A,$E1,$CB,$E0,$3E,$99,$8C,$3E,$CB,$F5,$BE,$96,$5F
	.dc.b	$C9,$8B,$5F,$07,$EA,$BF,$38,$DE,$DF,$F0,$27,$D5,$BE,$5B,$FC,$0D
	.dc.b	$5A,$CB,$E2,$DA,$1D,$7A,$37,$85,$CB,$F7,$93,$FE,$E2,$6F,$AD,$FE
	.dc.b	$11,$FA,$B7,$D5,$F6,$8E,$FF,$01,$2B,$67,$27,$C7,$C9,$23,$F1,$5E
	.dc.b	$C5,$D3,$D9,$C5,$E8,$31,$B7,$A3,$44,$7B,$3E,$02,$61,$E6,$8B,$7F
	.dc.b	$17,$E2,$3E,$09,$E7,$EA,$B5,$EC,$FA,$BF,$8A,$62,$BE,$63,$F2,$38
	.dc.b	$E4,$FF,$1D,$F5,$BF,$D6,$2F,$C4,$8F,$F1,$97,$23,$DC,$A7,$0C,$3F
	.dc.b	$2D,$19,$AF,$E4,$DF,$57,$FC,$AA,$4D,$F9,$6F,$EF,$73,$FE,$44,$8D
	.dc.b	$1F,$E9,$67,$8F,$C2,$7F,$90,$63,$2D,$F6,$64,$FF,$33,$A6,$98,$C3
	.dc.b	$F5,$18,$7F,$7F,$6F,$BD,$9E,$3F,$C2,$CE,$5B,$E5,$F7,$3F,$CA,$1F
	.dc.b	$4D,$EB,$BE,$71,$E3,$F3,$26,$7E,$E8,$B3,$FB,$7F,$6F,$FC,$6D,$87
	.dc.b	$E4,$9B,$5F,$E6,$2C,$3F,$BB,$4E,$7D,$BF,$F2,$F5,$7F,$97,$2B,$FA
	.dc.b	$73,$6F,$A4,$EA,$37,$E5,$CA,$FE,$A9,$87,$F4,$D7,$8C,$74,$F2,$A4
	.dc.b	$60,$EA,$F8,$3E,$0E,$47,$F5,$DF,$A8,$F0,$67,$E9,$9F,$2F,$CF,$CA
	.dc.b	$F2,$C5,$D3,$F8,$B5,$FF,$07,$6F,$93,$FE,$79,$87,$97,$62,$5F,$F6
	.dc.b	$BF,$91,$FE,$FC,$F8,$63,$FD,$01,$18,$BF,$E4,$4D,$7F,$D1,$61,$C3
	.dc.b	$FE,$8E,$6B,$FE,$CE,$6B,$BF,$7B,$FD,$32,$E5,$AE,$6B,$FD,$61,$3F
	.dc.b	$D0,$23,$BB,$FC,$15,$F0,$9E,$1F,$17,$8C,$66,$37,$18,$FA,$09,$5D
	.dc.b	$1B,$3C,$A9,$E0,$72,$25,$78,$9F,$2F,$13,$E1,$A9,$27,$FD,$8D,$19
	.dc.b	$6F,$9D,$FD,$19,$F8,$F0,$8C,$61,$FE,$B4,$8C,$4F,$F9,$0A,$3F,$DE
	.dc.b	$B1,$87,$FD,$7E,$BC,$DE,$31,$A6,$BF,$EB,$46,$C6,$E7,$A4,$FF,$AE
	.dc.b	$3C,$EC,$92,$33,$28,$E3,$F5,$72,$72,$DA,$FF,$68,$FD,$57,$F8,$5D
	.dc.b	$B3,$E2,$8C,$3F,$86,$6C,$7F,$B4,$9A,$83,$86,$C7,$33,$DB,$A7,$F9
	.dc.b	$35,$F5,$B6,$3F,$DC,$6D,$8F,$C3,$FC,$07,$0F,$1A,$47,$06,$D7,$AB
	.dc.b	$D7,$FE,$58,$FE,$33,$0F,$F3,$07,$8D,$FF,$00,$57,$CF,$7F,$BB,$E7
	.dc.b	$55,$C1,$F7,$72,$7F,$BB,$B1,$B1,$F4,$2F,$FC,$DE,$D9,$D3,$F0,$F9
	.dc.b	$2B,$F0,$E5,$18,$8F,$8A,$D8,$CC,$75,$F3,$F0,$47,$07,$87,$E8,$78
	.dc.b	$11,$C1,$E5,$E6,$E8,$47,$07,$E2,$3E,$AD,$E4,$F8,$DB,$19,$8C,$19
	.dc.b	$F8,$3F,$DA,$C7,$DD,$9F,$F3,$EB,$86,$BF,$BA,$6B,$FC,$31,$7C,$CF
	.dc.b	$28,$46,$1E,$51,$98,$C4,$73,$F9,$4C,$11,$F0,$3C,$F3,$71,$7E,$46
	.dc.b	$7F,$8E,$FA,$F7,$C2,$FF,$0C,$F3,$7F,$9F,$67,$FD,$21,$7C,$1B,$F1
	.dc.b	$3B,$E3,$FC,$E6,$36,$7E,$D3,$5F,$CE,$CD,$FA,$BF,$3E,$3F,$E2,$6F
	.dc.b	$09,$EF,$6D,$B7,$FE,$9F,$AD,$E3,$E7,$7E,$4D,$FF,$8B,$9C,$CF,$EA
	.dc.b	$DA,$9D,$CF,$1F,$91,$F2,$69,$FE,$34,$AF,$E8,$5E,$97,$96,$7C,$DC
	.dc.b	$76,$7C,$51,$87,$C9,$F0,$7E,$DF,$2C,$D3,$DC,$7B,$FE,$0E,$CF,$03
	.dc.b	$D8,$FC,$3F,$86,$E5,$CD,$27,$E0,$47,$C7,$F8,$8D,$46,$0F,$88,$FC
	.dc.b	$DF,$AB,$7A,$1E,$3C,$25,$7F,$B1,$E7,$2D,$F1,$B8,$54,$8C,$1F,$EC
	.dc.b	$DF,$A9,$7D,$5F,$D9,$F2,$9F,$0B,$4F,$B9,$3A,$AE,$1E,$E6,$2E,$46
	.dc.b	$0E,$9F,$27,$F8,$3F,$A6,$6B,$4E,$71,$D3,$D2,$B5,$C0,$F2,$3E,$0F
	.dc.b	$43,$F1,$9F,$56,$F5,$A9,$9A,$F5,$6D,$5F,$F4,$D6,$BF,$D6,$5C,$9E
	.dc.b	$FF,$4B,$D7,$32,$30,$7F,$BB,$FE,$AB,$FC,$F7,$DF,$FE,$7C,$FA,$97
	.dc.b	$9F,$F9,$D6,$D6,$FC,$0F,$D0,$5F,$EA,$63,$FD,$54,$F2,$8F,$D8,$36
	.dc.b	$3F,$A1,$3B,$BF,$F4,$4B,$FC,$A2,$3F,$D6,$5F,$56,$F5,$3F,$58,$D8
	.dc.b	$F3,$7A,$3E,$0F,$47,$C2,$C8,$F5,$47,$93,$CF,$FD,$A7,$D5,$BD,$DA
	.dc.b	$66,$BE,$07,$C4,$57,$D3,$5E,$C8,$F6,$FC,$2D,$9D,$9F,$1D,$F8,$3F
	.dc.b	$29,$F5,$BD,$E9,$3F,$2E,$3E,$2F,$B9,$A8,$C1,$F3,$DF,$93,$F1,$3F
	.dc.b	$37,$D8,$93,$F4,$E3,$E6,$FC,$3E,$A3,$07,$D7,$7F,$88,$7B,$6F,$AB
	.dc.b	$FC,$87,$9D,$2B,$FB,$87,$37,$FA,$DA,$B5,$18,$E7,$FE,$49,$1F,$C2
	.dc.b	$8C,$C7,$FB,$8D,$B3,$C7,$CC,$EF,$ED,$F8,$19,$5C,$CF,$EF,$D5,$FC
	.dc.b	$49,$BE,$D7,$E8,$FE,$AE,$46,$CF,$ED,$F1,$EA,$FA,$0F,$5B,$23,$54
	.dc.b	$9F,$7F,$A3,$FB,$7F,$83,$95,$A8,$F8,$7E,$4E,$A3,$07,$27,$E3,$FC
	.dc.b	$2F,$F3,$7F,$89,$74,$D3,$35,$E0,$EB,$AF,$C3,$AF,$9F,$1F,$43,$D8
	.dc.b	$B3,$B3,$CE,$FF,$43,$FA,$BF,$EF,$30,$E2,$FF,$73,$C7,$E7,$DB,$E0
	.dc.b	$6F,$C0,$E1,$EE,$F1,$F7,$A6,$1F,$1F,$04,$FF,$4F,$FA,$B7,$C6,$5E
	.dc.b	$39,$7F,$36,$72,$DE,$2E,$1F,$2D,$23,$07,$D3,$FA,$B7,$B3,$FD,$6D
	.dc.b	$FE,$57,$1E,$9E,$CF,$ED,$51,$8E,$0F,$F8,$02,$F0,$9E,$5F,$15,$C7
	.dc.b	$D4,$91,$83,$F0,$1D,$FF,$D4,$DE,$87,$F9,$15,$FE,$09,$87,$F7,$76
	.dc.b	$C6,$7E,$03,$FB,$74,$3C,$6B,$F1,$32,$BA,$C7,$F7,$3F,$AB,$55,$F0
	.dc.b	$E5,$7E,$46,$72,$DE,$2E,$1E,$0C,$8C,$1B,$F7,$FA,$7C,$04,$8C,$1F
	.dc.b	$9B,$FA,$B7,$87,$FE,$03,$7F,$A5,$C6,$7D,$FC,$4F,$F8,$1F,$EA,$DB
	.dc.b	$DF,$92,$97,$FA,$59,$CB,$FA,$7F,$03,$F5,$89,$FF,$00,$93,$7F,$F0
	.dc.b	$92,$F0,$EF,$D5,$CA,$E5,$1A,$D7,$7E,$1C,$9E,$D1,$FE,$1B,$FA,$BF
	.dc.b	$4C,$7D,$60,$BE,$0F,$46,$F7,$FF,$11,$7D,$5F,$F3,$FA,$DF,$B1,$CC
	.dc.b	$CF,$56,$B8,$37,$EA,$CA,$FE,$09,$69,$E7,$D5,$D3,$8E,$56,$A3,$BF
	.dc.b	$A9,$A8,$C1,$C9,$F1,$F8,$DE,$4F,$22,$4F,$68,$F3,$FA,$7A,$8C,$1E
	.dc.b	$27,$FC,$79,$F5,$5F,$E6,$28,$F8,$5F,$57,$FE,$B5,$7F,$8D,$35,$F8
	.dc.b	$D1,$BF,$15,$BF,$F9,$46,$1D,$DD,$3E,$2A,$5F,$F2,$39,$CB,$FC,$97
	.dc.b	$E3,$7F,$5D,$7F,$E5,$9F,$4B,$F8,$0C,$FF,$29,$6B,$F5,$67,$6F,$FA
	.dc.b	$BD,$FD,$A3,$9F,$89,$FF,$2B,$46,$A3,$FC,$A7,$E5,$7F,$96,$7E,$AF
	.dc.b	$FC,$97,$6E,$6F,$F2,$E3,$6B,$F9,$C4,$FF,$94,$0F,$C9,$97,$D7,$E5
	.dc.b	$E0,$F3,$F2,$2D,$70,$8C,$3C,$B6,$EB,$CD,$23,$07,$6E,$0F,$24,$C8
	.dc.b	$C1,$E2,$F1,$FF,$96,$3F,$E7,$D3,$FD,$9D,$97,$F4,$7E,$2F,$F4,$AB
	.dc.b	$63,$41,$C3,$C5,$FE,$57,$F5,$7C,$3D,$13,$49,$3F,$E9,$7F,$A4,$FF
	.dc.b	$4C,$26,$91,$FF,$52,$FD,$0F,$E0,$9F,$D5,$07,$18,$66,$3F,$D2,$D3
	.dc.b	$FE,$E7,$FA,$BF,$F4,$2C,$7F,$BA,$6B,$7C,$DE,$FE,$6B,$3F,$E9,$D6
	.dc.b	$BB,$8F,$ED,$CF,$F9,$A1,$BC,$BA,$8F,$E7,$F1,$F8,$BF,$96,$C4,$E7
	.dc.b	$E6,$64,$F0,$39,$BC,$7E,$2E,$46,$0E,$EF,$8B,$4F,$E6,$31,$FD,$FD
	.dc.b	$B8,$7F,$0B,$E7,$FD,$F4,$E7,$9B,$87,$1C,$8C,$1C,$5F,$6F,$D5,$FC
	.dc.b	$98,$FF,$05,$36,$FA,$77,$EB,$7D,$EF,$2F,$71,$6B,$F8,$E5,$FF,$A0
	.dc.b	$27,$9F,$8B,$DB,$C6,$EE,$F8,$3C,$E9,$3A,$3C,$2F,$93,$D3,$91,$E4
	.dc.b	$F8,$FB,$8F,$6F,$57,$D2,$63,$87,$F0,$19,$A3,$DB,$F1,$F3,$0F,$6F
	.dc.b	$D4,$DC,$3F,$99,$7C,$D3,$A3,$EF,$B6,$BF,$C5,$8B,$DE,$CC,$EB,$C8
	.dc.b	$95,$AF,$26,$E1,$FC,$F7,$EA,$DF,$3D,$F8,$A6,$D7,$F8,$DD,$FF,$07
	.dc.b	$B6,$35,$D5,$23,$5E,$F5,$BF,$E9,$1F,$56,$EE,$FC,$AB,$6B,$FC,$84
	.dc.b	$FE,$4A,$77,$3A,$DE,$56,$B8,$6D,$7F,$53,$FA,$BF,$F1,$1C,$FF,$92
	.dc.b	$DB,$19,$EC,$FD,$1C,$63,$97,$F5,$D5,$EA,$AF,$EB,$8D,$8C,$F8,$DF
	.dc.b	$AA,$8C,$09,$9F,$57,$63,$FC,$B4,$D8,$E7,$EE,$FE,$CA,$31,$7F,$BD
	.dc.b	$7D,$CD,$F0,$F5,$E0,$6D,$E1,$A4,$C7,$AF,$65,$CB,$FB,$97,$8B,$FE
	.dc.b	$01,$9F,$F3,$53,$63,$3F,$25,$FC,$0A,$31,$97,$F0,$D7,$FC,$51,$5C
	.dc.b	$3C,$BD,$7A,$7B,$7F,$84,$FD,$5F,$ED,$AF,$F8,$A3,$6B,$E1,$62,$87
	.dc.b	$9D,$C3,$C8,$E3,$F0,$D2,$FF,$CF,$D5,$FE,$13,$6C,$76,$E2,$91,$AF
	.dc.b	$6A,$E3,$FE,$0F,$F2,$7F,$C5,$73,$FE,$87,$6C,$67,$C8,$FE,$3F,$18
	.dc.b	$D3,$F9,$57,$D5,$BE,$83,$F9,$1B,$70,$FF,$48,$3F,$DD,$1B,$1F,$CA
	.dc.b	$3E,$AD,$F0,$F4,$FE,$57,$F5,$7B,$CF,$FC,$45,$3A,$BF,$F4,$D2,$F0
	.dc.b	$79,$1C,$63,$2F,$5F,$0A,$46,$0E,$2F,$87,$E0,$64,$7F,$A9,$70,$7C
	.dc.b	$3E,$31,$F0,$F1,$79,$CC,$6D,$FD,$27,$3F,$F6,$F6,$1F,$EA,$DF,$AB
	.dc.b	$FC,$8B,$F6,$1B,$7F,$9B,$C7,$FD,$FB,$3B,$8F,$E9,$33,$FC,$AE,$FF
	.dc.b	$AA,$36,$79,$4E,$B8,$37,$E2,$49,$F4,$7C,$2D,$AF,$F2,$45,$BD,$4E
	.dc.b	$A6,$71,$EA,$F8,$3D,$C9,$87,$A6,$2D,$F2,$F5,$3C,$A7,$D4,$F9,$2B
	.dc.b	$5E,$FE,$7F,$49,$8A,$F0,$BE,$CF,$06,$4F,$CD,$F8,$DB,$5E,$6F,$0F
	.dc.b	$E0,$98,$D3,$E0,$BF,$13,$E6,$65,$FD,$5F,$7E,$DF,$BB,$E6,$BE,$49
	.dc.b	$8C,$3E,$4B,$F3,$3E,$96,$5F,$D5,$8B,$5F,$E6,$DF,$A8,$FC,$B3,$7B
	.dc.b	$BF,$BD,$FD,$5B,$E5,$FF,$BE,$D6,$B2,$FF,$04,$3F,$86,$C7,$FD,$D8
	.dc.b	$DC,$39,$46,$3B,$38,$FB,$12,$3F,$DE,$5F,$56,$FA,$BC,$5F,$8F,$DA
	.dc.b	$EE,$63,$2E,$EF,$8B,$8E,$61,$F9,$7F,$6A,$DF,$AF,$8B,$CE,$63,$0F
	.dc.b	$3B,$EA,$F8,$A9,$7E,$88,$B5,$F0,$FE,$2B,$DE,$7F,$87,$34,$FF,$48
	.dc.b	$57,$BA,$3E,$1F,$D5,$C8,$C1,$F0,$9F,$8F,$C3,$23,$F9,$D6,$5F,$E2
	.dc.b	$FF,$63,$F8,$A6,$3F,$46,$BF,$C5,$3C,$2F,$7A,$9C,$67,$FD,$45,$19
	.dc.b	$C3,$99,$FC,$93,$EA,$FF,$93,$4F,$F4,$76,$AF,$98,$FE,$EC,$FF,$AA
	.dc.b	$E3,$23,$E1,$7F,$C7,$2D,$AE,$B3,$CF,$D2,$CD,$6B,$E2,$25,$6A,$BB
	.dc.b	$6D,$87,$B7,$C0,$74,$31,$BF,$F9,$3F,$2F,$EF,$B3,$E3,$F3,$7C,$27
	.dc.b	$F9,$2B,$E9,$FD,$D1,$9C,$7B,$57,$F0,$E8,$F5,$79,$DC,$27,$2F,$F9
	.dc.b	$4B,$DB,$FF,$28,$4F,$FB,$25,$B8,$FF,$97,$F0,$FE,$DB,$3F,$DB,$67
	.dc.b	$37,$FD,$39,$FE,$CE,$D7,$A7,$B9,$D5,$ED,$E6,$C9,$F8,$56,$F5,$71
	.dc.b	$87,$87,$CB,$F2,$18,$CB,$FC,$E1,$8F,$F1,$5C,$3E,$53,$FD,$2D,$F1
	.dc.b	$7C,$3F,$F9,$B3,$EA,$7C,$78,$79,$E3,$DD,$F3,$D2,$B8,$C5,$9F,$F7
	.dc.b	$45,$3E,$0F,$FC,$FD,$1F,$DA,$A3,$15,$FE,$94,$84,$61,$FF,$4B,$B5
	.dc.b	$FF,$4F,$60,$F8,$9E,$71,$F2,$5A,$A3,$55,$D5,$B1,$FE,$A2,$5E,$6F
	.dc.b	$D7,$3E,$C5,$5A,$FC,$65,$BE,$4B,$AB,$63,$D0,$DC,$3C,$F3,$E2,$FA
	.dc.b	$D7,$CF,$52,$5F,$BE,$32,$F9,$73,$F4,$E7,$EF,$C3,$F0,$51,$83,$3F
	.dc.b	$17,$2B,$D8,$D9,$FF,$5E,$FC,$B3,$C7,$FF,$58,$57,$FB,$16,$30,$FF
	.dc.b	$8D,$E7,$FC,$79,$18,$7E,$AB,$63,$87,$8D,$CA,$7F,$C4,$33,$FE,$E5
	.dc.b	$9E,$50,$8C,$3C,$A7,$83,$F9,$DF,$42,$BF,$55,$B9,$1A,$6B,$37,$CB
	.dc.b	$D2,$63,$E0,$8C,$BE,$2C,$7F,$C2,$CF,$BC,$EE,$E3,$0F,$4C,$78,$BE
	.dc.b	$AF,$FA,$7B,$FE,$E9,$6C,$7F,$8F,$9A,$FF,$98,$5A,$9A,$4F,$47,$A1
	.dc.b	$D7,$C5,$95,$F1,$A7,$FA,$A3,$DE,$3C,$5F,$11,$FC,$BE,$FF,$DE,$B4
	.dc.b	$8C,$47,$97,$E7,$FF,$DD,$0E,$A3,$2F,$D4,$F2,$7F,$CA,$B8,$7D,$F6
	.dc.b	$C6,$4F,$8C,$E1,$FF,$3E,$7D,$67,$69,$FF,$2E,$C2,$30,$FF,$98,$1A
	.dc.b	$F0,$F2,$B9,$3F,$0F,$D0,$B9,$7F,$AD,$63,$E0,$46,$0D,$D3,$E3,$29
	.dc.b	$27,$F1,$71,$96,$F9,$3C,$08,$E0,$F9,$9E,$8F,$80,$8E,$0F,$A1,$CB
	.dc.b	$E4,$23,$83,$BF,$93,$FC,$45,$F8,$0D,$BF,$F6,$1C,$7F,$15,$6B,$F1
	.dc.b	$87,$1F,$F2,$F7,$F7,$9E,$1F,$38,$C4,$6D,$AF,$E3,$FC,$FF,$06,$A6
	.dc.b	$92,$3F,$92,$43,$CF,$FF,$48,$F8,$8E,$1F,$D9,$F3,$61,$B7,$FD,$8E
	.dc.b	$F8,$37,$E4,$34,$FE,$6F,$0A,$FE,$6D,$B2,$9C,$FD,$63,$E5,$F7,$E1
	.dc.b	$BD,$FA,$6A,$37,$F1,$77,$2F,$EE,$97,$C5,$B1,$D3,$D3,$F0,$F7,$7F
	.dc.b	$9F,$A7,$FB,$D3,$56,$FE,$F7,$FD,$A5,$6A,$F2,$FA,$7C,$E7,$BC,$61
	.dc.b	$F0,$FC,$3F,$BB,$CE,$FF,$4B,$47,$F6,$68,$F6,$9E,$EF,$87,$B1,$C0
	.dc.b	$F4,$BF,$0B,$9B,$96,$B2,$7D,$B1,$F1,$BE,$2B,$51,$83,$E0,$BF,$3B
	.dc.b	$EB,$1E,$9F,$8F,$09,$5F,$E1,$99,$CB,$7C,$8E,$15,$23,$07,$F8,$7F
	.dc.b	$EA,$5F,$59,$F6,$BE,$09,$F0,$94,$FB,$B3,$AA,$E1,$F0,$D8,$B9,$18
	.dc.b	$3A,$7A,$7F,$84,$FA,$C6,$B4,$E7,$1D,$3D,$5B,$5C,$0F,$23,$E0,$F4
	.dc.b	$BF,$1B,$F5,$9F,$EA,$19,$FF,$1B,$35,$FB,$BE,$0F,$FB,$B3,$45,$7A
	.dc.b	$36,$E5,$5E,$F8,$C3,$F0,$7E,$1A,$BC,$80,$A5,$7C,$29,$FC,$FF,$D5
	.dc.b	$FE,$2E,$B8,$0F,$AC,$FE,$60,$FF,$92,$DA,$EA,$38,$76,$F1,$7F,$DD
	.dc.b	$D1,$FE,$51,$73,$97,$8A,$11,$87,$CA,$F0,$7A,$7C,$67,$EC,$BE,$B3
	.dc.b	$FC,$1D,$1F,$E5,$B6,$BA,$3B,$7C,$47,$AA,$64,$60,$F1,$7D,$5F,$EF
	.dc.b	$3E,$B1,$F0,$74,$F8,$31,$F0,$BD,$1B,$5C,$0F,$53,$F1,$BD,$1D,$46
	.dc.b	$0F,$73,$F2,$BE,$5B,$EE,$FC,$14,$9F,$85,$1F,$3B,$D0,$D4,$60,$F8
	.dc.b	$CF,$D2,$F7,$BF,$88,$F1,$3E,$3E,$99,$AF,$99,$ED,$57,$E9,$57,$DA
	.dc.b	$8F,$E3,$3E,$E3,$EB,$3F,$31,$E5,$CB,$FF,$42,$4E,$5B,$E4,$F0,$A9
	.dc.b	$18,$37,$E7,$F4,$99,$18,$39,$FD,$17,$E9,$7E,$0E,$46,$0E,$FE,$E7
	.dc.b	$92,$64,$60,$F1,$F9,$DF,$B7,$F8,$59,$18,$3C,$FE,$37,$B2,$64,$60
	.dc.b	$F5,$FC,$4F,$F3,$8F,$8B,$74,$53,$E1,$C7,$C4,$EA,$B5,$C0,$F8,$6F
	.dc.b	$C8,$F5,$75,$18,$3E,$3B,$FD,$13,$EB,$1D,$F7,$BE,$57,$FA,$BE,$72
	.dc.b	$DF,$67,$87,$B7,$23,$07,$FA,$D7,$EB,$1F,$03,$FD,$41,$FE,$FD,$1F
	.dc.b	$B2,$6F,$8F,$FB,$57,$71,$C7,$83,$87,$CC,$48,$C1,$FE,$C1,$FA,$CF
	.dc.b	$F4,$03,$7D,$5F,$EB,$B1,$FC,$B2,$3F,$7E,$E9,$BF,$2B,$AF,$46,$47
	.dc.b	$38,$FE,$CB,$DF,$FD,$4F,$3F,$E1,$0F,$EE,$5B,$F0,$3A,$F8,$8D,$F8
	.dc.b	$FC,$3E,$72,$63,$FE,$D5,$C1,$3F,$DC,$3E,$B3,$FA,$D4,$FF,$B6,$57
	.dc.b	$85,$93,$55,$BF,$03,$87,$81,$2F,$84,$73,$F7,$7A,$7B,$D2,$7A,$47
	.dc.b	$F7,$EF,$AC,$78,$77,$DD,$95,$FE,$E9,$9C,$BF,$EE,$BF,$AC,$6F,$FE
	.dc.b	$06,$5F,$2D,$4C,$47,$F8,$2B,$BF,$D3,$5F,$EF,$37,$FC,$00,$FF,$2E
	.dc.b	$6F,$27,$F3,$17,$71,$9F,$0F,$97,$C7,$48,$E5,$1D,$76,$ED,$E1,$C9
	.dc.b	$ED,$1F,$E1,$AF,$AC,$FE,$C3,$1F,$D0,$B1,$FE,$7D,$0F,$81,$D1,$BD
	.dc.b	$FF,$C4,$3F,$59,$FC,$D6,$D5,$FC,$C2,$F8,$FC,$9E,$BD,$BD,$F0,$70
	.dc.b	$F5,$E6,$1D,$E7,$FA,$06,$3F,$1A,$30,$7F,$54,$7F,$02,$E0,$3F,$39
	.dc.b	$C3,$5F,$EA,$AF,$06,$FA,$DE,$8E,$C9,$1C,$23,$D7,$E4,$7B,$3C,$89
	.dc.b	$3D,$23,$DF,$EA,$6A,$30,$78,$1F,$F2,$1F,$D5,$BE,$23,$FC,$8B,$F5
	.dc.b	$8F,$7E,$9A,$8D,$F8,$17,$0F,$65,$F1,$6C,$7F,$93,$BE,$2F,$F6,$5C
	.dc.b	$7F,$DC,$7D,$4F,$F2,$75,$7C,$36,$BF,$DD,$27,$7C,$3F,$DA,$DF,$DC
	.dc.b	$57,$F7,$97,$33,$FE,$00,$6B,$FE,$55,$F8,$8F,$F2,$EF,$D6,$7E,$42
	.dc.b	$FF,$BA,$60,$E4,$FF,$30,$B6,$FE,$A5,$6D,$FF,$05,$BF,$96,$C7,$FC
	.dc.b	$10,$E6,$0D,$35,$7F,$E0,$17,$F7,$B1,$FE,$0E,$72,$F4,$F1,$BC,$BE
	.dc.b	$37,$9B,$6F,$1F,$2C,$AF,$1C,$7A,$B6,$D4,$60,$F3,$BE,$EF,$1B,$FC
	.dc.b	$CF,$1F,$E9,$5C,$7F,$D5,$D9,$FF,$A5,$E3,$CF,$3A,$7F,$D7,$F3,$A7
	.dc.b	$87,$A2,$FA,$CE,$13,$FC,$E1,$B1,$99,$DC,$7F,$AA,$57,$D3,$52,$3F
	.dc.b	$EA,$D7,$FD,$5D,$F4,$5F,$E7,$BC,$7F,$C6,$52,$8C,$6F,$FE,$AD,$AF
	.dc.b	$F1,$DF,$D6,$7F,$A3,$EB,$FC,$59,$5A,$7F,$C5,$DE,$95,$7F,$C8,$2D
	.dc.b	$7A,$5F,$EB,$DF,$F3,$8B,$79,$34,$71,$E3,$7C,$B6,$27,$87,$CE,$73
	.dc.b	$31,$7F,$E4,$80,$F7,$F9,$C9,$3C,$0F,$03,$E5,$D7,$FD,$31,$1F,$C0
	.dc.b	$5B,$5F,$E5,$2C,$3F,$96,$CE,$79,$78,$72,$48,$C1,$C5,$F8,$1F,$59
	.dc.b	$EE,$BF,$F2,$DB,$70,$FD,$77,$7E,$B7,$DE,$F3,$1F,$E2,$57,$B7,$FC
	.dc.b	$B0,$FC,$6C,$39,$F8,$BD,$8E,$FE,$35,$AE,$3E,$7F,$88,$C5,$78,$9F
	.dc.b	$2F,$AB,$27,$BF,$C7,$DA,$F3,$FA,$DE,$A3,$1B,$7F,$9D,$73,$47,$BB
	.dc.b	$E3,$A6,$1E,$8F,$9D,$B8,$7F,$9E,$BE,$6F,$EC,$CF,$F2,$D6,$C6,$7D
	.dc.b	$FF,$C3,$46,$32,$FF,$41,$2F,$23,$26,$BC,$9B,$87,$F3,$7F,$AC,$7C
	.dc.b	$E7,$E3,$DB,$19,$F0,$71,$39,$EA,$93,$C7,$DF,$BA,$FF,$A4,$3E,$B3
	.dc.b	$F1,$33,$FD,$0D,$B1,$9F,$03,$F3,$91,$8D,$BF,$D3,$0B,$83,$26,$B8
	.dc.b	$AE,$1F,$D3,$7E,$B3,$FD,$F4,$DD,$FF,$4D,$0E,$CB,$1F,$A9,$9F,$F1
	.dc.b	$A9,$FD,$54,$B9,$FF,$A9,$97,$8D,$BE,$0F,$D8,$7D,$67,$FA,$C2,$7F
	.dc.b	$AE,$36,$B8,$7B,$BA,$77,$BC,$AD,$F7,$59,$FF,$59,$DF,$E0,$27,$9E
	.dc.b	$BD,$BA,$FB,$BF,$BF,$F1,$9E,$D5,$31,$F4,$16,$FF,$B7,$4F,$F8,$C1
	.dc.b	$BE,$67,$1F,$73,$93,$FC,$27,$EB,$3F,$67,$CF,$FE,$C5,$6D,$7F,$AE
	.dc.b	$DF,$F1,$AB,$FD,$E5,$CE,$BF,$ED,$07,$FD,$5F,$19,$F8,$7E,$9E,$EF
	.dc.b	$57,$B7,$14,$9E,$5E,$E5,$C3,$FD,$A7,$E4,$7F,$8D,$A7,$FC,$0C,$D8
	.dc.b	$CF,$91,$FC,$7E,$31,$9F,$FB,$6F,$EB,$1F,$09,$FC,$8D,$B8,$7F,$83
	.dc.b	$D3,$98,$FE,$51,$F5,$8F,$87,$A7,$F2,$BF,$AC,$DD,$7F,$85,$E7,$4D
	.dc.b	$F0,$B8,$9C,$1C,$1B,$F0,$FA,$9A,$70,$73,$E1,$F0,$9D,$FE,$1E,$C7
	.dc.b	$7E,$2F,$29,$8E,$0F,$F7,$BE,$3F,$39,$FF,$7C,$FD,$67,$E3,$F7,$FF
	.dc.b	$7E,$36,$FF,$DE,$5A,$3A,$3F,$A3,$4E,$6C,$1F,$E4,$4E,$1A,$E4,$E3
	.dc.b	$C3,$CB,$D3,$98,$7F,$8E,$F0,$FC,$9B,$FE,$3D,$BF,$DC,$3D,$AF,$B7
	.dc.b	$A7,$E2,$38,$BE,$A7,$CB,$ED,$C9,$E5,$16,$7D,$FE,$97,$A8,$FA,$DF
	.dc.b	$27,$6B,$C1,$CF,$EE,$31,$5F,$19,$F8,$3E,$04,$9F,$47,$C6,$DA,$F9
	.dc.b	$7E,$0F,$C6,$31,$C3,$F3,$1F,$93,$F3,$52,$FE,$07,$BD,$6F,$E9,$FC
	.dc.b	$07,$CE,$31,$87,$D4,$7E,$8F,$D3,$4B,$F8,$D1,$69,$5F,$96,$7E,$AF
	.dc.b	$D5,$FE,$69,$CD,$7C,$B8,$C4,$EE,$5F,$58,$F9,$7F,$F0,$25,$6A,$F8
	.dc.b	$75,$E0,$CF,$81,$63,$FC,$E9,$3F,$EE,$06,$FA,$FF,$E0,$DF,$AB,$7D
	.dc.b	$67,$48,$EB,$F0,$12,$B6,$72,$7C,$3C,$72,$3F,$0D,$EC,$DD,$BD,$67
	.dc.b	$BB,$8B,$CE,$67,$83,$D9,$A7,$F9,$5E,$FC,$BF,$15,$EC,$36,$E0,$F4
	.dc.b	$FC,$4F,$C0,$3D,$1D,$56,$3D,$BF,$59,$F1,$0C,$65,$E2,$18,$23,$E4
	.dc.b	$F1,$CA,$FF,$4C,$7A,$EF,$1B,$18,$71,$E5,$7B,$BF,$E3,$36,$DF,$CC
	.dc.b	$C3,$13,$FE,$37,$FA,$CF,$E5,$36,$6F,$CB,$FF,$75,$BF,$D4,$46,$8F
	.dc.b	$F4,$A3,$FE,$A9,$6C,$FA,$5D,$7C,$27,$5F,$66,$57,$FA,$BE,$FD,$8F
	.dc.b	$18,$F1,$7B,$F2,$30,$75,$7C,$DA,$48,$FF,$5C,$7D,$3F,$B3,$2F,$04
	.dc.b	$FA,$FC,$D9,$5E,$58,$B3,$F3,$7D,$BF,$F1,$67,$37,$E5,$DB,$5F,$C1
	.dc.b	$70,$FE,$DF,$39,$F6,$FF,$CB,$94,$29,$86,$D7,$FB,$2E,$3F,$87,$DF
	.dc.b	$F1,$C6,$AB,$7A,$9D,$3C,$DE,$BE,$54,$C3,$94,$5B,$F4,$7C,$2F,$88
	.dc.b	$FF,$37,$7D,$4F,$8E,$BC,$11,$E6,$F9,$F9,$18,$3C,$8F,$0F,$82,$F5
	.dc.b	$FC,$3D,$9F,$43,$EE,$F3,$64,$60,$F6,$38,$C7,$F9,$51,$ED,$7F,$D1
	.dc.b	$2D,$8F,$F4,$F6,$C9,$CC,$BE,$AD,$F3,$98,$6D,$FF,$A9,$5B,$3D,$8E
	.dc.b	$13,$CA,$FF,$C6,$15,$DB,$1C,$11,$F0,$FC,$46,$C8,$E0,$F2,$F9,$FF
	.dc.b	$C7,$5F,$5D,$23,$18,$7F,$47,$8C,$4F,$F8,$AE,$3E,$E4,$E9,$FE,$B3
	.dc.b	$F2,$BF,$C1,$5F,$F5,$5B,$67,$E3,$B6,$3F,$A9,$35,$FC,$37,$A1,$CD
	.dc.b	$FC,$73,$6C,$D3,$1F,$29,$2F,$C9,$5F,$95,$8C,$3E,$28,$E9,$7F,$E0
	.dc.b	$A8,$C4,$70,$8C,$3F,$73,$EA,$FF,$9F,$37,$D0,$F2,$46,$2B,$FB,$53
	.dc.b	$5F,$EF,$2B,$C3,$D9,$2B,$CB,$19,$6C,$66,$7F,$DA,$33,$96,$C7,$83
	.dc.b	$DA,$FE,$37,$1F,$83,$9F,$F0,$57,$87,$FE,$EB,$5F,$41,$FE,$D9,$9D
	.dc.b	$57,$07,$D9,$F0,$DF,$E0,$1C,$6C,$79,$23,$0F,$B1,$AF,$F8,$53,$EB
	.dc.b	$39,$4F,$E8,$1B,$1F,$E0,$B6,$BB,$C7,$8A,$C7,$FD,$B6,$FC,$2F,$47
	.dc.b	$31,$88,$F8,$31,$8F,$66,$47,$C8,$6B,$FE,$F6,$B7,$8B,$D9,$B1,$F8
	.dc.b	$26,$BB,$8E,$FE,$94,$8F,$A6,$D7,$E8,$FA,$3E,$42,$3F,$61,$F3,$FA
	.dc.b	$3F,$B3,$8F,$BE,$D7,$DC,$DC,$8E,$7A,$BF,$8F,$CF,$EF,$7C,$2C,$11
	.dc.b	$FB,$5F,$0B,$C4,$8E,$0F,$BD,$E9,$7F,$8E,$7C,$7F,$F0,$76,$28,$C7
	.dc.b	$0F,$E5,$1B,$5F,$94,$F4,$2B,$F4,$FD,$0B,$D3,$FE,$76,$8F,$F2,$8E
	.dc.b	$3F,$E1,$98,$C4,$7F,$73,$C7,$FD,$6C,$EB,$1F,$D0,$FC,$57,$49,$CC
	.dc.b	$EF,$0C,$4F,$83,$EB,$1F,$43,$E2,$F2,$1E,$F5,$35,$1B,$F9,$6B,$8F
	.dc.b	$73,$FC,$CB,$7C,$9C,$05,$1D,$FE,$4B,$DB,$DF,$2B,$74,$9F,$27,$A9
	.dc.b	$B9,$CC,$61,$DF,$C3,$7C,$8F,$40,$D3,$D8,$7B,$7E,$1A,$CF,$03,$BB
	.dc.b	$F0,$79,$F9,$69,$27,$C7,$1F,$17,$E2,$75,$18,$3C,$EF,$D8,$FA,$C7
	.dc.b	$A9,$E3,$C2,$57,$C2,$9C,$B7,$CA,$E1,$52,$30,$7C,$8F,$A9,$7D,$67
	.dc.b	$E1,$3C,$D7,$C3,$53,$F0,$F3,$AA,$E1,$EF,$E2,$E4,$60,$E9,$EA,$7E
	.dc.b	$3B,$EB,$1A,$D3,$9C,$74,$F5,$6D,$70,$3C,$8F,$83,$D3,$FC,$BF,$D6
	.dc.b	$3D,$AA,$66,$BD,$5B,$57,$FC,$91,$AA,$FD,$33,$EA,$F8,$7B,$5C,$0F
	.dc.b	$84,$FE,$93,$EB,$3F,$CF,$1E,$2F,$F9,$E7,$EA,$7F,$D6,$D5,$F8,$E6
	.dc.b	$BF,$E5,$59,$DF,$8B,$FE,$AB,$7F,$4F,$5F,$E8,$17,$33,$FE,$BC,$6A
	.dc.b	$77,$3C,$7D,$3F,$DA,$FD,$63,$D8,$A7,$96,$3C,$DF,$0F,$70,$FD,$0E
	.dc.b	$09,$CC,$63,$EA,$FF,$7F,$F5,$8F,$83,$A7,$BE,$3E,$07,$A9,$6B,$81
	.dc.b	$D1,$F8,$9E,$A6,$A3,$07,$81,$F9,$1E,$97,$D6,$F8,$09,$3E,$48,$F9
	.dc.b	$9E,$9E,$A3,$07,$A1,$FA,$1F,$13,$FC,$53,$8D,$EC,$D3,$35,$EE,$F8
	.dc.b	$8A,$FD,$1A,$FB,$11,$FC,$73,$DE,$7D,$67,$E4,$7C,$D9,$7F,$C1,$A7
	.dc.b	$2D,$F2,$F8,$54,$8C,$1B,$F4,$7A,$4C,$8C,$1C,$FE,$87,$F3,$DF,$09
	.dc.b	$23,$07,$7F,$9A,$F2,$4C,$8C,$1E,$3F,$43,$F6,$1F,$0D,$23,$07,$9F
	.dc.b	$C8,$F6,$4C,$8C,$1E,$BF,$88,$FE,$77,$F1,$4E,$8A,$7C,$38,$F8,$9D
	.dc.b	$56,$B8,$1F,$0D,$F9,$1E,$EE,$A3,$07,$C7,$7F,$A3,$FD,$63,$C0,$BD
	.dc.b	$F2,$BF,$98,$CE,$5B,$EE,$70,$F7,$64,$60,$FE,$6F,$F5,$9F,$DD,$3C
	.dc.b	$7F,$F7,$32,$E3,$FE,$AA,$D9,$DF,$C1,$E3,$87,$3F,$2F,$27,$9C,$7F
	.dc.b	$5C,$FA,$CF,$F4,$A4,$7F,$44,$7F,$91,$BF,$EC,$B6,$E3,$FE,$F5,$77
	.dc.b	$19,$F9,$4E,$5E,$94,$8E,$51,$FD,$A3,$BB,$FC,$79,$1F,$D4,$1F,$DC
	.dc.b	$A3,$FA,$67,$C5,$EF,$E0,$78,$78,$DC,$7E,$6E,$57,$18,$FE,$E7,$F5
	.dc.b	$9F,$D4,$E3,$FA,$DB,$FD,$69,$BE,$0E,$B1,$DF,$17,$0E,$F9,$5C,$23
	.dc.b	$9F,$BF,$D3,$E0,$24,$F4,$8F,$F0,$0F,$D6,$3C,$2B,$EF,$4A,$FE,$D7
	.dc.b	$39,$7F,$B6,$FD,$63,$7F,$F0,$4B,$FE,$1C,$8C,$FC,$5E,$27,$FC,$1B
	.dc.b	$DD,$E9,$9F,$F7,$D3,$FD,$DD,$BE,$3E,$B8,$37,$E1,$F0,$FA,$C9,$5F
	.dc.b	$DF,$2F,$FD,$F8,$DD,$FA,$F8,$72,$30,$7F,$80,$BE,$B3,$FA,$CB,$7E
	.dc.b	$0B,$FC,$3F,$06,$F0,$7F,$88,$BE,$B3,$F9,$9C,$FF,$2F,$AE,$5E,$C6
	.dc.b	$7D,$AD,$70,$EF,$D6,$95,$FE,$10,$C3,$E1,$2F,$05,$7F,$83,$84,$BB
	.dc.b	$3C,$1C,$92,$79,$47,$8F,$D5,$F2,$32,$3B,$47,$9F,$C8,$F4,$79,$12
	.dc.b	$7C,$51,$EB,$F4,$F5,$18,$3C,$CF,$F9,$03,$EA,$DF,$13,$FE,$42,$FA
	.dc.b	$C7,$BD,$4D,$46,$FC,$2B,$87,$F8,$BA,$F8,$B6,$3F,$C9,$7F,$19,$FA
	.dc.b	$CE,$0F,$81,$EA,$34,$FF,$27,$B6,$FF,$C6,$F5,$B7,$F5,$0F,$EC,$B8
	.dc.b	$BE,$4C,$63,$E2,$3A,$9D,$BE,$27,$FC,$B3,$F5,$9F,$8B,$AD,$D7,$F8
	.dc.b	$3B,$9B,$B9,$AF,$0B,$7E,$15,$9F,$F2,$3D,$F1,$D7,$EB,$3F,$97,$C3
	.dc.b	$9B,$FB,$39,$FB,$93,$AA,$E7,$1C,$BC,$8F,$1F,$0F,$BB,$9E,$4F,$68
	.dc.b	$F3,$F0,$EA,$30,$78,$9F,$5F,$91,$F8,$98,$FF,$7F,$65,$FC,$8F,$3F
	.dc.b	$F4,$73,$63,$33,$B7,$F1,$C2,$3F,$D4,$7F,$59,$C3,$D2,$34,$93,$FE
	.dc.b	$97,$FA,$6F,$F4,$C2,$69,$1F,$F4,$F7,$D1,$FC,$E7,$BB,$5F,$CF,$35
	.dc.b	$FE,$44,$D7,$FC,$AF,$BF,$FA,$97,$8F,$04,$7F,$D5,$1C,$7B,$23,$83
	.dc.b	$FD,$59,$F3,$7C,$C8,$E0,$FF,$58,$7C,$57,$72,$38,$3F,$D6,$DB,$7F
	.dc.b	$9F,$DB,$C5,$E7,$8C,$63,$F2,$8F,$DD,$46,$78,$65,$72,$6C,$EF,$EA
	.dc.b	$DA,$7C,$16,$D7,$FB,$15,$B3,$B9,$D6,$92,$B4,$D9,$E9,$E2,$F3,$24
	.dc.b	$F3,$7B,$79,$D2,$3B,$36,$7C,$9E,$A7,$8C,$93,$E3,$7C,$DE,$A4,$8F
	.dc.b	$33,$67,$D9,$F1,$5E,$B2,$4F,$AD,$FB,$DF,$32,$F9,$9A,$49,$FF,$72
	.dc.b	$46,$5B,$E2,$70,$23,$83,$FD,$D5,$F5,$6F,$9E,$FC,$53,$63,$31,$83
	.dc.b	$3E,$14,$8E,$2D,$CB,$ED,$7D,$5B,$9F,$F2,$8D,$AF,$F7,$C3,$67,$73
	.dc.b	$AF,$9F,$95,$A6,$CF,$E1,$7E,$AF,$F8,$E4,$FF,$C0,$16,$C6,$7C,$6F
	.dc.b	$D1,$46,$23,$F6,$B2,$EE,$FD,$33,$4F,$C8,$47,$F4,$88,$C9,$FA,$BF
	.dc.b	$AB,$FF,$06,$4E,$9B,$3F,$97,$6B,$FC,$09,$AF,$F5,$15,$DB,$CC,$93
	.dc.b	$CD,$ED,$C7,$23,$B3,$6B,$F4,$9D,$DF,$E8,$D9,$F3,$B6,$33,$EA,$FF
	.dc.b	$01,$8C,$47,$F1,$47,$FB,$04,$F0,$F1,$34,$47,$E0,$FD,$5F,$EC,$AB
	.dc.b	$F8,$DB,$5F,$ED,$4D,$7F,$C5,$CD,$4F,$FA,$76,$39,$FC,$2F,$F1,$38
	.dc.b	$C4,$FE,$F5,$AF,$6F,$73,$44,$AF,$1B,$FC,$6F,$BB,$FC,$4D,$1F,$C8
	.dc.b	$BB,$30,$47,$EC,$FD,$5B,$E0,$3F,$91,$36,$33,$18,$3F,$93,$7D,$5B
	.dc.b	$E1,$29,$27,$F0,$71,$96,$FC,$27,$02,$38,$37,$D5,$D0,$8E,$0E,$7E
	.dc.b	$FF,$80,$8E,$0E,$FB,$F9,$08,$E0,$FC,$AF,$D5,$7E,$33,$7F,$F2,$F3
	.dc.b	$53,$49,$B7,$F4,$37,$F9,$0C,$7F,$56,$6B,$FE,$25,$6B,$D3,$AF,$81
	.dc.b	$2B,$C2,$F3,$F2,$A4,$79,$1B,$3F,$CE,$69,$E9,$78,$49,$5E,$A7,$C9
	.dc.b	$EE,$C8,$F6,$B6,$B9,$78,$FE,$92,$4F,$C2,$7D,$9E,$3C,$8F,$8A,$D9
	.dc.b	$F1,$75,$7C,$12,$4F,$CA,$7E,$27,$CB,$48,$F9,$AD,$9F,$57,$AD,$F2
	.dc.b	$49,$3F,$49,$F9,$9F,$4B,$23,$EA,$B6,$7F,$AD,$7D,$3F,$E4,$5B,$CF
	.dc.b	$FD,$D6,$31,$1F,$E0,$86,$BF,$E5,$75,$F0,$FB,$25,$FF,$16,$8C,$B7
	.dc.b	$E2,$FA,$11,$C1,$BF,$33,$C0,$47,$07,$3F,$8C,$F2,$11,$C1,$DF,$DA
	.dc.b	$F4,$11,$C1,$E3,$F8,$1F,$61,$1C,$1E,$7F,$7F,$E0,$11,$C1,$EB,$FA
	.dc.b	$3F,$88,$47,$07,$BF,$E4,$BE,$41,$1C,$1F,$CE,$34,$7A,$5F,$E2,$16
	.dc.b	$C6,$63,$07,$F8,$97,$E9,$FF,$8C,$13,$D8,$C1,$9F,$62,$57,$F4,$49
	.dc.b	$FF,$3B,$B6,$39,$79,$D2,$34,$DE,$2F,$F0,$D5,$7F,$69,$6F,$A7,$E1
	.dc.b	$25,$75,$F0,$BC,$84,$70,$78,$7C,$2F,$41,$1C,$1F,$D6,$3C,$EF,$F1
	.dc.b	$ED,$FF,$8A,$5A,$FF,$95,$1A,$FF,$A8,$DA,$FF,$B9,$DF,$EA,$B1,$CD
	.dc.b	$FF,$25,$18,$8E,$8E,$3C,$29,$3E,$16,$CF,$87,$C2,$E2,$49,$F3,$3D
	.dc.b	$3C,$99,$1E,$96,$CF,$A7,$E5,$BC,$24,$9F,$73,$E4,$F9,$D9,$1F,$05
	.dc.b	$B3,$F0,$7C,$1F,$49,$27,$E3,$3E,$CF,$06,$47,$C9,$6C,$FC,$9F,$53
	.dc.b	$E0,$92,$7E,$73,$F1,$3D,$89,$1F,$45,$B3,$F4,$7C,$3F,$F5,$4C,$FD
	.dc.b	$16,$C7,$D1,$6B,$F4,$5A,$FF,$81,$BC,$3B,$BF,$FB,$12,$BF,$C1,$6D
	.dc.b	$9F,$F4,$13,$63,$EF,$78,$7C,$08,$F0,$79,$FC,$3C,$8E,$6D,$AF,$F4
	.dc.b	$67,$D2,$7F,$BA,$27,$FC,$32,$D8,$CC,$79,$A3,$0F,$FA,$81,$FE,$3C
	.dc.b	$DF,$17,$44,$79,$F6,$71,$23,$83,$BF,$CB,$7F,$9E,$AF,$FD,$05,$68
	.dc.b	$C4,$77,$6B,$FE,$FD,$8D,$79,$7F,$E3,$A5,$F4,$FF,$C1,$27,$9B,$5E
	.dc.b	$53,$B8,$C4,$73,$F5,$8F,$C2,$6B,$F1,$9B,$87,$C1,$6C,$78,$FC,$8A
	.dc.b	$F2,$C1,$18,$84,$61,$F9,$F1,$98,$FF,$3E,$47,$D2,$F2,$30,$47,$EA
	.dc.b	$FC,$C6,$C8,$E0,$FB,$3F,$35,$CC,$8E,$0F,$BB,$E5,$77,$23,$83,$F0
	.dc.b	$3E,$57,$8C,$8E,$0F,$C2,$F9,$5F,$E6,$87,$D6,$DC,$4D,$31,$F1,$72
	.dc.b	$7E,$B3,$5F,$BF,$EC,$71,$23,$F6,$DE,$9E,$64,$8F,$BC,$D7,$F0,$FE
	.dc.b	$67,$84,$8F,$E0,$DF,$27,$C5,$C8,$FC,$33,$5F,$C7,$FB,$BE,$92,$3F
	.dc.b	$8B,$7D,$DF,$48,$F8,$AA,$49,$FF,$44,$C6,$5B,$F1,$5C,$08,$E0,$DF
	.dc.b	$9F,$D0,$8E,$0E,$7E,$7F,$80,$8E,$0E,$FE,$C7,$90,$8E,$0F,$1F,$B1
	.dc.b	$E8,$23,$83,$CF,$F2,$1E,$C2,$38,$3D,$7E,$EF,$C0,$23,$83,$FD,$4F
	.dc.b	$F4,$6F,$86,$FC,$13,$63,$31,$83,$3E,$2C,$8E,$2D,$7F,$E0,$07,$D0
	.dc.b	$3E,$8F,$F1,$8D,$8F,$F5,$AB,$5D,$C6,$BD,$C9,$1A,$6B,$D3,$CF,$E6
	.dc.b	$47,$9B,$DB,$CF,$91,$D9,$AF,$93,$CD,$F1,$91,$F1,$BE,$6F,$36,$47
	.dc.b	$99,$AF,$B3,$CD,$F5,$91,$F5,$BF,$A9,$FA,$07,$8F,$49,$3F,$ED,$48
	.dc.b	$CB,$7C,$CE,$04,$70,$6F,$C8,$E8,$47,$07,$3F,$23,$C0,$47,$07,$7F
	.dc.b	$1B,$C8,$47,$07,$8F,$C6,$F4,$11,$C1,$E7,$F1,$BD,$84,$70,$7A,$FC
	.dc.b	$3F,$80,$47,$07,$BF,$C3,$F8,$84,$70,$7C,$3F,$89,$F9,$04,$70,$7C
	.dc.b	$7F,$73,$E6,$11,$C1,$FE,$FE,$F3,$5F,$3D,$FC,$65,$B1,$98,$C1,$9F
	.dc.b	$12,$47,$16,$BF,$D1,$0F,$E8,$F2,$F0,$FA,$12,$3C,$0F,$7F,$90,$91
	.dc.b	$E3,$6B,$CC,$FE,$C1,$F2,$1E,$52,$3E,$97,$D1,$E1,$C8,$F6,$35,$E5
	.dc.b	$6F,$0F,$DC,$48,$F8,$4F,$C1,$F6,$E4,$7C,$56,$BC,$BD,$BF,$8C,$47
	.dc.b	$E5,$3F,$27,$E5,$64,$7C,$D6,$BE,$BF,$9C,$FE,$BC,$DE,$6F,$E9,$11
	.dc.b	$88,$F9,$4D,$7F,$B7,$AF,$2F,$64,$8F,$8C,$23,$4E,$3E,$71,$F0,$3F
	.dc.b	$D5,$5B,$19,$8C,$19,$D6,$47,$16,$BF,$4F,$E7,$1E,$3F,$F5,$F6,$C7
	.dc.b	$89,$AE,$E3,$5E,$E4,$8D,$35,$FB,$DE,$77,$FA,$F9,$F5,$B5,$CF,$B3
	.dc.b	$FD,$B6,$31,$1F,$8F,$5D,$79,$24,$72,$7F,$BA,$7C,$E2,$69,$27,$E4
	.dc.b	$C6,$5B,$CB,$C0,$8E,$0D,$F9,$DD,$08,$E0,$FA,$7F,$38,$F8,$0F,$F0
	.dc.b	$23,$63,$31,$83,$FC,$11,$F3,$8F,$5E,$92,$7E,$EC,$65,$FB,$FE,$2F
	.dc.b	$EB,$5B,$DB,$FE,$11,$8C,$47,$EA,$5A,$FE,$D5,$75,$6C,$91,$D1,$E5
	.dc.b	$E9,$C8,$EE,$D7,$F6,$9F,$36,$F8,$9F,$F1,$03,$63,$F2,$6D,$77,$1F
	.dc.b	$E2,$5F,$9B,$FE,$73,$1F,$BD,$6B,$FC,$25,$AF,$F9,$39,$AF,$F9,$49
	.dc.b	$FF,$15,$BF,$A0,$6B,$CF,$DB,$EC,$77,$E3,$91,$B6,$BE,$1F,$1F,$C8
	.dc.b	$47,$A3,$E7,$ED,$91,$DD,$AF,$A7,$C7,$F6,$11,$F2,$3F,$E4,$3F,$9A
	.dc.b	$7B,$34,$93,$FB,$88,$CB,$7C,$6E,$04,$70,$7E,$FB,$E3,$3F,$C1,$8F
	.dc.b	$F2,$56,$BF,$CC,$5A,$FF,$9D,$9A,$FF,$9E,$9F,$DA,$BB,$ED,$FF,$2C
	.dc.b	$46,$23,$84,$7F,$96,$FE,$6B,$FC,$75,$1F,$CC,$DA,$FF,$3F,$6B,$FE
	.dc.b	$84,$6B,$FE,$88,$7F,$22,$E7,$93,$A1,$1F,$E3,$31,$FB,$17,$B7,$67
	.dc.b	$84,$8E,$B2,$F2,$11,$C1,$CB,$1F,$41,$1C,$1D,$BA,$DF,$33,$49,$1E
	.dc.b	$D8,$CB,$EF,$F0,$9F,$35,$E6,$6C,$66,$30,$7F,$A0,$57,$CA,$D2,$3F
	.dc.b	$E8,$57,$FD,$0D,$F0,$DF,$9C,$83,$49,$EA,$FF,$A2,$FD,$ED,$11,$C1
	.dc.b	$FE,$8F,$E0,$E4,$47,$07,$FA,$4F,$E3,$3B,$11,$C1,$FE,$97,$F8,$CF
	.dc.b	$11,$1C,$1F,$D3,$35,$FE,$50,$DE,$0F,$D9,$C6,$32,$FE,$EC,$D7,$FC
	.dc.b	$04,$FE,$82,$79,$70,$70,$23,$FD,$63,$E6,$BE,$0B,$74,$FD,$FC,$62
	.dc.b	$FF,$C0,$4D,$7F,$C1,$CF,$F9,$BE,$79,$69,$C0,$8F,$5E,$CE,$64,$70
	.dc.b	$78,$7C,$3E,$E4,$70,$79,$7D,$DF,$19,$1C,$1E,$9F,$27,$CE,$47,$07
	.dc.b	$B7,$DD,$F5,$91,$C1,$FD,$E3,$E4,$1F,$13,$FC,$79,$B1,$98,$C1,$9E
	.dc.b	$D9,$1C,$5B,$AF,$F8,$9F,$E6,$9F,$21,$FC,$A1,$B5,$FE,$06,$6C,$EE
	.dc.b	$75,$EB,$4A,$D3,$67,$FC,$65,$F3,$4E,$7F,$E6,$6D,$AF,$F0,$8B,$67
	.dc.b	$73,$AF,$90,$95,$A6,$CF,$F8,$F7,$E6,$BF,$C3,$93,$FE,$1A,$6C,$67
	.dc.b	$C4,$FE,$83,$18,$8F,$F2,$74,$FF,$87,$0F,$F1,$09,$5F,$F2,$43,$5D
	.dc.b	$F7,$7F,$4A,$F9,$A4,$52,$4F,$F8,$AA,$32,$DE,$DE,$04,$70,$6F,$8F
	.dc.b	$A1,$1C,$1F,$E3,$4E,$E7,$9B,$FD,$69,$B1,$98,$C1,$9F,$22,$47,$16
	.dc.b	$E3,$FE,$63,$F9,$AF,$D7,$3E,$A0,$E1,$F6,$B5,$CF,$81,$FD,$AA,$1E
	.dc.b	$DD,$25,$FF,$92,$A3,$2D,$F4,$F8,$11,$C1,$FE,$50,$F0,$3F,$C3,$CD
	.dc.b	$ED,$FE,$EF,$18,$C7,$FC,$F4,$D7,$FD,$0D,$F3,$4F,$84,$FE,$FA,$DB
	.dc.b	$FF,$2D,$B6,$7F,$D0,$4D,$7F,$D1,$DF,$32,$F9,$AF,$F0,$2B,$6B,$FC
	.dc.b	$C6,$D9,$FF,$45,$35,$E3,$ED,$F0,$24,$F0,$79,$F5,$C8,$E6,$D9,$E7
	.dc.b	$EA,$F8,$09,$3E,$07,$C7,$BC,$8F,$1B,$67,$FD,$39,$F3,$5F,$8E,$69
	.dc.b	$FE,$1E,$70,$FF,$AA,$DA,$FD,$F7,$F9,$13,$75,$E0,$4A,$FF,$41,$CF
	.dc.b	$EC,$E7,$97,$67,$52,$4E,$BC,$AF,$01,$1C,$1C,$BD,$0F,$21,$1C,$1D
	.dc.b	$BE,$33,$D0,$47,$07,$8B,$C8,$F6,$11,$C1,$E6,$F2,$3E,$01,$1C,$1E
	.dc.b	$AE,$BF,$88,$47,$07,$BB,$E2,$FE,$41,$1C,$1F,$0B,$E2,$FE,$61,$1C
	.dc.b	$1F,$1B,$E5,$FE,$81,$1C,$1F,$EA,$CF,$9A,$FC,$86,$DF,$AF,$6B,$FC
	.dc.b	$29,$A9,$A3,$E0,$3F,$CB,$2D,$9E,$91,$83,$3F,$09,$23,$C0,$DA,$FF
	.dc.b	$71,$57,$F4,$46,$F9,$9D,$09,$5E,$7F,$8B,$EE,$47,$07,$AF,$E2,$FC
	.dc.b	$64,$70,$7B,$FD,$CF,$39,$1C,$1F,$0F,$DC,$F5,$91,$C1,$F1,$FE,$53
	.dc.b	$DE,$47,$07,$CB,$F8,$CF,$86,$47,$07,$FB,$6B,$57,$A7,$FE,$80,$6C
	.dc.b	$66,$30,$7F,$A1,$3E,$6B,$F8,$F1,$3C,$7C,$19,$F6,$25,$7F,$B9,$A7
	.dc.b	$FC,$29,$BF,$F2,$93,$FA,$1B,$87,$0F,$F4,$97,$F9,$73,$5D,$F9,$DE
	.dc.b	$13,$C5,$E9,$CA,$EC,$D9,$EC,$2B,$FD,$63,$3E,$3F,$07,$D4,$47,$CD
	.dc.b	$E0,$FB,$48,$E0,$FF,$7F,$F9,$DF,$E2,$1B,$EA,$D7,$D8,$D7,$FA,$33
	.dc.b	$5F,$EC,$2F,$FB,$EA,$39,$BF,$EA,$E3,$11,$D1,$C7,$81,$27,$C2,$D9
	.dc.b	$F0,$F8,$1C,$49,$3E,$67,$A7,$93,$23,$D2,$D9,$F4,$FC,$D7,$84,$93
	.dc.b	$EE,$7C,$9F,$23,$23,$E0,$B6,$7E,$0F,$81,$E9,$24,$FC,$67,$D9,$E0
	.dc.b	$48,$F9,$2D,$9F,$93,$EA,$7C,$12,$4F,$CE,$7E,$27,$B1,$23,$E8,$B6
	.dc.b	$7E,$8F,$84,$9F,$E6,$FB,$23,$13,$FE,$DE,$8C,$4E,$A7,$2F,$D8,$F0
	.dc.b	$AD,$7D,$86,$BF,$61,$AA,$91,$85,$E1,$52,$57,$91,$D7,$CD,$48,$F3
	.dc.b	$B6,$FF,$DE,$7F,$31,$EB,$9F,$EA,$2D,$8C,$C7,$8A,$30,$FE,$AD,$FE
	.dc.b	$F0,$DF,$3F,$44,$79,$F6,$71,$23,$83,$BF,$CD,$7F,$02,$BF,$D4,$DA
	.dc.b	$31,$1D,$FE,$43,$FD,$52,$EA,$32,$FF,$89,$67,$87,$77,$EC,$A1,$18
	.dc.b	$78,$46,$63,$11,$EB,$F5,$7F,$C5,$D8,$58,$F8,$71,$87,$B3,$5F,$F2
	.dc.b	$6C,$BC,$8C,$47,$81,$B1,$98,$F9,$B1,$8F,$1E,$47,$D0,$6B,$F5,$3E
	.dc.b	$6B,$89,$1F,$AA,$F4,$F9,$49,$1F,$61,$AF,$DC,$F2,$7C,$24,$7E,$EB
	.dc.b	$E4,$F2,$64,$7E,$01,$AF,$E1,$3C,$9F,$94,$FC,$A6,$E4,$69,$8F,$8C
	.dc.b	$93,$F3,$5A,$FD,$CF,$63,$89,$1F,$A4,$F4,$F2,$E4,$7D,$56,$BF,$84
	.dc.b	$F2,$FC,$24,$7E,$D3,$E4,$F9,$A9,$1F,$75,$AF,$E3,$3D,$DF,$49,$1F
	.dc.b	$C1,$3E,$EF,$9A,$7C,$AD,$24,$FE,$2E,$32,$DF,$8A,$E0,$47,$06,$FC
	.dc.b	$FE,$84,$70,$73,$F3,$FC,$04,$70,$77,$F6,$3C,$84,$70,$78,$FD,$8F
	.dc.b	$41,$1C,$1E,$7F,$8E,$F6,$11,$C1,$EB,$F7,$7E,$01,$1C,$1F,$A9,$F9
	.dc.b	$A7,$C3,$7E,$09,$B1,$98,$C1,$9F,$16,$47,$16,$BF,$D9,$3E,$69,$F3
	.dc.b	$5F,$8C,$6C,$7E,$DD,$AE,$E3,$5E,$DC,$8D,$35,$E9,$E7,$F3,$23,$CD
	.dc.b	$ED,$E7,$C8,$EC,$D7,$C9,$E6,$F8,$C8,$F8,$DF,$37,$9B,$23,$CC,$D7
	.dc.b	$D9,$E6,$FA,$C8,$FA,$DF,$D4,$FC,$D7,$F4,$48,$FF,$02,$B5,$FF,$0A
	.dc.b	$35,$FF,$5F,$35,$FF,$6C,$BF,$EA,$97,$F8,$E3,$5E,$7E,$67,$63,$BF
	.dc.b	$8F,$23,$6D,$7C,$3E,$3F,$90,$8F,$47,$CF,$E2,$C8,$EE,$D7,$D3,$E2
	.dc.b	$FB,$08,$F9,$1F,$7F,$8B,$23,$CE,$D7,$E0,$F8,$5F,$10,$8F,$B1,$F8
	.dc.b	$FE,$14,$8F,$7B,$5F,$93,$F1,$3F,$30,$8F,$C4,$7E,$7F,$B7,$23,$E3
	.dc.b	$B5,$FF,$1A,$F9,$AF,$9A,$FE,$3A,$D8,$FE,$90,$D7,$71,$AF,$0E,$46
	.dc.b	$9A,$F4,$F0,$B9,$91,$E6,$F6,$F8,$F9,$1D,$9A,$F9,$3E,$3F,$C6,$47
	.dc.b	$C6,$F9,$BC,$29,$1E,$66,$BE,$CF,$0B,$D6,$47,$D6,$FB,$BD,$B9,$1E
	.dc.b	$E6,$BF,$13,$DB,$F8,$64,$7E,$1B,$F1,$BE,$6A,$47,$C6,$6B,$FE,$60
	.dc.b	$F9,$AF,$EB,$CD,$E6,$FE,$93,$18,$8F,$F3,$93,$51,$17,$95,$C0,$91
	.dc.b	$FD,$EE,$32,$FF,$7D,$F9,$A7,$C0,$FF,$56,$6C,$66,$30,$67,$49,$1C
	.dc.b	$5A,$FF,$9E,$3E,$6B,$FC,$5C,$FF,$82,$5A,$E7,$C6,$FE,$C7,$18,$8F
	.dc.b	$F4,$62,$F7,$32,$48,$E4,$FF,$6A,$F3,$7F,$D8,$D1,$FE,$15,$6B,$FE
	.dc.b	$88,$6B,$FE,$97,$6B,$F1,$82,$EA,$E8,$48,$E8,$FF,$78,$F9,$AD,$C7
	.dc.b	$F8,$8D,$AF,$FA,$59,$AF,$FA,$95,$AF,$75,$C7,$CC,$91,$CD,$ED,$E6
	.dc.b	$48,$EC,$D7,$FD,$47,$F3,$4F,$80,$FF,$05,$B6,$3F,$C6,$CD,$77,$1F
	.dc.b	$E1,$0F,$9A,$7A,$B4,$93,$FE,$3D,$8C,$BF,$E3,$FF,$11,$D7,$FE,$18
	.dc.b	$6C,$66,$30,$67,$A6,$47,$16,$BF,$EB,$A9,$7A,$3C,$C9,$1D,$DF,$F1
	.dc.b	$2F,$CD,$3E,$22,$92,$7F,$CA,$51,$97,$FC,$A9,$F3,$5F,$CF,$53,$49
	.dc.b	$EE,$FF,$8C,$1F,$F1,$74,$70,$F6,$B0,$47,$9F,$1E,$C8,$E0,$EF,$E3
	.dc.b	$F3,$23,$83,$C7,$DB,$DC,$8E,$0F,$3F,$8F,$E3,$23,$83,$FC,$DD,$F3
	.dc.b	$4F,$63,$FC,$8E,$D8,$CC,$60,$CF,$8B,$23,$8B,$5F,$F7,$4F,$C5,$FF
	.dc.b	$83,$9B,$EC,$7F,$94,$A3,$11,$FE,$FB,$6B,$FA,$57,$F6,$8D,$EF,$E0
	.dc.b	$47,$FD,$09,$F3,$5F,$E3,$A7,$FD,$EE,$D7,$F5,$C1,$C3,$FB,$16,$BF
	.dc.b	$C6,$1F,$C6,$BC,$B9,$BF,$CC,$F1,$88,$E6,$D7,$97,$76,$88,$F9,$1E
	.dc.b	$39,$48,$F3,$B5,$F3,$E5,$D8,$8F,$B1,$F0,$F6,$BE,$67,$DE,$D9,$F7
	.dc.b	$B5,$F2,$B5,$F7,$78,$6F,$9A,$F8,$ED,$8F,$8E,$D7,$DA,$D7,$E4,$8F
	.dc.b	$95,$FF,$44,$B4,$FF,$45,$B8,$3F,$D1,$BF,$23,$FE,$E8,$C3,$FC,$49
	.dc.b	$C1,$7A,$3E,$9C,$DB,$FE,$45,$35,$FE,$92,$DF,$31,$E7,$E6,$7F,$CB
	.dc.b	$37,$FD,$3B,$0F,$F1,$95,$6A,$30,$7E,$03,$8B,$87,$C6,$F1,$6C,$6A
	.dc.b	$FF,$99,$4F,$5F,$8D,$CC,$7C,$F9,$FC,$4C,$65,$E1,$0B,$FF,$6E,$46
	.dc.b	$21,$3E,$9F,$88,$F6,$63,$FC,$FA,$B8,$BF,$AD,$6F,$1E,$B8,$E4,$76
	.dc.b	$C5,$D5,$8C,$BC,$87,$F6,$E8,$CC,$F9,$3E,$69,$C1,$F2,$DF,$8F,$5C
	.dc.b	$03,$A9,$E1,$C1,$B3,$E8,$75,$58,$F1,$3C,$FC,$0E,$9E,$35,$9F,$03
	.dc.b	$87,$C5,$F1,$D7,$1F,$06,$28,$F2,$7B,$DE,$5F,$32,$DF,$A5,$C3,$EB
	.dc.b	$F8,$EB,$87,$27,$D9,$EF,$FB,$7D,$FB,$5F,$05,$FC,$57,$C5,$BE,$2E
	.dc.b	$FC,$5E,$1B,$5F,$6A,$5C,$1C,$3D,$CE,$2D,$9F,$BF,$F3,$4F,$91,$FC
	.dc.b	$B2,$F9,$5B,$18,$73,$CF,$AF,$9E,$B7,$C9,$FC,$F7,$CD,$BC,$0B,$E0
	.dc.b	$61,$B5,$F8,$E9,$71,$70,$E2,$E2,$D9,$FC,$9F,$CD,$FF,$8B,$9B,$B7
	.dc.b	$EA,$E3,$1C,$9F,$81,$7D,$F3,$F9,$A3,$FC,$5E,$5D,$BF,$64,$E2,$7F
	.dc.b	$10,$FC,$13,$F4,$3F,$37,$FE,$3E,$6F,$6F,$EE,$63,$13,$F8,$F7,$F1
	.dc.b	$CD,$F0,$F5,$AE,$DB,$7D,$3A,$39,$69,$CD,$B1,$FB,$0F,$15,$EA,$7F
	.dc.b	$07,$5E,$FD,$8C,$39,$F4,$75,$E1,$DB,$E4,$FF,$0F,$F9,$BF,$DA,$6B
	.dc.b	$FA,$57,$F6,$8D,$F1,$B5,$E3,$6F,$E0,$B8,$7C,$15,$C7,$F8,$1E,$08
	.dc.b	$E9,$EE,$FF,$1B,$E4,$F8,$7E,$57,$FB,$1E,$93,$9C,$F5,$E4,$6F,$C7
	.dc.b	$B9,$7F,$0E,$C7,$F5,$0F,$F1,$0F,$9B,$7C,$4F,$F2,$75,$F2,$B6,$30
	.dc.b	$FF,$2B,$F9,$BF,$CB,$57,$F1,$85,$F0,$99,$F1,$74,$C9,$D7,$8B,$75
	.dc.b	$FE,$11,$7F,$C7,$E7,$27,$F2,$2F,$A9,$78,$7D,$EF,$F9,$D4,$62,$7C
	.dc.b	$1F,$01,$E3,$E5,$F2,$36,$3C,$91,$C3,$CD,$C6,$5F,$C5,$ED,$1F,$D1
	.dc.b	$7E,$73,$E5,$DF,$F9,$A5,$FC,$6F,$37,$FB,$CD,$CD,$FF,$A4,$9C,$CE
	.dc.b	$F8,$BC,$BF,$1F,$CD,$B1,$FC,$F2,$7F,$68,$DF,$27,$C1,$E2,$78,$59
	.dc.b	$1B,$F7,$2E,$3F,$CB,$6D,$1E,$73,$E1,$FC,$D5,$AF,$90,$FA,$FE,$4B
	.dc.b	$D9,$EA,$D9,$F1,$38,$7E,$77,$CD,$5C,$7E,$23,$F0,$FD,$5F,$89,$E4
	.dc.b	$5A,$F5,$38,$7F,$AB,$FD,$51,$C5,$FF,$43,$8C,$47,$C1,$E2,$F9,$DC
	.dc.b	$3F,$3D,$B1,$F1,$63,$E9,$69,$8B,$F8,$0E,$27,$FB,$2F,$E0,$DF,$B6
	.dc.b	$7E,$64,$7F,$7D,$8C,$4F,$F5,$3F,$9E,$FC,$DD,$7F,$6C,$7F,$36,$DE
	.dc.b	$EF,$F0,$24,$62,$78,$77,$7E,$3B,$87,$2D,$BE,$5E,$4F,$0F,$5F,$8B
	.dc.b	$63,$B4,$7F,$83,$FA,$BC,$7D,$3F,$17,$1F,$F6,$CE,$5F,$D2,$6F,$FB
	.dc.b	$FD,$69,$FF,$5C,$43,$E4,$FC,$BF,$27,$E6,$6C,$FA,$E3,$D3,$F0,$F8
	.dc.b	$C3,$E0,$38,$9E,$B9,$D9,$FE,$ED,$8F,$F1,$C9,$F8,$9E,$67,$C2,$F9
	.dc.b	$CF,$86,$D8,$F9,$11,$FE,$29,$FA,$67,$CA,$E7,$0F,$F0,$04,$62,$BE
	.dc.b	$77,$B9,$FE,$2F,$E1,$7B,$BF,$3A,$30,$C9,$FF,$0F,$F4,$3E,$67,$18
	.dc.b	$FF,$82,$E3,$13,$FE,$24,$F9,$FF,$E5,$7C,$9F,$2E,$91,$FC,$B2,$30
	.dc.b	$EA,$7F,$DC,$4E,$5B,$E9,$7F,$90,$FE,$89,$C7,$FE,$45,$AD,$60,$13
	.dc.b	$AB,$F1,$7C,$36,$FD,$CE,$0D,$8F,$34,$75,$F9,$2C,$5F,$F8,$96,$FF
	.dc.b	$DA,$B5,$EB,$F2,$FC,$1C,$DE,$16,$C7,$BE,$3F,$CB,$1F,$4F,$FE,$D6
	.dc.b	$8C,$1F,$E4,$67,$17,$FE,$31,$F7,$7F,$C6,$7B,$FC,$54,$7F,$98,$4C
	.dc.b	$D7,$B6,$73,$19,$3B,$FD,$1D,$9E,$CF,$1F,$33,$97,$99,$67,$FC,$AF
	.dc.b	$5F,$E5,$46,$C7,$A3,$3D,$C7,$9D,$F1,$7A,$DE,$3F,$5A,$CF,$27,$0F
	.dc.b	$F9,$8B,$EA,$1F,$57,$8D,$3F,$C9,$D1,$89,$F0,$F9,$9E,$D3,$DD,$F3
	.dc.b	$F6,$3C,$BE,$6F,$E1,$EF,$E0,$6A,$8C,$47,$DF,$8C,$4F,$FA,$52,$3E
	.dc.b	$4C,$E9,$7D,$57,$F5,$E9,$D3,$63,$FD,$36,$D7,$FD,$8A,$D7,$C3,$EB
	.dc.b	$73,$BF,$EB,$51,$FE,$A2,$6B,$FE,$8D,$8D,$76,$7F,$09,$BF,$F4,$83
	.dc.b	$67,$D6,$D7,$D4,$D7,$F0,$3E,$75,$B3,$49,$5F,$E2,$C9,$FD,$07,$BB
	.dc.b	$F9,$F7,$E4,$36,$3F,$DB,$ED,$8F,$C5,$35,$F9,$5E,$8F,$72,$4F,$58
	.dc.b	$FC,$9C,$79,$7E,$97,$FA,$2C,$7D,$56,$CF,$FB,$B1,$B1,$F9,$76,$BE
	.dc.b	$6E,$BF,$75,$FE,$B2,$3C,$F1,$83,$B7,$89,$2F,$C5,$7F,$CF,$DE,$D1
	.dc.b	$E3,$F8,$DF,$E2,$F7,$FE,$D4,$A4,$62,$3C,$9F,$1D,$FE,$A2,$75,$19
	.dc.b	$7C,$DF,$07,$FA,$DC,$3D,$4D,$8C,$9E,$27,$0F,$F0,$FB,$FF,$28,$B9
	.dc.b	$21,$F1,$63,$6A,$3F,$CA,$AF,$B3,$CC,$CC,$62,$3E,$64,$70,$CF,$FA
	.dc.b	$64,$7D,$36,$BF,$7B,$CE,$D1,$CB,$E3,$24,$7C,$86,$BF,$5F,$E8,$3B
	.dc.b	$91,$F9,$AF,$9B,$EA,$DE,$4D,$24,$FE,$0E,$32,$FD,$4F,$47,$D2,$47
	.dc.b	$07,$D8,$F2,$BF,$68,$FC,$16,$E9,$FC,$EA,$3F,$70,$D7,$ED,$71,$68
	.dc.b	$8E,$9E,$7D,$0F,$1A,$92,$7F,$7A,$F6,$F4,$64,$76,$6B,$F8,$77,$F9
	.dc.b	$67,$8D,$FD,$57,$0F,$F7,$6C,$FF,$BF,$1A,$FF,$B9,$D7,$A3,$C3,$93
	.dc.b	$F9,$A6,$BF,$EC,$1A,$E1,$39,$7F,$0C,$D5,$1B,$7A,$7D,$36,$6C,$D3
	.dc.b	$1F,$31,$27,$C8,$D7,$FD,$F5,$5F,$E7,$A7,$FB,$16,$3F,$DB,$FE,$99
	.dc.b	$F4,$7E,$16,$CE,$B3,$FE,$99,$18,$85,$71,$F9,$9E,$04,$8E,$07,$F6
	.dc.b	$F9,$47,$B7,$07,$C1,$F7,$F8,$A9,$7C,$05,$A3,$C9,$CB,$27,$53,$F8
	.dc.b	$E7,$C7,$97,$9A,$7B,$7C,$97,$AF,$E1,$E5,$70,$7D,$D8,$C8,$E6,$D7
	.dc.b	$E3,$FA,$1F,$0F,$E7,$64,$78,$1F,$8D,$F1,$72,$3C,$6D,$7E,$7F,$A1
	.dc.b	$F6,$FE,$99,$F2,$F4,$93,$EF,$8F,$EF,$6D,$F2,$78,$11,$F8,$9C,$3F
	.dc.b	$9B,$7E,$37,$D3,$7F,$8B,$5B,$ED,$FE,$26,$31,$1F,$10,$FF,$10,$2F
	.dc.b	$88,$D1,$23,$9B,$F9,$1F,$AB,$F3,$C7,$6F,$47,$A1,$1F,$0F,$A5,$DC
	.dc.b	$8E,$0F,$B5,$F6,$FE,$77,$C4,$30,$FF,$60,$BE,$AF,$4B,$D0,$7A,$7D
	.dc.b	$29,$1E,$E6,$BC,$A7,$FA,$FB,$F0,$BD,$2F,$80,$47,$F1,$5F,$79,$FC
	.dc.b	$ED,$F0,$FD,$4F,$F9,$37,$BF,$FC,$9C,$DE,$CF,$F1,$43,$67,$FC,$8C
	.dc.b	$FF,$BA,$1F,$CA,$35,$E7,$F1,$1F,$B7,$8C,$57,$F9,$67,$EB,$7D,$B5
	.dc.b	$E0,$F4,$B4,$47,$9F,$A5,$C4,$8E,$0F,$D1,$71,$7D,$AB,$F3,$F9,$FD
	.dc.b	$C9,$3E,$4F,$47,$C4,$47,$07,$A3,$D0,$F3,$11,$C1,$FA,$FF,$25,$DD
	.dc.b	$EB,$F9,$59,$3F,$01,$F7,$79,$B2,$3E,$1B,$5F,$BB,$9B,$A7,$E2,$7C
	.dc.b	$D4,$9F,$92,$FC,$7F,$86,$91,$F3,$1A,$FE,$1B,$EB,$1B,$7F,$23,$F6
	.dc.b	$3F,$D2,$B3,$FC,$15,$AF,$E3,$1B,$C1,$FE,$89,$7F,$8B,$37,$CE,$E4
	.dc.b	$49,$E5,$E7,$75,$23,$83,$F8,$97,$C4,$7F,$0D,$AE,$9F,$43,$E3,$24
	.dc.b	$F8,$FE,$87,$CC,$47,$07,$2F,$2F,$D5,$E6,$C8,$F5,$3E,$DF,$36,$47
	.dc.b	$B5,$AF,$FA,$9F,$87,$E1,$79,$92,$3E,$13,$F1,$7D,$59,$1F,$15,$AF
	.dc.b	$C5,$F6,$BE,$51,$1F,$94,$FF,$33,$3F,$A8,$7D,$5F,$F8,$37,$5F,$F5
	.dc.b	$93,$7C,$2F,$F4,$63,$6B,$FD,$54,$BE,$3B,$8F,$AD,$27,$F9,$D4,$A3
	.dc.b	$AF,$47,$F5,$DF,$AB,$7C,$5D,$25,$FF,$41,$87,$C5,$EF,$86,$4E,$5E
	.dc.b	$5C,$3F,$D9,$FE,$AF,$F9,$4C,$FF,$0E,$6E,$DF,$EC,$B6,$ED,$FE,$A5
	.dc.b	$5F,$15,$C7,$E1,$65,$FF,$51,$9F,$F4,$53,$7E,$7B,$FB,$A7,$BA,$F8
	.dc.b	$5A,$49,$FE,$B1,$0E,$ED,$F8,$92,$72,$F2,$F1,$3F,$BF,$7D,$5F,$F5
	.dc.b	$D9,$FE,$C0,$D7,$F9,$3B,$76,$FF,$72,$3F,$93,$DB,$FD,$7C,$BB,$79
	.dc.b	$92,$B9,$F2,$76,$F1,$A4,$76,$7F,$C1,$DF,$57,$FE,$C7,$8F,$E6,$AB
	.dc.b	$E3,$32,$48,$FE,$E9,$0F,$8C,$FF,$0B,$7D,$5B,$E1,$E9,$2B,$FB,$D4
	.dc.b	$37,$FF,$0E,$7A,$1F,$E0,$29,$FE,$FC,$D7,$FA,$2B,$77,$6F,$FE,$DB
	.dc.b	$8F,$84,$DF,$0B,$99,$2B,$9F,$85,$DB,$E0,$64,$76,$7C,$3B,$FF,$8B
	.dc.b	$7E,$AD,$F3,$54,$95,$FE,$11,$8F,$84,$FF,$84,$FE,$AF,$FA,$22,$69
	.dc.b	$3D,$7F,$F1,$C3,$FE,$55,$D7,$D3,$23,$87,$C5,$E8,$DF,$0F,$0E,$09
	.dc.b	$3D,$DE,$7D,$12,$3C,$4D,$7F,$B5,$E1,$FE,$4E,$7C,$DE,$2F,$88,$8F
	.dc.b	$A7,$C5,$F2,$91,$C1,$ED,$F4,$7F,$CA,$1F,$57,$F8,$D8,$F4,$35,$34
	.dc.b	$79,$BC,$09,$1F,$E3,$98,$CB,$FE,$18,$E2,$FF,$2E,$7C,$CF,$F8,$32
	.dc.b	$3F,$C0,$6D,$DF,$E3,$C2,$31,$CD,$F2,$63,$1F,$0B,$C8,$8F,$F9,$1E
	.dc.b	$7F,$7A,$FC,$FF,$85,$FE,$B6,$FF,$92,$FE,$AF,$FA,$1D,$F3,$6C,$7D
	.dc.b	$56,$A7,$F2,$51,$BF,$4F,$FC,$E9,$18,$9E,$11,$8F,$12,$47,$76,$BF
	.dc.b	$E2,$0F,$DA,$7F,$5D,$9F,$BA,$FE,$61,$3E,$2F,$AE,$F2,$FD,$0D,$7F
	.dc.b	$C5,$33,$FB,$FA,$FC,$67,$D8,$7C,$79,$F5,$78,$BF,$E7,$47,$F7,$87
	.dc.b	$AD,$C3,$FB,$FA,$E4,$E9,$AE,$7C,$57,$D6,$F2,$8F,$89,$18,$78,$35
	.dc.b	$E1,$1F,$EB,$A5,$F5,$14,$8F,$FA,$F9,$FF,$5F,$FD,$1F,$E9,$B2,$FC
	.dc.b	$0C,$FF,$99,$B8,$3F,$D0,$4A,$75,$1F,$C5,$FE,$3A,$CF,$E1,$5A,$FF
	.dc.b	$A3,$3C,$BA,$F3,$3F,$CA,$3E,$8F,$F3,$5A,$EB,$CD,$AF,$47,$6D,$8F
	.dc.b	$0C,$72,$F9,$9C,$5F,$F9,$68,$39,$9F,$F4,$C5,$BE,$5B,$15,$E0,$7C
	.dc.b	$BA,$7F,$A6,$EB,$FD,$3C,$FF,$BA,$9B,$CD,$AE,$59,$1D,$3C,$0B,$87
	.dc.b	$A9,$F7,$FD,$6F,$C5,$9F,$F5,$43,$EF,$C3,$F5,$6E,$6B,$FC,$E2,$E6
	.dc.b	$7F,$58,$D8,$FE,$2D,$93,$BF,$AF,$07,$CC,$8C,$5F,$6F,$1F,$C3,$E7
	.dc.b	$F8,$9B,$1E,$28,$F2,$F9,$D8,$C7,$C2,$FA,$3E,$3B,$D3,$EA,$DB,$F4
	.dc.b	$B8,$7D,$7E,$A5,$BF,$4B,$F0,$3E,$37,$E0,$FC,$6D,$AF,$82,$FE,$0B
	.dc.b	$E7,$BF,$C2,$91,$FE,$CB,$5F,$31,$9F,$7B,$4E,$F8,$78,$79,$56,$F8
	.dc.b	$3F,$8C,$FA,$DF,$F0,$9C,$7F,$B5,$57,$D4,$67,$DE,$D3,$BE,$4E,$1F
	.dc.b	$2D,$6F,$83,$F9,$6F,$AD,$FF,$08,$C7,$FB,$7D,$FC,$5B,$7B,$F5,$DF
	.dc.b	$BE,$2E,$1C,$56,$B8,$3F,$A0,$FA,$DF,$F1,$74,$7F,$BA,$5F,$AF,$C9
	.dc.b	$FE,$B0,$72,$DE,$0F,$D3,$CF,$F8,$C8,$FF,$78,$1E,$E8,$FF,$5B,$B9
	.dc.b	$6E,$DF,$AE,$FA,$DF,$EB,$93,$FE,$F7,$7F,$1D,$DB,$FC,$E9,$CD,$7F
	.dc.b	$B0,$1C,$CF,$F3,$B5,$E1,$73,$D7,$A3,$67,$A7,$47,$7D,$7C,$0D,$8F
	.dc.b	$ED,$98,$FF,$48,$F1,$5E,$E7,$F0,$25,$F2,$16,$30,$E7,$D6,$D7,$87
	.dc.b	$70,$E6,$7F,$4F,$B4,$7F,$0C,$FA,$DF,$DC,$6F,$FE,$DA,$7F,$70,$DF
	.dc.b	$1B,$5E,$36,$FD,$DE,$1F,$0F,$72,$EB,$7F,$E0,$3C,$13,$D7,$DE,$FE
	.dc.b	$33,$E6,$7C,$8E,$0F,$EC,$75,$FE,$B4,$C3,$8E,$1F,$C8,$CD,$F8,$FC
	.dc.b	$3C,$7B,$3A,$C3,$FC,$6C,$F0,$FA,$D7,$BD,$FC,$99,$FF,$2D,$46,$63
	.dc.b	$1F,$09,$FC,$AB,$EB,$7F,$3B,$3E,$05,$F0,$F9,$F1,$74,$C9,$D7,$8B
	.dc.b	$75,$FF,$7F,$DA,$3A,$1F,$CD,$BE,$B5,$E1,$E2,$FD,$B1,$87,$C1,$F0
	.dc.b	$9E,$3E,$5F,$23,$63,$C9,$1C,$FC,$DC,$64,$E4,$FF,$05,$E0,$9F,$E8
	.dc.b	$BF,$5B,$F3,$70,$FF,$06,$3F,$94,$BF,$F1,$B3,$96,$F1,$70,$3D,$1F
	.dc.b	$33,$6B,$CE,$F4,$F0,$FA,$F9,$16,$7E,$BE,$1F,$BC,$C3,$C9,$1E,$09
	.dc.b	$F1,$FB,$FC,$23,$6E,$1F,$87,$F5,$96,$B8,$9F,$E4,$BD,$11,$EC,$F9
	.dc.b	$DF,$6F,$A9,$6F,$C2,$E1,$F9,$BF,$49,$70,$E6,$30,$47,$C6,$F5,$7E
	.dc.b	$3F,$91,$6F,$D6,$E1,$FC,$9F,$D6,$FF,$9D,$9C,$98,$3E,$7F,$17,$D0
	.dc.b	$E1,$B1,$F1,$1C,$3F,$1B,$5B,$9F,$93,$83,$DB,$39,$3F,$43,$F5,$BF
	.dc.b	$BE,$71,$3E,$A8,$C3,$EC,$FA,$DF,$CF,$CF,$EA,$1F,$CF,$C6,$7B,$BF
	.dc.b	$C0,$F1,$8E,$6F,$F2,$8B,$6B,$FC,$A1,$C1,$FE,$53,$79,$77,$7E,$67
	.dc.b	$93,$8F,$95,$CB,$D7,$B5,$DD,$C3,$FB,$8E,$AF,$2E,$5D,$A7,$FC,$2F
	.dc.b	$18,$9F,$F2,$F6,$4F,$9C,$F2,$FC,$E7,$99,$B3,$EB,$8F,$4F,$C6,$E3
	.dc.b	$7F,$80,$E2,$7C,$19,$D9,$FB,$F8,$FF,$24,$9F,$89,$E6,$7C,$2F,$A6
	.dc.b	$F8,$6D,$8F,$91,$1F,$E2,$CF,$98,$7D,$6E,$70,$FC,$4C,$62,$BE,$77
	.dc.b	$B9,$FE,$32,$E2,$7B,$DF,$3A,$30,$C9,$FE,$37,$D4,$FA,$4C,$63,$F9
	.dc.b	$48,$C4,$FF,$20,$FA,$DF,$E6,$3C,$BF,$E8,$67,$F9,$8B,$7D,$3F,$F2
	.dc.b	$0C,$62,$7F,$D1,$0D,$BF,$F5,$97,$D6,$B8,$F7,$EA,$7F,$92,$23,$17
	.dc.b	$DF,$E1,$F9,$FB,$9D,$1B,$1E,$38,$E1,$F2,$78,$C7,$F5,$77,$FE,$E3
	.dc.b	$AF,$47,$99,$E4,$E7,$F2,$B6,$3D,$91,$FE,$58,$FA,$D7,$C7,$E7,$0F
	.dc.b	$DA,$C6,$2B,$FA,$0F,$BD,$FE,$37,$9F,$ED,$2F,$F9,$80,$C1,$FD,$B2
	.dc.b	$36,$F5,$F9,$DD,$36,$39,$F9,$FC,$7C,$FE,$4D,$8E,$F1,$E5,$CF,$18
	.dc.b	$79,$1F,$07,$B1,$E1,$F6,$2D,$F9,$9C,$3F,$D6,$3E,$B1,$F5,$B8,$C3
	.dc.b	$F8,$8C,$62,$7D,$BE,$87,$FA,$0A,$7F,$DD,$1A,$23,$11,$FE,$08,$6C
	.dc.b	$66,$7F,$D2,$DF,$5B,$FD,$59,$BE,$CF,$08,$C5,$FF,$A9,$1A,$FF,$B3
	.dc.b	$4F,$EE,$7F,$13,$DB,$DF,$FF,$23,$DF,$FA,$85,$B3,$FE,$AC,$6B,$C7
	.dc.b	$0F,$E9,$D3,$FE,$20,$8F,$4F,$72,$31,$1F,$E2,$08,$C3,$FE,$A0,$9E
	.dc.b	$2D,$F4,$B3,$5D,$5B,$1F,$EE,$39,$46,$23,$E3,$B5,$FF,$06,$F6,$5C
	.dc.b	$3F,$3B,$1F,$EA,$E6,$DF,$C1,$6C,$7F,$BB,$D7,$8D,$C4,$93,$EA,$8C
	.dc.b	$BE,$CF,$AD,$FE,$85,$5F,$EB,$A6,$C7,$CD,$6C,$7F,$BE,$DA,$8F,$F1
	.dc.b	$1F,$5F,$9C,$93,$DD,$F5,$7D,$34,$8F,$13,$67,$C1,$F1,$DF,$C4,$1F
	.dc.b	$F7,$0D,$FF,$AC,$FC,$5F,$F6,$D2,$F8,$FF,$F1,$54,$F9,$6B,$83,$E7
	.dc.b	$F8,$6F,$23,$EB,$6E,$39,$8F,$81,$18,$7D,$9E,$6D,$7D,$98,$7F,$94
	.dc.b	$9F,$8D,$19,$51,$97,$E4,$46,$63,$15,$F2,$63,$07,$1D,$3F,$A3,$57
	.dc.b	$D3,$F3,$F0,$47,$F1,$FC,$8E,$BF,$03,$63,$31,$83,$C1,$F4,$92,$3E
	.dc.b	$E3,$5F,$F7,$AF,$CB,$BC,$7F,$53,$63,$F3,$AD,$77,$1E,$FF,$97,$F9
	.dc.b	$FA,$7E,$BD,$B3,$F3,$9A,$FE,$0D,$AF,$C6,$E5,$FD,$B3,$F8,$96,$BF
	.dc.b	$39,$AE,$CF,$BF,$C9,$8D,$CD,$24,$7F,$08,$F1,$78,$EA,$69,$23,$F8
	.dc.b	$17,$C4,$F6,$E0,$34,$91,$F8,$EF,$4B,$F1,$3E,$6F,$EF,$A3,$EF,$C6
	.dc.b	$35,$FD,$13,$63,$F0,$FD,$35,$FC,$2B,$E9,$23,$F9,$A7,$DF,$C1,$FE
	.dc.b	$D2,$8F,$86,$D7,$E3,$B5,$FE,$3C,$D7,$F6,$5D,$3D,$F6,$34,$C7,$CD
	.dc.b	$49,$FF,$48,$CA,$31,$1F,$E4,$AA,$FF,$42,$BF,$E9,$6F,$AD,$7D,$2F
	.dc.b	$AD,$B1,$98,$C7,$97,$23,$83,$5E,$9F,$0D,$C7,$D1,$91,$D5,$E9,$CF
	.dc.b	$23,$C0,$D7,$C9,$F2,$5E,$1F,$16,$47,$95,$F2,$5C,$8F,$43,$5F,$67
	.dc.b	$9B,$E9,$F9,$B9,$1E,$D7,$D9,$F1,$B2,$3E,$03,$5F,$89,$DD,$F6,$FE
	.dc.b	$B5,$F2,$D4,$93,$FE,$BC,$8F,$F7,$1B,$7C,$8E,$04,$7F,$D8,$3C,$5F
	.dc.b	$95,$7F,$D8,$7F,$59,$FE,$29,$6F,$B1,$F8,$98,$C4,$7F,$9F,$97,$C5
	.dc.b	$64,$91,$C9,$FC,$87,$D4,$F9,$A3,$AF,$D4,$73,$23,$E0,$F4,$3B,$11
	.dc.b	$C1,$FE,$D6,$FA,$DF,$33,$E1,$F8,$FF,$21,$1F,$4F,$A3,$E7,$23,$83
	.dc.b	$DB,$E8,$FA,$C8,$E0,$FF,$71,$FD,$6F,$F3,$AF,$13,$F5,$DF,$5B,$FA
	.dc.b	$F6,$F8,$BF,$AB,$8C,$3F,$CF,$1F,$F7,$1D,$7F,$BB,$1A,$FF,$A9,$1B
	.dc.b	$3D,$3E,$27,$F6,$9F,$5B,$FE,$D8,$AE,$1E,$97,$52,$3C,$FD,$2F,$01
	.dc.b	$1C,$1F,$EF,$AE,$1F,$AB,$87,$A3,$D1,$F3,$12,$7C,$9E,$97,$A4,$8E
	.dc.b	$0F,$47,$C3,$7B,$48,$E0,$FE,$D9,$5F,$C4,$7C,$A7,$77,$C3,$F3,$65
	.dc.b	$7C,$17,$E3,$79,$B2,$3E,$23,$5F,$F6,$3E,$6E,$8F,$99,$F3,$52,$7E
	.dc.b	$53,$F3,$FE,$12,$47,$CD,$6B,$FE,$D0,$FA,$D6,$FF,$C7,$FD,$8F,$F4
	.dc.b	$6C,$FA,$39,$BF,$B8,$C6,$23,$FD,$AE,$DD,$FF,$96,$3F,$C3,$23,$3E
	.dc.b	$6F,$32,$3C,$CF,$F0,$74,$7F,$9B,$9C,$1F,$E0,$3F,$37,$C4,$47,$3F
	.dc.b	$0F,$FD,$6E,$FB,$FD,$27,$9C,$93,$CA,$DF,$47,$EB,$24,$65,$FE,$79
	.dc.b	$6F,$2F,$E0,$79,$B2,$B9,$3F,$0F,$CD,$91,$D5,$AF,$FA,$97,$8B,$E4
	.dc.b	$79,$92,$3C,$4F,$CB,$F5,$A4,$79,$5A,$FC,$AF,$5B,$E8,$11,$F5,$3F
	.dc.b	$D4,$7C,$B3,$FD,$4F,$EB,$7F,$C2,$DC,$1F,$D6,$DB,$C7,$FE,$26,$6D
	.dc.b	$FF,$8E,$D7,$C6,$F1,$F5,$A4,$FD,$09,$FE,$94,$DE,$9F,$EC,$1F,$5A
	.dc.b	$F8,$AA,$49,$FA,$F0,$F8,$AD,$F1,$49,$CB,$CB,$8B,$FB,$57,$D6,$FF
	.dc.b	$2C,$96,$DF,$DC,$E3,$0D,$F4,$F7,$F1,$3C,$3E,$5A,$5F,$E1,$AD,$1D
	.dc.b	$3E,$7F,$FB,$B7,$B8,$F8,$FA,$4B,$FC,$6C,$7F,$93,$DB,$E3,$70,$23
	.dc.b	$AF,$1B,$FC,$01,$F5,$AA,$A4,$9F,$CB,$C7,$E4,$F9,$FF,$CA,$AF,$E4
	.dc.b	$23,$3D,$FC,$8E,$7C,$9D,$3C,$89,$3C,$9F,$F0,$77,$D6,$FF,$A8,$78
	.dc.b	$BF,$C2,$CB,$E3,$72,$4A,$FD,$3C,$3E,$37,$FC,$2D,$F5,$BF,$E2,$C9
	.dc.b	$FD,$5B,$5F,$9F,$A7,$F9,$AA,$31,$87,$F8,$8D,$FF,$14,$CF,$EC,$BD
	.dc.b	$1F,$DA,$3C,$0F,$F4,$23,$89,$FF,$15,$36,$BF,$CE,$AF,$C7,$9E,$5E
	.dc.b	$17,$F8,$A6,$3F,$CB,$D1,$CC,$38,$AE,$7E,$16,$FE,$02,$57,$8D,$E5
	.dc.b	$AF,$F8,$CF,$EB,$5F,$3B,$49,$5F,$C2,$A3,$E4,$3F,$C3,$3E,$B7,$FA
	.dc.b	$7B,$FE,$9D,$6A,$69,$FD,$3B,$4F,$F4,$7A,$F8,$BF,$F2,$13,$67,$A4
	.dc.b	$63,$E2,$F3,$D1,$27,$C0,$F0,$E8,$91,$E3,$6B,$F0,$EF,$FD,$7A,$F9
	.dc.b	$FC,$5E,$E4,$7D,$5E,$27,$88,$8E,$0F,$77,$A5,$FE,$57,$FA,$DF,$CA
	.dc.b	$47,$FA,$ED,$A9,$A3,$CF,$E0,$48,$FE,$67,$19,$7F,$C3,$FC,$3F,$E6
	.dc.b	$3F,$99,$FF,$0A,$47,$DB,$6F,$27,$F9,$B5,$BC,$1F,$EB,$87,$F7,$AD
	.dc.b	$F1,$39,$12,$BF,$C5,$7C,$9F,$E6,$FF,$AC,$FE,$95,$87,$F9,$DD,$B7
	.dc.b	$FE,$C2,$6D,$7F,$B5,$DF,$C9,$37,$D2,$E4,$47,$FA,$6C,$7E,$55,$FC
	.dc.b	$93,$6B,$B7,$83,$E1,$3F,$D0,$1F,$1F,$FB,$3C,$3C,$1F,$17,$E5,$26
	.dc.b	$1D,$37,$F4,$11,$C1,$E0,$F0,$BF,$8F,$AF,$AD,$6F,$EB,$9C,$CF,$FB
	.dc.b	$78,$69,$FD,$6B,$4F,$F5,$6B,$63,$FD,$62,$D7,$83,$FC,$DF,$EB,$77
	.dc.b	$E2,$1A,$49,$FF,$5A,$FD,$47,$FA,$D9,$34,$8F,$FA,$F3,$E9,$1B,$7F
	.dc.b	$76,$8F,$F3,$4B,$63,$31,$FB,$D3,$25,$F8,$EC,$E5,$FE,$B1,$F3,$BF
	.dc.b	$E0,$05,$F3,$9C,$BF,$EF,$E8,$CD,$73,$E5,$CA,$27,$9D,$87,$11,$54
	.dc.b	$53,$1F,$33,$6F,$D4,$E1,$FF,$08,$DB,$E5,$71,$7F,$C5,$E3,$13,$FE
	.dc.b	$16,$CB,$FA,$23,$79,$3C,$51,$8A,$FE,$3C,$FF,$BD,$67,$47,$6E,$FD
	.dc.b	$D7,$A6,$D1,$EB,$FA,$D6,$D7,$6C,$37,$0F,$F1,$3C,$FC,$EC,$7E,$67
	.dc.b	$76,$73,$F4,$3C,$FC,$5E,$9E,$65,$C3,$8B,$87,$9F,$97,$6F,$F9,$B6
	.dc.b	$3F,$D3,$67,$BF,$C6,$F9,$7D,$1F,$33,$63,$C7,$1E,$9F,$47,$17,$C9
	.dc.b	$F6,$7C,$67,$B7,$E3,$2D,$FA,$9F,$B5,$F2,$FF,$E1,$68,$FF,$21,$AF
	.dc.b	$98,$CF,$B7,$A7,$7C,$1C,$3C,$9B,$7C,$1F,$C2,$7D,$6F,$F8,$56,$3F
	.dc.b	$C9,$AB,$EA,$33,$ED,$E9,$DF,$4F,$0F,$96,$B7,$C1,$FC,$77,$D6,$FE
	.dc.b	$32,$3F,$CA,$EB,$BB,$3D,$FA,$77,$C5,$C3,$8A,$DF,$07,$F3,$1F,$5B
	.dc.b	$FE,$F9,$8F,$F3,$0A,$DF,$3E,$86,$9F,$CF,$4F,$F8,$CC,$FF,$33,$9E
	.dc.b	$DE,$5F,$ED,$8E,$5B,$BF,$E9,$7E,$B7,$F5,$13,$FE,$6E,$7F,$1A,$D9
	.dc.b	$D7,$76,$FC,$3E,$1B,$5A,$E0,$F3,$E9,$E9,$B5,$9E,$8F,$EC,$FC,$4F
	.dc.b	$EF,$B1,$FE,$7C,$5F,$1F,$9F,$67,$4E,$FD,$5E,$1E,$2D,$BE,$0F,$EF
	.dc.b	$7E,$B7,$F5,$9E,$2F,$F0,$ED,$D1,$FA,$D7,$FC,$18,$EA,$BF,$83,$46
	.dc.b	$BC,$7E,$5E,$D7,$3F,$84,$B5,$FE,$90,$BF,$F3,$6B,$7D,$5F,$E1,$9C
	.dc.b	$7F,$23,$1E,$B1,$87,$3E,$46,$BC,$8B,$7F,$E9,$AB,$FD,$45,$7F,$87
	.dc.b	$FE,$B5,$F2,$7F,$C5,$DF,$F2,$B5,$7F,$A8,$1C,$B7,$E0,$BF,$8E,$7D
	.dc.b	$6F,$E5,$E3,$FD,$4A,$BE,$17,$3E,$2E,$99,$3A,$F1,$6E,$5F,$E2,$FB
	.dc.b	$47,$43,$F9,$47,$D4,$3C,$3C,$5F,$FA,$CA,$30,$F8,$3E,$0B,$C7,$C9
	.dc.b	$E4,$6C,$79,$23,$9F,$9B,$8C,$BF,$C7,$D6,$8F,$E6,$DF,$55,$F2,$FA
	.dc.b	$3E,$D3,$F9,$2A,$FF,$22,$39,$C3,$EC,$B9,$9D,$F1,$7A,$3E,$67,$9B
	.dc.b	$6B,$FD,$93,$5F,$B5,$BE,$7E,$47,$83,$C2,$F0,$B2,$B7,$EC,$DA,$F0
	.dc.b	$BE,$63,$E2,$FD,$5D,$9F,$88,$FA,$BE,$73,$D7,$EA,$D9,$F0,$B8,$7E
	.dc.b	$57,$D6,$DC,$3E,$1B,$F0,$BD,$5F,$87,$E4,$5A,$F4,$B8,$7F,$DC,$7F
	.dc.b	$5B,$FD,$C1,$C9,$83,$E5,$F1,$7C,$CE,$1B,$1F,$09,$C3,$F1,$34,$B9
	.dc.b	$7C,$E7,$E9,$9F,$DA,$7E,$B7,$F6,$F1,$99,$FF,$79,$C6,$38,$FF,$61
	.dc.b	$F5,$BF,$96,$AF,$F3,$D3,$F9,$A6,$F7,$FF,$73,$8C,$4F,$0E,$EF,$C7
	.dc.b	$F1,$65,$B8,$FF,$C0,$0A,$FE,$0A,$DF,$5F,$97,$93,$CD,$91,$FE,$01
	.dc.b	$C3,$F8,$17,$57,$96,$3F,$7F,$7F,$D2,$73,$FF,$47,$4E,$5F,$F0,$44
	.dc.b	$3E,$73,$CD,$F3,$9E,$76,$DF,$8E,$3D,$5F,$17,$8B,$F7,$B8,$9F,$11
	.dc.b	$FC,$3B,$0F,$4C,$EA,$32,$FF,$85,$E3,$38,$FC,$AF,$2F,$E3,$7D,$37
	.dc.b	$C7,$6D,$FC,$38,$FF,$0E,$7C,$BB,$EB,$73,$8F,$FA,$A6,$31,$5F,$27
	.dc.b	$DC,$FF,$11,$70,$BD,$EF,$A5,$18,$64,$F9,$2F,$F6,$9F,$49,$87,$FD
	.dc.b	$69,$D2,$8F,$F1,$67,$D6,$FF,$2B,$9F,$F1,$83,$FC,$BF,$6F,$E6,$4E
	.dc.b	$60,$E6,$FF,$98,$B6,$3F,$C7,$1F,$5B,$FE,$AB,$9F,$14,$E6,$53,$97
	.dc.b	$FD,$8D,$1F,$D6,$EB,$BF,$B5,$E4,$F8,$6F,$2B,$30,$F6,$60,$F9,$1C
	.dc.b	$63,$FE,$CF,$8C,$4F,$83,$CA,$F6,$F2,$7B,$9B,$1E,$48,$FF,$26,$7D
	.dc.b	$6B,$E3,$F3,$87,$FB,$66,$31,$5F,$2F,$DD,$FF,$19,$3F,$D4,$11,$FE
	.dc.b	$55,$33,$3F,$E5,$67,$07,$6F,$9C,$B3,$F5,$69,$1C,$7C,$BE,$5E,$5D
	.dc.b	$AF,$03,$87,$86,$77,$1F,$33,$E1,$F5,$BC,$5E,$B5,$AF,$43,$87,$F0
	.dc.b	$9F,$58,$FA,$DC,$61,$FE,$F7,$8C,$4E,$BD,$3F,$EE,$53,$FD,$BA,$7F
	.dc.b	$A3,$BF,$E9,$90,$E1,$FF,$4E,$35,$E1,$E5,$72,$7F,$CF,$90,$8C,$47
	.dc.b	$26,$C6,$6B,$B4,$BD,$BF,$E3,$2F,$FA,$A5,$AF,$F8,$09,$AF,$FA,$25
	.dc.b	$AF,$1F,$0F,$FD,$33,$3C,$5B,$1C,$5A,$F1,$6B,$F7,$BC,$BB,$C4,$69
	.dc.b	$2B,$FC,$15,$3F,$E9,$6F,$3F,$64,$8C,$C7,$E2,$23,$AF,$D6,$7F,$41
	.dc.b	$8F,$9E,$D9,$FF,$63,$36,$3F,$C3,$ED,$7F,$D4,$CF,$F3,$78,$DF,$99
	.dc.b	$D0,$8F,$82,$3F,$9D,$3D,$3C,$2F,$09,$1D,$7C,$67,$F0,$DB,$FB,$D5
	.dc.b	$FE,$A7,$F0,$BE,$9A,$FA,$0F,$F4,$CC,$EA,$B8,$3F,$57,$C4,$FF,$1A
	.dc.b	$E1,$63,$E0,$EE,$9C,$C7,$D9,$6B,$FE,$C1,$F9,$FE,$75,$FE,$E7,$6C
	.dc.b	$7F,$90,$1A,$EF,$2E,$2B,$27,$9B,$86,$C7,$C3,$8C,$11,$C1,$0F,$A6
	.dc.b	$F1,$F8,$C9,$1F,$2D,$AF,$D8,$FA,$2E,$C4,$7E,$83,$E1,$F3,$A4,$7D
	.dc.b	$36,$BF,$80,$E4,$F3,$11,$FB,$0F,$A7,$CE,$FD,$1C,$7B,$5B,$89,$A6
	.dc.b	$BE,$55,$F2,$14,$95,$FA,$C8,$CB,$FA,$ED,$5F,$63,$D9,$B1,$98,$C1
	.dc.b	$9F,$3E,$47,$16,$BF,$C1,$67,$F9,$63,$D7,$C5,$E8,$47,$C5,$D5,$E7
	.dc.b	$23,$83,$CD,$E1,$FF,$9F,$BC,$BF,$F4,$26,$5F,$CC,$27,$FA,$F3,$54
	.dc.b	$6A,$78,$72,$7F,$A1,$A9,$18,$A0,$8C,$BD,$23,$4E,$7C,$4B,$BB,$E9
	.dc.b	$A9,$27,$FB,$84,$3E,$37,$64,$8C,$BD,$35,$FF,$3F,$47,$F7,$7F,$AD
	.dc.b	$7D,$2F,$81,$B1,$98,$C6,$F2,$38,$35,$F8,$3A,$FE,$97,$E0,$B9,$7C
	.dc.b	$D4,$9E,$CF,$5F,$8A,$91,$E1,$6B,$C8,$FD,$97,$C9,$F8,$FE,$1E,$47
	.dc.b	$9D,$F3,$5C,$8F,$53,$5E,$43,$4F,$EB,$EF,$BF,$E6,$FD,$C4,$7E,$17
	.dc.b	$C5,$FC,$12,$38,$3F,$C3,$92,$F3,$FE,$AF,$D6,$FE,$6A,$7B,$B7,$9F
	.dc.b	$FD,$94,$DD,$BF,$D8,$EF,$E1,$E3,$3D,$1C,$89,$5D,$BD,$3F,$C0,$FD
	.dc.b	$6F,$F8,$9E,$3F,$C5,$8D,$7D,$AD,$AF,$F6,$C2,$F8,$7E,$24,$9E,$2F
	.dc.b	$E3,$3E,$A7,$C9,$1F,$92,$97,$D4,$78,$09,$1E,$0F,$88,$F1,$91,$C1
	.dc.b	$FE,$3F,$FA,$D6,$5E,$8F,$52,$4F,$A1,$F5,$FC,$04,$8F,$5B,$5F,$5F
	.dc.b	$AB,$F0,$08,$FC,$07,$F4,$7F,$5B,$FD,$02,$3E,$EF,$D6,$FF,$BD,$F7
	.dc.b	$FF,$7C,$37,$8B,$FD,$EA,$D9,$FD,$7B,$FE,$E6,$7F,$CA,$CD,$79,$FA
	.dc.b	$9F,$B0,$8C,$52,$BF,$DF,$FF,$5B,$FE,$DB,$BE,$3F,$2D,$B2,$3D,$3E
	.dc.b	$27,$91,$1C,$1F,$E6,$5E,$3F,$B1,$7E,$6F,$3F,$C0,$49,$F2,$F9,$FE
	.dc.b	$32,$38,$3D,$3E,$7F,$9C,$8E,$0F,$F3,$97,$90,$ED,$F6,$7C,$B4,$9F
	.dc.b	$82,$FB,$FE,$26,$47,$C4,$6B,$FA,$AC,$FF,$0A,$FC,$8F,$9A,$F8,$C4
	.dc.b	$7E,$5F,$AD,$F2,$48,$E0,$FF,$42,$7D,$6B,$5F,$E3,$BE,$D7,$FA,$46
	.dc.b	$7F,$D1,$0D,$7F,$72,$DE,$3F,$A5,$1F,$CA,$23,$F8,$AB,$AF,$2F,$99
	.dc.b	$1E,$7E,$5F,$62,$38,$3F,$D2,$DF,$14,$F2,$FC,$7F,$49,$27,$C6,$F9
	.dc.b	$BE,$8E,$47,$99,$AF,$FA,$9F,$E1,$7D,$7E,$6C,$8F,$5B,$EE,$F3,$64
	.dc.b	$7B,$9A,$FF,$A9,$3C,$6F,$87,$E5,$C8,$F8,$6F,$C6,$F1,$64,$7C,$66
	.dc.b	$BF,$1B,$DD,$F9,$64,$7E,$5B,$FC,$FC,$FE,$97,$F5,$AE,$FA,$49,$FF
	.dc.b	$5C,$47,$F8,$59,$FE,$52,$FF,$4B,$73,$E9,$F2,$39,$F4,$7F,$56,$FA
	.dc.b	$D7,$C6,$52,$57,$FB,$0E,$1F,$19,$BD,$E4,$E5,$E5,$BF,$F6,$0F,$AD
	.dc.b	$FE,$4D,$E2,$7F,$66,$A4,$61,$FE,$D2,$D7,$3E,$D7,$1F,$95,$91,$FE
	.dc.b	$D3,$94,$75,$F9,$FF,$ED,$BE,$CB,$E1,$A9,$2F,$FD,$B5,$1F,$E4,$F6
	.dc.b	$F8,$BC,$08,$EB,$C5,$FE,$EF,$F5,$AA,$A4,$9F,$F7,$24,$7E,$3D,$BD
	.dc.b	$7C,$08,$EB,$8B,$9F,$8F,$23,$8B,$FE,$04,$FA,$DF,$89,$1F,$D7,$97
	.dc.b	$C7,$64,$91,$FE,$F3,$87,$C7,$7F,$83,$3E,$B7,$FD,$B5,$3F,$EF,$76
	.dc.b	$BF,$DA,$1B,$E3,$FE,$C2,$10,$DB,$FC,$2B,$E7,$FF,$8B,$A7,$FB,$7B
	.dc.b	$6B,$83,$79,$7E,$D3,$F1,$67,$FC,$43,$3A,$F0,$39,$92,$33,$5C,$3C
	.dc.b	$0E,$FE,$F4,$9C,$9F,$E2,$AC,$11,$E4,$DB,$FC,$53,$F5,$AF,$9B,$A4
	.dc.b	$BE,$22,$3E,$4B,$8F,$AD,$FE,$90,$9A,$4F,$5B,$FC,$6A,$FF,$98,$63
	.dc.b	$87,$C5,$60,$CF,$CA,$EB,$82,$4F,$57,$8F,$4C,$8F,$03,$5F,$F0,$CD
	.dc.b	$BE,$63,$BF,$8B,$27,$CC,$F8,$BC,$59,$1E,$96,$BC,$BD,$3F,$F2,$5F
	.dc.b	$D6,$FE,$46,$3F,$C4,$AD,$E6,$FB,$CD,$E5,$FD,$82,$F3,$78,$92,$BE
	.dc.b	$4C,$65,$FF,$12,$F2,$7F,$96,$3E,$6F,$FC,$1B,$06,$93,$E1,$34,$48
	.dc.b	$FA,$71,$FB,$BF,$86,$FE,$B1,$1F,$5B,$EB,$3F,$9F,$E7,$F9,$36,$E7
	.dc.b	$FC,$15,$B3,$FE,$6C,$7F,$20,$DF,$53,$FC,$D7,$18,$83,$4F,$CA,$3F
	.dc.b	$E4,$86,$D7,$7F,$13,$89,$FE,$74,$F8,$CF,$D8,$C7,$8F,$D4,$EA,$4B
	.dc.b	$F4,$6F,$E0,$23,$83,$D9,$E2,$FF,$A2,$17,$D6,$FC,$77,$D5,$19,$7F
	.dc.b	$7F,$92,$31,$1F,$05,$AF,$0C,$3F,$82,$BF,$0F,$C6,$7D,$6E,$23,$E0
	.dc.b	$C6,$1D,$DF,$F3,$A9,$D3,$63,$FD,$83,$F5,$1F,$EC,$24,$D2,$3D,$7E
	.dc.b	$8F,$FB,$BE,$1F,$C2,$63,$FD,$87,$C1,$FB,$05,$3A,$8F,$E3,$7F,$27
	.dc.b	$67,$FA,$03,$57,$A3,$FF,$00,$67,$94,$E6,$31,$3F,$E8,$01,$CB,$C5
	.dc.b	$FF,$58,$D7,$86,$31,$E8,$C8,$F2,$7C,$8D,$AE,$4F,$13,$B7,$CB,$D9
	.dc.b	$FE,$13,$82,$3C,$D9,$FF,$41,$9F,$E1,$AF,$F8,$35,$BC,$BA,$E4,$91
	.dc.b	$CB,$C0,$B9,$7F,$A6,$31,$47,$C1,$FA,$DF,$8F,$5F,$C6,$1F,$97,$5F
	.dc.b	$E9,$E7,$2D,$DB,$81,$F1,$7B,$AD,$7A,$1E,$9E,$3F,$5F,$3E,$CF,$57
	.dc.b	$0F,$7F,$3A,$E1,$DD,$F2,$7C,$6F,$97,$D4,$B5,$E5,$70,$F9,$FD,$4B
	.dc.b	$7E,$77,$D9,$F1,$9E,$DF,$8C,$B5,$ED,$7F,$01,$F3,$0F,$98,$BF,$31
	.dc.b	$86,$D7,$F3,$A9,$6F,$C3,$C9,$E2,$D9,$FE,$7F,$F5,$BF,$E1,$26,$FD
	.dc.b	$47,$E3,$63,$1A,$FF,$B2,$14,$67,$9B,$5F,$2D,$6F,$9B,$F9,$5F,$AD
	.dc.b	$FC,$6E,$1F,$ED,$05,$DB,$9E,$ED,$3B,$E2,$E1,$C5,$71,$FE,$A7,$68
	.dc.b	$FD,$07,$D6,$FF,$BE,$E7,$FA,$BB,$EE,$6F,$9F,$AD,$FF,$4D,$1F,$E3
	.dc.b	$23,$FA,$E9,$EC,$EF,$FE,$F4,$E7,$4F,$F7,$2B,$93,$FC,$5D,$F5,$A8
	.dc.b	$FD,$83,$F8,$C9,$FF,$01,$39,$9F,$ED,$2E,$5E,$3E,$17,$0D,$B8,$B6
	.dc.b	$37,$D1,$D3,$6E,$AD,$8F,$EE,$7E,$23,$D9,$FE,$00,$FF,$95,$E3,$31
	.dc.b	$8F,$85,$CF,$AB,$AF,$0E,$D7,$27,$F8,$47,$D6,$FE,$CF,$76,$FF,$E1
	.dc.b	$B8,$FD,$A4,$62,$BF,$C9,$D1,$B9,$D7,$8B,$CB,$DC,$E7,$F0,$F6,$FF
	.dc.b	$C1,$36,$8E,$FE,$EF,$F1,$7F,$2F,$E3,$65,$D4,$F9,$95,$FE,$B3,$BF
	.dc.b	$F0,$8C,$EB,$2F,$F1,$73,$FE,$3A,$8C,$F8,$BC,$FC,$6E,$8F,$F2,$3F
	.dc.b	$AD,$7B,$B7,$E0,$F0,$DB,$FF,$0E,$FD,$6F,$E6,$9B,$F0,$DF,$CA,$E3
	.dc.b	$1A,$F9,$42,$8D,$1B,$F0,$ED,$74,$70,$FF,$8A,$BE,$B7,$F3,$0E,$4C
	.dc.b	$78,$3D,$FE,$0F,$C1,$C9,$67,$CA,$E1,$E5,$E6,$5C,$FC,$98,$23,$FA
	.dc.b	$17,$D6,$FC,$7A,$FF,$42,$3F,$91,$AF,$CC,$39,$6F,$17,$03,$FA,$55
	.dc.b	$BE,$6B,$17,$E9,$7A,$F8,$3D,$BC,$7B,$3F,$E4,$6C,$7F,$76,$D8,$F1
	.dc.b	$FB,$DB,$8D,$38,$7E,$27,$D2,$5C,$FD,$6F,$AB,$E7,$BD,$7E,$9D,$AE
	.dc.b	$CE,$1F,$99,$F4,$96,$FE,$1B,$F0,$BD,$4F,$87,$E3,$5A,$F3,$38,$7F
	.dc.b	$CB,$BF,$5A,$71,$7F,$7E,$31,$1E,$DE,$2F,$9B,$C1,$F3,$9B,$1F,$06
	.dc.b	$3E,$8E,$98,$BF,$C4,$6E,$8C,$D7,$F9,$B3,$EB,$7F,$74,$FC,$98,$FE
	.dc.b	$F5,$18,$AF,$C8,$7D,$6F,$E6,$AB,$FC,$E6,$FF,$AC,$DB,$D7,$FE,$01
	.dc.b	$8C,$4F,$0E,$CF,$B7,$C5,$96,$DF,$2F,$1F,$87,$AD,$C5,$B1,$DA,$3F
	.dc.b	$C1,$9D,$3E,$3F,$0B,$FD,$3D,$97,$F4,$FA,$FF,$42,$CE,$5F,$F6,$94
	.dc.b	$3E,$77,$C9,$F3,$BE,$56,$CF,$AA,$3D,$1F,$19,$8D,$7D,$EE,$27,$AE
	.dc.b	$76,$7F,$65,$8F,$F1,$F7,$E2,$F9,$5F,$07,$E2,$BE,$13,$6F,$E3,$C7
	.dc.b	$F8,$97,$E5,$DF,$4D,$9B,$FD,$E4,$62,$BE,$6F,$B7,$FE,$2B,$E1,$7B
	.dc.b	$9F,$36,$30,$C9,$FF,$52,$F4,$BE,$B7,$18,$FF,$07,$8C,$4F,$FA,$AB
	.dc.b	$EB,$7F,$94,$3F,$EF,$77,$F9,$6F,$27,$FB,$25,$CD,$7F,$4D,$73,$3B
	.dc.b	$F4,$7F,$C7,$DF,$5A,$E3,$FF,$20,$56,$AF,$FA,$A4,$E6,$FC,$1F,$0F
	.dc.b	$AF,$97,$DB,$63,$C9,$1D,$3E,$4B,$17,$FC,$7E,$D1,$DB,$CB,$EF,$CD
	.dc.b	$70,$F6,$38,$7F,$D8,$FF,$5A,$F8,$FC,$63,$FC,$A6,$31,$3F,$EC,$CF
	.dc.b	$73,$FC,$6B,$C1,$FE,$E9,$7F,$CA,$71,$9F,$86,$FF,$2E,$46,$2F,$F9
	.dc.b	$93,$6B,$FD,$E1,$6F,$9C,$DC,$FF,$98,$E3,$18,$77,$F3,$B9,$F9,$DD
	.dc.b	$1B,$1E,$38,$E1,$AE,$32,$F4,$3E,$2F,$5F,$C7,$EB,$DB,$F5,$38,$7F
	.dc.b	$DC,$FF,$58,$FA,$DC,$61,$FD,$26,$31,$3A,$E9,$FF,$40,$4F,$F8,$2E
	.dc.b	$7F,$9D,$4F,$FA,$45,$B1,$FE,$96,$6B,$FE,$11,$3F,$C2,$5E,$E7,$F0
	.dc.b	$77,$07,$F6,$FF,$7B,$AB,$6B,$FD,$39,$3B,$8E,$4F,$EA,$DE,$2D,$73
	.dc.b	$E2,$72,$8C,$47,$F8,$87,$83,$F8,$E5,$7F,$7A,$8C,$79,$9F,$CA,$DF
	.dc.b	$F4,$E4,$E9,$B1,$99,$F3,$CB,$D2,$C5,$7B,$63,$13,$D6,$3F,$D2,$F3
	.dc.b	$A7,$C4,$D5,$F5,$9F,$CE,$1F,$F5,$63,$67,$E2,$B6,$3E,$53,$5F,$97
	.dc.b	$D3,$CF,$1F,$D0,$D7,$85,$B3,$FE,$B7,$6C,$74,$F0,$FC,$57,$FC,$CA
	.dc.b	$30,$7A,$5C,$57,$83,$E3,$7F,$86,$BF,$ED,$28,$FF,$53,$78,$7F,$ED
	.dc.b	$15,$F4,$3F,$EC,$29,$D5,$70,$7D,$7B,$FE,$0F,$0B,$1D,$E3,$0F,$B9
	.dc.b	$AF,$F9,$67,$E5,$79,$4F,$E2,$1B,$1F,$8D,$6B,$BC,$78,$BF,$3B,$CE
	.dc.b	$B9,$7F,$A4,$63,$DD,$18,$33,$F1,$B2,$3E,$13,$5F,$AA,$DC,$3E,$AB
	.dc.b	$63,$EA,$FD,$17,$62,$3F,$2D,$F1,$E2,$F1,$E9,$27,$F1,$11,$97,$E9
	.dc.b	$79,$DE,$82,$38,$3E,$B7,$91,$FF,$00,$5F,$C5,$B7,$3F,$E6,$11,$F5
	.dc.b	$63,$1C,$52,$3E,$C3,$5F,$D9,$8B,$7D,$6F,$56,$C6,$8D,$46,$4D,$79
	.dc.b	$B2,$35,$3F,$D6,$9B,$1C,$FC,$59,$1A,$6B,$E2,$97,$D1,$F8,$49,$1D
	.dc.b	$1F,$81,$F3,$0F,$0F,$F8,$77,$8D,$FB,$BA,$4B,$AF,$3C,$7F,$C4,$E3
	.dc.b	$13,$FE,$A5,$AE,$11,$CE,$31,$3B,$BE,$E4,$8F,$85,$3B,$7E,$4B,$FE
	.dc.b	$C6,$73,$E2,$79,$88,$FE,$9F,$E9,$BC,$6F,$FA,$D7,$85,$18,$8F,$E0
	.dc.b	$AD,$87,$C6,$D2,$4F,$98,$E4,$F1,$D7,$FD,$01,$19,$1F,$5B,$F3,$1C
	.dc.b	$35,$FF,$44,$37,$4F,$94,$FF,$65,$8D,$EF,$E5,$8C,$47,$08,$F0,$7C
	.dc.b	$26,$7E,$56,$4F,$67,$87,$3C,$8F,$0B,$5F,$F4,$C6,$2F,$92,$ED,$F0
	.dc.b	$F2,$7C,$EF,$87,$09,$1E,$A6,$BC,$BC,$DF,$37,$CD,$C8,$F7,$BE,$9F
	.dc.b	$8C,$91,$F0,$9A,$F8,$BC,$EF,$A5,$F5,$AF,$46,$92,$7E,$64,$7E,$29
	.dc.b	$FF,$58,$37,$87,$F2,$4D,$E8,$FE,$C4,$D9,$FE,$39,$E9,$7D,$FF,$AD
	.dc.b	$FF,$1A,$47,$D7,$6B,$FE,$BB,$6D,$FE,$75,$7C,$47,$12,$4F,$17,$F1
	.dc.b	$5F,$53,$E9,$8E,$9F,$51,$DC,$8F,$7F,$89,$F1,$11,$C1,$F8,$9F,$AD
	.dc.b	$65,$E7,$F8,$E9,$3E,$77,$D5,$EA,$C8,$F5,$35,$F5,$7C,$4F,$BC,$8F
	.dc.b	$BD,$FD,$07,$D6,$FF,$49,$8F,$F6,$D7,$D6,$FE,$EF,$7F,$DE,$37,$87
	.dc.b	$FC,$18,$D9,$07,$FD,$D6,$FE,$85,$AF,$4F,$89,$FD,$74,$62,$BF,$84
	.dc.b	$7D,$6B,$3C,$FA,$53,$0E,$4F,$0F,$66,$47,$56,$BF,$EF,$1E,$2F,$AF
	.dc.b	$9F,$9B,$D0,$EE,$49,$F2,$7C,$47,$88,$8E,$0F,$47,$A3,$E6,$23,$83
	.dc.b	$F7,$9E,$43,$B3,$D7,$F2,$B2,$7E,$03,$EE,$F8,$99,$1F,$0D,$AF,$9F
	.dc.b	$A7,$E1,$FC,$CC,$8F,$90,$FC,$6F,$88,$91,$F2,$DA,$B8,$FF,$98,$7D
	.dc.b	$6B,$6F,$E3,$5E,$DF,$F4,$1A,$D7,$07,$F9,$0E,$31,$7F,$CD,$DB,$19
	.dc.b	$F4,$F8,$92,$BF,$8D,$4F,$F1,$59,$E5,$E6,$76,$24,$E9,$FE,$55,$F1
	.dc.b	$5F,$DA,$E3,$C2,$7E,$1A,$9F,$49,$E6,$24,$76,$FA,$3F,$49,$1C,$1F
	.dc.b	$CA,$EB,$FC,$F4,$F9,$3C,$BF,$79,$1F,$3F,$97,$F0,$88,$E0,$FF,$59
	.dc.b	$F0,$FC,$6F,$26,$47,$B9,$F9,$3E,$1C,$8F,$82,$D7,$D6,$7F,$8C,$3E
	.dc.b	$03,$E7,$91,$F8,$EF,$F4,$33,$FA,$67,$D6,$BB,$69,$27,$FA,$34,$7F
	.dc.b	$89,$5E,$AF,$F5,$07,$3E,$9F,$23,$9F,$3F,$F5,$7F,$AD,$7C,$65,$25
	.dc.b	$7F,$51,$87,$C6,$6F,$79,$39,$79,$6F,$FD,$87,$EB,$7F,$93,$49,$A4
	.dc.b	$F6,$F5,$F2,$D2,$7F,$AF,$4A,$39,$7C,$F7,$F6,$BF,$65,$F0,$F4,$97
	.dc.b	$FD,$96,$3F,$CB,$2D,$F0,$B8,$11,$D7,$85,$FD,$D7,$EB,$55,$49,$3F
	.dc.b	$DB,$E3,$F2,$1E,$07,$F8,$4D,$75,$F1,$24,$EF,$8B,$A7,$89,$23,$93
	.dc.b	$FE,$03,$FA,$DF,$F5,$3E,$1F,$E1,$D5,$F1,$F9,$25,$7F,$7F,$87,$C7
	.dc.b	$FF,$82,$FE,$B7,$FD,$CB,$3F,$E0,$36,$BF,$8A,$6D,$FF,$8A,$56,$DF
	.dc.b	$E1,$3F,$37,$FB,$FC,$FF,$8B,$5B,$E2,$FF,$AF,$5B,$7F,$CC,$9F,$8A
	.dc.b	$FF,$84,$57,$81,$CC,$91,$9F,$03,$B7,$B7,$23,$83,$E1,$DB,$FC,$49
	.dc.b	$F5,$AF,$9C,$A4,$AF,$F0,$F4,$7C,$47,$FC,$41,$F5,$BF,$D1,$53,$49
	.dc.b	$EA,$FF,$8B,$DF,$F3,$3C,$7E,$B1,$BA,$FF,$92,$DB,$AF,$FB,$6D,$F8
	.dc.b	$ED,$DB,$91,$2B,$C3,$CF,$D4,$8E,$0F,$2F,$CC,$78,$7C,$39,$1E,$87
	.dc.b	$C9,$E1,$C8,$F5,$B5,$F3,$79,$FF,$E4,$7F,$AD,$FC,$7E,$5F,$EE,$F6
	.dc.b	$CF,$F6,$46,$A7,$F9,$0B,$FE,$45,$6B,$FC,$31,$BA,$7F,$9A,$DB,$3F
	.dc.b	$EA,$FE,$1F,$F2,$B7,$CD,$FF,$83,$6F,$F8,$9B,$6F,$FC,$E4,$DA,$FE
	.dc.b	$EB,$28,$FD,$E9,$3E,$2B,$99,$FE,$59,$E3,$FF,$31,$7D,$5F,$F3,$FA
	.dc.b	$FE,$FA,$D4,$D3,$F2,$2E,$3D,$0E,$24,$7F,$CC,$51,$F9,$47,$F9,$4B
	.dc.b	$71,$FF,$44,$37,$1F,$3B,$67,$FC,$1B,$F1,$7F,$B2,$9C,$FC,$47,$98
	.dc.b	$93,$DB,$D1,$F4,$91,$C1,$E2,$F0,$7F,$B4,$2F,$AD,$F8,$8F,$BA,$32
	.dc.b	$FC,$83,$DE,$E1,$F2,$B5,$E1,$97,$FA,$D7,$2F,$E6,$7F,$5B,$7C,$33
	.dc.b	$49,$5F,$EC,$9F,$A7,$FF,$65,$26,$91,$FF,$60,$7D,$23,$5F,$F4,$64
	.dc.b	$7F,$99,$5B,$19,$8F,$EA,$A6,$4B,$F1,$39,$CB,$FD,$6D,$E7,$3D,$7C
	.dc.b	$61,$EF,$AB,$7F,$E3,$CA,$D4,$F8,$39,$33,$3F,$E4,$0F,$3E,$BF,$EB
	.dc.b	$F9,$F2,$C6,$0C,$FC,$7D,$8F,$4B,$87,$FD,$4D,$6F,$97,$C5,$FF,$6E
	.dc.b	$8C,$4F,$FA,$AF,$3F,$E8,$2D,$E4,$F1,$C6,$2B,$FB,$AB,$FE,$FF,$9D
	.dc.b	$1D,$BB,$F7,$5E,$AB,$47,$B3,$EB,$5B,$DD,$B0,$DC,$3F,$D7,$93,$F3
	.dc.b	$71,$F9,$7D,$D9,$CF,$D0,$F3,$F1,$7A,$79,$97,$0E,$2E,$1E,$7E,$5D
	.dc.b	$BF,$F0,$6E,$3F,$D2,$A7,$BF,$C4,$F9,$7D,$2F,$33,$63,$C7,$1E,$9F
	.dc.b	$4B,$17,$C9,$F6,$7C,$1F,$B7,$E2,$2D,$FA,$9F,$B7,$F3,$0F,$98,$BF
	.dc.b	$31,$86,$D7,$FB,$6E,$5C,$1C,$3C,$9E,$2D,$9F,$F7,$07,$D6,$FF,$84
	.dc.b	$1B,$F5,$BF,$88,$8C,$6B,$FE,$2A,$51,$9E,$8D,$7C,$9D,$BE,$6F,$E4
	.dc.b	$3E,$B5,$DD,$7B,$B0,$DA,$FF,$78,$4B,$8B,$87,$17,$16,$CF,$FB,$D7
	.dc.b	$EB,$7F,$DF,$AD,$DB,$F3,$B1,$8D,$BF,$C7,$EF,$C0,$9F,$F7,$E1,$FE
	.dc.b	$30,$2F,$BF,$FA,$47,$13,$FE,$46,$7D,$FD,$7F,$B5,$87,$D6,$FF,$AA
	.dc.b	$27,$FD,$53,$39,$8F,$C8,$C6,$0F,$E7,$2E,$E7,$FC,$A2,$E9,$BE,$17
	.dc.b	$ED,$23,$E2,$CF,$19,$FF,$5A,$C7,$ED,$5E,$3A,$EF,$A3,$86,$96,$F8
	.dc.b	$57,$9A,$3F,$7F,$E2,$3D,$9B,$F0,$78,$6D,$F0,$9F,$F0,$CB,$7C,$3E
	.dc.b	$3E,$AF,$27,$F8,$4F,$D6,$FE,$DF,$7F,$88,$FE,$DF,$3F,$F3,$6B,$9B
	.dc.b	$FE,$B2,$E6,$77,$ED,$72,$F8,$5E,$6D,$AF,$4D,$7F,$AD,$5F,$5F,$8D
	.dc.b	$F2,$6F,$BF,$A7,$FC,$6A,$31,$3F,$D9,$9F,$F1,$EB,$C3,$C5,$D7,$8D
	.dc.b	$B7,$F9,$0F,$D6,$BD,$BB,$F0,$58,$6D,$FC,$AF,$AD,$FC,$EB,$7E,$17
	.dc.b	$F9,$4C,$63,$83,$FD,$16,$A3,$26,$BC,$2B,$5C,$DC,$3F,$5B,$EB,$7F
	.dc.b	$DD,$0E,$4C,$78,$1D,$BE,$07,$BF,$1D,$9F,$23,$87,$97,$97,$73,$FF
	.dc.b	$4E,$E0,$8F,$E7,$FF,$59,$F1,$A7,$FD,$FE,$FE,$47,$93,$F3,$8E,$60
	.dc.b	$E6,$FF,$C2,$2D,$80,$B7,$CB,$76,$E4,$FE,$9B,$18,$BF,$C9,$DF,$E7
	.dc.b	$2F,$FD,$62,$BC,$6F,$2F,$81,$E6,$65,$70,$F7,$6D,$F0,$7D,$67,$6F
	.dc.b	$A4,$B3,$F2,$DF,$7F,$CC,$FC,$0F,$4E,$CF,$89,$C3,$F4,$7E,$92,$E1
	.dc.b	$F2,$1F,$8F,$E9,$FC,$8F,$1A,$D7,$A9,$C3,$FA,$EF,$AD,$FE,$F8,$E4
	.dc.b	$C1,$F4,$38,$FE,$8F,$15,$8F,$86,$E1,$F8,$BA,$5C,$BE,$2B,$F6,$0F
	.dc.b	$EE,$FF,$5B,$FB,$99,$D9,$FB,$E7,$13,$FE,$DC,$FA,$DF,$CE,$4F,$F0
	.dc.b	$27,$FD,$72,$DE,$CF,$F0,$0C,$62,$78,$76,$FD,$8E,$1C,$B6,$F9,$79
	.dc.b	$1C,$3D,$5E,$2D,$8E,$D1,$FE,$0C,$EA,$F1,$F5,$7F,$27,$CB,$FA,$A5
	.dc.b	$7F,$14,$9C,$BF,$E5,$C8,$7C,$C7,$93,$E6,$3C,$AD,$9F,$54,$7A,$3E
	.dc.b	$9F,$1A,$FB,$DC,$4F,$5C,$EC,$FF,$BF,$B1,$FF,$7C,$4F,$C3,$F2,$7E
	.dc.b	$0F,$C5,$7C,$26,$C7,$C7,$8F,$F1,$2F,$CA,$BE,$97,$38,$36,$FE,$85
	.dc.b	$39,$BF,$9D,$ED,$7F,$8B,$38,$5E,$DF,$CE,$8C,$32,$7F,$9A,$F4,$3E
	.dc.b	$B7,$19,$75,$9F,$F3,$FC,$E5,$FE,$77,$F5,$BF,$CA,$E7,$F6,$2F,$F2
	.dc.b	$B8,$FF,$42,$B9,$83,$97,$7F,$03,$FE,$40,$FA,$D7,$1F,$F9,$0A,$B5
	.dc.b	$49,$CD,$F8,$3E,$1B,$5E,$D6,$DB,$1E,$48,$E9,$F5,$58,$BF,$51,$FC
	.dc.b	$1A,$D1,$DF,$CB,$F0,$73,$5C,$3D,$AE,$1F,$EB,$3F,$5A,$F8,$EC,$63
	.dc.b	$D0,$46,$67,$FA,$FF,$BB,$FE,$3B,$7F,$D4,$EF,$F6,$7E,$6F,$EF,$0E
	.dc.b	$6B,$F8,$CB,$99,$DB,$FE,$64,$A7,$D3,$FF,$99,$A3,$55,$FC,$82,$73
	.dc.b	$5E,$0F,$33,$5E,$66,$DB,$1E,$48,$E9,$A6,$2F,$D5,$7F,$EF,$CA,$F4
	.dc.b	$FA,$DD,$FD,$6F,$03,$63,$DB,$1F,$E7,$8F,$AC,$7D,$6E,$70,$F8,$91
	.dc.b	$8A,$FF,$00,$BF,$CC,$63,$FD,$44,$1C,$3F,$EA,$56,$BF,$EC,$E6,$BF
	.dc.b	$CF,$BD,$BF,$D4,$BC,$9A,$FF,$AB,$5A,$FF,$B5,$9A,$F1,$FA,$0F,$F1
	.dc.b	$43,$96,$B9,$AF,$2A,$94,$EA,$1B,$3D,$0F,$F5,$35,$FF,$AF,$63,$0F
	.dc.b	$F8,$B6,$7E,$EC,$E9,$FC,$1B,$5F,$EB,$3F,$59,$FC,$E1,$F2,$B6,$7F
	.dc.b	$D9,$0D,$8F,$F7,$8B,$5F,$89,$CB,$CD,$F4,$37,$1F,$CF,$D7,$8E,$72
	.dc.b	$D8,$7D,$1E,$C9,$19,$7A,$5F,$FA,$AF,$BF,$E2,$B6,$FE,$DB,$FE,$C1
	.dc.b	$F5,$F3,$F1,$BF,$C4,$27,$3F,$05,$FD,$D9,$F6,$B7,$1F,$6B,$63,$DA
	.dc.b	$D7,$E2,$F9,$35,$60,$8C,$36,$3D,$CD,$7E,$5F,$9B,$A2,$3A,$78,$FC
	.dc.b	$5C,$8E,$2D,$7E,$9F,$0F,$62,$3D,$9A,$E8,$C1,$0F,$C7,$FD,$5B,$C4
	.dc.b	$F5,$36,$3E,$A3,$5D,$C7,$BF,$E5,$7F,$D9,$31,$F6,$5A,$FE,$69,$BA
	.dc.b	$7C,$D8,$46,$1F,$C3,$F8,$BF,$9C,$7E,$4B,$73,$FE,$63,$1F,$72,$31
	.dc.b	$C7,$23,$EF,$B5,$FD,$CF,$D7,$3C,$2E,$8D,$8F,$D4,$B5,$DC,$78,$73
	.dc.b	$FE,$B5,$1F,$AE,$6B,$FC,$11,$AF,$AD,$AF,$E2,$5F,$E6,$4F,$1F,$0B
	.dc.b	$A1,$1E,$9F,$49,$DC,$8E,$0F,$07,$D0,$78,$C8,$E0,$FE,$03,$A7,$FB
	.dc.b	$11,$BE,$17,$C9,$8C,$47,$F1,$A6,$BF,$45,$CF,$89,$FE,$B5,$6E,$FF
	.dc.b	$A1,$6C,$FF,$96,$1F,$F2,$83,$FE,$A6,$5E,$A7,$E9,$EB,$97,$89,$6F
	.dc.b	$93,$6F,$95,$57,$F4,$BD,$77,$8B,$F7,$B1,$FE,$B2,$6B,$FD,$65,$7C
	.dc.b	$DE,$C9,$1B,$7A,$6F,$FE,$8B,$8F,$F5,$DF,$D6,$B7,$F0,$36,$33,$18
	.dc.b	$E0,$91,$C1,$AF,$F6,$9A,$7C,$1F,$2E,$19,$3D,$9E,$BC,$D2,$3C,$2D
	.dc.b	$79,$7C,$87,$8B,$E1,$64,$79,$9F,$2D,$C8,$F4,$B5,$F1,$79,$5E,$AF
	.dc.b	$9B,$91,$EE,$7D,$BF,$13,$23,$E0,$B5,$F5,$79,$9F,$47,$EB,$5F,$2D
	.dc.b	$49,$3F,$EE,$28,$FE,$92,$DF,$27,$81,$1F,$F7,$3E,$FF,$A9,$E9,$FF
	.dc.b	$7F,$7D,$6F,$F8,$D6,$7F,$DD,$6D,$7F,$C2,$4D,$8D,$FC,$3F,$E0,$E0
	.dc.b	$D2,$4F,$FB,$D3,$EA,$7C,$EF,$4F,$A8,$E0,$47,$BF,$C4,$73,$23,$83
	.dc.b	$FD,$FB,$F5,$BE,$7E,$FF,$E7,$D8,$BE,$37,$C4,$4A,$F4,$FB,$5E,$52
	.dc.b	$38,$3D,$BF,$11,$E9,$23,$83,$FD,$7F,$AB,$0F,$E9,$9F,$5B,$FC,$92
	.dc.b	$31,$3F,$EC,$4F,$AD,$AF,$F8,$DD,$B1,$86,$D7,$F5,$57,$FD,$E5,$3D
	.dc.b	$A3,$1F,$11,$FA,$D9,$CC,$F3,$8F,$D8,$7D,$6F,$BE,$38,$1E,$1F,$47
	.dc.b	$81,$1C,$8A,$FE,$AD,$87,$8F,$D1,$EC,$49,$FA,$DF,$01,$E1,$F4,$24
	.dc.b	$72,$BF,$F0,$AB,$7E,$23,$CC,$47,$06,$FE,$23,$D4,$47,$07,$93,$CA
	.dc.b	$77,$7B,$FE,$52,$4F,$77,$E1,$7C,$4C,$8F,$13,$5F,$F3,$27,$53,$AB
	.dc.b	$E4,$7C,$CC,$9F,$43,$F2,$FE,$1A,$47,$AD,$AF,$F9,$B7,$EB,$7D,$AF
	.dc.b	$F9,$BF,$E7,$FF,$A3,$F1,$FD,$C6,$F3,$FF,$83,$5B,$20,$FF,$15,$79
	.dc.b	$79,$BF,$C9,$E3,$15,$CE,$31,$E6,$CA,$FA,$F5,$FD,$41,$F1,$47,$F2
	.dc.b	$FF,$8A,$FE,$F1,$7E,$3F,$A4,$E8,$4A,$E3,$F4,$7D,$C8,$E0,$FC,$1D
	.dc.b	$7F,$9F,$9E,$FE,$6F,$94,$8F,$8B,$CD,$F4,$11,$C1,$FE,$BA,$E0,$F6
	.dc.b	$79,$72,$3D,$2F,$BF,$D8,$91,$EC,$6B,$F3,$7E,$07,$E2,$11,$F8,$2F
	.dc.b	$F4,$23,$FA,$77,$D6,$FE,$63,$87,$FC,$78,$D9,$FA,$0D,$4B,$DF,$E3
	.dc.b	$EB,$49,$FD,$1C,$FF,$52,$6F,$37,$F5,$BF,$AD,$7C,$65,$24,$FE,$AA
	.dc.b	$1F,$19,$BE,$09,$39,$79,$70,$7F,$65,$FA,$DF,$E6,$33,$FE,$B9,$6F
	.dc.b	$2F,$F0,$66,$ED,$FE,$57,$7F,$BB,$C9,$7C,$B7,$32,$47,$EF,$21,$F3
	.dc.b	$DF,$DC,$FD,$B7,$C8,$D2,$57,$F0,$28,$FC,$AB,$7C,$1E,$04,$75,$E0
	.dc.b	$FF,$7D,$FA,$D5,$52,$4F,$F0,$C8,$FC,$9E,$7F,$E6,$F5,$D7,$C4,$93
	.dc.b	$BE,$2E,$9E,$1C,$8E,$4F,$F8,$2F,$EB,$7F,$C9,$9C,$1F,$CA,$97,$C8
	.dc.b	$64,$95,$FC,$72,$1F,$21,$FE,$13,$FA,$DF,$F7,$7C,$FF,$20,$6B,$FE
	.dc.b	$E9,$6D,$FF,$36,$5B,$FF,$86,$FC,$FF,$D8,$4F,$F3,$A6,$ED,$FE,$8B
	.dc.b	$6D,$FE,$FD,$F8,$AF,$F2,$E5,$DF,$CC,$91,$9E,$FE,$DE,$F4,$8E,$0F
	.dc.b	$87,$6F,$F1,$67,$D6,$BE,$72,$92,$BF,$9D,$C7,$C4,$7F,$9E,$FD,$6F
	.dc.b	$F4,$C7,$F8,$93,$53,$4F,$E9,$7A,$7F,$56,$7F,$CE,$70,$8C,$7C,$6F
	.dc.b	$F8,$F6,$73,$1B,$F8,$DC,$6F,$2F,$C2,$EF,$A2,$47,$91,$AF,$5E,$0F
	.dc.b	$F5,$A6,$0F,$99,$ED,$E2,$4A,$F5,$BE,$1F,$0E,$47,$B9,$AF,$33,$DF
	.dc.b	$E7,$7F,$94,$FE,$B7,$F2,$71,$FC,$CD,$A9,$A3,$CB,$E0,$48,$FE,$CF
	.dc.b	$19,$7F,$C4,$BC,$7F,$E5,$FF,$8B,$FF,$0C,$C7,$84,$FF,$34,$FC,$36
	.dc.b	$4D,$7C,$2C,$8F,$EE,$33,$FB,$E7,$FA,$33,$57,$C2,$FF,$62,$7F,$BB
	.dc.b	$FD,$67,$F4,$6E,$0F,$F6,$D3,$63,$FA,$73,$53,$F9,$46,$FA,$3F,$E7
	.dc.b	$58,$C1,$3C,$4C,$12,$7B,$BF,$E7,$BF,$8E,$FD,$8D,$F8,$3E,$2F,$91
	.dc.b	$2B,$CB,$BF,$52,$38,$3D,$3E,$2F,$E2,$D7,$D6,$FF,$A3,$9F,$4C,$65
	.dc.b	$FF,$65,$65,$FC,$23,$AB,$FD,$99,$1A,$6B,$EE,$9E,$0F,$C4,$F1,$9F
	.dc.b	$5B,$AB,$FE,$72,$D8,$CC,$F4,$9C,$BF,$ED,$5F,$A7,$FF,$6B,$26,$91
	.dc.b	$F0,$FC,$B7,$F8,$FB,$1F,$F1,$EC,$7F,$9C,$35,$FF,$16,$A9,$D4,$7F
	.dc.b	$B9,$79,$2C,$FF,$81,$4E,$07,$E2,$7C,$FF,$F0,$3C,$F2,$9D,$CF,$FA
	.dc.b	$5B,$D8,$AB,$9A,$FC,$9E,$1B,$7E,$28,$DF,$C9,$E2,$BF,$2C,$1C,$CF
	.dc.b	$F9,$13,$17,$CB,$62,$BC,$8F,$8B,$4F,$E8,$35,$FE,$4A,$5C,$99,$E6
	.dc.b	$D3,$27,$A7,$83,$70,$F7,$3E,$CF,$AD,$F8,$D3,$FE,$54,$7E,$1E,$1F
	.dc.b	$E6,$87,$35,$FA,$D7,$33,$F2,$5B,$1F,$E2,$FC,$9D,$FD,$78,$3E,$4C
	.dc.b	$62,$FB,$78,$FE,$1F,$3F,$C4,$D8,$F1,$47,$97,$CE,$C6,$3E,$17,$D1
	.dc.b	$F1,$FE,$9F,$56,$DF,$A5,$C3,$EB,$F5,$2D,$FA,$5F,$81,$F1,$BF,$07
	.dc.b	$E3,$6D,$7C,$17,$EF,$FC,$C7,$F8,$5E,$3F,$CE,$EB,$E6,$33,$EF,$69
	.dc.b	$DF,$0F,$0F,$2A,$DF,$07,$F1,$3F,$5B,$FE,$16,$8F,$F4,$0A,$FA,$DC
	.dc.b	$FB,$DA,$77,$C9,$C3,$E5,$ED,$F0,$7F,$27,$F5,$AE,$FB,$DF,$86,$D7
	.dc.b	$FA,$3A,$7F,$C2,$6D,$E1,$E3,$C5,$C9,$FC,$EF,$D6,$BD,$1B,$C1,$86
	.dc.b	$DF,$FA,$60,$FF,$19,$97,$6F,$D2,$B8,$E5,$FE,$64,$FB,$DF,$E6,$9F
	.dc.b	$5B,$FE,$73,$8F,$F5,$0B,$F8,$CE,$EF,$F5,$AB,$99,$DF,$8B,$FB,$08
	.dc.b	$FD,$75,$70,$DB,$1E,$26,$5D,$75,$6F,$7B,$5D,$9F,$DC,$F8,$CF,$72
	.dc.b	$FC,$8E,$1B,$5F,$EB,$49,$FF,$08,$B7,$C4,$E3,$EB,$F2,$7F,$81,$FD
	.dc.b	$6F,$EC,$B7,$FE,$A0,$FE,$CA,$BF,$D7,$AE,$78,$3F,$DA,$2E,$67,$7E
	.dc.b	$EF,$2F,$87,$E6,$DB,$E3,$EF,$7F,$10,$E3,$F8,$98,$F2,$8C,$61,$9F
	.dc.b	$1B,$5E,$35,$C3,$FD,$9B,$87,$F8,$C9,$FF,$67,$FD,$6B,$E5,$7F,$8D
	.dc.b	$BF,$E5,$48,$CC,$63,$E1,$7F,$8F,$FD,$6F,$E5,$67,$FD,$AE,$BE,$23
	.dc.b	$3E,$26,$99,$3A,$F1,$2E,$BF,$DC,$ED,$1D,$0F,$E5,$7F,$5A,$F0,$B1
	.dc.b	$7F,$EE,$48,$C3,$E0,$F8,$5F,$1F,$1F,$91,$B1,$E4,$8E,$7E,$66,$32
	.dc.b	$FF,$01,$5A,$3F,$9C,$FD,$6F,$CA,$BF,$E8,$0F,$E4,$5B,$5A,$E1,$91
	.dc.b	$E3,$F9,$AB,$3D,$1E,$7E,$07,$4F,$22,$D7,$FB,$F3,$0F,$E7,$8D,$8F
	.dc.b	$0F,$C0,$6E,$34,$E1,$F8,$1F,$59,$71,$F4,$BE,$8F,$9E,$F4,$FA,$76
	.dc.b	$BB,$38,$7E,$47,$D2,$5B,$F8,$2F,$C0,$F5,$3E,$0F,$8F,$6B,$CC,$E1
	.dc.b	$FF,$73,$77,$9F,$EC,$9F,$5B,$5D,$9A,$32,$7C,$CE,$2F,$9B,$C3,$63
	.dc.b	$E1,$38,$7E,$26,$B7,$3F,$F1,$B7,$02,$3E,$A9,$FD,$BB,$EB,$7F,$6D
	.dc.b	$5C,$0F,$05,$7F,$BE,$E3,$2F,$F8,$FF,$EB,$7F,$39,$3A,$84,$7E,$75
	.dc.b	$D7,$77,$F7,$A8,$C6,$9E,$F9,$1C,$3B,$BF,$2B,$C7,$B6,$E1,$D6,$DE
	.dc.b	$4F,$3F,$5B,$A3,$67,$51,$FE,$08,$EA,$F3,$E5,$DF,$2F,$2C,$E7,$CA
	.dc.b	$F2,$7C,$DF,$97,$E6,$EE,$3D,$DC,$3E,$1F,$8E,$B8,$7C,$17,$D8,$77
	.dc.b	$CE,$CF,$C3,$CF,$CE,$FA,$7C,$BF,$85,$F1,$9F,$0D,$B1,$ED,$8F,$F0
	.dc.b	$FF,$CB,$BE,$B7,$39,$7F,$9B,$63,$15,$F0,$FD,$BF,$F1,$2F,$13,$DD
	.dc.b	$F9,$D1,$86,$4F,$D5,$EA,$7C,$FE,$31,$FF,$3C,$46,$27,$ED,$7D,$6F
	.dc.b	$F2,$97,$FC,$6C,$FF,$29,$E5,$F9,$4E,$6B,$EC,$B9,$9D,$FC,$17,$F8
	.dc.b	$E3,$EB,$5C,$7F,$E3,$AA,$D5,$FE,$02,$73,$7E,$0F,$88,$D7,$B7,$B6
	.dc.b	$C7,$92,$3A,$7C,$9E,$2F,$FD,$21,$68,$ED,$E6,$77,$E6,$B8,$7B,$1C
	.dc.b	$3F,$97,$FA,$D7,$C8,$63,$1F,$F4,$D4,$62,$7F,$39,$EE,$FF,$8B,$A7
	.dc.b	$FC,$A0,$FF,$95,$0C,$1F,$D9,$9F,$F5,$13,$78,$3F,$21,$4F,$A8,$E1
	.dc.b	$3F,$E5,$C8,$C5,$F6,$F3,$BA,$79,$DD,$5B,$1E,$28,$E1,$A6,$32,$F3
	.dc.b	$BE,$3F,$63,$C9,$EC,$5B,$F4,$B8,$7F,$6D,$F5,$8F,$AD,$C6,$1F,$EB
	.dc.b	$98,$C4,$EB,$CA,$FF,$3E,$CF,$ED,$A7,$F8,$44,$FF,$A4,$9B,$1F,$E9
	.dc.b	$66,$BF,$E7,$D1,$F5,$AF,$87,$E7,$18,$6C,$66,$BF,$D3,$AD,$9F,$F5
	.dc.b	$0E,$FF,$E2,$A6,$FC,$17,$28,$C4,$77,$6B,$FE,$8D,$F7,$FF,$95,$3F
	.dc.b	$E9,$66,$DF,$95,$B1,$B8,$FE,$28,$7A,$7E,$B7,$F8,$EC,$7F,$A7,$9B
	.dc.b	$19,$F3,$B7,$3F,$E9,$B9,$CB,$7C,$3C,$BF,$C7,$A3,$F9,$25,$75,$9C
	.dc.b	$CA,$72,$F3,$F8,$EF,$DE,$3F,$EC,$28,$FF,$4A,$F8,$7F,$EC,$25,$F1
	.dc.b	$FF,$EA,$F9,$D5,$70,$7D,$DE,$2F,$FA,$BF,$1B,$1F,$E3,$F3,$F9,$8E
	.dc.b	$69,$D4,$7C,$96,$BF,$CD,$BE,$5B,$A5,$7D,$D6,$C7,$FA,$ED,$AE,$F2
	.dc.b	$E2,$FB,$7C,$FB,$99,$A4,$8F,$20,$A7,$C9,$F2,$6C,$69,$AF,$FB,$49
	.dc.b	$B8,$79,$7C,$3E,$04,$9E,$0F,$8B,$CF,$91,$E0,$C5,$18,$8F,$C0,$36
	.dc.b	$BF,$08,$D8,$FC,$07,$9F,$EC,$23,$CD,$F7,$F8,$FF,$EF,$88,$FB,$CD
	.dc.b	$BF,$E4,$B1,$F6,$E3,$5F,$2C,$F9,$0A,$49,$FF,$80,$31,$97,$F5,$C7
	.dc.b	$D5,$FA,$EE,$DA,$1A,$48,$D9,$F3,$7C,$FD,$91,$CC,$FF,$51,$7F,$67
	.dc.b	$C5,$E9,$E9,$34,$91,$C3,$CA,$F0,$91,$C1,$FB,$CF,$98,$E4,$47,$07
	.dc.b	$E6,$FE,$2B,$E2,$11,$C1,$F0,$3C,$0F,$F0,$3A,$F2,$3F,$BE,$6B,$FC
	.dc.b	$6E,$3F,$C2,$4D,$75,$C7,$C2,$3F,$B0,$C6,$BD,$0F,$9E,$47,$E6,$BF
	.dc.b	$4B,$E0,$24,$7D,$06,$BF,$B3,$FA,$6F,$1E,$1F,$E0,$DE,$44,$63,$1F
	.dc.b	$E0,$8D,$AF,$E0,$5F,$1D,$5F,$E0,$0D,$AF,$F6,$14,$7F,$8A,$DA,$FA
	.dc.b	$8F,$D9,$7C,$47,$1F,$AE,$7D,$2D,$24,$FF,$8C,$E3,$FA,$EB,$7D,$0E
	.dc.b	$04,$7F,$C7,$1C,$BF,$E8,$67,$87,$CB,$F5,$23,$CF,$D6,$F0,$11,$C1
	.dc.b	$CB,$C9,$F2,$65,$23,$C4,$F9,$FE,$2A,$47,$95,$AF,$B4,$FE,$01,$E9
	.dc.b	$FB,$79,$E4,$7A,$DF,$81,$F1,$D2,$3D,$CD,$79,$66,$F4,$7E,$77,$D7
	.dc.b	$7E,$9E,$7B,$37,$97,$FC,$88,$DE,$5F,$C6,$AF,$37,$89,$2B,$FC,$BD
	.dc.b	$19,$7E,$4F,$AD,$F6,$FE,$B7,$F1,$B1,$EB,$6E,$1F,$E5,$16,$E1,$F9
	.dc.b	$A5,$D3,$C4,$95,$FE,$6B,$8C,$BF,$E6,$DF,$A9,$F3,$3F,$E6,$88,$7D
	.dc.b	$47,$84,$91,$C7,$E2,$BC,$84,$70,$7F,$9D,$FE,$B9,$9F,$A7,$E4,$24
	.dc.b	$F8,$5F,$67,$C0,$48,$F2,$35,$F7,$7C,$57,$C1,$23,$E9,$7F,$3D,$F5
	.dc.b	$DF,$D5,$63,$EF,$7D,$67,$F9,$CF,$83,$FC,$E8,$DE,$0F,$DE,$B6,$7F
	.dc.b	$D0,$AF,$FB,$A5,$FF,$47,$35,$E7,$EC,$7E,$AE,$31,$5F,$E7,$FF,$AE
	.dc.b	$FF,$6F,$57,$0F,$89,$D1,$1E,$7F,$13,$C4,$8E,$0F,$A1,$E3,$F5,$FA
	.dc.b	$E9,$1E,$27,$C1,$EA,$48,$F2,$B5,$F7,$FA,$9E,$52,$3E,$A7,$F8,$17
	.dc.b	$94,$EE,$F5,$79,$B2,$BD,$EF,$B7,$E2,$A4,$7C,$26,$BF,$A9,$EB,$FC
	.dc.b	$1B,$F1,$BE,$6F,$E2,$11,$F9,$3E,$DF,$C7,$23,$83,$FD,$75,$F5,$CD
	.dc.b	$7F,$8B,$7B,$DF,$D0,$E7,$F6,$CD,$E3,$FF,$56,$37,$9B,$F9,$AB,$FC
	.dc.b	$52,$7F,$D8,$8D,$78,$F9,$DD,$0E,$BE,$6C,$8D,$B5,$FF,$50,$79,$9E
	.dc.b	$2F,$A5,$91,$D1,$F2,$FB,$92,$3B,$B5,$FF,$54,$79,$7E,$AF,$36,$47
	.dc.b	$91,$F6,$F9,$B2,$3C,$ED,$7F,$D4,$BE,$37,$C2,$F2,$E4,$7B,$1F,$8B
	.dc.b	$E2,$C8,$F7,$B5,$F8,$FE,$B7,$CA,$23,$F1,$1F,$E8,$5E,$6F,$F6,$C8
	.dc.b	$FF,$75,$7D,$77,$FB,$79,$BE,$07,$F4,$78,$C4,$7F,$27,$7F,$C5,$4F
	.dc.b	$0F,$53,$44,$79,$F4,$FF,$4F,$FA,$E7,$BD,$49,$3F,$EF,$98,$7B,$3B
	.dc.b	$E0,$93,$97,$97,$07,$F5,$9F,$AE,$FE,$45,$3F,$CE,$5B,$D5,$FE,$EF
	.dc.b	$6F,$57,$F7,$85,$F1,$5C,$7E,$5A,$58,$E4,$F5,$F8,$6F,$EC,$DE,$CB
	.dc.b	$E1,$A9,$2B,$86,$5F,$EF,$D9,$FF,$31,$C6,$32,$FF,$03,$C3,$C5,$E6
	.dc.b	$49,$D1,$E1,$87,$8B,$FD,$D7,$EB,$91,$49,$3C,$04,$3B,$F8,$77,$49
	.dc.b	$D4,$FF,$95,$B2,$65,$FE,$1A,$8F,$A7,$3A,$F1,$FC,$04,$7B,$7D,$77
	.dc.b	$FA,$AE,$F8,$C7,$C7,$87,$C6,$7F,$82,$DB,$1C,$23,$1F,$19,$FE,$0F
	.dc.b	$FA,$EF,$F7,$34,$FA,$5A,$FF,$71,$6D,$F6,$84,$7C,$97,$E0,$7A,$1F
	.dc.b	$E0,$C7,$FB,$C3,$67,$83,$73,$FF,$1B,$BF,$2A,$78,$F8,$5F,$E2,$18
	.dc.b	$FE,$ED,$18,$3A,$F8,$59,$F7,$A4,$78,$1E,$1A,$FF,$8A,$7E,$B9,$F3
	.dc.b	$94,$95,$F5,$63,$E1,$BF,$E0,$DF,$AE,$FE,$98,$9A,$4F,$5F,$FC,$68
	.dc.b	$FF,$9D,$23,$87,$C5,$E0,$CF,$C9,$EB,$A6,$4F,$E0,$EB,$E3,$3D,$FA
	.dc.b	$79,$91,$D3,$DB,$E5,$FB,$F8,$D2,$79,$3E,$2F,$1A,$47,$56,$BE,$BF
	.dc.b	$4F,$FC,$95,$F5,$DF,$91,$8F,$CA,$35,$FF,$16,$37,$9B,$EE,$2F,$3F
	.dc.b	$89,$27,$8B,$FC,$53,$8F,$FC,$AD,$F3,$5F,$E1,$09,$34,$9F,$0D,$A2
	.dc.b	$47,$E9,$63,$F8,$0B,$FE,$C9,$E5,$FF,$30,$7D,$67,$F4,$3C,$BF,$1C
	.dc.b	$DE,$AF,$F3,$EB,$6B,$F8,$23,$F8,$B6,$FA,$7C,$88,$FE,$CE,$3F,$1C
	.dc.b	$FF,$92,$DB,$5D,$BC,$3F,$09,$FE,$71,$F8,$BF,$D6,$61,$E0,$F1,$3C
	.dc.b	$A4,$C3,$A6,$DE,$82,$38,$3C,$1E,$27,$FA,$4D,$7D,$77,$C6,$7D,$31
	.dc.b	$97,$FD,$32,$33,$FF,$6E,$E7,$FE,$B4,$6C,$7F,$AF,$1A,$F0,$7F,$9C
	.dc.b	$FD,$76,$FD,$03,$49,$3F,$EC,$FF,$A8,$FF,$68,$26,$91,$FF,$61,$FD
	.dc.b	$1F,$F0,$CC,$3F,$80,$47,$FB,$47,$5F,$E3,$CA,$75,$1E,$AF,$94,$B3
	.dc.b	$FE,$B4,$6A,$F4,$1E,$DE,$23,$A1,$FE,$7F,$8C,$D7,$6E,$6D,$7A,$1B
	.dc.b	$6C,$78,$A3,$97,$CD,$62,$FF,$D1,$01,$CC,$FF,$35,$B7,$CC,$62,$BC
	.dc.b	$0F,$9B,$4F,$E7,$B5,$FC,$ED,$FE,$74,$DE,$6D,$72,$C8,$E9,$E0,$5C
	.dc.b	$3D,$4F,$C0,$FA,$EF,$8D,$3F,$D1,$5F,$7E,$1F,$CA,$DC,$D7,$FA,$7D
	.dc.b	$CC,$FF,$B7,$1B,$1F,$D8,$F2,$77,$F5,$E0,$F9,$B1,$8B,$ED,$E4,$78
	.dc.b	$7C,$EF,$13,$63,$C5,$1E,$5F,$3B,$18,$F8,$5F,$47,$C6,$FA,$7D,$4B
	.dc.b	$7E,$97,$0F,$AF,$D4,$B7,$E9,$7E,$07,$C6,$7C,$1F,$8C,$B5,$F0,$5F
	.dc.b	$C1,$FC,$CB,$E6,$6F,$D7,$61,$B5,$FD,$AE,$7D,$98,$FF,$B0,$D7,$95
	.dc.b	$CB,$8B,$9B,$F8,$EF,$AE,$FF,$0B,$DF,$F7,$35,$F5,$39,$F7,$B4,$EF
	.dc.b	$93,$87,$CB,$5B,$E0,$FE,$63,$EB,$BF,$C2,$B1,$FD,$F5,$FC,$63,$7B
	.dc.b	$F5,$DF,$BE,$3E,$1C,$76,$B8,$3F,$A2,$FA,$E7,$A3,$78,$30,$DA,$FF
	.dc.b	$05,$1F,$E3,$42,$ED,$FA,$A7,$1C,$5F,$EE,$57,$DE,$FF,$B9,$FE,$BB
	.dc.b	$FC,$51,$1F,$E1,$37,$F1,$BD,$DF,$DE,$5C,$CE,$FC,$5F,$DA,$C7,$FB
	.dc.b	$76,$B8,$6D,$8F,$13,$2E,$BA,$B7,$BD,$AE,$CF,$EF,$BC,$67,$B7,$7E
	.dc.b	$17,$0D,$AF,$F1,$34,$FF,$8C,$5B,$E2,$71,$F5,$F9,$3F,$C2,$3E,$BB
	.dc.b	$F6,$7B,$B8,$3F,$C2,$51,$FB,$48,$C5,$FF,$91,$D4,$70,$9D,$F8,$FC
	.dc.b	$FD,$DE,$9F,$11,$70,$FF,$1C,$5A,$3C,$1E,$AF,$F1,$8E,$4F,$8F,$8F
	.dc.b	$82,$31,$96,$7C,$7D,$78,$F7,$2F,$F2,$16,$08,$FE,$43,$F5,$CF,$7A
	.dc.b	$FC,$26,$1B,$7F,$E4,$9F,$AE,$FC,$BB,$7E,$1F,$F9,$4C,$63,$3F,$79
	.dc.b	$FE,$65,$51,$A3,$7E,$25,$AE,$8E,$1F,$F2,$B7,$D7,$3C,$2C,$61,$D0
	.dc.b	$46,$67,$C9,$F0,$9E,$2E,$4F,$1B,$63,$D1,$1F,$CF,$A5,$E6,$E6,$FA
	.dc.b	$46,$2B,$FC,$CD,$F5,$9F,$31,$FB,$0F,$E5,$9B,$3F,$D2,$23,$0F,$FA
	.dc.b	$0D,$AF,$9B,$E6,$B5,$C5,$E1,$78,$78,$1C,$7C,$7B,$87,$81,$C3,$E2
	.dc.b	$F7,$EE,$1F,$07,$1F,$F0,$DC,$E4,$F8,$FF,$4D,$E3,$AF,$73,$E5,$F9
	.dc.b	$EF,$37,$A9,$6B,$DA,$E1,$F9,$FF,$5D,$71,$F8,$6F,$B7,$D5,$F7,$78
	.dc.b	$F6,$BE,$2B,$87,$FD,$19,$F5,$DF,$E8,$67,$26,$0F,$8D,$C5,$F1,$F8
	.dc.b	$2C,$7C,$E7,$0F,$D0,$D2,$E5,$EF,$7E,$71,$FD,$DB,$EB,$BF,$73,$3B
	.dc.b	$3F,$D3,$8E,$27,$F0,$DF,$5D,$F9,$B9,$FF,$50,$3F,$EB,$36,$F5,$FF
	.dc.b	$80,$23,$13,$C3,$B3,$F0,$DC,$39,$6D,$F2,$F2,$78,$7A,$FC,$5B,$1D
	.dc.b	$A3,$FC,$17,$D3,$E3,$F0,$FF,$D2,$99,$7F,$4B,$AF,$F5,$94,$E5,$FE
	.dc.b	$09,$0F,$9C,$F2,$7C,$E7,$95,$B3,$EA,$8F,$47,$C6,$E3,$5F,$7B,$89
	.dc.b	$EB,$9D,$9F,$D4,$62,$8F,$81,$E5,$7C,$1F,$8B,$B7,$F1,$9C,$3F,$EC
	.dc.b	$7F,$AE,$7D,$4E,$31,$FD,$94,$62,$7E,$6F,$B9,$FE,$2A,$E1,$7B,$BF
	.dc.b	$36,$30,$C9,$FF,$69,$74,$3E,$7F,$18,$FE,$F6,$31,$3F,$ED,$7F,$AE
	.dc.b	$FE,$4C,$FF,$29,$7F,$96,$72,$7F,$AF,$DC,$D7,$FB,$CD,$CC,$EF,$D1
	.dc.b	$FF,$1E,$FD,$73,$8F,$FC,$7F,$5A,$BF,$F7,$D4,$E6,$FC,$1F,$11,$AF
	.dc.b	$6B,$6D,$8F,$24,$74,$F9,$2C,$5F,$F1,$5B,$47,$6F,$2F,$BF,$3D,$C3
	.dc.b	$D8,$E1,$FF,$7A,$7D,$73,$E4,$31,$8F,$F1,$F8,$C4,$FF,$BE,$7D,$CF
	.dc.b	$F1,$7C,$9F,$EC,$8F,$F2,$57,$26,$38,$3F,$DC,$8D,$9F,$EA,$54,$F9
	.dc.b	$DF,$F3,$0C,$E6,$C2,$75,$5E,$1F,$37,$7E,$6F,$06,$C7,$96,$3F,$CD
	.dc.b	$72,$D3,$37,$EB,$7B,$FA,$DE,$0F,$5A,$E1,$ED,$70,$F8,$B4,$FF,$7E
	.dc.b	$7D,$73,$EB,$33,$8F,$F4,$38,$C5,$6C,$FE,$DC,$BC,$EF,$F4,$2C,$FB
	.dc.b	$23,$13,$FE,$9C,$8C,$4E,$94,$E9,$D6,$CF,$7F,$4D,$BE,$53,$F4,$E7
	.dc.b	$26,$BD,$59,$3A,$AF,$EB,$15,$D2,$7C,$1F,$59,$FC,$5E,$3F,$D3,$6D
	.dc.b	$AF,$F5,$8B,$63,$FC,$14,$D7,$EB,$BF,$C6,$A3,$7E,$01,$FF,$15,$CE
	.dc.b	$A3,$0F,$66,$C7,$6F,$8D,$FD,$CC,$FC,$5A,$FF,$49,$F8,$1E,$E7,$FD
	.dc.b	$4B,$EB,$5A,$DD,$79,$BE,$07,$FC,$3D,$85,$8F,$08,$BF,$F1,$8B,$86
	.dc.b	$BF,$EB,$AF,$9F,$E7,$3F,$ED,$06,$C7,$F8,$A5,$AF,$E5,$9A,$F5,$F3
	.dc.b	$7C,$2F,$C6,$6B,$F1,$9A,$EC,$F1,$7C,$5E,$2C,$D2,$47,$AA,$9E,$07
	.dc.b	$26,$C6,$63,$07,$3F,$36,$47,$BD,$AF,$DE,$F3,$7C,$04,$7E,$23,$5D
	.dc.b	$18,$21,$F8,$6F,$13,$F2,$2F,$CD,$6E,$3F,$C7,$E3,$EC,$46,$3C,$19
	.dc.b	$1F,$6D,$AF,$E8,$F1,$7D,$77,$46,$C7,$E7,$9A,$EE,$35,$E5,$C8,$D3
	.dc.b	$5E,$9E,$17,$32,$3C,$DE,$DE,$2C,$8E,$CD,$7F,$6D,$F5,$CF,$98,$F7
	.dc.b	$B6,$3F,$62,$D7,$71,$F0,$BE,$16,$47,$E6,$5A,$FE,$77,$D0,$F8,$E4
	.dc.b	$7F,$3C,$FC,$CF,$98,$FE,$AD,$1F,$BE,$6B,$FC,$25,$AF,$F9,$F4,$38
	.dc.b	$7F,$D0,$0B,$DD,$FF,$41,$78,$5F,$B2,$9D,$F9,$F8,$E8,$DB,$63,$F9
	.dc.b	$9D,$F1,$7F,$8F,$18,$BD,$3E,$06,$C6,$E3,$FC,$D0,$BD,$DD,$F2,$3E
	.dc.b	$9E,$92,$BA,$9F,$B7,$F8,$AD,$91,$CC,$FF,$B1,$9D,$F2,$7F,$9F,$9E
	.dc.b	$3F,$5C,$FA,$5F,$2B,$63,$31,$8E,$49,$1C,$1A,$FF,$9F,$E9,$F0,$9C
	.dc.b	$BE,$56,$4F,$67,$AF,$AB,$23,$C2,$D7,$97,$C8,$F8,$BE,$1E,$47,$99
	.dc.b	$F2,$DC,$8F,$4B,$5F,$17,$A9,$EA,$E8,$91,$EE,$7D,$BF,$15,$23,$E0
	.dc.b	$B5,$F5,$7A,$5F,$5B,$EB,$9F,$2B,$49,$3F,$42,$3F,$BE,$37,$CE,$E0
	.dc.b	$47,$E7,$FB,$1F,$80,$FA,$EF,$C9,$F6,$FE,$2D,$BD,$BF,$EC,$F6,$D7
	.dc.b	$F7,$C7,$F0,$93,$BE,$CE,$44,$9F,$C0,$FD,$57,$99,$FC,$1D,$3E,$A3
	.dc.b	$C0,$48,$CF,$C5,$F8,$C8,$E0,$FC,$5F,$D7,$33,$F4,$7A,$92,$7A,$BE
	.dc.b	$BF,$80,$91,$E0,$6B,$FE,$C1,$C5,$F1,$5F,$04,$91,$E6,$7F,$47,$F5
	.dc.b	$CF,$57,$F4,$BF,$5D,$FA,$FE,$0F,$D8,$37,$87,$FD,$E8,$DC,$3F,$C4
	.dc.b	$2F,$FB,$8D,$FD,$1B,$5F,$F6,$93,$67,$9F,$C5,$7E,$CB,$EB,$BF,$DB
	.dc.b	$15,$C3,$D7,$F0,$91,$E7,$F1,$3E,$42,$38,$3E,$8F,$97,$E8,$FA,$D9
	.dc.b	$1E,$27,$D7,$EB,$48,$F2,$B5,$F2,$FA,$DF,$00,$8F,$A9,$FE,$0F,$E6
	.dc.b	$3B,$FE,$2F,$A1,$2B,$DE,$FC,$8F,$42,$47,$C2,$6B,$FE,$F3,$EC,$FC
	.dc.b	$23,$F1,$BE,$67,$E7,$91,$F9,$3F,$13,$F4,$88,$E0,$FE,$1F,$F5,$CD
	.dc.b	$7F,$8D,$FD,$07,$FA,$42,$7F,$89,$B5,$72,$7F,$29,$8C,$37,$CF,$E2
	.dc.b	$49,$E2,$F4,$F3,$E4,$74,$6B,$FE,$9C,$F3,$7C,$3F,$4B,$23,$C2,$F9
	.dc.b	$3E,$92,$47,$91,$AF,$FA,$93,$CC,$F4,$F9,$D2,$3D,$2F,$B3,$CE,$91
	.dc.b	$EC,$6B,$FE,$A1,$F2,$3E,$0F,$9B,$23,$E0,$BF,$13,$C7,$91,$F1,$1A
	.dc.b	$FC,$4F,$07,$FC,$F2,$B8,$1F,$35,$F3,$09,$1F,$31,$FE,$93,$E9,$6A
	.dc.b	$3F,$A0,$FD,$77,$FB,$1D,$BE,$07,$F4,$E8,$C4,$77,$3F,$D0,$EB,$E0
	.dc.b	$75,$EB,$C9,$E6,$F1,$EA,$FE,$AF,$F5,$DF,$E2,$89,$E4,$1C,$74,$7F
	.dc.b	$65,$6E,$BE,$C5,$EE,$72,$E3,$97,$FD,$62,$7F,$A9,$3D,$1F,$E9,$2F
	.dc.b	$F5,$CF,$AE,$FE,$5A,$F8,$1B,$87,$F7,$36,$E1,$F2,$97,$C7,$FF,$6E
	.dc.b	$5F,$33,$49,$7D,$7E,$27,$FB,$9F,$B6,$F8,$9A,$49,$FE,$DB,$1F,$55
	.dc.b	$BE,$2F,$02,$3A,$F1,$7F,$BE,$FD,$72,$A9,$27,$FB,$BC,$3C,$0D,$F7
	.dc.b	$C9,$CB,$CB,$E1,$F9,$F9,$12,$78,$BF,$E0,$AF,$AE,$FF,$0D,$71,$7F
	.dc.b	$85,$57,$C8,$64,$95,$FE,$06,$87,$C8,$7F,$84,$BE,$BB,$FD,$C3,$3F
	.dc.b	$E0,$A6,$BF,$79,$B7,$FE,$21,$5B,$FF,$86,$BD,$1F,$F7,$15,$7F,$84
	.dc.b	$9A,$FE,$19,$B5,$FE,$29,$7E,$24,$6F,$C3,$E4,$47,$97,$87,$D7,$E0
	.dc.b	$64,$75,$7C,$1A,$FF,$8A,$BE,$BB,$FC,$4D,$3F,$E2,$26,$BF,$96,$6E
	.dc.b	$3F,$E3,$97,$E2,$C7,$F8,$A3,$EB,$BF,$A5,$26,$93,$D6,$FF,$1B,$BF
	.dc.b	$E7,$58,$FD,$0B,$6B,$FC,$88,$DE,$CF,$DD,$BF,$21,$BA,$F2,$25,$78
	.dc.b	$79,$FA,$91,$C1,$E5,$F9,$9F,$0F,$85,$23,$D0,$F9,$3C,$29,$1E,$B6
	.dc.b	$BE,$6F,$47,$FC,$9F,$F5,$DF,$8F,$8F,$F2,$3B,$5F,$DD,$B7,$4F,$F2
	.dc.b	$F2,$F3,$B8,$92,$78,$BF,$C4,$B8,$7F,$CB,$7F,$39,$FE,$11,$9F,$E0
	.dc.b	$AD,$BF,$F3,$43,$76,$FE,$3E,$FE,$E2,$41,$C3,$7E,$13,$A9,$FC,$83
	.dc.b	$8F,$FC,$D3,$F5,$BF,$E5,$59,$FE,$26,$DC,$3F,$CE,$CD,$C3,$F9,$6B
	.dc.b	$F9,$26,$FA,$5C,$89,$5F,$E6,$A9,$FC,$AB,$D3,$C1,$EE,$47,$FC,$DF
	.dc.b	$F1,$9F,$B0,$BE,$BE,$0F,$90,$93,$CF,$6F,$39,$1C,$1D,$FC,$1F,$EC
	.dc.b	$AB,$EB,$BE,$13,$F1,$32,$4E,$A3,$FD,$31,$18,$D3,$FD,$35,$1B,$9D
	.dc.b	$3F,$01,$AF,$3F,$09,$F5,$DC,$21,$7F,$EE,$68,$C3,$63,$AB,$5E,$1F
	.dc.b	$51,$FE,$DA,$4D,$23,$D7,$E0,$5A,$FF,$A5,$A3,$FC,$D6,$D8,$CC,$7F
	.dc.b	$4C,$32,$5F,$8B,$CE,$7F,$EB,$1F,$5F,$FB,$35,$F3,$9C,$BF,$DA,$A3
	.dc.b	$35,$CF,$93,$2F,$F6,$EF,$36,$BF,$90,$8A,$FF,$48,$AF,$34,$E3,$E7
	.dc.b	$32,$D8,$F5,$47,$2B,$7C,$BE,$6F,$FB,$2C,$62,$BF,$D6,$B9,$7F,$43
	.dc.b	$6F,$1F,$8A,$31,$5F,$DB,$1F,$E8,$93,$A3,$FD,$7B,$4E,$FE,$17,$E1
	.dc.b	$7D,$7F,$5D,$FE,$69,$AF,$F6,$1B,$F2,$AF,$FD,$B0,$E7,$0F,$EF,$2E
	.dc.b	$67,$FC,$22,$D8,$FD,$9E,$0E,$EE,$BB,$7B,$DE,$FF,$25,$E0,$F3,$2E
	.dc.b	$3E,$07,$0F,$8B,$CB,$B8,$F8,$1F,$3F,$C6,$FA,$3D,$1B,$5E,$87,$0F
	.dc.b	$AB,$D1,$B7,$E8,$7D,$FF,$19,$F0,$3E,$32,$D7,$C0,$7E,$EF,$CC,$FF
	.dc.b	$87,$63,$FD,$CA,$BE,$67,$3E,$D6,$9D,$EF,$C3,$C9,$B7,$C1,$FC,$3F
	.dc.b	$D7,$7F,$86,$A3,$FD,$DE,$BE,$BB,$3E,$E6,$9D,$F2,$F0,$F9,$4B,$7C
	.dc.b	$1F,$C8,$FD,$77,$E3,$F8,$7F,$C5,$6B,$B7,$3D,$DA,$77,$C7,$C3,$8E
	.dc.b	$E3,$FF,$00,$2D,$1F,$9C,$FA,$E7,$A1,$78,$70,$DB,$FE,$5F,$DB,$FB
	.dc.b	$0D,$3F,$6B,$E3,$7E,$92,$7D,$D1,$FE,$3E,$51,$9E,$1F,$D4,$FD,$77
	.dc.b	$F3,$89,$E6,$7F,$96,$9F,$C7,$5F,$57,$57,$FE,$48,$73,$3B,$F0,$F9
	.dc.b	$F0,$74,$6C,$74,$AF,$E8,$B5,$FE,$3A,$8C,$42,$1B,$F9,$3A,$BC,$AF
	.dc.b	$EF,$BC,$57,$B3,$7E,$0B,$0D,$C7,$9C,$FF,$87,$1B,$E2,$71,$F5,$79
	.dc.b	$3F,$C2,$3E,$BB,$F6,$FB,$FC,$97,$F6,$FA,$FF,$99,$1C,$DF,$F9,$F1
	.dc.b	$CC,$EF,$DA,$E5,$F0,$BC,$DB,$5E,$CA,$FF,$5F,$3E,$EF,$2F,$E4,$DF
	.dc.b	$7F,$53,$F8,$CC,$62,$7F,$D1,$2F,$F9,$01,$E1,$E3,$6B,$C7,$DB,$FC
	.dc.b	$83,$EB,$9E,$DD,$F8,$1C,$36,$FE,$6F,$D7,$7E,$75,$BF,$0B,$FC,$A2
	.dc.b	$31,$C1,$FE,$82,$51,$93,$5E,$25,$AE,$6E,$1F,$B3,$F5,$DF,$98,$72
	.dc.b	$63,$C2,$ED,$F0,$3D,$F8,$EC,$F9,$1C,$3C,$BC,$DB,$9F,$FA,$53,$04
	.dc.b	$7F,$3E,$FA,$EF,$8D,$3F,$E0,$E7,$F2,$2D,$35,$C5,$2F,$CB,$F3,$16
	.dc.b	$7A,$3C,$FC,$1E,$9E,$45,$AF,$C9,$E1,$FB,$96,$C7,$87,$DE,$DC,$69
	.dc.b	$C3,$F0,$BE,$B6,$E3,$E9,$7D,$1F,$39,$E9,$F5,$2D,$76,$70,$FC,$AF
	.dc.b	$A6,$B7,$F0,$5F,$81,$EA,$FC,$1F,$1A,$D7,$99,$C3,$FA,$AF,$AE,$FE
	.dc.b	$EA,$E4,$C1,$F2,$78,$7E,$57,$05,$8F,$7B,$87,$E0,$E9,$72,$FF,$62
	.dc.b	$EE,$8F,$A4,$7F,$70,$FA,$EF,$D9,$D7,$03,$F7,$6E,$2B,$FD,$9B,$F5
	.dc.b	$DF,$96,$9F,$DF,$BF,$EB,$56,$F5,$FF,$7A,$8C,$4F,$0E,$CF,$EF,$BC
	.dc.b	$DF,$E4,$BB,$E3,$18,$E3,$E1,$E4,$71,$F5,$ED,$76,$70,$FF,$0E,$E9
	.dc.b	$F2,$78,$7F,$C9,$F3,$FE,$99,$9F,$FB,$8A,$73,$5F,$EF,$A9,$7C,$D7
	.dc.b	$97,$E6,$BC,$CD,$9F,$5C,$7A,$7E,$A3,$18,$7C,$07,$13,$E1,$CE,$CF
	.dc.b	$77,$E0,$79,$3F,$07,$E2,$AD,$7C,$67,$0F,$F2,$6F,$AE,$7D,$3E,$30
	.dc.b	$FF,$7C,$C6,$27,$E6,$FB,$9F,$E2,$4E,$27,$BD,$F3,$63,$0C,$9F,$E6
	.dc.b	$3D,$4F,$9E,$C6,$2E,$3F,$E9,$93,$9A,$FE,$6F,$F5,$DF,$C9,$2B,$FC
	.dc.b	$FE,$FF,$24,$8F,$F4,$0C,$23,$4D,$F8,$3F,$F1,$BC,$6E,$3F,$D0,$FF
	.dc.b	$5C,$E2,$CD,$7F,$8F,$23,$17,$E0,$F6,$B8,$FB,$5C,$9B,$1E,$48,$E1
	.dc.b	$F5,$78,$C7,$D5,$3E,$4B,$4F,$8B,$CB,$F1,$F5,$5A,$F6,$B8,$7F,$AA
	.dc.b	$FD,$73,$E3,$F1,$8F,$43,$FD,$3F,$19,$9F,$EB,$7E,$EF,$F8,$C6,$4C
	.dc.b	$8C,$FC,$37,$F9,$62,$31,$3C,$CD,$BF,$E5,$CA,$7C,$EE,$E3,$0F,$1F
	.dc.b	$3F,$97,$9F,$71,$F0,$B8,$7D,$3A,$5C,$7B,$62,$8F,$17,$AF,$E3,$F5
	.dc.b	$ED,$FA,$9C,$3F,$DD,$BE,$B5,$F5,$D8,$C7,$D1,$18,$9F,$EF,$7F,$5A
	.dc.b	$F8,$7F,$F0,$A5,$7E,$12,$31,$3F,$E9,$76,$C7,$FA,$81,$AF,$E2,$8F
	.dc.b	$93,$EB,$F2,$7F,$CF,$CD,$73,$3F,$EA,$1E,$0F,$EA,$51,$FE,$AA,$7F
	.dc.b	$A4,$C7,$F8,$62,$3F,$D0,$DD,$7F,$E9,$68,$ED,$3A,$9F,$3B,$63,$C5
	.dc.b	$F5,$BF,$C5,$1B,$E0,$F1,$AF,$CF,$CE,$5B,$19,$7F,$DC,$71,$FE,$8C
	.dc.b	$F0,$3F,$D6,$CD,$8F,$F5,$FA,$FA,$07,$B3,$AF,$8E,$FD,$CD,$6B,$E1
	.dc.b	$3F,$5B,$97,$91,$B1,$E2,$6B,$E0,$6A,$C1,$E5,$61,$B1,$F1,$4D,$46
	.dc.b	$4D,$79,$D2,$35,$28,$C4,$7D,$E8,$FF,$3A,$3A,$F8,$DE,$C4,$78,$FD
	.dc.b	$1F,$84,$8E,$0E,$BE,$7F,$94,$8E,$0F,$0F,$9F,$E9,$23,$83,$CB,$E3
	.dc.b	$FF,$B8,$DF,$80,$DC,$3F,$87,$47,$D2,$8C,$11,$C1,$0F,$AB,$E1,$F1
	.dc.b	$23,$F6,$9E,$DC,$CF,$0E,$92,$7F,$3C,$F8,$FE,$87,$F9,$CC,$7F,$BD
	.dc.b	$DA,$FF,$C0,$06,$BF,$A1,$6B,$FC,$B1,$FE,$24,$F1,$F1,$3A,$11,$E9
	.dc.b	$F4,$7D,$C8,$E0,$FD,$87,$CC,$FA,$08,$E0,$FC,$AF,$C4,$7C,$72,$38
	.dc.b	$3F,$33,$E9,$FC,$B2,$38,$21,$FC,$6F,$5F,$F6,$F3,$FE,$7E,$6B,$9F
	.dc.b	$4F,$FA,$3F,$9F,$FE,$99,$F9,$87,$C1,$63,$A7,$FA,$5C,$7F,$55,$BF
	.dc.b	$F4,$3B,$5F,$E5,$0F,$07,$33,$F8,$18,$C4,$7F,$0B,$0E,$1F,$E1,$CD
	.dc.b	$7B,$FC,$17,$12,$3C,$4E,$2F,$E2,$FE,$9B,$CD,$1F,$BE,$6B,$FA,$F6
	.dc.b	$E5,$FE,$BA,$8F,$85,$F1,$D8,$23,$FE,$A2,$E1,$FF,$38,$BF,$EA,$4F
	.dc.b	$AD,$70,$F4,$6C,$66,$31,$C5,$23,$83,$5F,$E9,$94,$F8,$6E,$5F,$37
	.dc.b	$27,$B3,$D7,$D6,$91,$E1,$6B,$CB,$C9,$F1,$7C,$44,$8F,$33,$E5,$B9
	.dc.b	$1E,$96,$BE,$2F,$4F,$D5,$D1,$23,$DC,$FB,$7E,$2E,$47,$C1,$6B,$EA
	.dc.b	$F4,$7E,$6F,$D6,$BD,$6A,$49,$FF,$65,$C7,$F7,$06,$F9,$BC,$08,$FC
	.dc.b	$FE,$EF,$AF,$F5,$9F,$8B,$EC,$FF,$75,$37,$B3,$FC,$36,$DA,$07,$F1
	.dc.b	$B3,$C3,$B3,$99,$27,$FD,$BB,$F5,$3E,$77,$FD,$BF,$4F,$A8,$F0,$92
	.dc.b	$33,$F1,$9E,$42,$38,$3F,$DC,$FF,$5B,$FC,$3D,$E7,$EA,$7A,$88,$F6
	.dc.b	$F5,$3D,$A4,$70,$78,$BE,$2F,$E0,$91,$C1,$FE,$F4,$FA,$DF,$EA,$6F
	.dc.b	$FB,$DB,$EB,$5B,$FE,$79,$B1,$98,$C7,$C5,$48,$E0,$D7,$FC,$3D,$F5
	.dc.b	$AC,$B9,$7A,$92,$7B,$3D,$7D,$49,$1E,$16,$BF,$E2,$7E,$1F,$A8,$B2
	.dc.b	$7A,$5E,$52,$47,$A7,$D2,$F4,$11,$C1,$ED,$F4,$BD,$84,$70,$7E,$03
	.dc.b	$9B,$DE,$BC,$A7,$6F,$C3,$F3,$E4,$FC,$77,$E3,$79,$F2,$3E,$53,$5F
	.dc.b	$57,$47,$CB,$F9,$99,$1F,$3D,$F9,$DF,$0F,$23,$E9,$35,$FF,$24,$7D
	.dc.b	$6F,$C0,$7C,$DC,$DF,$E7,$2F,$6B,$FA,$6E,$BE,$86,$C6,$7E,$2B,$F8
	.dc.b	$BC,$7F,$0C,$AF,$F2,$93,$6B,$87,$9D,$A3,$7E,$74,$AE,$E7,$E9,$2B
	.dc.b	$FA,$AB,$CA,$3F,$91,$FC,$53,$CD,$EF,$F4,$92,$B8,$9F,$E2,$17,$C7
	.dc.b	$F4,$72,$33,$82,$31,$1F,$EA,$0F,$33,$D3,$E7,$48,$DB,$EC,$F3,$A4
	.dc.b	$72,$6B,$FE,$6F,$F1,$3F,$64,$F6,$F3,$7E,$19,$1F,$0F,$95,$F1,$88
	.dc.b	$E0,$F2,$FC,$CF,$CA,$23,$83,$F9,$E3,$FD,$13,$EB,$7F,$C5,$F1,$FE
	.dc.b	$7F,$6F,$17,$E3,$1B,$C9,$FE,$AA,$7F,$C5,$D3,$F1,$9F,$E9,$4F,$1F
	.dc.b	$57,$A1,$D7,$A3,$FA,$8F,$D6,$BE,$36,$92,$BE,$7C,$3E,$37,$7C,$12
	.dc.b	$72,$F2,$E0,$FE,$B7,$F5,$BF,$CA,$E7,$FD,$2C,$DD,$7F,$46,$DD,$7F
	.dc.b	$D8,$2B,$E3,$78,$FC,$C4,$BF,$BF,$28,$EB,$F3,$DF,$DA,$3D,$A7,$C4
	.dc.b	$D2,$5F,$E1,$A3,$FC,$CC,$DF,$13,$81,$1D,$78,$9F,$DC,$BE,$B5,$54
	.dc.b	$93,$F9,$08,$FC,$A3,$7A,$F8,$11,$D7,$1F,$3F,$1A,$47,$17,$FB,$FF
	.dc.b	$D6,$FF,$A5,$F8,$7F,$82,$AF,$8F,$C9,$2B,$F3,$F0,$F8,$FF,$F0,$3F
	.dc.b	$D6,$FF,$B5,$27,$F4,$6D,$7F,$D8,$8D,$BF,$E2,$2B,$6F,$F0,$7F,$A5
	.dc.b	$FB,$09,$FE,$2A,$DE,$5F,$F7,$9B,$6F,$FA,$CB,$F1,$1F,$D8,$BF,$EC
	.dc.b	$36,$F7,$F4,$3A,$F7,$F6,$F7,$E4,$F0,$7C,$3B,$7F,$88,$3E,$B5,$F0
	.dc.b	$B4,$95,$FB,$E8,$F8,$8F,$F0,$0F,$AD,$FE,$57,$A7,$F2,$C6,$E9,$FD
	.dc.b	$BA,$51,$87,$FB,$8A,$F6,$7F,$C5,$EF,$F9,$F2,$3F,$DD,$4D,$AE,$9F
	.dc.b	$17,$AF,$8B,$DF,$04,$BF,$13,$CB,$A6,$47,$95,$AF,$4F,$8A,$EF,$E2
	.dc.b	$C8,$F5,$3E,$2F,$16,$47,$B5,$AF,$93,$D2,$FF,$22,$FD,$6F,$E4,$23
	.dc.b	$FE,$00,$37,$4F,$E9,$AD,$DF,$E1,$2F,$37,$89,$2B,$F9,$4C,$65,$FF
	.dc.b	$25,$72,$7F,$94,$BE,$6B,$FC,$21,$0C,$3F,$AC,$C6,$1B,$F0,$FB,$F8
	.dc.b	$69,$5F,$CD,$2B,$F7,$8F,$F8,$81,$AB,$E2,$7F,$B0,$BF,$CE,$BE,$AF
	.dc.b	$FA,$46,$9F,$65,$B1,$FE,$28,$6A,$7F,$28,$DF,$57,$FC,$D3,$18,$27
	.dc.b	$8B,$82,$4F,$77,$FC,$DF,$F1,$BF,$B1,$BF,$07,$AB,$C8,$95,$E5,$E0
	.dc.b	$EA,$47,$07,$A7,$C6,$F8,$CB,$EB,$5A,$FA,$E7,$33,$F9,$53,$D4,$E1
	.dc.b	$D7,$07,$FA,$1F,$4F,$EB,$0D,$F1,$DF,$5B,$49,$5F,$E9,$5A,$FE,$71
	.dc.b	$1F,$ED,$98,$7D,$3D,$23,$FE,$DD,$7F,$DB,$DF,$2B,$FE,$32,$C3,$FC
	.dc.b	$2B,$3F,$E6,$5D,$BF,$B6,$A9,$D4,$7A,$79,$2C,$FE,$BD,$AF,$FA,$1F
	.dc.b	$D9,$AF,$33,$F1,$53,$FB,$5F,$49,$3A,$E6,$DF,$A5,$6B,$C2,$E1,$F1
	.dc.b	$FC,$D5,$C3,$99,$FB,$F7,$B9,$E2,$C1,$F2,$D8,$AF,$F0,$3C,$61,$FF
	.dc.b	$04,$E9,$FC,$F9,$BC,$BE,$A8,$C6,$1D,$8F,$D1,$46,$4E,$DE,$05,$AE
	.dc.b	$AF,$C2,$FA,$DF,$91,$1F,$E1,$67,$E1,$5F,$F9,$09,$CD,$F5,$73,$3F
	.dc.b	$C6,$5B,$1F,$E8,$8C,$1D,$FD,$77,$F7,$3D,$FC,$8F,$07,$A1,$71,$F0
	.dc.b	$38,$7C,$5E,$7D,$C7,$C0,$F9,$FE,$3B,$D1,$EA,$DA,$F4,$38,$7D,$5E
	.dc.b	$AD,$BF,$43,$EF,$F8,$AF,$81,$EE,$DA,$F8,$0F,$E1,$3E,$61,$F3,$17
	.dc.b	$E6,$30,$DA,$FF,$1D,$CF,$B9,$BE,$57,$1E,$0E,$4F,$E3,$BE,$B7,$FC
	.dc.b	$25,$3F,$E4,$25,$F5,$B9,$F7,$F4,$EF,$97,$87,$CB,$DB,$E0,$FE,$63
	.dc.b	$EB,$7F,$C2,$11,$FE,$4C,$5D,$B9,$EE,$D3,$BE,$2E,$1C,$56,$F8,$3F
	.dc.b	$A2,$FA,$D7,$A3,$78,$30,$DA,$FF,$2C,$9F,$E2,$F2,$ED,$FA,$A7,$1C
	.dc.b	$BF,$91,$7D,$EF,$E4,$FE,$B7,$FD,$97,$1F,$E6,$27,$F1,$7D,$DF,$E8
	.dc.b	$E7,$33,$BF,$13,$F6,$B0,$F0,$EE,$F8,$6E,$1C,$FA,$B7,$BF,$06,$C7
	.dc.b	$F9,$C3,$C5,$7B,$BF,$C0,$17,$C9,$D8,$C3,$9F,$5B,$5E,$2D,$BE,$4F
	.dc.b	$F0,$8F,$AD,$FD,$7F,$07,$FA,$99,$FF,$40,$4F,$F9,$F9,$CB,$FD,$F5
	.dc.b	$CF,$07,$ED,$57,$BD,$CF,$E2,$3A,$36,$B9,$7B,$FF,$C5,$B8,$FE,$2E
	.dc.b	$3D,$E3,$18,$E7,$C6,$D7,$8D,$70,$FF,$47,$E1,$FE,$31,$7F,$D2,$3F
	.dc.b	$5B,$FE,$0B,$6F,$C2,$FF,$22,$8C,$61,$FC,$25,$FF,$2A,$CF,$FA,$63
	.dc.b	$EB,$7F,$2E,$DF,$88,$FE,$53,$18,$AF,$E2,$0A,$32,$6B,$C2,$B5,$CD
	.dc.b	$C3,$FE,$A4,$FA,$8F,$F7,$33,$93,$1E,$07,$6F,$85,$EF,$C3,$67,$C8
	.dc.b	$E1,$E5,$E5,$DC,$FF,$91,$60,$8F,$E7,$FF,$57,$F2,$AB,$FC,$60,$FE
	.dc.b	$49,$B5,$AD,$E4,$79,$7E,$6E,$CF,$47,$9F,$7F,$4F,$16,$D7,$FA,$FB
	.dc.b	$0F,$D9,$B6,$3C,$3E,$CE,$E3,$4E,$1F,$85,$F5,$77,$1F,$4B,$E8,$F9
	.dc.b	$FF,$4F,$A5,$6B,$B3,$87,$E5,$7D,$6D,$BF,$82,$FC,$0F,$4F,$E0,$F8
	.dc.b	$96,$BC,$CE,$1F,$F6,$AF,$D6,$9C,$5F,$F4,$B8,$C4,$7B,$78,$7E,$5F
	.dc.b	$07,$CC,$6C,$7C,$18,$F9,$FA,$62,$FF,$AA,$6E,$8C,$D7,$FB,$93,$EB
	.dc.b	$7F,$6A,$FC,$98,$FE,$EB,$18,$AF,$EB,$9F,$5B,$F9,$6F,$03,$CD,$01
	.dc.b	$1F,$98,$77,$F1,$9F,$DF,$A3,$13,$C2,$77,$DD,$FE,$02,$E7,$FF,$10
	.dc.b	$E5,$FE,$F6,$BD,$BF,$E7,$58,$CD,$78,$BC,$6E,$9E,$B7,$56,$D7,$18
	.dc.b	$FF,$06,$F5,$7A,$72,$F0,$63,$FD,$32,$BD,$11,$E3,$9D,$47,$FA,$0E
	.dc.b	$5F,$39,$E9,$F9,$CF,$53,$67,$CD,$1E,$DF,$8D,$C6,$7F,$19,$C4,$F6
	.dc.b	$3F,$D1,$D8,$7F,$20,$7D,$D1,$E6,$CF,$E3,$79,$3F,$1F,$E3,$2D,$7C
	.dc.b	$37,$0F,$3C,$3E,$87,$CB,$BE,$67,$31,$FE,$11,$8C,$4F,$CB,$F6,$FF
	.dc.b	$C5,$FC,$4F,$73,$EA,$46,$19,$3D,$6F,$F7,$1F,$5B,$87,$FC,$39,$D4
	.dc.b	$8F,$F1,$CF,$D6,$FF,$2E,$9F,$C1,$3F,$CB,$B9,$3E,$F3,$98,$39,$BF
	.dc.b	$F5,$83,$63,$F1,$5F,$5A,$E1,$EB,$F0,$7F,$E4,$68,$C5,$F9,$6F,$FA
	.dc.b	$FE,$1D,$BD,$9F,$1F,$C4,$79,$19,$5E,$BB,$7C,$96,$33,$FF,$1B,$46
	.dc.b	$27,$A7,$97,$EC,$E7,$F6,$B6,$3C,$11,$FE,$58,$FA,$D7,$C8,$67,$0F
	.dc.b	$F2,$04,$62,$BE,$57,$B9,$FE,$34,$9F,$ED,$4F,$F9,$80,$C1,$FD,$B6
	.dc.b	$36,$F5,$F9,$CD,$36,$39,$F9,$DC,$7C,$EE,$4D,$8E,$F1,$E5,$CF,$18
	.dc.b	$79,$1F,$07,$AD,$E1,$F5,$AD,$F9,$9C,$3F,$81,$FA,$C7,$D6,$E3,$0F
	.dc.b	$F2,$E4,$62,$7D,$BE,$FF,$FA,$0A,$7F,$81,$68,$8C,$47,$F8,$25,$B1
	.dc.b	$99,$FF,$4E,$4F,$0E,$AF,$E1,$71,$FE,$9F,$9F,$E9,$0F,$F8,$52,$74
	.dc.b	$D7,$FC,$29,$3C,$1F,$F7,$D4,$BC,$CC,$47,$91,$AF,$F8,$6D,$B1,$BA
	.dc.b	$E9,$F5,$9F,$E8,$2E,$DF,$2D,$A3,$0F,$A1,$FF,$49,$FA,$BA,$73,$1F
	.dc.b	$EB,$77,$FC,$3E,$F4,$F7,$FF,$93,$3F,$EA,$88,$CB,$63,$37,$C7,$C6
	.dc.b	$B3,$C2,$BF,$C4,$EF,$9A,$32,$FF,$31,$8C,$BF,$EB,$4F,$77,$F8,$CB
	.dc.b	$F1,$DA,$E7,$C9,$E2,$A3,$37,$E9,$F8,$9C,$D9,$A4,$AF,$8F,$F4,$9A
	.dc.b	$23,$83,$EB,$7C,$B3,$C3,$EE,$D8,$CC,$60,$EF,$E6,$48,$FA,$6D,$7E
	.dc.b	$47,$8B,$F5,$1F,$A8,$DC,$7F,$84,$47,$D3,$8C,$78,$32,$3E,$B3,$5F
	.dc.b	$B4,$D7,$EC,$35,$FA,$FF,$5B,$FE,$46,$7F,$06,$D7,$F3,$2D,$7F,$DE
	.dc.b	$CD,$7E,$FE,$AF,$83,$F4,$36,$3F,$1A,$D7,$F4,$6D,$7F,$E0,$06,$7F
	.dc.b	$C3,$9F,$C9,$B5,$FD,$4B,$5D,$F8,$3F,$16,$1F,$45,$49,$3C,$9D,$7C
	.dc.b	$C4,$8E,$AD,$7F,$3D,$87,$CD,$23,$F9,$F7,$E8,$7A,$12,$3F,$44,$D7
	.dc.b	$F4,$CF,$E8,$3C,$9F,$F0,$06,$C8,$C7,$1F,$E7,$5B,$5F,$9B,$F4,$2B
	.dc.b	$FB,$17,$EF,$11,$FD,$83,$F8,$3F,$98,$7B,$D4,$93,$FB,$A8,$CB,$63
	.dc.b	$AC,$62,$3F,$DF,$0D,$C7,$FC,$2C,$BD,$ED,$F8,$9C,$19,$5F,$D9,$E5
	.dc.b	$3E,$8C,$DF,$4B,$49,$5F,$DB,$63,$FD,$7E,$DF,$8A,$E0,$47,$FB,$97
	.dc.b	$0F,$F9,$ED,$FE,$E9,$F5,$BF,$71,$BC,$5F,$1A,$31,$0D,$3F,$C5,$D1
	.dc.b	$F8,$76,$F1,$EC,$93,$DB,$E1,$39,$72,$48,$EE,$F5,$EA,$91,$E2,$6B
	.dc.b	$DF,$93,$F8,$EE,$6F,$8F,$F2,$7C,$34,$AF,$4B,$E7,$C6,$47,$B1,$AF
	.dc.b	$33,$FC,$99,$4F,$3B,$DD,$F3,$72,$7E,$1B,$F0,$7E,$2A,$47,$C6,$6B
	.dc.b	$CC,$53,$CE,$FC,$6F,$D6,$BE,$52,$92,$BF,$C4,$51,$FE,$E1,$6F,$95
	.dc.b	$C0,$8F,$D6,$F5,$3F,$2F,$F5,$BF,$E2,$58,$FF,$15,$35,$F1,$37,$93
	.dc.b	$FC,$C4,$FE,$3F,$87,$FA,$2B,$67,$FC,$D8,$D4,$DF,$AA,$F4,$47,$6F
	.dc.b	$A9,$F0,$91,$F0,$FC,$57,$90,$8E,$0F,$F1,$F7,$D6,$FF,$1B,$C7,$C7
	.dc.b	$F1,$DE,$A2,$4F,$AB,$D3,$F6,$91,$C1,$EE,$F4,$FE,$09,$1C,$1F,$E4
	.dc.b	$BF,$AD,$7A,$7F,$BC,$FA,$CF,$F4,$16,$3F,$E8,$36,$EF,$FD,$99,$B7
	.dc.b	$FE,$8D,$5F,$13,$C4,$91,$FE,$57,$8C,$BF,$E5,$9F,$AD,$69,$DF,$D6
	.dc.b	$93,$B7,$C5,$EB,$48,$E4,$D7,$F1,$1C,$5F,$5F,$2F,$27,$A9,$E9,$24
	.dc.b	$F8,$7D,$6F,$61,$1C,$1E,$5F,$57,$E0,$11,$C1,$FE,$6F,$F2,$9D,$DF
	.dc.b	$17,$D1,$93,$ED,$7E,$47,$A3,$23,$E0,$35,$F3,$F4,$FC,$DF,$97,$91
	.dc.b	$F1,$5F,$A1,$F0,$B2,$3E,$43,$5F,$D3,$7D,$6B,$5F,$E6,$3F,$3F,$FE
	.dc.b	$8D,$9F,$F4,$23,$5F,$D6,$37,$9F,$FD,$78,$FF,$12,$6F,$9D,$C8,$93
	.dc.b	$CB,$CD,$EA,$47,$07,$FA,$47,$E2,$7F,$C0,$CF,$6F,$A4,$F1,$91,$F1
	.dc.b	$FD,$1F,$98,$8E,$0E,$5E,$5F,$AB,$CD,$91,$EA,$7D,$BE,$6C,$8F,$6B
	.dc.b	$5F,$F4,$FF,$0F,$C2,$F2,$E4,$7C,$27,$E2,$F9,$12,$3E,$2B,$5F,$8B
	.dc.b	$EC,$FC,$A2,$3F,$29,$FE,$BB,$E6,$FF,$8C,$63,$FD,$61,$F5,$BF,$E3
	.dc.b	$AE,$1F,$F7,$33,$7B,$7F,$C9,$2D,$9F,$F7,$82,$F8,$FF,$EC,$EF,$F4
	.dc.b	$C8,$FE,$40,$DA,$E7,$EA,$EB,$A7,$FB,$6F,$D6,$BE,$2E,$93,$0F,$F6
	.dc.b	$14,$3E,$2F,$7C,$52,$72,$F2,$E2,$FE,$EF,$F5,$BF,$C9,$E7,$F9,$83
	.dc.b	$73,$FF,$80,$0D,$DB,$FC,$C6,$BE,$2B,$8F,$CB,$CB,$FF,$69,$4A,$3A
	.dc.b	$FC,$3F,$F8,$13,$DB,$7C,$85,$25,$FF,$B6,$61,$DF,$BF,$12,$4E,$5E
	.dc.b	$5E,$27,$F8,$3B,$EB,$57,$49,$5F,$EE,$38,$FC,$9E,$CD,$BF,$CF,$70
	.dc.b	$EE,$E4,$4A,$E1,$C9,$D7,$C6,$91,$CD,$FF,$0E,$FD,$6F,$FA,$12,$3F
	.dc.b	$AB,$3F,$E0,$A7,$FD,$E4,$D7,$3F,$03,$C3,$E3,$7F,$C4,$BF,$5B,$FE
	.dc.b	$DC,$9F,$F7,$C3,$5F,$EC,$0D,$D3,$D1,$01,$1F,$E5,$18,$CC,$3D,$0F
	.dc.b	$F0,$46,$BF,$E9,$86,$C7,$16,$BB,$DB,$FC,$6F,$1F,$DB,$63,$B5,$74
	.dc.b	$8C,$47,$F7,$36,$C6,$7C,$1E,$3E,$FC,$AE,$38,$78,$63,$AE,$BF,$E4
	.dc.b	$4F,$AD,$FF,$1A,$56,$8F,$63,$87,$FC,$00,$DC,$3D,$AF,$CB,$7D,$FF
	.dc.b	$5B,$FD,$3A,$4D,$27,$AF,$FE,$52,$7F,$D1,$B1,$FE,$0B,$6C,$F2,$F8
	.dc.b	$CD,$7C,$66,$F9,$E5,$F9,$4E,$AF,$3E,$79,$19,$C3,$FC,$C8,$D8,$F0
	.dc.b	$7C,$5F,$87,$C4,$93,$B7,$C9,$E2,$48,$E4,$D7,$CD,$E9,$7F,$9A,$BE
	.dc.b	$B7,$F3,$3C,$5F,$ED,$26,$CF,$E0,$DA,$97,$9D,$C4,$91,$EB,$8C,$BF
	.dc.b	$E5,$6E,$3F,$F3,$BF,$CF,$FF,$85,$23,$FC,$60,$DE,$1F,$C6,$B7,$0F
	.dc.b	$F7,$1B,$FC,$0E,$7F,$28,$D4,$D3,$1E,$5F,$F6,$68,$F9,$9F,$59,$FD
	.dc.b	$3E,$BF,$DD,$AD,$8F,$CF,$35,$3F,$98,$27,$A5,$FE,$8E,$71,$D5,$49
	.dc.b	$3D,$9F,$F4,$3B,$BF,$89,$8F,$89,$E5,$AC,$8F,$2B,$5F,$F2,$95,$7F
	.dc.b	$07,$CF,$F6,$5F,$5B,$EE,$9F,$3F,$85,$FE,$79,$7F,$82,$1E,$87,$0F
	.dc.b	$F0,$7A,$E4,$E9,$AE,$7C,$27,$D6,$F2,$8F,$83,$18,$78,$35,$E1,$1F
	.dc.b	$EE,$D5,$F4,$D4,$8F,$FB,$C5,$FF,$79,$7C,$17,$F0,$CC,$BF,$84,$CF
	.dc.b	$F9,$AF,$7F,$D0,$3F,$E4,$06,$FC,$5E,$DF,$E4,$0D,$8F,$E2,$9E,$77
	.dc.b	$F3,$89,$D9,$FA,$97,$11,$FE,$7B,$1F,$15,$C5,$FE,$39,$E7,$D7,$35
	.dc.b	$E2,$8C,$19,$F9,$7B,$1E,$67,$0F,$7F,$AD,$B5,$FE,$89,$C1,$1D,$F2
	.dc.b	$FE,$8F,$5F,$BD,$5C,$79,$E4,$D3,$27,$D5,$DD,$71,$E8,$FA,$7E,$B7
	.dc.b	$E3,$47,$FA,$5D,$F9,$4D,$DF,$5B,$CB,$F8,$9D,$B6,$BC,$CF,$3F,$93
	.dc.b	$E9,$E6,$5A,$FE,$25,$8F,$FB,$59,$B1,$E1,$F3,$37,$1A,$7C,$9F,$1B
	.dc.b	$E5,$F4,$EC,$F2,$70,$F4,$F4,$EE,$7E,$97,$D9,$F0,$FE,$DF,$8C,$B5
	.dc.b	$E2,$7E,$C7,$CB,$FF,$86,$63,$F9,$32,$F9,$7C,$FB,$BA,$77,$C1,$C3
	.dc.b	$C8,$B7,$C1,$FC,$17,$D6,$FF,$85,$E3,$F9,$82,$FA,$DC,$FB,$BA,$77
	.dc.b	$D9,$C3,$E5,$2D,$F0,$7F,$19,$F5,$BF,$E1,$58,$FE,$74,$BB,$F3,$E0
	.dc.b	$69,$DF,$17,$0E,$2B,$7C,$1F,$CB,$7D,$6B,$CF,$BC,$18,$6D,$7F,$47
	.dc.b	$3F,$C6,$85,$DB,$F3,$CE,$39,$7F,$DB,$8F,$BD,$FF,$6F,$7D,$6F,$F7
	.dc.b	$D8,$FE,$A0,$FE,$33,$C2,$FE,$FA,$E6,$77,$DF,$FA,$98,$FD,$3C,$72
	.dc.b	$DB,$1E,$26,$5D,$75,$6F,$7B,$97,$67,$F6,$5E,$2F,$F8,$26,$3F,$B0
	.dc.b	$2F,$90,$CF,$B7,$A7,$7E,$AF,$0F,$16,$DF,$07,$F7,$9F,$5B,$FA,$ED
	.dc.b	$7F,$DF,$4F,$EB,$A7,$FB,$63,$9E,$0F,$F0,$CB,$99,$DF,$B9,$CB,$E1
	.dc.b	$B9,$B6,$F8,$FB,$BF,$C2,$78,$FE,$26,$3C,$A3,$18,$67,$C6,$D7,$8D
	.dc.b	$70,$FE,$F9,$82,$3F,$89,$7D,$6F,$FB,$5A,$7F,$C0,$0F,$F9,$59,$BE
	.dc.b	$EE,$BE,$0F,$F8,$C7,$D6,$FE,$56,$7F,$C0,$CB,$E1,$B3,$E2,$69,$93
	.dc.b	$AF,$12,$EA,$DF,$FC,$73,$82,$7A,$9F,$C9,$7E,$A3,$FD,$C7,$7C,$4F
	.dc.b	$F0,$A3,$8B,$E5,$1F,$E4,$55,$4F,$8A,$F3,$F1,$7A,$1B,$5E,$88,$ED
	.dc.b	$E6,$62,$F9,$9F,$E4,$FC,$11,$FC,$DF,$EA,$FE,$7D,$F7,$7F,$2C,$DA
	.dc.b	$D7,$14,$8F,$3F,$CC,$D9,$EA,$30,$47,$5E,$FE,$DE,$3D,$BF,$F1,$96
	.dc.b	$3F,$B9,$6C,$78,$FD,$ED,$C6,$9C,$3F,$1B,$EA,$EE,$5E,$0C,$91,$EB
	.dc.b	$FA,$1F,$67,$A9,$6F,$BB,$87,$E7,$FD,$6D,$C3,$93,$F0,$FD,$4F,$89
	.dc.b	$E3,$5A,$F3,$B8,$7F,$C9,$5F,$5B,$FE,$91,$72,$60,$F9,$9C,$5F,$37
	.dc.b	$86,$C7,$C0,$70,$FC,$2D,$2E,$5F,$2B,$83,$D7,$39,$3F,$CB,$1F,$5B
	.dc.b	$FB,$B9,$D5,$7F,$6F,$8C,$74,$7F,$A2,$3E,$B7,$F4,$15,$F5,$1F,$D0
	.dc.b	$37,$B7,$FB,$B4,$62,$78,$76,$FE,$57,$93,$2D,$CB,$FC,$D1,$5F,$C2
	.dc.b	$5B,$EB,$72,$F1,$F9,$B2,$3F,$CD,$DD,$5E,$47,$F3,$55,$FD,$37,$3F
	.dc.b	$C3,$CE,$5F,$F5,$14,$3E,$6F,$CB,$F3,$7E,$66,$DF,$8A,$3D,$3F,$19
	.dc.b	$8B,$F7,$38,$9F,$0E,$56,$7B,$BF,$03,$C9,$F8,$3F,$15,$6B,$DA,$E1
	.dc.b	$FF,$43,$7C,$BB,$EB,$71,$87,$E7,$A3,$13,$F1,$3D,$BF,$F1,$07,$03
	.dc.b	$DD,$F9,$B1,$86,$4F,$FA,$47,$A5,$F5,$B8,$C7,$F5,$31,$89,$FF,$4B
	.dc.b	$7D,$6F,$F2,$B7,$F8,$13,$FC,$AF,$9B,$FD,$8A,$E6,$BF,$D9,$4E,$67
	.dc.b	$7F,$05,$FE,$33,$FA,$D7,$17,$F8,$D6,$B5,$7F,$ED,$19,$CD,$F8,$3E
	.dc.b	$1F,$5E,$D6,$DB,$1E,$48,$E9,$F2,$38,$BF,$E0,$36,$8E,$DE,$5F,$7E
	.dc.b	$AB,$87,$B1,$C3,$FE,$B3,$FA,$D7,$C7,$63,$1F,$E1,$71,$89,$FF,$5C
	.dc.b	$7B,$9F,$E3,$2E,$0F,$F7,$2B,$FD,$99,$BF,$0F,$FE,$51,$8C,$4F,$F2
	.dc.b	$96,$DF,$FB,$BE,$DF,$39,$B9,$FF,$2C,$46,$2B,$BF,$9F,$CF,$CF,$E8
	.dc.b	$D8,$F1,$C7,$0D,$71,$8F,$A1,$F1,$7A,$FE,$3F,$5E,$DF,$A9,$C3,$FE
	.dc.b	$D3,$FA,$C7,$D6,$E3,$0F,$E5,$71,$89,$FF,$6C,$7B,$7F,$C3,$E7,$FC
	.dc.b	$FA,$D8,$FE,$8D,$08,$C3,$FD,$25,$A8,$7D,$5F,$F0,$49,$E2,$D8,$FE
	.dc.b	$A0,$D7,$83,$5E,$3F,$1B,$F9,$F7,$FD,$4B,$1F,$E7,$4E,$EF,$F4,$F2
	.dc.b	$F9,$B7,$B1,$8B,$E1,$7E,$0F,$27,$FA,$F6,$36,$3F,$C4,$87,$82,$32
	.dc.b	$F9,$1A,$F9,$3C,$9A,$FB,$1B,$89,$A6,$3B,$FF,$D6,$53,$F0,$9B,$3F
	.dc.b	$A6,$8F,$4C,$63,$AE,$47,$B1,$AF,$FB,$13,$E5,$1F,$5D,$D1,$B1,$FE
	.dc.b	$BF,$6B,$B8,$D7,$91,$23,$4D,$7A,$76,$F3,$23,$CD,$ED,$E0,$48,$EC
	.dc.b	$D7,$FD,$AF,$F5,$CF,$97,$F7,$B6,$3F,$DA,$4D,$77,$1F,$0B,$E1,$24
	.dc.b	$7D,$A6,$BF,$79,$AF,$E2,$1A,$FE,$23,$CD,$F9,$64,$7F,$06,$D7,$FC
	.dc.b	$0F,$B2,$3F,$C3,$DE,$13,$8B,$95,$7E,$5E,$30,$FE,$4B,$CD,$DC,$FF
	.dc.b	$89,$7E,$BB,$FD,$A3,$3F,$EE,$D6,$BF,$EF,$86,$BF,$A6,$6B,$C7,$C0
	.dc.b	$E6,$FE,$96,$38,$DF,$82,$11,$8A,$E1,$1E,$56,$BD,$EB,$F1,$B2,$69
	.dc.b	$8E,$CE,$F3,$D1,$B3,$FC,$0A,$3A,$1E,$3F,$2F,$27,$42,$3E,$F7,$BF
	.dc.b	$BB,$A8,$E2,$7A,$23,$AC,$48,$F8,$CF,$06,$9F,$84,$F8,$4F,$C3,$91
	.dc.b	$C1,$FC,$38,$FE,$15,$F8,$2F,$F1,$0C,$FE,$37,$0A,$BD,$AD,$AF,$1D
	.dc.b	$6F,$4F,$F2,$E3,$63,$99,$F1,$1C,$27,$F8,$4B,$EC,$DD,$F4,$FF,$29
	.dc.b	$B1,$98,$C7,$C6,$48,$E0,$D7,$FC,$AF,$D1,$FE,$7D,$7E,$3F,$D7,$BE
	.dc.b	$9B,$EA,$B6,$33,$18,$E8,$91,$C1,$AF,$4F,$86,$E3,$F2,$F2,$3A,$BD
	.dc.b	$3D,$39,$1E,$06,$BE,$4F,$94,$F0,$FC,$4C,$8F,$2B,$E4,$B9,$1E,$86
	.dc.b	$BE,$CF,$43,$D3,$DB,$23,$DA,$FB,$39,$24,$7C,$06,$BF,$13,$CD,$FC
	.dc.b	$AF,$D7,$BD,$2A,$49,$FC,$6C,$7E,$8D,$FE,$98,$D4,$D3,$1C,$2F,$4F
	.dc.b	$F4,$3F,$5F,$FE,$22,$AF,$CB,$B5,$FF,$45,$B7,$AB,$F6,$4B,$AF,$89
	.dc.b	$27,$8B,$FA,$CF,$AB,$F4,$47,$4F,$A9,$EE,$47,$BF,$C4,$F8,$88,$E0
	.dc.b	$FD,$4F,$D7,$F9,$DF,$2F,$CC,$7A,$08,$FA,$3D,$DF,$59,$1C,$1E,$CF
	.dc.b	$89,$F7,$91,$C1,$FB,$6F,$AF,$FC,$EB,$FE,$AA,$FA,$FF,$E2,$1B,$7F
	.dc.b	$11,$6F,$87,$FE,$C3,$6C,$FF,$7B,$5F,$11,$FC,$31,$B1,$FC,$09,$AF
	.dc.b	$58,$FE,$23,$F5,$EC,$F8,$7C,$B4,$AE,$0F,$3F,$87,$91,$CD,$AF,$FB
	.dc.b	$03,$87,$EB,$65,$E5,$FA,$EF,$11,$27,$C5,$EB,$F9,$48,$E0,$F3,$7A
	.dc.b	$FE,$92,$38,$3F,$8D,$F9,$1F,$C2,$DF,$8D,$E9,$7B,$C8,$FB,$FD,$2F
	.dc.b	$84,$47,$07,$F2,$4C,$9C,$FF,$1F,$E6,$E4,$FC,$77,$E5,$7C,$3C,$8F
	.dc.b	$94,$D7,$FD,$BF,$F5,$ED,$7F,$9D,$FB,$BF,$E8,$79,$FE,$62,$D7,$FD
	.dc.b	$CA,$DE,$6F,$E9,$8F,$F0,$D8,$CF,$99,$C8,$8F,$2F,$33,$A9,$1C,$1F
	.dc.b	$E9,$6F,$2F,$C3,$F4,$D2,$3C,$2F,$93,$E9,$64,$79,$1A,$FF,$A6,$FC
	.dc.b	$AF,$4F,$99,$23,$D2,$FB,$3C,$C9,$1E,$C6,$BF,$E9,$8E,$1F,$83,$E5
	.dc.b	$48,$F8,$2F,$C4,$F1,$E4,$7C,$46,$BF,$13,$E0,$7E,$49,$1F,$92,$FF
	.dc.b	$66,$F2,$7F,$CB,$91,$FD,$73,$EB,$DD,$DF,$DA,$9B,$19,$8C,$7C,$7E
	.dc.b	$7D,$39,$3C,$5E,$1D,$3F,$DC,$BE,$BD,$F1,$94,$95,$FD,$AA,$1E,$F6
	.dc.b	$F8,$64,$E5,$E5,$C3,$FD,$F3,$EB,$FF,$91,$13,$E3,$31,$EF,$E7,$E5
	.dc.b	$E6,$1F,$DD,$A9,$1C,$7E,$83,$FC,$0B,$ED,$3E,$46,$92,$FF,$BE,$C7
	.dc.b	$F9,$B5,$BE,$37,$02,$3A,$F1,$BF,$C1,$FF,$5E,$AA,$49,$FF,$03,$C3
	.dc.b	$BF,$7D,$D2,$72,$F2,$E3,$E7,$ED,$49,$E2,$FF,$87,$3E,$BF,$FB,$C8
	.dc.b	$F8,$FC,$12,$7F,$C2,$B0,$F8,$FF,$F1,$1F,$D7,$FF,$B2,$67,$FC,$32
	.dc.b	$D5,$E4,$7F,$8C,$E3,$0D,$DF,$FC,$57,$E8,$FF,$B2,$AB,$FC,$42,$D7
	.dc.b	$8F,$A3,$FE,$A1,$51,$9A,$FF,$1F,$3F,$16,$38,$F8,$9D,$08,$F4,$F1
	.dc.b	$3B,$FB,$F2,$3B,$BE,$2D,$BF,$C7,$BF,$5E,$F9,$EA,$4A,$FF,$1A,$C7
	.dc.b	$C5,$7F,$C6,$FF,$5F,$FC,$F1,$34,$9E,$C7,$F9,$21,$FF,$4B,$47,$B0
	.dc.b	$FF,$5E,$38,$D3,$FC,$B0,$DD,$7E,$03,$F2,$1B,$B7,$32,$57,$8B,$A3
	.dc.b	$B1,$1C,$1E,$6F,$8B,$F1,$78,$B2,$3D,$2F,$97,$C5,$91,$EC,$6B,$E7
	.dc.b	$3D,$BE,$9F,$F9,$8B,$EB,$FF,$21,$1F,$E5,$86,$BB,$6E,$9F,$E7,$35
	.dc.b	$E7,$F1,$24,$F1,$7F,$84,$71,$7F,$9B,$BE,$6F,$FC,$07,$3E,$16,$DF
	.dc.b	$F9,$F5,$BB,$7E,$15,$FF,$0D,$13,$C5,$E4,$7F,$13,$E4,$FF,$3C,$FD
	.dc.b	$77,$F3,$9B,$FF,$43,$36,$FF,$1C,$DC,$3F,$DE,$0F,$E3,$E3,$3E,$97
	.dc.b	$22,$4F,$F9,$E2,$7F,$24,$FC,$76,$D7,$6F,$0B,$C2,$7F,$A2,$27,$C7
	.dc.b	$E0,$C7,$93,$E6,$D6,$47,$56,$BD,$7C,$1F,$D6,$AF,$AF,$6D,$EA,$9C
	.dc.b	$CF,$EA,$46,$48,$CC,$6F,$2F,$F4,$33,$A8,$CB,$F0,$BC,$27,$D7,$EA
	.dc.b	$FF,$98,$B6,$33,$3F,$E8,$F9,$CB,$FE,$F8,$FA,$9F,$F7,$CA,$69,$1F
	.dc.b	$0F,$CB,$FF,$B2,$31,$FF,$5D,$47,$F9,$6F,$83,$FD,$4A,$A7,$51,$ED
	.dc.b	$F8,$AB,$3F,$C1,$DA,$BC,$D7,$B3,$88,$E8,$7E,$76,$33,$5D,$B9,$75
	.dc.b	$E7,$6D,$B1,$E2,$8E,$5F,$35,$8B,$FD,$30,$73,$3F,$EB,$CB,$7C,$C6
	.dc.b	$2B,$C0,$F9,$B4,$FE,$7D,$5F,$EC,$27,$F9,$E3,$79,$B5,$CD,$23,$A7
	.dc.b	$7D,$C3,$D4,$FC,$0F,$AF,$F8,$D3,$FE,$CC,$7D,$F8,$7F,$B9,$5C,$D7
	.dc.b	$EF,$5C,$CF,$F2,$F6,$C0,$F9,$5D,$DD,$78,$3E,$6E,$88,$F0,$78,$BE
	.dc.b	$1F,$32,$E3,$E1,$70,$F8,$FC,$BB,$8F,$85,$F4,$7C,$57,$A7,$D1,$B5
	.dc.b	$E9,$70,$FA,$FD,$1B,$7E,$97,$E0,$7C,$27,$C1,$F8,$9B,$5F,$05,$FC
	.dc.b	$1F,$CD,$3D,$9B,$F5,$F8,$6D,$7F,$BC,$A7,$D8,$DF,$27,$8F,$07,$27
	.dc.b	$F1,$BF,$5F,$FE,$0E,$9F,$F7,$CA,$FA,$AC,$FB,$7A,$77,$C9,$C3,$E4
	.dc.b	$AD,$F0,$7F,$2F,$F5,$EE,$FB,$DF,$86,$D7,$D3,$EB,$3F,$D9,$69,$C7
	.dc.b	$CB,$8F,$9B,$67,$81,$FD,$AF,$EB,$DE,$87,$E9,$9F,$7F,$37,$F3,$F7
	.dc.b	$33,$C2,$3F,$05,$1A,$C7,$C3,$1F,$85,$9F,$F1,$99,$B7,$CB,$BF,$EC
	.dc.b	$A3,$31,$FD,$2D,$F9,$0F,$F4,$DF,$AF,$E4,$DF,$17,$F7,$31,$8B,$EF
	.dc.b	$28,$FC,$CC,$61,$78,$9C,$36,$E2,$DC,$3C,$93,$FE,$38,$6E,$9D,$BA
	.dc.b	$BB,$BF,$C2,$3C,$6F,$F0,$7C,$FA,$DF,$F2,$CB,$7D,$9D,$7C,$0E,$FD
	.dc.b	$5E,$1E,$2D,$AE,$0F,$F1,$2F,$AF,$FD,$EE,$FF,$DA,$DF,$DE,$B7,$C8
	.dc.b	$D7,$91,$BF,$6B,$87,$C2,$DC,$7E,$66,$08,$E9,$EA,$FF,$1E,$E5,$F8
	.dc.b	$D5,$FE,$27,$A4,$E7,$2D,$79,$1B,$F2,$2D,$FD,$7C,$3F,$C7,$8F,$D9
	.dc.b	$FA,$FF,$F6,$AB,$7E,$07,$F9,$5C,$63,$0F,$F0,$23,$FE,$5F,$9F,$C0
	.dc.b	$7D,$7F,$E8,$5B,$F0,$9F,$CC,$A3,$15,$FE,$0B,$51,$93,$5E,$1D,$AE
	.dc.b	$6E,$1F,$C6,$7D,$7F,$FB,$85,$C9,$8F,$07,$B7,$C0,$F7,$E2,$B3,$E4
	.dc.b	$70,$F2,$F3,$6E,$7F,$E1,$DC,$11,$FD,$23,$EB,$7E,$5C,$FF,$21,$7F
	.dc.b	$1F,$5F,$E5,$37,$30,$72,$EF,$7F,$EA,$36,$F9,$6E,$91,$87,$B7,$83
	.dc.b	$DF,$C7,$B8,$FE,$A7,$1F,$E9,$4D,$8F,$27,$B5,$B8,$D3,$87,$E2,$FD
	.dc.b	$6D,$CF,$D8,$FA,$FE,$67,$D9,$EA,$5A,$EC,$E1,$F9,$BF,$5F,$6F,$E2
	.dc.b	$3F,$0F,$D5,$F8,$9E,$35,$AF,$33,$87,$F8,$07,$D7,$FF,$14,$72,$60
	.dc.b	$F9,$9C,$5F,$37,$82,$C7,$BD,$C3,$F0,$74,$B9,$7F,$94,$77,$47,$D5
	.dc.b	$3F,$BE,$FD,$7F,$ED,$A7,$51,$8F,$1B,$FC,$03,$F5,$FF,$97,$F3,$3F
	.dc.b	$84,$BF,$97,$AF,$F2,$E3,$9C,$7F,$D1,$CE,$67,$7D,$DF,$6B,$8F,$9B
	.dc.b	$6B,$F8,$F5,$7F,$04,$BE,$7E,$BF,$83,$C9,$F0,$B2,$BF,$92,$75,$79
	.dc.b	$F2,$F1,$CF,$F8,$5A,$31,$3E,$2F,$99,$F4,$FC,$CF,$A9,$B1,$E0,$8F
	.dc.b	$6F,$D5,$63,$2F,$5B,$89,$EF,$95,$9F,$F3,$D6,$1E,$7C,$BE,$1F,$97
	.dc.b	$F1,$BE,$27,$E3,$B7,$0F,$6C,$7F,$8A,$BE,$BD,$F5,$39,$C7,$FD,$09
	.dc.b	$18,$AF,$87,$ED,$FF,$8C,$38,$5E,$EF,$D2,$8C,$32,$7F,$A3,$74,$BE
	.dc.b	$9F,$18,$FF,$A3,$E3,$13,$FD,$2F,$EB,$FF,$94,$ED,$FD,$09,$FE,$52
	.dc.b	$DF,$4B,$FC,$7F,$18,$9F,$F6,$43,$6F,$FC,$35,$F5,$EE,$3D,$FA,$7F
	.dc.b	$E4,$68,$C5,$F7,$F8,$7E,$7E,$D7,$46,$C7,$8E,$38,$7D,$6E,$31,$FF
	.dc.b	$53,$5A,$3C,$7E,$67,$93,$9E,$E1,$EB,$70,$FF,$67,$FA,$F7,$C7,$63
	.dc.b	$1F,$F5,$84,$62,$7F,$B6,$7B,$BF,$E3,$49,$FE,$BE,$FF,$97,$8C,$1F
	.dc.b	$D8,$A3,$6F,$F9,$8A,$5F,$39,$B8,$C3,$C7,$CE,$E5,$E7,$5B,$EE,$E1
	.dc.b	$F3,$67,$71,$F2,$BE,$1F,$5F,$C5,$EB,$DA,$F3,$B8,$7F,$C0,$9F,$5E
	.dc.b	$FA,$EC,$61,$FE,$CD,$8C,$4F,$F8,$27,$EA,$FF,$A7,$3F,$E9,$96,$CF
	.dc.b	$F9,$4C,$38,$7F,$CA,$CD,$73,$F1,$9F,$92,$7F,$D4,$46,$3B,$7F,$D3
	.dc.b	$F3,$FE,$84,$F5,$6D,$6E,$BC,$1E,$5F,$7C,$2C,$7F,$85,$8E,$F1,$97
	.dc.b	$C9,$68,$C4,$79,$3C,$7A,$FA,$1B,$26,$98,$EB,$FF,$59,$CF,$B9,$B3
	.dc.b	$F9,$98,$F4,$C6,$39,$E4,$7B,$1A,$FC,$0F,$AD,$74,$74,$6C,$7F,$B0
	.dc.b	$1A,$EE,$3C,$3F,$3F,$FC,$02,$3F,$D8,$AD,$7E,$5B,$5F,$F6,$73,$5F
	.dc.b	$AC,$FE,$71,$E3,$D3,$D0,$8F,$4F,$A1,$EE,$47,$07,$97,$E5,$BD,$04
	.dc.b	$70,$7D,$AF,$83,$F8,$A4,$70,$7D,$EF,$2F,$E4,$91,$C1,$0F,$C2,$EA
	.dc.b	$8D,$46,$1B,$E2,$7F,$94,$A7,$53,$F9,$79,$FC,$A4,$E7,$CB,$BC,$5F
	.dc.b	$84,$7E,$C9,$2B,$F2,$2F,$DE,$FA,$EF,$F4,$04,$7F,$9A,$5F,$F2,$DB
	.dc.b	$7B,$F9,$BF,$EB,$38,$DA,$BF,$F3,$2C,$62,$3F,$E0,$03,$5F,$D4,$35
	.dc.b	$E5,$DD,$E1,$3A,$A8,$EB,$3D,$27,$31,$FE,$DD,$9F,$37,$91,$E9,$23
	.dc.b	$FA,$E3,$93,$ED,$F6,$F8,$B2,$3D,$D1,$D0,$D3,$C6,$51,$88,$FD,$CC
	.dc.b	$61,$E2,$43,$E1,$65,$EE,$F6,$FD,$7F,$5D,$FC,$1E,$72,$7F,$B6,$F9
	.dc.b	$7F,$CC,$AF,$FA,$5D,$AE,$84,$71,$CF,$FD,$22,$FE,$D2,$32,$7E,$D9
	.dc.b	$C1,$1F,$09,$0F,$F5,$22,$EB,$F9,$FC,$0F,$A7,$A4,$AF,$E4,$5F,$17
	.dc.b	$A2,$38,$3F,$D5,$DC,$FF,$E6,$A7,$FD,$61,$F5,$CE,$7F,$B6,$D8,$CC
	.dc.b	$63,$A2,$47,$06,$BF,$06,$9F,$05,$CB,$CE,$93,$D9,$EB,$E6,$C8,$F0
	.dc.b	$B5,$E5,$F1,$DE,$2F,$84,$91,$E6,$7C,$B7,$23,$D2,$D7,$C5,$E5,$FA
	.dc.b	$BB,$A4,$7B,$9F,$6F,$C3,$C8,$F8,$2D,$7D,$5E,$5F,$E7,$3E,$B9,$F2
	.dc.b	$94,$93,$FE,$D7,$8F,$D7,$BF,$81,$6F,$3F,$FB,$41,$BD,$5F,$D4,$5B
	.dc.b	$21,$C1,$FC,$AA,$BF,$DA,$FF,$5D,$FE,$1D,$9F,$F7,$23,$5F,$C7,$36
	.dc.b	$37,$DB,$FB,$28,$FC,$F4,$EA,$30,$7E,$DB,$EA,$BD,$71,$D3,$EA,$38
	.dc.b	$11,$EF,$F0,$FC,$C8,$E0,$FF,$7B,$7D,$77,$9F,$4F,$ED,$98,$BE,$2F
	.dc.b	$C4,$4A,$F4,$FB,$5E,$52,$38,$3D,$BF,$0F,$E9,$23,$83,$E3,$6B,$FE
	.dc.b	$FA,$FA,$EF,$D1,$46,$31,$FE,$F9,$F5,$DF,$EF,$EA,$FF,$7F,$B6,$32
	.dc.b	$7F,$16,$5F,$49,$49,$5C,$CF,$A9,$1A,$70,$3E,$BB,$DC,$E2,$3F,$BC
	.dc.b	$37,$D3,$EC,$77,$F8,$49,$1B,$6B,$FC,$1F,$8B,$ED,$63,$EA,$FA,$DF
	.dc.b	$31,$27,$B7,$C3,$7A,$48,$E0,$F1,$7A,$DE,$D2,$38,$3D,$3E,$3F,$F1
	.dc.b	$6D,$BE,$2F,$A7,$F0,$C9,$3E,$BF,$4F,$E3,$11,$C1,$F0,$B3,$74,$7C
	.dc.b	$BF,$9C,$93,$F0,$DF,$9D,$F0,$92,$3E,$33,$5F,$E4,$DF,$5C,$DB,$FA
	.dc.b	$4F,$D0,$7F,$81,$74,$70,$7F,$8D,$E3,$0D,$F8,$9D,$F9,$93,$1F,$AB
	.dc.b	$5F,$C5,$DF,$E6,$4D,$BE,$71,$FA,$F8,$FF,$06,$3A,$F3,$3C,$24,$7E
	.dc.b	$FF,$C4,$BC,$CF,$2F,$D2,$C9,$EE,$FA,$3E,$92,$47,$89,$AF,$FA,$7B
	.dc.b	$CB,$F6,$F9,$B2,$3C,$EF,$C0,$F3,$64,$7A,$9A,$FF,$A6,$F9,$3E,$2F
	.dc.b	$97,$23,$DE,$FC,$8F,$72,$47,$C2,$6B,$F0,$BE,$67,$E6,$91,$F8,$EF
	.dc.b	$F7,$8F,$2C,$FF,$7A,$FA,$E7,$81,$49,$5F,$A2,$87,$C7,$6F,$D5,$93
	.dc.b	$97,$97,$67,$F8,$13,$EB,$9F,$1B,$49,$5F,$AC,$87,$C6,$EF,$86,$4E
	.dc.b	$5E,$5C,$3F,$E0,$EF,$AE,$FE,$53,$26,$93,$E0,$75,$F2,$F2,$7F,$73
	.dc.b	$28,$E5,$F3,$DF,$E1,$8F,$6D,$F2,$54,$97,$FC,$02,$1D,$DB,$F2,$24
	.dc.b	$E5,$E5,$E4,$7F,$88,$BE,$B9,$74,$95,$FC,$2A,$3F,$18,$DE,$DE,$04
	.dc.b	$75,$C5,$CF,$DB,$91,$C5,$FF,$17,$FD,$77,$F8,$37,$C7,$FF,$4E,$2F
	.dc.b	$90,$C9,$2B,$F8,$CC,$3E,$43,$FC,$6F,$F5,$DF,$EC,$F9,$FE,$3A,$D7
	.dc.b	$FC,$34,$DF,$23,$F5,$30,$86,$FF,$E4,$0F,$4B,$F5,$75,$A6,$E5,$FE
	.dc.b	$AF,$6D,$FF,$5F,$7E,$2B,$FC,$B1,$78,$7C,$C9,$19,$F0,$FB,$7C,$04
	.dc.b	$8E,$0F,$87,$6F,$F2,$8F,$D7,$3E,$22,$92,$BF,$9C,$C7,$C5,$7F,$9D
	.dc.b	$FD,$77,$F4,$57,$FB,$93,$53,$4F,$E8,$4F,$F8,$F5,$FF,$55,$37,$E3
	.dc.b	$FF,$CC,$31,$8D,$7E,$3B,$F2,$75,$FF,$68,$3F,$22,$BC,$3B,$70,$38
	.dc.b	$F3,$C8,$F2,$B5,$F0,$7C,$77,$6F,$12,$47,$A9,$F0,$F8,$92,$3D,$AD
	.dc.b	$7D,$1E,$A7,$F9,$DB,$EB,$BF,$23,$1F,$E5,$76,$E9,$F7,$9B,$A7,$FB
	.dc.b	$9D,$79,$DC,$49,$5F,$D9,$23,$2F,$F9,$33,$8B,$FD,$07,$F3,$5F,$E0
	.dc.b	$E8,$FF,$32,$37,$0F,$C5,$B7,$0F,$F7,$A3,$FB,$C2,$7C,$87,$23,$F8
	.dc.b	$B7,$27,$FA,$3B,$EB,$3F,$9E,$57,$FB,$ED,$B8,$7F,$8A,$DA,$9F,$C8
	.dc.b	$B7,$D4,$E4,$47,$FB,$EC,$7E,$51,$FF,$3D,$37,$1F,$D0,$B6,$BC,$1E
	.dc.b	$0F,$FA,$72,$BC,$9E,$9C,$79,$BE,$7D,$A4,$76,$6B,$DB,$C1,$FF,$59
	.dc.b	$2F,$AE,$6D,$ED,$9C,$CF,$FA,$4F,$24,$62,$35,$9F,$FA,$A6,$72,$DF
	.dc.b	$09,$F5,$D4,$95,$FE,$9D,$3E,$34,$69,$FF,$59,$0F,$A7,$FF,$4C,$11
	.dc.b	$CB,$AF,$90,$FE,$DB,$7F,$C1,$A7,$FC,$CF,$BF,$F8,$A1,$FF,$4C,$37
	.dc.b	$E3,$76,$FF,$AB,$5B,$1F,$E6,$3F,$3B,$F9,$7C,$EC,$FF,$18,$B8,$8F
	.dc.b	$F5,$70,$E4,$E2,$FF,$9A,$FC,$EA,$E6,$BC,$51,$83,$3F,$35,$63,$CC
	.dc.b	$E1,$EF,$F2,$F6,$BF,$D7,$98,$23,$BE,$5F,$CF,$AB,$FC,$8A,$FF,$81
	.dc.b	$DB,$CB,$AE,$59,$1E,$AE,$EB,$8F,$47,$D3,$F5,$DF,$1A,$3F,$D9,$4F
	.dc.b	$CA,$6E,$FA,$DE,$5F,$C4,$ED,$B5,$E6,$79,$F8,$9D,$3C,$BB,$5F,$E5
	.dc.b	$AC,$7F,$A3,$36,$3C,$3E,$56,$E3,$4F,$93,$E3,$BC,$BE,$8D,$9E,$4E
	.dc.b	$1E,$9E,$8D,$CF,$D2,$FB,$3E,$33,$DB,$F1,$96,$BC,$4F,$D8,$F9,$9F
	.dc.b	$F0,$5C,$7F,$9B,$97,$D7,$67,$DA,$D3,$BE,$0E,$1E,$45,$BE,$0F,$E0
	.dc.b	$BE,$BB,$FC,$17,$1F,$E7,$85,$F5,$D9,$F6,$F4,$EF,$B3,$87,$CA,$DB
	.dc.b	$E0,$FE,$33,$EB,$9E,$05,$F0,$30,$DA,$FF,$42,$CB,$8B,$87,$17,$16
	.dc.b	$CF,$FA,$2B,$EB,$9E,$7F,$E6,$17,$05,$8C,$3F,$9C,$9F,$F1,$49,$FE
	.dc.b	$92,$3D,$BC,$AF,$07,$FB,$AC,$6A,$30,$FF,$B3,$5F,$F3,$5F,$07,$E9
	.dc.b	$BE,$BB,$FD,$A1,$3F,$E9,$C7,$F2,$15,$FD,$F9,$CD,$72,$3F,$DA,$CE
	.dc.b	$A7,$A3,$FA,$D8,$D0,$7E,$3C,$75,$F1,$3C,$1B,$78,$7A,$7C,$5A,$DA
	.dc.b	$F1,$3F,$BA,$F1,$5E,$C5,$F8,$2C,$36,$BF,$D6,$73,$FE,$12,$6F,$89
	.dc.b	$C7,$D4,$E4,$FF,$04,$FA,$EF,$DB,$6D,$E6,$7F,$6C,$DF,$23,$5E,$46
	.dc.b	$FD,$9E,$1F,$09,$72,$FF,$60,$E1,$FE,$66,$6F,$A9,$FC,$43,$CB,$F8
	.dc.b	$FC,$BF,$E2,$6A,$4E,$73,$D7,$91,$BF,$22,$E3,$FE,$CC,$C7,$FC,$68
	.dc.b	$FF,$B3,$BE,$BB,$FD,$AA,$DF,$81,$FE,$3B,$18,$C3,$E7,$BF,$E5,$89
	.dc.b	$FF,$6A,$FD,$77,$E7,$9B,$F0,$9F,$C9,$23,$15,$F5,$94,$64,$D7,$89
	.dc.b	$6B,$9B,$87,$FD,$C3,$F5,$DF,$EF,$D7,$26,$3C,$2E,$DF,$01,$DF,$8E
	.dc.b	$CF,$91,$C3,$CB,$CD,$B9,$FE,$23,$04,$7F,$3A,$FA,$CF,$97,$5F,$99
	.dc.b	$7F,$21,$40,$E9,$C3,$A7,$3B,$FF,$44,$B7,$CC,$F5,$8C,$3D,$FC,$0F
	.dc.b	$07,$91,$71,$FF,$7F,$E2,$8C,$C6,$3D,$9B,$97,$E8,$2B,$FB,$BC,$E4
	.dc.b	$F8,$FF,$57,$EB,$AF,$13,$ED,$F9,$DF,$77,$AB,$6B,$BB,$87,$E7,$FD
	.dc.b	$75,$C7,$E3,$3F,$17,$D5,$F8,$DE,$3D,$AF,$3B,$87,$E3,$77,$7F,$9E
	.dc.b	$3E,$BB,$FD,$D7,$1A,$32,$7F,$05,$73,$87,$AF,$8F,$E9,$70,$FD,$36
	.dc.b	$D7,$C4,$8F,$AD,$AE,$2B,$E1,$38,$9E,$B1,$F5,$FE,$BB,$F7,$71,$F3
	.dc.b	$27,$FB,$BC,$62,$7F,$84,$FD,$77,$E7,$72,$FF,$47,$BF,$9D,$8C,$FC
	.dc.b	$F7,$F7,$E8,$C4,$EB,$BB,$EF,$71,$65,$B9,$78,$B0,$5F,$FB,$72,$30
	.dc.b	$FF,$50,$8F,$0F,$95,$D7,$D8,$95,$C7,$AB,$CD,$97,$92,$7F,$C1,$F1
	.dc.b	$89,$FE,$5D,$83,$E7,$3C,$DF,$39,$E7,$6C,$F8,$23,$D5,$F1,$58,$CF
	.dc.b	$E0,$B8,$9F,$06,$76,$7F,$96,$E3,$E9,$7E,$3F,$97,$F0,$FE,$27,$E2
	.dc.b	$36,$FD,$B1,$FE,$24,$F9,$67,$CC,$66,$FF,$9B,$C6,$2B,$E1,$FB,$BF
	.dc.b	$E2,$AE,$17,$BF,$F3,$E3,$0C,$9F,$99,$D0,$FA,$8C,$63,$FD,$0A,$31
	.dc.b	$3F,$47,$EB,$BF,$98,$F2,$FF,$B2,$9F,$E6,$2D,$F4,$BF,$C7,$31,$89
	.dc.b	$FE,$B0,$DB,$FE,$01,$F5,$CE,$3D,$FA,$7F,$E4,$08,$C5,$F7,$F8,$6E
	.dc.b	$7E,$DF,$46,$C7,$8E,$38,$7D,$5E,$31,$FE,$B3,$68,$F1,$F9,$9E,$4E
	.dc.b	$7B,$87,$AD,$C3,$F8,$DF,$AE,$7C,$76,$31,$FE,$C7,$18,$9F,$C9,$7B
	.dc.b	$BF,$E3,$BE,$1F,$F7,$5B,$FD,$A5,$BF,$11,$FE,$5A,$8C,$4F,$F7,$96
	.dc.b	$DF,$F2,$1B,$7C,$E6,$E7,$FC,$C3,$18,$AE,$FE,$7F,$3F,$3F,$A3,$63
	.dc.b	$C7,$1C,$35,$C6,$3E,$87,$C5,$EC,$78,$FD,$8B,$7E,$A7,$0F,$EB,$7E
	.dc.b	$B5,$F5,$D8,$C3,$FC,$07,$18,$9F,$D9,$7D,$5F,$F9,$A7,$B7,$FD,$21
	.dc.b	$48,$C3,$FE,$93,$5F,$34,$F5,$B7,$1A,$9F,$F0,$E1,$C7,$C6,$FF,$0F
	.dc.b	$61,$89,$46,$27,$FC,$45,$18,$7F,$C3,$A7,$F4,$B7,$0F,$FA,$67,$DF
	.dc.b	$FD,$D3,$A0,$E1,$D7,$93,$D5,$FE,$A9,$19,$BF,$17,$4F,$FA,$AA,$FF
	.dc.b	$19,$5E,$A8,$C1,$FE,$AE,$FA,$D7,$55,$24,$FC,$68,$CB,$F2,$30,$7C
	.dc.b	$0F,$FA,$D9,$B1,$98,$C1,$9F,$87,$91,$C5,$AF,$FB,$07,$E7,$DF,$43
	.dc.b	$FE,$C0,$6C,$7D,$66,$BB,$8D,$78,$B2,$34,$D7,$A7,$CB,$73,$23,$CD
	.dc.b	$F8,$9E,$74,$8F,$BA,$D7,$F0,$4F,$F9,$67,$C0,$70,$7D,$A8,$C3,$67
	.dc.b	$F1,$BE,$76,$1F,$B8,$FD,$12,$3F,$8A,$6B,$F8,$F6,$BF,$89,$7E,$17
	.dc.b	$C7,$73,$8F,$CB,$45,$E2,$39,$D3,$FD,$1B,$3F,$D3,$DA,$FC,$DA,$E6
	.dc.b	$FC,$B7,$FD,$DB,$F5,$DF,$E8,$C8,$FE,$AC,$D7,$FC,$D6,$DC,$BE,$0C
	.dc.b	$7F,$88,$63,$FC,$28,$F0,$EE,$E6,$47,$F5,$B1,$A7,$FD,$FC,$BD,$4F
	.dc.b	$63,$9A,$EB,$28,$C6,$7F,$E8,$97,$84,$79,$BD,$EF,$49,$1E,$ED,$7D
	.dc.b	$2F,$63,$BF,$BD,$23,$CA,$FC,$22,$3E,$8F,$03,$4F,$C2,$8F,$3C,$7D
	.dc.b	$3F,$AE,$FE,$03,$8F,$A1,$B8,$FF,$A7,$DB,$7F,$E4,$77,$F8,$53,$FE
	.dc.b	$03,$72,$9A,$4E,$CE,$C7,$F1,$68,$D3,$63,$9C,$7F,$12,$32,$7F,$16
	.dc.b	$FA,$7F,$C3,$11,$FE,$43,$CD,$F8,$92,$38,$3F,$11,$F0,$BF,$8F,$FA
	.dc.b	$EF,$D7,$E5,$F5,$1B,$97,$FA,$ED,$B5,$FE,$63,$7F,$A8,$BF,$E1,$F6
	.dc.b	$BF,$EB,$46,$CF,$3F,$2F,$F3,$1F,$0D,$DF,$A6,$5F,$17,$C5,$E8,$48
	.dc.b	$E8,$D7,$FD,$7F,$CC,$F9,$2F,$46,$92,$7C,$4F,$AF,$09,$1E,$56,$BC
	.dc.b	$BC,$DF,$81,$F3,$72,$3D,$4F,$C3,$F8,$B9,$1E,$D6,$BE,$2F,$33,$F6
	.dc.b	$DF,$5C,$F4,$69,$27,$FC,$85,$0F,$47,$64,$8C,$BF,$BB,$F1,$FF,$80
	.dc.b	$7D,$6F,$F8,$76,$7F,$C9,$4D,$7F,$DB,$ED,$E3,$FD,$4A,$EC,$E2,$49
	.dc.b	$E2,$FF,$0B,$FA,$AF,$34,$74,$FA,$8E,$E4,$7B,$FC,$4F,$88,$8E,$0F
	.dc.b	$F2,$E7,$D7,$7F,$44,$7C,$7F,$1D,$E8,$23,$E8,$F7,$7D,$64,$70,$7B
	.dc.b	$3E,$27,$DE,$47,$07,$F9,$A7,$EB,$BF,$3F,$AF,$F0,$6F,$AD,$FE,$0E
	.dc.b	$DE,$FF,$E4,$D1,$87,$FD,$6D,$48,$7C,$4E,$89,$1F,$E7,$58,$CB,$FE
	.dc.b	$77,$FA,$E6,$7D,$3E,$5A,$4E,$DE,$FE,$BC,$8E,$4D,$59,$7F,$1C,$E1
	.dc.b	$FA,$D9,$F9,$32,$7D,$6F,$A0,$97,$E3,$F6,$3D,$64,$70,$79,$FE,$2B
	.dc.b	$DE,$47,$07,$FA,$3F,$C8,$FF,$80,$2F,$B7,$D6,$F8,$A4,$7E,$07,$A9
	.dc.b	$F2,$08,$E0,$FF,$4C,$66,$E7,$F9,$BF,$35,$27,$E4,$3F,$43,$E2,$24
	.dc.b	$7C,$B6,$BF,$0B,$6F,$A7,$F5,$DF,$02,$3F,$D4,$9E,$EF,$FA,$11,$BF
	.dc.b	$19,$FD,$7E,$31,$1C,$4F,$F7,$83,$FC,$41,$FF,$1A,$FB,$7B,$23,$D3
	.dc.b	$CE,$E4,$47,$07,$FA,$CF,$E3,$1E,$77,$7F,$72,$4F,$91,$F1,$7D,$24
	.dc.b	$8F,$3B,$5F,$F4,$CF,$99,$E7,$F3,$A4,$7B,$1F,$57,$9D,$23,$DE,$D7
	.dc.b	$FD,$2D,$C5,$EF,$F3,$24,$7C,$47,$E1,$79,$32,$3E,$3B,$5F,$88,$7F
	.dc.b	$95,$BD,$EF,$90,$47,$E6,$BF,$E0,$53,$FC,$15,$F5,$DF,$E4,$EE,$FF
	.dc.b	$ED,$ED,$E5,$FD,$3B,$68,$17,$C8,$72,$F5,$A4,$FF,$B7,$65,$1D,$BA
	.dc.b	$7F,$C3,$1F,$5C,$F8,$CA,$4B,$FF,$72,$43,$E3,$37,$C3,$27,$2F,$2E
	.dc.b	$1F,$F1,$17,$D7,$7F,$29,$9F,$F7,$63,$5F,$86,$DD,$BF,$C1,$6F,$F8
	.dc.b	$67,$6F,$E0,$6B,$E5,$B9,$92,$B9,$FC,$F7,$F8,$BF,$DB,$7C,$9D,$24
	.dc.b	$FF,$BE,$E3,$FD,$0A,$DF,$13,$81,$1D,$78,$9F,$E3,$9F,$AE,$55,$24
	.dc.b	$FC,$ED,$3F,$C4,$33,$F9,$48,$C6,$9F,$C6,$A1,$D7,$CC,$93,$A3,$EB
	.dc.b	$C7,$CF,$70,$1F,$E0,$6F,$E2,$1B,$5E,$0F,$6B,$FC,$9D,$F5,$DF,$E1
	.dc.b	$41,$F0,$58,$25,$F4,$87,$C7,$FF,$95,$7E,$BB,$FD,$1B,$3F,$94,$6E
	.dc.b	$1F,$E3,$C6,$E7,$FC,$D1,$FF,$43,$CF,$E6,$BD,$1F,$F0,$2C,$FF,$38
	.dc.b	$6C,$7E,$49,$A9,$F9,$6F,$B9,$77,$FF,$99,$DB,$19,$8C,$77,$E7,$DF
	.dc.b	$93,$C5,$E1,$AF,$F9,$BF,$EB,$9F,$0D,$49,$5F,$32,$3E,$2B,$F3,$BE
	.dc.b	$BB,$FA,$83,$FA,$86,$A6,$9F,$D3,$9F,$DC,$AF,$96,$FF,$3F,$36,$3A
	.dc.b	$46,$3E,$43,$3C,$F2,$7C,$0F,$0E,$79,$1E,$36,$BD,$BE,$3B,$AF,$87
	.dc.b	$23,$D0,$F8,$3C,$39,$1E,$B6,$BE,$6F,$47,$FD,$25,$F5,$DF,$95,$8F
	.dc.b	$E2,$2D,$E6,$FF,$3C,$B7,$97,$FB,$8A,$F2,$F8,$92,$BF,$2B,$19,$7F
	.dc.b	$D3,$1C,$7F,$E9,$FF,$8D,$FF,$0A,$C7,$F1,$E6,$E1,$FE,$86,$6E,$1F
	.dc.b	$DF,$9F,$E0,$0D,$F8,$5E,$44,$AF,$F5,$27,$2F,$FA,$AF,$EB,$7F,$A5
	.dc.b	$13,$D2,$C7,$A5,$2B,$F5,$33,$F9,$02,$78,$5C,$49,$1A,$7F,$D3,$6F
	.dc.b	$5F,$46,$3C,$9F,$06,$F2,$3A,$B5,$FE,$77,$5F,$BE,$D7,$F9,$8F,$D7
	.dc.b	$7F,$A8,$25,$39,$7F,$D0,$0D,$7F,$C3,$E6,$BC,$4F,$F0,$8B,$A6,$BA
	.dc.b	$8F,$E6,$9F,$5D,$FE,$89,$8E,$BE,$72,$31,$3A,$9E,$F3,$97,$FD,$29
	.dc.b	$F4,$FF,$E9,$54,$D2,$3D,$7E,$43,$FD,$93,$97,$FA,$FA,$3F,$CC,$DC
	.dc.b	$5F,$C5,$94,$EA,$3D,$7F,$19,$67,$FC,$76,$D5,$E7,$3D,$8C,$47,$43
	.dc.b	$FB,$2C,$66,$BB,$72,$EB,$E6,$76,$D8,$F1,$47,$2F,$93,$C5,$FF,$6E
	.dc.b	$0E,$67,$F9,$6D,$BE,$BB,$15,$E0,$7C,$DA,$7F,$40,$AF,$E6,$6F,$F8
	.dc.b	$41,$BC,$BA,$E5,$91,$D3,$C0,$B8,$7A,$9F,$81,$F5,$DF,$1A,$7F,$9E
	.dc.b	$BE,$FC,$3F,$DD,$2E,$6B,$FC,$0A,$E6,$7F,$CC,$6D,$8F,$EF,$B9,$3B
	.dc.b	$FA,$F0,$7C,$D8,$C5,$F6,$F1,$FC,$3E,$77,$89,$B1,$E2,$8F,$2F,$9B
	.dc.b	$8C,$7C,$2F,$A3,$E3,$7D,$3E,$9D,$BF,$4B,$87,$D7,$E9,$DB,$F4,$BF
	.dc.b	$03,$E1,$FE,$0F,$C6,$5A,$F8,$2F,$E0,$FE,$67,$FC,$2B,$1F,$D8,$17
	.dc.b	$D7,$67,$DD,$D3,$BE,$1E,$1E,$55,$BE,$0F,$E3,$7E,$BB,$FC,$29,$1F
	.dc.b	$DA,$97,$CC,$E7,$DD,$D3,$BE,$4E,$1F,$2B,$6F,$83,$F9,$7F,$AE,$77
	.dc.b	$DE,$FC,$36,$BF,$BB,$CB,$8B,$87,$17,$16,$CF,$F7,$CF,$AE,$FF,$17
	.dc.b	$B7,$7F,$D2,$46,$39,$3F,$C8,$AF,$DA,$7F,$C8,$F3,$FE,$33,$3F,$C0
	.dc.b	$A7,$B9,$CF,$8B,$FA,$D8,$C7,$69,$FF,$55,$FD,$77,$F9,$11,$FF,$06
	.dc.b	$BF,$8F,$8D,$57,$F9,$22,$34,$FF,$95,$9D,$4F,$2F,$0F,$F7,$51,$F1
	.dc.b	$23,$9C,$63,$5C,$F4,$EB,$4B,$5E,$07,$F8,$17,$8D,$FE,$03,$8F,$F0
	.dc.b	$FA,$F9,$0C,$FB,$9A,$77,$EB,$70,$F1,$2D,$F0,$7F,$86,$7D,$77,$ED
	.dc.b	$F6,$FF,$36,$BF,$B7,$6F,$8F,$AF,$1F,$7E,$E7,$0F,$86,$B8,$FF,$8C
	.dc.b	$30,$47,$4F,$53,$F8,$C7,$9B,$F1,$39,$7E,$3E,$6B,$56,$A3,$7E,$47
	.dc.b	$0F,$22,$E5,$FE,$3C,$C7,$FC,$62,$FF,$8F,$BE,$BB,$FD,$BA,$DF,$83
	.dc.b	$FE,$4D,$18,$C3,$FD,$0E,$FF,$97,$27,$FC,$8B,$F5,$DF,$9C,$6F,$C3
	.dc.b	$7F,$2D,$8C,$57,$FA,$39,$46,$4D,$78,$76,$B9,$B8,$7F,$CA,$3F,$5D
	.dc.b	$FE,$EA,$72,$63,$C1,$ED,$F0,$7D,$F8,$EC,$F9,$1C,$3C,$BC,$DB,$9F
	.dc.b	$FA,$83,$04,$7F,$43,$FA,$EF,$99,$5F,$F0,$01,$FC,$7D,$7F,$A6,$9C
	.dc.b	$B7,$7E,$07,$F4,$BB,$7C,$CE,$2F,$D2,$F5,$F0,$3B,$78,$F6,$7F,$CD
	.dc.b	$B8,$FF,$44,$6C,$78,$FD,$DD,$C6,$9C,$3F,$13,$EB,$2E,$7E,$B7,$D5
	.dc.b	$F3,$9E,$BF,$52,$D7,$67,$0F,$CC,$FA,$EB,$7F,$0D,$F8,$5E,$AF,$C3
	.dc.b	$F1,$ED,$79,$9C,$3F,$E8,$4F,$AE,$FD,$E3,$93,$07,$CB,$E2,$F9,$9C
	.dc.b	$16,$3D,$EE,$1F,$83,$A5,$CB,$FD,$A7,$BA,$3E,$A1,$FD,$DF,$EB,$BF
	.dc.b	$6D,$3A,$8C,$78,$FF,$DE,$FE,$BB,$F2,$DE,$17,$E9,$1F,$CB,$57,$FB
	.dc.b	$79,$CE,$3F,$ED,$48,$46,$A7,$87,$77,$DB,$E3,$E8,$DB,$FF,$51,$D7
	.dc.b	$F0,$56,$FA,$FE,$1F,$27,$C4,$C8,$FF,$54,$F5,$7A,$31,$F2,$4F,$F8
	.dc.b	$46,$31,$3E,$3B,$7C,$D7,$AF,$E6,$BD,$8D,$9F,$0C,$7B,$FE,$2F,$1A
	.dc.b	$FB,$5C,$4F,$83,$3B,$3F,$EF,$EC,$7D,$0F,$C3,$F2,$FE,$47,$C5,$7C
	.dc.b	$96,$DF,$BA,$3F,$C4,$DF,$2C,$F9,$7C,$DB,$7F,$E2,$33,$9B,$F8,$BE
	.dc.b	$E7,$F8,$B7,$85,$EF,$7D,$58,$C3,$27,$FD,$9B,$D2,$FA,$EC,$65,$D6
	.dc.b	$7E,$B4,$E5,$FF,$69,$FD,$77,$F2,$DC,$FF,$90,$BF,$CB,$5B,$E9,$7F
	.dc.b	$8F,$63,$13,$F7,$DB,$61,$F5,$DF,$EA,$99,$EB,$E9,$FF,$91,$63,$13
	.dc.b	$E3,$38,$57,$F6,$17,$01,$FE,$BF,$5D,$FD,$9F,$17,$BF,$23,$FD,$D7
	.dc.b	$6F,$90,$C6,$1C,$A3,$13,$EE,$F3,$3D,$5C,$DE,$B6,$C7,$C2,$8F,$F2
	.dc.b	$D7,$D7,$3E,$37,$38,$79,$23,$15,$FE,$FA,$F7,$3F,$87,$4F,$A1,$BC
	.dc.b	$1F,$D2,$9F,$ED,$51,$9F,$33,$AE,$52,$F9,$45,$9E,$DE,$BF,$88,$FF
	.dc.b	$36,$7D,$6F,$F2,$68,$65,$FD,$52,$FC,$5E,$1F,$F9,$C7,$23,$C6,$32
	.dc.b	$7D,$5C,$FF,$AE,$E1,$E4,$C3,$FD,$2A,$D7,$F7,$4D,$7E,$E3,$57,$D6
	.dc.b	$7F,$9B,$7A,$3F,$D2,$CD,$8F,$F4,$F9,$AF,$8A,$FF,$44,$B8,$8D,$3F
	.dc.b	$EA,$45,$F4,$1F,$C2,$70,$FE,$E3,$EE,$FF,$BC,$DB,$F4,$99,$CF,$F8
	.dc.b	$75,$F0,$C3,$D9,$18,$C3,$51,$F9,$3C,$39,$4F,$F1,$56,$C7,$E5,$EB
	.dc.b	$93,$F9,$06,$E1,$E0,$8E,$A7,$3C,$3A,$BF,$0B,$0B,$5C,$30,$ED,$5A
	.dc.b	$C3,$BD,$EC,$E6,$53,$F3,$EE,$8D,$D7,$38,$F3,$5F,$3A,$91,$CA,$74
	.dc.b	$F3,$C3,$D0,$F4,$9F,$2D,$48,$F0,$DF,$79,$C4,$7C,$19,$EA,$F1,$3B
	.dc.b	$3A,$AE,$A4,$C3,$E8,$4E,$8E,$B5,$CA,$7F,$C5,$78,$79,$8F,$3E,$1E
	.dc.b	$1A,$E0,$78,$4F,$C5,$E1,$FE,$E1,$7E,$31,$F8,$77,$13,$F6,$23,$07
	.dc.b	$02,$1F,$8F,$8F,$46,$1E,$68,$FB,$DF,$2B,$87,$E6,$9E,$2C,$3C,$F1
	.dc.b	$98,$F9,$93,$C6,$73,$1B,$8C,$11,$FC,$64,$F1,$BF,$81,$7F,$86,$8C
	.dc.b	$39,$6C,$F3,$38,$94,$F0,$9E,$63,$D2,$66,$7A,$BF,$83,$8E,$6F,$CB
	.dc.b	$36,$5B,$FA,$75,$E1,$8E,$4D,$7F,$98,$FD,$5B,$E3,$3E,$93,$63,$F7
	.dc.b	$A1,$C3,$FC,$01,$AF,$EF,$7E,$1F,$F4,$F1,$AB,$FF,$36,$B7,$2F,$DD
	.dc.b	$CF,$EB,$3E,$AD,$DD,$7B,$E6,$3F,$A8,$9C,$B6,$78,$11,$FD,$6F,$D5
	.dc.b	$BE,$2F,$F6,$FF,$58,$F8,$4A,$4A,$E9,$F2,$DE,$13,$0F,$F5,$29,$E0
	.dc.b	$E8,$F1,$7C,$1F,$9C,$8F,$97,$E0,$FD,$44,$70,$79,$3E,$8F,$DD,$F3
	.dc.b	$32,$3D,$8F,$C1,$EF,$91,$EF,$6B,$FD,$93,$47,$83,$FC,$4B,$EB,$3E
	.dc.b	$54,$FF,$71,$6A,$68,$F0,$38,$12,$3F,$89,$C6,$5F,$93,$E2,$7F,$1E
	.dc.b	$FA,$CF,$F0,$54,$7F,$18,$6B,$FD,$D9,$BD,$1F,$EA,$85,$F0,$7C,$49
	.dc.b	$3C,$5F,$E5,$3F,$4F,$E6,$97,$D3,$77,$25,$77,$F8,$3F,$11,$1C,$1F
	.dc.b	$CA,$BE,$B1,$97,$9F,$E2,$A4,$F9,$DF,$57,$A3,$23,$D4,$D7,$D5,$F0
	.dc.b	$7E,$F2,$3E,$F7,$F9,$FF,$D6,$7F,$5A,$8F,$F0,$AF,$D5,$FF,$B1,$B7
	.dc.b	$FF,$63,$B7,$C1,$FF,$31,$36,$7F,$C4,$8F,$F4,$22,$79,$1F,$D2,$9F
	.dc.b	$E8,$BF,$58,$CF,$1F,$07,$2F,$FA,$3D,$23,$87,$BD,$27,$2D,$7F,$C5
	.dc.b	$3C,$7F,$53,$2F,$1F,$D6,$77,$24,$F4,$F8,$2F,$11,$1C,$1E,$0F,$82
	.dc.b	$F3,$11,$C1,$FD,$6B,$CB,$FF,$72,$BE,$6F,$93,$F6,$11,$F4,$FC,$17
	.dc.b	$BC,$8E,$0F,$EC,$79,$7E,$09,$F7,$F9,$7F,$14,$8F,$C2,$F7,$BE,$41
	.dc.b	$1C,$1F,$DB,$3E,$B1,$A7,$F7,$6F,$4F,$FC,$DD,$3F,$DC,$1A,$FF,$93
	.dc.b	$9B,$CB,$FF,$00,$17,$D1,$F1,$24,$F1,$7A,$7B,$52,$3A,$35,$FF,$2C
	.dc.b	$FC,$17,$F1,$67,$B7,$99,$E3,$23,$E3,$FA,$2F,$31,$1C,$1F,$E0,$69
	.dc.b	$79,$5E,$BF,$32,$4F,$AD,$F7,$79,$92,$3D,$CD,$7F,$CD,$7E,$07,$CD
	.dc.b	$7E,$07,$93,$F1,$48,$FC,$5F,$0B,$E4,$11,$C1,$F2,$7D,$7F,$98,$47
	.dc.b	$07,$F3,$07,$FC,$4B,$F5,$8E,$DA,$49,$FF,$10,$C3,$E2,$77,$EC,$49
	.dc.b	$CB,$CB,$A3,$FC,$5F,$F5,$8F,$83,$A4,$AF,$F1,$5C,$3E,$0F,$7C,$12
	.dc.b	$72,$F2,$E0,$FF,$1C,$FD,$67,$F1,$F9,$FF,$47,$B7,$A5,$E1,$FF,$B6
	.dc.b	$E3,$0D,$F9,$3E,$3F,$23,$2F,$FC,$77,$48,$EB,$F3,$7F,$E4,$6F,$75
	.dc.b	$F1,$54,$97,$FE,$42,$87,$7E,$FC,$59,$39,$79,$78,$BF,$E5,$0F,$AC
	.dc.b	$5D,$25,$7F,$93,$23,$F2,$1B,$7A,$A3,$FD,$69,$0E,$FE,$44,$8E,$1C
	.dc.b	$9D,$7D,$99,$1C,$DF,$F2,$FF,$D6,$7F,$BF,$63,$FD,$6A,$BE,$6F,$24
	.dc.b	$8F,$F2,$EC,$7F,$68,$7F,$D7,$7F,$59,$FE,$B5,$7F,$CC,$2D,$73,$F1
	.dc.b	$9F,$E6,$C8,$6D,$FE,$6E,$F3,$FF,$D6,$55,$FE,$68,$6B,$FE,$C6,$6E
	.dc.b	$BE,$F8,$FE,$4D,$B7,$1F,$0A,$47,$17,$A7,$85,$D7,$DE,$93,$D5,$F0
	.dc.b	$6D,$FE,$7E,$FA,$C7,$CC,$D2,$57,$F9,$EA,$3E,$13,$FE,$7C,$FA,$CF
	.dc.b	$E8,$09,$A4,$F5,$FF,$D1,$2B,$E5,$29,$27,$97,$0E,$B8,$24,$75,$78
	.dc.b	$F3,$C8,$F0,$35,$FF,$74,$DB,$E5,$3B,$F8,$B2,$7C,$CF,$8B,$C5,$91
	.dc.b	$E9,$6B,$CB,$D2,$FF,$4F,$FD,$67,$E3,$3C,$3F,$F7,$F4,$A3,$0F,$FC
	.dc.b	$01,$6B,$9F,$3F,$89,$1F,$F5,$0C,$23,$FD,$57,$F0,$8E,$4F,$F5,$6F
	.dc.b	$C7,$FF,$86,$89,$F1,$98,$F8,$69,$8F,$FA,$AA,$7F,$6A,$8F,$F7,$0C
	.dc.b	$74,$E7,$FF,$5B,$FD,$57,$F3,$C9,$FF,$58,$B5,$76,$7E,$26,$30,$DF
	.dc.b	$57,$89,$27,$8B,$D3,$B6,$47,$46,$BF,$E4,$93,$C3,$F1,$71,$F0,$BE
	.dc.b	$4D,$A4,$79,$1A,$F9,$38,$3E,$7E,$3F,$B9,$C7,$DD,$F5,$9F,$EF,$EA
	.dc.b	$F8,$13,$97,$FD,$E2,$D7,$F7,$0D,$79,$F8,$BF,$01,$DF,$8A,$FA,$CF
	.dc.b	$F3,$ED,$7F,$2F,$6C,$66,$77,$1F,$EB,$55,$F4,$94,$8F,$FA,$E5,$FF
	.dc.b	$5D,$7D,$0F,$F1,$3C,$BF,$6D,$3F,$E8,$DC,$FF,$DC,$6A,$75,$1F,$C3
	.dc.b	$F8,$EC,$FE,$41,$AB,$D0,$FE,$53,$3C,$A7,$31,$89,$F9,$47,$E6,$38
	.dc.b	$F8,$AF,$42,$C6,$19,$7E,$4F,$96,$B5,$C8,$FC,$73,$C8,$EF,$F5,$96
	.dc.b	$7F,$DF,$B8,$23,$CF,$9F,$F3,$89,$FF,$80,$2F,$FA,$79,$BC,$9A,$E4
	.dc.b	$91,$CB,$C0,$B9,$78,$71,$47,$C2,$FA,$CF,$91,$5F,$A1,$C7,$F9,$14
	.dc.b	$7C,$DB,$F4,$3A,$6E,$DC,$4F,$8F,$DD,$6F,$D2,$F5,$F1,$FB,$79,$B6
	.dc.b	$7B,$1F,$C0,$B0,$FE,$84,$E3,$0F,$D9,$46,$1F,$0F,$C8,$F2,$9D,$42
	.dc.b	$7F,$D6,$91,$85,$F1,$9E,$DF,$53,$DC,$D9,$D4,$7C,$1F,$53,$17,$EE
	.dc.b	$7E,$27,$C4,$7C,$5F,$7A,$DF,$57,$F1,$3F,$2D,$FE,$11,$8F,$1A,$FA
	.dc.b	$CC,$FB,$BA,$77,$C1,$C3,$C9,$B7,$C1,$FC,$9F,$D6,$7F,$84,$63,$DA
	.dc.b	$BE,$B3,$3E,$F6,$9D,$F4,$70,$F9,$6B,$7C,$1F,$CE,$FD,$67,$E5,$63
	.dc.b	$E3,$AE,$EC,$F7,$E9,$DF,$17,$0E,$2B,$7C,$1F,$D3,$FD,$63,$D0,$BC
	.dc.b	$18,$6D,$7D,$43,$FC,$5C,$5D,$BF,$5E,$E3,$97,$F4,$4F,$BD,$FD,$27
	.dc.b	$D6,$7F,$31,$8F,$BA,$FE,$3F,$BB,$FB,$A3,$99,$DF,$89,$FB,$A8,$FF
	.dc.b	$6A,$57,$0D,$B1,$E2,$65,$D7,$56,$F7,$B5,$D9,$FE,$07,$E2,$FF,$A9
	.dc.b	$63,$F1,$AF,$F9,$45,$BE,$DE,$BE,$17,$7E,$B7,$0F,$16,$D7,$07,$F8
	.dc.b	$6F,$D6,$7E,$DF,$7F,$E0,$0F,$ED,$DB,$E4,$6B,$C8,$DF,$B9,$C3,$E1
	.dc.b	$EE,$3F,$9E,$C1,$1D,$3D,$3F,$E3,$3C,$9F,$0F,$97,$FC,$31,$49,$CE
	.dc.b	$7A,$F2,$77,$E4,$DC,$BF,$53,$8F,$EA,$2B,$F8,$A7,$D6,$3D,$DF,$E4
	.dc.b	$AF,$F9,$42,$BF,$5E,$E5,$BF,$09,$FC,$AB,$EB,$3F,$2D,$1F,$B4,$5F
	.dc.b	$0F,$9F,$1B,$4C,$9D,$78,$D7,$2F,$E4,$97,$FC,$CE,$72,$7E,$FB,$EB
	.dc.b	$3F,$DB,$8E,$27,$F9,$54,$61,$E3,$F0,$7E,$3E,$4F,$23,$6B,$C9,$1F
	.dc.b	$D0,$29,$E7,$66,$F9,$BF,$D1,$3E,$AB,$E6,$5F,$F8,$59,$FC,$93,$67
	.dc.b	$5C,$12,$3C,$DF,$31,$67,$CA,$F3,$F0,$7A,$79,$36,$BF,$8C,$E1,$FB
	.dc.b	$36,$C7,$87,$DD,$DC,$69,$C3,$F0,$FE,$A6,$E3,$F0,$9F,$47,$CE,$FA
	.dc.b	$7D,$6B,$5D,$9C,$3F,$2F,$EB,$2D,$FC,$A7,$E0,$7A,$DF,$07,$C8,$B5
	.dc.b	$E6,$70,$FF,$30,$FA,$CF,$F4,$EB,$93,$07,$C9,$E1,$F9,$5C,$16,$3D
	.dc.b	$EE,$1F,$83,$A5,$CB,$FB,$26,$E8,$FA,$47,$F7,$6F,$AC,$FD,$A4,$EA
	.dc.b	$31,$E4,$7F,$7A,$FA,$CF,$CA,$4F,$F6,$D7,$F3,$0D,$EE,$D7,$6C,$CB
	.dc.b	$F1,$9C,$76,$BF,$16,$F3,$F2,$7A,$7B,$16,$BF,$A7,$E3,$FC,$FE,$FB
	.dc.b	$4F,$F8,$2B,$AB,$CB,$E2,$FF,$A2,$6B,$4F,$FA,$2E,$33,$7F,$E6,$09
	.dc.b	$7C,$DF,$A7,$E6,$FD,$4D,$9E,$F1,$ED,$F8,$CC,$65,$ED,$71,$3F,$D8
	.dc.b	$B0,$FE,$2D,$39,$8C,$67,$F1,$7C,$CF,$8D,$F1,$56,$7D,$6E,$1F,$ED
	.dc.b	$7F,$2C,$FA,$CC,$61,$FE,$17,$8C,$4F,$C2,$F7,$7F,$C5,$3C,$4F,$7F
	.dc.b	$E9,$46,$19,$3F,$DD,$BA,$9F,$59,$8C,$7F,$C4,$51,$89,$FE,$F7,$F5
	.dc.b	$9F,$C9,$F9,$3F,$D6,$4F,$F2,$76,$FA,$7F,$E3,$88,$C4,$FF,$A2,$9B
	.dc.b	$7F,$E7,$9F,$AC,$FF,$52,$4F,$4F,$53,$FC,$7F,$18,$9F,$F1,$95,$7F
	.dc.b	$53,$C3,$BF,$B5,$D7,$E1,$FB,$33,$1F,$F0,$7E,$0F,$91,$C6,$3F,$E3
	.dc.b	$B8,$C4,$FA,$BC,$DF,$37,$3F,$9D,$B1,$EE,$8F,$F2,$8F,$D6,$3E,$3B
	.dc.b	$38,$7F,$91,$63,$15,$FE,$1E,$F7,$7F,$83,$4F,$FB,$A4,$38,$7F,$DD
	.dc.b	$8D,$7F,$CA,$DE,$6F,$F9,$99,$FF,$4E,$53,$1C,$42,$31,$7D,$63,$13
	.dc.b	$A9,$CC,$78,$5F,$E9,$D3,$B6,$C6,$7D,$8E,$07,$6C,$1E,$15,$25,$F7
	.dc.b	$8C,$BE,$1F,$AB,$FF,$30,$73,$DF,$8A,$FC,$76,$3F,$06,$5F,$31,$FE
	.dc.b	$7F,$8C,$4E,$9F,$F6,$68,$FA,$CF,$C8,$BE,$D6,$BF,$2D,$AF,$06,$BF
	.dc.b	$ED,$6A,$FF,$10,$3F,$E5,$05,$F4,$5F,$ED,$AE,$66,$5B,$8C,$3C,$7E
	.dc.b	$12,$DF,$FB,$36,$F9,$BF,$E8,$37,$9B,$D6,$B6,$A3,$77,$DA,$39,$56
	.dc.b	$A7,$8C,$73,$AC,$4C,$8E,$84,$3C,$0E,$5E,$06,$6B,$B3,$E1,$8D,$46
	.dc.b	$2B,$BB,$6B,$37,$EF,$BE,$4F,$C3,$7E,$45,$E9,$F8,$53,$E9,$30,$F7
	.dc.b	$38,$9A,$38,$90,$F3,$92,$BC,$93,$CA,$FC,$B7,$E1,$AF,$79,$C6,$BD
	.dc.b	$0F,$63,$C0,$68,$FC,$75,$7E,$8D,$C9,$F8,$CA,$FB,$2D,$8D,$46,$23
	.dc.b	$E1,$1C,$8F,$CE,$47,$94,$FB,$BF,$29,$FE,$26,$8F,$25,$F0,$BF,$5C
	.dc.b	$7F,$A7,$A7,$4F,$96,$32,$D9,$32,$77,$3A,$24,$F5,$37,$29,$DC,$66
	.dc.b	$3F,$1A,$79,$0E,$9E,$DF,$A5,$C1,$C5,$C3,$E3,$36,$6B,$0F,$CC,$CF
	.dc.b	$58,$E0,$67,$0F,$C0,$DF,$8A,$3C,$27,$33,$EA,$E1,$F9,$48,$D4,$62
	.dc.b	$BA,$5F,$68,$F0,$9C,$5C,$3B,$33,$5F,$9B,$F8,$6F,$D7,$FD,$63,$E3
	.dc.b	$E9,$FA,$D8,$95,$FC,$55,$FB,$E4,$7F,$94,$7D,$63,$BF,$F1,$0D,$8C
	.dc.b	$C6,$08,$FF,$B6,$A3,$4B,$0F,$D8,$C6,$1F,$81,$F1,$BF,$C1,$21,$F2
	.dc.b	$BF,$C1,$BE,$B1,$F0,$B4,$D4,$4C,$3C,$AF,$90,$8F,$A3,$87,$A1,$1F
	.dc.b	$27,$C1,$F7,$23,$83,$F2,$9F,$43,$E3,$F9,$89,$1E,$B7,$CD,$D9,$23
	.dc.b	$DC,$D7,$DD,$C7,$F9,$2E,$4F,$EA,$91,$FE,$FC,$FA,$CF,$A4,$DE,$3F
	.dc.b	$E4,$71,$88,$E2,$7E,$81,$FE,$C6,$F1,$EC,$D9,$1F,$D5,$78,$5F,$CB
	.dc.b	$3E,$B3,$FA,$0F,$27,$F1,$E6,$E5,$FA,$46,$D7,$EB,$17,$C1,$F1,$24
	.dc.b	$7E,$C2,$B9,$47,$F3,$8F,$A7,$F3,$47,$5F,$AC,$F0,$12,$B8,$61,$DF
	.dc.b	$E0,$FC,$84,$8C,$9E,$17,$FA,$1F,$D6,$33,$F5,$7C,$AC,$AD,$8D,$11
	.dc.b	$EF,$F4,$24,$EA,$7F,$D4,$2D,$8F,$89,$F0,$72,$34,$D7,$E7,$8F,$AC
	.dc.b	$FE,$E0,$E3,$EA,$FF,$A3,$ED,$FD,$21,$BA,$FF,$0E,$6C,$FF,$16,$7F
	.dc.b	$D9,$CF,$A1,$AF,$3F,$81,$FE,$C3,$18,$AF,$EA,$7F,$59,$EC,$AE,$1F
	.dc.b	$07,$A2,$3C,$FD,$EE,$24,$70,$7C,$6E,$3F,$BA,$B0,$7D,$5F,$80,$91
	.dc.b	$E5,$F8,$2F,$19,$1C,$1E,$9F,$82,$F3,$91,$C1,$F5,$3C,$BF,$F7,$A3
	.dc.b	$EF,$F9,$5F,$69,$1F,$85,$F0,$5F,$00,$8E,$0F,$47,$4F,$C4,$F2,$A4
	.dc.b	$7C,$97,$E3,$FB,$D2,$3E,$63,$5F,$B3,$F5,$9E,$E7,$F0,$FE,$97,$F9
	.dc.b	$FD,$BF,$05,$FE,$10,$8C,$47,$E0,$17,$D6,$64,$91,$C9,$E1,$ED,$48
	.dc.b	$EA,$D7,$FD,$15,$E8,$75,$F3,$24,$78,$9F,$07,$D1,$48,$F2,$B5,$FF
	.dc.b	$49,$79,$7E,$5F,$36,$47,$A9,$F4,$79,$B2,$3D,$AD,$7F,$D1,$FC,$1E
	.dc.b	$DF,$6A,$47,$C2,$7E,$07,$8B,$23,$E2,$B5,$F8,$5F,$31,$F1,$48,$FC
	.dc.b	$A7,$FB,$D9,$FE,$38,$FA,$C7,$6D,$24,$FE,$DA,$1F,$15,$BF,$62,$4E
	.dc.b	$5E,$5C,$FF,$E4,$2F,$AC,$7C,$1D,$25,$7F,$00,$87,$C1,$EF,$82,$4E
	.dc.b	$5E,$5C,$1F,$E4,$CF,$AC,$FE,$4F,$3F,$BC,$6F,$8D,$FE,$40,$6F,$91
	.dc.b	$FE,$06,$5F,$29,$C7,$E4,$A5,$FF,$10,$94,$75,$F9,$DF,$F2,$CF,$C4
	.dc.b	$BE,$2E,$92,$FF,$8B,$C7,$F9,$E9,$BE,$27,$02,$3A,$F1,$3F,$CC,$7F
	.dc.b	$58,$AA,$49,$FE,$43,$1F,$D9,$F4,$FF,$0E,$AE,$DE,$24,$9D,$F1,$F4
	.dc.b	$F6,$E4,$72,$7F,$CE,$3F,$59,$FE,$8A,$E1,$FF,$31,$AF,$9D,$C9,$2B
	.dc.b	$F9,$8C,$7F,$72,$7F,$99,$FD,$67,$FB,$1D,$BF,$1B,$FE,$7C,$8C,$47
	.dc.b	$F2,$C5,$BF,$FA,$03,$D1,$FF,$62,$4F,$F3,$C6,$BF,$CC,$9B,$8F,$F9
	.dc.b	$D9,$F8,$91,$BF,$07,$91,$1E,$5E,$0F,$5F,$80,$91,$D5,$F0,$6B,$FE
	.dc.b	$90,$FA,$C7,$CD,$52,$57,$F4,$E8,$F8,$4F,$F5,$0F,$AC,$FE,$7A,$9A
	.dc.b	$4F,$5F,$FD,$36,$BE,$56,$92,$79,$70,$6B,$79,$1D,$5E,$3D,$12,$3C
	.dc.b	$0D,$7F,$AA,$DB,$E5,$7B,$F8,$92,$7C,$CF,$8B,$C4,$91,$E9,$6B,$CB
	.dc.b	$D2,$FF,$58,$FD,$67,$E2,$E3,$FC,$DA,$D4,$D3,$FC,$66,$FF,$72,$6B
	.dc.b	$FD,$A1,$BC,$BF,$EA,$C6,$F6,$7F,$88,$78,$FF,$D7,$FF,$21,$FE,$08
	.dc.b	$8F,$F3,$AB,$6F,$FC,$FA,$D7,$FD,$D7,$F0,$9C,$08,$FF,$80,$A3,$F7
	.dc.b	$2F,$FA,$0F,$97,$FD,$97,$F5,$7F,$CE,$A3,$FC,$0E,$D7,$FB,$EB,$72
	.dc.b	$FF,$60,$BF,$8E,$C3,$FD,$0E,$D9,$FF,$47,$B5,$34,$FC,$9B,$D3,$C0
	.dc.b	$F0,$91,$FF,$0B,$FC,$57,$EB,$6B,$AF,$A5,$E6,$24,$F9,$BD,$2F,$49
	.dc.b	$1C,$1E,$AF,$0B,$FC,$D0,$BE,$B1,$AF,$A2,$73,$3F,$EA,$03,$CE,$E1
	.dc.b	$D6,$99,$7D,$B1,$97,$DF,$E1,$BE,$B3,$51,$F0,$23,$0F,$1D,$BF,$9B
	.dc.b	$CE,$9B,$1F,$EC,$BF,$A6,$FF,$66,$26,$91,$EB,$F4,$3F,$EF,$FC,$7F
	.dc.b	$DE,$71,$FE,$95,$DB,$FC,$7E,$A7,$51,$ED,$F8,$CB,$3F,$EB,$E6,$AF
	.dc.b	$39,$ED,$62,$3A,$1F,$E3,$E8,$CD,$76,$E5,$D7,$9F,$B6,$C7,$8A,$39
	.dc.b	$7C,$BE,$2F,$FC,$8E,$1C,$CF,$F9,$5A,$DF,$2B,$8A,$F0,$3E,$6D,$3F
	.dc.b	$9F,$57,$F9,$71,$FF,$56,$37,$9F,$5C,$D2,$3A,$78,$17,$0F,$53,$F0
	.dc.b	$3E,$B3,$E3,$4F,$F9,$99,$F7,$E0,$8F,$F7,$1C,$6A,$BF,$CB,$EE,$A3
	.dc.b	$3B,$FC,$AC,$DD,$DD,$63,$0F,$7F,$1F,$C1,$E6,$DB,$F0,$38,$7C,$5F
	.dc.b	$23,$72,$EE,$F9,$FE,$3B,$D1,$E9,$DA,$F4,$38,$7D,$5E,$9D,$BF,$3B
	.dc.b	$EF,$F8,$DF,$81,$F1,$96,$BE,$03,$F8,$2F,$97,$FF,$08,$C7,$FA,$21
	.dc.b	$7D,$66,$7D,$CD,$3B,$DF,$87,$93,$6F,$83,$F8,$CF,$AC,$FF,$08,$C7
	.dc.b	$FA,$49,$7D,$3E,$7D,$DD,$3B,$E5,$E1,$F2,$F6,$F8,$3F,$96,$FA,$CF
	.dc.b	$F0,$7C,$7F,$A7,$1F,$C5,$37,$BF,$5D,$FB,$E2,$E1,$C5,$6B,$83,$FA
	.dc.b	$0F,$AC,$7A,$17,$83,$0D,$AF,$F5,$49,$FE,$2E,$2E,$DF,$A8,$71,$CB
	.dc.b	$FE,$A2,$7D,$EF,$FA,$93,$EB,$0F,$EB,$9F,$C5,$77,$7E,$36,$33,$1A
	.dc.b	$9F,$F5,$BB,$A7,$8C,$21,$E1,$F3,$DB,$A3,$6B,$8F,$4F,$7D,$BC,$0D
	.dc.b	$8F,$F6,$1F,$8C,$F6,$FF,$80,$BF,$E5,$48,$CC,$63,$E1,$33,$EB,$6B
	.dc.b	$C4,$B5,$C9,$FE,$13,$F5,$9F,$B4,$DB,$FD,$86,$FE,$D1,$BE,$3E,$BC
	.dc.b	$7D,$FB,$9C,$3E,$1A,$E3,$FE,$D6,$C1,$1D,$3D,$5F,$E2,$BE,$6F,$C5
	.dc.b	$CB,$AC,$61,$CF,$91,$AF,$22,$E3,$FE,$E0,$B4,$7F,$1E,$FA,$C7,$BB
	.dc.b	$7E,$0F,$0D,$BF,$F7,$3F,$D6,$7E,$59,$BF,$0D,$FC,$9A,$31,$9F,$FB
	.dc.b	$81,$46,$4D,$78,$B6,$B9,$B8,$7F,$DE,$7F,$59,$FE,$F2,$72,$63,$C3
	.dc.b	$ED,$F0,$7D,$F8,$EC,$F9,$1C,$3C,$BC,$EB,$9F,$FB,$CB,$04,$7F,$3C
	.dc.b	$FA,$AF,$91,$E1,$7E,$3E,$91,$F8,$D7,$F2,$AE,$AF,$F1,$EE,$63,$7C
	.dc.b	$5F,$D1,$B0,$7C,$CF,$58,$C3,$DF,$C2,$F0,$79,$17,$1F,$D0,$E5,$E1
	.dc.b	$CB,$F6,$F9,$23,$15,$E7,$F6,$F8,$DF,$03,$C6,$E5,$F9,$3F,$53,$6F
	.dc.b	$E0,$3E,$FF,$9D,$F8,$1E,$AD,$AE,$55,$EE,$8C,$C7,$D2,$FA,$CB,$3F
	.dc.b	$25,$F9,$1E,$B7,$C9,$F1,$ED,$72,$AF,$93,$19,$8C,$7D,$67,$F4,$77
	.dc.b	$5B,$7F,$A8,$1C,$9F,$4F,$8F,$EA,$70,$D8,$DB,$87,$8E,$B7,$1F,$8D
	.dc.b	$1F,$52,$3E,$E1,$FD,$E3,$EB,$3F,$75,$1C,$0F,$33,$89,$E2,$7F,$12
	.dc.b	$FA,$CF,$D0,$CF,$AD,$FF,$59,$37,$B3,$FC,$07,$18,$9E,$21,$C7,$1F
	.dc.b	$E2,$F9,$BF,$C5,$77,$CF,$B7,$F0,$DC,$9C,$BC,$8E,$7E,$C5,$AF,$03
	.dc.b	$87,$E6,$75,$79,$F1,$FC,$B6,$5F,$D4,$B3,$F0,$CE,$6B,$F9,$1C,$BE
	.dc.b	$6B,$CF,$F3,$5E,$86,$CF,$B6,$3D,$7F,$19,$8C,$3D,$77,$FE,$8F,$AF
	.dc.b	$85,$3F,$E2,$2C,$3F,$91,$56,$A3,$19,$FC,$4F,$2F,$E2,$FC,$55,$C3
	.dc.b	$E5,$38,$7F,$0D,$F5,$8F,$99,$C6,$3F,$2A,$31,$3F,$47,$DD,$FF,$18
	.dc.b	$F1,$3D,$FF,$A3,$18,$64,$FE,$47,$A5,$F5,$98,$C7,$EA,$C6,$27,$F2
	.dc.b	$DF,$59,$FC,$CF,$93,$F8,$03,$FC,$CD,$BE,$97,$F9,$02,$31,$3F,$D3
	.dc.b	$9B,$7F,$D6,$3E,$B3,$E3,$3F,$82,$8C,$A8,$CD,$F6,$F8,$9E,$9E,$DF
	.dc.b	$56,$E1,$E4,$8F,$F2,$7D,$3E,$47,$38,$7E,$3A,$31,$5E,$AF,$2F,$CD
	.dc.b	$CF,$E7,$6C,$7B,$A3,$FC,$B3,$F5,$8F,$9C,$CE,$1F,$98,$8C,$57,$EE
	.dc.b	$7D,$DF,$F7,$B3,$FE,$76,$6C,$FF,$9F,$5A,$FF,$A0,$DA,$FF,$72,$9F
	.dc.b	$EF,$AF,$FA,$1D,$AE,$73,$FF,$39,$BF,$E0,$08,$CB,$7E,$13,$2A,$3F
	.dc.b	$A5,$BC,$E3,$31,$88,$FF,$00,$FD,$5F,$F9,$93,$93,$FD,$1A,$D8,$FF
	.dc.b	$4A,$3F,$E7,$8F,$43,$3F,$15,$F8,$B9,$FF,$4C,$7D,$57,$E4,$DF,$F4
	.dc.b	$AB,$5C,$F5,$FF,$A6,$14,$65,$FF,$52,$3B,$E4,$FE,$94,$DD,$BF,$C2
	.dc.b	$AD,$9F,$F0,$FF,$D6,$7F,$9F,$5F,$E5,$11,$FE,$C9,$9F,$F0,$DC,$FF
	.dc.b	$AC,$A7,$F9,$63,$5F,$E1,$F3,$B7,$FC,$52,$D5,$1B,$AF,$EB,$8D,$AF
	.dc.b	$1F,$D0,$7F,$8A,$27,$CE,$D8,$F0,$F8,$D6,$FC,$ED,$7E,$0C,$DC,$3C
	.dc.b	$18,$75,$9D,$9E,$C9,$D1,$C6,$35,$1E,$6A,$CB,$B3,$DE,$EC,$CB,$F0
	.dc.b	$27,$0E,$8E,$06,$4C,$1F,$CD,$27,$BD,$7C,$53,$CF,$5C,$A7,$81,$30
	.dc.b	$F8,$F1,$97,$E2,$C7,$67,$2F,$67,$2F,$1C,$3E,$4B,$97,$E7,$BE,$63
	.dc.b	$F0,$F8,$7E,$2E,$CD,$24,$7E,$49,$AF,$E5,$9B,$5E,$02,$57,$E3,$B8
	.dc.b	$BE,$3C,$62,$BF,$0B,$5F,$5E,$73,$5E,$49,$F4,$1F,$8A,$8C,$BC,$0F
	.dc.b	$81,$1D,$5A,$64,$E2,$7E,$92,$38,$BF,$86,$36,$64,$B7,$F4,$EB,$C9
	.dc.b	$18,$78,$C7,$F1,$78,$C1,$C2,$73,$3A,$6C,$64,$D9,$E4,$9E,$4F,$EA
	.dc.b	$3D,$FF,$CC,$75,$48,$FD,$8C,$7E,$6C,$FA,$84,$7F,$81,$BF,$C0,$7E
	.dc.b	$B3,$FD,$1B,$1F,$E9,$03,$FD,$22,$E1,$C0,$70,$ED,$AF,$0E,$5F,$D6
	.dc.b	$47,$1C,$FF,$CA,$4D,$CF,$AF,$AD,$FA,$FF,$92,$FD,$97,$CA,$BE,$82
	.dc.b	$99,$CB,$F8,$5C,$FE,$34,$8F,$F1,$C7,$F2,$1E,$E7,$EE,$49,$1F,$C4
	.dc.b	$BD,$CF,$DE,$11,$C1,$FE,$3D,$FA,$C7,$D1,$7F,$01,$F9,$99,$3F,$C6
	.dc.b	$9F,$E0,$9F,$09,$23,$F8,$EB,$5F,$F1,$3D,$3C,$1F,$E1,$9F,$58,$F9
	.dc.b	$4A,$4A,$FF,$26,$47,$C3,$7F,$D8,$ED,$4D,$31,$4F,$0F,$F8,$C7,$D5
	.dc.b	$FF,$81,$2B,$FC,$AC,$D7,$FC,$72,$DE,$4F,$F4,$12,$F8,$1E,$24,$9E
	.dc.b	$2F,$F2,$3F,$A9,$F4,$47,$4E,$2E,$E4,$7B,$FC,$07,$88,$8E,$0F,$F3
	.dc.b	$47,$D6,$79,$DF,$2F,$C2,$FA,$08,$FA,$3C,$EF,$59,$1C,$1E,$CF,$7F
	.dc.b	$DE,$47,$07,$F9,$D7,$EB,$1E,$3F,$F3,$BF,$AB,$FF,$50,$6F,$FE,$A1
	.dc.b	$6F,$6F,$FB,$B1,$B7,$FE,$F6,$7F,$D8,$AF,$F9,$FD,$AF,$F9,$89,$B3
	.dc.b	$CF,$DF,$FE,$91,$F5,$8D,$3C,$1F,$21,$30,$E4,$F8,$FD,$B9,$1D,$5A
	.dc.b	$FF,$9B,$B8,$BE,$B6,$7E,$5F,$AC,$F5,$12,$7C,$9F,$01,$ED,$23,$83
	.dc.b	$D1,$F0,$1F,$04,$8E,$0F,$F4,$D7,$D6,$3B,$BE,$37,$C9,$49,$F8,$0F
	.dc.b	$C9,$F8,$09,$1F,$0D,$AF,$9F,$A7,$E7,$7C,$BC,$8F,$90,$FD,$1F,$7A
	.dc.b	$47,$CB,$6B,$FE,$8C,$FA,$C6,$BF,$DC,$7E,$7F,$FC,$15,$26,$93,$D2
	.dc.b	$D1,$23,$FD,$69,$1F,$E0,$53,$F9,$1F,$3F,$FB,$7D,$BC,$29,$91,$86
	.dc.b	$FA,$5F,$E0,$1F,$82,$7A,$3E,$3F,$A3,$95,$D9,$F3,$79,$D2,$3C,$2D
	.dc.b	$7F,$CF,$DE,$67,$AF,$CE,$91,$E6,$7D,$DE,$74,$8F,$4B,$5F,$F3,$DF
	.dc.b	$1F,$C3,$F3,$64,$7B,$9F,$8D,$E5,$48,$F8,$2D,$7F,$D6,$FB,$3E,$57
	.dc.b	$E6,$12,$3E,$3B,$FE,$1F,$F3,$35,$1F,$ED,$DF,$AC,$78,$1F,$E2,$36
	.dc.b	$C6,$63,$1F,$13,$9F,$9B,$93,$C5,$E1,$D7,$FE,$2B,$FA,$C7,$C2,$D2
	.dc.b	$57,$FB,$AE,$1F,$07,$BE,$29,$39,$79,$71,$7F,$8D,$7E,$B3,$F9,$21
	.dc.b	$3D,$AC,$7B,$39,$F9,$29,$87,$FB,$DE,$91,$C7,$E7,$3F,$C7,$DE,$EB
	.dc.b	$E2,$E9,$2F,$FD,$FD,$1F,$91,$6F,$8F,$C0,$8E,$BC,$7F,$F2,$47,$D6
	.dc.b	$2A,$92,$7F,$21,$E2,$3C,$4F,$C9,$42,$BF,$1E,$EA,$33,$0F,$03,$A9
	.dc.b	$27,$47,$E6,$E3,$E6,$B8,$3A,$79,$3E,$42,$39,$FA,$CF,$EF,$98,$78
	.dc.b	$E3,$F4,$90,$F9,$CF,$F3,$13,$63,$84,$63,$E7,$3F,$CC,$DF,$59,$FE
	.dc.b	$B9,$9F,$03,$55,$F5,$51,$98,$C2,$DB,$FC,$E1,$E9,$7E,$CE,$8D,$26
	.dc.b	$DA,$F1,$24,$FB,$E7,$FB,$73,$7C,$4E,$7F,$03,$23,$4F,$6D,$BF,$CF
	.dc.b	$DF,$58,$F9,$BA,$4A,$F9,$51,$F0,$5F,$5F,$1F,$D8,$FA,$CF,$E8,$F0
	.dc.b	$69,$3D,$9F,$F4,$52,$F9,$5A,$49,$FA,$D1,$F0,$EB,$A6,$FC,$09,$3A
	.dc.b	$E7,$E6,$47,$07,$2F,$95,$EF,$E2,$C8,$EA,$F8,$BC,$59,$1E,$06,$BE
	.dc.b	$C3,$F9,$26,$2F,$57,$FD,$45,$F5,$9F,$94,$9D,$07,$1C,$FF,$75,$BA
	.dc.b	$7E,$71,$7A,$1C,$89,$5F,$95,$8C,$BF,$E2,$FE,$3F,$F5,$87,$C5,$FF
	.dc.b	$83,$A3,$BB,$70,$FC,$53,$70,$FD,$4B,$FE,$25,$27,$89,$C8,$FE,$39
	.dc.b	$CB,$FE,$BB,$FA,$AF,$F6,$41,$3D,$6C,$7A,$D3,$2F,$D4,$4F,$E4,$49
	.dc.b	$D9,$C4,$91,$A7,$FC,$AA,$F5,$F5,$63,$C9,$F0,$6F,$23,$AB,$5F,$97
	.dc.b	$5F,$BD,$D7,$FA,$A7,$D6,$7F,$B8,$63,$13,$FE,$6E,$6B,$FE,$73,$6B
	.dc.b	$FC,$29,$AF,$F5,$D7,$F9,$7F,$D6,$7F,$9D,$A1,$18,$7A,$79,$E9,$D4
	.dc.b	$E6,$78,$7D,$2F,$17,$FB,$43,$53,$4D,$7C,$1F,$F0,$BC,$7F,$80,$4F
	.dc.b	$F9,$7F,$8B,$F8,$A3,$FE,$34,$6F,$C6,$6D,$FF,$47,$36,$3F,$BB,$79
	.dc.b	$FF,$E0,$19,$D9,$FC,$75,$C4,$7E,$50,$71,$F1,$7F,$BF,$79,$F5,$CD
	.dc.b	$78,$A3,$06,$7E,$8A,$C7,$99,$C3,$DF,$E4,$ED,$7E,$8B,$04,$77,$CB
	.dc.b	$FA,$05,$7F,$31,$7F,$C1,$2D,$E3,$D7,$1C,$8F,$57,$7D,$C7,$A3,$E9
	.dc.b	$FA,$CF,$8D,$1F,$B0,$7E,$53,$77,$D6,$F2,$FE,$27,$75,$AF,$33,$CF
	.dc.b	$C5,$E9,$E5,$DA,$FE,$8D,$8F,$F3,$F6,$C7,$87,$CB,$DC,$69,$F2,$7C
	.dc.b	$67,$97,$D2,$B3,$C9,$C3,$D3,$D2,$B9,$FA,$5F,$67,$C3,$7B,$7E,$2E
	.dc.b	$D7,$89,$FB,$1F,$2C,$F6,$AF,$D6,$61,$B5,$FD,$76,$5C,$1C,$3C,$8E
	.dc.b	$2D,$9F,$EC,$5F,$59,$FE,$0C,$6F,$CD,$7E,$16,$31,$AF,$F1,$C5,$19
	.dc.b	$EB,$D7,$D6,$5B,$E6,$FE,$37,$EB,$1D,$F7,$BF,$0D,$AF,$EE,$52,$E2
	.dc.b	$E1,$C5,$C5,$B3,$FD,$DF,$EB,$1E,$7F,$E6,$5F,$6C,$66,$31,$BF,$E7
	.dc.b	$63,$FC,$56,$7F,$7F,$3D,$7E,$07,$F6,$F7,$3C,$1F,$CE,$1C,$9F,$E1
	.dc.b	$0F,$AC,$FF,$31,$3F,$E0,$77,$F1,$F3,$FD,$DD,$CC,$FF,$40,$73,$3C
	.dc.b	$7C,$3F,$D7,$47,$C2,$8E,$51,$8D,$73,$D1,$AD,$2D,$77,$7F,$6D,$E2
	.dc.b	$7F,$7E,$8F,$F0,$C3,$FE,$50,$6F,$AF,$AF,$82,$DF,$A9,$C3,$C4,$B5
	.dc.b	$C1,$FE,$01,$F5,$9F,$B2,$DB,$FA,$BB,$FB,$26,$F8,$FA,$F1,$F7,$EC
	.dc.b	$F0,$F8,$6B,$8F,$F8,$A7,$0F,$F3,$43,$7D,$2F,$E1,$9E,$57,$C4,$E5
	.dc.b	$FF,$0E,$57,$FA,$A6,$BF,$C6,$13,$97,$FC,$7A,$FE,$B7,$3F,$ED,$2B
	.dc.b	$C6,$E7,$E3,$F4,$7F,$8C,$7D,$63,$DA,$BF,$03,$86,$E3,$FE,$3D,$FA
	.dc.b	$CF,$CB,$B7,$E1,$3F,$90,$46,$33,$FE,$EE,$A3,$26,$BC,$4B,$5C,$DC
	.dc.b	$3F,$E4,$9F,$A7,$FF,$6D,$39,$31,$E1,$76,$F8,$0E,$FC,$96,$7C,$8E
	.dc.b	$1E,$5E,$6D,$CF,$FC,$13,$82,$3F,$9B,$7D,$57,$C7,$9F,$F5,$A3,$F8
	.dc.b	$E6,$9A,$E2,$97,$E5,$F9,$8B,$3D,$1E,$7E,$0F,$4F,$22,$D7,$F9,$8F
	.dc.b	$04,$66,$31,$EE,$5C,$3F,$C3,$B4,$8F,$29,$FD,$2F,$37,$D5,$62,$FE
	.dc.b	$33,$EA,$F9,$BF,$5F,$AB,$67,$C0,$E1,$F9,$9F,$59,$72,$F8,$6F,$C2
	.dc.b	$F5,$7E,$1F,$91,$6B,$D0,$E1,$FF,$3E,$7D,$67,$FA,$25,$C9,$83,$E5
	.dc.b	$F1,$7C,$CE,$1B,$1F,$05,$C3,$F0,$F4,$B9,$7C,$07,$E9,$9F,$DA,$7E
	.dc.b	$B3,$F6,$91,$99,$FF,$46,$46,$3C,$0F,$F3,$E7,$D6,$7E,$5E,$7F,$D2
	.dc.b	$0F,$FA,$AD,$BD,$9F,$DC,$E3,$17,$FE,$4A,$6D,$7F,$A1,$78,$7E,$85
	.dc.b	$F2,$ED,$FA,$1C,$7C,$7C,$8E,$5E,$BD,$AE,$EE,$1F,$F5,$07,$57,$95
	.dc.b	$FF,$56,$E5,$FD,$2D,$CF,$95,$E3,$F9,$AF,$27,$CD,$59,$F4,$B8,$7D
	.dc.b	$7F,$13,$77,$F4,$3E,$B3,$FC,$1D,$8F,$F1,$78,$CC,$F8,$B3,$F8,$1E
	.dc.b	$67,$C1,$F8,$9B,$3F,$19,$C3,$FE,$B7,$F9,$77,$D3,$63,$0F,$F3,$7C
	.dc.b	$62,$7E,$6F,$B9,$FE,$1D,$E2,$7B,$BF,$36,$30,$C9,$FF,$61,$75,$3E
	.dc.b	$B3,$18,$FF,$9E,$A3,$13,$FE,$C7,$FA,$CF,$E5,$1D,$1F,$EB,$77,$F9
	.dc.b	$43,$7D,$3F,$F1,$64,$62,$7F,$DB,$2D,$B0,$FA,$BF,$F5,$04,$F5,$F5
	.dc.b	$3F,$C6,$91,$89,$FF,$45,$D7,$F5,$6C,$3C,$1F,$2B,$DB,$E2,$3B,$B3
	.dc.b	$1F,$F6,$E6,$0F,$AC,$C6,$3F,$E9,$48,$C4,$FA,$FC,$DF,$3F,$3F,$A1
	.dc.b	$B1,$EF,$8F,$F2,$37,$D6,$3E,$37,$38,$7F,$A7,$A3,$15,$FE,$EB,$7F
	.dc.b	$C0,$F1,$FE,$F8,$0E,$1F,$F7,$DB,$5F,$86,$D7,$75,$FE,$64,$72,$D7
	.dc.b	$33,$FE,$77,$94,$E7,$EB,$3F,$C9,$FC,$7F,$E7,$B6,$DF,$8C,$2F,$9C
	.dc.b	$7A,$3B,$F8,$BF,$C3,$D6,$F1,$7B,$9F,$E7,$D6,$C7,$86,$30,$7F,$A0
	.dc.b	$7E,$B3,$F9,$4C,$7F,$A2,$DA,$F8,$5A,$FC,$56,$BF,$84,$0F,$0E,$36
	.dc.b	$B9,$8C,$37,$D3,$E1,$3A,$AF,$C6,$D3,$D2,$CD,$F5,$FA,$CF,$B5,$5F
	.dc.b	$92,$BE,$0A,$1F,$25,$C2,$3F,$2F,$18,$78,$57,$3E,$2D,$47,$16,$CE
	.dc.b	$E3,$FD,$C6,$DB,$EE,$F8,$A3,$33,$C2,$E4,$76,$3C,$31,$D0,$CB,$D2
	.dc.b	$30,$EC,$F1,$16,$7C,$17,$F0,$67,$73,$E0,$34,$F4,$BF,$2D,$E9,$F3
	.dc.b	$C6,$6B,$E0,$57,$5A,$C3,$F1,$A7,$84,$EA,$3C,$C7,$63,$81,$0C,$92
	.dc.b	$BC,$2E,$5F,$0B,$C4,$F4,$11,$CB,$C6,$3C,$11,$CA,$3F,$41,$7F,$3D
	.dc.b	$FD,$A3,$5F,$31,$8B,$FF,$6D,$46,$63,$E8,$9F,$EE,$57,$0F,$CE,$3C
	.dc.b	$55,$E0,$3F,$35,$F3,$CF,$67,$E1,$CF,$23,$67,$F2,$28,$D3,$EF,$9E
	.dc.b	$4E,$F0,$FC,$AC,$EA,$38,$9E,$18,$F7,$1A,$AE,$F1,$9A,$E6,$AB,$95
	.dc.b	$70,$9C,$DF,$E4,$E3,$95,$78,$CF,$14,$F2,$AE,$2F,$0C,$3B,$C6,$27
	.dc.b	$64,$3C,$47,$BF,$0F,$2E,$1D,$EF,$A4,$62,$F9,$B6,$BC,$47,$A5,$FD
	.dc.b	$6F,$5C,$8F,$CF,$3F,$B2,$C1,$F1,$59,$9F,$D8,$CD,$CB,$E5,$CF,$EE
	.dc.b	$3E,$B1,$F0,$D4,$97,$FA,$B8,$FF,$5C,$B7,$DB,$FE,$54,$FF,$5A,$8C
	.dc.b	$BF,$CB,$5A,$FF,$41,$6B,$FC,$47,$E7,$3F,$1A,$47,$F1,$8D,$7F,$83
	.dc.b	$BF,$C2,$8F,$E1,$BF,$58,$F6,$BF,$86,$FB,$F2,$BF,$7C,$FF,$10,$DA
	.dc.b	$47,$F0,$16,$AD,$3F,$17,$F4,$5F,$C5,$BE,$62,$4F,$F0,$77,$F8,$CF
	.dc.b	$C3,$C8,$FE,$14,$D7,$97,$1F,$E7,$BC,$1F,$E4,$3F,$58,$F9,$5A,$4A
	.dc.b	$FE,$29,$1F,$39,$BC,$5C,$08,$FF,$19,$E5,$FE,$B6,$FF,$1B,$FA,$CF
	.dc.b	$F0,$A3,$7C,$CF,$E5,$F1,$88,$F1,$9F,$E0,$C7,$FA,$03,$C7,$AF,$64
	.dc.b	$7F,$93,$7D,$47,$9D,$ED,$F4,$BD,$08,$F8,$79,$3B,$91,$C1,$FC,$BF
	.dc.b	$EB,$3C,$EF,$88,$60,$F8,$8F,$39,$23,$D9,$E8,$7A,$88,$E0,$F8,$1F
	.dc.b	$07,$EE,$23,$83,$F9,$F7,$D6,$7F,$78,$7F,$A0,$7D,$67,$F5,$AE,$8F
	.dc.b	$EB,$6D,$E5,$FE,$D8,$D9,$FF,$18,$BF,$EC,$97,$8F,$C1,$FF,$5A,$8C
	.dc.b	$47,$AD,$B5,$FD,$9B,$EB,$3D,$93,$E0,$F8,$0E,$04,$7C,$7E,$EF,$32
	.dc.b	$38,$3E,$8F,$27,$7F,$AA,$91,$EA,$7C,$5F,$03,$23,$DA,$D7,$C9,$F0
	.dc.b	$3E,$72,$3F,$09,$FE,$E7,$F5,$8E,$EF,$67,$C9,$CA,$F8,$EF,$BF,$E0
	.dc.b	$A4,$7C,$A6,$BF,$6F,$A9,$D5,$F1,$7E,$5E,$4F,$D0,$7E,$47,$BF,$23
	.dc.b	$E9,$B5,$FC,$27,$D6,$36,$FF,$03,$FA,$FF,$E7,$A9,$FE,$F0,$D7,$F1
	.dc.b	$4D,$E4,$FF,$0A,$BF,$DE,$A3,$FC,$D2,$D7,$FC,$75,$E8,$F4,$3A,$FA
	.dc.b	$12,$3A,$B5,$FF,$43,$FA,$1E,$2F,$A3,$91,$E2,$7C,$BE,$74,$8F,$2B
	.dc.b	$5F,$F4,$7F,$97,$EA,$F3,$64,$7A,$9F,$6F,$9B,$23,$DA,$D7,$FD,$19
	.dc.b	$C3,$F0,$BC,$C9,$1F,$09,$F8,$BE,$DC,$8F,$8A,$D7,$E2,$FA,$DF,$28
	.dc.b	$8F,$CA,$7F,$A1,$1F,$E3,$4F,$AC,$77,$52,$4F,$F8,$9E,$1F,$13,$BF
	.dc.b	$9B,$93,$97,$97,$47,$F8,$FB,$EB,$1F,$07,$49,$5F,$E3,$18,$7C,$1E
	.dc.b	$F8,$24,$E5,$E5,$C1,$FE,$48,$FA,$CF,$E4,$93,$FC,$25,$BD,$3F,$E5
	.dc.b	$46,$F9,$3F,$E6,$57,$FA,$7C,$67,$E4,$B9,$12,$BF,$C8,$52,$F9,$CF
	.dc.b	$F2,$AF,$C4,$3E,$26,$92,$BF,$C9,$11,$FE,$83,$6F,$89,$C0,$8E,$BC
	.dc.b	$4F,$F2,$FF,$D6,$2A,$92,$7F,$CA,$50,$EE,$DF,$6C,$9C,$BC,$B8,$F9
	.dc.b	$F8,$D2,$78,$BF,$E6,$CF,$AC,$FF,$55,$70,$7F,$9D,$97,$CE,$64,$95
	.dc.b	$FE,$61,$87,$CE,$7F,$9D,$3E,$B3,$FD,$85,$3F,$E6,$56,$BF,$CD,$DB
	.dc.b	$7F,$E8,$25,$BF,$F9,$F3,$CF,$FF,$0B,$57,$FB,$61,$A9,$A7,$C3,$7F
	.dc.b	$CE,$4B,$C1,$E4,$48,$CF,$83,$D7,$DF,$91,$C1,$F0,$6B,$FE,$8C,$FA
	.dc.b	$C7,$CD,$D2,$57,$F9,$FE,$3E,$1B,$FE,$81,$FA,$CF,$E7,$C9,$A4,$F5
	.dc.b	$BF,$D2,$EB,$E5,$69,$27,$97,$C8,$EB,$69,$1D,$5E,$3D,$32,$3C,$0D
	.dc.b	$7F,$B1,$5B,$E5,$7B,$F8,$B2,$7C,$CF,$8B,$C5,$91,$E9,$6B,$CB,$D2
	.dc.b	$FF,$56,$FD,$67,$E2,$E3,$FD,$3C,$D7,$FB,$83,$79,$BF,$D7,$6B,$CE
	.dc.b	$E2,$49,$E2,$FF,$91,$F8,$FF,$D7,$3F,$31,$FE,$03,$9F,$EF,$4D,$BF
	.dc.b	$F6,$2B,$7C,$5F,$F6,$B3,$FB,$49,$07,$0D,$F8,$5E,$A7,$F9,$57,$9B
	.dc.b	$FD,$8F,$F5,$5F,$CE,$49,$E9,$E3,$D3,$97,$FE,$BB,$9F,$C7,$93,$C5
	.dc.b	$E2,$48,$D3,$FE,$46,$7A,$F8,$91,$E4,$F8,$36,$91,$D5,$AF,$F8,$66
	.dc.b	$9E,$27,$F9,$1E,$1F,$59,$ED,$7D,$F3,$AD,$FF,$CC,$6D,$8F,$F7,$D1
	.dc.b	$AA,$E4,$E2,$34,$DF,$15,$F5,$9C,$9B,$1C,$84,$FF,$30,$9D,$1F,$E7
	.dc.b	$28,$FA,$3F,$4B,$53,$48,$FF,$9E,$3E,$55,$C7,$FE,$E7,$8F,$F2,$C3
	.dc.b	$63,$31,$FE,$3B,$32,$5F,$8C,$CE,$7A,$8F,$F5,$F7,$B3,$8B,$F8,$35
	.dc.b	$67,$F0,$BE,$82,$B1,$1F,$88,$E5,$AB,$E3,$EC,$61,$98,$79,$FE,$4E
	.dc.b	$D7,$95,$E2,$74,$C1,$F2,$B8,$AF,$CA,$E7,$E7,$9C,$C6,$0F,$CC,$DF
	.dc.b	$FA,$17,$97,$CB,$5F,$CE,$67,$53,$CA,$73,$CB,$27,$33,$FE,$DF,$F0
	.dc.b	$70,$FF,$98,$F0,$47,$C0,$FA,$C7,$05,$E0,$C3,$70,$E7,$3F,$43,$3F
	.dc.b	$47,$7E,$72,$E8,$F3,$F2,$3A,$79,$F7,$0E,$2E,$1E,$7E,$75,$BF,$F3
	.dc.b	$B6,$3F,$D1,$A7,$BF,$C6,$F9,$7D,$5F,$33,$63,$C7,$1E,$9F,$53,$17
	.dc.b	$C9,$F6,$7C,$5F,$B7,$E2,$ED,$FA,$9F,$BF,$F3,$1F,$E0,$E8,$F8,$EB
	.dc.b	$EB,$33,$EE,$E9,$DF,$07,$0F,$2A,$DF,$07,$F1,$3F,$59,$FE,$0E,$8F
	.dc.b	$A2,$BE,$B3,$3E,$F6,$9D,$F4,$F0,$F9,$7B,$7C,$1F,$C9,$FD,$67,$E4
	.dc.b	$63,$ED,$AE,$FC,$F8,$1A,$77,$C5,$C3,$8A,$DF,$07,$F3,$BF,$58,$F4
	.dc.b	$6F,$06,$1B,$5F,$88,$3F,$C5,$65,$DB,$F4,$AE,$39,$7F,$D5,$8F,$BD
	.dc.b	$FF,$57,$7D,$67,$F3,$08,$FC,$8B,$F8,$C6,$C7,$EB,$A3,$0F,$0F,$13
	.dc.b	$1A,$E5,$B5,$CB,$A3,$86,$BC,$5B,$1F,$9E,$F1,$BF,$C0,$0D,$F8,$5F
	.dc.b	$DD,$C6,$31,$FF,$61,$3F,$E5,$09,$E1,$EB,$6B,$C4,$DB,$63,$F5,$3F
	.dc.b	$59,$FE,$2B,$6F,$8D,$FC,$1A,$31,$87,$FB,$35,$FD,$A3,$7D,$CD,$7C
	.dc.b	$3E,$DB,$7D,$3D,$4F,$E1,$FE,$47,$C7,$BE,$51,$87,$3E,$36,$BC,$6B
	.dc.b	$7F,$BC,$BF,$D3,$57,$FB,$77,$EB,$1E,$EF,$F1,$A7,$FC,$99,$5F,$C0
	.dc.b	$9C,$B7,$E1,$3F,$8F,$FD,$67,$E5,$63,$F8,$3A,$F8,$7C,$F8,$7A,$64
	.dc.b	$EB,$C3,$B9,$7F,$BB,$6D,$1D,$0F,$E5,$7F,$4F,$FE,$EF,$AE,$07,$F1
	.dc.b	$47,$15,$C7,$E0,$FC,$7C,$3E,$46,$D7,$92,$39,$F9,$78,$AF,$F7,$E5
	.dc.b	$A3,$F9,$CF,$D6,$7C,$BF,$43,$FA,$13,$F8,$E6,$D6,$B8,$66,$3E,$3F
	.dc.b	$98,$B3,$D1,$E7,$E0,$74,$F1,$ED,$7F,$2C,$C7,$F6,$4D,$8F,$0F,$BD
	.dc.b	$B8,$D3,$87,$E0,$7D,$55,$C7,$D2,$FA,$3E,$73,$D3,$EA,$5A,$EC,$E1
	.dc.b	$F9,$1F,$59,$6F,$E0,$BF,$03,$D4,$F8,$3E,$2D,$AF,$33,$87,$FA,$1F
	.dc.b	$D6,$7F,$6E,$72,$60,$F9,$3C,$5F,$2B,$82,$C7,$BD,$C3,$F0,$74,$B9
	.dc.b	$38,$BE,$7C,$7D,$23,$FB,$5F,$D6,$7E,$C6,$33,$5F,$D5,$23,$0F,$F5
	.dc.b	$EF,$AC,$FC,$AD,$F3,$8C,$76,$7F,$74,$9C,$DF,$F7,$D7,$F3,$B5,$B3
	.dc.b	$FC,$7E,$E2,$3F,$C0,$5C,$DF,$E4,$1B,$FE,$C7,$39,$7F,$B6,$C6,$6B
	.dc.b	$90,$9F,$E0,$CE,$1F,$E8,$CD,$F2,$3C,$9E,$B4,$BF,$ED,$FD,$3E,$CA
	.dc.b	$F1,$CF,$F8,$32,$31,$1D,$7E,$6B,$D9,$F3,$5E,$D6,$DF,$9A,$3E,$07
	.dc.b	$C6,$62,$BC,$4E,$27,$BE,$36,$7D,$38,$7A,$34,$F8,$DE,$5F,$C9,$F8
	.dc.b	$AF,$94,$DC,$3E,$1C,$7F,$88,$3E,$AD,$F5,$99,$C7,$E1,$C6,$2B,$E4
	.dc.b	$FB,$7F,$E2,$6E,$27,$BB,$F5,$63,$0C,$9F,$F0,$8F,$3B,$E7,$71,$8F
	.dc.b	$CE,$8C,$4F,$F8,$5B,$EB,$3F,$95,$ED,$FE,$39,$7F,$95,$B7,$E0,$BF
	.dc.b	$C6,$B1,$89,$FF,$12,$B6,$FF,$CF,$FF,$59,$FE,$A9,$9E,$9E,$9F,$F8
	.dc.b	$F2,$31,$3F,$7A,$BF,$AD,$61,$DF,$D9,$EB,$F1,$1D,$99,$8F,$F8,$AB
	.dc.b	$07,$C8,$E3,$1F,$C3,$C6,$27,$D5,$E5,$F9,$B9,$BC,$ED,$8F,$74,$7F
	.dc.b	$93,$BE,$AD,$F1,$D9,$C3,$F2,$31,$8A,$FF,$1C,$BF,$CD,$63,$FC,$E0
	.dc.b	$D7,$FC,$F6,$D7,$FC,$8E,$D7,$FD,$43,$F5,$9F,$E5,$0E,$3F,$F3,$C4
	.dc.b	$23,$0F,$F9,$ED,$FF,$31,$79,$DA,$73,$1F,$E8,$8F,$AC,$FF,$22,$BC
	.dc.b	$9A,$FF,$A0,$DA,$EE,$7F,$D0,$51,$FE,$B9,$8E,$5F,$13,$82,$3C,$FC
	.dc.b	$DD,$91,$C1,$0F,$DA,$CF,$FA,$7D,$B1,$FE,$66,$E2,$FF,$40,$7D,$63
	.dc.b	$E4,$B1,$AF,$F9,$AA,$1C,$F8,$C3,$FD,$8B,$EE,$3B,$31,$08,$C3,$FE
	.dc.b	$6A,$AE,$2D,$AD,$CE,$A3,$F8,$4B,$71,$F1,$CF,$4A,$E3,$5A,$BE,$26
	.dc.b	$66,$47,$F9,$E2,$F9,$61,$FE,$82,$6C,$7F,$B7,$2F,$9E,$1D,$A3,$86
	.dc.b	$1D,$1F,$31,$0F,$01,$A3,$11,$C5,$C9,$1C,$12,$FC,$53,$D6,$7F,$93
	.dc.b	$D7,$8A,$7E,$76,$19,$BF,$F5,$96,$1E,$8B,$5E,$DF,$55,$6E,$FA,$C7
	.dc.b	$91,$E6,$F9,$CF,$23,$87,$C3,$1C,$5F,$B7,$87,$CA,$72,$7E,$82,$BE
	.dc.b	$8B,$FE,$E2,$6B,$CC,$C4,$FF,$B6,$63,$31,$E6,$39,$46,$4E,$D8,$7C
	.dc.b	$58,$FA,$9D,$38,$C3,$CD,$39,$9F,$83,$39,$7A,$1E,$63,$E5,$B9,$7E
	.dc.b	$3D,$5B,$4D,$8C,$36,$7E,$41,$E2,$29,$C8,$B3,$C9,$B3,$C9,$A6,$A7
	.dc.b	$A3,$E5,$AF,$D6,$C6,$1F,$65,$73,$9D,$5F,$C0,$23,$E3,$3F,$B6,$FD
	.dc.b	$5B,$DB,$FD,$7B,$63,$FD,$AE,$D7,$F1,$BF,$27,$FA,$93,$73,$23,$F6
	.dc.b	$91,$FA,$DF,$92,$E7,$F3,$F2,$7F,$71,$19,$7F,$04,$FD,$D2,$3F,$C9
	.dc.b	$3E,$B3,$FD,$58,$DF,$5F,$F9,$3A,$8C,$BF,$EF,$40,$66,$3F,$11,$DF
	.dc.b	$FC,$2A,$77,$AB,$4F,$F7,$DC,$61,$FE,$29,$3F,$C1,$7D,$C9,$87,$F0
	.dc.b	$F7,$F8,$47,$B7,$23,$F8,$93,$5F,$F0,$BF,$D5,$BE,$8B,$F8,$77,$CC
	.dc.b	$49,$FE,$30,$FF,$11,$F8,$49,$1F,$C6,$DA,$F2,$F0,$BF,$90,$FD,$63
	.dc.b	$B6,$92,$7D,$FB,$FA,$DF,$83,$2F,$8A,$E2,$48,$DE,$FF,$19,$4E,$DC
	.dc.b	$9F,$2B,$47,$89,$FC,$CB,$EA,$FF,$C1,$13,$A0,$E1,$FF,$1B,$B7,$1F
	.dc.b	$24,$6B,$E0,$39,$11,$E5,$28,$FE,$7F,$F4,$FE,$F9,$CF,$3F,$84,$8F
	.dc.b	$0F,$7F,$C8,$47,$07,$97,$EB,$3D,$2F,$6F,$92,$F5,$11,$F0,$F9,$FE
	.dc.b	$D2,$38,$3C,$BE,$FF,$C1,$23,$83,$E1,$7D,$67,$F7,$0D,$7D,$FF,$59
	.dc.b	$F0,$1B,$C5,$FD,$6A,$30,$FD,$17,$FD,$8B,$5F,$2D,$AF,$F9,$81,$B3
	.dc.b	$D3,$E2,$FF,$B1,$FD,$63,$3E,$9F,$47,$30,$E4,$F7,$F7,$24,$75,$6B
	.dc.b	$FE,$6C,$E1,$FA,$D9,$F8,$FE,$AB,$CC,$49,$F2,$7C,$07,$A4,$8E,$0F
	.dc.b	$47,$C0,$7B,$48,$E0,$FC,$37,$D6,$3B,$BE,$17,$C9,$49,$F8,$0F,$C5
	.dc.b	$F8,$19,$1F,$0D,$AF,$9F,$A3,$E5,$7C,$BC,$8F,$90,$FC,$DF,$7E,$47
	.dc.b	$CB,$6B,$FE,$8A,$FA,$C6,$BF,$E0,$CF,$A0,$FF,$36,$CF,$E6,$DA,$FF
	.dc.b	$A4,$1B,$CD,$F9,$A7,$FC,$35,$1B,$F5,$79,$11,$E5,$EE,$75,$23,$83
	.dc.b	$F5,$1F,$05,$FC,$D6,$FA,$7A,$1E,$32,$4F,$8F,$D0,$F3,$11,$C1,$CB
	.dc.b	$CB,$F5,$79,$B2,$3D,$4F,$B7,$CD,$91,$ED,$6B,$FE,$7F,$E1,$F8,$5E
	.dc.b	$64,$8F,$84,$FC,$5F,$22,$47,$C5,$6B,$F1,$7D,$7F,$94,$47,$E5,$3F
	.dc.b	$E3,$8F,$2B,$FC,$71,$1F,$C1,$FE,$B1,$DF,$FE,$3C,$6C,$66,$31,$F0
	.dc.b	$F9,$F9,$A9,$3C,$5E,$1D,$3F,$E4,$4F,$AC,$7A,$74,$95,$FC,$56,$1F
	.dc.b	$09,$BE,$19,$39,$79,$70,$FF,$93,$BE,$B3,$F9,$14,$FF,$B4,$1B,$CD
	.dc.b	$FC,$69,$BE,$27,$F2,$95,$F2,$9C,$7E,$46,$5F,$F2,$59,$47,$5F,$9B
	.dc.b	$FF,$2D,$FB,$FF,$E3,$12,$7C,$06,$3C,$0C,$F8,$B3,$1F,$E5,$D4,$8E
	.dc.b	$3E,$2F,$F9,$93,$EB,$3F,$C7,$14,$69,$3E,$07,$5E,$04,$9F,$E7,$12
	.dc.b	$8E,$5C,$BC,$FC,$89,$5A,$7F,$CE,$3F,$59,$FD,$B7,$8F,$F9,$DA,$F9
	.dc.b	$BC,$92,$BF,$A2,$47,$F8,$1E,$7F,$A0,$FD,$67,$FA,$DA,$7F,$A4,$35
	.dc.b	$FF,$7E,$B6,$37,$AF,$FA,$03,$D0,$FF,$59,$D7,$F4,$F6,$A9,$FE,$9B
	.dc.b	$18,$6E,$FC,$7C,$39,$5C,$5E,$9E,$1F,$5F,$7E,$4F,$57,$C1,$AF,$FA
	.dc.b	$43,$EB,$1F,$33,$49,$5F,$D8,$63,$E1,$3F,$D8,$FE,$B3,$F9,$EB,$FE
	.dc.b	$12,$6A,$69,$FC,$F3,$B3,$FB,$92,$F9,$4F,$F4,$FB,$67,$A4,$63,$E4
	.dc.b	$33,$C1,$27,$C0,$F0,$E8,$91,$E3,$6B,$F1,$F0,$EF,$F2,$9D,$BC,$59
	.dc.b	$3E,$97,$C3,$E2,$C8,$F6,$35,$E4,$7B,$7D,$3F,$F5,$AF,$D6,$7F,$1B
	.dc.b	$C3,$FC,$18,$D9,$FF,$19,$B5,$2F,$3F,$89,$23,$FC,$19,$19,$7F,$C1
	.dc.b	$DE,$63,$8F,$FD,$87,$F1,$9F,$E0,$59,$F1,$87,$1D,$1F,$E1,$06,$E3
	.dc.b	$FE,$1F,$7F,$73,$3F,$E4,$66,$BC,$7E,$1B,$A9,$FC,$67,$9B,$FD,$A3
	.dc.b	$F5,$7F,$D0,$27,$BB,$70,$FF,$0E,$B7,$0F,$F1,$63,$F9,$36,$FA,$9C
	.dc.b	$89,$5F,$E2,$99,$FD,$C9,$3A,$FB,$9F,$E7,$28,$F0,$F8,$71,$E0,$F9
	.dc.b	$37,$91,$CD,$AF,$9B,$C2,$FD,$32,$FA,$CF,$F7,$B6,$1F,$C0,$73,$FF
	.dc.b	$34,$4F,$F9,$A2,$30,$E1,$AF,$F9,$B5,$AF,$BB,$C3,$7D,$67,$28,$4F
	.dc.b	$F3,$48,$C3,$63,$A4,$F0,$7F,$CF,$7F,$49,$FE,$7C,$4D,$23,$D7,$E4
	.dc.b	$BF,$CB,$D8,$7F,$93,$23,$FC,$C9,$BF,$F9,$4D,$4E,$A3,$FD,$09,$F1
	.dc.b	$96,$7F,$CC,$0D,$5E,$7B,$CA,$C4,$74,$3F,$1B,$19,$AE,$DC,$9A,$F3
	.dc.b	$F6,$D8,$F1,$47,$2F,$8E,$C5,$FE,$58,$39,$9F,$F3,$55,$BE,$B3,$15
	.dc.b	$E0,$7C,$DA,$39,$6F,$2E,$1B,$87,$F9,$CE,$7F,$DA,$6F,$4F,$03,$38
	.dc.b	$7A,$9F,$81,$F5,$9F,$1A,$7F,$CF,$4F,$BE,$BF,$CE,$EE,$6B,$F5,$AE
	.dc.b	$67,$FD,$10,$D8,$FF,$57,$E4,$EF,$EB,$C1,$F3,$63,$17,$DB,$C7,$F0
	.dc.b	$F9,$BE,$26,$C7,$8A,$3C,$BF,$23,$8C,$7C,$2F,$A3,$E3,$7D,$3E,$9D
	.dc.b	$BF,$4B,$87,$D7,$E9,$DB,$F4,$BF,$03,$E1,$BE,$0F,$C6,$5A,$F8,$2F
	.dc.b	$E0,$FE,$59,$ED,$DF,$AC,$C3,$6B,$FD,$41,$3E,$CC,$7F,$88,$AF,$2B
	.dc.b	$97,$17,$37,$F1,$DF,$59,$FE,$12,$BF,$F5,$42,$FA,$6C,$FB,$BA,$77
	.dc.b	$C9,$C3,$EB,$2D,$F0,$7F,$31,$F5,$9F,$E1,$08,$FF,$59,$3F,$8C,$6F
	.dc.b	$7E,$BB,$F7,$C7,$C3,$8E,$D7,$07,$F4,$5F,$58,$F4,$6F,$06,$1B,$5F
	.dc.b	$EB,$F3,$FC,$5E,$5D,$BF,$54,$E3,$8B,$F9,$9B,$EF,$7F,$9A,$FD,$67
	.dc.b	$FB,$D2,$3F,$D9,$0F,$E3,$7B,$BF,$D8,$AE,$67,$7E,$2F,$ED,$63,$FD
	.dc.b	$B5,$5C,$36,$C7,$89,$97,$5D,$5B,$DE,$D7,$67,$F7,$DE,$33,$DB,$BF
	.dc.b	$07,$86,$D7,$FB,$6E,$7F,$C5,$CD,$F1,$38,$FA,$DC,$9F,$E1,$1F,$59
	.dc.b	$FB,$39,$FF,$71,$3F,$B0,$DF,$FA,$9B,$96,$F9,$1C,$3D,$CE,$3F,$0D
	.dc.b	$6F,$8B,$D3,$D4,$FE,$29,$C7,$F0,$F9,$3F,$DD,$39,$7F,$AA,$6F,$FD
	.dc.b	$E1,$39,$7F,$DF,$4F,$F8,$CD,$CF,$8D,$CB,$C7,$E6,$FF,$1F,$FA,$C7
	.dc.b	$C9,$DF,$84,$C3,$6F,$FD,$F9,$F5,$9F,$95,$6F,$C3,$FF,$27,$8C,$70
	.dc.b	$FF,$70,$51,$93,$5E,$25,$AE,$6E,$15,$7F,$78,$FA,$7F,$F6,$D4,$68
	.dc.b	$CE,$E7,$72,$9D,$BE,$1F,$84,$F2,$71,$F9,$5B,$5E,$A8,$FE,$7D,$4F
	.dc.b	$37,$37,$E1,$7F,$A1,$7D,$67,$CB,$8F,$2B,$F9,$5B,$F0,$B9,$83,$97
	.dc.b	$7C,$5D,$BE,$67,$9B,$6B,$D6,$72,$EF,$EF,$E3,$F8,$1B,$1A,$BF,$0C
	.dc.b	$F9,$3D,$EC,$C7,$F8,$7B,$1F,$F0,$F5,$6E,$3C,$73,$EE,$FA,$AE,$13
	.dc.b	$9D,$11,$EE,$F9,$CF,$7F,$A9,$6B,$96,$3F,$E5,$E6,$C7,$C8,$FA,$CD
	.dc.b	$C7,$37,$E3,$FA,$BF,$23,$C6,$B3,$C9,$C3,$E5,$FA,$CF,$B4,$E4,$C1
	.dc.b	$F4,$38,$BE,$8F,$15,$8F,$1B,$87,$CB,$A5,$DF,$E1,$3F,$60,$FE,$F5
	.dc.b	$F5,$9F,$BD,$9D,$9F,$15,$C4,$FF,$91,$BE,$B3,$F4,$33,$F2,$57,$67
	.dc.b	$F8,$0D,$CC,$EE,$30,$7E,$2B,$8A,$D7,$CD,$78,$79,$1C,$7D,$7B,$3D
	.dc.b	$5C,$3F,$5F,$A3,$C7,$D5,$F9,$DC,$7F,$47,$97,$F5,$0A,$D7,$93,$E3
	.dc.b	$F9,$BF,$27,$CD,$D9,$F4,$38,$7D,$5F,$19,$70,$F7,$3E,$B3,$FC,$3D
	.dc.b	$97,$FB,$C2,$76,$78,$F4,$F8,$1E,$57,$C1,$F8,$BB,$3F,$15,$C3,$F8
	.dc.b	$FF,$97,$7D,$66,$32,$FF,$38,$C6,$27,$E6,$7B,$9F,$E2,$DE,$37,$BB
	.dc.b	$F3,$63,$0C,$9F,$CD,$74,$3E,$B3,$18,$FF,$9F,$23,$13,$F9,$FF,$AC
	.dc.b	$FE,$59,$CB,$F2,$69,$1F,$CC,$63,$0E,$A7,$F8,$4B,$96,$FA,$3F,$E4
	.dc.b	$1F,$AC,$FF,$58,$47,$58,$C7,$1F,$F9,$16,$BF,$AE,$5F,$4F,$63,$1F
	.dc.b	$0F,$96,$61,$FB,$3C,$5F,$23,$8C,$FF,$D2,$B1,$89,$F5,$79,$7D,$79
	.dc.b	$FB,$36,$3D,$D1,$FE,$55,$FA,$C7,$C7,$E7,$0F,$F4,$FC,$62,$BF,$81
	.dc.b	$7D,$67,$F9,$27,$8F,$FC,$E4,$DA,$FF,$3E,$BF,$E5,$DF,$3B,$3F,$15
	.dc.b	$F5,$27,$FD,$03,$83,$D9,$FF,$3D,$42,$30,$FF,$9F,$5A,$F0,$8D,$FD
	.dc.b	$67,$FA,$CE,$3F,$9B,$87,$0F,$F3,$A6,$BF,$D0,$9A,$B0,$FF,$09,$3E
	.dc.b	$AF,$80,$F1,$C6,$23,$D7,$1E,$6F,$AC,$FF,$5A,$47,$81,$AF,$F4,$86
	.dc.b	$BF,$D4,$1A,$9F,$F5,$C3,$9F,$A0,$FE,$9F,$CB,$F8,$D9,$CC,$EF,$7C
	.dc.b	$EF,$D7,$A2,$CF,$5F,$8D,$AF,$56,$DF,$FB,$84,$8E,$0C,$CC,$9F,$E7
	.dc.b	$93,$FE,$7F,$6C,$E5,$FF,$6F,$C7,$27,$27,$09,$D6,$1E,$9A,$ED,$39
	.dc.b	$21,$C9,$CB,$D2,$78,$DF,$C9,$9C,$D7,$C3,$9C,$D7,$BC,$F0,$9C,$4E
	.dc.b	$B1,$E1,$72,$F1,$73,$5D,$5C,$BF,$2D,$C9,$D4,$F3,$18,$3E,$C4,$F9
	.dc.b	$23,$94,$78,$48,$7E,$0E,$3F,$0C,$47,$FD,$AD,$68,$C4,$7F,$B8,$A3
	.dc.b	$0E,$A3,$35,$E9,$8E,$CF,$F1,$1F,$95,$7C,$4F,$C4,$9F,$29,$E7,$8C
	.dc.b	$BE,$9B,$FC,$15,$FE,$1A,$73,$3D,$30,$F8,$4D,$3F,$C4,$F1,$AC,$3C
	.dc.b	$44,$9D,$BE,$3C,$3C,$58,$76,$BF,$F0,$35,$F1,$BD,$92,$3F,$49,$87
	.dc.b	$8E,$F8,$BE,$BA,$E1,$3A,$3D,$51,$B6,$9E,$F9,$E2,$D3,$27,$AC,$4F
	.dc.b	$47,$F6,$5F,$29,$FA,$8F,$AC,$7B,$F4,$FD,$4C,$49,$FB,$2F,$F0,$12
	.dc.b	$3F,$73,$E5,$7F,$82,$11,$C1,$0F,$D4,$BF,$B4,$3C,$1E,$A7,$ED,$BB
	.dc.b	$BF,$73,$F5,$9F,$EF,$6A,$FF,$17,$35,$FD,$8C,$6D,$FC,$5B,$FC,$4C
	.dc.b	$8F,$E3,$BD,$EF,$E0,$24,$7F,$7B,$EF,$7F,$04,$23,$83,$EB,$FD,$17
	.dc.b	$F0,$8F,$98,$91,$FC,$11,$FE,$17,$F0,$D2,$3F,$83,$B5,$72,$7E,$57
	.dc.b	$C1,$FE,$27,$F5,$8F,$94,$A4,$AF,$F2,$54,$3E,$33,$64,$8C,$BF,$97
	.dc.b	$F1,$3F,$8E,$FD,$67,$F7,$79,$FF,$29,$35,$F0,$F4,$FF,$9E,$23,$19
	.dc.b	$7F,$77,$5F,$05,$C8,$95,$C9,$FE,$53,$F5,$1E,$78,$EB,$F4,$BE,$02
	.dc.b	$3E,$0E,$5F,$19,$1C,$1F,$E6,$5F,$AC,$67,$E8,$F9,$49,$3E,$87,$D7
	.dc.b	$E8,$48,$F5,$B5,$F5,$9F,$E9,$8C,$DF,$03,$F0,$89,$1F,$09,$FE,$83
	.dc.b	$F5,$8F,$33,$FA,$1F,$D6,$7F,$85,$F8,$3F,$C3,$0D,$E1,$FF,$4E,$B7
	.dc.b	$0F,$F5,$2B,$FE,$B6,$7F,$CF,$CD,$7C,$61,$C4,$F4,$F8,$DF,$EA,$3F
	.dc.b	$58,$D7,$C3,$F0,$13,$0E,$6F,$93,$DE,$91,$D9,$AF,$6E,$4F,$B3,$9F
	.dc.b	$9B,$EA,$BD,$64,$9F,$2F,$C0,$FB,$88,$E0,$F4,$FC,$0F,$C2,$23,$83
	.dc.b	$FD,$35,$F5,$8E,$EF,$8F,$F2,$92,$7E,$0B,$F2,$BE,$0E,$47,$C4,$6B
	.dc.b	$E7,$EB,$F9,$FF,$31,$23,$E4,$BF,$4B,$E0,$24,$7C,$C6,$BF,$33,$EB
	.dc.b	$1A,$FF,$7A,$FA,$0F,$F2,$FC,$FF,$AB,$DA,$FD,$26,$F3,$7F,$94,$DF
	.dc.b	$F0,$39,$3D,$5E,$47,$3F,$56,$57,$36,$BF,$E6,$AF,$4F,$C1,$E7,$C8
	.dc.b	$F0,$3E,$3F,$3E,$47,$8D,$AF,$F9,$CB,$CB,$F4,$79,$B2,$3D,$0F,$AF
	.dc.b	$CD,$91,$EB,$6B,$FE,$6F,$E1,$F8,$1E,$64,$8F,$80,$FC,$3F,$1A,$47
	.dc.b	$C3,$6B,$F0,$FD,$9F,$90,$47,$E4,$3F,$D7,$CF,$F1,$2F,$D6,$3B,$A9
	.dc.b	$27,$FD,$B9,$0F,$89,$DF,$B3,$27,$2F,$2E,$8F,$F1,$7F,$D6,$3D,$3A
	.dc.b	$4A,$FF,$73,$43,$E0,$B7,$C1,$27,$2F,$2E,$0F,$F1,$CF,$D6,$7F,$23
	.dc.b	$9F,$F7,$73,$5F,$D8,$B7,$A7,$FD,$24,$BE,$5B,$8F,$C8,$CA,$E2,$F4
	.dc.b	$F9,$CF,$F2,$2F,$BF,$FE,$D0,$9F,$F7,$D3,$5F,$DE,$B7,$0F,$F4,$F2
	.dc.b	$ED,$E3,$E1,$4A,$E2,$F4,$F0,$BF,$CA,$3F,$59,$FE,$49,$9F,$E0,$ED
	.dc.b	$C3,$FD,$50,$DF,$21,$32,$3F,$C4,$2F,$0E,$DE,$64,$77,$5C,$B8,$FB
	.dc.b	$F8,$B2,$73,$87,$45,$87,$EF,$5C,$9E,$0F,$AC,$FF,$43,$E1,$FE,$BB
	.dc.b	$5F,$39,$FE,$6A,$6C,$F0,$0E,$1F,$E4,$6F,$F8,$21,$D7,$D6,$7F,$AC
	.dc.b	$9B,$F1,$5F,$E7,$48,$C4,$7F,$2A,$5B,$7F,$9E,$3C,$EF,$F5,$7C,$F7
	.dc.b	$6B,$FC,$C1,$B8,$FF,$B4,$5F,$95,$1F,$29,$78,$3C,$C9,$1C,$FC,$1E
	.dc.b	$DE,$F4,$8E,$CF,$87,$7F,$F4,$4F,$D6,$3E,$66,$92,$BE,$14,$7C,$1C
	.dc.b	$BF,$DC,$5F,$59,$FD,$26,$4D,$27,$AB,$FE,$94,$5F,$1F,$49,$3F,$36
	.dc.b	$3E,$2D,$74,$DB,$81,$27,$5C,$FC,$C8,$E0,$E5,$F2,$9D,$FC,$49,$1D
	.dc.b	$5F,$17,$89,$23,$C0,$D7,$FA,$EE,$2F,$47,$FD,$57,$F5,$9F,$94,$CF
	.dc.b	$FE,$00,$37,$AB,$F1,$CD,$BF,$CD,$AF,$37,$89,$23,$F1,$51,$97,$FC
	.dc.b	$CB,$C7,$FE,$B8,$F9,$7F,$F0,$84,$7F,$6E,$6E,$0C,$3F,$39,$18,$6F
	.dc.b	$C1,$F1,$25,$7E,$5E,$7F,$83,$4F,$E8,$FE,$27,$F4,$53,$FD,$EB,$EA
	.dc.b	$BF,$D9,$A4,$F5,$3F,$D9,$2F,$E3,$AB,$F5,$4D,$9F,$DB,$B5,$34,$FC
	.dc.b	$9B,$8F,$13,$99,$1F,$D5,$7C,$57,$EC,$2B,$AF,$A9,$E0,$24,$FE,$BE
	.dc.b	$FE,$33,$E7,$DF,$CA,$47,$5E,$2F,$F9,$41,$7D,$63,$5F,$4C,$E6,$7F
	.dc.b	$82,$1E,$87,$0E,$B7,$CB,$EE,$8C,$BF,$03,$C6,$7D,$66,$A3,$E0,$C6
	.dc.b	$1E,$39,$7F,$43,$9D,$36,$38,$0F,$A5,$AD,$32,$E0,$DF,$C1,$7E,$06
	.dc.b	$3F,$C6,$27,$F9,$AF,$6F,$E2,$8F,$F8,$FD,$BF,$17,$BC,$F8,$4F,$F2
	.dc.b	$E8,$C4,$FF,$25,$F4,$3F,$9E,$3C,$0F,$E3,$CE,$27,$BF,$2F,$07,$F9
	.dc.b	$67,$9F,$5F,$E7,$31,$5F,$EA,$55,$E4,$9C,$7D,$16,$5B,$1E,$88,$E5
	.dc.b	$6F,$94,$CD,$FF,$93,$A3,$15,$FC,$C7,$27,$37,$85,$FF,$75,$46,$63
	.dc.b	$1C,$B2,$3D,$7E,$05,$C3,$A3,$E9,$FA,$CF,$91,$1F,$E5,$C7,$E6,$37
	.dc.b	$7D,$6F,$2F,$E2,$F7,$DA,$F3,$3C,$FC,$6E,$9E,$75,$AF,$E9,$18,$FF
	.dc.b	$4C,$6C,$78,$7E,$43,$71,$A7,$C9,$F1,$BE,$5F,$52,$CF,$27,$0F,$4F
	.dc.b	$52,$E7,$E9,$7D,$9F,$0D,$ED,$F8,$CB,$5E,$27,$EC,$7C,$B7,$F8,$76
	.dc.b	$3F,$AE,$2F,$AC,$CF,$BB,$A7,$7C,$1C,$3C,$9B,$7C,$1F,$C1,$7D,$67
	.dc.b	$F8,$76,$3F,$B3,$2F,$AC,$CF,$BD,$A7,$7D,$7C,$3E,$56,$DF,$07,$F1
	.dc.b	$9F,$59,$F8,$AE,$1F,$F4,$7A,$EF,$CF,$81,$A7,$7C,$7C,$38,$EE,$3F
	.dc.b	$DD,$EF,$FD,$24,$FF,$79,$FA,$C7,$A3,$F9,$95,$C3,$63,$0F,$E7,$67
	.dc.b	$FC,$78,$7F,$80,$4F,$76,$7F,$EA,$27,$2D,$DF,$F4,$9F,$59,$FE,$1C
	.dc.b	$9F,$F0,$43,$F8,$EF,$0B,$F6,$6E,$6B,$FD,$56,$E6,$7F,$C3,$0F,$FB
	.dc.b	$8D,$E3,$B7,$4F,$13,$AB,$DB,$A7,$BE,$F6,$BB,$BF,$B5,$F1,$5E,$DD
	.dc.b	$F8,$5C,$36,$BF,$C3,$93,$FE,$12,$6F,$89,$C7,$D6,$E4,$FE,$FF,$EB
	.dc.b	$3F,$63,$3F,$E2,$37,$F6,$3B,$7F,$B1,$1C,$B7,$C7,$E1,$EE,$71,$F8
	.dc.b	$7B,$7C,$5E,$9E,$EF,$F0,$BF,$2B,$E2,$72,$7F,$18,$CB,$FD,$69,$7F
	.dc.b	$E3,$19,$CB,$FE,$3D,$7F,$59,$97,$FB,$51,$78,$DC,$FC,$7E,$8F,$F1
	.dc.b	$7F,$AC,$7B,$B7,$E8,$B0,$DC,$7F,$C7,$DF,$59,$F9,$96,$FC,$47,$F1
	.dc.b	$F8,$C6,$7F,$EE,$25,$19,$35,$E2,$5A,$E6,$E1,$FF,$25,$7D,$67,$E5
	.dc.b	$2F,$94,$FF,$2B,$8C,$6F,$FC,$E3,$0F,$F1,$DD,$F9,$38,$BC,$5F,$07
	.dc.b	$E3,$64,$74,$F3,$6E,$5F,$EF,$6C,$51,$FC,$DF,$EB,$3E,$4C,$FF,$9C
	.dc.b	$1F,$C8,$34,$D7,$14,$BF,$3F,$CC,$D9,$E8,$F3,$EF,$E9,$E4,$5A,$FF
	.dc.b	$33,$61,$FB,$76,$C7,$87,$DA,$DC,$69,$C3,$F1,$3E,$AA,$E3,$E9,$7D
	.dc.b	$1F,$3B,$E9,$F5,$6D,$76,$70,$FC,$CF,$AC,$B7,$F0,$5F,$81,$EA,$FC
	.dc.b	$1F,$1A,$D7,$99,$C3,$FE,$7D,$FA,$CF,$F1,$73,$93,$07,$C9,$E1,$F9
	.dc.b	$5C,$16,$3D,$EE,$1F,$83,$A5,$C9,$CD,$FD,$CB,$81,$3F,$4C,$FE,$D5
	.dc.b	$F5,$9F,$B6,$BE,$27,$FA,$3D,$C5,$F2,$8F,$F4,$E7,$D6,$7E,$69,$FF
	.dc.b	$4A,$2E,$CF,$EE,$8E,$67,$81,$FE,$AE,$75,$3C,$27,$F2,$1C,$5C,$1B
	.dc.b	$1D,$3C,$8E,$7E,$BF,$46,$C7,$82,$3F,$C0,$5D,$3E,$6D,$7F,$C1,$F8
	.dc.b	$FF,$4D,$CB,$C2,$27,$57,$EA,$F9,$BF,$3F,$CD,$FA,$1B,$1E,$C8,$F5
	.dc.b	$FC,$66,$2F,$9E,$1F,$E8,$AA,$F8,$33,$E2,$CF,$71,$FE,$B7,$C7,$FD
	.dc.b	$FC,$FC,$6F,$27,$E2,$FC,$67,$C6,$6C,$7C,$A8,$FF,$0E,$7D,$5B,$EB
	.dc.b	$33,$9F,$AA,$31,$5F,$43,$DC,$FF,$11,$70,$BD,$EF,$A3,$18,$64,$FF
	.dc.b	$B2,$7A,$1F,$59,$8C,$7E,$2C,$62,$7F,$D9,$DF,$59,$FD,$3B,$93,$FC
	.dc.b	$72,$D8,$FF,$6E,$3F,$D4,$23,$3E,$BF,$F8,$CF,$EB,$1F,$29,$4F,$F1
	.dc.b	$B5,$71,$C7,$FD,$DF,$3A,$C3,$FD,$B8,$BE,$83,$91,$CF,$DC,$B3,$E4
	.dc.b	$F8,$1E,$66,$2F,$EE,$E1,$FC,$93,$3F,$F2,$7F,$D6,$3E,$BF,$FC,$A3
	.dc.b	$5F,$EF,$38,$C6,$1F,$E6,$96,$C7,$F9,$C1,$AF,$F9,$DB,$EB,$3F,$C8
	.dc.b	$9C,$3F,$E6,$F6,$C7,$F9,$D9,$7C,$DB,$CD,$D7,$C5,$FD,$3A,$FF,$3D
	.dc.b	$FD,$67,$F3,$67,$FD,$F5,$8D,$7F,$2A,$DD,$3F,$CC,$F3,$73,$54,$8C
	.dc.b	$38,$AA,$9B,$13,$D4,$E3,$3D,$9B,$7F,$19,$B7,$AA,$E1,$5E,$69,$CB
	.dc.b	$DD,$AA,$36,$F5,$38,$1E,$8A,$F5,$CF,$09,$F3,$DC,$8C,$12,$B8,$9C
	.dc.b	$CF,$DC,$CF,$6A,$F4,$46,$9E,$AF,$7B,$F5,$57,$9A,$FB,$4E,$4F,$95
	.dc.b	$1D,$DE,$D5,$D1,$EF,$19,$8C,$12,$BA,$CF,$33,$E1,$4F,$72,$39,$BF
	.dc.b	$44,$FE,$46,$73,$3D,$8D,$38,$AE,$D1,$E6,$32,$7B,$E7,$93,$83,$91
	.dc.b	$EA,$72,$E2,$3E,$2B,$93,$ED,$1C,$5C,$9E,$07,$F1,$53,$F0,$67,$DF
	.dc.b	$1F,$9A,$8F,$F6,$4D,$7E,$66,$3F,$DB,$6D,$76,$62,$36,$47,$FD,$B1
	.dc.b	$1F,$11,$F7,$1F,$7B,$D4,$F0,$38,$BF,$2D,$FE,$56,$73,$3F,$40,$F5
	.dc.b	$46,$5A,$71,$7D,$33,$FA,$C9,$E3,$5E,$1A,$FC,$D4,$62,$F2,$D8,$E6
	.dc.b	$68,$B3,$C4,$46,$E3,$E8,$5F,$8D,$A7,$8C,$87,$33,$27,$AC,$B3,$D1
	.dc.b	$EF,$5E,$46,$9E,$2A,$F0,$5F,$C6,$8F,$E0,$D1,$8B,$F8,$27,$8E,$75
	.dc.b	$3D,$DB,$7D,$E3,$C3,$7D,$DA,$77,$3F,$41,$F5,$8F,$1B,$F8,$6C,$48
	.dc.b	$C3,$83,$F8,$CB,$87,$0D,$7F,$13,$F1,$1F,$C3,$23,$58,$31,$FE,$33
	.dc.b	$18,$7F,$8B,$4F,$F1,$2F,$AC,$7C,$25,$F8,$D9,$8F,$ED,$67,$2D,$9E
	.dc.b	$04,$7F,$75,$F2,$FF,$E0,$57,$F7,$9F,$2C,$F4,$BF,$8F,$B6,$33,$1E
	.dc.b	$1F,$62,$46,$08,$7F,$84,$30,$47,$F2,$8F,$AB,$FE,$E3,$3F,$C2,$5A
	.dc.b	$FF,$85,$9B,$B7,$C1,$8F,$E8,$30,$F6,$39,$12,$39,$3F,$CD,$3E,$A1
	.dc.b	$AF,$7F,$A4,$95,$DD,$F1,$7B,$12,$3C,$4D,$7F,$C4,$FF,$59,$F0,$DF
	.dc.b	$27,$C8,$7A,$48,$FA,$7C,$7F,$61,$1C,$1E,$DF,$67,$E0,$11,$C1,$FC
	.dc.b	$A7,$EB,$3F,$B4,$3F,$CA,$FE,$B3,$E0,$3F,$E3,$86,$EF,$86,$F3,$FF
	.dc.b	$5C,$7F,$D5,$73,$C7,$E0,$FF,$AA,$C6,$23,$94,$7F,$58,$FA,$C6,$7A
	.dc.b	$F8,$E9,$5E,$17,$8F,$AD,$23,$C8,$D7,$FC,$97,$C5,$F5,$B2,$F1,$7D
	.dc.b	$5F,$80,$93,$EB,$F6,$FC,$64,$70,$7B,$FD,$BF,$39,$1C,$1F,$D3,$7C
	.dc.b	$E7,$C5,$FB,$3E,$3A,$4F,$C7,$7D,$FE,$D4,$8F,$94,$D7,$CF,$D1,$F1
	.dc.b	$3E,$4A,$47,$CF,$7E,$3F,$AF,$23,$E9,$35,$FF,$34,$FD,$63,$5F,$F0
	.dc.b	$0F,$9D,$FE,$55,$9F,$EC,$4D,$7F,$CD,$ED,$E7,$FB,$6B,$E8,$38,$92
	.dc.b	$78,$BD,$3E,$B2,$47,$46,$BF,$E6,$1F,$92,$F0,$FD,$0C,$8F,$0B,$E4
	.dc.b	$FA,$09,$1E,$46,$BF,$E6,$9F,$AC,$F4,$F9,$F2,$3D,$2F,$B3,$D1,$91
	.dc.b	$EC,$6B,$FE,$66,$E1,$F8,$3F,$1F,$23,$E0,$BF,$13,$E1,$24,$7C,$46
	.dc.b	$BF,$13,$E5,$7E,$49,$1F,$92,$FF,$8B,$3C,$F3,$FE,$2E,$FA,$C7,$7D
	.dc.b	$25,$7F,$84,$61,$F1,$3B,$F7,$E4,$E5,$E5,$D3,$FE,$38,$FA,$C7,$D1
	.dc.b	$D2,$57,$F8,$6E,$1F,$09,$BE,$19,$39,$79,$70,$FF,$90,$BE,$B3,$F9
	.dc.b	$6C,$FF,$AB,$9B,$D5,$FB,$A6,$F8,$9F,$E5,$27,$F9,$1C,$FF,$8A,$5F
	.dc.b	$E5,$8F,$1F,$57,$A1,$D7,$E6,$FF,$C9,$FF,$45,$FE,$CD,$9F,$F1,$83
	.dc.b	$5F,$F5,$EB,$70,$FE,$12,$FF,$9D,$23,$7E,$1F,$22,$3C,$BC,$3F,$F2
	.dc.b	$DF,$D6,$2A,$92,$7F,$C7,$D0,$EE,$DF,$6C,$9C,$BC,$B8,$B9,$F8,$B2
	.dc.b	$78,$BF,$E6,$8F,$AC,$FE,$CF,$A7,$F1,$E5,$F3,$99,$25,$7F,$93,$61
	.dc.b	$F3,$9F,$E7,$0F,$AC,$FF,$56,$CF,$F9,$45,$AF,$FB,$79,$B7,$FC,$A9
	.dc.b	$6F,$FE,$78,$F3,$9E,$2D,$25,$FF,$96,$E3,$E1,$37,$C2,$E0,$47,$5E
	.dc.b	$17,$3F,$A9,$91,$C5,$ED,$B7,$FA,$1B,$EB,$1F,$37,$49,$5F,$E6,$98
	.dc.b	$F8,$2F,$F9,$AF,$EB,$3F,$9E,$BF,$EA,$06,$A6,$9F,$CF,$76,$FE,$8E
	.dc.b	$BE,$5B,$FD,$2A,$D9,$E9,$18,$F9,$1C,$EF,$27,$C0,$F0,$E9,$91,$E3
	.dc.b	$6A,$C3,$BF,$CB,$76,$F1,$24,$FA,$5F,$0F,$89,$23,$D8,$D7,$97,$81
	.dc.b	$FE,$B4,$D1,$E8,$FF,$AB,$3E,$B3,$F1,$D5,$FE,$8D,$6B,$C0,$FF,$61
	.dc.b	$B8,$CB,$FB,$3A,$F3,$B9,$12,$79,$3F,$C3,$78,$FF,$D7,$3F,$13,$FE
	.dc.b	$06,$9F,$F4,$D3,$5E,$C7,$F8,$F7,$E0,$77,$F0,$F2,$36,$D7,$7E,$7F
	.dc.b	$F5,$67,$FD,$49,$F5,$5F,$CF,$DB,$E9,$FF,$B2,$23,$11,$D9,$FC,$9B
	.dc.b	$C3,$B7,$44,$79,$F8,$DC,$48,$E0,$FF,$1B,$9D,$7C,$58,$F8,$9F,$06
	.dc.b	$B2,$3C,$AD,$7C,$5E,$2F,$DD,$5F,$59,$FE,$57,$C5,$39,$9F,$F2,$FC
	.dc.b	$63,$87,$FC,$BD,$08,$C3,$FE,$61,$6B,$C3,$C6,$7D,$67,$09,$FE,$66
	.dc.b	$D8,$CC,$F1,$6B,$DB,$E9,$BB,$A0,$E1,$A6,$9D,$FC,$8F,$F9,$2F,$0F
	.dc.b	$F1,$6C,$FF,$91,$BA,$3F,$DB,$4A,$75,$1F,$E7,$CE,$3B,$3D,$CE,$07
	.dc.b	$E1,$BC,$E7,$B1,$A8,$F2,$4D,$9C,$D7,$6E,$5D,$57,$F9,$92,$30,$78
	.dc.b	$BE,$2B,$26,$BE,$8A,$C7,$99,$C3,$FE,$F1,$B7,$C9,$62,$FC,$8F,$8F
	.dc.b	$4F,$E7,$95,$FE,$F7,$7F,$9D,$37,$97,$5C,$B2,$3A,$78,$37,$0F,$73
	.dc.b	$ED,$FA,$CF,$8D,$3F,$F0,$05,$F8,$78,$7F,$96,$1C,$D7,$EA,$9C,$CF
	.dc.b	$C9,$6C,$7D,$9C,$9E,$07,$5D,$FE,$43,$DF,$C7,$F0,$79,$F7,$1F,$03
	.dc.b	$87,$C5,$E7,$5C,$7C,$0F,$9F,$E2,$FD,$1E,$9D,$AF,$43,$87,$D5,$E9
	.dc.b	$DB,$F4,$3E,$FF,$84,$F8,$1F,$15,$6B,$E0,$3F,$7F,$E5,$7F,$C2,$31
	.dc.b	$F9,$3E,$1F,$F3,$CC,$3E,$B3,$5E,$E6,$DE,$1C,$1C,$7C,$AB,$87,$13
	.dc.b	$F1,$D7,$FC,$91,$C6,$9F,$9F,$FA,$CF,$F0,$AC,$62,$10,$F9,$AF,$C9
	.dc.b	$3A,$8F,$E3,$6E,$67,$83,$FE,$F8,$6F,$CA,$70,$E6,$E2,$FE,$6F,$EB
	.dc.b	$3F,$2D,$7F,$C9,$17,$6E,$7B,$B4,$EF,$8F,$87,$1D,$CB,$CD,$7F,$E9
	.dc.b	$07,$D1,$F5,$8F,$47,$F5,$0F,$E0,$A3,$31,$8E,$1F,$D6,$47,$F8,$E0
	.dc.b	$F8,$07,$B7,$93,$FD,$4E,$E7,$4F,$E7,$2E,$4F,$DC,$7D,$67,$F0,$77
	.dc.b	$E4,$3F,$96,$9F,$F5,$7B,$99,$FE,$82,$E6,$78,$F8,$5F,$BD,$8F,$87
	.dc.b	$5C,$23,$1B,$67,$A3,$5B,$5C,$3B,$BF,$C1,$7C,$57,$B7,$7E,$13,$0D
	.dc.b	$AF,$B9,$3F,$E1,$86,$F8,$5C,$7D,$6E,$4F,$F0,$FF,$AC,$FD,$EE,$FF
	.dc.b	$D6,$1F,$DE,$B7,$C6,$D7,$8D,$BF,$73,$87,$C3,$5C,$BF,$15,$87,$F9
	.dc.b	$C1,$BE,$EF,$F1,$BF,$33,$E2,$72,$7F,$B5,$2B,$FD,$67,$5F,$91,$9C
	.dc.b	$BF,$C9,$5F,$F1,$DB,$9F,$1B,$97,$8F,$CD,$FE,$4D,$F5,$8F,$91,$BF
	.dc.b	$45,$86,$DF,$E7,$7E,$B3,$F3,$1C,$3F,$DD,$17,$C3,$FF,$2F,$73,$AF
	.dc.b	$FB,$A1,$CD,$71,$9D,$78,$9B,$6C,$75,$8F,$E7,$1F,$59,$F9,$5C,$3B
	.dc.b	$46,$3C,$3E,$DF,$09,$DF,$92,$E1,$E5,$70,$FE,$CF,$07,$9B,$8C,$7F
	.dc.b	$C1,$71,$89,$FD,$C7,$D5,$7C,$89,$FE,$B4,$FE,$3D,$A7,$F4,$B7,$09
	.dc.b	$B2,$F9,$8D,$72,$7A,$9E,$1E,$0F,$1F,$22,$E3,$E0,$70,$F8,$BD,$FB
	.dc.b	$97,$F8,$83,$24,$78,$4F,$8B,$F5,$56,$BD,$CF,$97,$E6,$FC,$DE,$A5
	.dc.b	$9F,$6B,$87,$E2,$7D,$65,$C7,$E1,$BE,$DF,$57,$DD,$E4,$5A,$F8,$AE
	.dc.b	$1F,$E3,$DF,$58,$71,$7F,$E3,$68,$C4,$7C,$CE,$2F,$91,$C1,$F2,$5B
	.dc.b	$1F,$42,$3E,$66,$98,$BF,$80,$E2,$7F,$96,$7D,$67,$ED,$9F,$AF,$1F
	.dc.b	$DF,$63,$13,$FE,$46,$FA,$CF,$CD,$57,$F3,$35,$DB,$FE,$03,$73,$3B
	.dc.b	$8C,$1F,$90,$E0,$B5,$F8,$97,$87,$95,$C7,$D8,$B3,$D5,$C3,$FD,$07
	.dc.b	$A3,$C7,$DD,$F1,$71,$FF,$12,$65,$FD,$51,$FF,$2D,$CE,$AB,$FC,$35
	.dc.b	$2F,$9A,$F3,$7C,$D7,$9D,$B3,$EB,$8F,$57,$C6,$E3,$0F,$80,$E2,$7C
	.dc.b	$59,$D9,$EE,$FC,$1F,$2F,$E1,$7C,$65,$AF,$8C,$E1,$FE,$B5,$F5,$6F
	.dc.b	$AC,$C6,$1F,$E7,$28,$C4,$FC,$DF,$77,$FC,$5D,$C4,$F7,$7E,$74,$61
	.dc.b	$93,$FD,$97,$9D,$F5,$98,$C7,$FC,$FB,$18,$9F,$ED,$5F,$57,$FD,$9E
	.dc.b	$7F,$C7,$C1,$C3,$FE,$41,$6B,$FE,$51,$6A,$FA,$C7,$C9,$7F,$93,$A3
	.dc.b	$FC,$AD,$18,$AF,$F3,$3B,$63,$B9,$C6,$78,$3C,$FE,$7E,$C9,$A4,$AF
	.dc.b	$F3,$8D,$FF,$26,$AF,$4B,$5F,$47,$C1,$74,$8C,$47,$F9,$A7,$EA,$FF
	.dc.b	$C9,$3A,$7F,$9E,$9B,$1E,$47,$FC,$BB,$E4,$67,$E1,$FE,$84,$FA,$3E
	.dc.b	$AF,$FD,$CE,$DF,$67,$FC,$EB,$18,$8F,$96,$D7,$FC,$D5,$F5,$6F,$43
	.dc.b	$FC,$D9,$DF,$F9,$99,$4F,$F3,$E7,$FC,$3C,$D7,$FD,$6A,$F1,$73,$3C
	.dc.b	$B6,$EB,$E8,$D9,$FF,$43,$57,$1B,$EC,$F1,$8E,$77,$B7,$C3,$56,$78
	.dc.b	$5F,$A2,$B0,$66,$65,$78,$88,$79,$9C,$BD,$8C,$DF,$8A,$32,$F9,$27
	.dc.b	$AC,$FC,$09,$CD,$78,$63,$8B,$EE,$74,$60,$F7,$CF,$07,$E0,$CF,$47
	.dc.b	$C4,$79,$4E,$26,$2B,$C4,$E1,$F3,$3D,$4F,$AB,$3D,$5F,$8E,$66,$FE
	.dc.b	$64,$F0,$8E,$C7,$5B,$F5,$38,$8D,$1E,$07,$47,$A0,$F1,$47,$08,$D1
	.dc.b	$C1,$F8,$51,$C0,$F2,$9C,$0E,$46,$67,$CB,$1C,$27,$94,$FC,$83,$EC
	.dc.b	$9F,$C0,$A7,$F0,$CD,$AF,$F6,$8C,$62,$3D,$66,$4F,$F7,$1B,$97,$A1
	.dc.b	$FB,$E7,$C2,$E2,$7D,$75,$FE,$CD,$F9,$37,$D0,$7D,$79,$CD,$6E,$B5
	.dc.b	$39,$7E,$21,$23,$FD,$C9,$5C,$9F,$A9,$86,$DC,$D5,$7E,$39,$F2,$67
	.dc.b	$83,$F1,$E7,$51,$8B,$F1,$CF,$06,$C6,$6F,$C5,$39,$6C,$66,$3D,$44
	.dc.b	$34,$79,$0F,$39,$F8,$F7,$83,$A2,$61,$F2,$94,$6A,$30,$4B,$FD,$E4
	.dc.b	$65,$F8,$11,$C6,$31,$3E,$01,$7F,$C4,$E3,$31,$F0,$CF,$F5,$83,$88
	.dc.b	$F8,$73,$CF,$0E,$B7,$F2,$1B,$7E,$B7,$F8,$67,$C3,$FF,$0F,$FA,$BF
	.dc.b	$F1,$7C,$FF,$AE,$DA,$FF,$1E,$8D,$BF,$8C,$7F,$22,$47,$FA,$48,$FA
	.dc.b	$A7,$C2,$62,$30,$D8,$CC,$70,$6B,$FD,$49,$B9,$7F,$33,$6E,$89,$91
	.dc.b	$87,$F1,$FE,$37,$F2,$98,$7A,$9F,$CA,$FE,$AD,$E7,$D3,$51,$2F,$C4
	.dc.b	$F9,$48,$FB,$3E,$AF,$FB,$CB,$FE,$DF,$6B,$F5,$B1,$F2,$28,$CE,$5F
	.dc.b	$D2,$D7,$A9,$FC,$F5,$B3,$CE,$30,$7F,$41,$FA,$77,$C7,$EF,$E9,$65
	.dc.b	$78,$5E,$5E,$AC,$8F,$23,$5F,$21,$F1,$FE,$AD,$C1,$E1,$F7,$64,$FA
	.dc.b	$DF,$27,$8B,$23,$DC,$D7,$99,$ED,$F5,$BD,$44,$7E,$23,$FD,$67,$EA
	.dc.b	$DE,$07,$F5,$BF,$AB,$FE,$D9,$C3,$FD,$B5,$BC,$5F,$21,$B8,$7D,$D7
	.dc.b	$FD,$68,$FD,$F9,$E1,$F0,$3C,$C9,$19,$AF,$A3,$F5,$6D,$3C,$1F,$19
	.dc.b	$2B,$47,$57,$C9,$E9,$C8,$C8,$C9,$19,$8F,$83,$C9,$F7,$31,$F4,$7D
	.dc.b	$57,$B4,$93,$CA,$FF,$CB,$8D,$F5,$FE,$11,$1D,$7A,$FF,$14,$8E,$0F
	.dc.b	$1F,$9A,$F8,$9F,$95,$F1,$92,$7B,$3F,$37,$D8,$91,$E1,$6B,$E8,$EA
	.dc.b	$FA,$5F,$21,$23,$CC,$FD,$5F,$56,$47,$A5,$AF,$E0,$3E,$AD,$B7,$F8
	.dc.b	$57,$E6,$3F,$CC,$D3,$F2,$1A,$FE,$19,$BC,$FF,$E2,$B5,$F4,$1C,$49
	.dc.b	$3C,$5E,$9F,$41,$23,$A3,$5F,$F3,$97,$C8,$F8,$7E,$82,$47,$85,$F2
	.dc.b	$7C,$FC,$8F,$23,$5F,$F3,$D7,$9D,$E9,$F4,$A4,$7A,$5F,$67,$A1,$23
	.dc.b	$D8,$D7,$FC,$EF,$C3,$F0,$7E,$36,$47,$C1,$7E,$27,$8B,$23,$E2,$35
	.dc.b	$F8,$9F,$27,$F2,$48,$FC,$97,$FA,$29,$FE,$41,$FA,$BF,$C8,$C7,$E5
	.dc.b	$1A,$FE,$C9,$BC,$3F,$E5,$75,$F0,$DC,$7E,$62,$57,$17,$A7,$3F,$F9
	.dc.b	$37,$EA,$DF,$3F,$49,$5F,$A2,$87,$D5,$EF,$82,$4E,$5E,$5C,$1F,$E5
	.dc.b	$7F,$AB,$FE,$65,$26,$93,$D3,$D7,$A9,$27,$F5,$F3,$FC,$D1,$BF,$37
	.dc.b	$FE,$61,$F7,$BF,$DB,$31,$FB,$46,$BF,$C4,$9B,$C3,$FE,$77,$5D,$9C
	.dc.b	$7C,$19,$5C,$5E,$9E,$0F,$F9,$B7,$EA,$FF,$D0,$F8,$7F,$9F,$DB,$E4
	.dc.b	$7F,$3A,$6D,$FF,$4F,$5D,$DC,$7B,$64,$FF,$08,$94,$75,$E4,$ED,$E2
	.dc.b	$4A,$D3,$FE,$7D,$FA,$BF,$F5,$EE,$5F,$E8,$F5,$F3,$79,$25,$7F,$14
	.dc.b	$87,$CD,$FF,$A1,$7E,$AF,$FD,$7B,$3F,$C6,$1A,$FF,$36,$6D,$FF,$A6
	.dc.b	$D6,$FF,$E8,$DF,$3D,$E2,$D2,$5F,$F2,$28,$F8,$29,$78,$5C,$49,$1B
	.dc.b	$F0,$BA,$7D,$44,$8E,$4F,$7D,$BF,$D3,$9F,$56,$F9,$9A,$4A,$FE,$63
	.dc.b	$1F,$09,$FE,$67,$F5,$7F,$D1,$13,$49,$EB,$7F,$A9,$D7,$CA,$52,$4F
	.dc.b	$2F,$8F,$D6,$F2,$3A,$BC,$7A,$A4,$78,$1A,$FF,$60,$B7,$CA,$77,$F1
	.dc.b	$E4,$F9,$9F,$17,$8F,$23,$D2,$D7,$97,$A3,$FE,$BD,$FA,$BF,$C5,$C7
	.dc.b	$F5,$06,$BF,$DB,$9B,$CF,$FE,$D3,$5E,$7F,$12,$4F,$17,$FC,$9B,$C7
	.dc.b	$FE,$C9,$F9,$7F,$F0,$4C,$FF,$5C,$6B,$FD,$E9,$B7,$FE,$DF,$7F,$72
	.dc.b	$4F,$87,$E4,$7F,$96,$39,$7F,$DA,$9F,$55,$FC,$EA,$FF,$B3,$B5,$FF
	.dc.b	$02,$B6,$FF,$DD,$4F,$E3,$A3,$7D,$5C,$88,$F2,$F1,$7A,$91,$C1,$FD
	.dc.b	$D3,$E1,$FF,$4F,$E2,$FF,$8B,$AF,$E2,$B7,$C5,$F2,$9E,$6D,$A5,$F9
	.dc.b	$9A,$F9,$BC,$4F,$F3,$6A,$FA,$BF,$F2,$FE,$1F,$BA,$D3,$FC,$B6,$D9
	.dc.b	$FF,$1E,$9F,$E5,$17,$0E,$1A,$F9,$7C,$57,$D5,$F2,$6C,$72,$13,$FC
	.dc.b	$CA,$74,$7F,$99,$E1,$F4,$B4,$8F,$F9,$AD,$FF,$36,$7C,$13,$5F,$F0
	.dc.b	$CC,$FF,$92,$1B,$19,$8F,$F1,$C9,$92,$FC,$46,$72,$D4,$3D,$0F,$E5
	.dc.b	$77,$D2,$B4,$FF,$97,$A3,$33,$D3,$93,$2F,$F9,$8B,$D0,$AF,$F9,$DE
	.dc.b	$2B,$FE,$74,$AF,$3C,$E3,$E8,$72,$D8,$F5,$C7,$2B,$7C,$96,$6F,$FC
	.dc.b	$A7,$18,$AF,$F1,$CE,$4E,$5F,$0B,$FC,$E2,$33,$18,$E4,$91,$FE,$40
	.dc.b	$A7,$81,$8C,$3A,$BE,$AF,$AB,$FF,$2D,$D7,$F9,$19,$F9,$58,$FC,$87
	.dc.b	$35,$FE,$66,$73,$3F,$E8,$46,$C7,$FA,$66,$DD,$DD,$77,$F7,$3D,$FC
	.dc.b	$7F,$07,$9D,$71,$F0,$38,$7C,$5E,$6D,$C7,$C0,$F9,$FE,$2F,$D1,$EA
	.dc.b	$5A,$F4,$38,$7D,$5E,$9D,$BF,$43,$EF,$F8,$6F,$81,$F1,$56,$BE,$03
	.dc.b	$F7,$3E,$57,$FC,$1F,$1F,$E6,$85,$F5,$79,$F7,$34,$EF,$7E,$1E,$4D
	.dc.b	$BE,$0F,$E1,$BE,$AF,$FC,$1F,$1F,$E7,$25,$F3,$39,$F7,$74,$EF,$97
	.dc.b	$87,$C9,$DB,$E0,$FE,$43,$EA,$FF,$1D,$C3,$FE,$98,$5D,$D9,$EF,$D3
	.dc.b	$BE,$3E,$1C,$77,$1F,$F4,$1D,$FF,$A2,$DF,$F4,$2F,$D5,$BD,$0F,$CE
	.dc.b	$BF,$6A,$33,$18,$E1,$FD,$0C,$7F,$8B,$8F,$F4,$69,$EE,$CF,$FD,$56
	.dc.b	$E5,$BB,$FE,$A3,$EA,$FF,$DB,$D3,$FE,$94,$7F,$1B,$DD,$FB,$67,$35
	.dc.b	$FE,$B4,$73,$3F,$ED,$27,$FD,$E2,$F1,$DB,$A7,$87,$D5,$ED,$D3,$DF
	.dc.b	$7B,$5D,$DF,$DC,$F8,$BF,$E0,$18,$FF,$52,$BF,$EA,$46,$FB,$7A,$F8
	.dc.b	$FD,$FA,$DC,$3C,$4B,$5C,$1F,$E0,$7F,$57,$FB,$58,$FF,$57,$BF,$B5
	.dc.b	$DB,$FD,$9A,$E5,$BE,$37,$0F,$73,$8F,$C3,$5B,$E2,$F4,$F5,$3F,$87
	.dc.b	$F9,$9F,$13,$93,$F8,$DE,$5F,$EA,$9B,$FF,$5E,$CE,$5F,$F7,$BB,$FE
	.dc.b	$30,$73,$E3,$72,$F1,$F9,$BF,$C6,$BE,$AD,$EE,$DF,$A1,$C3,$6F,$FD
	.dc.b	$93,$F5,$7F,$98,$6F,$C3,$FF,$22,$8C,$70,$FF,$BA,$94,$64,$D7,$89
	.dc.b	$6B,$9B,$87,$FD,$AB,$F5,$7F,$93,$BE,$53,$FC,$BA,$31,$BF,$F3,$AC
	.dc.b	$3F,$C6,$F7,$E4,$E2,$F1,$7C,$1F,$8D,$91,$D3,$CD,$B9,$7F,$BF,$B1
	.dc.b	$47,$F3,$AF,$AB,$F9,$12,$0F,$E4,$1C,$3A,$73,$C5,$C4,$F3,$7C,$C5
	.dc.b	$C3,$D6,$F5,$EF,$ED,$E4,$59,$FF,$79,$E1,$FB,$86,$C7,$8F,$DE,$DC
	.dc.b	$69,$C3,$F1,$BE,$A6,$E5,$EB,$7D,$5F,$37,$EB,$F5,$2D,$76,$70,$FC
	.dc.b	$EF,$AB,$B7,$F0,$DF,$85,$EA,$FC,$3F,$1A,$D7,$99,$C3,$F1,$79,$7F
	.dc.b	$B5,$FD,$5F,$F7,$58,$D1,$93,$E5,$60,$8D,$29,$FE,$E7,$B7,$D2,$AF
	.dc.b	$87,$C1,$F4,$F8,$BE,$A3,$23,$E3,$ED,$70,$E5,$1D,$AD,$3F,$6C,$FE
	.dc.b	$EB,$F5,$7F,$BD,$AD,$9F,$6A,$33,$E3,$7F,$7A,$FA,$BF,$D0,$E1,$E9
	.dc.b	$AF,$C1,$43,$B7,$5D,$BB,$65,$7E,$63,$92,$DF,$E3,$5E,$9E,$3F,$5F
	.dc.b	$5E,$D7,$5A,$63,$FD,$39,$F2,$98,$9F,$F0,$6F,$5F,$A7,$87,$FC,$33
	.dc.b	$5B,$7F,$C3,$71,$9B,$F9,$F2,$F9,$AF,$67,$CD,$7B,$5B,$3E,$19,$4F
	.dc.b	$C1,$F8,$CC,$E5,$E2,$9F,$F4,$7D,$6E,$7C,$79,$EE,$3D,$58,$FF,$27
	.dc.b	$BE,$1E,$57,$CB,$F8,$AF,$98,$DA,$F0,$47,$F8,$A3,$EA,$9F,$57,$9B
	.dc.b	$F8,$31,$8A,$F3,$7B,$9F,$E2,$EE,$27,$BD,$F5,$E3,$0C,$9F,$99,$D4
	.dc.b	$FA,$BC,$63,$F3,$23,$13,$F4,$7E,$AB,$F8,$4C,$FF,$95,$DA,$FF,$98
	.dc.b	$9A,$FE,$69,$AF,$F0,$5F,$AB,$FF,$01,$CF,$F9,$7E,$11,$87,$FC,$C4
	.dc.b	$D7,$FC,$DA,$D7,$F4,$AB,$E7,$7F,$CA,$B1,$86,$BD,$DA,$F6,$6F,$D5
	.dc.b	$FF,$92,$32,$F1,$36,$7F,$CE,$4B,$E6,$1E,$4E,$BE,$23,$E5,$D7,$F9
	.dc.b	$DF,$EA,$5E,$B7,$AE,$11,$87,$DC,$D7,$84,$6E,$31,$D9,$FE,$6E,$6E
	.dc.b	$BF,$C0,$9B,$5F,$CB,$7E,$AF,$F0,$2F,$E8,$E1,$F2,$7F,$C2,$2F,$FC
	.dc.b	$EF,$5F,$EA,$FB,$FE,$16,$DB,$FE,$6D,$7C,$1E,$57,$C6,$B7,$F4,$16
	.dc.b	$7C,$0D,$7D,$B3,$6B,$94,$7E,$CA,$FB,$26,$C6,$B0,$F7,$57,$39,$D9
	.dc.b	$A7,$A1,$9B,$97,$E6,$21,$DD,$CB,$FC,$0A,$FB,$61,$F2,$6B,$38,$7C
	.dc.b	$4B,$EB,$7C,$A3,$0F,$F0,$DC,$3C,$4E,$6F,$E1,$B9,$AE,$2F,$7B,$EB
	.dc.b	$3A,$8C,$1C,$4F,$92,$71,$30,$7C,$A3,$07,$AC,$CE,$1E,$A7,$89,$93
	.dc.b	$A9,$30,$F4,$B9,$7E,$6B,$DB,$0F,$9B,$87,$A2,$73,$3B,$BF,$44,$60
	.dc.b	$8E,$09,$1F,$88,$C3,$94,$62,$3E,$F5,$61,$D1,$FA,$09,$FC,$F1,$1C
	.dc.b	$10,$FD,$53,$96,$9F,$09,$FC,$77,$B3,$D6,$3E,$41,$F0,$6B,$F4,$D1
	.dc.b	$9C,$7E,$44,$EB,$0F,$D6,$4E,$6F,$4D,$9F,$9E,$D7,$A1,$FA,$67,$13
	.dc.b	$F1,$23,$07,$94,$E2,$53,$89,$FC,$0A,$38,$DF,$E4,$E3,$F7,$11,$87
	.dc.b	$E3,$9F,$99,$70,$FA,$63,$DA,$F5,$9F,$13,$D4,$C4,$75,$79,$18,$3A
	.dc.b	$C7,$C2,$70,$75,$3D,$67,$A2,$FE,$49,$2B,$C6,$D8,$FE,$E5,$F5,$2F
	.dc.b	$7F,$F8,$22,$F8,$BA,$62,$24,$FF,$3F,$9E,$0D,$8F,$E5,$11,$88,$FF
	.dc.b	$10,$9F,$E4,$BF,$AA,$7D,$7F,$F2,$D6,$C7,$F4,$96,$BC,$03,$87,$9B
	.dc.b	$AC,$99,$FF,$8B,$A3,$0D,$AF,$E3,$D1,$FD,$8E,$BD,$FE,$A7,$F2,$1F
	.dc.b	$96,$7C,$1D,$3E,$0C,$49,$F7,$3E,$C2,$3F,$D7,$FE,$AF,$A3,$FE,$18
	.dc.b	$6E,$5E,$48,$FF,$35,$46,$32,$FF,$21,$AF,$5F,$F9,$93,$67,$FC,$48
	.dc.b	$D7,$A4,$7F,$37,$FA,$8F,$A9,$1A,$FA,$4E,$24,$78,$FA,$FD,$08,$E0
	.dc.b	$FF,$18,$7D,$5B,$C9,$F0,$FC,$74,$9F,$0B,$E4,$F0,$E4,$79,$1A,$FB
	.dc.b	$FD,$7F,$49,$1F,$4B,$FD,$37,$EA,$FF,$B7,$C7,$F8,$FF,$EA,$BE,$23
	.dc.b	$FE,$3E,$6F,$06,$1B,$C1,$FE,$68,$7F,$D6,$73,$C7,$D5,$FE,$B1,$18
	.dc.b	$8E,$51,$FD,$6F,$EA,$D9,$EB,$D7,$95,$E1,$78,$FA,$32,$3C,$8D,$7F
	.dc.b	$CA,$3C,$5F,$5D,$FF,$51,$63,$FE,$15,$7D,$5E,$47,$88,$F1,$FA,$B2
	.dc.b	$3D,$CD,$7D,$DE,$AF,$A0,$8F,$C3,$7F,$B9,$7D,$5B,$C2,$F6,$FC,$5C
	.dc.b	$AF,$90,$FC,$0F,$56,$47,$CB,$6B,$E8,$E9,$F8,$BF,$1F,$23,$E8,$3F
	.dc.b	$23,$D2,$91,$F4,$DA,$FF,$9C,$3E,$AF,$DC,$FF,$9C,$BC,$EF,$F0,$AB
	.dc.b	$7D,$6F,$F0,$44,$63,$A3,$ED,$BF,$E0,$62,$79,$7A,$37,$F4,$12,$FF
	.dc.b	$D0,$12,$8C,$47,$F9,$83,$E3,$BA,$FD,$04,$8D,$BE,$0F,$A0,$91,$C9
	.dc.b	$AF,$F9,$A3,$EA,$FC,$BF,$1B,$23,$BB,$E8,$F8,$D9,$1E,$26,$BF,$E6
	.dc.b	$5E,$1F,$6F,$C6,$48,$F3,$BF,$03,$C5,$91,$EA,$6B,$F1,$FD,$5F,$8A
	.dc.b	$47,$DE,$FF,$8B,$FE,$B3,$FC,$83,$0F,$AC,$78,$1F,$E3,$36,$C7,$FA
	.dc.b	$95,$AF,$FA,$91,$7C,$26,$FE,$5E,$4E,$DE,$5D,$3F,$E3,$DF,$AC,$7D
	.dc.b	$1D,$25,$7F,$AC,$61,$F3,$3B,$E1,$93,$97,$97,$0F,$F9,$1F,$EB,$3F
	.dc.b	$9E,$75,$7E,$E9,$BE,$5F,$FB,$79,$B6,$0B,$E4,$B9,$7C,$84,$9F,$F6
	.dc.b	$14,$A3,$B7,$AD,$FE,$55,$FA,$2F,$F6,$85,$7F,$B1,$DA,$FF,$B1,$9B
	.dc.b	$8F,$F0,$C7,$FC,$FD,$1B,$F1,$39,$11,$E5,$E2,$7F,$98,$7E,$B1,$54
	.dc.b	$93,$FE,$D5,$87,$7E,$FB,$A4,$E5,$E5,$C7,$CF,$C6,$93,$C5,$FF,$37
	.dc.b	$7D,$67,$F6,$2D,$3F,$90,$AF,$9A,$C9,$2B,$FD,$C9,$0F,$9A,$FF,$3B
	.dc.b	$7D,$63,$EA,$E9,$2B,$FD,$D5,$0E,$0F,$F3,$E7,$A5,$FA,$D9,$FF,$77
	.dc.b	$35,$FF,$76,$B7,$2F,$E6,$6F,$C2,$8D,$F8,$7C,$88,$F2,$F0,$FA,$FC
	.dc.b	$2C,$8E,$AF,$83,$6F,$F4,$67,$D6,$3E,$6A,$92,$BF,$E0,$04,$7C,$17
	.dc.b	$F0,$D9,$7F,$44,$FA,$CF,$E8,$31,$F8,$96,$A6,$9F,$D0,$63,$87,$CA
	.dc.b	$7F,$A6,$E0,$D3,$1B,$E7,$69,$7E,$0E,$DF,$CB,$3D,$63,$97,$4C,$8D
	.dc.b	$1E,$37,$0B,$1F,$47,$CA,$78,$7C,$49,$3C,$0F,$43,$E5,$F1,$24,$67
	.dc.b	$44,$62,$3A,$65,$E4,$97,$A5,$FE,$B7,$FA,$CF,$C9,$D7,$56,$BC,$0F
	.dc.b	$D8,$38,$CB,$FB,$7A,$F3,$B9,$12,$79,$3F,$C5,$78,$FF,$D8,$9F,$2F
	.dc.b	$FE,$0E,$9E,$81,$C5,$FF,$7A,$6E,$5F,$29,$FE,$03,$3B,$F8,$AE,$64
	.dc.b	$7E,$27,$9B,$FD,$7D,$F8,$DF,$57,$FE,$CE,$27,$A9,$FE,$D5,$7F,$2D
	.dc.b	$87,$D6,$6C,$FF,$0D,$6A,$69,$F9,$B7,$D2,$DC,$7F,$C2,$2D,$9E,$9E
	.dc.b	$27,$F9,$0E,$BC,$1E,$1C,$7C,$AF,$8F,$59,$1E,$86,$BD,$BC,$3F,$C2
	.dc.b	$2F,$AC,$F9,$0F,$9E,$32,$FF,$99,$31,$46,$23,$DC,$D7,$86,$5F,$E6
	.dc.b	$37,$DD,$E2,$3E,$B3,$17,$FC,$F9,$B1,$99,$FF,$35,$4E,$5E,$43,$E9
	.dc.b	$2B,$4C,$B8,$37,$F2,$7F,$E5,$BC,$3F,$C7,$33,$FE,$85,$DB,$F3,$CA
	.dc.b	$75,$1F,$E8,$0F,$90,$B3,$FA,$E6,$AF,$39,$EC,$E2,$3A,$1F,$90,$8C
	.dc.b	$D7,$6E,$4D,$79,$FB,$6C,$78,$A3,$97,$C6,$62,$FF,$32,$1C,$CF,$ED
	.dc.b	$6D,$F2,$78,$AF,$03,$E6,$D3,$FA,$2D,$7E,$ED,$FE,$86,$DE,$5D,$72
	.dc.b	$C8,$E9,$E0,$5C,$3D,$4F,$C0,$FA,$CF,$8D,$3F,$C1,$5F,$7E,$1F,$E6
	.dc.b	$47,$35,$FB,$07,$33,$FC,$59,$B1,$FD,$B7,$2F,$E1,$33,$CF,$7F,$9B
	.dc.b	$18,$9E,$5E,$37,$87,$CE,$F1,$36,$FC,$51,$E5,$F3,$71,$5E,$17,$D1
	.dc.b	$F1,$9E,$9F,$4A,$DF,$A5,$C3,$EB,$F4,$AD,$FA,$5F,$81,$F0,$DF,$07
	.dc.b	$E2,$ED,$7C,$17,$F0,$7F,$2D,$FE,$1C,$8F,$E5,$0B,$EB,$33,$ED,$E9
	.dc.b	$DF,$95,$C3,$C9,$B7,$C1,$FC,$6F,$D6,$7F,$87,$23,$F9,$92,$FA,$CC
	.dc.b	$FB,$9A,$77,$C9,$C3,$EB,$2D,$F0,$7F,$2F,$F5,$9F,$E1,$98,$FE,$78
	.dc.b	$FE,$2D,$BD,$FA,$EF,$DF,$17,$0E,$2B,$5C,$1F,$D0,$FD,$63,$D1,$BC
	.dc.b	$18,$6D,$7F,$49,$3F,$C7,$25,$DB,$F5,$2E,$39,$7F,$9A,$BE,$F7,$F9
	.dc.b	$BF,$D6,$7F,$8C,$63,$FA,$93,$F8,$CE,$EF,$F5,$8B,$99,$DF,$8B,$FB
	.dc.b	$48,$78,$97,$7C,$37,$0E,$7D,$5B,$DF,$83,$63,$FB,$07,$8C,$F6,$FF
	.dc.b	$7E,$FF,$97,$A3,$31,$8F,$83,$CF,$AD,$AF,$16,$D7,$27,$F8,$3F,$D6
	.dc.b	$7E,$CB,$7F,$EA,$6F,$EC,$9B,$E4,$6B,$C8,$DF,$B9,$C3,$E1,$6E,$3F
	.dc.b	$DC,$F0,$47,$4F,$77,$F8,$9F,$27,$C3,$E5,$FF,$6A,$D7,$FA,$CE,$FF
	.dc.b	$BD,$4E,$5F,$F0,$3B,$FE,$39,$73,$E3,$F2,$F2,$39,$BF,$C7,$BE,$B1
	.dc.b	$F2,$77,$E8,$F0,$DB,$FF,$03,$7D,$67,$E5,$5B,$F0,$FF,$C9,$A3,$1C
	.dc.b	$5F,$DC,$94,$64,$D7,$8B,$6B,$9B,$87,$FC,$25,$F5,$9F,$EE,$47,$26
	.dc.b	$3C,$3E,$DF,$09,$DF,$92,$CF,$91,$C3,$CB,$CE,$B9,$FF,$81,$30,$47
	.dc.b	$F3,$CF,$AA,$F8,$D3,$FE,$81,$7F,$21,$5F,$E2,$E7,$30,$72,$EF,$8B
	.dc.b	$FA,$2D,$BE,$67,$A4,$61,$ED,$E1,$77,$F2,$2E,$3F,$E2,$FC,$7F,$69
	.dc.b	$8F,$F8,$61,$BC,$E9,$F1,$7B,$FC,$5E,$0F,$A4,$F0,$FD,$4D,$8F,$92
	.dc.b	$FB,$7E,$73,$DD,$EA,$D9,$F0,$38,$7E,$7F,$D6,$5C,$BE,$33,$F1,$7D
	.dc.b	$6F,$8D,$E3,$DA,$F4,$38,$7F,$C9,$1F,$59,$E7,$72,$60,$F9,$DC,$7F
	.dc.b	$3F,$86,$C7,$C1,$70,$FC,$3D,$2E,$5F,$0D,$FA,$C7,$F7,$2F,$AC,$FD
	.dc.b	$B4,$EC,$FF,$2C,$38,$9F,$F2,$0F,$D6,$7E,$6A,$7F,$CB,$8B,$B7,$FB
	.dc.b	$DB,$99,$DC,$60,$FB,$7C,$16,$BF,$26,$F0,$F2,$38,$FB,$16,$7A,$B8
	.dc.b	$7F,$CD,$5D,$3E,$3E,$4F,$B5,$8F,$FB,$37,$2F,$E9,$0F,$F9,$5A,$75
	.dc.b	$5F,$E7,$C9,$7C,$D7,$9B,$E6,$BC,$ED,$9F,$5C,$7A,$BE,$37,$18,$7C
	.dc.b	$07,$13,$E2,$CE,$CF,$77,$E0,$F9,$7F,$0B,$E3,$2D,$7C,$67,$0F,$FA
	.dc.b	$0F,$EB,$1F,$33,$8C,$3F,$CD,$B1,$89,$F9,$BE,$EF,$F8,$9B,$89,$EF
	.dc.b	$FC,$E8,$C3,$27,$FD,$1D,$CE,$FA,$CC,$63,$FE,$78,$8C,$4F,$FA,$4F
	.dc.b	$EA,$FF,$B6,$3F,$E4,$B6,$CF,$FA,$88,$FF,$7C,$38,$70,$D7,$FD,$4E
	.dc.b	$FF,$B3,$E3,$97,$A7,$FE,$44,$9E,$2D,$53,$B7,$B0,$FA,$BF,$F1,$8E
	.dc.b	$3F,$E5,$D6,$C6,$9F,$F2,$2F,$8B,$9F,$85,$F8,$73,$CA,$DC,$3F,$E5
	.dc.b	$C6,$C7,$B6,$30,$7F,$98,$3E,$AF,$E8,$C7,$F9,$AD,$AF,$91,$AF,$C3
	.dc.b	$6B,$FE,$C6,$77,$D3,$F9,$28,$46,$34,$D7,$17,$FB,$32,$B5,$3F,$EB
	.dc.b	$39,$FF,$67,$FD,$5F,$FA,$86,$FF,$D5,$AB,$E4,$78,$4E,$9B,$21,$3C
	.dc.b	$6F,$C4,$DA,$E1,$3F,$ED,$18,$E8,$DF,$57,$9B,$C9,$B1,$A8,$F1,$DE
	.dc.b	$26,$47,$38,$D9,$87,$D3,$18,$74,$77,$21,$D0,$C9,$C0,$CD,$FA,$A7
	.dc.b	$94,$64,$91,$97,$8B,$F0,$27,$C4,$E6,$3B,$1C,$2F,$E7,$CF,$49,$D1
	.dc.b	$C6,$7C,$33,$F2,$1E,$31,$97,$91,$F4,$6F,$E3,$47,$29,$EB,$3C,$8D
	.dc.b	$5F,$78,$CC,$72,$3E,$CC,$65,$E9,$7E,$A9,$F1,$57,$59,$C9,$D4,$E2
	.dc.b	$7C,$28,$EB,$1E,$73,$84,$68,$F0,$5F,$E1,$5F,$29,$C4,$EA,$7C,$F3
	.dc.b	$F8,$13,$F9,$56,$D7,$FB,$22,$31,$1E,$C3,$27,$FB,$55,$CB,$D0,$C9
	.dc.b	$E2,$8F,$C8,$FC,$8B,$CD,$F8,$13,$E2,$BF,$56,$1F,$9F,$85,$6D,$F8
	.dc.b	$58,$7A,$DF,$CD,$1F,$A5,$AF,$DA,$4E,$E3,$C0,$7C,$97,$11,$DA,$3C
	.dc.b	$15,$F2,$8F,$04,$E9,$C4,$68,$87,$92,$FD,$2F,$88,$FB,$D3,$C5,$DD
	.dc.b	$F9,$CF,$44,$66,$38,$5F,$8D,$B7,$E2,$BE,$B7,$C7,$0F,$93,$18,$7C
	.dc.b	$38,$73,$9D,$5F,$BE,$B4,$D3,$E4,$D7,$08,$CC,$7A,$1B,$87,$8E,$FA
	.dc.b	$56,$30,$F2,$5F,$23,$A0,$AF,$E4,$9F,$57,$FE,$8A,$8F,$A9,$3F,$6E
	.dc.b	$31,$1B,$8C,$7B,$DF,$84,$F0,$3F,$8A,$D6,$AF,$9C,$E0,$D5,$F4,$3F
	.dc.b	$92,$7D,$53,$EA,$FF,$8E,$B6,$34,$1C,$3E,$13,$8F,$4F,$03,$FA,$AB
	.dc.b	$82,$C7,$F6,$D6,$C6,$96,$0F,$3F,$0D,$8E,$6D,$7F,$B0,$B5,$FE,$E0
	.dc.b	$E7,$1F,$93,$3F,$08,$FE,$6F,$F5,$6F,$4F,$F9,$CF,$D5,$F8,$F0,$F5
	.dc.b	$07,$17,$FD,$55,$BB,$7E,$AD,$7A,$3C,$89,$5F,$1A,$32,$FC,$8F,$A8
	.dc.b	$7C,$77,$83,$E8,$A4,$ED,$F1,$FA,$52,$39,$35,$F2,$FD,$5B,$93,$D3
	.dc.b	$AC,$9F,$03,$EC,$F0,$E4,$78,$DA,$FB,$B3,$FF,$42,$3E,$7F,$47,$E1
	.dc.b	$91,$FB,$FF,$56,$F0,$BF,$AF,$FD,$57,$F7,$7E,$2F,$EF,$0D,$E2,$FE
	.dc.b	$1C,$DB,$FE,$DC,$FF,$AE,$1F,$C4,$B5,$F9,$ED,$9E,$7E,$F7,$F6,$CF
	.dc.b	$AB,$FA,$D5,$C3,$E1,$FC,$24,$79,$F9,$DE,$42,$38,$3E,$87,$57,$A3
	.dc.b	$EA,$E4,$78,$9F,$5F,$A5,$23,$CA,$D7,$CB,$E9,$7C,$02,$3E,$A7,$FC
	.dc.b	$03,$F5,$7F,$CA,$75,$79,$DF,$18,$95,$F0,$3C,$AF,$92,$47,$07,$E1
	.dc.b	$7A,$BE,$6F,$C7,$C8,$F8,$EF,$D0,$F4,$24,$7C,$A6,$BF,$98,$FA,$B6
	.dc.b	$BF,$E1,$4F,$98,$FF,$15,$4F,$F8,$7D,$A9,$A3,$E7,$F8,$12,$3F,$7F
	.dc.b	$19,$6F,$D5,$F5,$23,$83,$FC,$E9,$F1,$BE,$1F,$A0,$91,$C9,$F2,$7C
	.dc.b	$FC,$8E,$AD,$7F,$CF,$3F,$57,$E9,$F8,$A9,$1E,$27,$D9,$F1,$52,$3C
	.dc.b	$AD,$7F,$CE,$DC,$1F,$07,$D2,$91,$EA,$7E,$27,$8B,$23,$DA,$D7,$F8
	.dc.b	$0E,$AF,$90,$F9,$44,$8F,$86,$FF,$70,$3F,$C8,$3F,$56,$EC,$A4,$9F
	.dc.b	$E4,$F0,$F8,$2D,$FC,$BC,$9C,$BC,$B9,$FF,$C9,$7F,$56,$FA,$1A,$4A
	.dc.b	$FE,$63,$0F,$99,$DF,$0C,$9C,$BC,$B8,$7F,$CA,$BF,$57,$FD,$07,$C7
	.dc.b	$FF,$01,$36,$BF,$CC,$4D,$4B,$E4,$38,$FC,$6C,$9F,$E8,$32,$8E,$BF
	.dc.b	$33,$FE,$62,$F9,$8F,$F6,$D5,$7F,$46,$6B,$FC,$B1,$BE,$5F,$F9,$FD
	.dc.b	$FC,$94,$6F,$C3,$E4,$47,$97,$87,$FE,$6E,$FA,$B5,$52,$4F,$F5,$48
	.dc.b	$FF,$42,$3F,$E7,$D5,$DB,$C4,$91,$BE,$2E,$9E,$F4,$8E,$4F,$F9,$F3
	.dc.b	$EA,$FF,$D7,$59,$FF,$A5,$57,$CC,$E4,$95,$FD,$92,$1F,$33,$FE,$84
	.dc.b	$FA,$B7,$D5,$52,$57,$F6,$A8,$FE,$0E,$FF,$6C,$F4,$1E,$3F,$FA,$31
	.dc.b	$B1,$98,$C6,$F9,$F0,$E4,$F1,$78,$78,$7C,$7E,$1A,$4F,$57,$A6,$FF
	.dc.b	$E9,$8F,$AB,$7C,$C5,$25,$7F,$7E,$8F,$7B,$FE,$00,$FA,$BF,$E7,$89
	.dc.b	$A4,$F5,$BF,$D4,$6B,$E4,$E9,$27,$97,$06,$BA,$64,$75,$78,$F4,$C8
	.dc.b	$F0,$35,$FE,$EB,$6F,$93,$EF,$E2,$C9,$F3,$3E,$2F,$16,$47,$A5,$AF
	.dc.b	$2F,$4B,$FD,$73,$F5,$7F,$88,$83,$49,$E7,$E8,$91,$FE,$22,$84,$7F
	.dc.b	$9C,$B8,$BF,$D8,$7F,$2D,$FD,$FA,$BF,$C4,$ED,$7F,$C1,$AD,$ED,$FF
	.dc.b	$70,$3F,$B1,$27,$C5,$F2,$3F,$CF,$BC,$BF,$EC,$EF,$A9,$FE,$65,$7F
	.dc.b	$E1,$66,$DF,$FB,$A5,$BE,$3F,$F9,$FD,$FC,$4C,$EF,$AF,$91,$1F,$F1
	.dc.b	$DC,$7E,$B1,$BD,$7D,$C8,$FF,$90,$3E,$23,$F4,$77,$D7,$D4,$F2,$12
	.dc.b	$79,$ED,$E7,$23,$83,$BF,$89,$FE,$5D,$5F,$56,$DF,$D1,$39,$9F,$F4
	.dc.b	$C0,$CF,$F7,$7A,$FF,$93,$1B,$1A,$35,$3C,$5F,$85,$E2,$BE,$AF,$83
	.dc.b	$63,$85,$7F,$2F,$9C,$9F,$E5,$38,$7D,$1D,$23,$FE,$57,$7F,$CB,$1F
	.dc.b	$02,$DB,$FC,$CB,$3F,$E3,$E6,$C6,$63,$FC,$9A,$64,$BF,$13,$9C,$F5
	.dc.b	$0F,$3F,$FB,$B5,$F4,$AD,$3F,$EB,$B8,$CC,$F4,$E5,$CB,$FE,$BF,$F3
	.dc.b	$EA,$11,$87,$F9,$F5,$7A,$27,$1F,$19,$96,$C7,$B2,$39,$DB,$E5,$73
	.dc.b	$7F,$E6,$88,$C5,$7F,$A1,$B2,$FE,$6E,$DE,$4F,$1C,$62,$BF,$CD,$EA
	.dc.b	$32,$76,$F0,$2D,$7F,$A4,$2D,$1E,$CF,$AB,$FF,$2A,$4F,$FA,$51,$F9
	.dc.b	$97,$F2,$9C,$E1,$FE,$7C,$73,$3F,$EE,$36,$C7,$C5,$B7,$77,$5D,$BD
	.dc.b	$EF,$7F,$1F,$C1,$E7,$5C,$7C,$0E,$1F,$17,$9B,$71,$F0,$3E,$7F,$8C
	.dc.b	$F4,$7A,$56,$BD,$0E,$1F,$57,$A5,$6F,$D0,$FB,$FE,$2F,$E0,$7C,$5D
	.dc.b	$AF,$80,$FD,$EF,$98,$FF,$07,$47,$FA,$D9,$7D,$5E,$7D,$BD,$3B,$DF
	.dc.b	$87,$93,$6F,$83,$F8,$8F,$AB,$FF,$07,$47,$FB,$01,$7D,$5E,$7D,$CD
	.dc.b	$3B,$E5,$E1,$F5,$76,$F8,$3F,$92,$FA,$BF,$C7,$F0,$FF,$AB,$9F,$F5
	.dc.b	$F4,$FF,$B3,$1C,$F4,$FE,$89,$CC,$EF,$8B,$97,$17,$36,$DF,$FB,$53
	.dc.b	$EA,$DE,$87,$E8,$1F,$B7,$19,$8C,$70,$7E,$92,$3F,$C5,$C7,$FB,$70
	.dc.b	$F7,$E1,$FE,$C2,$72,$DD,$BF,$57,$F5,$7F,$AB,$9F,$F7,$1B,$F8,$ED
	.dc.b	$3F,$72,$E6,$BF,$D9,$6E,$67,$F6,$50,$8F,$D9,$C7,$2F,$07,$AF,$85
	.dc.b	$D9,$EF,$D1,$E0,$D6,$D7,$81,$FD,$EF,$88,$F6,$EF,$C1,$E1,$B5,$FE
	.dc.b	$F8,$9F,$F4,$AB,$7C,$2E,$3E,$B7,$27,$F8,$3F,$D5,$FE,$D7,$7F,$E2
	.dc.b	$6F,$ED,$6B,$FE,$00,$B9,$7D,$6E,$77,$FF,$74,$BF,$ED,$58,$DF,$C2
	.dc.b	$F4,$F7,$7A,$BD,$BD,$5F,$E2,$BE,$5F,$C7,$C7,$BC,$63,$1C,$F8,$BA
	.dc.b	$F1,$6E,$5F,$C6,$EF,$F9,$3E,$1F,$E3,$08,$C1,$F3,$BE,$AD,$EE,$FF
	.dc.b	$22,$5F,$45,$63,$0F,$F2,$6F,$AB,$FC,$ED,$72,$CF,$F9,$44,$3E,$1B
	.dc.b	$5E,$1E,$D9,$5B,$A5,$7F,$39,$8C,$3C,$5B,$E1,$F7,$3F,$9A,$FD,$5B
	.dc.b	$C2,$C4,$F3,$8C,$3C,$BE,$0F,$CD,$C9,$E7,$6C,$76,$8E,$7E,$5E,$32
	.dc.b	$70,$FF,$4C,$C1,$3F,$D1,$3E,$AF,$E8,$61,$F9,$47,$F2,$D7,$FD,$1D
	.dc.b	$CB,$78,$B8,$1E,$8F,$98,$B5,$E7,$7A,$78,$1D,$7D,$4B,$3F,$17,$0F
	.dc.b	$DD,$E7,$FD,$79,$B1,$E1,$F6,$78,$3B,$7C,$C7,$C6,$FA,$9B,$1F,$15
	.dc.b	$F5,$7C,$EF,$AF,$D3,$B3,$DD,$C3,$F2,$FE,$AE,$ED,$F1,$23,$F6,$7C
	.dc.b	$09,$F8,$9E,$A7,$C5,$F1,$ED,$7A,$5C,$3F,$7F,$EA,$FF,$C0,$0E,$4C
	.dc.b	$1F,$37,$8F,$E7,$71,$58,$F8,$4E,$1F,$89,$AD,$CF,$C8,$7D,$17,$EB
	.dc.b	$1F,$DE,$BE,$AF,$F7,$D3,$C0,$FC,$83,$89,$E2,$3E,$AF,$F4,$33,$F9
	.dc.b	$57,$F4,$2D,$EE,$FF,$02,$C6,$27,$8B,$78,$FF,$11,$C5,$AE,$EF,$CB
	.dc.b	$BC,$3C,$9E,$3E,$B5,$CB,$B3,$87,$F4,$9D,$5E,$5C,$3F,$C3,$B8,$FF
	.dc.b	$56,$CB,$CF,$39,$AF,$E4,$B2,$F9,$BF,$2F,$CD,$F9,$9B,$3E,$B8,$F4
	.dc.b	$FC,$66,$30,$F4,$B8,$9F,$0E,$76,$7B,$BF,$03,$CA,$F8,$3F,$15,$6B
	.dc.b	$E3,$38,$7F,$79,$F5,$6F,$A9,$C6,$1F,$3A,$31,$3F,$37,$DB,$FF,$17
	.dc.b	$F1,$3D,$DF,$9B,$18,$64,$FF,$07,$E8,$7C,$F6,$31,$FB,$31,$89,$FE
	.dc.b	$19,$F5,$4F,$9C,$FF,$22,$D7,$F4,$A8,$C4,$FF,$98,$61,$18,$7F,$CC
	.dc.b	$8D,$7F,$CD,$7F,$57,$FE,$39,$C7,$83,$63,$FC,$D4,$BE,$69,$E4,$6B
	.dc.b	$E1,$BD,$95,$D3,$DD,$F0,$3E,$06,$E3,$FE,$2A,$8F,$23,$5F,$1E,$DA
	.dc.b	$23,$A7,$9F,$D5,$FF,$A7,$A3,$FD,$02,$D7,$FB,$4B,$5F,$F2,$E3,$53
	.dc.b	$FE,$A2,$73,$F3,$DF,$DC,$3B,$3F,$CD,$B3,$99,$7D,$5B,$E4,$F5,$C1
	.dc.b	$FD,$DB,$EA,$3F,$D6,$CF,$83,$D2,$CD,$F8,$1A,$A7,$6F,$F9,$BD,$AF
	.dc.b	$92,$37,$87,$F9,$F6,$3F,$A4,$4E,$8D,$C6,$63,$F7,$6D,$C7,$FD,$0B
	.dc.b	$3B,$F8,$9E,$77,$2F,$9C,$79,$27,$9C,$E6,$7F,$82,$C7,$17,$FA,$C5
	.dc.b	$F5,$C3,$DD,$1C,$88,$78,$AB,$96,$1E,$2C,$3B,$18,$AE,$35,$DA,$32
	.dc.b	$F6,$AD,$3F,$16,$32,$F5,$34,$78,$2B,$B3,$83,$47,$53,$55,$E0,$8F
	.dc.b	$BB,$5D,$1E,$33,$C1,$CB,$C0,$F1,$1D,$5C,$BD,$48,$E0,$93,$F8,$5A
	.dc.b	$EC,$F8,$49,$3D,$0E,$2E,$1F,$C1,$1E,$13,$F1,$11,$F1,$1C,$9C,$0E
	.dc.b	$47,$FB,$8B,$04,$7E,$C1,$B1,$FC,$05,$AF,$FB,$8D,$AF,$18,$E0,$47
	.dc.b	$83,$EA,$7D,$15,$F8,$EF,$81,$C4,$79,$8F,$45,$7F,$B2,$9B,$5A,$8C
	.dc.b	$47,$A4,$E0,$57,$35,$F5,$09,$7F,$2E,$7F,$47,$7C,$EF,$A3,$63,$C2
	.dc.b	$D7,$A3,$6B,$9C,$7E,$C6,$31,$5D,$63,$F4,$31,$87,$DF,$1F,$BF,$8C
	.dc.b	$47,$A2,$7E,$13,$EE,$3C,$25,$3C,$25,$7C,$2F,$F0,$38,$E7,$7F,$BD
	.dc.b	$6D,$FE,$D5,$B3,$F0,$0F,$F2,$57,$C9,$3C,$BF,$E1,$AD,$8C,$C7,$EF
	.dc.b	$7D,$39,$18,$21,$FB,$E7,$F8,$17,$A7,$CB,$E0,$A4,$FF,$10,$8C,$BF
	.dc.b	$89,$7F,$18,$47,$FA,$67,$D4,$BD,$4F,$E7,$ED,$8C,$C6,$0F,$CA,$79
	.dc.b	$7F,$C9,$27,$53,$FE,$14,$6E,$DF,$C6,$A7,$F9,$4F,$D3,$FF,$62,$AF
	.dc.b	$F3,$53,$5F,$F0,$DB,$65,$C1,$FE,$AA,$8E,$6D,$C3,$99,$2B,$FC,$E9
	.dc.b	$F5,$3F,$CC,$A9,$F4,$5E,$02,$4F,$83,$D4,$F1,$91,$C1,$FE,$7D,$FA
	.dc.b	$97,$9F,$E8,$F6,$E4,$FA,$1F,$5F,$8B,$23,$D6,$D7,$D7,$EA,$FC,$02
	.dc.b	$3F,$01,$FE,$97,$F5,$3F,$DC,$23,$FD,$19,$F4,$FF,$01,$BB,$FF,$50
	.dc.b	$8C,$47,$F9,$01,$FF,$4F,$BC,$3D,$5D,$11,$FF,$4B,$7D,$4F,$FA,$3D
	.dc.b	$EB,$F1,$5C,$C8,$F8,$3D,$2E,$C4,$70,$7C,$CF,$37,$C5,$E6,$C8,$F3
	.dc.b	$BE,$5F,$56,$47,$A9,$AF,$9F,$D5,$F5,$11,$F7,$BF,$DB,$3E,$A5,$F0
	.dc.b	$FE,$FF,$8A,$95,$F1,$1F,$85,$EB,$C8,$F8,$ED,$7F,$CC,$7D,$5F,$80
	.dc.b	$7E,$5F,$9D,$F2,$48,$FC,$EF,$37,$E6,$11,$C1,$FE,$BD,$FA,$9F,$F1
	.dc.b	$93,$FE,$BF,$F4,$BF,$C8,$8D,$F5,$FF,$C0,$71,$88,$FF,$39,$BF,$E0
	.dc.b	$F2,$7A,$1A,$37,$F3,$D2,$7A,$35,$FF,$27,$7C,$BF,$4F,$9D,$91,$E1
	.dc.b	$7B,$FC,$E4,$8F,$23,$5F,$F2,$B7,$D4,$F9,$3E,$2A,$47,$A5,$F3,$FA
	.dc.b	$72,$3D,$8D,$7F,$CA,$7E,$37,$B3,$E2,$64,$7C,$17,$DF,$E2,$C8,$F8
	.dc.b	$8D,$7E,$0F,$C7,$FC,$42,$3F,$25,$FF,$14,$FD,$59,$FF,$15,$FD,$53
	.dc.b	$BE,$92,$BF,$DD,$D0,$F8,$5D,$FC,$B4,$9C,$BC,$BA,$3F,$C6,$BF,$54
	.dc.b	$FA,$0A,$4A,$FF,$7C,$43,$E6,$37,$C3,$27,$2F,$2E,$1F,$F1,$FF,$D5
	.dc.b	$7F,$5A,$F2,$95,$23,$0D,$F6,$F7,$EC,$F0,$F6,$A4,$F0,$F5,$BF,$73
	.dc.b	$7F,$CF,$63,$FD,$72,$BD,$AF,$F2,$6F,$D1,$BE,$1E,$92,$B9,$67,$E2
	.dc.b	$53,$FE,$79,$72,$79,$97,$8F,$D0,$91,$A3,$8F,$8F,$FE,$5C,$FA,$AE
	.dc.b	$71,$96,$BF,$EC,$E6,$E7,$FC,$49,$78,$3C,$7C,$09,$5C,$5E,$9C,$BD
	.dc.b	$7C,$99,$3D,$5F,$F3,$5F,$D5,$7A,$E3,$FD,$B6,$BE,$67,$24,$8F,$44
	.dc.b	$3E,$67,$FC,$E7,$F5,$4F,$A9,$A4,$AF,$74,$7C,$27,$E0,$7A,$8F,$27
	.dc.b	$FC,$F6,$D8,$CC,$63,$8B,$3E,$34,$9E,$2F,$0F,$1B,$8F,$C3,$C9,$EA
	.dc.b	$F4,$DB,$FD,$0F,$F5,$4F,$98,$A4,$AF,$A5,$1E,$F7,$EA,$7D,$57,$F3
	.dc.b	$EE,$6F,$E7,$ED,$E9,$FC,$03,$67,$F2,$6F,$F4,$25,$3C,$3E,$4F,$FD
	.dc.b	$2F,$26,$98,$E1,$CF,$54,$BF,$0B,$C3,$AA,$47,$91,$AF,$6C,$7C,$7F
	.dc.b	$27,$DB,$C7,$93,$EA,$7C,$3E,$3C,$8F,$6B,$5E,$47,$AB,$D4,$FF,$56
	.dc.b	$FD,$57,$E3,$23,$F2,$CD,$76,$1C,$6D,$FD,$91,$7A,$3C,$89,$3C,$9F
	.dc.b	$E2,$3C,$7F,$EB,$AF,$87,$FF,$5B,$93,$E2,$F1,$E4,$4B,$FD,$34,$FE
	.dc.b	$F9,$37,$D1,$FE,$AF,$1F,$AB,$FA,$9F,$F6,$71,$3D,$5F,$F6,$3B,$F8
	.dc.b	$D8,$F4,$B7,$4F,$EF,$6D,$E1,$FD,$E3,$67,$F8,$7B,$FB,$08,$CF,$67
	.dc.b	$42,$4F,$F9,$28,$EF,$E2,$C7,$C8,$F8,$B6,$91,$E7,$6B,$E4,$F1,$7F
	.dc.b	$12,$BE,$AB,$DE,$F8,$A3,$2F,$ED,$F1,$46,$23,$D2,$D7,$86,$5F,$BD
	.dc.b	$7D,$7E,$33,$EA,$B1,$0B,$FE,$71,$18,$6C,$71,$6B,$C3,$E8,$F8,$BF
	.dc.b	$CD,$1A,$9A,$6B,$E5,$7F,$C5,$98,$7F,$8A,$A7,$FD,$5B,$B7,$F2,$35
	.dc.b	$3A,$8F,$E2,$9C,$76,$7F,$98,$35,$7A,$0F,$6B,$11,$D0,$FC,$BC,$66
	.dc.b	$BB,$72,$EB,$D1,$DB,$63,$C5,$1C,$BE,$3F,$17,$FA,$00,$E6,$7F,$9D
	.dc.b	$5B,$E5,$71,$5E,$07,$CD,$A7,$F4,$6A,$FE,$80,$FF,$38,$6F,$2E,$B9
	.dc.b	$64,$74,$EF,$B8,$7A,$9F,$81,$F5,$5F,$1A,$7F,$A5,$3E,$FC,$3F,$CA
	.dc.b	$AE,$6B,$F6,$EE,$67,$FA,$C3,$63,$FB,$5E,$4E,$EE,$BC,$1F,$36,$31
	.dc.b	$7D,$BC,$6F,$0F,$9F,$E2,$6C,$78,$A3,$CB,$E7,$63,$1F,$0B,$E8,$F8
	.dc.b	$EF,$4F,$AB,$6F,$D2,$E1,$F5,$FA,$96,$FD,$2F,$C0,$F8,$DF,$83,$F1
	.dc.b	$B6,$BE,$0B,$F8,$3F,$98,$FF,$0B,$47,$F6,$D5,$F5,$59,$F7,$74,$EF
	.dc.b	$87,$87,$93,$6F,$83,$F8,$DF,$AA,$FF,$0B,$47,$F7,$75,$F4,$99,$F7
	.dc.b	$B4,$EF,$93,$87,$D5,$5B,$E0,$FE,$5F,$EA,$9D,$F7,$BF,$0D,$AF,$F0
	.dc.b	$24,$FF,$84,$DB,$C3,$C7,$8B,$93,$FA,$1F,$AA,$7A,$17,$83,$0D,$BF
	.dc.b	$F0,$71,$FE,$33,$2E,$DF,$A9,$71,$CB,$FC,$F1,$F7,$BF,$CF,$BE,$AB
	.dc.b	$FC,$5F,$1F,$E1,$77,$F1,$9D,$DF,$EA,$27,$33,$BF,$17,$F6,$90,$F1
	.dc.b	$2E,$F8,$6E,$1C,$FA,$B7,$BF,$06,$C7,$F8,$97,$C6,$7B,$9F,$BF,$7F
	.dc.b	$CA,$D1,$98,$C7,$C2,$E7,$D5,$D7,$8B,$6B,$93,$FC,$1F,$EA,$BF,$65
	.dc.b	$BF,$F5,$A7,$F6,$4D,$F2,$35,$E4,$6F,$DC,$E1,$F0,$F7,$1F,$F1,$C6
	.dc.b	$1F,$E6,$96,$FB,$BF,$C4,$F9,$3E,$1F,$2F,$FB,$1A,$BF,$D5,$75,$FE
	.dc.b	$3F,$9C,$BF,$E4,$97,$FC,$5C,$E7,$C7,$E5,$E4,$73,$7F,$8F,$7D,$53
	.dc.b	$E5,$2F,$C2,$E1,$B7,$FE,$4B,$FA,$AF,$CA,$B7,$E2,$3F,$93,$46,$38
	.dc.b	$BF,$BC,$A8,$C9,$AF,$16,$D7,$37,$0F,$F9,$5F,$EA,$5E,$1E,$30,$FF
	.dc.b	$01,$C6,$27,$C5,$F0,$9E,$0E,$4F,$0B,$63,$CD,$1F,$CE,$65,$E7,$66
	.dc.b	$F9,$BF,$CF,$7E,$AB,$E3,$C7,$FA,$01,$FC,$8D,$FF,$9C,$5C,$C1,$CB
	.dc.b	$BE,$2E,$DF,$35,$CD,$B5,$FE,$71,$AF,$D9,$B7,$C9,$F0,$78,$7E,$16
	.dc.b	$46,$FD,$FB,$8F,$F8,$8A,$FF,$C2,$73,$93,$E3,$7D,$47,$A6,$BD,$CF
	.dc.b	$AF,$E7,$BD,$9E,$AD,$AF,$13,$87,$E7,$7D,$55,$C7,$E6,$BF,$0F,$D6
	.dc.b	$F8,$9E,$3D,$AF,$53,$87,$FD,$19,$F5,$47,$17,$FE,$38,$8C,$47,$C1
	.dc.b	$E1,$F9,$DC,$3F,$3D,$B1,$F1,$63,$E9,$69,$8B,$F8,$0E,$27,$FD,$35
	.dc.b	$F5,$5F,$B6,$7E,$64,$7F,$75,$8C,$4F,$F9,$23,$EA,$BF,$35,$5F,$EA
	.dc.b	$17,$FD,$5C,$DE,$BF,$EF,$91,$89,$E1,$D9,$F8,$7E,$1C,$B6,$F9,$79
	.dc.b	$1C,$3D,$8E,$2D,$8E,$D1,$FE,$07,$E9,$F1,$F8,$7F,$17,$1F,$F6,$5E
	.dc.b	$5F,$D2,$1F,$F2,$EC,$EA,$FF,$D7,$D2,$F9,$BF,$2F,$CD,$F9,$9B,$3E
	.dc.b	$B8,$F4,$FC,$66,$30,$F8,$0E,$27,$FD,$81,$87,$F1,$A9,$CC,$63,$3F
	.dc.b	$83,$E5,$FC,$2F,$8B,$B3,$F1,$DC,$3F,$EC,$9F,$AA,$7C,$CE,$30,$FF
	.dc.b	$3A,$C6,$27,$E7,$7B,$BF,$E2,$7E,$17,$BF,$F3,$A3,$0C,$9F,$F6,$A7
	.dc.b	$43,$E7,$B1,$8F,$FA,$02,$31,$3F,$ED,$9F,$AA,$FF,$10,$5F,$F9,$29
	.dc.b	$B5,$FE,$4F,$7F,$C6,$9E,$26,$7E,$13,$C3,$3F,$E5,$3F,$AA,$70,$7F
	.dc.b	$92,$9B,$1F,$E5,$46,$BB,$8D,$69,$23,$4D,$7F,$CB,$DF,$55,$FE,$44
	.dc.b	$7F,$CB,$AD,$73,$E5,$FF,$96,$23,$11,$FE,$E1,$5F,$21,$92,$47,$26
	.dc.b	$BF,$EA,$28,$DB,$FE,$FB,$6B,$9F,$4C,$FF,$BF,$FE,$A7,$FD,$34,$3E
	.dc.b	$1F,$31,$FE,$AC,$8C,$3B,$F7,$73,$DD,$F9,$78,$4F,$2A,$CF,$17,$28
	.dc.b	$E3,$3C,$EF,$2F,$8A,$38,$3E,$3A,$CB,$D2,$6C,$EA,$3C,$93,$89,$91
	.dc.b	$C4,$87,$73,$D0,$62,$FD,$15,$E4,$6C,$F2,$9E,$A6,$EF,$D5,$19,$9E
	.dc.b	$37,$E4,$31,$3A,$AE,$07,$B6,$32,$FB,$EB,$83,$A3,$DC,$79,$8F,$09
	.dc.b	$2B,$CA,$E5,$F0,$BC,$6F,$EC,$CF,$48,$E1,$7E,$37,$E1,$18,$3C,$87
	.dc.b	$37,$55,$D2,$34,$79,$0E,$2E,$49,$1C,$09,$1D,$0E,$35,$E4,$7C,$D7
	.dc.b	$F0,$9E,$87,$E3,$8A,$79,$CC,$47,$F8,$DD,$FF,$61,$B5,$FF,$64,$35
	.dc.b	$DD,$FF,$B0,$23,$11,$E4,$3D,$57,$F2,$93,$F9,$9F,$91,$75,$F9,$6B
	.dc.b	$D1,$5A,$C7,$F4,$75,$A8,$F9,$07,$DB,$28,$74,$FC,$88,$F1,$4F,$A6
	.dc.b	$FC,$17,$DF,$0F,$89,$87,$09,$F3,$12,$FD,$51,$97,$9C,$F0,$AE,$58
	.dc.b	$78,$AB,$4F,$94,$F1,$16,$FC,$2F,$A3,$0F,$7C,$62,$36,$7C,$8C,$3C
	.dc.b	$37,$DA,$33,$7D,$9D,$1D,$67,$6F,$5C,$3C,$CD,$3C,$D5,$CE,$FD,$D1
	.dc.b	$FC,$52,$FA,$61,$E4,$C3,$9C,$F9,$23,$0E,$EF,$F8,$2F,$9D,$FC,$23
	.dc.b	$EA,$5E,$4D,$33,$5F,$C7,$E7,$F0,$C4,$7F,$1A,$FE,$27,$CF,$91,$F9
	.dc.b	$06,$BF,$90,$7F,$89,$91,$FE,$57,$9F,$F2,$3B,$FF,$29,$C7,$F2,$DF
	.dc.b	$17,$F8,$DF,$D4,$BC,$FA,$7F,$1C,$89,$3F,$9E,$7F,$34,$47,$F4,$C7
	.dc.b	$F3,$8F,$A9,$FE,$C4,$F8,$36,$FE,$6F,$18,$8F,$CC,$1B,$EE,$FE,$57
	.dc.b	$1E,$18,$E0,$73,$BC,$91,$C8,$FA,$3F,$E4,$D7,$DB,$E8,$79,$92,$77
	.dc.b	$6F,$43,$B9,$23,$2F,$F3,$CF,$A9,$FC,$44,$FC,$69,$F3,$87,$1B,$7F
	.dc.b	$4D,$6E,$1E,$EF,$0B,$D6,$4A,$F0,$BE,$EF,$42,$47,$91,$AF,$6F,$A9
	.dc.b	$78,$5F,$D3,$BE,$A7,$FA,$D6,$5F,$D6,$DB,$8F,$F6,$26,$DF,$F6,$87
	.dc.b	$FD,$40,$FC,$66,$BE,$A6,$CF,$3F,$43,$FA,$DF,$D4,$FA,$AB,$97,$C2
	.dc.b	$F8,$48,$F3,$F3,$7C,$84,$70,$7D,$4E,$2F,$BD,$87,$9F,$EA,$7D,$44
	.dc.b	$9F,$27,$A3,$ED,$23,$83,$D1,$E8,$FC,$12,$38,$3F,$03,$F5,$2F,$13
	.dc.b	$E3,$7C,$3C,$9F,$80,$FC,$9F,$2A,$47,$C3,$6B,$F7,$33,$74,$FC,$FF
	.dc.b	$8D,$93,$F2,$5F,$A5,$E5,$C8,$F9,$8D,$7F,$0B,$F5,$2D,$BF,$C0,$BF
	.dc.b	$2D,$FE,$4E,$9F,$CB,$B5,$FC,$5B,$78,$3F,$C2,$CB,$E7,$78,$92,$78
	.dc.b	$BD,$3E,$8A,$47,$46,$BF,$E5,$AF,$8B,$F0,$FC,$F4,$8F,$0B,$E4,$F9
	.dc.b	$D9,$1E,$46,$BF,$E6,$2F,$A9,$F4,$FC,$44,$8F,$4B,$EC,$F5,$24,$7B
	.dc.b	$1A,$FF,$97,$F8,$7E,$0F,$C3,$C8,$F8,$2F,$C4,$F3,$E4,$7C,$46,$BF
	.dc.b	$13,$E3,$BE,$49,$1F,$92,$FF,$8C,$3D,$23,$FE,$32,$FA,$97,$75,$25
	.dc.b	$7F,$06,$87,$C1,$6F,$E5,$64,$E5,$E5,$D3,$FE,$3C,$FA,$97,$D0,$52
	.dc.b	$57,$F1,$08,$7C,$BE,$F8,$64,$E5,$E5,$C3,$FE,$46,$FA,$9F,$ED,$9E
	.dc.b	$4F,$F9,$61,$B5,$FE,$5E,$6A,$5E,$BF,$1F,$62,$4F,$F2,$09,$FE,$92
	.dc.b	$DF,$97,$FF,$29,$FC,$B3,$E1,$29,$27,$F9,$3C,$3B,$F7,$E3,$49,$CB
	.dc.b	$CB,$C6,$FF,$2E,$FD,$4A,$E9,$2B,$F9,$8C,$3C,$0D,$F7,$C9,$CB,$CB
	.dc.b	$93,$9F,$91,$27,$8B,$FE,$6A,$FA,$9F,$F1,$77,$8B,$FE,$77,$5F,$2F
	.dc.b	$92,$57,$F4,$08,$7C,$BF,$F9,$CB,$EA,$5F,$4F,$49,$5F,$D1,$A1,$C1
	.dc.b	$FE,$78,$F4,$5E,$4D,$25,$7F,$4C,$8F,$80,$DF,$1B,$81,$1D,$78,$DC
	.dc.b	$FE,$1A,$47,$17,$B6,$FF,$E8,$6F,$A9,$7C,$AD,$25,$7F,$5A,$8F,$7B
	.dc.b	$FD,$73,$EA,$7F,$9D,$A6,$93,$D7,$FF,$48,$AF,$91,$A4,$9E,$5C,$1A
	.dc.b	$DE,$47,$57,$8F,$54,$8F,$03,$5F,$EB,$F6,$F9,$1E,$FE,$44,$9F,$33
	.dc.b	$E2,$F2,$24,$7A,$5A,$F2,$F5,$3F,$D4,$FF,$53,$F8,$78,$FE,$F2,$D7
	.dc.b	$FB,$6B,$7A,$BF,$D7,$0B,$D2,$E2,$49,$E2,$FF,$03,$E3,$FF,$59,$FC
	.dc.b	$2F,$F8,$0A,$7F,$D9,$2D,$4D,$3F,$5E,$4F,$8A,$E2,$7F,$0D,$E5,$FF
	.dc.b	$5E,$FD,$3F,$F3,$3A,$FF,$07,$B5,$FF,$01,$B7,$2F,$F6,$73,$F8,$C8
	.dc.b	$DF,$57,$22,$3C,$BC,$3E,$A4,$70,$7F,$87,$3E,$27,$F4,$7E,$5F,$FB
	.dc.b	$7A,$FE,$23,$7C,$4F,$29,$E6,$DA,$5F,$99,$AF,$9B,$C4,$FF,$45,$AF
	.dc.b	$A9,$F8,$6F,$FB,$BB,$04,$EA,$3F,$D4,$11,$8D,$3F,$C7,$B0,$8C,$3F
	.dc.b	$E4,$06,$BC,$3C,$57,$D4,$F0,$9F,$E5,$6D,$8C,$CF,$16,$BD,$BE,$87
	.dc.b	$BA,$0E,$1A,$69,$DF,$C1,$B5,$FF,$22,$CB,$E3,$F0,$FF,$8E,$8F,$F1
	.dc.b	$6C,$6A,$33,$96,$A7,$C1,$18,$9F,$14,$3D,$1F,$E5,$51,$C4,$FF,$29
	.dc.b	$B8,$8E,$DC,$9C,$5E,$3E,$8D,$7D,$91,$5F,$E7,$F5,$E8,$9C,$7C,$76
	.dc.b	$5B,$1E,$C8,$E5,$6F,$92,$CD,$FF,$95,$A3,$15,$FE,$66,$CB,$F9,$D3
	.dc.b	$78,$FC,$51,$8A,$FF,$2E,$BF,$CE,$E7,$47,$B3,$C0,$DD,$78,$1F,$57
	.dc.b	$D4,$FC,$AB,$FF,$32,$BF,$3E,$BF,$CE,$AE,$6F,$E6,$39,$9D,$EB,$EA
	.dc.b	$EE,$E6,$DB,$E3,$E3,$F6,$F3,$BB,$B6,$3A,$C7,$8B,$CD,$C6,$1E,$27
	.dc.b	$CB,$F1,$BE,$6F,$56,$DF,$91,$C3,$E6,$F5,$2D,$FA,$9F,$6F,$CF,$FB
	.dc.b	$BE,$32,$D7,$B1,$FB,$58,$7F,$A9,$E3,$FD,$26,$DB,$FF,$43,$BF,$D0
	.dc.b	$23,$3F,$1F,$C2,$F8,$DD,$9E,$2F,$E1,$7F,$17,$FA,$58,$7F,$A4,$1F
	.dc.b	$E4,$53,$FE,$A0,$73,$C3,$F9,$77,$33,$FE,$96,$FB,$A7,$3F,$E3,$DF
	.dc.b	$E4,$37,$FE,$A8,$72,$DF,$1B,$5F,$07,$23,$5A,$DC,$3F,$D4,$3E,$73
	.dc.b	$E6,$3F,$34,$DA,$FF,$59,$AF,$AF,$CC,$FE,$7F,$EE,$9E,$BD,$F5,$F0
	.dc.b	$DC,$3F,$D7,$7C,$AF,$07,$F4,$CD,$8C,$F8,$38,$9F,$D5,$7E,$2A,$F8
	.dc.b	$EA,$63,$D7,$B7,$FE,$C6,$FB,$B7,$CF,$7E,$C5,$B1,$9F,$8C,$C4,$E7
	.dc.b	$E1,$75,$F5,$B6,$B9,$3F,$B9,$FC,$21,$DD,$7B,$B0,$DA,$FF,$6A,$FD
	.dc.b	$E7,$EC,$BA,$3F,$D9,$2B,$E0,$7F,$81,$39,$F0,$3F,$88,$39,$AE,$3B
	.dc.b	$E8,$DF,$BF,$63,$FD,$C3,$F8,$9B,$DB,$FE,$18,$D8,$CF,$B7,$89,$FE
	.dc.b	$21,$F7,$BF,$B1,$CB,$F8,$F3,$6B,$F8,$EB,$F3,$F4,$FF,$6F,$B9,$7F
	.dc.b	$99,$3F,$9B,$73,$E4,$F4,$F2,$BA,$BF,$C7,$7B,$3F,$C5,$F7,$FE,$F6
	.dc.b	$5E,$76,$7C,$2D,$3F,$C8,$FE,$ED,$F0,$77,$E2,$70,$DC,$3F,$DF,$FF
	.dc.b	$59,$F8,$A6,$F8,$7F,$CA,$E3,$1A,$FF,$BD,$94,$64,$FE,$61,$F5,$8E
	.dc.b	$9C,$5F,$F4,$DE,$1F,$E8,$13,$98,$C1,$FD,$03,$EB,$3F,$A9,$38,$C7
	.dc.b	$FA,$A4,$7F,$1D,$8D,$47,$F3,$D8,$C3,$AF,$63,$C3,$F3,$D9,$6D,$73
	.dc.b	$95,$7F,$42,$58,$7F,$53,$8F,$DB,$C6,$27,$C9,$E3,$77,$F3,$E4,$73
	.dc.b	$AF,$DF,$D6,$CF,$3C,$F9,$7C,$EF,$A1,$F6,$9E,$7F,$4A,$E1,$C5,$C3
	.dc.b	$CF,$A2,$E1,$CC,$F3,$BE,$F3,$E6,$FD,$6D,$9F,$66,$E8,$F8,$9E,$87
	.dc.b	$C5,$F3,$6D,$79,$DC,$3F,$0F,$EB,$7F,$BD,$39,$30,$7C,$DE,$3F,$9D
	.dc.b	$C7,$63,$E0,$38,$7E,$17,$D6,$DD,$3D,$1C,$28,$FC,$87,$5F,$F7,$3C
	.dc.b	$5F,$13,$8C,$3C,$4F,$F7,$7F,$AC,$FC,$AC,$7D,$77,$FD,$5C,$DE,$AD
	.dc.b	$75,$FE,$3B,$96,$4E,$E2,$E9,$F2,$78,$9F,$6D,$D7,$E6,$7F,$2A,$DB
	.dc.b	$ED,$F3,$1E,$13,$17,$F4,$73,$FF,$6D,$CF,$8F,$D2,$F3,$1E,$7F,$A1
	.dc.b	$B1,$E7,$F2,$BF,$C3,$97,$FE,$1A,$BF,$F4,$44,$FF,$8F,$1B,$1F,$E1
	.dc.b	$E3,$FC,$60,$E1,$C7,$D0,$7F,$01,$6F,$57,$47,$FC,$6D,$19,$7F,$C4
	.dc.b	$ED,$79,$F8,$9B,$55,$FC,$12,$FC,$53,$98,$C4,$F8,$CD,$FC,$8E,$49
	.dc.b	$5E,$97,$A7,$D0,$FF,$02,$9F,$53,$5F,$F1,$CB,$5F,$F2,$63,$5F,$F7
	.dc.b	$47,$B5,$E5,$E2,$FE,$B9,$1E,$16,$BF,$15,$AF,$F9,$13,$AF,$FC,$00
	.dc.b	$FC,$96,$B9,$EF,$F8,$11,$88,$FF,$2E,$7D,$67,$F8,$E6,$FF,$CD,$AD
	.dc.b	$8F,$80,$FF,$BD,$FC,$5C,$FC,$3F,$F7,$19,$F8,$9F,$4C,$FA,$8F,$F2
	.dc.b	$AF,$91,$FE,$59,$CB,$FD,$9F,$5F,$C3,$5A,$FE,$E6,$B8,$3F,$C4,$5A
	.dc.b	$FE,$F6,$37,$3C,$F8,$FA,$F0,$7F,$98,$AF,$73,$FE,$67,$6C,$7C,$39
	.dc.b	$DD,$FF,$9E,$6F,$9C,$F1,$C3,$32,$BE,$2E,$EF,$37,$C2,$3F,$D0,$53
	.dc.b	$A7,$85,$6A,$A4,$7A,$E3,$AC,$EE,$B9,$4F,$F4,$1A,$CD,$48,$F8,$D1
	.dc.b	$C9,$F8,$D1,$97,$E2,$46,$70,$F8,$13,$97,$E3,$4F,$23,$C4,$43,$89
	.dc.b	$93,$A9,$97,$8B,$A3,$24,$72,$4C,$3E,$54,$71,$7E,$83,$D6,$3C,$AE
	.dc.b	$4F,$98,$66,$38,$BD,$5E,$2F,$86,$38,$B9,$38,$99,$8E,$31,$A3,$C9
	.dc.b	$1A,$71,$1A,$3A,$BA,$8F,$B2,$79,$4F,$A8,$79,$4F,$BF,$3F,$36,$78
	.dc.b	$92,$32,$71,$BF,$BA,$7D,$EC,$3F,$0C,$7E,$07,$0F,$B1,$5E,$3A,$E2
	.dc.b	$7A,$CF,$43,$97,$CA,$64,$F2,$9F,$0C,$C9,$FC,$9E,$7F,$52,$DB,$FF
	.dc.b	$6C,$C6,$23,$E7,$99,$3F,$DD,$6E,$5F,$19,$93,$B3,$EA,$D7,$E1,$47
	.dc.b	$A3,$1F,$9A,$FE,$92,$32,$FA,$23,$2F,$EB,$63,$33,$F1,$2B,$F4,$B1
	.dc.b	$87,$2D,$9F,$A9,$28,$C4,$79,$1B,$5D,$A7,$53,$C2,$B9,$B4,$E7,$5C
	.dc.b	$2F,$E7,$57,$09,$F7,$D7,$06,$9C,$0E,$47,$84,$B7,$E1,$BE,$91,$F1
	.dc.b	$6F,$F8,$B4,$62,$3E,$24,$61,$E2,$F8,$A3,$57,$F3,$E7,$27,$33,$E7
	.dc.b	$18,$39,$98,$3C,$E7,$94,$B5,$E5,$78,$47,$57,$F9,$97,$C0,$7E,$18
	.dc.b	$8F,$E1,$1A,$E8,$FE,$6C,$FF,$1C,$F1,$7F,$8E,$FA,$DF,$8E,$24,$FE
	.dc.b	$31,$FC,$8E,$92,$3F,$1E,$D7,$FB,$67,$D3,$7F,$73,$4D,$26,$7F,$DC
	.dc.b	$60,$69,$FE,$73,$06,$98,$F0,$3F,$9E,$FD,$37,$FA,$9A,$BF,$CB,$D5
	.dc.b	$FE,$7E,$6C,$EB,$0D,$F9,$32,$7F,$CC,$97,$FE,$49,$78,$FB,$1E,$0F
	.dc.b	$07,$C2,$C8,$EB,$1E,$4E,$EF,$2F,$D2,$C9,$F0,$C6,$5A,$43,$FC,$B9
	.dc.b	$9F,$FB,$2D,$F3,$F8,$FF,$D5,$BB,$DF,$0F,$4F,$83,$12,$76,$43,$9F
	.dc.b	$81,$F1,$88,$FC,$27,$A7,$1F,$CA,$F8,$99,$3F,$19,$F9,$BE,$44,$8F
	.dc.b	$92,$D7,$FC,$E5,$F4,$CD,$7F,$B5,$FC,$FF,$FB,$7A,$7F,$D1,$8D,$7F
	.dc.b	$CF,$0D,$E9,$FB,$11,$F7,$21,$F0,$5C,$89,$1C,$9E,$BF,$31,$FD,$E3
	.dc.b	$EB,$5F,$69,$49,$5E,$0F,$98,$F1,$91,$C1,$E4,$F9,$7F,$39,$1C,$1F
	.dc.b	$EA,$3A,$7D,$47,$B3,$E9,$24,$FB,$1F,$7F,$CA,$C8,$F7,$B5,$FF,$0D
	.dc.b	$71,$7C,$4F,$7E,$47,$C4,$7E,$3F,$BB,$23,$E3,$B5,$F8,$FF,$4F,$FE
	.dc.b	$16,$FC,$77,$FC,$C9,$C5,$F2,$25,$18,$F0,$FF,$4B,$18,$38,$79,$71
	.dc.b	$FF,$5F,$4E,$DF,$D5,$B5,$E7,$1F,$E2,$4F,$9D,$FF,$6B,$47,$08,$E4
	.dc.b	$FF,$AC,$63,$2F,$FA,$CF,$F0,$27,$17,$F8,$B1,$B1,$FE,$CD,$6B,$BE
	.dc.b	$2F,$CE,$79,$1F,$9D,$F7,$A4,$FF,$B4,$FC,$04,$F8,$FD,$EF,$27,$BB
	.dc.b	$2B,$FD,$B1,$9A,$7C,$1C,$9E,$1E,$89,$5E,$18,$FD,$57,$CB,$48,$EC
	.dc.b	$FE,$B7,$CA,$F8,$1E,$54,$9E,$AF,$F9,$1B,$F6,$AF,$A1,$CB,$F4,$78
	.dc.b	$5F,$11,$EF,$25,$FB,$FE,$27,$F7,$1E,$7F,$EE,$76,$93,$ED,$8E,$9E
	.dc.b	$7F,$5F,$5E,$4F,$FB,$CF,$3F,$E2,$CF,$FB,$7F,$57,$C2,$FF,$96,$17
	.dc.b	$C2,$D3,$11,$FC,$1F,$92,$57,$CB,$7F,$85,$70,$7A,$FC,$69,$3D,$DF
	.dc.b	$77,$AF,$D7,$D7,$ED,$F1,$12,$BF,$98,$F2,$7E,$47,$4F,$F2,$DF,$97
	.dc.b	$F1,$3C,$1F,$C8,$B6,$BF,$9B,$B5,$3E,$CA,$FC,$D6,$EF,$53,$E7,$12
	.dc.b	$BE,$17,$A9,$EF,$E9,$91,$E2,$C7,$F3,$F9,$27,$E2,$3F,$E7,$CC,$9C
	.dc.b	$77,$79,$87,$13,$CB,$1F,$77,$E0,$3E,$F7,$C3,$49,$FA,$51,$FE,$86
	.dc.b	$FA,$37,$C0,$D2,$4F,$31,$0F,$8C,$E1,$EB,$49,$D4,$A3,$A7,$AD,$27
	.dc.b	$2D,$5C,$7F,$BF,$A7,$C0,$F8,$FE,$8E,$57,$47,$CD,$EC,$48,$EE,$D7
	.dc.b	$9F,$1F,$F0,$6B,$7B,$1F,$EA,$2F,$A3,$7C,$ED,$25,$FB,$63,$F7,$39
	.dc.b	$FF,$11,$6E,$1F,$C4,$1B,$3F,$DD,$9A,$BE,$AF,$F4,$EF,$C7,$F8,$A7
	.dc.b	$CE,$FF,$AC,$9B,$19,$8C,$7C,$CC,$8E,$0D,$7D,$67,$F8,$12,$9E,$47
	.dc.b	$FA,$F3,$E8,$BF,$A5,$CF,$D5,$6B,$C0,$38,$CF,$F9,$62,$F7,$79,$12
	.dc.b	$79,$3D,$7E,$92,$47,$56,$BF,$C9,$4F,$16,$D1,$F1,$3E,$5F,$56,$47
	.dc.b	$95,$AF,$96,$DE,$AF,$E0,$A1,$F4,$8D,$FF,$9F,$D6,$B3,$FF,$13,$36
	.dc.b	$BF,$A1,$9F,$E2,$87,$0E,$1A,$F2,$FA,$37,$D2,$72,$6C,$72,$3E,$3B
	.dc.b	$87,$93,$5D,$7D,$27,$F8,$DD,$FF,$1B,$B5,$34,$D7,$B5,$FE,$02,$C3
	.dc.b	$FA,$E4,$FE,$13,$97,$F5,$8A,$75,$1E,$CE,$4B,$3F,$D8,$1A,$BD,$57
	.dc.b	$87,$88,$E8,$7E,$06,$33,$5D,$B9,$B5,$E7,$ED,$B1,$E2,$8E,$5E,$86
	.dc.b	$2F,$F1,$47,$F3,$29,$F3,$C7,$8A,$DE,$77,$09,$FE,$0D,$18,$7F,$84
	.dc.b	$68,$E6,$F4,$BF,$B2,$8C,$C6,$39,$A4,$72,$F0,$2E,$5E,$B7,$E0,$FD
	.dc.b	$27,$C7,$9F,$E2,$8F,$C0,$C3,$FC,$10,$E6,$BF,$44,$E6,$7F,$C1,$2D
	.dc.b	$8F,$F3,$7E,$2E,$FE,$BC,$1E,$D7,$BF,$9F,$E0,$F3,$EE,$3E,$07,$0F
	.dc.b	$8B,$E1,$AE,$3E,$07,$CF,$ED,$7A,$3D,$CB,$5E,$87,$0F,$AB,$CF,$B7
	.dc.b	$E8,$7D,$FE,$7F,$C0,$F5,$AD,$7C,$07,$F0,$78,$7F,$B2,$63,$F9,$B3
	.dc.b	$6F,$F8,$12,$CF,$7F,$17,$C3,$0E,$38,$5C,$38,$BF,$8E,$FA,$4F,$AD
	.dc.b	$8F,$F0,$B7,$FC,$43,$3F,$D0,$DC,$F1,$7F,$8D,$DC,$CF,$F1,$1F,$A4
	.dc.b	$FC,$43,$7C,$1F,$CC,$C6,$2B,$FC,$7C,$FE,$FE,$FF,$A7,$2F,$84,$D9
	.dc.b	$C3,$5B,$3F,$D4,$FC,$E7,$C8,$7E,$91,$B1,$9F,$A4,$C4,$FE,$A3,$E9
	.dc.b	$1F,$0F,$7E,$1F,$0D,$BF,$EB,$DC,$AF,$0B,$F5,$ED,$8C,$F8,$58,$9F
	.dc.b	$D9,$FD,$23,$E5,$A9,$8F,$5E,$DF,$F6,$9F,$A4,$7D,$27,$EE,$9B,$19
	.dc.b	$F8,$FC,$4E,$7E,$93,$5F,$49,$6B,$93,$FC,$0B,$E9,$1D,$B7,$B7,$0D
	.dc.b	$AF,$EF,$1F,$49,$FB,$3E,$DF,$E8,$8F,$F8,$4E,$33,$E5,$FF,$0C,$8C
	.dc.b	$4E,$B6,$C1,$9F,$7A,$DF,$F8,$0F,$E9,$3F,$C8,$AD,$F8,$BF,$E2,$B1
	.dc.b	$FD,$5A,$30,$7F,$82,$7E,$93,$F6,$18,$1A,$7C,$D7,$FA,$C3,$8F,$17
	.dc.b	$F8,$FA,$F2,$2C,$61,$FE,$47,$D5,$FE,$0F,$BF,$F0,$B2,$F3,$73,$E0
	.dc.b	$69,$FE,$53,$F4,$8F,$8C,$BF,$0D,$86,$E1,$FE,$1F,$FA,$4F,$C3,$B7
	.dc.b	$C1,$FE,$63,$18,$DB,$FB,$72,$8C,$9F,$CD,$BE,$89,$CF,$8B,$FF,$15
	.dc.b	$46,$1F,$F1,$67,$D2,$7F,$C0,$06,$FB,$1F,$CF,$A3,$18,$FF,$7B,$7F
	.dc.b	$8E,$79,$7F,$EA,$56,$D7,$FA,$92,$FF,$D0,$CF,$23,$BF,$C8,$73,$F1
	.dc.b	$7A,$7B,$16,$7C,$2E,$1F,$1F,$9D,$72,$F0,$3F,$23,$E2,$BC,$9E,$8D
	.dc.b	$AF,$4B,$87,$D7,$D1,$6F,$D0,$FA,$CF,$89,$F1,$F6,$7E,$1B,$EF,$F3
	.dc.b	$FE,$07,$99,$67,$E2,$B8,$7F,$CA,$5F,$48,$71,$87,$F8,$7E,$31,$1F
	.dc.b	$33,$8F,$E5,$6F,$F2,$DB,$1F,$43,$E4,$7E,$71,$8B,$F9,$CF,$F6,$EB
	.dc.b	$FF,$24,$CF,$83,$E2,$7F,$13,$D9,$97,$FB,$9F,$D1,$FF,$9C,$A7,$FC
	.dc.b	$CC,$DC,$3F,$C6,$4F,$E3,$E3,$3D,$7F,$85,$E2,$91,$83,$F0,$AF,$4F
	.dc.b	$90,$91,$F8,$8F,$A4,$B8,$74,$F9,$0F,$09,$8E,$AF,$F7,$46,$48,$F2
	.dc.b	$FA,$33,$0F,$2F,$CF,$DC,$3F,$C8,$DD,$CC,$FF,$C1,$AD,$AF,$5F,$93
	.dc.b	$89,$FC,$E7,$4C,$9F,$F6,$D7,$4D,$C3,$DF,$F2,$3F,$10,$F8,$BE,$25
	.dc.b	$AF,$8B,$1F,$E1,$AF,$9D,$FE,$43,$5C,$27,$FD,$1D,$18,$E0,$F5,$F2
	.dc.b	$7F,$8A,$28,$38,$7E,$AB,$FE,$39,$6B,$FE,$44,$6B,$C0,$FD,$9F,$D1
	.dc.b	$73,$6E,$1F,$E4,$16,$9F,$E4,$3E,$94,$EF,$E8,$FF,$C3,$FE,$D6,$DB
	.dc.b	$5F,$E4,$97,$FC,$67,$E2,$E7,$D0,$FE,$AB,$3F,$E4,$DF,$B1,$FF,$13
	.dc.b	$BD,$DA,$E6,$3F,$C9,$91,$88,$FF,$2E,$FD,$23,$8F,$D5,$BF,$E8,$67
	.dc.b	$E1,$CF,$99,$FF,$5B,$35,$FF,$41,$3C,$5C,$CF,$2D,$7E,$24,$EE,$BA
	.dc.b	$52,$31,$3A,$AE,$13,$E0,$6B,$BD,$2D,$70,$8D,$1E,$03,$CB,$1A,$37
	.dc.b	$32,$32,$43,$D2,$E5,$F5,$C7,$17,$A1,$A3,$85,$FA,$AB,$C5,$1E,$BA
	.dc.b	$E5,$39,$AE,$53,$C4,$91,$97,$9D,$FC,$17,$BC,$65,$F1,$98,$8D,$38
	.dc.b	$24,$78,$EB,$C3,$3F,$42,$BC,$33,$9A,$E4,$7D,$C8,$CC,$F4,$9D,$4F
	.dc.b	$48,$E4,$F0,$3D,$C7,$A0,$E4,$66,$FF,$04,$F1,$32,$7C,$08,$EC,$74
	.dc.b	$74,$74,$34,$70,$9F,$84,$79,$09,$1C,$5C,$BD,$23,$F0,$67,$49,$FA
	.dc.b	$A7,$92,$78,$C7,$59,$E4,$71,$7C,$51,$D1,$D1,$D1,$E2,$FE,$4C,$F8
	.dc.b	$47,$08,$E5,$3E,$0A,$FA,$73,$E4,$3F,$5F,$7F,$EB,$D9,$F8,$F1,$83
	.dc.b	$F8,$33,$87,$EF,$99,$3F,$D9,$4E,$5F,$25,$78,$27,$F4,$FF,$3A,$F8
	.dc.b	$5F,$0D,$79,$4D,$9F,$C7,$A3,$4F,$D9,$9D,$C6,$23,$E9,$9C,$48,$E7
	.dc.b	$1F,$87,$87,$CD,$C3,$E5,$4E,$5F,$25,$7E,$26,$30,$F8,$8E,$93,$C5
	.dc.b	$F2,$61,$E8,$8C,$47,$82,$31,$87,$99,$B1,$D5,$AF,$A2,$BB,$BD,$A7
	.dc.b	$83,$EF,$38,$14,$E1,$3C,$B0,$FB,$F5,$CA,$FC,$64,$BF,$9F,$5B,$C3
	.dc.b	$E0,$B7,$0F,$75,$F1,$AC,$42,$30,$F8,$70,$FE,$4B,$18,$8F,$A3,$87
	.dc.b	$C2,$8C,$4F,$D0,$AF,$0D,$71,$C3,$C2,$D3,$C3,$87,$F2,$8F,$80,$91
	.dc.b	$FC,$E9,$FE,$57,$E2,$C8,$FC,$8B,$F9,$52,$3F,$94,$E1,$FE,$63,$EB
	.dc.b	$48,$FE,$87,$1F,$CD,$08,$E0,$FE,$8F,$E2,$FF,$3A,$FA,$3F,$E8,$3A
	.dc.b	$7F,$42,$6E,$7F,$DA,$5B,$5F,$DC,$9F,$F3,$53,$FD,$D2,$78,$79,$FC
	.dc.b	$EA,$4E,$5E,$51,$FD,$23,$E9,$3F,$D3,$33,$93,$A4,$78,$CD,$65,$E5
	.dc.b	$F9,$BE,$1E,$5C,$AE,$03,$24,$FA,$FB,$F8,$4E,$A7,$2D,$8F,$81,$E1
	.dc.b	$7C,$1F,$A4,$93,$B8,$CB,$48,$35,$F8,$DE,$27,$C9,$F2,$7F,$7B,$F1
	.dc.b	$12,$BC,$11,$96,$90,$E5,$C1,$F3,$7C,$3F,$A4,$48,$F4,$3F,$C2,$39
	.dc.b	$7E,$B7,$C4,$C9,$F6,$3F,$67,$C2,$91,$EF,$6B,$D8,$FF,$14,$FC,$E3
	.dc.b	$6F,$EF,$3F,$1F,$FE,$D4,$9F,$96,$D7,$80,$71,$B7,$F8,$19,$7C,$1F
	.dc.b	$22,$4F,$27,$AF,$CB,$FF,$81,$BC,$77,$CF,$D2,$57,$83,$E6,$3C,$64
	.dc.b	$70,$79,$3E,$63,$CE,$47,$07,$F9,$23,$E9,$3D,$7F,$49,$23,$D6,$FB
	.dc.b	$BE,$4A,$47,$B9,$AF,$BB,$C1,$71,$FC,$4F,$80,$93,$F1,$1F,$8F,$E9
	.dc.b	$48,$F8,$ED,$7E,$3F,$03,$E9,$3F,$C4,$BF,$49,$FE,$51,$9F,$96,$DE
	.dc.b	$3F,$F1,$7B,$7A,$7F,$CA,$CB,$CE,$E2,$72,$A8,$7E,$82,$34,$FE,$8B
	.dc.b	$E9,$3F,$D9,$EE,$67,$8B,$9D,$7F,$C6,$FF,$48,$E3,$A4,$9F,$D5,$C3
	.dc.b	$8F,$FA,$BF,$93,$FD,$63,$DD,$95,$FB,$0F,$35,$3F,$E4,$2C,$DE,$5D
	.dc.b	$F7,$65,$F9,$63,$C3,$CD,$E2,$E9,$93,$E0,$8F,$EC,$BF,$2D,$23,$B3
	.dc.b	$FD,$A3,$CB,$ED,$E5,$C9,$F8,$0F,$F9,$4B,$E9,$3F,$8A,$C3,$81,$F1
	.dc.b	$5F,$00,$91,$EE,$FA,$0F,$EE,$7C,$9F,$DD,$35,$93,$ED,$8E,$9E,$7F
	.dc.b	$5F,$80,$93,$FC,$43,$2F,$E4,$2F,$E9,$B4,$7C,$27,$F9,$91,$7C,$25
	.dc.b	$31,$1F,$E0,$2E,$29,$5F,$2D,$FF,$02,$EF,$EB,$EE,$93,$DD,$F7,$7B
	.dc.b	$3D,$7D,$EE,$DF,$13,$2B,$F9,$1E,$69,$FF,$3A,$F9,$7F,$10,$9E,$06
	.dc.b	$38,$3E,$47,$AB,$30,$F6,$3F,$2F,$D4,$F5,$74,$49,$EA,$FB,$72,$EB
	.dc.b	$C7,$DB,$69,$5F,$CD,$33,$4F,$D5,$F8,$1F,$AD,$F0,$B2,$BE,$54,$7F
	.dc.b	$A2,$BE,$8D,$F0,$74,$93,$FC,$FA,$1F,$1B,$BF,$5A,$4E,$5E,$5E,$B4
	.dc.b	$8E,$0D,$7F,$8D,$D3,$E1,$3C,$1F,$49,$27,$B3,$E3,$F6,$24,$78,$5A
	.dc.b	$F2,$F6,$3F,$D4,$3F,$48,$F9,$DA,$49,$FE,$AB,$1F,$AF,$6F,$9B,$C0
	.dc.b	$8F,$F5,$AF,$A4,$FC,$BB,$FD,$73,$E2,$BF,$C4,$A4,$F8,$2F,$F5,$7A
	.dc.b	$F9,$9A,$49,$E0,$FF,$AD,$29,$E4,$7F,$AD,$BE,$8F,$F9,$BD,$7F,$68
	.dc.b	$6B,$FC,$F9,$BE,$AF,$FB,$09,$7B,$BC,$49,$3C,$5E,$9F,$41,$23,$A3
	.dc.b	$5F,$E9,$3E,$8F,$F1,$D7,$FD,$F3,$5F,$1D,$F1,$6F,$E5,$3C,$DB,$48
	.dc.b	$F3,$35,$F3,$7A,$BF,$DC,$57,$D2,$37,$FF,$7F,$61,$FC,$67,$4F,$F7
	.dc.b	$E3,$6B,$F9,$90,$70,$FF,$35,$6B,$CB,$E8,$DF,$49,$C9,$B1,$CA,$51
	.dc.b	$88,$E4,$D7,$BF,$D2,$78,$11,$1C,$34,$AE,$3D,$AF,$F7,$F6,$1F,$EF
	.dc.b	$49,$FF,$44,$79,$3F,$E2,$07,$FA,$33,$79,$76,$FF,$4F,$6C,$78,$3D
	.dc.b	$6F,$EF,$53,$B3,$FC,$52,$E2,$3F,$BF,$9F,$8D,$8E,$DE,$87,$28,$F4
	.dc.b	$F3,$D7,$F8,$FC,$E8,$F2,$7A,$39,$AD,$32,$3B,$F9,$F6,$BF,$C1,$B8
	.dc.b	$23,$B6,$4E,$6B,$CB,$86,$E1,$FE,$40,$9F,$D8,$3E,$AF,$03,$38,$F4
	.dc.b	$7D,$1F,$49,$F1,$A3,$FC,$36,$FC,$A6,$EF,$AD,$E5,$FC,$4E,$EB,$5E
	.dc.b	$67,$9F,$9F,$D3,$D0,$B5,$FE,$52,$C5,$19,$8C,$7C,$35,$C3,$C2,$F9
	.dc.b	$3D,$AF,$2F,$B9,$6B,$93,$E8,$F4,$24,$60,$EC,$FB,$3D,$79,$1E,$1F
	.dc.b	$46,$E3,$FE,$38,$C3,$C7,$3F,$E6,$65,$8F,$DB,$73,$3B,$8C,$7B,$12
	.dc.b	$B8,$7A,$F6,$BF,$C8,$7F,$49,$FE,$05,$9F,$F3,$8B,$63,$3D,$BF,$87
	.dc.b	$8C,$72,$AB,$F4,$34,$FA,$47,$CA,$7E,$35,$B3,$FE,$7C,$5F,$21,$99
	.dc.b	$D7,$0E,$F8,$6D,$ED,$FC,$B7,$A0,$FA,$0B,$F1,$58,$6D,$7F,$A2,$3E
	.dc.b	$8F,$FC,$6C,$DF,$85,$FC,$F4,$7F,$82,$23,$07,$FA,$3B,$1F,$EF,$6D
	.dc.b	$F2,$3F,$49,$18,$DB,$FC,$CE,$FF,$13,$9F,$F4,$AF,$D1,$FF,$58,$6E
	.dc.b	$7F,$AB,$8C,$57,$F9,$B9,$47,$EB,$FE,$8F,$F9,$8D,$7F,$A7,$9F,$D2
	.dc.b	$37,$CF,$D7,$85,$BF,$8B,$93,$BF,$1A,$E5,$FE,$7A,$FA,$1F,$F1,$AC
	.dc.b	$FF,$AA,$1B,$19,$E9,$FD,$F4,$63,$9F,$EC,$4F,$F1,$8F,$A1,$7C,$0F
	.dc.b	$F0,$36,$E1,$FE,$B2,$5E,$EE,$67,$5B,$EF,$C8,$B7,$B7,$F8,$67,$47
	.dc.b	$F2,$48,$FF,$5D,$BF,$B7,$CB,$C3,$92,$34,$FF,$A5,$1D,$4F,$E3,$3E
	.dc.b	$83,$F8,$9C,$FF,$A6,$1F,$A9,$5C,$9C,$B7,$CC,$FE,$39,$1F,$A4,$AF
	.dc.b	$F6,$5B,$6B,$97,$C7,$6B,$C2,$FE,$45,$F1,$DF,$E0,$9B,$FF,$68,$B6
	.dc.b	$BF,$D4,$EB,$E9,$77,$EF,$7F,$29,$FA,$1F,$E5,$DA,$7B,$1F,$E4,$18
	.dc.b	$7F,$B6,$9C,$BF,$A6,$73,$87,$FA,$CB,$E8,$3E,$0D,$7C,$47,$E0,$C7
	.dc.b	$F4,$C7,$35,$FA,$E7,$30,$FA,$0F,$B9,$3F,$B2,$6C,$71,$F4,$7F,$9F
	.dc.b	$46,$21,$5F,$2A,$BA,$6F,$FD,$0E,$3E,$5C,$60,$FF,$79,$BB,$9F,$27
	.dc.b	$7E,$EB,$83,$F0,$FE,$17,$97,$CD,$5C,$38,$BD,$7D,$19,$18,$3A,$BE
	.dc.b	$1F,$36,$47,$C9,$F9,$8B,$97,$C7,$9F,$37,$89,$8D,$FE,$43,$FD,$6F
	.dc.b	$E8,$3F,$8A,$61,$FD,$D2,$7F,$1F,$18,$BF,$A1,$5B,$AF,$81,$E7,$C7
	.dc.b	$DF,$E8,$DA,$FA,$DE,$0F,$F7,$77,$EA,$35,$F9,$6D,$E1,$39,$C5,$F0
	.dc.b	$94,$98,$7F,$82,$63,$2F,$F8,$2F,$E8,$3F,$CD,$AD,$F8,$AE,$D1,$8C
	.dc.b	$7F,$C3,$ED,$7F,$DA,$5E,$A7,$82,$BF,$C2,$8D,$8F,$F1,$2B,$5D,$E3
	.dc.b	$D3,$C2,$95,$E1,$6C,$F6,$F0,$FC,$24,$9F,$33,$E4,$E2,$91,$E9,$6C
	.dc.b	$F9,$BE,$6B,$D2,$49,$F7,$3F,$27,$A5,$CD,$49,$3F,$E2,$D8,$CB,$F1
	.dc.b	$7B,$3E,$19,$1C,$1B,$F5,$BA,$11,$C1,$FE,$37,$EE,$77,$FD,$96,$C6
	.dc.b	$63,$07,$DD,$E1,$7D,$05,$24,$FF,$90,$63,$2F,$F9,$0F,$E6,$BF,$DE
	.dc.b	$4D,$F8,$5F,$C3,$46,$36,$FF,$2B,$B5,$FE,$D2,$3E,$83,$FC,$51,$87
	.dc.b	$F9,$35,$B1,$FE,$5E,$5F,$20,$F0,$F5,$F0,$1E,$6A,$FF,$31,$F6,$7F
	.dc.b	$B1,$DF,$F2,$B3,$5C,$F8,$3F,$E5,$68,$C4,$78,$BE,$7F,$FB,$AB,$FE
	.dc.b	$5D,$6B,$FE,$6E,$6B,$BF,$6B,$FC,$BF,$0F,$27,$FC,$07,$E1,$FF,$3E
	.dc.b	$D9,$E8,$63,$87,$FC,$0F,$18,$C7,$F6,$D1,$FE,$0A,$C3,$94,$60,$FF
	.dc.b	$36,$B8,$57,$D6,$FA,$47,$F9,$CA,$FA,$3A,$BE,$98,$6A,$18,$73,$C3
	.dc.b	$94,$77,$17,$CA,$3F,$AB,$CE,$EF,$9D,$6E,$B5,$1E,$CC,$3B,$56,$EB
	.dc.b	$D3,$87,$86,$F7,$1E,$7A,$DC,$C8,$E9,$1B,$30,$FA,$CE,$66,$88,$64
	.dc.b	$F6,$C7,$0C,$3D,$B1,$9A,$F0,$BA,$3E,$05,$F9,$DF,$83,$7E,$17,$37
	.dc.b	$CF,$0F,$74,$78,$30,$F8,$B3,$98,$F6,$19,$9F,$8E,$F9,$6F,$B6,$1E
	.dc.b	$23,$C6,$F8,$4C,$E1,$E1,$9E,$17,$D5,$F2,$9D,$4F,$86,$68,$C1,$A3
	.dc.b	$04,$C3,$E1,$93,$0F,$2B,$EA,$8E,$38,$79,$2B,$C0,$F9,$EF,$C5,$87
	.dc.b	$9A,$BB,$BE,$9A,$F4,$5F,$59,$F2,$BD,$5F,$A5,$5A,$C3,$C2,$48,$F9
	.dc.b	$26,$8F,$A2,$43,$04,$9E,$84,$8F,$AB,$19,$7A,$3F,$7C,$FD,$D5,$7E
	.dc.b	$E0,$8F,$EF,$DA,$E8,$C1,$0F,$95,$1D,$2D,$E2,$FC,$4A,$F8,$AD,$AF
	.dc.b	$45,$FE,$AA,$31,$1C,$8F,$DE,$B8,$8F,$54,$60,$E3,$7D,$EB,$F5,$B5
	.dc.b	$B8,$F1,$36,$7E,$7C,$7F,$0B,$8C,$47,$C1,$AF,$13,$D0,$D9,$0F,$E0
	.dc.b	$B1,$C2,$FF,$28,$DB,$F0,$CF,$13,$A5,$F8,$9C,$57,$E0,$DF,$D6,$D7
	.dc.b	$16,$9A,$AF,$8E,$D3,$8C,$6D,$FD,$43,$6F,$D3,$1F,$24,$F0,$9E,$9A
	.dc.b	$FA,$6F,$33,$4E,$1B,$5E,$06,$9E,$03,$84,$FD,$27,$E1,$9F,$C1,$BB
	.dc.b	$FF,$9F,$91,$FE,$8C,$FF,$42,$D3,$F9,$0B,$23,$F8,$6C,$7F,$31,$F9
	.dc.b	$EF,$F1,$44,$1A,$4C,$BF,$B1,$C7,$FB,$32,$3F,$D0,$F7,$FE,$88,$6C
	.dc.b	$E4,$E9,$F2,$9F,$CF,$7E,$7D,$F2,$D4,$98,$7F,$3D,$B7,$CB,$78,$CE
	.dc.b	$F0,$3C,$8E,$1A,$6A,$38,$B7,$C3,$D4,$7F,$AA,$A3,$04,$3D,$6B,$E7
	.dc.b	$F6,$48,$E8,$F2,$F0,$E4,$77,$6B,$CB,$D4,$EE,$47,$C8,$F8,$BD,$4F
	.dc.b	$9C,$C8,$F8,$11,$F4,$7E,$17,$CE,$48,$F6,$35,$FE,$95,$82,$38,$47
	.dc.b	$C9,$F2,$FF,$B5,$7C,$BF,$AE,$30,$43,$FA,$93,$97,$FA,$A7,$6B,$E0
	.dc.b	$7F,$B9,$3F,$EF,$B8,$FE,$E0,$D7,$71,$A2,$3A,$78,$FB,$9F,$DA,$FE
	.dc.b	$6E,$4F,$F7,$37,$AF,$87,$DB,$E4,$2C,$F6,$8C,$3E,$0F,$77,$FC,$09
	.dc.b	$D2,$F2,$69,$E6,$89,$3B,$21,$EC,$EF,$E3,$DB,$EC,$64,$7F,$82,$EF
	.dc.b	$FB,$66,$AF,$9E,$F8,$3E,$4C,$AF,$83,$1F,$10,$8E,$0F,$8B,$F0,$7F
	.dc.b	$E1,$9F,$9E,$FF,$23,$F8,$1F,$EE,$A6,$F8,$1F,$99,$A4,$61,$FC,$E3
	.dc.b	$FE,$47,$7F,$C3,$CD,$7F,$C1,$0E,$DE,$27,$7F,$23,$C1,$ED,$7F,$8A
	.dc.b	$7E,$6F,$EE,$61,$FE,$FA,$6E,$1E,$1C,$15,$C3,$0E,$ED,$7B,$FB,$7E
	.dc.b	$B2,$3E,$47,$B7,$77,$F8,$DF,$F0,$9F,$D6,$4F,$F8,$75,$7A,$B9,$35
	.dc.b	$E1,$7E,$53,$D5,$95,$FE,$3B,$ED,$47,$E6,$3E,$73,$D4,$C9,$FF,$20
	.dc.b	$6D,$A6,$CF,$84,$8F,$F9,$17,$14,$79,$7E,$73,$F4,$67,$E9,$3C,$49
	.dc.b	$5E,$18,$F9,$07,$C9,$89,$1E,$08,$F7,$7C,$DF,$BF,$C3,$93,$DE,$3F
	.dc.b	$5F,$E2,$C8,$F6,$3F,$B2,$F5,$BE,$A3,$23,$D6,$FD,$78,$FF,$30,$7D
	.dc.b	$7F,$EC,$67,$FC,$BC,$D7,$FC,$90,$DE,$67,$37,$E2,$E1,$E2,$7F,$00
	.dc.b	$F5,$FF,$80,$E5,$30,$F9,$D1,$F8,$53,$F0,$D1,$23,$E6,$C7,$8B,$E3
	.dc.b	$E4,$78,$9F,$2F,$D3,$7F,$0C,$D7,$F8,$6F,$93,$2B,$ED,$47,$E4,$4F
	.dc.b	$C9,$44,$8F,$B3,$1E,$EF,$A2,$F7,$F8,$12,$7D,$F1,$FE,$71,$F6,$64
	.dc.b	$7F,$9E,$5F,$F3,$A7,$D3,$FF,$A1,$74,$FF,$55,$4F,$F9,$AD,$BC,$1F
	.dc.b	$12,$3F,$A3,$46,$38,$3F,$60,$AB,$FD,$1D,$F4,$0E,$4B,$CD,$30,$FF
	.dc.b	$47,$4E,$5B,$CD,$C0,$8E,$B9,$B9,$91,$C1,$CB,$9B,$BE,$52,$3A,$BE
	.dc.b	$2E,$09,$1E,$06,$BF,$E8,$7C,$1C,$3F,$EA,$5F,$A0,$7C,$BD,$25,$7F
	.dc.b	$A9,$61,$E6,$6C,$91,$97,$FD,$5D,$F4,$0F,$A1,$FF,$58,$78,$6F,$99
	.dc.b	$A4,$BF,$F5,$84,$3E,$6F,$64,$8C,$BD,$3E,$87,$FD,$75,$F3,$DF,$97
	.dc.b	$E7,$FE,$40,$DA,$FE,$F4,$D4,$FE,$69,$FF,$5F,$35,$FF,$54,$37,$A3
	.dc.b	$EF,$47,$0E,$FE,$FF,$29,$23,$8B,$5F,$F5,$87,$D0,$FF,$29,$BF,$93
	.dc.b	$1D,$B8,$7C,$E7,$A3,$82,$47,$89,$AF,$8B,$D5,$FF,$6E,$AF,$A1,$6F
	.dc.b	$EE,$9C,$CF,$F3,$F3,$DA,$E1,$D6,$9F,$E1,$CC,$53,$99,$F8,$BE,$13
	.dc.b	$E8,$74,$D8,$FF,$18,$D2,$31,$1D,$E3,$13,$FE,$25,$87,$D0,$D2,$3F
	.dc.b	$E2,$97,$FC,$55,$E8,$7E,$A7,$1F,$9B,$3F,$83,$E1,$FF,$78,$29,$D4
	.dc.b	$7B,$79,$2C,$FF,$5D,$6A,$F5,$5E,$A6,$23,$A1,$FE,$E3,$8C,$D7,$6E
	.dc.b	$6D,$7D,$0E,$DB,$1E,$28,$E5,$E9,$62,$FF,$DD,$87,$F5,$C9,$F3,$C7
	.dc.b	$8A,$DF,$11,$C2,$7F,$C9,$5A,$7E,$2A,$73,$18,$3F,$BB,$43,$9F,$D6
	.dc.b	$FE,$C2,$35,$87,$29,$CF,$34,$9C,$CF,$C1,$F0,$B0,$FB,$DF,$8B,$F4
	.dc.b	$3F,$2E,$76,$FC,$4C,$3F,$3A,$E6,$91,$8D,$F8,$9E,$DF,$06,$DF,$83
	.dc.b	$3F,$F2,$23,$D3,$E2,$7B,$FB,$9E,$06,$C7,$82,$3C,$7F,$2D,$8C,$BB
	.dc.b	$CF,$F8,$5F,$34,$FA,$3E,$23,$D3,$F2,$96,$BD,$2F,$B3,$E6,$24,$60
	.dc.b	$F6,$BF,$03,$DE,$91,$F0,$3E,$26,$E5,$E0,$3F,$CF,$D8,$7F,$8F,$E7
	.dc.b	$E5,$BE,$26,$E7,$F8,$E8,$C4,$F1,$3F,$DB,$FF,$2B,$93,$5F,$25,$6F
	.dc.b	$EA,$FD,$0F,$EC,$1B,$E0,$FE,$62,$3F,$59,$18,$3E,$DF,$D0,$BE,$87
	.dc.b	$F3,$AD,$8C,$FC,$8E,$27,$3C,$1A,$E0,$B5,$C9,$FD,$2F,$A4,$FA,$1B
	.dc.b	$F1,$58,$6D,$7E,$2B,$E8,$5F,$0F,$FA,$C6,$C6,$7D,$0C,$4F,$EC,$32
	.dc.b	$FD,$17,$17,$C3,$7F,$8A,$D7,$E5,$1C,$F6,$FF,$2A,$73,$3F,$27,$E8
	.dc.b	$7F,$B2,$57,$E6,$96,$9F,$BC,$73,$3B,$8F,$DF,$FD,$0F,$F8,$4A,$BF
	.dc.b	$42,$FE,$A5,$BE,$8E,$BC,$3D,$FC,$64,$9D,$F8,$17,$1F,$B5,$F4,$2E
	.dc.b	$AF,$E1,$6D,$AF,$D6,$AE,$AC,$CF,$F1,$0F,$A1,$7C,$4D,$31,$F1,$57
	.dc.b	$0F,$DA,$CF,$B1,$BE,$3F,$1E,$0E,$4F,$F1,$AE,$9F,$F6,$84,$FE,$F1
	.dc.b	$FD,$A7,$27,$F5,$27,$3A,$FE,$45,$CC,$FF,$9F,$FE,$85,$E6,$7F,$24
	.dc.b	$7E,$8C,$66,$31,$E6,$67,$C6,$93,$FC,$2A,$9F,$1F,$8C,$7F,$3D,$F0
	.dc.b	$AF,$80,$FE,$5E,$DB,$FE,$22,$BE,$1F,$33,$FC,$D3,$E8,$7F,$8F,$DF
	.dc.b	$E9,$DB,$B7,$F6,$C7,$F9,$9C,$67,$D1,$FE,$77,$F4,$1E,$C8,$FF,$79
	.dc.b	$35,$3F,$0A,$BF,$BA,$B9,$F9,$DF,$B1,$3F,$B5,$6C,$EB,$D6,$FE,$8F
	.dc.b	$18,$D3,$FD,$4D,$7F,$1A,$FB,$F0,$78,$B8,$7C,$6F,$93,$DB,$97,$FC
	.dc.b	$C2,$F8,$5F,$92,$7D,$3F,$3D,$2B,$F9,$AE,$AF,$A1,$C5,$FF,$09,$C1
	.dc.b	$EA,$7C,$03,$3B,$7B,$1F,$87,$E8,$4C,$79,$FC,$95,$C3,$97,$8F,$F2
	.dc.b	$0C,$61,$F1,$1F,$EC,$BF,$43,$FE,$7A,$BF,$E8,$ED,$AE,$31,$83,$E8
	.dc.b	$F8,$11,$F7,$7A,$B6,$7F,$90,$7D,$9B,$E6,$BF,$B8,$B6,$BE,$27,$89
	.dc.b	$89,$FE,$EB,$F4,$4F,$9E,$A7,$93,$E2,$AD,$FF,$5A,$E9,$FE,$0B,$BF
	.dc.b	$F8,$C1,$B1,$FE,$DB,$B7,$D1,$7D,$AE,$8F,$B7,$EA,$CA,$FE,$C7,$ED
	.dc.b	$FE,$DD,$E9,$1E,$2F,$26,$47,$F6,$8C,$3F,$72,$D8,$CF,$07,$F3,$DF
	.dc.b	$9A,$F6,$3F,$CF,$EB,$FD,$9F,$5F,$DC,$27,$3A,$7F,$91,$A7,$FD,$9B
	.dc.b	$3D,$7E,$63,$E0,$F3,$FC,$27,$E1,$FC,$87,$F8,$83,$10,$BF,$F6,$5D
	.dc.b	$FF,$8E,$E3,$31,$87,$0D,$7F,$C8,$0B,$CA,$E9,$1C,$DB,$19,$9E,$93
	.dc.b	$96,$FD,$0F,$F4,$56,$FB,$3C,$A3,$13,$E3,$6B,$FE,$4F,$F9,$AF,$92
	.dc.b	$F9,$DA,$E7,$D0,$F1,$C6,$23,$DE,$7D,$AF,$95,$F3,$F2,$1A,$4C,$7C
	.dc.b	$FC,$BF,$DC,$2B,$FC,$BC,$D7,$9B,$5F,$96,$D5,$F4,$5F,$E3,$4A,$F8
	.dc.b	$2D,$8F,$9C,$FF,$92,$3C,$8C,$F9,$3F,$EC,$09,$F3,$E9,$FE,$68,$7E
	.dc.b	$3F,$C0,$FC,$A5,$19,$7E,$34,$67,$E8,$9E,$4D,$24,$FD,$28,$CB,$F5
	.dc.b	$3E,$93,$FD,$48,$DF,$87,$ED,$18,$8F,$C2,$35,$FF,$40,$4B,$C7,$FF
	.dc.b	$25,$EF,$F3,$A3,$F9,$DC,$FF,$91,$9A,$FF,$86,$5E,$0E,$67,$8B,$F9
	.dc.b	$48,$DF,$4F,$FA,$89,$B5,$D6,$B7,$49,$E6,$F5,$F4,$73,$1C,$23,$A4
	.dc.b	$E0,$CC,$C8,$F1,$10,$E4,$7A,$5E,$0E,$6F,$C9,$1C,$63,$D9,$3A,$78
	.dc.b	$92,$7D,$51,$97,$8C,$75,$3A,$5F,$B5,$C1,$C5,$C3,$D8,$CC,$72,$70
	.dc.b	$47,$04,$8E,$2F,$8A,$38,$C7,$D7,$AF,$23,$A3,$11,$A3,$E4,$9E,$63
	.dc.b	$EB,$47,$C1,$9E,$91,$E1,$9E,$2F,$A4,$E2,$64,$E2,$75,$3E,$D5,$78
	.dc.b	$9C,$9C,$4C,$9E,$13,$F0,$15,$E1,$9F,$44,$76,$AF,$80,$F2,$7C,$A4
	.dc.b	$8F,$2B,$A8,$EA,$79,$4E,$A7,$96,$FF,$1B,$1C,$63,$A4,$F2,$3C,$2F
	.dc.b	$28,$E9,$3F,$91,$8C,$4E,$A3,$04,$BF,$C4,$4F,$EB,$1F,$DD,$11,$FD
	.dc.b	$C3,$5F,$E6,$D1,$FA,$07,$81,$A3,$E3,$3E,$89,$FE,$43,$F3,$6E,$CF
	.dc.b	$9D,$5E,$43,$E4,$4E,$A3,$35,$C6,$72,$7F,$48,$AE,$4F,$6B,$E7,$18
	.dc.b	$8F,$95,$3C,$5E,$C2,$35,$87,$CB,$8C,$CE,$E7,$38,$7C,$7B,$EB,$87
	.dc.b	$91,$B7,$E1,$3F,$09,$5C,$1F,$39,$D2,$3F,$83,$C6,$63,$07,$09,$E8
	.dc.b	$D3,$A1,$F8,$38,$E1,$7E,$B2,$57,$4A,$E3,$5B,$6E,$1E,$DB,$E3,$58
	.dc.b	$C3,$E7,$5F,$6A,$D5,$E6,$3D,$37,$C1,$D5,$70,$7E,$6D,$FC,$26,$9E
	.dc.b	$92,$9E,$33,$89,$D2,$BF,$92,$F8,$DF,$D1,$C9,$1F,$D1,$5A,$FF,$37
	.dc.b	$76,$FF,$3B,$F1,$FF,$98,$FD,$03,$0A,$49,$FF,$06,$E5,$97,$FC,$1F
	.dc.b	$3F,$E3,$37,$33,$FD,$D2,$31,$1F,$E1,$16,$BF,$DE,$03,$87,$87,$8B
	.dc.b	$FD,$8E,$3F,$C8,$91,$FE,$0E,$6B,$CF,$C6,$D1,$BE,$BF,$4B,$23,$C6
	.dc.b	$79,$A3,$DA,$47,$27,$89,$FE,$A3,$C0,$F8,$4F,$87,$E1,$7C,$4F,$A1
	.dc.b	$97,$C3,$14,$EA,$30,$D7,$FD,$F9,$0F,$0F,$E6,$FB,$32,$79,$3F,$43
	.dc.b	$C3,$91,$82,$1C,$9F,$AB,$F1,$52,$3C,$4F,$D8,$F2,$A4,$60,$87,$89
	.dc.b	$FE,$DB,$E1,$B3,$DC,$E4,$F8,$71,$97,$E2,$FC,$6B,$E1,$FF,$BA,$BF
	.dc.b	$F0,$06,$33,$E3,$62,$73,$F0,$56,$78,$C6,$1E,$7F,$27,$FD,$FB,$C2
	.dc.b	$7D,$0D,$3A,$C7,$6F,$0E,$D7,$03,$C6,$F8,$BE,$0F,$FC,$0F,$DA,$F8
	.dc.b	$CA,$66,$BD,$3F,$09,$57,$77,$E6,$56,$EC,$3D,$B5,$D3,$C4,$B5,$C8
	.dc.b	$FC,$3E,$3F,$D0,$6F,$F8,$80,$F7,$FE,$29,$F1,$BC,$29,$5F,$29,$AF
	.dc.b	$8B,$C1,$FF,$2C,$D3,$E2,$3F,$C4,$3F,$43,$FA,$5D,$7C,$CD,$AF,$F2
	.dc.b	$FB,$6F,$FC,$FE,$BC,$BE,$27,$28,$87,$E7,$A3,$4D,$F7,$BF,$C5,$FF
	.dc.b	$43,$F7,$E3,$C8,$7F,$A2,$9C,$61,$E2,$9F,$29,$27,$91,$D7,$D7,$F4
	.dc.b	$91,$C1,$FA,$EA,$FD,$3B,$FB,$0F,$A1,$78,$DF,$E4,$05,$EB,$53,$11
	.dc.b	$F9,$8F,$0E,$57,$E5,$9F,$CD,$FA,$DE,$A6,$47,$77,$DB,$1D,$49,$1E
	.dc.b	$17,$C1,$F3,$DF,$A4,$3F,$4B,$E2,$C9,$FD,$1C,$7C,$63,$E3,$C4,$8F
	.dc.b	$64,$7A,$7E,$7B,$D5,$E2,$C9,$F7,$47,$EC,$3C,$69,$1F,$AE,$7F,$67
	.dc.b	$F0,$9F,$49,$91,$F1,$DF,$AB,$1F,$D7,$BC,$29,$3F,$C6,$3D,$0F,$DA
	.dc.b	$37,$C4,$FD,$EF,$C2,$FE,$FB,$19,$3F,$BC,$8F,$BE,$7E,$02,$24,$7D
	.dc.b	$58,$FE,$D7,$F3,$12,$3B,$3F,$DB,$BE,$87,$F8,$4E,$9F,$C2,$BC,$B9
	.dc.b	$5F,$C2,$23,$F1,$67,$E3,$22,$47,$E0,$63,$FB,$BF,$D0,$FF,$78,$F0
	.dc.b	$24,$FA,$E3,$F8,$27,$B5,$23,$F9,$AF,$95,$FE,$BB,$6F,$D0,$FF,$A1
	.dc.b	$B5,$C4,$7E,$60,$38,$ED,$FD,$F3,$7C,$0F,$F6,$3A,$AF,$F4,$7F,$D0
	.dc.b	$B9,$AF,$44,$C7,$FA,$2C,$E5,$BD,$1C,$08,$EB,$A3,$99,$1C,$1C,$BA
	.dc.b	$3B,$E5,$23,$AB,$E2,$E1,$91,$E0,$6B,$ED,$C1,$C5,$FE,$A6,$FA,$17
	.dc.b	$CC,$D2,$57,$F5,$C8,$79,$BB,$24,$65,$FF,$57,$FD,$0B,$E5,$BF,$D6
	.dc.b	$3E,$1F,$FB,$17,$93,$F9,$33,$79,$3F,$DD,$ED,$C0,$17,$CD,$F2,$24
	.dc.b	$7F,$6E,$8C,$BF,$DC,$30,$7D,$0F,$FB,$03,$E7,$FF,$37,$3F,$6D,$B8
	.dc.b	$FF,$37,$6E,$3F,$F0,$01,$77,$F1,$25,$7F,$7C,$8C,$B7,$E5,$FB,$11
	.dc.b	$C1,$FE,$02,$FA,$1F,$E5,$8A,$7E,$6B,$CF,$83,$CC,$79,$F7,$91,$DD
	.dc.b	$AF,$A7,$D4,$FE,$9E,$BE,$87,$E3,$3F,$22,$D3,$A8,$FF,$12,$46,$34
	.dc.b	$FF,$13,$46,$E7,$4F,$C1,$6B,$CF,$C2,$7D,$0F,$08,$F5,$46,$1E,$76
	.dc.b	$8C,$47,$81,$AF,$83,$E8,$79,$3F,$79,$A9,$A6,$BC,$FF,$F2,$26,$3F
	.dc.b	$DB,$65,$F0,$B8,$E1,$FF,$20,$55,$9F,$C4,$9A,$8F,$F1,$9B,$4F,$DE
	.dc.b	$7A,$DF,$DD,$A3,$35,$C6,$53,$97,$F8,$11,$F9,$29,$E7,$F2,$DB,$8C
	.dc.b	$1A,$70,$67,$D1,$B3,$E2,$70,$FF,$91,$A9,$F1,$18,$BF,$2B,$E3,$C9
	.dc.b	$C9,$78,$F0,$DC,$3F,$C9,$F3,$FB,$87,$A7,$83,$9C,$3D,$EF,$B7,$E8
	.dc.b	$7F,$BE,$4F,$F9,$59,$F8,$95,$FE,$19,$73,$5F,$C8,$5C,$CF,$CC,$6C
	.dc.b	$7E,$DB,$27,$81,$D7,$6F,$95,$18,$BE,$DF,$29,$E1,$F7,$3C,$4D,$8F
	.dc.b	$14,$79,$7E,$67,$18,$F8,$5F,$47,$C4,$FA,$7E,$56,$DF,$A5,$F6,$7C
	.dc.b	$D4,$8C,$1E,$D7,$E0,$7B,$D2,$3E,$07,$C9,$DC,$7F,$9F,$7A,$1F,$CD
	.dc.b	$E7,$FC,$F2,$D8,$CF,$9B,$F8,$58,$C7,$17,$F8,$FF,$EE,$7F,$83,$DF
	.dc.b	$FA,$01,$B7,$FD,$25,$FF,$2F,$46,$7C,$3F,$C7,$7D,$13,$D5,$A6,$3E
	.dc.b	$A6,$D7,$3F,$2F,$F2,$98,$7F,$2B,$8E,$DF,$01,$89,$FC,$C7,$DC,$B9
	.dc.b	$A9,$8F,$92,$B7,$FE,$94,$F5,$3C,$BB,$FF,$95,$DB,$1F,$A8,$7F,$1E
	.dc.b	$EF,$6F,$D2,$3E,$AD,$FF,$B2,$36,$BF,$CC,$0B,$E2,$37,$EE,$F0,$F6
	.dc.b	$A4,$FF,$A9,$0F,$F3,$A9,$5F,$D8,$7D,$13,$C1,$A6,$3D,$8B,$5F,$EA
	.dc.b	$CF,$B8,$6D,$FB,$76,$C6,$7E,$3F,$13,$9F,$A6,$93,$C7,$E6,$EE,$FF
	.dc.b	$DF,$7E,$8D,$5F,$C0,$5B,$5F,$EB,$AF,$C5,$BF,$DB,$0F,$FA,$F2,$33
	.dc.b	$5F,$E0,$5F,$3F,$FB,$4C,$FF,$B0,$1B,$19,$F2,$3F,$86,$46,$38,$FF
	.dc.b	$D0,$DF,$72,$EC,$FE,$20,$DC,$3F,$D9,$2F,$E1,$5B,$1F,$C5,$7C,$3F
	.dc.b	$D8,$E2,$69,$FC,$4A,$FF,$D2,$4E,$BD,$07,$9F,$FC,$75,$B8,$7F,$B5
	.dc.b	$97,$A1,$99,$D7,$8B,$2B,$57,$71,$FF,$12,$7C,$1F,$F8,$92,$7F,$DB
	.dc.b	$ED,$8C,$ED,$FC,$AE,$31,$9F,$FA,$97,$E8,$DF,$2F,$FC,$BD,$B8,$7F
	.dc.b	$BA,$1F,$F3,$CB,$63,$F9,$A7,$D1,$7A,$31,$FF,$1A,$36,$FF,$D6,$0F
	.dc.b	$AA,$BF,$DE,$2E,$5F,$E8,$6F,$F8,$2D,$CF,$BD,$D3,$5E,$AF,$6E,$79
	.dc.b	$3E,$3F,$A3,$B8,$F8,$3C,$3F,$19,$8C,$BC,$2F,$9B,$CE,$97,$EB,$F9
	.dc.b	$8B,$7E,$8F,$3F,$D6,$63,$0F,$4B,$EE,$F3,$E5,$FC,$3F,$8B,$B7,$F0
	.dc.b	$3E,$6F,$E1,$98,$C3,$E0,$BF,$D6,$7E,$8B,$E5,$CF,$F9,$65,$B3,$F1
	.dc.b	$BC,$0D,$71,$FF,$5F,$FA,$17,$CE,$D3,$FB,$17,$D1,$FE,$5F,$4F,$ED
	.dc.b	$73,$A7,$FC,$6B,$87,$FB,$7F,$E8,$7E,$66,$7F,$E6,$76,$C7,$FB,$8E
	.dc.b	$9F,$45,$FD,$B6,$73,$7F,$DD,$97,$D1,$E4,$FE,$E5,$DC,$D7,$15,$E0
	.dc.b	$3C,$7F,$2B,$E9,$3D,$5F,$2B,$63,$F1,$5D,$FF,$8D,$3F,$BE,$7C,$AF
	.dc.b	$F2,$9D,$BF,$CF,$9D,$FF,$D5,$5F,$F3,$F6,$CE,$5F,$88,$7C,$5E,$8B
	.dc.b	$8F,$86,$BF,$CA,$B4,$F6,$FF,$C2,$91,$86,$A8,$D4,$FF,$8A,$5B,$1F
	.dc.b	$E3,$2F,$A1,$FE,$71,$3F,$E2,$D6,$C6,$7B,$39,$C6,$23,$FC,$3A,$75
	.dc.b	$F3,$FB,$DF,$F2,$9B,$ED,$18,$3C,$7F,$3F,$FA,$CA,$FF,$20,$B5,$E6
	.dc.b	$D7,$FC,$52,$D7,$FC,$E7,$CA,$F7,$7D,$AD,$BF,$F2,$23,$67,$FC,$9A
	.dc.b	$D7,$CF,$F2,$DF,$D2,$5B,$EA,$FC,$68,$C4,$7F,$8E,$1A,$FF,$A0,$9F
	.dc.b	$F4,$7D,$7F,$94,$5B,$1F,$E5,$A6,$BD,$BE,$33,$E8,$FD,$1F,$F8,$B3
	.dc.b	$1F,$F2,$C3,$67,$FC,$C2,$BE,$71,$E6,$EB,$E2,$7F,$D8,$35,$FE,$4A
	.dc.b	$FA,$AF,$F5,$E3,$FE,$61,$6B,$FE,$6A,$6B,$F5,$1A,$FF,$94,$BE,$91
	.dc.b	$EB,$FF,$98,$1B,$1F,$E6,$B6,$BF,$E7,$46,$BF,$79,$7D,$07,$F9,$73
	.dc.b	$C2,$6F,$F1,$FE,$23,$0D,$BF,$CB,$0B,$E2,$E1,$AF,$7B,$F2,$11,$F2
	.dc.b	$78,$3D,$1F,$F5,$9F,$3F,$F4,$A7,$AB,$5F,$F7,$35,$27,$84,$78,$5B
	.dc.b	$3B,$C7,$D3,$3D,$3C,$FD,$DF,$18,$F3,$5F,$09,$91,$82,$1F,$EF,$5B
	.dc.b	$E9,$8F,$FA,$01,$B1,$FE,$CA,$BE,$8F,$9A,$F9,$BC,$6F,$C3,$12,$3D
	.dc.b	$B1,$AC,$3C,$F1,$EB,$3C,$2E,$2F,$BC,$1B,$7E,$09,$E2,$DC,$3C,$D7
	.dc.b	$D5,$FC,$36,$1E,$1C,$3C,$87,$88,$EF,$1A,$70,$68,$F6,$1A,$3E,$61
	.dc.b	$AB,$F0,$CF,$57,$D1,$5E,$19,$F8,$D3,$9A,$F2,$C7,$17,$C0,$F6,$3C
	.dc.b	$15,$C9,$F0,$1F,$14,$CC,$72,$78,$47,$89,$E8,$79,$8C,$C7,$68,$E9
	.dc.b	$1C,$8F,$29,$DB,$0F,$9E,$7C,$EA,$F4,$47,$39,$E2,$77,$7A,$99,$AF
	.dc.b	$24,$64,$8E,$09,$1E,$03,$89,$F7,$6F,$E4,$3E,$F9,$E2,$F5,$21,$99
	.dc.b	$F6,$1D,$1E,$47,$C8,$C3,$F2,$AF,$EF,$EF,$F8,$31,$1F,$D9,$90,$FE
	.dc.b	$0C,$E5,$A7,$37,$F5,$7E,$57,$C4,$8F,$11,$EB,$BF,$C9,$C6,$6B,$F2
	.dc.b	$93,$9B,$FE,$31,$39,$BD,$36,$7D,$0D,$7E,$2D,$23,$17,$F1,$27,$F8
	.dc.b	$94,$62,$FE,$1C,$7E,$D6,$30,$F8,$4C,$1E,$B3,$B9,$4E,$EF,$07,$DA
	.dc.b	$76,$35,$3E,$1A,$F2,$46,$2B,$C2,$EC,$F4,$DF,$B9,$D1,$E3,$35,$5C
	.dc.b	$E3,$51,$C0,$F2,$94,$F2,$9C,$8F,$2D,$78,$9D,$4F,$96,$FD,$53,$F4
	.dc.b	$A7,$BD,$F9,$A3,$04,$9D,$DF,$DE,$AF,$E7,$FF,$49,$FE,$10,$4D,$25
	.dc.b	$FF,$82,$20,$48,$FE,$3B,$87,$F1,$E6,$CE,$A7,$9B,$FC,$F2,$3F,$C8
	.dc.b	$F1,$FD,$3B,$EA,$BF,$D5,$2D,$F6,$FF,$A0,$C6,$21,$7D,$7C,$BE,$47
	.dc.b	$8E,$3C,$4D,$7D,$2D,$BF,$F4,$3C,$6D,$7D,$2E,$49,$19,$78,$78,$3E
	.dc.b	$76,$47,$C6,$8F,$51,$1C,$1F,$2B,$C2,$F7,$7C,$1F,$F5,$8F,$8A,$93
	.dc.b	$F3,$A3,$2D,$21,$EE,$D7,$E4,$63,$FE,$DB,$8F,$AB,$E3,$FC,$92,$3F
	.dc.b	$63,$E6,$BE,$61,$87,$91,$F5,$63,$2F,$F7,$CE,$BF,$F7,$73,$7E,$13
	.dc.b	$FB,$7C,$62,$38,$9F,$E1,$DF,$07,$26,$A6,$47,$38,$E3,$F0,$BA,$8C
	.dc.b	$1D,$DE,$BE,$0F,$F7,$CF,$09,$F4,$B4,$CD,$79,$7E,$56,$BE,$2C,$97
	.dc.b	$E6,$9F,$3F,$D2,$FF,$81,$BA,$9F,$1F,$4C,$E1,$EF,$F0,$2B,$CE,$BD
	.dc.b	$F1,$F0,$3D,$EF,$F0,$7F,$D5,$BA,$29,$9B,$F9,$1F,$4D,$5F,$25,$7C
	.dc.b	$88,$91,$A2,$1E,$9F,$8E,$FF,$0D,$FD,$37,$F9,$C3,$BB,$FC,$06,$DE
	.dc.b	$9F,$D3,$36,$BF,$5A,$FF,$93,$1F,$F1,$9B,$5F,$AA,$ED,$34,$D4,$F6
	.dc.b	$F0,$FF,$C5,$7F,$4D,$FA,$0A,$FF,$1E,$B5,$E7,$1E,$62,$3E,$07,$D3
	.dc.b	$F0,$5E,$AF,$82,$93,$E4,$7D,$BE,$9C,$8C,$10,$F3,$BF,$0B,$8B,$51
	.dc.b	$83,$DA,$FF,$90,$3E,$D3,$FD,$49,$1F,$E5,$27,$FD,$58,$DF,$0B,$5E
	.dc.b	$AF,$F9,$1B,$ED,$1D,$D7,$BE,$5F,$F9,$6A,$72,$FF,$97,$3E,$93,$F3
	.dc.b	$DC,$1F,$E3,$47,$31,$FA,$06,$C6,$6B,$B7,$1F,$7E,$49,$7D,$63,$FC
	.dc.b	$AF,$F6,$8E,$0B,$E7,$CA,$FF,$35,$CE,$5F,$17,$85,$FE,$5E,$CD,$F3
	.dc.b	$54,$C4,$7F,$98,$97,$C3,$DF,$7A,$5F,$F9,$DA,$72,$DF,$1F,$87,$8F
	.dc.b	$23,$07,$C1,$F9,$BF,$07,$8B,$23,$07,$FA,$03,$ED,$3F,$6A,$DF,$0F
	.dc.b	$FC,$E7,$06,$F9,$1F,$E7,$6F,$B4,$FF,$2C,$8F,$8D,$C1,$9F,$03,$DD
	.dc.b	$E4,$FB,$FE,$7A,$65,$F4,$E3,$FC,$FF,$F6,$8D,$AF,$24,$AF,$F4,$9C
	.dc.b	$E5,$FF,$4A,$FA,$4F,$9C,$FF,$43,$AC,$69,$88,$F9,$C7,$CF,$F9,$19
	.dc.b	$5F,$83,$8F,$F4,$7F,$DA,$7A,$63,$FD,$42,$BE,$9F,$26,$B6,$DF,$C9
	.dc.b	$F0,$F9,$19,$7C,$23,$EE,$7A,$3F,$77,$92,$4F,$E4,$63,$FD,$3B,$F5
	.dc.b	$2F,$AA,$BE,$44,$AF,$F5,$8C,$E5,$BE,$C7,$02,$3A,$F8,$8E,$64,$70
	.dc.b	$72,$F9,$BE,$FC,$B2,$3A,$BE,$2E,$69,$1E,$06,$BF,$D0,$B0,$77,$7F
	.dc.b	$AD,$3E,$A9,$DF,$49,$5F,$EC,$88,$FE,$6E,$DE,$FE,$04,$7F,$D9,$BF
	.dc.b	$54,$FA,$1F,$F5,$FF,$A6,$FA,$4A,$4A,$FF,$69,$C7,$D3,$6F,$47,$02
	.dc.b	$3D,$3E,$AB,$FD,$93,$F4,$FF,$BA,$F6,$7F,$C2,$4D,$9F,$F0,$CB,$52
	.dc.b	$F2,$B8,$92,$3F,$DC,$71,$96,$F9,$1D,$88,$E0,$FF,$74,$FD,$57,$F3
	.dc.b	$59,$F8,$F1,$CB,$83,$CA,$79,$B7,$91,$D9,$AF,$F7,$29,$78,$BF,$E2
	.dc.b	$28,$7D,$53,$6F,$6C,$E6,$7F,$C6,$07,$B1,$C3,$AF,$4F,$FC,$51,$7D
	.dc.b	$1B,$E2,$BE,$AA,$92,$BF,$C8,$15,$FA,$D8,$FF,$17,$43,$E8,$69,$1F
	.dc.b	$F1,$93,$FE,$33,$F8,$FF,$F7,$BE,$1F,$EE,$29,$FB,$DC,$3F,$E4,$A8
	.dc.b	$C4,$EA,$3D,$3C,$96,$BF,$C9,$4D,$7F,$89,$7C,$1D,$79,$9F,$E1,$8F
	.dc.b	$0D,$19,$E6,$D7,$8D,$70,$F0,$1F,$19,$CB,$8B,$FF,$2C,$79,$98,$7F
	.dc.b	$C4,$61,$F0,$1E,$3C,$1F,$1F,$8A,$E5,$5F,$C9,$E3,$D1,$A7,$F3,$E9
	.dc.b	$CB,$FA,$A6,$F3,$EB,$9A,$47,$2C,$1E,$16,$31,$F7,$3F,$0F,$EA,$9C
	.dc.b	$17,$83,$0D,$C3,$C7,$3F,$1B,$34,$7F,$9D,$71,$FE,$60,$EA,$FD,$9E
	.dc.b	$0F,$5F,$7F,$B7,$BF,$6B,$9B,$E1,$F8,$89,$18,$3B,$BE,$5F,$32,$47
	.dc.b	$8B,$E8,$2E,$FF,$E4,$ED,$5E,$6F,$E1,$63,$FD,$1B,$18,$AF,$F3,$F3
	.dc.b	$63,$FD,$24,$D7,$F3,$EF,$FB,$26,$37,$F0,$FC,$E5,$39,$78,$C6,$7E
	.dc.b	$4F,$F9,$95,$7F,$A2,$DB,$3F,$E9,$F6,$BF,$AC,$6B,$FE,$9D,$FA,$9F
	.dc.b	$E1,$13,$FE,$93,$6C,$7F,$A9,$DA,$FE,$D1,$AF,$FA,$97,$EA,$BF,$AD
	.dc.b	$4F,$FA,$6D,$B1,$FE,$AF,$6B,$FB,$C6,$BE,$1F,$1F,$DD,$5F,$CA,$EB
	.dc.b	$C5,$18,$3E,$17,$D5,$38,$29,$27,$FD,$55,$19,$7F,$D5,$9E,$7F,$9B
	.dc.b	$0F,$F5,$FB,$53,$4C,$47,$F8,$C8,$FF,$5A,$16,$3F,$D9,$0D,$7F,$8A
	.dc.b	$B5,$FF,$63,$2F,$AA,$FB,$0D,$39,$B8,$3E,$E7,$D5,$7E,$92,$3F,$D8
	.dc.b	$4D,$9F,$F6,$B3,$5F,$E4,$6D,$7F,$DA,$9F,$55,$FE,$37,$9F,$F6,$53
	.dc.b	$63,$FD,$BC,$D7,$F9,$5B,$5E,$3F,$27,$F8,$D8,$34,$93,$FE,$D5,$FA
	.dc.b	$9C,$B6,$3F,$29,$18,$AF,$F7,$63,$5F,$E7,$3E,$27,$F7,$0A,$FF,$6F
	.dc.b	$36,$3F,$DE,$4D,$77,$E8,$FE,$7F,$EA,$5E,$DD,$25,$FF,$B9,$E3,$2F
	.dc.b	$FB,$A7,$9B,$F9,$4D,$FF,$48,$6B,$FE,$F9,$6B,$FF,$00,$1A,$FF,$10
	.dc.b	$F1,$3F,$5D,$E0,$7F,$25,$AF,$F7,$A3,$67,$FA,$99,$B8,$CF,$CB,$70
	.dc.b	$23,$83,$FD,$F9,$F2,$0F,$5F,$F7,$4D,$8C,$C6,$0F,$DF,$7C,$FB,$DF
	.dc.b	$A4,$9F,$0E,$09,$D4,$63,$EA,$7C,$D3,$FD,$A1,$A8,$8E,$1B,$9F,$F0
	.dc.b	$D8,$FE,$E5,$3C,$FD,$2C,$11,$ED,$F1,$3B,$23,$83,$C5,$D5,$CC,$8E
	.dc.b	$0F,$37,$83,$DC,$8E,$0F,$57,$85,$E3,$23,$83,$DD,$E0,$F9,$C8,$E0
	.dc.b	$F8,$5F,$53,$EB,$23,$83,$E4,$FD,$4F,$C8,$D7,$ED,$B5,$FF,$07,$B5
	.dc.b	$F9,$01,$C3,$F2,$FE,$89,$F5,$3F,$CD,$5B,$5F,$59,$B3,$F8,$86,$BF
	.dc.b	$E1,$CF,$A8,$F9,$EF,$E3,$1A,$E7,$6F,$E8,$11,$88,$F1,$AF,$A8,$C9
	.dc.b	$2F,$F0,$B1,$97,$C5,$F1,$9E,$52,$38,$3F,$17,$C7,$FD,$A2,$FF,$3A
	.dc.b	$D7,$FC,$62,$D7,$E1,$B5,$F8,$5F,$2B,$F0,$49,$3F,$05,$FE,$B1,$F2
	.dc.b	$4F,$A7,$A4,$9F,$CD,$C6,$5F,$4F,$27,$C0,$23,$83,$E6,$7C,$37,$CF
	.dc.b	$23,$83,$F4,$BF,$58,$EB,$FE,$D0,$D8,$CC,$60,$F9,$BE,$7C,$8F,$92
	.dc.b	$DC,$FF,$7F,$F7,$8F,$3F,$FB,$93,$6B,$F6,$6D,$9D,$CF,$F7,$7F,$A6
	.dc.b	$FF,$15,$BF,$B8,$6C,$7F,$0A,$7F,$C7,$BE,$0E,$7D,$AE,$D3,$FC,$3B
	.dc.b	$E7,$FF,$CC,$EF,$F0,$06,$B9,$F8,$2F,$E0,$6A,$32,$FF,$9A,$3E,$8D
	.dc.b	$F2,$FA,$6C,$7F,$0A,$6B,$FC,$71,$AF,$F9,$C3,$E9,$3F,$D3,$4F,$F1
	.dc.b	$06,$BF,$C8,$9A,$EF,$DA,$ED,$F4,$BF,$A8,$F7,$7F,$C6,$B3,$FC,$46
	.dc.b	$D3,$9E,$0F,$D6,$C6,$27,$FC,$9E,$BC,$4D,$47,$17,$D4,$47,$D4,$FB
	.dc.b	$7C,$3D,$C2,$5E,$56,$1B,$1D,$A7,$06,$66,$46,$48,$31,$F5,$BD,$63
	.dc.b	$6E,$70,$F5,$D7,$48,$F5,$CE,$AB,$8D,$AB,$55,$C4,$F8,$D5,$BB,$E7
	.dc.b	$1E,$2B,$E3,$39,$8F,$99,$5C,$A7,$31,$E0,$38,$57,$27,$B1,$E9,$34
	.dc.b	$62,$FC,$8F,$03,$C6,$6A,$7D,$35,$DE,$BE,$CD,$F8,$EB,$E8,$4E,$5E
	.dc.b	$27,$58,$C3,$A3,$A4,$F2,$7C,$06,$4E,$06,$8F,$79,$AB,$FB,$6E,$8F
	.dc.b	$55,$76,$8C,$DF,$C4,$9C,$9C,$6F,$D1,$3F,$02,$32,$73,$9E,$07,$C0
	.dc.b	$78,$1E,$33,$CC,$E5,$E1,$7E,$B7,$9C,$64,$E7,$19,$70,$FD,$57,$31
	.dc.b	$F2,$1C,$BF,$00,$EA,$66,$38,$11,$C1,$0F,$2B,$97,$F0,$07,$23,$27
	.dc.b	$53,$F8,$45,$FE,$AD,$FE,$0C,$D7,$F8,$33,$5D,$9A,$3E,$A3,$FA,$FF
	.dc.b	$13,$E9,$3F,$4D,$B1,$F6,$5A,$FC,$69,$FE,$07,$18,$AF,$83,$3F,$BB
	.dc.b	$8C,$4F,$AE,$D1,$89,$F8,$D5,$FA,$E8,$C4,$7B,$8E,$47,$98,$F1,$14
	.dc.b	$F1,$1F,$C5,$DE,$33,$F8,$59,$E9,$3B,$7C,$B1,$E2,$8C,$57,$96,$B8
	.dc.b	$DF,$C3,$9C,$9E,$69,$F7,$B8,$3C,$D5,$CD,$F1,$98,$9E,$4E,$23,$C9
	.dc.b	$3D,$E3,$94,$61,$F5,$18,$9E,$37,$E8,$7E,$A4,$64,$FE,$63,$7E,$86
	.dc.b	$99,$3F,$83,$3F,$CB,$BE,$8F,$F9,$1D,$7F,$24,$6E,$1F,$E2,$86,$D7
	.dc.b	$F9,$45,$FF,$13,$BF,$E5,$39,$E1,$E2,$73,$A9,$39,$79,$47,$7F,$2B
	.dc.b	$FA,$FD,$FF,$99,$6B,$FC,$C6,$D7,$FC,$EA,$D7,$FC,$72,$1F,$F0,$93
	.dc.b	$D4,$F2,$47,$57,$06,$FB,$BD,$DF,$45,$EF,$FA,$49,$3A,$B4,$EA,$3C
	.dc.b	$6F,$98,$8E,$0E,$0E,$DF,$E6,$70,$F0,$7E,$67,$B1,$EE,$24,$FB,$9F
	.dc.b	$83,$F0,$3A,$8F,$14,$7D,$38,$FA,$9F,$01,$27,$CC,$FD,$7F,$52,$47
	.dc.b	$C8,$7E,$59,$1F,$F4,$6F,$C3,$3E,$83,$FB,$5A,$F9,$FA,$62,$24,$F0
	.dc.b	$7F,$C2,$73,$C6,$30,$74,$F8,$FC,$91,$C1,$E0,$F8,$EE,$24,$70,$79
	.dc.b	$3B,$FA,$91,$C1,$E8,$F9,$6F,$09,$1C,$1E,$CF,$87,$F2,$91,$C1,$F0
	.dc.b	$3C,$8F,$49,$1C,$1F,$13,$CC,$F6,$91,$C1,$0F,$ED,$B9,$23,$E4,$9F
	.dc.b	$2B,$E7,$FF,$C2,$BF,$47,$F9,$4E,$BF,$F0,$5B,$7A,$FF,$D1,$ED,$C3
	.dc.b	$FD,$70,$BC,$7E,$27,$F9,$01,$C4,$7F,$B1,$1A,$FF,$B5,$9A,$FE,$89
	.dc.b	$7B,$3C,$49,$1C,$5E,$9E,$A7,$99,$91,$FA,$49,$4F,$AB,$D8,$F5,$FA
	.dc.b	$32,$B5,$19,$7C,$8F,$9C,$8F,$9F,$E2,$7D,$44,$70,$7F,$B7,$BE,$89
	.dc.b	$F3,$1F,$E5,$E6,$C6,$63,$07,$F9,$8B,$E8,$BF,$21,$1F,$EF,$66,$BF
	.dc.b	$C0,$9A,$FF,$AF,$1A,$FF,$B9,$7D,$EF,$CF,$3F,$EE,$E6,$BF,$F0,$01
	.dc.b	$AF,$F0,$C6,$BC,$F8,$3C,$44,$7F,$DE,$F2,$8F,$F3,$BF,$D1,$33,$A4
	.dc.b	$AF,$F7,$EC,$65,$F1,$74,$FF,$9F,$B0,$34,$91,$FC,$78,$FF,$7F,$7C
	.dc.b	$7B,$E8,$BF,$D0,$AD,$8D,$07,$0F,$16,$BF,$CA,$17,$77,$22,$47,$28
	.dc.b	$CB,$F1,$FE,$43,$CC,$47,$07,$83,$E8,$9F,$45,$FE,$98,$6C,$66,$30
	.dc.b	$7F,$A7,$3E,$8B,$F2,$F1,$E8,$6B,$EF,$6B,$FC,$F9,$AF,$FB,$E3,$3F
	.dc.b	$F5,$1B,$7B,$FE,$51,$1F,$87,$F4,$4F,$13,$FD,$56,$D8,$CC,$60,$FF
	.dc.b	$57,$F9,$8F,$A2,$A4,$9F,$9B,$19,$7F,$15,$1F,$64,$8E,$0F,$A7,$F4
	.dc.b	$4F,$97,$FF,$5D,$B6,$33,$18,$33,$F1,$52,$38,$B5,$FC,$B7,$8D,$F8
	.dc.b	$72,$3F,$9B,$7F,$D8,$FF,$42,$FA,$2A,$49,$FC,$2C,$65,$BE,$6F,$FA
	.dc.b	$B2,$0D,$31,$E6,$E7,$E1,$2D,$7F,$6A,$A4,$E6,$31,$B7,$3E,$19,$18
	.dc.b	$21,$FE,$96,$AF,$A4,$D9,$F1,$73,$FF,$AF,$BE,$89,$CF,$49,$5F,$DE
	.dc.b	$A1,$E6,$EC,$91,$97,$FD,$8F,$F4,$4F,$98,$FF,$64,$F8,$DF,$ED,$EF
	.dc.b	$2F,$FD,$7E,$DF,$2D,$84,$8C,$37,$E7,$38,$91,$FF,$05,$C2,$3B,$FD
	.dc.b	$17,$FB,$5B,$E7,$FF,$8B,$E1,$FE,$CC,$6E,$1D,$B0,$46,$1F,$0B,$FC
	.dc.b	$09,$FF,$0B,$B5,$FF,$5E,$B6,$7A,$78,$7D,$FC,$29,$5C,$1A,$FF,$B1
	.dc.b	$3E,$8B,$FA,$83,$E0,$E0,$F3,$11,$EF,$BF,$A4,$8E,$0F,$1F,$81,$FE
	.dc.b	$E7,$5F,$45,$FD,$21,$F4,$C6,$5F,$F1,$AE,$68,$C4,$7C,$18,$C6,$7F
	.dc.b	$C6,$1B,$1F,$C7,$D7,$D1,$3E,$8B,$4D,$9D,$35,$EE,$D7,$F9,$47,$CC
	.dc.b	$FF,$24,$7E,$EB,$53,$4D,$7A,$6E,$0F,$F2,$3C,$BE,$1B,$0F,$FB,$B8
	.dc.b	$FE,$F1,$1A,$8C,$E5,$A9,$FC,$5C,$62,$7F,$01,$EC,$3D,$7C,$57,$E0
	.dc.b	$EA,$A9,$D5,$78,$B9,$73,$3F,$C2,$E3,$07,$9B,$C4,$C9,$AF,$5A,$C7
	.dc.b	$A9,$C3,$FE,$62,$B7,$C5,$62,$D9,$FA,$67,$35,$FE,$68,$8F,$E9,$8D
	.dc.b	$E4,$F3,$C6,$2B,$8C,$FE,$76,$3F,$72,$EC,$F0,$78,$3C,$2B,$DF,$7F
	.dc.b	$C2,$A7,$FC,$EB,$F4,$5F,$33,$0E,$67,$ED,$1F,$A5,$1A,$DF,$E3,$C6
	.dc.b	$27,$61,$C6,$3E,$BF,$03,$7B,$F8,$1E,$3E,$B4,$C7,$C3,$EB,$DC,$79
	.dc.b	$E2,$F8,$FE,$E6,$72,$FB,$D1,$9B,$F4,$FC,$BF,$90,$F2,$FC,$4D,$8F
	.dc.b	$6F,$99,$E8,$F8,$9F,$4B,$63,$E0,$FC,$47,$E0,$BF,$1B,$7A,$5E,$E8
	.dc.b	$C3,$F8,$6F,$90,$FE,$7D,$5F,$E9,$C7,$F9,$CB,$7D,$0D,$7A,$1F,$8C
	.dc.b	$FA,$2F,$F0,$6C,$FF,$A8,$9F,$B1,$BF,$E5,$DC,$F0,$FC,$A7,$33,$FC
	.dc.b	$5F,$E8,$FF,$B2,$3F,$EA,$A5,$E9,$FE,$61,$CD,$7D,$27,$33,$DF,$CD
	.dc.b	$F1,$79,$9E,$36,$C7,$FA,$D3,$E8,$DC,$FF,$A2,$5D,$16,$30,$FE,$9B
	.dc.b	$5F,$3E,$BF,$AF,$7F,$1F,$9F,$E0,$9C,$E1,$FC,$BD,$CC,$FE,$D6,$3F
	.dc.b	$D4,$47,$FB,$10,$DC,$FF,$42,$72,$FE,$31,$CB,$FC,$17,$E8,$FF,$6F
	.dc.b	$0F,$85,$B1,$87,$F7,$7F,$47,$F9,$DA,$FC,$AA,$F9,$7C,$F8,$5A,$7F
	.dc.b	$80,$FD,$1F,$F9,$27,$1F,$F6,$BB,$F1,$1B,$F0,$5A,$DB,$7F,$47,$FC
	.dc.b	$23,$CB,$FE,$31,$5C,$23,$0F,$F0,$CF,$A3,$47,$F0,$EF,$AD,$7C,$9E
	.dc.b	$63,$0F,$3F,$6F,$F8,$9F,$D1,$BB,$71,$8F,$FB,$AA,$30,$DF,$A4,$DF
	.dc.b	$67,$06,$C7,$FB,$C3,$C2,$FD,$9B,$FD,$85,$FE,$35,$C9,$FB,$87,$2F
	.dc.b	$F7,$27,$3C,$FF,$CA,$FC,$AF,$F4,$ED,$7F,$BE,$9F,$F4,$BB,$7C,$EF
	.dc.b	$E5,$31,$8B,$FE,$06,$BC,$5C,$F8,$9A,$6D,$FF,$80,$30,$FE,$71,$D4
	.dc.b	$F7,$FF,$99,$BF,$E6,$69,$D1,$FD,$E5,$D3,$8C,$7F,$C1,$31,$B7,$F8
	.dc.b	$9F,$CF,$7E,$EA,$75,$0F,$A4,$F6,$C7,$F8,$0E,$75,$3F,$C6,$A7,$2F
	.dc.b	$F8,$51,$F8,$2E,$7D,$EF,$E8,$F1,$8C,$7F,$A8,$3F,$05,$FE,$A5,$F4
	.dc.b	$2F,$AC,$FE,$9A,$DC,$7C,$75,$B7,$FC,$29,$19,$C3,$C5,$D7,$D8,$EF
	.dc.b	$D9,$6B,$8C,$74,$E9,$C5,$7F,$32,$BF,$C7,$D7,$6F,$33,$CF,$E6,$FA
	.dc.b	$1B,$1F,$13,$EB,$7F,$33,$7F,$CE,$9B,$AF,$B2,$FF,$B2,$C6,$2B,$FC
	.dc.b	$55,$A3,$E6,$FE,$19,$F1,$39,$2C,$FB,$B7,$FE,$E9,$97,$06,$BF,$35
	.dc.b	$AF,$D5,$6F,$CA,$7F,$72,$8F,$F5,$93,$5F,$F1,$D0,$70,$FF,$8F,$5A
	.dc.b	$FF,$93,$3E,$95,$F5,$7D,$9B,$7F,$EB,$86,$CF,$FA,$FD,$AF,$6A,$FE
	.dc.b	$E6,$F8,$BD,$9F,$29,$2B,$97,$CE,$F5,$23,$83,$FD,$8B,$5F,$8E,$AE
	.dc.b	$9E,$9F,$B8,$93,$E3,$F9,$CF,$31,$1C,$1C,$AF,$E2,$91,$C1,$FE,$D2
	.dc.b	$E3,$7D,$0F,$CC,$6C,$66,$30,$7C,$3F,$8C,$91,$F0,$DB,$8F,$FB,$86
	.dc.b	$BF,$91,$37,$9F,$A1,$2B,$C1,$EC,$F7,$23,$83,$FD,$CB,$F4,$3F,$CF
	.dc.b	$6B,$FD,$D8,$D7,$ED,$B5,$FF,$3B,$35,$FF,$42,$FD,$1F,$F8,$A3,$C7
	.dc.b	$FF,$77,$B6,$3F,$DE,$CF,$F8,$F7,$C2,$CF,$0F,$F2,$99,$FF,$7D,$7D
	.dc.b	$CF,$C4,$7F,$DE,$ED,$73,$B7,$F9,$5E,$31,$1F,$8E,$FA,$4F,$F3,$1B
	.dc.b	$FE,$FE,$6B,$F9,$20,$E1,$E1,$F0,$BF,$E6,$18,$FF,$6C,$46,$DA,$F5
	.dc.b	$F9,$2D,$1F,$E6,$7F,$19,$D5,$49,$3D,$CF,$E5,$8F,$F9,$B7,$EB,$BF
	.dc.b	$45,$1F,$C7,$9A,$FF,$8D,$1B,$E0,$7F,$99,$70,$79,$5F,$E4,$7E,$2F
	.dc.b	$F5,$24,$62,$FD,$D3,$E9,$23,$E1,$FA,$4F,$E0,$0F,$F2,$C7,$CE,$DC
	.dc.b	$7A,$3F,$EB,$C8,$DB,$FE,$46,$6E,$1F,$E7,$77,$8C,$66,$78,$A8,$DE
	.dc.b	$1F,$EB,$96,$C7,$59,$DC,$72,$AE,$9D,$F8,$33,$32,$BD,$11,$97,$D0
	.dc.b	$43,$CC,$E5,$F2,$19,$9E,$AF,$23,$17,$E5,$AE,$13,$A9,$E2,$47,$04
	.dc.b	$9E,$11,$C8,$F9,$15,$CA,$33,$1C,$5F,$85,$19,$8F,$84,$71,$3E,$43
	.dc.b	$97,$E5,$57,$C1,$7D,$A7,$A4,$E8,$72,$32,$4A,$FA,$D1,$C5,$F4,$47
	.dc.b	$98,$F0,$5F,$D1,$BF,$2C,$75,$9E,$2E,$AF,$C9,$5B,$8F,$04,$F4,$7D
	.dc.b	$27,$4B,$DD,$78,$6F,$E1,$D6,$EF,$F0,$77,$EE,$3B,$1E,$19,$D9,$1F
	.dc.b	$41,$ED,$70,$F1,$30,$6A,$7D,$2F,$C6,$3C,$A6,$8E,$0E,$8F,$25,$7F
	.dc.b	$B8,$6B,$F4,$33,$FD,$61,$AF,$C6,$3F,$85,$38,$7C,$A6,$E3,$91,$1F
	.dc.b	$61,$E5,$A3,$3E,$E1,$C5,$F2,$23,$62,$32,$E2,$FE,$49,$F2,$A3,$31
	.dc.b	$ED,$8C,$DF,$ED,$27,$38,$7B,$AB,$AC,$63,$0F,$64,$E6,$7F,$D8,$B3
	.dc.b	$97,$D0,$43,$E5,$C7,$FB,$E6,$35,$18,$9E,$64,$7D,$D8,$7B,$5F,$A5
	.dc.b	$3C,$30,$F3,$C7,$92,$75,$7F,$8D,$AD,$61,$8C,$3E,$4D,$FC,$5B,$C3
	.dc.b	$E5,$AD,$BF,$1E,$BE,$14,$EA,$3C,$C7,$90,$D1,$E1,$29,$E6,$78,$61
	.dc.b	$F3,$DB,$3D,$27,$78,$79,$E7,$31,$FE,$5A,$BF,$1B,$F5,$6B,$83,$4F
	.dc.b	$01,$4F,$01,$FE,$62,$FA,$DF,$F3,$A3,$FE,$64,$FA,$5F,$F4,$9B,$7D
	.dc.b	$2F,$E6,$33,$A8,$FF,$7E,$B7,$16,$7F,$22,$F8,$D2,$72,$FE,$52,$33
	.dc.b	$18,$85,$73,$AE,$5D,$BF,$D7,$65,$68,$C3,$63,$FB,$13,$5F,$EE,$4D
	.dc.b	$BE,$CD,$AE,$B3,$FE,$82,$E9,$FF,$00,$56,$BB,$BF,$A7,$D9,$A6,$3E
	.dc.b	$96,$DF,$C7,$8A,$FB,$F0,$F8,$53,$B8,$F2,$BE,$82,$3E,$8F,$67,$BF
	.dc.b	$B9,$23,$E9,$47,$88,$F7,$91,$E0,$7C,$0F,$57,$E1,$91,$C1,$FE,$05
	.dc.b	$FA,$F7,$D3,$FB,$26,$4F,$FA,$8A,$72,$FF,$A9,$3E,$39,$F4,$DF,$45
	.dc.b	$B1,$FE,$0F,$6B,$FE,$1F,$6B,$FE,$72,$73,$B7,$DD,$6E,$9F,$AC,$6C
	.dc.b	$FF,$16,$5F,$2F,$FD,$FB,$A9,$F1,$DF,$E0,$09,$CD,$79,$6B,$53,$9F
	.dc.b	$93,$D7,$CA,$CA,$F5,$46,$5A,$43,$EB,$1F,$C9,$B0,$7C,$27,$7F,$76
	.dc.b	$4F,$C3,$7C,$5F,$21,$FE,$12,$EF,$7D,$46,$2B,$E4,$4E,$63,$D1,$E8
	.dc.b	$7A,$7D,$C9,$5F,$36,$32,$D2,$1E,$41,$BB,$E8,$FE,$11,$C2,$35,$1F
	.dc.b	$E3,$AE,$1F,$99,$3F,$E2,$6F,$A7,$FE,$4E,$7F,$A9,$38,$FF,$5C,$D9
	.dc.b	$FE,$3A,$EB,$C3,$E0,$4C,$3F,$DC,$BF,$33,$FD,$CD,$4B,$E0,$7B,$12
	.dc.b	$3F,$DD,$70,$F7,$FC,$46,$63,$D9,$1C,$5D,$C7,$9F,$D3,$F4,$7B,$F2
	.dc.b	$79,$C6,$5A,$43,$9F,$C5,$7B,$C8,$F8,$9F,$F2,$3F,$C1,$3E,$87,$FC
	.dc.b	$93,$F4,$EF,$A7,$A4,$BF,$EC,$FC,$1F,$C4,$E5,$F2,$7F,$E5,$0F,$A7
	.dc.b	$FC,$BD,$1A,$4F,$03,$BF,$D4,$49,$E7,$C9,$FD,$2B,$44,$F9,$3B,$7A
	.dc.b	$73,$4B,$D9,$FD,$E3,$1F,$C0,$38,$3C,$1F,$25,$FE,$62,$FA,$87,$1D
	.dc.b	$24,$EE,$1C,$7F,$0B,$59,$3C,$F4,$FD,$5B,$F8,$F5,$D7,$FE,$6E,$FA
	.dc.b	$07,$D4,$52,$57,$8E,$3F,$83,$BF,$97,$5E,$47,$12,$47,$A7,$EA,$3F
	.dc.b	$D6,$0F,$7F,$1F,$EA,$91,$F2,$78,$1F,$E7,$E9,$FB,$11,$C7,$A7,$06
	.dc.b	$78,$FF,$D0,$7F,$51,$F8,$B9,$FD,$43,$7A,$7F,$A6,$37,$9F,$FB,$DB
	.dc.b	$F9,$59,$F8,$5E,$3F,$E1,$88,$FC,$5E,$7F,$F4,$7F,$D4,$3E,$7A,$92
	.dc.b	$7E,$A4,$7C,$07,$EB,$7C,$3F,$FA,$F3,$2F,$EB,$8D,$CB,$FC,$16,$D9
	.dc.b	$FF,$1A,$BF,$EA,$D7,$E9,$7B,$DF,$98,$23,$F5,$FA,$BF,$D4,$3F,$50
	.dc.b	$FA,$1A,$49,$FC,$34,$7B,$DF,$F2,$0A,$F9,$3E,$24,$8D,$FC,$0F,$E9
	.dc.b	$BC,$E9,$1F,$85,$7F,$53,$F3,$9F,$EB,$0F,$A8,$FD,$1C,$FE,$4D,$AF
	.dc.b	$F1,$76,$ED,$FD,$ED,$7A,$FC,$7D,$7E,$4C,$AE,$51,$D7,$E3,$35,$18
	.dc.b	$3B,$3E,$1E,$3F,$17,$81,$27,$C5,$19,$69,$0F,$27,$6F,$FB,$23,$EA
	.dc.b	$3F,$A5,$C7,$EA,$DA,$FF,$2F,$6E,$7F,$E1,$35,$EB,$71,$24,$F1,$7F
	.dc.b	$DA,$7F,$50,$E4,$FF,$6A,$F9,$CF,$A4,$A4,$BF,$DD,$43,$E9,$76,$48
	.dc.b	$CB,$D3,$8B,$FD,$BF,$F4,$FF,$C4,$09,$ED,$E3,$C9,$97,$FC,$16,$7F
	.dc.b	$77,$B7,$F8,$AD,$BD,$BF,$E5,$A5,$E3,$F4,$25,$71,$7F,$DD,$BE,$AF
	.dc.b	$F4,$58,$FF,$44,$CF,$C4,$AE,$5E,$47,$90,$F2,$F0,$CA,$F1,$35,$FE
	.dc.b	$B5,$4F,$0F,$FC,$85,$0F,$A8,$FD,$92,$9D,$67,$FE,$70,$8F,$45,$70
	.dc.b	$71,$3A,$74,$71,$8D,$3E,$FF,$8B,$7D,$46,$A3,$B4,$61,$E3,$48,$C4
	.dc.b	$76,$6B,$DB,$E8,$3F,$CA,$49,$A4,$7B,$7B,$5F,$E5,$EC,$7F,$CB,$91
	.dc.b	$F8,$3D,$7F,$9D,$3F,$EB,$26,$F2,$ED,$F8,$0D,$8F,$F5,$CF,$BB,$FE
	.dc.b	$02,$9D,$9F,$D0,$DC,$47,$F8,$1C,$FC,$64,$76,$F1,$79,$47,$FB,$0B
	.dc.b	$9A,$BF,$C2,$67,$47,$93,$DB,$CD,$69,$91,$DF,$E4,$2D,$7F,$85,$70
	.dc.b	$47,$6C,$BF,$1B,$5F,$D5,$DF,$F1,$CB,$79,$75,$C9,$23,$D5,$E0,$DC
	.dc.b	$7A,$3E,$8F,$A8,$F8,$D1,$FE,$23,$7E,$53,$77,$D6,$F2,$FE,$27,$81
	.dc.b	$6B,$CC,$F3,$F7,$24,$FF,$6B,$C3,$FB,$2B,$63,$C1,$F2,$92,$30,$6D
	.dc.b	$F2,$7C,$DC,$8E,$5F,$39,$71,$F3,$79,$7E,$9F,$2F,$D4,$DA,$F0,$7C
	.dc.b	$1F,$D4,$FA,$87,$A1,$EF,$8C,$3F,$63,$1F,$F2,$6D,$7F,$80,$5B,$8F
	.dc.b	$F9,$15,$1B,$F8,$4E,$1F,$3D,$F8,$0F,$C1,$3F,$D6,$77,$C6,$30,$FE
	.dc.b	$13,$EA,$1D,$57,$D9,$C3,$67,$FC,$21,$F6,$FF,$E0,$E6,$FB,$FF,$8C
	.dc.b	$8C,$69,$FE,$57,$7E,$5B,$78,$74,$6F,$7B,$7F,$E1,$BF,$4F,$F9,$6B
	.dc.b	$7C,$8F,$CB,$C7,$F0,$78,$C1,$FE,$21,$FB,$77,$9B,$F9,$CE,$66,$5F
	.dc.b	$9E,$9C,$D7,$F8,$9E,$72,$DF,$AE,$E0,$71,$F9,$BB,$1F,$E2,$DF,$AA
	.dc.b	$7A,$DF,$A8,$6C,$67,$BB,$13,$FA,$CF,$B7,$FC,$EE,$7F,$E7,$76,$F8
	.dc.b	$BF,$ED,$B7,$DB,$19,$F1,$38,$73,$71,$F8,$0B,$97,$F8,$F6,$FF,$D2
	.dc.b	$2F,$F8,$FF,$EA,$1F,$47,$FB,$A7,$F7,$F9,$7F,$B9,$DC,$CE,$FE,$3F
	.dc.b	$F7,$FF,$6F,$FE,$38,$8F,$F2,$43,$FE,$C0,$6F,$C3,$EB,$E4,$F7,$EF
	.dc.b	$C9,$DF,$D6,$5E,$0F,$F4,$97,$8F,$FE,$7F,$9F,$F2,$93,$63,$3E,$E7
	.dc.b	$F0,$E8,$C6,$BF,$EF,$BF,$B7,$FF,$41,$5F,$F9,$65,$B7,$FE,$9D,$5D
	.dc.b	$1F,$C5,$9C,$D7,$1F,$1F,$1D,$59,$6E,$1F,$E6,$3F,$C5,$DF,$4B,$FC
	.dc.b	$79,$7D,$BD,$8C,$3F,$C8,$BE,$B1,$F6,$7F,$CD,$35,$FC,$4C,$63,$0F
	.dc.b	$E8,$CD,$8F,$E9,$AD,$7F,$AA,$7D,$67,$E8,$1B,$F2,$7C,$D4,$65,$FC
	.dc.b	$A9,$88,$D4,$7E,$9A,$39,$F9,$38,$25,$FF,$57,$9F,$DE,$B7,$4E,$24
	.dc.b	$75,$AF,$42,$38,$39,$78,$FE,$02,$38,$3B,$7D,$1F,$90,$8E,$0F,$87
	.dc.b	$E3,$FB,$74,$FF,$1D,$C6,$09,$9F,$F7,$2F,$AB,$7D,$27,$C3,$6C,$FF
	.dc.b	$71,$6C,$EE,$7E,$47,$D5,$FE,$12,$3F,$C0,$0D,$7E,$4B,$5F,$DF,$B5
	.dc.b	$3F,$8A,$AF,$EF,$CD,$8F,$F0,$2B,$5E,$BC,$DE,$5E,$19,$5E,$56,$CF
	.dc.b	$E0,$3C,$6F,$80,$49,$FC,$2B,$ED,$FA,$09,$1E,$D6,$CF,$F8,$53,$B3
	.dc.b	$F2,$B3,$FE,$11,$6C,$67,$9B,$F1,$11,$88,$F8,$3C,$BF,$10,$95,$EC
	.dc.b	$7E,$64,$C8,$F9,$8D,$9F,$7F,$AF,$F3,$09,$3F,$11,$FC,$B7,$6F,$FA
	.dc.b	$FA,$3F,$C4,$8D,$9F,$F1,$5B,$5F,$F1,$93,$5F,$CD,$FD,$9F,$9D,$B8
	.dc.b	$FE,$86,$31,$1F,$CD,$1A,$BE,$B3,$FC,$0F,$3F,$E3,$36,$C7,$F8,$E9
	.dc.b	$78,$6F,$73,$5E,$FF,$FB,$22,$BF,$57,$F4,$AD,$3F,$C7,$90,$8C,$3F
	.dc.b	$E3,$F6,$BF,$E4,$66,$BF,$E4,$DF,$B7,$FE,$0E,$FF,$91,$5A,$FF,$93
	.dc.b	$1A,$EF,$BF,$BF,$D5,$3D,$1A,$49,$FF,$27,$46,$5F,$F2,$87,$D2,$7F
	.dc.b	$6E,$6F,$B1,$E8,$8C,$47,$F9,$69,$AF,$F9,$86,$5E,$E7,$80,$91,$D9
	.dc.b	$F8,$39,$7F,$AF,$E3,$E1,$7C,$27,$F1,$0F,$1D,$CB,$F2,$23,$0F,$F1
	.dc.b	$4C,$1E,$CE,$2F,$FB,$5C,$61,$F5,$C7,$C0,$23,$FE,$6B,$FA,$7E,$AF
	.dc.b	$F7,$2C,$2A,$FA,$6F,$F2,$2B,$9C,$3F,$91,$C6,$21,$3C,$2B,$8B,$6B
	.dc.b	$C9,$3C,$F8,$AD,$F1,$AD,$4F,$53,$FC,$07,$7C,$AF,$BD,$F0,$39,$C6
	.dc.b	$67,$73,$97,$91,$DC,$B1,$B4,$DB,$F1,$4F,$69,$DB,$D8,$E8,$FB,$9D
	.dc.b	$34,$F0,$9C,$0F,$0C,$74,$7F,$C3,$AD,$7C,$31,$D5,$DD,$76,$BF,$87
	.dc.b	$1D,$4D,$9A,$21,$9B,$F6,$1E,$B9,$E5,$5F,$D6,$23,$2E,$27,$C7,$3E
	.dc.b	$17,$94,$65,$F9,$06,$0F,$61,$E3,$BF,$51,$EB,$AE,$B1,$E4,$BF,$C4
	.dc.b	$CE,$6B,$B4,$62,$FB,$5F,$0B,$E9,$18,$AD,$9F,$3A,$B8,$4F,$89,$F8
	.dc.b	$A6,$1E,$26,$63,$E0,$1E,$BB,$F1,$46,$4F,$82,$79,$E3,$6E,$4F,$83
	.dc.b	$3D,$4E,$F1,$A8,$EE,$48,$CB,$F8,$23,$11,$E9,$38,$46,$A3,$AC,$65
	.dc.b	$E1,$1F,$80,$7F,$DD,$17,$FE,$13,$7F,$02,$7F,$B9,$9C,$12,$7F,$94
	.dc.b	$B6,$3F,$79,$1F,$29,$EA,$FE,$0F,$C0,$C4,$74,$AF,$25,$7F,$80,$23
	.dc.b	$E1,$3F,$56,$3F,$79,$3C,$23,$15,$CE,$B6,$E6,$74,$D8,$C9,$D2,$32
	.dc.b	$D3,$27,$E3,$CF,$CE,$3A,$71,$5E,$27,$6F,$A2,$34,$D8,$CC,$FD,$03
	.dc.b	$CC,$F2,$70,$FC,$43,$F8,$7C,$F2,$8C,$C7,$F0,$68,$C3,$B3,$07,$63
	.dc.b	$15,$F2,$5E,$C6,$27,$F8,$BC,$F2,$8F,$A1,$1E,$4A,$FA,$2F,$C7,$8E
	.dc.b	$26,$8E,$B3,$C6,$BF,$7E,$D9,$EB,$3E,$33,$81,$E6,$3F,$85,$7D,$4B
	.dc.b	$DF,$FE,$66,$D8,$FE,$4E,$D7,$F8,$67,$C0,$FF,$2E,$37,$32,$3F,$95
	.dc.b	$C7,$F3,$2F,$47,$F2,$24,$8F,$E9,$8F,$E5,$3E,$A3,$51,$FC,$CA,$3F
	.dc.b	$9C,$C7,$F3,$AF,$23,$FA,$2F,$13,$E1,$29,$9B,$FE,$73,$3F,$9E,$23
	.dc.b	$FD,$65,$FD,$0F,$C3,$EE,$3F,$92,$47,$F4,$68,$FE,$8F,$E2,$FE,$98
	.dc.b	$93,$FD,$85,$AF,$F2,$B7,$6F,$F4,$8F,$97,$FE,$B3,$DE,$F9,$4B,$F4
	.dc.b	$92,$BF,$A7,$CE,$5B,$3F,$E1,$C8,$C4,$7F,$A2,$8F,$F6,$FB,$87,$0D
	.dc.b	$7F,$D1,$FD,$6F,$7B,$FB,$4B,$63,$FA,$EB,$5F,$E8,$D2,$F8,$4F,$39
	.dc.b	$E1,$24,$78,$4F,$0B,$E5,$F9,$29,$19,$3F,$C7,$2E,$1E,$12,$F9,$AF
	.dc.b	$61,$23,$84,$65,$BE,$07,$C1,$23,$83,$7F,$45,$F1,$48,$E0,$E7,$F4
	.dc.b	$5F,$24,$8E,$0F,$37,$D4,$7F,$92,$13,$48,$FF,$93,$1F,$A1,$DB,$FE
	.dc.b	$1C,$6F,$5F,$FA,$F1,$B2,$39,$F9,$1F,$E1,$88,$FF,$2D,$47,$82,$3A
	.dc.b	$11,$F1,$F8,$7D,$C8,$E0,$F3,$FB,$DC,$E2,$47,$F8,$5A,$93,$DF,$DB
	.dc.b	$F0,$7A,$12,$B5,$19,$7E,$0B,$ED,$23,$F1,$3D,$AF,$80,$47,$04,$3F
	.dc.b	$C1,$96,$8F,$71,$EF,$F6,$BF,$C6,$FF,$B5,$FC,$4A,$F9,$B1,$97,$FC
	.dc.b	$5B,$19,$FC,$1D,$F0,$FF,$E3,$E6,$CF,$F8,$C9,$AF,$F8,$6D,$FC,$B3
	.dc.b	$F9,$AE,$0F,$CE,$FD,$1C,$8C,$1D,$B7,$F1,$FC,$94,$8C,$1D,$79,$BC
	.dc.b	$1C,$B2,$30,$7E,$9B,$E7,$7F,$52,$47,$F5,$3F,$1D,$EE,$F9,$69,$1E
	.dc.b	$C7,$E0,$FC,$B7,$AF,$39,$3E,$B7,$DD,$7F,$E5,$CF,$83,$F8,$C7,$E1
	.dc.b	$A7,$FC,$70,$DE,$9F,$F3,$5B,$F2,$F8,$7F,$CF,$C7,$E0,$27,$E5,$F3
	.dc.b	$FD,$0E,$8F,$E6,$B9,$FF,$36,$CE,$57,$F3,$68,$F0,$6F,$23,$FC,$BF
	.dc.b	$8A,$3C,$93,$F6,$BD,$F9,$5F,$61,$FB,$BE,$AF,$F4,$2E,$AF,$E8,$7D
	.dc.b	$52,$BF,$A1,$C7,$B7,$29,$1D,$9F,$81,$9F,$E1,$A6,$4F,$E1,$1F,$C4
	.dc.b	$D7,$F4,$EE,$89,$3F,$D3,$9F,$EA,$3F,$0F,$F2,$76,$F9,$59,$CA,$F5
	.dc.b	$C7,$E4,$BD,$9F,$C9,$F9,$52,$7F,$23,$1F,$E8,$CF,$A0,$7D,$45,$24
	.dc.b	$FF,$A0,$A1,$E6,$6F,$CC,$E0,$C9,$D4,$73,$F8,$0D,$46,$0E,$4F,$7F
	.dc.b	$37,$C1,$F2,$12,$7B,$46,$5A,$43,$FC,$F5,$83,$CE,$FF,$51,$FD,$0F
	.dc.b	$E9,$3A,$FF,$D5,$ED,$F1,$FF,$89,$5A,$30,$FF,$15,$7F,$52,$FF,$A1
	.dc.b	$5B,$3C,$BC,$BE,$A7,$FA,$C3,$E8,$5D,$FF,$EB,$2E,$3F,$C3,$61,$FE
	.dc.b	$A6,$6B,$FE,$8F,$6D,$FF,$B0,$97,$67,$12,$4F,$17,$B7,$83,$FE,$BE
	.dc.b	$F9,$DF,$EB,$F7,$FE,$CA,$6E,$9E,$C6,$DF,$F2,$F5,$F4,$3C,$49,$1F
	.dc.b	$EB,$A8,$CB,$7E,$73,$B1,$1C,$1F,$EC,$0F,$A1,$FE,$31,$3F,$22,$0D
	.dc.b	$26,$DE,$72,$47,$77,$D5,$E8,$7F,$B7,$21,$F4,$3F,$B7,$79,$C6,$5F
	.dc.b	$EF,$94,$8C,$47,$BD,$AF,$BD,$AF,$93,$E8,$1F,$43,$FD,$C6,$91,$89
	.dc.b	$FE,$F1,$18,$9D,$4E,$67,$FB,$C7,$D0,$FF,$63,$23,$FD,$EC,$37,$CD
	.dc.b	$FF,$7F,$EF,$FE,$FE,$9F,$E1,$1B,$FF,$BA,$14,$EA,$3F,$C3,$1E,$45
	.dc.b	$9F,$CA,$B5,$79,$8F,$53,$11,$D0,$FF,$6D,$46,$6B,$B7,$26,$BC,$9D
	.dc.b	$B6,$3C,$51,$CB,$C9,$C5,$FF,$B8,$C3,$99,$FF,$7F,$DB,$E0,$B1,$5E
	.dc.b	$07,$CD,$A3,$96,$FA,$58,$6E,$1F,$E0,$19,$3F,$B2,$7F,$AE,$BD,$BC
	.dc.b	$2D,$E1,$EC,$7E,$17,$D0,$FC,$89,$E4,$7E,$E9,$F8,$55,$D2,$BF,$C1
	.dc.b	$D1,$B7,$FD,$FC,$EA,$78,$6F,$F3,$73,$78,$3D,$A3,$0F,$83,$C6,$95
	.dc.b	$E0,$C3,$AF,$97,$87,$AF,$91,$E6,$31,$9F,$8E,$19,$FF,$9C,$6F,$94
	.dc.b	$63,$CB,$F6,$9E,$EF,$52,$CE,$A9,$E6,$7C,$2F,$84,$F8,$6D,$9D,$7A
	.dc.b	$1F,$87,$FA,$17,$C0,$7C,$88,$C3,$F8,$BC,$5F,$0F,$4C,$36,$FC,$AF
	.dc.b	$FA,$6D,$F3,$FD,$0F,$D8,$73,$F0,$7F,$95,$8C,$6C,$DF,$E0,$7D,$0F
	.dc.b	$E6,$2F,$DE,$FF,$9B,$DB,$CD,$F9,$D8,$C5,$71,$9F,$E6,$BF,$42,$F5
	.dc.b	$BF,$44,$B9,$6C,$61,$CF,$14,$AF,$9B,$4E,$0C,$61,$D4,$FF,$29,$7A
	.dc.b	$9F,$E7,$BA,$FA,$6D,$8D,$1F,$D7,$3C,$7F,$D9,$46,$75,$FB,$9F,$42
	.dc.b	$F3,$7F,$6D,$D1,$FE,$5D,$C7,$EE,$CE,$5F,$BF,$19,$C7,$AA,$F5,$F8
	.dc.b	$9C,$BD,$8B,$5F,$86,$FA,$17,$AF,$FC,$0D,$B1,$9E,$FC,$4F,$F0,$7F
	.dc.b	$A0,$FD,$5E,$5E,$A6,$EB,$F9,$37,$D5,$7F,$D8,$9C,$B7,$E0,$F9,$7D
	.dc.b	$0F,$36,$C7,$E6,$3E,$81,$F4,$3F,$C5,$5F,$E2,$18,$FC,$87,$33,$BF
	.dc.b	$37,$F8,$D7,$D0,$BE,$86,$FC,$6E,$1B,$5F,$A4,$97,$C0,$70,$38,$FC
	.dc.b	$C5,$9F,$D4,$F8,$FF,$C5,$9B,$E8,$7F,$27,$8F,$F3,$FC,$60,$FD,$7F
	.dc.b	$D0,$E3,$5F,$BA,$DE,$2F,$DB,$3F,$4E,$77,$EE,$FF,$31,$8F,$F3,$D4
	.dc.b	$72,$E9,$C7,$8F,$97,$F9,$B7,$D0,$BE,$66,$FC,$BE,$1B,$7F,$C0,$3E
	.dc.b	$7F,$E1,$37,$8B,$F9,$F4,$3E,$76,$C7,$F4,$1F,$A1,$FD,$05,$7F,$08
	.dc.b	$6E,$FF,$96,$7F,$4D,$19,$F2,$38,$1F,$D2,$7E,$7F,$F7,$55,$B8,$C7
	.dc.b	$9D,$D7,$7E,$DB,$DC,$3B,$3F,$D4,$7E,$C7,$F7,$BC,$7F,$C6,$DB,$3F
	.dc.b	$E2,$AC,$D5,$B8,$F4,$7D,$14,$7D,$1E,$55,$D3,$D1,$F6,$3F,$D9,$5F
	.dc.b	$F6,$26,$08,$C4,$23,$15,$FE,$E9,$6C,$7F,$BB,$F1,$FF,$20,$4F,$FB
	.dc.b	$B5,$B1,$9F,$8B,$E7,$18,$8F,$F2,$08,$FA,$3F,$CB,$BF,$EF,$96,$BA
	.dc.b	$6B,$FE,$48,$6B,$FE,$DB,$F9,$2F,$EC,$97,$FE,$FD,$6C,$76,$0E,$1F
	.dc.b	$0B,$5E,$9F,$5B,$F0,$28,$D2,$57,$5F,$5F,$E2,$DF,$F3,$E8,$C1,$2F
	.dc.b	$D5,$F3,$7C,$C9,$1E,$38,$F8,$11,$CA,$E4,$FA,$5B,$3E,$9A,$F9,$0F
	.dc.b	$F9,$A1,$AF,$C1,$CB,$C6,$7A,$7C,$09,$7D,$F2,$FD,$DB,$63,$EE,$F2
	.dc.b	$FE,$6A,$FE,$9B,$5F,$F3,$AB,$50,$70,$BC,$07,$CB,$FE,$21,$B7,$EF
	.dc.b	$6C,$FD,$36,$BE,$0F,$73,$C6,$49,$F8,$D7,$FD,$DE,$7E,$D7,$D1,$7F
	.dc.b	$88,$7B,$7E,$53,$63,$3E,$37,$FB,$55,$FF,$46,$C6,$5F,$C7,$FD,$8B
	.dc.b	$D0,$FF,$24,$36,$3E,$A3,$5F,$C4,$35,$FC,$BF,$D1,$BC,$2F,$F2,$83
	.dc.b	$63,$EE,$35,$FC,$83,$5F,$CF,$FC,$F7,$FA,$49,$BF,$05,$FE,$59,$8C
	.dc.b	$47,$FA,$91,$AF,$D0,$F9,$C7,$D3,$FF,$97,$DB,$1F,$8B,$6B,$F9,$F6
	.dc.b	$BF,$AE,$FA,$47,$85,$FE,$67,$6C,$7E,$4D,$AF,$E9,$DA,$FE,$DB,$EC
	.dc.b	$7F,$52,$FE,$65,$AF,$EB,$5A,$EF,$A7,$B7,$81,$FE,$4D,$E4,$FF,$5F
	.dc.b	$DF,$F8,$0E,$BF,$C6,$2D,$7F,$BF,$AF,$07,$84,$7E,$42,$35,$E1,$FF
	.dc.b	$8E,$DB,$5D,$2F,$74,$9E,$4F,$F8,$FB,$E9,$1F,$49,$A7,$FC,$89,$87
	.dc.b	$F9,$15,$B8,$7F,$B5,$1D,$D7,$1A,$E1,$32,$30,$43,$FC,$2F,$3C,$9F
	.dc.b	$D3,$B5,$3D,$1D,$CF,$37,$8C,$F9,$62,$47,$A6,$35,$5E,$79,$CC,$F0
	.dc.b	$34,$78,$E3,$4E,$0D,$1E,$03,$53,$EF,$3A,$9F,$06,$7A,$CF,$C5,$8C
	.dc.b	$CF,$B2,$73,$1C,$63,$B4,$F8,$1E,$C7,$81,$E4,$F8,$0E,$47,$08,$E4
	.dc.b	$F0,$8E,$47,$07,$94,$79,$5E,$47,$CC,$7C,$AF,$CE,$9F,$2B,$F4,$A3
	.dc.b	$35,$F4,$6B,$AC,$78,$1E,$C6,$27,$CA,$78,$0E,$47,$82,$79,$4F,$12
	.dc.b	$47,$09,$E4,$F9,$4F,$71,$98,$F5,$3D,$4D,$1F,$05,$CC,$79,$1E,$B1
	.dc.b	$ED,$3C,$4F,$B2,$BD,$F1,$CA,$78,$3D,$C8,$66,$79,$D7,$AA,$3D,$E7
	.dc.b	$F7,$47,$F2,$CD,$AF,$F7,$74,$62,$3D,$26,$4F,$EB,$8E,$5E,$67,$AA
	.dc.b	$63,$E2,$FA,$47,$C0,$FA,$2B,$C5,$3E,$D3,$FD,$03,$3A,$7E,$85,$FB
	.dc.b	$A3,$11,$F1,$CF,$6C,$F2,$8C,$E1,$F4,$A3,$A9,$9B,$F0,$5F,$36,$DF
	.dc.b	$E3,$2F,$8E,$1E,$88,$F2,$61,$F7,$27,$31,$E0,$BE,$4F,$98,$E5,$7D
	.dc.b	$1F,$30,$9E,$2F,$4B,$ED,$87,$96,$39,$1C,$27,$71,$EA,$BE,$F3,$D2
	.dc.b	$30,$FF,$B7,$27,$73,$E1,$C3,$FD,$C5,$18,$9E,$B1,$8C,$35,$87,$BE
	.dc.b	$B5,$5D,$6F,$F9,$6C,$62,$FE,$64,$62,$F8,$D7,$DD,$6B,$ED,$3F,$C1
	.dc.b	$35,$C6,$FD,$F5,$C6,$7C,$71,$E1,$AF,$E6,$DE,$0F,$F3,$7F,$0E,$57
	.dc.b	$F3,$B8,$CB,$FC,$F6,$32,$FF,$32,$8F,$E7,$DE,$1F,$E6,$09,$1F,$96
	.dc.b	$6B,$FC,$E1,$DB,$FD,$1F,$E9,$FF,$A3,$7C,$04,$8F,$E9,$4F,$F4,$9F
	.dc.b	$8B,$91,$82,$1F,$D0,$9C,$BF,$E5,$1D,$91,$FA,$A2,$47,$EA,$3D,$DF
	.dc.b	$EA,$84,$7F,$AC,$78,$FF,$D5,$CF,$D8,$91,$FD,$7B,$FE,$4B,$5E,$03
	.dc.b	$E9,$30,$D8,$FF,$28,$35,$FF,$32,$35,$FF,$39,$B9,$E2,$70,$7F,$90
	.dc.b	$27,$AD,$7F,$6E,$AD,$4E,$57,$CD,$67,$E6,$7C,$24,$BF,$0B,$E4,$F8
	.dc.b	$5D,$47,$F9,$33,$1E,$EF,$F9,$3A,$D5,$DE,$A4,$EF,$0F,$04,$62,$17
	.dc.b	$C9,$76,$3D,$FF,$47,$D1,$49,$E4,$77,$7D,$9F,$15,$FE,$0A,$EB,$7C
	.dc.b	$66,$2B,$67,$CE,$9D,$47,$C3,$EE,$F8,$9E,$6C,$AD,$EF,$FE,$26,$D3
	.dc.b	$C8,$DD,$BE,$45,$F1,$AD,$FD,$17,$F8,$6B,$E9,$3F,$D4,$B8,$7A,$DB
	.dc.b	$7F,$4D,$B5,$FE,$74,$5F,$2F,$C4,$E5,$10,$F4,$C6,$9B,$1B,$69,$EE
	.dc.b	$F9,$DF,$F1,$6B,$D7,$E0,$7C,$84,$79,$FC,$07,$9C,$C3,$F4,$B0,$F3
	.dc.b	$4E,$63,$C3,$ED,$FB,$7E,$3A,$46,$08,$71,$3D,$91,$F0,$CF,$89,$F2
	.dc.b	$7F,$17,$CA,$93,$EC,$7E,$47,$B9,$23,$04,$3A,$0D,$7F,$DC,$EF,$DC
	.dc.b	$FA,$3F,$96,$E6,$79,$3C,$23,$3F,$48,$F9,$7F,$F2,$7B,$67,$F0,$AD
	.dc.b	$7B,$BF,$9A,$7F,$13,$D7,$F8,$17,$FC,$F9,$C7,$FE,$58,$83,$79,$3F
	.dc.b	$CB,$78,$7D,$7A,$F1,$7A,$38,$33,$C9,$E2,$EB,$F1,$F5,$4A,$F1,$C7
	.dc.b	$5F,$A0,$91,$E2,$7C,$1F,$25,$DB,$E6,$E4,$FC,$37,$C3,$F3,$7E,$ED
	.dc.b	$64,$FB,$9F,$83,$7F,$A7,$8F,$D4,$65,$1F,$D3,$F7,$FC,$CD,$5D,$1E
	.dc.b	$CE,$A9,$5F,$46,$3F,$A6,$E7,$FD,$3B,$39,$3F,$D6,$E3,$F6,$9B,$C8
	.dc.b	$EA,$FE,$DE,$3E,$3F,$C9,$C9,$FB,$8F,$CA,$F6,$BF,$AC,$F5,$FF,$5A
	.dc.b	$EB,$95,$FD,$9E,$3F,$80,$63,$23,$D2,$FF,$03,$CB,$E9,$CC,$9F,$C4
	.dc.b	$3F,$5A,$BF,$B2,$F4,$49,$FE,$E4,$FF,$68,$F8,$BF,$E1,$9A,$FF,$0D
	.dc.b	$CE,$57,$C6,$8F,$BF,$D9,$F8,$0F,$7E,$4F,$E5,$63,$FD,$3B,$F4,$6F
	.dc.b	$A4,$A4,$9F,$E2,$B0,$F3,$F7,$E7,$70,$64,$EA,$39,$FC,$1E,$A3,$07
	.dc.b	$27,$BF,$9F,$E0,$F9,$49,$3D,$A3,$2D,$21,$FA,$EC,$1E,$87,$FA,$D3
	.dc.b	$E9,$3F,$49,$E4,$FF,$35,$6F,$91,$FD,$35,$B6,$0F,$EA,$5F,$DC,$B6
	.dc.b	$79,$79,$FD,$4F,$F6,$0F,$D2,$3C,$0F,$F6,$1F,$1B,$B6,$93,$1F,$E7
	.dc.b	$70,$ED,$D9,$23,$2F,$4F,$0B,$FD,$9B,$F4,$7F,$D8,$27,$CC,$D4,$D3
	.dc.b	$FB,$1B,$FD,$21,$AF,$F1,$26,$E7,$FD,$5D,$BA,$FF,$6F,$5F,$3F,$E0
	.dc.b	$25,$73,$7F,$DB,$FE,$6F,$F1,$89,$F9,$33,$E2,$F2,$FC,$E7,$A3,$79
	.dc.b	$1E,$36,$BE,$3F,$4B,$FB,$4A,$FA,$46,$DE,$39,$CC,$FF,$80,$CF,$7B
	.dc.b	$87,$5A,$65,$F8,$B1,$97,$E3,$FD,$1B,$E9,$35,$1F,$22,$30,$F1,$CD
	.dc.b	$18,$8F,$04,$62,$7F,$C4,$10,$FA,$4A,$47,$FC,$48,$FF,$89,$7C,$AF
	.dc.b	$F0,$6E,$3F,$E0,$59,$FC,$27,$07,$F7,$C7,$FB,$4B,$78,$F6,$FB,$DB
	.dc.b	$1F,$95,$F3,$7F,$C5,$F3,$B3,$FC,$0A,$E2,$3F,$A7,$9F,$8D,$8E,$DE
	.dc.b	$47,$28,$FC,$F7,$2D,$51,$4C,$79,$16,$BC,$CE,$1E,$FF,$49,$6B,$FA
	.dc.b	$EE,$08,$ED,$97,$E2,$AB,$FC,$32,$FF,$59,$6F,$36,$B9,$64,$7A,$BC
	.dc.b	$0B,$8F,$47,$D1,$F4,$9F,$1A,$3F,$B5,$BF,$29,$BB,$EB,$79,$7F,$13
	.dc.b	$BE,$D7,$99,$E7,$E2,$49,$FF,$16,$E0,$F2,$71,$7D,$BC,$6F,$09,$8D
	.dc.b	$BF,$C7,$D4,$8F,$2F,$93,$30,$E7,$EA,$5C,$3C,$FF,$01,$EA,$3D,$7E
	.dc.b	$5D,$AF,$F1,$EF,$D2,$3E,$07,$EB,$36,$3C,$BE,$8E,$27,$ED,$62,$CA
	.dc.b	$E1,$86,$DF,$F9,$23,$E9,$1F,$07,$F8,$16,$C6,$7E,$7B,$13,$F8,$5F
	.dc.b	$A4,$FF,$2C,$4F,$F9,$41,$BB,$7F,$86,$97,$BB,$BF,$7F,$F1,$9F,$48
	.dc.b	$E6,$A6,$38,$EE,$3F,$E5,$A9,$7C,$4F,$0F,$97,$E2,$D9,$FF,$2F,$FA
	.dc.b	$BF,$B2,$6F,$C6,$7E,$66,$31,$AF,$F8,$B1,$47,$E7,$78,$7F,$0D,$D3
	.dc.b	$FE,$73,$6D,$FF,$A2,$DF,$C4,$46,$7B,$78,$7B,$3C,$7B,$2D,$FF,$9C
	.dc.b	$2F,$F0,$8F,$F9,$CB,$E8,$FF,$A6,$BF,$E7,$4F,$C6,$5F,$29,$FA,$D9
	.dc.b	$CE,$BF,$E4,$09,$CD,$7F,$9E,$BE,$91,$E7,$FE,$CD,$F5,$EB,$FE,$89
	.dc.b	$73,$3B,$F3,$F5,$CF,$23,$97,$C1,$5C,$BF,$C9,$BF,$47,$FB,$19,$FF
	.dc.b	$43,$B6,$33,$E0,$7F,$01,$8C,$65,$FE,$97,$FA,$27,$C7,$7F,$05,$6E
	.dc.b	$1F,$E9,$05,$F1,$D9,$9D,$7B,$FB,$F2,$6D,$ED,$FE,$1D,$E2,$3E,$0A
	.dc.b	$FC,$1E,$1B,$5F,$E9,$CF,$A2,$4F,$F1,$55,$D5,$63,$0E,$6E,$57,$1F
	.dc.b	$2E,$F0,$FF,$9B,$BE,$89,$F4,$DF,$C7,$DB,$5F,$EA,$A7,$FD,$AC,$D8
	.dc.b	$FE,$49,$F4,$6F,$A4,$BC,$18,$6D,$7F,$AC,$BE,$85,$E7,$7F,$2B,$7F
	.dc.b	$65,$19,$8C,$78,$1F,$CB,$FE,$87,$FD,$B7,$3F,$EB,$A6,$E9,$FE,$84
	.dc.b	$BE,$2F,$FA,$1A,$33,$CF,$FE,$CB,$9F,$D3,$D7,$87,$C6,$F0,$7A,$1E
	.dc.b	$17,$C5,$E4,$7F,$4D,$C7,$FC,$35,$68,$C4,$23,$17,$FD,$89,$B1,$FD
	.dc.b	$A7,$1F,$CC,$BF,$DB,$DA,$FF,$BC,$DA,$FF,$89,$DA,$BE,$85,$E8,$78
	.dc.b	$1B,$7F,$DC,$1B,$3F,$DD,$DA,$FF,$7D,$E3,$F1,$B7,$1F,$4C,$62,$3F
	.dc.b	$C6,$C1,$C3,$FE,$39,$97,$CB,$F8,$49,$7E,$17,$E1,$F8,$1F,$CB,$6B
	.dc.b	$FC,$10,$D7,$FC,$1C,$D7,$BB,$5E,$DC,$5E,$D2,$57,$B5,$EB,$52,$3F
	.dc.b	$C1,$F5,$E4,$6C,$78,$BE,$26,$46,$9B,$7F,$17,$E2,$FE,$51,$27,$E5
	.dc.b	$3F,$73,$E6,$3F,$0B,$3F,$E2,$16,$BF,$E2,$76,$BF,$21,$AF,$F9,$7F
	.dc.b	$E8,$7F,$9D,$D7,$F8,$91,$B1,$FE,$2B,$6B,$FE,$31,$6B,$F4,$BE,$87
	.dc.b	$FB,$14,$FF,$8B,$1B,$1F,$E3,$36,$BF,$E3,$96,$BF,$6B,$EA,$3F,$92
	.dc.b	$BF,$E3,$96,$BF,$E3,$F6,$BB,$F3,$7F,$37,$F4,$5F,$E0,$F7,$FC,$76
	.dc.b	$D8,$FF,$21,$2F,$98,$FF,$6D,$47,$E2,$23,$4F,$F9,$2F,$E8,$1F,$43
	.dc.b	$FE,$44,$84,$61,$FF,$23,$B5,$FF,$27,$35,$FF,$2A,$7D,$0F,$F9,$FD
	.dc.b	$FF,$26,$B5,$FF,$2A,$35,$DF,$C0,$77,$FB,$BF,$DA,$C1,$A4,$D7,$C9
	.dc.b	$F4,$2F,$9E,$A4,$AF,$F2,$DC,$65,$BE,$0F,$02,$38,$37,$D3,$D0,$8E
	.dc.b	$0E,$7D,$FE,$02,$38,$3F,$CD,$3F,$42,$C7,$F6,$7E,$0D,$9F,$F5,$54
	.dc.b	$D5,$16,$30,$DF,$81,$E3,$5F,$BC,$9C,$B4,$CD,$74,$6A,$8F,$F0,$EB
	.dc.b	$FE,$B9,$AE,$DE,$96,$BA,$1F,$49,$4E,$37,$FE,$2F,$BE,$18,$78,$6F
	.dc.b	$51,$83,$FD,$89,$7C,$9C,$1E,$18,$E6,$70,$7D,$F3,$FA,$FB,$CC,$70
	.dc.b	$BF,$03,$A3,$AC,$71,$C3,$D3,$1B,$BC,$5F,$38,$C1,$0E,$43,$0F,$33
	.dc.b	$86,$D7,$48,$DE,$1E,$28,$FA,$D3,$AA,$ED,$5A,$BF,$4B,$CF,$0F,$1B
	.dc.b	$F7,$A3,$37,$CA,$73,$3F,$14,$F3,$46,$5F,$93,$7D,$DE,$46,$AF,$C3
	.dc.b	$3B,$8F,$11,$93,$E3,$9C,$AB,$C3,$3F,$66,$30,$43,$89,$27,$A1,$23
	.dc.b	$84,$65,$F0,$91,$F2,$99,$3C,$27,$C3,$3C,$87,$BC,$C1,$23,$E8,$12
	.dc.b	$3A,$47,$D7,$9F,$4B,$DA,$7C,$8F,$67,$83,$99,$E3,$5F,$26,$3C,$13
	.dc.b	$A7,$A1,$F0,$5C,$61,$F7,$EB,$8D,$F8,$2B,$8C,$61,$F5,$CE,$DF,$14
	.dc.b	$7A,$23,$11,$EF,$3F,$DE,$8F,$E7,$1B,$87,$F2,$B8,$C4,$7C,$E3,$27
	.dc.b	$F0,$A8,$D1,$1F,$0B,$FB,$6E,$CF,$34,$F8,$08,$E2,$32,$D9,$F4,$B5
	.dc.b	$FB,$07,$F0,$77,$18,$7C,$46,$DF,$E0,$A3,$F8,$B4,$62,$7E,$81,$FC
	.dc.b	$71,$C3,$F2,$0C,$1E,$83,$C2,$53,$C3,$7F,$E1,$47,$13,$E1,$6D,$78
	.dc.b	$4F,$E4,$55,$DA,$BD,$F3,$A3,$C6,$7A,$CC,$1E,$33,$07,$82,$FE,$2B
	.dc.b	$A8,$E6,$E4,$F3,$94,$F3,$CF,$86,$7E,$94,$71,$6B,$BB,$F5,$36,$FD
	.dc.b	$4F,$C4,$3F,$87,$F7,$7F,$30,$F1,$A4,$7F,$23,$8F,$E6,$66,$1F,$E2
	.dc.b	$B1,$87,$F9,$47,$BB,$F8,$B2,$3F,$D4,$9A,$FF,$1C,$7F,$9E,$9F,$CF
	.dc.b	$BE,$1B,$F9,$FF,$A3,$27,$F9,$93,$FD,$0B,$C1,$C4,$7F,$24,$8B,$97
	.dc.b	$F5,$79,$D4,$FF,$60,$7F,$38,$47,$FC,$A6,$7F,$95,$3B,$1F,$45,$FD
	.dc.b	$59,$FF,$49,$C7,$F4,$86,$BF,$CE,$1D,$B9,$8F,$F0,$94,$62,$3A,$9F
	.dc.b	$E1,$B8,$C9,$AF,$94,$91,$A3,$FC,$40,$E1,$E3,$3F,$ED,$07,$C5,$1E
	.dc.b	$53,$CD,$ED,$F9,$FE,$7B,$C2,$49,$E9,$48,$C4,$7A,$5E,$0F,$B7,$E3
	.dc.b	$7F,$BA,$F3,$FF,$84,$E3,$DD,$1F,$0C,$CF,$5F,$C5,$F1,$FE,$37,$B7
	.dc.b	$2B,$E3,$46,$5A,$43,$E4,$7C,$F7,$F8,$0B,$E8,$7E,$8F,$07,$F8,$3D
	.dc.b	$BC,$DF,$E8,$66,$D0,$2F,$83,$E4,$73,$88,$7F,$7E,$8D,$37,$E4,$3F
	.dc.b	$C2,$3F,$75,$FE,$C7,$8F,$F0,$1B,$5F,$EE,$72,$9F,$31,$23,$91,$D7
	.dc.b	$E5,$FD,$46,$5E,$31,$87,$C5,$F0,$DE,$FF,$6E,$46,$08,$78,$23,$E2
	.dc.b	$1F,$17,$E2,$BE,$37,$8B,$27,$D8,$FC,$9F,$62,$46,$08,$7A,$9F,$9D
	.dc.b	$EC,$7F,$8C,$3E,$83,$EE,$70,$7F,$8F,$5B,$C1,$F3,$E7,$77,$F2,$5A
	.dc.b	$FC,$4F,$5B,$EC,$11,$F9,$CF,$5C,$7F,$C7,$BF,$83,$FF,$4E,$3F,$94
	.dc.b	$4F,$F8,$79,$B8,$FF,$93,$5F,$EA,$7D,$1F,$EF,$49,$FF,$1A,$B5,$FF
	.dc.b	$12,$B6,$7F,$CA,$8B,$3F,$61,$27,$FC,$7B,$9E,$5B,$EF,$79,$3E,$0E
	.dc.b	$47,$F9,$0B,$14,$7A,$3E,$CF,$F4,$35,$FA,$2F,$36,$5F,$92,$3E,$54
	.dc.b	$7C,$BC,$A4,$F8,$23,$E0,$FB,$1F,$0B,$E1,$24,$F7,$8F,$D6,$D4,$8F
	.dc.b	$7B,$FB,$0E,$0F,$AC,$48,$F5,$BF,$67,$83,$FC,$D3,$5F,$C4,$B7,$FF
	.dc.b	$3B,$37,$87,$FD,$DA,$DB,$FF,$63,$78,$7F,$BC,$BF,$DE,$F9,$B2,$7E
	.dc.b	$7C,$7E,$16,$3F,$CE,$BF,$B6,$7F,$9E,$A7,$F6,$0D,$5C,$5F,$E1,$9C
	.dc.b	$3F,$8A,$47,$F9,$69,$B9,$79,$3C,$4F,$3F,$C4,$7F,$A0,$3E,$F7,$91
	.dc.b	$FE,$80,$BF,$F3,$0B,$6B,$FD,$20,$F7,$C7,$1A,$FF,$01,$A4,$3D,$FF
	.dc.b	$39,$E8,$F4,$FE,$2F,$D6,$7C,$6F,$91,$97,$FE,$80,$D5,$3E,$3B,$93
	.dc.b	$FE,$84,$CB,$A3,$7C,$AF,$86,$47,$FD,$13,$97,$FB,$55,$BF,$31,$FE
	.dc.b	$9C,$FA,$47,$A3,$7E,$96,$57,$FA,$42,$72,$DF,$4B,$85,$48,$C1,$BF
	.dc.b	$83,$E9,$32,$30,$73,$F4,$FC,$1D,$92,$30,$43,$FD,$1B,$82,$7C,$C7
	.dc.b	$9F,$D3,$FF,$59,$FD,$2B,$E6,$A9,$2F,$FD,$4F,$0F,$57,$64,$8C,$BF
	.dc.b	$EB,$CF,$A5,$7D,$2F,$FA,$FB,$CE,$FF,$75,$75,$FF,$B2,$DB,$D7,$F6
	.dc.b	$DB,$87,$F3,$D5,$F3,$FC,$49,$1F,$EB,$88,$CB,$FE,$BA,$C1,$F4,$BF
	.dc.b	$EC,$CF,$A3,$FE,$E1,$8F,$FB,$65,$B8,$FE,$25,$B7,$FD,$2D,$7D,$47
	.dc.b	$12,$47,$FB,$1A,$32,$DF,$A1,$EC,$47,$07,$FB,$33,$E9,$BF,$AB,$A9
	.dc.b	$F9,$8F,$3E,$0F,$31,$E7,$DE,$47,$76,$BF,$EC,$49,$78,$7F,$EF,$28
	.dc.b	$7D,$33,$7F,$64,$E6,$7F,$C3,$75,$C3,$5F,$F0,$63,$63,$FC,$26,$D7
	.dc.b	$7E,$23,$E9,$B6,$D8,$DB,$5D,$C7,$F8,$5D,$FE,$E9,$F4,$D5,$34,$8F
	.dc.b	$F8,$77,$DA,$70,$FF,$BC,$A1,$F1,$38,$7F,$DC,$67,$F7,$D8,$D4,$67
	.dc.b	$2D,$4F,$F8,$0E,$31,$3F,$C1,$7E,$01,$E1,$E2,$BF,$84,$55,$53,$AA
	.dc.b	$F1,$73,$E6,$5F,$4D,$63,$0C,$9F,$47,$B7,$6B,$CE,$F1,$3B,$5B,$C4
	.dc.b	$C5,$7F,$86,$73,$F4,$CE,$63,$07,$F1,$D8,$73,$F9,$DF,$DA,$46,$B0
	.dc.b	$E5,$39,$E6,$93,$99,$F6,$F8,$38,$57,$E0,$8F,$85,$F4,$DF,$36,$73
	.dc.b	$5F,$CB,$63,$E9,$BC,$DD,$57,$D0,$73,$3C,$0F,$F1,$A3,$88,$F6,$F8
	.dc.b	$1D,$B7,$F8,$6F,$83,$DA,$98,$75,$F6,$EE,$3C,$C6,$2F,$A3,$F3,$9A
	.dc.b	$CB,$C4,$FA,$BD,$B9,$87,$9F,$DC,$B8,$73,$C1,$EC,$7C,$03,$E0,$FB
	.dc.b	$36,$FE,$47,$D3,$3E,$23,$F0,$CD,$8F,$85,$ED,$E2,$7F,$15,$97,$F8
	.dc.b	$46,$7E,$82,$BC,$E7,$A7,$F2,$1F,$4C,$F5,$E9,$8F,$52,$E1,$F6,$7E
	.dc.b	$99,$EC,$7E,$5D,$B1,$9F,$86,$C4,$FE,$6F,$E9,$9C,$D4,$C7,$25,$BF
	.dc.b	$C2,$4B,$DF,$E1,$EF,$71,$6C,$FE,$27,$D7,$79,$1F,$A7,$5F,$4D,$63
	.dc.b	$0F,$EA,$F8,$BF,$0B,$5F,$DD,$5B,$D3,$FE,$77,$7F,$09,$19,$F1,$F8
	.dc.b	$7C,$27,$1E,$DB,$3F,$99,$A4,$7E,$DF,$E9,$9E,$BF,$EE,$7E,$99,$F3
	.dc.b	$19,$8C,$3F,$BD,$FA,$6F,$E4,$F8,$FE,$8D,$F5,$B7,$D1,$D6,$FB,$E7
	.dc.b	$93,$BF,$8C,$BC,$1F,$93,$FA,$6F,$DB,$37,$CC,$FE,$15,$1F,$85,$8C
	.dc.b	$57,$EC,$7E,$99,$F2,$5F,$C3,$DB,$19,$F8,$BC,$4E,$7E,$07,$5E,$5D
	.dc.b	$AE,$4F,$F1,$7F,$11,$F4,$D7,$E9,$70,$DA,$FE,$01,$F4,$C9,$FE,$3C
	.dc.b	$BA,$AC,$61,$CD,$CA,$E3,$E6,$5D,$BF,$65,$F4,$CF,$A6,$FE,$50,$DA
	.dc.b	$FE,$1C,$FF,$B9,$9B,$1F,$CB,$7E,$9B,$DF,$5F,$EE,$06,$BF,$EE,$E6
	.dc.b	$BF,$E3,$B6,$BF,$EB,$DF,$A6,$FD,$63,$FE,$E8,$6B,$FE,$F6,$6B,$FE
	.dc.b	$43,$6B,$FE,$C5,$FA,$6F,$E9,$B7,$FE,$EB,$6C,$7F,$BC,$9A,$FF,$C0
	.dc.b	$06,$BC,$FC,$4F,$E9,$AD,$9F,$F7,$A5,$7E,$A5,$B3,$A9,$FF,$6A,$7D
	.dc.b	$EF,$E0,$DF,$F2,$B0,$70,$FF,$96,$5A,$F0,$F4,$39,$2F,$A7,$FE,$0D
	.dc.b	$7C,$A3,$FD,$C1,$18,$70,$D7,$BB,$5F,$F3,$37,$D3,$FF,$81,$67,$94
	.dc.b	$23,$0F,$56,$BE,$86,$BC,$B8,$FC,$B1,$F9,$29,$E6,$D9,$E5,$3E,$AA
	.dc.b	$7D,$6D,$25,$F8,$A3,$2F,$D5,$F8,$0F,$83,$AF,$F7,$3C,$3E,$B4,$60
	.dc.b	$E9,$D3,$2B,$93,$67,$D5,$CB,$E6,$24,$F9,$9F,$4F,$2C,$8F,$4B,$67
	.dc.b	$E1,$78,$DE,$E2,$4F,$B9,$F8,$3E,$74,$8F,$82,$D9,$F9,$5E,$77,$C6
	.dc.b	$24,$FC,$67,$EE,$FD,$22,$69,$27,$EA,$C6,$5F,$AF,$F4,$FF,$E0,$9F
	.dc.b	$1E,$F8,$5F,$C8,$2F,$F0,$93,$FE,$1E,$F3,$33,$5F,$7B,$EB,$3F,$81
	.dc.b	$55,$DB,$FC,$6D,$18,$8F,$C6,$35,$FC,$BE,$5F,$E3,$27,$C7,$EE,$7F
	.dc.b	$8F,$63,$11,$F9,$36,$BF,$9E,$EE,$7B,$3F,$E4,$36,$C7,$E4,$1A,$FE
	.dc.b	$6D,$AF,$E9,$BE,$95,$F4,$1F,$E4,$B6,$C7,$E6,$1A,$FE,$8D,$AF,$EB
	.dc.b	$BE,$B1,$E5,$7F,$94,$DB,$1F,$A0,$6B,$FA,$B6,$BF,$B6,$FA,$B7,$DC
	.dc.b	$7F,$96,$DB,$1F,$A8,$6B,$FB,$36,$BE,$9C,$BF,$C7,$1C,$7F,$ED,$5C
	.dc.b	$7D,$FF,$03,$FE,$56,$6C,$67,$C2,$DD,$FE,$0E,$7F,$BA,$3E,$F8,$F8
	.dc.b	$44,$7F,$73,$F4,$FF,$EE,$27,$F1,$4D,$4D,$FA,$8F,$F0,$FC,$7F,$90
	.dc.b	$DA,$FF,$8B,$63,$81,$BD,$FC,$71,$CB,$EA,$2C,$FF,$88,$E3,$FC,$98
	.dc.b	$D7,$94,$68,$DF,$89,$23,$71,$97,$8C,$65,$E2,$68,$C9,$0C,$93,$3F
	.dc.b	$F3,$FD,$78,$5B,$3D,$67,$59,$FE,$B5,$F8,$71,$DE,$78,$DE,$DE,$33
	.dc.b	$DD,$CB,$E3,$3E,$34,$F1,$AE,$96,$BE,$51,$89,$D5,$7D,$49,$CB,$F5
	.dc.b	$A3,$2F,$CD,$3B,$5F,$89,$EE,$62,$BB,$D7,$0B,$F8,$11,$89,$D1,$D4
	.dc.b	$F5,$39,$7E,$19,$93,$A9,$83,$C4,$62,$BC,$47,$C2,$75,$3E,$07,$94
	.dc.b	$64,$93,$F0,$63,$C2,$7C,$13,$CA,$7D,$23,$27,$82,$38,$BD,$23,$93
	.dc.b	$E0,$AF,$81,$7F,$2A,$38,$4E,$AB,$CE,$FB,$DE,$71,$DC,$F6,$46,$48
	.dc.b	$F7,$21,$D0,$F4,$1C,$0C,$DF,$DF,$7F,$08,$7F,$7C,$70,$FF,$75,$6B
	.dc.b	$9B,$FF,$72,$46,$23,$EA,$19,$3F,$77,$1A,$3D,$D1,$D2,$BF,$DB,$BC
	.dc.b	$5F,$CD,$DF,$6C,$66,$FD,$76,$AD,$46,$27,$53,$EC,$AE,$7C,$BD,$B0
	.dc.b	$F9,$98,$70,$AD,$61,$C6,$7F,$81,$CE,$67,$E3,$13,$0E,$AF,$86,$FA
	.dc.b	$3F,$80,$9E,$13,$A2,$57,$4A,$DD,$7E,$4E,$B7,$38,$C3,$CF,$3F,$8A
	.dc.b	$BE,$2F,$FB,$76,$B6,$E6,$BA,$34,$F1,$5F,$3C,$3E,$3D,$70,$9F,$C7
	.dc.b	$1C,$E7,$4F,$C2,$3F,$BB,$46,$F0,$FA,$37,$DA,$78,$9E,$9A,$E2,$FC
	.dc.b	$6A,$E1,$7F,$2E,$3B,$1E,$53,$81,$6F,$F3,$33,$C2,$3C,$47,$3A,$EB
	.dc.b	$87,$D0,$9F,$9B,$E7,$7F,$36,$25,$7F,$36,$EF,$AF,$F3,$F8,$B8,$7F
	.dc.b	$2A,$9F,$E7,$BF,$0B,$FC,$FB,$DB,$97,$FC,$FA,$32,$FF,$40,$8C,$BF
	.dc.b	$D2,$63,$FA,$2F,$AF,$FD,$1B,$D0,$93,$FD,$1A,$32,$FF,$58,$7F,$46
	.dc.b	$47,$FC,$A7,$F6,$0F,$A9,$DC,$7F,$55,$9F,$EA,$73,$98,$FF,$14,$64
	.dc.b	$DF,$F9,$F7,$8C,$FA,$3F,$F2,$ED,$23,$0F,$F9,$81,$AF,$F9,$AD,$AF
	.dc.b	$F9,$D9,$CD,$75,$8F,$F1,$B4,$62,$BB,$D7,$F6,$68,$7C,$B5,$F6,$B1
	.dc.b	$F2,$B3,$0F,$3E,$1E,$69,$D4,$78,$DF,$3D,$9A,$63,$EA,$7B,$12,$79
	.dc.b	$9E,$4F,$8E,$F1,$1C,$88,$FB,$DF,$87,$DB,$FE,$26,$8F,$EC,$8D,$7F
	.dc.b	$CA,$4D,$7F,$C7,$3E,$17,$93,$C2,$91,$F2,$5A,$FC,$97,$E1,$11,$F2
	.dc.b	$9F,$47,$EA,$7F,$C1,$7F,$51,$FE,$A2,$E0,$F9,$92,$8C,$6B,$FD,$DD
	.dc.b	$B1,$FE,$65,$5E,$47,$23,$EE,$B8,$8F,$F4,$DB,$5F,$F5,$33,$5F,$F5
	.dc.b	$8A,$F9,$1E,$24,$8E,$2F,$4F,$97,$F3,$B2,$3D,$72,$9F,$59,$B8,$F0
	.dc.b	$B5,$F0,$F9,$FE,$62,$3E,$66,$AC,$7E,$1D,$27,$E3,$9F,$23,$CF,$F9
	.dc.b	$3F,$1D,$2F,$C6,$FC,$CF,$86,$91,$F0,$9F,$8A,$47,$E2,$F8,$FF,$20
	.dc.b	$8E,$0F,$93,$F1,$1F,$5A,$24,$7C,$08,$FB,$24,$70,$7D,$EF,$A5,$5F
	.dc.b	$F9,$23,$EA,$3F,$C0,$D2,$7F,$A2,$93,$C8,$FF,$26,$F2,$F0,$9F,$C3
	.dc.b	$B5,$F8,$9C,$5F,$96,$BD,$E7,$FE,$54,$C4,$D2,$57,$8F,$E0,$3C,$5E
	.dc.b	$24,$8F,$C9,$E6,$8F,$37,$D4,$78,$AB,$C7,$E5,$CB,$F1,$C7,$78,$F0
	.dc.b	$69,$27,$E0,$47,$BF,$E9,$FE,$07,$C2,$49,$EF,$1E,$EA,$91,$EE,$7E
	.dc.b	$0F,$0F,$AC,$91,$F2,$DF,$77,$0F,$F9,$CA,$9F,$05,$49,$5F,$AE,$8F
	.dc.b	$E2,$8F,$EE,$A7,$E6,$FD,$44,$8F,$9B,$1F,$26,$3F,$CF,$BF,$4D,$FB
	.dc.b	$2E,$8F,$D9,$B7,$93,$FC,$D1,$C6,$AF,$17,$83,$31,$F0,$3E,$5F,$8A
	.dc.b	$FF,$43,$FD,$3F,$19,$FF,$6E,$3F,$E9,$32,$61,$E0,$F9,$4F,$0F,$CA
	.dc.b	$4A,$FE,$0F,$9F,$FA,$67,$67,$D4,$7C,$1F,$94,$95,$EF,$8E,$F7,$23
	.dc.b	$F8,$7E,$48,$F1,$F9,$1E,$E2,$4F,$67,$E0,$FC,$07,$FA,$7B,$E9,$9E
	.dc.b	$9D,$FA,$89,$7F,$C6,$A7,$2D,$F5,$38,$54,$8C,$1B,$FA,$8E,$93,$23
	.dc.b	$07,$3F,$57,$C1,$D9,$23,$04,$3F,$42,$06,$09,$F4,$1E,$9F,$53,$FD
	.dc.b	$6F,$F5,$0F,$9D,$A4,$BF,$E6,$70,$F4,$F6,$48,$CB,$FE,$C0,$FA,$77
	.dc.b	$D4,$7F,$B0,$BC,$EF,$F7,$FF,$7F,$F3,$86,$F5,$FF,$AB,$5B,$87,$9D
	.dc.b	$7D,$17,$12,$47,$F4,$48,$CB,$FD,$1B,$07,$D4,$7F,$B4,$3E,$9B,$F3
	.dc.b	$72,$69,$3D,$9D,$12,$3F,$A7,$C7,$F7,$26,$FD,$1F,$32,$38,$3F,$AA
	.dc.b	$FD,$47,$F4,$E7,$B7,$07,$84,$8F,$4D,$FC,$84,$70,$78,$3C,$3F,$E7
	.dc.b	$4B,$EA,$3F,$A3,$3F,$67,$1F,$EF,$9C,$1F,$ED,$38,$FF,$6A,$46,$1C
	.dc.b	$1E,$FA,$E6,$EE,$34,$DF,$11,$F5,$1C,$23,$E3,$46,$1E,$72,$8C,$47
	.dc.b	$19,$FF,$12,$AF,$A8,$A4,$7F,$C5,$2F,$F8,$AB,$D9,$FF,$11,$63,$FD
	.dc.b	$E2,$5F,$19,$8E,$5F,$65,$59,$FF,$78,$4D,$7F,$34,$DF,$7B,$FC,$69
	.dc.b	$5F,$E0,$69,$E6,$FF,$82,$23,$31,$FA,$23,$F4,$5C,$DC,$D7,$D4,$58
	.dc.b	$C3,$2B,$A7,$B9,$6B,$91,$FE,$01,$8C,$9D,$FC,$3E,$93,$E5,$7D,$19
	.dc.b	$B9,$6F,$26,$1B,$7F,$E1,$E9,$FD,$83,$E9,$F0,$33,$8F,$A7,$44,$7C
	.dc.b	$3F,$A8,$FF,$0D,$D7,$F8,$99,$F8,$35,$EF,$72,$DD,$B8,$1F,$27,$BE
	.dc.b	$D7,$C7,$7A,$7B,$72,$3A,$7B,$96,$FA,$7C,$97,$84,$C6,$9F,$E3,$2C
	.dc.b	$91,$E5,$F6,$E6,$1E,$5F,$72,$E1,$E5,$F6,$3D,$47,$AF,$D9,$B5,$FE
	.dc.b	$40,$FA,$87,$BD,$F8,$16,$C7,$BF,$DB,$C4,$FE,$17,$E4,$BF,$8E,$4F
	.dc.b	$F9,$21,$B8,$FE,$05,$79,$9B,$E2,$FC,$67,$DE,$BE,$9A,$98,$F0,$6E
	.dc.b	$3F,$E5,$2F,$C4,$DE,$97,$E4,$DB,$1F,$E5,$6F,$A9,$7C,$36,$6B,$F3
	.dc.b	$1F,$7D,$F8,$EA,$FF,$2D,$B7,$6F,$CA,$3F,$C7,$A3,$3E,$17,$0E,$39
	.dc.b	$1C,$39,$AD,$FE,$6F,$D9,$F8,$6D,$E3,$FD,$37,$AC,$F9,$5F,$D4,$56
	.dc.b	$BB,$3F,$CC,$F8,$FE,$EC,$FF,$36,$97,$0F,$D3,$3F,$E8,$93,$3C,$3F
	.dc.b	$B1,$FB,$EF,$E9,$51,$FE,$74,$6C,$F0,$F4,$74,$FC,$5F,$76,$57,$F9
	.dc.b	$EB,$EA,$5F,$61,$8A,$FD,$A6,$1F,$E3,$2A,$E9,$BF,$7E,$FF,$03,$FC
	.dc.b	$07,$CD,$F5,$4F,$F5,$BB,$FE,$21,$C3,$9F,$2B,$4F,$F0,$7F,$C6,$1F
	.dc.b	$43,$4F,$E1,$5F,$52,$F9,$8C,$63,$FE,$8E,$FB,$EF,$DB,$37,$CE,$FE
	.dc.b	$21,$1F,$EB,$48,$C1,$BE,$DC,$9A,$FA,$BB,$1F,$E9,$8F,$C2,$1E,$37
	.dc.b	$F1,$96,$C7,$FA,$73,$B5,$F4,$59,$AF,$E3,$DE,$AB,$CF,$A6,$3C,$FB
	.dc.b	$87,$FA,$93,$F0,$57,$C5,$7F,$25,$6C,$64,$C4,$7F,$29,$FC,$61,$F3
	.dc.b	$74,$FE,$59,$F5,$9E,$DE,$9F,$E4,$D3,$AB,$FF,$58,$FD,$67,$F9,$7D
	.dc.b	$BF,$2B,$FC,$CE,$1F,$2B,$63,$F9,$B7,$D6,$3E,$5A,$98,$F5,$6E,$1F
	.dc.b	$EB,$C9,$7C,$AF,$F4,$3C,$FF,$C2,$D1,$89,$FE,$A8,$D8,$FE,$BA,$D7
	.dc.b	$FB,$17,$D4,$3E,$37,$94,$23,$15,$FD,$75,$B1,$FD,$99,$AF,$F6,$BF
	.dc.b	$A9,$FF,$20,$4F,$F6,$36,$C7,$F6,$C6,$BB,$F9,$2E,$FE,$E7,$8E,$3F
	.dc.b	$28,$D7,$FC,$60,$D4,$D9,$FC,$95,$77,$F0,$3D,$31,$8A,$F0,$47,$B3
	.dc.b	$E9,$FF,$C5,$75,$FD,$DD,$B3,$FE,$00,$6B,$FE,$05,$6B,$FA,$7A,$FB
	.dc.b	$8F,$AF,$93,$B9,$2B,$FC,$05,$DF,$E1,$9F,$F0,$73,$5F,$D8,$35,$FF
	.dc.b	$26,$35,$FE,$43,$A7,$D2,$8F,$9B,$1C,$23,$07,$BB,$E4,$A5,$7C,$F6
	.dc.b	$CF,$F8,$6B,$E9,$7F,$15,$3F,$E1,$56,$C6,$78,$BE,$FC,$62,$3F,$C4
	.dc.b	$9E,$7B,$C2,$FC,$1B,$6F,$FC,$3C,$D9,$FF,$13,$B5,$E3,$F3,$BC,$09
	.dc.b	$3C,$1F,$C6,$FD,$B7,$EE,$A7,$FC,$5E,$D7,$FC,$68,$D7,$F8,$B3,$5F
	.dc.b	$F3,$97,$D5,$3C,$CF,$E0,$B1,$FE,$38,$8C,$78,$7F,$E3,$35,$F2,$EF
	.dc.b	$13,$17,$C2,$FF,$C8,$5F,$4B,$FC,$1D,$FF,$1D,$35,$FF,$21,$B5,$DF
	.dc.b	$8B,$FE,$3F,$5F,$6D,$FC,$D1,$FF,$21,$35,$FF,$25,$B5,$FF,$28,$35
	.dc.b	$FE,$65,$F6,$DF,$B0,$4D,$26,$DD,$FE,$A9,$F4,$94,$95,$FE,$51,$8C
	.dc.b	$BF,$E5,$3F,$C0,$7F,$C9,$A4,$7D,$0F,$F9,$43,$27,$C8,$7A,$DB,$3F
	.dc.b	$0A,$30,$78,$7B,$A4,$77,$6B,$FE,$68,$FA,$AF,$F6,$82,$FA,$9F,$AC
	.dc.b	$FF,$AA,$7D,$F6,$9F,$D4,$67,$33,$FE,$4D,$AE,$27,$C3,$24,$FC,$F7
	.dc.b	$E3,$74,$FF,$B8,$E7,$A4,$FF,$5B,$8C,$47,$F9,$5D,$B7,$E0,$8F,$F5
	.dc.b	$D5,$7F,$89,$1E,$27,$FA,$F6,$77,$5C,$FE,$8F,$FB,$35,$F1,$A5,$69
	.dc.b	$E7,$EC,$6A,$FF,$B6,$60,$FA,$AC,$57,$FB,$27,$F0,$BF,$F0,$EB,$E0
	.dc.b	$0E,$1F,$85,$1C,$14,$F3,$78,$4E,$F1,$F2,$36,$FC,$37,$C6,$F8,$5F
	.dc.b	$F9,$E6,$B4,$F8,$FC,$1D,$54,$8E,$B1,$E1,$9D,$C7,$8A,$74,$F3,$9D
	.dc.b	$18,$C3,$E3,$5F,$81,$F2,$C6,$A7,$E0,$4E,$70,$F3,$C7,$80,$87,$C2
	.dc.b	$37,$7C,$5F,$01,$82,$18,$25,$F8,$A3,$C9,$3F,$E1,$FC,$3D,$67,$B2
	.dc.b	$FC,$57,$F8,$4A,$ED,$7E,$88,$EA,$F3,$7C,$AE,$5F,$31,$1C,$12,$FD
	.dc.b	$B1,$97,$C5,$87,$67,$33,$D6,$7D,$31,$A7,$C2,$E8,$C3,$A3,$D0,$68
	.dc.b	$FB,$87,$99,$E0,$E8,$E0,$7E,$42,$7E,$2B,$99,$F4,$3C,$8F,$41,$99
	.dc.b	$F0,$C6,$A3,$F0,$87,$84,$E8,$78,$4E,$04,$32,$78,$AF,$E8,$BC,$8F
	.dc.b	$2E,$1F,$57,$0F,$2C,$70,$AF,$05,$70,$3F,$3B,$1F,$EF,$B6,$BF,$40
	.dc.b	$C4,$FF,$09,$23,$FE,$F7,$8D,$BF,$14,$D9,$EB,$7B,$78,$5E,$27,$C1
	.dc.b	$3F,$AD,$8C,$39,$6C,$F1,$6B,$F2,$8F,$DD,$B8,$AF,$C0,$1F,$A8,$71
	.dc.b	$87,$D8,$9D,$9E,$3A,$EC,$D3,$B1,$4E,$D7,$FE,$56,$71,$7D,$AB,$F8
	.dc.b	$AD,$F6,$9F,$A9,$5B,$78,$DF,$66,$9D,$8A,$76,$9E,$4F,$D9,$BF,$53
	.dc.b	$E0,$BF,$55,$F5,$9F,$14,$F8,$2F,$E5,$3B,$3C,$77,$D2,$BE,$8C,$64
	.dc.b	$ED,$7E,$A7,$11,$A7,$CB,$7D,$9F,$92,$62,$7C,$75,$FC,$97,$E1,$7F
	.dc.b	$97,$70,$FF,$BC,$2B,$F9,$A3,$5F,$EA,$91,$B7,$3F,$21,$C0,$C3,$FD
	.dc.b	$67,$E3,$BF,$AB,$11,$FC,$73,$5D,$1F,$C8,$9F,$E6,$DE,$2F,$F3,$FE
	.dc.b	$D7,$D3,$D2,$57,$F4,$78,$7D,$5E,$CC,$C3,$5F,$EA,$D3,$CA,$38,$4E
	.dc.b	$DF,$F6,$01,$FE,$C1,$8E,$6E,$01,$FE,$28,$8F,$F0,$F4,$6A,$35,$F5
	.dc.b	$AF,$89,$FE,$AC,$D8,$FE,$B8,$D7,$C9,$BF,$7F,$9E,$F7,$1D,$09,$1E
	.dc.b	$13,$FC,$6E,$F8,$3E,$B2,$46,$70,$46,$23,$D1,$18,$7E,$34,$7C,$B3
	.dc.b	$E6,$7C,$CF,$CD,$F2,$BD,$64,$9E,$8F,$BB,$9F,$51,$F3,$E3,$EA,$44
	.dc.b	$8F,$A4,$D7,$DA,$7C,$8B,$79,$3F,$DE,$7E,$B3,$E7,$EF,$FE,$09,$6E
	.dc.b	$9F,$EF,$16,$DF,$D0,$7F,$CA,$CF,$F8,$15,$AF,$60,$F0,$69,$DF,$C7
	.dc.b	$F0,$78,$1F,$E0,$9F,$B8,$FE,$5A,$7F,$2E,$AF,$F1,$ED,$73,$6E,$1E
	.dc.b	$5A,$E3,$3E,$1F,$A5,$F5,$7C,$04,$8F,$23,$ED,$F8,$09,$18,$21,$D5
	.dc.b	$CB,$4F,$63,$F1,$7C,$AF,$8D,$F1,$52,$7E,$03,$F2,$7E,$2A,$46,$08
	.dc.b	$7A,$5F,$9D,$E9,$7F,$89,$FE,$B3,$FC,$C3,$3F,$1D,$BA,$7D,$19,$DC
	.dc.b	$FC,$D6,$BF,$0F,$DB,$FF,$18,$7E,$40,$F8,$3A,$49,$FF,$1C,$43,$DB
	.dc.b	$D9,$98,$FA,$D1,$C5,$DC,$74,$E6,$DC,$60,$E6,$FF,$8F,$FF,$03,$7C
	.dc.b	$DD,$F6,$65,$7F,$91,$E7,$2F,$F9,$27,$F0,$37,$CA,$7F,$91,$D7,$D2
	.dc.b	$D3,$11,$FE,$4B,$FC,$59,$BF,$F9,$37,$EC,$3F,$CC,$58,$7E,$35,$BC
	.dc.b	$3F,$E5,$FA,$DC,$FF,$95,$FF,$04,$7D,$37,$F9,$55,$7D,$25,$31,$19
	.dc.b	$E6,$D7,$3C,$BE,$51,$FE,$5D,$FC,$7D,$F3,$7E,$3F,$87,$95,$FE,$65
	.dc.b	$9C,$BF,$E6,$7F,$94,$7A,$5F,$E6,$55,$F3,$F4,$C4,$7F,$9A,$7F,$05
	.dc.b	$4D,$FA,$D9,$7F,$E7,$19,$CB,$FE,$73,$FA,$87,$D0,$7C,$5F,$B7,$93
	.dc.b	$98,$E9,$AF,$5F,$B0,$93,$CE,$3F,$CF,$1F,$82,$FF,$B2,$79,$BF,$D1
	.dc.b	$0B,$C6,$C9,$AF,$1B,$FC,$FD,$F8,$4F,$E5,$BC,$EF,$E6,$38,$BA,$DF
	.dc.b	$81,$FD,$73,$7F,$13,$FE,$85,$F8,$37,$1D,$F6,$A6,$1F,$E8,$D9,$CB
	.dc.b	$7D,$5E,$1F,$6F,$23,$07,$FA,$4F,$F1,$17,$CF,$7E,$17,$E1,$64,$E6
	.dc.b	$3F,$D2,$9F,$83,$7A,$E3,$FD,$34,$BD,$3C,$9A,$E3,$FF,$4D,$7D,$C7
	.dc.b	$FB,$12,$7F,$D5,$14,$6C,$7F,$B2,$31,$FE,$E1,$CD,$FE,$AF,$E2,$FF
	.dc.b	$04,$71,$FC,$46,$E1,$FD,$5A,$B7,$3F,$22,$FF,$0B,$3F,$EA,$D7,$F0
	.dc.b	$AD,$EF,$F1,$F7,$F9,$3B,$6C,$EE,$30,$F1,$F9,$5A,$FF,$31,$B5,$7B
	.dc.b	$27,$DB,$F0,$9E,$EF,$42,$5F,$86,$32,$D2,$1C,$FE,$13,$FD,$7D,$F7
	.dc.b	$0F,$B3,$A4,$9F,$F6,$34,$3E,$0F,$64,$8C,$BF,$EC,$7F,$B8,$78,$7F
	.dc.b	$EC,$9F,$73,$F6,$B5,$FE,$CF,$6B,$FD,$2D,$BC,$9F,$ED,$D7,$F7,$B1
	.dc.b	$BF,$37,$91,$1F,$F6,$C5,$FE,$D1,$FE,$A9,$F6,$DF,$E0,$5B,$FF,$73
	.dc.b	$37,$93,$F3,$6D,$9F,$F0,$6B,$F9,$47,$8F,$81,$FE,$DF,$8C,$57,$28
	.dc.b	$C7,$95,$2B,$FD,$D1,$5F,$B8,$7C,$31,$FE,$E7,$FB,$8F,$EC,$93,$F2
	.dc.b	$E7,$7B,$F5,$3B,$6D,$23,$9B,$5F,$5F,$AB,$FE,$2B,$5F,$70,$DF,$D7
	.dc.b	$39,$9F,$F1,$5E,$1F,$CC,$34,$FF,$0D,$36,$3F,$C5,$8D,$77,$EB,$3E
	.dc.b	$E3,$6D,$8D,$B5,$DC,$7F,$89,$5F,$F1,$AF,$D6,$D4,$D2,$3F,$E2,$BF
	.dc.b	$85,$70,$FF,$8F,$31,$B3,$FC,$3B,$D6,$B1,$88,$FF,$20,$D6,$2B,$FC
	.dc.b	$8F,$56,$78,$47,$33,$F8,$BF,$C4,$3D,$ED,$4F,$F9,$2A,$AC,$E6,$B8
	.dc.b	$4F,$F1,$EE,$6D,$AF,$B1,$B1,$86,$4F,$09,$FF,$2A,$FC,$2E,$1F,$F1
	.dc.b	$20,$79,$9E,$0F,$9B,$B3,$C7,$04,$79,$B4,$FE,$8B,$3D,$5F,$D9,$37
	.dc.b	$9F,$5C,$D2,$39,$78,$17,$2F,$F1,$A6,$3F,$C1,$27,$CD,$F7,$0E,$0F
	.dc.b	$86,$FF,$66,$8C,$C6,$37,$91,$E5,$EF,$B8,$74,$78,$FA,$7F,$56,$3F
	.dc.b	$A2,$E0,$8C,$47,$F9,$E2,$31,$5F,$E6,$B6,$C7,$F9,$F5,$7C,$87,$08
	.dc.b	$46,$27,$84,$E6,$31,$3C,$7D,$CC,$12,$BB,$3B,$F9,$89,$1E,$16,$CF
	.dc.b	$FA,$37,$EE,$1E,$9F,$85,$B5,$FE,$81,$6C,$EE,$7C,$DF,$3E,$D2,$92
	.dc.b	$FF,$D1,$11,$97,$FD,$15,$F7,$0F,$B8,$F7,$36,$33,$18,$3E,$17,$DC
	.dc.b	$7F,$82,$A3,$FD,$24,$DC,$7F,$D4,$8D,$7F,$D5,$2D,$7F,$BB,$FD,$BF
	.dc.b	$E6,$67,$FD,$34,$D8,$FF,$56,$35,$FF,$58,$B5,$E3,$EA,$FD,$38,$FE
	.dc.b	$E3,$3C,$A3,$07,$D7,$F9,$6F,$8B,$5F,$EB,$A6,$BF,$E0,$86,$BF,$EB
	.dc.b	$A6,$BF,$C0,$BC,$1F,$C0,$57,$F9,$20,$FF,$58,$16,$3F,$D8,$4D,$7F
	.dc.b	$D8,$FF,$6B,$FE,$C4,$8F,$F5,$B3,$63,$FD,$90,$D7,$67,$CC,$F8,$C9
	.dc.b	$3F,$31,$B1,$C3,$8B,$F2,$54,$69,$2B,$FD,$89,$E1,$7A,$89,$C3,$F9
	.dc.b	$97,$DB,$F6,$AF,$B5,$FC,$E3,$6B,$FD,$98,$DC,$7F,$DB,$CD,$7F,$DC
	.dc.b	$7F,$69,$FC,$09,$FF,$71,$35,$CF,$7F,$E9,$E3,$11,$FE,$36,$5F,$31
	.dc.b	$92,$5F,$FB,$6E,$32,$FF,$B7,$7E,$D1,$CD,$FB,$16,$C6,$63,$07,$ED
	.dc.b	$7B,$BF,$73,$87,$FB,$E1,$AF,$F9,$19,$AF,$F3,$8E,$DF,$DE,$7D,$A7
	.dc.b	$FA,$66,$FF,$DD,$8D,$9F,$F7,$F3,$5F,$F2,$78,$70,$FF,$94,$BE,$CD
	.dc.b	$F4,$1F,$C1,$DB,$5F,$EF,$76,$CF,$20,$E1,$EB,$F6,$6F,$B2,$FE,$20
	.dc.b	$DA,$FF,$80,$0D,$9E,$4D,$7C,$BF,$63,$FE,$98,$9E,$D0,$8C,$3E,$16
	.dc.b	$BC,$3E,$57,$F8,$F7,$DF,$7E,$88,$9C,$D8,$3F,$91,$7D,$87,$C0,$A0
	.dc.b	$70,$DF,$9C,$D9,$FC,$A7,$EF,$BF,$C0,$99,$79,$1B,$3F,$0D,$AF,$CE
	.dc.b	$6B,$FE,$7E,$7F,$24,$DE,$8E,$84,$9F,$7D,$BE,$4F,$CE,$48,$CB,$F3
	.dc.b	$3C,$59,$1F,$31,$B7,$F5,$6B,$E6,$3F,$E8,$E6,$BF,$01,$AF,$C1,$DF
	.dc.b	$DD,$F7,$B2,$FD,$CD,$9F,$77,$C5,$7C,$32,$4F,$C3,$7E,$37,$C5,$48
	.dc.b	$F8,$CD,$9F,$8D,$CD,$F2,$C9,$3F,$2D,$F9,$DE,$64,$8F,$9C,$D9,$FC
	.dc.b	$87,$E3,$0F,$83,$FE,$B8,$DA,$FC,$3B,$67,$73,$FD,$87,$EC,$BF,$C0
	.dc.b	$7D,$9F,$8C,$6C,$FE,$6D,$FF,$0E,$76,$67,$D5,$FD,$CC,$FE,$7F,$EF
	.dc.b	$DB,$FF,$8D,$5B,$1F,$96,$6B,$B8,$FF,$1D,$7D,$9B,$EE,$E9,$27,$F3
	.dc.b	$B1,$97,$F3,$FF,$59,$FE,$78,$7F,$6E,$D7,$FD,$8E,$D7,$F3,$2D,$5F
	.dc.b	$66,$F9,$6F,$F2,$53,$63,$F5,$6D,$7F,$74,$D7,$F8,$2F,$DB,$7F,$1E
	.dc.b	$7F,$62,$D7,$F8,$03,$5D,$F8,$FF,$E5,$7B,$FF,$1A,$47,$AB,$DD,$C1
	.dc.b	$1F,$0F,$DC,$79,$08,$E0,$83,$DE,$FF,$2B,$62,$9F,$F7,$3C,$61,$BC
	.dc.b	$BF,$EE,$A9,$78,$78,$BF,$DE,$FD,$87,$F9,$7F,$0F,$F3,$03,$6F,$FD
	.dc.b	$E2,$D9,$DF,$85,$C5,$42,$3F,$CE,$8D,$8F,$E2,$AD,$7F,$92,$35,$FE
	.dc.b	$5E,$FF,$88,$9F,$F2,$E3,$5F,$74,$F8,$BC,$3F,$04,$BE,$86,$98,$89
	.dc.b	$5C,$A3,$7E,$4F,$0F,$97,$93,$DA,$32,$F8,$23,$2F,$80,$F8,$B1,$A7
	.dc.b	$07,$C0,$8D,$E5,$F2,$27,$4F,$CA,$8C,$E3,$F2,$A7,$2F,$CC,$8C,$D7
	.dc.b	$63,$C8,$78,$C8,$78,$31,$FF,$7F,$BF,$42,$B8,$BE,$33,$89,$B7,$30
	.dc.b	$9F,$0C,$F7,$3A,$3A,$9C,$C7,$33,$27,$4B,$F7,$3E,$47,$DC,$E6,$FD
	.dc.b	$CE,$5F,$73,$93,$C1,$1C,$1E,$A7,$9C,$D1,$E1,$74,$78,$1D,$1C,$1C
	.dc.b	$C7,$82,$33,$18,$7F,$3B,$5E,$CB,$F9,$27,$C3,$8F,$31,$E4,$AE,$55
	.dc.b	$C6,$3C,$51,$82,$3D,$0E,$27,$D3,$3D,$33,$D5,$F3,$47,$07,$46,$0E
	.dc.b	$47,$A1,$D4,$71,$9F,$13,$E8,$BF,$B7,$7E,$B6,$DF,$81,$FD,$89,$1F
	.dc.b	$D8,$B5,$FD,$8B,$5F,$D8,$B9,$AF,$9C,$4B,$FE,$03,$E0,$F7,$AF,$6B
	.dc.b	$5F,$41,$FC,$29,$C4,$F1,$8C,$1E,$53,$89,$4E,$27,$F0,$A8,$E3,$3F
	.dc.b	$09,$B1,$F5,$4E,$67,$0C,$3E,$23,$FC,$22,$36,$FC,$EA,$FE,$37,$18
	.dc.b	$9F,$C0,$B6,$FF,$85,$B6,$B9,$1E,$5A,$F5,$3E,$53,$17,$E5,$69,$E5
	.dc.b	$34,$79,$6B,$F2,$15,$D6,$3F,$74,$DA,$F2,$C7,$23,$D1,$5F,$41,$C5
	.dc.b	$FA,$1B,$7E,$86,$9E,$83,$C0,$7C,$E3,$F8,$47,$C9,$FF,$28,$23,$FC
	.dc.b	$BB,$E0,$6B,$A3,$F9,$8C,$7F,$2E,$8F,$E5,$FE,$2C,$9F,$E6,$AF,$F3
	.dc.b	$2E,$F9,$18,$21,$FC,$5D,$FE,$8D,$F0,$AF,$9A,$BF,$61,$2B,$FC,$9B
	.dc.b	$AB,$5F,$F2,$7C,$2F,$83,$A8,$CC,$62,$3F,$CD,$CD,$7F,$CF,$2D,$77
	.dc.b	$7F,$E6,$67,$A3,$E4,$23,$D4,$FF,$2F,$C3,$E3,$FD,$3F,$1D,$C6,$0D
	.dc.b	$24,$FF,$9C,$A1,$F1,$9C,$23,$A1,$EB,$9F,$87,$1F,$13,$EB,$E4,$EA
	.dc.b	$91,$F2,$3E,$22,$4F,$57,$C4,$47,$C1,$DB,$F3,$BE,$72,$47,$38,$FA
	.dc.b	$27,$A4,$8F,$02,$1F,$D7,$2F,$FC,$35,$AF,$D3,$8F,$87,$3F,$6B,$EB
	.dc.b	$FF,$C0,$1F,$5D,$FE,$A6,$AD,$52,$31,$BF,$DA,$6C,$7F,$A2,$97,$91
	.dc.b	$C8,$FA,$8E,$23,$FD,$36,$D7,$FD,$54,$D7,$FD,$6A,$FF,$8A,$9C,$F8
	.dc.b	$1C,$88,$F2,$F9,$1F,$3C,$48,$FA,$72,$9F,$5F,$9F,$EC,$F4,$25,$6A
	.dc.b	$32,$F9,$1F,$39,$1C,$1B,$21,$FE,$06,$A7,$A1,$F2,$3E,$9A,$4F,$89
	.dc.b	$F9,$7F,$01,$23,$E0,$BF,$10,$8F,$C4,$F1,$3E,$39,$1C,$1F,$23,$EB
	.dc.b	$7E,$AC,$48,$F7,$C7,$F8,$DF,$EB,$5F,$5B,$7C,$79,$5F,$96,$9C,$B6
	.dc.b	$7F,$51,$18,$8F,$F6,$E3,$5F,$F7,$3B,$5F,$DC,$BF,$B9,$7F,$DA,$AD
	.dc.b	$78,$F2,$F4,$3F,$69,$F5,$AF,$9C,$A4,$9F,$F6,$EC,$65,$FF,$6F,$FD
	.dc.b	$6B,$E2,$7F,$78,$D8,$CC,$60,$FE,$01,$F5,$BF,$E6,$08,$FF,$73,$B5
	.dc.b	$FF,$7B,$35,$FF,$80,$2D,$7F,$89,$FD,$6B,$E6,$3F,$85,$36,$3F,$DE
	.dc.b	$2D,$7F,$E0,$03,$5F,$E3,$61,$70,$72,$24,$72,$7F,$8B,$7D,$6B,$EA
	.dc.b	$E9,$27,$FD,$FF,$19,$7F,$E0,$0F,$C5,$3C,$BF,$E3,$CD,$8C,$C6,$0F
	.dc.b	$E4,$5F,$5A,$9A,$49,$F3,$9F,$F0,$02,$34,$E0,$79,$8F,$AD,$FE,$58
	.dc.b	$D8,$D3,$5E,$6D,$7E,$1C,$FA,$5E,$ED,$73,$8F,$98,$FE,$71,$F5,$BF
	.dc.b	$AA,$8F,$13,$5F,$3B,$5F,$9E,$D7,$FA,$5F,$D6,$FF,$AE,$5B,$ED,$7F
	.dc.b	$43,$8C,$40,$38,$5E,$B7,$F4,$37,$E1,$35,$F8,$ED,$7E,$FB,$5F,$EB
	.dc.b	$6B,$C8,$FE,$A4,$D8,$E7,$18,$3F,$AB,$7D,$4B,$E5,$69,$27,$E9,$46
	.dc.b	$5F,$A9,$F5,$6F,$0F,$FA,$F3,$63,$31,$83,$FB,$17,$D4,$FF,$38,$8F
	.dc.b	$CA,$B5,$FE,$E2,$D7,$E2,$35,$53,$E1,$FC,$E4,$8F,$2B,$DB,$8E,$47
	.dc.b	$E2,$6B,$F2,$8F,$E3,$9A,$F8,$B8,$7C,$A4,$78,$3E,$8F,$A9,$91,$CD
	.dc.b	$AF,$3F,$5F,$DA,$47,$C0,$D7,$FD,$73,$48,$FF,$6D,$36,$3F,$BB,$AF
	.dc.b	$5F,$FD,$BD,$F5,$2F,$A2,$A4,$AF,$EF,$B1,$FE,$04,$6F,$AD,$C0,$8F
	.dc.b	$F8,$0B,$EA,$1F,$53,$FE,$E9,$F3,$DF,$49,$49,$5F,$E0,$88,$FE,$19
	.dc.b	$EA,$7F,$6E,$6F,$A9,$FE,$26,$6C,$FE,$65,$AB,$0F,$D7,$57,$F7,$6F
	.dc.b	$A7,$FF,$14,$37,$DC,$FF,$7D,$46,$30,$FF,$18,$BF,$9D,$9E,$1E,$E6
	.dc.b	$89,$87,$F8,$72,$7F,$82,$BF,$EE,$46,$CF,$7F,$0B,$A9,$FE,$3C,$7F
	.dc.b	$C0,$FF,$53,$FE,$04,$A5,$7C,$D7,$A7,$07,$90,$F2,$EF,$23,$C0,$D7
	.dc.b	$C1,$EE,$B0,$FD,$D7,$D4,$FF,$7D,$9F,$DE,$61,$FE,$34,$9F,$E3,$4D
	.dc.b	$8F,$E3,$ED,$7F,$9B,$B5,$F8,$7E,$5B,$EA,$78,$B6,$38,$B5,$E2,$D7
	.dc.b	$F9,$5F,$CF,$7F,$94,$53,$48,$FF,$2D,$F7,$3F,$C6,$18,$7E,$D7,$92
	.dc.b	$F0,$FF,$4A,$9B,$5F,$E6,$69,$AF,$F1,$96,$FB,$FF,$E6,$9A,$FF,$28
	.dc.b	$CF,$E3,$9F,$F2,$9C,$66,$3E,$44,$7F,$56,$F3,$B9,$C3,$EA,$6C,$61
	.dc.b	$93,$D3,$DE,$B5,$C8,$FF,$3C,$BC,$8F,$1F,$C8,$59,$F0,$BE,$8C,$FF
	.dc.b	$A9,$47,$F9,$99,$FD,$73,$79,$75,$C9,$23,$D3,$E0,$5C,$7C,$03,$4F
	.dc.b	$EA,$B3,$FE,$70,$FA,$9F,$F2,$7B,$76,$F8,$F1,$8A,$E2,$FC,$56,$C7
	.dc.b	$CA,$F0,$36,$DA,$E9,$EF,$72,$F1,$39,$B6,$3C,$11,$F5,$3F,$10,$FE
	.dc.b	$99,$18,$3F,$D0,$6E,$30,$F0,$FA,$7E,$4F,$2B,$CA,$DA,$F5,$7D,$57
	.dc.b	$A3,$E3,$7D,$2D,$8F,$F4,$6F,$D5,$7F,$56,$7F,$D1,$FF,$8B,$3E,$17
	.dc.b	$F0,$B3,$9C,$BE,$3C,$E6,$BF,$D2,$BF,$2F,$FC,$8D,$BE,$57,$E3,$23
	.dc.b	$15,$F3,$9F,$E4,$93,$FE,$9C,$FA,$B7,$D5,$FE,$49,$FF,$02,$46,$63
	.dc.b	$1E,$D7,$E5,$BE,$AD,$F5,$9F,$98,$FB,$F7,$C6,$66,$30,$FE,$73,$EB
	.dc.b	$7F,$25,$8F,$FA,$A9,$78,$39,$F0,$B4,$EF,$9B,$87,$25,$BE,$0F,$E9
	.dc.b	$BE,$B7,$E1,$78,$5F,$E9,$8F,$5B,$FD,$5F,$5F,$EB,$69,$CB,$FE,$D2
	.dc.b	$8C,$F1,$FE,$3A,$3F,$CA,$07,$FA,$F0,$F8,$25,$F2,$3F,$66,$E2,$3F
	.dc.b	$2B,$F5,$9F,$7F,$0F,$F5,$3B,$F9,$16,$C7,$EE,$A3,$13,$FE,$E2,$E1
	.dc.b	$7D,$6F,$EF,$BE,$FD,$F2,$9F,$2E,$30,$F1,$EE,$E5,$BD,$C3,$FD,$A3
	.dc.b	$8F,$F8,$2D,$FF,$69,$F9,$9E,$EC,$FF,$C4,$0B,$E6,$7F,$85,$B9,$8F
	.dc.b	$E2,$F1,$FC,$3B,$EB,$9F,$4D,$FC,$43,$EF,$DF,$4F,$98,$C3,$FC,$53
	.dc.b	$EB,$FF,$7B,$87,$ED,$9F,$DE,$B7,$CF,$D7,$9F,$BF,$99,$E1,$DB,$76
	.dc.b	$FF,$74,$E1,$F8,$67,$FD,$D7,$F5,$EF,$13,$F9,$0F,$17,$F8,$86,$73
	.dc.b	$18,$F8,$4F,$E4,$BE,$AF,$FA,$CA,$7F,$DE,$6F,$F8,$65,$BE,$76,$BC
	.dc.b	$EF,$E5,$7F,$5E,$F9,$7B,$F2,$F8,$6D,$FF,$BF,$3E,$BD,$F3,$DF,$CC
	.dc.b	$BE,$FF,$DF,$E1,$7F,$B9,$27,$35,$FE,$4D,$8F,$F7,$3C,$EA,$3F,$CA
	.dc.b	$1C,$CF,$AF,$7A,$9F,$E3,$16,$CF,$F9,$C5,$B3,$FE,$76,$6B,$FE,$8E
	.dc.b	$FA,$FF,$DF,$37,$E3,$FF,$C7,$51,$88,$EE,$D7,$FC,$C1,$F7,$DE,$FB
	.dc.b	$FF,$3F,$36,$3F,$D0,$6D,$7F,$D3,$6D,$7D,$3F,$61,$FE,$EB,$9F,$F4
	.dc.b	$33,$63,$FD,$16,$D7,$FD,$46,$D7,$E0,$CF,$FB,$52,$7C,$3B,$FF,$94
	.dc.b	$63,$13,$E2,$8F,$F2,$A7,$D8,$7E,$42,$BF,$D2,$8D,$9F,$F4,$CB,$5F
	.dc.b	$F5,$8B,$5E,$FF,$3D,$D4,$93,$F0,$1E,$BF,$4D,$23,$AB,$67,$D7,$F5
	.dc.b	$BF,$E6,$5C,$8D,$25,$7C,$AF,$83,$F5,$11,$C1,$F1,$BA,$79,$11,$C1
	.dc.b	$F4,$B6,$F8,$44,$70,$7C,$EE,$7F,$11,$1C,$1F,$6B,$C6,$F9,$44,$70
	.dc.b	$7F,$AE,$FE,$BD,$87,$EA,$F3,$FF,$60,$C6,$3B,$35,$B3,$B3,$85,$EE
	.dc.b	$FF,$DB,$BF,$5F,$FD,$3D,$FF,$63,$35,$FF,$65,$B5,$DF,$BF,$FE,$CB
	.dc.b	$5F,$59,$D5,$FF,$67,$35,$FF,$69,$B5,$FF,$75,$B5,$FD,$4F,$AB,$FA
	.dc.b	$87,$FD,$AC,$D7,$FD,$B6,$D7,$FD,$E6,$D7,$F6,$31,$FE,$F7,$6F,$C3
	.dc.b	$F9,$63,$11,$FE,$C5,$6B,$E8,$F6,$34,$47,$D2,$FB,$7E,$F7,$F8,$54
	.dc.b	$7F,$BA,$5A,$FF,$BB,$9A,$FF,$02,$0E,$1F,$E0,$BF,$5A,$FA,$8F,$90
	.dc.b	$D8,$FF,$7A,$35,$FF,$7C,$B5,$E9,$F7,$BF,$1D,$FF,$6F,$35,$34,$9B
	.dc.b	$7F,$9D,$F8,$BF,$8A,$C7,$D3,$F0,$51,$89,$D5,$7C,$5F,$95,$AB,$EE
	.dc.b	$DB,$E1,$B1,$FC,$8E,$FF,$90,$C6,$23,$FC,$77,$B7,$F4,$E7,$47,$D1
	.dc.b	$F4,$7E,$A1,$1F,$C3,$BF,$5F,$AF,$C3,$3F,$B4,$9E,$27,$E3,$A3,$CB
	.dc.b	$F8,$0F,$AA,$32,$D7,$C3,$1C,$8D,$E1,$EC,$FA,$F7,$7D,$F2,$25,$F6
	.dc.b	$9C,$B6,$72,$FF,$95,$A3,$FA,$3B,$92,$19,$9D,$EF,$E3,$9E,$73,$C6
	.dc.b	$B3,$7F,$D2,$27,$9C,$E4,$E7,$5C,$6F,$70,$13,$D2,$7B,$5E,$A7,$35
	.dc.b	$EF,$89,$1E,$D8,$D5,$71,$BF,$15,$62,$77,$5E,$33,$D8,$71,$3C,$B3
	.dc.b	$C1,$C1,$A3,$85,$F9,$1C,$BE,$B7,$89,$EB,$BF,$13,$6B,$CB,$7E,$2A
	.dc.b	$E1,$3D,$5E,$8F,$6B,$F9,$55,$C2,$3A,$9F,$08,$F0,$C7,$9D,$E0,$77
	.dc.b	$32,$78,$EB,$C1,$1D,$23,$27,$D0,$8C,$BC,$4F,$24,$75,$AF,$29,$E8
	.dc.b	$23,$92,$47,$18,$CB,$E1,$3D,$37,$F8,$48,$CC,$78,$4C,$DF,$BD,$F5
	.dc.b	$9E,$F3,$C6,$73,$33,$3C,$EB,$D9,$18,$BF,$E7,$53,$F9,$68,$FE,$34
	.dc.b	$D7,$E4,$18,$8D,$91,$FE,$40,$77,$34,$7A,$9F,$E1,$7E,$D3,$E8,$FE
	.dc.b	$4C,$FA,$CF,$4C,$65,$E0,$7D,$A7,$0F,$9C,$C4,$72,$AF,$C4,$C7,$3C
	.dc.b	$7F,$81,$57,$07,$E9,$E1,$F4,$27,$4F,$F8,$E2,$74,$E6,$B9,$9D,$23
	.dc.b	$33,$F2,$EB,$9B,$6B,$95,$F8,$F0,$F6,$DF,$0A,$EE,$6C,$87,$FA,$12
	.dc.b	$78,$3E,$C3,$E0,$9F,$11,$CB,$4C,$DF,$89,$E4,$62,$FD,$D5,$E1,$C3
	.dc.b	$D7,$87,$93,$0E,$78,$71,$AD,$4E,$2F,$D7,$87,$33,$CF,$3A,$70,$78
	.dc.b	$0F,$A8,$79,$4B,$5D,$6F,$A4,$72,$6B,$F1,$0E,$51,$FC,$9F,$E4,$3F
	.dc.b	$9B,$12,$3F,$A0,$EA,$FA,$2C,$47,$F3,$39,$B5,$FD,$06,$B5,$3F,$D2
	.dc.b	$3E,$F9,$F6,$3F,$D0,$7E,$32,$4E,$63,$FA,$FC,$EA,$3F,$A1,$4E,$95
	.dc.b	$FF,$4D,$9D,$46,$34,$FF,$40,$F8,$0F,$9A,$FE,$D2,$BE,$6E,$98,$89
	.dc.b	$5F,$A8,$13,$C5,$C3,$5F,$D6,$B5,$FF,$07,$7D,$E3,$DB,$E7,$1E,$33
	.dc.b	$FA,$34,$F1,$CB,$FB,$25,$6E,$7F,$B5,$CB,$E6,$BF,$C7,$4D,$8F,$0C
	.dc.b	$60,$8F,$07,$DA,$7F,$81,$A5,$F7,$3F,$E0,$8F,$4F,$E8,$D7,$8C,$F6
	.dc.b	$FC,$C7,$C4,$35,$C7,$23,$E1,$4F,$07,$FC,$54,$D7,$FD,$5A,$D7,$DB
	.dc.b	$EF,$7C,$E2,$3E,$CF,$7A,$BE,$1C,$D3,$8D,$3F,$C0,$D0,$AF,$AE,$7C
	.dc.b	$22,$57,$D2,$7F,$CB,$DF,$65,$FE,$2F,$27,$85,$83,$FC,$C5,$F3,$DF
	.dc.b	$E5,$7C,$3F,$C2,$53,$FE,$2F,$E3,$FF,$62,$B7,$0F,$EF,$55,$B9,$CF
	.dc.b	$87,$FE,$30,$A7,$C7,$53,$11,$2F,$F1,$12,$8F,$29,$23,$3E,$9F,$A0
	.dc.b	$8E,$08,$7D,$59,$47,$84,$F1,$7B,$D2,$7B,$BF,$08,$8F,$8B,$E2,$3E
	.dc.b	$29,$1C,$1F,$9D,$FB,$BF,$F3,$26,$9F,$12,$7D,$C7,$F9,$3A,$33,$9F
	.dc.b	$C6,$FB,$2F,$85,$E6,$4C,$3F,$C8,$39,$A7,$E3,$1F,$50,$91,$F1,$1F
	.dc.b	$F4,$A7,$D9,$3D,$BA,$49,$FD,$84,$65,$B1,$FE,$60,$8C,$47,$E6,$5B
	.dc.b	$9F,$F9,$35,$FD,$BC,$7F,$95,$3E,$EB,$FA,$8C,$62,$38,$72,$7F,$9A
	.dc.b	$A7,$FA,$AF,$D9,$3E,$CF,$FC,$DA,$DA,$FF,$2E,$B5,$FD,$42,$F9,$FF
	.dc.b	$F3,$9F,$D9,$BE,$52,$92,$BF,$CC,$B0,$F9,$BF,$F3,$CF,$D9,$FF,$A2
	.dc.b	$F5,$FF,$32,$37,$8B,$FC,$95,$68,$C3,$FE,$4D,$D1,$F8,$07,$FA,$0B
	.dc.b	$ED,$1E,$DD,$25,$7F,$9D,$21,$ED,$EF,$A2,$4E,$5E,$5F,$1B,$FE,$8C
	.dc.b	$FB,$4F,$EC,$84,$F4,$7E,$37,$CE,$FF,$A4,$3E,$49,$F3,$74,$98,$FF
	.dc.b	$A0,$A1,$EA,$7F,$A5,$BE,$D1,$14,$93,$FE,$88,$8F,$E4,$0F,$F1,$39
	.dc.b	$7D,$07,$FA,$79,$B1,$FE,$8E,$6B,$FC,$67,$87,$C6,$F3,$CB,$C0,$47
	.dc.b	$FD,$2B,$F6,$9F,$9A,$7F,$CE,$CB,$C3,$FF,$55,$B6,$37,$18,$F8,$AF
	.dc.b	$F5,$7F,$DA,$7F,$28,$E8,$FF,$4D,$37,$A3,$EB,$B6,$FE,$B6,$7F,$EB
	.dc.b	$6F,$8C,$FE,$7D,$5F,$71,$B3,$FE,$87,$6A,$7E,$3B,$FE,$AC,$5F,$69
	.dc.b	$CC,$91,$9F,$3B,$FD,$87,$F6,$6E,$6A,$49,$FD,$17,$D9,$FF,$B2,$3E
	.dc.b	$D3,$F8,$5C,$7F,$AE,$1A,$FF,$43,$6E,$BF,$EB,$77,$E2,$47,$F4,$7F
	.dc.b	$B4,$FF,$5D,$4F,$FA,$ED,$BA,$7E,$61,$B3,$FE,$A3,$7F,$DE,$73,$FE
	.dc.b	$C1,$C5,$BF,$A0,$95,$FB,$7F,$80,$FF,$6E,$DF,$DC,$8F,$07,$46,$0C
	.dc.b	$F0,$48,$FF,$68,$CF,$E4,$1B,$1C,$B4,$91,$A6,$BF,$89,$F1,$BB,$F8
	.dc.b	$B2,$39,$BE,$2F,$7A,$46,$08,$7C,$37,$FD,$E7,$F6,$9F,$9A,$DB,$FD
	.dc.b	$BC,$DD,$FF,$76,$DA,$FF,$63,$2F,$B2,$E2,$48,$FF,$75,$C6,$5F,$F7
	.dc.b	$6F,$D9,$BE,$D3,$FE,$00,$FA,$8F,$B0,$A4,$AF,$F7,$9C,$7F,$04,$6F
	.dc.b	$8F,$C0,$8F,$4E,$EF,$E1,$FA,$7F,$BD,$7E,$CF,$FC,$49,$3F,$EF,$C6
	.dc.b	$BF,$E0,$76,$C6,$FE,$0B,$8E,$DF,$C6,$27,$F3,$64,$8C,$D7,$FB,$96
	.dc.b	$3F,$3A,$F3,$F1,$32,$A3,$37,$E3,$9C,$FB,$3F,$B9,$AD,$7C,$17,$0B
	.dc.b	$F9,$91,$83,$D1,$C3,$92,$38,$3D,$9F,$05,$FE,$48,$5F,$69,$FD,$DD
	.dc.b	$F3,$46,$5F,$E6,$67,$95,$C3,$EF,$6B,$C3,$2F,$F2,$45,$FF,$43,$FB
	.dc.b	$4B,$E9,$9A,$4A,$FF,$27,$0F,$A8,$AD,$32,$E0,$DF,$BE,$E0,$4F,$F8
	.dc.b	$EC,$FC,$23,$8C,$75,$1F,$D3,$26,$C6,$23,$8F,$26,$2B,$DE,$D8,$EB
	.dc.b	$F0,$7F,$E6,$29,$E8,$68,$B1,$AF,$63,$83,$E4,$E5,$AF,$F5,$29,$DB
	.dc.b	$96,$B9,$3D,$5E,$97,$0F,$80,$E2,$D9,$DC,$76,$C1,$F2,$F9,$BF,$F2
	.dc.b	$E4,$62,$BB,$E7,$FD,$49,$BC,$9E,$68,$C5,$7F,$98,$DF,$F3,$C4,$E8
	.dc.b	$F3,$D7,$F5,$C7,$13,$DB,$C1,$F7,$FD,$A3,$7B,$BE,$1B,$7F,$02,$7E
	.dc.b	$AD,$F9,$BC,$0C,$B6,$75,$E7,$74,$F1,$FA,$B6,$39,$47,$CE,$FB,$4F
	.dc.b	$C0,$C6,$0F,$9A,$E2,$BF,$CF,$D8,$3D,$6F,$3F,$99,$E8,$6D,$F9,$7E
	.dc.b	$53,$D7,$F0,$DE,$C6,$C7,$D8,$FB,$4F,$EA,$CF,$DA,$FB,$47,$CD,$7E
	.dc.b	$02,$73,$9F,$FA,$36,$73,$5F,$80,$FB,$56,$1F,$86,$58,$D8,$C3,$F8
	.dc.b	$AF,$B4,$FA,$78,$7F,$A5,$DF,$E5,$17,$F8,$A7,$3C,$DF,$9B,$73,$3F
	.dc.b	$E9,$DF,$C7,$BC,$2D,$F8,$BF,$CA,$C6,$2B,$F4,$4F,$F9,$BA,$FF,$2B
	.dc.b	$F8,$3F,$FB,$6D,$BF,$1D,$F9,$C8,$69,$63,$F3,$FE,$A3,$E8,$29,$8F
	.dc.b	$BB,$B8,$7E,$83,$F1,$EF,$C0,$E7,$FB,$16,$C7,$F9,$01,$FF,$1D,$BB
	.dc.b	$EC,$FD,$5F,$2E,$37,$FF,$5C,$36,$FF,$74,$FE,$A5,$BE,$1F,$EC,$BF
	.dc.b	$08,$7C,$CD,$31,$F1,$17,$2F,$D9,$FD,$B3,$EE,$FF,$72,$D8,$CF,$89
	.dc.b	$89,$FD,$EF,$E1,$1F,$E5,$B9,$FD,$DB,$FE,$38,$6F,$C6,$EB,$E2,$3F
	.dc.b	$81,$FE,$1B,$F8,$09,$FE,$02,$FE,$01,$BD,$BA,$ED,$FE,$13,$F6,$CF
	.dc.b	$80,$BF,$05,$86,$DF,$F0,$AF,$C2,$7F,$CC,$6D,$E0,$FE,$23,$0F,$8D
	.dc.b	$B1,$FC,$53,$F0,$FF,$ED,$55,$FC,$49,$BD,$1F,$EE,$35,$EF,$6F,$E0
	.dc.b	$3F,$8D,$F9,$3E,$7A,$FF,$74,$37,$AB,$F9,$52,$F9,$8D,$F0,$7F,$22
	.dc.b	$FC,$2B,$F1,$97,$FE,$EF,$7F,$CA,$19,$FF,$22,$73,$87,$F3,$37,$33
	.dc.b	$FE,$F5,$F8,$BF,$F3,$7D,$7F,$28,$7F,$89,$46,$7C,$3F,$E5,$F1,$8B
	.dc.b	$FE,$79,$F6,$7F,$66,$BF,$DF,$CB,$AF,$F9,$AB,$99,$DC,$7F,$38,$FC
	.dc.b	$13,$E4,$63,$FD,$82,$FF,$A6,$C6,$1C,$35,$FE,$CA,$D7,$FD,$33,$F6
	.dc.b	$2F,$B6,$FE,$9D,$18,$6B,$FD,$62,$35,$5D,$E1,$18,$7C,$5F,$65,$FE
	.dc.b	$EB,$9D,$36,$33,$F1,$1D,$14,$65,$FE,$F2,$BE,$AF,$4D,$AE,$71,$83
	.dc.b	$95,$BE,$EA,$92,$7C,$D1,$97,$D1,$97,$FB,$0D,$BE,$CF,$8E,$31,$8F
	.dc.b	$C8,$8E,$BE,$2C,$AE,$6D,$9F,$A1,$D9,$FC,$2E,$7E,$23,$63,$3E,$67
	.dc.b	$BE,$31,$1E,$EF,$57,$DA,$4A,$F8,$EF,$C0,$F4,$A4,$7C,$A6,$CF,$CE
	.dc.b	$F6,$7E,$29,$27,$E7,$BF,$23,$D9,$91,$F4,$9B,$3F,$87,$FC,$11,$F1
	.dc.b	$DF,$4D,$B5,$F7,$5B,$3B,$9F,$B1,$F6,$7F,$DE,$B4,$FC,$13,$67,$F2
	.dc.b	$4B,$E0,$FF,$B7,$46,$3E,$07,$F9,$83,$F9,$7F,$C1,$2B,$6F,$FC,$5B
	.dc.b	$18,$8F,$F1,$DB,$5F,$F7,$7A,$F5,$B6,$48,$FC,$A4,$65,$FC,$B7,$D5
	.dc.b	$FF,$9C,$5B,$EB,$7F,$8E,$A3,$11,$FA,$96,$BF,$E4,$AF,$B3,$FC,$B2
	.dc.b	$69,$36,$FF,$21,$FD,$9B,$DA,$A4,$AF,$D3,$46,$5F,$1F,$C3,$78,$48
	.dc.b	$E0,$DF,$8D,$D0,$8E,$0F,$D8,$7D,$67,$F8,$A9,$BB,$FF,$95,$23,$11
	.dc.b	$FC,$0D,$AF,$F9,$93,$EB,$BD,$64,$CB,$FC,$B6,$D8,$FF,$1F,$BF,$EC
	.dc.b	$44,$D3,$FD,$8B,$1F,$E5,$A8,$7D,$93,$F0,$8F,$F4,$2B,$63,$FC,$92
	.dc.b	$D5,$E3,$7F,$96,$EF,$FD,$D5,$1E,$0F,$9D,$F1,$91,$FE,$1D,$1A,$7F
	.dc.b	$8F,$2F,$B0,$FF,$49,$B6,$3D,$31,$8F,$B1,$91,$82,$1E,$2C,$BD,$54
	.dc.b	$9E,$7E,$67,$FA,$7B,$8F,$FD,$C3,$7F,$E6,$26,$BC,$4F,$F3,$DD,$7F
	.dc.b	$B8,$72,$F6,$57,$1E,$C9,$5F,$32,$3A,$BC,$E3,$2F,$33,$E7,$C6,$9C
	.dc.b	$5F,$82,$BF,$D0,$4D,$C7,$D3,$5B,$9C,$DF,$47,$C7,$7E,$13,$C0,$43
	.dc.b	$C1,$87,$A6,$F9,$C6,$67,$9B,$C4,$DD,$F5,$9F,$C1,$61,$E3,$9F,$87
	.dc.b	$3E,$39,$CE,$1B,$9C,$9C,$EF,$AB,$EE,$8E,$0E,$8C,$1E,$39,$F2,$BE
	.dc.b	$23,$07,$FA,$B2,$FC,$4F,$CA,$3B,$47,$89,$D9,$93,$91,$D8,$95,$F2
	.dc.b	$1C,$BE,$62,$47,$89,$D1,$88,$F1,$57,$A9,$ED,$19,$55,$E6,$C3,$E8
	.dc.b	$47,$BE,$75,$39,$C3,$93,$EC,$31,$3E,$C3,$B9,$EC,$AE,$6F,$83,$0F
	.dc.b	$80,$FB,$1F,$41,$1E,$26,$4F,$13,$97,$CF,$87,$D9,$7A,$1A,$38,$E1
	.dc.b	$ED,$AF,$08,$C3,$F8,$69,$27,$F7,$6D,$7F,$76,$D7,$F8,$B3,$99,$F9
	.dc.b	$B7,$E3,$A7,$C6,$7C,$D7,$BD,$7F,$02,$8C,$39,$6D,$FC,$73,$F8,$DB
	.dc.b	$87,$EE,$D7,$F0,$28,$C5,$F8,$27,$F7,$B1,$89,$F1,$36,$3E,$11,$D0
	.dc.b	$ED,$7E,$47,$F9,$5C,$6D,$F9,$26,$CF,$09,$CA,$3E,$4C,$6D,$B1,$D2
	.dc.b	$FD,$8D,$3B,$CE,$E3,$E1,$46,$4E,$A7,$7B,$F6,$3E,$2A,$EA,$E9,$F4
	.dc.b	$3F,$C1,$2B,$A3,$EA,$3A,$4F,$29,$FA,$71,$8B,$F8,$AD,$3D,$95,$C9
	.dc.b	$E8,$64,$F1,$3F,$CE,$BC,$39,$1F,$D4,$9F,$E7,$BD,$B8,$8F,$E2,$F1
	.dc.b	$70,$FE,$95,$3F,$D4,$D5,$FF,$55,$EC,$7D,$8E,$63,$1F,$80,$5C,$47
	.dc.b	$F8,$5D,$C3,$FE,$19,$9B,$5C,$A3,$77,$D9,$E3,$0A,$D9,$C2,$B8,$FC
	.dc.b	$BF,$4F,$B2,$7E,$05,$4C,$E1,$D3,$2F,$45,$7F,$9B,$63,$FC,$54,$DC
	.dc.b	$4D,$31,$F3,$1C,$A7,$FA,$54,$E9,$E8,$79,$A3,$D7,$F0,$9C,$C9,$1F
	.dc.b	$E7,$B9,$FF,$66,$36,$36,$F2,$3F,$C7,$8F,$C4,$F4,$BE,$2F,$D1,$FC
	.dc.b	$6F,$56,$56,$ED,$3A,$8E,$CF,$AC,$8F,$FA,$2F,$EC,$3F,$C1,$8F,$F6
	.dc.b	$36,$F0,$7C,$F8,$46,$34,$FF,$09,$3F,$E4,$E8,$E0,$7D,$A7,$4F,$DE
	.dc.b	$6B,$CB,$E5,$34,$7C,$F8,$E5,$1E,$26,$BD,$A3,$5E,$0C,$8F,$43,$5F
	.dc.b	$F5,$94,$BD,$5F,$14,$C9,$F9,$53,$96,$F9,$BE,$7F,$3A,$46,$08,$74
	.dc.b	$8F,$31,$1F,$91,$F1,$BE,$92,$38,$21,$F2,$DF,$88,$7C,$5F,$8C,$91
	.dc.b	$F4,$5F,$84,$47,$EA,$7B,$7F,$14,$8E,$0F,$B1,$F6,$1F,$3E,$24,$7B
	.dc.b	$A3,$F5,$7F,$5E,$FB,$0B,$EE,$4A,$FC,$64,$E5,$B3,$FA,$18,$C4,$7F
	.dc.b	$B7,$5A,$FE,$19,$AF,$F9,$4D,$7D,$C7,$12,$47,$FB,$5E,$32,$D8,$F2
	.dc.b	$91,$F1,$FB,$7F,$C0,$BE,$AD,$E0,$FA,$A2,$4F,$E8,$E7,$2F,$F9,$17
	.dc.b	$5F,$F2,$46,$B2,$30,$7E,$A7,$E8,$5B,$7F,$0E,$5B,$53,$11,$E1,$F0
	.dc.b	$BF,$89,$FE,$DB,$FE,$A6,$E8,$FD,$13,$7A,$3F,$CE,$5A,$7F,$AE,$B2
	.dc.b	$FF,$2B,$7C,$1F,$F9,$64,$8F,$F9,$67,$E3,$3C,$1E,$7C,$8F,$23,$FC
	.dc.b	$83,$F0,$8F,$F4,$06,$3F,$EA,$99,$FF,$2A,$37,$2F,$0F,$B3,$E4,$F2
	.dc.b	$3F,$CC,$DE,$A7,$F9,$9F,$79,$87,$F9,$9E,$3A,$74,$F5,$E9,$93,$FC
	.dc.b	$2F,$3F,$F4,$5B,$FC,$CB,$3F,$E5,$CF,$83,$E9,$7E,$09,$F0,$BC,$AF
	.dc.b	$F3,$A7,$24,$9F,$F3,$A3,$FE,$77,$DF,$D5,$E1,$C9,$EC,$FB,$7D,$5F
	.dc.b	$97,$F4,$B2,$7F,$8E,$64,$8F,$9F,$F3,$FF,$D0,$BD,$7F,$F5,$11,$3A
	.dc.b	$31,$C1,$F1,$FD,$69,$8F,$AD,$F9,$5E,$07,$D8,$F3,$64,$F6,$7E,$DF
	.dc.b	$9B,$D7,$93,$B6,$F2,$BF,$97,$67,$FA,$0E,$9F,$E4,$1C,$AF,$07,$EB
	.dc.b	$7E,$11,$2F,$E5,$47,$E1,$BE,$93,$F0,$FF,$2B,$27,$DB,$1F,$D7,$3E
	.dc.b	$AB,$FC,$8A,$4F,$5B,$1E,$AC,$AF,$E7,$D3,$FD,$F1,$FF,$4A,$37,$3F
	.dc.b	$E6,$4B,$E1,$7A,$12,$78,$BF,$A9,$FA,$AF,$07,$95,$27,$AB,$E3,$F1
	.dc.b	$E4,$60,$87,$81,$FE,$E7,$F5,$4E,$AA,$49,$FE,$A9,$1F,$D7,$DF,$F5
	.dc.b	$53,$72,$FE,$90,$DE,$AF,$F6,$C3,$67,$FC,$03,$F5,$4F,$90,$FF,$00
	.dc.b	$FA,$6F,$84,$A4,$AF,$EC,$50,$F9,$DD,$92,$32,$FF,$82,$AD,$F2,$5F
	.dc.b	$E0,$CF,$A9,$FC,$EE,$9F,$D6,$DB,$A7,$FB,$AD,$B8,$7F,$85,$17,$91
	.dc.b	$C4,$91,$FD,$CE,$32,$DF,$0F,$B1,$1C,$1F,$C4,$4F,$16,$F1,$E4,$F9
	.dc.b	$7C,$09,$1D,$5A,$FF,$B5,$69,$E0,$7F,$BE,$A1,$F5,$5F,$D6,$D4,$EB
	.dc.b	$3F,$EF,$71,$EB,$AE,$0F,$F7,$E9,$D3,$A6,$B9,$EA,$7D,$57,$28,$EB
	.dc.b	$18,$78,$35,$E1,$1F,$E2,$95,$F5,$54,$8F,$F8,$B5,$FF,$17,$7B,$7F
	.dc.b	$E3,$BC,$7F,$C1,$D3,$F8,$1D,$FF,$C4,$0A,$75,$1F,$E1,$6E,$4B,$3F
	.dc.b	$E4,$A3,$81,$FE,$FB,$F7,$9E,$26,$A3,$FC,$9F,$36,$73,$5D,$B9,$75
	.dc.b	$4F,$AA,$B1,$86,$4F,$93,$DD,$B4,$CF,$D3,$1C,$8E,$DE,$65,$AF,$F1
	.dc.b	$D6,$28,$F2,$57,$F2,$19,$FF,$1F,$3F,$AF,$6F,$2E,$B9,$64,$72,$F0
	.dc.b	$6E,$5E,$F7,$DD,$F5,$3F,$1E,$7F,$C9,$0F,$C4,$C3,$FC,$74,$E6,$BD
	.dc.b	$13,$F2,$63,$4D,$DB,$E3,$77,$F4,$6C,$74,$F9,$1E,$FF,$0D,$E0,$6C
	.dc.b	$78,$23,$E8,$7D,$57,$F9,$FA,$30,$7F,$96,$DC,$63,$E2,$3F,$D0,$78
	.dc.b	$BD,$CF,$5F,$B5,$EC,$6D,$FB,$3D,$AF,$7F,$AF,$F0,$1B,$1F,$E6,$7F
	.dc.b	$AA,$7A,$7F,$81,$FA,$A7,$C1,$7E,$0E,$73,$5F,$E6,$D9,$CB,$FE,$6F
	.dc.b	$F8,$06,$1F,$89,$58,$D8,$C3,$F8,$DF,$AA,$FB,$1B,$7A,$C3,$FE,$88
	.dc.b	$BF,$F3,$D3,$AE,$4F,$F3,$63,$99,$E1,$F5,$5E,$26,$FC,$27,$E6,$63
	.dc.b	$15,$FE,$73,$51,$F9,$DF,$AA,$7C,$95,$31,$A5,$CB,$FD,$15,$EA,$3E
	.dc.b	$57,$F4,$8D,$8C,$FC,$26,$27,$F5,$1F,$55,$F8,$4D,$7F,$D6,$ED,$4F
	.dc.b	$F3,$3A,$FF,$4B,$39,$E5,$C4,$FC,$36,$ED,$FE,$8B,$7F,$5D,$5F,$E9
	.dc.b	$D7,$2F,$EE,$FE,$A9,$F3,$5F,$B9,$6C,$7F,$A8,$D7,$D1,$66,$7F,$7D
	.dc.b	$F5,$4F,$90,$A6,$3C,$6B,$87,$FA,$AF,$EA,$9E,$F7,$F0,$55,$F5,$56
	.dc.b	$30,$FF,$09,$FA,$AF,$C1,$D7,$FA,$CD,$FC,$1B,$7A,$F5,$D7,$FC,$3F
	.dc.b	$EA,$BF,$BB,$4F,$FA,$E5,$7D,$56,$7E,$07,$4F,$F1,$5F,$AA,$F5,$D7
	.dc.b	$FA,$FD,$BC,$BF,$8B,$5F,$55,$BE,$3F,$E3,$9F,$55,$FD,$A6,$FF,$D8
	.dc.b	$CD,$BF,$C9,$3F,$D8,$A3,$3F,$03,$FC,$8F,$C7,$F4,$D7,$E5,$DB,$D3
	.dc.b	$FE,$BF,$7F,$CC,$F5,$FE,$D1,$72,$FF,$2F,$FA,$AF,$C8,$BF,$ED,$55
	.dc.b	$E6,$7F,$2F,$73,$7F,$A3,$73,$3F,$ED,$BF,$99,$78,$1F,$CD,$57,$A1
	.dc.b	$63,$0F,$F3,$9F,$AA,$FB,$95,$FE,$E3,$5D,$59,$EB,$D3,$FC,$FF,$E9
	.dc.b	$FA,$D7,$F5,$96,$E7,$FD,$91,$AF,$FB,$81,$AF,$F4,$BF,$AA,$FF,$65
	.dc.b	$4F,$F6,$16,$C7,$F6,$C6,$BF,$EE,$86,$BF,$D4,$FE,$F3,$FD,$61,$3F
	.dc.b	$DA,$5B,$1F,$DD,$1A,$FF,$BC,$1A,$FF,$5B,$FA,$AF,$F2,$84,$FF,$71
	.dc.b	$6C,$7F,$7C,$6B,$FE,$F8,$6B,$E1,$F7,$3F,$B5,$A1,$5F,$D8,$1F,$14
	.dc.b	$69,$CB,$67,$F5,$F1,$DF,$C2,$C1,$2F,$E2,$7D,$CF,$1C,$4D,$24,$7C
	.dc.b	$9F,$47,$44,$70,$7C,$DF,$37,$91,$1C,$1F,$27,$D2,$F3,$69,$FC,$E3
	.dc.b	$2F,$95,$18,$3C,$5E,$AC,$AF,$B0,$D9,$FA,$1F,$27,$C8,$93,$F5,$5F
	.dc.b	$85,$F5,$5F,$DD,$7C,$DF,$F1,$03,$6B,$FC,$56,$FF,$83,$3E,$03,$3E
	.dc.b	$EF,$F2,$B9,$FF,$17,$FD,$E7,$F4,$E7,$FC,$50,$D7,$3B,$7F,$8A,$D4
	.dc.b	$65,$FF,$0D,$0F,$AB,$FF,$61,$38,$6B,$FE,$1F,$6B,$F7,$5A,$FF,$58
	.dc.b	$F2,$3B,$79,$3F,$D6,$23,$FC,$74,$D7,$FC,$8A,$D7,$C3,$F4,$DF,$E6
	.dc.b	$D7,$FC,$80,$D7,$FC,$96,$D7,$7F,$07,$EB,$FC,$21,$F4,$94,$93,$FE
	.dc.b	$49,$8C,$BD,$3E,$17,$E1,$5F,$F8,$BE,$30,$43,$FC,$B3,$F5,$7F,$EF
	.dc.b	$87,$FC,$A4,$D7,$3F,$41,$F2,$E3,$11,$F2,$BE,$F7,$FC,$90,$FF,$96
	.dc.b	$5A,$FF,$99,$1A,$EF,$87,$EA,$7D,$4B,$E7,$A9,$27,$FC,$C5,$19,$7F
	.dc.b	$CC,$9F,$59,$FE,$06,$6E,$FF,$76,$31,$1F,$E7,$26,$BF,$79,$CF,$98
	.dc.b	$E9,$FF,$2D,$46,$1B,$5F,$E5,$D9,$D5,$7E,$EF,$57,$D7,$E6,$FF,$CC
	.dc.b	$58,$BA,$B1,$7F,$C0,$63,$0F,$F0,$3C,$39,$A8,$CB,$FE,$19,$C3,$F7
	.dc.b	$6E,$EF,$FC,$DD,$18,$8F,$C1,$F7,$FD,$A2,$3F,$1B,$C1,$F0,$BD,$7E
	.dc.b	$B2,$E0,$6C,$C9,$EC,$79,$E3,$66,$5F,$F1,$64,$FF,$B8,$F0,$F7,$F9
	.dc.b	$5D,$0E,$BE,$54,$AE,$67,$B6,$34,$E2,$B9,$3F,$16,$72,$F6,$9E,$4F
	.dc.b	$82,$33,$9F,$C8,$AF,$EB,$33,$CE,$77,$39,$AD,$DF,$D4,$9C,$BE,$8A
	.dc.b	$F1,$4E,$1F,$11,$82,$1E,$13,$B1,$8A,$EC,$E0,$D1,$E9,$35,$3E,$17
	.dc.b	$51,$C1,$D1,$C1,$F6,$57,$A2,$3F,$11,$5E,$57,$F0,$D1,$97,$E6,$1D
	.dc.b	$2B,$B3,$D2,$BB,$1D,$1D,$1D,$0F,$29,$F0,$0F,$14,$7C,$57,$93,$D2
	.dc.b	$39,$46,$0E,$47,$03,$93,$E0,$8E,$4F,$02,$38,$24,$70,$7B,$18,$7D
	.dc.b	$A7,$53,$E0,$9E,$83,$31,$D1,$E4,$79,$1F,$13,$83,$89,$E9,$38,$C7
	.dc.b	$81,$D5,$FE,$31,$D3,$F9,$67,$8C,$F4,$78,$9C,$2B,$F8,$C9,$1F,$E2
	.dc.b	$ED,$7F,$A6,$9E,$23,$FA,$7C,$FF,$8C,$1D,$90,$F8,$0F,$87,$3F,$F4
	.dc.b	$6C,$7D,$A3,$C7,$7E,$95,$1A,$78,$D7,$E2,$63,$17,$F1,$E7,$89,$F1
	.dc.b	$63,$35,$CE,$FE,$01,$E1,$F8,$1F,$C9,$38,$C3,$F8,$6D,$70,$BF,$84
	.dc.b	$79,$5C,$3D,$89,$3C,$DF,$53,$CD,$FE,$0F,$1C,$23,$11,$F8,$18,$C4
	.dc.b	$F8,$8F,$0D,$59,$F0,$35,$E0,$DC,$3E,$05,$72,$9F,$0D,$72,$69,$C8
	.dc.b	$A7,$22,$CF,$2A,$E9,$87,$D1,$BE,$95,$F4,$0F,$5C,$E9,$F9,$CD,$7E
	.dc.b	$24,$70,$AF,$A9,$87,$C5,$9E,$47,$84,$FE,$63,$F3,$5F,$CE,$88,$FF
	.dc.b	$2E,$FA,$07,$D6,$E2,$36,$7F,$2F,$9F,$E8,$73,$A8,$FE,$6B,$E5,$3E
	.dc.b	$9B,$FA,$3B,$63,$31,$8F,$99,$91,$82,$1F,$EE,$F9,$47,$2F,$97,$E7
	.dc.b	$F3,$32,$BB,$C6,$5F,$D1,$8A,$47,$F8,$D5,$B1,$A6,$BF,$AA,$6E,$9F
	.dc.b	$DA,$9B,$BF,$F9,$BE,$7A,$57,$F6,$2F,$AD,$7D,$E7,$BB,$D2,$EC,$4A
	.dc.b	$EC,$FF,$91,$7C,$5F,$CC,$E3,$FE,$9E,$9F,$8F,$3F,$DB,$27,$3E,$27
	.dc.b	$C8,$24,$FC,$CF,$BC,$AF,$C5,$8B,$A7,$9E,$7F,$BA,$FD,$77,$F8,$92
	.dc.b	$3F,$C0,$6F,$F9,$49,$B5,$AF,$06,$3F,$22,$7F,$C5,$2F,$BD,$AF,$FA
	.dc.b	$3D,$AB,$EB,$7F,$CA,$15,$DA,$B9,$CA,$3F,$C2,$73,$AD,$FA,$7D,$0F
	.dc.b	$9B,$E6,$A5,$FF,$7C,$C5,$3E,$A3,$FC,$D8,$E2,$3C,$D1,$83,$3E,$A4
	.dc.b	$8F,$53,$5F,$E4,$B2,$8F,$8A,$48,$F0,$7B,$9F,$23,$E3,$E4,$78,$DF
	.dc.b	$97,$F1,$B2,$3E,$33,$E4,$23,$F9,$0F,$BC,$F9,$D9,$FC,$99,$E1,$3F
	.dc.b	$0E,$3F,$15,$3A,$8F,$99,$F5,$9F,$63,$CD,$91,$F0,$E3,$ED,$9F,$08
	.dc.b	$8F,$D7,$7F,$D2,$1F,$5E,$FA,$BA,$49,$FD,$1C,$65,$B1,$FE,$3A,$8C
	.dc.b	$47,$94,$FF,$66,$38,$E1,$FF,$25,$BF,$AB,$8F,$F1,$8B,$FE,$4C,$8C
	.dc.b	$F9,$3C,$CE,$9F,$53,$2B,$C2,$F9,$88,$FE,$DB,$EB,$DF,$3D,$FE,$A9
	.dc.b	$6C,$66,$30,$7F,$A3,$37,$FE,$8B,$BC,$9F,$E7,$B1,$FE,$54,$E5,$FF
	.dc.b	$45,$65,$F9,$0D,$70,$67,$6F,$F2,$CE,$4F,$0F,$FC,$B7,$F5,$BF,$EB
	.dc.b	$FC,$BE,$10,$71,$87,$F9,$B2,$B8,$4F,$F4,$CF,$87,$FE,$AE,$47,$FA
	.dc.b	$7F,$A1,$E3,$FA,$E9,$1F,$E5,$8D,$11,$FE,$6A,$FA,$EF,$F3,$BE,$3F
	.dc.b	$E7,$8C,$9F,$33,$E4,$26,$1F,$E6,$08,$7D,$7F,$F6,$6F,$07,$FB,$3E
	.dc.b	$F2,$BF,$AF,$C7,$5F,$5B,$B7,$44,$9F,$F3,$4E,$7F,$E9,$F7,$DF,$8F
	.dc.b	$F3,$56,$F9,$3F,$10,$F8,$BE,$57,$F7,$4E,$49,$3F,$DB,$1F,$EE,$FB
	.dc.b	$FA,$FC,$69,$3D,$9F,$77,$BF,$F3,$BE,$92,$4F,$AD,$FA,$3F,$49,$FD
	.dc.b	$2B,$C0,$93,$FD,$45,$FE,$9B,$BF,$C7,$F8,$59,$3E,$A7,$E5,$77,$FD
	.dc.b	$AF,$3A,$4F,$C7,$7E,$EF,$9D,$FD,$5B,$8F,$FA,$BE,$B2,$BF,$AF,$47
	.dc.b	$D3,$F0,$7E,$A7,$D7,$C9,$F8,$F1,$F8,$6F,$A8,$FC,$3F,$CD,$49,$FA
	.dc.b	$91,$FE,$A4,$FA,$FF,$F1,$F7,$6F,$FA,$C9,$BC,$1F,$D2,$5B,$40,$FF
	.dc.b	$90,$C9,$EC,$73,$3A,$7C,$C4,$9F,$F5,$0C,$A3,$11,$F9,$8F,$0D,$F5
	.dc.b	$FE,$3F,$22,$4F,$07,$CD,$E3,$48,$C1,$0F,$D0,$E1,$F9,$E7,$F4,$7F
	.dc.b	$5E,$E8,$FF,$60,$B6,$3F,$D6,$6D,$77,$CB,$A2,$3A,$7F,$D9,$3F,$5E
	.dc.b	$F9,$6F,$F6,$5F,$A2,$FA,$AA,$4B,$FF,$5F,$C3,$E2,$76,$48,$CB,$D3
	.dc.b	$EB,$FF,$DA,$BF,$5D,$F9,$BE,$5F,$F7,$1B,$79,$7F,$BD,$36,$FE,$93
	.dc.b	$FE,$0E,$7F,$D9,$AD,$7F,$81,$B6,$79,$F8,$BD,$BC,$19,$5C,$1A,$FF
	.dc.b	$0A,$F3,$3F,$9F,$2B,$F8,$D7,$E4,$E1,$F3,$9E,$8E,$09,$5E,$26,$BE
	.dc.b	$2E,$FF,$ED,$8B,$EB,$FF,$B2,$3D,$63,$2F,$F8,$5C,$C5,$FF,$89,$30
	.dc.b	$FF,$11,$B6,$77,$3F,$13,$AD,$F5,$FB,$6C,$7F,$A3,$AD,$18,$8E,$D1
	.dc.b	$89,$FF,$18,$C7,$E7,$7E,$BE,$A6,$91,$FF,$1B,$FB,$4E,$0F,$F7,$EC
	.dc.b	$7E,$0D,$B1,$98,$FE,$6A,$64,$BC,$99,$CF,$50,$F7,$7F,$C8,$57,$D2
	.dc.b	$B4,$FF,$8E,$E3,$33,$D3,$93,$2F,$F8,$FF,$E3,$6B,$FB,$78,$AA,$29
	.dc.b	$8F,$72,$DF,$A9,$C3,$FB,$FC,$7C,$B8,$3C,$AC,$E1,$FD,$42,$31,$5B
	.dc.b	$3F,$CA,$14,$E5,$F1,$BF,$B3,$8D,$57,$59,$CF,$24,$9C,$CF,$C3,$F0
	.dc.b	$B0,$F7,$7D,$BF,$5C,$DE,$EF,$86,$DF,$29,$FA,$19,$7A,$BC,$0C,$E3
	.dc.b	$E9,$79,$FB,$7D,$3E,$56,$E1,$C5,$C3,$E4,$FA,$FF,$E4,$4E,$4C,$5F
	.dc.b	$93,$D9,$91,$E1,$F6,$EC,$FF,$76,$CD,$EB,$7A,$8F,$5F,$B5,$6F,$E0
	.dc.b	$FD,$73,$E5,$7E,$DB,$63,$E2,$FD,$7B,$D4,$CD,$7E,$03,$DE,$7B,$94
	.dc.b	$C7,$B3,$70,$F9,$BF,$5F,$FE,$8E,$6F,$89,$F8,$88,$FD,$3C,$60,$FA
	.dc.b	$7F,$5E,$E5,$FC,$6B,$63,$3C,$B8,$9F,$C8,$FD,$7B,$92,$F2,$E1,$B7
	.dc.b	$F7,$BE,$3F,$FC,$0A,$DF,$7B,$F3,$31,$8E,$7F,$F0,$EB,$FE,$B5,$9F
	.dc.b	$C2,$FD,$7F,$E6,$DB,$E1,$FE,$82,$31,$5F,$E2,$45,$1F,$A4,$C2,$D7
	.dc.b	$E3,$9F,$F3,$CB,$63,$5E,$7F,$EA,$BE,$BD,$F5,$17,$2C,$36,$FF,$2F
	.dc.b	$F5,$EF,$B0,$FD,$8B,$FE,$AE,$8C,$C6,$3E,$3B,$F6,$BF,$60,$FA,$6A
	.dc.b	$63,$CA,$B7,$FA,$2F,$AF,$FB,$6D,$EB,$FD,$EC,$3A,$EC,$7F,$00,$FA
	.dc.b	$FF,$ED,$75,$FA,$A6,$F2,$7F,$91,$9F,$F2,$CC,$67,$DE,$FE,$0F,$F5
	.dc.b	$CE,$1B,$E6,$61,$B5,$FB,$4E,$9F,$D0,$E3,$FE,$50,$7F,$65,$19,$8F
	.dc.b	$E2,$31,$87,$59,$E3,$E2,$72,$DA,$FD,$FF,$D6,$3B,$BF,$8C,$B6,$33
	.dc.b	$F3,$58,$9F,$E3,$BF,$59,$FE,$0B,$9F,$E0,$ED,$C7,$FC,$C6,$BE,$BF
	.dc.b	$7F,$0D,$FC,$93,$EB,$7E,$B5,$FF,$0E,$6D,$FF,$9A,$9F,$B0,$DE,$9F
	.dc.b	$E5,$7F,$5A,$E1,$FE,$6F,$3F,$EE,$D8,$C5,$FF,$BE,$9B,$1F,$D6,$C3
	.dc.b	$87,$FA,$F7,$D6,$3E,$8F,$9C,$7F,$8B,$63,$15,$C6,$11,$87,$A3,$5F
	.dc.b	$ED,$1F,$5A,$F8,$DE,$70,$8C,$57,$36,$C7,$95,$AF,$F7,$1C,$3F,$0A
	.dc.b	$FA,$5A,$E7,$D5,$EB,$18,$8F,$F2,$21,$DF,$E8,$FC,$34,$69,$2F,$D9
	.dc.b	$B3,$EB,$3C,$ED,$8C,$C6,$0F,$5E,$7F,$C8,$70,$F8,$3E,$86,$09,$5D
	.dc.b	$3E,$5F,$99,$1C,$1F,$2E,$BF,$8C,$57,$CA,$F3,$7A,$92,$75,$CB,$E0
	.dc.b	$23,$83,$EA,$7D,$97,$FA,$CD,$BF,$1F,$F4,$23,$17,$F8,$16,$BF,$E1
	.dc.b	$DF,$AD,$FE,$D5,$A7,$DD,$6C,$7E,$1D,$7C,$AF,$F5,$D8,$C7,$C1,$7F
	.dc.b	$AF,$DF,$9F,$F5,$95,$B7,$FE,$1A,$8C,$47,$F9,$DD,$A9,$FF,$5B,$BF
	.dc.b	$8D,$6B,$F9,$76,$BF,$E3,$36,$BF,$7B,$DB,$FF,$61,$BF,$95,$6B,$F9
	.dc.b	$F6,$BF,$E3,$B6,$BF,$86,$FB,$2F,$CA,$A6,$93,$4F,$F1,$9F,$D6,$3B
	.dc.b	$E9,$2B,$F4,$91,$97,$F4,$DF,$59,$FE,$38,$23,$FE,$3C,$7F,$C7,$7F
	.dc.b	$53,$FD,$0A,$3F,$58,$D7,$F6,$CD,$7F,$CA,$4D,$7F,$3B,$F4,$FF,$E2
	.dc.b	$17,$F6,$8D,$7F,$7C,$D7,$FC,$B4,$D7,$F4,$BF,$53,$F0,$09,$8F,$F9
	.dc.b	$41,$F8,$73,$FB,$99,$EB,$DB,$FE,$54,$8C,$47,$68,$FF,$2B,$FD,$67
	.dc.b	$CC,$8F,$F5,$A3,$53,$49,$E3,$FF,$8D,$30,$FD,$BC,$7D,$8F,$86,$FF
	.dc.b	$28,$B6,$3C,$FF,$03,$BB,$FF,$60,$7E,$12,$FA,$9C,$42,$30,$FF,$8E
	.dc.b	$1A,$B0,$FF,$2E,$E8,$FA,$BE,$2D,$9F,$E3,$CD,$7F,$94,$35,$FF,$42
	.dc.b	$35,$F9,$7E,$77,$8B,$7F,$E9,$11,$FE,$48,$6B,$E1,$CF,$6F,$F9,$32
	.dc.b	$BF,$DB,$6A,$77,$87,$F9,$C9,$B5,$C2,$79,$1F,$EE,$1A,$78,$BF,$0B
	.dc.b	$81,$F3,$F4,$F6,$C4,$AF,$F2,$DC,$E9,$B3,$97,$A8,$AE,$AE,$88,$69
	.dc.b	$CB,$6B,$C0,$EB,$1F,$9F,$1D,$5F,$F4,$6D,$F8,$67,$BB,$63,$71,$F1
	.dc.b	$67,$9B,$AB,$E5,$87,$E6,$A7,$35,$D2,$73,$5F,$16,$73,$5F,$A8,$9C
	.dc.b	$CF,$B6,$72,$85,$F8,$EB,$B4,$78,$CF,$63,$CD,$C1,$F0,$4C,$9C,$E3
	.dc.b	$93,$E3,$37,$7E,$77,$67,$8C,$CC,$F5,$7E,$01,$83,$91,$82,$47,$C8
	.dc.b	$7C,$46,$23,$94,$64,$8E,$09,$1D,$23,$2F,$A6,$FE,$2B,$E5,$3A,$C7
	.dc.b	$A5,$EC,$71,$9F,$8E,$E0,$F9,$73,$F4,$30,$FC,$3D,$71,$25,$70,$7C
	.dc.b	$35,$C0,$F8,$B3,$C2,$74,$7A,$23,$47,$91,$CB,$E9,$38,$E1,$F8,$0C
	.dc.b	$3E,$9C,$70,$BE,$F7,$F9,$A8,$FD,$ED,$7F,$23,$23,$FC,$11,$AF,$F0
	.dc.b	$46,$BC,$0F,$AC,$7C,$79,$FE,$1F,$DB,$F0,$67,$E1,$36,$77,$48,$C4
	.dc.b	$7A,$CF,$DF,$38,$AF,$3B,$70,$F7,$CF,$ED,$E3,$0F,$D3,$38,$9D,$AF
	.dc.b	$C2,$D3,$C3,$1C,$1F,$44,$FF,$46,$8C,$5F,$D1,$9E,$67,$94,$D9,$DC
	.dc.b	$F1,$1D,$4D,$3E,$D7,$17,$EA,$6D,$7A,$A7,$51,$C2,$FC,$ED,$3C,$E7
	.dc.b	$03,$C1,$1C,$2B,$ED,$CF,$06,$CF,$03,$91,$E4,$BF,$6C,$FF,$42,$6C
	.dc.b	$7F,$35,$6B,$FC,$7D,$FE,$9F,$D6,$F0,$69,$27,$F8,$55,$B1,$FF,$29
	.dc.b	$D3,$E0,$35,$5F,$CB,$A3,$93,$C2,$77,$7D,$AB,$FA,$EF,$D5,$FF,$AE
	.dc.b	$EB,$47,$F1,$77,$0F,$F3,$86,$E1,$E1,$8F,$E3,$98,$3E,$5F,$D2,$74
	.dc.b	$F3,$24,$F4,$3F,$C7,$F1,$EE,$F6,$7F,$B6,$FC,$57,$FB,$BE,$74,$D7
	.dc.b	$D5,$1B,$7F,$D0,$D7,$FE,$CB,$6C,$7F,$A2,$5F,$F2,$5B,$63,$51,$C8
	.dc.b	$F1,$38,$C3,$D5,$1F,$E5,$2F,$6B,$E8,$FA,$5C,$49,$1E,$07,$A7,$46
	.dc.b	$A3,$9C,$7D,$88,$FF,$03,$FD,$57,$FA,$1A,$7F,$AF,$B6,$FB,$B7,$83
	.dc.b	$FD,$16,$FF,$98,$67,$67,$22,$1F,$25,$D4,$F2,$F8,$2F,$F0,$AF,$D3
	.dc.b	$7F,$82,$E3,$D6,$DB,$EB,$3E,$43,$E3,$57,$F8,$FE,$FF,$AA,$35,$34
	.dc.b	$7C,$1F,$18,$F1,$4F,$BA,$3D,$FF,$55,$F0,$3C,$F9,$5E,$78,$CB,$FE
	.dc.b	$BD,$B4,$78,$89,$19,$F8,$BF,$29,$1C,$10,$F9,$6F,$CE,$3E,$7F,$C3
	.dc.b	$7F,$8D,$BE,$AB,$FC,$CD,$3F,$E0,$E6,$EB,$F7,$9B,$A7,$D0,$F8,$1F
	.dc.b	$AE,$7C,$42,$57,$9D,$F8,$FE,$EE,$A3,$E9,$47,$DE,$8F,$F2,$2F,$D5
	.dc.b	$3E,$9A,$FD,$44,$BF,$D0,$4E,$5B,$3F,$97,$8C,$47,$FB,$A5,$AF,$E2
	.dc.b	$5A,$FF,$B4,$17,$D4,$71,$24,$7F,$BA,$63,$2D,$8F,$29,$1F,$1F,$97
	.dc.b	$FE,$5C,$FA,$97,$D9,$D2,$4F,$A3,$DC,$F6,$9E,$72,$3D,$DF,$57,$D9
	.dc.b	$C8,$F1,$35,$FF,$43,$C7,$F9,$24,$8F,$F9,$1A,$3F,$33,$12,$3F,$83
	.dc.b	$74,$7F,$BA,$DB,$3E,$7E,$D9,$1F,$C2,$F0,$FD,$D7,$07,$FB,$D1,$76
	.dc.b	$7F,$95,$FE,$37,$FC,$B1,$E5,$CB,$FF,$2A,$C7,$EA,$0F,$D4,$C4,$8F
	.dc.b	$0C,$7C,$5F,$0B,$E3,$7C,$64,$9F,$04,$7F,$98,$FE,$0E,$47,$F9,$81
	.dc.b	$FF,$33,$7C,$07,$ED,$19,$1E,$C7,$F6,$F1,$F4,$7C,$29,$3E,$B7,$EA
	.dc.b	$78,$5F,$E6,$FF,$85,$FF,$38,$7C,$4C,$AF,$F3,$6C,$7F,$00,$3F,$80
	.dc.b	$C4,$8F,$93,$1F,$77,$EB,$7E,$F7,$D3,$C9,$F9,$11,$FE,$7B,$F3,$BF
	.dc.b	$CF,$9E,$74,$9F,$F3,$CC,$7F,$0C,$3F,$86,$C4,$8F,$AB,$1F,$89,$FA
	.dc.b	$7F,$F5,$2F,$81,$FC,$6E,$7F,$CB,$CD,$5C,$89,$FE,$9F,$EF,$C9,$FE
	.dc.b	$89,$E7,$23,$FA,$9F,$D6,$FF,$AC,$36,$C5,$7F,$49,$6B,$FE,$8A,$6F
	.dc.b	$8B,$FC,$E9,$FF,$7E,$BF,$CF,$32,$FE,$4F,$D3,$FE,$71,$A4,$61,$F3
	.dc.b	$FC,$E7,$AB,$E7,$7F,$D7,$FF,$57,$FE,$A8,$27,$5E,$3A,$25,$FF,$5B
	.dc.b	$94,$62,$3F,$D3,$B0,$E8,$E6,$48,$E0,$FF,$B3,$3E,$B1,$AF,$83,$8E
	.dc.b	$57,$57,$C7,$C9,$23,$04,$3B,$BF,$ED,$6F,$AC,$7D,$0D,$24,$FF,$71
	.dc.b	$8F,$EE,$CD,$F5,$78,$11,$FE,$ED,$F5,$8E,$AF,$F7,$0F,$A6,$E1,$A4
	.dc.b	$AF,$EF,$90,$FA,$5D,$92,$32,$FF,$BA,$AF,$F0,$51,$FE,$03,$FA,$8F
	.dc.b	$F0,$C3,$7D,$9F,$F7,$7C,$62,$3F,$D8,$EB,$EB,$32,$48,$E4,$F0,$FA
	.dc.b	$69,$1D,$5A,$FF,$B4,$3E,$B3,$F2,$95,$F1,$A3,$C5,$BF,$80,$F0,$ED
	.dc.b	$23,$CC,$D7,$C7,$F4,$7F,$EF,$F5,$F5,$9F,$D9,$9F,$F5,$F5,$A7,$51
	.dc.b	$FE,$03,$8C,$76,$FF,$81,$63,$73,$A7,$F2,$D3,$A7,$4B,$EB,$1F,$59
	.dc.b	$C6,$53,$98,$C3,$63,$8C,$F0,$7F,$C4,$DF,$59,$FE,$27,$4D,$23,$D7
	.dc.b	$BB,$FC,$7D,$8F,$F8,$9A,$3F,$05,$E0,$7F,$8E,$D4,$EA,$3F,$C4,$1C
	.dc.b	$B6,$7F,$60,$D7,$F9,$AF,$C0,$57,$98,$BF,$F1,$9F,$9C,$9D,$73,$6F
	.dc.b	$EB,$2E,$3E,$17,$0F,$8F,$E0,$6E,$3C,$E3,$FC,$65,$1D,$CF,$16,$0F
	.dc.b	$93,$C4,$FF,$95,$63,$0F,$F9,$5F,$27,$37,$A5,$FD,$6C,$66,$31,$CB
	.dc.b	$23,$97,$87,$72,$F1,$1F,$F0,$07,$0F,$E0,$F3,$FE,$63,$FA,$BF,$92
	.dc.b	$DD,$FE,$34,$62,$B8,$87,$E3,$E5,$FE,$58,$6D,$7F,$1D,$B7,$81,$CB
	.dc.b	$0F,$7B,$D3,$E6,$3A,$FC,$C5,$BF,$13,$87,$FC,$ED,$F5,$8C,$F0,$78
	.dc.b	$9C,$47,$A7,$E8,$3C,$E7,$A3,$E0,$EC,$7B,$7E,$23,$D6,$7B,$3E,$5E
	.dc.b	$C7,$FA,$0F,$EB,$1F,$15,$F8,$56,$C7,$FA,$1F,$EB,$1F,$27,$9A,$FC
	.dc.b	$57,$C5,$7F,$3A,$AF,$F4,$63,$75,$F8,$EB,$E2,$77,$E9,$FE,$47,$EB
	.dc.b	$3F,$69,$7F,$E9,$36,$DF,$CE,$7F,$5D,$19,$F1,$7F,$31,$F5,$8E,$4A
	.dc.b	$63,$92,$D7,$FA,$77,$EB,$1C,$7F,$9F,$5C,$96,$30,$FE,$8F,$D6,$FF
	.dc.b	$5F,$57,$FA,$95,$FF,$5D,$37,$E4,$35,$E7,$7E,$A7,$EB,$1E,$25,$F1
	.dc.b	$30,$DB,$FF,$57,$64,$9F,$D8,$3F,$E7,$A8,$CC,$63,$D1,$FD,$A7,$D6
	.dc.b	$3E,$16,$E9,$86,$DF,$FA,$E7,$EB,$1F,$0B,$FB,$B5,$F0,$D6,$30,$FE
	.dc.b	$FF,$EB,$1F,$11,$4C,$78,$D7,$0F,$F6,$27,$D6,$7D,$9F,$27,$FD,$66
	.dc.b	$D8,$FE,$F0,$FD,$B7,$7D,$BF,$C2,$FE,$B3,$FB,$4C,$7F,$B3,$5B,$DD
	.dc.b	$FA,$87,$F4,$71,$9F,$7B,$F8,$97,$D6,$38,$6F,$9B,$86,$D7,$FB,$5F
	.dc.b	$AF,$F3,$F8,$FE,$C9,$FD,$A6,$9F,$EC,$87,$2D,$F4,$FF,$8F,$47,$9E
	.dc.b	$78,$7C,$FE,$3C,$4C,$BF,$C8,$FE,$B1,$E0,$D3,$1F,$47,$73,$FF,$74
	.dc.b	$7C,$77,$F8,$2D,$BF,$0B,$FC,$AE,$1F,$1B,$63,$F9,$77,$D6,$7D,$6A
	.dc.b	$FF,$78,$37,$4F,$E1,$6F,$D8,$8C,$F5,$FF,$35,$FA,$BE,$D5,$FD,$55
	.dc.b	$AF,$FB,$6D,$AF,$F8,$AD,$AF,$F0,$7F,$AB,$7C,$B7,$F4,$78,$C3,$54
	.dc.b	$6A,$FF,$AD,$36,$3F,$B2,$7D,$97,$F4,$19,$FE,$BC,$D8,$CF,$BD,$CE
	.dc.b	$31,$1F,$EE,$FA,$7C,$B7,$76,$DF,$68,$C1,$E3,$FA,$C7,$D8,$52,$4F
	.dc.b	$F6,$D8,$CB,$7D,$4E,$04,$70,$6F,$E7,$FA,$11,$C1,$CF,$D1,$F0,$11
	.dc.b	$C1,$DF,$97,$C8,$47,$07,$C2,$F9,$BF,$80,$47,$07,$F8,$0F,$EB,$3F
	.dc.b	$AE,$7C,$A5,$F8,$4F,$F3,$3E,$FF,$E6,$19,$FE,$D9,$DD,$9A,$FF,$05
	.dc.b	$7D,$97,$F3,$2A,$BA,$FF,$83,$54,$65,$FF,$0C,$35,$FF,$36,$0F,$AD
	.dc.b	$7C,$B7,$16,$C7,$F8,$5D,$AF,$F8,$85,$AF,$2F,$80,$7C,$0F,$85,$B1
	.dc.b	$FE,$1F,$6B,$FE,$29,$6B,$E2,$FA,$DF,$C7,$26,$93,$6F,$67,$D9,$3E
	.dc.b	$76,$92,$BF,$C5,$91,$97,$FC,$5D,$F5,$3F,$CC,$1B,$ED,$7C,$58,$C4
	.dc.b	$7F,$8E,$DA,$FC,$4F,$AE,$F2,$BF,$E9,$16,$BF,$DB,$5A,$9A,$7F,$A5
	.dc.b	$9F,$F1,$F3,$5F,$F2,$43,$5E,$DE,$87,$D7,$FA,$87,$D3,$52,$4F,$F9
	.dc.b	$1E,$32,$FF,$92,$7E,$B5,$B7,$DF,$6C,$66,$30,$75,$F8,$79,$1C,$DA
	.dc.b	$FF,$96,$E9,$D3,$E3,$24,$78,$5A,$FF,$87,$23,$6F,$F8,$3F,$CF,$70
	.dc.b	$7F,$AD,$E7,$33,$FE,$22,$D7,$F6,$EE,$E7,$5F,$65,$FE,$BD,$BE,$33
	.dc.b	$FE,$19,$B4,$E6,$B7,$3C,$4D,$78,$39,$78,$4D,$55,$C9,$E0,$62,$35
	.dc.b	$1F,$C4,$23,$2F,$FB,$11,$BE,$8E,$8D,$FA,$52,$32,$43,$E2,$3D,$3E
	.dc.b	$77,$AF,$C6,$C9,$EB,$19,$7B,$C6,$5F,$99,$19,$7E,$74,$67,$3F,$8B
	.dc.b	$1E,$C9,$D4,$63,$0F,$F2,$4C,$F4,$3B,$90,$E6,$64,$EE,$66,$39,$57
	.dc.b	$7C,$FE,$D5,$F6,$7E,$E4,$66,$BE,$B4,$E5,$FB,$11,$97,$CC,$75,$9E
	.dc.b	$2F,$86,$BD,$15,$C6,$3E,$BD,$71,$B4,$E5,$F6,$CF,$A8,$F4,$92,$BD
	.dc.b	$2F,$5B,$F4,$BE,$13,$CA,$64,$F4,$99,$3C,$26,$4F,$49,$D5,$F4,$BE
	.dc.b	$18,$F4,$BC,$67,$CB,$1E,$28,$E0,$76,$38,$4F,$28,$E3,$1A,$3C,$26
	.dc.b	$8E,$0E,$8E,$84,$8F,$6C,$71,$9F,$0C,$75,$33,$3C,$5C,$4F,$C2,$38
	.dc.b	$57,$CF,$BF,$9C,$D8,$F6,$35,$D9,$FC,$4D,$FE,$30,$47,$04,$3F,$86
	.dc.b	$39,$69,$D0,$FD,$9F,$D5,$BE,$B7,$C3,$1F,$4E,$30,$EC,$F8,$CE,$1F
	.dc.b	$19,$88,$E2,$D3,$86,$1F,$05,$FF,$52,$4F,$18,$C4,$7D,$E6,$C6,$70
	.dc.b	$F9,$24,$9F,$A6,$FF,$AB,$54,$6D,$C9,$0E,$91,$B8,$D4,$60,$D6,$3F
	.dc.b	$9E,$7E,$1D,$FA,$AF,$C6,$DB,$F1,$4F,$82,$30,$F7,$BE,$B7,$CE,$3E
	.dc.b	$2D,$FF,$08,$9C,$E1,$C2,$79,$1F,$1B,$0E,$B3,$A8,$C1,$F2,$30,$E1
	.dc.b	$87,$5C,$3A,$46,$30,$FE,$61,$7C,$6F,$C3,$87,$C1,$AD,$47,$B8,$EA
	.dc.b	$7C,$33,$C6,$70,$3D,$8F,$07,$FD,$D7,$3F,$2B,$CA,$7C,$37,$F3,$F6
	.dc.b	$CF,$F3,$E6,$BF,$EF,$25,$F3,$9B,$99,$3B,$79,$47,$3F,$8B,$E9,$F3
	.dc.b	$72,$BA,$46,$5E,$D1,$97,$B4,$78,$FD,$FF,$27,$C0,$49,$F2,$46,$5F
	.dc.b	$F1,$89,$FE,$F7,$C5,$1F,$E3,$96,$C6,$83,$87,$8F,$82,$F6,$3A,$47
	.dc.b	$BE,$BF,$B1,$4E,$63,$E2,$78,$3D,$48,$F5,$BF,$F1,$43,$E4,$6F,$0F
	.dc.b	$F5,$06,$F0,$7E,$85,$B3,$E9,$FA,$5F,$EE,$1F,$58,$F2,$AC,$C3,$FB
	.dc.b	$8C,$65,$B1,$FE,$57,$8C,$47,$C4,$6B,$FE,$56,$6B,$FB,$17,$FD,$CC
	.dc.b	$F4,$F8,$9F,$09,$E2,$F8,$99,$19,$20,$CB,$FB,$5D,$27,$73,$E1,$7C
	.dc.b	$84,$7E,$94,$3D,$1F,$41,$23,$2D,$7B,$67,$E7,$8F,$8A,$7C,$6F,$47
	.dc.b	$E3,$FC,$84,$AF,$8A,$FC,$AF,$8B,$91,$E0,$7E,$31,$1F,$1F,$8F,$F2
	.dc.b	$48,$E0,$F3,$FC,$0F,$D4,$89,$1F,$4A,$3E,$B9,$F4,$48,$F0,$3F,$21
	.dc.b	$F5,$4F,$AB,$FF,$46,$36,$33,$18,$23,$FE,$2E,$8D,$3F,$08,$07,$3B
	.dc.b	$7F,$87,$1F,$CE,$C7,$1F,$1B,$FC,$75,$1F,$E4,$A9,$D4,$63,$DB,$95
	.dc.b	$E6,$7C,$A4,$7D,$3E,$37,$A0,$8E,$0F,$6F,$D4,$F6,$89,$1E,$98,$F0
	.dc.b	$91,$C1,$FE,$4A,$CD,$F3,$7F,$C8,$2A,$4F,$F5,$C7,$F9,$1D,$7F,$94
	.dc.b	$FE,$AB,$FC,$9D,$3F,$0C,$38,$CB,$FC,$7A,$DE,$2F,$E9,$CA,$BF,$CB
	.dc.b	$9F,$54,$EE,$A4,$C3,$FC,$BB,$1F,$C1,$70,$FE,$81,$F2,$1F,$CD,$3C
	.dc.b	$A9,$5F,$DC,$23,$C8,$79,$62,$47,$F9,$9B,$14,$F8,$3E,$0B,$C3,$F1
	.dc.b	$F2,$BB,$C7,$F4,$0F,$32,$47,$F7,$C7,$FA,$1F,$C2,$FC,$16,$47,$57
	.dc.b	$E2,$47,$B3,$E1,$E4,$FA,$9F,$7F,$85,$FD,$33,$E1,$FF,$A6,$FC,$64
	.dc.b	$AF,$F0,$4C,$7C,$D3,$E7,$44,$8F,$54,$7C,$8F,$0F,$E4,$FD,$2C,$9F
	.dc.b	$8F,$1F,$D6,$3D,$1F,$EB,$3E,$84,$9F,$F0,$AC,$7D,$93,$ED,$44,$8F
	.dc.b	$8D,$1F,$53,$EA,$BF,$D2,$1D,$5F,$EE,$CE,$4F,$F3,$BB,$7A,$3F,$C1
	.dc.b	$1C,$CA,$FF,$4B,$7D,$53,$E0,$A9,$32,$FF,$4B,$C7,$FB,$C3,$1F,$F4
	.dc.b	$0E,$FB,$9F,$F4,$DB,$5F,$C6,$36,$37,$F0,$1E,$4F,$2A,$57,$E6,$5F
	.dc.b	$3F,$CE,$FF,$A9,$FE,$A9,$D9,$49,$5F,$EA,$88,$FF,$52,$3F,$99,$6E
	.dc.b	$1F,$E9,$46,$ED,$FE,$25,$6C,$FF,$00,$5C,$FD,$C9,$1D,$5F,$F5,$C7
	.dc.b	$D5,$34,$F2,$71,$4A,$F1,$3E,$7E,$39,$18,$21,$E5,$7F,$D8,$5F,$54
	.dc.b	$FA,$2A,$49,$FF,$61,$C3,$D5,$D9,$23,$2F,$FB,$2F,$EA,$9D,$3F,$EC
	.dc.b	$DF,$41,$C5,$49,$7F,$EC,$E8,$FE,$0A,$DF,$1B,$81,$1F,$F6,$9D,$FE
	.dc.b	$0D,$FF,$6A,$FD,$3F,$F8,$63,$8F,$FD,$7E,$DE,$3F,$F2,$83,$67,$F9
	.dc.b	$5B,$FE,$1C,$78,$FB,$1F,$EE,$18,$C4,$7F,$0F,$6D,$74,$FA,$5D,$12
	.dc.b	$7C,$2F,$FB,$A7,$EA,$BF,$31,$1D,$B8,$3A,$11,$F3,$EF,$DC,$8E,$0F
	.dc.b	$5F,$B5,$FD,$BD,$7D,$57,$F6,$F7,$D3,$19,$7F,$CD,$83,$24,$66,$3E
	.dc.b	$0B,$5E,$39,$7F,$89,$65,$F5,$2F,$AA,$C3,$6B,$FC,$5F,$3A,$9C,$9F
	.dc.b	$E2,$D8,$FE,$AB,$F5,$55,$34,$8F,$F8,$CF,$B5,$C1,$FD,$73,$1F,$F7
	.dc.b	$AC,$FE,$12,$30,$7B,$27,$71,$FD,$04,$D1,$79,$33,$5F,$EF,$EF,$77
	.dc.b	$FC,$7F,$7C,$AB,$FD,$21,$5A,$8F,$F4,$94,$EA,$7A,$72,$65,$FF,$4B
	.dc.b	$7C,$55,$7F,$B8,$C5,$51,$4C,$7B,$96,$FC,$13,$FD,$E2,$33,$18,$B7
	.dc.b	$C8,$E2,$BF,$AE,$46,$27,$8E,$4E,$5F,$13,$FB,$38,$CC,$63,$92,$47
	.dc.b	$5F,$32,$E1,$D1,$F5,$7D,$57,$F9,$36,$7C,$AF,$CA,$C3,$E6,$D2,$34
	.dc.b	$FF,$6F,$75,$3F,$EB,$96,$C7,$F8,$53,$17,$77,$6D,$FE,$44,$62,$FB
	.dc.b	$FB,$9E,$2F,$93,$F1,$B6,$3C,$71,$F3,$FE,$AB,$FC,$3F,$18,$3E,$5B
	.dc.b	$8C,$7C,$9F,$3B,$EC,$3D,$BE,$F5,$AF,$6F,$C1,$FC,$03,$E0,$FC,$65
	.dc.b	$8F,$AB,$F5,$4F,$95,$FB,$CD,$8F,$B3,$F5,$4F,$8F,$CD,$7E,$0F,$F0
	.dc.b	$9F,$F5,$1D,$7D,$E6,$E9,$FE,$15,$7F,$D5,$B1,$9F,$17,$F1,$5F,$56
	.dc.b	$FB,$3A,$63,$EB,$6D,$7E,$23,$EB,$7F,$D2,$8D,$F8,$EF,$C9,$43,$E2
	.dc.b	$EC,$7E,$57,$F0,$9F,$F3,$65,$7E,$41,$BA,$7F,$89,$DF,$63,$7E,$37
	.dc.b	$F3,$9E,$BF,$A3,$BB,$F0,$2D,$BF,$F7,$B3,$F9,$17,$F3,$6E,$6B,$FC
	.dc.b	$A7,$F5,$9F,$E8,$19,$FC,$FB,$63,$FC,$6C,$BE,$D3,$F5,$4E,$6F,$F1
	.dc.b	$F4,$F8,$5F,$D7,$34,$FD,$48,$F8,$5C,$C7,$EC,$FE,$B5,$E2,$53,$1E
	.dc.b	$DD,$BF,$D8,$FD,$6B,$E1,$FF,$74,$D8,$CF,$C3,$E2,$7F,$7D,$F8,$43
	.dc.b	$EA,$A9,$8F,$93,$B7,$FB,$CF,$AD,$D3,$67,$F8,$37,$D6,$FF,$1C,$8C
	.dc.b	$1F,$C0,$97,$C0,$E8,$FE,$19,$AF,$F8,$66,$75,$1F,$C3,$FF,$07,$76
	.dc.b	$53,$1D,$B6,$BF,$86,$F9,$1F,$B2,$E8,$34,$FF,$36,$C6,$BC,$DF,$E3
	.dc.b	$5F,$82,$F9,$6B,$FC,$D4,$DE,$8F,$E0,$2B,$E8,$77,$7F,$C8,$7D,$87
	.dc.b	$95,$4C,$65,$67,$F8,$FF,$2F,$F8,$45,$BF,$0B,$FC,$A6,$3F,$81,$67
	.dc.b	$FE,$79,$74,$FF,$9E,$FE,$B7,$EA,$36,$3F,$98,$43,$A2,$C7,$F3,$3F
	.dc.b	$AD,$7C,$8F,$F4,$8B,$FF,$52,$C6,$31,$FF,$80,$0D,$8F,$EC,$A1,$C3
	.dc.b	$FD,$A1,$F2,$B7,$D5,$E8,$ED,$B1,$A8,$FF,$59,$8F,$AC,$7C,$BF,$06
	.dc.b	$DF,$82,$75,$39,$3A,$F7,$FE,$06,$3C,$AD,$7F,$BD,$B5,$EE,$D7,$FD
	.dc.b	$87,$F5,$7F,$BC,$7F,$C0,$2D,$7C,$AD,$77,$E7,$F8,$3E,$C6,$61,$E5
	.dc.b	$B4,$62,$BE,$3D,$BD,$8F,$86,$DB,$F6,$36,$77,$3F,$23,$ED,$FE,$34
	.dc.b	$7F,$B6,$1A,$9A,$4E,$DE,$04,$BF,$91,$1F,$A8,$7E,$57,$D6,$FF,$68
	.dc.b	$F9,$3B,$F0,$1F,$E7,$9C,$3E,$9C,$BE,$85,$D5,$AB,$FA,$5F,$59,$BB
	.dc.b	$FB,$2D,$7F,$DD,$AD,$4D,$1F,$29,$FE,$1F,$6C,$74,$8C,$1F,$E2,$3E
	.dc.b	$CF,$E1,$51,$F8,$16,$BF,$92,$6B,$FE,$38,$6B,$F8,$CF,$AC,$FC,$8A
	.dc.b	$69,$34,$FF,$18,$7D,$67,$F9,$86,$7F,$3A,$D7,$FC,$86,$D7,$F2,$ED
	.dc.b	$43,$ED,$FF,$C7,$04,$7F,$C7,$8F,$F8,$EF,$EA,$DF,$11,$FE,$3F,$6C
	.dc.b	$FE,$75,$AF,$EB,$9A,$FF,$95,$3F,$0D,$F9,$84,$D7,$FC,$90,$FC,$FF
	.dc.b	$A9,$FF,$11,$CF,$E9,$DA,$FE,$E1,$AF,$F9,$79,$AF,$EC,$97,$D2,$7F
	.dc.b	$94,$DB,$1C,$E3,$07,$F9,$5F,$EB,$7E,$7C,$7C,$26,$A6,$98,$F9,$DF
	.dc.b	$F4,$1C,$7F,$9B,$9A,$FF,$02,$6B,$F3,$1A,$BE,$E3,$F8,$D3,$FC,$0D
	.dc.b	$AF,$F1,$46,$BF,$E7,$A6,$BF,$C3,$9D,$F9,$5F,$E3,$66,$EF,$9F,$17
	.dc.b	$F8,$9C,$FF,$8E,$BE,$B5,$F5,$B8,$AF,$E2,$FB,$BE,$33,$19,$7E,$0A
	.dc.b	$30,$FF,$96,$1A,$B0,$E9,$EF,$70,$84,$EA,$7F,$C9,$2D,$BE,$F7,$BC
	.dc.b	$7F,$DC,$B7,$FE,$64,$7F,$C8,$EF,$F9,$9B,$77,$DC,$75,$6C,$6F,$C8
	.dc.b	$E4,$66,$64,$64,$FF,$37,$F6,$7F,$1A,$7F,$E0,$0F,$C8,$F9,$F8,$7F
	.dc.b	$CD,$51,$AF,$13,$D5,$5B,$8F,$76,$9F,$6E,$F8,$1A,$A9,$3D,$08,$70
	.dc.b	$3F,$57,$5C,$F6,$FF,$35,$B6,$FA,$4F,$09,$CD,$74,$8E,$43,$0F,$3B
	.dc.b	$86,$CF,$08,$F5,$1C,$EB,$AB,$CA,$FC,$0F,$92,$FD,$6E,$CC,$1C,$8E
	.dc.b	$D8,$69,$E5,$1E,$5B,$FF,$42,$E1,$E2,$C3,$BC,$78,$EB,$67,$53,$D5
	.dc.b	$7C,$67,$FB,$3E,$1E,$49,$C9,$B7,$06,$6F,$B4,$7E,$AE,$B9,$3E,$83
	.dc.b	$CE,$71,$C3,$2F,$43,$27,$90,$98,$7A,$A7,$DC,$FA,$49,$87,$B0,$93
	.dc.b	$DE,$32,$F8,$CF,$01,$F2,$5E,$8E,$0E,$F1,$C9,$F2,$47,$27,$FA,$B5
	.dc.b	$78,$A7,$58,$7C,$28,$C4,$F7,$39,$9E,$39,$E9,$87,$C6,$8E,$67,$8E
	.dc.b	$33,$87,$C4,$BF,$98,$F4,$9D,$9F,$14,$FF,$0D,$98,$2C,$7E,$95,$B7
	.dc.b	$FE,$FF,$8C,$47,$DC,$8C,$11,$FF,$10,$47,$AD,$F2,$1F,$A6,$FA,$EF
	.dc.b	$24,$F9,$5B,$1A,$3F,$BD,$38,$7E,$09,$83,$53,$D5,$FD,$04,$6F,$1F
	.dc.b	$9D,$28,$C6,$1E,$7A,$FE,$0F,$18,$7E,$61,$E5,$3A,$DF,$95,$A7,$82
	.dc.b	$38,$3F,$2A,$FF,$7F,$18,$BF,$CD,$B6,$BF,$87,$B6,$BC,$27,$43,$DB
	.dc.b	$3E,$77,$15,$EA,$9E,$11,$EE,$33,$7E,$B7,$07,$2B,$F4,$34,$F4,$1C
	.dc.b	$0F,$79,$F2,$CD,$1E,$29,$EF,$5E,$C8,$E0,$7B,$EB,$F7,$73,$FC,$B7
	.dc.b	$DF,$93,$FE,$0C,$AF,$E0,$F1,$FC,$BB,$E6,$F3,$3F,$C6,$E2,$CF,$F2
	.dc.b	$69,$FE,$67,$F5,$3F,$D3,$7E,$3F,$FC,$A1,$7C,$F1,$FF,$37,$C6,$23
	.dc.b	$F8,$F4,$F0,$74,$7F,$87,$23,$56,$AE,$3E,$46,$EB,$F9,$37,$A1,$FD
	.dc.b	$6F,$F0,$C7,$D8,$53,$FA,$2C,$49,$FC,$DB,$FA,$12,$3F,$9E,$EA,$D3
	.dc.b	$D0,$FE,$A6,$FF,$67,$FB,$0F,$F4,$14,$7F,$8C,$1A,$9A,$47,$81,$0F
	.dc.b	$69,$BF,$23,$A7,$85,$FD,$CF,$EB,$3F,$BF,$CE,$9A,$FF,$4C,$AF,$F1
	.dc.b	$AE,$BF,$0E,$3C,$55,$E4,$9F,$F1,$BC,$7F,$A7,$1A,$E8,$FF,$4D,$2F
	.dc.b	$43,$D8,$48,$F2,$C3,$E2,$7E,$01,$98,$F1,$4E,$1D,$C7,$84,$8F,$67
	.dc.b	$C9,$F1,$92,$3C,$2D,$7F,$D7,$74,$8F,$9C,$48,$F1,$FC,$A7,$F8,$5F
	.dc.b	$EE,$5F,$1D,$49,$3E,$8F,$AB,$FA,$C7,$C3,$23,$F0,$1F,$B3,$F6,$1F
	.dc.b	$EE,$58,$FF,$5E,$35,$FF,$66,$B5,$FE,$56,$D7,$FD,$8E,$BE,$C7,$FC
	.dc.b	$54,$F1,$DB,$EA,$4A,$31,$CD,$C2,$B6,$78,$89,$3E,$27,$CB,$F6,$32
	.dc.b	$3C,$AD,$7C,$AF,$80,$8F,$E3,$2B,$FC,$CE,$FF,$84,$BE,$B9,$F0,$DF
	.dc.b	$E3,$F6,$C7,$E4,$9A,$FF,$86,$27,$FD,$C2,$F3,$F7,$3E,$39,$1F,$8E
	.dc.b	$FC,$AF,$88,$91,$F2,$9A,$FC,$AF,$77,$E2,$C4,$8F,$34,$7C,$82,$38
	.dc.b	$3D,$5F,$01,$FE,$55,$FC,$0F,$F9,$54,$7F,$8B,$1F,$F0,$81,$3C,$3F
	.dc.b	$F2,$D7,$C1,$FF,$84,$CD,$F0,$7D,$26,$F0,$7F,$BF,$9F,$F0,$C7,$0B
	.dc.b	$E8,$3F,$CC,$4F,$EB,$27,$FD,$57,$8A,$30,$FF,$AB,$9E,$0F,$EE,$BF
	.dc.b	$02,$FD,$53,$7E,$DB,$FC,$D9,$06,$F5,$F9,$BE,$6B,$CF,$E5,$CB,$FE
	.dc.b	$07,$9A,$7C,$7F,$59,$E4,$F9,$89,$5F,$C2,$33,$4F,$F9,$E7,$07,$C2
	.dc.b	$DF,$46,$5F,$F0,$E9,$CB,$DB,$E8,$FE,$3F,$9B,$23,$07,$5D,$BE,$1E
	.dc.b	$D2,$30,$7F,$18,$FC,$03,$E1,$A7,$E1,$BF,$E6,$06,$F1,$FE,$89,$DF
	.dc.b	$03,$EA,$BE,$BF,$D5,$4B,$F8,$31,$F4,$BC,$6F,$A7,$E7,$49,$F8,$11
	.dc.b	$FE,$97,$FA,$17,$C9,$DC,$25,$7F,$2C,$9C,$BF,$32,$7F,$0B,$E1,$48
	.dc.b	$C1,$F2,$FD,$AF,$C0,$F8,$D2,$30,$7F,$36,$FB,$F7,$C8,$FF,$A9,$D7
	.dc.b	$95,$4C,$47,$E4,$BD,$5F,$C9,$FD,$34,$BF,$BB,$1F,$8E,$F3,$7F,$1F
	.dc.b	$72,$7E,$E4,$7F,$AC,$FE,$CB,$FD,$5B,$1F,$E8,$A6,$F2,$FF,$0C,$6F
	.dc.b	$85,$F9,$65,$E4,$71,$25,$7F,$4E,$8C,$B7,$C8,$EC,$47,$07,$F5,$4F
	.dc.b	$B2,$74,$78,$FB,$E4,$F3,$7C,$DE,$04,$8C,$10,$F4,$3F,$EC,$DF,$B2
	.dc.b	$78,$14,$93,$FD,$8A,$1E,$06,$C9,$19,$7F,$DA,$DF,$62,$FB,$2F,$F6
	.dc.b	$C7,$88,$F9,$AA,$4B,$FE,$DB,$0F,$AD,$D9,$23,$2F,$4F,$B3,$FF,$71
	.dc.b	$7D,$8F,$F0,$69,$07,$0D,$F3,$B7,$E6,$48,$FE,$F7,$3F,$B9,$6C,$74
	.dc.b	$F2,$A4,$69,$AF,$FB,$1B,$EC,$FF,$CE,$FD,$3F,$D2,$2F,$E3,$CF,$4E
	.dc.b	$1F,$29,$E6,$E0,$91,$E0,$6B,$FE,$D0,$9F,$D7,$57,$AB,$EC,$FF,$83
	.dc.b	$CF,$B2,$72,$FF,$A1,$8F,$77,$92,$AC,$46,$1B,$3C,$23,$FD,$1A,$BC
	.dc.b	$77,$D9,$E9,$B3,$D6,$75,$39,$3F,$C4,$FF,$67,$27,$FC,$56,$D7,$C9
	.dc.b	$EF,$7F,$87,$70,$FE,$D7,$1F,$7F,$BF,$FC,$52,$A7,$51,$FE,$A0,$E5
	.dc.b	$B3,$FC,$E9,$AB,$C3,$7C,$36,$23,$A1,$FE,$F8,$8C,$D7,$6E,$6D,$7A
	.dc.b	$9B,$6C,$78,$A3,$97,$BD,$8B,$FF,$80,$21,$CC,$FF,$90,$2D,$F3,$58
	.dc.b	$AF,$03,$E6,$D1,$CD,$79,$70,$DC,$3F,$C9,$33,$FA,$C7,$A7,$89,$9C
	.dc.b	$3D,$4F,$C0,$FB,$3F,$8D,$3F,$E5,$07,$DF,$5F,$E1,$67,$35,$F8,$19
	.dc.b	$FE,$C3,$1A,$6E,$FF,$2B,$37,$83,$D6,$30,$F7,$F3,$25,$77,$F4,$6E
	.dc.b	$3D,$4F,$D6,$E2,$AF,$AB,$F6,$7F,$B0,$BF,$34,$62,$BD,$5E,$67,$E0
	.dc.b	$9E,$2D,$57,$FC,$DA,$BF,$32,$D8,$FC,$F2,$F8,$7E,$0D,$9D,$C6,$0E
	.dc.b	$9F,$67,$FE,$65,$8F,$D1,$B6,$BF,$50,$D7,$FD,$DE,$D7,$FC,$13,$F6
	.dc.b	$5F,$DE,$E7,$F5,$6D,$8F,$D8,$35,$FF,$7B,$B5,$FF,$08,$FD,$97,$F5
	.dc.b	$E9,$FD,$9B,$63,$F7,$0D,$7F,$DF,$ED,$7F,$C3,$3F,$65,$FE,$66,$9F
	.dc.b	$DD,$B6,$3F,$80,$35,$FF,$0F,$87,$0F,$F8,$8B,$EC,$BF,$D0,$53,$FC
	.dc.b	$11,$B1,$FC,$25,$AF,$26,$BF,$E2,$AF,$94,$F6,$3F,$CB,$1A,$9A,$63
	.dc.b	$EC,$BF,$C3,$B5,$FC,$49,$B3,$FC,$5D,$AF,$A1,$AF,$F8,$E2,$1F,$65
	.dc.b	$F8,$56,$9F,$C7,$0B,$1F,$C8,$5A,$FC,$2F,$B2,$FE,$D7,$1F,$C8,$9B
	.dc.b	$1F,$C9,$DA,$EF,$9F,$F2,$5F,$64,$F8,$FA,$4B,$FE,$57,$19,$7F,$96
	.dc.b	$FD,$93,$EC,$3F,$34,$D8,$CC,$60,$FC,$F7,$D8,$E7,$1F,$E7,$0D,$7E
	.dc.b	$CB,$5F,$F2,$CB,$5F,$EB,$7F,$63,$FE,$46,$AF,$E7,$AD,$8F,$E8,$4D
	.dc.b	$7F,$06,$D7,$FC,$C7,$F6,$3F,$E8,$A9,$FE,$8A,$D8,$FE,$94,$D7,$F1
	.dc.b	$6D,$7F,$CD,$7C,$FF,$CB,$9F,$C7,$B5,$CF,$67,$EE,$E3,$11,$FD,$CB
	.dc.b	$EC,$7F,$E0,$05,$FF,$54,$6C,$7F,$59,$6B,$F9,$A6,$BF,$E7,$AF,$1D
	.dc.b	$F5,$1F,$C1,$DB,$5F,$D7,$9B,$3F,$D4,$27,$F8,$66,$FF,$E2,$2A,$FE
	.dc.b	$C6,$D9,$FE,$D0,$D7,$F5,$2D,$7F,$D1,$3F,$61,$F7,$DF,$ED,$CD,$7F
	.dc.b	$5E,$D7,$71,$FC,$6F,$EC,$39,$47,$F8,$49,$A9,$A6,$3E,$C1,$F2,$3F
	.dc.b	$C8,$DB,$87,$F7,$A6,$D7,$F7,$F6,$BF,$C0,$3E,$C3,$FD,$2D,$3F,$E0
	.dc.b	$06,$C7,$F8,$15,$AE,$FE,$37,$F9,$84,$7F,$1F,$8F,$F1,$3B,$53,$49
	.dc.b	$F1,$7C,$09,$5D,$DE,$7E,$2C,$8F,$13,$6B,$C5,$E2,$78,$09,$3E,$77
	.dc.b	$FA,$17,$D8,$7E,$5E,$7F,$C7,$2D,$4D,$31,$F6,$1F,$DC,$BE,$AF,$FC
	.dc.b	$3E,$D9,$FF,$12,$3F,$E0,$6E,$2C,$F9,$CA,$FF,$12,$FD,$C7,$FB,$02
	.dc.b	$FF,$C5,$4D,$7F,$95,$B5,$FF,$65,$B5,$FF,$26,$FD,$8B,$EC,$7F,$C5
	.dc.b	$F0,$8C,$3F,$E3,$26,$BF,$E3,$76,$BF,$CE,$39,$BF,$AE,$3F,$E3,$86
	.dc.b	$BF,$E3,$D6,$BB,$F2,$BB,$FD,$8F,$F6,$98,$34,$9B,$F9,$3E,$C5,$F4
	.dc.b	$54,$95,$FE,$46,$8C,$BF,$E4,$8F,$B1,$F5,$A6,$91,$F7,$FD,$C7,$91
	.dc.b	$FF,$77,$B5,$FF,$38,$B5,$34,$FF,$7A,$3F,$E5,$66,$BF,$E5,$B6,$BF
	.dc.b	$D7,$1A,$FF,$BF,$3E,$B3,$E0,$13,$0F,$A2,$E3,$EC,$E9,$27,$FC,$C7
	.dc.b	$08,$DF,$CC,$49,$CB,$5F,$F3,$64,$BE,$8B,$A9,$23,$9B,$E0,$FA,$F9
	.dc.b	$1D,$9A,$F2,$FA,$8F,$29,$1F,$1B,$5F,$F1,$60,$A4,$7F,$A6,$A3,$31
	.dc.b	$87,$FD,$3B,$ED,$32,$F8,$9E,$06,$E3,$13,$FE,$A2,$AF,$F6,$7B,$7D
	.dc.b	$1D,$5F,$CB,$6C,$6F,$CE,$AF,$8D,$B5,$CA,$3C,$D3,$89,$CF,$C8,$6B
	.dc.b	$DA,$95,$E6,$C3,$EC,$4E,$A3,$07,$96,$76,$FB,$A3,$C3,$F0,$3F,$EB
	.dc.b	$81,$F5,$78,$33,$F1,$92,$B9,$9F,$E8,$48,$D3,$89,$E5,$38,$8F,$9F
	.dc.b	$1D,$1D,$47,$37,$33,$BD,$3E,$5C,$7F,$A4,$27,$51,$8A,$E5,$3C,$0E
	.dc.b	$64,$30,$48,$E1,$3C,$B1,$FB,$D5,$CA,$7A,$5E,$6F,$FD,$B7,$3F,$0E
	.dc.b	$3C,$95,$C2,$F5,$5F,$04,$F8,$13,$8A,$E1,$18,$3C,$97,$E2,$70,$79
	.dc.b	$6F,$A3,$A7,$CB,$5E,$08,$F1,$DF,$58,$C4,$7C,$53,$81,$E8,$73,$7F
	.dc.b	$4A,$3B,$DF,$B9,$D9,$C4,$F1,$1C,$4C,$C7,$A9,$C9,$B8,$F5,$38,$8F
	.dc.b	$95,$7E,$A8,$FB,$F7,$E2,$BF,$88,$D8,$EA,$D7,$67,$82,$3D,$31,$EF
	.dc.b	$8F,$4C,$F5,$BF,$09,$EB,$8F,$14,$FE,$2D,$FF,$3A,$DA,$3F,$A3,$B6
	.dc.b	$3F,$DE,$ED,$7F,$DF,$8D,$78,$C7,$02,$3C,$1F,$5B,$F9,$FF,$8C,$AF
	.dc.b	$A8,$FC,$FB,$8A,$F3,$DF,$E4,$23,$0F,$AC,$C4,$FC,$43,$E8,$C6,$9F
	.dc.b	$C0,$9D,$CD,$1D,$E7,$F4,$91,$B8,$FA,$D3,$D1,$D3,$DD,$E6,$79,$8E
	.dc.b	$73,$E6,$7F,$43,$3C,$EF,$F8,$23,$6F,$F8,$3B,$67,$C8,$76,$3A,$D7
	.dc.b	$35,$49,$E1,$1E,$13,$81,$E8,$38,$DF,$A6,$B9,$57,$E6,$E7,$84,$62
	.dc.b	$FF,$61,$7D,$23,$E1,$CF,$B2,$FE,$03,$E2,$9D,$47,$53,$37,$D6,$B8
	.dc.b	$D7,$92,$31,$3B,$8F,$29,$0F,$41,$D4,$FE,$45,$F1,$7F,$CB,$08,$FF
	.dc.b	$25,$F8,$AA,$FF,$24,$8B,$7F,$C7,$67,$F9,$9F,$CC,$4A,$FE,$58,$FF
	.dc.b	$36,$F8,$29,$18,$21,$FC,$B9,$FE,$77,$D7,$FC,$F3,$CE,$93,$FC,$D6
	.dc.b	$32,$FE,$89,$FD,$29,$1C,$1B,$21,$FC,$EB,$ED,$BF,$AB,$FD,$4F,$FB
	.dc.b	$8B,$2F,$E6,$ED,$D3,$FB,$53,$68,$17,$8F,$C8,$E7,$10,$FE,$9B,$1A
	.dc.b	$6F,$C3,$F8,$08,$EB,$E4,$7C,$66,$1F,$EF,$B2,$9C,$C7,$4F,$3B,$D3
	.dc.b	$F4,$12,$30,$43,$FD,$FA,$79,$E3,$FC,$9F,$46,$92,$38,$65,$F0,$A3
	.dc.b	$E2,$F9,$FF,$1B,$E6,$A5,$7B,$1F,$93,$F3,$72,$3B,$9F,$15,$EE,$47
	.dc.b	$38,$7F,$8D,$1F,$96,$D7,$E6,$7D,$0F,$98,$F3,$FD,$AF,$D6,$64,$7C
	.dc.b	$C8,$FF,$07,$FD,$3B,$EA,$6F,$8F,$2B,$FB,$EC,$E5,$B3,$FE,$70,$8C
	.dc.b	$47,$C6,$6B,$FE,$29,$0E,$1F,$F1,$63,$C3,$5F,$EF,$AD,$E2,$CD,$79
	.dc.b	$3A,$BC,$BE,$34,$C3,$C6,$FA,$3E,$D2,$47,$C8,$C1,$1E,$52,$46,$7C
	.dc.b	$5F,$41,$1C,$1C,$7E,$E3,$E1,$C4,$8F,$83,$1F,$1B,$E8,$3E,$3F,$8B
	.dc.b	$27,$E2,$C6,$5F,$13,$F1,$48,$F9,$7E,$AB,$E4,$11,$C1,$F8,$0F,$AA
	.dc.b	$FE,$A0,$FF,$94,$5A,$FE,$A1,$A9,$BF,$43,$FE,$22,$8F,$A0,$D7,$FC
	.dc.b	$B4,$D7,$F6,$2D,$5C,$6F,$96,$FF,$55,$B6,$3F,$26,$D7,$F4,$6D,$7E
	.dc.b	$F7,$D5,$7F,$AD,$53,$49,$D7,$FE,$B6,$BF,$E8,$13,$F9,$E6,$BF,$AE
	.dc.b	$6B,$EA,$F1,$BC,$7D,$D2,$3C,$6D,$7F,$6B,$F2,$CF,$2B,$FD,$86,$D8
	.dc.b	$FD,$5B,$5D,$C7,$8F,$E5,$E4,$78,$DA,$FB,$BE,$07,$E1,$91,$F8,$6F
	.dc.b	$FB,$43,$EA,$7F,$86,$47,$EE,$1A,$FF,$08,$6B,$FA,$36,$BF,$53,$CA
	.dc.b	$F8,$84,$7E,$23,$F4,$BC,$39,$1F,$49,$AF,$F1,$4F,$4D,$F5,$3F,$EE
	.dc.b	$26,$C7,$F0,$A6,$BB,$8F,$A4,$47,$E9,$35,$FB,$BF,$05,$F8,$22,$3F
	.dc.b	$82,$7F,$DD,$FF,$4E,$F8,$9A,$49,$FE,$33,$19,$7E,$F7,$9F,$F8,$12
	.dc.b	$38,$3F,$1B,$E4,$7E,$40,$8E,$0F,$E4,$7F,$4F,$FE,$A1,$27,$AB,$FE
	.dc.b	$A6,$5D,$B7,$B7,$0C,$AF,$F4,$AC,$A7,$51,$86,$BF,$EC,$2F,$A7,$74
	.dc.b	$75,$F5,$A4,$F2,$7C,$1C,$D2,$30,$43,$FD,$97,$87,$E3,$FC,$8F,$F4
	.dc.b	$CF,$D3,$BA,$3F,$D7,$AD,$AF,$F5,$4B,$5F,$F6,$9A,$F5,$76,$48,$DB
	.dc.b	$FE,$C6,$FA,$77,$D3,$FF,$B2,$3B,$5F,$1F,$49,$7F,$EB,$68,$7D,$06
	.dc.b	$C9,$19,$7A,$7D,$3F,$FB,$4B,$E9,$3F,$CC,$9A,$FF,$AC,$9B,$E8,$7F
	.dc.b	$B9,$5B,$7F,$DF,$1F,$DD,$CE,$FC,$3E,$44,$9F,$F6,$3C,$BC,$3E,$C4
	.dc.b	$8C,$BF,$EE,$2F,$A7,$FF,$04,$42,$7E,$35,$F8,$F8,$7C,$C7,$9F,$82
	.dc.b	$57,$66,$BF,$EF,$EA,$78,$4B,$FF,$09,$FD,$3B,$7F,$64,$E6,$FF,$C0
	.dc.b	$37,$C2,$FF,$BA,$B6,$33,$3B,$7F,$B0,$43,$E9,$DF,$4F,$C1,$B3,$D1
	.dc.b	$AF,$06,$BF,$E1,$0F,$A7,$FF,$01,$BF,$E2,$86,$A6,$9A,$F7,$1C,$3F
	.dc.b	$EF,$49,$FC,$1B,$63,$31,$F0,$63,$FB,$B7,$26,$33,$FF,$77,$CF,$03
	.dc.b	$47,$F8,$E3,$E0,$3F,$CF,$35,$D6,$B8,$3F,$E3,$C9,$4E,$9E,$27,$F9
	.dc.b	$03,$97,$6E,$FC,$EA,$E6,$BD,$11,$88,$FE,$A9,$1E,$BF,$77,$13,$96
	.dc.b	$47,$F9,$33,$1F,$6D,$78,$BE,$43,$2F,$8B,$14,$78,$27,$F2,$55,$C4
	.dc.b	$F5,$3F,$B7,$71,$5D,$79,$B8,$72,$C9,$CC,$FC,$0F,$0B,$0F,$77,$DB
	.dc.b	$F4,$DF,$E5,$E9,$DB,$F3,$70,$F3,$B9,$AF,$8E,$E6,$7E,$63,$63,$E9
	.dc.b	$78,$1D,$37,$F9,$3A,$23,$BF,$93,$30,$EF,$EF,$DC,$3E,$9F,$D3,$FF
	.dc.b	$9C,$A7,$D2,$D8,$F0,$D7,$D4,$8C,$67,$E6,$F2,$FD,$7F,$1D,$EC,$6E
	.dc.b	$1E,$CF,$8E,$FB,$7F,$6E,$F4,$3E,$0C,$61,$FB,$FF,$50,$F5,$7F,$03
	.dc.b	$F9,$07,$E6,$30,$F9,$73,$97,$E6,$C6,$76,$FC,$5F,$D4,$FF,$B2,$67
	.dc.b	$E8,$BF,$6A,$33,$E3,$7E,$32,$31,$B7,$E1,$3E,$A7,$FB,$3D,$FD,$85
	.dc.b	$E9,$FE,$49,$CE,$1F,$98,$73,$3F,$7B,$EA,$5F,$55,$F9,$85,$F1,$36
	.dc.b	$30,$FE,$73,$EA,$BF,$D0,$D5,$F8,$65,$CF,$9F,$6B,$4F,$E8,$BD,$7F
	.dc.b	$66,$7F,$A8,$7F,$24,$DD,$F5,$DF,$FA,$8F,$AA,$FE,$31,$97,$E4,$97
	.dc.b	$BB,$9F,$37,$4F,$EB,$A5,$F5,$57,$E3,$F0,$D8,$FC,$DF,$D5,$7F,$6C
	.dc.b	$6F,$C2,$7E,$DA,$31,$A7,$EF,$14,$7E,$EB,$EA,$98,$5F,$90,$C3,$70
	.dc.b	$FD,$3F,$D5,$3E,$AB,$F8,$0F,$A3,$F7,$7A,$FF,$73,$39,$AF,$D6,$CE
	.dc.b	$5F,$E1,$DF,$55,$9D,$7F,$86,$FD,$53,$E4,$BF,$85,$CE,$71,$FE,$07
	.dc.b	$39,$BF,$DB,$BF,$E4,$77,$F7,$3E,$73,$E8,$3F,$89,$CE,$71,$FE,$33
	.dc.b	$39,$AF,$DF,$FD,$53,$EA,$FF,$8C,$BF,$86,$AF,$E3,$EE,$67,$7E,$17
	.dc.b	$F1,$EF,$2B,$F7,$55,$FE,$29,$F5,$71,$E8,$E5,$FE,$47,$F5,$7E,$F9
	.dc.b	$FF,$73,$BF,$EB,$5C,$3F,$87,$B9,$83,$97,$F9,$7F,$B0,$F3,$FF,$96
	.dc.b	$AF,$3A,$C6,$1F,$E6,$3C,$FF,$E2,$0C,$3F,$8D,$BF,$07,$8B,$F9,$5B
	.dc.b	$9C,$3F,$9D,$B9,$9F,$E8,$1F,$57,$F8,$07,$F9,$1A,$E9,$FE,$7A,$E6
	.dc.b	$BF,$A1,$B9,$9F,$E5,$3F,$57,$FC,$65,$FF,$70,$B6,$7F,$DE,$2D,$7F
	.dc.b	$D9,$8D,$7F,$A8,$7D,$57,$E8,$27,$FD,$D2,$D8,$FF,$7C,$B5,$FF,$6A
	.dc.b	$35,$FE,$B1,$F7,$5F,$D1,$27,$FD,$E2,$D8,$FF,$80,$2D,$7F,$DB,$8D
	.dc.b	$7C,$3E,$47,$F6,$84,$2B,$F8,$0B,$E2,$8D,$39,$6C,$FF,$27,$AF,$0F
	.dc.b	$A5,$82,$3E,$CF,$57,$8E,$06,$92,$3D,$3F,$31,$D0,$8E,$0F,$F7,$A6
	.dc.b	$3E,$6F,$AA,$FE,$C7,$F5,$58,$84,$61,$DA,$F9,$B7,$26,$E3,$53,$FE
	.dc.b	$FF,$FA,$AF,$F6,$11,$3D,$8F,$F0,$CB,$8D,$E9,$27,$C1,$08,$FF,$0F
	.dc.b	$F3,$7E,$5E,$7C,$8D,$7D,$2D,$7F,$C2,$07,$E7,$F9,$BF,$C4,$FF,$55
	.dc.b	$FE,$DD,$27,$D1,$E0,$FF,$15,$FD,$4B,$E8,$E9,$2F,$E1,$46,$5F,$89
	.dc.b	$F5,$5F,$C3,$5B,$E3,$FF,$8D,$23,$11,$F3,$1A,$F9,$3E,$A7,$CC,$FF
	.dc.b	$89,$5A,$FE,$19,$A9,$BF,$55,$F3,$A3,$EA,$35,$FB,$2D,$7E,$0B,$5F
	.dc.b	$F1,$97,$DB,$7F,$96,$DF,$B8,$D7,$F0,$2D,$7E,$4B,$5F,$F1,$D7,$DC
	.dc.b	$7F,$09,$7F,$08,$D7,$F1,$2D,$7E,$8B,$5F,$0F,$D7,$FF,$96,$24,$D2
	.dc.b	$47,$93,$D4,$D1,$1C,$1F,$93,$FC,$1B,$F4,$E9,$A4,$E9,$FF,$32,$FD
	.dc.b	$63,$EB,$7F,$C7,$31,$FE,$53,$8C,$57,$F8,$D9,$AF,$F9,$51,$7D,$3F
	.dc.b	$08,$FD,$CC,$61,$FD,$23,$5F,$D4,$B5,$FC,$8B,$5F,$04,$73,$51,$C7
	.dc.b	$C6,$FF,$2A,$B6,$BF,$CD,$1E,$B3,$E5,$F8,$4F,$FB,$72,$FD,$1E,$86
	.dc.b	$0E,$9B,$7F,$84,$A3,$24,$3D,$91,$AF,$82,$F1,$7D,$C3,$EB,$FA,$FC
	.dc.b	$F4,$BF,$F2,$CC,$E5,$B3,$97,$E3,$4F,$28,$F6,$78,$7F,$EB,$ED,$7F
	.dc.b	$D0,$0F,$FB,$0A,$BF,$83,$7D,$7E,$BE,$0A,$47,$F9,$9A,$74,$48,$CC
	.dc.b	$7F,$A1,$A3,$33,$8E,$0F,$F3,$DC,$77,$9F,$E5,$AE,$9E,$95,$BB,$F1
	.dc.b	$DF,$98,$C4,$F0,$9C,$3E,$38,$CB,$FC,$76,$32,$F8,$E3,$35,$E1,$8F
	.dc.b	$69,$C4,$F0,$90,$E0,$64,$F6,$4F,$29,$EB,$7E,$C8,$C4,$EE,$30,$F3
	.dc.b	$7E,$0C,$F0,$BF,$9F,$5C,$2B,$B9,$D8,$E1,$1F,$09,$E8,$F0,$72,$7C
	.dc.b	$A9,$E8,$FC,$53,$27,$B8,$EB,$5E,$47,$B1,$89,$EC,$E2,$34,$7A,$1F
	.dc.b	$09,$D4,$D4,$70,$8D,$46,$0D,$38,$24,$7A,$63,$47,$96,$32,$F0,$8D
	.dc.b	$46,$23,$51,$82,$3F,$04,$F2,$9F,$81,$3E,$0D,$FE,$35,$EE,$AF,$FD
	.dc.b	$E6,$D8,$FE,$D0,$D7,$FB,$6B,$5F,$EA,$AD,$7D,$46,$23,$99,$1F,$0B
	.dc.b	$FC,$27,$ED,$DF,$07,$F9,$28,$FC,$BC,$62,$76,$7D,$87,$0F,$A2,$37
	.dc.b	$3C,$63,$4D,$7E,$5C,$F0,$C3,$EC,$E1,$F1,$A7,$E2,$1F,$EF,$A9,$D4
	.dc.b	$67,$0F,$D2,$1F,$A5,$8C,$CF,$D3,$21,$CA,$BF,$DE,$B1,$A8,$C5,$73
	.dc.b	$C3,$E0,$61,$E5,$BF,$1D,$7F,$8B,$2F,$86,$1E,$0B,$F6,$CF,$C2,$3D
	.dc.b	$77,$C7,$0F,$55,$F0,$AC,$57,$3A,$AF,$88,$C1,$CF,$0F,$0A,$AE,$4E
	.dc.b	$6F,$FC,$7E,$E6,$B9,$E1,$F0,$DB,$87,$B2,$78,$1F,$12,$F9,$BF,$5C
	.dc.b	$FF,$24,$E1,$F1,$AB,$8C,$63,$0E,$F5,$C2,$BC,$B8,$77,$AF,$E5,$DF
	.dc.b	$33,$FC,$D8,$8F,$13,$E5,$7C,$16,$1F,$E6,$11,$6B,$EA,$CF,$F5,$9F
	.dc.b	$0F,$FA,$75,$7F,$35,$F9,$1F,$E8,$26,$78,$64,$7F,$3D,$8C,$BD,$CF
	.dc.b	$F4,$84,$7F,$48,$F0,$A4,$7F,$43,$7F,$A5,$FA,$78,$8F,$E8,$51,$71
	.dc.b	$FE,$83,$3A,$9E,$04,$3F,$A4,$3F,$DB,$FE,$C3,$FC,$45,$1F,$E7,$26
	.dc.b	$A6,$93,$CC,$E0,$48,$F4,$7C,$EF,$F9,$3E,$DC,$1F,$7A,$FF,$C5,$51
	.dc.b	$8A,$ED,$F2,$FE,$13,$C5,$F0,$F2,$3F,$CD,$95,$FE,$2D,$7C,$91,$FE
	.dc.b	$6F,$7C,$8D,$8F,$F3,$14,$7C,$96,$AA,$F0,$3E,$BF,$33,$E0,$7C,$57
	.dc.b	$C1,$64,$71,$72,$F3,$8E,$C4,$7C,$AF,$87,$E3,$24,$7A,$1A,$FF,$A2
	.dc.b	$3E,$D3,$FC,$DF,$7E,$2C,$15,$F4,$4F,$F1,$04,$EB,$3F,$91,$F6,$3F
	.dc.b	$57,$EB,$26,$3E,$D8,$FB,$07,$C1,$23,$F2,$1F,$8D,$F6,$0F,$B1,$A4
	.dc.b	$9F,$F4,$C4,$65,$B1,$FE,$31,$8C,$47,$C9,$BF,$DB,$C6,$33,$F9,$6F
	.dc.b	$FA,$76,$7F,$30,$D7,$FD,$4E,$D7,$FD,$1E,$D7,$9F,$4F,$9B,$EC,$64
	.dc.b	$79,$1A,$FF,$AC,$3E,$A5,$F3,$7E,$59,$93,$FA,$79,$CB,$E2,$FB,$1F
	.dc.b	$57,$95,$23,$04,$3C,$B1,$F1,$88,$FC,$4F,$AB,$F9,$24,$70,$43,$E2
	.dc.b	$1F,$ED,$8C,$91,$F2,$4F,$95,$E5,$FF,$96,$F8,$1F,$11,$8B,$F8,$F3
	.dc.b	$98,$FF,$30,$7D,$8F,$F9,$0C,$9E,$36,$3C,$6F,$F3,$27,$D6,$3D,$9A
	.dc.b	$4C,$7F,$84,$43,$C5,$F0,$7C,$4C,$9F,$E1,$9A,$23,$C9,$EE,$7F,$9B
	.dc.b	$FE,$C5,$EF,$D2,$5F,$F1,$38,$FD,$5F,$47,$EA,$B2,$FF,$1E,$13,$4F
	.dc.b	$2F,$C4,$FB,$FC,$29,$87,$81,$F8,$5E,$57,$F9,$F7,$E8,$9E,$85,$25
	.dc.b	$7F,$21,$FC,$0D,$F0,$9F,$03,$D1,$93,$EE,$7E,$1F,$9B,$F3,$7D,$A9
	.dc.b	$3E,$E7,$E8,$72,$7F,$A2,$FE,$C9,$E6,$52,$57,$F2,$F8,$FE,$2B,$E2
	.dc.b	$FF,$54,$E2,$7A,$5F,$40,$95,$F3,$BD,$0F,$B9,$E7,$48,$F9,$AF,$DF
	.dc.b	$F2,$BF,$D3,$5F,$12,$FB,$2A,$4A,$FE,$7D,$0F,$92,$FB,$B3,$27,$ED
	.dc.b	$BF,$80,$FA,$DF,$C5,$7C,$5C,$9F,$B6,$FE,$37,$E6,$7F,$D5,$1F,$65
	.dc.b	$F8,$E9,$FE,$94,$D7,$F2,$E1,$C7,$77,$F2,$47,$FD,$AC,$FD,$DF,$94
	.dc.b	$FC,$79,$27,$F1,$BF,$03,$F9,$BF,$B2,$91,$F8,$C7,$FD,$71,$FB,$77
	.dc.b	$F2,$EB,$FE,$5F,$F6,$7F,$EB,$3A,$F0,$AF,$A1,$FF,$5F,$36,$3F,$B0
	.dc.b	$3B,$9E,$3E,$7E,$BE,$0E,$56,$A3,$FD,$8D,$EA,$3A,$39,$C4,$9F,$E0
	.dc.b	$91,$DB,$D6,$91,$D9,$F0,$F8,$72,$30,$43,$CE,$FF,$B4,$FE,$CF,$FC
	.dc.b	$60,$4F,$0B,$1F,$65,$2B,$FB,$C4,$A3,$11,$F4,$3E,$CD,$CF,$FE,$DD
	.dc.b	$F2,$5F,$37,$49,$5F,$E0,$18,$7D,$7E,$C9,$19,$7F,$DC,$D7,$FE,$E1
	.dc.b	$8F,$F0,$47,$D9,$7F,$B2,$C9,$EA,$FF,$BA,$DF,$F5,$91,$3C,$FC,$19
	.dc.b	$F3,$65,$72,$6B,$F8,$8F,$83,$FE,$A8,$85,$7F,$94,$C9,$BF,$7D,$A4
	.dc.b	$F9,$1A,$FE,$4A,$7F,$5B,$CD,$CF,$EC,$DB,$7F,$91,$27,$35,$FE,$46
	.dc.b	$3C,$0E,$1C,$35,$E1,$E3,$FF,$7F,$78,$1E,$6F,$B2,$7D,$9E,$23,$0D
	.dc.b	$A4,$EA,$78,$C6,$27,$FC,$21,$0F,$A5,$A4,$7F,$C2,$8F,$F8,$57,$E2
	.dc.b	$3F,$C3,$D9,$FF,$84,$27,$EF,$F9,$9F,$E3,$D5,$3A,$8F,$77,$2D,$9F
	.dc.b	$F3,$2B,$57,$C5,$3E,$13,$11,$D0,$FD,$EC,$66,$BB,$73,$EB,$DB,$DB
	.dc.b	$63,$C5,$1C,$BD,$1C,$5F,$F0,$63,$F9,$3C,$F9,$E3,$C5,$6F,$B3,$E1
	.dc.b	$3F,$E5,$78,$C3,$FE,$59,$D1,$CF,$E9,$7F,$D6,$71,$98,$C7,$34,$8E
	.dc.b	$5E,$15,$CB,$D6,$FC,$1F,$B3,$F8,$F3,$FE,$65,$7E,$06,$1F,$E3,$C7
	.dc.b	$35,$FC,$79,$CC,$FF,$A3,$9B,$1F,$99,$C7,$FC,$17,$3C,$F7,$F6,$F8
	.dc.b	$1D,$FC,$C9,$3D,$FE,$2E,$E1,$FC,$AB,$EC,$FF,$45,$3F,$E7,$76,$C7
	.dc.b	$86,$BE,$C4,$63,$3F,$37,$C1,$FA,$FE,$03,$D8,$DC,$3D,$9E,$EF,$DF
	.dc.b	$FB,$37,$AB,$F0,$63,$0F,$E0,$FE,$CF,$FB,$BD,$7F,$A1,$FE,$CF,$FC
	.dc.b	$DE,$FF,$A2,$63,$3A,$FF,$40,$9C,$BF,$D0,$BD,$17,$AB,$F8,$DF,$C3
	.dc.b	$F3,$B7,$F9,$9A,$73,$8F,$FA,$4E,$72,$FF,$AE,$3E,$CF,$F2,$6F,$FA
	.dc.b	$5D,$7D,$D7,$E5,$DC,$E3,$FD,$41,$CC,$FF,$A7,$7E,$D3,$FC,$46,$DE
	.dc.b	$1F,$CF,$46,$2B,$FA,$BA,$8F,$D1,$7E,$05,$6B,$FD,$4C,$DA,$FE,$BA
	.dc.b	$B8,$37,$5F,$A9,$F5,$BF,$9A,$5F,$FA,$B5,$B7,$FD,$91,$FE,$FF,$19
	.dc.b	$F2,$FF,$61,$F6,$9F,$2B,$7F,$F4,$4B,$67,$FD,$B2,$BD,$FD,$FC,$4F
	.dc.b	$ED,$FE,$51,$F3,$D4,$C7,$C3,$5C,$3F,$D7,$DF,$69,$FD,$65,$BE,$D7
	.dc.b	$EF,$E3,$F2,$BB,$7F,$78,$74,$FF,$7A,$FB,$4F,$E9,$75,$FE,$C7,$6C
	.dc.b	$67,$E0,$3F,$84,$46,$36,$FF,$50,$FE,$05,$F4,$AB,$FC,$06,$FF,$91
	.dc.b	$EF,$FD,$A0,$E6,$77,$F1,$FF,$C4,$7E,$CB,$FD,$15,$5F,$ED,$57,$F8
	.dc.b	$75,$FF,$83,$1C,$B7,$D0,$FE,$31,$E5,$7F,$95,$EB,$FD,$B8,$BD,$BC
	.dc.b	$FC,$8E,$9F,$E3,$DF,$81,$7F,$9A,$AB,$FD,$C4,$DC,$3F,$C3,$2B,$E5
	.dc.b	$37,$F0,$7F,$C9,$7C,$8F,$E2,$F5,$FE,$1E,$6E,$DF,$EC,$07,$FA,$6E
	.dc.b	$1F,$EE,$D7,$2F,$F5,$B9,$7D,$97,$B2,$BF,$D8,$ED,$9D,$3F,$C7,$A7
	.dc.b	$FC,$56,$E5,$FE,$C9,$F8,$13,$C5,$FE,$70,$FF,$AF,$AB,$FD,$F8,$E5
	.dc.b	$BD,$3F,$CF,$7E,$CF,$FC,$BF,$1F,$F0,$05,$7D,$AE,$7E,$53,$4E,$FE
	.dc.b	$1B,$87,$4D,$BE,$0F,$F4,$7F,$B4,$FF,$36,$C7,$F5,$F6,$E9,$FD,$91
	.dc.b	$AF,$F6,$D6,$BF,$0F,$EC,$FF,$84,$BF,$DC,$5A,$E7,$C0,$FE,$B8,$FF
	.dc.b	$80,$E3,$2D,$F0,$72,$A3,$F4,$D7,$FD,$C5,$B1,$FD,$D9,$AF,$7F,$4F
	.dc.b	$8C,$1A,$49,$E5,$E9,$E8,$8E,$0F,$EF,$DF,$75,$EC,$BF,$F0,$C3,$53
	.dc.b	$4C,$7D,$9F,$F5,$DD,$FF,$C0,$CD,$9F,$F0,$5B,$FD,$C7,$E0,$33,$EE
	.dc.b	$FF,$A5,$A7,$FC,$1F,$F6,$55,$B7,$FE,$0F,$51,$97,$FC,$38,$D7,$F2
	.dc.b	$4F,$E0,$1F,$F0,$B3,$5E,$BC,$5C,$E3,$11,$FE,$25,$F3,$3F,$90,$3F
	.dc.b	$E1,$D6,$BF,$E2,$26,$BB,$F3,$7C,$9F,$61,$F5,$A0,$D2,$6B,$E9,$FB
	.dc.b	$27,$CD,$D2,$57,$F8,$AE,$32,$FF,$8B,$7E,$C3,$FC,$4E,$47,$E1,$3F
	.dc.b	$E2,$AF,$B1,$F1,$C7,$EC,$5A,$FF,$93,$5A,$9B,$F8,$0F,$C7,$8F,$F2
	.dc.b	$93,$53,$4C,$7D,$7F,$B4,$98,$FD,$37,$1F,$67,$49,$5F,$E4,$58,$47
	.dc.b	$D9,$FB,$3F,$EC,$F3,$FE,$49,$6B,$FE,$4E,$6B,$FE,$57,$6B,$FC,$36
	.dc.b	$7F,$BD,$BD,$7D,$FF,$11,$1F,$07,$D7,$F9,$48,$E0,$F2,$7D,$97,$A4
	.dc.b	$8E,$0F,$F2,$E7,$D9,$3A,$BF,$8E,$78,$ED,$3F,$8F,$4E,$6B,$FC,$F5
	.dc.b	$39,$7F,$CF,$91,$97,$FD,$8F,$97,$FA,$D5,$BE,$D7,$C9,$23,$F1,$DA
	.dc.b	$E8,$EB,$7F,$06,$3C,$6F,$C3,$F3,$EC,$FF,$A2,$E9,$F0,$B8,$84,$61
	.dc.b	$E6,$6E,$35,$0F,$99,$B1,$27,$67,$27,$71,$D7,$D2,$FE,$A5,$3A,$5E
	.dc.b	$96,$4D,$7A,$52,$75,$19,$78,$47,$3F,$73,$D1,$F6,$3F,$E2,$29,$F0
	.dc.b	$B5,$FF,$78,$DF,$FB,$15,$F0,$4F,$0F,$81,$91,$82,$1F,$EF,$8A,$E5
	.dc.b	$C1,$F5,$E1,$18,$DB,$FA,$D6,$1D,$67,$BE,$1D,$1E,$58,$79,$DE,$B8
	.dc.b	$73,$C3,$94,$7F,$AF,$5A,$FF,$A3,$EB,$83,$C8,$F6,$D6,$2B,$C9,$7E
	.dc.b	$17,$A4,$78,$5C,$D6,$F0,$FE,$E1,$87,$63,$BD,$F1,$8F,$5C,$F1,$77
	.dc.b	$7C,$70,$FF,$69,$56,$DF,$44,$6D,$CD,$6D,$F8,$F1,$97,$D8,$7C,$2A
	.dc.b	$EA,$FB,$0C,$1C,$8C,$1F,$04,$F9,$A6,$8E,$A6,$B0,$F7,$B9,$C3,$DD
	.dc.b	$5E,$3C,$3D,$13,$C0,$E9,$5C,$1E,$A6,$2F,$CA,$E2,$34,$7C,$E7,$C1
	.dc.b	$1F,$21,$F0,$57,$BE,$BB,$9E,$0B,$F1,$CF,$A6,$31,$3B,$C3,$E4,$56
	.dc.b	$AB,$C1,$7D,$5F,$21,$F4,$0E,$53,$F3,$67,$ED,$BF,$88,$8F,$F7,$D5
	.dc.b	$FE,$55,$B1,$FD,$96,$31,$1D,$4C,$9F,$EF,$77,$2F,$D3,$3F,$DF,$AF
	.dc.b	$B2,$3E,$67,$CA,$7C,$D6,$CF,$90,$F7,$46,$5A,$64,$F6,$9F,$8C,$74
	.dc.b	$E0,$FF,$78,$38,$C7,$E8,$E1,$FB,$08,$C3,$F1,$63,$F4,$71,$89,$F4
	.dc.b	$46,$1F,$03,$C9,$F0,$18,$9E,$47,$57,$07,$27,$0F,$96,$FE,$13,$FC
	.dc.b	$6E,$BC,$0F,$E7,$8F,$E2,$2E,$23,$F0,$6D,$BF,$E4,$8D,$9E,$47,$80
	.dc.b	$E5,$7E,$97,$C1,$3F,$39,$A7,$CC,$29,$E5,$9E,$8F,$88,$E9,$5E,$26
	.dc.b	$CF,$89,$FE,$69,$F3,$DF,$CD,$C8,$FF,$37,$F8,$BA,$FF,$2C,$8B,$87
	.dc.b	$F1,$29,$FE,$7D,$DD,$FC,$FF,$E9,$E5,$FF,$3F,$8C,$BF,$D0,$63,$2F
	.dc.b	$F3,$68,$FE,$A5,$F6,$1F,$E8,$58,$5F,$F3,$18,$C3,$63,$7E,$47,$06
	.dc.b	$1F,$D4,$A7,$7A,$FF,$A3,$D4,$F0,$9D,$1C,$0E,$F3,$B8,$F1,$FC,$8F
	.dc.b	$F9,$2E,$8D,$25,$79,$C0,$8D,$BE,$38,$E6,$F9,$7E,$63,$DF,$E7,$FC
	.dc.b	$0F,$42,$4F,$A6,$32,$F5,$3A,$3E,$32,$39,$BF,$F5,$13,$F0,$7D,$7F
	.dc.b	$96,$7C,$CF,$5A,$C6,$48,$7C,$03,$D3,$28,$E1,$1D,$5F,$84,$47,$C1
	.dc.b	$F4,$FF,$14,$8E,$0F,$27,$D2,$FD,$C8,$91,$F5,$63,$FC,$21,$F5,$6F
	.dc.b	$AF,$BE,$3C,$AF,$F0,$5C,$E5,$B3,$FE,$7D,$8C,$47,$D1,$6B,$F8,$96
	.dc.b	$BF,$EC,$F5,$F6,$BC,$49,$1F,$36,$32,$D8,$F2,$91,$F1,$FD,$87,$A0
	.dc.b	$F1,$11,$E4,$F9,$7E,$C2,$47,$56,$BE,$7F,$B0,$F8,$31,$23,$DF,$1F
	.dc.b	$13,$E0,$BE,$2F,$8B,$27,$E1,$C6,$5F,$43,$F0,$C8,$FA,$FC,$3F,$8C
	.dc.b	$47,$04,$3F,$00,$32,$FF,$1E,$3F,$42,$3F,$C8,$9F,$91,$7F,$AB,$7A
	.dc.b	$7F,$C4,$5F,$53,$FE,$76,$AF,$F1,$FC,$E7,$AB,$FD,$F8,$D8,$FD,$1C
	.dc.b	$ED,$FF,$16,$75,$FF,$8E,$3E,$B6,$47,$F8,$E2,$3F,$CA,$5F,$65,$FB
	.dc.b	$68,$FF,$25,$AF,$9B,$C9,$AE,$EF,$F2,$C6,$1F,$82,$CF,$CB,$E1,$E0
	.dc.b	$CF,$2F,$F9,$1B,$E0,$E6,$5F,$E4,$67,$FC,$93,$F0,$7E,$0F,$A0,$93
	.dc.b	$E0,$7C,$7F,$41,$DB,$E0,$24,$FF,$97,$B3,$47,$8B,$D5,$FF,$35,$7B
	.dc.b	$9F,$C4,$2B,$F8,$1B,$EB,$7E,$7B,$75,$FF,$26,$3B,$87,$07,$F0,$87
	.dc.b	$EA,$AF,$17,$E2,$9F,$1B,$C6,$F8,$3D,$DF,$0B,$BA,$5F,$96,$3C,$7C
	.dc.b	$92,$3F,$CE,$B8,$FC,$B6,$EF,$F4,$3E,$4E,$47,$D0,$7E,$9F,$C9,$FC
	.dc.b	$EF,$C0,$64,$FC,$17,$E8,$FB,$3F,$03,$8E,$4F,$67,$E1,$EB,$F7,$3C
	.dc.b	$A9,$3F,$71,$FB,$FE,$57,$DA,$F6,$64,$FC,$D7,$EE,$F8,$BF,$33,$8B
	.dc.b	$E6,$EB,$2B,$D7,$1F,$88,$F8,$6F,$C4,$FB,$52,$7F,$13,$1F,$86,$F1
	.dc.b	$FF,$0F,$F8,$1C,$9F,$B5,$1F,$EA,$1F,$AD,$79,$37,$DB,$95,$FE,$9C
	.dc.b	$9C,$B7,$CA,$E1,$52,$30,$7F,$A8,$BE,$B5,$F2,$FD,$6A,$4E,$63,$C1
	.dc.b	$EE,$48,$EC,$F8,$FC,$C9,$18,$21,$FB,$8C,$7F,$6E,$FE,$EB,$EB,$5F
	.dc.b	$49,$FE,$BB,$6C,$7F,$AD,$1A,$EF,$D9,$D1,$1D,$3F,$EC,$3F,$AD,$79
	.dc.b	$3F,$EC,$5E,$4F,$C4,$57,$F0,$86,$F8,$BF,$EA,$56,$FA,$1F,$D8,$A7
	.dc.b	$F3,$1E,$0F,$12,$3F,$EC,$69,$47,$7F,$2B,$FD,$A5,$F5,$3F,$EC,$6A
	.dc.b	$FE,$2A,$DC,$3F,$D6,$4D,$C3,$B2,$FA,$DE,$24,$AF,$F6,$9C,$65,$BF
	.dc.b	$51,$D8,$8E,$0F,$F6,$CF,$D6,$FE,$CE,$7E,$3E,$1E,$0E,$0F,$29,$E6
	.dc.b	$DE,$5F,$66,$BE,$8F,$7F,$FC,$32,$BE,$B7,$F7,$4F,$F8,$1B,$04,$EA
	.dc.b	$3F,$C4,$11,$8C,$3F,$C0,$0D,$8F,$F0,$73,$5D,$FD,$6B,$EB,$76,$D8
	.dc.b	$DB,$5D,$C7,$F8,$2D,$FF,$0B,$FD,$6D,$4D,$23,$FE,$11,$F5,$9C,$1F
	.dc.b	$EF,$E8,$FE,$68,$D8,$CC,$7F,$4C,$32,$5F,$3B,$39,$7F,$BA,$FE,$09
	.dc.b	$EB,$E3,$0F,$F2,$05,$5B,$FF,$15,$D6,$A7,$C1,$CB,$99,$FF,$17,$F9
	.dc.b	$95,$FC,$5C,$F9,$63,$06,$7C,$CB,$1E,$97,$0F,$F8,$E3,$2F,$27,$AB
	.dc.b	$8B,$FE,$CB,$8F,$F3,$EA,$D1,$F9,$DA,$73,$78,$DF,$D8,$C6,$AB,$A4
	.dc.b	$E7,$96,$4E,$65,$7F,$C4,$23,$17,$CA,$7C,$3E,$1F,$BF,$EB,$7F,$CB
	.dc.b	$35,$C5,$F9,$B5,$F4,$29,$1A,$7F,$BE,$BA,$9F,$DA,$36,$3E,$6F,$83
	.dc.b	$D7,$7F,$99,$8A,$3C,$1E,$4C,$C3,$C1,$E4,$DC,$3F,$C1,$3F,$5B,$FE
	.dc.b	$60,$9F,$63,$63,$C3,$5F,$5E,$31,$9F,$9F,$37,$CE,$FB,$7E,$4F,$DC
	.dc.b	$DC,$7D,$DE,$AF,$DF,$FA,$D7,$B5,$F0,$E3,$0F,$E0,$FE,$B5,$EA,$FE
	.dc.b	$17,$EB,$7F,$2F,$87,$CF,$9C,$BF,$46,$33,$BF,$F8,$8B,$E2,$1F,$5B
	.dc.b	$F8,$DF,$AD,$CF,$07,$CD,$9C,$E1,$F6,$67,$2F,$F1,$FF,$AD,$FB,$6F
	.dc.b	$DD,$5F,$1F,$F9,$77,$38,$FF,$8C,$1C,$CF,$E0,$FE,$B7,$E1,$B7,$8B
	.dc.b	$F3,$D1,$8A,$FF,$1B,$A8,$FD,$17,$D6,$DA,$FC,$63,$6B,$FC,$7A,$B8
	.dc.b	$77,$5F,$A9,$F5,$DE,$5D,$31,$E5,$5C,$7F,$2D,$F5,$BF,$23,$2F,$F2
	.dc.b	$3B,$78,$3F,$20,$BD,$EF,$DA,$39,$AF,$E8,$5F,$1C,$F7,$7F,$70,$D8
	.dc.b	$FD,$12,$F8,$2C,$CF,$EF,$7E,$B7,$FB,$4D,$7E,$9D,$B9,$7F,$96,$1F
	.dc.b	$CB,$46,$7D,$BF,$E0,$9F,$5B,$FD,$62,$3F,$5C,$DA,$FF,$2F,$AF,$5B
	.dc.b	$7E,$C7,$F0,$BF,$AD,$FA,$75,$FE,$64,$5F,$5B,$9E,$9D,$3F,$C4,$7E
	.dc.b	$B7,$FD,$31,$97,$EE,$DF,$E2,$D7,$FE,$6C,$72,$DF,$4B,$F8,$C7,$98
	.dc.b	$F9,$9B,$F3,$38,$6E,$1F,$C1,$7E,$B5,$F0,$7F,$C7,$DB,$19,$FA,$DC
	.dc.b	$4F,$F2,$3F,$2B,$FD,$AB,$3F,$C3,$5B,$9F,$F9,$F9,$FF,$25,$46,$7D
	.dc.b	$5F,$E5,$5F,$5B,$E8,$EC,$FE,$14,$D9,$FE,$F4,$FF,$21,$BF,$F4,$3B
	.dc.b	$97,$FD,$BF,$F5,$AF,$1F,$F9,$A3,$FE,$C9,$AF,$E3,$CE,$5B,$D9,$FC
	.dc.b	$E7,$EB,$5F,$53,$7E,$A7,$0D,$AF,$E4,$D3,$F6,$73,$FE,$3C,$FD,$98
	.dc.b	$CF,$4F,$3E,$AF,$E8,$9C,$1D,$27,$FD,$D0,$DD,$7F,$DE,$CD,$7F,$C2
	.dc.b	$C7,$CE,$C7,$F9,$53,$FE,$19,$FA,$CF,$CE,$57,$FB,$C5,$B1,$FE,$FF
	.dc.b	$6B,$C9,$AF,$F8,$87,$EE,$7F,$8B,$CF,$FB,$E5,$B1,$E0,$0E,$1F,$1B
	.dc.b	$5F,$17,$AB,$E4,$42,$91,$98,$C3,$63,$35,$E1,$F9,$3D,$5F,$FB,$8A
	.dc.b	$30,$4B,$F2,$7D,$67,$FA,$73,$1F,$34,$23,$0E,$BA,$FF,$D3,$8F,$F8
	.dc.b	$FE,$32,$FC,$CF,$B9,$FE,$66,$F2,$6B,$F0,$5A,$EF,$D4,$FF,$0A,$7D
	.dc.b	$6B,$8E,$92,$7C,$71,$97,$CB,$E6,$FF,$1D,$6F,$9D,$FE,$1D,$8C,$47
	.dc.b	$CF,$6B,$F8,$09,$FD,$93,$D3,$BB,$C2,$47,$E0,$FD,$63,$E7,$3F,$C5
	.dc.b	$0D,$8C,$C6,$0F,$F1,$67,$D6,$FF,$B7,$E3,$F1,$AD,$7F,$CC,$8D,$7E
	.dc.b	$DB,$57,$D6,$79,$9F,$F3,$53,$5F,$C0,$B5,$DE,$5F,$E3,$BF,$AD,$FA
	.dc.b	$71,$FE,$71,$6B,$F8,$86,$A6,$FD,$57,$E2,$60,$D2,$71,$FF,$91,$FE
	.dc.b	$B3,$FA,$84,$FE,$11,$AF,$E6,$DA,$FE,$B1,$AF,$FA,$0F,$EB,$3F,$A1
	.dc.b	$3F,$8C,$6B,$FA,$36,$BF,$B4,$6B,$FE,$8B,$A7,$3F,$F9,$65,$B1,$F9
	.dc.b	$66,$BF,$AD,$6B,$CF,$D1,$FF,$30,$57,$F4,$F8,$F1,$46,$0E,$9F,$39
	.dc.b	$FE,$3D,$C7,$C0,$D9,$FD,$8D,$A3,$11,$F2,$3E,$72,$C9,$F3,$12,$3C
	.dc.b	$CF,$A7,$CB,$DF,$B1,$27,$FC,$89,$5F,$EC,$E5,$E1,$FF,$8F,$63,$11
	.dc.b	$FC,$61,$AF,$C5,$AE,$8F,$C3,$7C,$64,$7F,$CA,$19,$BE,$DF,$CF,$F3
	.dc.b	$52,$7C,$2F,$B3,$EE,$1F,$3F,$4D,$4F,$93,$6F,$A9,$59,$8F,$F2,$E7
	.dc.b	$7F,$FA,$BD,$D1,$FC,$61,$FF,$5C,$37,$D0,$F9,$31,$BD,$7F,$CB,$4D
	.dc.b	$BF,$F6,$53,$C2,$BF,$D0,$F5,$C3,$E6,$38,$FB,$9C,$BD,$D9,$5E,$28
	.dc.b	$CB,$E4,$8C,$BE,$43,$FD,$23,$1A,$71,$7F,$E2,$47,$81,$CC,$86,$A7
	.dc.b	$9E,$3E,$8A,$E4,$FA,$A3,$38,$FA,$A7,$2F,$B2,$33,$5F,$16,$72,$FC
	.dc.b	$78,$CB,$E6,$3C,$35,$D5,$F2,$9E,$89,$E2,$FF,$AB,$27,$8D,$FF,$B3
	.dc.b	$A3,$8D,$F8,$5E,$37,$E9,$72,$78,$4C,$C7,$A5,$C9,$D4,$C9,$E9,$33
	.dc.b	$3F,$25,$F8,$C6,$2F,$C7,$7E,$69,$F2,$4E,$89,$7E,$37,$D8,$76,$3C
	.dc.b	$06,$08,$FC,$37,$8D,$F9,$DC,$BE,$F7,$27,$9C,$C9,$EF,$32,$7C,$F3
	.dc.b	$27,$82,$3B,$1E,$52,$47,$57,$2F,$85,$CC,$75,$72,$F8,$5C,$BD,$63
	.dc.b	$91,$C4,$8F,$17,$2F,$17,$DC,$FF,$86,$A7,$E5,$B6,$FF,$DF,$D1,$88
	.dc.b	$D9,$0F,$84,$6C,$FE,$FE,$FD,$4A,$F2,$B6,$BF,$C2,$32,$FA,$EF,$F3
	.dc.b	$2B,$F7,$4F,$C2,$B8,$7E,$4C,$F1,$7D,$07,$F8,$6A,$35,$5D,$AF,$F3
	.dc.b	$31,$8B,$F8,$35,$F4,$EB,$E3,$CE,$63,$99,$99,$DC,$66,$30,$6A,$FC
	.dc.b	$4F,$A7,$D6,$FC,$96,$1E,$06,$C7,$A9,$AF,$E0,$AB,$A5,$72,$25,$7F
	.dc.b	$42,$8C,$BC,$30,$FF,$7A,$46,$2B,$E3,$1F,$C4,$5C,$3F,$1A,$79,$5F
	.dc.b	$1C,$3D,$6D,$9F,$7D,$70,$BF,$91,$1F,$E4,$47,$BC,$EE,$3E,$09,$EF
	.dc.b	$3C,$31,$F2,$A3,$58,$7C,$4A,$E9,$7F,$42,$31,$3A,$C3,$CB,$87,$2A
	.dc.b	$F2,$B4,$F2,$CF,$F3,$CF,$8C,$91,$FD,$11,$FE,$CF,$D7,$FD,$46,$30
	.dc.b	$43,$F9,$8C,$6B,$87,$FA,$2F,$74,$9F,$E9,$AF,$F4,$8F,$A6,$C4,$7F
	.dc.b	$38,$8B,$AF,$F4,$A9,$FE,$EB,$F5,$8F,$95,$B7,$0F,$91,$A7,$E7,$27
	.dc.b	$51,$83,$FC,$DE,$FF,$8A,$09,$3F,$E6,$48,$46,$1E,$06,$23,$C2,$7F
	.dc.b	$81,$BE,$6F,$FC,$A7,$82,$FC,$1F,$39,$E6,$F6,$26,$1C,$B1,$4F,$A8
	.dc.b	$CC,$7F,$59,$84,$E6,$38,$7C,$27,$C0,$23,$CF,$E4,$2B,$C6,$2E,$3E
	.dc.b	$38,$F7,$CF,$C8,$F5,$FE,$4F,$B3,$27,$C9,$19,$7F,$D0,$B8,$A3,$E2
	.dc.b	$36,$33,$1F,$1F,$E9,$7F,$D5,$F8,$77,$9F,$93,$3C,$DF,$EE,$93,$AD
	.dc.b	$3E,$A7,$D5,$FD,$AF,$AC,$95,$F1,$63,$EE,$9E,$02,$3E,$47,$EB,$7D
	.dc.b	$5B,$EA,$A9,$27,$FD,$3B,$19,$6C,$7F,$8F,$23,$11,$FE,$07,$6E,$5E
	.dc.b	$95,$D5,$BF,$26,$57,$E0,$A7,$FC,$A7,$5B,$FA,$9E,$87,$90,$93,$E2
	.dc.b	$7F,$19,$F4,$CF,$98,$DC,$F3,$9F,$F2,$7C,$E6,$3B,$7D,$3F,$9F,$CB
	.dc.b	$91,$82,$1F,$EC,$FC,$53,$F1,$09,$1F,$0B,$DE,$F8,$E4,$70,$43,$E1
	.dc.b	$3F,$14,$F8,$DF,$0B,$F1,$FC,$79,$3F,$05,$FF,$32,$FD,$5F,$FB,$FA
	.dc.b	$3F,$4A,$D7,$FD,$C4,$D7,$FD,$E6,$D7,$FA,$7B,$5F,$F7,$0F,$D5,$FF
	.dc.b	$A9,$C9,$F5,$9F,$B3,$7F,$D5,$5E,$3F,$FA,$D6,$7F,$97,$78,$3F,$CD
	.dc.b	$4F,$E6,$DE,$17,$F9,$E3,$A7,$F7,$F3,$F9,$0F,$9D,$C1,$9F,$0F,$FC
	.dc.b	$FB,$8F,$E0,$3A,$FF,$B0,$BF,$80,$AF,$DF,$B5,$FF,$26,$3B,$7F,$AF
	.dc.b	$7C,$2F,$F4,$52,$3F,$C1,$B3,$7C,$2F,$8F,$E8,$A4,$FF,$0A,$CD,$1E
	.dc.b	$7F,$A2,$F1,$FD,$14,$AF,$0B,$E6,$F9,$DF,$E5,$B1,$FC,$BB,$58,$FF
	.dc.b	$14,$F3,$3F,$58,$86,$3F,$C2,$9F,$07,$93,$F1,$4F,$8D,$E4,$FC,$2F
	.dc.b	$03,$E1,$F7,$4A,$F8,$31,$FC,$F3,$8E,$47,$77,$F9,$FE,$DF,$3F,$E5
	.dc.b	$64,$FB,$5F,$A5,$ED,$7C,$EF,$AC,$93,$F3,$1F,$A3,$EA,$FF,$48,$E2
	.dc.b	$93,$EB,$7F,$A5,$E9,$F6,$FE,$16,$4F,$C9,$7E,$F7,$9D,$F6,$BD,$C9
	.dc.b	$3F,$61,$FB,$BE,$27,$F5,$4E,$1F,$EA,$BA,$CA,$F9,$11,$F8,$7F,$3F
	.dc.b	$F1,$1F,$17,$27,$EB,$47,$E1,$BC,$5F,$C3,$FD,$64,$9F,$C2,$C7,$FB
	.dc.b	$1B,$EB,$3F,$D7,$9E,$5F,$F2,$15,$EF,$E4,$D7,$BD,$BF,$2A,$E1,$FD
	.dc.b	$32,$B8,$F8,$9F,$E3,$25,$3F,$C1,$3E,$5F,$85,$6E,$30,$F1,$F3,$BC
	.dc.b	$44,$7A,$79,$7E,$53,$0F,$FA,$96,$FF,$87,$36,$3F,$DB,$FF,$58,$FA
	.dc.b	$5A,$49,$FE,$C3,$1F,$E8,$1D,$7F,$9F,$37,$3F,$F2,$4B,$67,$F2,$4D
	.dc.b	$5F,$58,$F2,$7F,$DD,$BC,$8F,$02,$92,$BF,$B7,$C7,$E9,$9B,$E0,$70
	.dc.b	$23,$FD,$D2,$FF,$5E,$FF,$76,$FA,$BF,$F6,$61,$3E,$07,$FD,$F6,$FF
	.dc.b	$B4,$63,$F4,$CD,$4D,$27,$D4,$70,$24,$75,$7F,$C0,$0F,$4D,$A3,$E1
	.dc.b	$7B,$FC,$0C,$8F,$23,$5F,$F6,$B5,$3E,$05,$C1,$FB,$9F,$AC,$FF,$02
	.dc.b	$D7,$9A,$72,$FF,$9D,$8C,$4D,$69,$6B,$A4,$66,$3D,$BF,$58,$FA,$CC
	.dc.b	$52,$31,$3C,$5A,$F7,$8D,$4F,$F8,$96,$1F,$59,$48,$FF,$8A,$5F,$F1
	.dc.b	$57,$9D,$FE,$11,$C3,$FC,$09,$3F,$7F,$97,$FC,$56,$FF,$8E,$5B,$C7
	.dc.b	$B7,$DC,$D8,$FE,$39,$F0,$9F,$DD,$27,$67,$F8,$D1,$C4,$7F,$C0,$13
	.dc.b	$F9,$6C,$76,$F2,$39,$47,$F2,$5E,$4A,$A2,$98,$F2,$2D,$79,$9C,$3D
	.dc.b	$FD,$1B,$5F,$62,$3F,$D5,$38,$27,$BE,$4E,$4B,$C9,$86,$DF,$F9,$32
	.dc.b	$7F,$60,$FA,$FB,$F3,$8F,$53,$E0,$BE,$AF,$AB,$F9,$11,$90,$FC,$D7
	.dc.b	$0E,$AB,$FC,$68,$E5,$BB,$7C,$4C,$FF,$C2,$98,$7F,$97,$AB,$4F,$B3
	.dc.b	$BB,$C1,$E2,$C8,$D7,$8B,$72,$F1,$7D,$67,$F5,$19,$FF,$33,$B6,$3C
	.dc.b	$75,$F4,$A3,$19,$F9,$3E,$4F,$D8,$7B,$7E,$76,$E1,$FE,$70,$FA,$C7
	.dc.b	$AD,$F6,$9B,$1E,$7F,$5B,$13,$F7,$BE,$AF,$FC,$E5,$3F,$E7,$86,$E9
	.dc.b	$F1,$FE,$B1,$E9,$6E,$BF,$0B,$F5,$6F,$7E,$98,$F8,$5B,$3F,$E8,$2F
	.dc.b	$AB,$7C,$77,$E3,$1B,$19,$FA,$DC,$4F,$E4,$3E,$B1,$BD,$31,$F1,$76
	.dc.b	$FF,$D1,$BF,$58,$8F,$CB,$AF,$8C,$B1,$87,$F3,$7E,$B7,$EE,$6B,$FD
	.dc.b	$2A,$BE,$CF,$3E,$5E,$9F,$D0,$FD,$6B,$B6,$F6,$E1,$B8,$7F,$A7,$BE
	.dc.b	$3F,$E2,$B7,$DF,$FD,$4C,$63,$A3,$F2,$0A,$3F,$5B,$F5,$AF,$63,$F5
	.dc.b	$FF,$5A,$FA,$4C,$C6,$1F,$D9,$FE,$1D,$F8,$2C,$BF,$C3,$19,$23,$F3
	.dc.b	$11,$87,$51,$FE,$1D,$72,$FF,$87,$FE,$B9,$D1,$FB,$E6,$D7,$FA,$E1
	.dc.b	$7C,$76,$67,$F8,$1F,$D6,$FE,$DE,$BF,$D7,$AD,$BF,$D6,$3F,$8E,$6F
	.dc.b	$9F,$FC,$2B,$CD,$7C,$D5,$31,$E7,$DC,$3F,$D8,$FF,$5A,$F8,$3F,$E2
	.dc.b	$2F,$EB,$B1,$FD,$C3,$99,$DF,$9D,$FC,$5B,$CB,$7C,$35,$F8,$BC,$36
	.dc.b	$BF,$DA,$7F,$5B,$E5,$F4,$3F,$D6,$0B,$E3,$7F,$8F,$B9,$9F,$9A,$E5
	.dc.b	$FF,$6D,$FD,$6B,$1F,$E4,$AD,$8C,$FA,$58,$9F,$E5,$3F,$5A,$FA,$7A
	.dc.b	$63,$E1,$2D,$FF,$B9,$BF,$18,$FA,$FC,$26,$9F,$61,$C1,$FC,$D3,$E2
	.dc.b	$DF,$55,$FD,$1A,$3F,$CB,$51,$8A,$FE,$B4,$D8,$FE,$C0,$D7,$FB,$27
	.dc.b	$D7,$7E,$76,$7F,$B0,$36,$3F,$B3,$35,$DF,$8B,$DD,$43,$C6,$C3,$6B
	.dc.b	$9C,$60,$E1,$F6,$7F,$9F,$8F,$ED,$CD,$AF,$EE,$8D,$7F,$BC,$35,$FC
	.dc.b	$FF,$D7,$7F,$47,$F0,$FF,$BB,$36,$3F,$BE,$3F,$ED,$2F,$84,$CF,$7F
	.dc.b	$FA,$9F,$EB,$7F,$D5,$14,$69,$2F,$FC,$01,$19,$DA,$CF,$CE,$8F,$F0
	.dc.b	$23,$5F,$F0,$53,$5F,$F0,$77,$23,$EC,$79,$B6,$3F,$C1,$8D,$7F,$C2
	.dc.b	$4D,$7F,$C2,$DF,$5B,$F3,$53,$49,$AF,$9B,$EB,$BF,$D7,$93,$FE,$1D
	.dc.b	$6B,$FC,$11,$AF,$FA,$75,$AB,$EB,$7F,$C2,$84,$7E,$0B,$FE,$17,$FA
	.dc.b	$EF,$E6,$31,$FE,$23,$6B,$FE,$29,$6B,$FE,$2D,$6B,$FC,$53,$EC,$FE
	.dc.b	$79,$35,$F9,$AF,$D2,$FA,$EF,$71,$31,$FA,$2F,$C1,$FA,$FF,$E7,$57
	.dc.b	$FE,$33,$6B,$FE,$39,$6B,$FE,$3D,$6B,$FC,$91,$F9,$8D,$D7,$EE,$C6
	.dc.b	$23,$FD,$80,$D5,$3F,$DA,$1F,$F2,$13,$5F,$F2,$43,$5F,$F2,$63,$5F
	.dc.b	$E6,$7F,$82,$FC,$E2,$79,$9F,$8C,$7E,$8C,$BE,$7B,$DE,$4A,$EE,$FE
	.dc.b	$4B,$EC,$3F,$59,$1F,$ED,$5F,$11,$9F,$F4,$09,$CC,$FF,$86,$DB,$3A
	.dc.b	$8F,$F6,$E5,$FF,$AB,$1E,$55,$FD,$22,$31,$1E,$CF,$67,$E8,$91,$F8
	.dc.b	$3E,$47,$F4,$EB,$E9,$0F,$B0,$7D,$0E,$71,$FE,$A7,$18,$9E,$E7,$8A
	.dc.b	$F4,$D3,$A9,$0F,$9A,$FE,$08,$DB,$F5,$1B,$8F,$FA,$CB,$0F,$1B,$FD
	.dc.b	$A6,$7B,$8F,$AC,$FE,$60,$E1,$AF,$F9,$05,$B3,$EC,$5E,$1F,$81,$B3
	.dc.b	$C7,$C4,$C1,$99,$D7,$CC,$F9,$BE,$99,$F6,$74,$CD,$F8,$67,$0F,$9A
	.dc.b	$76,$E4,$F3,$7D,$A7,$C0,$F8,$4F,$F0,$84,$7B,$1A,$FA,$63,$CC,$6F
	.dc.b	$C7,$91,$C8,$87,$F9,$7E,$78,$E9,$F6,$9B,$C1,$FE,$0A,$D5,$87,$4B
	.dc.b	$F0,$61,$D1,$F0,$1D,$30,$E5,$19,$7E,$54,$F1,$9F,$04,$E6,$FB,$4E
	.dc.b	$6B,$8D,$76,$3A,$4E,$9C,$1C,$8C,$83,$EF,$4F,$57,$18,$7B,$1D,$61
	.dc.b	$EC,$78,$9F,$E4,$CC,$38,$3C,$8C,$CF,$17,$47,$09,$E4,$F8,$C9,$87
	.dc.b	$C0,$78,$D7,$49,$DC,$E6,$77,$87,$F8,$E2,$FB,$56,$A3,$E0,$4E,$8F
	.dc.b	$5D,$FC,$78,$CD,$FC,$89,$D1,$E9,$3E,$A9,$8B,$EB,$87,$18,$E6,$7C
	.dc.b	$C9,$F2,$12,$BC,$97,$E9,$AE,$15,$F0,$8E,$F8,$79,$DF,$9B,$1F,$2B
	.dc.b	$0E,$F1,$F0,$8E,$D7,$E8,$23,$E2,$39,$CF,$69,$FA,$07,$28,$F5,$9F
	.dc.b	$EF,$9A,$FA,$2D,$9F,$EE,$F1,$88,$F9,$26,$4F,$F7,$D3,$97,$CE,$62
	.dc.b	$3D,$B1,$EB,$8F,$E1,$1F,$25,$89,$CB,$63,$F0,$4D,$7E,$7E,$3F,$C2
	.dc.b	$63,$0F,$CB,$8F,$C8,$C6,$30,$F8,$31,$D8,$FC,$65,$76,$69,$D8,$A7
	.dc.b	$68,$FD,$B4,$F3,$8F,$59,$FC,$6D,$C3,$F8,$3A,$F2,$3C,$9F,$23,$E1
	.dc.b	$3A,$1C,$9C,$DA,$79,$BF,$14,$C4,$F3,$AF,$5C,$62,$39,$9E,$18,$CC
	.dc.b	$F2,$BF,$E5,$8D,$9F,$E6,$8D,$7D,$47,$53,$C9,$3F,$01,$C1,$D4,$F9
	.dc.b	$A6,$2B,$CD,$3C,$EF,$E0,$CF,$92,$7D,$57,$E2,$8F,$E8,$9E,$C7,$F3
	.dc.b	$F2,$3F,$D1,$FA,$2B,$FD,$1E,$2D,$FF,$47,$9F,$EA,$BF,$5F,$FE,$B2
	.dc.b	$A4,$FF,$66,$8C,$36,$37,$E3,$70,$61,$FD,$62,$77,$9F,$FA,$F1,$4F
	.dc.b	$6F,$96,$FE,$C7,$F8,$27,$F8,$26,$FF,$AF,$B5,$F0,$DF,$19,$E1,$1E
	.dc.b	$53,$F9,$45,$A3,$FC,$BE,$D8,$D0,$70,$F1,$6E,$1E,$D9,$EB,$3E,$03
	.dc.b	$AF,$D0,$7C,$4F,$82,$F8,$BF,$21,$27,$DD,$19,$7E,$04,$65,$E4,$78
	.dc.b	$63,$E6,$FD,$8C,$8F,$8C,$FD,$0F,$85,$C4,$70,$C9,$38,$CB,$E9,$57
	.dc.b	$D6,$F8,$AF,$F0,$5F,$D5,$3E,$CE,$99,$C7,$E9,$4F,$C4,$23,$F1,$31
	.dc.b	$47,$C8,$F8,$CE,$13,$E1,$8F,$C1,$C7,$F8,$63,$EC,$5F,$25,$7E,$4E
	.dc.b	$5F,$F8,$5A,$72,$D9,$FF,$48,$C6,$23,$EC,$35,$FC,$73,$5F,$F8,$02
	.dc.b	$FF,$B6,$5F,$B2,$D7,$8F,$BF,$D0,$8F,$85,$F3,$1E,$7F,$AC,$FF,$17
	.dc.b	$FE,$08,$FB,$9A,$4A,$F4,$7D,$6F,$B4,$F3,$91,$F0,$BE,$AF,$AC,$D4
	.dc.b	$70,$3E,$03,$F1,$7D,$FF,$8D,$F1,$72,$7E,$24,$65,$F5,$BF,$10,$8F
	.dc.b	$BB,$E3,$FE,$39,$1C,$10,$FC,$5E,$9F,$E4,$56,$C7,$F9,$33,$EC,$9E
	.dc.b	$B7,$F9,$3B,$EC,$7F,$CC,$7D,$1F,$E5,$E6,$F7,$7F,$83,$F0,$46,$1F
	.dc.b	$F0,$95,$BE,$3B,$A1,$FE,$CA,$71,$1F,$AC,$6B,$FB,$36,$BF,$C2,$57
	.dc.b	$C7,$71,$24,$71,$7F,$DA,$BE,$47,$E7,$A3,$FC,$89,$DB,$FE,$47,$3F
	.dc.b	$C9,$1D,$BF,$9E,$24,$7F,$99,$3B,$51,$FA,$2F,$88,$F8,$1C,$52,$BF
	.dc.b	$CD,$3A,$23,$E2,$71,$7F,$94,$EB,$FC,$A9,$EF,$CB,$FF,$2A,$47,$EB
	.dc.b	$23,$F5,$B9,$C9,$F0,$47,$CD,$FB,$3F,$9D,$F1,$F2,$7B,$C7,$F9,$7E
	.dc.b	$A4,$7F,$97,$DF,$F3,$17,$7F,$EE,$09,$1E,$B7,$F7,$7D,$FF,$67,$D4
	.dc.b	$93,$EA,$7E,$E7,$3F,$F9,$AE,$7F,$CD,$9E,$E4,$AF,$F3,$64,$7F,$04
	.dc.b	$8F,$F4,$7F,$DA,$7F,$A7,$BA,$3F,$D3,$ED,$E8,$FC,$17,$27,$FB,$E6
	.dc.b	$7F,$D4,$7F,$B7,$FE,$2B,$1F,$F4,$8A,$F8,$5F,$F4,$CB,$63,$3D,$7F
	.dc.b	$E9,$CF,$C1,$BA,$47,$F9,$6D,$78,$B9,$35,$97,$91,$F2,$F8,$B2,$BF
	.dc.b	$D3,$FA,$70,$7C,$3E,$C7,$83,$D8,$91,$83,$FD,$4B,$8F,$F0,$DD,$7F
	.dc.b	$D7,$6D,$CB,$F3,$CD,$9F,$27,$A7,$F1,$08,$F8,$1F,$8F,$F0,$FE,$EF
	.dc.b	$0A,$4F,$C2,$7E,$0F,$83,$FE,$BA,$FB,$2F,$C6,$4F,$FA,$D9,$FD,$52
	.dc.b	$69,$A7,$FA,$1A,$EE,$E5,$DF,$27,$94,$7D,$8E,$DE,$CC,$8F,$8B,$1E
	.dc.b	$1E,$A9,$1E,$17,$C9,$D5,$23,$04,$3C,$2F,$FB,$4B,$EC,$9F,$53,$49
	.dc.b	$3F,$ED,$08,$7B,$7B,$24,$65,$FF,$6D,$FD,$8B,$EC,$BF,$DB,$BE,$A3
	.dc.b	$EA,$E9,$2F,$FD,$B7,$0F,$AD,$D9,$23,$2F,$4F,$B5,$FF,$73,$FD,$8F
	.dc.b	$FB,$62,$7F,$B5,$B7,$A7,$FD,$F2,$DE,$9F,$E3,$8F,$F1,$48,$CF,$37
	.dc.b	$22,$57,$FB,$B2,$7F,$8A,$B7,$9B,$B9,$1F,$F7,$8F,$DB,$FF,$86,$9F
	.dc.b	$F5,$B4,$FC,$87,$97,$0F,$98,$F3,$F0,$48,$EC,$D7,$D3,$F6,$7F,$E2
	.dc.b	$C5,$F6,$8D,$FD,$33,$99,$FF,$19,$60,$8C,$46,$B4,$FF,$0E,$CE,$5B
	.dc.b	$F4,$CF,$B4,$A4,$AF,$F1,$45,$27,$51,$FE,$26,$1F,$69,$5A,$65,$C1
	.dc.b	$BF,$81,$70,$FF,$B6,$71,$B5,$FD,$37,$92,$C6,$23,$FD,$BD,$58,$AF
	.dc.b	$F2,$75,$59,$E1,$1C,$CF,$EA,$DF,$0A,$F0,$35,$3F,$EE,$7A,$B3,$9A
	.dc.b	$E1,$3F,$D7,$79,$B6,$BE,$D2,$C6,$19,$3C,$27,$FD,$E5,$F0,$58,$7F
	.dc.b	$C7,$27,$F9,$36,$34,$D3,$FC,$7B,$6F,$0B,$BD,$79,$1F,$3E,$9F,$D8
	.dc.b	$AF,$B3,$FB,$06,$F3,$EB,$9A,$47,$8F,$0F,$EB,$0F,$08,$F7,$F8,$3F
	.dc.b	$0B,$ED,$3F,$CD,$73,$E9,$7F,$60,$DE,$0D,$6F,$23,$A7,$81,$76,$F4
	.dc.b	$BC,$FE,$6E,$4F,$3F,$89,$B5,$FE,$5A,$FB,$43,$F4,$9B,$3E,$1A,$C4
	.dc.b	$F9,$7D,$69,$3E,$5F,$5E,$CF,$F9,$93,$ED,$1F,$69,$F7,$1B,$5E,$BF
	.dc.b	$63,$13,$F8,$0F,$B4,$FC,$7C,$FD,$76,$DF,$F9,$BB,$ED,$1F,$15,$BA
	.dc.b	$FC,$3F,$DA,$7F,$52,$7E,$F3,$6B,$FC,$EA,$FF,$97,$23,$3E,$AF,$E3
	.dc.b	$BE,$D1,$F6,$94,$C7,$C8,$5A,$FC,$47,$DA,$36,$FC,$AB,$63,$3F,$01
	.dc.b	$89,$FC,$CF,$DA,$7F,$9C,$27,$F2,$0B,$E0,$73,$5A,$7F,$3F,$F3,$3F
	.dc.b	$B7,$AF,$CB,$2F,$B4,$CF,$93,$A7,$F4,$BF,$68,$EB,$BD,$58,$6E,$1F
	.dc.b	$9F,$F8,$7F,$8E,$DF,$93,$FD,$6C,$63,$97,$FD,$32,$A3,$F6,$3F,$68
	.dc.b	$F6,$BF,$67,$F6,$9F,$E5,$0C,$3F,$55,$39,$7F,$59,$19,$C3,$FD,$47
	.dc.b	$F6,$9F,$88,$6F,$89,$FB,$C8,$C7,$95,$FC,$1D,$FC,$EC,$FF,$AA,$BE
	.dc.b	$D1,$D1,$FC,$09,$B8,$FE,$E5,$7C,$CE,$67,$F8,$3F,$DA,$7E,$D6,$BF
	.dc.b	$7C,$DB,$FF,$5A,$BF,$90,$6F,$9D,$FC,$3B,$CB,$FF,$31,$57,$F0,$56
	.dc.b	$CF,$FA,$F1,$FD,$E3,$7D,$BF,$E2,$BF,$68,$F8,$1B,$E6,$61,$B8,$7F
	.dc.b	$0F,$F2,$9F,$67,$FC,$71,$7D,$A5,$8C,$3F,$C8,$3E,$D3,$CD,$C7,$FC
	.dc.b	$AD,$7C,$DE,$7C,$DD,$3F,$C9,$BE,$D1,$8D,$31,$E8,$5C,$FF,$90,$7D
	.dc.b	$A3,$E3,$3F,$96,$B6,$33,$F1,$98,$9F,$E6,$3F,$69,$F5,$A7,$FD,$14
	.dc.b	$DE,$4F,$E0,$AF,$D8,$DB,$FA,$03,$98,$FF,$10,$57,$F4,$37,$FD,$25
	.dc.b	$F1,$7D,$DF,$F7,$E3,$5F,$F4,$B1,$C2,$FF,$DF,$7F,$68,$F1,$79,$4F
	.dc.b	$F1,$48,$C4,$EA,$73,$29,$CB,$D0,$7D,$CF,$CA,$AC,$46,$38,$38,$CE
	.dc.b	$63,$83,$5F,$2F,$DA,$7F,$41,$F9,$3E,$2D,$8E,$CF,$F5,$9F,$0B,$3D
	.dc.b	$BF,$E6,$D9,$F6,$FD,$CF,$F1,$87,$C2,$D7,$CC,$D7,$7E,$BF,$F7,$7F
	.dc.b	$B5,$FF,$39,$47,$AD,$AF,$C8,$6B,$FE,$BF,$6B,$FC,$EB,$97,$F8,$C3
	.dc.b	$F0,$1A,$FC,$56,$BF,$49,$AF,$FB,$27,$ED,$7E,$6A,$69,$36,$FF,$06
	.dc.b	$7D,$AF,$ED,$27,$E6,$B5,$FA,$4D,$7E,$FB,$5E,$DE,$CF,$F8,$56,$BF
	.dc.b	$D0,$91,$F6,$5A,$FE,$19,$AF,$FB,$77,$ED,$3C,$EF,$F5,$86,$A6,$98
	.dc.b	$FB,$56,$DF,$E2,$36,$CF,$E1,$1A,$FE,$25,$AF,$E5,$BE,$BB,$E1,$93
	.dc.b	$2F,$F1,$63,$F1,$BE,$D1,$A7,$F8,$BD,$B5,$F9,$06,$BF,$95,$6B,$FA
	.dc.b	$25,$F3,$FC,$49,$1C,$5E,$FC,$B2,$3F,$37,$7F,$C1,$9B,$1F,$E3,$FF
	.dc.b	$B2,$FE,$1F,$1F,$A0,$6B,$FA,$56,$BF,$B2,$6B,$FE,$02,$12,$FB,$8F
	.dc.b	$79,$23,$C2,$D5,$DB,$FE,$43,$9E,$11,$FE,$31,$6B,$AC,$FB,$43,$EA
	.dc.b	$3C,$15,$C7,$CB,$B5,$AF,$6A,$E1,$FE,$16,$C3,$FD,$66,$7F,$AC,$6F
	.dc.b	$84,$7C,$4B,$FA,$A4,$7E,$3F,$95,$C9,$F6,$B6,$8D,$F5,$24,$FC,$E7
	.dc.b	$EF,$91,$C1,$ED,$76,$FB,$4F,$F1,$5F,$D9,$3E,$D3,$C1,$F2,$3F,$E9
	.dc.b	$B9,$FF,$17,$B5,$FF,$31,$78,$DA,$F4,$36,$C8,$FF,$2D,$4E,$CF,$F5
	.dc.b	$16,$DF,$5D,$FF,$45,$46,$BE,$63,$B9,$E0,$F9,$69,$19,$21,$E7,$3F
	.dc.b	$CE,$4F,$9B,$E3,$FC,$FF,$63,$27,$C3,$19,$7C,$71,$97,$95,$A7,$51
	.dc.b	$89,$D7,$0F,$FA,$1E,$B5,$3D,$0E,$64,$39,$D7,$29,$E1,$5C,$B1,$F8
	.dc.b	$13,$97,$E1,$46,$6B,$E6,$CE,$5F,$9F,$19,$7F,$D2,$11,$97,$FD,$25
	.dc.b	$19,$BF,$2C,$78,$8F,$25,$F8,$9F,$41,$93,$DC,$6A,$FC,$AE,$A3,$83
	.dc.b	$A3,$83,$F0,$EF,$C1,$7E,$3B,$F0,$CF,$02,$5F,$4B,$E2,$FC,$07,$C0
	.dc.b	$7C,$03,$C0,$E5,$F0,$47,$77,$27,$C4,$38,$9F,$10,$C9,$F1,$0C,$CF
	.dc.b	$92,$7C,$53,$E4,$9E,$47,$90,$F1,$1E,$08,$F1,$47,$86,$34,$75,$34
	.dc.b	$70,$74,$75,$74,$E2,$3C,$27,$42,$47,$17,$2F,$92,$39,$4F,$A6,$34
	.dc.b	$66,$78,$BF,$04,$FB,$87,$F8,$5A,$3F,$06,$DB,$FF,$76,$46,$23,$EA
	.dc.b	$19,$3F,$C1,$EE,$5F,$86,$64,$F8,$2F,$8D,$FF,$47,$7D,$87,$E0,$67
	.dc.b	$F2,$9F,$0C,$8C,$4F,$D4,$8C,$3F,$5A,$B9,$1C,$E3,$31,$DA,$B9,$D7
	.dc.b	$28,$FA,$33,$AC,$3E,$64,$7C,$43,$E1,$3F,$EF,$D8,$D6,$1F,$32,$72
	.dc.b	$D8,$CC,$71,$C3,$C7,$87,$F2,$F9,$CC,$F0,$35,$19,$C3,$C9,$87,$8A
	.dc.b	$7E,$19,$ED,$9F,$4B,$4F,$85,$3C,$30,$F1,$36,$3D,$57,$E0,$C3,$C7
	.dc.b	$1C,$0E,$47,$C2,$9E,$0F,$94,$F0,$9F,$24,$E0,$53,$81,$CB,$0F,$45
	.dc.b	$77,$C3,$9B,$67,$58,$7F,$6D,$8C,$47,$3A,$F9,$D7,$FD,$1B,$D9,$93
	.dc.b	$FD,$01,$FE,$93,$F7,$38,$8F,$E5,$F1,$6B,$F9,$64,$FF,$5D,$FB,$1F
	.dc.b	$F0,$DC,$31,$FE,$F9,$1F,$E7,$08,$C5,$FF,$C0,$08,$C4,$EB,$C0,$E4
	.dc.b	$73,$98,$0F,$E9,$13,$DA,$7A,$FC,$AF,$84,$93,$BB,$EE,$A9,$F3,$7E
	.dc.b	$73,$71,$D2,$B0,$F1,$BF,$4C,$EA,$F9,$DA,$F7,$7E,$4F,$82,$FF,$3F
	.dc.b	$BD,$DB,$3F,$ED,$F8,$FF,$41,$B5,$FF,$44,$36,$BA,$CF,$CB,$3E,$67
	.dc.b	$C2,$7C,$DF,$84,$F9,$DF,$57,$30,$F3,$1F,$2E,$34,$F6,$7C,$24,$7C
	.dc.b	$3F,$15,$F5,$C8,$F2,$3E,$37,$C5,$7D,$B3,$2F,$D0,$8C,$3B,$C9,$3E
	.dc.b	$E2,$47,$B9,$F8,$7F,$5A,$FA,$7A,$49,$FF,$52,$C6,$5B,$1F,$E2,$78
	.dc.b	$C4,$7F,$81,$DB,$A7,$DC,$5C,$DB,$F1,$A5,$7A,$E7,$FC,$AC,$F1,$FB
	.dc.b	$1E,$87,$90,$8F,$89,$F3,$F8,$B2,$3C,$AD,$7F,$D7,$D2,$FA,$FF,$34
	.dc.b	$C9,$F0,$47,$A4,$F8,$24,$7D,$EF,$C4,$F6,$A4,$7C,$26,$BD,$9F,$84
	.dc.b	$47,$D9,$F5,$BF,$17,$C4,$91,$EF,$7F,$C9,$DF,$5F,$FE,$C3,$D7,$F4
	.dc.b	$2F,$FB,$16,$7F,$DC,$2D,$7F,$CE,$0D,$7F,$5C,$D4,$BC,$3F,$CE,$B6
	.dc.b	$3F,$DB,$AD,$7F,$DC,$4D,$7B,$FC,$EF,$92,$64,$7E,$3E,$53,$E8,$24
	.dc.b	$65,$FF,$34,$7D,$63,$E7,$BF,$9E,$73,$CA,$FC,$C7,$82,$8F,$E8,$1D
	.dc.b	$1F,$E6,$FF,$AE,$34,$95,$F9,$F8,$7C,$27,$F9,$D7,$EB,$FF,$CF,$9E
	.dc.b	$07,$FA,$4D,$B5,$FC,$45,$A9,$71,$7F,$4F,$AF,$EA,$1E,$A4,$AF,$2C
	.dc.b	$78,$E3,$C9,$9C,$9F,$14,$78,$78,$FC,$5F,$17,$27,$F6,$98,$A7,$FA
	.dc.b	$E5,$C9,$F8,$2F,$F5,$FE,$EF,$80,$48,$F7,$3F,$0F,$B7,$DD,$E7,$49
	.dc.b	$EA,$FC,$1E,$7F,$ED,$13,$FD,$A7,$D4,$95,$F3,$A3,$E6,$47,$FA,$6F
	.dc.b	$EB,$BF,$96,$F4,$FF,$2E,$6F,$0F,$FA,$EF,$8B,$FD,$9B,$87,$F0,$DE
	.dc.b	$6F,$CA,$4F,$FA,$FD,$7D,$7F,$DE,$3E,$FF,$C1,$7F,$A9,$FE,$BF,$9D
	.dc.b	$FF,$41,$7F,$DE,$F9,$7F,$3C,$6E,$1F,$EC,$97,$72,$CB,$2F,$F4,$97
	.dc.b	$F8,$4C,$FF,$B3,$DB,$19,$F7,$7F,$1F,$F0,$FF,$90,$F4,$E4,$FA,$63
	.dc.b	$C9,$E3,$48,$FE,$51,$8F,$24,$FF,$14,$7F,$96,$61,$5F,$F5,$03,$72
	.dc.b	$F9,$1E,$37,$CB,$F8,$3F,$CF,$F8,$52,$FE,$43,$FA,$2F,$07,$FD,$95
	.dc.b	$F5,$FF,$92,$1E,$D6,$0C,$FB,$5A,$F0,$6E,$1F,$CF,$6B,$84,$7D,$9E
	.dc.b	$FD,$4E,$63,$0E,$FA,$FB,$91,$E5,$D7,$E2,$30,$FF,$AE,$EF,$FD,$B6
	.dc.b	$E8,$FE,$99,$F5,$FF,$B9,$6F,$8B,$FE,$E3,$8C,$47,$FB,$0D,$7B,$19
	.dc.b	$24,$72,$7F,$DD,$5F,$5E,$FA,$FF,$F7,$67,$A6,$FA,$6A,$4B,$FE,$B9
	.dc.b	$0F,$A9,$D9,$23,$2F,$4F,$AF,$FF,$7B,$7D,$67,$FB,$94,$9D,$B8,$E6
	.dc.b	$97,$FD,$9E,$7F,$DC,$6F,$F8,$55,$A9,$FE,$3A,$EF,$9B,$A9,$1F,$EE
	.dc.b	$1F,$5F,$FE,$22,$79,$70,$78,$88,$F7,$DF,$CA,$47,$07,$8F,$DF,$75
	.dc.b	$FF,$89,$FE,$BF,$FC,$3F,$3E,$59,$CB,$FE,$2B,$26,$48,$CC,$64,$F5
	.dc.b	$38,$76,$D7,$E0,$7D,$13,$EB,$F8,$46,$1A,$FC,$38,$D3,$AA,$46,$23
	.dc.b	$7F,$5F,$FE,$2A,$4D,$23,$D3,$DE,$FF,$25,$E5,$FE,$3E,$8F,$C1,$78
	.dc.b	$DF,$E1,$B5,$3A,$8F,$67,$2D,$9F,$84,$D5,$F0,$5F,$E0,$C9,$E5,$39
	.dc.b	$8C,$4F,$CE,$9F,$8D,$CB,$C6,$1F,$5F,$63,$0C,$9F,$27,$C0,$5A,$E4
	.dc.b	$7D,$37,$91,$DF,$C3,$B3,$FE,$35,$C1,$1E,$7C,$FF,$B4,$CF,$F8,$E5
	.dc.b	$FE,$CC,$DE,$5D,$72,$48,$E5,$E1,$5C,$BC,$27,$C0,$7E,$17,$D7,$FC
	.dc.b	$89,$FF,$22,$3F,$32,$76,$E5,$BB,$70,$3E,$37,$83,$6B,$D0,$F4,$F9
	.dc.b	$69,$1D,$3E,$06,$DF,$8B,$EB,$FF,$CF,$D3,$FE,$53,$6C,$78,$2B,$EB
	.dc.b	$46,$35,$FF,$3E,$E4,$F6,$3D,$27,$AB,$D9,$B8,$FF,$97,$7E,$BD,$F3
	.dc.b	$9F,$7D,$B1,$EE,$F6,$71,$3F,$83,$FA,$FF,$F4,$D4,$FF,$99,$5B,$AF
	.dc.b	$C8,$FA,$F7,$CB,$EE,$BF,$15,$F8,$2B,$EA,$E9,$8E,$FB,$3F,$E6,$FF
	.dc.b	$C4,$5F,$77,$F9,$16,$C7,$F9,$CF,$EC,$5F,$1F,$9A,$FC,$B7,$D8,$7F
	.dc.b	$9C,$2B,$FC,$F0,$DD,$7E,$DA,$F8,$1D,$FC,$17,$E7,$7F,$06,$FF,$0F
	.dc.b	$DF,$F9,$FD,$B7,$F8,$27,$FC,$4F,$19,$E5,$FD,$27,$DC,$7C,$6A,$FC
	.dc.b	$3B,$7A,$7F,$D5,$BE,$D3,$ED,$77,$CF,$FA,$BF,$B2,$7C,$AD,$31,$F1
	.dc.b	$97,$4F,$F4,$87,$E0,$DF,$6D,$FE,$28,$DD,$3F,$07,$EB,$FE,$D6,$31
	.dc.b	$0E,$6F,$F1,$25,$7C,$BF,$8B,$FD,$D7,$D9,$7F,$83,$23,$07,$FA,$77
	.dc.b	$EC,$BF,$35,$C1,$FE,$C4,$6C,$7F,$23,$7F,$A3,$3B,$F2,$3F,$82,$7A
	.dc.b	$5E,$F8,$FE,$50,$DC,$BF,$DA,$8F,$ED,$23,$FD,$54,$E6,$14,$FA,$EF
	.dc.b	$E1,$CD,$AF,$F5,$87,$D8,$3E,$5F,$35,$FC,$4F,$CD,$FD,$DE,$1F,$B6
	.dc.b	$6E,$7F,$ED,$C7,$FB,$3E,$1F,$EB,$97,$2F,$F3,$DF,$C1,$7F,$12,$FF
	.dc.b	$AF,$5B,$5F,$C0,$17,$D9,$7F,$21,$73,$7F,$EE,$8F,$C4,$5F,$1F,$FC
	.dc.b	$95,$B8,$7F,$B1,$FB,$3F,$42,$FF,$B2,$63,$38,$FF,$0C,$FB,$17,$CD
	.dc.b	$FF,$2D,$6D,$7F,$B3,$D7,$DE,$66,$7F,$99,$7E,$22,$F9,$3A,$62,$E9
	.dc.b	$C7,$E9,$7A,$1F,$CE,$7E,$CB,$FD,$2D,$19,$3F,$DB,$9F,$65,$F7,$B3
	.dc.b	$07,$0D,$F8,$0F,$E8,$51,$8F,$90,$7C,$57,$F4,$DA,$E3,$18,$BF,$EB
	.dc.b	$AD,$8F,$EC,$2D,$7F,$B5,$7E,$00,$F9,$1E,$50,$8C,$57,$F6,$66,$C7
	.dc.b	$F6,$A6,$BF,$DC,$FE,$C7,$FA,$0F,$27,$F6,$C6,$C7,$F7,$07,$FD,$C7
	.dc.b	$F0,$D9,$F8,$3F,$F5,$E7,$D8,$7F,$AE,$C9,$7F,$DD,$2D,$19,$F9,$AC
	.dc.b	$3F,$E7,$48,$FE,$F2,$D7,$FB,$EB,$5F,$F0,$37,$BD,$FE,$7E,$7F,$C0
	.dc.b	$2D,$73,$E9,$77,$8C,$47,$F8,$47,$EB,$FE,$1A,$69,$35,$F3,$7D,$83
	.dc.b	$CB,$A4,$AF,$F0,$6C,$65,$FF,$07,$FD,$7F,$F8,$24,$8F,$BD,$FF,$06
	.dc.b	$7D,$83,$E0,$7E,$13,$67,$FC,$2E,$D7,$FC,$36,$D7,$FC,$49,$F8,$07
	.dc.b	$FB,$94,$9B,$FC,$B7,$FD,$D1,$F5,$FE,$C2,$67,$F3,$DF,$7F,$D8,$BD
	.dc.b	$AF,$A4,$DC,$3F,$C4,$ED,$7F,$C5,$6D,$7F,$C6,$8F,$CA,$7F,$36,$D4
	.dc.b	$D3,$1E,$87,$F9,$9E,$3F,$C6,$6D,$7F,$C6,$ED,$7F,$C8,$0D,$7F,$C8
	.dc.b	$DF,$80,$7F,$52,$7F,$C7,$6D,$7F,$C7,$ED,$7F,$C9,$0D,$7F,$C9,$D6
	.dc.b	$FB,$0F,$84,$48,$FF,$22,$C6,$5F,$F2,$3F,$D9,$3E,$2F,$F2,$CD,$8C
	.dc.b	$C6,$08,$FE,$F3,$CA,$69,$FE,$5D,$AE,$13,$FE,$1E,$6C,$EA,$3F,$81
	.dc.b	$DF,$FA,$A9,$E5,$5E,$8F,$8D,$FF,$33,$56,$E5,$F6,$5F,$C6,$1D,$9F
	.dc.b	$C3,$1C,$3F,$E2,$CB,$FF,$51,$B6,$BF,$04,$47,$E0,$BF,$B3,$FC,$03
	.dc.b	$FD,$0F,$1F,$E3,$46,$DF,$85,$B8,$7F,$9E,$7E,$89,$F4,$DE,$68,$C3
	.dc.b	$FB,$D8,$CD,$FF,$9C,$9A,$FF,$9D,$9A,$FF,$9F,$DC,$DF,$BD,$B5,$D2
	.dc.b	$7D,$E2,$3F,$92,$B8,$B7,$C1,$78,$E3,$F9,$7C,$61,$D4,$65,$E5,$F3
	.dc.b	$76,$75,$1F,$06,$70,$6E,$BF,$D8,$B1,$E2,$5F,$67,$CF,$A1,$F2,$74
	.dc.b	$E1,$12,$BD,$33,$A6,$CE,$5F,$F3,$1D,$7F,$64,$7F,$DF,$BE,$67,$9F
	.dc.b	$E9,$8F,$8F,$D6,$93,$A2,$1C,$9C,$BF,$E6,$E9,$E2,$EB,$6F,$9D,$3A
	.dc.b	$7F,$1D,$3A,$9E,$15,$9D,$BF,$01,$7E,$38,$93,$F0,$E3,$4F,$F5,$EA
	.dc.b	$E5,$7D,$2B,$33,$F6,$6B,$8B,$D8,$E4,$43,$04,$9F,$3B,$97,$A3,$EB
	.dc.b	$BF,$2B,$97,$FD,$9D,$7E,$26,$DF,$8B,$0F,$9D,$5D,$67,$2F,$13,$FB
	.dc.b	$BC,$6A,$FB,$C6,$CC,$E1,$EC,$C3,$B4,$7C,$1C,$3B,$4F,$BA,$B8,$9E
	.dc.b	$A9,$FA,$C7,$A4,$E0,$78,$4F,$8C,$64,$57,$67,$27,$23,$C8,$6A,$BC
	.dc.b	$2F,$18,$E1,$7E,$BC,$3E,$3C,$F0,$9C,$E1,$EB,$BF,$45,$6B,$0F,$41
	.dc.b	$EA,$7E,$3D,$F8,$63,$31,$87,$51,$F1,$8F,$65,$FA,$4E,$EE,$2B,$BB
	.dc.b	$B3,$D8,$7F,$05,$BF,$E1,$84,$7F,$86,$35,$FE,$18,$D7,$81,$F2,$F0
	.dc.b	$F8,$0F,$DC,$F8,$AF,$8A,$FA,$71,$FE,$39,$18,$C7,$D7,$3B,$39,$5F
	.dc.b	$E2,$67,$55,$F1,$4E,$33,$A3,$C1,$1A,$7D,$95,$FC,$76,$31,$1F,$6A
	.dc.b	$30,$F1,$3C,$87,$84,$A6,$E7,$C0,$E6,$3C,$EE,$4E,$F1,$E4,$78,$DF
	.dc.b	$C2,$74,$7F,$2B,$AF,$1C,$64,$FE,$DD,$5D,$1C,$5F,$F2,$39,$DB,$6B
	.dc.b	$D1,$5B,$9C,$C7,$F2,$76,$BF,$C7,$9B,$1F,$02,$3C,$B3,$F3,$5E,$77
	.dc.b	$EE,$7C,$C6,$0F,$69,$F0,$5D,$5F,$9A,$3C,$06,$CF,$7B,$EC,$8F,$F3
	.dc.b	$9C,$3E,$C6,$AB,$FE,$E1,$18,$6C,$7F,$49,$87,$91,$62,$1F,$D5,$A3
	.dc.b	$85,$76,$77,$19,$F9,$9F,$EA,$5F,$80,$7F,$84,$63,$C1,$5E,$48,$C6
	.dc.b	$1E,$2A,$F0,$1C,$3A,$A5,$F3,$3C,$1F,$25,$E5,$23,$E3,$FB,$3F,$41
	.dc.b	$87,$B9,$FE,$94,$C9,$3A,$8E,$07,$42,$1C,$0F,$91,$3F,$0F,$E1,$FF
	.dc.b	$CF,$94,$69,$27,$E6,$9D,$DE,$0F,$C7,$7E,$67,$D7,$FC,$DF,$AE,$FE
	.dc.b	$ED,$F5,$DF,$EF,$AA,$FF,$00,$B5,$F1,$5A,$78,$47,$63,$EE,$3E,$42
	.dc.b	$39,$FB,$3F,$F5,$F3,$FE,$0C,$6B,$ED,$6E,$BF,$E3,$15,$F5,$DF,$E0
	.dc.b	$96,$CF,$1F,$6F,$06,$66,$47,$77,$FD,$41,$3C,$23,$E1,$B5,$FC,$8B
	.dc.b	$FE,$F3,$7C,$DF,$0B,$C0,$8E,$0F,$43,$E3,$3C,$9E,$F7,$F8,$77,$F0
	.dc.b	$07,$C8,$D2,$57,$BF,$ED,$3D,$47,$90,$8F,$C5,$7C,$FF,$0F,$23,$E4
	.dc.b	$35,$F4,$FB,$BF,$0E,$24,$7C,$B8,$F8,$C7,$C0,$23,$F4,$5F,$87,$EE
	.dc.b	$C8,$FA,$8D,$7E,$2B,$F3,$C8,$FF,$90,$BE,$CD,$E9,$7F,$8E,$FE,$CF
	.dc.b	$FC,$D5,$3F,$95,$6F,$0F,$F9,$A5,$BD,$9F,$C0,$2D,$F5,$7F,$E4,$36
	.dc.b	$DF,$E2,$5A,$FE,$61,$AF,$F0,$66,$BC,$FE,$A7,$FD,$B9,$06,$92,$3A
	.dc.b	$7C,$E7,$78,$91,$EB,$8F,$11,$1C,$1E,$FF,$9D,$FF,$2C,$7E,$0C,$9B
	.dc.b	$52,$BF,$CC,$B3,$97,$FC,$CF,$F8,$77,$E9,$5F,$F3,$4F,$DA,$3E,$6F
	.dc.b	$FC,$C4,$D8,$DF,$5F,$F9,$97,$ED,$7F,$CB,$F1,$FB,$F7,$D8,$FF,$9C
	.dc.b	$9A,$E7,$D1,$FF,$36,$7E,$0E,$F0,$6F,$85,$2B,$FC,$F3,$39,$7F,$CF
	.dc.b	$7F,$87,$7F,$BA,$5F,$E1,$99,$7E,$85,$BD,$DF,$E7,$79,$D7,$37,$FA
	.dc.b	$8F,$3F,$75,$79,$78,$BD,$07,$A7,$8B,$FC,$FD,$F8,$47,$FA,$62,$41
	.dc.b	$FB,$B7,$FE,$AD,$6F,$37,$F4,$47,$84,$BA,$FF,$56,$FC,$AE,$FF,$92
	.dc.b	$7F,$A2,$E3,$31,$F1,$FB,$BE,$57,$A5,$23,$E4,$47,$FA,$47,$F0,$8F
	.dc.b	$F4,$14,$7F,$2B,$5E,$3E,$4D,$7A,$5F,$E9,$7F,$C3,$BF,$90,$6B,$FE
	.dc.b	$C3,$E3,$FF,$5F,$B7,$E4,$F5,$F4,$1F,$67,$EB,$FE,$D7,$B1,$34,$FB
	.dc.b	$11,$FE,$A1,$ED,$FF,$01,$0F,$C1,$B0,$67,$D5,$FF,$53,$7E,$1C,$F9
	.dc.b	$AF,$C3,$FA,$73,$1F,$F5,$9C,$E5,$FC,$25,$FE,$26,$E4,$60,$FF,$5C
	.dc.b	$FD,$DF,$FB,$6F,$6F,$F6,$E7,$E1,$4F,$90,$FF,$5A,$36,$B7,$D7,$E2
	.dc.b	$F8,$5F,$CB,$FB,$B2,$7C,$31,$F9,$7F,$77,$F3,$1F,$7F,$27,$F2,$D1
	.dc.b	$FE,$BF,$CF,$ED,$47,$8B,$EA,$30,$67,$BB,$5E,$9D,$9F,$3C,$61,$FD
	.dc.b	$4F,$DF,$57,$FB,$15,$AB,$F5,$73,$DF,$E1,$65,$7C,$07,$C5,$EA,$C8
	.dc.b	$C1,$0E,$6F,$FB,$57,$EF,$9F,$6B,$49,$3F,$EE,$28,$FD,$AF,$1B,$87
	.dc.b	$FC,$37,$18,$6F,$9F,$CC,$FF,$70,$7D,$F3,$EF,$BF,$DC,$5F,$7C,$FA
	.dc.b	$AA,$4C,$BF,$DD,$F0,$FA,$1D,$92,$32,$F4,$F9,$8F,$F7,$67,$DD,$FF
	.dc.b	$BD,$27,$FC,$0C,$DD,$7C,$D1,$FB,$28,$C6,$9F,$E3,$27,$F6,$31,$9F
	.dc.b	$73,$99,$1F,$F8,$03,$1F,$B4,$27,$DD,$78,$0F,$F7,$E7,$DF,$7F,$8C
	.dc.b	$27,$B7,$0F,$90,$8F,$2E,$0F,$39,$1C,$1D,$BD,$DF,$F2,$12,$FB,$E6
	.dc.b	$FE,$79,$CC,$FF,$92,$32,$FE,$81,$AF,$F8,$45,$B1,$FE,$20,$6B,$BF
	.dc.b	$51,$F7,$DB,$6C,$6D,$AE,$E3,$FC,$38,$FF,$1A,$FA,$FA,$9A,$47,$FC
	.dc.b	$45,$F1,$DF,$43,$0F,$A3,$1F,$C8,$B9,$31,$C5,$FE,$60,$AB,$3F,$E6
	.dc.b	$49,$AE,$E7,$F9,$4D,$72,$F2,$DF,$1F,$AA,$FC,$55,$68,$FF,$36,$CE
	.dc.b	$9C,$47,$F9,$C3,$9B,$87,$DF,$71,$6C,$EA,$7F,$9B,$CF,$5F,$8F,$CC
	.dc.b	$7F,$94,$43,$99,$E1,$6F,$A4,$C5,$7F,$95,$E3,$13,$D7,$27,$3F,$B9
	.dc.b	$FD,$7C,$66,$31,$CD,$23,$AF,$85,$70,$E8,$FC,$6F,$BD,$FF,$41,$CF
	.dc.b	$A5,$F8,$73,$FE,$67,$72,$DD,$F8,$1F,$3B,$27,$81,$8B,$F5,$BD,$7C
	.dc.b	$7F,$B3,$7F,$E9,$8C,$11,$88,$FE,$BB,$18,$AF,$F3,$13,$63,$FC,$E3
	.dc.b	$F7,$DF,$D8,$A7,$FC,$CC,$D8,$C9,$CC,$39,$7F,$D4,$B2,$FA,$4D,$36
	.dc.b	$BB,$C6,$0E,$5F,$7D,$FE,$0F,$8F,$F3,$A3,$6B,$FD,$06,$D7,$FD,$62
	.dc.b	$D7,$FD,$6B,$F7,$DF,$C7,$67,$FC,$F8,$D8,$FF,$45,$B5,$FF,$5C,$B5
	.dc.b	$FF,$5E,$FD,$EB,$E0,$BD,$8D,$AF,$F4,$33,$67,$FD,$2A,$D7,$FD,$87
	.dc.b	$F6,$EF,$BC,$F8,$8D,$AF,$F4,$73,$67,$FD,$3A,$D7,$FD,$97,$F7,$8F
	.dc.b	$A5,$F9,$8D,$AF,$F4,$B3,$67,$FD,$4A,$D7,$FD,$A7,$F7,$9F,$DC,$67
	.dc.b	$FD,$38,$D8,$FF,$55,$B5,$DF,$B5,$F6,$7D,$CF,$95,$1F,$E2,$06,$A6
	.dc.b	$98,$FB,$C7,$BD,$F8,$16,$E1,$FE,$AB,$6D,$7F,$AE,$9A,$FF,$BA,$3E
	.dc.b	$F3,$F0,$CF,$FB,$B1,$AF,$F8,$C1,$A9,$BA,$3E,$9F,$E6,$92,$FE,$6B
	.dc.b	$F9,$3F,$BB,$FD,$5D,$FF,$BD,$9A,$FF,$BE,$9A,$FF,$8F,$9A,$FF,$25
	.dc.b	$F8,$1F,$CE,$78,$DF,$08,$8F,$5F,$BA,$98,$7F,$B2,$6D,$18,$BF,$F6
	.dc.b	$CF,$77,$F0,$66,$F7,$FE,$9E,$31,$1F,$E4,$E3,$FC,$9E,$E1,$C7,$DD
	.dc.b	$3D,$BF,$D7,$37,$1F,$F6,$C3,$67,$FD,$CE,$D7,$A0,$FB,$AF,$F5,$0C
	.dc.b	$FF,$B7,$DB,$1F,$EE,$E6,$BC,$1A,$F9,$7E,$E9,$F1,$DF,$BF,$6D,$7F
	.dc.b	$BA,$5B,$3F,$EF,$86,$BE,$2F,$BA,$7D,$0F,$F0,$76,$D7,$FB,$C5,$B3
	.dc.b	$FF,$00,$1A,$FA,$A9,$F0,$5E,$12,$57,$85,$FE,$25,$F7,$3F,$68,$9E
	.dc.b	$46,$0F,$E2,$DF,$71,$FC,$A2,$BF,$E0,$03,$70,$F8,$E1,$C3,$F2,$9A
	.dc.b	$FD,$7F,$B8,$FE,$91,$F4,$1C,$21,$18,$79,$BF,$D8,$3D,$3D,$39,$8F
	.dc.b	$AB,$F8,$03,$EA,$3F,$BB,$B6,$3A,$35,$F2,$B5,$FB,$BF,$70,$FB,$9F
	.dc.b	$EF,$ED,$8F,$23,$5F,$6B,$5F,$C2,$F9,$6F,$AD,$FF,$03,$B6,$3D,$8D
	.dc.b	$7E,$2B,$5F,$C6,$FD,$CF,$F4,$14,$D2,$6D,$FE,$12,$FB,$9F,$CD,$4F
	.dc.b	$C7,$6B,$F4,$1A,$FE,$61,$AF,$E9,$7E,$E7,$FC,$32,$47,$FC,$3A,$FF
	.dc.b	$87,$3E,$E7,$C9,$1F,$AB,$6B,$FE,$BA,$6A,$6C,$37,$E6,$49,$FB,$91
	.dc.b	$97,$EF,$7D,$A7,$BC,$98,$7F,$8B,$9F,$85,$F7,$1F,$D7,$E7,$F0,$8D
	.dc.b	$7F,$1A,$D7,$F6,$AD,$7F,$81,$3F,$E9,$B7,$7F,$27,$FE,$3A,$8C,$47
	.dc.b	$08,$F0,$7D,$1C,$8F,$2B,$5F,$CD,$FD,$C7,$CA,$7F,$DB,$6D,$4D,$31
	.dc.b	$7F,$AE,$8F,$7F,$4F,$B4,$8E,$0F,$F1,$84,$6D,$7B,$9F,$E4,$88,$C3
	.dc.b	$7E,$B7,$F8,$F3,$9F,$94,$F9,$04,$AF,$9A,$FC,$BF,$13,$7A,$FF,$25
	.dc.b	$C1,$38,$AF,$F7,$AC,$61,$F1,$D6,$54,$7F,$89,$9F,$3F,$B9,$98,$C4
	.dc.b	$7A,$2F,$B7,$C8,$E4,$FB,$84,$8F,$C0,$BF,$E6,$CF,$84,$FF,$49,$C0
	.dc.b	$FF,$2E,$FD,$B7,$86,$0D,$FB,$A9,$3F,$C1,$E7,$FD,$40,$FF,$36,$6B
	.dc.b	$FD,$0D,$AF,$57,$BB,$FE,$6A,$8E,$6F,$A4,$D7,$C7,$F8,$7E,$3A,$47
	.dc.b	$F9,$96,$BF,$D7,$CF,$3F,$8C,$F3,$1A,$79,$1F,$E7,$69,$CB,$C7,$E1
	.dc.b	$3D,$9F,$49,$23,$04,$38,$DA,$35,$18,$9F,$89,$E7,$7C,$5F,$9A,$93
	.dc.b	$E5,$8C,$BE,$78,$CB,$D2,$32,$F6,$8C,$F1,$FF,$A5,$B0,$FE,$16,$DA
	.dc.b	$CE,$FF,$2E,$B8,$CE,$6F,$B1,$0E,$C7,$8C,$C3,$D6,$38,$CF,$D5,$AE
	.dc.b	$B5,$F5,$67,$2F,$D7,$8C,$BF,$EA,$08,$CB,$FE,$A2,$8C,$DA,$72,$FA
	.dc.b	$AB,$D1,$5E,$27,$C0,$F9,$4F,$84,$64,$F0,$19,$BF,$43,$93,$C0,$66
	.dc.b	$3A,$BC,$8F,$82,$48,$E2,$E5,$F9,$0E,$63,$8C,$68,$C9,$1C,$92,$3C
	.dc.b	$A7,$13,$E4,$1E,$59,$F8,$0E,$A7,$A4,$F1,$8F,$49,$C4,$F9,$87,$1B
	.dc.b	$F9,$37,$F3,$1B,$19,$3E,$01,$E0,$3D,$93,$F3,$1E,$97,$F4,$DE,$06
	.dc.b	$2B,$A3,$F4,$4F,$DC,$57,$F0,$32,$3F,$BD,$6B,$FE,$0C,$9F,$E6,$04
	.dc.b	$8F,$BE,$7F,$07,$F9,$9B,$3E,$63,$FD,$CE,$E1,$F2,$B7,$0F,$9D,$1F
	.dc.b	$C1,$63,$18,$7A,$E7,$F4,$91,$89,$F9,$93,$FC,$36,$31,$1F,$32,$76
	.dc.b	$F8,$6F,$E9,$B9,$AE,$0E,$6F,$F2,$F7,$CA,$B6,$1C,$47,$B6,$78,$1D
	.dc.b	$63,$2F,$F1,$56,$BF,$C2,$9B,$7E,$48,$FA,$B3,$E0,$79,$9F,$10,$B1
	.dc.b	$F1,$1F,$92,$62,$FE,$21,$EA,$79,$9E,$33,$47,$C6,$3E,$61,$83,$E1
	.dc.b	$18,$38,$9E,$12,$C7,$86,$FE,$14,$7F,$45,$84,$FF,$9F,$FE,$CF,$FC
	.dc.b	$19,$3F,$CC,$1F,$F3,$BC,$EA,$3F,$A6,$46,$27,$87,$6F,$F5,$D8,$E3
	.dc.b	$5B,$A4,$6D,$E2,$6E,$78,$7C,$5C,$8F,$25,$F8,$E3,$A7,$C8,$E6,$7C
	.dc.b	$91,$70,$ED,$82,$BC,$5F,$47,$E3,$F2,$66,$1D,$0F,$4C,$69,$FF,$33
	.dc.b	$52,$3F,$CD,$11,$E8,$6C,$66,$37,$1C,$5F,$8C,$FC,$1F,$A3,$F8,$5F
	.dc.b	$55,$27,$C0,$7C,$27,$E3,$7D,$5C,$8C,$DA,$78,$47,$D1,$9F,$99,$F5
	.dc.b	$92,$36,$FC,$FF,$87,$91,$82,$1F,$65,$FF,$08,$7D,$93,$EC,$EF,$B1
	.dc.b	$2B,$CB,$39,$6C,$FC,$28,$C4,$7F,$A4,$9A,$FF,$AE,$9A,$FF,$B1,$DD
	.dc.b	$EF,$F8,$16,$F0,$E6,$BC,$7C,$7E,$4F,$1A,$61,$E2,$7C,$FE,$CC,$8F
	.dc.b	$F4,$D6,$1F,$EB,$E6,$C7,$94,$F3,$78,$B2,$36,$D7,$D1,$F8,$07,$C2
	.dc.b	$89,$1F,$02,$3E,$2F,$D0,$FC,$6F,$1A,$4F,$C4,$8C,$BE,$27,$E2,$11
	.dc.b	$C1,$B2,$1F,$E1,$8C,$1F,$07,$F4,$BC,$59,$3F,$41,$FF,$22,$7D,$93
	.dc.b	$D8,$A7,$E4,$9C,$4F,$FA,$ED,$AF,$FB,$11,$AF,$FB,$B1,$78,$BC,$49
	.dc.b	$1C,$5E,$9F,$2B,$E0,$64,$7E,$5A,$53,$E4,$F9,$AF,$2F,$8F,$2B,$51
	.dc.b	$97,$C8,$F9,$C8,$E0,$D9,$0F,$F1,$DD,$3E,$47,$FC,$C9,$F6,$4A,$FD
	.dc.b	$8B,$6B,$FD,$B4,$D7,$FD,$C6,$D7,$FC,$BC,$7F,$97,$FE,$C9,$D7,$FB
	.dc.b	$A6,$C7,$FB,$8D,$AF,$FB,$B1,$AF,$21,$F6,$0F,$2F,$F8,$1B,$63,$FD
	.dc.b	$D8,$D7,$FD,$EA,$D7,$97,$D9,$3B,$FF,$85,$B6,$3F,$DE,$8D,$7F,$DF
	.dc.b	$AD,$7C,$5F,$64,$FA,$3F,$E2,$6D,$8F,$F7,$E3,$5F,$40,$70,$FA,$EF
	.dc.b	$E3,$3F,$19,$AF,$FA,$15,$AF,$C0,$D3,$F9,$0F,$D9,$7E,$0A,$3A,$87
	.dc.b	$0F,$81,$AF,$C9,$6B,$F5,$3A,$1F,$65,$FC,$B1,$B1,$D1,$AF,$A9,$AF
	.dc.b	$A6,$3E,$89,$1F,$A2,$FF,$36,$FB,$27,$8F,$49,$3E,$88,$CB,$EA,$FB
	.dc.b	$27,$CA,$FF,$3E,$6C,$66,$30,$7E,$03,$D2,$91,$F8,$06,$BF,$43,$E4
	.dc.b	$7F,$C6,$8D,$F4,$FF,$A4,$C6,$23,$F1,$8D,$7F,$2F,$F5,$EF,$98,$FE
	.dc.b	$9E,$D8,$F9,$ED,$7E,$EB,$5F,$B9,$F5,$FF,$94,$23,$F9,$47,$FA,$CF
	.dc.b	$D7,$7F,$71,$8F,$CE,$35,$FD,$3B,$5F,$F6,$33,$54,$BE,$8F,$CA,$48
	.dc.b	$F2,$BF,$A4,$F5,$E4,$7E,$91,$AF,$E5,$29,$C3,$E4,$24,$7E,$32,$32
	.dc.b	$DF,$1B,$D2,$47,$07,$E4,$BE,$C9,$F5,$FF,$DD,$9B,$19,$8C,$1F,$0B
	.dc.b	$DF,$FF,$78,$5F,$FB,$B3,$D1,$FF,$0A,$37,$D1,$F8,$7E,$4E,$8F,$92
	.dc.b	$4C,$3C,$CF,$F8,$13,$EB,$DF,$2F,$FE,$FB,$6C,$FF,$7E,$6B,$FE,$D8
	.dc.b	$7F,$DB,$A9,$A4,$F0,$BF,$8C,$E3,$FE,$27,$FA,$F7,$D7,$E3,$2F,$E3
	.dc.b	$BE,$E3,$E5,$31,$97,$FB,$F6,$30,$FF,$83,$9A,$FF,$BA,$17,$D5,$F0
	.dc.b	$8F,$F1,$33,$63,$31,$C6,$35,$87,$F0,$DC,$7F,$C7,$5F,$5B,$FE,$FE
	.dc.b	$27,$C2,$73,$7F,$A7,$DF,$F8,$79,$AF,$FB,$E1,$B3,$CB,$D1,$E1,$E8
	.dc.b	$CA,$E0,$D7,$FD,$FF,$F5,$FF,$E4,$75,$7F,$32,$FC,$7C,$3E,$13,$C5
	.dc.b	$C1,$2B,$C4,$D7,$C5,$E9,$29,$0F,$AF,$FF,$24,$38,$C1,$5B,$8F,$E5
	.dc.b	$B1,$8A,$FE,$46,$D8,$CC,$ED,$FE,$95,$1F,$D6,$FE,$BF,$7E,$51,$A4
	.dc.b	$9F,$F2,$7F,$D0,$FF,$94,$13,$48,$FF,$32,$F8,$6F,$F3,$96,$3F,$BD
	.dc.b	$E4,$BC,$BF,$D6,$66,$D7,$F6,$49,$AF,$C0,$6F,$8C,$F7,$B3,$3F,$8D
	.dc.b	$BF,$F2,$BC,$E5,$FF,$2C,$C7,$E2,$9B,$F5,$FC,$39,$F8,$B2,$37,$F0
	.dc.b	$F6,$BE,$84,$7F,$6D,$8E,$E7,$8B,$EB,$EC,$78,$1F,$3E,$9F,$E1,$28
	.dc.b	$FF,$34,$BF,$B0,$6F,$36,$B9,$64,$7A,$3C,$3B,$8F,$80,$FB,$FA,$7F
	.dc.b	$12,$9F,$F3,$9F,$D7,$38,$3E,$2B,$F0,$6B,$80,$75,$3C,$38,$36,$7C
	.dc.b	$AF,$06,$C7,$C8,$79,$F8,$DF,$43,$EB,$FF,$1F,$3E,$08,$C6,$12,$3F
	.dc.b	$D0,$B2,$F8,$6C,$4F,$8A,$31,$3E,$8F,$98,$F3,$FC,$CF,$A1,$B1,$FE
	.dc.b	$8E,$FA,$F7,$9F,$F8,$0F,$C0,$1F,$1D,$F8,$29,$CD,$7F,$A5,$27,$2F
	.dc.b	$FA,$5B,$EC,$3F,$D8,$6D,$F8,$CF,$C4,$C6,$34,$F8,$AA,$3F,$1B,$F6
	.dc.b	$1F,$E3,$BA,$FF,$4F,$BF,$E0,$06,$FC,$26,$BD,$BF,$C9,$FD,$53,$EC
	.dc.b	$3F,$2B,$F8,$07,$FB,$D6,$FF,$D4,$F3,$97,$FD,$53,$19,$CF,$EA,$FD
	.dc.b	$8F,$C4,$7E,$BF,$D8,$79,$7B,$FF,$DA,$13,$97,$FD,$A5,$19,$BF,$F2
	.dc.b	$3F,$D8,$FF,$71,$7F,$D6,$AB,$93,$F4,$EE,$70,$FC,$1B,$99,$FF,$5D
	.dc.b	$D7,$CA,$9F,$C3,$FC,$DB,$E6,$7F,$5F,$39,$C7,$FD,$C5,$39,$BF,$F6
	.dc.b	$1F,$D8,$FF,$84,$9B,$F1,$5F,$B6,$8C,$63,$F9,$27,$FC,$95,$3F,$EC
	.dc.b	$AF,$B1,$FC,$5C,$FE,$59,$FC,$AB,$4F,$DF,$B8,$C3,$FD,$E5,$BF,$FA
	.dc.b	$4A,$BF,$DA,$7F,$80,$3E,$77,$E9,$FC,$9F,$F0,$68,$C4,$FF,$B6,$3E
	.dc.b	$CB,$F4,$98,$7F,$BE,$D7,$99,$FC,$31,$CC,$FF,$9C,$9C,$BF,$EE,$0F
	.dc.b	$3F,$E1,$3F,$E7,$65,$F3,$3F,$C5,$1C,$EB,$FC,$D1,$CC,$F8,$7E,$CB
	.dc.b	$F8,$CF,$E0,$0F,$9B,$F0,$46,$1F,$E3,$BE,$6F,$F0,$5C,$3F,$72,$FF
	.dc.b	$3A,$BF,$F7,$8B,$9D,$11,$FE,$55,$F5,$FF,$92,$FD,$9B,$E1,$EF,$D0
	.dc.b	$E1,$B5,$FE,$FA,$FB,$3F,$F5,$B3,$FE,$FC,$F8,$C7,$D1,$7F,$2E,$9C
	.dc.b	$E9,$FC,$2A,$73,$5F,$D4,$33,$F0,$CF,$F5,$2F,$B2,$F6,$4E,$4F,$F4
	.dc.b	$DA,$FB,$5F,$E7,$2E,$A7,$F8,$AB,$99,$E1,$1F,$D7,$3E,$D3,$F6,$91
	.dc.b	$88,$47,$FC,$01,$8C,$3A,$8F,$E3,$EE,$5E,$BF,$4F,$FD,$23,$F0,$0F
	.dc.b	$E6,$51,$8A,$E9,$18,$C7,$D7,$2F,$B5,$FC,$1B,$FD,$B9,$FC,$1B,$FC
	.dc.b	$B1,$CB,$BE,$9F,$EA,$BF,$68,$F7,$BF,$DB,$73,$FD,$DA,$31,$F4,$7F
	.dc.b	$E9,$27,$FA,$C7,$B9,$6B,$74,$FB,$4F,$A6,$FF,$A6,$1A,$FF,$A9,$9A
	.dc.b	$FF,$AF,$DA,$FF,$B2,$BF,$01,$FC,$73,$FE,$A0,$6B,$FE,$AE,$6B,$FE
	.dc.b	$C7,$6B,$FE,$D2,$FB,$4F,$D6,$3F,$EA,$86,$BF,$EB,$66,$BF,$EC,$F6
	.dc.b	$BF,$ED,$AF,$33,$FD,$A6,$DF,$A4,$FF,$58,$BF,$C5,$5B,$1A,$8F,$F0
	.dc.b	$DF,$DA,$7F,$33,$7F,$D6,$ED,$7F,$D8,$AD,$7F,$DB,$CD,$7F,$DD,$3F
	.dc.b	$68,$EB,$EE,$D8,$FF,$60,$B5,$FF,$67,$B5,$FF,$74,$7D,$9F,$F6,$B7
	.dc.b	$FD,$8E,$D7,$FD,$AA,$D7,$7A,$FB,$3B,$1F,$5B,$49,$3F,$ED,$08,$CB
	.dc.b	$7C,$1E,$04,$70,$7F,$B5,$BF,$01,$7D,$B7,$C8,$6C,$66,$30,$7C,$CF
	.dc.b	$B5,$7C,$0D,$24,$FF,$B8,$23,$2F,$FB,$8B,$F0,$BF,$F5,$31,$3B,$3E
	.dc.b	$AB,$FE,$A9,$F3,$7F,$85,$4F,$FB,$A1,$AF,$FB,$D9,$AF,$F9,$48,$FF
	.dc.b	$29,$B8,$71,$F6,$8F,$B2,$FC,$0B,$63,$FD,$E6,$D7,$FE,$00,$B5,$E8
	.dc.b	$25,$F5,$9F,$8A,$6C,$79,$63,$04,$7F,$84,$47,$FA,$07,$0F,$33,$63
	.dc.b	$FC,$D3,$F8,$17,$F8,$31,$FE,$1C,$FF,$BA,$9F,$08,$70,$F9,$1A,$FC
	.dc.b	$16,$BF,$E7,$4F,$C0,$BF,$C3,$2F,$4F,$A8,$FF,$4C,$4F,$5F,$0B,$F8
	.dc.b	$7D,$6A,$BF,$C6,$EF,$F2,$77,$F9,$03,$5C,$F8,$FF,$EA,$18,$78,$B4
	.dc.b	$FD,$6B,$89,$F4,$87,$0F,$B1,$AF,$CE,$FC,$0B,$FD,$29,$9F,$DB,$F9
	.dc.b	$5F,$F4,$FB,$7E,$73,$FD,$61,$3A,$AF,$F4,$8B,$FE,$C6,$79,$7C,$9F
	.dc.b	$EF,$A3,$11,$F8,$56,$BC,$23,$99,$1F,$E7,$D1,$FE,$11,$6F,$D4,$78
	.dc.b	$08,$EB,$EA,$7C,$66,$1F,$F2,$FC,$A7,$31,$FD,$27,$99,$F2,$3F,$EC
	.dc.b	$97,$E9,$65,$F9,$96,$CF,$FA,$B1,$DB,$C6,$78,$70,$C8,$C1,$FD,$57
	.dc.b	$4F,$E1,$9C,$1F,$D2,$9F,$D3,$B7,$AF,$FD,$B1,$1A,$C7,$E7,$78,$18
	.dc.b	$99,$7C,$E3,$66,$1F,$F3,$DD,$74,$AE,$31,$B3,$C5,$5B,$BF,$15,$F2
	.dc.b	$9C,$90,$E6,$E5,$EB,$3C,$6B,$AC,$68,$E1,$5E,$B7,$C0,$64,$E2,$6A
	.dc.b	$38,$C7,$8A,$30,$7F,$B6,$2F,$D4,$DB,$F1,$4F,$58,$E6,$75,$79,$38
	.dc.b	$74,$60,$95,$F9,$9B,$F5,$4F,$53,$CF,$3C,$27,$51,$87,$CE,$61,$EC
	.dc.b	$E1,$E4,$60,$F5,$4F,$A2,$35,$1F,$ED,$38,$F0,$BF,$ED,$5A,$D3,$C0
	.dc.b	$F8,$66,$4F,$11,$93,$B1,$93,$E1,$4F,$96,$BD,$51,$E4,$7E,$0B,$C5
	.dc.b	$CB,$C5,$CB,$E5,$3C,$37,$F9,$37,$94,$F5,$23,$82,$1C,$4F,$2C,$71
	.dc.b	$7B,$47,$07,$53,$E0,$9F,$ED,$D1,$F8,$C3,$F9,$EB,$87,$E6,$98,$AF
	.dc.b	$F0,$3C,$66,$3E,$11,$93,$F8,$EB,$E1,$24,$7F,$53,$9F,$F1,$DF,$D0
	.dc.b	$62,$BE,$69,$E6,$01,$D3,$F3,$8F,$7B,$88,$F5,$4F,$5B,$FE,$C1,$39
	.dc.b	$7F,$03,$7F,$0E,$31,$7F,$22,$31,$5F,$82,$8F,$11,$27,$6E,$27,$6F
	.dc.b	$A1,$CC,$69,$B3,$CD,$B3,$B8,$ED,$3B,$69,$B2,$C6,$CF,$8B,$1B,$6C
	.dc.b	$6D,$A6,$CF,$8C,$7C,$97,$4E,$23,$93,$63,$37,$F9,$29,$CB,$63,$37
	.dc.b	$FA,$A9,$CB,$63,$E8,$57,$D8,$F9,$FF,$F0,$DA,$BF,$F3,$CF,$DB,$FF
	.dc.b	$8D,$DB,$E0,$F0,$95,$70,$9F,$F3,$FB,$70,$FF,$0D,$5F,$16,$DF,$19
	.dc.b	$D5,$FE,$22,$9F,$45,$D5,$B1,$FE,$97,$6B,$E5,$BF,$F1,$37,$E0,$3F
	.dc.b	$DC,$B1,$F3,$DF,$12,$57,$FA,$7E,$75,$5E,$4F,$AE,$ED,$EB,$CA,$ED
	.dc.b	$1E,$13,$FC,$DB,$89,$A4,$8F,$F0,$1C,$65,$E2,$43,$CC,$7E,$93,$EA
	.dc.b	$5F,$11,$F2,$7F,$01,$FF,$19,$E9,$F3,$BE,$23,$E0,$19,$B8,$7C,$17
	.dc.b	$2F,$FA,$E6,$78,$E7,$F3,$BE,$67,$E4,$12,$75,$F7,$35,$F1,$8C,$D3
	.dc.b	$9C,$7F,$D6,$F8,$2F,$EF,$7D,$C3,$96,$F9,$93,$2F,$F6,$54,$E5,$B3
	.dc.b	$FE,$94,$8C,$47,$F9,$01,$AF,$F9,$49,$AF,$F9,$7F,$EB,$7F,$CB,$18
	.dc.b	$FF,$B2,$27,$8B,$6B,$F2,$13,$AD,$38,$7C,$EF,$9B,$CE,$98,$FF,$B7
	.dc.b	$69,$3C,$0F,$F5,$3B,$88,$FF,$1D,$B5,$FF,$2A,$B5,$F3,$7D,$CF,$AC
	.dc.b	$8F,$F9,$0B,$24,$62,$3E,$6C,$61,$F8,$B1,$FA,$7F,$51,$F4,$9F,$33
	.dc.b	$EB,$E5,$7C,$A7,$E7,$F9,$72,$3B,$3F,$2C,$8F,$F9,$4A,$9F,$05,$B8
	.dc.b	$F0,$CF,$78,$FD,$B7,$E0,$1F,$EB,$FE,$0F,$F7,$9F,$DD,$3D,$FF,$DD
	.dc.b	$37,$0F,$E7,$33,$FE,$F8,$AE,$1C,$3F,$E9,$37,$FD,$89,$3F,$CF,$9A
	.dc.b	$9A,$62,$9F,$4D,$FE,$D6,$6C,$FF,$9A,$9A,$FF,$A1,$1A,$FF,$A5,$9C
	.dc.b	$DF,$A2,$7C,$E7,$F1,$1F,$B6,$FF,$34,$52,$9F,$55,$EC,$24,$7B,$31
	.dc.b	$FF,$56,$4B,$E7,$73,$3F,$A4,$8E,$2E,$CF,$EB,$6E,$1D,$47,$C8,$3D
	.dc.b	$84,$7D,$8F,$FB,$D3,$EE,$B8,$64,$BF,$F5,$E4,$61,$B8,$7F,$2B,$8F
	.dc.b	$25,$F8,$47,$DD,$38,$3F,$98,$36,$34,$D7,$F7,$BC,$BF,$91,$7F,$7F
	.dc.b	$A3,$EB,$FE,$41,$23,$B4,$3E,$C7,$F9,$E7,$DD,$3C,$1A,$4A,$F2,$47
	.dc.b	$E4,$1F,$37,$DD,$7F,$B0,$53,$4C,$62,$FA,$2F,$B1,$B4,$AF,$0B,$F6
	.dc.b	$FD,$BF,$E9,$9F,$75,$F8,$0E,$1F,$F7,$13,$6B,$FC,$1A,$D4,$FE,$0B
	.dc.b	$E5,$1F,$49,$FD,$55,$B3,$F2,$DA,$FC,$6F,$A0,$FC,$54,$C8,$F8,$6F
	.dc.b	$F5,$DF,$DC,$3F,$96,$C7,$F8,$73,$EE,$FF,$9F,$3F,$CA,$9B,$CF,$F7
	.dc.b	$63,$5E,$6F,$F6,$59,$FF,$4F,$C7,$F3,$0F,$BB,$FF,$51,$A6,$98,$E0
	.dc.b	$67,$F9,$DF,$B2,$95,$F5,$9F,$27,$A9,$FD,$C7,$EB,$DE,$F5,$25,$7E
	.dc.b	$26,$3F,$D0,$6B,$EE,$FE,$C3,$FD,$09,$BA,$FA,$63,$5E,$27,$F2,$EF
	.dc.b	$1B,$F5,$DE,$F4,$9F,$C3,$BF,$17,$EC,$BF,$C0,$5F,$77,$C7,$0E,$4D
	.dc.b	$E7,$FF,$21,$DA,$30,$FF,$91,$9F,$E1,$AF,$F9,$23,$2F,$8F,$5E,$36
	.dc.b	$D7,$26,$CF,$F9,$3B,$81,$E4,$FF,$03,$F7,$64,$FE,$71,$FB,$5E,$5F
	.dc.b	$9B,$8E,$4F,$EB,$73,$F9,$ED,$DB,$D6,$47,$07,$F4,$BF,$5F,$FA,$89
	.dc.b	$1C,$1F,$B7,$E8,$7C,$BF,$C4,$FB,$BF,$8B,$E8,$CA,$F0,$46,$5A,$43
	.dc.b	$FB,$AD,$3E,$DB,$FC,$5D,$F7,$7F,$C7,$89,$F0,$D8,$F2,$A5,$FF,$07
	.dc.b	$9F,$DF,$36,$3F,$C6,$DF,$05,$FB,$D8,$FE,$19,$F0,$0E,$8F,$F1,$D3
	.dc.b	$63,$31,$8E,$29,$1C,$1A,$FF,$82,$AB,$F8,$07,$DC,$BE,$EF,$FC,$88
	.dc.b	$D9,$FE,$34,$D7,$FC,$22,$BE,$27,$64,$8D,$BC,$BD,$59,$1C,$9A,$FF
	.dc.b	$86,$BE,$13,$F6,$AA,$BF,$A1,$93,$6F,$1E,$B2,$7C,$6D,$7C,$7F,$47
	.dc.b	$FE,$5F,$5F,$77,$FE,$4E,$B4,$E6,$7F,$36,$77,$70,$FE,$7A,$37,$DF
	.dc.b	$A8,$FF,$51,$C6,$1C,$3F,$D5,$BE,$EF,$51,$E9,$F3,$91,$89,$D4,$F0
	.dc.b	$9C,$BC,$07,$D4,$56,$99,$70,$6F,$C7,$FE,$2D,$87,$EF,$E7,$F0,$5E
	.dc.b	$1F,$F4,$F5,$3A,$8F,$7F,$25,$9F,$DD,$B5,$79,$1F,$E4,$59,$E5,$39
	.dc.b	$8C,$4F,$F9,$40,$FF,$10,$CF,$7E,$CE,$51,$83,$4E,$0C,$F4,$D9,$F3
	.dc.b	$38,$7F,$B3,$53,$E7,$F1,$7F,$E5,$E8,$C4,$FF,$6B,$CB,$FC,$26,$DE
	.dc.b	$2F,$54,$62,$BF,$CC,$AA,$32,$7A,$FC,$1B,$5E,$07,$E1,$7D,$DF,$C8
	.dc.b	$8F,$EF,$0F,$CC,$BF,$F3,$7B,$96,$E9,$C0,$F8,$DD,$F6,$BD,$6F,$CD
	.dc.b	$FB,$BE,$15,$F9,$06,$BF,$CB,$1A,$FF,$BC,$1A,$9F,$13,$C6,$BF,$00
	.dc.b	$ED,$AE,$DB,$E9,$FC,$38,$FC,$FB,$5F,$E7,$6D,$7F,$DF,$ED,$7F,$BF
	.dc.b	$65,$C8,$98,$7E,$7A,$7E,$63,$79,$FB,$91,$C1,$FA,$4F,$C2,$FF,$13
	.dc.b	$7F,$B0,$6B,$FD,$31,$A8,$38,$6E,$FE,$E8,$FC,$1C,$7E,$D5,$AF,$58
	.dc.b	$D7,$99,$2B,$C4,$D9,$FD,$E7,$E1,$6F,$BF,$F9,$2D,$AF,$DD,$B6,$77
	.dc.b	$3F,$3B,$F2,$0F,$F2,$15,$7F,$00,$6C,$FF,$07,$6B,$FC,$39,$AF,$F6
	.dc.b	$A5,$F7,$7C,$C9,$3C,$DF,$B7,$ED,$FF,$B5,$E3,$F8,$6B,$6B,$F8,$A3
	.dc.b	$5F,$E3,$6D,$7F,$BB,$AF,$93,$E6,$49,$E6,$F6,$F9,$39,$1D,$9B,$5F
	.dc.b	$C8,$7F,$0B,$7C,$A7,$E3,$9B,$5F,$C8,$1B,$3B,$9D,$7C,$04,$AD,$36
	.dc.b	$7A,$7B,$DC,$C9,$3C,$DF,$CD,$FE,$16,$F3,$E9,$27,$F9,$7C,$65,$FE
	.dc.b	$63,$F5,$EA,$FD,$1B,$63,$31,$83,$F4,$FE,$0F,$C7,$C3,$F2,$0D,$7F
	.dc.b	$94,$FB,$BA,$3F,$5B,$F8,$BB,$EE,$A9,$30,$FE,$83,$19,$7F,$A1,$FE
	.dc.b	$29,$FE,$7F,$6F,$C4,$7E,$DA,$31,$87,$F4,$D6,$BF,$D5,$3E,$B7,$F9
	.dc.b	$17,$C7,$7F,$4C,$6C,$7F,$53,$7F,$DF,$59,$E7,$AF,$FD,$81,$3F,$D7
	.dc.b	$3D,$9F,$EE,$8F,$FA,$45,$FF,$37,$13,$D2,$FF,$1D,$F8,$FF,$E4,$78
	.dc.b	$DC,$FF,$5F,$6B,$CE,$7F,$AB,$46,$23,$FC,$86,$F0,$9E,$53,$D3,$4C
	.dc.b	$B6,$BF,$79,$F5,$8F,$C2,$31,$0F,$AD,$FF,$06,$61,$FE,$59,$6E,$1F
	.dc.b	$E5,$06,$CF,$F0,$57,$C1,$19,$F5,$B9,$C6,$A7,$A4,$63,$1E,$FF,$58
	.dc.b	$F9,$7B,$F2,$F2,$FF,$D6,$D3,$96,$CE,$56,$5F,$E6,$D9,$57,$4A,$F5
	.dc.b	$FD,$5F,$F8,$C6,$F7,$9F,$F9,$A2,$1E,$E6,$8D,$F9,$92,$BF,$C1,$33
	.dc.b	$C5,$FF,$0A,$B5,$E5,$1F,$E0,$78,$ED,$97,$F9,$FD,$B7,$AA,$FF,$68
	.dc.b	$7C,$17,$CF,$7F,$DA,$5F,$56,$FB,$9F,$A2,$D8,$CC,$63,$7C,$B2,$32
	.dc.b	$72,$8F,$B3,$F5,$5F,$E8,$08,$3F,$25,$F8,$27,$F3,$B7,$19,$7F,$A5
	.dc.b	$5E,$04,$D7,$99,$D3,$C8,$FC,$2F,$E1,$8F,$A4,$FC,$3F,$3F,$F7,$BC
	.dc.b	$7F,$39,$E9,$6A,$64,$7F,$8C,$AB,$83,$63,$FC,$5D,$18,$8F,$F1,$E3
	.dc.b	$72,$FF,$54,$37,$7F,$F4,$FB,$67,$F3,$3F,$51,$F8,$EF,$82,$91,$FE
	.dc.b	$FD,$E8,$FF,$6D,$1E,$57,$E0,$9F,$9D,$8D,$B8,$8D,$E5,$F9,$4E,$BF
	.dc.b	$CA,$FA,$32,$BF,$27,$19,$7F,$2B,$19,$7A,$F4,$FF,$89,$DF,$ED,$1A
	.dc.b	$7F,$B0,$7F,$1D,$FF,$51,$C7,$8B,$9F,$FD,$36,$EB,$6F,$CA,$7B,$9F
	.dc.b	$B2,$AD,$DF,$1F,$6B,$FC,$A9,$1C,$AF,$FC,$CE,$D7,$FD,$86,$D7,$FB
	.dc.b	$C3,$CA,$F3,$3D,$4F,$DF,$F4,$B8,$3F,$23,$DB,$30,$FC,$AB,$F9,$4E
	.dc.b	$1C,$47,$CA,$EA,$FD,$FB,$FE,$05,$6A,$51,$F9,$4E,$DF,$CA,$F0,$CA
	.dc.b	$F8,$07,$E4,$E3,$4E,$0F,$CC,$C6,$DF,$B3,$D3,$FB,$E7,$5F,$8A,$3E
	.dc.b	$EB,$F8,$CA,$E4,$A6,$22,$4F,$08,$CB,$CA,$32,$FE,$1A,$79,$3B,$34
	.dc.b	$43,$53,$C9,$E1,$3C,$B8,$7F,$E0,$05,$72,$9D,$9C,$63,$33,$C5,$DD
	.dc.b	$71,$E0,$FF,$7E,$DF,$B6,$FC,$47,$23,$B9,$CE,$37,$3C,$DE,$37,$CE
	.dc.b	$7F,$C7,$55,$E3,$BE,$F1,$C4,$CC,$F1,$77,$7C,$55,$78,$DF,$15,$72
	.dc.b	$7C,$64,$70,$4B,$F1,$C7,$73,$9D,$77,$7A,$DF,$7B,$F3,$57,$B2,$78
	.dc.b	$47,$86,$FA,$47,$98,$CD,$73,$9E,$31,$EC,$38,$DE,$67,$B3,$63,$27
	.dc.b	$CE,$AF,$AA,$E0,$F7,$DF,$A9,$F7,$3E,$57,$27,$73,$27,$94,$C9,$EE
	.dc.b	$3A,$47,$A9,$C1,$C0,$F8,$A7,$8E,$3C,$11,$E4,$8F,$F7,$15,$FD,$37
	.dc.b	$FD,$D2,$D7,$CC,$62,$FF,$DB,$91,$98,$F3,$19,$3F,$DC,$EF,$CD,$AE
	.dc.b	$E4,$8F,$C4,$FD,$0F,$D5,$AF,$AC,$D8,$FA,$AD,$7D,$54,$8C,$5F,$A4
	.dc.b	$FD,$2B,$8B,$ED,$5F,$8E,$8C,$3F,$2E,$76,$FD,$E3,$B9,$4E,$F1,$C1
	.dc.b	$F9,$71,$C1,$F1,$DF,$A1,$E1,$1E,$3A,$FD,$7D,$71,$BF,$85,$5F,$00
	.dc.b	$DC,$F9,$A3,$D9,$18,$3C,$D1,$C9,$C1,$E2,$70,$F8,$8F,$49,$4F,$49
	.dc.b	$C2,$3C,$2F,$C3,$31,$7F,$C7,$AB,$CC,$D8,$C9,$FC,$92,$32,$D3,$27
	.dc.b	$F1,$D7,$FC,$6B,$EC,$C8,$FE,$D0,$FF,$8E,$3E,$13,$11,$FD,$96,$2E
	.dc.b	$3F,$E4,$39,$FF,$1C,$DF,$F8,$EA,$E5,$FF,$8E,$23,$2F,$F8,$EA,$32
	.dc.b	$FF,$66,$8F,$E6,$FF,$41,$FC,$E3,$EB,$A4,$FF,$3C,$8C,$BF,$94,$7F
	.dc.b	$32,$47,$06,$C8,$7F,$39,$F7,$7F,$C5,$7D,$72,$3F,$C5,$EF,$F8,$BB
	.dc.b	$3C,$47,$F5,$B8,$B9,$B5,$FF,$70,$4F,$F5,$BA,$FF,$16,$ED,$FE,$2E
	.dc.b	$CE,$4F,$F5,$7D,$FF,$87,$BF,$E3,$09,$CB,$FD,$62,$78,$BF,$D6,$3E
	.dc.b	$37,$FA,$BF,$C7,$C8,$C1,$0F,$DE,$C7,$EE,$08,$FE,$D9,$FE,$BE,$47
	.dc.b	$FB,$1F,$C0,$FF,$9E,$BE,$B5,$F4,$97,$DE,$95,$FD,$B6,$72,$D9,$CB
	.dc.b	$FE,$26,$84,$F3,$9F,$F4,$2F,$D6,$38,$E9,$2F,$FB,$C4,$38,$F1,$1F
	.dc.b	$E3,$08,$E0,$EA,$79,$CF,$FA,$43,$E8,$7F,$D0,$C7,$F9,$86,$7F,$C6
	.dc.b	$ED,$E2,$FA,$90,$11,$C3,$CA,$FA,$F1,$21,$18,$8F,$B7,$18,$70,$72
	.dc.b	$8A,$D3,$FC,$25,$1D,$5B,$DB,$FE,$A6,$FA,$C7,$81,$7D,$E9,$5F,$E1
	.dc.b	$99,$CB,$67,$2F,$F9,$4A,$3F,$DF,$2E,$4F,$F1,$07,$93,$F5,$1F,$F2
	.dc.b	$BB,$72,$F8,$31,$FE,$71,$DB,$FD,$6F,$3F,$5A,$38,$C6,$08,$F4,$DF
	.dc.b	$E7,$7C,$5F,$CF,$ED,$93,$EE,$8C,$BF,$13,$54,$7C,$46,$C6,$63,$07
	.dc.b	$07,$EB,$BF,$6F,$E8,$FF,$D9,$7F,$46,$FB,$2F,$9A,$DA,$FA,$2D,$7E
	.dc.b	$B3,$5F,$B5,$EE,$78,$48,$F9,$5F,$AF,$F4,$7F,$EB,$A8,$FB,$6D,$7F
	.dc.b	$00,$D7,$F0,$AD,$7F,$7B,$F2,$EF,$96,$FC,$0B,$63,$F0,$AD,$7F,$14
	.dc.b	$D7,$F1,$EB,$EE,$78,$92,$38,$BD,$3F,$0F,$91,$D1,$AF,$4F,$C2,$3C
	.dc.b	$24,$7C,$2F,$93,$EB,$E4,$79,$1A,$F9,$3E,$13,$D2,$47,$D2,$FB,$3E
	.dc.b	$32,$47,$B1,$AF,$B3,$E3,$3E,$09,$1F,$82,$FC,$4E,$59,$1F,$11,$AF
	.dc.b	$C4,$E5,$F9,$24,$7E,$4B,$F3,$3E,$82,$47,$CC,$6B,$F3,$38,$BE,$89
	.dc.b	$1F,$A2,$FD,$4E,$79,$1F,$51,$AF,$D4,$F8,$6F,$B2,$47,$EC,$BF,$73
	.dc.b	$EB,$64,$7D,$C6,$BF,$73,$E4,$3F,$02,$47,$F0,$2F,$F0,$4F,$8E,$FD
	.dc.b	$8C,$7F,$0E,$6B,$FD,$19,$AF,$F1,$56,$A5,$EF,$F3,$24,$7F,$12,$8C
	.dc.b	$B7,$E0,$3C,$04,$70,$6F,$D5,$FE,$AB,$DF,$FC,$12,$BF,$DF,$7E,$87
	.dc.b	$8E,$F2,$F9,$6A,$DF,$F2,$9F,$9D,$FF,$72,$61,$FC,$31,$B5,$FD,$81
	.dc.b	$B3,$BF,$47,$8A,$84,$61,$B1,$FE,$B9,$6B,$FE,$F0,$F9,$DF,$84,$FF
	.dc.b	$AF,$7B,$1F,$09,$D1,$B1,$98,$C7,$C2,$C8,$E0,$D7,$A7,$C6,$F9,$3E
	.dc.b	$6B,$F9,$E4,$7F,$BF,$9B,$A7,$F2,$86,$E9,$FD,$F1,$FF,$6B,$93,$CB
	.dc.b	$E4,$73,$F7,$25,$FF,$B4,$E7,$FD,$B6,$D8,$F8,$7F,$3B,$FD,$43,$0E
	.dc.b	$DE,$CF,$8C,$95,$C7,$83,$CC,$47,$07,$5F,$2D,$5F,$D7,$7E,$77,$F8
	.dc.b	$86,$49,$CD,$7F,$5B,$8C,$57,$F5,$A6,$C6,$67,$6F,$F6,$48,$FE,$77
	.dc.b	$F3,$BB,$F3,$0D,$24,$FF,$62,$F9,$DF,$EC,$69,$A4,$7F,$B4,$FC,$77
	.dc.b	$F7,$4C,$7F,$4B,$C7,$78,$7F,$B9,$CD,$AF,$F1,$5C,$D7,$FB,$D3,$7E
	.dc.b	$71,$D1,$99,$FC,$2D,$FF,$7F,$C3,$FB,$0D,$6A,$30,$7F,$80,$39,$78
	.dc.b	$7C,$3F,$16,$C6,$AF,$FC,$77,$3D,$7E,$77,$31,$F4,$27,$FC,$13,$19
	.dc.b	$78,$42,$DF,$09,$98,$E0,$7F,$84,$E3,$33,$C3,$17,$37,$BD,$72,$D8
	.dc.b	$C3,$27,$BD,$3E,$1F,$18,$76,$3C,$4F,$C9,$F9,$D7,$05,$E0,$C3,$70
	.dc.b	$F4,$CF,$CA,$BE,$58,$3D,$ED,$46,$1F,$A9,$F3,$BB,$8F,$F1,$5B,$5F
	.dc.b	$F3,$1B,$5F,$F1,$93,$52,$E1,$FE,$46,$DC,$3F,$C5,$ED,$9F,$F1,$A3
	.dc.b	$5F,$F1,$CF,$C2,$FC,$69,$FF,$1A,$B6,$3F,$C7,$4D,$77,$AF,$F3,$08
	.dc.b	$FE,$27,$3C,$3A,$B0,$47,$FC,$7F,$F8,$3B,$D0,$FE,$6E,$D8,$CC,$60
	.dc.b	$FE,$77,$F3,$DF,$73,$1F,$F2,$53,$5F,$F2,$7B,$5F,$F4,$3B,$5F,$F2
	.dc.b	$B2,$F8,$6E,$64,$AF,$F2,$8C,$65,$FF,$29,$FE,$0B,$FD,$1D,$BE,$3F
	.dc.b	$F4,$C8,$C5,$FF,$96,$DA,$FF,$98,$3E,$7B,$FD,$31,$5F,$E5,$C6,$C7
	.dc.b	$F9,$85,$AE,$E3,$FA,$BC,$7F,$54,$78,$79,$38,$23,$FE,$66,$FA,$8F
	.dc.b	$D1,$37,$8F,$FA,$FC,$62,$FF,$CD,$CD,$7F,$CE,$7E,$8B,$D4,$FE,$CA
	.dc.b	$DB,$FF,$38,$B6,$7F,$CE,$CD,$78,$FA,$9C,$09,$3C,$1F,$ED,$FF,$80
	.dc.b	$7F,$9F,$A3,$FC,$F6,$D9,$FF,$40,$35,$FF,$42,$B5,$FF,$5E,$FD,$07
	.dc.b	$CA,$6E,$DF,$DE,$A3,$11,$FE,$8F,$6A,$5E,$5E,$C9,$7F,$E8,$C8,$CB
	.dc.b	$FE,$8E,$FB,$F7,$CC,$FF,$81,$9B,$19,$8C,$1F,$E0,$AF,$9F,$E3,$87
	.dc.b	$FA,$65,$AF,$FA,$79,$AF,$FB,$59,$AF,$FA,$97,$B9,$ED,$7F,$85,$1B
	.dc.b	$7F,$CE,$23,$07,$F8,$63,$EF,$7F,$05,$5F,$EA,$76,$BF,$EA,$C6,$BF
	.dc.b	$EE,$46,$BF,$EB,$3F,$9F,$7B,$1F,$E2,$26,$DF,$FA,$B9,$B3,$FE,$B3
	.dc.b	$6B,$FE,$B8,$F9,$EF,$E5,$BF,$3F,$FE,$B4,$6C,$7F,$AE,$5F,$E8,$B9
	.dc.b	$67,$C9,$57,$FA,$EB,$ED,$DF,$6F,$FE,$BA,$6E,$19,$8C,$1F,$EB,$EF
	.dc.b	$9E,$7B,$DF,$E1,$CF,$9F,$FF,$0F,$D1,$B3,$A8,$FE,$86,$D7,$FD,$98
	.dc.b	$D7,$FD,$A2,$D7,$FD,$AC,$1C,$3F,$ED,$97,$97,$9D,$FE,$BC,$6E,$DC
	.dc.b	$1B,$3F,$ED,$FF,$A1,$7C,$E7,$F8,$CB,$E8,$3F,$C2,$73,$FD,$15,$B8
	.dc.b	$7F,$B1,$9B,$87,$89,$F6,$98,$9F,$F7,$53,$53,$4F,$87,$F4,$1F,$D8
	.dc.b	$23,$FB,$1B,$FC,$6D,$BE,$37,$F9,$12,$35,$1F,$D5,$DB,$7F,$ED,$87
	.dc.b	$84,$66,$78,$70,$7F,$93,$3E,$7F,$FB,$04,$FF,$70,$FA,$07,$AD,$FE
	.dc.b	$50,$6C,$7F,$74,$9D,$BF,$D9,$5A,$FD,$00,$E1,$FA,$6D,$46,$BD,$4F
	.dc.b	$F8,$03,$1D,$32,$FF,$77,$B7,$4E,$33,$FE,$62,$F9,$3F,$A7,$7F,$E0
	.dc.b	$6F,$A0,$7C,$0F,$F9,$99,$B1,$98,$C7,$B1,$96,$46,$4E,$51,$FE,$6E
	.dc.b	$F9,$EF,$F1,$1C,$7F,$84,$9F,$97,$8F,$E3,$1B,$1E,$A7,$6F,$F8,$1D
	.dc.b	$B9,$A7,$86,$DD,$4E,$DE,$DF,$F9,$F3,$E8,$1F,$3F,$78,$A6,$5F,$E2
	.dc.b	$29,$CB,$67,$E7,$3F,$9E,$8C,$B8,$6B,$F4,$DA,$EE,$3C,$04,$7B,$7D
	.dc.b	$07,$E7,$7D,$69,$1F,$E3,$2E,$6F,$F4,$5E,$9F,$9D,$F8,$3C,$CF,$E8
	.dc.b	$6A,$CE,$63,$F2,$F8,$7E,$57,$09,$18,$21,$F9,$77,$2F,$E6,$63,$E2
	.dc.b	$7C,$DF,$C5,$F9,$C9,$3F,$0E,$32,$FC,$21,$9A,$31,$18,$6C,$66,$38
	.dc.b	$35,$FF,$1F,$37,$0F,$B1,$1F,$96,$8C,$6F,$F3,$FD,$AF,$C7,$7B,$72
	.dc.b	$BF,$24,$FE,$47,$4C,$47,$F9,$27,$97,$F8,$E9,$23,$F1,$DC,$1F,$8F
	.dc.b	$D6,$57,$E3,$E3,$2F,$E4,$63,$2F,$E3,$63,$F1,$1F,$3F,$F8,$9F,$A0
	.dc.b	$93,$F8,$78,$CB,$F8,$98,$CB,$EE,$8C,$BF,$02,$33,$AF,$C0,$8F,$97
	.dc.b	$3A,$8C,$6A,$AD,$EB,$F9,$1A,$D3,$F9,$38,$CD,$FC,$99,$CB,$F2,$E3
	.dc.b	$2F,$63,$D0,$79,$C8,$79,$CC,$9E,$73,$37,$E4,$9E,$B1,$F8,$29,$EB
	.dc.b	$69,$CB,$E9,$9E,$37,$E9,$7A,$DF,$BA,$F8,$DF,$06,$CE,$AB,$DC,$F2
	.dc.b	$3D,$C6,$4F,$71,$93,$DC,$7A,$CF,$51,$2B,$D4,$F1,$3D,$47,$1B,$F5
	.dc.b	$39,$7D,$4F,$19,$F5,$3D,$67,$D4,$F1,$BF,$53,$4E,$57,$EC,$78,$57
	.dc.b	$B1,$E8,$7B,$0C,$9E,$C3,$27,$B2,$BD,$2E,$2F,$CB,$18,$24,$F5,$7C
	.dc.b	$B3,$D0,$F8,$A7,$F1,$7B,$FD,$B9,$1F,$DB,$B5,$FE,$3C,$D7,$81,$F1
	.dc.b	$5F,$E5,$9F,$01,$F8,$07,$F0,$2D,$8F,$94,$D7,$C5,$7F,$EB,$57,$91
	.dc.b	$A2,$9A,$8E,$15,$EC,$8F,$E0,$D1,$8B,$FA,$73,$FA,$88,$C4,$FC,$B8
	.dc.b	$C3,$D1,$E4,$FF,$1A,$8D,$BF,$36,$7F,$9F,$46,$2B,$F0,$73,$FC,$62
	.dc.b	$31,$7F,$02,$3A,$57,$88,$E2,$E0,$F1,$38,$7E,$0C,$7C,$67,$47,$84
	.dc.b	$F6,$98,$AF,$0D,$FA,$A3,$89,$F1,$CA,$7C,$73,$84,$F9,$5C,$C7,$C0
	.dc.b	$9F,$A1,$18,$AF,$43,$E0,$30,$7A,$0C,$4F,$F1,$7F,$A0,$7C,$9F,$F9
	.dc.b	$02,$31,$EB,$DB,$FF,$21,$D2,$31,$7F,$E1,$28,$E2,$FF,$85,$2B,$A5
	.dc.b	$EA,$13,$C5,$D5,$F4,$BD,$4F,$29,$5F,$F2,$BF,$A0,$72,$73,$6C,$F2
	.dc.b	$3F,$C4,$6E,$1E,$38,$7F,$96,$A3,$FA,$8C,$6A,$79,$11,$E4,$23,$6B
	.dc.b	$E7,$5F,$41,$EB,$5F,$43,$4C,$44,$9F,$ED,$93,$C1,$B1,$FD,$B2,$31
	.dc.b	$1F,$E1,$26,$E3,$FD,$25,$B8,$FF,$9C,$9B,$3F,$E4,$07,$FC,$6A,$FB
	.dc.b	$8F,$F2,$1F,$D0,$3C,$1F,$A0,$BE,$86,$98,$89,$3F,$02,$74,$D9,$CB
	.dc.b	$D4,$FF,$47,$EF,$F6,$FC,$CF,$AD,$97,$F5,$F6,$E5,$FE,$88,$7E,$F5
	.dc.b	$71,$72,$FF,$96,$1A,$F4,$3E,$8B,$C2,$7E,$17,$C7,$7C,$FE,$E9,$5F
	.dc.b	$5B,$44,$FD,$33,$E6,$12,$3E,$8B,$5F,$54,$70,$79,$47,$DB,$FA,$1F
	.dc.b	$CB,$7C,$97,$FB,$0A,$7F,$06,$BE,$63,$26,$BD,$79,$3A,$8C,$BC,$23
	.dc.b	$F4,$1F,$6F,$F9,$3C,$FF,$C0,$8D,$D7,$FD,$5A,$DA,$FF,$3D,$BF,$90
	.dc.b	$30,$9A,$4F,$0B,$F5,$9D,$5F,$92,$9F,$CC,$BF,$86,$7F,$BC,$72,$6B
	.dc.b	$96,$47,$F8,$FE,$78,$3F,$E6,$26,$BF,$E9,$16,$BF,$EC,$27,$78,$FF
	.dc.b	$87,$5B,$8E,$6B,$C7,$F5,$5F,$BF,$FB,$57,$B1,$7E,$46,$65,$FA,$E9
	.dc.b	$CB,$67,$2F,$E3,$E1,$3C,$E7,$A7,$6C,$AE,$4F,$7F,$9D,$C4,$71,$8B
	.dc.b	$A7,$F8,$CF,$05,$7F,$13,$FA,$4F,$BD,$7F,$9D,$FA,$AF,$F5,$ED,$FF
	.dc.b	$8E,$5B,$AF,$FB,$8A,$B7,$3F,$E3,$D7,$8D,$7F,$AD,$DA,$EF,$8B,$FC
	.dc.b	$D3,$18,$8F,$F5,$BB,$CA,$B5,$1F,$C4,$23,$0F,$FA,$F2,$FF,$88,$F3
	.dc.b	$BE,$97,$F8,$41,$27,$F8,$6B,$FE,$76,$FA,$57,$CA,$D2,$4F,$F9,$F6
	.dc.b	$32,$DD,$B8,$11,$C1,$FE,$83,$FA,$57,$D2,$FF,$A0,$9B,$19,$8C,$19
	.dc.b	$FA,$59,$1C,$5A,$FF,$AB,$BE,$33,$FD,$DA,$FF,$B7,$5A,$FF,$B6,$83
	.dc.b	$87,$FD,$BA,$D5,$4F,$96,$F2,$12,$3D,$0F,$9F,$E2,$24,$7A,$DA,$F8
	.dc.b	$FC,$DF,$F4,$ED,$FF,$BF,$23,$C9,$18,$3E,$07,$C7,$C8,$F8,$8D,$7F
	.dc.b	$D9,$38,$7F,$57,$7F,$DF,$0D,$7D,$8D,$7F,$DE,$ED,$7D,$FE,$7F,$88
	.dc.b	$8F,$FA,$CE,$FF,$D9,$EF,$0E,$6F,$A4,$47,$EB,$3F,$57,$E9,$24,$7D
	.dc.b	$96,$BF,$47,$D5,$F7,$91,$EC,$FD,$DE,$79,$1F,$81,$6B,$EE,$F0,$BF
	.dc.b	$04,$47,$F0,$CF,$CA,$F8,$C9,$1E,$96,$BF,$81,$E9,$F9,$E4,$7D,$CF
	.dc.b	$D2,$F9,$49,$1F,$05,$AF,$E2,$7E,$0F,$EB,$91,$F8,$CF,$DA,$F6,$BF
	.dc.b	$C1,$DA,$BE,$8E,$98,$89,$5F,$35,$E0,$D3,$FD,$C5,$F4,$6F,$6B,$FC
	.dc.b	$30,$D8,$FF,$02,$B5,$FE,$C8,$FF,$BF,$32,$FC,$D3,$60,$1C,$37,$D9
	.dc.b	$FF,$12,$7D,$1F,$E8,$23,$FC,$23,$F4,$7F,$C8,$1B,$E3,$7F,$8A,$63
	.dc.b	$11,$FD,$CD,$FD,$DB,$7B,$F4,$47,$FC,$37,$4F,$94,$FF,$19,$7D,$17
	.dc.b	$F4,$52,$79,$98,$F8,$49,$7F,$E2,$29,$FF,$06,$F9,$1F,$ED,$B6,$ED
	.dc.b	$FB,$75,$F0,$9D,$09,$5C,$5F,$ED,$6F,$83,$82,$3D,$5F,$1E,$F2,$3C
	.dc.b	$0D,$7F,$C2,$14,$F5,$3F,$C8,$70,$FA,$37,$07,$8E,$73,$3F,$C2,$C5
	.dc.b	$71,$CF,$11,$F5,$23,$0E,$1A,$F1,$7F,$C1,$BF,$47,$C2,$39,$FA,$C8
	.dc.b	$C4,$EA,$79,$4E,$5E,$03,$E8,$EB,$4C,$B8,$37,$E8,$37,$FF,$29,$CF
	.dc.b	$F6,$26,$C6,$63,$FC,$68,$64,$BC,$B9,$CB,$50,$F9,$C7,$46,$70,$F8
	.dc.b	$D7,$6B,$F2,$F5,$A9,$F0,$F3,$E6,$7F,$37,$F1,$95,$FE,$9F,$3E,$68
	.dc.b	$C1,$9F,$A3,$B1,$EA,$70,$FF,$9B,$AD,$F1,$38,$BF,$F2,$74,$62,$7F
	.dc.b	$CE,$79,$FE,$39,$BC,$BE,$38,$C5,$7F,$95,$5F,$DC,$4E,$8E,$DF,$17
	.dc.b	$BA,$F5,$5A,$3D,$9F,$47,$FE,$16,$AF,$F4,$0B,$F2,$EF,$E4,$B9,$C3
	.dc.b	$FC,$C4,$E6,$7F,$AC,$36,$3E,$4F,$C3,$F4,$DB,$BB,$F2,$7E,$8F,$E3
	.dc.b	$CF,$E4,$9A,$FE,$5D,$AF,$F6,$36,$BF,$C2,$97,$D2,$7D,$C8,$D3,$5D
	.dc.b	$35,$47,$0C,$3F,$35,$F1,$8E,$BD,$36,$33,$18,$39,$7D,$27,$F6,$F8
	.dc.b	$FD,$13,$6F,$F5,$6D,$7F,$60,$D7,$FB,$E7,$E0,$FF,$CC,$E7,$F5,$4D
	.dc.b	$8F,$D9,$B5,$FD,$C3,$5F,$F0,$27,$D2,$BD,$2F,$63,$6B,$F6,$8D,$9F
	.dc.b	$DE,$B5,$FE,$05,$F4,$BF,$E0,$C9,$FD,$D3,$63,$F8,$1B,$5D,$F9,$1F
	.dc.b	$27,$EE,$1E,$5D,$25,$FF,$03,$8C,$BF,$C1,$7E,$9A,$DF,$F0,$E6,$BF
	.dc.b	$C4,$DA,$FF,$87,$DA,$FF,$46,$FA,$6F,$DB,$4F,$F0,$F6,$C7,$F1,$86
	.dc.b	$BF,$C7,$5A,$FF,$8A,$7F,$04,$FF,$2E,$CF,$F1,$76,$C7,$F2,$06,$BF
	.dc.b	$C9,$5A,$FF,$8C,$7D,$C7,$B1,$F8,$96,$D7,$F2,$16,$CF,$F2,$A6,$BF
	.dc.b	$CB,$FE,$9B,$FD,$CB,$3F,$C9,$DB,$1F,$CC,$1A,$EF,$CC,$FC,$B7,$E0
	.dc.b	$4F,$02,$92,$FF,$98,$46,$5F,$E6,$5F,$4F,$FD,$6D,$BE,$7F,$E7,$E3
	.dc.b	$18,$7F,$3E,$6B,$FD,$0F,$E9,$DE,$AF,$E9,$5B,$7F,$CF,$5B,$3F,$D1
	.dc.b	$9A,$FF,$4B,$F0,$7F,$D9,$F3,$FD,$0D,$B1,$FD,$31,$AE,$FC,$AF,$D9
	.dc.b	$79,$2F,$76,$92,$FF,$A6,$46,$5F,$E9,$DF,$4D,$F8,$4B,$34,$9C,$FF
	.dc.b	$BD,$FC,$05,$F4,$74,$97,$FD,$5E,$32,$FF,$59,$FA,$6F,$E5,$FF,$4D
	.dc.b	$7C,$C3,$9F,$3F,$FD,$67,$9F,$F6,$9A,$DB,$FD,$A3,$E9,$BB,$BF,$D8
	.dc.b	$DA,$E7,$2F,$EB,$F1,$88,$FE,$E5,$4F,$98,$FE,$C6,$D8,$FE,$D8,$D7
	.dc.b	$FB,$AB,$5F,$EF,$5F,$38,$F8,$4F,$F4,$37,$9F,$FD,$72,$3F,$D6,$B3
	.dc.b	$FE,$2E,$6A,$78,$3F,$E3,$28,$E1,$3F,$DD,$27,$FC,$DB,$3B,$8C,$11
	.dc.b	$E1,$AF,$FA,$56,$79,$C2,$B9,$BA,$3C,$5F,$05,$E6,$AF,$F4,$E7,$CF
	.dc.b	$FF,$90,$09,$E5,$E6,$BF,$C8,$B1,$FE,$1D,$79,$07,$6E,$1E,$5E,$CF
	.dc.b	$18,$7D,$1F,$93,$1F,$EF,$EB,$1D,$36,$7A,$BB,$C7,$FD,$60,$D8,$FF
	.dc.b	$2B,$BB,$7D,$D1,$A9,$E7,$9F,$87,$37,$CA,$5F,$9D,$98,$7A,$A7,$2D
	.dc.b	$9F,$A0,$47,$FC,$59,$1A,$7F,$C6,$AD,$CB,$EF,$4F,$73,$C1,$E8,$7C
	.dc.b	$1F,$A0,$FF,$39,$4F,$F8,$79,$B9,$FF,$B2,$DB,$87,$F9,$C9,$FF,$38
	.dc.b	$98,$4D,$26,$FF,$37,$E8,$7F,$16,$7C,$59,$34,$CF,$A1,$BD,$64,$FD
	.dc.b	$A9,$FC,$9E,$8F,$97,$F9,$1F,$43,$2B,$E3,$47,$CB,$3F,$28,$47,$E2
	.dc.b	$B5,$FA,$67,$E7,$DE,$0F,$CF,$F5,$FF,$09,$F4,$2F,$9E,$BE,$6C,$AF
	.dc.b	$C6,$CE,$5B,$39,$7A,$88,$4F,$49,$FC,$87,$D0,$B9,$29,$2F,$F3,$31
	.dc.b	$F6,$78,$7F,$DF,$8E,$E3,$33,$B7,$BC,$6D,$FF,$53,$FC,$B3,$CE,$FC
	.dc.b	$F2,$F9,$BA,$62,$24,$FF,$9D,$67,$83,$63,$FC,$E3,$18,$8F,$F4,$0B
	.dc.b	$79,$7F,$CC,$AD,$C7,$FD,$6F,$28,$C3,$E4,$F8,$9F,$D8,$7D,$0F,$F1
	.dc.b	$7B,$FF,$36,$AF,$A1,$C9,$AF,$02,$5F,$EF,$27,$4D,$9C,$BF,$43,$E8
	.dc.b	$7E,$A3,$77,$FD,$F4,$63,$0F,$F3,$DB,$F5,$27,$F8,$3B,$9C,$FD,$2D
	.dc.b	$8F,$F6,$8B,$53,$C5,$FB,$F9,$BC,$CF,$9F,$E0,$49,$FE,$27,$39,$6C
	.dc.b	$E5,$E3,$3F,$5A,$24,$70,$8D,$61,$F3,$A7,$33,$C0,$D1,$D0,$86,$49
	.dc.b	$1C,$5C,$BF,$62,$7A,$E1,$F6,$2F,$CD,$5E,$CB,$CB,$F8,$CA,$F0,$DF
	.dc.b	$38,$EB,$7C,$67,$2F,$3A,$E3,$3F,$83,$9C,$BF,$96,$8C,$DF,$F3,$98
	.dc.b	$F0,$1E,$6A,$E8,$FB,$8C,$92,$FD,$4E,$5F,$21,$23,$8C,$79,$8F,$01
	.dc.b	$D4,$F6,$1C,$5F,$43,$93,$D0,$64,$E2,$78,$5E,$B1,$E2,$8C,$1A,$70
	.dc.b	$48,$EB,$1A,$3A,$9E,$53,$F5,$15,$F1,$9C,$CF,$17,$A4,$F2,$8C,$46
	.dc.b	$A3,$04,$7D,$A7,$5A,$FB,$B3,$F1,$63,$2F,$28,$CC,$F1,$9E,$86,$9C
	.dc.b	$46,$A3,$04,$3D,$27,$E4,$8F,$E0,$66,$0F,$E1,$35,$FC,$30,$8F,$F1
	.dc.b	$16,$BA,$30,$7D,$38,$F0,$3F,$C3,$7E,$87,$E3,$46,$08,$E2,$72,$D9
	.dc.b	$E4,$D7,$E9,$4A,$31,$7F,$16,$7B,$1E,$1A,$E6,$D3,$99,$4E,$73,$CA
	.dc.b	$FE,$3B,$63,$EC,$9F,$CD,$9C,$3F,$16,$BF,$95,$C6,$2F,$D7,$3D,$67
	.dc.b	$C9,$1E,$78,$C5,$76,$69,$D8,$F4,$9E,$EB,$F4,$B4,$F4,$94,$F4,$9C
	.dc.b	$8F,$24,$7F,$05,$BF,$24,$FF,$4C,$6C,$FF,$1E,$6B,$FC,$91,$B7,$EC
	.dc.b	$8F,$F1,$AF,$D0,$BE,$53,$FC,$64,$BE,$52,$98,$89,$3C,$23,$2F,$28
	.dc.b	$FF,$1D,$7D,$0F,$F6,$58,$FE,$E2,$DB,$65,$FD,$52,$30,$DF,$0F,$8F
	.dc.b	$3C,$BF,$F2,$4D,$69,$07,$FB,$73,$FE,$13,$E6,$FF,$25,$75,$FF,$5E
	.dc.b	$8F,$F0,$B7,$3E,$26,$47,$E8,$8E,$73,$FD,$CD,$C1,$CC,$AE,$5E,$0D
	.dc.b	$CB,$D7,$1C,$A7,$AB,$FE,$10,$97,$D0,$FF,$97,$7E,$87,$FD,$27,$7C
	.dc.b	$29,$0F,$88,$D1,$BF,$4E,$4F,$F9,$AA,$74,$D9,$CB,$FE,$1E,$8F,$EC
	.dc.b	$55,$D5,$BB,$7F,$80,$9B,$3F,$29,$7C,$CF,$97,$BA,$4F,$77,$D1,$1F
	.dc.b	$E7,$5F,$A1,$FC,$34,$FF,$9F,$3E,$87,$FD,$2A,$7F,$A5,$A3,$31,$FE
	.dc.b	$30,$6B,$FC,$41,$AF,$F8,$51,$AF,$DC,$6A,$79,$E3,$F1,$DB,$8F,$1E
	.dc.b	$3F,$F4,$45,$FE,$08,$7D,$6F,$FA,$2F,$F0,$DF,$F6,$1E,$3F,$E9,$26
	.dc.b	$BF,$E4,$49,$DB,$EB,$9E,$1E,$A4,$8C,$1C,$7E,$8F,$A1,$1E,$9F,$21
	.dc.b	$5F,$F2,$7E,$09,$C6,$9F,$7E,$15,$FE,$9F,$C9,$F2,$57,$E0,$E6,$3F
	.dc.b	$EA,$89,$CB,$67,$F2,$AF,$ED,$DB,$1A,$8F,$CE,$35,$FE,$7E,$E7,$0F
	.dc.b	$C8,$37,$5F,$F2,$2B,$67,$F5,$AB,$E7,$BF,$D6,$9F,$74,$F9,$EB,$EC
	.dc.b	$4B,$FF,$5E,$CE,$5B,$39,$7F,$CE,$30,$9E,$73,$D3,$E0,$65,$72,$7B
	.dc.b	$FD,$4E,$23,$8C,$5C,$FF,$51,$82,$BF,$D9,$1F,$72,$F5,$AF,$91,$31
	.dc.b	$FF,$68,$CE,$5B,$3F,$2A,$31,$1F,$C0,$DA,$FF,$66,$6B,$FE,$6A,$77
	.dc.b	$87,$EE,$5B,$B6,$6B,$C7,$F1,$FF,$ED,$BF,$A6,$7C,$CF,$D8,$6E,$3F
	.dc.b	$C3,$9A,$FF,$13,$6B,$FD,$D9,$FF,$2E,$37,$B3,$91,$1F,$E2,$FF,$72
	.dc.b	$F8,$BF,$C2,$B6,$33,$18,$3F,$13,$F4,$8F,$A6,$A4,$9F,$E4,$31,$97
	.dc.b	$F9,$1E,$4F,$86,$FC,$8B,$63,$31,$83,$3F,$0B,$23,$8B,$5F,$E6,$14
	.dc.b	$F8,$AF,$19,$23,$C2,$F6,$F8,$09,$1E,$26,$BF,$CE,$70,$FE,$9C,$FF
	.dc.b	$87,$1A,$FF,$A8,$1A,$FF,$48,$6A,$AF,$F4,$AB,$FE,$24,$6B,$FE,$A8
	.dc.b	$6B,$FD,$41,$AA,$1F,$01,$F1,$09,$1F,$21,$F8,$FE,$7C,$8F,$96,$D7
	.dc.b	$E1,$F5,$FD,$42,$3F,$49,$F9,$FF,$07,$23,$E9,$B5,$F8,$5F,$07,$F5
	.dc.b	$08,$FD,$87,$EF,$FA,$52,$3E,$EB,$5F,$95,$F1,$9F,$70,$8F,$E0,$1F
	.dc.b	$BF,$F2,$52,$3F,$06,$D7,$E9,$7C,$5F,$E1,$08,$FE,$21,$FC,$3F,$D2
	.dc.b	$C8,$FC,$5B,$5F,$B5,$E0,$7E,$30,$8F,$E4,$1F,$C7,$F8,$52,$3F,$26
	.dc.b	$D7,$F0,$4F,$FB,$7B,$BB,$FB,$67,$07,$F0,$F9,$FF,$03,$C6,$27,$5D
	.dc.b	$BF,$85,$F0,$2D,$7F,$83,$3E,$95,$F3,$D8,$84,$61,$E7,$5F,$EE,$C8
	.dc.b	$C4,$30,$FE,$F9,$1F,$C1,$23,$A7,$81,$CE,$31,$1D,$23,$B7,$D2,$FE
	.dc.b	$2E,$3A,$9F,$E1,$EF,$A3,$FE,$FE,$F5,$C3,$FD,$F7,$18,$C3,$FC,$14
	.dc.b	$D8,$CF,$85,$E9,$8F,$E1,$B3,$A8,$C1,$E0,$EF,$F7,$7D,$17,$F8,$9C
	.dc.b	$9E,$CE,$3E,$16,$65,$E4,$12,$8C,$C7,$94,$FF,$1A,$C3,$E9,$7A,$92
	.dc.b	$39,$46,$5F,$F2,$89,$E2,$DE,$3A,$7C,$BB,$48,$E2,$D7,$C0,$79,$3C
	.dc.b	$8F,$A4,$BE,$95,$B7,$F9,$22,$35,$B7,$F8,$FB,$15,$F0,$9F,$F5,$2C
	.dc.b	$62,$75,$39,$9E,$52,$F8,$17,$D2,$F3,$6C,$FF,$93,$67,$53,$93,$FB
	.dc.b	$CC,$3E,$96,$91,$FE,$FA,$FF,$7E,$F7,$3F,$C9,$B8,$FF,$91,$E7,$FC
	.dc.b	$8D,$C7,$F4,$D4,$EA,$3F,$D7,$BE,$E5,$9F,$80,$D7,$FC,$23,$D1,$5E
	.dc.b	$67,$CA,$F5,$3F,$BB,$57,$5E,$6D,$7A,$7B,$6C,$78,$63,$97,$A9,$8B
	.dc.b	$FA,$61,$CC,$FE,$26,$FF,$A5,$3E,$98,$F2,$46,$27,$ED,$E7,$F8,$9A
	.dc.b	$FC,$83,$FB,$46,$F2,$FA,$E3,$13,$C3,$97,$9F,$6E,$5B,$7F,$99,$FA
	.dc.b	$5F,$8E,$DD,$BE,$24,$62,$FF,$16,$FC,$4C,$3F,$CE,$CD,$AF,$C5,$64
	.dc.b	$F8,$DE,$38,$7A,$9F,$9D,$F4,$9F,$E3,$CA,$FF,$45,$B7,$1F,$F4,$83
	.dc.b	$5F,$F7,$9B,$5F,$CB,$3F,$26,$38,$EB,$F8,$19,$FF,$50,$CE,$5E,$71
	.dc.b	$95,$E3,$7E,$72,$BF,$D3,$6D,$8F,$F5,$03,$5F,$D2,$1C,$39,$F9,$47
	.dc.b	$E7,$67,$8F,$3E,$08,$F4,$F8,$6D,$91,$C1,$FE,$AA,$FA,$2F,$EC,$AD
	.dc.b	$F3,$3C,$D1,$8C,$7F,$D6,$4D,$7C,$6B,$CA,$D9,$2B,$A3,$CB,$C9,$91
	.dc.b	$DD,$B5,$FE,$BD,$FA,$0F,$F2,$C4,$FF,$AF,$1B,$19,$F6,$7E,$44,$62
	.dc.b	$3E,$1B,$FA,$27,$97,$56,$89,$5C,$FD,$3E,$24,$70,$7F,$B2,$BD,$67
	.dc.b	$C5,$FD,$56,$C6,$63,$06,$62,$47,$16,$DF,$FB,$5D,$FF,$5B,$C7,$5F
	.dc.b	$82,$E8,$47,$FD,$B3,$F8,$97,$FB,$99,$B5,$F8,$58,$C3,$FE,$E1,$6B
	.dc.b	$F8,$27,$F4,$4D,$F0,$38,$12,$7A,$F4,$73,$23,$83,$FD,$D3,$F6,$3F
	.dc.b	$EA,$16,$FC,$07,$E5,$63,$17,$FE,$F1,$6B,$F9,$2F,$AE,$7C,$07,$E6
	.dc.b	$DB,$7F,$EF,$46,$CF,$FB,$E5,$AF,$E6,$BE,$6F,$FC,$AD,$F0,$7F,$EF
	.dc.b	$A6,$C7,$FB,$FD,$FF,$49,$67,$9E,$0F,$F6,$1C,$FF,$45,$FA,$E7,$C8
	.dc.b	$7F,$AB,$9B,$1F,$A6,$01,$CB,$96,$BF,$AC,$F9,$7F,$F2,$2B,$C4,$38
	.dc.b	$7A,$35,$E1,$D5,$FE,$BE,$87,$C6,$52,$4F,$38,$CB,$D7,$D6,$FF,$06
	.dc.b	$9F,$D7,$1F,$EF,$0D,$7F,$B3,$BA,$F4,$7F,$D9,$6F,$FA,$7A,$3D,$AD
	.dc.b	$7F,$D5,$D1,$A3,$FC,$49,$1C,$63,$FC,$09,$F5,$CF,$82,$FF,$14,$7C
	.dc.b	$E7,$F6,$59,$FF,$04,$2F,$07,$26,$BC,$29,$3A,$8C,$BC,$23,$FC,$65
	.dc.b	$F3,$8F,$06,$92,$7F,$C2,$90,$F7,$31,$1F,$E0,$48,$E0,$EA,$79,$CF
	.dc.b	$F8,$F7,$EB,$BF,$AB,$4F,$F8,$75,$FF,$03,$B7,$D6,$D7,$93,$23,$ED
	.dc.b	$4F,$07,$F0,$8D,$7F,$95,$B5,$FF,$08,$3B,$F3,$9D,$78,$AE,$95,$FE
	.dc.b	$1A,$F9,$8F,$94,$FF,$8B,$FE,$77,$EC,$5F,$08,$FC,$8C,$7C,$D7,$FC
	.dc.b	$42,$D8,$D4,$23,$8D,$FF,$88,$5B,$1C,$B5,$FF,$2F,$47,$19,$D7,$99
	.dc.b	$FE,$61,$F9,$DF,$E6,$57,$FE,$41,$6B,$FE,$30,$6E,$1E,$07,$E7,$C6
	.dc.b	$7C,$BE,$44,$79,$6F,$FE,$6D,$F9,$17,$C2,$FE,$33,$49,$5F,$E4,$F9
	.dc.b	$CB,$67,$F5,$4A,$32,$FE,$D5,$AF,$F5,$C6,$BF,$E4,$17,$78,$FC,$66
	.dc.b	$E7,$9A,$FF,$40,$73,$3E,$5B,$F3,$7F,$73,$F9,$CF,$1E,$65,$F9,$C8
	.dc.b	$CB,$F9,$E8,$CB,$FE,$54,$E8,$4F,$E6,$BA,$64,$FE,$61,$FC,$ED,$E2
	.dc.b	$3F,$31,$17,$4F,$B7,$D1,$E3,$9F,$CB,$7D,$77,$E6,$7E,$36,$46,$08
	.dc.b	$7E,$52,$3E,$3F,$41,$A4,$8D,$DF,$E2,$27,$51,$FE,$7F,$E7,$F0,$3F
	.dc.b	$8E,$EB,$FC,$3E,$92,$3F,$2D,$1F,$8A,$30,$FE,$52,$30,$FE,$27,$AF
	.dc.b	$F1,$84,$7F,$19,$A5,$7F,$19,$17,$1F,$C6,$4F,$FA,$CF,$E6,$9C,$17
	.dc.b	$09,$87,$FA,$56,$72,$D9,$CB,$FE,$86,$8F,$A6,$E4,$86,$67,$77,$FA
	.dc.b	$A9,$E7,$3C,$6B,$3C,$7F,$E8,$B9,$E7,$39,$39,$D7,$1B,$DC,$7F,$48
	.dc.b	$9E,$73,$D6,$F3,$29,$E7,$39,$8E,$77,$DE,$F7,$5C,$A3,$C5,$78,$9F
	.dc.b	$8C,$CA,$F8,$91,$C2,$31,$1E,$F3,$27,$63,$CC,$F3,$70,$79,$5C,$3E
	.dc.b	$F2,$5F,$BD,$E6,$62,$BC,$AE,$23,$4E,$4E,$E6,$63,$DC,$F8,$2F,$DC
	.dc.b	$FC,$9B,$F5,$39,$6B,$CC,$F9,$4E,$2B,$E5,$47,$BC,$8E,$09,$5E,$08
	.dc.b	$CB,$E8,$7A,$57,$D4,$8D,$4F,$09,$F4,$9C,$0F,$31,$E9,$BF,$8F,$1B
	.dc.b	$8F,$94,$7B,$CD,$9D,$A3,$F8,$ED,$7E,$42,$3F,$CE,$AD,$7C,$86,$23
	.dc.b	$64,$7F,$C1,$D1,$B7,$CB,$3E,$18,$F9,$71,$F8,$FF,$5F,$F0,$73,$9A
	.dc.b	$FD,$74,$62,$FE,$14,$F1,$8C,$57,$49,$C9,$FA,$38,$CC,$7B,$8F,$92
	.dc.b	$E2,$FC,$24,$71,$79,$6C,$FE,$05,$AF,$23,$E6,$19,$29,$93,$E5,$19
	.dc.b	$29,$9B,$F7,$39,$6B,$9B,$F5,$BE,$D2,$1F,$1C,$EF,$1F,$0D,$B1,$C1
	.dc.b	$D1,$E0,$8D,$36,$35,$6A,$F6,$D6,$8F,$19,$67,$77,$F1,$6F,$B9,$E4
	.dc.b	$BE,$4F,$F8,$BA,$FF,$C6,$3F,$53,$23,$04,$3F,$AA,$39,$7F,$AB,$47
	.dc.b	$F8,$BB,$E3,$24,$7F,$8B,$1F,$F1,$8F,$AF,$88,$FF,$16,$45,$CB,$FA
	.dc.b	$AC,$FF,$35,$F9,$CF,$E6,$DC,$12,$FF,$9A,$46,$5F,$CA,$BF,$9E,$23
	.dc.b	$83,$64,$3F,$9C,$7C,$57,$F8,$5B,$4F,$F0,$BE,$52,$7F,$C4,$D1,$97
	.dc.b	$FC,$51,$19,$7F,$C4,$B1,$FE,$1B,$E6,$91,$FE,$1B,$7F,$C3,$D9,$E2
	.dc.b	$3F,$C3,$91,$73,$FF,$0E,$4F,$FA,$02,$51,$FE,$2E,$F9,$27,$C1,$66
	.dc.b	$31,$E9,$DE,$1E,$14,$8C,$5F,$FA,$2A,$38,$BF,$E8,$CA,$E9,$7F,$E1
	.dc.b	$E9,$CC,$FF,$A4,$5A,$FF,$BE,$1A,$F6,$8E,$AA,$38,$CF,$48,$ED,$7E
	.dc.b	$4F,$82,$DF,$CF,$BD,$7A,$4A,$D9,$FE,$A0,$8F,$F1,$3B,$63,$4F,$C4
	.dc.b	$AF,$8D,$3C,$E3,$88,$78,$AF,$9F,$71,$FA,$17,$C7,$D3,$11,$27,$9C
	.dc.b	$E9,$B3,$97,$C4,$7C,$DF,$B2,$F8,$7A,$3B,$A9,$88,$93,$FE,$B5,$94
	.dc.b	$7D,$12,$46,$5D,$BD,$8F,$9C,$F6,$FA,$FF,$A3,$F4,$2E,$6A,$4A,$F8
	.dc.b	$70,$F8,$FC,$47,$38,$E0,$EA,$79,$CF,$DC,$F5,$7E,$64,$F8,$5F,$96
	.dc.b	$4E,$0D,$1B,$F1,$A4,$FD,$29,$E3,$C7,$F9,$9F,$7B,$E2,$FB,$F2,$BE
	.dc.b	$BE,$C9,$F9,$27,$E4,$C9,$1F,$1D,$AF,$C3,$8E,$0F,$28,$FA,$38,$FE
	.dc.b	$53,$E5,$3F,$D5,$B3,$FE,$EB,$6B,$FA,$D0,$E1,$FD,$83,$5F,$F7,$8B
	.dc.b	$FE,$56,$79,$78,$DF,$E8,$48,$C4,$73,$6B,$CB,$87,$44,$7F,$DE,$91
	.dc.b	$F8,$57,$FD,$ED,$F7,$0E,$1F,$F4,$7B,$63,$31,$83,$FD,$27,$C0,$FA
	.dc.b	$3A,$49,$FF,$80,$11,$97,$F1,$DC,$1F,$BD,$23,$83,$7E,$C7,$42,$38
	.dc.b	$39,$FC,$47,$80,$8E,$0E,$FF,$3B,$E4,$23,$83,$E7,$01,$FB,$8B,$E0
	.dc.b	$7F,$D5,$AD,$8D,$9A,$8C,$9A,$FC,$02,$46,$A5,$18,$8E,$F0,$FA,$3E
	.dc.b	$A4,$8E,$0F,$83,$E5,$A4,$73,$6B,$CB,$E5,$BC,$A4,$7C,$0F,$A3,$E4
	.dc.b	$A4,$78,$DA,$F8,$BE,$47,$DA,$47,$D0,$FC,$0F,$81,$91,$EB,$6B,$EA
	.dc.b	$F7,$7E,$29,$1F,$80,$FC,$8F,$6A,$47,$C3,$6B,$F0,$BD,$5F,$9A,$47
	.dc.b	$E4,$3F,$43,$D3,$91,$F2,$DA,$FC,$AF,$3B,$EA,$91,$FA,$0F,$D8,$F3
	.dc.b	$FF,$A0,$EF,$FE,$3E,$23,$FD,$13,$E6,$1A,$52,$5F,$F3,$68,$69,$B2
	.dc.b	$46,$5F,$E9,$BF,$31,$FC,$5A,$3F,$91,$7C,$17,$F5,$D7,$F9,$FB,$5C
	.dc.b	$FA,$1F,$D5,$23,$FA,$9F,$9D,$FD,$AD,$B3,$F9,$36,$A6,$D7,$F2,$78
	.dc.b	$FE,$97,$F2,$EE,$4F,$EB,$ED,$8C,$C6,$39,$24,$70,$6B,$FC,$E6,$5C
	.dc.b	$9C,$89,$1D,$9F,$ED,$7F,$4B,$FC,$57,$C8,$67,$FE,$03,$BF,$97,$7E
	.dc.b	$4D,$FC,$67,$93,$69,$1E,$86,$BE,$1E,$6F,$F0,$5A,$F9,$9F,$E4,$B6
	.dc.b	$9C,$CF,$F2,$58,$C6,$3F,$B4,$6C,$7E,$F1,$AE,$F9,$DF,$33,$B6,$C6
	.dc.b	$DA,$EE,$3F,$7E,$FF,$05,$F9,$9A,$9A,$47,$F8,$2F,$C6,$FC,$47,$FB
	.dc.b	$21,$FE,$1F,$D7,$FC,$02,$A3,$58,$EE,$3F,$89,$72,$62,$BF,$A1,$36
	.dc.b	$3F,$8B,$73,$3E,$4F,$11,$C2,$BF,$C1,$B3,$97,$BF,$AF,$A9,$7A,$F6
	.dc.b	$30,$C9,$F2,$FC,$4D,$AF,$4B,$C4,$EF,$6F,$8E,$C5,$78,$1F,$1E,$7F
	.dc.b	$8A,$8F,$F0,$FB,$FA,$E6,$F3,$6B,$96,$47,$A3,$E6,$6E,$5F,$E0,$BC
	.dc.b	$51,$EE,$F9,$9F,$91,$5F,$E2,$A7,$E6,$57,$F8,$45,$CB,$77,$E0,$7A
	.dc.b	$3C,$6B,$5F,$19,$F9,$3F,$32,$DB,$E9,$5F,$F9,$76,$31,$5F,$D0,$1B
	.dc.b	$1F,$D2,$1A,$FF,$77,$7F,$6D,$1B,$EE,$E7,$3F,$D3,$67,$2F,$18,$CB
	.dc.b	$FE,$40,$E6,$BE,$D9,$A4,$BF,$EA,$30,$F1,$36,$48,$CB,$D3,$EC,$5F
	.dc.b	$3D,$49,$3F,$D6,$A3,$2F,$F5,$CF,$9B,$7C,$77,$91,$B1,$98,$C1,$9F
	.dc.b	$8E,$91,$C5,$B9,$FF,$6C,$FB,$0F,$F1,$E4,$FF,$68,$6C,$67,$DD,$F8
	.dc.b	$31,$88,$FF,$10,$7C,$DF,$E6,$5F,$EE,$AD,$73,$DB,$F2,$23,$11,$FE
	.dc.b	$93,$7F,$29,$7F,$DD,$9B,$1C,$BA,$F8,$1F,$43,$E0,$5F,$11,$49,$7F
	.dc.b	$DF,$A3,$2F,$F8,$03,$DD,$FF,$21,$35,$FB,$26,$2F,$FC,$16,$D7,$FC
	.dc.b	$74,$FF,$A2,$E3,$97,$BB,$C0,$8F,$F8,$33,$EB,$9E,$5F,$E1,$1B,$19
	.dc.b	$8C,$1F,$88,$F9,$CF,$CF,$E1,$FE,$47,$6B,$FE,$B5,$6B,$FE,$1D,$6A
	.dc.b	$7F,$52,$F1,$EC,$E8,$4A,$FF,$0F,$FD,$63,$E3,$BF,$2E,$D8,$CC,$60
	.dc.b	$FC,$DF,$CD,$BE,$72,$92,$7F,$C5,$31,$97,$FC,$57,$F3,$9F,$C2,$FD
	.dc.b	$AB,$C5,$FE,$2F,$CF,$FC,$C5,$3F,$C7,$FC,$BC,$D7,$F8,$CB,$EA,$9F
	.dc.b	$25,$FC,$D9,$B7,$98,$C1,$FC,$EB,$E6,$DF,$39,$49,$3F,$E3,$C8,$CB
	.dc.b	$7E,$03,$81,$1C,$1F,$E4,$2F,$94,$FD,$B3,$7C,$5F,$E8,$F1,$88,$FF
	.dc.b	$26,$B5,$FF,$3E,$7C,$CF,$FB,$03,$E7,$35,$E8,$FF,$41,$6F,$0F,$FB
	.dc.b	$B6,$7F,$D8,$A3,$E1,$3F,$A9,$B6,$FF,$CA,$6D,$7F,$CB,$6D,$7B,$C7
	.dc.b	$FB,$8A,$39,$BF,$EE,$4F,$9A,$7C,$AF,$FB,$9B,$E6,$9F,$09,$7E,$12
	.dc.b	$63,$FE,$D8,$9C,$B6,$72,$FF,$B0,$61,$3C,$E7,$FD,$E3,$F3,$5F,$52
	.dc.b	$BF,$D8,$CD,$D7,$FB,$03,$74,$FF,$39,$4A,$3C,$AF,$FB,$A1,$D3,$FE
	.dc.b	$77,$6B,$BE,$3E,$F9,$7F,$C0,$0F,$9A,$FF,$52,$61,$FD,$B5,$FF,$4E
	.dc.b	$57,$FB,$5D,$B1,$9F,$90,$E0,$7F,$80,$23,$FD,$4F,$3F,$E8,$56,$A6
	.dc.b	$9D,$B1,$F4,$B6,$38,$4F,$B2,$BF,$BF,$7C,$D7,$B2,$7F,$DD,$1F,$39
	.dc.b	$FC,$ED,$FF,$49,$4F,$F8,$0E,$3E,$7C,$6B,$4E,$58,$FC,$69,$E1,$3C
	.dc.b	$BD,$2C,$BF,$E9,$B8,$E3,$7C,$EF,$32,$9E,$33,$FE,$11,$F9,$CF,$DF
	.dc.b	$4F,$8B,$0F,$A7,$18,$8F,$F7,$F3,$67,$86,$BD,$23,$15,$C3,$C1,$C1
	.dc.b	$99,$F0,$F3,$7E,$1B,$0E,$C7,$FA,$C5,$F3,$D7,$E0,$9B,$19,$E1,$E2
	.dc.b	$7E,$B8,$8F,$EC,$DA,$B1,$FF,$16,$4F,$78,$C5,$75,$AF,$1F,$B7,$FA
	.dc.b	$13,$F4,$5E,$5C,$AF,$D0,$46,$5F,$D1,$46,$5F,$D0,$47,$E7,$BE,$6E
	.dc.b	$47,$E5,$DF,$D0,$FA,$D8,$8F,$D0,$C5,$CB,$F4,$33,$F1,$3E,$73,$E2
	.dc.b	$F1,$4B,$F8,$A3,$44,$EE,$3F,$89,$3F,$C4,$08,$E0,$E3,$1C,$1B,$1F
	.dc.b	$4B,$E4,$3F,$07,$C3,$F1,$71,$FC,$86,$78,$33,$AC,$C3,$C7,$49,$D4
	.dc.b	$62,$7F,$05,$CD,$23,$D3,$C5,$F9,$FC,$3F,$1D,$B6,$67,$F1,$75,$67
	.dc.b	$31,$F0,$3E,$5B,$FD,$42,$7F,$AA,$63,$FC,$CC,$DE,$1F,$E0,$EE,$BE
	.dc.b	$0F,$F3,$AE,$6B,$FD,$F8,$D4,$ED,$ED,$E5,$F0,$38,$CC,$8C,$90,$FC
	.dc.b	$49,$D1,$D3,$83,$AC,$73,$7D,$51,$AB,$F4,$CE,$67,$85,$78,$DB,$19
	.dc.b	$9F,$7D,$71,$BF,$7C,$E5,$F8,$31,$9A,$F7,$CE,$5F,$31,$E2,$3D,$55
	.dc.b	$E2,$70,$78,$A7,$83,$EA,$33,$1E,$27,$8D,$78,$9C,$BE,$27,$27,$88
	.dc.b	$C9,$E2,$3D,$A6,$8F,$A7,$3E,$97,$C0,$68,$F2,$1C,$8C,$CF,$A5,$C9
	.dc.b	$1C,$12,$7C,$91,$97,$D0,$F5,$8F,$63,$93,$D0,$66,$38,$C6,$8E,$91
	.dc.b	$A7,$11,$A3,$AB,$A8,$F6,$9E,$9B,$FC,$91,$2B,$E0,$38,$BB,$1F,$4A
	.dc.b	$35,$3F,$58,$FE,$CA,$74,$29,$4C,$A8,$FF,$6C,$B6,$36,$6C,$F8,$A7
	.dc.b	$08,$D9,$E2,$7F,$65,$F0,$75,$F8,$C7,$EA,$9C,$5F,$C2,$8F,$D7,$C6
	.dc.b	$30,$F8,$91,$83,$E1,$57,$C0,$9D,$47,$E1,$6B,$F7,$B1,$88,$F7,$46
	.dc.b	$1E,$27,$CB,$8D,$5F,$F0,$66,$D7,$F0,$26,$CF,$A8,$E2,$7A,$A7,$FD
	.dc.b	$C6,$E2,$BD,$4F,$B4,$C5,$FA,$9C,$1C,$6F,$D4,$E8,$F5,$47,$F1,$4B
	.dc.b	$F0,$C7,$EF,$9B,$5E,$B8,$E4,$7A,$CE,$47,$C0,$9E,$4E,$2F,$DC,$DB
	.dc.b	$F7,$34,$F7,$1D,$0F,$75,$FF,$80,$3E,$71,$E8,$D2,$4F,$F4,$E8,$FF
	.dc.b	$25,$62,$C7,$FC,$67,$1C,$27,$55,$C1,$FE,$41,$1C,$27,$FC,$73,$F3
	.dc.b	$8E,$6F,$F0,$7A,$F5,$29,$88,$97,$FD,$7A,$74,$D9,$CB,$FC,$9B,$E7
	.dc.b	$3F,$9E,$FC,$CC,$8F,$E8,$D1,$FD,$00,$FC,$F9,$1F,$CF,$EC,$8C,$47
	.dc.b	$33,$89,$FD,$3F,$E5,$FF,$C4,$BF,$38,$E8,$A4,$9F,$EE,$70,$F7,$31
	.dc.b	$1F,$D0,$63,$83,$A9,$E7,$3F,$E3,$0F,$9C,$F9,$53,$FD,$1D,$F9,$08
	.dc.b	$38,$6F,$07,$13,$97,$91,$23,$FC,$0B,$3D,$38,$BF,$B2,$FB,$5F,$D8
	.dc.b	$BE,$72,$57,$F6,$B8,$FE,$CA,$7F,$04,$23,$F1,$9A,$FF,$54,$76,$FF
	.dc.b	$73,$C3,$FC,$93,$F2,$4F,$B2,$A4,$9F,$F0,$BF,$D5,$FE,$25,$FF,$0C
	.dc.b	$CE,$DB,$3B,$7F,$B4,$4E,$DB,$CB,$FE,$56,$8F,$C1,$C6,$B9,$31,$32
	.dc.b	$3B,$C6,$CC,$3F,$DC,$FD,$1F,$F7,$5B,$FE,$2A,$5F,$1D,$FE,$62,$6C
	.dc.b	$67,$DE,$C1,$FC,$99,$47,$27,$F9,$7B,$5F,$E6,$6D,$7F,$D3,$2E,$F6
	.dc.b	$F6,$37,$A7,$35,$E1,$FB,$AF,$F3,$9F,$D5,$FF,$35,$C3,$FC,$14,$DC
	.dc.b	$3E,$23,$6F,$FD,$4A,$FF,$18,$9D,$F7,$72,$23,$FE,$48,$8D,$57,$2E
	.dc.b	$4E,$FC,$92,$B5,$1E,$23,$0F,$8A,$30,$FF,$94,$BE,$9F,$F4,$0F,$F9
	.dc.b	$53,$EA,$FF,$DC,$EF,$F8,$85,$B7,$F6,$E7,$73,$FE,$25,$6B,$FD,$69
	.dc.b	$AF,$FB,$01,$AF,$E8,$DA,$9E,$71,$D7,$9F,$FD,$37,$18,$CF,$FD,$93
	.dc.b	$7F,$C4,$6F,$FC,$D3,$F4,$5F,$EB,$82,$7A,$3F,$75,$C7,$7D,$24,$FF
	.dc.b	$66,$8F,$F6,$3B,$FD,$C3,$AD,$EF,$FE,$15,$B1,$FD,$B1,$AF,$F7,$46
	.dc.b	$BC,$7E,$87,$81,$1E,$0F,$F7,$6D,$E4,$7F,$18,$6B,$FD,$FB,$E7,$9E
	.dc.b	$87,$E5,$5B,$1F,$DF,$1A,$EE,$3C,$DF,$25,$23,$CC,$D7,$CD,$F4,$3E
	.dc.b	$B2,$3E,$B7,$B7,$D6,$48,$EC,$D7,$C9,$F4,$5E,$32,$3E,$37,$CD,$F4
	.dc.b	$52,$3C,$CD,$7D,$9D,$3E,$B2,$3E,$B7,$DD,$D5,$23,$DC,$D7,$E2,$76
	.dc.b	$7C,$32,$3F,$0D,$F8,$DE,$AC,$8F,$8C,$D7,$E6,$7C,$FF,$CB,$23,$F2
	.dc.b	$DF,$9D,$F3,$D2,$3E,$73,$5F,$A9,$EB,$7D,$32,$3F,$4D,$FA,$DE,$DC
	.dc.b	$8F,$AC,$D7,$EE,$7B,$7F,$6C,$8F,$DB,$7E,$F7,$2C,$8F,$BC,$D7,$F0
	.dc.b	$9E,$BF,$FB,$F3,$93,$FD,$75,$D9,$FD,$2E,$51,$87,$FA,$73,$FE,$B9
	.dc.b	$7F,$D9,$AD,$79,$9F,$8B,$9F,$C7,$1F,$C8,$FE,$89,$F3,$9D,$7C,$1F
	.dc.b	$E7,$11,$8A,$FF,$6B,$B5,$FF,$52,$2D,$B8,$63,$E5,$8C,$4F,$09,$CB
	.dc.b	$FD,$8E,$32,$FF,$AB,$7E,$8B,$FB,$6B,$FE,$AF,$F9,$9F,$F5,$E1,$3D
	.dc.b	$7E,$30,$8F,$E5,$0F,$F6,$D6,$B9,$F1,$F7,$18,$AF,$E1,$53,$FD,$CF
	.dc.b	$E8,$1F,$45,$D1,$B1,$FE,$F4,$6B,$FE,$C1,$7F,$9F,$3C,$79,$F8,$11
	.dc.b	$E1,$CF,$CC,$8E,$0F,$EC,$E7,$7D,$E3,$DD,$F1,$6B,$23,$C4,$D7,$C5
	.dc.b	$CC,$E0,$FF,$06,$7D,$17,$C3,$9F,$64,$E5,$FF,$44,$E4,$8C,$47,$FA
	.dc.b	$3A,$31,$5F,$E8,$86,$C7,$0F,$89,$7D,$16,$5B,$19,$8C,$1F,$DA,$D7
	.dc.b	$D1,$52,$3F,$DB,$DF,$EE,$1F,$21,$FE,$09,$C3,$FC,$09,$2E,$3C,$72
	.dc.b	$FF,$A8,$AA,$CF,$E7,$E6,$BF,$E2,$B6,$F3,$3E,$63,$33,$CE,$FF,$C6
	.dc.b	$18,$7F,$84,$6B,$51,$83,$FD,$5F,$CD,$C3,$DF,$E2,$D8,$D5,$FE,$BE
	.dc.b	$7A,$FD,$06,$63,$E9,$4F,$FA,$E2,$32,$F0,$85,$FF,$4C,$8C,$42,$7D
	.dc.b	$3E,$4F,$B3,$1F,$F0,$CD,$71,$7F,$66,$DE,$5D,$72,$C8,$ED,$8B,$E8
	.dc.b	$B1,$97,$90,$FF,$28,$46,$67,$C9,$F4,$4E,$0F,$96,$FC,$7A,$E0,$1D
	.dc.b	$4F,$0E,$0D,$9F,$43,$C8,$B1,$E2,$7E,$AF,$D1,$7E,$D6,$3F,$DA,$6D
	.dc.b	$BF,$F6,$DB,$5F,$F6,$FB,$5F,$E3,$9F,$85,$FF,$A5,$67,$FD,$B6,$D8
	.dc.b	$FF,$71,$B5,$FF,$73,$B5,$FE,$49,$EC,$3C,$4F,$C6,$C7,$F9,$C6,$31
	.dc.b	$5F,$EE,$66,$C7,$FB,$B9,$AF,$FB,$CB,$E8,$9F,$47,$CA,$11,$8A,$FF
	.dc.b	$77,$B6,$3F,$DE,$ED,$7F,$DF,$7F,$69,$F9,$07,$FD,$F6,$D7,$3D,$1D
	.dc.b	$63,$11,$FC,$F7,$E9,$1F,$47,$E1,$6E,$1F,$EF,$E6,$CF,$F4,$33,$FD
	.dc.b	$1C,$E1,$C7,$D1,$FF,$A6,$E7,$51,$FE,$93,$8C,$38,$6B,$C3,$DA,$F8
	.dc.b	$1E,$9B,$EA,$29,$2F,$9D,$7F,$86,$23,$E3,$7D,$27,$E7,$E3,$BB,$5F
	.dc.b	$33,$5F,$EB,$4D,$7F,$D4,$DF,$01,$FE,$22,$AE,$AD,$8F,$4B,$5F,$80
	.dc.b	$D7,$FB,$27,$DA,$3D,$8F,$B0,$DA,$F3,$B6,$7E,$1B,$5F,$95,$F0,$DF
	.dc.b	$EE,$A9,$F6,$B6,$3E,$4B,$5D,$FC,$37,$E1,$7E,$95,$E6,$D2,$5F,$C6
	.dc.b	$8C,$BF,$23,$EC,$FF,$41,$7F,$EC,$56,$A6,$98,$FA,$6F,$DB,$47,$DF
	.dc.b	$6B,$FE,$03,$6B,$FE,$CF,$6A,$FA,$67,$CC,$FE,$6D,$B7,$F6,$9B,$3F
	.dc.b	$86,$6B,$F8,$DF,$0F,$F4,$AF,$F8,$45,$AF,$FB,$75,$AF,$F1,$3E,$CF
	.dc.b	$D4,$7D,$27,$F9,$B3,$E9,$3F,$0E,$D9,$FC,$9B,$FC,$E7,$C4,$CF,$92
	.dc.b	$AF,$C6,$FD,$23,$E7,$3F,$D7,$0D,$C3,$31,$83,$FD,$79,$DE,$F7,$A9
	.dc.b	$27,$F2,$F1,$97,$F3,$3F,$47,$FE,$C2,$6F,$C6,$FF,$B1,$63,$11,$FA
	.dc.b	$86,$BF,$AF,$F9,$D7,$AB,$FE,$CB,$6C,$7E,$91,$AF,$EC,$1A,$FE,$DF
	.dc.b	$E8,$9F,$4D,$FE,$D3,$6C,$7E,$B1,$AF,$EE,$1A,$F8,$7E,$7B,$FC,$47
	.dc.b	$4F,$83,$A6,$22,$4F,$8D,$F3,$11,$C1,$FD,$5E,$7F,$C9,$49,$A7,$F9
	.dc.b	$27,$D4,$FF,$7D,$7D,$15,$E8,$FF,$29,$3F,$E9,$48,$CF,$8B,$FE,$36
	.dc.b	$8D,$58,$38,$6C,$6C,$E1,$E9,$FF,$8F,$BE,$89,$F3,$F4,$95,$FD,$B6
	.dc.b	$3F,$D8,$4F,$F5,$87,$6E,$63,$6E,$7D,$0F,$8F,$F0,$FF,$23,$96,$4F
	.dc.b	$C1,$8C,$BF,$65,$FB,$84,$70,$6C,$87,$F6,$8D,$7F,$CB,$3A,$35,$FF
	.dc.b	$3B,$7D,$0F,$CC,$3E,$01,$37,$FF,$2D,$B9,$D6,$98,$8F,$F2,$FF,$D0
	.dc.b	$BE,$86,$93,$2F,$F0,$74,$7C,$07,$FB,$CB,$B7,$31,$B7,$3B,$7F,$9A
	.dc.b	$BE,$7F,$FC,$CE,$7F,$8C,$63,$FC,$30,$D7,$FC,$06,$EB,$DF,$FE,$7F
	.dc.b	$08,$46,$1B,$1F,$D1,$1A,$FF,$50,$79,$3D,$A3,$C5,$EF,$FF,$9F,$3E
	.dc.b	$85,$E5,$D2,$57,$F8,$AE,$3F,$9E,$36,$32,$FF,$86,$23,$8B,$BF,$3B
	.dc.b	$FD,$0F,$F4,$3F,$86,$E1,$FA,$93,$F9,$38,$7A,$1A,$37,$CD,$27,$FC
	.dc.b	$75,$3A,$6C,$E5,$F8,$BF,$43,$F2,$FE,$5A,$47,$CA,$8F,$9C,$7C,$52
	.dc.b	$3F,$15,$AF,$23,$81,$F4,$BE,$87,$FD,$3F,$F4,$3F,$8C,$E7,$FF,$60
	.dc.b	$36,$7F,$58,$D4,$FF,$A7,$9B,$F2,$3C,$88,$FF,$95,$63,$5F,$39,$F2
	.dc.b	$23,$FC,$82,$FC,$E6,$ED,$FE,$B0,$8D,$6D,$F2,$A5,$18,$C3,$FD,$A8
	.dc.b	$F1,$9D,$36,$33,$C1,$F9,$1F,$57,$F2,$5F,$43,$30,$FC,$8C,$65,$FC
	.dc.b	$9C,$65,$FC,$9C,$65,$FC,$AC,$67,$0F,$27,$2A,$B5,$18,$9D,$61,$FE
	.dc.b	$E9,$AD,$5F,$80,$EE,$43,$B9,$93,$B9,$9A,$EF,$7D,$AB,$A5,$F6,$AF
	.dc.b	$E4,$17,$D9,$E6,$72,$25,$F2,$7B,$3E,$77,$B5,$79,$EF,$B4,$74,$BE
	.dc.b	$CF,$EC,$E3,$2F,$ED,$63,$37,$E1,$BE,$D3,$E0,$BF,$3B,$E0,$7B,$3E
	.dc.b	$03,$B1,$93,$D7,$7D,$2F,$BD,$F5,$AE,$75,$D4,$D9,$C4,$F2,$1D,$4F
	.dc.b	$41,$98,$4F,$A2,$BA,$B6,$74,$D7,$E5,$DF,$C7,$7C,$07,$43,$37,$E4
	.dc.b	$8D,$1F,$08,$8E,$49,$1E,$83,$89,$F0,$8E,$B7,$EF,$75,$3D,$27,$8C
	.dc.b	$78,$63,$E7,$BF,$E7,$59,$FC,$13,$6F,$F9,$41,$1F,$E5,$0D,$7F,$94
	.dc.b	$36,$7C,$F1,$E0,$AF,$CD,$61,$F4,$AB,$E9,$B6,$7E,$63,$5F,$59,$FC
	.dc.b	$E5,$C5,$74,$8C,$1F,$4E,$79,$B4,$E6,$53,$9D,$FF,$01,$7C,$84,$3E
	.dc.b	$15,$F9,$9E,$15,$DE,$78,$46,$8E,$0F,$CD,$7C,$67,$6A,$F1,$B4,$F1
	.dc.b	$94,$F1,$DF,$B2,$FF,$37,$1F,$D6,$A3,$0F,$66,$BE,$13,$BD,$F9,$6B
	.dc.b	$BB,$63,$BB,$5E,$E7,$B2,$BB,$C7,$F1,$2A,$EB,$7E,$B3,$DE,$5B,$EB
	.dc.b	$5C,$09,$7E,$53,$FC,$51,$7F,$CB,$FE,$85,$F4,$3F,$E2,$B9,$FF,$7B
	.dc.b	$C6,$1D,$D6,$63,$11,$FE,$FC,$FA,$1F,$E7,$17,$CE,$F9,$36,$78,$9F
	.dc.b	$E3,$7A,$E3,$1F,$CF,$EB,$8D,$F8,$5F,$ED,$8D,$E3,$E6,$DB,$E6,$27
	.dc.b	$6F,$F9,$0D,$AE,$67,$8B,$96,$BC,$79,$6F,$95,$FE,$AE,$8E,$AB,$E2
	.dc.b	$B2,$6B,$DA,$93,$FE,$D6,$9E,$0F,$FB,$89,$AF,$FB,$A5,$AF,$CD,$77
	.dc.b	$97,$F5,$F6,$E5,$9A,$F1,$FC,$BF,$9B,$E8,$7F,$C3,$F7,$FD,$95,$B5
	.dc.b	$F5,$DB,$87,$F9,$8D,$76,$71,$ED,$93,$F2,$27,$69,$DB,$9E,$6F,$07
	.dc.b	$34,$8C,$10,$63,$E3,$8D,$46,$31,$F2,$57,$CE,$FA,$1F,$F0,$47,$D0
	.dc.b	$BE,$97,$E8,$B6,$FE,$DC,$FF,$83,$47,$C1,$FF,$91,$1B,$1F,$95,$3F
	.dc.b	$2C,$E1,$C0,$70,$ED,$E9,$1E,$33,$C3,$C5,$F8,$2F,$B0,$7C,$F5,$25
	.dc.b	$FE,$3A,$1F,$3D,$88,$F7,$1F,$E9,$18,$E2,$EE,$7A,$4F,$5F,$3F,$B7
	.dc.b	$CF,$CA,$E9,$19,$7B,$46,$5E,$A2,$FF,$D4,$5B,$7E,$D7,$D4,$7C,$EF
	.dc.b	$E6,$97,$D1,$D3,$11,$2F,$E2,$9F,$EC,$47,$FC,$DF,$3C,$A3,$2D,$7E
	.dc.b	$5B,$78,$FF,$C6,$8D,$BF,$15,$78,$78,$FF,$57,$F6,$0F,$7A,$93,$1F
	.dc.b	$DB,$C3,$EB,$F1,$1E,$E8,$E0,$EA,$79,$CF,$4F,$A4,$EB,$DB,$2B,$9C
	.dc.b	$65,$EB,$19,$7A,$47,$EF,$BE,$C3,$FD,$CF,$AF,$F9,$35,$7C,$26,$4D
	.dc.b	$7B,$B2,$FE,$3C,$FF,$BB,$1F,$E4,$8D,$7F,$DA,$0D,$4F,$6D,$B7,$3C
	.dc.b	$E7,$ED,$CF,$F9,$A7,$E9,$9E,$7F,$FA,$81,$B3,$F5,$9A,$FE,$71,$AF
	.dc.b	$E8,$7C,$DF,$8E,$FE,$81,$AE,$76,$FF,$55,$C6,$23,$F9,$E2,$F8,$1C
	.dc.b	$92,$3F,$0B,$19,$7F,$D5,$DD,$FF,$EB,$12,$38,$3F,$17,$F4,$AE,$9F
	.dc.b	$F5,$CB,$63,$31,$83,$FD,$7B,$F5,$8F,$96,$A4,$9F,$CA,$C6,$5F,$1F
	.dc.b	$D3,$FA,$88,$E0,$F0,$F6,$78,$08,$E0,$F2,$F8,$1E,$42,$38,$3D,$3D
	.dc.b	$DE,$82,$38,$3D,$BE,$3F,$B0,$8E,$0F,$D5,$FE,$E5,$FE,$D8,$4D,$23
	.dc.b	$F1,$3C,$A8,$FC,$86,$BF,$C6,$35,$7C,$07,$CA,$24,$7C,$E7,$FD,$CB
	.dc.b	$4F,$89,$A4,$9F,$1B,$F4,$BE,$8E,$47,$D6,$6B,$DF,$E0,$BE,$B9,$1F
	.dc.b	$B6,$F6,$F9,$D9,$1E,$E6,$BE,$7F,$86,$FB,$E4,$7F,$06,$FE,$0B,$96
	.dc.b	$47,$E1,$9A,$FF,$36,$C7,$FD,$F2,$FF,$3B,$6B,$FE,$11,$6A,$3E,$6F
	.dc.b	$9D,$F8,$EF,$82,$91,$F9,$26,$BF,$2D,$FE,$FB,$A9,$A7,$F8,$02,$BF
	.dc.b	$AF,$CF,$F2,$EF,$A8,$7D,$07,$F8,$11,$B1,$FE,$0E,$6B,$BF,$A6,$D1
	.dc.b	$1D,$3F,$E0,$CF,$A8,$FF,$74,$79,$4D,$BF,$C1,$ED,$AF,$F0,$CB,$5F
	.dc.b	$E7,$EF,$F7,$42,$69,$C0,$E9,$CD,$FE,$18,$FA,$67,$D4,$52,$5F,$F8
	.dc.b	$92,$3F,$C1,$6F,$C9,$6A,$69,$39,$BA,$12,$39,$3F,$E2,$6F,$6B,$FB
	.dc.b	$87,$91,$FE,$F5,$AF,$86,$DD,$FC,$87,$97,$69,$87,$8D,$AF,$F5,$CA
	.dc.b	$73,$7F,$82,$A1,$F5,$1F,$D4,$67,$38,$7F,$58,$6C,$FF,$7E,$6B,$FE
	.dc.b	$03,$6B,$FE,$59,$5E,$3B,$EA,$39,$36,$79,$35,$E4,$D7,$FC,$17,$F5
	.dc.b	$1F,$BD,$23,$FD,$A8,$37,$AD,$E5,$7F,$95,$27,$F0,$0D,$8C,$C7,$F7
	.dc.b	$C3,$25,$E4,$CE,$5A,$8F,$CC,$7C,$96,$2F,$FC,$41,$56,$7F,$49,$D2
	.dc.b	$AC,$47,$E9,$F9,$6B,$FE,$42,$BF,$24,$63,$DB,$91,$E7,$FA,$8B,$5E
	.dc.b	$57,$89,$D3,$0F,$EF,$93,$B3,$FC,$E2,$E2,$7F,$C2,$F9,$7F,$99,$E7
	.dc.b	$FC,$EA,$FE,$B1,$BC,$BE,$68,$C4,$F0,$E4,$ED,$F2,$F9,$6D,$FF,$9F
	.dc.b	$BE,$A3,$F9,$03,$76,$F7,$C6,$2F,$FC,$4E,$FC,$CC,$3E,$53,$6B,$FD
	.dc.b	$39,$93,$EA,$38,$E1,$ED,$7E,$4F,$D4,$7F,$8A,$AB,$FC,$4E,$DC,$7F
	.dc.b	$C9,$AD,$7F,$CA,$4D,$7F,$D5,$0F,$F7,$C8,$E3,$E8,$7D,$B9,$CB,$5F
	.dc.b	$E3,$93,$B7,$FC,$B3,$EE,$3C,$4D,$42,$31,$7F,$E3,$A6,$C7,$09,$DC
	.dc.b	$3E,$46,$92,$B8,$3A,$CE,$47,$36,$DF,$F9,$AB,$E9,$1F,$3B,$E2,$6D
	.dc.b	$7F,$92,$9B,$3B,$9D,$7A,$52,$B4,$D9,$E9,$F2,$5C,$C9,$3C,$DF,$7F
	.dc.b	$D1,$7F,$91,$63,$FC,$AE,$D9,$FF,$3E,$B5,$FF,$41,$35,$FF,$6C,$3F
	.dc.b	$E8,$D8,$E3,$F0,$9D,$08,$F4,$FA,$DE,$E4,$70,$7F,$99,$BE,$2D,$EB
	.dc.b	$FD,$26,$C6,$63,$07,$D6,$FC,$25,$DB,$49,$3F,$E7,$08,$CB,$6B,$81
	.dc.b	$1C,$1F,$E7,$6F,$B0,$FB,$4D,$E1,$FC,$1C,$63,$4F,$F4,$FB,$5F,$F5
	.dc.b	$27,$D7,$BE,$13,$F1,$2D,$BF,$F4,$0B,$67,$FD,$4E,$D7,$FD,$59,$F4
	.dc.b	$3F,$E7,$9F,$A1,$FF,$43,$B6,$3F,$D5,$AF,$F0,$FF,$13,$3E,$F2,$BF
	.dc.b	$D1,$BF,$8B,$5B,$B1,$FC,$9E,$31,$1F,$EB,$66,$BF,$EB,$B7,$FD,$52
	.dc.b	$E7,$E2,$B8,$11,$EB,$ED,$F3,$23,$83,$FD,$3B,$F6,$3F,$05,$BA,$FF
	.dc.b	$37,$8C,$47,$FB,$19,$AF,$FB,$2D,$7D,$9E,$C9,$1D,$1E,$5F,$27,$23
	.dc.b	$BB,$5F,$F6,$8F,$D9,$7F,$A3,$1F,$F0,$C7,$8D,$FC,$EE,$3F,$79,$18
	.dc.b	$E1,$FF,$0E,$28,$CB,$FE,$C0,$79,$37,$E1,$FF,$A8,$46,$23,$FD,$C4
	.dc.b	$D7,$9B,$5F,$F1,$43,$CA,$BC,$53,$E0,$6C,$F8,$CF,$F1,$77,$D9,$3E
	.dc.b	$7B,$58,$79,$BE,$8D,$F0,$98,$BF,$F1,$A4,$61,$E4,$1C,$3F,$ED,$A5
	.dc.b	$F0,$9C,$63,$53,$67,$78,$FF,$61,$6D,$F8,$6F,$87,$D1,$BE,$5E,$FA
	.dc.b	$13,$2F,$0C,$E5,$B3,$97,$FD,$D7,$09,$E7,$3E,$3F,$B2,$7C,$FD,$F9
	.dc.b	$F9,$87,$B2,$72,$D9,$FF,$01,$46,$23,$ED,$9F,$E5,$77,$0E,$1A,$FE
	.dc.b	$01,$E1,$A7,$F7,$F6,$E7,$9A,$F8,$DF,$13,$F4,$EF,$E6,$FD,$1F,$FA
	.dc.b	$B5,$61,$FE,$07,$8C,$37,$E3,$BE,$74,$7C,$A9,$FF,$35,$3A,$C3,$F2
	.dc.b	$2D,$8C,$CF,$0E,$0F,$AD,$F4,$7F,$D4,$67,$EE,$BF,$35,$F0,$DF,$E6
	.dc.b	$A3,$18,$7F,$86,$5E,$13,$A3,$FC,$F4,$E2,$35,$AF,$63,$BF,$A1,$F8
	.dc.b	$5F,$A9,$FE,$57,$7F,$8C,$E7,$FE,$B8,$DF,$17,$5E,$44,$8F,$A2,$7F
	.dc.b	$A1,$EB,$FD,$12,$9A,$48,$E1,$1E,$53,$4F,$E5,$7A,$3C,$DA,$7E,$87
	.dc.b	$83,$F2,$C4,$AF,$CF,$F3,$BE,$37,$11,$F9,$F9,$BB,$7E,$6E,$7F,$33
	.dc.b	$A7,$E6,$BE,$4E,$5F,$E6,$23,$2F,$E6,$A3,$2F,$E6,$23,$F2,$DF,$23
	.dc.b	$F9,$7F,$BA,$93,$F9,$58,$CB,$F0,$87,$3F,$FB,$AD,$F8,$AD,$7F,$29
	.dc.b	$1F,$9A,$38,$BF,$0E,$38,$63,$FE,$53,$CB,$E1,$CF,$E2,$BA,$FF,$1A
	.dc.b	$47,$F2,$B1,$5F,$99,$CF,$FC,$01,$FC,$CC,$FE,$52,$7F,$1B,$D5,$F8
	.dc.b	$EC,$E5,$7E,$3A,$32,$FE,$42,$32,$FD,$7E,$6F,$DD,$3F,$C4,$3D,$D7
	.dc.b	$D7,$7F,$0D,$5F,$69,$4C,$44,$9E,$11,$97,$94,$65,$FC,$24,$F2,$76
	.dc.b	$68,$86,$A7,$93,$C2,$79,$69,$FE,$7B,$AE,$53,$B3,$8C,$66,$78,$BB
	.dc.b	$AE,$3C,$7F,$D1,$AF,$DB,$7E,$23,$91,$DC,$E7,$1B,$9E,$6F,$1B,$E7
	.dc.b	$3F,$EE,$BA,$F1,$DF,$78,$E2,$66,$78,$BB,$BE,$2A,$BC,$6F,$8A,$B9
	.dc.b	$3E,$32,$38,$25,$F8,$E3,$B9,$CE,$BB,$BD,$6F,$BD,$F9,$AB,$D9,$3C
	.dc.b	$23,$C3,$7D,$23,$CC,$66,$B9,$CF,$18,$F6,$1C,$6F,$33,$D9,$B1,$93
	.dc.b	$E7,$57,$D5,$70,$7B,$EF,$D4,$FB,$9F,$2B,$93,$B9,$93,$CA,$64,$F7
	.dc.b	$1D,$23,$D4,$E0,$E0,$7C,$53,$C7,$1E,$08,$F2,$47,$F2,$0B,$FA,$6F
	.dc.b	$F5,$26,$BE,$63,$17,$FE,$D9,$8C,$C7,$98,$C9,$FC,$25,$F9,$B5,$DC
	.dc.b	$91,$F8,$9F,$82,$FA,$B5,$F5,$9B,$1F,$55,$AF,$AA,$BF,$DA,$CF,$D5
	.dc.b	$34,$53,$51,$C2,$FD,$A7,$EA,$5C,$5F,$8A,$7F,$25,$18,$7E,$7C,$ED
	.dc.b	$E8,$F2,$9F,$D5,$4E,$E7,$E7,$4E,$DF,$35,$FA,$DE,$35,$DE,$FD,$31
	.dc.b	$A3,$D2,$53,$D2,$70,$3C,$71,$E0,$70,$78,$1C,$3C,$6B,$CC,$FC,$D3
	.dc.b	$73,$F4,$9C,$1B,$9F,$25,$FA,$E7,$53,$F1,$0E,$37,$FC,$8E,$74,$D8
	.dc.b	$C9,$FC,$A6,$3F,$C6,$5F,$25,$23,$FA,$EB,$FE,$35,$F8,$DC,$47,$F8
	.dc.b	$EE,$2E,$3F,$E3,$79,$FF,$1A,$FC,$BF,$F8,$DB,$E0,$25,$FF,$8D,$23
	.dc.b	$2F,$F8,$DA,$32,$FF,$8D,$23,$F9,$C7,$D8,$FF,$39,$F5,$E4,$FF,$3D
	.dc.b	$8C,$BF,$95,$7F,$34,$47,$06,$C8,$7F,$3A,$F7,$3F,$C4,$7D,$52,$3F
	.dc.b	$C5,$6F,$F8,$9B,$3C,$47,$F5,$88,$B9,$FF,$8B,$67,$FC,$47,$D5,$FE
	.dc.b	$24,$CA,$5F,$F8,$92,$32,$FF,$89,$A3,$2F,$F5,$78,$FE,$A9,$E0,$FF
	.dc.b	$55,$F0,$64,$FF,$59,$8C,$BF,$B1,$7F,$6C,$47,$06,$C8,$7F,$56,$E6
	.dc.b	$FF,$36,$7D,$43,$E7,$2F,$9D,$2B,$FC,$EB,$39,$6C,$E5,$FF,$11,$47
	.dc.b	$FA,$9B,$47,$37,$FA,$5A,$38,$4F,$F8,$9F,$EA,$3F,$B8,$B7,$1F,$F3
	.dc.b	$E4,$62,$F8,$C8,$47,$FA,$7D,$FF,$43,$BB,$BF,$F5,$1B,$63,$83,$5E
	.dc.b	$3D,$5F,$DF,$7E,$63,$FB,$F7,$2C,$9F,$F0,$14,$65,$FE,$34,$FF,$1F
	.dc.b	$23,$83,$64,$3F,$C8,$5F,$80,$7F,$B5,$B1,$F4,$C7,$C0,$3F,$C8,$DA
	.dc.b	$3D,$7F,$F4,$D3,$63,$FD,$42,$FF,$A0,$A7,$8B,$73,$FF,$5F,$BB,$9E
	.dc.b	$75,$C9,$E7,$11,$E6,$F6,$DB,$11,$E2,$8B,$7D,$A7,$FD,$59,$87,$F7
	.dc.b	$5A,$FF,$59,$AF,$A4,$C9,$AE,$C9,$3F,$E4,$79,$E0,$FF,$9A,$9A,$FF
	.dc.b	$A7,$1A,$FF,$B6,$1D,$DF,$C6,$6E,$59,$AF,$1F,$C5,$7F,$AF,$FE,$73
	.dc.b	$FD,$43,$87,$E0,$DA,$9A,$4F,$3F,$FC,$B5,$19,$6C,$1A,$63,$C6,$93
	.dc.b	$FE,$61,$9F,$F2,$F3,$FE,$EE,$6B,$F9,$3E,$97,$4F,$66,$C7,$F9,$A5
	.dc.b	$AF,$FA,$31,$AF,$FA,$F9,$71,$71,$24,$71,$7A,$71,$48,$E8,$D7,$FD
	.dc.b	$2F,$F3,$8F,$23,$DE,$D8,$FF,$3D,$35,$DC,$7C,$4D,$DF,$29,$49,$3F
	.dc.b	$6D,$D7,$83,$23,$EF,$35,$FF,$53,$FC,$DB,$E3,$FE,$7B,$63,$FD,$14
	.dc.b	$D7,$71,$AF,$99,$91,$A6,$BD,$3E,$83,$99,$1E,$6F,$6F,$8F,$91,$D9
	.dc.b	$AF,$93,$D6,$F1,$91,$F1,$BE,$6E,$49,$1E,$66,$BE,$CE,$4F,$59,$1F
	.dc.b	$5B,$EE,$DA,$47,$B9,$AF,$C4,$FB,$9F,$86,$47,$E1,$BF,$1B,$EA,$64
	.dc.b	$7C,$66,$BF,$33,$E1,$FE,$59,$1F,$96,$FC,$EF,$A9,$91,$F3,$9A,$FD
	.dc.b	$4F,$8A,$FA,$64,$7E,$9B,$F5,$BC,$0F,$E5,$91,$FE,$D9,$E1,$4E,$3B
	.dc.b	$BF,$8C,$C6,$27,$F9,$94,$62,$B5,$DD,$CB,$E7,$DD,$D8,$BF,$E7,$11
	.dc.b	$87,$8C,$EE,$75,$09,$7C,$E6,$1B,$1C,$0E,$71,$93,$8F,$CF,$BB,$B9
	.dc.b	$F3,$BE,$5A,$92,$F9,$08,$7B,$7C,$09,$1A,$8C,$BC,$2D,$D7,$E8,$F9
	.dc.b	$C7,$CF,$D2,$57,$68,$FE,$50,$DF,$1B,$81,$1D,$78,$DC,$C8,$E0,$F3
	.dc.b	$FC,$FF,$F4,$17,$FB,$15,$7C,$B5,$C7,$FD,$3A,$30,$DE,$1F,$37,$04
	.dc.b	$AF,$23,$5E,$3C,$5F,$82,$BF,$E0,$77,$FD,$BF,$E7,$FE,$35,$78,$3C
	.dc.b	$2E,$EF,$E2,$0C,$4D,$69,$6B,$9B,$7E,$79,$F3,$FB,$84,$A7,$31,$B9
	.dc.b	$D4,$E5,$FE,$77,$0F,$9F,$A4,$7F,$9F,$BF,$D0,$3C,$DF,$F0,$4E,$1F
	.dc.b	$DC,$A7,$F0,$99,$FE,$09,$FE,$E4,$DE,$6D,$BE,$D6,$C7,$F8,$47,$E6
	.dc.b	$DE,$D6,$23,$FC,$29,$36,$73,$5D,$7A,$35,$4F,$9F,$B1,$86,$4F,$8F
	.dc.b	$C9,$B5,$F5,$0F,$DA,$BC,$8E,$DF,$27,$67,$F3,$78,$23,$C9,$9F,$FB
	.dc.b	$B6,$7F,$3E,$FE,$D1,$BC,$DA,$E5,$91,$CB,$D4,$B9,$78,$46,$28,$F8
	.dc.b	$1F,$3F,$F2,$27,$4F,$CC,$6F,$06,$B7,$98,$79,$BD,$0B,$5D,$71,$FE
	.dc.b	$C3,$3F,$B4,$F9,$F7,$9B,$F5,$EB,$FC,$81,$18,$9F,$C4,$36,$3F,$20
	.dc.b	$D7,$F2,$8B,$CD,$E5,$08,$C4,$F2,$9C,$C6,$27,$F3,$5E,$CB,$E2,$78
	.dc.b	$B6,$FF,$30,$D9,$DC,$EB,$19,$5A,$6C,$FE,$97,$ED,$7F,$BB,$4F,$E8
	.dc.b	$5B,$19,$F3,$BC,$F1,$88,$FD,$77,$CF,$BE,$7F,$D6,$DB,$FD,$53,$67
	.dc.b	$F6,$4D,$78,$FC,$FF,$02,$4F,$07,$E2,$FE,$19,$FE,$6D,$8F,$DA,$36
	.dc.b	$7F,$78,$D7,$F8,$03,$5F,$E9,$3F,$40,$FA,$0F,$9E,$DA,$FD,$F3,$67
	.dc.b	$F8,$33,$5F,$E1,$4B,$E1,$B8,$92,$78,$BF,$6B,$E0,$9F,$23,$49,$3F
	.dc.b	$C2,$E3,$2F,$F0,$DF,$A0,$FE,$7A,$DF,$07,$F0,$71,$8C,$7F,$8C,$35
	.dc.b	$FE,$38,$BD,$6D,$92,$BA,$3F,$8C,$F1,$1F,$49,$49,$3F,$C7,$63,$2F
	.dc.b	$F1,$FF,$A0,$7C,$1F,$E5,$1B,$19,$8C,$1F,$98,$FA,$0F,$DE,$7D,$07
	.dc.b	$F2,$76,$E7,$FC,$BD,$FE,$27,$E0,$67,$C6,$57,$F2,$DF,$B0,$FC,$45
	.dc.b	$DE,$8F,$F3,$1C,$62,$3F,$9C,$35,$FE,$79,$F4,$2F,$A1,$FF,$34,$B6
	.dc.b	$3F,$9C,$35,$FE,$80,$D7,$8F,$D0,$F0,$23,$C1,$FF,$39,$FB,$FF,$E6
	.dc.b	$A8,$34,$9F,$05,$FE,$78,$FA,$17,$D0,$D2,$57,$F4,$98,$CB,$7E,$87
	.dc.b	$81,$1C,$1F,$D3,$FF,$09,$FE,$D0,$FF,$59,$6B,$FE,$22,$6B,$FE,$CF
	.dc.b	$6A,$F9,$EF,$F0,$1B,$FC,$97,$E3,$FF,$83,$3F,$E0,$C6,$BF,$CA,$63
	.dc.b	$7E,$97,$FB,$09,$E3,$1F,$DA,$5A,$FF,$8C,$9A,$FF,$B7,$DA,$F8,$7B
	.dc.b	$3F,$C4,$F1,$D3,$9F,$84,$7F,$8A,$BE,$79,$E5,$7F,$8B,$3E,$79,$F3
	.dc.b	$74,$98,$FF,$88,$A1,$F2,$D8,$8F,$E8,$11,$C1,$D4,$F3,$9F,$F1,$BF
	.dc.b	$D0,$7F,$BC,$67,$FA,$32,$F9,$7C,$9A,$DA,$4F,$F8,$C2,$74,$D9,$CB
	.dc.b	$FD,$3F,$E7,$DC,$9F,$E4,$47,$FD,$31,$C1,$FE,$EA,$6E,$FF,$E5,$87
	.dc.b	$75,$FE,$3C,$7F,$18,$4A,$FF,$60,$B6,$35,$1D,$4F,$F2,$8E,$7F,$22
	.dc.b	$3F,$C8,$BF,$3E,$F9,$FF,$F2,$AB,$63,$31,$89,$CB,$23,$27,$28,$FF
	.dc.b	$2E,$FC,$FF,$F6,$98,$FF,$28,$BF,$27,$26,$5F,$E7,$28,$C3,$7D,$AE
	.dc.b	$27,$2D,$A5,$F2,$8F,$F3,$4F,$B8,$F9,$EB,$F3,$F2,$BF,$CB,$F3,$96
	.dc.b	$CF,$FB,$92,$31,$1F,$E3,$96,$BF,$E3,$D6,$BF,$E8,$67,$79,$7A,$20
	.dc.b	$23,$38,$EE,$BC,$BF,$03,$F9,$3F,$8E,$97,$F9,$17,$F2,$DF,$19,$88
	.dc.b	$FC,$8C,$5C,$7B,$43,$9B,$FC,$E2,$FE,$2F,$E6,$BF,$27,$EF,$C8,$C1
	.dc.b	$0F,$CA,$B9,$7F,$2F,$1F,$23,$E7,$FE,$4F,$A9,$27,$E3,$C6,$5F,$F2
	.dc.b	$B6,$C8,$F9,$04,$8C,$BB,$7F,$C5,$11,$B7,$E8,$FC,$FF,$E2,$BA,$E4
	.dc.b	$7E,$39,$FC,$6E,$98,$8F,$F1,$87,$3F,$ED,$1D,$1F,$94,$8F,$C5,$F5
	.dc.b	$7E,$33,$39,$3F,$8C,$8C,$BF,$8E,$8C,$BF,$E3,$BE,$6F,$D8,$BE,$6F
	.dc.b	$53,$F1,$3D,$B2,$3F,$0F,$1F,$8C,$30,$FB,$23,$0E,$8F,$C6,$BA,$E1
	.dc.b	$FA,$FA,$AB,$75,$C4,$D9,$C4,$F1,$18,$34,$60,$93,$D2,$BB,$72,$FE
	.dc.b	$4E,$BA,$CE,$6B,$93,$E2,$8C,$CF,$E6,$6B,$8D,$7E,$62,$FC,$CF,$8E
	.dc.b	$F9,$3E,$27,$93,$9A,$E4,$F6,$7C,$4E,$4E,$95,$D9,$FD,$25,$F8,$9C
	.dc.b	$9C,$4F,$13,$97,$F5,$51,$C6,$FC,$B1,$F0,$AB,$C2,$F9,$5F,$0B,$93
	.dc.b	$A9,$93,$A9,$E5,$3A,$9E,$13,$CA,$66,$3F,$91,$3C,$4C,$9D,$4E,$26
	.dc.b	$5F,$2D,$FC,$68,$C5,$7A,$E7,$35,$E8,$AF,$9A,$6C,$8E,$49,$5E,$83
	.dc.b	$C1,$3E,$49,$E3,$3F,$01,$E3,$3F,$01,$CB,$C5,$CB,$E8,$72,$FB,$1F
	.dc.b	$43,$FC,$7E,$FF,$92,$11,$FE,$02,$D7,$F9,$33,$5E,$07,$C9,$7F,$99
	.dc.b	$F3,$FC,$17,$E1,$36,$3E,$73,$5F,$6B,$67,$CF,$7F,$D1,$E7,$E0,$D7
	.dc.b	$27,$B1,$C8,$A7,$22,$CF,$2A,$E9,$7F,$1C,$FE,$74,$E2,$7E,$34,$61
	.dc.b	$E2,$ED,$FD,$34,$6D,$F9,$11,$B7,$CA,$78,$CA,$78,$CB,$3E,$37,$A9
	.dc.b	$E0,$3A,$98,$3A,$98,$38,$C7,$8D,$FE,$9D,$3E,$3B,$FC,$6C,$6E,$BE
	.dc.b	$4D,$FC,$78,$E2,$77,$38,$9D,$E7,$8B,$83,$67,$88,$C4,$FF,$52,$F9
	.dc.b	$F7,$CA,$7F,$60,$9F,$F7,$34,$7F,$9F,$1F,$EC,$73,$BF,$47,$4F,$1A
	.dc.b	$B9,$71,$BE,$77,$C2,$31,$7A,$8F,$ED,$B3,$D5,$EF,$59,$5F,$3F,$F8
	.dc.b	$7C,$3F,$B3,$B7,$0F,$33,$67,$FB,$BB,$FE,$0E,$6E,$7C,$E3,$51,$D2
	.dc.b	$31,$D9,$FE,$0E,$F9,$EF,$F0,$7F,$44,$AF,$F0,$6C,$65,$FE,$CE,$FF
	.dc.b	$83,$48,$E0,$D9,$06,$3F,$DF,$BE,$77,$FD,$33,$1F,$0C,$FF,$06,$E7
	.dc.b	$F3,$DD,$E7,$F4,$61,$E9,$FC,$36,$CF,$13,$E6,$95,$E5,$7F,$E0,$C7
	.dc.b	$71,$D0,$EF,$3C,$DE,$91,$1E,$4F,$7D,$B1,$1C,$4F,$C2,$C6,$2B,$B4
	.dc.b	$FD,$7C,$3F,$C4,$D5,$E5,$5F,$27,$93,$5C,$12,$7E,$01,$FE,$2E,$9E
	.dc.b	$2E,$1A,$FC,$56,$BF,$95,$77,$7F,$36,$7A,$4F,$9E,$5F,$15,$F8,$BF
	.dc.b	$AC,$78,$57,$C2,$97,$F2,$E7,$2D,$9C,$BF,$06,$3F,$C8,$2F,$F9,$11
	.dc.b	$DB,$9E,$2E,$C4,$7A,$71,$57,$47,$78,$F2,$47,$E7,$FA,$3C,$F3,$F7
	.dc.b	$57,$C9,$E4,$D6,$52,$7E,$48,$9F,$F5,$D9,$23,$31,$97,$83,$F6,$E3
	.dc.b	$BB,$7C,$4F,$D8,$7D,$63,$E3,$AF,$CC,$4A,$FC,$84,$E5,$B3,$97,$F0
	.dc.b	$D0,$9E,$73,$D3,$8E,$57,$27,$BF,$A5,$88,$E3,$17,$8F,$FC,$CB,$87
	.dc.b	$FB,$8A,$7F,$25,$B3,$E9,$BF,$83,$7D,$63,$C7,$A7,$F0,$99,$D5,$7E
	.dc.b	$9D,$AF,$E4,$1A,$FE,$C5,$AF,$F2,$16,$BF,$E8,$27,$8E,$1F,$E7,$86
	.dc.b	$F5,$6F,$0F,$1F,$7F,$F1,$9F,$A3,$7D,$0F,$D2,$6E,$3F,$61,$AF,$E8
	.dc.b	$9A,$FF,$03,$7F,$D8,$2E,$7C,$EE,$44,$7F,$03,$E3,$3E,$0B,$EF,$B6
	.dc.b	$33,$18,$33,$E1,$48,$E2,$D7,$F6,$72,$F0,$B9,$92,$3B,$BF,$8C,$FA
	.dc.b	$87,$C3,$52,$4F,$E5,$23,$2F,$E5,$BC,$07,$CC,$FE,$51,$B1,$98,$C1
	.dc.b	$9F,$42,$47,$16,$BF,$C2,$3F,$02,$F8,$AF,$F2,$96,$BF,$D6,$5A,$FF
	.dc.b	$AE,$9A,$BE,$83,$F2,$CF,$EA,$9A,$FF,$14,$6B,$FC,$D5,$AF,$F6,$55
	.dc.b	$F4,$9F,$AE,$6C,$73,$8C,$19,$EA,$91,$DD,$AF,$7F,$95,$E4,$47,$C8
	.dc.b	$F5,$E3,$91,$E7,$6B,$E7,$EC,$F1,$11,$F6,$3E,$5F,$92,$91,$EF,$6B
	.dc.b	$EF,$FA,$4F,$51,$1F,$88,$FB,$7D,$F9,$1F,$1D,$AF,$C7,$F8,$CF,$84
	.dc.b	$47,$E6,$3F,$17,$E0,$A4,$7C,$F6,$BF,$D0,$BF,$73,$FF,$71,$BF,$D8
	.dc.b	$DA,$FF,$84,$DA,$FF,$BD,$1A,$FD,$BF,$3F,$FC,$11,$BB,$E9,$29,$88
	.dc.b	$93,$F7,$DF,$B0,$47,$F9,$5F,$D2,$7F,$24,$F0,$BF,$DE,$AD,$F0,$7F
	.dc.b	$C4,$8D,$90,$7F,$9B,$BF,$DD,$5A,$FF,$69,$6C,$F4,$F2,$FF,$C3,$DF
	.dc.b	$48,$ED,$FF,$10,$79,$2F,$9C,$A4,$C7,$FC,$01,$0F,$84,$D9,$23,$2F
	.dc.b	$4E,$BF,$F1,$57,$D0,$3E,$92,$92,$BF,$C1,$51,$FD,$D9,$BE,$B7,$02
	.dc.b	$3A,$F5,$B9,$91,$C1,$FE,$13,$FA,$4F,$F0,$8F,$0F,$F9,$26,$FE,$34
	.dc.b	$9A,$4D,$FC,$A4,$8F,$1B,$E8,$F0,$7F,$C5,$30,$FA,$4F,$DA,$D7,$F0
	.dc.b	$AE,$2F,$EC,$2D,$9F,$F0,$5B,$5F,$F1,$63,$5F,$27,$D2,$3E,$93,$8B
	.dc.b	$63,$8B,$5E,$2D,$7F,$C2,$FF,$2F,$FD,$B1,$34,$8F,$F7,$4F,$71,$A7
	.dc.b	$F8,$DE,$3E,$E3,$63,$31,$FE,$2D,$32,$5E,$4C,$E5,$A8,$7C,$FB,$E2
	.dc.b	$33,$87,$F9,$26,$ED,$7E,$52,$B5,$3E,$1E,$5C,$CF,$E6,$3E,$92,$BF
	.dc.b	$8B,$9F,$34,$60,$CF,$B1,$63,$D4,$E1,$FF,$29,$5B,$E5,$F1,$7F,$E4
	.dc.b	$78,$C4,$FF,$95,$F3,$72,$F8,$9F,$F1,$AC,$66,$31,$EE,$48,$EB,$EB
	.dc.b	$DC,$3A,$3E,$AF,$A4,$FF,$32,$CF,$F9,$8D,$F9,$38,$7C,$77,$35,$FE
	.dc.b	$59,$73,$3F,$E9,$D6,$C7,$C7,$F4,$FA,$6F,$F1,$B3,$47,$C8,$FA,$37
	.dc.b	$9B,$FC,$49,$B8,$FF,$8C,$9B,$87,$F9,$71,$AF,$F9,$F7,$EE,$BF,$CA
	.dc.b	$33,$FE,$38,$6C,$7F,$98,$9A,$EF,$E0,$BF,$8F,$FC,$37,$E0,$2B,$FC
	.dc.b	$7E,$D9,$FF,$34,$B5,$FF,$45,$35,$FF,$5D,$7D,$1B,$CF,$FE,$52,$DA
	.dc.b	$FF,$24,$B6,$7F,$CE,$6D,$7F,$D2,$9F,$75,$F6,$9F,$F3,$BB,$5C,$F2
	.dc.b	$7F,$34,$8C,$47,$FB,$27,$E9,$1F,$49,$FC,$E1,$B8,$7F,$96,$1B,$3F
	.dc.b	$E8,$26,$BF,$EA,$3F,$A4,$FE,$5B,$3F,$E5,$E6,$C7,$FA,$21,$AE,$CF
	.dc.b	$E8,$BF,$53,$FA,$07,$FD,$1A,$D7,$FD,$5C,$D7,$FD,$BC,$D7,$F9,$DF
	.dc.b	$D2,$7F,$92,$EB,$FC,$DA,$D8,$FF,$4B,$B5,$FF,$5C,$35,$FF,$74,$7C
	.dc.b	$17,$F8,$92,$7F,$CE,$AD,$8F,$F4,$FB,$5F,$F6,$03,$5F,$F7,$87,$DC
	.dc.b	$FF,$BC,$A7,$FC,$FA,$D8,$FF,$53,$B5,$FF,$64,$35,$FF,$7C,$7D,$2B
	.dc.b	$D6,$FE,$CE,$DA,$FF,$43,$B6,$7F,$D6,$2D,$7F,$DA,$5E,$1F,$F9,$2E
	.dc.b	$7F,$D1,$AD,$8F,$E4,$3F,$05,$FD,$CA,$31,$1F,$ED,$9F,$B8,$FF,$11
	.dc.b	$D7,$FA,$51,$B1,$FE,$BE,$6B,$BF,$7B,$FB,$DF,$D2,$FF,$09,$FA,$5F
	.dc.b	$F4,$DB,$67,$FD,$8A,$FF,$1F,$F2,$33,$CB,$FE,$EA,$9F,$F7,$47,$D2
	.dc.b	$3E,$97,$FD,$3E,$D8,$FF,$52,$B5,$FF,$67,$B5,$FF,$78,$79,$4F,$8E
	.dc.b	$FF,$53,$B6,$3F,$D5,$AD,$7F,$DA,$ED,$7F,$DF,$1F,$48,$FA,$4F,$F5
	.dc.b	$7B,$63,$FD,$6A,$D7,$FD,$BE,$D7,$FE,$00,$7C,$07,$F9,$50,$9F,$17
	.dc.b	$FE,$B8,$7F,$CA,$FF,$77,$FA,$A9,$FF,$5F,$B5,$FF,$74,$B5,$FF,$09
	.dc.b	$9F,$E1,$47,$0E,$3E,$95,$F0,$9F,$EC,$46,$C7,$FB,$29,$AF,$FB,$D1
	.dc.b	$AF,$41,$F4,$BF,$EC,$87,$0D,$7C,$2D,$7F,$C4,$AD,$5F,$85,$7F,$B2
	.dc.b	$5F,$F8,$02,$D7,$B3,$5F,$5B,$5F,$F1,$7F,$A3,$FE,$F2,$7F,$A5,$F8
	.dc.b	$FF,$03,$CF,$FE,$CD,$3F,$EF,$46,$FC,$C6,$7A,$BE,$2C,$A3,$FD,$CA
	.dc.b	$DB,$FF,$75,$B5,$F8,$07,$F5,$88,$D9,$CD,$F1,$B5,$D1,$FE,$44,$AF
	.dc.b	$1F,$0E,$7E,$87,$FC,$7D,$39,$C3,$EB,$C6,$21,$5F,$DF,$E3,$0D,$F8
	.dc.b	$0E,$5F,$05,$2B,$91,$FE,$52,$AD,$DF,$4A,$F2,$7D,$23,$E6,$2F,$A5
	.dc.b	$30,$E2,$27,$4E,$1A,$F9,$2F,$F1,$B1,$FD,$92,$35,$3D,$AB,$E1,$6A
	.dc.b	$FA,$2B,$F2,$F2,$FB,$CE,$5B,$3F,$48,$8F,$D2,$6B,$E1,$3F,$02,$F6
	.dc.b	$3B,$FD,$17,$D1,$FA,$17,$D1,$52,$57,$BA,$3E,$2B,$FE,$78,$76,$E6
	.dc.b	$36,$F7,$11,$C1,$7D,$0B,$E8,$BF,$00,$FC,$6E,$2F,$F1,$B3,$77,$FD
	.dc.b	$9B,$BA,$F9,$ED,$7B,$3B,$7F,$D1,$6D,$7B,$D7,$CC,$F3,$7E,$6F,$AD
	.dc.b	$27,$E5,$C6,$5F,$CC,$BF,$9C,$23,$83,$64,$3D,$DA,$BD,$DF,$CD,$7D
	.dc.b	$13,$E3,$EF,$CB,$4B,$FC,$3C,$E5,$B3,$97,$E4,$42,$79,$CF,$E9,$7E
	.dc.b	$8B,$B5,$7E,$41,$AF,$CF,$6E,$DF,$E5,$F5,$F4,$3C,$71,$95,$C5,$E9
	.dc.b	$1F,$90,$F9,$DF,$C8,$F3,$CA,$FC,$7C,$65,$FD,$EB,$FC,$00,$8E,$0D
	.dc.b	$90,$FC,$17,$CD,$FF,$4F,$C7,$FC,$E5,$97,$FA,$7A,$7F,$D8,$2B,$CD
	.dc.b	$FE,$04,$D8,$FD,$6B,$B7,$E4,$18,$8D,$9C,$1E,$31,$1E,$2F,$4D,$31
	.dc.b	$1F,$95,$8E,$75,$E4,$75,$1F,$BC,$BF,$F3,$7B,$7E,$43,$F8,$B4,$7F
	.dc.b	$AF,$E3,$DB,$BE,$0C,$CC,$8E,$51,$97,$91,$0E,$26,$8C,$5F,$FA,$5E
	.dc.b	$BB,$36,$79,$4F,$CC,$AE,$CF,$22,$4F,$18,$CB,$C6,$3B,$1C,$4E,$FC
	.dc.b	$9F,$D8,$2F,$DF,$1C,$A3,$15,$C9,$D5,$F8,$AF,$B5,$72,$3F,$DF,$77
	.dc.b	$CA,$3C,$11,$DC,$8E,$09,$5D,$A3,$90,$9E,$51,$E0,$8C,$C7,$C4,$8C
	.dc.b	$47,$9C,$F0,$DF,$9D,$F1,$DF,$9E,$7B,$47,$89,$E3,$7E,$28,$E2,$76
	.dc.b	$9E,$0E,$0E,$07,$43,$E5,$C7,$C2,$72,$7A,$63,$8B,$EB,$AF,$23,$93
	.dc.b	$8D,$79,$2B,$91,$E3,$8D,$1D,$2F,$E0,$3E,$13,$E0,$1D,$4F,$41,$D0
	.dc.b	$F4,$47,$49,$F7,$92,$32,$F8,$4C,$3E,$08,$E0,$E8,$F0,$3F,$71,$FF
	.dc.b	$0E,$4F,$E2,$1B,$7F,$E0,$07,$CC,$60,$C4,$6D,$FB,$26,$4F,$F0,$3B
	.dc.b	$F8,$07,$BD,$79,$21,$F0,$DF,$31,$D4,$F9,$E7,$F3,$53,$9A,$D3,$67
	.dc.b	$E2,$1F,$8D,$70,$FC,$AA,$EA,$E9,$EA,$48,$DD,$FD,$58,$DC,$F1,$AE
	.dc.b	$EF,$16,$B9,$3A,$57,$2B,$F6,$36,$3D,$E7,$EA,$9C,$3E,$D3,$07,$53
	.dc.b	$99,$4F,$53,$C2,$BD,$11,$9B,$F4,$4E,$6F,$F8,$DB,$67,$F9,$03,$5F
	.dc.b	$E7,$2D,$9F,$E2,$8D,$BF,$9F,$3D,$5F,$5B,$4F,$59,$F3,$CF,$F1,$3F
	.dc.b	$D1,$7F,$93,$63,$FC,$DE,$FF,$94,$C9,$E1,$FF,$90,$DD,$48,$3B,$F0
	.dc.b	$F5,$3B,$E2,$95,$D9,$E5,$C5,$88,$FF,$0A,$5F,$F6,$6C,$3C,$35,$A9
	.dc.b	$FF,$1D,$F7,$7D,$CC,$D5,$FE,$84,$87,$C8,$7F,$95,$DB,$5C,$34,$C1
	.dc.b	$FA,$17,$FC,$90,$D8,$FF,$3D,$11,$DC,$71,$3C,$3E,$77,$F9,$8F,$E8
	.dc.b	$BF,$CF,$D8,$F8,$A7,$E1,$43,$E5,$34,$6F,$E2,$E4,$FF,$95,$27,$4D
	.dc.b	$9C,$BF,$E3,$58,$FF,$07,$3B,$E5,$F0,$1E,$1F,$9B,$B1,$92,$1C,$5F
	.dc.b	$F3,$C6,$CF,$A1,$FF,$3D,$7D,$13,$C6,$A6,$A2,$5F,$F0,$5A,$E0,$D8
	.dc.b	$F9,$11,$88,$FF,$44,$B7,$4F,$85,$1F,$E2,$F9,$E6,$FF,$93,$ED,$DF
	.dc.b	$FE,$8B,$FB,$37,$CC,$5E,$B9,$7F,$E7,$89,$CB,$67,$2F,$F9,$66,$3F
	.dc.b	$97,$A7,$6E,$7E,$93,$B1,$1E,$9C,$55,$D1,$DE,$3C,$91,$FE,$9F,$DD
	.dc.b	$F5,$97,$E3,$E5,$FF,$A3,$27,$2D,$9F,$C6,$46,$23,$FD,$5A,$D7,$FD
	.dc.b	$9C,$D7,$FD,$B6,$EF,$6F,$92,$DE,$8C,$D7,$8E,$3F,$D6,$5F,$66,$FA
	.dc.b	$CB,$C7,$31,$FF,$4F,$CE,$5B,$39,$7F,$D0,$11,$F8,$17,$27,$0F,$13
	.dc.b	$A9,$1E,$7E,$25,$78,$C5,$CF,$EF,$D2,$BF,$D8,$DF,$67,$F8,$3D,$3F
	.dc.b	$DE,$EB,$EB,$B2,$6B,$9A,$65,$FD,$4E,$7F,$18,$FF,$BC,$DA,$F2,$8F
	.dc.b	$DC,$47,$58,$FE,$20,$D4,$DF,$A9,$FF,$2A,$0F,$73,$F8,$03,$67,$FB
	.dc.b	$13,$5F,$F7,$1B,$5F,$E3,$87,$F1,$DF,$55,$F0,$FF,$C2,$5B,$1F,$DA
	.dc.b	$DA,$FF,$BB,$1A,$F2,$0B,$E3,$F9,$12,$39,$3D,$74,$91,$D5,$AF,$FB
	.dc.b	$E3,$EA,$1F,$53,$FC,$6D,$B1,$FD,$F5,$AE,$E3,$5E,$C4,$8D,$35,$FE
	.dc.b	$FA,$FF,$25,$F3,$0D,$24,$7F,$82,$7E,$A7,$E3,$3F,$19,$AF,$F4,$36
	.dc.b	$A6,$FD,$83,$EA,$7F,$99,$36,$7F,$C2,$8D,$7E,$40,$70,$FC,$BF,$A8
	.dc.b	$7C,$67,$F3,$B6,$C7,$F8,$75,$AF,$26,$BC,$7A,$F8,$11,$E0,$FF,$45
	.dc.b	$FA,$97,$A1,$49,$3F,$E2,$98,$CB,$DF,$EA,$BC,$44,$70,$78,$FC,$9F
	.dc.b	$31,$1C,$1C,$FD,$2F,$01,$1C,$1E,$BD,$3D,$C4,$70,$7B,$FE,$17,$E1
	.dc.b	$11,$C1,$E7,$F8,$2F,$61,$1C,$1F,$1F,$E1,$FE,$51,$1C,$1E,$FE,$EF
	.dc.b	$88,$47,$07,$C3,$F8,$CF,$90,$47,$07,$D3,$F6,$3E,$B1,$1C,$10,$78
	.dc.b	$5F,$D2,$B8,$7F,$D8,$37,$FB,$26,$CF,$F8,$3F,$7F,$D7,$D7,$F5,$0F
	.dc.b	$A9,$FF,$02,$5F,$EE,$1B,$5F,$E1,$56,$A7,$FC,$0A,$F8,$23,$87,$AD
	.dc.b	$FC,$02,$BA,$36,$33,$3F,$E1,$F9,$FE,$BF,$F5,$3F,$BC,$8C,$4F,$F0
	.dc.b	$8F,$51,$F4,$1C,$67,$FC,$4D,$18,$76,$62,$34,$FF,$82,$7E,$8E,$C9
	.dc.b	$A4,$8E,$22,$FF,$6E,$E3,$E9,$BF,$B1,$E3,$FD,$D1,$B8,$FF,$1C,$6C
	.dc.b	$FF,$8E,$5F,$ED,$64,$F4,$FC,$6E,$3D,$DA,$49,$EC,$FA,$3E,$A7,$F8
	.dc.b	$F6,$1D,$FD,$BE,$24,$AF,$26,$DD,$08,$E0,$F4,$7D,$4F,$E0,$97,$D4
	.dc.b	$FF,$89,$96,$49,$D4,$7F,$2D,$8C,$61,$FE,$9D,$6C,$7F,$0A,$6B,$BF
	.dc.b	$A8,$7D,$4E,$DB,$1B,$6B,$B8,$FF,$24,$2F,$9A,$A4,$7F,$C9,$8F,$F9
	.dc.b	$37,$E1,$3F,$C8,$98,$7F,$89,$27,$F0,$7C,$1F,$61,$4E,$A3,$FA,$37
	.dc.b	$25,$9F,$EA,$2D,$5F,$4E,$EC,$C4,$74,$3F,$34,$7F,$9C,$E3,$55,$DF
	.dc.b	$97,$7F,$53,$C1,$B1,$E3,$8E,$7F,$07,$8B,$E6,$7F,$9F,$5C,$CF,$E4
	.dc.b	$2D,$DB,$8A,$DC,$62,$7F,$2B,$A3,$9B,$D4,$FF,$B1,$A3,$31,$8E,$59
	.dc.b	$1D,$7E,$0A,$E1,$D1,$F8,$5F,$53,$F2,$27,$F4,$2F,$C1,$C3,$FC,$A2
	.dc.b	$E6,$BD,$2E,$67,$FB,$DB,$63,$E3,$7B,$3D,$38,$3E,$76,$68,$FA,$1F
	.dc.b	$52,$F1,$FE,$D5,$FF,$A7,$E3,$18,$FF,$BA,$DB,$1F,$D6,$1A,$FF,$80
	.dc.b	$5F,$F0,$94,$6F,$D6,$E7,$29,$CB,$C6,$33,$F1,$6F,$A3,$A4,$BF,$F7
	.dc.b	$D4,$65,$BE,$97,$02,$38,$37,$F0,$5D,$08,$E0,$FE,$DF,$83,$E9,$5F
	.dc.b	$63,$E6,$6C,$E9,$B3,$FD,$D9,$AF,$F8,$7D,$61,$C4,$93,$C5,$E9,$F0
	.dc.b	$F2,$3A,$36,$BF,$C0,$3F,$4A,$F9,$5F,$8E,$DA,$F3,$B6,$77,$3A,$F9
	.dc.b	$59,$5A,$6C,$F4,$F9,$FE,$64,$9E,$6F,$D4,$FB,$3F,$F2,$C4,$7C,$46
	.dc.b	$CF,$F8,$51,$AF,$F8,$F5,$AF,$F9,$23,$EC,$5E,$BF,$DE,$6D,$7C,$E6
	.dc.b	$CF,$F8,$79,$AF,$F9,$1F,$E8,$BF,$5B,$F4,$7F,$51,$B1,$FE,$24,$7F
	.dc.b	$81,$78,$99,$F0,$15,$F6,$BE,$C3,$FA,$45,$DF,$13,$FD,$C1,$18,$8F
	.dc.b	$F1,$6B,$5F,$F2,$B7,$84,$F6,$3F,$DC,$CD,$8F,$C2,$B5,$FF,$1A,$B5
	.dc.b	$E3,$C7,$C0,$8F,$07,$FD,$E1,$F4,$5F,$EB,$68,$FF,$32,$35,$FF,$37
	.dc.b	$B5,$07,$0D,$F4,$39,$91,$FC,$CC,$23,$BD,$49,$CB,$5F,$F2,$6F,$4F
	.dc.b	$EC,$93,$49,$C1,$FA,$48,$CF,$B7,$67,$FC,$83,$1C,$A3,$06,$72,$91
	.dc.b	$FA,$C9,$F8,$CF,$FA,$31,$AF,$BB,$E5,$BF,$9E,$BF,$DF,$1F,$E0,$0D
	.dc.b	$ED,$FF,$02,$C6,$A3,$F9,$AB,$5F,$F4,$73,$5F,$F4,$CB,$5E,$DE,$17
	.dc.b	$AE,$39,$78,$BF,$EF,$86,$E5,$C2,$7F,$C2,$7F,$61,$FE,$F8,$BF,$F0
	.dc.b	$7F,$D7,$FF,$7E,$AF,$F5,$2B,$FD,$6A,$33,$CB,$FE,$1A,$8D,$47,$DA
	.dc.b	$6A,$78,$3F,$D3,$9B,$8A,$BA,$CF,$F8,$93,$E7,$FF,$D9,$CF,$F8,$81
	.dc.b	$AF,$F5,$66,$CF,$18,$FC,$4F,$E1,$3F,$A4,$78,$7C,$3F,$32,$3F,$E2
	.dc.b	$A8,$D3,$FD,$83,$EC,$7F,$D2,$55,$CE,$10,$FB,$3F,$F1,$BB,$63,$FC
	.dc.b	$66,$F0,$C7,$FD,$92,$FF,$A6,$22,$FC,$17,$E2,$E3,$11,$F9,$E6,$BD
	.dc.b	$A3,$A9,$1F,$F1,$FF,$CB,$7D,$37,$FB,$AF,$D0,$FF,$90,$1F,$EE,$FF
	.dc.b	$84,$FC,$E3,$E7,$C6,$72,$F5,$37,$1F,$F6,$EC,$EE,$BF,$CA,$33,$97
	.dc.b	$F6,$3F,$45,$FD,$8D,$FF,$27,$B5,$FF,$02,$B6,$FE,$33,$F4,$AF,$FD
	.dc.b	$D2,$E5,$BE,$7F,$F9,$7A,$31,$5D,$B6,$FF,$30,$FC,$4B,$EC,$BF,$3F
	.dc.b	$E7,$4A,$FF,$2F,$CE,$5B,$3F,$EC,$A2,$3F,$EC,$A6,$BF,$E1,$A8,$E0
	.dc.b	$D8,$FC,$EF,$A9,$F9,$7F,$4F,$F3,$1E,$1C,$AF,$CF,$C6,$5F,$D0,$C6
	.dc.b	$5F,$CC,$47,$F9,$E3,$91,$F6,$94,$93,$F8,$DF,$AB,$AF,$F8,$B2,$76
	.dc.b	$FE,$46,$79,$4F,$E3,$3E,$BF,$F1,$BF,$15,$2B,$F1,$B1,$97,$FB,$BF
	.dc.b	$3A,$3F,$18,$48,$CB,$B7,$F3,$8F,$FA,$3B,$99,$A7,$FA,$43,$9B,$F6
	.dc.b	$3C,$9F,$84,$D7,$06,$7A,$E6,$3F,$E9,$0A,$FF,$4D,$3F,$90,$9C,$BF
	.dc.b	$E4,$8F,$A6,$FD,$5F,$3F,$EA,$5B,$87,$F3,$C9,$46,$1F,$E7,$EF,$C4
	.dc.b	$79,$79,$3F,$EA,$48,$D4,$70,$8C,$70,$7F,$AA,$3E,$47,$FD,$09,$3F
	.dc.b	$EA,$47,$FD,$7B,$87,$EE,$5B,$87,$85,$DC,$FF,$97,$9B,$3B,$F8,$5E
	.dc.b	$87,$5F,$84,$95,$D6,$32,$F7,$8C,$BC,$E3,$2F,$58,$CC,$F5,$8E,$47
	.dc.b	$12,$1C,$4C,$9C,$4C,$BC,$5D,$19,$23,$92,$5F,$BE,$7C,$2F,$C1,$AF
	.dc.b	$2C,$FC,$59,$CD,$7F,$70,$BF,$8E,$DA,$CA,$8F,$8F,$7C,$E3,$CF,$7C
	.dc.b	$DF,$3C,$65,$F3,$C6,$5F,$3C,$66,$3C,$F1,$E1,$BE,$F5,$E2,$8F,$05
	.dc.b	$E2,$7D,$37,$C4,$F7,$4F,$80,$F4,$1E,$47,$2F,$80,$C9,$E4,$73,$7E
	.dc.b	$53,$C7,$1E,$03,$A1,$F0,$67,$C2,$F2,$8E,$13,$C8,$F2,$46,$5C,$3E
	.dc.b	$D3,$C8,$64,$C1,$E5,$3C,$07,$94,$CC,$76,$78,$1A,$70,$F9,$4E,$A6
	.dc.b	$63,$C2,$E4,$FD,$91,$E0,$2D,$FE,$91,$B5,$FE,$21,$8C,$47,$CA,$32
	.dc.b	$7F,$A2,$1C,$BE,$69,$EF,$1F,$86,$8F,$C6,$7C,$1F,$D9,$7E,$5C,$23
	.dc.b	$11,$F0,$63,$0E,$AB,$F0,$73,$AB,$F7,$D6,$AB,$85,$69,$E0,$76,$AE
	.dc.b	$17,$E4,$6C,$79,$CF,$29,$C0,$F1,$3A,$7E,$45,$FF,$0B,$8C,$5F,$B6
	.dc.b	$7F,$49,$18,$BF,$E2,$8D,$8F,$E2,$8D,$7F,$88,$36,$BA,$47,$85,$F6
	.dc.b	$C6,$6F,$F8,$C3,$67,$CE,$7C,$03,$DC,$71,$8F,$73,$4F,$71,$E6,$30
	.dc.b	$7B,$1E,$AE,$2F,$D8,$7C,$DA,$FF,$11,$78,$2F,$80,$FF,$1F,$7C,$04
	.dc.b	$BF,$F2,$2C,$7F,$90,$4C,$3F,$E3,$D8,$C3,$FE,$2F,$FA,$0F,$ED,$04
	.dc.b	$7F,$C6,$5E,$6D,$7F,$B3,$45,$C3,$FC,$65,$D5,$FE,$7D,$9F,$E7,$DE
	.dc.b	$4F,$F4,$1F,$22,$46,$08,$7E,$78,$FE,$31,$1F,$9B,$23,$F9,$97,$FA
	.dc.b	$51,$1F,$E9,$5E,$07,$F6,$5E,$EF,$EC,$DA,$49,$FE,$C5,$19,$7F,$B2
	.dc.b	$46,$5F,$F1,$9C,$7F,$59,$EE,$91,$FE,$9C,$CB,$FB,$24,$7F,$5D,$CF
	.dc.b	$33,$FD,$A6,$2C,$FF,$69,$9F,$EB,$DF,$25,$FD,$7F,$E5,$65,$FF,$5F
	.dc.b	$8C,$BF,$B3,$7F,$70,$47,$06,$C8,$2B,$FC,$BD,$F2,$7F,$E7,$6F,$A6
	.dc.b	$7C,$9F,$F7,$79,$FF,$31,$46,$2B,$D5,$CB,$FD,$EE,$31,$1C,$E3,$1F
	.dc.b	$27,$C5,$D5,$D9,$FF,$5E,$D7,$0B,$EB,$39,$94,$AF,$1F,$4D,$FE,$08
	.dc.b	$9F,$09,$FE,$C6,$5F,$4B,$A3,$7D,$D2,$76,$27,$6E,$1A,$FA,$4E,$91
	.dc.b	$E3,$CD,$CF,$78,$A5,$74,$9C,$B6,$7F,$D9,$11,$88,$FC,$93,$5F,$E6
	.dc.b	$2D,$7F,$C4,$8E,$F4,$FA,$73,$E1,$3C,$5F,$23,$F1,$7E,$57,$FC,$63
	.dc.b	$3F,$E9,$56,$EF,$FE,$2D,$6E,$1F,$EE,$A7,$FE,$00,$4E,$FC,$1E,$44
	.dc.b	$9F,$8B,$3A,$AE,$5C,$1D,$F7,$95,$A8,$F1,$18,$7E,$40,$AF,$B0,$EC
	.dc.b	$F8,$DF,$0F,$F2,$FE,$26,$46,$08,$7C,$68,$F9,$24,$7F,$74,$35,$47
	.dc.b	$D5,$24,$7D,$3F,$2F,$F1,$5F,$35,$FE,$80,$8F,$EB,$ED,$7F,$CA,$4D
	.dc.b	$41,$C3,$7D,$2F,$0B,$4F,$1C,$7F,$AE,$09,$E9,$E9,$FF,$62,$47,$F1
	.dc.b	$16,$BE,$28,$E7,$F3,$1F,$ED,$08,$FE,$E6,$D7,$FC,$CC,$D7,$CA,$D4
	.dc.b	$FF,$99,$DE,$3E,$3F,$42,$3F,$22,$1D,$9E,$02,$46,$5F,$6F,$CB,$FE
	.dc.b	$4E,$3E,$1B,$53,$49,$C9,$C0,$91,$F5,$A1,$1D,$3E,$62,$4E,$5A,$FF
	.dc.b	$35,$F4,$BE,$2B,$FE,$85,$6B,$F5,$1A,$9B,$F3,$0F,$9F,$FA,$AD,$9F
	.dc.b	$C3,$B5,$FE,$84,$D7,$F6,$3B,$7E,$D8,$8F,$E3,$B8,$7F,$DE,$72,$FA
	.dc.b	$0E,$A4,$9E,$AF,$83,$E2,$E4,$78,$1A,$F8,$3E,$2F,$CA,$47,$CA,$FA
	.dc.b	$3E,$2E,$47,$A1,$AF,$A3,$E6,$FD,$A4,$7D,$AF,$C0,$F9,$09,$1F,$01
	.dc.b	$AF,$C0,$F8,$FF,$8A,$47,$E2,$BF,$23,$E2,$24,$7C,$86,$BF,$23,$D8
	.dc.b	$F9,$A4,$7E,$6B,$F4,$3E,$0E,$47,$D0,$6B,$F4,$3C,$5F,$E7,$9B,$26
	.dc.b	$98,$8F,$E8,$1F,$36,$F9,$4A,$4B,$FE,$73,$1F,$E8,$32,$7B,$1C,$0F
	.dc.b	$E9,$1F,$36,$F3,$BF,$A5,$74,$FF,$22,$BF,$C9,$F8,$F8,$33,$E8,$48
	.dc.b	$FE,$8B,$3F,$C8,$5F,$E0,$6D,$56,$F3,$FF,$AB,$7C,$CB,$D1,$A4,$AF
	.dc.b	$E9,$F1,$FE,$17,$7F,$92,$B7,$CC,$79,$9F,$EE,$38,$C3,$7D,$6E,$BF
	.dc.b	$17,$2F,$9B,$5F,$E5,$FF,$37,$FE,$7A,$C3,$9C,$7F,$1A,$C3,$E5,$3E
	.dc.b	$2E,$2F,$39,$E8,$E1,$91,$E6,$6B,$E6,$F7,$3F,$C0,$8B,$E6,$FF,$76
	.dc.b	$F5,$8C,$BF,$E8,$AB,$46,$23,$D8,$D7,$86,$1F,$CE,$A7,$FD,$01,$F3
	.dc.b	$77,$DE,$34,$95,$FC,$C3,$E6,$FF,$98,$A6,$91,$FE,$8B,$F0,$5F,$0B
	.dc.b	$0F,$87,$1F,$CF,$39,$B1,$C3,$FE,$A0,$AB,$3F,$D0,$04,$E1,$E1,$3F
	.dc.b	$E0,$8F,$75,$F3,$79,$9D,$CE,$63,$1D,$19,$F6,$6C,$FF,$84,$2B,$FC
	.dc.b	$D0,$D8,$E5,$EC,$EE,$34,$E1,$F0,$7C,$9D,$AF,$ED,$B8,$A3,$CD,$A3
	.dc.b	$9A,$F2,$E1,$B8,$7F,$88,$27,$F5,$8F,$AF,$C1,$CE,$3D,$1F,$81,$F3
	.dc.b	$6E,$0B,$C1,$86,$DF,$F8,$AA,$7E,$0D,$F9,$3E,$33,$2D,$9F,$F1,$7F
	.dc.b	$CD,$FF,$76,$7F,$BC,$B6,$3F,$CE,$8D,$7F,$DA,$6D,$7E,$A2,$F6,$3E
	.dc.b	$DC,$61,$AF,$F7,$88,$D4,$F6,$9C,$CB,$C7,$7C,$0E,$1B,$5F,$E0,$C6
	.dc.b	$CF,$FA,$1D,$AF,$FB,$99,$7C,$8F,$12,$4F,$17,$BF,$D9,$BD,$2A,$49
	.dc.b	$FF,$0C,$C6,$5F,$F0,$DF,$CE,$79,$5B,$7E,$88,$C6,$1F,$E9,$96,$BF
	.dc.b	$EF,$85,$E6,$6C,$95,$D1,$F8,$3F,$65,$F6,$E3,$FC,$52,$DA,$FF,$52
	.dc.b	$35,$FD,$10,$70,$FE,$97,$E7,$5E,$97,$CC,$6D,$7F,$8D,$1B,$3F,$EA
	.dc.b	$F6,$BD,$17,$A5,$C4,$93,$C5,$FA,$F8,$7C,$E8,$FF,$1F,$B6,$BF,$D7
	.dc.b	$4D,$7C,$ED,$7F,$75,$F3,$6F,$9C,$FC,$0B,$6B,$FC,$92,$D9,$FF,$62
	.dc.b	$35,$F8,$3F,$39,$FA,$DF,$8E,$FF,$27,$B6,$3F,$D9,$4F,$F0,$2E,$7C
	.dc.b	$F6,$FF,$8D,$7E,$BF,$DB,$5F,$E5,$66,$BF,$ED,$16,$BF,$39,$AF,$F1
	.dc.b	$1E,$D7,$CD,$7F,$93,$9B,$1F,$E6,$06,$BF,$ED,$B6,$BF,$61,$FC,$1B
	.dc.b	$BD,$79,$11,$FF,$34,$FC,$87,$F0,$76,$CF,$F9,$72,$31,$1F,$EE,$76
	.dc.b	$BF,$84,$F9,$CF,$EB,$4D,$F3,$FF,$CC,$71,$88,$FE,$54,$D7,$FC,$D6
	.dc.b	$F2,$73,$5C,$08,$FF,$9F,$3E,$5B,$FC,$DE,$4E,$CF,$F3,$7B,$FE,$72
	.dc.b	$EB,$7C,$E7,$F9,$C9,$B5,$FE,$84,$6B,$FE,$FF,$6B,$F9,$C7,$E3,$BF
	.dc.b	$CF,$DA,$FF,$A0,$1A,$FE,$FB,$4F,$F3,$FF,$D6,$BD,$1A,$49,$FF,$49
	.dc.b	$C6,$5F,$F4,$AF,$C1,$7F,$17,$7F,$D9,$FE,$DF,$F1,$06,$F8,$3F,$ED
	.dc.b	$D9,$D4,$7F,$AA,$1F,$F4,$79,$34,$FF,$47,$B6,$3F,$DB,$2F,$0F,$1F
	.dc.b	$FA,$EB,$7A,$7F,$AD,$7C,$EB,$C8,$FF,$75,$7C,$AF,$F8,$84,$FF,$0D
	.dc.b	$13,$C9,$C9,$AF,$3A,$65,$FE,$E6,$9F,$F1,$47,$CA,$BE,$2B,$FD,$E8
	.dc.b	$D9,$FF,$75,$B5,$FF,$60,$BF,$E8,$E7,$31,$C0,$8F,$0F,$89,$FF,$7F
	.dc.b	$7C,$CB,$E6,$6F,$44,$AF,$F7,$BC,$E5,$B3,$FE,$B2,$8C,$47,$F1,$80
	.dc.b	$E1,$FE,$36,$D7,$FB,$F3,$C3,$83,$FC,$0A,$DD,$F3,$5F,$E0,$39,$FF
	.dc.b	$04,$FC,$CF,$CB,$BF,$F6,$2C,$7F,$83,$3E,$67,$FB,$34,$62,$FF,$C1
	.dc.b	$51,$FE,$4F,$27,$85,$CE,$32,$D9,$FF,$67,$CE,$F8,$31,$08,$EA,$FF
	.dc.b	$BB,$3E,$11,$E2,$E9,$B1,$E1,$3F,$DB,$2E,$1E,$1B,$EE,$64,$6C,$4E
	.dc.b	$F2,$FF,$13,$BF,$6D,$FF,$14,$F8,$3F,$E3,$C9,$E0,$FE,$7D,$FF,$7F
	.dc.b	$B6,$37,$EE,$7A,$A3,$84,$FF,$0B,$8F,$E1,$C4,$7F,$89,$BB,$23,$DB
	.dc.b	$DB,$FC,$CF,$B1,$F9,$AF,$62,$4F,$E5,$63,$2F,$E5,$E3,$2B,$0F,$F2
	.dc.b	$0F,$32,$BF,$33,$EF,$CA,$FC,$CB,$F9,$CF,$1B,$11,$F9,$A8,$B7,$FE
	.dc.b	$4A,$E8,$FF,$3A,$C8,$68,$F9,$9F,$81,$ED,$49,$F8,$F1,$F0,$CF,$E5
	.dc.b	$C4,$7F,$99,$B5,$F9,$6E,$DF,$9F,$E0,$7E,$13,$8F,$FD,$24,$7A,$E3
	.dc.b	$ED,$35,$F8,$1C,$DF,$92,$EC,$E0,$71,$DA,$4F,$08,$FC,$57,$64,$8F
	.dc.b	$C2,$72,$7E,$E5,$FC,$86,$F9,$7F,$25,$37,$7F,$C3,$CF,$DF,$F8,$1F
	.dc.b	$C0,$7C,$24,$BF,$C3,$C6,$5F,$C4,$C6,$5F,$C4,$C6,$5F,$C5,$C6,$6B
	.dc.b	$F2,$33,$97,$F2,$71,$9C,$FF,$CF,$1C,$1F,$EB,$3B,$EE,$76,$21,$D8
	.dc.b	$C9,$D8,$CD,$76,$72,$F6,$73,$3D,$AF,$C4,$F8,$2F,$C5,$3F,$83,$BF
	.dc.b	$13,$DC,$EC,$4B,$EC,$E5,$EC,$E6,$BB,$3E,$27,$B3,$E2,$3B,$5F,$89
	.dc.b	$F0,$5F,$89,$FC,$9C,$65,$FC,$AC,$66,$FC,$B7,$E2,$3C,$24,$BF,$0D
	.dc.b	$79,$5F,$0B,$E5,$3C,$27,$94,$F0,$9E,$53,$C2,$64,$F0,$99,$9F,$0C
	.dc.b	$76,$9F,$04,$F2,$3C,$07,$63,$C0,$72,$32,$76,$33,$3F,$EF,$F7,$C0
	.dc.b	$F6,$7A,$47,$67,$83,$F0,$67,$DA,$E6,$78,$4E,$A7,$24,$7C,$93,$C5
	.dc.b	$EB,$19,$78,$1E,$93,$FC,$D7,$3F,$84,$6D,$FF,$8B,$C8,$FF,$8D,$1A
	.dc.b	$FF,$8C,$5B,$5D,$E3,$A8,$F8,$BF,$F8,$01,$7F,$2C,$91,$E6,$7F,$DB
	.dc.b	$91,$A8,$C1,$0E,$6E,$5E,$B7,$F0,$8F,$87,$19,$71,$7F,$8D,$8C,$3F
	.dc.b	$8F,$8D,$CD,$8D,$34,$FC,$24,$F0,$7E,$BC,$7F,$60,$8C,$47,$DE,$9E
	.dc.b	$6E,$8E,$27,$33,$F0,$71,$C2,$FE,$34,$F0,$9E,$55,$E2,$79,$1C,$0E
	.dc.b	$53,$F3,$6B,$95,$FC,$A9,$F8,$A6,$9C,$46,$A3,$E5,$5F,$7A,$F0,$DF
	.dc.b	$79,$F0,$9E,$83,$C3,$1D,$DE,$87,$D3,$AE,$8F,$D2,$9F,$88,$FF,$81
	.dc.b	$3E,$65,$F2,$D4,$91,$FD,$EE,$1F,$2F,$88,$FE,$86,$7F,$57,$8E,$2E
	.dc.b	$E7,$A4,$FF,$84,$3E,$63,$FC,$07,$3F,$E0,$55,$F3,$39,$35,$C1,$27
	.dc.b	$51,$97,$84,$7D,$BE,$AF,$B9,$C7,$27,$FC,$21,$A2,$75,$1F,$E3,$93
	.dc.b	$F3,$8D,$69,$98,$E0,$D8,$E2,$FB,$8D,$96,$1F,$15,$FE,$2C,$F9,$97
	.dc.b	$83,$49,$5F,$E2,$38,$73,$62,$3A,$47,$07,$53,$CE,$7A,$6D,$D7,$69
	.dc.b	$5C,$E3,$2F,$58,$CB,$D2,$3F,$41,$F3,$1F,$A1,$F9,$99,$3F,$1A,$32
	.dc.b	$FC,$17,$E2,$66,$69,$23,$81,$0F,$85,$E6,$7F,$93,$3E,$A9,$F1,$14
	.dc.b	$93,$FE,$43,$87,$B3,$88,$F9,$91,$C1,$D4,$F3,$9E,$9C,$5D,$78,$E5
	.dc.b	$73,$8C,$BD,$63,$2F,$48,$FF,$2F,$C7,$F9,$1B,$9D,$D3,$FE,$2D,$AF
	.dc.b	$C8,$F8,$9A,$37,$E2,$C8,$FC,$45,$FF,$93,$9E,$07,$E6,$23,$F1,$51
	.dc.b	$98,$E8,$F7,$AB,$3C,$E7,$F1,$52,$FA,$BF,$F3,$9F,$D5,$BE,$62,$92
	.dc.b	$BF,$CD,$50,$FA,$8C,$47,$E4,$E3,$83,$A9,$E7,$3D,$39,$7F,$CF,$FF
	.dc.b	$57,$F0,$2B,$FC,$E8,$D7,$94,$6D,$CE,$7C,$0C,$3F,$EC,$9F,$89,$E7
	.dc.b	$F3,$12,$3F,$D8,$F1,$FE,$CF,$3F,$D9,$64,$7F,$DA,$ED,$79,$1C,$0F
	.dc.b	$1F,$A9,$FE,$94,$F9,$E4,$FE,$D5,$B3,$FB,$96,$BF,$EB,$36,$BF,$51
	.dc.b	$7C,$F7,$12,$47,$17,$F8,$07,$8C,$F9,$CA,$49,$FE,$0D,$19,$7F,$84
	.dc.b	$7C,$F7,$FB,$45,$FC,$0B,$5F,$C5,$35,$FF,$3C,$35,$3F,$E5,$F7,$8F
	.dc.b	$95,$FC,$46,$31,$1C,$A3,$1E,$FC,$8F,$0B,$5F,$F6,$97,$CF,$7F,$10
	.dc.b	$7F,$8E,$35,$CF,$8B,$FC,$76,$31,$1F,$98,$7F,$0E,$F2,$E3,$D1,$1E
	.dc.b	$7F,$11,$C4,$8E,$0F,$E5,$5F,$37,$F1,$5F,$F7,$23,$5F,$D2,$B5,$FD
	.dc.b	$73,$5F,$F4,$E7,$85,$F6,$1F,$F0,$4F,$07,$F1,$02,$3D,$BA,$7F,$9B
	.dc.b	$57,$F4,$68,$C1,$0F,$0F,$1E,$88,$F9,$9E,$3F,$19,$23,$D2,$D7,$D3
	.dc.b	$F2,$5D,$88,$FB,$9F,$0F,$CE,$C8,$F8,$2D,$7E,$0F,$CE,$F9,$88,$FC
	.dc.b	$67,$D3,$F3,$F2,$3E,$4B,$5F,$93,$F1,$7E,$E2,$3F,$39,$F8,$3D,$32
	.dc.b	$3E,$8B,$5F,$E2,$63,$0F,$F7,$A3,$FC,$59,$AF,$F1,$F6,$BF,$EC,$D6
	.dc.b	$AC,$FF,$BA,$BF,$D7,$D7,$C2,$FF,$27,$F0,$FF,$84,$46,$27,$FD,$FC
	.dc.b	$F0,$8C,$F8,$9F,$60,$93,$F8,$87,$FB,$87,$CF,$3E,$63,$A7,$7F,$F3
	.dc.b	$38,$C5,$7F,$B8,$6F,$F8,$B4,$62,$3F,$B9,$2F,$89,$E3,$1F,$EE,$78
	.dc.b	$C3,$C6,$BA,$4E,$A1,$3A,$7F,$BE,$7C,$F7,$F9,$F5,$FE,$FD,$D5,$FD
	.dc.b	$CD,$BE,$8F,$18,$C6,$3F,$CA,$5F,$EF,$73,$DC,$FF,$7C,$38,$79,$FA
	.dc.b	$3C,$4F,$15,$FF,$A0,$BE,$71,$ED,$79,$5B,$5C,$C3,$87,$FC,$24,$BE
	.dc.b	$03,$81,$23,$84,$65,$BF,$1D,$D4,$8E,$0E,$DF,$3D,$FE,$75,$9F,$9D
	.dc.b	$9F,$F4,$56,$EF,$FD,$9D,$B4,$3C,$BB,$FA,$48,$F8,$9F,$67,$9C,$BE
	.dc.b	$1F,$3D,$FC,$5E,$D3,$9B,$FC,$DC,$62,$BF,$C7,$0D,$8C,$CE,$DF,$EE
	.dc.b	$70,$F4,$1F,$3D,$C1,$B3,$D1,$AF,$06,$BF,$AA,$F9,$EF,$F2,$22,$69
	.dc.b	$1F,$F2,$57,$C4,$FF,$6B,$C7,$FB,$54,$7F,$90,$34,$FB,$0A,$75,$1F
	.dc.b	$E0,$8E,$4B,$3F,$E1,$16,$AF,$85,$FF,$14,$CF,$29,$CC,$62,$7F,$CA
	.dc.b	$47,$F9,$2A,$7B,$FB,$DC,$A3,$06,$9C,$19,$F7,$AC,$F9,$9C,$3F,$8F
	.dc.b	$AF,$F1,$AB,$93,$1E,$BF,$9B,$27,$1D,$EA,$C3,$6B,$F3,$13,$FB,$D7
	.dc.b	$A7,$7E,$73,$FF,$35,$62,$8F,$83,$F3,$DF,$1E,$BF,$40,$FC,$BA,$FF
	.dc.b	$38,$39,$6E,$BC,$0F,$8B,$E0,$5A,$F5,$3F,$33,$E7,$9E,$E7,$D6,$BF
	.dc.b	$CC,$46,$2B,$FD,$08,$D8,$FE,$56,$D7,$FC,$7F,$F6,$8F,$94,$E5,$08
	.dc.b	$C5,$7F,$A3,$5B,$1F,$CD,$1A,$FF,$92,$3E,$79,$E9,$72,$6D,$7F,$A5
	.dc.b	$5B,$3F,$CF,$1A,$FF,$94,$3E,$7F,$FC,$4B,$3F,$E9,$B6,$C7,$F4,$26
	.dc.b	$BB,$F3,$FC,$FF,$68,$FA,$7A,$4B,$FF,$52,$46,$5F,$F5,$37,$D0,$BC
	.dc.b	$CF,$7B,$63,$31,$83,$E1,$FD,$0F,$D5,$C7,$FC,$CA,$D7,$FC,$DC,$D7
	.dc.b	$F8,$43,$57,$C7,$FE,$41,$FE,$B6,$D7,$FC,$DE,$D7,$FC,$F0,$D7,$F8
	.dc.b	$97,$D0,$BE,$87,$EA,$B7,$0F,$F6,$03,$67,$FB,$3B,$5F,$F3,$DE,$1F
	.dc.b	$39,$FF,$42,$35,$FE,$3C,$D7,$E2,$6F,$F8,$2F,$B4,$7B,$34,$98,$7F
	.dc.b	$B3,$E3,$2F,$FB,$47,$E8,$7F,$D4,$3F,$2D,$7C,$0F,$EE,$71,$FE,$A8
	.dc.b	$F0,$F1,$3F,$ED,$7F,$A1,$F6,$67,$76,$FF,$6C,$46,$23,$FC,$02,$D7
	.dc.b	$FD,$2B,$F1,$4F,$99,$FF,$6F,$36,$3F,$DC,$4D,$7F,$C1,$2D,$7F,$D3
	.dc.b	$BE,$1F,$E0,$5F,$F7,$43,$5F,$F0,$7B,$5D,$F0,$7F,$BA,$FE,$53,$F8
	.dc.b	$2C,$7F,$BB,$DA,$FF,$85,$9A,$FF,$AA,$9A,$FF,$AC,$FC,$27,$D5,$7F
	.dc.b	$BD,$5B,$1F,$EF,$96,$BF,$E2,$06,$BF,$EB,$4F,$A2,$E6,$FF,$55,$6A
	.dc.b	$69,$8F,$9E,$FE,$0B,$1F,$E2,$A0,$E1,$FF,$16,$B5,$FF,$60,$B5,$FF
	.dc.b	$65,$7D,$93,$DE,$E5,$08,$C3,$D5,$AF,$F8,$E1,$AF,$FB,$33,$9F,$F8
	.dc.b	$73,$E3,$6B,$FE,$3E,$6B,$BF,$0B,$BF,$D1,$7F,$86,$E3,$FD,$AA,$D7
	.dc.b	$FD,$BC,$D7,$FB,$C3,$57,$D8,$FF,$A1,$9F,$F7,$23,$5F,$EF,$CD,$77
	.dc.b	$ED,$FD,$FF,$53,$FC,$59,$BF,$F8,$AF,$CD,$FF,$66,$37,$AF,$FD,$DC
	.dc.b	$FF,$B8,$EF,$FC,$10,$FF,$B9,$DB,$5C,$DF,$C4,$35,$FF,$04,$4F,$28
	.dc.b	$EE,$78,$3D,$2F,$F1,$E7,$CF,$3D,$2B,$E9,$4C,$3F,$C5,$53,$96,$CE
	.dc.b	$5E,$A7,$F8,$6E,$13,$D2,$7F,$C9,$1F,$3D,$FE,$31,$CF,$FD,$D0,$DC
	.dc.b	$BF,$C4,$58,$23,$0F,$F8,$95,$FF,$19,$3D,$8D,$7B,$7D,$08,$FF,$8F
	.dc.b	$E3,$75,$F2,$BC,$5F,$9B,$E9,$CA,$FF,$22,$6D,$FC,$A1,$B3,$FB,$12
	.dc.b	$3F,$B1,$6B,$E1,$8E,$0D,$8F,$B3,$F3,$DF,$E6,$5F,$9E,$FF,$2C,$60
	.dc.b	$CF,$DB,$1F,$26,$FD,$4D,$9D,$47,$F9,$0E,$38,$CF,$0F,$7F,$A6,$B2
	.dc.b	$BF,$CB,$95,$A7,$FC,$95,$F0,$2F,$8C,$FF,$3B,$36,$3F,$CC,$4D,$7E
	.dc.b	$53,$F3,$93,$9C,$7C,$71,$89,$E6,$74,$DF,$F2,$FF,$3D,$F9,$8F,$8F
	.dc.b	$95,$F3,$23,$2F,$F1,$97,$F8,$E1,$1C,$1B,$21,$F7,$7E,$33,$F0,$6F
	.dc.b	$DF,$F9,$E7,$CE,$7F,$A4,$5B,$1F,$E7,$D6,$BB,$F9,$5D,$4C,$8D,$3C
	.dc.b	$63,$FD,$35,$F3,$DF,$EA,$FE,$1F,$C8,$BF,$EE,$5A,$FC,$73,$75,$F8
	.dc.b	$6E,$E3,$FC,$F0,$D4,$EE,$79,$FC,$87,$6D,$FF,$D5,$19,$FE,$92,$FF
	.dc.b	$D2,$8F,$E0,$6B,$F4,$2D,$8F,$F7,$EF,$47,$03,$F9,$1C,$71,$8F,$F6
	.dc.b	$BB,$5F,$8E,$D7,$9C,$71,$7B,$BE,$2E,$CF,$F5,$D7,$CF,$7F,$AC,$AF
	.dc.b	$F6,$2D,$AF,$F4,$A3,$53,$F8,$F6,$FC,$07,$22,$3F,$EA,$F8,$D1,$F5
	.dc.b	$34,$FC,$5A,$D7,$C2,$78,$B6,$95,$C6,$32,$F3,$8F,$93,$F3,$BF,$2B
	.dc.b	$E7,$A4,$FC,$A8,$CB,$F3,$23,$2F,$E2,$E3,$2F,$E3,$63,$38,$7E,$7B
	.dc.b	$55,$6A,$31,$3A,$CF,$EA,$CE,$5F,$AF,$19,$BF,$F5,$D5,$F2,$7C,$87
	.dc.b	$8C,$87,$8C,$C9,$E3,$33,$7E,$3B,$ED,$1D,$2B,$B2,$AE,$CF,$33,$91
	.dc.b	$2B,$93,$D9,$F5,$BD,$AF,$D6,$E5,$F5,$B9,$3D,$77,$E2,$7C,$17,$E2
	.dc.b	$BF,$0D,$F6,$9F,$05,$F2,$7C,$91,$93,$C2,$E8,$EA,$78,$4F,$71,$E1
	.dc.b	$3D,$D7,$E1,$A5,$F8,$63,$0D,$9E,$47,$98,$E5,$3E,$6B,$F7,$C7,$84
	.dc.b	$D1,$E6,$73,$7E,$62,$38,$24,$79,$8E,$53,$D5,$F5,$98,$AF,$45,$FC
	.dc.b	$76,$C7,$91,$AF,$EE,$A3,$F7,$A4,$7F,$C5,$53,$F7,$9F,$DE,$C6,$5E
	.dc.b	$06,$9E,$77,$F9,$BF,$81,$B8,$7A,$2B,$F1,$11,$87,$CF,$3B,$7E,$4D
	.dc.b	$FA,$DF,$A2,$7F,$16,$9E,$0E,$0F,$E6,$CE,$23,$81,$FE,$3F,$9F,$C0
	.dc.b	$57,$86,$33,$3B,$BF,$E4,$6D,$BE,$C7,$8A,$3F,$01,$19,$8E,$E6,$4F
	.dc.b	$3B,$9B,$F5,$47,$DB,$3D,$B7,$E1,$9D,$3F,$CB,$9B,$1F,$B8,$6B,$F1
	.dc.b	$8E,$C7,$33,$05,$BF,$7D,$FA,$A3,$77,$F0,$1E,$86,$2F,$C4,$D3,$CA
	.dc.b	$68,$E9,$3F,$01,$D1,$F3,$E7,$DA,$7F,$74,$F9,$E7,$9F,$FE,$28,$8C
	.dc.b	$36,$FF,$C5,$73,$AA,$1F,$E2,$95,$EB,$6E,$33,$1B,$75,$59,$51,$9B
	.dc.b	$F1,$56,$63,$FB,$11,$FD,$86,$79,$F1,$E7,$8E,$47,$8A,$FD,$53,$C4
	.dc.b	$CC,$7F,$8F,$5F,$F2,$F2,$5E,$1F,$8D,$B1,$CF,$2F,$F2,$4C,$70,$3F
	.dc.b	$CC,$91,$C7,$C4,$FF,$45,$C6,$08,$F3,$24,$73,$3E,$3C,$3E,$7B,$E1
	.dc.b	$7C,$F3,$E5,$E9,$2B,$80,$87,$CF,$66,$3F,$CB,$11,$C5,$DB,$5C,$C7
	.dc.b	$6E,$5F,$A1,$F3,$DF,$E3,$2D,$7F,$C6,$6D,$D3,$FC,$CF,$87,$F6,$BC
	.dc.b	$39,$DF,$09,$91,$83,$CB,$F4,$6F,$84,$EB,$F3,$52,$73,$1F,$7E,$75
	.dc.b	$1F,$EB,$B6,$BF,$E3,$40,$E1,$FF,$1B,$B5,$FF,$3D,$BC,$32,$FF,$25
	.dc.b	$B6,$F3,$5E,$EF,$57,$F1,$DF,$61,$FE,$DE,$BF,$F2,$8B,$6B,$FD,$0A
	.dc.b	$DC,$3F,$36,$BE,$37,$8F,$B5,$27,$E9,$CE,$D3,$B7,$3C,$3E,$0F,$16
	.dc.b	$46,$08,$7F,$A3,$E5,$1A,$8C,$4F,$E9,$7E,$0B,$FA,$AC,$7E,$09,$FE
	.dc.b	$54,$4E,$ED,$1F,$ED,$C7,$71,$F2,$DA,$FF,$98,$1A,$FF,$A9,$1D,$E7
	.dc.b	$FE,$74,$6E,$79,$AF,$17,$7F,$ED,$FE,$BD,$C3,$49,$8F,$E5,$A1,$F3
	.dc.b	$98,$8F,$F5,$64,$70,$75,$3C,$E7,$F8,$17,$D7,$3E,$0A,$FD,$1C,$BF
	.dc.b	$D0,$CE,$5B,$39,$7F,$D7,$10,$9E,$73,$D3,$D4,$EB,$E9,$CB,$E7,$19
	.dc.b	$7F,$20,$2D,$18,$8C,$36,$33,$1C,$1A,$F8,$A3,$A3,$FE,$CB,$FA,$37
	.dc.b	$C2,$FF,$1F,$5F,$47,$D5,$B3,$E2,$6B,$FA,$36,$BF,$E9,$8E,$47,$D1
	.dc.b	$F9,$5B,$1E,$A6,$BF,$AB,$6B,$C7,$E0,$38,$11,$E0,$FB,$FE,$8F,$FD
	.dc.b	$75,$1F,$EA,$66,$BF,$EE,$A6,$BF,$D0,$1A,$97,$C0,$73,$24,$7C,$E8
	.dc.b	$CB,$F4,$3E,$8F,$E7,$26,$91,$FA,$4F,$F1,$B7,$FD,$72,$D7,$FD,$FA
	.dc.b	$D7,$FA,$9B,$5E,$9C,$FC,$88,$F8,$5F,$BD,$F4,$0F,$A7,$A4,$9F,$C2
	.dc.b	$C6,$5F,$F1,$4F,$81,$FE,$A8,$23,$83,$7E,$3F,$42,$38,$3F,$90,$7D
	.dc.b	$2F,$E3,$7C,$33,$49,$1E,$0E,$8F,$29,$1C,$1E,$4F,$98,$F4,$91,$C1
	.dc.b	$F9,$C7,$F4,$8D,$E9,$FC,$E4,$62,$3F,$98,$35,$FF,$72,$E4,$E0,$F8
	.dc.b	$A4,$8F,$86,$FC,$8F,$A4,$91,$F1,$9A,$F4,$F7,$3B,$91,$F2,$3F,$43
	.dc.b	$C6,$91,$F3,$9A,$F9,$3B,$3C,$E4,$7D,$8F,$AB,$DC,$91,$EF,$6B,$FD
	.dc.b	$2F,$17,$9F,$F7,$89,$1F,$71,$AB,$C4,$78,$9F,$E1,$C8,$57,$FB,$CC
	.dc.b	$FF,$7A,$6A,$FA,$1F,$F0,$D7,$D1,$FF,$41,$AE,$57,$C6,$31,$87,$49
	.dc.b	$FF,$13,$C6,$27,$5F,$43,$C8,$93,$FE,$00,$94,$7F,$89,$FE,$8F,$FA
	.dc.b	$3C,$FF,$81,$3D,$CF,$F7,$51,$3D,$CF,$F1,$63,$FE,$EB,$27,$C1,$E0
	.dc.b	$E5,$F0,$3F,$E3,$1F,$A1,$70,$D2,$61,$FE,$10,$8F,$73,$76,$E0,$47
	.dc.b	$5E,$87,$32,$38,$3F,$C3,$3F,$47,$F2,$DF,$8A,$7F,$92,$2B,$E1,$37
	.dc.b	$D2,$F2,$1E,$5D,$A4,$78,$DA,$F2,$12,$F6,$3E,$D4,$3E,$8D,$B7,$DB
	.dc.b	$AD,$71,$FF,$40,$6D,$7F,$84,$5A,$F9,$DA,$F1,$F6,$9F,$47,$C5,$B1
	.dc.b	$C5,$AF,$16,$BF,$E1,$DF,$A3,$FE,$3E,$47,$FA,$73,$FD,$3F,$CE,$FC
	.dc.b	$36,$1F,$87,$9F,$F3,$6F,$26,$39,$FE,$55,$59,$FC,$B4,$D7,$73,$FE
	.dc.b	$48,$F3,$FF,$CF,$6F,$F9,$2A,$13,$B9,$CC,$7C,$B1,$CD,$CD,$7C,$45
	.dc.b	$8C,$32,$FA,$7C,$2D,$AE,$4F,$13,$B7,$C0,$59,$F2,$3E,$7C,$DC,$D7
	.dc.b	$97,$0D,$AF,$F3,$0C,$FF,$8C,$DF,$67,$81,$9C,$7F,$05,$A2,$3E,$17
	.dc.b	$D1,$FF,$70,$AF,$F3,$4B,$F0,$1B,$BE,$B7,$91,$F1,$FB,$AE,$3D,$1F
	.dc.b	$99,$F4,$7F,$E5,$D9,$FF,$03,$B6,$7F,$CB,$0D,$7F,$3C,$D7,$F8,$53
	.dc.b	$FE,$00,$8E,$3E,$A7,$DD,$9C,$B5,$FF,$55,$CE,$DF,$F3,$27,$D1,$3E
	.dc.b	$63,$50,$8C,$5F,$F8,$69,$B1,$C2,$77,$1F,$E8,$F9,$D7,$BB,$82,$3C
	.dc.b	$7D,$5D,$91,$C1,$FE,$26,$F9,$FF,$9C,$DD,$FC,$71,$8C,$7F,$CF,$4D
	.dc.b	$7F,$80,$7E,$1D,$FE,$2C,$AF,$F1,$7B,$63,$FD,$00,$D7,$7C,$3E,$EF
	.dc.b	$C3,$BF,$23,$5F,$E3,$66,$CF,$FA,$1D,$AF,$F0,$D6,$BF,$CE,$7E,$C1
	.dc.b	$F6,$3F,$21,$B5,$FE,$40,$6C,$FF,$A4,$5A,$FF,$17,$FA,$C7,$B1,$F4
	.dc.b	$1B,$5F,$E4,$86,$CF,$FA,$65,$AF,$F1,$FF,$A1,$FF,$71,$F9,$FF,$E4
	.dc.b	$E6,$C7,$FA,$7D,$FF,$77,$F3,$E7,$B3,$FC,$BD,$3F,$C9,$FD,$57,$C8
	.dc.b	$7F,$94,$DB,$1F,$E5,$86,$BF,$EA,$96,$BC,$73,$E0,$47,$83,$CF,$E1
	.dc.b	$24,$73,$6B,$FE,$B3,$FB,$1F,$E0,$2F,$F9,$99,$AE,$7C,$2F,$F3,$34
	.dc.b	$62,$3F,$9F,$2F,$0B,$24,$8E,$4F,$0F,$22,$47,$56,$BF,$EC,$3F,$9E
	.dc.b	$FE,$D8,$FF,$49,$6B,$FD,$D5,$A8,$38,$6F,$99,$FE,$7B,$8F,$ED,$71
	.dc.b	$FE,$80,$6B,$DE,$35,$E6,$48,$D3,$5F,$F6,$B7,$C0,$3E,$3F,$FD,$0C
	.dc.b	$D8,$FF,$45,$B5,$DC,$6B,$82,$46,$9A,$F4,$FA,$7E,$64,$79,$BF,$E9
	.dc.b	$3F,$AA,$FF,$34,$C7,$F6,$17,$F8,$23,$F0,$7C,$8F,$E9,$11,$88,$D6
	.dc.b	$1F,$E1,$B9,$CB,$CC,$91,$FE,$A2,$9E,$AF,$F7,$77,$BC,$F0,$8E,$DD
	.dc.b	$1F,$E2,$59,$CB,$FE,$27,$FA,$D4,$77,$8F,$F1,$57,$CE,$7F,$72,$AE
	.dc.b	$27,$D7,$5F,$25,$B8,$C3,$54,$71,$7F,$C0,$0D,$C3,$FB,$33,$CA,$3A
	.dc.b	$9C,$27,$94,$6E,$BF,$C7,$9F,$39,$FE,$F9,$27,$9B,$DD,$FF,$09,$D9
	.dc.b	$CB,$D1,$E1,$E7,$79,$7E,$B7,$CB,$3E,$07,$F6,$79,$7F,$79,$6E,$1F
	.dc.b	$E4,$B7,$73,$FE,$19,$6B,$FE,$25,$6B,$BC,$7F,$DA,$51,$88,$FC,$F3
	.dc.b	$CA,$BB,$4F,$8C,$F9,$1E,$57,$D2,$AF,$87,$F3,$7F,$B3,$6F,$37,$CD
	.dc.b	$8F,$89,$1F,$A9,$74,$FF,$8B,$5B,$7F,$E1,$37,$70,$69,$AC,$7F,$CD
	.dc.b	$1E,$F7,$F9,$22,$7F,$C6,$CD,$AF,$F0,$E3,$67,$7A,$FD,$F8,$C5,$7E
	.dc.b	$ED,$AF,$F9,$EB,$D9,$D4,$FE,$17,$EA,$1F,$5B,$F8,$7E,$27,$9D,$4D
	.dc.b	$44,$C3,$FD,$FF,$5C,$1B,$1F,$EF,$F8,$C4,$7F,$95,$C3,$8C,$FF,$C6
	.dc.b	$8D,$CF,$FD,$0C,$D9,$FE,$25,$C9,$FE,$90,$7F,$27,$C0,$F9,$7F,$C9
	.dc.b	$FD,$94,$9C,$C7,$E5,$8C,$3F,$9E,$8C,$3F,$91,$EC,$FC,$B1,$1F,$CA
	.dc.b	$76,$57,$FC,$C9,$CF,$FC,$73,$4F,$F2,$53,$74,$FD,$0D,$7E,$B7,$F0
	.dc.b	$8F,$CA,$DF,$E3,$3C,$3F,$CE,$19,$F1,$64,$7C,$31,$3F,$D3,$9F,$F5
	.dc.b	$33,$5F,$E6,$4D,$77,$1C,$9F,$F3,$C4,$78,$1F,$F0,$E7,$2F,$E4,$FA
	.dc.b	$FF,$29,$B4,$9F,$CC,$C6,$5F,$CD,$C6,$5F,$CB,$47,$E2,$7A,$E4,$7F
	.dc.b	$02,$AF,$DD,$9F,$05,$FC,$64,$69,$FC,$AB,$F9,$1D,$BF,$88,$7E,$37
	.dc.b	$FE,$2D,$9F,$E1,$6F,$FB,$8D,$BF,$0F,$AF,$52,$46,$08,$7F,$A5,$67
	.dc.b	$93,$63,$AC,$C8,$EB,$1C,$9E,$11,$97,$81,$A3,$AF,$0F,$FA,$16,$BC
	.dc.b	$91,$A2,$46,$67,$93,$BA,$EC,$E6,$79,$F0,$7F,$B7,$EF,$CF,$3B,$9C
	.dc.b	$9F,$DB,$63,$BC,$66,$3B,$CF,$1B,$EE,$F3,$BE,$EE,$67,$98,$BE,$F1
	.dc.b	$B8,$CC,$76,$8C,$47,$63,$C5,$5D,$9F,$25,$76,$72,$72,$70,$F7,$8E
	.dc.b	$0F,$86,$33,$5D,$E3,$89,$DC,$D1,$E4,$9D,$CF,$29,$E1,$3A,$38,$38
	.dc.b	$24,$76,$9E,$0D,$9F,$31,$F7,$4F,$C0,$98,$39,$C7,$07,$57,$F6,$5D
	.dc.b	$9E,$09,$F2,$4F,$9A,$3F,$71,$5F,$92,$7F,$DE,$4D,$7E,$69,$89,$FD
	.dc.b	$F3,$EB,$30,$72,$73,$1F,$62,$3E,$89,$E4,$BF,$C5,$7D,$4F,$86,$3E
	.dc.b	$1B,$70,$FC,$41,$F9,$97,$13,$F7,$A3,$13,$CC,$EA,$72,$30,$7C,$F7
	.dc.b	$9C,$78,$23,$35,$E5,$75,$19,$7E,$31,$AA,$F0,$BD,$AB,$C2,$E8,$F0
	.dc.b	$15,$EE,$FE,$8E,$38,$BD,$8C,$1E,$2B,$FC,$6C,$6A,$FC,$81,$C4,$7F
	.dc.b	$10,$6C,$FF,$10,$6B,$FC,$2D,$B5,$E6,$3C,$71,$E1,$25,$75,$7C,$47
	.dc.b	$BC,$A7,$B2,$38,$C7,$92,$BC,$53,$A9,$CD,$78,$2F,$E5,$47,$F8,$C3
	.dc.b	$E5,$FF,$C5,$BF,$3B,$2F,$FA,$FC,$7F,$8C,$0C,$3F,$E3,$88,$C3,$FE
	.dc.b	$2B,$F7,$FF,$C6,$04,$7F,$C5,$DF,$73,$5F,$EB,$71,$6F,$FA,$DC,$FF
	.dc.b	$36,$F7,$3F,$9B,$FC,$7C,$BF,$E6,$F1,$97,$F4,$2F,$E6,$08,$E0,$D9
	.dc.b	$0F,$E7,$1F,$19,$FE,$25,$EA,$FF,$13,$67,$27,$FC,$57,$19,$7F,$C5
	.dc.b	$B1,$97,$FC,$51,$1F,$E1,$BE,$99,$1F,$D4,$DF,$F0,$F6,$78,$8F,$F1
	.dc.b	$2C,$5C,$FF,$C4,$B3,$FE,$61,$F2,$3F,$C4,$23,$E1,$F0,$67,$C4,$98
	.dc.b	$7F,$75,$9F,$F1,$83,$FE,$A3,$3F,$D4,$8E,$1C,$35,$CC,$72,$56,$E0
	.dc.b	$DC,$F8,$FC,$4F,$F3,$A7,$D0,$E6,$FF,$CD,$ED,$78,$D7,$86,$31,$79
	.dc.b	$5F,$25,$CA,$65,$72,$7A,$C7,$6E,$0E,$FC,$12,$BB,$C6,$5F,$0C,$65
	.dc.b	$F0,$9F,$EC,$3B,$4F,$CB,$E2,$F9,$9B,$4A,$FF,$43,$EA,$9D,$47,$CB
	.dc.b	$7F,$91,$C6,$63,$04,$3B,$87,$4E,$1B,$9F,$C5,$5F,$55,$FE,$9D,$EF
	.dc.b	$FF,$75,$CF,$C9,$6E,$BF,$65,$B7,$FE,$DD,$7F,$CD,$CD,$F4,$B9,$11
	.dc.b	$FF,$50,$C6,$97,$E2,$9F,$4A,$BF,$08,$FD,$28,$FF,$53,$B5,$F5,$BB
	.dc.b	$7F,$DD,$8D,$79,$EF,$FE,$B5,$8E,$18,$7E,$05,$D6,$5F,$EC,$4A,$FC
	.dc.b	$8F,$D2,$FE,$6C,$FC,$E7,$9F,$23,$F3,$0F,$CE,$23,$83,$EA,$BB,$7E
	.dc.b	$67,$9D,$FE,$C5,$F9,$97,$AF,$FD,$61,$B3,$FE,$57,$6B,$FE,$7B,$6B
	.dc.b	$FA,$50,$BE,$AF,$91,$23,$93,$D7,$CA,$91,$D5,$AF,$FA,$1F,$E6,$7F
	.dc.b	$6C,$FF,$99,$DA,$E7,$AB,$FB,$6C,$62,$3C,$4F,$EC,$D3,$49,$D5,$C3
	.dc.b	$AA,$4F,$56,$BF,$E9,$7F,$99,$FF,$47,$3F,$E7,$76,$B9,$F2,$7F,$BE
	.dc.b	$C6,$23,$E1,$2E,$EC,$92,$39,$3C,$3B,$E4,$75,$6B,$FE,$A7,$F9,$47
	.dc.b	$D9,$FF,$82,$9B,$1F,$E8,$96,$BB,$8D,$7C,$F4,$8D,$35,$E9,$F3,$3C
	.dc.b	$C8,$F3,$7F,$C2,$FE,$73,$E3,$29,$27,$FD,$31,$19,$7F,$93,$7C,$5F
	.dc.b	$EF,$08,$E0,$FF,$4F,$7E,$28,$CF,$FC,$46,$D8,$CC,$60,$FF,$13,$F8
	.dc.b	$8E,$0A,$49,$FF,$54,$46,$5B,$E7,$F0,$23,$83,$7E,$7F,$42,$38,$39
	.dc.b	$F9,$3E,$02,$38,$3B,$F8,$FE,$42,$38,$3C,$7E,$3F,$A0,$8E,$0F,$3F
	.dc.b	$8F,$EC,$23,$83,$D7,$F0,$FF,$D3,$7C,$4F,$E8,$19,$FF,$1C,$91,$FC
	.dc.b	$91,$C3,$67,$87,$8D,$FD,$57,$E6,$1A,$E2,$B8,$C6,$1E,$78,$7F,$4D
	.dc.b	$8C,$47,$F3,$07,$FD,$78,$9A,$4D,$7A,$43,$E6,$1C,$58,$F7,$5F,$21
	.dc.b	$49,$5E,$0C,$FF,$B5,$4B,$E4,$B8,$12,$74,$7C,$08,$D3,$83,$C1,$F3
	.dc.b	$1E,$2F,$96,$FE,$07,$AF,$F6,$76,$E1,$FD,$DD,$B5,$F2,$97,$83,$C4
	.dc.b	$91,$D6,$32,$DE,$6E,$C4,$70,$78,$FE,$63,$F9,$06,$1E,$1E,$1F,$21
	.dc.b	$27,$A7,$07,$9C,$8E,$0F,$07,$3F,$F8,$19,$7C,$C3,$6F,$54,$E6,$7F
	.dc.b	$97,$66,$8C,$46,$B1,$FE,$64,$FB,$E3,$2F,$C1,$E8,$7C,$C6,$A1,$69
	.dc.b	$CC,$72,$8C,$4E,$DB,$1F,$E9,$CF,$98,$FF,$4E,$A6,$91,$ED,$F1,$BF
	.dc.b	$E1,$DC,$7F,$C3,$91,$F8,$6E,$1F,$B8,$A7,$51,$FD,$2F,$92,$CF,$F8
	.dc.b	$C5,$AB,$E4,$DF,$15,$88,$E8,$7F,$82,$A3,$35,$DB,$9B,$5E,$B6,$DB
	.dc.b	$1E,$28,$E5,$F1,$98,$BF,$F0,$98,$73,$3F,$93,$B7,$C5,$E2,$BC,$0F
	.dc.b	$9B,$4F,$E3,$55,$F9,$97,$F6,$AD,$E7,$D7,$34,$8E,$9F,$2D,$70,$F5
	.dc.b	$3F,$03,$E6,$3E,$34,$FE,$8D,$F7,$E1,$FE,$47,$73,$5F,$E2,$A7,$33
	.dc.b	$FE,$59,$6C,$7F,$0C,$CB,$FC,$77,$3C,$F7,$F9,$B1,$89,$FF,$19,$FC
	.dc.b	$C7,$F4,$3A,$FC,$43,$63,$F1,$ED,$7F,$C0,$4D,$7F,$CD,$0F,$F8,$32
	.dc.b	$38,$FA,$9F,$82,$9F,$E3,$B3,$97,$9C,$65,$7C,$0B,$E2,$F0,$DB,$FC
	.dc.b	$D3,$67,$F4,$2D,$7F,$C2,$8F,$F1,$68,$DF,$81,$C8,$8F,$E8,$FE,$CF
	.dc.b	$D1,$5F,$CC,$1A,$9A,$63,$E6,$3F,$25,$3F,$AB,$6C,$FE,$C9,$AF,$F8
	.dc.b	$95,$AF,$FA,$2F,$E6,$1C,$BE,$F6,$D7,$ED,$5B,$3F,$BC,$6B,$FE,$2F
	.dc.b	$7F,$C5,$D1,$BF,$63,$91,$1F,$DF,$7C,$BB,$E2,$FE,$63,$63,$31,$83
	.dc.b	$E8,$7C,$B7,$E7,$3E,$5F,$F8,$33,$71,$FE,$1A,$FE,$C3,$B7,$3D,$4A
	.dc.b	$FE,$19,$F5,$EF,$B0,$FE,$1D,$82,$30,$EA,$33,$18,$8F,$E3,$3C,$9E
	.dc.b	$77,$F8,$B3,$5C,$E5,$D2,$31,$1F,$E5,$15,$F3,$39,$24,$72,$7C,$7F
	.dc.b	$81,$FF,$82,$63,$F9,$3B,$5F,$F2,$D3,$5F,$F6,$03,$5F,$EE,$7F,$35
	.dc.b	$FC,$0D,$FE,$54,$D7,$F9,$8B,$5F,$F3,$23,$5F,$F6,$53,$FC,$F5,$E3
	.dc.b	$E4,$7C,$68,$C4,$72,$8F,$93,$F8,$1F,$C5,$8F,$F3,$73,$5F,$F6,$A3
	.dc.b	$5F,$F0,$3B,$57,$CD,$BC,$CF,$A6,$D8,$FE,$84,$D7,$FA,$4B,$5F,$F3
	.dc.b	$E3,$FD,$3D,$DF,$97,$C8,$8F,$F4,$BE,$57,$D0,$FE,$01,$B1,$98,$C1
	.dc.b	$F8,$4F,$87,$FE,$53,$1F,$D4,$DA,$FF,$5A,$6B,$FE,$8D,$6B,$FE,$F3
	.dc.b	$7F,$D8,$CF,$1F,$89,$E8,$47,$FA,$F7,$D8,$7A,$5F,$EC,$AD,$7F,$D3
	.dc.b	$0D,$7F,$DF,$CD,$7F,$C6,$3F,$32,$F9,$AF,$F1,$0F,$CD,$FF,$13,$9F
	.dc.b	$F0,$A3,$5F,$F6,$CC,$ED,$FF,$0C,$35,$FE,$EA,$D7,$FD,$50,$D7,$FC
	.dc.b	$7E,$1C,$3E,$29,$FF,$18,$C7,$3F,$21,$D9,$C2,$30,$F2,$F9,$B7,$CC
	.dc.b	$FF,$8E,$3E,$33,$FD,$33,$7C,$31,$FF,$26,$C7,$FA,$5D,$FF,$17,$35
	.dc.b	$FF,$79,$BC,$2F,$CA,$D8,$E3,$F1,$1D,$4E,$DF,$15,$FE,$46,$F9,$BF
	.dc.b	$F2,$59,$39,$71,$C9,$96,$5F,$F8,$F6,$77,$3F,$E4,$EF,$85,$7C,$9F
	.dc.b	$F9,$43,$E7,$7F,$81,$6B,$E9,$C1,$18,$D7,$E4,$DF,$09,$FF,$24,$35
	.dc.b	$FF,$12,$B5,$FF,$6C,$35,$E3,$7F,$EA,$23,$C4,$47,$FC,$A5,$F2,$3F
	.dc.b	$31,$FF,$2A,$7C,$EB,$E7,$7F,$CC,$AF,$C5,$83,$4D,$3C,$23,$7B,$48
	.dc.b	$C1,$FE,$5F,$E1,$FF,$3A,$E9,$EF,$6E,$7F,$86,$6C,$FF,$A1,$9F,$F3
	.dc.b	$C9,$87,$F0,$3E,$B7,$F9,$E2,$30,$7F,$9F,$BE,$73,$FB,$6C,$FF,$9B
	.dc.b	$57,$CE,$FF,$9F,$9B,$19,$F4,$B0,$7F,$3C,$8E,$31,$FE,$50,$6B,$F9
	.dc.b	$A0,$E1,$FC,$EB,$C3,$1F,$B4,$DC,$73,$5E,$1F,$9D,$FC,$C7,$C6,$7E
	.dc.b	$67,$E7,$66,$3F,$96,$8C,$BF,$98,$8C,$BF,$98,$8F,$C9,$7C,$1C,$8F
	.dc.b	$C7,$3F,$95,$F8,$DC,$43,$3F,$C6,$73,$FF,$98,$E3,$67,$E5,$67,$E2
	.dc.b	$7C,$67,$C5,$F9,$D9,$3F,$26,$32,$FF,$9B,$76,$47,$C7,$24,$65,$DB
	.dc.b	$E1,$DB,$E2,$C7,$D1,$F9,$DF,$C5,$F6,$7E,$33,$59,$7F,$90,$8C,$BF
	.dc.b	$92,$8C,$BC,$CF,$F6,$4F,$37,$EE,$F5,$FC,$57,$5F,$E0,$C9,$3F,$8F
	.dc.b	$DE,$BC,$4F,$E1,$31,$89,$FC,$94,$FF,$B0,$BC,$3F,$F5,$3D,$6D,$FF
	.dc.b	$57,$A6,$9A,$9F,$DC,$37,$4F,$E3,$EE,$BE,$63,$A7,$C5,$CB,$FF,$60
	.dc.b	$4E,$5B,$19,$7C,$13,$E3,$31,$7D,$A7,$13,$CA,$73,$7E,$09,$CC,$F2
	.dc.b	$3A,$90,$C1,$2B,$A3,$97,$D7,$5C,$9C,$C7,$C3,$9C,$57,$BA,$BC,$33
	.dc.b	$FE,$FC,$BF,$84,$D8,$DB,$9A,$DB,$F2,$E3,$2F,$CD,$8C,$CF,$C5,$8F
	.dc.b	$01,$E5,$AF,$03,$F0,$4C,$1D,$4C,$1E,$D7,$E2,$3A,$3C,$06,$AF,$DA
	.dc.b	$48,$EB,$1C,$4E,$87,$88,$C1,$C8,$C4,$71,$7E,$01,$D6,$3C,$2E,$9E
	.dc.b	$B1,$F3,$A3,$C9,$3F,$21,$E9,$19,$78,$1E,$27,$04,$9E,$AF,$86,$38
	.dc.b	$C7,$63,$24,$7C,$A7,$AA,$7C,$04,$BF,$A6,$E2,$FF,$CB,$EF,$E9,$27
	.dc.b	$F9,$13,$5F,$69,$88,$D9,$1F,$F7,$A4,$6D,$F5,$57,$8D,$FC,$01,$FA
	.dc.b	$4F,$83,$FC,$0D,$7E,$09,$B1,$F1,$9A,$FC,$43,$F6,$8E,$30,$F8,$57
	.dc.b	$FE,$63,$6D,$73,$AF,$9E,$D7,$94,$FB,$DF,$11,$A3,$C0,$53,$75,$DE
	.dc.b	$FD,$53,$DD,$B5,$D6,$FC,$37,$C8,$EF,$19,$9D,$C6,$67,$6F,$63,$C0
	.dc.b	$FB,$2F,$C1,$1F,$28,$FD,$83,$87,$E5,$9F,$C6,$5C,$5F,$DE,$6D,$FF
	.dc.b	$1F,$6C,$FC,$03,$CA,$78,$EF,$CA,$F8,$CF,$E5,$5F,$3B,$FE,$17,$C3
	.dc.b	$FC,$30,$D9,$FF,$1F,$B6,$41,$7C,$EF,$F9,$12,$B8,$D5,$38,$96,$B8
	.dc.b	$D7,$37,$95,$65,$E9,$C3,$65,$C3,$4C,$44,$9F,$1E,$1E,$89,$D4,$63
	.dc.b	$0F,$4D,$7F,$43,$E4,$FE,$89,$B4,$AF,$E7,$31,$97,$F3,$AF,$E9,$88
	.dc.b	$E0,$D9,$0F,$E9,$9F,$25,$E9,$F0,$3F,$DC,$F1,$E1,$3E,$3B,$87,$FA
	.dc.b	$C3,$72,$FF,$23,$3F,$EB,$56,$FB,$BC,$C9,$3C,$CF,$F2,$7F,$C3,$FC
	.dc.b	$7F,$9D,$F9,$47,$D2,$8D,$35,$FE,$CC,$EB,$4E,$1B,$C9,$DC,$7F,$67
	.dc.b	$F3,$7F,$B4,$7C,$EC,$9F,$EB,$B1,$97,$F8,$0B,$FC,$28,$8E,$0D,$90
	.dc.b	$FE,$E1,$E2,$FD,$DF,$AB,$FF,$51,$C7,$F7,$96,$E9,$FE,$67,$6E,$1F
	.dc.b	$E8,$E5,$F4,$9C,$79,$E5,$FC,$39,$DB,$F9,$47,$6E,$78,$FC,$1C,$92
	.dc.b	$30,$43,$C4,$E5,$F2,$47,$E4,$FB,$9C,$97,$92,$57,$D3,$9C,$B6,$7F
	.dc.b	$C9,$8F,$FA,$9A,$32,$E1,$AF,$F9,$45,$AE,$E3,$C0,$47,$B7,$C2,$FE
	.dc.b	$97,$EA,$BF,$C4,$DA,$FF,$AC,$9B,$3F,$B4,$6A,$5D,$DC,$7B,$E4,$FE
	.dc.b	$2E,$76,$69,$FC,$77,$D5,$FE,$27,$87,$FD,$34,$BE,$C7,$F6,$CD,$9D
	.dc.b	$F3,$60,$CC,$C8,$C9,$FE,$9E,$F5,$B8,$FC,$E4,$8C,$10,$FF,$50,$C7
	.dc.b	$CA,$7F,$87,$F9,$C8,$C4,$23,$0E,$DA,$FF,$91,$5B,$8F,$FA,$CD,$BA
	.dc.b	$FF,$B6,$B2,$7A,$BF,$C4,$7E,$8B,$8E,$1F,$E8,$56,$BF,$EE,$06,$BF
	.dc.b	$C8,$9A,$97,$A7,$E6,$6C,$7A,$DA,$FE,$19,$AF,$FA,$4F,$85,$F3,$FE
	.dc.b	$E6,$C7,$C3,$6B,$F8,$E6,$BF,$E9,$BF,$A2,$FE,$E8,$9A,$4F,$1B,$E4
	.dc.b	$AF,$1A,$92,$7E,$84,$65,$BD,$DC,$08,$E0,$FA,$DF,$45,$FE,$F3,$7F
	.dc.b	$D5,$8D,$7F,$A3,$87,$0F,$F4,$A6,$B9,$E7,$FB,$B1,$F9,$38,$FD,$5B
	.dc.b	$5E,$B1,$AF,$81,$91,$E2,$6B,$FB,$2F,$A2,$7D,$17,$E2,$1B,$1F,$8D
	.dc.b	$6B,$B8,$D7,$34,$8D,$35,$E9,$DD,$CC,$8F,$37,$FC,$5B,$B4,$8F,$F1
	.dc.b	$3B,$5F,$E0,$BF,$43,$EB,$7E,$2B,$5F,$ED,$AD,$4D,$FC,$2D,$CD,$FA
	.dc.b	$26,$CF,$E9,$DA,$FF,$12,$6B,$FE,$DF,$B7,$AF,$FA,$A6,$C7,$EB,$DA
	.dc.b	$FF,$1A,$6B,$C7,$DB,$E0,$47,$83,$CF,$E5,$24,$73,$6B,$CF,$CE,$F0
	.dc.b	$11,$F0,$3E,$3F,$0A,$47,$8D,$AF,$8F,$EA,$BD,$04,$7D,$0F,$AF,$E8
	.dc.b	$FF,$C1,$58,$BC,$6A,$62,$25,$7B,$DE,$0D,$3F,$88,$7D,$1F,$C2,$F0
	.dc.b	$7E,$5B,$77,$FC,$DB,$67,$FC,$46,$B5,$FF,$0D,$36,$3F,$BC,$35,$FF
	.dc.b	$6C,$36,$BE,$BF,$D1,$B8,$BF,$C4,$5F,$10,$F9,$9A,$4C,$3F,$C0,$50
	.dc.b	$F9,$CD,$92,$32,$FF,$8A,$ED,$F4,$7F,$E2,$DF,$A0,$FF,$70,$93,$D9
	.dc.b	$C7,$85,$30,$FF,$07,$4F,$FA,$64,$9E,$17,$12,$46,$9F,$F1,$C7,$D1
	.dc.b	$FE,$FA,$3F,$DE,$93,$F1,$53,$49,$C1,$E3,$24,$78,$1F,$37,$B6,$E0
	.dc.b	$FF,$27,$FD,$1B,$7F,$E1,$78,$2B,$58,$7F,$70,$8C,$5F,$F4,$96,$C6
	.dc.b	$67,$6F,$9E,$1D,$2F,$A3,$E0,$D9,$E8,$D7,$83,$5F,$EF,$FF,$47,$FD
	.dc.b	$51,$34,$8F,$F5,$BF,$99,$FD,$BE,$3F,$B4,$E4,$BC,$FF,$26,$6D,$7F
	.dc.b	$28,$9A,$FF,$9C,$1B,$DA,$FA,$3C,$CF,$E1,$EF,$FC,$8D,$39,$7F,$C9
	.dc.b	$11,$F8,$76,$FB,$DC,$3A,$38,$B2,$37,$EF,$5A,$FA,$51,$F7,$63,$B9
	.dc.b	$E2,$F9,$AB,$1E,$07,$CF,$A7,$F4,$D8,$FF,$2C,$3F,$C4,$1B,$CD,$AF
	.dc.b	$0A,$47,$A3,$E8,$EE,$3E,$03,$FD,$1F,$A2,$3E,$1F,$D1,$B8,$2F,$06
	.dc.b	$1B,$87,$F9,$9E,$7E,$26,$1C,$C6,$0F,$0B,$71,$98,$C5,$7F,$9B,$FE
	.dc.b	$8D,$F3,$3F,$CA,$AB,$FD,$43,$18,$9F,$F0,$B3,$63,$FC,$82,$D7,$F4
	.dc.b	$9F,$6F,$F4,$DF,$F2,$2B,$5C,$F2,$75,$51,$97,$FB,$4F,$C6,$7F,$05
	.dc.b	$BF,$F1,$23,$63,$FC,$9E,$D7,$F6,$4D,$7F,$B8,$FD,$1B,$E9,$7B,$36
	.dc.b	$BF,$C5,$CD,$9F,$F2,$CB,$5F,$DE,$7D,$BF,$F9,$CA,$7F,$C6,$8D,$8F
	.dc.b	$F2,$FB,$5D,$FC,$2F,$AF,$E9,$1D,$34,$97,$FE,$3D,$8C,$B7,$DE,$E0
	.dc.b	$47,$07,$F9,$0F,$CD,$7C,$37,$C7,$6C,$66,$30,$7C,$BF,$A3,$76,$3E
	.dc.b	$8F,$FC,$99,$18,$6F,$17,$E2,$F5,$FF,$3C,$CE,$5F,$F3,$DF,$D1,$3E
	.dc.b	$8F,$FC,$A2,$D8,$FF,$2B,$35,$DC,$7F,$95,$A9,$EE,$52,$4F,$F9,$76
	.dc.b	$32,$FF,$97,$F9,$DE,$D7,$F9,$79,$B1,$98,$C1,$FE,$62,$FA,$2F,$F1
	.dc.b	$B4,$7F,$A4,$1A,$FF,$33,$6B,$FE,$32,$6B,$FE,$FA,$FB,$D7,$CC,$FF
	.dc.b	$9B,$1B,$1F,$E7,$16,$BF,$E9,$C6,$BF,$CF,$FE,$8F,$F9,$A3,$FE,$77
	.dc.b	$6B,$FE,$A2,$6B,$BF,$37,$FC,$F3,$F4,$7F,$1E,$3F,$A4,$35,$FF,$23
	.dc.b	$B5,$07,$0F,$F0,$3F,$63,$FD,$07,$E2,$9A,$48,$FF,$45,$7D,$1F,$F4
	.dc.b	$76,$FA,$1F,$E8,$A8,$C4,$7F,$AE,$1A,$FF,$5B,$E9,$7D,$1F,$FA,$41
	.dc.b	$B1,$FE,$95,$6B,$FE,$C0,$6B,$FD,$8F,$ED,$9E,$17,$FA,$61,$B1,$FE
	.dc.b	$9D,$6B,$FE,$C8,$6B,$FD,$AF,$E4,$FF,$DC,$ED,$EB,$FF,$50,$C6,$23
	.dc.b	$FC,$DA,$D7,$E3,$FD,$27,$A9,$FF,$38,$B5,$CE,$7F,$EA,$A8,$C4,$3E
	.dc.b	$D9,$F4,$3F,$EA,$E6,$C7,$FA,$CD,$AF,$FB,$79,$AF,$F8,$07,$E7,$3D
	.dc.b	$7E,$8F,$F9,$49,$7D,$07,$FA,$0A,$30,$DF,$0B,$FD,$0B,$3C,$1B,$B7
	.dc.b	$FA,$FA,$3E,$1D,$E0,$FF,$62,$38,$7F,$0A,$F1,$6C,$75,$69,$F8,$AB
	.dc.b	$F2,$7D,$06,$63,$FD,$6F,$F1,$57,$E3,$DF,$41,$4C,$44,$9E,$59,$7F
	.dc.b	$A7,$2B,$71,$83,$F2,$EE,$1F,$EA,$0F,$FA,$6F,$0F,$F3,$AF,$D0,$7F
	.dc.b	$84,$27,$80,$70,$FF,$57,$6C,$70,$F8,$8F,$44,$62,$FF,$D5,$AD,$AF
	.dc.b	$D3,$B5,$E9,$3E,$FD,$3F,$B0,$47,$81,$7C,$CE,$4D,$7A,$92,$7E,$7B
	.dc.b	$F4,$C8,$E0,$E4,$ED,$FE,$DD,$2F,$9A,$F9,$DF,$40,$F9,$BB,$B4,$BF
	.dc.b	$83,$39,$6C,$E5,$FE,$F1,$1F,$27,$4F,$F6,$5B,$B8,$FE,$F9,$F4,$3F
	.dc.b	$BA,$DB,$FD,$42,$DA,$FF,$68,$B6,$7F,$85,$BF,$14,$C2,$69,$3E,$07
	.dc.b	$F0,$DD,$4F,$87,$BF,$11,$2F,$EC,$CE,$5B,$3F,$E6,$37,$FD,$C3,$1F
	.dc.b	$EF,$42,$46,$63,$F8,$EC,$71,$7F,$C2,$CD,$C7,$FD,$76,$DC,$79,$FC
	.dc.b	$07,$E6,$7E,$87,$FD,$7F,$5F,$EC,$06,$CF,$FB,$C1,$B7,$FC,$A9,$7C
	.dc.b	$E7,$19,$93,$F9,$29,$DA,$76,$E7,$8B,$C1,$C3,$23,$04,$3F,$C5,$D3
	.dc.b	$F8,$17,$A4,$79,$E7,$E5,$F2,$7C,$CD,$A5,$7C,$A8,$CB,$F1,$74,$47
	.dc.b	$C7,$8F,$F0,$0B,$63,$31,$B8,$E2,$FC,$A7,$EC,$FC,$7F,$EF,$FE,$87
	.dc.b	$FC,$27,$3F,$B0,$6B,$FE,$45,$6E,$BF,$EE,$A5,$F0,$BC,$7D,$A9,$5C
	.dc.b	$5E,$91,$FC,$2F,$E8,$7E,$49,$F2,$27,$FC,$A0,$DB,$FF,$7A,$BA,$DF
	.dc.b	$87,$04,$AF,$E0,$73,$F4,$B4,$78,$9F,$92,$FA,$19,$5F,$90,$8F,$CA
	.dc.b	$98,$7F,$13,$18,$74,$7E,$59,$D6,$3F,$8C,$9D,$4F,$03,$47,$03,$F2
	.dc.b	$6F,$0C,$7F,$37,$5C,$2B,$95,$70,$39,$1C,$0E,$4F,$E6,$A3,$94,$7E
	.dc.b	$82,$B9,$57,$48,$E5,$5D,$1E,$47,$47,$93,$D2,$3F,$5B,$3D,$23,$FC
	.dc.b	$8B,$5D,$1E,$C7,$49,$EC,$F4,$AE,$C7,$47,$B4,$74,$7B,$1F,$B6,$9E
	.dc.b	$D7,$F5,$23,$27,$80,$EC,$78,$0E,$D3,$E0,$AE,$C7,$81,$ED,$1E,$03
	.dc.b	$B1,$D2,$7B,$39,$3B,$1D,$0E,$D3,$D2,$7B,$1D,$1E,$D1,$D0,$E5,$3D
	.dc.b	$23,$31,$D1,$E4,$F4,$9E,$47,$47,$94,$70,$9E,$53,$99,$E4,$F0,$79
	.dc.b	$1C,$1D,$47,$08,$C9,$C0,$D1,$87,$E0,$BF,$E4,$23,$E6,$14,$FF,$41
	.dc.b	$DF,$F0,$32,$3F,$C2,$5A,$FF,$33,$8E,$6F,$94,$E2,$7C,$FF,$57,$F1
	.dc.b	$0E,$0F,$9C,$7F,$A4,$1C,$C7,$A6,$35,$7F,$CD,$A7,$35,$A6,$CF,$53
	.dc.b	$EE,$46,$5A,$64,$EF,$5D,$49,$3C,$A3,$2D,$AF,$A4,$DB,$F6,$57,$FA
	.dc.b	$32,$B9,$DF,$76,$9D,$EF,$ED,$46,$5B,$19,$AE,$73,$96,$C6,$4F,$AD
	.dc.b	$19,$69,$93,$E2,$C7,$07,$FA,$ED,$F9,$DF,$BC,$D7,$67,$F6,$47,$17
	.dc.b	$F0,$A3,$E3,$9F,$08,$FF,$14,$FD,$0F,$FB,$11,$B1,$FE,$3D,$8C,$47
	.dc.b	$FC,$01,$5E,$F6,$66,$4F,$03,$9C,$71,$E3,$E5,$C9,$27,$BC,$65,$F0
	.dc.b	$C6,$5E,$D1,$FE,$4D,$EE,$FE,$AB,$0D,$19,$FE,$6E,$5C,$7B,$38,$72
	.dc.b	$4B,$FF,$03,$4F,$27,$F4,$0D,$79,$47,$F6,$58,$EF,$7D,$A7,$FD,$53
	.dc.b	$18,$BD,$C7,$F9,$42,$1F,$25,$FE,$63,$FA,$1F,$F1,$65,$F1,$A4,$62
	.dc.b	$FF,$64,$D8,$FF,$5A,$BF,$E3,$56,$F7,$73,$23,$FE,$5E,$8D,$2F,$9F
	.dc.b	$FE,$C9,$5F,$BD,$5F,$43,$FE,$75,$6C,$FF,$99,$9D,$BF,$EC,$46,$BC
	.dc.b	$E3,$81,$C7,$9B,$FC,$FD,$F4,$2F,$3F,$A7,$CD,$4B,$FF,$38,$CE,$5B
	.dc.b	$3F,$E0,$F8,$C4,$7F,$8E,$DA,$FF,$11,$6B,$FC,$55,$DE,$9F,$25,$B9
	.dc.b	$66,$BD,$9F,$43,$FE,$91,$FB,$17,$CB,$52,$63,$FE,$84,$8F,$F2,$A9
	.dc.b	$A7,$11,$A7,$F1,$97,$FC,$96,$77,$F3,$1F,$E9,$DF,$A7,$FE,$4D,$9F
	.dc.b	$DB,$6D,$6E,$7C,$46,$EE,$1E,$17,$97,$DA,$73,$F5,$E4,$FF,$A6,$2D
	.dc.b	$3A,$8F,$F2,$E4,$61,$B1,$FE,$39,$8C,$47,$F9,$95,$B8,$79,$67,$94
	.dc.b	$FB,$0F,$F8,$03,$2F,$B3,$FF,$5C,$7D,$9B,$E1,$A9,$2B,$FD,$57,$0F
	.dc.b	$AA,$C4,$73,$11,$C5,$DC,$F4,$9F,$F6,$27,$D9,$BE,$C6,$F9,$D2,$FF
	.dc.b	$D6,$F3,$96,$CE,$5E,$D0,$9E,$73,$D3,$C6,$EB,$E3,$4B,$E7,$19,$7F
	.dc.b	$D0,$96,$8F,$F3,$03,$63,$31,$83,$83,$E0,$7F,$DB,$9F,$52,$F9,$9B
	.dc.b	$EB,$7F,$9A,$60,$D2,$57,$FA,$4F,$C0,$7D,$4F,$F9,$BD,$B1,$98,$C1
	.dc.b	$9F,$6A,$47,$16,$BF,$EA,$1F,$A9,$7A,$7F,$E7,$96,$C7,$FA,$85,$AE
	.dc.b	$E3,$5E,$6C,$8D,$35,$FF,$56,$7D,$4B,$E7,$3F,$D0,$8D,$8F,$F5,$63
	.dc.b	$5D,$C6,$BD,$99,$1A,$6B,$D3,$E3,$79,$91,$E6,$FF,$A4,$7E,$A7,$F8
	.dc.b	$9C,$7F,$AF,$5A,$FF,$84,$DA,$FF,$85,$DA,$82,$F1,$7A,$12,$3F,$D8
	.dc.b	$91,$96,$F9,$DE,$12,$38,$3F,$D9,$4B,$EC,$BF,$D4,$CD,$8C,$C6,$0F
	.dc.b	$F5,$57,$D4,$B1,$A4,$9F,$F6,$A4,$65,$FF,$6B,$7D,$4F,$F4,$37,$FC
	.dc.b	$60,$D7,$FC,$68,$D7,$DC,$D5,$E4,$FF,$00,$7F,$94,$78,$3F,$CB,$48
	.dc.b	$FA,$BC,$7F,$49,$1C,$1E,$EF,$9F,$F6,$91,$C1,$F0,$BA,$BE,$09,$1C
	.dc.b	$1F,$1B,$D1,$F8,$A4,$70,$7C,$AF,$99,$F9,$24,$70,$7C,$EE,$EF,$9A
	.dc.b	$47,$07,$D2,$F8,$CF,$A2,$47,$04,$3F,$89,$70,$23,$FC,$B9,$E3,$FF
	.dc.b	$6F,$8C,$4F,$E1,$E3,$1C,$FA,$F2,$BF,$CC,$5F,$52,$FA,$2C,$5F,$E3
	.dc.b	$63,$0F,$19,$DC,$EA,$13,$FD,$2D,$FE,$48,$DB,$FF,$00,$36,$BB,$EF
	.dc.b	$C7,$EA,$7F,$C0,$F7,$D4,$FC,$E7,$C9,$7F,$0F,$70,$1C,$3F,$CC,$5B
	.dc.b	$87,$F8,$35,$7D,$27,$95,$B3,$C9,$AF,$18,$F1,$79,$FE,$BF,$A7,$FE
	.dc.b	$5D,$3E,$06,$BF,$CF,$DB,$7F,$E1,$B5,$F0,$BC,$49,$3C,$5E,$9E,$0C
	.dc.b	$8E,$8D,$7F,$A5,$7D,$4F,$F6,$A7,$F7,$95,$F4,$1F,$06,$FE,$53,$CD
	.dc.b	$AC,$8F,$33,$5F,$37,$C8,$FF,$91,$57,$D4,$FC,$97,$AC,$65,$FF,$50
	.dc.b	$5A,$31,$1F,$EA,$58,$C6,$7F,$EA,$58,$E2,$D5,$0F,$87,$7D,$4E,$63
	.dc.b	$0D,$8C,$CF,$39,$CB,$CC,$7D,$4D,$69,$97,$06,$FC,$6F,$F2,$A6,$1F
	.dc.b	$C2,$FC,$EB,$CB,$FE,$BB,$9B,$7F,$EB,$F9,$AF,$F2,$96,$F7,$3E,$7B
	.dc.b	$33,$FD,$7E,$FF,$23,$39,$7F,$27,$1F,$8E,$6F,$C4,$F0,$E6,$E2,$C8
	.dc.b	$DF,$C4,$DA,$FF,$04,$07,$B9,$E2,$FA,$2B,$3E,$07,$CF,$A7,$F6,$08
	.dc.b	$FD,$1B,$FB,$76,$F3,$6B,$96,$47,$A3,$E6,$6E,$3F,$E1,$7D,$11,$F0
	.dc.b	$BE,$A7,$FB,$05,$7E,$B9,$F8,$15,$FE,$1D,$72,$DD,$F8,$1F,$23,$BE
	.dc.b	$D7,$C6,$7E,$6F,$D4,$FF,$99,$63,$F0,$CD,$AF,$C5,$35,$FF,$78,$35
	.dc.b	$FF,$7A,$3F,$E0,$C8,$E3,$EA,$7F,$43,$8C,$4F,$28,$FE,$8D,$E9,$3B
	.dc.b	$A9,$2F,$F2,$D1,$97,$F3,$1F,$4F,$FD,$A5,$B7,$FD,$3E,$30,$FE,$7D
	.dc.b	$AF,$F6,$F3,$FB,$83,$F1,$1C,$07,$0E,$DA,$F0,$E6,$FE,$B9,$F6,$3F
	.dc.b	$CD,$EB,$F5,$4D,$9F,$D7,$35,$EA,$D7,$C5,$F5,$EF,$80,$FE,$CC,$DA
	.dc.b	$FD,$A3,$67,$F7,$0D,$7C,$FF,$4C,$E6,$7D,$3F,$EE,$E3,$0D,$EB,$FF
	.dc.b	$10,$57,$C1,$9C,$BF,$C0,$FE,$BF,$FC,$C8,$FF,$03,$6B,$9F,$80,$FD
	.dc.b	$E4,$62,$3E,$3F,$CC,$38,$3F,$80,$36,$3F,$86,$35,$FE,$20,$D7,$E7
	.dc.b	$FC,$97,$F3,$07,$F8,$93,$5F,$E2,$CD,$77,$E4,$7F,$0D,$FA,$7F,$F4
	.dc.b	$04,$7F,$19,$6B,$FC,$75,$AF,$DC,$6B,$F8,$1F,$C5,$BF,$C2,$44,$E0
	.dc.b	$FE,$30,$E3,$E3,$29,$27,$F9,$34,$23,$F8,$F7,$D2,$BE,$86,$92,$BF
	.dc.b	$96,$46,$5B,$F4,$FC,$08,$E0,$DF,$C9,$F4,$23,$83,$F9,$B7,$D2,$FF
	.dc.b	$35,$6F,$97,$FC,$BE,$31,$1F,$CF,$5A,$FE,$71,$7B,$1B,$24,$74,$79
	.dc.b	$7C,$B4,$8E,$ED,$7F,$A4,$7B,$0F,$98,$FE,$7C,$D8,$FE,$94,$D7,$71
	.dc.b	$AF,$87,$91,$A6,$BD,$3E,$BB,$99,$1E,$6F,$F4,$BF,$93,$FF,$06,$C7
	.dc.b	$FA,$DD,$FE,$0C,$FF,$83,$1A,$E7,$E2,$BF,$C3,$91,$C2,$3F,$7A,$D7
	.dc.b	$F8,$13,$5F,$F3,$93,$53,$E0,$F3,$9C,$9C,$2B,$A9,$FE,$29,$FB,$07
	.dc.b	$CC,$FF,$8A,$FC,$8F,$F6,$CD,$7F,$0F,$6E,$3F,$E7,$F6,$BF,$EB,$0F
	.dc.b	$94,$E1,$F0,$72,$3F,$C4,$F3,$A1,$3F,$E3,$7F,$A1,$70,$DF,$82,$97
	.dc.b	$FE,$2E,$9C,$B6,$72,$FF,$B7,$E3,$FC,$96,$D8,$FF,$21,$7D,$83,$E3
	.dc.b	$6F,$CB,$4A,$FF,$1C,$CE,$5B,$3F,$DF,$E3,$11,$FE,$0D,$6B,$FC,$B1
	.dc.b	$AF,$F3,$17,$7C,$1F,$32,$3F,$D4,$51,$8D,$75,$5F,$E5,$5F,$85,$F9
	.dc.b	$B5,$FE,$4A,$FA,$27,$D1,$FF,$96,$5B,$19,$8C,$70,$65,$91,$93,$94
	.dc.b	$7F,$98,$7E,$8F,$F4,$B1,$FE,$56,$7E,$4A,$CF,$CC,$A3,$31,$87,$8C
	.dc.b	$6F,$2F,$E9,$D1,$89,$D6,$BD,$CF,$07,$57,$F9,$C7,$E2,$DF,$5F,$FE
	.dc.b	$73,$FC,$5B,$FC,$95,$91,$B3,$A8,$FF,$35,$B5,$FF,$1E,$35,$FE,$B6
	.dc.b	$D7,$FB,$03,$5F,$F6,$83,$C7,$0F,$6B,$72,$DD,$FE,$77,$D8,$FC,$E3
	.dc.b	$2F,$F2,$AF,$E7,$FC,$7C,$47,$C1,$3F,$3D,$18,$C3,$E4,$F3,$AB,$F3
	.dc.b	$5E,$E7,$E6,$FE,$32,$61,$F9,$98,$CB,$F9,$B8,$CB,$F9,$98,$FC,$26
	.dc.b	$5F,$85,$F2,$24,$FF,$A3,$B9,$3E,$BE,$5F,$61,$FF,$68,$12,$BF,$CA
	.dc.b	$CD,$9E,$11,$D1,$E5,$A7,$E3,$7A,$A4,$7E,$45,$FC,$86,$B8,$8F,$80
	.dc.b	$39,$7F,$6E,$EC,$FC,$C4,$7E,$3F,$AB,$F2,$1A,$49,$FC,$84,$65,$FC
	.dc.b	$94,$65,$EB,$C9,$FB,$8D,$3F,$27,$E5,$FF,$15,$D3,$FC,$38,$BE,$CF
	.dc.b	$FD,$78,$DB,$FF,$58,$BB,$41,$C3,$63,$89,$CB,$BE,$47,$28,$CB,$D2
	.dc.b	$32,$FC,$29,$F0,$36,$73,$7E,$8A,$E3,$39,$39,$10,$C1,$23,$83,$97
	.dc.b	$E0,$CF,$27,$31,$E1,$9C,$5F,$C4,$AF,$0B,$F1,$A3,$33,$F0,$AF,$CF
	.dc.b	$7C,$89,$5C,$67,$E0,$1E,$17,$93,$83,$CC,$E1,$F1,$19,$8F,$3B,$C4
	.dc.b	$F0,$57,$89,$FA,$06,$4F,$A8,$E6,$BC,$B5,$D6,$BC,$B5,$E9,$BD,$5F
	.dc.b	$BA,$30,$78,$6B,$81,$1C,$9E,$83,$C6,$78,$6F,$BB,$AB,$EE,$F9,$0E
	.dc.b	$86,$67,$A3,$93,$A1,$EC,$3E,$19,$23,$E1,$BA,$9F,$C8,$92,$FD,$71
	.dc.b	$E3,$6B,$C6,$BE,$A3,$C6,$FF,$22,$FF,$93,$1F,$F6,$95,$7F,$7B,$23
	.dc.b	$82,$47,$F9,$41,$AF,$53,$E1,$9F,$9B,$F6,$BE,$63,$ED,$C3,$E1,$D7
	.dc.b	$F8,$41,$B5,$A8,$C5,$FC,$47,$DB,$1F,$B4,$BD,$3B,$6D,$FD,$16,$BE
	.dc.b	$33,$F6,$CE,$2F,$C2,$DB,$EC,$D9,$F4,$CF,$F0,$68,$C4,$7A,$67,$F8
	.dc.b	$3C,$62,$FE,$8B,$6F,$CD,$5F,$83,$7C,$CD,$3C,$C5,$3C,$D3,$F0,$1F
	.dc.b	$0D,$7C,$19,$E7,$1A,$79,$B4,$E6,$53,$9D,$78,$DF,$E3,$2D,$7E,$A1
	.dc.b	$E1,$BF,$86,$DA,$F8,$6F,$F8,$AB,$E7,$BF,$C6,$84,$7F,$C5,$BC,$B5
	.dc.b	$FE,$72,$A3,$15,$FD,$62,$3F,$C6,$1F,$1F,$FE,$31,$F1,$E5,$FF,$8B
	.dc.b	$E3,$2F,$F8,$C6,$32,$FF,$8B,$E3,$FD,$A1,$5F,$E0,$AF,$91,$7C,$2E
	.dc.b	$32,$FF,$6A,$43,$E1,$B1,$87,$F3,$68,$C3,$FC,$E2,$F9,$5E,$A1,$3D
	.dc.b	$1F,$F6,$D9,$C6,$2B,$4B,$1B,$76,$D9,$EB,$1D,$EF,$CB,$5D,$E9,$F0
	.dc.b	$DF,$E3,$1E,$B9,$7F,$E3,$57,$FC,$69,$AE,$23,$FB,$2F,$37,$F8,$05
	.dc.b	$D1,$FE,$3B,$8F,$F1,$97,$5F,$F8,$CF,$59,$3F,$E3,$38,$CB,$FE,$35
	.dc.b	$8C,$BF,$D9,$23,$C9,$F2,$5E,$5F,$3E,$4F,$A4,$F9,$51,$A7,$E0,$BF
	.dc.b	$11,$F9,$2D,$CC,$D3,$84,$7F,$7E,$8C,$BD,$4F,$A3,$E1,$FD,$3F,$A6
	.dc.b	$7D,$05,$F7,$25,$6E,$72,$D9,$CB,$FE,$0A,$84,$F3,$9F,$BD,$F8,$1F
	.dc.b	$FB,$FA,$BC,$0D,$7F,$C2,$2D,$E3,$FF,$2B,$AF,$96,$E3,$EF,$CA,$E2
	.dc.b	$F4,$8F,$B9,$C1,$F7,$78,$A5,$7B,$75,$FD,$23,$67,$E3,$11,$FC,$5B
	.dc.b	$5F,$8C,$F4,$79,$74,$7E,$5F,$E9,$DF,$29,$7D,$A9,$5F,$2E,$72,$D9
	.dc.b	$CB,$FE,$2E,$8F,$F7,$63,$63,$F4,$5F,$4F,$F3,$23,$FC,$66,$DD,$BF
	.dc.b	$CF,$2D,$EA,$FF,$10,$CA,$3E,$7B,$F6,$DD,$3F,$E2,$66,$BB,$E0,$EF
	.dc.b	$C3,$FB,$1F,$98,$FF,$27,$DF,$EE,$BD,$1F,$D3,$19,$65,$FE,$CA,$7F
	.dc.b	$D1,$6F,$88,$38,$7C,$AD,$7F,$C6,$EF,$66,$C7,$47,$FC,$A3,$0F,$2B
	.dc.b	$F8,$27,$CE,$B5,$FC,$F3,$6B,$F8,$1B,$5F,$E0,$ED,$7C,$8B,$E1,$F8
	.dc.b	$92,$38,$BD,$3D,$29,$1D,$1A,$FF,$12,$F9,$D7,$C8,$7E,$BD,$B1,$FC
	.dc.b	$51,$AE,$E3,$5F,$2F,$23,$4D,$7A,$7C,$B7,$32,$3C,$DF,$DE,$7C,$EF
	.dc.b	$FB,$12,$3F,$91,$B5,$FA,$0D,$7F,$06,$D7,$FC,$CA,$BD,$7E,$64,$8F
	.dc.b	$E5,$71,$96,$FA,$3E,$02,$38,$3F,$98,$FC,$E3,$E8,$7F,$88,$36,$33
	.dc.b	$18,$33,$F1,$72,$38,$B5,$FE,$7B,$2F,$A9,$E6,$48,$EE,$FF,$1D,$F5
	.dc.b	$9F,$2F,$49,$3F,$D1,$23,$2D,$DB,$81,$1C,$1B,$DB,$A1,$1C,$1F,$D3
	.dc.b	$BE,$DB,$F8,$3B,$FD,$49,$AF,$E7,$DA,$FE,$C9,$AF,$FA,$52,$BF,$D1
	.dc.b	$2F,$E9,$5A,$FE,$E1,$AE,$FD,$4F,$E7,$5C,$FF,$C0,$63,$F5,$8D,$7F
	.dc.b	$57,$D7,$A3,$E2,$7A,$72,$3E,$13,$5F,$ED,$36,$F4,$FE,$51,$23,$E4
	.dc.b	$3F,$D2,$7F,$74,$FE,$95,$06,$91,$D7,$CA,$47,$FB,$9C,$23,$11,$FD
	.dc.b	$E2,$1F,$2F,$FD,$F5,$D4,$7F,$4F,$6B,$FE,$D0,$76,$FF,$53,$FA,$0F
	.dc.b	$D5,$F8,$6F,$03,$F8,$FC,$61,$BC,$3F,$E0,$98,$70,$D2,$5F,$37,$FC
	.dc.b	$1D,$F4,$1F,$EC,$6F,$17,$BC,$FF,$B8,$1B,$8F,$78,$FF,$66,$C6,$31
	.dc.b	$FE,$5E,$FF,$86,$A3,$87,$AF,$FE,$19,$8C,$57,$19,$FE,$8B,$3E,$0D
	.dc.b	$BF,$C3,$DF,$3A,$FA,$4A,$4A,$FE,$E5,$0F,$92,$D9,$23,$2F,$2E,$89
	.dc.b	$1C,$1A,$FF,$BE,$BE,$87,$F9,$A6,$5E,$3B,$F9,$13,$FE,$E8,$6A,$69
	.dc.b	$F2,$9F,$1E,$DE,$82,$3E,$6F,$05,$5F,$EF,$2F,$A1,$FE,$BF,$3C,$E7
	.dc.b	$2F,$F8,$EA,$D1,$88,$FE,$91,$18,$AF,$F1,$EB,$63,$87,$C8,$BE,$87
	.dc.b	$2D,$8C,$C6,$0F,$CC,$FD,$0C,$8F,$F1,$23,$5F,$F2,$5F,$47,$DD,$C3
	.dc.b	$FA,$CF,$45,$E4,$FE,$B5,$36,$BF,$B6,$CD,$47,$E7,$27,$FC,$51,$F0
	.dc.b	$7F,$E1,$E7,$FC,$55,$19,$3E,$2B,$FD,$DA,$35,$18,$3F,$BC,$F2,$75
	.dc.b	$7F,$CD,$73,$CA,$31,$F0,$12,$3A,$FC,$05,$AE,$47,$91,$E4,$79,$ED
	.dc.b	$F3,$78,$AF,$3B,$EB,$CF,$F9,$5C,$7F,$90,$5F,$EB,$AD,$E4,$D7,$5C
	.dc.b	$8F,$7F,$D0,$DC,$BC,$38,$A3,$E3,$7D,$0F,$F0,$3A,$FF,$26,$3F,$0F
	.dc.b	$1F,$D2,$39,$BF,$4B,$99,$FF,$0D,$B6,$3E,$7D,$BB,$FA,$EF,$ED,$7E
	.dc.b	$A7,$D0,$BD,$7F,$F0,$73,$71,$FF,$43,$B7,$0F,$F4,$5B,$5F,$F0,$C3
	.dc.b	$FE,$0D,$8D,$FA,$5C,$88,$FF,$A4,$3E,$87,$F0,$2D,$AF,$F0,$E4,$61
	.dc.b	$FF,$4B,$B5,$FF,$12,$3E,$F6,$F2,$70,$24,$F5,$F6,$F9,$91,$C1,$FE
	.dc.b	$A0,$FA,$0F,$F0,$D3,$7D,$4F,$F1,$64,$62,$FF,$D4,$ED,$7F,$C6,$9F
	.dc.b	$41,$F8,$2F,$7F,$FD,$52,$D8,$FF,$56,$BF,$E7,$6C,$73,$CD,$FE,$73
	.dc.b	$9F,$F2,$6F,$D9,$7F,$7D,$7F,$D6,$6D,$7F,$C8,$0D,$77,$E9,$7F,$A4
	.dc.b	$FE,$7D,$F4,$14,$93,$FE,$BA,$8C,$BF,$EB,$CF,$8C,$FE,$40,$DF,$1F
	.dc.b	$FD,$39,$18,$8F,$F6,$1B,$5F,$F2,$82,$F8,$CD,$92,$3A,$3F,$EA,$5F
	.dc.b	$A0,$FF,$29,$41,$A4,$E1,$D7,$BD,$27,$FD,$9F,$3F,$E5,$36,$C7,$FA
	.dc.b	$BF,$EC,$BF,$C7,$F1,$FE,$61,$6B,$FE,$74,$6B,$FC,$21,$AB,$E8,$5F
	.dc.b	$43,$FE,$B9,$6C,$7F,$B7,$9A,$FF,$B8,$9A,$FF,$9B,$D7,$C6,$F1,$24
	.dc.b	$71,$7F,$D8,$9F,$62,$F7,$69,$27,$FD,$D7,$19,$7F,$DD,$BF,$44,$F9
	.dc.b	$9F,$F6,$63,$63,$31,$83,$3F,$35,$23,$8B,$5F,$F7,$CF,$C5,$BE,$5F
	.dc.b	$FD,$AC,$D8,$FF,$7D,$B5,$DC,$7F,$B6,$F8,$5F,$43,$49,$3F,$E8,$F3
	.dc.b	$F9,$9C,$69,$C1,$F8,$3F,$A1,$E4,$47,$30,$8F,$F7,$47,$D8,$3D,$3A
	.dc.b	$4A,$E7,$19,$7A,$FC,$E7,$9F,$CC,$FE,$FD,$F3,$BF,$DB,$BA,$3F,$D6
	.dc.b	$6D,$AF,$E8,$F5,$B9,$FF,$5B,$BE,$F8,$FF,$54,$35,$CE,$9F,$EF,$B8
	.dc.b	$C4,$72,$8F,$F5,$FD,$FF,$83,$67,$A4,$E4,$FF,$7D,$7D,$0F,$F2,$77
	.dc.b	$FD,$F9,$E5,$69,$FE,$4D,$7C,$B0,$FF,$64,$CE,$47,$8F,$9A,$FE,$BF
	.dc.b	$19,$AC,$47,$23,$AC,$E3,$E8,$1E,$75,$F9,$A9,$5F,$DA,$34,$FF,$12
	.dc.b	$56,$A3,$19,$7F,$B6,$A3,$11,$FD,$71,$57,$28,$C5,$F8,$BE,$7F,$FD
	.dc.b	$95,$07,$F6,$CF,$F6,$46,$E1,$FE,$33,$76,$7F,$02,$26,$1F,$88,$9E
	.dc.b	$4D,$8D,$BD,$CF,$07,$9F,$F0,$BE,$75,$F1,$B4,$97,$EC,$3F,$C0,$F1
	.dc.b	$F4,$4D,$B9,$0E,$1F,$EF,$6D,$C3,$FC,$94,$FD,$9F,$9E,$7C,$FF,$D2
	.dc.b	$7E,$3C,$FF,$93,$DB,$8F,$F8,$5A,$D1,$C1,$F0,$35,$FF,$04,$BC,$1F
	.dc.b	$F0,$EB,$5E,$35,$F8,$0F,$15,$F1,$37,$E3,$25,$7A,$E7,$2D,$9F,$E7
	.dc.b	$24,$7F,$9F,$35,$FF,$0C,$C7,$06,$C7,$7F,$86,$FC,$7F,$27,$FA,$CB
	.dc.b	$3F,$F3,$63,$73,$FC,$CF,$6E,$A3,$6C,$C3,$F4,$2E,$5F,$A1,$C6,$F9
	.dc.b	$FF,$CC,$76,$C9,$CC,$7E,$6C,$C3,$F9,$38,$C3,$F6,$7E,$69,$EF,$FE
	.dc.b	$91,$FF,$40,$E5,$F4,$5B,$3F,$E4,$87,$6F,$19,$FC,$84,$62,$3F,$8E
	.dc.b	$35,$FB,$41,$C3,$D6,$3C,$04,$7C,$9E,$FF,$E5,$7A,$64,$7E,$3F,$93
	.dc.b	$F7,$AD,$D3,$2F,$E4,$26,$ED,$F9,$09,$FC,$77,$4F,$E3,$F5,$97,$F8
	.dc.b	$A8,$CB,$F8,$C8,$CB,$F8,$C8,$F8,$9F,$3F,$F1,$7C,$89,$3F,$26,$32
	.dc.b	$FC,$B8,$CB,$FE,$5C,$D5,$3A,$8C,$4E,$B1,$FF,$57,$56,$A7,$A1,$CC
	.dc.b	$87,$3A,$E5,$3C,$2B,$96,$DF,$5E,$72,$FD,$98,$CD,$7D,$99,$CB,$F6
	.dc.b	$E3,$2F,$D5,$8C,$BF,$5E,$33,$7E,$58,$F1,$1E,$4B,$F1,$3E,$83,$27
	.dc.b	$B8,$D5,$F9,$5D,$47,$07,$47,$07,$E1,$DF,$82,$FC,$77,$E1,$9E,$04
	.dc.b	$BE,$97,$C5,$F8,$0F,$80,$F8,$07,$81,$CB,$E0,$8E,$EE,$4F,$88,$71
	.dc.b	$3E,$21,$93,$E2,$19,$9F,$24,$F8,$A7,$C9,$3C,$8F,$21,$E2,$3C,$11
	.dc.b	$E2,$8F,$0C,$68,$EA,$68,$E0,$E8,$EA,$E9,$C4,$78,$4E,$84,$8E,$2E
	.dc.b	$5F,$24,$72,$9F,$4C,$68,$CC,$F1,$7E,$09,$F7,$0F,$DD,$C7,$E0,$DB
	.dc.b	$7F,$C0,$88,$FF,$86,$A3,$4F,$D6,$36,$66,$3E,$E9,$FC,$47,$E7,$EC
	.dc.b	$7C,$F9,$46,$23,$E3,$46,$1D,$4F,$E1,$A7,$57,$F1,$27,$F8,$C4,$62
	.dc.b	$FE,$14,$7F,$13,$8C,$3F,$04,$C1,$D0,$F1,$3A,$7E,$1C,$FE,$E2,$31
	.dc.b	$7F,$2A,$78,$9E,$6B,$F2,$3C,$CF,$4C,$F0,$8F,$81,$19,$AF,$43,$C9
	.dc.b	$B5,$D4,$95,$C6,$7F,$9D,$35,$FE,$20,$D9,$F3,$47,$5B,$F8,$35,$F2
	.dc.b	$63,$CC,$7B,$8D,$1F,$0D,$D4,$F8,$AF,$D7,$3C,$0F,$59,$C0,$F8,$B5
	.dc.b	$FD,$26,$FF,$C5,$FF,$3E,$F9,$0B,$F3,$72,$FF,$C1,$53,$96,$CE,$56
	.dc.b	$2C,$3F,$C0,$94,$BE,$B7,$FE,$3C,$F9,$F7,$B7,$49,$8F,$F8,$5E,$1F
	.dc.b	$3F,$88,$EB,$8F,$F8,$3E,$78,$C6,$EB,$A5,$7F,$65,$E1,$FE,$CD,$C5
	.dc.b	$2B,$FB,$54,$65,$FD,$03,$FD,$90,$8E,$0D,$90,$FE,$8D,$D1,$FE,$54
	.dc.b	$F8,$07,$C8,$5F,$9F,$95,$FE,$2F,$9C,$B6,$72,$FA,$80,$84,$F5,$9F
	.dc.b	$F3,$0F,$CF,$FC,$FA,$E0,$1C,$70,$FC,$F8,$E7,$18,$DB,$FC,$68,$FD
	.dc.b	$18,$FF,$21,$39,$77,$BF,$63,$BF,$07,$F9,$C3,$E5,$BF,$CB,$FC,$AF
	.dc.b	$9F,$FF,$08,$19,$C6,$61,$FE,$A2,$AF,$F6,$F9,$23,$FC,$BC,$FF,$9A
	.dc.b	$08,$F5,$F6,$7F,$CF,$BF,$5D,$FE,$AB,$AF,$6B,$FE,$A7,$7F,$CB,$0D
	.dc.b	$73,$E9,$70,$38,$FA,$52,$37,$1F,$E8,$8F,$A2,$FC,$2C,$7C,$86,$EF
	.dc.b	$F1,$E1,$18,$C7,$FC,$B4,$FE,$42,$3F,$CD,$0E,$5F,$AE,$D7,$8F,$3F
	.dc.b	$7E,$7F,$07,$91,$E1,$F3,$65,$72,$8C,$BF,$EC,$8B,$FF,$32,$3F,$E3
	.dc.b	$06,$BE,$28,$FF,$43,$47,$F9,$A7,$1F,$43,$71,$D5,$79,$BC,$6F,$F5
	.dc.b	$3F,$D6,$FF,$02,$9F,$F3,$A2,$F9,$FC,$9A,$EF,$97,$FE,$85,$9D,$36
	.dc.b	$72,$FE,$43,$EB,$7F,$14,$DE,$6F,$F5,$AC,$63,$3F,$98,$BE,$17,$33
	.dc.b	$2B,$FD,$25,$3B,$CB,$FD,$10,$EE,$3F,$D2,$DE,$A7,$89,$F0,$7C,$27
	.dc.b	$63,$FD,$87,$19,$C7,$FD,$A9,$97,$FB,$64,$95,$FE,$E8,$6B,$FA,$89
	.dc.b	$E1,$87,$FB,$92,$7D,$3E,$87,$FB,$3B,$E8,$DF,$15,$FE,$B5,$6E,$1F
	.dc.b	$EF,$E6,$BF,$E6,$60,$E1,$FF,$34,$AF,$A3,$E4,$48,$E4,$FF,$B0,$7D
	.dc.b	$1F,$CE,$47,$F9,$C1,$AF,$FA,$AD,$AF,$F0,$66,$AF,$A3,$7C,$D7,$FB
	.dc.b	$2D,$B1,$E5,$35,$19,$35,$F3,$32,$35,$28,$C4,$7B,$61,$EB,$75,$24
	.dc.b	$70,$7F,$DB,$5F,$47,$F5,$E3,$FD,$12,$D7,$FD,$88,$D7,$F8,$FB,$53
	.dc.b	$FE,$D3,$78,$FB,$1D,$08,$FA,$E3,$F8,$93,$7C,$3F,$09,$1F,$81,$F4
	.dc.b	$6F,$85,$FF,$77,$36,$33,$18,$33,$F4,$72,$38,$B5,$FB,$D2,$FA,$EE
	.dc.b	$64,$8E,$EF,$FB,$EF,$DD,$7C,$C5,$24,$FD,$38,$CB,$77,$E0,$47,$06
	.dc.b	$FC,$2E,$84,$70,$73,$DB,$C0,$47,$07,$F3,$3F,$2F,$F9,$89,$1C,$1F
	.dc.b	$85,$FA,$37,$CD,$FE,$1E,$11,$87,$51,$98,$C4,$7E,$87,$17,$87,$F0
	.dc.b	$49,$1E,$F7,$83,$0F,$01,$4F,$D5,$4F,$FB,$39,$FE,$C6,$D7,$C7,$EE
	.dc.b	$F8,$4F,$17,$CC,$7F,$68,$8F,$EC,$72,$F9,$4C,$78,$1F,$CF,$63,$13
	.dc.b	$FD,$BA,$31,$5A,$EF,$E5,$48,$EB,$5F,$15,$AF,$EF,$FE,$8F,$EC,$BF
	.dc.b	$DC,$9B,$7F,$D2,$DB,$7E,$57,$F6,$73,$BE,$0F,$03,$FE,$00,$8C,$D7
	.dc.b	$19,$CC,$BE,$8D,$D9,$8F,$53,$CD,$3E,$BC,$7E,$14,$62,$3F,$C0,$AD
	.dc.b	$C3,$FA,$F3,$F6,$1B,$C1,$CC,$93,$CC,$FF,$09,$E1,$F7,$1C,$0F,$A0
	.dc.b	$7D,$1F,$99,$B1,$A6,$BF,$E1,$65,$F2,$BB,$24,$6D,$E5,$F3,$F2,$39
	.dc.b	$35,$FF,$0F,$FD,$1F,$DF,$7A,$74,$F8,$48,$F8,$77,$F2,$11,$C1,$E5
	.dc.b	$F0,$BF,$CA,$AB,$E8,$FF,$C6,$AF,$B6,$32,$F7,$CB,$F9,$26,$DF,$CA
	.dc.b	$DB,$1F,$D0,$5A,$EF,$E4,$1F,$47,$B6,$C6,$DA,$EE,$3F,$60,$BE,$8E
	.dc.b	$91,$FD,$A3,$FB,$5E,$8F,$F1,$F6,$1F,$E1,$99,$FC,$C7,$27,$D6,$7F
	.dc.b	$44,$DE,$7D,$BF,$25,$B1,$FE,$50,$EE,$FF,$3A,$4E,$CF,$BE,$E2,$3F
	.dc.b	$C9,$E7,$F0,$68,$ED,$D5,$CA,$3F,$CB,$3C,$D5,$45,$31,$F1,$56,$BC
	.dc.b	$CE,$1E,$FE,$8D,$AF,$F2,$EE,$1F,$E0,$99,$FC,$8E,$4E,$5F,$03,$FC
	.dc.b	$BA,$33,$18,$E4,$91,$EA,$FA,$3B,$87,$47,$D1,$F4,$7F,$1A,$3F,$CD
	.dc.b	$6F,$CA,$6E,$FA,$DE,$5F,$C4,$F0,$2D,$79,$9F,$85,$F4,$7F,$E7,$99
	.dc.b	$FE,$E0,$D9,$FF,$03,$35,$FC,$D3,$5F,$E3,$5F,$6B,$EF,$7F,$C1,$8D
	.dc.b	$7F,$40,$D7,$7B,$FE,$05,$FF,$44,$FD,$27,$AD,$FD,$33,$5F,$E4,$CD
	.dc.b	$77,$97,$15,$0F,$88,$C3,$71,$FF,$05,$36,$7F,$C4,$0D,$7F,$69,$F6
	.dc.b	$AF,$7F,$A3,$6B,$FC,$24,$D9,$FF,$14,$35,$FD,$E7,$D2,$7F,$91,$FE
	.dc.b	$03,$FC,$30,$D8,$FF,$17,$3F,$EC,$9C,$B3,$9F,$F9,$F2,$7F,$A1,$7D
	.dc.b	$23,$DA,$FE,$B4,$D8,$FF,$11,$35,$FF,$1C,$35,$FE,$19,$EE,$3E,$5B
	.dc.b	$FB,$13,$63,$FC,$54,$D7,$FC,$80,$D7,$F8,$A7,$CB,$7F,$1E,$7F,$C6
	.dc.b	$0D,$7F,$C8,$CD,$77,$E3,$7F,$6F,$87,$CA,$D2,$4F,$27,$FB,$A7,$BC
	.dc.b	$F8,$EA,$49,$FF,$1F,$46,$5F,$F2,$07,$D1,$3E,$27,$FB,$E3,$63,$31
	.dc.b	$83,$FC,$01,$F4,$5F,$D7,$23,$F9,$73,$5F,$ED,$CD,$7F,$DA,$ED,$5F
	.dc.b	$68,$F9,$5F,$F0,$53,$63,$FC,$A4,$D7,$FC,$D0,$D7,$F9,$E7,$D1,$57
	.dc.b	$FB,$E3,$5F,$F7,$2B,$53,$7E,$8B,$FC,$79,$1F,$E7,$36,$BF,$D1,$DA
	.dc.b	$FF,$81,$DA,$FF,$BC,$3C,$17,$8B,$FE,$21,$6C,$7F,$9A,$5A,$FF,$9F
	.dc.b	$DA,$FF,$53,$FA,$27,$D0,$FF,$8A,$5B,$1F,$E7,$16,$BF,$E8,$76,$BF
	.dc.b	$D6,$F7,$7C,$F7,$F8,$C5,$B1,$FE,$79,$6B,$FE,$8F,$6B,$FD,$8F,$ED
	.dc.b	$1F,$67,$FE,$39,$6C,$7F,$A0,$5A,$FF,$A5,$DA,$FF,$6B,$E3,$FF,$45
	.dc.b	$93,$E4,$BF,$C8,$4F,$FA,$3B,$E8,$DE,$97,$F9,$15,$B5,$FE,$8D,$6B
	.dc.b	$FE,$A3,$6B,$FD,$DF,$ED,$1F,$4D,$FE,$4D,$6C,$7F,$A5,$5A,$FF,$AA
	.dc.b	$DA,$FE,$B3,$DD,$FF,$11,$F5,$7F,$A4,$A3,$FC,$26,$D5,$E8,$7F,$BE
	.dc.b	$E3,$FC,$4F,$1A,$9F,$D7,$11,$FD,$AB,$5E,$3E,$37,$F9,$2E,$78,$C6
	.dc.b	$00,$73,$5D,$FE,$03,$FC,$6F,$F3,$9F,$B1,$AE,$01,$C4,$AF,$FC,$B1
	.dc.b	$18,$6B,$CB,$86,$57,$F8,$C2,$34,$F8,$9E,$8F,$F8,$D3,$E8,$DF,$45
	.dc.b	$FE,$48,$7F,$85,$C6,$78,$31,$32,$30,$43,$CF,$3C,$9F,$F1,$FF,$CD
	.dc.b	$FF,$7A,$6F,$B1,$FE,$54,$8F,$F2,$56,$5F,$E7,$17,$53,$F0,$1F,$F3
	.dc.b	$C1,$3D,$8F,$F4,$63,$63,$AC,$73,$3A,$7B,$1F,$E6,$0F,$A3,$FE,$A5
	.dc.b	$3F,$E5,$06,$BF,$31,$BB,$7C,$B8,$FA,$B1,$F1,$93,$AC,$3F,$D0,$CD
	.dc.b	$8D,$FA,$5D,$F7,$FF,$37,$FC,$9B,$E8,$EE,$F2,$FF,$CC,$33,$96,$CE
	.dc.b	$5F,$BF,$1F,$25,$B1,$F1,$FE,$8F,$E4,$78,$B2,$7E,$6C,$65,$FF,$7B
	.dc.b	$6E,$8F,$8E,$48,$CB,$B7,$E7,$3F,$4F,$E6,$7F,$D0,$FF,$46,$F8,$BB
	.dc.b	$E7,$4B,$FF,$3C,$CE,$5B,$39,$7F,$2F,$09,$E7,$3F,$E9,$2F,$A3,$7A
	.dc.b	$D4,$97,$FE,$84,$87,$CF,$62,$3F,$43,$1C,$1D,$4F,$39,$FC,$57,$07
	.dc.b	$E2,$F8,$65,$7E,$42,$32,$FC,$47,$F1,$24,$70,$6C,$87,$E4,$7A,$BF
	.dc.b	$D5,$3F,$47,$FE,$7C,$E6,$FC,$34,$A1,$F4,$7A,$37,$F1,$B2,$7F,$D3
	.dc.b	$F3,$A6,$CE,$5F,$DE,$FD,$1F,$CB,$6E,$DF,$EB,$88,$C6,$1E,$07,$E6
	.dc.b	$4F,$FA,$A9,$CF,$4F,$FB,$65,$B1,$FC,$39,$A9,$E2,$FF,$0C,$FA,$3F
	.dc.b	$5D,$7B,$39,$5F,$2F,$FB,$53,$FD,$93,$3A,$C3,$FD,$CC,$DB,$FE,$37
	.dc.b	$5C,$1F,$DF,$D7,$0A,$91,$82,$1E,$17,$2F,$EF,$E3,$C0,$FF,$B1,$AB
	.dc.b	$B2,$6D,$77,$8E,$06,$EB,$9D,$FA,$EF,$9C,$F7,$34,$78,$A3,$4E,$0D
	.dc.b	$1D,$08,$FE,$02,$FC,$67,$8A,$7A,$C6,$6B,$C2,$7F,$1B,$8C,$BF,$20
	.dc.b	$CD,$7C,$88,$CB,$FD,$12,$BA,$CE,$21,$7E,$FB,$ED,$1E,$77,$D2,$F9
	.dc.b	$A7,$B3,$97,$C4,$FB,$DC,$BE,$E8,$E2,$48,$CB,$DA,$3A,$1E,$73,$A1
	.dc.b	$E7,$8F,$34,$F9,$5E,$24,$8E,$07,$23,$ED,$3F,$09,$F6,$DF,$BE,$FE
	.dc.b	$0C,$79,$CE,$37,$E3,$25,$7B,$23,$A4,$F0,$7B,$5F,$8D,$F3,$1E,$83
	.dc.b	$B9,$D4,$F1,$99,$78,$3A,$3E,$34,$FC,$87,$FB,$C9,$D0,$A5,$2C,$FF
	.dc.b	$A8,$DE,$A6,$8A,$6A,$30,$F1,$3E,$41,$C2,$BD,$EF,$E1,$7D,$4F,$91
	.dc.b	$5E,$F9,$FC,$A4,$62,$B2,$DA,$F8,$15,$FE,$77,$7D,$67,$A6,$BA,$BF
	.dc.b	$AD,$9E,$0E,$8C,$1C,$AB,$AB,$4E,$B1,$C2,$79,$D7,$76,$D7,$0A,$E9
	.dc.b	$3E,$98,$FE,$17,$5C,$5F,$83,$1C,$5A,$71,$AE,$75,$E0,$3C,$47,$4A
	.dc.b	$E4,$FB,$67,$AB,$A9,$F2,$CE,$A3,$E7,$4E,$9F,$2B,$DC,$F0,$BD,$CC
	.dc.b	$96,$7C,$6F,$84,$ED,$1E,$A9,$ED,$5F,$C2,$1F,$C4,$D7,$F4,$AF,$A3
	.dc.b	$FF,$6F,$5F,$F9,$12,$31,$E5,$6B,$0F,$C7,$D2,$31,$5F,$E4,$A8,$E2
	.dc.b	$FF,$91,$9B,$5C,$A3,$A5,$72,$AE,$30,$AD,$CF,$E7,$3E,$8B,$FB,$74
	.dc.b	$FF,$94,$1B,$7F,$D8,$1B,$3B,$E4,$E7,$18,$AF,$F2,$EB,$53,$4F,$EF
	.dc.b	$AF,$90,$DF,$D1,$F2,$F4,$E4,$64,$E3,$1D,$8F,$D3,$B8,$3F,$53,$28
	.dc.b	$CB,$63,$31,$C5,$E9,$19,$7D,$DE,$3F,$C4,$FA,$37,$6D,$FA,$29,$5E
	.dc.b	$39,$CB,$67,$2F,$F9,$DA,$13,$CE,$7E,$7F,$D1,$FF,$70,$D7,$FC,$12
	.dc.b	$DD,$7F,$85,$B7,$0F,$F4,$2A,$F8,$1E,$3D,$F2,$7E,$24,$ED,$3B,$7E
	.dc.b	$47,$BB,$E2,$77,$F0,$3E,$23,$F0,$31,$98,$FF,$46,$B5,$FE,$36,$D7
	.dc.b	$F8,$FB,$5F,$F4,$CB,$53,$CF,$1F,$F1,$2B,$71,$E3,$9F,$C2,$F1,$BF
	.dc.b	$23,$F6,$9F,$EB,$EB,$FF,$15,$AF,$A7,$C9,$AF,$56,$65,$F8,$39,$D3
	.dc.b	$67,$2F,$FA,$9B,$E9,$5F,$0B,$F9,$E6,$C7,$E2,$DA,$EF,$E3,$75,$32
	.dc.b	$34,$F1,$8E,$5E,$DF,$3F,$1E,$57,$38,$CB,$FC,$EA,$FF,$DD,$6F,$F4
	.dc.b	$66,$BF,$EC,$06,$A7,$FD,$E7,$9F,$F9,$31,$BB,$71,$AF,$DB,$FD,$A7
	.dc.b	$F8,$88,$7D,$06,$3E,$AA,$63,$FA,$79,$CB,$67,$2F,$FB,$2A,$13,$D2
	.dc.b	$7F,$81,$FD,$9B,$E2,$29,$2F,$F6,$50,$F9,$4C,$47,$FB,$52,$38,$3A
	.dc.b	$9E,$73,$D3,$E2,$BA,$FD,$04,$AE,$71,$97,$FB,$0D,$A3,$F7,$AD,$8C
	.dc.b	$C6,$0E,$0F,$85,$FE,$37,$F5,$1F,$EC,$A8,$FF,$76,$AF,$A8,$FE,$0E
	.dc.b	$D8,$FE,$DE,$D7,$FB,$A3,$5F,$EF,$7F,$01,$FC,$BD,$FE,$EA,$D7,$FB
	.dc.b	$D3,$5D,$F8,$FF,$C5,$21,$F0,$B4,$93,$C9,$FE,$33,$F5,$0F,$9C,$A4
	.dc.b	$9F,$F0,$1C,$65,$BA,$70,$23,$83,$FC,$13,$F4,$EE,$8F,$E4,$AD,$8C
	.dc.b	$C6,$0C,$FD,$44,$8E,$2D,$7F,$C2,$D3,$FE,$6D,$7F,$C3,$AD,$7F,$C4
	.dc.b	$2D,$41,$C2,$FA,$7F,$F4,$7B,$FE,$26,$6B,$FE,$2A,$6B,$C9,$AF,$1E
	.dc.b	$3F,$E7,$F1,$F9,$C8,$FF,$15,$35,$E9,$1A,$FA,$79,$1F,$E2,$C9,$46
	.dc.b	$23,$FC,$65,$F3,$6F,$B4,$FE,$96,$D8,$FF,$19,$B5,$DC,$6B,$E8,$E4
	.dc.b	$69,$AF,$4F,$3F,$99,$1E,$6F,$6F,$83,$91,$D9,$AF,$F9,$17,$D4,$FF
	.dc.b	$61,$3F,$E4,$AF,$63,$FA,$FC,$62,$3F,$D6,$31,$FD,$8B,$EA,$1F,$11
	.dc.b	$49,$3F,$E4,$C8,$CB,$FE,$4E,$FA,$E7,$BF,$FD,$A9,$B1,$98,$C1,$D5
	.dc.b	$87,$32,$9F,$E5,$B9,$7C,$4F,$8C,$91,$E1,$7C,$DF,$13,$23,$C8,$D7
	.dc.b	$97,$2F,$AC,$8F,$A5,$AB,$C7,$FF,$78,$DF,$E7,$63,$1E,$37,$F9,$D2
	.dc.b	$31,$3A,$F8,$1F,$F0,$FF,$D4,$BE,$8A,$93,$0F,$F0,$74,$38,$F6,$48
	.dc.b	$CB,$FE,$28,$FA,$97,$81,$FE,$2A,$C3,$BE,$7F,$E8,$46,$DF,$E9,$1A
	.dc.b	$9F,$E2,$CD,$E1,$E4,$47,$FC,$41,$6F,$6B,$FC,$6F,$F4,$CF,$A9,$A4
	.dc.b	$AF,$F1,$34,$7F,$32,$7E,$19,$FB,$47,$1B,$32,$FF,$4C,$C6,$1B,$E5
	.dc.b	$76,$F2,$A5,$F4,$6B,$D0,$7D,$4F,$FA,$83,$0E,$91,$DF,$87,$CE,$47
	.dc.b	$CB,$C1,$EA,$23,$83,$D3,$E2,$FD,$D5,$F5,$3F,$1D,$F5,$46,$5F,$F5
	.dc.b	$7E,$9F,$CA,$70,$FF,$54,$B6,$3F,$D5,$CD,$77,$EA,$BE,$A7,$6D,$8D
	.dc.b	$B5,$DC,$7F,$93,$97,$D4,$D2,$3F,$E5,$27,$FC,$A7,$EA,$FC,$2C,$3E
	.dc.b	$1C,$FF,$24,$E9,$C7,$0F,$FB,$1A,$AC,$FF,$B2,$A6,$BB,$9F,$F3,$07
	.dc.b	$82,$F7,$73,$3F,$CF,$67,$31,$8E,$7C,$F6,$D9,$FF,$34,$57,$ED,$1B
	.dc.b	$1C,$BB,$77,$1A,$70,$F8,$3D,$EB,$5F,$7F,$1F,$F2,$94,$FF,$9C,$F4
	.dc.b	$FF,$01,$37,$93,$D3,$18,$AF,$C3,$3F,$AF,$9D,$1E,$CF,$9C,$DD,$7C
	.dc.b	$0B,$47,$C2,$FA,$9F,$F1,$CD,$7F,$A0,$5F,$81,$87,$E4,$5C,$B7,$6E
	.dc.b	$07,$3F,$1E,$D7,$CA,$7A,$7A,$FF,$52,$BF,$DF,$36,$8C,$47,$F5,$A8
	.dc.b	$C5,$7F,$BD,$1B,$1F,$EF,$95,$EC,$F0,$84,$62,$78,$4E,$63,$13,$FF
	.dc.b	$00,$7E,$CF,$FD,$2D,$5F,$F0,$01,$B1,$9F,$85,$E9,$18,$8F,$ED,$A7
	.dc.b	$F6,$EF,$AF,$FF,$24,$56,$23,$31,$87,$01,$C3,$B6,$BD,$3E,$9B,$FD
	.dc.b	$1D,$F2,$3C,$21,$18,$79,$BF,$93,$E7,$D3,$98,$F1,$7D,$BB,$EC,$BF
	.dc.b	$D6,$AD,$8E,$8D,$7C,$AD,$78,$FC,$1F,$02,$3C,$1E,$7F,$15,$23,$9B
	.dc.b	$5F,$7F,$DA,$7F,$80,$DF,$8E,$D7,$FC,$2A,$D4,$D3,$FC,$C6,$FC,$16
	.dc.b	$BD,$3E,$07,$FD,$9F,$18,$8F,$A0,$FF,$8F,$9E,$1E,$E6,$88,$FC,$BF
	.dc.b	$A8,$FE,$F0,$FF,$89,$9A,$9A,$4E,$FF,$F6,$EB,$FE,$22,$23,$8E,$C9
	.dc.b	$3D,$5A,$FE,$03,$E9,$9F,$33,$FE,$E6,$6C,$7D,$D6,$BB,$8D,$7C,$6C
	.dc.b	$8D,$35,$E9,$F1,$5C,$C8,$F3,$7F,$DE,$7F,$49,$FE,$EB,$8F,$CA,$35
	.dc.b	$FC,$CB,$5F,$F2,$3B,$52,$EF,$E6,$48,$FC,$A4,$65,$BF,$39,$E0,$23
	.dc.b	$83,$F3,$5F,$3E,$FB,$5F,$CE,$42,$30,$EA,$33,$18,$8F,$D4,$AF,$90
	.dc.b	$C9,$23,$AB,$C3,$EC,$24,$78,$1A,$FE,$CB,$EC,$BF,$D9,$FE,$3F,$F4
	.dc.b	$47,$FC,$06,$4E,$BF,$DE,$C6,$5F,$3C,$7F,$B5,$A7,$F7,$CD,$7F,$81
	.dc.b	$B5,$FF,$38,$BD,$BB,$B3,$3D,$1B,$3F,$E7,$6A,$F1,$79,$59,$8E,$35
	.dc.b	$55,$F3,$72,$71,$BF,$0D,$6E,$75,$3C,$2B,$06,$66,$47,$13,$99,$FE
	.dc.b	$0B,$9E,$27,$F8,$90,$93,$CA,$38,$1E,$49,$E6,$F1,$8D,$3C,$E3,$2F
	.dc.b	$C8,$7E,$59,$1F,$E3,$D3,$DE,$39,$DF,$F7,$8F,$AF,$FA,$B1,$8C,$3D
	.dc.b	$F7,$E2,$9C,$BE,$4B,$E4,$E7,$C7,$F9,$53,$A8,$F2,$19,$21,$82,$61
	.dc.b	$E1,$9E,$27,$86,$3E,$EE,$18,$78,$9F,$D0,$6B,$FC,$C4,$D9,$FC,$FC
	.dc.b	$62,$3F,$A4,$BF,$E5,$22,$47,$DF,$F8,$06,$7D,$6B,$FC,$7D,$5A,$7E
	.dc.b	$8C,$E9,$F4,$E1,$E7,$6E,$5F,$D5,$A7,$51,$FE,$47,$6C,$EE,$3A,$E1
	.dc.b	$E1,$C2,$47,$23,$FD,$97,$87,$F9,$A2,$F8,$E1,$D6,$BC,$D8,$73,$3F
	.dc.b	$C5,$F3,$AB,$FF,$69,$BE,$27,$FB,$33,$5E,$47,$F5,$F7,$0F,$09,$F1
	.dc.b	$BE,$97,$E3,$7C,$2E,$2B,$BC,$5B,$F8,$13,$6B,$91,$F2,$27,$8B,$F2
	.dc.b	$A3,$4A,$FD,$0D,$8D,$BE,$13,$91,$C4,$86,$49,$7F,$0A,$38,$BE,$68
	.dc.b	$E2,$7D,$38,$CD,$FC,$28,$E2,$FA,$9E,$B3,$C5,$C9,$F3,$27,$8B,$EE
	.dc.b	$32,$7A,$4E,$33,$D6,$79,$1E,$58,$C9,$C5,$CC,$78,$63,$CC,$7A,$23
	.dc.b	$2F,$84,$E2,$7C,$9B,$F9,$76,$BF,$2C,$E4,$E5,$7D,$6F,$C5,$1E,$8A
	.dc.b	$FA,$27,$17,$AB,$8B,$E6,$E6,$38,$46,$2F,$C2,$D8,$EE,$7B,$6F,$C0
	.dc.b	$47,$04,$3E,$41,$F0,$0E,$A7,$8A,$7A,$B4,$EA,$7B,$4E,$17,$EC,$3E
	.dc.b	$3B,$97,$13,$E1,$8C,$3F,$54,$C5,$FB,$67,$8B,$F1,$2B,$E0,$DF,$CE
	.dc.b	$6D,$F8,$A7,$85,$FD,$98,$FD,$7C,$FF,$70,$9F,$F1,$81,$1C,$12,$3F
	.dc.b	$64,$D7,$D7,$3E,$4E,$CF,$99,$18,$3F,$7A,$E3,$0F,$6C,$60,$F8,$C6
	.dc.b	$CF,$D7,$46,$E7,$A4,$6A,$78,$BA,$31,$1A,$79,$5F,$B9,$FC,$91,$FC
	.dc.b	$76,$BA,$38,$AF,$E2,$71,$88,$FB,$B1,$87,$A9,$BA,$FE,$51,$3B,$AE
	.dc.b	$D3,$E3,$3C,$37,$F9,$29,$D3,$FC,$E5,$B1,$FC,$D9,$AF,$94,$EE,$78
	.dc.b	$AB,$BB,$8A,$EE,$7F,$98,$2F,$E1,$B8,$BE,$D3,$B6,$BE,$43,$F4,$F7
	.dc.b	$E0,$AF,$74,$64,$F9,$31,$9B,$F0,$C6,$67,$BC,$E6,$3E,$D1,$E9,$35
	.dc.b	$3C,$27,$A9,$E6,$AF,$1B,$88,$FC,$DC,$61,$EF,$3F,$D0,$AB,$F2,$84
	.dc.b	$7F,$A1,$3B,$7C,$35,$FD,$33,$AF,$11,$F4,$E3,$B3,$FD,$3A,$33,$87
	.dc.b	$C9,$BF,$31,$FD,$8E,$3F,$B7,$1A,$BB,$5F,$DB,$23,$0D,$21,$F0,$6F
	.dc.b	$BB,$FD,$1E,$38,$BF,$EB,$EB,$FF,$49,$36,$3F,$60,$47,$9C,$FA,$DF
	.dc.b	$EF,$7E,$57,$28,$C4,$7F,$77,$9F,$F0,$24,$E7,$0F,$54,$F2,$7F,$BF
	.dc.b	$46,$63,$E1,$19,$3F,$B8,$B9,$7D,$06,$4F,$EC,$0E,$6F,$D3,$7E,$23
	.dc.b	$FB,$DC,$66,$F4,$F9,$2B,$95,$70,$BF,$1C,$7B,$0C,$D6,$DF,$F0,$95
	.dc.b	$73,$74,$60,$D9,$8B,$F0,$B8,$34,$47,$CD,$5F,$12,$30,$F8,$9F,$A9
	.dc.b	$3F,$C8,$48,$FF,$83,$DF,$24,$F6,$8F,$F1,$37,$83,$FE,$71,$F8,$B7
	.dc.b	$CD,$7F,$A6,$1B,$3F,$EF,$56,$BF,$EF,$E6,$A2,$0F,$79,$8F,$FA,$89
	.dc.b	$B1,$FE,$FF,$6B,$C0,$E4,$E1,$C2,$F9,$9E,$44,$8E,$4F,$FA,$BB,$E3
	.dc.b	$1E,$AD,$24,$F5,$3C,$D1,$A7,$01,$7C,$9F,$32,$46,$A3,$2D,$FA,$9F
	.dc.b	$01,$1C,$1D,$3E,$31,$E9,$7F,$B0,$9B,$19,$8C,$19,$F4,$A4,$71,$6B
	.dc.b	$E9,$97,$CF,$F3,$24,$77,$7F,$D9,$FF,$1A,$E6,$A4,$9F,$7C,$65,$BC
	.dc.b	$BC,$08,$E0,$DF,$2F,$42,$38,$3E,$4F,$AA,$F7,$3F,$DB,$ED,$8C,$C6
	.dc.b	$0C,$F0,$48,$E2,$D7,$EB,$4F,$CE,$6F,$07,$42,$3E,$0E,$2E,$E4,$70
	.dc.b	$79,$3E,$0F,$C6,$47,$07,$E0,$3E,$1F,$FD,$EA,$9A,$47,$FC,$15,$E8
	.dc.b	$C7,$FC,$F0,$D7,$F9,$17,$A3,$FE,$0F,$23,$FE,$7D,$7F,$C2,$58,$C8
	.dc.b	$FF,$40,$35,$FC,$A5,$3E,$7B,$F2,$71,$E3,$6C,$66,$3D,$11,$8F,$A7
	.dc.b	$FF,$74,$4E,$65,$DB,$FA,$C8,$C3,$67,$FD,$DD,$F1,$2F,$72,$92,$FF
	.dc.b	$DB,$F0,$F8,$3D,$92,$32,$FF,$BD,$FE,$23,$F7,$D1,$FE,$E6,$F8,$97
	.dc.b	$8F,$FE,$FB,$6C,$66,$31,$E4,$48,$E0,$D7,$D3,$E0,$FF,$0A,$B7,$17
	.dc.b	$F0,$CD,$7F,$83,$FC,$23,$E2,$71,$7F,$C4,$23,$18,$7F,$BE,$1A,$F4
	.dc.b	$3F,$8B,$2F,$89,$E3,$18,$6A,$8D,$3C,$E3,$30,$F8,$9A,$49,$F0,$3C
	.dc.b	$BE,$27,$F8,$E6,$3E,$8E,$1E,$24,$AF,$37,$07,$42,$38,$3D,$5D,$7F
	.dc.b	$71,$7C,$4B,$8B,$E0,$4E,$67,$F9,$56,$7F,$B3,$C7,$F8,$EB,$63,$F9
	.dc.b	$13,$5D,$F6,$3E,$27,$6D,$8D,$B5,$DC,$7E,$A3,$E2,$64,$7E,$61,$AF
	.dc.b	$F2,$FF,$4B,$F9,$FE,$1F,$CE,$F9,$AF,$17,$F3,$F9,$B5,$FD,$22,$6B
	.dc.b	$FD,$2D,$BE,$C3,$DE,$CC,$FE,$3A,$FF,$A8,$61,$FD,$2A,$B5,$18,$3F
	.dc.b	$A6,$74,$70,$F8,$9E,$2D,$8D,$5F,$F5,$89,$EB,$F1,$39,$8F,$C4,$1F
	.dc.b	$D7,$5C,$CF,$08,$5B,$E0,$33,$1C,$23,$13,$BC,$7F,$10,$DE,$5F,$7C
	.dc.b	$62,$BB,$BF,$B5,$9D,$1D,$E9,$E2,$F0,$BF,$8A,$FC,$8F,$89,$F8,$98
	.dc.b	$7A,$1B,$87,$F6,$E5,$BE,$F7,$99,$75,$F0,$AE,$1E,$97,$A7,$99,$FD
	.dc.b	$CE,$3F,$BF,$60,$8C,$42,$31,$5F,$DB,$1B,$1F,$DC,$57,$C5,$70,$6C
	.dc.b	$EE,$30,$74,$F8,$87,$CB,$D2,$4F,$F7,$88,$CB,$FD,$EB,$E2,$3E,$97
	.dc.b	$C4,$DE,$23,$9E,$6F,$83,$9F,$F8,$0E,$B6,$FF,$81,$7E,$B5,$EF,$FF
	.dc.b	$32,$6C,$7F,$81,$9A,$EE,$3F,$9B,$FC,$53,$CD,$A4,$9F,$F0,$6C,$65
	.dc.b	$BE,$6F,$02,$38,$3F,$C2,$7F,$79,$FC,$A1,$FF,$0C,$35,$FF,$13,$35
	.dc.b	$FF,$17,$B5,$FF,$19,$7C,$5F,$F7,$D7,$FC,$40,$D7,$FC,$5C,$D7,$FC
	.dc.b	$6E,$D7,$8F,$A5,$FD,$4A,$3F,$CF,$B1,$C2,$30,$7F,$56,$1F,$2F,$49
	.dc.b	$1F,$E2,$C8,$CB,$FE,$2E,$F8,$CF,$DD,$37,$6F,$EB,$F1,$88,$FF,$1B
	.dc.b	$35,$FF,$22,$3F,$AD,$79,$70,$70,$23,$FE,$3A,$FA,$EF,$EB,$0D,$F2
	.dc.b	$7F,$B5,$C6,$23,$FC,$84,$D7,$FC,$A1,$F1,$AF,$8C,$FE,$E2,$D8,$FF
	.dc.b	$22,$B5,$FF,$25,$35,$E3,$F1,$9C,$08,$F0,$7F,$BD,$FD,$6B,$E1,$A9
	.dc.b	$27,$FC,$A3,$19,$7F,$CA,$7F,$1B,$FC,$31,$BE,$2F,$F8,$12,$31,$1F
	.dc.b	$E5,$B6,$BF,$E6,$95,$F1,$BB,$24,$74,$7F,$C1,$BE,$FF,$EE,$63,$FC
	.dc.b	$C2,$D7,$FC,$CC,$D7,$FC,$E8,$D7,$FC,$FB,$F1,$9F,$A2,$7F,$CF,$0D
	.dc.b	$73,$D3,$FE,$1A,$8C,$47,$FA,$15,$7C,$56,$49,$1F,$E7,$28,$CB,$FE
	.dc.b	$74,$FA,$CF,$F2,$13,$FE,$85,$6B,$FE,$8E,$6B,$FE,$91,$6A,$F8,$CF
	.dc.b	$D5,$79,$3F,$E8,$EF,$8D,$79,$FF,$DF,$9B,$5F,$CE,$EB,$7E,$1F,$FA
	.dc.b	$69,$FD,$64,$73,$F2,$BF,$C6,$B1,$82,$3F,$E0,$87,$8C,$7F,$A8,$DA
	.dc.b	$BE,$37,$DA,$FF,$4C,$8C,$92,$7F,$C2,$33,$AC,$7F,$D4,$B3,$9A,$F3
	.dc.b	$56,$B4,$91,$82,$1F,$D6,$2F,$96,$1F,$EB,$46,$C7,$96,$BD,$07,$27
	.dc.b	$85,$61,$E7,$87,$FA,$DA,$B8,$C7,$AE,$7F,$CA,$F1,$EC,$8F,$F5,$D3
	.dc.b	$5F,$04,$68,$ED,$3C,$DF,$7F,$1E,$30,$DD,$7F,$6D,$BE,$E7,$E9,$5F
	.dc.b	$29,$9A,$E0,$FF,$B4,$63,$57,$E7,$BC,$CF,$9C,$E8,$F8,$1E,$44,$3E
	.dc.b	$43,$97,$B9,$23,$A3,$C8,$FF,$64,$CF,$95,$F8,$33,$F1,$08,$EC,$EA
	.dc.b	$7B,$7E,$17,$B5,$74,$8C,$9F,$66,$33,$87,$D9,$AB,$3E,$F9,$FC,$39
	.dc.b	$87,$F1,$D1,$87,$B1,$E1,$AE,$C7,$DB,$9D,$47,$82,$33,$5F,$87,$9C
	.dc.b	$CF,$DC,$7E,$F9,$1F,$82,$F2,$9E,$91,$F6,$7B,$71,$5C,$E3,$C6,$7E
	.dc.b	$56,$32,$FA,$CC,$CF,$E3,$E3,$F4,$66,$A3,$F5,$55,$87,$B1,$C2,$BB
	.dc.b	$3D,$4C,$9F,$2C,$D1,$F8,$79,$E4,$E4,$95,$E1,$78,$BE,$C7,$33,$FA
	.dc.b	$B9,$E3,$39,$24,$66,$78,$BE,$48,$CC,$78,$4C,$9F,$04,$EA,$64,$F5
	.dc.b	$18,$3E,$09,$C0,$FE,$19,$3F,$17,$0F,$F0,$04,$FA,$67,$8B,$E0,$78
	.dc.b	$98,$AF,$33,$D6,$78,$C6,$27,$91,$C6,$33,$3E,$58,$CC,$FB,$49,$19
	.dc.b	$79,$18,$7C,$31,$C1,$D1,$F0,$5E,$D5,$F3,$27,$85,$78,$2B,$DE,$7C
	.dc.b	$17,$04,$9E,$11,$CE,$78,$C6,$23,$C3,$18,$79,$5F,$C9,$9D,$57,$48
	.dc.b	$B1,$F3,$5B,$3F,$8A,$39,$9E,$A3,$CA,$7C,$B7,$6D,$3D,$11,$FC,$A0
	.dc.b	$8F,$F0,$A6,$A3,$F9,$5B,$86,$9B,$39,$9F,$C9,$FC,$EE,$F8,$7C,$D8
	.dc.b	$FE,$A1,$18,$C3,$E9,$C7,$F2,$78,$C3,$E9,$C3,$FD,$1A,$F4,$3F,$0C
	.dc.b	$6C,$C5,$75,$8D,$C6,$8F,$83,$19,$9E,$B1,$8A,$FE,$A5,$18,$A5,$FF
	.dc.b	$60,$8C,$36,$BB,$D7,$A1,$F5,$98,$3D,$13,$C1,$F8,$46,$4F,$01,$E9
	.dc.b	$3C,$D5,$E0,$7F,$15,$1C,$5F,$ED,$0D,$BF,$EC,$0D,$AF,$4C,$F5,$AE
	.dc.b	$4F,$03,$17,$E9,$7B,$BF,$8C,$AE,$51,$E3,$BF,$95,$39,$8E,$27,$A8
	.dc.b	$E2,$79,$8C,$9D,$CC,$9E,$C9,$EA,$FC,$63,$33,$F4,$DE,$2F,$B1,$EC
	.dc.b	$F8,$9B,$7E,$E7,$8B,$63,$84,$60,$F5,$46,$4F,$AD,$7F,$12,$76,$48
	.dc.b	$CD,$FC,$58,$C5,$7E,$46,$31,$3E,$18,$FF,$05,$DF,$F8,$38,$8F,$F7
	.dc.b	$07,$6F,$88,$FF,$01,$7C,$6E,$1F,$85,$7D,$9F,$F1,$94,$67,$0F,$CC
	.dc.b	$CE,$54,$66,$B4,$DA,$F8,$A7,$62,$9D,$AF,$E1,$39,$3F,$C6,$51,$C5
	.dc.b	$A7,$F8,$75,$C5,$FF,$8B,$5F,$F2,$09,$89,$FF,$1C,$38,$9E,$F7,$E8
	.dc.b	$9C,$C7,$33,$37,$B7,$B9,$F8,$7B,$F7,$5E,$5F,$81,$36,$B9,$D7,$F1
	.dc.b	$12,$9F,$E3,$CA,$FF,$1F,$91,$F0,$1D,$A9,$3C,$7E,$7D,$E2,$73,$9C
	.dc.b	$D7,$F1,$DF,$85,$AF,$73,$17,$5F,$14,$69,$D3,$5C,$FC,$66,$DE,$8D
	.dc.b	$50,$8C,$36,$38,$2F,$8F,$C9,$23,$2F,$3F,$8D,$FF,$46,$93,$E2,$70
	.dc.b	$67,$E3,$65,$78,$E5,$18,$8F,$64,$79,$1D,$5F,$52,$3E,$BF,$8C,$FF
	.dc.b	$58,$3F,$D0,$DA,$9A,$4E,$CF,$80,$9A,$48,$F8,$DF,$19,$F2,$5F,$E9
	.dc.b	$CD,$4D,$27,$1F,$CB,$7F,$8F,$C7,$D7,$6B,$F8,$76,$BD,$38,$B8,$11
	.dc.b	$F0,$BF,$57,$D7,$7A,$54,$93,$F6,$E3,$2D,$F7,$B8,$11,$C1,$BD,$BA
	.dc.b	$11,$C1,$CF,$C1,$F0,$11,$C1,$DF,$E1,$FC,$84,$70,$7E,$37,$E2,$FF
	.dc.b	$18,$9A,$47,$F2,$0B,$DE,$A4,$8E,$4F,$FA,$A3,$9E,$47,$FA,$29,$AF
	.dc.b	$E8,$7A,$99,$FF,$AB,$89,$1F,$E9,$06,$BF,$EB,$F7,$FD,$92,$7F,$B2
	.dc.b	$BE,$5F,$F6,$1D,$1F,$BB,$C5,$F1,$98,$BF,$F0,$74,$3E,$37,$11,$ED
	.dc.b	$6C,$70,$8F,$F0,$A7,$5F,$EE,$23,$11,$FB,$76,$BF,$BB,$6B,$E1,$AF
	.dc.b	$DA,$7C,$6B,$E0,$7B,$B6,$7A,$E3,$FC,$46,$31,$0C,$3F,$83,$C3,$D1
	.dc.b	$F3,$FC,$6F,$EA,$EB,$57,$C6,$7A,$9A,$EB,$F6,$FC,$47,$FB,$D0,$9E
	.dc.b	$36,$3C,$99,$87,$88,$FF,$19,$4A,$33,$1D,$8F,$F1,$B6,$2E,$5F,$93
	.dc.b	$F0,$EF,$8D,$A4,$AE,$82,$1F,$1B,$C0,$91,$A8,$CB,$7E,$37,$A9,$1C
	.dc.b	$1D,$3E,$37,$F9,$57,$0F,$F2,$CA,$F9,$2D,$E2,$F2,$9E,$6E,$09,$7D
	.dc.b	$DA,$FB,$4F,$37,$C6,$7D,$15,$F1,$AE,$0F,$E8,$D8,$AB,$55,$FC,$EE
	.dc.b	$31,$8F,$F3,$56,$C7,$F3,$D6,$BB,$EF,$7C,$6E,$DB,$1B,$6B,$B8,$FD
	.dc.b	$9F,$C6,$C8,$FE,$80,$D7,$FA,$57,$C6,$7F,$50,$C3,$FA,$57,$2D,$E7
	.dc.b	$FE,$A1,$36,$BF,$B0,$4D,$7F,$CE,$4D,$F6,$1F,$01,$99,$4E,$63,$1D
	.dc.b	$19,$F0,$AE,$3F,$80,$BF,$E0,$8D,$8E,$5F,$1B,$B8,$D3,$87,$F0,$F8
	.dc.b	$BE,$03,$17,$F5,$63,$13,$F8,$CD,$1C,$FE,$97,$F5,$D1,$98,$C7,$34
	.dc.b	$8E,$BE,$45,$C3,$A3,$F0,$7E,$33,$DB,$3F,$98,$6C,$FE,$15,$70,$6F
	.dc.b	$83,$E3,$78,$92,$F8,$45,$BE,$1F,$19,$FD,$47,$03,$49,$5F,$DE,$FE
	.dc.b	$C5,$F5,$9F,$D5,$DB,$19,$8C,$1F,$D6,$FE,$33,$E3,$79,$BF,$E0,$36
	.dc.b	$E3,$FE,$0B,$7E,$CE,$99,$E2,$FE,$9B,$3F,$E1,$1F,$8C,$7C,$07,$F8
	.dc.b	$25,$B1,$FE,$0D,$6B,$FE,$15,$6B,$FE,$19,$E8,$7C,$47,$F8,$45,$B1
	.dc.b	$FE,$15,$6B,$FE,$1D,$6B,$FE,$21,$F8,$BF,$F1,$3B,$FE,$1B,$6B,$FE
	.dc.b	$23,$6B,$BF,$3F,$FC,$3D,$F1,$7F,$92,$8F,$F1,$13,$5F,$F1,$53,$5F
	.dc.b	$F1,$73,$5F,$F1,$F7,$A2,$FB,$6F,$F1,$43,$63,$FC,$58,$D7,$FC,$68
	.dc.b	$D7,$FC,$71,$F1,$9F,$E0,$04,$D2,$7A,$DF,$E3,$33,$DF,$1F,$E4,$D6
	.dc.b	$BF,$ED,$F6,$BF,$EC,$4C,$FF,$C7,$5F,$78,$F9,$FA,$49,$FF,$20,$46
	.dc.b	$5F,$F2,$17,$C7,$7E,$9D,$BC,$3F,$E4,$38,$C4,$7F,$93,$5A,$FF,$94
	.dc.b	$7E,$3B,$FA,$63,$FE,$4B,$6B,$FE,$53,$6B,$BF,$33,$FC,$9D,$F6,$2F
	.dc.b	$97,$A4,$9F,$F2,$A4,$65,$FF,$2B,$7C,$7B,$E3,$FF,$CA,$CD,$8C,$C6
	.dc.b	$0F,$F2,$D7,$C7,$BE,$26,$92,$7F,$CC,$11,$97,$FC,$C5,$E2,$38,$FF
	.dc.b	$CC,$4D,$8C,$C6,$0F,$F3,$37,$C7,$E6,$3F,$CE,$0D,$7F,$CE,$8D,$7F
	.dc.b	$D0,$8D,$47,$F9,$3F,$E5,$3F,$CE,$1F,$07,$FE,$56,$23,$FE,$73,$FB
	.dc.b	$16,$14,$95,$FE,$79,$8C,$BF,$E7,$BF,$91,$7C,$77,$F9,$ED,$B1,$98
	.dc.b	$C1,$FE,$7F,$F9,$17,$C3,$52,$4F,$FA,$16,$32,$FF,$A1,$FE,$C5,$F5
	.dc.b	$FF,$E8,$76,$C6,$63,$07,$FA,$2F,$E2,$3E,$74,$7F,$65,$7C,$6D,$C3
	.dc.b	$FC,$53,$1A,$F4,$BF,$D3,$AF,$09,$FF,$4F,$35,$FF,$57,$35,$F9,$6D
	.dc.b	$4F,$3F,$2F,$7C,$7D,$8E,$FF,$23,$6B,$FC,$1B,$5B,$79,$CE,$9C,$BB
	.dc.b	$7D,$4E,$5F,$64,$E6,$FD,$33,$97,$D7,$19,$BF,$1D,$78,$23,$51,$88
	.dc.b	$EA,$64,$87,$F9,$76,$3E,$61,$1F,$F5,$D5,$A3,$9C,$7C,$47,$E6,$F8
	.dc.b	$18,$8F,$8B,$16,$53,$63,$0F,$09,$F0,$DF,$A9,$F8,$F1,$97,$E0,$9A
	.dc.b	$BF,$AB,$19,$9E,$13,$A3,$F0,$51,$C5,$FB,$AF,$E1,$C8,$FB,$1E,$53
	.dc.b	$D2,$3F,$DF,$71,$87,$F0,$2F,$E2,$B8,$31,$1F,$82,$8B,$87,$FC,$01
	.dc.b	$8D,$9E,$A2,$1F,$EF,$FA,$EC,$F9,$8C,$DF,$E5,$2F,$9D,$F4,$6C,$EA
	.dc.b	$BF,$2B,$1E,$38,$C1,$E0,$3B,$1F,$98,$78,$B8,$3F,$32,$E2,$BC,$11
	.dc.b	$F0,$0E,$66,$8C,$1E,$73,$95,$FB,$DF,$84,$70,$70,$4B,$F5,$47,$29
	.dc.b	$FD,$8C,$7A,$63,$9B,$A6,$9C,$EB,$53,$C2,$30,$73,$BF,$97,$1E,$A8
	.dc.b	$E2,$79,$DC,$12,$36,$48,$F8,$C7,$33,$E5,$46,$DC,$9C,$E3,$83,$F1
	.dc.b	$4E,$F1,$E5,$7E,$2C,$FC,$D9,$FA,$F1,$D4,$F8,$E6,$8F,$91,$5E,$E8
	.dc.b	$DC,$77,$3C,$6E,$2B,$C9,$1C,$9F,$01,$8B,$FA,$11,$F1,$8F,$2C,$FA
	.dc.b	$23,$EA,$DF,$96,$7D,$B5,$C6,$FE,$93,$63,$93,$5F,$21,$F0,$5C,$1E
	.dc.b	$2B,$F3,$1E,$A8,$FE,$5B,$5F,$CD,$C8,$FF,$0C,$6B,$FE,$ED,$6B,$C0
	.dc.b	$FE,$52,$FC,$57,$F9,$9F,$C8,$D7,$C5,$3E,$07,$27,$02,$1F,$98,$8E
	.dc.b	$18,$7C,$28,$FE,$6B,$18,$AC,$B6,$38,$DF,$F4,$88,$C4,$73,$33,$7E
	.dc.b	$17,$C6,$78,$8F,$BC,$6C,$B3,$F5,$E7,$9B,$63,$99,$F5,$27,$A5,$AB
	.dc.b	$71,$86,$CE,$CF,$05,$7E,$0E,$FD,$71,$D0,$F2,$92,$FA,$39,$7E,$23
	.dc.b	$FD,$AE,$BA,$BB,$6C,$7E,$F9,$B3,$FD,$B9,$B3,$F3,$0E,$91,$E2,$7E
	.dc.b	$0D,$FA,$5C,$9E,$19,$E2,$F8,$4C,$9F,$26,$FD,$8E,$CC,$BF,$25,$EA
	.dc.b	$7C,$33,$FB,$BC,$79,$5D,$CF,$F7,$E8,$C4,$70,$36,$7A,$4C,$5F,$C6
	.dc.b	$7C,$86,$23,$8D,$79,$23,$2F,$E1,$DE,$E7,$D4,$AF,$03,$83,$B9,$D0
	.dc.b	$FF,$09,$CF,$F8,$74,$8F,$F7,$27,$BC,$72,$84,$A3,$FC,$83,$0F,$91
	.dc.b	$FF,$4B,$CF,$8E,$3C,$B3,$C6,$38,$47,$F9,$22,$77,$39,$24,$7A,$4D
	.dc.b	$78,$DE,$27,$07,$AA,$B9,$BE,$18,$D3,$99,$E6,$E4,$F9,$D3,$E5,$24
	.dc.b	$7A,$AB,$C7,$3D,$B0,$FA,$55,$A8,$F1,$9F,$E6,$07,$C3,$3C,$23,$C0
	.dc.b	$79,$9C,$19,$9E,$75,$E4,$8C,$DF,$A2,$73,$5D,$8F,$9F,$5E,$59,$D6
	.dc.b	$1F,$43,$0F,$0C,$59,$D5,$7B,$9E,$D1,$98,$E9,$3E,$B7,$CE,$7C,$53
	.dc.b	$B9,$E1,$3E,$E9,$92,$E1,$FE,$D2,$7B,$9A,$39,$47,$91,$F8,$D5,$A8
	.dc.b	$FA,$FF,$32,$F6,$7C,$94,$F8,$77,$C9,$73,$8F,$F4,$6C,$61,$C3,$5F
	.dc.b	$03,$5F,$3A,$F9,$2E,$51,$D5,$B1,$98,$E7,$18,$DA,$47,$66,$BE,$DF
	.dc.b	$91,$7C,$97,$46,$C7,$9D,$AE,$E3,$57,$23,$4D,$7E,$47,$C9,$3E,$03
	.dc.b	$CE,$D8,$F8,$4D,$77,$1A,$EE,$91,$A6,$BF,$4F,$E4,$BF,$4A,$FD,$A6
	.dc.b	$BF,$B8,$6B,$FE,$BA,$6A,$7F,$DC,$AE,$FC,$AF,$97,$18,$8E,$11,$8C
	.dc.b	$E4,$7D,$99,$FF,$5D,$BE,$38,$FA,$7E,$EB,$D8,$A4,$9F,$C0,$46,$5B
	.dc.b	$F1,$7C,$08,$E0,$DF,$81,$D0,$8E,$0E,$7E,$2F,$80,$8E,$0E,$FE,$17
	.dc.b	$90,$8E,$0F,$C9,$7C,$87,$E2,$93,$48,$FE,$39,$FC,$E3,$C3,$97,$04
	.dc.b	$79,$FC,$1E,$C8,$E0,$FF,$50,$E5,$FE,$CF,$23,$82,$1F,$ED,$17,$FD
	.dc.b	$9E,$7F,$B4,$3E,$4A,$47,$FA,$A9,$FF,$6C,$11,$F5,$FC,$8F,$9C,$8E
	.dc.b	$08,$7F,$02,$3D,$EF,$ED,$B0,$37,$C4,$FD,$CF,$C8,$BD,$0B,$F0,$93
	.dc.b	$0F,$E0,$20,$4E,$DC,$8F,$91,$FE,$04,$E2,$34,$F8,$DB,$DF,$FD,$02
	.dc.b	$77,$3E,$7F,$8D,$7C,$37,$F0,$C6,$C7,$66,$BB,$EA,$D1,$1D,$3F,$C5
	.dc.b	$29,$F2,$3F,$C5,$BE,$35,$CF,$49,$7E,$A8,$7C,$8E,$C9,$19,$79,$7C
	.dc.b	$54,$8E,$0D,$7E,$87,$C8,$FF,$30,$D7,$FA,$ED,$B7,$F1,$12,$BC,$1C
	.dc.b	$1E,$52,$38,$3C,$9F,$21,$FD,$E1,$7C,$8B,$83,$FB,$CE,$0A,$D5,$7F
	.dc.b	$40,$8C,$63,$FC,$CD,$B1,$FC,$FD,$AE,$FE,$45,$F2,$3B,$6C,$6D,$AE
	.dc.b	$E3,$F5,$9F,$23,$23,$F9,$BB,$5F,$E9,$9F,$17,$FE,$08,$C3,$F8,$DF
	.dc.b	$2D,$F3,$7F,$AA,$CD,$AF,$F0,$9C,$D7,$F2,$0D,$F9,$17,$C5,$66,$7F
	.dc.b	$0F,$7F,$92,$9C,$BF,$94,$8F,$E6,$8D,$F0,$38,$74,$F1,$64,$6F,$C0
	.dc.b	$B5,$FA,$90,$F7,$3C,$5E,$3D,$9F,$03,$E7,$D1,$CF,$79,$B0,$DA,$FD
	.dc.b	$3C,$FE,$0D,$F4,$78,$D9,$C7,$F7,$1A,$23,$E1,$7C,$8F,$BA,$7F,$78
	.dc.b	$DB,$FF,$14,$AE,$0D,$F0,$7C,$7F,$0E,$65,$FB,$6A,$E0,$FF,$90,$32
	.dc.b	$4F,$CF,$F9,$77,$CC,$7F,$4C,$6D,$7F,$7B,$6E,$7F,$E0,$76,$BF,$E0
	.dc.b	$EF,$90,$70,$7F,$41,$8C,$35,$FF,$27,$C7,$DA,$D3,$FC,$0C,$F0,$F2
	.dc.b	$BF,$C1,$1F,$2B,$FB,$EA,$FF,$0C,$B5,$FF,$2B,$B5,$FE,$28,$D5,$EC
	.dc.b	$3D,$BF,$F0,$4B,$63,$FC,$28,$D7,$FC,$40,$D7,$FC,$4E,$BD,$AE,$24
	.dc.b	$8E,$2F,$F8,$53,$E3,$FE,$4C,$7F,$87,$DA,$FF,$8A,$DA,$FF,$8C,$9A
	.dc.b	$FF,$9C,$17,$99,$CC,$91,$CD,$FF,$10,$FC,$7F,$F8,$26,$3F,$C7,$6D
	.dc.b	$7F,$CF,$2D,$7F,$9B,$35,$7C,$77,$F7,$A7,$FC,$86,$D7,$FD,$02,$D7
	.dc.b	$F9,$F3,$5E,$3E,$EF,$F8,$CA,$3F,$D3,$91,$CA,$30,$7F,$8D,$B0,$7D
	.dc.b	$15,$24,$FF,$90,$E3,$2F,$F9,$17,$E3,$1F,$27,$FE,$3E,$6C,$66,$30
	.dc.b	$67,$E4,$64,$71,$6B,$FE,$59,$F8,$EF,$E4,$4F,$F9,$45,$AE,$7C,$8F
	.dc.b	$F2,$5C,$62,$3F,$CC,$BF,$18,$FA,$2F,$F2,$7B,$63,$FC,$B4,$D7,$FC
	.dc.b	$CC,$D7,$8F,$CD,$F0,$23,$C1,$FF,$2C,$FC,$6B,$E1,$E9,$27,$FC,$CD
	.dc.b	$19,$7F,$CD,$1F,$1A,$FA,$1F,$F3,$0B,$63,$31,$83,$3E,$E4,$8E,$2D
	.dc.b	$7F,$CF,$9F,$00,$3F,$CD,$65,$7F,$CE,$CD,$76,$FF,$9C,$23,$F4,$91
	.dc.b	$FE,$C6,$6B,$FE,$03,$6A,$69,$EB,$7F,$D1,$8D,$7F,$D2,$6D,$7A,$61
	.dc.b	$FE,$7E,$F8,$CF,$F3,$8C,$7F,$A6,$1A,$FF,$B5,$9A,$FF,$84,$DA,$BE
	.dc.b	$6B,$FC,$72,$FF,$88,$5F,$E7,$2F,$F8,$59,$AE,$7C,$DF,$F4,$CC,$70
	.dc.b	$8F,$F5,$3B,$5F,$F7,$2B,$5F,$F1,$13,$5F,$04,$72,$23,$FE,$24,$EB
	.dc.b	$7C,$E6,$23,$C0,$78,$E7,$FD,$53,$3A,$8C,$10,$D9,$FE,$3B,$74,$E3
	.dc.b	$CA,$FF,$74,$DF,$C0,$6D,$74,$BF,$04,$62,$BA,$3F,$EF,$0B,$E6,$E6
	.dc.b	$3A,$78,$BF,$02,$F9,$47,$AE,$F9,$CE,$63,$04,$38,$9F,$1A,$78,$BE
	.dc.b	$99,$E2,$E4,$E1,$5E,$38,$CC,$7C,$C3,$35,$FE,$51,$7F,$9E,$C6,$B0
	.dc.b	$F8,$13,$9C,$3D,$6D,$9D,$61,$E2,$C3,$B6,$1F,$0A,$BD,$4F,$94,$F5
	.dc.b	$1F,$D0,$AB,$FA,$49,$1F,$E9,$2F,$38,$ED,$1F,$7F,$E3,$30,$FF,$99
	.dc.b	$F0,$F1,$B6,$BE,$FC,$EA,$3D,$67,$B8,$E0,$E6,$78,$C6,$8F,$C1,$46
	.dc.b	$67,$4D,$8E,$47,$F6,$27,$0F,$B4,$E2,$79,$CD,$CF,$61,$1E,$68,$CC
	.dc.b	$D7,$89,$CE,$FE,$5B,$87,$F1,$93,$E0,$6C,$77,$38,$5F,$C8,$78,$3F
	.dc.b	$66,$78,$36,$38,$4F,$62,$46,$63,$89,$F3,$4F,$5C,$F9,$E7,$71,$93
	.dc.b	$67,$C5,$3D,$A7,$69,$E9,$1C,$23,$E8,$3E,$A7,$8D,$FC,$7B,$ED,$5C
	.dc.b	$0F,$65,$71,$9F,$97,$5D,$4F,$14,$66,$F8,$C6,$2B,$89,$CD,$FC,$19
	.dc.b	$E6,$3C,$31,$F1,$9C,$9E,$4B,$FC,$1B,$DC,$F4,$CF,$47,$75,$EB,$3C
	.dc.b	$D3,$E2,$8F,$88,$7D,$63,$C8,$70,$9F,$23,$D0,$C9,$C2,$7A,$47,$78
	.dc.b	$D1,$97,$8D,$FC,$98,$F3,$3F,$19,$F0,$1C,$1C,$4F,$48,$EF,$1F,$97
	.dc.b	$BF,$43,$CE,$3E,$29,$F1,$E7,$83,$E1,$38,$DF,$0A,$FF,$22,$92,$3F
	.dc.b	$C8,$AD,$7F,$C8,$AD,$7F,$C8,$AE,$67,$AD,$79,$AF,$F8,$97,$91,$D6
	.dc.b	$FD,$AD,$7F,$12,$DA,$E2,$D9,$F8,$AD,$9F,$A0,$7F,$0B,$70,$FA,$CF
	.dc.b	$F2,$FB,$8A,$F1,$46,$CF,$9B,$7F,$1A,$74,$FF,$50,$6C,$7F,$57,$6B
	.dc.b	$FD,$3D,$B7,$FD,$1D,$B7,$EE,$9E,$A7,$B8,$A7,$B8,$B3,$EE,$6C,$7B
	.dc.b	$9A,$7B,$8F,$59,$83,$DD,$7F,$E6,$C9,$F9,$F7,$EB,$9D,$4F,$C0,$78
	.dc.b	$9F,$82,$3C,$45,$9F,$13,$6B,$C4,$D3,$C5,$3D,$E3,$F2,$F5,$DE,$74
	.dc.b	$79,$2F,$E1,$BF,$3C,$F9,$07,$91,$F9,$CD,$3E,$71,$4D,$9D,$2B,$C0
	.dc.b	$F3,$BF,$7B,$C0,$DB,$86,$DF,$D3,$79,$57,$81,$FF,$4E,$BF,$EA,$22
	.dc.b	$3F,$EA,$27,$53,$F7,$63,$FC,$31,$DF,$8A,$E9,$39,$9F,$F0,$94,$E5
	.dc.b	$F3,$4E,$8F,$EE,$51,$98,$FA,$73,$C4,$FF,$08,$46,$6F,$DF,$3A,$AE
	.dc.b	$2F,$BE,$B8,$47,$58,$FE,$D9,$5C,$6F,$93,$63,$84,$F2,$7E,$29,$E6
	.dc.b	$21,$FE,$C6,$8E,$0E,$8C,$1C,$8D,$DF,$98,$FF,$5B,$B9,$3D,$06,$A3
	.dc.b	$9D,$FA,$23,$CE,$6E,$79,$5F,$F9,$02,$A4,$E6,$2A,$A5,$FF,$88,$7E
	.dc.b	$19,$F4,$3F,$E2,$3A,$91,$87,$00,$39,$72,$D7,$8F,$A7,$FE,$2D,$BE
	.dc.b	$F1,$C6,$78,$46,$23,$9D,$6E,$FC,$97,$67,$83,$E2,$69,$6D,$19,$8F
	.dc.b	$2B,$7E,$25,$F4,$14,$93,$CE,$32,$DF,$3B,$C0,$A3,$2F,$C9,$C3,$FC
	.dc.b	$91,$18,$CB,$D4,$AB,$8B,$6B,$A1,$0F,$96,$E5,$F9,$BF,$12,$F9,$CF
	.dc.b	$53,$63,$E0,$35,$DC,$6B,$EB,$7C,$F3,$9C,$3E,$63,$71,$FF,$2F,$29
	.dc.b	$D9,$C6,$FA,$B5,$FB,$FF,$17,$FE,$F4,$7F,$CC,$CD,$7F,$04,$D4,$D3
	.dc.b	$FC,$9A,$3C,$7F,$96,$D8,$FC,$56,$1F,$81,$8D,$F2,$7D,$97,$FD,$27
	.dc.b	$3F,$93,$6B,$E2,$8F,$BD,$C3,$FE,$B6,$27,$C6,$60,$CE,$FF,$6A,$11
	.dc.b	$4C,$44,$AF,$C6,$4F,$DE,$7F,$44,$D7,$FD,$0C,$D7,$F4,$5F,$14,$F4
	.dc.b	$7F,$20,$D8,$FC,$B3,$5F,$D4,$B5,$FD,$72,$F5,$BF,$1E,$EF,$3F,$D2
	.dc.b	$37,$7F,$F4,$B3,$BA,$F9,$31,$8A,$F0,$92,$3C,$2F,$93,$E7,$A4,$79
	.dc.b	$1A,$F9,$3E,$2F,$C3,$12,$3A,$9E,$E8,$F2,$9E,$E2,$3E,$E7,$F6,$1F
	.dc.b	$45,$F8,$D8,$FD,$B3,$5F,$E1,$2D,$7F,$87,$35,$FF,$59,$2F,$63,$F6
	.dc.b	$2F,$1C,$BF,$83,$37,$1D,$CF,$ED,$BE,$2D,$F0,$D7,$DE,$98,$FA,$84
	.dc.b	$E9,$C7,$C5,$FF,$1A,$75,$19,$7F,$7C,$7F,$1F,$71,$97,$F1,$79,$E1
	.dc.b	$3A,$1F,$0D,$F5,$5C,$36,$7F,$8F,$37,$2F,$F6,$82,$F8,$2F,$E1,$ED
	.dc.b	$AF,$43,$5E,$71,$FC,$52,$DF,$17,$FC,$5B,$E1,$DF,$17,$49,$7F,$06
	.dc.b	$1F,$17,$B2,$46,$5E,$5F,$0B,$23,$83,$5F,$77,$C5,$FF,$45,$D7,$F9
	.dc.b	$C5,$B7,$F1,$12,$BC,$1C,$1E,$52,$38,$3C,$9E,$B7,$F5,$35,$F1,$7F
	.dc.b	$29,$FE,$9D,$69,$D4,$7F,$55,$8C,$61,$FD,$0D,$B1,$FD,$59,$AE,$FE
	.dc.b	$2D,$F1,$7B,$6C,$6D,$AE,$E3,$FA,$52,$F8,$BA,$47,$FA,$73,$FD,$3F
	.dc.b	$B3,$FA,$EE,$1F,$D6,$E7,$F7,$9C,$BF,$91,$7F,$83,$37,$AF,$6F,$F6
	.dc.b	$F6,$C7,$F6,$DF,$1F,$FA,$DC,$EC,$FC,$D3,$88,$FC,$81,$F8,$98,$ED
	.dc.b	$E7,$72,$8F,$EE,$DC,$F5,$45,$31,$F1,$76,$BC,$CE,$1E,$FF,$11,$6B
	.dc.b	$F3,$D8,$23,$B6,$4E,$5B,$C9,$86,$E1,$FB,$19,$FC,$33,$EA,$F8,$BC
	.dc.b	$E3,$D1,$F4,$7C,$5F,$C8,$8F,$F0,$7B,$5F,$F0,$A3,$5F,$F0,$90,$70
	.dc.b	$FC,$9F,$0F,$E9,$36,$BF,$C1,$19,$A7,$33,$FD,$DD,$BF,$15,$ED,$F8
	.dc.b	$AF,$F0,$73,$67,$FC,$42,$FF,$83,$B8,$B3,$B2,$BF,$C2,$9F,$3A,$F9
	.dc.b	$EF,$F0,$93,$70,$CC,$60,$FF,$0B,$7A,$AF,$8B,$A4,$9F,$F0,$F4,$65
	.dc.b	$BF,$0F,$C0,$8E,$0D,$F0,$74,$23,$83,$FC,$51,$F1,$7F,$E4,$E7,$FC
	.dc.b	$76,$D7,$FC,$80,$D7,$E6,$B5,$FF,$20,$7C,$5F,$F1,$37,$FC,$8C,$D7
	.dc.b	$E9,$B5,$DF,$47,$F8,$CF,$E3,$DF,$67,$49,$3F,$E3,$B8,$CB,$FE,$3D
	.dc.b	$F8,$DF,$F0,$61,$3D,$4F,$F1,$E3,$FE,$0F,$CD,$F1,$BF,$E3,$F6,$D7
	.dc.b	$F9,$15,$AF,$F9,$61,$AF,$F9,$77,$E7,$5F,$45,$FE,$47,$6C,$7F,$93
	.dc.b	$5A,$FF,$98,$1A,$FF,$99,$7E,$3D,$F1,$DF,$E4,$F6,$C7,$F9,$55,$AF
	.dc.b	$F9,$A1,$AF,$F9,$B7,$E3,$BF,$8E,$3F,$E5,$B6,$BF,$E6,$E6,$BB,$F2
	.dc.b	$3F,$CB,$9F,$3A,$FA,$2A,$49,$FF,$32,$46,$5F,$F3,$37,$C7,$BE,$3F
	.dc.b	$FC,$C8,$D8,$CC,$60,$FF,$34,$7C,$7B,$E2,$E9,$27,$FC,$E1,$19,$7F
	.dc.b	$CE,$5D,$0F,$8E,$FF,$38,$36,$33,$18,$3F,$CE,$9F,$1E,$F8,$EA,$49
	.dc.b	$FF,$3E,$46,$5F,$F3,$F7,$C7,$8F,$F3,$E1,$5C,$B8,$3F,$D0,$1F,$39
	.dc.b	$FE,$53,$27,$46,$0F,$F4,$25,$FA,$6B,$FD,$36,$D7,$FD,$40,$D7,$F8
	.dc.b	$6B,$5F,$F5,$07,$C8,$7F,$92,$9F,$F4,$93,$5F,$F5,$2B,$5F,$F5,$53
	.dc.b	$5F,$E2,$FF,$38,$F9,$FF,$F4,$B3,$63,$FD,$38,$D7,$FD,$5E,$D7,$FD
	.dc.b	$69,$E0,$BE,$1F,$FD,$2B,$F2,$5F,$CF,$27,$FC,$3C,$D7,$FC,$29,$3B
	.dc.b	$7F,$C4,$4F,$F9,$A9,$FF,$5D,$35,$34,$93,$FE,$A2,$79,$47,$48,$C7
	.dc.b	$C9,$65,$91,$D9,$CB,$48,$71,$3C,$71,$C1,$C1,$3D,$0F,$F6,$25,$27
	.dc.b	$51,$8A,$BE,$7F,$F4,$2A,$D5,$F6,$AD,$36,$75,$7C,$23,$E0,$DD,$7E
	.dc.b	$0C,$7A,$1F,$14,$F5,$9C,$57,$92,$B8,$5F,$C4,$8C,$5F,$B0,$F0,$1D
	.dc.b	$8C,$1F,$52,$78,$D7,$4B,$F2,$CF,$03,$E9,$C6,$5F,$AF,$3C,$1E,$75
	.dc.b	$E0,$7F,$DE,$B7,$F0,$9B,$3F,$D3,$88,$F0,$BE,$EF,$E1,$BC,$EE,$CF
	.dc.b	$C2,$34,$7E,$19,$CD,$75,$BF,$01,$F5,$A3,$37,$A6,$D7,$23,$F1,$AE
	.dc.b	$1F,$84,$7E,$51,$FC,$71,$99,$E9,$19,$39,$57,$86,$FF,$01,$5C,$27
	.dc.b	$37,$5F,$61,$83,$91,$83,$C2,$4B,$EE,$E5,$F6,$D7,$C6,$72,$47,$04
	.dc.b	$8F,$8C,$76,$8E,$2F,$D8,$31,$1E,$87,$C8,$77,$70,$F8,$DE,$8F,$C8
	.dc.b	$38,$9C,$CD,$1E,$03,$91,$98,$F0,$BC,$4C,$9F,$74,$E2,$65,$F6,$3C
	.dc.b	$63,$C3,$7F,$E2,$78,$F0,$9C,$67,$AC,$66,$3A,$BF,$55,$E4,$79,$63
	.dc.b	$AB,$C6,$7A,$DF,$D7,$9E,$77,$67,$E1,$C6,$2B,$C3,$7D,$E3,$11,$E0
	.dc.b	$BF,$C4,$57,$02,$47,$88,$F0,$1E,$37,$2E,$09,$5D,$63,$2F,$AE,$33
	.dc.b	$3E,$57,$A4,$64,$E0,$F5,$33,$7B,$7E,$75,$7C,$C8,$C4,$F0,$9F,$0C
	.dc.b	$62,$BD,$71,$CC,$E2,$74,$3E,$1B,$FE,$4A,$23,$FE,$48,$6B,$FC,$11
	.dc.b	$AF,$F4,$A7,$37,$F5,$2B,$C4,$7F,$91,$5C,$47,$9B,$E4,$BC,$95,$F4
	.dc.b	$9A,$FE,$01,$B3,$CD,$B8,$7C,$A6,$CF,$AE,$FB,$4F,$CE,$BE,$D3,$EC
	.dc.b	$3B,$14,$EC,$5B,$EC,$D9,$EC,$D3,$B5,$F8,$1F,$92,$7F,$5E,$71,$7F
	.dc.b	$01,$B7,$FD,$7D,$B3,$FD,$A1,$AF,$F6,$C6,$DF,$F5,$66,$DF,$F5,$36
	.dc.b	$CF,$F6,$E6,$BE,$13,$CC,$79,$2F,$E1,$BF,$16,$FE,$1B,$F9,$D3,$E1
	.dc.b	$94,$F8,$65,$BF,$86,$DB,$F8,$6D,$3E,$19,$E0,$AF,$80,$DA,$F8,$0D
	.dc.b	$3E,$01,$4F,$80,$5A,$F8,$0D,$AF,$80,$D3,$E0,$14,$F8,$07,$23,$C8
	.dc.b	$62,$39,$B8,$AF,$86,$E6,$7F,$0C,$F9,$0E,$53,$F9,$67,$0F,$23,$0E
	.dc.b	$A3,$CA,$F5,$BF,$7B,$4E,$31,$D9,$E0,$7A,$8D,$18,$3E,$C5,$F2,$AF
	.dc.b	$04,$F1,$6C,$F1,$7C,$66,$1F,$61,$F7,$CF,$09,$DC,$FF,$1F,$D7,$FB
	.dc.b	$18,$8F,$FB,$11,$F0,$4F,$B2,$4F,$FB,$93,$E6,$DE,$AF,$FB,$9A,$73
	.dc.b	$61,$F2,$4F,$92,$D6,$1F,$EE,$C8,$7C,$9E,$2B,$84,$61,$E5,$7D,$1E
	.dc.b	$77,$B6,$9F,$EF,$63,$A3,$5A,$53,$66,$67,$C8,$F6,$BD,$AA,$F0,$4E
	.dc.b	$A3,$13,$E9,$E4,$F1,$7C,$9B,$D6,$BF,$1D,$2F,$A1,$F0,$E7,$4E,$03
	.dc.b	$87,$D0,$D7,$D8,$D7,$E6,$1C,$27,$D8,$D3,$A4,$73,$C7,$E4,$37,$0E
	.dc.b	$3F,$25,$FE,$79,$9F,$98,$BE,$4B,$E6,$2F,$93,$A6,$22,$61,$F2,$E7
	.dc.b	$83,$63,$C7,$18,$8F,$A4,$DA,$EF,$3C,$A7,$E5,$9F,$87,$F9,$37,$CA
	.dc.b	$FD,$D7,$FC,$C1,$8F,$C0,$5A,$E4,$D7,$3C,$BF,$BD,$3F,$E8,$F2,$47
	.dc.b	$C6,$79,$C7,$78,$E0,$F8,$87,$2B,$E2,$7F,$1C,$B8,$2F,$CA,$4A,$FB
	.dc.b	$53,$96,$CF,$D4,$8C,$47,$E4,$DA,$FE,$5D,$AF,$DC,$77,$AF,$E7,$DB
	.dc.b	$9E,$6B,$C7,$F4,$3F,$A1,$F9,$57,$C2,$DF,$80,$99,$7E,$36,$72,$D9
	.dc.b	$FC,$3C,$62,$3F,$4C,$D7,$F5,$4D,$7F,$1E,$EF,$2F,$D9,$37,$1C,$D7
	.dc.b	$8F,$E4,$3C,$9F,$29,$E5,$F9,$59,$8F,$8E,$32,$FE,$CF,$04,$79,$09
	.dc.b	$19,$76,$FD,$58,$DB,$ED,$F9,$9F,$E0,$5F,$2C,$F9,$6B,$CF,$2B,$F5
	.dc.b	$D3,$96,$CF,$EA,$23,$11,$FC,$1D,$AF,$F0,$B6,$BF,$B0,$F9,$67,$C9
	.dc.b	$7E,$C1,$B1,$FC,$35,$AF,$F1,$16,$BF,$C5,$7E,$5B,$F9,$1B,$FC,$4D
	.dc.b	$AF,$F1,$76,$BB,$EF,$FD,$EE,$DF,$3E,$3F,$96,$35,$34,$7C,$17,$F2
	.dc.b	$56,$C7,$F0,$F6,$BF,$9F,$6F,$17,$F2,$DB,$7C,$B7,$F2,$DF,$93,$7C
	.dc.b	$B5,$26,$1F,$C6,$63,$F4,$CD,$EB,$E0,$47,$5F,$1F,$CC,$8E,$0F,$E4
	.dc.b	$5F,$2D,$FC,$E7,$2F,$E7,$D6,$F9,$1F,$11,$2B,$C1,$C1,$E5,$23,$83
	.dc.b	$C9,$EF,$7F,$39,$5F,$2D,$F1,$9F,$EA,$96,$9D,$47,$F4,$98,$C6,$1F
	.dc.b	$D1,$DB,$1F,$D2,$DA,$EF,$E5,$9F,$2D,$B6,$C6,$DA,$EE,$3F,$9B,$AF
	.dc.b	$96,$A4,$7F,$9D,$BF,$CF,$3B,$FF,$B5,$E1,$FD,$86,$7F,$7F,$DF,$FD
	.dc.b	$25,$4E,$A3,$FB,$17,$7D,$9F,$E8,$ED,$7F,$AB,$79,$95,$E6,$7F,$21
	.dc.b	$F3,$3F,$BB,$57,$5E,$8D,$7C,$B6,$DB,$1E,$18,$E5,$E7,$E2,$FF,$95
	.dc.b	$1F,$E0,$49,$F3,$47,$86,$DF,$29,$C2,$7F,$B1,$46,$1F,$EC,$9A,$39
	.dc.b	$BD,$0F,$E0,$E3,$31,$8E,$59,$1C,$BD,$EB,$97,$AD,$F8,$1F,$2C,$F6
	.dc.b	$7E,$54,$FF,$85,$63,$15,$FE,$05,$6C,$7F,$82,$5A,$FF,$83,$7E,$5B
	.dc.b	$D3,$F2,$DF,$E0,$B6,$C7,$F8,$3D,$FF,$0F,$71,$E7,$65,$7F,$84,$7E
	.dc.b	$D3,$F8,$5D,$FF,$89,$5A,$9A,$4F,$17,$FC,$3C,$FF,$16,$DF,$FC,$44
	.dc.b	$DC,$FF,$8A,$6D,$AA,$97,$FE,$0D,$AD,$3F,$E3,$3F,$13,$F5,$8F,$F8
	.dc.b	$45,$AF,$F6,$18,$DB,$E3,$9F,$F1,$74,$EB,$0F,$F1,$93,$70,$FF,$09
	.dc.b	$DF,$03,$8C,$66,$FC,$18,$79,$1E,$D8,$74,$89,$1E,$78,$D6,$1D,$A7
	.dc.b	$33,$C0,$D1,$E9,$73,$1F,$12,$78,$E1,$F1,$2B,$47,$03,$91,$FD,$FE
	.dc.b	$FC,$C9,$B8,$7A,$EB,$AD,$73,$AE,$37,$DD,$EA,$73,$C3,$C2,$FD,$DA
	.dc.b	$E2,$FC,$23,$C5,$3C,$5E,$07,$27,$04,$AF,$03,$F0,$E3,$8C,$7D,$09
	.dc.b	$EB,$1E,$9A,$F0,$BF,$64,$E9,$3C,$8E,$0F,$23,$C8,$78,$8F,$6B,$A7
	.dc.b	$15,$E1,$72,$7B,$4E,$A7,$27,$D0,$E4,$E3,$1E,$17,$11,$A7,$D0,$72
	.dc.b	$3D,$07,$63,$27,$23,$8C,$79,$63,$C8,$47,$04,$8F,$23,$E1,$7A,$C7
	.dc.b	$87,$0F,$C0,$9F,$80,$BF,$92,$E2,$FE,$5E,$1F,$9C,$7E,$7B,$F8,$5B
	.dc.b	$F4,$4F,$17,$E6,$1E,$8B,$F2,$3E,$82,$19,$C3,$C9,$5C,$E3,$DC,$64
	.dc.b	$DB,$E1,$7E,$33,$B9,$EB,$1B,$3D,$B3,$F0,$5C,$57,$95,$F9,$8F,$B5
	.dc.b	$F3,$D7,$57,$C2,$F5,$9F,$91,$3C,$6B,$E4,$5F,$C2,$C3,$E3,$B7,$0F
	.dc.b	$83,$5D,$5F,$B5,$87,$5B,$F3,$BD,$63,$EA,$90,$F4,$9F,$1A,$79,$4F
	.dc.b	$E1,$E7,$96,$1F,$23,$0F,$0D,$FC,$46,$BE,$A6,$D7,$81,$B3,$C6,$7C
	.dc.b	$F8,$7A,$5F,$8A,$DA,$F7,$C7,$53,$65,$36,$57,$6F,$84,$D9,$4D,$95
	.dc.b	$D9,$E5,$76,$D7,$6D,$37,$3E,$98,$CB,$63,$31,$F2,$EF,$E7,$3F,$82
	.dc.b	$AF,$CD,$DF,$F0,$8C,$3E,$E6,$1F,$87,$7E,$7B,$F8,$9B,$F9,$F7,$CE
	.dc.b	$BE,$4B,$4F,$3C,$F8,$4E,$57,$F2,$5B,$1F,$15,$AF,$C1,$6E,$1E,$D6
	.dc.b	$9C,$E3,$F0,$53,$FD,$09,$FF,$7F,$61,$FD,$61,$B1,$FE,$73,$23,$FE
	.dc.b	$73,$6B,$FE,$73,$6D,$7E,$A2,$7F,$80,$F6,$FC,$39,$F8,$8D,$AF,$DE
	.dc.b	$B5,$FD,$AC,$FF,$9F,$AB,$95,$70,$9F,$E0,$6D,$76,$D8,$DC,$74,$36
	.dc.b	$53,$65,$76,$F8,$2B,$C5,$1E,$96,$DF,$ED,$9B,$3F,$AB,$BF,$34,$FE
	.dc.b	$EA,$FC,$4D,$9E,$CD,$7B,$1E,$42,$9E,$42,$DF,$90,$F1,$3B,$6B,$B6
	.dc.b	$9B,$8F,$34,$6D,$AE,$DB,$1B,$8F,$55,$FB,$1F,$E3,$6D,$BF,$B3,$1C
	.dc.b	$CD,$94,$D9,$5D,$BD,$CD,$94,$D9,$5D,$CF,$8E,$32,$D8,$CD,$7C,$98
	.dc.b	$F7,$B4,$F5,$96,$7B,$C7,$1B,$FF,$16,$DF,$AD,$B3,$E7,$6B,$E3,$6C
	.dc.b	$F1,$8F,$F1,$D4,$F1,$69,$C4,$A7,$12,$C7,$17,$FC,$45,$1C,$5A,$71
	.dc.b	$29,$C4,$B1,$C4,$FF,$21,$C7,$16,$C7,$16,$BC,$5A,$71,$AF,$F7,$64
	.dc.b	$F1,$6C,$F1,$6B,$C5,$B3,$C6,$3F,$CB,$93,$C5,$A7,$12,$9C,$4B,$1C
	.dc.b	$5F,$F2,$04,$71,$69,$C4,$A7,$12,$C7,$1B,$FD,$6C,$7E,$BE,$FF,$5A
	.dc.b	$D3,$F5,$A5,$3F,$53,$7F,$83,$7F,$16,$FB,$0C,$4F,$48,$EF,$18,$BF
	.dc.b	$C1,$C7,$F2,$24,$C3,$05,$3E,$49,$7C,$57,$D1,$66,$38,$46,$1C,$07
	.dc.b	$0E,$DA,$F3,$8F,$06,$2B,$E7,$18,$6C,$F9,$A7,$4F,$A2,$9F,$1D,$B8
	.dc.b	$F5,$F8,$CE,$BC,$42,$30,$F3,$C3,$E0,$46,$21,$87,$67,$87,$57,$29
	.dc.b	$34,$91,$CC,$F9,$7E,$33,$C6,$F0,$F8,$2F,$1A,$92,$B8,$08,$78,$DC
	.dc.b	$09,$1A,$8C,$B7,$B3,$A9,$1C,$1D,$3C,$6F,$B5,$A7,$AE,$BF,$D2,$E4
	.dc.b	$E4,$F2,$F1,$CB,$EC,$D7,$E1,$EF,$F8,$5A,$78,$BF,$1A,$5E,$37,$CD
	.dc.b	$7F,$17,$3A,$C7,$F0,$6D,$8F,$C4,$35,$FC,$73,$5E,$DE,$33,$C6,$E2
	.dc.b	$D8,$E2,$D7,$8B,$5F,$CA,$78,$DF,$92,$7F,$3A,$D4,$D3,$5D,$DF,$8C
	.dc.b	$C3,$F0,$13,$F8,$9C,$FE,$AA,$9D,$47,$E9,$79,$AC,$FE,$A5,$AB,$C3
	.dc.b	$78,$D8,$8E,$87,$CC,$3F,$63,$1A,$AE,$FC,$FB,$F1,$B8,$36,$3C,$71
	.dc.b	$CF,$C6,$C5,$F3,$3F,$76,$E6,$7F,$1B,$6F,$0B,$15,$B8,$C4,$FE,$4B
	.dc.b	$47,$37,$A9,$72,$D8,$C3,$27,$AF,$8D,$70,$E8,$FC,$2F,$13,$C1,$E2
	.dc.b	$7F,$04,$6C,$FF,$09,$7E,$2F,$0E,$74,$57,$F0,$AF,$6D,$E8,$7D,$B6
	.dc.b	$E1,$98,$C1,$FE,$36,$7F,$C7,$67,$F8,$E4,$87,$F8,$E4,$FE,$F5,$D5
	.dc.b	$FC,$DF,$C4,$FE,$4C,$7F,$96,$23,$FB,$6E,$9F,$CA,$DF,$E6,$1B,$FF
	.dc.b	$50,$8C,$47,$F9,$8C,$FE,$C6,$E2,$79,$47,$F3,$03,$FA,$3C,$7F,$31
	.dc.b	$76,$74,$78,$57,$F8,$4C,$FE,$EA,$EE,$3F,$B3,$B6,$BC,$51,$C4,$EA
	.dc.b	$43,$C8,$70,$31,$5E,$43,$DF,$5D,$E3,$C6,$70,$9E,$AF,$A4,$C5,$79
	.dc.b	$9F,$92,$70,$AE,$95,$E8,$8F,$21,$C6,$3E,$6E,$DF,$D7,$23,$0D,$8C
	.dc.b	$C7,$F5,$E3,$25,$72,$FF,$60,$3D,$71,$C6,$7D,$51,$EF,$3C,$E7,$C0
	.dc.b	$73,$7E,$CA,$F3,$CF,$A6,$B9,$3E,$C3,$94,$69,$B1,$A7,$FC,$79,$5E
	.dc.b	$F7,$B9,$EF,$7B,$DF,$E0,$E7,$85,$FF,$8F,$E7,$85,$F5,$8F,$24,$EA
	.dc.b	$BE,$F1,$C6,$FE,$15,$F4,$7E,$09,$83,$D1,$5C,$9C,$1E,$63,$DC,$61
	.dc.b	$F1,$38,$3E,$61,$F4,$0B,$7F,$42,$3F,$CA,$71,$E0,$33,$3E,$37,$84
	.dc.b	$69,$B5,$FE,$58,$78,$46,$6B,$A4,$E6,$76,$7E,$C0,$DB,$98,$DC,$64
	.dc.b	$D9,$FE,$20,$36,$65,$DB,$93,$6F,$F9,$B9,$D9,$97,$66,$5D,$DF,$D6
	.dc.b	$BF,$74,$EA,$31,$3F,$42,$BE,$B3,$F1,$0C,$1F,$E4,$83,$FC,$56,$7F
	.dc.b	$89,$E5,$5F,$D5,$5B,$1F,$E2,$86,$BB,$8D,$11,$D3,$E2,$7F,$A0,$78
	.dc.b	$0F,$27,$18,$FE,$EA,$31,$3F,$B7,$AF,$CC,$46,$23,$F3,$F1,$8A,$E3
	.dc.b	$87,$F1,$EA,$D5,$FE,$A6,$B5,$7C,$CF,$E1,$33,$AA,$FE,$0D,$1F,$B3
	.dc.b	$8C,$3F,$B3,$37,$7D,$9F,$E3,$11,$A7,$F5,$46,$8E,$C7,$38,$E0,$FF
	.dc.b	$15,$8E,$18,$7F,$1C,$8E,$07,$B0,$B8,$7A,$5B,$7E,$68,$F8,$D7,$D2
	.dc.b	$BF,$CD,$17,$D3,$0E,$ED,$C3,$C7,$7F,$E4,$58,$C4,$77,$3E,$9D,$F3
	.dc.b	$AF,$F2,$8C,$EA,$BC,$55,$A3,$9E,$1F,$3E,$F8,$47,$F9,$12,$B8,$4F
	.dc.b	$9C,$FF,$3B,$38,$8F,$98,$73,$3C,$55,$EC,$7A,$9F,$3A,$BA,$BB,$8F
	.dc.b	$19,$B7,$8D,$78,$DE,$27,$CF,$9E,$33,$F0,$5B,$3E,$46,$CF,$F9,$89
	.dc.b	$AA,$BD,$0F,$DE,$3C,$53,$ED,$7C,$C7,$83,$0F,$A7,$5C,$AB,$8C,$7D
	.dc.b	$F9,$E2,$7C,$2B,$F5,$1E,$CB,$F5,$47,$B0,$EA,$53,$D8,$77,$C3,$EB
	.dc.b	$E1,$DC,$91,$93,$EA,$38,$79,$5F,$99,$D1,$E4,$3D,$D7,$E4,$BF,$74
	.dc.b	$78,$0E,$37,$E0,$78,$CF,$CD,$6C,$FC,$D6,$9F,$34,$A7,$CD,$BF,$24
	.dc.b	$FC,$18,$F7,$B5,$F0,$36,$FD,$11,$E2,$3D,$A7,$4C,$3E,$7E,$1F,$C0
	.dc.b	$23,$47,$B6,$7E,$95,$7B,$63,$E9,$61,$F4,$23,$B5,$7B,$67,$83,$CE
	.dc.b	$3D,$D1,$F4,$EF,$F8,$11,$E0,$7C,$67,$E7,$4F,$F6,$A3,$FC,$65,$B8
	.dc.b	$7F,$9F,$E3,$11,$FE,$C8,$3F,$CC,$9E,$27,$FB,$4A,$BF,$D7,$8D,$C3
	.dc.b	$FD,$B3,$E2,$7F,$B1,$A3,$FD,$97,$5F,$ED,$A6,$BF,$ED,$98,$FF,$46
	.dc.b	$46,$23,$FD,$CD,$E2,$7F,$B2,$6F,$FD,$90,$DB,$FF,$66,$F5,$7F,$B1
	.dc.b	$AB,$FC,$ED,$3F,$EC,$A7,$FD,$82,$D7,$3D,$3F,$C2,$3C,$4F,$F6,$34
	.dc.b	$7F,$9F,$7B,$FF,$D9,$4F,$FB,$0F,$C4,$FF,$62,$B9,$F1,$2E,$7F,$EB
	.dc.b	$F7,$FD,$89,$1C,$7A,$AC,$FF,$B0,$DF,$F7,$2F,$89,$FE,$C6,$D7,$FD
	.dc.b	$7D,$C9,$FE,$D8,$25,$FF,$B1,$08,$78,$F8,$FF,$D8,$D8,$7F,$B3,$7C
	.dc.b	$4F,$F6,$27,$89,$FE,$C3,$AF,$F6,$E4,$7F,$B1,$DF,$F5,$FB,$5F,$F6
	.dc.b	$63,$5F,$F6,$4B,$5F,$F7,$27,$87,$FE,$BD,$7F,$D9,$AD,$73,$E2,$7F
	.dc.b	$B1,$8F,$F5,$E3,$FE,$C1,$F1,$2A,$3F,$D8,$A5,$21,$FE,$BD,$CF,$FD
	.dc.b	$7F,$C9,$FE,$B7,$EE,$FF,$5C,$32,$3F,$D5,$7E,$25,$AF,$F5,$53,$FE
	.dc.b	$B6,$8F,$F5,$A9,$FE,$BA,$2C,$7F,$AD,$47,$89,$FE,$B2,$D7,$FD,$9E
	.dc.b	$D9,$F1,$F8,$9F,$EC,$AF,$13,$FD,$6B,$7C,$63,$1E,$16,$7E,$1B,$FD
	.dc.b	$AF,$3F,$EC,$A6,$BF,$ED,$16,$D1,$CF,$89,$FE,$B2,$9F,$82,$D7,$E2
	.dc.b	$3F,$EB,$47,$FD,$6D,$5F,$EB,$47,$E3,$8E,$CF,$CF,$3F,$ED,$AF,$77
	.dc.b	$E6,$68,$CF,$E7,$64,$9C,$C7,$E9,$2B,$F4,$3B,$61,$E4,$7F,$B6,$27
	.dc.b	$0F,$78,$B8,$78,$23,$FD,$DD,$3A,$BE,$D1,$D0,$87,$FB,$7E,$34,$72
	.dc.b	$A9,$1D,$C8,$7E,$D9,$CB,$C4,$CE,$1F,$BE,$8C,$BE,$F8,$F4,$9B,$39
	.dc.b	$1E,$27,$4F,$A0,$E7,$87,$C3,$76,$71,$8D,$DF,$B2,$33,$1E,$17,$89
	.dc.b	$E0,$70,$EA,$7E,$8D,$FC,$08,$F1,$47,$79,$EB,$39,$9E,$07,$E9,$F0
	.dc.b	$F9,$D7,$E1,$AD,$CF,$23,$58,$71,$8C,$C7,$23,$EC,$1E,$63,$E0,$19
	.dc.b	$AE,$B7,$C7,$0F,$21,$27,$D8,$FA,$9F,$01,$D0,$F8,$CE,$5E,$97,$CD
	.dc.b	$E2,$79,$8E,$38,$7A,$E7,$86,$1F,$20,$8E,$09,$3C,$4F,$64,$F1,$7A
	.dc.b	$9E,$39,$E2,$E2,$34,$E6,$7D,$8F,$C2,$9F,$8E,$70,$74,$71,$BF,$C6
	.dc.b	$CF,$0A,$E6,$74,$C3,$F4,$47,$CA,$AF,$9C,$E3,$0F,$85,$1D,$EB,$C3
	.dc.b	$3B,$8F,$92,$7B,$70,$F1,$5F,$6C,$3C,$33,$9B,$F8,$6F,$33,$04,$30
	.dc.b	$4B,$F1,$3F,$81,$AF,$46,$1D,$23,$71,$87,$EA,$DF,$A2,$B8,$C7,$98
	.dc.b	$E9,$5C,$6B,$CD,$1F,$84,$9F,$66,$1F,$33,$0F,$7D,$FC,$2B,$ED,$7F
	.dc.b	$6B,$0E,$D3,$CD,$B1,$E2,$3E,$5C,$65,$A6,$4F,$26,$1E,$6C,$3C,$11
	.dc.b	$F8,$48,$C6,$1F,$23,$0F,$0D,$7C,$EC,$3B,$DF,$8C,$EA,$53,$AE,$1E
	.dc.b	$16,$9E,$19,$EE,$FE,$20,$95,$F1,$1F,$D1,$9F,$46,$BE,$86,$1F,$1B
	.dc.b	$0F,$B6,$F9,$A7,$F2,$07,$CB,$C3,$E4,$57,$06,$C6,$4E,$85,$3E,$B5
	.dc.b	$FB,$AF,$F3,$15,$EE,$C3,$E0,$57,$B9,$AF,$B0,$F8,$0F,$B3,$0F,$24
	.dc.b	$7E,$1A,$7C,$4F,$80,$F6,$9E,$07,$2D,$3C,$93,$F1,$AF,$C9,$1F,$97
	.dc.b	$7E,$34,$7E,$C6,$BE,$24,$FD,$9C,$3E,$84,$F8,$DB,$87,$C8,$6C,$EC
	.dc.b	$FA,$91,$96,$99,$3D,$AE,$5A,$66,$7F,$9C,$BF,$E4,$C9,$FE,$DE,$FF
	.dc.b	$11,$30,$62,$91,$A8,$FE,$48,$68,$FE,$EF,$1A,$78,$47,$C6,$9F,$F2
	.dc.b	$BF,$6D,$BF,$E0,$AD,$AF,$DE,$5F,$F9,$56,$BE,$35,$FE,$2A,$B5,$5F
	.dc.b	$91,$AD,$5F,$EC,$23,$F6,$93,$A9,$FE,$2E,$DA,$FF,$32,$5F,$4A,$D9
	.dc.b	$F6,$2F,$C6,$D3,$C7,$5C,$5F,$F3,$4B,$5F,$F3,$49,$B3,$C8,$E6,$7D
	.dc.b	$F1,$9B,$F1,$B6,$FC,$77,$E9,$8F,$EA,$D7,$C6,$3F,$C9,$15,$EE,$9F
	.dc.b	$8B,$5F,$0D,$E4,$79,$0A,$79,$0F,$CB,$DF,$E3,$EF,$A4,$7F,$6F,$6B
	.dc.b	$FE,$2D,$6D,$77,$8F,$9D,$5D,$DE,$77,$F1,$9A,$7C,$63,$C3,$7F,$01
	.dc.b	$C9,$F1,$0C,$C7,$B2,$FB,$92,$32,$79,$4A,$79,$4B,$7E,$53,$F6,$E7
	.dc.b	$CC,$7B,$B6,$BE,$F4,$7C,$C3,$8C,$F7,$77,$7F,$49,$F9,$A7,$E1,$0C
	.dc.b	$9F,$34,$A7,$CD,$9E,$A7,$CD,$7A,$90,$DB,$F0,$CD,$DF,$C9,$6C,$7C
	.dc.b	$93,$71,$F1,$5E,$51,$A7,$E5,$1A,$3E,$31,$4F,$8C,$5B,$F8,$D5,$FA
	.dc.b	$3B,$F8,$15,$CE,$7E,$35,$FE,$46,$3D,$57,$C6,$B7,$39,$3D,$05,$8F
	.dc.b	$44,$F3,$BF,$95,$1F,$88,$3E,$54,$65,$F9,$4D,$7E,$54,$F3,$7D,$B1
	.dc.b	$93,$E4,$46,$6F,$CF,$3F,$58,$FB,$87,$9C,$F8,$F1,$BA,$F1,$E1,$FE
	.dc.b	$96,$93,$49,$3F,$C6,$DF,$F3,$3D,$FF,$98,$A3,$FC,$FC,$7F,$9F,$FE
	.dc.b	$7B,$FD,$CB,$1F,$E6,$CE,$DF,$F7,$2D,$FF,$B9,$FB,$7F,$DC,$77,$FE
	.dc.b	$BF,$97,$6E,$2B,$FD,$89,$18,$7F,$D8,$D7,$FE,$E5,$6D,$71,$ED,$FF
	.dc.b	$74,$B6,$32,$7F,$B3,$7B,$7F,$DD,$4F,$4E,$DF,$F7,$4B,$FE,$F9,$7F
	.dc.b	$DC,$F3,$E3,$ED,$B3,$E2,$7F,$D3,$F8,$BB,$3F,$DD,$7C,$1F,$EE,$6C
	.dc.b	$FF,$A1,$CE,$5B,$3F,$EE,$78,$C5,$FF,$AD,$97,$67,$FB,$9D,$BD,$9F
	.dc.b	$EE,$79,$D4,$FF,$BA,$DF,$F7,$37,$67,$FB,$A1,$FF,$73,$9F,$D4,$9F
	.dc.b	$E9,$DD,$9F,$EE,$1C,$7F,$D3,$73,$FE,$E9,$3F,$DB,$DC,$99,$D7,$FD
	.dc.b	$83,$D9,$67,$FD,$E5,$CA,$8F,$F7,$44,$7F,$B9,$7B,$2E,$1F,$EE,$4B
	.dc.b	$FF,$74,$BD,$BB,$3F,$D5,$CE,$6B,$FD,$C9,$3F,$49,$FF,$7F,$4F,$FB
	.dc.b	$9A,$31,$08,$C2,$9F,$F5,$9F,$5F,$FB,$86,$01,$C3,$70,$67,$FD,$D2
	.dc.b	$73,$93,$D7,$FC,$C1,$2F,$FB,$CC,$65,$FE,$F1,$D5,$FC,$05,$FD,$8E
	.dc.b	$7F,$AB,$8F,$DB,$F1,$FF,$2A,$6E,$DF,$CF,$F4,$FD,$8D,$FE,$BD,$AF
	.dc.b	$F0,$E7,$F8,$54,$F0,$8F,$DF,$DF,$F1,$97,$F9,$9C,$65,$FD,$BC,$F2
	.dc.b	$BF,$04,$79,$0E,$D5,$C6,$75,$7E,$38,$C1,$E9,$9E,$07,$F0,$F3,$F8
	.dc.b	$59,$FC,$92,$7C,$AE,$AB,$BB,$A3,$84,$78,$EF,$E3,$46,$67,$73,$E2
	.dc.b	$8F,$69,$D6,$BC,$4E,$EF,$CC,$E4,$F5,$94,$F9,$11,$C2,$7A,$47,$C4
	.dc.b	$8F,$6B,$C0,$D9,$C6,$BC,$71,$A8,$F4,$B6,$3F,$9F,$C7,$98,$E2,$68
	.dc.b	$E5,$7E,$EA,$F0,$9F,$D0,$0B,$5F,$CF,$A7,$C1,$7E,$C7,$9D,$F8,$9C
	.dc.b	$9D,$DF,$0B,$EB,$32,$53,$8C,$73,$8D,$D7,$86,$BD,$53,$9B,$F3,$46
	.dc.b	$2F,$B4,$61,$EB,$5C,$5B,$3E,$97,$B9,$C0,$FF,$19,$4F,$E4,$E7,$84
	.dc.b	$68,$F3,$5F,$92,$B9,$36,$3F,$AF,$C7,$8D,$DD,$78,$DE,$55,$D2,$B8
	.dc.b	$47,$F5,$CA,$E1,$39,$AE,$15,$A3,$FA,$ED,$72,$AF,$0C,$E8,$91,$EA
	.dc.b	$3F,$39,$7F,$1D,$C1,$F9,$33,$E0,$1F,$B2,$36,$62,$FF,$4A,$EC,$F5
	.dc.b	$9C,$9F,$C1,$BF,$00,$F9,$A5,$3C,$C5,$BF,$54,$FF,$0C,$AF,$3C,$F8
	.dc.b	$4F,$F2,$A1,$FC,$80,$FF,$0C,$94,$FF,$0C,$96,$51,$FE,$8F,$6C,$7F
	.dc.b	$1C,$21,$FE,$90,$8D,$1E,$17,$F9,$B7,$84,$F0,$31,$1F,$C0,$63,$0F
	.dc.b	$F0,$59,$FF,$0C,$C6,$23,$F8,$14,$62,$78,$E1,$FC,$8A,$B4,$FF,$09
	.dc.b	$AF,$E2,$71,$8C,$7F,$69,$7C,$67,$B1,$AB,$E7,$3F,$BB,$AE,$13,$FB
	.dc.b	$63,$47,$03,$D1,$3C,$1F,$E1,$91,$C1,$FD,$AD,$77,$7D,$27,$22,$9C
	.dc.b	$8E,$97,$EB,$AE,$91,$FC,$22,$BA,$61,$FE,$6F,$BE,$97,$EF,$AE,$47
	.dc.b	$47,$93,$D3,$0F,$66,$1D,$2F,$FD,$19,$87,$4C,$3B,$B6,$F8,$1F,$D9
	.dc.b	$1C,$3D,$6F,$F1,$11,$8C,$3C,$42,$33,$87,$7B,$FC,$74,$E6,$FA,$4E
	.dc.b	$5F,$C2,$B5,$FE,$ED,$5C,$9F,$19,$F0,$E3,$53,$CE,$17,$E3,$AF,$BF
	.dc.b	$1D,$0E,$38,$7C,$0B,$E3,$3D,$67,$D8,$7D,$A8,$CB,$E6,$38,$E1,$EA
	.dc.b	$9E,$18,$7C,$07,$E6,$CE,$8F,$01,$D0,$F4,$19,$3D,$A7,$53,$57,$E5
	.dc.b	$75,$3E,$DC,$3D,$91,$F9,$33,$89,$F0,$30,$F3,$B4,$E3,$7E,$63,$E1
	.dc.b	$BF,$86,$AF,$43,$EF,$36,$78,$DC,$BE,$83,$CB,$87,$C1,$75,$3D,$27
	.dc.b	$F1,$F1,$A3,$E0,$BC,$B0,$F6,$46,$4F,$4E,$1E,$C7,$F4,$E7,$0A,$E8
	.dc.b	$7C,$C7,$85,$7C,$3C,$3D,$F5,$C6,$FF,$49,$5C,$4F,$8D,$5E,$43,$C5
	.dc.b	$87,$A6,$3E,$31,$D3,$0F,$8C,$F0,$7C,$13,$C2,$3A,$1E,$63,$E5,$47
	.dc.b	$99,$F9,$47,$33,$E2,$9C,$CC,$C7,$37,$E0,$E1,$EA,$23,$F4,$8F,$2C
	.dc.b	$6C,$F8,$0E,$70,$F8,$17,$FC,$0A,$FC,$53,$F5,$EB,$FD,$50,$FF,$99
	.dc.b	$72,$FF,$55,$CF,$F3,$F6,$C7,$F9,$D3,$B3,$FD,$58,$FF,$96,$1F,$F5
	.dc.b	$91,$1F,$F4,$8F,$67,$FA,$C2,$BF,$D5,$8F,$FA,$AE,$1C,$DF,$EA,$B7
	.dc.b	$F2,$AD,$C3,$FD,$6D,$D9,$FE,$AA,$9F,$F5,$8F,$65,$78,$BF,$57,$B3
	.dc.b	$FD,$57,$1D,$3B,$3F,$D5,$77,$FE,$BF,$7F,$D5,$31,$E0,$EC,$B8,$9F
	.dc.b	$F5,$5F,$67,$FA,$AA,$7F,$D7,$EF,$FA,$9C,$97,$FE,$AA,$3F,$D4,$91
	.dc.b	$FD,$86,$BF,$D4,$EF,$FA,$9B,$B3,$2F,$FA,$90,$87,$F6,$7E,$BF,$F5
	.dc.b	$43,$FD,$AF,$87,$FD,$BC,$4D,$FF,$B6,$46,$5A,$FF,$A6,$1A,$FF,$A5
	.dc.b	$DB,$D5,$FE,$E7,$3F,$D5,$F1,$FE,$AA,$F7,$2B,$FE,$EA,$6F,$19,$A6
	.dc.b	$3E,$9A,$5F,$FA,$51,$B5,$E4,$EA,$FF,$4F,$93,$B3,$FD,$3F,$19,$F0
	.dc.b	$1E,$4F,$E0,$A3,$F1,$9F,$43,$FA,$BC,$CF,$E1,$63,$FD,$DB,$D1,$98
	.dc.b	$91,$FE,$B7,$E1,$FF,$7B,$F1,$7F,$BE,$6B,$33,$FC,$72,$F9,$BB,$8E
	.dc.b	$8A,$32,$EE,$F3,$7D,$35,$FC,$F4,$74,$C3,$B5,$73,$3D,$24,$3F,$2F
	.dc.b	$87,$6C,$3A,$CF,$F4,$D6,$DF,$E8,$2B,$91,$E1,$39,$4F,$67,$99,$88
	.dc.b	$F1,$BC,$23,$47,$5B,$F3,$4F,$59,$E2,$75,$24,$F0,$8D,$3C,$63,$B1
	.dc.b	$F4,$0E,$D1,$E0,$78,$BC,$08,$FF,$BD,$27,$C2,$F1,$3C,$33,$D5,$E2
	.dc.b	$65,$F1,$3C,$4F,$09,$FC,$0A,$FE,$54,$F4,$8C,$1E,$43,$81,$D0,$C9
	.dc.b	$E0,$AF,$45,$78,$4E,$F1,$A7,$11,$A3,$D7,$19,$70,$F6,$8E,$38,$7E
	.dc.b	$16,$31,$3A,$78,$1E,$D7,$18,$7E,$02,$30,$43,$F2,$57,$ED,$AE,$91
	.dc.b	$C8,$C9,$F1,$1C,$9A,$9E,$91,$A3,$E0,$1D,$5C,$9A,$8E,$38,$7C,$B8
	.dc.b	$C3,$F3,$A7,$B3,$5F,$99,$3C,$1F,$AB,$3C,$1F,$A4,$78,$70,$F9,$6F
	.dc.b	$B6,$7C,$D7,$ED,$BE,$95,$AA,$F7,$56,$9F,$09,$D2,$FD,$37,$FB,$4B
	.dc.b	$F4,$36,$7C,$6D,$7F,$05,$3E,$57,$E1,$9E,$5C,$3E,$B5,$FA,$E3,$75
	.dc.b	$C7,$0F,$25,$F1,$7F,$0E,$F8,$A7,$DF,$5E,$77,$E2,$E1,$EC,$C3,$AB
	.dc.b	$63,$F0,$A7,$C3,$7D,$2F,$26,$BE,$69,$E9,$7F,$15,$F3,$3E,$4C,$3C
	.dc.b	$D7,$F8,$C6,$D7,$63,$DC,$77,$3E,$29,$DC,$A7,$7B,$F0,$D7,$E7,$1B
	.dc.b	$5F,$A2,$BF,$1B,$EC,$AF,$A9,$7F,$6D,$A7,$D8,$3E,$7E,$1F,$1D,$EC
	.dc.b	$78,$30,$EC,$DA,$ED,$87,$82,$3D,$07,$D4,$3D,$05,$39,$96,$79,$B6
	.dc.b	$79,$BD,$4E,$71,$F2,$5E,$6D,$79,$B4,$E6,$58,$E7,$1D,$4E,$6F,$D1
	.dc.b	$79,$B5,$E6,$D8,$E6,$D3,$9B,$D6,$39,$9F,$66,$BF,$21,$3F,$CE,$63
	.dc.b	$FC,$15,$5F,$DF,$30,$34,$91,$FD,$AC,$FF,$02,$C6,$9C,$1C,$8F,$38
	.dc.b	$3F,$C5,$FD,$1F,$8A,$7F,$66,$D8,$FE,$1E,$D9,$FF,$1B,$47,$E5,$2B
	.dc.b	$F0,$D3,$9A,$FE,$E8,$D7,$FC,$72,$DC,$3F,$C7,$31,$E2,$3B,$96,$BB
	.dc.b	$B6,$FB,$9E,$3A,$E5,$7E,$99,$F4,$9B,$3C,$25,$3C,$25,$BF,$0B,$F5
	.dc.b	$8F,$E2,$D3,$D1,$F8,$97,$FE,$4F,$AF,$AE,$7B,$A3,$93,$E1,$8D,$47
	.dc.b	$DA,$3D,$67,$52,$D7,$E0,$9E,$E7,$23,$BD,$78,$5F,$59,$C6,$7E,$1B
	.dc.b	$6F,$E1,$BF,$1E,$78,$4F,$8A,$3F,$95,$CF,$F8,$6A,$72,$48,$F0,$1D
	.dc.b	$AB,$C8,$F9,$AF,$DE,$F9,$8F,$01,$F3,$0A,$7C,$C2,$C7,$CC,$39,$DF
	.dc.b	$A1,$B7,$E8,$9F,$65,$F8,$DE,$37,$F2,$E3,$15,$C2,$7C,$0D,$7C,$0D
	.dc.b	$3C,$13,$E8,$8F,$51,$F4,$EF,$C5,$19,$6C,$66,$FC,$95,$E8,$BF,$0D
	.dc.b	$EE,$72,$F1,$8C,$D7,$13,$CB,$1B,$69,$93,$CE,$E5,$A6,$6B,$D2,$D8
	.dc.b	$F8,$4F,$D6,$BF,$B8,$7E,$BD,$FA,$93,$F0,$67,$71,$F2,$8F,$C5,$9C
	.dc.b	$6B,$CE,$FE,$10,$FE,$B2,$E2,$BA,$47,$C8,$34,$7F,$53,$8D,$5F,$EC
	.dc.b	$63,$A1,$0F,$01,$F8,$BA,$F7,$5F,$B2,$39,$BF,$E2,$8A,$FF,$74,$5B
	.dc.b	$9B,$F8,$E1,$2B,$FD,$37,$19,$6C,$7F,$AE,$2F,$F6,$0F,$FA,$DB,$9B
	.dc.b	$F9,$21,$27,$8B,$FC,$9F,$9B,$FD,$71,$1F,$EB,$D3,$F9,$07,$88,$ED
	.dc.b	$FD,$D5,$7F,$0B,$8C,$4F,$FA,$FB,$3D,$D7,$EF,$ED,$1F,$CA,$7E,$B3
	.dc.b	$75,$D4,$38,$CF,$FD,$95,$1B,$BE,$55,$63,$B1,$FC,$E2,$FF,$D9,$6D
	.dc.b	$8F,$F6,$3B,$FC,$53,$C3,$75,$62,$3F,$DC,$F1,$87,$FD,$83,$D5,$FE
	.dc.b	$CA,$7F,$93,$35,$78,$7F,$EC,$06,$EF,$FE,$C8,$8C,$1F,$EC,$2E,$59
	.dc.b	$1F,$EC,$EE,$34,$62,$BF,$DA,$70,$E8,$FF,$61,$C7,$F3,$53,$FD,$5C
	.dc.b	$E1,$FE,$71,$D1,$FE,$C6,$AF,$F6,$1C,$7F,$B2,$7A,$3F,$D8,$EF,$19
	.dc.b	$EF,$D1,$E1,$F9,$BF,$F6,$EE,$26,$92,$BF,$13,$CF,$F9,$77,$F5,$E7
	.dc.b	$E4,$B3,$FD,$3F,$07,$E5,$78,$FF,$42,$7E,$BA,$B8,$46,$7A,$3F,$95
	.dc.b	$5F,$EF,$0E,$47,$EB,$8E,$B5,$CA,$BF,$83,$C7,$ED,$AF,$B9,$FC,$25
	.dc.b	$E5,$3B,$6D,$76,$79,$B6,$FC,$51,$FB,$83,$65,$36,$57,$6F,$EE,$4D
	.dc.b	$94,$D9,$5D,$9F,$B5,$76,$D7,$6D,$37,$5F,$D8,$E3,$2D,$8C,$C7,$C1
	.dc.b	$3E,$33,$FC,$B5,$EB,$1B,$3E,$2B,$FC,$DE,$FE,$4B,$FC,$4E,$FF,$90
	.dc.b	$DF,$C3,$AD,$CF,$D2,$3F,$88,$1D,$9F,$9C,$F9,$0F,$11,$A3,$84,$FD
	.dc.b	$F7,$FB,$51,$E4,$36,$F8,$1F,$99,$1E,$08,$FB,$11,$D4,$F3,$57,$E2
	.dc.b	$27,$55,$E7,$9F,$21,$ED,$3B,$1C,$0F,$09,$CE,$BF,$AA,$BE,$77,$BD
	.dc.b	$FC,$C8,$E0,$7F,$6A,$39,$C7,$03,$D0,$FD,$DA,$F1,$BE,$A3,$C6,$7C
	.dc.b	$2B,$F9,$F3,$C5,$F5,$46,$A3,$C3,$3C,$A3,$A1,$9B,$F7,$3E,$76,$CF
	.dc.b	$8A,$FF,$11,$5C,$5B,$5C,$4F,$21,$C4,$B3,$C6,$7F,$0D,$1B,$3C,$26
	.dc.b	$CB,$1B,$7E,$CD,$FA,$9D,$57,$81,$B5,$D1,$A7,$03,$E9,$C7,$CF,$AF
	.dc.b	$89,$18,$9F,$51,$C6,$3C,$11,$F5,$27,$F3,$CF,$DB,$BF,$F3,$3D,$F8
	.dc.b	$EB,$67,$49,$FB,$47,$F8,$AC,$FE,$50,$A3,$FD,$00,$48,$FE,$3A,$43
	.dc.b	$FC,$FE,$E5,$A7,$F0,$E7,$FB,$7F,$77,$8A,$3F,$88,$B5,$FE,$32,$DB
	.dc.b	$FE,$48,$DC,$3F,$40,$D9,$FC,$23,$5F,$CB,$36,$FF,$C2,$ED,$BF,$0C
	.dc.b	$F8,$CF,$09,$4F,$09,$6F,$C2,$DB,$F0,$B4,$F0,$94,$F0,$96,$FC,$2D
	.dc.b	$AF,$0B,$FC,$D9,$FF,$4C,$DF,$96,$3F,$D4,$56,$BF,$F3,$E4,$E6,$7F
	.dc.b	$BF,$C6,$1E,$87,$07,$0F,$53,$DA,$53,$DA,$59,$F6,$B6,$3D,$AD,$3D
	.dc.b	$A5,$3D,$A5,$BF,$6B,$6F,$DB,$5F,$D6,$DF,$6C,$79,$8F,$49,$FE,$9C
	.dc.b	$BF,$64,$FB,$E3,$06,$DC,$2A,$4F,$0B,$F7,$D7,$7A,$F7,$B5,$F7,$B4
	.dc.b	$F7,$96,$3D,$ED,$8F,$7B,$5F,$7B,$6B,$DE,$DB,$F7,$D7,$F3,$37,$DF
	.dc.b	$5F,$0C,$FA,$87,$B0,$C1,$F0,$8F,$65,$FA,$DD,$57,$57,$89,$A3,$E2
	.dc.b	$57,$C0,$3C,$0F,$C0,$6B,$F0,$1B,$1F,$01,$A7,$C0,$2C,$7C,$06,$BF
	.dc.b	$01,$B7,$F0,$1B,$7F,$02,$7F,$9A,$3F,$02,$BD,$11,$ED,$AF,$AE,$FA
	.dc.b	$0F,$6D,$72,$7D,$B1,$A7,$E2,$CF,$E0,$23,$0F,$43,$17,$F1,$E7,$E2
	.dc.b	$1C,$2B,$E2,$3E,$43,$E2,$14,$F8,$85,$9F,$88,$D8,$F8,$8D,$3E,$21
	.dc.b	$4F,$88,$5A,$F8,$8D,$BF,$88,$FF,$67,$7E,$24,$FB,$E3,$E2,$5F,$A6
	.dc.b	$BD,$B3,$9B,$F6,$D7,$C7,$3D,$AF,$C1,$8F,$6C,$75,$3D,$A7,$E9,$7A
	.dc.b	$7F,$D2,$88,$C1,$F3,$BA,$7F,$D3,$E7,$FA,$55,$FF,$32,$1F,$E8,$73
	.dc.b	$F9,$FB,$FE,$97,$AF,$F4,$7C,$73,$BF,$F4,$4B,$F7,$67,$FD,$53,$CD
	.dc.b	$FE,$9C,$93,$63,$FD,$35,$CF,$FE,$9C,$AE,$67,$EE,$E7,$81,$B7,$35
	.dc.b	$DB,$A7,$FD,$31,$3A,$C3,$FD,$66,$FF,$A5,$2B,$FA,$AB,$FE,$A2,$7D
	.dc.b	$74,$E7,$FF,$49,$47,$F5,$A6,$BF,$14,$FF,$4B,$F4,$7F,$A6,$5F,$91
	.dc.b	$D1,$FE,$95,$4F,$FA,$5F,$B3,$FD,$2C,$4D,$BE,$4C,$FC,$F8,$FE,$2E
	.dc.b	$F5,$E3,$7B,$B8,$8F,$A9,$18,$AF,$F4,$D4,$E6,$53,$97,$EA,$F3,$7F
	.dc.b	$A5,$79,$BB,$AE,$7F,$F4,$AC,$FF,$A6,$BA,$3F,$D3,$2F,$FA,$7E,$FF
	.dc.b	$D2,$8F,$FA,$5C,$87,$FA,$9B,$A3,$FD,$33,$F2,$52,$3F,$D5,$8D,$7E
	.dc.b	$5F,$3F,$F6,$B3,$F3,$BC,$3F,$E5,$78,$FC,$D9,$FB,$A8,$D3,$4F,$EC
	.dc.b	$4F,$0E,$8F,$D1,$C7,$F0,$23,$F5,$51,$97,$F3,$73,$FA,$47,$27,$33
	.dc.b	$FA,$84,$6D,$FD,$D9,$FA,$6D,$BF,$8B,$C6,$79,$3F,$7B,$39,$39,$9E
	.dc.b	$83,$99,$D5,$FE,$7D,$1C,$67,$F8,$1C,$F1,$3F,$60,$71,$9F,$DB,$BC
	.dc.b	$8C,$47,$89,$E8,$D7,$27,$8A,$38,$9E,$07,$66,$63,$F7,$F5,$F3,$E7
	.dc.b	$71,$8B,$F9,$75,$E5,$AD,$4F,$0B,$FA,$73,$E0,$3D,$37,$C9,$FA,$46
	.dc.b	$27,$83,$F6,$0F,$19,$93,$75,$F5,$5C,$1E,$1A,$EF,$5D,$63,$33,$B2
	.dc.b	$5F,$A1,$B1,$F8,$0A,$E3,$7E,$D8,$C1,$E3,$3D,$53,$EE,$7C,$73,$C1
	.dc.b	$F2,$D7,$58,$C4,$7C,$58,$C3,$CE,$FC,$D5,$CE,$3D,$46,$6F,$C4,$FE
	.dc.b	$3A,$3C,$ED,$BE,$B1,$8A,$FC,$7B,$6B,$F1,$2D,$AF,$E9,$67,$8C,$F6
	.dc.b	$9F,$89,$9F,$1D,$F9,$A3,$15,$E2,$9E,$B7,$E0,$24,$F9,$CF,$82,$74
	.dc.b	$8F,$C6,$3F,$22,$BA,$46,$0E,$F1,$BA,$FC,$B4,$62,$3D,$C7,$53,$D0
	.dc.b	$60,$F2,$9E,$03,$0F,$58,$F4,$1F,$0A,$BC,$F3,$C0,$E7,$3E,$C7,$B5
	.dc.b	$FC,$77,$FB,$79,$FD,$F8,$FF,$49,$CF,$F3,$07,$07,$F9,$80,$C9,$FC
	.dc.b	$D2,$3F,$C1,$8E,$0F,$E3,$66,$63,$F8,$D9,$FD,$0E,$7F,$C9,$FE,$07
	.dc.b	$F4,$59,$FC,$7C,$FF,$7C,$8C,$4F,$E2,$23,$13,$C8,$FD,$3C,$EA,$FF
	.dc.b	$57,$1F,$B1,$8C,$3F,$98,$36,$71,$34,$5A,$FD,$54,$70,$8F,$CA,$9A
	.dc.b	$3F,$6B,$7F,$AD,$BF,$55,$FF,$5D,$AE,$13,$9A,$EE,$70,$AF,$4B,$4C
	.dc.b	$9F,$00,$B7,$EF,$6D,$FB,$23,$CB,$1E,$78,$FF,$0A,$DF,$82,$B1,$3F
	.dc.b	$E1,$38,$C3,$DC,$FF,$31,$38,$7C,$06,$0E,$27,$F3,$17,$55,$E3,$8D
	.dc.b	$1C,$DF,$1B,$6B,$C6,$F2,$BF,$E3,$15,$CA,$7D,$47,$C7,$9D,$5F,$CA
	.dc.b	$7D,$06,$0F,$F3,$9C,$F1,$7E,$D1,$E2,$3D,$37,$E2,$6D,$78,$9F,$21
	.dc.b	$FC,$C2,$79,$3F,$72,$7F,$22,$F7,$3A,$3E,$F3,$A1,$67,$A4,$76,$3A
	.dc.b	$1C,$63,$9B,$E0,$3E,$51,$98,$F8,$0E,$6F,$E2,$B9,$3E,$09,$9B,$F8
	.dc.b	$8A,$FD,$33,$F8,$D8,$F3,$D7,$CC,$7C,$67,$13,$E4,$9D,$4E,$45,$9E
	.dc.b	$75,$F9,$48,$F6,$1E,$83,$91,$D8,$FA,$37,$F1,$2B,$C1,$1B,$38,$DF
	.dc.b	$B1,$E3,$3F,$11,$CD,$7B,$2F,$C3,$1F,$9E,$3A,$DF,$CE,$AF,$A1,$1E
	.dc.b	$C3,$89,$EC,$3A,$C7,$CC,$72,$79,$CF,$BC,$F3,$3C,$F7,$E5,$8D,$CF
	.dc.b	$94,$EE,$FA,$DB,$1B,$6D,$FA,$67,$E5,$10,$F1,$9F,$68,$EB,$5E,$59
	.dc.b	$E3,$1E,$93,$26,$DC,$BE,$9A,$FB,$D1,$F3,$E3,$E2,$9F,$0D,$D9,$E9
	.dc.b	$9F,$6D,$79,$67,$FD,$24,$7F,$35,$E8,$FF,$49,$3F,$E9,$03,$FC,$FB
	.dc.b	$5F,$E8,$28,$FF,$4B,$C7,$FB,$1B,$9B,$FD,$04,$BB,$5D,$3F,$E8,$D8
	.dc.b	$C3,$63,$FD,$13,$7F,$EA,$36,$CE,$8F,$F4,$E5,$FF,$A0,$E3,$9C,$60
	.dc.b	$E1,$5F,$E8,$7C,$BF,$D3,$B7,$89,$FF,$52,$61,$FE,$99,$5D,$3C,$95
	.dc.b	$7F,$A6,$EF,$95,$7A,$A7,$FD,$33,$7F,$01,$FF,$4C,$C7,$FA,$6D,$FF
	.dc.b	$4C,$74,$7F,$5B,$E9,$FF,$4C,$C7,$FA,$CD,$AF,$FA,$68,$FE,$C9,$D3
	.dc.b	$FE,$99,$3F,$B3,$3F,$E9,$A3,$FD,$7F,$C6,$EF,$FF,$48,$37,$83,$FD
	.dc.b	$36,$D9,$DC,$FF,$A5,$79,$FF,$D2,$51,$FE,$94,$E7,$FF,$48,$BF,$EC
	.dc.b	$F8,$73,$FF,$A4,$A3,$FD,$45,$F4,$3F,$DD,$F0,$34,$91,$E0,$F0,$34
	.dc.b	$47,$07,$F7,$BE,$EF,$F0,$07,$11,$A4,$8F,$F0,$07,$77,$F1,$6C,$1F
	.dc.b	$2B,$4F,$F7,$2C,$FF,$B9,$07,$1B,$7F,$F5,$C4,$EE,$7F,$DD,$53,$97
	.dc.b	$FD,$C5,$7C,$AF,$AC,$6A,$F8,$B4,$D9,$4F,$05,$73,$7F,$D7,$F7,$DD
	.dc.b	$F1,$CF,$E3,$EF,$B4,$E6,$B8,$62,$9C,$DF,$5C,$3C,$6F,$23,$C2,$53
	.dc.b	$C3,$87,$99,$C4,$70,$BE,$B5,$D2,$73,$87,$C8,$C3,$C1,$1D,$E3,$C9
	.dc.b	$87,$9E,$3D,$D3,$A7,$89,$B3,$35,$D6,$FE,$16,$1E,$1B,$E5,$5E,$7B
	.dc.b	$E5,$5A,$C3,$BC,$FB,$23,$CE,$F5,$3B,$9C,$70,$ED,$7E,$F9,$DC,$E4
	.dc.b	$E0,$F5,$7C,$A4,$39,$B9,$78,$11,$C1,$23,$D2,$F6,$31,$1F,$F0,$02
	.dc.b	$BE,$55,$7A,$A3,$F0,$27,$79,$E1,$5E,$72,$47,$BD,$EA,$61,$E2,$78
	.dc.b	$DC,$BD,$CF,$6B,$87,$F0,$26,$4F,$6B,$9B,$F9,$EF,$5C,$3E,$A4,$F8
	.dc.b	$48,$78,$2B,$DA,$F4,$24,$70,$75,$3D,$6F,$E0,$4E,$6F,$BD,$74,$8E
	.dc.b	$06,$67,$EC,$C7,$9E,$FE,$2B,$BC,$3E,$55,$F2,$8F,$0C,$62,$B8,$C7
	.dc.b	$88,$F4,$CE,$EF,$BD,$7D,$49,$F8,$E7,$7C,$3D,$0E,$0E,$47,$BE,$B9
	.dc.b	$4F,$78,$FB,$53,$F2,$63,$A4,$7C,$E3,$A1,$E2,$BF,$87,$1A,$C3,$E5
	.dc.b	$E1,$F0,$6B,$99,$C0,$F0,$E1,$EF,$3C,$2F,$95,$C5,$F4,$8E,$27,$C4
	.dc.b	$34,$FC,$49,$ED,$1F,$89,$3B,$38,$AF,$9F,$1F,$02,$31,$87,$E2,$DF
	.dc.b	$93,$1D,$AF,$C6,$FB,$CF,$8D,$7E,$37,$B1,$E3,$35,$1E,$87,$E7,$4F
	.dc.b	$67,$EF,$1E,$83,$D6,$F8,$1E,$71,$E0,$BF,$C9,$BC,$E3,$C5,$3C,$DF
	.dc.b	$04,$EA,$BE,$16,$1F,$47,$0F,$B1,$5E,$B9,$F4,$57,$C2,$C3,$B5,$7E
	.dc.b	$C5,$ED,$1F,$62,$F9,$5F,$CA,$6B,$F2,$5B,$87,$A6,$7A,$61,$DB,$0F
	.dc.b	$C7,$DF,$A2,$F6,$7E,$35,$F9,$37,$C6,$77,$7E,$19,$F4,$1F,$2A,$B8
	.dc.b	$BF,$24,$E2,$F3,$8F,$BB,$5C,$E7,$F2,$87,$13,$99,$E3,$2C,$77,$7E
	.dc.b	$8D,$FA,$5E,$26,$CC,$9F,$DD,$08,$7F,$BB,$4F,$EE,$47,$FB,$BF,$0F
	.dc.b	$E9,$0D,$8F,$EF,$E4,$78,$CF,$C8,$7F,$C9,$67,$63,$F0,$7C,$FF,$42
	.dc.b	$BF,$4C,$DA,$FF,$25,$CE,$EB,$EE,$4E,$63,$9C,$75,$79,$9F,$03,$1F
	.dc.b	$E4,$D3,$B8,$FE,$AC,$D9,$FE,$EA,$D9,$FF,$2A,$47,$33,$BC,$F3,$3F
	.dc.b	$89,$CE,$DF,$F2,$81,$E5,$9E,$EE,$CF,$E2,$11,$B7,$FC,$9A,$DB,$34
	.dc.b	$91,$C6,$3E,$1B,$6F,$E1,$DF,$96,$FB,$CF,$84,$F2,$9B,$3E,$2B,$F9
	.dc.b	$03,$71,$C6,$36,$71,$3C,$E7,$13,$E3,$95,$F8,$ED,$8F,$8E,$75,$79
	.dc.b	$C7,$57,$99,$D6,$BD,$6D,$AF,$5D,$71,$8D,$D7,$1A,$DB,$6B,$73,$F9
	.dc.b	$88,$FA,$AF,$33,$DE,$66,$BD,$E7,$5A,$E6,$DA,$E7,$3F,$C8,$0F,$E4
	.dc.b	$97,$F4,$9E,$87,$C2,$2D,$FC,$26,$D7,$C2,$7C,$07,$C2,$7B,$3E,$43
	.dc.b	$B1,$F9,$63,$89,$F8,$A8,$F7,$3B,$BF,$E2,$F3,$EA,$8E,$A7,$C6,$7A
	.dc.b	$BE,$59,$E2,$D8,$CC,$79,$CC,$95,$C9,$E9,$8F,$2C,$7C,$C3,$71,$99
	.dc.b	$F1,$CF,$B4,$F1,$BF,$31,$B7,$F3,$23,$71,$E7,$32,$7B,$63,$F8,$05
	.dc.b	$FC,$FB,$EF,$3F,$0E,$FC,$AD,$38,$9E,$D2,$C7,$AD,$B1,$E7,$AE,$B1
	.dc.b	$F9,$73,$91,$EE,$8F,$98,$D3,$D8,$53,$D0,$7F,$B1,$39,$3F,$D9,$1C
	.dc.b	$DF,$EC,$7C,$7F,$D8,$D3,$FE,$CC,$C0,$FF,$B3,$39,$BF,$D9,$31,$FE
	.dc.b	$E5,$88,$64,$E0,$53,$B3,$FE,$D7,$7F,$83,$96,$7F,$83,$B6,$FF,$D4
	.dc.b	$6F,$FB,$94,$E2,$58,$DB,$72,$FF,$74,$5F,$92,$35,$5E,$C8,$FF,$7A
	.dc.b	$61,$FE,$F0,$3F,$DE,$33,$FE,$E0,$3F,$DB,$F8,$7F,$B8,$5F,$F6,$F9
	.dc.b	$FD,$17,$07,$2F,$FB,$79,$BC,$DF,$EE,$33,$FD,$BD,$3F,$EE,$02,$1F
	.dc.b	$20,$FF,$6A,$71,$7F,$B4,$DB,$DB,$FD,$45,$72,$39,$7F,$DB,$71,$8A
	.dc.b	$E0,$B8,$5D,$9F,$ED,$B2,$47,$FB,$A1,$FF,$69,$3E,$9E,$4A,$FF,$5D
	.dc.b	$7F,$DB,$3D,$92,$3C,$5A,$23,$11,$F0,$7A,$FF,$DB,$E4,$76,$FF,$B8
	.dc.b	$B8,$FF,$27,$1F,$99,$3F,$1D,$5F,$5B,$83,$F3,$51,$89,$FD,$57,$2F
	.dc.b	$FA,$ED,$B1,$FC,$58,$FD,$51,$4E,$95,$FB,$03,$A7,$2F,$E8,$E3,$57
	.dc.b	$DA,$75,$5D,$27,$F8,$A4,$E6,$FC,$11,$D8,$CD,$F3,$69,$FA,$CB,$E6
	.dc.b	$EE,$BF,$85,$4E,$DB,$1B,$AE,$AF,$A1,$B1,$E1,$8F,$3C,$F1,$3F,$6F
	.dc.b	$5E,$C8,$FE,$23,$7E,$9A,$EB,$1F,$A8,$3F,$A8,$46,$6B,$8B,$F4,$A7
	.dc.b	$B4,$FA,$8E,$0F,$F5,$18,$E1,$1F,$C1,$2B,$C0,$FD,$53,$DD,$5D,$1F
	.dc.b	$EF,$13,$C1,$FE,$13,$3C,$0F,$8A,$FB,$CF,$69,$F0,$CA,$78,$A7,$FA
	.dc.b	$C4,$F3,$9F,$E5,$D3,$93,$E0,$9E,$23,$CD,$1E,$D9,$F9,$11,$83,$CA
	.dc.b	$6C,$F2,$C6,$DF,$11,$F0,$E7,$D1,$7F,$52,$B9,$CF,$03,$99,$EB,$38
	.dc.b	$57,$37,$CC,$7C,$73,$57,$F5,$AF,$AC,$7C,$06,$BC,$8E,$B1,$F5,$8E
	.dc.b	$13,$D6,$77,$18,$9D,$D7,$9A,$7E,$39,$A7,$EB,$47,$E1,$0F,$85,$1A
	.dc.b	$AE,$84,$8E,$67,$C4,$AE,$CF,$13,$D7,$5D,$23,$F1,$11,$D2,$FF,$B1
	.dc.b	$57,$38,$E1,$3E,$BA,$E0,$7D,$E8,$E0,$7F,$80,$67,$84,$F4,$8E,$E7
	.dc.b	$DB,$3F,$98,$47,$F0,$D7,$F8,$AB,$5C,$98,$21,$83,$F8,$3B,$FC,$C6
	.dc.b	$13,$FD,$17,$8F,$FA,$F4,$FA,$AB,$FA,$ED,$69,$FE,$B3,$3A,$7D,$06
	.dc.b	$AB,$97,$27,$F8,$26,$FF,$C0,$9C,$BF,$D6,$2F,$85,$F8,$8E,$07,$4A
	.dc.b	$E0,$FF,$88,$E7,$84,$62,$3F,$8F,$C6,$1F,$E2,$55,$DB,$0F,$ED,$E7
	.dc.b	$F1,$F7,$DD,$87,$92,$76,$75,$9D,$C7,$57,$DD,$1F,$DB,$30,$F3,$C7
	.dc.b	$38,$F3,$99,$C3,$85,$7F,$89,$F0,$F2,$4F,$E8,$B2,$FE,$05,$8F,$8A
	.dc.b	$3D,$C7,$88,$F5,$E1,$E2,$70,$F8,$A5,$5D,$1E,$73,$E9,$3F,$C1,$4E
	.dc.b	$1E,$66,$2B,$ED,$2A,$E9,$5C,$E7,$31,$83,$C5,$3E,$98,$F4,$3E,$F3
	.dc.b	$CB,$7E,$F9,$DD,$F8,$5E,$53,$E1,$3F,$0A,$78,$67,$E9,$3D,$4F,$56
	.dc.b	$3F,$3A,$3E,$5C,$7C,$63,$47,$B2,$34,$F8,$8D,$57,$86,$7C,$91,$E1
	.dc.b	$3F,$1C,$78,$4D,$DF,$C0,$8F,$1C,$F1,$9F,$EB,$98,$7C,$57,$D3,$87
	.dc.b	$B3,$0F,$0C,$79,$23,$18,$FA,$23,$F0,$15,$C5,$F0,$CE,$9F,$41,$A7
	.dc.b	$E4,$E3,$F9,$98,$F4,$47,$E3,$DF,$41,$F4,$4F,$06,$3E,$EC,$3C,$78
	.dc.b	$FA,$67,$4D,$3F,$A9,$61,$E3,$BE,$87,$8E,$79,$9F,$01,$F1,$BF,$53
	.dc.b	$1F,$65,$72,$9C,$CF,$D3,$9C,$E3,$F8,$6A,$FC,$BD,$68,$FF,$08,$5F
	.dc.b	$87,$1E,$6D,$C3,$E1,$D7,$07,$E0,$CE,$AB,$93,$EF,$C3,$BC,$FE,$CA
	.dc.b	$B8,$63,$F0,$B1,$F9,$4D,$9F,$51,$DA,$3C,$D8,$F8,$63,$E0,$9F,$22
	.dc.b	$3F,$08,$FE,$18,$FC,$0E,$1F,$86,$8C,$C6,$CC,$9B,$3E,$FC,$6D,$C9
	.dc.b	$B3,$26,$F1,$FA,$71,$B8,$C9,$B8,$CB,$B8,$F8,$8F,$C6,$3E,$4D,$7A
	.dc.b	$F0,$FC,$03,$6F,$E1,$E3,$F2,$2B,$F1,$75,$F2,$23,$18,$7E,$0E,$3F
	.dc.b	$11,$5F,$1A,$3E,$8D,$7E,$14,$F0,$1F,$22,$79,$3F,$34,$FA,$97,$E1
	.dc.b	$7E,$89,$FE,$96,$EB,$FE,$82,$4C,$3F,$D2,$8F,$FA,$45,$FF,$49,$92
	.dc.b	$3F,$7B,$3F,$E9,$77,$FD,$5F,$5F,$EB,$27,$F3,$F1,$97,$FC,$85,$97
	.dc.b	$37,$FD,$5F,$7D,$DE,$38,$FF,$A9,$1C,$E1,$FE,$A5,$7F,$D4,$04,$30
	.dc.b	$7F,$A5,$F0,$FF,$4E,$BF,$EA,$1E,$6F,$F5,$01,$39,$BF,$D4,$10,$EA
	.dc.b	$76,$F1,$A5,$7F,$B2,$27,$35,$EC,$9C,$BC,$6B,$85,$6D,$F1,$C7,$FB
	.dc.b	$28,$F3,$14,$C9,$FE,$C4,$32,$53,$37,$FA,$F7,$2D,$73,$A7,$FB,$39
	.dc.b	$CB,$5C,$EB,$FE,$C3,$69,$FE,$C6,$21,$C1,$D3,$E5,$21,$E3,$2C,$F8
	.dc.b	$5B,$3D,$DB,$3C,$9E,$04,$9E,$0D,$9E,$0D,$9E,$13,$A2,$4E,$9B,$1A
	.dc.b	$8C,$11,$C1,$27,$DA,$7C,$03,$DC,$60,$F3,$94,$F1,$16,$7B,$B6,$7B
	.dc.b	$36,$7B,$3D,$08,$77,$21,$D8,$C9,$4C,$9E,$03,$25,$33,$3E,$27,$2D
	.dc.b	$73,$3E,$47,$2D,$73,$3E,$66,$9E,$92,$1C,$1D,$3E,$52,$1E,$32,$CF
	.dc.b	$85,$B3,$DD,$B3,$C9,$E0,$49,$E0,$D9,$E0,$D9,$E1,$3A,$24,$E9,$B1
	.dc.b	$A8,$C1,$1C,$12,$7D,$A7,$C0,$3D,$C6,$0F,$39,$4F,$11,$67,$BB,$67
	.dc.b	$B3,$67,$B3,$D0,$87,$72,$1D,$8C,$94,$C9,$E0,$32,$53,$33,$E2,$72
	.dc.b	$D7,$33,$E4,$72,$D7,$33,$E6,$69,$E9,$21,$C1,$D3,$E5,$21,$E3,$2C
	.dc.b	$F8,$5B,$3D,$DB,$3C,$9E,$04,$9E,$0D,$9E,$0D,$9E,$13,$A2,$4E,$9B
	.dc.b	$1A,$8C,$11,$C1,$27,$DA,$7C,$03,$DC,$60,$F3,$94,$F1,$16,$7B,$B6
	.dc.b	$7B,$36,$7B,$3D,$08,$77,$21,$D8,$C9,$4C,$9E,$03,$25,$33,$3E,$27
	.dc.b	$2D,$73,$3E,$47,$2D,$73,$3E,$66,$9E,$92,$1C,$1D,$3E,$52,$1E,$32
	.dc.b	$CF,$85,$B3,$DD,$B3,$C9,$E0,$49,$E0,$D9,$E0,$D9,$E1,$3A,$24,$E9
	.dc.b	$B1,$A8,$C1,$1C,$12,$7D,$A7,$C0,$3D,$C6,$0F,$39,$4F,$11,$67,$F2
	.dc.b	$58,$7F,$81,$E3,$FC,$15,$87,$F8,$21,$1F,$E0,$03,$24,$32,$7F,$81
	.dc.b	$DE,$0E,$CC,$89,$FF,$0B,$B8,$6B,$B2,$1C,$CE,$C6,$0E,$53,$E7,$79
	.dc.b	$35,$D4,$74,$79,$1A,$9F,$2B,$83,$73,$C5,$C1,$0E,$F1,$C1,$A7,$90
	.dc.b	$D4,$F7,$72,$77,$34,$60,$EF,$1C,$E3,$81,$27,$C5,$3B,$3A,$B8,$8D
	.dc.b	$4F,$18,$C1,$1C,$12,$7C,$33,$C1,$B3,$B6,$CE,$DF,$41,$0F,$39,$A2
	.dc.b	$19,$3C,$AF,$17,$66,$4F,$51,$27,$C4,$E5,$AF,$02,$1D,$0E,$E6,$0E
	.dc.b	$53,$E9,$79,$35,$D4,$74,$79,$1A,$9F,$33,$83,$73,$C5,$C1,$0E,$F1
	.dc.b	$C1,$A7,$90,$D4,$F7,$72,$77,$34,$60,$EF,$1C,$E3,$81,$27,$C5,$3B
	.dc.b	$3A,$B8,$8D,$4F,$18,$C1,$1C,$12,$7C,$33,$C1,$B3,$B6,$CE,$DF,$41
	.dc.b	$0F,$39,$A2,$19,$3C,$AF,$17,$66,$4F,$51,$27,$C4,$E5,$AF,$02,$1D
	.dc.b	$0E,$E6,$0E,$53,$E9,$79,$35,$D4,$74,$79,$1A,$9F,$33,$83,$73,$C5
	.dc.b	$C1,$0E,$F1,$C1,$A7,$90,$D4,$F7,$72,$77,$34,$60,$EF,$1C,$E3,$81
	.dc.b	$27,$C5,$3B,$3A,$B8,$8D,$4F,$18,$C1,$1C,$12,$7C,$33,$C1,$B3,$B6
	.dc.b	$CE,$DF,$41,$0F,$39,$A2,$19,$3C,$AF,$17,$66,$4F,$51,$27,$C4,$E5
	.dc.b	$AF,$02,$1D,$0E,$E6,$0E,$53,$E9,$79,$35,$D4,$74,$79,$1A,$9F,$33
	.dc.b	$83,$73,$C5,$C1,$0E,$F1,$C1,$A7,$90,$D4,$F7,$72,$77,$34,$60,$EF
	.dc.b	$1C,$E3,$81,$27,$C5,$3B,$3A,$B8,$8D,$4F,$18,$C1,$1C,$12,$7C,$33
	.dc.b	$C1,$B3,$B6,$CE,$DF,$41,$0F,$39,$A3,$F9,$9F,$E5,$22,$00

;--------------------------------------------------------------------------------
;	FASIN.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fasin_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FASIN.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FASIN.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fasin_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1687+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fasin.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FASIN.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1687+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fasin.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FASIN.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fasin_data_compressed::
	.dc.b	$00,$02,$93,$04,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$FA,$77,$E2,$7F,$C2
	.dc.b	$CF,$C1,$F0,$9F,$C1,$BF,$D7,$5F,$EA,$2D,$F3,$38,$F9,$9C,$BC,$CE
	.dc.b	$7E,$67,$4F,$33,$AF,$99,$DB,$CC,$EF,$E6,$78,$3C,$CF,$0F,$99,$E2
	.dc.b	$F3,$3C,$7E,$67,$F6,$1F,$C5,$1F,$2F,$7E,$2B,$E7,$FA,$7F,$D9,$7E
	.dc.b	$6D,$F2,$9B,$F5,$71,$F0,$F9,$F9,$5D,$7C,$9E,$FE,$3B,$87,$C5,$F1
	.dc.b	$F7,$F9,$7B,$7C,$FD,$BE,$9D,$9D,$7B,$3B,$76,$77,$EC,$F0,$76,$78
	.dc.b	$7B,$3C,$5D,$9E,$3E,$CF,$27,$67,$97,$B3,$CD,$D9,$E7,$EC,$FE,$F9
	.dc.b	$C9,$EC,$ED,$FF,$04,$3F,$EC,$4E,$DF,$F1,$AB,$FE,$7F,$2F,$C5,$F0
	.dc.b	$F8,$8E,$3F,$01,$CB,$DD,$E7,$EE,$F4,$F7,$7A,$FB,$BD,$BD,$DE,$FE
	.dc.b	$EF,$83,$DD,$F0,$FB,$BE,$2F,$77,$C7,$EE,$F9,$3D,$DF,$2F,$BB,$F4
	.dc.b	$78,$3C,$FE,$EF,$A3,$DD,$FF,$0A,$7C,$E7,$F6,$D1,$F1,$38,$F8,$9C
	.dc.b	$FC,$4E,$BE,$27,$7F,$13,$C3,$E2,$78,$FC,$37,$2F,$84,$E7,$F0,$9D
	.dc.b	$3E,$13,$AF,$C2,$76,$F8,$4E,$FF,$09,$E0,$F8,$4F,$0F,$C2,$78,$BE
	.dc.b	$13,$C7,$F0,$9E,$4F,$84,$F2,$FC,$27,$9B,$E1,$3C,$FF,$03,$FE,$2B
	.dc.b	$F9,$DF,$C3,$F9,$9F,$E3,$75,$EF,$E7,$CC,$D7,$BF,$BF,$7F,$87,$BF
	.dc.b	$C7,$DF,$E5,$EF,$F3,$F7,$FF,$2D,$E9,$75,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$47,$BF,$FE,$43,$F9,$EF,$C4,$74,$BE,$03,$F9,$0A,$87
	.dc.b	$C2,$6F,$E0,$B8,$7C,$9F,$1F,$5B,$97,$2F,$3F,$83,$E9,$F0,$7D,$7E
	.dc.b	$0F,$B7,$C1,$F7,$F8,$3F,$07,$C1,$F8,$7E,$0F,$C5,$F0,$7E,$3F,$83
	.dc.b	$F2,$7C,$1F,$97,$E0,$FC,$DF,$07,$E7,$F8,$3F,$DD,$6B,$FB,$BD,$7F
	.dc.b	$79,$AF,$EF,$75,$FD,$F6,$BE,$DF,$83,$F7,$7C,$1F,$F9,$87,$E0,$3F
	.dc.b	$C7,$47,$E3,$BA,$3F,$CD,$CB,$E0,$74,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$FF,$0D,$E8,$ED,$F0,$3D,$FE,$07,$C1
	.dc.b	$F0,$3F,$C4,$7A,$FF,$89,$75,$FF,$13,$EB,$FE,$29,$D7,$FC,$57,$AF
	.dc.b	$F8,$B7,$5F,$F1,$7E,$6F,$E3,$1C,$DF,$E7,$9F,$C5,$7F,$CA,$47,$A7
	.dc.b	$9C,$FA,$5F,$F4,$43,$FE,$E4,$3F,$D3,$C7,$F8,$38,$FE,$8C,$7F,$45
	.dc.b	$78,$F0,$F5,$E1,$ED,$C1,$DF,$83,$E4,$7C,$17,$C9,$F8,$2F,$95,$F0
	.dc.b	$5F,$2F,$E0,$BE,$67,$C1,$7C,$DF,$82,$F9,$DF,$05,$F3,$FE,$0B,$E8
	.dc.b	$7C,$17,$FA,$47,$E8,$3F,$4F,$C0,$7F,$4E,$FF,$A7,$D7,$C0,$EF,$E0
	.dc.b	$78,$7C,$0F,$1F,$81,$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF
	.dc.b	$03,$DF,$E0,$7C,$1F,$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0
	.dc.b	$3F,$37,$C9,$F9,$DD,$9F,$3F,$B3,$E8,$76,$7D,$1E,$CF,$A5,$D9,$F4
	.dc.b	$FB,$3E,$A7,$67,$D5,$EC,$FA,$DD,$9F,$EA,$FF,$4F,$DE,$7B,$BB,$4F
	.dc.b	$B9,$FF,$5F,$2E,$BE,$1D,$7C,$7A,$F9,$75,$F3,$EB,$E9,$D7,$D7,$AF
	.dc.b	$E6,$7C,$0F,$CD,$F8,$1F,$9D,$F0,$3E,$1E,$EF,$A1,$F0,$3F,$EC,$1F
	.dc.b	$A1,$FD,$F9,$FB,$CE,$0F,$F6,$72,$F7,$F5,$EF,$EF,$DF,$E1,$EF,$F1
	.dc.b	$F7,$F9,$7B,$FC,$FD,$FE,$9C,$FD,$7D,$FF,$81,$E4,$7C,$1F,$23,$E1
	.dc.b	$79,$1F,$0F,$C8,$F8,$9E,$47,$8F,$DF,$F2,$7B,$3E,$5F,$67,$CD,$EC
	.dc.b	$F9,$FD,$9F,$47,$B3,$E9,$F6,$7D,$5E,$CF,$AF,$D9,$F6,$7B,$3F,$ED
	.dc.b	$DF,$A2,$FF,$5B,$1F,$A5,$3F,$49,$D8,$5E,$FE,$FD,$FE,$1E,$FF,$1F
	.dc.b	$7F,$97,$BF,$F6,$7D,$8E,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B
	.dc.b	$9E,$1F,$73,$C5,$EE,$78,$FD,$CF,$27,$B9,$E5,$F7,$3C,$DE,$E7,$9F
	.dc.b	$DC,$F4,$7B,$9F,$87,$F5,$3D,$5E,$E7,$FB,$DF,$CF,$FD,$39,$F8,$93
	.dc.b	$F1,$1D,$27,$F1,$08,$C7,$B7,$C7,$DB,$E5,$ED,$F3,$F6,$FA,$7B,$7D
	.dc.b	$7D,$BE,$DE,$CF,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC
	.dc.b	$9D,$FE,$5E,$2F,$37,$B7,$E7,$F6,$FD,$1E,$DF,$8F,$8D,$F4,$7F,$D3
	.dc.b	$8F,$C8,$38,$F6,$F5,$1E,$B8,$F5,$3F,$EB,$D3,$FC,$EA,$7F,$54,$3F
	.dc.b	$CE,$6E,$BD,$7E,$3E,$BF,$2F,$5F,$F6,$19,$74,$F5,$FA,$FA,$FD,$BD
	.dc.b	$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F
	.dc.b	$AF,$EB,$FA,$4F,$67,$09,$FE,$64,$7E,$38,$FC,$6B,$F4,$D7,$AD,$C7
	.dc.b	$D6,$E5,$EB,$7F,$05,$E8,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07
	.dc.b	$AD,$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$79
	.dc.b	$FD,$6F,$47,$AD,$E9,$F5,$BD,$5E,$B7,$D5,$FA,$5F,$E6,$67,$F1,$53
	.dc.b	$F1,$A7,$E3,$39,$CB,$D6,$E1,$EB,$71,$F5,$BF,$8F,$79,$7C,$FD,$6E
	.dc.b	$9E,$B7,$E0,$7F,$16,$FF,$06,$9F,$80,$3F,$87,$65,$F9,$77,$F0,$09
	.dc.b	$FF,$7E,$1F,$EB,$43,$FC,$26,$7E,$01,$D7,$99,$D7,$CC,$ED,$E6,$77
	.dc.b	$F2,$FC,$1E,$5F,$87,$CB,$F1,$79,$7E,$3F,$2F,$C9,$E5,$F9,$7C,$BF
	.dc.b	$85,$EB,$FC,$3F,$5F,$E2,$7A,$FF,$17,$D7,$F8,$DE,$BF,$C7,$F5,$FE
	.dc.b	$47,$AF,$F2,$7D,$7F,$95,$EB,$FC,$BF,$57,$E6,$79,$DF,$37,$AF,$F4
	.dc.b	$5F,$01,$FE,$2E,$3F,$42,$7C,$2E,$9F,$E8,$1E,$CE,$BD,$FD,$FB,$FC
	.dc.b	$3D,$FE,$3E,$FF,$2E,$EE,$7D,$DD,$3B,$BA,$F7,$76,$EE,$EF,$DD,$E0
	.dc.b	$EE,$F0,$F7,$78,$BB,$BC,$7D,$DE,$4F,$6F,$CB,$ED,$F9,$BD,$BF,$DC
	.dc.b	$7D,$3F,$E7,$8F,$F0,$71,$FC,$A8,$FC,$EF,$29,$FC,$EB,$FC,$45,$7B
	.dc.b	$5C,$BD,$AE,$7E,$D7,$4F,$6B,$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF
	.dc.b	$0F,$B5,$E2,$F6,$BC,$7E,$D7,$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED
	.dc.b	$7A,$3D,$AF,$4F,$B5,$FC,$47,$EA,$3F,$C1,$E7,$F7,$13,$F9,$49,$F9
	.dc.b	$D3,$F3,$9C,$BF,$CA,$57,$B3,$C7,$D9,$E5,$EC,$F3,$F6,$7A,$7B,$1D
	.dc.b	$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C
	.dc.b	$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$F9,$47,$D4,$FF,$9D,$0F,$E4
	.dc.b	$E7,$E1,$CF,$D3,$8F,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB
	.dc.b	$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F
	.dc.b	$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6,$F5,$FF,$9F,$7B,$BF,$C9,$0F,$F2
	.dc.b	$A1,$FD,$C0,$FC,$39,$F8,$6F,$1B,$F9,$1F,$AF,$C3,$D7,$E3,$EB,$F2
	.dc.b	$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5,$FC,$3E
	.dc.b	$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F,$AF,$FD
	.dc.b	$57,$EA,$BF,$CA,$67,$F8,$D4,$FE,$DA,$7F,$21,$3F,$0B,$C8,$7F,$0A
	.dc.b	$FF,$6F,$5E,$B7,$3F,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E
	.dc.b	$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7
	.dc.b	$F5,$BD,$1E,$B7,$F6,$DF,$AB,$FF,$5C,$9F,$DB,$0F,$E1,$87,$F2,$C3
	.dc.b	$F0,$DC,$7F,$E0,$55,$EA,$F1,$F5,$79,$7A,$BC,$FD,$5E,$9E,$AF,$5F
	.dc.b	$57,$B7,$AB,$DF,$D5,$F0,$7A,$BE,$1F,$57,$C5,$EA,$F8,$FD,$5F,$27
	.dc.b	$AB,$E5,$F5,$7C,$DE,$AF,$9F,$D5,$FF,$02,$7D,$67,$F6,$A3,$FD,$34
	.dc.b	$7F,$8D,$8F,$E1,$A7,$E1,$CF,$E1,$7C,$9F,$E1,$D5,$EA,$72,$F5,$39
	.dc.b	$FA,$9D,$3D,$4E,$BE,$A7,$6F,$53,$BF,$A9,$E0,$F5,$3C,$3E,$A7,$8B
	.dc.b	$D4,$F1,$FA,$9E,$4F,$53,$CB,$EA,$79,$BD,$4F,$3F,$A9,$E8,$F5,$3F
	.dc.b	$C3,$9E,$BF,$FA,$68,$FF,$3E,$1F,$E3,$43,$FB,$41,$F8,$83,$F0,$FC
	.dc.b	$9F,$B7,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7,$4F,$53,$AF,$A9,$DB,$D4
	.dc.b	$EF,$EA,$78,$3D,$4F,$0F,$A9,$E2,$F5,$3C,$7E,$A7,$93,$D4,$F2,$FA
	.dc.b	$9E,$6F,$53,$FC,$65,$F5,$BF,$EF,$E3,$FC,$62,$7F,$51,$3F,$B8,$9F
	.dc.b	$86,$E2,$3F,$86,$7F,$C9,$8B,$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD
	.dc.b	$3E,$FE,$9F,$83,$D3,$F0,$FA,$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F
	.dc.b	$A7,$E6,$F4,$FC,$FE,$9F,$A3,$D3,$FF,$25,$7D,$77,$F8,$C0,$FF,$70
	.dc.b	$1F,$E7,$C3,$FA,$99,$FC,$34,$FC,$3F,$19,$FC,$3B,$FE,$65,$7F,$DC
	.dc.b	$03,$D1,$EB,$E8,$F6,$F4,$7B,$FA,$3E,$0F,$47,$C3,$E8,$F8,$BD,$1F
	.dc.b	$1F,$A3,$E4,$F4,$7C,$BE,$8F,$9B,$D1,$F3,$FA,$3E,$8F,$47,$D3,$E8
	.dc.b	$FF,$98,$FE,$BF,$FD,$C2,$7F,$B0,$CF,$F3,$D9,$FE,$2F,$3F,$87,$9F
	.dc.b	$88,$3F,$86,$F1,$7F,$9F,$57,$A1,$CF,$D0,$E9,$E8,$75,$F4,$3B,$7A
	.dc.b	$1D,$FD,$0F,$07,$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F
	.dc.b	$43,$CD,$E8,$79,$FD,$0F,$F3,$D7,$C1,$3E,$5E,$FC,$7E,$3E,$23,$3F
	.dc.b	$07,$AF,$5F,$7E,$DF,$0F,$5B,$8F,$AD,$CB,$D6,$E7,$EB,$74,$F5,$BA
	.dc.b	$FA,$DD,$BD,$6E,$FE,$B7,$83,$D6,$F0,$FA,$DE,$2F,$5B,$C7,$EB,$79
	.dc.b	$3D,$6F,$2F,$AD,$E6,$F5,$BF,$D2,$1F,$61,$FE,$76,$1F,$2B,$8F,$91
	.dc.b	$CF,$C6,$6B,$E0,$F7,$F0,$7C,$3E,$03,$8F,$C0,$72,$F7,$79,$FB,$BD
	.dc.b	$3D,$DE,$BE,$EF,$6F,$77,$BF,$BB,$E0,$F7,$7C,$3E,$EF,$8B,$DD,$F1
	.dc.b	$FB,$BE,$4F,$77,$CB,$EE,$F9,$BD,$DF,$3F,$BB,$FE,$A7,$FB,$17,$CC
	.dc.b	$5F,$96,$C7,$C9,$67,$E3,$35,$E9,$6F,$E2,$B8,$7C,$37,$1F,$82,$E5
	.dc.b	$F0,$5C,$FE,$0B,$A7,$C1,$75,$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05
	.dc.b	$E1,$F8,$2F,$17,$C1,$78,$FE,$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17
	.dc.b	$FA,$F7,$EC,$BF,$09,$F0,$BF,$ED,$17,$FD,$7A,$7F,$97,$4F,$F3,$E9
	.dc.b	$FC,$34,$FE,$18,$9F,$E1,$8F,$1F,$77,$A7,$BB,$D7,$DD,$ED,$EE,$F7
	.dc.b	$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE,$3F,$77,$C9,$EE,$F9,$7D,$DF
	.dc.b	$37,$BB,$E7,$F7,$7D,$1E,$EF,$FB,$57,$D8,$FF,$5D,$9F,$86,$E1,$FF
	.dc.b	$75,$2F,$6F,$5E,$DE,$FD,$BE,$13,$C7,$DB,$E5,$ED,$F3,$F6,$FA,$7B
	.dc.b	$7D,$7D,$BE,$DE,$DF,$7F,$6F,$C1,$ED,$F8,$7D,$BF,$17,$B7,$E3,$F6
	.dc.b	$FC,$9E,$DF,$97,$DB,$F3,$7B,$7E,$7F,$6F,$D1,$ED,$FF,$BB,$7E,$CF
	.dc.b	$F0,$FC,$FF,$F0,$05,$7B,$59,$F6,$B5,$ED,$6F,$DA,$E1,$ED,$71,$F6
	.dc.b	$B9,$7B,$5C,$FD,$AE,$9E,$D7,$5F,$6B,$B7,$B5,$DF,$DA,$F0,$7B,$5E
	.dc.b	$1F,$6B,$C5,$ED,$78,$FD,$AF,$27,$B5,$E5,$F6,$BC,$DE,$D7,$9F,$DA
	.dc.b	$73,$CF,$B4,$FC,$4A,$5F,$0B,$A8,$C7,$C2,$D9,$FF,$63,$9F,$D5,$CF
	.dc.b	$C5,$9F,$8A,$8E,$5E,$E6,$FD,$CE,$1E,$E7,$1F,$73,$97,$B9,$CF,$DC
	.dc.b	$E9,$EE,$75,$F7,$3B,$7B,$9D,$FD,$CF,$07,$B9,$E1,$F7,$3C,$5E,$E7
	.dc.b	$8F,$DC,$F3,$7D,$AF,$9F,$84,$FE,$2D,$BE,$EE,$BD,$DD,$FB,$BC,$3D
	.dc.b	$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B,$FB,$BE
	.dc.b	$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF,$9B,$DD
	.dc.b	$F3,$FB,$BE,$8F,$77,$E7,$7C,$27,$E2,$B9,$FF,$C8,$BE,$EE,$3D,$DC
	.dc.b	$FB,$BA,$F7,$77,$EE,$F0,$F7,$78,$FB,$BC,$BD,$DE,$7E,$EF,$4F,$77
	.dc.b	$AF,$BB,$DB,$DD,$EF,$EE,$F8,$3D,$DF,$0F,$BB,$E2,$F7,$7C,$7E,$EF
	.dc.b	$93,$DD,$F2,$FB,$BE,$6F,$77,$F0,$9F,$6D,$F8,$83,$F0,$FC,$FF,$8C
	.dc.b	$7F,$DD,$83,$E1,$F7,$F0,$FC,$3A,$B8,$FC,$3F,$2F,$7B,$9F,$BD,$D3
	.dc.b	$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$7B,$C3,$EF,$78,$BD,$EF,$1F
	.dc.b	$BD,$E4,$F7,$BC,$BE,$F7,$9B,$DE,$F3,$FB,$DF,$9B,$FB,$7F,$C3,$1F
	.dc.b	$85,$3F,$87,$73,$7E,$91,$FF,$76,$9F,$EA,$73,$FC,$36,$7F,$86,$92
	.dc.b	$F6,$7A,$7B,$3D,$7D,$9E,$DE,$CF,$7F,$67,$C1,$EC,$F8,$7D,$9F,$17
	.dc.b	$B3,$E3,$F6,$7C,$9E,$CF,$97,$D9,$F3,$7B,$3E,$7F,$67,$D1,$EC,$FA
	.dc.b	$7D,$9F,$DA,$7D,$C7,$E1,$CF,$E1,$FC,$7F,$BC,$5E,$CE,$BD,$9D,$FB
	.dc.b	$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B
	.dc.b	$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF
	.dc.b	$9B,$D9,$F3,$FB,$3F,$C2,$F8,$7F,$87,$9F,$88,$E6,$FF,$7F,$FC,$16
	.dc.b	$7D,$9D,$7B,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3
	.dc.b	$D7,$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67
	.dc.b	$C9,$EC,$F9,$7D,$9F,$37,$B3,$FC,$87,$EE,$7F,$86,$1F,$C2,$CF,$C3
	.dc.b	$F3,$17,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE,$FF,$39,$FB,$AF
	.dc.b	$E1,$87,$E1,$B9,$CF,$E1,$9F,$E8,$6B,$DC,$E1,$EE,$71,$F3,$79,$7B
	.dc.b	$9C,$FD,$CE,$9E,$E7,$5F,$73,$B7,$B9,$DF,$DC,$F0,$7B,$9E,$1F,$73
	.dc.b	$C5,$EE,$78,$FD,$CF,$27,$B9,$E5,$F7,$3C,$DE,$E7,$9F,$DC,$F4,$7B
	.dc.b	$9F,$D4,$7E,$EF,$FA,$A9,$F8,$71,$EE,$E7,$DD,$D7,$BB,$BF,$77,$87
	.dc.b	$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B,$BD,$7D,$DE,$DE,$EF,$7F,$77
	.dc.b	$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7,$7C,$9E,$EF,$97,$DD,$F3,$7B
	.dc.b	$BF,$D9,$FD,$9F,$EA,$47,$F5,$13,$F0,$E7,$E1,$BC,$AF,$EE,$2B,$DB
	.dc.b	$E1,$ED,$F1,$8E,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$37
	.dc.b	$B7,$E7,$F6,$FD,$1E,$DF,$F7,$FF,$BC,$FE,$A4,$7E,$1C,$FC,$37,$29
	.dc.b	$FC,$33,$FE,$0A,$5E,$D7,$1F,$6B,$97,$B5,$CF,$DA,$E9,$ED,$75,$F6
	.dc.b	$BB,$7B,$5D,$FD,$AF,$07,$B5,$E1,$F6,$BC,$5E,$D7,$8F,$DA,$F2,$7B
	.dc.b	$5E,$5F,$6B,$CD,$ED,$79,$FD,$AF,$47,$B5,$FE,$17,$FB,$DF,$F0,$D9
	.dc.b	$FC,$34,$FC,$3F,$2F,$F8,$81,$FE,$A8,$DF,$63,$87,$B1,$C7,$D8,$E5
	.dc.b	$EC,$73,$F6,$3A,$7B,$0F,$E0,$18,$5C,$7A,$43,$EF,$BF,$C5,$BF,$04
	.dc.b	$E7,$E3,$18,$6D,$7F,$36,$6C,$EE,$39,$CE,$DA,$6C,$B1,$B5,$AA,$FC
	.dc.b	$51,$B3,$2E,$27,$84,$EE,$BB,$56,$9B,$19,$8E,$4D,$9E,$35,$CE,$FD
	.dc.b	$24,$AF,$2B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD,$F8,$A7,$73,$9A,$E9
	.dc.b	$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57,$72,$47,$9D,$B7,$E4
	.dc.b	$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C,$73,$B6,$9B,$2C,$6D
	.dc.b	$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E,$EB,$C1,$5A,$6C,$66
	.dc.b	$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C,$A3,$83,$A2,$1A,$9E
	.dc.b	$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F,$46,$BC,$4C,$46,$C9
	.dc.b	$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C,$F0,$6C,$ED,$B3,$B8
	.dc.b	$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD,$71,$C4,$D3,$89,$E5
	.dc.b	$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF,$D6,$4A,$F3,$B5,$E8
	.dc.b	$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE,$97,$E3,$6C,$EA,$31
	.dc.b	$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E,$48,$C1,$1C,$12,$FC
	.dc.b	$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6,$DF,$51,$2F,$DA,$4B
	.dc.b	$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6,$63,$93,$67,$8D,$73
	.dc.b	$BF,$59,$2B,$CE,$D7,$A1,$CA,$38,$3A,$21,$A9,$EB,$7E,$29,$DC,$E6
	.dc.b	$BA,$5F,$8D,$B3,$A8,$C4,$F4,$6B,$C4,$C4,$6C,$95,$DC,$91,$E7,$6D
	.dc.b	$F9,$23,$04,$70,$4B,$F1,$CF,$06,$CE,$DB,$3B,$8F,$1C,$ED,$A6,$CB
	.dc.b	$1B,$7D,$44,$BF,$69,$2F,$D7,$1C,$4D,$38,$9E,$53,$BA,$F0,$56,$9B
	.dc.b	$19,$8E,$4D,$9E,$35,$CE,$FD,$64,$AF,$3B,$5E,$87,$28,$E0,$E8,$86
	.dc.b	$A7,$AD,$F8,$A7,$73,$9A,$E9,$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11
	.dc.b	$B2,$57,$72,$47,$9D,$B7,$E4,$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C
	.dc.b	$EE,$3C,$73,$B6,$9B,$2C,$6D,$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2
	.dc.b	$79,$4E,$EB,$C1,$5A,$6C,$66,$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED
	.dc.b	$7A,$1C,$A3,$83,$A2,$1A,$9E,$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A
	.dc.b	$8C,$4F,$46,$BC,$4C,$46,$C9,$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04
	.dc.b	$BF,$1C,$F0,$6C,$ED,$B3,$B8,$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6
	.dc.b	$92,$FD,$71,$C4,$D3,$89,$E5,$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3
	.dc.b	$5C,$EF,$D6,$4A,$F3,$B5,$E8,$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77
	.dc.b	$39,$AE,$97,$E3,$6C,$EA,$31,$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79
	.dc.b	$DB,$7E,$48,$C1,$1C,$12,$FC,$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69
	.dc.b	$B2,$C6,$DF,$51,$2F,$DA,$4B,$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15
	.dc.b	$A6,$C6,$63,$93,$67,$8D,$73,$BF,$59,$2B,$CE,$D7,$A1,$CA,$38,$3A
	.dc.b	$21,$A9,$EB,$7E,$29,$DC,$E6,$BA,$5F,$8D,$B3,$A8,$C4,$F4,$6B,$C4
	.dc.b	$C4,$6C,$95,$DC,$91,$E7,$6D,$F9,$23,$04,$70,$4B,$F1,$CF,$06,$CE
	.dc.b	$DB,$3B,$8F,$1C,$ED,$A6,$CB,$1B,$7D,$44,$BF,$69,$2F,$D7,$1C,$4D
	.dc.b	$38,$9E,$53,$BA,$F0,$56,$9B,$19,$8E,$4D,$9E,$35,$CE,$FD,$64,$AF
	.dc.b	$3B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD,$F8,$A7,$73,$9A,$E9,$7E,$36
	.dc.b	$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57,$72,$47,$9D,$B7,$E4,$8C,$11
	.dc.b	$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C,$73,$B6,$9B,$2C,$6D,$F5,$12
	.dc.b	$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E,$EB,$C1,$5A,$6C,$66,$39,$36
	.dc.b	$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C,$A3,$83,$A2,$1A,$9E,$B7,$E2
	.dc.b	$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F,$46,$BC,$4C,$46,$C9,$5D,$C9
	.dc.b	$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C,$F0,$6C,$ED,$B3,$B8,$F1,$CE
	.dc.b	$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD,$71,$C4,$D3,$89,$E5,$3B,$AF
	.dc.b	$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF,$D6,$4A,$F3,$B5,$E8,$72,$8E
	.dc.b	$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE,$97,$E3,$6C,$EA,$31,$3D,$1A
	.dc.b	$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E,$48,$C1,$1C,$12,$FC,$73,$C1
	.dc.b	$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6,$DF,$51,$2F,$DA,$4B,$F5,$C7
	.dc.b	$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6,$63,$93,$67,$8D,$73,$BF,$59
	.dc.b	$2B,$CE,$D7,$A1,$CA,$38,$3A,$21,$A9,$EB,$7E,$29,$DC,$E6,$BA,$5F
	.dc.b	$8D,$B3,$A8,$C4,$F4,$6B,$C4,$C4,$6C,$95,$DC,$91,$E7,$6D,$F9,$23
	.dc.b	$04,$70,$4B,$F1,$CF,$06,$CE,$DB,$3B,$8F,$1C,$ED,$A6,$CB,$1B,$7D
	.dc.b	$44,$BF,$69,$2F,$D7,$1C,$4D,$38,$9E,$53,$BA,$F0,$56,$9B,$19,$8E
	.dc.b	$4D,$9E,$35,$CE,$FD,$64,$AF,$3B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD
	.dc.b	$F8,$A7,$73,$9A,$E9,$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57
	.dc.b	$72,$47,$9D,$B7,$E4,$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C
	.dc.b	$73,$B6,$9B,$2C,$6D,$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E
	.dc.b	$EB,$C1,$5A,$6C,$66,$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C
	.dc.b	$A3,$83,$A2,$1A,$9E,$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F
	.dc.b	$46,$BC,$4C,$46,$C9,$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C
	.dc.b	$F0,$6C,$ED,$B3,$B8,$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD
	.dc.b	$71,$C4,$D3,$89,$E5,$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF
	.dc.b	$D6,$4A,$F3,$B5,$E8,$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE
	.dc.b	$97,$E3,$6C,$EA,$31,$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E
	.dc.b	$48,$C1,$1C,$12,$FC,$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6
	.dc.b	$DF,$51,$2F,$DA,$4B,$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6
	.dc.b	$63,$93,$67,$8D,$73,$BF,$59,$28,$29,$2B,$CD,$AF,$77,$BA,$D2,$9D
	.dc.b	$3D,$3F,$74,$79,$6F,$8E,$EF,$F7,$F6,$B5,$E6,$D8,$C4,$2F,$8D,$70
	.dc.b	$3C,$2A,$B8,$46,$3C,$DB,$5E,$1F,$36,$BC,$E2,$DA,$6F,$9B,$2F,$B5
	.dc.b	$FA,$7C,$DC,$2C,$BC,$B8,$3C,$DE,$7E,$6F,$46,$E1,$CA,$FC,$0D,$AF
	.dc.b	$03,$73,$F1,$E7,$E4,$F3,$7C,$BE,$6D,$AE,$0F,$A3,$CD,$F4,$F9,$B6
	.dc.b	$7A,$3E,$CC,$39,$C7,$53,$C1,$87,$C2,$FD,$D7,$E9,$47,$95,$B9,$FB
	.dc.b	$63,$EE,$C3,$E2,$B8,$47,$1E,$4E,$5F,$09,$63,$93,$E1,$7B,$9E,$1B
	.dc.b	$E8,$7D,$37,$C1,$E7,$F8,$FC,$FF,$23,$63,$C9,$E7,$F9,$CF,$47,$9F
	.dc.b	$63,$A7,$2F,$D2,$D7,$C1,$8F,$63,$F1,$0F,$BF,$CF,$F8,$1E,$7D,$BF
	.dc.b	$80,$FC,$3E,$DF,$89,$E7,$D9,$F8,$8F,$4C,$7B,$47,$80,$FC,$8E,$1F
	.dc.b	$2F,$CF,$F9,$9E,$7D,$9F,$98,$FC,$FF,$3E,$47,$CF,$F3,$EE,$BE,$7E
	.dc.b	$BF,$C1,$B7,$1F,$05,$76,$3E,$B7,$9F,$F6,$4F,$B5,$E7,$D8,$FB,$5E
	.dc.b	$7F,$DD,$3E,$F7,$9F,$63,$EF,$79,$FF,$81,$3F,$0B,$8F,$C7,$EB,$47
	.dc.b	$E9,$BF,$76,$7C,$AD,$31,$F2,$17,$0F,$CE,$4B,$E1,$B8,$1C,$7E,$1A
	.dc.b	$CE,$BD,$4E,$9E,$A7,$56,$C7,$2F,$53,$C1,$EA,$78,$5B,$1D,$BD,$4F
	.dc.b	$21,$E5,$F4,$6C,$78,$BE,$1B,$D1,$F0,$DE,$96,$C7,$9B,$E1,$BD,$87
	.dc.b	$B7,$E1,$AC,$7A,$BE,$1B,$E0,$7C,$37,$C1,$6C,$7B,$BE,$1B,$E2,$7C
	.dc.b	$37,$C5,$6C,$7C,$2F,$86,$F9,$07,$C9,$F8,$6B,$1F,$1B,$E1,$BF,$88
	.dc.b	$FA,$9F,$CB,$A3,$07,$F0,$57,$1D,$9F,$99,$5F,$29,$C3,$E3,$F8,$B6
	.dc.b	$B8,$7C,$5F,$43,$AF,$C4,$D8,$E9,$F1,$3E,$0F,$89,$F0,$B6,$3C,$1F
	.dc.b	$13,$E4,$F8,$9F,$2B,$63,$C9,$F1,$3E,$83,$D3,$F0,$B6,$3D,$1F,$0B
	.dc.b	$EC,$3D,$BE,$F5,$8F,$67,$9B,$F0,$3E,$17,$E0,$B6,$3E,$07,$C2,$FC
	.dc.b	$4F,$84,$F8,$AD,$8F,$89,$F0,$BF,$23,$E1,$7E,$4B,$63,$E4,$7C,$2F
	.dc.b	$CC,$3E,$6F,$35,$8F,$99,$F0,$BF,$D0,$3F,$77,$FE,$3B,$ED,$FF,$4C
	.dc.b	$A4,$E6,$B5,$F0,$DB,$F8,$6B,$1F,$CE,$70,$FE,$B8,$DF,$86,$E9,$F0
	.dc.b	$BD,$5B,$1B,$F8,$6F,$01,$E1,$F8,$6B,$1C,$FE,$1B,$C9,$F0,$DE,$56
	.dc.b	$C7,$7F,$86,$F4,$7C,$37,$A5,$B1,$E3,$F8,$6F,$61,$ED,$F8,$6B,$1E
	.dc.b	$7F,$86,$F8,$07,$C1,$F8,$6B,$1E,$BF,$86,$F8,$9F,$0B,$F1,$5B,$1E
	.dc.b	$FF,$85,$F9,$1F,$0B,$F2,$5B,$1F,$0F,$E1,$7E,$67,$C2,$7C,$D6,$C7
	.dc.b	$F5,$FF,$DE,$3F,$84,$B7,$E1,$7F,$B8,$C3,$E1,$6C,$63,$E1,$33,$F0
	.dc.b	$96,$F9,$3C,$3E,$0F,$8F,$C2,$59,$EC,$F4,$F8,$39,$1E,$1F,$84,$BE
	.dc.b	$BF,$F4,$3D,$1F,$09,$E2,$3C,$7F,$09,$6F,$D1,$F0,$7E,$63,$CF,$F0
	.dc.b	$96,$3D,$9F,$09,$EA,$3D,$7F,$09,$63,$E0,$7C,$1F,$BB,$E1,$3D,$ED
	.dc.b	$8F,$89,$F0,$7F,$0B,$E1,$3E,$1B,$63,$A9,$36,$F4,$61,$D5,$F5,$5F
	.dc.b	$36,$CF,$CF,$C3,$E6,$FC,$24,$8F,$A5,$F0,$96,$BE,$17,$C2,$7D,$23
	.dc.b	$FC,$47,$FB,$CF,$F1,$0F,$4B,$FC,$5F,$C6,$9C,$FC,$24,$C7,$FC,$35
	.dc.b	$6F,$85,$C6,$3F,$DB,$E1,$F0,$BD,$0E,$BF,$0B,$6F,$8F,$C2,$78,$0F
	.dc.b	$0F,$C2,$D8,$EB,$F0,$9E,$4F,$85,$F2,$B6,$3C,$3F,$0B,$E8,$3D,$3F
	.dc.b	$0B,$63,$A9,$34,$F2,$61,$D5,$F3,$5F,$36,$CF,$B3,$0F,$87,$F0,$B2
	.dc.b	$3D,$FF,$0B,$6B,$DD,$F0,$9F,$23,$E1,$7E,$4B,$6B,$E2,$FC,$27,$CC
	.dc.b	$3E,$6F,$C2,$D8,$F9,$3F,$09,$F4,$3E,$17,$E8,$B6,$3E,$6D,$F3,$C7
	.dc.b	$BC,$7D,$0C,$79,$47,$CF,$C7,$AC,$F8,$0F,$AB,$F0,$BF,$70,$E9,$3F
	.dc.b	$55,$F3,$C7,$33,$B9,$63,$A3,$CD,$FB,$99,$78,$3A,$BF,$CE,$7C,$CF
	.dc.b	$85,$FC,$4F,$C2,$FE,$29,$B8,$7F,$97,$FA,$FF,$97,$BF,$E8,$67,$FC
	.dc.b	$94,$E4,$FF,$39,$B8,$C7,$FC,$70,$FF,$2E,$6F,$C2,$6B,$E0,$B6,$F6
	.dc.b	$8D,$FC,$1D,$CF,$FD,$0F,$68,$EB,$87,$6F,$85,$AF,$95,$E6,$F9,$F0
	.dc.b	$EC,$D3,$D5,$F0,$9E,$63,$CF,$F0,$96,$3D,$DF,$07,$EA,$3D,$7F,$09
	.dc.b	$63,$E1,$7C,$1F,$BB,$E1,$3D,$ED,$8F,$8D,$F0,$9F,$0B,$E1,$3E,$1B
	.dc.b	$63,$A9,$38,$3E,$2E,$1D,$5F,$8F,$7C,$DB,$3F,$47,$0F,$9B,$F0,$92
	.dc.b	$3E,$A7,$C2,$5A,$F9,$FF,$07,$F4,$BE,$13,$E9,$B6,$BE,$D7,$C1,$FD
	.dc.b	$6F,$84,$FA,$ED,$8F,$BD,$F0,$7F,$68,$E9,$CF,$F4,$9E,$71,$F8,$3C
	.dc.b	$2E,$3F,$79,$E7,$1F,$87,$C2,$CF,$E0,$9E,$91,$F8,$BF,$84,$B3,$F8
	.dc.b	$67,$9E,$1F,$80,$BB,$5F,$8F,$79,$C7,$E1,$2E,$CF,$E4,$DF,$F6,$1F
	.dc.b	$EF,$5F,$D5,$B0,$FF,$44,$36,$BF,$DA,$2F,$F5,$7A,$FF,$5D,$39,$7F
	.dc.b	$DC,$2F,$F9,$85,$CF,$C1,$F4,$F8,$3E,$AF,$6F,$83,$93,$CB,$E0,$EE
	.dc.b	$3E,$0F,$83,$F1,$9E,$4F,$84,$B5,$D0,$93,$E6,$C7,$AC,$79,$30,$E7
	.dc.b	$9F,$A3,$1B,$3E,$6F,$84,$F7,$9F,$03,$E1,$2C,$7B,$3E,$13,$E1,$9F
	.dc.b	$13,$E1,$2C,$7C,$0F,$84,$F8,$E7,$C8,$F8,$4B,$1F,$13,$E0,$BE,$5F
	.dc.b	$C1,$7C,$C6,$C7,$52,$6F,$F3,$B0,$EA,$FD,$0C,$39,$BF,$4B,$0B,$AF
	.dc.b	$D2,$F8,$2F,$AF,$F0,$7F,$61,$B8,$FD,$4F,$84,$FB,$67,$DC,$F8,$4B
	.dc.b	$1F,$63,$E1,$3E,$F9,$F8,$0F,$32,$C7,$DC,$C0,$FF,$BD,$FF,$7B,$FE
	.dc.b	$C8,$FD,$E8,$CE,$9F,$EC,$D9,$CA,$1D,$4C,$BC,$31,$F8,$CC,$AE,$9F
	.dc.b	$8C,$CF,$C3,$3F,$90,$CE,$CF,$68,$63,$D9,$A7,$C0,$3C,$47,$4C,$FF
	.dc.b	$27,$8F,$C1,$7E,$21,$F0,$49,$1D,$DB,$5D,$EB,$FD,$1F,$1F,$07,$E1
	.dc.b	$3D,$F3,$9F,$84,$B3,$EF,$BF,$2D,$61,$F8,$57,$FE,$02,$9F,$95,$F0
	.dc.b	$99,$9F,$5F,$C2,$5A,$D4,$7C,$BF,$84,$C1,$9F,$82,$B8,$7D,$3F,$82
	.dc.b	$E0,$71,$CA,$C7,$62,$5F,$D5,$C7,$B6,$5C,$A3,$B4,$FD,$DC,$31,$A7
	.dc.b	$C1,$F8,$2F,$21,$E5,$F8,$4B,$5F,$83,$F8,$4F,$41,$E9,$F8,$4B,$1F
	.dc.b	$87,$F8,$4F,$61,$ED,$F8,$2B,$1F,$8B,$F8,$2F,$80,$7C,$1F,$82,$B1
	.dc.b	$F8,$FF,$81,$F8,$9F,$07,$F1,$5B,$1F,$93,$F8,$4F,$91,$F0,$9F,$25
	.dc.b	$B1,$F9,$7F,$82,$F9,$9F,$07,$F3,$5B,$1F,$95,$E1,$FC,$67,$EF,$9F
	.dc.b	$DD,$39,$3F,$30,$FF,$71,$8C,$CF,$D7,$8C,$3F,$7D,$FE,$E9,$1C,$BE
	.dc.b	$07,$5F,$09,$B7,$BD,$A7,$BF,$C2,$61,$B8,$78,$CE,$87,$7F,$84,$F0
	.dc.b	$34,$D5,$79,$AF,$B4,$62,$18,$B1,$CC,$78,$67,$D1,$F0,$9E,$B3,$D9
	.dc.b	$F0,$96,$39,$5B,$E1,$3E,$07,$C1,$7C,$16,$CE,$BE,$0B,$E2,$1F,$17
	.dc.b	$E0,$AC,$72,$F8,$3F,$90,$7C,$9F,$84,$B1,$DB,$E1,$3E,$67,$C2,$7C
	.dc.b	$D6,$C7,$8B,$E1,$3E,$81,$F4,$7E,$06,$C7,$9B,$E1,$3E,$A7,$C1,$7D
	.dc.b	$56,$C7,$AB,$E1,$3E,$C1,$F6,$7E,$12,$C7,$C7,$7F,$AC,$72,$7F,$17
	.dc.b	$6A,$7F,$AF,$3F,$24,$D7,$C1,$7E,$E6,$3F,$AF,$46,$4E,$1F,$05,$9F
	.dc.b	$84,$D3,$BF,$83,$98,$78,$7E,$0E,$DF,$1C,$1F,$09,$D7,$E1,$3B,$36
	.dc.b	$FD,$1F,$09,$E1,$F8,$4F,$13,$63,$D9,$F0,$5E,$53,$CD,$F0,$56,$3E
	.dc.b	$07,$C1,$7A,$7E,$0F,$D4,$D8,$F8,$9F,$09,$ED,$F8,$4F,$73,$63,$E4
	.dc.b	$7C,$27,$C1,$3E,$17,$C0,$58,$F9,$9F,$07,$F1,$7E,$0F,$E3,$36,$3E
	.dc.b	$87,$C1,$FC,$93,$F9,$17,$EF,$BF,$CE,$F8,$FF,$D5,$D3,$F9,$B8,$FE
	.dc.b	$63,$C9,$F3,$5B,$E6,$FF,$43,$7F,$D6,$13,$BF,$80,$E7,$EF,$F4,$F8
	.dc.b	$0B,$3C,$DE,$FF,$05,$E0,$F8,$1B,$3D,$DF,$E6,$9E,$8B,$9A,$9E,$6F
	.dc.b	$82,$B5,$E5,$F8,$2F,$4F,$C1,$7A,$9B,$1E,$9F,$82,$F6,$FC,$17,$B9
	.dc.b	$B1,$FB,$7C,$5D,$5E,$03,$E1,$FC,$15,$9F,$85,$F0,$5F,$1B,$E0,$BE
	.dc.b	$3B,$63,$E3,$7C,$17,$CA,$F8,$1F,$96,$D8,$ED,$E0,$7B,$0F,$9F,$F0
	.dc.b	$56,$3E,$77,$C0,$FD,$2F,$82,$FA,$6D,$8F,$A5,$F0,$3F,$5B,$E0,$BE
	.dc.b	$BB,$63,$D5,$E2,$7C,$83,$FA,$E7,$EF,$DF,$CA,$FB,$DC,$49,$CF,$C1
	.dc.b	$6A,$2D,$FF,$19,$BF,$E5,$6F,$F6,$B7,$F8,$D3,$9F,$83,$EB,$F0,$7D
	.dc.b	$9F,$AF,$E4,$C9,$E5,$F0,$77,$3F,$DA,$60,$F8,$3F,$2F,$C0,$F9,$9B
	.dc.b	$7E,$1F,$82,$F4,$9E,$AF,$82,$B1,$F5,$FC,$EF,$C2,$9E,$EF,$82,$B1
	.dc.b	$E9,$F8,$2F,$83,$F0,$5F,$09,$B1,$ED,$F8,$1F,$8A,$7C,$6F,$82,$B1
	.dc.b	$F8,$3F,$4F,$F2,$27,$CA,$F8,$2B,$1F,$17,$E0,$BE,$6F,$C0,$FC,$E6
	.dc.b	$C7,$C9,$F8,$2F,$A2,$7D,$2F,$82,$B1,$F8,$FF,$5F,$F3,$A7,$D6,$F5
	.dc.b	$AC,$7F,$49,$FD,$FF,$F9,$09,$FC,$83,$D7,$FF,$18,$38,$F8,$AF,$F0
	.dc.b	$C3,$FC,$82,$33,$F0,$58,$F8,$3C,$BA,$F8,$39,$87,$5F,$83,$BD,$BF
	.dc.b	$C9,$F1,$FE,$3F,$3E,$1F,$83,$EB,$F0,$7D,$9E,$FF,$05,$27,$CD,$F0
	.dc.b	$77,$0F,$E6,$1E,$33,$E0,$BF,$72,$79,$BE,$0E,$DF,$B7,$DC,$F4,$FA
	.dc.b	$FE,$A6,$C7,$C1,$F4,$BD,$A7,$BB,$CD,$B1,$FB,$7F,$83,$FE,$0A,$7C
	.dc.b	$2F,$2E,$C7,$C9,$F1,$7E,$2F,$81,$F1,$9B,$1F,$37,$AF,$E4,$9F,$2B
	.dc.b	$5B,$1F,$DE,$7E,$CF,$FC,$B8,$7F,$97,$79,$7F,$91,$F0,$FF,$4D,$8F
	.dc.b	$F3,$07,$A3,$3E,$0F,$FC,$94,$DB,$FF,$01,$3F,$C9,$2B,$FA,$A3,$7B
	.dc.b	$3F,$CD,$4F,$F9,$39,$BF,$11,$FE,$53,$C3,$F9,$6D,$FF,$96,$DB,$7F
	.dc.b	$E7,$BB,$7C,$1E,$F1,$F0,$FC,$1F,$8B,$E0,$EE,$3E,$27,$CB,$F0,$72
	.dc.b	$3C,$BF,$05,$72,$F6,$69,$E1,$3D,$7F,$07,$6B,$D7,$F0,$5E,$EF,$83
	.dc.b	$F7,$B6,$3D,$FF,$07,$F0,$8F,$87,$F0,$56,$3E,$67,$C2,$7B,$4F,$8F
	.dc.b	$F0,$96,$3E,$3F,$C1,$FC,$AF,$84,$F9,$6D,$8F,$97,$F0,$9F,$38,$F9
	.dc.b	$FF,$07,$63,$EC,$72,$FC,$93,$E9,$FC,$25,$8F,$A7,$F0,$97,$8B,$E9
	.dc.b	$B7,$AB,$E9,$F4,$FD,$A3,$FD,$09,$FC,$01,$FC,$A5,$1F,$57,$A6,$61
	.dc.b	$FE,$3F,$9F,$F2,$9E,$1F,$EA,$FE,$B5,$6A,$3F,$10,$F4,$F2,$31,$8F
	.dc.b	$FA,$AE,$FF,$96,$D7,$F8,$65,$B1,$9F,$83,$F1,$FC,$27,$E1,$7B,$66
	.dc.b	$7E,$1F,$84,$B8,$FE,$56,$6E,$1D,$7E,$0A,$DF,$FA,$97,$2F,$F2,$63
	.dc.b	$E9,$F8,$3F,$7F,$C1,$FC,$07,$F2,$5E,$0C,$9F,$77,$C1,$DC,$FF,$45
	.dc.b	$37,$2F,$57,$C1,$DB,$ED,$F0,$5F,$28,$F9,$7F,$07,$63,$F5,$DE,$2F
	.dc.b	$E7,$CF,$9F,$F0,$76,$32,$4C,$BE,$27,$C1,$E5,$F8,$DF,$07,$F5,$8F
	.dc.b	$AF,$F0,$56,$BF,$75,$E4,$7E,$AC,$FB,$7F,$07,$63,$24,$BF,$A1,$F0
	.dc.b	$79,$7F,$7D,$E4,$FE,$CC,$FC,$1F,$C1,$DA,$FE,$0D,$16,$7E,$C7,$C1
	.dc.b	$DE,$2F,$F7,$37,$7A,$B1,$7F,$ED,$5F,$E0,$1F,$F2,$EC,$E6,$3F,$88
	.dc.b	$79,$BF,$C0,$4F,$F6,$A4,$FF,$41,$8F,$E0,$51,$97,$67,$FA,$3E,$3F
	.dc.b	$A1,$3A,$33,$8F,$F1,$79,$E7,$F0,$5C,$63,$F8,$D7,$9D,$FC,$3C,$EF
	.dc.b	$F0,$56,$FF,$91,$45,$9F,$F3,$66,$0F,$82,$C6,$9E,$EA,$B7,$ED,$F8
	.dc.b	$2B,$87,$E8,$FD,$1F,$E3,$87,$B3,$E0,$AD,$7F,$2F,$8B,$3D,$BE,$0A
	.dc.b	$E3,$F2,$A6,$D7,$C9,$F8,$2B,$7F,$AF,$F4,$FF,$94,$1F,$23,$E0,$AD
	.dc.b	$7F,$3B,$8B,$3E,$AF,$81,$B8,$FD,$69,$B5,$F5,$7E,$0A,$DF,$EE,$FD
	.dc.b	$5F,$E6,$87,$D4,$F8,$2B,$5F,$D1,$E2,$CF,$C6,$F8,$2B,$8F,$E0,$A6
	.dc.b	$D7,$E0,$7E,$06,$DF,$F0,$7F,$67,$FA,$01,$F8,$0F,$80,$B5,$FD,$4E
	.dc.b	$2C,$FD,$2F,$81,$B8,$FE,$3A,$6D,$7E,$37,$E0,$6D,$FF,$13,$F7,$3F
	.dc.b	$A6,$1F,$8C,$F8,$1B,$5F,$D7,$E2,$CF,$DE,$F8,$2B,$8F,$E6,$A6,$D2
	.dc.b	$A7,$CA,$7F,$66,$9F,$EB,$1E,$FC,$AF,$F7,$14,$FF,$58,$A7,$82,$AD
	.dc.b	$47,$F6,$B5,$5C,$3E,$07,$38,$FE,$AA,$6C,$F8,$69,$87,$F6,$98,$FE
	.dc.b	$CD,$3F,$DA,$7D,$F9,$1E,$2F,$80,$B7,$FD,$E6,$6E,$1E,$08,$F5,$FB
	.dc.b	$D8,$BF,$EE,$15,$67,$DF,$ED,$57,$D7,$F0,$1F,$DD,$CE,$58,$73,$8F
	.dc.b	$2F,$AD,$9C,$3F,$C1,$95,$67,$FD,$AD,$A3,$D3,$C6,$1F,$C2,$2A,$DF
	.dc.b	$BE,$9E,$7E,$30,$FE,$9D,$F0,$1F,$E0,$73,$C3,$5B,$F3,$30,$FF,$86
	.dc.b	$E2,$CF,$83,$CC,$B9,$7F,$16,$9B,$5D,$FC,$8B,$7F,$7F,$ED,$7F,$D1
	.dc.b	$47,$FA,$2F,$37,$8D,$E0,$9D,$5D,$55,$C2,$31,$4F,$0F,$53,$99,$FC
	.dc.b	$66,$5F,$26,$3B,$78,$1A,$9D,$FC,$44,$7F,$1B,$DD,$72,$FF,$80,$17
	.dc.b	$FD,$C9,$F0,$D7,$83,$AF,$58,$F5,$8B,$87,$CD,$E9,$B7,$EF,$F8,$2F
	.dc.b	$29,$E8,$E8,$B5,$E6,$3F,$25,$97,$FA,$9A,$B8,$47,$E8,$39,$77,$1E
	.dc.b	$4B,$B9,$7E,$17,$7F,$F0,$5D,$F3,$AE,$27,$AB,$E0,$7E,$29,$F3,$7E
	.dc.b	$53,$F6,$13,$83,$84,$FC,$E9,$C2,$BF,$CE,$57,$ED,$F5,$CC,$7C,$DA
	.dc.b	$B7,$FB,$BA,$FD,$E6,$38,$C3,$F0,$1F,$07,$F5,$8F,$13,$FE,$14,$AF
	.dc.b	$83,$19,$7C,$7A,$23,$66,$67,$EA,$B7,$1F,$2C,$3E,$0F,$17,$F3,$1F
	.dc.b	$C2,$9E,$0F,$83,$B1,$DA,$2C,$FB,$E3,$D4,$FF,$87,$0F,$C1,$C6,$AB
	.dc.b	$F1,$8F,$AB,$E0,$B1,$5F,$6A,$FF,$26,$7B,$BE,$06,$C7,$B2,$2E,$1D
	.dc.b	$C6,$9F,$C8,$FE,$07,$39,$7E,$66,$AD,$F9,$3E,$06,$E1,$FA,$BF,$82
	.dc.b	$FD,$09,$F3,$3E,$0A,$D7,$CA,$8B,$3D,$F2,$FE,$61,$F0,$58,$CB,$F4
	.dc.b	$D5,$6F,$E0,$FC,$15,$C3,$F7,$1B,$7E,$B8,$FB,$3F,$05,$6B,$EB,$C5
	.dc.b	$9E,$C3,$E0,$B1,$8F,$EC,$E6,$D7,$CE,$F8,$2B,$7F,$D0,$FE,$D7,$FC
	.dc.b	$09,$3F,$94,$6C,$74,$F8,$3F,$E0,$91,$8C,$3F,$03,$F0,$7F,$8C,$3F
	.dc.b	$1B,$F0,$77,$0F,$D2,$70,$7C,$EF,$8C,$FD,$12,$6F,$C5,$61,$FE,$17
	.dc.b	$CF,$2B,$8D,$7F,$89,$34,$FE,$35,$39,$7F,$83,$BD,$3E,$17,$19,$FF
	.dc.b	$11,$8B,$7F,$A5,$B7,$C2,$E3,$0F,$E4,$9D,$3F,$C6,$8F,$2F,$C2,$5B
	.dc.b	$FD,$1C,$59,$FB,$57,$FE,$EE,$7F,$8B,$C6,$71,$FE,$40,$FB,$BE,$0F
	.dc.b	$17,$FC,$3B,$AB,$F9,$39,$F0,$BE,$0E,$C7,$EC,$22,$E1,$DB,$E0,$EE
	.dc.b	$3F,$CA,$E6,$D7,$AF,$E0,$ED,$FF,$3D,$ED,$FE,$66,$7C,$EF,$83,$B5
	.dc.b	$FB,$C8,$B3,$EA,$F8,$3B,$8F,$F3,$79,$B5,$F1,$FE,$0E,$DF,$F4,$9F
	.dc.b	$03,$F9,$F9,$F6,$BE,$0E,$D7,$F0,$88,$B3,$F1,$BE,$0E,$E3,$FD,$0E
	.dc.b	$6D,$7D,$3F,$83,$B7,$FD,$57,$C2,$FE,$96,$7E,$1B,$E0,$ED,$7F,$14
	.dc.b	$8B,$3F,$4B,$E0,$EE,$3F,$D3,$E6,$D7,$DF,$F8,$3B,$7F,$D8,$7C,$5F
	.dc.b	$EA,$E7,$E4,$BE,$0E,$D7,$F2,$08,$B3,$FB,$7C,$3F,$9B,$BF,$D5,$23
	.dc.b	$31,$FD,$75,$FE,$78,$FF,$89,$7C,$B4,$7F,$64,$F1,$E5,$FE,$8A,$9F
	.dc.b	$05,$8C,$3F,$AF,$D5,$9F,$E7,$E7,$F8,$8E,$3F,$BC,$46,$6B,$FA,$BB
	.dc.b	$E0,$F8,$2C,$4F,$F7,$DF,$23,$FB,$99,$E4,$F8,$2B,$7F,$CF,$22,$CF
	.dc.b	$B7,$E0,$AE,$3F,$DD,$E6,$D7,$F4,$B8,$7C,$16,$2F,$FC,$15,$E5,$FF
	.dc.b	$80,$0F,$83,$F0,$56,$FF,$A4,$C5,$9F,$95,$F0,$57,$0F,$F0,$24,$DA
	.dc.b	$ED,$F0,$56,$FF,$C2,$DE,$6F,$F8,$30,$F9,$BF,$05,$6B,$FA,$AC,$59
	.dc.b	$FA,$DF,$03,$71,$FF,$08,$4D,$AF,$57,$C0,$DB,$FF,$11,$7A,$1F,$E1
	.dc.b	$83,$EC,$FC,$0D,$AF,$EC,$31,$67,$F0,$5F,$03,$71,$FF,$0E,$4D,$AF
	.dc.b	$8D,$F0,$36,$FF,$CD,$9F,$6B,$FC,$76,$7F,$BE,$B6,$3F,$19,$F0,$5F
	.dc.b	$E2,$D8,$C6,$1F,$E4,$F6,$D7,$F9,$95,$B5,$FD,$5A,$7F,$CE,$13,$D3
	.dc.b	$D4,$E4,$47,$F1,$DF,$03,$6B,$FA,$F0,$F5,$FF,$8C,$5F,$F9,$75,$B5
	.dc.b	$F7,$FE,$07,$F9,$EF,$C2,$FE,$73,$E1,$3F,$C3,$1F,$86,$7F,$9F,$2B
	.dc.b	$F3,$51,$87,$3F,$07,$AF,$82,$B1,$FD,$02,$BF,$C5,$99,$B6,$FF,$71
	.dc.b	$3F,$1B,$DD,$EB,$F0,$7F,$0D,$97,$F2,$1F,$0F,$C1,$E2,$BD,$77,$C9
	.dc.b	$E9,$3E,$6F,$83,$D7,$07,$CB,$F8,$CF,$9C,$7A,$BE,$12,$E5,$F5,$22
	.dc.b	$E1,$D3,$0B,$7C,$6B,$47,$AF,$E1,$32,$FD,$5F,$83,$FB,$47,$F9,$9F
	.dc.b	$F7,$3F,$F1,$14,$7E,$02,$32,$D3,$9E,$1E,$A8,$CB,$63,$E4,$7C,$2E
	.dc.b	$30,$FC,$27,$C2,$7E,$1C,$FA,$7F,$0B,$6B,$F1,$B1,$6F,$FC,$1D,$C4
	.dc.b	$F8,$3C,$61,$CA,$F6,$7D,$5F,$80,$CB,$F8,$FF,$85,$FC,$A1,$EB,$DA
	.dc.b	$C7,$E6,$A2,$E3,$F4,$B2,$B8,$7E,$66,$6D,$7F,$94,$F2,$7B,$D8,$C3
	.dc.b	$F4,$BD,$5F,$A1,$30,$FE,$15,$BC,$AF,$81,$FF,$5E,$7E,$12,$E2,$F3
	.dc.b	$65,$FA,$C9,$CC,$63,$0F,$C2,$FC,$2F,$23,$F5,$C4,$AF,$CA,$7B,$D6
	.dc.b	$7F,$17,$AF,$EC,$F0,$E7,$1F,$BA,$8C,$57,$FB,$9F,$1F,$14,$FE,$FB
	.dc.b	$0E,$5B,$7C,$1A,$FD,$FC,$7E,$F7,$E1,$64,$7E,$8F,$DC,$B7,$F9,$AC
	.dc.b	$BB,$BF,$A5,$6E,$BF,$E2,$DA,$FE,$05,$18,$BE,$38,$FF,$A1,$38,$3F
	.dc.b	$A7,$D7,$F0,$E9,$FE,$19,$F0,$B2,$3F,$D1,$BD,$1F,$E4,$37,$F8,$9C
	.dc.b	$7B,$32,$6B,$EC,$7C,$53,$EF,$E0,$F9,$99,$7F,$15,$CF,$CA,$71,$7F
	.dc.b	$8C,$65,$6B,$CE,$7D,$53,$B9,$6F,$E0,$3D,$5F,$E4,$F5,$F0,$37,$F2
	.dc.b	$63,$E3,$D7,$F0,$57,$C7,$1E,$ED,$3F,$93,$7C,$2F,$F2,$B3,$F6,$5E
	.dc.b	$AD,$8F,$E6,$31,$F3,$0F,$4B,$E1,$CB,$F8,$A4,$66,$36,$7F,$19,$F6
	.dc.b	$79,$4F,$F1,$0F,$85,$FE,$70,$7F,$24,$BB,$5F,$CF,$A3,$EC,$9F,$8A
	.dc.b	$F8,$67,$13,$D5,$CD,$78,$74,$5F,$2C,$7F,$73,$E9,$5D,$7F,$A2,$7C
	.dc.b	$37,$F4,$73,$F9,$9F,$A1,$71,$FE,$9B,$1F,$88,$3A,$DF,$DC,$E4,$FE
	.dc.b	$95,$19,$9D,$9F,$E7,$4F,$3B,$FA,$6D,$7F,$35,$9F,$EA,$9F,$0D,$2B
	.dc.b	$5E,$25,$AF,$E6,$33,$F9,$63,$A3,$DC,$FE,$C2,$E6,$7F,$B0,$BF,$EE
	.dc.b	$1B,$FF,$3D,$4F,$F4,$38,$FE,$C5,$F0,$D2,$BC,$98,$DA,$FE,$7F,$3F
	.dc.b	$A4,$3E,$65,$FF,$9E,$AB,$E5,$46,$5F,$ED,$EF,$7E,$6E,$5A,$7F,$6D
	.dc.b	$F8,$6F,$EE,$02,$BF,$DE,$3F,$BA,$7C,$2E,$14,$EA,$7E,$B3,$D7,$2F
	.dc.b	$71,$FD,$6D,$FD,$B9,$B7,$FD,$7D,$97,$F6,$C7,$FC,$07,$1F,$DF,$FE
	.dc.b	$1E,$57,$F8,$12,$2C,$FF,$5C,$7F,$81,$FD,$C3,$CE,$7C,$7F,$8A,$FF
	.dc.b	$31,$39,$8C,$5F,$F6,$4F,$89,$F1,$6B,$FE,$0B,$7F,$C1,$C4,$BF,$C0
	.dc.b	$78,$76,$BC,$11,$D3,$1E,$F1,$EE,$D7,$E1,$CE,$71,$FC,$C6,$1C,$71
	.dc.b	$F9,$93,$FE,$1E,$8F,$F6,$2E,$32,$3F,$1D,$E0,$5C,$7F,$C2,$79,$FA
	.dc.b	$5F,$96,$DE,$FF,$E0,$1A,$FC,$2D,$7C,$71,$FC,$C3,$B3,$FD,$91,$E0
	.dc.b	$FE,$2C,$FC,$67,$83,$73,$FF,$6C,$67,$FE,$D6,$3F,$21,$E0,$D8,$69
	.dc.b	$E2,$CB,$AF,$3E,$21,$AF,$8E,$3F,$C6,$5A,$F8,$F8,$7F,$8C,$6B,$FE
	.dc.b	$D9,$ED,$FC,$E7,$5F,$E7,$5B,$9F,$FB,$A7,$7F,$F7,$39,$FA,$1E,$AB
	.dc.b	$1E,$2E,$4F,$C9,$6B,$E4,$D7,$C0,$DD,$BC,$AF,$E2,$AB,$E9,$3F,$AF
	.dc.b	$CF,$FD,$D5,$D5,$FA,$D3,$F5,$DB,$5A,$FF,$7D,$71,$7F,$BE,$0F,$D9
	.dc.b	$73,$D8,$F1,$1F,$07,$2F,$0E,$3F,$0B,$2F,$0B,$F9,$F9,$F0,$F2,$7D
	.dc.b	$3E,$87,$3F,$EF,$4F,$DF,$7C,$4D,$AE,$B7,$F4,$3D,$67,$3F,$C0,$E1
	.dc.b	$FF,$29,$3F,$C1,$7E,$27,$17,$F4,$EB,$6F,$2B,$F2,$73,$F8,$BA,$3E
	.dc.b	$AC,$D9,$72,$3F,$0E,$FF,$5C,$E7,$F3,$A3,$13,$D8,$DC,$FE,$BA,$BC
	.dc.b	$F3,$A7,$FC,$F3,$A7,$F9,$DC,$E1,$1D,$7E,$23,$13,$E5,$E2,$F9,$27
	.dc.b	$CA,$ED,$FD,$4E,$3F,$9E,$E3,$F9,$70,$C7,$8C,$B3,$F3,$67,$FE,$C1
	.dc.b	$8F,$F3,$44,$FF,$A1,$F7,$99,$79,$A9,$F1,$18,$BF,$CE,$6F,$F0,$DF
	.dc.b	$B0,$78,$5F,$EA,$17,$BC,$7F,$D2,$18,$DC,$7F,$45,$F2,$1F,$21,$B8
	.dc.b	$7C,$71,$E0,$3C,$0C,$56,$E7,$F3,$19,$78,$75,$F8,$3C,$FF,$E0,$DE
	.dc.b	$0F,$C6,$DE,$EC,$DA,$6B,$D2,$7A,$E1,$33,$F8,$1F,$13,$67,$DE,$7B
	.dc.b	$4F,$E3,$1D,$7F,$B7,$7F,$9B,$E3,$F7,$5F,$F5,$1B,$77,$F9,$FF,$13
	.dc.b	$CD,$FD,$9C,$F9,$39,$3D,$A7,$C1,$E0,$77,$E3,$3F,$C0,$72,$FE,$9B
	.dc.b	$83,$F0,$1A,$78,$78,$7F,$C5,$9A,$FF,$C0,$0B,$FC,$D3,$73,$F1,$EF
	.dc.b	$CD,$F2,$C7,$CC,$31,$AF,$91,$FC,$26,$3F,$8E,$CB,$F1,$79,$77,$E7
	.dc.b	$FB,$3B,$7D,$4E,$0F,$F2,$6E,$1D,$F1,$E5,$18,$E5,$FB,$27,$B7,$96
	.dc.b	$ED,$EC,$DB,$ED,$67,$D7,$6F,$E5,$39,$31,$EA,$76,$CF,$C1,$87,$F8
	.dc.b	$B6,$74,$FC,$09,$EB,$5F,$49,$FC,$66,$DE,$BC,$7F,$AD,$EF,$F8,$7D
	.dc.b	$FF,$88,$E3,$CB,$0F,$B7,$87,$7C,$7A,$1D,$F8,$BF,$26,$7D,$7E,$3B
	.dc.b	$1E,$BA,$FC,$36,$3F,$D3,$A3,$F4,$D8,$2F,$E3,$9E,$C8,$E8,$7A,$FE
	.dc.b	$27,$F3,$E7,$F0,$2E,$73,$88,$ED,$7E,$9D,$3D,$7C,$7F,$CB,$75,$F5
	.dc.b	$DF,$D4,$3F,$C8,$D8,$3E,$27,$18,$FC,$97,$F5,$A7,$7F,$89,$B1,$DB
	.dc.b	$0E,$98,$F6,$D7,$F6,$27,$E8,$2F,$F0,$27,$47,$15,$F5,$6F,$F7,$27
	.dc.b	$AF,$E2,$6C,$75,$9E,$7A,$7C,$77,$E4,$E3,$FC,$1A,$FF,$1A,$76,$AE
	.dc.b	$87,$F9,$B7,$B3,$FC,$BA,$FF,$39,$6D,$7F,$1F,$8F,$EE,$7D,$9F,$CD
	.dc.b	$FE,$27,$F8,$75,$6B,$A7,$F6,$7C,$BF,$85,$C3,$F7,$B8,$7F,$3F,$C3
	.dc.b	$F9,$F7,$C4,$FC,$A3,$E5,$FC,$4D,$9F,$C9,$6D,$FC,$58,$FB,$1F,$13
	.dc.b	$63,$DD,$87,$F4,$C3,$A9,$EE,$3F,$3F,$3F,$AA,$3E,$BF,$C4,$D9,$FC
	.dc.b	$F6,$BF,$C8,$8F,$E4,$DF,$BA,$FF,$39,$E8,$47,$E7,$B9,$FF,$83,$1F
	.dc.b	$B8,$AF,$C8,$C7,$ED,$8E,$98,$FF,$18,$7F,$73,$A7,$F2,$E3,$B4,$FF
	.dc.b	$0A,$7A,$72,$FE,$DA,$39,$92,$BD,$43,$85,$E5,$FC,$E3,$F3,$5F,$15
	.dc.b	$6B,$E1,$F1,$7F,$1A,$7F,$A5,$3F,$47,$9B,$F8,$DE,$7F,$C0,$F1,$E9
	.dc.b	$E5,$FD,$83,$F4,$9F,$15,$70,$E7,$B7,$BE,$BC,$78,$F3,$E8,$FE,$9F
	.dc.b	$8F,$3D,$5E,$67,$E0,$8F,$D7,$7C,$75,$CB,$9F,$07,$E7,$F2,$FB,$1B
	.dc.b	$7E,$BD,$F7,$F1,$7D,$0A,$FE,$C5,$8F,$F7,$2E,$9F,$E7,$AD,$8F,$C8
	.dc.b	$F6,$3E,$2F,$F2,$3E,$1F,$CC,$C7,$E4,$F9,$1F,$D6,$B8,$3F,$5F,$1E
	.dc.b	$AE,$2F,$E2,$BC,$3E,$9F,$37,$F3,$27,$F0,$5F,$8B,$BA,$7A,$2F,$F1
	.dc.b	$BC,$1E,$72,$63,$D1,$F9,$B3,$E8,$F3,$7F,$46,$7F,$0F,$F8,$BB,$1C
	.dc.b	$E7,$F6,$87,$2B,$FD,$BD,$F3,$3E,$6C,$FC,$2F,$27,$F5,$A7,$F8,$6B
	.dc.b	$F0,$FF,$F5,$14,$7E,$D6,$3E,$05,$F3,$BF,$F0,$4B,$7A,$3E,$C4,$7F
	.dc.b	$42,$D7,$EC,$79,$1F,$B8,$3A,$7C,$65,$C7,$CD,$9F,$F8,$DA,$FE,$47
	.dc.b	$47,$E5,$72,$ED,$F7,$9F,$E0,$33,$FC,$15,$E3,$36,$7C,$6F,$F8,$46
	.dc.b	$7F,$C1,$1B,$78,$B1,$F2,$E5,$EF,$7E,$09,$0F,$F0,$6E,$DF,$97,$AF
	.dc.b	$F0,$06,$9F,$E7,$F6,$CF,$F9,$23,$1F,$F5,$84,$FF,$93,$E7,$33,$F8
	.dc.b	$EB,$EC,$F8,$0E,$D8,$78,$FA,$FF,$21,$1F,$B0,$AF,$EE,$17,$DF,$CD
	.dc.b	$FA,$1F,$1B,$EB,$7E,$7E,$3D,$3A,$3F,$1D,$87,$4E,$8F,$E0,$FC,$3F
	.dc.b	$C3,$B1,$FF,$1E,$65,$FE,$45,$3F,$C5,$2F,$F9,$63,$57,$C6,$FC,$7D
	.dc.b	$7A,$E3,$F0,$F2,$EA,$FA,$CE,$D8,$F5,$7D,$1E,$87,$CF,$3F,$0D,$F1
	.dc.b	$B7,$2F,$B7,$3F,$E9,$33,$ED,$9F,$B0,$E1,$F9,$DE,$8F,$D7,$3F,$1D
	.dc.b	$F1,$16,$B8,$D7,$C9,$3C,$44,$E3,$FF,$22,$DF,$D7,$F4,$BF,$00,$7E
	.dc.b	$5F,$E3,$2D,$F2,$D7,$D7,$7F,$83,$38,$C7,$C6,$F4,$FF,$10,$7E,$7F
	.dc.b	$E3,$2C,$71,$F0,$BF,$8D,$57,$D9,$CB,$F0,$DE,$A7,$E3,$CF,$D3,$7C
	.dc.b	$65,$BE,$07,$F2,$4E,$7F,$AB,$B7,$F3,$78,$FA,$BE,$AF,$E5,$CF,$D7
	.dc.b	$7C,$65,$8E,$33,$F2,$3B,$FF,$5D,$CF,$F8,$87,$F4,$7E,$A7,$E7,$CF
	.dc.b	$DB,$7C,$5D,$BE,$33,$FB,$3C,$BC,$2F,$F0,$8C,$3F,$21,$AF,$E9,$CF
	.dc.b	$DF,$7C,$55,$BE,$3D,$DF,$9A,$E0,$F8,$11,$FE,$22,$F6,$3F,$CF,$67
	.dc.b	$F5,$AF,$4E,$C7,$0C,$7F,$0E,$7D,$F7,$FC,$53,$EC,$7F,$A0,$8D,$C7
	.dc.b	$F9,$B9,$FC,$8B,$F7,$B9,$FE,$BE,$BE,$8C,$BF,$57,$C1,$F0,$5B,$EC
	.dc.b	$7F,$A3,$0E,$98,$FF,$9A,$1E,$7A,$7F,$55,$C7,$E5,$57,$2D,$BF,$AA
	.dc.b	$60,$D3,$FD,$2A,$7F,$A6,$7F,$4A,$7A,$D8,$CF,$D9,$8F,$E4,$74,$F2
	.dc.b	$3F,$C7,$63,$FC,$8B,$EB,$7F,$A7,$8E,$11,$FE,$8A,$7C,$73,$FD,$77
	.dc.b	$8B,$EA,$78,$5F,$C7,$75,$F6,$FA,$BF,$E7,$23,$FC,$E7,$7F,$EA,$87
	.dc.b	$F2,$37,$C6,$7E,$A3,$E2,$C3,$F8,$CE,$3F,$E5,$6F,$4B,$FC,$F6,$73
	.dc.b	$7F,$CD,$CF,$2C,$7F,$B2,$70,$7E,$8E,$FF,$AA,$D7,$F9,$7B,$D1,$FF
	.dc.b	$41,$85,$83,$46,$58,$8F,$F7,$EF,$4F,$0A,$FD,$06,$5D,$B1,$F3,$65
	.dc.b	$FB,$0B,$FE,$E3,$E8,$7F,$A4,$0E,$0F,$FA,$C9,$F3,$3F,$D8,$B1,$F7
	.dc.b	$DF,$C7,$8F,$E5,$78,$79,$5F,$24,$F8,$EF,$FA,$BC,$73,$9F,$4E,$5F
	.dc.b	$EC,$D7,$AB,$FD,$92,$BA,$CF,$E5,$CF,$F0,$8F,$9D,$D6,$7F,$DA,$8D
	.dc.b	$9F,$81,$E8,$F6,$9C,$DF,$C0,$BF,$41,$FA,$28,$FD,$CF,$2F,$F0,$AF
	.dc.b	$3B,$94,$FC,$CC,$BF,$AC,$DF,$9B,$1F,$56,$5A,$A5,$69,$FC,$D9,$FC
	.dc.b	$B7,$1F,$C5,$BE,$96,$BF,$54,$FF,$2C,$16,$1A,$B4,$D4,$FF,$A1,$0D
	.dc.b	$46,$5C,$CF,$09,$DC,$74,$3F,$C3,$E7,$07,$31,$B8,$E3,$3C,$DD,$1D
	.dc.b	$3C,$E7,$BD,$C1,$B1,$98,$C1,$1C,$10,$F6,$47,$06,$C6,$DA,$6C,$FC
	.dc.b	$37,$99,$F9,$9C,$7F,$37,$1F,$4E,$31,$9E,$A3,$EA,$1F,$5F,$CB,$CB
	.dc.b	$F6,$70,$F3,$DF,$28,$FA,$F1,$8A,$FC,$A3,$63,$F5,$7F,$7B,$FE,$6E
	.dc.b	$7A,$9C,$31,$B9,$AB,$84,$E4,$FC,$F9,$F8,$9F,$87,$F1,$BF,$86,$CA
	.dc.b	$D1,$A6,$B8,$9E,$2E,$B0,$EB,$5D,$F1,$F0,$C7,$83,$2F,$36,$3E,$03
	.dc.b	$C5,$A7,$BE,$F8,$46,$5B,$19,$CF,$C7,$E6,$4A,$F6,$BE,$6F,$2F,$C5
	.dc.b	$97,$80,$93,$DF,$1F,$0E,$1D,$F2,$F3,$92,$BE,$3E,$3C,$EF,$E4,$E1
	.dc.b	$5F,$27,$95,$D9,$F3,$5F,$17,$B7,$B5,$9C,$7B,$63,$E2,$7D,$53,$F5
	.dc.b	$2F,$E6,$91,$FB,$3F,$BB,$3E,$3B,$97,$17,$E0,$30,$E4,$FF,$27,$D7
	.dc.b	$D6,$E6,$FC,$73,$A7,$EE,$1F,$82,$E0,$FB,$64,$7F,$0B,$D7,$D9,$FC
	.dc.b	$94,$F1,$57,$D9,$C9,$D8,$99,$FB,$1F,$B1,$5F,$88,$23,$F9,$2F,$27
	.dc.b	$F9,$F7,$B7,$FE,$EE,$DF,$E4,$5F,$C6,$AF,$92,$78,$78,$FF,$03,$A7
	.dc.b	$B6,$7B,$1F,$87,$C3,$9F,$57,$F3,$76,$DF,$F2,$36,$BF,$C8,$3C,$AB
	.dc.b	$87,$48,$F5,$DF,$84,$FD,$86,$9E,$F8,$FC,$AE,$3E,$12,$3F,$30,$FC
	.dc.b	$B5,$F4,$23,$FD,$7B,$3F,$E9,$8B,$C4,$7B,$3E,$2D,$7E,$0E,$1D,$2B
	.dc.b	$E7,$F2,$7D,$37,$C2,$79,$A3,$FA,$A7,$37,$ED,$3E,$3B,$D6,$49,$F0
	.dc.b	$70,$7F,$59,$F9,$1E,$DC,$7F,$2F,$2F,$D9,$65,$D7,$87,$E1,$61,$D8
	.dc.b	$FE,$8B,$97,$F0,$0C,$3E,$39,$23,$A9,$FD,$3F,$E3,$BA,$BF,$82,$D3
	.dc.b	$D5,$C3,$F2,$70,$EA,$4C,$BF,$82,$E3,$FC,$33,$0F,$A6,$4C,$3A,$8E
	.dc.b	$77,$C2,$F6,$E4,$FA,$D9,$F6,$D7,$D5,$97,$CA,$C3,$EA,$57,$F2,$9F
	.dc.b	$65,$F0,$3F,$82,$6C,$FF,$AB,$E3,$F0,$FE,$E7,$FB,$07,$8F,$B6,$5F
	.dc.b	$4F,$7F,$8B,$C1,$EB,$CF,$B6,$3F,$1E,$5C,$1F,$CA,$33,$7A,$5D,$C9
	.dc.b	$5F,$EA,$7E,$2E,$AF,$DB,$21,$F8,$03,$EB,$71,$FE,$86,$FA,$7A,$5E
	.dc.b	$92,$4F,$FA,$C7,$93,$9B,$E3,$CF,$EE,$BC,$8F,$A8,$7C,$4D,$7F,$DA
	.dc.b	$5E,$8F,$C3,$24,$7F,$AE,$FE,$3B,$9E,$DF,$5F,$3F,$E9,$7B,$FC,$17
	.dc.b	$A7,$1F,$EC,$B1,$FF,$6E,$7A,$3F,$34,$98,$FF,$B1,$76,$E7,$A7,$C8
	.dc.b	$C7,$94,$F9,$0F,$95,$3F,$27,$D0,$FA,$C4,$8F,$F6,$77,$C6,$F2,$CF
	.dc.b	$F8,$A5,$F2,$CF,$F0,$75,$F2,$B4,$FA,$6F,$F6,$2E,$37,$A3,$F8,$12
	.dc.b	$47,$FB,$67,$E3,$7A,$3E,$3C,$3A,$6F,$F8,$1D,$BF,$2F,$3C,$FD,$2F
	.dc.b	$C5,$11,$FD,$1F,$43,$F1,$1F,$F0,$47,$83,$FC,$77,$7F,$8F,$8F,$F3
	.dc.b	$7E,$CF,$CB,$C7,$DD,$C3,$CD,$E9,$FE,$54,$8F,$4C,$1F,$1D,$E6,$E2
	.dc.b	$F7,$CF,$E1,$CF,$95,$C5,$F5,$DF,$ED,$1C,$BF,$93,$F5,$3F,$40,$48
	.dc.b	$E9,$F2,$5C,$F5,$FD,$54,$79,$8F,$95,$5F,$E0,$2F,$89,$FE,$33,$BF
	.dc.b	$F1,$A6,$F8,$9F,$E7,$76,$D3,$87,$E1,$61,$D6,$3F,$3B,$C9,$F3,$1F
	.dc.b	$AD,$C9,$F3,$1F,$F0,$15,$BD,$4F,$E6,$ED,$8F,$E0,$6D,$59,$7F,$9D
	.dc.b	$F1,$FA,$1F,$25,$DE,$7F,$8D,$63,$DD,$FD,$27,$27,$F7,$C7,$FC,$4B
	.dc.b	$83,$D6,$EE,$48,$FA,$1F,$21,$CF,$8B,$F8,$0F,$3F,$D8,$9F,$9F,$C3
	.dc.b	$F1,$7D,$6F,$39,$23,$EC,$7A,$1C,$78,$7F,$67,$1F,$59,$FE,$13,$87
	.dc.b	$E5,$A3,$E0,$C7,$C0,$23,$F8,$1F,$8D,$E5,$AF,$EC,$B2,$F1,$73,$FE
	.dc.b	$0B,$0E,$4F,$F9,$1F,$67,$AF,$F2,$89,$8F,$E2,$FE,$43,$9B,$FD,$B8
	.dc.b	$F2,$63,$F9,$37,$E9,$FB,$1F,$48,$8F,$E4,$EF,$8E,$1F,$D9,$B2,$F1
	.dc.b	$F3,$FF,$27,$E0,$FC,$47,$AF,$F6,$89,$87,$E6,$FE,$3F,$8D,$7E,$8E
	.dc.b	$3F,$20,$7F,$02,$3E,$AF,$B1,$F8,$22,$3F,$A3,$F4,$38,$BF,$E6,$0C
	.dc.b	$7C,$2F,$E6,$0F,$BB,$EC,$7E,$28,$8F,$EA,$FE,$43,$8B,$FA,$9D,$3E
	.dc.b	$57,$27,$D4,$D7,$FC,$BF,$E5,$FF,$81,$89,$5F,$D5,$39,$3E,$C7,$6F
	.dc.b	$EF,$63,$E4,$9F,$2F,$0F,$E4,$8F,$E7,$B8,$FF,$20,$FE,$D0,$E5,$87
	.dc.b	$F8,$48,$FF,$0A,$FA,$6C,$FC,$BD,$9F,$DA,$67,$FC,$E2,$4A,$FF,$3D
	.dc.b	$F9,$5F,$E7,$43,$FC,$36,$47,$F8,$46,$FC,$8D,$3C,$08,$7F,$A0,$3B
	.dc.b	$FF,$8C,$7B,$BF,$BF,$F6,$BF,$62,$49,$FE,$25,$F1,$DC,$DF,$F4,$CD
	.dc.b	$7F,$1F,$F8,$4F,$F4,$CB,$63,$3E,$4A,$72,$FF,$88,$23,$FD,$27,$E6
	.dc.b	$7F,$8C,$88,$FF,$1F,$F7,$7B,$3F,$E8,$18,$FF,$48,$91,$FE,$A7,$C3
	.dc.b	$FE,$96,$DF,$F9,$7F,$91,$FE,$63,$8F,$F1,$FC,$62,$7F,$8A,$60,$F8
	.dc.b	$CE,$BC,$9F,$E3,$EB,$E4,$4A,$FE,$BF,$E6,$75,$39,$91,$FE,$69,$F1
	.dc.b	$9C,$BB,$BF,$CE,$53,$B9,$E2,$7C,$6C,$FF,$DD,$18,$7F,$93,$3C,$CF
	.dc.b	$29,$27,$FC,$3F,$F1,$EF,$8A,$E9,$1F,$EB,$C9,$F7,$11,$FE,$E7,$8F
	.dc.b	$C1,$3D,$E4,$7F,$C5,$3B,$3C,$DE,$71,$FE,$7A,$9E,$0F,$23,$FB,$EF
	.dc.b	$95,$F2,$26,$47,$27,$F2,$BD,$1C,$AF,$FD,$A9,$9F,$F9,$9B,$0E,$77
	.dc.b	$FE,$AD,$C3,$E8,$F9,$92,$3C,$6F,$D4,$8E,$47,$9E,$3F,$D6,$93,$F6
	.dc.b	$E3,$51,$EA,$6B,$FD,$03,$53,$3B,$5D,$5D,$23,$81,$0E,$07,$FB,$9B
	.dc.b	$CB,$FC,$29,$1F,$8E,$FF,$90,$A9,$F0,$5E,$F8,$D4,$F2,$3F,$D4,$6E
	.dc.b	$FE,$37,$F1,$FE,$64,$9F,$A2,$FF,$A4,$3F,$77,$7B,$7F,$26,$35,$3C
	.dc.b	$F8,$11,$F9,$8F,$36,$4F,$DA,$6B,$EA,$EF,$68,$F8,$6F,$A9,$5A,$F1
	.dc.b	$BF,$D1,$F8,$A3,$F4,$7E,$6C,$C7,$F0,$AD,$7A,$60,$F8,$DF,$BD,$39
	.dc.b	$C3,$9B,$F1,$7A,$BF,$BA,$FC,$34,$C3,$FB,$AB,$F2,$7E,$37,$F3,$3A
	.dc.b	$F9,$BE,$37,$E6,$12,$BF,$C4,$F9,$BE,$37,$FB,$FF,$C4,$C9,$FE,$FE
	.dc.b	$FD,$3F,$8D,$E7,$4F,$8D,$FA,$C4,$AE,$7F,$19,$FE,$08,$F8,$C9,$1F
	.dc.b	$E0,$87,$EE,$7C,$57,$2F,$89,$FB,$C4,$9F,$1F,$47,$F8,$43,$E2,$A4
	.dc.b	$7F,$84,$1F,$C2,$7C,$47,$8B,$E2,$3F,$0D,$F1,$12,$BD,$71,$FE,$18
	.dc.b	$F8,$A9,$1F,$E1,$87,$F1,$9F,$11,$EA,$F8,$8F,$C7,$12,$7E,$1F,$C4
	.dc.b	$7F,$18,$F8,$09,$1F,$BA,$7F,$29,$F1,$1F,$C7,$3C,$DF,$EB,$F3,$FD
	.dc.b	$67,$CB,$FD,$1B,$FE,$F6,$AE,$2F,$E3,$5E,$0F,$F0,$4F,$77,$F9,$29
	.dc.b	$1F,$E0,$F5,$A7,$FA,$DF,$9D,$FD,$F5,$FF,$08,$1E,$32,$1F,$99,$8E
	.dc.b	$1E,$0F,$E4,$1B,$D1,$E4,$F3,$1E,$D7,$5C,$3A,$61,$FD,$9F,$3F,$4F
	.dc.b	$8D,$FB,$A8,$EE,$75,$3C,$A7,$F9,$BB,$F1,$2F,$F2,$7A,$8E,$0F,$EA
	.dc.b	$FC,$EF,$E8,$27,$FA,$F2,$7F,$C9,$1A,$7E,$92,$3F,$A4,$12,$7F,$AD
	.dc.b	$CF,$F9,$0D,$FF,$4F,$C7,$27,$F9,$5B,$CF,$E1,$FF,$A6,$E5,$27,$F9
	.dc.b	$7B,$DF,$E0,$B1,$0C,$D9,$FB,$AB,$FA,$C1,$FD,$67,$DD,$B5,$C8,$87
	.dc.b	$EF,$F2,$FF,$1D,$BF,$CF,$23,$FA,$E9,$1F,$89,$F0,$D5,$F0,$5F,$CC
	.dc.b	$9E,$D1,$FE,$BA,$F8,$CA,$FF,$11,$8E,$1A,$FF,$9A,$F0,$7C,$3F,$F6
	.dc.b	$AB,$97,$FD,$31,$F9,$9F,$0F,$88,$EE,$7F,$34,$E0,$FE,$FC,$E0,$FA
	.dc.b	$D3,$FE,$06,$B7,$C3,$FF,$79,$8E,$58,$FF,$18,$7E,$C7,$C1,$FF,$7A
	.dc.b	$F8,$39,$87,$F5,$C7,$F3,$F5,$FC,$8A,$3F,$A1,$FC,$93,$DD,$DC,$78
	.dc.b	$73,$5F,$F8,$1C,$FF,$04,$7C,$15,$9E,$84,$3E,$87,$A3,$FE,$0A,$89
	.dc.b	$1F,$DB,$1F,$F0,$6F,$C0,$E2,$3B,$47,$47,$FA,$AF,$2F,$E0,$1F,$EA
	.dc.b	$FF,$80,$FF,$37,$79,$10,$FE,$85,$3E,$CE,$BF,$DD,$CF,$EF,$27,$FA
	.dc.b	$35,$F8,$1F,$EF,$AF,$F9,$23,$93,$C5,$87,$79,$FD,$26,$1E,$37,$E9
	.dc.b	$5F,$85,$EE,$78,$76,$FE,$8F,$3E,$18,$F3,$94,$4F,$8E,$7E,$84,$60
	.dc.b	$D4,$67,$1F,$EE,$BE,$93,$8D,$E8,$79,$B9,$FF,$7F,$97,$97,$3F,$46
	.dc.b	$3E,$5C,$FA,$BC,$F3,$F1,$9F,$E0,$08,$CB,$FB,$DF,$37,$C3,$EE,$C8
	.dc.b	$D6,$BF,$2A,$3F,$81,$79,$7E,$6F,$67,$C5,$5F,$97,$CF,$F9,$4E,$7E
	.dc.b	$29,$E5,$3F,$DF,$F3,$FF,$01,$BD,$0F,$B7,$87,$37,$07,$C6,$F7,$3E
	.dc.b	$19,$1C,$CF,$B7,$2F,$BB,$7E,$29,$FA,$E7,$E2,$35,$CB,$C2,$3C,$7A
	.dc.b	$FC,$4C,$7C,$13,$E5,$9B,$3F,$21,$FF,$45,$F4,$E3,$0E,$75,$B3,$C7
	.dc.b	$EC,$E5,$FC,$AD,$5E,$2F,$B1,$E5,$FD,$CA,$95,$C0,$FE,$65,$F1,$D8
	.dc.b	$7E,$34,$5C,$BF,$85,$7C,$55,$CB,$F3,$73,$69,$8F,$E1,$7E,$27,$F1
	.dc.b	$3E,$DC,$9F,$8E,$FF,$8A,$7D,$1C,$47,$D3,$8B,$97,$E1,$BD,$BB,$97
	.dc.b	$EA,$26,$D7,$6E,$97,$C4,$FE,$5A,$E4,$FD,$47,$FA,$37,$C4,$E2,$3F
	.dc.b	$01,$17,$1F,$C9,$FB,$B7,$2F,$DA,$CD,$AF,$5F,$47,$FA,$3D,$FD,$3D
	.dc.b	$7E,$90,$8F,$E2,$76,$FE,$89,$F0,$F8,$8F,$C7,$4D,$C7,$F8,$E5,$7F
	.dc.b	$49,$F0,$F1,$97,$F0,$4A,$B7,$E0,$E8,$FD,$B7,$C0,$39,$69,$27,$C1
	.dc.b	$A7,$FA,$0D,$BF,$0B,$FB,$FC,$FF,$AB,$7C,$2D,$A3,$67,$1F,$0B,$9F
	.dc.b	$84,$98,$74,$8F,$E0,$3F,$07,$23,$F3,$EF,$3C,$F1,$1F,$AB,$8B,$C1
	.dc.b	$EE,$C9,$F0,$98,$E0,$FE,$43,$56,$FE,$6F,$B1,$FC,$3B,$E1,$24,$7E
	.dc.b	$B1,$F4,$7B,$38,$8F,$DC,$45,$CB,$B7,$B3,$71,$FE,$5B,$36,$BE,$CF
	.dc.b	$B1,$FC,$6B,$E1,$24,$7E,$E1,$F8,$9E,$BE,$23,$F8,$24,$5C,$7D,$5E
	.dc.b	$6D,$C7,$F9,$CC,$DA,$FC,$2F,$8D,$FC,$9B,$E1,$24,$7F,$04,$7E,$87
	.dc.b	$A3,$88,$FE,$21,$17,$1F,$8D,$E6,$5C,$7F,$A2,$CD,$AF,$C8,$FA,$1F
	.dc.b	$CC,$BE,$12,$47,$F1,$07,$EF,$7E,$25,$FC,$AE,$3F,$9D,$C6,$9F,$E3
	.dc.b	$CE,$BC,$DC,$65,$FC,$CA,$2E,$3F,$CF,$2D,$E7,$FF,$40,$F8,$59,$8F
	.dc.b	$F1,$E7,$BF,$9F,$88,$FE,$55,$17,$1F,$CE,$DF,$FA,$3E,$78,$10,$D9
	.dc.b	$FA,$77,$FD,$C8,$FF,$2A,$F8,$5F,$E9,$E4,$75,$E3,$D7,$F9,$C4,$5A
	.dc.b	$EF,$E2,$DC,$FF,$B6,$CD,$AF,$E2,$78,$7F,$A4,$DF,$E6,$FF,$0B,$FD
	.dc.b	$70,$8F,$8F,$A2,$BF,$D1,$62,$E3,$EC,$E5,$B8,$FF,$7A,$9B,$5D,$B8
	.dc.b	$BF,$B3,$FC,$2C,$8F,$E8,$AF,$D6,$E1,$C4,$7F,$51,$8B,$8F,$C8,$DE
	.dc.b	$E3,$FE,$06,$9B,$5E,$AD,$7F,$B9,$F9,$D2,$3F,$A8,$BF,$82,$CF,$11
	.dc.b	$FD,$76,$2E,$3F,$52,$EE,$1F,$E1,$29,$B5,$F1,$9F,$EF,$FF,$0B,$23
	.dc.b	$FA,$EB,$F8,$E3,$0F,$F6,$96,$DF,$E0,$3E,$16,$DF,$F8,$7A,$6D,$7D
	.dc.b	$2F,$85,$FF,$05,$D4,$8F,$ED,$2F,$E6,$BE,$17,$11,$FD,$D6,$2E,$1F
	.dc.b	$CA,$39,$FF,$D0,$BE,$BA,$73,$97,$F8,$B1,$FE,$41,$3F,$E8,$48,$FF
	.dc.b	$17,$E3,$FE,$1A,$26,$1F,$CD,$E7,$FD,$08,$FF,$80,$E3,$2D,$35,$F0
	.dc.b	$98,$C7,$FC,$33,$36,$FF,$C4,$76,$F8,$4F,$E6,$FE,$C4,$C7,$FC,$13
	.dc.b	$E4,$A3,$CB,$F1,$9E,$6F,$85,$95,$E3,$7D,$3F,$0B,$EA,$F8,$59,$3E
	.dc.b	$77,$DB,$F0,$BF,$E3,$4E,$29,$3F,$E1,$27,$E0,$FC,$2E,$23,$FC,$3D
	.dc.b	$17,$83,$E5,$7C,$1D,$E0,$FF,$29,$CD,$AF,$EF,$3C,$1F,$E3,$C7,$FC
	.dc.b	$43,$EB,$FF,$AD,$09,$5F,$3B,$E0,$BF,$C9,$51,$89,$DC,$FD,$2F,$82
	.dc.b	$CB,$23,$27,$D1,$F8,$2E,$3F,$1D,$23,$BB,$F6,$BD,$BC,$43,$5F,$F2
	.dc.b	$AE,$3F,$76,$07,$FA,$CA,$BF,$03,$C9,$C2,$B7,$17,$1F,$AF,$F0,$5E
	.dc.b	$8F,$8E,$98,$7B,$9F,$C4,$FC,$0E,$23,$C1,$5F,$EF,$2C,$BF,$0F,$8F
	.dc.b	$8A,$FF,$1F,$F0,$3B,$8C,$5F,$33,$FC,$09,$F8,$97,$F8,$26,$DD,$BF
	.dc.b	$20,$93,$AF,$82,$AF,$80,$F6,$CF,$07,$E9,$CF,$F7,$DA,$FF,$52,$46
	.dc.b	$08,$71,$3F,$DA,$B1,$B9,$FA,$FD,$B2,$7E,$EB,$CF,$E2,$F1,$1C,$67
	.dc.b	$F3,$8E,$8F,$5F,$C0,$6B,$2F,$A7,$77,$5F,$F4,$2E,$5F,$E0,$EB,$FC
	.dc.b	$4F,$77,$E1,$88,$FC,$4F,$7E,$BD,$E2,$E1,$FE,$26,$A7,$C0,$63,$0F
	.dc.b	$A5,$56,$FF,$C9,$F4,$F0,$7F,$27,$E0,$C9,$FC,$CB,$F1,$FD,$FC,$47
	.dc.b	$B6,$2E,$5D,$FD,$EB,$97,$DF,$9B,$5D,$BD,$EF,$CF,$F8,$92,$3F,$4A
	.dc.b	$FD,$3F,$7B,$11,$F2,$62,$E3,$EB,$F7,$AE,$3F,$8B,$9B,$5E,$AF,$7B
	.dc.b	$F5,$7E,$3C,$8F,$D8,$BF,$7F,$DE,$C4,$7D,$58,$B8,$FC,$7F,$7A,$E3
	.dc.b	$F9,$79,$B5,$F1,$BD,$EF,$DB,$F9,$52,$3F,$7A,$FE,$2F,$DE,$C4,$7E
	.dc.b	$06,$2E,$3F,$4F,$DE,$B8,$FE,$8E,$6D,$7F,$A0,$3F,$12,$FF,$09,$76
	.dc.b	$3C,$DF,$E0,$A4,$9F,$F3,$34,$FF,$AC,$1F,$C7,$46,$5A,$6B,$DF,$C6
	.dc.b	$5F,$C0,$26,$E3,$FE,$D2,$BF,$F5,$85,$FE,$9B,$CE,$FE,$26,$47,$BF
	.dc.b	$C3,$D7,$F3,$91,$70,$FF,$39,$D7,$F8,$4A,$72,$7F,$20,$9D,$5F,$7F
	.dc.b	$7F,$F8,$FF,$C4,$4A,$FE,$4A,$FF,$9B,$ED,$F0,$19,$9F,$D5,$4D,$C7
	.dc.b	$BF,$C1,$5C,$7F,$86,$4D,$AF,$95,$F0,$1F,$CB,$3D,$29,$1F,$CC,$9E
	.dc.b	$DF,$01,$88,$FD,$B4,$5C,$7D,$7F,$01,$71,$FE,$31,$36,$BE,$B7,$C0
	.dc.b	$7F,$38,$F5,$64,$7F,$3E,$7D,$5F,$01,$88,$FE,$05,$17,$1F,$8F,$F0
	.dc.b	$17,$1F,$E4,$93,$6B,$F0,$5F,$01,$FD,$13,$D8,$91,$FD,$29,$F8,$DF
	.dc.b	$01,$88,$FE,$1D,$17,$1F,$A7,$F0,$17,$1F,$E6,$13,$6B,$F1,$DF,$01
	.dc.b	$FD,$43,$DB,$91,$FD,$59,$FA,$5E,$FE,$23,$F8,$D4,$5C,$7E,$FF,$BF
	.dc.b	$71,$FE,$79,$36,$BF,$35,$EF,$FF,$5C,$F7,$A4,$7F,$62,$7F,$D5,$9F
	.dc.b	$89,$7F,$63,$73,$1F,$C9,$DC,$FB,$F8,$CB,$FA,$BC,$5C,$7F,$AB,$FC
	.dc.b	$33,$0F,$E9,$71,$FD,$37,$E0,$3F,$B5,$93,$1F,$F6,$64,$FF,$41,$7F
	.dc.b	$9A,$46,$5A,$6B,$D9,$C6,$3F,$D9,$A6,$DF,$82,$DE,$BF,$F7,$8F,$80
	.dc.b	$98,$FF,$7E,$7F,$B5,$BF,$D2,$7A,$1C,$9A,$8F,$E8,$93,$72,$63,$C2
	.dc.b	$DC,$79,$8F,$ED,$15,$6F,$FB,$BD,$7F,$8E,$5F,$6F,$A9,$F3,$08,$FC
	.dc.b	$0F,$47,$FC,$1B,$EF,$C8,$FF,$09,$BF,$4B,$CF,$C4,$7F,$57,$8B,$9F
	.dc.b	$FB,$BB,$67,$99,$F6,$3C,$A9,$7F,$31,$FB,$7E,$4F,$C0,$BF,$5C,$FF
	.dc.b	$83,$3F,$05,$FF,$80,$04,$F9,$0F,$C0,$F8,$FB,$E1,$F8,$F3,$9A,$FF
	.dc.b	$09,$DF,$25,$0B,$FC,$4F,$89,$F8,$AF,$0A,$57,$DE,$7F,$1B,$E0,$75
	.dc.b	$8F,$17,$C0,$D7,$9B,$70,$3F,$94,$EF,$C3,$27,$C7,$9F,$9A,$3F,$33
	.dc.b	$DB,$F9,$AE,$B9,$5F,$8E,$7F,$3B,$D1,$D7,$E2,$64,$F2,$3E,$23,$FA
	.dc.b	$2E,$5C,$C7,$F8,$3F,$3F,$F0,$9B,$67,$D5,$DD,$E6,$E0,$B9,$FA,$2B
	.dc.b	$A4,$FA,$A3,$26,$AB,$DD,$E8,$4A,$F1,$1F,$31,$E6,$65,$F1,$37,$8B
	.dc.b	$ED,$DF,$C6,$F8,$1F,$0F,$C0,$CB,$F0,$E2,$8F,$2F,$C4,$FC,$DA,$95
	.dc.b	$E0,$D1,$1E,$AF,$82,$F5,$FC,$16,$AB,$E1,$4E,$1E,$BF,$03,$F0,$08
	.dc.b	$F8,$3E,$07,$ED,$65,$23,$B3,$F1,$BE,$07,$E3,$FC,$0F,$C8,$F8,$19
	.dc.b	$5E,$88,$F9,$7F,$03,$23,$D6,$FC,$EF,$81,$FC,$36,$F2,$7D,$4F,$D2
	.dc.b	$F8,$1F,$A7,$F0,$3F,$53,$E0,$65,$7C,$48,$FA,$FF,$03,$23,$E3,$BF
	.dc.b	$6B,$E0,$7F,$25,$C7,$27,$E3,$3F,$80,$FC,$4B,$F8,$5C,$7F,$96,$F9
	.dc.b	$D1,$FE,$01,$E8,$FA,$AB,$E0,$B8,$1C,$7E,$1B,$97,$C4,$CC,$3F,$0F
	.dc.b	$5F,$E0,$0C,$39,$FC,$27,$E8,$FA,$A5,$7D,$77,$C5,$F1,$9E,$3F,$85
	.dc.b	$F2,$7C,$24,$AE,$71,$E7,$F8,$49,$1D,$9F,$57,$C2,$7E,$BF,$BA,$4F
	.dc.b	$E0,$DF,$77,$C2,$7B,$FE,$13,$E0,$7C,$24,$AF,$3C,$7C,$3F,$66,$47
	.dc.b	$A9,$F8,$DF,$09,$FB,$BF,$0E,$4F,$E3,$DF,$95,$F0,$9F,$2F,$E1,$3E
	.dc.b	$67,$BD,$2B,$E1,$C7,$CF,$F7,$BE,$87,$C2,$49,$F8,$F1,$F4,$FE,$12
	.dc.b	$47,$CA,$7E,$B7,$C2,$7E,$DB,$F1,$2F,$E3,$1D,$DF,$A8,$7F,$D5,$DD
	.dc.b	$46,$9A,$F5,$BF,$AD,$BF,$D6,$B0,$4E,$FE,$27,$A1,$D7,$E2,$7F,$8C
	.dc.b	$79,$B2,$BF,$58,$F8,$7E,$1B,$C5,$F0,$DE,$3F,$86,$95,$FC,$1B,$2F
	.dc.b	$EB,$2D,$F8,$6F,$41,$1D,$7C,$37,$F2,$5F,$46,$47,$EE,$5F,$77,$C2
	.dc.b	$FB,$FE,$17,$E0,$7C,$2C,$AE,$D1,$F0,$FE,$16,$47,$85,$F8,$DF,$0B
	.dc.b	$FC,$C7,$D7,$93,$FC,$15,$F9,$5F,$0B,$F2,$FE,$17,$E6,$7C,$2C,$AF
	.dc.b	$54,$7C,$FF,$85,$FA,$1F,$0B,$27,$DD,$1F,$CF,$7D,$C9,$1F,$C4,$5F
	.dc.b	$AD,$F0,$BF,$5F,$E1,$7F,$96,$FE,$25,$FD,$36,$BF,$95,$B5,$FD,$87
	.dc.b	$2A,$77,$F0,$7C,$3E,$16,$57,$08,$FE,$99,$F0,$32,$3F,$90,$3D,$BE
	.dc.b	$17,$BF,$C2,$78,$3E,$12,$57,$82,$3C,$7F,$09,$E4,$F7,$E4,$F9,$23
	.dc.b	$FA,$C7,$C2,$48,$FE,$58,$FA,$BD,$BF,$5F,$AD,$EC,$F4,$65,$7B,$23
	.dc.b	$DF,$E2,$FC,$0F,$02,$4F,$C0,$8F,$EC,$9F,$07,$23,$F9,$C3,$F1,$BB
	.dc.b	$3E,$3F,$3F,$C8,$E1,$95,$F2,$23,$E5,$E9,$F3,$2E,$4F,$CC,$8F,$EE
	.dc.b	$1F,$09,$23,$FA,$23,$F4,$8F,$A7,$F0,$9F,$53,$E1,$24,$FD,$48,$FE
	.dc.b	$C7,$F1,$1F,$C9,$3C,$FF,$EA,$AD,$F3,$FF,$C8,$AD,$AF,$D5,$BF,$C9
	.dc.b	$E3,$3F,$0B,$C8,$93,$FD,$AA,$7F,$92,$0F,$82,$EF,$F0,$52,$75,$1E
	.dc.b	$2F,$81,$F1,$FC,$14,$9E,$51,$E6,$F8,$29,$1D,$5F,$4F,$C1,$7A,$BE
	.dc.b	$0B,$D7,$F0,$32,$BC,$51,$EE,$F8,$2F,$7F,$C1,$49,$F3,$47,$C2,$F8
	.dc.b	$29,$1E,$97,$E2,$FC,$17,$C6,$F8,$1F,$8F,$F0,$52,$BD,$D1,$F2,$BE
	.dc.b	$07,$E5,$FC,$14,$9F,$85,$1F,$3B,$E0,$A4,$7C,$57,$E8,$FC,$17,$F8
	.dc.b	$53,$F1,$2F,$E3,$53,$FC,$61,$FE,$3E,$4E,$8D,$7C,$16,$FE,$0A,$57
	.dc.b	$F8,$62,$BF,$90,$7C,$19,$7C,$1F,$52,$4E,$FE,$0F,$C1,$F0,$72,$39
	.dc.b	$3E,$3F,$83,$F2,$7C,$1C,$9E,$CF,$9F,$E0,$FD,$1F,$05,$E9,$F8,$39
	.dc.b	$5E,$38,$F6,$7C,$1C,$8F,$33,$EF,$F8,$3F,$81,$F0,$72,$7D,$4F,$C3
	.dc.b	$F8,$2F,$89,$F0,$7F,$17,$E0,$A5,$7B,$E3,$E4,$7C,$1C,$8F,$84,$FC
	.dc.b	$BF,$83,$F9,$9F,$03,$FE,$40,$FC,$4B,$F8,$97,$9D,$FE,$32,$6F,$57
	.dc.b	$FA,$8B,$9F,$F8,$A6,$3F,$E4,$29,$FF,$22,$3F,$E0,$77,$FC,$97,$5C
	.dc.b	$7E,$0B,$A7,$C1,$75,$F8,$29,$7C,$5F,$07,$C0,$F8,$7E,$0B,$C7,$3F
	.dc.b	$5F,$4C,$3D,$EF,$CD,$84,$8C,$1C,$C8,$78,$F0,$EA,$F9,$FE,$07,$DC
	.dc.b	$47,$D5,$F0,$5F,$07,$E0,$BE,$17,$C0,$C9,$F7,$47,$C5,$F8,$2F,$8D
	.dc.b	$F0,$32,$7E,$14,$7C,$9F,$82,$91,$F1,$5F,$99,$F0,$5F,$37,$E0,$64
	.dc.b	$FC,$97,$E8,$7C,$17,$D1,$F8,$19,$3F,$35,$FA,$9F,$05,$F5,$7E,$0A
	.dc.b	$4F,$D1,$7E,$C7,$C0,$FD,$9F,$82,$93,$F5,$5F,$F3,$DF,$E2,$5F,$C5
	.dc.b	$7C,$D3,$FC,$56,$7F,$C0,$2D,$8C,$FC,$47,$0F,$84,$EB,$5F,$7F,$0A
	.dc.b	$FF,$91,$6F,$FC,$79,$DA,$C7,$C3,$3F,$83,$7C,$2F,$8B,$08,$FF,$A2
	.dc.b	$AB,$C0,$7F,$92,$DC,$7C,$17,$A0,$8E,$FE,$0F,$D7,$F0,$52,$39,$3E
	.dc.b	$EF,$83,$F7,$FC,$1C,$9E,$CF,$C2,$F8,$3F,$87,$7D,$27,$D5,$86,$1E
	.dc.b	$64,$3D,$D8,$75,$79,$90,$F8,$57,$D5,$F6,$7C,$17,$37,$E3,$5F,$D2
	.dc.b	$8F,$44,$74,$3B,$78,$9F,$EA,$0C,$D9,$FA,$67,$A9,$DE,$75,$9F,$DC
	.dc.b	$F8,$3F,$BB,$F0,$52,$BE,$64,$7E,$03,$E0,$E4,$7C,$F7,$F0,$7F,$07
	.dc.b	$F8,$4F,$83,$93,$F4,$DF,$C3,$FC,$17,$E2,$3E,$0E,$4F,$D7,$7F,$17
	.dc.b	$F0,$5D,$23,$F0,$58,$5D,$3D,$58,$F4,$9F,$C3,$63,$67,$DD,$8F,$E5
	.dc.b	$3E,$0E,$57,$E0,$DF,$CB,$FC,$1F,$E6,$3E,$0E,$4F,$E1,$DF,$F6,$AF
	.dc.b	$5F,$F4,$2E,$CF,$F4,$97,$C2,$60,$95,$FE,$D6,$9F,$E7,$FC,$8C,$3A
	.dc.b	$CF,$E4,$5E,$7F,$07,$DA,$3F,$39,$8D,$D3,$FD,$A7,$87,$73,$FC,$DA
	.dc.b	$4C,$7C,$7F,$09,$E4,$F8,$4F,$2F,$C2,$4C,$7C,$51,$E8,$F8,$4F,$4F
	.dc.b	$C2,$49,$F3,$47,$B3,$E1,$3D,$BF,$09,$27,$D5,$1D,$5E,$64,$3F,$D6
	.dc.b	$D9,$BE,$0F,$B4,$7C,$1C,$3B,$3A,$C3,$B3,$F1,$70,$EC,$F2,$C3,$E6
	.dc.b	$DC,$8F,$94,$FD,$0F,$84,$EB,$1E,$0C,$3A,$ED,$F1,$B1,$EB,$3E,$4C
	.dc.b	$7A,$CF,$CA,$C7,$EC,$FC,$27,$DA,$F8,$49,$7F,$5E,$3E,$EF,$C2,$48
	.dc.b	$FB,$4F,$E0,$3E,$13,$F0,$3F,$07,$F8,$2F,$84,$95,$F7,$E3,$F0,$BF
	.dc.b	$09,$23,$F0,$4F,$E2,$3E,$13,$A4,$7C,$8C,$2E,$9F,$7B,$1E,$B3,$F8
	.dc.b	$9C,$7A,$BF,$3B,$1F,$C8,$FC,$02,$25,$BE,$13,$1D,$AF,$C4,$BF,$B4
	.dc.b	$D7,$FA,$B9,$B1,$FE,$E9,$BE,$53,$FE,$EF,$9F,$ED,$77,$CE,$38,$7C
	.dc.b	$0E,$A3,$17,$D6,$31,$87,$82,$F8,$B7,$E0,$73,$87,$85,$B7,$E5,$6D
	.dc.b	$3E,$12,$FC,$27,$BA,$FF,$55,$9E,$1F,$7E,$3D,$FE,$13,$8E,$72,$72
	.dc.b	$7B,$E3,$AE,$52,$33,$9F,$85,$B9,$F8,$5B,$F0,$9E,$42,$3A,$F8,$4F
	.dc.b	$3F,$C1,$FA,$3E,$12,$4F,$28,$F5,$FC,$1C,$8E,$AF,$BB,$E1,$3D,$FF
	.dc.b	$09,$F0,$30,$95,$E2,$8E,$AF,$93,$0E,$AF,$D5,$C3,$A3,$F5,$F0,$BC
	.dc.b	$7F,$53,$36,$5F,$33,$E1,$25,$FB,$9F,$9F,$F0,$9D,$A3,$F0,$19,$76
	.dc.b	$C3,$96,$5D,$38,$BE,$E6,$76,$FF,$15,$97,$D8,$F8,$49,$1F,$29,$FB
	.dc.b	$7F,$09,$F7,$3E,$12,$4F,$CE,$7E,$FF,$C2,$7E,$03,$E1,$3F,$03,$F0
	.dc.b	$72,$BE,$9C,$7E,$1B,$F1,$2F,$F0,$41,$23,$1F,$09,$D1,$FC,$C6,$17
	.dc.b	$7F,$85,$8F,$59,$FB,$B8,$F5,$7F,$3D,$8F,$5F,$32,$47,$E1,$32,$FF
	.dc.b	$01,$2C,$3F,$3F,$0F,$8A,$F2,$7C,$2C,$AE,$0F,$9F,$E0,$FD,$1F,$0B
	.dc.b	$27,$A3,$EB,$F8,$3F,$67,$C2,$FB,$7E,$12,$57,$86,$3E,$07,$C2,$C8
	.dc.b	$F2,$3F,$0F,$E1,$3E,$27,$C2,$FC,$5F,$83,$95,$E9,$8F,$91,$F0,$72
	.dc.b	$3D,$8F,$CB,$F8,$3E,$91,$ED,$EC,$FE,$0F,$C2,$CB,$3C,$5F,$03,$3E
	.dc.b	$F7,$F1,$73,$EE,$F1,$1E,$2B,$06,$9F,$67,$E0,$FE,$D7,$C2,$CB,$F9
	.dc.b	$D1,$F7,$7E,$16,$47,$D1,$7F,$01,$F0,$BF,$81,$F8,$49,$3F,$55,$FC
	.dc.b	$27,$C1,$FF,$0A,$FC,$4B,$FB,$90,$F8,$4C,$19,$E2,$D7,$C2,$4C,$3F
	.dc.b	$80,$D7,$FA,$C8,$7C,$2F,$3F,$84,$E9,$F0,$B2,$B7,$1D,$FE,$0A,$47
	.dc.b	$27,$C5,$F0,$7E,$3F,$82,$F2,$7C,$1C,$AE,$F1,$E7,$F8,$5F,$47,$C2
	.dc.b	$C9,$F1,$C7,$AF,$E1,$3D,$9F,$0B,$27,$CF,$1E,$FF,$83,$F8,$1F,$07
	.dc.b	$27,$D7,$1F,$0F,$E0,$A4,$7B,$9F,$8D,$F0,$7F,$1F,$E1,$7E,$47,$C2
	.dc.b	$CA,$F8,$71,$F2,$FE,$0F,$F9,$AF,$C1,$FF,$46,$3F,$A3,$F3,$7F,$AC
	.dc.b	$2B,$F0,$1E,$12,$7F,$9F,$C7,$E0,$9B,$19,$F8,$3E,$71,$D3,$E0,$FA
	.dc.b	$FC,$2C,$BF,$E6,$72,$9F,$0F,$C2,$C9,$CB,$E4,$F8,$5F,$2F,$C2,$F9
	.dc.b	$BE,$02,$57,$28,$F4,$FC,$0C,$8E,$AF,$B3,$E0,$7D,$BF,$03,$EE,$F8
	.dc.b	$59,$5E,$28,$F8,$3F,$09,$F0,$BE,$12,$4F,$9A,$3E,$2F,$C2,$7C,$6F
	.dc.b	$84,$93,$EA,$8F,$93,$F0,$7F,$2B,$E0,$E4,$FB,$A3,$E6,$FC,$14,$8F
	.dc.b	$82,$FD,$0F,$83,$FE,$CB,$F8,$97,$FA,$9C,$FF,$52,$12,$73,$F0,$9A
	.dc.b	$F8,$49,$1F,$D7,$27,$F8,$CF,$B7,$FD,$D9,$7C,$27,$42,$4E,$FE,$13
	.dc.b	$BF,$C2,$78,$3E,$12,$4F,$38,$F1,$FC,$1F,$93,$E1,$24,$F7,$8F,$3F
	.dc.b	$C1,$48,$F1,$3E,$AF,$82,$F5,$FC,$04,$9F,$33,$EE,$F8,$2F,$7F,$C2
	.dc.b	$7C,$0F,$84,$95,$EB,$8F,$87,$F0,$52,$3D,$CF,$C6,$F8,$3F,$8F,$F0
	.dc.b	$72,$7E,$13,$F2,$BE,$0F,$FC,$23,$0F,$D0,$5E,$F6,$17,$BB,$4C,$47
	.dc.b	$D2,$F8,$4F,$A7,$F0,$92,$FE,$74,$7F,$86,$BF,$13,$FF,$41,$13,$E1
	.dc.b	$31,$F0,$BE,$0F,$D0,$5B,$52,$61,$FE,$16,$1F,$9C,$FF,$B9,$70,$7C
	.dc.b	$3F,$52,$4F,$F8,$62,$3F,$8B,$B7,$E1,$BC,$44,$75,$AF,$C0,$BF,$F1
	.dc.b	$EC,$7F,$37,$3B,$99,$3E,$2F,$C1,$3C,$DF,$67,$04,$9F,$03,$EF,$F8
	.dc.b	$6F,$81,$F0,$BF,$07,$E1,$A5,$79,$63,$9D,$F5,$8F,$76,$3E,$2F,$17
	.dc.b	$FC,$99,$1F,$E8,$79,$F9,$59,$7C,$BF,$86,$93,$EF,$8F,$9D,$F0,$B2
	.dc.b	$3E,$13,$F4,$7E,$1B,$BC,$76,$3C,$07,$BE,$7D,$8F,$C9,$F8,$5F,$B1
	.dc.b	$F0,$D2,$30,$7C,$DF,$86,$FB,$9F,$0D,$23,$07,$D7,$BE,$6F,$CB,$D7
	.dc.b	$D9,$D3,$FE,$6A,$6C,$7F,$9D,$67,$7B,$7F,$9A,$39,$5F,$0B,$F8,$9F
	.dc.b	$86,$97,$F7,$63,$F2,$FC,$1F,$E5,$38,$FB,$0B,$DF,$C9,$AF,$82,$F8
	.dc.b	$33,$E3,$3E,$0E,$3F,$3B,$2F,$0E,$9F,$E7,$EC,$1F,$01,$D8,$97,$F9
	.dc.b	$38,$7B,$FF,$9D,$F8,$6F,$F3,$BF,$E2,$9F,$E8,$83,$FD,$13,$BF,$F4
	.dc.b	$3C,$FB,$73,$F8,$33,$F8,$59,$79,$27,$FD,$0F,$7F,$D1,$0F,$E9,$11
	.dc.b	$FE,$72,$6B,$FD,$31,$FE,$96,$BF,$3E,$7A,$7F,$BA,$5E,$95,$31,$1D
	.dc.b	$F0,$F0,$D7,$E0,$CF,$B1,$8F,$F0,$2F,$CF,$BD,$B1,$FB,$16,$BF,$A5
	.dc.b	$6E,$5F,$96,$7C,$AF,$E6,$3E,$F5,$F0,$9F,$E2,$0A,$FD,$BF,$17,$FA
	.dc.b	$AA,$73,$F1,$3F,$EB,$5F,$17,$FD,$C9,$5F,$EB,$79,$CB,$C0,$C5,$F3
	.dc.b	$9D,$C2,$77,$18,$6D,$F7,$AE,$0D,$BF,$03,$6F,$B1,$CE,$76,$F8,$27
	.dc.b	$6D,$37,$5D,$9C,$5F,$86,$B8,$C7,$3A,$E2,$D3,$8C,$6A,$B9,$46,$6B
	.dc.b	$95,$70,$9C,$92,$B8,$12,$34,$DA,$EB,$18,$23,$82,$57,$66,$DF,$C3
	.dc.b	$8C,$11,$C1,$23,$DE,$D7,$D0,$DB,$F3,$57,$82,$B9,$57,$02,$1C,$0B
	.dc.b	$3C,$1F,$19,$8B,$F6,$4F,$28,$E0,$49,$E1,$19,$6D,$F8,$EB,$8B,$6F
	.dc.b	$C8,$DB,$EE,$73,$9D,$BE,$19,$DB,$4D,$D7,$67,$17,$E1,$AE,$31,$CE
	.dc.b	$B8,$B4,$E3,$1A,$AE,$51,$9A,$E5,$5C,$27,$24,$AE,$04,$8D,$36,$BA
	.dc.b	$C6,$08,$E0,$95,$D9,$B7,$F0,$E3,$04,$70,$48,$F7,$B5,$F4,$36,$FC
	.dc.b	$D5,$E0,$AE,$55,$C0,$87,$02,$CF,$07,$C6,$62,$FD,$93,$CA,$38,$12
	.dc.b	$78,$46,$5B,$7E,$3A,$E2,$DB,$F2,$36,$FB,$9C,$E7,$6F,$86,$76,$D3
	.dc.b	$75,$D9,$C5,$F8,$6B,$8C,$73,$AE,$2D,$38,$C6,$AB,$94,$66,$B9,$57
	.dc.b	$09,$C9,$2B,$81,$23,$4D,$AE,$B1,$82,$38,$25,$76,$6D,$FC,$38,$C1
	.dc.b	$1C,$12,$3D,$ED,$7D,$0D,$BF,$35,$78,$2B,$95,$70,$21,$C0,$B3,$C1
	.dc.b	$F1,$98,$BF,$64,$F2,$8E,$04,$9E,$11,$96,$DF,$8E,$B8,$B6,$FC,$8D
	.dc.b	$BE,$E7,$39,$DB,$E1,$9D,$B4,$DD,$76,$71,$7E,$1A,$E3,$1C,$EB,$8B
	.dc.b	$4E,$31,$AA,$E5,$19,$AE,$55,$C2,$72,$4A,$E0,$48,$D3,$6B,$AC,$60
	.dc.b	$8E,$09,$5D,$9B,$7F,$0E,$30,$47,$04,$8F,$7B,$5F,$43,$6F,$CD,$5E
	.dc.b	$0A,$E5,$5C,$08,$70,$2C,$F0,$7C,$66,$2F,$D9,$3C,$A3,$81,$27,$84
	.dc.b	$65,$B7,$E3,$AE,$2D,$BF,$23,$6F,$B9,$CE,$76,$F8,$67,$6D,$37,$5D
	.dc.b	$9C,$5F,$86,$B8,$C7,$3A,$E2,$D3,$8C,$6A,$B9,$46,$6B,$95,$70,$9C
	.dc.b	$92,$B8,$12,$34,$DA,$EB,$18,$23,$82,$57,$66,$DF,$C3,$8C,$11,$C1
	.dc.b	$23,$DE,$D7,$D0,$DB,$F3,$57,$82,$B9,$57,$02,$1C,$0B,$3C,$1F,$19
	.dc.b	$8B,$F6,$4F,$28,$E0,$49,$E1,$19,$6D,$F8,$EB,$8B,$6F,$C8,$DB,$EE
	.dc.b	$73,$9D,$BE,$19,$DB,$4D,$D7,$67,$17,$E1,$AE,$31,$CE,$B8,$B4,$E3
	.dc.b	$1A,$AE,$51,$9A,$E5,$5C,$27,$24,$AE,$04,$8D,$36,$BA,$C6,$08,$E0
	.dc.b	$95,$D9,$B7,$F0,$E3,$04,$70,$48,$F7,$B5,$F4,$36,$FC,$D5,$E0,$AE
	.dc.b	$55,$C0,$87,$02,$CF,$07,$C6,$62,$FD,$93,$CA,$38,$12,$78,$46,$5B
	.dc.b	$7E,$3A,$E2,$DB,$F2,$36,$FB,$9C,$E7,$6F,$86,$76,$D3,$75,$D9,$C5
	.dc.b	$F8,$6B,$8C,$73,$AE,$2D,$38,$C6,$AB,$94,$66,$B9,$57,$09,$C9,$2B
	.dc.b	$81,$23,$4D,$AE,$B1,$82,$38,$25,$76,$6D,$FC,$38,$C1,$1C,$12,$3D
	.dc.b	$ED,$7D,$0D,$BF,$35,$78,$2B,$95,$70,$21,$C0,$B3,$C1,$F1,$98,$BF
	.dc.b	$64,$F2,$8E,$04,$9E,$11,$96,$DF,$8E,$B8,$B6,$FC,$8D,$BE,$E7,$39
	.dc.b	$DB,$E1,$9D,$B4,$DD,$76,$71,$7E,$1A,$E3,$1C,$EB,$8B,$4E,$31,$AA
	.dc.b	$E5,$19,$AE,$55,$C2,$72,$4A,$E0,$48,$D3,$6B,$AC,$60,$8E,$09,$5D
	.dc.b	$9B,$7F,$0E,$30,$47,$04,$8F,$7B,$5F,$43,$6F,$CD,$5E,$0A,$E5,$5C
	.dc.b	$08,$70,$2C,$F0,$7C,$66,$2F,$D9,$3C,$A3,$81,$27,$84,$65,$B7,$E3
	.dc.b	$AE,$2D,$BF,$23,$6F,$B9,$CE,$76,$F8,$67,$6D,$37,$5D,$9C,$5F,$86
	.dc.b	$B8,$C7,$3A,$E2,$D3,$8C,$6A,$B9,$46,$6B,$95,$70,$9C,$92,$B8,$12
	.dc.b	$34,$DA,$EB,$18,$23,$82,$57,$66,$DF,$C3,$8C,$11,$C1,$23,$DE,$D7
	.dc.b	$D0,$DB,$F3,$57,$82,$B9,$57,$02,$1C,$0B,$3C,$1F,$19,$8B,$F6,$4F
	.dc.b	$28,$E0,$49,$E1,$19,$6D,$F8,$EB,$8B,$6F,$C8,$DB,$EE,$73,$9D,$BE
	.dc.b	$19,$DB,$4D,$D7,$67,$17,$E1,$AE,$31,$CE,$B8,$B4,$E3,$1A,$AE,$51
	.dc.b	$9A,$E5,$5C,$27,$24,$AE,$04,$8D,$36,$BA,$C6,$08,$E0,$95,$D9,$B7
	.dc.b	$F0,$E3,$04,$70,$48,$F7,$B5,$F4,$36,$FC,$D5,$E0,$AE,$55,$C0,$87
	.dc.b	$02,$CF,$07,$C6,$62,$FD,$93,$CA,$38,$12,$78,$46,$5B,$7E,$3A,$E2
	.dc.b	$DB,$F2,$36,$FB,$9C,$E7,$6F,$86,$76,$D3,$75,$D9,$C5,$F8,$6B,$8C
	.dc.b	$73,$AE,$2D,$38,$C6,$AB,$94,$66,$B9,$57,$09,$C9,$2B,$81,$23,$4D
	.dc.b	$AE,$B1,$82,$38,$25,$76,$6D,$FC,$38,$C1,$1C,$12,$3D,$ED,$7D,$0D
	.dc.b	$BF,$35,$78,$2B,$95,$70,$21,$C0,$B3,$C1,$F1,$98,$BF,$64,$F2,$8E
	.dc.b	$04,$9E,$11,$96,$DF,$8E,$B8,$B6,$FC,$8D,$BE,$E7,$39,$DB,$E1,$9D
	.dc.b	$B4,$DD,$76,$71,$7E,$1A,$E3,$1C,$EB,$8B,$4E,$31,$AA,$E5,$19,$AE
	.dc.b	$55,$C2,$72,$4A,$E0,$48,$D3,$6B,$AC,$60,$8E,$09,$5D,$9B,$7F,$0E
	.dc.b	$30,$47,$04,$8F,$7B,$5F,$43,$6F,$CD,$5E,$0A,$E5,$5C,$08,$70,$2C
	.dc.b	$F0,$7C,$66,$2F,$D9,$3C,$A3,$81,$27,$84,$65,$B7,$E3,$AE,$2D,$BF
	.dc.b	$23,$6F,$B9,$CE,$76,$F8,$67,$6D,$37,$5D,$9C,$5F,$86,$B8,$C7,$3A
	.dc.b	$E2,$D3,$8C,$6A,$B9,$46,$6B,$95,$70,$9C,$92,$B8,$12,$34,$DA,$EB
	.dc.b	$18,$23,$82,$57,$66,$DF,$C3,$8C,$11,$C1,$23,$DE,$D7,$D0,$DB,$F3
	.dc.b	$57,$82,$B9,$57,$02,$1C,$0B,$3C,$1F,$19,$8B,$F6,$4F,$28,$E0,$49
	.dc.b	$E1,$19,$6D,$F8,$EB,$8B,$6F,$C8,$DB,$EE,$73,$9D,$BE,$19,$DB,$4D
	.dc.b	$D7,$67,$17,$E1,$AE,$31,$CE,$B8,$B4,$E3,$1A,$AE,$51,$9A,$E5,$5C
	.dc.b	$27,$24,$AE,$04,$8D,$36,$BA,$C6,$08,$E0,$95,$D9,$B7,$F0,$E3,$04
	.dc.b	$70,$48,$F7,$B5,$F4,$36,$FC,$D5,$E0,$AE,$55,$C0,$87,$02,$CF,$07
	.dc.b	$C6,$62,$FD,$93,$CA,$38,$12,$78,$46,$5B,$7E,$3A,$E2,$DB,$F2,$36
	.dc.b	$FB,$9C,$E7,$6F,$86,$76,$D3,$75,$D9,$C5,$F8,$6B,$8C,$73,$AE,$2D
	.dc.b	$38,$C6,$AB,$94,$66,$B9,$57,$09,$C9,$2B,$81,$23,$4D,$AE,$B1,$82
	.dc.b	$38,$25,$76,$6D,$FC,$38,$C1,$1C,$12,$3D,$ED,$7D,$0D,$BF,$35,$78
	.dc.b	$2B,$95,$70,$21,$C0,$B3,$C1,$F1,$98,$BF,$64,$F2,$8E,$04,$9E,$11
	.dc.b	$96,$DF,$8E,$B8,$B6,$FC,$8D,$BE,$E7,$39,$DB,$E1,$9D,$B4,$DD,$76
	.dc.b	$71,$7E,$1A,$E3,$1C,$EB,$8A,$A4,$A5,$2A,$5E,$23,$85,$6D,$D4,$66
	.dc.b	$F9,$DE,$E7,$15,$B7,$13,$A7,$47,$49,$AD,$33,$1B,$78,$CF,$37,$AC
	.dc.b	$F0,$9D,$46,$0E,$33,$B7,$DB,$1A,$FA,$E1,$00

;--------------------------------------------------------------------------------
;	FATAN.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fatan_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FATAN.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FATAN.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fatan_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fatan.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FATAN.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fatan.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FATAN.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fatan_data_compressed::
	.dc.b	$00,$02,$84,$2C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CF,$ED,$1F,$8A,$7F,$87,$A7,$FC,$D0,$5F,$8A,$E1
	.dc.b	$F0,$FC,$7D,$FE,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$1F
	.dc.b	$07,$9F,$DB,$F4,$7B,$7F,$DF,$7E,$6F,$FA,$B0,$F8,$8C,$7C,$46,$7E
	.dc.b	$23,$5F,$11,$BF,$88,$E1,$F1,$1C,$7E,$17,$97,$C0,$F3,$F8,$1E,$9F
	.dc.b	$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03
	.dc.b	$E3,$F8,$1F,$27,$C0,$F9,$7E,$07,$CD,$F0,$3E,$7F,$81,$FF,$0A,$7C
	.dc.b	$E7,$E1,$BE,$13,$FC,$40,$BE,$07,$3E,$66,$BE,$07,$7F,$03,$C3,$E0
	.dc.b	$78,$FC,$0F,$2F,$81,$E7,$F0,$3F,$63,$D2,$EB,$F0,$3D,$BE,$07,$BF
	.dc.b	$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93,$E0,$7C,$BF
	.dc.b	$03,$E6,$F8,$1F,$3F,$C0,$FA,$3E,$07,$FC,$59,$F3,$BF,$88,$E9,$FF
	.dc.b	$1C,$7C,$07,$F0,$A5,$1F,$EF,$63,$FD,$60,$7F,$9A,$0F,$EB,$47,$F1
	.dc.b	$23,$81,$FC,$41,$FF,$21,$2F,$7B,$BF,$BD,$E0,$F7,$BC,$3E,$F7,$8B
	.dc.b	$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD,$E8,$F7,$BF
	.dc.b	$3B,$EF,$FE,$7B,$5F,$CF,$EB,$FA,$0D,$7F,$43,$AF,$BB,$DE,$F7,$FB
	.dc.b	$DF,$E4,$BF,$6F,$FC,$48,$7E,$3F,$8B,$FC,$B0,$BD,$CD,$3B,$F7,$38
	.dc.b	$7B,$9C,$7D,$CE,$5E,$E7,$3F,$73,$A7,$B9,$FB,$0E,$8E,$DE,$E7,$7F
	.dc.b	$73,$C1,$EE,$7E,$D3,$AF,$F6,$BD,$7F,$B6,$EB,$FD,$BF,$5F,$EE,$3A
	.dc.b	$FF,$73,$D7,$FB,$AE,$6F,$DD,$F3,$7F,$98,$FF,$15,$FF,$1E,$1E,$9E
	.dc.b	$73,$E9,$7F,$CE,$4F,$FB,$38,$FF,$44,$1F,$DE,$0F,$E6,$07,$F2,$F7
	.dc.b	$8F,$0F,$5E,$1E,$DC,$1D,$F8,$3E,$47,$BB,$F2,$7D,$DF,$95,$EE,$FC
	.dc.b	$BF,$77,$E6,$7B,$BF,$37,$DD,$F9,$DE,$EF,$CF,$F7,$7E,$87,$BB,$FE
	.dc.b	$7B,$F9,$FF,$D4,$70,$1F,$D4,$3F,$E8,$95,$EE,$6F,$DC,$E1,$EE,$71
	.dc.b	$F7,$39,$7B,$9C,$FD,$CE,$9E,$E7,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B
	.dc.b	$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$F3,$7C,$9F,$9D,$D9,$F3
	.dc.b	$FB,$3E,$87,$67,$D1,$EC,$FA,$5D,$9F,$4F,$B3,$EA,$76,$7D,$5E,$CF
	.dc.b	$AD,$D9,$FE,$99,$F4,$FD,$E7,$BB,$B4,$FB,$9F,$F5,$42,$EB,$E1,$D7
	.dc.b	$C7,$AF,$97,$5F,$3E,$BE,$9D,$7D,$7A,$FE,$67,$B7,$F3,$7D,$BF,$9D
	.dc.b	$ED,$F8,$7B,$BE,$87,$B7,$FE,$AB,$FA,$0F,$E0,$1C,$05,$F0,$99,$F8
	.dc.b	$AD,$7C,$AE,$FE,$17,$87,$C0,$F1,$F8,$6E,$5F,$01,$CF,$E0,$3A,$74
	.dc.b	$75,$F8,$0F,$81,$E4,$FC,$1F,$27,$E1,$79,$3F,$0F,$C9,$F8,$9E,$4F
	.dc.b	$8F,$E0,$3C,$9E,$D7,$97,$DA,$F3,$7B,$5E,$7F,$6B,$D1,$ED,$7A,$7D
	.dc.b	$AF,$57,$B5,$EB,$F6,$BD,$9E,$D7,$FB,$1F,$E8,$7F,$D4,$07,$E9,$4F
	.dc.b	$D2,$76,$97,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1F,$67,$D9
	.dc.b	$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BB,$E1,$F7,$7C,$5E,$EF
	.dc.b	$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE,$F9,$FD,$DF,$47,$BB,$F8,$7F
	.dc.b	$57,$D5,$EE,$FF,$B8,$7C,$FF,$D3,$9F,$89,$3F,$11,$D2,$7F,$10,$FF
	.dc.b	$BC,$5F,$F7,$61,$FE,$A5,$3F,$CB,$AE,$FD,$9E,$9E,$CF,$5F,$67,$B7
	.dc.b	$AF,$DF,$CC,$F0,$79,$9E,$1F,$33,$C5,$E6,$78,$FC,$CF,$27,$67,$97
	.dc.b	$7F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$FB,$E7,$E8,$BF,$1E,$7E,$3B,$84
	.dc.b	$17,$B1,$BE,$3E,$1E,$C7,$1F,$63,$97,$B1,$CF,$D8,$E9,$EC,$75,$F6
	.dc.b	$3B,$7B,$1F,$AF,$DF,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E
	.dc.b	$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7A,$7D,$8F,$3F,$3B,$E8
	.dc.b	$FF,$DD,$87,$F3,$13,$F1,$A7,$E3,$1F,$6A,$F5,$F8,$BF,$EA,$83,$FC
	.dc.b	$C1,$C2,$EB,$CF,$AB,$AF,$57,$7E,$AF,$0F,$57,$8F,$AB,$CB,$D5,$E7
	.dc.b	$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E,$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F
	.dc.b	$57,$C7,$EA,$FB,$3E,$93,$F9,$99,$FC,$54,$FC,$69,$F8,$CE,$62,$F7
	.dc.b	$F8,$7B,$FC,$7D,$FE,$5E,$CF,$3F,$67,$A7,$B3,$D7,$D9,$ED,$EC,$F7
	.dc.b	$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9,$EC,$F9,$7C,$FF
	.dc.b	$37,$85,$E7,$F6,$7E,$9F,$C0,$7F,$1A,$3F,$19,$CB,$FC,$A7,$D9,$CF
	.dc.b	$B3,$AF,$67,$7E,$CF,$0F,$67,$8F,$B3,$CB,$D9,$E7,$E6,$74,$F6,$7A
	.dc.b	$FB,$3D,$BD,$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F,$67,$C7,$EC,$F9
	.dc.b	$3D,$9F,$2F,$B3,$E6,$F6,$7F,$13,$F4,$BF,$8A,$E6,$3F,$DE,$4F,$E2
	.dc.b	$A7,$E2,$5C,$FB,$5C,$3D,$DE,$3E,$EF,$2F,$6F,$9F,$B7,$D3,$DB,$EB
	.dc.b	$ED,$F6,$F6,$FB,$FB,$7E,$0F,$6F,$C3,$ED,$F8,$BD,$BF,$1F,$B7,$E4
	.dc.b	$F6,$FC,$BE,$DF,$9B,$DB,$F3,$FB,$7E,$8F,$6F,$F4,$3F,$4D,$FD,$E0
	.dc.b	$FE,$B4,$7F,$12,$3F,$0C,$7E,$17,$97,$F7,$AB,$DA,$E3,$ED,$72,$F6
	.dc.b	$B9,$FB,$5D,$3D,$AE,$BE,$D7,$6F,$6B,$BF,$B5,$E0,$F6,$BC,$3E,$D7
	.dc.b	$8B,$DA,$F1,$FB,$5E,$4F,$6B,$CB,$ED,$79,$BD,$AF,$3F,$B5,$FB,$8F
	.dc.b	$A7,$FE,$BE,$7E,$18,$FE,$1F,$C9,$FC,$49,$FE,$54,$DF,$63,$87,$B1
	.dc.b	$C7,$D8,$E5,$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1
	.dc.b	$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E
	.dc.b	$7F,$63,$F8,$7F,$AF,$F8,$73,$FC,$7C,$7F,$5A,$3F,$0C,$7F,$0D,$E4
	.dc.b	$FE,$56,$BD,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6,$F5
	.dc.b	$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC,$BE
	.dc.b	$BF,$9B,$D7,$F3,$FA,$FE,$8F,$5F,$F9,$3F,$D4,$7F,$8F,$8F,$F1,$01
	.dc.b	$FD,$60,$FE,$18,$7E,$1F,$94,$FE,$1D,$FE,$8A,$FF,$58,$1E,$AF,$4F
	.dc.b	$57,$AF,$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E
	.dc.b	$AF,$93,$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FA,$3D,$5F,$E7,$FF,$53
	.dc.b	$FE,$26,$3F,$86,$9F,$88,$E2,$3F,$88,$6F,$AB,$C3,$D5,$E3,$EA,$F2
	.dc.b	$F5,$79,$FA,$BD,$3D,$5E,$BE,$AF,$6F,$57,$BF,$AB,$E0,$F5,$7C,$3E
	.dc.b	$AF,$8B,$D5,$F1,$FA,$BE,$4F,$57,$CB,$EA,$F9,$BD,$5F,$3F,$AB,$FD
	.dc.b	$5B,$EA,$BF,$87,$1F,$E8,$53,$FC,$42,$7F,$0D,$3F,$11,$C8,$7F,$10
	.dc.b	$FF,$74,$5E,$A7,$3F,$53,$A7,$A9,$D7,$D4,$ED,$EA,$77,$F5,$3C,$1E
	.dc.b	$A7,$87,$D4,$F1,$7A,$9E,$3F,$53,$C9,$EA,$79,$7D,$4F,$37,$A9,$E7
	.dc.b	$F5,$3D,$1E,$A7,$A7,$D4,$FE,$DF,$E2,$7F,$A1,$8F,$F3,$41,$FE,$20
	.dc.b	$3F,$AA,$1F,$89,$3F,$11,$C8,$7F,$C3,$E7,$E2,$1F,$F0,$7A,$F4,$BA
	.dc.b	$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA,$5E,$2F,$4B,$C7,$E9,$79
	.dc.b	$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$A3,$D2,$FF,$03,$FD,$5F,$F9
	.dc.b	$B8,$FE,$AA,$7E,$24,$FC,$41,$E7,$E1,$FF,$11,$AF,$4B,$8F,$A5,$CB
	.dc.b	$D2,$E7,$E9,$74,$F4,$BA,$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA
	.dc.b	$5E,$2F,$4B,$C7,$E9,$79,$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$F8
	.dc.b	$7B,$EB,$3F,$AB,$1F,$EC,$D3,$FC,$D2,$7F,$55,$3F,$12,$7E,$23,$90
	.dc.b	$FE,$21,$FF,$1E,$BF,$EC,$D1,$E8,$75,$F4,$3B,$7A,$1D,$FD,$0F,$07
	.dc.b	$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F,$43,$CD,$E8,$79
	.dc.b	$FD,$0F,$47,$A1,$E9,$F4,$3F,$C6,$FF,$5B,$FE,$CE,$3F,$D6,$07,$F9
	.dc.b	$A0,$FF,$0E,$1F,$C4,$4F,$C4,$9F,$E6,$63,$F1,$1C,$47,$F1,$0F,$F9
	.dc.b	$61,$79,$FD,$BC,$FE,$FE,$7F,$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7
	.dc.b	$F9,$3C,$FF,$2F,$9F,$E6,$F3,$FC,$FE,$7F,$A3,$CF,$FF,$27,$FC,$17
	.dc.b	$FA,$D8,$FF,$0E,$9F,$C4,$4F,$E2,$07,$E2,$78,$3F,$C6,$9E,$7F,$0F
	.dc.b	$3F,$8F,$9F,$CB,$CF,$E7,$E7,$F4,$F3,$FA,$F9,$FD,$BC,$FE,$FE,$7F
	.dc.b	$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7,$F9,$3C,$FF,$2F,$9F,$E6,$F3
	.dc.b	$FF,$CC,$BF,$5D,$FE,$1D,$1F,$27,$8F,$8F,$CF,$C4,$EB,$DE,$DF,$BD
	.dc.b	$C3,$DA,$E3,$ED,$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$FE,$7C,$FA,$F7,$CB,$DF,$94,$C7,$C8,$67,$E2,$B5
	.dc.b	$E7,$6F,$E1,$B8,$7C,$37,$1F,$84,$E5,$F0,$3C,$FD,$EE,$9E,$F7,$5F
	.dc.b	$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27
	.dc.b	$BD,$E5,$F7,$BC,$DE,$F7,$FA,$4B,$EC,$3F,$DF,$27,$F9,$80,$FE,$A8
	.dc.b	$7F,$53,$3F,$08,$7F,$0B,$F7,$FF,$D5,$EB,$DE,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$FD,$53,$EC,$FF,$98
	.dc.b	$4F,$C2,$F1,$FF,$B0,$9F,$F7,$A9,$FE,$61,$3F,$AA,$9F,$C3,$4F,$E1
	.dc.b	$8F,$F8,$77,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8E,$FE,$C7,$83
	.dc.b	$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79,$3D,$8F,$2F,$B1,$E6,$F6,$3C
	.dc.b	$FE,$C7,$A3,$D8,$FF,$60,$FD,$8F,$E1,$B8,$BF,$DB,$2B,$D7,$CF,$AF
	.dc.b	$AF,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D
	.dc.b	$7F,$2F,$AF,$E6,$F5,$FC,$FE,$BF,$FB,$63,$EC,$9F,$35,$7E,$4F,$1F
	.dc.b	$1F,$9F,$8F,$D7,$C4,$EF,$E1,$B8,$7B,$BC,$7D,$DE,$5E,$EF,$3F,$77
	.dc.b	$A7,$BB,$D7,$DD,$ED,$EE,$F7,$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE
	.dc.b	$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$FE,$EE,$FB,$3F,$F5,$31,$F8
	.dc.b	$6F,$38,$BE,$17,$5F,$0B,$BF,$85,$E1,$F0,$3C,$7E,$17,$97,$C1,$73
	.dc.b	$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F,$82,$F0,$7C,$17,$87,$E0
	.dc.b	$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E,$0B,$CD,$F0,$5E,$7F,$82
	.dc.b	$F4,$7C,$13,$9E,$78,$FF,$87,$7F,$BE,$7C,$0E,$5F,$EF,$DF,$03,$67
	.dc.b	$FD,$DC,$7F,$98,$CF,$F0,$D8,$F7,$B5,$EF,$6F,$DE,$E1,$EF,$71,$F7
	.dc.b	$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE
	.dc.b	$1F,$7B,$C5,$EF,$78,$FD,$EF,$37,$DA,$7E,$23,$93,$D8,$FF,$BB,$8F
	.dc.b	$F3,$11,$FE,$61,$3F,$A9,$1F,$C3,$D3,$FC,$3D,$E3,$EC,$74,$F6,$3A
	.dc.b	$FB,$1D,$BD,$8E,$FE,$C7,$83,$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79
	.dc.b	$3D,$8F,$2F,$B1,$E6,$F6,$3C,$FE,$C7,$A3,$D8,$F9,$DF,$6B,$FE,$EF
	.dc.b	$3F,$14,$7E,$27,$87,$EB,$3F,$EE,$F3,$FC,$C6,$7F,$86,$CF,$E1,$E3
	.dc.b	$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD,$3E,$FE,$9F,$83,$D3,$F0,$FA
	.dc.b	$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F,$A7,$E6,$F4,$FC,$FE,$9F,$A3
	.dc.b	$D3,$FC,$37,$DB,$7E,$24,$FE,$1D,$C7,$F8,$D5,$E9,$EB,$E1,$37,$F0
	.dc.b	$9C,$3E,$13,$8F,$AB,$CB,$D5,$E7,$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E
	.dc.b	$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F,$57,$C7,$EA,$F9,$3D,$5F,$2F,$AB
	.dc.b	$E6,$F5,$7C,$FE,$AF,$E7,$78,$7F,$11,$CD,$FF,$00,$7E,$0B,$1F,$05
	.dc.b	$9F,$82,$D7,$C1,$6F,$DA,$E1,$F0,$5C,$7D,$AE,$5E,$D7,$3F,$6B,$A7
	.dc.b	$B5,$D7,$DA,$ED,$ED,$77,$F6,$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F
	.dc.b	$6B,$C9,$ED,$79,$7D,$AF,$37,$B5,$FB,$3F,$B7,$FD,$A7,$39,$FE,$20
	.dc.b	$7E,$1D,$C7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$D1,$EF,$7F,$0B,$FB
	.dc.b	$8F,$E1,$C7,$E1,$B9,$8F,$E1,$9F,$E2,$8B,$DD,$E1,$EE,$F1,$F3,$B9
	.dc.b	$7B,$BC,$FD,$DE,$9E,$EF,$5F,$77,$B7,$BB,$DF,$DD,$F0,$7B,$BE,$1F
	.dc.b	$77,$C5,$EE,$F8,$FD,$DF,$27,$BB,$E5,$F7,$7C,$DE,$EF,$9F,$DD,$F4
	.dc.b	$7B,$BF,$C8,$FE,$E7,$F0,$E3,$DE,$C7,$BD,$9F,$7B,$5E,$F6,$FD,$EE
	.dc.b	$1E,$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$75,$F7,$BB,$7B,$DD,$FD
	.dc.b	$EF,$07,$BD,$E1,$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD
	.dc.b	$EF,$7F,$39,$F6,$BF,$0E,$7F,$55,$3F,$0C,$3D,$DD,$7B,$BB,$F7,$78
	.dc.b	$4F,$1F,$77,$97,$BB,$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7B,$BD,$FD,$DF
	.dc.b	$07,$BB,$E1,$F7,$7C,$5E,$EF,$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE
	.dc.b	$F9,$FD,$DF,$EA,$1F,$75,$FD,$4C,$FC,$31,$F8,$5F,$78,$FE,$15,$FE
	.dc.b	$B4,$BD,$CE,$3E,$E7,$2F,$73,$9F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B
	.dc.b	$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7
	.dc.b	$9B,$DC,$F3,$FB,$9E,$8F,$73,$FB,$47,$DD,$FF,$0C,$3F,$0D,$CB,$FD
	.dc.b	$BD,$7B,$7A,$F6,$F7,$ED,$F0,$F6,$F8,$FB,$7C,$BD,$BE,$7E,$DF,$4F
	.dc.b	$6D,$FC,$03,$0B,$A7,$F7,$BF,$DD,$1E,$5F,$F9,$07,$EF,$3F,$BE,$43
	.dc.b	$CC,$C7,$C7,$FF,$7D,$8D,$BF,$E0,$39,$DC,$EA,$FA,$79,$B5,$5B,$CD
	.dc.b	$C2,$9B,$18,$F3,$66,$3D,$4F,$07,$9B,$8C,$19,$79,$B3,$E7,$9F,$46
	.dc.b	$DF,$1A,$F1,$B5,$F1,$E3,$E2,$CB,$C3,$39,$CD,$9F,$7C,$BE,$1E,$76
	.dc.b	$BE,$16,$5D,$B2,$F1,$C6,$0E,$59,$F5,$7A,$67,$67,$67,$D1,$7A,$3F
	.dc.b	$4F,$1F,$8B,$86,$23,$E4,$E1,$DF,$1F,$81,$87,$63,$D1,$9F,$AA,$BA
	.dc.b	$9F,$41,$C4,$FD,$CC,$3A,$66,$C6,$E3,$D7,$1F,$56,$3C,$AB,$F0,$51
	.dc.b	$D8,$F5,$96,$35,$97,$4C,$3D,$33,$F2,$DF,$A8,$F9,$E7,$A5,$F7,$7E
	.dc.b	$A9,$CA,$3A,$63,$F9,$2B,$E9,$1F,$57,$1F,$3B,$CE,$7E,$1D,$79,$F3
	.dc.b	$F9,$D1,$C6,$79,$9F,$99,$BE,$75,$ED,$BE,$6F,$19,$FC,$16,$1E,$58
	.dc.b	$F9,$51,$F0,$1F,$89,$9F,$CC,$8F,$D2,$65,$F3,$2F,$9E,$3F,$A4,$9E
	.dc.b	$67,$9C,$FD,$55,$71,$BF,$83,$3E,$77,$E8,$5F,$3A,$F9,$47,$28,$F3
	.dc.b	$1F,$29,$E8,$F2,$7D,$C7,$C0,$7F,$1C,$F2,$7A,$BF,$2C,$C4,$FC,$A3
	.dc.b	$D1,$1D,$4E,$EF,$F0,$4A,$F4,$47,$C4,$3E,$C1,$D0,$91,$FA,$89,$EA
	.dc.b	$7E,$31,$EA,$77,$7F,$86,$65,$F6,$1F,$4C,$FC,$E7,$E8,$4F,$E6,$5E
	.dc.b	$71,$FC,$67,$0F,$9F,$1C,$8F,$8D,$3D,$9F,$9E,$72,$2C,$FA,$23,$A3
	.dc.b	$E8,$8F,$E1,$E7,$67,$F6,$2F,$CE,$69,$F1,$1E,$71,$F6,$4F,$8C,$F7
	.dc.b	$8F,$DE,$3C,$F3,$FD,$83,$72,$FC,$74,$7F,$2C,$9E,$87,$DB,$3E,$C9
	.dc.b	$F0,$CF,$44,$FE,$40,$E9,$7F,$0A,$3F,$9E,$61,$FB,$07,$99,$F5,$F0
	.dc.b	$F3,$57,$A0,$F9,$EF,$C2,$23,$F0,$A3,$E2,$1F,$D1,$F0,$F3,$9E,$8C
	.dc.b	$7F,$07,$8F,$F3,$BA,$F8,$15,$FD,$03,$2F,$CD,$3C,$8F,$93,$87,$D5
	.dc.b	$7A,$4F,$CD,$AE,$51,$EE,$3E,$11,$FC,$E2,$3E,$55,$7F,$0E,$8E,$93
	.dc.b	$F2,$6B,$A4,$76,$AF,$C2,$3F,$26,$3E,$0B,$D2,$7F,$AA,$4F,$43,$F4
	.dc.b	$91,$D3,$0F,$7C,$79,$70,$F7,$9F,$03,$0F,$6E,$1F,$22,$FD,$67,$D0
	.dc.b	$C3,$C9,$7E,$29,$FC,$2C,$FB,$4F,$A9,$1F,$48,$FB,$8F,$B3,$2F,$B8
	.dc.b	$FF,$6A,$CB,$EE,$57,$BA,$FE,$3B,$EC,$9E,$67,$E2,$1F,$23,$F4,$EF
	.dc.b	$DB,$18,$9F,$2C,$FE,$22,$7F,$52,$79,$CF,$C3,$3C,$DA,$FD,$33,$E8
	.dc.b	$E1,$F4,$EB,$A6,$5F,$22,$7E,$16,$5F,$22,$FF,$0B,$87,$F6,$E7,$F3
	.dc.b	$27,$E8,$CF,$C1,$57,$E3,$5D,$CF,$CF,$77,$5F,$55,$D9,$F5,$4D,$9D
	.dc.b	$B0,$FA,$AF,$CC,$C3,$F8,$84,$FB,$CF,$E3,$CF,$CB,$AF,$34,$7D,$5C
	.dc.b	$3C,$CF,$D3,$AF,$33,$F7,$70,$F3,$3B,$AF,$33,$B3,$CC,$6E,$7C,$CE
	.dc.b	$DF,$31,$EC,$BF,$92,$FB,$23,$E4,$BE,$C7,$C9,$87,$B1,$FE,$1A,$7A
	.dc.b	$A3,$EC,$3F,$92,$9F,$64,$71,$7F,$0D,$87,$B1,$FC,$16,$5F,$8E,$8F
	.dc.b	$EF,$F5,$CC,$F8,$D3,$F9,$83,$8E,$1E,$E7,$8D,$FC,$47,$F2,$98,$FC
	.dc.b	$47,$71,$EE,$AF,$89,$1E,$E7,$E2,$1E,$E3,$E2,$1E,$E8,$F8,$87,$D9
	.dc.b	$3F,$A5,$1E,$E7,$F7,$C7,$99,$FC,$AD,$FE,$F2,$3F,$80,$47,$EB,$5F
	.dc.b	$8F,$3E,$87,$EA,$BE,$88,$FA,$AF,$A2,$3C,$47,$A2,$3C,$4F,$CD,$7C
	.dc.b	$53,$FA,$A8,$FD,$B4,$7F,$71,$7F,$4D,$3F,$26,$3C,$51,$FB,$33,$A1
	.dc.b	$E4,$7D,$AF,$AC,$EC,$53,$F7,$A7,$47,$DF,$9F,$F8,$72,$7F,$B8,$4F
	.dc.b	$B2,$7C,$B7,$F4,$DF,$0D,$FF,$08,$6E,$5F,$8B,$AF,$25,$FE,$99,$F5
	.dc.b	$9F,$83,$2B,$EB,$7F,$82,$1F,$CA,$F0,$F1,$57,$F4,$99,$FB,$C7,$CD
	.dc.b	$9F,$BC,$FC,$D3,$F2,$75,$E1,$BF,$CA,$57,$C2,$AE,$ED,$8F,$CA,$D7
	.dc.b	$86,$7E,$11,$E1,$9F,$E7,$4F,$F8,$F5,$FC,$A1,$F0,$8F,$BA,$7C,$2C
	.dc.b	$7F,$34,$FD,$C8,$FF,$4B,$E5,$FA,$3C,$FF,$97,$CF,$DE,$AF,$E0,$B7
	.dc.b	$E8,$8F,$CA,$DF,$D9,$3E,$B9,$E8,$2E,$1E,$88,$FA,$E7,$E1,$5F,$CB
	.dc.b	$4F,$F3,$3C,$7F,$95,$4F,$8E,$7F,$18,$71,$AE,$78,$7F,$89,$DF,$A2
	.dc.b	$7E,$12,$39,$CF,$1A,$E7,$3E,$53,$F4,$E7,$94,$B3,$D5,$E5,$97,$57
	.dc.b	$E6,$9F,$84,$AF,$26,$5F,$15,$FE,$8F,$1F,$95,$AF,$C6,$3E,$46,$BC
	.dc.b	$9F,$BA,$F8,$DF,$F1,$1B,$F9,$73,$96,$7F,$97,$79,$67,$F5,$9E,$17
	.dc.b	$E1,$7F,$29,$5E,$93,$B9,$F0,$9F,$C2,$1F,$D5,$AF,$FB,$09,$F3,$6F
	.dc.b	$FA,$B6,$7F,$C9,$A7,$F8,$8B,$F7,$5F,$DA,$9F,$04,$FA,$33,$F0,$1F
	.dc.b	$95,$5E,$0C,$7E,$43,$E1,$8E,$04,$7D,$F5,$FA,$4A,$FC,$A4,$FE,$91
	.dc.b	$FC,$A1,$FC,$9E,$3F,$8B,$1B,$8F,$8B,$1F,$9A,$7E,$2B,$B3,$EA,$9F
	.dc.b	$A0,$76,$75,$36,$7F,$1C,$8E,$B1,$F0,$CF,$EA,$0F,$C3,$3F,$6C,$FF
	.dc.b	$06,$3F,$3F,$1E,$BB,$F9,$31,$FA,$53,$FA,$71,$B8,$F9,$35,$F8,$07
	.dc.b	$F1,$F9,$FE,$B1,$FC,$46,$5F,$AC,$AE,$18,$79,$27,$F0,$D8,$76,$77
	.dc.b	$3D,$9F,$C7,$67,$FC,$24,$E8,$FD,$49,$FC,$53,$F9,$1B,$FC,$53,$F4
	.dc.b	$F3,$FC,$56,$1F,$C7,$F2,$FC,$55,$78,$EB,$AE,$3F,$02,$7D,$47,$ED
	.dc.b	$32,$EE,$F0,$8F,$24,$FE,$7F,$2F,$24,$7E,$1E,$7F,$2D,$97,$91,$FC
	.dc.b	$31,$F8,$98,$FE,$33,$8F,$F3,$A3,$F0,$17,$EB,$8F,$C0,$3E,$B7,$F0
	.dc.b	$0F,$E9,$1F,$C0,$3B,$3F,$00,$6C,$FC,$03,$F8,$67,$F6,$06,$E3,$F0
	.dc.b	$11,$F9,$28,$D9,$D4,$DC,$FA,$A3,$C6,$FA,$B1,$F9,$AF,$C6,$CB,$E6
	.dc.b	$CF,$F3,$9C,$BF,$4B,$9C,$8F,$79,$FD,$1C,$FC,$96,$1F,$E0,$4C,$7C
	.dc.b	$6F,$E6,$1F,$DA,$1F,$85,$7F,$25,$1F,$98,$3F,$24,$FE,$E9,$FC,$59
	.dc.b	$EF,$8F,$97,$87,$A6,$7E,$2B,$E9,$9F,$97,$1F,$85,$7E,$5C,$6E,$7C
	.dc.b	$73,$F2,$DF,$1C,$F5,$3D,$2F,$F5,$F3,$73,$D6,$7E,$59,$D4,$FF,$4C
	.dc.b	$DF,$E7,$5F,$E1,$78,$FE,$45,$FC,$55,$FE,$47,$3F,$F6,$3D,$7E,$40
	.dc.b	$F9,$53,$FE,$1E,$3F,$2E,$7E,$C2,$37,$3F,$91,$9D,$E1,$FD,$A7,$0F
	.dc.b	$C8,$D7,$E5,$AF,$FD,$31,$5D,$6F,$FA,$B6,$3C,$1F,$E9,$87,$F2,$A9
	.dc.b	$FE,$75,$28,$43,$0C,$2F,$BD,$7D,$43,$97,$35,$FE,$C3,$9C,$C7,$7A
	.dc.b	$FE,$73,$3F,$D9,$6F,$C1,$1F,$CA,$F0,$E3,$1E,$0B,$E2,$F8,$AB,$A6
	.dc.b	$1D,$E3,$D6,$4B,$FE,$88,$7F,$81,$CF,$A6,$F1,$3A,$47,$18,$FF,$7A
	.dc.b	$D7,$67,$E8,$1F,$83,$9F,$C7,$BF,$A7,$AF,$F5,$75,$F9,$A3,$E1,$5F
	.dc.b	$39,$F2,$E1,$EF,$9F,$EE,$37,$FA,$47,$E5,$E7,$FD,$22,$3F,$C0,$98
	.dc.b	$73,$C7,$C5,$87,$29,$F6,$3F,$77,$2F,$F1,$9E,$1C,$B5,$F1,$3F,$26
	.dc.b	$BD,$CF,$93,$2C,$E5,$FE,$0F,$C7,$AD,$7F,$86,$2B,$DF,$3F,$88,$8F
	.dc.b	$CA,$1C,$A7,$73,$EE,$E5,$F3,$69,$EF,$AF,$97,$1E,$FB,$FE,$41,$5F
	.dc.b	$4B,$3F,$E4,$15,$FE,$6E,$C3,$F0,$FC,$BC,$B9,$79,$B6,$7D,$E7,$96
	.dc.b	$BF,$0D,$7F,$6D,$FF,$7A,$5F,$E4,$39,$7C,$7C,$BE,$46,$DF,$2B,$FC
	.dc.b	$7B,$F9,$E7,$93,$E3,$D1,$CB,$EC,$F8,$1F,$6B,$67,$95,$7D,$A3,$D8
	.dc.b	$FF,$1D,$7F,$C8,$D8,$FE,$8B,$E0,$7E,$2F,$C0,$FC,$66,$D7,$2D,$FE
	.dc.b	$2E,$BE,$C3,$E2,$E1,$CB,$37,$C0,$FC,$EF,$81,$F9,$ED,$C3,$97,$47
	.dc.b	$F3,$4E,$8F,$F2,$1E,$1F,$E1,$BD,$7F,$67,$F0,$3F,$5B,$E0,$7E,$BB
	.dc.b	$70,$E2,$7F,$24,$D7,$F9,$E3,$F8,$8B,$E5,$A3,$E0,$7E,$FF,$C0,$FE
	.dc.b	$01,$B5,$CA,$BF,$90,$DF,$27,$F8,$34,$7C,$DF,$81,$FC,$3F,$C0,$FE
	.dc.b	$21,$B1,$C7,$87,$F9,$7D,$7E,$B6,$3F,$0D,$F0,$FF,$C4,$1F,$C9,$2F
	.dc.b	$93,$FD,$33,$99,$EE,$7F,$13,$75,$3D,$4F,$DC,$EE,$87,$C6,$FE,$93
	.dc.b	$87,$9B,$D2,$FA,$E1,$70,$F0,$BF,$9D,$E2,$FF,$2C,$63,$CE,$F9,$57
	.dc.b	$99,$E1,$7E,$49,$F0,$5E,$32,$FE,$53,$18,$BF,$35,$7F,$3E,$9C,$E5
	.dc.b	$E6,$9F,$A9,$AF,$F9,$BA,$FC,$11,$F1,$2F,$CE,$D3,$CE,$74,$7F,$A7
	.dc.b	$9E,$3A,$E8,$FC,$83,$F1,$0E,$2F,$C3,$1D,$0E,$05,$7F,$89,$4F,$E6
	.dc.b	$F3,$F9,$2F,$99,$B3,$C2,$3C,$6D,$3C,$58,$FC,$B8,$FF,$0F,$E1,$EC
	.dc.b	$9F,$26,$3E,$FC,$7C,$95,$FD,$0D,$FC,$06,$3F,$1B,$0F,$99,$39,$7A
	.dc.b	$5F,$F8,$7B,$0E,$95,$FC,$22,$F9,$CF,$E1,$CF,$65,$F5,$69,$FC,$20
	.dc.b	$87,$C6,$CB,$EF,$BF,$2D,$F0,$BF,$91,$C2,$E1,$F7,$63,$F9,$6F,$81
	.dc.b	$FC,$97,$3F,$8B,$97,$8F,$0D,$E5,$F9,$8B,$F8,$D3,$ED,$CB,$ED,$67
	.dc.b	$FC,$D6,$7C,$11,$F3,$CF,$C8,$1F,$81,$7E,$6B,$67,$94,$78,$A3,$EC
	.dc.b	$1C,$1F,$9F,$87,$92,$3E,$11,$F6,$27,$F0,$6D,$3F,$9A,$12,$7F,$90
	.dc.b	$DF,$CC,$7E,$34,$78,$4F,$A8,$57,$E3,$3E,$CC,$7E,$5D,$7E,$83,$1B
	.dc.b	$3F,$7D,$F8,$B9,$C8,$E0,$7C,$1B,$F8,$67,$A9,$F6,$57,$A2,$3E,$19
	.dc.b	$F3,$23,$C6,$FC,$32,$47,$D3,$8E,$8F,$8D,$F4,$67,$F6,$E3,$A1,$F3
	.dc.b	$0F,$D1,$CF,$29,$FE,$33,$7E,$A8,$FE,$28,$FA,$5F,$11,$F0,$8E,$07
	.dc.b	$E3,$B3,$FD,$04,$71,$BF,$CC,$D7,$D4,$AF,$5D,$F8,$A3,$D6,$F8,$88
	.dc.b	$F8,$8F,$59,$F2,$CF,$5E,$5F,$8F,$8E,$53,$E3,$78,$BE,$3C,$FF,$08
	.dc.b	$FD,$EC,$3D,$53,$C3,$0F,$89,$87,$E0,$EB,$FB,$34,$FC,$B7,$EA,$1F
	.dc.b	$59,$E1,$87,$86,$7D,$91,$F2,$DE,$58,$FF,$66,$7D,$55,$CA,$7E,$A3
	.dc.b	$F2,$AF,$C2,$FE,$EE,$7F,$2C,$FC,$BD,$3F,$9B,$E1,$F9,$5C,$3E,$6D
	.dc.b	$7C,$2C,$FD,$6F,$82,$B7,$97,$D1,$BF,$E0,$F8,$FF,$49,$C7,$F8,$2B
	.dc.b	$6F,$BC,$7C,$07,$D3,$9F,$C8,$8F,$D1,$9E,$7A,$F8,$0F,$9C,$FB,$73
	.dc.b	$E1,$3F,$91,$C7,$9E,$7E,$8D,$77,$AD,$9C,$73,$F9,$0F,$BE,$F8,$E1
	.dc.b	$E9,$3D,$B1,$B8,$FC,$39,$DD,$FE,$4D,$87,$78,$FD,$49,$FC,$BF,$E1
	.dc.b	$3F,$B7,$FC,$27,$F7,$06,$C7,$EC,$AB,$EE,$3F,$8D,$7E,$B3,$70,$FE
	.dc.b	$B7,$3C,$27,$FA,$93,$63,$F1,$F1,$F2,$4F,$D3,$CF,$89,$FC,$83,$FE
	.dc.b	$02,$9E,$0F,$94,$FF,$02,$D7,$F4,$0A,$FF,$4A,$D3,$56,$DF,$2B,$1F
	.dc.b	$74,$FA,$0F,$F0,$46,$3D,$1B,$8F,$43,$DD,$87,$E4,$08,$FE,$51,$F6
	.dc.b	$BB,$8F,$6B,$63,$DB,$87,$E6,$E7,$DA,$78,$4F,$C2,$C7,$27,$FB,$B6
	.dc.b	$3F,$3E,$B9,$1F,$5E,$7D,$B5,$EC,$8F,$8D,$9F,$F0,$6B,$F8,$E7,$88
	.dc.b	$E7,$1F,$24,$E6,$FE,$1C,$FE,$DF,$97,$C5,$CF,$F1,$13,$B9,$E2,$73
	.dc.b	$C3,$FC,$2B,$8F,$CF,$9F,$F6,$2C,$FC,$FB,$E5,$1F,$C4,$23,$DF,$5E
	.dc.b	$D3,$8C,$75,$6C,$7F,$22,$3F,$05,$87,$AC,$F6,$3F,$1A,$7F,$00,$F9
	.dc.b	$27,$A4,$7F,$81,$6F,$C9,$3F,$EA,$EB,$F1,$BF,$94,$3C,$73,$CA,$BC
	.dc.b	$27,$28,$FF,$25,$57,$27,$D9,$3F,$3B,$3F,$C4,$D7,$E0,$A7,$F0,$E7
	.dc.b	$8E,$7A,$1E,$33,$F1,$27,$E8,$C8,$FD,$68,$F0,$61,$FE,$55,$BF,$C0
	.dc.b	$4F,$33,$EB,$4F,$E0,$30,$F5,$BF,$CB,$4E,$38,$2B,$8C,$7B,$30,$FC
	.dc.b	$41,$F5,$B2,$FC,$46,$1F,$A1,$C7,$E6,$C7,$F2,$48,$FB,$D3,$E9,$69
	.dc.b	$F1,$1F,$5B,$F2,$88,$7D,$53,$E1,$D7,$85,$B1,$CE,$3F,$47,$AF,$FB
	.dc.b	$53,$2F,$5C,$FE,$AC,$F9,$35,$DC,$F0,$E1,$F1,$63,$C2,$FE,$17,$2F
	.dc.b	$A7,$1F,$32,$3F,$CF,$DA,$7E,$ED,$B1,$FA,$0A,$FE,$1C,$7E,$3A,$BF
	.dc.b	$9E,$9C,$9F,$64,$F0,$7E,$93,$71,$F6,$47,$2C,$BF,$79,$3E,$3D,$3F
	.dc.b	$53,$3E,$8B,$FA,$98,$7E,$7C,$F8,$47,$D4,$8F,$CB,$3F,$8D,$9E,$38
	.dc.b	$7D,$CD,$3E,$DC,$F0,$AF,$C1,$BF,$6F,$3F,$93,$1F,$8B,$8F,$93,$8F
	.dc.b	$C7,$8F,$15,$7E,$6F,$2F,$C0,$E1,$F9,$9A,$FC,$0E,$BF,$C9,$74,$F9
	.dc.b	$1A,$1F,$E4,$98,$FB,$F5,$FB,$19,$F9,$CF,$DF,$61,$F4,$EB,$89,$C3
	.dc.b	$2F,$E4,$17,$F3,$F5,$FE,$C9,$1F,$8F,$CF,$DD,$87,$C5,$AF,$C0,$47
	.dc.b	$E1,$A3,$FD,$43,$AF,$BA,$7F,$1D,$8F,$D3,$9F,$88,$D8,$E3,$1F,$35
	.dc.b	$FD,$C9,$F9,$FC,$FE,$F4,$F9,$E7,$F9,$1B,$C4,$F9,$17,$E1,$AF,$E4
	.dc.b	$B3,$EF,$7F,$85,$4F,$E4,$7E,$13,$FD,$A7,$F0,$9F,$ED,$46,$D7,$E5
	.dc.b	$B0,$FC,$6E,$7F,$72,$3F,$D3,$53,$AC,$3F,$D6,$8D,$C7,$C0,$7E,$FE
	.dc.b	$7C,$0F,$EF,$DF,$F5,$17,$C2,$7F,$B8,$BE,$13,$FD,$C6,$D8,$F0,$7C
	.dc.b	$27,$FB,$98,$FE,$C0,$7F,$BA,$3C,$9F,$F7,$4F,$93,$67,$FD,$DA,$FE
	.dc.b	$17,$7F,$F7,$6F,$14,$9F,$F7,$8C,$5D,$3F,$D9,$5C,$7F,$EF,$2F,$27
	.dc.b	$FD,$E6,$DA,$F8,$1F,$09,$FE,$F6,$3F,$DE,$FF,$09,$63,$FD,$F7,$F0
	.dc.b	$9F,$EF,$AF,$84,$FF,$7D,$B6,$3E,$47,$C2,$7F,$BF,$8F,$F7,$FF,$C2
	.dc.b	$58,$67,$FE,$ED,$FA,$84,$48,$C7,$C7,$5E,$5F,$F7,$7D,$6A,$B8,$C6
	.dc.b	$27,$E0,$46,$0E,$BD,$4A,$74,$F3,$9C,$C6,$1E,$18,$2B,$87,$C4,$EE
	.dc.b	$31,$5B,$7D,$1F,$0D,$9A,$ED,$F0,$D6,$93,$63,$CB,$5E,$FF,$86,$D1
	.dc.b	$BF,$86,$B9,$78,$B1,$F4,$7C,$37,$3F,$86,$E8,$D9,$FE,$53,$9B,$E1
	.dc.b	$BC,$07,$87,$E1,$AC,$F3,$97,$C3,$79,$7E,$1B,$CC,$D9,$E7,$F0,$DE
	.dc.b	$93,$D5,$F0,$D6,$39,$7C,$37,$B7,$E1,$BD,$CD,$8F,$1F,$C3,$7C,$13
	.dc.b	$E1,$7C,$35,$8F,$17,$C3,$7C,$7F,$A8,$7C,$8F,$C7,$8C,$3F,$1B,$B3
	.dc.b	$F0,$5C,$8A,$7C,$57,$13,$97,$C4,$5C,$7E,$44,$BE,$1F,$B7,$C3,$F7
	.dc.b	$6C,$FF,$50,$97,$C3,$78,$CF,$27,$C3,$D9,$E7,$F0,$DE,$7F,$87,$F4
	.dc.b	$36,$39,$7C,$37,$AC,$F6,$7C,$3D,$8F,$1F,$C3,$7B,$FE,$1F,$E0,$36
	.dc.b	$3C,$5F,$0D,$F0,$CF,$89,$F0,$F6,$3D,$7F,$0D,$F1,$FE,$1F,$E4,$36
	.dc.b	$3D,$5F,$0B,$F2,$CF,$99,$F0,$D6,$3E,$1F,$C2,$FC,$FF,$86,$FA,$0D
	.dc.b	$8F,$EE,$3F,$4F,$FB,$27,$F2,$0D,$F3,$1D,$7F,$4A,$3F,$D3,$FD,$67
	.dc.b	$51,$9F,$82,$FD,$14,$7F,$4F,$BE,$71,$8F,$85,$FB,$91,$FB,$49,$FE
	.dc.b	$D7,$87,$68,$C7,$C3,$72,$F8,$4E,$6F,$4F,$85,$EB,$F0,$96,$BC,$EF
	.dc.b	$4B,$FD,$C6,$0F,$85,$F1,$9E,$4F,$85,$B7,$EE,$F8,$5F,$3F,$C2,$FA
	.dc.b	$1B,$1F,$13,$E1,$7D,$67,$B3,$E1,$6C,$7C,$6F,$84,$F7,$FC,$2F,$C0
	.dc.b	$6C,$7C,$CF,$85,$F8,$67,$C4,$F8,$5B,$1F,$3B,$E1,$3E,$39,$F2,$3E
	.dc.b	$16,$C7,$D4,$F8,$5F,$96,$7C,$CF,$85,$B1,$C4,$9C,$5F,$24,$F8,$35
	.dc.b	$C9,$FF,$39,$7C,$37,$F8,$73,$8D,$87,$29,$E1,$D7,$F6,$F1,$C7,$5F
	.dc.b	$DF,$CF,$93,$F8,$1C,$AE,$7F,$DC,$F3,$FF,$08,$BF,$6F,$2E,$2F,$DD
	.dc.b	$C2,$E5,$F5,$31,$E2,$F9,$A7,$D3,$8F,$F6,$67,$FC,$44,$7E,$4A,$F8
	.dc.b	$E5,$E5,$8C,$1C,$0F,$C5,$5F,$C9,$71,$7F,$E2,$C3,$FC,$66,$7B,$B3
	.dc.b	$F8,$31,$CA,$31,$8F,$69,$57,$9A,$FF,$3D,$9F,$A2,$FF,$C5,$18,$7F
	.dc.b	$86,$8E,$59,$7E,$97,$2F,$AD,$18,$3C,$A7,$92,$F9,$BE,$6A,$E6,$7F
	.dc.b	$85,$0F,$34,$F8,$F2,$F2,$63,$E3,$C9,$97,$8F,$3F,$F1,$99,$F7,$6F
	.dc.b	$86,$3E,$1C,$7E,$83,$E1,$BE,$2F,$86,$B8,$F0,$FF,$90,$AB,$C3,$3C
	.dc.b	$73,$FA,$11,$83,$E5,$65,$FD,$7F,$AB,$FB,$94,$62,$BF,$AC,$46,$35
	.dc.b	$FB,$93,$F4,$9F,$8D,$3D,$72,$6B,$EE,$8E,$27,$D1,$BE,$4F,$E3,$F8
	.dc.b	$BE,$B6,$BF,$3B,$1F,$26,$3F,$97,$AE,$27,$CF,$BE,$33,$F8,$33,$FC
	.dc.b	$93,$3E,$27,$8C,$FE,$5C,$F2,$65,$ED,$7C,$47,$F0,$DC,$FF,$90,$61
	.dc.b	$F3,$AB,$89,$F3,$8F,$D5,$E1,$E0,$9F,$E4,$D7,$CE,$FF,$78,$7C,$1C
	.dc.b	$BE,$05,$F0,$8E,$E6,$9E,$B5,$F4,$B1,$F4,$1E,$F8,$E1,$3E,$F3,$CD
	.dc.b	$9F,$E7,$A3,$8B,$F3,$A3,$DA,$7C,$68,$FB,$07,$23,$F8,$3E,$9F,$83
	.dc.b	$AE,$F3,$FB,$D3,$FA,$2E,$1E,$88,$FE,$EB,$97,$BF,$0F,$D4,$D7,$59
	.dc.b	$FC,$0E,$1F,$42,$39,$3E,$4B,$E5,$1F,$E3,$6B,$E5,$97,$C7,$AF,$1C
	.dc.b	$FE,$BD,$F5,$E7,$F8,$1C,$3C,$B1,$F4,$5F,$1C,$FC,$A8,$E2,$7E,$B6
	.dc.b	$37,$9F,$DA,$9F,$59,$F1,$27,$E1,$4F,$CD,$39,$C7,$D3,$78,$E5,$F2
	.dc.b	$E7,$EA,$E3,$F7,$CF,$4D,$7C,$13,$F3,$07,$E7,$F0,$F4,$3F,$0A,$FA
	.dc.b	$D7,$82,$3F,$83,$B4,$FC,$C6,$BF,$CB,$9F,$EB,$B5,$F9,$1A,$F5,$DF
	.dc.b	$86,$FE,$04,$7D,$0B,$E1,$8F,$BF,$0F,$A1,$5F,$BD,$BE,$0F,$E2,$4F
	.dc.b	$71,$E5,$3F,$49,$1E,$E7,$EA,$5F,$C5,$7E,$C1,$F9,$0C,$3C,$8F,$19
	.dc.b	$FF,$35,$57,$67,$F1,$F5,$F0,$0F,$36,$BF,$BC,$AF,$A7,$3F,$33,$E1
	.dc.b	$BF,$87,$7C,$37,$F0,$F6,$CF,$F9,$DA,$7F,$94,$DF,$FA,$B5,$FE,$C4
	.dc.b	$3F,$75,$FE,$02,$D9,$FF,$3D,$39,$AB,$3F,$65,$EB,$1F,$BA,$6D,$79
	.dc.b	$AB,$C1,$5D,$09,$3C,$A3,$CC,$7F,$22,$BB,$7F,$C1,$1F,$F0,$B3,$E3
	.dc.b	$8E,$67,$8A,$79,$3C,$49,$1E,$36,$DF,$8E,$3D,$F7,$E3,$9E,$47,$43
	.dc.b	$67,$EB,$6F,$C7,$5F,$B8,$7C,$73,$C6,$7D,$B7,$E2,$BF,$44,$72,$7E
	.dc.b	$DC,$7C,$48,$FC,$EC,$78,$9E,$17,$FE,$B9,$8F,$9D,$1E,$37,$99,$F5
	.dc.b	$B4,$FE,$2D,$3E,$39,$FB,$91,$87,$AC,$73,$BF,$35,$7C,$0B,$EF,$5E
	.dc.b	$D3,$8E,$7E,$59,$FF,$66,$E1,$CD,$FD,$5E,$5C,$1F,$F6,$8E,$1F,$6A
	.dc.b	$FC,$D8,$F8,$E7,$CD,$87,$35,$87,$37,$F2,$35,$FE,$18,$7E,$13,$C8
	.dc.b	$FF,$6E,$BF,$08,$C9,$D1,$F3,$9F,$35,$EB,$3F,$52,$3C,$D1,$C6,$BE
	.dc.b	$14,$7E,$6A,$BC,$47,$B7,$46,$9F,$ED,$5A,$FA,$53,$AE,$0F,$E7,$97
	.dc.b	$E4,$84,$63,$0F,$87,$3E,$0C,$9A,$78,$AF,$BE,$7F,$0D,$E9,$5F,$CD
	.dc.b	$71,$EF,$8F,$3C,$7F,$3C,$7C,$08,$FB,$58,$73,$C3,$D5,$C1,$EB,$6E
	.dc.b	$1F,$82,$C3,$96,$3E,$68,$EC,$7C,$CC,$BF,$07,$8F,$48,$FA,$19,$7E
	.dc.b	$03,$0E,$6F,$CF,$3F,$21,$8F,$CD,$9D,$C7,$CD,$7A,$1E,$33,$F2,$71
	.dc.b	$63,$8C,$79,$4F,$1B,$F8,$EC,$7E,$06,$1E,$49,$F3,$65,$F2,$6F,$C7
	.dc.b	$8F,$D4,$C6,$D7,$E7,$8F,$09,$E4,$BF,$A2,$F9,$B2,$FE,$A1,$5F,$72
	.dc.b	$FC,$8D,$9F,$26,$5F,$D7,$DF,$BD,$97,$91,$FB,$99,$78,$DF,$69,$F9
	.dc.b	$29,$F6,$9F,$DA,$2B,$F9,$61,$F9,$08,$F2,$4F,$19,$FF,$0F,$FC,$3F
	.dc.b	$A9,$FE,$7E,$FC,$D7,$F0,$13,$F0,$8F,$56,$1F,$E4,$99,$E1,$AF,$E6
	.dc.b	$1E,$18,$77,$BF,$DF,$5F,$03,$FC,$9B,$3C,$35,$3C,$23,$E4,$CF,$03
	.dc.b	$C2,$79,$1E,$27,$C0,$7D,$2F,$C8,$AF,$7B,$F2,$0F,$BB,$1F,$9A,$8F
	.dc.b	$21,$F8,$F7,$AD,$FA,$67,$AC,$7C,$A9,$FF,$12,$D7,$96,$3F,$5B,$3F
	.dc.b	$CB,$8F,$CB,$D7,$18,$F1,$9C,$63,$FC,$49,$3C,$63,$C5,$1F,$2D,$F0
	.dc.b	$DF,$C2,$8F,$F3,$8F,$C3,$FC,$D6,$9E,$D3,$F8,$B6,$1F,$B6,$7D,$BA
	.dc.b	$FE,$5E,$FF,$91,$E7,$FC,$97,$2F,$25,$FF,$8F,$B0,$E9,$5F,$E5,$E7
	.dc.b	$C9,$7F,$CA,$63,$C7,$97,$FA,$16,$77,$5C,$71,$F1,$57,$FA,$0D,$F9
	.dc.b	$CF,$E5,$CF,$8A,$7D,$83,$F6,$86,$23,$FC,$9F,$F0,$FF,$E9,$1F,$87
	.dc.b	$FF,$49,$36,$BA,$D7,$F9,$3C,$FD,$A3,$FB,$6C,$BF,$92,$FC,$3F,$FA
	.dc.b	$64,$FF,$4D,$7C,$35,$9F,$F4,$CF,$C3,$FF,$A7,$7E,$1F,$FD,$3C,$D8
	.dc.b	$EB,$C7,$FE,$33,$C7,$FA,$1F,$1F,$F2,$9C,$BF,$98,$FC,$3F,$FA,$94
	.dc.b	$FF,$53,$7C,$2D,$C3,$FD,$4B,$F0,$DF,$EA,$9F,$86,$FF,$55,$36,$3A
	.dc.b	$F4,$FF,$3D,$C7,$A7,$3B,$E1,$BF,$D6,$07,$FA,$C7,$E1,$6D,$FF,$AC
	.dc.b	$3E,$1B,$FD,$69,$F0,$DF,$EB,$56,$C7,$42,$6F,$FE,$96,$F8,$6F,$F5
	.dc.b	$D1,$FE,$BB,$F8,$6B,$29,$74,$A3,$1C,$4F,$A7,$7C,$3E,$6F,$75,$A9
	.dc.b	$7C,$3E,$70,$D6,$1C,$70,$C5,$EE,$BF,$D5,$D6,$F8,$7C,$E2,$C3,$C1
	.dc.b	$18,$7C,$5F,$0D,$6B,$86,$1E,$9C,$BB,$C3,$E1,$F5,$F0,$DB,$6D,$7E
	.dc.b	$FB,$AB,$EE,$1F,$D3,$F3,$FC,$A6,$5E,$FF,$86,$EC,$77,$F8,$5B,$7D
	.dc.b	$CF,$81,$F0,$DE,$3F,$86,$F2,$36,$3B,$75,$7D,$D7,$F2,$9A,$BE,$1B
	.dc.b	$D4,$7A,$FE,$16,$CF,$5D,$3E,$15,$30,$E9,$1F,$40,$CB,$8E,$0F,$E3
	.dc.b	$71,$FA,$F3,$F4,$67,$B2,$7C,$15,$D7,$5F,$B9,$8F,$86,$31,$3A,$AE
	.dc.b	$D5,$F8,$1C,$3B,$57,$C0,$37,$9F,$F7,$8C,$FD,$D7,$DF,$1F,$5B,$CB
	.dc.b	$16,$78,$CF,$EB,$FC,$37,$DB,$F8,$6F,$B8,$DC,$FE,$49,$F9,$A9,$FB
	.dc.b	$BF,$0D,$F8,$13,$F0,$5F,$0B,$67,$D3,$5C,$72,$F0,$36,$FE,$2C,$7F
	.dc.b	$07,$3F,$77,$3F,$52,$F9,$47,$9E,$ED,$79,$AB,$E2,$E1,$E6,$69,$E6
	.dc.b	$AF,$F2,$F7,$81,$FC,$17,$36,$3C,$AB,$F4,$D8,$DB,$FB,$BF,$0D,$F9
	.dc.b	$BF,$86,$FC,$E3,$6B,$A6,$7F,$CE,$6B,$ED,$7C,$37,$E8,$4F,$D1,$7C
	.dc.b	$2D,$AF,$B3,$7C,$6B,$E8,$36,$FE,$0F,$3F,$E8,$35,$F3,$63,$C6,$FE
	.dc.b	$B6,$36,$FF,$57,$87,$1C,$7F,$7C,$DA,$FC,$71,$F2,$F1,$FD,$55,$7E
	.dc.b	$4B,$97,$F8,$46,$3C,$F0,$FE,$0D,$8D,$C3,$F3,$BF,$0D,$FB,$DF,$A7
	.dc.b	$FE,$25,$5F,$92,$8C,$4F,$E3,$63,$F0,$47,$F1,$49,$7C,$57,$13,$97
	.dc.b	$C4,$59,$E4,$7E,$7E,$7A,$D7,$E4,$EF,$AC,$FC,$E3,$FA,$AE,$7E,$9A
	.dc.b	$EB,$1C,$70,$EA,$D3,$AD,$FE,$9A,$7A,$D7,$F1,$2B,$EB,$3F,$D7,$CF
	.dc.b	$8D,$B7,$F4,$9A,$EB,$1C,$70,$EA,$DC,$3D,$1E,$47,$C5,$F8,$7F,$8C
	.dc.b	$D8,$F4,$4F,$EF,$B2,$F4,$7C,$47,$CB,$3E,$67,$C4,$5A,$E6,$7C,$3C
	.dc.b	$39,$B4,$E7,$7F,$A4,$3F,$91,$E4,$BE,$51,$EF,$BB,$87,$F3,$EA,$FF
	.dc.b	$49,$DF,$3A,$F8,$98,$F9,$DF,$85,$C3,$F9,$19,$FE,$ED,$5D,$5F,$AD
	.dc.b	$8F,$A1,$A7,$D9,$F8,$8F,$C3,$7C,$47,$E1,$DB,$1E,$8B,$FC,$8B,$FE
	.dc.b	$20,$3F,$0D,$F1,$1F,$8D,$3F,$1D,$F0,$F6,$3A,$1F,$85,$BE,$8D,$3A
	.dc.b	$70,$7D,$7A,$E1,$9F,$DF,$BB,$AF,$C1,$CB,$C7,$3F,$83,$6C,$7E,$C3
	.dc.b	$BF,$F9,$2B,$F8,$DB,$E3,$8F,$E6,$F0,$B8,$7F,$61,$D5,$F1,$1F,$A5
	.dc.b	$CF,$F4,$CD,$BF,$29,$F2,$B2,$FD,$51,$FA,$BF,$87,$B1,$FA,$19,$E1
	.dc.b	$87,$E5,$DB,$9F,$96,$7F,$5B,$3E,$1B,$E2,$FE,$76,$ED,$7D,$FB,$E3
	.dc.b	$7F,$B3,$6D,$7F,$04,$DB,$EF,$3F,$8A,$C3,$8D,$7E,$E7,$0B,$3E,$FF
	.dc.b	$88,$FE,$05,$F1,$1F,$C0,$DB,$5E,$E9,$FC,$9F,$C4,$7F,$07,$3F,$84
	.dc.b	$7C,$3D,$9F,$DF,$57,$0A,$FD,$A3,$74,$F7,$47,$F3,$69,$FC,$8D,$F1
	.dc.b	$8F,$DD,$5D,$AF,$CE,$5F,$1B,$FE,$20,$DA,$F7,$65,$F7,$27,$F2,$11
	.dc.b	$F6,$70,$E5,$1F,$C5,$F0,$B3,$F7,$7E,$23,$F9,$07,$C3,$7F,$21,$6D
	.dc.b	$7B,$F2,$FE,$25,$F0,$DF,$C9,$4F,$F1,$17,$D3,$FF,$8D,$67,$F8,$E4
	.dc.b	$71,$7F,$8F,$5D,$C3,$F8,$A3,$FC,$A5,$F8,$07,$EF,$27,$A4,$FF,$80
	.dc.b	$EB,$A4,$FF,$1B,$9E,$57,$FB,$DB,$B7,$FC,$C5,$E9,$1F,$D5,$4F,$3C
	.dc.b	$79,$0E,$87,$6A,$FF,$05,$DF,$FB,$79,$BF,$11,$EC,$F8,$8F,$6B,$FE
	.dc.b	$EC,$F5,$A5,$F9,$7E,$1F,$E0,$9F,$0B,$E1,$EC,$7A,$6F,$8B,$4F,$65
	.dc.b	$F2,$C3,$F9,$97,$C4,$31,$E3,$1E,$4C,$7A,$5F,$73,$94,$79,$9B,$9F
	.dc.b	$F9,$73,$36,$3D,$63,$FC,$1B,$8F,$5A,$E5,$7D,$5F,$9D,$9F,$82,$B9
	.dc.b	$7C,$3F,$DC,$F8,$8F,$BA,$DA,$FF,$5B,$E9,$D5,$FF,$39,$E6,$F8,$8F
	.dc.b	$C1,$9F,$84,$F8,$8B,$3F,$6E,$F9,$67,$F4,$1B,$97,$9B,$1E,$93,$FE
	.dc.b	$6F,$C7,$A5,$FE,$0F,$1E,$8D,$3A,$1F,$87,$CB,$CE,$D3,$CE,$4A,$FB
	.dc.b	$F3,$E4,$7F,$03,$95,$C7,$E5,$7C,$3F,$E7,$3E,$23,$F3,$AD,$C3,$94
	.dc.b	$F9,$DF,$CB,$7C,$47,$E8,$8F,$D1,$FC,$45,$8F,$CD,$57,$1C,$7F,$1E
	.dc.b	$DB,$E5,$1F,$82,$D7,$83,$F9,$4C,$39,$DF,$63,$99,$FA,$5C,$7C,$AD
	.dc.b	$3C,$A7,$F0,$09,$70,$7E,$05,$E9,$1C,$DF,$D0,$E5,$E0,$C7,$F1,$5F
	.dc.b	$0F,$FB,$CF,$88,$FD,$EB,$6F,$AD,$FF,$8B,$B0,$E9,$7F,$91,$F8,$8F
	.dc.b	$E0,$67,$F0,$4F,$88,$B7,$FB,$BC,$39,$61,$FB,$06,$DF,$C2,$24,$FE
	.dc.b	$32,$F8,$D7,$ED,$AE,$D7,$9A,$BF,$83,$63,$E6,$69,$E6,$AF,$F3,$66
	.dc.b	$3E,$68,$E8,$7F,$0D,$C3,$CC,$D3,$F8,$87,$C3,$FF,$1C,$F8,$8F,$E3
	.dc.b	$AD,$8E,$9E,$1F,$FB,$C6,$FA,$61,$F9,$7F,$88,$FE,$46,$7F,$24,$E1
	.dc.b	$B7,$D0,$FE,$39,$8F,$46,$9D,$3C,$1F,$F6,$7D,$FD,$8A,$E2,$FF,$16
	.dc.b	$BE,$95,$FC,$95,$E5,$1F,$C6,$AE,$EB,$FC,$0D,$FD,$5C,$7C,$2B,$E7
	.dc.b	$3F,$C1,$EB,$9C,$F5,$BF,$F5,$97,$C2,$FF,$BB,$1B,$F0,$FF,$ED,$F8
	.dc.b	$C4,$FF,$AC,$9F,$F7,$6C,$FA,$4F,$F6,$34,$7A,$5E,$7F,$0F,$C4,$E5
	.dc.b	$36,$3B,$FC,$47,$5B,$EC,$D8,$E9,$7F,$EC,$D3,$A1,$E1,$BE,$4F,$8E
	.dc.b	$EE,$3F,$EF,$E7,$CE,$DC,$3A,$DF,$B0,$EB,$05,$53,$5F,$BB,$EC,$B0
	.dc.b	$FF,$7E,$CF,$86,$BE,$3C,$78,$1E,$33,$BF,$87,$F9,$1F,$11,$F2,$5B
	.dc.b	$3E,$03,$BD,$FB,$31,$F4,$7C,$47,$CF,$3E,$87,$C4,$5A,$FA,$B7,$CA
	.dc.b	$FE,$93,$6F,$F5,$31,$FB,$5E,$0E,$0F,$D7,$C3,$9D,$FD,$E7,$C8,$FE
	.dc.b	$03,$1B,$3F,$75,$FC,$4F,$1F,$E7,$EF,$9E,$1E,$6B,$FC,$3E,$58,$7E
	.dc.b	$0E,$8F,$88,$FC,$47,$C4,$7E,$25,$B3,$F0,$73,$F9,$D8,$FE,$37,$E2
	.dc.b	$24,$FE,$4B,$E2,$2D,$7E,$3E,$F8,$B6,$7F,$2F,$87,$9B,$9F,$F5,$3C
	.dc.b	$5F,$63,$36,$5C,$B5,$FC,$26,$77,$1F,$3E,$7D,$27,$F1,$0D,$8F,$E5
	.dc.b	$58,$7E,$FB,$3F,$D5,$47,$E9,$B0,$E5,$0E,$2F,$C3,$63,$87,$F1,$1F
	.dc.b	$11,$FA,$9F,$88,$FD,$53,$70,$F8,$57,$F8,$17,$E4,$FC,$3F,$EB,$CF
	.dc.b	$D8,$7C,$3D,$8F,$DA,$D7,$1A,$FD,$93,$75,$FC,$0C,$79,$8F,$DB,$5F
	.dc.b	$17,$F7,$17,$67,$F4,$D5,$C6,$FE,$CB,$6B,$F1,$51,$FD,$03,$1F,$D1
	.dc.b	$61,$C6,$3E,$FE,$16,$BE,$16,$CF,$86,$FE,$11,$F0,$FF,$C2,$5B,$7F
	.dc.b	$C1,$B7,$FA,$DF,$0F,$F6,$39,$3F,$87,$3F,$C3,$FE,$1E,$57,$F1,$4F
	.dc.b	$86,$B9,$B8,$3F,$70,$FD,$1F,$A7,$FF,$6F,$D7,$EC,$9B,$19,$F8,$7D
	.dc.b	$D7,$0E,$15,$6A,$9F,$BB,$3B,$BF,$99,$AE,$DB,$18,$79,$E5,$DF,$E2
	.dc.b	$6D,$F8,$DB,$8F,$6F,$87,$B9,$26,$FC,$4C,$C3,$F8,$8E,$9E,$5A,$FF
	.dc.b	$69,$63,$EF,$9F,$65,$72,$F8,$9E,$7F,$0F,$67,$89,$F0,$5E,$FF,$11
	.dc.b	$23,$37,$F1,$FE,$23,$0F,$53,$E5,$7C,$47,$98,$F3,$FC,$35,$AE,$3B
	.dc.b	$FC,$8F,$87,$F5,$F8,$9E,$C6,$CE,$B6,$7C,$3F,$C0,$3E,$0F,$C3,$59
	.dc.b	$D7,$C3,$FC,$4F,$87,$F8,$AD,$8E,$5F,$0F,$F2,$0F,$93,$F0,$F6,$3B
	.dc.b	$7C,$37,$CC,$3E,$6F,$C3,$D8,$EA,$78,$F0,$EA,$D3,$A9,$38,$7E,$16
	.dc.b	$FF,$84,$C3,$B4,$63,$56,$3D,$63,$18,$DE,$7F,$C6,$7D,$47,$FA,$61
	.dc.b	$B9,$7E,$0A,$3F,$D1,$D1,$88,$F9,$71,$FE,$78,$6F,$C3,$EF,$E1,$F8
	.dc.b	$3C,$7E,$1E,$4F,$0F,$87,$B9,$7A,$30,$7C,$37,$73,$C1,$F0,$D6,$FB
	.dc.b	$FC,$37,$8F,$E1,$BC,$8D,$8F,$1F,$C2,$F9,$FE,$1B,$D0,$D8,$F3,$FC
	.dc.b	$37,$AC,$F6,$7C,$35,$8F,$5F,$C2,$FB,$FE,$17,$E0,$36,$3D,$FF,$0D
	.dc.b	$F0,$FE,$1B,$E2,$36,$3A,$93,$83,$D9,$87,$57,$DD,$7C,$DB,$3F,$2B
	.dc.b	$0E,$6F,$CC,$C2,$CF,$C5,$79,$72,$FD,$6C,$BD,$35,$DE,$3E,$CE,$1D
	.dc.b	$5B,$1D,$4F,$49,$D4,$B7,$D5,$FB,$7F,$0D,$F7,$CF,$C0,$7C,$35,$8F
	.dc.b	$8A,$73,$9F,$49,$F6,$72,$FC,$37,$A9,$FC,$23,$EA,$3F,$C3,$35,$FE
	.dc.b	$0C,$E5,$FF,$09,$C6,$2F,$FA,$84,$31,$ED,$1F,$83,$C6,$E5,$EA,$8E
	.dc.b	$D8,$78,$9A,$7E,$27,$1F,$07,$C3,$F8,$A7,$F7,$F9,$77,$EF,$67,$DE
	.dc.b	$7D,$E7,$E6,$B2,$EE,$F9,$F5,$F9,$4D,$3E,$06,$7F,$A6,$CB,$BC,$7E
	.dc.b	$89,$B9,$FE,$CB,$2F,$F2,$6B,$7E,$1B,$E2,$7C,$3F,$C5,$7F,$1B,$7D
	.dc.b	$0E,$D3,$EC,$C3,$F9,$F5,$FF,$8E,$67,$D7,$E9,$FC,$DF,$73,$E7,$3C
	.dc.b	$E3,$DF,$87,$66,$9F,$1E,$FF,$7B,$C3,$E0,$C7,$B4,$FF,$90,$31,$ED
	.dc.b	$1E,$9C,$3B,$4F,$80,$E7,$1E,$B7,$E6,$E5,$E1,$8F,$C1,$E1,$DA,$7C
	.dc.b	$07,$E0,$70,$E7,$3D,$63,$EA,$E5,$EA,$7E,$BF,$A9,$F8,$93,$F1,$5F
	.dc.b	$0F,$63,$9D,$7E,$22,$FD,$58,$77,$8F,$A1,$F0,$DF,$91,$F8,$5F,$C9
	.dc.b	$36,$FA,$92,$7E,$B6,$1D,$5F,$4C,$79,$CE,$FC,$7F,$88,$C7,$9C,$7E
	.dc.b	$33,$0B,$5F,$7D,$E5,$8F,$D8,$CB,$D3,$5D,$E3,$F3,$47,$9C,$AF,$7A
	.dc.b	$F4,$BD,$4F,$CC,$E5,$E9,$8F,$C6,$7C,$37,$EB,$4F,$D7,$7C,$35,$AF
	.dc.b	$3F,$2B,$3F,$8B,$1D,$71,$FC,$5F,$C3,$7E,$D7,$E1,$7F,$6C,$D9,$FD
	.dc.b	$3E,$3C,$F2,$FC,$D6,$57,$0F,$3C,$7D,$1A,$EF,$8F,$27,$F5,$F8,$D9
	.dc.b	$FD,$13,$C6,$3F,$1D,$87,$A1,$EE,$72,$24,$FE,$A7,$1F,$23,$63,$A9
	.dc.b	$FB,$CF,$86,$FE,$1D,$F0,$DF,$C3,$DB,$1C,$B2,$F3,$F0,$7D,$98,$FE
	.dc.b	$07,$F0,$DF,$C5,$7E,$17,$F8,$B3,$63,$A1,$FC,$23,$0E,$8D,$3A,$1F
	.dc.b	$B6,$BF,$B4,$FE,$A6,$79,$63,$FC,$05,$B3,$FB,$AC,$7C,$B3,$FC,$15
	.dc.b	$B1,$E6,$DF,$EC,$5F,$6A,$FF,$3A,$73,$B0,$E7,$3D,$CF,$E3,$B9,$7F
	.dc.b	$2F,$F8,$6F,$E6,$1F,$0B,$67,$F9,$0B,$F8,$2C,$3E,$C1,$F9,$E2,$E5
	.dc.b	$FC,$3B,$E1,$7F,$9C,$7C,$2F,$F3,$96,$DF,$F2,$AC,$3A,$B4,$EA,$79
	.dc.b	$F7,$FE,$5B,$DD,$FC,$A3,$1F,$4D,$F2,$7F,$92,$37,$1F,$E3,$B8,$74
	.dc.b	$9F,$E5,$0D,$8F,$89,$C9,$E9,$C3,$ED,$57,$2C,$7F,$9E,$E5,$E7,$69
	.dc.b	$FE,$3E,$F3,$BF,$CD,$34,$5F,$0B,$FE,$5C,$72,$62,$7D,$5C,$2D,$7D
	.dc.b	$57,$C7,$E0,$F8,$7C,$27,$16,$DF,$7C,$FA,$65,$D5,$B1,$D3,$0E,$79
	.dc.b	$7F,$9A,$63,$CD,$87,$36,$9C,$CF,$55,$F3,$69,$CC,$FE,$91,$7F,$EB
	.dc.b	$9A,$E8,$FF,$90,$95,$75,$8C,$3F,$07,$E1,$25,$7C,$7F,$8B,$B5,$F8
	.dc.b	$5C,$FE,$7E,$FF,$4A,$BE,$67,$C2,$FC,$9F,$85,$F9,$4D,$8F,$A1,$8F
	.dc.b	$2A,$FF,$4B,$67,$FE,$82,$6C,$7D,$BD,$FF,$3F,$97,$76,$CF,$0C,$3A
	.dc.b	$4F,$AD,$B1,$CB,$0E,$E7,$A4,$B3,$CF,$87,$B1,$E6,$8F,$C2,$65,$67
	.dc.b	$E3,$3F,$83,$F8,$59,$3F,$89,$F8,$5B,$5F,$2F,$E1,$7F,$11,$F0,$BF
	.dc.b	$89,$6D,$7E,$3F,$E1,$3F,$19,$F0,$BF,$8D,$6C,$7E,$4F,$1E,$7A,$7C
	.dc.b	$4C,$AC,$B2,$FB,$58,$F4,$7C,$11,$F9,$BC,$BA,$3F,$9D,$CA,$CF,$DF
	.dc.b	$79,$BF,$83,$C7,$F3,$D8,$63,$87,$F4,$58,$F5,$CB,$EA,$B7,$1F,$14
	.dc.b	$FE,$9F,$3F,$4E,$3F,$8B,$6D,$FD,$CF,$85,$FD,$51,$FA,$BF,$85,$B8
	.dc.b	$7E,$C7,$1E,$98,$7B,$23,$D4,$7E,$D7,$E1,$7F,$65,$F0,$9F,$B3,$6C
	.dc.b	$7E,$E7,$0E,$75,$F9,$BC,$2E,$3E,$8A,$E7,$3F,$18,$FD,$06,$FC,$E3
	.dc.b	$F8,$16,$36,$7F,$48,$F2,$8F,$7F,$2F,$FA,$F9,$FE,$11,$86,$6B,$D0
	.dc.b	$74,$8F,$61,$E0,$C3,$F3,$B9,$7A,$27,$F6,$0D,$8F,$D0,$FC,$2F,$F0
	.dc.b	$F3,$F8,$87,$C2,$DB,$E7,$7D,$4F,$55,$7E,$A3,$3F,$E2,$DF,$09,$FC
	.dc.b	$5F,$E1,$2D,$FF,$1D,$79,$5F,$EF,$F1,$EB,$3F,$C8,$5F,$CB,$3E,$73
	.dc.b	$DF,$BF,$D4,$E1,$FE,$49,$87,$59,$FE,$50,$DB,$E9,$1C,$CF,$E5,$99
	.dc.b	$F8,$33,$FA,$53,$E9,$69,$E9,$9E,$67,$9C,$B1,$FC,$57,$24,$7F,$BD
	.dc.b	$3E,$A3,$FD,$A0,$FE,$49,$B2,$C5,$1F,$ED,$29,$D1,$EE,$9F,$13,$EE
	.dc.b	$3E,$03,$C6,$3F,$DA,$CF,$FB,$99,$FE,$89,$1E,$1F,$85,$F1,$CE,$78
	.dc.b	$2E,$38,$C3,$C3,$78,$3C,$27,$A7,$E2,$F1,$7D,$E3,$C9,$5F,$0E,$3E
	.dc.b	$EC,$7E,$43,$7D,$DF,$B3,$1B,$97,$92,$FC,$D1,$8C,$B3,$84,$3C,$07
	.dc.b	$C9,$3B,$9D,$4F,$C6,$E3,$E1,$3D,$57,$F0,$9B,$3E,$0A,$ED,$5F,$4F
	.dc.b	$E1,$BC,$67,$93,$E1,$AC,$74,$3B,$63,$EC,$3F,$A9,$67,$EA,$F8,$6F
	.dc.b	$5F,$C3,$5A,$FB,$CF,$2A,$FA,$58,$75,$9F,$01,$D6,$39,$B7,$5F,$EF
	.dc.b	$58,$73,$9F,$B5,$77,$4F,$6B,$D5,$F7,$E1,$D6,$7B,$E1,$CF,$46,$1F
	.dc.b	$22,$7B,$9F,$85,$CB,$D3,$5F,$95,$6C,$7E,$27,$E1,$7E,$BF,$C3,$7D
	.dc.b	$86,$D7,$5B,$E9,$A7,$A2,$3F,$3D,$F0,$BF,$7B,$E1,$BE,$FB,$63,$F4
	.dc.b	$57,$CB,$1E,$D1,$FA,$5C,$79,$E7,$E8,$8E,$C7,$43,$F5,$58,$74,$69
	.dc.b	$D0,$E4,$5A,$FD,$15,$74,$CA,$4F,$39,$FD,$35,$73,$6C,$7E,$DF,$E0
	.dc.b	$FF,$2E,$7E,$63,$E1,$6C,$79,$27,$F0,$9A,$7C,$17,$F6,$1E,$57,$E7
	.dc.b	$BC,$4F,$CF,$B6,$BA,$12,$7F,$6D,$7D,$1E,$7A,$7C,$16,$E7,$CA,$BF
	.dc.b	$85,$61,$67,$F7,$CF,$1E,$1F,$D0,$E7,$E7,$AE,$D1,$FC,$61,$FC,$DB
	.dc.b	$6B,$F0,$6F,$EA,$32,$F3,$47,$AA,$7B,$9F,$C6,$7E,$1B,$F7,$3F,$0D
	.dc.b	$FB,$A6,$C7,$C3,$BE,$9A,$FC,$37,$F6,$7F,$0D,$FC,$03,$E1,$7F,$80
	.dc.b	$B6,$BA,$1F,$C9,$70,$E8,$D3,$A1,$FC,$7F,$1F,$E4,$27,$F1,$F9,$E4
	.dc.b	$FF,$1F,$6D,$FF,$02,$C7,$CB,$3F,$C8,$DB,$1E,$6D,$BE,$73,$76,$F2
	.dc.b	$E1,$FC,$DF,$2B,$87,$F2,$B7,$F8,$CF,$C2,$C9,$FF,$0E,$5F,$F2,$9F
	.dc.b	$A8,$7C,$3E,$70,$FB,$35,$EA,$7F,$9F,$4B,$E1,$FF,$92,$AF,$86,$B1
	.dc.b	$87,$96,$1F,$C7,$31,$DE,$38,$C7,$E2,$4F,$F2,$16,$9F,$78,$FF,$0F
	.dc.b	$E1,$DD,$F3,$98,$BE,$EE,$31,$F2,$C7,$57,$FA,$66,$1E,$09,$FE,$B4
	.dc.b	$FE,$AA,$3D,$A2,$FE,$16,$3E,$9A,$E9,$C0,$C7,$C5,$7B,$9F,$11,$A8
	.dc.b	$7C,$37,$C4,$3E,$2F,$C3,$D9,$D7,$C2,$FC,$8F,$87,$F9,$2D,$8E,$5F
	.dc.b	$0D,$F3,$3E,$1F,$E6,$B6,$3B,$7C,$2F,$D0,$3E,$8F,$C3,$D8,$F1,$7C
	.dc.b	$37,$D4,$F8,$7F,$AA,$D8,$F3,$7C,$37,$D8,$F8,$7F,$B2,$D8,$F5,$7C
	.dc.b	$37,$DC,$F8,$6F,$BA,$D8,$F7,$7C,$27,$E0,$3E,$1B,$F0,$2D,$8F,$85
	.dc.b	$F0,$DF,$84,$F8,$6F,$C2,$B6,$3E,$5F,$D4,$7F,$B0,$0F,$F5,$E7,$3F
	.dc.b	$CB,$3F,$C1,$4E,$7E,$1B,$FB,$8B,$FE,$A8,$EB,$FE,$FC,$FF,$AA,$2B
	.dc.b	$9F,$C3,$EB,$E1,$F7,$F0,$DC,$3E,$1E,$D7,$81,$E7,$F0,$D2,$3C,$7F
	.dc.b	$0D,$74,$F1,$78,$0C,$DF,$0B,$E2,$F8,$5F,$1B,$70,$F5,$FB,$FE,$6F
	.dc.b	$27,$CE,$D8,$F7,$FC,$2F,$A8,$F5,$FC,$35,$8F,$87,$F0,$DE,$EF,$85
	.dc.b	$F7,$B6,$3E,$3F,$C2,$FC,$23,$E1,$FC,$35,$8F,$97,$F0,$DF,$18,$F8
	.dc.b	$FF,$0B,$63,$E7,$FC,$37,$CA,$3E,$5F,$C3,$58,$FC,$EF,$3F,$F8,$EB
	.dc.b	$8C,$D3,$FC,$75,$C9,$FE,$3D,$75,$C8,$CB,$B4,$FC,$88,$65,$8E,$AF
	.dc.b	$AF,$9F,$7A,$D6,$76,$7F,$C7,$18,$F7,$CF,$FC,$5E,$F8,$33,$F1,$C7
	.dc.b	$33,$FD,$43,$97,$85,$FC,$5E,$59,$C7,$F1,$39,$75,$9D,$E5,$67,$D5
	.dc.b	$1D,$8E,$FF,$0D,$5F,$87,$F0,$BE,$23,$C7,$EB,$58,$F8,$FF,$07,$E6
	.dc.b	$F8,$5F,$3B,$63,$E5,$FC,$27,$AB,$E1,$7D,$6D,$8E,$A4,$CB,$F1,$FC
	.dc.b	$1F,$E4,$BE,$1F,$87,$9F,$4A,$FA,$B9,$5B,$DB,$D2,$3E,$CE,$36,$79
	.dc.b	$BE,$77,$AD,$7A,$A7,$C1,$7F,$43,$2F,$45,$78,$9B,$1F,$53,$2F,$44
	.dc.b	$F9,$9B,$1F,$63,$E1,$3E,$BF,$C2,$7D,$86,$E1,$F8,$CF,$85,$FB,$7F
	.dc.b	$0B,$F7,$1B,$1F,$90,$F8,$3F,$BE,$7E,$03,$E1,$6C,$7E,$52,$F9,$E7
	.dc.b	$F1,$9B,$97,$E2,$30,$E7,$3F,$29,$B1,$F8,$CC,$39,$CF,$CE,$C2,$CF
	.dc.b	$A7,$0F,$D0,$E5,$E9,$69,$E9,$3F,$4B,$97,$A5,$A7,$A4,$97,$D5,$B5
	.dc.b	$F9,$CF,$84,$FC,$DF,$C2,$FE,$71,$B3,$FB,$0F,$83,$FC,$FF,$C2,$FE
	.dc.b	$81,$B1,$FB,$4F,$85,$FD,$1F,$C2,$FE,$91,$B1,$FB,$8B,$E7,$97,$E2
	.dc.b	$9B,$97,$EB,$30,$E7,$3F,$8E,$6C,$7E,$C2,$F9,$CF,$E4,$9B,$1E,$9B
	.dc.b	$FE,$0B,$97,$A5,$A7,$A4,$FE,$13,$97,$A5,$A7,$A4,$FE,$1B,$97,$A5
	.dc.b	$A7,$F1,$0F,$83,$FD,$F9,$FE,$AE,$FA,$8F,$F0,$8C,$64,$FE,$E2,$FF
	.dc.b	$70,$C3,$FD,$7C,$FF,$83,$E3,$3F,$09,$C7,$E1,$79,$7C,$24,$9E,$3F
	.dc.b	$09,$76,$FD,$7E,$5E,$09,$EB,$F0,$D6,$7F,$A0,$E5,$DE,$51,$FE,$DD
	.dc.b	$8F,$03,$AC,$BC,$13,$E6,$C7,$B6,$7E,$6C,$AC,$F2,$F1,$3F,$96,$E7
	.dc.b	$DE,$BD,$79,$DA,$D6,$7E,$1C,$FC,$1A,$7C,$5D,$3B,$C7,$9B,$4E,$AF
	.dc.b	$97,$4B,$A3,$07,$C2,$7C,$D3,$E7,$7C,$2D,$CF,$E6,$FC,$27,$D1,$F8
	.dc.b	$5F,$A4,$D8,$FA,$3F,$0B,$F5,$4F,$AD,$F0,$B6,$3E,$AE,$1C,$F4,$F9
	.dc.b	$38,$DE,$1F,$AF,$8F,$39,$F9,$AD,$8F,$B7,$8F,$39,$FA,$38,$D9,$FB
	.dc.b	$F8,$73,$C7,$E2,$E3,$6F,$EA,$63,$CE,$7E,$4B,$63,$EC,$63,$CE,$7E
	.dc.b	$6E,$36,$7E,$E7,$C2,$7E,$44,$FC,$97,$C2,$DA,$FC,$8F,$C2,$7E,$57
	.dc.b	$E1,$7F,$2C,$D8,$FC,$AF,$C2,$7E,$67,$E1,$7F,$34,$D8,$FC,$CE,$1C
	.dc.b	$F1,$FC,$6E,$37,$7F,$CD,$E3,$CE,$7F,$22,$D8,$FC,$FE,$1C,$E7,$F2
	.dc.b	$AD,$8F,$D1,$DF,$3C,$3F,$13,$85,$BF,$CC,$61,$CE,$7F,$1A,$D8,$FC
	.dc.b	$E6,$1C,$E7,$F2,$2D,$8F,$D0,$7C,$27,$ED,$7E,$17,$F6,$CD,$AF,$DA
	.dc.b	$FC,$27,$EE,$7E,$17,$F7,$4D,$8F,$F3,$4F,$D4,$7E,$A9,$B8,$1C,$FC
	.dc.b	$35,$8F,$F1,$8C,$ED,$FD,$9E,$1D,$DB,$5F,$C1,$31,$EE,$FF,$06,$C6
	.dc.b	$CF,$EE,$9E,$9B,$75,$E8,$63,$D6,$30,$DB,$FD,$B3,$FE,$54,$8C,$DF
	.dc.b	$85,$FE,$23,$87,$C2,$AF,$11,$F0,$8F,$11,$67,$C4,$FC,$02,$3E,$EC
	.dc.b	$BC,$B1,$E0,$8A,$30,$F3,$FC,$36,$1F,$4F,$C3,$5C,$BF,$CC,$97,$E8
	.dc.b	$8C,$7A,$52,$3E,$8F,$C3,$5C,$7E,$6F,$C3,$71,$F8,$6E,$4D,$AF,$AF
	.dc.b	$9F,$5C,$7B,$36,$3E,$DE,$3C,$F6,$F9,$39,$59,$F9,$38,$73,$CB,$BC
	.dc.b	$7E,$0B,$2E,$6F,$E1,$32,$B3,$F8,$07,$93,$F0,$73,$FF,$20,$BF,$84
	.dc.b	$C3,$3A,$7D,$AC,$7B,$65,$F6,$9B,$8F,$DA,$F8,$5F,$8C,$7C,$7F,$86
	.dc.b	$B1,$F9,$4F,$85,$F9,$5F,$0D,$F2,$DB,$1F,$98,$F8,$6F,$9C,$7C,$FF
	.dc.b	$86,$B1,$F9,$CC,$39,$EB,$F9,$8C,$6E,$7F,$8C,$C7,$9C,$FE,$71,$B1
	.dc.b	$F9,$0C,$79,$CF,$E8,$31,$B3,$F9,$4C,$39,$E3,$F9,$4C,$6D,$FC,$5C
	.dc.b	$79,$CF,$E6,$1B,$1F,$27,$1E,$73,$F9,$CC,$6C,$FC,$DF,$85,$FC,$51
	.dc.b	$F8,$BF,$86,$B5,$FB,$8F,$86,$FC,$77,$C3,$7E,$3D,$B1,$FB,$CF,$86
	.dc.b	$FC,$91,$F9,$3F,$85,$B1,$FC,$03,$0E,$78,$FE,$F3,$1B,$BF,$EC,$31
	.dc.b	$E7,$3F,$C0,$1B,$1F,$B4,$C7,$9C,$FF,$04,$C6,$CF,$EE,$30,$E7,$8F
	.dc.b	$EE,$31,$B7,$F8,$9C,$79,$CF,$EF,$1B,$1F,$8D,$C7,$9C,$FF,$00,$C6
	.dc.b	$CF,$E4,$7E,$17,$F5,$C7,$EB,$FE,$16,$D7,$F8,$13,$3F,$E1,$9F,$51
	.dc.b	$FE,$A9,$3F,$D3,$05,$F8,$6F,$DC,$2F,$83,$A6,$3E,$1A,$DF,$F2,$6C
	.dc.b	$7A,$E5,$FC,$8B,$2B,$97,$6A,$E7,$97,$46,$9E,$33,$FC,$1D,$97,$77
	.dc.b	$17,$FC,$2E,$30,$FF,$B7,$9F,$F4,$84,$70,$CB,$BB,$FC,$8B,$1F,$80
	.dc.b	$F9,$4E,$7D,$0A,$F7,$46,$1B,$7D,$6F,$DA,$F7,$CE,$E5,$F2,$3E,$13
	.dc.b	$E2,$9F,$1B,$E1,$AE,$3C,$FE,$17,$E4,$FC,$37,$CA,$6C,$77,$F8,$6F
	.dc.b	$9B,$F0,$7F,$39,$B1,$E3,$C3,$9E,$7E,$FC,$6E,$5F,$63,$1E,$73,$F0
	.dc.b	$DB,$1F,$73,$0E,$73,$F1,$DB,$1E,$9C,$3B,$1F,$06,$FA,$BF,$4A,$F9
	.dc.b	$B4,$F8,$D9,$79,$DF,$91,$95,$C7,$EA,$3F,$88,$F8,$69,$3F,$37,$E1
	.dc.b	$2D,$7D,$9F,$86,$FC,$6F,$C3,$7E,$39,$B5,$F4,$FE,$1B,$F2,$3F,$09
	.dc.b	$F9,$26,$C7,$D7,$C7,$9E,$DF,$9A,$CA,$CF,$E2,$B0,$E7,$97,$78,$FB
	.dc.b	$D9,$73,$7F,$01,$95,$9F,$C6,$3C,$9F,$C8,$63,$FA,$2C,$31,$C9,$F8
	.dc.b	$9C,$7A,$E5,$FA,$A6,$E3,$F9,$AC,$FC,$F8,$FE,$5D,$B5,$FB,$0F,$86
	.dc.b	$FD,$69,$FA,$EF,$85,$B7,$F9,$3F,$86,$FD,$8F,$C3,$7E,$C9,$B1,$F9
	.dc.b	$7F,$86,$FE,$D9,$F0,$5F,$E9,$0C,$FF,$49,$18,$F8,$6E,$71,$FA,$7C
	.dc.b	$BB,$5F,$E7,$DE,$8F,$E8,$71,$B3,$FA,$D7,$F8,$0D,$F3,$9F,$01,$FA
	.dc.b	$AC,$3C,$19,$FA,$E3,$A1,$FA,$BC,$FC,$31,$FC,$3B,$3F,$4E,$3F,$B9
	.dc.b	$6C,$7F,$12,$C3,$B4,$FA,$AF,$B6,$7E,$EF,$85,$F7,$FC,$35,$9F,$13
	.dc.b	$F0,$BE,$1B,$E1,$E9,$67,$CC,$FC,$6F,$84,$F8,$FF,$09,$67,$D4,$F3
	.dc.b	$C9,$A7,$F8,$BE,$FF,$D4,$ED,$CB,$C3,$7F,$07,$1E,$ED,$3B,$9F,$17
	.dc.b	$1E,$ED,$3B,$9F,$26,$79,$E1,$E0,$8F,$F5,$76,$1F,$EA,$9E,$33,$E2
	.dc.b	$8C,$E1,$E0,$AE,$07,$F8,$BE,$51,$9C,$7D,$58,$F8,$AB,$78,$74,$8C
	.dc.b	$4F,$BF,$E1,$3F,$16,$7E,$33,$E1,$6E,$5F,$7F,$E1,$7F,$1F,$F0,$BF
	.dc.b	$90,$6C,$7E,$0F,$E0,$FF,$26,$7E,$53,$E0,$EC,$7E,$1F,$0E,$7A,$78
	.dc.b	$B6,$65,$8C,$FE,$86,$5D,$2B,$4D,$8F,$A9,$97,$49,$E5,$95,$9F,$56
	.dc.b	$5D,$CF,$CB,$61,$D5,$F9,$B8,$73,$69,$F9,$BC,$F9,$3F,$4B,$1F,$53
	.dc.b	$ED,$3F,$41,$F0,$7F,$B0,$3F,$63,$F0,$B6,$3F,$49,$F0,$9F,$B4,$F8
	.dc.b	$5F,$DA,$B6,$3F,$51,$F0,$9F,$B8,$3F,$73,$F0,$76,$3A,$93,$7F,$8D
	.dc.b	$87,$57,$E4,$5F,$36,$CF,$EC,$F0,$E5,$1F,$37,$0E,$B3,$E0,$3A,$6D
	.dc.b	$F9,$DC,$BD,$31,$83,$F2,$38,$75,$9F,$D1,$36,$3F,$2B,$5D,$A7,$F4
	.dc.b	$D9,$59,$FC,$CF,$C2,$FF,$14,$3F,$8A,$FC,$2D,$8F,$EE,$3F,$51,$FE
	.dc.b	$97,$CF,$FD,$0C,$DE,$63,$FE,$99,$94,$E8,$FF,$45,$47,$FA,$21,$E3
	.dc.b	$F0,$B8,$F8,$3C,$BA,$F8,$5D,$FC,$1D,$C3,$C2,$F2,$F8,$3E,$7F,$0B
	.dc.b	$67,$CA,$F6,$F8,$3E,$FF,$07,$67,$D2,$F8,$BE,$0E,$47,$B3,$E1,$2E
	.dc.b	$BF,$AE,$D9,$F0,$9E,$8F,$84,$F4,$B6,$FE,$1F,$C1,$FB,$3E,$0F,$DA
	.dc.b	$D8,$F8,$FF,$07,$F0,$0F,$83,$F0,$96,$3E,$5F,$C2,$7C,$43,$E2,$FC
	.dc.b	$25,$8F,$9F,$F0,$5F,$23,$E1,$3E,$4B,$63,$A9,$36,$F4,$61,$D5,$F5
	.dc.b	$61,$CD,$F6,$61,$7B,$7F,$DA,$9B,$B1,$E9,$5F,$05,$B3,$AC,$FA,$3C
	.dc.b	$31,$B9,$7D,$EE,$AF,$C0,$E7,$D6,$31,$93,$3C,$67,$F8,$8C,$FA,$D6
	.dc.b	$9B,$1F,$ED,$1F,$A8,$FF,$7B,$4F,$F9,$01,$B1,$FC,$89,$7C,$2F,$FB
	.dc.b	$55,$CE,$9F,$EE,$F7,$FC,$9E,$DF,$83,$D7,$C2,$6D,$E1,$F0,$9C,$7E
	.dc.b	$0E,$E5,$DD,$E8,$F5,$F8,$3A,$F8,$DE,$6F,$96,$AE,$7E,$57,$9B,$E7
	.dc.b	$AB,$8F,$95,$F4,$7C,$1C,$BF,$77,$C2,$5B,$F7,$7C,$27,$B7,$E0,$BD
	.dc.b	$CD,$AF,$89,$F0,$5F,$04,$F8,$5F,$05,$63,$E4,$63,$F1,$7E,$13,$A6
	.dc.b	$9F,$15,$E7,$3F,$37,$1B,$97,$CD,$79,$3F,$3F,$1E,$B3,$E0,$3E,$A7
	.dc.b	$C1,$FD,$1F,$84,$FA,$4D,$8F,$B1,$F0,$9F,$57,$E0,$FE,$B3,$63,$EE
	.dc.b	$7C,$1F,$D9,$F8,$4F,$B4,$D8,$FC,$06,$1C,$F2,$FC,$06,$37,$5F,$BD
	.dc.b	$8F,$39,$FC,$23,$63,$F0,$58,$73,$9F,$C4,$36,$3F,$0D,$7C,$F0,$FC
	.dc.b	$D7,$3B,$1C,$61,$F9,$3C,$7A,$56,$9B,$1F,$97,$C7,$A4,$F2,$6C,$7E
	.dc.b	$6F,$E1,$3F,$2C,$27,$FD,$37,$F5,$1F,$E3,$27,$C3,$19,$9F,$EF,$F3
	.dc.b	$FE,$0B,$27,$C2,$67,$4D,$78,$09,$D7,$C1,$67,$1F,$DA,$47,$F8,$31
	.dc.b	$EF,$F0,$19,$C5,$7C,$6F,$CB,$5F,$AF,$CB,$53,$D7,$E1,$73,$7E,$4F
	.dc.b	$84,$CE,$5E,$4C,$3D,$71,$8C,$64,$78,$0F,$7B,$C3,$E1,$24,$7C,$2F
	.dc.b	$85,$B8,$7C,$0F,$85,$EA,$76,$F8,$4B,$5F,$23,$E1,$3C,$3F,$09,$E2
	.dc.b	$6C,$7C,$CC,$79,$F0,$FC,$6D,$59,$E2,$BE,$2E,$5D,$33,$F9,$F9,$DC
	.dc.b	$38,$65,$D2,$7E,$9B,$63,$A7,$C2,$7C,$23,$E1,$FC,$2D,$C7,$EE,$FC
	.dc.b	$1F,$C6,$3E,$3F,$C2,$D8,$FC,$0F,$C2,$7C,$AF,$85,$F9,$6D,$8F,$C2
	.dc.b	$E1,$CF,$3F,$C1,$E3,$72,$F6,$65,$CE,$7F,$0F,$95,$9F,$81,$8F,$39
	.dc.b	$FC,$5B,$63,$D3,$8F,$63,$F2,$78,$75,$7F,$13,$87,$37,$F1,$78,$5C
	.dc.b	$3F,$13,$BB,$1E,$95,$F9,$06,$CE,$BE,$17,$F0,$E7,$E2,$3E,$16,$D7
	.dc.b	$E8,$BE,$17,$F1,$67,$E3,$3E,$16,$C7,$FC,$00,$F3,$BF,$C1,$C7,$F8
	.dc.b	$37,$0F,$CC,$B8,$F8,$8E,$71,$FA,$8C,$7B,$57,$EB,$9E,$71,$FA,$CC
	.dc.b	$3B,$4F,$84,$ED,$1F,$A0,$BE,$DD,$67,$B4,$F8,$FE,$17,$5F,$0D,$B6
	.dc.b	$C7,$9F,$E1,$B9,$7C,$37,$36,$C7,$AF,$E1,$BB,$7C,$37,$76,$C7,$BF
	.dc.b	$0E,$79,$FF,$A0,$32,$FF,$14,$67,$E3,$9E,$95,$F1,$72,$B7,$B7,$A4
	.dc.b	$7C,$9C,$6C,$F3,$7D,$DF,$0D,$27,$E7,$7C,$35,$AF,$07,$C3,$7C,$3F
	.dc.b	$86,$F8,$8D,$AF,$A9,$F0,$BF,$1C,$F9,$1F,$0D,$63,$EC,$63,$CF,$B3
	.dc.b	$F8,$BE,$EC,$F1,$5F,$E0,$98,$63,$D7,$3F,$0C,$FE,$07,$3E,$91,$F0
	.dc.b	$F3,$EF,$3E,$23,$B9,$23,$E4,$61,$D9,$FC,$3D,$F5,$C7,$C6,$DA,$F1
	.dc.b	$5F,$69,$F3,$E7,$67,$CD,$F0,$BF,$85,$3F,$0D,$F0,$D6,$3F,$2D,$F0
	.dc.b	$BF,$89,$3F,$15,$F0,$B6,$3F,$35,$F0,$BF,$8D,$F8,$4F,$C7,$36,$3F
	.dc.b	$3D,$87,$3D,$7F,$01,$8D,$CF,$E4,$E5,$CE,$7F,$09,$95,$9F,$9B,$8F
	.dc.b	$39,$FC,$43,$63,$E8,$FC,$27,$E7,$7E,$1B,$F3,$CD,$CF,$FC,$EF,$3F
	.dc.b	$D3,$3E,$A3,$FC,$57,$1F,$E3,$D6,$E1,$FD,$21,$FF,$18,$C7,$03,$47
	.dc.b	$F4,$87,$FC,$62,$F1,$F8,$5C,$7C,$26,$5F,$0C,$7E,$F7,$1B,$7F,$9F
	.dc.b	$79,$44,$3B,$47,$87,$E1,$B0,$FF,$B9,$30,$EE,$F9,$6F,$BD,$70,$EB
	.dc.b	$56,$9B,$CF,$E7,$F8,$5F,$43,$6B,$85,$3E,$1B,$D9,$F0,$7E,$D6,$CE
	.dc.b	$BE,$1B,$E0,$1F,$07,$E1,$AC,$72,$C7,$A6,$9E,$7C,$AF,$07,$CF,$C7
	.dc.b	$9C,$FA,$DB,$1F,$4F,$1E,$73,$EF,$C6,$CF,$D7,$C3,$9E,$3F,$5F,$1B
	.dc.b	$7F,$5F,$1E,$73,$F6,$DB,$1F,$6F,$1E,$73,$F7,$F1,$B3,$F7,$FE,$13
	.dc.b	$EE,$1F,$77,$E1,$2D,$7C,$6F,$86,$FC,$07,$C3,$7E,$05,$B1,$F2,$BE
	.dc.b	$0B,$F0,$87,$E1,$7E,$12,$C7,$CE,$C3,$F1,$1F,$07,$D3,$2F,$B8,$F3
	.dc.b	$9F,$A7,$8D,$9F,$C0,$3C,$9F,$C1,$61,$D6,$7C,$07,$DA,$F8,$6F,$CA
	.dc.b	$7C,$27,$E5,$5B,$1F,$7B,$E1,$BF,$31,$F0,$7F,$99,$6C,$7E,$27,$EA
	.dc.b	$3F,$C6,$D8,$FF,$8A,$17,$C1,$FF,$89,$9C,$F4,$9C,$D7,$84,$FC,$5E
	.dc.b	$5E,$16,$9E,$13,$F1,$F8,$F8,$5A,$78,$7E,$13,$B6,$1D,$DB,$1E,$5F
	.dc.b	$86,$F1,$7C,$2F,$8D,$B1,$E9,$F8,$6F,$31,$E7,$F8,$5B,$1E,$DF,$85
	.dc.b	$F5,$7C,$2F,$AD,$B1,$D4,$F8,$58,$75,$69,$D4,$9B,$BB,$BF,$CA,$BB
	.dc.b	$B1,$EF,$39,$AF,$F6,$44,$FF,$8D,$1F,$9D,$8F,$85,$A7,$69,$E0,$7D
	.dc.b	$1C,$3C,$4E,$38,$3E,$84,$76,$C7,$C3,$9F,$BE,$FC,$B1,$E4,$CF,$DB
	.dc.b	$87,$1C,$BC,$4D,$CB,$EE,$7C,$2F,$DD,$F8,$5B,$8F,$E0,$DF,$C0,$7C
	.dc.b	$37,$E0,$7E,$16,$CF,$E1,$DF,$C2,$7C,$2F,$E1,$7E,$16,$CF,$E2,$DE
	.dc.b	$79,$78,$70,$BA,$FC,$87,$9C,$79,$70,$B3,$F3,$1E,$71,$E9,$AB,$3F
	.dc.b	$41,$FC,$A7,$C2,$CC,$3F,$35,$F0,$96,$FE,$07,$C3,$7E,$64,$FC,$D7
	.dc.b	$C1,$5A,$FD,$07,$C2,$FE,$74,$FC,$F7,$C1,$D8,$FD,$26,$3C,$F3,$FB
	.dc.b	$F9,$59,$FB,$B8,$F3,$CB,$E7,$65,$6F,$F0,$B9,$73,$9F,$A5,$95,$9F
	.dc.b	$C4,$FC,$2F,$EB,$4F,$F4,$9F,$D4,$2B,$19,$7F,$3D,$8F,$F1,$B9,$3E
	.dc.b	$17,$7F,$0D,$C3,$E1,$38,$FC,$35,$BE,$0F,$2C,$3F,$05,$8F,$67,$19
	.dc.b	$7F,$AB,$B2,$EB,$8F,$8B,$2E,$B8,$F8,$72,$EA,$D3,$A9,$FC,$27,$1F
	.dc.b	$4F,$C2,$F4,$8F,$57,$6B,$14,$F4,$CF,$F3,$91,$EB,$AF,$76,$5E,$99
	.dc.b	$CE,$97,$4F,$33,$F1,$BE,$1A,$5F,$C5,$F8,$6B,$7D,$3E,$1B,$E5,$FC
	.dc.b	$2F,$CC,$6D,$7C,$BF,$85,$F9,$E7,$D0,$F8,$6B,$1F,$3F,$1F,$A7,$F0
	.dc.b	$DD,$34,$F8,$CF,$39,$FA,$98,$DC,$BD,$6F,$27,$DB,$8F,$59,$F0,$1F
	.dc.b	$6F,$E1,$BE,$FF,$C2,$FE,$01,$B1,$F7,$FE,$17,$F0,$67,$E1,$3E,$16
	.dc.b	$C7,$E0,$FE,$17,$F0,$FF,$0B,$F8,$86,$C7,$E1,$F0,$E7,$97,$CE,$C6
	.dc.b	$EB,$F7,$F2,$E7,$3F,$4B,$2B,$3F,$83,$C7,$9C,$FD,$66,$C7,$E1,$FE
	.dc.b	$1B,$F2,$E7,$E6,$3E,$1A,$E7,$F9,$7F,$85,$FF,$76,$F4,$7F,$8B,$3A
	.dc.b	$7F,$05,$18,$9C,$FC,$2C,$7F,$C3,$55,$FE,$19,$3F,$C6,$39,$FE,$1B
	.dc.b	$1E,$AF,$E2,$31,$B9,$7E,$67,$1E,$98,$78,$4E,$67,$5A,$FE,$55,$87
	.dc.b	$4A,$4F,$E5,$A3,$C1,$8F,$48,$FC,$CF,$53,$2F,$74,$7E,$CD,$FD,$BE
	.dc.b	$3E,$0C,$7F,$36,$DC,$7A,$C3,$3F,$D0,$E3,$E3,$9C,$C7,$E9,$3E,$17
	.dc.b	$18,$F8,$FE,$1B,$E4,$FC,$1F,$CA,$6C,$7C,$6F,$84,$F9,$A7,$CE,$F8
	.dc.b	$4B,$1F,$2B,$E1,$7E,$89,$F4,$BE,$16,$C7,$CE,$BE,$7A,$FC,$46,$E9
	.dc.b	$E4,$E1,$65,$D2,$BE,$4E,$56,$B5,$97,$49,$F9,$B9,$59,$E5,$F0,$BF
	.dc.b	$82,$3F,$07,$F0,$57,$4F,$BF,$F0,$BF,$86,$F8,$5F,$C3,$B6,$3F,$07
	.dc.b	$F0,$7F,$8A,$3F,$17,$F0,$76,$3F,$0F,$87,$3C,$FE,$EE,$37,$2F,$56
	.dc.b	$3C,$E7,$F0,$2D,$8F,$76,$3C,$E7,$F0,$B8,$D9,$F8,$5F,$0B,$F9,$AF
	.dc.b	$85,$FC,$DB,$73,$FC,$BF,$C2,$FE,$78,$FC,$FF,$C2,$58,$FC,$DF,$C2
	.dc.b	$FE,$88,$FD,$1F,$C2,$D8,$FC,$FD,$F3,$CF,$F2,$AD,$CB,$E9,$63,$CE
	.dc.b	$7F,$33,$8D,$9F,$AD,$8F,$39,$FC,$EE,$36,$7D,$38,$FE,$B3,$2F,$4B
	.dc.b	$4F,$49,$2B,$AB,$67,$AB,$EA,$3F,$6B,$5D,$5F,$E7,$BF,$51,$FE,$44
	.dc.b	$3F,$C8,$75,$FE,$17,$71,$F0,$DF,$A2,$BF,$CF,$C3,$E1,$71,$7C,$9B
	.dc.b	$7F,$A7,$97,$C2,$63,$B5,$1C,$6F,$FD,$40,$BE,$13,$4D,$8C,$FC,$2E
	.dc.b	$27,$3F,$0D,$AD,$2D,$72,$78,$E7,$23,$AE,$57,$3F,$66,$1C,$E7,$A9
	.dc.b	$E2,$7C,$3F,$0D,$E2,$F8,$5B,$3E,$67,$CB,$F0,$BE,$6F,$83,$B3,$EA
	.dc.b	$78,$3B,$3D,$BF,$0B,$8C,$7E,$66,$17,$87,$E5,$B6,$BE,$9E,$1D,$5F
	.dc.b	$8A,$DA,$FA,$98,$ED,$F9,$D8,$5E,$1F,$9B,$F0,$9F,$2B,$E1,$7E,$5B
	.dc.b	$71,$F9,$FF,$0B,$F3,$8F,$9F,$F0,$B6,$3E,$9F,$C2,$FD,$2F,$85,$FA
	.dc.b	$6D,$8F,$AF,$7C,$F3,$FC,$33,$77,$FC,$2E,$3C,$E7,$F1,$58,$D9,$FC
	.dc.b	$4E,$1C,$E7,$F1,$CD,$8F,$C6,$DF,$3C,$3F,$2A,$DA,$FC,$66,$1C,$E7
	.dc.b	$F3,$38,$59,$FC,$86,$1C,$E7,$F3,$AD,$8F,$CA,$7C,$2F,$E3,$8F,$C7
	.dc.b	$FC,$2D,$AF,$C9,$FC,$2F,$E4,$BE,$0F,$F2,$6D,$8F,$F7,$B7,$D4,$7F
	.dc.b	$8F,$5B,$F0,$9F,$9D,$8C,$63,$FE,$7F,$51,$D6,$3F,$59,$8F,$82,$BF
	.dc.b	$42,$F6,$7F,$47,$85,$9F,$D9,$3D,$5E,$D5,$67,$C2,$79,$3E,$1B,$C5
	.dc.b	$F0,$DE,$36,$C7,$A3,$E1,$BC,$C7,$9F,$D1,$B1,$EC,$F8,$3F,$57,$C3
	.dc.b	$7A,$DB,$1D,$4F,$83,$87,$56,$9D,$49,$97,$C1,$BE,$AF,$C3,$AE,$6D
	.dc.b	$9F,$97,$7F,$27,$E1,$A4,$7C,$EF,$84,$B5,$F2,$FE,$13,$E7,$7C,$37
	.dc.b	$CF,$6D,$7D,$4F,$86,$FA,$5F,$09,$F4,$DB,$1F,$63,$1E,$7D,$5F,$DE
	.dc.b	$39,$19,$E2,$BF,$CB,$F0,$C7,$AE,$7E,$19,$FC,$0E,$7D,$5F,$C1,$E7
	.dc.b	$67,$F0,$0F,$37,$FC,$11,$6F,$85,$C7,$72,$31,$7F,$8C,$C7,$BE,$7E
	.dc.b	$56,$C7,$93,$5F,$5E,$3F,$8D,$6D,$7A,$FE,$17,$F2,$5F,$0B,$F9,$36
	.dc.b	$DF,$E6,$3E,$17,$F2,$DF,$0D,$F9,$76,$C7,$E7,$3E,$13,$F3,$5F,$0D
	.dc.b	$F9,$B6,$C7,$E8,$2F,$9E,$7F,$9D,$6E,$5F,$2F,$0E,$73,$FA,$16,$C7
	.dc.b	$CF,$BE,$73,$FA,$56,$C7,$D3,$F8,$5F,$D5,$1F,$D1,$FE,$A3,$FC,$A1
	.dc.b	$BA,$31,$F0,$D9,$F8,$6B,$A7,$FA,$BF,$0F,$F1,$9E,$09,$E5,$F0,$B2
	.dc.b	$B4,$7F,$A7,$25,$F0,$D9,$D3,$F7,$19,$78,$AB,$71,$FE,$41,$AF,$1C
	.dc.b	$F8,$23,$FD,$F9,$1D,$E7,$7C,$88,$FF,$2B,$47,$7C,$7B,$65,$E6,$CB
	.dc.b	$DD,$96,$2B,$C9,$F0,$DF,$07,$E1,$BE,$13,$63,$AE,$2F,$86,$F8,$C7
	.dc.b	$C7,$F8,$6B,$3A,$F8,$5F,$95,$F0,$BF,$2D,$B1,$CB,$0E,$9B,$FA,$9B
	.dc.b	$AF,$CA,$C7,$9C,$FB,$B1,$B3,$F3,$B0,$E7,$3F,$09,$B1,$F4,$B3,$E7
	.dc.b	$8F,$B0,$F4,$BC,$9F,$A5,$87,$A5,$F6,$BE,$77,$DF,$95,$C3,$E7,$BF
	.dc.b	$85,$F8,$59,$3F,$13,$E1,$AD,$7D,$4F,$86,$FC,$51,$F8,$BF,$86,$B5
	.dc.b	$F2,$BE,$1B,$F1,$C7,$E3,$FE,$0E,$C7,$CE,$C7,$F2,$5F,$0D,$D3,$6F
	.dc.b	$C6,$BC,$B2,$ED,$5F,$53,$1C,$57,$E4,$B1,$E7,$5F,$86,$C6,$CF,$E5
	.dc.b	$BE,$1B,$F3,$C7,$E7,$FE,$1A,$E3,$F7,$BE,$1B,$FC,$51,$F0,$AF,$85
	.dc.b	$FD,$24,$61,$CF,$C2,$EB,$E1,$6D,$FE,$4A,$FF,$C5,$DA,$7E,$5B,$2E
	.dc.b	$AF,$E6,$32,$B9,$FE,$A3,$3E,$98,$78,$4E,$59,$F4,$B5,$7F,$A8,$31
	.dc.b	$EA,$DC,$FC,$57,$E1,$F8,$5F,$4F,$C2,$FA,$9B,$1E,$5B,$F6,$FC,$27
	.dc.b	$B9,$B1,$E9,$F8,$3F,$83,$F0,$9F,$09,$B1,$ED,$C3,$9E,$7E,$48,$3F
	.dc.b	$EB,$7B,$FF,$28,$C6,$CF,$01,$F1,$70,$F0,$34,$F0,$1F,$27,$0F,$03
	.dc.b	$4E,$55,$DC,$ED,$7E,$D8,$CE,$7F,$17,$0F,$5D,$6F,$C2,$44,$AF,$04
	.dc.b	$E6,$19,$7B,$A3,$0D,$BF,$A1,$F0,$BF,$84,$F8,$5F,$C2,$B7,$0F,$C0
	.dc.b	$7C,$1F,$E2,$3E,$17,$F1,$2D,$8F,$C2,$7C,$27,$E3,$3E,$13,$F1,$AD
	.dc.b	$8F,$C4,$61,$CF,$3F,$26,$37,$2F,$C0,$63,$CE,$7D,$0D,$8F,$C2,$61
	.dc.b	$CE,$7D,$8D,$8F,$C4,$7C,$27,$E7,$3E,$17,$F3,$AD,$CB,$F3,$1F,$07
	.dc.b	$FA,$0F,$85,$FD,$0B,$63,$F3,$9F,$0B,$FA,$43,$F4,$BF,$0B,$63,$F4
	.dc.b	$18,$73,$CF,$E6,$63,$72,$FC,$C6,$3C,$E7,$E8,$36,$3F,$39,$8F,$39
	.dc.b	$FA,$98,$D9,$FF,$7F,$FD,$47,$F8,$F9,$BF,$0D,$FE,$9D,$8C,$67,$F9
	.dc.b	$2E,$54,$7F,$90,$A3,$FA,$2A,$F8,$6E,$1F,$0B,$C5,$B3,$AF,$86,$E8
	.dc.b	$75,$F8,$6B,$1C,$B1,$F0,$1E,$1D,$2C,$76,$BE,$71,$F9,$1C,$7C,$13
	.dc.b	$E2,$C7,$9C,$E7,$1E,$D3,$E1,$3B,$12,$39,$1D,$8E,$97,$DB,$3F,$8A
	.dc.b	$D9,$F8,$58,$2F,$BD,$7C,$AC,$AD,$6B,$E1,$BE,$51,$F2,$FE,$1A,$C7
	.dc.b	$C8,$F8,$6F,$9C,$7C,$FF,$86,$B1,$F3,$3E,$0F,$E9,$7C,$37,$D3,$6C
	.dc.b	$7D,$0C,$39,$ED,$F0,$F1,$B9,$FD,$CC,$B9,$CF,$C7,$CA,$CF,$E0,$31
	.dc.b	$E7,$3F,$2D,$B1,$F8,$4F,$84,$FC,$17,$C3,$7E,$0D,$B9,$FE,$03,$E1
	.dc.b	$BF,$0C,$7E,$1F,$E0,$EC,$7E,$13,$E1,$BF,$15,$F0,$9F,$8B,$6C,$7E
	.dc.b	$22,$F9,$E7,$F6,$DB,$97,$E5,$31,$E7,$3F,$7F,$1B,$3F,$98,$C3,$9C
	.dc.b	$FE,$0D,$B1,$F9,$CF,$86,$FC,$D7,$C2,$FE,$6D,$B9,$7E,$63,$E1,$BF
	.dc.b	$3C,$7E,$7F,$E1,$2C,$7F,$88,$3E,$A0,$FF,$BF,$DF,$F1,$0E,$4E,$C4
	.dc.b	$6A,$78,$7C,$37,$77,$F4,$B9,$5B,$FC,$C3,$D9,$FC,$D6,$3E,$19,$F2
	.dc.b	$1E,$13,$C1,$5E,$16,$9F,$B0,$CB,$D7,$5F,$A3,$6E,$7F,$BC,$BE,$CF
	.dc.b	$EF,$B3,$B7,$FB,$BC,$BD,$18,$7F,$3B,$7D,$B9,$FB,$27,$BC,$E7,$E1
	.dc.b	$BE,$17,$C3,$7C,$36,$CF,$C7,$F8,$6F,$8C,$7C,$7F,$85,$B1,$F2,$FE
	.dc.b	$0F,$E5,$7C,$1F,$CB,$6C,$7C,$FB,$E7,$8F,$78,$FA,$58,$F2,$8F,$36
	.dc.b	$3D,$67,$C0,$75,$3E,$C6,$1D,$5A,$7D,$BF,$85,$FB,$5F,$0B,$F6,$DB
	.dc.b	$1F,$7F,$E1,$7E,$F7,$C1,$FD,$F6,$C7,$E0,$FE,$17,$F0,$5F,$0B,$F8
	.dc.b	$36,$C7,$E1,$F0,$E7,$9F,$CB,$C6,$F1,$FE,$2A,$3F,$11,$C4,$CB,$AC
	.dc.b	$7D,$16,$CE,$F8,$BF,$1B,$97,$6A,$FA,$CD,$9D,$5B,$E1,$3F,$2C,$7E
	.dc.b	$5F,$E1,$2E,$9F,$9B,$F8,$4F,$CD,$7C,$2F,$E6,$DB,$1F,$9F,$F8,$4F
	.dc.b	$CF,$7C,$27,$E7,$DB,$1F,$A3,$C3,$F4,$5F,$09,$D3,$5F,$C4,$3C,$E7
	.dc.b	$F5,$18,$D9,$FC,$63,$CE,$3F,$59,$85,$9F,$C8,$3C,$F4,$F5,$DD,$AF
	.dc.b	$CB,$3C,$E3,$DF,$76,$7F,$34,$F3,$8F,$87,$56,$7F,$3C,$FD,$18,$F9
	.dc.b	$9F,$51,$FE,$51,$9F,$F4,$EA,$F8,$4D,$47,$F5,$F8,$D6,$3F,$E1,$F7
	.dc.b	$51,$C3,$E1,$B7,$3F,$C0,$E7,$FC,$93,$3C,$7E,$17,$3C,$49,$CD,$6F
	.dc.b	$E0,$F3,$06,$FC,$1E,$1C,$FC,$36,$B3,$B9,$7A,$1E,$39,$C8,$F5,$E7
	.dc.b	$72,$F4,$61,$CE,$7D,$F8,$D9,$F6,$7C,$37,$8B,$E1,$7C,$6D,$AF,$8B
	.dc.b	$F0,$DE,$63,$CF,$F0,$76,$3E,$4F,$C2,$7A,$8F,$5F,$C2,$58,$F9,$B7
	.dc.b	$CF,$1E,$F1,$F4,$31,$E5,$1F,$37,$1E,$B3,$E0,$3A,$92,$3E,$95,$F5
	.dc.b	$7E,$A7,$C3,$7C,$A3,$E5,$FC,$2D,$AF,$BB,$F0,$DF,$38,$F9,$FF,$09
	.dc.b	$63,$F0,$3F,$0B,$F4,$BE,$17,$E9,$B6,$3F,$0B,$87,$3C,$7F,$05,$8D
	.dc.b	$E8,$FC,$5E,$5C,$E7,$F0,$D9,$59,$FC,$7E,$3C,$E7,$F1,$4D,$8F,$C9
	.dc.b	$E1,$F8,$2F,$86,$E9,$97,$E4,$5E,$4F,$E4,$F1,$F4,$BE,$C3,$D2,$7E
	.dc.b	$6B,$3F,$4B,$4F,$CE,$FC,$37,$E3,$BE,$1B,$F1,$ED,$8F,$F7,$EF,$9F
	.dc.b	$FE,$3E,$D3,$FA,$E3,$7B,$0F,$F8,$FE,$3F,$D0,$0E,$4F,$F1,$8B,$8F
	.dc.b	$85,$FC,$FC,$63,$5F,$E3,$4A,$3C,$37,$FB,$0C,$6D,$FE,$B5,$E2,$C8
	.dc.b	$F2,$FC,$35,$8F,$0F,$C3,$76,$BE,$ED,$AF,$5F,$C2,$F8,$BE,$17,$C6
	.dc.b	$D8,$F7,$FC,$27,$98,$F3,$FC,$25,$8F,$87,$F0,$9E,$A3,$D7,$F0,$96
	.dc.b	$3A,$9F,$23,$0E,$AD,$3A,$93,$3F,$8F,$7D,$5F,$93,$7C,$DF,$97,$77
	.dc.b	$7F,$8F,$F0,$9F,$2B,$E1,$7E,$5B,$71,$FA,$FF,$07,$F3,$8F,$9F,$F0
	.dc.b	$B6,$3E,$DF,$C2,$7D,$2F,$84,$FA,$6D,$8F,$BF,$7C,$F3,$FB,$6D,$CB
	.dc.b	$EB,$E3,$CE,$7E,$FE,$36,$7E,$DE,$1C,$E7,$F0,$6D,$8F,$4E,$1F,$8C
	.dc.b	$CB,$D2,$D3,$D2,$4A,$EA,$D9,$EA,$FE,$53,$2F,$4B,$4F,$CB,$FC,$27
	.dc.b	$E3,$8F,$C7,$FC,$2D,$8F,$CD,$FC,$1F,$E4,$8F,$C9,$FC,$2D,$8F,$CF
	.dc.b	$FC,$27,$E5,$BE,$13,$F2,$ED,$8E,$A4,$CF,$F3,$F8,$75,$7F,$43,$87
	.dc.b	$37,$F4,$78,$5C,$FF,$3F,$87,$39,$FD,$3B,$63,$FB,$E7,$D4,$7F,$8C
	.dc.b	$1B,$8F,$F4,$68,$FF,$18,$99,$7F,$D6,$0B,$E1,$34,$6F,$E1,$2C,$B0
	.dc.b	$FF,$7A,$53,$E1,$3A,$1D,$7B,$ED,$6F,$1F,$0B,$4E,$FF,$0B,$E3,$7C
	.dc.b	$8F,$6C,$3A,$E5,$E1,$9E,$11,$FE,$63,$C7,$CD,$18,$F0,$51,$E3,$8C
	.dc.b	$36,$7C,$F8,$F9,$1F,$4F,$C2,$FC,$33,$E2,$7C,$15,$BE,$3F,$07,$F1
	.dc.b	$CF,$91,$F0,$76,$3A,$FC,$2F,$CB,$F8,$4F,$98,$D8,$F0,$E1,$CF,$1F
	.dc.b	$8F,$8D,$DF,$EA,$E5,$CE,$7E,$5E,$56,$7E,$CE,$3C,$E7,$E7,$B6,$3E
	.dc.b	$EE,$1F,$6F,$E1,$7A,$65,$F6,$9E,$4F,$DC,$C7,$D2,$FB,$0F,$49,$F1
	.dc.b	$B3,$F4,$B4,$F9,$3F,$03,$F8,$7F,$84,$FC,$43,$63,$E6,$FC,$27,$E2
	.dc.b	$FE,$0B,$F1,$8D,$8F,$A3,$F0,$9F,$8F,$F8,$3F,$C8,$36,$3A,$9F,$5B
	.dc.b	$0E,$AD,$3A,$9F,$6A,$FA,$B4,$EA,$7D,$EA,$EA,$D3,$F0,$3F,$09,$F9
	.dc.b	$F3,$FC,$4B,$F5,$1F,$E2,$ED,$FF,$4B,$18,$F8,$39,$5F,$8B,$AF,$F1
	.dc.b	$4F,$39,$FF,$19,$F1,$B0,$FF,$34,$43,$E0,$BA,$9D,$BE,$0A,$E1,$E2
	.dc.b	$25,$F1,$7F,$55,$87,$5F,$84,$F3,$47,$87,$2F,$1D,$6C,$FF,$61,$63
	.dc.b	$E4,$3C,$79,$78,$B2,$EF,$3B,$C7,$D1,$96,$2B,$CF,$F0,$7F,$0B,$E0
	.dc.b	$7E,$1B,$73,$F8,$1F,$03,$F1,$8F,$8F,$F0,$36,$3E,$27,$C1,$FC,$AF
	.dc.b	$83,$F9,$6D,$8F,$91,$7C,$F4,$F1,$6C,$8C,$67,$F2,$32,$E9,$5A,$CA
	.dc.b	$CF,$CC,$C7,$A4,$F2,$6C,$7D,$0C,$F9,$E5,$F2,$B2,$EC,$FD,$98,$E6
	.dc.b	$E7,$1F,$5B,$EE,$3D,$67,$DF,$CF,$D6,$D3,$F0,$5F,$07,$F8,$83,$F1
	.dc.b	$3F,$01,$63,$F0,$DF,$07,$F8,$C3,$F1,$BF,$01,$63,$F1,$5F,$07,$F9
	.dc.b	$03,$F2,$3F,$07,$63,$A9,$37,$F8,$F8,$75,$7E,$4E,$1C,$DF,$97,$85
	.dc.b	$DB,$F2,$38,$F3,$9F,$9F,$8D,$9F,$CA,$FC,$17,$E8,$0F,$D0,$FC,$1D
	.dc.b	$CF,$F3,$DF,$07,$FA,$43,$F4,$BF,$07,$63,$F9,$67,$D4,$7F,$8D,$4F
	.dc.b	$F1,$A6,$1F,$EC,$77,$1F,$09,$CF,$3F,$BF,$97,$6B,$FD,$53,$CE,$3F
	.dc.b	$07,$8F,$69,$F0,$9D,$A3,$FD,$33,$7D,$A3,$F4,$AF,$67,$C7,$F0,$7A
	.dc.b	$37,$F0,$96,$3C,$FF,$09,$CB,$E0,$F9,$B6,$3D,$7F,$09,$DB,$E1,$3B
	.dc.b	$B6,$3A,$9F,$03,$E1,$3A,$67,$FA,$BC,$D1,$D2,$7E,$2E,$36,$B4,$F4
	.dc.b	$8F,$93,$85,$9E,$4F,$4D,$FE,$55,$DA,$F7,$BC,$E3,$E7,$5D,$9F,$86
	.dc.b	$F3,$8F,$A5,$56,$7E,$3B,$F2,$BE,$12,$47,$DA,$F8,$4B,$8F,$AF,$E0
	.dc.b	$7E,$79,$F4,$3E,$12,$D7,$E0,$3E,$13,$E9,$9F,$53,$E1,$2C,$7E,$13
	.dc.b	$1E,$79,$FE,$0F,$2B,$3F,$63,$1E,$79,$7C,$9C,$AD,$FE,$27,$2E,$73
	.dc.b	$F3,$72,$B3,$F8,$DF,$84,$FC,$1F,$C2,$7E,$11,$B9,$7E,$53,$E0,$FF
	.dc.b	$0F,$F0,$7F,$88,$6C,$7E,$63,$E0,$FF,$16,$7E,$33,$E0,$EC,$7E,$72
	.dc.b	$F9,$E7,$F7,$5B,$97,$E7,$70,$E7,$3F,$81,$6C,$7E,$87,$0E,$73,$F8
	.dc.b	$5C,$2C,$FE,$97,$E1,$3F,$DE,$7C,$DF,$E2,$DA,$79,$AE,$75,$6B,$4D
	.dc.b	$FC,$44,$7F,$C3,$98,$7F,$87,$B0,$FE,$78,$A7,$AF,$C2,$CA,$DF,$C2
	.dc.b	$DB,$DC,$75,$3F,$77,$85,$8E,$77,$DD,$EB,$F0,$97,$45,$FF,$56,$C9
	.dc.b	$97,$59,$CD,$FB,$27,$FC,$5D,$18,$D1,$1C,$2F,$D6,$78,$67,$E3,$38
	.dc.b	$3D,$F8,$F8,$5F,$06,$37,$0E,$D8,$F8,$67,$C9,$9D,$9F,$17,$C1,$FD
	.dc.b	$03,$E8,$FC,$2D,$AF,$97,$F0,$3F,$50,$FA,$BF,$09,$63,$E7,$FC,$0F
	.dc.b	$D8,$F8,$4F,$B2,$D8,$FA,$78,$7D,$CF,$84,$E9,$97,$D4,$79,$3F,$5B
	.dc.b	$1E,$B5,$E0,$3A,$9F,$73,$0E,$AD,$3E,$FF,$C1,$FE,$20,$FC,$4F,$C2
	.dc.b	$58,$FC,$1F,$C2,$7E,$30,$FC,$6F,$C1,$58,$FC,$3F,$C2,$7E,$43,$E0
	.dc.b	$FF,$22,$D8,$EA,$4C,$FF,$17,$87,$57,$F1,$B8,$73,$7F,$1F,$85,$E5
	.dc.b	$FB,$9C,$CC,$7A,$57,$E5,$1B,$3A,$F8,$4F,$D0,$9F,$A2,$F8,$4B,$9F
	.dc.b	$E7,$3E,$13,$F1,$35,$F8,$7F,$A8,$FF,$1A,$4F,$E3,$E7,$2F,$F8,$9A
	.dc.b	$33,$8F,$EA,$9F,$F1,$0C,$67,$E1,$37,$39,$F8,$7B,$DA,$BD,$CF,$E5
	.dc.b	$F3,$F1,$BF,$AC,$CE,$E7,$EE,$CF,$8D,$F8,$30,$FC,$DC,$63,$E1,$B3
	.dc.b	$9C,$D3,$FD,$25,$7F,$E4,$83,$BD,$FE,$8B,$3E,$7F,$0D,$DA,$BD,$8F
	.dc.b	$F0,$6C,$7A,$47,$C5,$39,$E3,$D7,$4F,$F1,$E6,$3D,$0E,$D9,$79,$BE
	.dc.b	$1B,$CF,$F0,$D7,$0F,$1B,$EA,$F8,$39,$1E,$6F,$83,$B8,$F4,$F8,$6F
	.dc.b	$7F,$C3,$7C,$06,$D7,$22,$4F,$8B,$1E,$F1,$F0,$30,$E7,$C3,$F6,$F1
	.dc.b	$B3,$F6,$70,$E7,$87,$78,$F8,$F8,$FC,$EF,$85,$91,$F2,$BE,$16,$D7
	.dc.b	$C0,$F8,$5F,$A6,$7D,$4F,$85,$B5,$F4,$3E,$17,$EB,$9F,$63,$E1,$6C
	.dc.b	$7D,$4C,$7E,$DF,$C2,$F4,$DB,$F2,$0F,$2C,$BB,$57,$DA,$C7,$15,$F9
	.dc.b	$5C,$79,$D7,$D2,$C6,$CF,$E6,$7E,$0B,$F0,$E7,$E2,$3E,$16,$E3,$F8
	.dc.b	$4F,$83,$FC,$5F,$C2,$FE,$31,$B1,$FE,$D9,$FA,$8F,$F7,$63,$7E,$1B
	.dc.b	$F3,$31,$8C,$FF,$CF,$8F,$F8,$95,$FD,$16,$3D,$9F,$C2,$E3,$72,$FD
	.dc.b	$66,$3D,$A7,$F1,$2D,$8F,$E6,$D7,$D5,$B8,$78,$2B,$C2,$F3,$C7,$F6
	.dc.b	$58,$FB,$5F,$CF,$BE,$B7,$F4,$39,$5A,$FC,$BB,$CE,$3D,$87,$BE,$3A
	.dc.b	$BF,$07,$E1,$BD,$FF,$0D,$F0,$1B,$1F,$17,$E1,$BE,$19,$F1,$3E,$16
	.dc.b	$C7,$C9,$F1,$BE,$39,$F2,$3E,$16,$C7,$CD,$C3,$9E,$3F,$2B,$1B,$AF
	.dc.b	$C2,$CB,$9C,$FC,$EC,$AC,$FC,$6C,$B9,$CF,$D2,$CA,$CF,$CA,$F8,$6F
	.dc.b	$AF,$F0,$DF,$61,$BA,$7D,$DF,$86,$FB,$67,$DC,$F8,$6B,$1F,$81,$F8
	.dc.b	$6F,$BE,$7E,$03,$E1,$AC,$7E,$17,$0E,$79,$FE,$0B,$1B,$97,$DA,$CB
	.dc.b	$9C,$FE,$1B,$2B,$3F,$7B,$2E,$73,$F8,$AC,$AC,$FE,$0B,$E1,$BF,$1E
	.dc.b	$7E,$43,$E1,$2E,$9F,$95,$F8,$6F,$C9,$FC,$2F,$E5,$1B,$1F,$99,$F8
	.dc.b	$6F,$CB,$FC,$37,$E6,$1B,$1F,$9D,$C3,$9E,$7F,$9A,$C6,$E5,$F9,$2C
	.dc.b	$79,$CF,$E7,$9B,$1F,$96,$C3,$9C,$FE,$89,$B1,$E9,$C3,$F5,$5F,$0D
	.dc.b	$D1,$FF,$76,$F5,$A3,$D1,$1F,$AF,$D2,$E1,$FA,$C7,$D1,$1F,$B3,$D2
	.dc.b	$CF,$EC,$1F,$D9,$FC,$2F,$F4,$CF,$C0,$3F,$C7,$9E,$9B,$2F,$F1,$DE
	.dc.b	$0C,$3F,$DE,$55,$A3,$FC,$71,$19,$CF,$FC,$4D,$1F,$E3,$D6,$FC,$37
	.dc.b	$5F,$86,$ED,$F0,$DD,$FE,$1A,$E1,$C4,$ED,$9B,$1F,$F1,$8C,$7C,$0D
	.dc.b	$3C,$CF,$A0,$E1,$9F,$F9,$4A,$7C,$99,$77,$9C,$DF,$38,$C5,$7A,$B1
	.dc.b	$F1,$56,$AF,$FC,$79,$3E,$EC,$7E,$27,$C1,$CA,$D7,$C3,$5A,$F8,$7F
	.dc.b	$07,$F2,$7E,$1B,$E5,$36,$BA,$7C,$37,$CD,$3E,$77,$C2,$D8,$F0,$63
	.dc.b	$CF,$87,$E4,$E5,$67,$E1,$E8,$C7,$A6,$7E,$09,$F4,$67,$CE,$3E,$AE
	.dc.b	$7D,$A7,$C2,$7B,$7E,$1B,$EF,$7C,$37,$DF,$6C,$7C,$1F,$85,$FC,$11
	.dc.b	$F8,$3F,$85,$B1,$F1,$7E,$17,$F0,$C7,$E1,$FE,$16,$C7,$53,$E5,$61
	.dc.b	$D5,$A7,$52,$65,$F8,$AB,$EA,$FE,$32,$F9,$BF,$8E,$BB,$BF,$E3,$30
	.dc.b	$F3,$CF,$E4,$B2,$B3,$F9,$0C,$B9,$BF,$92,$C6,$E5,$F2,$B2,$E6,$FC
	.dc.b	$CC,$AE,$3F,$89,$F8,$6F,$D1,$1F,$A3,$F8,$4B,$5F,$85,$F8,$5F,$D3
	.dc.b	$7C,$27,$E9,$DB,$1F,$8E,$F7,$FF,$C5,$87,$F8,$98,$BF,$05,$CB,$1F
	.dc.b	$D4,$E3,$D6,$BF,$20,$F3,$7F,$25,$85,$9F,$D7,$BF,$E4,$0B,$E7,$3D
	.dc.b	$FE,$13,$18,$67,$E1,$A4,$F8,$FE,$1A,$E2,$E0,$FF,$4D,$62,$F8,$4E
	.dc.b	$87,$5F,$85,$B8,$7A,$E7,$C0,$78,$7E,$16,$C7,$BF,$2E,$B8,$74,$CA
	.dc.b	$CF,$EE,$AF,$B3,$F1,$71,$EC,$D3,$B1,$32,$F3,$3C,$71,$EF,$18,$3F
	.dc.b	$D9,$F4,$F8,$5F,$88,$7C,$5F,$85,$B7,$F4,$FE,$0F,$E4,$7C,$2F,$C9
	.dc.b	$6C,$7D,$7F,$85,$F9,$9F,$09,$F3,$5B,$1F,$6F,$0F,$A1,$F0,$9D,$33
	.dc.b	$F1,$79,$0D,$13,$D6,$FF,$05,$9D,$BD,$D2,$3B,$4F,$E1,$F2,$B5,$A7
	.dc.b	$B7,$07,$D6,$C2,$D7,$CB,$7A,$47,$DA,$BB,$3F,$3D,$E7,$1F,$7A,$AC
	.dc.b	$FD,$37,$F1,$5F,$09,$F8,$BF,$85,$B3,$F9,$97,$F1,$DF,$05,$F8,$FF
	.dc.b	$85,$B3,$F9,$D7,$F2,$5F,$0B,$F9,$3F,$84,$B3,$FA,$17,$A3,$CC,$FD
	.dc.b	$26,$18,$D7,$E1,$E1,$D6,$7F,$50,$D8,$F8,$F7,$D6,$7F,$58,$D8,$FA
	.dc.b	$3F,$80,$FF,$8F,$67,$FA,$3B,$FE,$49,$E0,$39,$94,$63,$E1,$BF,$D5
	.dc.b	$32,$F8,$5B,$F0,$D8,$6E,$1F,$D2,$25,$F0,$DC,$3E,$1B,$8B,$67,$58
	.dc.b	$74,$3A,$E5,$63,$96,$1D,$23,$D2,$DE,$24,$FA,$AF,$A5,$7B,$9B,$3A
	.dc.b	$8F,$5E,$2F,$86,$F5,$9E,$CF,$86,$B8,$7A,$7E,$17,$DE,$7C,$0F,$86
	.dc.b	$B1,$ED,$F8,$6F,$86,$7C,$4F,$86,$B1,$F0,$70,$E7,$97,$D0,$C6,$E7
	.dc.b	$E3,$E1,$F9,$F9,$F4,$AF,$AB,$9D,$AD,$68,$D3,$AD,$7D,$AD,$2D,$6B
	.dc.b	$3F,$AB,$F0,$DD,$B5,$FB,$AF,$37,$EF,$E7,$EB,$7D,$C7,$AC,$98,$75
	.dc.b	$6D,$F8,$BE,$1B,$F0,$3F,$0D,$F8,$26,$CF,$DF,$F8,$6F,$C2,$FC,$37
	.dc.b	$E1,$9B,$1F,$83,$F8,$4F,$C4,$9F,$8A,$F8,$6B,$1F,$87,$BE,$79,$7E
	.dc.b	$59,$B9,$7C,$2C,$39,$CF,$E6,$9B,$1F,$1B,$0E,$73,$F9,$EC,$2C,$FC
	.dc.b	$AF,$85,$FC,$CF,$C3,$7E,$69,$B9,$7E,$5F,$E1,$3F,$0D,$5F,$57,$F0
	.dc.b	$2F,$F1,$DF,$27,$F8,$EB,$2F,$E3,$31,$FE,$3E,$3F,$C7,$2E,$3E,$1B
	.dc.b	$8E,$1F,$E3,$29,$CD,$F3,$9F,$04,$FE,$97,$2B,$7F,$B4,$7F,$5B,$8F
	.dc.b	$28,$FD,$8E,$1E,$28,$C7,$3A,$3C,$5F,$0D,$87,$FC,$91,$87,$86,$31
	.dc.b	$8E,$7E,$1B,$5F,$0D,$6B,$FC,$7B,$7F,$E4,$6C,$BB,$FC,$2F,$33,$A7
	.dc.b	$C2,$DA,$DF,$C2,$77,$F8,$4F,$03,$63,$9E,$5D,$9A,$76,$CB,$96,$3F
	.dc.b	$2B,$1E,$B3,$E2,$79,$BE,$4C,$2C,$FD,$07,$D0,$FD,$2A,$BC,$1F,$5D
	.dc.b	$EA,$FD,$9B,$EB,$3D,$EF,$D2,$75,$2B,$D5,$F8,$7F,$09,$F2,$CF,$99
	.dc.b	$F0,$D6,$3E,$3F,$C2,$7C,$FF,$85,$FA,$0D,$8F,$97,$F0,$BF,$4F,$E1
	.dc.b	$7E,$A3,$63,$A9,$31,$FC,$5E,$1D,$5F,$C6,$DF,$36,$CF,$D6,$C3,$9B
	.dc.b	$F6,$30,$B3,$F9,$47,$F0,$5F,$0B,$F8,$3F,$85,$B5,$F7,$9F,$C3,$7C
	.dc.b	$1F,$E1,$FE,$16,$CF,$E0,$9F,$C5,$7C,$2C,$8F,$C2,$FC,$2D,$CB,$F4
	.dc.b	$38,$F5,$9F,$C4,$B6,$3F,$4B,$87,$59,$FC,$6B,$63,$F5,$37,$CF,$87
	.dc.b	$F5,$D8,$5B,$FF,$4C,$FE,$06,$F8,$6F,$CF,$B6,$3F,$D2,$6B,$E1,$B3
	.dc.b	$3A,$F8,$59,$5A,$F8,$5B,$9F,$F8,$43,$0F,$F1,$5B,$7E,$1B,$AF,$C3
	.dc.b	$76,$79,$DF,$6C,$7C,$39,$62,$F9,$65,$D3,$4F,$E7,$59,$33,$C6,$3E
	.dc.b	$AC,$FA,$3E,$86,$E1,$E3,$CB,$DB,$F0,$3D,$71,$E6,$F2,$9F,$EE,$38
	.dc.b	$B2,$F8,$79,$67,$C2,$4F,$B3,$0F,$1E,$BE,$CD,$3E,$3B,$F2,$BE,$1B
	.dc.b	$E5,$FC,$1D,$9F,$96,$FC,$EF,$86,$91,$F3,$BE,$1A,$EB,$F2,$7E,$13
	.dc.b	$E9,$9F,$53,$E1,$AD,$7D,$4C,$7A,$B4,$EA,$7D,$8C,$39,$E7,$F2,$31
	.dc.b	$B3,$F8,$0C,$39,$E3,$F5,$F1,$B7,$F3,$BE,$17,$F0,$67,$E1,$3E,$1A
	.dc.b	$E1,$F8,$4F,$83,$FC,$39,$F8,$8F,$86,$B1,$F8,$8F,$84,$FC,$5F,$C3
	.dc.b	$7E,$31,$B1,$F8,$CC,$3F,$1F,$F0,$BD,$34,$FC,$43,$C9,$FC,$56,$3D
	.dc.b	$6B,$C0,$75,$3F,$2B,$87,$56,$9F,$98,$F8,$6F,$E7,$3E,$6F,$F8,$B7
	.dc.b	$9B,$EC,$C6,$30,$CF,$C2,$6B,$E1,$2E,$1F,$EF,$2C,$3F,$C6,$0D,$F8
	.dc.b	$6E,$67,$4F,$85,$B1,$DC,$E1,$87,$EA,$B1,$ED,$1C,$F1,$EC,$D3,$B6
	.dc.b	$3C,$AB,$FC,$EF,$8B,$1E,$D5,$E1,$3A,$92,$7F,$5D,$87,$67,$8E,$1D
	.dc.b	$9B,$3D,$8E,$4F,$6A,$EA,$D8,$F8,$1F,$0D,$F2,$0F,$93,$F0,$56,$3E
	.dc.b	$27,$C3,$7C,$CF,$82,$F9,$AD,$8F,$91,$F0,$DF,$43,$E1,$BE,$8B,$63
	.dc.b	$A9,$31,$F7,$E1,$D5,$F8,$37,$CD,$B3,$F4,$F0,$E6,$FD,$5C,$2C,$FC
	.dc.b	$87,$EF,$FC,$37,$E0,$3E,$1A,$D7,$DB,$7F,$07,$F0,$92,$3E,$F7,$C3
	.dc.b	$5C,$3E,$8F,$C2,$FE,$20,$FC,$4F,$C3,$5A,$E4,$49,$FA,$F8,$F7,$8F
	.dc.b	$C4,$61,$CF,$BB,$FC,$A1,$C8,$CB,$14,$C3,$FC,$FF,$2C,$BB,$47,$E0
	.dc.b	$B2,$B8,$72,$F8,$6F,$CD,$9F,$9C,$F8,$6B,$8F,$E5,$FE,$1B,$F3,$FF
	.dc.b	$0B,$FA,$06,$C7,$E6,$FE,$17,$FC,$37,$F8,$27,$F9,$EA,$30,$78,$F1
	.dc.b	$FF,$33,$46,$73,$FC,$9E,$5D,$9F,$CA,$E5,$6B,$DB,$8F,$69,$FC,$CB
	.dc.b	$63,$E0,$E3,$E0,$C7,$F6,$6D,$7F,$5D,$87,$93,$E1,$BA,$47,$C5,$EE
	.dc.b	$67,$FE,$3C,$8C,$6E,$CB,$B4,$61,$B9,$FC,$CC,$BB,$3F,$AC,$6E,$7E
	.dc.b	$EF,$86,$F8,$7F,$09,$F1,$1B,$87,$C3,$3C,$59,$7F,$8D,$F3,$F1,$FC
	.dc.b	$1F,$CA,$F8,$3F,$96,$FC,$CD,$7E,$6F,$C2,$DC,$39,$6B,$FE,$BC,$C5
	.dc.b	$3D,$A3,$FD,$EB,$0C,$FC,$55,$C3,$2F,$83,$1D,$A3,$13,$E3,$19,$66
	.dc.b	$F9,$E3,$E2,$AD,$B6,$3B,$FC,$2F,$E0,$8F,$C1,$FC,$2D,$AE,$BF,$09
	.dc.b	$F8,$6F,$84,$FC,$3B,$63,$C3,$F0,$9F,$8A,$3F,$17,$F0,$B6,$3C,$B8
	.dc.b	$7E,$3B,$E1,$7A,$65,$EE,$79,$CF,$AB,$1B,$3F,$09,$E4,$FC,$4C,$3A
	.dc.b	$CF,$80,$F8,$3F,$07,$F9,$AF,$85,$FC,$DB,$63,$E2,$FC,$27,$E7,$BE
	.dc.b	$17,$F3,$ED,$8F,$93,$F0,$BF,$A2,$3F,$47,$F0,$B6,$3E,$6E,$1C,$F2
	.dc.b	$FA,$B8,$DE,$9F,$D2,$E3,$CE,$7E,$CB,$63,$F5,$38,$F3,$9F,$BB,$8D
	.dc.b	$9F,$D6,$E1,$CF,$1F,$C6,$E3,$6F,$F0,$D8,$F3,$9F,$C8,$B6,$3F,$15
	.dc.b	$8F,$39,$FC,$AE,$36,$5C,$CF,$C1,$7F,$C7,$50,$CF,$FD,$C5,$1F,$E3
	.dc.b	$BB,$FC,$73,$A8,$CF,$C3,$7F,$A8,$23,$FC,$45,$3A,$F8,$2C,$7C,$1E
	.dc.b	$5D,$7C,$36,$FE,$0E,$D7,$47,$96,$12,$3B,$FC,$35,$CF,$C9,$1F,$BF
	.dc.b	$CB,$BC,$60,$F5,$BF,$E3,$F8,$F3,$E5,$E1,$7D,$38,$F8,$5A,$78,$4F
	.dc.b	$6F,$C3,$7A,$FE,$13,$D8,$D8,$F8,$3F,$0B,$EF,$F8,$5F,$80,$D8,$F8
	.dc.b	$BF,$0B,$F0,$FE,$17,$E2,$36,$3A,$9F,$2B,$0E,$AD,$3A,$9F,$3A,$FA
	.dc.b	$B4,$EA,$7D,$2A,$EA,$D3,$EA,$FC,$2F,$D3,$3E,$A7,$C1,$58,$FB,$3F
	.dc.b	$0B,$F5,$CF,$B1,$F0,$56,$3E,$EF,$C2,$FD,$BF,$85,$FB,$8D,$8E,$A7
	.dc.b	$E0,$BE,$13,$A7,$1F,$D1,$E9,$FC,$54,$73,$8C,$72,$32,$EF,$7E,$23
	.dc.b	$B9,$F8,$CC,$7B,$B4,$EF,$5C,$AF,$BF,$4F,$E5,$1E,$CF,$E5,$AF,$AC
	.dc.b	$F8,$1E,$6F,$8A,$EC,$FE,$71,$FC,$C7,$C2,$CC,$3F,$3B,$F0,$76,$FD
	.dc.b	$9F,$83,$7F,$90,$C9,$F0,$9F,$EF,$E7,$1F,$0B,$7E,$17,$0D,$C3,$F4
	.dc.b	$F8,$EF,$E1,$BA,$68,$9F,$EB,$18,$27,$A6,$52,$7F,$9E,$4B,$2C,$4F
	.dc.b	$93,$1E,$CD,$AD,$E5,$E5,$F8,$69,$1C,$BE,$16,$D7,$B3,$E1,$7D,$47
	.dc.b	$AF,$E1,$6D,$78,$3E,$1B,$DD,$F0,$DE,$F6,$C7,$93,$1F,$85,$F0,$BD
	.dc.b	$38,$3D,$99,$7F,$0C,$DD,$3D,$23,$C1,$7E,$CC,$B1,$7C,$39,$D8,$32
	.dc.b	$EF,$87,$D2,$6D,$6E,$5F,$0D,$F4,$CF,$A9,$F0,$D7,$2F,$93,$F0,$DF
	.dc.b	$5F,$E1,$BE,$C3,$63,$E6,$FC,$37,$DB,$3E,$E7,$C2,$58,$FA,$3F,$0B
	.dc.b	$F7,$CE,$9A,$FE,$11,$E7,$1F,$5B,$0B,$3F,$88,$79,$3F,$8A,$BE,$B3
	.dc.b	$E0,$3A,$9F,$7B,$E1,$3B,$34,$F4,$9D,$4B,$3D,$5E,$4F,$57,$F0,$F8
	.dc.b	$78,$35,$F8,$DF,$0D,$F9,$7F,$85,$FC,$C3,$6B,$FC,$11,$CD,$FE,$39
	.dc.b	$AF,$A4,$D8,$FF,$16,$3F,$E2,$BC,$3F,$A3,$B9,$6F,$C2,$FE,$02,$B8
	.dc.b	$B6,$B8,$FC,$1D,$AD,$C7,$E7,$31,$C7,$23,$1F,$F1,$47,$1A,$73,$1F
	.dc.b	$E2,$B3,$C1,$3B,$AC,$FC,$36,$9D,$E5,$1F,$6E,$57,$2F,$E2,$D8,$7F
	.dc.b	$88,$A7,$E0,$FC,$17,$6F,$83,$EE,$F8,$3E,$16,$4F,$C6,$F8,$1B,$87
	.dc.b	$2F,$84,$F2,$9E,$6F,$84,$B5,$D0,$F9,$BF,$05,$D1,$A7,$D0,$C3,$9F
	.dc.b	$17,$D0,$C6,$CF,$D0,$C3,$9E,$3E,$8C,$6D,$FD,$6F,$82,$F8,$A7,$C6
	.dc.b	$F8,$4B,$87,$DC,$F8,$4F,$93,$F0,$5F,$29,$B1,$F8,$0F,$83,$F9,$A7
	.dc.b	$CE,$F8,$3B,$1F,$84,$F8,$4F,$A2,$74,$D3,$E4,$BC,$E3,$F1,$38,$59
	.dc.b	$F9,$AF,$27,$E7,$DF,$59,$F0,$1D,$4F,$C8,$FC,$27,$66,$9E,$93,$A9
	.dc.b	$67,$AB,$C9,$EA,$FE,$6B,$0F,$06,$9F,$9A,$F8,$4F,$C4,$9F,$8E,$FC
	.dc.b	$1F,$FC,$31,$5F,$9C,$8C,$7C,$1C,$8F,$F3,$54,$FF,$83,$2B,$F8,$A3
	.dc.b	$FE,$21,$C3,$FA,$F3,$FE,$0D,$6F,$C0,$F5,$F8,$5E,$CF,$7F,$85,$98
	.dc.b	$7E,$BB,$2B,$7C,$B2,$F0,$3D,$3E,$12,$E7,$E0,$CB,$9E,$1E,$46,$D7
	.dc.b	$8B,$E1,$3D,$9F,$05,$ED,$6D,$7A,$BE,$13,$E0,$7C,$1F,$C1,$6C,$7B
	.dc.b	$BE,$13,$E2,$7C,$27,$C5,$6C,$7C,$2C,$3E,$47,$C2,$74,$CB,$E0,$BC
	.dc.b	$E7,$E3,$E3,$67,$E2,$BC,$E3,$E5,$E1,$67,$E4,$BF,$4F,$E1,$3E,$A7
	.dc.b	$C0,$5A,$FA,$2F,$D7,$F8,$4F,$B1,$F0,$96,$7E,$AB,$F6,$FE,$0A,$47
	.dc.b	$D8,$F8,$3B,$97,$D6,$C3,$F0,$1F,$05,$DA,$BE,$EB,$CE,$7E,$DE,$17
	.dc.b	$8F,$EF,$BC,$9F,$C0,$E3,$F8,$9C,$71,$87,$FB,$3F,$A5,$8F,$68,$F4
	.dc.b	$9D,$8F,$C5,$E7,$EA,$69,$EA,$3F,$1F,$7D,$32,$FF,$39,$DB,$0C,$79
	.dc.b	$A8,$7C,$1F,$E6,$CF,$E5,$3F,$84,$2F,$18,$FF,$B0,$E3,$FC,$10,$DF
	.dc.b	$82,$DF,$C1,$F0,$78,$FC,$1F,$2F,$84,$B5,$C5,$E7,$87,$F8,$F7,$07
	.dc.b	$C1,$63,$87,$FB,$C4,$62,$FC,$19,$75,$69,$E1,$3C,$98,$F7,$69,$DC
	.dc.b	$F4,$7C,$27,$B0,$F6,$FA,$56,$3D,$9F,$05,$F0,$0F,$83,$F0,$56,$3E
	.dc.b	$07,$C2,$7C,$43,$E2,$FC,$25,$8F,$89,$F0,$9F,$20,$E9,$BF,$B5,$E4
	.dc.b	$FB,$F0,$EB,$3E,$03,$A9,$23,$E2,$57,$57,$E3,$7C,$27,$D4,$F8,$4F
	.dc.b	$AA,$DA,$FA,$9F,$07,$F6,$0F,$B3,$F0,$76,$3E,$C7,$C1,$7D,$C3,$EE
	.dc.b	$FC,$25,$8F,$B9,$87,$3C,$7B,$C6,$0E,$A4,$7E,$C5,$F5,$7E,$D5,$F3
	.dc.b	$7E,$E5,$DE,$4F,$B7,$F0,$9F,$8C,$3B,$38,$C3,$A3,$F8,$0C,$BD,$31
	.dc.b	$83,$F2,$5D,$6E,$66,$5D,$67,$30,$C7,$19,$7F,$B4,$A1,$F0,$9F,$9C
	.dc.b	$3F,$3B,$F0,$76,$FF,$3D,$9F,$F9,$1F,$F0,$9F,$F0,$19,$FE,$15,$CC
	.dc.b	$E3,$E1,$3F,$C1,$AB,$E1,$29,$8F,$84,$B8,$7E,$9F,$2D,$FC,$4F,$7D
	.dc.b	$7F,$D3,$98,$23,$9D,$74,$C6,$BE,$43,$E0,$CF,$8C,$F0,$39,$AF,$F7
	.dc.b	$7C,$61,$EB,$87,$89,$EF,$F0,$BE,$8F,$85,$F4,$B6,$F9,$69,$EE,$F8
	.dc.b	$4F,$69,$EE,$F8,$4B,$3A,$C9,$F0,$9F,$08,$F8,$7F,$09,$67,$59,$7C
	.dc.b	$6F,$85,$ED,$A7,$C0,$79,$BF,$0B,$2E,$D5,$D5,$E5,$1F,$1B,$0E,$B3
	.dc.b	$E1,$7E,$8F,$C2,$C8,$F2,$7C,$2D,$AF,$9B,$F0,$BF,$58,$FA,$FF,$09
	.dc.b	$6B,$D5,$F0,$9F,$68,$FB,$7F,$07,$63,$DD,$F0,$BF,$78,$FB,$FF,$0B
	.dc.b	$63,$E1,$61,$CF,$1F,$B7,$8D,$EA,$FC,$56,$5C,$E7,$EF,$E5,$67,$F1
	.dc.b	$D9,$7E,$2B,$E0,$FA,$57,$E1,$1E,$59,$7E,$03,$2F,$4B,$E0,$3D,$24
	.dc.b	$9E,$AD,$BF,$F5,$07,$BB,$FE,$14,$97,$32,$30,$DF,$83,$FF,$33,$C6
	.dc.b	$0F,$F0,$B5,$FE,$C1,$FF,$0A,$5F,$4F,$83,$DF,$C1,$F0,$F8,$59,$7F
	.dc.b	$8A,$B7,$C2,$E3,$3F,$F7,$D4,$75,$F8,$5E,$FF,$11,$E0,$7C,$39,$74
	.dc.b	$C3,$93,$6F,$8E,$3C,$F3,$F6,$E5,$6F,$CD,$97,$A8,$F5,$FC,$35,$8F
	.dc.b	$27,$C2,$7B,$8F,$7F,$C3,$58,$F4,$7C,$2F,$C2,$3E,$1F,$C3,$58,$F6
	.dc.b	$61,$F1,$BE,$1B,$A6,$BE,$C7,$93,$EE,$C7,$AD,$78,$0E,$A4,$8F,$87
	.dc.b	$7D,$5F,$8B,$F0,$DF,$48,$FA,$7F,$09,$6B,$E6,$7C,$37,$D6,$F8,$5F
	.dc.b	$AE,$D8,$FA,$1F,$0B,$F6,$8F,$B7,$F0,$D6,$3E,$A7,$C2,$FD,$E3,$FD
	.dc.b	$4B,$F8,$57,$F7,$D8,$C9,$D8,$FB,$58,$76,$69,$D8,$99,$FD,$CB,$EC
	.dc.b	$B9,$3F,$DE,$17,$FE,$06,$8C,$57,$F3,$A8,$C4,$FE,$5B,$0F,$1F,$C1
	.dc.b	$F7,$BF,$C3,$E3,$6B,$F3,$6F,$67,$F3,$B9,$7B,$B1,$F1,$C7,$83,$E1
	.dc.b	$BD,$A7,$BB,$E1,$2C,$FF,$1E,$CD,$F0,$DF,$08,$F8,$7F,$0D,$67,$5F
	.dc.b	$0D,$F1,$8F,$8F,$F0,$D6,$39,$7C,$1F,$CA,$3A,$ED,$EE,$3F,$5F,$8A
	.dc.b	$3A,$46,$7C,$34,$F8,$AB,$C9,$8E,$B5,$F9,$39,$77,$C3,$85,$FF,$B5
	.dc.b	$5B,$1F,$6B,$E1,$BE,$DF,$C3,$5C,$7D,$CF,$DE,$F8,$5F,$BF,$F0,$D6
	.dc.b	$7E,$13,$F8,$2F,$86,$FC,$1F,$C3,$59,$F8,$CF,$E1,$BE,$12,$47,$C9
	.dc.b	$F8,$6B,$A7,$93,$E1,$7F,$16,$7E,$33,$E1,$6D,$7C,$FF,$86,$FC,$79
	.dc.b	$F9,$0F,$83,$B1,$F4,$FE,$17,$F2,$67,$FB,$E3,$F0,$BF,$F7,$9F,$1F
	.dc.b	$FB,$DE,$7F,$AA,$C7,$F7,$32,$7C,$36,$FE,$17,$87,$C2,$F1,$F8,$6B
	.dc.b	$7C,$1E,$9F,$0D,$23,$C0,$74,$79,$3D,$AB,$C5,$78,$DF,$EC,$62,$C7
	.dc.b	$B5,$F9,$1B,$3A,$F8,$3F,$57,$C1,$7A,$DB,$5E,$AF,$86,$F7,$7C,$1F
	.dc.b	$BD,$B1,$EE,$F8,$5F,$85,$F0,$BF,$0D,$B1,$F0,$BE,$17,$E3,$1F,$1F
	.dc.b	$E1,$6C,$7C,$6F,$7F,$E5,$7C,$27,$CB,$6C,$7C,$AF,$84,$F9,$DF,$09
	.dc.b	$F3,$DB,$1F,$3B,$E1,$3E,$97,$C2,$FD,$36,$C7,$D2,$F8,$53,$FE,$08
	.dc.b	$FC,$33,$FA,$5F,$2A,$73,$C9,$F9,$F9,$FE,$9E,$7F,$B7,$FC,$9C,$F9
	.dc.b	$68,$FE,$9F,$3F,$D5,$9D,$61,$FE,$40,$73,$3B,$F8,$5E,$B7,$CA,$7F
	.dc.b	$AA,$D6,$FE,$0B,$2F,$1A,$B8,$75,$F8,$2B,$8F,$87,$E1,$2E,$3F,$D6
	.dc.b	$30,$F6,$46,$3E,$02,$47,$BF,$E0,$EE,$3E,$DF,$80,$E3,$F0,$7C,$9B
	.dc.b	$5F,$17,$E0,$FA,$FC,$17,$66,$C7,$C9,$F8,$3F,$09,$E2,$F8,$3B,$1F
	.dc.b	$37,$E0,$FC,$BF,$07,$E6,$6C,$7D,$1F,$80,$F4,$FC,$07,$A9,$B1,$F5
	.dc.b	$7E,$0F,$DB,$F0,$7E,$E6,$C7,$F9,$03,$C8,$FE,$85,$C1,$FC,$55,$FE
	.dc.b	$6B,$1F,$E7,$07,$3D,$3F,$61,$CC,$FF,$83,$17,$C1,$7C,$C5,$F0,$56
	.dc.b	$30,$E7,$E0,$B5,$F0,$17,$1F,$13,$C7,$E0,$39,$7C,$15,$9F,$33,$D7
	.dc.b	$E0,$BB,$7C,$15,$9F,$53,$E1,$F8,$2F,$17,$C1,$59,$F7,$3E,$5F,$82
	.dc.b	$F3,$7C,$15,$9F,$84,$FE,$1B,$83,$FC,$F9,$D7,$EB,$F8,$2C,$7A,$E8
	.dc.b	$C5,$FC,$9F,$82,$F8,$1F,$05,$F0,$5B,$1F,$37,$E0,$BE,$27,$C1,$7C
	.dc.b	$56,$C7,$F9,$E7,$2F,$F4,$57,$81,$F2,$7E,$0B,$BB,$67,$EB,$7C,$17
	.dc.b	$CD,$F8,$2F,$9C,$D8,$FF,$3C,$FE,$1B,$FB,$46,$FC,$07,$E7,$A3,$18
	.dc.b	$FF,$51,$51,$EC,$F1,$75,$C1,$6F,$9B,$C7,$E0,$24,$76,$F8,$2B,$8F
	.dc.b	$3F,$82,$ED,$F0,$5D,$DB,$5E,$DC,$1E,$47,$8F,$C3,$F9,$6D,$9F,$4F
	.dc.b	$C0,$F9,$CF,$47,$C1,$58,$F6,$FC,$0F,$AF,$E0,$BD,$8D,$8E,$DE,$6F
	.dc.b	$BF,$E0,$BE,$BB,$63,$E2,$FC,$17,$C3,$3E,$27,$C0,$D8,$F9,$3F,$05
	.dc.b	$F1,$FE,$07,$E4,$36,$3D,$5E,$8F,$CB,$F8,$2F,$C1,$B6,$3E,$8F,$C0
	.dc.b	$FC,$F3,$E8,$7C,$15,$8F,$AB,$F0,$3F,$C4,$3F,$0E,$FF,$38,$F6,$FF
	.dc.b	$27,$9F,$F5,$8C,$7F,$9D,$FA,$7F,$D9,$9E,$BE,$FE,$23,$F2,$2D,$AD
	.dc.b	$FC,$3F,$33,$A7,$C1,$D8,$E7,$F0,$9D,$FE,$13,$C0,$D8,$FC,$87,$EF
	.dc.b	$E8,$FD,$03,$63,$33,$89,$F4,$7C,$1F,$A7,$E0,$ED,$7A,$1F,$67,$C1
	.dc.b	$FB,$7E,$0A,$CF,$B1,$FD,$4F,$C0,$7C,$1F,$83,$B3,$F0,$1F,$89,$F0
	.dc.b	$72,$3E,$1F,$A9,$78,$BF,$C5,$1C,$6F,$3F,$E5,$79,$9F,$2D,$B7,$F8
	.dc.b	$FF,$83,$F9,$DE,$3F,$EE,$1B,$1F,$3B,$C2,$FA,$47,$D3,$ED,$B1,$F4
	.dc.b	$BA,$3E,$B6,$FF,$5D,$B1,$FB,$7F,$B6,$FE,$46,$B5,$FF,$56,$EA,$CB
	.dc.b	$F8,$55,$6A,$73,$1F,$D4,$3F,$10,$C7,$C3,$9A,$7F,$2C,$71,$3B,$F8
	.dc.b	$5E,$1F,$0B,$72,$FF,$7B,$DF,$F2,$E7,$C9,$F1,$5D,$7E,$1B,$C2,$D8
	.dc.b	$DF,$C3,$78,$4F,$17,$C3,$58,$E7,$F0,$DE,$5F,$86,$F3,$36,$3E,$06
	.dc.b	$7E,$9F,$86,$F6,$B6,$3C,$7F,$0D,$ED,$3D,$DF,$0D,$63,$CF,$F0,$DF
	.dc.b	$07,$E1,$BE,$13,$63,$E6,$70,$7C,$5F,$86,$F9,$2D,$8F,$7F,$C2,$FC
	.dc.b	$93,$E5,$7C,$35,$8F,$87,$F0,$BF,$37,$E1,$BE,$73,$63,$EC,$72,$7D
	.dc.b	$1F,$86,$FA,$AD,$8F,$97,$F0,$BF,$54,$FA,$DF,$0D,$63,$EF,$7C,$3F
	.dc.b	$F8,$63,$97,$F9,$5B,$7E,$1F,$FC,$17,$18,$9F,$C2,$76,$7F,$A0,$67
	.dc.b	$F0,$7D,$3A,$F8,$3B,$5D,$0D,$DA,$39,$FC,$3C,$9F,$C8,$55,$EE,$FF
	.dc.b	$45,$5B,$E1,$FC,$3F,$0D,$E2,$6D,$FE,$5F,$BB,$CB,$F0,$FF,$94,$6C
	.dc.b	$64,$F1,$60,$D7,$FD,$31,$F0,$FB,$9C,$C6,$1B,$93,$E1,$FE,$0F,$C3
	.dc.b	$FC,$26,$CF,$E9,$FC,$2F,$8B,$F0,$DF,$A4,$6C,$64,$E7,$F0,$F9,$69
	.dc.b	$DB,$E1,$FE,$6F,$C3,$FC,$E6,$C7,$EC,$F2,$FA,$3F,$0F,$FB,$06,$C6
	.dc.b	$4F,$3F,$C3,$65,$A7,$AB,$E1,$FE,$CF,$C3,$FD,$A6,$C7,$EF,$FC,$AF
	.dc.b	$BB,$F0,$FF,$BC,$6C,$64,$F8,$7F,$0F,$96,$9F,$1B,$E1,$FF,$C9,$FF
	.dc.b	$88,$BE,$33,$F0,$F1,$87,$F8,$3F,$9D,$AF,$88,$B7,$F4,$2F,$FC,$C7
	.dc.b	$D3,$FC,$3F,$D4,$5F,$4F,$88,$DC,$FF,$B3,$6B,$FC,$D6,$A9,$E8,$F8
	.dc.b	$BE,$23,$F8,$AB,$71,$FE,$38,$F9,$BE,$1B,$31,$E1,$6D,$F5,$F8,$7B
	.dc.b	$E4,$7F,$9E,$B2,$FF,$34,$4F,$FB,$84,$F0,$7A,$7F,$03,$E1,$FF,$91
	.dc.b	$B7,$1F,$E5,$4F,$C4,$F8,$7C,$C7,$BD,$B5,$EB,$F8,$7B,$8F,$77,$F9
	.dc.b	$6F,$AF,$F3,$3E,$1E,$E1,$F1,$DB,$7F,$0F,$E1,$EE,$1E,$76,$D7,$ED
	.dc.b	$B4,$7C,$3E,$2F,$F4,$51,$89,$FE,$81,$ED,$7D,$AF,$87,$FE,$7A,$D8
	.dc.b	$C9,$F5,$FE,$1B,$2D,$33,$7F,$57,$E1,$F2,$D8,$FE,$99,$EE,$7E,$1B
	.dc.b	$E1,$FF,$A4,$B6,$32,$7E,$0F,$E1,$F2,$D3,$35,$F8,$1F,$87,$CB,$63
	.dc.b	$FA,$C7,$BF,$F9,$2F,$86,$FE,$AA,$D8,$C9,$F8,$FF,$87,$CB,$4C,$D7
	.dc.b	$E3,$7E,$1B,$2D,$8F,$EC,$9F,$03,$FE,$AC,$FC,$4B,$FB,$04,$FE,$46
	.dc.b	$30,$FF,$64,$75,$F0,$D6,$BF,$86,$D7,$FA,$32,$B8,$46,$3E,$1B,$50
	.dc.b	$E7,$FF,$64,$D7,$FA,$42,$BF,$BA,$7C,$2F,$87,$E1,$BF,$B8,$B6,$32
	.dc.b	$76,$F8,$5C,$B4,$CE,$1D,$5F,$E7,$36,$F8,$0D,$CE,$4F,$E4,$D1,$99
	.dc.b	$FF,$02,$7C,$37,$C1,$F7,$BF,$C0,$2D,$8C,$9E,$EF,$67,$2D,$33,$7D
	.dc.b	$7D,$D6,$0F,$5B,$73,$98,$C5,$7F,$84,$FE,$1B,$E8,$7A,$7F,$E1,$06
	.dc.b	$C6,$4F,$9B,$E7,$E5,$A6,$6F,$A7,$99,$96,$C7,$F8,$7F,$E1,$BE,$E7
	.dc.b	$95,$FE,$1C,$6C,$64,$FB,$3E,$46,$5A,$66,$BD,$1E,$1E,$5B,$1F,$EA
	.dc.b	$1F,$B9,$FE,$D1,$1F,$33,$C1,$FF,$59,$46,$38,$3E,$1C,$62,$BF,$1D
	.dc.b	$DD,$5F,$C4,$B6,$FE,$37,$65,$EA,$FC,$3B,$D3,$E1,$FF,$2D,$D5,$70
	.dc.b	$FC,$8B,$6F,$E7,$73,$DC,$3F,$1E,$DA,$FC,$2F,$1D,$AF,$C1,$BE,$8F
	.dc.b	$87,$FD,$17,$0D,$AF,$CE,$B6,$BE,$F7,$E2,$7F,$01,$70,$7F,$80,$6B
	.dc.b	$FB,$7C,$E6,$31,$3C,$34,$E5,$9A,$FF,$C1,$F7,$FE,$BC,$7E,$67,$C4
	.dc.b	$66,$FB,$E5,$F3,$23,$17,$9E,$16,$1D,$A3,$85,$F9,$2B,$78,$76,$79
	.dc.b	$3E,$66,$C7,$B3,$E1,$ED,$75,$BF,$F0,$B5,$FD,$F9,$E7,$5F,$E0,$F9
	.dc.b	$7C,$47,$E0,$27,$38,$F9,$23,$17,$F1,$3E,$22,$E1,$F2,$9B,$8F,$7F
	.dc.b	$88,$B5,$F1,$1F,$C5,$DF,$4C,$BF,$C3,$13,$F9,$18,$C7,$C3,$EA,$3E
	.dc.b	$8F,$C4,$5C,$BE,$BB,$71,$F6,$7C,$45,$AF,$A2,$FE,$63,$3E,$DF,$11
	.dc.b	$67,$ED,$B7,$0F,$C0,$FC,$3D,$C3,$E3,$36,$BE,$DF,$C4,$5A,$F8,$CF
	.dc.b	$E9,$36,$F5,$7C,$45,$AF,$C3,$B6,$BF,$1B,$F1,$17,$1F,$A4,$DA,$FC
	.dc.b	$3F,$C3,$DA,$FA,$4F,$EC,$37,$F8,$DF,$11,$6B,$F2,$6D,$AF,$CC,$FC
	.dc.b	$45,$C7,$EF,$36,$BF,$27,$F1,$16,$BE,$F3,$F8,$05,$F7,$DF,$E0,$8B
	.dc.b	$FC,$37,$C4,$E6,$31,$7F,$9C,$8F,$C1,$D2,$E9,$88,$B3,$FE,$56,$E7
	.dc.b	$FF,$17,$37,$E2,$7F,$17,$1F,$E2,$A3,$FC,$BC,$FF,$0E,$E6,$FC,$7F
	.dc.b	$C0,$FF,$BC,$7B,$91,$89,$D5,$FF,$98,$AB,$FC,$67,$97,$EE,$2B,$8F
	.dc.b	$C4,$EA,$3F,$9D,$4F,$F8,$CB,$0E,$77,$C7,$E2,$7F,$8E,$F5,$78,$3E
	.dc.b	$22,$CF,$56,$E1,$CB,$E2,$2F,$0F,$ED,$F2,$FF,$7B,$4F,$F8,$BF,$26
	.dc.b	$3E,$AF,$88,$DE,$5F,$E9,$2B,$F9,$77,$FE,$D7,$B7,$67,$C0,$F8,$8F
	.dc.b	$E5,$ED,$BF,$E6,$8F,$C4,$F8,$8C,$C7,$C1,$6D,$FB,$B0,$B9,$F9,$5F
	.dc.b	$E7,$3D,$DF,$33,$E2,$2E,$1F,$25,$B7,$F1,$BE,$22,$E1,$ED,$6D,$7A
	.dc.b	$7E,$22,$D7,$FA,$57,$1F,$F1,$A3,$F0,$7C,$1F,$B3,$F0,$FF,$D2,$1B
	.dc.b	$7F,$D3,$5F,$BB,$F0,$F9,$8F,$B4,$DA,$FA,$FF,$0F,$73,$F9,$CF,$F5
	.dc.b	$2F,$0F,$F0,$BF,$0F,$70,$FC,$13,$6F,$EF,$FC,$3D,$C3,$EB,$36,$BE
	.dc.b	$97,$C3,$DA,$F5,$3F,$D7,$BC,$6F,$C8,$FC,$3D,$AF,$C7,$36,$BF,$17
	.dc.b	$F0,$F7,$1F,$C1,$36,$BE,$F7,$C3,$DA,$F8,$CF,$F6,$AF,$27,$F9,$5F
	.dc.b	$E2,$9F,$E3,$4E,$B4,$62,$B5,$7F,$EB,$EA,$FF,$19,$56,$A3,$1F,$11
	.dc.b	$AA,$FF,$0B,$57,$F8,$C5,$FC,$95,$F6,$F8,$8F,$EF,$3E,$5F,$83,$E2
	.dc.b	$2E,$3D,$5B,$7C,$BE,$22,$ED,$FD,$4F,$24,$62,$FF,$DD,$71,$FE,$2F
	.dc.b	$AF,$D0,$61,$EC,$F8,$8F,$F0,$3F,$9D,$EE,$F8,$8B,$87,$AD,$B7,$E8
	.dc.b	$F8,$8B,$9F,$76,$D7,$6F,$88,$B5,$FE,$E3,$CF,$FC,$5E,$FE,$BF,$D0
	.dc.b	$F9,$7F,$0F,$FE,$16,$6D,$FF,$86,$DF,$9F,$F0,$F9,$8F,$98,$DA,$F9
	.dc.b	$3F,$0F,$73,$F8,$0F,$F8,$83,$D3,$FA,$FF,$0F,$70,$FA,$8D,$BF,$A3
	.dc.b	$F0,$F7,$0F,$90,$DA,$F8,$FF,$0F,$6B,$D4,$FF,$31,$FB,$CF,$E4,$B1
	.dc.b	$F6,$BE,$23,$1E,$BF,$E1,$9B,$7F,$83,$F8,$8B,$5F,$80,$6E,$5F,$5F
	.dc.b	$E2,$2D,$7D,$67,$F9,$D7,$A5,$F9,$0F,$88,$B3,$F8,$D7,$FB,$ED,$3E
	.dc.b	$1B,$FB,$FF,$E2,$9F,$E3,$F8,$C5,$32,$FF,$32,$4F,$F9,$0C,$FF,$0F
	.dc.b	$38,$F8,$6E,$3F,$0F,$C9,$F7,$7B,$BF,$A0,$F8,$9B,$87,$E7,$5B,$7F
	.dc.b	$ED,$7E,$1F,$F7,$39,$FC,$F0,$B5,$FA,$8F,$89,$DE,$7F,$95,$7E,$57
	.dc.b	$BF,$FD,$6F,$F3,$DE,$0D,$AC,$E5,$E9,$BF,$F1,$5E,$0F,$80,$CE,$5E
	.dc.b	$16,$CF,$E8,$77,$B7,$FA,$07,$EC,$4F,$F6,$9F,$DA,$7F,$9C,$CF,$53
	.dc.b	$E3,$5B,$0E,$D1,$98,$F9,$53,$C1,$F0,$46,$56,$7C,$1C,$9F,$EF,$B7
	.dc.b	$18,$FA,$B0,$FF,$1A,$FE,$2F,$FE,$59,$C7,$38,$7D,$7F,$84,$4E,$7E
	.dc.b	$0F,$D1,$F1,$1F,$91,$6D,$78,$6F,$FA,$DE,$3F,$1B,$E0,$F5,$39,$6E
	.dc.b	$9F,$EA,$5C,$7F,$03,$5F,$33,$0F,$1F,$27,$E7,$6F,$F3,$DC,$F7,$1E
	.dc.b	$CD,$C7,$F0,$1E,$FF,$F9,$97,$F1,$3F,$A9,$1F,$8F,$D3,$FB,$3B,$F2
	.dc.b	$E7,$5B,$FC,$E9,$C8,$F8,$3E,$9A,$FD,$E7,$C3,$97,$CD,$E2,$7C,$1E
	.dc.b	$32,$FF,$65,$08,$CC,$FB,$CF,$89,$97,$82,$7C,$9A,$FE,$E6,$31,$1F
	.dc.b	$7E,$3F,$79,$87,$8A,$BC,$CE,$1F,$C1,$71,$7D,$4E,$9C,$5F,$CB,$78
	.dc.b	$3F,$07,$0E,$BA,$7B,$9B,$87,$E0,$3E,$37,$8E,$BF,$E8,$9E,$34,$7F
	.dc.b	$0D,$F8,$2F,$C2,$FC,$6D,$D7,$C3,$5F,$E8,$2C,$FE,$8C,$7E,$F7,$8B
	.dc.b	$E8,$C7,$CF,$8C,$71,$7C,$FE,$0F,$81,$18,$9F,$AD,$8E,$1F,$DD,$37
	.dc.b	$3F,$D7,$FC,$2D,$AF,$B0,$DA,$F7,$47,$37,$C0,$31,$F1,$EF,$F0,$67
	.dc.b	$4B,$7E,$55,$FA,$FC,$6C,$E6,$BF,$01,$F0,$7A,$E3,$3E,$7C,$7A,$FC
	.dc.b	$2F,$CE,$CB,$E7,$B6,$FE,$3F,$C6,$79,$F9,$7F,$56,$FF,$39,$9F,$D8
	.dc.b	$F5,$5A,$F9,$AD,$D7,$D5,$C3,$FB,$3E,$27,$C6,$76,$BF,$95,$97,$ED
	.dc.b	$A3,$17,$FA,$DF,$8C,$E5,$C7,$F7,$5F,$E9,$1F,$11,$FC,$03,$E2,$EE
	.dc.b	$3F,$71,$B7,$E0,$3F,$55,$F1,$3F,$8A,$AF,$C5,$B6,$BF,$07,$F1,$7C
	.dc.b	$31,$FC,$53,$FD,$56,$7F,$86,$FC,$5D,$AF,$C4,$B7,$0E,$C7,$F1,$FF
	.dc.b	$8B,$EB,$3F,$59,$B1,$FC,$1F,$BF,$86,$3F,$97,$7F,$B1,$4F,$F1,$AE
	.dc.b	$6B,$7F,$96,$6D,$FA,$F7,$FE,$37,$F1,$7F,$A4,$BF,$DA,$46,$2F,$C0
	.dc.b	$7E,$7B,$E2,$38,$E5,$FA,$57,$FB,$87,$C5,$FF,$47,$F1,$6D,$7F,$31
	.dc.b	$F1,$BF,$90,$BF,$C3,$67,$F6,$3F,$17,$FB,$26,$C7,$E4,$9F,$E6,$3F
	.dc.b	$0B,$CB,$5F,$E4,$8F,$F7,$DA,$FE,$FD,$F1,$76,$FF,$97,$36,$BE,$4F
	.dc.b	$E0,$3F,$B8,$F8,$8F,$CE,$DF,$F0,$C8,$C4,$FC,$EF,$09,$3F,$E0,$9F
	.dc.b	$8C,$E5,$BF,$F5,$C7,$BE,$5F,$D2,$FF,$6A,$F5,$E1,$FC,$AF,$3F,$F6
	.dc.b	$04,$62,$3F,$AB,$61,$DF,$C9,$69,$E2,$C3,$F6,$9A,$78,$BC,$4F,$D9
	.dc.b	$DF,$86,$7F,$A4,$37,$0F,$EF,$59,$FF,$54,$F8,$7C,$6B,$FC,$55,$F8
	.dc.b	$18,$7F,$0E,$C7,$A7,$83,$FC,$EF,$1F,$16,$7F,$BE,$C7,$FA,$EF,$C3
	.dc.b	$72,$D3,$F5,$4F,$FB,$57,$CC,$FE,$C3,$F0,$36,$7F,$8E,$BF,$ED,$DC
	.dc.b	$BA,$ED,$FE,$19,$6B,$F0,$27,$DB,$BF,$C0,$8F,$C7,$7B,$6F,$31,$5F
	.dc.b	$CE,$23,$0D,$F2,$B3,$4E,$2B,$D6,$BD,$CF,$F4,$EF,$8E,$79,$BF,$DE
	.dc.b	$3E,$03,$FD,$DC,$DC,$BF,$C0,$5A,$F9,$8F,$3F,$4F,$E0,$23,$CF,$C7
	.dc.b	$FC,$23,$27,$8D,$E2,$D7,$C3,$0F,$17,$17,$FE,$03,$8D,$BE,$B8,$FF
	.dc.b	$7F,$70,$6B,$66,$EF,$C6,$7F,$CA,$B8,$FB,$A5,$C1,$EC,$8F,$7C,$EB
	.dc.b	$1F,$F1,$0E,$DE,$9E,$2F,$99,$D3,$F8,$89,$F4,$E3,$F3,$BB,$BC,$19
	.dc.b	$FC,$76,$ED,$EF,$9F,$99,$5E,$99,$F3,$47,$8B,$93,$8E,$6F,$5B,$3B
	.dc.b	$7F,$41,$AE,$27,$B2,$75,$87,$CE,$D3,$CB,$D7,$E8,$6D,$7F,$84,$7A
	.dc.b	$FF,$26,$FE,$13,$AF,$CA,$79,$BE,$3E,$E1,$FE,$39,$CB,$FD,$C7,$C5
	.dc.b	$F1,$3D,$8F,$1F,$A1,$E6,$7F,$29,$9F,$6C,$9C,$9F,$7B,$4E,$FC,$9F
	.dc.b	$E2,$2E,$47,$E3,$3F,$E7,$1D,$7F,$C1,$2D,$C3,$DB,$3A,$E9,$F9,$73
	.dc.b	$9D,$7F,$CB,$FB,$FC,$6E,$1F,$F4,$83,$71,$F8,$57,$F5,$32,$FA,$D9
	.dc.b	$7F,$14,$D7,$CD,$D4,$C3,$FD,$31,$F2,$1C,$B1,$FF,$24,$38,$7F,$95
	.dc.b	$7A,$B7,$0F,$F2,$C3,$F0,$32,$ED,$9F,$EF,$EA,$D7,$F8,$FE,$FE,$EF
	.dc.b	$C7,$7A,$67,$FC,$7D,$18,$99,$1F,$C7,$7C,$AB,$1F,$E3,$AF,$15,$1D
	.dc.b	$E3,$D5,$8F,$F1,$B6,$9F,$D6,$30,$EC,$FF,$A7,$7C,$8F,$F2,$56,$1F
	.dc.b	$E7,$39,$CC,$FF,$9D,$3C,$4F,$2D,$FF,$2A,$F0,$AD,$FF,$2F,$7C,$1C
	.dc.b	$3F,$C7,$78,$7E,$4F,$2F,$DC,$DF,$FD,$0B,$9F,$F2,$EC,$3B,$7C,$7F
	.dc.b	$2D,$3F,$CA,$73,$FE,$FB,$E4,$FD,$83,$6B,$8B,$F1,$AB,$E3,$BE,$0B
	.dc.b	$FC,$D6,$9E,$FC,$FC,$4F,$F8,$B6,$79,$63,$FE,$BB,$7F,$DF,$53,$FE
	.dc.b	$40,$73,$07,$35,$E0,$C3,$FA,$67,$7F,$81,$A7,$E2,$AB,$E6,$DF,$E2
	.dc.b	$8F,$0B,$F3,$F1,$FB,$BF,$1D,$FE,$68,$C7,$B9,$DA,$7C,$9F,$1F,$AC
	.dc.b	$3E,$93,$63,$FC,$EF,$D5,$FC,$37,$4F,$76,$9F,$02,$7E,$E6,$3D,$4E
	.dc.b	$D9,$71,$CF,$DD,$1F,$D8,$B6,$E9,$B7,$F1,$DE,$4F,$E2,$0F,$F1,$FC
	.dc.b	$FB,$63,$FD,$C6,$BE,$D3,$F2,$6F,$D5,$F1,$FD,$F6,$F9,$CE,$1F,$F3
	.dc.b	$FF,$81,$FD,$BF,$81,$3F,$C6,$3C,$DF,$B1,$87,$F5,$5E,$3F,$C3,$E1
	.dc.b	$FA,$48,$FA,$55,$FE,$1E,$C1,$D3,$D6,$7F,$C4,$91,$8A,$FB,$DF,$1F
	.dc.b	$FC,$93,$DE,$FD,$5B,$63,$EB,$70,$FD,$43,$EB,$71,$FC,$97,$F0,$53
	.dc.b	$F5,$7E,$3F,$EB,$36,$BF,$15,$1F,$CC,$3A,$7F,$6E,$D8,$EB,$7C,$CF
	.dc.b	$DF,$65,$F8,$47,$F4,$33,$F8,$1F,$8E,$E9,$3F,$B3,$71,$5F,$23,$AB
	.dc.b	$F8,$1C,$7F,$3E,$7E,$3C,$FF,$28,$DF,$E3,$BD,$4F,$CE,$4F,$C7,$F2
	.dc.b	$79,$C7,$C6,$7F,$D6,$FE,$2F,$F6,$28,$FE,$53,$BF,$F4,$7E,$7F,$EB
	.dc.b	$D3,$97,$FA,$EF,$91,$E0,$E8,$FE,$6D,$A7,$F1,$8F,$13,$F9,$AF,$C6
	.dc.b	$7F,$94,$BC,$53,$7E,$33,$0F,$E7,$FE,$3B,$FA,$CF,$87,$FE,$02,$C3
	.dc.b	$F8,$74,$63,$8B,$B7,$17,$B3,$9F,$EB,$72,$FF,$08,$AF,$D4,$7C,$77
	.dc.b	$3C,$3F,$78,$E1,$FE,$31,$97,$F2,$A8,$FE,$56,$F4,$F0,$BF,$4F,$C9
	.dc.b	$F8,$F7,$8D,$7E,$3F,$B9,$7F,$9E,$AF,$E0,$57,$FD,$C3,$F6,$B7,$5F
	.dc.b	$4C,$15,$FE,$EA,$9C,$C7,$E0,$A3,$B1,$FA,$97,$B5,$F5,$C3,$BC,$7E
	.dc.b	$7F,$9B,$FC,$30,$FE,$A3,$CB,$FE,$F7,$E7,$FF,$7E,$7D,$58,$73,$3F
	.dc.b	$00,$7E,$0F,$4E,$B5,$CF,$C8,$FF,$02,$FA,$1F,$E0,$46,$C7,$96,$B8
	.dc.b	$1C,$F8,$7E,$7B,$E8,$6B,$FC,$8F,$4F,$89,$9B,$C9,$FF,$08,$E9,$FE
	.dc.b	$10,$6E,$1D,$7A,$7F,$0F,$87,$99,$FE,$17,$8F,$F2,$FC,$79,$F8,$DF
	.dc.b	$E1,$AC,$FF,$C3,$2D,$C3,$97,$2F,$E0,$F9,$7F,$59,$97,$95,$AF,$96
	.dc.b	$7E,$57,$AB,$FE,$24,$D3,$FC,$46,$D8,$E7,$C1,$F8,$2A,$F8,$3A,$7B
	.dc.b	$B5,$FF,$32,$F1,$7D,$49,$F9,$BC,$0F,$3F,$FC,$5F,$B7,$F8,$B9,$B9
	.dc.b	$74,$AF,$95,$AF,$93,$1F,$7B,$FD,$83,$83,$E4,$F9,$BF,$E3,$8D,$BF
	.dc.b	$C6,$ED,$C3,$96,$BF,$C5,$31,$F7,$10,$F3,$E1,$CE,$7F,$11,$E6,$FF
	.dc.b	$90,$B3,$FF,$20,$B6,$39,$CF,$F1,$2B,$F7,$70,$FF,$8C,$74,$F7,$E9
	.dc.b	$F9,$FF,$C2,$3F,$C0,$7D,$FF,$E6,$1D,$7F,$C9,$F1,$8F,$0B,$FC,$45
	.dc.b	$9F,$4D,$BF,$C8,$CF,$F3,$7C,$7D,$F1,$FD,$B7,$1E,$9C,$7F,$E5,$6D
	.dc.b	$B9,$B6,$F9,$57,$FA,$4C,$FC,$36,$9E,$53,$FC,$8D,$C7,$F4,$F6,$FC
	.dc.b	$9E,$9F,$DE,$B6,$7B,$FF,$E1,$68,$C3,$FE,$18,$D7,$C7,$8F,$5B,$F8
	.dc.b	$39,$FE,$9A,$3F,$53,$7F,$DF,$F3,$FD,$DF,$87,$FE,$9A,$C7,$FC,$2D
	.dc.b	$5F,$E6,$A1,$5B,$7F,$CA,$31,$A9,$FE,$D1,$A7,$A7,$0F,$C7,$65,$E2
	.dc.b	$3E,$7E,$3F,$D7,$71,$79,$FF,$E8,$2C,$7F,$C2,$31,$89,$F1,$9F,$C3
	.dc.b	$B4,$F7,$F9,$FF,$C0,$B5,$FC,$36,$BF,$84,$3F,$0D,$97,$67,$DB,$E7
	.dc.b	$75,$DF,$ED,$B7,$1F,$A3,$E8,$7F,$A3,$F0,$F8,$B7,$FC,$32,$FE,$87
	.dc.b	$2F,$EB,$B4,$FE,$F3,$87,$92,$BF,$D4,$FA,$F9,$DF,$8D,$3A,$7E,$2E
	.dc.b	$DE,$7B,$FC,$91,$FA,$5E,$0F,$8D,$3F,$E5,$3E,$3F,$C9,$78,$1F,$0F
	.dc.b	$6F,$C8,$B7,$3F,$3F,$4F,$F3,$5E,$0F,$BB,$B7,$9B,$1F,$D3,$6A,$F2
	.dc.b	$BE,$67,$9D,$F9,$A6,$E3,$C8,$FC,$C6,$BE,$2E,$1F,$E2,$91,$F5,$7C
	.dc.b	$BF,$A9,$AF,$E8,$9B,$1C,$72,$FC,$BF,$37,$E2,$EF,$F2,$BE,$2F,$DB
	.dc.b	$CB,$F5,$2D,$BE,$13,$F3,$8F,$E5,$BA,$FA,$70,$FB,$47,$E3,$B8,$7F
	.dc.b	$09,$97,$EC,$9B,$7C,$B8,$BE,$56,$7E,$17,$F9,$1E,$0F,$23,$FD,$17
	.dc.b	$F2,$1F,$E8,$C6,$DF,$1C,$BF,$0F,$BF,$C1,$C3,$F8,$64,$7E,$06,$FF
	.dc.b	$D2,$7F,$21,$C2,$3F,$CE,$AF,$DB,$7F,$80,$9F,$08,$FC,$46,$7F,$D7
	.dc.b	$63,$FB,$FF,$8B,$CB,$1F,$C4,$53,$CA,$FF,$07,$47,$AA,$3F,$94,$9E
	.dc.b	$4D,$7D,$8F,$F2,$1D,$5F,$21,$FE,$A6,$F9,$0F,$F5,$3B,$6F,$D9,$5F
	.dc.b	$2B,$9F,$F0,$BA,$7F,$6F,$BF,$F5,$6F,$C8,$7F,$9C,$5B,$7F,$82,$CB
	.dc.b	$F6,$FC,$1F,$0F,$1F,$06,$1F,$DD,$F0,$FF,$3C,$7C,$7F,$F9,$E5,$B8
	.dc.b	$71,$AF,$49,$E1,$BE,$4F,$F5,$53,$F5,$94,$9F,$F0,$4F,$23,$8B,$FC
	.dc.b	$17,$5F,$D3,$FE,$0B,$AA,$9D,$3F,$E1,$0F,$19,$3F,$A3,$C7,$FB,$34
	.dc.b	$FC,$93,$E9,$E1,$FE,$1D,$F8,$FF,$F4,$87,$C7,$FF,$A4,$5B,$1E,$9C
	.dc.b	$7F,$85,$F0,$7F,$17,$C3,$F9,$8E,$DF,$E1,$7C,$3D,$18,$F8,$F4,$F3
	.dc.b	$DF,$AF,$1F,$F5,$75,$6B,$1F,$3D,$FF,$BF,$A7,$CD,$BF,$CE,$3C,$18
	.dc.b	$7E,$42,$BF,$B9,$4F,$18,$F8,$7A,$BE,$13,$18,$7F,$8C,$32,$F0,$3E
	.dc.b	$33,$E2,$47,$67,$E9,$C7,$AF,$1F,$F1,$FC,$7F,$84,$FE,$3F,$9A,$97
	.dc.b	$C7,$E1,$4D,$8E,$99,$FC,$6D,$BE,$ED,$7F,$2F,$8F,$F0,$54,$FF,$15
	.dc.b	$37,$9F,$F8,$5B,$C5,$FE,$5C,$FE,$7E,$7F,$80,$EF,$E7,$CB,$F0,$27
	.dc.b	$F8,$D0,$B8,$31,$67,$A9,$FF,$26,$9A,$8C,$B9,$9E,$13,$B8,$E8,$7F
	.dc.b	$B1,$CE,$0E,$63,$71,$C6,$79,$BA,$3A,$7C,$7B,$E0,$38,$36,$33,$18
	.dc.b	$23,$82,$1E,$C8,$E0,$D8,$DB,$4D,$9F,$8E,$F8,$EF,$D2,$E5,$F8,$CC
	.dc.b	$17,$E1,$8C,$37,$3E,$17,$FA,$E7,$F2,$F1,$98,$FC,$53,$5F,$C1,$37
	.dc.b	$F0,$67,$A9,$E0,$AE,$39,$73,$BE,$2D,$78,$B7,$17,$C2,$F1,$9F,$CC
	.dc.b	$7C,$7D,$7F,$21,$A5,$A3,$67,$19,$F9,$67,$DF,$87,$58,$F3,$65,$A8
	.dc.b	$C1,$30,$FD,$21,$F9,$DC,$7B,$7C,$77,$87,$F6,$CF,$A3,$B7,$ED,$9B
	.dc.b	$AF,$EF,$27,$A5,$F2,$CF,$C2,$DC,$FC,$2F,$7C,$3E,$8E,$57,$0F,$E0
	.dc.b	$B9,$7B,$7C,$EF,$0D,$FC,$5C,$B0,$EC,$F9,$67,$5B,$F9,$9B,$FD,$28
	.dc.b	$D1,$F3,$8D,$1E,$5B,$F9,$5F,$1F,$D2,$3D,$18,$61,$E3,$A7,$AB,$3F
	.dc.b	$3C,$7C,$6C,$3B,$3D,$23,$24,$CF,$D7,$1F,$77,$E3,$FE,$57,$C0,$C8
	.dc.b	$C1,$DB,$6F,$B0,$7C,$67,$A5,$7E,$29,$A7,$F1,$BD,$3F,$05,$F1,$FF
	.dc.b	$83,$C2,$4F,$E2,$E3,$E9,$E3,$E3,$8F,$8E,$FF,$2E,$CB,$D2,$E6,$3E
	.dc.b	$33,$73,$FC,$9C,$7E,$57,$BF,$FA,$7F,$AB,$FC,$F2,$3F,$91,$61,$E2
	.dc.b	$74,$7F,$78,$AF,$91,$9F,$A7,$0E,$D8,$F8,$23,$04,$AF,$2B,$F2,$1F
	.dc.b	$C5,$76,$78,$3B,$BF,$A4,$E7,$67,$F9,$E3,$5F,$E6,$93,$71,$F0,$C7
	.dc.b	$42,$0C,$3C,$4E,$23,$6F,$C2,$C3,$C5,$8F,$F0,$9F,$8F,$ED,$E3,$BD
	.dc.b	$9A,$77,$8F,$07,$0F,$DA,$E5,$FC,$8E,$7D,$B2,$E8,$77,$9F,$DB,$71
	.dc.b	$7E,$EF,$E3,$F9,$4F,$C9,$C3,$D5,$1D,$30,$FC,$31,$F9,$EE,$9F,$BB
	.dc.b	$AF,$C9,$BE,$A7,$EE,$4F,$E1,$9F,$1F,$C9,$FA,$97,$F1,$63,$A4,$7E
	.dc.b	$30,$E3,$9F,$9F,$7F,$86,$4D,$3C,$F3,$FC,$67,$D4,$7C,$17,$D3,$F6
	.dc.b	$BF,$D5,$BE,$9D,$EB,$3F,$56,$B7,$7F,$85,$C3,$E3,$3E,$1C,$BC,$11
	.dc.b	$8E,$4F,$C9,$71,$7F,$21,$BF,$F2,$B5,$FF,$A2,$F2,$FF,$5D,$F8,$7D
	.dc.b	$CE,$DE,$97,$B3,$0F,$CD,$63,$FA,$B6,$F1,$FE,$F3,$2E,$9E,$97,$9F
	.dc.b	$79,$9F,$F9,$43,$44,$F3,$E3,$F9,$DC,$3F,$86,$DF,$CF,$97,$D4,$C3
	.dc.b	$F8,$61,$F1,$7D,$2F,$85,$B4,$8E,$91,$CD,$FC,$14,$FD,$B7,$96,$BF
	.dc.b	$01,$F9,$DE,$87,$CB,$CE,$47,$96,$39,$3F,$80,$CB,$C9,$CD,$F8,$48
	.dc.b	$EA,$79,$0F,$E7,$BA,$BD,$0F,$AD,$9C,$9F,$81,$1D,$2F,$F0,$19,$F9
	.dc.b	$67,$F0,$D1,$87,$A6,$7F,$E3,$8C,$1E,$87,$E0,$73,$95,$F2,$E3,$A7
	.dc.b	$17,$DD,$E0,$FF,$2C,$F0,$7C,$AA,$EB,$3F,$8E,$7C,$FE,$87,$E2,$F5
	.dc.b	$91,$F5,$63,$9E,$BF,$AE,$DF,$E0,$4F,$C3,$BE,$5A,$F9,$7D,$0F,$CA
	.dc.b	$69,$2B,$EF,$47,$27,$F2,$57,$CB,$7F,$75,$FF,$82,$79,$3F,$25,$E8
	.dc.b	$7E,$77,$49,$87,$E1,$E3,$95,$7F,$1E,$C3,$96,$7E,$FA,$FE,$70,$7F
	.dc.b	$80,$7C,$5F,$E5,$5E,$27,$F9,$9B,$A5,$D6,$DB,$F5,$47,$F1,$DA,$DF
	.dc.b	$87,$F9,$2C,$7A,$63,$FC,$6F,$2E,$FC,$3F,$9D,$8E,$DE,$7F,$EC,$F6
	.dc.b	$91,$FE,$AC,$C7,$FC,$AE,$FE,$CE,$BF,$23,$CB,$F5,$30,$E7,$3F,$80
	.dc.b	$E1,$FF,$09,$5F,$F9,$82,$3F,$65,$F8,$E7,$91,$6E,$8C,$42,$78,$2E
	.dc.b	$0F,$E9,$5C,$7F,$3B,$2F,$E3,$5C,$DF,$83,$E3,$FD,$BD,$FE,$3A,$7F
	.dc.b	$D3,$F8,$7F,$97,$63,$C3,$BF,$F1,$1C,$BD,$71,$D7,$87,$7E,$6F,$D4
	.dc.b	$C3,$AE,$FE,$CC,$BF,$5D,$7F,$E2,$0C,$BF,$C8,$AF,$B3,$D0,$E8,$7D
	.dc.b	$1E,$3E,$51,$FD,$2B,$5F,$36,$3F,$4F,$8B,$FC,$75,$C5,$F1,$BC,$EF
	.dc.b	$26,$B3,$0F,$A1,$1C,$9F,$E5,$FA,$F8,$F0,$FE,$1B,$7F,$3B,$CE,$F6
	.dc.b	$67,$2F,$EC,$47,$1E,$2F,$DF,$C7,$D8,$3F,$AD,$57,$D6,$F3,$7E,$26
	.dc.b	$52,$7F,$01,$1C,$70,$FE,$49,$8F,$A6,$7E,$F1,$F9,$5E,$8F,$99,$94
	.dc.b	$8F,$C4,$47,$1D,$7F,$82,$70,$FD,$63,$E5,$9F,$EB,$AE,$17,$9B,$F5
	.dc.b	$73,$98,$7E,$46,$39,$3F,$E2,$BD,$FE,$13,$EA,$3F,$CC,$9F,$21,$FD
	.dc.b	$F7,$D2,$FB,$BD,$1F,$E1,$16,$F3,$7F,$B3,$27,$9E,$7F,$E1,$DC,$3F
	.dc.b	$CC,$DC,$5D,$31,$FE,$C8,$7F,$23,$C7,$F2,$37,$FE,$07,$3F,$C1,$13
	.dc.b	$CB,$83,$F4,$55,$F7,$0F,$9C,$72,$C3,$FD,$19,$F2,$1F,$E1,$1F,$D0
	.dc.b	$9E,$F5,$3F,$C2,$51,$CE,$87,$ED,$CF,$F4,$03,$E6,$7B,$1F,$E6,$97
	.dc.b	$F5,$FE,$97,$E7,$78,$E4,$FE,$E6,$3A,$3C,$CF,$F6,$E7,$6B,$F1,$C7
	.dc.b	$BB,$FE,$8B,$8C,$36,$33,$3F,$E9,$59,$E1,$F0,$5F,$E8,$DF,$92,$FD
	.dc.b	$67,$2C,$AF,$F1,$E4,$78,$3B,$BF,$C0,$CF,$5F,$01,$9D,$39,$57,$F9
	.dc.b	$D3,$AB,$F6,$DE,$B4,$8F,$87,$CD,$FE,$4B,$C7,$FD,$7A,$FF,$04,$ED
	.dc.b	$FE,$3B,$DF,$F9,$EA,$FF,$38,$61,$D1,$FF,$47,$67,$FE,$A5,$C3,$FD
	.dc.b	$27,$1F,$39,$7C,$97,$F0,$5F,$5E,$4F,$83,$B3,$FA,$8C,$FA,$FB,$39
	.dc.b	$3C,$F1,$95,$FE,$8C,$8E,$DF,$25,$FC,$3F,$D8,$93,$CC,$F0,$47,$37
	.dc.b	$FD,$26,$FC,$5F,$5A,$46,$0F,$F4,$BF,$6F,$F9,$67,$E2,$9F,$05,$4F
	.dc.b	$F2,$E6,$C9,$E9,$5F,$E9,$17,$C6,$F3,$30,$7F,$A8,$BE,$47,$8C,$48
	.dc.b	$F3,$1F,$E6,$18,$E7,$E0,$7F,$A0,$E7,$E5,$12,$7F,$D5,$7F,$21,$E2
	.dc.b	$F9,$1F,$D9,$79,$72,$79,$E2,$9E,$77,$FD,$F2,$FC,$15,$E4,$3F,$D5
	.dc.b	$0F,$B0,$E8,$F2,$9C,$3F,$D4,$CF,$F0,$1F,$07,$F0,$EA,$E9,$1E,$07
	.dc.b	$CC,$F4,$3C,$7F,$23,$FE,$83,$FD,$B7,$FC,$9F,$1F,$D0,$CF,$D8,$B8
	.dc.b	$E9,$FC,$84,$FA,$27,$F1,$0F,$89,$E0,$7F,$A1,$F1,$FA,$71,$F5,$3E
	.dc.b	$43,$B7,$B3,$2B,$E1,$C7,$37,$FD,$1C,$FF,$05,$F8,$49,$1A,$3E,$4F
	.dc.b	$9B,$FC,$27,$E2,$64,$60,$F8,$5C,$3F,$CB,$75,$FE,$27,$1F,$ED,$1A
	.dc.b	$F1,$3C,$CF,$3F,$07,$F6,$F7,$EA,$FC,$97,$F6,$5F,$81,$91,$83,$AD
	.dc.b	$FF,$41,$C3,$F9,$44,$7F,$BA,$AB,$F2,$38,$48,$E0,$7C,$45,$E0,$FF
	.dc.b	$71,$F8,$19,$3F,$E0,$D8,$F9,$9E,$3B,$BD,$F1,$5F,$E8,$76,$D7,$F9
	.dc.b	$ED,$AF,$D7,$9E,$1F,$15,$FE,$21,$F0,$BF,$BF,$7A,$32,$BF,$C3,$51
	.dc.b	$DA,$FF,$D1,$F1,$CF,$E0,$FB,$1D,$FE,$0F,$FC,$55,$E1,$FF,$81,$FD
	.dc.b	$39,$5F,$E2,$38,$E5,$D1,$E5,$24,$79,$7E,$0B,$FC,$65,$E2,$7F,$83
	.dc.b	$FD,$29,$3F,$E2,$B8,$F1,$7C,$07,$B7,$E0,$24,$FB,$63,$FC,$75,$D3
	.dc.b	$FE,$17,$F2,$E4,$FF,$8C,$E3,$D5,$EF,$FC,$52,$47,$C5,$F7,$FF,$81
	.dc.b	$FC,$8F,$FA,$BF,$F1,$FF,$E4,$FC,$7F,$D3,$39,$D1,$8B,$FE,$45,$5F
	.dc.b	$C2,$BC,$FF,$EC,$5E,$07,$FB,$4C,$47,$3F,$3E,$57,$EF,$23,$F8,$8F
	.dc.b	$C5,$FF,$BB,$7E,$50,$D3,$33,$FC,$0B,$0E,$A7,$F3,$33,$CE,$F3,$34
	.dc.b	$43,$F2,$F1,$C9,$FC,$CC,$79,$DF,$ED,$F8,$79,$E3,$9C,$7E,$7A,$F4
	.dc.b	$FE,$83,$E4,$FF,$B9,$7A,$12,$3F,$8A,$C7,$F2,$A3,$C8,$FF,$0B,$3F
	.dc.b	$A3,$3F,$D2,$4F,$99,$E9,$48,$E2,$7F,$1E,$F9,$3F,$A1,$E3,$C8,$C1
	.dc.b	$0F,$EE,$3B,$3D,$2F,$EE,$0D,$8F,$F0,$0F,$91,$FC,$ED,$FE,$AD,$7F
	.dc.b	$D6,$4F,$B9,$E8,$C8,$E9,$E7,$FF,$65,$F6,$FF,$01,$E8,$49,$FE,$63
	.dc.b	$1F,$D1,$8E,$4F,$6C,$FF,$D5,$B8,$7F,$65,$99,$3F,$EE,$9C,$9E,$87
	.dc.b	$E2,$CE,$51,$FC,$F3,$DA,$FC,$77,$A3,$23,$04,$1A,$FA,$D7,$27,$F8
	.dc.b	$31,$CE,$7F,$EB,$EC,$BF,$BB,$61,$27,$C5,$E7,$7E,$6C,$FE,$BF,$1F
	.dc.b	$D8,$3D,$FF,$CF,$7A,$12,$BF,$A7,$C6,$5F,$F0,$C6,$BF,$9A,$6F,$37
	.dc.b	$F8,$0E,$64,$7B,$FA,$3F,$50,$71,$7F,$93,$FC,$57,$E7,$74,$91,$FD
	.dc.b	$72,$3F,$B8,$9C,$9F,$59,$FE,$0D,$AF,$F0,$84,$48,$F9,$9D,$3F,$B5
	.dc.b	$38,$BF,$D9,$BE,$46,$BE,$03,$07,$73,$EC,$1F,$E1,$99,$E8,$F8,$07
	.dc.b	$E1,$8F,$57,$9F,$3F,$F8,$A6,$AF,$81,$FE,$4E,$BD,$96,$7F,$99,$3F
	.dc.b	$E5,$2B,$D3,$FC,$BF,$67,$67,$F0,$D3,$85,$F3,$AD,$DF,$7B,$F3,$9F
	.dc.b	$80,$C3,$CF,$CB,$F8,$3A,$F3,$4F,$F8,$41,$E3,$7F,$F0,$02,$BC,$F3
	.dc.b	$E3,$3B,$9E,$5F,$8F,$FE,$3D,$D3,$23,$94,$69,$F8,$7A,$7C,$CC,$7F
	.dc.b	$25,$CE,$C3,$E1,$47,$8E,$78,$64,$C3,$F9,$57,$3C,$C7,$A9,$E7,$F8
	.dc.b	$2F,$E4,$1C,$32,$30,$7A,$F0,$F0,$71,$FA,$E3,$E1,$E3,$F0,$1F,$F0
	.dc.b	$4B,$77,$FA,$31,$F6,$B2,$F1,$9E,$4B,$91,$D8,$E3,$8E,$1E,$71,$E0
	.dc.b	$7F,$0C,$78,$1D,$57,$6B,$F7,$4F,$B0,$E2,$7D,$3F,$92,$FF,$13,$7C
	.dc.b	$6C,$8C,$10,$D9,$2B,$A4,$FC,$A3,$FA,$97,$4C,$9E,$13,$E1,$F6,$7F
	.dc.b	$C6,$5E,$54,$9F,$C1,$47,$D2,$3F,$D5,$1E,$BC,$8C,$90,$FC,$A7,$4F
	.dc.b	$F5,$F3,$83,$F8,$BF,$13,$FD,$61,$F2,$52,$30,$43,$E8,$DF,$07,$E9
	.dc.b	$CF,$E0,$8F,$ED,$5C,$B2,$38,$65,$F2,$71,$FE,$DD,$D7,$27,$F3,$51
	.dc.b	$F8,$A3,$93,$C3,$A7,$F4,$D1,$98,$FF,$42,$79,$0F,$C8,$3F,$43,$08
	.dc.b	$C3,$B3,$FC,$59,$ED,$FE,$8A,$38,$73,$5A,$FB,$75,$A3,$AF,$57,$F9
	.dc.b	$C5,$BE,$2F,$18,$37,$C6,$C7,$8D,$9F,$1A,$5F,$28,$FD,$2F,$B5,$DF
	.dc.b	$B2,$4F,$EE,$63,$2F,$B3,$2F,$CE,$6F,$FE,$11,$6C,$7F,$0D,$9D,$BF
	.dc.b	$E3,$B8,$74,$FA,$CC,$F1,$7E,$5F,$9B,$DD,$CD,$2F,$F8,$44,$7E,$E0
	.dc.b	$F8,$5C,$D2,$32,$43,$B7,$27,$C7,$38,$3F,$C4,$7D,$8F,$95,$9C,$8C
	.dc.b	$10,$F3,$5F,$CF,$3F,$83,$CC,$8F,$80,$FD,$33,$0F,$EF,$BD,$8F,$AC
	.dc.b	$47,$F9,$03,$FC,$40,$FB,$5F,$21,$23,$24,$3E,$37,$AF,$F7,$CE,$0F
	.dc.b	$F2,$98,$FC,$17,$AF,$23,$04,$3E,$77,$7F,$E1,$CF,$E3,$F3,$23,$EC
	.dc.b	$77,$7E,$2C,$C3,$FC,$5A,$FF,$1D,$EA,$C8,$FE,$71,$1F,$CA,$0F,$C9
	.dc.b	$7C,$7C,$8C,$90,$FB,$DE,$A7,$E5,$CE,$0F,$F4,$1C,$BF,$35,$EA,$48
	.dc.b	$C1,$0F,$C3,$7C,$7F,$E7,$CF,$E6,$F3,$23,$F5,$7B,$FF,$2A,$E4,$FE
	.dc.b	$B7,$DE,$8C,$74,$7F,$57,$BF,$E7,$FC,$1F,$AB,$97,$C7,$D3,$38,$FF
	.dc.b	$55,$9D,$7C,$14,$8C,$10,$FF,$4A,$53,$E3,$FA,$1C,$A3,$F9,$FF,$17
	.dc.b	$7F,$86,$91,$FD,$86,$3F,$A8,$9E,$3F,$8F,$91,$92,$1D,$BE,$3F,$D0
	.dc.b	$70,$7F,$B5,$72,$FA,$FE,$3E,$46,$08,$79,$BE,$3B,$E0,$1F,$D8,$26
	.dc.b	$47,$CA,$F8,$DF,$88,$61,$FE,$AF,$CF,$F1,$FE,$32,$47,$F7,$98,$FE
	.dc.b	$D2,$7C,$BF,$88,$91,$92,$1F,$1B,$E1,$FE,$81,$C1,$FF,$01,$75,$7D
	.dc.b	$3F,$86,$91,$82,$1F,$3B,$E1,$7E,$C1,$FD,$E2,$64,$7E,$0B,$E1,$3E
	.dc.b	$E1,$87,$FB,$7F,$6F,$DF,$F3,$E4,$7F,$84,$63,$FC,$02,$7E,$0F,$E2
	.dc.b	$24,$64,$87,$DE,$F8,$8F,$C4,$1C,$11,$FE,$E1,$8F,$C6,$7C,$44,$8C
	.dc.b	$F8,$1F,$DF,$BE,$0D,$5F,$CE,$FD,$37,$3D,$25,$FF,$46,$87,$3E,$F9
	.dc.b	$E4,$E5,$E5,$CD,$FE,$1A,$F0,$FF,$39,$F1,$12,$BF,$C5,$D1,$97,$FA
	.dc.b	$8F,$E4,$3F,$EC,$3F,$14,$FF,$B1,$63,$FC,$97,$3B,$7F,$DE,$10,$F8
	.dc.b	$9F,$EA,$4D,$8E,$A7,$BB,$C6,$FF,$27,$7B,$BF,$B0,$8F,$EB,$92,$F8
	.dc.b	$8E,$27,$F8,$DE,$75,$27,$86,$3E,$47,$B9,$BB,$93,$EA,$F8,$8F,$19
	.dc.b	$85,$C7,$FE,$2C,$AF,$DE,$C7,$C0,$F5,$FD,$27,$87,$C9,$91,$F1,$3D
	.dc.b	$8F,$69,$87,$9E,$3E,$C9,$F8,$3E,$BC,$9F,$80,$7E,$0B,$0E,$EF,$03
	.dc.b	$1E,$57,$C8,$F6,$24,$70,$3E,$04,$7C,$33,$E6,$F4,$C8,$D5,$27,$51
	.dc.b	$F6,$BE,$23,$E9,$98,$7C,$07,$CB,$F2,$FE,$BF,$C4,$48,$E7,$1F,$44
	.dc.b	$FB,$7F,$11,$23,$24,$3F,$03,$F0,$7F,$80,$38,$3E,$4F,$3B,$F0,$7F
	.dc.b	$0F,$23,$04,$3F,$13,$F0,$FF,$88,$3F,$01,$32,$3F,$27,$F0,$DF,$8C
	.dc.b	$30,$FA,$B5,$7A,$1F,$90,$F8,$49,$3F,$12,$3F,$14,$7E,$53,$E1,$24
	.dc.b	$64,$87,$E6,$BE,$13,$F3,$27,$07,$E6,$F8,$BF,$9C,$F8,$49,$18,$21
	.dc.b	$F6,$FB,$5F,$2D,$FE,$38,$6C,$7E,$6E,$A3,$FE,$3E,$9F,$F1,$AF,$6F
	.dc.b	$F5,$46,$F5,$FE,$9A,$BF,$45,$E3,$7F,$8F,$65,$F0,$94,$CE,$1F,$82
	.dc.b	$9D,$7C,$3C,$8C,$10,$F3,$57,$F8,$B1,$F1,$7C,$27,$27,$E6,$79,$3D
	.dc.b	$FE,$12,$47,$E3,$A3,$F6,$87,$8F,$E1,$24,$64,$87,$6F,$84,$F4,$1C
	.dc.b	$1F,$CB,$F9,$9E,$BF,$84,$91,$82,$1E,$6F,$7B,$E0,$1F,$C0,$E6,$47
	.dc.b	$C9,$F8,$4F,$88,$61,$FC,$27,$BB,$F1,$FE,$12,$47,$E9,$A3,$F8,$61
	.dc.b	$F2,$FE,$12,$46,$48,$7C,$6F,$84,$FA,$07,$07,$F5,$FE,$FF,$D3,$F8
	.dc.b	$49,$18,$21,$F3,$BE,$13,$EC,$1F,$C6,$E6,$47,$E0,$7E,$13,$EE,$18
	.dc.b	$7F,$39,$F0,$5F,$7F,$E1,$24,$7E,$FA,$3F,$92,$1F,$83,$F8,$49,$19
	.dc.b	$21,$F7,$BE,$0B,$F1,$07,$07,$F8,$3F,$C2,$7E,$2F,$E0,$64,$60,$87
	.dc.b	$E1,$BD,$CF,$C8,$1F,$CC,$E6,$47,$FA,$5B,$E4,$BF,$99,$9F,$CD,$BC
	.dc.b	$94,$FF,$39,$F8,$5F,$F4,$A4,$FF,$20,$F1,$3F,$C6,$2D,$F1,$3F,$A1
	.dc.b	$CE,$F0,$E9,$EA,$70,$30,$FE,$FE,$39,$FA,$52,$3F,$90,$C7,$F4,$7F
	.dc.b	$85,$EF,$E7,$C9,$FE,$4B,$19,$7D,$1E,$6F,$90,$C3,$FC,$2E,$3C,$FE
	.dc.b	$4C,$8E,$27,$F2,$EF,$85,$F5,$F9,$32,$30,$43,$F8,$16,$AF,$17,$E0
	.dc.b	$9F,$D6,$EA,$4F,$C9,$F0,$7E,$29,$87,$F8,$E7,$C1,$FC,$8F,$02,$47
	.dc.b	$F3,$E8,$FE,$C8,$7C,$CE,$C9,$19,$21,$E2,$EA,$FA,$27,$07,$FA,$3F
	.dc.b	$C2,$FD,$4E,$69,$18,$21,$EA,$E2,$FB,$27,$F7,$39,$91,$F8,$1E,$0F
	.dc.b	$BA,$61,$F8,$3A,$7E,$03,$29,$1F,$87,$8F,$EF,$9F,$0B,$F8,$4B,$93
	.dc.b	$FD,$6A,$32,$FE,$35,$FC,$49,$87,$E6,$FC,$27,$FB,$7F,$DB,$FF,$0B
	.dc.b	$1F,$E1,$7E,$14,$7F,$82,$4D,$7C,$24,$AF,$EC,$F1,$97,$FD,$DF,$5F
	.dc.b	$E1,$16,$FC,$1E,$5F,$E7,$91,$FE,$11,$77,$F0,$7E,$13,$C5,$DB,$FE
	.dc.b	$3E,$8F,$ED,$BF,$85,$F9,$3E,$13,$CC,$67,$BD,$DF,$B9,$FF,$16,$1E
	.dc.b	$AF,$84,$F6,$1B,$8D,$7C,$77,$BF,$E1,$3E,$07,$C2,$4C,$BD,$31,$B7
	.dc.b	$FC,$9F,$87,$F8,$C5,$F4,$EB,$F1,$CD,$28,$D6,$5E,$48,$E3,$C1,$F3
	.dc.b	$7E,$12,$47,$F8,$4E,$B7,$9F,$F9,$73,$DA,$62,$F8,$4F,$A9,$F0,$93
	.dc.b	$0F,$9B,$1E,$0C,$3E,$74,$7C,$CF,$84,$FB,$46,$79,$24,$7D,$4F,$84
	.dc.b	$FB,$C6,$1F,$A7,$F0,$9F,$81,$F8,$49,$1F,$6A,$3B,$73,$7E,$17,$E1
	.dc.b	$24,$FA,$63,$F1,$1F,$09,$23,$F0,$2F,$07,$F0,$7F,$09,$F8,$DF,$84
	.dc.b	$91,$83,$E1,$75,$FC,$FE,$27,$E4,$5F,$DF,$CF,$C1,$4F,$E0,$A7,$51
	.dc.b	$FD,$F2,$13,$C3,$E2,$32,$6B,$E1,$B7,$F0,$DC,$3E,$16,$5F,$F9,$2A
	.dc.b	$FB,$CF,$CB,$F0,$3A,$FC,$2C,$AF,$AD,$1E,$0F,$85,$91,$C7,$34,$79
	.dc.b	$3E,$17,$CB,$F0,$BE,$6F,$85,$97,$B8,$FB,$1E,$2F,$AB,$E1,$64,$FE
	.dc.b	$0E,$3D,$BF,$0B,$23,$B3,$F0,$3E,$17,$E0,$FC,$2F,$C2,$F8,$59,$5E
	.dc.b	$38,$FC,$27,$91,$F1,$BE,$16,$4F,$E3,$E3,$E4,$FC,$2C,$8F,$53,$F3
	.dc.b	$3E,$17,$E6,$FC,$2F,$CE,$F8,$59,$5E,$F8,$FC,$87,$9D,$F4,$BE,$16
	.dc.b	$4F,$E6,$E3,$EA,$FC,$2C,$8F,$8C,$FD,$DF,$C8,$FF,$D1,$E7,$F5,$78
	.dc.b	$FB,$2F,$F9,$BD,$FA,$8D,$7F,$C8,$4E,$DF,$D7,$79,$BC,$FE,$16,$47
	.dc.b	$EA,$E3,$B7,$C2,$C8,$E6,$F8,$7E,$17,$C5,$F0,$BE,$3F,$85,$95,$E0
	.dc.b	$8F,$D6,$7A,$9E,$7F,$85,$93,$FB,$78,$F5,$7C,$2C,$8F,$3B,$ED,$F8
	.dc.b	$5F,$77,$C2,$FB,$FE,$16,$57,$B2,$3F,$3D,$F8,$EB,$8E,$92,$7F,$29
	.dc.b	$0E,$2F,$91,$F0,$DF,$27,$E1,$A5,$7C,$68,$F9,$9F,$0D,$F3,$7E,$16
	.dc.b	$4F,$CA,$8F,$E0,$DE,$E7,$D1,$F8,$69,$3F,$C4,$63,$EA,$7C,$34,$8F
	.dc.b	$A2,$FD,$7F,$86,$FB,$1F,$0B,$FB,$5F,$C9,$3F,$DB,$47,$FB,$6F,$B3
	.dc.b	$FA,$2D,$FF,$1B,$F8,$2D,$7C,$24,$8C,$1F,$AF,$AF,$E2,$FF,$0B,$FC
	.dc.b	$97,$D5,$46,$3E,$17,$F9,$13,$6B,$F7,$AF,$F4,$77,$FD,$08,$BD,$AF
	.dc.b	$27,$AB,$27,$B4,$7F,$25,$F8,$4F,$47,$A1,$27,$F9,$84,$7A,$FC,$B9
	.dc.b	$1E,$57,$DD,$E3,$FB,$FC,$6F,$81,$E0,$CA,$F5,$47,$F3,$1F,$83,$F8
	.dc.b	$9C,$D2,$7F,$9E,$47,$C7,$E2,$91,$F0,$5F,$95,$97,$CB,$9F,$99,$F0
	.dc.b	$52,$BE,$34,$7F,$3D,$8F,$A1,$F0,$72,$7F,$A4,$47,$D3,$F8,$39,$1F
	.dc.b	$35,$FA,$DF,$05,$F5,$FE,$0B,$EC,$7C,$14,$AF,$A5,$1F,$D2,$70,$FE
	.dc.b	$61,$E4,$7F,$AB,$4F,$E5,$FD,$1F,$D6,$6B,$FA,$E4,$EB,$E0,$B7,$F0
	.dc.b	$52,$7F,$95,$5F,$F3,$3E,$A2,$F8,$1E,$BF,$05,$2B,$71,$FD,$67,$6F
	.dc.b	$0F,$C1,$49,$FE,$C9,$1E,$4F,$81,$F2,$FC,$0C,$9E,$F1,$E8,$F8,$0F
	.dc.b	$4F,$C0,$C9,$F1,$C7,$B3,$E0,$7D,$BF,$03,$27,$CF,$1F,$03,$E0,$7E
	.dc.b	$0F,$C0,$49,$F5,$C7,$C4,$F8,$1F,$8B,$F0,$32,$7D,$F1,$F2,$3E,$07
	.dc.b	$E4,$FC,$0C,$9F,$87,$1F,$33,$E0,$3E,$6F,$C0,$C9,$F8,$F1,$F4,$3D
	.dc.b	$8F,$A3,$EB,$C9,$F9,$71,$FD,$8F,$F2,$5F,$A5,$DB,$FA,$77,$F9,$47
	.dc.b	$51,$A6,$BC,$EF,$F0,$53,$FC,$A6,$78,$7C,$17,$33,$A7,$C4,$F5,$F8
	.dc.b	$3E,$DF,$07,$33,$FE,$C9,$82,$7C,$5F,$05,$E3,$F8,$39,$5A,$8F,$37
	.dc.b	$C1,$C8,$E2,$FA,$7C,$2F,$57,$C1,$C9,$EA,$FB,$7E,$0B,$DD,$F0,$7E
	.dc.b	$FF,$81,$95,$E2,$8F,$85,$F0,$52,$3C,$AF,$C5,$F8,$1F,$8D,$F0,$5F
	.dc.b	$1F,$E0,$A5,$7A,$A3,$E5,$7C,$17,$CB,$F8,$29,$3E,$E8,$F9,$DF,$05
	.dc.b	$FE,$11,$FC,$99,$F1,$34,$CD,$7F,$81,$E3,$FC,$6E,$FF,$84,$1B,$D6
	.dc.b	$69,$FC,$65,$C7,$C3,$F5,$F8,$19,$5C,$E3,$C1,$F0,$5E,$1F,$82,$93
	.dc.b	$DE,$3C,$9F,$05,$E5,$F8,$29,$3E,$38,$F4,$7C,$0F,$A7,$E0,$A4,$F9
	.dc.b	$E3,$D9,$F0,$5E,$DF,$82,$93,$EB,$8F,$81,$F0,$52,$3D,$CF,$C3,$F8
	.dc.b	$2F,$89,$F0,$52,$7E,$13,$F1,$FD,$FF,$91,$ED,$7C,$9F,$82,$95,$F1
	.dc.b	$E3,$93,$F2,$6E,$AF,$2B,$FA,$17,$2B,$E0,$7B,$4E,$7C,$CF,$F3,$24
	.dc.b	$B3,$CD,$F0,$CF,$C3,$7F,$57,$3F,$0B,$C4,$FA,$D9,$FD,$DF,$82,$91
	.dc.b	$F6,$DF,$F2,$87,$E4,$FF,$C8,$33,$FF,$39,$3F,$C6,$C9,$F0,$7A,$F8
	.dc.b	$3D,$FC,$14,$C3,$FC,$95,$5F,$C7,$DB,$F0,$7D,$3E,$0F,$AF,$C1,$49
	.dc.b	$DC,$78,$3E,$0F,$C3,$F0,$52,$79,$C7,$93,$E0,$FC,$BF,$05,$27,$BC
	.dc.b	$7A,$3E,$0E,$47,$89,$F5,$FC,$17,$B3,$E0,$E4,$F9,$9F,$7F,$C1,$7C
	.dc.b	$0F,$83,$93,$EA,$7E,$1F,$C1,$72,$8F,$6D,$D7,$E6,$79,$2E,$17,$C1
	.dc.b	$F6,$9C,$C3,$2C,$75,$FF,$7F,$86,$7E,$0A,$DE,$76,$B5,$9F,$D4,$F8
	.dc.b	$29,$5F,$39,$FA,$FF,$07,$F6,$3E,$0B,$EC,$FC,$1C,$AF,$A7,$1F,$E8
	.dc.b	$BE,$2F,$E4,$91,$FE,$86,$6E,$E6,$9F,$CB,$35,$61,$DA,$BE,$EB,$CF
	.dc.b	$E0,$BC,$11,$F8,$0C,$7C,$0F,$C0,$C7,$C1,$D9,$23,$FD,$1D,$8F,$6F
	.dc.b	$8D,$F2,$7C,$24,$9C,$BE,$7F,$82,$EB,$1F,$2B,$1E,$BC,$1F,$27,$2E
	.dc.b	$93,$F4,$B1,$F4,$E5,$EF,$CB,$E0,$7C,$27,$C1,$F8,$29,$7E,$28,$F8
	.dc.b	$9F,$07,$23,$CA,$FC,$7F,$83,$F9,$1F,$07,$27,$D2,$FC,$BF,$83,$EB
	.dc.b	$1C,$C8,$78,$FE,$0F,$A3,$E5,$BB,$B7,$F8,$FF,$56,$3D,$6B,$D5,$8D
	.dc.b	$AD,$63,$F6,$7E,$0B,$ED,$7C,$1C,$BF,$97,$1F,$77,$E0,$FE,$F7,$C1
	.dc.b	$C9,$F9,$F1,$F8,$1F,$82,$91,$F4,$9F,$C2,$7C,$1F,$58,$E6,$43,$E6
	.dc.b	$7C,$17,$57,$EC,$DF,$57,$E8,$5F,$47,$E9,$5D,$CF,$DD,$87,$E4,$7E
	.dc.b	$0F,$F2,$5F,$07,$2F,$F0,$71,$FE,$DF,$FC,$A3,$FA,$7C,$7F,$B4,$DA
	.dc.b	$FE,$B9,$B7,$FE,$67,$7F,$A7,$9F,$D3,$49,$F1,$7C,$FE,$0F,$C1,$5D
	.dc.b	$88,$73,$F8,$4F,$07,$C2,$C8,$F0,$3E,$3F,$88,$F2,$5C,$9F,$23,$D5
	.dc.b	$F3,$7C,$2F,$A4,$8F,$A7,$E1,$BA,$3E,$9B,$B9,$FA,$70,$E9,$3E,$DC
	.dc.b	$2C,$FB,$72,$F8,$9F,$0B,$F1,$7E,$1A,$5F,$C5,$8F,$91,$F0,$B2,$3E
	.dc.b	$43,$F2,$FE,$1B,$E6,$7C,$2C,$9F,$98,$FC,$FF,$86,$FA,$1F,$0D,$DA
	.dc.b	$7A,$1F,$45,$E8,$FD,$1B,$B9,$FD,$1C,$3A,$4F,$D5,$C2,$CF,$D5,$C3
	.dc.b	$EE,$7C,$37,$DD,$F8,$49,$7F,$76,$3F,$01,$F0,$9F,$81,$F8,$39,$3F
	.dc.b	$81,$8F,$C2,$7C,$24,$8F,$C2,$3F,$87,$F8,$6E,$B1,$CC,$87,$E1,$FE
	.dc.b	$17,$AB,$F8,$BB,$EA,$FE,$2E,$FA,$3F,$8D,$BB,$9F,$E3,$70,$4C,$8F
	.dc.b	$F5,$B7,$E5,$3F,$D9,$61,$EA,$A3,$0D,$8E,$27,$F4,$68,$FF,$42,$0F
	.dc.b	$85,$CD,$74,$6F,$9F,$E0,$BE,$91,$8B,$EF,$7F,$DB,$5E,$9F,$09,$9A
	.dc.b	$F1,$36,$4D,$F8,$3C,$7C,$34,$BF,$66,$3D,$EF,$CB,$F0,$BC,$72,$95
	.dc.b	$A3,$C2,$F5,$F8,$6E,$D9,$49,$D6,$7E,$26,$E0,$CF,$C9,$8F,$86,$7E
	.dc.b	$26,$78,$77,$97,$81,$F9,$0F,$77,$9E,$3D,$DF,$82,$76,$BF,$81,$F0
	.dc.b	$B2,$3C,$2F,$C3,$F8,$6F,$89,$F0,$D2,$7C,$AF,$C7,$F4,$BE,$47,$C3
	.dc.b	$49,$F4,$BF,$2F,$E1,$B9,$47,$B2,$EB,$F7,$7E,$17,$A3,$F7,$F0,$BC
	.dc.b	$7F,$88,$C7,$A4,$FE,$0F,$1B,$3F,$8C,$C7,$EC,$7C,$2F,$D9,$F8,$59
	.dc.b	$7F,$2A,$3E,$E7,$C2,$FD,$DF,$86,$93,$F3,$A3,$F0,$1F,$0D,$F8,$1F
	.dc.b	$85,$93,$F4,$A3,$AB,$CC,$87,$E5,$7E,$07,$AB,$F6,$2F,$AB,$F9,$9B
	.dc.b	$EA,$FD,$CB,$EA,$FE,$76,$FF,$21,$F0,$52,$3F,$02,$FE,$57,$F2,$AF
	.dc.b	$F0,$79,$FE,$B8,$3F,$D6,$9D,$9F,$9F,$6C,$FE,$B1,$D7,$C2,$F1,$24
	.dc.b	$7E,$3A,$3F,$D6,$8D,$9F,$03,$FA,$FA,$F0,$47,$87,$E1,$3C,$5F,$0D
	.dc.b	$2B,$94,$74,$7F,$C7,$76,$C3,$1C,$2C,$3C,$DF,$0D,$EB,$F8,$5E,$B8
	.dc.b	$78,$4F,$23,$DD,$F3,$65,$DD,$EB,$97,$4B,$F8,$18,$D9,$F8,$18,$7C
	.dc.b	$7F,$86,$F9,$1F,$0B,$27,$E0,$47,$CB,$F8,$5F,$99,$F0,$B2,$7E,$24
	.dc.b	$7C,$FF,$85,$FA,$1F,$0D,$27,$E4,$47,$D3,$F8,$5E,$D1,$D0,$F9,$8F
	.dc.b	$57,$E7,$5F,$57,$E2,$DF,$57,$E9,$5F,$57,$E4,$DF,$DF,$F8,$69,$1F
	.dc.b	$5D,$FC,$17,$C2,$7E,$0F,$E1,$64,$FD,$B7,$F0,$DF,$0B,$F8,$7F,$85
	.dc.b	$FC,$47,$C3,$4A,$FC,$04,$72,$7F,$05,$75,$FB,$3F,$0B,$D1,$FB,$78
	.dc.b	$5E,$7F,$B5,$1F,$8D,$C7,$B3,$F8,$BC,$7B,$3F,$82,$C7,$F3,$1F,$0D
	.dc.b	$FC,$B7,$DA,$FF,$08,$F3,$7F,$35,$6D,$7E,$5E,$75,$F0,$3B,$F8,$19
	.dc.b	$3F,$C9,$2B,$FD,$CA,$3E,$07,$A1,$23,$5F,$0B,$E0,$7F,$3D,$85,$7F
	.dc.b	$D1,$F4,$F8,$5E,$F1,$D7,$0E,$EE,$B1,$E8,$F0,$C6,$ED,$F0,$BD,$56
	.dc.b	$2C,$FD,$D9,$4C,$3D,$0F,$C1,$F8,$5E,$F1,$E2,$CF,$BE,$3C,$F0,$F6
	.dc.b	$E5,$DB,$5F,$8C,$FC,$9C,$BE,$57,$C1,$FC,$BF,$84,$99,$7C,$58,$F9
	.dc.b	$DF,$07,$23,$E4,$3F,$47,$1F,$A5,$F0,$72,$7E,$63,$F5,$7E,$0F,$EB
	.dc.b	$7C,$27,$69,$E8,$7D,$17,$A3,$F2,$AE,$EB,$E9,$C3,$A4,$FC,$EC,$2C
	.dc.b	$FB,$70,$FC,$17,$C1,$FE,$0F,$E1,$25,$FD,$D8,$FC,$37,$C1,$48,$FC
	.dc.b	$03,$F8,$9F,$83,$FC,$57,$C1,$C9,$FC,$23,$F8,$DF,$83,$FC,$77,$C2
	.dc.b	$76,$9E,$87,$E2,$5E,$8F,$E0,$AE,$E7,$F4,$70,$E9,$3F,$86,$C2,$CF
	.dc.b	$D5,$C3,$FC,$0D,$F9,$5F,$F8,$58,$7C,$16,$0C,$DE,$BE,$0B,$7F,$0B
	.dc.b	$23,$FB,$ED,$7F,$86,$5B,$F0,$7D,$08,$EB,$E0,$F9,$BC,$0A,$7F,$BA
	.dc.b	$25,$F0,$9D,$E3,$AD,$F7,$75,$87,$47,$86,$17,$3F,$CD,$61,$FE,$0D
	.dc.b	$9F,$2E,$1E,$E2,$3E,$8F,$84,$EC,$F8,$71,$ED,$7C,$B1,$E9,$D9,$FE
	.dc.b	$31,$C7,$FC,$53,$7E,$E3,$D8,$7C,$2F,$84,$F9,$7F,$07,$23,$07,$C6
	.dc.b	$F8,$3F,$9E,$47,$E4,$FC,$1F,$D2,$F8,$3F,$A7,$F0,$92,$7E,$6C,$7D
	.dc.b	$6C,$24,$74,$3E,$47,$C1,$F3,$7E,$56,$17,$5F,$87,$8F,$59,$FA,$F8
	.dc.b	$F5,$7E,$86,$3F,$82,$F8,$4F,$C1,$FC,$14,$9F,$BB,$1F,$86,$F8,$39
	.dc.b	$1F,$80,$7F,$13,$F0,$5F,$8A,$F8,$3F,$C5,$FC,$24,$AF,$C2,$C7,$57
	.dc.b	$99,$0F,$C0,$7C,$27,$47,$F0,$57,$75,$FB,$78,$75,$9F,$C7,$E1,$D5
	.dc.b	$FC,$46,$1F,$E5,$FF,$CB,$3F,$9D,$C2,$FF,$8E,$43,$E0,$B4,$6F,$97
	.dc.b	$87,$C1,$CB,$FF,$2B,$D7,$F8,$43,$D1,$3F,$E0,$18,$CF,$C1,$77,$F8
	.dc.b	$49,$18,$38,$E1,$D9,$E7,$8F,$77,$AE,$3D,$DF,$34,$FF,$9F,$70,$F0
	.dc.b	$1D,$E3,$27,$FA,$31,$FF,$0E,$B8,$C7,$CB,$87,$8A,$31,$3A,$CB,$E0
	.dc.b	$DF,$86,$72,$FC,$7F,$82,$F1,$4F,$C6,$C6,$CF,$8F,$E1,$3E,$66,$9D
	.dc.b	$6B,$E5,$E5,$8C,$7E,$3F,$C2,$7D,$1F,$82,$91,$83,$E5,$FC,$1F,$D5
	.dc.b	$F8,$29,$18,$3E,$7F,$C1,$FD,$9F,$84,$91,$83,$91,$EC,$BE,$AF,$BB
	.dc.b	$0F,$C0,$61,$D2,$3E,$0E,$18,$79,$1F,$12,$FA,$BF,$1B,$0F,$C4,$7C
	.dc.b	$24,$8F,$C0,$BF,$8B,$F8,$4F,$C6,$7C,$1F,$E3,$7E,$0E,$57,$E1,$A3
	.dc.b	$F2,$1F,$05,$F9,$1F,$83,$93,$F8,$A8,$E4,$F4,$21,$F5,$AF,$AB,$F9
	.dc.b	$0B,$EA,$FD,$AB,$EA,$FE,$52,$FA,$BF,$7A,$FF,$41,$F0,$92,$3F,$32
	.dc.b	$FE,$8F,$E1,$3F,$D8,$5F,$96,$FF,$85,$3A,$8B,$E1,$72,$6B,$E2,$37
	.dc.b	$F0,$93,$1F,$F5,$CD,$23,$9F,$C2,$77,$9F,$C7,$E5,$7A,$3F,$5F,$9F
	.dc.b	$39,$E3,$97,$67,$9E,$5D,$5F,$CD,$65,$D5,$EF,$97,$67,$F6,$77,$D7
	.dc.b	$AF,$F4,$D9,$7B,$67,$99,$D8,$87,$4C,$3B,$65,$FE,$1D,$C7,$B4,$F8
	.dc.b	$31,$EC,$F8,$BE,$13,$E3,$FC,$34,$8F,$83,$1F,$2B,$E1,$BE,$5F,$C2
	.dc.b	$C9,$F8,$B1,$F3,$BE,$1A,$47,$C8,$7E,$8F,$C2,$FD,$2B,$93,$D0,$F9
	.dc.b	$BF,$0B,$D5,$F9,$F8,$75,$7E,$2E,$1D,$1F,$8F,$85,$D3,$D9,$8F,$DE
	.dc.b	$F8,$59,$5F,$61,$FC,$0F,$C2,$FE,$0B,$E1,$A4,$FD,$C7,$F0,$BF,$0B
	.dc.b	$F8,$6F,$86,$FC,$3F,$C2,$CA,$FC,$0C,$72,$7A,$10,$FB,$37,$D1,$FB
	.dc.b	$77,$72,$FA,$18,$75,$9F,$C5,$E1,$D5,$FC,$0E,$1F,$96,$F8,$5F,$CB
	.dc.b	$FC,$2C,$9F,$C8,$C7,$E6,$BE,$17,$F3,$7F,$0D,$27,$F2,$B0,$66,$5F
	.dc.b	$E5,$BF,$1B,$11,$8E,$AD,$77,$27,$FC,$3F,$39,$11,$9F,$86,$E3,$7C
	.dc.b	$E7,$CC,$79,$08,$7F,$33,$AF,$F1,$13,$F9,$CC,$3C,$CF,$E7,$B1,$F5
	.dc.b	$3D,$F0,$F1,$3D,$B1,$F2,$DF,$92,$3A,$63,$97,$CB,$1E,$C7,$F5,$39
	.dc.b	$79,$1A,$75,$8F,$1E,$5F,$A1,$CB,$C6,$FE,$8F,$2F,$1D,$7A,$4E,$D9
	.dc.b	$7C,$17,$96,$38,$C3,$C1,$7F,$37,$E0,$B0,$4A,$F0,$7C,$2E,$FE,$1A
	.dc.b	$47,$8D,$E5,$F0,$DC,$FE,$1A,$4F,$9D,$ED,$F0,$7D,$EF,$A4,$FC,$DC
	.dc.b	$30,$F3,$21,$F4,$70,$EA,$F3,$21,$F5,$6F,$AB,$F1,$7E,$13,$D8,$47
	.dc.b	$E4,$7C,$1F,$BF,$E0,$FE,$07,$C1,$49,$F9,$91,$F0,$FE,$0F,$E2,$7C
	.dc.b	$24,$9F,$A1,$1C,$DF,$C2,$61,$77,$FC,$A6,$3D,$67,$EB,$63,$D5,$FC
	.dc.b	$56,$3D,$5F,$B5,$8F,$57,$F1,$D8,$FD,$3F,$83,$91,$F7,$DF,$AD,$F0
	.dc.b	$9F,$5F,$E0,$FE,$C7,$C1,$CA,$FC,$24,$7D,$BF,$82,$91,$F8,$77,$EF
	.dc.b	$7C,$1F,$28,$E8,$43,$F3,$D7,$D5,$FC,$75,$F5,$7F,$45,$7D,$1F,$D2
	.dc.b	$5D,$DB,$F6,$98,$75,$9F,$D4,$63,$EA,$7A,$9E,$63,$A9,$0E,$A7,$9C
	.dc.b	$87,$63,$F4,$11,$F9,$7F,$84,$FC,$DF,$E5,$DF,$E4,$4B,$FF,$6F,$37
	.dc.b	$BB,$F5,$F3,$AF,$85,$98,$7E,$D6,$BF,$C9,$0D,$F8,$3E,$7F,$07,$D3
	.dc.b	$E0,$E4,$EE,$3C,$0F,$EC,$70,$AF,$F8,$DE,$DF,$07,$DA,$31,$8D,$CF
	.dc.b	$CB,$97,$69,$F1,$65,$D9,$E5,$97,$B3,$E1,$3D,$BF,$07,$27,$D1,$1F
	.dc.b	$03,$E1,$24,$7A,$DF,$87,$F0,$9F,$13,$C2,$93,$EF,$7E,$3F,$C1,$F5
	.dc.b	$8F,$85,$87,$57,$D5,$87,$47,$D9,$85,$D3,$E6,$E3,$D2,$7E,$06,$36
	.dc.b	$7E,$8E,$3F,$53,$E1,$3E,$AF,$C2,$4B,$FA,$11,$F6,$3E,$12,$47,$D3
	.dc.b	$7E,$DF,$C1,$FD,$CF,$84,$FB,$BF,$07,$2B,$EC,$47,$57,$99,$0F,$A5
	.dc.b	$F0,$9D,$1F,$A9,$77,$3F,$C2,$E1,$D6,$7F,$05,$87,$57,$ED,$61,$D5
	.dc.b	$FB,$98,$DB,$FC,$6E,$3C,$FA,$3F,$A7,$F3,$B3,$C6,$1F,$D3,$A1,$9F
	.dc.b	$78,$FC,$36,$7D,$F1,$FC,$BE,$7F,$9A,$F8,$1F,$CD,$FC,$14,$BF,$CB
	.dc.b	$47,$F3,$8F,$CB,$FF,$C7,$C7,$F8,$F6,$3F,$A7,$35,$FF,$44,$3A,$F8
	.dc.b	$1E,$04,$8D,$FC,$2F,$77,$F4,$57,$DD,$FC,$D5,$F7,$F8,$4F,$07,$C2
	.dc.b	$C9,$EF,$1C,$DE,$F8,$5C,$DC,$1F,$DC,$B3,$67,$E9,$F8,$39,$87,$A1
	.dc.b	$F6,$7C,$27,$B7,$E0,$BD,$DF,$07,$2B,$DB,$1F,$07,$E0,$FE,$17,$C1
	.dc.b	$C9,$F8,$31,$D1,$F8,$38,$5D,$3C,$71,$FE,$00,$CD,$97,$78,$F9,$39
	.dc.b	$77,$7E,$4E,$5D,$DF,$9B,$97,$77,$E6,$E5,$F4,$FE,$13,$EA,$7C,$24
	.dc.b	$9F,$A7,$1F,$5F,$E1,$24,$7D,$67,$ED,$7C,$27,$DB,$F8,$4F,$B9,$F0
	.dc.b	$72,$BE,$DC,$75,$7E,$EE,$1D,$5F,$BB,$87,$47,$EF,$E1,$77,$F7,$FB
	.dc.b	$EE,$06,$7D,$EF,$F0,$F9,$F7,$7F,$0F,$9F,$77,$F1,$39,$F7,$C7,$9E
	.dc.b	$0C,$FB,$71,$FD,$DD,$2E,$1F,$37,$3E,$CF,$E4,$B3,$ED,$8F,$2C,$FF
	.dc.b	$35,$F0,$9F,$9B,$F8,$39,$7F,$9A,$8F,$CF,$7C,$27,$E7,$FE,$12,$4F
	.dc.b	$E7,$A3,$F4,$5F,$09,$FE,$1A,$FC,$C3,$FC,$2F,$B7,$F9,$01,$B5,$F7
	.dc.b	$E7,$BB,$FA,$6C,$7B,$BF,$A6,$C7,$BB,$FA,$AC,$7B,$BF,$AA,$C7,$BB
	.dc.b	$FE,$3B,$C3,$AF,$11,$BC,$2C,$BC,$9F,$0B,$E5,$F8,$99,$87,$18,$F4
	.dc.b	$7C,$1F,$A7,$E1,$24,$F5,$8F,$67,$C2,$48,$F0,$3E,$FF,$86,$EB,$1E
	.dc.b	$3C,$3A,$AE,$3F,$C0,$E9,$FE,$4D,$6E,$1D,$A7,$5C,$5F,$1F,$2F,$1C
	.dc.b	$76,$21,$A3,$FC,$CF,$6C,$BC,$51,$F0,$FE,$17,$E9,$7C,$34,$8C,$1F
	.dc.b	$1F,$E1,$BE,$B1,$1F,$95,$F0,$DF,$67,$E1,$64,$7C,$D7,$EE,$7C,$37
	.dc.b	$DD,$C2,$4F,$43,$C9,$C9,$F4,$7E,$17,$A4,$63,$46,$58,$E0,$F8,$79
	.dc.b	$F6,$AD,$67,$67,$E3,$E7,$DA,$79,$67,$67,$E5,$E5,$CF,$7F,$C7,$65
	.dc.b	$6F,$E2,$E5,$CE,$7F,$25,$95,$9F,$93,$97,$E5,$FE,$1A,$47,$E2,$DF
	.dc.b	$CD,$7C,$17,$E6,$FE,$1B,$F3,$9F,$0B,$2B,$F2,$11,$F9,$FF,$86,$FD
	.dc.b	$07,$C3,$49,$FC,$A4,$72,$7A,$10,$F8,$97,$D5,$FC,$D5,$F5,$7E,$45
	.dc.b	$F5,$7F,$3D,$7D,$5F,$99,$7F,$AF,$F8,$69,$1F,$A3,$7F,$65,$F0,$DF
	.dc.b	$E8,$CD,$3F,$C6,$13,$FE,$AE,$7F,$C5,$AF,$D8,$6F,$0F,$F1,$67,$7C
	.dc.b	$85,$F0,$9C,$FE,$0E,$5F,$FA,$AE,$72,$FE,$CF,$0F,$1B,$F8,$1C,$7C
	.dc.b	$6E,$F8,$7C,$9F,$0B,$23,$07,$3C,$7A,$BE,$5C,$7D,$51,$EB,$F8,$5F
	.dc.b	$67,$C3,$4A,$F1,$47,$BF,$E1,$7E,$07,$C3,$49,$F3,$47,$C3,$F8,$6F
	.dc.b	$89,$F0,$D2,$7D,$51,$D5,$F6,$61,$D5,$F8,$D8,$75,$7E,$06,$1D,$5F
	.dc.b	$95,$87,$57,$E2,$61,$D5,$F9,$D8,$75,$79,$9E,$83,$E5,$7C,$37,$57
	.dc.b	$EA,$5F,$5E,$2F,$A9,$87,$59,$FB,$18,$75,$9F,$B1,$87,$DF,$F8,$69
	.dc.b	$5F,$55,$FC,$17,$C2,$FE,$0F,$E1,$A4,$FD,$97,$F0,$DF,$0B,$F8,$7F
	.dc.b	$84,$93,$F7,$5F,$C5,$7C,$1F,$48,$FC,$46,$17,$2F,$C4,$63,$D2,$7F
	.dc.b	$19,$8D,$9F,$C6,$63,$D2,$7F,$21,$8D,$9F,$C8,$63,$F9,$7F,$84,$95
	.dc.b	$F8,$D7,$F3,$5F,$0D,$F9,$BF,$84,$FC,$E7,$C2,$CA,$FC,$94,$7F,$BA
	.dc.b	$BF,$31,$7C,$2D,$F8,$69,$5F,$F0,$06,$72,$F5,$21,$FA,$3C,$7B,$BD
	.dc.b	$0F,$CF,$61,$DD,$FD,$46,$1D,$F0,$EB,$1F,$C8,$AF,$FC,$81,$9F,$F8
	.dc.b	$01,$EF,$19,$F8,$6F,$31,$1F,$27,$C3,$7A,$7E,$1A,$47,$9D,$F6,$7C
	.dc.b	$37,$B7,$E1,$64,$FA,$DF,$81,$F0,$DF,$06,$FA,$4F,$73,$A9,$CC,$87
	.dc.b	$9E,$FA,$BC,$C8,$7A,$EB,$AB,$C8,$E8,$4C,$BD,$D5,$E8,$8E,$87,$68
	.dc.b	$F8,$58,$7A,$23,$A1,$DA,$3E,$36,$1F,$67,$E1,$64,$FD,$77,$EE,$7C
	.dc.b	$37,$DD,$F8,$5F,$BD,$F0,$D2,$BE,$E4,$7E,$07,$E1,$A4,$7D,$F7,$F0
	.dc.b	$9F,$0D,$D2,$3E,$B6,$17,$2F,$AD,$8F,$59,$FC,$36,$3D,$5F,$B9,$8F
	.dc.b	$47,$EF,$63,$6F,$EF,$63,$F9,$1F,$85,$FC,$97,$C3,$4B,$FC,$84,$7E
	.dc.b	$57,$E1,$7F,$2D,$F0,$D2,$7F,$29,$1F,$99,$F8,$5F,$CD,$7C,$34,$9F
	.dc.b	$CC,$47,$57,$F3,$58,$75,$7F,$21,$87,$57,$F3,$D8,$75,$7F,$29,$87
	.dc.b	$57,$F4,$58,$75,$7F,$31,$87,$EA,$7E,$15,$13,$3F,$F2,$CF,$E6,$5F
	.dc.b	$E2,$E3,$FC,$5F,$D1,$FE,$2D,$9D,$79,$FC,$3D,$35,$7F,$87,$E3,$0B
	.dc.b	$E1,$3A,$D7,$F8,$22,$30,$F2,$15,$CB,$E1,$B7,$8F,$8B,$E1,$6E,$3D
	.dc.b	$3E,$16,$E1,$C3,$14,$63,$1F,$1D,$FB,$6B,$F8,$16,$95,$F1,$7C,$2E
	.dc.b	$F0,$F4,$47,$3D,$30,$67,$4D,$61,$BC,$A5,$7A,$72,$EE,$F6,$21,$FC
	.dc.b	$17,$2E,$D9,$73,$8E,$19,$7C,$27,$E2,$65,$97,$E3,$63,$D9,$F9,$18
	.dc.b	$F8,$9E,$3F,$0D,$E9,$23,$D3,$E1,$3D,$9F,$0D,$23,$BB,$EF,$F8,$4F
	.dc.b	$81,$F0,$92,$7C,$6F,$C3,$F7,$7E,$25,$F4,$9F,$B1,$86,$1E,$64,$3E
	.dc.b	$E6,$1D,$5E,$64,$3F,$01,$7D,$5F,$83,$F0,$9F,$44,$8F,$C4,$F8,$5F
	.dc.b	$A9,$F0,$B2,$3E,$3B,$F5,$FE,$17,$EC,$7C,$24,$9F,$96,$FD,$BF,$85
	.dc.b	$FB,$97,$D2,$7F,$21,$86,$1E,$64,$3F,$29,$87,$57,$99,$0F,$CC,$5F
	.dc.b	$57,$EC,$FC,$27,$E2,$48,$FD,$CF,$85,$FC,$67,$C2,$7E,$37,$E1,$64
	.dc.b	$FE,$02,$3F,$2D,$F9,$9F,$F8,$C0,$FF,$19,$F6,$7E,$AD,$B3,$FA,$47
	.dc.b	$FC,$6B,$D0,$C7,$BC,$FE,$AD,$FC,$D6,$3E,$08,$FC,$6E,$3E,$07,$F6
	.dc.b	$18,$F8,$31,$F0,$FC,$34,$9F,$CD,$65,$DB,$1E,$BD,$DF,$C0,$9E,$B3
	.dc.b	$E9,$3A,$90,$EA,$72,$BE,$8F,$FA,$1A,$BD,$7D,$A7,$E0,$7C,$2F,$C2
	.dc.b	$F8,$69,$5E,$28,$F8,$BF,$0B,$23,$CA,$FC,$8F,$86,$F9,$3F,$09,$27
	.dc.b	$D2,$FC,$CF,$85,$F9,$B7,$27,$A1,$EE,$F8,$4E,$8F,$97,$0B,$B7,$FB
	.dc.b	$E3,$36,$5D,$6B,$D5,$95,$AD,$65,$F6,$BE,$16,$57,$CA,$7E,$EF,$C3
	.dc.b	$7D,$EF,$86,$93,$F3,$9F,$C0,$FC,$37,$E0,$BE,$1B,$F0,$7F,$0D,$2B
	.dc.b	$E9,$C7,$27,$A1,$0F,$99,$7D,$1F,$9D,$77,$2F,$56,$1D,$67,$EE,$E1
	.dc.b	$D5,$FA,$98,$7E,$4B,$E1,$BF,$27,$F0,$D2,$7F,$07,$1F,$96,$F8,$4F
	.dc.b	$CB,$FC,$34,$9F,$C3,$C7,$E6,$BE,$12,$47,$E2,$9F,$CE,$FC,$2F,$58
	.dc.b	$E6,$43,$FD,$93,$EE,$7F,$89,$B5,$FA,$17,$D6,$7F,$11,$87,$85,$E8
	.dc.b	$43,$F1,$98,$76,$7A,$57,$D9,$C7,$C1,$1D,$BE,$17,$93,$D0,$FC,$FD
	.dc.b	$E2,$BF,$01,$7E,$A9,$EA,$7E,$93,$2F,$E9,$99,$7F,$8F,$BA,$FF,$B0
	.dc.b	$36,$7F,$AC,$B7,$AC,$BE,$17,$8F,$C1,$CA,$FE,$91,$28,$EB,$F0,$BD
	.dc.b	$BE,$13,$BF,$C2,$CB,$DC,$78,$BE,$1A,$47,$27,$CB,$8F,$58,$F9,$B8
	.dc.b	$5D,$9A,$7A,$32,$ED,$5E,$2C,$BB,$3F,$53,$2F,$7F,$C3,$7C,$0F,$86
	.dc.b	$93,$E8,$8F,$87,$F0,$D2,$3D,$6F,$C6,$F8,$6F,$8F,$BF,$C8,$F8,$69
	.dc.b	$5F,$02,$3A,$BC,$C8,$7E,$13,$E1,$BA,$3F,$86,$BB,$B7,$AA,$3E,$7E
	.dc.b	$1D,$9F,$97,$87,$67,$F1,$B8,$7D,$8F,$85,$91,$F4,$1F,$B7,$F0,$DF
	.dc.b	$73,$E1,$64,$FD,$47,$EF,$FC,$37,$E0,$3E,$16,$4F,$D8,$7F,$07,$F0
	.dc.b	$DD,$23,$F3,$78,$5D,$3C,$DD,$1F,$82,$CB,$AD,$7E,$83,$2B,$5A,$CD
	.dc.b	$9F,$6A,$FD,$2E,$76,$B5,$9F,$A2,$7D,$44,$3F,$57,$8F,$4E,$5F,$CA
	.dc.b	$E5,$67,$F2,$B8,$F3,$9F,$CC,$E3,$67,$F3,$38,$7E,$7B,$E1,$3F,$C5
	.dc.b	$5F,$9A,$7F,$8E,$25,$9F,$B2,$30,$FA,$67,$7F,$0D,$C3,$E1,$65,$FF
	.dc.b	$89,$6F,$FC,$5B,$B9,$7C,$2F,$62,$4E,$FE,$17,$C4,$F1,$3C,$66,$0F
	.dc.b	$1E,$BF,$E5,$1A,$EA,$7F,$8A,$F3,$F3,$63,$E4,$AF,$16,$1D,$9E,$27
	.dc.b	$F9,$42,$BC,$6F,$84,$F4,$7C,$37,$C2,$23,$EB,$F8,$3F,$8B,$F0,$92
	.dc.b	$3D,$CF,$C8,$F8,$4F,$93,$F0,$D2,$7E,$13,$F3,$3E,$17,$9C,$78,$B7
	.dc.b	$F8,$98,$E3,$5F,$9D,$97,$5A,$D6,$6C,$F1,$5F,$4F,$3E,$D5,$AC,$EC
	.dc.b	$FD,$7C,$FE,$DF,$C3,$7D,$CF,$83,$97,$F5,$63,$EF,$FC,$27,$E0,$3E
	.dc.b	$12,$4F,$D9,$8F,$C1,$FC,$37,$E1,$3E,$16,$4F,$DD,$8E,$AF,$DF,$C3
	.dc.b	$AB,$E8,$C3,$AB,$F8,$3C,$3A,$BE,$CC,$3A,$BF,$87,$C3,$AB,$F0,$30
	.dc.b	$EA,$F3,$3D,$07,$E3,$7E,$1B,$AB,$F1,$AF,$AE,$7F,$97,$C3,$AC,$FC
	.dc.b	$AC,$3A,$CF,$E6,$F0,$FC,$FF,$C0,$CA,$FC,$C3,$FA,$2F,$85,$FD,$1F
	.dc.b	$C2,$49,$FC,$E3,$FA,$6F,$85,$FF,$48,$7E,$6B,$FC,$56,$7E,$E3,$FE
	.dc.b	$20,$27,$57,$58,$FB,$78,$5C,$BF,$59,$8F,$59,$FB,$F8,$D9,$FD,$86
	.dc.b	$3D,$B0,$EF,$F0,$D2,$FF,$D2,$59,$78,$6F,$F8,$8A,$F8,$5F,$29,$27
	.dc.b	$7F,$0D,$E8,$F8,$59,$1C,$9F,$5F,$C2,$7B,$3E,$12,$4F,$67,$DF,$F0
	.dc.b	$9C,$E3,$DF,$85,$CB,$C3,$8F,$49,$F8,$78,$D9,$F2,$E3,$D2,$7E,$3E
	.dc.b	$36,$7D,$38,$FC,$CF,$86,$95,$EE,$7E,$7F,$C2,$FD,$0F,$85,$93,$F0
	.dc.b	$9F,$A7,$F0,$BF,$53,$E1,$7E,$AF,$C2,$4A,$F8,$F1,$C9,$E8,$43,$EC
	.dc.b	$DF,$47,$ED,$DD,$CB,$E6,$E1,$D6,$7E,$8E,$1D,$5F,$C0,$E1,$D5,$FC
	.dc.b	$1E,$3E,$A7,$A9,$E7,$21,$D8,$FB,$75,$D5,$FC,$5D,$75,$BF,$BB,$7F
	.dc.b	$90,$F8,$59,$5F,$82,$7F,$27,$F0,$7F,$94,$F8,$59,$3F,$86,$7F,$2F
	.dc.b	$F0,$BF,$98,$F8,$49,$3F,$8A,$7F,$37,$F0,$9D,$23,$F3,$78,$5C,$BF
	.dc.b	$1B,$8F,$49,$FC,$FE,$36,$7F,$23,$8F,$49,$FD,$1E,$36,$7F,$2B,$8F
	.dc.b	$EA,$3E,$13,$FD,$F1,$F9,$B7,$F9,$0F,$A9,$83,$CD,$4E,$63,$1D,$4B
	.dc.b	$E3,$F0,$BC,$BE,$16,$57,$FB,$D6,$FF,$C6,$87,$F8,$BD,$C7,$C2,$78
	.dc.b	$0F,$0F,$C2,$F8,$E3,$87,$C3,$79,$0C,$3F,$CD,$EB,$9D,$F6,$8C,$61
	.dc.b	$C9,$EA,$43,$85,$FB,$BE,$17,$DF,$F0,$D2,$7C,$D1,$F0,$BE,$16,$47
	.dc.b	$A5,$F8,$BF,$0D,$F1,$BE,$17,$E3,$FC,$2C,$AF,$74,$75,$79,$90,$F4
	.dc.b	$7C,$1F,$47,$D5,$77,$97,$FC,$2F,$B3,$1E,$D5,$F2,$71,$EC,$FC,$1C
	.dc.b	$7B,$3F,$0F,$2B,$87,$2C,$BA,$72,$7D,$AE,$4F,$B3,$A6,$30,$FB,$9A
	.dc.b	$77,$7D,$27,$72,$1D,$CF,$B5,$F0,$BF,$86,$F8,$69,$18,$3E,$F7,$C3
	.dc.b	$7E,$28,$8F,$E0,$7E,$1B,$F1,$BF,$0D,$23,$F0,$8F,$E4,$3E,$1B,$F2
	.dc.b	$38,$49,$E8,$7D,$FF,$86,$E6,$FE,$07,$0B,$BF,$C8,$C7,$A4,$FE,$17
	.dc.b	$1B,$3F,$33,$1F,$CE,$FC,$34,$AF,$CA,$3F,$A0,$F8,$6F,$CF,$C7,$F8
	.dc.b	$0F,$C6,$FF,$16,$F0,$FC,$96,$C7,$E8,$DB,$C2,$7F,$C5,$A7,$F8,$86
	.dc.b	$78,$7C,$1F,$4C,$BF,$CA,$73,$FE,$1E,$BE,$DF,$09,$E3,$7F,$39,$95
	.dc.b	$AF,$C1,$E5,$C7,$1F,$15,$53,$C4,$5C,$BC,$4D,$C7,$C7,$8D,$CA,$4F
	.dc.b	$EF,$31,$F1,$CE,$FE,$17,$87,$C2,$CA,$D6,$5F,$06,$7A,$7C,$2C,$9C
	.dc.b	$9F,$0D,$F0,$7C,$27,$87,$E1,$3C,$5F,$09,$2B,$7B,$7C,$79,$F0,$46
	.dc.b	$36,$67,$E1,$8F,$A5,$97,$77,$EA,$65,$75,$F9,$79,$F7,$9E,$59,$F7
	.dc.b	$7E,$D6,$5D,$9F,$B9,$8F,$5B,$FA,$78,$F3,$D3,$EF,$65,$6F,$EF,$65
	.dc.b	$D5,$FC,$26,$5D,$70,$FB,$79,$7C,$FF,$84,$FA,$1F,$07,$2F,$D9,$1F
	.dc.b	$4F,$E1,$24,$7B,$DF,$AD,$F0,$5F,$5F,$E0,$E4,$FC,$37,$ED,$7C,$17
	.dc.b	$58,$F8,$D8,$75,$7F,$2D,$87,$47,$F3,$18,$5C,$BF,$1F,$8F,$49,$FC
	.dc.b	$E6,$36,$7F,$27,$8F,$E1,$FE,$0F,$F1,$1F,$09,$2F,$EA,$47,$E2,$FE
	.dc.b	$12,$47,$D7,$7F,$1D,$F0,$9F,$8F,$F8,$4F,$C8,$7C,$24,$AF,$B9,$1D
	.dc.b	$5F,$BD,$87,$57,$F5,$D8,$74,$7F,$61,$85,$CB,$F4,$F8,$F5,$9F,$C3
	.dc.b	$63,$D5,$FD,$B6,$3F,$A1,$FC,$DF,$FC,$61,$CB,$FC,$7D,$7C,$2E,$4D
	.dc.b	$46,$FE,$16,$5F,$E5,$2B,$FC,$6C,$DF,$85,$E9,$F0,$BD,$7E,$16,$4E
	.dc.b	$E3,$9B,$C7,$0A,$FF,$A2,$ED,$F0,$BD,$A3,$18,$DD,$BF,$DA,$B4,$CB
	.dc.b	$BD,$78,$F2,$EE,$F3,$CB,$AD,$F0,$CA,$9E,$FC,$BA,$3F,$07,$2B,$7F
	.dc.b	$DE,$72,$EC,$FA,$4E,$C4,$3B,$1F,$0F,$E1,$7E,$67,$C1,$C8,$C1,$F1
	.dc.b	$FE,$17,$E8,$11,$F9,$5F,$0B,$F4,$FC,$29,$1F,$35,$FA,$DF,$09,$F5
	.dc.b	$F0,$E9,$3F,$1F,$0C,$3D,$48,$7C,$BC,$3E,$F6,$1D,$23,$E7,$61,$75
	.dc.b	$F9,$78,$FE,$0F,$E1,$7F,$09,$F0,$B2,$FE,$F4,$7E,$1F,$E1,$64,$7E
	.dc.b	$05,$FC,$57,$C2,$FE,$2F,$E1,$64,$FE,$15,$FC,$77,$C2,$F5,$8F,$C4
	.dc.b	$61,$D5,$FC,$06,$1D,$1F,$C1,$61,$72,$FB,$F8,$F4,$9F,$C3,$63,$67
	.dc.b	$F0,$78,$FE,$6F,$E1,$7F,$39,$F0,$92,$FF,$2D,$1F,$CF,$FF,$38,$FF
	.dc.b	$19,$79,$A5,$F0,$59,$35,$F0,$9B,$F8,$4E,$1F,$09,$31,$FE,$67,$69
	.dc.b	$F0,$47,$E8,$F2,$F0,$3F,$9B,$CB,$9B,$C3,$1E,$CF,$2C,$7B,$3D,$31
	.dc.b	$EC,$FF,$BB,$71,$EC,$FA,$88,$75,$3C,$7F,$0B,$D9,$E1,$7D,$9E,$57
	.dc.b	$D9,$E9,$7D,$B5,$75,$FF,$87,$F1,$EF,$5E,$1F,$81,$F9,$04,$7E,$0F
	.dc.b	$C2,$FC,$BF,$81,$F9,$9F,$0B,$27,$E2,$C7,$CF,$F8,$4F,$A1,$F0,$B2
	.dc.b	$7E,$4C,$7D,$3B,$E9,$1D,$CE,$A7,$23,$E7,$DF,$57,$E2,$5F,$57,$E9
	.dc.b	$DF,$57,$E4,$5F,$DF,$F8,$5F,$C0,$7C,$1C,$9F,$B3,$1F,$83,$F8,$49
	.dc.b	$1F,$71,$FC,$37,$C1,$FE,$1F,$E0,$FF,$11,$F0,$72,$BF,$03,$1D,$5F
	.dc.b	$C1,$E1,$D5,$FB,$18,$74,$7E,$D6,$17,$87,$E7,$C7,$FA,$17,$81,$97
	.dc.b	$78,$FC,$6E,$5D,$DF,$C1,$65,$F9,$9F,$83,$FC,$D7,$C2,$49,$FC,$9C
	.dc.b	$7E,$77,$E1,$24,$7E,$59,$FD,$07,$C2,$7F,$87,$FF,$39,$FF,$18,$8F
	.dc.b	$89,$C1,$9A,$EC,$7E,$7B,$0E,$CF,$E4,$F0,$EA,$FE,$57,$0B,$AF,$B3
	.dc.b	$BF,$F5,$99,$73,$AF,$F0,$CE,$5D,$DB,$96,$5E,$87,$72,$18,$3B,$1E
	.dc.b	$B2,$1F,$A5,$9E,$57,$E1,$AE,$D2,$8C,$56,$BA,$3F,$CF,$95,$FE,$4F
	.dc.b	$AF,$F5,$B3,$FE,$4D,$AE,$5F,$0D,$F1,$FE,$1B,$E4,$7C,$2C,$BF,$5B
	.dc.b	$F2,$FE,$1B,$E6,$7C,$37,$CD,$F8,$69,$5F,$02,$39,$3F,$0B,$0F,$A4
	.dc.b	$61,$F3,$67,$E4,$C3,$B4,$60,$F2,$BD,$DE,$87,$CC,$BE,$EE,$F6,$68
	.dc.b	$C7,$F0,$1F,$0B,$2B,$E9,$BF,$83,$F8,$6F,$C2,$7C,$2C,$9F,$AE,$FE
	.dc.b	$1F,$E1,$BF,$11,$F0,$DF,$89,$F8,$69,$5F,$72,$3A,$3E,$1C,$98,$E3
	.dc.b	$8B,$EA,$65,$D6,$B5,$95,$9F,$B1,$97,$69,$FC,$3E,$5D,$9E,$99,$7E
	.dc.b	$67,$E1,$A4,$7E,$31,$FC,$E7,$C3,$7E,$77,$E1,$7F,$3D,$F0,$D2,$BF
	.dc.b	$23,$1F,$A1,$F8,$69,$1F,$94,$7F,$49,$F0,$DD,$23,$D5,$85,$D3,$F1
	.dc.b	$98,$F5,$9F,$CD,$E3,$D5,$F8,$18,$F4,$7E,$16,$36,$FF,$29,$8F,$9A
	.dc.b	$7D,$04,$3E,$46,$1D,$5F,$95,$87,$5A,$FC,$E6,$1C,$F9,$7F,$33,$8D
	.dc.b	$BF,$CC,$E3,$FB,$DF,$85,$91,$FB,$07,$F8,$07,$C3,$7F,$A7,$F5,$FF
	.dc.b	$1F,$CB,$95,$1F,$E4,$08,$FF,$47,$B5,$FC,$63,$C2,$37,$F0,$3C,$C9
	.dc.b	$3C,$3E,$17,$BB,$E1,$21,$F8,$5B,$F1,$67,$E3,$F8,$69,$3E,$08,$E6
	.dc.b	$F8,$2E,$ED,$FC,$37,$2F,$F1,$9C,$F9,$3E,$17,$D8,$47,$D3,$F0,$DE
	.dc.b	$FF,$86,$F8,$1F,$0B,$27,$DB,$1F,$0F,$E1,$7E,$27,$C2,$C9,$F8,$31
	.dc.b	$F1,$EF,$A4,$77,$3A,$9C,$8F,$8F,$7D,$5F,$8F,$7D,$5F,$97,$7D,$5F
	.dc.b	$97,$7F,$4F,$E1,$64,$7D,$07,$EB,$7C,$2F,$D7,$F8,$5F,$B1,$F0,$92
	.dc.b	$BE,$AC,$7D,$BF,$85,$FB,$9F,$0B,$27,$EC,$C7,$47,$EC,$E1,$78,$3E
	.dc.b	$56,$3D,$67,$EF,$E3,$D5,$FB,$F8,$F5,$7F,$07,$8F,$57,$F0,$78,$F9
	.dc.b	$9F,$41,$0F,$C3,$DF,$57,$F1,$37,$D7,$0F,$B5,$87,$59,$FC,$6E,$1D
	.dc.b	$67,$EF,$61,$F9,$7F,$85,$95,$F9,$47,$F3,$5F,$0B,$F9,$BF,$85,$FC
	.dc.b	$E7,$C1,$CA,$FC,$CC,$7E,$7F,$E1,$3F,$DF,$DF,$9D,$7F,$90,$0F,$F2
	.dc.b	$0F,$32,$3A,$D7,$E7,$B1,$B9,$FE,$3F,$2E,$D3,$FA,$4C,$BB,$3F,$A4
	.dc.b	$CB,$B3,$FE,$F6,$CB,$B7,$C1,$E2,$3F,$9A,$47,$F9,$39,$D7,$C3,$79
	.dc.b	$0F,$2F,$C2,$F9,$BE,$16,$4F,$37,$D3,$F0,$DE,$AF,$86,$93,$DD,$F6
	.dc.b	$FC,$37,$28,$F1,$5D,$7C,$3F,$0B,$D1,$F1,$E1,$77,$FE,$F1,$9B,$2E
	.dc.b	$B5,$E8,$CA,$D6,$B2,$F9,$7F,$0D,$2B,$E0,$3F,$3B,$E1,$7E,$7F,$C3
	.dc.b	$49,$F8,$8F,$D2,$F8,$5F,$A7,$F0,$DF,$53,$E1,$A5,$7C,$98,$E4,$F4
	.dc.b	$21,$F2,$AF,$A3,$F3,$2E,$E5,$EA,$C3,$AC,$FD,$3C,$3A,$BF,$4B,$0E
	.dc.b	$AF,$D4,$C7,$D4,$F5,$3C,$E4,$3B,$1F,$76,$BA,$BF,$72,$BA,$DF,$CA
	.dc.b	$BF,$C7,$FC,$2C,$AF,$C2,$3F,$92,$F8,$4F,$C9,$FC,$2F,$E5,$3E,$12
	.dc.b	$57,$E2,$63,$F2,$FF,$0B,$23,$F1,$8F,$E6,$BE,$17,$A4,$7E,$33,$0B
	.dc.b	$97,$DA,$C7,$AC,$FE,$4F,$1E,$AF,$E4,$B1,$E8,$FE,$53,$1B,$7F,$82
	.dc.b	$C7,$F4,$FF,$0A,$A9,$3F,$E3,$DF,$CE,$F1,$18,$F3,$B5,$E9,$27,$33
	.dc.b	$FA,$0B,$DD,$F1,$F8,$6B,$87,$0B,$FF,$19,$61,$C7,$E1,$B3,$8F,$F5
	.dc.b	$59,$FF,$1D,$E1,$E1,$F8,$5C,$DF,$7F,$85,$B4,$6C,$FA,$AF,$CC,$73
	.dc.b	$36,$64,$FF,$14,$4F,$6C,$7C,$98,$F4,$3A,$E5,$D2,$B9,$E3,$E1,$6D
	.dc.b	$F8,$5B,$87,$93,$E1,$BC,$BF,$0B,$27,$71,$E8,$F8,$6F,$4F,$C2,$C9
	.dc.b	$E7,$1E,$CF,$85,$F6,$FC,$2C,$9E,$F1,$D5,$E6,$43,$D7,$F0,$BD,$5F
	.dc.b	$2D,$F5,$7D,$F7,$D5,$F4,$DF,$57,$E1,$DF,$57,$E2,$E1,$78,$7F,$05
	.dc.b	$8F,$59,$F9,$38,$F5,$9F,$C3,$63,$D6,$7E,$6E,$3D,$67,$F1,$58,$FD
	.dc.b	$8F,$37,$EC,$FC,$2C,$BF,$97,$1F,$73,$E1,$64,$7C,$E7,$EF,$FC,$2F
	.dc.b	$E0,$3E,$13,$F0,$3F,$0B,$2B,$E9,$C7,$57,$EA,$E1,$D5,$FB,$F8,$74
	.dc.b	$7F,$03,$85,$CB,$F3,$D8,$F5,$9F,$BB,$8F,$57,$F0,$F8,$FE,$43,$E1
	.dc.b	$64,$7E,$09,$FC,$9F,$C2,$FE,$53,$E1,$7F,$2B,$F0,$92,$BF,$0F,$1F
	.dc.b	$9A,$FC,$F3,$FC,$58,$7F,$8B,$F3,$FF,$55,$37,$AF,$F5,$AF,$F8,$C7
	.dc.b	$A1,$97,$83,$0F,$CC,$BE,$07,$A9,$F9,$4C,$7C,$0F,$E7,$B1,$F0,$61
	.dc.b	$E1,$F8,$59,$3F,$9C,$C7,$C7,$F0,$BE,$5F,$86,$93,$98,$F4,$7C,$34
	.dc.b	$8E,$0F,$AF,$E1,$BD,$9F,$0B,$ED,$F8,$69,$5D,$63,$E0,$5F,$48,$FF
	.dc.b	$4C,$63,$FE,$3C,$E0,$F8,$33,$D6,$73,$8D,$9F,$8B,$8F,$69,$F4,$63
	.dc.b	$D9,$E7,$8F,$67,$D9,$97,$7B,$EB,$1E,$82,$1D,$8F,$85,$7D,$5F,$2D
	.dc.b	$F5,$9F,$AB,$7F,$67,$E1,$7E,$D7,$C3,$4B,$F9,$51,$F7,$7E,$1B,$EF
	.dc.b	$7C,$34,$9F,$9D,$1B,$7E,$0D,$5C,$3F,$03,$F0,$BD,$67,$EA,$4E,$78
	.dc.b	$7F,$07,$8F,$59,$FB,$18,$F5,$7E,$5E,$3B,$7F,$CF,$DC,$0C,$31,$E2
	.dc.b	$A1,$97,$E4,$FE,$1B,$F2,$9F,$0B,$30,$FC,$24,$7E,$5F,$E1,$A4,$7E
	.dc.b	$1D,$FC,$D7,$C2,$FE,$6F,$E1,$64,$FE,$2D,$FC,$F7,$C2,$75,$8F,$C7
	.dc.b	$61,$D5,$FC,$0E,$1D,$1F,$C1,$E1,$78,$7F,$47,$8F,$49,$FC,$3E,$36
	.dc.b	$7F,$4F,$8F,$EA,$FE,$17,$F5,$9F,$0D,$2F,$F3,$91,$FD,$23,$D8,$7C
	.dc.b	$35,$24,$FF,$3C,$8F,$F1,$B3,$FE,$EB,$5F,$0D,$C4,$91,$BF,$84,$F0
	.dc.b	$3F,$A9,$C3,$C0,$FE,$73,$0E,$AF,$3C,$2E,$4E,$7F,$ED,$39,$78,$23
	.dc.b	$15,$FE,$B5,$8F,$0B,$E5,$CB,$C2,$F9,$72,$F0,$BF,$AA,$CB,$C3,$7F
	.dc.b	$C2,$B2,$EF,$8F,$6C,$AD,$F5,$F8,$6E,$77,$E0,$E4,$69,$E0,$AF,$85
	.dc.b	$A7,$C9,$D3,$E5,$7C,$24,$9F,$90,$FC,$DF,$84,$F9,$DF,$09,$F3,$FE
	.dc.b	$1A,$57,$CD,$8F,$A5,$F0,$72,$3E,$83,$F5,$7E,$17,$A4,$7D,$1C,$2E
	.dc.b	$BE,$5D,$19,$76,$AF,$B1,$97,$67,$EC,$65,$D5,$FB,$59,$5C,$39,$65
	.dc.b	$F8,$3F,$85,$FC,$27,$C2,$CB,$FC,$14,$7E,$1F,$39,$1F,$85,$7F,$15
	.dc.b	$F0,$BF,$8B,$F8,$59,$3F,$89,$7F,$1D,$F0,$BD,$63,$F1,$98,$75,$7F
	.dc.b	$19,$87,$47,$F1,$D8,$5C,$BD,$D8,$F4,$9F,$C9,$63,$67,$E1,$63,$F9
	.dc.b	$BF,$84,$95,$F9,$97,$F3,$DF,$0B,$F9,$FF,$85,$FD,$07,$C2,$CA,$FC
	.dc.b	$F4,$7E,$8F,$E1,$64,$7E,$85,$FF,$16,$FE,$7B,$FE,$3A,$CB,$E8,$63
	.dc.b	$DA,$BF,$48,$F6,$79,$9F,$A9,$C3,$B3,$FA,$9C,$3A,$BF,$AB,$C2,$DF
	.dc.b	$E3,$F0,$ED,$D7,$FA,$2C,$7B,$4F,$82,$FF,$69,$97,$AD,$EC,$43,$B1
	.dc.b	$E9,$21,$DC,$FF,$1B,$DF,$65,$D2,$8F,$F2,$04,$61,$B1,$9F,$84,$F8
	.dc.b	$7F,$0B,$F1,$3E,$16,$5F,$47,$E3,$FC,$37,$C8,$F8,$5F,$93,$F0,$D2
	.dc.b	$BC,$31,$C9,$F1,$E1,$F3,$8C,$2D,$3E,$26,$1D,$A3,$1B,$7F,$9B,$B4
	.dc.b	$4F,$82,$7A,$9E,$EC,$7C,$0F,$03,$D9,$8F,$DC,$F8,$4F,$BB,$F0,$B2
	.dc.b	$7E,$2C,$7E,$03,$E1,$3F,$03,$F0,$B2,$7E,$4C,$7E,$13,$E1,$A4,$7C
	.dc.b	$C7,$F0,$FF,$0B,$D6,$39,$90,$F1,$71,$7C,$FF,$86,$ED,$1F,$53,$0E
	.dc.b	$CE,$B6,$63,$DA,$31,$8D,$E0,$FC,$46,$5F,$96,$F8,$6F,$CB,$FC,$34
	.dc.b	$BF,$BF,$1F,$9A,$F8,$59,$1F,$82,$7F,$3B,$F0,$DF,$9E,$F8,$4F,$CF
	.dc.b	$FC,$34,$AF,$C3,$C7,$57,$99,$0F,$37,$C2,$74,$7D,$17,$73,$FC,$E6
	.dc.b	$1D,$67,$F2,$38,$75,$7D,$D8,$75,$7E,$06,$36,$FF,$49,$8F,$3E,$7F
	.dc.b	$C0,$65,$6F,$F2,$D9,$75,$7E,$46,$5D,$70,$FD,$66,$5F,$BA,$F8,$59
	.dc.b	$5F,$A2,$7F,$D5,$3F,$9F,$7F,$4E,$8F,$AA,$FF,$93,$DF,$D2,$B7,$90
	.dc.b	$D1,$F0,$BC,$BE,$12,$57,$FA,$92,$7F,$C7,$5A,$FF,$B1,$17,$C2,$F8
	.dc.b	$A7,$EF,$63,$73,$FE,$A1,$5E,$36,$FC,$37,$98,$95,$D7,$2E,$EF,$A3
	.dc.b	$2B,$5C,$B1,$F6,$FC,$37,$BB,$E1,$65,$F9,$63,$E0,$FC,$37,$C2,$F8
	.dc.b	$69,$3E,$98,$F8,$BF,$09,$F1,$BE,$16,$4F,$B6,$3A,$BE,$FC,$3A,$BF
	.dc.b	$2B,$0E,$AF,$C3,$C3,$AB,$F3,$B0,$EA,$FC,$7C,$3A,$BF,$4B,$0F,$AB
	.dc.b	$F0,$D2,$3E,$63,$F6,$3E,$17,$EC,$FC,$24,$9F,$A0,$FD,$CF,$85,$FB
	.dc.b	$BF,$0B,$F7,$BE,$16,$57,$D5,$8E,$4F,$42,$1F,$82,$BE,$8F,$E1,$2E
	.dc.b	$F0,$7D,$6C,$3A,$CF,$DF,$C3,$AB,$F8,$AC,$3A,$BF,$8C,$C7,$D4,$F5
	.dc.b	$3C,$E4,$3B,$1F,$89,$AE,$AF,$E5,$2B,$AD,$FE,$1A,$FF,$33,$F0,$BF
	.dc.b	$9A,$F8,$59,$7F,$91,$8F,$CE,$FC,$27,$E7,$BE,$16,$4F,$E5,$60,$FF
	.dc.b	$BE,$BF,$3F,$FF,$21,$B7,$E1,$72,$FF,$18,$8F,$F1,$A7,$67,$0E,$D4
	.dc.b	$70,$9F,$D4,$E3,$E2,$7F,$57,$97,$99,$F0,$1F,$A5,$CB,$C6,$FE,$C3
	.dc.b	$2E,$FA,$7F,$8C,$70,$E8,$DC,$F3,$9F,$86,$7C,$77,$FD,$69,$DC,$67
	.dc.b	$E1,$77,$39,$F8,$5B,$AF,$A1,$B7,$FE,$3D,$C7,$C1,$F0,$B8,$F8,$59
	.dc.b	$87,$C5,$8D,$FC,$24,$8F,$90,$F2,$F8,$5E,$79,$C9,$F9,$8F,$6F,$85
	.dc.b	$EF,$7D,$27,$E6,$61,$87,$99,$0F,$A1,$87,$57,$EB,$7C,$2F,$A0,$8F
	.dc.b	$D9,$F8,$4F,$5F,$C2,$FB,$3E,$16,$4F,$DD,$8F,$7F,$C2,$C8,$FC,$03
	.dc.b	$F0,$BE,$17,$9C,$7D,$FC,$2E,$FF,$4F,$1E,$B3,$F8,$7C,$7A,$BF,$85
	.dc.b	$C7,$A3,$F8,$7C,$6D,$FD,$BC,$7E,$7E,$1D,$38,$FE,$D6,$58,$C3,$CC
	.dc.b	$75,$21,$D4,$F3,$90,$EC,$7E,$66,$BE,$DF,$C2,$FD,$CF,$82,$93,$F9
	.dc.b	$D8,$FB,$FF,$05,$23,$F4,$0F,$E0,$BE,$13,$F0,$7F,$0B,$27,$F4,$8F
	.dc.b	$E1,$BE,$17,$AC,$7E,$9F,$0E,$AF,$E9,$70,$E8,$FE,$9F,$0B,$AF,$E6
	.dc.b	$F1,$E9,$3F,$AB,$C6,$CF,$E7,$F1,$FC,$AF,$89,$FE,$35,$EA,$FF,$69
	.dc.b	$DE,$0C,$FC,$1E,$BE,$16,$47,$EE,$B0,$FF,$4C,$5A,$79,$FC,$27,$4F
	.dc.b	$85,$EB,$F0,$93,$0E,$1C,$7F,$BB,$C1,$5D,$E7,$27,$F9,$1A,$3C,$84
	.dc.b	$CB,$C3,$1F,$E1,$C8,$F0,$3C,$33,$FF,$75,$E5,$E2,$8C,$4F,$66,$E3
	.dc.b	$E3,$79,$61,$F0,$BE,$16,$47,$07,$E2,$FC,$27,$C6,$F8,$5F,$8F,$F0
	.dc.b	$52,$BA,$C7,$CA,$F8,$49,$1E,$07,$E6,$FC,$1F,$48,$F4,$61,$78,$7E
	.dc.b	$4E,$3D,$67,$CF,$8F,$57,$DD,$8F,$47,$E0,$63,$6F,$E8,$E3,$CF,$83
	.dc.b	$E6,$65,$6F,$E5,$65,$D5,$F9,$19,$75,$C3,$EC,$E5,$CF,$0F,$A9,$95
	.dc.b	$BF,$A5,$97,$E1,$BE,$0E,$47,$CC,$7F,$13,$F0,$BF,$8A,$F8,$3F,$C5
	.dc.b	$FC,$24,$AF,$A3,$1F,$8E,$F8,$3F,$C7,$FC,$1C,$9F,$AB,$1C,$9E,$84
	.dc.b	$3E,$F5,$F5,$7E,$DD,$F5,$7F,$05,$7D,$5F,$BF,$7D,$5F,$C3,$5F,$E7
	.dc.b	$BE,$0A,$47,$E1,$1F,$D0,$FC,$1F,$E8,$BE,$12,$4F,$E2,$1F,$D2,$FC
	.dc.b	$27,$F3,$AF,$D0,$3F,$C4,$DC,$1F,$D4,$DF,$F1,$B1,$1E,$B5,$F9,$5C
	.dc.b	$2F,$17,$E9,$B1,$EB,$3F,$99,$C6,$CF,$EA,$B1,$ED,$87,$7F,$84,$97
	.dc.b	$F9,$DC,$BC,$2F,$FA,$A9,$FF,$21,$B9,$F8,$4F,$31,$E7,$F8,$4F,$47
	.dc.b	$C2,$49,$E4,$FA,$FE,$0F,$D9,$F0,$9E,$DF,$83,$95,$DE,$39,$BD,$48
	.dc.b	$7F,$A3,$31,$FF,$27,$BE,$3B,$EC,$E5,$EC,$F3,$3D,$55,$D9,$E5,$5D
	.dc.b	$9E,$98,$DD,$5A,$7C,$FC,$FA,$C6,$2B,$C0,$7B,$0F,$8D,$9F,$B1,$EE
	.dc.b	$43,$B9,$F2,$FE,$13,$EE,$11,$F9,$DF,$07,$F7,$FE,$12,$47,$D1,$7F
	.dc.b	$05,$F0,$7F,$83,$F0,$FF,$09,$F0,$92,$BE,$B4,$73,$7E,$36,$17,$2F
	.dc.b	$47,$8C,$E2,$67,$DA,$FE,$66,$76,$F7,$4C,$FC,$15,$F8,$3C,$FC,$0F
	.dc.b	$D3,$CF,$F2,$DF,$09,$23,$F1,$0F,$E6,$7E,$0F,$F3,$5F,$09,$27,$F1
	.dc.b	$8F,$E7,$7E,$13,$F3,$DF,$09,$27,$F2,$0F,$E8,$7E,$0F,$A4,$7E,$07
	.dc.b	$0B,$AF,$9B,$1E,$93,$F8,$5C,$6C,$FA,$B1,$E9,$3F,$89,$C6,$CF,$BB
	.dc.b	$1F,$D7,$7C,$24,$AF,$D0,$3F,$E2,$AF,$D0,$7F,$C5,$F0,$E2,$47,$F8
	.dc.b	$D1,$BF,$03,$B3,$87,$C2,$F1,$F8,$5E,$5F,$05,$30,$FE,$91,$82,$7C
	.dc.b	$11,$F9,$DC,$6E,$7F,$62,$BC,$39,$78,$31,$ED,$7C,$B2,$EE,$F4,$C7
	.dc.b	$B3,$FE,$F2,$B6,$5E,$38,$F0,$FC,$47,$6E,$BF,$E1,$9A,$7B,$E3,$9C
	.dc.b	$F1,$C7,$BB,$CF,$2E,$4F,$5C,$3B,$E9,$E7,$CF,$E4,$7C,$1F,$C9,$F8
	.dc.b	$39,$7F,$02,$3E,$67,$C2,$FC,$DF,$85,$93,$F1,$23,$E8,$7C,$0C,$8F
	.dc.b	$8E,$FD,$3F,$82,$EB,$1C,$C8,$7B,$FE,$13,$AB,$F3,$AF,$AB,$F0,$EF
	.dc.b	$A3,$F1,$6E,$E7,$F3,$F0,$FC,$07,$C2,$FE,$07,$E1,$65,$FD,$88,$FC
	.dc.b	$27,$C2,$FE,$17,$E1,$24,$FD,$C8,$FC,$47,$C2,$FE,$27,$E1,$64,$FE
	.dc.b	$02,$3A,$BC,$C8,$7D,$7F,$85,$EA,$FE,$1A,$FA,$BF,$6E,$FA,$BF,$8A
	.dc.b	$BE,$AF,$DF,$BF,$CC,$7C,$2C,$8F,$C7,$BF,$9B,$F8,$5F,$F4,$2F,$E8
	.dc.b	$5F,$E3,$12,$4E,$3E,$17,$3F,$0B,$AF,$86,$93,$FE,$1F,$AF,$F1,$0B
	.dc.b	$F8,$DC,$3B,$34,$FC,$EE,$1D,$98,$71,$3B,$1C,$EF,$AA,$E1,$FE,$2F
	.dc.b	$6C,$3C,$73,$EA,$21,$DC,$FC,$F5,$F6,$79,$77,$7F,$AE,$6B,$DB,$1E
	.dc.b	$9C,$B9,$BA,$3F,$79,$1D,$DF,$75,$EB,$0F,$1F,$C2,$FC,$82,$4F,$C3
	.dc.b	$F8,$6F,$97,$F0,$DF,$33,$E1,$A4,$FC,$78,$F9,$FF,$0B,$F4,$3E,$1A
	.dc.b	$4F,$CB,$8F,$A7,$F0,$DF,$53,$E1,$A4,$FC,$F8,$FA,$FF,$0D,$F6,$3E
	.dc.b	$1A,$4F,$D3,$8F,$B7,$F0,$BF,$73,$E1,$A4,$FD,$78,$FB,$FF,$0B,$23
	.dc.b	$ED,$3F,$82,$F8,$6F,$C1,$FC,$34,$9F,$BC,$FE,$1B,$E1,$BF,$0F,$F0
	.dc.b	$D2,$7F,$04,$FE,$2B,$E1,$BF,$D9,$FA,$7F,$81,$BD,$23,$49,$F0,$DA
	.dc.b	$F8,$3D,$FC,$1C,$C7,$FC,$D5,$5F,$E0,$C7,$EA,$5F,$2F,$86,$EA,$48
	.dc.b	$DF,$C2,$78,$3B,$E4,$72,$7C,$79,$77,$8E,$B8,$F7,$7A,$E3,$DF,$B3
	.dc.b	$FB,$2A,$C7,$B4,$E9,$F6,$E3,$DA,$38,$E3,$6F,$8E,$3D,$E7,$AE,$3D
	.dc.b	$E7,$AE,$7D,$1E,$F8,$DC,$17,$FD,$37,$26,$5F,$33,$E1,$64,$FC,$67
	.dc.b	$E7,$FC,$37,$D0,$F8,$6F,$A3,$F0,$D2,$BE,$5C,$7D,$4F,$86,$91,$F3
	.dc.b	$9F,$AF,$F0,$BD,$23,$E1,$E1,$75,$F8,$78,$F5,$9F,$AB,$8F,$57,$E4
	.dc.b	$E3,$D1,$F9,$78,$DB,$F9,$78,$F3,$D3,$DB,$C3,$F8,$4C,$F1,$87,$BF
	.dc.b	$3E,$CF,$D4,$CF,$B6,$3F,$53,$3E,$98,$F2,$D5,$A6,$30,$F9,$1A,$7E
	.dc.b	$4B,$E1,$3F,$27,$F0,$B2,$7F,$1B,$1F,$96,$F8,$5F,$CB,$FC,$2C,$9F
	.dc.b	$C8,$C7,$E6,$BE,$17,$F3,$7F,$0B,$27,$F2,$B1,$D5,$FC,$BE,$1D,$5F
	.dc.b	$C5,$E1,$D5,$FC,$DE,$1D,$5F,$C7,$E1,$D5,$FC,$FE,$1D,$5F,$C9,$E1
	.dc.b	$D5,$FC,$AE,$1D,$76,$FC,$AE,$3D,$67,$F3,$38,$F5,$9F,$CC,$E3,$D6
	.dc.b	$7F,$3B,$8F,$59,$FC,$EE,$3F,$B6,$F8,$59,$5F,$B0,$7F,$73,$F0,$BF
	.dc.b	$BA,$F8,$4F,$C9,$CF,$EC,$BF,$43,$FF,$08,$9F,$E1,$2E,$04,$EB,$1F
	.dc.b	$F2,$6C,$FF,$88,$C7,$C2,$67,$8B,$F0,$31,$8C,$3F,$DD,$0F,$F8,$A7
	.dc.b	$06,$5E,$49,$FD,$AB,$E4,$7F,$82,$E3,$CD,$FF,$19,$57,$93,$E1,$7D
	.dc.b	$11,$D3,$E1,$75,$D2,$AF,$04,$FA,$F3,$CC,$78,$71,$F0,$69,$FC,$09
	.dc.b	$F0,$57,$F0,$6C,$B7,$94,$8F,$19,$DC,$F5,$E1,$D0,$EB,$97,$6F,$84
	.dc.b	$93,$F0,$1F,$0F,$C2,$F8,$BE,$13,$C7,$F0,$B2,$BE,$2C,$79,$BE,$0F
	.dc.b	$CF,$F0,$92,$7E,$4C,$74,$7E,$B6,$17,$6F,$A7,$8F,$59,$F9,$F8,$F5
	.dc.b	$7E,$E6,$3D,$5F,$A7,$8F,$57,$F0,$18,$F3,$E2,$FB,$39,$59,$FA,$38
	.dc.b	$F5,$7F,$0D,$8F,$5C,$BF,$07,$97,$59,$FC,$56,$5D,$67,$F0,$F9,$7D
	.dc.b	$2F,$84,$95,$F8,$47,$EA,$FC,$27,$D6,$F8,$5F,$AF,$F0,$92,$BF,$13
	.dc.b	$1F,$6B,$E1,$7E,$DF,$C1,$C9,$FC,$6C,$74,$7F,$35,$85,$CB,$F2,$F8
	.dc.b	$F5,$9F,$C9,$E3,$D5,$FD,$06,$3D,$5F,$CB,$E3,$D5,$FD,$26,$3C,$F8
	.dc.b	$3F,$3B,$95,$9F,$CA,$E3,$D5,$FD,$56,$3D,$72,$FD,$3E,$5D,$67,$F5
	.dc.b	$D9,$75,$9F,$D5,$E5,$F9,$6F,$84,$95,$FA,$87,$F3,$3F,$07,$F9,$AF
	.dc.b	$84,$FC,$DF,$C2,$CA,$FD,$6C,$7E,$7B,$E1,$64,$7E,$C1,$FD,$0F,$C2
	.dc.b	$74,$8F,$DF,$61,$72,$FD,$DE,$3D,$67,$F6,$F8,$F5,$7F,$82,$63,$D1
	.dc.b	$FE,$0D,$8D,$BF,$E0,$78,$F3,$E0,$FE,$03,$95,$BF,$DC,$E5,$D5,$FE
	.dc.b	$1D,$97,$5C,$3F,$85,$E5,$CF,$0F,$E1,$39,$5B,$FE,$03,$97,$EF,$BE
	.dc.b	$DF,$F8,$BC,$7F,$82,$1B,$87,$FB,$81,$B7,$FD,$49,$7C,$2F,$1F,$86
	.dc.b	$E5,$F0,$B3,$0F,$EE,$78,$71,$9E,$DF,$0B,$27,$38,$F3,$8F,$17,$C2
	.dc.b	$F8,$E6,$1A,$3F,$D9,$19,$78,$9B,$97,$85,$D7,$86,$9F,$F2,$94,$7A
	.dc.b	$F3,$D3,$ED,$C3,$C4,$D3,$8C,$B0,$F0,$C7,$C3,$2E,$7F,$0F,$1E,$6F
	.dc.b	$C5,$C3,$E0,$4F,$C9,$CF,$AC,$FC,$9C,$6E,$7F,$27,$1F,$A1,$F0,$93
	.dc.b	$0F,$0B,$F4,$FE,$17,$EA,$7C,$27,$D5,$F8,$59,$5E,$68,$FB,$1F,$09
	.dc.b	$23,$D2,$FD,$BF,$85,$E9,$1F,$6F,$0B,$A7,$DB,$C7,$AC,$FC,$0C,$7A
	.dc.b	$BF,$81,$C7,$A3,$F8,$3C,$6D,$FE,$0F,$1E,$79,$FE,$0F,$2B,$7F,$5B
	.dc.b	$2E,$AF,$E2,$F2,$EB,$87,$E2,$F2,$E7,$87,$E2,$F2,$B7,$F7,$B2,$FC
	.dc.b	$A7,$C2,$FE,$57,$CB,$93,$F5,$A3,$F3,$1F,$09,$23,$EC,$BF,$9B,$F8
	.dc.b	$5F,$CE,$7C,$2F,$E7,$7E,$16,$57,$DE,$8E,$AF,$E0,$30,$EA,$FE,$87
	.dc.b	$0E,$8F,$E8,$F0,$BA,$FE,$8F,$1E,$B3,$F8,$8C,$7A,$BF,$A9,$C7,$AB
	.dc.b	$FA,$BC,$7A,$DF,$EA,$F1,$E7,$A7,$EA,$F2,$B7,$F9,$EC,$BA,$BF,$B3
	.dc.b	$CB,$AE,$1F,$B3,$CB,$F7,$1F,$0B,$2B,$F3,$2F,$EE,$FE,$17,$F7,$9F
	.dc.b	$0B,$FB,$DF,$85,$95,$F9,$E8,$FE,$FB,$F6,$FF,$E8,$F2,$6B,$8F,$86
	.dc.b	$E5,$5D,$08,$7F,$05,$C3,$AB,$FA,$8C,$3A,$BF,$C2,$70,$EB,$72,$3F
	.dc.b	$6B,$87,$7C,$3A,$C7,$F0,$FC,$7D,$4F,$53,$CC,$75,$21,$D4,$E7,$5D
	.dc.b	$5F,$F4,$FD,$75,$7B,$FC,$2F,$B8,$8F,$8B,$E1,$BE,$0F,$C3,$7C,$2F
	.dc.b	$86,$93,$E6,$8F,$8B,$F0,$DF,$1B,$E1,$A4,$FA,$A3,$9B,$E0,$C2,$F0
	.dc.b	$B3,$FF,$3B,$E6,$CF,$BD,$FC,$2C,$FB,$BE,$8C,$FB,$BF,$1B,$3E,$EF
	.dc.b	$B3,$3E,$B8,$F4,$CE,$CB,$C3,$67,$DE,$3D,$47,$72,$1D,$CF,$51,$DC
	.dc.b	$87,$73,$EB,$FC,$2F,$E1,$88,$FD,$AF,$86,$FC,$4F,$C3,$7E,$2B,$E1
	.dc.b	$A4,$FD,$E8,$FC,$6F,$C2,$C8,$FC,$0B,$F9,$0F,$86,$E7,$1F,$63,$0B
	.dc.b	$C5,$F2,$63,$D9,$93,$2E,$F1,$F8,$AC,$BB,$BF,$80,$CB,$B3,$F8,$2C
	.dc.b	$AE,$1C,$F3,$F8,$B9,$F6,$D7,$EA,$E9,$70,$EF,$9F,$67,$F1,$79,$F6
	.dc.b	$C7,$96,$9F,$EA,$89,$F5,$90,$F0,$1F,$9E,$C7,$F5,$BF,$0B,$23,$F4
	.dc.b	$2F,$EC,$3E,$1B,$FC,$E5,$0F,$8E,$7B,$B8,$5E,$E5,$31,$1F,$B8,$F8
	.dc.b	$69,$5F,$AC,$7F,$77,$F0,$FD,$E3,$F4,$78,$DD,$3D,$39,$72,$9F,$ED
	.dc.b	$87,$87,$1E,$CE,$32,$F1,$BD,$48,$7E,$BB,$1E,$EF,$B0,$87,$80,$FD
	.dc.b	$FD,$77,$57,$F7,$5F,$01,$DE,$F5,$1F,$3A,$BD,$B3,$DC,$FE,$13,$8F
	.dc.b	$F1,$8F,$85,$FE,$33,$F0,$F2,$7F,$86,$C7,$F1,$CF,$86,$FF,$4C,$FB
	.dc.b	$4F,$86,$A6,$6B,$FC,$0F,$0F,$85,$DF,$C3,$49,$CC,$77,$7F,$8C,$E3
	.dc.b	$DD,$FE,$1F,$8F,$37,$DB,$CF,$FE,$F6,$C3,$C1,$1D,$B1,$D6,$BF,$93
	.dc.b	$9F,$04,$F8,$B2,$F0,$3C,$8F,$DE,$46,$9E,$E7,$88,$86,$0F,$E4,$39
	.dc.b	$FB,$9F,$01,$FC,$C1,$F8,$79,$F6,$BF,$15,$E2,$BE,$1F,$C3,$7C,$8E
	.dc.b	$39,$18,$3E,$3F,$07,$CC,$F8,$69,$18,$3E,$5F,$C2,$FD,$02,$3F,$3B
	.dc.b	$E1,$BE,$9E,$1D,$23,$DD,$D8,$DB,$F1,$F1,$85,$3F,$6B,$1E,$F3,$9C
	.dc.b	$70,$F8,$08,$71,$C7,$F0,$38,$F8,$23,$EF,$E5,$E1,$D3,$E9,$63,$CA
	.dc.b	$7A,$1F,$84,$C3,$13,$F5,$F0,$F5,$CF,$53,$D8,$7E,$2B,$E1,$BF,$1E
	.dc.b	$47,$F1,$BF,$0D,$F9,$2F,$85,$FC,$9F,$C3,$49,$FC,$8C,$7E,$5B,$6F
	.dc.b	$CB,$FC,$34,$9F,$CA,$C7,$37,$E2,$61,$72,$FC,$56,$3D,$67,$F3,$78
	.dc.b	$F5,$7E,$56,$3D,$5F,$CF,$E3,$D5,$F9,$D8,$F3,$E5,$FC,$C6,$56,$7F
	.dc.b	$23,$8F,$57,$EA,$63,$D7,$2F,$CD,$65,$D6,$7E,$C6,$5D,$67,$F3,$D9
	.dc.b	$7E,$CB,$E1,$A5,$7E,$C1,$FD,$AF,$C3,$7E,$DB,$E1,$64,$FE,$D1,$FD
	.dc.b	$CF,$C3,$7E,$EB,$E1,$A4,$FE,$E1,$FD,$EF,$C3,$74,$8F,$C4,$61,$72
	.dc.b	$FD,$76,$3D,$27,$F1,$98,$D9,$FD,$96,$3D,$27,$F2,$18,$D9,$FD,$B6
	.dc.b	$3C,$F8,$3F,$79,$95,$BF,$DA,$E5,$CE,$7F,$80,$65,$67,$F7,$39,$73
	.dc.b	$9F,$E0,$99,$59,$FD,$EE,$27,$FD,$33,$FA,$27,$FA,$77,$A0,$D2,$7C
	.dc.b	$37,$F4,$8F,$90,$78,$17,$BA,$61,$FC,$F2,$72,$F4,$F8,$6E,$CF,$EC
	.dc.b	$30,$46,$38,$7F,$B3,$46,$27,$C7,$F0,$F9,$DB,$CD,$A7,$93,$E1,$F1
	.dc.b	$5F,$E1,$1B,$46,$27,$53,$EC,$F8,$7F,$47,$95,$E9,$F2,$A4,$F9,$E3
	.dc.b	$86,$0F,$87,$BF,$0F,$2B,$E2,$C6,$BE,$1B,$7F,$0F,$27,$E4,$C7,$C5
	.dc.b	$F5,$FE,$37,$A7,$27,$E2,$47,$6F,$87,$EF,$F0,$F2,$7E,$8C,$78,$BE
	.dc.b	$1B,$C7,$F0,$F2,$7E,$AC,$7D,$1F,$80,$FA,$5E,$DC,$9F,$A1,$1E,$AF
	.dc.b	$87,$F5,$FC,$3C,$9F,$BB,$1E,$EF,$86,$91,$F8,$07,$E0,$FC,$3F,$F8
	.dc.b	$13,$13,$7E,$12,$57,$F7,$89,$CB,$F8,$88,$F9,$3F,$0F,$23,$07,$E3
	.dc.b	$3E,$1B,$E6,$91,$FC,$7F,$C3,$F7,$E0,$F0,$61,$27,$B4,$7D,$BF,$D1
	.dc.b	$7F,$BC,$78,$1F,$E2,$B5,$BE,$4D,$77,$EF,$86,$61,$FE,$F4,$A4,$73
	.dc.b	$F8,$8F,$77,$77,$BF,$AA,$5F,$B6,$3C,$1F,$0F,$E1,$F8,$89,$3C,$E3
	.dc.b	$AB,$FD,$2B,$1E,$AF,$7C,$7B,$3C,$CF,$91,$87,$67,$FB,$70,$DB,$FE
	.dc.b	$00,$3D,$4F,$46,$1E,$07,$B6,$1E,$0D,$7E,$16,$17,$4F,$D2,$E5,$FE
	.dc.b	$71,$BE,$B8,$F4,$6E,$5E,$1C,$3F,$C4,$53,$97,$E6,$63,$DD,$F9,$D9
	.dc.b	$7C,$2B,$F9,$5F,$11,$F4,$FA,$A4,$60,$F7,$1D,$8F,$79,$8A,$FB,$1D
	.dc.b	$7F,$6F,$E1,$A4,$FD,$58,$FB,$D7,$D2,$3E,$65,$E1,$E5,$AF,$F0,$AC
	.dc.b	$FF,$6D,$1D,$8F,$7E,$FF,$87,$C3,$AC,$7D,$5C,$70,$F9,$CE,$A7,$E1
	.dc.b	$70,$EC,$F2,$3A,$65,$F1,$B5,$FC,$3C,$79,$CF,$C6,$57,$69,$F6,$C7
	.dc.b	$E3,$FE,$23,$F3,$04,$7F,$25,$F0,$FC,$9F,$8C,$49,$FA,$79,$7E,$7F
	.dc.b	$E2,$3F,$41,$E2,$4A,$FC,$D4,$73,$7F,$21,$85,$BF,$D2,$61,$D5,$F4
	.dc.b	$4F,$AB,$6F,$91,$8F,$58,$FD,$26,$3D,$5F,$CD,$63,$CE,$BF,$17,$8D
	.dc.b	$9F,$F5,$F7,$13,$2E,$D5,$F4,$F4,$FC,$63,$DA,$BC,$E7,$62,$1D,$8F
	.dc.b	$DA,$78,$BF,$BD,$F8,$69,$18,$3E,$F9,$FD,$57,$87,$EF,$BF,$1E,$BB
	.dc.b	$3F,$B1,$F8,$6F,$E0,$DF,$0D,$23,$F8,$04,$7F,$09,$C3,$AC,$7E,$DF
	.dc.b	$0C,$3D,$8F,$D9,$63,$EA,$7C,$1B,$FF,$11,$BE,$91,$FB,$FC,$30,$F9
	.dc.b	$CE,$87,$F0,$FB,$EA,$FE,$2C,$F4,$1E,$BB,$F9,$2F,$CC,$AE,$BC,$1F
	.dc.b	$C7,$6F,$AC,$FF,$0D,$F8,$6F,$E4,$44,$7F,$8E,$7A,$1C,$9F,$49,$F7
	.dc.b	$CF,$49,$FC,$87,$D0,$FE,$C9,$FA,$37,$FA,$12,$3F,$CC,$2D,$C3,$F9
	.dc.b	$64,$F5,$9F,$01,$FC,$A6,$FF,$99,$5F,$7D,$FE,$9D,$FE,$3A,$BD,$CF
	.dc.b	$27,$FC,$18,$FF,$B8,$70,$ED,$BF,$F7,$A6,$CF,$E3,$31,$F0,$57,$F3
	.dc.b	$DC,$6D,$FF,$1C,$C7,$C1,$A7,$D5,$C3,$F9,$46,$DF,$D1,$71,$F6,$FC
	.dc.b	$3C,$9F,$80,$78,$B0,$F0,$7C,$3F,$C2,$24,$64,$79,$ED,$5F,$0F,$E0
	.dc.b	$9F,$98,$78,$0E,$13,$BF,$84,$F9,$BF,$0B,$27,$37,$FE,$8A,$6E,$1E
	.dc.b	$07,$E8,$BF,$53,$0E,$D1,$F8,$63,$B1,$F5,$AF,$A3,$F6,$2F,$ED,$C7
	.dc.b	$3E,$0F,$9D,$77,$2F,$9D,$87,$49,$FA,$C7,$F0,$EC,$3C,$B7,$D2,$3E
	.dc.b	$B6,$1D,$27,$D3,$F0,$FF,$88,$23,$EC,$F8,$6E,$98,$FF,$01,$8F,$DB
	.dc.b	$E9,$F2,$7E,$1B,$F1,$FF,$0F,$27,$E1,$C7,$27,$F2,$38,$5C,$BF,$0D
	.dc.b	$8F,$49,$FD,$31,$D3,$65,$F3,$8F,$CC,$E1,$F9,$C8,$E5,$97,$E3,$71
	.dc.b	$B5,$F7,$2D,$87,$58,$FD,$39,$F9,$8A,$FC,$A5,$F5,$C7,$F3,$18,$F5
	.dc.b	$AF,$B7,$F0,$FF,$AA,$23,$F7,$BE,$1F,$A7,$3F,$E9,$31,$FE,$2B,$87
	.dc.b	$E8,$7E,$1B,$F6,$3F,$0F,$30,$FC,$2C,$72,$7F,$67,$8D,$AF,$D4,$63
	.dc.b	$D3,$9F,$F8,$3F,$07,$E5,$70,$E5,$87,$EE,$B1,$FD,$EC,$71,$CB,$F5
	.dc.b	$D9,$5A,$F8,$58,$F3,$7E,$65,$72,$CB,$F6,$D8,$72,$9F,$CC,$FC,$37
	.dc.b	$F0,$82,$3F,$9C,$F8,$7E,$4F,$C9,$3E,$77,$C3,$FF,$0E,$F8,$79,$1F
	.dc.b	$A2,$8E,$2F,$F1,$0A,$B8,$7F,$05,$BE,$6F,$E5,$67,$F8,$AD,$72,$7F
	.dc.b	$8B,$DF,$F1,$A8,$E3,$A7,$F0,$BC,$2D,$7D,$FC,$39,$CF,$D2,$3F,$88
	.dc.b	$57,$3C,$3F,$8A,$61,$CE,$FF,$6F,$F0,$FF,$C9,$88,$FE,$EB,$E1,$FF
	.dc.b	$D1,$EB,$E4,$DF,$0F,$FE,$67,$88,$7F,$0B,$34,$FF,$30,$F8,$8F,$E6
	.dc.b	$2C,$8F,$E0,$B1,$D9,$FE,$67,$77,$1F,$E5,$18,$78,$30,$EF,$5F,$E7
	.dc.b	$13,$B9,$FE,$4A,$72,$FF,$3C,$BE,$CF,$F3,$EC,$3C,$4F,$B8,$87,$80
	.dc.b	$F7,$1E,$99,$E7,$7F,$E4,$68,$EE,$B3,$FF,$5D,$CE,$9F,$E7,$97,$E0
	.dc.b	$9F,$E8,$17,$E0,$B4,$71,$7F,$97,$F9,$7F,$EC,$F1,$F1,$1F,$EB,$96
	.dc.b	$C7,$F2,$E7,$FD,$92,$FB,$EB,$C7,$87,$F9,$0E,$3E,$34,$71,$F8,$7E
	.dc.b	$5F,$0F,$27,$94,$75,$7B,$61,$1E,$C7,$23,$E7,$10,$F0,$1C,$8F,$24
	.dc.b	$F5,$7D,$B7,$FE,$D6,$C3,$FD,$B5,$18,$3D,$E7,$73,$1C,$C8,$F7,$C6
	.dc.b	$1F,$7E,$5F,$1E,$3F,$16,$78,$DF,$A8,$FD,$7C,$BE,$03,$E4,$21,$CF
	.dc.b	$2F,$99,$F0,$FF,$37,$C5,$93,$F3,$63,$E0,$3F,$E8,$83,$E6,$83,$C2
	.dc.b	$79,$7A,$7E,$B7,$C3,$48,$FA,$D1,$D9,$FB,$15,$D9,$F6,$57,$6C,$7D
	.dc.b	$31,$DC,$F5,$1C,$8F,$C0,$5F,$57,$C0,$79,$CE,$A4,$3A,$9C,$8E,$84
	.dc.b	$8F,$95,$C1,$F9,$78,$FC,$8E,$1D,$67,$F2,$9A,$7D,$3B,$FC,$94,$FE
	.dc.b	$4F,$E1,$7F,$29,$F0,$D3,$0F,$CA,$47,$AF,$7F,$C9,$47,$D5,$7F,$4B
	.dc.b	$B7,$E6,$FE,$1B,$F3,$9E,$EC,$AF,$CE,$47,$57,$F3,$D8,$75,$7F,$05
	.dc.b	$87,$57,$FC,$C4,$7D,$43,$E9,$E9,$E8,$E1,$5F,$59,$FC,$56,$1E,$17
	.dc.b	$CE,$76,$21,$D8,$F9,$15,$F5,$09,$97,$AA,$3C,$E7,$EC,$E7,$B4,$F8
	.dc.b	$6B,$F7,$1D,$F2,$BF,$70,$FE,$EF,$E1,$BD,$51,$FE,$DF,$3D,$5C,$1F
	.dc.b	$4F,$83,$F4,$5F,$0B,$FC,$07,$E1,$A5,$7F,$01,$8E,$8F,$E9,$70,$B9
	.dc.b	$7C,$1C,$7A,$C7,$EA,$34,$FA,$75,$FB,$83,$F5,$97,$D1,$FD,$76,$1F
	.dc.b	$C3,$E3,$9C,$FE,$E6,$AD,$7F,$16,$BE,$AF,$A3,$83,$83,$FC,$57,$0E
	.dc.b	$98,$FD,$3C,$7F,$8D,$7C,$34,$AF,$E3,$4F,$F1,$DF,$86,$E7,$1F,$87
	.dc.b	$3F,$5D,$AF,$C3,$E0,$7C,$37,$F2,$3F,$86,$95,$FC,$8E,$3A,$3F,$C1
	.dc.b	$F0,$B9,$7E,$0F,$1E,$B3,$FA,$1D,$FE,$D3,$FC,$36,$F9,$BF,$C3,$F0
	.dc.b	$FE,$61,$1C,$A7,$F8,$D6,$16,$BF,$9B,$DF,$4C,$3F,$0B,$1E,$8D,$78
	.dc.b	$BF,$CD,$F1,$EB,$3F,$95,$C7,$F9,$E7,$C3,$4A,$FE,$78,$FF,$3F,$F8
	.dc.b	$6E,$99,$FD,$8B,$FD,$BE,$3F,$C4,$6F,$F9,$3F,$E8,$FF,$C8,$E7,$F6
	.dc.b	$EB,$E1,$F5,$18,$6D,$FF,$19,$E4,$57,$18,$FE,$6D,$97,$8F,$0F,$E6
	.dc.b	$0F,$8D,$FD,$1C,$78,$C8,$FB,$CF,$F4,$0C,$F3,$7F,$D1,$2F,$8D,$CD
	.dc.b	$F8,$9C,$CF,$7C,$3F,$71,$76,$3F,$A4,$65,$F0,$AB,$F7,$E7,$88,$E2
	.dc.b	$74,$21,$F1,$23,$B6,$3E,$EF,$86,$B9,$1A,$63,$E1,$FB,$5F,$EC,$CE
	.dc.b	$D5,$F1,$BE,$1F,$8F,$C3,$C9,$F7,$C7,$27,$E0,$DD,$7E,$AD,$F4,$E8
	.dc.b	$FD,$94,$7C,$73,$EC,$D7,$37,$ED,$DF,$9A,$39,$6B,$F7,$30,$B9,$FD
	.dc.b	$9C,$D9,$75,$C7,$F9,$26,$7E,$77,$E9,$E5,$D7,$1F,$C2,$E5,$F0,$BE
	.dc.b	$1E,$57,$D8,$7E,$2F,$C3,$73,$8F,$83,$93,$E1,$FE,$4F,$C3,$C9,$FB
	.dc.b	$F1,$C9,$FC,$9D,$DC,$7F,$27,$87,$39,$FE,$21,$1F,$96,$AE,$51,$F9
	.dc.b	$AB,$FA,$B1,$C7,$5F,$76,$16,$BF,$17,$87,$3C,$3F,$47,$1E,$23,$F2
	.dc.b	$38,$74,$8E,$D1,$F9,$4F,$86,$FC,$09,$1F,$CB,$FC,$2F,$F9,$2F,$CF
	.dc.b	$FF,$26,$79,$F2,$7F,$C8,$11,$F8,$8F,$87,$FF,$28,$7A,$12,$7F,$3F
	.dc.b	$1F,$8C,$EE,$91,$FA,$27,$F1,$FF,$0D,$FE,$57,$F5,$FF,$CB,$1E,$BC
	.dc.b	$AF,$F2,$64,$7F,$96,$FD,$79,$1F,$E4,$F7,$FC,$BD,$EB,$FE,$63,$E1
	.dc.b	$BF,$33,$F0,$D2,$BF,$5F,$1F,$E6,$3F,$80,$91,$FE,$57,$7F,$CC,$DF
	.dc.b	$01,$FE,$67,$F8,$09,$3F,$E5,$B7,$FC,$D5,$F0,$1D,$63,$F7,$38,$75
	.dc.b	$7F,$86,$E1,$D1,$FF,$36,$61,$79,$7F,$C1,$38,$F4,$9F,$F3,$86,$36
	.dc.b	$7F,$C1,$B8,$FE,$C3,$E1,$A5,$7F,$06,$7F,$67,$F0,$DF,$E8,$AF,$95
	.dc.b	$45,$89,$3F,$E8,$99,$CB,$7E,$1F,$81,$1D,$7C,$3F,$EF,$A2,$47,$F1
	.dc.b	$47,$F8,$0C,$77,$E9,$F0,$73,$4A,$ED,$1E,$3E,$69,$1E,$17,$CD,$CD
	.dc.b	$E7,$F1,$24,$F9,$5F,$E1,$FF,$A4,$3E,$23,$D9,$E2,$FB,$7C,$19,$7E
	.dc.b	$B8,$F8,$1E,$0C,$8F,$73,$F0,$FC,$3F,$F5,$6F,$CB,$BC,$BB,$E5,$CB
	.dc.b	$FF,$56,$4E,$5B,$E8,$7C,$BF,$1E,$46,$0F,$95,$E6,$FC,$F2,$3F,$37
	.dc.b	$CC,$F4,$FC,$4F,$AB,$BE,$4F,$F9,$7F,$B7,$FC,$87,$C4,$AF,$F6,$0F
	.dc.b	$D0,$7F,$1C,$A6,$8D,$3D,$26,$A3,$E2,$F8,$11,$E9,$AF,$F5,$3C,$53
	.dc.b	$F2,$BC,$0F,$1D,$68,$79,$CA,$6B,$E6,$9F,$F4,$4C,$79,$5F,$F6,$86
	.dc.b	$7E,$5C,$FC,$91,$F0,$CF,$F4,$90,$8F,$E6,$B7,$D7,$1F,$34,$F4,$7C
	.dc.b	$8F,$8E,$3C,$5A,$78,$75,$F8,$B9,$7A,$2B,$AE,$1E,$1D,$7E,$1C,$79
	.dc.b	$0E,$B5,$F3,$31,$BA,$FC,$3D,$7F,$17,$E0,$FE,$33,$E2,$66,$1E,$1D
	.dc.b	$D3,$F3,$63,$D9,$8F,$FA,$97,$7F,$7C,$78,$8F,$CA,$FC,$4C,$79,$3F
	.dc.b	$98,$F2,$3B,$C7,$5C,$7B,$BF,$33,$0E,$CF,$C8,$BE,$B8,$77,$D7,$2F
	.dc.b	$E1,$B1,$ED,$5F,$51,$EC,$FA,$6F,$AC,$FE,$33,$0F,$0C,$F9,$CF,$5E
	.dc.b	$9F,$E1,$FC,$7B,$BF,$E2,$1D,$BE,$BE,$8C,$BB,$E3,$E1,$C3,$E3,$69
	.dc.b	$FB,$4F,$27,$F6,$BE,$4C,$9F,$95,$1E,$8C,$BB,$6D,$F5,$A7,$C4,$7E
	.dc.b	$33,$C9,$FD,$E9,$1F,$A5,$F1,$3D,$5F,$A9,$87,$57,$F2,$78,$75,$7F
	.dc.b	$D9,$1B,$7A,$9B,$86,$79,$7F,$39,$8F,$59,$F0,$1D,$8F,$39,$D8,$87
	.dc.b	$63,$F1,$87,$C8,$3A,$98,$39,$93,$4F,$DD,$C7,$F1,$8C,$BF,$8C,$FC
	.dc.b	$4F,$F1,$AF,$36,$5F,$E3,$E3,$D4,$FE,$FB,$1F,$8D,$87,$FB,$CF,$87
	.dc.b	$E9,$F9,$DF,$C8,$89,$5F,$97,$F3,$BA,$BF,$99,$C7,$AB,$FB,$8C,$7A
	.dc.b	$EB,$F1,$63,$B1,$E9,$C7,$F8,$8E,$1D,$23,$F7,$EF,$F3,$0C,$3A,$C7
	.dc.b	$F0,$2C,$3A,$D7,$F1,$8C,$3A,$E5,$E8,$E0,$FC,$55,$F3,$AE,$AF,$EB
	.dc.b	$71,$E8,$FF,$2E,$F6,$BF,$99,$FB,$12,$30,$79,$6F,$F1,$2F,$D1,$3F
	.dc.b	$86,$1F,$CE,$BD,$78,$FF,$38,$7F,$9E,$FB,$1F,$CF,$AF,$A4,$FE,$B6
	.dc.b	$F0,$F5,$3E,$D9,$FB,$1C,$BE,$3D,$F4,$8F,$DA,$3D,$1F,$29,$CC,$FE
	.dc.b	$8F,$3D,$1F,$CE,$E5,$FB,$97,$D1,$BF,$F2,$7A,$E9,$3F,$C2,$DE,$93
	.dc.b	$FC,$BB,$E2,$7F,$AF,$FC,$4C,$8F,$E2,$91,$F3,$9E,$67,$F6,$4F,$89
	.dc.b	$8F,$F1,$97,$FB,$3F,$B7,$FD,$A2,$79,$4F,$F3,$AA,$C3,$F0,$4F,$C1
	.dc.b	$9F,$04,$F8,$9C,$DF,$D0,$EF,$AC,$7F,$42,$7A,$BE,$43,$F9,$44,$F6
	.dc.b	$C7,$FA,$DD,$F9,$E7,$A1,$F1,$37,$FE,$A9,$7D,$2B,$FB,$0B,$D2,$50
	.dc.b	$9F,$F5,$67,$E9,$3F,$D2,$87,$C5,$7F,$A4,$9B,$5F,$D2,$5F,$ED,$1A
	.dc.b	$FF,$45,$8F,$8E,$67,$E2,$78,$11,$E1,$F1,$5C,$DE,$97,$1E,$87,$AF
	.dc.b	$1F,$B8,$FA,$CF,$ED,$38,$2F,$A4,$7A,$B1,$F4,$C7,$9F,$0E,$D3,$E7
	.dc.b	$C3,$B3,$B3,$99,$DE,$BF,$42,$F7,$D5,$87,$78,$F6,$10,$E5,$1A,$BF
	.dc.b	$8D,$F1,$3F,$1F,$E2,$A5,$7C,$78,$EC,$FE,$8B,$3F,$E2,$0F,$CD,$F8
	.dc.b	$98,$FC,$D7,$E8,$7C,$57,$48,$FA,$17,$D1,$F0,$E4,$C3,$AC,$7E,$3B
	.dc.b	$0E,$AF,$32,$1C,$30,$EA,$F9,$4E,$A4,$3A,$CF,$4E,$CF,$E3,$9B,$FE
	.dc.b	$16,$BC,$F5,$CC,$FC,$26,$5F,$87,$F8,$9F,$C4,$7C,$54,$9F,$C4,$47
	.dc.b	$3E,$8F,$E9,$1A,$3E,$27,$F1,$C4,$AF,$C7,$7C,$57,$37,$F2,$18,$73
	.dc.b	$7E,$0E,$1C,$DF,$D5,$DF,$37,$F2,$D7,$CD,$F8,$F7,$CD,$F9,$37,$CF
	.dc.b	$6F,$D1,$47,$E3,$B0,$E8,$FE,$0F,$0E,$8F,$D0,$C3,$A5,$70,$E1,$FC
	.dc.b	$3E,$3F,$A7,$F8,$7F,$D4,$7C,$54,$C3,$F5,$11,$FD,$4B,$7F,$D4,$C0
	.dc.b	$FF,$23,$0F,$85,$78,$9F,$B1,$F8,$A9,$3F,$B1,$7F,$69,$E0,$77,$8F
	.dc.b	$DA,$63,$DD,$FC,$2E,$3D,$DF,$EC,$38,$74,$79,$10,$FC,$5D,$F4,$7C
	.dc.b	$E7,$42,$1D,$0F,$E3,$98,$FA,$9F,$4C,$F5,$79,$9D,$74,$F8,$DB,$2F
	.dc.b	$B4,$FE,$6B,$C0,$FE,$1D,$C5,$23,$F8,$74,$7A,$5F,$E5,$58,$75,$C3
	.dc.b	$9F,$81,$FC,$54,$93,$FC,$57,$C1,$FE,$31,$7C,$E3,$F4,$F8,$61,$F4
	.dc.b	$9F,$13,$7E,$2F,$EB,$2E,$E1,$E8,$C3,$84,$F2,$3F,$91,$5E,$27,$F5
	.dc.b	$F1,$FC,$5C,$FE,$D2,$7C,$4B,$E7,$D7,$FC,$BB,$0B,$3F,$C7,$30,$FE
	.dc.b	$5D,$EA,$FF,$2F,$F5,$64,$FF,$2F,$8F,$84,$F5,$D3,$E1,$61,$F8,$9E
	.dc.b	$4F,$98,$FE,$2E,$7F,$8A,$ED,$F9,$C6,$C6,$74,$FC,$C4,$BD,$5F,$0D
	.dc.b	$7F,$3D,$CB,$C2,$FF,$0E,$CB,$C3,$7E,$E3,$A6,$5E,$29,$FF,$4E,$E1
	.dc.b	$DE,$7A,$E3,$DC,$F3,$46,$5B,$8F,$82,$BF,$8D,$E3,$E0,$BF,$C3,$E3
	.dc.b	$E0,$BF,$F0,$FD,$7C,$B3,$FD,$35,$87,$37,$C3,$1D,$DA,$6B,$0C,$65
	.dc.b	$F9,$1F,$56,$DB,$DB,$A6,$23,$C1,$7F,$B5,$E3,$FF,$63,$6B,$EF,$C3
	.dc.b	$F3,$5F,$13,$C8,$95,$E4,$F6,$B9,$3E,$6C,$7A,$BE,$8C,$7A,$BF,$86
	.dc.b	$3F,$D7,$78,$75,$C9,$87,$49,$FB,$6F,$A3,$0E,$B1,$F7,$B0,$EA,$FE
	.dc.b	$03,$0E,$B9,$7A,$38,$B9,$6F,$F8,$0C,$6E,$DF,$53,$2F,$F7,$BF,$4F
	.dc.b	$FB,$E3,$9E,$67,$FE,$F6,$8E,$6F,$CD,$3F,$8B,$63,$FE,$FE,$E8,$91
	.dc.b	$FE,$FC,$7F,$E0,$0E,$DD,$23,$FD,$FF,$7D,$1F,$F5,$95,$F4,$7E,$E1
	.dc.b	$FE,$B5,$AE,$6D,$3F,$92,$F7,$FF,$09,$7C,$87,$42,$1D,$0F,$DE,$9F
	.dc.b	$C2,$30,$E7,$A7,$FB,$36,$AE,$7F,$EB,$5C,$BF,$07,$ED,$FE,$13,$DB
	.dc.b	$98,$FE,$3A,$3C,$CF,$D6,$C7,$F1,$3E,$DC,$8F,$C9,$BF,$8C,$F8,$8E
	.dc.b	$71,$F9,$6B,$E6,$FE,$D2,$F9,$E7,$FA,$AC,$39,$4F,$EE,$30,$B3,$FB
	.dc.b	$6C,$39,$CF,$EF,$30,$E7,$3F,$BA,$C3,$9E,$5F,$C0,$79,$3F,$15,$8F
	.dc.b	$28,$FD,$8B,$E5,$C3,$F5,$33,$FA,$2F,$80,$91,$83,$E4,$9F,$C3,$F3
	.dc.b	$FD,$3F,$C0,$C8,$FD,$8B,$FA,$AF,$81,$FD,$5D,$F2,$9F,$E2,$37,$87
	.dc.b	$9E,$7F,$C2,$72,$F3,$63,$CE,$7F,$8B,$E3,$6B,$F8,$AE,$3C,$67,$99
	.dc.b	$FC,$03,$0C,$4F,$38,$FD,$B1,$FC,$EC,$F2,$90,$EA,$7F,$08,$9F,$E0
	.dc.b	$3F,$03,$FC,$0B,$E1,$E4,$FF,$0C,$8F,$96,$FF,$75,$CF,$F6,$9F,$11
	.dc.b	$FC,$20,$91,$FC,$4B,$E2,$39,$BF,$C5,$2F,$9B,$FC,$CA,$F9,$BF,$A8
	.dc.b	$3F,$8C,$D7,$27,$F9,$BB,$FC,$52,$B9,$C7,$F3,$AA,$E7,$8F,$F3,$8B
	.dc.b	$E7,$87,$92,$B9,$90,$FE,$83,$7F,$C7,$BE,$23,$F8,$FF,$C3,$C9,$FE
	.dc.b	$53,$1C,$B6,$FE,$47,$F0,$F2,$7F,$96,$BF,$CA,$3F,$4A,$7C,$47,$39
	.dc.b	$FE,$63,$87,$37,$FA,$7E,$1C,$DE,$24,$3F,$A9,$DF,$3A,$E5,$1F,$D5
	.dc.b	$AB,$96,$9F,$D5,$2F,$94,$F0,$39,$47,$F5,$8B,$E5,$1F,$C1,$3C,$EF
	.dc.b	$E2,$9D,$4F,$36,$98,$8E,$77,$FD,$8E,$5E,$77,$02,$5F,$F0,$EF,$3B
	.dc.b	$9F,$A5,$D3,$D1,$93,$FC,$4A,$3F,$91,$71,$7F,$61,$8F,$E7,$58,$FE
	.dc.b	$4D,$FB,$C7,$FA,$3C,$62,$F4,$BC,$E4,$8F,$E3,$BE,$97,$F2,$DE,$97
	.dc.b	$B1,$EC,$98,$7F,$91,$A3,$2F,$F2,$8C,$3C,$1E,$3F,$F3,$F8,$FE,$4F
	.dc.b	$EA,$7C,$42,$3F,$CA,$BD,$5E,$DE,$E7,$7F,$62,$4F,$58,$F6,$F5,$7F
	.dc.b	$56,$DF,$DD,$9B,$D9,$F3,$12,$FC,$9E,$CF,$A6,$FD,$5E,$DC,$9F,$44
	.dc.b	$74,$7F,$81,$F8,$9F,$C0,$9F,$67,$B9,$F0,$48,$FB,$FD,$CF,$89,$F0
	.dc.b	$1F,$17,$DF,$93,$F0,$E3,$A3,$FE,$31,$E2,$F8,$4F,$D1,$D5,$F0,$1F
	.dc.b	$34,$91,$F2,$FE,$03,$F3,$DF,$11,$F9,$FF,$88,$93,$FE,$75,$E7,$FF
	.dc.b	$3F,$78,$CA,$EF,$1F,$42,$6E,$FF,$E3,$1C,$9F,$0B,$F6,$C9,$3F,$67
	.dc.b	$E1,$3B,$BD,$6B,$FC,$75,$8B,$2F,$0C,$FF,$B3,$A3,$C0,$FD,$F9,$FF
	.dc.b	$11,$BF,$CD,$4F,$C2,$5F,$67,$E0,$3F,$8B,$BE,$F1,$CE,$FA,$E3,$EC
	.dc.b	$CB,$D3,$3F,$CA,$8F,$ED,$07,$42,$1D,$0F,$51,$E9,$1E,$DB,$2F,$3C
	.dc.b	$7A,$67,$C3,$5F,$E5,$3C,$DF,$0D,$E0,$AF,$D4,$E9,$EC,$AF,$EC,$91
	.dc.b	$E0,$78,$69,$ED,$79,$69,$D2,$BE,$4E,$36,$FF,$D5,$98,$F6,$7E,$41
	.dc.b	$ED,$3E,$41,$EB,$25,$F6,$7D,$A7,$A4,$F7,$67,$ED,$AE,$D0,$AE,$CF
	.dc.b	$F8,$14,$FC,$E9,$FD,$27,$93,$D1,$87,$E1,$72,$B3,$E6,$CB,$F8,$06
	.dc.b	$7E,$9C,$BB,$4F,$EF,$E3,$E3,$70,$7F,$AB,$E3,$B1,$E8,$C3,$A3,$F1
	.dc.b	$DF,$C9,$5F,$48,$F4,$EB,$F5,$35,$FC,$04,$75,$3E,$29,$F4,$8F,$8C
	.dc.b	$FD,$27,$B1,$F3,$32,$F4,$BF,$06,$BB,$3F,$1C,$93,$F7,$F1,$F4,$C7
	.dc.b	$F7,$53,$99,$F2,$6B,$9E,$8B,$E7,$7C,$6B,$CC,$7E,$EF,$E1,$F9,$BF
	.dc.b	$96,$39,$D7,$EC,$1F,$84,$7C,$D3,$8C,$F3,$7F,$82,$61,$FC,$BE,$B9
	.dc.b	$C7,$E1,$27,$F0,$4F,$4C,$3F,$54,$F2,$CB,$F7,$D7,$E6,$7F,$80,$5F
	.dc.b	$2B,$FE,$85,$D3,$FC,$A3,$0F,$69,$CB,$0F,$51,$CC,$FC,$BE,$1E,$37
	.dc.b	$E9,$E9,$F7,$EF,$E9,$1E,$42,$63,$F8,$09,$FE,$3D,$9F,$17,$DA,$75
	.dc.b	$3F,$2D,$97,$96,$7F,$5C,$7F,$31,$8F,$71,$F2,$0F,$B8,$7F,$4F,$8F
	.dc.b	$73,$E6,$30,$79,$8F,$B9,$5E,$47,$F0,$C7,$3C,$FE,$E3,$C0,$F8,$24
	.dc.b	$8F,$E5,$BF,$11,$E5,$8F,$C8,$DF,$96,$7A,$47,$E1,$CE,$98,$7B,$E3
	.dc.b	$F8,$66,$16,$BF,$95,$DF,$3B,$FC,$E1,$FD,$D2,$B9,$E1,$F8,$78,$FC
	.dc.b	$4E,$1C,$5F,$E1,$38,$5D,$BF,$8E,$61,$CF,$2F,$E9,$B8,$F2,$D3,$F3
	.dc.b	$38,$FF,$89,$EB,$96,$3F,$40,$FD,$8B,$C5,$FE,$41,$97,$86,$F9,$47
	.dc.b	$F1,$07,$E1,$BF,$1D,$F6,$BC,$8F,$A0,$7E,$85,$E2,$FA,$8D,$1F,$37
	.dc.b	$3F,$CA,$B8,$8F,$C5,$C7,$E0,$A3,$E7,$9F,$C5,$CF,$B1,$87,$B1,$E1
	.dc.b	$1F,$31,$A7,$F4,$0D,$9A,$7B,$63,$F8,$5E,$1E,$DA,$FC,$B6,$1F,$43
	.dc.b	$2F,$8D,$5C,$B4,$F9,$AF,$F1,$6C,$39,$3F,$C7,$AB,$DA,$71,$78,$62
	.dc.b	$CF,$9C,$7F,$67,$AE,$35,$EE,$8F,$C9,$9F,$9E,$3F,$4E,$77,$39,$CF
	.dc.b	$F9,$7E,$78,$E3,$F9,$13,$D2,$49,$E7,$5F,$D4,$9E,$47,$F7,$7A,$F1
	.dc.b	$C7,$F8,$22,$B9,$CF,$CD,$3C,$51,$EB,$7D,$07,$88,$E6,$FA,$5F,$CC
	.dc.b	$E9,$F9,$0B,$F1,$47,$E8,$CF,$C0,$1F,$E1,$FE,$0F,$F2,$51,$1F,$F2
	.dc.b	$47,$99,$E2,$7F,$45,$3F,$E1,$7E,$2F,$F2,$81,$23,$FC,$9D,$97,$17
	.dc.b	$A1,$FE,$51,$BF,$F2,$B5,$F3,$E1,$FF,$22,$DF,$1A,$FF,$18,$3F,$E5
	.dc.b	$BB,$E5,$97,$F8,$07,$0B,$8F,$F8,$9F,$1E,$5C,$5C,$6B,$FC,$0D,$85
	.dc.b	$AF,$F1,$66,$1C,$31,$FE,$FB,$8D,$7F,$B9,$63,$C9,$FF,$33,$5F,$17
	.dc.b	$FC,$06,$F8,$DF,$51,$0F,$69,$FE,$47,$8E,$4F,$A8,$E6,$43,$FC,$99
	.dc.b	$5E,$1E,$0F,$ED,$98,$7F,$90,$70,$F8,$B5,$FB,$63,$FA,$64,$7C,$F7
	.dc.b	$96,$3E,$E7,$E7,$9F,$58,$F2,$3F,$1D,$C1,$FA,$38,$FD,$C3,$FC,$73
	.dc.b	$3F,$A2,$F4,$8E,$CF,$98,$E3,$3C,$DF,$7B,$FE,$33,$C7,$DE,$F8,$E3
	.dc.b	$E0,$11,$FF,$1C,$63,$EC,$8E,$35,$EF,$7A,$BF,$A7,$7F,$85,$9E,$5C
	.dc.b	$FE,$93,$FE,$27,$CE,$E7,$FB,$98,$E2,$7D,$C3,$E0,$D7,$E1,$5F,$95
	.dc.b	$87,$F6,$CB,$E3,$7C,$B0,$F2,$4F,$B4,$97,$C6,$FE,$41,$E4,$D3,$F5
	.dc.b	$4F,$E1,$0F,$90,$53,$E4,$1F,$83,$3C,$67,$94,$C1,$D0,$FD,$A9,$FC
	.dc.b	$F2,$FC,$AF,$DC,$37,$5E,$47,$93,$F4,$5F,$6D,$7D,$CC,$FE,$DC,$7D
	.dc.b	$18,$FB,$73,$F4,$A3,$E3,$9F,$E5,$1D,$3F,$25,$A7,$DA,$AD,$92,$7D
	.dc.b	$95,$C5,$F1,$C7,$D9,$7E,$D9,$FE,$68,$9E,$4F,$08,$E4,$FD,$93,$D9
	.dc.b	$3E,$97,$89,$E3,$3C,$B7,$FE,$E5,$C7,$FC,$35,$5F,$85,$9E,$47,$B4
	.dc.b	$EC,$7F,$9E,$32,$FB,$71,$E5,$BF,$91,$3F,$BE,$3E,$39,$F8,$EB,$F8
	.dc.b	$F1,$F9,$92,$4F,$F0,$89,$F8,$FA,$7E,$D9,$FA,$31,$F8,$59,$FE,$5D
	.dc.b	$87,$9A,$7F,$9B,$47,$27,$CA,$7A,$1F,$94,$7C,$12,$4F,$33,$FD,$77
	.dc.b	$C5,$F2,$A7,$84,$7C,$AB,$FF,$5F,$47,$CA,$8E,$0F,$D1,$3F,$47,$C1
	.dc.b	$FE,$AB,$AE,$55,$C2,$7E,$41,$D0,$F9,$07,$BB,$3F,$92,$FF,$75,$9E
	.dc.b	$3A,$FE,$52,$3F,$3B,$8F,$A1,$F0,$1F,$84,$38,$1F,$38,$FD,$D1,$FA
	.dc.b	$63,$C0,$4B,$FD,$D3,$C4,$FA,$C7,$3C,$3D,$4F,$3C,$BF,$62,$F1,$BF
	.dc.b	$DD,$3E,$93,$E0,$F0,$7F,$AB,$30,$FA,$B1,$FE,$4E,$CF,$E5,$57,$08
	.dc.b	$F0,$62,$CF,$E0,$E5,$E0,$9F,$DB,$65,$F7,$5F,$EC,$27,$DC,$AF,$99
	.dc.b	$8F,$DF,$8F,$F5,$E6,$BF,$29,$F5,$C7,$C7,$BE,$8F,$92,$3F,$C6,$59
	.dc.b	$79,$1F,$8D,$7F,$12,$7E,$74,$72,$9F,$BF,$5C,$9F,$D2,$61,$E5,$D7
	.dc.b	$F3,$71,$CE,$BE,$BA,$67,$CD,$BC,$EC,$43,$15,$62,$9E,$6E,$A7,$35
	.dc.b	$CA,$31,$5B,$BF,$CC,$5B,$CE,$CD,$F8,$1A,$A3,$4F,$97,$0E,$AE,$7C
	.dc.b	$ED,$CB,$D2,$BE,$84,$AE,$19,$79,$E7,$C3,$E8,$F0,$24,$E8,$F8,$3E
	.dc.b	$97,$4F,$5F,$AE,$F2,$74,$79,$63,$D7,$C5,$E2,$24,$6B,$37,$A9,$E6
	.dc.b	$F7,$3C,$FE,$BC,$AD,$47,$C4,$F6,$3D,$64,$8E,$5E,$C7,$BB,$E0,$3D
	.dc.b	$FE,$FC,$9E,$D1,$F3,$3E,$03,$E1,$92,$3C,$5F,$01,$F3,$3E,$71,$F0
	.dc.b	$97,$E0,$E5,$7B,$27,$2F,$EA,$CF,$A7,$C4,$F1,$79,$7C,$24,$9E,$0F
	.dc.b	$D2,$E8,$ED,$F1,$3D,$FE,$22,$57,$58,$FE,$31,$8B,$E2,$3C,$84,$9F
	.dc.b	$17,$C4,$79,$FE,$1F,$D1,$F1,$12,$7C,$D1,$CB,$E2,$3D,$84,$8F,$57
	.dc.b	$C4,$7B,$FE,$1F,$E0,$7C,$44,$9F,$74,$78,$BE,$23,$E2,$12,$3E,$17
	.dc.b	$C4,$7C,$7F,$87,$F9,$1F,$11,$27,$E3,$47,$AB,$E1,$FE,$61,$23,$E5
	.dc.b	$7C,$3F,$E6,$BE,$75,$F1,$1F,$46,$61,$F8,$A8,$CB,$F3,$CF,$C6,$F4
	.dc.b	$FC,$F7,$EA,$7C,$47,$D9,$23,$F5,$FE,$23,$B7,$C3,$F7,$F8,$89,$3D
	.dc.b	$63,$FA,$36,$0F,$88,$F2,$12,$7C,$5F,$11,$E7,$EF,$F4,$7C,$44,$9F
	.dc.b	$34,$72,$F8,$8F,$61,$23,$D5,$F1,$1E,$FF,$87,$F8,$1F,$11,$27,$DD
	.dc.b	$1E,$2F,$88,$F8,$84,$8F,$85,$F1,$1F,$1F,$E1,$BE,$47,$C2,$C9,$F8
	.dc.b	$D1,$EA,$F8,$8F,$98,$48,$F9,$5F,$11,$F3,$FE,$17,$E8,$7C,$44,$9F
	.dc.b	$9D,$1F,$0B,$E1,$FE,$A1,$23,$E9,$7C,$3F,$F1,$0C,$FF,$90,$47,$EF
	.dc.b	$9F,$E3,$AD,$F8,$5D,$7C,$2F,$F7,$19,$FE,$2B,$3F,$C9,$7C,$FF,$E4
	.dc.b	$F1,$8F,$84,$FE,$2E,$D8,$CF,$C2,$77,$F8,$3F,$07,$C2,$4C,$3C,$11
	.dc.b	$D7,$E1,$3C,$84,$8F,$27,$C2,$79,$FC,$3F,$47,$C1,$C9,$F4,$47,$97
	.dc.b	$E0,$BD,$84,$8F,$67,$C0,$FB,$FE,$03,$E0,$7C,$24,$9F,$81,$1E,$D9
	.dc.b	$F8,$84,$8F,$E4,$3F,$3E,$E8,$E5,$1F,$1E,$F9,$3F,$0C,$FE,$99,$C2
	.dc.b	$C3,$A4,$78,$88,$6C,$5F,$57,$E9,$7C,$27,$D5,$F8,$49,$18,$3F,$50
	.dc.b	$7F,$99,$ED,$97,$9E,$3A,$46,$1F,$4E,$5F,$7E,$F9,$C6,$AF,$0F,$4E
	.dc.b	$4F,$F7,$35,$F1,$8E,$78,$75,$9F,$EA,$78,$F9,$E7,$BE,$1D,$27,$F0
	.dc.b	$38,$74,$9F,$F0,$39,$FE,$0F,$3E,$2D,$F3,$E4,$FC,$4B,$FD,$8B,$0E
	.dc.b	$97,$F8,$AC,$3A,$57,$E3,$B0,$E9,$3F,$E0,$7C,$3F,$D8,$67,$43,$E6
	.dc.b	$BF,$92,$C3,$A6,$5C,$2F,$95,$7E,$7E,$F0,$FF,$95,$A3,$FD,$87,$F0
	.dc.b	$FF,$E2,$CC,$FE,$16,$3C,$EB,$AD,$7F,$81,$B3,$F3,$BF,$A2,$CF,$CF
	.dc.b	$3F,$82,$C7,$AB,$FE,$26,$3F,$D3,$A7,$F2,$0F,$13,$E1,$3E,$93,$E7
	.dc.b	$9F,$60,$FC,$01,$F5,$CF,$41,$8C,$3E,$11,$FE,$BB,$7C,$A7,$37,$D0
	.dc.b	$FE,$3F,$0F,$3B,$F9,$6B,$F3,$AC,$58,$7A,$38,$3F,$D2,$13,$F6,$4D
	.dc.b	$6B,$E8,$69,$E7,$3F,$06,$79,$8F,$41,$83,$F0,$67,$F8,$F8,$F3,$67
	.dc.b	$F8,$27,$D2,$7E,$93,$5F,$93,$87,$92,$3A,$67,$F8,$57,$F3,$15,$FE
	.dc.b	$A3,$9F,$AE,$F3,$8F,$96,$7C,$1A,$E7,$8F,$C0,$A6,$5F,$B6,$AD,$E1
	.dc.b	$F2,$0F,$CF,$DF,$08,$F0,$E1,$F7,$DF,$09,$F8,$93,$D6,$43,$E2,$9E
	.dc.b	$13,$EC,$3F,$92,$7E,$2C,$FA,$DF,$8A,$7E,$70,$E2,$F8,$E3,$D8,$F9
	.dc.b	$23,$99,$F8,$F3,$CA,$7D,$D3,$CA,$7E,$46,$FE,$2B,$E5,$6B,$F3,$CE
	.dc.b	$67,$E0,$F0,$F8,$D8,$7C,$5A,$F1,$4F,$B7,$3F,$C5,$BF,$89,$9F,$99
	.dc.b	$5F,$89,$D3,$FC,$BB,$AF,$F9,$CD,$FC,$46,$3E,$03,$C3,$87,$27,$E9
	.dc.b	$E1,$EA,$3C,$47,$93,$5F,$C9,$BF,$89,$3E,$2E,$7F,$10,$8F,$F9,$F6
	.dc.b	$BA,$65,$E4,$8E,$58,$79,$1F,$CE,$61,$E3,$D7,$F3,$97,$C2,$7C,$53
	.dc.b	$F9,$17,$F3,$5A,$7E,$C3,$1F,$F4,$3E,$7F,$C1,$F0,$F6,$9F,$C4,$2B
	.dc.b	$E5,$DF,$1B,$FE,$4B,$1E,$23,$FC,$DB,$5F,$15,$FD,$0C,$F8,$DF,$D6
	.dc.b	$4F,$E1,$4E,$2F,$E2,$2F,$F2,$4F,$E2,$23,$F8,$13,$F9,$A8,$F8,$6F
	.dc.b	$C6,$3E,$26,$BF,$9B,$79,$1F,$87,$39,$47,$B1,$FF,$31,$E5,$EC,$7F
	.dc.b	$6B,$97,$C5,$78,$9F,$58,$F8,$B3,$FE,$B2,$7E,$6E,$5C,$5F,$D1,$1A
	.dc.b	$3F,$30,$72,$C7,$F0,$2F,$88,$FF,$5B,$65,$F3,$5F,$8B,$3F,$35,$E4
	.dc.b	$78,$72,$F8,$A7,$C7,$BF,$14,$7E,$1E,$FE,$63,$F1,$4F,$86,$7D,$6D
	.dc.b	$7F,$3E,$FE,$A3,$5F,$F2,$A3,$B3,$E5,$13,$2F,$85,$87,$83,$0F,$DE
	.dc.b	$3E,$19,$FE,$08,$7F,$40,$3F,$BC,$63,$F0,$E3,$F7,$51,$FD,$E3,$E0
	.dc.b	$FF,$C6,$9F,$07,$23,$FC,$69,$1F,$49,$FC,$39,$FC,$CF,$0F,$D4,$5F
	.dc.b	$E6,$36,$FF,$1E,$57,$1A,$FF,$01,$DE,$1F,$66,$DF,$27,$1F,$DD,$C7
	.dc.b	$AE,$F4,$F0,$3F,$C8,$95,$89,$FD,$04,$7D,$43,$C8,$7D,$C7,$9E,$BF
	.dc.b	$BC,$AE,$19,$FF,$91,$AF,$C9,$5F,$E4,$6C,$78,$BF,$ED,$ED,$BF,$45
	.dc.b	$8F,$3B,$FC,$CE,$3C,$E3,$E1,$9E,$42,$1F,$24,$F6,$9F,$A8,$8E,$39
	.dc.b	$FB,$DF,$F2,$BE,$1E,$27,$C9,$3D,$0F,$C8,$0B,$3E,$68,$FF,$00,$65
	.dc.b	$F6,$5B,$1C,$71,$FF,$30,$E5,$D6,$3F,$C8,$78,$71,$7E,$41,$FC,$67
	.dc.b	$4F,$B2,$FE,$B3,$0F,$8E,$FF,$92,$AB,$93,$FD,$D6,$B9,$3F,$8B,$B6
	.dc.b	$1E,$68,$FF,$07,$61,$D1,$FC,$C3,$FC,$B6,$B9,$E6,$C3,$BD,$7F,$85
	.dc.b	$72,$FA,$38,$FE,$50,$FB,$79,$FC,$99,$F5,$9F,$42,$3D,$EF,$C1,$CB
	.dc.b	$F5,$EF,$F5,$9C,$BC,$53,$F8,$E7,$C5,$1C,$67,$E8,$BF,$64,$FA,$37
	.dc.b	$EB,$7F,$CC,$DA,$FE,$CA,$FC,$95,$F7,$27,$EF,$1E,$43,$F3,$04,$8F
	.dc.b	$C7,$3F,$84,$3F,$18,$7F,$61,$C3,$D8,$FC,$D3,$C8,$7D,$13,$D8,$7E
	.dc.b	$EA,$FE,$2D,$7E,$02,$3F,$82,$65,$F9,$57,$D5,$7F,$11,$F1,$4F,$E4
	.dc.b	$1F,$14,$7A,$9F,$11,$EA,$3C,$47,$03,$C4,$74,$8F,$9B,$1F,$CF,$8E
	.dc.b	$51,$FC,$52,$7C,$51,$F0,$23,$91,$FE,$3D,$CB,$EF,$E3,$FD,$AA,$3E
	.dc.b	$B6,$9F,$9B,$7F,$91,$9D,$CF,$0D,$7A,$CE,$87,$F0,$27,$F4,$C7,$E5
	.dc.b	$33,$FB,$F8,$FF,$50,$7E,$FC,$78,$B1,$FC,$83,$F3,$0F,$C8,$1F,$C6
	.dc.b	$EF,$C4,$FF,$84,$CF,$11,$E9,$3E,$A9,$F1,$AF,$C5,$97,$E1,$DF,$D1
	.dc.b	$C7,$F3,$0B,$F1,$BF,$E1,$7A,$FC,$FD,$F8,$A3,$C9,$A7,$EB,$30,$FB
	.dc.b	$B1,$F0,$B0,$FD,$0B,$F5,$69,$5F,$AA,$CF,$E5,$EB,$FE,$0B,$AF,$DE
	.dc.b	$63,$EA,$CB,$E0,$E3,$F2,$73,$FC,$FB,$F2,$9F,$B6,$7F,$6F,$AF,$F4
	.dc.b	$21,$DF,$5F,$E7,$8F,$C7,$CF,$D1,$5C,$E3,$F4,$0F,$C2,$7E,$96,$1E
	.dc.b	$13,$DF,$97,$F0,$0C,$FD,$D3,$F3,$CF,$ED,$67,$B5,$F6,$3F,$57,$5F
	.dc.b	$C3,$5F,$E3,$F2,$FD,$03,$F1,$27,$F5,$67,$D3,$8F,$69,$1F,$2B,$EE
	.dc.b	$7F,$90,$9F,$BC,$D3,$F6,$39,$7F,$74,$AF,$CD,$BF,$3A,$7E,$FB,$F0
	.dc.b	$2B,$F5,$51,$FC,$18,$F7,$9E,$A3,$C7,$0F,$84,$FE,$49,$5F,$84,$9F
	.dc.b	$F4,$F7,$C2,$E2,$FF,$94,$5A,$31,$3F,$EF,$AB,$F8,$EF,$DC,$3F,$6D
	.dc.b	$B7,$F9,$80,$77,$70,$87,$CF,$BE,$13,$18,$F8,$73,$F1,$37,$3F,$E7
	.dc.b	$15,$BB,$FE,$0B,$1F,$38,$8F,$E4,$DF,$F6,$86,$1D,$E5,$F0,$D7,$E1
	.dc.b	$A5,$FF,$58,$C7,$E0,$56,$A3,$A9,$CC,$87,$EB,$E3,$FD,$FB,$7D,$7E
	.dc.b	$0F,$B7,$C1,$CA,$E9,$97,$A2,$7E,$2C,$7F,$0A,$3E,$2F,$27,$FB,$EE
	.dc.b	$FC,$DF,$0B,$E7,$F8,$69,$5C,$8F,$ED,$B9,$A7,$9E,$7F,$8F,$9F,$2F
	.dc.b	$07,$CF,$C7,$A4,$71,$3E,$7F,$C1,$FC,$3F,$83,$91,$9C,$53,$D3,$5F
	.dc.b	$D7,$4F,$E4,$4F,$39,$F7,$34,$F9,$9F,$09,$F3,$7E,$16,$4F,$38,$E4
	.dc.b	$FE,$6B,$4F,$C6,$74,$7F,$55,$C3,$EB,$CF,$CB,$6E,$BF,$86,$8F,$27
	.dc.b	$C1,$FD,$9F,$83,$91,$83,$EA,$9C,$4F,$83,$BF,$C0,$58,$FD,$CF,$84
	.dc.b	$FC,$17,$C2,$C9,$F8,$11,$E6,$7F,$9A,$9F,$C3,$0F,$DB,$E3,$F2,$DF
	.dc.b	$94,$F9,$5E,$B9,$7E,$53,$E0,$BF,$1D,$52,$7E,$64,$75,$F0,$7F,$98
	.dc.b	$57,$F8,$47,$5F,$85,$5E,$83,$E6,$1C,$0F,$29,$0E,$E7,$E0,$AB,$A3
	.dc.b	$FB,$C3,$CF,$9F,$C1,$CF,$F3,$73,$E3,$9E,$27,$EC,$31,$FD,$24,$72
	.dc.b	$8F,$81,$3F,$D9,$B2,$F8,$11,$C1,$F8,$11,$D2,$3C,$8F,$CD,$7D,$07
	.dc.b	$F0,$F3,$9E,$3F,$73,$0F,$F0,$4D,$72,$8F,$91,$3C,$9F,$DE,$E1,$E4
	.dc.b	$8F,$BB,$3E,$2B,$F2,$70,$66,$7D,$7C,$4B,$F7,$57,$96,$7C,$47,$E6
	.dc.b	$4F,$8A,$7C,$E3,$8E,$5C,$E3,$81,$FC,$0B,$E0,$FE,$24,$F1,$8F,$89
	.dc.b	$9F,$AB,$1E,$19,$FA,$AB,$99,$FA,$BB,$E3,$7F,$A0,$EF,$F8,$38,$71
	.dc.b	$AF,$8F,$3E,$0B,$E4,$F8,$63,$E0,$CF,$07,$F2,$38,$78,$1F,$C1,$1C
	.dc.b	$AB,$F9,$2E,$1E,$07,$F1,$D8,$71,$9F,$CD,$D7,$F2,$3A,$E5,$1F,$CC
	.dc.b	$B1,$C3,$E0,$3F,$7E,$7E,$06,$72,$FB,$89,$8F,$9D,$F9,$EF,$BD,$E2
	.dc.b	$FD,$87,$E2,$1E,$73,$D2,$7D,$AA,$F4,$3F,$E1,$3C,$7D,$13,$FC,$BA
	.dc.b	$38,$9E,$0C,$7E,$39,$F0,$1F,$41,$F7,$8F,$41,$C2,$3E,$73,$E0,$7D
	.dc.b	$07,$F8,$0D,$FC,$B9,$FB,$33,$E3,$9C,$63,$BB,$C5,$F0,$3F,$01,$F9
	.dc.b	$EF,$C0,$7F,$8D,$DF,$C8,$9F,$0C,$FC,$F3,$A1,$FE,$67,$3F,$C1,$C7
	.dc.b	$A5,$F8,$0F,$C5,$3E,$41,$C4,$C1,$E1,$3E,$41,$E1,$3F,$D6,$7A,$7E
	.dc.b	$BD,$FE,$93,$87,$E0,$34,$65,$E5,$C7,$FC,$21,$5F,$C8,$8F,$D0,$3F
	.dc.b	$B9,$7F,$51,$87,$C8,$8F,$87,$1E,$38,$C3,$EB,$3F,$C0,$47,$AE,$3F
	.dc.b	$B9,$BF,$00,$E0,$F2,$A5,$FE,$0A,$7F,$D6,$B9,$7E,$0A,$B8,$C7,$C0
	.dc.b	$BF,$97,$7E,$A8,$E9,$AF,$F8,$CA,$BF,$AC,$D7,$E4,$E7,$C4,$7F,$66
	.dc.b	$CF,$E0,$3F,$6C,$87,$F8,$53,$1F,$15,$FF,$36,$8F,$E9,$C4,$CF,$FC
	.dc.b	$39,$5C,$A3,$D6,$72,$C7,$D6,$FE,$53,$0F,$E1,$2F,$EE,$8F,$85,$1E
	.dc.b	$97,$F3,$87,$D6,$38,$63,$F4,$5F,$CD,$9F,$05,$F9,$19,$F8,$71,$F8
	.dc.b	$13,$FD,$BE,$7F,$26,$FC,$03,$D1,$87,$9E,$FC,$18,$7C,$AC,$BD,$FA
	.dc.b	$7C,$FB,$F8,$C7,$9B,$2F,$C4,$4E,$A3,$F1,$C7,$43,$47,$E5,$AF,$E0
	.dc.b	$BF,$95,$3E,$D9,$C3,$3F,$CA,$3F,$3C,$E1,$5E,$BB,$FF,$43,$47,$F8
	.dc.b	$9C,$F5,$4F,$0A,$EB,$87,$E0,$AF,$85,$7F,$87,$F6,$FF,$4A,$BF,$5F
	.dc.b	$5F,$E6,$0F,$C9,$3E,$B4,$F0,$7F,$B2,$CF,$0C,$3F,$A8,$69,$FB,$A9
	.dc.b	$F6,$9F,$58,$F4,$1E,$DD,$7F,$0D,$3C,$23,$E4,$9D,$CF,$C2,$93,$2F
	.dc.b	$4D,$7E,$C2,$FF,$C0,$B3,$FB,$A3,$DF,$AF,$D7,$C7,$E8,$C7,$E1,$0F
	.dc.b	$E2,$D9,$FF,$06,$3F,$62,$F5,$8F,$64,$7F,$9E,$EB,$D9,$9F,$DD,$A6
	.dc.b	$5F,$2F,$5F,$E0,$98,$70,$CB,$F5,$24,$AF,$26,$3F,$9A,$7F,$26,$F8
	.dc.b	$33,$FC,$5C,$FE,$EE,$FF,$0D,$5F,$63,$1F,$F0,$A4,$FF,$28,$BF,$B1
	.dc.b	$3F,$B8,$D7,$EF,$C7,$F9,$6E,$3E,$C1,$FC,$42,$78,$1F,$A0,$78,$1F
	.dc.b	$7C,$FE,$53,$1E,$53,$E2,$47,$08,$FB,$E7,$03,$EF,$9C,$1F,$D7,$47
	.dc.b	$08,$FD,$91,$F7,$F0,$F6,$B8,$8F,$E9,$0F,$F2,$63,$E9,$CF,$E1,$DF
	.dc.b	$C0,$4F,$07,$F0,$18,$7C,$4A,$F3,$56,$DF,$95,$8F,$DE,$AF,$C6,$61
	.dc.b	$F5,$F2,$FE,$1F,$1F,$92,$AF,$DC,$3F,$83,$3E,$39,$F5,$2F,$E0,$3F
	.dc.b	$83,$3E,$04,$7E,$0D,$F8,$07,$A0,$FE,$8F,$A7,$E3,$DF,$DD,$63,$E7
	.dc.b	$AF,$8A,$79,$CE,$B9,$7B,$49,$1B,$9E,$87,$F0,$53,$E1,$E5,$ED,$C3
	.dc.b	$E4,$9F,$CD,$4F,$87,$1F,$A8,$7E,$1B,$E2,$3D,$D8,$FE,$B5,$FC,$CE
	.dc.b	$3E,$77,$EB,$5F,$85,$F8,$75,$FE,$11,$7C,$27,$33,$E6,$9F,$C4,$F3
	.dc.b	$FC,$AB,$F7,$72,$FC,$73,$FB,$9D,$7F,$96,$BF,$D9,$6B,$D1,$3F,$12
	.dc.b	$3D,$0F,$C4,$CB,$DC,$FE,$F6,$3F,$CE,$D3,$EE,$7D,$27,$DB,$8F,$D4
	.dc.b	$3E,$EB,$FE,$2F,$3F,$0A,$3D,$A6,$4F,$D2,$1E,$73,$F3,$A7,$9F,$4F
	.dc.b	$D5,$D7,$EE,$63,$F9,$89,$FC,$30,$F0,$1F,$D5,$F1,$F9,$19,$FC,$D3
	.dc.b	$F1,$F1,$F2,$0F,$09,$F2,$0F,$DE,$1F,$D6,$CD,$5F,$E5,$1D,$3E,$F8
	.dc.b	$E6,$FA,$CD,$1F,$82,$AF,$EA,$EE,$F0,$FA,$75,$FC,$BA,$3F,$83,$BE
	.dc.b	$D3,$F8,$01,$FE,$0E,$7B,$67,$F8,$37,$AD,$7D,$19,$FF,$1E,$9F,$84
	.dc.b	$7F,$34,$FC,$93,$F4,$C7,$53,$F8,$29,$D5,$FB,$CF,$D8,$3D,$EF,$D8
	.dc.b	$7D,$EE,$AF,$DE,$FE,$0E,$3F,$21,$3E,$F7,$F9,$5C,$6E,$BD,$0F,$39
	.dc.b	$D4,$FD,$87,$DA,$FD,$83,$F8,$59,$F8,$33,$FC,$AE,$79,$0F,$B8,$65
	.dc.b	$FD,$11,$F4,$63,$B1,$F1,$63,$F8,$D1,$F1,$4F,$CD,$9F,$17,$4F,$DF
	.dc.b	$67,$FD,$16,$3F,$C1,$C4,$3C,$D3,$EB,$7C,$D3,$FC,$11,$F1,$47,$D0
	.dc.b	$7E,$E3,$FA,$EA,$FE,$B3,$A7,$EB,$B5,$FE,$C7,$5D,$0F,$E1,$79,$7F
	.dc.b	$57,$7F,$C0,$E7,$F1,$CB,$FC,$F4,$F0,$AF,$EE,$79,$FF,$29,$CF,$E6
	.dc.b	$DF,$C4,$3E,$B9,$D2,$3F,$1F,$97,$F7,$48,$FD,$5E,$5E,$4D,$7F,$57
	.dc.b	$1E,$FC,$3A,$BF,$E1,$BC,$3A,$BF,$50,$EA,$7C,$B3,$F9,$15,$7D,$17
	.dc.b	$C4,$FD,$E9,$E7,$1F,$B5,$3E,$D1,$CE,$35,$5F,$D0,$F0,$F3,$EB,$FD
	.dc.b	$E2,$BE,$11,$F4,$F2,$F8,$55,$B8,$FD,$B5,$F8,$5E,$99,$FE,$06,$3F
	.dc.b	$8D,$C7,$C3,$37,$1F,$99,$7F,$80,$C7,$89,$FE,$50,$68,$FC,$36,$BF
	.dc.b	$C8,$1F,$F4,$11,$FE,$39,$8F,$CC,$A5,$28,$9F,$8E,$BD,$97,$C5,$66
	.dc.b	$60,$9D,$46,$23,$8C,$F0,$71,$3A,$6C,$65,$E7,$19,$6C,$65,$F0,$46
	.dc.b	$5B,$19,$7B,$CF,$85,$F2,$46,$5B,$1E,$8F,$CE,$84,$00

;--------------------------------------------------------------------------------
;	FATANH.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fatanh_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FATANH.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FATANH.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fatanh_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1687+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fatanh.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FATANH.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1687+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fatanh.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FATANH.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fatanh_data_compressed::
	.dc.b	$00,$02,$93,$04,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$FA,$77,$E2,$7F,$C2
	.dc.b	$CF,$C1,$F0,$9F,$C1,$BF,$D7,$5F,$EA,$2D,$F3,$38,$F9,$9C,$BC,$CE
	.dc.b	$7E,$67,$4F,$33,$AF,$99,$DB,$CC,$EF,$E6,$78,$3C,$CF,$0F,$99,$E2
	.dc.b	$F3,$3C,$7E,$67,$F6,$1F,$C5,$1F,$2F,$7E,$2B,$E7,$FA,$7F,$D9,$7E
	.dc.b	$6D,$F2,$9B,$F5,$71,$F0,$F9,$F9,$5D,$7C,$9E,$FE,$3B,$87,$C5,$F1
	.dc.b	$F7,$F9,$7B,$7C,$FD,$BE,$9D,$9D,$7B,$3B,$76,$77,$EC,$F0,$76,$78
	.dc.b	$7B,$3C,$5D,$9E,$3E,$CF,$27,$67,$97,$B3,$CD,$D9,$E7,$EC,$FE,$F9
	.dc.b	$C9,$EC,$ED,$FF,$04,$3F,$EC,$4E,$DF,$F1,$AB,$FE,$7F,$2F,$C5,$F0
	.dc.b	$F8,$8E,$3F,$01,$CB,$DD,$E7,$EE,$F4,$F7,$7A,$FB,$BD,$BD,$DE,$FE
	.dc.b	$EF,$83,$DD,$F0,$FB,$BE,$2F,$77,$C7,$EE,$F9,$3D,$DF,$2F,$BB,$F4
	.dc.b	$78,$3C,$FE,$EF,$A3,$DD,$FF,$0A,$7C,$E7,$F6,$D1,$F1,$38,$F8,$9C
	.dc.b	$FC,$4E,$BE,$27,$7F,$13,$C3,$E2,$78,$FC,$37,$2F,$84,$E7,$F0,$9D
	.dc.b	$3E,$13,$AF,$C2,$76,$F8,$4E,$FF,$09,$E0,$F8,$4F,$0F,$C2,$78,$BE
	.dc.b	$13,$C7,$F0,$9E,$4F,$84,$F2,$FC,$27,$9B,$E1,$3C,$FF,$03,$FE,$2B
	.dc.b	$F9,$DF,$C3,$F9,$9F,$E3,$75,$EF,$E7,$CC,$D7,$BF,$BF,$7F,$87,$BF
	.dc.b	$C7,$DF,$E5,$EF,$F3,$F7,$FF,$2D,$E9,$75,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$47,$BF,$FE,$43,$F9,$EF,$C4,$74,$BE,$03,$F9,$0A,$87
	.dc.b	$C2,$6F,$E0,$B8,$7C,$9F,$1F,$5B,$97,$2F,$3F,$83,$E9,$F0,$7D,$7E
	.dc.b	$0F,$B7,$C1,$F7,$F8,$3F,$07,$C1,$F8,$7E,$0F,$C5,$F0,$7E,$3F,$83
	.dc.b	$F2,$7C,$1F,$97,$E0,$FC,$DF,$07,$E7,$F8,$3F,$DD,$6B,$FB,$BD,$7F
	.dc.b	$79,$AF,$EF,$75,$FD,$F6,$BE,$DF,$83,$F7,$7C,$1F,$F9,$87,$E0,$3F
	.dc.b	$C7,$47,$E3,$BA,$3F,$CD,$CB,$E0,$74,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$FF,$0D,$E8,$ED,$F0,$3D,$FE,$07,$C1
	.dc.b	$F0,$3F,$C4,$7A,$FF,$89,$75,$FF,$13,$EB,$FE,$29,$D7,$FC,$57,$AF
	.dc.b	$F8,$B7,$5F,$F1,$7E,$6F,$E3,$1C,$DF,$E7,$9F,$C5,$7F,$CA,$47,$A7
	.dc.b	$9C,$FA,$5F,$F4,$43,$FE,$E4,$3F,$D3,$C7,$F8,$38,$FE,$8C,$7F,$45
	.dc.b	$78,$F0,$F5,$E1,$ED,$C1,$DF,$83,$E4,$7C,$17,$C9,$F8,$2F,$95,$F0
	.dc.b	$5F,$2F,$E0,$BE,$67,$C1,$7C,$DF,$82,$F9,$DF,$05,$F3,$FE,$0B,$E8
	.dc.b	$7C,$17,$FA,$47,$E8,$3F,$4F,$C0,$7F,$4E,$FF,$A7,$D7,$C0,$EF,$E0
	.dc.b	$78,$7C,$0F,$1F,$81,$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF
	.dc.b	$03,$DF,$E0,$7C,$1F,$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0
	.dc.b	$3F,$37,$C9,$F9,$DD,$9F,$3F,$B3,$E8,$76,$7D,$1E,$CF,$A5,$D9,$F4
	.dc.b	$FB,$3E,$A7,$67,$D5,$EC,$FA,$DD,$9F,$EA,$FF,$4F,$DE,$7B,$BB,$4F
	.dc.b	$B9,$FF,$5F,$2E,$BE,$1D,$7C,$7A,$F9,$75,$F3,$EB,$E9,$D7,$D7,$AF
	.dc.b	$E6,$7C,$0F,$CD,$F8,$1F,$9D,$F0,$3E,$1E,$EF,$A1,$F0,$3F,$EC,$1F
	.dc.b	$A1,$FD,$F9,$FB,$CE,$0F,$F6,$72,$F7,$F5,$EF,$EF,$DF,$E1,$EF,$F1
	.dc.b	$F7,$F9,$7B,$FC,$FD,$FE,$9C,$FD,$7D,$FF,$81,$E4,$7C,$1F,$23,$E1
	.dc.b	$79,$1F,$0F,$C8,$F8,$9E,$47,$8F,$DF,$F2,$7B,$3E,$5F,$67,$CD,$EC
	.dc.b	$F9,$FD,$9F,$47,$B3,$E9,$F6,$7D,$5E,$CF,$AF,$D9,$F6,$7B,$3F,$ED
	.dc.b	$DF,$A2,$FF,$5B,$1F,$A5,$3F,$49,$D8,$5E,$FE,$FD,$FE,$1E,$FF,$1F
	.dc.b	$7F,$97,$BF,$F6,$7D,$8E,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B
	.dc.b	$9E,$1F,$73,$C5,$EE,$78,$FD,$CF,$27,$B9,$E5,$F7,$3C,$DE,$E7,$9F
	.dc.b	$DC,$F4,$7B,$9F,$87,$F5,$3D,$5E,$E7,$FB,$DF,$CF,$FD,$39,$F8,$93
	.dc.b	$F1,$1D,$27,$F1,$08,$C7,$B7,$C7,$DB,$E5,$ED,$F3,$F6,$FA,$7B,$7D
	.dc.b	$7D,$BE,$DE,$CF,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC
	.dc.b	$9D,$FE,$5E,$2F,$37,$B7,$E7,$F6,$FD,$1E,$DF,$8F,$8D,$F4,$7F,$D3
	.dc.b	$8F,$C8,$38,$F6,$F5,$1E,$B8,$F5,$3F,$EB,$D3,$FC,$EA,$7F,$54,$3F
	.dc.b	$CE,$6E,$BD,$7E,$3E,$BF,$2F,$5F,$F6,$19,$74,$F5,$FA,$FA,$FD,$BD
	.dc.b	$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F
	.dc.b	$AF,$EB,$FA,$4F,$67,$09,$FE,$64,$7E,$38,$FC,$6B,$F4,$D7,$AD,$C7
	.dc.b	$D6,$E5,$EB,$7F,$05,$E8,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07
	.dc.b	$AD,$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$79
	.dc.b	$FD,$6F,$47,$AD,$E9,$F5,$BD,$5E,$B7,$D5,$FA,$5F,$E6,$67,$F1,$53
	.dc.b	$F1,$A7,$E3,$39,$CB,$D6,$E1,$EB,$71,$F5,$BF,$8F,$79,$7C,$FD,$6E
	.dc.b	$9E,$B7,$E0,$7F,$16,$FF,$06,$9F,$80,$3F,$87,$65,$F9,$77,$F0,$09
	.dc.b	$FF,$7E,$1F,$EB,$43,$FC,$26,$7E,$01,$D7,$99,$D7,$CC,$ED,$E6,$77
	.dc.b	$F2,$FC,$1E,$5F,$87,$CB,$F1,$79,$7E,$3F,$2F,$C9,$E5,$F9,$7C,$BF
	.dc.b	$85,$EB,$FC,$3F,$5F,$E2,$7A,$FF,$17,$D7,$F8,$DE,$BF,$C7,$F5,$FE
	.dc.b	$47,$AF,$F2,$7D,$7F,$95,$EB,$FC,$BF,$57,$E6,$79,$DF,$37,$AF,$F4
	.dc.b	$5F,$01,$FE,$2E,$3F,$42,$7C,$2E,$9F,$E8,$1E,$CE,$BD,$FD,$FB,$FC
	.dc.b	$3D,$FE,$3E,$FF,$2E,$EE,$7D,$DD,$3B,$BA,$F7,$76,$EE,$EF,$DD,$E0
	.dc.b	$EE,$F0,$F7,$78,$BB,$BC,$7D,$DE,$4F,$6F,$CB,$ED,$F9,$BD,$BF,$DC
	.dc.b	$7D,$3F,$E7,$8F,$F0,$71,$FC,$A8,$FC,$EF,$29,$FC,$EB,$FC,$45,$7B
	.dc.b	$5C,$BD,$AE,$7E,$D7,$4F,$6B,$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF
	.dc.b	$0F,$B5,$E2,$F6,$BC,$7E,$D7,$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED
	.dc.b	$7A,$3D,$AF,$4F,$B5,$FC,$47,$EA,$3F,$C1,$E7,$F7,$13,$F9,$49,$F9
	.dc.b	$D3,$F3,$9C,$BF,$CA,$57,$B3,$C7,$D9,$E5,$EC,$F3,$F6,$7A,$7B,$1D
	.dc.b	$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C
	.dc.b	$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$F9,$47,$D4,$FF,$9D,$0F,$E4
	.dc.b	$E7,$E1,$CF,$D3,$8F,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB
	.dc.b	$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F
	.dc.b	$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6,$F5,$FF,$9F,$7B,$BF,$C9,$0F,$F2
	.dc.b	$A1,$FD,$C0,$FC,$39,$F8,$6F,$1B,$F9,$1F,$AF,$C3,$D7,$E3,$EB,$F2
	.dc.b	$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5,$FC,$3E
	.dc.b	$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F,$AF,$FD
	.dc.b	$57,$EA,$BF,$CA,$67,$F8,$D4,$FE,$DA,$7F,$21,$3F,$0B,$C8,$7F,$0A
	.dc.b	$FF,$6F,$5E,$B7,$3F,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E
	.dc.b	$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7
	.dc.b	$F5,$BD,$1E,$B7,$F6,$DF,$AB,$FF,$5C,$9F,$DB,$0F,$E1,$87,$F2,$C3
	.dc.b	$F0,$DC,$7F,$E0,$55,$EA,$F1,$F5,$79,$7A,$BC,$FD,$5E,$9E,$AF,$5F
	.dc.b	$57,$B7,$AB,$DF,$D5,$F0,$7A,$BE,$1F,$57,$C5,$EA,$F8,$FD,$5F,$27
	.dc.b	$AB,$E5,$F5,$7C,$DE,$AF,$9F,$D5,$FF,$02,$7D,$67,$F6,$A3,$FD,$34
	.dc.b	$7F,$8D,$8F,$E1,$A7,$E1,$CF,$E1,$7C,$9F,$E1,$D5,$EA,$72,$F5,$39
	.dc.b	$FA,$9D,$3D,$4E,$BE,$A7,$6F,$53,$BF,$A9,$E0,$F5,$3C,$3E,$A7,$8B
	.dc.b	$D4,$F1,$FA,$9E,$4F,$53,$CB,$EA,$79,$BD,$4F,$3F,$A9,$E8,$F5,$3F
	.dc.b	$C3,$9E,$BF,$FA,$68,$FF,$3E,$1F,$E3,$43,$FB,$41,$F8,$83,$F0,$FC
	.dc.b	$9F,$B7,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7,$4F,$53,$AF,$A9,$DB,$D4
	.dc.b	$EF,$EA,$78,$3D,$4F,$0F,$A9,$E2,$F5,$3C,$7E,$A7,$93,$D4,$F2,$FA
	.dc.b	$9E,$6F,$53,$FC,$65,$F5,$BF,$EF,$E3,$FC,$62,$7F,$51,$3F,$B8,$9F
	.dc.b	$86,$E2,$3F,$86,$7F,$C9,$8B,$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD
	.dc.b	$3E,$FE,$9F,$83,$D3,$F0,$FA,$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F
	.dc.b	$A7,$E6,$F4,$FC,$FE,$9F,$A3,$D3,$FF,$25,$7D,$77,$F8,$C0,$FF,$70
	.dc.b	$1F,$E7,$C3,$FA,$99,$FC,$34,$FC,$3F,$19,$FC,$3B,$FE,$65,$7F,$DC
	.dc.b	$03,$D1,$EB,$E8,$F6,$F4,$7B,$FA,$3E,$0F,$47,$C3,$E8,$F8,$BD,$1F
	.dc.b	$1F,$A3,$E4,$F4,$7C,$BE,$8F,$9B,$D1,$F3,$FA,$3E,$8F,$47,$D3,$E8
	.dc.b	$FF,$98,$FE,$BF,$FD,$C2,$7F,$B0,$CF,$F3,$D9,$FE,$2F,$3F,$87,$9F
	.dc.b	$88,$3F,$86,$F1,$7F,$9F,$57,$A1,$CF,$D0,$E9,$E8,$75,$F4,$3B,$7A
	.dc.b	$1D,$FD,$0F,$07,$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F
	.dc.b	$43,$CD,$E8,$79,$FD,$0F,$F3,$D7,$C1,$3E,$5E,$FC,$7E,$3E,$23,$3F
	.dc.b	$07,$AF,$5F,$7E,$DF,$0F,$5B,$8F,$AD,$CB,$D6,$E7,$EB,$74,$F5,$BA
	.dc.b	$FA,$DD,$BD,$6E,$FE,$B7,$83,$D6,$F0,$FA,$DE,$2F,$5B,$C7,$EB,$79
	.dc.b	$3D,$6F,$2F,$AD,$E6,$F5,$BF,$D2,$1F,$61,$FE,$76,$1F,$2B,$8F,$91
	.dc.b	$CF,$C6,$6B,$E0,$F7,$F0,$7C,$3E,$03,$8F,$C0,$72,$F7,$79,$FB,$BD
	.dc.b	$3D,$DE,$BE,$EF,$6F,$77,$BF,$BB,$E0,$F7,$7C,$3E,$EF,$8B,$DD,$F1
	.dc.b	$FB,$BE,$4F,$77,$CB,$EE,$F9,$BD,$DF,$3F,$BB,$FE,$A7,$FB,$17,$CC
	.dc.b	$5F,$96,$C7,$C9,$67,$E3,$35,$E9,$6F,$E2,$B8,$7C,$37,$1F,$82,$E5
	.dc.b	$F0,$5C,$FE,$0B,$A7,$C1,$75,$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05
	.dc.b	$E1,$F8,$2F,$17,$C1,$78,$FE,$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17
	.dc.b	$FA,$F7,$EC,$BF,$09,$F0,$BF,$ED,$17,$FD,$7A,$7F,$97,$4F,$F3,$E9
	.dc.b	$FC,$34,$FE,$18,$9F,$E1,$8F,$1F,$77,$A7,$BB,$D7,$DD,$ED,$EE,$F7
	.dc.b	$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE,$3F,$77,$C9,$EE,$F9,$7D,$DF
	.dc.b	$37,$BB,$E7,$F7,$7D,$1E,$EF,$FB,$57,$D8,$FF,$5D,$9F,$86,$E1,$FF
	.dc.b	$75,$2F,$6F,$5E,$DE,$FD,$BE,$13,$C7,$DB,$E5,$ED,$F3,$F6,$FA,$7B
	.dc.b	$7D,$7D,$BE,$DE,$DF,$7F,$6F,$C1,$ED,$F8,$7D,$BF,$17,$B7,$E3,$F6
	.dc.b	$FC,$9E,$DF,$97,$DB,$F3,$7B,$7E,$7F,$6F,$D1,$ED,$FF,$BB,$7E,$CF
	.dc.b	$F0,$FC,$FF,$F0,$05,$7B,$59,$F6,$B5,$ED,$6F,$DA,$E1,$ED,$71,$F6
	.dc.b	$B9,$7B,$5C,$FD,$AE,$9E,$D7,$5F,$6B,$B7,$B5,$DF,$DA,$F0,$7B,$5E
	.dc.b	$1F,$6B,$C5,$ED,$78,$FD,$AF,$27,$B5,$E5,$F6,$BC,$DE,$D7,$9F,$DA
	.dc.b	$73,$CF,$B4,$FC,$4A,$5F,$0B,$A8,$C7,$C2,$D9,$FF,$63,$9F,$D5,$CF
	.dc.b	$C5,$9F,$8A,$8E,$5E,$E6,$FD,$CE,$1E,$E7,$1F,$73,$97,$B9,$CF,$DC
	.dc.b	$E9,$EE,$75,$F7,$3B,$7B,$9D,$FD,$CF,$07,$B9,$E1,$F7,$3C,$5E,$E7
	.dc.b	$8F,$DC,$F3,$7D,$AF,$9F,$84,$FE,$2D,$BE,$EE,$BD,$DD,$FB,$BC,$3D
	.dc.b	$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B,$FB,$BE
	.dc.b	$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF,$9B,$DD
	.dc.b	$F3,$FB,$BE,$8F,$77,$E7,$7C,$27,$E2,$B9,$FF,$C8,$BE,$EE,$3D,$DC
	.dc.b	$FB,$BA,$F7,$77,$EE,$F0,$F7,$78,$FB,$BC,$BD,$DE,$7E,$EF,$4F,$77
	.dc.b	$AF,$BB,$DB,$DD,$EF,$EE,$F8,$3D,$DF,$0F,$BB,$E2,$F7,$7C,$7E,$EF
	.dc.b	$93,$DD,$F2,$FB,$BE,$6F,$77,$F0,$9F,$6D,$F8,$83,$F0,$FC,$FF,$8C
	.dc.b	$7F,$DD,$83,$E1,$F7,$F0,$FC,$3A,$B8,$FC,$3F,$2F,$7B,$9F,$BD,$D3
	.dc.b	$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$7B,$C3,$EF,$78,$BD,$EF,$1F
	.dc.b	$BD,$E4,$F7,$BC,$BE,$F7,$9B,$DE,$F3,$FB,$DF,$9B,$FB,$7F,$C3,$1F
	.dc.b	$85,$3F,$87,$73,$7E,$91,$FF,$76,$9F,$EA,$73,$FC,$36,$7F,$86,$92
	.dc.b	$F6,$7A,$7B,$3D,$7D,$9E,$DE,$CF,$7F,$67,$C1,$EC,$F8,$7D,$9F,$17
	.dc.b	$B3,$E3,$F6,$7C,$9E,$CF,$97,$D9,$F3,$7B,$3E,$7F,$67,$D1,$EC,$FA
	.dc.b	$7D,$9F,$DA,$7D,$C7,$E1,$CF,$E1,$FC,$7F,$BC,$5E,$CE,$BD,$9D,$FB
	.dc.b	$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B
	.dc.b	$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF
	.dc.b	$9B,$D9,$F3,$FB,$3F,$C2,$F8,$7F,$87,$9F,$88,$E6,$FF,$7F,$FC,$16
	.dc.b	$7D,$9D,$7B,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3
	.dc.b	$D7,$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67
	.dc.b	$C9,$EC,$F9,$7D,$9F,$37,$B3,$FC,$87,$EE,$7F,$86,$1F,$C2,$CF,$C3
	.dc.b	$F3,$17,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE,$FF,$39,$FB,$AF
	.dc.b	$E1,$87,$E1,$B9,$CF,$E1,$9F,$E8,$6B,$DC,$E1,$EE,$71,$F3,$79,$7B
	.dc.b	$9C,$FD,$CE,$9E,$E7,$5F,$73,$B7,$B9,$DF,$DC,$F0,$7B,$9E,$1F,$73
	.dc.b	$C5,$EE,$78,$FD,$CF,$27,$B9,$E5,$F7,$3C,$DE,$E7,$9F,$DC,$F4,$7B
	.dc.b	$9F,$D4,$7E,$EF,$FA,$A9,$F8,$71,$EE,$E7,$DD,$D7,$BB,$BF,$77,$87
	.dc.b	$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B,$BD,$7D,$DE,$DE,$EF,$7F,$77
	.dc.b	$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7,$7C,$9E,$EF,$97,$DD,$F3,$7B
	.dc.b	$BF,$D9,$FD,$9F,$EA,$47,$F5,$13,$F0,$E7,$E1,$BC,$AF,$EE,$2B,$DB
	.dc.b	$E1,$ED,$F1,$8E,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$37
	.dc.b	$B7,$E7,$F6,$FD,$1E,$DF,$F7,$FF,$BC,$FE,$A4,$7E,$1C,$FC,$37,$29
	.dc.b	$FC,$33,$FE,$0A,$5E,$D7,$1F,$6B,$97,$B5,$CF,$DA,$E9,$ED,$75,$F6
	.dc.b	$BB,$7B,$5D,$FD,$AF,$07,$B5,$E1,$F6,$BC,$5E,$D7,$8F,$DA,$F2,$7B
	.dc.b	$5E,$5F,$6B,$CD,$ED,$79,$FD,$AF,$47,$B5,$FE,$17,$FB,$DF,$F0,$D9
	.dc.b	$FC,$34,$FC,$3F,$2F,$F8,$81,$FE,$A8,$DF,$63,$87,$B1,$C7,$D8,$E5
	.dc.b	$EC,$73,$F6,$3A,$7B,$0F,$E0,$18,$5C,$7A,$43,$EF,$BF,$C5,$BF,$04
	.dc.b	$E7,$E3,$18,$6D,$7F,$36,$6C,$EE,$39,$CE,$DA,$6C,$B1,$B5,$AA,$FC
	.dc.b	$51,$B3,$2E,$27,$84,$EE,$BB,$56,$9B,$19,$8E,$4D,$9E,$35,$CE,$FD
	.dc.b	$24,$AF,$2B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD,$F8,$A7,$73,$9A,$E9
	.dc.b	$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57,$72,$47,$9D,$B7,$E4
	.dc.b	$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C,$73,$B6,$9B,$2C,$6D
	.dc.b	$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E,$EB,$C1,$5A,$6C,$66
	.dc.b	$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C,$A3,$83,$A2,$1A,$9E
	.dc.b	$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F,$46,$BC,$4C,$46,$C9
	.dc.b	$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C,$F0,$6C,$ED,$B3,$B8
	.dc.b	$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD,$71,$C4,$D3,$89,$E5
	.dc.b	$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF,$D6,$4A,$F3,$B5,$E8
	.dc.b	$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE,$97,$E3,$6C,$EA,$31
	.dc.b	$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E,$48,$C1,$1C,$12,$FC
	.dc.b	$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6,$DF,$51,$2F,$DA,$4B
	.dc.b	$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6,$63,$93,$67,$8D,$73
	.dc.b	$BF,$59,$2B,$CE,$D7,$A1,$CA,$38,$3A,$21,$A9,$EB,$7E,$29,$DC,$E6
	.dc.b	$BA,$5F,$8D,$B3,$A8,$C4,$F4,$6B,$C4,$C4,$6C,$95,$DC,$91,$E7,$6D
	.dc.b	$F9,$23,$04,$70,$4B,$F1,$CF,$06,$CE,$DB,$3B,$8F,$1C,$ED,$A6,$CB
	.dc.b	$1B,$7D,$44,$BF,$69,$2F,$D7,$1C,$4D,$38,$9E,$53,$BA,$F0,$56,$9B
	.dc.b	$19,$8E,$4D,$9E,$35,$CE,$FD,$64,$AF,$3B,$5E,$87,$28,$E0,$E8,$86
	.dc.b	$A7,$AD,$F8,$A7,$73,$9A,$E9,$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11
	.dc.b	$B2,$57,$72,$47,$9D,$B7,$E4,$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C
	.dc.b	$EE,$3C,$73,$B6,$9B,$2C,$6D,$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2
	.dc.b	$79,$4E,$EB,$C1,$5A,$6C,$66,$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED
	.dc.b	$7A,$1C,$A3,$83,$A2,$1A,$9E,$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A
	.dc.b	$8C,$4F,$46,$BC,$4C,$46,$C9,$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04
	.dc.b	$BF,$1C,$F0,$6C,$ED,$B3,$B8,$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6
	.dc.b	$92,$FD,$71,$C4,$D3,$89,$E5,$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3
	.dc.b	$5C,$EF,$D6,$4A,$F3,$B5,$E8,$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77
	.dc.b	$39,$AE,$97,$E3,$6C,$EA,$31,$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79
	.dc.b	$DB,$7E,$48,$C1,$1C,$12,$FC,$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69
	.dc.b	$B2,$C6,$DF,$51,$2F,$DA,$4B,$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15
	.dc.b	$A6,$C6,$63,$93,$67,$8D,$73,$BF,$59,$2B,$CE,$D7,$A1,$CA,$38,$3A
	.dc.b	$21,$A9,$EB,$7E,$29,$DC,$E6,$BA,$5F,$8D,$B3,$A8,$C4,$F4,$6B,$C4
	.dc.b	$C4,$6C,$95,$DC,$91,$E7,$6D,$F9,$23,$04,$70,$4B,$F1,$CF,$06,$CE
	.dc.b	$DB,$3B,$8F,$1C,$ED,$A6,$CB,$1B,$7D,$44,$BF,$69,$2F,$D7,$1C,$4D
	.dc.b	$38,$9E,$53,$BA,$F0,$56,$9B,$19,$8E,$4D,$9E,$35,$CE,$FD,$64,$AF
	.dc.b	$3B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD,$F8,$A7,$73,$9A,$E9,$7E,$36
	.dc.b	$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57,$72,$47,$9D,$B7,$E4,$8C,$11
	.dc.b	$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C,$73,$B6,$9B,$2C,$6D,$F5,$12
	.dc.b	$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E,$EB,$C1,$5A,$6C,$66,$39,$36
	.dc.b	$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C,$A3,$83,$A2,$1A,$9E,$B7,$E2
	.dc.b	$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F,$46,$BC,$4C,$46,$C9,$5D,$C9
	.dc.b	$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C,$F0,$6C,$ED,$B3,$B8,$F1,$CE
	.dc.b	$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD,$71,$C4,$D3,$89,$E5,$3B,$AF
	.dc.b	$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF,$D6,$4A,$F3,$B5,$E8,$72,$8E
	.dc.b	$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE,$97,$E3,$6C,$EA,$31,$3D,$1A
	.dc.b	$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E,$48,$C1,$1C,$12,$FC,$73,$C1
	.dc.b	$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6,$DF,$51,$2F,$DA,$4B,$F5,$C7
	.dc.b	$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6,$63,$93,$67,$8D,$73,$BF,$59
	.dc.b	$2B,$CE,$D7,$A1,$CA,$38,$3A,$21,$A9,$EB,$7E,$29,$DC,$E6,$BA,$5F
	.dc.b	$8D,$B3,$A8,$C4,$F4,$6B,$C4,$C4,$6C,$95,$DC,$91,$E7,$6D,$F9,$23
	.dc.b	$04,$70,$4B,$F1,$CF,$06,$CE,$DB,$3B,$8F,$1C,$ED,$A6,$CB,$1B,$7D
	.dc.b	$44,$BF,$69,$2F,$D7,$1C,$4D,$38,$9E,$53,$BA,$F0,$56,$9B,$19,$8E
	.dc.b	$4D,$9E,$35,$CE,$FD,$64,$AF,$3B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD
	.dc.b	$F8,$A7,$73,$9A,$E9,$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57
	.dc.b	$72,$47,$9D,$B7,$E4,$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C
	.dc.b	$73,$B6,$9B,$2C,$6D,$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E
	.dc.b	$EB,$C1,$5A,$6C,$66,$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C
	.dc.b	$A3,$83,$A2,$1A,$9E,$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F
	.dc.b	$46,$BC,$4C,$46,$C9,$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C
	.dc.b	$F0,$6C,$ED,$B3,$B8,$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD
	.dc.b	$71,$C4,$D3,$89,$E5,$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF
	.dc.b	$D6,$4A,$F3,$B5,$E8,$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE
	.dc.b	$97,$E3,$6C,$EA,$31,$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E
	.dc.b	$48,$C1,$1C,$12,$FC,$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6
	.dc.b	$DF,$51,$2F,$DA,$4B,$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6
	.dc.b	$63,$93,$67,$8D,$73,$BF,$59,$28,$29,$2B,$CD,$AF,$77,$BA,$D2,$9D
	.dc.b	$3A,$1F,$71,$DB,$E4,$9F,$19,$99,$C8,$8C,$BC,$67,$84,$ED,$B1,$93
	.dc.b	$84,$65,$A6,$67,$9B,$E8,$9E,$2D,$36,$58,$DB,$F1,$23,$6D,$3A,$E2
	.dc.b	$FD,$05,$BB,$E4,$3E,$5F,$DF,$DB,$F9,$55,$A9,$C9,$F4,$2F,$97,$0E
	.dc.b	$EF,$84,$75,$C3,$15,$BB,$70,$E3,$1E,$AD,$BF,$26,$D6,$B7,$7E,$0B
	.dc.b	$F1,$FC,$0E,$7E,$07,$4D,$9D,$61,$ED,$F8,$1E,$5F,$03,$CD,$B3,$AC
	.dc.b	$5F,$03,$DF,$E0,$7C,$0D,$9D,$7C,$0F,$8C,$F2,$7C,$0D,$8E,$5F,$03
	.dc.b	$E7,$F8,$1F,$43,$63,$B7,$C0,$FA,$CF,$77,$E8,$5F,$98,$8C,$9E,$85
	.dc.b	$F2,$9A,$F8,$FD,$B6,$75,$F1,$5C,$BE,$17,$9B,$63,$97,$C2,$F6,$F8
	.dc.b	$5E,$ED,$8E,$DF,$0B,$E2,$F8,$5F,$1B,$63,$C5,$F0,$BE,$63,$CF,$F0
	.dc.b	$B6,$3C,$DF,$0B,$EA,$F8,$5F,$5B,$63,$D5,$F0,$BE,$EF,$85,$F7,$B6
	.dc.b	$3D,$DF,$0B,$F0,$8F,$87,$F0,$B6,$3E,$17,$C2,$FC,$63,$E3,$FC,$2D
	.dc.b	$8F,$8D,$F0,$BF,$28,$F9,$7F,$0B,$63,$E5,$7C,$2F,$E2,$FF,$43,$FC
	.dc.b	$E4,$60,$FC,$2B,$8F,$1D,$E5,$FE,$62,$1F,$21,$C4,$E5,$E8,$DB,$E3
	.dc.b	$F1,$7D,$4E,$DF,$0F,$63,$AF,$C3,$F8,$7E,$1F,$C4,$D8,$F0,$FC,$3F
	.dc.b	$94,$F3,$7C,$3D,$8F,$2F,$C3,$FA,$4F,$57,$B3,$63,$D3,$F0,$FE,$DF
	.dc.b	$87,$F7,$36,$3D,$BF,$0F,$F0,$7E,$1F,$E1,$36,$3E,$0F,$C3,$FC,$5F
	.dc.b	$87,$F8,$CD,$8F,$8B,$F0,$FF,$24,$F9,$5F,$0F,$63,$E4,$FC,$3F,$CD
	.dc.b	$3E,$77,$C3,$D8,$FD,$8F,$AB,$FC,$8D,$BF,$05,$FB,$D8,$FE,$53,$18
	.dc.b	$37,$F1,$59,$35,$F1,$56,$39,$FC,$4F,$1F,$89,$E4,$D8,$EF,$F0,$9D
	.dc.b	$4E,$DF,$0B,$63,$C7,$F0,$BE,$13,$C5,$F0,$B6,$3C,$FF,$09,$E5,$3C
	.dc.b	$DF,$0B,$63,$D7,$F0,$7E,$93,$D5,$F0,$96,$3D,$FB,$FB,$4F,$77,$C1
	.dc.b	$D8,$F8,$7F,$07,$F0,$4F,$85,$F0,$76,$3E,$3F,$C1,$7C,$53,$E3,$7C
	.dc.b	$1D,$8F,$97,$F0,$7F,$C9,$7F,$44,$FE,$11,$E7,$7E,$62,$7F,$99,$66
	.dc.b	$AD,$4E,$FE,$0B,$87,$C1,$DB,$FE,$4B,$82,$3A,$7C,$17,$5F,$83,$B5
	.dc.b	$A7,$C1,$F0,$5E,$1F,$83,$B3,$C9,$F2,$7C,$17,$97,$E0,$EC,$F6,$7D
	.dc.b	$1F,$07,$23,$C3,$F0,$77,$6F,$51,$FD,$DF,$57,$C1,$7B,$FE,$0F,$E0
	.dc.b	$36,$FD,$3F,$05,$F0,$FE,$0F,$E2,$36,$3D,$BF,$07,$F1,$FE,$0F,$E4
	.dc.b	$36,$3E,$0F,$C1,$FC,$BF,$83,$F9,$8D,$8F,$8B,$F0,$7F,$3C,$FA,$1F
	.dc.b	$07,$63,$E4,$FC,$1F,$D3,$3F,$AC,$7E,$8B,$FC,$2F,$E1,$3F,$B0,$F0
	.dc.b	$27,$3F,$07,$AF,$83,$B8,$7F,$56,$C3,$FC,$20,$DF,$83,$E6,$74,$F8
	.dc.b	$3B,$1B,$F8,$2E,$FF,$07,$E0,$6C,$73,$F8,$3F,$19,$E4,$F8,$3B,$1D
	.dc.b	$FE,$0F,$CF,$F0,$7E,$86,$C7,$8F,$E0,$BD,$7F,$07,$EC,$6C,$79,$FE
	.dc.b	$0B,$DF,$F0,$7F,$01,$B1,$EB,$F8,$2F,$87,$F0,$7F,$11,$B1,$EF,$F8
	.dc.b	$3F,$8E,$7C,$8F,$83,$B1,$F0,$FE,$0F,$E5,$FC,$1F,$CC,$6C,$75,$3E
	.dc.b	$46,$1D,$5A,$75,$27,$8B,$F8,$CF,$87,$73,$B1,$EF,$39,$FD,$1B,$F8
	.dc.b	$B3,$7E,$17,$FC,$25,$1F,$C5,$A3,$18,$7F,$83,$21,$F0,$BA,$F8,$5D
	.dc.b	$B6,$75,$F0,$9C,$BE,$17,$9B,$63,$97,$C2,$F6,$F8,$5E,$ED,$8E,$DF
	.dc.b	$0B,$E2,$F8,$5F,$1B,$63,$C5,$F0,$BE,$6F,$85,$F3,$B6,$3C,$DF,$0B
	.dc.b	$EA,$3A,$6F,$EF,$79,$C7,$AF,$0B,$3F,$0D,$E7,$1E,$FB,$B3,$F1,$DF
	.dc.b	$8B,$F0,$BF,$1B,$E0,$2D,$7C,$57,$E4,$FC,$2C,$8F,$91,$F0,$B7,$0F
	.dc.b	$A1,$F0,$BF,$38,$F9,$FF,$0B,$6B,$E7,$7C,$2F,$D2,$3E,$9F,$C2,$D8
	.dc.b	$FA,$58,$F3,$F4,$7F,$C9,$FD,$0C,$F1,$5F,$E5,$58,$65,$D7,$3F,$C0
	.dc.b	$67,$70,$FF,$2B,$F8,$7F,$CB,$1B,$3F,$E5,$38,$FF,$19,$E1,$FE,$4C
	.dc.b	$74,$F5,$CB,$BC,$FF,$A3,$F1,$F1,$4F,$98,$FC,$46,$3D,$DF,$26,$36
	.dc.b	$BF,$D2,$D8,$7F,$33,$9E,$DF,$07,$E1,$F8,$4F,$13,$E3,$F8,$4F,$27
	.dc.b	$C2,$5A,$F2,$3E,$7F,$84,$91,$E7,$F8,$4B,$97,$E8,$F2,$7C,$27,$B7
	.dc.b	$E1,$3D,$CD,$BF,$76,$5D,$9A,$76,$3E,$16,$3D,$38,$7E,$4E,$56,$7F
	.dc.b	$D0,$98,$BE,$13,$E5,$1F,$2F,$E0,$EC,$FC,$BF,$84,$F9,$DF,$07,$F3
	.dc.b	$DB,$1F,$3F,$E1,$3E,$91,$F4,$FE,$0E,$C7,$D3,$F8,$4F,$AD,$F0,$7F
	.dc.b	$5D,$B1,$D4,$FB,$18,$75,$69,$D4,$9C,$5F,$16,$FA,$BF,$1F,$E1,$3F
	.dc.b	$05,$F0,$9F,$83,$6D,$7E,$0F,$E1,$3F,$0D,$F0,$7F,$87,$6C,$7F,$A4
	.dc.b	$BF,$47,$FE,$88,$FF,$A9,$DF,$F2,$6B,$93,$FD,$40,$E3,$0F,$E3,$0F
	.dc.b	$F4,$46,$FC,$1F,$81,$B3,$E1,$3F,$29,$7E,$16,$9D,$4F,$17,$C2,$76
	.dc.b	$7B,$BE,$0A,$F0,$FC,$25,$9F,$43,$CE,$FF,$11,$77,$9B,$F4,$EF,$3A
	.dc.b	$EA,$77,$9F,$4E,$1E,$CF,$84,$F6,$FC,$1D,$AF,$88,$FC,$0F,$84,$91
	.dc.b	$F1,$FE,$12,$E1,$F0,$BE,$13,$E2,$FC,$1F,$C6,$6D,$72,$3E,$77,$C2
	.dc.b	$76,$69,$F4,$70,$E7,$AF,$FB,$53,$81,$96,$2B,$EA,$61,$D3,$2F,$04
	.dc.b	$7D,$9C,$BE,$AF,$C1,$C8,$FB,$9F,$09,$6B,$EB,$FC,$27,$DA,$EF,$FB
	.dc.b	$6D,$AF,$C1,$7C,$27,$DE,$3E,$FF,$C0,$D8,$FC,$37,$C2,$7E,$08,$E9
	.dc.b	$BF,$9D,$E7,$1F,$8B,$C2,$E3,$F8,$67,$93,$F8,$8C,$3A,$CF,$80,$FC
	.dc.b	$97,$C0,$7E,$38,$FC,$7F,$C1,$D8,$FC,$B7,$C2,$7F,$BC,$7F,$48,$FE
	.dc.b	$C7,$9F,$C7,$8C,$7C,$1E,$7E,$33,$5F,$07,$70,$FF,$77,$E1,$FD,$91
	.dc.b	$BF,$09,$CC,$ED,$AF,$F7,$39,$47,$E8,$31,$EB,$5E,$23,$9E,$3D,$6F
	.dc.b	$86,$17,$25,$25,$F0,$CC,$BC,$51,$DF,$2F,$67,$C2,$78,$EF,$CE,$F8
	.dc.b	$5E,$E7,$86,$7C,$D9,$77,$8E,$D8,$75,$CB,$15,$E0,$C7,$BE,$5E,$48
	.dc.b	$C7,$C2,$FC,$BF,$84,$F9,$8D,$C7,$E3,$FC,$2F,$CF,$F8,$5F,$A0,$D8
	.dc.b	$F9,$7F,$07,$F4,$FE,$17,$EA,$36,$3E,$7F,$C2,$7D,$73,$A6,$5F,$35
	.dc.b	$E7,$1F,$53,$0B,$3F,$45,$E7,$1F,$62,$EC,$FD,$57,$F0,$5F,$0B,$F8
	.dc.b	$3F,$85,$B5,$F7,$9F,$C3,$7C,$2F,$E1,$FE,$0A,$CF,$E0,$9F,$C5,$7C
	.dc.b	$1C,$8F,$C2,$FC,$15,$CB,$F0,$5F,$05,$F8,$F3,$F2,$1F,$0B,$6B,$F1
	.dc.b	$7F,$0B,$F9,$33,$E6,$72,$7C,$DE,$77,$E9,$3F,$DC,$B1,$FA,$55,$AB
	.dc.b	$FA,$F3,$FD,$E2,$7F,$C8,$4D,$BE,$4F,$F6,$F8,$CF,$C4,$F1,$7F,$17
	.dc.b	$0F,$85,$C6,$3E,$06,$C3,$E0,$AC,$63,$E1,$73,$F0,$77,$3F,$CB,$67
	.dc.b	$E0,$8E,$3F,$07,$27,$27,$A3,$E1,$31,$A7,$A3,$E1,$7B,$9D,$A7,$71
	.dc.b	$E6,$8E,$CF,$BA,$EE,$1D,$EB,$7A,$FC,$8C,$7A,$CF,$C2,$C6,$D7,$C8
	.dc.b	$F8,$5F,$6F,$C2,$FB,$9B,$1C,$B7,$7C,$2F,$C2,$3E,$1F,$C1,$59,$D7
	.dc.b	$C2,$FC,$6F,$83,$F8,$ED,$8E,$5F,$09,$F2,$8F,$97,$F0,$B6,$3B,$7C
	.dc.b	$27,$CE,$3E,$7F,$C2,$D8,$F1,$7C,$2F,$D2,$3E,$9F,$C1,$D8,$F7,$E5
	.dc.b	$FD,$C3,$97,$F8,$2B,$79,$0F,$F6,$FE,$3F,$C6,$39,$3F,$B8,$38,$F8
	.dc.b	$4F,$C0,$C6,$2F,$F8,$1A,$8C,$FC,$1E,$BE,$0E,$D7,$89,$E3,$F0,$5C
	.dc.b	$BE,$06,$CF,$99,$EB,$F0,$3D,$BE,$06,$CF,$A9,$F0,$FC,$1F,$8B,$E0
	.dc.b	$AC,$FB,$9F,$2F,$C1,$48,$F8,$3F,$05,$78,$7F,$D4,$5B,$BE,$03,$D7
	.dc.b	$F0,$1E,$C6,$DF,$C8,$F8,$2F,$7F,$C1,$7C,$06,$C7,$CC,$F8,$2F,$86
	.dc.b	$7C,$4F,$82,$B1,$F4,$3E,$07,$F4,$FF,$A5,$7F,$4A,$8F,$8F,$18,$F8
	.dc.b	$0C,$FC,$1E,$BE,$06,$D7,$E5,$AF,$F9,$5B,$FC,$C1,$FE,$8E,$E7,$E0
	.dc.b	$FA,$FC,$1F,$67,$BF,$C0,$FE,$E7,$89,$FB,$EF,$EF,$7B,$FF,$84,$62
	.dc.b	$AD,$4E,$71,$ED,$F0,$7E,$8F,$83,$F4,$B6,$3C,$9F,$03,$EC,$3D,$BF
	.dc.b	$07,$63,$F6,$79,$F8,$3D,$2F,$17,$C3,$BE,$13,$BC,$61,$F8,$BF,$09
	.dc.b	$F1,$9B,$7F,$05,$F9,$3F,$09,$23,$E2,$7C,$25,$C5,$C3,$FD,$3D,$64
	.dc.b	$BF,$09,$EE,$F4,$EE,$1F,$37,$E1,$3E,$AF,$C2,$7D,$66,$C7,$D1,$F8
	.dc.b	$4F,$B2,$7D,$AF,$84,$B1,$EE,$C5,$A7,$B8,$F9,$7E,$A5,$9F,$B5,$F0
	.dc.b	$9F,$82,$F8,$4F,$C1,$B6,$3E,$F7,$C2,$7E,$18,$FE,$5B,$FA,$67,$F3
	.dc.b	$BF,$51,$D8,$9F,$97,$B4,$BF,$9F,$FB,$F7,$EF,$78,$FE,$0C,$F2,$F8
	.dc.b	$69,$5F,$C8,$32,$FB,$D5,$FE,$06,$7F,$A1,$47,$F7,$C7,$F9,$F3,$BF
	.dc.b	$85,$F1,$FC,$37,$91,$FB,$7C,$92,$7C,$3E,$B5,$C3,$E9,$C7,$3F,$86
	.dc.b	$F5,$9E,$CF,$85,$B1,$E3,$ED,$F7,$FC,$2F,$C0,$6C,$7C,$DE,$9F,$C4
	.dc.b	$9F,$9B,$E0,$7A,$B8,$9F,$67,$C3,$7C,$83,$E4,$FC,$2D,$8F,$81,$F0
	.dc.b	$DF,$33,$BB,$E6,$B6,$3E,$D7,$77,$E5,$8E,$DE,$A5,$8F,$91,$F0,$BF
	.dc.b	$50,$FA,$BF,$0B,$63,$E6,$7A,$FF,$63,$E1,$7E,$CB,$63,$F0,$DD,$FF
	.dc.b	$A2,$3D,$5E,$9D,$8F,$A9,$E1,$7E,$00,$FC,$0F,$C2,$D8,$FB,$1F,$0B
	.dc.b	$F8,$43,$F0,$BF,$0B,$63,$F5,$FF,$BF,$BD,$4F,$E0,$0D,$8C,$FA,$78
	.dc.b	$9F,$C6,$FC,$34,$9F,$EC,$7F,$A7,$FF,$43,$F1,$13,$FD,$FA,$3F,$A2
	.dc.b	$4F,$F9,$01,$FE,$81,$E0,$7F,$93,$9C,$B7,$CC,$FD,$FF,$97,$2B,$CB
	.dc.b	$E8,$5A,$FC,$29,$C7,$E1,$FB,$1D,$FE,$1E,$D7,$93,$C6,$F1,$7C,$3F
	.dc.b	$8D,$B1,$FB,$1F,$37,$F8,$69,$F0,$3C,$9B,$1E,$CF,$87,$F5,$1E,$BF
	.dc.b	$87,$B1,$F0,$3E,$1F,$DD,$E3,$7B,$DB,$1F,$C6,$BE,$DF,$FC,$59,$B1
	.dc.b	$BC,$FF,$B3,$9D,$4F,$C7,$F8,$89,$1F,$33,$E2,$2C,$FC,$0F,$55,$CC
	.dc.b	$F8,$8F,$9D,$F1,$1F,$3D,$B8,$78,$7D,$5F,$E7,$5A,$7E,$DB,$1F,$F3
	.dc.b	$F7,$57,$F8,$BE,$B4,$7D,$AF,$23,$EC,$1F,$67,$E1,$EC,$7D,$EF,$88
	.dc.b	$FB,$9F,$11,$F7,$5B,$1E,$FF,$8D,$EE,$7F,$49,$C1,$E8,$E2,$7F,$0F
	.dc.b	$F1,$3F,$85,$3F,$0D,$E4,$D8,$FF,$16,$F5,$7F,$5F,$D7,$FD,$26,$BE
	.dc.b	$23,$FC,$92,$E7,$90,$E6,$BE,$87,$B5,$F0,$0F,$17,$9F,$63,$AF,$C4
	.dc.b	$72,$39,$FC,$45,$8F,$0F,$93,$DB,$CA,$EE,$D8,$FB,$9E,$EF,$CC,$3D
	.dc.b	$DE,$6D,$8F,$EB,$3C,$4F,$37,$E1,$1E,$8F,$86,$B3,$EE,$F8,$7F,$5F
	.dc.b	$C3,$FB,$1B,$1F,$89,$F8,$0F,$B2,$7F,$73,$FE,$00,$FD,$3F,$2F,$E2
	.dc.b	$E3,$33,$B3,$FB,$4C,$FE,$BA,$75,$1F,$33,$E2,$3E,$4F,$A1,$F2,$9B
	.dc.b	$1F,$97,$F8,$2F,$C3,$9D,$FC,$EB,$19,$25,$FE,$C8,$60,$F3,$F7,$1F
	.dc.b	$97,$E8,$FE,$F1,$5F,$73,$E1,$BE,$CF,$C4,$7D,$A7,$F2,$BF,$07,$27
	.dc.b	$D3,$E7,$5C,$3B,$78,$EA,$C7,$C4,$7F,$77,$D3,$F9,$DE,$1F,$AF,$9F
	.dc.b	$C2,$FC,$47,$E1,$BE,$1E,$CF,$E3,$1F,$D1,$FC,$3C,$8F,$8F,$E6,$DC
	.dc.b	$FB,$CD,$CB,$B7,$9D,$6F,$FD,$B3,$C1,$F9,$07,$E7,$F9,$9F,$94,$8F
	.dc.b	$AC,$FE,$C3,$C9,$93,$F5,$3C,$BB,$97,$B2,$6E,$3E,$BF,$32,$DF,$FB
	.dc.b	$CE,$BE,$6F,$97,$FB,$AF,$85,$FC,$13,$67,$F8,$74,$59,$FA,$9C,$4F
	.dc.b	$33,$18,$FF,$87,$79,$17,$8C,$3B,$7F,$00,$FE,$BF,$1F,$E2,$D3,$FC
	.dc.b	$23,$E1,$65,$7F,$8D,$2B,$F5,$D5,$F5,$27,$2D,$9D,$79,$58,$C7,$C1
	.dc.b	$56,$7B,$F9,$56,$C7,$F2,$5F,$86,$FE,$2E,$7F,$94,$71,$FE,$A5,$1F
	.dc.b	$82,$8C,$D7,$F2,$D7,$5E,$5E,$27,$53,$93,$D1,$E3,$67,$2F,$85,$F0
	.dc.b	$BF,$C9,$0E,$DE,$35,$AF,$E7,$11,$6F,$F3,$D9,$3C,$7C,$61,$CA,$F6
	.dc.b	$7C,$5F,$17,$2F,$D9,$E1,$79,$BF,$CC,$8F,$67,$8F,$67,$FA,$34,$5C
	.dc.b	$BB,$F8,$97,$1F,$B1,$36,$BE,$BF,$89,$6F,$E4,$E3,$FE,$B2,$9F,$EA
	.dc.b	$1F,$09,$F2,$7C,$3F,$E8,$2C,$8C,$9F,$6F,$C3,$C5,$FE,$BA,$6E,$5F
	.dc.b	$E6,$3D,$FF,$3D,$39,$BF,$EB,$9F,$0B,$FD,$30,$FA,$DF,$11,$63,$24
	.dc.b	$9F,$93,$F0,$F9,$7F,$77,$57,$0E,$DF,$0F,$6F,$E4,$7C,$3F,$F5,$83
	.dc.b	$F0,$5F,$09,$63,$FB,$64,$5C,$3E,$AF,$C1,$5C,$3F,$3D,$36,$BF,$55
	.dc.b	$F3,$1F,$DC,$B8,$53,$FD,$93,$E2,$25,$7F,$A4,$2B,$FB,$36,$3F,$DD
	.dc.b	$E7,$2F,$F1,$27,$FD,$49,$1F,$E6,$28,$D4,$FF,$7F,$9D,$E1,$FE,$8B
	.dc.b	$AF,$F3,$1D,$71,$AF,$F0,$27,$AB,$FD,$D0,$E7,$EB,$58,$C9,$27,$F9
	.dc.b	$75,$3D,$4D,$47,$F2,$4B,$B8,$77,$F5,$2E,$1F,$91,$F5,$7F,$C0,$27
	.dc.b	$A3,$D2,$B1,$FE,$15,$8B,$87,$6F,$46,$E1,$FC,$1E,$6D,$7C,$AF,$36
	.dc.b	$DF,$E4,$3D,$6F,$F0,$69,$F1,$3C,$AB,$5F,$E2,$18,$B3,$EA,$F2,$6E
	.dc.b	$3F,$C4,$E6,$D7,$D6,$F2,$2D,$FF,$04,$FB,$9F,$F4,$A1,$FE,$95,$E1
	.dc.b	$78,$DF,$ED,$58,$CD,$D3,$EA,$F8,$55,$FF,$65,$C6,$9F,$BD,$DF,$9C
	.dc.b	$3F,$AD,$4F,$F6,$C3,$FB,$6E,$0E,$CE,$DF,$23,$30,$E6,$FE,$0B,$A7
	.dc.b	$18,$72,$8B,$B7,$E0,$7A,$6D,$FF,$3B,$CB,$FB,$A1,$FD,$D7,$27,$3F
	.dc.b	$AF,$E2,$65,$7A,$1F,$C7,$F2,$E3,$3F,$3C,$5C,$BF,$1D,$C5,$6F,$B7
	.dc.b	$C4,$FC,$50,$FF,$2E,$9B,$6F,$77,$F9,$8C,$EC,$F8,$F5,$67,$FC,$A0
	.dc.b	$7C,$57,$15,$FE,$E7,$D1,$1D,$23,$AD,$FF,$3C,$C7,$78,$7C,$1F,$89
	.dc.b	$FB,$06,$DF,$E8,$33,$FD,$0A,$F3,$1F,$62,$6E,$3E,$26,$E1,$E1,$6C
	.dc.b	$F2,$97,$B9,$87,$F1,$4F,$E1,$CF,$03,$F0,$4F,$F0,$C4,$7E,$1A,$34
	.dc.b	$D3,$E4,$1D,$7E,$27,$33,$F8,$89,$B7,$F0,$21,$F1,$38,$BF,$C0,$7B
	.dc.b	$DF,$97,$3C,$39,$FF,$60,$F7,$B1,$3F,$96,$9B,$7F,$43,$E2,$6E,$3F
	.dc.b	$3E,$6D,$7B,$FE,$22,$DF,$E9,$F4,$FD,$21,$D9,$7C,$46,$2B,$F4,$73
	.dc.b	$67,$EE,$FC,$0D,$C7,$EE,$4D,$AF,$99,$F0,$36,$FF,$69,$B7,$EC,$4E
	.dc.b	$DF,$0D,$6B,$F5,$F1,$67,$F1,$3F,$03,$71,$FC,$44,$DA,$FB,$1F,$03
	.dc.b	$6F,$FC,$07,$F7,$2F,$73,$EB,$C2,$31,$5F,$52,$3E,$D7,$C1,$6A,$73
	.dc.b	$1F,$98,$F8,$2F,$84,$7C,$3F,$81,$B7,$C7,$9B,$F4,$5F,$1F,$F5,$1F
	.dc.b	$F4,$61,$FE,$1E,$D7,$F3,$6F,$F8,$97,$B1,$D1,$FC,$44,$E1,$F0,$57
	.dc.b	$0F,$E1,$F1,$73,$FF,$5C,$5B,$E0,$B1,$8F,$EA,$6A,$DF,$E8,$AB,$FC
	.dc.b	$3F,$39,$8F,$E3,$3D,$5F,$C7,$CF,$37,$05,$8C,$93,$1E,$FF,$05,$97
	.dc.b	$F8,$ED,$5B,$F5,$FC,$15,$BF,$E2,$9D,$7F,$CA,$CF,$85,$F0,$56,$3F
	.dc.b	$94,$C5,$C3,$D7,$F0,$57,$0F,$E0,$53,$6B,$D5,$F0,$56,$FF,$9D,$77
	.dc.b	$7F,$37,$3E,$77,$C1,$5A,$FE,$6B,$16,$7E,$3F,$C1,$5C,$7F,$87,$4D
	.dc.b	$AF,$8D,$F0,$56,$FF,$A3,$78,$3F,$D0,$CF,$B5,$F0,$56,$BF,$A0,$C5
	.dc.b	$9F,$A7,$F0,$97,$1F,$E3,$53,$6B,$E9,$7C,$15,$BF,$EA,$5E,$1F,$F4
	.dc.b	$F3,$F0,$DF,$05,$6B,$FA,$6C,$59,$FB,$FF,$05,$71,$FE,$4D,$36,$BE
	.dc.b	$F7,$C1,$5B,$FE,$BD,$E3,$7F,$5B,$3F,$25,$F0,$56,$BF,$AC,$C5,$96
	.dc.b	$1F,$CB,$8F,$E7,$2F,$F5,$A8,$D4,$7F,$35,$7F,$87,$BF,$E1,$CE,$F4
	.dc.b	$7F,$69,$F1,$E5,$FE,$8E,$BF,$C3,$94,$9C,$DF,$F6,$67,$9F,$C3,$62
	.dc.b	$BF,$A1,$45,$BF,$E7,$70,$F8,$2C,$5F,$F7,$BF,$23,$FB,$C1,$E4,$F8
	.dc.b	$2B,$7F,$DD,$A2,$CF,$BA,$39,$FC,$16,$30,$FE,$A1,$56,$7B,$FC,$15
	.dc.b	$BF,$F0,$47,$97,$FE,$05,$3E,$0F,$C1,$5A,$FF,$01,$C5,$9E,$C7,$F6
	.dc.b	$3C,$1F,$05,$97,$FA,$F4,$DA,$F6,$FC,$15,$BF,$F0,$B7,$9B,$FE,$13
	.dc.b	$3E,$7F,$C1,$5A,$FF,$08,$C5,$9E,$FE,$4F,$DA,$F8,$2C,$65,$FD,$AE
	.dc.b	$AD,$FC,$AF,$82,$B8,$7F,$88,$FD,$0F,$F1,$01,$F7,$3E,$0A,$D7,$F8
	.dc.b	$7A,$2C,$F6,$C3,$FD,$08,$FF,$87,$A3,$31,$FD,$E5,$FC,$2F,$C0,$E2
	.dc.b	$BF,$DA,$1F,$73,$FE,$C3,$1E,$9F,$E2,$BE,$0B,$F9,$9C,$66,$29,$F8
	.dc.b	$EF,$81,$B5,$FC,$D7,$25,$6E,$3F,$1D,$F0,$1A,$9F,$A2,$7F,$3D,$F4
	.dc.b	$BF,$9D,$1F,$98,$F8,$0B,$5F,$95,$F8,$0F,$ED,$3F,$87,$FF,$87,$23
	.dc.b	$07,$F3,$B7,$17,$F3,$97,$C1,$F0,$F8,$2E,$2D,$AF,$36,$7F,$CF,$DF
	.dc.b	$83,$F1,$3D,$BE,$0F,$DE,$C8,$DB,$E2,$F8,$3C,$65,$EA,$BE,$4F,$FB
	.dc.b	$71,$BD,$2C,$FF,$C2,$B8,$72,$3C,$3E,$F7,$CC,$3D,$BF,$05,$63,$24
	.dc.b	$C3,$B6,$59,$7C,$15,$79,$FF,$C6,$38,$3E,$0B,$18,$7D,$CF,$80,$FB
	.dc.b	$27,$F6,$CF,$DA,$BC,$78,$FA,$63,$33,$F1,$1E,$31,$EB,$72,$EE,$3E
	.dc.b	$87,$C1,$E3,$1F,$F2,$6C,$FE,$03,$E2,$BF,$0E,$7D,$8F,$83,$B7,$F4
	.dc.b	$22,$CF,$D7,$F8,$2B,$7C,$B8,$7F,$B1,$DF,$E5,$2F,$F0,$3E,$FE,$A3
	.dc.b	$F2,$FF,$07,$F9,$43,$D9,$B5,$8F,$BB,$17,$2F,$AD,$77,$1E,$D3,$6B
	.dc.b	$FC,$A7,$BF,$F8,$A1,$E2,$7E,$1B,$AB,$F4,$26,$F0,$FB,$CD,$7F,$3F
	.dc.b	$F0,$1F,$E6,$EF,$9C,$FF,$00,$47,$E0,$F0,$FD,$4C,$E7,$87,$D3,$87
	.dc.b	$E4,$BE,$13,$96,$E7,$F5,$E4,$8F,$CF,$7B,$97,$1F,$C8,$63,$D2,$50
	.dc.b	$CB,$C2,$FE,$CA,$74,$E7,$1F,$DF,$E1,$CF,$5F,$87,$87,$8C,$FD,$2F
	.dc.b	$C2,$7F,$00,$3F,$63,$ED,$58,$FD,$4E,$BE,$BC,$7F,$40,$D7,$C1,$1A
	.dc.b	$3C,$B9,$F9,$35,$61,$CE,$BF,$71,$EC,$5C,$BF,$05,$F0,$9F,$C4,$0F
	.dc.b	$CC,$7A,$F7,$2E,$97,$EB,$3C,$2F,$6D,$BE,$6C,$7E,$D7,$4F,$26,$67
	.dc.b	$E1,$E3,$F8,$FC,$EE,$1F,$C8,$A6,$DF,$EE,$F2,$AF,$DE,$9E,$CF,$80
	.dc.b	$F1,$1F,$00,$F0,$3A,$78,$18,$3F,$84,$57,$4C,$1F,$09,$FC,$B8,$FD
	.dc.b	$B7,$AB,$6F,$F8,$BC,$7A,$8F,$5E,$36,$3B,$57,$27,$67,$F2,$5D,$39
	.dc.b	$63,$FC,$5F,$E1,$3F,$9D,$9F,$CB,$31,$B5,$FC,$92,$3E,$39,$F7,$7C
	.dc.b	$0F,$E3,$5E,$9E,$27,$BC,$F0,$BF,$E6,$1E,$85,$DF,$F7,$5F,$07,$FD
	.dc.b	$20,$FE,$73,$E7,$DC,$3F,$98,$C7,$D4,$3A,$5F,$CA,$D5,$19,$9D,$9F
	.dc.b	$E7,$9F,$37,$FA,$B4,$7F,$54,$F8,$39,$39,$F1,$6E,$1F,$CD,$67,$F0
	.dc.b	$27,$37,$B3,$89,$FE,$86,$FF,$AD,$AD,$E2,$F2,$CF,$F8,$A7,$C1,$FF
	.dc.b	$62,$35,$CF,$70,$FE,$97,$1F,$8E,$3E,$45,$BC,$0C,$4F,$D6,$9E,$19
	.dc.b	$7F,$81,$2B,$FD,$33,$87,$6A,$FE,$AB,$F0,$7F,$DC,$0F,$F5,$1F,$ED
	.dc.b	$7F,$DD,$5C,$47,$E0,$1E,$B7,$F2,$5F,$CF,$9A,$7B,$78,$36,$FF,$97
	.dc.b	$7C,$27,$F7,$C3,$17,$FE,$02,$78,$77,$7F,$0A,$FB,$BF,$F7,$FF,$C7
	.dc.b	$F7,$AF,$CE,$46,$31,$F0,$47,$F6,$CF,$86,$E7,$1F,$D8,$DF,$F0,$49
	.dc.b	$23,$ED,$78,$76,$7F,$B5,$69,$FC,$26,$FA,$63,$FC,$1E,$31,$E7,$B2
	.dc.b	$FC,$FE,$5C,$74,$FE,$91,$57,$3F,$E4,$79,$7F,$B2,$73,$FF,$63,$1F
	.dc.b	$77,$C1,$B3,$FE,$17,$D3,$BF,$1F,$06,$DF,$F7,$BC,$3C,$39,$FF,$00
	.dc.b	$ED,$FF,$12,$F8,$5F,$87,$3F,$11,$E0,$DC,$BF,$C4,$5F,$11,$FC,$79
	.dc.b	$BA,$7E,$37,$C3,$FF,$6C,$3F,$ED,$ED,$3B,$E8,$6F,$6B,$5F,$0C,$FF
	.dc.b	$1F,$D3,$FC,$6B,$E0,$7E,$5F,$BB,$F3,$0D,$CF,$C3,$B7,$FB,$94,$FC
	.dc.b	$E7,$6D,$8F,$81,$CB,$F2,$F1,$FE,$33,$A7,$83,$3E,$FC,$DF,$D6,$B4
	.dc.b	$FF,$21,$F5,$FE,$94,$FD,$34,$DC,$7D,$FC,$5F,$EF,$33,$F5,$5D,$16
	.dc.b	$3B,$F2,$7E,$23,$4F,$06,$5F,$76,$71,$3F,$37,$AD,$D3,$FB,$33,$F6
	.dc.b	$9F,$0F,$6B,$E6,$F3,$7F,$C0,$13,$F7,$1F,$0F,$63,$BF,$0F,$B3,$3E
	.dc.b	$FD,$3F,$E5,$ED,$FD,$AF,$F6,$8E,$17,$E2,$3F,$EA,$CE,$C6,$39,$8C
	.dc.b	$37,$7C,$E5,$FE,$7C,$9C,$BF,$E4,$1D,$7F,$CC,$E6,$AF,$78,$76,$F9
	.dc.b	$0C,$E3,$E6,$E1,$FD,$76,$BE,$6C,$7F,$21,$3F,$11,$FE,$49,$8F,$07
	.dc.b	$C1,$F1,$17,$2F,$F1,$46,$FF,$E7,$63,$C1,$7E,$9F,$88,$C3,$DF,$7F
	.dc.b	$CF,$65,$DF,$1F,$F3,$BF,$07,$C4,$7F,$CB,$58,$70,$7F,$CB,$D9,$DC
	.dc.b	$BF,$C7,$F9,$FC,$09,$37,$C1,$FF,$0D,$67,$FE,$83,$8D,$74,$7E,$78
	.dc.b	$FC,$74,$7B,$76,$FF,$0B,$CF,$33,$A7,$6B,$3F,$5E,$3F,$E1,$E9,$D7
	.dc.b	$1E,$23,$C2,$EB,$8B,$ED,$FC,$46,$6B,$F8,$46,$5F,$06,$FF,$C4,$F9
	.dc.b	$F8,$9F,$C3,$61,$FD,$DE,$B7,$1F,$E9,$66,$EB,$F6,$B9,$1F,$11,$D6
	.dc.b	$FB,$72,$CC,$BF,$D4,$5C,$0E,$DC,$67,$FE,$A0,$D3,$E1,$72,$7D,$1E
	.dc.b	$1F,$DF,$61,$F0,$F9,$7F,$8C,$EE,$CF,$8E,$FF,$43,$5B,$7A,$AF,$71
	.dc.b	$8C,$FE,$33,$FC,$37,$2F,$BD,$9F,$BF,$8F,$F1,$F8,$7B,$F2,$F8,$DC
	.dc.b	$3F,$E3,$9C,$3C,$19,$7D,$98,$C7,$37,$CB,$3F,$A9,$F2,$D9,$F9,$9A
	.dc.b	$7F,$19,$D3,$D9,$CD,$F8,$B8,$FC,$2E,$1D,$4E,$D9,$78,$2F,$E8,$45
	.dc.b	$9F,$A1,$8D,$DB,$FD,$1B,$B7,$C8,$D7,$FA,$95,$7D,$83,$FC,$63,$AB
	.dc.b	$0E,$D8,$FB,$4F,$51,$F7,$78,$FF,$08,$7F,$6C,$E6,$B1,$EA,$C7,$E3
	.dc.b	$70,$FE,$9B,$2E,$13,$32,$EF,$8F,$C7,$C7,$E4,$7C,$47,$E3,$CF,$C9
	.dc.b	$73,$1C,$5F,$5E,$1F,$8A,$FD,$FC,$3F,$39,$3C,$A3,$FA,$75,$3E,$23
	.dc.b	$15,$EE,$7F,$38,$76,$F8,$8B,$5D,$73,$F8,$BC,$BF,$7A,$F8,$4F,$E7
	.dc.b	$6F,$9C,$FE,$C1,$FC,$05,$FE,$9C,$F4,$FC,$45,$AE,$97,$FD,$F8,$FA
	.dc.b	$17,$FB,$8A,$E3,$3D,$4F,$D5,$3F,$E8,$3F,$15,$E1,$FE,$DD,$B3,$FE
	.dc.b	$98,$E4,$FD,$67,$C4,$66,$BF,$31,$A7,$C5,$CB,$F8,$AD,$FC,$7E,$DF
	.dc.b	$E6,$FF,$11,$FE,$55,$DC,$DF,$88,$C3,$F9,$BD,$FF,$81,$1F,$53,$E2
	.dc.b	$2E,$1E,$CE,$0F,$EA,$59,$7B,$39,$BF,$63,$7C,$6B,$E9,$FC,$6D,$AF
	.dc.b	$C7,$EF,$FC,$34,$FE,$25,$FB,$67,$F2,$2A,$FE,$41,$1E,$A9,$E8,$7F
	.dc.b	$3C,$29,$FC,$A1,$FD,$81,$D7,$1F,$E4,$6F,$EC,$F3,$FE,$34,$77,$AF
	.dc.b	$E5,$8F,$59,$FB,$C7,$A8,$B9,$7F,$0A,$AF,$D9,$75,$9F,$DA,$71,$FF
	.dc.b	$9B,$B0,$FF,$0B,$F9,$5F,$98,$F8,$AF,$AC,$F5,$C3,$F4,$59,$FE,$07
	.dc.b	$2E,$AF,$F2,$AE,$9F,$E1,$A7,$F4,$BC,$FB,$63,$F6,$72,$79,$7F,$81
	.dc.b	$3F,$4B,$F1,$56,$7E,$96,$5F,$DD,$5F,$C0,$77,$7E,$FF,$2F,$9F,$1C
	.dc.b	$FC,$CF,$C5,$1F,$AE,$F8,$AB,$1C,$A7,$E7,$6B,$E6,$7F,$8A,$74,$FE
	.dc.b	$2F,$5F,$2C,$7D,$2D,$7F,$98,$AE,$BF,$EF,$AA,$33,$1F,$D0,$7E,$2B
	.dc.b	$F3,$5E,$2F,$C5,$AF,$85,$87,$EC,$36,$F8,$F9,$78,$76,$FF,$1E,$E7
	.dc.b	$E0,$8F,$26,$6F,$37,$F4,$87,$F0,$8F,$8A,$BB,$7A,$F3,$FD,$09,$F8
	.dc.b	$8C,$7E,$26,$1F,$86,$9F,$9D,$E6,$FE,$B4,$FE,$23,$F1,$56,$39,$79
	.dc.b	$7E,$EA,$F2,$73,$7D,$8D,$FF,$B0,$74,$FE,$13,$CB,$FD,$B1,$FD,$AF
	.dc.b	$F1,$2F,$F1,$1E,$3F,$E5,$C8,$F8,$1A,$FC,$C3,$F0,$9C,$5F,$00,$8F
	.dc.b	$C4,$3E,$C9,$F3,$FC,$AF,$E0,$47,$5F,$8A,$B7,$E6,$E5,$FE,$91,$C3
	.dc.b	$EE,$C3,$F3,$53,$FC,$0F,$EF,$FF,$C2,$07,$F8,$C3,$C6,$68,$F8,$CF
	.dc.b	$F0,$24,$FF,$84,$38,$3D,$DA,$F9,$8F,$27,$0F,$C2,$E1,$F8,$39,$75
	.dc.b	$78,$9F,$06,$AC,$79,$4F,$CA,$BD,$27,$E9,$57,$EF,$2F,$FB,$4B,$F8
	.dc.b	$BA,$FF,$49,$C6,$3A,$BF,$C8,$79,$3E,$33,$CF,$E8,$7C,$CD,$FF,$A8
	.dc.b	$E3,$E6,$DF,$E5,$BF,$3F,$3E,$E7,$E7,$72,$FE,$21,$1D,$C0,$73,$97
	.dc.b	$83,$E3,$FE,$2F,$67,$7D,$BF,$A9,$DF,$E3,$5F,$66,$3F,$42,$3D,$9E
	.dc.b	$8F,$CC,$3F,$09,$F1,$B7,$83,$E2,$4F,$1E,$1F,$D7,$74,$7F,$80,$75
	.dc.b	$FE,$E9,$5F,$CF,$CF,$87,$E9,$7E,$37,$E2,$3E,$BB,$E3,$BF,$D5,$61
	.dc.b	$F0,$48,$7C,$1C,$FC,$9C,$1F,$07,$D3,$FB,$E7,$E5,$FE,$32,$E5,$E4
	.dc.b	$9F,$BD,$7F,$CC,$38,$3F,$65,$87,$CC,$D3,$FB,$B5,$F9,$6B,$E4,$C6
	.dc.b	$1F,$99,$D1,$F8,$EF,$8C,$FC,$6F,$AB,$E7,$9F,$C6,$1E,$53,$CF,$CD
	.dc.b	$FC,$03,$87,$E1,$7A,$BF,$94,$3F,$59,$F1,$97,$3F,$36,$9F,$7A,$FF
	.dc.b	$41,$87,$F9,$7F,$0F,$95,$EB,$7E,$70,$FD,$A7,$C6,$5C,$38,$D7,$83
	.dc.b	$A3,$EC,$BC,$63,$F5,$1E,$9F,$E9,$0F,$DE,$7C,$7D,$8E,$33,$E9,$26
	.dc.b	$5E,$98,$FD,$86,$5F,$AC,$3F,$82,$7C,$65,$8E,$33,$E8,$AE,$5C,$7F
	.dc.b	$39,$FB,$2B,$D7,$FF,$3A,$1F,$DA,$3D,$3B,$7C,$B4,$F4,$DF,$2C,$BF
	.dc.b	$AC,$FA,$FF,$E7,$B3,$71,$FE,$75,$7E,$97,$3F,$F6,$1E,$1F,$95,$7C
	.dc.b	$B3,$F1,$5F,$93,$D7,$FF,$42,$1C,$F2,$FF,$4B,$3C,$AF,$FA,$94,$FE
	.dc.b	$6C,$FC,$46,$5F,$DA,$B3,$FF,$46,$1F,$E9,$1F,$D3,$7F,$AE,$E1,$FE
	.dc.b	$9F,$8E,$55,$FD,$4A,$FC,$7D,$9F,$A4,$AF,$EA,$5E,$A7,$FA,$54,$E6
	.dc.b	$FF,$A9,$DE,$35,$FD,$03,$5F,$CC,$9F,$D6,$F4,$F5,$F0,$7F,$4C,$F4
	.dc.b	$FF,$CF,$07,$F9,$E6,$FF,$D4,$0F,$F2,$AD,$FF,$7B,$7C,$B3,$FE,$17
	.dc.b	$87,$29,$FE,$F1,$E9,$FF,$A0,$4E,$79,$FF,$A8,$DE,$5A,$7E,$D3,$4F
	.dc.b	$C2,$D7,$CD,$DF,$FC,$1B,$E8,$FF,$A2,$4F,$F7,$2E,$6C,$FF,$89,$DF
	.dc.b	$FB,$F3,$9F,$59,$FF,$2A,$CB,$A6,$FF,$89,$3F,$21,$5F,$B7,$F4,$3F
	.dc.b	$D2,$87,$07,$FC,$FE,$F9,$1F,$D8,$C7,$1D,$3F,$C2,$5A,$7F,$4A,$C3
	.dc.b	$C7,$97,$8A,$BF,$AD,$DF,$29,$FE,$4F,$9F,$9E,$3F,$D8,$B1,$D6,$3F
	.dc.b	$19,$1F,$AD,$CB,$ED,$3F,$DC,$FC,$DE,$AF,$FB,$31,$AB,$D1,$EF,$5D
	.dc.b	$61,$82,$BB,$63,$F9,$0A,$FA,$D5,$FA,$8C,$BF,$B1,$F9,$BD,$30,$FE
	.dc.b	$BD,$9F,$A3,$0F,$66,$39,$AE,$78,$7E,$44,$F4,$1F,$9A,$AF,$A0,$FD
	.dc.b	$43,$FB,$B1,$6D,$A3,$4D,$4E,$8F,$F2,$19,$B8,$D3,$A9,$E3,$3C,$23
	.dc.b	$FC,$0C,$6D,$C4,$6A,$38,$4F,$27,$47,$3F,$31,$EF,$70,$6C,$66,$30
	.dc.b	$47,$04,$3D,$91,$C1,$B1,$B6,$9B,$3F,$0D,$E5,$FE,$63,$6F,$B1,$82
	.dc.b	$31,$AE,$AB,$E9,$1F,$DC,$25,$E5,$6A,$31,$8F,$A3,$0E,$6D,$BF,$CA
	.dc.b	$36,$7F,$BC,$FE,$00,$F6,$39,$47,$E2,$27,$4D,$CF,$C2,$D1,$87,$A3
	.dc.b	$E2,$78,$E7,$E4,$9D,$B5,$DE,$5F,$A5,$CF,$BF,$1B,$C4,$E3,$9F,$83
	.dc.b	$0F,$F0,$C6,$7C,$A3,$C7,$AF,$9F,$2C,$4C,$7A,$9F,$09,$F0,$E7,$E4
	.dc.b	$F3,$64,$F9,$1F,$3F,$97,$E0,$AE,$5B,$7C,$87,$DF,$87,$7B,$EA,$7C
	.dc.b	$19,$D3,$8C,$BC,$ED,$C7,$AB,$F0,$B0,$AE,$FC,$9F,$03,$C7,$3F,$80
	.dc.b	$7B,$B8,$33,$8F,$62,$70,$7F,$0D,$C7,$E7,$E1,$30,$F9,$EF,$DA,$FD
	.dc.b	$B5,$F0,$BC,$AF,$F0,$3B,$7B,$9F,$E4,$FB,$7B,$72,$EB,$8F,$C9,$3C
	.dc.b	$95,$F2,$9F,$B7,$C1,$F7,$88,$FD,$EE,$BE,$CF,$1A,$E6,$AF,$B6,$1E
	.dc.b	$28,$EA,$49,$FC,$53,$CC,$87,$CD,$AF,$C6,$F8,$FF,$CF,$BD,$1F,$F5
	.dc.b	$4F,$17,$D0,$BE,$ED,$C3,$DB,$C1,$E4,$C7,$EC,$47,$86,$BB,$C7,$D9
	.dc.b	$C3,$A7,$4F,$F3,$96,$C7,$EB,$DA,$FF,$2C,$F2,$2E,$3D,$E3,$E0,$61
	.dc.b	$D4,$F0,$BF,$3C,$C9,$0F,$E6,$BC,$5F,$A8,$AE,$64,$8F,$F2,$30,$F1
	.dc.b	$7F,$D7,$4E,$2F,$C0,$FB,$0E,$9B,$7C,$E7,$B4,$68,$97,$FB,$6F,$8E
	.dc.b	$F4,$92,$7A,$7C,$77,$B3,$9B,$F2,$7C,$7E,$99,$FB,$5C,$1F,$2F,$1E
	.dc.b	$95,$FB,$7C,$7F,$80,$E1,$F1,$09,$7E,$7F,$8E,$E9,$8F,$3C,$3D,$39
	.dc.b	$F9,$EB,$D4,$48,$F3,$BF,$C2,$AB,$E7,$91,$F8,$BC,$CF,$85,$EB,$1F
	.dc.b	$8B,$E0,$FA,$4F,$F1,$ED,$7D,$1A,$FB,$23,$F6,$C7,$F1,$CF,$49,$F0
	.dc.b	$1F,$79,$B1,$F4,$9A,$FF,$9D,$BD,$87,$0F,$83,$9B,$F2,$38,$7E,$E6
	.dc.b	$BD,$CF,$78,$FC,$79,$FD,$D6,$FF,$CA,$D3,$D7,$CF,$F0,$1F,$90,$F5
	.dc.b	$3F,$D9,$FC,$7D,$75,$F5,$D7,$C6,$E2,$F9,$1C,$5F,$4B,$8F,$E6,$72
	.dc.b	$7F,$3A,$CD,$E8,$7A,$C9,$87,$FA,$47,$3F,$F6,$DF,$2F,$58,$FC,$3E
	.dc.b	$FE,$97,$E5,$18,$C7,$F8,$13,$FC,$8B,$07,$A1,$F1,$C9,$1D,$78,$9F
	.dc.b	$1D,$D6,$3E,$E6,$1D,$78,$BE,$3E,$3E,$89,$FA,$5E,$8F,$D1,$23,$CF
	.dc.b	$6E,$5B,$FB,$E3,$F9,$9F,$07,$C2,$C7,$E7,$70,$FC,$6A,$5F,$C3,$F4
	.dc.b	$7E,$E1,$30,$F3,$7C,$77,$4C,$38,$C7,$E1,$CF,$B3,$9F,$95,$F8,$1E
	.dc.b	$97,$E1,$48,$FB,$FE,$3B,$94,$FE,$9F,$4F,$23,$CA,$7F,$1E,$7C,$9F
	.dc.b	$4B,$F1,$C4,$7E,$A7,$E2,$9F,$E0,$9F,$0F,$F8,$EE,$5E,$49,$FC,$07
	.dc.b	$6F,$E6,$B0,$F2,$71,$7E,$2F,$D3,$FC,$C1,$2B,$9F,$C7,$F9,$1F,$B6
	.dc.b	$7E,$E8,$F2,$4F,$F2,$8E,$1F,$C4,$7A,$9F,$A1,$24,$F9,$3E,$3F,$95
	.dc.b	$7F,$3C,$DF,$DB,$3D,$39,$3F,$6E,$FF,$66,$F8,$A7,$91,$FC,$A1,$B1
	.dc.b	$FC,$71,$AF,$FB,$83,$87,$FC,$B1,$C3,$F0,$37,$F2,$F3,$FF,$2A,$AE
	.dc.b	$DC,$5F,$0F,$2F,$BB,$1F,$E1,$4C,$7F,$CB,$CF,$5F,$57,$C0,$7E,$AF
	.dc.b	$27,$C7,$FA,$B1,$FD,$5E,$BE,$AD,$BF,$27,$87,$D6,$7F,$CD,$58,$BD
	.dc.b	$6F,$51,$27,$EB,$FC,$87,$37,$F9,$8F,$37,$D5,$7E,$27,$27,$F8,$B2
	.dc.b	$FF,$CB,$F1,$F0,$7D,$6F,$88,$79,$7D,$0E,$51,$F7,$8F,$82,$7F,$3A
	.dc.b	$AE,$5C,$3F,$2C,$91,$F8,$8F,$90,$E3,$C5,$FC,$B3,$2F,$16,$BF,$76
	.dc.b	$BE,$37,$AF,$F4,$C8,$FE,$43,$E4,$38,$E3,$FB,$5E,$1F,$D5,$3F,$3F
	.dc.b	$3F,$57,$B1,$F6,$C9,$7F,$98,$DF,$8F,$37,$F8,$A3,$4F,$0E,$3F,$B3
	.dc.b	$C3,$E1,$7A,$FF,$83,$26,$1F,$A0,$F9,$0E,$2F,$EF,$2F,$8B,$F0,$4F
	.dc.b	$BD,$EC,$7E,$2C,$8F,$EA,$3D,$1E,$3C,$1F,$EA,$1E,$7F,$B9,$9F,$AB
	.dc.b	$4F,$E3,$33,$F7,$3D,$8F,$CA,$13,$0F,$D8,$FC,$7F,$2B,$FC,$85,$7E
	.dc.b	$10,$F9,$53,$FE,$3F,$F2,$7F,$C0,$24,$7F,$BA,$6F,$C5,$FE,$14,$7E
	.dc.b	$2E,$BC,$4D,$3F,$CF,$51,$FE,$0A,$23,$FD,$72,$F9,$3F,$B7,$38,$1F
	.dc.b	$E0,$83,$D7,$3F,$EC,$FF,$1F,$FC,$E2,$7F,$85,$09,$1F,$C3,$36,$E7
	.dc.b	$A7,$F4,$89,$DC,$FF,$9D,$9F,$F4,$57,$9C,$EE,$7C,$6F,$E6,$7D,$AF
	.dc.b	$DD,$12,$FF,$8B,$FC,$6F,$4C,$B8,$DB,$E2,$29,$2B,$FD,$33,$8F,$5F
	.dc.b	$8C,$FF,$43,$F9,$9F,$E2,$A2,$4F,$F2,$2F,$7B,$A3,$FE,$8A,$8D,$B4
	.dc.b	$FF,$3A,$F2,$FF,$A6,$37,$FE,$51,$E4,$7F,$97,$23,$FC,$6F,$18,$96
	.dc.b	$3E,$DF,$8B,$EB,$1F,$D2,$AB,$91,$1F,$F4,$0F,$99,$D4,$E6,$47,$F9
	.dc.b	$BF,$C5,$F2,$E8,$FC,$84,$EE,$78,$9F,$03,$2F,$F7,$14,$FE,$57,$CC
	.dc.b	$F2,$92,$7F,$C7,$FF,$1C,$F8,$AE,$9D,$1F,$E7,$6A,$F7,$12,$BF,$D2
	.dc.b	$B1,$F0,$4F,$79,$1F,$F2,$4E,$DF,$C8,$1F,$F2,$47,$87,$FC,$A4,$FD
	.dc.b	$F3,$FE,$C2,$9E,$4F,$43,$FD,$43,$E4,$7C,$AF,$32,$47,$47,$FC,$C7
	.dc.b	$EF,$7F,$A9,$7C,$FF,$E7,$13,$FC,$DB,$87,$F8,$13,$CC,$FE,$EF,$87
	.dc.b	$67,$FB,$CE,$1F,$5B,$CD,$91,$E8,$7E,$CC,$7C,$28,$FF,$6B,$BE,$D7
	.dc.b	$99,$E6,$3D,$EF,$06,$9F,$E4,$6E,$DE,$4F,$FA,$4E,$34,$B0,$FE,$A7
	.dc.b	$A7,$FA,$A2,$3E,$4F,$99,$F8,$A2,$38,$3F,$C5,$5C,$4F,$83,$EB,$4A
	.dc.b	$DD,$75,$AF,$F5,$7C,$7F,$8F,$57,$C6,$7E,$5B,$CC,$95,$F6,$1F,$F4
	.dc.b	$7F,$ED,$DF,$E3,$A8,$FB,$3B,$FF,$44,$7F,$82,$73,$7F,$1F,$AE,$31
	.dc.b	$DC,$FF,$7E,$BF,$A2,$F3,$24,$7E,$11,$AE,$FA,$FF,$BB,$FC,$3F,$AA
	.dc.b	$73,$BF,$C4,$7F,$57,$E6,$4B,$FC,$6B,$5E,$5A,$BE,$37,$E1,$CE,$6F
	.dc.b	$E6,$BF,$07,$AB,$FC,$09,$F0,$F3,$0F,$F0,$23,$F1,$7E,$37,$F0,$1A
	.dc.b	$7F,$82,$BE,$37,$E4,$12,$BE,$8F,$C6,$FF,$83,$7E,$2A,$47,$F8,$35
	.dc.b	$F9,$DF,$1B,$CB,$67,$C6,$FD,$12,$57,$DA,$F8,$DF,$F0,$B7,$C6,$C8
	.dc.b	$FF,$0B,$3F,$5F,$E3,$39,$7C,$5F,$D9,$24,$FE,$0B,$E2,$BF,$C3,$DF
	.dc.b	$19,$23,$FC,$3C,$FD,$FF,$89,$F1,$7C,$47,$E0,$7E,$22,$57,$E2,$A3
	.dc.b	$FC,$4D,$F1,$92,$3F,$C4,$CF,$E1,$FE,$27,$D5,$F1,$3F,$89,$24,$FE
	.dc.b	$4B,$E2,$7F,$8F,$FC,$04,$8F,$E0,$2F,$E3,$FE,$27,$F9,$1F,$B5,$FD
	.dc.b	$B2,$7F,$13,$E6,$99,$E4,$FF,$AA,$DE,$24,$3F,$18,$7F,$53,$F0,$3F
	.dc.b	$4F,$EF,$7F,$2E,$24,$7F,$10,$BE,$EF,$E5,$EB,$FD,$7C,$7F,$07,$E8
	.dc.b	$F0,$42,$3C,$0F,$66,$DF,$ED,$2B,$9C,$E5,$B1,$FE,$4A,$F3,$DE,$E6
	.dc.b	$BC,$8F,$F0,$15,$7F,$45,$3F,$A3,$79,$F7,$0F,$31,$0F,$D1,$7E,$2B
	.dc.b	$FE,$54,$C7,$F8,$1C,$FF,$4A,$24,$FF,$AA,$67,$FC,$9E,$F9,$E3,$AB
	.dc.b	$F2,$EF,$FC,$25,$1F,$DB,$27,$FC,$95,$39,$D7,$E0,$E7,$FD,$BA,$3F
	.dc.b	$9B,$D7,$67,$C7,$F1,$1F,$D6,$B3,$91,$FC,$F1,$FF,$63,$60,$F5,$33
	.dc.b	$3F,$EA,$49,$FE,$C8,$7F,$65,$F7,$6C,$71,$21,$F0,$3E,$1F,$FB,$3C
	.dc.b	$48,$FE,$92,$F6,$F8,$7C,$47,$FA,$C2,$3A,$3E,$B3,$F1,$3E,$FE,$1F
	.dc.b	$B5,$1C,$37,$FF,$40,$F9,$0C,$DF,$0F,$FD,$DB,$09,$87,$F5,$87,$DB
	.dc.b	$F0,$F8,$8F,$F5,$FC,$7F,$7F,$3E,$67,$C3,$D8,$E0,$43,$ED,$FC,$27
	.dc.b	$F8,$0B,$E0,$E4,$7F,$64,$7F,$23,$9F,$FA,$4B,$E4,$9E,$FE,$AB,$FD
	.dc.b	$9B,$5F,$E0,$D3,$FC,$1D,$F0,$76,$39,$10,$FC,$4F,$A5,$FE,$10,$89
	.dc.b	$1F,$DC,$9F,$F0,$9F,$C1,$E2,$3F,$DB,$31,$D1,$F9,$43,$F0,$47,$B3
	.dc.b	$BC,$FF,$AA,$D5,$EF,$FD,$4D,$7E,$AA,$FF,$51,$87,$67,$FC,$02,$FF
	.dc.b	$76,$E4,$F0,$76,$FF,$99,$70,$F0,$52,$1D,$BD,$33,$ED,$3F,$C1,$2A
	.dc.b	$BE,$53,$E4,$E5,$E2,$69,$FC,$2F,$2F,$26,$7F,$EB,$E6,$CF,$F0,$FE
	.dc.b	$6F,$C1,$7A,$5E,$4C,$BF,$47,$8F,$93,$3F,$1E,$5C,$F3,$EA,$F8,$4F
	.dc.b	$E3,$19,$F5,$73,$9F,$67,$FC,$7F,$CE,$F3,$F3,$EF,$C9,$E3,$B7,$CA
	.dc.b	$8F,$D5,$F9,$BC,$B2,$F2,$57,$C1,$3E,$AE,$3F,$51,$E2,$7B,$EB,$A3
	.dc.b	$FC,$9A,$3F,$C0,$8E,$8F,$B9,$3E,$6F,$7A,$46,$67,$FC,$06,$D9,$F1
	.dc.b	$3F,$89,$3F,$0B,$E1,$36,$FE,$53,$B7,$C6,$9E,$A7,$13,$A6,$3E,$4D
	.dc.b	$BE,$96,$1D,$27,$CC,$D8,$FE,$55,$D5,$75,$FC,$15,$73,$3D,$9E,$D6
	.dc.b	$5F,$A8,$66,$BF,$7B,$C0,$F3,$3E,$9D,$4A,$F7,$3F,$C6,$3E,$3F,$11
	.dc.b	$D0,$C9,$FC,$B7,$E2,$B3,$87,$E2,$6A,$E3,$F8,$9C,$7F,$E0,$0D,$FC
	.dc.b	$FF,$6F,$F0,$24,$7E,$87,$C0,$3D,$2C,$46,$CF,$A1,$3F,$88,$F7,$33
	.dc.b	$97,$E5,$E2,$E3,$DF,$E2,$75,$7F,$5E,$EB,$FD,$23,$E2,$2E,$1F,$9F
	.dc.b	$9B,$7F,$B6,$F7,$AD,$FE,$96,$6F,$6F,$EC,$7E,$0B,$F3,$35,$2B,$F0
	.dc.b	$8F,$EB,$27,$F6,$DF,$11,$99,$FC,$74,$DC,$7F,$96,$F3,$FF,$1A,$9E
	.dc.b	$24,$36,$7E,$62,$FF,$80,$77,$7E,$B3,$E0,$9C,$B4,$95,$BB,$FF,$45
	.dc.b	$3F,$E4,$75,$F0,$DE,$57,$FC,$D1,$19,$F8,$6C,$4E,$7E,$1B,$5F,$0D
	.dc.b	$2B,$A4,$7E,$E3,$E1,$64,$7E,$8D,$E9,$96,$23,$F5,$91,$75,$F7,$FC
	.dc.b	$35,$E3,$FE,$47,$36,$BF,$D7,$FA,$7F,$82,$A7,$F9,$3F,$C3,$7F,$06
	.dc.b	$24,$FB,$7D,$AA,$FE,$E6,$2E,$3F,$33,$D9,$B7,$CA,$BF,$71,$5F,$07
	.dc.b	$D9,$FE,$23,$F0,$D2,$7F,$78,$FC,$5F,$63,$11,$FC,$16,$2E,$3F,$63
	.dc.b	$CD,$B8,$FF,$3C,$9B,$5E,$AF,$13,$F8,$EF,$C3,$48,$FE,$10,$FD,$1F
	.dc.b	$47,$11,$FC,$46,$2E,$3F,$84,$F2,$6E,$3F,$D2,$26,$D7,$C6,$F4,$3F
	.dc.b	$94,$FC,$34,$8F,$E2,$8F,$E4,$7F,$15,$FE,$59,$1F,$D5,$63,$4F,$F1
	.dc.b	$F7,$5E,$66,$32,$FE,$81,$17,$1F,$E8,$17,$FC,$7A,$FF,$93,$7C,$3F
	.dc.b	$F3,$82,$3F,$95,$8F,$E5,$CE,$8F,$E4,$D1,$A7,$FD,$31,$1F,$E8,$F7
	.dc.b	$24,$38,$1A,$F1,$BF,$A2,$FC,$3C,$8F,$E6,$0F,$B7,$C5,$C4,$7F,$39
	.dc.b	$8B,$97,$F8,$62,$DE,$1E,$32,$FE,$E3,$56,$DE,$1A,$DD,$7F,$D5,$3E
	.dc.b	$1E,$57,$F3,$F7,$E6,$74,$E2,$3F,$A4,$45,$CF,$C1,$CD,$73,$FE,$FF
	.dc.b	$36,$BB,$E2,$E3,$FE,$C3,$F0,$F2,$7F,$A6,$3F,$67,$8B,$11,$FD,$56
	.dc.b	$2E,$5E,$DE,$2B,$97,$F8,$32,$6D,$76,$DF,$FB,$6F,$95,$23,$FA,$C3
	.dc.b	$F8,$5D,$31,$1F,$D8,$62,$E3,$F2,$6E,$E1,$FE,$18,$9B,$5E,$A7,$FB
	.dc.b	$CF,$C3,$C8,$FE,$C8,$FE,$44,$C3,$FD,$B5,$B7,$F5,$7E,$1E,$DF,$F8
	.dc.b	$92,$6D,$7C,$6F,$87,$FF,$02,$D4,$8F,$EE,$0F,$E7,$7E,$1B,$11,$FD
	.dc.b	$E6,$2E,$1F,$E5,$8E,$8F,$F4,$26,$3F,$E3,$09,$CB,$6B,$F9,$64,$FF
	.dc.b	$A1,$2B,$FC,$41,$87,$F8,$50,$93,$AF,$8D,$AF,$F8,$1E,$2C,$F7,$F8
	.dc.b	$6B,$87,$F8,$FA,$6D,$7E,$3B,$0F,$F4,$03,$FE,$0D,$F1,$9E,$A7,$F3
	.dc.b	$B2,$47,$93,$E1,$7F,$A5,$53,$E1,$69,$27,$D3,$F0,$BA,$F8,$59,$1E
	.dc.b	$C7,$8F,$C2,$FF,$8B,$B8,$64,$FF,$86,$5E,$BF,$0B,$88,$FF,$11,$45
	.dc.b	$E0,$F9,$5F,$07,$78,$3F,$CB,$33,$6B,$D9,$F0,$7F,$E3,$EF,$57,$E7
	.dc.b	$C7,$FB,$55,$AF,$F9,$5A,$BD,$7F,$07,$B6,$47,$F8,$BE,$78,$F8,$6E
	.dc.b	$07,$C1,$EA,$BF,$DB,$75,$D0,$DC,$FF,$86,$63,$97,$C7,$78,$C8,$FD
	.dc.b	$EF,$62,$BC,$A2,$ED,$F8,$7E,$5B,$9F,$FB,$7B,$1F,$C2,$36,$BE,$9F
	.dc.b	$C6,$FB,$FA,$A6,$3E,$97,$EB,$FB,$B8,$8F,$3C,$5C,$BF,$29,$EE,$DC
	.dc.b	$7B,$4D,$AF,$E2,$7F,$8A,$BD,$EF,$99,$DD,$27,$E3,$3F,$E0,$D9,$7B
	.dc.b	$D9,$9F,$8B,$37,$2F,$F5,$CD,$7F,$83,$67,$27,$22,$08,$DE,$DF,$6F
	.dc.b	$E2,$64,$FD,$57,$C1,$F1,$98,$8F,$A7,$16,$BB,$FB,$F7,$5F,$95,$36
	.dc.b	$BD,$BF,$01,$F8,$7F,$06,$47,$E0,$5F,$67,$BF,$88,$FB,$F1,$71,$F5
	.dc.b	$FC,$05,$C7,$EB,$4D,$AF,$93,$E2,$7E,$4F,$C4,$91,$F8,$D7,$E4,$7C
	.dc.b	$06,$23,$F1,$71,$71,$F8,$FF,$19,$71,$FC,$14,$DA,$FA,$BF,$01,$F9
	.dc.b	$FF,$1E,$47,$E6,$5F,$A9,$F0,$18,$8F,$CB,$C5,$C7,$E9,$FC,$05,$C7
	.dc.b	$F1,$D3,$6B,$F0,$3F,$01,$FA,$BF,$2A,$47,$E9,$5F,$C0,$7C,$06,$23
	.dc.b	$F4,$71,$71,$FB,$FF,$01,$71,$FC,$D4,$DA,$FC,$6F,$C0,$7E,$DF,$CD
	.dc.b	$91,$FB,$17,$F1,$9F,$01,$88,$FD,$7C,$5C,$7F,$17,$F0,$17,$1F,$D3
	.dc.b	$4D,$AF,$ED,$7F,$8A,$FF,$84,$3B,$5E,$87,$F0,$52,$4F,$F9,$6E,$7F
	.dc.b	$D5,$EF,$EF,$23,$2D,$35,$F1,$58,$C9,$8F,$ED,$6B,$0F,$FB,$56,$FF
	.dc.b	$C2,$33,$AB,$FE,$25,$E8,$FF,$15,$23,$E1,$F8,$6A,$FF,$0C,$8B,$3C
	.dc.b	$7C,$F5,$BE,$0F,$38,$FF,$05,$BB,$87,$C0,$F8,$3F,$E4,$5E,$9C,$8F
	.dc.b	$E3,$8F,$FA,$0F,$0F,$F0,$9C,$60,$87,$F1,$C8,$D7,$C1,$71,$8C,$1A
	.dc.b	$CF,$C9,$4F,$82,$FE,$61,$EA,$CA,$FE,$54,$F7,$F8,$2C,$47,$F2,$88
	.dc.b	$B7,$CB,$E0,$6E,$BF,$C7,$E6,$D7,$DA,$F8,$1F,$E7,$9E,$C4,$8F,$E6
	.dc.b	$CF,$AF,$E0,$71,$1F,$CD,$22,$E3,$E6,$F8,$1B,$8F,$F2,$B9,$B5,$F8
	.dc.b	$6F,$81,$FE,$91,$F0,$F2,$3F,$A1,$3F,$1F,$E0,$B1,$1F,$D0,$22,$E3
	.dc.b	$F0,$BE,$0A,$E3,$FC,$DE,$6D,$7E,$4B,$E0,$BF,$AA,$7B,$D2,$3F,$A7
	.dc.b	$3F,$4F,$E0,$B1,$1F,$D3,$22,$E3,$F3,$BE,$0A,$E3,$FD,$0E,$6D,$7E
	.dc.b	$7B,$E0,$BF,$B0,$7C,$0C,$8F,$EB,$4F,$FB,$07,$F1,$5F,$EA,$CE,$63
	.dc.b	$FA,$CB,$9F,$82,$C6,$5F,$97,$97,$87,$FE,$04,$79,$9F,$EC,$1B,$7B
	.dc.b	$FF,$DB,$FE,$0A,$4F,$F6,$97,$C1,$EF,$62,$3F,$B3,$C5,$C3,$F3,$97
	.dc.b	$FE,$12,$9C,$9C,$48,$7F,$43,$D4,$FF,$99,$A1,$EB,$E2,$34,$6B,$A3
	.dc.b	$FD,$F0,$FF,$7B,$F8,$2F,$F0,$21,$BC,$3D,$D3,$A9,$FE,$F7,$3A,$78
	.dc.b	$1E,$09,$DF,$1E,$63,$66,$8F,$08,$A7,$1F,$FB,$C7,$D0,$7A,$1F,$4F
	.dc.b	$D1,$97,$FE,$EB,$9C,$BF,$E1,$1F,$82,$FF,$0D,$11,$FA,$1E,$8D,$7F
	.dc.b	$C2,$31,$73,$F9,$7E,$77,$7F,$4B,$C1,$E9,$4B,$ED,$1F,$07,$8B,$E4
	.dc.b	$7E,$11,$FD,$F6,$7D,$D1,$8F,$83,$D4,$7D,$09,$D3,$F3,$BC,$F7,$2B
	.dc.b	$CC,$DC,$AB,$7A,$7E,$2D,$B8,$F5,$9F,$C5,$78,$FF,$17,$D4,$F8,$DE
	.dc.b	$A4,$9F,$89,$1D,$2F,$FC,$1B,$1E,$63,$C9,$F0,$79,$3F,$3D,$E1,$58
	.dc.b	$D1,$E3,$8C,$3E,$9D,$3F,$38,$FA,$FF,$57,$FC,$53,$7E,$17,$E0,$EE
	.dc.b	$9F,$E8,$BB,$7F,$33,$7F,$DB,$79,$7F,$86,$E3,$DD,$F1,$DE,$73,$F6
	.dc.b	$1C,$B7,$1F,$1D,$6F,$83,$E8,$D7,$6D,$3E,$2D,$7F,$B0,$B1,$FF,$6F
	.dc.b	$63,$DF,$E3,$3C,$71,$FD,$33,$0F,$D1,$E7,$E5,$F6,$7E,$37,$89,$31
	.dc.b	$EE,$F5,$30,$FE,$5B,$4F,$DE,$BA,$3E,$65,$7F,$8E,$BE,$1B,$BF,$B3
	.dc.b	$E0,$F6,$65,$6B,$3F,$F7,$43,$CE,$BE,$B1,$1E,$7B,$3E,$0F,$DD,$F0
	.dc.b	$72,$72,$FB,$3D,$AF,$6F,$C4,$FB,$BD,$B9,$5D,$E3,$F0,$5F,$0F,$23
	.dc.b	$E5,$BF,$27,$E0,$FE,$2F,$B9,$27,$CC,$FC,$8F,$73,$E4,$FB,$9F,$2B
	.dc.b	$DC,$95,$EB,$8F,$C7,$7C,$34,$8F,$AE,$FD,$5F,$2F,$E8,$F9,$B2,$7E
	.dc.b	$13,$F5,$3D,$DF,$AB,$EE,$7D,$6F,$72,$57,$C7,$8F,$CD,$71,$C8,$FC
	.dc.b	$1B,$F8,$EF,$C5,$BF,$23,$E0,$7E,$37,$DD,$FB,$E4,$AF,$A1,$E9,$FE
	.dc.b	$0B,$D3,$FC,$1F,$B9,$27,$EA,$47,$E9,$FA,$64,$7E,$41,$ED,$DB,$F8
	.dc.b	$AF,$72,$4F,$DB,$7F,$3F,$FA,$C7,$FA,$80,$FF,$47,$C3,$C1,$47,$F4
	.dc.b	$08,$FC,$3B,$5F,$E0,$8F,$07,$F8,$17,$C4,$7E,$E4,$8F,$E3,$B6,$74
	.dc.b	$78,$3D,$B9,$3D,$1F,$1F,$C4,$F9,$3D,$BF,$2F,$C4,$4A,$F0,$C7,$F0
	.dc.b	$5F,$89,$91,$FA,$E7,$E6,$73,$7B,$3D,$C9,$3E,$87,$DF,$ED,$FC,$0F
	.dc.b	$6E,$4F,$B1,$F8,$7F,$0D,$F5,$79,$A4,$FA,$5F,$B1,$C7,$F2,$3E,$1A
	.dc.b	$4F,$C4,$7E,$5F,$B5,$F3,$3D,$AF,$9B,$F0,$B2,$BE,$4C,$7F,$1D,$F1
	.dc.b	$E4,$7F,$06,$7F,$88,$FE,$2F,$FD,$97,$BF,$FC,$85,$D3,$FE,$CD,$AF
	.dc.b	$9F,$F0,$DF,$60,$FB,$3E,$DF,$DA,$F8,$6F,$B7,$F0,$92,$BE,$BC,$7F
	.dc.b	$2B,$F8,$A9,$1F,$C5,$1E,$FC,$7F,$82,$F8,$59,$3F,$79,$FC,$2F,$BB
	.dc.b	$FC,$93,$F5,$9F,$E2,$FE,$11,$FE,$6B,$5F,$D1,$5B,$B7,$F9,$65,$FE
	.dc.b	$3B,$1F,$D1,$7E,$2F,$F9,$E1,$1F,$E2,$3C,$1F,$EE,$E7,$F8,$9C,$FF
	.dc.b	$27,$7F,$24,$DB,$F0,$7B,$FE,$3F,$73,$C9,$EC,$CB,$E2,$F9,$FD,$5F
	.dc.b	$E9,$5F,$17,$27,$F9,$9B,$F3,$B6,$F6,$7A,$72,$7C,$2F,$BF,$CF,$F8
	.dc.b	$1E,$64,$9F,$2B,$F0,$FC,$BF,$EA,$DF,$11,$27,$F9,$FB,$FD,$2B,$F1
	.dc.b	$8F,$E5,$F1,$F1,$7C,$7F,$94,$47,$DF,$DB,$F3,$7A,$24,$7C,$27,$E8
	.dc.b	$72,$FF,$63,$F5,$64,$FF,$4C,$7B,$67,$F5,$77,$FA,$D9,$CA,$F9,$71
	.dc.b	$F6,$6A,$47,$CE,$7E,$E7,$C4,$7F,$6F,$24,$7F,$58,$7D,$58,$7E,$07
	.dc.b	$E2,$7F,$05,$F0,$F2,$BE,$BC,$7E,$17,$E2,$24,$7D,$A7,$FB,$5F,$83
	.dc.b	$FC,$7B,$E1,$7F,$AC,$F0,$AF,$FB,$F1,$2F,$FA,$9C,$65,$FE,$AB,$5F
	.dc.b	$DC,$AB,$FB,$0B,$FD,$99,$BF,$0B,$DB,$E1,$FB,$FC,$3C,$BE,$2F,$8B
	.dc.b	$E1,$BC,$7E,$C7,$F7,$EF,$8B,$FE,$D9,$5F,$C7,$DB,$E0,$33,$FF,$58
	.dc.b	$4F,$8B,$E1,$7D,$7F,$09,$23,$07,$9B,$E1,$7D,$E4,$7D,$3F,$0B,$F0
	.dc.b	$BD,$CF,$87,$F0,$92,$7D,$B1,$F1,$BE,$17,$E3,$FC,$2C,$9F,$83,$1F
	.dc.b	$2B,$E1,$64,$7C,$47,$E6,$FB,$9F,$3B,$E1,$7E,$7F,$C2,$CA,$F9,$31
	.dc.b	$F4,$BE,$17,$E9,$FC,$2C,$9F,$9B,$1F,$5B,$DD,$FA,$FE,$EC,$9F,$A3
	.dc.b	$1F,$E2,$4F,$D6,$BF,$D0,$DE,$A7,$FB,$55,$FE,$8F,$CC,$69,$AE,$DF
	.dc.b	$F0,$BB,$FD,$1E,$78,$7C,$2F,$33,$A7,$C6,$75,$F8,$69,$97,$F8,$52
	.dc.b	$FF,$91,$61,$CB,$E1,$7C,$5F,$0D,$E3,$F8,$69,$7B,$8F,$37,$C3,$79
	.dc.b	$FD,$D9,$3C,$E3,$D5,$F0,$D2,$3B,$3E,$DF,$86,$F7,$7C,$34,$9F,$13
	.dc.b	$F0,$7E,$17,$E1,$7C,$34,$9F,$33,$F1,$7C,$4F,$8D,$F0,$BF,$1F,$E1
	.dc.b	$25,$7A,$E3,$E5,$7C,$2C,$8F,$73,$F3,$7D,$EF,$9D,$F0,$BF,$E5,$2F
	.dc.b	$D6,$FF,$97,$9F,$CC,$39,$51,$9F,$89,$98,$7F,$8E,$AB,$F9,$8F,$37
	.dc.b	$F9,$19,$FE,$50,$7F,$12,$EA,$46,$7E,$0F,$B1,$DF,$E1,$3C,$1F,$09
	.dc.b	$30,$E8,$F8,$FE,$13,$C9,$F0,$B2,$7C,$0F,$9F,$E1,$79,$47,$8F,$0E
	.dc.b	$AF,$97,$0E,$8F,$FA,$2F,$36,$38,$CF,$CB,$3F,$E1,$19,$65,$F0,$FE
	.dc.b	$06,$4F,$BD,$F8,$DF,$0B,$F1,$FE,$16,$4F,$C3,$7E,$57,$C2,$FC,$BF
	.dc.b	$85,$93,$F1,$DF,$9D,$F0,$3F,$3F,$E1,$3E,$87,$C2,$CA,$F9,$91,$F4
	.dc.b	$FE,$17,$EA,$7C,$2C,$9F,$A1,$1F,$5F,$E0,$E4,$7D,$37,$ED,$7C,$2F
	.dc.b	$DB,$F8,$59,$3F,$5D,$FF,$47,$FE,$B9,$FC,$96,$3F,$C6,$0F,$F1,$F6
	.dc.b	$FC,$2E,$8D,$FC,$2F,$0F,$85,$E3,$F0,$B2,$BF,$CF,$77,$FC,$68,$FE
	.dc.b	$32,$FF,$AE,$5E,$DE,$99,$A6,$BA,$3E,$8C,$67,$0F,$05,$74,$C7,$C2
	.dc.b	$F6,$C7,$C0,$F5,$C7,$C8,$F8,$BE,$0F,$DB,$F0,$72,$30,$79,$BE,$0F
	.dc.b	$E0,$FC,$1C,$8C,$1E,$AF,$82,$F8,$BF,$07,$23,$07,$23,$D9,$87,$CA
	.dc.b	$3E,$5F,$C1,$7C,$CC,$3A,$4F,$C1,$C3,$0F,$23,$E2,$5F,$57,$E3,$61
	.dc.b	$C9,$F9,$77,$5F,$95,$F0,$7C,$9E,$87,$D0,$BC,$70,$7C,$AE,$1F,$F2
	.dc.b	$FC,$FD,$4C,$7B,$4E,$71,$FC,$0F,$C1,$CA,$FB,$2F,$E1,$3E,$0F,$F0
	.dc.b	$BF,$05,$F8,$6F,$83,$95,$F7,$A3,$F1,$3F,$05,$23,$F0,$2F,$E3,$3E
	.dc.b	$0F,$94,$7E,$12,$EB,$F8,$3F,$82,$EA,$FE,$23,$0E,$AF,$E1,$F0,$E8
	.dc.b	$FE,$27,$0B,$9F,$B7,$1F,$CC,$FC,$1F,$FB,$A2,$FF,$A0,$5F,$FA,$25
	.dc.b	$BD,$3F,$8F,$9D,$7C,$24,$AF,$F6,$8D,$7F,$4A,$3F,$A3,$79,$FF,$ED
	.dc.b	$E5,$F0,$9D,$7E,$0E,$57,$08,$F0,$7C,$1F,$8A,$3F,$43,$8E,$1E,$B8
	.dc.b	$F7,$5A,$7E,$7B,$1F,$07,$C2,$7A,$71,$EB,$5A,$CF,$C5,$9E,$63,$CF
	.dc.b	$F0,$7E,$FF,$84,$91,$83,$D7,$F0,$7F,$0F,$E1,$24,$60,$F7,$FC,$27
	.dc.b	$C7,$F8,$49,$18,$39,$1E,$13,$E5,$61,$F3,$4F,$9D,$F0,$9F,$3F,$0E
	.dc.b	$B3,$BE,$36,$EC,$B5,$3D,$4D,$43,$1F,$0C,$63,$2F,$B5,$F0,$9D,$A3
	.dc.b	$EA,$E5,$87,$D4,$72,$C3,$C3,$C3,$F2,$F3,$FC,$16,$1D,$23,$E7,$63
	.dc.b	$8B,$FB,$FF,$09,$F8,$7F,$84,$91,$83,$F0,$7F,$07,$F8,$BF,$84,$91
	.dc.b	$83,$F0,$FF,$09,$F8,$F2,$3F,$8A,$F8,$4F,$C9,$7C,$1C,$8F,$C6,$A6
	.dc.b	$67,$FC,$7B,$F5,$DF,$F4,$DF,$13,$2E,$F5,$F1,$B2,$FC,$E4,$76,$8F
	.dc.b	$93,$95,$C3,$F1,$59,$75,$9F,$A5,$8F,$F7,$27,$FD,$D4,$FF,$A6,$49
	.dc.b	$F0,$9B,$AF,$17,$C0,$DE,$9F,$37,$32,$DF,$05,$9C,$79,$B6,$75,$76
	.dc.b	$BF,$B5,$D7,$BA,$F9,$C6,$23,$CB,$F0,$5B,$F8,$1E,$1F,$07,$C7,$E0
	.dc.b	$A5,$71,$3D,$31,$D7,$1E,$D9,$C9,$D6,$7E,$26,$E5,$DF,$4F,$9D,$9F
	.dc.b	$94,$F0,$4E,$BE,$0B,$D0,$47,$8F,$C1,$7A,$FE,$0A,$47,$47,$DD,$F0
	.dc.b	$5E,$FF,$82,$93,$E0,$7E,$17,$C0,$FC,$3F,$82,$93,$E4,$7E,$37,$C0
	.dc.b	$F4,$8F,$AF,$85,$D3,$F0,$98,$F4,$9F,$B7,$8D,$9F,$C4,$63,$F3,$FE
	.dc.b	$0B,$E8,$7C,$14,$BF,$83,$1C,$F9,$BE,$8E,$56,$7F,$07,$97,$39,$FA
	.dc.b	$B9,$59,$FC,$3E,$3F,$6F,$E0,$A4,$7C,$C7,$EF,$7C,$17,$E0,$7F,$5E
	.dc.b	$FE,$F6,$7F,$81,$F9,$0D,$33,$F0,$5B,$F8,$29,$87,$D8,$9F,$F5,$52
	.dc.b	$AE,$3F,$05,$DA,$3E,$DE,$1D,$E3,$C1,$1E,$1F,$83,$8F,$33,$C9,$8F
	.dc.b	$67,$F8,$A6,$1F,$EA,$C3,$BE,$1E,$6F,$87,$F5,$12,$7C,$9F,$07,$ED
	.dc.b	$F8,$4F,$77,$C2,$49,$F4,$47,$C1,$F8,$2F,$85,$F0,$12,$7D,$91,$F1
	.dc.b	$7E,$07,$E3,$7C,$24,$9F,$81,$1F,$27,$E1,$3E,$57,$C1,$C9,$F8,$91
	.dc.b	$F3,$7E,$0E,$47,$C7,$7E,$87,$9D,$F4,$7E,$0B,$E9,$7C,$14,$AF,$99
	.dc.b	$1D,$5F,$9D,$87,$57,$E1,$E1,$D1,$F8,$B8,$5F,$03,$FC,$63,$C7,$FC
	.dc.b	$EA,$7E,$B7,$C1,$7D,$F2,$3F,$67,$E0,$FF,$05,$F0,$5F,$C0,$FF,$5F
	.dc.b	$FE,$DE,$7F,$70,$B4,$67,$E0,$F5,$F0,$73,$0F,$DC,$DF,$FA,$E8,$FE
	.dc.b	$DE,$E3,$E0,$BA,$1D,$7E,$0F,$B7,$C2,$49,$E2,$F8,$7E,$0F,$C5,$F0
	.dc.b	$9E,$3F,$81,$95,$DA,$3C,$DF,$05,$E7,$F8,$29,$3E,$28,$F5,$7C,$14
	.dc.b	$8F,$2B,$ED,$F8,$3F,$77,$C2,$49,$F4,$BF,$07,$E1,$3E,$17,$C2,$7C
	.dc.b	$3F,$82,$95,$EE,$8F,$8D,$F0,$7F,$1F,$E0,$E4,$FC,$28,$F9,$5F,$07
	.dc.b	$F2,$FE,$12,$4F,$C6,$8F,$E6,$1E,$CB,$E4,$6F,$C2,$CA,$FE,$47,$39
	.dc.b	$6F,$C4,$70,$F8,$39,$18,$37,$F0,$7D,$08,$F2,$F8,$2E,$FF,$0D,$E0
	.dc.b	$F8,$59,$3D,$A3,$C7,$F0,$DE,$4F,$85,$93,$E2,$8F,$3F,$C2,$FA,$3E
	.dc.b	$0A,$4F,$9A,$3D,$7F,$05,$23,$D2,$FB,$BE,$07,$DF,$F0,$32,$7D,$AF
	.dc.b	$C2,$F8,$5F,$87,$C3,$27,$E0,$BF,$1B,$E1,$3E,$3F,$C2,$C9,$F8,$AF
	.dc.b	$CA,$F8,$4F,$97,$F0,$72,$7E,$4B,$FD,$77,$F6,$0F,$F4,$B9,$FC,$74
	.dc.b	$9F,$07,$9F,$83,$D7,$C2,$CB,$FE,$A7,$48,$E5,$F0,$BC,$FE,$13,$A7
	.dc.b	$C2,$4B,$DC,$77,$F8,$39,$1C,$9F,$17,$C2,$78,$FE,$0E,$4F,$67,$CD
	.dc.b	$F0,$9E,$7F,$83,$93,$E2,$7D,$5F,$09,$EB,$F8,$09,$3E,$67,$DD,$F0
	.dc.b	$3E,$FF,$81,$F8,$1F,$09,$2B,$D7,$1F,$0F,$E0,$FE,$27,$C2,$49,$F7
	.dc.b	$C7,$C7,$F8,$3F,$91,$F0,$72,$7E,$1C,$7C,$BF,$82,$FF,$04,$7E,$C3
	.dc.b	$FE,$74,$3F,$CE,$BC,$FF,$C1,$AB,$FB,$DC,$3E,$0F,$64,$8C,$FC,$1F
	.dc.b	$2F,$81,$91,$C1,$EB,$F0,$5D,$BE,$0B,$BF,$C1,$4A,$EB,$1E,$2F,$81
	.dc.b	$91,$E0,$7C,$BF,$05,$E6,$F8,$19,$3E,$47,$D3,$F0,$5E,$AF,$82,$93
	.dc.b	$E8,$7D,$BF,$05,$EE,$F8,$29,$3E,$C7,$E0,$FC,$17,$C2,$F8,$19,$3F
	.dc.b	$01,$F8,$BF,$05,$F1,$BE,$06,$4F,$C4,$7E,$4F,$C1,$7C,$AF,$81,$F9
	.dc.b	$7F,$05,$2B,$E4,$C7,$F8,$D7,$F6,$2F,$E0,$51,$FE,$75,$7F,$C8,$9F
	.dc.b	$0E,$69,$AF,$87,$FF,$78,$AF,$83,$E5,$F0,$72,$FF,$C5,$F3,$96,$FC
	.dc.b	$1F,$7F,$83,$91,$83,$7F,$05,$E3,$F8,$39,$18,$39,$FC,$17,$9F,$E0
	.dc.b	$E4,$60,$EF,$F0,$5E,$B2,$3E,$2F,$83,$F7,$7C,$1F,$BF,$E0,$E4,$F9
	.dc.b	$A3,$E1,$7C,$1F,$C3,$F8,$39,$3E,$A8,$F8,$DF,$07,$F1,$FE,$0E,$4F
	.dc.b	$BA,$3E,$57,$C1,$C8,$F8,$2F,$CD,$F8,$3F,$9D,$F0,$72,$7E,$2B,$FE
	.dc.b	$65,$ED,$FF,$3A,$47,$FA,$95,$FF,$13,$13,$E0,$F5,$F0,$7B,$F8,$3E
	.dc.b	$1F,$07,$2F,$FC,$BB,$7F,$C2,$DF,$F3,$20,$FB,$E7,$2F,$FB,$32,$3F
	.dc.b	$A2,$7C,$2F,$FA,$0E,$73,$DD,$FE,$91,$F8,$64,$E6,$F8,$46,$2F,$FD
	.dc.b	$2D,$7C,$1E,$66,$27,$75,$C6,$FA,$36,$76,$79,$67,$6F,$19,$DB,$4D
	.dc.b	$D7,$7B,$F1,$47,$19,$DB,$EF,$E8,$FF,$23,$2F,$49,$E5,$FB,$E7,$31
	.dc.b	$89,$F5,$C6,$27,$5E,$17,$FA,$CE,$B5,$5D,$63,$13,$A8,$EC,$D9,$DE
	.dc.b	$2B,$E9,$5D,$AF,$8B,$4E,$25,$AE,$33,$C2,$3C,$06,$8B,$1A,$9E,$71
	.dc.b	$8B,$F1,$11,$E8,$D7,$46,$08,$60,$97,$E9,$9E,$AD,$8F,$13,$6B,$C5
	.dc.b	$18,$BF,$44,$E6,$FC,$0D,$9E,$ED,$7B,$CF,$02,$5F,$81,$B1,$D6,$30
	.dc.b	$47,$04,$BF,$19,$2F,$C3,$5C,$A3,$13,$E2,$8C,$4E,$AB,$A5,$6A,$BC
	.dc.b	$B1,$89,$D4,$77,$6C,$EE,$FD,$84,$BE,$B5,$DE,$F9,$34,$E4,$5A,$E5
	.dc.b	$3C,$23,$C2,$68,$B1,$A9,$E7,$18,$BF,$19,$1E,$8D,$74,$60,$86,$09
	.dc.b	$7E,$99,$EA,$D8,$F1,$36,$BC,$51,$8B,$F4,$4E,$6F,$C0,$D9,$EE,$D7
	.dc.b	$BC,$F0,$25,$F8,$1B,$1D,$63,$04,$70,$4B,$F1,$92,$FC,$35,$CA,$31
	.dc.b	$3E,$28,$C4,$EA,$BA,$56,$AB,$CB,$18,$9D,$47,$76,$CE,$EF,$D8,$4B
	.dc.b	$EB,$5D,$EF,$93,$4E,$45,$AE,$53,$C2,$3C,$26,$8B,$1A,$9E,$71,$8B
	.dc.b	$F1,$91,$E8,$D7,$46,$08,$60,$97,$E9,$9E,$AD,$8F,$13,$6B,$C5,$18
	.dc.b	$BF,$44,$E6,$FC,$0D,$9E,$ED,$7B,$CF,$02,$5F,$81,$B1,$D6,$30,$47
	.dc.b	$04,$BF,$19,$2F,$C3,$5C,$A3,$13,$E2,$8C,$4E,$AB,$A5,$6A,$BC,$B1
	.dc.b	$89,$D4,$77,$6C,$EE,$FD,$84,$BE,$B5,$DE,$F9,$34,$E4,$5A,$E5,$3C
	.dc.b	$23,$C2,$68,$B1,$A9,$E7,$18,$BF,$19,$1E,$8D,$74,$60,$86,$09,$7E
	.dc.b	$99,$EA,$D8,$F1,$36,$BC,$51,$8B,$F4,$4E,$6F,$C0,$D9,$EE,$D7,$BC
	.dc.b	$F0,$25,$F8,$1B,$1D,$63,$04,$70,$4B,$F1,$92,$FC,$35,$CA,$31,$3E
	.dc.b	$28,$C4,$EA,$BA,$56,$AB,$CB,$18,$9D,$47,$76,$CE,$EF,$D8,$4B,$EB
	.dc.b	$5D,$EF,$93,$4E,$45,$AE,$53,$C2,$3C,$26,$8B,$1A,$9E,$71,$8B,$F1
	.dc.b	$91,$E8,$D7,$46,$08,$60,$97,$E9,$9E,$AD,$8F,$13,$6B,$C5,$18,$BF
	.dc.b	$44,$E6,$FC,$0D,$9E,$ED,$7B,$CF,$02,$5F,$81,$B1,$D6,$30,$47,$04
	.dc.b	$BF,$19,$2F,$C3,$5C,$A3,$13,$E2,$8C,$4E,$AB,$A5,$6A,$BC,$B1,$89
	.dc.b	$D4,$77,$6C,$EE,$FD,$84,$BE,$B5,$DE,$F9,$34,$E4,$5A,$E5,$3C,$23
	.dc.b	$C2,$68,$B1,$A9,$E7,$18,$BF,$19,$1E,$8D,$74,$60,$86,$09,$7E,$99
	.dc.b	$EA,$D8,$F1,$36,$BC,$51,$8B,$F4,$4E,$6F,$C0,$D9,$EE,$D7,$BC,$F0
	.dc.b	$25,$F8,$1B,$1D,$63,$04,$70,$4B,$F1,$92,$FC,$35,$CA,$31,$3E,$28
	.dc.b	$C4,$EA,$BA,$56,$AB,$CB,$18,$9D,$47,$76,$CE,$EF,$D8,$4B,$EB,$5D
	.dc.b	$EF,$93,$4E,$45,$AE,$53,$C2,$3C,$26,$8B,$1A,$9E,$71,$8B,$F1,$91
	.dc.b	$E8,$D7,$46,$08,$60,$97,$E9,$9E,$AD,$8F,$13,$6B,$C5,$18,$BF,$44
	.dc.b	$E6,$FC,$0D,$9E,$ED,$7B,$CF,$02,$5F,$81,$B1,$D6,$30,$47,$04,$BF
	.dc.b	$19,$2F,$C3,$5C,$A3,$13,$E2,$8C,$4E,$AB,$A5,$6A,$BC,$B1,$89,$D4
	.dc.b	$77,$6C,$EE,$FD,$84,$BE,$B5,$DE,$F9,$34,$E4,$5A,$E5,$3C,$23,$C2
	.dc.b	$68,$B1,$A9,$E7,$18,$BF,$19,$1E,$8D,$74,$60,$86,$09,$7E,$99,$EA
	.dc.b	$D8,$F1,$36,$BC,$51,$8B,$F4,$4E,$6F,$C0,$D9,$EE,$D7,$BC,$F0,$25
	.dc.b	$F8,$1B,$1D,$63,$04,$70,$4B,$F1,$92,$FC,$35,$CA,$31,$3E,$28,$C4
	.dc.b	$EA,$BA,$56,$AB,$CB,$18,$9D,$47,$76,$CE,$EF,$D8,$4B,$EB,$5D,$EF
	.dc.b	$93,$4E,$45,$AE,$53,$C2,$3C,$26,$8B,$1A,$9E,$71,$8B,$F1,$91,$E8
	.dc.b	$D7,$46,$08,$60,$97,$E9,$9E,$AD,$8F,$13,$6B,$C5,$18,$BF,$44,$E6
	.dc.b	$FC,$0D,$9E,$ED,$7B,$CF,$02,$5F,$81,$B1,$D6,$30,$47,$04,$BF,$19
	.dc.b	$2F,$C3,$5C,$A3,$13,$E2,$8C,$4E,$AB,$A5,$6A,$BC,$B1,$89,$D4,$77
	.dc.b	$6C,$EE,$FD,$84,$BE,$B5,$DE,$F9,$34,$E4,$5A,$E5,$3C,$23,$C2,$68
	.dc.b	$B1,$A9,$E7,$18,$BF,$19,$1E,$8D,$74,$60,$86,$09,$7E,$99,$EA,$D8
	.dc.b	$F1,$36,$BC,$51,$8B,$F4,$4E,$6F,$C0,$D9,$EE,$D7,$BC,$F0,$25,$F8
	.dc.b	$1B,$1D,$63,$04,$70,$4B,$F1,$92,$FC,$35,$CA,$31,$3E,$28,$C4,$EA
	.dc.b	$BA,$56,$AB,$CB,$18,$9D,$47,$76,$CE,$EF,$D8,$4B,$EB,$5D,$EF,$93
	.dc.b	$4E,$45,$AE,$53,$C2,$3C,$26,$8B,$1A,$9E,$71,$8B,$F1,$91,$E8,$D7
	.dc.b	$46,$08,$60,$97,$E9,$9E,$AD,$8F,$13,$6B,$C5,$18,$BF,$44,$E6,$FC
	.dc.b	$0D,$9E,$ED,$7B,$CF,$02,$5F,$81,$B1,$D6,$30,$47,$04,$BF,$19,$2F
	.dc.b	$C3,$5C,$A3,$13,$E2,$8C,$4E,$AB,$A5,$6A,$BC,$B1,$89,$D4,$77,$6C
	.dc.b	$EE,$FD,$84,$BE,$B5,$DE,$F9,$34,$E4,$5A,$E5,$3C,$23,$C2,$68,$B1
	.dc.b	$A9,$E7,$18,$BF,$19,$1E,$8D,$74,$60,$86,$09,$7E,$99,$EA,$D8,$F1
	.dc.b	$36,$BC,$51,$8B,$F4,$4E,$6F,$C0,$D9,$EE,$D7,$BC,$F0,$25,$F8,$1B
	.dc.b	$1D,$63,$04,$70,$4B,$F1,$92,$FC,$35,$CA,$31,$3E,$28,$C4,$EA,$BA
	.dc.b	$56,$AB,$CB,$18,$9D,$47,$76,$CE,$EF,$D8,$4B,$EB,$5D,$EF,$93,$4E
	.dc.b	$45,$AE,$53,$C2,$3C,$26,$8B,$1A,$9E,$71,$8B,$F1,$91,$E8,$D7,$46
	.dc.b	$08,$60,$97,$E9,$9E,$AD,$8F,$13,$6B,$C5,$18,$BF,$44,$E6,$FC,$0D
	.dc.b	$9E,$ED,$7B,$CF,$02,$5F,$81,$B1,$D6,$30,$47,$04,$BF,$19,$2F,$C3
	.dc.b	$5C,$A3,$13,$E2,$8C,$4E,$AB,$A5,$6A,$BC,$B1,$89,$D4,$77,$6C,$EE
	.dc.b	$FD,$84,$BE,$B5,$DE,$F9,$34,$E4,$5A,$E5,$3C,$23,$C2,$68,$B1,$A9
	.dc.b	$E7,$18,$BF,$19,$1E,$8D,$74,$60,$86,$09,$7E,$99,$EA,$D8,$F1,$36
	.dc.b	$BC,$51,$8B,$F4,$4E,$6F,$C0,$D9,$EE,$D7,$BC,$F0,$25,$F8,$1B,$1D
	.dc.b	$63,$04,$70,$4B,$F1,$92,$FC,$35,$CA,$31,$3E,$28,$C4,$EA,$BA,$56
	.dc.b	$AB,$CB,$18,$9D,$47,$76,$CE,$EF,$D8,$4B,$EB,$5D,$EF,$93,$4E,$45
	.dc.b	$AE,$53,$C2,$3C,$26,$8B,$1A,$9E,$71,$8B,$F1,$91,$E8,$D7,$46,$08
	.dc.b	$60,$97,$E9,$9E,$AD,$8F,$13,$6B,$C5,$18,$BF,$44,$E6,$FC,$0D,$9E
	.dc.b	$ED,$7B,$CF,$02,$5F,$81,$B1,$D6,$30,$47,$04,$BF,$19,$2F,$C3,$5C
	.dc.b	$A3,$13,$E2,$8C,$4E,$AB,$A5,$6A,$BC,$B1,$89,$D4,$77,$6C,$EE,$FD
	.dc.b	$84,$BE,$B5,$DE,$F9,$34,$E4,$5A,$E5,$3C,$23,$C2,$68,$B1,$A9,$E7
	.dc.b	$18,$BF,$18,$65,$AD,$59,$B5,$EE,$D7,$BA,$C4,$66,$30,$E0,$DD,$F4
	.dc.b	$71,$1A,$78,$B5,$E5,$1B,$72,$D7,$33,$D5,$EA,$D9,$E5,$1E,$C9,$E0
	.dc.b	$D9,$F3,$FD,$60,$80

;--------------------------------------------------------------------------------
;	FBcc.L <label>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fbccl060_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68060+FPSP060,-(sp)
	peamsg	'FBCC.L <LABEL>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FBcc.L <label>',13,10
;------------------------------------------------
;	d1	actual result
;	d4	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result,expected status,...
;	a1	expected result
;	a4	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	fbccl060_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FBcc.L <forward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	jsr	([fbccl060_execute_forward,za0,d5.l*4])	;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.L <forward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
;FBcc.L <backward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	jsr	([fbccl060_execute_backward,za0,d5.l*4])	;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.L <backward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
fbccl060_execute_forward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fbccl060_execute_forward_&cc
  .endm
fbccl060_execute_backward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fbccl060_execute_backward_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
fbccl060_execute_forward_&cc::
	moveq.l	#2,d1			;actual result. 2=too long
	FB&cc.L	@f			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts

	subq.l	#2,d1			;actual result. -1=too short(forward)/too long(backward)
@@:
	subq.l	#1,d1			;actual result. 1=taken
	rts

fbccl060_execute_backward_&cc::
	moveq.l	#2,d1			;actual result. 2=too short
	FB&cc.L	@b			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts
  .endm
	.cpu	68000

	.align	4
fbccl060_expected_compressed::
	.dc.b	$00,$00,$10,$00,$10,$84,$21,$5A,$F0,$6B,$08,$E0,$C3,$0E,$47,$27
	.dc.b	$0C,$0F,$17,$07,$27,$0E,$0A,$74,$30,$75,$21,$83,$D6,$F0,$78,$10
	.dc.b	$F8,$2E,$5F,$88,$D7,$E0,$1D,$0D,$BE,$B2,$3D,$9A,$FC,$96,$BC,$0F
	.dc.b	$37,$E0,$43,$E2,$95,$F1,$99,$72,$47,$2E,$DF,$39,$B3,$2F,$73,$4F
	.dc.b	$CE,$8F,$C0,$C6,$1F,$A3,$1C,$4F,$BC,$EA,$39,$12,$3E,$2C,$65,$A7
	.dc.b	$13,$F1,$B3,$F8,$63,$F0,$04,$7F,$16,$E5,$FA,$AE,$5F,$C8,$B9,$7F
	.dc.b	$24,$E5,$FC,$B3,$97,$EC,$B9,$7F,$36,$E5,$FB,$6E,$5F,$D0,$39,$7F
	.dc.b	$38,$E5,$F4,$9E,$63,$CA,$43,$CA,$64,$F2,$99,$3F,$55,$5F,$A6,$23
	.dc.b	$FA,$A6,$BF,$AE,$6B,$C0,$FC,$BB,$FA,$72,$38,$21,$FB,$57,$2D,$3B
	.dc.b	$1F,$4C,$3F,$5D,$FE,$00,$D7,$F8,$1B,$5D,$9C,$5C,$11,$FB,$D3,$AA
	.dc.b	$FC,$54,$F3,$8C,$47,$EB,$23,$04,$9E,$2D,$8F,$CA,$C2,$FF,$83,$1F
	.dc.b	$C3,$8A,$FF,$12,$6B,$FC,$59,$AF,$F1,$96,$BF,$C7,$1A,$FF,$1F,$6B
	.dc.b	$FC,$89,$AF,$84,$FE,$0E,$F8,$5A,$78,$4A,$78,$4A,$78,$4A,$78,$4A
	.dc.b	$78,$4A,$78,$4A,$30,$FE,$4C,$7F,$2E,$2B,$FC,$C9,$AF,$F3,$67,$4F
	.dc.b	$F1,$E7,$2F,$F1,$C6,$BF,$C7,$1A,$FF,$41,$6B,$FD,$11,$FB,$47,$E0
	.dc.b	$CA,$FE,$2D,$B3,$F6,$1D,$47,$E1,$1C,$C7,$86,$30,$47,$F8,$84,$70
	.dc.b	$6C,$7E,$5F,$1F,$E8,$C7,$F4,$E2,$BF,$D4,$9A,$FF,$56,$6B,$FD,$65
	.dc.b	$AF,$F5,$C6,$BF,$D7,$DA,$FF,$62,$6A,$69,$1F,$E9,$2F,$8D,$A7,$8C
	.dc.b	$A7,$8C,$A7,$8C,$A7,$8C,$A7,$F7,$1C,$BF,$B2,$1F,$DB,$0A,$FF,$70
	.dc.b	$6B,$FD,$D1,$D3,$FD,$71,$CB,$FD,$69,$AF,$F5,$A6,$BD,$CE,$07,$53
	.dc.b	$F7,$6F,$F0,$13,$F7,$45,$72,$7F,$0B,$71,$3F,$D2,$5D,$C6,$A3,$0D
	.dc.b	$7C,$8F,$13,$07,$F8,$43,$3F,$F0,$59,$1F,$F0,$7B,$FE,$07,$23,$FD
	.dc.b	$E9,$AF,$F8,$59,$E0,$D3,$89,$B2,$1F,$E1,$B7,$8B,$FE,$1F,$78,$B4
	.dc.b	$E2,$6C,$E2,$7F,$81,$9E,$2F,$A4,$D9,$C4,$F1,$1C,$48,$71,$39,$9F
	.dc.b	$E3,$0D,$3F,$C5,$64,$7F,$C5,$ED,$7F,$C6,$6D,$7F,$C6,$CF,$F8,$AC
	.dc.b	$8F,$F8,$79,$AF,$F8,$E5,$AF,$F8,$79,$AF,$84,$FE,$8C,$FF,$41,$23
	.dc.b	$82,$4F,$F3,$68,$CB,$FE,$09,$7F,$A0,$12,$3F,$A2,$B6,$35,$18,$23
	.dc.b	$FD,$03,$5F,$F2,$59,$23,$FC,$9E,$FF,$91,$C8,$FF,$8E,$9A,$FF,$95
	.dc.b	$9E,$0D,$38,$9B,$21,$FE,$5B,$78,$BF,$E5,$F7,$8B,$4E,$26,$CE,$27
	.dc.b	$F9,$19,$E2,$FF,$9A,$1E,$2D,$38,$9B,$38,$9F,$E4,$C7,$8B,$FE,$71
	.dc.b	$DB,$FC,$D6,$47,$FC,$DE,$D7,$FC,$E6,$D7,$FC,$EC,$FF,$9A,$C8,$FF
	.dc.b	$97,$9A,$FF,$9E,$5A,$FF,$97,$9A,$FF,$A0,$1F,$F0,$01,$1F,$F0,$43
	.dc.b	$67,$FB,$AB,$5F,$0C,$7F,$80,$A3,$FC,$00,$47,$04,$8F,$F2,$74,$65
	.dc.b	$B1,$FE,$15,$DF,$FD,$08,$7F,$A3,$0A,$FF,$A4,$1A,$FF,$A5,$1A,$FF
	.dc.b	$A5,$DA,$FF,$A6,$9A,$FF,$A7,$5A,$FF,$A8,$1A,$9A,$47,$FD,$0E,$F8
	.dc.b	$DA,$78,$CA,$78,$CA,$78,$CA,$78,$CA,$7F,$AB,$F8,$3F,$D4,$47,$FA
	.dc.b	$A8,$AF,$FA,$B9,$AF,$FA,$C9,$D3,$FE,$9A,$72,$FF,$A6,$1A,$FF,$A6
	.dc.b	$1A,$F7,$38,$1D,$4F,$F1,$CB,$FE,$3F,$3F,$C7,$05,$72,$7F,$92,$5C
	.dc.b	$4F,$FA,$1D,$DC,$6A,$30,$D7,$C8,$F1,$30,$7F,$B3,$B8,$7F,$D8,$47
	.dc.b	$FB,$28,$AF,$FB,$39,$AF,$FB,$49,$AF,$FB,$55,$AF,$33,$FD,$7A,$F3
	.dc.b	$69,$CC,$A7,$32,$9C,$CA,$73,$3F,$DC,$0F,$36,$9C,$CA,$73,$29,$CC
	.dc.b	$A7,$33,$FD,$9A,$FF,$BA,$38,$BF,$DC,$45,$8F,$F6,$AB,$5F,$F7,$83
	.dc.b	$5F,$F7,$A3,$C1,$D9,$0D,$94,$FF,$6D,$BC,$5F,$F6,$DB,$DE,$7F,$D8
	.dc.b	$0F,$F9,$FC,$FF,$3C,$16,$3F,$CF,$AD,$8F,$F3,$AB,$A8,$C9,$23,$9B
	.dc.b	$5F,$0B,$FE,$91,$8E,$10,$00

;--------------------------------------------------------------------------------
;	FBcc.L <label>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fbccl88x_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040,-(sp)
	peamsg	'FBCC.L <LABEL>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FBcc.L <label>',13,10
;------------------------------------------------
;	d1	actual result
;	d4	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result,expected status,...
;	a1	expected result
;	a4	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	fbccl88x_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FBcc.L <forward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	jsr	([fbccl88x_execute_forward,za0,d5.l*4])	;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.L <forward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
;FBcc.L <backward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	jsr	([fbccl88x_execute_backward,za0,d5.l*4])	;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.L <backward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
fbccl88x_execute_forward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fbccl88x_execute_forward_&cc
  .endm
fbccl88x_execute_backward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fbccl88x_execute_backward_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
fbccl88x_execute_forward_&cc::
	moveq.l	#2,d1			;actual result. 2=too long
	FB&cc.L	@f			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts

	subq.l	#2,d1			;actual result. -1=too short(forward)/too long(backward)
@@:
	subq.l	#1,d1			;actual result. 1=taken
	rts

fbccl88x_execute_backward_&cc::
	moveq.l	#2,d1			;actual result. 2=too short
	FB&cc.L	@b			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts
  .endm
	.cpu	68000

	.align	4
fbccl88x_expected_compressed::
	.dc.b	$00,$00,$10,$00,$10,$84,$21,$5A,$F0,$6B,$08,$E0,$C3,$0E,$47,$27
	.dc.b	$0C,$0F,$17,$07,$27,$0E,$0A,$74,$30,$75,$21,$83,$D6,$F0,$78,$10
	.dc.b	$F8,$2E,$5F,$88,$D7,$E0,$1D,$0D,$BE,$B2,$3D,$9A,$FC,$96,$BC,$0F
	.dc.b	$37,$E0,$43,$E2,$95,$F1,$99,$72,$47,$2E,$DF,$39,$B3,$2F,$73,$4F
	.dc.b	$CE,$8F,$C0,$C6,$1F,$A3,$1C,$4F,$BC,$EA,$39,$12,$3E,$2C,$65,$A7
	.dc.b	$13,$F1,$B3,$F8,$63,$F0,$04,$7F,$16,$E5,$FA,$AE,$5F,$C8,$B9,$7F
	.dc.b	$24,$E5,$FC,$B3,$97,$EC,$B9,$7F,$36,$E5,$FB,$6E,$5F,$D0,$39,$7F
	.dc.b	$38,$E5,$F4,$9E,$63,$CA,$43,$CA,$64,$F2,$99,$3F,$55,$5F,$A6,$23
	.dc.b	$FA,$A6,$BF,$AE,$6B,$C0,$FC,$BB,$FA,$72,$38,$21,$FB,$57,$2D,$3B
	.dc.b	$1F,$4C,$3F,$5D,$FE,$00,$D7,$F8,$1B,$5D,$9C,$5C,$11,$FB,$D3,$AA
	.dc.b	$FC,$54,$F3,$8C,$47,$EB,$23,$04,$9E,$2D,$8F,$CA,$C2,$FF,$83,$1F
	.dc.b	$C3,$8A,$FF,$12,$6B,$FC,$59,$AF,$F1,$96,$BF,$C7,$1A,$FF,$1F,$6B
	.dc.b	$FC,$89,$AF,$84,$FE,$0E,$F8,$5A,$78,$4A,$78,$4A,$78,$4A,$78,$4A
	.dc.b	$78,$4A,$78,$4A,$30,$FE,$4C,$7F,$2E,$2B,$FC,$C9,$AF,$F3,$66,$BF
	.dc.b	$C6,$9A,$FF,$1A,$6B,$FC,$E5,$AF,$F3,$06,$BE,$13,$EE,$BF,$83,$3E
	.dc.b	$E1,$5C,$9F,$8D,$72,$E2,$7F,$86,$46,$1B,$19,$7C,$CD,$3C,$C7,$E5
	.dc.b	$F1,$FE,$8C,$7F,$4E,$2B,$FD,$49,$AF,$F5,$66,$BF,$D6,$5A,$FF,$5C
	.dc.b	$6B,$FD,$7D,$AF,$F6,$26,$A6,$91,$FE,$92,$F8,$DA,$78,$CA,$78,$CA
	.dc.b	$78,$CA,$78,$CA,$7F,$71,$CB,$FB,$21,$FD,$B0,$AF,$F7,$06,$BF,$DD
	.dc.b	$1A,$FF,$58,$6B,$FD,$61,$AF,$F7,$76,$BF,$DB,$9A,$9F,$DC,$1F,$C0
	.dc.b	$4A,$FF,$09,$6C,$FE,$D9,$AF,$86,$3F,$81,$C6,$1F,$11,$82,$47,$F4
	.dc.b	$F8,$D4,$7F,$84,$73,$FF,$05,$91,$FF,$07,$BF,$E0,$72,$3F,$DE,$9A
	.dc.b	$FF,$85,$9E,$0D,$38,$9B,$21,$FE,$1B,$78,$BF,$E1,$F7,$8B,$4E,$26
	.dc.b	$CE,$27,$F8,$19,$E2,$FF,$8A,$1E,$2D,$38,$9B,$38,$9F,$E0,$C7,$8B
	.dc.b	$FE,$31,$D3,$FC,$56,$47,$FC,$5E,$D7,$FC,$66,$D7,$FC,$6C,$FF,$8A
	.dc.b	$C8,$FF,$87,$9A,$FF,$8E,$5A,$FF,$87,$9A,$FF,$90,$1F,$E8,$04,$7F
	.dc.b	$A4,$36,$BF,$9A,$B5,$F0,$C7,$F4,$28,$FE,$80,$47,$04,$8F,$F0,$74
	.dc.b	$65,$B1,$FD,$57,$5F,$F2,$59,$1F,$F2,$7B,$FE,$47,$23,$FE,$3A,$6B
	.dc.b	$FE,$56,$78,$34,$E2,$6C,$87,$F9,$6D,$E2,$FF,$97,$DE,$2D,$38,$9B
	.dc.b	$38,$9F,$E4,$67,$8B,$FE,$68,$78,$B4,$E2,$6C,$E2,$7F,$93,$1E,$2F
	.dc.b	$F9,$C7,$6F,$F3,$59,$1F,$F3,$7B,$5F,$F3,$9B,$5F,$F3,$B3,$FE,$6B
	.dc.b	$23,$FE,$5E,$6B,$FE,$79,$6B,$FE,$5E,$6B,$FE,$80,$7F,$C0,$04,$7F
	.dc.b	$C1,$0D,$9F,$EE,$AD,$7C,$31,$FE,$02,$8F,$F0,$01,$1C,$12,$3F,$C9
	.dc.b	$D1,$96,$C7,$F8,$57,$7F,$F4,$21,$FE,$8C,$2B,$FE,$90,$6B,$FE,$94
	.dc.b	$6B,$FE,$97,$6B,$FE,$9A,$6B,$FE,$9D,$6B,$FE,$A0,$6A,$69,$1F,$F4
	.dc.b	$3B,$E3,$69,$E3,$29,$E3,$29,$E3,$29,$E3,$29,$FE,$AF,$E0,$FF,$51
	.dc.b	$1F,$EA,$A2,$BF,$EA,$E6,$BF,$EB,$26,$BF,$E9,$66,$BF,$E9,$66,$BF
	.dc.b	$EB,$56,$BF,$EA,$C6,$A7,$FC,$6C,$7F,$8F,$CA,$FF,$91,$DB,$3F,$E3
	.dc.b	$46,$BE,$18,$FF,$20,$C6,$1F,$11,$82,$47,$FA,$36,$35,$1F,$EC,$FE
	.dc.b	$1F,$F6,$11,$FE,$CA,$2B,$FE,$CE,$6B,$FE,$D2,$6B,$FE,$D5,$6B,$FE
	.dc.b	$D8,$6B,$D4,$FF,$5F,$BD,$5A,$75,$29,$D4,$A7,$52,$9D,$4A,$75,$3F
	.dc.b	$DC,$8F,$56,$9D,$4A,$75,$29,$FE,$ED,$E2,$FF,$71,$1F,$EE,$A2,$BF
	.dc.b	$EE,$E6,$BF,$EF,$26,$BF,$ED,$66,$BF,$ED,$66,$BF,$EF,$56,$BF,$EE
	.dc.b	$C6,$A7,$FC,$EC,$7F,$9F,$CA,$FF,$A1,$DB,$5F,$E7,$46,$BE,$18,$FF
	.dc.b	$40,$C6,$1F,$11,$82,$47,$FB,$36,$39,$A0

;--------------------------------------------------------------------------------
;	FBcc.W <label>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fbccw060_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68060+FPSP060,-(sp)
	peamsg	'FBCC.W <LABEL>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FBcc.W <label>',13,10
;------------------------------------------------
;	d1	actual result
;	d4	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result,expected status,...
;	a1	expected result
;	a4	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	fbccw060_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FBcc.W <forward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	jsr	([fbccw060_execute_forward,za0,d5.l*4])	;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.W <forward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
;FBcc.W <backward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	jsr	([fbccw060_execute_backward,za0,d5.l*4])	;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.W <backward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
fbccw060_execute_forward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fbccw060_execute_forward_&cc
  .endm
fbccw060_execute_backward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fbccw060_execute_backward_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
fbccw060_execute_forward_&cc::
	moveq.l	#2,d1			;actual result. 2=too long
	FB&cc.W	@f			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts

	subq.l	#2,d1			;actual result. -1=too short(forward)/too long(backward)
@@:
	subq.l	#1,d1			;actual result. 1=taken
	rts

fbccw060_execute_backward_&cc::
	moveq.l	#2,d1			;actual result. 2=too short
	FB&cc.W	@b			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts
  .endm
	.cpu	68000

	.align	4
fbccw060_expected_compressed::
	.dc.b	$00,$00,$10,$00,$10,$84,$21,$5A,$F0,$6B,$08,$E0,$C3,$0E,$47,$27
	.dc.b	$0C,$0F,$17,$07,$27,$0E,$0A,$74,$30,$75,$21,$83,$D6,$F0,$78,$10
	.dc.b	$F8,$2E,$5F,$88,$D7,$E0,$1D,$0D,$BE,$B2,$3D,$9A,$FC,$96,$BC,$0F
	.dc.b	$37,$E0,$43,$E2,$95,$F1,$99,$72,$47,$2E,$DF,$39,$B3,$2F,$73,$4F
	.dc.b	$CE,$8F,$C0,$C6,$1F,$A3,$1C,$4F,$BC,$EA,$39,$12,$3E,$2C,$65,$A7
	.dc.b	$13,$F1,$B3,$F8,$63,$F0,$04,$7F,$16,$E5,$FA,$AE,$5F,$C8,$B9,$7F
	.dc.b	$24,$E5,$FC,$B3,$97,$EC,$B9,$7F,$36,$E5,$FB,$6E,$5F,$D0,$39,$7F
	.dc.b	$38,$E5,$F4,$9E,$63,$CA,$43,$CA,$64,$F2,$99,$3F,$55,$5F,$A6,$23
	.dc.b	$FA,$A6,$BF,$AE,$6B,$C0,$FC,$BB,$FA,$72,$38,$21,$FB,$57,$2D,$3B
	.dc.b	$1F,$4C,$3F,$5D,$FE,$00,$D7,$F8,$1B,$5D,$9C,$5C,$11,$FB,$D3,$AA
	.dc.b	$FC,$54,$F3,$8C,$47,$EB,$23,$04,$9E,$2D,$8F,$CA,$C2,$FF,$83,$1F
	.dc.b	$C3,$8A,$FF,$12,$6B,$FC,$59,$AF,$F1,$96,$BF,$C7,$1A,$FF,$1F,$6B
	.dc.b	$FC,$89,$AF,$84,$FE,$0E,$F8,$5A,$78,$4A,$78,$4A,$78,$4A,$78,$4A
	.dc.b	$78,$4A,$78,$4A,$30,$FE,$4C,$7F,$2E,$2B,$FC,$C9,$AF,$F3,$67,$4F
	.dc.b	$F1,$E7,$2F,$F1,$C6,$BF,$C7,$1A,$FF,$41,$6B,$FD,$11,$FB,$47,$E0
	.dc.b	$CA,$FE,$2D,$B3,$F6,$1D,$47,$E1,$1C,$C7,$86,$30,$47,$F8,$84,$70
	.dc.b	$6C,$7E,$5F,$1F,$E8,$C7,$F4,$E2,$BF,$D4,$9A,$FF,$56,$6B,$FD,$65
	.dc.b	$AF,$F5,$C6,$BF,$D7,$DA,$FF,$62,$6A,$69,$1F,$E9,$2F,$8D,$A7,$8C
	.dc.b	$A7,$8C,$A7,$8C,$A7,$8C,$A7,$F7,$1C,$BF,$B2,$1F,$DB,$0A,$FF,$70
	.dc.b	$6B,$FD,$D1,$D3,$FD,$71,$CB,$FD,$69,$AF,$F5,$A6,$BD,$CE,$07,$53
	.dc.b	$F7,$6F,$F0,$13,$F7,$45,$72,$7F,$0B,$71,$3F,$D2,$5D,$C6,$A3,$0D
	.dc.b	$7C,$8F,$13,$07,$F8,$43,$3F,$F0,$59,$1F,$F0,$7B,$FE,$07,$23,$FD
	.dc.b	$E9,$AF,$F8,$59,$E0,$D3,$89,$B2,$1F,$E1,$B7,$8B,$FE,$1F,$78,$B4
	.dc.b	$E2,$6C,$E2,$7F,$81,$9E,$2F,$A4,$D9,$C4,$F1,$1C,$48,$71,$39,$9F
	.dc.b	$E3,$0D,$3F,$C5,$64,$7F,$C5,$ED,$7F,$C6,$6D,$7F,$C6,$CF,$F8,$AC
	.dc.b	$8F,$F8,$79,$AF,$F8,$E5,$AF,$F8,$79,$AF,$84,$FE,$8C,$FF,$41,$23
	.dc.b	$82,$4F,$F3,$68,$CB,$FE,$09,$7F,$A0,$12,$3F,$A2,$B6,$35,$18,$23
	.dc.b	$FD,$03,$5F,$F2,$59,$23,$FC,$9E,$FF,$91,$C8,$FF,$8E,$9A,$FF,$95
	.dc.b	$9E,$0D,$38,$9B,$21,$FE,$5B,$78,$BF,$E5,$F7,$8B,$4E,$26,$CE,$27
	.dc.b	$F9,$19,$E2,$FF,$9A,$1E,$2D,$38,$9B,$38,$9F,$E4,$C7,$8B,$FE,$71
	.dc.b	$DB,$FC,$D6,$47,$FC,$DE,$D7,$FC,$E6,$D7,$FC,$EC,$FF,$9A,$C8,$FF
	.dc.b	$97,$9A,$FF,$9E,$5A,$FF,$97,$9A,$FF,$A0,$1F,$F0,$01,$1F,$F0,$43
	.dc.b	$67,$FB,$AB,$5F,$0C,$7F,$80,$A3,$FC,$00,$47,$04,$8F,$F2,$74,$65
	.dc.b	$B1,$FE,$15,$DF,$FD,$08,$7F,$A3,$0A,$FF,$A4,$1A,$FF,$A5,$1A,$FF
	.dc.b	$A5,$DA,$FF,$A6,$9A,$FF,$A7,$5A,$FF,$A8,$1A,$9A,$47,$FD,$0E,$F8
	.dc.b	$DA,$78,$CA,$78,$CA,$78,$CA,$78,$CA,$7F,$AB,$F8,$3F,$D4,$47,$FA
	.dc.b	$A8,$AF,$FA,$B9,$AF,$FA,$C9,$D3,$FE,$9A,$72,$FF,$A6,$1A,$FF,$A6
	.dc.b	$1A,$F7,$38,$1D,$4F,$F1,$CB,$FE,$3F,$3F,$C7,$05,$72,$7F,$92,$5C
	.dc.b	$4F,$FA,$1D,$DC,$6A,$30,$D7,$C8,$F1,$30,$7F,$B3,$B8,$7F,$D8,$47
	.dc.b	$FB,$28,$AF,$FB,$39,$AF,$FB,$49,$AF,$FB,$55,$AF,$33,$FD,$7A,$F3
	.dc.b	$69,$CC,$A7,$32,$9C,$CA,$73,$3F,$DC,$0F,$36,$9C,$CA,$73,$29,$CC
	.dc.b	$A7,$33,$FD,$9A,$FF,$BA,$38,$BF,$DC,$45,$8F,$F6,$AB,$5F,$F7,$83
	.dc.b	$5F,$F7,$A3,$C1,$D9,$0D,$94,$FF,$6D,$BC,$5F,$F6,$DB,$DE,$7F,$D8
	.dc.b	$0F,$F9,$FC,$FF,$3C,$16,$3F,$CF,$AD,$8F,$F3,$AB,$A8,$C9,$23,$9B
	.dc.b	$5F,$0B,$FE,$91,$8E,$10,$00

;--------------------------------------------------------------------------------
;	FBcc.W <label>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fbccw88x_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040,-(sp)
	peamsg	'FBCC.W <LABEL>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FBcc.W <label>',13,10
;------------------------------------------------
;	d1	actual result
;	d4	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result,expected status,...
;	a1	expected result
;	a4	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	fbccw88x_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FBcc.W <forward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	jsr	([fbccw88x_execute_forward,za0,d5.l*4])	;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.W <forward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
;FBcc.W <backward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	jsr	([fbccw88x_execute_backward,za0,d5.l*4])	;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.W <backward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
fbccw88x_execute_forward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fbccw88x_execute_forward_&cc
  .endm
fbccw88x_execute_backward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fbccw88x_execute_backward_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
fbccw88x_execute_forward_&cc::
	moveq.l	#2,d1			;actual result. 2=too long
	FB&cc.W	@f			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts

	subq.l	#2,d1			;actual result. -1=too short(forward)/too long(backward)
@@:
	subq.l	#1,d1			;actual result. 1=taken
	rts

fbccw88x_execute_backward_&cc::
	moveq.l	#2,d1			;actual result. 2=too short
	FB&cc.W	@b			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts
  .endm
	.cpu	68000

	.align	4
fbccw88x_expected_compressed::
	.dc.b	$00,$00,$10,$00,$10,$84,$21,$5A,$F0,$6B,$08,$E0,$C3,$0E,$47,$27
	.dc.b	$0C,$0F,$17,$07,$27,$0E,$0A,$74,$30,$75,$21,$83,$D6,$F0,$78,$10
	.dc.b	$F8,$2E,$5F,$88,$D7,$E0,$1D,$0D,$BE,$B2,$3D,$9A,$FC,$96,$BC,$0F
	.dc.b	$37,$E0,$43,$E2,$95,$F1,$99,$72,$47,$2E,$DF,$39,$B3,$2F,$73,$4F
	.dc.b	$CE,$8F,$C0,$C6,$1F,$A3,$1C,$4F,$BC,$EA,$39,$12,$3E,$2C,$65,$A7
	.dc.b	$13,$F1,$B3,$F8,$63,$F0,$04,$7F,$16,$E5,$FA,$AE,$5F,$C8,$B9,$7F
	.dc.b	$24,$E5,$FC,$B3,$97,$EC,$B9,$7F,$36,$E5,$FB,$6E,$5F,$D0,$39,$7F
	.dc.b	$38,$E5,$F4,$9E,$63,$CA,$43,$CA,$64,$F2,$99,$3F,$55,$5F,$A6,$23
	.dc.b	$FA,$A6,$BF,$AE,$6B,$C0,$FC,$BB,$FA,$72,$38,$21,$FB,$57,$2D,$3B
	.dc.b	$1F,$4C,$3F,$5D,$FE,$00,$D7,$F8,$1B,$5D,$9C,$5C,$11,$FB,$D3,$AA
	.dc.b	$FC,$54,$F3,$8C,$47,$EB,$23,$04,$9E,$2D,$8F,$CA,$C2,$FF,$83,$1F
	.dc.b	$C3,$8A,$FF,$12,$6B,$FC,$59,$AF,$F1,$96,$BF,$C7,$1A,$FF,$1F,$6B
	.dc.b	$FC,$89,$AF,$84,$FE,$0E,$F8,$5A,$78,$4A,$78,$4A,$78,$4A,$78,$4A
	.dc.b	$78,$4A,$78,$4A,$30,$FE,$4C,$7F,$2E,$2B,$FC,$C9,$AF,$F3,$66,$BF
	.dc.b	$C6,$9A,$FF,$1A,$6B,$FC,$E5,$AF,$F3,$06,$BE,$13,$EE,$BF,$83,$3E
	.dc.b	$E1,$5C,$9F,$8D,$72,$E2,$7F,$86,$46,$1B,$19,$7C,$CD,$3C,$C7,$E5
	.dc.b	$F1,$FE,$8C,$7F,$4E,$2B,$FD,$49,$AF,$F5,$66,$BF,$D6,$5A,$FF,$5C
	.dc.b	$6B,$FD,$7D,$AF,$F6,$26,$A6,$91,$FE,$92,$F8,$DA,$78,$CA,$78,$CA
	.dc.b	$78,$CA,$78,$CA,$7F,$71,$CB,$FB,$21,$FD,$B0,$AF,$F7,$06,$BF,$DD
	.dc.b	$1A,$FF,$58,$6B,$FD,$61,$AF,$F7,$76,$BF,$DB,$9A,$9F,$DC,$1F,$C0
	.dc.b	$4A,$FF,$09,$6C,$FE,$D9,$AF,$86,$3F,$81,$C6,$1F,$11,$82,$47,$F4
	.dc.b	$F8,$D4,$7F,$84,$73,$FF,$05,$91,$FF,$07,$BF,$E0,$72,$3F,$DE,$9A
	.dc.b	$FF,$85,$9E,$0D,$38,$9B,$21,$FE,$1B,$78,$BF,$E1,$F7,$8B,$4E,$26
	.dc.b	$CE,$27,$F8,$19,$E2,$FF,$8A,$1E,$2D,$38,$9B,$38,$9F,$E0,$C7,$8B
	.dc.b	$FE,$31,$D3,$FC,$56,$47,$FC,$5E,$D7,$FC,$66,$D7,$FC,$6C,$FF,$8A
	.dc.b	$C8,$FF,$87,$9A,$FF,$8E,$5A,$FF,$87,$9A,$FF,$90,$1F,$E8,$04,$7F
	.dc.b	$A4,$36,$BF,$9A,$B5,$F0,$C7,$F4,$28,$FE,$80,$47,$04,$8F,$F0,$74
	.dc.b	$65,$B1,$FD,$57,$5F,$F2,$59,$1F,$F2,$7B,$FE,$47,$23,$FE,$3A,$6B
	.dc.b	$FE,$56,$78,$34,$E2,$6C,$87,$F9,$6D,$E2,$FF,$97,$DE,$2D,$38,$9B
	.dc.b	$38,$9F,$E4,$67,$8B,$FE,$68,$78,$B4,$E2,$6C,$E2,$7F,$93,$1E,$2F
	.dc.b	$F9,$C7,$6F,$F3,$59,$1F,$F3,$7B,$5F,$F3,$9B,$5F,$F3,$B3,$FE,$6B
	.dc.b	$23,$FE,$5E,$6B,$FE,$79,$6B,$FE,$5E,$6B,$FE,$80,$7F,$C0,$04,$7F
	.dc.b	$C1,$0D,$9F,$EE,$AD,$7C,$31,$FE,$02,$8F,$F0,$01,$1C,$12,$3F,$C9
	.dc.b	$D1,$96,$C7,$F8,$57,$7F,$F4,$21,$FE,$8C,$2B,$FE,$90,$6B,$FE,$94
	.dc.b	$6B,$FE,$97,$6B,$FE,$9A,$6B,$FE,$9D,$6B,$FE,$A0,$6A,$69,$1F,$F4
	.dc.b	$3B,$E3,$69,$E3,$29,$E3,$29,$E3,$29,$E3,$29,$FE,$AF,$E0,$FF,$51
	.dc.b	$1F,$EA,$A2,$BF,$EA,$E6,$BF,$EB,$26,$BF,$E9,$66,$BF,$E9,$66,$BF
	.dc.b	$EB,$56,$BF,$EA,$C6,$A7,$FC,$6C,$7F,$8F,$CA,$FF,$91,$DB,$3F,$E3
	.dc.b	$46,$BE,$18,$FF,$20,$C6,$1F,$11,$82,$47,$FA,$36,$35,$1F,$EC,$FE
	.dc.b	$1F,$F6,$11,$FE,$CA,$2B,$FE,$CE,$6B,$FE,$D2,$6B,$FE,$D5,$6B,$FE
	.dc.b	$D8,$6B,$D4,$FF,$5F,$BD,$5A,$75,$29,$D4,$A7,$52,$9D,$4A,$75,$3F
	.dc.b	$DC,$8F,$56,$9D,$4A,$75,$29,$FE,$ED,$E2,$FF,$71,$1F,$EE,$A2,$BF
	.dc.b	$EE,$E6,$BF,$EF,$26,$BF,$ED,$66,$BF,$ED,$66,$BF,$EF,$56,$BF,$EE
	.dc.b	$C6,$A7,$FC,$EC,$7F,$9F,$CA,$FF,$A1,$DB,$5F,$E7,$46,$BE,$18,$FF
	.dc.b	$40,$C6,$1F,$11,$82,$47,$FB,$36,$39,$A0

;--------------------------------------------------------------------------------
;	FCMP.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fcmp_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#63,-(sp)
	peamsg	'FCMP.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FCMP.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected status,...
;	fp2	source
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fcmp_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected status,...
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fcmp.x	fp3,fp3		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	move.l	(a4),d5			;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	move.l	d7,-(sp)		;actual status
	jbsr	test_status
	lea.l	(12,sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FCMP.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	move.l	d7,-(sp)		;actual status
	jbsr	output_status
	lea.l	(12,sp),sp
@@:
;
	addq.l	#4,a4			;expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fcmp.x	fp2,fp3		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	move.l	(a4),d5			;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	move.l	d7,-(sp)		;actual status
	jbsr	test_status
	lea.l	(12,sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FCMP.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	move.l	d7,-(sp)		;actual status
	jbsr	output_status
	lea.l	(12,sp),sp
@@:
;
	addq.l	#4,a4			;expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fcmp_data_compressed::
	.dc.b	$00,$07,$5C,$18,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$F2,$5F,$D9,$89,$5F,$CF,$88,$7F,$69,$29,$FC,$D8
	.dc.b	$C1,$0C,$11,$E6,$E5,$AE,$4E,$6E,$08,$E0,$87,$12,$9C,$4C,$10,$C1
	.dc.b	$1E,$6E,$5A,$E4,$E6,$E0,$8E,$08,$71,$29,$C4,$C1,$0C,$11,$E6,$E5
	.dc.b	$AE,$4E,$6E,$08,$E0,$87,$12,$9C,$4C,$10,$C1,$1E,$6E,$5A,$E4,$E6
	.dc.b	$E0,$8E,$08,$71,$29,$C4,$C1,$0C,$11,$E6,$E5,$AE,$4E,$6E,$08,$E0
	.dc.b	$87,$12,$9C,$4C,$10,$C1,$1E,$6E,$5F,$A3,$F1,$BF,$E4,$42,$47,$F8
	.dc.b	$D4,$87,$F9,$29,$CB,$4C,$9F,$DF,$4A,$E9,$C1,$B2,$18,$23,$B2,$3C
	.dc.b	$88,$70,$72,$D3,$27,$22,$BA,$70,$6C,$86,$08,$EC,$8F,$22,$1C,$1C
	.dc.b	$B4,$C9,$C8,$AE,$9C,$1B,$21,$82,$3B,$23,$C8,$87,$07,$2D,$32,$72
	.dc.b	$2B,$A7,$06,$C8,$60,$8E,$C8,$F2,$21,$C1,$CB,$4C,$9C,$8A,$E9,$C1
	.dc.b	$B2,$18,$23,$B2,$3C,$88,$70,$72,$D3,$27,$22,$BA,$70,$6C,$86,$0F
	.dc.b	$C1,$7C,$DF,$FA,$90,$8F,$FA,$9D,$FE,$0B,$F0,$5F,$EA,$82,$47,$FA
	.dc.b	$38,$87,$FA,$B9,$CB,$4C,$9F,$D9,$C8,$E9,$A6,$9C,$11,$C1,$0E,$84
	.dc.b	$38,$95,$E0,$D3,$81,$A2,$3A,$69,$A7,$04,$70,$43,$A1,$0E,$25,$78
	.dc.b	$34,$E0,$68,$8E,$9A,$69,$C1,$1C,$10,$E8,$43,$89,$5E,$0D,$38,$1A
	.dc.b	$23,$A6,$9A,$70,$47,$04,$3A,$10,$E2,$57,$83,$4E,$06,$88,$E9,$A6
	.dc.b	$9C,$11,$C1,$0E,$84,$38,$95,$E0,$D3,$81,$A2,$3A,$69,$A7,$04,$70
	.dc.b	$43,$A1,$0E,$25,$78,$34,$E0,$68,$8E,$9A,$69,$C1,$1C,$10,$E8,$43
	.dc.b	$89,$5E,$0D,$38,$1A,$23,$A6,$9A,$70,$47,$04,$3A,$10,$E2,$57,$83
	.dc.b	$4E,$07,$E8,$63,$F8,$6A,$0E,$1C,$10,$84,$3F,$C3,$F1,$D9,$F0,$91
	.dc.b	$C1,$1F,$0B,$5F,$21,$0E,$06,$88,$65,$E0,$E8,$C1,$A3,$04,$76,$47
	.dc.b	$4D,$78,$90,$E0,$68,$86,$5E,$0E,$8C,$1A,$30,$47,$64,$74,$D7,$89
	.dc.b	$0E,$06,$88,$65,$E0,$E8,$C1,$A3,$04,$76,$47,$4D,$78,$90,$E0,$68
	.dc.b	$87,$96,$3F,$08,$7E,$15,$FC,$11,$08,$43,$FC,$C1,$1D,$DE,$A7,$77
	.dc.b	$04,$7B,$9A,$23,$97,$81,$A7,$0E,$8C,$10,$D9,$0D,$15,$E2,$47,$81
	.dc.b	$A2,$39,$78,$1A,$70,$E8,$C1,$0D,$90,$D1,$5E,$24,$78,$1A,$23,$97
	.dc.b	$81,$A7,$0E,$8C,$1F,$BE,$8F,$AE,$FF,$82,$4F,$A8,$42,$10,$FD,$ED
	.dc.b	$7F,$A8,$A7,$A9,$23,$A9,$E2,$21,$E1,$2B,$B6,$9C,$0D,$11,$D3,$4D
	.dc.b	$38,$23,$82,$1D,$08,$71,$2B,$C1,$A7,$03,$44,$74,$D3,$4E,$08,$E0
	.dc.b	$87,$42,$1C,$4A,$FF,$6B,$9F,$A3,$1F,$CB,$5F,$9C,$C2,$10,$FE,$59
	.dc.b	$3F,$45,$EE,$75,$24,$F9,$1E,$E6,$9C,$3A,$30,$43,$64,$34,$57,$89
	.dc.b	$1E,$06,$88,$E5,$E0,$69,$C3,$A3,$04,$36,$43,$45,$78,$91,$E0,$68
	.dc.b	$FE,$9B,$3F,$31,$FE,$AE,$7C,$92,$10,$87,$E7,$0F,$EB,$73,$FD,$6C
	.dc.b	$8F,$69,$F0,$3E,$33,$04,$30,$43,$C8,$F0,$69,$B2,$9B,$74,$47,$4D
	.dc.b	$74,$60,$86,$08,$74,$23,$C5,$A7,$02,$9C,$1D,$11,$FE,$08,$7E,$54
	.dc.b	$A7,$CC,$3E,$49,$08,$43,$FD,$26,$77,$7A,$3D,$CC,$10,$C1,$0F,$29
	.dc.b	$1E,$2D,$38,$14,$E0,$E8,$8E,$9A,$E8,$C1,$0C,$10,$E8,$47,$8B,$4E
	.dc.b	$07,$E4,$0F,$8B,$3F,$D8,$5F,$84,$42,$10,$4F,$EE,$DF,$04,$F6,$23
	.dc.b	$D9,$AE,$8C,$10,$C1,$0F,$31,$5E,$2E,$08,$E0,$8F,$33,$25,$72,$F3
	.dc.b	$30,$43,$04,$3F,$D1,$53,$FB,$33,$E0,$30,$84,$3F,$8E,$9F,$B2,$21
	.dc.b	$FE,$51,$9F,$0B,$E4,$3B,$4E,$08,$F8,$CD,$11,$CB,$C0,$D3,$87,$46
	.dc.b	$08,$6C,$86,$8A,$F1,$23,$C0,$D1,$1F,$EE,$73,$F0,$E3,$E0,$10,$84
	.dc.b	$3F,$8A,$1F,$04,$F8,$67,$E2,$0E,$D1,$E0,$7C,$66,$08,$60,$87,$8C
	.dc.b	$C9,$4C,$BC,$DC,$11,$C1,$1E,$2D,$78,$98,$21,$82,$1F,$ED,$D8,$F8
	.dc.b	$4F,$BC,$84,$21,$FC,$28,$E3,$1D,$DF,$21,$D8,$A6,$8C,$3B,$23,$82
	.dc.b	$1B,$21,$C8,$8F,$07,$2D,$72,$72,$29,$EE,$8F,$6B,$EA,$21,$08,$7F
	.dc.b	$A9,$8E,$51,$D4,$8E,$08,$78,$8A,$78,$08,$F0,$74,$43,$27,$07,$4E
	.dc.b	$0D,$38,$23,$B3,$F9,$7C,$7B,$1F,$49,$08,$43,$FC,$FA,$71,$3A,$94
	.dc.b	$F0,$12,$3B,$BA,$21,$93,$83,$A7,$06,$9C,$11,$D9,$0D,$14,$F9,$11
	.dc.b	$EB,$7D,$04,$21,$0F,$D8,$1F,$B3,$21,$83,$C6,$48,$F1,$3E,$02,$19
	.dc.b	$3C,$4E,$9C,$1A,$70,$47,$64,$34,$53,$89,$1E,$0F,$FB,$E5,$FF,$28
	.dc.b	$3F,$D5,$8F,$EF,$67,$B4,$84,$21,$F6,$8E,$E4,$3B,$1E,$17,$BB,$83
	.dc.b	$4E,$08,$EC,$86,$8A,$71,$23,$C1,$D1,$0F,$F6,$14,$7D,$97,$CE,$42
	.dc.b	$12,$7F,$B7,$3D,$0E,$F3,$D5,$C1,$D5,$C1,$1D,$90,$D1,$4E,$24,$78
	.dc.b	$3A,$21,$93,$FC,$13,$3E,$97,$CE,$42,$10,$FE,$0C,$72,$8E,$AF,$33
	.dc.b	$07,$8D,$C9,$1C,$10,$D9,$0E,$44,$78,$39,$6B,$90,$7E,$34,$FC,$61
	.dc.b	$FE,$47,$21,$E9,$21,$08,$72,$3F,$AF,$9F,$AB,$9F,$F3,$8B,$E1,$3D
	.dc.b	$07,$84,$C1,$0C,$12,$3C,$E4,$78,$B5,$E0,$D7,$81,$A2,$1F,$EA,$39
	.dc.b	$FA,$CF,$A0,$84,$21,$FA,$73,$99,$D8,$E6,$60,$91,$E1,$23,$A6,$BC
	.dc.b	$48,$70,$34,$43,$2F,$E1,$23,$CF,$1E,$42,$10,$87,$CF,$3F,$C1,$B3
	.dc.b	$E0,$7B,$1D,$0E,$C6,$09,$3B,$23,$A6,$BC,$48,$70,$34,$43,$FD,$81
	.dc.b	$3E,$87,$CA,$42,$12,$7F,$77,$3D,$A3,$A1,$D9,$C1,$1E,$C6,$88,$E5
	.dc.b	$E0,$69,$C3,$A3,$07,$DA,$9F,$33,$E4,$21,$08,$7E,$E0,$E4,$76,$21
	.dc.b	$82,$47,$80,$8F,$80,$DB,$86,$99,$38,$95,$FF,$3F,$C7,$9A,$3C,$64
	.dc.b	$21,$0F,$CC,$47,$57,$99,$D8,$86,$08,$EC,$8F,$84,$87,$07,$2D,$32
	.dc.b	$1F,$E0,$4D,$3F,$82,$38,$23,$82,$1F,$C3,$A3,$FE,$00,$3E,$B2,$10
	.dc.b	$87,$63,$FD,$1E,$7F,$7A,$9F,$E1,$EF,$B0,$F2,$CF,$A1,$C1,$E9,$21
	.dc.b	$82,$3B,$23,$C8,$87,$07,$F6,$CF,$F2,$83,$FD,$D8,$79,$88,$42,$1E
	.dc.b	$93,$D2,$7E,$E4,$FF,$09,$BE,$07,$B1,$E0,$3C,$A5,$74,$E0,$D9,$0C
	.dc.b	$11,$D9,$1F,$A2,$7D,$73,$F9,$31,$EB,$9F,$3B,$08,$43,$D4,$43,$D4
	.dc.b	$48,$EC,$F8,$DA,$F8,$CF,$51,$0D,$14,$D1,$82,$18,$23,$F8,$C7,$FA
	.dc.b	$D1,$0F,$F6,$C9,$EB,$3D,$04,$21,$0F,$51,$FD,$6A,$7F,$16,$7A,$22
	.dc.b	$47,$95,$AF,$91,$AF,$90,$D1,$0D,$14,$D1,$FD,$E6,$7F,$7C,$F8,$C8
	.dc.b	$42,$1F,$3A,$7F,$3F,$3E,$02,$47,$56,$BE,$02,$1E,$02,$3A,$69,$C4
	.dc.b	$87,$CB,$7E,$73,$F8,$72,$3F,$EE,$47,$E7,$1E,$62,$10,$87,$CE,$3F
	.dc.b	$96,$3E,$12,$3E,$03,$C6,$7A,$1C,$B4,$D3,$87,$64,$70,$7F,$9A,$1F
	.dc.b	$E1,$C7,$F9,$A0,$F1,$90,$84,$3F,$07,$1F,$D5,$9E,$E4,$3C,$44,$3C
	.dc.b	$06,$4A,$66,$3C,$6E,$5A,$F1,$9F,$C0,$BE,$12,$10,$93,$E3,$7A,$1C
	.dc.b	$C8,$60,$87,$88,$93,$E0,$69,$C0,$A7,$F8,$CE,$7C,$6F,$84,$84,$21
	.dc.b	$F6,$8E,$51,$D4,$8E,$08,$79,$1C,$B4,$C9,$E2,$23,$F9,$02,$3F,$C6
	.dc.b	$1A,$FE,$54,$C1,$0C,$10,$FC,$C4,$7B,$1F,$49,$08,$43,$FB,$59,$E4
	.dc.b	$8F,$3B,$E4,$30,$43,$D4,$43,$45,$78,$91,$F3,$1F,$AA,$2B,$EC,$7F
	.dc.b	$A5,$C7,$F6,$47,$C8,$42,$10,$FD,$61,$F4,$CF,$01,$E5,$3C,$64,$8F
	.dc.b	$13,$A2,$19,$38,$3A,$70,$7F,$2D,$8F,$94,$F8,$48,$42,$1F,$C5,$CF
	.dc.b	$F4,$34,$F3,$99,$1D,$88,$78,$9C,$B4,$C9,$E2,$23,$FE,$81,$9F,$1C
	.dc.b	$78,$08,$42,$4F,$EA,$5E,$27,$49,$EE,$E0,$F0,$10,$C1,$1D,$91,$FB
	.dc.b	$27,$FB,$10,$FE,$46,$78,$48,$42,$1F,$60,$FE,$D5,$1E,$58,$E9,$1E
	.dc.b	$16,$BE,$13,$D0,$43,$45,$34,$60,$FE,$FA,$7F,$22,$3F,$D7,$A7,$88
	.dc.b	$84,$21,$E5,$3F,$C2,$F1,$FD,$91,$ED,$1E,$42,$38,$21,$E1,$3C,$64
	.dc.b	$76,$E5,$FD,$01,$1F,$F0,$AC,$FD,$C7,$C6,$42,$10,$FD,$01,$D4,$F0
	.dc.b	$94,$EC,$62,$3C,$2E,$48,$E0,$FE,$4D,$1E,$18,$EC,$42,$10,$FE,$0E
	.dc.b	$76,$21,$D0,$F0,$10,$CC,$77,$70,$6C,$FF,$0F,$C7,$85,$EC,$42,$10
	.dc.b	$FD,$71,$FE,$4A,$39,$C7,$57,$B9,$82,$18,$21,$E1,$23,$F4,$DF,$E5
	.dc.b	$4D,$7F,$8C,$18,$3F,$42,$7E,$38,$FE,$32,$7F,$6F,$3C,$C4,$21,$0F
	.dc.b	$C7,$1F,$69,$F4,$3E,$23,$CE,$53,$CE,$E8,$8E,$9A,$FF,$B5,$0F,$D9
	.dc.b	$9D,$C8,$42,$1E,$23,$F3,$C7,$FA,$34,$E8,$43,$C2,$53,$C8,$47,$C2
	.dc.b	$D3,$4F,$C9,$23,$F2,$0F,$96,$7F,$56,$3C,$E7,$90,$84,$21,$E7,$3F
	.dc.b	$22,$77,$3C,$A4,$30,$47,$CA,$47,$CA,$43,$EE,$1F,$DA,$63,$F2,$4F
	.dc.b	$80,$84,$24,$FF,$19,$79,$1D,$CB,$3D,$DF,$23,$5D,$18,$3F,$8D,$4F
	.dc.b	$81,$EC,$42,$10,$FC,$41,$D8,$87,$32,$1D,$89,$1E,$12,$3F,$E9,$18
	.dc.b	$F0,$3D,$48,$42,$1F,$42,$38,$BD,$9A,$F6,$3C,$65,$34,$61,$FF,$65
	.dc.b	$BF,$CD,$4F,$F6,$59,$D8,$84,$21,$F8,$13,$F9,$A9,$FD,$B8,$87,$62
	.dc.b	$1E,$12,$3E,$62,$3E,$32,$1F,$82,$3F,$2A,$FE,$61,$C3,$5C,$9F,$9B
	.dc.b	$70,$7F,$A2,$23,$E5,$C7,$A0,$84,$21,$FC,$68,$4F,$91,$F4,$CD,$7D
	.dc.b	$27,$98,$A6,$8C,$3E,$27,$C8,$43,$D2,$53,$F6,$33,$E6,$7C,$64,$21
	.dc.b	$27,$F3,$CF,$72,$1E,$22,$4F,$9C,$8F,$90,$87,$88,$87,$CA,$3F,$C5
	.dc.b	$C7,$F0,$23,$C2,$42,$10,$FC,$11,$FE,$F4,$8F,$2B,$E2,$3B,$94,$F3
	.dc.b	$C7,$91,$A6,$9C,$3E,$33,$F8,$A1,$F9,$59,$F0,$30,$84,$3C,$67,$F8
	.dc.b	$76,$7C,$73,$D0,$91,$DC,$87,$88,$87,$9C,$FE,$33,$3F,$55,$EC,$42
	.dc.b	$10,$FE,$06,$72,$8E,$ED,$3B,$9D,$5C,$B4,$FF,$34,$47,$77,$A1,$08
	.dc.b	$43,$F8,$11,$D2,$39,$91,$C1,$0F,$11,$0F,$9C,$FE,$5D,$FC,$1B,$5C
	.dc.b	$9F,$97,$70,$47,$FC,$4E,$7E,$1C,$F2,$10,$84,$3E,$A9,$FD,$78,$FD
	.dc.b	$79,$FC,$20,$F8,$47,$95,$F5,$34,$F5,$1A,$23,$E3,$7D,$27,$9C,$FF
	.dc.b	$32,$1F,$89,$3C,$44,$21,$0F,$29,$FE,$F2,$3C,$92,$7C,$D1,$E4,$7C
	.dc.b	$27,$91,$C1,$1C,$10,$F2,$9F,$E0,$53,$C8,$78,$48,$42,$1E,$43,$F5
	.dc.b	$C7,$90,$F1,$9D,$4F,$19,$F1,$C8,$60,$8F,$95,$F1,$9F,$78,$AF,$CB
	.dc.b	$70,$7E,$C8,$FF,$16,$9F,$E7,$A3,$CA,$42,$10,$F4,$1F,$A4,$3C,$27
	.dc.b	$9D,$F1,$38,$3C,$EE,$08,$FD,$B3,$CA,$7F,$A7,$A7,$B3,$08,$43,$C2
	.dc.b	$7C,$83,$FC,$58,$7F,$95,$CE,$E7,$8C,$86,$08,$79,$49,$1F,$14,$F8
	.dc.b	$D1,$FB,$78,$F0,$10,$84,$3F,$A7,$4F,$B1,$EB,$35,$EA,$78,$8A,$68
	.dc.b	$FE,$91,$3D,$DE,$84,$21,$27,$F9,$0B,$CE,$7B,$11,$C1,$0E,$C7,$F9
	.dc.b	$D2,$7B,$3D,$08,$42,$1F,$76,$3A,$3C,$8F,$49,$1C,$B4,$FE,$3E,$7F
	.dc.b	$09,$39,$90,$84,$3E,$D9,$FC,$80,$EC,$F4,$23,$EB,$3B,$1A,$3F,$14
	.dc.b	$7E,$71,$FB,$4E,$1A,$FE,$38,$87,$F5,$38,$F8,$6F,$8C,$84,$21,$E5
	.dc.b	$3C,$24,$8F,$2B,$E3,$21,$93,$C4,$43,$C2,$F9,$1F,$41,$93,$C8,$7C
	.dc.b	$93,$C2,$42,$10,$F9,$33,$F8,$47,$C3,$1E,$47,$C4,$60,$F9,$4E,$4F
	.dc.b	$11,$1F,$13,$5F,$E3,$67,$F9,$28,$F0,$10,$84,$3E,$D9,$FD,$0E,$3F
	.dc.b	$97,$3F,$B5,$3C,$64,$8F,$13,$E4,$21,$E7,$3E,$39,$E4,$21,$F6,$A7
	.dc.b	$EC,$BE,$12,$10,$87,$ED,$8E,$A7,$84,$B1,$E2,$7C,$06,$4F,$09,$0F
	.dc.b	$AE,$7F,$96,$E3,$C2,$F6,$21,$09,$3F,$55,$E8,$49,$F0,$34,$F1,$90
	.dc.b	$F0,$3F,$29,$FE,$6E,$43,$B1,$08,$43,$E9,$9C,$4F,$A8,$7F,$4E,$3D
	.dc.b	$EF,$88,$8F,$84,$F2,$1A,$3E,$89,$F9,$63,$F0,$A7,$72,$10,$87,$88
	.dc.b	$FE,$70,$7F,$9D,$5E,$EF,$52,$1E,$02,$3A,$7F,$AC,$CF,$D0,$7A,$90
	.dc.b	$84,$3F,$8D,$1E,$B8,$E4,$47,$04,$3F,$C8,$F1,$D5,$E4,$42,$10,$FE
	.dc.b	$09,$1C,$DF,$4B,$5F,$49,$D4,$FF,$7E,$C7,$57,$91,$08,$43,$F3,$07
	.dc.b	$43,$94,$75,$23,$83,$FD,$BE,$7F,$09,$39,$10,$84,$3E,$A4,$FE,$56
	.dc.b	$39,$C7,$72,$1E,$A2,$BF,$C1,$5F,$D1,$91,$C1,$1F,$C3,$BF,$EB,$D9
	.dc.b	$F8,$11,$E1,$21,$09,$3E,$37,$D8,$78,$0A,$78,$C9,$3D,$DF,$D3,$9E
	.dc.b	$02,$1F,$C1,$CF,$88,$77,$21,$08,$7E,$5C,$FC,$49,$E1,$21,$E0,$3C
	.dc.b	$6F,$C8,$7E,$11,$E0,$29,$E4,$21,$FB,$A9,$FD,$E1,$E0,$61,$08,$78
	.dc.b	$CF,$DD,$10,$FC,$89,$E2,$21,$E6,$8F,$21,$1F,$23,$F1,$DF,$29,$FE
	.dc.b	$D8,$3E,$29,$E0,$21,$08,$78,$E7,$EB,$3D,$CF,$09,$F1,$4C,$10,$EF
	.dc.b	$1E,$12,$3F,$E4,$13,$A9,$08,$43,$E1,$1F,$04,$FE,$36,$7F,$4D,$7D
	.dc.b	$CF,$76,$BE,$32,$3E,$07,$E9,$CF,$E5,$1E,$C4,$21,$0F,$C1,$9C,$8F
	.dc.b	$7C,$76,$23,$83,$F9,$3C,$75,$79,$10,$84,$3E,$D9,$E9,$39,$14,$87
	.dc.b	$84,$FC,$5C,$7F,$45,$23,$F2,$CE,$84,$21,$0F,$96,$7F,$A9,$4F,$F3
	.dc.b	$59,$FE,$BA,$3B,$BE,$C6,$9E,$43,$C4,$7E,$2A,$3D,$CF,$52,$10,$87
	.dc.b	$F1,$C3,$8C,$75,$69,$D5,$FE,$81,$1D,$23,$89,$08,$49,$FC,$9B,$C8
	.dc.b	$F3,$90,$C1,$FE,$81,$8E,$8F,$22,$10,$87,$E0,$27,$CC,$47,$CC,$43
	.dc.b	$FC,$DC,$7E,$A8,$E2,$42,$10,$FA,$27,$C3,$9F,$4C,$75,$8F,$43,$5F
	.dc.b	$4B,$F9,$97,$F5,$67,$EF,$08,$7E,$10,$FC,$11,$DC,$84,$21,$FE,$06
	.dc.b	$AF,$F0,$71,$1F,$F1,$CB,$5F,$F7,$43,$53,$E5,$89,$1F,$0C,$F3,$9E
	.dc.b	$43,$CE,$79,$48,$60,$FC,$51,$F2,$4F,$29,$08,$43,$F3,$31,$E3,$3C
	.dc.b	$91,$E2,$8F,$92,$78,$DC,$3D,$DF,$94,$47,$CC,$FE,$A9,$1F,$C9,$4F
	.dc.b	$F0,$F9,$E1,$21,$08,$7D,$C3,$F8,$B4,$79,$5F,$96,$78,$C8,$79,$88
	.dc.b	$79,$4F,$19,$0F,$8F,$1F,$ED,$67,$E3,$1E,$02,$10,$87,$F2,$88,$F1
	.dc.b	$3F,$68,$F0,$9F,$58,$F0,$C7,$C6,$3E,$23,$E0,$3C,$47,$F8,$C8,$F0
	.dc.b	$9D,$88,$42,$1F,$D6,$63,$C2,$FE,$20,$F0,$C7,$76,$90,$F0,$10,$FC
	.dc.b	$35,$7E,$E1,$EA,$42,$10,$F7,$10,$F0,$16,$3C,$0F,$C0,$3C,$07,$C3
	.dc.b	$3E,$91,$E1,$7E,$83,$F8,$23,$B1,$08,$43,$F0,$47,$F7,$89,$F8,$27
	.dc.b	$81,$ED,$14,$8F,$CE,$7F,$C5,$0F,$CE,$3A,$10,$84,$3F,$C3,$11,$D9
	.dc.b	$F5,$C7,$47,$DC,$43,$DA,$47,$FC,$26,$F7,$3A,$10,$84,$3F,$CF,$91
	.dc.b	$DD,$E9,$1D,$DF,$69,$E2,$3F,$13,$3F,$65,$E4,$42,$10,$FA,$E4,$9F
	.dc.b	$DE,$BC,$CE,$E5,$3B,$9F,$E3,$F3,$EB,$1C,$C8,$42,$1F,$58,$F5,$1E
	.dc.b	$09,$F5,$11,$EC,$FF,$9F,$67,$D5,$1C,$88,$42,$4F,$49,$E5,$1E,$82
	.dc.b	$1F,$D5,$27,$9B,$C4,$84,$21,$FC,$28,$FC,$E4,$FC,$47,$9C,$F6,$7E
	.dc.b	$19,$FE,$F0,$9E,$AF,$22,$10,$87,$43,$B1,$F0,$CF,$82,$43,$B9,$F9
	.dc.b	$72,$1F,$A7,$29,$F8,$D3,$FC,$C8,$F8,$9E,$E4,$21,$0F,$C0,$1E,$29
	.dc.b	$F6,$4C,$8F,$11,$D4,$F0,$10,$C1,$1F,$E7,$89,$FC,$FB,$F8,$63,$C0
	.dc.b	$42,$10,$FC,$31,$FB,$B3,$C8,$7C,$67,$C8,$FC,$42,$1E,$13,$E1,$BF
	.dc.b	$CB,$1C,$1F,$C7,$67,$E2,$3D,$C8,$42,$4F,$E2,$DF,$83,$3E,$07,$EF
	.dc.b	$BD,$48,$F8,$0F,$EE,$53,$D9,$E8,$42,$10,$F6,$9F,$83,$3D,$71,$DC
	.dc.b	$8F,$53,$E8,$3E,$07,$FB,$4B,$F4,$0F,$9E,$FF,$A8,$4E,$C4,$21,$0F
	.dc.b	$9E,$7F,$9F,$CF,$ED,$67,$90,$FC,$A1,$E4,$3E,$19,$0F,$19,$E5,$3E
	.dc.b	$79,$FD,$64,$FC,$01,$DC,$84,$21,$E2,$3F,$D6,$13,$F4,$5F,$01,$0F
	.dc.b	$A2,$77,$8F,$84,$F8,$1F,$84,$FF,$8D,$0E,$A4,$21,$0F,$01,$F0,$4F
	.dc.b	$A4,$7E,$90,$F0,$3F,$49,$A7,$84,$87,$D2,$3F,$C3,$67,$F8,$6C,$FA
	.dc.b	$47,$72,$10,$87,$88,$F1,$1E,$28,$FE,$20,$FC,$03,$C2,$43,$E0,$1E
	.dc.b	$23,$E8,$1E,$03,$A9,$08,$43,$C0,$7C,$17,$E8,$3D,$C9,$1F,$05,$EE
	.dc.b	$7E,$AC,$FE,$B4,$73,$21,$08,$7F,$76,$8E,$B1,$CC,$91,$F8,$B3,$F8
	.dc.b	$5C,$FE,$A5,$E2,$42,$10,$F8,$C7,$F0,$53,$94,$74,$7E,$33,$F8,$47
	.dc.b	$FA,$C1,$F1,$4E,$44,$21,$0F,$8A,$7E,$90,$EE,$74,$3B,$90,$FC,$11
	.dc.b	$DA,$3F,$66,$73,$61,$08,$76,$3F,$03,$3F,$19,$E8,$7D,$22,$CF,$F2
	.dc.b	$E9,$F8,$D1,$C8,$84,$21,$F9,$B2,$47,$C5,$7A,$1F,$AD,$8E,$6F,$02
	.dc.b	$10,$93,$E5,$7E,$11,$0F,$ED,$C7,$E1,$0E,$04,$21,$0F,$E2,$2F,$37
	.dc.b	$C6,$49,$FF,$01,$3C,$CE,$04,$21,$0F,$E9,$31,$CE,$3D,$C4,$8F,$D3
	.dc.b	$3F,$C2,$08,$E0,$FD,$CB,$F6,$1F,$DD,$CF,$F2,$C7,$B1,$08,$43,$ED
	.dc.b	$1F,$AB,$3F,$95,$4F,$85,$F8,$D3,$E0,$7F,$94,$90,$FD,$69,$FA,$23
	.dc.b	$E0,$1D,$48,$42,$1F,$E2,$D8,$F8,$0F,$E1,$88,$75,$8F,$03,$D9,$B3
	.dc.b	$F8,$08,$E6,$42,$10,$F5,$9F,$AF,$3D,$67,$63,$A1,$F2,$CF,$01,$FE
	.dc.b	$16,$7B,$3C,$C8,$42,$1F,$AE,$3F,$A8,$9F,$83,$3B,$47,$A8,$8F,$73
	.dc.b	$F9,$0C,$7D,$97,$99,$08,$43,$EB,$9F,$B0,$3D,$A7,$B0,$87,$B0,$F5
	.dc.b	$90,$FF,$75,$9F,$D7,$8E,$84,$21,$0E,$E7,$B6,$7F,$08,$FB,$4E,$93
	.dc.b	$EC,$7F,$97,$1D,$4E,$44,$21,$0E,$B1,$F8,$27,$D3,$1D,$8E,$8F,$AC
	.dc.b	$FE,$AD,$3F,$1D,$E4,$42,$10,$F4,$9F,$C8,$CF,$E1,$93,$D5,$FA,$A7
	.dc.b	$D3,$9F,$F1,$94,$FE,$02,$39,$10,$84,$39,$9F,$8D,$3E,$21,$23,$D2
	.dc.b	$7E,$61,$FD,$29,$C4,$84,$21,$FD,$44,$FB,$A7,$33,$CE,$7D,$03,$FC
	.dc.b	$D2,$74,$38,$90,$84,$3F,$BB,$90,$FC,$51,$F9,$97,$E2,$3F,$48,$FC
	.dc.b	$B4,$7C,$37,$91,$08,$49,$FB,$4F,$9C,$E4,$7F,$07,$8E,$4F,$02,$10
	.dc.b	$87,$C0,$8F,$1B,$CC,$FF,$6A,$C7,$27,$64,$21,$0F,$CC,$1F,$20,$E5
	.dc.b	$3F,$DB,$5F,$F4,$01,$FC,$A8,$80,$FF,$0A,$47,$36,$10,$87,$E3,$8F
	.dc.b	$98,$7E,$10,$E5,$3E,$A9,$FF,$59,$4F,$A6,$78,$90,$84,$39,$1F,$C3
	.dc.b	$23,$93,$F9,$72,$1F,$83,$21,$FA,$77,$F4,$D1,$FD,$0E,$3A,$10,$84
	.dc.b	$3D,$47,$E9,$E7,$FC,$8A,$FF,$72,$3E,$54,$FE,$71,$F0,$9D,$CA,$FE
	.dc.b	$A9,$FC,$A1,$D4,$84,$21,$FD,$88,$FC,$FB,$EE,$7D,$6F,$76,$9D,$8F
	.dc.b	$AE,$7F,$04,$3F,$B0,$90,$EA,$42,$10,$F7,$9F,$D8,$4F,$D1,$1E,$E3
	.dc.b	$B1,$E0,$21,$EF,$3F,$C5,$27,$84,$FC,$44,$FC,$F7,$B1,$08,$43,$C2
	.dc.b	$7C,$E3,$EF,$CF,$C1,$7B,$12,$7E,$13,$4F,$83,$3F,$8B,$7A,$10,$84
	.dc.b	$9F,$DA,$CF,$53,$99,$23,$DE,$7E,$80,$E2,$42,$10,$F4,$9F,$C2,$E3
	.dc.b	$AB,$F2,$4F,$8F,$1D,$5F,$93,$3E,$77,$91,$08,$43,$F8,$11,$CC,$F8
	.dc.b	$07,$43,$CF,$1F,$9E,$7F,$00,$71,$21,$08,$7A,$0F,$C0,$1F,$AE,$7E
	.dc.b	$23,$D0,$87,$DE,$7F,$84,$1D,$DF,$BE,$74,$21,$08,$77,$3F,$97,$9D
	.dc.b	$CF,$63,$F1,$5F,$69,$FC,$50,$FF,$3A,$1E,$C8,$E6,$C2,$10,$EC,$76
	.dc.b	$9F,$DC,$9F,$EB,$49,$F6,$C7,$C9,$3F,$83,$CF,$D3,$79,$10,$84,$3F
	.dc.b	$95,$9F,$CF,$0F,$3C,$7A,$4F,$F3,$F3,$FC,$00,$E0,$42,$10,$FF,$6D
	.dc.b	$CF,$E2,$1E,$91,$CD,$FD,$EB,$F4,$1A,$7F,$B9,$CE,$44,$21,$0F,$86
	.dc.b	$7E,$B6,$3B,$3F,$D0,$8F,$BA,$7F,$8C,$8F,$9C,$71,$21,$08,$74,$3F
	.dc.b	$93,$1F,$EE,$A7,$E7,$3F,$70,$FD,$5C,$FC,$D7,$89,$08,$43,$F8,$09
	.dc.b	$C4,$FA,$A7,$CB,$34,$42,$10,$FE,$A2,$43,$F3,$B1,$F8,$27,$FC,$2D
	.dc.b	$3F,$82,$7F,$C8,$67,$FB,$34,$E4,$42,$10,$F8,$07,$ED,$0F,$AE,$7F
	.dc.b	$84,$8F,$F2,$01,$B2,$10,$87,$EC,$63,$E9,$C7,$FB,$DE,$7E,$34,$E8
	.dc.b	$84,$24,$F0,$7E,$CC,$FE,$1D,$FC,$E1,$F9,$43,$F0,$E7,$EE,$4A,$7E
	.dc.b	$B4,$FE,$70,$27,$F9,$C3,$DC,$84,$21,$F0,$8F,$98,$7F,$48,$3F,$DA
	.dc.b	$87,$D7,$36,$42,$10,$FA,$C4,$3F,$A1,$1D,$0F,$E0,$C7,$B2,$3D,$CF
	.dc.b	$B4,$C1,$EC,$7F,$C3,$91,$F2,$A3,$B1,$08,$43,$ED,$1C,$C9,$1D,$88
	.dc.b	$FF,$4E,$3D,$87,$12,$10,$87,$AC,$87,$AE,$63,$F6,$08,$7A,$8F,$AC
	.dc.b	$43,$EB,$47,$F0,$17,$13,$ED,$7B,$90,$84,$3F,$8B,$9E,$C3,$B9,$4F
	.dc.b	$F4,$B9,$FA,$13,$81,$08,$43,$F0,$07,$A8,$FF,$1E,$9F,$AC,$3A,$C7
	.dc.b	$C8,$7D,$87,$F6,$B3,$91,$08,$43,$A9,$F8,$23,$F8,$19,$C9,$E8,$FA
	.dc.b	$4E,$87,$12,$10,$87,$43,$F9,$11,$FE,$2E,$3C,$E4,$3E,$83,$F3,$DF
	.dc.b	$A2,$7E,$04,$FD,$E9,$D0,$84,$21,$FC,$C2,$7F,$6C,$FA,$4E,$C7,$EC
	.dc.b	$27,$F5,$EF,$02,$10,$87,$C4,$8E,$6F,$C1,$9F,$87,$3F,$83,$8D,$90
	.dc.b	$84,$3E,$59,$C4,$F1,$BC,$5D,$10,$84,$9F,$0C,$FD,$23,$E3,$1F,$1E
	.dc.b	$79,$4F,$02,$10,$87,$EC,$23,$93,$FC,$A4,$FC,$91,$A2,$10,$87,$DB
	.dc.b	$3F,$24,$7F,$9D,$4F,$9C,$7D,$A3,$F8,$FC,$79,$1E,$24,$21,$0F,$D3
	.dc.b	$C7,$91,$FF,$0E,$9F,$80,$34,$42,$10,$FE,$49,$3F,$19,$FE,$28,$43
	.dc.b	$C8,$7F,$9A,$27,$C8,$F1,$21,$08,$7D,$89,$FF,$46,$3E,$33,$24,$21
	.dc.b	$0F,$EA,$F3,$E3,$7F,$84,$9F,$C6,$8D,$10,$84,$3F,$8C,$90,$FB,$07
	.dc.b	$AC,$FD,$89,$B2,$10,$87,$D7,$3F,$9E,$BF,$60,$FF,$25,$3F,$BA,$9A
	.dc.b	$47,$FC,$26,$7D,$23,$46,$0A,$65,$E0,$D3,$4E,$1D,$91,$C1,$1D,$90
	.dc.b	$E4,$43,$81,$92,$99,$79,$34,$D3,$87,$64,$70,$47,$64,$39,$10,$E0
	.dc.b	$64,$A6,$5E,$4D,$34,$E1,$D9,$1C,$11,$D9,$0E,$44,$38,$19,$29,$97
	.dc.b	$93,$4D,$38,$76,$47,$04,$76,$43,$91,$0E,$06,$4A,$65,$E4,$D3,$4E
	.dc.b	$1D,$91,$C1,$1D,$90,$E4,$43,$81,$92,$99,$79,$34,$D3,$87,$64,$70
	.dc.b	$47,$64,$39,$10,$E0,$64,$A6,$5E,$4D,$34,$E1,$D9,$FE,$3F,$3F,$5A
	.dc.b	$00

;--------------------------------------------------------------------------------
;	FCOS.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fcos_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FCOS.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FCOS.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fcos_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1953+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fcos.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FCOS.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1953+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fcos.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FCOS.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fcos_data_compressed::
	.dc.b	$00,$02,$FA,$EC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$F3,$FE,$4D,$CD,$F8,$3E,$27,$13,$D5,$F5,$7A,$BB
	.dc.b	$AF,$C3,$52,$F8,$FA,$59,$F4,$B5,$E9,$6F,$D2,$E1,$E9,$71,$F4,$B9
	.dc.b	$7A,$5C,$FD,$2E,$9E,$97,$5F,$4B,$B7,$A5,$DF,$D2,$F0,$7A,$5E,$1F
	.dc.b	$4B,$C5,$E9,$7A,$7F,$58,$FF,$77,$F1,$1F,$F5,$4B,$7E,$37,$5F,$15
	.dc.b	$BF,$80,$E1,$EE,$71,$F5,$F9,$7A,$FC,$FD,$6E,$9E,$B7,$5F,$5B,$B7
	.dc.b	$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$78,$FC,$2F,$27,$AD,$E5
	.dc.b	$F5,$BC,$DE,$B7,$E0,$FD,$AF,$CA,$7C,$7F,$E5,$7C,$7F,$F6,$B7,$C4
	.dc.b	$FA,$FD,$9C,$CE,$BE,$27,$7F,$0B,$C3,$E0,$B8,$FB,$DC,$BC,$DE,$7E
	.dc.b	$6F,$4F,$37,$AF,$9B,$DB,$CD,$EF,$E6,$F8,$3C,$DF,$0F,$9B,$E2,$F3
	.dc.b	$7C,$7E,$6F,$93,$CD,$F2,$F9,$BE,$6F,$37,$CF,$E6,$FA,$3C,$DF,$D6
	.dc.b	$7C,$AF,$FB,$60,$FF,$48,$9F,$E4,$B3,$FC,$06,$7E,$D4,$FD,$A7,$49
	.dc.b	$79,$1C,$BC,$8E,$7E,$47,$4F,$23,$AF,$91,$DB,$C8,$EF,$E4,$78,$3C
	.dc.b	$8F,$0F,$91,$E2,$F2,$3C,$7E,$47,$93,$C8,$F2,$F9,$1F,$9D,$FD,$67
	.dc.b	$FD,$AB,$C0,$F2,$7D,$3E,$3F,$F0,$5F,$97,$7C,$AE,$FE,$27,$87,$C7
	.dc.b	$71,$F8,$CE,$5F,$0F,$CF,$E1,$7A,$7C,$2F,$5F,$85,$ED,$F0,$BD,$FE
	.dc.b	$17,$C1,$F0,$BE,$1F,$85,$F1,$7C,$2F,$8F,$E1,$7C,$9F,$0B,$E5,$F8
	.dc.b	$5F,$37,$C2,$F9,$FE,$17,$D1,$F0,$9E,$9F,$84,$F5,$7B,$DF,$B7,$FD
	.dc.b	$6B,$F8,$E7,$C2,$17,$C0,$67,$E0,$35,$F0,$1B,$F8,$0E,$1F,$01,$C7
	.dc.b	$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7,$C0,$77,$F8,$0F
	.dc.b	$07,$C0,$78,$7E,$03,$C5,$F0,$1F,$27,$C6,$F9,$5E,$37,$97,$E0,$3E
	.dc.b	$67,$8F,$F3,$7C,$7F,$9D,$E3,$FC,$FF,$1F,$E8,$78,$FF,$47,$C7,$FA
	.dc.b	$5E,$3F,$D3,$F1,$FE,$A7,$8F,$F5,$7C,$7F,$AD,$E3,$FD,$7F,$1F,$EC
	.dc.b	$78,$FF,$67,$C7,$FB,$5E,$3F,$DB,$F1,$FE,$E7,$8F,$F7,$7C,$7F,$E8
	.dc.b	$5F,$33,$FA,$F3,$FE,$00,$1F,$EB,$83,$FC,$56,$7F,$65,$3F,$91,$1F
	.dc.b	$C7,$BD,$C2,$F0,$79,$F8,$3D,$3C,$1E,$BE,$0F,$6F,$07,$BF,$83,$E0
	.dc.b	$F0,$7C,$3E,$0F,$8B,$C1,$F1,$F7,$F9,$3E,$07,$CB,$F0,$3E,$6F,$81
	.dc.b	$F3,$FC,$0F,$A3,$E0,$7F,$01,$5F,$D6,$FC,$57,$CA,$5F,$8B,$FC,$26
	.dc.b	$3F,$85,$C7,$F0,$D8,$FE,$1F,$1F,$E7,$D8,$FF,$26,$3F,$92,$8F,$87
	.dc.b	$CF,$C3,$EB,$E1,$F7,$F0,$FC,$3E,$1F,$8F,$C3,$F2,$F8,$7E,$7F,$0F
	.dc.b	$D3,$E1,$FA,$FC,$3F,$6F,$87,$EF,$F0,$FE,$0F,$87,$F0,$FC,$3F,$8B
	.dc.b	$E1,$FC,$7F,$0F,$E4,$F8,$7F,$2F,$C2,$F9,$BE,$17,$FA,$A7,$EB,$7F
	.dc.b	$B4,$F2,$7F,$A7,$2F,$73,$3D,$FA,$F7,$37,$EE,$70,$F7,$38,$FB,$9C
	.dc.b	$BD,$CE,$7E,$E7,$4F,$73,$AF,$B9,$F2,$FC,$0F,$99,$E0,$78,$3D,$CF
	.dc.b	$9D,$E0,$FC,$FF,$07,$E8,$78,$3F,$47,$C1,$FA,$5E,$0F,$D3,$F0,$7E
	.dc.b	$A7,$83,$F5,$7C,$1F,$AD,$E0,$FD,$7F,$07,$EC,$78,$3F,$67,$C1,$FB
	.dc.b	$5E,$0F,$DB,$F0,$7E,$E7,$83,$F7,$7C,$1F,$BD,$E0,$FF,$86,$7E,$6F
	.dc.b	$E1,$78,$47,$FD,$C2,$7F,$A6,$CF,$F2,$99,$FE,$0A,$75,$DB,$C7,$B7
	.dc.b	$97,$6F,$3E,$DE,$9D,$BD,$7B,$7B,$75,$F7,$F7,$7C,$1E,$EF,$87,$DD
	.dc.b	$F1,$7B,$BE,$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$E7,$F7,$7E,$CE
	.dc.b	$5F,$E3,$1F,$9C,$FE,$BC,$7F,$16,$E3,$FF,$1F,$2F,$83,$D7,$C1,$FE
	.dc.b	$03,$6F,$C0,$ED,$F8,$2D,$BF,$07,$B7,$E1,$36,$FC,$2E,$DF,$86,$DB
	.dc.b	$F0,$FB,$7F,$86,$FF,$5C,$FE,$4A,$77,$F8,$73,$DF,$7F,$F0,$C3,$FE
	.dc.b	$F0,$8C,$FC,$2F,$1F,$85,$E5,$F0,$BC,$FE,$13,$A7,$C2,$75,$F8,$4F
	.dc.b	$F2,$67,$EE,$6F,$84,$F0,$7C,$2F,$85,$F1,$7C,$2F,$8F,$E1,$7C,$9F
	.dc.b	$0B,$E5,$F8,$5F,$37,$C2,$F9,$FD,$FF,$47,$C1,$FA,$7E,$0F,$D5,$F0
	.dc.b	$7E,$BF,$83,$FF,$1B,$7E,$BD,$FE,$1E,$3F,$C3,$A7,$EE,$7A,$3F,$C6
	.dc.b	$4B,$DE,$DF,$BD,$F4,$7A,$3E,$97,$47,$D3,$E8,$E7,$EF,$7D,$5E,$9F
	.dc.b	$AD,$D3,$F5,$FA,$7E,$C7,$47,$D9,$E8,$FB,$5D,$1F,$6F,$A3,$EE,$74
	.dc.b	$7D,$DE,$8F,$BD,$D1,$F7,$FA,$3F,$01,$D1,$F8,$1E,$8F,$C1,$74,$7E
	.dc.b	$0F,$A3,$FC,$E9,$F4,$1F,$C6,$BA,$7F,$D0,$0B,$E1,$B3,$F0,$3A,$EA
	.dc.b	$DF,$57,$0E,$7E,$3F,$03,$CB,$E0,$79,$FC,$0F,$4F,$81,$EB,$F0,$3D
	.dc.b	$BE,$07,$BF,$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93
	.dc.b	$E0,$7C,$BF,$03,$E6,$F8,$1F,$3F,$C0,$FF,$A2,$F9,$3F,$66,$7F,$13
	.dc.b	$E7,$FF,$4C,$7E,$37,$FE,$C4,$87,$C2,$EF,$E1,$78,$7C,$2F,$1F,$84
	.dc.b	$E5,$F0,$9C,$FE,$13,$A7,$C2,$75,$F8,$4E,$DF,$09,$DF,$E1,$3C,$1F
	.dc.b	$09,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$FA,$87,$93,$F8,$27,$23,$E0,$7D,$5F,$03,$FE,$A5,$FA,$2F,$F2
	.dc.b	$91,$FB,$5E,$83,$FA,$76,$FC,$1E,$FD,$FE,$1E,$FF,$1F,$7F,$97,$BF
	.dc.b	$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BF,$E1,$F7,$FC
	.dc.b	$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF,$F9,$FD,$FF,$47,$BF
	.dc.b	$FE,$BD,$FA,$3F,$C5,$73,$1F,$C5,$3F,$EC,$B7,$F4,$CE,$7D,$EE,$1E
	.dc.b	$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$7C,$9F,$3F,$B7,$BD,$DF,$DE
	.dc.b	$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$E5,$F7,$BC,$DE
	.dc.b	$F7,$9F,$DE,$F4,$7B,$DE,$9F,$7B,$FD,$AD,$F4,$9F,$CF,$BC,$F7,$B9
	.dc.b	$8F,$89,$CF,$C0,$6B,$E0,$37,$F0,$1C,$39,$78,$DF,$2F,$80,$E7,$F0
	.dc.b	$1D,$3E,$03,$AF,$C0,$76,$F8,$0E,$FF,$01,$E0,$F8,$0F,$0F,$C0,$78
	.dc.b	$BE,$03,$C7,$F0,$1E,$4F,$80,$F2,$FC,$07,$E2,$F5,$FC,$66,$BF,$8D
	.dc.b	$D7,$F1,$DA,$FE,$3F,$5F,$5F,$BF,$EC,$F7,$FF,$DE,$5E,$F7,$FA,$7C
	.dc.b	$FC,$71,$F8,$D0,$6B,$DE,$DF,$BD,$C3,$DE,$E3,$EF,$72,$F7,$B9,$FB
	.dc.b	$DF,$9D,$E6,$EB,$EF,$76,$F7,$BB,$FB,$DF,$A1,$E9,$FD,$17,$4F,$E8
	.dc.b	$FA,$7F,$49,$D3,$FA,$5E,$9F,$D3,$74,$FE,$9F,$93,$F5,$1C,$8F,$52
	.dc.b	$7E,$39,$FE,$B7,$3F,$4C,$7E,$94,$F4,$BA,$E5,$E0,$A3,$FD,$D8,$7F
	.dc.b	$92,$CF,$F2,$53,$D3,$8B,$70,$E2,$BC,$38,$E1,$F9,$3E,$FF,$CA,$F7
	.dc.b	$FE,$5F,$BF,$F3,$3D,$FF,$9B,$EF,$FC,$EF,$7F,$E7,$FB,$FF,$43,$DF
	.dc.b	$FA,$3E,$FF,$93,$E9,$BF,$4E,$7E,$9B,$84,$FE,$98,$FE,$94,$FC,$15
	.dc.b	$EF,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE
	.dc.b	$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$F0,$BC,$8F,$87
	.dc.b	$D5,$F1,$3A,$BE,$2F,$57,$C6,$EA,$F8,$FD,$5F,$23,$AB,$E4,$F5,$7C
	.dc.b	$AE,$AF,$97,$D5,$F4,$BD,$0F,$E0,$A7,$BC,$FE,$06,$7B,$BA,$CF,$B9
	.dc.b	$FC,$03,$FE,$D8,$3F,$DA,$EE,$39,$FA,$73,$F5,$E7,$F9,$9E,$F7,$CD
	.dc.b	$F7,$BE,$77,$BD,$E1,$EA,$FA,$1E,$F7,$E1,$3E,$9F,$FC,$76,$7F,$49
	.dc.b	$CC,$FF,$48,$7F,$20,$BD,$BE,$1E,$DF,$1F,$6F,$97,$B7,$CF,$DB,$E9
	.dc.b	$ED,$F5,$E2,$ED,$ED,$FC,$1F,$0F,$E1,$78,$7F,$0F,$C3,$F8,$9E,$1F
	.dc.b	$C5,$F0,$FC,$9E,$DF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$D1,$EB,$FA,$7D
	.dc.b	$7F,$57,$AF,$EB,$F5,$FD,$9E,$BF,$B7,$D7,$FD,$0F,$D4,$7E,$8B,$AC
	.dc.b	$FF,$4D,$3F,$A6,$1F,$A8,$73,$EC,$70,$F6,$F8,$FB,$7C,$BD,$BF,$B5
	.dc.b	$E9,$74,$F6,$FA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$D6,$F0,$FA,$DE,$2F
	.dc.b	$5B,$C7,$EB,$79,$3D,$6F,$2F,$AD,$E6,$F5,$BC,$FE,$B7,$A3,$D6,$FC
	.dc.b	$47,$A5,$EA,$F5,$BF,$77,$E7,$7F,$4E,$3F,$50,$7F,$4C,$3F,$14,$7E
	.dc.b	$27,$9F,$EC,$7B,$1C,$3D,$8E,$3E,$C7,$2F,$63,$9F,$B1,$D3,$D6,$EB
	.dc.b	$E6,$F6,$F3,$7B,$F9,$BE,$0F,$37,$C3,$E6,$F8,$BB,$BC,$7C,$3E,$4F
	.dc.b	$63,$CB,$EC,$79,$BD,$8F,$E2,$3F,$53,$FE,$88,$3F,$1C,$7E,$37,$8C
	.dc.b	$FE,$35,$FE,$3C,$FF,$A2,$47,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7F,$D7,$69,$E0,$F5,$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F
	.dc.b	$5F,$CB,$EB,$F9,$BD,$7F,$3F,$AF,$E8,$F5,$FD,$3E,$BF,$F2,$AF,$AA
	.dc.b	$FC,$51,$FE,$15,$3F,$C2,$87,$E2,$47,$AF,$BF,$5F,$87,$AF,$FC,$07
	.dc.b	$AB,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F,$07,$AF
	.dc.b	$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB,$F9,$FD
	.dc.b	$7F,$47,$AF,$FD,$0B,$EA,$FF,$C2,$A7,$F2,$F3,$FC,$26,$7F,$11,$3F
	.dc.b	$10,$3D,$7E,$1E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB
	.dc.b	$7A,$FD,$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$79,$DE
	.dc.b	$5F,$03,$CD,$EB,$FF,$58,$F8,$4F,$F6,$79,$FE,$5D,$F8,$C7,$C5,$FF
	.dc.b	$22,$F6,$BF,$D1,$DF,$07,$BF,$73,$87,$BD,$FE,$96,$E6,$FF,$4B,$F3
	.dc.b	$7F,$A6,$39,$BF,$D3,$3C,$DF,$E9,$AE,$6F,$F4,$DF,$37,$FA,$73,$9B
	.dc.b	$FD,$3B,$CD,$FE,$9E,$E6,$FF,$4F,$F3,$7F,$A8,$39,$BF,$D4,$3C,$DF
	.dc.b	$EA,$2E,$6F,$F5,$1F,$37,$F7,$5E,$57,$A0,$F9,$6C,$3E,$93,$F4,$1F
	.dc.b	$97,$3C,$CF,$EE,$AF,$E2,$3A,$17,$CB,$9F,$8F,$3F,$2F,$7B,$9F,$BD
	.dc.b	$D3,$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$7B,$C3,$EF,$78,$BD,$EF
	.dc.b	$1F,$BD,$E4,$F7,$BC,$BE,$F7,$9B,$93,$CF,$EF,$7A,$3D,$EF,$A5,$C1
	.dc.b	$F4,$F8,$3E,$A7,$07,$D5,$E0,$F6,$FB,$DF,$5F,$87,$EC,$70,$FD,$9E
	.dc.b	$1F,$B5,$C3,$F6,$F8,$7E,$E7,$0F,$DD,$E1,$FB,$DC,$3F,$7F,$87,$F0
	.dc.b	$1C,$3F,$81,$E1,$FC,$16,$BF,$E1,$AF,$AD,$FE,$74,$7F,$A1,$4F,$F4
	.dc.b	$21,$FC,$E4,$F9,$5C,$C5,$EE,$F1,$E9,$E5,$EE,$F3,$F7,$7A,$7B,$BD
	.dc.b	$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7,$7C
	.dc.b	$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE,$FF,$8C,$BE,$BB,$FD
	.dc.b	$0A,$7F,$87,$CF,$F4,$19,$FE,$1E,$F2,$1F,$AF,$FF,$57,$8F,$E7,$47
	.dc.b	$EC,$F9,$FF,$0B,$0F,$99,$FC,$82,$F9,$4B,$F1,$58,$F8,$3C,$F5,$6B
	.dc.b	$AB,$7D,$5C,$3A,$B8,$F5,$72,$EA,$E7,$D5,$D3,$AB,$AF,$57,$6E,$AE
	.dc.b	$FD,$5E,$0E,$AF,$0F,$57,$F9,$3B,$F1,$FF,$F5,$E7,$2F,$F9,$5D,$77
	.dc.b	$F9,$BB,$3C,$FD,$9E,$8E,$CF,$4F,$67,$EA,$7F,$60,$FF,$79,$76,$9F
	.dc.b	$F5,$4B,$7E,$4B,$5F,$07,$BF,$8B,$E1,$F0,$DC,$7E,$1B,$97,$C2,$73
	.dc.b	$F8,$4E,$9F,$09,$D7,$E1,$3B,$7C,$27,$7F,$84,$FA,$98,$7D,$5C,$3C
	.dc.b	$5F,$09,$F5,$F1,$FB,$18,$FD,$9C,$7E,$D6,$3F,$6F,$1F,$B9,$7F,$77
	.dc.b	$E1,$3E,$F7,$C2,$7D,$FF,$84,$FC,$07,$C2,$7E,$07,$E1,$3F,$05,$F0
	.dc.b	$9F,$83,$F8,$4F,$F3,$AF,$D9,$7F,$07,$EE,$FF,$3F,$3F,$E3,$33,$FB
	.dc.b	$31,$FD,$95,$2F,$82,$E3,$F0,$5C,$BE,$0B,$9F,$C1,$74,$F8,$2E,$BF
	.dc.b	$05,$DB,$E0,$BB,$FC,$17,$83,$E0,$BC,$3F,$05,$E2,$F8,$2F,$1F,$C1
	.dc.b	$79,$3E,$0B,$CB,$F0,$5E,$6F,$2F,$CF,$F0,$5E,$8F,$82,$FF,$46,$FB
	.dc.b	$9F,$E2,$D3,$FD,$F0,$7F,$BD,$CF,$F1,$61,$FB,$43,$F6,$7C,$BF,$C7
	.dc.b	$BC,$A7,$6F,$2F,$6F,$9F,$B7,$F8,$3E,$FF,$C2,$77,$FE,$17,$BF,$F0
	.dc.b	$DD,$FF,$87,$EF,$FC,$47,$7F,$F3,$1F,$D8,$7E,$57,$9A,$7F,$A4,$1F
	.dc.b	$E5,$B7,$1F,$09,$BF,$07,$87,$83,$C7,$C1,$E5,$E0,$F3,$F0,$7A,$78
	.dc.b	$3D,$7C,$1E,$DE,$0F,$7F,$07,$C1,$E0,$F8,$7C,$1F,$17,$83,$E3,$F0
	.dc.b	$7C,$9E,$0F,$97,$C1,$F3,$78,$3E,$7F,$07,$FA,$47,$4F,$FA,$DB,$94
	.dc.b	$FE,$F7,$C0,$FE,$86,$FE,$F5,$BE,$FF,$0F,$0B,$8F,$BF,$CB,$C1,$E7
	.dc.b	$E0,$F4,$F7,$FE,$A9,$F5,$BD,$FF,$AF,$EF,$F8,$1F,$B3,$EF,$FD,$AF
	.dc.b	$7F,$ED,$FB,$FF,$73,$DF,$FB,$BE,$FF,$DE,$F7,$FE,$FF,$BF,$F8,$0F
	.dc.b	$7B,$F0,$3E,$F7,$E0,$BD,$EF,$C1,$FB,$DF,$ED,$0F,$B4,$7C,$ED,$F9
	.dc.b	$7C,$7C,$0E,$7E,$07,$5F,$03,$BF,$81,$E1,$F0,$3C,$7E,$07,$97,$C0
	.dc.b	$F3,$F8,$1E,$9F,$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7B,$FE,$1F
	.dc.b	$7F,$C5,$EF,$F8,$FC,$8F,$27,$BF,$E5,$F7,$FC,$DE,$FF,$FB,$9B,$ED
	.dc.b	$7F,$65,$E6,$1F,$F5,$21,$FE,$57,$3F,$64,$7E,$C5,$FF,$7A,$3F,$E5
	.dc.b	$66,$FB,$9F,$73,$C7,$E9,$EE,$75,$F7,$3B,$7B,$9F,$80,$F2,$7F,$03
	.dc.b	$E4,$FE,$0B,$C9,$FC,$1F,$93,$F8,$4F,$27,$F0,$BE,$4F,$E1,$BC,$9F
	.dc.b	$C3,$F9,$3F,$88,$F2,$7F,$13,$E4,$FE,$2B,$C9,$FF,$09,$7E,$C5,$FC
	.dc.b	$A8,$FE,$52,$3C,$7C,$F8,$FA,$F1,$F7,$E3,$F0,$F1,$F8,$F8,$FC,$BC
	.dc.b	$7E,$7E,$3F,$4F,$1F,$AF,$8F,$DB,$C7,$EF,$E3,$F8,$3C,$7F,$0F,$8F
	.dc.b	$E2,$F1,$FC,$7E,$3F,$93,$C7,$F2,$F8,$FE,$6F,$1F,$FC,$51,$FB,$1F
	.dc.b	$EE,$0F,$F1,$F1,$FB,$73,$E7,$FB,$DF,$57,$DE,$DF,$C0,$7D,$7F,$7F
	.dc.b	$EC,$7B,$FF,$67,$DF,$FB,$5E,$FF,$DB,$F7,$FE,$E7,$BF,$F7,$7D,$FF
	.dc.b	$BD,$EF,$FD,$FF,$7D,$E9,$CF,$B8,$FF,$63,$1F,$B1,$4F,$EC,$5B,$EF
	.dc.b	$F0,$73,$EF,$D9,$FE,$3B,$1C,$3D,$EC,$7B,$D9,$F7,$B5,$EF,$6F,$D9
	.dc.b	$E1,$EB,$71,$F5,$B9,$7A,$DC,$FD,$6E,$9E,$B7,$5F,$5B,$B7,$AD,$DF
	.dc.b	$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$79,$7B,$FF,$63,$C7,$EE,$5E,$E6
	.dc.b	$7D,$6D,$7A,$DB,$F5,$B8,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7,$AD
	.dc.b	$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B
	.dc.b	$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5,$BE,$67,$DC,$FE,$C8,$7C,$C6
	.dc.b	$3E,$43,$3F,$0B,$AE,$AD,$FC,$2F,$0F,$80,$E3,$F0,$1C,$BE,$03,$9F
	.dc.b	$C0,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$05,$F7,$5F,$83,$F3
	.dc.b	$0F,$E2,$4F,$C4,$0F,$7F,$7E,$FF,$0F,$85,$E3,$EF,$F2,$F7,$F9,$FB
	.dc.b	$FD,$3D,$FE,$BE,$FF,$6F,$7F,$BF,$BF,$E0,$F7,$FC,$3E,$FF,$8B,$DF
	.dc.b	$F1,$FB,$FE,$4F,$7F,$CB,$EF,$F9,$BD,$FF,$3F,$BF,$E8,$F7,$FF,$35
	.dc.b	$F7,$7F,$86,$E7,$3F,$86,$7F,$4A,$FF,$09,$6F,$BD,$C3,$DE,$E3,$EF
	.dc.b	$72,$F7,$B9,$FB,$DD,$3D,$EE,$BE,$F7,$6F,$7B,$BF,$BD,$E0,$F7,$BC
	.dc.b	$3E,$F7,$8B,$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD
	.dc.b	$E8,$F7,$BF,$65,$C5,$FC,$E0,$FC,$3F,$27,$EF,$17,$B9,$AF,$73,$7E
	.dc.b	$E7,$0F,$73,$8F,$B9,$CB,$DC,$E7,$EE,$74,$F7,$3A,$FB,$9D,$BD,$CE
	.dc.b	$FE,$E7,$83,$DC,$F0,$FB,$9E,$2F,$73,$C7,$EE,$79,$3D,$CF,$2F,$B9
	.dc.b	$E6,$F7,$3C,$FE,$E7,$F0,$9F,$BC,$FC,$41,$FC,$3C,$FE,$1C,$7E,$1F
	.dc.b	$9F,$F8,$AB,$FE,$9C,$3F,$86,$9F,$EC,$97,$3E,$C7,$3F,$63,$A7,$B1
	.dc.b	$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1,$7B,$1E,$3F,$63
	.dc.b	$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7,$F2,$1F,$BD,$FC
	.dc.b	$41,$F8,$7E,$23,$F8,$76,$FB,$1B,$F2,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3F,$9C,$7D
	.dc.b	$F7,$F7,$E3,$F0,$FC,$A7,$F0,$EF,$F4,$47,$FC,$E0,$7F,$7F,$75,$EB
	.dc.b	$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF,$AD,$E0,$F5,$BC
	.dc.b	$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79,$BD,$6F,$3F,$AD
	.dc.b	$E8,$F5,$BF,$A8,$7B,$9F,$C3,$CF,$EA,$C7,$F5,$53,$F1,$1C,$7F,$EE
	.dc.b	$0F,$77,$71,$C3,$DD,$E3,$EC,$72,$F6,$39,$FB,$1D,$3D,$8E,$BE,$C7
	.dc.b	$6F,$63,$BF,$B1,$E0,$F6,$3C,$3E,$C7,$8B,$D8,$F1,$FB,$1E,$4F,$63
	.dc.b	$CB,$EC,$79,$BD,$8F,$3F,$B1,$FD,$9F,$EF,$FF,$87,$1F,$87,$3F,$0C
	.dc.b	$3D,$8D,$7B,$1B,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7,$3F,$63,$A7,$B1
	.dc.b	$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1,$7B,$1E,$3F,$63
	.dc.b	$C9,$EC,$79,$7D,$8F,$37,$B1,$FD,$F7,$F0,$0F,$F2,$19,$F8,$63,$F0
	.dc.b	$BE,$41,$FF,$5A,$1F,$85,$3F,$C8,$4E,$7D,$7E,$5E,$BF,$3F,$5F,$A7
	.dc.b	$AF,$D7,$D7,$ED,$EB,$F7,$F5,$FC,$1E,$BF,$87,$D7,$F1,$7A,$FE,$3F
	.dc.b	$5F,$C9,$EB,$F9,$7D,$7F,$37,$AF,$E7,$F5,$FF,$C2,$9F,$80,$FF,$53
	.dc.b	$3F,$C3,$47,$F8,$64,$FC,$28,$F6,$37,$EC,$70,$F6,$38,$FB,$1C,$BD
	.dc.b	$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F,$B1
	.dc.b	$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7F,$8A
	.dc.b	$FD,$9F,$EA,$47,$F0,$B3,$F8,$51,$F8,$5F,$17,$FC,$6E,$BD,$7E,$1E
	.dc.b	$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F
	.dc.b	$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB
	.dc.b	$F9,$FD,$7F,$F2,$17,$E0,$5F,$E8,$83,$F8,$59,$F8,$63,$FD,$F0,$7F
	.dc.b	$09,$3F,$0A,$3D,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6
	.dc.b	$F5,$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC
	.dc.b	$BE,$BF,$9B,$D7,$FF,$2C,$7E,$07,$FE,$17,$3F,$CB,$C7,$F9,$74,$FE
	.dc.b	$12,$7E,$17,$C5,$FF,$32,$BF,$E5,$C3,$FA,$82,$5E,$A7,$4F,$53,$AF
	.dc.b	$A9,$DB,$D4,$EF,$EA,$78,$3D,$4F,$0F,$A9,$E2,$F5,$3C,$7E,$A7,$93
	.dc.b	$D4,$F2,$FA,$9E,$6F,$53,$CF,$EA,$7A,$3D,$4F,$4F,$A9,$FE,$71,$FC
	.dc.b	$13,$FC,$30,$7F,$51,$3F,$A8,$1F,$87,$E3,$3F,$87,$7F,$CF,$EB,$D3
	.dc.b	$E5,$E9,$F3,$F4,$FA,$7A,$7D,$7D,$3E,$DE,$9F,$7F,$4F,$C1,$E9,$F8
	.dc.b	$7D,$3F,$17,$A7,$E3,$F4,$FC,$9E,$9F,$97,$D3,$F3,$7A,$7E,$7F,$4F
	.dc.b	$D1,$E9,$FF,$A2,$78,$7F,$D9,$E7,$F5,$23,$F8,$68,$F8,$1D,$7C,$0E
	.dc.b	$FD,$7E,$1E,$AF,$1F,$57,$97,$AB,$CF,$D5,$E9,$EA,$F5,$F5,$7B,$7A
	.dc.b	$BD,$FD,$5F,$07,$AB,$E1,$F5,$7C,$5E,$AF,$8F,$D5,$F2,$7A,$BE,$5F
	.dc.b	$57,$CD,$EA,$FF,$A7,$FF,$05,$FF,$2F,$1F,$EA,$53,$FD,$48,$7F,$50
	.dc.b	$3F,$11,$E3,$1F,$F5,$19,$F8,$83,$F0,$EF,$FA,$C1,$7A,$9D,$7D,$4E
	.dc.b	$DE,$A7,$7F,$53,$C1,$EA,$78,$7D,$4F,$17,$A9,$E3,$F5,$3C,$9E,$A7
	.dc.b	$97,$D4,$F3,$7A,$9E,$7F,$53,$D1,$EA,$7A,$7D,$4F,$F5,$C7,$E0,$DF
	.dc.b	$E5,$F3,$FC,$30,$7F,$85,$CF,$C4,$1F,$87,$E2,$3F,$87,$7F,$D8,$EB
	.dc.b	$D4,$E7,$EA,$74,$F5,$3A,$FA,$9D,$BD,$4E,$FE,$A7,$83,$D4,$F0,$FA
	.dc.b	$9E,$2F,$53,$C7,$EA,$79,$3D,$4F,$2F,$A9,$E6,$F5,$3C,$FE,$A7,$A3
	.dc.b	$D4,$FF,$68,$7E,$0E,$F9,$EB,$F1,$58,$F8,$7C,$FA,$FA,$F5,$F7,$EB
	.dc.b	$F0,$F5,$F8,$FA,$FC,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF
	.dc.b	$EB,$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE
	.dc.b	$6F,$5F,$FD,$CD,$EF,$7F,$A9,$4F,$F7,$61,$FE,$EB,$3F,$C2,$E7,$E2
	.dc.b	$0F,$F7,$49,$FC,$38,$FC,$3F,$92,$7F,$0E,$FF,$BE,$17,$AD,$DB,$D6
	.dc.b	$EF,$EB,$78,$3D,$6F,$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6,$F2,$FA
	.dc.b	$DE,$6F,$5B,$CF,$EB,$7A,$3D,$6F,$4F,$AD,$FE,$FF,$FC,$23,$FD,$4C
	.dc.b	$7F,$97,$CF,$F2,$F1,$FC,$3C,$FC,$41,$F8,$7E,$13,$F8,$74,$63,$D5
	.dc.b	$E9,$EA,$F5,$F5,$7B,$7A,$BD,$FD,$5F,$07,$AB,$E1,$F5,$7C,$5E,$AF
	.dc.b	$8F,$D5,$F2,$7A,$BE,$5F,$57,$CD,$EA,$F9,$FD,$5F,$47,$AB,$E0,$E0
	.dc.b	$7E,$13,$F8,$97,$C6,$FF,$98,$8F,$F0,$E9,$FC,$45,$CF,$A3,$C6,$3F
	.dc.b	$C3,$A7,$F5,$98,$E3,$E7,$E7,$CF,$D7,$9E,$FE,$01,$85,$AE,$F0,$F9
	.dc.b	$7F,$87,$F2,$0E,$5E,$31,$86,$D2,$74,$F2,$9D,$34,$D4,$F0,$9D,$C7
	.dc.b	$C0,$AF,$03,$A3,$84,$6B,$24,$EA,$30,$DA,$E8,$4A,$F5,$7C,$9B,$D5
	.dc.b	$7C,$B7,$DE,$FC,$2A,$D6,$BD,$7B,$18,$D1,$87,$1B,$E0,$7E,$1A,$33
	.dc.b	$5C,$23,$12,$DF,$18,$78,$9B,$7D,$37,$B5,$BB,$EB,$7E,$8D,$33,$EB
	.dc.b	$E9,$B3,$AC,$3C,$BE,$BF,$23,$9F,$AF,$67,$58,$BD,$7E,$FE,$BF,$81
	.dc.b	$B3,$AF,$5F,$C7,$EB,$F9,$1B,$1C,$BD,$7F,$CB,$71,$7C,$5C,$5E,$BE
	.dc.b	$2B,$E0,$B7,$83,$E8,$DD,$DB,$E5,$FA,$F7,$07,$AF,$4C,$46,$FF,$65
	.dc.b	$FD,$26,$0C,$FE,$A4,$3E,$83,$46,$E7,$87,$CF,$FC,$E9,$7C,$A5,$8C
	.dc.b	$39,$F8,$C9,$BF,$CB,$A7,$C6,$63,$0F,$04,$FD,$F9,$FE,$3C,$EF,$E2
	.dc.b	$7B,$7B,$DD,$DF,$07,$C3,$7A,$7D,$F9,$87,$58,$F6,$7B,$FE,$DF,$7E
	.dc.b	$4F,$86,$3E,$07,$BF,$23,$C8,$FC,$3F,$7F,$E2,$7B,$FE,$D3,$DD,$EF
	.dc.b	$DD,$7D,$5E,$FF,$C1,$F7,$FE,$13,$63,$DD,$EF,$FC,$5F,$7F,$E3,$36
	.dc.b	$3E,$17,$BF,$F4,$7D,$F9,$AE,$0F,$C8,$57,$EA,$1F,$8A,$7F,$07,$5F
	.dc.b	$2F,$F7,$DB,$1C,$2F,$1E,$7C,$8E,$6E,$FE,$43,$87,$C7,$71,$F8,$B9
	.dc.b	$5E,$08,$E9,$F1,$32,$3C,$6F,$7F,$83,$F0,$7C,$1F,$87,$E0,$E5,$7A
	.dc.b	$23,$C9,$EB,$79,$7E,$0E,$4F,$B2,$3D,$1F,$07,$23,$DE,$FA,$FE,$0F
	.dc.b	$D9,$F0,$7E,$DF,$83,$95,$F1,$23,$E0,$7C,$1F,$C1,$F8,$39,$3F,$22
	.dc.b	$3E,$27,$C1,$FC,$5F,$83,$93,$F3,$23,$E4,$7C,$1F,$E9,$FF,$66,$FE
	.dc.b	$85,$3F,$3B,$D3,$46,$3D,$0F,$E2,$DD,$4B,$E1,$F1,$1C,$7E,$22,$63
	.dc.b	$FC,$1F,$04,$F5,$F8,$8E,$DF,$0F,$2B,$51,$E1,$F8,$59,$1C,$5F,$27
	.dc.b	$C2,$F9,$7E,$17,$F8,$DE,$3F,$C6,$6B,$CC,$BE,$17,$33,$AF,$84,$9E
	.dc.b	$1F,$8F,$E1,$6E,$1C,$3E,$17,$99,$D3,$E1,$6D,$7A,$7E,$13,$E2,$12
	.dc.b	$BD,$9F,$0B,$F1,$FE,$12,$47,$BD,$F9,$5F,$0B,$F2,$FE,$13,$E6,$7C
	.dc.b	$2C,$AF,$89,$1F,$3F,$E0,$FE,$87,$C2,$49,$F9,$11,$FC,$67,$F6,$7F
	.dc.b	$E2,$7D,$9F,$D1,$5B,$AF,$F4,$36,$D7,$F6,$97,$F8,$B3,$FC,$DD,$7C
	.dc.b	$27,$3F,$84,$93,$98,$ED,$F0,$9F,$D1,$A8,$DF,$84,$C3,$9F,$83,$98
	.dc.b	$F5,$F8,$3B,$8F,$E1,$F0,$FF,$3C,$CF,$87,$E0,$BA,$7C,$1F,$57,$B7
	.dc.b	$C1,$7B,$7E,$0E,$61,$E7,$8F,$81,$F0,$72,$3D,$4F,$C3,$F8,$3F,$89
	.dc.b	$F0,$72,$7D,$CF,$C7,$F8,$3F,$91,$F0,$52,$7E,$13,$F2,$FE,$0F,$E6
	.dc.b	$79,$92,$7E,$33,$F3,$FE,$0F,$E8,$7C,$1F,$D1,$F8,$39,$5F,$2E,$3F
	.dc.b	$A6,$7E,$D1,$FD,$F3,$8B,$FB,$23,$FC,$49,$A7,$F7,$07,$13,$FA,$B7
	.dc.b	$FC,$31,$1C,$3E,$0B,$5F,$07,$B7,$87,$C1,$71,$F8,$3B,$5D,$9E,$9F
	.dc.b	$07,$E2,$F8,$39,$97,$8A,$3C,$BF,$05,$23,$CA,$FA,$3E,$0F,$D3,$F0
	.dc.b	$52,$7D,$2F,$B3,$E0,$FD,$BF,$07,$EE,$F8,$39,$5E,$E8,$F8,$3F,$05
	.dc.b	$23,$E0,$BF,$13,$E0,$FE,$2F,$C1,$49,$F8,$AF,$C8,$F8,$3F,$93,$F0
	.dc.b	$5F,$17,$E0,$FE,$33,$78,$BE,$5F,$C1,$7C,$93,$FC,$33,$E8,$FF,$A7
	.dc.b	$63,$27,$F8,$4D,$FE,$1A,$DF,$82,$DF,$C1,$70,$78,$FC,$0C,$9E,$1F
	.dc.b	$05,$75,$FE,$3D,$87,$F8,$BD,$BF,$03,$E0,$F8,$2F,$0B,$E2,$F8,$0F
	.dc.b	$1F,$C0,$DC,$7C,$4F,$9B,$E0,$7C,$FF,$03,$67,$CC,$FA,$BE,$06,$47
	.dc.b	$A7,$E0,$6E,$7E,$2F,$81,$F7,$FC,$0F,$C0,$6D,$7B,$FE,$07,$E1,$9F
	.dc.b	$13,$E0,$6C,$7C,$3F,$81,$F8,$E7,$C8,$E1,$B1,$F1,$FE,$03,$E5,$9F
	.dc.b	$33,$E0,$6C,$7F,$8F,$7F,$69,$FF,$49,$B7,$E1,$FF,$C9,$31,$8D,$7F
	.dc.b	$A2,$3F,$E2,$39,$E1,$F0,$3A,$37,$F0,$56,$3A,$7C,$0F,$23,$9F,$C1
	.dc.b	$58,$F0,$7C,$17,$6F,$82,$EE,$D8,$F2,$7C,$07,$8B,$E0,$7C,$6D,$8F
	.dc.b	$47,$C0,$79,$BE,$07,$CE,$D8,$F6,$7C,$0F,$AB,$E0,$7D,$6D,$8F,$81
	.dc.b	$F0,$3E,$E3,$DF,$F0,$36,$3E,$27,$C0,$FC,$2F,$81,$F8,$6D,$8F,$91
	.dc.b	$F0,$3F,$18,$F8,$FF,$03,$63,$E6,$7C,$07,$CA,$3E,$5F,$C0,$D8,$FF
	.dc.b	$36,$7E,$D5,$FC,$2F,$A7,$FB,$83,$7E,$13,$FD,$08,$FF,$0A,$6C,$FF
	.dc.b	$9F,$23,$F8,$5C,$F0,$F8,$2C,$7C,$16,$5D,$7C,$16,$FE,$0A,$E1,$E0
	.dc.b	$79,$7C,$0F,$3F,$82,$B3,$E4,$7B,$7C,$17,$F9,$AF,$33,$4C,$7C,$16
	.dc.b	$7E,$0A,$6B,$EC,$78,$7C,$17,$1F,$81,$93,$F0,$1E,$9F,$05,$D7,$E0
	.dc.b	$BD,$87,$B7,$E0,$AE,$BF,$1B,$E0,$BE,$01,$F0,$7E,$0A,$C7,$CA,$F8
	.dc.b	$1F,$89,$F0,$5F,$15,$B1,$F3,$BE,$0B,$E4,$1F,$EA,$4F,$DA,$FF,$86
	.dc.b	$6B,$F4,$23,$1F,$07,$9F,$83,$B7,$FE,$A1,$B4,$71,$F8,$3E,$5F,$07
	.dc.b	$6B,$4F,$5F,$83,$91,$C7,$E0,$EE,$7F,$E2,$DC,$7F,$C3,$5D,$FF,$EC
	.dc.b	$17,$FC,$35,$5D,$FE,$0F,$CD,$F0,$7E,$7F,$83,$F4,$7C,$1D,$AF,$23
	.dc.b	$EB,$F8,$3F,$F5,$06,$0F,$83,$A6,$23,$3F,$05,$34,$F6,$3C,$3E,$0F
	.dc.b	$8F,$C1,$F2,$F8,$39,$5F,$06,$3A,$FC,$1F,$C9,$3E,$57,$C1,$DD,$7E
	.dc.b	$37,$C1,$FC,$DF,$83,$F9,$CD,$8F,$F6,$57,$5F,$F0,$BE,$4F,$F7,$03
	.dc.b	$FC,$2F,$2F,$F2,$83,$99,$FF,$70,$39,$9D,$FC,$0F,$FA,$F6,$3F,$D5
	.dc.b	$F5,$D3,$E0,$71,$39,$F8,$1D,$7C,$0C,$AF,$0C,$71,$F8,$0E,$5F,$03
	.dc.b	$27,$CB,$1D,$7E,$03,$B7,$C0,$C9,$F4,$C7,$87,$E0,$3C,$5F,$03,$27
	.dc.b	$DB,$1E,$5F,$80,$F3,$7C,$0C,$9F,$83,$1E,$9F,$81,$FF,$7A,$F0,$1B
	.dc.b	$F0,$38,$73,$F0,$13,$93,$E4,$FC,$0D,$C7,$E3,$FC,$0F,$23,$9F,$C0
	.dc.b	$DA,$F9,$FF,$03,$DB,$E0,$7B,$B6,$3F,$DF,$9F,$B6,$7F,$0F,$58,$3E
	.dc.b	$21,$39,$8C,$3F,$C4,$E7,$73,$A4,$74,$F8,$2B,$5F,$E8,$BA,$FE,$2F
	.dc.b	$D9,$8B,$EB,$F0,$7A,$8C,$4E,$B0,$EA,$BE,$0F,$FD,$F8,$DC,$3D,$31
	.dc.b	$8F,$83,$CF,$C1,$C9,$F6,$BC,$3E,$0F,$8F,$C1,$49,$F8,$2F,$4F,$82
	.dc.b	$EB,$F0,$5E,$3C,$FC,$3F,$05,$6F,$E0,$CD,$DF,$E2,$CD,$D5,$F0,$54
	.dc.b	$C7,$C1,$5C,$7E,$7F,$C1,$6C,$E1,$F0,$56,$3E,$9F,$C1,$7B,$BE,$0A
	.dc.b	$6B,$83,$EB,$F7,$FC,$22,$3F,$6B,$E0,$BE,$2F,$C1,$7D,$A8,$FF,$5D
	.dc.b	$7E,$DB,$FE,$B9,$C7,$EB,$B6,$3E,$DC,$EF,$A3,$EA,$4F,$F8,$B2,$78
	.dc.b	$FC,$1F,$BE,$3F,$90,$DF,$E1,$AB,$73,$9F,$82,$93,$93,$BF,$C1,$63
	.dc.b	$0F,$1F,$C1,$73,$3A,$7C,$15,$AE,$13,$F9,$1C,$BF,$92,$C7,$A3,$E0
	.dc.b	$7C,$5F,$05,$E3,$7C,$9F,$05,$27,$8F,$4A,$DF,$05,$9C,$7E,$1F,$C0
	.dc.b	$7A,$FE,$07,$D8,$DC,$38,$DB,$E0,$7E,$24,$7C,$8F,$81,$B8,$79,$5B
	.dc.b	$97,$7A,$B6,$F8,$0A,$4B,$EB,$F0,$3A,$F8,$19,$1E,$07,$8F,$C0,$F2
	.dc.b	$F8,$0F,$A1,$F0,$3F,$45,$BA,$FB,$BF,$6E,$FE,$4D,$87,$ED,$DB,$A9
	.dc.b	$FE,$51,$3F,$B6,$72,$7F,$B1,$1E,$5F,$05,$F9,$B8,$C5,$7F,$C0,$15
	.dc.b	$19,$F8,$0F,$7C,$3E,$06,$98,$8C,$FC,$06,$BE,$06,$61,$E7,$8E,$3F
	.dc.b	$01,$23,$D4,$F4,$F8,$1E,$BF,$01,$27,$DC,$F8,$3E,$07,$C3,$F0,$12
	.dc.b	$7E,$13,$E4,$F8,$1F,$2F,$C0,$79,$BE,$06,$57,$C7,$8F,$4F,$C0,$FB
	.dc.b	$CF,$81,$F0,$37,$8F,$E6,$7C,$0F,$C3,$3E,$27,$C0,$D8,$FA,$1F,$01
	.dc.b	$F1,$CF,$91,$F0,$36,$3F,$23,$3F,$C8,$32,$79,$BF,$C7,$E1,$F1,$7F
	.dc.b	$88,$6C,$F0,$AC,$7C,$16,$7E,$0A,$4F,$48,$E1,$F0,$52,$3B,$BC,$FE
	.dc.b	$0B,$A7,$C1,$7F,$15,$C4,$DF,$82,$C3,$9F,$82,$99,$FA,$3E,$0A,$E5
	.dc.b	$E8,$8F,$E7,$B8,$3E,$0B,$A7,$C1,$75,$6C,$FC,$0F,$81,$F6,$7C,$0C
	.dc.b	$C3,$07,$C4,$F8,$1F,$81,$F0,$32,$30,$7C,$8F,$81,$F8,$84,$7E,$5F
	.dc.b	$C0,$FC,$7F,$81,$F9,$1F,$03,$27,$E7,$C7,$CB,$F8,$1F,$D9,$FE,$DF
	.dc.b	$FC,$77,$CE,$37,$85,$3A,$F8,$2D,$FC,$34,$C7,$F8,$4D,$A7,$9F,$C1
	.dc.b	$C9,$CB,$DB,$E0,$FB,$FC,$1F,$83,$E0,$E5,$72,$8F,$1F,$C1,$C8,$EA
	.dc.b	$F9,$BE,$0F,$CF,$F0,$7F,$D2,$F3,$FE,$2F,$D1,$E2,$8C,$7C,$1E,$7E
	.dc.b	$0B,$5F,$07,$6B,$CE,$F1,$F8,$2E,$5F,$07,$67,$D6,$F5,$F8,$2E,$DF
	.dc.b	$05,$67,$DE,$F8,$7E,$0A,$47,$C2,$F8,$2B,$C1,$F8,$33,$FC,$09,$B3
	.dc.b	$E0,$7D,$1F,$05,$E9,$6D,$FF,$32,$FD,$C3,$FA,$95,$F1,$EC,$43,$E2
	.dc.b	$BF,$94,$B6,$78,$18,$F8,$78,$F3,$77,$F0,$BC,$3E,$0E,$4F,$77,$9F
	.dc.b	$C2,$74,$F8,$39,$3E,$37,$BF,$C2,$78,$3E,$12,$4F,$9D,$F1,$FC,$27
	.dc.b	$93,$E1,$24,$FA,$DF,$3F,$C2,$7A,$3E,$13,$FC,$05,$B1,$BF,$09,$87
	.dc.b	$3F,$07,$38,$FE,$27,$C2,$5C,$BE,$26,$0F,$84,$E7,$F0,$9D,$1B,$7F
	.dc.b	$2F,$E1,$3B,$F9,$DE,$06,$C7,$CF,$F8,$4F,$19,$E4,$F8,$4B,$1F,$D9
	.dc.b	$3F,$71,$FF,$1D,$75,$9A,$7F,$8E,$F4,$FF,$20,$38,$F8,$2F,$F0,$F3
	.dc.b	$FE,$33,$8C,$FC,$1E,$27,$3F,$07,$AF,$83,$B8,$F7,$78,$FC,$17,$2F
	.dc.b	$83,$B3,$E3,$7A,$FC,$1F,$F6,$AC,$DF,$07,$4C,$46,$7E,$0A,$6B,$EB
	.dc.b	$78,$7C,$1F,$1F,$83,$93,$EF,$7A,$7C,$1F,$5F,$83,$F5,$9E,$CF,$83
	.dc.b	$BA,$FC,$5F,$83,$F7,$FC,$1F,$C0,$6C,$7C,$9F,$82,$F8,$7F,$07,$F1
	.dc.b	$1B,$1F,$37,$E0,$FD,$44,$D3,$FC,$27,$E7,$FF,$13,$3F,$86,$38,$F8
	.dc.b	$3F,$F0,$4C,$7F,$B3,$23,$7F,$07,$8F,$83,$CF,$C1,$C9,$E7,$1C,$3E
	.dc.b	$0B,$FC,$A5,$66,$FC,$1E,$1C,$FC,$1E,$BE,$0E,$EB,$E4,$78,$FC,$1F
	.dc.b	$2F,$83,$B3,$E8,$7A,$FC,$1F,$9B,$E0,$E6,$BE,$C8,$F4,$FC,$1C,$8F
	.dc.b	$7B,$EC,$F8,$3F,$6F,$C1,$FB,$BE,$0E,$57,$C4,$8F,$83,$F0,$7E,$AF
	.dc.b	$83,$F5,$B7,$4F,$93,$C5,$EE,$3D,$FF,$05,$63,$FC,$7F,$FB,$97,$F8
	.dc.b	$80,$FF,$0F,$E3,$FE,$99,$71,$F0,$7F,$E7,$75,$F0,$7F,$E3,$87,$FD
	.dc.b	$93,$3F,$C5,$5B,$F0,$DF,$E9,$57,$71,$DB,$E0,$F8,$7C,$1C,$8C,$1E
	.dc.b	$2F,$83,$E8,$47,$CB,$F0,$7D,$FE,$0F,$C5,$F0,$7E,$36,$EB,$EA,$F8
	.dc.b	$3F,$31,$E7,$F8,$3B,$1E,$EF,$82,$F5,$1E,$BF,$83,$B1,$F0,$BE,$0F
	.dc.b	$D8,$4C,$FE,$2F,$C1,$FB,$FE,$06,$47,$C8,$7E,$17,$C1,$7C,$3F,$81
	.dc.b	$93,$F3,$1F,$8D,$F0,$5F,$1F,$E0,$BE,$57,$C1,$7C,$B6,$E7,$FE,$70
	.dc.b	$FD,$CF,$F8,$69,$3E,$1F,$FD,$4C,$E3,$E0,$AF,$C1,$E1,$B5,$C7,$E0
	.dc.b	$FF,$CD,$75,$FE,$BD,$D3,$FA,$43,$7B,$3F,$D9,$0B,$E0,$F7,$F0,$72
	.dc.b	$BC,$4F,$2F,$83,$E7,$F0,$32,$7C,$CF,$6F,$82,$EF,$F0,$52,$7D,$4F
	.dc.b	$8B,$E0,$BC,$7F,$05,$E4,$F8,$29,$5E,$F8,$F3,$FC,$17,$A3,$E0,$A4
	.dc.b	$FC,$38,$F5,$FC,$0F,$C2,$F8,$2F,$86,$DE,$3F,$91,$F0,$3F,$1B,$E0
	.dc.b	$BE,$3B,$63,$E6,$7C,$17,$CA,$3F,$DA,$1F,$BA,$3E,$5F,$13,$FE,$A0
	.dc.b	$8F,$F5,$C7,$3F,$FB,$B5,$FF,$5B,$E9,$FD,$C5,$CC,$FF,$BE,$1C,$CE
	.dc.b	$FE,$07,$A9,$DB,$E0,$AD,$75,$F8,$1F,$0F,$C1,$78,$9B,$1E,$1F,$82
	.dc.b	$FF,$51,$E2,$69,$8F,$82,$CF,$C1,$4D,$3D,$0F,$0F,$82,$E3,$F0,$32
	.dc.b	$7D,$8F,$4F,$82,$EB,$F0,$5F,$08,$F8,$7F,$05,$74,$F8,$5F,$03,$F1
	.dc.b	$BE,$0B,$E3,$B6,$3E,$37,$C0,$FC,$AF,$82,$F9,$6D,$8F,$95,$F0,$5F
	.dc.b	$3B,$E0,$BE,$7B,$63,$FD,$9B,$D7,$FE,$AE,$3F,$87,$FC,$31,$C7,$C1
	.dc.b	$A7,$7F,$07,$78,$7F,$C4,$55,$FE,$AE,$EB,$C6,$1C,$7E,$23,$51,$89
	.dc.b	$D6,$0B,$5F,$7F,$83,$FF,$68,$D1,$A6,$3E,$0F,$3F,$07,$AF,$83,$98
	.dc.b	$7A,$A3,$8F,$C1,$48,$F6,$BD,$3E,$0F,$AF,$C1,$F6,$F8,$39,$5F,$0A
	.dc.b	$3C,$3F,$07,$E2,$F8,$39,$3F,$1A,$3C,$BF,$07,$23,$E4,$BE,$8F,$83
	.dc.b	$F4,$FC,$1F,$AB,$E0,$E5,$7C,$E8,$F6,$FC,$1F,$BB,$E0,$E4,$FD,$28
	.dc.b	$F8,$3F,$05,$23,$EA,$BF,$13,$E0,$FF,$DE,$BF,$80,$7F,$AC,$FA,$7E
	.dc.b	$7B,$7A,$7E,$67,$4A,$31,$87,$CE,$7F,$D6,$AF,$F9,$65,$7C,$17,$4F
	.dc.b	$83,$93,$F5,$70,$EB,$3E,$0F,$83,$93,$9C,$7B,$E1,$FC,$6D,$BF,$05
	.dc.b	$E6,$F8,$2F,$C0,$C6,$9B,$5E,$3F,$82,$BA,$79,$E6,$E0,$F8,$2B,$F0
	.dc.b	$58,$6E,$3E,$1C,$7D,$9F,$05,$C3,$E0,$B8,$B6,$75,$97,$C3,$F8,$1E
	.dc.b	$A7,$6F,$82,$B3,$AC,$DF,$03,$E2,$3C,$7F,$05,$67,$5F,$03,$E6,$3C
	.dc.b	$FF,$05,$63,$97,$C1,$7A,$BC,$AF,$5B,$63,$B7,$C1,$7B,$8F,$83,$FB
	.dc.b	$AF,$F2,$18,$C9,$E8,$7F,$8E,$F0,$FE,$99,$BB,$9F,$E3,$F3,$F8,$07
	.dc.b	$23,$E0,$BA,$9D,$BE,$0A,$CF,$5F,$81,$F0,$9E,$2F,$82,$B1,$E1,$F8
	.dc.b	$1F,$29,$E6,$F8,$1B,$1E,$5F,$81,$FD,$D6,$9F,$EA,$3C,$DF,$03,$8F
	.dc.b	$07,$F8,$1C,$62,$78,$1E,$DF,$81,$D6,$1E,$9F,$81,$B3,$EF,$6F,$07
	.dc.b	$BE,$AD,$BE,$06,$92,$FE,$47,$C0,$EB,$E0,$3E,$69,$F3,$BE,$06,$E7
	.dc.b	$F3,$7E,$03,$E8,$FC,$0F,$D2,$6C,$7D,$1F,$80,$FA,$A7,$D6,$F8,$1B
	.dc.b	$1F,$8C,$FD,$DB,$F8,$D3,$73,$FC,$DC,$7F,$2A,$32,$FF,$B1,$DF,$E4
	.dc.b	$2F,$F2,$06,$EE,$5F,$01,$CB,$E0,$79,$FC,$0F,$4F,$81,$B8,$F8,$5E
	.dc.b	$FF,$01,$23,$C9,$F0,$37,$3F,$07,$C0,$F9,$3E,$07,$CA,$DA,$F5,$7C
	.dc.b	$07,$A0,$F4,$FC,$0D,$8F,$77,$C0,$7B,$0F,$6F,$C0,$D8,$F8,$5F,$03
	.dc.b	$F0,$0F,$83,$F0,$36,$3E,$37,$C0,$7C,$4F,$81,$F8,$AD,$8F,$95,$F0
	.dc.b	$1F,$7B,$94,$D3,$1F,$03,$9F,$81,$9C,$3F,$45,$E1,$F0,$3F,$86,$7F
	.dc.b	$8F,$F4,$9F,$E3,$FE,$2F,$F5,$46,$E5,$FC,$99,$D9,$FC,$8A,$FF,$83
	.dc.b	$AF,$81,$E8,$75,$F8,$1E,$DF,$03,$2B,$6F,$87,$E0,$7C,$5F,$01,$E3
	.dc.b	$F8,$19,$5D,$23,$CD,$F0,$1E,$7F,$81,$93,$E0,$8F,$57,$C0,$7A,$FE
	.dc.b	$06,$4F,$92,$3D,$DF,$03,$23,$CE,$FC,$1F,$81,$F8,$5F,$01,$F0,$FE
	.dc.b	$02,$57,$B2,$3E,$37,$C0,$7C,$7F,$80,$93,$F0,$23,$E5,$7B,$FF,$2F
	.dc.b	$E0,$24,$FC,$48,$F9,$DF,$01,$FB,$4F,$DD,$FF,$C7,$13,$FC,$B9,$AB
	.dc.b	$A7,$E8,$4E,$FE,$27,$87,$C5,$4A,$E1,$1C,$FE,$0E,$47,$37,$B7,$C1
	.dc.b	$F7,$F8,$39,$3D,$DF,$17,$C1,$F8,$FE,$0E,$4F,$8D,$F3,$7C,$1F,$9F
	.dc.b	$E0,$E4,$F9,$DF,$57,$C1,$FA,$FE,$0B,$FA,$9F,$01,$BF,$07,$87,$3F
	.dc.b	$05,$3C,$6F,$81,$F0,$57,$2F,$81,$1F,$DC,$B0,$7C,$17,$43,$AF,$C1
	.dc.b	$59,$F9,$1F,$03,$E0,$F8,$2F,$0B,$63,$E6,$7C,$17,$90,$F2,$FC,$15
	.dc.b	$8F,$EB,$BF,$BC,$7F,$0E,$EB,$FF,$18,$BF,$E3,$A3,$11,$FE,$04,$71
	.dc.b	$F0,$5F,$DC,$1F,$E1,$F1,$CA,$31,$F0,$79,$F8,$39,$3D,$BE,$0E,$F0
	.dc.b	$7B,$8F,$F1,$CE,$0F,$83,$E9,$F0,$7D,$5B,$7E,$6F,$82,$F0,$1E,$1F
	.dc.b	$83,$B1,$EA,$F8,$3F,$27,$C1,$F9,$5B,$1E,$EF,$83,$F4,$1E,$9F,$83
	.dc.b	$B1,$F0,$BE,$0F,$D9,$F0,$7E,$D6,$C7,$C6,$F8,$2F,$80,$7C,$1F,$83
	.dc.b	$B1,$F2,$BE,$0B,$E2,$1F,$17,$E0,$AC,$7C,$EF,$80,$F9,$1F,$05,$F2
	.dc.b	$5B,$1F,$E0,$EF,$DE,$7F,$CC,$CD,$F8,$BF,$F0,$EC,$63,$9B,$E4,$7C
	.dc.b	$22,$13,$AF,$83,$97,$CF,$E1,$2E,$1C,$2D,$F0,$9D,$3E,$13,$AB,$6F
	.dc.b	$C5,$F0,$7E,$03,$C3,$F0,$96,$3C,$DF,$07,$E4,$3C,$BF,$09,$63,$D5
	.dc.b	$F0,$7E,$83,$D3,$F0,$96,$3D,$DF,$09,$EC,$3D,$BF,$09,$63,$E1,$7C
	.dc.b	$1F,$C0,$3E,$0F,$C2,$58,$F8,$DF,$07,$F1,$0F,$8B,$F0,$96,$3E,$57
	.dc.b	$C1,$FC,$8F,$84,$F9,$2D,$8F,$9D,$F0,$9F,$30,$FF,$24,$FA,$1F,$E2
	.dc.b	$9E,$6F,$8D,$18,$F8,$3C,$FC,$25,$BF,$F2,$1D,$FF,$8A,$2F,$FD,$22
	.dc.b	$FF,$89,$E3,$3F,$07,$D3,$E1,$3A,$BD,$BE,$13,$BF,$C2,$5A,$E6,$F8
	.dc.b	$BE,$13,$C7,$F0,$96,$7B,$BE,$6F,$84,$91,$E2,$F8,$4B,$A7,$E1,$34
	.dc.b	$FE,$21,$3E,$6F,$83,$F6,$FC,$27,$B9,$F7,$FC,$24,$9F,$5F,$C2,$5C
	.dc.b	$39,$7C,$1F,$C4,$F8,$3F,$8A,$DA,$F8,$3F,$05,$F2,$3E,$0F,$E4,$B6
	.dc.b	$3E,$2F,$C1,$FC,$CD,$7E,$6B,$63,$FC,$D7,$FB,$D7,$F0,$CC,$BF,$D7
	.dc.b	$2D,$F3,$8F,$FB,$28,$C1,$FE,$1E,$3F,$87,$C7,$E8,$9C,$7C,$17,$FA
	.dc.b	$09,$FF,$60,$57,$6F,$82,$C7,$C1,$65,$D7,$C0,$49,$F1,$FC,$0D,$C7
	.dc.b	$A7,$C0,$F3,$3A,$7C,$0D,$AF,$4F,$C0,$77,$F8,$1F,$03,$63,$DB,$F0
	.dc.b	$3E,$3F,$81,$F2,$36,$3E,$0F,$C0,$79,$FE,$07,$D0,$D8,$F8,$BF,$03
	.dc.b	$EB,$3D,$9F,$03,$63,$E4,$FC,$0F,$BC,$F8,$1F,$03,$63,$E6,$FC,$07
	.dc.b	$C3,$3E,$27,$C0,$D8,$FF,$4D,$FE,$F7,$FC,$2F,$E1,$7F,$DD,$ED,$8F
	.dc.b	$F6,$0B,$FC,$31,$D9,$FE,$AE,$7F,$D7,$EF,$2F,$82,$C7,$C1,$65,$D7
	.dc.b	$C0,$49,$EF,$F0,$37,$9F,$F8,$CE,$0F,$81,$E8,$75,$F8,$1B,$7E,$6F
	.dc.b	$81,$F0,$7C,$0F,$85,$B1,$EA,$F8,$1F,$21,$E5,$F8,$1B,$1E,$EF,$81
	.dc.b	$F4,$7C,$0F,$A5,$B1,$F0,$BE,$07,$D8,$7B,$7E,$06,$C7,$C6,$F8,$1F
	.dc.b	$81,$F0,$3F,$05,$B1,$F2,$BE,$03,$E2,$1F,$17,$E0,$6C,$7C,$EF,$81
	.dc.b	$FF,$6A,$FE,$F9,$FC,$37,$A7,$E7,$C6,$3E,$37,$3F,$07,$AF,$83,$B8
	.dc.b	$7F,$B4,$30,$FE,$1D,$E8,$9F,$E1,$DD,$3F,$D0,$97,$C1,$76,$F8,$3E
	.dc.b	$EF,$83,$E0,$FC,$3F,$07,$71,$E8,$F9,$3E,$0E,$47,$7F,$83,$B9,$75
	.dc.b	$F8,$3F,$49,$EA,$F8,$3B,$5E,$5F,$82,$F6,$FC,$1F,$B9,$B1,$E9,$F8
	.dc.b	$3F,$82,$7C,$2F,$83,$B1,$ED,$F8,$2F,$8A,$7C,$6F,$83,$B1,$F0,$7E
	.dc.b	$0B,$E4,$9F,$2B,$E0,$EC,$7C,$5F,$82,$F9,$BF,$07,$F3,$9B,$1F,$EF
	.dc.b	$8E,$BF,$E1,$84,$F8,$C4,$67,$86,$9F,$DB,$27,$FD,$56,$7F,$0E,$ED
	.dc.b	$CF,$52,$33,$F0,$9C,$27,$31,$FC,$42,$B9,$7C,$16,$70,$43,$0E,$5F
	.dc.b	$07,$98,$C5,$F1,$C3,$9F,$C1,$E6,$FC,$1F,$07,$6B,$FC,$4D,$8F,$B6
	.dc.b	$31,$F0,$72,$3E,$07,$C1,$DC,$BD,$9F,$05,$C4,$E5,$F0,$76,$BE,$37
	.dc.b	$C1,$75,$3B,$7C,$1D,$8F,$95,$F0,$5E,$13,$C5,$F0,$76,$3E,$77,$C1
	.dc.b	$F9,$4F,$37,$C1,$D8,$FA,$5F,$07,$E9,$3D,$5F,$07,$63,$EB,$7C,$1F
	.dc.b	$B4,$F7,$7C,$1D,$8F,$93,$CD,$F4,$BF,$7D,$FE,$2D,$C3,$F2,$DB,$CC
	.dc.b	$7F,$8C,$47,$F8,$99,$C9,$FC,$61,$C7,$C2,$7C,$D8,$C5,$75,$8C,$7C
	.dc.b	$1E,$7E,$0E,$E3,$E2,$3E,$DE,$3D,$E3,$97,$C1,$49,$D6,$5E,$5F,$82
	.dc.b	$C3,$E3,$F8,$1F,$01,$E1,$F8,$2B,$5C,$31,$F5,$7C,$17,$94,$F3,$7C
	.dc.b	$15,$9D,$6E,$F8,$1F,$51,$EB,$F8,$2B,$3A,$F8,$2F,$71,$EF,$F8,$2B
	.dc.b	$1C,$BE,$07,$E1,$7C,$17,$C3,$6C,$76,$F8,$2F,$8D,$F0,$5F,$1D,$B1
	.dc.b	$E2,$F1,$FE,$51,$F9,$7F,$DF,$BF,$8F,$47,$C6,$D7,$F5,$0F,$B6,$1F
	.dc.b	$05,$B3,$87,$C1,$59,$D7,$C0,$F3,$3A,$7C,$0D,$8E,$5F,$03,$DC,$F0
	.dc.b	$7C,$0D,$8E,$DF,$03,$E3,$3C,$9F,$03,$63,$C5,$F0,$3E,$7F,$81,$F4
	.dc.b	$36,$3C,$DF,$01,$EB,$F8,$1F,$63,$63,$D5,$F0,$3E,$F3,$E0,$7C,$0D
	.dc.b	$8F,$77,$C0,$FC,$33,$E2,$7C,$0D,$8F,$85,$F0,$3F,$1C,$F9,$1F,$03
	.dc.b	$63,$E3,$7C,$07,$CB,$F8,$1F,$98,$D8,$FC,$37,$EF,$FF,$E2,$71,$F0
	.dc.b	$5F,$C1,$DC,$37,$E2,$F1,$F0,$76,$B8,$BB,$F8,$3E,$1F,$07,$67,$AB
	.dc.b	$CF,$E0,$E4,$78,$3E,$0E,$F9,$FF,$E9,$2C,$DF,$07,$E1,$F8,$3F,$13
	.dc.b	$6F,$CF,$F0,$7E,$5F,$83,$F3,$36,$3D,$7F,$07,$E9,$F8,$3F,$53,$63
	.dc.b	$DF,$F0,$7E,$DF,$83,$F7,$36,$3E,$1F,$C1,$FC,$13,$E1,$7C,$1D,$8F
	.dc.b	$8F,$F0,$7F,$14,$F8,$DF,$07,$63,$E5,$FC,$1F,$C9,$3E,$57,$C1,$D8
	.dc.b	$FD,$A3,$FC,$38,$9F,$15,$FC,$D5,$C7,$C4,$5F,$84,$C3,$6B,$8F,$C2
	.dc.b	$6F,$E1,$38,$36,$3A,$FC,$27,$33,$A7,$C2,$58,$F0,$FC,$27,$73,$C1
	.dc.b	$F0,$96,$3C,$BF,$07,$E3,$F8,$3F,$23,$63,$D3,$F0,$7E,$73,$D1,$F0
	.dc.b	$76,$3D,$BF,$07,$EB,$3D,$9F,$07,$63,$E0,$FC,$17,$BC,$F8,$1F,$07
	.dc.b	$63,$E2,$FC,$17,$C3,$F8,$3F,$88,$D8,$F9,$3F,$07,$FD,$6F,$F8,$03
	.dc.b	$F8,$2F,$8F,$FD,$8F,$A3,$F8,$2B,$FE,$DC,$8C,$FC,$29,$FE,$0F,$7F
	.dc.b	$C9,$5B,$3B,$F8,$2E,$9F,$07,$D7,$E0,$FB,$7C,$1D,$C3,$83,$E1,$F8
	.dc.b	$39,$1C,$FE,$0E,$E3,$E1,$F8,$3F,$EE,$EF,$8F,$E0,$EF,$C3,$7F,$80
	.dc.b	$B5,$46,$30,$FE,$87,$0B,$C5,$FF,$7F,$87,$C1,$E3,$7F,$67,$C2,$59
	.dc.b	$36,$71,$F0,$72,$FE,$03,$BF,$84,$E1,$F0,$9C,$7E,$12,$57,$C5,$8E
	.dc.b	$9F,$09,$23,$E4,$3D,$FE,$13,$C1,$F0,$9E,$1F,$84,$95,$F3,$63,$C9
	.dc.b	$F0,$72,$3E,$83,$E7,$F8,$4F,$57,$F0,$0F,$C6,$97,$2F,$F8,$9A,$30
	.dc.b	$D8,$DF,$C2,$F0,$F8,$59,$1F,$D7,$E9,$1D,$3E,$17,$AF,$C2,$F6,$F8
	.dc.b	$59,$7B,$8F,$0F,$C2,$C8,$E4,$F9,$3E,$17,$CB,$F0,$72,$7B,$3E,$8F
	.dc.b	$84,$F4,$FC,$1F,$AB,$E1,$25,$78,$E3,$DB,$F0,$7E,$EF,$84,$93,$E7
	.dc.b	$8F,$83,$F0,$7F,$0B,$E1,$24,$FA,$E3,$E2,$FC,$27,$C6,$F8,$49,$3E
	.dc.b	$F8,$F9,$3F,$09,$F2,$BE,$12,$4F,$C3,$8F,$9B,$F0,$9F,$3B,$D0,$93
	.dc.b	$F1,$E3,$F1,$5F,$49,$FE,$58,$8F,$A3,$E5,$7F,$85,$23,$1D,$7F,$E5
	.dc.b	$86,$C6,$7E,$17,$8F,$C2,$4B,$FF,$08,$52,$3A,$FC,$2F,$6F,$84,$95
	.dc.b	$A7,$C3,$F0,$BE,$2F,$85,$F1,$FC,$2C,$AE,$71,$E6,$F8,$59,$1D,$9F
	.dc.b	$4F,$C2,$FA,$BE,$13,$D7,$F0,$B2,$BC,$71,$EE,$F8,$5F,$7F,$C2,$C9
	.dc.b	$F3,$C7,$C2,$F8,$59,$1E,$A7,$E2,$FC,$2F,$C6,$F8,$5F,$8F,$F0,$B2
	.dc.b	$BD,$F1,$F2,$BE,$12,$47,$C2,$7E,$6F,$C2,$FC,$EF,$85,$FD,$DF,$4F
	.dc.b	$F9,$EB,$E2,$AD,$9D,$3F,$E4,$08,$FF,$3F,$FC,$2E,$BE,$1A,$4E,$4F
	.dc.b	$F1,$09,$7E,$1B,$9F,$C3,$48,$C1,$BF,$86,$EE,$47,$97,$C3,$78,$BE
	.dc.b	$16,$47,$57,$CB,$F0,$DE,$6F,$83,$F3,$FC,$24,$AF,$14,$7A,$BE,$13
	.dc.b	$D7,$F0,$92,$7C,$D1,$EE,$F8,$39,$1E,$97,$E0,$FC,$27,$C2,$F8,$4F
	.dc.b	$F4,$D7,$17,$FA,$79,$B4,$F8,$4B,$18,$F8,$39,$E5,$FC,$3F,$84,$B7
	.dc.b	$F9,$A3,$F8,$58,$F8,$4E,$7F,$09,$D1,$B5,$F3,$FE,$1B,$FD,$5A,$DF
	.dc.b	$85,$F1,$C6,$2F,$A6,$9F,$EB,$19,$CF,$C2,$4B,$E5,$F0,$96,$F7,$87
	.dc.b	$FA,$CE,$7F,$D2,$2D,$8F,$E3,$EB,$E1,$3B,$FC,$27,$83,$E1,$3F,$0F
	.dc.b	$2F,$84,$A6,$23,$3F,$09,$AF,$84,$98,$FA,$E3,$8F,$C1,$C8,$F7,$3D
	.dc.b	$3E,$13,$AF,$C1,$C9,$F8,$4F,$83,$E1,$3C,$3F,$07,$E2,$F8,$49,$5F
	.dc.b	$1E,$3C,$BF,$07,$E6,$F8,$49,$3F,$2E,$3D,$3F,$07,$F2,$BE,$13,$E5
	.dc.b	$B7,$93,$E8,$7C,$1F,$CE,$3E,$7F,$C2,$58,$FD,$17,$57,$FA,$D8,$FE
	.dc.b	$2A,$E3,$E1,$3F,$0F,$19,$3F,$D6,$31,$FE,$BE,$6E,$5F,$D3,$17,$C1
	.dc.b	$70,$F8,$3E,$3F,$07,$CB,$E0,$EE,$1E,$07,$AF,$C1,$76,$F8,$3B,$3E
	.dc.b	$47,$C3,$F0,$7F,$BC,$C9,$F0,$74,$C4,$67,$E0,$A6,$BE,$C7,$87,$C1
	.dc.b	$F1,$F8,$29,$3F,$01,$E9,$F0,$7D,$7E,$0F,$DF,$F0,$7F,$01,$BA,$FC
	.dc.b	$6F,$82,$F8,$7F,$07,$F1,$1B,$1F,$2B,$E0,$FE,$39,$F2,$3E,$0E,$C7
	.dc.b	$CE,$F8,$3F,$51,$34,$FE,$25,$F9,$3F,$F1,$56,$FC,$1F,$F1,$A8,$7C
	.dc.b	$1D,$31,$19,$F8,$39,$5C,$9E,$1F,$07,$C7,$E0,$F9,$7C,$1C,$AE,$F1
	.dc.b	$D7,$E0,$E4,$78,$9F,$07,$C1,$F8,$7E,$0A,$4F,$99,$F2,$7C,$1F,$97
	.dc.b	$E0,$BF,$85,$EC,$F8,$3B,$18,$73,$F0,$73,$8B,$DF,$F0,$77,$2F,$DF
	.dc.b	$E0,$F8,$2E,$7F,$07,$D1,$B7,$F1,$BE,$0F,$E2,$7C,$1C,$C3,$07,$CA
	.dc.b	$F8,$2F,$91,$F0,$72,$30,$7F,$3F,$F8,$4F,$F0,$FE,$08,$FF,$11,$C7
	.dc.b	$FA,$D1,$D7,$C2,$7F,$4A,$8D,$C6,$BE,$13,$F9,$4C,$7F,$88,$AB,$A7
	.dc.b	$C2,$62,$73,$F0,$7A,$F8,$4B,$5E,$17,$8F,$C1,$48,$F2,$7C,$1D,$CB
	.dc.b	$F9,$0E,$5F,$EB,$69,$F4,$7C,$1F,$83,$E0,$FC,$2F,$8B,$E0,$FC,$7F
	.dc.b	$07,$6B,$DC,$F9,$BE,$0F,$CF,$F0,$76,$7E,$13,$EA,$F8,$2F,$77,$C1
	.dc.b	$CD,$3E,$34,$7C,$1F,$82,$F8,$5F,$07,$27,$E5,$47,$C5,$F8,$2F,$8D
	.dc.b	$F0,$72,$7E,$74,$7C,$9F,$83,$F8,$FF,$07,$F2,$1B,$A7,$F7,$1F,$C9
	.dc.b	$FF,$88,$0F,$84,$FE,$AA,$E1,$BF,$07,$27,$87,$C2,$5C,$BF,$A3,$DB
	.dc.b	$E1,$38,$9C,$BE,$12,$DF,$73,$AE,$3D,$9A,$78,$FE,$13,$C3,$F0,$9E
	.dc.b	$26,$C7,$9F,$E0,$BC,$BF,$07,$E6,$6C,$75,$3D,$97,$D5,$A7,$BF,$E0
	.dc.b	$FD,$A7,$BB,$E0,$EC,$7C,$3F,$83,$F8,$27,$C2,$F8,$3B,$1F,$1F,$E0
	.dc.b	$BE,$2F,$C1,$FC,$66,$C7,$CB,$F8,$2F,$92,$7C,$AF,$83,$B1,$F3,$FE
	.dc.b	$0F,$E6,$9F,$3B,$E0,$EC,$75,$3E,$A6,$1D,$5A,$7D,$7F,$83,$FF,$04
	.dc.b	$FE,$4F,$FE,$C4,$F8,$57,$3F,$FB,$11,$57,$EB,$54,$EC,$FF,$61,$BA
	.dc.b	$E8,$FF,$49,$C7,$F1,$E9,$FF,$1A,$3A,$79,$39,$8F,$BF,$AF,$C0,$6C
	.dc.b	$F9,$9F,$F0,$4C,$BE,$0F,$CE,$7A,$3E,$0E,$D7,$6F,$82,$F5,$9E,$CF
	.dc.b	$83,$B1,$E2,$F8,$3F,$79,$F0,$3E,$0E,$C7,$9B,$E0,$FE,$19,$F1,$3E
	.dc.b	$0E,$C7,$AB,$E0,$FE,$3F,$C1,$FC,$86,$C7,$52,$6D,$E8,$C3,$AB,$EA
	.dc.b	$F8,$3F,$9F,$F0,$7F,$41,$B5,$F1,$BE,$07,$E9,$FC,$17,$D4,$6C,$74
	.dc.b	$E4,$F9,$C7,$D2,$DB,$D5,$19,$3E,$6F,$C1,$7D,$CF,$82,$FB,$AD,$8F
	.dc.b	$A7,$F0,$5F,$80,$F8,$2F,$C0,$B6,$3F,$CD,$5F,$93,$FF,$8E,$71,$EE
	.dc.b	$D8,$FC,$12,$F8,$3F,$F2,$8B,$9D,$FF,$DA,$6F,$F2,$B7,$FC,$82,$DA
	.dc.b	$E7,$F0,$7C,$3E,$0F,$8F,$C1,$F2,$F8,$3B,$97,$89,$EB,$8C,$8F,$2E
	.dc.b	$56,$3A,$FC,$1F,$88,$F1,$FC,$1D,$AF,$5F,$C1,$79,$BE,$0B,$CE,$D8
	.dc.b	$F7,$DF,$3D,$7D,$F8,$D9,$F7,$FC,$17,$BB,$E0,$BD,$ED,$8F,$8F,$F0
	.dc.b	$5F,$08,$F8,$7F,$05,$63,$E5,$FC,$0F,$C6,$3E,$3F,$C1,$58,$F9,$FF
	.dc.b	$03,$F2,$8F,$97,$F0,$56,$3E,$9F,$C0,$FC,$E3,$E7,$FC,$15,$8E,$A4
	.dc.b	$DF,$E7,$E1,$D5,$FA,$3F,$03,$F5,$8F,$AF,$F0,$56,$BF,$D5,$07,$F9
	.dc.b	$1C,$9E,$D7,$FA,$31,$FF,$25,$75,$B5,$E9,$C7,$F8,$4C,$AD,$7F,$05
	.dc.b	$3F,$C9,$31,$FC,$23,$E0,$B9,$7C,$17,$37,$A7,$C1,$75,$F8,$2B,$7E
	.dc.b	$27,$C1,$F0,$3E,$1F,$82,$B3,$E6,$7C,$9F,$03,$23,$D3,$F0,$57,$8B
	.dc.b	$FD,$1F,$0F,$82,$F5,$7C,$17,$AD,$B1,$F0,$31,$EC,$D3,$B1,$F1,$3E
	.dc.b	$0B,$E1,$1F,$0F,$E0,$AC,$7C,$8F,$6F,$E3,$7B,$DF,$1D,$B1,$CF,$2F
	.dc.b	$1E,$BE,$A6,$C7,$D0,$F8,$2F,$9C,$7C,$FF,$82,$B1,$F5,$3E,$0B,$E9
	.dc.b	$1F,$4F,$E0,$AC,$7D,$8F,$82,$FF,$60,$FE,$4F,$FC,$E6,$3F,$9C,$1F
	.dc.b	$CD,$DF,$E7,$87,$F3,$1C,$5D,$08,$C3,$7E,$1F,$9F,$C3,$F4,$F8,$39
	.dc.b	$87,$FB,$B6,$DF,$07,$8C,$7E,$7E,$5E,$2F,$86,$FC,$43,$6F,$87,$C2
	.dc.b	$79,$8F,$3F,$C2,$58,$E9,$ED,$FA,$BE,$03,$D6,$D8,$F7,$1E,$1C,$3A
	.dc.b	$B4,$F2,$7C,$27,$C2,$3E,$1F,$C2,$58,$F4,$7C,$1F,$C6,$3E,$3F,$C2
	.dc.b	$58,$F6,$7C,$1F,$CA,$F8,$4F,$96,$D8,$F8,$1F,$07,$F3,$8F,$9F,$F0
	.dc.b	$96,$3E,$27,$C1,$FD,$2F,$84,$FA,$6D,$8E,$A7,$C9,$C3,$AB,$4F,$99
	.dc.b	$F0,$7F,$6B,$E1,$3E,$DB,$63,$F5,$9C,$BF,$63,$F2,$7F,$E5,$DD,$86
	.dc.b	$9F,$CC,$E7,$71,$99,$EE,$7F,$BC,$BA,$58,$E2,$7F,$A5,$C6,$2B,$8D
	.dc.b	$FF,$37,$7F,$C0,$0B,$E1,$7F,$A8,$C6,$3E,$16,$E3,$E5,$9B,$87,$97
	.dc.b	$E1,$6E,$1F,$EF,$98,$F4,$7C,$2E,$27,$3F,$0B,$AF,$85,$B9,$F8,$9E
	.dc.b	$3F,$09,$23,$CB,$F0,$B7,$4F,$CD,$C7,$9B,$E0,$FB,$9E,$0F,$84,$B1
	.dc.b	$EC,$C7,$B3,$4E,$C7,$C0,$F8,$4F,$39,$E8,$F8,$4B,$1F,$13,$DC,$F5
	.dc.b	$FB,$9E,$C6,$C7,$3C,$BC,$67,$D7,$D7,$D6,$E3,$83,$EC,$47,$C4,$F8
	.dc.b	$4F,$8B,$F0,$96,$FE,$8B,$F2,$3E,$13,$FD,$39,$87,$1E,$47,$C2,$63
	.dc.b	$2F,$F0,$1C,$62,$7E,$CD,$67,$0F,$B3,$F0,$96,$7F,$14,$DB,$FC,$53
	.dc.b	$6D,$F0,$99,$FC,$9F,$FC,$79,$7E,$97,$FC,$F8,$FE,$39,$AE,$70,$E1
	.dc.b	$F0,$52,$34,$F3,$F8,$3E,$D8,$C9,$FC,$7B,$E1,$C3,$C1,$F0,$72,$7B
	.dc.b	$3E,$3F,$83,$F2,$7B,$FE,$5F,$6E,$57,$8E,$3A,$3F,$1F,$0B,$B7,$C6
	.dc.b	$C7,$D9,$F0,$72,$BD,$4F,$BF,$E0,$FE,$07,$C1,$7C,$1F,$83,$95,$EF
	.dc.b	$8F,$89,$F0,$7F,$17,$E0,$E4,$FC,$38,$F9,$1F,$05,$F2,$7E,$0E,$4F
	.dc.b	$C7,$8F,$99,$F0,$7F,$37,$E0,$E4,$FC,$B8,$EA,$FC,$DC,$3A,$BF,$47
	.dc.b	$0F,$A9,$F0,$7F,$57,$E0,$E4,$FD,$38,$FB,$1E,$FC,$8F,$AC,$FD,$FF
	.dc.b	$86,$7B,$9D,$A7,$EE,$63,$DB,$8F,$FC,$9B,$D8,$CF,$87,$C1,$F1,$F8
	.dc.b	$49,$87,$F0,$EC,$93,$D7,$E1,$3B,$7C,$24,$AD,$47,$87,$E1,$24,$71
	.dc.b	$7C,$9F,$09,$E5,$F8,$39,$3D,$5F,$47,$C2,$7A,$7E,$12,$4F,$85,$F6
	.dc.b	$7C,$27,$48,$FD,$76,$8C,$71,$C1,$FE,$68,$86,$5F,$0F,$E1,$25,$FA
	.dc.b	$DF,$8D,$F0,$9F,$1F,$E1,$24,$FB,$DF,$95,$EF,$74,$F2,$BF,$D4,$98
	.dc.b	$F7,$AF,$61,$F1,$71,$FA,$1F,$09,$F4,$7E,$12,$4F,$C9,$8F,$A9,$F0
	.dc.b	$7F,$57,$E1,$24,$FC,$D8,$FB,$1F,$07,$F6,$7E,$12,$4F,$D1,$8F,$DD
	.dc.b	$7E,$4F,$FC,$1A,$3F,$87,$37,$87,$FD,$0E,$DD,$7F,$D7,$4F,$F9,$E6
	.dc.b	$7F,$A3,$AF,$84,$E6,$48,$CF,$C2,$78,$9F,$C0,$63,$E2,$7E,$7E,$3E
	.dc.b	$2F,$83,$F1,$FC,$24,$9E,$B1,$E6,$F8,$0F,$3F,$B9,$27,$C3,$1D,$5F
	.dc.b	$B3,$87,$5D,$7E,$C6,$3E,$EF,$84,$F7,$FC,$24,$BF,$4C,$7C,$2F,$84
	.dc.b	$F8,$7F,$09,$27,$DB,$1F,$1B,$E1,$3E,$3F,$C2,$49,$F8,$31,$F2,$BE
	.dc.b	$0F,$E5,$FC,$24,$9F,$8B,$1F,$3B,$E1,$3E,$7F,$C2,$49,$F9,$31,$D5
	.dc.b	$F9,$78,$75,$7E,$9E,$1F,$5B,$E0,$FE,$BF,$C2,$49,$FA,$31,$F6,$BD
	.dc.b	$D9,$1F,$51,$FB,$BE,$EF,$58,$FB,$B8,$75,$E2,$FB,$B8,$FF,$3B,$FC
	.dc.b	$9F,$F9,$BC,$7F,$B1,$9B,$D9,$FE,$F9,$6D,$FF,$90,$9F,$E7,$33,$BF
	.dc.b	$84,$FE,$F3,$1F,$E7,$A1,$F0,$98,$8C,$FC,$24,$C3,$FC,$1F,$4F,$84
	.dc.b	$C6,$3F,$D7,$E1,$F0,$7D,$0E,$BF,$09,$6F,$8F,$C2,$78,$0F,$0F,$C2
	.dc.b	$58,$EB,$8F,$4D,$7D,$8D,$8F,$67,$C1,$FA,$3E,$13,$D2,$D8,$F2,$FB
	.dc.b	$5E,$C3,$DB,$EE,$58,$EA,$7A,$B7,$F8,$38,$75,$DF,$E3,$7C,$1F,$C4
	.dc.b	$F8,$4F,$8A,$DB,$F8,$3F,$07,$F2,$0F,$93,$F0,$96,$3E,$2F,$C1,$FC
	.dc.b	$C3,$E6,$FC,$25,$8F,$93,$F0,$9F,$40,$FF,$14,$FE,$4F,$FC,$DA,$32
	.dc.b	$7F,$8F,$1F,$F3,$B9,$3E,$0F,$7F,$07,$CF,$5F,$A1,$8F,$6B,$F0,$9C
	.dc.b	$7E,$0B,$A9,$DB,$E0,$EC,$75,$F7,$FC,$27,$8B,$DA,$B1,$CB,$0E,$AF
	.dc.b	$E0,$35,$F0,$74,$21,$3F,$E8,$ED,$5F,$07,$EC,$F8,$3F,$6B,$6B,$D9
	.dc.b	$F0,$7F,$00,$F8,$3F,$07,$63,$E0,$7C,$1F,$C4,$3E,$2F,$C1,$D8,$F8
	.dc.b	$9F,$07,$F2,$3E,$0F,$E4,$B6,$3E,$47,$C0,$FC,$CF,$82,$F9,$AD,$8E
	.dc.b	$A4,$DB,$FC,$6D,$B3,$1E,$D1,$8F,$81,$FA,$BF,$05,$F5,$9B,$7F,$57
	.dc.b	$DA,$FB,$3E,$DF,$DA,$6C,$75,$E3,$F8,$D8,$7C,$5D,$FD,$73,$93,$E4
	.dc.b	$CF,$F6,$C9,$FF,$35,$F5,$7F,$A0,$0F,$F3,$F3,$FE,$4B,$71,$F0,$3A
	.dc.b	$37,$F0,$36,$FF,$D0,$73,$FE,$71,$1F,$03,$D3,$E0,$7A,$BD,$BE,$03
	.dc.b	$BF,$C0,$DB,$E2,$F8,$BE,$03,$C7,$F0,$36,$7A,$BC,$DE,$F7,$74,$F3
	.dc.b	$3E,$AF,$80,$93,$E4,$F8,$1B,$5E,$BF,$63,$DF,$EC,$FC,$06,$D7,$AB
	.dc.b	$5F,$3B,$4E,$E7,$BB,$E0,$3E,$39,$F2,$3E,$06,$C7,$C2,$F8,$0F,$F3
	.dc.b	$07,$0B,$E0,$69,$88,$CF,$C0,$4D,$3E,$4B,$C3,$E0,$78,$FC,$0F,$2F
	.dc.b	$81,$95,$F3,$A3,$FC,$D9,$F9,$3F,$E8,$F3,$FF,$59,$B6,$7F,$D0,$4B
	.dc.b	$E2,$BF,$D4,$EE,$7A,$FF,$0A,$17,$C1,$68,$DF,$C1,$DC,$3A,$FB,$5C
	.dc.b	$8E,$7E,$D5,$8F,$0F,$C1,$76,$F8,$2E,$ED,$8F,$2F,$C1,$78,$8F,$1F
	.dc.b	$C1,$58,$F4,$FC,$0F,$98,$F3,$FC,$15,$8F,$6F,$C1,$7C,$02,$70,$FC
	.dc.b	$0F,$82,$F8,$7F,$03,$F1,$3E,$0A,$4F,$C4,$8F,$8F,$F0,$5F,$23,$E0
	.dc.b	$A4,$FC,$88,$F9,$7F,$01,$F1,$BD,$CF,$8E,$DC,$FE,$6F,$C1,$7C,$A3
	.dc.b	$FD,$A3,$F9,$3F,$F2,$9C,$3E,$7C,$63,$C1,$95,$FE,$E6,$A7,$C1,$E2
	.dc.b	$FF,$D6,$D0,$F8,$3E,$7F,$07,$D1,$B7,$C7,$E0,$FB,$9E,$0F,$83,$B1
	.dc.b	$D7,$E0,$BC,$67,$93,$E0,$EC,$78,$7E,$0F,$FD,$65,$93,$E0,$E9,$88
	.dc.b	$CF,$B3,$AF,$7A,$6D,$E9,$8E,$3F,$05,$CB,$E0,$E4,$FB,$63,$AF,$C1
	.dc.b	$48,$F8,$0F,$83,$E0,$FA,$C7,$C3,$C3,$AB,$F0,$B0,$F2,$FC,$0F,$9B
	.dc.b	$E0,$A4,$FC,$98,$F4,$FC,$0F,$AB,$E0,$A4,$FC,$D8,$EA,$FC,$CC,$3A
	.dc.b	$F9,$BF,$9E,$E0,$65,$FE,$D3,$FC,$9F,$FC,$AD,$6A,$47,$F9,$5E,$3F
	.dc.b	$DD,$4D,$8F,$D3,$BC,$1D,$FC,$1F,$3F,$83,$95,$FC,$B3,$93,$F9,$6D
	.dc.b	$FF,$97,$E3,$07,$75,$ED,$78,$C9,$1A,$C3,$C7,$EF,$F9,$FE,$0F,$D1
	.dc.b	$F0,$72,$B5,$82,$7D,$9F,$07,$ED,$F8,$39,$5A,$8E,$AF,$B3,$2B,$97
	.dc.b	$B3,$2F,$89,$F0,$72,$BA,$BF,$1F,$E0,$FE,$47,$C1,$49,$F0,$BF,$2F
	.dc.b	$E0,$F9,$EF,$F0,$32,$B5,$EF,$C7,$E8,$7C,$1C,$8F,$4B,$F4,$FE,$0F
	.dc.b	$EA,$7C,$17,$D5,$F8,$39,$5E,$E8,$FB,$1F,$07,$F6,$7E,$0E,$4F,$C2
	.dc.b	$8F,$B9,$F0,$5F,$77,$E0,$E4,$FC,$68,$FF,$03,$BF,$7F,$5F,$EC,$3F
	.dc.b	$07,$FE,$56,$E3,$FE,$4D,$1A,$7F,$92,$C6,$4F,$F2,$CC,$7D,$CC,$3C
	.dc.b	$35,$F8,$6C,$7C,$AF,$D9,$AF,$E5,$6D,$F8,$2E,$97,$E0,$6C,$7E,$46
	.dc.b	$7F,$CC,$2F,$1F,$81,$CE,$3F,$CE,$7E,$06,$D7,$F9,$86,$B4,$7C,$13
	.dc.b	$F0,$78,$F9,$2B,$F2,$F9,$70,$F8,$09,$5E,$57,$9F,$C0,$F4,$F8,$1E
	.dc.b	$BF,$03,$2B,$D5,$1E,$0F,$81,$F0,$FC,$0C,$9F,$74,$79,$3E,$06,$47
	.dc.b	$C1,$7C,$FF,$03,$E8,$F8,$09,$3F,$15,$F5,$FC,$0F,$58,$F9,$18,$75
	.dc.b	$7E,$B1,$F1,$70,$F8,$3F,$01,$F0,$BE,$06,$4F,$CF,$8F,$8B,$F0,$12
	.dc.b	$3E,$93,$F2,$3E,$07,$B4,$74,$E2,$FE,$C9,$8F,$77,$EF,$E3,$F3,$BE
	.dc.b	$06,$57,$DB,$7E,$9F,$E4,$EF,$8A,$BF,$07,$2B,$F7,$13,$FE,$2B,$7B
	.dc.b	$3F,$D0,$1C,$FC,$0F,$2F,$82,$91,$83,$8F,$C1,$76,$23,$D3,$E0,$BC
	.dc.b	$3F,$05,$E2,$F8,$29,$3E,$08,$F2,$E1,$D2,$3C,$38,$61,$F2,$FC,$0F
	.dc.b	$A8,$8F,$A3,$E0,$BD,$BE,$C7,$BB,$DA,$93,$EC,$8E,$4F,$43,$DF,$78
	.dc.b	$E2,$FF,$05,$EC,$F8,$1F,$8F,$F0,$52,$FE,$2C,$7C,$AF,$82,$91,$F2
	.dc.b	$1F,$9B,$F0,$5F,$3B,$E0,$7E,$7F,$C1,$4A,$F9,$B1,$F4,$BE,$06,$47
	.dc.b	$D0,$7E,$AF,$C1,$7E,$A3,$F2,$7F,$E6,$C4,$9C,$7C,$57,$57,$EB,$E1
	.dc.b	$D5,$FA,$F8,$71,$F8,$39,$1F,$CC,$EF,$F9,$D3,$F8,$95,$F0,$7D,$FD
	.dc.b	$A9,$3B,$7C,$5E,$D7,$78,$FC,$1E,$3D,$F8,$BE,$3E,$5E,$7F,$81,$F4
	.dc.b	$7C,$1C,$BF,$04,$7A,$FE,$07,$D9,$F0,$52,$7C,$91,$EF,$F8,$29,$1E
	.dc.b	$77,$E1,$7C,$17,$C3,$F8,$2F,$89,$F0,$52,$BD,$91,$F1,$FE,$0B,$E4
	.dc.b	$7C,$14,$9F,$81,$1D,$1F,$81,$85,$DB,$FC,$93,$C6,$CB,$E8,$7C,$0F
	.dc.b	$D1,$F8,$29,$87,$C9,$8F,$A9,$ED,$48,$F9,$8F,$D7,$F6,$FA,$7C,$07
	.dc.b	$FA,$53,$16,$78,$BF,$EE,$D1,$DD,$FA,$9F,$03,$F7,$FE,$0A,$46,$0F
	.dc.b	$ED,$5F,$93,$FF,$97,$F0,$FF,$4F,$B6,$3F,$C0,$0F,$F4,$07,$F8,$23
	.dc.b	$9D,$8F,$F9,$92,$79,$FC,$1F,$F9,$A1,$E7,$F0,$F6,$F9,$CD,$BE,$3F
	.dc.b	$09,$70,$7C,$25,$8C,$39,$F8,$3D,$7C,$1D,$CF,$D2,$F1,$C6,$47,$B3
	.dc.b	$3B,$9F,$A3,$E0,$BB,$1D,$FE,$0E,$D7,$C2,$F6,$BC,$5E,$DF,$8D,$B1
	.dc.b	$EC,$25,$FC,$2C,$3A,$EF,$FE,$86,$DD,$F0,$7E,$BF,$83,$F6,$37,$0F
	.dc.b	$9F,$F0,$7E,$FF,$83,$F8,$0D,$8F,$A7,$F0,$5F,$0F,$E0,$FE,$23,$63
	.dc.b	$EB,$FC,$1F,$C7,$3E,$47,$C1,$D8,$FB,$7F,$07,$F2,$CF,$9D,$E7,$FF
	.dc.b	$9C,$76,$F3,$47,$49,$FC,$0E,$37,$83,$EE,$BC,$3E,$0A,$4F,$F8,$BF
	.dc.b	$0F,$F3,$AB,$63,$AF,$BD,$DB,$DA,$B3,$B7,$A6,$5F,$17,$0B,$3F,$8E
	.dc.b	$7C,$BF,$05,$23,$B7,$C1,$5C,$FF,$A9,$66,$F8,$1F,$5F,$C1,$7B,$1B
	.dc.b	$7E,$5F,$82,$F7,$FC,$17,$C0,$6C,$7A,$7E,$07,$E1,$FC,$17,$C4,$6C
	.dc.b	$7B,$7E,$03,$E3,$9F,$23,$E0,$6C,$75,$3E,$16,$1D,$5A,$7C,$5F,$81
	.dc.b	$F9,$E7,$D0,$F8,$1B,$1F,$27,$DA,$FA,$67,$D4,$F6,$EC,$75,$E0,$F7
	.dc.b	$E1,$D5,$B1,$F4,$7E,$07,$ED,$9F,$73,$E0,$6C,$7F,$98,$FF,$25,$57
	.dc.b	$EB,$5B,$19,$F8,$2C,$4E,$79,$35,$F0,$56,$B9,$3C,$7E,$0A,$47,$5F
	.dc.b	$82,$B7,$FE,$05,$C7,$FC,$D9,$3E,$1F,$81,$F0,$7C,$17,$85,$E9,$5B
	.dc.b	$C6,$F6,$FF,$BB,$30,$FE,$8B,$87,$A7,$E0,$7D,$1F,$05,$E9,$6C,$7B
	.dc.b	$7D,$9F,$61,$ED,$F7,$AC,$7C,$1D,$FD,$6D,$3D,$67,$C5,$F8,$1F,$88
	.dc.b	$7C,$5F,$82,$B1,$F2,$7E,$07,$E4,$1F,$27,$E0,$AC,$7C,$DF,$81,$F9
	.dc.b	$87,$CD,$F8,$2B,$1F,$47,$E0,$BE,$87,$C1,$7D,$16,$C7,$D5,$F8,$1F
	.dc.b	$A8,$7D,$5F,$81,$B1,$D4,$9C,$BF,$37,$0E,$AF,$F8,$CF,$F2,$5F,$E6
	.dc.b	$07,$F3,$3E,$55,$7E,$01,$46,$BE,$1B,$F9,$73,$FC,$D6,$B9,$7B,$F8
	.dc.b	$F6,$F2,$F6,$BF,$A9,$B7,$B7,$1F,$1C,$76,$BF,$F6,$16,$0F,$83,$EB
	.dc.b	$F0,$7D,$9B,$7F,$EB,$59,$7C,$1F,$88,$F1,$FC,$1D,$9D,$7C,$1F,$98
	.dc.b	$F3,$FC,$1D,$8E,$5F,$05,$EA,$F8,$3F,$5B,$63,$B7,$C0,$FB,$BE,$0B
	.dc.b	$DE,$D8,$EA,$4D,$BE,$56,$1D,$5F,$99,$F0,$3F,$1B,$E0,$BE,$3B,$6B
	.dc.b	$D5,$F0,$5F,$2B,$DC,$F9,$6D,$8E,$9C,$7F,$14,$FA,$3B,$7A,$A3,$27
	.dc.b	$D5,$E1,$7C,$17,$D4,$3E,$AF,$C1,$59,$F9,$1F,$05,$F6,$0F,$EC,$DF
	.dc.b	$92,$7F,$39,$8E,$51,$8F,$83,$CF,$C5,$59,$FF,$72,$57,$F3,$6E,$0F
	.dc.b	$F3,$D2,$F8,$4E,$67,$4F,$84,$B5,$C3,$E1,$3B,$FC,$27,$81,$B1,$D3
	.dc.b	$2E,$ED,$3B,$E5,$E6,$F8,$39,$1E,$2F,$84,$BA,$78,$BE,$13,$D6,$7B
	.dc.b	$3D,$CB,$5C,$E7,$CF,$87,$AD,$B1,$EC,$F8,$3F,$87,$F0,$9F,$11,$B1
	.dc.b	$F0,$3E,$13,$E3,$FC,$27,$C8,$6C,$7C,$4F,$84,$F9,$7F,$09,$F3,$1B
	.dc.b	$1F,$23,$E1,$3E,$7F,$C2,$7D,$06,$C7,$CC,$F8,$4F,$A7,$F0,$9F,$51
	.dc.b	$B1,$D4,$FA,$38,$75,$69,$FB,$53,$F6,$DB,$7F,$34,$DB,$FD,$5C,$DE
	.dc.b	$EF,$C5,$2F,$84,$FF,$13,$39,$AE,$3E,$D6,$0C,$FC,$05,$8F,$16,$1F
	.dc.b	$80,$C7,$C4,$D8,$F1,$2C,$58,$3E,$13,$B9,$E0,$F8,$4B,$7C,$2B,$7F
	.dc.b	$07,$E4,$F8,$4F,$2B,$67,$58,$3E,$0B,$D2,$7A,$BE,$0E,$CE,$BE,$0F
	.dc.b	$DB,$F0,$7E,$E6,$C7,$2F,$83,$F8,$3F,$07,$F0,$9B,$1D,$B0,$E7,$BF
	.dc.b	$CA,$C6,$EB,$F2,$BE,$0F,$E4,$9F,$2B,$E0,$EE,$1E,$6F,$6F,$E6,$9F
	.dc.b	$3B,$DF,$B1,$D4,$98,$F5,$6E,$1F,$5B,$E0,$BE,$AF,$C1,$FD,$66,$CF
	.dc.b	$C2,$F8,$3F,$B2,$7D,$AF,$83,$B1,$F2,$FF,$23,$FF,$35,$13,$E1,$3F
	.dc.b	$CC,$8E,$3E,$0E,$FC,$26,$1B,$5C,$7E,$0F,$7F,$09,$C1,$B1,$D4,$E6
	.dc.b	$74,$C2,$BE,$1F,$84,$EF,$F0,$9E,$06,$C7,$97,$DB,$F1,$FC,$07,$91
	.dc.b	$B1,$D4,$9C,$5F,$8B,$7F,$A9,$6B,$FD,$55,$A7,$B3,$E0,$E3,$F0,$7E
	.dc.b	$12,$D7,$13,$F2,$95,$FE,$6A,$9F,$8D,$F0,$9F,$13,$E1,$3E,$2B,$F1
	.dc.b	$BE,$0F,$E3,$FC,$25,$AF,$98,$FC,$AF,$83,$F9,$7F,$09,$67,$E8,$3F
	.dc.b	$3B,$E0,$FE,$7F,$C1,$D9,$FA,$8F,$3D,$3C,$9D,$8D,$59,$66,$3F,$A0
	.dc.b	$46,$2F,$EE,$7C,$1F,$FA,$77,$F2,$3F,$E7,$18,$7B,$23,$1F,$15,$9F
	.dc.b	$80,$D7,$97,$70,$FD,$76,$1F,$CE,$72,$EB,$C1,$EE,$9F,$C4,$37,$2F
	.dc.b	$2D,$7F,$3D,$9E,$3F,$07,$E1,$F8,$4F,$13,$E3,$F8,$3F,$F2,$2D,$F0
	.dc.b	$6E,$3F,$D6,$70,$46,$30,$FF,$06,$43,$E1,$31,$7F,$C2,$63,$FC,$E8
	.dc.b	$FA,$3E,$13,$38,$27,$33,$E6,$F8,$4C,$C3,$E1,$29,$27,$E0,$FC,$26
	.dc.b	$BE,$0F,$7F,$09,$27,$E2,$C7,$2F,$83,$91,$F2,$1E,$BF,$09,$DB,$E1
	.dc.b	$3B,$FB,$B2,$BE,$6C,$78,$BE,$0F,$C7,$F0,$92,$7E,$8C,$79,$BE,$0E
	.dc.b	$47,$D4,$7D,$3F,$09,$EE,$FC,$8B,$E5,$CF,$E1,$9F,$E7,$0F,$F2,$76
	.dc.b	$B9,$38,$7C,$27,$1F,$84,$91,$B8,$E9,$F0,$92,$39,$3D,$FE,$13,$C1
	.dc.b	$F0,$72,$7B,$3E,$3F,$84,$F2,$7C,$04,$9F,$13,$E7,$F8,$0F,$47,$C1
	.dc.b	$FA,$7E,$12,$57,$9E,$3D,$9F,$07,$ED,$F8,$49,$3E,$B8,$F8,$1F,$09
	.dc.b	$F0,$7E,$12,$4F,$BE,$3A,$BF,$07,$0E,$AF,$C5,$C3,$E4,$7C,$24,$8F
	.dc.b	$8C,$FC,$BF,$84,$F9,$9F,$07,$FE,$2B,$F1,$BF,$CB,$9C,$A8,$C5,$FF
	.dc.b	$59,$87,$C2,$63,$C5,$46,$EF,$E9,$7C,$26,$61,$F0,$96,$30,$F1,$F8
	.dc.b	$3F,$F3,$5D,$7F,$5F,$7F,$8D,$63,$FD,$9D,$CB,$FC,$11,$CE,$06,$47
	.dc.b	$1F,$84,$FD,$1C,$1A,$E3,$E0,$E4,$F7,$77,$F0,$9C,$3E,$0A,$4F,$8D
	.dc.b	$E7,$F0,$7D,$3E,$0B,$D0,$7A,$7E,$0E,$E9,$E9,$F8,$3F,$67,$C1,$FB
	.dc.b	$5B,$1E,$DF,$83,$F8,$1F,$07,$F0,$5B,$1F,$07,$D9,$F8,$9E,$FF,$C5
	.dc.b	$6C,$75,$3E,$36,$1D,$5A,$7C,$9F,$83,$F9,$9F,$07,$F3,$5B,$1F,$37
	.dc.b	$E0,$FE,$81,$F4,$7E,$0E,$C7,$5E,$1F,$9D,$87,$56,$C7,$D5,$F8,$2F
	.dc.b	$C0,$7E,$45,$FC,$BB,$A3,$E9,$46,$3E,$0F,$3F,$07,$AF,$83,$B8,$7F
	.dc.b	$86,$F0,$FE,$5C,$FF,$AA,$5F,$E5,$58,$FF,$40,$6E,$3F,$E7,$D5,$F0
	.dc.b	$5E,$0F,$83,$F0,$FC,$1F,$8B,$E0,$EE,$3D,$5F,$2F,$C1,$C8,$F0,$7C
	.dc.b	$1D,$CF,$B6,$1C,$E7,$C9,$8D,$9F,$16,$1E,$EF,$82,$ED,$5E,$97,$E0
	.dc.b	$FC,$17,$C2,$F8,$1B,$3E,$D7,$96,$DE,$9C,$BE,$3E,$58,$AF,$81,$87
	.dc.b	$57,$D0,$75,$3E,$37,$C1,$7C,$EF,$82,$F9,$ED,$8F,$95,$F0,$5F,$48
	.dc.b	$FA,$7F,$03,$63,$E7,$7C,$17,$D6,$F8,$1F,$AE,$D8,$FA,$5F,$05,$F6
	.dc.b	$8F,$B7,$F0,$56,$3A,$93,$8B,$EA,$61,$D5,$FA,$DF,$05,$FC,$07,$F2
	.dc.b	$2F,$F3,$34,$7F,$97,$7A,$13,$98,$D7,$C3,$EF,$E0,$EE,$3F,$E5,$DC
	.dc.b	$11,$D7,$93,$F0,$1B,$FB,$2B,$C5,$1F,$EB,$FA,$7C,$27,$84,$F1,$7C
	.dc.b	$25,$AE,$9F,$09,$E5,$F8,$4F,$33,$63,$C1,$F0,$9E,$9F,$84,$F5,$36
	.dc.b	$3C,$9F,$09,$ED,$3D,$DE,$FD,$8F,$47,$C0,$7C,$1F,$83,$F8,$4D,$8E
	.dc.b	$A7,$B7,$0E,$AD,$3E,$07,$C2,$7C,$9F,$83,$F9,$4D,$8F,$89,$F0,$9F
	.dc.b	$37,$E1,$3E,$73,$63,$AF,$37,$F8,$9B,$8B,$FD,$01,$18,$E3,$F8,$D9
	.dc.b	$7D,$5F,$84,$93,$F3,$FE,$12,$E1,$C7,$E1,$3E,$D1,$F6,$FE,$0E,$D7
	.dc.b	$D5,$F8,$3F,$E7,$DF,$91,$7F,$32,$CF,$E9,$46,$27,$3F,$11,$AF,$84
	.dc.b	$B1,$FE,$99,$BF,$E6,$CD,$F8,$5E,$7F,$0B,$D1,$B1,$DC,$E0,$7E,$37
	.dc.b	$0E,$CF,$3F,$85,$F1,$9E,$4F,$85,$B1,$DF,$E1,$7C,$FF,$0B,$E8,$6C
	.dc.b	$75,$E0,$FF,$26,$E7,$EB,$84,$7F,$AD,$79,$3F,$AB,$C7,$88,$F5,$7C
	.dc.b	$2F,$C2,$3E,$1F,$C2,$D8,$F7,$7B,$BF,$18,$F8,$FF,$09,$63,$E1,$7C
	.dc.b	$2F,$CA,$F8,$4F,$96,$D8,$F8,$DF,$0B,$F3,$8F,$9F,$F0,$96,$3E,$57
	.dc.b	$C2,$FD,$23,$E9,$FC,$2D,$8E,$A7,$CF,$C3,$AB,$4F,$A5,$F0,$BF,$68
	.dc.b	$FB,$7F,$0B,$63,$FD,$9D,$E6,$FF,$3B,$27,$C2,$7F,$7D,$7F,$9F,$71
	.dc.b	$B5,$F4,$71,$FC,$A6,$F6,$B0,$F9,$6F,$ED,$BE,$13,$97,$BD,$CD,$B3
	.dc.b	$E0,$F7,$BB,$7C,$27,$76,$C7,$93,$E0,$BC,$47,$8F,$E1,$2C,$7A,$3E
	.dc.b	$13,$CC,$79,$FE,$0A,$C7,$B3,$E1,$3D,$5F,$09,$EB,$6C,$75,$26,$FF
	.dc.b	$E9,$0E,$16,$3D,$A3,$1F,$07,$F0,$CF,$89,$F0,$96,$FE,$4F,$B7,$F1
	.dc.b	$CF,$91,$F0,$16,$3A,$F1,$FC,$5C,$0F,$F1,$AD,$FD,$93,$A3,$E4,$E6
	.dc.b	$F8,$3F,$A4,$7D,$3F,$84,$B3,$F6,$3E,$13,$EB,$7C,$27,$D7,$6C,$7D
	.dc.b	$CF,$83,$FF,$1B,$7E,$41,$FC,$DB,$3F,$E5,$31,$8D,$33,$F0,$9A,$F8
	.dc.b	$4B,$87,$DA,$3E,$DE,$1F,$CD,$9C,$57,$AA,$3F,$99,$C6,$30,$46,$27
	.dc.b	$FD,$76,$BE,$13,$C7,$B7,$E3,$72,$F1,$61,$E7,$39,$57,$9C,$7C,$1F
	.dc.b	$AB,$E1,$3D,$6D,$9D,$FC,$27,$B8,$F7,$FB,$96,$3A,$67,$E2,$3C,$0F
	.dc.b	$67,$E2,$FC,$24,$8F,$0F,$C2,$59,$F4,$FC,$1F,$FA,$BB,$33,$4C,$7C
	.dc.b	$26,$7E,$0F,$5F,$09,$2F,$D7,$1C,$7E,$0F,$97,$C2,$49,$F7,$C7,$5F
	.dc.b	$84,$91,$F0,$9F,$07,$C2,$75,$8F,$8B,$87,$57,$E6,$E1,$E5,$F8,$4F
	.dc.b	$37,$C2,$49,$F9,$71,$E9,$F8,$09,$1F,$39,$FF,$64,$FE,$41,$F9,$58
	.dc.b	$E6,$76,$21,$83,$FD,$EA,$BE,$17,$89,$23,$F0,$51,$FE,$74,$77,$F0
	.dc.b	$7F,$E8,$78,$FF,$3A,$71,$EB,$E1,$31,$39,$F8,$49,$3C,$BE,$12,$E1
	.dc.b	$E0,$F8,$3E,$47,$3F,$84,$B5,$E0,$F8,$3E,$C7,$7F,$83,$B1,$E4,$C7
	.dc.b	$9F,$87,$FB,$31,$AB,$3C,$D7,$F1,$A8,$7C,$1F,$A4,$F5,$7C,$1D,$9F
	.dc.b	$7F,$B5,$ED,$3D,$DE,$FD,$8F,$01,$37,$F9,$38,$F6,$7F,$81,$E2,$F8
	.dc.b	$3F,$8C,$7C,$7F,$83,$B7,$F3,$3E,$0F,$E5,$7C,$1F,$CB,$6C,$7D,$0F
	.dc.b	$83,$F9,$DF,$07,$F3,$DB,$1F,$B5,$FC,$83,$FC,$F0,$3E,$13,$FD,$74
	.dc.b	$E1,$BF,$05,$8F,$83,$B5,$C5,$E8,$FD,$EC,$2E,$DF,$84,$79,$FC,$14
	.dc.b	$9F,$07,$C1,$DA,$FE,$43,$1F,$D1,$B2,$7B,$5E,$23,$C7,$EF,$D9,$F6
	.dc.b	$1D,$CB,$5D,$DF,$67,$C1,$FA,$8F,$5F,$C1,$D8,$F8,$1F,$05,$EE,$F8
	.dc.b	$3F,$7B,$63,$E2,$7C,$1F,$C2,$3E,$1F,$C1,$D8,$F9,$1F,$07,$F1,$8F
	.dc.b	$8F,$F0,$76,$3E,$67,$C1,$FC,$AF,$83,$F9,$6D,$8E,$A4,$D7,$E0,$9F
	.dc.b	$D0,$B8,$18,$F7,$8C,$F8,$49,$7C,$17,$D6,$F8,$3F,$AE,$DC,$3E,$E7
	.dc.b	$BF,$FE,$F3,$E8,$FE,$7D,$7F,$E7,$29,$FE,$69,$18,$DF,$B6,$3F,$A1
	.dc.b	$AF,$0D,$79,$63,$F4,$94,$F8,$3E,$9F,$07,$D5,$B5,$D3,$E0,$BF,$C5
	.dc.b	$3A,$FA,$2F,$FC,$EA,$E4,$F0,$FC,$1E,$B1,$EB,$F0,$57,$0F,$F0,$3D
	.dc.b	$FF,$3E,$BF,$3C,$E5,$FE,$CF,$0F,$83,$C5,$A1,$F0,$78,$6D,$7C,$08
	.dc.b	$C7,$C1,$E7,$E0,$F5,$F0,$72,$BE,$2C,$71,$CA,$47,$88,$F9,$39,$75
	.dc.b	$F8,$29,$1F,$31,$F0,$7C,$17,$87,$DD,$93,$F4,$1F,$27,$BD,$D2,$3E
	.dc.b	$36,$17,$6F,$9F,$8F,$A7,$E0,$65,$7D,$87,$D9,$F0,$5E,$DF,$82,$F7
	.dc.b	$7C,$14,$AF,$BB,$1F,$07,$E0,$64,$7E,$01,$F8,$9F,$05,$F1,$7E,$0A
	.dc.b	$4F,$E1,$1F,$97,$F9,$07,$F4,$1D,$3F,$92,$3F,$CF,$67,$FB,$E3,$FC
	.dc.b	$EE,$33,$F0,$FC,$3E,$13,$BD,$7E,$07,$1B,$5F,$8A,$C7,$AF,$C1,$F6
	.dc.b	$F8,$49,$7D,$23,$C3,$F0,$1E,$2F,$72,$4F,$82,$39,$F6,$FE,$1B,$2B
	.dc.b	$3F,$84,$C7,$D3,$F0,$9E,$AF,$84,$93,$E8,$8F,$6F,$C2,$48,$F5,$BF
	.dc.b	$03,$E1,$3E,$0F,$C2,$49,$F7,$BF,$13,$E1,$3E,$2F,$C2,$49,$F8,$6F
	.dc.b	$C8,$F8,$4F,$93,$F0,$7F,$2B,$E0,$E5,$7C,$88,$E8,$FE,$F3,$9F,$FA
	.dc.b	$5B,$D8,$F9,$B8,$7D,$2F,$82,$91,$F4,$1F,$AB,$F0,$7F,$5B,$E0,$E4
	.dc.b	$FD,$47,$EC,$FC,$0F,$6E,$9F,$B3,$B5,$E5,$FF,$39,$E6,$CF,$F5,$3F
	.dc.b	$90,$3E,$2A,$93,$3F,$F1,$D4,$7F,$45,$6F,$C2,$70,$F3,$A4,$60,$DF
	.dc.b	$C2,$74,$F8,$49,$18,$39,$FC,$27,$83,$E1,$24,$60,$EF,$F0,$9E,$4F
	.dc.b	$84,$91,$83,$91,$FC,$93,$36,$1D,$A3,$18,$FB,$3E,$0F,$DB,$F0,$92
	.dc.b	$7D,$71,$F0,$3D,$CF,$83,$F0,$12,$7D,$F1,$F1,$2F,$A7,$2F,$F8,$17
	.dc.b	$2E,$F5,$E2,$8C,$7C,$27,$CA,$F8,$49,$3F,$22,$3E,$6F,$C1,$C8,$F9
	.dc.b	$6F,$D0,$F8,$4F,$A3,$F0,$72,$7E,$7B,$F5,$3E,$13,$EA,$FC,$24,$9F
	.dc.b	$A6,$FF,$20,$FC,$83,$F9,$7C,$7F,$46,$7F,$9B,$3F,$E7,$15,$F0,$9B
	.dc.b	$38,$7C,$27,$79,$F9,$58,$5D,$7D,$D8,$F5,$F8,$49,$5C,$DF,$07,$C2
	.dc.b	$78,$7D,$CF,$17,$B9,$2B,$C1,$1C,$FC,$2F,$F6,$0F,$1F,$F5,$89,$F6
	.dc.b	$9F,$8C,$C7,$D5,$F0,$9E,$BF,$84,$93,$E9,$8F,$77,$C1,$FB,$FE,$12
	.dc.b	$4F,$B6,$3E,$17,$C2,$48,$F8,$0F,$C5,$F8,$4F,$8D,$F0,$92,$7E,$23
	.dc.b	$F2,$7E,$13,$E5,$7C,$24,$9F,$90,$FC,$DF,$84,$E9,$1F,$37,$0B,$C9
	.dc.b	$FE,$47,$E8,$65,$F4,$FE,$0E,$5F,$D1,$7E,$B7,$C2,$7D,$7F,$76,$4F
	.dc.b	$D5,$7E,$D7,$BB,$D3,$97,$FD,$CF,$8F,$FA,$32,$BD,$67,$DB,$C7,$F0
	.dc.b	$1F,$09,$23,$EF,$3F,$D8,$BE,$07,$F9,$C5,$FF,$BC,$1F,$F3,$A3,$FE
	.dc.b	$B3,$5F,$09,$FE,$5F,$8D,$7C,$25,$BF,$F6,$9D,$6A,$78,$39,$D8,$FF
	.dc.b	$9E,$27,$FC,$2A,$BE,$13,$4E,$67,$FC,$D8,$BE,$0F,$26,$BE,$12,$E1
	.dc.b	$E5,$F8,$3F,$51,$31,$F4,$7C,$27,$B7,$E1,$3D,$DF,$09,$27,$D9,$1F
	.dc.b	$07,$DC,$91,$EF,$7E,$27,$B7,$F1,$7E,$12,$4F,$C3,$7E,$47,$C2,$7C
	.dc.b	$9F,$83,$93,$F1,$DF,$99,$F0,$9F,$37,$E1,$24,$FC,$B7,$E8,$7C,$27
	.dc.b	$D1,$F8,$4F,$A5,$F0,$92,$BE,$84,$7F,$85,$BF,$1E,$FE,$5D,$1F,$EE
	.dc.b	$56,$BF,$E3,$E6,$F3,$7F,$A4,$DF,$E6,$11,$9F,$83,$E4,$47,$97,$C1
	.dc.b	$F5,$F6,$BB,$7B,$F2,$7B,$47,$87,$E0,$FC,$5F,$07,$27,$C5,$1E,$5F
	.dc.b	$83,$F3,$7C,$1C,$9F,$34,$7A,$7E,$0B,$F0,$5A,$1B,$F0,$78,$73,$F0
	.dc.b	$53,$BB,$DD,$F0,$77,$1F,$77,$C1,$72,$F8,$3E,$6D,$AF,$89,$F0,$7F
	.dc.b	$1C,$95,$F1,$FE,$0F,$E5,$7C,$07,$CB,$F6,$E4,$FC,$B8,$F9,$DF,$07
	.dc.b	$FE,$44,$FC,$7B,$F9,$2F,$4F,$FA,$F5,$B9,$2D,$2E,$15,$F1,$F8,$3E
	.dc.b	$5F,$09,$2F,$85,$F3,$9F,$1C,$7D,$BC,$FC,$6F,$DB,$CF,$C6,$F0,$CB
	.dc.b	$8E,$7E,$28,$F0,$B7,$0F,$16,$5E,$9F,$84,$95,$C4,$DB,$ED,$F8,$4F
	.dc.b	$0C,$7E,$27,$2F,$0D,$FE,$27,$2F,$0C,$FF,$A2,$B1,$EC,$77,$D7,$DF
	.dc.b	$1D,$69,$3F,$E7,$18,$E1,$F0,$9F,$2C,$F9,$9F,$09,$F3,$7E,$0A,$4F
	.dc.b	$1A,$FF,$37,$BD,$FE,$0F,$E9,$7C,$17,$D3,$F8,$39,$3B,$8F,$AD,$EF
	.dc.b	$FD,$7F,$6A,$4F,$38,$EA,$FA,$71,$BA,$7D,$BC,$7A,$CF,$87,$1E,$AF
	.dc.b	$BF,$1F,$C1,$7C,$1C,$8F,$33,$F8,$5F,$83,$E7,$CB,$F5,$32,$B5,$F5
	.dc.b	$31,$E7,$97,$E2,$72,$B7,$F6,$B2,$FC,$77,$C1,$CC,$3E,$13,$F9,$1F
	.dc.b	$83,$FC,$97,$C1,$C9,$F8,$CF,$E5,$7E,$0F,$F2,$DF,$07,$27,$E5,$3F
	.dc.b	$99,$F8,$3F,$CD,$7C,$17,$FA,$7B,$F1,$CF,$E7,$B5,$F9,$86,$ED,$F6
	.dc.b	$27,$B4,$FD,$FC,$6C,$FE,$87,$1E,$D3,$F8,$3C,$6C,$FE,$97,$1E,$DF
	.dc.b	$05,$2B,$F0,$99,$7F,$A0,$5F,$F6,$9B,$FE,$81,$DB,$F3,$79,$76,$6C
	.dc.b	$FE,$0F,$2E,$B1,$FA,$8A,$FE,$93,$3D,$3E,$07,$2F,$AF,$E0,$BD,$DF
	.dc.b	$05,$23,$C3,$1F,$07,$E0,$BE,$17,$C1,$49,$F2,$C7,$C5,$F8,$2F,$8D
	.dc.b	$F0,$32,$7D,$31,$F2,$7E,$0B,$E5,$7C,$14,$9F,$6C,$75,$7D,$F8,$75
	.dc.b	$7D,$F8,$73,$7E,$0F,$C1,$7D,$33,$1C,$6D,$FD,$DF,$05,$F5,$C9,$5F
	.dc.b	$2B,$E0,$BB,$3F,$2B,$1E,$D7,$C6,$3E,$E6,$3D,$5F,$A3,$F0,$5F,$81
	.dc.b	$30,$FD,$2F,$82,$FE,$E9,$C6,$F8,$2B,$18,$73,$F0,$5A,$F8,$1B,$B7
	.dc.b	$DC,$78,$FC,$14,$8F,$E0,$5E,$77,$FA,$0D,$FC,$0C,$7F,$80,$61,$F1
	.dc.b	$34,$C7,$C1,$DA,$FC,$3E,$1D,$DA,$77,$3A,$FC,$1F,$33,$A7,$C1,$58
	.dc.b	$F0,$FC,$1F,$73,$C1,$F0,$56,$39,$F0,$7E,$4B,$5F,$5B,$63,$D3,$F0
	.dc.b	$7E,$73,$D1,$F0,$56,$3D,$BF,$07,$EB,$F8,$2F,$63,$63,$E0,$FC,$1F
	.dc.b	$BC,$F8,$1F,$07,$63,$E2,$FC,$1F,$C3,$3E,$27,$C1,$58,$F9,$3F,$07
	.dc.b	$F1,$CF,$91,$F0,$76,$3A,$93,$5F,$EA,$2F,$F4,$CE,$66,$3D,$E7,$27
	.dc.b	$F5,$68,$61,$DE,$33,$DA,$9C,$DF,$98,$FB,$7F,$07,$F6,$4F,$B5,$EF
	.dc.b	$58,$F1,$57,$E4,$4E,$5A,$FC,$08,$C9,$E5,$C7,$C1,$3B,$C3,$BE,$3E
	.dc.b	$25,$AF,$E1,$FD,$F9,$87,$E3,$3E,$06,$DF,$A3,$F1,$AF,$EE,$53,$FD
	.dc.b	$45,$7C,$1F,$F9,$3D,$CC,$EE,$31,$F0,$99,$F8,$4B,$7C,$5E,$19,$F1
	.dc.b	$F8,$2B,$3D,$5E,$98,$75,$C2,$CF,$85,$E7,$1E,$5B,$BA,$FC,$5C,$91
	.dc.b	$E5,$F8,$39,$87,$A7,$DE,$B8,$7B,$4E,$D1,$ED,$69,$F0,$74,$F5,$47
	.dc.b	$C1,$6C,$7C,$5F,$7B,$E1,$1F,$0F,$DD,$B5,$F1,$FD,$EF,$8C,$7C,$7F
	.dc.b	$81,$B1,$F2,$FE,$0B,$E5,$1F,$2F,$E0,$AC,$7C,$FF,$82,$F9,$DF,$05
	.dc.b	$F3,$DB,$1D,$49,$9F,$D7,$C3,$AB,$F6,$70,$E6,$FD,$BC,$2F,$0F,$C7
	.dc.b	$F8,$2F,$B5,$EF,$7D,$B6,$E3,$D2,$7E,$66,$9E,$98,$C1,$FE,$2E,$E3
	.dc.b	$FF,$56,$BE,$B3,$F0,$D8,$76,$75,$EF,$7E,$1F,$E0,$7F,$10,$D8,$FC
	.dc.b	$67,$C0,$FE,$2F,$E0,$BF,$18,$D8,$FF,$10,$7E,$35,$FC,$E5,$BF,$07
	.dc.b	$FE,$9F,$8C,$6F,$FA,$E7,$FD,$56,$DF,$86,$D1,$BF,$84,$B7,$E0,$3F
	.dc.b	$39,$8F,$81,$A7,$80,$ED,$87,$76,$9E,$4F,$83,$F1,$1E,$3F,$84,$B1
	.dc.b	$D7,$1F,$D5,$5F,$56,$C7,$53,$D6,$57,$D5,$7E,$DF,$73,$DD,$EE,$DA
	.dc.b	$F8,$6F,$C1,$F8,$39,$1F,$1B,$E1,$2E,$1F,$0F,$E0,$FE,$31,$F1,$FE
	.dc.b	$12,$D7,$CC,$F8,$4F,$94,$7C,$BF,$84,$B1,$F4,$31,$E7,$D3,$FC,$47
	.dc.b	$8D,$9E,$2B,$FC,$DF,$0C,$BA,$E7,$F4,$F3,$B8,$71,$F8,$4F,$B1,$F0
	.dc.b	$9F,$65,$B8,$F7,$25,$FD,$BC,$BB,$3F,$E8,$2C,$BB,$4F,$E1,$32,$F6
	.dc.b	$34,$FC,$3F,$C1,$FE,$14,$FC,$37,$C1,$D8,$FC,$5F,$C0,$7E,$27,$DC
	.dc.b	$FC,$53,$63,$F1,$FF,$07,$F8,$D3,$FD,$D9,$F8,$CF,$F6,$FC,$FF,$1D
	.dc.b	$18,$F8,$4C,$FC,$25,$BF,$F3,$B5,$A3,$AC,$7E,$6B,$1B,$87,$E5,$DE
	.dc.b	$D1,$E0,$38,$FC,$56,$27,$FA,$4D,$BE,$17,$C7,$F0,$BE,$46,$DF,$B0
	.dc.b	$EC,$5A,$EC,$F5,$7D,$2F,$63,$1C,$9F,$D3,$6D,$F0,$BF,$00,$F8,$3F
	.dc.b	$0B,$6B,$DB,$F0,$BF,$10,$F8,$BF,$0B,$63,$E0,$FC,$27,$C8,$3E,$4F
	.dc.b	$C2,$D8,$F8,$BF,$01,$F3,$0F,$9B,$F0,$76,$3A,$93,$5F,$83,$87,$57
	.dc.b	$E1,$E1,$CD,$F8,$B8,$5D,$7D,$5F,$01,$F6,$0F,$B3,$F0,$37,$1E,$93
	.dc.b	$EF,$D3,$D3,$18,$3F,$D9,$3B,$B2,$E9,$18,$C6,$E5,$FE,$D4,$A7,$C2
	.dc.b	$FE,$18,$FC,$3F,$C2,$DB,$FC,$1F,$C2,$FE,$28,$FC,$5F,$C2,$D8,$FC
	.dc.b	$3F,$C2,$FE,$3B,$E1,$7F,$1E,$D8,$FF,$60,$73,$FF,$6E,$DF,$FC,$04
	.dc.b	$D8,$FF,$6F,$BF,$DD,$1D,$E9,$D4,$FC,$4E,$3E,$08,$F1,$9E,$03,$F2
	.dc.b	$F8,$78,$1A,$78,$3D,$FF,$F0,$83,$D9,$B5,$FE,$F4,$BF,$ED,$F3,$B3
	.dc.b	$AC,$78,$1F,$D1,$E5,$E3,$C7,$F1,$D9,$7B,$6B,$F3,$AD,$8F,$6F,$C2
	.dc.b	$F6,$58,$7F,$8A,$70,$F8,$7F,$0B,$98,$C6,$1F,$E2,$E5,$F0,$79,$F8
	.dc.b	$3D,$36,$FE,$5F,$C1,$F1,$39,$7C,$1D,$8F,$9F,$EE,$F5,$3B,$7C,$0D
	.dc.b	$8F,$A7,$7C,$F2,$F0,$37,$4F,$AB,$8F,$39,$F2,$63,$67,$EC,$FC,$0F
	.dc.b	$A4,$F5,$7B,$D7,$1E,$A7,$E0,$34,$F4,$B4,$F4,$93,$0E,$AD,$BF,$C2
	.dc.b	$FC,$17,$C5,$3E,$37,$C1,$D9,$FC,$5F,$C1,$7C,$93,$E5,$7C,$1D,$8F
	.dc.b	$C7,$FC,$0F,$CD,$F8,$2F,$9C,$D8,$FC,$9F,$C1,$7D,$13,$E9,$7B,$D6
	.dc.b	$3A,$93,$4F,$A9,$87,$57,$EB,$5F,$36,$CF,$E7,$B0,$FC,$07,$E3,$3F
	.dc.b	$DF,$35,$FF,$30,$3F,$DF,$1B,$3A,$F7,$FA,$47,$E9,$34,$F5,$CF,$86
	.dc.b	$3F,$4F,$7D,$DF,$D4,$E3,$65,$8A,$58,$BE,$0F,$C7,$F0,$7E,$46,$DF
	.dc.b	$0A,$F2,$FC,$1F,$A0,$F4,$FC,$1D,$9D,$1B,$C1,$F0,$7E,$E3,$DF,$F0
	.dc.b	$76,$77,$9F,$8E,$9F,$03,$F1,$3D,$EF,$8A,$DA,$DE,$7E,$8C,$FC,$18
	.dc.b	$7F,$A4,$30,$67,$8A,$F9,$79,$F7,$AD,$67,$67,$E7,$FB,$BF,$4B,$E0
	.dc.b	$7E,$9B,$73,$EF,$93,$3E,$91,$ED,$CE,$E5,$E9,$CF,$A4,$FC,$1C,$EC
	.dc.b	$FB,$7E,$0F,$EF,$FC,$1F,$E0,$1B,$5D,$BE,$0B,$F0,$7F,$07,$F8,$46
	.dc.b	$C7,$8B,$E0,$BF,$0F,$F0,$5F,$88,$6C,$79,$BE,$0B,$F1,$7E,$FF,$E3
	.dc.b	$1B,$1D,$4F,$5E,$1D,$5A,$7B,$BE,$0B,$F8,$97,$E3,$3F,$E0,$1E,$5F
	.dc.b	$C9,$C6,$3E,$07,$3F,$07,$AF,$83,$B8,$7D,$4C,$3F,$C0,$4F,$7E,$1F
	.dc.b	$A9,$97,$76,$C6,$FE,$0F,$BF,$9F,$E0,$6C,$73,$F8,$3F,$CB,$D7,$FB
	.dc.b	$33,$07,$C1,$E3,$6F,$F6,$34,$62,$78,$1E,$3F,$7B,$58,$7E,$66,$51
	.dc.b	$89,$FD,$14,$61,$F5,$7C,$0E,$6D,$3F,$D9,$67,$DD,$EF,$66,$0D,$F7
	.dc.b	$F1,$F0,$79,$F8,$39,$7F,$16,$38,$65,$23,$CC,$78,$0F,$95,$5C,$DA
	.dc.b	$7C,$DF,$83,$F0,$7C,$1C,$8F,$A0,$F8,$FE,$0B,$A4,$7C,$7B,$BA,$7C
	.dc.b	$6D,$7A,$4F,$CB,$C6,$CF,$CA,$C7,$D9,$F0,$5E,$DF,$82,$97,$F7,$63
	.dc.b	$E0,$7C,$07,$C1,$F8,$09,$3F,$81,$8F,$89,$F0,$5F,$27,$F1,$9F,$B3
	.dc.b	$3F,$C3,$F9,$3F,$85,$46,$33,$FB,$35,$BF,$88,$97,$FB,$EA,$FF,$78
	.dc.b	$B7,$E1,$3A,$7C,$47,$5F,$85,$93,$B8,$F0,$7C,$2C,$8E,$4F,$8F,$E0
	.dc.b	$FC,$9E,$FF,$F6,$FC,$4D,$F8,$3C,$39,$F8,$3D,$7C,$2D,$E0,$F2,$3C
	.dc.b	$7E,$17,$97,$C2,$D9,$F4,$3D,$7E,$16,$47,$AF,$E1,$6E,$DF,$0C,$FE
	.dc.b	$65,$A3,$E1,$7C,$87,$97,$E0,$6D,$FC,$3F,$83,$F4,$1E,$9F,$81,$B1
	.dc.b	$F1,$FE,$0B,$D9,$F0,$BE,$D6,$C7,$CB,$C3,$9E,$5E,$1E,$A6,$6C,$F3
	.dc.b	$0B,$FA,$F9,$F6,$BE,$16,$CF,$17,$B9,$F5,$FC,$2F,$F8,$6F,$DD,$FE
	.dc.b	$B9,$8F,$E0,$23,$0F,$63,$F0,$79,$FB,$DF,$31,$EF,$25,$78,$30,$EF
	.dc.b	$1F,$D4,$EF,$FB,$33,$7E,$0F,$C1,$F0,$BE,$17,$C5,$F0,$BE,$3A,$B5
	.dc.b	$FD,$29,$FE,$78,$DF,$80,$F4,$7C,$07,$A5,$B1,$BF,$80,$F6,$1E,$DF
	.dc.b	$84,$B1,$CF,$0E,$CD,$3B,$61,$CD,$F0,$61,$74,$FA,$53,$FD,$9A,$7C
	.dc.b	$9F,$0B,$F2,$7E,$13,$E5,$36,$3D,$1F,$0B,$F3,$4F,$9D,$F0,$76,$3D
	.dc.b	$9F,$09,$F4,$4F,$A5,$F0,$96,$3E,$07,$B9,$F5,$7E,$0F,$EB,$36,$3A
	.dc.b	$93,$4F,$2E,$1D,$5F,$3E,$1C,$DF,$4E,$17,$9F,$F0,$47,$F6,$7F,$84
	.dc.b	$FC,$17,$C1,$FE,$0D,$B8,$FD,$1F,$84,$FC,$31,$F8,$7F,$84,$B1,$FD
	.dc.b	$FB,$F1,$9F,$EA,$4F,$F8,$45,$7C,$27,$F9,$9D,$CE,$7F,$E1,$B7,$33
	.dc.b	$C7,$E1,$34,$6F,$E0,$EC,$78,$49,$5F,$4B,$0F,$0B,$FA,$5B,$EB,$A1
	.dc.b	$EB,$19,$F8,$5F,$16,$5E,$3F,$85,$98,$FA,$7E,$16,$DF,$DE,$D7,$AE
	.dc.b	$5F,$1B,$A1,$A6,$30,$F7,$67,$D1,$F8,$4D,$C3,$C9,$F0,$BF,$07,$DF
	.dc.b	$F8,$4D,$AF,$91,$EF,$FC,$5F,$7F,$E3,$36,$3E,$67,$C2,$7C,$93,$E5
	.dc.b	$7C,$1D,$8F,$A1,$F0,$9F,$37,$E0,$BE,$73,$63,$A9,$F5,$70,$EA,$D3
	.dc.b	$A9,$33,$F9,$57,$D5,$F9,$9F,$09,$F6,$7E,$13,$ED,$36,$BF,$01,$F0
	.dc.b	$7F,$74,$FB,$DF,$09,$63,$F0,$9F,$07,$F8,$1F,$84,$FC,$13,$63,$F1
	.dc.b	$1F,$09,$F8,$53,$FD,$59,$F8,$CF,$FB,$27,$3F,$C3,$47,$37,$EF,$63
	.dc.b	$DA,$BC,$27,$63,$F2,$58,$76,$69,$FE,$52,$BF,$F6,$95,$7F,$2B,$5F
	.dc.b	$09,$E0,$F8,$5F,$0B,$E2,$F8,$4F,$1F,$C2,$DC,$39,$3E,$6F,$85,$91
	.dc.b	$D7,$E1,$6E,$5D,$7E,$13,$A4,$F8,$5B,$1E,$1C,$3A,$CF,$95,$B1,$E5
	.dc.b	$BE,$7D,$DF,$B6,$D3,$FD,$93,$7E,$B9,$F8,$FF,$0B,$F2,$31,$B3,$DC
	.dc.b	$E8,$7A,$0B,$87,$C1,$CF,$AC,$7F,$A1,$F0,$61,$F4,$B3,$CD,$FF,$A4
	.dc.b	$65,$F0,$BF,$58,$FA,$FF,$0B,$70,$F9,$DF,$0B,$F6,$8F,$B7,$F0,$B6
	.dc.b	$3E,$97,$C2,$FD,$E3,$EF,$FC,$2D,$8F,$AD,$87,$3C,$BE,$B6,$37,$83
	.dc.b	$E1,$65,$CE,$7E,$D6,$56,$7E,$36,$5C,$E7,$EF,$65,$67,$E5,$7B,$FF
	.dc.b	$8E,$3F,$1F,$F0,$77,$4F,$C3,$7C,$1F,$E4,$8F,$C9,$FC,$0D,$8F,$C5
	.dc.b	$7C,$27,$E5,$8F,$CB,$FC,$2D,$8F,$C7,$7C,$27,$F2,$C7,$FD,$65,$F8
	.dc.b	$CF,$F6,$CD,$7E,$54,$E5,$FF,$5E,$46,$74,$71,$B3,$F0,$57,$E5,$F3
	.dc.b	$B5,$F8,$3C,$FC,$13,$F9,$BC,$EC,$FF,$3C,$C7,$FB,$7C,$F6,$F8,$5F
	.dc.b	$04,$A7,$FD,$95,$A7,$FA,$D9,$DC,$67,$E0,$77,$06,$F9,$F8,$73,$F0
	.dc.b	$92,$F9,$7C,$15,$CB,$FC,$31,$83,$E1,$39,$9D,$3E,$12,$DF,$87,$1E
	.dc.b	$8F,$8F,$2B,$87,$2C,$BA,$67,$ED,$CE,$DF,$CE,$F8,$3F,$3F,$C2,$7A
	.dc.b	$1B,$1D,$49,$87,$93,$1E,$B8,$7C,$7C,$79,$CF,$D4,$C6,$CF,$C7,$F8
	.dc.b	$4F,$87,$F0,$5F,$11,$B8,$FC,$9F,$82,$F8,$FF,$05,$F2,$1B,$1F,$37
	.dc.b	$E0,$FE,$59,$F3,$3E,$12,$C7,$D1,$F8,$3F,$9F,$F0,$9F,$41,$B1,$D4
	.dc.b	$FA,$D8,$75,$69,$D4,$9A,$FD,$0B,$EA,$FD,$2F,$7B,$ED,$FB,$DF,$71
	.dc.b	$B5,$FE,$F6,$F2,$3F,$B4,$93,$E0,$7F,$0E,$E3,$E1,$2F,$C2,$61,$B5
	.dc.b	$C7,$E0,$F6,$70,$F8,$3B,$1D,$7E,$0F,$99,$D3,$1B,$1E,$1B,$E7,$A7
	.dc.b	$F9,$7F,$2F,$F6,$AB,$72,$F1,$FC,$14,$BF,$47,$C1,$DA,$E1,$F0,$7E
	.dc.b	$83,$D3,$F0,$16,$BD,$DF,$07,$EC,$3D,$BF,$03,$63,$E1,$7C,$17,$C0
	.dc.b	$3E,$0F,$C1,$D8,$F8,$D8,$73,$C7,$CD,$8D,$ED,$FF,$15,$ED,$FE,$CE
	.dc.b	$BE,$E7,$CD,$CB,$B5,$FB,$7E,$0B,$E6,$FC,$1F,$CE,$6D,$76,$24,$F6
	.dc.b	$6E,$1D,$B5,$E6,$F8,$30,$B2,$A4,$FF,$3C,$F8,$0F,$B6,$7D,$CF,$83
	.dc.b	$B5,$F8,$1F,$83,$FB,$FF,$05,$F8,$06,$C7,$E1,$7E,$0F,$F8,$4F,$E3
	.dc.b	$3F,$DD,$71,$FE,$8F,$18,$CB,$3F,$09,$2F,$C5,$5F,$E4,$EB,$FE,$DF
	.dc.b	$18,$32,$FF,$74,$8F,$93,$87,$82,$FF,$04,$D8,$F9,$B8,$77,$BF,$21
	.dc.b	$CE,$BD,$F7,$E5,$1F,$07,$E8,$F8,$1F,$4B,$6B,$86,$0A,$FE,$C3,$18
	.dc.b	$6D,$FF,$AC,$5F,$EE,$2D,$E2,$F8,$7F,$05,$2B,$87,$C1,$DA,$E1,$F0
	.dc.b	$5F,$23,$E0,$BE,$4B,$6B,$B6,$1F,$30,$F9,$BF,$07,$63,$C5,$8F,$3C
	.dc.b	$FE,$56,$56,$7E,$4F,$C1,$FD,$4F,$83,$FA,$AD,$8F,$57,$C1,$FD,$8F
	.dc.b	$83,$FB,$2D,$8F,$77,$C1,$FD,$CF,$83,$FB,$AD,$8F,$85,$F0,$3F,$80
	.dc.b	$F8,$1F,$C0,$B6,$3A,$9F,$1F,$0E,$AD,$3A,$9F,$2E,$FA,$B4,$F9,$DF
	.dc.b	$03,$F8,$CF,$81,$FC,$6B,$63,$AF,$27,$D1,$BE,$AD,$8E,$AF,$D5,$BE
	.dc.b	$AD,$8F,$B5,$F0,$7F,$98,$3F,$B0,$FE,$33,$FD,$BF,$9D,$18,$F8,$3C
	.dc.b	$FC,$3D,$BF,$C2,$DF,$F6,$6C,$3F,$DE,$4F,$F6,$38,$CF,$C0,$F4,$F7
	.dc.b	$FA,$BD,$BD,$F9,$3C,$BD,$FB,$87,$2C,$BB,$CF,$66,$C7,$6B,$E7,$3E
	.dc.b	$2C,$2C,$F8,$BE,$0F,$D7,$F0,$7E,$C6,$D7,$A3,$E0,$BD,$FF,$07,$F0
	.dc.b	$1B,$1E,$CF,$82,$F8,$7F,$07,$F1,$1B,$1F,$03,$E0,$FE,$3F,$C0,$7C
	.dc.b	$86,$C7,$53,$E2,$E1,$D5,$A7,$53,$E4,$DF,$56,$9F,$33,$E0,$FE,$9F
	.dc.b	$C0,$FD,$46,$C7,$D0,$F8,$2F,$AF,$F0,$7F,$61,$B1,$F5,$3E,$0B,$ED
	.dc.b	$FC,$1F,$DC,$6C,$7E,$F3,$F1,$9F,$F6,$19,$FE,$C2,$F1,$FF,$DB,$EE
	.dc.b	$3E,$07,$A4,$FD,$EC,$6F,$27,$DE,$7A,$47,$E0,$B1,$B3,$F8,$27,$FC
	.dc.b	$57,$7F,$D5,$B0,$EB,$F0,$38,$F8,$1C,$BD,$23,$FD,$DF,$D8,$CB,$19
	.dc.b	$7F,$26,$8C,$3E,$D3,$C0,$5C,$BC,$0F,$B7,$E1,$3C,$3F,$09,$E2,$6C
	.dc.b	$7C,$1F,$83,$F2,$9E,$6F,$83,$B1,$F1,$7E,$07,$D3,$F0,$3E,$A6,$C7
	.dc.b	$C9,$F7,$FD,$A7,$BB,$DF,$B1,$D4,$99,$FF,$4E,$7F,$9F,$F0,$B1,$EF
	.dc.b	$39,$3F,$A9,$C3,$0E,$F1,$9D,$D3,$9B,$FB,$59,$7C,$EF,$83,$91,$F7
	.dc.b	$7E,$0E,$E1,$C6,$DF,$05,$F5,$3E,$0F,$EA,$B6,$FF,$09,$F0,$7F,$63
	.dc.b	$E0,$FE,$CB,$63,$F1,$1F,$03,$F7,$3E,$0F,$EE,$B6,$3F,$19,$7C,$F2
	.dc.b	$F0,$37,$6F,$C6,$E1,$CE,$7C,$8D,$8F,$C8,$FC,$17,$E2,$0F,$F3,$9F
	.dc.b	$37,$F6,$BD,$51,$8F,$83,$CF,$C3,$DC,$FF,$AB,$DF,$F6,$AC,$BE,$43
	.dc.b	$FD,$7A,$33,$F0,$7D,$3E,$13,$AB,$DB,$E0,$BB,$FC,$15,$AE,$6F,$5C
	.dc.b	$39,$E3,$75,$57,$F7,$AC,$BC,$11,$8A,$FF,$06,$C7,$85,$F1,$E3,$E1
	.dc.b	$AF,$19,$E8,$F8,$2F,$79,$F0,$3E,$0A,$C7,$B3,$5F,$7B,$4F,$7E,$5E
	.dc.b	$A7,$9E,$1F,$03,$5C,$3F,$0B,$E0,$FE,$59,$F3,$3E,$12,$D7,$C8,$F8
	.dc.b	$3F,$9F,$F0,$9F,$41,$B1,$F3,$3E,$13,$E9,$9F,$53,$DF,$B1,$F4,$3E
	.dc.b	$13,$EB,$FC,$07,$D8,$6C,$75,$3E,$AE,$1D,$5A,$75,$26,$7F,$62,$FA
	.dc.b	$BF,$6B,$E0,$FF,$07,$F0,$9F,$84,$6D,$7E,$03,$E0,$FF,$0E,$7E,$23
	.dc.b	$E1,$2C,$7E,$13,$E0,$FF,$D7,$BF,$8C,$FF,$B2,$FA,$3F,$AE,$CF,$F8
	.dc.b	$7A,$3F,$D9,$C7,$FA,$F8,$9F,$09,$C3,$E0,$B9,$E9,$F8,$CC,$BB,$5F
	.dc.b	$84,$E4,$76,$3F,$33,$8D,$7B,$7C,$17,$88,$F1,$FC,$15,$8F,$17,$C2
	.dc.b	$79,$BE,$13,$CE,$D8,$F3,$7C,$1F,$AB,$E1,$3D,$6D,$8F,$57,$C2,$7B
	.dc.b	$8F,$7F,$C0,$D8,$EA,$7B,$F0,$EA,$D3,$A9,$F0,$EF,$AB,$4F,$8D,$F0
	.dc.b	$9F,$28,$F9,$7F,$05,$63,$AE,$FF,$C5,$F8,$98,$76,$8C,$3D,$1D,$61
	.dc.b	$7C,$5F,$D4,$0A,$47,$D8,$F8,$39,$87,$D7,$F8,$4B,$87,$7F,$83,$FB
	.dc.b	$BF,$09,$F7,$9B,$5F,$77,$E1,$3F,$02,$7E,$0B,$DF,$B1,$F8,$1F,$84
	.dc.b	$FC,$29,$F8,$6F,$83,$B1,$F8,$5B,$E7,$8F,$78,$FC,$46,$3C,$A3,$E0
	.dc.b	$63,$D6,$7C,$07,$F9,$AB,$F1,$97,$C3,$FF,$A4,$6B,$F9,$E4,$3E,$1B
	.dc.b	$1C,$3F,$0A,$31,$3C,$6B,$37,$FE,$88,$97,$C2,$E2,$BF,$D3,$71,$87
	.dc.b	$C1,$19,$B4,$BE,$1B,$0B,$E1,$69,$2F,$D1,$E1,$6B,$E1,$64,$7A,$DE
	.dc.b	$3F,$07,$CB,$29,$3E,$03,$E0,$63,$CD,$F7,$E1,$72,$F7,$63,$E2,$F8
	.dc.b	$3F,$1F,$C1,$CB,$F9,$11,$E6,$F8,$5F,$3F,$C2,$C9,$F9,$91,$EA,$F8
	.dc.b	$59,$1F,$3D,$F6,$FC,$2F,$BB,$E0,$7D,$FF,$0B,$2B,$EA,$47,$57,$EB
	.dc.b	$61,$D5,$FA,$B8,$74,$7E,$BE,$17,$4F,$AD,$8F,$CA,$F8,$29,$5F,$7D
	.dc.b	$F9,$BF,$0B,$D6,$3E,$FE,$3D,$6F,$EF,$63,$CF,$97,$EE,$65,$6F,$ED
	.dc.b	$E5,$F5,$BE,$12,$47,$E2,$DF,$B3,$F0,$BF,$6B,$E1,$3E,$F7,$E3,$3F
	.dc.b	$8D,$AF,$F0,$43,$73,$FC,$64,$EB,$E0,$25,$7F,$AF,$A9,$1C,$BE,$03
	.dc.b	$9F,$C1,$F4,$F8,$19,$7B,$8F,$03,$C4,$F0,$98,$3C,$35,$D9,$A7,$7A
	.dc.b	$F3,$FC,$2F,$A3,$E1,$64,$F8,$E3,$D7,$F0,$B2,$3C,$CF,$BB,$E1,$7D
	.dc.b	$FF,$0B,$F0,$3E,$16,$57,$AE,$3E,$1F,$C2,$C8,$F7,$3F,$1B,$E0,$BA
	.dc.b	$47,$C6,$C2,$F0,$FF,$12,$E0,$65,$DA,$BE,$36,$5D,$9F,$9B,$97,$D0
	.dc.b	$F8,$5F,$A3,$F0,$72,$7E,$64,$7D,$4F,$84,$91,$F3,$DF,$AF,$F0,$BF
	.dc.b	$63,$E1,$3E,$CF,$C2,$CA,$FA,$91,$F7,$3E,$0B,$EE,$FC,$2C,$9F,$B1
	.dc.b	$1D,$1F,$BF,$85,$DF,$D5,$8F,$59,$FB,$D8,$F5,$7F,$0B,$8F,$E2,$3E
	.dc.b	$0F,$F8,$37,$C0,$FF,$86,$67,$FC,$A0,$DC,$3E,$24,$F5,$F0,$14,$97
	.dc.b	$13,$3D,$4E,$61,$9F,$8E,$3F,$2B,$9F,$8E,$FE,$86,$7E,$2F,$84,$95
	.dc.b	$DA,$7F,$A2,$E8,$9F,$3F,$C2,$FA,$3E,$16,$5E,$F5,$E1,$69,$F6,$FC
	.dc.b	$2F,$BA,$25,$6C,$F5,$23,$E1,$FC,$07,$C4,$F8,$5F,$8B,$F0,$12,$B8
	.dc.b	$69,$B9,$F0,$C7,$C8,$C6,$CF,$A3,$1E,$73,$C3,$1E,$CD,$72,$FD,$0F
	.dc.b	$85,$FA,$7F,$0B,$23,$84,$7D,$6F,$85,$91,$CD,$FB,$3F,$0B,$F6,$BE
	.dc.b	$12,$4F,$77,$EE,$FC,$27,$DE,$F8,$19,$3E,$37,$F0,$3F,$03,$D2,$3F
	.dc.b	$03,$85,$D7,$E8,$E3,$D2,$7F,$0B,$8D,$9F,$AB,$8F,$E2,$BE,$07,$F1
	.dc.b	$7F,$09,$2F,$E0,$47,$E3,$BE,$13,$F1,$FF,$09,$27,$E2,$47,$F5,$7F
	.dc.b	$C6,$7F,$B3,$C7,$CE,$6B,$FE,$A3,$6E,$3F,$E0,$87,$FB,$3C,$67,$E0
	.dc.b	$39,$11,$E5,$F0,$1E,$17,$E9,$61,$E1,$7F,$3F,$87,$87,$E1,$24,$78
	.dc.b	$5F,$27,$C2,$F9,$7E,$0B,$CD,$F0,$52,$BC,$D1,$D5,$FD,$4E,$1D,$74
	.dc.b	$FC,$CE,$3D,$27,$D5,$8D,$9F,$56,$3F,$07,$E1,$3E,$17,$C2,$4B,$F8
	.dc.b	$51,$F1,$7E,$0F,$E3,$7C,$24,$9F,$8D,$1F,$27,$DF,$F9,$5F,$01,$27
	.dc.b	$E5,$47,$CD,$F8,$0F,$9D,$F0,$12,$7E,$74,$75,$7E,$86,$1D,$5F,$A1
	.dc.b	$87,$57,$EA,$61,$D5,$FA,$98,$7D,$9F,$83,$91,$F6,$5F,$B9,$F0,$9F
	.dc.b	$77,$E1,$24,$FD,$D7,$F0,$1F,$09,$F8,$1F,$84,$93,$F8,$17,$F0,$9F
	.dc.b	$09,$F8,$5F,$84,$93,$F8,$57,$FC,$3D,$F8,$CF,$F5,$FF,$37,$F8,$47
	.dc.b	$6B,$2E,$F7,$F8,$97,$BB,$D4,$87,$E3,$71,$EE,$FE,$FA,$FF,$B2,$B7
	.dc.b	$E0,$FB,$9E,$0F,$83,$F0,$FC,$27,$8B,$E1,$65,$71,$8F,$2F,$C2,$48
	.dc.b	$E8,$FA,$3E,$17,$D3,$F0,$7E,$AF,$83,$95,$E1,$8E,$AF,$8F,$0E,$AF
	.dc.b	$F3,$1D,$FF,$93,$68,$CB,$B4,$E4,$FE,$6D,$0C,$B3,$DD,$FE,$70,$86
	.dc.b	$7F,$2B,$E0,$BE,$5F,$C1,$CC,$3E,$14,$78,$9E,$96,$B6,$7E,$48,$E9
	.dc.b	$97,$6C,$FC,$B9,$DC,$3C,$79,$FD,$8F,$85,$91,$F3,$DF,$B7,$F0,$BF
	.dc.b	$73,$E1,$24,$FD,$37,$EF,$FC,$2F,$E0,$3E,$0A,$4F,$D7,$7F,$07,$F0
	.dc.b	$3F,$84,$F8,$09,$3F,$6D,$FC,$3F,$C0,$F4,$8F,$5F,$0B,$1C,$70,$7C
	.dc.b	$5C,$BA,$D6,$B2,$B3,$F2,$72,$FC,$8F,$C2,$4A,$FC,$43,$F9,$4F,$85
	.dc.b	$FC,$AF,$C2,$FF,$9C,$FF,$19,$FE,$D9,$C7,$FD,$75,$B7,$F5,$A7,$5F
	.dc.b	$0B,$BF,$85,$93,$FD,$42,$BF,$B1,$37,$E0,$BA,$7C,$14,$8C,$1F,$99
	.dc.b	$C3,$C1,$8C,$8C,$1C,$F1,$EA,$FD,$CC,$18,$F8,$A3,$C1,$F0,$5E,$9F
	.dc.b	$85,$91,$83,$C9,$F0,$9E,$D2,$3E,$7F,$85,$F8,$1F,$09,$F0,$7E,$16
	.dc.b	$4F,$AE,$3E,$27,$C1,$C8,$F7,$3F,$1F,$E0,$F9,$C7,$97,$0B,$CB,$FE
	.dc.b	$C8,$D7,$FD,$8F,$3C,$CF,$8D,$8F,$67,$D9,$8F,$D1,$F8,$3F,$A5,$F0
	.dc.b	$72,$7E,$64,$74,$E1,$FA,$99,$59,$FF,$0F,$63,$D9,$E6,$76,$21,$83
	.dc.b	$E5,$7C,$2F,$DF,$23,$F6,$7E,$17,$F0,$5F,$09,$23,$EE,$3F,$85,$F8
	.dc.b	$5F,$C3,$7C,$07,$E1,$FE,$06,$57,$E0,$63,$FD,$79,$E3,$7F,$62,$D3
	.dc.b	$FC,$4C,$D9,$FF,$1A,$37,$87,$F7,$BF,$01,$DE,$7E,$F6,$56,$7E,$56
	.dc.b	$5E,$09,$FC,$76,$5E,$07,$F0,$99,$78,$3E,$17,$C3,$F0,$B2,$7F,$C4
	.dc.b	$79,$FF,$B4,$5F,$E0,$AF,$FB,$35,$CF,$C2,$FA,$0F,$4F,$C2,$FA,$BE
	.dc.b	$0F,$D7,$F0,$92,$B9,$C7,$BB,$E0,$64,$76,$7E,$0F,$C1,$72,$8E,$A4
	.dc.b	$3C,$37,$D5,$F2,$DF,$57,$CB,$7F,$2B,$E0,$7E,$5F,$C1,$49,$F5,$C7
	.dc.b	$CE,$F8,$3F,$9F,$F0,$92,$7D,$F1,$F4,$BE,$12,$47,$C2,$7E,$AF,$C2
	.dc.b	$7D,$6F,$80,$FA,$FF,$09,$2B,$E3,$C7,$57,$E4,$E1,$D5,$F9,$38,$74
	.dc.b	$7E,$5E,$17,$9D,$D5,$F8,$0C,$BF,$07,$F0,$32,$FE,$9B,$F8,$6F,$84
	.dc.b	$ED,$1F,$53,$2E,$D8,$72,$8F,$C3,$E5,$D7,$1F,$C5,$E5,$67,$F1,$79
	.dc.b	$7E,$43,$E1,$3F,$D0,$31,$FD,$0B,$F1,$9F,$EE,$0F,$FA,$0D,$B8,$FF
	.dc.b	$4C,$9D,$BD,$C4,$7F,$6C,$9E,$5F,$07,$AC,$7F,$D0,$12,$8C,$4F,$5F
	.dc.b	$84,$B7,$CB,$DF,$B7,$FD,$86,$BF,$CE,$AD,$7C,$DF,$01,$AF,$80,$F5
	.dc.b	$47,$E6,$33,$C3,$C3,$1E,$79,$48,$C1,$D3,$1E,$E4,$7B,$E1,$E2,$EC
	.dc.b	$F1,$FC,$1C,$9F,$1C,$79,$BE,$13,$CF,$F0,$72,$7C,$F1,$EA,$F8,$4F
	.dc.b	$5F,$C2,$49,$F5,$C7,$BB,$E1,$24,$7B,$9F,$83,$F0,$3F,$0B,$E0,$7E
	.dc.b	$1F,$C0,$CA,$F8,$71,$F1,$BE,$07,$E3,$FC,$24,$9F,$8F,$1C,$9E,$84
	.dc.b	$3E,$F5,$F5,$7E,$6D,$F5,$7F,$05,$7D,$5F,$A3,$7D,$5F,$C3,$5F,$D6
	.dc.b	$F8,$4F,$AF,$F0,$72,$7E,$BC,$7D,$AF,$83,$91,$F6,$9F,$BB,$F0,$3F
	.dc.b	$80,$FC,$67,$FB,$3F,$59,$A4,$F8,$2D,$7C,$16,$FE,$12,$63,$FE,$B8
	.dc.b	$AF,$EB,$AD,$F8,$4E,$9F,$09,$23,$06,$FE,$13,$C0,$47,$97,$BB,$E3
	.dc.b	$F8,$4F,$27,$C0,$C9,$ED,$1F,$98,$5F,$34,$F8,$3C,$7E,$81,$FC,$9F
	.dc.b	$83,$FD,$E0,$DF,$23,$F5,$B5,$BB,$F4,$5F,$C1,$23,$EB,$F8,$6F,$89
	.dc.b	$F0,$D2,$3D,$CF,$C7,$F8,$5F,$15,$F3,$9F,$F2,$20,$C7,$C0,$FF,$BF
	.dc.b	$1A,$FA,$30,$FA,$1F,$0D,$F4,$7E,$16,$4F,$CB,$8F,$A9,$F0,$B2,$3E
	.dc.b	$73,$F5,$FE,$13,$C3,$7F,$03,$E0,$65,$7F,$BC,$4E,$C7,$D5,$F8,$4F
	.dc.b	$BD,$F0,$72,$30,$7D,$9F,$84,$FE,$03,$F9,$4F,$F5,$D3,$FA,$F7,$52
	.dc.b	$3B,$CF,$5D,$BD,$67,$B3,$7F,$85,$C9,$FC,$4F,$0F,$F6,$4B,$73,$EA
	.dc.b	$76,$F8,$6E,$FF,$0D,$E0,$F8,$39,$90,$FF,$31,$E0,$9F,$0C,$7C,$EC
	.dc.b	$BC,$38,$63,$1F,$BF,$5F,$82,$25,$7C,$3C,$BC,$BC,$FD,$69,$8B,$E1
	.dc.b	$BE,$0F,$C1,$CC,$35,$7B,$AD,$47,$3E,$2F,$9E,$F8,$F3,$F1,$57,$E3
	.dc.b	$A7,$C3,$3E,$7C,$F2,$FB,$73,$F1,$B7,$AF,$F1,$66,$9F,$B2,$75,$3E
	.dc.b	$86,$19,$9E,$31,$F8,$FA,$EF,$A7,$F5,$18,$E2,$D9,$D8,$BE,$F3,$EB
	.dc.b	$21,$DC,$93,$C7,$0F,$26,$BF,$4E,$7F,$05,$87,$8E,$79,$C7,$E8,$B0
	.dc.b	$EC,$F4,$BE,$4F,$A8,$86,$E7,$AD,$7A,$B2,$F9,$53,$C5,$FD,$CE,$3D
	.dc.b	$AF,$9C,$F1,$39,$BF,$C0,$32,$F2,$47,$98,$86,$C9,$1F,$16,$7A,$57
	.dc.b	$82,$72,$FF,$74,$CB,$F1,$F1,$E3,$3E,$D6,$3E,$77,$F7,$D8,$F8,$1F
	.dc.b	$85,$7D,$1D,$9C,$4F,$E0,$D8,$FE,$2B,$1F,$59,$31,$F0,$5F,$5A,$FB
	.dc.b	$71,$E5,$9F,$44,$7E,$76,$FC,$11,$F0,$63,$E1,$CE,$C8,$F9,$8F,$87
	.dc.b	$1D,$DF,$29,$E8,$3A,$1E,$7C,$BF,$1F,$97,$C2,$8F,$8A,$70,$9F,$D8
	.dc.b	$BD,$2B,$BB,$E8,$3F,$3C,$74,$38,$99,$3E,$19,$E8,$3A,$3C,$A7,$F2
	.dc.b	$11,$FA,$B3,$C9,$97,$CD,$7F,$A1,$E1,$F0,$AB,$EC,$65,$F3,$5F,$1E
	.dc.b	$3E,$AC,$BE,$63,$FA,$4C,$39,$3F,$98,$C3,$CA,$FC,$0B,$E5,$1E,$42
	.dc.b	$1F,$B1,$3E,$01,$88,$FD,$7E,$1D,$2B,$F5,$35,$CA,$BE,$05,$78,$DE
	.dc.b	$26,$27,$C7,$5E,$B8,$FE,$95,$5C,$6B,$D1,$1C,$31,$F8,$B9,$7D,$B8
	.dc.b	$EB,$7E,$17,$F2,$13,$F6,$67,$55,$E1,$9F,$98,$7C,$0C,$3C,$2F,$CE
	.dc.b	$3E,$A1,$F5,$8D,$1E,$1C,$7C,$CF,$D2,$3E,$F5,$76,$9F,$81,$7F,$D1
	.dc.b	$DF,$DF,$CF,$C0,$BF,$5B,$F9,$12,$57,$17,$C3,$5C,$5F,$5D,$FD,$AB
	.dc.b	$FE,$49,$87,$BE,$7F,$1B,$5E,$FC,$BF,$03,$5F,$AB,$8F,$8F,$18,$9F
	.dc.b	$3B,$FC,$39,$F8,$07,$F2,$27,$F1,$6E,$30,$FC,$FB,$F2,$1F,$33,$E8
	.dc.b	$8E,$4F,$C6,$39,$1F,$68,$E4,$FE,$F9,$F8,$55,$FD,$CB,$5F,$EA,$51
	.dc.b	$FD,$8B,$5C,$3F,$6B,$4F,$C1,$57,$CA,$8F,$15,$FE,$FB,$0F,$EB,$B9
	.dc.b	$FF,$6C,$9F,$ED,$7B,$7F,$70,$79,$1F,$73,$5F,$C1,$57,$B0,$97,$CF
	.dc.b	$87,$FC,$05,$5F,$D9,$04,$FF,$6A,$F8,$AF,$F0,$49,$FE,$4B,$E2,$FC
	.dc.b	$FC,$F5,$CF,$FC,$00,$F1,$AF,$EE,$3A,$7F,$81,$23,$1C,$3F,$E0,$8C
	.dc.b	$BF,$BB,$64,$E6,$FF,$11,$57,$82,$73,$1F,$E2,$38,$F5,$9F,$38,$FE
	.dc.b	$21,$87,$F9,$6F,$97,$FC,$41,$1C,$4F,$4D,$B9,$7F,$CC,$53,$FE,$3F
	.dc.b	$9D,$57,$F9,$8F,$AB,$0F,$FA,$BE,$DD,$5E,$EA,$F0,$70,$7F,$7C,$AF
	.dc.b	$F3,$57,$4F,$F8,$F2,$79,$DB,$AB,$FC,$83,$1F,$E5,$C3,$FC,$D5,$D1
	.dc.b	$FE,$5F,$77,$D9,$D7,$0F,$E7,$91,$FC,$C1,$FF,$3B,$71,$7F,$96,$23
	.dc.b	$AB,$FE,$7B,$EC,$FF,$3B,$C7,$F9,$D7,$C1,$FF,$3B,$BE,$2F,$0B,$FD
	.dc.b	$01,$E1,$7F,$9D,$63,$CB,$E1,$61,$FF,$3B,$10,$F4,$F8,$5F,$E8,$7F
	.dc.b	$0B,$FD,$21,$1F,$EC,$4F,$1F,$FD,$20,$FF,$B6,$7C,$7F,$F4,$7B,$FE
	.dc.b	$DB,$F1,$EE,$DF,$ED,$7F,$69,$E4,$7F,$A4,$7B,$25,$FF,$A2,$DF,$F4
	.dc.b	$A7,$91,$FE,$54,$F8,$59,$3F,$86,$7F,$CA,$FF,$0F,$FE,$58,$F8,$7F
	.dc.b	$F2,$CF,$C3,$4A,$FC,$5C,$7F,$97,$3E,$1E,$47,$E3,$9F,$F2,$FF,$C1
	.dc.b	$74,$8F,$F2,$7F,$C1,$FF,$98,$8C,$6D,$FE,$5B,$BE,$93,$FE,$54,$C3
	.dc.b	$C0,$F3,$21,$FE,$58,$C3,$AB,$E7,$3F,$3B,$7D,$5C,$1F,$E5,$EA,$EB
	.dc.b	$7F,$E6,$9B,$EB,$3C,$CE,$B1,$FE,$6D,$BE,$B1,$F9,$6D,$9E,$DF,$FB
	.dc.b	$96,$30,$FF,$BA,$5F,$F7,$19,$FE,$FA,$F6,$BF,$DC,$CF,$1F,$6B,$FD
	.dc.b	$CC,$FF,$BF,$BD,$AF,$F7,$33,$D7,$DA,$FF,$42,$FC,$1C,$8F,$DB,$BF
	.dc.b	$E8,$8F,$83,$FF,$44,$FC,$34,$9F,$DD,$BF,$E8,$CF,$87,$FF,$46,$FC
	.dc.b	$3F,$FA,$3B,$E1,$E5,$7F,$00,$8F,$F4,$8F,$C3,$48,$FE,$06,$F7,$7F
	.dc.b	$D0,$77,$D5,$A7,$F0,$8B,$EA,$FF,$0A,$BE,$AF,$FA,$2A,$FF,$D1,$8F
	.dc.b	$EF,$3D,$7A,$D3,$3C,$4E,$AF,$F4,$34,$EA,$31,$3F,$C8,$21,$F0,$DC
	.dc.b	$2B,$7F,$0D,$6B,$5F,$0D,$70,$60,$8C,$2A,$BF,$0D,$8F,$87,$98,$7F
	.dc.b	$27,$A4,$F0,$F8,$7E,$3C,$F2,$B5,$1D,$30,$EB,$8C,$9E,$51,$D9,$E9
	.dc.b	$8F,$67,$E2,$61,$E4,$E6,$F2,$F3,$C9,$F1,$47,$A3,$9A,$47,$95,$F5
	.dc.b	$F3,$FB,$3E,$1E,$4F,$A5,$F7,$FC,$3F,$C0,$F8,$69,$3E,$D7,$E1,$FC
	.dc.b	$3F,$48,$FA,$D8,$5E,$2F,$A2,$7C,$AC,$7A,$CF,$DB,$C6,$CE,$B8,$99
	.dc.b	$7C,$EF,$86,$97,$F2,$DF,$A3,$F0,$FD,$79,$7E,$FE,$76,$BE,$F6,$7D
	.dc.b	$27,$F0,$79,$D9,$FC,$16,$3F,$6B,$C9,$91,$F5,$DF,$BB,$E5,$7D,$EF
	.dc.b	$2A,$4F,$DB,$7F,$03,$E5,$7E,$17,$F2,$AF,$F5,$C7,$07,$B9,$79,$79
	.dc.b	$35,$A6,$FD,$09,$8F,$F6,$BA,$47,$3F,$47,$BC,$FE,$7F,$2B,$8F,$C5
	.dc.b	$CB,$9C,$F1,$CB,$B3,$CF,$2F,$27,$A3,$E5,$F4,$64,$F7,$8F,$47,$A3
	.dc.b	$E9,$F4,$64,$F8,$E3,$D9,$EA,$48,$F3,$3E,$FF,$57,$E0,$7A,$B2,$7D
	.dc.b	$4F,$C3,$F5,$7A,$C7,$B7,$0E,$AF,$F7,$4E,$26,$3D,$63,$18,$DD,$FE
	.dc.b	$56,$5F,$37,$DB,$F9,$DE,$DC,$BF,$91,$1F,$47,$DA,$91,$F2,$DF,$A9
	.dc.b	$EC,$FD,$5F,$7A,$4F,$CF,$7E,$C7,$C4,$7D,$9F,$7B,$ED,$7B,$F2,$BE
	.dc.b	$A4,$74,$7D,$18,$5D,$3E,$F6,$3D,$27,$D9,$8D,$9F,$C1,$63,$F8,$5F
	.dc.b	$82,$FE,$15,$F9,$5B,$E1,$29,$9C,$3D,$04,$3B,$9F,$85,$C7,$A7,$42
	.dc.b	$3F,$C7,$3C,$7F,$E2,$D7,$D8,$7E,$2F,$1E,$FF,$13,$23,$FC,$71,$8F
	.dc.b	$F8,$D5,$FF,$20,$AF,$85,$F2,$FC,$3F,$9B,$E1,$E5,$70,$8F,$4F,$C4
	.dc.b	$7A,$BC,$B9,$3D,$23,$DB,$F0,$BE,$EF,$85,$93,$E0,$8E,$AF,$8B,$0E
	.dc.b	$AF,$F8,$DB,$2F,$F3,$DB,$E4,$BE,$AE,$70,$F0,$BE,$AF,$83,$F9,$7F
	.dc.b	$07,$23,$07,$BB,$E0,$FE,$79,$1F,$83,$F0,$BF,$4B,$E1,$7E,$9F,$C2
	.dc.b	$49,$F8,$B1,$F5,$BE,$13,$EB,$FC,$1C,$9F,$93,$1C,$DF,$46,$17,$AB
	.dc.b	$F8,$EE,$CC,$BB,$57,$D0,$CB,$B3,$EF,$CB,$F0,$7F,$07,$23,$EA,$BF
	.dc.b	$86,$F8,$2F,$C3,$FC,$1F,$E2,$3E,$12,$57,$DA,$8F,$EA,$9F,$96,$7F
	.dc.b	$5D,$EC,$FC,$26,$A8,$C5,$7F,$9F,$9B,$19,$F8,$5E,$3F,$05,$CB,$E0
	.dc.b	$E5,$FF,$9F,$6F,$FA,$E3,$DC,$FC,$4E,$5E,$27,$EC,$65,$E2,$61,$C4
	.dc.b	$F2,$E1,$E6,$F8,$29,$1D,$5F,$4F,$C1,$FB,$A3,$A9,$0E,$A7,$32,$1F
	.dc.b	$E7,$CC,$3D,$67,$F7,$A7,$D1,$F0,$9F,$13,$E1,$64,$60,$F6,$7C,$27
	.dc.b	$C8,$23,$EF,$F8,$5F,$97,$F0,$32,$3E,$13,$F3,$BE,$07,$E7,$FC,$0F
	.dc.b	$D0,$F8,$19,$5F,$1E,$39,$BE,$5C,$2F,$1F,$F6,$0D,$19,$75,$AF,$56
	.dc.b	$56,$B5,$97,$DC,$F8,$59,$5F,$4D,$FB,$FF,$0B,$F8,$0F,$85,$FC,$0F
	.dc.b	$C2,$CA,$FB,$11,$F8,$4F,$85,$91,$F6,$DF,$C3,$FC,$2F,$E2,$3E,$0F
	.dc.b	$F1,$3F,$07,$2B,$F0,$11,$D5,$FC,$16,$1D,$5F,$A1,$87,$47,$E9,$61
	.dc.b	$74,$F7,$63,$FE,$27,$F7,$7F,$B1,$F6,$FE,$51,$B7,$FE,$D0,$6D,$7F
	.dc.b	$42,$F8,$3E,$1F,$0B,$C7,$E1,$65,$FF,$B3,$EF,$FB,$2B,$7E,$13,$B7
	.dc.b	$C2,$C8,$C1,$B9,$F1,$11,$E5,$F0,$7D,$9E,$97,$D9,$FF,$63,$CF,$66
	.dc.b	$DF,$57,$37,$ED,$8F,$77,$C1,$FB,$FE,$0E,$57,$9E,$3B,$3D,$30,$ED
	.dc.b	$A7,$F8,$8B,$1E,$CD,$C3,$B3,$9C,$7B,$5F,$C1,$F8,$5F,$98,$47,$E2
	.dc.b	$7C,$2F,$CF,$F8,$49,$1F,$1D,$FA,$5F,$0B,$F4,$FE,$0E,$4F,$CB,$7E
	.dc.b	$B7,$C0,$7D,$7F,$83,$FB,$1F,$01,$2B,$E8,$47,$47,$E2,$E1,$74,$F5
	.dc.b	$63,$D2,$7E,$4E,$36,$7D,$D8,$FE,$0F,$E1,$3F,$09,$F0,$B2,$FE,$E4
	.dc.b	$7E,$1F,$E1,$7F,$CD,$DF,$96,$FF,$65,$9F,$E3,$7D,$3F,$C7,$23,$11
	.dc.b	$F5,$A7,$7F,$09,$C3,$E1,$24,$F0,$19,$AF,$FB,$03,$7E,$13,$BF,$C1
	.dc.b	$4A,$E1,$39,$7F,$21,$8F,$93,$E1,$A4,$60,$DE,$1D,$0F,$ED,$4E,$31
	.dc.b	$F2,$BA,$F8,$3F,$6F,$C1,$C8,$C1,$CB,$E1,$3E,$0F,$C2,$C8,$C1,$DB
	.dc.b	$E1,$7E,$29,$1F,$0F,$C2,$FC,$8F,$83,$91,$E4,$7E,$5F,$C1,$7C,$CC
	.dc.b	$3A,$4F,$C4,$E3,$F9,$98,$E6,$3B,$10,$D1,$F2,$B5,$65,$F5,$B2,$FA
	.dc.b	$FF,$03,$F6,$3E,$0A,$5F,$C4,$8E,$8F,$2E,$2F,$A5,$8F,$89,$FF,$41
	.dc.b	$E2,$CF,$D9,$5B,$8E,$EF,$8E,$77,$8F,$E1,$BE,$13,$F0,$FF,$0B,$2B
	.dc.b	$E9,$C7,$E2,$BE,$13,$F1,$7F,$09,$27,$EB,$C7,$E3,$BE,$06,$47,$DA
	.dc.b	$7F,$23,$F0,$5F,$EB,$FF,$CB,$BF,$B3,$92,$71,$F0,$7D,$5F,$C0,$E1
	.dc.b	$D5,$F4,$61,$D1,$F5,$61,$74,$F7,$E3,$D7,$E1,$65,$7E,$43,$2F,$ED
	.dc.b	$EF,$C7,$7F,$D9,$CE,$7E,$13,$C9,$F0,$92,$30,$71,$F8,$4F,$41,$1E
	.dc.b	$9F,$09,$EB,$F8,$2F,$67,$C0,$C9,$F0,$47,$BF,$0E,$91,$E1,$C3,$0F
	.dc.b	$52,$1E,$5C,$3E,$36,$1F,$1F,$E0,$7E,$47,$C0,$CA,$F6,$47,$CB,$F8
	.dc.b	$39,$1E,$F7,$E7,$7C,$27,$CF,$F8,$39,$3F,$0D,$FA,$5F,$07,$F4,$FE
	.dc.b	$02,$4F,$C7,$7E,$B7,$C0,$FD,$7F,$81,$93,$F2,$DF,$B5,$F0,$3F,$6F
	.dc.b	$E0,$A4,$FC,$F7,$EF,$7C,$1F,$FB,$F7,$F2,$FF,$EA,$64,$F1,$F1,$F0
	.dc.b	$9D,$2F,$EA,$63,$7A,$FF,$DB,$7D,$2C,$FA,$D7,$D9,$CE,$D6,$B3,$ED
	.dc.b	$F0,$92,$BF,$87,$E9,$FD,$0B,$4F,$F1,$CA,$F8,$1E,$D1,$FE,$3C,$CD
	.dc.b	$A6,$2F,$F8,$EC,$78,$5F,$61,$0F,$11,$F8,$DC,$7D,$DF,$01,$EF,$F8
	.dc.b	$09,$3E,$28,$F8,$5F,$03,$23,$CA,$FC,$5F,$84,$F8,$DF,$07,$27,$D2
	.dc.b	$FC,$9F,$84,$F9,$5F,$09,$F2,$FE,$12,$57,$BA,$3E,$77,$C1,$7C,$FF
	.dc.b	$82,$93,$F0,$A3,$E9,$7C,$17,$D3,$F8,$29,$3F,$1A,$3E,$B7,$C1,$C8
	.dc.b	$F9,$2F,$D9,$F8,$4F,$B5,$F0,$7F,$6F,$E1,$25,$7C,$E8,$FB,$DF,$03
	.dc.b	$FE,$A9,$8F,$E2,$9E,$3F,$F4,$CE,$C7,$4F,$F9,$AA,$30,$FD,$4A,$E1
	.dc.b	$3A,$AF,$E9,$B5,$FD,$30,$9F,$09,$AB,$E5,$F0,$36,$B7,$F0,$97,$06
	.dc.b	$1F,$EA,$06,$C6,$61,$F0,$97,$E0,$E5,$FF,$56,$94,$6F,$E0,$F8,$7B
	.dc.b	$BC,$7D,$E9,$7B,$8E,$9F,$03,$D7,$DE,$93,$CE,$3C,$1F,$07,$23,$B3
	.dc.b	$E3,$F8,$3F,$27,$C1,$49,$F1,$3E,$7F,$83,$F4,$7C,$0C,$9F,$33,$EB
	.dc.b	$F8,$2F,$67,$C0,$49,$F5,$3E,$FF,$81,$F8,$1F,$03,$F0,$7E,$06,$57
	.dc.b	$BE,$3E,$47,$E6,$1F,$E9,$A2,$46,$3E,$0F,$3F,$07,$AF,$87,$93,$FD
	.dc.b	$F2,$BF,$96,$F5,$7D,$90,$BE,$0B,$A9,$27,$87,$C1,$78,$3E,$0B,$C3
	.dc.b	$D7,$27,$A4,$79,$3E,$0F,$CB,$F0,$92,$7C,$11,$E8,$F8,$3F,$4F,$C2
	.dc.b	$49,$F2,$47,$B3,$E1,$3D,$BF,$09,$27,$D1,$1F,$03,$E1,$24,$7A,$DF
	.dc.b	$87,$F0,$9F,$13,$E1,$3E,$2F,$C2,$4A,$F8,$11,$F2,$3D,$EF,$93,$F0
	.dc.b	$72,$7E,$24,$7C,$CF,$83,$91,$F1,$DF,$9F,$EF,$FE,$5F,$F3,$1F,$F3
	.dc.b	$B9,$3C,$EC,$7C,$2E,$7E,$17,$5F,$09,$31,$FF,$15,$DF,$F1,$D7,$FC
	.dc.b	$66,$BE,$0B,$A7,$C1,$C9,$DC,$77,$F8,$39,$1C,$9F,$17,$C1,$F8,$FE
	.dc.b	$0F,$C9,$F0,$72,$BB,$C7,$9F,$E0,$A4,$78,$9F,$57,$C1,$FA,$FE,$0B
	.dc.b	$D9,$F0,$72,$BC,$F1,$EF,$F8,$2F,$81,$F0,$72,$7D,$71,$F0,$FE,$0F
	.dc.b	$E2,$7C,$1C,$9F,$7C,$7C,$7F,$82,$F9,$1F,$07,$27,$E1,$C7,$CB,$F8
	.dc.b	$3F,$E0,$5F,$99,$7F,$0E,$9F,$F2,$BB,$7C,$D7,$32,$B7,$F0,$9C,$3E
	.dc.b	$12,$61,$FE,$58,$C3,$FC,$BA,$7F,$0B,$27,$C2,$76,$F8,$4E,$FF,$09
	.dc.b	$E0,$F8,$49,$7C,$63,$C7,$F0,$7E,$4F,$84,$93,$D6,$3C,$FF,$07,$23
	.dc.b	$C0,$FA,$BE,$13,$D7,$F0,$9E,$CF,$84,$95,$E5,$8F,$7F,$C2,$7C,$0F
	.dc.b	$84,$93,$E9,$8F,$87,$F0,$9F,$13,$E1,$24,$FB,$63,$E3,$FC,$27,$C8
	.dc.b	$F8,$49,$3F,$06,$3E,$5F,$C1,$C8,$F8,$8F,$CE,$F8,$4F,$E6,$1B,$FF
	.dc.b	$41,$9F,$F2,$A3,$7A,$BF,$82,$37,$CE,$4F,$FA,$2A,$1F,$09,$CB,$E1
	.dc.b	$24,$FF,$A2,$27,$F8,$6D,$70,$BF,$F2,$E7,$C2,$78,$3E,$0F,$C3,$F0
	.dc.b	$73,$0E,$11,$E4,$F8,$2F,$E2,$98,$71,$C7,$FA,$2B,$D6,$33,$5E,$3F
	.dc.b	$83,$B9,$7F,$9A,$7E,$0A,$E0,$F8,$3B,$18,$73,$F0,$5A,$F8,$3B,$9F
	.dc.b	$A5,$E3,$F0,$52,$3D,$9F,$07,$74,$F8,$FF,$07,$D8,$EF,$F0,$76,$BE
	.dc.b	$17,$C1,$78,$BE,$0F,$C6,$D8,$F8,$DF,$05,$E6,$F8,$3F,$3B,$63,$E5
	.dc.b	$7C,$1F,$A8,$F5,$FC,$1D,$8F,$F5,$AF,$E6,$7F,$DD,$9A,$7C,$25,$F0
	.dc.b	$96,$31,$F0,$99,$F8,$4B,$1C,$9E,$1F,$07,$23,$AF,$C2,$5E,$1F,$F4
	.dc.b	$C6,$2F,$84,$EC,$77,$F8,$4B,$7E,$4F,$83,$F1,$1E,$3F,$84,$B1,$E8
	.dc.b	$F8,$4F,$31,$E7,$F8,$4B,$1E,$CF,$83,$F5,$1E,$BF,$84,$B1,$F0,$3E
	.dc.b	$0F,$DC,$7B,$FE,$12,$C7,$C4,$F8,$3F,$84,$7C,$3F,$84,$B1,$F2,$3E
	.dc.b	$0F,$E3,$1F,$1F,$E1,$2C,$7C,$CF,$83,$FC,$5F,$E6,$8F,$8A,$F9,$91
	.dc.b	$87,$3F,$0B,$2B,$FD,$CF,$4F,$85,$C7,$3F,$FA,$FE,$1F,$09,$D3,$E1
	.dc.b	$3A,$B6,$F8,$FC,$1F,$83,$E1,$3C,$2D,$8E,$BE,$7F,$90,$F2,$FC,$25
	.dc.b	$8F,$0F,$C2,$7A,$0F,$4F,$C2,$58,$F2,$FC,$1F,$B0,$F6,$FC,$25,$8F
	.dc.b	$4F,$C1,$FC,$03,$E0,$FC,$25,$8F,$6F,$C1,$FC,$4F,$84,$F8,$AD,$8F
	.dc.b	$83,$F0,$7F,$20,$F9,$3F,$09,$63,$E2,$FC,$27,$CC,$F8,$4F,$9A,$D8
	.dc.b	$FE,$0C,$7F,$07,$FC,$D7,$F8,$7B,$8E,$77,$BC,$9F,$E2,$33,$98,$C4
	.dc.b	$FF,$02,$9C,$BF,$EE,$F5,$F0,$99,$35,$F0,$96,$F9,$4F,$F8,$85,$BF
	.dc.b	$09,$CF,$E1,$3A,$3D,$7E,$12,$4F,$0F,$84,$B9,$77,$F8,$4F,$17,$C2
	.dc.b	$78,$DB,$5D,$BE,$13,$CC,$79,$FE,$12,$C7,$8B,$E1,$3D,$47,$AF,$E1
	.dc.b	$2C,$79,$BE,$13,$DC,$7B,$FE,$12,$C7,$AB,$E0,$FE,$11,$F0,$FE,$12
	.dc.b	$C7,$BB,$E0,$FE,$31,$F1,$FE,$12,$C7,$C2,$F8,$4F,$95,$F0,$9F,$2D
	.dc.b	$B1,$F3,$FD,$DF,$E2,$1C,$5F,$52,$3F,$9F,$47,$F1,$07,$47,$F3,$47
	.dc.b	$11,$B3,$ED,$AF,$83,$D1,$BF,$83,$B5,$DF,$E0,$F9,$7C,$1F,$36,$C7
	.dc.b	$8F,$E0,$BB,$1D,$F0,$B1,$E7,$F8,$2F,$11,$E3,$F8,$2B,$1E,$BF,$82
	.dc.b	$F3,$7C,$17,$9D,$B1,$EF,$F8,$2F,$57,$C1,$7A,$DB,$1F,$0F,$E0,$BD
	.dc.b	$DF,$05,$EF,$6C,$7C,$7F,$82,$F8,$47,$C3,$F8,$2B,$1F,$2F,$E0,$BE
	.dc.b	$37,$C1,$7C,$76,$C7,$CF,$F8,$2F,$EB,$BF,$9B,$7F,$13,$8C,$1F,$A1
	.dc.b	$71,$CF,$F3,$7C,$D4,$3E,$0B,$8F,$C1,$72,$6D,$F1,$F8,$0E,$A7,$6F
	.dc.b	$81,$B1,$D7,$E0,$3C,$3F,$03,$E2,$6C,$78,$7E,$07,$CB,$F0,$3E,$66
	.dc.b	$C7,$97,$E0,$7D,$27,$AB,$E0,$6C,$7A,$7E,$07,$DB,$F0,$3E,$E6,$C7
	.dc.b	$B7,$E0,$7E,$0F,$C0,$FC,$26,$C7,$52,$71,$FC,$1E,$26,$3D,$A3,$18
	.dc.b	$74,$6D,$7C,$9C,$7E,$6F,$C0,$48,$F9,$9F,$03,$6B,$A7,$C0,$FD,$23
	.dc.b	$E9,$FC,$0D,$AF,$A5,$F0,$1F,$58,$FA,$FF,$03,$63,$F8,$B7,$E6,$FF
	.dc.b	$C7,$30,$FA,$CD,$EA,$FF,$59,$3F,$C6,$1B,$1F,$E1,$C8,$FE,$33,$3C
	.dc.b	$BE,$0B,$1F,$05,$97,$5F,$03,$BF,$81,$B5,$E0,$79,$7C,$0C,$8F,$1F
	.dc.b	$C0,$DC,$BC,$78,$73,$9F,$3E,$36,$7C,$F8,$F5,$9F,$5B,$63,$D7,$F0
	.dc.b	$1E,$7F,$81,$F4,$36,$BE,$0F,$C0,$FA,$FE,$07,$D8,$D8,$F8,$BF,$03
	.dc.b	$EF,$3E,$07,$C0,$D8,$F9,$3F,$03,$F0,$FE,$07,$E2,$36,$3A,$9F,$3B
	.dc.b	$0E,$AD,$3A,$93,$3F,$A3,$7D,$5F,$A7,$F0,$3F,$3C,$FA,$1F,$03,$6B
	.dc.b	$AC,$FD,$8B,$EA,$D8,$E9,$CB,$F6,$B3,$F4,$C6,$0F,$F3,$17,$2B,$E0
	.dc.b	$7E,$E1,$F7,$7E,$06,$CF,$E1,$BE,$03,$F0,$1F,$03,$F8,$16,$C7,$F5
	.dc.b	$FF,$CE,$3F,$98,$E5,$FE,$F9,$7F,$96,$B6,$3F,$CC,$71,$87,$FC,$E4
	.dc.b	$FF,$2F,$79,$7C,$0E,$BE,$0B,$6F,$78,$FC,$BE,$17,$5F,$CB,$BD,$23
	.dc.b	$AF,$C3,$D7,$C8,$F8,$3E,$0A,$47,$9F,$E0,$ED,$FF,$15,$CB,$FC,$95
	.dc.b	$3E,$BF,$83,$F3,$FC,$1F,$A1,$F4,$FC,$1C,$9F,$81,$F0,$77,$0E,$5F
	.dc.b	$05,$EE,$3D,$FF,$07,$6B,$A1,$27,$C1,$8F,$58,$F9,$58,$73,$E1,$F9
	.dc.b	$58,$D9,$F8,$FF,$05,$F2,$8F,$97,$F0,$76,$3E,$97,$C1,$7C,$E3,$E7
	.dc.b	$FC,$1D,$8F,$AD,$F0,$7F,$4B,$E0,$FE,$9B,$63,$ED,$7C,$17,$D6,$F8
	.dc.b	$3F,$AE,$D8,$EA,$4D,$FE,$46,$1D,$5F,$95,$7C,$DB,$3F,$85,$C3,$F0
	.dc.b	$3F,$07,$FE,$A4,$F1,$FF,$A2,$EF,$F8,$78,$C7,$C3,$73,$8F,$C5,$6B
	.dc.b	$EA,$C3,$C1,$1F,$8D,$CB,$B3,$63,$B1,$FE,$31,$C3,$FA,$73,$7E,$0F
	.dc.b	$C1,$F0,$9E,$17,$C5,$F0,$9E,$3F,$84,$B5,$C5,$F3,$7C,$1F,$9F,$E1
	.dc.b	$2C,$F5,$7D,$5F,$07,$EB,$F8,$4B,$3E,$17,$9F,$17,$F6,$1D,$18,$E3
	.dc.b	$55,$FF,$B4,$E7,$2F,$A7,$2F,$03,$4F,$01,$ED,$F8,$4F,$95,$F0,$3F
	.dc.b	$2D,$B1,$F0,$7E,$07,$E7,$1F,$3F,$E0,$6C,$7C,$5F,$83,$FA,$5F,$09
	.dc.b	$F4,$DB,$1F,$27,$E0,$FE,$B1,$F5,$FE,$0E,$C7,$52,$6F,$ED,$C3,$AB
	.dc.b	$EF,$BE,$6D,$9F,$A9,$87,$E0,$7E,$0E,$47,$D7,$F8,$3B,$5F,$1F,$E0
	.dc.b	$FF,$0C,$7E,$1F,$E0,$ED,$7F,$9B,$7F,$39,$FE,$AC,$FF,$BE,$DF,$EA
	.dc.b	$CD,$AF,$F7,$8C,$61,$FF,$3B,$BF,$D6,$1F,$F0,$A3,$78,$FE,$9C,$6B
	.dc.b	$E0,$38,$FC,$0F,$82,$3C,$97,$F8,$DC,$31,$7C,$06,$1E,$4A,$EF,$F1
	.dc.b	$3E,$27,$C6,$F9,$3E,$0A,$61,$EB,$F8,$5B,$5D,$89,$87,$5E,$86,$5E
	.dc.b	$0B,$F7,$1F,$0F,$1E,$CD,$3E,$37,$C2,$FC,$23,$E1,$FC,$2D,$8F,$95
	.dc.b	$7F,$18,$F8,$FF,$0B,$63,$E7,$7C,$2F,$CA,$F8,$0F,$96,$D8,$FA,$5F
	.dc.b	$01,$F3,$BE,$0F,$E7,$B6,$3A,$93,$6F,$9F,$87,$57,$E8,$DF,$36,$CF
	.dc.b	$DD,$C3,$EC,$FC,$24,$8F,$C0,$7C,$2D,$AF,$B3,$F0,$BF,$78,$FB,$FF
	.dc.b	$0B,$6B,$F0,$DF,$09,$F8,$2F,$84,$FC,$1B,$63,$F1,$5F,$09,$F8,$6F
	.dc.b	$82,$FC,$3B,$63,$F1,$D8,$73,$C7,$F0,$F8,$DE,$6F,$C0,$74,$FF,$69
	.dc.b	$BF,$CA,$67,$D9,$A7,$FB,$0F,$F3,$AF,$EC,$AB,$0F,$F6,$4C,$61,$BF
	.dc.b	$07,$F8,$B9,$D6,$0A,$D3,$7E,$17,$57,$CB,$E1,$6C,$E6,$FC,$11,$8B
	.dc.b	$7C,$2D,$F8,$5C,$36,$BC,$DF,$09,$BF,$82,$E0,$D8,$F5,$67,$CF,$1E
	.dc.b	$8D,$8F,$75,$F3,$CF,$FD,$01,$97,$82,$7E,$16,$5E,$3F,$82,$91,$F1
	.dc.b	$7E,$0A,$DF,$0C,$7A,$36,$7D,$67,$63,$99,$5E,$F7,$EF,$F8,$4F,$81
	.dc.b	$F0,$96,$BE,$93,$F0,$FE,$12,$47,$D5,$F8,$4B,$87,$A3,$DE,$F9,$1F
	.dc.b	$05,$F2,$5B,$5F,$6F,$DF,$F9,$9F,$05,$F3,$5B,$1F,$7F,$1E,$7C,$BF
	.dc.b	$C1,$78,$59,$E2,$BE,$EC,$31,$EB,$9F,$86,$7F,$11,$9F,$D9,$F8,$39
	.dc.b	$1F,$8B,$F8,$4B,$5F,$3B,$E1,$3E,$F1,$F8,$2F,$CE,$FF,$B6,$D7,$68
	.dc.b	$C7,$C1,$C8,$FC,$69,$F8,$E9,$7C,$26,$6F,$FD,$71,$0F,$84,$E8,$75
	.dc.b	$F8,$1B,$7F,$9D,$C7,$C2,$F4,$B7,$03,$0F,$1C,$7E,$82,$BB,$4F,$F2
	.dc.b	$99,$FE,$EF,$3E,$0C,$73,$7E,$5F,$81,$F6,$1E,$DF,$84,$B1,$DF,$0E
	.dc.b	$DF,$09,$F0,$4F,$85,$F0,$96,$75,$93,$E0,$FE,$37,$C2,$7C,$76,$CE
	.dc.b	$BE,$13,$B3,$4E,$39,$73,$D7,$E2,$E5,$DA,$7A,$BC,$DE,$F8,$59,$F9
	.dc.b	$6F,$D5,$F8,$29,$3E,$4F,$82,$B5,$D2,$39,$9E,$8D,$3C,$1A,$FC,$7D
	.dc.b	$3C,$F5,$F5,$9B,$1F,$2F,$E0,$BF,$05,$F0,$9F,$83,$6E,$1F,$0B,$E1
	.dc.b	$3F,$0C,$7E,$1F,$E1,$2C,$7C,$6F,$84,$FC,$51,$F8,$BF,$84,$B1,$F2
	.dc.b	$BE,$03,$A3,$4F,$9B,$87,$2D,$3F,$13,$8F,$59,$F0,$1D,$49,$1F,$8E
	.dc.b	$BE,$AF,$E2,$23,$F9,$0E,$BF,$DA,$F4,$34,$FE,$F8,$E2,$7F,$90,$C7
	.dc.b	$F7,$B9,$E3,$F0,$98,$F8,$4C,$BA,$F8,$49,$3F,$89,$A7,$C2,$63,$0F
	.dc.b	$11,$F9,$A8,$EE,$4F,$83,$EF,$F0,$9D,$6F,$96,$36,$F8,$E7,$E8,$8E
	.dc.b	$8F,$F8,$1F,$06,$5E,$8C,$B3,$87,$C0,$F8,$4F,$5F,$C2,$7B,$1B,$1E
	.dc.b	$7F,$82,$F7,$9F,$03,$E0,$AC,$7A,$FE,$0B,$E1,$FC,$1F,$C4,$6C,$7B
	.dc.b	$EF,$9E,$DE,$6D,$11,$8D,$7C,$D9,$74,$AD,$65,$67,$D5,$8F,$49,$E4
	.dc.b	$D8,$F7,$5F,$A7,$1F,$9D,$F0,$96,$FC,$18,$F3,$9F,$BF,$8D,$9F,$A1
	.dc.b	$8F,$39,$FC,$1B,$63,$EA,$7C,$27,$E0,$0F,$C0,$FC,$25,$AF,$B9,$F0
	.dc.b	$3F,$84,$F8,$1F,$C2,$B6,$3F,$01,$F0,$9F,$88,$F8,$1F,$C4,$B6,$3F
	.dc.b	$09,$F0,$7F,$8C,$3F,$1B,$F0,$96,$3F,$88,$7E,$79,$FD,$D0,$9F,$09
	.dc.b	$FD,$C9,$C7,$BD,$7E,$0B,$0D,$AE,$A4,$DF,$E9,$61,$D5,$FC,$4D,$F5
	.dc.b	$9E,$98,$75,$79,$9E,$2C,$2D,$70,$7C,$5F,$05,$E3,$F8,$0B,$5E,$87
	.dc.b	$CD,$F0,$72,$3D,$7F,$09,$70,$F0,$FC,$27,$AC,$F6,$7C,$25,$AF,$83
	.dc.b	$F0,$9E,$FF,$84,$F8,$0D,$8F,$8B,$F0,$5F,$0F,$E0,$BE,$23,$63,$E4
	.dc.b	$FC,$0F,$C7,$3E,$47,$C0,$D8,$F9,$BF,$07,$F2,$FE,$13,$E6,$36,$3F
	.dc.b	$A7,$6A,$FB,$97,$57,$D1,$F8,$4F,$F1,$2C,$E6,$31,$5F,$63,$E1,$3E
	.dc.b	$B1,$F5,$FE,$06,$C7,$F4,$BF,$CF,$7F,$AB,$F9,$06,$9F,$D5,$9F,$F4
	.dc.b	$03,$FD,$73,$0E,$D3,$F8,$4C,$7C,$39,$79,$23,$07,$5F,$82,$FF,$1B
	.dc.b	$DF,$F6,$D8,$C1,$E1,$F8,$4C,$FC,$24,$8F,$27,$C2,$5C,$BF,$8A,$9E
	.dc.b	$03,$F0,$F9,$78,$36,$F7,$C7,$F1,$2C,$5F,$05,$E0,$F8,$2F,$0B,$71
	.dc.b	$F8,$1E,$7F,$93,$DF,$F2,$B6,$3E,$7E,$FE,$BC,$7B,$C6,$32,$EF,$C1
	.dc.b	$F1,$3E,$0B,$DB,$F0,$7E,$E6,$D7,$CD,$F8,$3F,$83,$F0,$7F,$09,$B1
	.dc.b	$CC,$F8,$F8,$FB,$B2,$F0,$63,$FE,$AF,$F1,$DC,$3F,$2B,$DF,$98,$71
	.dc.b	$C3,$E0,$47,$CE,$F8,$3F,$9F,$F0,$16,$BE,$F3,$F1,$DE,$F8,$FF,$B0
	.dc.b	$AF,$E6,$3D,$F0,$D4,$74,$9E,$D3,$F8,$2C,$32,$F1,$C9,$E6,$E2,$BF
	.dc.b	$D0,$17,$F9,$78,$C3,$FE,$26,$CF,$D7,$3D,$0F,$F4,$0B,$88,$F8,$5C
	.dc.b	$1E,$E6,$CF,$E5,$F1,$E7,$1D,$63,$93,$FE,$8F,$C3,$C7,$18,$C5,$87
	.dc.b	$57,$D1,$87,$DA,$9F,$3D,$7D,$07,$FD,$27,$CA,$C7,$D3,$7F,$A5,$7C
	.dc.b	$35,$DB,$1F,$29,$F6,$72,$DE,$5D,$F1,$EB,$97,$7A,$F3,$3E,$DC,$7C
	.dc.b	$73,$ED,$3D,$73,$ED,$3A,$C7,$CF,$BE,$B9,$CA,$F7,$67,$E7,$6E,$7E
	.dc.b	$6A,$91,$F0,$F2,$F2,$63,$E2,$8F,$C1,$9E,$63,$C4,$F8,$1F,$87,$5C
	.dc.b	$A3,$07,$CF,$AF,$13,$8C,$7E,$7E,$3F,$23,$1F,$24,$E5,$B3,$F0,$23
	.dc.b	$E4,$61,$F0,$33,$F9,$93,$93,$F2,$91,$E9,$BE,$98,$C8,$E1,$8F,$96
	.dc.b	$B8,$47,$E3,$2B,$CB,$18,$C3,$FC,$EF,$8F,$8A,$BE,$04,$7E,$7C,$F6
	.dc.b	$9F,$46,$78,$36,$37,$1E,$48,$F8,$19,$7C,$C7,$F5,$73,$F2,$0F,$A6
	.dc.b	$79,$88,$F0,$BF,$2C,$74,$38,$47,$C7,$C7,$D9,$19,$8F,$A0,$FC,$32
	.dc.b	$3B,$7B,$BE,$AC,$BE,$5C,$71,$8F,$51,$F8,$13,$F8,$E9,$E3,$78,$E1
	.dc.b	$F1,$CF,$D0,$3E,$87,$E1,$CF,$49,$E5,$1F,$0C,$E1,$87,$C3,$9F,$D6
	.dc.b	$4F,$D0,$BF,$B0,$F9,$72,$F6,$BC,$B0,$FB,$07,$B3,$0F,$16,$5F,$5E
	.dc.b	$74,$73,$9F,$21,$ED,$C3,$C9,$1F,$48,$FE,$21,$97,$CE,$78,$DF,$B1
	.dc.b	$F9,$07,$8E,$B9,$BA,$CB,$D5,$9F,$EE,$63,$A5,$71,$3C,$B9,$FA,$A3
	.dc.b	$C0,$7C,$E3,$75,$CA,$FF,$5F,$1E,$2C,$BD,$8F,$46,$E5,$F2,$23,$93
	.dc.b	$4F,$69,$F3,$30,$FD,$7D,$7C,$73,$F0,$CF,$95,$B5,$D3,$1F,$EE,$F9
	.dc.b	$F9,$EF,$96,$3E,$5C,$7F,$07,$8F,$B3,$1F,$C1,$9F,$36,$F8,$61,$F0
	.dc.b	$F0,$E3,$9F,$DC,$BE,$18,$7F,$73,$D7,$FA,$9E,$3F,$E0,$F6,$C7,$E3
	.dc.b	$C9,$3C,$09,$3F,$CC,$B6,$FE,$CF,$3F,$DF,$5B,$3F,$8E,$BF,$B8,$FB
	.dc.b	$E7,$F0,$87,$DC,$7F,$C3,$3C,$5F,$DF,$63,$FC,$0C,$D8,$E5,$7F,$26
	.dc.b	$7F,$82,$F0,$FF,$7A,$7F,$C1,$8D,$9F,$F1,$1F,$3F,$F8,$55,$FF,$12
	.dc.b	$F3,$DB,$FA,$3B,$FC,$B9,$FD,$24,$7F,$4C,$E6,$FF,$09,$63,$FE,$17
	.dc.b	$6C,$FF,$8B,$BE,$23,$FC,$4E,$7F,$8A,$3E,$1A,$DF,$96,$FF,$17,$AF
	.dc.b	$F8,$B3,$AF,$FC,$5B,$D7,$6B,$FC,$58,$7F,$93,$0F,$F3,$CF,$7F,$F9
	.dc.b	$31,$A7,$3E,$3F,$B2,$FF,$91,$C9,$87,$F9,$D3,$B7,$FC,$8E,$FF,$9D
	.dc.b	$BC,$4F,$F3,$03,$67,$FD,$0B,$E2,$FF,$98,$1F,$F4,$3F,$8B,$67,$FC
	.dc.b	$FA,$FF,$97,$DF,$F3,$FF,$89,$FE,$62,$9F,$F3,$29,$FE,$8E,$F2,$FF
	.dc.b	$CF,$4D,$3F,$CF,$47,$FA,$4B,$CA,$FF,$3C,$71,$7F,$89,$7C,$AB,$3F
	.dc.b	$E0,$AF,$2B,$FC,$F1,$8F,$FA,$33,$CA,$B7,$FE,$5A,$F8,$5F,$F2,$9F
	.dc.b	$C2,$FF,$95,$1B,$87,$F9,$8F,$E1,$3F,$CA,$FF,$0B,$FE,$58,$6C,$31
	.dc.b	$FF,$6F,$FE,$7D,$FE,$F7,$3F,$DE,$A5,$F8,$4E,$93,$FE,$6D,$C6,$D7
	.dc.b	$F9,$55,$E9,$1F,$E7,$1C,$2C,$FF,$96,$54,$77,$CA,$C9,$61,$DE,$74
	.dc.b	$6D,$C4,$FF,$B7,$FD,$AF,$F6,$33,$FE,$E1,$F6,$AD,$FF,$BD,$7D,$9F
	.dc.b	$F6,$34,$FF,$B9,$5B,$1F,$EF,$9F,$67,$FD,$8D,$3F,$EE,$96,$C7,$90
	.dc.b	$F1,$FC,$35,$CB,$C4,$DA,$F4,$64,$F8,$5C,$5F,$0B,$F4,$D6,$BE,$17
	.dc.b	$36,$F8,$5B,$18,$73,$F0,$D2,$F8,$E1,$F1,$3E,$13,$18,$FD,$6C,$F9
	.dc.b	$FC,$37,$46,$DF,$0C,$D9,$F5,$8C,$36,$FE,$FE,$5D,$32,$F8,$16,$CB
	.dc.b	$18,$7D,$1C,$BA,$E1,$E1,$8E,$D1,$EA,$7C,$19,$D9,$FC,$5B,$C9,$F8
	.dc.b	$F8,$FC,$2B,$C6,$3F,$80,$F8,$4F,$8B,$F0,$9F,$19,$B8,$FA,$FE,$13
	.dc.b	$E4,$9F,$2B,$E1,$AC,$7B,$FE,$1B,$E6,$F2,$FC,$E6,$C7,$C3,$C3,$E8
	.dc.b	$F2,$F4,$D7,$F3,$EF,$27,$F4,$38,$F5,$AF,$01,$D4,$F9,$98,$75,$69
	.dc.b	$F3,$F9,$7E,$E9,$F7,$BE,$1A,$C7,$D3,$E6,$FC,$0F,$AB,$F8,$26,$C7
	.dc.b	$D7,$F5,$BF,$0A,$7E,$1B,$E1,$6C,$7E,$0B,$C0,$FD,$4B,$4E,$BA,$7E
	.dc.b	$A5,$FD,$64,$6A,$3F,$77,$7D,$9B,$3F,$85,$C3,$AC,$7F,$00,$C3,$C1
	.dc.b	$3E,$33,$C1,$F0,$BF,$9C,$C0,$DF,$86,$C3,$9F,$86,$D7,$8D,$6F,$D0
	.dc.b	$F1,$F8,$6E,$5E,$3D,$9F,$63,$D7,$C7,$ED,$95,$9F,$80,$F3,$D3,$DC
	.dc.b	$7F,$07,$58,$67,$A7,$F7,$11,$8A,$F9,$18,$F8,$1A,$78,$0F,$99,$E3
	.dc.b	$FB,$4F,$0B,$83,$E8,$67,$EC,$69,$EC,$39,$E3,$E4,$C1,$83,$3F,$74
	.dc.b	$E6,$31,$5F,$6F,$CB,$F9,$87,$CD,$F2,$EC,$7D,$FF,$86,$FA,$1E,$67
	.dc.b	$D1,$6C,$7E,$0F,$CC,$FA,$9F,$0D,$F5,$5B,$1F,$87,$F8,$6F,$B1,$E8
	.dc.b	$FD,$96,$C7,$53,$F1,$98,$75,$69,$D4,$FC,$85,$F5,$69,$F9,$3F,$53
	.dc.b	$F0,$87,$E1,$7E,$1A,$C7,$E5,$FD,$2F,$C4,$1F,$C9,$3F,$3F,$7B,$18
	.dc.b	$9F,$E3,$B0,$F6,$35,$EC,$6D,$B3,$AF,$63,$91,$CF,$DC,$B1,$DC,$9C
	.dc.b	$BF,$4F,$0E,$F1,$8E,$7F,$E7,$91,$DE,$3C,$38,$F7,$69,$E4,$F7,$3D
	.dc.b	$1E,$DF,$A5,$B1,$E8,$F8,$7F,$61,$ED,$F7,$EC,$7B,$3E,$1F,$E0,$1F
	.dc.b	$07,$E0,$2C,$7C,$0F,$87,$F8,$87,$C5,$F8,$2B,$1D,$49,$B7,$AF,$0E
	.dc.b	$AF,$B7,$0E,$6F,$BF,$0B,$B7,$BB,$E1,$FE,$81,$F4,$7E,$1A,$E3,$D2
	.dc.b	$7E,$1E,$9E,$98,$C1,$FD,$D7,$81,$97,$48,$C6,$37,$2F,$EF,$14,$F8
	.dc.b	$5F,$BD,$F0,$FF,$7D,$B7,$F7,$BE,$13,$F0,$5F,$05,$F8,$36,$C7,$E0
	.dc.b	$BE,$13,$F0,$C7,$E1,$FE,$0E,$C7,$F6,$2F,$D0,$3F,$C2,$CD,$CF,$FB
	.dc.b	$7C,$3E,$16,$C7,$59,$FC,$6E,$36,$3F,$08,$F4,$7F,$0D,$87,$79,$F1
	.dc.b	$1D,$FE,$13,$AF,$C2,$F6,$6C,$78,$FE,$17,$C3,$F0,$7E,$26,$C7,$9F
	.dc.b	$E0,$FC,$BF,$07,$E6,$6C,$7A,$FE,$13,$D3,$F0,$9E,$A6,$C7,$53,$E0
	.dc.b	$61,$D5,$A7,$53,$E2,$5F,$56,$9F,$1F,$E0,$BE,$2F,$C0,$FC,$66,$C7
	.dc.b	$CB,$F8,$4F,$93,$F0,$9F,$29,$B1,$F3,$FE,$0F,$E6,$FC,$27,$CE,$6C
	.dc.b	$7D,$3F,$84,$FA,$3F,$09,$F4,$9B,$1D,$4F,$B1,$87,$56,$9D,$4F,$B9
	.dc.b	$7D,$5A,$7D,$FF,$83,$FB,$BF,$07,$F7,$9B,$1D,$79,$BF,$98,$74,$B0
	.dc.b	$ED,$18,$7A,$3A,$C2,$F8,$5F,$E3,$DA,$47,$E2,$BE,$0F,$F1,$7F,$09
	.dc.b	$71,$FC,$83,$FE,$3C,$F9,$FF,$ED,$98,$7F,$91,$5B,$3F,$E5,$B7,$FB
	.dc.b	$2B,$FE,$38,$73,$8F,$F9,$75,$7C,$2F,$3F,$80,$E8,$DA,$E1,$F0,$7D
	.dc.b	$FE,$0F,$C0,$D8,$E8,$78,$CF,$27,$C2,$D7,$C1,$5C,$F1,$FE,$A9,$8F
	.dc.b	$82,$7C,$98,$FA,$FE,$0E,$47,$9F,$E1,$6D,$F0,$F8,$4F,$80,$7C,$1F
	.dc.b	$85,$B5,$ED,$F8,$5F,$89,$F0,$BF,$15,$B1,$F0,$7E,$17,$E4,$1F,$27
	.dc.b	$E0,$AC,$7C,$5C,$39,$E3,$E6,$C6,$F1,$7C,$DC,$79,$CF,$A9,$B1,$F4
	.dc.b	$7E,$17,$EA,$7C,$1F,$D5,$6E,$3F,$47,$E1,$3E,$C7,$C2,$FD,$96,$C7
	.dc.b	$D5,$F8,$5F,$B9,$F0,$BF,$75,$B1,$F6,$7E,$17,$FC,$F7,$E5,$7F,$4A
	.dc.b	$D7,$E5,$C6,$3E,$17,$A5,$7D,$FC,$AD,$7D,$07,$A4,$7E,$0F,$1B,$3F
	.dc.b	$51,$EB,$F0,$BD,$BE,$0E,$D7,$FA,$03,$24,$74,$DF,$F2,$18,$DA,$FC
	.dc.b	$83,$D2,$3F,$29,$85,$9F,$CA,$3E,$AF,$85,$F5,$FC,$2D,$9E,$CF,$BB
	.dc.b	$E1,$24,$78,$7E,$16,$EB,$FE,$1D,$D1,$F0,$1F,$13,$E0,$FE,$2B,$6F
	.dc.b	$D1,$F0,$3F,$23,$E0,$7E,$4B,$63,$D9,$97,$4D,$BF,$D5,$D8,$B3,$C5
	.dc.b	$7F,$B1,$63,$BB,$F0,$F2,$EE,$D3,$E3,$7C,$2F,$D6,$3E,$BF,$C2,$D8
	.dc.b	$F9,$5F,$0B,$F6,$BE,$17,$ED,$B6,$3E,$77,$C1,$FD,$EF,$84,$FB,$ED
	.dc.b	$8F,$A5,$F0,$9F,$82,$3F,$07,$F0,$56,$3A,$9F,$5F,$0E,$AD,$3A,$9F
	.dc.b	$6E,$FA,$B4,$FB,$DF,$05,$F8,$E3,$F1,$FF,$09,$63,$FD,$83,$FA,$0F
	.dc.b	$F6,$0E,$6F,$F2,$73,$78,$BF,$DA,$CF,$F6,$19,$DC,$FF,$B4,$A1,$F0
	.dc.b	$74,$C7,$C2,$5A,$EB,$F0,$9B,$38,$7C,$15,$8E,$64,$C3,$C3,$F0,$BD
	.dc.b	$5E,$47,$93,$0E,$B3,$E8,$7C,$5F,$0B,$23,$D7,$F0,$96,$BD,$67,$38
	.dc.b	$F7,$DD,$8E,$91,$CC,$F8,$79,$78,$34,$F8,$3F,$0B,$EF,$3E,$07,$4D
	.dc.b	$AF,$F7,$4F,$C8,$39,$7F,$DE,$AD,$8C,$F2,$E2,$73,$C7,$27,$8F,$1D
	.dc.b	$D3,$E8,$FC,$2F,$CD,$F8,$6F,$9C,$DA,$FA,$FF,$0D,$F4,$4F,$A5,$F0
	.dc.b	$36,$3E,$DF,$C3,$7D,$5F,$81,$FA,$CD,$8F,$BF,$F0,$BF,$67,$E1,$BE
	.dc.b	$D3,$63,$A9,$33,$FC,$1E,$1D,$5F,$C2,$DF,$36,$CF,$E2,$B0,$FC,$27
	.dc.b	$C2,$FE,$17,$E1,$AC,$FE,$39,$FC,$47,$C2,$FE,$27,$E1,$6C,$FE,$49
	.dc.b	$2C,$3F,$D4,$1F,$9E,$7F,$5E,$E8,$34,$FE,$C7,$1A,$9E,$2F,$0F,$83
	.dc.b	$B5,$29,$F0,$58,$BE,$D3,$D2,$B1,$68,$EF,$5E,$68,$FD,$2E,$16,$7F
	.dc.b	$4A,$F0,$7C,$B7,$70,$F3,$7C,$25,$8C,$7C,$2C,$7C,$27,$8A,$5F,$0D
	.dc.b	$9C,$7F,$D8,$1D,$6F,$4F,$F1,$A7,$E3,$BD,$3B,$87,$E4,$BD,$0F,$C8
	.dc.b	$FA,$1F,$92,$6C,$76,$C3,$D9,$97,$FB,$19,$F9,$BF,$0B,$E7,$F8,$5F
	.dc.b	$43,$EB,$5E,$E5,$F5,$F0,$DC,$3F,$43,$ED,$FB,$FE,$0B,$F4,$2D,$8F
	.dc.b	$07,$1F,$D8,$F8,$2F,$88,$7C,$5F,$82,$B3,$AC,$9F,$09,$E1,$F8,$4F
	.dc.b	$13,$67,$9E,$EF,$85,$F9,$D3,$F3,$DB,$3A,$AF,$A5,$F0,$9F,$4D,$B1
	.dc.b	$DF,$83,$DC,$7B,$F1,$B1,$DB,$E1,$7E,$D7,$C2,$FD,$B6,$C7,$8B,$E0
	.dc.b	$BE,$F7,$17,$DF,$6C,$7A,$CF,$95,$D9,$F2,$DA,$7A,$B9,$3F,$0C,$7E
	.dc.b	$1F,$E1,$6C,$7C,$8F,$CF,$3F,$C0,$9C,$9F,$9A,$6C,$7E,$61,$FF,$67
	.dc.b	$73,$FE,$5E,$31,$CD,$F9,$3F,$93,$79,$34,$C7,$95,$6B,$B7,$C3,$7E
	.dc.b	$67,$1F,$F0,$1C,$78,$7B,$F1,$39,$EF,$93,$E5,$EE,$BA,$78,$4F,$D4
	.dc.b	$66,$F3,$7C,$47,$8F,$CE,$B7,$ED,$F0,$7C,$07,$87,$E1,$AC,$7C,$1F
	.dc.b	$84,$F2,$7C,$2F,$95,$B1,$F8,$CC,$5E,$F7,$BF,$DE,$F8,$0D,$9F,$95
	.dc.b	$F0,$DE,$DF,$86,$F7,$36,$3E,$77,$C3,$7C,$1F,$86,$F8,$4D,$8F,$CC
	.dc.b	$FC,$A3,$DA,$FD,$F2,$F6,$AC,$61,$F9,$5F,$0D,$F2,$FE,$1E,$DF,$DB
	.dc.b	$7E,$77,$C3,$FC,$FF,$7A,$CF,$DF,$7A,$D7,$6C,$EC,$F8,$9F,$AD,$EF
	.dc.b	$FD,$7C,$AC,$FE,$1D,$FB,$59,$48,$FD,$8F,$1F,$F7,$B7,$F1,$11,$FC
	.dc.b	$5E,$3F,$BE,$74,$7D,$9F,$07,$FD,$E7,$F0,$79,$F8,$6D,$3B,$F8,$4E
	.dc.b	$1C,$B7,$2E,$AF,$3E,$69,$1E,$0F,$86,$B9,$75,$CD,$F0,$DE,$1F,$86
	.dc.b	$F1,$36,$FC,$FF,$0D,$E5,$3C,$DE,$05,$8F,$5F,$81,$E9,$F8,$6F,$53
	.dc.b	$63,$DF,$F0,$DE,$DF,$86,$F7,$36,$3E,$1F,$C3,$7C,$1F,$2F,$E1,$36
	.dc.b	$3E,$3F,$97,$F1,$7E,$1B,$E3,$36,$3E,$5F,$C3,$7C,$9F,$86,$F9,$4D
	.dc.b	$8F,$9F,$F0,$9F,$D0,$FF,$39,$FE,$B1,$C6,$8C,$4E,$7D,$48,$FF,$26
	.dc.b	$AF,$F5,$57,$2F,$D3,$7F,$AE,$F6,$7F,$5A,$7F,$AC,$CE,$FE,$1B,$B7
	.dc.b	$C3,$77,$7C,$1F,$0D,$E1,$F7,$6D,$75,$7C,$9E,$EC,$8F,$07,$C2,$DC
	.dc.b	$FC,$1F,$0B,$E9,$3D,$5F,$0B,$6B,$CD,$F0,$7E,$DF,$84,$F7,$36,$3D
	.dc.b	$5F,$07,$F0,$7E,$13,$E1,$36,$3D,$DF,$07,$F1,$7E,$13,$E3,$36,$3E
	.dc.b	$17,$C1,$FC,$9F,$84,$F9,$4D,$8F,$8D,$F0,$9F,$34,$F9,$DF,$09,$63
	.dc.b	$FA,$7F,$E8,$5F,$C9,$B8,$3F,$BD,$3F,$C9,$BA,$3F,$C0,$6E,$53,$99
	.dc.b	$FF,$01,$BF,$C8,$A7,$F4,$8D,$9E,$3F,$07,$FD,$FE,$31,$5C,$3E,$0B
	.dc.b	$FC,$31,$7C,$1B,$7E,$06,$CF,$6F,$82,$B5,$E6,$F8,$2B,$7E,$4F,$82
	.dc.b	$B8,$FF,$A9,$2B,$DB,$58,$F8,$2C,$F8,$32,$7E,$0C,$70,$F8,$09,$1F
	.dc.b	$11,$E7,$F0,$3D,$3E,$06,$4F,$C8,$7B,$FB,$9E,$0F,$76,$4F,$CC,$7C
	.dc.b	$7F,$01,$E4,$F8,$0F,$2F,$C0,$4A,$FA,$31,$E8,$F8,$0F,$4F,$C0,$49
	.dc.b	$FA,$B1,$EC,$F8,$09,$1F,$61,$F8,$3F,$A1,$FC,$A8,$FE,$28,$BE,$03
	.dc.b	$3F,$03,$27,$FC,$1B,$39,$6F,$C0,$F2,$F8,$19,$18,$37,$F0,$3D,$88
	.dc.b	$F2,$F8,$1F,$0F,$C0,$48,$EA,$F9,$3E,$07,$CB,$F0,$32,$7C,$2F,$A3
	.dc.b	$E0,$7D,$3F,$01,$27,$CA,$FB,$3E,$07,$DB,$F0,$12,$7D,$2F,$C0,$F8
	.dc.b	$1F,$83,$F0,$12,$7D,$AF,$C4,$F8,$1F,$8B,$F0,$12,$7E,$0B,$F2,$3E
	.dc.b	$07,$E4,$FC,$07,$CA,$F8,$19,$5F,$1A,$3E,$6F,$C0,$C8,$F9,$2F,$E4
	.dc.b	$3D,$9F,$F3,$4C,$7F,$43,$5F,$03,$9F,$89,$93,$FE,$3C,$9C,$B7,$E0
	.dc.b	$39,$7C,$0C,$8C,$1B,$F8,$1E,$DF,$03,$23,$07,$3F,$81,$F1,$11,$ED
	.dc.b	$F0,$3E,$5C,$64,$78,$5F,$47,$C0,$FE,$A3,$E6,$DB,$D2,$57,$F9,$4A
	.dc.b	$1C,$3E,$FF,$81,$93,$EB,$7E,$17,$C1,$7C,$3F,$81,$F8,$9F,$05,$2B
	.dc.b	$E0,$47,$7E,$BF,$07,$5C,$9E,$D1,$F2,$FE,$06,$47,$C7,$7E,$77,$C1
	.dc.b	$7C,$FF,$81,$FA,$1F,$05,$2B,$E6,$47,$AF,$C5,$F6,$78,$B2,$7D,$51
	.dc.b	$F5,$FE,$0A,$47,$D3,$7E,$D7,$C1,$7F,$0A,$FD,$11,$F1,$34,$95,$FE
	.dc.b	$74,$87,$C6,$FC,$98,$87,$C7,$3E,$67,$A1,$CF,$E2,$64,$71,$7B,$7C
	.dc.b	$37,$7F,$86,$93,$D5,$F1,$7C,$2F,$F1,$CD,$9E,$D5,$25,$7F,$A1,$A1
	.dc.b	$EC,$FA,$7E,$0B,$D5,$F0,$92,$BD,$11,$ED,$F8,$29,$1E,$B7,$E0,$7C
	.dc.b	$1F,$73,$C1,$F0,$72,$7B,$47,$C5,$89,$1F,$0D,$F9,$1F,$07,$F2,$7E
	.dc.b	$0E,$4F,$C7,$7E,$67,$C1,$FA,$F8,$24,$FA,$5F,$77,$07,$D1,$F8,$3F
	.dc.b	$A5,$C1,$2B,$E8,$47,$D5,$DF,$EB,$7C,$1C,$9F,$A9,$1F,$D0,$BE,$7B
	.dc.b	$FD,$63,$1F,$EA,$A5,$D5,$93,$5D,$BF,$7F,$E0,$E5,$7D,$D7,$F0,$5F
	.dc.b	$09,$F8,$3F,$84,$93,$F8,$17,$FA,$7F,$E8,$BF,$CB,$FE,$1F,$FD,$CF
	.dc.b	$97,$FA,$CA,$BC,$1E,$3F,$8C,$F2,$79,$3E,$5F,$1A,$4F,$95,$F4,$79
	.dc.b	$3D,$3C,$09,$3F,$EC,$2C,$BF,$9B,$3F,$E4,$F5,$F0,$DE,$FF,$4E,$4F
	.dc.b	$BD,$F8,$5E,$9F,$C3,$F4,$E4,$FC,$37,$E3,$7A,$7E,$AF,$85,$F5,$FC
	.dc.b	$2C,$AF,$04,$7B,$BE,$16,$47,$8D,$F8,$3E,$6F,$C2,$F1,$E4,$F9,$DF
	.dc.b	$8B,$F0,$BF,$1B,$E1,$7E,$3F,$C2,$CA,$F6,$47,$47,$D9,$85,$E7,$7C
	.dc.b	$42,$50,$E2,$67,$E0,$9F,$89,$9D,$AE,$31,$C2,$99,$FD,$8F,$86,$FB
	.dc.b	$3E,$AC,$AF,$9D,$1E,$1C,$BC,$18,$77,$CF,$15,$CF,$3E,$D5,$AC,$BF
	.dc.b	$CE,$B3,$F0,$4F,$B1,$97,$E1,$BD,$6F,$F0,$AF,$CF,$FF,$AF,$67,$FD
	.dc.b	$FA,$D7,$EE,$46,$B4,$91,$A7,$8C,$72,$F8,$79,$3C,$9E,$BF,$0F,$F9
	.dc.b	$47,$F2,$AC,$8F,$C5,$47,$E6,$3E,$1F,$F3,$3F,$0F,$27,$F1,$D1,$F9
	.dc.b	$CF,$77,$F3,$BE,$FC,$9F,$C9,$47,$E8,$3E,$1F,$F4,$3F,$0F,$27,$F2
	.dc.b	$D1,$D5,$FC,$C6,$1D,$5F,$CC,$61,$D5,$FC,$E6,$1D,$5F,$CE,$61,$FE
	.dc.b	$3A,$FD,$1B,$FB,$5F,$27,$DA,$0E,$38,$BE,$AB,$6B,$F0,$AF,$F6,$C7
	.dc.b	$FD,$68,$BE,$1F,$A1,$23,$EE,$CF,$2F,$87,$F0,$7C,$3C,$9C,$9D,$1F
	.dc.b	$27,$C3,$F9,$7C,$2F,$37,$85,$2B,$79,$F9,$A7,$B4,$78,$B2,$B3,$EB
	.dc.b	$CD,$97,$7A,$F3,$E5,$6B,$59,$7C,$3F,$0B,$E2,$78,$52,$FB,$E2,$9F
	.dc.b	$91,$F0,$FF,$27,$E1,$A5,$6A,$3E,$67,$93,$F3,$7C,$A9,$3C,$A3,$E8
	.dc.b	$79,$3F,$47,$CA,$93,$DA,$3A,$BE,$0C,$3A,$BF,$3B,$0E,$AF,$93,$0E
	.dc.b	$AF,$D2,$C3,$EE,$7C,$37,$DD,$F8,$69,$3E,$A8,$EA,$FD,$8C,$3A,$EF
	.dc.b	$F1,$F1,$EB,$3F,$73,$1E,$B3,$F2,$F1,$FC,$47,$A3,$2B,$E2,$BF,$8B
	.dc.b	$F4,$BF,$19,$F0,$D2,$7E,$4B,$FE,$88,$F2,$FF,$BF,$C7,$DA,$7F,$BE
	.dc.b	$B7,$E1,$34,$6F,$E1,$38,$7B,$7C,$7D,$B9,$5C,$63,$B3,$F9,$2C,$2E
	.dc.b	$9F,$9B,$C7,$BC,$FF,$1C,$ED,$FE,$1F,$1E,$08,$C1,$FC,$81,$C4,$F1
	.dc.b	$CB,$C5,$3E,$7F,$84,$F5,$7C,$1C,$8C,$1E,$BF,$84,$F7,$7C,$24,$8C
	.dc.b	$1E,$FF,$84,$F8,$44,$7E,$17,$C2,$7C,$5F,$84,$91,$F1,$5F,$91,$F0
	.dc.b	$9F,$27,$0E,$93,$E8,$D1,$AB,$2D,$4F,$72,$1B,$86,$5F,$4F,$2F,$A9
	.dc.b	$F0,$92,$BE,$A3,$F5,$FE,$0F,$EC,$7C,$24,$9F,$B0,$FD,$BF,$83,$FB
	.dc.b	$9F,$09,$F7,$7E,$12,$57,$DD,$8F,$C0,$7C,$27,$38,$FC,$06,$18,$79
	.dc.b	$90,$F3,$61,$D5,$E4,$7E,$1E,$FA,$BE,$BB,$FC,$67,$C2,$73,$8F,$76
	.dc.b	$18,$7A,$6F,$F8,$3C,$3D,$51,$F0,$F2,$F4,$BD,$48,$75,$3F,$87,$FE
	.dc.b	$8F,$FE,$11,$E7,$FE,$AE,$D8,$FE,$CE,$DB,$3F,$E0,$A7,$F9,$63,$FE
	.dc.b	$0A,$8E,$7F,$07,$FE,$EA,$8C,$47,$DB,$5F,$05,$9F,$83,$93,$E2,$78
	.dc.b	$7C,$07,$1F,$84,$93,$E6,$7A,$7C,$0F,$5C,$64,$F8,$0F,$5E,$1D,$1F
	.dc.b	$06,$17,$2F,$3E,$3E,$5F,$83,$95,$F0,$9F,$47,$C0,$FA,$7E,$12,$4F
	.dc.b	$C6,$7D,$9D,$3E,$DF,$84,$93,$F2,$9F,$81,$F0,$5C,$A3,$A1,$0F,$85
	.dc.b	$7D,$1F,$89,$77,$2F,$97,$87,$49,$F9,$18,$59,$F9,$F8,$7C,$DF,$82
	.dc.b	$95,$F6,$9F,$A1,$F0,$3F,$47,$E0,$A4,$FD,$E7,$EA,$7B,$FF,$57,$E0
	.dc.b	$3E,$B7,$BF,$2B,$F0,$71,$C9,$E8,$43,$ED,$5F,$47,$EE,$5D,$CB,$F0
	.dc.b	$78,$75,$9F,$C6,$E1,$D5,$FC,$16,$1D,$5E,$67,$E4,$B0,$FC,$47,$C0
	.dc.b	$73,$E8,$FC,$9E,$56,$BF,$27,$8F,$57,$F1,$98,$F5,$CB,$F2,$79,$7E
	.dc.b	$47,$E0,$65,$7E,$79,$FC,$A7,$C1,$7F,$90,$E3,$F8,$77,$E9,$1F,$E0
	.dc.b	$1B,$34,$9F,$15,$B8,$FF,$05,$47,$F8,$0E,$7F,$BB,$BF,$DC,$9B,$F0
	.dc.b	$3B,$D7,$FC,$83,$3B,$9F,$D1,$61,$E4,$7F,$49,$8F,$91,$E7,$5C,$5F
	.dc.b	$F2,$BD,$78,$E3,$18,$79,$B1,$B3,$D6,$3C,$B8,$E0,$CF,$C1,$6B,$E0
	.dc.b	$B7,$F0,$53,$2F,$4C,$72,$F8,$3E,$7F,$09,$27,$DB,$1D,$BE,$0A,$47
	.dc.b	$C0,$7C,$3F,$09,$D6,$39,$90,$F9,$BE,$F7,$57,$E3,$DF,$57,$E8,$DF
	.dc.b	$47,$E9,$DD,$D3,$E7,$E1,$EE,$F7,$E5,$7D,$07,$E0,$FB,$DF,$0B,$DF
	.dc.b	$93,$F5,$1F,$8B,$F0,$5F,$1B,$E0,$FE,$3F,$C1,$4A,$FB,$31,$C9,$E8
	.dc.b	$43,$F0,$B7,$D1,$FC,$3D,$DC,$BF,$07,$87,$59,$FC,$1E,$1D,$5F,$C6
	.dc.b	$E1,$D5,$FC,$7E,$3E,$A7,$A9,$E7,$21,$D8,$FC,$6D,$75,$7F,$2F,$5D
	.dc.b	$6F,$F2,$77,$F8,$2F,$83,$FC,$1F,$C1,$CB,$FC,$AC,$7E,$1B,$E0,$FF
	.dc.b	$0F,$F0,$12,$7F,$33,$1F,$8C,$FD,$27,$FC,$17,$8F,$F9,$9D,$FF,$03
	.dc.b	$F5,$7F,$A6,$9B,$5F,$EA,$A7,$71,$F9,$EC,$3C,$13,$FA,$CC,$7C,$8F
	.dc.b	$63,$F4,$F8,$F8,$5F,$D8,$E3,$E1,$F8,$AF,$17,$C1,$49,$FF,$31,$E5
	.dc.b	$D9,$F0,$7C,$2F,$9C,$8E,$BE,$13,$D5,$F0,$BE,$BF,$85,$93,$CA,$3D
	.dc.b	$DF,$0B,$EF,$F8,$39,$3D,$A3,$E1,$5F,$48,$FF,$09,$62,$C7,$31,$CC
	.dc.b	$F2,$E3,$D9,$DE,$3D,$9F,$4E,$3D,$9E,$78,$FC,$FF,$83,$91,$EE,$7E
	.dc.b	$97,$C1,$FD,$3F,$83,$FA,$9F,$0B,$2B,$E1,$C7,$D7,$F8,$59,$1F,$19
	.dc.b	$FB,$5F,$0B,$CA,$3A,$10,$F5,$DF,$57,$E6,$DF,$57,$DF,$7D,$1F,$83
	.dc.b	$77,$9F,$F0,$98,$75,$9F,$8B,$8F,$A9,$EA,$79,$8E,$A4,$3A,$9E,$72
	.dc.b	$1D,$8F,$BF,$1F,$93,$F8,$59,$1F,$82,$7F,$2D,$F0,$3F,$97,$F8,$19
	.dc.b	$3F,$86,$7F,$35,$F0,$9F,$9B,$F8,$59,$3F,$8A,$7F,$3D,$F0,$9C,$A3
	.dc.b	$A1,$0F,$BF,$7D,$1F,$C0,$DD,$CF,$F4,$98,$74,$9F,$C2,$E1,$67,$F5
	.dc.b	$18,$7E,$AF,$E1,$65,$7E,$69,$FE,$73,$B7,$F8,$AC,$FF,$15,$77,$25
	.dc.b	$69,$FF,$14,$BF,$ED,$96,$CE,$E3,$5F,$09,$D3,$E0,$A5,$FF,$B2,$AF
	.dc.b	$FC,$5E,$DF,$82,$F0,$FC,$14,$8D,$3E,$4F,$84,$F2,$FC,$1F,$9B,$E0
	.dc.b	$E5,$73,$8F,$4F,$C1,$FA,$BE,$0E,$4F,$78,$F6,$FC,$1C,$8F,$13,$F0
	.dc.b	$3E,$0F,$E0,$FC,$1C,$9F,$33,$F1,$3E,$0F,$E2,$FC,$17,$C6,$F8,$29
	.dc.b	$5E,$B8,$E4,$F4,$21,$FE,$87,$E0,$61,$D6,$31,$85,$E3,$F9,$58,$F6
	.dc.b	$9F,$8D,$8F,$67,$96,$3F,$5B,$E0,$64,$7C,$B7,$EC,$FC,$17,$DA,$F8
	.dc.b	$3F,$B7,$F0,$52,$BE,$84,$7D,$EF,$83,$91,$F4,$DF,$EF,$3F,$A5,$7F
	.dc.b	$83,$C7,$C2,$60,$93,$F7,$7A,$FE,$F4,$FF,$7D,$6F,$C1,$F2,$F8,$19
	.dc.b	$5B,$3F,$13,$5F,$E1,$37,$FD,$54,$BE,$0B,$C3,$F0,$12,$78,$71,$2B
	.dc.b	$51,$85,$F0,$3E,$82,$46,$7E,$07,$D7,$F0,$32,$38,$3E,$EF,$82,$F7
	.dc.b	$FC,$04,$9E,$8F,$C2,$F8,$2F,$87,$F0,$52,$7C,$0F,$C6,$F8,$2F,$8F
	.dc.b	$F0,$32,$7C,$8F,$CA,$F8,$2F,$97,$F0,$52,$7D,$0F,$CE,$F8,$2F,$9F
	.dc.b	$F0,$3F,$43,$E0,$A5,$7B,$63,$FC,$57,$FA,$5F,$F4,$DE,$33,$49,$F0
	.dc.b	$5A,$F8,$09,$7F,$22,$7F,$A2,$37,$E0,$B9,$FC,$04,$8D,$3D,$BE,$07
	.dc.b	$BF,$C0,$49,$E4,$F8,$BE,$07,$C7,$EF,$49,$EC,$F9,$BE,$07,$CF,$E6
	.dc.b	$FA,$3E,$06,$57,$8E,$3D,$7F,$03,$23,$CC,$FB,$BE,$07,$DF,$F0,$32
	.dc.b	$7D,$4F,$C2,$F8,$1F,$87,$F0,$3F,$13,$E0,$65,$7B,$E3,$E3,$FC,$0C
	.dc.b	$8F,$84,$FF,$95,$D7,$CB,$FC,$47,$FD,$F5,$C8,$F8,$2F,$9E,$48,$F9
	.dc.b	$5F,$05,$FE,$5C,$FD,$33,$FD,$9C,$7F,$24,$8F,$E4,$6D,$EA,$3F,$ED
	.dc.b	$58,$7D,$6B,$C6,$F1,$65,$E1,$B7,$9F,$FE,$7F,$8C,$3F,$E8,$1A,$E1
	.dc.b	$7F,$A2,$C3,$F9,$6B,$E0,$F8,$9F,$21,$E5,$F8,$4F,$37,$C2,$79,$FE
	.dc.b	$12,$57,$18,$F8,$58,$74,$8F,$E7,$78,$F8,$7C,$0F,$29,$AD,$7E,$07
	.dc.b	$C2,$4A,$F1,$3F,$0F,$E0,$FE,$27,$C2,$7C,$5F,$83,$95,$E7,$8E,$4F
	.dc.b	$F0,$13,$E8,$E1,$D9,$C6,$5E,$3E,$1F,$87,$1F,$33,$E1,$3E,$81,$1F
	.dc.b	$85,$F0,$9F,$4F,$E1,$3E,$A7,$C2,$49,$F8,$D1,$E0,$7D,$A7,$F5,$9E
	.dc.b	$5F,$9A,$DC,$B5,$8F,$80,$FB,$37,$CE,$78,$C6,$1F,$A5,$F0,$9F,$83
	.dc.b	$F8,$39,$1F,$55,$FC,$37,$C2,$74,$E3,$F9,$93,$F0,$72,$F0,$BF,$84
	.dc.b	$E3,$E2,$F6,$E8,$61,$8C,$BC,$90,$F8,$4F,$C9,$7C,$24,$BF,$C2,$3F
	.dc.b	$95,$F8,$3F,$04,$7B,$4F,$06,$F8,$C3,$F2,$B5,$C6,$7C,$0F,$E7,$7E
	.dc.b	$13,$F3,$DF,$09,$27,$F2,$0F,$FA,$F3,$4F,$AB,$1F,$30,$F5,$1D,$88
	.dc.b	$67,$97,$EE,$6F,$FE,$03,$D5,$7D,$6B,$E0,$3D,$7E,$13,$B7,$C4,$77
	.dc.b	$F8,$59,$5F,$D0,$B2,$FE,$9C,$FF,$84,$F2,$EA,$FB,$CF,$27,$17,$CE
	.dc.b	$8C,$ED,$FB,$1F,$85,$F5,$7C,$2C,$BE,$91,$ED,$F8,$59,$1D,$DF,$81
	.dc.b	$F0,$9F,$32,$3F,$32,$7B,$AF,$E1,$E1,$DB,$25,$F2,$AF,$C4,$C6,$1F
	.dc.b	$57,$C2,$7C,$BF,$84,$91,$ED,$7E,$74,$F7,$D3,$F0,$E7,$AD,$DC,$F4
	.dc.b	$9B,$8F,$E6,$65,$F0,$9F,$5F,$E1,$25,$FC,$97,$ED,$7C,$27,$5E,$8F
	.dc.b	$D9,$4F,$D7,$BD,$BE,$8A,$B8,$75,$F8,$4F,$C1,$7C,$24,$AF,$A8,$FE
	.dc.b	$17,$E1,$3A,$47,$C5,$E9,$FE,$35,$8E,$E3,$E1,$D5,$C3,$D1,$F0,$9F
	.dc.b	$8D,$F8,$49,$5F,$7D,$FC,$87,$C1,$F4,$8F,$BB,$AF,$E6,$F1,$57,$09
	.dc.b	$F9,$EF,$E5,$FE,$13,$FC,$B1,$1F,$BA,$FD,$37,$FB,$51,$3E,$27,$39
	.dc.b	$FF,$01,$8F,$F6,$4C,$FE,$70,$EE,$7A,$CF,$C8,$61,$C3,$5F,$F2,$F4
	.dc.b	$F3,$9F,$F0,$BB,$DB,$2D,$47,$F9,$A2,$1F,$0B,$8C,$11,$89,$E3,$39
	.dc.b	$87,$C2,$FC,$1A,$FD,$8F,$37,$E0,$EF,$64,$3D,$D7,$C7,$E1,$7F,$A7
	.dc.b	$B2,$3F,$89,$3D,$7E,$17,$A6,$7F,$C5,$A7,$F0,$46,$88,$7E,$B2,$7F
	.dc.b	$85,$64,$F8,$5F,$31,$27,$E0,$FC,$2F,$47,$F6,$99,$FE,$9B,$6F,$CE
	.dc.b	$61,$E3,$9E,$E7,$D5,$F8,$5F,$80,$47,$E5,$FC,$27,$37,$F6,$27,$0B
	.dc.b	$E8,$F9,$9F,$8B,$5E,$6F,$85,$F9,$44,$9F,$AB,$F0,$BC,$F8,$3E,$54
	.dc.b	$73,$DB,$FB,$95,$F9,$70,$EB,$3F,$89,$F8,$5F,$A8,$47,$EF,$FC,$2F
	.dc.b	$3E,$4F,$E0,$DE,$1F,$F1,$49,$FF,$07,$E1,$CF,$1F,$D1,$C7,$F6,$3F
	.dc.b	$94,$7C,$2F,$F8,$55,$B1,$98,$FE,$EF,$E1,$F5,$8F,$EC,$FC,$5F,$29
	.dc.b	$F8,$17,$D1,$F5,$1F,$C9,$67,$8B,$C0,$87,$E7,$09,$AF,$B8,$93,$F0
	.dc.b	$A3,$A9,$FA,$03,$8C,$7F,$20,$9E,$B1,$E0,$3F,$51,$1F,$96,$CF,$F8
	.dc.b	$6B,$D2,$39,$1F,$8B,$3A,$9C,$2B,$D5,$1D,$63,$FC,$0B,$5F,$32,$F9
	.dc.b	$D7,$E5,$EB,$A5,$71,$3F,$BF,$65,$F3,$DE,$F1,$AC,$D8,$7A,$23,$18
	.dc.b	$74,$AF,$CB,$E1,$E2,$23,$D0,$F3,$DF,$DD,$25,$F3,$9D,$CF,$8C,$FC
	.dc.b	$BD,$FC,$37,$E4,$C6,$B6,$FE,$01,$1F,$0C,$F9,$57,$D1,$F8,$27,$03
	.dc.b	$D7,$3F,$04,$8E,$08,$7E,$40,$F3,$E1,$F9,$4C,$3C,$D1,$AA,$FC,$B5
	.dc.b	$FF,$4E,$CB,$E7,$4F,$D4,$8F,$2E,$3F,$1A,$3C,$67,$F7,$DB,$DB,$F8
	.dc.b	$AC,$7C,$6F,$C1,$C3,$AE,$5F,$4E,$31,$97,$89,$FF,$32,$DF,$2C,$3E
	.dc.b	$34,$73,$3C,$D3,$C7,$1F,$1E,$3F,$3B,$1F,$8D,$5E,$33,$E3,$1F,$46
	.dc.b	$7D,$8E,$23,$C6,$FC,$43,$18,$78,$E3,$EE,$5F,$E0,$E7,$EC,$DF,$78
	.dc.b	$F1,$63,$E7,$8F,$F3,$B5,$F1,$9F,$DB,$57,$DA,$76,$F8,$5E,$D1,$C6
	.dc.b	$3F,$1E,$43,$E2,$9F,$46,$B4,$FC,$63,$E8,$E3,$F4,$DF,$8C,$79,$8F
	.dc.b	$8E,$71,$3C,$07,$0B,$F3,$B8,$38,$E3,$FC,$8A,$7F,$56,$FC,$E8,$FD
	.dc.b	$01,$EB,$39,$1F,$C3,$4E,$51,$F6,$09,$3F,$1E,$3C,$47,$13,$EB,$12
	.dc.b	$3D,$47,$07,$E1,$63,$F4,$9E,$31,$A9,$F4,$C7,$17,$B3,$EA,$CB,$E3
	.dc.b	$DF,$A6,$3F,$44,$78,$6B,$F8,$63,$C4,$F8,$E7,$13,$D7,$5F,$C8,$5F
	.dc.b	$DA,$47,$85,$F8,$07,$13,$EB,$9C,$4E,$D1,$C6,$3F,$A9,$C7,$43,$8B
	.dc.b	$E2,$3A,$3F,$64,$97,$F0,$5F,$0C,$F3,$3D,$F3,$C5,$F8,$6F,$2A,$FC
	.dc.b	$FC,$F8,$8F,$B3,$1E,$27,$07,$E2,$CF,$55,$77,$8F,$A5,$1E,$2C,$BF
	.dc.b	$A7,$3E,$23,$E3,$C7,$85,$F5,$DF,$90,$FE,$03,$7F,$8B,$7F,$8F,$63
	.dc.b	$FC,$07,$0E,$D1,$E8,$25,$74,$78,$E3,$F6,$AB,$86,$3F,$C6,$6B,$F2
	.dc.b	$27,$D0,$C7,$CF,$8F,$78,$FE,$42,$FE,$04,$FC,$16,$3E,$08,$F3,$90
	.dc.b	$E5,$8F,$81,$FA,$C7,$E2,$2B,$E4,$1F,$5A,$7F,$A4,$C7,$89,$E8,$FD
	.dc.b	$9A,$F1,$3F,$C4,$70,$F4,$BF,$C3,$8E,$18,$7F,$06,$BF,$B1,$3A,$9F
	.dc.b	$E8,$39,$7F,$0B,$7F,$8D,$DF,$DC,$AD,$C7,$F1,$93,$67,$43,$79,$7F
	.dc.b	$19,$C7,$CB,$5F,$0F,$0F,$E1,$71,$EA,$AF,$C0,$E3,$E3,$CB,$ED,$47
	.dc.b	$E2,$1F,$C8,$3F,$54,$F9,$13,$C7,$1F,$C8,$63,$DE,$FF,$0F,$1F,$22
	.dc.b	$F9,$DF,$E3,$E3,$F1,$0F,$D0,$3F,$90,$63,$F2,$30,$FE,$6B,$1F,$8F
	.dc.b	$34,$71,$AF,$C3,$B8,$9F,$89,$1D,$9F,$74,$7D,$A9,$F3,$4F,$07,$E1
	.dc.b	$E1,$E7,$9F,$9B,$8F,$D1,$C3,$FC,$0D,$97,$D1,$9F,$F0,$E6,$3D,$9F
	.dc.b	$84,$7E,$17,$1F,$8A,$F9,$B1,$EE,$FC,$6C,$BE,$63,$C2,$BC,$AF,$EC
	.dc.b	$A7,$6F,$B4,$D6,$5E,$57,$C3,$97,$97,$0F,$EC,$18,$F9,$72,$FE,$6F
	.dc.b	$7B,$7F,$9C,$D6,$EF,$FB,$F5,$6E,$FF,$C2,$17,$B7,$F6,$B3,$F8,$BB
	.dc.b	$F8,$2F,$E2,$CF,$83,$3B,$7F,$5B,$7D,$5F,$EF,$D7,$D5,$FD,$8C,$ED
	.dc.b	$FD,$11,$D6,$FD,$2F,$5B,$F4,$BF,$EA,$D3,$EE,$5F,$5C,$3F,$BE,$CF
	.dc.b	$E2,$1F,$B3,$3B,$7F,$3B,$3B,$CB,$FB,$6C,$FE,$EB,$2F,$B0,$F4,$3C
	.dc.b	$47,$C4,$3C,$58,$7C,$47,$A4,$FE,$6A,$3E,$ED,$FE,$F8,$FA,$E4,$8F
	.dc.b	$BB,$3D,$4F,$CC,$46,$F2,$F8,$91,$FE,$13,$CB,$E2,$4F,$E7,$B2,$FD
	.dc.b	$34,$FE,$32,$3A,$D7,$F5,$AA,$DD,$FF,$82,$AB,$F1,$13,$D2,$BF,$C7
	.dc.b	$35,$A9,$FE,$CF,$87,$EC,$27,$9C,$7C,$39,$D3,$FA,$DC,$7F,$92,$E1
	.dc.b	$FC,$B6,$3F,$81,$9F,$C2,$8F,$86,$7E,$98,$F9,$27,$E0,$0F,$92,$7E
	.dc.b	$00,$D9,$F9,$83,$75,$F6,$AF,$F6,$53,$F2,$23,$EC,$CE,$A3,$EC,$BF
	.dc.b	$D5,$AB,$EC,$CE,$A3,$E6,$CF,$9F,$1F,$C2,$3C,$CF,$B1,$1F,$C6,$4F
	.dc.b	$99,$8F,$C7,$8F,$F1,$E5,$FF,$0F,$7E,$65,$7F,$3D,$9D,$57,$E4,$EB
	.dc.b	$4F,$CC,$39,$9F,$00,$FB,$87,$C0,$9F,$D8,$BE,$B3,$FA,$34,$FD,$B7
	.dc.b	$58,$7E,$C1,$F5,$9F,$D1,$4F,$A2,$7E,$C0,$D9,$FB,$03,$A1,$FD,$E2
	.dc.b	$3A,$3F,$E4,$2C,$3E,$35,$7E,$32,$3F,$8A,$4F,$B1,$FF,$0B,$47,$F4
	.dc.b	$A7,$C2,$7F,$4A,$3C,$2F,$ED,$1E,$98,$7C,$C7,$F9,$01,$F0,$8D,$DF
	.dc.b	$EA,$5D,$9F,$87,$3A,$CF,$F0,$C7,$67,$E1,$CD,$9F,$87,$37,$1F,$C1
	.dc.b	$9D,$9F,$38,$DC,$7C,$E7,$67,$EE,$CD,$DF,$E9,$1F,$92,$7F,$12,$36
	.dc.b	$7E,$64,$DD,$7E,$13,$1F,$F0,$84,$7C,$EA,$FE,$21,$87,$CE,$7F,$C4
	.dc.b	$B7,$F3,$B0,$FE,$0B,$87,$E4,$DF,$CE,$DF,$EA,$4F,$3B,$F7,$EB,$EA
	.dc.b	$57,$F0,$59,$FC,$53,$E7,$76,$7D,$F3,$CE,$FE,$A5,$FE,$98,$75,$8F
	.dc.b	$1D,$7D,$5C,$7F,$77,$3F,$8B,$9F,$D2,$BB,$BF,$F5,$AD,$FA,$63,$F4
	.dc.b	$C7,$E1,$8F,$1D,$FC,$37,$F9,$56,$5F,$CB,$5F,$CD,$9F,$C2,$CF,$9B
	.dc.b	$3F,$C6,$EB,$E1,$C7,$85,$F8,$6F,$E6,$CF,$69,$F4,$EF,$E1,$DF,$C9
	.dc.b	$8F,$69,$B8,$F6,$BF,$8B,$76,$F5,$36,$FE,$5C,$DB,$F8,$B7,$6F,$E7
	.dc.b	$5F,$86,$F5,$9F,$87,$3F,$26,$FF,$C4,$B8,$FE,$9E,$FF,$C9,$B9,$7E
	.dc.b	$13,$0F,$A7,$7F,$C5,$A7,$FC,$43,$1F,$0A,$3F,$85,$E1,$FC,$C6,$F9
	.dc.b	$3F,$E4,$D9,$FE,$39,$87,$CB,$7F,$74,$7C,$B3,$E2,$E3,$F2,$DF,$AA
	.dc.b	$FE,$1D,$FC,$B9,$FC,$57,$1F,$26,$3E,$E3,$F5,$C7,$B8,$F8,$97,$EE
	.dc.b	$9F,$94,$FD,$67,$E2,$3F,$5A,$37,$1F,$EB,$2C,$7E,$23,$FE,$2B,$C7
	.dc.b	$D6,$FF,$8E,$2F,$C3,$7C,$EF,$C2,$F3,$3F,$82,$3F,$50,$FD,$C1,$E7
	.dc.b	$7F,$7C,$F9,$DD,$9E,$79,$FB,$4F,$53,$67,$57,$F8,$54,$75,$8F,$D7
	.dc.b	$1B,$8F,$E0,$91,$D5,$DC,$75,$8F,$1B,$B8,$FE,$09,$1B,$7F,$88,$3D
	.dc.b	$5F,$DA,$3E,$97,$AC,$78,$DD,$BD,$4F,$1B,$F6,$AF,$F3,$CF,$F8,$FC
	.dc.b	$FF,$2D,$57,$F7,$4A,$FE,$71,$97,$EE,$A7,$EA,$9F,$9E,$BF,$8A,$FD
	.dc.b	$53,$ED,$9D,$4F,$D2,$1D,$71,$FF,$13,$BF,$8C,$C3,$F2,$AF,$F3,$9B
	.dc.b	$F7,$4F,$F1,$17,$DD,$1F,$B5,$AF,$89,$3F,$C4,$63,$A5,$7F,$6F,$7C
	.dc.b	$27,$F4,$43,$F8,$51,$F9,$63,$FA,$7C,$FF,$9D,$72,$FD,$05,$FF,$2C
	.dc.b	$9F,$D0,$3E,$0B,$FD,$04,$73,$7F,$41,$1E,$6B,$FE,$29,$3F,$AA,$3F
	.dc.b	$9B,$5F,$F9,$F7,$2F,$E6,$CF,$D1,$8F,$AE,$FA,$A7,$F7,$0F,$FA,$1F
	.dc.b	$2F,$A0,$FF,$9E,$B2,$FA,$18,$7E,$6A,$FE,$81,$F8,$83,$E8,$1F,$94
	.dc.b	$3E,$80,$52,$89,$8B,$0C,$44,$32,$D7,$31,$B7,$01,$C3,$B6,$BC,$DA
	.dc.b	$F2,$8E,$2F,$16,$C6,$A3,$C2,$76,$70,$EC,$C4,$70,$23,$82,$1E,$37
	.dc.b	$2D,$32,$79,$5C,$B4,$E3,$5F,$83,$7F,$C8,$18,$7D,$47,$FA,$AE,$1F
	.dc.b	$53,$0F,$E0,$31,$F5,$0F,$D7,$9F,$52,$BF,$76,$FD,$47,$ED,$3F,$98
	.dc.b	$3F,$B1,$1B,$8F,$DA,$BF,$8F,$BF,$DA,$E1,$FE,$7D,$8D,$8B,$FF,$21
	.dc.b	$BF,$E6,$9C,$BF,$A3,$BF,$DA,$0F,$E6,$87,$F6,$83,$F2,$A7,$FB,$A3
	.dc.b	$2F,$CA,$BE,$13,$F2,$A7,$9F,$0F,$6B,$E7,$3F,$AB,$1B,$3F,$68,$6C
	.dc.b	$FE,$22,$7C,$3B,$FD,$D4,$7E,$86,$36,$7B,$4F,$E0,$B1,$ED,$7A,$9E
	.dc.b	$DC,$BF,$9D,$BD,$4F,$E4,$73,$FC,$76,$7A,$C7,$8D,$FD,$BE,$1F,$56
	.dc.b	$3C,$67,$A7,$2F,$E9,$6F,$53,$E2,$1F,$DA,$8F,$F0,$59,$F8,$03,$FB
	.dc.b	$51,$FE,$6F,$3F,$85,$BF,$C0,$5F,$9A,$7F,$01,$39,$1F,$DA,$8E,$47
	.dc.b	$F1,$F3,$94,$FF,$0A,$AF,$56,$5F,$C2,$AB,$FC,$C4,$7F,$3B,$AF,$79
	.dc.b	$F2,$E3,$53,$FC,$66,$3E,$D9,$FC,$B7,$1F,$B6,$FB,$4F,$25,$73,$BF
	.dc.b	$E5,$07,$CF,$7E,$1E,$1F,$8A,$7F,$9F,$E5,$F8,$AA,$FE,$51,$1F,$E1
	.dc.b	$07,$E3,$BC,$F1,$FE,$33,$1F,$B8,$3E,$CD,$73,$C7,$F8,$D5,$F3,$BF
	.dc.b	$91,$5F,$CC,$8E,$0F,$E9,$9F,$C2,$1F,$D3,$8F,$B1,$3F,$D3,$9F,$64
	.dc.b	$7C,$68,$FE,$55,$1F,$62,$3F,$C4,$2F,$CF,$3F,$48,$7F,$81,$CF,$D2
	.dc.b	$1E,$13,$FD,$CC,$FA,$DE,$51,$FC,$CD,$FC,$09,$C8,$FF,$6F,$9C,$A3
	.dc.b	$E2,$3F,$3A,$3C,$6F,$E0,$63,$C6,$7F,$4B,$BF,$EB,$F1,$F8,$A9,$FC
	.dc.b	$0B,$F8,$A9,$F6,$3F,$9F,$C3,$EB,$E3,$F7,$31,$F9,$93,$FC,$4B,$1F
	.dc.b	$99,$87,$28,$F6,$5F,$F9,$AD,$F9,$78,$7F,$10,$8F,$C0,$4F,$CB,$BE
	.dc.b	$75,$FD,$9D,$FD,$43,$F9,$03,$E8,$9F,$D3,$CF,$A3,$1F,$91,$7F,$3F
	.dc.b	$5F,$C9,$CF,$7C,$7E,$FE,$BF,$67,$7E,$6A,$FC,$D1,$EF,$9E,$55,$F4
	.dc.b	$63,$E4,$47,$D1,$7E,$D4,$7E,$51,$FB,$47,$E5,$23,$FC,$B5,$1F,$22
	.dc.b	$3D,$F1,$FC,$62,$BD,$F1,$E9,$BF,$78,$BF,$F0,$0D,$7F,$77,$9F,$F0
	.dc.b	$0B,$E9,$C7,$EB,$DF,$ED,$E7,$D2,$7C,$E8,$FF,$29,$BF,$84,$3C,$87
	.dc.b	$F9,$4C,$FD,$3C,$FE,$13,$0F,$8C,$FB,$5F,$C2,$47,$F0,$49,$FE,$7F
	.dc.b	$8F,$DF,$CB,$F9,$0C,$FB,$0F,$E3,$78,$7B,$1F,$E9,$E7,$D4,$3F,$57
	.dc.b	$87,$F4,$AC,$BE,$DC,$FE,$2A,$7D,$A7,$F2,$7C,$3D,$A7,$81,$C9,$CC
	.dc.b	$FD,$2D,$7A,$4F,$8C,$FF,$1C,$C7,$D2,$FE,$95,$FF,$07,$BF,$E3,$B8
	.dc.b	$FC,$A4,$7E,$98,$FA,$D7,$E4,$AF,$E1,$35,$FE,$50,$3F,$5F,$3F,$7D
	.dc.b	$F9,$8F,$F0,$83,$95,$7E,$20,$F2,$47,$F5,$B3,$D3,$5F,$7B,$2F,$BB
	.dc.b	$19,$3D,$13,$F7,$5F,$A0,$7F,$52,$CB,$E8,$3F,$60,$FD,$96,$1E,$6C
	.dc.b	$7F,$CB,$38,$FC,$03,$CC,$7C,$09,$FD,$E1,$F8,$67,$CE,$FD,$43,$FB
	.dc.b	$69,$F6,$CF,$D7,$47,$DB,$3F,$AD,$BF,$45,$FD,$D6,$3F,$1E,$3F,$8A
	.dc.b	$9C,$CF,$C0,$BE,$F7,$FA,$1C,$2F,$FD,$4B,$1F,$6E,$74,$FE,$E3,$0F
	.dc.b	$B7,$7F,$B8,$8F,$C2,$1F,$10,$FE,$65,$7E,$5C,$3F,$A8,$C7,$86,$3F
	.dc.b	$7A,$FE,$D2,$7E,$B5,$7E,$D0,$FA,$D1,$F2,$8F,$49,$F7,$CF,$DE,$BF
	.dc.b	$D1,$9E,$4F,$E2,$0E,$4F,$7A,$FC,$BB,$F5,$CF,$BF,$97,$E9,$B0,$F8
	.dc.b	$D1,$FC,$28,$E0,$7F,$0F,$29,$F9,$7A,$FF,$72,$3F,$CE,$DE,$EF,$E8
	.dc.b	$0F,$41,$CC,$F4,$5F,$E0,$63,$D1,$1F,$96,$7D,$0F,$F0,$13,$FA,$21
	.dc.b	$CA,$3F,$80,$BE,$27,$F0,$AF,$73,$F0,$98,$7E,$DD,$FC,$B9,$F5,$0F
	.dc.b	$41,$FA,$D3,$F1,$D3,$E8,$8F,$C3,$E5,$FC,$BE,$7F,$9C,$9F,$86,$C3
	.dc.b	$E3,$46,$67,$EE,$4F,$C5,$3C,$58,$7C,$AC,$3C,$F1,$F0,$30,$FD,$43
	.dc.b	$E2,$3F,$3E,$76,$AF,$CC,$D7,$E3,$AB,$F9,$E1,$CA,$3F,$53,$5C,$A3
	.dc.b	$F5,$2F,$D2,$AF,$E5,$27,$E5,$F0,$F3,$BF,$84,$3D,$E7,$CD,$9F,$E1
	.dc.b	$78,$7A,$F2,$FD,$CC,$FD,$03,$E5,$E5,$FB,$93,$FD,$8C,$FF,$A9,$CC
	.dc.b	$9F,$94,$C7,$CA,$FC,$FB,$F2,$BF,$3C,$F7,$4F,$E8,$9F,$8F,$7E,$17
	.dc.b	$F2,$67,$F0,$93,$A9,$F8,$9A,$EA,$FE,$F6,$7F,$A4,$D7,$F0,$99,$FE
	.dc.b	$9D,$3D,$8F,$A8,$FD,$C7,$E1,$19,$3F,$A9,$1F,$EE,$0C,$BE,$C6,$5F
	.dc.b	$BC,$3F,$A7,$1F,$3C,$FC,$BD,$77,$BF,$E8,$15,$F8,$E3,$C8,$73,$C3
	.dc.b	$D8,$E6,$FE,$E5,$FF,$30,$8F,$F2,$81,$F0,$0F,$A1,$1E,$27,$F9,$85
	.dc.b	$FF,$0D,$7A,$9F,$46,$34,$F2,$3B,$65,$F8,$07,$F5,$87,$E0,$0F,$C2
	.dc.b	$1F,$36,$38,$E3,$FC,$7E,$3F,$74,$7E,$16,$FF,$88,$BF,$4A,$FF,$25
	.dc.b	$8F,$E0,$9F,$F2,$85,$F2,$CB,$F9,$3A,$C7,$FD,$81,$7F,$E2,$5C,$BE
	.dc.b	$1D,$FA,$EB,$E1,$CF,$EF,$71,$FA,$0F,$EF,$5F,$DA,$63,$F7,$9F,$CF
	.dc.b	$57,$F2,$69,$F3,$1B,$3F,$C7,$87,$DE,$CB,$F6,$AF,$DE,$3E,$46,$1E
	.dc.b	$2A,$FB,$B1,$F9,$77,$EE,$9D,$8F,$E2,$6F,$C8,$3E,$AC,$71,$7C,$06
	.dc.b	$6B,$F5,$6F,$87,$2F,$D5,$C6,$23,$E9,$9F,$91,$7F,$8C,$DF,$E5,$63
	.dc.b	$2F,$A2,$33,$87,$C1,$9F,$D5,$4E,$70,$F8,$B1,$F8,$F0,$2B,$F2,$CE
	.dc.b	$1F,$ED,$DA,$34,$DC,$EE,$AC,$E6,$38,$46,$27,$75,$FA,$FC,$7F,$05
	.dc.b	$3F,$E2,$7C,$B2,$F4,$C5,$96,$31,$F2,$72,$B7,$C6,$48,$C4,$F8,$0D
	.dc.b	$CE,$A1,$BF,$82,$7F,$C0,$7B,$7F,$B5,$EF,$97,$0E,$C9,$3C,$E3,$B7
	.dc.b	$0F,$18,$FF,$75,$D7,$1C,$1C,$7D,$C9,$3C,$61,$C7,$C7,$5F,$EF,$33
	.dc.b	$F2,$4F,$F0,$B6,$3E,$BE,$5F,$41,$23,$97,$2F,$1D,$FF,$B9,$C7,$F4
	.dc.b	$3C,$FF,$CD,$8F,$7E,$6F,$34,$62,$3F,$A4,$63,$F3,$1F,$27,$17,$C5
	.dc.b	$23,$E7,$E8,$F8,$D9,$FC,$D9,$CC,$F9,$E3,$0F,$B7,$A3,$E7,$74,$C8
	.dc.b	$F8,$0F,$D1,$F8,$7E,$7C,$08,$D4,$F3,$8C,$3F,$1F,$AB,$EE,$EC,$EC
	.dc.b	$A4,$9F,$A9,$0E,$DF,$3B,$B8,$CC,$ED,$F4,$43,$B7,$AF,$7C,$9E,$6F
	.dc.b	$83,$C0,$E6,$E6,$38,$BE,$4F,$03,$CB,$E1,$49,$F1,$BE,$8F,$0F,$C6
	.dc.b	$E6,$3C,$2F,$B3,$C3,$F6,$F8,$32,$7D,$6F,$C0,$CF,$D6,$E6,$3D,$2F
	.dc.b	$C4,$F1,$3E,$2F,$8D,$27,$E1,$BF,$23,$C7,$F8,$6E,$63,$E0,$BF,$33
	.dc.b	$C7,$F9,$BE,$5F,$CE,$F2,$E5,$7C,$C8,$F9,$7E,$57,$D2,$F2,$E4,$FD
	.dc.b	$08,$FD,$8F,$9C,$F3,$A9,$27,$EA,$F9,$7F,$8A,$9D,$47,$EA,$DA,$FE
	.dc.b	$2D,$BD,$DF,$E9,$A3,$F8,$0F,$9D,$DB,$D0,$95,$C5,$F0,$FA,$3D,$5C
	.dc.b	$C7,$37,$CB,$E8,$F9,$BD,$39,$3E,$17,$D3,$EA,$79,$5C,$C7,$8D,$F6
	.dc.b	$FA,$9E,$EF,$5F,$DF,$EB,$CA,$F5,$47,$B6,$31,$1E,$B7,$E2,$FA,$FF
	.dc.b	$1B,$DA,$F8,$FE,$DC,$AF,$85,$1F,$17,$DB,$F9,$64,$8F,$8D,$ED,$FC
	.dc.b	$EF,$7B,$E7,$FB,$F2,$7E,$54,$7C,$DF,$7F,$E9,$92,$3E,$77,$BF,$F5
	.dc.b	$9F,$AF,$F1,$31,$FA,$47,$D5,$F8,$2F,$B6,$47,$EB,$7C,$17,$F2,$EF
	.dc.b	$85,$7C,$3D,$F8,$39,$5F,$C9,$67,$2D,$F8,$7E,$04,$7F,$9A,$1F,$EE
	.dc.b	$7E,$C4,$BE,$1F,$B1,$27,$A7,$C3,$F8,$7E,$07,$C5,$F0,$92,$7C,$11
	.dc.b	$E5,$F8,$49,$1E,$37,$D1,$F0,$DE,$9F,$86,$93,$E7,$7D,$9F,$11,$ED
	.dc.b	$F8,$19,$3E,$B7,$E0,$7C,$07,$C1,$F8,$49,$3E,$F7,$E2,$7C,$37,$C5
	.dc.b	$F8,$69,$3F,$0D,$F9,$1F,$0F,$FD,$5B,$E2,$7F,$8B,$4F,$F4,$B7,$F8
	.dc.b	$93,$7D,$FD,$7C,$0F,$CF,$F8,$49,$3F,$35,$FA,$5F,$0B,$F4,$FE,$12
	.dc.b	$4F,$D1,$7E,$B7,$C3,$77,$F8,$19,$3D,$DF,$17,$C0,$78,$FE,$0E,$4F
	.dc.b	$8D,$F3,$7C,$27,$9F,$E1,$24,$F9,$DF,$57,$C2,$FA,$FE,$06,$4F,$AD
	.dc.b	$F7,$7C,$07,$BF,$E0,$E4,$FB,$DF,$85,$F0,$9F,$0F,$E1,$24,$FC,$37
	.dc.b	$E3,$7C,$2F,$C7,$F7,$64,$FC,$77,$E5,$7C,$1F,$CB,$F8,$4F,$99,$F0
	.dc.b	$72,$BE,$64,$7C,$FF,$84,$FA,$1F,$0B,$27,$E8,$47,$F8,$23,$C6,$7B
	.dc.b	$B4,$93,$FD,$FE,$1F,$03,$BF,$82,$E1,$F0,$52,$B5,$1C,$FE,$16,$47
	.dc.b	$17,$F0,$1F,$0B,$DF,$E0,$BC,$1E,$EC,$AE,$D1,$E3,$F8,$29,$1E,$17
	.dc.b	$CD,$F0,$5E,$7F,$83,$F4,$7C,$24,$AF,$34,$73,$7F,$BA,$6C,$C7,$1E
	.dc.b	$BF,$F2,$B8,$65,$DA,$B7,$95,$AD,$65,$DE,$7E,$16,$5D,$DE,$B9,$7C
	.dc.b	$9F,$81,$91,$F1,$DF,$99,$F0,$1F,$37,$E0,$3E,$77,$C1,$CA,$F9,$91
	.dc.b	$F4,$7E,$13,$E9,$7C,$24,$9F,$A1,$1F,$E3,$CF,$8A,$7B,$94,$93,$F5
	.dc.b	$4F,$B5,$F0,$9F,$6F,$E1,$24,$7D,$98,$FB,$DF,$09,$23,$EE,$3F,$81
	.dc.b	$F8,$5E,$DE,$E7,$7F,$76,$57,$F8,$FF,$24,$F8,$FE,$16,$4E,$5F,$37
	.dc.b	$C2,$F9,$FE,$16,$4F,$17,$D5,$F0,$9D,$23,$FC,$51,$87,$F9,$49,$ED
	.dc.b	$8F,$89,$E8,$43,$78,$76,$7A,$10,$E7,$7D,$9F,$47,$B9,$F2,$7D,$D9
	.dc.b	$18,$3D,$9D,$BF,$37,$E1,$24,$60,$F8,$1F,$09,$F4,$48,$FC,$3F,$84
	.dc.b	$FA,$98,$74,$8F,$5E,$18,$79,$1E,$EB,$EA,$FC,$0C,$39,$BF,$0B,$0B
	.dc.b	$DD,$FD,$FF,$5F,$F3,$CC,$F4,$3E,$96,$1D,$E7,$58,$FE,$17,$0E,$B1
	.dc.b	$C7,$1C,$5F,$A8,$87,$73,$D4,$7D,$DF,$6B,$F1,$DE,$DC,$8C,$1F,$81
	.dc.b	$F8,$4F,$C9,$7C,$24,$8C,$1F,$85,$F8,$4F,$CB,$7C,$24,$8C,$1F,$89
	.dc.b	$BE,$8F,$E0,$1F,$CE,$61,$F9,$DC,$3A,$4F,$E1,$30,$C3,$C8,$FC,$3D
	.dc.b	$F5,$7F,$13,$87,$FA,$83,$E2,$DE,$E5,$F7,$25,$7F,$A5,$E7,$2D,$F8
	.dc.b	$3E,$1F,$09,$23,$07,$E7,$FE,$13,$F6,$1F,$09,$23,$07,$3C,$7A,$BF
	.dc.b	$E8,$1C,$3F,$CD,$CD,$8F,$26,$1D,$63,$58,$61,$E6,$7E,$8A,$FB,$3F
	.dc.b	$A4,$C3,$D9,$ED,$FB,$7D,$C9,$3E,$B8,$F8,$1F,$09,$23,$DC,$FC,$3F
	.dc.b	$84,$F8,$9F,$09,$F1,$7E,$12,$57,$C3,$8E,$AF,$C5,$C3,$AB,$EB,$C3
	.dc.b	$A3,$ED,$C2,$F3,$FF,$94,$F7,$65,$DA,$BE,$76,$5D,$9F,$89,$97,$4B
	.dc.b	$E1,$95,$F5,$7F,$DD,$58,$B3,$EB,$18,$CE,$E3,$FE,$F1,$AE,$EF,$A4
	.dc.b	$EE,$43,$B9,$F7,$FD,$EF,$C2,$11,$FC,$17,$C0,$FE,$1F,$DE,$FC,$47
	.dc.b	$C2,$49,$FC,$34,$7E,$2F,$E1,$24,$7E,$25,$FC,$77,$C2,$73,$8F,$C0
	.dc.b	$7C,$0F,$E4,$4C,$6F,$F1,$6F,$A4,$FE,$17,$0F,$03,$CC,$87,$E2,$70
	.dc.b	$EA,$FF,$BB,$7E,$33,$FD,$03,$B2,$18,$25,$F0,$1F,$F0,$01,$B3,$C7
	.dc.b	$DF,$FD,$25,$C8,$FD,$13,$FA,$7F,$85,$FD,$47,$C2,$FE,$A7,$E1,$65
	.dc.b	$7E,$9E,$3A,$63,$23,$C8,$F7,$C3,$A4,$7E,$87,$0B,$97,$E1,$71,$EB
	.dc.b	$3F,$B1,$C7,$AB,$FA,$7C,$7D,$1E,$FC,$8F,$80,$FA,$FE,$0F,$D9,$EF
	.dc.b	$C9,$F8,$8F,$BF,$E0,$3E,$07,$C2,$FC,$1F,$85,$95,$F2,$63,$A3,$FE
	.dc.b	$F5,$E0,$7C,$1F,$C7,$33,$B7,$FA,$B6,$18,$F7,$AE,$0F,$77,$A1,$F5
	.dc.b	$70,$EE,$F5,$C3,$A3,$F6,$70,$AF,$87,$3E,$98,$F7,$C2,$EF,$F7,$63
	.dc.b	$EE,$67,$EC,$BE,$C4,$3B,$1F,$85,$F8,$1F,$C0,$7C,$0C,$8C,$1F,$89
	.dc.b	$F8,$0F,$C2,$11,$FC,$67,$C0,$FE,$1F,$E1,$64,$7E,$3D,$FC,$57,$C2
	.dc.b	$F3,$8F,$91,$77,$2F,$8D,$8F,$39,$F9,$98,$D9,$F9,$58,$F4,$9F,$A1
	.dc.b	$8D,$9F,$9D,$8F,$E5,$FE,$06,$57,$E7,$DF,$CD,$7C,$27,$E6,$FE,$0E
	.dc.b	$4F,$E8,$DF,$CF,$7C,$1F,$E7,$FE,$0E,$4F,$E9,$DF,$D1,$7C,$2F,$E8
	.dc.b	$E2,$1F,$AB,$3B,$9F,$83,$AE,$AD,$3F,$61,$5D,$1F,$C4,$55,$CB,$F0
	.dc.b	$D7,$FB,$39,$FD,$9F,$0F,$D9,$D7,$FC,$6B,$1F,$6E,$31,$F0,$5B,$BF
	.dc.b	$F7,$D5,$7F,$AF,$6B,$47,$EF,$1E,$2F,$43,$1F,$09,$C6,$FB,$36,$32
	.dc.b	$72,$A5,$7F,$B4,$23,$94,$63,$E1,$38,$3A,$F8,$2B,$46,$CE,$31,$CF
	.dc.b	$C2,$4C,$3B,$5F,$A5,$EA,$FF,$91,$1C,$E3,$D0,$EB,$F0,$9D,$23,$8E
	.dc.b	$3E,$03,$0F,$2F,$84,$F1,$91,$EF,$87,$37,$A9,$0F,$0D,$72,$CF,$E6
	.dc.b	$DF,$59,$F3,$E7,$EE,$F7,$E4,$7A,$9F,$83,$F0,$9F,$0B,$DC,$F8,$7F
	.dc.b	$07,$2B,$DF,$1F,$1B,$E0,$E4,$7C,$27,$E4,$FC,$1F,$CA,$F8,$0F,$97
	.dc.b	$F0,$12,$BE,$3C,$75,$7E,$4E,$1D,$5F,$93,$87,$47,$E5,$E1,$75,$FC
	.dc.b	$3E,$3F,$5B,$E0,$BE,$BF,$C1,$4B,$FA,$71,$F7,$3E,$37,$FD,$D3,$1F
	.dc.b	$80,$6D,$B9,$FF,$87,$46,$1F,$CC,$7B,$DF,$82,$F7,$E5,$FD,$E7,$F0
	.dc.b	$BF,$09,$D7,$E1,$24,$FE,$3F,$0F,$F0,$FE,$3D,$63,$FA,$AC,$7F,$BF
	.dc.b	$1F,$17,$C1,$79,$48,$FE,$2B,$1E,$EF,$E2,$B1,$B7,$FA,$6C,$7D,$9F
	.dc.b	$03,$ED,$F8,$49,$7E,$18,$F8,$1E,$E7,$C1,$F8,$19,$3E,$58,$F8,$9F
	.dc.b	$03,$F1,$7E,$06,$4F,$A6,$3E,$47,$C1,$FC,$9F,$83,$93,$ED,$8E,$AF
	.dc.b	$BF,$0E,$AF,$CD,$C3,$AB,$F0,$F0,$EA,$FD,$1C,$3E,$A7,$BB,$F5,$7E
	.dc.b	$06,$4F,$C9,$8E,$AF,$D7,$C3,$AE,$DF,$2C,$71,$32,$EF,$5F,$77,$2E
	.dc.b	$F5,$BA,$FF,$15,$4F,$D5,$F8,$0F,$C2,$7C,$04,$8C,$1F,$67,$DC,$FC
	.dc.b	$41,$1F,$B9,$EE,$FE,$2F,$E0,$A4,$7D,$F7,$F1,$DF,$05,$F8,$FE,$A9
	.dc.b	$3F,$83,$7F,$25,$F0,$1C,$E3,$F2,$58,$5D,$3C,$D8,$F4,$9F,$CB,$63
	.dc.b	$67,$D5,$8F,$F1,$6F,$8F,$FF,$0D,$D7,$F1,$A7,$FC,$46,$FF,$88,$9B
	.dc.b	$27,$6F,$C9,$5E,$D7,$3F,$7A,$57,$F2,$79,$CB,$7D,$FF,$07,$C0,$C8
	.dc.b	$C1,$BF,$7F,$C9,$EF,$C8,$C1,$CF,$1E,$AF,$9F,$1F,$54,$7A,$F0,$E9
	.dc.b	$3E,$BC,$30,$F8,$FE,$03,$F9,$AF,$C7,$FF,$3A,$8F,$E7,$8D,$7E,$14
	.dc.b	$6B,$DD,$91,$A7,$8C,$72,$CE,$4F,$27,$AE,$7F,$37,$DE,$F9,$D7,$2B
	.dc.b	$E2,$47,$D1,$F6,$64,$7C,$77,$EA,$7B,$BF,$57,$E0,$64,$FC,$B7,$EC
	.dc.b	$7C,$0F,$D9,$F8,$1F,$B5,$EE,$4A,$FA,$11,$D1,$FB,$98,$5E,$4F,$B1
	.dc.b	$8F,$49,$FC,$06,$36,$7E,$E6,$3F,$85,$F7,$FF,$0D,$EF,$CB,$FB,$91
	.dc.b	$CF,$B3,$FC,$59,$CE,$CF,$15,$F8,$9C,$FA,$56,$B3,$B3,$F8,$DC,$BF
	.dc.b	$25,$F0,$32,$3F,$10,$FE,$57,$E0,$7F,$B5,$FC,$83,$C1,$A4,$AF,$EE
	.dc.b	$90,$F0,$77,$E3,$49,$CB,$CB,$C7,$FE,$EB,$F2,$3F,$E0,$2E,$3F,$E0
	.dc.b	$4D,$DB,$FC,$8C,$DB,$FD,$11,$E2,$F2,$7C,$7E,$3C,$8F,$0C,$79,$BC
	.dc.b	$B9,$1E,$47,$D3,$E5,$F8,$3D,$FF,$0F,$BD,$2B,$FC,$19,$8F,$F8,$2F
	.dc.b	$3F,$E2,$9B,$FF,$40,$D3,$FC,$B0,$FF,$3F,$8C,$FC,$47,$C4,$F5,$7D
	.dc.b	$7F,$01,$31,$E6,$FB,$BE,$13,$DF,$F0,$12,$7B,$BF,$0B,$E0,$FF,$C2
	.dc.b	$FD,$4F,$6A,$92,$BF,$C4,$10,$F6,$FE,$4F,$C1,$FC,$AF,$84,$95,$EA
	.dc.b	$8F,$9B,$F0,$32,$3D,$AF,$D0,$F8,$0E,$E7,$83,$E0,$24,$F6,$8F,$AB
	.dc.b	$12,$3E,$2B,$F6,$3E,$13,$EC,$E3,$27,$E4,$BF,$73,$E0,$FD,$79,$49
	.dc.b	$F4,$BE,$EE,$8F,$C0,$FC,$27,$E0,$BE,$12,$57,$D2,$8F,$C2,$FC,$1C
	.dc.b	$8F,$AA,$FE,$23,$E0,$FF,$C7,$BF,$25,$FE,$55,$9F,$F2,$0B,$7A,$7F
	.dc.b	$A6,$37,$AF,$FD,$20,$BC,$7F,$C8,$FC,$24,$BF,$C0,$BF,$94,$E6,$FF
	.dc.b	$24,$F8,$8F,$84,$BD,$B2,$FF,$CA,$53,$97,$FC,$A9,$8B,$CB,$F2,$F5
	.dc.b	$49,$CC,$74,$EF,$91,$C9,$EF,$F0,$BE,$0F,$1F,$C3,$E2,$CA,$EF,$1F
	.dc.b	$03,$D4,$F8,$3E,$BC,$9F,$0C,$7A,$3C,$89,$1E,$67,$D7,$F0,$BE,$CD
	.dc.b	$A4,$FA,$9F,$7F,$C2,$7C,$CF,$6E,$4F,$B1,$F9,$FE,$E7,$C4,$F8,$59
	.dc.b	$3F,$09,$F8,$FE,$77,$C8,$F8,$2F,$93,$F0,$52,$BE,$3C,$7C,$CF,$82
	.dc.b	$F9,$BE,$A4,$9F,$97,$1F,$43,$D4,$91,$F3,$9F,$A7,$EA,$7F,$9F,$3E
	.dc.b	$4D,$AD,$F8,$99,$7F,$E8,$59,$CB,$7E,$27,$81,$1F,$B3,$E1,$7D,$EF
	.dc.b	$87,$FB,$FF,$0F,$27,$EE,$C7,$7E,$89,$1D,$5F,$17,$C2,$78,$FE,$12
	.dc.b	$4F,$85,$F3,$7C,$37,$9F,$E1,$E4,$F9,$5F,$57,$3F,$AF,$C2,$93,$E9
	.dc.b	$7D,$DE,$FF,$BF,$E1,$24,$FB,$5F,$85,$F0,$BF,$0F,$E1,$A4,$FC,$17
	.dc.b	$E3,$78,$3F,$1F,$CC,$F9,$1F,$05,$2B,$E3,$47,$CB,$F8,$5F,$99,$F0
	.dc.b	$D2,$7E,$54,$7C,$FF,$85,$91,$F3,$5F,$A5,$E5,$FF,$AB,$BE,$53,$F9
	.dc.b	$47,$47,$FB,$F5,$7C,$36,$4D,$7B,$1B,$F8,$4E,$1F,$09,$32,$FF,$5E
	.dc.b	$5F,$FA,$D5,$BF,$1B,$D4,$8F,$DD,$3C,$1F,$01,$1E,$2F,$8F,$E0,$3C
	.dc.b	$9F,$0B,$27,$AB,$E7,$F8,$5F,$47,$C2,$C9,$F0,$BE,$BF,$81,$F6,$7C
	.dc.b	$1C,$9F,$2B,$EF,$F8,$3F,$81,$F0,$5F,$07,$E0,$E5,$7A,$A3,$E2,$7C
	.dc.b	$2F,$C5,$F8,$59,$3E,$E8,$FF,$6A,$FC,$A3,$DE,$FF,$31,$73,$A9,$F0
	.dc.b	$8A,$D4,$FF,$B8,$57,$BF,$C2,$B7,$1A,$7F,$DC,$7C,$AF,$86,$D6,$08
	.dc.b	$7C,$36,$1D,$63,$F4,$A3,$1F,$0D,$23,$EA,$FC,$35,$C7,$FD,$DF,$93
	.dc.b	$E0,$BA,$AF,$83,$B1,$87,$3F,$07,$AF,$83,$B8,$FA,$DE,$3F,$0D,$CB
	.dc.b	$E1,$AC,$FB,$DE,$BE,$EC,$8F,$85,$EE,$DD,$3C,$7F,$0B,$E2,$3C,$7F
	.dc.b	$0B,$6B,$E4,$7C,$2F,$9B,$E1,$7C,$ED,$8F,$99,$EE,$7A,$8F,$5F,$BB
	.dc.b	$63,$E8,$7C,$27,$B8,$F7,$FC,$2D,$8F,$A9,$F0,$9F,$0B,$E1,$7E,$1B
	.dc.b	$63,$EE,$75,$FF,$A7,$FC,$07,$BB,$F5,$17,$BB,$63,$0E,$7E,$17,$5F
	.dc.b	$0B,$70,$E6,$74,$B4,$73,$F8,$4F,$A7,$34,$FC,$41,$DF,$DC,$F0,$7B
	.dc.b	$96,$39,$F3,$7E,$43,$04,$F9,$7E,$13,$CD,$F0,$96,$F7,$B2,$D3,$EB
	.dc.b	$F8,$4F,$67,$C2,$5B,$DD,$7F,$A5,$9B,$EE,$7C,$13,$E1,$7B,$96,$37
	.dc.b	$F0,$9F,$17,$E1,$3E,$33,$63,$9F,$C2,$7C,$93,$E5,$7C,$25,$8F,$2F
	.dc.b	$CB,$BD,$DF,$D0,$B6,$3C,$7C,$58,$9F,$A5,$F0,$92,$7C,$FF,$0B,$6B
	.dc.b	$F6,$71,$F3,$FE,$13,$EC,$7C,$2F,$D9,$6C,$78,$3E,$13,$BF,$C2,$F8
	.dc.b	$1B,$1F,$23,$E5,$9F,$01,$FB,$45,$F0,$36,$30,$FA,$3E,$1B,$D3,$F0
	.dc.b	$D6,$FD,$AF,$B3,$E1,$3D,$B5,$67,$E0,$BD,$67,$B7,$05,$9F,$13,$F1
	.dc.b	$30,$F8,$BF,$09,$67,$E4,$BF,$23,$E0,$3E,$4F,$55,$9F,$9A,$FF,$0C
	.dc.b	$F9,$77,$0D,$EB,$C3,$6B,$E9,$FC,$3F,$D2,$F8,$7F,$A6,$D8,$FA,$FF
	.dc.b	$0F,$F5,$8F,$AF,$F0,$F6,$3F,$1F,$8B,$AF,$C0,$78,$7C,$6B,$3F,$80
	.dc.b	$F8,$6F,$BE,$7E,$03,$E1,$EC,$7E,$13,$E1,$7F,$06,$7E,$13,$E1,$2C
	.dc.b	$76,$F1,$7D,$87,$B7,$CF,$B1,$F8,$CF,$84,$FC,$59,$F8,$CF,$84,$B1
	.dc.b	$FA,$4F,$98,$FF,$24,$37,$DF,$FE,$63,$18,$F0,$FF,$16,$7F,$3E,$51
	.dc.b	$FC,$DB,$DF,$55,$F6,$B0,$D3,$B7,$C4,$74,$3A,$FC,$45,$8F,$17,$C3
	.dc.b	$78,$3E,$1B,$C2,$D8,$FD,$CE,$1F,$EB,$B6,$FB,$7E,$18,$F1,$3E,$8F
	.dc.b	$86,$F4,$FC,$35,$AF,$73,$EC,$F8,$49,$1F,$07,$E1,$2F,$07,$BB,$D8
	.dc.b	$FF,$62,$7C,$27,$B7,$4F,$73,$6F,$E4,$7C,$27,$C6,$8F,$8E,$D8,$F9
	.dc.b	$9A,$7C,$A3,$E5,$EB,$63,$D7,$C5,$F2,$78,$7E,$53,$63,$F8,$BF,$CC
	.dc.b	$FF,$5E,$D3,$C9,$C8,$F8,$6F,$EC,$B3,$9D,$BE,$8D,$7D,$59,$FB,$3F
	.dc.b	$09,$30,$FB,$DF,$09,$6B,$DD,$87,$FA,$E6,$7B,$F4,$74,$F1,$3A,$BD
	.dc.b	$BC,$4E,$FE,$0D,$AF,$2B,$F8,$6F,$86,$FC,$3F,$C3,$59,$FC,$63,$FD
	.dc.b	$EB,$E6,$BF,$9F,$C7,$F3,$25,$F0,$99,$F8,$5D,$3F,$90,$98,$7E,$57
	.dc.b	$C0,$BB,$7D,$5B,$FE,$A3,$3C,$FE,$1B,$B7,$C3,$F7,$7A,$47,$7C,$2D
	.dc.b	$F1,$79,$F2,$7C,$81,$CB,$F9,$AA,$F0,$1F,$A5,$3D,$5E,$2F,$AF,$C5
	.dc.b	$B1,$EB,$7D,$DF,$0F,$23,$DD,$F0,$F7,$2F,$47,$C3,$FC,$3F,$87,$F8
	.dc.b	$8D,$AF,$89,$E7,$7C,$7F,$3B,$E4,$36,$3E,$46,$3C,$F5,$F4,$70,$B3
	.dc.b	$C5,$7F,$78,$8F,$EB,$AF,$CF,$CB,$BB,$4F,$A5,$F0,$DF,$54,$FA,$DF
	.dc.b	$0B,$63,$F2,$3F,$03,$F8,$E3,$F1,$FE,$FD,$8F,$CA,$FB,$BF,$92,$3F
	.dc.b	$27,$EE,$D8,$FB,$DF,$0F,$F8,$13,$F0,$5F,$0F,$63,$F0,$5C,$7F,$85
	.dc.b	$3F,$0D,$F0,$D6,$3F,$0D,$F0,$DF,$89,$3F,$15,$F0,$D6,$3F,$15,$EE
	.dc.b	$FE,$34,$FC,$77,$BD,$63,$A9,$30,$FC,$46,$1D,$5F,$C5,$61,$CD,$FC
	.dc.b	$66,$17,$93,$F0,$9C,$0D,$BF,$CA,$6D,$C7,$FC,$16,$FF,$75,$BF,$E1
	.dc.b	$0E,$5F,$CB,$1D,$EB,$EB,$E3,$6B,$EB,$63,$DD,$FB,$18,$DB,$FF,$35
	.dc.b	$E1,$FE,$94,$9E,$FC,$1E,$2E,$0F,$1B,$E4,$F8,$3F,$2F,$C1,$DA,$F2
	.dc.b	$BE,$8F,$83,$F4,$FC,$35,$9F,$4B,$EC,$F8,$29,$1E,$CF,$85,$BA,$7B
	.dc.b	$31,$EB,$3F,$01,$B1,$F0,$30,$EB,$3F,$11,$B1,$F1,$3E,$13,$E4,$FC
	.dc.b	$27,$CA,$6D,$7C,$AF,$33,$E6,$9F,$3B,$CC,$B1,$F3,$BE,$0F,$E8,$FC
	.dc.b	$2F,$D2,$6C,$7D,$2F,$3F,$EA,$FA,$1F,$59,$B1,$D4,$99,$FD,$6C,$3A
	.dc.b	$BF,$62,$F9,$B6,$7E,$EE,$1F,$80,$F8,$3F,$C0,$FC,$1D,$9F,$C0,$BF
	.dc.b	$E9,$6F,$9A,$7C,$0D,$F8,$9C,$36,$BF,$CA,$33,$FD,$A1,$BE,$DF,$1F
	.dc.b	$6F,$93,$CF,$E1,$7F,$1F,$F0,$D6,$BF,$1E,$F4,$79,$E3,$75,$FE,$B1
	.dc.b	$8F,$F7,$7F,$07,$F8,$E3,$DE,$33,$8F,$7B,$F0,$9E,$4F,$86,$F5,$7C
	.dc.b	$37,$AD,$B1,$E8,$D7,$D5,$7D,$9B,$8F,$6D,$7D,$35,$EE,$6D,$79,$3E
	.dc.b	$0F,$E3,$1F,$1F,$E1,$2D,$7C,$4F,$81,$F9,$47,$CB,$F8,$6B,$1F,$23
	.dc.b	$E1,$BE,$77,$C3,$7C,$F6,$C7,$CC,$F8,$6F,$A4,$7D,$3F,$86,$B1,$D4
	.dc.b	$99,$FC,$6C,$3A,$BF,$22,$F9,$B6,$7E,$BE,$1F,$77,$E1,$BE,$F7,$C3
	.dc.b	$59,$FB,$6F,$E0,$58,$7D,$EF,$82,$B7,$F5,$3E,$17,$F0,$DF,$0B,$F8
	.dc.b	$76,$D7,$FA,$53,$E6,$FF,$B2,$D7,$F8,$45,$FE,$CB,$19,$F8,$6F,$F7
	.dc.b	$1C,$62,$7E,$FC,$F6,$E4,$FB,$F8,$D8,$FF,$5C,$F4,$B1,$EF,$7E,$39
	.dc.b	$FC,$9E,$5D,$3E,$1B,$AF,$C3,$59,$F2,$3E,$0F,$7F,$C3,$EF,$D9,$F4
	.dc.b	$3E,$4F,$82,$F2,$FB,$F6,$7D,$8F,$A3,$E1,$3D,$3F,$09,$67,$E0,$3C
	.dc.b	$F4,$F8,$18,$5E,$0F,$EE,$9C,$3F,$D7,$6F,$E3,$65,$D9,$A7,$63,$E5
	.dc.b	$7C,$1F,$C6,$F8,$4F,$8E,$D8,$F9,$DA,$FA,$9A,$7A,$8E,$79,$7D,$1D
	.dc.b	$7D,$4D,$8F,$AD,$F0,$7F,$4B,$DA,$FA,$6D,$8F,$B5,$F0,$1F,$5B,$E0
	.dc.b	$3E,$BB,$63,$EF,$7C,$17,$DA,$F8,$2F,$B6,$D8,$FC,$17,$C0,$7D,$EF
	.dc.b	$80,$FB,$ED,$8E,$A7,$E1,$F0,$EA,$D3,$A9,$F8,$BB,$EA,$D3,$F1,$D9
	.dc.b	$FE,$28,$FC,$5F,$C0,$D8,$FF,$72,$FC,$E7,$F4,$F9,$FF,$31,$2F,$82
	.dc.b	$FF,$13,$4A,$5C,$AA,$E0,$FF,$8A,$DE,$13,$D3,$E0,$F8,$D7,$09,$FE
	.dc.b	$CD,$5B,$F8,$4C,$BC,$EA,$CB,$DE,$BF,$05,$86,$E3,$E8,$F8,$4D,$E7
	.dc.b	$C1,$B1,$E1,$26,$7E,$CC,$FA,$E5,$DB,$E0,$65,$7C,$2F,$84,$B7,$F0
	.dc.b	$BD,$CF,$19,$E4,$F7,$2D,$7C,$8F,$81,$F3,$9E,$8F,$82,$B1,$F3,$3E
	.dc.b	$13,$D6,$7B,$3E,$12,$C7,$D0,$C3,$9E,$3F,$43,$1B,$C7,$F2,$F2,$E7
	.dc.b	$3F,$53,$2B,$3F,$3F,$E1,$3E,$3F,$C1,$7C,$86,$E5,$D4,$97,$D5,$B5
	.dc.b	$D5,$F5,$1F,$82,$BE,$AF,$DF,$8F,$BD,$F0,$7F,$50,$FA,$BF,$09,$67
	.dc.b	$46,$FD,$87,$C1,$7D,$C6,$C7,$E3,$BE,$0B,$13,$F7,$BE,$13,$EF,$FC
	.dc.b	$25,$AF,$C9,$BC,$7D,$F9,$1D,$B5,$FC,$8E,$2F,$83,$CC,$7F,$AE,$63
	.dc.b	$BB,$71,$FC,$5E,$5E,$0B,$FC,$FE,$5E,$06,$9E,$03,$B6,$BF,$EA,$FF
	.dc.b	$81,$F5,$7C,$27,$AD,$B3,$AC,$9F,$05,$EF,$F8,$2F,$80,$D9,$D7,$C1
	.dc.b	$FC,$3F,$84,$F8,$8D,$8E,$5F,$03,$F1,$CF,$91,$F0,$36,$3B,$5F,$3E
	.dc.b	$0F,$73,$75,$FF,$4D,$66,$C7,$A5,$7C,$36,$CE,$BE,$0B,$EA,$1F,$56
	.dc.b	$AE,$1E,$BF,$73,$EC,$7C,$17,$D9,$6C,$7B,$FE,$0B,$EE,$7C,$27,$DD
	.dc.b	$6C,$7C,$3F,$84,$E8,$D3,$E3,$61,$CD,$F9,$18,$5D,$3E,$C3,$CE,$3E
	.dc.b	$66,$16,$7E,$E3,$F8,$BF,$83,$93,$F4,$7E,$0A,$D7,$E0,$73,$F3,$B6
	.dc.b	$7B,$9E,$93,$A9,$5E,$AF,$DB,$3F,$30,$7F,$14,$F9,$D7,$C0,$E2,$3F
	.dc.b	$0B,$1F,$ED,$29,$FE,$47,$1F,$A7,$87,$C1,$71,$F8,$3E,$4F,$3F,$85
	.dc.b	$95,$CB,$E2,$2E,$1C,$8F,$E5,$B8,$78,$27,$BE,$3E,$3F,$85,$F2,$3D
	.dc.b	$63,$C7,$8D,$BE,$5C,$8C,$3C,$75,$E8,$CB,$BC,$FF,$B8,$67,$FC,$18
	.dc.b	$D8,$F8,$1F,$03,$F0,$7E,$02,$DF,$C0,$7E,$27,$C1,$48,$F8,$7F,$05
	.dc.b	$72,$F0,$E5,$FD,$E5,$BF,$07,$F2,$FE,$17,$E6,$3F,$37,$E1,$3A,$CF
	.dc.b	$06,$EB,$E5,$CB,$A3,$E7,$C6,$E5,$F3,$F1,$E6,$FD,$46,$E1,$E2,$F8
	.dc.b	$4F,$B4,$7D,$BF,$81,$B5,$F6,$BE,$13,$EF,$7C,$17,$DF,$6C,$7D,$EF
	.dc.b	$84,$FC,$11,$F8,$3F,$83,$B1,$F8,$2B,$E7,$8F,$C0,$6E,$9F,$37,$0E
	.dc.b	$73,$F1,$1B,$1F,$47,$0E,$73,$F2,$30,$B3,$F5,$6F,$D1,$87,$E4,$3E
	.dc.b	$12,$DF,$38,$FC,$9E,$5E,$96,$9E,$92,$47,$56,$D9,$FE,$7D,$EB,$B8
	.dc.b	$7F,$B9,$B6,$3F,$91,$AE,$3C,$CE,$B9,$F7,$D1,$6F,$6F,$2E,$AE,$7D
	.dc.b	$36,$79,$BF,$E0,$0F,$A0,$FF,$04,$C7,$F4,$F7,$FC,$13,$A7,$E6,$9C
	.dc.b	$BF,$E1,$07,$3A,$7A,$63,$1E,$F7,$3F,$7F,$A3,$6F,$F6,$F3,$FB,$BF
	.dc.b	$85,$FD,$E3,$63,$F9,$FE,$1D,$B0,$3D,$A3,$C9,$F0,$3E,$7F,$82,$EB
	.dc.b	$1C,$32,$B8,$79,$DE,$B1,$FC,$1B,$1B,$3F,$A4,$7F,$C3,$FF,$40,$F1
	.dc.b	$AF,$93,$86,$DF,$F5,$29,$78,$FC,$0E,$3E,$4D,$9D,$79,$3D,$0E,$BE
	.dc.b	$55,$8E,$5E,$77,$83,$CE,$F0,$B6,$3B,$79,$DE,$4F,$43,$CA,$D8,$F1
	.dc.b	$79,$FE,$83,$D3,$E8,$58,$F3,$7A,$3E,$CF,$4B,$DA,$D8,$FE,$F1,$C6
	.dc.b	$FA,$0F,$E2,$6F,$F1,$6E,$1F,$F6,$B7,$4F,$DB,$8D,$75,$FC,$6F,$63
	.dc.b	$E3,$FB,$13,$1F,$89,$EC,$5B,$FC,$37,$B1,$F8,$B3,$F1,$9E,$C5,$AF
	.dc.b	$0E,$9E,$0C,$FF,$10,$34,$F0,$CF,$F9,$77,$0F,$0B,$F4,$73,$FC,$B7
	.dc.b	$B3,$F9,$7F,$6E,$CF,$E2,$1F,$CD,$7B,$72,$3F,$17,$EE,$DC,$3F,$9E
	.dc.b	$1D,$0F,$81,$87,$F9,$91,$C6,$5E,$4C,$BC,$71,$F9,$FF,$7B,$F4,$DF
	.dc.b	$01,$FA,$76,$E3,$F9,$8F,$33,$F5,$5F,$01,$FA,$B6,$C7,$C0,$E4,$F9
	.dc.b	$4D,$3D,$87,$E8,$3E,$03,$F6,$5F,$05,$FB,$36,$C7,$E9,$3E,$0B,$F6
	.dc.b	$DF,$09,$FB,$76,$C7,$3D,$3D,$6D,$DB,$D5,$BF,$EF,$7E,$12,$4E,$4F
	.dc.b	$DF,$3F,$C0,$7E,$17,$F8,$17,$C3,$59,$FD,$9B,$F1,$79,$FE,$C4,$7F
	.dc.b	$94,$F8,$FE,$2E,$5F,$22,$31,$8E,$F7,$EF,$3F,$42,$F8,$BF,$C2,$72
	.dc.b	$D5,$FE,$13,$E0,$AC,$ED,$BA,$7A,$A7,$D1,$3F,$EA,$0C,$BC,$38,$74
	.dc.b	$AF,$E0,$F8,$70,$BF,$69,$C2,$FF,$41,$86,$5F,$F5,$55,$FF,$C0,$0B
	.dc.b	$FD,$03,$CF,$0E,$B8,$FD,$8A,$F5,$3E,$CA,$DB,$70,$FC,$C5,$FB,$AB
	.dc.b	$BC,$78,$63,$B3,$A9,$E9,$5C,$A3,$F0,$75,$D9,$FE,$47,$8F,$9E,$7D
	.dc.b	$B1,$CF,$1F,$55,$76,$38,$1F,$82,$2C,$FC,$96,$CF,$C4,$69,$F4,$DE
	.dc.b	$84,$73,$1F,$B5,$CB,$C5,$97,$C4,$AF,$29,$F5,$27,$CB,$5D,$0F,$23
	.dc.b	$ED,$7E,$E4,$6E,$79,$47,$9E,$74,$7F,$04,$F3,$D8,$FC,$16,$DF,$96
	.dc.b	$76,$FD,$0C,$7C,$B8,$C8,$F2,$57,$0C,$3D,$6D,$BE,$31,$F8,$2B,$F9
	.dc.b	$79,$FD,$68,$FC,$2E,$7F,$5B,$0F,$B0,$7D,$6B,$E2,$7C,$02,$E1,$D6
	.dc.b	$37,$5F,$00,$F8,$97,$D5,$FD,$41,$F9,$BC,$7D,$0F,$9B,$2F,$80,$79
	.dc.b	$32,$97,$F2,$6B,$78,$F8,$2B,$73,$F1,$4F,$0C,$7E,$72,$B7,$18,$3F
	.dc.b	$6E,$7C,$69,$E3,$7E,$69,$EE,$FE,$0B,$3F,$AF,$7F,$8B,$3C,$17,$C4
	.dc.b	$F0,$4F,$C8,$7F,$78,$79,$C8,$FB,$5E,$0F,$77,$8B,$F1,$23,$F2,$11
	.dc.b	$C1,$EF,$1C,$4E,$04,$7E,$9C,$F9,$A7,$DB,$3B,$9E,$27,$E0,$1F,$CA
	.dc.b	$C7,$C3,$7F,$04,$F7,$31,$1E,$28,$E5,$18,$CB,$EF,$47,$BC,$FB,$D3
	.dc.b	$D7,$2F,$BC,$7F,$0B,$CB,$E8,$CE,$E7,$F3,$D9,$7D,$07,$86,$5F,$41
	.dc.b	$F5,$D7,$0A,$F8,$98,$78,$1B,$7B,$C7,$D5,$1E,$C3,$B3,$F0,$AF,$9B
	.dc.b	$F4,$1F,$86,$F0,$C7,$CF,$9F,$F3,$FC,$BC,$11,$E9,$9E,$07,$C1,$C3
	.dc.b	$FA,$36,$1F,$E4,$5D,$2E,$3F,$B9,$78,$E3,$FA,$99,$F1,$9F,$C3,$CF
	.dc.b	$96,$7F,$92,$F6,$FE,$B6,$FF,$93,$DB,$5E,$88,$FC,$96,$9F,$CB,$DE
	.dc.b	$51,$EA,$7F,$CA,$9C,$5F,$DC,$1F,$F2,$C3,$63,$91,$32,$FF,$2F,$70
	.dc.b	$FF,$6E,$AF,$F2,$F3,$67,$F8,$97,$3F,$F8,$25,$FE,$29,$CF,$6F,$F0
	.dc.b	$BC,$3F,$8F,$C3,$F7,$51,$FE,$5A,$E6,$FF,$02,$C7,$F9,$A5,$B7,$FC
	.dc.b	$77,$AF,$FC,$49,$3F,$E6,$DE,$BB,$3E,$58,$F8,$EF,$E4,$4F,$E4,$9D
	.dc.b	$7F,$E2,$1C,$3F,$CC,$ED,$9F,$F1,$D7,$77,$F8,$FD,$B3,$FE,$59,$EF
	.dc.b	$FA,$6F,$F1,$93,$FB,$D7,$7F,$F8,$E7,$0F,$F1,$07,$7D,$9F,$EF,$DE
	.dc.b	$27,$F9,$57,$0F,$F2,$33,$6B,$FC,$4D,$E1,$FF,$95,$67,$FC,$94,$D8
	.dc.b	$FF,$15,$78,$7F,$E5,$59,$FF,$27,$78,$76,$7F,$C5,$DE,$57,$F9,$BA
	.dc.b	$7F,$CA,$4D,$8F,$F1,$97,$95,$FE,$6D,$9F,$F2,$B7,$95,$67,$FC,$6D
	.dc.b	$E5,$7F,$9B,$67,$FC,$B5,$E5,$59,$FF,$26,$F5,$7F,$3B,$4D,$3F,$DD
	.dc.b	$FB,$7F,$AC,$3C,$1F,$F4,$1B,$0F,$F4,$07,$C3,$5C,$3F,$CC,$DF,$0D
	.dc.b	$FE,$88,$F8,$6F,$F4,$4B,$6B,$FD,$6B,$7F,$EB,$83,$FD,$73,$F0,$D6
	.dc.b	$3F,$D1,$47,$3C,$BF,$C9,$D8,$D8,$FF,$16,$5F,$3C,$7B,$C7,$FA,$47
	.dc.b	$1F,$F5,$FB,$FE,$C2,$F6,$3F,$D8,$BE,$CE,$29,$8F,$C0,$F6,$7F,$D8
	.dc.b	$77,$FE,$C8,$6C,$EA,$3F,$D9,$FE,$C7,$FB,$11,$FF,$66,$B6,$34,$74
	.dc.b	$D9,$EE,$FF,$B5,$0F,$F6,$AF,$C0,$DB,$FF,$6A,$FC,$27,$FB,$60,$FF
	.dc.b	$6C,$FC,$35,$8F,$F6,$CF,$C2,$FF,$B7,$3E,$17,$FD,$BA,$D8,$FF,$6E
	.dc.b	$E1,$FE,$E0,$F8,$5E,$99,$F8,$B6,$FB,$31,$CE,$31,$A3,$3E,$F8,$78
	.dc.b	$8E,$E7,$FB,$A3,$2E,$ED,$3B,$9F,$EE,$CB,$EF,$87,$E2,$1E,$CF,$E2
	.dc.b	$AB,$AE,$1E,$07,$9B,$E2,$BB,$3F,$90,$7F,$DF,$3E,$DC,$C3,$FD,$F3
	.dc.b	$F0,$96,$FD,$3F,$09,$FE,$FE,$F8,$4F,$F7,$FB,$6B,$FD,$FF,$F0,$BF
	.dc.b	$EA,$65,$5F,$4E,$30,$FD,$5F,$85,$B9,$FE,$7A,$6E,$DC,$2F,$8C,$FE
	.dc.b	$92,$FC,$4D,$8C,$9A,$96,$1E,$3A,$F3,$CE,$9F,$C2,$63,$DB,$09,$39
	.dc.b	$C3,$F5,$F8,$E2,$FE,$D7,$D0,$BE,$0B,$CE,$FF,$6A,$A3,$9B,$E4,$E4
	.dc.b	$7A,$DC,$0E,$3F,$09,$67,$7F,$09,$D0,$EB,$F0,$B6,$39,$FB,$5E,$0F
	.dc.b	$77,$C2,$D8,$EF,$87,$3D,$3C,$F8,$DC,$FF,$DA,$F9,$B3,$E9,$5E,$CC
	.dc.b	$ED,$6B,$DA,$F6,$9E,$EF,$7A,$E7,$EC,$F8,$1F,$82,$7C,$2E,$4B,$1F
	.dc.b	$03,$E0,$7E,$2F,$B9,$F1,$9B,$1F,$13,$E0,$7E,$4F,$B5,$F2,$9B,$1D
	.dc.b	$49,$AF,$D0,$C3,$AB,$F4,$AF,$9B,$67,$E7,$E1,$F5,$3D,$C9,$1F,$4B
	.dc.b	$1B,$5F,$6B,$2E,$7A,$FC,$4C,$AD,$FE,$3B,$81,$9F,$6A,$F9,$2D,$9D
	.dc.b	$7C,$17,$E0,$8F,$C1,$FB,$D6,$3F,$15,$F4,$4F,$63,$F3,$AD,$8F,$C2
	.dc.b	$FB,$D8,$9F,$C5,$FB,$D2,$7F,$13,$EF,$5D,$FF,$85,$61,$FE,$FC,$D0
	.dc.b	$D1,$EE,$76,$F8,$0E,$98,$78,$B2,$FC,$BE,$78,$C7,$F3,$79,$F5,$CB
	.dc.b	$F0,$B9,$DC,$3F,$5F,$9B,$DF,$F4,$FB,$FE,$A6,$CF,$BB,$DE,$F6,$9E
	.dc.b	$EF,$84,$B1,$F0,$BD,$FF,$83,$EF,$FC,$26,$C7,$C6,$F8,$0F,$8B,$F0
	.dc.b	$BF,$19,$B1,$D4,$9B,$FC,$2C,$3A,$BF,$12,$F9,$B6,$7E,$8E,$1F,$43
	.dc.b	$DE,$91,$F5,$3D,$EB,$5F,$33,$DB,$FA,$BE,$DF,$D6,$6D,$7D,$AF,$81
	.dc.b	$FB,$3F,$0D,$F6,$9B,$1F,$7B,$E0,$BE,$EF,$C1,$7D,$E6,$C7,$E0,$AF
	.dc.b	$9E,$3D,$E3,$F0,$B8,$F3,$7F,$0F,$8D,$9F,$BE,$FE,$22,$21,$F8,$CF
	.dc.b	$80,$B3,$CC,$FC,$2D,$7A,$67,$A9,$E9,$3A,$96,$3A,$BF,$C7,$FE,$8F
	.dc.b	$FC,$1C,$7F,$83,$23,$FB,$9B,$8F,$82,$FE,$6A,$BE,$02,$98,$F8,$8B
	.dc.b	$7C,$FE,$07,$7F,$0D,$C1,$B1,$DF,$E1,$39,$9D,$3E,$12,$C7,$8F,$0E
	.dc.b	$79,$FF,$69,$C9,$96,$3B,$D7,$FD,$12,$59,$76,$BE,$0D,$AD,$C7,$B3
	.dc.b	$DE,$F5,$FC,$27,$B1,$B8,$7C,$3F,$81,$F7,$9F,$03,$E0,$6C,$7C,$7F
	.dc.b	$84,$F8,$7F,$09,$F1,$1B,$1F,$2F,$E0,$BE,$39,$F2,$3E,$12,$C7,$52
	.dc.b	$67,$ED,$C3,$AB,$EF,$BE,$6D,$9F,$AD,$87,$D2,$F8,$19,$1F,$67,$E0
	.dc.b	$6D,$7F,$83,$F3,$7B,$7F,$DB,$1B,$5F,$D7,$5F,$EA,$D3,$B9,$FB,$AF
	.dc.b	$DE,$F8,$4B,$5F,$85,$7F,$03,$F0,$92,$3F,$11,$F0,$96,$B9,$47,$5B
	.dc.b	$F0,$7C,$27,$85,$E7,$3F,$6F,$0B,$C7,$F5,$07,$2A,$3D,$1F,$03,$31
	.dc.b	$F4,$7C,$0D,$C3,$B9,$30,$E5,$AF,$B1,$B9,$76,$BF,$C3,$63,$6F,$F3
	.dc.b	$AF,$C5,$F8,$49,$3F,$17,$E0,$ED,$7C,$BF,$81,$F9,$47,$CB,$F8,$3B
	.dc.b	$5F,$2F,$E0,$FE,$77,$C2,$7C,$F6,$C7,$CF,$F7,$BE,$97,$C2,$7D,$36
	.dc.b	$C7,$D3,$C3,$9E,$3F,$67,$1B,$B7,$B3,$1E,$73,$F7,$5B,$1F,$03,$E0
	.dc.b	$BE,$F7,$C1,$7D,$F6,$E3,$F7,$FE,$07,$F0,$5F,$03,$F8,$36,$C7,$E0
	.dc.b	$FE,$13,$F0,$DF,$09,$F8,$76,$C7,$E1,$FD,$FF,$C5,$1F,$8B,$F7,$EC
	.dc.b	$75,$3F,$19,$87,$56,$9D,$4F,$C8,$5F,$56,$9F,$E2,$FF,$A4,$FE,$C9
	.dc.b	$B7,$E6,$7E,$07,$FC,$87,$18,$DB,$EE,$C6,$3D,$A9,$7C,$7D,$AB,$87
	.dc.b	$F9,$BF,$0F,$EB,$13,$D7,$E0,$39,$FC,$07,$47,$AF,$C1,$F6,$F8,$4B
	.dc.b	$5E,$37,$9E,$9F,$95,$C1,$9F,$93,$3C,$DF,$F9,$52,$7B,$E7,$E1,$CB
	.dc.b	$BC,$F8,$4F,$77,$C1,$7B,$0F,$6F,$C0,$58,$F8,$58,$F4,$BF,$F5,$1E
	.dc.b	$0C,$71,$92,$3B,$BE,$5C,$BB,$D7,$87,$0F,$99,$F0,$3F,$28,$F9,$7F
	.dc.b	$05,$63,$E8,$7B,$DF,$38,$F9,$FE,$FD,$8F,$A9,$F0,$5F,$48,$FA,$7F
	.dc.b	$05,$63,$EC,$7C,$17,$D6,$3E,$BF,$C0,$D8,$EA,$4C,$FE,$A6,$1D,$5F
	.dc.b	$AD,$87,$37,$EC,$61,$75,$FA,$1F,$05,$F8,$2F,$81,$FC,$1B,$71,$FF
	.dc.b	$3B,$7D,$2F,$FB,$94,$FE,$B1,$47,$1F,$05,$F8,$A9,$FC,$5F,$C1,$D7
	.dc.b	$F2,$0F,$FA,$42,$7F,$A4,$C7,$4F,$7B,$1F,$07,$97,$5F,$0F,$BC,$2D
	.dc.b	$78,$9E,$9A,$FE,$47,$1B,$97,$E2,$5E,$DF,$05,$2F,$D3,$F0,$56,$FF
	.dc.b	$D8,$DA,$3D,$DF,$27,$BB,$E5,$6D,$FC,$0F,$81,$F4,$1E,$9F,$82,$B1
	.dc.b	$F1,$3E,$13,$D9,$F0,$9E,$D6,$C7,$C8,$BE,$79,$78,$1B,$B7,$F3,$3C
	.dc.b	$D9,$74,$AF,$2E,$56,$B5,$EC,$FC,$9F,$82,$F9,$4D,$CB,$B1,$F5,$B5
	.dc.b	$F5,$34,$F5,$13,$0E,$AD,$BF,$17,$C2,$7D,$53,$EB,$7C,$25,$9F,$C0
	.dc.b	$FC,$17,$D9,$3E,$D7,$C1,$58,$FC,$2F,$C2,$7D,$DF,$84,$FB,$CD,$8F
	.dc.b	$C4,$E5,$F8,$13,$F0,$5E,$FD,$8E,$A4,$8F,$AB,$7D,$5F,$AF,$5C,$DB
	.dc.b	$3F,$94,$BF,$F6,$DF,$C8,$BE,$02,$98,$F8,$0B,$5F,$EC,$F9,$7C,$1F
	.dc.b	$03,$8F,$C1,$D9,$D7,$C1,$F4,$3A,$FC,$2D,$8E,$5E,$FF,$83,$E1,$3C
	.dc.b	$2D,$8E,$DF,$07,$E4,$3C,$BF,$09,$63,$C5,$F0,$9E,$8F,$84,$F4,$B6
	.dc.b	$3C,$DF,$07,$EC,$F8,$4F,$6B,$63,$A9,$38,$DD,$5F,$DF,$78,$59,$77
	.dc.b	$9C,$D7,$FB,$F2,$58,$F7,$9C,$9C,$A3,$33,$F1,$F2,$F9,$BF,$05,$F3
	.dc.b	$BE,$0A,$CF,$CB,$7E,$8F,$C2,$48,$F9,$DF,$07,$71,$EB,$A3,$27,$C2
	.dc.b	$7D,$8F,$84,$FB,$2D,$C1,$7C,$3D,$5F,$E6,$31,$8C,$11,$89,$F8,$2B
	.dc.b	$DB,$FF,$0F,$56,$E3,$55,$C3,$E0,$B3,$7C,$3E,$0A,$CE,$6F,$B1,$C6
	.dc.b	$33,$7F,$D9,$63,$E5,$7C,$26,$3E,$1F,$F2,$2E,$BD,$FD,$FB,$F6,$B9
	.dc.b	$E5,$E6,$8E,$7F,$09,$D3,$E1,$2D,$6B,$4F,$6C,$78,$3E,$13,$C3,$F0
	.dc.b	$96,$B5,$AF,$F5,$76,$FC,$17,$9B,$E0,$BC,$ED,$8D,$FC,$1F,$A8,$F5
	.dc.b	$D5,$8E,$7F,$07,$EE,$3D,$FF,$07,$63,$C9,$F2,$4F,$80,$F8,$EF,$F4
	.dc.b	$23,$11,$AF,$80,$F8,$EF,$C8,$F8,$0A,$FA,$1F,$97,$F0,$72,$3D,$7F
	.dc.b	$09,$79,$7F,$0F,$83,$DF,$EE,$78,$3E,$02,$DF,$93,$E0,$BC,$67,$93
	.dc.b	$E0,$AC,$7A,$3E,$0F,$CE,$7A,$3E,$12,$C7,$B3,$DF,$F5,$FC,$07,$B1
	.dc.b	$B1,$F0,$3E,$13,$DE,$7C,$0F,$81,$B1,$F1,$3E,$0F,$E1,$9F,$13,$E0
	.dc.b	$EC,$7C,$8F,$81,$F8,$E7,$C8,$F8,$0B,$1F,$33,$E0,$3E,$5F,$C0,$7C
	.dc.b	$C6,$C7,$D0,$F8,$3F,$9E,$7D,$0F,$83,$B1,$F8,$7F,$93,$FE,$36,$DF
	.dc.b	$81,$FC,$DC,$3E,$06,$C6,$3E,$07,$3F,$03,$6F,$93,$F7,$3A,$FE,$EF
	.dc.b	$C2,$59,$FC,$0B,$D3,$DC,$91,$E1,$F7,$2D,$7E,$FB,$37,$C0,$F8,$8F
	.dc.b	$1F,$C1,$5B,$F4,$7C,$27,$9B,$E1,$3C,$ED,$8F,$67,$B5,$EA,$F6,$FD
	.dc.b	$6D,$8F,$81,$F0,$5E,$E3,$DF,$F0,$76,$3E,$27,$C2,$7C,$23,$E1,$FC
	.dc.b	$25,$8F,$D7,$7D,$37,$EE,$F7,$FE,$53,$BB,$DC,$FE,$07,$39,$F6,$53
	.dc.b	$9B,$FA,$1F,$0B,$F3,$8F,$9F,$F0,$B6,$3E,$A7,$C2,$FD,$2F,$85,$FA
	.dc.b	$6D,$8F,$0F,$B9,$D8,$EF,$EE,$58,$F2,$FC,$47,$88,$F1,$FC,$2D,$8F
	.dc.b	$4F,$C2,$F9,$8F,$3F,$C2,$D8,$F6,$FB,$9E,$A3,$D7,$EE,$58,$F8,$3F
	.dc.b	$09,$EE,$3D,$FF,$09,$63,$E2,$FC,$27,$C2,$3E,$1F,$C2,$58,$F9,$3E
	.dc.b	$EF,$C6,$F7,$7E,$3B,$63,$E6,$FC,$1F,$CA,$3E,$5F,$C2,$58,$FA,$3F
	.dc.b	$09,$F3,$8F,$9F,$F0,$96,$3F,$91,$78,$3F,$E4,$6E,$EF,$E9,$2B,$DB
	.dc.b	$FE,$6E,$E7,$B7,$FD,$B8,$E6,$B8,$FC,$26,$8D,$FC,$25,$8F,$C0,$9F
	.dc.b	$78,$FB,$FF,$09,$5F,$0F,$B5,$D8,$EF,$ED,$58,$F2,$FC,$1F,$88,$F1
	.dc.b	$FC,$1D,$8F,$4F,$C1,$F9,$BE,$0F,$CE,$D8,$F6,$FB,$5E,$AF,$6B,$D6
	.dc.b	$D8,$F8,$3F,$03,$EE,$F8,$2F,$7B,$63,$E2,$FC,$17,$C2,$F8,$2F,$86
	.dc.b	$D8,$F9,$3C,$BF,$DA,$8F,$92,$5E,$4F,$8E,$D8,$FE,$D9,$C8,$F6,$B1
	.dc.b	$3F,$E0,$C8,$7C,$16,$2D,$18,$9F,$AF,$39,$87,$C1,$5F,$81,$C7,$C1
	.dc.b	$4B,$FB,$B1,$CF,$E9,$FF,$C7,$11,$FD,$85,$7B,$79,$35,$ED,$EF,$E0
	.dc.b	$B8,$7C,$1C,$BE,$11,$CF,$E0,$E4,$73,$7B,$7C,$1F,$7F,$72,$4F,$77
	.dc.b	$C5,$EE,$78,$FE,$0F,$C9,$F0,$72,$BC,$91,$E7,$F8,$3F,$47,$C1,$C9
	.dc.b	$F4,$47,$AF,$DA,$91,$EB,$7D,$DE,$DF,$BF,$E0,$BE,$07,$C1,$CA,$F8
	.dc.b	$11,$F0,$F9,$FE,$27,$C1,$C9,$F8,$91,$F4,$7E,$A1,$ED,$DF,$6E,$57
	.dc.b	$F8,$26,$72,$DF,$87,$F9,$DF,$0B,$23,$07,$CF,$F8,$3F,$A5,$F0,$72
	.dc.b	$30,$73,$F6,$FC,$1E,$DC,$8C,$1D,$FE,$0F,$C9,$F0,$72,$30,$78,$FE
	.dc.b	$0F,$D1,$F0,$72,$30,$79,$FD,$BF,$61,$1F,$57,$B7,$EF,$F8,$3F,$81
	.dc.b	$F0,$72,$7D,$D1,$F0,$FE,$0E,$47,$C1,$7E,$37,$C1,$FC,$7F,$3E,$4F
	.dc.b	$C5,$7E,$57,$B9,$F2,$FE,$0B,$E6,$7C,$1C,$AF,$95,$1F,$3F,$E0,$FE
	.dc.b	$87,$C1,$C9,$F9,$D1,$FA,$3F,$9B,$7B,$97,$DD,$95,$FE,$3B,$9C,$B7
	.dc.b	$E1,$F8,$7C,$2C,$8C,$1B,$F8,$5E,$84,$79,$7C,$2F,$7F,$73,$C1,$EE
	.dc.b	$C9,$ED,$1E,$3F,$85,$F2,$7C,$2C,$9F,$14,$79,$FE,$16,$47,$95,$F5
	.dc.b	$7C,$2F,$AF,$DC,$93,$E9,$7D,$DE,$EF,$BF,$E1,$3E,$07,$C2,$CA,$F7
	.dc.b	$47,$C3,$F8,$4F,$89,$F0,$B2,$7E,$14,$7F,$09,$A7,$B9,$7D,$C9,$5F
	.dc.b	$E6,$19,$CB,$F2,$4F,$9C,$43,$E6,$7C,$2F,$D1,$F8,$4F,$A5,$F0,$92
	.dc.b	$7E,$84,$77,$F6,$A4,$75,$7C,$5E,$DF,$8F,$E0,$A4,$F8,$5F,$37,$C1
	.dc.b	$F9,$FE,$0A,$4F,$95,$F5,$7C,$1F,$AF,$DA,$F6,$7B,$72,$BD,$51,$EF
	.dc.b	$F8,$3F,$81,$F0,$72,$7D,$D1,$F0,$FE,$0A,$47,$C1,$7E,$37,$C1,$FC
	.dc.b	$7F,$6A,$4F,$C5,$7E,$57,$B5,$F2,$FE,$0E,$4F,$C9,$7E,$77,$C1,$FC
	.dc.b	$FF,$82,$93,$F3,$5F,$A5,$F0,$7F,$CE,$FE,$71,$ED,$D2,$57,$FA,$4E
	.dc.b	$3F,$D5,$8D,$F8,$4E,$1F,$09,$23,$06,$FE,$0F,$EF,$7C,$1C,$8C,$1C
	.dc.b	$FD,$AF,$07,$B5,$23,$07,$7F,$83,$F2,$7C,$1C,$8C,$1E,$3F,$83,$F4
	.dc.b	$7C,$1C,$8C,$1E,$7F,$67,$D9,$ED,$48,$C1,$EB,$F8,$2F,$80,$47,$DD
	.dc.b	$F0,$5F,$0F,$E0,$64,$7C,$17,$E3,$7C,$17,$C7,$F7,$E4,$FC,$57,$E5
	.dc.b	$7C,$07,$F6,$5F,$9F,$7B,$D4,$95,$FE,$BD,$8F,$F0,$D3,$7D,$DE,$04
	.dc.b	$75,$EE,$FD,$5F,$83,$91,$F4,$DF,$B1,$F0,$7D,$23,$FD,$DB,$7E,$07
	.dc.b	$B6,$1E,$3F,$80,$E9,$1A,$C6,$FC,$3B,$1F,$F7,$66,$0C,$BD,$5F,$09
	.dc.b	$30,$F4,$3E,$DF,$84,$F0,$61,$C6,$4A,$FC,$51,$E4,$CF,$C4,$FB,$C8
	.dc.b	$78,$CF,$79,$F1,$7E,$0B,$E5,$7C,$1C,$8C,$1F,$27,$E0,$BE,$71,$1F
	.dc.b	$99,$F0,$7F,$47,$E0,$E4,$7C,$F7,$EA,$7C,$1F,$D5,$F7,$FE,$B7,$C0
	.dc.b	$4A,$FA,$91,$F6,$7E,$0F,$ED,$7C,$1C,$9F,$B1,$1F,$77,$E0,$E4,$7D
	.dc.b	$B7,$F0,$1F,$03,$F8,$1F,$7E,$4F,$DF,$7F,$09,$F0,$1D,$63,$F0,$58
	.dc.b	$75,$7E,$CE,$1D,$1F,$B7,$85,$E3,$F9,$C7,$CD,$C7,$F1,$D3,$0F,$C6
	.dc.b	$1F,$91,$F8,$1F,$F1,$B7,$D4,$BD,$EB,$EF,$4A,$1F,$63,$AE,$E2,$BE
	.dc.b	$13,$89,$CC,$7F,$34,$8E,$07,$E0,$1C,$4F,$99,$B3,$C3,$E0,$FC,$7F
	.dc.b	$03,$E4,$F8,$19,$87,$53,$94,$7A,$32,$F3,$47,$6C,$7C,$99,$66,$38
	.dc.b	$47,$9F,$E1,$3D,$E4,$8C,$E6,$CB,$E1,$FC,$27,$C4,$F8,$49,$5A,$8F
	.dc.b	$8F,$F0,$7F,$23,$E1,$24,$F2,$8F,$97,$F0,$72,$3A,$BF,$3B,$E1,$3E
	.dc.b	$7F,$C2,$49,$F0,$BF,$4B,$E1,$3A,$C7,$93,$0E,$AF,$B7,$4F,$F3,$5B
	.dc.b	$D0,$86,$70,$EC,$FB,$3E,$03,$EE,$FC,$04,$8C,$1E,$73,$B1,$0E,$D9
	.dc.b	$FA,$5E,$87,$6E,$0F,$B1,$A7,$E2,$3E,$0B,$F1,$3F,$05,$2B,$E5,$C7
	.dc.b	$FA,$07,$D4,$7B,$DF,$8E,$F7,$A5,$7D,$49,$CB,$F4,$BE,$13,$F2,$7F
	.dc.b	$09,$23,$06,$FE,$0B,$A7,$C1,$48,$C1,$CF,$1E,$AF,$C7,$C7,$C5,$1F
	.dc.b	$9E,$C3,$A4,$FC,$BC,$30,$F8,$FE,$13,$D1,$F0,$D2,$30,$79,$FE,$0B
	.dc.b	$D8,$47,$D5,$F0,$5E,$FF,$82,$91,$ED,$7E,$17,$C1,$7C,$3F,$84,$93
	.dc.b	$F0,$5F,$8D,$F0,$9C,$E3,$F3,$5C,$2C,$71,$CB,$FB,$C8,$65,$DA,$B7
	.dc.b	$95,$AD,$65,$F4,$7E,$07,$E9,$7C,$0C,$BF,$A1,$1F,$57,$E0,$7E,$B7
	.dc.b	$C0,$C9,$FA,$91,$F6,$7E,$13,$ED,$7C,$24,$9F,$B1,$1F,$77,$59,$1F
	.dc.b	$6D,$FC,$07,$C2,$75,$8F,$BD,$87,$57,$DB,$87,$57,$F0,$58,$75,$7E
	.dc.b	$0E,$1F,$89,$F3,$64,$7E,$1D,$FC,$67,$C2,$75,$8F,$8F,$87,$5D,$BE
	.dc.b	$36,$3D,$67,$E5,$E3,$D6,$7E,$56,$3F,$ED,$5F,$AA,$7C,$25,$F8,$49
	.dc.b	$87,$F0,$89,$CB,$7E,$13,$87,$C2,$48,$C1,$BF,$83,$E8,$47,$97,$C2
	.dc.b	$F7,$F8,$09,$1D,$5F,$17,$C0,$78,$F0,$E9,$3F,$C6,$F1,$63,$98,$EC
	.dc.b	$43,$58,$FA,$F1,$F6,$7C,$17,$B7,$E0,$A5,$7A,$E3,$E0,$7C,$0C,$8F
	.dc.b	$73,$F0,$FE,$0B,$E2,$7C,$24,$9F,$84,$FC,$7F,$84,$F9,$1F,$01,$F2
	.dc.b	$7E,$02,$57,$C7,$8E,$8F,$A7,$0B,$CB,$FD,$53,$76,$5D,$6B,$DD,$95
	.dc.b	$AD,$65,$F5,$7E,$06,$57,$D3,$7E,$C7,$C1,$7F,$54,$8F,$E8,$1F,$57
	.dc.b	$FD,$29,$FF,$06,$BF,$CE,$DF,$E8,$2D,$77,$EF,$7E,$05,$87,$DF,$3F
	.dc.b	$09,$F0,$7F,$85,$F8,$5F,$C3,$7C,$2C,$9F,$C2,$47,$87,$3F,$ED,$F8
	.dc.b	$78,$53,$DD,$CF,$C0,$EF,$C2,$FE,$D3,$3B,$E0,$F5,$61,$F7,$CF,$19
	.dc.b	$F8,$1C,$B8,$3E,$CF,$83,$B8,$F6,$6D,$FF,$72,$9F,$87,$F0,$78,$D7
	.dc.b	$39,$C9,$F8,$71,$DB,$0F,$A9,$17,$3F,$EF,$B9,$B3,$EB,$F0,$9D,$BE
	.dc.b	$16,$5F,$CC,$8F,$0F,$C1,$C8,$F9,$EF,$93,$E0,$FC,$BF,$09,$E6,$F8
	.dc.b	$59,$5F,$52,$3D,$3F,$07,$EA,$F8,$09,$3F,$62,$3A,$3F,$7B,$0B,$B7
	.dc.b	$D0,$C7,$AC,$FD,$EC,$7A,$BF,$84,$C7,$E2,$FC,$0F,$C6,$F8,$59,$3F
	.dc.b	$84,$8F,$93,$F0,$3F,$2B,$E0,$A4,$FE,$22,$3E,$6F,$C2,$7C,$EF,$84
	.dc.b	$93,$F8,$C8,$FA,$3F,$01,$23,$F1,$EF,$D4,$F8,$0E,$B1,$F9,$2C,$3A
	.dc.b	$BF,$98,$C3,$AB,$F9,$6C,$3A,$BF,$9C,$C3,$EF,$FD,$67,$F6,$68,$FC
	.dc.b	$FF,$C1,$FE,$00,$CF,$C1,$EB,$DD,$93,$FE,$20,$A4,$72,$F7,$79,$FC
	.dc.b	$1C,$AD,$3D,$BE,$0F,$BF,$C2,$F8,$3E,$16,$57,$38,$EA,$FF,$8B,$F1
	.dc.b	$EA,$F7,$C7,$B3,$D0,$86,$F0,$EC,$F9,$7E,$0F,$DB,$F0,$72,$30,$73
	.dc.b	$21,$DE,$BA,$BC,$CE,$BD,$3F,$E6,$0C,$3F,$DA,$B1,$F0,$73,$F9,$3F
	.dc.b	$0B,$23,$E2,$3F,$33,$E1,$7E,$6F,$C1,$FC,$EF,$83,$95,$F2,$63,$E8
	.dc.b	$FC,$24,$8F,$98,$FD,$4F,$85,$FA,$BF,$03,$F5,$BE,$0A,$57,$D1,$8E
	.dc.b	$AF,$D3,$C3,$AB,$F2,$70,$E8,$FC,$BC,$2E,$BE,$BC,$7F,$03,$F0,$72
	.dc.b	$BE,$E3,$F8,$4F,$84,$FE,$07,$D5,$FD,$5A,$7F,$CC,$4D,$BF,$F4,$1B
	.dc.b	$7C,$BF,$F3,$CB,$FE,$B9,$9F,$C3,$47,$E3,$BE,$16,$46,$0F,$F3,$4C
	.dc.b	$FF,$AD,$AB,$9F,$BF,$E0,$3C,$3E,$FF,$8B,$E1,$65,$F0,$7C,$BF,$0B
	.dc.b	$F9,$AC,$BA,$CF,$E2,$72,$C3,$DB,$E0,$FD,$64,$7C,$3F,$07,$EE,$F7
	.dc.b	$64,$79,$1F,$83,$EE,$FC,$2F,$82,$93,$E8,$7E,$2F,$C1,$FC,$6F,$85
	.dc.b	$93,$EC,$7E,$4F,$C2,$F4,$8F,$93,$85,$E1,$F9,$38,$F4,$9F,$9B,$8D
	.dc.b	$9F,$9B,$8F,$D2,$F8,$3F,$A7,$F0,$72,$FE,$4C,$73,$DF,$E8,$E5,$67
	.dc.b	$FD,$C7,$C8,$CF,$A5,$7D,$6C,$ED,$6B,$2F,$BF,$F0,$9F,$80,$F8,$59
	.dc.b	$3F,$5A,$3F,$07,$F0,$5F,$84,$F8,$39,$3F,$6A,$3F,$0F,$F0,$BF,$88
	.dc.b	$F8,$59,$3F,$7A,$3F,$16,$C3,$F0,$27,$E3,$BD,$FE,$AF,$E1,$2F,$AB
	.dc.b	$F9,$0B,$EA,$FE,$22,$FA,$BF,$94,$BF,$EB,$3F,$5F,$FD,$97,$5F,$F6
	.dc.b	$5B,$65,$C6,$8C,$37,$DA,$E3,$F0,$12,$BF,$D7,$B4,$8E,$BF,$0D,$DB
	.dc.b	$E1,$7B,$FC,$24,$BD,$C7,$8B,$09,$1C,$9F,$2F,$C1,$F5,$8F,$F6,$66
	.dc.b	$3D,$5E,$F8,$F6,$79,$9E,$2C,$3B,$3C,$30,$FE,$EF,$F5,$CE,$3A,$49
	.dc.b	$FF,$6A,$C3,$8F,$7D,$3C,$3A,$A5,$6A,$39,$F5,$74,$E9,$93,$CA,$3E
	.dc.b	$6F,$BD,$F3,$BE,$02,$4F,$C6,$8F,$A3,$F0,$7F,$4B,$E1,$24,$FC,$A8
	.dc.b	$FA,$BF,$09,$23,$E6,$BF,$63,$E1,$3A,$C7,$D0,$C3,$AB,$F1,$B0,$EA
	.dc.b	$FD,$4C,$3A,$BF,$2B,$0E,$8F,$CC,$C2,$F3,$BA,$3F,$07,$97,$6A,$E8
	.dc.b	$7A,$4F,$BF,$F0,$9D,$9F,$AB,$8F,$6B,$E5,$1F,$8A,$E6,$65,$DB,$1F
	.dc.b	$C7,$E5,$6B,$59,$7D,$6F,$2B,$EB,$F9,$52,$7E,$AC,$7D,$AF,$2A,$47
	.dc.b	$D8,$7E,$EF,$95,$FE,$34,$FA,$FF,$0C,$FE,$6B,$23,$F9,$C8,$CE,$BE
	.dc.b	$60,$E2,$77,$18,$F2,$FA,$79,$BD,$7C,$D9,$86,$AF,$78,$2B,$C5,$E9
	.dc.b	$78,$FD,$39,$7B,$3C,$51,$E7,$F4,$A4,$67,$3F,$1C,$7A,$FD,$3F,$67
	.dc.b	$A7,$ED,$F4,$E5,$EF,$34,$FF,$95,$A1,$F5,$EF,$5B,$1F,$3E,$F8,$6A
	.dc.b	$6E,$75,$5C,$EB,$E5,$11,$ED,$EC,$7C,$DF,$76,$47,$85,$FA,$1E,$EF
	.dc.b	$3A,$EB,$3E,$F1,$8F,$85,$C6,$1E,$07,$EB,$1F,$57,$DD,$FB,$5F,$03
	.dc.b	$23,$DD,$1F,$77,$DF,$91,$F0,$5F,$C0,$7B,$DE,$18,$F0,$61,$DD,$EA
	.dc.b	$78,$B5,$D3,$E0,$E3,$FC,$0F,$C0,$7E,$27,$E0,$64,$FC,$D8,$FC,$67
	.dc.b	$C1,$7F,$28,$F8,$39,$3F,$46,$39,$DF,$E4,$31,$F0,$37,$1C,$F5,$FF
	.dc.b	$2E,$F8,$7F,$E6,$31,$30,$FE,$53,$1F,$CD,$3E,$1B,$F9,$AF,$C3,$C9
	.dc.b	$FE,$5B,$1C,$9F,$CD,$E3,$DB,$0F,$F5,$93,$63,$F3,$FC,$5F,$37,$2F
	.dc.b	$25,$FE,$08,$93,$F4,$B8,$FF,$A1,$FC,$3F,$F4,$4F,$82,$93,$FC,$F2
	.dc.b	$3C,$0F,$C9,$3E,$6E,$B8,$C7,$B6,$7E,$2C,$7A,$D7,$D0,$9C,$BE,$4C
	.dc.b	$7F,$21,$D7,$23,$C4,$43,$EB,$E9,$EB,$7C,$07,$39,$F1,$C2,$FB,$E9
	.dc.b	$FC,$03,$0E,$F7,$FA,$2B,$EC,$F8,$F0,$F4,$47,$E5,$A3,$BE,$58,$7F
	.dc.b	$13,$87,$4C,$BF,$DB,$D8,$74,$26,$1D,$31,$F0,$63,$E4,$C7,$A4,$63
	.dc.b	$1F,$63,$F4,$9E,$A6,$30,$EB,$5C,$DE,$F7,$E3,$9F,$29,$F4,$0E,$07
	.dc.b	$CB,$39,$9F,$23,$1E,$17,$F0,$E7,$A4,$EC,$F9,$47,$33,$A9,$82,$47
	.dc.b	$9D,$E8,$78,$0F,$98,$7E,$EA,$3A,$B8,$38,$D7,$E5,$30,$F8,$6F,$BA
	.dc.b	$39,$65,$F6,$5F,$4D,$FC,$77,$89,$F9,$AB,$EB,$97,$D1,$CB,$9C,$FE
	.dc.b	$8B,$1F,$7E,$3F,$91,$8E,$6F,$77,$99,$FC,$92,$F9,$D7,$D7,$8F,$85
	.dc.b	$87,$84,$95,$CE,$3E,$09,$CB,$0F,$C0,$BF,$04,$F8,$8F,$E4,$31,$FC
	.dc.b	$C1,$F9,$97,$D1,$1E,$43,$F7,$39,$7A,$1F,$71,$CC,$ED,$87,$B9,$FD
	.dc.b	$0E,$1E,$47,$ED,$3F,$54,$98,$7B,$DD,$47,$BC,$E8,$E2,$FC,$D9,$75
	.dc.b	$7C,$B3,$EE,$9E,$6F,$6A,$E8,$7E,$51,$E6,$E1,$D1,$F4,$CF,$96,$E5
	.dc.b	$FC,$24,$74,$7F,$A1,$E1,$F0,$5F,$31,$FB,$53,$E5,$E1,$F0,$71,$F9
	.dc.b	$4F,$F0,$09,$FD,$A4,$F8,$0E,$B7,$E5,$79,$1F,$28,$F1,$9E,$53,$07
	.dc.b	$BD,$FB,$D7,$F1,$6B,$F1,$F9,$79,$70,$FB,$79,$5C,$7C,$38,$F8,$0F
	.dc.b	$16,$4B,$FC,$6C,$7E,$7A,$BE,$76,$1E,$98,$FC,$11,$C4,$95,$C5,$F9
	.dc.b	$A7,$48,$F8,$53,$E1,$8E,$33,$F6,$31,$FD,$F4,$70,$39,$D7,$07,$E5
	.dc.b	$9E,$0C,$7F,$60,$FC,$29,$E2,$FC,$87,$E1,$3F,$20,$FC,$8E,$5F,$6E
	.dc.b	$FF,$3D,$97,$C2,$7C,$51,$F4,$6B,$99,$CA,$BF,$86,$BF,$D8,$DF,$19
	.dc.b	$F4,$4F,$D1,$9F,$C7,$1F,$F4,$86,$7F,$D4,$A3,$FD,$03,$A7,$F5,$67
	.dc.b	$E6,$1E,$5B,$F0,$BC,$8E,$87,$D5,$AF,$F4,$D6,$1F,$DA,$E3,$FD,$17
	.dc.b	$AF,$F6,$F7,$99,$E6,$C7,$F3,$19,$7C,$6B,$FC,$C3,$63,$F4,$11,$FE
	.dc.b	$48,$DF,$FC,$07,$5F,$E9,$6E,$0F,$F0,$3B,$D2,$7E,$03,$FC,$71,$FF
	.dc.b	$56,$70,$7F,$80,$63,$FD,$3B,$BF,$FA,$D2,$3F,$C3,$B1,$FE,$15,$EB
	.dc.b	$72,$E2,$FF,$CB,$10,$E5,$FE,$35,$5E,$93,$F6,$73,$FE,$1C,$7F,$C8
	.dc.b	$DC,$BF,$E1,$C7,$A7,$2F,$F8,$E9,$ED,$D3,$6B,$AF,$4F,$47,$D2,$7F
	.dc.b	$11,$3F,$CC,$FD,$1F,$E3,$5A,$F3,$74,$FF,$9C,$63,$FC,$AF,$1F,$E5
	.dc.b	$5E,$9F,$F2,$BB,$EB,$EB,$E9,$C7,$FD,$16,$3F,$01,$B7,$F8,$1B,$1F
	.dc.b	$F3,$9F,$57,$F9,$46,$7E,$17,$5F,$F9,$C9,$FF,$30,$78,$3F,$E7,$27
	.dc.b	$E3,$78,$3F,$E7,$07,$FC,$C9,$E0,$FF,$9C,$1F,$95,$E0,$FF,$9B,$9F
	.dc.b	$99,$E0,$DC,$BE,$5F,$83,$FE,$91,$9F,$F6,$1F,$8D,$FE,$91,$7F,$C8
	.dc.b	$DE,$A3,$C7,$FF,$48,$C6,$3C,$7B,$87,$FA,$2E,$9E,$3F,$FA,$4F,$8A
	.dc.b	$5F,$F9,$DD,$FF,$4B,$78,$FF,$ED,$4F,$81,$FF,$6A,$FC,$14,$AF,$F6
	.dc.b	$9C,$7F,$B6,$3E,$13,$FD,$B3,$F0,$92,$7F,$DA,$F1,$FE,$DC,$F8,$6F
	.dc.b	$F6,$EF,$C3,$49,$FF,$6D,$C7,$37,$D3,$85,$D7,$CD,$8F,$59,$FF,$70
	.dc.b	$E3,$D5,$F7,$E3,$D5,$FF,$72,$E3,$D5,$F8,$78,$FF,$BB,$3D,$AF,$F7
	.dc.b	$54,$7C,$6F,$6F,$0F,$FB,$9C,$FF,$76,$FB,$7F,$EE,$A7,$E5,$FB,$7F
	.dc.b	$EE,$A7,$FD,$E3,$ED,$FF,$BA,$9F,$9F,$ED,$FF,$BE,$3E,$0F,$FD,$73
	.dc.b	$1F,$E7,$7C,$5F,$13,$8E,$0F,$F5,$D4,$62,$7F,$DF,$75,$FF,$00,$3E
	.dc.b	$1F,$FE,$00,$FC,$34,$9F,$F7,$FC,$71,$E5,$FF,$5F,$F3,$BE,$1F,$14
	.dc.b	$C9,$39,$8F,$F5,$B3,$5F,$C2,$FC,$3F,$09,$67,$E3,$9F,$C4,$99,$AF
	.dc.b	$5B,$EB,$D3,$CC,$F4,$CF,$CC,$FE,$43,$3E,$19,$48,$F0,$BC,$F2,$F0
	.dc.b	$62,$F8,$2A,$4A,$FF,$66,$43,$E0,$F7,$EE,$70,$F8,$39,$5A,$8E,$7F
	.dc.b	$0D,$23,$8B,$DB,$E1,$BB,$FC,$37,$83,$E1,$E5,$76,$8E,$AF,$83,$0E
	.dc.b	$AF,$D9,$C3,$A3,$F6,$F0,$BC,$5F,$6B,$1F,$5F,$C2,$FB,$3E,$0E,$5F
	.dc.b	$AA,$3A,$BF,$83,$C7,$AD,$FE,$0B,$1E,$7C,$BF,$EF,$AD,$99,$E3,$0F
	.dc.b	$B9,$0C,$FE,$4F,$C3,$49,$F8,$EF,$CC,$F8,$5F,$9B,$F0,$3F,$3B,$E1
	.dc.b	$65,$7C,$C8,$FA,$3F,$0B,$F4,$BE,$1E,$4F,$D0,$8F,$AB,$F0,$3F,$5B
	.dc.b	$E1,$64,$FD,$48,$EA,$FD,$6C,$3A,$BF,$9F,$C3,$AB,$F6,$B0,$EA,$FE
	.dc.b	$8F,$0F,$C0,$FC,$1C,$8F,$BE,$FE,$13,$E0,$FF,$29,$F6,$1F,$D1,$E4
	.dc.b	$7F,$B8,$5C,$37,$E0,$F7,$F0,$5C,$3E,$0F,$8F,$C2,$CA,$E3,$CE,$E6
	.dc.b	$B8,$AF,$BF,$C1,$4B,$DD,$6A,$3C,$7F,$0B,$E4,$CE,$56,$8F,$D8,$61
	.dc.b	$E4,$C7,$B4,$63,$3F,$23,$B8,$4B,$3F,$34,$FB,$73,$F8,$3F,$09,$F0
	.dc.b	$BE,$12,$4F,$1B,$4F,$C6,$F6,$A4,$E5,$F9,$3F,$07,$F2,$BE,$13,$E5
	.dc.b	$FC,$1C,$AE,$51,$F3,$BE,$0E,$47,$57,$E8,$FC,$2F,$48,$FA,$18,$5D
	.dc.b	$BD,$A7,$AB,$1E,$D3,$E5,$C7,$B3,$F5,$F1,$FB,$7F,$05,$F7,$3E,$06
	.dc.b	$4F,$AE,$3A,$72,$FD,$4C,$AC,$FC,$CC,$7A,$BF,$81,$C7,$AE,$5E,$2D
	.dc.b	$BF,$CA,$53,$F8,$5F,$84,$FC,$4F,$C1,$C8,$F9,$11,$FC,$7F,$EC,$7F
	.dc.b	$AF,$C7,$CB,$CF,$F1,$E6,$7E,$07,$F2,$3F,$03,$F9,$2F,$85,$95,$F4
	.dc.b	$63,$97,$C2,$C8,$FB,$98,$23,$B7,$C2,$73,$9E,$C4,$3F,$9B,$63,$F9
	.dc.b	$4A,$F1,$C7,$E7,$6F,$BB,$F9,$FC,$3C,$6F,$6F,$80,$F5,$11,$F0,$FC
	.dc.b	$2F,$B7,$E1,$24,$79,$1F,$81,$F0,$7F,$07,$E1,$24,$FA,$1F,$89,$F0
	.dc.b	$9F,$17,$E0,$64,$FB,$1F,$91,$F0,$BD,$23,$D2,$7F,$4F,$D5,$8E,$79
	.dc.b	$7F,$0B,$0C,$BB,$D7,$09,$4B,$2C,$C7,$0C,$BE,$B7,$C2,$7D,$7F,$83
	.dc.b	$98,$7C,$E8,$FB,$5E,$F7,$DB,$F7,$A4,$FD,$28,$FB,$DF,$07,$23,$EA
	.dc.b	$BF,$81,$F5,$BF,$05,$F0,$BF,$83,$F8,$59,$5F,$6A,$3A,$BF,$73,$0E
	.dc.b	$AF,$A7,$6F,$ED,$5A,$FB,$B1,$ED,$39,$7B,$3F,$84,$F8,$2F,$C8,$11
	.dc.b	$FC,$3F,$C1,$F6,$7A,$61,$DA,$FE,$4E,$9D,$1E,$D3,$EA,$3F,$1F,$A7
	.dc.b	$F7,$9F,$B2,$FE,$F1,$3F,$B4,$6F,$57,$F2,$06,$D7,$D7,$38,$7C,$0C
	.dc.b	$7F,$8A,$42,$3A,$7C,$17,$5F,$84,$95,$A7,$C1,$F0,$9E,$1F,$80,$93
	.dc.b	$C9,$F2,$7C,$07,$58,$FC,$06,$3D,$5E,$F8,$F6,$7A,$10,$DE,$1D,$9F
	.dc.b	$2F,$C0,$FB,$FE,$0A,$46,$0F,$4F,$C1,$7C,$3F,$82,$91,$83,$DB,$F0
	.dc.b	$7F,$1F,$E1,$24,$60,$F8,$3E,$FF,$CB,$F8,$09,$18,$39,$1E,$AB,$EA
	.dc.b	$FB,$30,$FA,$58,$74,$8F,$7E,$18,$7E,$6F,$C0,$FD,$7F,$82,$91,$83
	.dc.b	$FA,$27,$A9,$FD,$4D,$BE,$E7,$F8,$D6,$3F,$A3,$46,$BD,$CF,$C0,$7C
	.dc.b	$14,$8F,$B2,$FE,$0F,$E0,$BF,$09,$F0,$72,$7E,$EB,$F8,$7F,$84,$E9
	.dc.b	$7D,$71,$95,$D4,$E6,$43,$EC,$E1,$D5,$F3,$FC,$17,$97,$E0,$A4,$60
	.dc.b	$E4,$7E,$FF,$3F,$4B,$6F,$C2,$F5,$EE,$FC,$C6,$1E,$B8,$FC,$3E,$9F
	.dc.b	$07,$E1,$24,$7C,$67,$E2,$7C,$27,$C5,$F8,$29,$3F,$29,$F9,$1F,$05
	.dc.b	$F2,$7E,$0E,$4F,$CE,$7E,$67,$C2,$7C,$DF,$84,$F9,$DF,$03,$2B,$E9
	.dc.b	$C7,$47,$D5,$85,$D7,$FB,$0E,$EC,$BA,$D7,$BF,$2B,$5A,$CB,$ED,$7C
	.dc.b	$1F,$DB,$F8,$39,$7F,$80,$8F,$BD,$F0,$32,$3F,$06,$FE,$07,$E0,$BF
	.dc.b	$05,$F0,$9F,$83,$D2,$57,$E2,$23,$F0,$DF,$01,$23,$F1,$6F,$E2,$7E
	.dc.b	$03,$A4,$7D,$3C,$2E,$9E,$AC,$79,$CF,$D7,$C6,$CF,$BB,$1F,$F4,$37
	.dc.b	$D9,$FF,$56,$8F,$83,$E7,$7E,$54,$CF,$C2,$6B,$DC,$98,$FF,$87,$EF
	.dc.b	$F9,$F3,$7D,$CE,$7F,$05,$D3,$E0,$A4,$EE,$3B,$FC,$27,$83,$E1,$24
	.dc.b	$F3,$8E,$AF,$F4,$BC,$3A,$B7,$1F,$3E,$3D,$A3,$C5,$8F,$67,$86,$3E
	.dc.b	$CF,$82,$F6,$FC,$14,$9F,$44,$76,$DB,$FC,$3F,$97,$66,$E5,$DA,$BD
	.dc.b	$07,$62,$1D,$8F,$85,$F0,$9F,$28,$8F,$C5,$F8,$4F,$9B,$F0,$32,$3E
	.dc.b	$43,$F4,$3E,$0B,$E8,$FC,$1F,$D2,$F8,$49,$5F,$36,$3E,$AF,$C1,$FD
	.dc.b	$6F,$7E,$4F,$D1,$8E,$8F,$C7,$C2,$E9,$EC,$C7,$AC,$FD,$7C,$7A,$BF
	.dc.b	$37,$1F,$C0,$FC,$07,$E0,$BE,$06,$4F,$DD,$8F,$F6,$57,$DA,$7F,$4E
	.dc.b	$C5,$C1,$FE,$C8,$7E,$4D,$6B,$E1,$FF,$17,$F0,$B2,$3F,$0C,$FE,$3B
	.dc.b	$E1,$79,$FB,$F2,$7F,$CD,$18,$FF,$B1,$DB,$F0,$9C,$DF,$C7,$1F,$96
	.dc.b	$30,$7E,$16,$FB,$34,$FC,$A5,$FA,$3E,$06,$47,$67,$D7,$F0,$BE,$CF
	.dc.b	$73,$DB,$EE,$4A,$F1,$C7,$C0,$F8,$29,$1E,$67,$E1,$FC,$17,$C4,$F8
	.dc.b	$5F,$8B,$F0,$B2,$BD,$71,$D5,$F6,$E1,$D5,$FF,$13,$6C,$C7,$AC,$63
	.dc.b	$1B,$C3,$FE,$72,$A6,$5F,$4B,$E0,$7E,$9F,$C1,$4C,$3E,$4C,$77,$7A
	.dc.b	$E5,$DF,$0E,$59,$74,$F2,$3F,$D4,$F9,$33,$C6,$0B,$4F,$82,$3E,$AF
	.dc.b	$C2,$7E,$08,$8F,$D8,$F8,$4F,$C2,$FC,$14,$8F,$B6,$FE,$23,$E0,$BF
	.dc.b	$13,$F0,$92,$7E,$FB,$F8,$CF,$84,$FC,$6F,$C0,$7E,$3B,$E0,$25,$7E
	.dc.b	$12,$3A,$3F,$23,$0B,$A7,$C5,$C7,$A4,$FC,$CC,$6C,$FC,$9C,$7F,$D9
	.dc.b	$B3,$FE,$55,$FB,$5F,$F6,$26,$FF,$EB,$76,$C6,$7E,$0B,$FD,$43,$3F
	.dc.b	$EC,$C9,$CD,$FF,$A9,$17,$C1,$6A,$FE,$44,$3E,$0B,$14,$8D,$DF,$79
	.dc.b	$CC,$3E,$12,$98,$F8,$4B,$5E,$5F,$80,$D9,$C3,$E0,$2C,$7A,$73,$E7
	.dc.b	$8F,$46,$C7,$B7,$0E,$79,$FC,$1C,$6E,$7E,$BF,$82,$F1,$9E,$4F,$82
	.dc.b	$B8,$7C,$5F,$73,$CF,$EE,$7A,$1B,$1F,$27,$E0,$7D,$7F,$05,$EC,$6C
	.dc.b	$7C,$DF,$83,$F7,$9F,$03,$E1,$2C,$75,$3E,$96,$1D,$5A,$75,$3E,$B5
	.dc.b	$F5,$69,$F6,$7E,$0B,$E5,$9F,$33,$E0,$AC,$75,$DF,$F0,$17,$D5,$B1
	.dc.b	$D5,$FC,$25,$F5,$6C,$7E,$17,$E1,$3E,$B9,$F6,$3E,$12,$C7,$FB,$FF
	.dc.b	$A7,$FD,$88,$DF,$82,$FC,$0C,$3E,$0E,$C7,$E0,$26,$1C,$7E,$12,$E3
	.dc.b	$FE,$A9,$B7,$C1,$72,$F8,$3E,$6D,$BE,$C4,$AE,$18,$F6,$8E,$87,$E7
	.dc.b	$4F,$1E,$1D,$B9,$BF,$DF,$B8,$FF,$AD,$67,$D5,$F0,$9E,$8F,$82,$F4
	.dc.b	$BE,$AF,$82,$F5,$FC,$15,$AF,$80,$FB,$BE,$0E,$47,$C3,$F8,$4B,$AF
	.dc.b	$AF,$DF,$F8,$67,$C4,$F8,$0B,$5F,$27,$1E,$AD,$3A,$9F,$37,$0E,$7A
	.dc.b	$FA,$31,$B2,$C7,$F8,$C6,$AF,$82,$FA,$27,$D2,$F8,$2B,$5F,$5F,$DC
	.dc.b	$FA,$BE,$E7,$D6,$6C,$7D,$BF,$81,$FB,$27,$DA,$F8,$2B,$1F,$7F,$E0
	.dc.b	$FE,$EF,$C2,$7D,$E6,$C7,$53,$F0,$98,$75,$69,$D4,$9A,$7D,$FB,$EA
	.dc.b	$FF,$BA,$7E,$D9,$F0,$9F,$C2,$9B,$5F,$8F,$9C,$4F,$57,$F1,$17,$78
	.dc.b	$3E,$EB,$D1,$FB,$F9,$FA,$EF,$C3,$07,$E5,$2C,$1F,$09,$E1,$F8,$4F
	.dc.b	$13,$6B,$67,$09,$7C,$0F,$9F,$E0,$BD,$0D,$9D,$E7,$FD,$36,$BF,$DA
	.dc.b	$51,$8C,$7C,$1F,$07,$EF,$F8,$4F,$80,$FC,$1F,$7E,$56,$EF,$FB,$1B
	.dc.b	$63,$BB,$9C,$AE,$3F,$2D,$E9,$1C,$6E,$CF,$CF,$7E,$77,$C0,$49,$ED
	.dc.b	$F0,$36,$BE,$A7,$C0,$7D,$33,$EA,$7C,$05,$AF,$27,$C1,$7D,$7F,$83
	.dc.b	$FB,$0D,$8F,$47,$BD,$F6,$FB,$7E,$E3,$63,$D9,$87,$3C,$7F,$05,$8D
	.dc.b	$E1,$FA,$D1,$F2,$71,$E8,$FE,$1D,$B1,$A3,$ED,$7C,$07,$E2,$4F,$C5
	.dc.b	$7C,$05,$C7,$E8,$7D,$BF,$F5,$16,$FB,$5F,$D6,$23,$18,$FF,$80,$E7
	.dc.b	$F0,$B0,$9D,$7C,$3E,$FE,$0A,$D7,$47,$F3,$1F,$07,$F9,$9F,$84,$B3
	.dc.b	$F6,$5E,$DE,$FF,$7C,$2C,$F8,$8F,$2E,$1D,$5A,$75,$3D,$3F,$01,$E7
	.dc.b	$F8,$1F,$43,$63,$DB,$7D,$5A,$75,$D3,$D8,$7F,$87,$F4,$FE,$FB,$1E
	.dc.b	$ED,$3D,$73,$AD,$D4,$9F,$8D,$F0,$7F,$1F,$E0,$EE,$1F,$31,$F9,$5E
	.dc.b	$FF,$CB,$F7,$EC,$FD,$07,$E7,$7C,$1F,$CF,$F8,$3B,$3F,$51,$FA,$5F
	.dc.b	$05,$F4,$FD,$FB,$3F,$61,$E7,$B7,$97,$56,$38,$DB,$F0,$51,$8B,$FB
	.dc.b	$F9,$77,$69,$DC,$FC,$1F,$C1,$FE,$04,$FC,$17,$C0,$D8,$FC,$3F,$C0
	.dc.b	$FE,$17,$E0,$7F,$0C,$D8,$FC,$5F,$C1,$FE,$24,$FC,$57,$C1,$D8,$FD
	.dc.b	$F7,$DC,$3E,$03,$FC,$4A,$D8,$FC,$99,$88,$EA,$FE,$5B,$0B,$BF,$BD
	.dc.b	$E8,$FC,$1C,$3B,$CF,$88,$EF,$F0,$9D,$BE,$0B,$BB,$63,$C7,$EE,$78
	.dc.b	$BD,$CF,$1B,$63,$CF,$F0,$5E,$6F,$82,$F3,$B6,$3D,$7F,$07,$EA,$F8
	.dc.b	$4F,$5B,$63,$A9,$F0,$30,$EA,$D3,$A9,$F1,$2F,$AB,$4F,$8F,$F0,$3F
	.dc.b	$1B,$E1,$3E,$3B,$63,$AE,$DF,$95,$E1,$61,$DA,$30,$F4,$75,$85,$E8
	.dc.b	$FE,$0D,$48,$FA,$7F,$09,$F5,$3E,$12,$E3,$F6,$1F,$AF,$F0,$12,$3E
	.dc.b	$DF,$C0,$DC,$BC,$5F,$09,$F7,$0F,$BB,$F0,$96,$BF,$03,$F0,$1F,$80
	.dc.b	$F8,$3F,$C0,$B6,$3F,$0B,$8F,$3D,$BC,$B9,$59,$FE,$AB,$B7,$F7,$D7
	.dc.b	$F1,$79,$76,$69,$F8,$EF,$84,$FC,$69,$F8,$EF,$83,$B1,$FD,$3F,$1F
	.dc.b	$EB,$AD,$F8,$3F,$F3,$BC,$63,$6F,$E0,$0A,$33,$F0,$9F,$98,$9B,$5F
	.dc.b	$9C,$78,$FC,$14,$8E,$DE,$FD,$9E,$7F,$09,$DB,$0E,$ED,$AF,$26,$1D
	.dc.b	$5A,$75,$3D,$1F,$09,$E6,$F8,$2F,$3B,$63,$D9,$ED,$FA,$8F,$5F,$C0
	.dc.b	$58,$F8,$1F,$05,$EE,$3D,$FF,$05,$63,$E2,$7C,$1F,$C2,$3E,$1F,$C3
	.dc.b	$D8,$EA,$4D,$BE,$2E,$1D,$5F,$8F,$87,$37,$E4,$E1,$79,$FF,$AA,$71
	.dc.b	$3E,$0B,$E7,$9F,$43,$E0,$AE,$5D,$6B,$7A,$FA,$A3,$07,$F8,$5A,$FF
	.dc.b	$DD,$0F,$B0,$FB,$98,$77,$75,$F0,$7F,$70,$FB,$BF,$09,$63,$F0,$5F
	.dc.b	$05,$F8,$03,$F0,$3F,$05,$63,$F0,$DF,$09,$F8,$4F,$84,$FC,$2B,$63
	.dc.b	$FC,$31,$F7,$3F,$D2,$38,$3F,$DD,$0F,$F4,$58,$CF,$C2,$7F,$B0,$A3
	.dc.b	$13,$F8,$CC,$7B,$3F,$8E,$C6,$EF,$F1,$B1,$ED,$3F,$92,$6C,$7F,$7D
	.dc.b	$BF,$EE,$AD,$F8,$5E,$DF,$07,$DD,$F0,$7C,$17,$87,$E0,$AE,$BE,$87
	.dc.b	$C9,$F0,$92,$3D,$7F,$0B,$74,$FE,$C1,$9B,$DE,$F5,$7C,$1F,$AD,$B7
	.dc.b	$F0,$B2,$EC,$D3,$B1,$F1,$B1,$E9,$AF,$A7,$2B,$3F,$E5,$0C,$5F,$09
	.dc.b	$F1,$FE,$0B,$E4,$36,$7E,$7F,$B7,$F2,$CF,$99,$EF,$58,$FA,$7F,$03
	.dc.b	$F3,$CF,$A1,$F0,$36,$3E,$BF,$C2,$7D,$33,$EA,$7C,$25,$8E,$A4,$E2
	.dc.b	$F8,$38,$75,$7E,$1E,$1C,$DF,$8B,$85,$DB,$EA,$7C,$1F,$DF,$3F,$01
	.dc.b	$F0,$57,$1E,$93,$F6,$74,$F4,$C6,$0F,$85,$8F,$37,$E2,$61,$71,$5B
	.dc.b	$EE,$BF,$B4,$47,$D6,$AF,$F2,$01,$FE,$7A,$8C,$CF,$E1,$E3,$33,$CA
	.dc.b	$7F,$D9,$B1,$FE,$71,$6C,$7F,$87,$D7,$BF,$C2,$BC,$1E,$FD,$9F,$F6
	.dc.b	$FD,$6F,$E0,$B1,$F0,$79,$75,$F0,$53,$2E,$DE,$FD,$BE,$D8,$F3,$3A
	.dc.b	$67,$6B,$C9,$87,$37,$CD,$8D,$BF,$37,$BF,$E3,$3C,$9E,$FD,$AF,$67
	.dc.b	$C0,$F9,$CF,$47,$C1,$58,$F8,$1F,$07,$EB,$3D,$9F,$07,$63,$E2,$61
	.dc.b	$CF,$1F,$89,$8D,$E1,$FF,$37,$70,$B3,$E9,$5F,$27,$3B,$5A,$CF,$A4
	.dc.b	$FC,$DC,$EC,$F2,$F8,$3F,$99,$EF,$FC,$D6,$EB,$F5,$7E,$0B,$E8,$7C
	.dc.b	$1F,$D1,$6C,$7D,$9F,$47,$EA,$1F,$57,$E0,$EC,$7D,$DC,$3E,$C7,$C1
	.dc.b	$74,$D3,$EF,$3C,$E7,$F0,$58,$D9,$FC,$13,$C9,$FC,$26,$1D,$67,$C0
	.dc.b	$7E,$26,$BF,$09,$F0,$5D,$A3,$F1,$8F,$3D,$3E,$3E,$16,$7F,$2C,$F5
	.dc.b	$8F,$97,$7D,$70,$EF,$5F,$EC,$5F,$BB,$FF,$70,$C7,$F9,$CD,$B5,$F4
	.dc.b	$A3,$F2,$D1,$89,$FF,$6D,$BF,$EC,$E8,$E1,$F0,$7A,$F8,$3D,$BC,$3E
	.dc.b	$0B,$8F,$C1,$DC,$3C,$0F,$4F,$81,$EB,$F0,$56,$7C,$8F,$27,$CD,$7D
	.dc.b	$5A,$7A,$6F,$AB,$4E,$A7,$B7,$E0,$BD,$3F,$05,$EA,$6C,$7C,$1F,$7B
	.dc.b	$DB,$F0,$5E,$E6,$C7,$C5,$F8,$2F,$82,$7C,$2F,$80,$B1,$F2,$7E,$0B
	.dc.b	$E2,$FC,$17,$C6,$6C,$75,$3E,$76,$1D,$5A,$75,$27,$0F,$D1,$BE,$AF
	.dc.b	$D3,$F8,$2F,$A2,$7D,$2F,$82,$B5,$FE,$FD,$F4,$3F,$B3,$73,$1A,$7F
	.dc.b	$64,$7F,$DB,$0E,$B9,$9F,$01,$F7,$3E,$03,$EE,$B6,$FF,$09,$F0,$1F
	.dc.b	$80,$3F,$03,$F0,$56,$3C,$1F,$05,$FC,$67,$1E,$FF,$0F,$63,$9C,$FE
	.dc.b	$37,$1B,$B7,$E3,$5E,$1F,$07,$C7,$E0,$ED,$7A,$DE,$4F,$63,$DD,$86
	.dc.b	$2F,$F2,$DA,$79,$DB,$3D,$CF,$89,$EF,$F9,$7D,$FF,$33,$63,$E4,$7C
	.dc.b	$0F,$A4,$F5,$7C,$15,$8F,$99,$F0,$5E,$D3,$DD,$F0,$76,$3E,$87,$C0
	.dc.b	$FC,$1F,$82,$F8,$4D,$8E,$A4,$CF,$FD,$BD,$C4,$C7,$B4,$63,$1E,$8E
	.dc.b	$B1,$BC,$DF,$63,$E0,$BE,$71,$F3,$FE,$0A,$E5,$F8,$1F,$80,$FA,$59
	.dc.b	$FD,$36,$C7,$E1,$7E,$03,$EB,$7C,$07,$D7,$6C,$7E,$27,$E0,$BE,$D7
	.dc.b	$C1,$7D,$B6,$C7,$53,$F1,$D8,$75,$69,$D4,$FC,$95,$F5,$68,$D7,$FD
	.dc.b	$FD,$F7,$9F,$DA,$23,$1A,$FE,$5D,$FE,$C5,$19,$9F,$EE,$76,$9C,$FC
	.dc.b	$1C,$78,$9C,$AB,$FD,$5C,$E2,$BF,$BD,$C2,$7A,$FC,$14,$AD,$60,$AF
	.dc.b	$EE,$71,$87,$FB,$BC,$78,$6F,$9F,$C0,$F9,$BE,$07,$CE,$DB,$E1,$8F
	.dc.b	$96,$BD,$FF,$07,$70,$E4,$77,$CB,$C5,$3F,$13,$E0,$EC,$EB,$37,$BD
	.dc.b	$F1,$0F,$8B,$EF,$5C,$FB,$93,$0E,$ED,$BE,$58,$F4,$8F,$8D,$8D,$AF
	.dc.b	$8B,$ED,$FD,$03,$E8,$FB,$D6,$BE,$17,$C1,$FD,$4F,$83,$FA,$AD,$8F
	.dc.b	$8D,$F0,$5F,$63,$E0,$BE,$CB,$63,$E5,$7C,$07,$E0,$72,$FC,$0B,$75
	.dc.b	$F7,$36,$7E,$D7,$C0,$5A,$FB,$3F,$05,$6F,$EB,$7C,$0D,$E3,$36,$71
	.dc.b	$F0,$32,$FE,$D3,$BF,$82,$E5,$F7,$AF,$7E,$FC,$1C,$BF,$C1,$CE,$5B
	.dc.b	$F0,$1C,$3E,$0E,$46,$0D,$FC,$37,$8D,$86,$0E,$7F,$09,$E0,$23,$DB
	.dc.b	$E1,$3C,$7F,$01,$23,$C2,$F9,$BE,$0B,$CF,$F0,$12,$7C,$AF,$AB,$E1
	.dc.b	$3D,$7F,$03,$EC,$F8,$49,$5E,$A8,$E8,$FF,$97,$3A,$3F,$5F,$BB,$2C
	.dc.b	$F3,$FE,$52,$19,$F7,$BE,$07,$F9,$92,$19,$E6,$B8,$67,$F3,$3E,$07
	.dc.b	$E6,$FC,$1C,$C3,$E5,$C7,$6C,$FC,$39,$D9,$E9,$9F,$39,$F2,$FC,$27
	.dc.b	$D6,$31,$3E,$1F,$7F,$FC,$CD,$AB,$E0,$AC,$61,$FF,$35,$7D,$F3,$D6
	.dc.b	$BE,$E6,$1B,$97,$DE,$F8,$3E,$7F,$0B,$D1,$B1,$FC,$12,$DF,$05,$D3
	.dc.b	$E1,$3A,$B6,$75,$87,$4D,$7E,$3F,$23,$0F,$15,$73,$CF,$CB,$F0,$7E
	.dc.b	$AF,$83,$B3,$F9,$07,$D3,$EF,$FA,$BD,$FB,$3E,$37,$D4,$F3,$3C,$25
	.dc.b	$CB,$C3,$8B,$E0,$FE,$1F,$C2,$7C,$46,$D7,$B7,$E0,$FE,$3F,$C2,$7C
	.dc.b	$86,$C7,$C1,$F7,$3E,$59,$F3,$3E,$02,$C7,$C5,$F7,$BE,$79,$F4,$3E
	.dc.b	$0E,$C7,$C9,$F8,$3F,$A7,$F0,$9F,$51,$B1,$D4,$9B,$7A,$30,$EA,$FA
	.dc.b	$B0,$E6,$FB,$30,$BC,$9F,$CA,$B7,$7C,$1F,$E0,$3E,$13,$F0,$2D,$CB
	.dc.b	$AD,$6F,$5F,$54,$60,$FE,$A3,$6C,$BA,$C6,$31,$B9,$7F,$B1,$2B,$FA
	.dc.b	$94,$FE,$17,$E1,$3F,$1E,$FE,$41,$FF,$5D,$76,$7F,$52,$AF,$C1,$B7
	.dc.b	$2F,$E8,$2F,$F4,$CA,$FE,$C2,$E5,$BF,$07,$C8,$E7,$EF,$58,$FE,$AF
	.dc.b	$2F,$84,$EE,$78,$3E,$26,$CE,$BE,$13,$C6,$79,$3E,$0A,$C7,$2C,$3A
	.dc.b	$67,$E6,$6E,$7E,$6F,$83,$F5,$9E,$CF,$85,$B7,$D4,$97,$EE,$C3,$AB
	.dc.b	$CE,$3C,$FA,$FA,$9A,$7A,$BE,$03,$E3,$9F,$23,$E1,$6C,$7B,$BD,$AF
	.dc.b	$97,$F0,$BF,$31,$B1,$F0,$BE,$17,$E7,$9F,$43,$E1,$2C,$7C,$6F,$85
	.dc.b	$FA,$67,$D4,$F8,$5B,$1D,$4F,$97,$87,$56,$9D,$49,$A7,$DA,$BE,$AF
	.dc.b	$DC,$F8,$2F,$BE,$7E,$03,$E0,$6D,$75,$9F,$C0,$DF,$56,$C7,$4E,$2F
	.dc.b	$C1,$E7,$E9,$8C,$1F,$84,$F8,$4F,$C5,$9F,$8C,$F8,$3B,$1F,$E1,$8F
	.dc.b	$BF,$7C,$1F,$F8,$32,$3F,$C5,$71,$87,$5F,$07,$9C,$99,$2B,$51,$FC
	.dc.b	$61,$7C,$1E,$E7,$5F,$07,$67,$38,$78,$63,$F3,$F7,$23,$F2,$BF,$0D
	.dc.b	$63,$A6,$5C,$4E,$59,$5A,$F5,$65,$D7,$0E,$CD,$8F,$77,$C1,$F8,$4F
	.dc.b	$17,$C2,$D8,$F8,$58,$73,$C7,$D5,$8D,$D3,$FC,$5F,$A7,$A2,$7D,$D9
	.dc.b	$F6,$C3,$D4,$7C,$BF,$81,$F7,$1E,$FF,$85,$B1,$F3,$FE,$03,$E1,$1F
	.dc.b	$0F,$E0,$2C,$7D,$3F,$83,$F8,$C7,$C7,$F8,$1B,$1F,$5F,$E1,$3E,$51
	.dc.b	$F2,$FE,$16,$C7,$52,$69,$F4,$F0,$EA,$FD,$5C,$39,$BF,$5F,$0B,$B7
	.dc.b	$BB,$E1,$3E,$B1,$F5,$FE,$16,$E3,$D2,$7E,$1E,$9E,$98,$C1,$F8,$8C
	.dc.b	$79,$BF,$8A,$C2,$E3,$F8,$8E,$07,$C1,$7E,$0F,$CC,$FC,$23,$6F,$FC
	.dc.b	$D7,$F8,$07,$F6,$86,$FC,$1F,$E2,$E3,$18,$7F,$66,$7F,$DA,$E7,$FB
	.dc.b	$56,$7F,$D0,$6D,$8F,$07,$BF,$C7,$E0,$3F,$25,$F0,$BF,$93,$F8,$5B
	.dc.b	$87,$E8,$DE,$DF,$0B,$DF,$2B,$3E,$23,$CB,$7C,$DF,$D5,$61,$71,$F2
	.dc.b	$7C,$07,$9C,$F4,$7C,$0D,$AF,$6E,$3D,$5A,$75,$AF,$66,$1C,$F8,$FF
	.dc.b	$23,$8D,$AF,$6F,$C2,$FC,$3F,$85,$F8,$8D,$8F,$93,$F0,$1F,$1C,$F9
	.dc.b	$1F,$03,$63,$E6,$FC,$17,$CB,$3E,$67,$C1,$D8,$FA,$3F,$0B,$F3,$CF
	.dc.b	$A1,$F0,$B6,$3A,$93,$6F,$A5,$87,$57,$EA,$61,$CD,$FA,$D8,$5D,$7F
	.dc.b	$D3,$3C,$BF,$D9,$F0,$FB,$FF,$01,$F7,$7E,$07,$EF,$3C,$EB,$7A,$FA
	.dc.b	$AB,$C1,$1C,$B2,$E9,$1F,$80,$C6,$E5,$FC,$5B,$2F,$F6,$23,$F8,$7F
	.dc.b	$82,$FC,$57,$C2,$FE,$2D,$FE,$2D,$F8,$0F,$F6,$3C,$3F,$D7,$6F,$FB
	.dc.b	$27,$23,$9B,$FF,$1A,$B9,$C3,$FB,$DC,$7E,$4F,$E0,$FF,$2D,$7F,$97
	.dc.b	$7F,$31,$F0,$B2,$3F,$3B,$F0,$D6,$B9,$FC,$2F,$8B,$E1,$7C,$6D,$AF
	.dc.b	$D1,$9C,$CA,$7E,$9B,$0F,$4F,$C2,$7A,$BE,$06,$CF,$A9,$E6,$FB,$2E
	.dc.b	$E1,$EA,$7C,$F1,$CC,$EB,$1E,$B7,$E1,$FC,$2F,$C6,$F8,$3F,$8E,$D8
	.dc.b	$F8,$FF,$01,$F2,$8F,$97,$F0,$16,$3E,$5F,$C1,$FC,$E3,$E7,$FC,$1D
	.dc.b	$8F,$9F,$F0,$BF,$4B,$E1,$7E,$9B,$63,$A9,$33,$FA,$38,$75,$7E,$9E
	.dc.b	$1C,$DF,$AB,$85,$E5,$FF,$73,$F2,$BE,$03,$EF,$9F,$80,$F8,$4B,$97
	.dc.b	$5A,$DE,$BE,$A8,$C0,$C7,$FB,$0E,$0C,$BB,$4E,$6F,$F0,$F9,$62,$F7
	.dc.b	$F0,$BF,$8E,$3F,$1F,$F0,$76,$FF,$1F,$3F,$92,$F8,$1F,$C9,$B6,$3F
	.dc.b	$27,$F0,$7F,$96,$F8,$3F,$CB,$B6,$3F,$2F,$F0,$BF,$9A,$F8,$5F,$CD
	.dc.b	$B6,$3A,$9F,$9C,$C3,$AB,$4E,$A7,$E8,$2F,$AB,$4F,$61,$ED,$9F,$EC
	.dc.b	$5A,$23,$1F,$0B,$FD,$FE,$5F,$03,$63,$BE,$7F,$A8,$C6,$E1,$F5,$1E
	.dc.b	$F1,$FA,$CC,$6C,$FD,$87,$A7,$C1,$CC,$BC,$75,$FE,$12,$C9,$F0,$79
	.dc.b	$8F,$F5,$4C,$3E,$07,$C8,$79,$7E,$06,$DF,$2C,$BF,$E0,$0F,$C1,$7A
	.dc.b	$7E,$0F,$D4,$D9,$D6,$0F,$85,$F7,$1E,$FF,$85,$B3,$AC,$BA,$E7,$E2
	.dc.b	$CE,$E7,$F1,$E3,$D1,$97,$57,$CE,$D8,$D0,$CB,$FD,$CF,$87,$8B,$E0
	.dc.b	$3E,$77,$C0,$FC,$F7,$AC,$7D,$5C,$BD,$CF,$8C,$F4,$EB,$D5,$FB,$58
	.dc.b	$DC,$BE,$57,$C2,$7D,$93,$ED,$7C,$25,$AF,$8D,$F8,$17,$F7,$0D,$7C
	.dc.b	$AB,$DF,$FF,$1F,$39,$9F,$D1,$39,$7E,$4F,$C2,$FE,$14,$FC,$37,$C3
	.dc.b	$58,$F9,$BF,$07,$F8,$93,$F1,$5F,$07,$63,$C3,$F0,$3D,$B1,$EE,$D8
	.dc.b	$FA,$BF,$09,$F9,$13,$F2,$5F,$0B,$63,$D3,$F0,$BE,$6F,$85,$F3,$B6
	.dc.b	$3A,$9E,$EC,$3A,$B4,$F8,$3F,$0B,$EE,$3F,$3D,$F0,$B6,$3E,$2F,$C1
	.dc.b	$7F,$05,$E4,$FF,$7C,$F0,$A3,$1C,$5F,$21,$B5,$F1,$BE,$12,$DF,$F9
	.dc.b	$B2,$FF,$D9,$4F,$D0,$F8,$1C,$E0,$94,$61,$DC,$F8,$3E,$13,$1F,$0B
	.dc.b	$23,$EC,$C6,$FE,$0B,$87,$C1,$49,$FB,$B1,$CF,$E1,$3A,$63,$27,$F0
	.dc.b	$31,$DF,$E1,$7C,$18,$49,$FC,$2C,$78,$FE,$0F,$C9,$F0,$92,$7F,$13
	.dc.b	$1D,$5F,$61,$0F,$8F,$7E,$BF,$85,$ED,$1D,$0F,$C8,$BE,$FF,$81,$91
	.dc.b	$F9,$47,$E1,$7C,$27,$C6,$FC,$0D,$F0,$14,$95,$FB,$F8,$7C,$1E,$FE
	.dc.b	$0F,$E5,$7C,$2C,$AD,$47,$3F,$84,$F9,$DD,$12,$79,$47,$47,$FD,$47
	.dc.b	$1F,$D0,$70,$F0,$C7,$7C,$7C,$98,$F7,$8E,$A4,$38,$9F,$EA,$C9,$FF
	.dc.b	$7A,$3E,$7F,$83,$F6,$11,$F5,$7C,$27,$87,$A7,$F8,$3D,$78,$27,$9C
	.dc.b	$23,$0E,$AF,$C5,$3A,$96,$78,$AE,$7F,$0B,$F2,$FD,$E9,$7F,$25,$F9
	.dc.b	$DF,$0B,$F3,$FD,$E9,$3F,$35,$FA,$5F,$09,$F4,$FE,$0E,$4F,$D1,$7E
	.dc.b	$B7,$C2,$7D,$7F,$85,$FB,$1F,$09,$2B,$EB,$47,$47,$CD,$85,$D3,$DF
	.dc.b	$8F,$29,$E8,$43,$D9,$87,$E0,$FE,$0B,$F0,$9F,$09,$27,$F0,$51,$CF
	.dc.b	$97,$EC,$63,$67,$E9,$61,$D5,$F8,$DA,$7A,$9E,$A7,$E2,$FE,$03,$F2
	.dc.b	$04,$7F,$1D,$ED,$FE,$E3,$F0,$4F,$EA,$B1,$FD,$3D,$FF,$72,$B7,$E0
	.dc.b	$F5,$F0,$3B,$F8,$4F,$CD,$4C,$38,$3F,$9D,$F8,$4F,$CF,$7C,$2C,$9F
	.dc.b	$CE,$47,$27,$A1,$0F,$C1,$E1,$E2,$6D,$75,$7F,$04,$78,$4F,$37,$C1
	.dc.b	$FA,$48,$FA,$7E,$17,$B3,$D3,$2F,$71,$8E,$4F,$EB,$38,$7F,$80,$A7
	.dc.b	$F6,$07,$78,$CE,$1F,$17,$E0,$BF,$AF,$F1,$7F,$8A,$6D,$18,$B5,$A7
	.dc.b	$FB,$E5,$7C,$BA,$D4,$E7,$E1,$2D,$1B,$F0,$58,$73,$F0,$5A,$F8,$5B
	.dc.b	$8F,$D5,$78,$FB,$7F,$67,$E0,$E6,$9F,$66,$3E,$E7,$C1,$FD,$DF,$83
	.dc.b	$93,$F7,$63,$F0,$1F,$07,$F8,$1F,$85,$93,$F8,$18,$FE,$77,$E6,$7F
	.dc.b	$69,$8F,$F0,$92,$F7,$72,$6B,$DE,$FC,$57,$BD,$2B,$F1,$4F,$E3,$7E
	.dc.b	$13,$F1,$DF,$09,$27,$F1,$CF,$E4,$7E,$0F,$BF,$C0,$C9,$EE,$F8,$BE
	.dc.b	$07,$C7,$F0,$12,$7C,$6F,$9B,$E0,$BC,$FF,$07,$27,$CE,$FA,$BE,$13
	.dc.b	$D7,$F0,$32,$7D,$6F,$BB,$DC,$F7,$FC,$07,$C0,$F8,$29,$5F,$02,$3E
	.dc.b	$1F,$C1,$C8,$F8,$6F,$C6,$F8,$3F,$8F,$F0,$1F,$7B,$99,$F0,$16,$30
	.dc.b	$E6,$27,$7F,$CC,$F8,$2B,$8F,$F2,$49,$7C,$1F,$33,$A7,$C1,$DB,$FF
	.dc.b	$22,$7C,$DF,$F2,$4B,$BE,$F7,$F7,$28,$C4,$7F,$2E,$7F,$9B,$B7,$DA
	.dc.b	$D1,$1E,$7E,$DF,$1F,$84,$91,$D9,$FC,$1E,$1D,$7E,$0E,$4F,$89,$F0
	.dc.b	$7C,$1F,$38,$EC,$79,$B0,$F2,$61,$E5,$F8,$49,$3E,$A7,$D1,$F0,$9E
	.dc.b	$9F,$80,$F5,$7B,$92,$BD,$F1,$D7,$D2,$FF,$28,$EC,$CB,$1A,$FF,$A7
	.dc.b	$63,$FD,$41,$7F,$1B,$E0,$BE,$21,$1F,$93,$F0,$5F,$1F,$E0,$7E,$47
	.dc.b	$C1,$49,$F9,$B1,$FE,$1C,$FC,$17,$F8,$A6,$7F,$E6,$06,$E5,$FE,$AB
	.dc.b	$6D,$7F,$02,$7F,$94,$BF,$47,$E0,$FE,$A7,$C1,$C8,$FB,$31,$F5,$FD
	.dc.b	$FF,$B1,$F0,$12,$7E,$EC,$78,$7E,$06,$47,$FA,$0F,$0E,$F9,$1E,$EF
	.dc.b	$C4,$F8,$3F,$C1,$12,$7F,$0F,$F0,$7E,$BF,$83,$91,$CD,$F7,$7C,$1F
	.dc.b	$68,$FE,$EB,$87,$62,$7C,$1F,$C3,$F8,$39,$3E,$37,$F2,$0F,$C7,$F8
	.dc.b	$0F,$91,$F0,$32,$3D,$11,$F2,$FD,$EF,$99,$EF,$C9,$F6,$47,$CF,$F8
	.dc.b	$1F,$A1,$F0,$32,$7E,$04,$7D,$3F,$7B,$EA,$7B,$F2,$7E,$24,$75,$7E
	.dc.b	$36,$1D,$5F,$5E,$1F,$6F,$E0,$FE,$E7,$C3,$C9,$F9,$91,$F7,$FD,$DF
	.dc.b	$C0,$7B,$B2,$7E,$84,$75,$7E,$2E,$1D,$79,$BF,$09,$8F,$E1,$FE,$0B
	.dc.b	$F1,$1F,$07,$2F,$EC,$47,$F9,$9B,$F0,$6F,$E5,$91,$FE,$22,$6D,$FF
	.dc.b	$AB,$1B,$DF,$FE,$F7,$7F,$98,$4F,$FB,$0D,$7A,$DC,$C9,$19,$F5,$BF
	.dc.b	$31,$F0,$52,$3F,$10,$FE,$6F,$E0,$BC,$5F,$01,$E3,$F8,$09,$5D,$63
	.dc.b	$9B,$E5,$C2,$E7,$E5,$C7,$D5,$F0,$12,$BC,$8F,$B7,$E0,$3D,$DE,$EC
	.dc.b	$9F,$43,$F0,$7D,$EE,$7B,$FB,$72,$B5,$FD,$AB,$76,$5F,$1F,$E0,$A4
	.dc.b	$FC,$17,$E5,$7C,$17,$CB,$F7,$3E,$67,$B9,$2B,$E3,$47,$FA,$53,$A7
	.dc.b	$F9,$1F,$07,$FB,$61,$BA,$2B,$FB,$CC,$61,$BF,$0D,$F5,$FE,$0F,$EC
	.dc.b	$7C,$1C,$9F,$A5,$1F,$6F,$E0,$FE,$E7,$C1,$C9,$FA,$D1,$E1,$F8,$19
	.dc.b	$1F,$EF,$BC,$7B,$66,$7F,$9A,$BF,$81,$F8,$2F,$C3,$7C,$1C,$AF,$C0
	.dc.b	$47,$AF,$DF,$F6,$7B,$F2,$7A,$47,$67,$FC,$DB,$87,$62,$6D,$F0,$FE
	.dc.b	$0E,$4F,$8D,$FC,$A7,$07,$C7,$F7,$24,$F9,$DF,$95,$EE,$7C,$BF,$6E
	.dc.b	$4F,$AD,$F9,$DE,$E7,$CF,$F8,$39,$3E,$F7,$E9,$7C,$1F,$D3,$F8,$1F
	.dc.b	$A9,$F0,$52,$BE,$24,$74,$7D,$38,$5E,$3F,$B1,$8F,$DB,$F8,$29,$5F
	.dc.b	$2D,$FB,$DF,$07,$F7,$FD,$D9,$3F,$3D,$FC,$17,$BB,$CF,$83,$F0,$18
	.dc.b	$FD,$1C,$B1,$7F,$82,$CB,$F1,$1F,$03,$23,$EC,$3F,$8B,$F8,$2F,$F6
	.dc.b	$FF,$E0,$EF,$73,$FC,$7B,$CF,$F8,$BC,$FA,$9D,$A7,$4F,$E0,$70,$FF
	.dc.b	$21,$C7,$F9,$06,$37,$1F,$E4,$A8,$C3,$FE,$A6,$5E,$E7,$0F,$73,$BF
	.dc.b	$C1,$78,$3E,$0E,$E1,$F8,$D7,$AF,$C1,$76,$F8,$2B,$3D,$B8,$59,$27
	.dc.b	$A4,$E6,$3C,$32,$8F,$56,$5A,$CF,$C9,$F0,$7E,$AF,$82,$F5,$B6,$B9
	.dc.b	$7B,$BE,$EF,$7B,$DE,$D8,$ED,$BF,$3C,$BB,$63,$F0,$72,$C5,$7C,$9F
	.dc.b	$81,$F8,$E7,$C8,$F8,$2B,$1E,$7F,$73,$E5,$9F,$33,$DD,$B1,$EF,$FC
	.dc.b	$21,$F0,$DF,$E7,$27,$FC,$B5,$19,$8C,$7C,$3F,$D4,$F8,$49,$3F,$0F
	.dc.b	$E1,$6E,$BF,$4F,$E1,$7E,$CF,$57,$DA,$6D,$73,$3C,$18,$77,$C7,$AC
	.dc.b	$7C,$DF,$84,$FC,$09,$F8,$2F,$85,$B1,$E7,$F8,$1F,$37,$C0,$F9,$DB
	.dc.b	$1D,$73,$FF,$74,$ED,$EA,$6E,$3D,$9F,$C4,$7C,$2F,$E3,$BE,$17,$F1
	.dc.b	$ED,$BF,$89,$EF,$7C,$33,$E2,$7B,$D6,$3E,$47,$BB,$F1,$CF,$91,$EF
	.dc.b	$58,$F9,$9F,$09,$F2,$CF,$99,$F0,$B6,$3E,$87,$C2,$FC,$F3,$E8,$7C
	.dc.b	$2D,$8E,$A4,$D7,$D7,$87,$57,$DB,$F0,$BF,$5C,$FB,$1F,$09,$6B,$EE
	.dc.b	$7B,$DF,$6C,$FB,$9E,$FD,$8E,$9E,$17,$E3,$07,$27,$F9,$B2,$3F,$09
	.dc.b	$97,$79,$D4,$BE,$0F,$F0,$C7,$E1,$FE,$12,$DF,$EA,$7F,$09,$7B,$FF
	.dc.b	$D2,$76,$FF,$60,$CF,$F0,$C9,$CC,$6E,$75,$8F,$6A,$A6,$FD,$FB,$5F
	.dc.b	$EC,$AB,$7B,$F8,$B4,$62,$7C,$55,$9C,$DF,$09,$49,$3F,$A1,$F8,$5E
	.dc.b	$56,$F8,$3A,$49,$F6,$7C,$26,$B1,$DE,$72,$7B,$9F,$07,$E1,$79,$FC
	.dc.b	$2C,$8C,$1F,$17,$E0,$7B,$91,$F9,$1F,$03,$D5,$F9,$57,$D5,$F9,$27
	.dc.b	$C1,$C3,$CF,$F0,$BE,$8F,$85,$93,$F4,$63,$D7,$EF,$FF,$BB,$36,$37
	.dc.b	$DF,$C3,$FE,$F0,$F8,$3F,$E6,$37,$DB,$93,$E5,$B9,$F8,$0E,$4F,$3F
	.dc.b	$84,$9C,$FF,$7F,$E1,$6C,$F1,$CD,$8F,$84,$F1,$7C,$2D,$BF,$E5,$D1
	.dc.b	$E0,$CB,$F2,$55,$FD,$16,$FC,$53,$9A,$FC,$44,$7A,$FE,$17,$D9,$F0
	.dc.b	$B6,$7F,$20,$FA,$FD,$FF,$67,$C0,$59,$EE,$FA,$DF,$CB,$5D,$3D,$07
	.dc.b	$C7,$F8,$5F,$91,$F0,$B6,$7F,$38,$FC,$7F,$7A,$47,$AB,$DE,$B7,$E8
	.dc.b	$F7,$BE,$61,$F3,$7D,$EB,$5F,$03,$E1,$7E,$81,$F4,$7E,$16,$C7,$C4
	.dc.b	$F8,$3F,$A9,$F0,$9F,$55,$B1,$F2,$30,$E7,$8F,$C6,$C6,$F2,$7E,$EF
	.dc.b	$8F,$FA,$AD,$FF,$7A,$6D,$27,$EA,$FC,$1F,$E0,$7E,$13,$F0,$5E,$FF
	.dc.b	$E0,$FD,$FB,$8F,$D6,$7A,$BC,$F3,$F6,$3E,$33,$ED,$FC,$27,$E2,$CF
	.dc.b	$C6,$7C,$25,$8F,$BF,$EE,$FE,$3C,$FC,$87,$BD,$63,$FB,$87,$E1,$4F
	.dc.b	$7F,$F8,$A4,$7F,$84,$A1,$EF,$E3,$83,$D5,$1B,$9E,$33,$F9,$BF,$84
	.dc.b	$FC,$E7,$C2,$D9,$FC,$7B,$C7,$89,$6F,$83,$C3,$6B,$C7,$F0,$98,$9C
	.dc.b	$E7,$AC,$AD,$78,$0F,$4F,$C2,$72,$39,$FC,$2D,$8F,$6F,$C0,$76,$3B
	.dc.b	$FC,$05,$8F,$83,$87,$56,$9D,$77,$FD,$A6,$8F,$84,$F3,$9E,$8F,$84
	.dc.b	$B5,$F2,$BD,$FF,$5F,$BF,$EC,$6C,$7F,$87,$7F,$0B,$FE,$6A,$FD,$03
	.dc.b	$E2,$38,$C7,$BC,$7C,$4F,$85,$8F,$D6,$F8,$5B,$3E,$4F,$85,$F9,$3F
	.dc.b	$0B,$F2,$9B,$5D,$FE,$13,$E6,$BD,$9A,$FE,$03,$E1,$BE,$89,$F4,$BE
	.dc.b	$1A,$C7,$9F,$E0,$7C,$A7,$9B,$E0,$6C,$75,$3D,$45,$CB,$D2,$FE,$33
	.dc.b	$E1,$7E,$EF,$C2,$FD,$E6,$C7,$C3,$F7,$BE,$0F,$BF,$F0,$9B,$1F,$1F
	.dc.b	$DE,$F8,$BE,$FF,$C6,$6C,$7C,$BF,$85,$F9,$27,$CA,$F8,$5B,$1F,$3F
	.dc.b	$E1,$7E,$6F,$C2,$FC,$E6,$C7,$53,$EA,$61,$D5,$A7,$D7,$F8,$4F,$AB
	.dc.b	$F0,$BF,$59,$B1,$F6,$FE,$03,$EC,$FC,$07,$DA,$6C,$75,$E1,$FE,$79
	.dc.b	$D6,$C7,$B4,$61,$FC,$0F,$C2,$4A,$FC,$37,$C2,$5A,$E3,$F0,$1F,$86
	.dc.b	$F8,$0F,$C3,$B6,$BF,$CD,$1F,$86,$3E,$07,$F5,$CD,$8C,$FC,$0E,$27
	.dc.b	$F1,$FE,$E4,$9F,$CA,$7C,$2D,$CB,$FA,$96,$0F,$84,$E6,$74,$F8,$4B
	.dc.b	$7E,$0C,$7A,$3D,$8C,$1E,$4F,$86,$FC,$EE,$1E,$46,$C7,$A3,$DF,$F3
	.dc.b	$FC,$07,$A1,$B1,$CA,$BC,$77,$EC,$D7,$BC,$79,$BE,$17,$DF,$F0,$BF
	.dc.b	$01,$B7,$F1,$3E,$13,$E1,$FC,$2F,$C4,$6C,$7C,$8F,$73,$E3,$9F,$23
	.dc.b	$E1,$2C,$7C,$CF,$81,$F9,$67,$CC,$F8,$3B,$1F,$43,$E1,$7E,$79,$F4
	.dc.b	$3E,$12,$C7,$52,$6B,$F3,$30,$EA,$FC,$EB,$E6,$D9,$F0,$1F,$73,$E0
	.dc.b	$FE,$DF,$C1,$FD,$C6,$C7,$4F,$83,$FF,$25,$74,$7F,$BD,$E3,$F0,$78
	.dc.b	$F6,$9C,$FC,$2F,$A6,$7F,$0E,$DC,$7F,$C2,$1E,$73,$E1,$3F,$9B,$36
	.dc.b	$BF,$D6,$8F,$FA,$59,$B1,$AF,$5F,$7F,$01,$67,$6F,$2F,$81,$FC,$9F
	.dc.b	$C2,$D9,$FC,$BB,$DB,$E1,$3B,$FC,$05,$9E,$EF,$3C,$FC,$B8,$5D,$7F
	.dc.b	$CD,$1A,$23,$CF,$F0,$1E,$8F,$85,$B8,$FA,$1F,$5F,$C0,$C8,$F5,$FC
	.dc.b	$05,$CB,$E0,$EF,$EA,$9F,$7B,$63,$E2,$FC,$27,$C4,$F8,$4F,$8A,$DA
	.dc.b	$F8,$BA,$7F,$98,$C7,$C0,$D3,$0E,$7E,$0A,$61,$F3,$1E,$1F,$03,$C7
	.dc.b	$E0,$64,$FD,$07,$A7,$C1,$FF,$9B,$FF,$0D,$7B,$9F,$DA,$DB,$9F,$FB
	.dc.b	$A5,$7C,$1E,$67,$5F,$07,$BF,$84,$B7,$B7,$EF,$BF,$80,$F8,$4A,$FE
	.dc.b	$01,$ED,$EF,$78,$7E,$0E,$6B,$DE,$3C,$9E,$CF,$97,$E0,$24,$F8,$E3
	.dc.b	$D1,$F0,$92,$3C,$CF,$AF,$E0,$FD,$9E,$F4,$9F,$53,$EF,$F7,$BA,$C7
	.dc.b	$B7,$0E,$AF,$F9,$9F,$85,$8F,$C5,$F8,$3F,$8D,$F0,$72,$BE,$24,$7C
	.dc.b	$9F,$80,$F9,$5E,$FC,$9F,$91,$1F,$37,$DF,$91,$F2,$DF,$A1,$F0,$5F
	.dc.b	$47,$E0,$E4,$FC,$F7,$EA,$7C,$27,$FA,$8F,$EE,$5E,$E5,$F8,$39,$63
	.dc.b	$FC,$FD,$3A,$71,$E5,$FF,$B0,$E1,$F0,$7C,$C9,$3A,$EA,$DC,$BE,$0F
	.dc.b	$F0,$93,$F8,$5F,$84,$8F,$E0,$CF,$27,$BF,$E5,$F8,$39,$1C,$CF,$F2
	.dc.b	$A5,$A7,$D5,$F0,$92,$75,$9A,$93,$EF,$F8,$4F,$81,$F0,$9F,$07,$E1
	.dc.b	$26,$1C,$2D,$3F,$17,$DA,$93,$97,$E4,$7B,$FF,$27,$E0,$64,$F1,$7E
	.dc.b	$67,$C0,$FC,$DF,$82,$93,$D5,$FA,$1F,$09,$F4,$7E,$0B,$FC,$9D,$9B
	.dc.b	$E0,$EC,$61,$FF,$29,$7E,$1D,$FE,$5C,$E6,$F6,$BF,$E3,$D6,$FB,$FA
	.dc.b	$F7,$F9,$FC,$2F,$4F,$84,$B5,$ED,$79,$FC,$27,$4F,$88,$B3,$D1,$EF
	.dc.b	$87,$38,$EA,$78,$30,$F1,$FC,$2F,$A3,$E1,$6C,$FC,$97,$CF,$EF,$7A
	.dc.b	$3D,$FB,$3E,$87,$AB,$CC,$EB,$C7,$FE,$CB,$DB,$D8,$DF,$84,$F8,$3F
	.dc.b	$09,$F0,$9B,$5F,$0B,$E1,$3E,$2F,$C2,$7C,$66,$C7,$C6,$3E,$49,$F2
	.dc.b	$BD,$9A,$FC,$AF,$84,$F9,$A7,$CE,$F8,$5B,$1F,$3B,$E0,$FE,$8F,$C1
	.dc.b	$FD,$26,$C7,$52,$69,$E9,$C3,$AB,$EB,$F8,$3F,$B3,$F0,$7F,$69,$B5
	.dc.b	$F9,$6E,$A7,$C0,$FF,$A3,$DB,$19,$F8,$1C,$4F,$4F,$25,$3F,$B2,$E3
	.dc.b	$AF,$E1,$33,$D3,$9F,$84,$FC,$4F,$C2,$FE,$29,$B1,$E1,$F8,$1F,$01
	.dc.b	$E1,$F8,$1B,$1E,$5F,$81,$F2,$1E,$5F,$81,$B1,$E9,$F8,$4F,$47,$C2
	.dc.b	$FA,$5B,$1E,$DF,$7F,$D9,$F0,$1E,$D6,$C7,$52,$72,$FF,$B3,$DF,$E1
	.dc.b	$BC,$6C,$7B,$CE,$4F,$F6,$A4,$3E,$13,$E4,$1F,$27,$E1,$6D,$FC,$DF
	.dc.b	$7B,$E6,$7B,$DF,$35,$B1,$E0,$E5,$F9,$78,$F4,$DF,$DD,$39,$3C,$7F
	.dc.b	$07,$F5,$7E,$17,$EB,$36,$3E,$D7,$C1,$FD,$93,$ED,$7C,$25,$8F,$E2
	.dc.b	$7F,$87,$FF,$0C,$6F,$B5,$FE,$CD,$8C,$65,$ED,$CD,$09,$FC,$27,$C2
	.dc.b	$CB,$FC,$47,$C2,$DC,$38,$5F,$F3,$29,$EF,$F0,$5D,$7E,$0F,$B3,$DF
	.dc.b	$DE,$F0,$65,$6B,$C6,$7E,$4F,$0F,$C9,$1F,$93,$F8,$3B,$1E,$AF,$81
	.dc.b	$F4,$7C,$0F,$A5,$B1,$EE,$C3,$AB,$4E,$B7,$E6,$F8,$5F,$81,$F0,$BF
	.dc.b	$05,$B3,$F1,$BD,$FF,$88,$7C,$5F,$7E,$C7,$CA,$F7,$BE,$47,$6F,$C9
	.dc.b	$6C,$7C,$EF,$84,$F9,$9F,$0B,$F3,$5B,$1F,$4B,$E1,$7E,$87,$C2,$FD
	.dc.b	$16,$C7,$53,$EB,$E1,$D5,$A7,$DA,$F8,$5F,$B0,$7D,$9F,$85,$B1,$FD
	.dc.b	$43,$F1,$07,$BF,$F8,$CE,$C9,$18,$9F,$C1,$FB,$F9,$7F,$B5,$C6,$78
	.dc.b	$FE,$DE,$FE,$C6,$CF,$89,$FB,$FF,$0B,$F8,$93,$F1,$5F,$0B,$67,$C5
	.dc.b	$EF,$75,$A6,$4C,$5E,$FE,$63,$15,$67,$37,$EB,$F7,$B0,$67,$DF,$B1
	.dc.b	$EF,$F8,$5E,$07,$1F,$85,$B1,$F0,$FE,$0F,$A7,$C1,$F5,$6C,$75,$26
	.dc.b	$BF,$A1,$C9,$8F,$68,$C7,$C2,$F9,$7E,$17,$CC,$DB,$FA,$1F,$01,$E9
	.dc.b	$3D,$5F,$03,$63,$AF,$07,$D2,$DF,$D5,$18,$3E,$8F,$C2,$FC,$1F,$85
	.dc.b	$F8,$4D,$8F,$B9,$EF,$7C,$53,$E3,$7B,$F6,$3F,$C1,$FF,$88,$BE,$03
	.dc.b	$E6,$36,$33,$F0,$18,$9F,$9D,$F0,$BF,$3F,$E1,$AD,$7E,$25,$E3,$F0
	.dc.b	$B2,$3A,$FC,$35,$E4,$FE,$0D,$97,$56,$E3,$E0,$CB,$B5,$FE,$4C,$FB
	.dc.b	$9F,$0B,$F7,$5A,$7A,$3E,$03,$CE,$7A,$3E,$02,$C7,$B2,$F9,$E1,$E0
	.dc.b	$C2,$F1,$FE,$77,$2F,$F3,$DC,$F9,$3E,$17,$E0,$FC,$2F,$C2,$7E,$1F
	.dc.b	$BD,$32,$F8,$FE,$FD,$BF,$5F,$BF,$F2,$3D,$FF,$92,$DA,$F9,$BF,$09
	.dc.b	$F3,$0F,$9B,$F0,$B6,$3E,$8F,$C1,$FD,$03,$E8,$FC,$25,$8F,$AB,$7C
	.dc.b	$F1,$EF,$1F,$63,$1F,$AF,$F0,$9F,$63,$E1,$6C,$FD,$C7,$ED,$D4,$8F
	.dc.b	$BF,$EF,$D8,$E6,$7D,$4B,$F4,$CF,$53,$F0,$BF,$0B,$F8,$43,$F0,$BF
	.dc.b	$0B,$63,$FC,$AF,$97,$F9,$C9,$BE,$EF,$ED,$61,$EF,$58,$C7,$BD,$9F
	.dc.b	$7A,$DF,$27,$87,$C2,$48,$EB,$F0,$97,$5F,$E9,$D9,$7F,$99,$27,$C3
	.dc.b	$F0,$5D,$FE,$0F,$C0,$F4,$AD,$E3,$7B,$3F,$8E,$60,$8F,$37,$C1,$CC
	.dc.b	$7D,$7F,$09,$70,$EF,$EF,$FA,$FD,$FF,$63,$6B,$E0,$EF,$EB,$69,$EB
	.dc.b	$3E,$2F,$C1,$FC,$3F,$84,$F8,$8D,$8F,$93,$EF,$7C,$7F,$7B,$E4,$36
	.dc.b	$3F,$D0,$7F,$89,$3D,$EF,$E4,$0F,$F1,$ED,$3F,$B8,$39,$9D,$FB,$FF
	.dc.b	$47,$E1,$24,$7D,$6F,$85,$B7,$C3,$E1,$39,$9F,$5F,$E1,$2D,$78,$32
	.dc.b	$EF,$8F,$81,$B1,$F8,$0F,$84,$FB,$C7,$DF,$F8,$5B,$1E,$8F,$7B,$CF
	.dc.b	$EF,$FA,$1B,$1D,$4F,$39,$73,$F5,$E3,$EE,$F8,$59,$3F,$0B,$E1,$6D
	.dc.b	$7C,$0F,$77,$E1,$FB,$DF,$11,$B5,$F2,$3D,$EF,$8E,$7C,$8F,$7A,$C7
	.dc.b	$CC,$F8,$4F,$97,$F0,$BF,$31,$B1,$F4,$3E,$0F,$E7,$FC,$1F,$D0,$6C
	.dc.b	$75,$26,$7F,$4B,$0E,$AF,$D4,$F8,$3F,$AE,$7D,$8F,$84,$B5,$F7,$0F
	.dc.b	$B6,$7D,$CF,$80,$AF,$4F,$1F,$FB,$97,$63,$6F,$54,$E4,$FF,$6D,$C3
	.dc.b	$E1,$3F,$0B,$F0,$9F,$86,$6C,$FF,$B3,$3F,$13,$7B,$DF,$EE,$5D,$7F
	.dc.b	$C7,$11,$87,$51,$9C,$BF,$C0,$52,$F7,$B1,$7F,$EE,$A8,$7B,$D8,$C3
	.dc.b	$FB,$BC,$61,$F0,$46,$6D,$2F,$7B,$0F,$FA,$2A,$BC,$DF,$09,$83,$3F
	.dc.b	$0B,$AF,$85,$DF,$C2,$CA,$F6,$47,$2C,$A4,$78,$0F,$81,$8F,$6F,$85
	.dc.b	$EF,$F0,$B2,$7E,$24,$78,$BE,$03,$C7,$F0,$12,$7E,$44,$74,$7E,$3E
	.dc.b	$17,$6F,$8B,$8F,$AB,$E1,$3D,$7F,$09,$2F,$E8,$47,$BB,$E0,$24,$7D
	.dc.b	$37,$E0,$FC,$07,$C5,$FC,$51,$F0,$17,$E0,$65,$8F,$F3,$74,$E9,$FB
	.dc.b	$9F,$0B,$F2,$BE,$16,$46,$0F,$C0,$1F,$38,$87,$E0,$FE,$17,$C1,$87
	.dc.b	$68,$EB,$86,$1F,$C4,$7C,$07,$D6,$23,$F8,$BF,$85,$F4,$7C,$04,$8F
	.dc.b	$37,$0F,$F9,$27,$44,$FB,$7E,$03,$A5,$76,$37,$DB,$C2,$BF,$A3,$C6
	.dc.b	$0F,$F2,$0B,$EF,$7E,$0F,$C2,$FC,$72,$3B,$C1,$2F,$85,$F9,$9E,$F7
	.dc.b	$CD,$F7,$E5,$EE,$93,$F4,$7D,$EF,$A5,$EF,$CA,$D4,$7D,$5F,$84,$FA
	.dc.b	$DF,$0B,$27,$94,$7D,$9F,$83,$FB,$5F,$09,$27,$B4,$75,$7C,$18,$75
	.dc.b	$7E,$DE,$1F,$81,$F8,$49,$1E,$57,$F0,$9F,$0B,$F8,$5F,$80,$93,$E9
	.dc.b	$7F,$11,$F0,$1D,$63,$F0,$B8,$75,$E1,$FA,$58,$FE,$37,$E1,$65,$7C
	.dc.b	$17,$F2,$1F,$0B,$FB,$5F,$51,$EF,$D2,$57,$CD,$87,$BF,$BF,$7B,$87
	.dc.b	$BF,$2B,$51,$CF,$E1,$64,$71,$7F,$3D,$F0,$BD,$FE,$0E,$4F,$57,$C5
	.dc.b	$F0,$9D,$63,$C1,$87,$57,$FD,$9D,$93,$1F,$47,$C2,$FA,$7E,$16,$57
	.dc.b	$9E,$3D,$9F,$01,$ED,$F8,$09,$3E,$B8,$EC,$F4,$C7,$B7,$1F,$FA,$87
	.dc.b	$2E,$D3,$F0,$7E,$13,$E3,$11,$F8,$9F,$0B,$F2,$7D,$FF,$95,$F0,$12
	.dc.b	$7E,$44,$7F,$21,$FC,$57,$F9,$74,$7F,$B9,$DF,$E5,$8F,$E6,$5A,$E7
	.dc.b	$E0,$7E,$A7,$C2,$C8,$FA,$4F,$D7,$F8,$6F,$B1,$E5,$49,$FA,$CF,$DB
	.dc.b	$D3,$C1,$7E,$1F,$83,$95,$DE,$3F,$01,$F0,$B2,$3E,$F3,$F8,$3F,$86
	.dc.b	$F4,$7B,$FE,$9F,$80,$95,$E7,$8E,$AF,$A7,$0E,$AF,$A4,$F6,$7C,$1F
	.dc.b	$C1,$F8,$49,$18,$3E,$07,$BB,$F1,$48,$FC,$3F,$7B,$E4,$7B,$D2,$3E
	.dc.b	$33,$F2,$FD,$EF,$99,$F0,$72,$7E,$53,$F3,$FE,$13,$E8,$7C,$1F,$D1
	.dc.b	$F8,$49,$5F,$3E,$3A,$3F,$4F,$0B,$D9,$FE,$E1,$E8,$65,$F6,$7E,$12
	.dc.b	$5F,$D7,$7E,$E7,$C2,$7D,$DF,$7F,$EF,$7B,$F2,$BE,$E4,$74,$EA,$FE
	.dc.b	$4B,$8B,$3C,$57,$FB,$7A,$3B,$BF,$83,$F8,$4F,$C4,$7C,$24,$8C,$1F
	.dc.b	$C7,$FF,$16,$7C,$07,$F6,$D7,$F9,$04,$67,$E0,$31,$39,$F7,$B5,$F0
	.dc.b	$12,$B9,$47,$E5,$5F,$CB,$7C,$2C,$7F,$28,$75,$F8,$38,$F8,$5F,$07
	.dc.b	$C1,$F5,$8F,$25,$F5,$7F,$DF,$3B,$FF,$44,$CF,$F9,$7C,$7A,$7E,$17
	.dc.b	$D1,$F0,$B2,$30,$7B,$7E,$03,$D9,$F0,$12,$30,$7A,$0E,$E4,$3B,$8E
	.dc.b	$96,$1E,$08,$CC,$63,$E0,$FE,$47,$C2,$49,$F9,$91,$F2,$FD,$F6,$9F
	.dc.b	$3F,$E0,$2D,$72,$C5,$58,$C6,$42,$B1,$7F,$07,$E0,$2E,$1E,$5A,$46
	.dc.b	$2F,$CF,$F0,$16,$73,$7F,$7F,$E0,$F0,$67,$E1,$2C,$7F,$61,$FC,$5F
	.dc.b	$F9,$CB,$7E,$0B,$A4,$63,$3F,$82,$FF,$3A,$9E,$1F,$05,$A3,$C1,$F0
	.dc.b	$56,$3F,$1D,$F0,$BE,$33,$C9,$F0,$B6,$3C,$1E,$E7,$F8,$F7,$03,$4C
	.dc.b	$7B,$99,$F8,$4D,$7C,$24,$D3,$D1,$1C,$7E,$0B,$97,$C1,$C9,$F6,$47
	.dc.b	$5F,$7B,$D7,$EF,$7B,$1B,$A7,$C1,$F8,$2F,$79,$F0,$3E,$0E,$C7,$C5
	.dc.b	$F8,$4F,$86,$7C,$4F,$84,$B1,$F2,$4F,$8E,$7C,$8F,$82,$AF,$CD,$F8
	.dc.b	$1F,$96,$7C,$CF,$81,$B1,$F4,$7E,$0F,$E7,$9F,$43,$E1,$2C,$7F,$85
	.dc.b	$79,$DF,$05,$FE,$63,$5F,$07,$4C,$46,$7E,$12,$71,$F1,$78,$7C,$27
	.dc.b	$1F,$84,$E5,$F0,$92,$BB,$47,$5F,$84,$FC,$66,$4F,$84,$B1,$87,$3E
	.dc.b	$DE,$BE,$02,$ED,$E7,$78,$FC,$24,$8F,$57,$C2,$5C,$FF,$57,$93,$E0
	.dc.b	$FB,$FC,$27,$81,$B7,$F0,$72,$EC,$D3,$B1,$F1,$7E,$13,$CE,$7A,$3E
	.dc.b	$12,$C7,$C9,$F8,$3F,$59,$EC,$F8,$4B,$1C,$F3,$F1,$ED,$EA,$6C,$7D
	.dc.b	$1F,$83,$F8,$7F,$07,$F1,$1B,$1F,$E4,$9F,$C6,$3F,$95,$B7,$E1,$3F
	.dc.b	$69,$18,$BF,$E1,$2F,$F9,$36,$78,$7C,$1E,$8D,$FC,$05,$8F,$C0,$57
	.dc.b	$D2,$3E,$9F,$C1,$D8,$F0,$7C,$1F,$6F,$83,$EE,$D8,$EA,$4A,$F0,$DF
	.dc.b	$57,$C7,$F0,$7E,$6F,$83,$F3,$B6,$BD,$9F,$07,$EA,$3D,$7F,$07,$63
	.dc.b	$A7,$BD,$FC,$67,$6F,$F3,$1C,$7C,$3C,$BB,$4E,$7E,$0B,$E1,$FC,$1F
	.dc.b	$C4,$6D,$7C,$9F,$81,$F8,$E7,$C8,$F8,$1B,$1F,$E7,$5F,$C6,$7F,$89
	.dc.b	$61,$FC,$9D,$B1,$FE,$88,$7F,$8A,$6F,$FC,$D9,$CB,$FC,$EF,$37,$C0
	.dc.b	$FD,$4F,$81,$FA,$AD,$BF,$B5,$E0,$FD,$83,$EC,$FB,$D6,$3C,$5F,$0F
	.dc.b	$FC,$BF,$2E,$B9,$9F,$0B,$E5,$8F,$C0,$FC,$2C,$AF,$C3,$7C,$2D,$D7
	.dc.b	$CD,$ED,$F3,$3A,$7B,$96,$BB,$92,$7D,$9F,$09,$E1,$C7,$C5,$F0,$92
	.dc.b	$BF,$25,$F0,$96,$FE,$17,$C1,$F9,$CF,$47,$C1,$DA,$F9,$9E,$BF,$AF
	.dc.b	$D8,$F6,$36,$3E,$87,$C1,$7B,$FE,$0F,$E0,$36,$3E,$A7,$C0,$FC,$33
	.dc.b	$E2,$7C,$0D,$8E,$A7,$D9,$C3,$AB,$4F,$B9,$F0,$3F,$2F,$E0,$7E,$63
	.dc.b	$63,$F0,$1F,$01,$F3,$CF,$A1,$F0,$16,$39,$6F,$F7,$F0,$EA,$D8,$FC
	.dc.b	$47,$C1,$7D,$73,$EC,$7C,$1D,$8F,$F6,$07,$E3,$4F,$7B,$FD,$CD,$7F
	.dc.b	$DF,$A3,$0E,$A3,$3D,$9F,$DD,$E5,$EF,$E2,$FF,$DB,$30,$F7,$F1,$87
	.dc.b	$F8,$2E,$3F,$95,$3E,$0F,$7F,$38,$27,$84,$FE,$7E,$F3,$E0,$7F,$82
	.dc.b	$23,$F3,$91,$8F,$84,$E3,$5F,$D1,$27,$D7,$F0,$98,$33,$F0,$9A,$CF
	.dc.b	$79,$CA,$F8,$31,$CB,$E1,$39,$FC,$24,$9F,$8B,$1D,$BD,$FE,$FE,$FC
	.dc.b	$9F,$93,$1D,$5F,$61,$0F,$95,$7E,$6F,$84,$F3,$FC,$24,$9F,$A3,$1E
	.dc.b	$AF,$77,$D7,$EF,$49,$FA,$B1,$F1,$BD,$CF,$F2,$BC,$7F,$BD,$DF,$F2
	.dc.b	$BB,$FE,$34,$6F,$27,$F8,$AD,$DC,$7D,$BF,$83,$F8,$FF,$09,$23,$F0
	.dc.b	$31,$F2,$BE,$0E,$47,$E1,$1F,$9B,$F0,$7E,$1F,$87,$F1,$7C,$1C,$AF
	.dc.b	$14,$7D,$2F,$85,$FA,$7F,$0B,$27,$F1,$B1,$E9,$F8,$1F,$57,$C1,$49
	.dc.b	$F5,$47,$57,$D9,$87,$57,$D9,$87,$DE,$61,$F9,$83,$F0,$3A,$FC,$5F
	.dc.b	$7A,$47,$C5,$7E,$47,$BF,$F2,$7D,$F9,$3F,$25,$F9,$9E,$FF,$CD,$F8
	.dc.b	$3F,$9D,$F0,$92,$BE,$74,$7D,$1F,$84,$91,$F4,$5F,$A9,$F0,$9D,$23
	.dc.b	$E9,$61,$79,$BE,$96,$3F,$67,$E1,$65,$7D,$97,$EE,$7C,$2F,$DD,$F8
	.dc.b	$0F,$BD,$F0,$12,$BE,$F4,$73,$F2,$51,$FE,$75,$E1,$FF,$30,$BE,$B3
	.dc.b	$F0,$98,$FE,$1F,$E0,$E4,$7E,$1D,$FC,$57,$C2,$7E,$C3,$F1,$B7,$BF
	.dc.b	$49,$5F,$8D,$F7,$F5,$EF,$EF,$DF,$93,$E2,$E1,$FF,$3C,$61,$FE,$58
	.dc.b	$8F,$CA,$7C,$2F,$E5,$FE,$16,$46,$0E,$1B,$F1,$A7,$C2,$F8,$FE,$16
	.dc.b	$56,$8F,$F4,$94,$E9,$C6,$28,$F0,$47,$FA,$51,$65,$E6,$8C,$67,$EF
	.dc.b	$F8,$4F,$81,$F0,$B2,$79,$DA,$7E,$27,$BF,$F1,$7E,$02,$56,$A3,$E1
	.dc.b	$BC,$31,$ED,$97,$C2,$CB,$E6,$7C,$1F,$CD,$F8,$49,$7D,$A3,$E8,$7B
	.dc.b	$FF,$47,$DF,$93,$E2,$8F,$E4,$7F,$8E,$7F,$30,$8F,$F5,$7B,$71,$FF
	.dc.b	$53,$37,$8B,$E9,$3F,$CC,$A7,$D7,$F0,$BF,$74,$8F,$BB,$E1,$7F,$01
	.dc.b	$F0,$B2,$3E,$0B,$F8,$3F,$85,$F0,$FC,$24,$9F,$9F,$87,$77,$FD,$88
	.dc.b	$F7,$7E,$87,$C2,$FE,$2C,$8F,$CC,$F8,$6F,$5F,$BD,$EC,$F7,$E4,$F4
	.dc.b	$8E,$CF,$BB,$0B,$B7,$BB,$E1,$BF,$2D,$BF,$E5,$FE,$16,$5F,$D8,$8F
	.dc.b	$8F,$EE,$FF,$B0,$75,$7B,$D6,$30,$E7,$DD,$D7,$BD,$6F,$DA,$F1,$F8
	.dc.b	$3E,$5F,$09,$67,$E0,$BD,$7E,$0A,$47,$C4,$F8,$3B,$A7,$D6,$C7,$AC
	.dc.b	$FC,$86,$C7,$DA,$F8,$3F,$31,$E7,$F8,$4B,$5F,$3B,$DF,$F5,$1E,$BF
	.dc.b	$80,$B1,$CC,$F3,$96,$FF,$0B,$AF,$C1,$F8,$39,$3F,$63,$E1,$2D,$7E
	.dc.b	$28,$F8,$C7,$C7,$F4,$EC,$FE,$B3,$F1,$D7,$BF,$FE,$EB,$5E,$FD,$8C
	.dc.b	$3F,$3F,$E1,$7E,$87,$C2,$DB,$FC,$3B,$C7,$E1,$39,$7C,$2D,$9E,$CF
	.dc.b	$37,$C1,$77,$37,$6A,$7F,$87,$67,$E2,$7C,$DF,$09,$E5,$F8,$5F,$33
	.dc.b	$63,$D5,$F0,$1E,$93,$D5,$F0,$16,$3D,$DB,$FA,$9A,$78,$8F,$85,$F0
	.dc.b	$9F,$04,$F8,$5F,$0B,$63,$E3,$7B,$FF,$17,$DF,$F8,$CD,$8F,$E2,$A7
	.dc.b	$F2,$C3,$E6,$7B,$DE,$BD,$FD,$B1,$FC,$8B,$DF,$F9,$D3,$F3,$FE,$16
	.dc.b	$4F,$D3,$F8,$5B,$7C,$0F,$8B,$C2,$F8,$3F,$AC,$7D,$7F,$84,$B7,$E2
	.dc.b	$CB,$AB,$4F,$19,$F7,$FE,$17,$EF,$1F,$7F,$E1,$6C,$7A,$BD,$FF,$47
	.dc.b	$3C,$8F,$F3,$F9,$FE,$7E,$3F,$9D,$C6,$73,$F6,$E1,$E0,$7D,$15,$F6
	.dc.b	$E3,$0F,$4C,$DF,$0B,$F8,$F3,$F2,$1F,$0B,$71,$F9,$7E,$E7,$5B,$FC
	.dc.b	$27,$BB,$65,$36,$31,$EE,$67,$DD,$B5,$F5,$1E,$1F,$0B,$23,$EB,$FC
	.dc.b	$2D,$DB,$CF,$0F,$82,$ED,$F0,$7D,$DB,$1F,$7B,$1E,$CD,$3B,$1F,$82
	.dc.b	$F8,$2F,$31,$E7,$F8,$3B,$1F,$86,$F7,$FD,$5E,$FF,$AD,$B1,$CF,$2F
	.dc.b	$1E,$BE,$A6,$C7,$E3,$BE,$0F,$E1,$1F,$0F,$E0,$EC,$7E,$48,$F8,$C7
	.dc.b	$C7,$F7,$6B,$FD,$AB,$F1,$E7,$BB,$F3,$5B,$19,$F7,$71,$3F,$3F,$E1
	.dc.b	$24,$FE,$7B,$E1,$2E,$7F,$7B,$07,$C1,$F3,$3A,$7C,$25,$BF,$06,$3D
	.dc.b	$1E,$C6,$0F,$27,$C2,$78,$CF,$27,$C2,$58,$F4,$7B,$DE,$73,$D1,$EF
	.dc.b	$58,$E5,$9F,$8F,$0F,$66,$DD,$E3,$CD,$F0,$7E,$F3,$E0,$7C,$1D,$BF
	.dc.b	$89,$EE,$FC,$3F,$77,$E2,$36,$3E,$45,$FC,$73,$E4,$7B,$B6,$3E,$67
	.dc.b	$C1,$FC,$BF,$83,$F9,$8D,$8F,$A1,$F0,$5F,$3F,$E0,$BE,$83,$63,$A9
	.dc.b	$30,$F9,$98,$75,$7E,$77,$C1,$FD,$73,$EC,$7C,$1D,$AF,$F1,$2F,$E3
	.dc.b	$EF,$7B,$F5,$0B,$DF,$B1,$87,$A7,$99,$FB,$EE,$BF,$F2,$6D,$FE,$17
	.dc.b	$3C,$BF,$E3,$F8,$7C,$27,$E1,$CF,$C4,$7C,$25,$9F,$0F,$BD,$E0,$F7
	.dc.b	$BC,$2D,$8F,$2F,$BB,$E4,$3C,$BE,$F5,$8F,$4F,$C2,$7A,$3E,$13,$D2
	.dc.b	$D8,$F6,$FC,$1F,$B3,$E0,$FD,$AD,$8E,$A4,$E5,$FE,$23,$C6,$C7,$B4
	.dc.b	$63,$E0,$FE,$2F,$C2,$7C,$66,$DF,$CA,$F7,$FE,$49,$F2,$BD,$FB,$1D
	.dc.b	$79,$7F,$92,$E4,$DF,$D7,$39,$3F,$CC,$D0,$F8,$3F,$A6,$7D,$4F,$83
	.dc.b	$B3,$F6,$3D,$EF,$AF,$EF,$7D,$86,$C7,$F9,$83,$F2,$0F,$E3,$ED,$F7
	.dc.b	$BF,$92,$46,$38,$3F,$97,$A8,$FC,$0F,$C2,$4A,$FC,$2F,$C2,$5B,$DF
	.dc.b	$C2,$72,$39,$FC,$45,$AF,$07,$C1,$F6,$C7,$BB,$63,$C9,$F0,$BF,$8E
	.dc.b	$F8,$6F,$1B,$63,$D1,$F0,$1E,$63,$CF,$F0,$16,$3A,$9E,$B2,$E7,$EA
	.dc.b	$C7,$DB,$F0,$B2,$7E,$17,$C2,$DA,$F7,$FB,$FF,$0B,$DF,$F8,$6D,$AF
	.dc.b	$90,$7C,$6F,$7F,$E3,$B4,$F9,$9F,$09,$F2,$8F,$97,$F0,$B6,$3E,$87
	.dc.b	$C1,$FC,$E3,$E7,$FC,$25,$8E,$A7,$D5,$BE,$AD,$3E,$C7,$C2,$FD,$6F
	.dc.b	$85,$FA,$ED,$8F,$F4,$DF,$1B,$E0,$7F,$AF,$2F,$81,$B1,$87,$AE,$FF
	.dc.b	$7B,$1B,$E1,$7F,$76,$EC,$47,$E1,$3E,$16,$65,$F8,$9F,$85,$B8,$7F
	.dc.b	$74,$CF,$FA,$0B,$E2,$F8,$0F,$F6,$6B,$96,$E5,$FE,$02,$9D,$CF,$1F
	.dc.b	$80,$CE,$1F,$EC,$39,$7C,$06,$2B,$FC,$77,$18,$7D,$F5,$9B,$4F,$F4
	.dc.b	$89,$F8,$7F,$09,$98,$37,$E1,$71,$F0,$3F,$1B,$E0,$BE,$3B,$73,$F9
	.dc.b	$7F,$03,$F2,$BE,$0B,$E5,$B6,$3E,$7F,$C2,$FC,$EF,$85,$F9,$ED,$8F
	.dc.b	$A7,$EE,$7D,$23,$E9,$FB,$96,$3E,$BF,$C1,$FD,$6F,$84,$FA,$ED,$8F
	.dc.b	$B7,$F0,$9F,$68,$FB,$7F,$09,$63,$FD,$B3,$F9,$0B,$DE,$F8,$4B,$DF
	.dc.b	$A6,$23,$3F,$0B,$AF,$85,$9C,$FC,$A3,$E3,$3F,$1F,$E1,$63,$F8,$C3
	.dc.b	$AF,$C0,$FF,$8A,$71,$7C,$15,$8C,$39,$F8,$09,$A7,$9B,$E0,$6E,$1F
	.dc.b	$E3,$AC,$1F,$07,$CF,$E1,$3A,$36,$FD,$BF,$07,$EC,$F8,$39,$86,$0F
	.dc.b	$83,$F0,$7F,$00,$8F,$C4,$F8,$4F,$87,$F0,$72,$3E,$3B,$F1,$BE,$0F
	.dc.b	$F1,$1E,$2B,$DC,$A4,$AF,$9B,$E3,$7C,$CF,$82,$F9,$BF,$07,$27,$E8
	.dc.b	$C7,$D0,$F8,$39,$1F,$51,$FA,$7F,$07,$DB,$E0,$7F,$CA,$19,$7F,$2B
	.dc.b	$E1,$5F,$F9,$72,$3F,$98,$1F,$E6,$63,$4E,$7D,$EE,$3E,$F7,$26,$D7
	.dc.b	$2A,$E1,$F0,$7D,$8E,$FF,$07,$67,$46,$E5,$EF,$FC,$2F,$80,$99,$68
	.dc.b	$FF,$54,$CE,$9C,$3F,$EA,$CF,$80,$F9,$24,$8C,$C3,$E0,$3E,$6F,$C1
	.dc.b	$C9,$CB,$F4,$3E,$0F,$E8,$FB,$5F,$10,$F8,$BE,$D5,$D3,$9F,$C1,$7C
	.dc.b	$8F,$82,$F9,$2D,$8E,$FF,$03,$F3,$3E,$0B,$E6,$B6,$3C,$FF,$91,$7F
	.dc.b	$92,$9B,$EE,$7F,$A0,$63,$1A,$FF,$A3,$1F,$E0,$B3,$C3,$E0,$F4,$6F
	.dc.b	$E0,$EC,$7B,$8F,$B6,$7D,$CF,$82,$AF,$83,$D9,$FE,$29,$66,$98,$F6
	.dc.b	$B3,$F0,$FA,$F8,$39,$9F,$A2,$38,$FC,$17,$2F,$83,$93,$EC,$8E,$BE
	.dc.b	$D7,$6F,$6A,$4F,$C0,$8F,$0F,$C0,$F8,$BE,$0A,$4F,$C4,$8F,$2F,$C0
	.dc.b	$C8,$F8,$EF,$A3,$E0,$BD,$31,$F1,$FD,$AF,$90,$DE,$1F,$9B,$F0,$5F
	.dc.b	$2C,$F9,$9F,$05,$63,$E8,$FC,$0F,$CF,$3E,$87,$C1,$58,$FC,$5D,$FF
	.dc.b	$02,$6F,$B1,$FE,$B5,$87,$B3,$63,$1F,$03,$2B,$8F,$C1,$5B,$FF,$62
	.dc.b	$9F,$E5,$78,$EB,$F0,$5C,$FE,$0B,$A3,$D7,$D9,$FE,$91,$4F,$66,$98
	.dc.b	$8C,$FC,$16,$BE,$0A,$57,$A2,$38,$FC,$14,$8F,$5B,$D3,$E0,$BA,$FB
	.dc.b	$3D,$BD,$99,$5F,$02,$3C,$3F,$03,$23,$E1,$BE,$4F,$81,$F2,$FC,$0C
	.dc.b	$9F,$8E,$FA,$3E,$07,$FB,$27,$8E,$FC,$8D,$ED,$E1,$B7,$FB,$B9,$D4
	.dc.b	$E7,$DB,$D7,$C1,$7C,$2F,$82,$97,$BC,$3F,$92,$3F,$5B,$E0,$7E,$3F
	.dc.b	$C1,$48,$C1,$BF,$6B,$C1,$ED,$48,$C1,$CF,$E0,$BC,$84,$7B,$7C,$37
	.dc.b	$9F,$E0,$BD,$1F,$07,$27,$C5,$1E,$BF,$67,$D9,$ED,$49,$F3,$47,$BF
	.dc.b	$E0,$E4,$7A,$5F,$85,$F0,$7F,$0F,$E0,$FE,$27,$C1,$4A,$F7,$47,$C7
	.dc.b	$F1,$64,$7C,$17,$E5,$7B,$7F,$2F,$E0,$A4,$FC,$57,$E7,$7C,$17,$CF
	.dc.b	$F8,$19,$3F,$25,$FA,$5F,$05,$F5,$BF,$24,$7B,$77,$DC,$97,$F5,$67
	.dc.b	$2D,$F8,$6E,$1F,$09,$23,$06,$FE,$13,$A7,$C2,$48,$C1,$CF,$DC,$F0
	.dc.b	$7B,$92,$30,$77,$F8,$4F,$27,$C2,$48,$C1,$E3,$F8,$3F,$47,$C2,$48
	.dc.b	$C1,$E7,$F7,$7D,$9E,$EC,$8C,$1E,$BF,$83,$F8,$04,$7D,$DF,$09,$F0
	.dc.b	$FE,$0E,$47,$C1,$7E,$37,$C2,$7E,$73,$F2,$5F,$F6,$84,$FE,$65,$AB
	.dc.b	$E1,$9D,$BF,$DA,$BD,$15,$6A,$73,$49,$CF,$BD,$F4,$E2,$1F,$44,$FA
	.dc.b	$DF,$0D,$F5,$FE,$1A,$47,$D5,$7E,$D7,$C3,$79,$7D,$DF,$37,$BB,$2B
	.dc.b	$CD,$1E,$9F,$85,$91,$E9,$7D,$9F,$0D,$ED,$F8,$6F,$77,$C3,$4A,$F7
	.dc.b	$47,$C1,$F7,$7E,$17,$BB,$27,$E1,$47,$C5,$F8,$4F,$8D,$F0,$B2,$7E
	.dc.b	$34,$7C,$9F,$84,$91,$F2,$5F,$99,$F0,$BF,$37,$DC,$F8,$BC,$BF,$1B
	.dc.b	$97,$E4,$C6,$39,$FE,$67,$3B,$DC,$C6,$5F,$2E,$15,$8B,$FA,$3F,$09
	.dc.b	$65,$F0,$B6,$30,$E7,$E1,$66,$5F,$77,$E1,$6E,$5F,$76,$3E,$E7,$92
	.dc.b	$F6,$FB,$36,$3F,$8E,$2F,$73,$33,$AF,$85,$DF,$C2,$DB,$DB,$CB,$E1
	.dc.b	$7C,$DF,$0B,$67,$F1,$CF,$6F,$6F,$BF,$B9,$67,$BB,$E2,$F8,$49,$1E
	.dc.b	$2F,$84,$BB,$7A,$4F,$C8,$6D,$FE,$3C,$9F,$5F,$C2,$7A,$BE,$13,$D6
	.dc.b	$FB,$3D,$DF,$6F,$BB,$6B,$DA,$FC,$0F,$83,$91,$F0,$3E,$12,$E5,$DF
	.dc.b	$AB,$FD,$7D,$F0,$9F,$18,$F8,$FF,$09,$6F,$FA,$17,$E4,$DF,$EC,$56
	.dc.b	$FB,$BF,$66,$3F,$D8,$D1,$83,$E6,$9F,$40,$FA,$3F,$0B,$5F,$A3,$F0
	.dc.b	$BF,$53,$E1,$7E,$AB,$63,$BF,$B9,$D7,$DC,$EC,$D8,$F1,$FC,$2F,$84
	.dc.b	$F1,$7C,$2D,$8F,$3F,$C2,$79,$4F,$37,$C2,$58,$F5,$FB,$9E,$9F,$73
	.dc.b	$D4,$D8,$F7,$FC,$27,$B7,$E1,$3D,$CD,$8F,$87,$F0,$9F,$07,$E1,$3E
	.dc.b	$13,$63,$E3,$FB,$DF,$17,$DE,$F8,$CD,$8F,$97,$F0,$9F,$27,$E1,$3E
	.dc.b	$53,$63,$E7,$FC,$27,$CD,$3E,$77,$C2,$58,$FE,$E9,$F9,$3F,$FA,$E5
	.dc.b	$BE,$F7,$EA,$E3,$FD,$75,$18,$37,$F1,$19,$F8,$6D,$36,$3E,$EF,$83
	.dc.b	$F6,$FE,$1B,$EE,$36,$3B,$FB,$BD,$7D,$DE,$CD,$8F,$1F,$C2,$F8,$4F
	.dc.b	$17,$C3,$58,$F3,$FC,$37,$94,$F3,$7C,$35,$8F,$5F,$B7,$E9,$3D,$5E
	.dc.b	$DD,$8F,$7F,$C2,$7B,$4F,$77,$C2,$D8,$F8,$7F,$09,$F0,$7E,$17,$E1
	.dc.b	$36,$3F,$C3,$5F,$94,$7F,$AD,$DB,$EE,$7C,$7F,$73,$F8,$94,$E7,$8B
	.dc.b	$F4,$9C,$DF,$C5,$AF,$E7,$FC,$27,$CD,$F8,$5F,$9C,$FC,$FF,$84,$93
	.dc.b	$F5,$3E,$12,$E1,$FA,$DC,$7F,$74,$FF,$4B,$6F,$AA,$7F,$D6,$FF,$0D
	.dc.b	$FC,$AD,$C9,$FC,$5E,$79,$7C,$2F,$9B,$E1,$7C,$EF,$A3,$E1,$7D,$3F
	.dc.b	$0B,$73,$F7,$BE,$CF,$6A,$47,$C2,$F6,$AE,$3E,$5F,$85,$F8,$27,$C2
	.dc.b	$F8,$5B,$5F,$23,$E1,$7E,$29,$F1,$BE,$16,$C7,$CC,$F6,$BE,$4F,$B7
	.dc.b	$F2,$9B,$1F,$43,$E0,$BE,$6F,$C1,$7C,$E6,$C7,$D4,$F8,$2F,$A2,$7D
	.dc.b	$2F,$82,$B1,$FE,$51,$ED,$FF,$57,$F3,$7F,$4F,$6C,$7F,$79,$7F,$D6
	.dc.b	$3B,$FF,$1B,$72,$FF,$79,$7F,$8C,$39,$F8,$2F,$EA,$11,$FC,$62,$31
	.dc.b	$3F,$86,$3F,$06,$7E,$13,$E0,$6B,$E3,$F5,$B8,$9C,$BD,$7B,$1E,$7F
	.dc.b	$80,$EA,$76,$F8,$1B,$1E,$BF,$81,$F0,$9E,$2F,$81,$B1,$EF,$F5,$FC
	.dc.b	$BE,$BF,$99,$B1,$F0,$FD,$EF,$49,$EA,$F7,$EC,$7C,$7F,$7B,$DA,$7B
	.dc.b	$BD,$FB,$1F,$2F,$83,$E0,$9F,$0B,$D8,$B1,$F3,$FD,$EF,$8B,$EF,$FC
	.dc.b	$66,$C7,$D3,$F7,$BE,$49,$F2,$BD,$FB,$1F,$E8,$7F,$CA,$7F,$D6,$98
	.dc.b	$FF,$4D,$6F,$4F,$F8,$1D,$FF,$5A,$CE,$FD,$9F,$F0,$14,$7F,$1B,$AE
	.dc.b	$1F,$01,$8F,$88,$CB,$AF,$80,$DF,$C0,$DC,$7C,$0F,$2F,$5E,$47,$8F
	.dc.b	$D8,$B8,$F8,$FE,$03,$BF,$C0,$F8,$1B,$5E,$9F,$81,$F1,$9E,$4F,$81
	.dc.b	$B1,$ED,$F6,$3C,$FE,$C7,$A1,$B1,$F0,$7E,$03,$D7,$F0,$1E,$C6,$C7
	.dc.b	$C5,$F8,$0F,$79,$F0,$3E,$02,$C7,$FA,$BF,$F2,$AF,$F5,$9B,$7D,$9F
	.dc.b	$F1,$34,$3D,$9B,$18,$F3,$3E,$4F,$C0,$DB,$FA,$4F,$CC,$F7,$FE,$6F
	.dc.b	$C0,$59,$FA,$CF,$4F,$63,$AF,$B1,$67,$C4,$F8,$3E,$1A,$47,$97,$E1
	.dc.b	$EF,$47,$F8,$27,$6F,$E2,$DE,$0F,$F9,$0D,$FF,$16,$57,$AF,$E0,$3D
	.dc.b	$3F,$01,$EA,$F6,$24,$FB,$FD,$8B,$8F,$3F,$80,$F7,$FC,$07,$C0,$6D
	.dc.b	$7C,$5F,$80,$F8,$67,$C4,$F8,$0B,$1F,$24,$F8,$E7,$C8,$F6,$EB,$F3
	.dc.b	$7D,$FF,$96,$7C,$CF,$80,$B1,$F4,$7D,$FF,$9E,$7D,$0F,$80,$B1,$FE
	.dc.b	$E3,$FC,$AF,$FD,$66,$DF,$73,$FC,$B5,$0F,$72,$C6,$3E,$1E,$57,$1F
	.dc.b	$87,$BC,$3F,$E3,$1C,$3F,$8A,$F3,$FF,$9A,$DB,$5F,$E9,$57,$F8,$B4
	.dc.b	$67,$E0,$FB,$FB,$7E,$0F,$6E,$DF,$95,$F1,$FC,$1F,$93,$E0,$EC,$FA
	.dc.b	$5F,$3F,$C1,$FA,$3E,$0E,$CF,$B5,$F5,$FB,$92,$3E,$07,$B9,$76,$F4
	.dc.b	$7C,$17,$C0,$3E,$0F,$C1,$DA,$F8,$DF,$07,$F1,$0F,$8B,$F0,$76,$17
	.dc.b	$FE,$51,$F6,$DE,$E7,$FA,$0D,$15,$C3,$F3,$A2,$4F,$D3,$F8,$3B,$1F
	.dc.b	$2F,$E0,$BE,$9F,$C1,$FD,$46,$D7,$83,$9B,$FD,$31,$83,$DB,$F0,$1E
	.dc.b	$1F,$6E,$D6,$CF,$F6,$34,$FF,$1A,$6F,$C3,$F9,$FE,$13,$D0,$FA,$7E
	.dc.b	$0B,$D5,$F0,$76,$B9,$6B,$FE,$CB,$C3,$FC,$6A,$FC,$1F,$6F,$E0,$7B
	.dc.b	$7F,$05,$B5,$C3,$E1,$BE,$1F,$C2,$7C,$53,$E3,$7C,$25,$9D,$62,$F8
	.dc.b	$3F,$94,$7C,$BF,$84,$B3,$AF,$73,$E7,$7B,$9F,$3D,$B1,$CB,$E0,$FE
	.dc.b	$97,$C2,$7D,$36,$C7,$6F,$83,$FA,$DF,$09,$F5,$DB,$1E,$7F,$CB,$3F
	.dc.b	$81,$BF,$EE,$D6,$F1,$7F,$BA,$DF,$E0,$AF,$FA,$45,$CF,$01,$FE,$43
	.dc.b	$19,$F8,$5F,$F6,$34,$63,$AB,$FD,$A0,$FF,$23,$BF,$07,$C2,$68,$FC
	.dc.b	$57,$C2,$58,$F2,$7B,$3C,$8E,$7E,$CD,$8F,$47,$C2,$76,$3B,$FC,$25
	.dc.b	$8F,$67,$C2,$78,$8F,$1F,$C2,$58,$F8,$1E,$D7,$9B,$DA,$F3,$B6,$3E
	.dc.b	$27,$C0,$FA,$8F,$5F,$C0,$D8,$F9,$1F,$01,$EE,$3D,$FF,$03,$63,$F0
	.dc.b	$5F,$96,$FF,$04,$E5,$73,$23,$F8,$2C,$FE,$B2,$31,$ED,$7F,$BB,$67
	.dc.b	$35,$F4,$6B,$E5,$FC,$17,$CC,$69,$F6,$7E,$0B,$E7,$9F,$43,$E0,$AC
	.dc.b	$78,$BD,$9E,$9E,$CF,$56,$C7,$9B,$E0,$7C,$07,$87,$E0,$AC,$7A,$BE
	.dc.b	$0B,$C8,$79,$7E,$0A,$C7,$BB,$D9,$F4,$1E,$9F,$66,$C7,$C2,$F8,$1F
	.dc.b	$67,$C0,$FB,$5B,$1F,$1B,$E0,$7E,$01,$F0,$7E,$06,$C7,$CA,$F6,$BE
	.dc.b	$27,$B7,$F1,$5B,$1F,$3B,$E0,$7E,$41,$F2,$7E,$06,$C7,$D2,$F8,$1F
	.dc.b	$99,$F0,$3F,$35,$B1,$FB,$7F,$CB,$BF,$C3,$5E,$62,$3E,$A4,$7F,$00
	.dc.b	$8C,$75,$7D,$A8,$DC,$EB,$DB,$CD,$6B,$B9,$5F,$C9,$AB,$B7,$C1,$69
	.dc.b	$EB,$EC,$5C,$7A,$4F,$E0,$5F,$C1,$7C,$15,$AF,$C5,$BB,$C3,$AF,$B3
	.dc.b	$71,$FE,$15,$5E,$BF,$67,$13,$9F,$82,$93,$EF,$F8,$2B,$93,$5F,$46
	.dc.b	$3F,$C9,$E3,$DD,$F0,$5D,$7E,$0B,$B3,$DF,$D8,$FE,$87,$3C,$71,$F8
	.dc.b	$F1,$8C,$39,$B6,$3E,$5F,$B3,$6B,$E5,$7C,$05,$DF,$E3,$B7,$33,$7E
	.dc.b	$07,$1F,$03,$2F,$EC,$3B,$F8,$1E,$5E,$6B,$DB,$E5,$ED,$CB,$FE,$69
	.dc.b	$39,$7F,$01,$F0,$1D,$FE,$06,$46,$0F,$C2,$7C,$07,$8C,$8F,$E1,$FE
	.dc.b	$07,$BF,$B7,$E0,$F6,$E4,$F6,$8F,$1F,$C2,$F9,$3E,$16,$4F,$8A,$3C
	.dc.b	$FF,$0F,$E8,$F8,$49,$3E,$68,$F5,$FB,$7E,$CF,$6E,$4F,$AA,$3D,$FF
	.dc.b	$07,$F0,$3E,$0E,$4F,$BA,$3E,$1F,$C1,$C8,$F8,$2F,$C6,$F8,$3F,$8F
	.dc.b	$EE,$FC,$8F,$76,$57,$C6,$8F,$97,$F0,$52,$3E,$4B,$F3,$BE,$0F,$E7
	.dc.b	$FC,$1F,$D0,$F8,$39,$5F,$3A,$3E,$E7,$E5,$EF,$7A,$92,$7F,$B3,$43
	.dc.b	$DE,$DF,$64,$9C,$BC,$BE,$37,$EE,$F8,$FF,$7B,$E1,$25,$7D,$C8,$EF
	.dc.b	$EE,$7E,$97,$13,$7D,$CC,$39,$F8,$39,$5E,$2F,$84,$B8,$7E,$6C,$FE
	.dc.b	$6F,$87,$F9,$21,$C7,$96,$8F,$F2,$45,$7A,$BE,$13,$C1,$F0,$9E,$1F
	.dc.b	$6E,$4F,$BB,$DC,$B8,$F2,$C1,$F0,$9E,$7F,$84,$F4,$36,$FE,$2F,$C1
	.dc.b	$FA,$CF,$67,$C2,$58,$F9,$38,$7B,$CF,$81,$EE,$D8,$F9,$BF,$07,$F0
	.dc.b	$CF,$89,$F0,$96,$3E,$8F,$C2,$7C,$7F,$84,$F9,$0D,$8F,$F0,$D7,$E6
	.dc.b	$1F,$C0,$F8,$CD,$3F,$D8,$79,$7C,$47,$1F,$11,$FC,$55,$FE,$51,$5F
	.dc.b	$D7,$5F,$E4,$F3,$FD,$89,$CB,$7E,$13,$81,$C7,$E1,$6D,$78,$BD,$9F
	.dc.b	$E0,$D6,$F6,$69,$88,$CF,$C2,$6B,$E1,$66,$5E,$A8,$E3,$F0,$B2,$3D
	.dc.b	$AF,$4F,$85,$EB,$EC,$FA,$4F,$57,$B3,$74,$F8,$7F,$05,$ED,$3D,$DF
	.dc.b	$05,$63,$E3,$FC,$0F,$C1,$3E,$17,$C0,$D8,$FF,$22,$7E,$63,$FE,$23
	.dc.b	$7F,$C0,$0F,$F8,$8F,$4F,$C6,$B9,$7F,$C0,$EE,$7E,$13,$FA,$63,$FC
	.dc.b	$76,$79,$FC,$1F,$CF,$F8,$3F,$E8,$93,$9B,$FA,$9F,$07,$F4,$CF,$A9
	.dc.b	$F0,$77,$0F,$1F,$B3,$C0,$E3,$EC,$D8,$F3,$FC,$17,$43,$AF,$C1,$D8
	.dc.b	$F5,$FC,$17,$80,$F0,$FC,$1D,$8F,$7F,$B1,$E4,$3C,$BE,$CD,$8F,$87
	.dc.b	$F0,$1E,$8F,$81,$F4,$B6,$3E,$3F,$C0,$7B,$3E,$07,$DA,$D8,$F9,$67
	.dc.b	$F4,$9E,$2F,$EE,$BC,$5F,$2A,$31,$EC,$E7,$E0,$35,$F0,$32,$BE,$9C
	.dc.b	$71,$F8,$19,$1F,$59,$E9,$F0,$3F,$D5,$7E,$0D,$ED,$5F,$6A,$5F,$F9
	.dc.b	$FE,$72,$DF,$82,$E1,$F0,$52,$30,$6F,$E0,$BA,$11,$E5,$F0,$5D,$FD
	.dc.b	$9F,$07,$B3,$27,$B4,$78,$FE,$07,$C9,$F0,$52,$7C,$51,$E7,$F8,$29
	.dc.b	$1E,$57,$D5,$F0,$5E,$BF,$66,$4F,$A5,$F7,$7B,$5E,$FF,$82,$F8,$1F
	.dc.b	$05,$2B,$DD,$1F,$0F,$E0,$24,$7C,$17,$E3,$7C,$0F,$F7,$BF,$CC,$BF
	.dc.b	$C4,$13,$FC,$A1,$FF,$62,$F6,$7F,$A9,$5C,$F9,$DF,$E2,$77,$33,$FC
	.dc.b	$BB,$9B,$FD,$9F,$5F,$37,$E0,$BF,$C4,$11,$89,$D4,$63,$E0,$A5,$7D
	.dc.b	$3F,$82,$B5,$E1,$F6,$7F,$C5,$18,$FF,$B2,$6B,$C5,$EC,$E2,$73,$F0
	.dc.b	$DA,$F8,$7B,$5E,$77,$8F,$C1,$F2,$F8,$8B,$3E,$B7,$AF,$B3,$F0,$49
	.dc.b	$9F,$BF,$D9,$F8,$9F,$01,$23,$E1,$3F,$1F,$E0,$7E,$47,$C0,$49,$F8
	.dc.b	$CF,$CB,$F8,$1F,$99,$ED,$FA,$BD,$BF,$5B,$74,$F9,$9F,$03,$EE,$3D
	.dc.b	$FF,$03,$63,$E8,$7C,$0F,$C2,$F8,$1F,$86,$D8,$FF,$6C,$7E,$67,$FE
	.dc.b	$BC,$3F,$D7,$FD,$9F,$E5,$B7,$1E,$DF,$F9,$45,$FE,$2D,$F0,$C7,$F8
	.dc.b	$B5,$7F,$A6,$DC,$98,$DF,$FD,$42,$FF,$15,$BE,$9F,$07,$F5,$3E,$0F
	.dc.b	$A3,$D7,$D9,$FF,$22,$51,$A6,$3D,$AC,$FC,$1E,$BE,$0E,$65,$EA,$8E
	.dc.b	$3F,$07,$CB,$E0,$E4,$FB,$A3,$AF,$B5,$23,$E0,$BE,$0F,$6B,$C3,$F0
	.dc.b	$52,$7E,$2B,$E4,$F8,$2F,$2F,$C0,$C9,$F9,$2F,$A3,$E0,$BF,$CB,$9F
	.dc.b	$9A,$7F,$B0,$79,$4F,$FA,$F1,$BE,$DF,$F9,$EE,$31,$C5,$FE,$91,$E6
	.dc.b	$FA,$67,$CF,$F8,$2F,$A0,$FD,$1F,$81,$FA,$5F,$05,$63,$EC,$BC,$FD
	.dc.b	$9F,$09,$35,$1F,$E1,$AC,$5E,$CF,$9B,$E0,$A4,$EB,$B3,$FD,$35,$5F
	.dc.b	$C7,$23,$87,$C1,$7B,$3E,$0A,$46,$F3,$FF,$6D,$57,$F8,$FE,$30,$7F
	.dc.b	$B8,$1F,$F1,$F3,$8E,$0F,$87,$EC,$7C,$1F,$63,$E1,$36,$F5,$6F,$7F
	.dc.b	$E3,$1F,$1F,$E0,$2C,$EB,$DF,$F9,$47,$CB,$F8,$0B,$1C,$BD,$AF,$A7
	.dc.b	$ED,$4D,$70,$76,$F8,$0F,$AF,$F0,$12,$30,$78,$BD,$FF,$B6,$47,$CB
	.dc.b	$F0,$1F,$E9,$9F,$87,$FF,$0E,$F1,$1F,$F0,$EB,$7D,$AF,$F5,$C4,$60
	.dc.b	$FE,$43,$A7,$FC,$01,$6D,$AB,$97,$C3,$F0,$9F,$C3,$4F,$E1,$FE,$06
	.dc.b	$DF,$C6,$7A,$FA,$FE,$3F,$5E,$57,$CF,$B4,$F9,$FD,$F9,$39,$7D,$5F
	.dc.b	$01,$EB,$F8,$0F,$67,$C0,$4A,$E5,$1E,$FF,$5B,$D6,$7B,$3D,$7B,$AF
	.dc.b	$7F,$7F,$DF,$EF,$FC,$06,$C7,$8F,$DF,$F8,$7E,$FF,$C4,$6C,$7E,$17
	.dc.b	$F3,$5F,$F5,$B6,$7F,$E7,$06,$F8,$BF,$EE,$D6,$CF,$C9,$F6,$FE,$97
	.dc.b	$BF,$F4,$FE,$02,$61,$F0,$63,$EB,$7C,$07,$D7,$F8,$09,$3F,$16,$3F
	.dc.b	$DA,$58,$BD,$9F,$F7,$73,$72,$F2,$7B,$58,$9C,$FC,$3E,$BE,$06,$D7
	.dc.b	$A5,$E3,$F1,$5C,$BE,$16,$CF,$B5,$EB,$ED,$78,$3D,$A9,$AF,$C1,$8F
	.dc.b	$1F,$C1,$48,$F8,$8F,$9B,$E0,$BC,$FF,$03,$27,$E4,$3E,$AF,$82,$F5
	.dc.b	$FB,$7E,$A3,$D7,$ED,$DD,$3E,$77,$C1,$7B,$BE,$0B,$DE,$D8,$FA,$5F
	.dc.b	$05,$F0,$BE,$0B,$E1,$B6,$3F,$75,$F9,$B7,$FA,$F5,$BF,$10,$73,$EE
	.dc.b	$DD,$BF,$D5,$35,$FE,$BF,$9E,$3F,$09,$9F,$03,$5E,$82,$BF,$94,$1F
	.dc.b	$C9,$CF,$E5,$5D,$9F,$C9,$70,$FC,$26,$1F,$5B,$E0,$FE,$BB,$CC,$FE
	.dc.b	$3F,$7C,$FD,$AC,$57,$EB,$70,$FF,$62,$D7,$AB,$DA,$CD,$EE,$3D,$BF
	.dc.b	$07,$8A,$37,$E0,$F0,$E7,$E0,$F5,$F0,$76,$FE,$2B,$C7,$DA,$FE,$63
	.dc.b	$9F,$C2,$6E,$FF,$03,$24,$62,$FD,$30,$F6,$B1,$7E,$9F,$82,$BA,$FC
	.dc.b	$76,$D7,$FB,$32,$7E,$AF,$C1,$E3,$E0,$A4,$7D,$87,$7F,$05,$C2,$38
	.dc.b	$FB,$91,$FB,$A7,$4F,$82,$8F,$E0,$1E,$FF,$05,$E0,$F8,$29,$3F,$84
	.dc.b	$7C,$7F,$05,$E6,$FC,$DF,$FD,$94,$7F,$B2,$38,$BE,$FB,$8F,$73,$E8
	.dc.b	$BF,$CA,$89,$F0,$F8,$F8,$4C,$FC,$16,$BE,$12,$F3,$7F,$4B,$BF,$F2
	.dc.b	$4E,$3D,$3D,$AE,$7E,$DF,$46,$D6,$FE,$0F,$BF,$C2,$78,$1B,$1C,$FE
	.dc.b	$13,$C7,$F0,$9E,$46,$C7,$7F,$6F,$D8,$4E,$3F,$17,$B9,$EF,$F8,$3F
	.dc.b	$81,$F0,$72,$7C,$D1,$F0,$FE,$0A,$47,$A5,$F8,$DF,$07,$F7,$3C,$E7
	.dc.b	$B9,$7D,$C9,$7F,$DA,$67,$2D,$F8,$7F,$9D,$F0,$92,$30,$7C,$5F,$83
	.dc.b	$FA,$5F,$09,$23,$07,$3F,$73,$C1,$EE,$48,$C1,$DF,$E0,$FC,$9F,$09
	.dc.b	$23,$07,$8F,$E1,$3D,$04,$7C,$DF,$09,$EB,$F6,$FF,$5B,$B1,$BE,$E6
	.dc.b	$1C,$FC,$1C,$E9,$F7,$7C,$25,$C5,$E0,$FF,$06,$C7,$FC,$89,$5F,$0F
	.dc.b	$E0,$FA,$FC,$27,$67,$BC,$7C,$BF,$76,$61,$F2,$63,$E7,$7C,$27,$CF
	.dc.b	$F8,$49,$3F,$36,$3E,$97,$C1,$C8,$FA,$0F,$D5,$F8,$4F,$D8,$FE,$71
	.dc.b	$FE,$C4,$C3,$FC,$38,$DA,$F3,$C7,$F1,$28,$7B,$BF,$C3,$9D,$6D,$FC
	.dc.b	$81,$FE,$55,$5C,$0F,$F0,$13,$88,$D6,$DC,$7E,$17,$97,$C2,$CC,$7C
	.dc.b	$3F,$15,$71,$E5,$EF,$32,$FF,$15,$47,$9B,$DB,$F4,$9E,$AF,$6F,$D7
	.dc.b	$F0,$B2,$7D,$6F,$BB,$E1,$7D,$FF,$09,$F0,$3E,$16,$57,$C0,$8F,$87
	.dc.b	$EE,$7C,$4F,$72,$4F,$C4,$8F,$8F,$F0,$9F,$23,$E1,$24,$FC,$88,$F9
	.dc.b	$7F,$09,$23,$E5,$BF,$3B,$E0,$FF,$8F,$FE,$73,$FE,$26,$E1,$3F,$C0
	.dc.b	$38,$FF,$9F,$B9,$31,$7F,$28,$FA,$07,$D1,$F8,$4B,$3F,$63,$E1,$3E
	.dc.b	$A7,$C2,$7D,$56,$C7,$F9,$5E,$DE,$CF,$8B,$DA,$98,$E6,$3C,$BF,$11
	.dc.b	$23,$83,$E8,$F8,$5F,$4F,$C2,$49,$E8,$FB,$3E,$17,$DB,$ED,$7F,$52
	.dc.b	$C9,$ED,$58,$C3,$9F,$81,$9A,$F9,$7E,$0A,$E5,$F8,$0C,$3F,$C7,$F3
	.dc.b	$E9,$F8,$2E,$9F,$05,$D5,$ED,$ED,$49,$F7,$7B,$57,$0E,$5F,$05,$E3
	.dc.b	$3C,$9F,$05,$6B,$E2,$7C,$0F,$9C,$F4,$7C,$15,$8F,$F3,$F7,$E7,$4F
	.dc.b	$6F,$FA,$BA,$F6,$E9,$88,$CF,$C1,$CC,$F8,$BC,$3E,$0F,$8F,$C1,$FE
	.dc.b	$0B,$E0,$E5,$76,$8E,$BE,$DF,$F7,$AC,$9E,$DD,$8C,$39,$F8,$39,$AF
	.dc.b	$9B,$E0,$EE,$3F,$A8,$C3,$FC,$80,$DF,$83,$E9,$F0,$7D,$5E,$DE,$E7
	.dc.b	$B0,$93,$F0,$3D,$CF,$7F,$C1,$48,$F8,$6F,$C2,$F8,$3F,$87,$F0,$72
	.dc.b	$7E,$3B,$F1,$BE,$0F,$FB,$EF,$27,$FB,$24,$FF,$64,$17,$DB,$F7,$6F
	.dc.b	$0F,$9F,$F0,$77,$0F,$57,$C1,$7C,$33,$E2,$7C,$1D,$AF,$F5,$96,$1F
	.dc.b	$E1,$C6,$FB,$1F,$E2,$28,$FF,$62,$C6,$0D,$FC,$46,$4D,$7C,$25,$8E
	.dc.b	$7F,$05,$C4,$E5,$F0,$76,$3B,$FB,$1E,$CF,$62,$6B,$83,$C7,$F0,$3F
	.dc.b	$03,$E0,$64,$60,$F3,$FC,$0F,$C4,$F8,$19,$18,$3D,$7E,$CF,$C8,$F6
	.dc.b	$A4,$60,$F7,$FB,$FF,$33,$E0,$24,$60,$F8,$7E,$FF,$D0,$F8,$09,$18
	.dc.b	$3F,$DB,$1F,$9D,$FF,$AF,$89,$ED,$7F,$92,$1C,$7C,$45,$F8,$8C,$37
	.dc.b	$8F,$E7,$DF,$CD,$F8,$2F,$9C,$D8,$EB,$EC,$7F,$8F,$30,$7B,$14,$C4
	.dc.b	$67,$E0,$65,$F9,$1E,$1F,$05,$C7,$E0,$B9,$7C,$14,$AF,$4C,$75,$F6
	.dc.b	$3D,$07,$A7,$D8,$BA,$7B,$BE,$07,$D9,$F0,$3E,$D6,$C7,$C2,$F8,$1F
	.dc.b	$81,$F0,$3F,$05,$B1,$FE,$FD,$FC,$F1,$EC,$FF,$9B,$1B,$1F,$2B,$DA
	.dc.b	$C4,$FC,$9F,$81,$F9,$5F,$03,$6B,$E7,$BF,$37,$E0,$7E,$77,$C0,$D9
	.dc.b	$FA,$6F,$5F,$62,$47,$87,$D9,$BC,$7F,$D8,$B4,$FE,$25,$3E,$5F,$80
	.dc.b	$F2,$7C,$0F,$95,$F3,$7A,$1E,$7F,$81,$B5,$EB,$7D,$5E,$C7,$F9,$C3
	.dc.b	$03,$4C,$7B,$19,$F8,$19,$9F,$C3,$78,$7C,$0F,$1F,$81,$93,$F1,$DE
	.dc.b	$9F,$03,$D7,$DA,$F9,$07,$C9,$F6,$AE,$BF,$37,$E0,$3E,$67,$C0,$7C
	.dc.b	$D6,$C7,$D1,$F8,$0F,$A1,$F0,$1F,$45,$B0,$CF,$FC,$0B,$F9,$EF,$F8
	.dc.b	$36,$3F,$D4,$ED,$87,$C2,$A3,$FD,$71,$1C,$3D,$DF,$F5,$3C,$3E,$22
	.dc.b	$98,$F8,$3B,$87,$E0,$CF,$C1,$1F,$83,$F8,$3A,$F8,$FE,$27,$FD,$83
	.dc.b	$7F,$EB,$4C,$7C,$BE,$DF,$6F,$73,$BB,$E0,$F8,$49,$87,$13,$D3,$F1
	.dc.b	$18,$BF,$6F,$C2,$F9,$BE,$17,$CE,$DA,$E1,$3E,$EC,$BF,$D6,$31,$8E
	.dc.b	$0F,$81,$ED,$FB,$BD,$CF,$7B,$F0,$3E,$13,$E0,$FC,$25,$BE,$18,$23
	.dc.b	$E2,$FC,$27,$C6,$F8,$4B,$5A,$7E,$4E,$7F,$2B,$DD,$B3,$C9,$F9,$BF
	.dc.b	$09,$23,$AF,$C2,$5A,$FB,$3F,$09,$F4,$8F,$A7,$F0,$96,$BD,$5F,$08
	.dc.b	$F7,$3F,$DB,$AD,$8C,$FB,$B8,$9C,$FC,$44,$9E,$35,$71,$FF,$25,$BF
	.dc.b	$E3,$12,$7C,$2F,$4F,$85,$EB,$ED,$F6,$F7,$2E,$1E,$27,$C3,$F0,$72
	.dc.b	$3C,$BF,$09,$72,$EB,$F0,$9E,$6F,$84,$F3,$B6,$BD,$7E,$E7,$A8,$F5
	.dc.b	$FB,$96,$3D,$FF,$09,$EE,$F8,$4F,$7B,$63,$E1,$FC,$1F,$C2,$F8,$4F
	.dc.b	$86,$D8,$FC,$07,$E7,$DF,$EB,$06,$FB,$87,$3E,$ED,$DB,$FC,$DB,$4A
	.dc.b	$C7,$99,$9F,$19,$3F,$C8,$2B,$7F,$09,$F4,$BE,$17,$79,$7E,$CE,$F7
	.dc.b	$3F,$5B,$E1,$7E,$BF,$C2,$DA,$FB,$6F,$0C,$3F,$6B,$8B,$DC,$FD,$E4
	.dc.b	$79,$3D,$CB,$A7,$8E,$91,$8C,$3C,$11,$C6,$BE,$0F,$C2,$E5,$EA,$DD
	.dc.b	$3D,$6D,$97,$C2,$DF,$84,$97,$F2,$9D,$7C,$2E,$FD,$BE,$1E,$E4,$AF
	.dc.b	$9F,$1C,$FE,$12,$47,$D2,$7B,$7C,$27,$7F,$84,$93,$F5,$9F,$17,$C2
	.dc.b	$78,$F7,$F8,$9B,$FD,$0F,$76,$EF,$F6,$67,$EE,$C6,$3E,$13,$3F,$09
	.dc.b	$67,$F0,$2F,$0F,$83,$E3,$F0,$96,$7F,$0A,$F6,$FC,$FD,$EE,$FC,$56
	.dc.b	$ED,$FC,$25,$AF,$FA,$C9,$FF,$60,$B9,$F8,$5E,$04,$78,$7C,$47,$3F
	.dc.b	$86,$E9,$F0,$D2,$7A,$47,$7F,$73,$C1,$EE,$49,$F0,$47,$8F,$E1,$A4
	.dc.b	$78,$DF,$37,$C3,$79,$FE,$1A,$4F,$9D,$F5,$7C,$37,$AF,$DD,$F7,$1E
	.dc.b	$FF,$76,$F1,$7B,$7E,$17,$E1,$1F,$0F,$E1,$6C,$7C,$1F,$85,$F8,$C7
	.dc.b	$C7,$F8,$4B,$1F,$CB,$7F,$40,$7B,$BF,$8D,$5E,$F5,$31,$1F,$37,$68
	.dc.b	$7C,$D3,$E8,$7C,$2F,$D1,$F8,$5F,$A5,$F0,$B2,$7E,$94,$75,$F7,$24
	.dc.b	$78,$5F,$07,$BB,$E1,$F8,$49,$3E,$57,$C9,$F0,$BE,$5F,$85,$F3,$7C
	.dc.b	$2C,$AF,$54,$7A,$7D,$CF,$C8,$EE,$6F,$BB,$87,$3F,$0B,$38,$BE,$17
	.dc.b	$C2,$DC,$7F,$DE,$F8,$7F,$AE,$1B,$F0,$BD,$3E,$17,$AB,$DB,$DD,$F9
	.dc.b	$1E,$EC,$AF,$99,$1F,$2F,$E1,$7E,$67,$C2,$C9,$FA,$11,$F3,$FE,$12
	.dc.b	$47,$D3,$7E,$97,$C2,$FE,$FF,$CB,$FF,$5C,$4F,$F6,$37,$FD,$70,$DF
	.dc.b	$6F,$5E,$E6,$FE,$17,$87,$C2,$CA,$E1,$1C,$FE,$13,$EF,$7C,$2C,$9E
	.dc.b	$91,$DF,$DC,$F0,$7B,$92,$7C,$11,$E3,$F8,$59,$1E,$37,$CD,$F0,$9E
	.dc.b	$7F,$84,$F4,$7C,$24,$AF,$44,$7A,$FD,$BF,$67,$B7,$27,$D9,$1E,$FF
	.dc.b	$83,$91,$EF,$7E,$17,$C2,$7C,$3F,$84,$F8,$9F,$09,$2B,$E2,$47,$F2
	.dc.b	$5F,$D0,$7F,$C4,$51,$FE,$09,$5E,$EE,$4D,$7B,$BF,$37,$C3,$F9,$DF
	.dc.b	$07,$2F,$E7,$47,$D1,$F8,$39,$1F,$45,$FA,$9F,$07,$DF,$DC,$FE,$3F
	.dc.b	$97,$FA,$E6,$9F,$0D,$FD,$46,$3F,$D7,$31,$8C,$3F,$A7,$46,$E7,$5F
	.dc.b	$09,$CB,$E2,$39,$B6,$BD,$5E,$07,$6F,$84,$EE,$D8,$F7,$7B,$7F,$07
	.dc.b	$DB,$99,$E0,$F8,$5F,$07,$F1,$7E,$12,$46,$0F,$8D,$F0,$9F,$24,$8F
	.dc.b	$C9,$F8,$4F,$99,$ED,$FB,$BD,$CF,$7B,$73,$F9,$DF,$07,$F0,$8F,$87
	.dc.b	$F0,$76,$3E,$97,$C1,$FC,$6F,$83,$F8,$ED,$8F,$F1,$F7,$E8,$4F,$73
	.dc.b	$FA,$F2,$F7,$69,$88,$CF,$C1,$EB,$E1,$26,$FC,$A3,$F0,$4F,$E0,$FE
	.dc.b	$12,$3F,$83,$3A,$FB,$5F,$D5,$B1,$FD,$F7,$5F,$F8,$1D,$FF,$5F,$E7
	.dc.b	$E4,$9D,$7C,$26,$FE,$12,$CF,$A1,$E5,$F0,$9C,$FE,$12,$CF,$B1,$ED
	.dc.b	$ED,$77,$F6,$AC,$FC,$07,$C5,$F0,$72,$3E,$1F,$C2,$5D,$BF,$10,$7F
	.dc.b	$A0,$B5,$7C,$27,$A7,$E1,$3D,$4D,$BF,$F3,$5D,$3F,$19,$DF,$C4,$FF
	.dc.b	$A3,$FC,$B7,$B3,$FE,$16,$9C,$D2,$72,$FF,$A5,$63,$31,$88,$E9,$87
	.dc.b	$FA,$51,$E1,$5C,$E3,$13,$BA,$E3,$39,$BF,$13,$6F,$89,$DE,$F8,$C6
	.dc.b	$6F,$94,$62,$75,$5C,$9A,$7F,$8C,$0A,$5B,$F6,$B6,$7D,$0D,$9E,$0F
	.dc.b	$03,$C8,$64,$CC,$F2,$72,$43,$87,$D2,$08,$00

;--------------------------------------------------------------------------------
;	FCOSH.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fcosh_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FCOSH.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FCOSH.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fcosh_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fcosh.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FCOSH.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fcosh.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FCOSH.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fcosh_data_compressed::
	.dc.b	$00,$02,$84,$2C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CF,$ED,$1F,$8A,$7F,$87,$A7,$FC,$D0,$5F,$8A,$E1
	.dc.b	$F0,$FC,$7D,$FE,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$1F
	.dc.b	$07,$9F,$DB,$F4,$7B,$7F,$DF,$7E,$6F,$FA,$B0,$F8,$8C,$7C,$46,$7E
	.dc.b	$23,$5F,$11,$BF,$88,$E1,$F1,$1C,$7E,$17,$97,$C0,$F3,$F8,$1E,$9F
	.dc.b	$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03
	.dc.b	$E3,$F8,$1F,$27,$C0,$F9,$7E,$07,$CD,$F0,$3E,$7F,$81,$FF,$0A,$7C
	.dc.b	$E7,$E1,$BE,$13,$FC,$40,$BE,$07,$3E,$66,$BE,$07,$7F,$03,$C3,$E0
	.dc.b	$78,$FC,$0F,$2F,$81,$E7,$F0,$3F,$63,$D2,$EB,$F0,$3D,$BE,$07,$BF
	.dc.b	$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93,$E0,$7C,$BF
	.dc.b	$03,$E6,$F8,$1F,$3F,$C0,$FA,$3E,$07,$FC,$59,$F3,$BF,$88,$E9,$FF
	.dc.b	$1C,$7C,$07,$F0,$A5,$1F,$EF,$63,$FD,$60,$7F,$9A,$0F,$EB,$47,$F1
	.dc.b	$23,$81,$FC,$41,$FF,$21,$2F,$7B,$BF,$BD,$E0,$F7,$BC,$3E,$F7,$8B
	.dc.b	$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD,$E8,$F7,$BF
	.dc.b	$3B,$EF,$FE,$7B,$5F,$CF,$EB,$FA,$0D,$7F,$43,$AF,$BB,$DE,$F7,$FB
	.dc.b	$DF,$E4,$BF,$6F,$FC,$48,$7E,$3F,$8B,$FC,$B0,$BD,$CD,$3B,$F7,$38
	.dc.b	$7B,$9C,$7D,$CE,$5E,$E7,$3F,$73,$A7,$B9,$FB,$0E,$8E,$DE,$E7,$7F
	.dc.b	$73,$C1,$EE,$7E,$D3,$AF,$F6,$BD,$7F,$B6,$EB,$FD,$BF,$5F,$EE,$3A
	.dc.b	$FF,$73,$D7,$FB,$AE,$6F,$DD,$F3,$7F,$98,$FF,$15,$FF,$1E,$1E,$9E
	.dc.b	$73,$E9,$7F,$CE,$4F,$FB,$38,$FF,$44,$1F,$DE,$0F,$E6,$07,$F2,$F7
	.dc.b	$8F,$0F,$5E,$1E,$DC,$1D,$F8,$3E,$47,$BB,$F2,$7D,$DF,$95,$EE,$FC
	.dc.b	$BF,$77,$E6,$7B,$BF,$37,$DD,$F9,$DE,$EF,$CF,$F7,$7E,$87,$BB,$FE
	.dc.b	$7B,$F9,$FF,$D4,$70,$1F,$D4,$3F,$E8,$95,$EE,$6F,$DC,$E1,$EE,$71
	.dc.b	$F7,$39,$7B,$9C,$FD,$CE,$9E,$E7,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B
	.dc.b	$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$F3,$7C,$9F,$9D,$D9,$F3
	.dc.b	$FB,$3E,$87,$67,$D1,$EC,$FA,$5D,$9F,$4F,$B3,$EA,$76,$7D,$5E,$CF
	.dc.b	$AD,$D9,$FE,$99,$F4,$FD,$E7,$BB,$B4,$FB,$9F,$F5,$42,$EB,$E1,$D7
	.dc.b	$C7,$AF,$97,$5F,$3E,$BE,$9D,$7D,$7A,$FE,$67,$B7,$F3,$7D,$BF,$9D
	.dc.b	$ED,$F8,$7B,$BE,$87,$B7,$FE,$AB,$FA,$0F,$E0,$1C,$05,$F0,$99,$F8
	.dc.b	$AD,$7C,$AE,$FE,$17,$87,$C0,$F1,$F8,$6E,$5F,$01,$CF,$E0,$3A,$74
	.dc.b	$75,$F8,$0F,$81,$E4,$FC,$1F,$27,$E1,$79,$3F,$0F,$C9,$F8,$9E,$4F
	.dc.b	$8F,$E0,$3C,$9E,$D7,$97,$DA,$F3,$7B,$5E,$7F,$6B,$D1,$ED,$7A,$7D
	.dc.b	$AF,$57,$B5,$EB,$F6,$BD,$9E,$D7,$FB,$1F,$E8,$7F,$D4,$07,$E9,$4F
	.dc.b	$D2,$76,$97,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1F,$67,$D9
	.dc.b	$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BB,$E1,$F7,$7C,$5E,$EF
	.dc.b	$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE,$F9,$FD,$DF,$47,$BB,$F8,$7F
	.dc.b	$57,$D5,$EE,$FF,$B8,$7C,$FF,$D3,$9F,$89,$3F,$11,$D2,$7F,$10,$FF
	.dc.b	$BC,$5F,$F7,$61,$FE,$A5,$3F,$CB,$AE,$FD,$9E,$9E,$CF,$5F,$67,$B7
	.dc.b	$AF,$DF,$CC,$F0,$79,$9E,$1F,$33,$C5,$E6,$78,$FC,$CF,$27,$67,$97
	.dc.b	$7F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$FB,$E7,$E8,$BF,$1E,$7E,$3B,$84
	.dc.b	$17,$B1,$BE,$3E,$1E,$C7,$1F,$63,$97,$B1,$CF,$D8,$E9,$EC,$75,$F6
	.dc.b	$3B,$7B,$1F,$AF,$DF,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E
	.dc.b	$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7A,$7D,$8F,$3F,$3B,$E8
	.dc.b	$FF,$DD,$87,$F3,$13,$F1,$A7,$E3,$1F,$6A,$F5,$F8,$BF,$EA,$83,$FC
	.dc.b	$C1,$C2,$EB,$CF,$AB,$AF,$57,$7E,$AF,$0F,$57,$8F,$AB,$CB,$D5,$E7
	.dc.b	$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E,$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F
	.dc.b	$57,$C7,$EA,$FB,$3E,$93,$F9,$99,$FC,$54,$FC,$69,$F8,$CE,$62,$F7
	.dc.b	$F8,$7B,$FC,$7D,$FE,$5E,$CF,$3F,$67,$A7,$B3,$D7,$D9,$ED,$EC,$F7
	.dc.b	$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9,$EC,$F9,$7C,$FF
	.dc.b	$37,$85,$E7,$F6,$7E,$9F,$C0,$7F,$1A,$3F,$19,$CB,$FC,$A7,$D9,$CF
	.dc.b	$B3,$AF,$67,$7E,$CF,$0F,$67,$8F,$B3,$CB,$D9,$E7,$E6,$74,$F6,$7A
	.dc.b	$FB,$3D,$BD,$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F,$67,$C7,$EC,$F9
	.dc.b	$3D,$9F,$2F,$B3,$E6,$F6,$7F,$13,$F4,$BF,$8A,$E6,$3F,$DE,$4F,$E2
	.dc.b	$A7,$E2,$5C,$FB,$5C,$3D,$DE,$3E,$EF,$2F,$6F,$9F,$B7,$D3,$DB,$EB
	.dc.b	$ED,$F6,$F6,$FB,$FB,$7E,$0F,$6F,$C3,$ED,$F8,$BD,$BF,$1F,$B7,$E4
	.dc.b	$F6,$FC,$BE,$DF,$9B,$DB,$F3,$FB,$7E,$8F,$6F,$F4,$3F,$4D,$FD,$E0
	.dc.b	$FE,$B4,$7F,$12,$3F,$0C,$7E,$17,$97,$F7,$AB,$DA,$E3,$ED,$72,$F6
	.dc.b	$B9,$FB,$5D,$3D,$AE,$BE,$D7,$6F,$6B,$BF,$B5,$E0,$F6,$BC,$3E,$D7
	.dc.b	$8B,$DA,$F1,$FB,$5E,$4F,$6B,$CB,$ED,$79,$BD,$AF,$3F,$B5,$FB,$8F
	.dc.b	$A7,$FE,$BE,$7E,$18,$FE,$1F,$C9,$FC,$49,$FE,$54,$DF,$63,$87,$B1
	.dc.b	$C7,$D8,$E5,$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1
	.dc.b	$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E
	.dc.b	$7F,$63,$F8,$7F,$AF,$F8,$73,$FC,$7C,$7F,$5A,$3F,$0C,$7F,$0D,$E4
	.dc.b	$FE,$56,$BD,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6,$F5
	.dc.b	$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC,$BE
	.dc.b	$BF,$9B,$D7,$F3,$FA,$FE,$8F,$5F,$F9,$3F,$D4,$7F,$8F,$8F,$F1,$01
	.dc.b	$FD,$60,$FE,$18,$7E,$1F,$94,$FE,$1D,$FE,$8A,$FF,$58,$1E,$AF,$4F
	.dc.b	$57,$AF,$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E
	.dc.b	$AF,$93,$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FA,$3D,$5F,$E7,$FF,$53
	.dc.b	$FE,$26,$3F,$86,$9F,$88,$E2,$3F,$88,$6F,$AB,$C3,$D5,$E3,$EA,$F2
	.dc.b	$F5,$79,$FA,$BD,$3D,$5E,$BE,$AF,$6F,$57,$BF,$AB,$E0,$F5,$7C,$3E
	.dc.b	$AF,$8B,$D5,$F1,$FA,$BE,$4F,$57,$CB,$EA,$F9,$BD,$5F,$3F,$AB,$FD
	.dc.b	$5B,$EA,$BF,$87,$1F,$E8,$53,$FC,$42,$7F,$0D,$3F,$11,$C8,$7F,$10
	.dc.b	$FF,$74,$5E,$A7,$3F,$53,$A7,$A9,$D7,$D4,$ED,$EA,$77,$F5,$3C,$1E
	.dc.b	$A7,$87,$D4,$F1,$7A,$9E,$3F,$53,$C9,$EA,$79,$7D,$4F,$37,$A9,$E7
	.dc.b	$F5,$3D,$1E,$A7,$A7,$D4,$FE,$DF,$E2,$7F,$A1,$8F,$F3,$41,$FE,$20
	.dc.b	$3F,$AA,$1F,$89,$3F,$11,$C8,$7F,$C3,$E7,$E2,$1F,$F0,$7A,$F4,$BA
	.dc.b	$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA,$5E,$2F,$4B,$C7,$E9,$79
	.dc.b	$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$A3,$D2,$FF,$03,$FD,$5F,$F9
	.dc.b	$B8,$FE,$AA,$7E,$24,$FC,$41,$E7,$E1,$FF,$11,$AF,$4B,$8F,$A5,$CB
	.dc.b	$D2,$E7,$E9,$74,$F4,$BA,$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA
	.dc.b	$5E,$2F,$4B,$C7,$E9,$79,$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$F8
	.dc.b	$7B,$EB,$3F,$AB,$1F,$EC,$D3,$FC,$D2,$7F,$55,$3F,$12,$7E,$23,$90
	.dc.b	$FE,$21,$FF,$1E,$BF,$EC,$D1,$E8,$75,$F4,$3B,$7A,$1D,$FD,$0F,$07
	.dc.b	$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F,$43,$CD,$E8,$79
	.dc.b	$FD,$0F,$47,$A1,$E9,$F4,$3F,$C6,$FF,$5B,$FE,$CE,$3F,$D6,$07,$F9
	.dc.b	$A0,$FF,$0E,$1F,$C4,$4F,$C4,$9F,$E6,$63,$F1,$1C,$47,$F1,$0F,$F9
	.dc.b	$61,$79,$FD,$BC,$FE,$FE,$7F,$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7
	.dc.b	$F9,$3C,$FF,$2F,$9F,$E6,$F3,$FC,$FE,$7F,$A3,$CF,$FF,$27,$FC,$17
	.dc.b	$FA,$D8,$FF,$0E,$9F,$C4,$4F,$E2,$07,$E2,$78,$3F,$C6,$9E,$7F,$0F
	.dc.b	$3F,$8F,$9F,$CB,$CF,$E7,$E7,$F4,$F3,$FA,$F9,$FD,$BC,$FE,$FE,$7F
	.dc.b	$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7,$F9,$3C,$FF,$2F,$9F,$E6,$F3
	.dc.b	$FF,$CC,$BF,$5D,$FE,$1D,$1F,$27,$8F,$8F,$CF,$C4,$EB,$DE,$DF,$BD
	.dc.b	$C3,$DA,$E3,$ED,$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$FE,$7C,$FA,$F7,$CB,$DF,$94,$C7,$C8,$67,$E2,$B5
	.dc.b	$E7,$6F,$E1,$B8,$7C,$37,$1F,$84,$E5,$F0,$3C,$FD,$EE,$9E,$F7,$5F
	.dc.b	$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27
	.dc.b	$BD,$E5,$F7,$BC,$DE,$F7,$FA,$4B,$EC,$3F,$DF,$27,$F9,$80,$FE,$A8
	.dc.b	$7F,$53,$3F,$08,$7F,$0B,$F7,$FF,$D5,$EB,$DE,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$FD,$53,$EC,$FF,$98
	.dc.b	$4F,$C2,$F1,$FF,$B0,$9F,$F7,$A9,$FE,$61,$3F,$AA,$9F,$C3,$4F,$E1
	.dc.b	$8F,$F8,$77,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8E,$FE,$C7,$83
	.dc.b	$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79,$3D,$8F,$2F,$B1,$E6,$F6,$3C
	.dc.b	$FE,$C7,$A3,$D8,$FF,$60,$FD,$8F,$E1,$B8,$BF,$DB,$2B,$D7,$CF,$AF
	.dc.b	$AF,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D
	.dc.b	$7F,$2F,$AF,$E6,$F5,$FC,$FE,$BF,$FB,$63,$EC,$9F,$35,$7E,$4F,$1F
	.dc.b	$1F,$9F,$8F,$D7,$C4,$EF,$E1,$B8,$7B,$BC,$7D,$DE,$5E,$EF,$3F,$77
	.dc.b	$A7,$BB,$D7,$DD,$ED,$EE,$F7,$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE
	.dc.b	$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$FE,$EE,$FB,$3F,$F5,$31,$F8
	.dc.b	$6F,$38,$BE,$17,$5F,$0B,$BF,$85,$E1,$F0,$3C,$7E,$17,$97,$C1,$73
	.dc.b	$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F,$82,$F0,$7C,$17,$87,$E0
	.dc.b	$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E,$0B,$CD,$F0,$5E,$7F,$82
	.dc.b	$F4,$7C,$13,$9E,$78,$FF,$87,$7F,$BE,$7C,$0E,$5F,$EF,$DF,$03,$67
	.dc.b	$FD,$DC,$7F,$98,$CF,$F0,$D8,$F7,$B5,$EF,$6F,$DE,$E1,$EF,$71,$F7
	.dc.b	$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE
	.dc.b	$1F,$7B,$C5,$EF,$78,$FD,$EF,$37,$DA,$7E,$23,$93,$D8,$FF,$BB,$8F
	.dc.b	$F3,$11,$FE,$61,$3F,$A9,$1F,$C3,$D3,$FC,$3D,$E3,$EC,$74,$F6,$3A
	.dc.b	$FB,$1D,$BD,$8E,$FE,$C7,$83,$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79
	.dc.b	$3D,$8F,$2F,$B1,$E6,$F6,$3C,$FE,$C7,$A3,$D8,$F9,$DF,$6B,$FE,$EF
	.dc.b	$3F,$14,$7E,$27,$87,$EB,$3F,$EE,$F3,$FC,$C6,$7F,$86,$CF,$E1,$E3
	.dc.b	$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD,$3E,$FE,$9F,$83,$D3,$F0,$FA
	.dc.b	$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F,$A7,$E6,$F4,$FC,$FE,$9F,$A3
	.dc.b	$D3,$FC,$37,$DB,$7E,$24,$FE,$1D,$C7,$F8,$D5,$E9,$EB,$E1,$37,$F0
	.dc.b	$9C,$3E,$13,$8F,$AB,$CB,$D5,$E7,$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E
	.dc.b	$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F,$57,$C7,$EA,$F9,$3D,$5F,$2F,$AB
	.dc.b	$E6,$F5,$7C,$FE,$AF,$E7,$78,$7F,$11,$CD,$FF,$00,$7E,$0B,$1F,$05
	.dc.b	$9F,$82,$D7,$C1,$6F,$DA,$E1,$F0,$5C,$7D,$AE,$5E,$D7,$3F,$6B,$A7
	.dc.b	$B5,$D7,$DA,$ED,$ED,$77,$F6,$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F
	.dc.b	$6B,$C9,$ED,$79,$7D,$AF,$37,$B5,$FB,$3F,$B7,$FD,$A7,$39,$FE,$20
	.dc.b	$7E,$1D,$C7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$D1,$EF,$7F,$0B,$FB
	.dc.b	$8F,$E1,$C7,$E1,$B9,$8F,$E1,$9F,$E2,$8B,$DD,$E1,$EE,$F1,$F3,$B9
	.dc.b	$7B,$BC,$FD,$DE,$9E,$EF,$5F,$77,$B7,$BB,$DF,$DD,$F0,$7B,$BE,$1F
	.dc.b	$77,$C5,$EE,$F8,$FD,$DF,$27,$BB,$E5,$F7,$7C,$DE,$EF,$9F,$DD,$F4
	.dc.b	$7B,$BF,$C8,$FE,$E7,$F0,$E3,$DE,$C7,$BD,$9F,$7B,$5E,$F6,$FD,$EE
	.dc.b	$1E,$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$75,$F7,$BB,$7B,$DD,$FD
	.dc.b	$EF,$07,$BD,$E1,$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD
	.dc.b	$EF,$7F,$39,$F6,$BF,$0E,$7F,$55,$3F,$0C,$3D,$DD,$7B,$BB,$F7,$78
	.dc.b	$4F,$1F,$77,$97,$BB,$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7B,$BD,$FD,$DF
	.dc.b	$07,$BB,$E1,$F7,$7C,$5E,$EF,$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE
	.dc.b	$F9,$FD,$DF,$EA,$1F,$75,$FD,$4C,$FC,$31,$F8,$5F,$78,$FE,$15,$FE
	.dc.b	$B4,$BD,$CE,$3E,$E7,$2F,$73,$9F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B
	.dc.b	$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7
	.dc.b	$9B,$DC,$F3,$FB,$9E,$8F,$73,$FB,$47,$DD,$FF,$0C,$3F,$0D,$CB,$FD
	.dc.b	$BD,$7B,$7A,$F6,$F7,$ED,$F0,$F6,$F8,$FB,$7C,$BD,$BE,$7E,$DF,$4F
	.dc.b	$6D,$FC,$03,$0B,$A7,$59,$7D,$DF,$F0,$68,$C3,$56,$A9,$DC,$6E,$31
	.dc.b	$3B,$6C,$66,$76,$FF,$12,$8C,$B6,$32,$FA,$27,$94,$62,$3C,$67,$38
	.dc.b	$C9,$FE,$13,$FB,$0F,$F3,$4D,$31,$FF,$09,$46,$1B,$E8,$75,$B4,$F0
	.dc.b	$C3,$D6,$FF,$9C,$BE,$03,$FB,$E4,$EA,$BF,$66,$F7,$79,$4F,$1C,$71
	.dc.b	$3F,$E2,$59,$FE,$59,$3F,$E7,$C6,$9C,$09,$1F,$77,$0F,$2B,$4E,$67
	.dc.b	$D0,$C3,$CB,$3C,$21,$87,$9A,$FD,$1A,$72,$9E,$18,$F2,$CB,$E9,$D7
	.dc.b	$8F,$E2,$75,$8F,$92,$FE,$6C,$FA,$DF,$80,$78,$CA,$7E,$0B,$3E,$D7
	.dc.b	$D2,$BC,$99,$78,$DF,$8F,$5C,$27,$4F,$68,$FE,$1D,$87,$88,$E7,$9F
	.dc.b	$BA,$74,$78,$71,$FB,$4F,$AE,$34,$FB,$8F,$2E,$1E,$17,$FC,$AC,$D7
	.dc.b	$53,$ED,$3E,$13,$F3,$AF,$E1,$3A,$7E,$76,$1F,$8B,$8F,$95,$8F,$9B
	.dc.b	$2F,$11,$F5,$B0,$EC,$77,$AF,$7D,$73,$3F,$09,$97,$BD,$E2,$7C,$5C
	.dc.b	$BD,$73,$DE,$3E,$46,$1D,$DD,$61,$DF,$3F,$BD,$8F,$7C,$38,$9F,$13
	.dc.b	$2F,$59,$EC,$C7,$C6,$4C,$7E,$C9,$D7,$3F,$89,$3F,$61,$FC,$5E,$5E
	.dc.b	$67,$2F,$A8,$F9,$19,$F1,$9F,$8F,$5D,$27,$84,$7E,$13,$2F,$39,$B8
	.dc.b	$FB,$A2,$3C,$F5,$98,$F9,$07,$E3,$CF,$A2,$74,$38,$E5,$F4,$2B,$86
	.dc.b	$3F,$42,$F9,$D6,$CE,$71,$F3,$71,$FC,$5C,$7C,$C2,$3F,$01,$FC,$5D
	.dc.b	$74,$CB,$E6,$63,$F6,$67,$78,$7C,$07,$CA,$60,$F2,$D7,$C0,$7C,$58
	.dc.b	$79,$6B,$96,$1F,$32,$7C,$B9,$7C,$97,$91,$F0,$B2,$F7,$3C,$B2,$F1
	.dc.b	$BF,$48,$C6,$7E,$2B,$F6,$9D,$71,$FC,$9D,$F2,$AE,$B1,$E3,$79,$18
	.dc.b	$BF,$C9,$E1,$E3,$BE,$51,$FB,$0B,$E5,$18,$C3,$8B,$FA,$CB,$E5,$1E
	.dc.b	$3C,$FF,$AA,$47,$98,$F9,$A7,$E6,$CC,$4F,$12,$47,$EC,$AF,$C7,$3C
	.dc.b	$CF,$C4,$E3,$EC,$CF,$E7,$47,$80,$FE,$2F,$5F,$82,$71,$7F,$12,$3C
	.dc.b	$71,$87,$F4,$04,$7A,$DF,$CE,$79,$E7,$F8,$3C,$7C,$51,$F9,$53,$91
	.dc.b	$FC,$43,$0F,$1E,$1F,$16,$38,$9F,$1B,$0F,$65,$72,$C7,$F2,$91,$89
	.dc.b	$F8,$A4,$AE,$47,$C4,$7F,$64,$78,$0F,$13,$ED,$7D,$35,$F3,$5E,$53
	.dc.b	$D6,$79,$47,$EC,$70,$F6,$3F,$C4,$0F,$26,$3F,$C3,$2F,$95,$7E,$94
	.dc.b	$F1,$DF,$F3,$AA,$F6,$1E,$03,$91,$23,$6F,$C5,$39,$C7,$8C,$E4,$FC
	.dc.b	$63,$D9,$87,$C4,$7F,$A1,$46,$DE,$47,$B0,$F0,$4F,$B2,$3F,$90,$5F
	.dc.b	$A6,$7C,$51,$F8,$09,$F1,$C7,$F6,$AC,$7E,$BC,$F5,$8F,$88,$7B,$0F
	.dc.b	$E1,$35,$F9,$27,$99,$E0,$8F,$C4,$3C,$CF,$CD,$BE,$CA,$F9,$F7,$FC
	.dc.b	$CA,$B9,$47,$D7,$AE,$B8,$7C,$47,$DB,$1E,$6B,$FC,$D3,$E8,$BF,$89
	.dc.b	$3F,$80,$9F,$46,$5F,$B4,$7D,$47,$03,$D9,$3C,$09,$87,$C5,$BF,$B9
	.dc.b	$3E,$A9,$F3,$61,$F0,$63,$E5,$BF,$C6,$B3,$FE,$09,$1E,$0C,$BE,$4E
	.dc.b	$5F,$0C,$FC,$AE,$1F,$03,$1E,$D5,$F4,$49,$3B,$3E,$41,$F6,$CF,$86
	.dc.b	$7E,$0E,$79,$61,$E1,$8F,$CA,$BF,$CF,$72,$F8,$11,$CE,$BE,$03,$D7
	.dc.b	$2F,$81,$97,$E7,$B0,$F7,$CF,$CE,$AF,$8F,$9F,$F3,$47,$E3,$D7,$BA
	.dc.b	$7C,$D8,$76,$7C,$B7,$F8,$FB,$E9,$7E,$C9,$FA,$87,$0C,$FE,$2B,$C2
	.dc.b	$3F,$08,$47,$CD,$87,$77,$47,$94,$86,$0F,$CA,$9F,$16,$3C,$EF,$C4
	.dc.b	$3F,$A4,$E1,$E7,$C7,$ED,$C6,$CF,$AB,$3C,$1F,$CD,$9D,$67,$86,$7F
	.dc.b	$62,$36,$43,$C5,$5F,$31,$FC,$F9,$E6,$9F,$F0,$A6,$7F,$DC,$6F,$E9
	.dc.b	$CF,$0C,$3D,$D7,$E4,$CF,$E9,$CF,$E0,$9F,$D0,$67,$F4,$E3,$E2,$9E
	.dc.b	$D8,$FE,$D1,$8F,$9B,$1E,$D5,$BA,$EC,$6F,$1F,$F0,$93,$FA,$D3,$E5
	.dc.b	$3E,$4C,$3C,$4F,$D9,$7F,$4B,$5F,$8D,$CF,$EC,$C6,$AB,$E1,$57,$E9
	.dc.b	$1F,$ED,$79,$7B,$33,$F9,$11,$EB,$9F,$D6,$46,$1F,$03,$AA,$F9,$53
	.dc.b	$FC,$EF,$2F,$63,$FE,$02,$AD,$E3,$FA,$18,$FA,$F5,$E3,$7B,$1F,$86
	.dc.b	$71,$3C,$23,$CA,$7C,$33,$FC,$2A,$7C,$38,$DE,$5F,$25,$E8,$79,$33
	.dc.b	$FE,$4F,$87,$91,$FF,$0C,$5F,$57,$FC,$1D,$97,$C1,$C7,$E4,$63,$D7
	.dc.b	$2F,$81,$97,$57,$E0,$56,$E7,$F2,$F9,$FC,$E9,$DC,$7F,$31,$CF,$C7
	.dc.b	$5B,$C7,$F0,$37,$BC,$7F,$03,$7B,$C3,$AB,$EF,$3F,$C5,$53,$B7,$E6
	.dc.b	$63,$E3,$AF,$99,$7F,$CC,$F1,$F9,$99,$FB,$F3,$FF,$58,$1E,$F3,$FC
	.dc.b	$81,$87,$D2,$CB,$EF,$CE,$E3,$FC,$BB,$3B,$7F,$C2,$F9,$FA,$63,$F2
	.dc.b	$71,$FC,$7E,$FE,$8B,$E9,$C7,$C7,$3F,$93,$C3,$AE,$7E,$9C,$BE,$8E
	.dc.b	$7E,$9C,$3E,$E9,$E3,$C3,$E8,$BD,$6F,$E5,$BF,$C0,$4F,$D2,$9B,$3E
	.dc.b	$89,$F5,$E3,$E5,$BB,$C7,$E8,$BF,$2D,$FF,$18,$3B,$3E,$59,$E9,$3C
	.dc.b	$77,$F2,$EB,$D3,$1B,$8F,$4B,$B7,$F1,$2F,$58,$FF,$27,$61,$EF,$C7
	.dc.b	$E8,$E3,$EF,$BD,$DF,$A5,$FC,$1B,$D6,$76,$F5,$7E,$BC,$78,$F1,$F8
	.dc.b	$AF,$F1,$0C,$7E,$8D,$7B,$F1,$FC,$4B,$D6,$FE,$BD,$6E,$7E,$5B,$FD
	.dc.b	$52,$7D,$F3,$F5,$EF,$AB,$FC,$01,$F9,$73,$B3,$D2,$78,$E7,$E8,$9F
	.dc.b	$5E,$37,$3F,$D1,$63,$EB,$CE,$CF,$7B,$B3,$E8,$9E,$99,$FC,$4B,$FE
	.dc.b	$28,$3F,$A9,$1F,$14,$FC,$AE,$1F,$2E,$BE,$29,$E9,$8F,$1D,$7B,$EB
	.dc.b	$C7,$1F,$D0,$70,$FC,$1B,$F1,$4F,$E1,$67,$8E,$7E,$EB,$E3,$3E,$29
	.dc.b	$E3,$3F,$A7,$63,$F7,$5F,$19,$F5,$CF,$D3,$9F,$83,$AF,$E7,$19,$FF
	.dc.b	$4C,$8F,$7D,$7C,$BC,$3D,$31,$D7,$0F,$DC,$E1,$D6,$3F,$D9,$73,$FB
	.dc.b	$37,$FD,$5E,$75,$3F,$C2,$17,$D6,$FD,$27,$58,$F1,$9E,$F3,$F8,$59
	.dc.b	$F2,$CE,$B8,$FC,$B7,$71,$F5,$EB,$F1,$27,$E0,$E3,$D2,$FD,$DC,$3E
	.dc.b	$5C,$EE,$7E,$5B,$F4,$5D,$9F,$47,$2F,$CA,$C6,$E3,$C6,$7D,$1A,$FD
	.dc.b	$0B,$F4,$63,$F1,$F1,$F7,$5F,$C4,$9D,$4F,$7B,$E3,$9E,$AE,$CF,$4C
	.dc.b	$FF,$4D,$8F,$D0,$C6,$F2,$FC,$7C,$FF,$88,$0F,$BA,$7C,$5B,$FF,$67
	.dc.b	$3E,$39,$FA,$EE,$E3,$EB,$BB,$AF,$F5,$85,$FD,$18,$FA,$E7,$D1,$9F
	.dc.b	$E2,$2F,$BC,$FC,$49,$D4,$FA,$35,$E9,$7E,$5B,$F4,$67,$DF,$1F,$2E
	.dc.b	$FD,$35,$FE,$00,$8F,$D0,$D6,$CF,$97,$5E,$39,$5F,$F9,$EA,$7B,$57
	.dc.b	$D2,$AF,$99,$1C,$0F,$8D,$1F,$A2,$3E,$04,$FF,$24,$7D,$47,$E2,$8F
	.dc.b	$D1,$1F,$02,$7E,$F5,$7C,$07,$AC,$7C,$07,$C9,$3E,$38,$FD,$13,$D6
	.dc.b	$37,$1D,$4F,$8C,$75,$AF,$4D,$75,$7E,$04,$75,$7D,$33,$F8,$A8,$EB
	.dc.b	$7F,$C7,$9F,$C2,$3F,$C1,$9F,$F2,$AC,$7F,$A5,$F3,$FE,$C1,$1F,$48
	.dc.b	$D9,$F9,$C3,$75,$FE,$53,$9F,$7E,$1F,$D0,$A7,$F0,$97,$FE,$93,$C7
	.dc.b	$FC,$DF,$1F,$BF,$C2,$BF,$E2,$22,$63,$FC,$F3,$0F,$8B,$7F,$EE,$FA
	.dc.b	$DE,$5F,$DD,$EF,$E2,$56,$9F,$F1,$6C,$EB,$3F,$E1,$58,$7E,$6A,$B4
	.dc.b	$7F,$81,$E7,$55,$E6,$3F,$B7,$67,$FE,$45,$9F,$91,$1F,$80,$C7,$E4
	.dc.b	$3E,$6C,$FF,$56,$FE,$4C,$FC,$36,$1F,$93,$AF,$DB,$DF,$D4,$9F,$3B
	.dc.b	$F5,$0F,$D5,$CE,$DF,$DB,$C7,$5B,$FA,$91,$E3,$37,$87,$A2,$7F,$23
	.dc.b	$5E,$D9,$E9,$1F,$AA,$3E,$F9,$FA,$43,$EC,$DF,$EA,$9F,$B2,$7F,$9D
	.dc.b	$8F,$DB,$61,$F6,$5E,$AF,$EF,$27,$AC,$7E,$44,$EB,$3E,$E7,$AC,$7E
	.dc.b	$A5,$F8,$91,$F2,$AB,$DD,$1F,$8B,$3D,$C7,$E2,$E7,$DC,$FA,$4F,$A0
	.dc.b	$7A,$4D,$CF,$A6,$37,$1E,$33,$71,$E3,$77,$1E,$37,$DE,$7F,$74,$7F
	.dc.b	$53,$5E,$F9,$FD,$7D,$7C,$BA,$F4,$D7,$D0,$7F,$D4,$E7,$E5,$F2,$FE
	.dc.b	$A1,$1D,$4F,$1C,$FF,$0E,$C7,$F3,$15,$C2,$BF,$31,$3D,$63,$C6,$FF
	.dc.b	$4A,$AF,$1D,$7F,$68,$CB,$F4,$B1,$FD,$86,$7A,$57,$E7,$8E,$93,$F9
	.dc.b	$EC,$7E,$B4,$7A,$0F,$95,$3F,$41,$FE,$3D,$5F,$93,$8F,$13,$E8,$9F
	.dc.b	$E2,$53,$F2,$A3,$F2,$78,$7B,$A3,$F8,$61,$D4,$FA,$15,$D4,$FE,$24
	.dc.b	$FE,$08,$D9,$EE,$AF,$EA,$AF,$E2,$2B,$AD,$FB,$B3,$FE,$E5,$3D,$33
	.dc.b	$FC,$F6,$7F,$53,$0F,$6C,$F3,$CF,$FA,$0D,$6B,$1F,$65,$69,$F6,$4F
	.dc.b	$D2,$AF,$F5,$B6,$3F,$1E,$3F,$8A,$47,$DC,$CB,$FD,$51,$48,$FF,$18
	.dc.b	$9F,$A2,$9D,$AC,$7F,$93,$3F,$E1,$37,$FD,$C7,$9F,$C7,$8F,$DF,$9F
	.dc.b	$1E,$FE,$CC,$FF,$11,$8F,$B2,$7F,$11,$57,$FE,$49,$3D,$31,$F7,$CF
	.dc.b	$49,$B7,$59,$FF,$2A,$AF,$4C,$FA,$2B,$E4,$61,$E8,$7C,$F1,$FB,$E8
	.dc.b	$FC,$73,$F9,$CB,$FC,$76,$5F,$82,$BF,$B8,$FE,$96,$3E,$6D,$73,$7F
	.dc.b	$A0,$1F,$98,$AF,$F1,$9E,$5F,$6D,$FE,$C7,$7F,$02,$7F,$DE,$D9,$FD
	.dc.b	$79,$FC,$31,$F1,$8F,$B8,$7F,$2A,$74,$FE,$E8,$FC,$5C,$7C,$4C,$7E
	.dc.b	$54,$7C,$43,$EB,$1F,$CE,$B2,$F8,$95,$FD,$6A,$3F,$42,$78,$2F,$F4
	.dc.b	$33,$A8,$FB,$57,$F8,$17,$F7,$D8,$7A,$DF,$EE,$35,$EB,$7F,$AD,$1F
	.dc.b	$68,$F0,$E1,$F6,$9F,$C5,$57,$CC,$C7,$F2,$53,$FE,$F4,$BF,$C9,$57
	.dc.b	$9C,$FC,$96,$3E,$7C,$3F,$57,$1F,$04,$F8,$D8,$7C,$13,$E5,$BE,$1B
	.dc.b	$FB,$EF,$33,$F2,$D1,$F5,$9F,$D8,$1F,$CC,$73,$FC,$B5,$7C,$28,$F0
	.dc.b	$9F,$08,$FD,$14,$7F,$18,$CF,$F5,$99,$FC,$79,$FC,$51,$F8,$03,$F1
	.dc.b	$4F,$EE,$AB,$E3,$47,$F9,$1F,$1F,$B7,$7E,$0B,$E7,$3E,$03,$9E,$3F
	.dc.b	$AA,$CB,$E8,$3F,$DE,$32,$F7,$CF,$DF,$8F,$79,$F3,$DF,$E7,$B1,$F1
	.dc.b	$5F,$AA,$6D,$FC,$CB,$F1,$63,$F4,$67,$F8,$63,$3F,$D1,$C7,$C8,$3F
	.dc.b	$81,$65,$FE,$13,$AF,$EC,$2F,$B0,$FD,$7C,$7F,$9A,$0F,$C0,$C7,$ED
	.dc.b	$1F,$65,$75,$7E,$31,$D4,$4F,$57,$F8,$36,$1F,$B1,$CB,$E3,$E3,$F0
	.dc.b	$B0,$F0,$CE,$6B,$F0,$11,$EC,$C3,$E1,$3F,$8F,$8F,$E1,$11,$F8,$F7
	.dc.b	$F8,$41,$F9,$4C,$FE,$B5,$73,$8F,$8C,$7E,$95,$FA,$AF,$9F,$1F,$EA
	.dc.b	$CF,$E2,$8F,$D5,$9F,$95,$7F,$14,$EE,$3F,$9F,$BF,$D2,$CF,$C8,$61
	.dc.b	$EB,$7D,$F9,$7E,$93,$1F,$D8,$5F,$FA,$32,$7E,$59,$E7,$3F,$85,$DF
	.dc.b	$9F,$1F,$64,$68,$FC,$0E,$3F,$95,$8F,$A0,$7E,$2E,$3F,$2A,$72,$C3
	.dc.b	$C0,$FF,$27,$C3,$E5,$1F,$65,$F5,$47,$F4,$E7,$E7,$9F,$15,$FE,$44
	.dc.b	$FF,$8A,$67,$BC,$FF,$09,$CF,$F5,$15,$AA,$FF,$08,$1F,$27,$1F,$DE
	.dc.b	$CF,$E8,$E7,$51,$F3,$A7,$E4,$1F,$3A,$74,$FE,$66,$BF,$78,$EA,$79
	.dc.b	$E5,$F3,$63,$F9,$5E,$1F,$E3,$1C,$FF,$17,$1F,$08,$FE,$37,$9F,$CB
	.dc.b	$C3,$E8,$BF,$E9,$17,$E1,$C7,$D1,$37,$97,$C8,$7F,$41,$7F,$B2,$9F
	.dc.b	$90,$FE,$B1,$FE,$88,$FD,$18,$F9,$67,$F8,$FF,$3F,$AD,$3A,$3E,$46
	.dc.b	$3F,$C9,$2B,$F3,$73,$F9,$6B,$F0,$65,$F8,$59,$F0,$61,$E5,$7F,$83
	.dc.b	$3F,$C0,$1F,$C2,$9A,$7F,$0A,$7E,$D3,$0F,$03,$FC,$7B,$0F,$04,$7A
	.dc.b	$8F,$9B,$5F,$D8,$AF,$F8,$2D,$FF,$0F,$8F,$DC,$5F,$CE,$AF,$F2,$33
	.dc.b	$FA,$A3,$F4,$98,$FE,$AA,$BE,$24,$7E,$06,$3F,$B6,$3F,$8B,$3E,$E1
	.dc.b	$F8,$B9,$F2,$BF,$D4,$8F,$B1,$3E,$E7,$F9,$61,$F9,$43,$BD,$FE,$2A
	.dc.b	$BF,$63,$7E,$99,$FE,$23,$5E,$99,$FF,$4C,$C7,$A4,$FF,$23,$E5,$F8
	.dc.b	$67,$EB,$CF,$F9,$BF,$2F,$AF,$1E,$9C,$3E,$1E,$3F,$88,$7F,$D2,$F3
	.dc.b	$A9,$FC,$A6,$1E,$E7,$CA,$7B,$8F,$D1,$9A,$9E,$6F,$F3,$08,$F2,$D7
	.dc.b	$F3,$B8,$F7,$9F,$37,$3F,$D1,$E1,$F8,$68,$FD,$E1,$FC,$82,$7C,$8F
	.dc.b	$FA,$26,$77,$1F,$76,$3E,$71,$EB,$3E,$2B,$C2,$7E,$04,$7E,$1C,$D3
	.dc.b	$F9,$B3,$E5,$5F,$C0,$8D,$1E,$B3,$99,$EB,$74,$FC,$C9,$FC,$3C,$FE
	.dc.b	$59,$FD,$7C,$7E,$58,$D1,$F5,$0F,$BD,$5E,$A7,$F4,$11,$FD,$12,$BF
	.dc.b	$4B,$7F,$B9,$9F,$C6,$5F,$C5,$7F,$72,$7C,$53,$F9,$01,$AA,$F8,$8F
	.dc.b	$D4,$CB,$F1,$13,$EF,$9F,$88,$7F,$A7,$F1,$FB,$D3,$A8,$FD,$CD,$7F
	.dc.b	$4E,$3D,$55,$FA,$07,$E2,$3C,$DF,$88,$7F,$03,$3E,$4C,$F3,$7F,$68
	.dc.b	$73,$9F,$EB,$93,$FE,$40,$3E,$44,$7A,$4F,$86,$7F,$4C,$3D,$C7,$F1
	.dc.b	$68,$F9,$A7,$23,$FC,$17,$1C,$A3,$F2,$AF,$C7,$3F,$5D,$3F,$DF,$8F
	.dc.b	$E2,$13,$CA,$BF,$8A,$57,$69,$CB,$E3,$3E,$0C,$FF,$0A,$CB,$F5,$97
	.dc.b	$F8,$89,$F8,$07,$DE,$3F,$D4,$07,$DE,$CF,$F9,$D5,$7E,$86,$FC,$58
	.dc.b	$FD,$DA,$F3,$CF,$D8,$9F,$E9,$07,$C9,$8F,$E1,$07,$57,$FB,$AD,$FD
	.dc.b	$37,$59,$FF,$56,$9F,$DA,$47,$2C,$7C,$F7,$F8,$09,$FF,$38,$E5,$F8
	.dc.b	$09,$CC,$7D,$AC,$3F,$4E,$F7,$C7,$FA,$63,$F8,$E3,$CE,$7D,$63,$F8
	.dc.b	$81,$F0,$0F,$DE,$65,$F0,$2F,$B4,$7C,$33,$F9,$0E,$1F,$9F,$7E,$85
	.dc.b	$7C,$3C,$BF,$49,$1F,$40,$FD,$E9,$F9,$09,$FC,$AB,$FE,$07,$7F,$04
	.dc.b	$FF,$47,$7D,$CF,$F6,$FB,$F1,$3F,$94,$C7,$DE,$FF,$8E,$2F,$F8,$AE
	.dc.b	$1E,$68,$F4,$9E,$63,$F2,$58,$FE,$5F,$0E,$B9,$7F,$76,$9D,$1F,$09
	.dc.b	$FF,$26,$61,$FE,$81,$C3,$F9,$46,$5E,$CB,$F3,$DF,$F2,$2C,$3F,$31
	.dc.b	$39,$C7,$F2,$71,$DE,$7E,$84,$E5,$FE,$4B,$19,$7F,$48,$7E,$5E,$BA
	.dc.b	$E3,$F8,$EC,$FF,$DC,$55,$F9,$73,$F7,$15,$F0,$63,$F8,$DC,$7E,$2F
	.dc.b	$2F,$8B,$5F,$D9,$2F,$F3,$11,$EB,$AF,$E7,$B7,$C4,$FD,$0D,$F1,$7E
	.dc.b	$F1,$F8,$98,$FE,$B4,$7F,$8C,$5F,$9A,$64,$FC,$BE,$5F,$D7,$AB,$C2
	.dc.b	$FD,$48,$FB,$07,$4A,$F5,$5F,$F1,$98,$C9,$D2,$32,$FE,$78,$CB,$F3
	.dc.b	$B3,$FC,$1C,$78,$63,$E5,$63,$FC,$02,$3F,$3A,$7E,$44,$FE,$55,$87
	.dc.b	$93,$1F,$8F,$87,$F0,$39,$F2,$3F,$84,$C7,$FC,$0C,$E6,$3C,$8E,$67
	.dc.b	$E1,$C7,$FB,$94,$E2,$7E,$C6,$3D,$8F,$95,$FE,$3A,$71,$7F,$A9,$9F
	.dc.b	$47,$0F,$DB,$9F,$45,$EE,$73,$3E,$11,$CC,$FD,$31,$F8,$CC,$7E,$1C
	.dc.b	$71,$3E,$34,$7E,$72,$7F,$02,$FF,$B2,$CF,$E2,$71,$F3,$9F,$F6,$DE
	.dc.b	$7F,$A9,$8E,$31,$DC,$F9,$59,$7F,$13,$7D,$E7,$A6,$BD,$F3,$E1,$7D
	.dc.b	$CF,$E5,$A7,$F2,$55,$D6,$72,$7F,$02,$9F,$F2,$23,$FA,$0C,$BF,$2D
	.dc.b	$7F,$1A,$F8,$BF,$82,$9E,$2F,$F5,$43,$F8,$55,$FE,$C2,$3E,$7C,$7C
	.dc.b	$53,$E1,$47,$85,$C1,$ED,$8F,$03,$F1,$73,$FE,$65,$87,$F5,$19,$FE
	.dc.b	$3C,$FE,$3B,$3F,$E3,$D9,$7D,$09,$FC,$34,$FC,$83,$E6,$63,$FC,$FA
	.dc.b	$32,$7A,$30,$FA,$2D,$3C,$2F,$FB,$C1,$F8,$87,$F5,$D7,$F8,$3B,$EC
	.dc.b	$7F,$83,$9F,$9E,$78,$CF,$FA,$82,$3E,$23,$F3,$31,$F8,$91,$F8,$83
	.dc.b	$8E,$3F,$C4,$DF,$F0,$9C,$FB,$73,$FE,$1F,$7F,$CC,$AB,$E1,$BF,$D7
	.dc.b	$72,$FD,$F5,$FE,$0B,$1F,$AA,$FE,$1D,$FE,$30,$FF,$8F,$F2,$F8,$CF
	.dc.b	$F4,$28,$FE,$BF,$1F,$E7,$E8,$FE,$A5,$5F,$96,$7E,$11,$EA,$7E,$13
	.dc.b	$E6,$7E,$14,$FE,$FE,$FD,$58,$FF,$81,$B3,$FC,$4B,$F9,$1C,$7F,$89
	.dc.b	$61,$E5,$9C,$E1,$F3,$A3,$F9,$86,$7F,$19,$FE,$79,$1F,$5F,$0F,$83
	.dc.b	$1E,$69,$FC,$53,$F2,$B2,$F6,$BF,$C3,$8F,$89,$87,$83,$1F,$D2,$47
	.dc.b	$E4,$1F,$C2,$C7,$F0,$E8,$F8,$98,$FF,$15,$7F,$99,$E3,$F8,$87,$83
	.dc.b	$F8,$7C,$3C,$D9,$7C,$EC,$3F,$75,$97,$9A,$FF,$6F,$8F,$D6,$7F,$85
	.dc.b	$9F,$B7,$3F,$85,$CF,$91,$F5,$1F,$A6,$3F,$5C,$78,$0F,$E2,$47,$F0
	.dc.b	$63,$27,$03,$F5,$57,$E4,$C7,$E6,$C7,$D9,$3E,$69,$F9,$A3,$F2,$58
	.dc.b	$7D,$9C,$BE,$4B,$FC,$36,$3C,$EF,$E3,$2F,$CC,$EE,$3C,$C7,$DB,$3F
	.dc.b	$71,$3F,$8B,$7E,$0E,$7F,$D5,$EB,$F6,$99,$FC,$99,$F1,$1F,$78,$F7
	.dc.b	$1F,$14,$F7,$1D,$4F,$C9,$9F,$B8,$3E,$1D,$FE,$F5,$FB,$27,$E0,$2B
	.dc.b	$FA,$6B,$83,$D7,$97,$E0,$EB,$EE,$47,$C9,$BF,$C1,$D7,$F4,$C7,$F8
	.dc.b	$4E,$3F,$2A,$7C,$07,$97,$0F,$03,$F8,$F3,$AD,$78,$1C,$57,$C0,$7C
	.dc.b	$F8,$7E,$3D,$CE,$1F,$9D,$C7,$E4,$3F,$C2,$8F,$F6,$FE,$7F,$8B,$71
	.dc.b	$3F,$EB,$07,$F5,$0F,$F1,$1C,$FF,$00,$FD,$23,$FD,$0C,$E1,$F9,$33
	.dc.b	$CE,$31,$87,$F1,$79,$F6,$E5,$F7,$EF,$E2,$BF,$7E,$3F,$1A,$7D,$87
	.dc.b	$F4,$11,$FC,$57,$0F,$4D,$F4,$8F,$4C,$62,$B8,$67,$F2,$F1,$F1,$46
	.dc.b	$0F,$41,$FD,$28,$C1,$E8,$9E,$4F,$F1,$13,$F3,$39,$7F,$42,$7F,$27
	.dc.b	$97,$E5,$A3,$11,$E8,$38,$1F,$8F,$3F,$01,$87,$E4,$DC,$E5,$EC,$7F
	.dc.b	$5F,$87,$99,$CE,$3E,$9C,$7C,$78,$FF,$1F,$9F,$4B,$FC,$FD,$F4,$BF
	.dc.b	$36,$B9,$3B,$8F,$61,$F5,$8D,$E1,$F2,$1C,$1E,$03,$AB,$FC,$CA,$39
	.dc.b	$1D,$6B,$C7,$1F,$D9,$8F,$E6,$B8,$FC,$29,$FC,$ED,$7F,$91,$B3,$FD
	.dc.b	$A5,$7E,$46,$3B,$CF,$48,$F1,$47,$D3,$3C,$47,$A0,$C4,$6D,$FC,$A1
	.dc.b	$FC,$3C,$FC,$A1,$EC,$32,$7E,$3A,$BF,$64,$FE,$D6,$3F,$4C,$FE,$B6
	.dc.b	$3F,$4D,$97,$F1,$39,$CB,$F9,$59,$E1,$5E,$97,$C6,$FC,$17,$F0,$B7
	.dc.b	$FD,$A9,$FB,$78,$7C,$67,$FD,$27,$3F,$1A,$75,$1F,$E9,$F3,$F1,$71
	.dc.b	$FA,$2C,$3E,$53,$F8,$7C,$3E,$04,$FF,$AA,$CF,$81,$8F,$F6,$29,$F1
	.dc.b	$47,$E3,$F1,$FD,$55,$7D,$23,$FC,$39,$1E,$23,$CE,$7F,$77,$3C,$27
	.dc.b	$B8,$FE,$4B,$1F,$05,$FE,$12,$62,$FE,$5B,$83,$F3,$A7,$F1,$0C,$3F
	.dc.b	$8E,$5F,$FA,$C3,$2F,$CD,$4F,$E0,$23,$CF,$7F,$A1,$8F,$84,$7E,$A7
	.dc.b	$0F,$7C,$7F,$82,$23,$81,$E9,$3D,$D8,$7C,$67,$C0,$7E,$3C,$FF,$3B
	.dc.b	$1F,$C9,$8F,$ED,$A7,$B4,$F3,$9F,$81,$C3,$CE,$FF,$3E,$C3,$CE,$FD
	.dc.b	$E3,$C1,$5E,$73,$FD,$94,$F9,$F0,$F5,$CF,$0C,$3F,$69,$3F,$9A,$7F
	.dc.b	$50,$FA,$8F,$F6,$86,$1F,$EF,$67,$C7,$7F,$3C,$FC,$95,$FE,$3E,$B9
	.dc.b	$9F,$12,$FF,$C1,$0F,$FA,$2A,$3F,$11,$87,$E3,$2B,$CB,$3F,$81,$9F
	.dc.b	$F1,$D4,$FE,$CA,$FE,$D3,$FD,$26,$BB,$C6,$0F,$8E,$7D,$D7,$F1,$73
	.dc.b	$F9,$33,$DF,$1F,$C1,$8F,$EF,$E2,$BF,$1E,$FF,$84,$69,$F9,$1B,$DC
	.dc.b	$4C,$97,$C8,$5A,$D4,$E6,$31,$5C,$21,$5C,$1B,$4F,$77,$53,$99,$FC
	.dc.b	$6E,$1F,$0D,$FF,$1E,$1F,$1F,$4E,$6F,$A7,$2F,$17,$BB,$CB,$3F,$27
	.dc.b	$BB,$6B,$BB,$71,$F6,$D6,$67,$BD,$FC,$77,$DE,$B8,$7D,$59,$7F,$B8
	.dc.b	$2F,$E2,$FB,$BC,$F5,$7B,$B4,$C4,$75,$E0,$FB,$B8,$F5,$C8,$F5,$7D
	.dc.b	$98,$7F,$13,$1F,$92,$7C,$98,$F9,$3B,$7D,$18,$C4,$FC,$B3,$73,$ED
	.dc.b	$87,$C9,$F8,$78,$3E,$3E,$BF,$E6,$FB,$FC,$54,$63,$1E,$9B,$71,$26
	.dc.b	$3F,$6C,$FB,$38,$BD,$FE,$98,$7F,$24,$E5,$73,$7C,$6C,$7F,$C2,$3B
	.dc.b	$33,$ED,$97,$2F,$7F,$CC,$47,$B6,$3F,$03,$DF,$F5,$61,$27,$3A,$7C
	.dc.b	$86,$FB,$FE,$27,$FD,$A5,$8F,$6C,$3E,$4F,$BF,$F0,$FD,$F9,$5C,$A3
	.dc.b	$E3,$7B,$FF,$1F,$DF,$93,$DA,$3A,$71,$FC,$78,$FC,$71,$F0,$AB,$9B
	.dc.b	$F5,$0F,$8B,$7F,$46,$3E,$97,$BF,$27,$D2,$FD,$5F,$7F,$AE,$DF,$E3
	.dc.b	$6E,$3F,$93,$5F,$82,$3B,$E9,$F8,$2C,$7B,$BF,$7F,$74,$77,$9F,$8B
	.dc.b	$EF,$FE,$04,$8F,$C8,$F7,$FB,$BF,$D4,$B9,$7E,$B3,$FB,$CE,$5F,$AD
	.dc.b	$8F,$E4,$71,$7B,$FF,$8B,$25,$7D,$2F,$7F,$F1,$DE,$FF,$E3,$FD,$F9
	.dc.b	$3F,$5A,$3B,$BF,$DE,$F0,$E9,$7C,$A3,$E4,$47,$E7,$A3,$F3,$17,$F9
	.dc.b	$9F,$7F,$F3,$5E,$FC,$BF,$C0,$C7,$4E,$0F,$8F,$1F,$50,$38,$E5,$FD
	.dc.b	$B6,$9E,$A9,$FC,$4E,$3E,$1C,$BE,$56,$5F,$C0,$FF,$25,$FE,$21,$1F
	.dc.b	$AB,$5F,$1D,$93,$5E,$0F,$87,$87,$E4,$E3,$E0,$6F,$0F,$CA,$8F,$B3
	.dc.b	$6F,$8B,$EC,$48,$FC,$FC,$3E,$27,$C5,$F0,$52,$72,$F9,$7E,$0B,$C1
	.dc.b	$1F,$AF,$C7,$BF,$7F,$E7,$F0,$ED,$3D,$3E,$0B,$D8,$77,$7E,$7C,$3E
	.dc.b	$0B,$E0,$12,$3C,$7F,$05,$D9,$F9,$9C,$BF,$2F,$B7,$F9,$D6,$FE,$DE
	.dc.b	$6F,$D9,$71,$B2,$F0,$E1,$DB,$3F,$0B,$EE,$E1,$F9,$DF,$05,$23,$E0
	.dc.b	$BF,$47,$E0,$BC,$18,$FE,$F6,$7E,$61,$FA,$37,$BE,$0F,$82,$FB,$04
	.dc.b	$9F,$97,$F0,$5F,$6F,$E0,$BE,$E7,$C1,$49,$F9,$F1,$DB,$7F,$E2,$BC
	.dc.b	$7F,$27,$0E,$55,$F1,$9F,$C2,$61,$F8,$5F,$82,$FC,$37,$C1,$4A,$FB
	.dc.b	$31,$CD,$FF,$2E,$71,$7C,$5C,$7F,$B3,$69,$E7,$9F,$BF,$F0,$5F,$8F
	.dc.b	$3B,$F1,$FE,$3F,$E0,$BF,$24,$49,$FC,$2F,$C1,$75,$7E,$19,$F9,$FC
	.dc.b	$7B,$3F,$DB,$30,$7C,$17,$E6,$C9,$1F,$CB,$3C,$E7,$C8,$7F,$61,$5F
	.dc.b	$15,$7E,$12,$56,$E3,$C1,$D7,$F8,$18,$FE,$0D,$C9,$F8,$FC,$7A,$C7
	.dc.b	$E3,$1E,$98,$F5,$F8,$4E,$DF,$09,$2B,$C5,$1D,$3C,$0F,$CC,$F0,$FF
	.dc.b	$5F,$DB,$D3,$5E,$7F,$84,$F3,$1D,$A7,$FA,$F6,$FF,$C0,$1B,$7E,$B3
	.dc.b	$D9,$F0,$5D,$6F,$F1,$1C,$BF,$E1,$7B,$FE,$A5,$1F,$0F,$6F,$84,$7C
	.dc.b	$3F,$84,$F8,$9F,$15,$27,$E4,$3F,$1F,$E2,$BA,$47,$D1,$AF,$E9,$D7
	.dc.b	$D2,$3F,$8D,$B7,$3F,$3C,$F5,$9F,$A5,$F1,$5F,$47,$E2,$A4,$60,$FE
	.dc.b	$DC,$7F,$84,$8F,$9F,$5F,$E0,$3B,$E9,$E5,$FA,$F0,$F4,$D7,$DD,$CF
	.dc.b	$EE,$7C,$57,$DD,$F8,$89,$3F,$81,$8F,$4F,$89,$FE,$19,$C7,$FC,$61
	.dc.b	$3F,$85,$F8,$8F,$C2,$1F,$85,$F8,$7F,$C3,$7C,$47,$E1,$FE,$1A,$57
	.dc.b	$E2,$E3,$9F,$1F,$F0,$4D,$7E,$1E,$1C,$AB,$A1,$F9,$1C,$3F,$21,$F0
	.dc.b	$FF,$E3,$1F,$C9,$BF,$C3,$3C,$1F,$E4,$B6,$D7,$F9,$1B,$4F,$F1,$5B
	.dc.b	$F1,$3C,$1F,$F1,$4B,$F1,$0F,$EC,$B1,$EA,$7E,$8E,$3E,$17,$E9,$E3
	.dc.b	$DF,$E2,$3C,$1F,$11,$27,$FC,$1F,$8F,$F8,$99,$ED,$A7,$F8,$59,$FE
	.dc.b	$B0,$7F,$72,$8F,$F1,$03,$C3,$E1,$FD,$47,$AF,$E2,$3D,$9F,$11,$ED
	.dc.b	$E0,$95,$DA,$3E,$3B,$E9,$E3,$FE,$53,$87,$48,$FE,$C6,$B0,$EB,$1E
	.dc.b	$6C,$3E,$47,$C4,$7C,$9F,$87,$93,$EA,$8E,$AF,$F9,$E6,$FA,$74,$FF
	.dc.b	$91,$5E,$91,$E9,$3A,$1F,$0B,$E1,$7E,$A7,$C4,$48,$C1,$F1,$2B,$FA
	.dc.b	$83,$0F,$D1,$3F,$E5,$7A,$71,$FD,$DF,$88,$95,$F3,$A3,$F0,$1F,$11
	.dc.b	$F8,$1F,$88,$93,$F4,$A3,$D3,$CF,$FC,$73,$16,$3D,$2B,$EB,$E3,$D1
	.dc.b	$F8,$F8,$FE,$2F,$E2,$24,$7D,$C7,$F1,$DF,$07,$D3,$97,$FC,$89,$87
	.dc.b	$3A,$F9,$F8,$73,$C7,$E3,$63,$F9,$5F,$83,$95,$F8,$67,$F3,$1F,$0F
	.dc.b	$CE,$BE,$07,$C4,$7E,$6F,$E1,$E5,$FE,$32,$3F,$3D,$F1,$12,$3F,$1E
	.dc.b	$FF,$A4,$7F,$27,$FF,$07,$F8,$1F,$A9,$C3,$E6,$75,$7F,$61,$F4,$FF
	.dc.b	$98,$61,$F6,$F2,$EA,$FE,$19,$FD,$56,$5C,$FE,$22,$4F,$F9,$AB,$2F
	.dc.b	$F0,$CB,$7E,$0B,$D0,$FE,$6E,$F9,$E5,$F8,$57,$9C,$71,$F8,$8F,$37
	.dc.b	$C4,$48,$C1,$F8,$33,$FC,$DF,$5F,$E9,$7C,$1F,$11,$ED,$24,$F8,$FE
	.dc.b	$1B,$E0,$7C,$44,$8F,$33,$F0,$FE,$23,$CB,$C1,$FE,$0C,$C3,$9D,$7A
	.dc.b	$F0,$E6,$FF,$92,$21,$87,$CB,$F8,$89,$3F,$05,$F9,$DF,$11,$D3,$C7
	.dc.b	$FF,$6C,$5F,$4A,$E5,$87,$4D,$3F,$D5,$53,$D2,$7E,$57,$C3,$7D,$7F
	.dc.b	$86,$91,$83,$E1,$69,$FE,$85,$3D,$1F,$0F,$F7,$7E,$22,$4F,$D3,$8F
	.dc.b	$C0,$7C,$3F,$E0,$7E,$22,$4F,$D7,$8F,$33,$FE,$E6,$C9,$5C,$A3,$E0
	.dc.b	$5F,$E2,$63,$F1,$5F,$11,$27,$F0,$0F,$E3,$7E,$23,$9E,$AC,$39,$D7
	.dc.b	$E1,$70,$E9,$3F,$03,$0F,$CA,$7C,$44,$AF,$C5,$3F,$97,$F8,$6E,$7F
	.dc.b	$94,$7F,$82,$A7,$FD,$6E,$FF,$82,$9B,$F1,$1F,$EE,$D8,$D4,$7F,$9F
	.dc.b	$97,$C4,$E7,$E1,$E4,$F2,$8E,$BF,$13,$C6,$E4,$F6,$8E,$9F,$13,$D7
	.dc.b	$E2,$24,$F8,$A3,$8B,$E2,$21,$E1,$9F,$27,$C4,$79,$7E,$22,$4F,$AA
	.dc.b	$3D,$BE,$77,$A7,$E1,$24,$FB,$A3,$D9,$F1,$1E,$DF,$88,$93,$F0,$A3
	.dc.b	$E2,$D7,$08,$F8,$0F,$C3,$AF,$89,$F1,$12,$7E,$4B,$F1,$FE,$23,$8C
	.dc.b	$7C,$CA,$E5,$CB,$FE,$FD,$E2,$FF,$00,$CA,$10,$F8,$7F,$A0,$49,$FA
	.dc.b	$BF,$11,$F6,$3E,$17,$EA,$12,$3E,$CF,$C2,$FD,$7F,$87,$91,$F7,$1F
	.dc.b	$B5,$F1,$1C,$63,$EF,$FC,$47,$DD,$39,$3F,$EB,$AE,$1F,$EF,$3A,$1A
	.dc.b	$4F,$88,$E5,$3F,$83,$AE,$4F,$E2,$B2,$32,$DE,$AD,$C1,$F1,$1A,$AF
	.dc.b	$03,$6B,$71,$F1,$67,$6F,$3A,$E5,$F0,$DF,$A4,$9F,$EC,$73,$E9,$F8
	.dc.b	$8E,$55,$DD,$B1,$CF,$E2,$2E,$1E,$DF,$36,$E0,$F8,$7A,$62,$3B,$56
	.dc.b	$B0,$98,$78,$CE,$85,$C3,$E4,$7C,$47,$4F,$88,$93,$9C,$93,$D2,$31
	.dc.b	$7D,$1F,$A5,$87,$93,$E2,$24,$71,$7C,$FF,$03,$F0,$7E,$0B,$D2,$49
	.dc.b	$ED,$F0,$FE,$CF,$88,$91,$E1,$7D,$FF,$0F,$C6,$3C,$95,$C5,$FC,$15
	.dc.b	$7C,$4F,$87,$91,$E9,$7E,$3F,$C3,$71,$8F,$C4,$57,$1E,$0F,$B2,$7E
	.dc.b	$26,$FE,$6F,$C3,$4A,$F8,$4F,$D0,$F8,$4F,$B5,$F0,$5F,$4B,$E1,$A5
	.dc.b	$7C,$78,$FA,$FF,$94,$FF,$6E,$8F,$D4,$37,$13,$4F,$EC,$F1,$FA,$DE
	.dc.b	$8F,$EE,$8F,$F5,$E1,$ED,$7F,$B8,$39,$98,$78,$27,$F4,$58,$FE,$12
	.dc.b	$3C,$3F,$0B,$E2,$F8,$59,$5F,$7F,$2F,$F7,$3E,$7F,$98,$8F,$F4,$00
	.dc.b	$F8,$9F,$F1,$16,$5D,$A3,$F0,$59,$77,$AF,$8F,$39,$9F,$41,$EE,$F8
	.dc.b	$6F,$83,$F0,$32,$3C,$11,$F1,$3E,$12,$47,$8D,$F8,$FF,$0D,$F2,$3E
	.dc.b	$1A,$4F,$9D,$F9,$7F,$0D,$F3,$3E,$1A,$4F,$AD,$F9,$FF,$0D,$F4,$3E
	.dc.b	$1A,$4F,$BD,$FA,$7F,$01,$F5,$3E,$06,$4F,$C3,$7E,$BF,$C2,$FD,$8F
	.dc.b	$86,$FB,$3F,$0D,$2B,$E4,$47,$DC,$F8,$6F,$BB,$F0,$B2,$7E,$64,$7E
	.dc.b	$03,$E1,$7F,$6F,$F9,$57,$F5,$2E,$6F,$C6,$B7,$BB,$F8,$77,$57,$FB
	.dc.b	$43,$0F,$93,$4A,$FE,$B9,$18,$6C,$72,$F8,$4E,$BF,$0B,$DB,$E1,$BB
	.dc.b	$FC,$2C,$C3,$F7,$98,$27,$C7,$F0,$D2,$72,$F9,$BE,$13,$CF,$F0,$B2
	.dc.b	$78,$BE,$AF,$84,$ED,$1D,$31,$EC,$F9,$70,$E8,$F9,$F0,$BA,$7C,$0C
	.dc.b	$7E,$1F,$C1,$7C,$4F,$85,$97,$E6,$8E,$AF,$BF,$1E,$B7,$F2,$31,$E7
	.dc.b	$C1,$FC,$D7,$56,$78,$C3,$FA,$A4,$7F,$BD,$67,$5E,$17,$D2,$24,$FC
	.dc.b	$4F,$85,$FA,$BF,$0B,$23,$E3,$BF,$63,$E1,$3E,$CF,$C2,$7D,$AF,$7E
	.dc.b	$57,$CC,$8F,$BB,$F0,$92,$3E,$7B,$F8,$0F,$84,$E9,$1F,$67,$0B,$BF
	.dc.b	$E0,$B1,$EB,$3F,$5B,$1E,$AF,$DF,$C7,$F1,$3F,$09,$F8,$AF,$83,$93
	.dc.b	$F7,$23,$FA,$1F,$E5,$7F,$D8,$09,$18,$F8,$5C,$FC,$2C,$8F,$E7,$D3
	.dc.b	$FD,$51,$FC,$7B,$FE,$B6,$E1,$FE,$C0,$FF,$54,$D3,$F9,$1A,$F8,$0E
	.dc.b	$E7,$83,$E0,$3C,$37,$0E,$A7,$84,$EF,$87,$85,$EE,$FA,$3E,$17,$D3
	.dc.b	$F0,$9E,$AF,$85,$95,$E6,$8F,$6F,$C2,$48,$F4,$BF,$03,$E1,$7E,$0F
	.dc.b	$C2,$C9,$F6,$BF,$13,$E1,$7E,$2F,$C0,$FC,$6F,$81,$95,$F0,$A3,$A3
	.dc.b	$F0,$B0,$BC,$1F,$03,$1E,$93,$F1,$B1,$B3,$F1,$31,$FA,$3F,$0B,$2B
	.dc.b	$E6,$BF,$53,$E1,$79,$F1,$FC,$CC,$AD,$7F,$8A,$78,$D9,$F4,$AF,$A3
	.dc.b	$9D,$AD,$65,$F7,$BE,$16,$47,$DA,$7F,$03,$F0,$BF,$82,$F8,$59,$3F
	.dc.b	$79,$FC,$2F,$C2,$FE,$1B,$E1,$24,$FE,$09,$FC,$4F,$C0,$FF,$83,$3D
	.dc.b	$AF,$EA,$87,$F5,$2E,$2F,$EC,$6D,$D7,$FC,$48,$FF,$59,$DD,$97,$83
	.dc.b	$1F,$C5,$BE,$07,$B1,$0F,$C7,$E3,$E0,$7F,$C2,$15,$FD,$4A,$FA,$FC
	.dc.b	$27,$8C,$F2,$7C,$27,$97,$E0,$A5,$F0,$7D,$1F,$07,$E9,$F8,$3F,$57
	.dc.b	$C1,$CA,$EB,$1E,$DF,$83,$91,$E0,$7E,$07,$C1,$F4,$8F,$E6,$3A,$31
	.dc.b	$C7,$07,$C2,$CB,$B5,$7A,$32,$EC,$F0,$CB,$E5,$7C,$1F,$CB,$C6,$4F
	.dc.b	$BA,$3E,$77,$C1,$C8,$F8,$2F,$D1,$F8,$3F,$A5,$F0,$72,$7E,$2B,$F5
	.dc.b	$7E,$0F,$EB,$7C,$17,$D7,$F8,$39,$5F,$2A,$3A,$3E,$AC,$2E,$DF,$6B
	.dc.b	$1E,$93,$EE,$C6,$CF,$DE,$C7,$F0,$5F,$05,$F8,$3F,$83,$97,$F5,$A3
	.dc.b	$9F,$37,$E0,$B2,$B3,$FE,$83,$DB,$FD,$7A,$F5,$3D,$24,$3E,$6E,$1F
	.dc.b	$8F,$F8,$3F,$F2,$DF,$E5,$9F,$D8,$E7,$FD,$18,$DC,$D7,$FD,$95,$FF
	.dc.b	$1A,$C7,$F6,$2C,$7F,$CE,$CD,$AE,$3F,$07,$D3,$E0,$FA,$FB,$BD,$BD
	.dc.b	$E9,$5F,$E5,$FB,$FE,$C8,$DF,$7B,$C7,$F0,$32,$30,$6F,$1F,$39,$1E
	.dc.b	$58,$F6,$7A,$61,$D9,$FE,$F7,$87,$66,$FC,$1F,$C0,$F8,$39,$1E,$38
	.dc.b	$F8,$7F,$05,$23,$CC,$FC,$6F,$83,$F8,$FF,$05,$27,$D4,$FC,$AF,$82
	.dc.b	$F9,$7F,$01,$F3,$3E,$02,$57,$BE,$3A,$3E,$4C,$2F,$1F,$CC,$C7,$A4
	.dc.b	$FA,$31,$B3,$F4,$31,$FA,$FE,$FF,$D8,$F8,$09,$7F,$2E,$3E,$DF,$C1
	.dc.b	$48,$F9,$CF,$DE,$F8,$2F,$BF,$F0,$3F,$80,$F8,$29,$5F,$4E,$3F,$D3
	.dc.b	$BF,$96,$FF,$AE,$87,$C1,$E0,$CC,$F5,$3E,$6E,$17,$4F,$C2,$E3,$DA
	.dc.b	$7E,$F6,$3D,$9F,$A7,$8F,$6F,$81,$EF,$F0,$32,$7F,$D4,$59,$FF,$65
	.dc.b	$C7,$F1,$F9,$75,$ED,$FC,$7B,$EB,$AF,$41,$D8,$87,$63,$9F,$C1,$7B
	.dc.b	$3E,$0E,$46,$0E,$FF,$05,$F0,$3E,$0E,$46,$0F,$1F,$BF,$F1,$3D,$F9
	.dc.b	$18,$3C,$FE,$FF,$C8,$F7,$E4,$60,$E4,$7F,$93,$36,$FF,$63,$3E,$EC
	.dc.b	$3E,$7E,$3F,$43,$1E,$B3,$BC,$70,$FC,$4F,$82,$FA,$BF,$07,$23,$07
	.dc.b	$C8,$F8,$3F,$B2,$47,$E5,$FC,$1F,$DC,$F8,$3F,$BB,$F0,$72,$7E,$7C
	.dc.b	$7E,$03,$E0,$24,$7D,$27,$F0,$7F,$07,$CE,$3D,$58,$5E,$6F,$F5,$BF
	.dc.b	$03,$2E,$D5,$F6,$B2,$EC,$FC,$1C,$BF,$1B,$F0,$32,$3E,$FB,$F9,$0F
	.dc.b	$83,$FF,$76,$FE,$5D,$FE,$B5,$27,$4E,$3E,$33,$3F,$09,$30,$FF,$74
	.dc.b	$57,$F5,$95,$AA,$1F,$09,$D3,$E0,$A5,$70,$7B,$FC,$17,$83,$29,$3D
	.dc.b	$1F,$03,$DB,$1F,$03,$FE,$86,$C3,$C1,$8F,$7C,$7D,$3F,$03,$EA,$F8
	.dc.b	$29,$7E,$68,$EA,$F1,$C6,$DF,$FB,$83,$0F,$EF,$D3,$EA,$7A,$93,$1F
	.dc.b	$8B,$8F,$C6,$F8,$4F,$8F,$F0,$92,$FE,$1C,$7C,$AF,$84,$F9,$7E,$DC
	.dc.b	$9F,$8F,$1F,$3B,$DE,$F9,$FF,$05,$27,$E5,$C7,$D2,$F8,$2F,$A7,$F0
	.dc.b	$12,$7E,$7C,$75,$7E,$8E,$1D,$5F,$89,$87,$57,$EA,$E1,$D5,$F9,$18
	.dc.b	$7D,$EF,$83,$91,$F6,$9F,$C0,$FC,$27,$E0,$BE,$0F,$F0,$7F,$09,$2B
	.dc.b	$EF,$C7,$E1,$BE,$11,$13,$8D,$C9,$7C,$4F,$EB,$47,$F5,$BE,$5F,$EB
	.dc.b	$51,$9F,$81,$E3,$7F,$EC,$0A,$7C,$0E,$30,$E7,$3F,$8C,$9C,$BD,$EF
	.dc.b	$93,$E4,$3F,$27,$9F,$9D,$FC,$36,$7E,$77,$81,$CF,$E2,$FB,$57,$9D
	.dc.b	$D4,$78,$FE,$17,$2F,$87,$E0,$EE,$4C,$91,$87,$FA,$F4,$7A,$0D,$7C
	.dc.b	$2E,$7E,$16,$5F,$8A,$72,$DA,$7C,$1F,$32,$46,$7E,$13,$B7,$C1,$48
	.dc.b	$E0,$F8,$7E,$0B,$C5,$97,$69,$F9,$99,$61,$ED,$87,$47,$E8,$3E,$9C
	.dc.b	$3D,$5F,$05,$27,$C8,$FB,$7E,$0B,$94,$74,$3D,$17,$8D,$3E,$06,$9E
	.dc.b	$A9,$EE,$43,$EF,$61,$F1,$FE,$0E,$47,$C1,$7E,$57,$C2,$7C,$BF,$83
	.dc.b	$93,$F1,$5F,$9D,$F0,$9F,$3F,$DE,$93,$F2,$5F,$A5,$F0,$7F,$4F,$E0
	.dc.b	$BE,$A7,$C1,$CA,$F9,$D1,$D1,$FC,$86,$17,$5F,$4E,$3D,$27,$F2,$98
	.dc.b	$D9,$F6,$E3,$F8,$2F,$CB,$FF,$33,$5F,$C8,$3A,$7F,$45,$1F,$DE,$23
	.dc.b	$EF,$35,$D1,$FC,$99,$E3,$5F,$87,$7F,$B8,$4F,$2F,$84,$EC,$77,$F8
	.dc.b	$4F,$07,$C2,$78,$7E,$12,$63,$DA,$3C,$9F,$03,$23,$C2,$F9,$FE,$0B
	.dc.b	$9C,$79,$30,$AF,$93,$36,$7F,$EE,$38,$F3,$61,$D9,$D4,$7F,$B8,$DF
	.dc.b	$01,$EE,$F8,$2F,$89,$F0,$52,$30,$7C,$2F,$84,$F9,$04,$7E,$2F,$C2
	.dc.b	$7C,$BE,$9F,$99,$F0,$92,$7E,$4C,$7C,$FF,$82,$FA,$1F,$03,$27,$E6
	.dc.b	$C7,$37,$CB,$85,$D7,$E6,$63,$D6,$7E,$9E,$3D,$5F,$5E,$3F,$6F,$E0
	.dc.b	$24,$7D,$87,$EF,$7C,$0F,$3E,$5F,$BB,$95,$AF,$B5,$8F,$57,$E2,$E3
	.dc.b	$D7,$2F,$B9,$97,$E1,$FE,$13,$F1,$1F,$09,$2F,$F0,$B1,$F8,$BF,$83
	.dc.b	$91,$F8,$87,$F1,$DF,$07,$FC,$43,$F3,$0F,$E4,$73,$F5,$DF,$EA,$F8
	.dc.b	$9A,$69,$FE,$24,$FF,$66,$3F,$B9,$C7,$F1,$16,$BC,$FD,$FF,$0B,$F7
	.dc.b	$AE,$D7,$E5,$6F,$C5,$B7,$8F,$E1,$25,$F3,$8F,$37,$C2,$48,$EC,$FA
	.dc.b	$7E,$13,$D5,$F0,$72,$7C,$4F,$B7,$E1,$3D,$DF,$05,$27,$CC,$FC,$1F
	.dc.b	$83,$F8,$5F,$01,$27,$D4,$FC,$5F,$81,$E9,$1F,$17,$0B,$B7,$C2,$C7
	.dc.b	$A4,$FC,$9C,$6C,$FC,$6C,$7E,$77,$C0,$7C,$FF,$81,$97,$F1,$E3,$E9
	.dc.b	$7C,$14,$8F,$94,$FD,$5F,$83,$FA,$DF,$07,$27,$E7,$3F,$67,$E0,$FE
	.dc.b	$D7,$BD,$F6,$FD,$F9,$5F,$4E,$3A,$3F,$77,$0B,$A7,$DA,$C7,$A4,$FE
	.dc.b	$07,$1B,$3F,$7B,$1F,$C3,$7C,$1F,$F5,$6F,$CC,$7F,$D9,$47,$FB,$2F
	.dc.b	$AF,$FB,$8E,$1F,$E3,$65,$97,$6E,$7F,$EF,$51,$89,$FC,$36,$7D,$A7
	.dc.b	$39,$D9,$E5,$97,$7F,$83,$91,$FD,$77,$2F,$F6,$A6,$87,$FB,$CC,$7F
	.dc.b	$0F,$7F,$B3,$E1,$FE,$09,$6C,$7F,$92,$9B,$3C,$7E,$03,$D9,$EF,$7B
	.dc.b	$7D,$B9,$7D,$E3,$E0,$7B,$DF,$07,$DC,$93,$E3,$8F,$89,$F0,$32,$3C
	.dc.b	$CF,$C7,$C3,$AC,$7A,$70,$EA,$F9,$F0,$F9,$9F,$03,$F3,$7E,$06,$4F
	.dc.b	$BE,$3E,$87,$C0,$FD,$1F,$81,$93,$F0,$E3,$EA,$7C,$0C,$8F,$8C,$FD
	.dc.b	$7F,$81,$FB,$1E,$FC,$9F,$94,$FD,$BF,$7F,$AC,$7C,$DC,$3A,$BF,$2F
	.dc.b	$0E,$8F,$CD,$C2,$F0,$FF,$84,$F8,$99,$7E,$17,$DF,$97,$F5,$DF,$C4
	.dc.b	$7B,$FF,$89,$F8,$09,$3F,$6D,$FC,$67,$C0,$FF,$88,$B4,$FE,$B3,$CF
	.dc.b	$FE,$41,$6D,$7F,$A0,$9B,$99,$FE,$C4,$FF,$89,$5F,$EB,$F1,$BF,$77
	.dc.b	$A1,$D7,$E0,$7C,$31,$F8,$7C,$6D,$FA,$F1,$F1,$4C,$3F,$C5,$98,$78
	.dc.b	$0E,$F8,$F9,$7D,$EF,$47,$C0,$CA,$DC,$74,$9F,$C9,$63,$67,$F2,$58
	.dc.b	$75,$7F,$B3,$61,$D5,$F8,$5F,$0F,$F0,$FE,$0A,$47,$89,$F8,$DF,$05
	.dc.b	$F1,$FE,$06,$4F,$99,$F9,$5F,$03,$F2,$FD,$D9,$3E,$A7,$E7,$7B,$BF
	.dc.b	$3F,$DC,$FA,$1E,$EC,$AF,$7C,$74,$7C,$B8,$5E,$2F,$9D,$8F,$49,$F4
	.dc.b	$E3,$67,$E9,$63,$F6,$FE,$03,$EE,$7C,$0C,$BF,$97,$1F,$7F,$E0,$7F
	.dc.b	$01,$F0,$32,$7E,$7C,$7E,$0F,$E0,$24,$7D,$27,$F0,$DF,$03,$F8,$7F
	.dc.b	$80,$FC,$47,$C0,$4A,$FA,$F1,$D5,$FB,$38,$75,$7E,$7E,$1D,$1F,$A3
	.dc.b	$85,$D3,$F1,$58,$FF,$9D,$3F,$32,$FE,$C1,$D3,$FE,$93,$6D,$FF,$B5
	.dc.b	$1B,$47,$FB,$49,$FE,$CA,$27,$C1,$73,$F8,$7E,$9F,$05,$2F,$FC,$F1
	.dc.b	$48,$F0,$7C,$17,$87,$E0,$3C,$5F,$01,$2F,$71,$E5,$C6,$47,$27,$B3
	.dc.b	$D2,$FB,$3F,$E3,$3C,$3B,$3E,$0C,$3D,$DF,$01,$23,$C6,$FC,$1F,$80
	.dc.b	$EB,$1C,$70,$BA,$7F,$A4,$F0,$FE,$FE,$DC,$7B,$B6,$BB,$1E,$EF,$82
	.dc.b	$F9,$9F,$05,$23,$07,$C2,$F8,$1F,$A0,$47,$E2,$FC,$17,$D3,$F7,$FE
	.dc.b	$A7,$BD,$27,$E4,$C7,$D7,$F7,$BE,$C7,$BD,$27,$E6,$C7,$37,$E1,$61
	.dc.b	$73,$F5,$63,$D6,$7E,$9E,$3D,$5F,$91,$8F,$E0,$FE,$06,$47,$D8,$7F
	.dc.b	$0D,$F0,$5F,$87,$F8,$2F,$F5,$DF,$E6,$6F,$86,$A6,$6F,$FD,$6D,$1F
	.dc.b	$D6,$89,$F0,$7C,$3E,$1F,$8F,$C1,$72,$F8,$29,$7C,$23,$93,$FE,$6D
	.dc.b	$BE,$CF,$4C,$33,$BF,$FB,$1E,$7A,$CF,$83,$1E,$CF,$8B,$E0,$7C,$E4
	.dc.b	$7C,$BF,$05,$EA,$F8,$3F,$5F,$C2,$49,$F4,$C7,$BB,$E0,$FD,$FF,$09
	.dc.b	$27,$DB,$1F,$0B,$E0,$A4,$7C,$07,$E2,$FA,$FD,$63,$E1,$E1,$D5,$F6
	.dc.b	$61,$D5,$F8,$F8,$75,$7E,$06,$1F,$3B,$E0,$A4,$7C,$C7,$E8,$FC,$0F
	.dc.b	$58,$F8,$D8,$75,$DF,$E2,$63,$D6,$7E,$56,$3D,$67,$E4,$63,$F6,$BE
	.dc.b	$12,$57,$D8,$7E,$EF,$C2,$7D,$EF,$83,$FB,$FF,$07,$2B,$EE,$C7,$E0
	.dc.b	$BD,$F9,$1F,$80,$7F,$0B,$EF,$FF,$C0,$0F,$CD,$3F,$AD,$9F,$D3,$25
	.dc.b	$9A,$30,$D8,$F0,$1F,$8A,$C3,$C0,$FE,$0F,$0E,$EF,$E1,$70,$B5,$F8
	.dc.b	$3C,$3C,$1F,$09,$28,$7F,$AD,$F1,$FE,$D3,$5E,$1F,$84,$F3,$7C,$1C
	.dc.b	$AD,$CB,$D2,$B6,$AF,$67,$C2,$7B,$7E,$0F,$DD,$F0,$73,$1E,$38,$70
	.dc.b	$9F,$85,$F0,$32,$72,$7C,$2C,$51,$F1,$FE,$07,$C5,$3E,$7C,$AE,$7E
	.dc.b	$38,$E1,$9F,$85,$E7,$1F,$2F,$3F,$14,$7B,$F2,$FA,$5F,$03,$F4,$FE
	.dc.b	$06,$4F,$0D,$13,$F5,$FE,$0B,$EC,$7C,$1C,$AD,$47,$DB,$F8,$29,$1C
	.dc.b	$5F,$BD,$F0,$7F,$7F,$E0,$24,$F5,$7F,$05,$F0,$3D,$63,$C1,$87,$57
	.dc.b	$EA,$61,$D1,$FA,$D8,$5D,$BE,$01,$F8,$7C,$7F,$19,$F0,$3F,$8D,$F8
	.dc.b	$19,$7E,$B8,$EC,$FD,$FC,$7B,$5F,$2D,$59,$75,$E5,$FB,$B9,$DC,$3E
	.dc.b	$A6,$7F,$99,$F8,$29,$1F,$1D,$FC,$E7,$C1,$FE,$32,$3F,$37,$7F,$88
	.dc.b	$F5,$BF,$B3,$9F,$EC,$E2,$3B,$CF,$D7,$0F,$77,$13,$C7,$E0,$ED,$FE
	.dc.b	$9A,$BF,$A8,$CF,$5A,$CD,$A7,$FA,$FD,$78,$8E,$1E,$EE,$A3,$33,$E0
	.dc.b	$F8,$1C,$61,$9F,$87,$98,$7E,$2A,$FD,$38,$F1,$CA,$4E,$63,$A7,$C1
	.dc.b	$C8,$E0,$F7,$F8,$4F,$07,$C2,$78,$7E,$12,$57,$58,$F2,$7C,$1C,$8F
	.dc.b	$03,$E7,$F8,$3F,$46,$7E,$9F,$81,$95,$E5,$8E,$AF,$9F,$0E,$AF,$C1
	.dc.b	$E0,$F8,$98,$F5,$8C,$53,$1C,$57,$D2,$C7,$E3,$7C,$07,$C7,$F8,$19
	.dc.b	$87,$C3,$8F,$95,$F0,$52,$3E,$33,$F3,$7E,$0F,$E7,$7C,$1F,$CF,$F8
	.dc.b	$39,$5F,$2E,$3E,$97,$C1,$C8,$F9,$CF,$DA,$FC,$D7,$FA,$EC,$79,$B0
	.dc.b	$EC,$D3,$EA,$61,$D9,$FA,$D8,$76,$7D,$B8,$73,$F7,$E4,$7F,$04,$C3
	.dc.b	$FD,$96,$7F,$B2,$B5,$FC,$8E,$3D,$9B,$3D,$8F,$41,$D8,$87,$63,$A7
	.dc.b	$C1,$7A,$48,$F7,$F8,$3F,$67,$C0,$FB,$7E,$06,$4F,$8E,$3E,$07,$B9
	.dc.b	$23,$CC,$FC,$3F,$77,$E2,$7B,$92,$7D,$4F,$C7,$F7,$7A,$C7,$B7,$0E
	.dc.b	$AF,$E0,$75,$63,$D6,$31,$8D,$D3,$F8,$5D,$32,$FA,$5F,$03,$F4,$FE
	.dc.b	$06,$61,$F2,$63,$EB,$7C,$0F,$D7,$F8,$19,$3F,$36,$3E,$D7,$C0,$C8
	.dc.b	$FA,$0F,$DD,$F8,$1F,$BD,$F0,$1F,$7F,$E0,$65,$7D,$58,$EA,$FD,$7C
	.dc.b	$3A,$BE,$BC,$3A,$3E,$DC,$2E,$BE,$AC,$7F,$15,$F0,$3F,$8B,$F8,$19
	.dc.b	$7F,$81,$8F,$E0,$FF,$9B,$7F,$B3,$0F,$ED,$E4,$8C,$FC,$14,$7F,$A5
	.dc.b	$C7,$F6,$61,$F0,$5C,$BE,$03,$9F,$C0,$4A,$DC,$76,$7B,$FC,$3C,$79
	.dc.b	$9D,$4F,$DB,$5F,$5C,$3C,$B8,$79,$BE,$02,$57,$89,$F4,$FC,$0F,$58
	.dc.b	$F2,$E1,$D5,$E6,$75,$DB,$FA,$3E,$3D,$63,$D7,$87,$C2,$F8,$39,$1E
	.dc.b	$E7,$E2,$FC,$1F,$C6,$F8,$29,$3F,$09,$F9,$3F,$07,$F2,$BD,$DF,$97
	.dc.b	$EF,$CA,$F8,$F1,$F3,$BD,$EF,$9F,$EF,$49,$F9,$71,$D1,$F8,$38,$5D
	.dc.b	$7C,$F8,$F5,$9F,$A3,$8F,$57,$E3,$E3,$F6,$BE,$0F,$ED,$FC,$1C,$9F
	.dc.b	$AF,$1F,$7B,$E0,$BE,$FF,$C1,$C9,$FB,$71,$F8,$2F,$82,$FC,$1F,$C1
	.dc.b	$49,$FB,$F1,$FD,$33,$F3,$7F,$F5,$F8,$F8,$0C,$19,$F8,$1E,$D5,$F8
	.dc.b	$6C,$7B,$3F,$73,$1E,$CF,$E2,$B1,$EC,$FE,$03,$1E,$DF,$01,$DF,$E0
	.dc.b	$64,$FF,$86,$B3,$FE,$C8,$DF,$83,$F2,$11,$D7,$C1,$F9,$FE,$0F,$D1
	.dc.b	$F0,$72,$79,$47,$AF,$E0,$7D,$9F,$05,$27,$B4,$73,$7F,$A4,$63,$FD
	.dc.b	$9B,$83,$FC,$82,$D8,$FF,$2B,$AC,$7C,$11,$E7,$C7,$C0,$F2,$C7,$E5
	.dc.b	$7C,$0C,$8F,$63,$F3,$7E,$07,$C1,$D3,$FE,$1C,$C7,$B2,$69,$DE,$BD
	.dc.b	$67,$72,$1D,$CF,$91,$F0,$5F,$67,$E0,$E4,$60,$F9,$9F,$05,$F7,$7E
	.dc.b	$0E,$46,$0F,$A1,$EF,$7E,$04,$8F,$D3,$F8,$2F,$C2,$7C,$0C,$8F,$AC
	.dc.b	$FE,$1F,$E0,$BF,$11,$87,$49,$F9,$78,$61,$EA,$43,$E7,$E1,$F8,$FC
	.dc.b	$3F,$21,$F0,$72,$7F,$04,$FF,$87,$BC,$1F,$EC,$BB,$7F,$90,$97,$C1
	.dc.b	$E7,$E1,$E5,$FF,$98,$A7,$2D,$F8,$4E,$44,$75,$F0,$9D,$7E,$06,$47
	.dc.b	$17,$C1,$F0,$5E,$1C,$7A,$CF,$F8,$37,$0F,$ED,$5D,$5F,$E8,$67,$BC
	.dc.b	$67,$1E,$CD,$3B,$9E,$5F,$82,$F6,$7C,$0C,$8C,$1E,$9F,$84,$F8,$1F
	.dc.b	$09,$23,$07,$B7,$E0,$BE,$27,$C1,$C8,$C1,$F0,$7E,$03,$E4,$7C,$0C
	.dc.b	$8C,$1C,$8F,$55,$F5,$7D,$98,$7C,$FC,$3A,$47,$BF,$0C,$3F,$37,$E0
	.dc.b	$7E,$A7,$C0,$C8,$C1,$D7,$97,$EB,$63,$D5,$FB,$18,$F2,$9E,$87,$D7
	.dc.b	$C3,$13,$F7,$7E,$0B,$F0,$1F,$07,$2B,$EE,$47,$E0,$FE,$0A,$47,$DF
	.dc.b	$7F,$0D,$F0,$7F,$E6,$4F,$CE,$3F,$CE,$93,$FE,$43,$7F,$D9,$A4,$75
	.dc.b	$F0,$5B,$F7,$A4,$7F,$AA,$A7,$FD,$8B,$CC,$C3,$BC,$FE,$05,$EE,$EC
	.dc.b	$F0,$7C,$04,$70,$73,$F8,$3F,$21,$1E,$DF,$09,$E7,$F8,$49,$1E,$17
	.dc.b	$D5,$F0,$9E,$BF,$83,$93,$E5,$7D,$DF,$07,$EF,$F8,$19,$3E,$97,$E1
	.dc.b	$7C,$0F,$48,$F8,$58,$5E,$4F,$F5,$5E,$EC,$BA,$D7,$C7,$CA,$D6,$B2
	.dc.b	$F9,$9F,$03,$F3,$78,$25,$FC,$78,$FA,$1F,$05,$F4,$7E,$0E,$4F,$CB
	.dc.b	$8F,$A9,$F0,$52,$3E,$73,$F5,$FE,$0F,$EC,$7C,$0F,$D9,$F8,$29,$5F
	.dc.b	$4E,$3A,$BF,$57,$0E,$AF,$DD,$C3,$A3,$F7,$F0,$BA,$7B,$B1,$FC,$27
	.dc.b	$C0,$CA,$FB,$CF,$E1,$FE,$0B,$AC,$7E,$1F,$1E,$B7,$F1,$B1,$E7,$E4
	.dc.b	$A9,$2E,$26,$7A,$9C,$C3,$3F,$F5,$BF,$E7,$3F,$EF,$99,$FF,$4B,$36
	.dc.b	$FF,$DB,$AD,$C3,$B4,$23,$FB,$94,$FF,$BF,$9F,$EC,$AF,$2F,$82,$EA
	.dc.b	$76,$F8,$3E,$FF,$07,$27,$83,$E2,$F8,$1F,$1F,$C0,$49,$E8,$F9,$BE
	.dc.b	$07,$94,$77,$C3,$AB,$E1,$C3,$A3,$EA,$C2,$E5,$EA,$C7,$DF,$F0,$52
	.dc.b	$BD,$0F,$C2,$F8,$3F,$87,$F0,$72,$7D,$8F,$C6,$F8,$3F,$8F,$F0,$5F
	.dc.b	$23,$E0,$A5,$7C,$18,$F9,$7F,$01,$23,$E2,$3F,$3B,$E0,$3A,$47,$CE
	.dc.b	$C2,$E9,$F3,$B1,$EB,$3F,$2F,$1E,$AF,$D4,$C7,$EB,$FC,$07,$D8,$F8
	.dc.b	$09,$3F,$46,$3E,$DF,$C1,$7D,$CF,$82,$93,$F5,$63,$EF,$FC,$17,$FB
	.dc.b	$DB,$F3,$AF,$EB,$3D,$3F,$71,$B5,$F8,$AE,$95,$6F,$DF,$E1,$F0,$72
	.dc.b	$BF,$07,$87,$19,$F0,$C7,$E0,$F2,$F0,$BF,$8C,$CB,$C2,$F0,$3F,$2F
	.dc.b	$97,$83,$80,$DC,$FC,$33,$B9,$CF,$C0,$7A,$7E,$0E,$4E,$6D,$3E,$28
	.dc.b	$FC,$DE,$3E,$2B,$FC,$DE,$3E,$29,$FF,$7A,$5F,$84,$F0,$6B,$F1,$7E
	.dc.b	$0F,$E3,$7C,$1C,$AE,$D1,$F2,$7E,$0E,$47,$85,$F9,$9F,$07,$F3,$7D
	.dc.b	$D9,$3E,$57,$E8,$7B,$DF,$47,$E0,$64,$FA,$5F,$A9,$EF,$F4,$8F,$2E
	.dc.b	$17,$67,$27,$D7,$CB,$AD,$7A,$B2,$B5,$A8,$F9,$B9,$7D,$FF,$83,$91
	.dc.b	$F1,$DF,$C1,$7C,$1F,$E0,$FE,$0F,$F0,$9F,$07,$2B,$E6,$47,$E1,$FE
	.dc.b	$0B,$F1,$1F,$07,$27,$E8,$47,$E2,$FE,$02,$47,$D3,$7F,$1D,$F0,$7D
	.dc.b	$63,$EB,$61,$D5,$FA,$38,$75,$7E,$D6,$1D,$5F,$AB,$87,$F1,$D7,$F8
	.dc.b	$EE,$FF,$C4,$B3,$FE,$D7,$AF,$BC,$2F,$81,$E1,$19,$8C,$57,$E2,$DE
	.dc.b	$35,$CF,$E0,$F6,$D9,$D4,$61,$FE,$5B,$3F,$ED,$07,$97,$C1,$66,$FC
	.dc.b	$2D,$93,$7E,$0B,$1F,$01,$2F,$C4,$EF,$E0,$38,$63,$27,$C2,$79,$FE
	.dc.b	$0F,$A6,$12,$3C,$27,$AF,$0F,$07,$C0,$C8,$F7,$3E,$3F,$81,$E9,$1F
	.dc.b	$28,$F5,$E1,$8D,$3E,$36,$3D,$67,$51,$EC,$7B,$98,$D1,$8F,$BF,$E0
	.dc.b	$FE,$07,$C1,$CA,$F9,$D1,$F0,$FE,$0A,$47,$D1,$7E,$37,$C1,$FC,$7F
	.dc.b	$7B,$E4,$7B,$F2,$BE,$B4,$7C,$BF,$7F,$E6,$7B,$D2,$7E,$D4,$74,$7C
	.dc.b	$98,$5D,$7F,$01,$8F,$59,$FC,$06,$3D,$5F,$56,$3F,$5F,$E0,$A4,$7E
	.dc.b	$15,$FB,$5F,$07,$F6,$FE,$0F,$EE,$7C,$1C,$AF,$C5,$47,$E0,$7F,$3B
	.dc.b	$FE,$CD,$1F,$DC,$9F,$EC,$AF,$CA,$6E,$7F,$97,$76,$9F,$ED,$2E,$7E
	.dc.b	$07,$A1,$D7,$E0,$7C,$55,$F9,$9C,$3C,$4F,$D3,$C3,$C5,$D9,$E3,$F8
	.dc.b	$69,$3F,$CA,$72,$EE,$7F,$AF,$09,$F0,$5E,$9F,$82,$F5,$7C,$27,$AF
	.dc.b	$E1,$25,$F1,$8F,$77,$C2,$7B,$FE,$12,$4F,$58,$F8,$5F,$05,$23,$C0
	.dc.b	$FC,$5F,$82,$EB,$1E,$3C,$3A,$BF,$17,$0E,$AF,$9F,$0E,$AF,$C9,$C3
	.dc.b	$E7,$7C,$0C,$8F,$63,$F4,$7E,$0B,$AC,$7C,$FC,$3A,$F2,$7C,$FC,$7A
	.dc.b	$CF,$D3,$C7,$AC,$FD,$3C,$7E,$D7,$C2,$4A,$F9,$0F,$DD,$F8,$4F,$BD
	.dc.b	$F0,$72,$7E,$63,$F8,$1F,$84,$FC,$17,$BD,$F8,$3F,$7E,$57,$D1,$8F
	.dc.b	$C3,$7B,$DF,$C1,$FF,$3C,$FE,$C0,$7F,$61,$E6,$FE,$BF,$5F,$87,$C7
	.dc.b	$B3,$4F,$B5,$8F,$67,$EE,$63,$D9,$FC,$76,$3D,$7E,$13,$B7,$C2,$C9
	.dc.b	$FE,$D5,$97,$FB,$13,$7F,$E5,$0F,$F6,$88,$CF,$C1,$F9,$4F,$37,$C2
	.dc.b	$79,$FE,$0A,$4F,$27,$D5,$F0,$7E,$BF,$81,$F6,$7C,$0C,$AE,$F1,$CD
	.dc.b	$EA,$43,$F8,$6E,$2C,$3A,$C6,$30,$B9,$79,$B1,$F9,$1F,$03,$2B,$D6
	.dc.b	$FC,$BF,$81,$F9,$9F,$07,$27,$DE,$FC,$FF,$83,$FA,$1E,$A7,$D1,$F8
	.dc.b	$39,$5F,$12,$3E,$A7,$C0,$7D,$5F,$81,$93,$F2,$23,$A3,$E8,$C2,$E9
	.dc.b	$F2,$B1,$EB,$3F,$3B,$1E,$AF,$BB,$1F,$C0,$7C,$07,$E0,$7E,$06,$4F
	.dc.b	$D4,$8E,$7E,$57,$F9,$BB,$89,$9E,$2B,$FC,$FB,$0C,$BB,$C7,$CA,$CB
	.dc.b	$BE,$7F,$73,$3F,$C7,$7C,$1F,$E3,$FE,$0E,$5F,$E0,$A3,$FA,$C7,$E7
	.dc.b	$BF,$D8,$A3,$FD,$12,$DB,$34,$FE,$D3,$1F,$D1,$17,$BD,$CB,$DE,$93
	.dc.b	$FE,$37,$9C,$B7,$E0,$7B,$91,$D7,$BF,$E2,$61,$C4,$F2,$FC,$27,$9B
	.dc.b	$0E,$B1,$E1,$C3,$0F,$7F,$84,$F5,$91,$F1,$7C,$27,$BB,$E1,$24,$79
	.dc.b	$5F,$83,$F0,$9F,$0B,$E0,$E4,$FA,$5F,$8B,$F0,$9F,$1B,$E0,$A4,$FB
	.dc.b	$5F,$93,$F0,$5D,$23,$E2,$61,$78,$7F,$C2,$9C,$2C,$BA,$D7,$C9,$CA
	.dc.b	$D6,$B2,$FA,$7F,$03,$F5,$3E,$0A,$5F,$CB,$8F,$AF,$F0,$72,$3E,$73
	.dc.b	$F6,$BE,$13,$ED,$FC,$24,$9F,$A4,$FD,$EF,$84,$FB,$FF,$03,$F8,$0F
	.dc.b	$82,$95,$F5,$E3,$A3,$F7,$70,$BA,$7A,$B1,$E9,$3F,$81,$C6,$CF,$BB
	.dc.b	$1F,$F1,$17,$B9,$FD,$70,$FE,$BF,$E1,$7F,$B2,$1B,$7F,$ED,$67,$FB
	.dc.b	$0F,$0F,$FB,$5A,$3D,$64,$3F,$1D,$8F,$6A,$FF,$68,$CB,$1C,$57,$E6
	.dc.b	$30,$F0,$FC,$1F,$8B,$E1,$24,$FF,$A2,$72,$FE,$DB,$8F,$8F,$E0,$FC
	.dc.b	$FF,$07,$2B,$51,$EA,$F7,$FD,$7F,$01,$27,$94,$7B,$BD,$F9,$1D,$5F
	.dc.b	$83,$F0,$1F,$0A,$FA,$4F,$F8,$AB,$26,$39,$8E,$67,$97,$1E,$CE,$F1
	.dc.b	$F9,$7F,$07,$F3,$3E,$0E,$4F,$AE,$3E,$7F,$C1,$C8,$F7,$3F,$4B,$1F
	.dc.b	$A7,$F0,$72,$7E,$13,$F5,$BE,$0F,$EB,$FC,$0F,$D8,$F8,$19,$5F,$1E
	.dc.b	$3A,$3E,$9C,$2F,$47,$F9,$BF,$11,$D6,$7D,$D8,$D9,$D6,$3F,$84,$F8
	.dc.b	$09,$5F,$4D,$FC,$3F,$C0,$FE,$23,$E0,$FF,$CD,$5F,$9F,$7F,$B0,$CF
	.dc.b	$F6,$26,$5F,$D9,$31,$FF,$3F,$47,$F5,$9E,$66,$29,$4E,$63,$0D,$8C
	.dc.b	$FC,$17,$6F,$82,$95,$FE,$CD,$AF,$EC,$4F,$F9,$81,$FE,$A8,$FE,$27
	.dc.b	$1F,$2F,$C7,$48,$C1,$C7,$E1,$3D,$24,$7A,$61,$EC,$F8,$1F,$6F,$C0
	.dc.b	$C9,$F0,$47,$C1,$7C,$57,$D5,$FC,$46,$1C,$DF,$26,$17,$3F,$87,$8F
	.dc.b	$C8,$F8,$39,$5E,$B7,$E5,$FC,$27,$CC,$F8,$39,$3E,$F7,$E7,$FC,$1F
	.dc.b	$D0,$F7,$E4,$FC,$37,$E9,$FB,$DF,$53,$DD,$93,$F1,$DF,$AF,$EF,$74
	.dc.b	$8F,$91,$85,$D3,$EB,$E3,$D2,$7E,$66,$36,$7E,$DE,$3F,$80,$F8,$29
	.dc.b	$5F,$4D,$FC,$1F,$C1,$FE,$13,$E0,$BF,$0B,$F0,$72,$BE,$C4,$7E,$23
	.dc.b	$E0,$7F,$13,$F0,$52,$7E,$E4,$7F,$AD,$7F,$3F,$FE,$B8,$48,$C7,$C0
	.dc.b	$F5,$7F,$05,$87,$57,$F0,$58,$75,$7F,$0D,$87,$57,$F0,$D8,$75,$F8
	.dc.b	$1E,$DF,$03,$27,$F2,$BB,$7F,$B5,$73,$FE,$C7,$1F,$EA,$C7,$FD,$8A
	.dc.b	$EB,$BB,$71,$88,$D7,$C1,$FA,$7E,$0F,$D5,$F0,$7E,$BF,$83,$95,$C4
	.dc.b	$FD,$15,$A7,$E0,$7C,$0F,$C1,$F8,$19,$5B,$CD,$6A,$F8,$D9,$4A,$D4
	.dc.b	$F8,$EB,$D5,$97,$8A,$73,$97,$89,$F6,$65,$F3,$FE,$06,$47,$27,$E9
	.dc.b	$7C,$0F,$58,$F6,$9D,$B2,$F7,$3E,$0A,$E9,$8F,$C9,$C3,$C1,$3F,$27
	.dc.b	$E0,$FE,$E1,$1F,$3F,$C1,$FD,$FF,$82,$91,$EA,$7F,$05,$F0,$7F,$83
	.dc.b	$F8,$1F,$C2,$7C,$0C,$AF,$7C,$7E,$1F,$E0,$24,$7C,$27,$F1,$5F,$01
	.dc.b	$D2,$3E,$D6,$17,$4F,$BD,$8F,$59,$F9,$38,$F5,$7F,$01,$8F,$E4,$FE
	.dc.b	$0B,$F2,$9F,$07,$27,$E7,$C7,$F1,$C7,$FD,$E3,$FA,$07,$FB,$18,$FE
	.dc.b	$D3,$89,$A6,$7E,$0F,$84,$3A,$7F,$23,$2F,$83,$CC,$61,$B7,$FC,$A6
	.dc.b	$7F,$B2,$B7,$E0,$73,$7D,$FE,$06,$D7,$F5,$AA,$FC,$52,$AC,$E1,$AF
	.dc.b	$88,$95,$A8,$E3,$97,$2C,$A4,$F2,$8E,$BF,$03,$DB,$E0,$A4,$F6,$8F
	.dc.b	$0F,$C2,$78,$BE,$12,$4F,$8A,$3C,$BF,$07,$23,$CA,$FA,$3E,$13,$D3
	.dc.b	$F0,$5E,$AF,$7A,$57,$AA,$3A,$BE,$CC,$3A,$BF,$4B,$0E,$8F,$D4,$C2
	.dc.b	$F1,$7C,$F8,$F9,$58,$FC,$6F,$82,$91,$F1,$9F,$93,$F0,$5D,$A3,$EE
	.dc.b	$E3,$DA,$F9,$1F,$77,$1E,$BC,$1F,$6B,$2B,$7D,$8F,$A3,$F0,$7F,$50
	.dc.b	$8F,$D4,$F8,$4F,$AF,$F0,$9F,$63,$E1,$24,$FD,$88,$FB,$7E,$FC,$8F
	.dc.b	$B6,$FE,$03,$BD,$F0,$17,$E0,$25,$7F,$09,$9F,$ED,$83,$E0,$3B,$C7
	.dc.b	$E1,$31,$EE,$FE,$6B,$1E,$CF,$E7,$31,$BB,$FC,$0C,$3B,$CF,$5F,$84
	.dc.b	$F1,$7C,$3C,$8C,$1E,$1F,$83,$F3,$11,$F2,$7C,$27,$A7,$E1,$24,$79
	.dc.b	$DF,$67,$C2,$7B,$7E,$0B,$DD,$F0,$52,$BD,$91,$CD,$FF,$06,$73,$7F
	.dc.b	$3A,$D1,$96,$75,$F4,$67,$DA,$F6,$7F,$84,$E9,$9E,$6B,$DD,$9F,$CD
	.dc.b	$F8,$2F,$9D,$F0,$53,$0F,$99,$1F,$47,$E1,$3E,$97,$C2,$49,$FA,$11
	.dc.b	$F5,$7E,$13,$EB,$7C,$24,$9F,$A9,$1F,$67,$E0,$A4,$7D,$77,$EE,$7C
	.dc.b	$1F,$58,$FB,$58,$75,$7D,$3A,$36,$65,$DE,$7F,$01,$97,$77,$72,$CB
	.dc.b	$F1,$1F,$05,$27,$F0,$CF,$E2,$FE,$0F,$C1,$1C,$32,$F0,$65,$F8,$1C
	.dc.b	$BB,$4F,$4C,$BB,$4F,$E1,$71,$FE,$27,$FA,$0F,$FB,$54,$95,$8F,$85
	.dc.b	$CF,$C2,$EB,$E1,$64,$FF,$5F,$AF,$EE,$38,$7E,$B9,$FE,$B9,$CF,$FD
	.dc.b	$D9,$B3,$FE,$4B,$6C,$97,$C1,$78,$BD,$FF,$1F,$BF,$2F,$B4,$78,$9F
	.dc.b	$06,$3E,$27,$E6,$DF,$8B,$1F,$07,$C0,$7B,$09,$3E,$7C,$7D,$FF,$09
	.dc.b	$F0,$3E,$16,$4F,$AE,$3E,$1F,$C2,$7C,$4F,$84,$93,$EF,$8F,$8F,$F0
	.dc.b	$92,$3E,$13,$F2,$BE,$13,$E5,$F4,$FC,$CF,$82,$95,$F1,$E3,$AB,$F2
	.dc.b	$70,$EA,$FA,$B0,$E8,$FB,$30,$BC,$5F,$37,$1F,$AF,$F0,$3F,$63,$E0
	.dc.b	$A5,$FD,$38,$FB,$7F,$09,$23,$EB,$3F,$7B,$E1,$3E,$FF,$C1,$C9,$FB
	.dc.b	$4F,$E0,$BE,$13,$FA,$AF,$E8,$5F,$CC,$67,$F3,$2F,$F6,$62,$72,$F5
	.dc.b	$8F,$A3,$85,$D3,$F0,$58,$F5,$9F,$AB,$8D,$9F,$C3,$63,$DB,$E0,$3B
	.dc.b	$FC,$0C,$BF,$F1,$56,$7F,$DA,$2F,$F0,$D9,$75,$CF,$F1,$8F,$AE,$BD
	.dc.b	$24,$3B,$9F,$95,$C7,$D7,$F0,$72,$3A,$BE,$EF,$85,$F7,$FC,$14,$9F
	.dc.b	$0B,$F0,$BE,$0F,$E1,$FB,$BF,$13,$DE,$95,$E6,$8F,$8F,$EF,$48,$F4
	.dc.b	$BF,$2B,$E0,$7A,$47,$CA,$C2,$ED,$FD,$53,$76,$5D,$AB,$E0,$E5,$D9
	.dc.b	$FA,$39,$7D,$4F,$82,$FA,$BF,$07,$27,$E3,$C7,$D8,$F8,$3F,$B3,$F0
	.dc.b	$72,$7E,$5C,$7D,$CF,$83,$FB,$BF,$07,$27,$E7,$C7,$E0,$3D,$F9,$1F
	.dc.b	$49,$FC,$1F,$C0,$75,$8F,$AB,$87,$57,$F0,$B8,$75,$7E,$CE,$1D,$5F
	.dc.b	$C4,$E1,$F8,$CF,$80,$91,$F7,$9F,$F1,$1F,$E8,$7F,$D5,$A3,$F9,$3B
	.dc.b	$FE,$CD,$3F,$B1,$3F,$E6,$D6,$BA,$F8,$3E,$3F,$05,$23,$4F,$4F,$82
	.dc.b	$EB,$F0,$12,$79,$3E,$0F,$80,$F0,$DF,$8B,$E0,$E5,$77,$8E,$8F,$93
	.dc.b	$0B,$CF,$FE,$0C,$CB,$FD,$B7,$3E,$8F,$81,$F5,$FC,$0C,$8F,$44,$76
	.dc.b	$7D,$58,$F6,$7D,$47,$A0,$87,$73,$E0,$69,$F1,$BE,$0A,$47,$C3,$7E
	.dc.b	$4F,$C1,$FC,$AF,$82,$F9,$7F,$07,$2B,$E4,$47,$CE,$F7,$BE,$7F,$BF
	.dc.b	$27,$E6,$47,$D2,$F7,$A4,$7C,$F7,$EA,$FB,$FD,$63,$E9,$61,$D5,$FA
	.dc.b	$F8,$75,$7E,$B6,$1D,$5F,$B7,$87,$DE,$F8,$3F,$BF,$F0,$72,$7E,$E4
	.dc.b	$7E,$0B,$E0,$FF,$07,$F0,$72,$7F,$01,$1F,$E6,$8C,$3F,$CF,$51,$FD
	.dc.b	$B1,$BC,$9F,$E3,$E6,$F9,$BF,$EC,$65,$F0,$7C,$7E,$0E,$5F,$FA,$DE
	.dc.b	$7F,$A4,$2B,$43,$E0,$7C,$93,$F9,$1C,$BC,$8F,$E5,$B2,$F2,$30,$E8
	.dc.b	$79,$F1,$F4,$7C,$04,$8F,$03,$EB,$F8,$0F,$67,$C1,$7B,$7E,$0E,$57
	.dc.b	$96,$3E,$07,$C0,$C8,$F4,$3F,$0F,$E0,$BE,$27,$BF,$F1,$7E,$02,$57
	.dc.b	$B6,$3A,$BE,$FC,$3A,$BF,$27,$0E,$8F,$CB,$C2,$F0,$7C,$BC,$7E,$87
	.dc.b	$C0,$4A,$F9,$0F,$D3,$F8,$0E,$B1,$F4,$F1,$EB,$7F,$4F,$1E,$7C,$9F
	.dc.b	$4F,$2B,$7F,$3F,$8D,$9F,$DD,$F8,$1F,$BD,$F0,$52,$BE,$B4,$7E,$07
	.dc.b	$E0,$64,$7D,$97,$F0,$9F,$05,$F8,$5F,$7A,$4F,$DD,$7F,$11,$F0,$5F
	.dc.b	$89,$F8,$29,$3F,$81,$7F,$19,$F0,$1D,$23,$F1,$98,$5E,$0F,$C6,$63
	.dc.b	$D2,$7F,$21,$8D,$9F,$C8,$63,$FE,$BC,$FD,$13,$FB,$44,$7F,$A4,$1B
	.dc.b	$C0,$BF,$C7,$46,$1B,$F0,$DC,$7E,$12,$46,$EF,$94,$75,$F8,$4E,$DF
	.dc.b	$09,$DF,$E1,$25,$EF,$2E,$19,$7F,$B2,$63,$07,$83,$E0,$BC,$DF,$03
	.dc.b	$23,$39,$A9,$5E,$29,$FF,$2C,$C7,$F7,$32,$65,$E2,$8E,$31,$FE,$D9
	.dc.b	$77,$8E,$5E,$3F,$01,$F1,$48,$F0,$F8,$1F,$91,$F0,$9F,$27,$E1,$24
	.dc.b	$F4,$8F,$99,$F0,$52,$3B,$BF,$3F,$E0,$FE,$87,$C0,$49,$F1,$BF,$4F
	.dc.b	$E0,$7A,$C7,$9B,$0E,$AF,$B3,$0E,$8F,$BB,$0B,$BF,$D9,$C7,$EE,$7C
	.dc.b	$04,$AF,$7B,$F7,$FE,$07,$AC,$7C,$6C,$7A,$DF,$E0,$71,$E7,$C7,$F5
	.dc.b	$32,$B7,$F8,$2C,$BF,$11,$F0,$7F,$89,$F8,$39,$3F,$32,$3F,$19,$F0
	.dc.b	$72,$3E,$7B,$F8,$B7,$FD,$1B,$FA,$2F,$F6,$9C,$FF,$DB,$EB,$E0,$B5
	.dc.b	$18,$F7,$AE,$9F,$8D,$AF,$ED,$73,$C2,$B3,$7F,$C8,$E5,$F0,$58,$A4
	.dc.b	$62,$7F,$12,$D7,$C7,$EF,$F9,$DF,$C4,$E5,$6B,$CB,$F0,$FC,$EF,$CD
	.dc.b	$5C,$B3,$C1,$9C,$B5,$F0,$B2,$7B,$BC,$7E,$17,$97,$C1,$C9,$F1,$BD
	.dc.b	$7E,$13,$B7,$C2,$49,$F3,$BE,$1F,$84,$F1,$7C,$0F,$8F,$E0,$A5,$7B
	.dc.b	$23,$A3,$F3,$B0,$BB,$FC,$EC,$7A,$4F,$D2,$C6,$CF,$D2,$C7,$DD,$F0
	.dc.b	$5E,$FF,$4A,$5F,$C8,$8F,$85,$F0,$92,$3E,$5B,$F1,$7E,$13,$E3,$7C
	.dc.b	$1C,$9F,$9E,$FC,$9F,$84,$F9,$5F,$03,$F2,$FE,$0A,$57,$D4,$8E,$8F
	.dc.b	$E1,$B0,$BA,$7E,$1B,$1E,$93,$F8,$AC,$6C,$FE,$2B,$1F,$C0,$7E,$8D
	.dc.b	$FD,$B2,$BF,$B3,$3F,$DB,$1F,$E1,$2D,$7F,$DB,$EE,$E7,$F0,$F8,$78
	.dc.b	$3B,$3F,$20,$FB,$6B,$D6,$43,$C4,$7E,$32,$FC,$3F,$0B,$23,$BB,$E4
	.dc.b	$F8,$5F,$2F,$C2,$79,$BE,$12,$57,$92,$3D,$3E,$FC,$8F,$3B,$EC,$F8
	.dc.b	$0F,$6F,$C0,$7B,$BE,$02,$57,$B2,$3A,$BE,$EC,$3A,$BE,$EC,$3A,$3F
	.dc.b	$03,$0B,$B3,$6F,$E7,$5B,$B3,$F9,$7F,$07,$30,$F9,$2F,$CE,$F8,$4F
	.dc.b	$9F,$F0,$92,$7E,$6B,$F4,$BE,$13,$E9,$FC,$1F,$D4,$F8,$49,$5F,$4A
	.dc.b	$3E,$BF,$C1,$7D,$8F,$82,$93,$F5,$A3,$A3,$F5,$B0,$BB,$79,$E3,$EE
	.dc.b	$E3,$D9,$FB,$B8,$F6,$7E,$EE,$3F,$84,$F8,$2F,$C2,$FC,$14,$9F,$C1
	.dc.b	$C7,$E2,$3E,$12,$47,$E1,$9F,$C5,$FC,$27,$F0,$BF,$6B,$FB,$16,$65
	.dc.b	$F0,$99,$35,$C9,$BF,$84,$98,$FF,$58,$AF,$F6,$30,$F8,$1E,$F1,$F9
	.dc.b	$2C,$AE,$3E,$D1,$CC,$CF,$AD,$73,$CF,$C0,$F5,$CF,$CD,$F0,$9E,$7F
	.dc.b	$81,$93,$E1,$8E,$B3,$F7,$B2,$B3,$FA,$6C,$3A,$BF,$E4,$D9,$FE,$F9
	.dc.b	$A7,$BA,$3C,$07,$73,$D9,$F0,$7F,$14,$8F,$BF,$E1,$3E,$47,$C1,$FC
	.dc.b	$9F,$83,$93,$F0,$E3,$E6,$7B,$F2,$3E,$33,$F3,$FE,$03,$E8,$7C,$04
	.dc.b	$9F,$94,$FD,$3F,$80,$EB,$1F,$37,$0E,$AF,$A4,$FF,$33,$66,$C7,$B4
	.dc.b	$67,$8F,$ED,$65,$87,$E8,$53,$2F,$C0,$32,$BE,$D3,$F8,$3F,$83,$FC
	.dc.b	$27,$C1,$C9,$FB,$CF,$E1,$FE,$0F,$F1,$1F,$05,$27,$F0,$4F,$E2,$FE
	.dc.b	$0F,$F1,$9F,$03,$27,$F0,$CF,$F5,$9F,$D1,$FF,$B3,$72,$7D,$2C,$7B
	.dc.b	$57,$A7,$34,$77,$8E,$A4,$35,$8F,$77,$FC,$53,$7F,$D6,$9F,$84,$DD
	.dc.b	$7F,$20,$DC,$3F,$C7,$8F,$F7,$06,$FC,$27,$94,$95,$D3,$E1,$3D,$1F
	.dc.b	$01,$23,$BB,$EB,$F8,$0F,$67,$BD,$ED,$F7,$A5,$79,$23,$9B,$E4,$C2
	.dc.b	$E7,$E4,$C7,$A4,$FA,$31,$B3,$E8,$C7,$E4,$7B,$DF,$27,$DE,$97,$F0
	.dc.b	$23,$9E,$9E,$A8,$F8,$79,$7C,$FC,$B9,$CF,$C0,$CA,$D7,$FA,$BF,$56
	.dc.b	$5F,$57,$DF,$FA,$DF,$01,$2B,$E6,$C7,$D9,$F7,$FE,$D7,$C0,$49,$FA
	.dc.b	$31,$F7,$7D,$C9,$1F,$51,$FC,$07,$BD,$F8,$1F,$7A,$4F,$D8,$7B,$3F
	.dc.b	$62,$BA,$3F,$6A,$FC,$0F,$32,$1F,$7A,$FA,$BF,$82,$F7,$BF,$1D,$EF
	.dc.b	$C8,$C1,$F8,$6F,$80,$FF,$15,$7E,$91,$FD,$94,$FE,$CD,$BA,$33,$F0
	.dc.b	$FA,$F8,$29,$7F,$E7,$6B,$FE,$DD,$D1,$FC,$6D,$FF,$6B,$46,$7E,$0B
	.dc.b	$A9,$E1,$7F,$27,$87,$81,$87,$8C,$E9,$87,$67,$F7,$55,$FE,$E4,$B3
	.dc.b	$E0,$E5,$FF,$4A,$2F,$7F,$D6,$4A,$F3,$7C,$07,$BB,$E0,$3D,$FF,$01
	.dc.b	$27,$D5,$1F,$0B,$E0,$64,$7B,$5F,$8B,$F0,$5F,$1B,$E0,$BE,$3F,$C1
	.dc.b	$4A,$F8,$51,$D5,$F8,$98,$75,$7D,$38,$74,$7D,$78,$5D,$7E,$46,$3F
	.dc.b	$4B,$E0,$25,$7C,$D7,$B3,$C8,$EA,$43,$07,$43,$D0,$43,$E4,$C7,$DE
	.dc.b	$F8,$0F,$BF,$EF,$C9,$FB,$51,$F8,$2F,$80,$91,$F7,$5F,$C2,$FC,$0F
	.dc.b	$E1,$BE,$0A,$4F,$E0,$5F,$C4,$FC,$0F,$38,$FA,$FF,$05,$F8,$C3,$1A
	.dc.b	$7E,$17,$0E,$B3,$F7,$1E,$AF,$32,$1F,$80,$BE,$AF,$E3,$FD,$CF,$CB
	.dc.b	$91,$FC,$97,$B9,$FE,$83,$FD,$27,$FB,$C9,$3A,$31,$F0,$19,$F8,$3D
	.dc.b	$7C,$1C,$C3,$FD,$9F,$69,$E5,$F0,$3C,$FE,$0A,$56,$A3,$B6,$32,$38
	.dc.b	$BD,$9E,$75,$D5,$FE,$A5,$7F,$E0,$37,$27,$83,$E0,$BD,$04,$7C,$7F
	.dc.b	$05,$EB,$F7,$E4,$79,$9F,$77,$C0,$7B,$FE,$02,$4F,$A9,$F8,$5F,$01
	.dc.b	$F0,$FE,$0B,$E2,$7C,$14,$AF,$7C,$74,$7C,$B8,$5E,$6F,$F0,$A6,$8C
	.dc.b	$BA,$D7,$AB,$2B,$5A,$CB,$A4,$FB,$BD,$EF,$A4,$62,$79,$E3,$D7,$75
	.dc.b	$25,$97,$F7,$98,$FA,$79,$76,$AD,$D6,$B2,$C5,$71,$BF,$EF,$4F,$DA
	.dc.b	$F8,$2F,$C1,$11,$FB,$BE,$FF,$E1,$7E,$0B,$F0,$DF,$01,$27,$F0,$31
	.dc.b	$F8,$9F,$82,$FC,$57,$C1,$49,$FC,$2C,$7F,$B3,$FB,$7F,$BB,$C7,$A5
	.dc.b	$BA,$1B,$9B,$3F,$05,$7E,$3B,$2F,$03,$F8,$4C,$BC,$0F,$E4,$B2,$F0
	.dc.b	$3F,$88,$CB,$C1,$F0,$92,$3E,$1E,$DF,$13,$3F,$EF,$51,$E3,$F8,$5F
	.dc.b	$37,$C0,$48,$D9,$E7,$7D,$5F,$01,$EB,$F8,$09,$3A,$E0,$F5,$47,$BF
	.dc.b	$E0,$3E,$07,$C1,$CA,$D6,$48,$F2,$47,$F8,$BE,$3F,$DE,$17,$FE,$5D
	.dc.b	$6E,$7E,$7C,$BC,$35,$CB,$1F,$0B,$C5,$F0,$3D,$BE,$13,$E8,$11,$F0
	.dc.b	$FC,$1F,$D3,$F8,$4F,$A9,$F0,$72,$7C,$B1,$F5,$FE,$13,$EC,$7C,$14
	.dc.b	$9F,$4C,$7D,$BF,$82,$FB,$9F,$05,$27,$DB,$1D,$5F,$7E,$1D,$5F,$76
	.dc.b	$1D,$5F,$87,$87,$57,$E1,$61,$F8,$7F,$84,$91,$F2,$1E,$CF,$23,$A9
	.dc.b	$0C,$1C,$8E,$A4,$30,$7C,$FF,$84,$FC,$A1,$1F,$AB,$F0,$9F,$97,$F8
	.dc.b	$4F,$CC,$7C,$0C,$9F,$B3,$1F,$B2,$7F,$CC,$DF,$A5,$7F,$B9,$4F,$ED
	.dc.b	$8F,$ED,$5B,$CA,$75,$F0,$9C,$63,$FD,$35,$1F,$E0,$09,$E7,$F0,$99
	.dc.b	$E6,$FD,$CC,$FF,$79,$E7,$63,$E3,$BF,$C4,$65,$E7,$7F,$25,$5E,$3F
	.dc.b	$84,$E5,$1E,$36,$E5,$E3,$8B,$3F,$E0,$17,$8F,$C1,$60,$CF,$C1,$EB
	.dc.b	$E0,$24,$F6,$78,$FC,$0F,$2F,$81,$E7,$F0,$12,$BC,$71,$DB,$E0,$A4
	.dc.b	$79,$9E,$CF,$32,$1F,$32,$FA,$BC,$8F,$6D,$75,$7E,$8D,$7A,$BE,$0F
	.dc.b	$D7,$F0,$72,$7E,$1C,$7B,$BE,$0B,$DF,$F0,$72,$7E,$3C,$7C,$2E,$39
	.dc.b	$1F,$29,$F8,$BF,$07,$F1,$BE,$06,$4F,$CE,$7B,$3F,$82,$AE,$8F,$E1
	.dc.b	$2F,$C0,$F3,$21,$F8,$8B,$EA,$F3,$3E,$CF,$C1,$FD,$43,$BB,$D4,$FB
	.dc.b	$B3,$DB,$8F,$F1,$37,$E8,$9E,$87,$68,$FC,$6E,$3F,$7B,$E0,$FE,$FF
	.dc.b	$C1,$CA,$FC,$3C,$7E,$0B,$E0,$FF,$0B,$FA,$5F,$FB,$E8,$FF,$7D,$F3
	.dc.b	$FF,$BB,$6B,$F8,$F4,$7F,$B7,$DF,$F6,$33,$71,$FF,$7A,$37,$AF,$FA
	.dc.b	$8B,$FE,$05,$9F,$CA,$E3,$E1,$7F,$52,$F9,$5F,$CC,$E1,$E2,$A9,$18
	.dc.b	$3C,$3F,$09,$E6,$23,$E4,$F8,$1F,$4F,$C2,$48,$F3,$BE,$CF,$83,$F6
	.dc.b	$FC,$24,$9F,$5B,$F0,$3E,$13,$E0,$FC,$24,$9F,$7B,$F1,$3E,$03,$A4
	.dc.b	$7C,$0F,$80,$F8,$E6,$38,$3E,$0E,$1D,$67,$E2,$BD,$5E,$64,$3E,$4D
	.dc.b	$F5,$7E,$6F,$C0,$7D,$2F,$82,$91,$83,$E8,$FC,$1F,$D6,$F8,$49,$18
	.dc.b	$3E,$AF,$C2,$7D,$AF,$82,$91,$83,$EC,$DF,$47,$EC,$BF,$80,$C3,$F0
	.dc.b	$38,$74,$9F,$BB,$86,$1E,$47,$E0,$2F,$AB,$F8,$2C,$3A,$67,$F8,$2F
	.dc.b	$84,$FC,$59,$8E,$4F,$C1,$E3,$D6,$7F,$0F,$8F,$5C,$39,$1F,$8E,$C7
	.dc.b	$AC,$FE,$33,$1F,$CA,$FC,$14,$AF,$C9,$BF,$98,$F8,$3F,$CC,$FC,$14
	.dc.b	$9F,$CB,$BF,$9C,$F8,$3F,$E4,$3F,$A6,$7F,$84,$0F,$F0,$0E,$26,$9F
	.dc.b	$E1,$4E,$C6,$3C,$3E,$0E,$63,$E0,$3A,$90,$FD,$1E,$3D,$DF,$F0,$25
	.dc.b	$F7,$F8,$4F,$09,$E2,$F8,$AF,$1F,$C1,$4A,$DB,$E6,$F8,$5F,$3F,$C1
	.dc.b	$C9,$E6,$FA,$BE,$17,$D7,$F0,$BE,$CF,$85,$95,$E0,$8E,$8F,$F9,$1F
	.dc.b	$9F,$FC,$87,$A3,$E1,$7E,$19,$AE,$24,$EB,$2F,$05,$F1,$3F,$CA,$16
	.dc.b	$8F,$14,$77,$3D,$F8,$F8,$9E,$17,$F3,$71,$FA,$3F,$05,$27,$E2,$BF
	.dc.b	$53,$E0,$FE,$AF,$C1,$49,$F9,$2F,$D8,$F8,$3F,$B3,$F0,$92,$7E,$6B
	.dc.b	$F7,$3E,$17,$A4,$78,$35,$7C,$2F,$E0,$0C,$E9,$EE,$9F,$5E,$3D,$DE
	.dc.b	$0F,$77,$A9,$0E,$98,$77,$7D,$27,$DE,$AE,$CE,$0E,$67,$A8,$87,$97
	.dc.b	$1F,$53,$D0,$ED,$7E,$53,$F2,$98,$FE,$63,$E1,$7F,$33,$F0,$B2,$BF
	.dc.b	$1B,$1F,$9C,$F8,$3F,$CE,$FC,$1C,$9F,$C8,$C7,$F8,$02,$FF,$C3,$11
	.dc.b	$F5,$1B,$87,$76,$F1,$FF,$A2,$1F,$F0,$CE,$CC,$7C,$17,$F3,$72,$F2
	.dc.b	$3D,$8F,$CF,$E5,$E1,$7E,$96,$5E,$16,$1F,$E6,$7C,$3B,$BF,$E9,$B7
	.dc.b	$FC,$4E,$3E,$17,$D0,$48,$E1,$F0,$BE,$BF,$83,$F6,$7C,$24,$9E,$91
	.dc.b	$EF,$F8,$2F,$81,$F0,$92,$7C,$11,$CD,$F8,$58,$5E,$0F,$85,$8F,$59
	.dc.b	$F3,$63,$D5,$F9,$18,$F5,$7D,$58,$F5,$7E,$66,$3F,$3F,$E0,$A4,$7B
	.dc.b	$DF,$A5,$F0,$7F,$4F,$E1,$24,$FC,$37,$EB,$7C,$1F,$D7,$F8,$4F,$B1
	.dc.b	$F0,$12,$BE,$44,$74,$7E,$D6,$17,$6F,$B5,$8F,$49,$FB,$D8,$D9,$FB
	.dc.b	$D8,$F5,$9F,$A5,$8F,$57,$F0,$98,$F3,$E6,$FC,$16,$56,$7E,$B6,$5C
	.dc.b	$E7,$F0,$D9,$59,$FB,$58,$F5,$7F,$1D,$8F,$5C,$BF,$1D,$97,$E4,$FE
	.dc.b	$07,$F2,$9F,$07,$2F,$F0,$91,$F9,$7F,$82,$91,$F8,$77,$F3,$5F,$09
	.dc.b	$F9,$BF,$84,$93,$F8,$B7,$F3,$DF,$09,$D6,$3F,$1D,$87,$57,$F4,$18
	.dc.b	$74,$7F,$45,$85,$CB,$F4,$58,$F4,$9F,$D3,$63,$67,$F4,$D8,$FE,$AF
	.dc.b	$E0,$E5,$7E,$6D,$FD,$77,$C1,$FF,$9B,$BF,$4D,$FE,$1D,$3F,$C8,$9B
	.dc.b	$7F,$BD,$6F,$0F,$FB,$D9,$FF,$14,$4F,$FB,$F5,$FF,$14,$3C,$7E,$0B
	.dc.b	$A1,$D7,$E0,$FB,$7C,$1C,$9F,$F6,$95,$23,$C5,$F0,$1E,$3F,$83,$95
	.dc.b	$A7,$CD,$F0,$3E,$7F,$83,$93,$C9,$F5,$7C,$1F,$AF,$E0,$BD,$9F,$07
	.dc.b	$2B,$BC,$7B,$FE,$0B,$E0,$7C,$14,$9F,$1C,$7C,$3F,$81,$F8,$9F,$05
	.dc.b	$27,$CF,$1D,$5F,$4E,$1D,$5F,$91,$87,$57,$DB,$87,$57,$E6,$61,$D5
	.dc.b	$F8,$38,$75,$7E,$86,$1F,$4F,$E0,$E4,$7C,$67,$EB,$7C,$07,$D7,$F8
	.dc.b	$39,$3F,$29,$FB,$5F,$01,$F6,$FE,$0F,$FD,$53,$FA,$73,$E0,$69,$9B
	.dc.b	$1F,$D8,$23,$FC,$1A,$4F,$82,$E3,$F0,$7C,$BE,$0E,$57,$0E,$DF,$E8
	.dc.b	$17,$FF,$00,$21,$08,$7C,$1F,$8B,$E0,$65,$71,$EF,$E5,$5F,$DD,$63
	.dc.b	$07,$25,$F0,$7E,$AF,$81,$F5,$FC,$14,$AD,$E5,$DE,$7D,$FF,$07,$27
	.dc.b	$38,$7F,$82,$5B,$F0,$7F,$13,$E0,$FE,$2F,$C1,$C9,$DC,$7C,$8F,$81
	.dc.b	$91,$C9,$F9,$7F,$07,$F3,$3D,$D9,$3D,$9F,$9F,$F0,$7F,$43,$E0,$A4
	.dc.b	$F8,$9F,$A7,$F0,$5F,$53,$E0,$7E,$AF,$C1,$4A,$F3,$C7,$FB,$83,$F4
	.dc.b	$FF,$E9,$FC,$46,$93,$E2,$35,$F0,$9B,$F8,$49,$87,$C4,$AF,$EA,$8D
	.dc.b	$F8,$3E,$84,$75,$F0,$9D,$FE,$0E,$47,$17,$C5,$F0,$9E,$3F,$84,$F2
	.dc.b	$7C,$24,$AE,$D1,$E7,$F8,$49,$1E,$17,$D5,$F0,$9E,$BF,$83,$93,$E5
	.dc.b	$7D,$DF,$09,$EF,$F8,$3F,$81,$F0,$92,$BD,$51,$F0,$FE,$12,$47,$B5
	.dc.b	$F8,$DF,$09,$F9,$A8,$FF,$20,$0F,$D0,$BF,$40,$FE,$30,$C7,$9B,$BE
	.dc.b	$7F,$E3,$F3,$A9,$FF,$20,$F2,$3E,$17,$E9,$7C,$2F,$D3,$F8,$59,$5F
	.dc.b	$36,$3E,$B7,$C2,$C8,$FA,$0F,$5F,$33,$F4,$DE,$C7,$F2,$89,$FD,$7F
	.dc.b	$E1,$4F,$73,$51,$86,$A8,$D3,$E9,$D3,$CD,$AF,$F8,$BE,$FF,$98,$A7
	.dc.b	$F9,$8B,$9F,$85,$EA,$E7,$E2,$6C,$BE,$1E,$93,$0F,$D5,$C7,$F3,$37
	.dc.b	$E4,$63,$E2,$7E,$6B,$6F,$FA,$3B,$73,$FA,$59,$79,$9F,$83,$F0,$BE
	.dc.b	$02,$5F,$6F,$85,$F1,$FC,$2C,$8F,$0B,$E6,$F8,$5F,$87,$8F,$59,$F8
	.dc.b	$C7,$93,$7F,$8F,$97,$78,$FE,$23,$A3,$E1,$7D,$E4,$8F,$6D,$7C,$2F
	.dc.b	$85,$91,$F0,$1F,$8B,$F0,$9E,$09,$93,$DC,$FA,$5D,$1C,$1F,$F4,$3E
	.dc.b	$0B,$C7,$07,$D5,$F8,$5F,$9F,$F0,$B2,$FE,$6B,$F4,$AF,$AC,$7C,$CC
	.dc.b	$6D,$FE,$53,$1E,$13,$DA,$EC,$FE,$07,$E1,$7E,$DF,$C2,$CA,$FB,$2F
	.dc.b	$DE,$F8,$5E,$7C,$BF,$A5,$E1,$F9,$55,$F8,$DC,$FE,$4E,$7C,$63,$D7
	.dc.b	$85,$CB,$F1,$DF,$0B,$F8,$8F,$85,$95,$F8,$67,$F1,$BF,$A8,$7D,$18
	.dc.b	$E8,$7C,$73,$D8,$7C,$83,$81,$0F,$93,$1C,$BE,$1B,$9F,$C3,$74,$F8
	.dc.b	$69,$5F,$CF,$AD,$3E,$C8,$FC,$19,$D8,$FE,$03,$CF,$F8,$3C,$95,$FD
	.dc.b	$4A,$79,$7C,$37,$A0,$F4,$FC,$37,$AB,$E1,$A4,$F7,$7D,$BF,$0B,$DE
	.dc.b	$3D,$A7,$7C,$3E,$B3,$B2,$1F,$EC,$AC,$3F,$AC,$BE,$9F,$86,$F8,$FF
	.dc.b	$0D,$23,$D8,$FC,$AF,$85,$ED,$CB,$F5,$8F,$AF,$AF,$AF,$0F,$EB,$57
	.dc.b	$D6,$F9,$3D,$FE,$13,$E9,$FC,$2C,$8F,$90,$FD,$6F,$85,$EB,$D1,$EB
	.dc.b	$C7,$A4,$EE,$BD,$0F,$DC,$F8,$5F,$BB,$F0,$B2,$7E,$93,$F8,$0F,$85
	.dc.b	$E7,$1F,$7F,$83,$E1,$BF,$06,$78,$34,$FB,$5F,$0B,$F8,$8F,$85,$91
	.dc.b	$F7,$5F,$C5,$FC,$27,$48,$F9,$DD,$FF,$C3,$3B,$F7,$1F,$36,$AE,$3F
	.dc.b	$ED,$1E,$27,$C2,$FF,$20,$FD,$47,$FB,$49,$2F,$1F,$0D,$D6,$FF,$8C
	.dc.b	$F1,$7E,$8A,$7F,$C1,$75,$B6,$9C,$FE,$1B,$A7,$C3,$48,$DB,$DF,$E1
	.dc.b	$BA,$6D,$FA,$FC,$3A,$57,$F8,$92,$3F,$C5,$18,$2F,$94,$E4,$FF,$14
	.dc.b	$C6,$5F,$1F,$C3,$7A,$FE,$1A,$47,$99,$F7,$7C,$37,$82,$3F,$47,$A7
	.dc.b	$D2,$3F,$AE,$F8,$B9,$8F,$EC,$8F,$C5,$F8,$59,$1F,$01,$F9,$1F,$0D
	.dc.b	$BE,$FF,$C8,$F2,$DD,$FF,$97,$E1,$F3,$7E,$1A,$63,$F2,$1F,$A1,$F0
	.dc.b	$DB,$8F,$E0,$07,$DA,$3C,$BC,$3F,$15,$FE,$B9,$5F,$5F,$E1,$A4,$7D
	.dc.b	$27,$ED,$7C,$37,$38,$FD,$EE,$57,$6F,$6C,$7D,$8F,$85,$FC,$01,$1F
	.dc.b	$B7,$F0,$DB,$CB,$F1,$51,$F9,$77,$C4,$F9,$4F,$8A,$7F,$60,$F5,$BF
	.dc.b	$A8,$91,$FF,$05,$7A,$3F,$8E,$79,$1E,$C2,$1F,$E0,$E3,$27,$F0,$8C
	.dc.b	$3F,$1E,$FC,$03,$3C,$1F,$BF,$E0,$ED,$1E,$38,$C3,$F2,$A7,$8B,$99
	.dc.b	$FA,$6F,$27,$F8,$4D,$FB,$E3,$FC,$FB,$E2,$7F,$35,$BE,$71,$EF,$86
	.dc.b	$FB,$8F,$F0,$85,$59,$E1,$7F,$89,$77,$3D,$CF,$4C,$EF,$5E,$17,$CF
	.dc.b	$0F,$EC,$95,$B3,$C1,$3A,$71,$8F,$B0,$FF,$05,$DF,$38,$CE,$7F,$5A
	.dc.b	$7E,$FC,$7C,$13,$C2,$68,$FA,$38,$FA,$1C,$1F,$70,$E0,$7C,$DA,$EC
	.dc.b	$E2,$7C,$55,$FE,$1A,$C7,$D5,$5A,$9C,$D7,$DE,$C7,$F9,$AD,$6A,$3B
	.dc.b	$C7,$EA,$1B,$97,$D6,$7B,$61,$EB,$3F,$BA,$61,$D5,$C4,$75,$9E,$18
	.dc.b	$78,$23,$F1,$53,$A7,$EB,$9D,$8F,$C3,$39,$BF,$BF,$3F,$20,$E4,$E2
	.dc.b	$3C,$33,$C4,$FD,$69,$67,$84,$73,$AF,$C6,$CE,$63,$F5,$6F,$B0,$FA
	.dc.b	$38,$7C,$37,$07,$A0,$FC,$9E,$5F,$81,$73,$87,$F5,$FC,$7E,$14,$7C
	.dc.b	$13,$F5,$39,$7E,$41,$DE,$3F,$A6,$CB,$D6,$FF,$5B,$C3,$A4,$62,$7F
	.dc.b	$0A,$70,$3E,$29,$D4,$FD,$16,$3E,$78,$E1,$1F,$96,$3D,$C7,$EA,$A3
	.dc.b	$59,$7C,$38,$F2,$D7,$82,$31,$1F,$8D,$38,$9F,$C0,$CF,$11,$FE,$16
	.dc.b	$9E,$6F,$8E,$BE,$F6,$1F,$99,$8F,$92,$FC,$18,$F2,$9F,$6B,$1F,$8F
	.dc.b	$3F,$34,$99,$7E,$51,$F0,$BF,$B2,$7C,$33,$FD,$17,$0F,$3B,$E0,$C3
	.dc.b	$C7,$3C,$B2,$F9,$8F,$AB,$1F,$15,$F0,$3F,$1D,$8F,$8B,$1F,$C6,$57
	.dc.b	$E0,$8F,$0C,$66,$FF,$85,$61,$DA,$B7,$5E,$9C,$3E,$06,$3E,$EC,$7F
	.dc.b	$65,$3C,$8F,$16,$5F,$3D,$F9,$11,$E2,$72,$79,$F0,$ED,$8F,$F1,$9A
	.dc.b	$E1,$3E,$D3,$F0,$B7,$DD,$FE,$13,$1A,$C7,$CA,$7E,$56,$3A,$9F,$2B
	.dc.b	$0F,$2B,$4F,$F5,$EE,$5E,$F7,$F8,$8E,$5F,$21,$F3,$9F,$C2,$63,$4F
	.dc.b	$DE,$CB,$CE,$F0,$C3,$C5,$97,$F7,$3A,$DE,$5E,$7A,$FE,$38,$DC,$BC
	.dc.b	$D1,$E3,$3F,$21,$19,$AF,$E1,$79,$7D,$99,$F1,$1E,$C9,$D3,$FD,$54
	.dc.b	$E5,$5F,$89,$8C,$C7,$F0,$63,$B6,$5E,$D7,$E6,$1F,$A1,$33,$8F,$A9
	.dc.b	$FE,$4B,$8F,$AB,$0F,$14,$7A,$9B,$87,$AB,$0E,$E7,$E5,$E7,$E4,$9F
	.dc.b	$87,$CB,$E3,$BF,$A6,$C3,$B3,$FC,$33,$0E,$CF,$0C,$7C,$D8,$7D,$CC
	.dc.b	$BF,$85,$65,$E0,$7F,$C6,$F9,$78,$23,$EE,$B6,$FE,$E9,$F1,$E3,$83
	.dc.b	$F1,$E3,$4F,$F3,$43,$E0,$9F,$93,$3F,$34,$60,$F9,$39,$7A,$A3,$EE
	.dc.b	$61,$FA,$2A,$F8,$AF,$A2,$FF,$05,$7E,$13,$E1,$E1,$FB,$F7,$EB,$E3
	.dc.b	$F0,$1F,$A7,$87,$F3,$A7,$75,$F0,$1A,$7F,$24,$37,$87,$F4,$DC,$3C
	.dc.b	$11,$F6,$0F,$01,$A3,$F7,$C7,$80,$CE,$1F,$35,$F8,$93,$F1,$9F,$F2
	.dc.b	$F1,$F1,$6F,$E3,$D7,$BE,$3E,$B6,$3F,$0E,$BE,$0E,$5F,$72,$FF,$3D
	.dc.b	$97,$B7,$1F,$E1,$38,$F9,$B1,$FC,$01,$E2,$CB,$F8,$7B,$F4,$2F,$BE
	.dc.b	$1C,$F1,$FC,$7E,$3C,$EF,$FD,$49,$97,$87,$0F,$C4,$5E,$DF,$9F,$5B
	.dc.b	$7F,$AA,$CE,$DF,$88,$7E,$C2,$FE,$06,$1F,$46,$33,$97,$E7,$67,$27
	.dc.b	$F6,$A9,$F8,$75,$CE,$7D,$B8,$78,$6F,$8E,$1E,$17,$FA,$C9,$ED,$C3
	.dc.b	$9B,$FD,$B8,$FC,$81,$F5,$4F,$C8,$61,$D9,$E0,$7E,$08,$FD,$E3,$97
	.dc.b	$83,$F2,$1E,$6F,$E7,$AF,$F8,$24,$F8,$27,$E0,$1C,$E7,$E0,$4F,$E0
	.dc.b	$E3,$F0,$2F,$07,$53,$F0,$4E,$59,$7D,$BA,$FB,$F3,$F6,$CF,$E5,$78
	.dc.b	$FC,$5C,$BF,$AA,$C7,$EA,$8F,$59,$F3,$9F,$DD,$DF,$49,$E1,$3F,$DE
	.dc.b	$0F,$5C,$7A,$70,$F2,$47,$13,$F4,$8F,$A6,$FD,$77,$FB,$29,$F9,$75
	.dc.b	$D5,$F0,$D6,$E3,$D9,$1E,$E3,$FC,$14,$6A,$7F,$AA,$3E,$23,$47,$E9
	.dc.b	$63,$E1,$E5,$F8,$1A,$FE,$A7,$97,$E0,$70,$FE,$F1,$7C,$CF,$E9,$07
	.dc.b	$33,$D4,$7D,$08,$E1,$7F,$BD,$C7,$E1,$CF,$F8,$7C,$F9,$47,$E0,$23
	.dc.b	$C2,$F3,$BF,$63,$FC,$88,$F8,$27,$A3,$2F,$B1,$3F,$AC,$8F,$B0,$7F
	.dc.b	$5D,$3C,$C7,$F3,$73,$A6,$5F,$2F,$1F,$C7,$47,$89,$FC,$76,$3E,$88
	.dc.b	$FE,$1F,$18,$7A,$BF,$CA,$31,$F5,$BF,$E2,$AC,$BE,$45,$FF,$96,$2B
	.dc.b	$CE,$7C,$49,$FB,$6F,$F1,$D3,$39,$7B,$9F,$EA,$87,$D4,$7F,$0A,$67
	.dc.b	$2F,$E3,$11,$FC,$01,$F8,$2F,$F0,$63,$CB,$97,$F3,$AC,$7C,$AF,$F2
	.dc.b	$DC,$7C,$B3,$FE,$12,$8F,$9E,$FC,$17,$F2,$99,$7C,$18,$FE,$C0,$7C
	.dc.b	$AC,$7F,$A3,$D7,$28,$FE,$E1,$1F,$28,$FD,$D1,$F1,$DF,$6E,$5F,$C9
	.dc.b	$1F,$BC,$7D,$99,$E5,$3F,$E4,$D8,$F4,$9F,$CD,$01,$FB,$A9,$AC,$33
	.dc.b	$92,$31,$1F,$EE,$E3,$51,$97,$21,$C3,$C1,$AF,$08,$DB,$B6,$C6,$A3
	.dc.b	$B1,$0F,$1C,$68,$8F,$82,$38,$3A,$31,$1A,$7C,$4D,$7C,$46,$88,$68
	.dc.b	$B3,$FC,$61,$F9,$E7,$F1,$83,$EC,$61,$FD,$71,$FD,$09,$F9,$18,$FB
	.dc.b	$11,$FC,$2E,$3E,$A9,$F9,$77,$F8,$5B,$4F,$E1,$66,$AB,$F2,$97,$F6
	.dc.b	$A7,$FC,$C9,$3E,$23,$F6,$46,$EF,$F8,$8B,$F8,$F3,$F3,$D8,$7E,$3E
	.dc.b	$FF,$C1,$98,$7E,$24,$F2,$CF,$EF,$F1,$FC,$DB,$FE,$68,$3D,$67,$F7
	.dc.b	$23,$C1,$1F,$B5,$7E,$09,$FD,$B0,$DB,$F1,$CE,$AF,$B4,$FD,$89,$67
	.dc.b	$F6,$37,$E8,$9F,$CE,$E1,$FA,$6A,$FC,$B9,$FB,$3B,$F2,$BF,$E8,$DC
	.dc.b	$7E,$23,$FE,$60,$3C,$B3,$FB,$57,$CA,$7E,$EE,$36,$FF,$AB,$31,$F8
	.dc.b	$98,$7F,$9A,$A3,$E3,$9E,$0C,$7F,$C2,$78,$78,$1F,$CD,$E3,$FE,$59
	.dc.b	$9F,$66,$5F,$C1,$9F,$EF,$67,$1B,$FF,$08,$BE,$2A,$FF,$00,$BF,$1B
	.dc.b	$2F,$E5,$8F,$A0,$FD,$A1,$F6,$CF,$E8,$27,$DB,$3E,$46,$3E,$87,$FC
	.dc.b	$66,$7A,$0E,$47,$C7,$3C,$15,$FD,$B3,$0F,$9B,$1B,$7E,$AD,$FF,$21
	.dc.b	$C7,$DD,$5F,$DB,$AB,$F0,$87,$23,$FB,$71,$E3,$9F,$E5,$87,$DA,$7C
	.dc.b	$F5,$D9,$F3,$9F,$0C,$FF,$16,$9C,$CF,$EA,$CE,$E3,$9D,$FB,$67,$FD
	.dc.b	$D8,$6E,$7F,$A1,$3F,$24,$FE,$68,$7C,$09,$FA,$EE,$F0,$FE,$68,$EC
	.dc.b	$FA,$A6,$EF,$EA,$BF,$3C,$FE,$B4,$7B,$5F,$EB,$51,$FC,$EA,$BC,$D1
	.dc.b	$89,$FE,$5C,$75,$AF,$EF,$6F,$BC,$FF,$1D,$9D,$67,$F9,$EB,$B9,$FE
	.dc.b	$B5,$1F,$49,$FE,$74,$FE,$D6,$BC,$07,$F0,$97,$13,$E0,$3C,$C7,$F0
	.dc.b	$F8,$DB,$F6,$0F,$03,$FC,$55,$D9,$FB,$B8,$F0,$3F,$C6,$CD,$9F,$BB
	.dc.b	$36,$FE,$EE,$3A,$C7,$E8,$30,$FC,$3B,$FE,$2C,$AF,$4C,$FE,$E6,$3E
	.dc.b	$71,$F8,$03,$A9,$FB,$33,$E1,$1F,$D9,$CF,$65,$7C,$E7,$D8,$62,$3E
	.dc.b	$11,$D6,$3D,$31,$FC,$89,$F8,$AF,$ED,$32,$FD,$FE,$1F,$E7,$58,$E4
	.dc.b	$7C,$59,$DC,$7F,$2E,$8F,$19,$FD,$C2,$BD,$F1,$FA,$58,$FC,$61,$FC
	.dc.b	$94,$F8,$B1,$FB,$A9,$F1,$9F,$C9,$63,$DF,$3F,$C0,$1D,$9F,$98,$8D
	.dc.b	$CF,$F1,$78,$DD,$FE,$DE,$3D,$F1,$FB,$77,$C7,$1F,$3A,$30,$7B,$0F
	.dc.b	$A8,$79,$8F,$E4,$C7,$B2,$7F,$C0,$4F,$D4,$3F,$2A,$FC,$28,$FE,$65
	.dc.b	$3E,$09,$FD,$C9,$B9,$FF,$2A,$BE,$63,$F8,$1B,$F5,$DF,$C6,$47,$99
	.dc.b	$FC,$64,$75,$7F,$82,$C7,$5A,$FD,$FC,$7E,$9E,$7A,$9B,$7F,$66,$F5
	.dc.b	$3F,$24,$FB,$E3,$EA,$3E,$33,$F8,$5C,$F5,$AF,$DB,$47,$E0,$CF,$85
	.dc.b	$1F,$DA,$AB,$E8,$C7,$F3,$07,$C0,$7E,$D9,$FA,$E7,$F0,$87,$6F,$F2
	.dc.b	$D3,$C0,$FF,$14,$77,$5F,$C0,$9F,$91,$3F,$E7,$57,$EA,$47,$F4,$A8
	.dc.b	$FF,$00,$BE,$A9,$F0,$C7,$ED,$DF,$80,$FE,$34,$F9,$93,$F9,$F7,$F4
	.dc.b	$46,$23,$F2,$07,$28,$FD,$D3,$C9,$FC,$84,$6C,$FC,$84,$6E,$3F,$68
	.dc.b	$7A,$AB,$F8,$AC,$EE,$3E,$33,$E0,$7E,$94,$7E,$4D,$FC,$3C,$FC,$6A
	.dc.b	$FE,$67,$1F,$1A,$BF,$99,$BD,$4F,$E0,$07,$D5,$30,$7E,$C4,$EA,$7E
	.dc.b	$29,$F0,$3B,$9F,$63,$F5,$63,$CC,$FD,$58,$F6,$4F,$EC,$5D,$BF,$55
	.dc.b	$F3,$57,$F4,$27,$6F,$EC,$67,$F8,$35,$78,$2B,$F8,$34,$79,$AB,$F5
	.dc.b	$67,$27,$F8,$F2,$9F,$F3,$5C,$FF,$A3,$B1,$FF,$36,$CE,$6B,$FD,$60
	.dc.b	$D5,$9B,$B3,$FB,$14,$E7,$46,$9C,$B2,$FF,$3C,$CB,$A1,$BE,$72,$4E
	.dc.b	$63,$0D,$8E,$35,$CA,$9B,$F8,$23,$FD,$07,$8F,$82,$B6,$74,$A7,$0F
	.dc.b	$46,$C6,$B0,$F2,$4E,$78,$7A,$4F,$F9,$7A,$7E,$11,$E9,$E3,$E4,$48
	.dc.b	$E5,$6E,$3E,$31,$FE,$BA,$AF,$81,$CB,$E1,$24,$70,$E5,$DE,$7E,$98
	.dc.b	$F8,$7A,$7B,$F9,$FD,$04,$8E,$FC,$FF,$25,$FC,$74,$FC,$2E,$1F,$E6
	.dc.b	$19,$7C,$FE,$9F,$80,$4A,$F3,$F5,$71,$E7,$FC,$3D,$7B,$B7,$5E,$EF
	.dc.b	$C6,$7C,$B7,$E5,$75,$7C,$B8,$87,$C2,$3E,$77,$67,$5E,$AF,$E6,$F1
	.dc.b	$F4,$CD,$9D,$30,$FC,$23,$F5,$7B,$3F,$03,$DA,$EE,$A4,$9F,$B1,$0E
	.dc.b	$FE,$D1,$FC,$E0,$FC,$5C,$71,$77,$3C,$5F,$04,$7E,$65,$E7,$E0,$F8
	.dc.b	$08,$F6,$F0,$BA,$BF,$CF,$B8,$BF,$37,$4A,$DC,$F4,$7D,$3E,$17,$AB
	.dc.b	$C4,$93,$E8,$7D,$BE,$2F,$4D,$7F,$9D,$1F,$8E,$C1,$3C,$2B,$CE,$FC
	.dc.b	$4F,$17,$E2,$F8,$F2,$7E,$1B,$F2,$3C,$8E,$9A,$7F,$35,$3C,$F1,$B9
	.dc.b	$F8,$2F,$CE,$F2,$3E,$7F,$97,$27,$E6,$BF,$4B,$CB,$E7,$1F,$CD,$AF
	.dc.b	$E0,$CE,$E3,$E5,$BF,$63,$CB,$FB,$3E,$77,$DA,$F3,$E5,$7D,$88,$E6
	.dc.b	$FF,$3A,$DF,$E0,$79,$FF,$80,$24,$7D,$EF,$3F,$F8,$1F,$AA,$F4,$69
	.dc.b	$27,$F0,$7E,$8F,$4F,$07,$F9,$56,$DF,$C4,$2F,$85,$7E,$FD,$AF,$E4
	.dc.b	$5B,$97,$D1,$F0,$FB,$FA,$12,$B8,$3E,$2F,$4F,$B7,$81,$FC,$C6,$D8
	.dc.b	$7F,$24,$BD,$90,$ED,$E9,$FA,$BD,$59,$1E,$37,$DB,$3D,$23,$FC,$2F
	.dc.b	$9F,$A6,$77,$1E,$77,$E1,$FA,$BF,$13,$D5,$93,$ED,$7E,$3F,$AF,$CE
	.dc.b	$3F,$C1,$E7,$F2,$EC,$13,$C2,$3E,$13,$F3,$BD,$7F,$9F,$ED,$49,$F9
	.dc.b	$0F,$D2,$F6,$BA,$6B,$FE,$0B,$3F,$92,$E0,$F6,$BE,$B9,$2F,$E8,$6F
	.dc.b	$F6,$BD,$D9,$1F,$4D,$FB,$BF,$11,$D3,$83,$FB,$8E,$3D,$3D,$EF,$C0
	.dc.b	$92,$FE,$D7,$BD,$FD,$1B,$E1,$5F,$05,$49,$3F,$7F,$0E,$8F,$F9,$8F
	.dc.b	$0F,$3F,$C0,$FE,$33,$E0,$A4,$FE,$16,$3A,$7C,$3C,$8F,$E8,$D8,$A3
	.dc.b	$C1,$F0,$FF,$ED,$19,$7B,$6E,$FF,$F6,$A6,$4C,$1F,$0F,$E7,$F8,$6F
	.dc.b	$47,$C3,$CC,$7A,$C7,$AF,$E0,$A4,$78,$1F,$77,$C1,$7B,$FE,$12,$4F
	.dc.b	$91,$F8,$5F,$09,$F0,$FE,$1B,$E2,$7C,$34,$AF,$4C,$7C,$7F,$82,$91
	.dc.b	$EC,$7E,$57,$C0,$7C,$BF,$84,$93,$F0,$1F,$9D,$F0,$9F,$3F,$E1,$BE
	.dc.b	$87,$C3,$4A,$F8,$B1,$FD,$E3,$E2,$3F,$9A,$47,$F6,$D7,$F9,$73,$FC
	.dc.b	$FD,$BC,$9F,$DB,$1D,$C7,$F4,$3E,$43,$F7,$48,$7D,$2F,$85,$FC,$07
	.dc.b	$C2,$FE,$07,$E1,$A4,$FD,$68,$F1,$7C,$0C,$8F,$13,$E5,$F7,$FC,$DF
	.dc.b	$07,$27,$CC,$FA,$7E,$13,$D5,$F0,$92,$7D,$4F,$B7,$E1,$7D,$DF,$01
	.dc.b	$EF,$F7,$E5,$7B,$E3,$E1,$7C,$14,$8F,$84,$FC,$5F,$83,$F8,$DF,$07
	.dc.b	$F1,$FE,$12,$57,$C7,$8F,$95,$EE,$C8,$F9,$4F,$CD,$F8,$2F,$9D,$F0
	.dc.b	$72,$7E,$73,$F4,$7E,$0F,$E9,$7C,$1C,$9F,$A4,$FD,$5F,$84,$FF,$16
	.dc.b	$FC,$4B,$DF,$A4,$AF,$F1,$24,$7F,$1B,$7F,$BB,$2F,$82,$E2,$48,$DF
	.dc.b	$C1,$74,$F8,$5F,$C1,$79,$F2,$79,$C7,$83,$DB,$F0,$FC,$0C,$9E,$F1
	.dc.b	$E4,$F8,$2F,$2F,$C1,$49,$F1,$C7,$A3,$E0,$FD,$3F,$03,$27,$CF,$1D
	.dc.b	$5F,$4E,$1D,$5F,$4E,$1D,$5F,$6E,$1D,$5F,$6E,$1D,$5F,$83,$87,$57
	.dc.b	$E0,$E1,$F2,$3D,$F9,$1F,$19,$F9,$7E,$FF,$CC,$F7,$E4,$FC,$A7,$E7
	.dc.b	$FC,$1F,$D0,$F8,$39,$3F,$39,$FA,$7F,$07,$FE,$60,$F1,$1E,$D5,$25
	.dc.b	$7F,$95,$E3,$F8,$FB,$F5,$BE,$07,$ED,$91,$FB,$3F,$01,$F7,$BE,$07
	.dc.b	$EF,$FC,$1C,$9F,$BB,$1D,$FD,$8F,$07,$B3,$27,$B4,$78,$FE,$0A,$47
	.dc.b	$85,$F3,$7C,$17,$9F,$E0,$64,$F9,$5F,$57,$C0,$F5,$8F,$46,$1D,$5F
	.dc.b	$F2,$06,$BF,$E8,$27,$A1,$0C,$E1,$D9,$E8,$43,$8D,$F6,$7E,$27,$B1
	.dc.b	$F2,$48,$FC,$7F,$67,$E6,$7C,$0C,$8F,$94,$FC,$FF,$81,$FA,$1F,$03
	.dc.b	$F4,$7E,$06,$57,$CF,$8E,$6F,$A3,$0B,$EE,$FF,$A1,$34,$65,$D6,$BD
	.dc.b	$B9,$5A,$D6,$5D,$A7,$ED,$65,$D9,$F8,$79,$74,$E4,$FF,$53,$E4,$CF
	.dc.b	$15,$FE,$A6,$8E,$EF,$A8,$87,$80,$FC,$36,$1D,$DF,$9F,$87,$7C,$7D
	.dc.b	$18,$FE,$3F,$D9,$95,$F8,$D7,$F2,$5E,$D7,$E4,$FE,$07,$F2,$9F,$03
	.dc.b	$2B,$F2,$51,$F9,$7F,$81,$FC,$C7,$C0,$C9,$FC,$B4,$74,$7E,$FE,$17
	.dc.b	$2F,$91,$8F,$59,$FC,$E6,$3D,$5F,$C2,$E3,$D5,$FD,$06,$3D,$5F,$C4
	.dc.b	$E3,$FE,$C6,$F8,$A7,$B9,$7D,$C9,$5F,$EB,$D9,$CB,$7E,$23,$87,$C5
	.dc.b	$48,$C1,$FA,$D3,$F6,$44,$3F,$61,$F0,$5D,$F0,$EB,$1F,$E9,$9B,$FF
	.dc.b	$47,$36,$3C,$97,$D6,$35,$78,$7B,$10,$FD,$19,$FC,$06,$FD,$9E,$E7
	.dc.b	$B7,$E0,$24,$FA,$E3,$E0,$7B,$B2,$3D,$CF,$C3,$F8,$2F,$89,$F0,$52
	.dc.b	$7E,$13,$F1,$FA,$3A,$C7,$C5,$C3,$AB,$EB,$C3,$A3,$ED,$C2,$F2,$7F
	.dc.b	$B8,$B6,$65,$D6,$BE,$16,$56,$B5,$97,$69,$E8,$7A,$4F,$A9,$EF,$F4
	.dc.b	$EB,$FE,$FF,$4C,$7F,$D4,$35,$EC,$3E,$D6,$3D,$6B,$72,$C7,$14,$BE
	.dc.b	$58,$FE,$17,$DD,$FC,$37,$C0,$4A,$FC,$24,$7E,$27,$DD,$91,$F8,$77
	.dc.b	$F1,$9E,$EF,$E3,$7E,$07,$F1,$DF,$05,$2B,$F1,$91,$D5,$E6,$43,$F0
	.dc.b	$9F,$01,$D1,$FC,$35,$DC,$FE,$56,$1D,$67,$F2,$D8,$75,$7F,$19,$87
	.dc.b	$F8,$89,$FF,$73,$7C,$5F,$F7,$3B,$FA,$5F,$05,$FA,$13,$4B,$83,$E5
	.dc.b	$CE,$DB,$EF,$E9,$B7,$F4,$1A,$F6,$F7,$BF,$59,$EF,$FE,$B7,$E0,$E4
	.dc.b	$FE,$AE,$39,$3F,$3F,$0E,$CD,$F8,$8C,$E4,$9C,$B7,$E0,$B3,$0C,$E9
	.dc.b	$2F,$C8,$75,$CB,$C6,$FE,$EF,$2F,$1B,$FB,$1C,$B9,$FC,$17,$4F,$84
	.dc.b	$93,$CB,$3F,$14,$F8,$3E,$07,$C3,$F0,$52,$B5,$A2,$7C,$BF,$03,$E6
	.dc.b	$F8,$29,$5A,$8E,$EF,$0C,$2B,$F5,$BE,$0F,$B3,$D3,$1E,$CF,$DA,$C3
	.dc.b	$AB,$E0,$C3,$AB,$F7,$B0,$F8,$B7,$D3,$2F,$B9,$3F,$3F,$1C,$BE,$8C
	.dc.b	$3A,$CE,$8F,$A1,$8F,$8A,$7A,$1D,$C8,$60,$FC,$6F,$B7,$F5,$3D,$E9
	.dc.b	$1F,$12,$3E,$BF,$C1,$48,$F8,$EF,$DA,$F8,$2F,$B7,$F0,$52,$7E,$5B
	.dc.b	$F7,$BE,$0B,$EF,$DC,$9E,$87,$D0,$F8,$3E,$8F,$E7,$B0,$B9,$FA,$F8
	.dc.b	$59,$75,$AF,$D1,$E5,$6B,$59,$7E,$43,$E3,$3F,$C5,$35,$F8,$16,$BF
	.dc.b	$C1,$7A,$BF,$87,$C6,$2B,$E4,$2F,$7F,$97,$C4,$FE,$5A,$A5,$F3,$8F
	.dc.b	$CC,$FC,$24,$8F,$C4,$BF,$9C,$F8,$4F,$15,$F8,$FE,$16,$57,$8E,$3A
	.dc.b	$BF,$90,$C3,$AB,$E5,$3F,$8D,$63,$FE,$3E,$7F,$80,$E1,$DD,$A7,$E6
	.dc.b	$30,$F8,$1F,$09,$F0,$7E,$0A,$4F,$C1,$8F,$89,$EF,$7C,$5F,$85,$93
	.dc.b	$F1,$63,$E4,$7C,$2C,$8F,$90,$FC,$BF,$85,$EB,$1C,$C8,$79,$F8,$1A
	.dc.b	$3E,$07,$BC,$FD,$1C,$7B,$BB,$96,$3D,$E3,$18,$DC,$3E,$9E,$3F,$6F
	.dc.b	$E0,$BE,$E6,$52,$FE,$E4,$76,$7A,$65,$DA,$FE,$DE,$3C,$F7,$FB,$F9
	.dc.b	$5B,$F9,$B9,$7E,$1F,$E1,$24,$7E,$1D,$FC,$57,$C0,$FE,$2F,$DF,$93
	.dc.b	$F8,$B7,$F1,$DF,$01,$F8,$FF,$84,$FC,$87,$C2,$CA,$FC,$84,$7E,$4F
	.dc.b	$E0,$F9,$C7,$E4,$F0,$C3,$CC,$87,$C5,$C3,$AB,$C8,$FC,$D5,$F5,$7E
	.dc.b	$55,$FE,$7F,$E1,$7F,$41,$F0,$B2,$7F,$41,$1F,$CC,$B5,$FF,$76,$F0
	.dc.b	$7F,$05,$7F,$DE,$3B,$FF,$70,$6D,$7F,$69,$77,$1F,$A6,$8F,$D6,$7C
	.dc.b	$0C,$7F,$58,$75,$F8,$48,$FF,$2F,$AF,$F1,$7B,$F8,$57,$FC,$4B,$7E
	.dc.b	$1F,$84,$F2,$7C,$24,$AE,$11,$FB,$97,$B1,$F8,$7C,$30,$F5,$F8,$4F
	.dc.b	$61,$1F,$07,$BD,$EF,$F8,$4F,$81,$F0,$72,$7C,$91,$F0,$FE,$13,$E2
	.dc.b	$7C,$24,$9F,$44,$7C,$7F,$83,$F9,$18,$49,$F6,$47,$57,$DD,$87,$57
	.dc.b	$DD,$87,$57,$E1,$61,$D5,$F8,$58,$7D,$3F,$7A,$47,$C7,$7E,$B7,$C1
	.dc.b	$F5,$8F,$91,$87,$5D,$FE,$C6,$3D,$67,$E6,$63,$D6,$7E,$E6,$3F,$7F
	.dc.b	$DF,$95,$F4,$DF,$C1,$7B,$5F,$83,$F8,$39,$3F,$5D,$FC,$37,$C1,$FE
	.dc.b	$1F,$E0,$E4,$FD,$B7,$F1,$5F,$07,$CA,$3E,$F5,$D7,$EF,$7C,$1F,$47
	.dc.b	$F0,$18,$5C,$FF,$21,$8F,$49,$FC,$26,$36,$7F,$29,$8F,$E5,$FD,$F9
	.dc.b	$5F,$8B,$7F,$C1,$7F,$1A,$F7,$EF,$C2,$63,$E1,$25,$FF,$7F,$AF,$F0
	.dc.b	$79,$FE,$0D,$71,$F0,$7C,$8E,$7F,$09,$D3,$09,$3C,$5E,$FF,$09,$DA
	.dc.b	$3A,$61,$D9,$FE,$41,$87,$6B,$E9,$1D,$8F,$13,$E8,$F8,$0F,$4F,$C2
	.dc.b	$49,$F3,$47,$B3,$DA,$91,$E9,$7D,$FF,$03,$F0,$3E,$12,$4F,$B5,$F8
	.dc.b	$7F,$09,$F1,$3E,$13,$B4,$F4,$3E,$13,$D1,$F3,$DD,$E3,$FF,$03,$68
	.dc.b	$C7,$AD,$7B,$31,$B5,$AC,$7E,$8F,$C0,$4A,$F9,$CF,$D4,$F8,$0E,$9E
	.dc.b	$57,$F9,$EB,$26,$78,$BF,$F3,$A4,$77,$7D,$44,$3C,$07,$D9,$C3,$EF
	.dc.b	$FC,$1F,$E0,$3E,$0E,$4F,$DD,$8F,$C1,$E1,$23,$F0,$0F,$E1,$BD,$FF
	.dc.b	$C3,$FB,$FF,$88,$F8,$49,$5F,$85,$8F,$F2,$AF,$C7,$7F,$80,$6F,$FD
	.dc.b	$22,$D9,$FF,$45,$37,$73,$FD,$CD,$F9,$33,$E1,$9F,$C1,$BE,$17,$B1
	.dc.b	$F9,$3B,$F0,$BF,$88,$BF,$0F,$B9,$E2,$F8,$29,$3F,$E5,$7C,$53,$E6
	.dc.b	$F4,$64,$E5,$F4,$FB,$9E,$AF,$81,$93,$C5,$F6,$FC,$17,$BB,$DF,$EF
	.dc.b	$3D,$70,$C3,$CC,$87,$F7,$BC,$58,$F6,$8E,$84,$35,$87,$67,$D1,$F0
	.dc.b	$3F,$30,$8F,$AF,$E0,$7E,$7F,$BB,$F4,$3E,$02,$4F,$BE,$3E,$9F,$BB
	.dc.b	$23,$E1,$3F,$5B,$DE,$FA,$FF,$03,$F6,$30,$95,$F1,$E3,$AB,$F2,$70
	.dc.b	$EA,$FA,$F0,$E8,$FB,$70,$BC,$BF,$E6,$7D,$59,$7E,$13,$E0,$65,$FD
	.dc.b	$37,$F0,$FF,$03,$DA,$3E,$36,$5D,$B0,$E5,$97,$4E,$7F,$C5,$E7,$6F
	.dc.b	$F1,$79,$7E,$43,$E0,$64,$7D,$F7,$FD,$4F,$F1,$FF,$DC,$23,$FD,$B4
	.dc.b	$FF,$74,$C7,$F2,$31,$F9,$83,$7F,$01,$F9,$AF,$7F,$F3,$7F,$05,$27
	.dc.b	$F1,$51,$D3,$E2,$3A,$FC,$14,$9F,$F5,$26,$3F,$DD,$35,$FF,$78,$BD
	.dc.b	$F8,$BF,$DA,$13,$F8,$33,$B9,$D0,$FC,$B6,$3D,$DF,$C4,$63,$EA,$F8
	.dc.b	$39,$1D,$9F,$6F,$B7,$EE,$F7,$3D,$FF,$03,$2B,$C7,$1F,$0B,$E0,$A4
	.dc.b	$79,$9F,$8B,$F0,$7F,$1B,$E0,$79,$CF,$AB,$0C,$3C,$8F,$6E,$1D,$5F
	.dc.b	$F2,$B6,$AC,$7A,$C6,$31,$BC,$5F,$EA,$9A,$65,$F5,$3E,$0B,$EA,$FC
	.dc.b	$14,$C3,$E4,$C7,$D8,$F7,$E4,$7C,$C7,$ED,$FB,$FF,$73,$DF,$93,$F4
	.dc.b	$1F,$BF,$EF,$FE,$03,$E0,$BF,$03,$F0,$72,$BE,$AC,$74,$7D,$38,$5D
	.dc.b	$7C,$D8,$F4,$9F,$6E,$36,$7D,$58,$FE,$33,$E0,$E5,$7E,$01,$FC,$7F
	.dc.b	$C1,$F3,$E9,$FC,$0E,$56,$BF,$23,$97,$39,$FC,$2E,$56,$7F,$2B,$8F
	.dc.b	$FB,$FB,$DC,$7C,$0D,$24,$FE,$3B,$E0,$F5,$F0,$32,$3F,$DE,$94,$8E
	.dc.b	$5F,$07,$CF,$E0,$A5,$69,$ED,$F0,$5D,$EE,$4F,$27,$C5,$F0,$7D,$63
	.dc.b	$FD,$29,$1D,$5E,$F8,$76,$7A,$10,$DD,$F6,$7C,$BF,$05,$ED,$23,$E8
	.dc.b	$F8,$3F,$81,$EF,$FC,$1F,$83,$93,$EC,$8F,$89,$F0,$72,$3D,$EF,$C7
	.dc.b	$F8,$0F,$91,$F0,$72,$7E,$1B,$F2,$FE,$0F,$AC,$7C,$6C,$3A,$BE,$CC
	.dc.b	$3A,$3E,$EC,$2F,$57,$FC,$01,$D9,$97,$D5,$F8,$1F,$AD,$F0,$53,$0F
	.dc.b	$A3,$06,$5B,$A2,$FC,$87,$F6,$12,$7C,$36,$EB,$FD,$81,$1F,$D8,$E7
	.dc.b	$8E,$5C,$67,$EE,$FC,$27,$E1,$4F,$C3,$4F,$E1,$FE,$0E,$4F,$E0,$9F
	.dc.b	$C5,$7C,$27,$5C,$B9,$0C,$7C,$35,$E6,$BF,$2E,$5B,$7C,$FF,$0F,$DA
	.dc.b	$7C,$24,$3F,$07,$1B,$57,$D7,$E0,$F0,$F7,$F8,$5B,$89,$B3,$8C,$65
	.dc.b	$F8,$CE,$75,$F1,$F1,$F1,$47,$E4,$B4,$F8,$4F,$E8,$32,$EB,$F0,$BD
	.dc.b	$BE,$0A,$4F,$3D,$3E,$04,$F8,$BE,$0A,$4E,$72,$FF,$77,$37,$E1,$7C
	.dc.b	$FF,$0B,$E8,$F8,$59,$3B,$8F,$5F,$C2,$48,$E4,$FB,$BE,$17,$AC,$7D
	.dc.b	$CC,$6E,$7F,$4B,$2F,$F7,$BC,$F3,$3C,$58,$F6,$7F,$07,$8F,$C8,$F8
	.dc.b	$1F,$93,$F0,$32,$7D,$11,$F3,$3E,$17,$E6,$FC,$2C,$9F,$64,$7D,$0F
	.dc.b	$84,$91,$EF,$7E,$9F,$C2,$FD,$4F,$80,$93,$F0,$DF,$AF,$F0,$BD,$63
	.dc.b	$E3,$61,$D5,$FC,$BE,$1D,$1F,$CC,$E1,$76,$F7,$63,$F8,$09,$FC,$0F
	.dc.b	$C0,$C7,$E8,$1F,$86,$F9,$17,$BB,$49,$1F,$66,$3F,$A4,$B7,$E0,$B8
	.dc.b	$11,$D7,$C2,$F3,$F8,$5E,$9F,$0D,$27,$94,$74,$7F,$2F,$D9,$FB,$6C
	.dc.b	$3C,$31,$DF,$1F,$27,$C0,$75,$8D,$9F,$9E,$9F,$F7,$9E,$BF,$AC,$6D
	.dc.b	$F3,$CB,$D9,$F0,$5E,$DF,$83,$98,$7A,$E3,$C5,$87,$1C,$AE,$1E,$3C
	.dc.b	$BC,$13,$D7,$5B,$3D,$70,$F8,$9F,$0B,$F2,$7E,$16,$46,$0F,$91,$F0
	.dc.b	$3F,$34,$8F,$CB,$F8,$2F,$A1,$F0,$72,$3E,$73,$F4,$FE,$13,$EA,$7C
	.dc.b	$04,$9F,$A4,$FD,$7F,$80,$E7,$1F,$13,$57,$1B,$2C,$C7,$C5,$CB,$B5
	.dc.b	$6E,$99,$62,$BE,$56,$5F,$83,$F8,$29,$7F,$81,$7F,$0D,$F0,$5F,$C3
	.dc.b	$3E,$4B,$FB,$2C,$FE,$F5,$AF,$8D,$B7,$F5,$7E,$0B,$F1,$DF,$0B,$F8
	.dc.b	$FF,$85,$97,$F8,$D8,$E9,$F0,$1D,$7E,$0E,$4F,$58,$F0,$7C,$2C,$8F
	.dc.b	$03,$F9,$9C,$3A,$C7,$E5,$F0,$EA,$FB,$F0,$F4,$7C,$1C,$8F,$43,$EB
	.dc.b	$F8,$5F,$67,$BD,$ED,$F7,$A5,$7B,$63,$E0,$7C,$24,$8F,$80,$FC,$3F
	.dc.b	$84,$F8,$9E,$F4,$9F,$88,$FC,$7F,$85,$EB,$1F,$1F,$0E,$AF,$C7,$C3
	.dc.b	$A3,$F2,$70,$BC,$3F,$D2,$38,$99,$7D,$1F,$82,$97,$F4,$5F,$A9,$F0
	.dc.b	$7D,$A3,$E9,$65,$DB,$0E,$59,$74,$F0,$3F,$B2,$E4,$CF,$18,$7F,$4A
	.dc.b	$8E,$EF,$DD,$F8,$5F,$C0,$7C,$2C,$8C,$1F,$81,$F8,$3F,$C2,$11,$FC
	.dc.b	$27,$C1,$FE,$1F,$E1,$3F,$11,$F0,$B2,$7F,$11,$1F,$8B,$CE,$47,$E2
	.dc.b	$DF,$C7,$7C,$0F,$38,$FC,$66,$17,$6F,$89,$8F,$59,$FC,$96,$3D,$5F
	.dc.b	$C9,$63,$FD,$8B,$CC,$FF,$70,$13,$E0,$F1,$F0,$79,$F8,$29,$7F,$D5
	.dc.b	$E9,$1C,$7E,$0F,$97,$C2,$4A,$D3,$D7,$E1,$7B,$7B,$FD,$FE,$02,$57
	.dc.b	$38,$EA,$FE,$DF,$1E,$AF,$7C,$7B,$3D,$08,$6F,$0E,$CF,$97,$E0,$BD
	.dc.b	$84,$7D,$1F,$0B,$EF,$F7,$3E,$07,$B9,$27,$D9,$1F,$0F,$E0,$64,$7B
	.dc.b	$DF,$8D,$F0,$5F,$1F,$E1,$7E,$47,$C2,$CA,$F8,$91,$D5,$F8,$D8,$75
	.dc.b	$7D,$98,$74,$7D,$D8,$5E,$7F,$F1,$0E,$DF,$ED,$49,$F9,$DF,$03,$F5
	.dc.b	$48,$FD,$1F,$82,$EC,$FC,$7C,$7B,$5F,$2C,$7A,$76,$7F,$8F,$B1,$65
	.dc.b	$8C,$3F,$C4,$71,$DD,$FA,$1F,$09,$F8,$32,$3F,$7F,$E1,$3F,$C4,$FF
	.dc.b	$27,$FE,$D7,$C8,$FF,$B5,$C9,$F0,$7F,$8B,$A8,$7E,$1C,$FC,$77,$C2
	.dc.b	$F2,$F8,$0E,$7F,$09,$27,$FC,$45,$7F,$D5,$9F,$F1,$BB,$DC,$ED,$F0
	.dc.b	$BE,$2C,$3A,$CF,$E1,$CB,$8F,$CA,$C7,$CF,$F0,$92,$7A,$BE,$AF,$84
	.dc.b	$F5,$FC,$27,$B3,$E1,$25,$78,$A3,$DF,$F0,$BF,$03,$86,$4F,$9A,$3E
	.dc.b	$1F,$C0,$48,$F4,$BF,$1B,$E0,$3A,$C7,$B3,$0E,$AF,$F5,$3D,$D8,$F6
	.dc.b	$8F,$83,$8F,$67,$58,$FD,$0F,$83,$FA,$3F,$07,$27,$E3,$C7,$D4,$F6
	.dc.b	$FE,$AF,$B9,$27,$E5,$C7,$D8,$F8,$29,$1F,$39,$FB,$7F,$05,$F7,$3E
	.dc.b	$16,$4F,$D2,$7E,$FF,$C2,$F5,$8F,$AB,$87,$57,$D5,$87,$47,$D9,$85
	.dc.b	$E5,$FF,$3D,$EE,$CB,$FC,$DF,$F2,$9F,$D8,$3B,$BF,$D2,$2D,$C3,$FC
	.dc.b	$EA,$DB,$FC,$6F,$C1,$78,$67,$E4,$67,$E1,$9E,$B9,$F7,$C3,$FD,$37
	.dc.b	$93,$3C,$61,$F8,$AC,$FC,$3F,$09,$23,$FC,$E1,$A2,$3C,$BF,$0B,$E6
	.dc.b	$F8,$2F,$3F,$C1,$4B,$DC,$7A,$BE,$13,$D7,$F0,$92,$79,$C7,$BB,$DF
	.dc.b	$F7,$FC,$24,$9E,$F1,$D5,$F0,$E1,$D5,$FF,$0B,$66,$C7,$B4,$79,$B1
	.dc.b	$EC,$EB,$1F,$97,$F0,$32,$3D,$6F,$CE,$F8,$2F,$9F,$F0,$32,$7D,$EF
	.dc.b	$D2,$F8,$1F,$A7,$F0,$9F,$53,$E1,$25,$7C,$48,$FA,$FE,$FC,$8F,$8E
	.dc.b	$FD,$AF,$80,$E9,$1E,$8C,$2F,$4F,$FA,$EB,$76,$5D,$AB,$E7,$E5,$D9
	.dc.b	$F7,$E5,$F8,$4F,$82,$FC,$2F,$C1,$49,$FA,$B1,$FE,$B6,$F9,$57,$B9
	.dc.b	$49,$3F,$EA,$98,$FF,$5C,$37,$E0,$F8,$11,$D7,$C1,$FE,$4B,$BB,$F2
	.dc.b	$7F,$09,$27,$F0,$71,$D1,$FF,$24,$5F,$81,$ED,$87,$8F,$E0,$FA,$46
	.dc.b	$B1,$BC,$1F,$ED,$AB,$FF,$73,$4F,$FB,$41,$FE,$B6,$F4,$F8,$4F,$61
	.dc.b	$ED,$F8,$2F,$05,$71,$C6,$E5,$FE,$F6,$C3,$DB,$87,$73,$DC,$43,$CF
	.dc.b	$A7,$C7,$F8,$3F,$91,$F0,$92,$7E,$34,$7C,$BF,$80,$F9,$9F,$03,$27
	.dc.b	$E5,$47,$CF,$F8,$49,$1F,$35,$FA,$5F,$09,$F4,$FD,$FF,$A9,$F0,$92
	.dc.b	$BE,$94,$75,$7E,$A6,$1D,$5F,$97,$87,$47,$E6,$E1,$77,$F9,$18,$FD
	.dc.b	$FF,$81,$FC,$07,$C1,$4B,$FB,$D1,$F8,$3F,$82,$FC,$27,$C1,$49,$FC
	.dc.b	$14,$7E,$1F,$E1,$24,$7E,$15,$FC,$57,$C2,$7E,$2E,$21,$F8,$93,$F1
	.dc.b	$DF,$01,$D5,$FC,$65,$F5,$7F,$07,$7D,$1F,$C2,$DD,$CF,$F0,$18,$27
	.dc.b	$FD,$F9,$C3,$FD,$9C,$9F,$03,$98,$FF,$4A,$43,$E0,$B1,$6E,$7E,$4A
	.dc.b	$5E,$DE,$A7,$30,$F7,$31,$82,$F8,$46,$1E,$57,$6B,$FD,$97,$5E,$2F
	.dc.b	$81,$C1,$9F,$82,$D7,$C2,$49,$F5,$BC,$7E,$13,$96,$1C,$F2,$95,$F0
	.dc.b	$23,$93,$F0,$B0,$EA,$FC,$4C,$3C,$5F,$03,$E3,$F8,$29,$3F,$22,$3A
	.dc.b	$BF,$2B,$0E,$AF,$C8,$C3,$AB,$E9,$3B,$1D,$4F,$A3,$F0,$9E,$FF,$84
	.dc.b	$91,$83,$EA,$FC,$17,$C3,$23,$F6,$3E,$0B,$E3,$7C,$1C,$8F,$B6,$FC
	.dc.b	$9F,$84,$F9,$5F,$01,$F2,$FE,$02,$57,$E0,$23,$9B,$F7,$B0,$BC,$9F
	.dc.b	$76,$3E,$B7,$23,$2E,$D1,$F8,$4C,$AD,$6C,$FC,$16,$5F,$67,$E0,$A5
	.dc.b	$7E,$39,$FB,$9F,$07,$F7,$FE,$5B,$FA,$E4,$FE,$2B,$A3,$F1,$70,$F8
	.dc.b	$2D,$1B,$F8,$3E,$1F,$0F,$2F,$81,$F8,$BA,$FE,$E9,$B3,$E1,$BB,$7C
	.dc.b	$04,$AD,$EC,$B4,$F8,$FE,$07,$B5,$6A,$7C,$4D,$CF,$2F,$FB,$DE,$3F
	.dc.b	$C0,$44,$CB,$F3,$18,$FB,$7E,$0A,$57,$37,$E0,$7C,$1F,$C1,$F6,$FE
	.dc.b	$17,$B9,$2B,$C1,$1F,$17,$E0,$E4,$78,$DF,$91,$F0,$7F,$27,$E1,$3E
	.dc.b	$57,$C2,$CA,$F4,$47,$57,$D5,$87,$57,$D1,$87,$47,$D5,$85,$E1,$FB
	.dc.b	$5B,$32,$FA,$DF,$07,$F5,$FE,$0A,$61,$F1,$63,$B3,$F0,$F2,$ED,$87
	.dc.b	$2C,$BA,$73,$7D,$CC,$ED,$FD,$7C,$BF,$05,$F0,$BF,$83,$F8,$59,$3F
	.dc.b	$46,$3F,$0D,$7F,$87,$F8,$29,$3F,$56,$3F,$15,$F0,$7F,$8B,$F8,$49
	.dc.b	$3F,$66,$3F,$86,$7C,$BB,$E1,$2F,$C0,$4A,$FC,$1C,$E5,$E8,$7E,$07
	.dc.b	$1E,$EF,$DF,$C7,$BB,$F8,$5C,$7B,$BF,$83,$C7,$BF,$C1,$48,$EA,$F8
	.dc.b	$BE,$17,$C7,$DF,$27,$C2,$F9,$BE,$0B,$CF,$F0,$9E,$8F,$85,$95,$E6
	.dc.b	$8F,$5F,$C0,$48,$F4,$BE,$EF,$80,$E9,$1F,$CB,$75,$FE,$6B,$BF,$F3
	.dc.b	$58,$EE,$7C,$0C,$7B,$BF,$0B,$1E,$EF,$13,$F9,$B4,$FF,$81,$39,$FF
	.dc.b	$7C,$BE,$07,$E6,$92,$7E,$5F,$C1,$7D,$0F,$76,$47,$CE,$7E,$9F,$BB
	.dc.b	$F5,$3E,$07,$EA,$FC,$14,$AF,$A7,$1F,$63,$E0,$FE,$CF,$C2,$49,$FA
	.dc.b	$F1,$D1,$F3,$68,$D1,$96,$63,$CF,$97,$7A,$FB,$F9,$77,$78,$DB,$2F
	.dc.b	$C3,$FC,$17,$E2,$3E,$0E,$57,$E1,$A3,$A7,$0F,$CC,$CE,$CF,$C6,$CB
	.dc.b	$B3,$D3,$2E,$D9,$FC,$8C,$BF,$27,$30,$FC,$89,$F9,$6F,$84,$FE,$C5
	.dc.b	$F3,$1F,$EC,$DA,$2F,$82,$C9,$AF,$83,$DF,$C2,$F0,$F8,$59,$8F,$F1
	.dc.b	$EB,$FF,$71,$1F,$DA,$B0,$FE,$6C,$DD,$BF,$BC,$B6,$7F,$BA,$2F,$85
	.dc.b	$ED,$1C,$F2,$F0,$BD,$72,$F0,$BD,$F1,$F0,$BF,$52,$FC,$3F,$03,$EC
	.dc.b	$24,$79,$7E,$0B,$DF,$F0,$52,$3D,$0F,$C2,$F8,$2F,$87,$F0,$72,$7D
	.dc.b	$8F,$C6,$F8,$4F,$8F,$F0,$12,$7E,$03,$F2,$BE,$03,$A4,$78,$F0,$BC
	.dc.b	$3F,$0B,$1E,$93,$E7,$C6,$CF,$C6,$C7,$E9,$FC,$17,$D4,$F8,$29,$7F
	.dc.b	$36,$3F,$C3,$1E,$C7,$FB,$0F,$7F,$EA,$13,$F4,$7D,$BC,$FB,$9F,$73
	.dc.b	$E0,$A4,$FD,$57,$EF,$FC,$17,$E0,$3E,$0E,$4F,$D9,$7F,$07,$F0,$9C
	.dc.b	$A3,$A1,$0F,$D0,$E5,$E1,$BE,$B1,$F4,$5E,$AF,$F6,$08,$FE,$DE,$F0
	.dc.b	$F8,$4F,$49,$EA,$F8,$2E,$D1,$F6,$73,$FC,$99,$8E,$0F,$C4,$E1,$D9
	.dc.b	$B3,$F8,$FC,$FE,$1F,$C2,$7C,$4F,$84,$93,$E1,$8F,$8F,$F0,$52,$3C
	.dc.b	$8F,$CA,$D3,$E5,$FC,$24,$9F,$43,$F3,$BE,$13,$E7,$FB,$FF,$43,$DF
	.dc.b	$95,$ED,$8E,$8F,$C4,$C2,$E9,$FE,$15,$C4,$75,$9F,$93,$8D,$9D,$63
	.dc.b	$F7,$3E,$02,$57,$C9,$7E,$FF,$C0,$FE,$03,$E0,$3F,$03,$F0,$32,$BE
	.dc.b	$74,$7E,$13,$E0,$BF,$0B,$F0,$72,$7E,$94,$7E,$23,$DF,$FC,$4F,$C0
	.dc.b	$49,$FA,$D1,$D5,$FB,$18,$75,$7E,$FE,$1D,$5F,$B9,$87,$57,$F0,$78
	.dc.b	$7E,$53,$CB,$FC,$AF,$C0,$C9,$FC,$14,$7F,$9B,$FE,$67,$F9,$E3,$FE
	.dc.b	$7A,$6C,$7F,$9D,$5B,$B9,$7B,$7C,$7E,$0A,$57,$F8,$B2,$7F,$DA,$4F
	.dc.b	$F3,$47,$FD,$87,$5F,$E4,$57,$FB,$74,$67,$E0,$FC,$47,$8F,$E1,$3C
	.dc.b	$9F,$05,$27,$A3,$E7,$C7,$BC,$75,$C6,$E3,$D7,$E1,$3D,$9F,$09,$ED
	.dc.b	$F8,$29,$7E,$58,$E8,$F9,$71,$B7,$E5,$C7,$9E,$BF,$0F,$2B,$7F,$0F
	.dc.b	$2F,$91,$F0,$92,$3E,$03,$F2,$FE,$13,$E6,$7C,$0C,$9F,$88,$FC,$FF
	.dc.b	$82,$FA,$1F,$09,$27,$E4,$3F,$4F,$E1,$3E,$A7,$BD,$27,$E6,$3F,$5F
	.dc.b	$DF,$E9,$1F,$37,$0B,$BF,$CD,$C7,$A4,$FD,$1C,$6C,$FD,$1C,$7F,$01
	.dc.b	$F0,$32,$BE,$C3,$F8,$3F,$81,$FC,$24,$43,$EE,$1F,$87,$F8,$1F,$C4
	.dc.b	$7C,$17,$E2,$7E,$0E,$4F,$E0,$63,$FD,$75,$F3,$4F,$84,$BF,$03,$2B
	.dc.b	$FC,$E3,$39,$7A,$90,$FC,$41,$DC,$E8,$7E,$36,$FB,$BF,$8D,$BE,$FF
	.dc.b	$05,$E0,$F8,$39,$3D,$A3,$C7,$EE,$C8,$F0,$BE,$6F,$77,$CF,$F0,$7E
	.dc.b	$8F,$83,$95,$E6,$8F,$5F,$C3,$FB,$3E,$16,$4F,$AA,$3A,$3F,$E1,$2D
	.dc.b	$58,$E3,$93,$FD,$49,$0C,$BB,$D7,$C2,$CB,$BB,$C7,$2F,$93,$F0,$72
	.dc.b	$3E,$3B,$F3,$3E,$0F,$A7,$81,$FE,$F0,$C7,$BD,$7B,$0F,$9B,$8F,$67
	.dc.b	$C9,$8F,$6C,$BC,$79,$7D,$6F,$85,$FA,$FF,$0B,$2F,$EA,$C7,$DA,$F8
	.dc.b	$19,$1F,$61,$FB,$BF,$03,$F7,$BE,$12,$4F,$DC,$7F,$03,$F0,$BF,$82
	.dc.b	$F8,$0F,$C1,$FC,$04,$AF,$C0,$C7,$47,$E4,$61,$74,$F8,$F8,$F4,$9F
	.dc.b	$99,$8D,$9F,$97,$8A,$7F,$D7,$7F,$36,$F8,$5F,$C8,$FC,$1C,$9C,$C6
	.dc.b	$DF,$F2,$3C,$FF,$B0,$1E,$3F,$05,$9C,$3F,$DB,$53,$99,$DF,$C1,$E7
	.dc.b	$97,$C1,$4A,$FE,$D6,$F0,$F8,$4D,$5A,$35,$18,$78,$7C,$2D,$A7,$C0
	.dc.b	$D2,$5F,$B3,$E0,$75,$9C,$8F,$39,$F0,$33,$E5,$87,$3C,$A4,$FC,$48
	.dc.b	$ED,$F0,$72,$3E,$3B,$E1,$F8,$3F,$17,$B9,$E3,$F7,$65,$7C,$C8,$F3
	.dc.b	$7C,$1F,$9F,$E0,$E4,$FD,$08,$F5,$7C,$24,$8F,$A6,$FB,$7E,$17,$AC
	.dc.b	$7D,$6C,$3A,$BF,$4B,$0E,$AF,$DA,$C3,$AB,$F5,$B0,$F8,$DF,$05,$23
	.dc.b	$EF,$BF,$27,$E0,$FA,$C7,$DC,$C3,$AF,$27,$DA,$C7,$AC,$FE,$03,$1E
	.dc.b	$B3,$F7,$B1,$FA,$5F,$09,$F4,$FE,$16,$5F,$E3,$23,$EC,$78,$5F,$D6
	.dc.b	$2D,$D1,$F8,$28,$7C,$1E,$8D,$FC,$1F,$DD,$9F,$BD,$F0,$B3,$1F,$CB
	.dc.b	$C7,$4F,$80,$91,$F8,$C8,$E9,$1E,$0F,$80,$E7,$39,$3F,$19,$8F,$81
	.dc.b	$C6,$48,$C4,$7F,$A9,$1F,$45,$FE,$63,$B7,$D3,$3F,$A8,$C7,$C8,$FE
	.dc.b	$92,$7D,$FF,$05,$F0,$3E,$0A,$4F,$6C,$3F,$C0,$63,$E0,$BE,$29,$23
	.dc.b	$5F,$05,$F2,$3E,$0F,$E4,$FC,$24,$9E,$51,$F3,$3E,$03,$E6,$FC,$04
	.dc.b	$9E,$D1,$CD,$F5,$61,$77,$F4,$9E,$FC,$7B,$4F,$97,$1E,$CF,$C1,$C7
	.dc.b	$EC,$FC,$14,$8F,$53,$F7,$3E,$0B,$EE,$FB,$92,$7D,$CF,$E0,$3D,$DF
	.dc.b	$C0,$FC,$14,$9F,$84,$FE,$13,$E0,$BF,$0B,$F0,$92,$7E,$33,$F8,$8F
	.dc.b	$84,$E9,$1F,$4F,$0B,$B7,$AB,$89,$97,$5A,$FB,$19,$5A,$D6,$5F,$92
	.dc.b	$9F,$C9,$FC,$24,$7E,$A1,$D3,$9B,$F2,$99,$58,$FC,$3E,$3C,$E7,$F3
	.dc.b	$18,$D9,$FC,$5E,$1F,$C5,$7E,$73,$FB,$A4,$7F,$BB,$1B,$97,$D0,$6D
	.dc.b	$FF,$22,$5F,$09,$C7,$E0,$79,$7C,$14,$C3,$F2,$B5,$FE,$CD,$C0,$FF
	.dc.b	$B7,$4F,$F6,$E4,$6B,$E1,$3C,$5F,$09,$E3,$E0,$93,$C9,$F3,$7B,$FD
	.dc.b	$E3,$AE,$3D,$DF,$46,$5C,$9E,$84,$3D,$97,$EF,$F8,$49,$1E,$67,$E1
	.dc.b	$7C,$0F,$C3,$F8,$09,$3E,$A7,$E3,$7C,$0F,$C7,$F8,$49,$3E,$E7,$E5
	.dc.b	$7C,$27,$CB,$F7,$FE,$67,$BF,$2B,$E1,$C7,$47,$E7,$61,$78,$BE,$76
	.dc.b	$3D,$27,$E9,$63,$67,$E9,$63,$F5,$FE,$07,$EC,$7C,$0C,$BF,$9F,$1F
	.dc.b	$D3,$FE,$77,$FA,$AF,$3F,$F5,$85,$EE,$64,$D7,$B9,$BF,$81,$FC,$17
	.dc.b	$C1,$4C,$7F,$86,$DF,$F6,$D7,$ED,$FC,$1F,$E1,$FE,$12,$46,$0D,$E1
	.dc.b	$D5,$E3,$8F,$67,$9E,$3D,$5F,$C7,$63,$E1,$7B,$FC,$17,$A7,$E0,$A4
	.dc.b	$60,$EA,$74,$21,$FD,$BF,$4F,$76,$3C,$E3,$A9,$F9,$2B,$C7,$07,$F0
	.dc.b	$AD,$7F,$B8,$CF,$BB,$E1,$3E,$41,$F2,$7E,$13,$E5,$7C,$14,$9F,$86
	.dc.b	$FC,$DF,$82,$F9,$DF,$09,$F3,$FE,$12,$57,$C8,$8F,$A5,$EF,$C8,$F9
	.dc.b	$6F,$D5,$F8,$0E,$91,$E9,$3F,$C1,$19,$B1,$CE,$9E,$7C,$BB,$D7,$D4
	.dc.b	$CB,$BB,$C6,$56,$CB,$F0,$5F,$03,$2B,$ED,$BF,$85,$F8,$2F,$C3,$7C
	.dc.b	$0F,$E1,$FE,$0A,$57,$E0,$23,$F1,$5F,$09,$23,$F0,$6F,$E3,$7E,$13
	.dc.b	$F1,$DC,$B2,$7F,$0E,$FE,$47,$DF,$EB,$1F,$8A,$C3,$AB,$E7,$D1,$8F
	.dc.b	$58,$C6,$37,$5F,$AD,$97,$F8,$A7,$E7,$BF,$AD,$D7,$F8,$B5,$B2,$69
	.dc.b	$FD,$89,$FE,$8E,$BD,$9E,$5E,$D4,$AF,$ED,$33,$96,$FC,$07,$72,$3A
	.dc.b	$F8,$0F,$17,$C1,$F8,$FE,$0E,$4F,$28,$F3,$63,$D6,$3E,$A6,$0C,$73
	.dc.b	$1D,$88,$6B,$1F,$76,$3E,$FF,$80,$F8,$1F,$01,$2B,$CF,$1C,$9E,$B7
	.dc.b	$DF,$4F,$7E,$5E,$89,$E8,$76,$8F,$87,$87,$CB,$F8,$19,$3F,$09,$F9
	.dc.b	$DF,$05,$F3,$FE,$02,$4F,$C6,$7E,$97,$C0,$7D,$3F,$82,$93,$F2,$9F
	.dc.b	$AD,$F0,$5F,$5F,$DF,$93,$F3,$9F,$B5,$EF,$F4,$8F,$8B,$85,$D3,$EB
	.dc.b	$E3,$D2,$7E,$4E,$36,$7E,$DE,$3F,$83,$F8,$1F,$C2,$7C,$0C,$BF,$B7
	.dc.b	$1F,$E6,$FE,$2F,$EA,$F1,$FE,$73,$6D,$9A,$7F,$5A,$8F,$C1,$9F,$8F
	.dc.b	$F8,$28,$E0,$FF,$17,$C3,$DB,$ED,$F0,$12,$73,$1E,$1F,$87,$91,$C1
	.dc.b	$F2,$7C,$1F,$E6,$E3,$A9,$F8,$6C,$30,$F6,$F7,$FD,$5F,$01,$23,$07
	.dc.b	$8B,$DA,$F7,$7B,$52,$30,$79,$BE,$03,$E1,$7C,$04,$8C,$1E,$AF,$82
	.dc.b	$F8,$DF,$05,$23,$07,$23,$F8,$67,$03,$0E,$D1,$8C,$7E,$76,$3D,$63
	.dc.b	$78,$E1,$F8,$FF,$01,$F4,$FE,$02,$46,$0E,$DC,$7F,$E3,$CC,$BB,$47
	.dc.b	$CE,$CB,$ED,$65,$CE,$3A,$9F,$4B,$1C,$5F,$1F,$81,$FC,$01,$27,$EB
	.dc.b	$FC,$17,$E0,$FE,$03,$F0,$9F,$01,$27,$ED,$C7,$E1,$FE,$06,$47,$DE
	.dc.b	$7F,$15,$F0,$5F,$EB,$9F,$9F,$FE,$C5,$3F,$CB,$DB,$97,$F9,$51,$BE
	.dc.b	$9F,$FB,$1B,$77,$BD,$E1,$BF,$C4,$E7,$E1,$7E,$6E,$7E,$07,$E7,$E7
	.dc.b	$72,$F8,$99,$F8,$7D,$F9,$5F,$E8,$0D,$3F,$BE,$BF,$EC,$F7,$FB,$EB
	.dc.b	$9F,$7F,$D1,$EF,$C8,$C1,$C7,$E0,$7D,$84,$7A,$7C,$0F,$BF,$DC,$F8
	.dc.b	$1E,$EC,$9F,$04,$7C,$3C,$3A,$47,$C3,$C3,$0F,$52,$1F,$1F,$0F,$95
	.dc.b	$87,$CB,$F7,$E4,$FA,$DF,$9D,$EF,$FC,$FF,$66,$4F,$BD,$FA,$5E,$CF
	.dc.b	$D3,$F7,$FE,$A7,$BF,$2B,$E2,$47,$D7,$F8,$1F,$B1,$F0,$32,$7E,$44
	.dc.b	$74,$7E,$D6,$17,$93,$ED,$63,$D6,$7E,$76,$3D,$5F,$C0,$63,$F8,$3F
	.dc.b	$7B,$F0,$9F,$01,$27,$EA,$47,$3E,$C4,$7D,$5C,$B0,$FD,$7C,$7B,$3F
	.dc.b	$8B,$C7,$B6,$5F,$8B,$CB,$F2,$18,$7E,$47,$E0,$65,$FE,$06,$3A,$BF
	.dc.b	$E0,$1F,$A0,$7C,$06,$30,$FF,$53,$C3,$E0,$71,$7E,$49,$4E,$6F,$FD
	.dc.b	$96,$F2,$6F,$C3,$F0,$53,$C3,$E2,$30,$D8,$D9,$CB,$E0,$B3,$6F,$7E
	.dc.b	$98,$8C,$E7,$AC,$E5,$F8,$A3,$8E,$5C,$B2,$93,$D8,$F8,$1F,$01,$DB
	.dc.b	$E0,$64,$7A,$A3,$C3,$F0,$32,$3D,$AF,$93,$E0,$7C,$BF,$05,$E6,$F2
	.dc.b	$65,$7C,$28,$F4,$FB,$D2,$3E,$2B,$EC,$F7,$BA,$47,$D1,$C2,$F1,$7D
	.dc.b	$0C,$7A,$CF,$CC,$C7,$AB,$F5,$F1,$F8,$BF,$01,$23,$E8,$BF,$23,$E0
	.dc.b	$3E,$4C,$7C,$AF,$6A,$3F,$58,$F9,$BE,$FC,$7E,$CB,$F4,$3E,$03,$E8
	.dc.b	$FC,$17,$D2,$F8,$29,$5F,$7A,$3A,$BF,$80,$BE,$AF,$E2,$CE,$87,$E3
	.dc.b	$6E,$EB,$F8,$CC,$3F,$03,$F4,$3F,$DB,$0F,$ED,$55,$FE,$53,$6C,$FF
	.dc.b	$BE,$1F,$EE,$3D,$5F,$98,$9F,$CC,$61,$E0,$AF,$CB,$C7,$6E,$BF,$CB
	.dc.b	$53,$1C,$4F,$E4,$A3,$FD,$D2,$FE,$6E,$7F,$39,$87,$F7,$58,$C1,$F9
	.dc.b	$87,$FD,$9E,$F9,$7E,$0F,$D3,$F0,$7E,$AF,$7F,$D7,$EF,$CA,$E3,$9B
	.dc.b	$05,$7C,$0F,$81,$95,$A9,$FF,$6F,$0F,$81,$F8,$BF,$03,$F1,$BD,$D9
	.dc.b	$5B,$8F,$03,$C7,$2F,$03,$F2,$71,$EA,$FC,$BC,$6E,$5F,$2F,$1F,$A3
	.dc.b	$EF,$4A,$F1,$3F,$53,$DF,$FA,$BE,$FF,$D6,$F7,$E5,$79,$E3,$EC,$FC
	.dc.b	$0C,$8F,$53,$F7,$3E,$07,$EE,$FB,$D2,$7D,$CF,$E0,$3D,$EE,$B1,$F0
	.dc.b	$70,$EA,$FE,$07,$0E,$8F,$E0,$F0,$BA,$7E,$0F,$1F,$C4,$FB,$F2,$BE
	.dc.b	$53,$F8,$CF,$80,$FE,$1B,$EE,$BD,$BA,$4A,$FA,$B0,$F6,$F7,$F0,$1C
	.dc.b	$3E,$1E,$56,$A3,$9F,$C3,$C8,$E2,$FE,$6B,$E0,$FA,$47,$FA,$AF,$0F
	.dc.b	$03,$DB,$1F,$1E,$3D,$A3,$C3,$8F,$67,$86,$3E,$8F,$81,$F4,$FC,$0C
	.dc.b	$9F,$3C,$76,$DD,$E0,$7F,$33,$CF,$BD,$78,$73,$EE,$FA,$4E,$E4,$3B
	.dc.b	$9F,$0A,$3E,$3F,$C1,$48,$C1,$F1,$BD,$FF,$97,$F0,$12,$30,$7C,$AF
	.dc.b	$82,$F9,$E4,$7E,$6F,$C1,$7D,$2F,$76,$47,$D0,$7E,$AF,$BD,$F5,$B0
	.dc.b	$E9,$3F,$27,$0C,$3D,$48,$7C,$DC,$3E,$EE,$1F,$7B,$E0,$24,$FD,$C7
	.dc.b	$F0,$3F,$01,$F8,$2D,$BF,$07,$F0,$12,$BF,$03,$1F,$86,$F8,$1F,$C3
	.dc.b	$FC,$14,$9F,$C2,$C7,$F5,$6F,$A2,$FF,$63,$93,$DF,$C7,$BF,$D6,$BF
	.dc.b	$17,$8F,$57,$F0,$58,$F5,$7F,$1F,$8F,$57,$F0,$D8,$F5,$F8,$0E,$DF
	.dc.b	$03,$27,$F8,$4E,$7F,$DA,$DF,$F7,$B2,$F6,$BC,$9E,$DC,$9D,$C7,$9F
	.dc.b	$E0,$24,$72,$7D,$5F,$03,$EB,$F8,$29,$3D,$9F,$77,$C1,$FB,$EF,$A4
	.dc.b	$F8,$30,$C3,$CC,$87,$93,$0E,$AF,$A3,$E0,$7E,$47,$C0,$C8,$C1,$E6
	.dc.b	$3D,$B5,$DB,$BF,$F3,$71,$FE,$06,$D3,$DC,$F6,$8C,$88,$FF,$78,$47
	.dc.b	$84,$F8,$FA,$7D,$7F,$83,$91,$F2,$9F,$B5,$F0,$7F,$6F,$E0,$24,$FC
	.dc.b	$E7,$EF,$7C,$0F,$DF,$F8,$3F,$C0,$7C,$1C,$AF,$A7,$1F,$83,$F8,$19
	.dc.b	$1F,$59,$FC,$37,$C1,$F4,$8F,$AF,$85,$D7,$D3,$AB,$2E,$D5,$F7,$B2
	.dc.b	$EC,$FD,$EC,$BF,$20,$FE,$47,$E0,$E3,$F8,$43,$FC,$53,$F4,$7F,$D5
	.dc.b	$DF,$F1,$6F,$5F,$F8,$D6,$3F,$AB,$E5,$F2,$1B,$19,$F8,$2E,$3F,$0F
	.dc.b	$2F,$FA,$BD,$7F,$B7,$8F,$EE,$35,$CF,$DE,$EE,$49,$DF,$BD,$E2,$C3
	.dc.b	$C7,$F0,$92,$79,$C7,$57,$CA,$7F,$90,$B0,$ED,$8F,$AB,$E1,$FD,$7F
	.dc.b	$03,$2B,$C6,$FB,$BE,$0B,$DF,$EE,$7C,$0F,$72,$57,$A2,$3E,$1F,$C1
	.dc.b	$7C,$4F,$82,$93,$EC,$8F,$8F,$F0,$72,$3D,$EF,$CA,$F8,$4E,$B1,$F0
	.dc.b	$B0,$EA,$FA,$38,$DB,$32,$EF,$3F,$23,$2E,$EE,$E5,$97,$D6,$F8,$19
	.dc.b	$3F,$39,$FB,$3F,$05,$E0,$8E,$19,$78,$32,$FB,$19,$76,$9E,$99,$76
	.dc.b	$9F,$B9,$8F,$E0,$BE,$13,$F0,$7F,$09,$2F,$ED,$C7,$F9,$BB,$E9,$3F
	.dc.b	$D9,$B1,$F4,$7E,$0B,$F1,$26,$7E,$0F,$F1,$7F,$0B,$27,$F0,$6F,$E3
	.dc.b	$BE,$17,$97,$C1,$C9,$FF,$0E,$E1,$FE,$CA,$D0,$FF,$B2,$A3,$E4,$3E
	.dc.b	$08,$F9,$38,$F8,$27,$8E,$3D,$DF,$91,$8D,$AF,$CC,$63,$E9,$F8,$1F
	.dc.b	$57,$C2,$CB,$F0,$C7,$B7,$E0,$A4,$79,$1F,$81,$F0,$5F,$07,$E0,$FE
	.dc.b	$17,$C2,$4A,$F4,$C7,$C5,$F8,$0F,$8D,$F0,$12,$7D,$B1,$D1,$F5,$E1
	.dc.b	$74,$FE,$27,$C0,$CB,$B5,$7C,$4C,$BB,$3F,$0B,$2F,$A5,$F0,$52,$3E
	.dc.b	$4B,$F5,$7E,$0B,$EB,$7B,$92,$7E,$6B,$F6,$7D,$DF,$B5,$F0,$52,$7E
	.dc.b	$8B,$F7,$7E,$0B,$EF,$7C,$27,$DF,$F8,$49,$5F,$5A,$3A,$3F,$53,$0B
	.dc.b	$BF,$AB,$1E,$93,$F6,$31,$B3,$EE,$C7,$F1,$5E,$6C,$AF,$C0,$BF,$8D
	.dc.b	$F8,$2E,$6F,$FB,$17,$85,$96,$2B,$FD,$83,$0C,$BA,$D6,$F2,$B5,$AC
	.dc.b	$7F,$D8,$BE,$2B,$E1,$2F,$C2,$4A,$FF,$36,$CE,$5B,$F0,$5C,$3E,$0A
	.dc.b	$46,$0D,$FC,$27,$42,$3C,$BE,$13,$BF,$C1,$F8,$3E,$02,$4F,$68,$F1
	.dc.b	$E1,$D2,$3F,$82,$63,$FE,$D5,$7B,$18,$C3,$AB,$4E,$C7,$A3,$E1,$3D
	.dc.b	$BF,$05,$23,$07,$B3,$E0,$BE,$09,$1F,$7F,$C1,$7C,$4F,$84,$91,$F0
	.dc.b	$9F,$8F,$F0,$9F,$23,$DE,$93,$F1,$9F,$97,$EF,$F3,$8F,$4E,$17,$AF
	.dc.b	$FB,$6E,$AC,$BA,$D7,$BB,$2B,$5A,$CB,$EA,$FC,$17,$D6,$F8,$19,$7F
	.dc.b	$52,$3F,$DF,$BF,$4A,$F7,$69,$27,$FD,$4B,$1F,$D3,$96,$0C,$FF,$D7
	.dc.b	$CF,$92,$3F,$06,$7E,$13,$E0,$BF,$0B,$F0,$B2,$3F,$06,$FE,$23,$E1
	.dc.b	$7B,$47,$F3,$9C,$3C,$4F,$83,$1F,$2F,$C0,$74,$8F,$C5,$94,$FB,$B8
	.dc.b	$7A,$BE,$06,$47,$A1,$F6,$FC,$17,$6C,$FF,$C2,$B8,$F6,$AF,$06,$3D
	.dc.b	$9E,$A4,$3F,$12,$7A,$CF,$87,$F0,$9F,$20,$8F,$C6,$F8,$4F,$97,$F0
	.dc.b	$5F,$33,$79,$3F,$2A,$3E,$7F,$C2,$48,$F9,$AF,$D2,$F8,$4F,$A7,$EF
	.dc.b	$7D,$4F,$7E,$57,$D2,$8E,$AF,$D4,$C3,$AB,$F2,$70,$E8,$FC,$BC,$2E
	.dc.b	$BF,$03,$1F,$BF,$F0,$1F,$80,$F8,$19,$7F,$7A,$3F,$07,$F0,$12,$3F
	.dc.b	$02,$FE,$1B,$E0,$7F,$0F,$F0,$7F,$88,$F8,$39,$5F,$86,$8F,$C5,$FC
	.dc.b	$04,$8F,$C4,$BF,$8E,$F8,$0E,$91,$F7,$F0,$BA,$7D,$4C,$7A,$CF,$E4
	.dc.b	$31,$EA,$FE,$17,$13,$FE,$F8,$FA,$6F,$EB,$1C,$CF,$83,$FC,$D9,$A9
	.dc.b	$FF,$02,$4F,$F5,$D1,$ED,$E7,$2F,$F2,$BC,$62,$7A,$4E,$7D,$3E,$F4
	.dc.b	$A7,$B9,$9B,$FD,$1C,$7F,$AF,$27,$9F,$C0,$E5,$E0,$D8,$FF,$6D,$4F
	.dc.b	$A7,$E1,$31,$F0,$F2,$30,$78,$88,$7A,$B2,$E5,$8F,$3C,$24,$FC,$37
	.dc.b	$B7,$C2,$77,$F8,$2F,$07,$C1,$4A,$F9,$11,$E3,$C2,$47,$6D,$3E,$1E
	.dc.b	$3D,$23,$E4,$E3,$67,$E3,$E3,$EB,$F8,$39,$5F,$4D,$F7,$7C,$27,$BF
	.dc.b	$E0,$64,$FD,$77,$E1,$7C,$17,$C3,$F8,$49,$3F,$6D,$F8,$DF,$09,$F1
	.dc.b	$FE,$02,$4F,$DF,$7E,$57,$C0,$74,$8F,$BB,$85,$D3,$ED,$E3,$D2,$7F
	.dc.b	$03,$8D,$9F,$BF,$8F,$D3,$F8,$3F,$A9,$F0,$52,$FF,$19,$1F,$67,$E9
	.dc.b	$FF,$AC,$47,$F9,$35,$B7,$FE,$9E,$6F,$47,$E1,$97,$C1,$7D,$F8,$FC
	.dc.b	$07,$C2,$CB,$FC,$C4,$75,$F7,$BB,$7C,$14,$9F,$C7,$CF,$39,$F1,$7C
	.dc.b	$24,$9C,$9F,$CC,$B1,$F0,$B7,$1F,$03,$FA,$5C,$7C,$0F,$E8,$B0,$F6
	.dc.b	$7C,$0C,$8E,$79,$B0,$4F,$C1,$F8,$4F,$85,$EE,$4B,$DD,$23,$E3,$7B
	.dc.b	$9F,$1F,$E0,$BE,$47,$C1,$4B,$DC,$7C,$BF,$83,$F9,$9F,$09,$27,$9C
	.dc.b	$75,$7E,$76,$17,$4F,$9D,$8F,$59,$F0,$E3,$D5,$FA,$98,$FD,$7F,$82
	.dc.b	$FB,$1F,$05,$27,$CF,$1C,$F7,$FA,$D9,$59,$F9,$58,$F5,$7E,$F6,$3D
	.dc.b	$72,$FB,$D9,$7E,$0F,$E1,$25,$7C,$27,$F0,$DF,$09,$F8,$7F,$82,$93
	.dc.b	$F1,$9F,$C5,$7C,$17,$E2,$FE,$0F,$F1,$9F,$09,$2B,$E5,$C7,$E3,$DF
	.dc.b	$C8,$7C,$1C,$7E,$79,$D0,$FC,$95,$DC,$FF,$25,$87,$59,$FA,$B8,$75
	.dc.b	$7F,$31,$87,$F1,$3C,$FF,$B0,$5F,$FB,$D9,$FE,$F8,$9A,$6B,$7F,$E4
	.dc.b	$2B,$E0,$39,$12,$FF,$13,$0F,$80,$ED,$F0,$72,$72,$F8,$7E,$0F,$C5
	.dc.b	$EF,$49,$E2,$F9,$7D,$EE,$91,$FE,$BB,$C3,$C0,$F6,$C7,$D5,$8F,$58
	.dc.b	$D6,$37,$3F,$66,$5E,$FF,$7E,$57,$99,$F8,$5F,$01,$F0,$FD,$9F,$89
	.dc.b	$ED,$4A,$F5,$C7,$C7,$F7,$E4,$7B,$9F,$95,$F0,$1F,$2F,$E0,$64,$FC
	.dc.b	$27,$E7,$7C,$17,$58,$F8,$B8,$75,$7D,$98,$74,$7D,$D8,$5D,$3E,$A6
	.dc.b	$3F,$5F,$E0,$3E,$C7,$C0,$4B,$F9,$F1,$D5,$F8,$D8,$F5,$BF,$B9,$8F
	.dc.b	$3E,$5F,$B3,$95,$BF,$F1,$CF,$1B,$3F,$C2,$7C,$17,$E1,$7E,$0A,$57
	.dc.b	$DC,$8F,$EB,$5F,$51,$FE,$C9,$3F,$D9,$19,$B2,$FE,$CB,$18,$6F,$C0
	.dc.b	$70,$F8,$7F,$C8,$7C,$1C,$C7,$F8,$7D,$FF,$B3,$CF,$ED,$17,$CB,$DD
	.dc.b	$EE,$78,$3D,$DF,$0E,$3E,$2F,$82,$98,$72,$8E,$CF,$E4,$B1,$EC,$FF
	.dc.b	$95,$B0,$F2,$E5,$EA,$F8,$0F,$5F,$C0,$4A,$F1,$C7,$BB,$E0,$3D,$FF
	.dc.b	$01,$27,$CF,$1F,$0B,$E0,$7E,$1F,$C1,$49,$F5,$C7,$C6,$F7,$BE,$3F
	.dc.b	$BD,$27,$DF,$1D,$5F,$83,$87,$57,$CC,$7F,$9B,$38,$98,$F7,$8F,$8F
	.dc.b	$8F,$77,$7C,$6A,$65,$F5,$BE,$03,$EB,$FC,$04,$BF,$9F,$1F,$6B,$82
	.dc.b	$47,$D2,$7E,$EF,$B5,$F7,$BD,$F9,$3F,$59,$FC,$0F,$C0,$7E,$0B,$E0
	.dc.b	$A4,$FD,$A7,$F0,$BF,$05,$D2,$3C,$BA,$7E,$15,$EC,$7E,$03,$0E,$AE
	.dc.b	$B0,$BA,$EC,$F6,$63,$FE,$29,$FA,$9F,$EB,$B0,$69,$3E,$03,$C1,$1C
	.dc.b	$71,$F0,$61,$E2,$D3,$DA,$FA,$C8,$78,$4F,$C9,$E1,$DF,$E0,$A4,$7F
	.dc.b	$64,$C5,$1E,$3F,$83,$F2,$7B,$D2,$B4,$F9,$FD,$FF,$47,$C1,$7A,$7E
	.dc.b	$0A,$57,$38,$F6,$78,$DE,$DF,$76,$4F,$78,$E8,$FD,$EC,$D8,$E3,$7F
	.dc.b	$8D,$0C,$BB,$D7,$9F,$2E,$EF,$1C,$BE,$57,$BF,$F2,$FD,$F9,$3E,$D8
	.dc.b	$F9,$DE,$FF,$CF,$F7,$E4,$FC,$18,$FA,$5F,$01,$F4,$FE,$06,$4F,$C5
	.dc.b	$8F,$AD,$EE,$FD,$7F,$7A,$4F,$C9,$8E,$AF,$CB,$C3,$AB,$ED,$C3,$AB
	.dc.b	$F3,$F0,$EA,$FC,$1C,$3F,$05,$F0,$1F,$83,$F8,$09,$3F,$56,$3F,$CD
	.dc.b	$DF,$54,$F7,$69,$27,$FC,$3B,$0F,$77,$7F,$0F,$C3,$E2,$25,$6A,$3F
	.dc.b	$20,$FE,$47,$E1,$23,$F8,$33,$A1,$FB,$2B,$F0,$3D,$B0,$F1,$E1,$DA
	.dc.b	$3C,$38,$76,$78,$61,$E8,$F8,$2F,$4F,$C1,$49,$F3,$C7,$6E,$9F,$D2
	.dc.b	$65,$DA,$7C,$19,$76,$7D,$07,$62,$1D,$8F,$83,$F0,$9F,$1B,$E1,$24
	.dc.b	$60,$F8,$BF,$05,$F2,$88,$FC,$8F,$82,$F9,$BF,$09,$23,$E5,$BF,$43
	.dc.b	$E1,$3E,$8F,$C0,$7D,$2F,$80,$95,$F4,$23,$9B,$F1,$F0,$BA,$FB,$31
	.dc.b	$E9,$3F,$2F,$1B,$3F,$03,$1F,$BB,$F0,$3F,$7B,$E0,$A5,$FD,$C8,$FC
	.dc.b	$0F,$C0,$C8,$FB,$EF,$E1,$3E,$0B,$F0,$BF,$07,$F8,$6F,$84,$95,$F8
	.dc.b	$48,$FF,$5B,$7B,$4F,$7F,$F1,$7F,$01,$2B,$FC,$CB,$39,$7A,$90,$FC
	.dc.b	$16,$3D,$DE,$87,$E4,$30,$EE,$FE,$1F,$0E,$FF,$03,$E0,$F8,$29,$3D
	.dc.b	$A3,$C7,$EE,$48,$F0,$BE,$6F,$77,$CF,$F0,$32,$7C,$AF,$AB,$E0,$BD
	.dc.b	$7F,$09,$EC,$F8,$49,$5E,$A8,$E8,$FF,$4A,$DB,$FA,$4B,$D8,$F7,$E1
	.dc.b	$D5,$D6,$17,$97,$F9,$B5,$7F,$B8,$A7,$E2,$FC,$17,$CA,$89,$18,$3A
	.dc.b	$1D,$F3,$F0,$D7,$2C,$3A,$7B,$5F,$EE,$EB,$FF,$79,$5F,$80,$FA,$3A
	.dc.b	$7D,$6F,$83,$91,$F5,$1F,$B3,$F0,$9F,$6B,$E0,$A4,$FD,$87,$EE,$FC
	.dc.b	$17,$DE,$F8,$3F,$BF,$F0,$92,$BE,$EC,$7E,$0B,$E0,$E4,$7E,$01,$FC
	.dc.b	$2F,$C0,$F4,$8F,$91,$85,$DF,$E1,$E3,$D6,$7F,$0F,$8F,$57,$E7,$63
	.dc.b	$F8,$ED,$BF,$1F,$F0,$92,$7F,$1B,$07,$FD,$67,$F5,$7F,$D4,$1B,$F0
	.dc.b	$79,$7F,$C0,$31,$96,$FC,$1E,$73,$FE,$D5,$3F,$DA,$09,$F0,$F9,$C3
	.dc.b	$B4,$A7,$FD,$AF,$7E,$0F,$7B,$4A,$75,$18,$6C,$79,$BD,$FC,$57,$8E
	.dc.b	$7C,$39,$60,$CE,$5A,$CA,$61,$EE,$78,$FC,$37,$2F,$83,$93,$F0,$9E
	.dc.b	$BF,$05,$DB,$DB,$93,$F1,$9F,$0F,$B9,$E2,$F8,$19,$3F,$29,$F2,$FC
	.dc.b	$17,$9B,$E1,$3C,$FF,$0B,$2B,$E7,$C7,$47,$E6,$E1,$77,$F9,$78,$F4
	.dc.b	$9F,$A3,$8D,$9F,$9F,$8F,$C2,$F8,$1F,$87,$F0,$52,$FE,$DC,$73,$E5
	.dc.b	$FA,$3C,$6C,$F1,$5F,$3E,$19,$F5,$AD,$E7,$6B,$59,$7D,$0F,$83,$FA
	.dc.b	$3F,$09,$27,$F1,$31,$F5,$18,$7E,$30,$FA,$FF,$05,$F6,$3E,$0E,$47
	.dc.b	$E4,$1F,$B7,$F0,$9F,$7F,$EB,$3F,$D8,$53,$F8,$37,$FD,$8E,$4F,$83
	.dc.b	$EB,$1F,$94,$2F,$17,$E4,$B0,$EB,$1F,$98,$C2,$CF,$E5,$B0,$ED,$F0
	.dc.b	$12,$BF,$2B,$B7,$E5,$B2,$FE,$C1,$7F,$E4,$57,$FB,$D4,$E8,$F2,$7C
	.dc.b	$07,$9C,$8E,$65,$2F,$80,$F6,$7C,$17,$B7,$E0,$E5,$EE,$BF,$D9,$EF
	.dc.b	$F9,$B9,$7B,$DF,$0C,$91,$BF,$7B,$BB,$BC,$AE,$5E,$7C,$BB,$CF,$5C
	.dc.b	$7B,$BD,$72,$F9,$DF,$01,$F3,$FE,$02,$4F,$8E,$3E,$97,$B3,$F4,$FD
	.dc.b	$A9,$3E,$78,$FA,$DE,$FC,$8F,$53,$F6,$7E,$03,$ED,$7C,$0C,$9F,$73
	.dc.b	$F7,$7E,$0B,$AC,$7C,$1C,$3A,$BF,$07,$0E,$8F,$C3,$C2,$ED,$F4,$F1
	.dc.b	$FC,$37,$BF,$2B,$E5,$3F,$89,$F8,$2E,$B1,$F2,$F1,$EB,$7F,$6F,$1E
	.dc.b	$7C,$BF,$7F,$2B,$7F,$8F,$CB,$F8,$87,$D6,$FF,$B1,$A3,$F8,$A2,$F7
	.dc.b	$F2,$6B,$DF,$DF,$C0,$70,$F8,$19,$7F,$83,$B4,$F4,$F8,$3F,$CF,$7C
	.dc.b	$1C,$AD,$47,$83,$E0,$A4,$71,$7C,$7F,$01,$E4,$F8,$29,$3D,$5F,$3F
	.dc.b	$C1,$7A,$3E,$0E,$4F,$85,$F5,$FC,$1F,$B3,$DF,$F6,$FB,$F2,$BC,$D1
	.dc.b	$F0,$3D,$E9,$1E,$97,$E1,$FB,$FF,$13,$E0,$E4,$FB,$5F,$8F,$F0,$7F
	.dc.b	$23,$E0,$FE,$4F,$C1,$CA,$F8,$51,$FC,$F9,$7D,$73,$C1,$FE,$81,$0F
	.dc.b	$AF,$7B,$3F,$D4,$7E,$C9,$DB,$FD,$4E,$73,$5F,$D6,$67,$2F,$F5,$B8
	.dc.b	$CB,$DA,$FA,$C7,$0F,$07,$F5,$BF,$03,$FD,$37,$CE,$FE,$A7,$7F,$ED
	.dc.b	$E6,$CF,$FB,$15,$FF,$61,$F6,$FF,$AF,$A3,$FD,$90,$DF,$0B,$FA,$EB
	.dc.b	$6B,$F0,$FF,$0D,$EF,$DB,$DD,$7E,$5D,$BD,$9B,$7A,$DB,$79,$3F,$BA
	.dc.b	$57,$07,$E2,$47,$F6,$AC,$BF,$DA,$A7,$9B,$2C,$7C,$17,$AF,$4F,$67
	.dc.b	$C2,$4C,$3F,$8A,$5F,$F6,$7B,$E5,$F0,$9E,$5E,$0F,$8B,$97,$8A,$FB
	.dc.b	$9F,$37,$4F,$33,$E4,$9D,$BF,$DD,$F0,$F1,$E7,$F2,$CF,$14,$78,$3E
	.dc.b	$1B,$E7,$11,$F1,$FC,$37,$87,$7F,$81,$0E,$3F,$B6,$F8,$8C,$65,$E2
	.dc.b	$7E,$E4,$FC,$73,$F9,$3F,$C1,$FD,$B2,$47,$C0,$F8,$4F,$BD,$F0,$D2
	.dc.b	$3E,$1B,$F8,$1F,$86,$F1,$47,$7C,$7C,$3C,$FF,$8A,$6F,$07,$F8,$8F
	.dc.b	$97,$EA,$4F,$E2,$F8,$19,$F8,$30,$FC,$53,$F8,$DC,$FF,$1D,$F0,$52
	.dc.b	$7E,$9B,$F9,$1F,$82,$F0,$E1,$DF,$7F,$76,$9E,$0B,$FA,$1B,$FB,$F3
	.dc.b	$EF,$1F,$73,$3F,$04,$FE,$5F,$3F,$CE,$7C,$2F,$E7,$7E,$12,$5F,$E0
	.dc.b	$A3,$BE,$FF,$7F,$4F,$04,$FD,$2D,$FD,$B3,$F8,$FF,$83,$FD,$37,$C2
	.dc.b	$4A,$FC,$5C,$7E,$A6,$7F,$55,$F0,$D2,$7F,$1F,$1D,$9F,$C5,$1F,$7B
	.dc.b	$7F,$63,$F6,$8F,$C7,$EB,$FD,$82,$FB,$47,$EC,$F0,$FD,$B4,$7E,$DF
	.dc.b	$E0,$E4,$FE,$71,$FD,$D7,$C1,$F7,$8F,$C3,$9D,$F8,$BF,$11,$A7,$B2
	.dc.b	$3F,$31,$7D,$B1,$FD,$F6,$3D,$A7,$F4,$E7,$F0,$52,$1F,$AA,$F8,$4E
	.dc.b	$D9,$FA,$8F,$5E,$1D,$6D,$F6,$2F,$80,$FF,$2D,$2F,$89,$A6,$23,$3F
	.dc.b	$15,$32,$FF,$05,$57,$F8,$41,$FF,$32,$78,$3F,$E6,$A8,$FF,$12,$C7
	.dc.b	$C9,$E3,$FB,$2F,$E6,$77,$F7,$F8,$1F,$B6,$CF,$C7,$97,$82,$BA,$E7
	.dc.b	$E1,$7B,$FC,$1F,$98,$8F,$8B,$E0,$FD,$F8,$FE,$50,$EB,$87,$EC,$73
	.dc.b	$F7,$57,$A3,$1E,$8F,$A0,$FF,$3E,$47,$F8,$B3,$4F,$57,$C2,$FC,$42
	.dc.b	$4F,$C0,$F8,$5E,$EF,$D9,$3F,$29,$AF,$AE,$DF,$09,$F3,$09,$3F,$22
	.dc.b	$7E,$7F,$C2,$FD,$0E,$F9,$3F,$32,$3A,$BF,$1C,$EB,$A7,$AE,$F9,$C7
	.dc.b	$A6,$3E,$DC,$7D,$AB,$FB,$7A,$FD,$CF,$85,$97,$F6,$23,$AF,$17,$D4
	.dc.b	$3F,$A9,$F2,$7E,$17,$4F,$EC,$15,$F7,$B1,$EF,$93,$7F,$BD,$9F,$E2
	.dc.b	$3E,$0F,$F1,$3F,$05,$30,$FC,$2C,$76,$ED,$FD,$D1,$FC,$56,$BE,$37
	.dc.b	$BD,$F9,$0F,$85,$97,$F8,$C8,$FC,$9F,$C2,$FE,$53,$E1,$64,$FE,$42
	.dc.b	$3A,$3F,$B8,$3F,$99,$6F,$E9,$AE,$71,$F1,$77,$61,$F9,$E9,$FF,$5C
	.dc.b	$7D,$93,$E1,$29,$2F,$FC,$CD,$0F,$83,$EF,$1F,$B8,$3B,$E9,$EA,$7F
	.dc.b	$47,$87,$83,$87,$E3,$47,$E3,$31,$EF,$F0,$D2,$3A,$BE,$2F,$85,$EB
	.dc.b	$E0,$FE,$DB,$1F,$D2,$66,$F8,$2F,$3F,$C1,$4C,$7C,$B1,$EA,$F8,$2F
	.dc.b	$5F,$C3,$49,$F4,$C7,$4E,$BF,$D1,$C7,$F1,$AC,$3A,$3E,$EC,$3A,$3F
	.dc.b	$8B,$8C,$B7,$E1,$7E,$3F,$C2,$C8,$F8,$B1,$D5,$FC,$A9,$D7,$2E,$7C
	.dc.b	$1F,$08,$5B,$0F,$A3,$3F,$4B,$E0,$25,$FD,$07,$EA,$FC,$37,$58,$F9
	.dc.b	$F8,$7F,$6B,$C1,$F0,$DF,$68,$93,$F6,$3E,$17,$EE,$FC,$2C,$8F,$B6
	.dc.b	$FE,$03,$E1,$7A,$78,$7F,$A0,$E3,$F9,$F8,$72,$BF,$84,$FE,$1B,$0F
	.dc.b	$C3,$FC,$2F,$E2,$3E,$1A,$57,$E1,$A3,$9F,$83,$F9,$DD,$7C,$D8,$73
	.dc.b	$7E,$16,$BE,$8B,$F3,$1F,$8E,$F8,$4F,$C9,$91,$FC,$8F,$C3,$74,$7F
	.dc.b	$B6,$9F,$D4,$BE,$CD,$F0,$C8,$FF,$5E,$46,$1B,$19,$7F,$83,$43,$E1
	.dc.b	$71,$8F,$D5,$8F,$F1,$B4,$A1,$18,$6F,$C2,$FE,$A3,$CA,$F2,$57,$F3
	.dc.b	$DD,$7C,$55,$C3,$1E,$B5,$8C,$3C,$97,$E2,$9C,$BD,$E3,$C8,$FA,$BE
	.dc.b	$13,$37,$E2,$6C,$9B,$F0,$DE,$2F,$23,$FA,$E4,$F2,$BE,$C7,$C2,$3E
	.dc.b	$2E,$9C,$BE,$12,$47,$C8,$7A,$FC,$37,$3E,$6F,$EA,$96,$F8,$4F,$0F
	.dc.b	$C2,$4C,$3E,$7C,$79,$3E,$13,$CB,$F0,$D2,$7E,$9C,$73,$DF,$FC,$85
	.dc.b	$87,$39,$FA,$F8,$73,$7E,$AE,$1E,$EF,$85,$F7,$FC,$34,$9F,$BB,$1C
	.dc.b	$DF,$C1,$4F,$1C,$3E,$D5,$FC,$68,$F8,$FF,$0B,$F2,$3E,$1A,$5F,$E1
	.dc.b	$A3,$97,$7F,$FA,$3F,$37,$C2,$FC,$EF,$85,$97,$F8,$C8,$FA,$3F,$0B
	.dc.b	$F4,$BA,$A4,$FE,$42,$39,$C7,$DF,$9E,$33,$F9,$0A,$FB,$31,$F6,$BE
	.dc.b	$17,$ED,$FC,$34,$AF,$CC,$C7,$27,$F0,$93,$CB,$7F,$CB,$5F,$29,$FC
	.dc.b	$7F,$67,$E4,$37,$FF,$0C,$47,$E8,$8F,$C3,$9F,$8A,$72,$FF,$7F,$3F
	.dc.b	$30,$7F,$49,$D7,$FB,$16,$7F,$B9,$8F,$F1,$FC,$FF,$82,$1F,$ED,$CD
	.dc.b	$C3,$F8,$E3,$6F,$C5,$F1,$3D,$FE,$1F,$C1,$F1,$53,$0F,$39,$FA,$DB
	.dc.b	$FF,$14,$61,$FD,$1A,$3C,$67,$F4,$DC,$BC,$6F,$F0,$07,$FD,$89,$3D
	.dc.b	$E3,$97,$C2,$FB,$4F,$76,$3E,$FF,$87,$F8,$1F,$0F,$2B,$C3,$8F,$E6
	.dc.b	$E7,$E7,$3F,$9D,$BE,$53,$AC,$3B,$4F,$E8,$70,$F9,$5F,$0F,$F2,$FE
	.dc.b	$1E,$5E,$E3,$9D,$FF,$82,$23,$E3,$F6,$FD,$13,$E9,$7C,$27,$D3,$F8
	.dc.b	$69,$3D,$DF,$AD,$F0,$FC,$AB,$84,$F8,$DE,$2F,$97,$D8,$FB,$84,$7D
	.dc.b	$1F,$0F,$C5,$E0,$43,$EC,$E1,$F8,$3F,$85,$FC,$27,$C3,$49,$F8,$11
	.dc.b	$F1,$3E,$D3,$FC,$01,$1F,$0F,$E2,$3F,$14,$7F,$14,$8C,$C7,$E1,$3E
	.dc.b	$13,$F1,$DF,$11,$23,$E5,$47,$3F,$86,$E0,$48,$F9,$DF,$0D,$CF,$E1
	.dc.b	$64,$7D,$6C,$11,$DF,$E1,$F9,$4E,$6F,$93,$F9,$4B,$F2,$7C,$27,$97
	.dc.b	$E1,$64,$F3,$8F,$67,$C2,$7A,$7E,$1E,$4F,$78,$F6,$7C,$27,$B7,$E1
	.dc.b	$64,$F8,$E3,$8B,$B2,$1F,$E5,$1D,$5F,$0D,$F1,$7E,$1A,$4F,$AD,$F9
	.dc.b	$1F,$0D,$CA,$38,$63,$F2,$CF,$99,$5F,$37,$E1,$A4,$FC,$37,$E8,$7C
	.dc.b	$37,$D5,$F8,$0F,$A5,$F0,$D2,$BE,$44,$7D,$5F,$86,$91,$F2,$DF,$B1
	.dc.b	$F0,$BC,$63,$E7,$57,$17,$D5,$5F,$77,$CD,$FB,$DF,$09,$27,$EA,$47
	.dc.b	$17,$67,$80,$FB,$1F,$0D,$FD,$63,$ED,$7F,$AF,$F9,$5F,$E2,$17,$FB
	.dc.b	$51,$3E,$17,$F6,$2B,$E0,$38,$12,$FF,$6F,$1F,$D7,$DB,$F0,$9D,$7E
	.dc.b	$0E,$46,$0F,$E1,$A7,$F0,$AF,$88,$FF,$60,$79,$1F,$E4,$0F,$4D,$EB
	.dc.b	$7F,$76,$C7,$C1,$7D,$B2,$F0,$3F,$E8,$B8,$F0,$37,$E0,$FD,$84,$7C
	.dc.b	$BF,$09,$EA,$AE,$3B,$7B,$DF,$85,$F0,$DF,$0F,$E0,$A5,$FB,$1F,$8D
	.dc.b	$F0,$BF,$0B,$E1,$3E,$41,$27,$E0,$FC,$27,$CB,$F8,$5F,$99,$EB,$49
	.dc.b	$F8,$B1,$F2,$A7,$84,$7A,$6B,$E9,$47,$D3,$DE,$4F,$CC,$7E,$B7,$C2
	.dc.b	$F1,$8F,$9F,$5C,$B5,$FB,$17,$F6,$FE,$0E,$57,$D4,$7E,$F7,$C1,$7D
	.dc.b	$AF,$82,$FC,$01,$27,$EC,$FC,$27,$E0,$D8,$7D,$C3,$F0,$DF,$09,$FD
	.dc.b	$7D,$FE,$83,$A3,$0F,$F1,$C7,$37,$D8,$9F,$F1,$F4,$E5,$FF,$37,$C6
	.dc.b	$73,$7D,$B3,$E1,$3F,$C5,$2F,$F6,$08,$FF,$39,$37,$BF,$FC,$6E,$FF
	.dc.b	$BD,$BE,$1D,$EE,$FF,$90,$71,$FF,$68,$13,$1F,$F3,$36,$1E,$0C,$BF
	.dc.b	$C9,$87,$8E,$53,$E7,$7F,$1F,$9F,$E4,$67,$BF,$C1,$79,$3E,$1A,$57
	.dc.b	$F5,$4C,$FF,$DE,$0F,$F5,$FA,$FF,$63,$BF,$DD,$5F,$EE,$98,$78,$A3
	.dc.b	$BC,$78,$B0,$F8,$ED,$8F,$31,$FE,$5B,$A7,$C3,$7C,$42,$47,$87,$E1
	.dc.b	$7E,$96,$BF,$0B,$4F,$F3,$66,$5E,$BC,$7D,$CD,$CB,$D9,$87,$9C,$FF
	.dc.b	$39,$5B,$E0,$7C,$71,$87,$6F,$ED,$7C,$1C,$BF,$DF,$F0,$F6,$91,$FB
	.dc.b	$F7,$A8,$F0,$99,$8F,$F0,$3D,$F7,$7C,$F1,$FE,$7F,$8C,$47,$FA,$A0
	.dc.b	$98,$7F,$83,$1F,$01,$E9,$3F,$23,$5D,$CF,$3C,$FC,$38,$FF,$76,$E1
	.dc.b	$E8,$8E,$B5,$F4,$C9,$7F,$62,$BA,$9E,$6D,$FF,$19,$3F,$26,$3C,$E7
	.dc.b	$B8,$FA,$E7,$B2,$FB,$1E,$29,$E3,$5D,$7C,$2F,$F7,$36,$7C,$23,$E2
	.dc.b	$DF,$33,$DE,$FC,$7C,$7C,$EF,$C5,$C3,$A5,$78,$AB,$CF,$7D,$1E,$27
	.dc.b	$3C,$7D,$0E,$0F,$31,$C7,$0E,$AF,$CB,$7E,$44,$7C,$F9,$E8,$FC,$69
	.dc.b	$E9,$87,$E5,$DE,$95,$C2,$3C,$D1,$D7,$0F,$3B,$FB,$38,$F9,$D3,$B8
	.dc.b	$FB,$51,$F0,$9F,$D8,$9F,$32,$34,$9F,$3B,$67,$AD,$79,$F3,$B7,$E5
	.dc.b	$8F,$82,$7A,$6B,$E0,$9F,$84,$3B,$65,$F5,$9F,$DC,$10,$FC,$16,$5F
	.dc.b	$3E,$BC,$C7,$9F,$1F,$29,$31,$FC,$0D,$74,$AF,$AD,$97,$F5,$C8,$C1
	.dc.b	$FC,$08,$F9,$64,$3B,$9E,$52,$63,$E9,$BF,$87,$3C,$E7,$E8,$E3,$E9
	.dc.b	$7E,$49,$D4,$B5,$C6,$35,$5F,$28,$8E,$0E,$55,$D6,$78,$5F,$C1,$CF
	.dc.b	$F5,$D8,$7A,$A7,$F9,$46,$1C,$F1,$FA,$33,$F2,$70,$FC,$4E,$5F,$C0
	.dc.b	$4F,$93,$86,$A7,$ED,$1F,$87,$3A,$47,$E9,$F3,$D3,$C2,$32,$7F,$15
	.dc.b	$38,$1F,$0B,$2F,$9F,$3F,$C5,$AF,$C3,$3F,$8E,$BF,$4B,$B3,$E6,$DF
	.dc.b	$E6,$1C,$65,$F8,$B7,$B9,$F2,$A3,$F1,$87,$ED,$73,$F9,$79,$7E,$02
	.dc.b	$78,$CF,$A5,$E2,$FA,$5C,$9E,$93,$C2,$7E,$C6,$7E,$8E,$3E,$93,$E0
	.dc.b	$E1,$E8,$C3,$E9,$D7,$03,$E4,$9D,$0F,$E0,$73,$C1,$FB,$07,$AB,$3F
	.dc.b	$AC,$FB,$32,$FB,$D8,$FA,$F3,$DC,$FE,$26,$3F,$3D,$97,$E1,$67,$E7
	.dc.b	$E3,$F8,$0C,$FE,$D6,$1C,$B0,$F5,$57,$2A,$FD,$8E,$1D,$67,$D1,$8F
	.dc.b	$E4,$2B,$F9,$4D,$F1,$9F,$D0,$DF,$07,$F4,$95,$FA,$EC,$7E,$7C,$78
	.dc.b	$B0,$E9,$5F,$5E,$3E,$14,$7C,$F8,$F8,$E7,$9E,$38,$1E,$63,$F8,$A1
	.dc.b	$FB,$FA,$ED,$87,$A6,$BE,$FC,$EE,$B9,$3E,$AB,$F8,$8F,$23,$FA,$4D
	.dc.b	$F2,$7E,$56,$7F,$66,$FC,$E7,$E3,$EF,$99,$E6,$7F,$11,$3E,$A9,$E0
	.dc.b	$F8,$63,$81,$1C,$1F,$AF,$CF,$F0,$0F,$80,$FC,$06,$7F,$95,$BE,$D1
	.dc.b	$B3,$F2,$B7,$D9,$FA,$37,$D5,$FC,$6C,$F0,$7F,$1A,$70,$3E,$31,$E2
	.dc.b	$37,$1C,$70,$F8,$53,$F5,$4E,$51,$F5,$4F,$87,$3F,$A3,$8F,$87,$18
	.dc.b	$8E,$E7,$C3,$3B,$9E,$63,$83,$DF,$1F,$A7,$3D,$E3,$91,$DE,$7C,$D5
	.dc.b	$FA,$27,$F5,$19,$FE,$46,$3C,$F3,$F8,$09,$FE,$00,$70,$C7,$E5,$47
	.dc.b	$96,$B8,$61,$E0,$8E,$07,$EA,$33,$FE,$B5,$5C,$2F,$E3,$5F,$09,$FF
	.dc.b	$01,$65,$F4,$71,$F0,$D7,$B0,$87,$BC,$F8,$79,$7C,$A9,$E1,$97,$86
	.dc.b	$77,$5F,$AD,$7F,$C0,$18,$77,$7F,$C0,$59,$7C,$37,$A4,$FC,$3C,$3E
	.dc.b	$F5,$6D,$FE,$D1,$5B,$9F,$2D,$7D,$E3,$BD,$F8,$5F,$2C,$FD,$0C,$3C
	.dc.b	$B8,$FC,$29,$FD,$89,$EB,$C7,$B5,$7F,$45,$AD,$CF,$F8,$33,$2F,$90
	.dc.b	$EF,$3F,$E4,$8F,$E4,$4F,$CF,$1F,$C8,$72,$F5,$5F,$D5,$9D,$CF,$E2
	.dc.b	$4D,$E7,$F9,$BA,$DC,$7D,$49,$DC,$FC,$D8,$FC,$7C,$FC,$03,$6F,$8C
	.dc.b	$FE,$9E,$78,$E3,$F8,$C3,$F3,$2F,$D3,$3F,$23,$1F,$7C,$FF,$53,$3A
	.dc.b	$E7,$F2,$67,$A9,$BC,$3C,$78,$FC,$BA,$DE,$7F,$2D,$FE,$24,$FF,$2A
	.dc.b	$9D,$C7,$EC,$E7,$EC,$5F,$5C,$7C,$6F,$D7,$C7,$C7,$5F,$E6,$A3,$A9
	.dc.b	$F4,$4F,$AF,$8F,$8D,$F7,$9F,$2F,$2F,$7C,$7C,$5B,$EB,$97,$8E,$B7
	.dc.b	$5F,$2D,$FA,$38,$FE,$F6,$FD,$F8,$F8,$F2,$F4,$D7,$E7,$EB,$D3,$87
	.dc.b	$BF,$0F,$E0,$B3,$F5,$CE,$B3,$F1,$5E,$A7,$ED,$CF,$4D,$F5,$76,$7C
	.dc.b	$5C,$FF,$1F,$5F,$E2,$DC,$BD,$33,$FA,$8C,$7C,$79,$7E,$83,$1F,$1C
	.dc.b	$FC,$53,$E8,$9B,$3A,$9E,$9C,$3A,$C7,$F9,$EE,$FA,$E1,$FE,$53,$CF
	.dc.b	$E8,$E1,$FE,$5E,$AE,$B3,$E3,$C3,$D3,$9F,$CB,$7F,$A7,$57,$8F,$1E
	.dc.b	$B1,$FE,$79,$C7,$AE,$1F,$EA,$FC,$7A,$E7,$FE,$45,$8E,$A7,$8E,$BA
	.dc.b	$9E,$93,$AC,$7E,$EE,$37,$1D,$4D,$E5,$E9,$CF,$E2,$D6,$F1,$F1,$D7
	.dc.b	$F9,$BA,$BE,$BB,$FA,$73,$AE,$7F,$C8,$1F,$C7,$C7,$BD,$F9,$6F,$57
	.dc.b	$E5,$E7,$EF,$77,$8F,$A5,$DB,$EF,$C7,$C7,$7D,$67,$F0,$67,$EC,$CF
	.dc.b	$4C,$ED,$EB,$8F,$A5,$EA,$7B,$CE,$A7,$FB,$3F,$2F,$97,$5F,$D4,$AB
	.dc.b	$F9,$F4,$FE,$0F,$1F,$7E,$3F,$A7,$8F,$79,$FC,$C3,$2F,$8B,$1F,$E8
	.dc.b	$78,$F1,$9D,$72,$F4,$BB,$3C,$67,$CB,$BF,$EB,$D8,$7E,$0F,$0F,$F7
	.dc.b	$25,$FC,$5A,$F7,$BF,$B3,$7A,$9F,$EF,$93,$DF,$3F,$16,$7F,$73,$97
	.dc.b	$F0,$59,$DD,$7F,$A3,$B0,$F1,$D7,$F4,$2C,$7D,$37,$FE,$32,$BF,$8A
	.dc.b	$75,$3D,$F3,$B7,$D3,$7E,$F7,$6F,$57,$C6,$7B,$CF,$4C,$F8,$DD,$BF
	.dc.b	$E6,$47,$73,$D5,$DE,$1F,$C1,$71,$F8,$B1,$F9,$B9,$F8,$B5,$F7,$63
	.dc.b	$67,$D7,$9D,$CF,$E3,$E3,$D3,$1F,$D4,$DF,$4C,$FB,$E7,$EE,$BF,$C2
	.dc.b	$E3,$67,$F3,$B3,$AC,$7E,$6C,$DD,$7D,$7A,$FF,$70,$D7,$8E,$7E,$8C
	.dc.b	$FA,$6B,$FB,$5B,$B7,$FC,$A9,$3B,$AF,$F0,$A5,$FD,$D9,$F8,$B5,$FC
	.dc.b	$43,$2F,$F4,$FC,$7F,$3B,$3C,$75,$FA,$7C,$3C,$71,$F1,$63,$C7,$5F
	.dc.b	$83,$7A,$E1,$F4,$6B,$67,$C5,$3E,$5C,$EE,$BF,$CE,$B3,$BA,$F9,$67
	.dc.b	$EB,$63,$E2,$9E,$3A,$F7,$BF,$44,$F7,$9E,$33,$DF,$5D,$5F,$F1,$89
	.dc.b	$FB,$93,$DE,$7F,$9B,$EF,$DF,$87,$F2,$9C,$7F,$2B,$5F,$D6,$33,$FF
	.dc.b	$4A,$E7,$F9,$AC,$BE,$E6,$3F,$10,$FD,$F6,$5E,$E8,$F4,$3E,$E8,$FC
	.dc.b	$A3,$E8,$3E,$E4,$7B,$A3,$F7,$07,$E5,$1E,$B1,$B7,$E8,$1B,$8F,$D3
	.dc.b	$47,$53,$F3,$46,$DF,$1C,$6F,$1F,$D3,$47,$CA,$7D,$CF,$E5,$30,$F7
	.dc.b	$3F,$EB,$BB,$F7,$3F,$93,$BE,$8F,$F0,$93,$E7,$BF,$6E,$FF,$03,$1E
	.dc.b	$E7,$F0,$35,$F9,$38,$FF,$1F,$1F,$CE,$67,$D1,$3E,$E3,$71,$F6,$CD
	.dc.b	$DF,$F2,$19,$DC,$78,$E7,$71,$F2,$A3,$C7,$1F,$40,$F1,$C7,$C4,$76
	.dc.b	$7C,$A8,$D9,$F4,$0F,$89,$9F,$F5,$DC,$BF,$03,$1F,$EC,$EC,$BF,$02
	.dc.b	$FD,$53,$C4,$7E,$63,$1F,$76,$3F,$E6,$7C,$7E,$24,$FE,$48,$F8,$86
	.dc.b	$CF,$74,$6F,$1F,$44,$6E,$7D,$0E,$DF,$73,$F1,$31,$FE,$11,$1F,$12
	.dc.b	$3F,$49,$3E,$E7,$F0,$27,$B8,$FC,$3B,$F1,$0F,$A0,$F5,$7E,$23,$D6
	.dc.b	$FF,$84,$3F,$ED,$2A,$FE,$10,$EC,$EA,$6C,$FD,$81,$B8,$FD,$84,$FF
	.dc.b	$5C,$AF,$4D,$FB,$EB,$E8,$47,$A6,$3F,$27,$1F,$AB,$3F,$1C,$75,$8F
	.dc.b	$96,$F5,$3E,$59,$FB,$C8,$FE,$35,$1F,$B1,$8E,$B3,$EF,$7E,$B9,$BB
	.dc.b	$FE,$1C,$FC,$5A,$FD,$2D,$78,$E7,$F0,$53,$FA,$53,$F0,$47,$E9,$4F
	.dc.b	$C1,$1F,$40,$F8,$87,$A2,$7D,$CF,$D0,$3E,$E5,$7D,$07,$C5,$7F,$88
	.dc.b	$C3,$F1,$B1,$F9,$28,$FF,$02,$62,$8F,$F7,$89,$FD,$B3,$0F,$EB,$4E
	.dc.b	$CF,$F0,$AD,$FC,$F3,$ED,$E1,$E2,$3D,$AF,$C3,$8F,$41,$E2,$3F,$27
	.dc.b	$87,$8B,$1F,$85,$8F,$8A,$7C,$27,$F2,$88,$FF,$72,$3F,$E6,$13,$E3
	.dc.b	$3F,$E1,$89,$E5,$5D,$EF,$EE,$CF,$F0,$A7,$FB,$D3,$F5,$EB,$E3,$1F
	.dc.b	$75,$FD,$0B,$E5,$7F,$13,$5D,$23,$E5,$9F,$72,$BE,$03,$F9,$4A,$F8
	.dc.b	$18,$7C,$6B,$F5,$CF,$C6,$C7,$E6,$3F,$EB,$FC,$3F,$CB,$EF,$E2,$0E
	.dc.b	$A7,$C6,$C3,$AB,$F7,$63,$C4,$F9,$E3,$EE,$1A,$C3,$DA,$FF,$31,$C3
	.dc.b	$C4,$FC,$37,$F1,$B3,$F9,$F9,$DB,$F9,$38,$DB,$E2,$8F,$93,$3F,$BD
	.dc.b	$8F,$11,$F9,$33,$F0,$58,$7F,$B2,$23,$C5,$5F,$E0,$F7,$C5,$8F,$F7
	.dc.b	$DC,$BE,$0D,$7F,$AA,$CF,$A9,$5B,$BF,$15,$6D,$FF,$6A,$5F,$87,$0F
	.dc.b	$3C,$7E,$D6,$3F,$8C,$BF,$0C,$FA,$99,$76,$8F,$F5,$B6,$5F,$3A,$7E
	.dc.b	$FE,$7F,$DC,$31,$F8,$79,$78,$B2,$FE,$A7,$48,$61,$FE,$F3,$8F,$D8
	.dc.b	$E5,$F1,$1F,$D8,$BE,$43,$F0,$35,$E9,$8F,$F6,$74,$FA,$6F,$DD,$1E
	.dc.b	$93,$F0,$07,$A4,$FC,$06,$5E,$97,$FD,$84,$7E,$86,$7F,$13,$8F,$F9
	.dc.b	$EA,$3F,$18,$7E,$9B,$1F,$CE,$63,$FB,$08,$FB,$58,$7B,$AF,$C5,$8F
	.dc.b	$BA,$FD,$B5,$FE,$D3,$3E,$DC,$FE,$58,$FF,$54,$BF,$80,$3E,$23,$F0
	.dc.b	$DD,$9F,$C1,$DD,$DF,$C3,$CB,$F2,$F8,$FC,$FC,$3F,$79,$7F,$A1,$3F
	.dc.b	$86,$3F,$1F,$0F,$C2,$61,$E6,$9F,$E4,$27,$80,$F3,$47,$EB,$0F,$C8
	.dc.b	$65,$F1,$E3,$FC,$5D,$1F,$0B,$2F,$CD,$C7,$CD,$CB,$EC,$3F,$BD,$BF
	.dc.b	$E2,$F9,$79,$B3,$E5,$5F,$E9,$3C,$3F,$09,$7E,$27,$E3,$BF,$E0,$BB
	.dc.b	$FD,$66,$5E,$C9,$CD,$7F,$44,$9F,$D0,$D7,$F8,$FE,$FA,$E3,$FD,$72
	.dc.b	$3C,$47,$F0,$9C,$7C,$4F,$E8,$0F,$C6,$C7,$E2,$2F,$F1,$B1,$CC,$D1
	.dc.b	$C8,$FA,$D8,$78,$A7,$FB,$BE,$1E,$69,$FC,$39,$F4,$4F,$C6,$1F,$CF
	.dc.b	$F3,$FC,$65,$7C,$C8,$FC,$64,$FA,$A3,$F9,$8C,$2F,$E3,$4F,$D7,$9F
	.dc.b	$E6,$53,$FC,$15,$FA,$D1,$F8,$BC,$FF,$C3,$B1,$A9,$FB,$87,$CB,$3F
	.dc.b	$2F,$9F,$CB,$AF,$DC,$C7,$E1,$8F,$80,$7E,$B0,$F8,$B1,$F9,$B8,$FC
	.dc.b	$A4,$F8,$27,$F2,$95,$E0,$9F,$E6,$36,$CB,$F8,$23,$FA,$7A,$FC,$36
	.dc.b	$5F,$A7,$7A,$9F,$81,$AE,$A7,$CB,$C3,$E3,$47,$CB,$7F,$21,$8F,$E6
	.dc.b	$E7,$BD,$7D,$F7,$F9,$2C,$79,$EB,$F3,$71,$CC,$F8,$CF,$F5,$C7,$CF
	.dc.b	$87,$9A,$BF,$37,$97,$EA,$EB,$ED,$E1,$F7,$F0,$FE,$F9,$1F,$47,$0E
	.dc.b	$8F,$C2,$3A,$1F,$DF,$71,$F8,$4F,$CD,$AF,$BB,$1F,$72,$3E,$0C,$7D
	.dc.b	$73,$B3,$FE,$30,$BF,$14,$7C,$23,$27,$F6,$88,$F1,$1F,$70,$F3,$D7
	.dc.b	$DC,$7F,$22,$27,$F2,$31,$F0,$A3,$FA,$B6,$1F,$A1,$C7,$F4,$B1,$F8
	.dc.b	$13,$F4,$18,$7F,$0F,$7B,$9F,$6B,$1F,$C5,$3F,$CA,$B1,$FA,$AF,$E2
	.dc.b	$9F,$E1,$EF,$BB,$0F,$9E,$FC,$37,$F9,$7E,$7F,$E9,$48,$FE,$6A,$7E
	.dc.b	$90,$FE,$AE,$73,$8F,$BF,$5F,$E5,$2C,$FF,$7E,$FD,$49,$FC,$FE,$1F
	.dc.b	$E8,$DA,$FA,$B1,$FE,$C6,$7F,$0A,$FE,$3C,$FC,$29,$F2,$70,$F9,$D1
	.dc.b	$F2,$5F,$59,$F6,$32,$FC,$76,$1E,$0A,$FF,$16,$9F,$13,$3F,$AF,$7F
	.dc.b	$36,$FF,$8E,$9F,$75,$FC,$EC,$F7,$74,$7E,$F4,$F8,$AF,$D8,$9F,$AE
	.dc.b	$FE,$34,$FA,$F8,$FE,$75,$FD,$E1,$F4,$CF,$C0,$9F,$4E,$7F,$86,$1F
	.dc.b	$39,$FE,$DB,$87,$EC,$5F,$BA,$FE,$05,$F5,$65,$ED,$C3,$FC,$6F,$5F
	.dc.b	$21,$F5,$4F,$C8,$3C,$0F,$C8,$7C,$07,$D8,$7E,$45,$FD,$78,$FB,$87
	.dc.b	$F7,$53,$FC,$0F,$5D,$EB,$FC,$5F,$8D,$9F,$E7,$07,$90,$FE,$67,$1E
	.dc.b	$33,$E3,$DF,$D6,$7D,$D1,$FD,$A0,$FC,$04,$F3,$7F,$9C,$C7,$D2,$7D
	.dc.b	$2F,$D2,$7F,$2C,$7E,$E7,$3F,$E0,$98,$7F,$0F,$8F,$F3,$21,$F4,$8F
	.dc.b	$C5,$1F,$4B,$2F,$C0,$BF,$5F,$0F,$C0,$DF,$98,$FA,$71,$FD,$0F,$1F
	.dc.b	$D0,$67,$FD,$A2,$7D,$93,$E1,$3F,$4C,$6C,$FB,$38,$7E,$C1,$FE,$E7
	.dc.b	$1D,$1F,$DF,$57,$4C,$FE,$F4,$F4,$9F,$F7,$8E,$5F,$BE,$C7,$F6,$B1
	.dc.b	$F8,$2C,$3F,$19,$8F,$EA,$63,$F2,$C7,$AE,$FF,$2C,$F9,$4F,$CB,$63
	.dc.b	$DE,$BC,$95,$FD,$D6,$7F,$05,$5F,$52,$FF,$A4,$47,$D4,$3F,$07,$3F
	.dc.b	$DC,$23,$F0,$F1,$FE,$C2,$9F,$44,$FF,$2A,$8E,$E7,$C2,$3F,$3E,$7C
	.dc.b	$23,$B9,$F1,$CF,$C1,$47,$F8,$A3,$1F,$CF,$47,$F0,$67,$FB,$CB,$EE
	.dc.b	$3C,$87,$F4,$F8,$F9,$99,$7E,$CA,$33,$7F,$CC,$A7,$FC,$3B,$1F,$95
	.dc.b	$3F,$94,$3E,$8C,$BF,$13,$1F,$C6,$4F,$C4,$CF,$E1,$1F,$94,$FE,$65
	.dc.b	$F4,$3F,$1A,$7D,$18,$FD,$4C,$3C,$F5,$F9,$83,$EA,$1F,$98,$3E,$E9
	.dc.b	$E3,$AF,$BB,$5E,$38,$FC,$59,$F8,$63,$E5,$47,$EE,$5F,$94,$7E,$D7
	.dc.b	$1F,$EF,$18,$FE,$72,$3E,$0D,$7F,$25,$3E,$2E,$5F,$81,$8F,$A7,$5F
	.dc.b	$81,$79,$61,$FD,$B2,$7F,$65,$3F,$9F,$CE,$CF,$D4,$C7,$E6,$5F,$F9
	.dc.b	$E3,$1F,$06,$3F,$0A,$3F,$5F,$87,$E5,$5F,$84,$79,$0F,$E2,$27,$83
	.dc.b	$2F,$DC,$3F,$E2,$BC,$3F,$9C,$BF,$90,$CF,$F4,$57,$FD,$32,$3F,$21
	.dc.b	$97,$92,$FE,$81,$FE,$F7,$CB,$CB,$3F,$24,$FC,$6C,$FF,$88,$5F,$B7
	.dc.b	$5F,$11,$F9,$AF,$F0,$D7,$F5,$C7,$AB,$0F,$B7,$97,$BA,$B9,$4F,$DA
	.dc.b	$3F,$2D,$87,$86,$FF,$D6,$75,$B9,$FF,$07,$9B,$C3,$CB,$97,$D8,$CF
	.dc.b	$D5,$3F,$53,$26,$3F,$D9,$EF,$71,$F5,$67,$F9,$6E,$5C,$27,$F9,$FC
	.dc.b	$7F,$1D,$BC,$65,$EB,$3F,$19,$1F,$B4,$3E,$19,$E4,$C7,$F7,$15,$F9
	.dc.b	$D8,$F9,$0F,$E7,$6F,$DF,$3A,$AF,$91,$5C,$8F,$E5,$D5,$E2,$9F,$C0
	.dc.b	$65,$F0,$67,$35,$FC,$B5,$FE,$02,$76,$7F,$C0,$27,$E2,$6B,$96,$3F
	.dc.b	$43,$1F,$5D,$7E,$A8,$FA,$B3,$AC,$BE,$84,$7E,$4E,$76,$FE,$3C,$FD
	.dc.b	$3E,$3F,$C6,$70,$FE,$3C,$78,$32,$F6,$E7,$FC,$0B,$1F,$E3,$A7,$F6
	.dc.b	$C8,$E7,$5F,$06,$FD,$39,$FF,$0F,$C3,$D0,$7B,$63,$F3,$87,$F0,$B3
	.dc.b	$F3,$99,$7F,$6D,$7F,$8C,$1C,$B2,$FD,$A6,$3F,$BD,$C3,$F6,$87,$9E
	.dc.b	$7F,$87,$DF,$9D,$FB,$19,$79,$E3,$D7,$1F,$C9,$5F,$1B,$F9,$E3,$B1
	.dc.b	$F9,$E3,$E5,$3F,$84,$AF,$C0,$63,$EB,$CB,$E1,$D7,$F1,$28,$C3,$E0
	.dc.b	$31,$8F,$A2,$FF,$1D,$5F,$1B,$0F,$25,$7A,$0F,$24,$74,$7F,$25,$87
	.dc.b	$E3,$5E,$86,$2B,$DC,$FE,$20,$F5,$1F,$32,$3E,$AC,$7F,$84,$A3,$F8
	.dc.b	$31,$FC,$AD,$FE,$0D,$1F,$66,$32,$7E,$D8,$FF,$3C,$5F,$E1,$A7,$E3
	.dc.b	$C7,$ED,$5F,$F0,$86,$1F,$99,$AF,$E8,$53,$93,$FC,$59,$19,$9F,$ED
	.dc.b	$F1,$93,$F6,$47,$ED,$8F,$0C,$FE,$EB,$2F,$E3,$F9,$7F,$A3,$67,$F3
	.dc.b	$75,$F8,$98,$FC,$04,$7E,$94,$F9,$27,$C6,$AF,$AF,$87,$D1,$8F,$B7
	.dc.b	$8F,$D1,$7E,$A9,$F7,$CF,$D1,$57,$E3,$32,$FE,$57,$1F,$D7,$B0,$F4
	.dc.b	$CF,$E7,$63,$E7,$47,$E3,$1F,$E7,$2F,$C6,$8D,$1E,$6C,$BF,$0B,$3F
	.dc.b	$5C,$F8,$51,$FC,$0B,$0F,$9D,$87,$E0,$71,$F4,$D7,$97,$0E,$98,$FE
	.dc.b	$22,$3E,$01,$E0,$AE,$13,$FC,$42,$BF,$11,$3D,$F0,$F1,$DF,$B2,$3F
	.dc.b	$5E,$7F,$46,$3E,$8D,$7D,$CC,$3C,$71,$F2,$E3,$A1,$F9,$0B,$FB,$F7
	.dc.b	$D2,$7F,$18,$7C,$73,$FB,$D3,$97,$F1,$91,$FD,$43,$1F,$86,$F2,$C3
	.dc.b	$D9,$18,$BF,$9F,$18,$C3,$ED,$5F,$D6,$CF,$FA,$0B,$FE,$98,$3A,$D7
	.dc.b	$F4,$C9,$F4,$C7,$C1,$AE,$4F,$CB,$CB,$E3,$BF,$C0,$0F,$E6,$D8,$F8
	.dc.b	$DF,$C2,$98,$BF,$1B,$88,$F9,$A6,$DF,$DA,$1F,$92,$3C,$06,$E3,$E4
	.dc.b	$1C,$9D,$9F,$64,$F9,$C7,$8E,$3A,$C6,$DF,$C1,$C7,$8C,$F9,$B5,$EC
	.dc.b	$70,$7E,$98,$FE,$07,$97,$E4,$70,$F9,$37,$F8,$A9,$F2,$1E,$7B,$F5
	.dc.b	$1D,$8F,$0B,$D8,$D9,$E7,$3D,$A7,$AB,$2F,$DF,$3B,$31,$5F,$4A,$7A
	.dc.b	$D7,$E2,$A7,$C0,$FB,$4C,$3D,$63,$6F,$C3,$8F,$92,$F5,$8F,$34,$62
	.dc.b	$3E,$01,$FA,$CC,$BF,$B2,$3F,$C3,$4F,$E9,$67,$EE,$0F,$E6,$07,$78
	.dc.b	$F1,$3D,$CF,$29,$DD,$F2,$DF,$BA,$3B,$E1,$F9,$DB,$FE,$31,$1F,$D6
	.dc.b	$E3,$F8,$C5,$7F,$49,$38,$E1,$F0,$DE,$F3,$E3,$3F,$77,$7D,$63,$84
	.dc.b	$78,$CF,$31,$E3,$31,$1B,$CB,$F8,$E4,$7D,$03,$E0,$61,$E4,$3F,$C0
	.dc.b	$4F,$F3,$C3,$35,$FE,$61,$79,$DF,$95,$EC,$79,$6F,$F6,$8E,$30,$FC
	.dc.b	$63,$FA,$B3,$F3,$26,$E7,$F1,$F8,$7E,$BA,$30,$6C,$FC,$4E,$1F,$69
	.dc.b	$F9,$E7,$CC,$CB,$F3,$91,$F9,$28,$C1,$F9,$2C,$3E,$03,$FC,$60,$FE
	.dc.b	$9F,$7E,$97,$67,$C3,$3F,$D3,$77,$F1,$AB,$C1,$1F,$30,$DB,$8A,$EB
	.dc.b	$1F,$88,$C3,$DE,$FE,$BB,$1F,$AA,$FB,$0C,$5F,$58,$FB,$F1,$F3,$8F
	.dc.b	$F3,$DC,$F2,$BF,$64,$F2,$8C,$3F,$9B,$31,$8F,$F3,$78,$FC,$5D,$F5
	.dc.b	$BF,$AE,$75,$3F,$C7,$33,$BC,$BF,$D0,$B1,$E9,$8C,$67,$FE,$2D,$C7
	.dc.b	$F0,$98,$FF,$6B,$9C,$D7,$F4,$33,$81,$FC,$F0,$FE,$8F,$3F,$E0,$77
	.dc.b	$27,$F3,$03,$FC,$98,$78,$F1,$FE,$D4,$F1,$7D,$2F,$D7,$3F,$C0,$27
	.dc.b	$D4,$30,$FF,$63,$79,$61,$FC,$FA,$7F,$D0,$75,$F8,$4B,$F8,$D8,$FC
	.dc.b	$3B,$FA,$55,$F3,$23,$18,$7C,$C9,$FC,$6D,$7C,$C7,$C9,$87,$C6,$BF
	.dc.b	$87,$1F,$CA,$30,$F8,$71,$EF,$33,$1F,$C3,$71,$FF,$35,$BF,$12,$FF
	.dc.b	$C0,$59,$7E,$22,$B7,$8F,$F0,$59,$FE,$D9,$5E,$CB,$FF,$4C,$46,$23
	.dc.b	$E1,$1F,$32,$BB,$DF,$EC,$23,$F2,$B3,$F8,$C7,$73,$DB,$0F,$89,$1E
	.dc.b	$EC,$BE,$0C,$7A,$AB,$FC,$85,$3D,$5F,$05,$7F,$BB,$2B,$73,$F1,$30
	.dc.b	$FC,$35,$7F,$05,$8F,$9A,$FD,$07,$EB,$E5,$FC,$B9,$F1,$C6,$F2,$FA
	.dc.b	$11,$FA,$D8,$FA,$0F,$E1,$CF,$90,$7F,$32,$CB,$E4,$47,$1C,$BD,$12
	.dc.b	$36,$7F,$80,$70,$F9,$2F,$F0,$03,$E6,$1F,$E0,$3C,$BE,$4E,$7F,$C4
	.dc.b	$A3,$F8,$AE,$3E,$4C,$7B,$C7,$ED,$30,$EE,$F6,$C7,$FB,$33,$FC,$27
	.dc.b	$0F,$8C,$FC,$0C,$7F,$03,$87,$FB,$56,$3C,$33,$F9,$67,$F8,$B6,$3F
	.dc.b	$C4,$ED,$9F,$FB,$6A,$7D,$87,$B6,$D9,$FF,$87,$23,$27,$F8,$D8,$B1
	.dc.b	$AC,$73,$1C,$67,$83,$FF,$00,$4D,$B8,$75,$1C,$1E,$86,$C8,$68,$EB
	.dc.b	$2F,$81,$7E,$59,$C6,$30,$D8,$CC,$F0,$6B,$E8,$6B,$F0,$27,$8B,$B9
	.dc.b	$E4,$47,$04,$39,$3F,$13,$EE,$04,$00

;--------------------------------------------------------------------------------
;	FDABS.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fdabs_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FDABS.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FDABS.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fdabs_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdabs.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDABS.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdabs.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDABS.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fdabs_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$E6,$3A,$29,$F7,$3F,$80,$9C,$C6,$1C,$4E,$E7,$51
	.dc.b	$E6,$9D,$3F,$88,$9D,$3C,$23,$2F,$96,$78,$CE,$A3,$84,$E9,$F0,$CE
	.dc.b	$A3,$11,$DE,$30,$7F,$4A,$FA,$EF,$F3,$4F,$A3,$FC,$76,$51,$8F,$57
	.dc.b	$F9,$DB,$FE,$6E,$C3,$CA,$DC,$3D,$CE,$EB,$A5,$FE,$81,$A7,$90,$EA
	.dc.b	$FA,$DA,$7F,$5C,$7C,$51,$CE,$32,$7A,$4B,$5E,$29,$ED,$1A,$78,$9D
	.dc.b	$67,$CD,$87,$8A,$53,$A7,$C1,$87,$BA,$38,$12,$32,$71,$78,$3E,$3A
	.dc.b	$EA,$FF,$5D,$8F,$1F,$C3,$7A,$4C,$91,$C9,$23,$E2,$46,$5E,$E7,$32
	.dc.b	$B9,$7C,$11,$F0,$27,$E3,$1E,$22,$39,$24,$7D,$18,$CB,$D9,$E3,$1E
	.dc.b	$F8,$EA,$65,$E2,$E0,$8F,$C2,$3E,$1E,$1F,$4E,$3B,$47,$03,$38,$FE
	.dc.b	$19,$EC,$71,$C3,$E8,$4F,$02,$1F,$59,$CB,$4F,$B5,$5D,$DF,$1C,$6A
	.dc.b	$7B,$E3,$AB,$F0,$57,$AC,$87,$AE,$F9,$57,$8E,$38,$54,$7F,$19,$7D
	.dc.b	$4F,$1D,$78,$4E,$2E,$63,$86,$1F,$39,$F4,$1C,$8C,$DF,$C2,$9E,$11
	.dc.b	$F9,$CC,$3C,$EF,$B0,$FC,$84,$78,$23,$37,$E0,$9C,$9D,$0F,$93,$5E
	.dc.b	$B7,$A6,$19,$7C,$78,$7E,$01,$F9,$F5,$CD,$F3,$1C,$09,$1B,$78,$D7
	.dc.b	$E1,$E7,$84,$E5,$F8,$97,$EA,$9F,$14,$FC,$39,$DC,$7E,$31,$DD,$FA
	.dc.b	$AF,$C3,$5B,$BE,$58,$7C,$68,$C5,$6D,$F0,$4F,$C4,$37,$59,$C3,$CD
	.dc.b	$18,$8E,$B1,$8A,$D4,$7E,$13,$0F,$1B,$A3,$15,$D6,$77,$1F,$08,$F9
	.dc.b	$31,$D6,$FC,$73,$F3,$EF,$E0,$9D,$EB,$A1,$C4,$FC,$95,$71,$8F,$88
	.dc.b	$F7,$8D,$C7,$E7,$1F,$A6,$FE,$B7,$0F,$C2,$57,$77,$E3,$47,$7B,$FB
	.dc.b	$0F,$17,$F2,$8F,$92,$BD,$35,$EC,$8D,$1F,$95,$8F,$7E,$1F,$12,$B9
	.dc.b	$DF,$82,$7C,$67,$EB,$2B,$8B,$F9,$28,$F1,$BD,$8F,$5B,$C6,$7C,$6E
	.dc.b	$A3,$E1,$DF,$8E,$35,$5E,$39,$F9,$37,$E3,$8F,$85,$1F,$9D,$8F,$C6
	.dc.b	$BF,$62,$BE,$23,$EA,$C3,$E6,$61,$EA,$8E,$11,$F4,$5E,$07,$B0,$FA
	.dc.b	$D7,$EC,$7F,$68,$78,$A3,$9C,$78,$A3,$85,$F8,$9F,$BC,$F9,$DF,$8D
	.dc.b	$87,$B1,$F8,$C7,$97,$0F,$3B,$D5,$B8,$7E,$32,$3C,$97,$F6,$5F,$8E
	.dc.b	$7C,$C3,$D7,$3F,$25,$A6,$4F,$2E,$1E,$F7,$CA,$FA,$9F,$29,$C2,$BE
	.dc.b	$33,$C2,$7A,$34,$E9,$1F,$B4,$6D,$7E,$44,$F8,$91,$93,$99,$5C,$9B
	.dc.b	$8F,$09,$D6,$37,$19,$8F,$A7,$7F,$25,$FA,$73,$F1,$4F,$C5,$BF,$2C
	.dc.b	$F8,$F1,$F8,$DC,$3E,$94,$72,$3D,$D5,$E2,$7F,$2D,$5E,$06,$C6,$5F
	.dc.b	$9F,$19,$6C,$66,$7A,$BF,$55,$B5,$E1,$C3,$E7,$57,$84,$F7,$57,$18
	.dc.b	$CB,$C5,$B3,$C7,$0F,$C1,$CF,$F0,$66,$D7,$E6,$30,$F8,$2F,$E2,$4F
	.dc.b	$82,$7C,$37,$27,$C7,$9F,$D6,$36,$3F,$26,$71,$77,$1D,$E3,$99,$97
	.dc.b	$89,$98,$DC,$78,$E3,$2D,$3A,$C7,$77,$EC,$C7,$13,$6F,$E5,$27,$DA
	.dc.b	$D3,$C3,$3C,$70,$FD,$75,$7E,$D5,$B7,$F8,$56,$D7,$DE,$3F,$1B,$5F
	.dc.b	$91,$6D,$7E,$34,$FA,$17,$E0,$7A,$61,$FA,$76,$C7,$E8,$AB,$95,$7D
	.dc.b	$B6,$CF,$28,$F3,$BE,$D7,$E0,$B5,$F8,$27,$C5,$29,$EF,$8F,$96,$D8
	.dc.b	$F8,$A7,$89,$FD,$79,$F9,$63,$D6,$53,$C8,$FB,$70,$FC,$54,$7F,$7E
	.dc.b	$A7,$C7,$7F,$9A,$9B,$1F,$E5,$26,$BF,$C8,$1B,$5F,$E8,$76,$D2,$64
	.dc.b	$70,$C7,$FC,$18,$EA,$33,$7D,$CF,$E7,$8F,$FB,$A2,$FF,$D2,$4D,$BF
	.dc.b	$F5,$3B,$FE,$2A,$30,$7F,$AA,$A3,$6F,$FA,$08,$C9,$FE,$B3,$72,$FF
	.dc.b	$9B,$0C,$90,$FF,$7C,$7B,$1D,$A0,$D2,$47,$76,$BD,$5A,$FA,$5A,$F4
	.dc.b	$38,$1E,$73,$B1,$0C,$10,$F4,$19,$29,$93,$E3,$19,$29,$FE,$CC,$F8
	.dc.b	$47,$B5,$F2,$1B,$19,$8C,$11,$C1,$0F,$F6,$D4,$70,$6C,$6D,$A6,$CF
	.dc.b	$C0,$1F,$81,$87,$B7,$E1,$24,$79,$5A,$F6,$6B,$C8,$47,$27,$81,$0E
	.dc.b	$05,$3F,$DD,$1F,$09,$E2,$CF,$F2,$1F,$09,$74,$F2,$EA,$8C,$5F,$FB
	.dc.b	$EA,$EA,$94,$61,$E1,$1F,$98,$3F,$E0,$0C,$E6,$64,$64,$E5,$1C,$4C
	.dc.b	$AB,$FD,$24,$E6,$38,$57,$67,$57,$E8,$9E,$35,$EC,$9D,$DE,$2B,$51
	.dc.b	$82,$1C,$8F,$71,$DB,$E1,$3D,$84,$7C,$3F,$09,$5E,$07,$C3,$8C,$5F
	.dc.b	$EE,$E6,$D7,$CC,$9B,$5F,$C0,$71,$56,$23,$91,$E1,$3B,$94,$EF,$09
	.dc.b	$E0,$F2,$8D,$5F,$F0,$B9,$CD,$7D,$5F,$84,$FA,$DF,$09,$2F,$E5,$C6
	.dc.b	$5F,$9B,$19,$7B,$C6,$5F,$0C,$66,$FF,$8C,$4E,$5F,$E3,$51,$9C,$7F
	.dc.b	$17,$39,$7F,$1B,$19,$AF,$85,$39,$7E,$24,$65,$FC,$74,$7E,$37,$E1
	.dc.b	$3F,$1D,$F0,$92,$7F,$07,$19,$7F,$0B,$19,$7E,$3C,$65,$F9,$31,$9C
	.dc.b	$3F,$98,$4E,$5F,$E6,$51,$9C,$3F,$47,$39,$7F,$4B,$19,$AF,$B5,$39
	.dc.b	$7E,$E4,$65,$FD,$34,$7E,$97,$E1,$24,$7E,$69,$FD,$47,$06,$23,$EE
	.dc.b	$C5,$C3,$FA,$24,$DC,$3F,$6B,$36,$BF,$15,$35,$FD,$A4,$5D,$7F,$82
	.dc.b	$47,$43,$C0,$43,$F1,$73,$B9,$E0,$FE,$FB,$81,$CD,$FC,$94,$95,$FC
	.dc.b	$A5,$AF,$F2,$26,$BF,$C5,$63,$8B,$B2,$1B,$29,$BE,$BF,$01,$1F,$E0
	.dc.b	$D1,$C9,$FD,$7A,$8D,$46,$1F,$44,$E3,$2F,$EE,$75,$BA,$E7,$5A,$69
	.dc.b	$FD,$0D,$FE,$84,$47,$04,$3F,$9F,$B9,$69,$93,$D4,$53,$F9,$FC,$3B
	.dc.b	$3F,$A7,$12,$3F,$A9,$35,$FE,$96,$D7,$E5,$35,$E6,$D7,$FA,$53,$FD
	.dc.b	$67,$B6,$47,$F5,$D6,$BF,$D5,$9A,$FD,$56,$BE,$66,$BF,$D5,$5A,$FF
	.dc.b	$4E,$EE,$FE,$CC,$47,$FB,$53,$5F,$EC,$6D,$7F,$04,$D7,$E1,$B5,$FE
	.dc.b	$C4,$FF,$71,$EF,$91,$FD,$D5,$AF,$F6,$E6,$BF,$8D,$6B,$F3,$9A,$FF
	.dc.b	$6D,$6A,$FB,$BF,$F0,$BB,$FD,$45,$FF,$1A,$6F,$FD,$D1,$B1,$FE,$35
	.dc.b	$76,$FF,$3E,$6E,$7F,$E3,$E7,$71,$D7,$AF,$FB,$2C,$6A,$BA,$75,$62
	.dc.b	$65,$78,$63,$66,$1E,$D1,$87,$47,$9D,$D5,$F8,$67,$53,$C0,$D1,$C0
	.dc.b	$FE,$AB,$F0,$EF,$0F,$FB,$CB,$63,$FB,$93,$5F,$77,$C2,$EA,$64,$69
	.dc.b	$F6,$C6,$23,$AC,$5B,$F8,$93,$99,$FF,$0B,$57,$56,$D7,$F9,$5A,$31
	.dc.b	$1F,$E2,$46,$BF,$E6,$56,$A6,$92,$39,$12,$3F,$C4,$87,$17,$0F,$D9
	.dc.b	$EE,$FA,$A7,$9F,$2F,$04,$E2,$7F,$09,$39,$79,$18,$21,$F0,$0F,$88
	.dc.b	$63,$0F,$C2,$47,$18,$F4,$4F,$57,$C4,$70,$39,$18,$24,$FF,$A2,$AF
	.dc.b	$EA,$37,$C9,$FB,$24,$9F,$CD,$7C,$1D,$7E,$FC,$59,$FC,$B4,$DC,$7E
	.dc.b	$74,$DA,$FC,$F4,$D7,$E7,$45,$D5,$39,$C1,$39,$8C,$36,$BF,$16,$FD
	.dc.b	$D2,$3F,$50,$EA,$59,$E6,$D8,$DB,$F6,$A3,$2D,$8C,$BF,$8C,$6B,$EC
	.dc.b	$3F,$79,$F0,$5F,$B2,$3C,$AE,$08,$7C,$63,$CE,$63,$1F,$DE,$47,$08
	.dc.b	$FE,$00,$F3,$76,$71,$32,$71,$C7,$C4,$E6,$79,$B9,$3F,$39,$5F,$36
	.dc.b	$BA,$46,$EB,$8D,$70,$8E,$A6,$4E,$27,$33,$F8,$E4,$FD,$5F,$3F,$F5
	.dc.b	$84,$8F,$9C,$D7,$E6,$35,$FD,$AB,$5F,$D9,$35,$FD,$93,$FB,$7F,$42
	.dc.b	$47,$D9,$6B,$F5,$DA,$FF,$00,$6B,$FB,$D6,$BF,$BD,$6B,$FB,$DF,$47
	.dc.b	$F8,$21,$1F,$C3,$35,$FC,$23,$5F,$E1,$AD,$7F,$85,$35,$FE,$14,$FF
	.dc.b	$0F,$F4,$A4,$7E,$45,$AF,$E3,$DA,$FF,$18,$6B,$FC,$55,$AF,$F1,$56
	.dc.b	$BF,$C5,$7D,$3F,$E3,$84,$7F,$3C,$D7,$F3,$8D,$7F,$92,$B5,$FE,$44
	.dc.b	$D7,$F9,$13,$FC,$9F,$D4,$91,$FA,$96,$BF,$A7,$6B,$FC,$C1,$AF,$F2
	.dc.b	$D6,$BF,$CB,$5A,$FF,$2D,$F5,$7F,$9A,$11,$FD,$B3,$5F,$DA,$35,$FE
	.dc.b	$7A,$D7,$F9,$D3,$5F,$E7,$4F,$F3,$FF,$5A,$47,$F0,$16,$BF,$BF,$6B
	.dc.b	$FD,$21,$AF,$F4,$56,$BF,$D1,$5A,$FF,$45,$F8,$7F,$E9,$84,$7F,$87
	.dc.b	$35,$FE,$18,$D7,$FA,$AB,$5F,$EA,$4D,$7F,$A9,$3F,$C6,$5F,$F0,$57
	.dc.b	$C1,$7F,$B8,$BE,$21,$B5,$3F,$C0,$31,$27,$FD,$BD,$3A,$6C,$E5,$FF
	.dc.b	$06,$4F,$17,$27,$F8,$55,$C3,$FE,$CE,$9E,$4D,$9E,$5B,$B8,$13,$2B
	.dc.b	$73,$98,$C5,$74,$8F,$F0,$D5,$DB,$E3,$4B,$E3,$18,$6C,$7F,$89,$67
	.dc.b	$83,$D0,$F4,$9E,$C2,$D7,$85,$E9,$7C,$1F,$96,$6A,$FB,$46,$79,$17
	.dc.b	$BC,$3C,$37,$C8,$F8,$C7,$13,$C4,$7C,$53,$25,$26,$1E,$38,$E7,$5D
	.dc.b	$A7,$91,$D4,$DE,$1F,$E4,$5B,$F1,$61,$E5,$6C,$F0,$AE,$4F,$03,$C7
	.dc.b	$87,$7A,$FA,$F7,$EE,$9E,$F8,$78,$B0,$E7,$1C,$4F,$8A,$58,$F8,$26
	.dc.b	$DE,$35,$E3,$69,$E3,$9E,$77,$E1,$8F,$86,$D7,$A4,$F9,$1F,$3D,$F9
	.dc.b	$5F,$16,$1E,$CA,$E0,$49,$D3,$F8,$A9,$F0,$9C,$2F,$C3,$3A,$3E,$54
	.dc.b	$6A,$79,$3F,$8B,$31,$1B,$70,$79,$8F,$61,$A3,$CE,$6A,$39,$BF,$65
	.dc.b	$B5,$E9,$9E,$93,$96,$C6,$63,$99,$92,$B9,$70,$4C,$3E,$93,$E5,$3E
	.dc.b	$01,$E5,$37,$1C,$63,$EC,$E1,$F5,$E7,$85,$F9,$B0,$FD,$FC,$7E,$C8
	.dc.b	$FC,$91,$6B,$F8,$0B,$6B,$F8,$A4,$62,$3E,$C1,$93,$FC,$60,$E5,$FD
	.dc.b	$29,$93,$FC,$CE,$E5,$EE,$7C,$6D,$1E,$0F,$F0,$4C,$3F,$75,$18,$9F
	.dc.b	$73,$54,$6A,$7F,$03,$1F,$C5,$27,$53,$98,$E3,$3D,$72,$FE,$23,$5A
	.dc.b	$7E,$BD,$FF,$00,$8C,$57,$3A,$EA,$A3,$2D,$8C,$C6,$0C,$CF,$D2,$9F
	.dc.b	$E4,$F7,$E3,$7D,$37,$E2,$BE,$27,$96,$3C,$07,$A0,$E3,$1B,$70,$72
	.dc.b	$36,$7D,$E7,$4E,$4F,$31,$C2,$7C,$47,$E6,$AB,$E0,$57,$86,$76,$73
	.dc.b	$3A,$CF,$0A,$F8,$F8,$6F,$0F,$81,$7E,$AB,$F3,$9C,$2B,$6E,$0E,$B7
	.dc.b	$CD,$FD,$1D,$72,$74,$FC,$AB,$ED,$72,$72,$FF,$3D,$9E,$11,$EF,$78
	.dc.b	$3A,$21,$98,$E0,$73,$70,$7F,$34,$8E,$75,$E5,$8F,$D9,$C6,$2F,$E6
	.dc.b	$CF,$F0,$28,$C4,$FB,$CF,$5D,$75,$9E,$58,$7C,$E9,$F1,$C6,$30,$D6
	.dc.b	$1D,$09,$3C,$8D,$B8,$7C,$77,$C1,$F0,$18,$36,$62,$FC,$0F,$B8,$97
	.dc.b	$E9,$9F,$F0,$86,$1F,$AD,$9F,$E0,$8D,$7F,$16,$62,$36,$47,$FD,$61
	.dc.b	$1B,$7F,$26,$6C,$CC,$6D,$FF,$74,$C7,$EA,$B2,$7C,$17,$E4,$5B,$3E
	.dc.b	$A3,$F0,$F8,$FF,$82,$23,$31,$F9,$D6,$E1,$F5,$EF,$A5,$F3,$3E,$35
	.dc.b	$7E,$E7,$C5,$FC,$C1,$1F,$CA,$5B,$E0,$71,$1F,$8B,$9B,$87,$F8,$36
	.dc.b	$B5,$3F,$96,$9C,$E3,$FC,$31,$F5,$5F,$29,$F9,$71,$F3,$88,$FC,$86
	.dc.b	$BF,$11,$AF,$03,$44,$34,$5B,$FE,$4F,$E0,$7E,$CC,$F1,$C7,$F2,$39
	.dc.b	$C3,$FA,$C8,$CB,$F0,$CC,$10,$F8,$07,$B4,$C4,$F6,$70,$68,$FD,$84
	.dc.b	$76,$71,$1F,$CE,$A7,$0F,$18,$CD,$FC,$77,$AB,$FC,$D2,$BC,$B9,$2F
	.dc.b	$8C,$E6,$BC,$37,$D5,$E2,$7F,$41,$BE,$35,$C2,$30,$70,$AF,$03,$E8
	.dc.b	$38,$9F,$0C,$F2,$9C,$0D,$18,$35,$5E,$B2,$47,$47,$C9,$1C,$1C,$BF
	.dc.b	$4C,$F1,$1D,$5C,$BD,$23,$DC,$F1,$6C,$74,$37,$6B,$F7,$57,$13,$A5
	.dc.b	$70,$23,$E1,$3A,$1A,$21,$C4,$FE,$BD,$1D,$27,$B4,$7A,$0C,$9F,$21
	.dc.b	$CB,$F1,$5B,$5E,$69,$FB,$67,$32,$CF,$16,$BE,$43,$A9,$CD,$E0,$F8
	.dc.b	$67,$AD,$7B,$23,$75,$F1,$63,$13,$A9,$F6,$BF,$8B,$7C,$07,$A6,$FA
	.dc.b	$BE,$18,$D3,$6F,$E1,$4E,$AF,$E8,$B6,$FB,$BF,$A1,$AF,$E5,$35,$FC
	.dc.b	$B0,$8F,$F2,$66,$BF,$C8,$9A,$F0,$34,$43,$45,$AF,$F4,$F7,$C5,$3A
	.dc.b	$FF,$BC,$36,$7D,$A7,$F8,$05,$C3,$63,$D0,$FF,$AE,$A3,$33,$D2,$73
	.dc.b	$29,$CE,$1F,$EB,$86,$C7,$CC,$BF,$0D,$48,$F0,$DF,$37,$FC,$01,$5C
	.dc.b	$1F,$AF,$7F,$66,$30,$67,$1F,$B1,$3A,$3C,$D7,$D5,$FA,$87,$47,$17
	.dc.b	$FA,$AA,$E7,$18,$BE,$CD,$AE,$67,$F0,$09,$DB,$CC,$C4,$6D,$F8,$4F
	.dc.b	$EB,$A3,$71,$F2,$CD,$9C,$4C,$1F,$9B,$7E,$8B,$FE,$6F,$6B,$F2,$E3
	.dc.b	$FE,$00,$46,$23,$CA,$7F,$9D,$4F,$09,$4A,$65,$E0,$E0,$ED,$1D,$08
	.dc.b	$F1,$8F,$9F,$ED,$7E,$20,$F7,$38,$8F,$AF,$18,$9D,$46,$0F,$75,$FD
	.dc.b	$D7,$F1,$D3,$A7,$E0,$C6,$4F,$CB,$C6,$5C,$5F,$C1,$9E,$87,$23,$AB
	.dc.b	$F8,$F9,$E0,$D8,$C9,$F9,$98,$CB,$4C,$9F,$99,$72,$D3,$36,$8C,$B6
	.dc.b	$33,$18,$3E,$29,$5C,$9E,$8A,$F6,$B8,$3D,$44,$7D,$EF,$94,$C5,$7A
	.dc.b	$DF,$8E,$D7,$F0,$D3,$F8,$42,$3F,$65,$AF,$E0,$5A,$F0,$34,$43,$45
	.dc.b	$9F,$F0,$D7,$C0,$FF,$8B,$CF,$64,$7F,$8A,$DC,$3E,$DA,$F3,$4F,$08
	.dc.b	$FF,$13,$CF,$CC,$D5,$EF,$FF,$8E,$49,$3F,$E7,$FF,$1D,$86,$23,$FC
	.dc.b	$71,$37,$0C,$C0,$56,$1D,$9F,$E3,$69,$E1,$5B,$8D,$B8,$79,$5F,$91
	.dc.b	$E0,$78,$E3,$4A,$34,$F6,$3F,$D5,$15,$E1,$BF,$F2,$05,$F0,$AD,$C7
	.dc.b	$96,$75,$7F,$C9,$7A,$FF,$76,$4C,$3F,$60,$D7,$F7,$0D,$7F,$82,$35
	.dc.b	$E4,$70,$21,$FB,$FE,$BF,$E0,$E4,$7F,$78,$D7,$46,$08,$60,$87,$F1
	.dc.b	$37,$83,$FC,$57,$AF,$F8,$99,$1F,$E1,$0D,$7F,$88,$35,$E0,$68,$86
	.dc.b	$8A,$7F,$21,$EB,$FE,$3E,$47,$F8,$A3,$5F,$E3,$8D,$78,$1A,$21,$A2
	.dc.b	$9F,$CB,$78,$3F,$95,$91,$FE,$40,$D7,$F9,$43,$5E,$06,$88,$68,$A7
	.dc.b	$F3,$9E,$BF,$E6,$E4,$7F,$96,$35,$FE,$68,$D7,$81,$A2,$1A,$29,$FD
	.dc.b	$17,$AF,$FA,$19,$1F,$E7,$0D,$7F,$A0,$35,$E0,$68,$86,$8A,$7F,$51
	.dc.b	$EB,$FE,$9E,$47,$FA,$23,$5F,$E9,$8D,$78,$1A,$21,$A2,$9F,$D7,$7E
	.dc.b	$37,$FA,$D9,$1F,$EA,$0D,$7F,$AC,$35,$E0,$68,$86,$8A,$7F,$69,$D3
	.dc.b	$FB,$39,$1F,$EB,$8D,$7F,$B2,$35,$E0,$68,$86,$8A,$7F,$0B,$1F,$19
	.dc.b	$FC,$11,$C3,$5F,$E2,$0D,$7F,$8C,$35,$E3,$C3,$FD,$DD,$A7,$F7,$98
	.dc.b	$D1,$23,$4D,$8D,$46,$3E,$0A,$47,$16,$BF,$E0,$59,$46,$23,$CE,$D7
	.dc.b	$CE,$D7,$CE,$D7,$DF,$F0,$7E,$22,$3E,$76,$BD,$DA,$FC,$46,$BF,$11
	.dc.b	$AF,$C4,$7D,$BC,$92,$3E,$13,$5F,$53,$5F,$9E,$D7,$E7,$B5,$F9,$ED
	.dc.b	$7E,$BF,$C2,$7C,$A2,$3F,$3D,$AF,$C7,$6B,$F7,$1A,$FD,$C6,$BF,$71
	.dc.b	$FA,$BD,$72,$3E,$D3,$5F,$A4,$D7,$F0,$ED,$7F,$0E,$D7,$F0,$ED,$7F
	.dc.b	$1F,$F0,$BF,$82,$23,$F8,$76,$BF,$7D,$AF,$E5,$1A,$FE,$51,$AF,$E5
	.dc.b	$1F,$C6,$F6,$C8,$FC,$93,$5F,$C5,$35,$FC,$FB,$5F,$CF,$B5,$FC,$FB
	.dc.b	$5F,$D3,$FC,$37,$E6,$88,$FE,$7D,$AF,$E5,$DA,$FE,$B1,$AF,$EB,$1A
	.dc.b	$FE,$B1,$FD,$2F,$C3,$FF,$7F,$E9,$FF,$02,$6A,$8C,$6F,$FB,$46,$DF
	.dc.b	$F8,$2F,$A5,$C5,$88,$FD,$EC,$DB,$E5,$1F,$E1,$1C,$3A,$CE,$EF,$8C
	.dc.b	$E7,$0E,$CD,$3F,$80,$CF,$33,$86,$16,$3F,$80,$47,$A9,$A4,$3F,$74
	.dc.b	$E5,$A6,$4F,$E2,$2E,$5A,$71,$BF,$F0,$97,$1F,$9C,$F0,$42,$3D,$6F
	.dc.b	$F8,$CE,$34,$FA,$4F,$86,$5C,$7E,$1B,$FE,$22,$8D,$BC,$F0,$F8,$2D
	.dc.b	$3F,$90,$F8,$FF,$2C,$8F,$F9,$07,$96,$BB,$3F,$96,$C6,$30,$FF,$15
	.dc.b	$4D,$AF,$E5,$B3,$6B,$ED,$4F,$27,$F9,$9C,$7F,$28,$23,$FC,$B5,$AF
	.dc.b	$F2,$56,$BC,$0D,$10,$D1,$6B,$FC,$79,$CD,$F8,$13,$C5,$1E,$83,$FC
	.dc.b	$B8,$E5,$F9,$13,$AD,$13,$99,$F8,$19,$2B,$F2,$1F,$15,$27,$E8,$3F
	.dc.b	$8F,$E8,$C4,$70,$8B,$8F,$F9,$3E,$6E,$7F,$D4,$A6,$D7,$E7,$E7,$8B
	.dc.b	$FD,$62,$3F,$A6,$91,$FE,$A4,$D7,$FA,$53,$5E,$06,$88,$68,$B5,$FE
	.dc.b	$61,$E9,$FD,$31,$E2,$8F,$ED,$33,$87,$FD,$05,$19,$C5,$39,$9F,$F3
	.dc.b	$7C,$E7,$CC,$FD,$A1,$23,$FD,$15,$D7,$5F,$97,$17,$0F,$F3,$94,$DC
	.dc.b	$BF,$BA,$CD,$AF,$E0,$53,$C5,$FE,$F7,$1F,$DC,$08,$FF,$75,$6B,$FD
	.dc.b	$B5,$AF,$03,$44,$34,$5A,$FF,$42,$76,$7F,$09,$3C,$51,$FE,$0C,$9C
	.dc.b	$3F,$EA,$58,$CE,$49,$CC,$FF,$A4,$A7,$8F,$C6,$BD,$CF,$E3,$4D,$9F
	.dc.b	$DD,$35,$FF,$4C,$F9,$3A,$99,$1A,$7F,$8E,$46,$23,$F6,$93,$C7,$1F
	.dc.b	$F5,$83,$73,$FF,$57,$70,$AF,$85,$FF,$89,$EB,$C0,$DC,$3F,$C4,$11
	.dc.b	$88,$FF,$5B,$35,$FF,$5F,$35,$34,$91,$C8,$91,$FE,$35,$38,$B8,$7F
	.dc.b	$D6,$FF,$0F,$FC,$DC,$F6,$63,$EA,$9C,$4F,$FB,$4E,$73,$8F,$92,$73
	.dc.b	$3C,$8F,$8C,$7C,$CF,$5F,$E7,$91,$FF,$6E,$78,$75,$F9,$31,$70,$FF
	.dc.b	$62,$4D,$BF,$83,$36,$BF,$AA,$4F,$17,$E5,$47,$C8,$23,$EF,$3E,$2B
	.dc.b	$87,$E1,$62,$8C,$47,$23,$84,$65,$B1,$A7,$C6,$7F,$BD,$7C,$CF,$EB
	.dc.b	$C7,$95,$EE,$64,$87,$B8,$F8,$86,$30,$FF,$71,$47,$1A,$7B,$1F,$96
	.dc.b	$25,$7F,$4F,$D5,$F0,$F8,$8F,$C8,$CD,$C3,$FD,$D9,$37,$3F,$C4,$CD
	.dc.b	$AF,$EF,$93,$C5,$FC,$94,$7E,$40,$8F,$C2,$6B,$F2,$1A,$F0,$34,$43
	.dc.b	$45,$AF,$F7,$F7,$C3,$FF,$80,$CF,$14,$7E,$62,$70,$FB,$63,$38,$FC
	.dc.b	$28,$C1,$0F,$79,$FB,$8F,$7B,$F7,$64,$7E,$37,$C3,$D7,$F6,$B1,$70
	.dc.b	$65,$FB,$4A,$C3,$FA,$E9,$B5,$FE,$21,$9E,$4F,$ED,$E3,$F6,$A4,$7F
	.dc.b	$0E,$D7,$F1,$AD,$78,$1A,$21,$A2,$D7,$91,$FF,$18,$7C,$3F,$F8,$A4
	.dc.b	$F1,$BF,$BE,$9C,$3F,$83,$8C,$E0,$9C,$CF,$F8,$1F,$45,$6B,$DB,$FE
	.dc.b	$42,$49,$FC,$67,$9B,$5F,$E3,$71,$71,$F2,$42,$D5,$9C,$FF,$8A,$DD
	.dc.b	$AF,$F2,$44,$F3,$7F,$8F,$47,$F1,$C2,$3F,$AE,$6B,$FB,$46,$BC,$0D
	.dc.b	$10,$D1,$6F,$CB,$6F,$87,$FF,$28,$1E,$39,$FE,$4B,$58,$7F,$55,$19
	.dc.b	$CD,$39,$9F,$91,$39,$F6,$FF,$A3,$92,$3F,$C8,$1F,$78,$F8,$5C,$47
	.dc.b	$F4,$29,$B9,$7C,$8A,$D4,$FF,$4E,$9C,$E9,$FC,$F1,$FF,$36,$4E,$A7
	.dc.b	$91,$FC,$4A,$3F,$A1,$91,$C1,$0F,$E3,$6E,$5A,$64,$FE,$9C,$E5,$A7
	.dc.b	$AB,$DC,$F0,$9E,$47,$FA,$6C,$61,$D1,$E9,$21,$FE,$53,$94,$EE,$3D
	.dc.b	$27,$F6,$47,$FB,$79,$1F,$93,$EE,$D7,$47,$F7,$28,$F8,$71,$9C,$3F
	.dc.b	$B5,$C5,$C1,$52,$D7,$CE,$3F,$B6,$57,$F6,$A2,$3F,$CF,$1A,$FF,$42
	.dc.b	$6B,$C0,$D1,$0D,$16,$FE,$E7,$A9,$F6,$0F,$24,$7F,$74,$9C,$3F,$86
	.dc.b	$8C,$BF,$DD,$A3,$2F,$C9,$8C,$E3,$F9,$98,$FF,$46,$35,$FF,$40,$36
	.dc.b	$BF,$18,$FF,$AC,$CC,$63,$F1,$25,$18,$7E,$3B,$56,$3A,$7B,$E7,$CC
	.dc.b	$8F,$E7,$A6,$BF,$E0,$97,$C8,$F7,$3B,$C7,$F6,$F6,$9F,$DC,$63,$6F
	.dc.b	$95,$AF,$92,$38,$46,$23,$89,$DE,$32,$47,$A3,$FA,$8D,$A4,$7A,$88
	.dc.b	$7E,$55,$CB,$F2,$A3,$D7,$87,$ED,$67,$2F,$F1,$EA,$F8,$C3,$8B,$E7
	.dc.b	$12,$FF,$80,$FC,$45,$7E,$AC,$5A,$FE,$01,$37,$2F,$81,$36,$BF,$31
	.dc.b	$3C,$5F,$95,$7D,$DF,$91,$3F,$10,$8E,$08,$7D,$37,$2D,$32,$74,$73
	.dc.b	$7F,$C4,$F7,$FE,$1E,$78,$E7,$F0,$F3,$87,$F8,$EC,$66,$BC,$91,$FB
	.dc.b	$F8,$D3,$F0,$0F,$CB,$F3,$FE,$4C,$8F,$F2,$6F,$88,$AF,$E6,$A2,$E1
	.dc.b	$FC,$96,$6E,$5F,$85,$9B,$5F,$BD,$9E,$2F,$C6,$8F,$C3,$91,$FC,$73
	.dc.b	$5F,$CB,$35,$E0,$68,$86,$8B,$5F,$CD,$F9,$7F,$99,$9E,$28,$FD,$5C
	.dc.b	$E1,$FE,$83,$19,$C5,$39,$9F,$E4,$93,$9F,$89,$FD,$B1,$23,$FA,$4F
	.dc.b	$C3,$57,$F7,$B1,$70,$FE,$91,$37,$2F,$D4,$4D,$AF,$E4,$13,$C5,$FC
	.dc.b	$4C,$7E,$A8,$8F,$EC,$5A,$FE,$E5,$AF,$03,$44,$34,$5A,$FE,$B5,$F0
	.dc.b	$DF,$D5,$8F,$14,$7F,$0E,$9C,$3F,$D9,$23,$39,$27,$33,$FD,$1E,$73
	.dc.b	$E2,$7F,$19,$24,$7F,$6C,$EB,$AF,$F2,$08,$B8,$7F,$6B,$9B,$97,$F0
	.dc.b	$B9,$B5,$FD,$0E,$78,$BF,$AC,$8F,$E1,$A4,$7F,$8A,$35,$FE,$38,$D7
	.dc.b	$81,$A2,$1A,$2D,$7F,$7D,$F8,$6F,$EF,$27,$8A,$3F,$98,$CE,$1F,$F0
	.dc.b	$3C,$67,$24,$E6,$7F,$B5,$4E,$7C,$AF,$E7,$84,$8F,$F0,$7F,$B7,$5F
	.dc.b	$E8,$71,$70,$FF,$07,$4D,$CB,$F9,$74,$DA,$FE,$CD,$3C,$5F,$E1,$F1
	.dc.b	$FC,$C0,$8F,$F3,$76,$BF,$CF,$DA,$F0,$34,$43,$45,$AF,$F1,$07,$81
	.dc.b	$FE,$1C,$3C,$51,$FD,$52,$70,$FF,$8A,$A3,$39,$27,$33,$FE,$0D,$9C
	.dc.b	$F9,$FF,$D7,$C9,$1F,$E3,$2F,$85,$AF,$F6,$68,$B8,$7F,$8C,$66,$E5
	.dc.b	$FD,$46,$6D,$7F,$82,$67,$8B,$FC,$CA,$3F,$A9,$91,$FE,$B4,$D7,$FB
	.dc.b	$13,$5E,$06,$88,$68,$B5,$FE,$43,$F7,$7F,$C7,$E7,$8A,$3F,$BB,$CE
	.dc.b	$1F,$F2,$6C,$67,$24,$E6,$7F,$C6,$13,$9C,$3F,$B4,$92,$3F,$B8,$35
	.dc.b	$FE,$F0,$D7,$FA,$7B,$5F,$EE,$8D,$7F,$C0,$EF,$F7,$BF,$8F,$7B,$74
	.dc.b	$93,$FE,$ED,$8C,$B6,$32,$FF,$81,$44,$69,$C3,$5F,$F0,$53,$5F,$F0
	.dc.b	$D3,$5F,$F1,$FB,$5E,$71,$E5,$F6,$E4,$78,$DA,$F6,$6B,$D9,$AF,$66
	.dc.b	$BD,$9A,$FA,$3D,$BF,$84,$47,$E0,$35,$F5,$B5,$F5,$B5,$F5,$B5,$F5
	.dc.b	$BF,$37,$DB,$91,$F2,$DA,$FC,$66,$BF,$19,$AF,$C6,$6B,$F1,$9A,$FD
	.dc.b	$0F,$6F,$ED,$11,$FB,$0D,$7E,$9B,$5F,$A6,$D7,$E9,$B5,$FA,$6F,$E1
	.dc.b	$7D,$AF,$F2,$86,$DF,$EC,$6C,$91,$8C,$FE,$DB,$6F,$FC,$F3,$F7,$BF
	.dc.b	$E4,$9B,$E1,$39,$9F,$C3,$B6,$F9,$C7,$FA,$02,$7F,$DA,$75,$C6,$FA
	.dc.b	$CE,$5A,$A7,$AB,$93,$C2,$75,$BF,$53,$F9,$B6,$BF,$96,$6B,$F9,$66
	.dc.b	$BD,$8C,$1C,$48,$71,$3C,$02,$9F,$0D,$EF,$6C,$6B,$0F,$7B,$E9,$C3
	.dc.b	$D7,$87,$C0,$BE,$07,$A6,$78,$61,$87,$BE,$1D,$1E,$47,$EA,$A3,$E6
	.dc.b	$7C,$1C,$8F,$03,$E3,$F4,$F1,$1C,$0F,$DC,$C6,$30,$FA,$33,$71,$FD
	.dc.b	$BC,$DA,$FB,$B3,$C9,$FD,$EC,$7F,$0B,$23,$FC,$25,$AF,$F0,$46,$BC
	.dc.b	$0D,$10,$D1,$6B,$F0,$7F,$1F,$F2,$CF,$14,$7A,$0F,$86,$E5,$F7,$90
	.dc.b	$F7,$62,$8D,$47,$C1,$CD,$7F,$93,$F8,$E9,$3F,$5D,$FB,$5F,$11,$88
	.dc.b	$E1,$17,$1F,$CB,$CD,$CF,$F8,$EC,$DA,$FD,$1C,$F1,$7F,$91,$C7,$F3
	.dc.b	$12,$3F,$CB,$DA,$FF,$2A,$6B,$C0,$D1,$0D,$16,$BF,$53,$EC,$FE,$44
	.dc.b	$F1,$47,$F3,$69,$C3,$F8,$A8,$CE,$29,$CC,$FC,$89,$CF,$C4,$7E,$DC
	.dc.b	$91,$F9,$26,$BF,$2D,$B7,$FB,$A6,$E3,$FD,$01,$B3,$FC,$19,$E2,$FF
	.dc.b	$46,$7F,$AA,$11,$FE,$A4,$D7,$FA,$6B,$5E,$06,$88,$68,$B3,$FC,$33
	.dc.b	$E3,$FF,$68,$78,$A3,$FA,$CC,$E1,$FD,$6C,$67,$14,$E6,$7F,$17,$39
	.dc.b	$D3,$F8,$D1,$23,$F6,$BE,$ED,$7F,$1D,$17,$0F,$E3,$B3,$71,$FE,$C7
	.dc.b	$36,$BF,$94,$CF,$17,$FB,$4C,$7F,$77,$23,$FD,$D5,$AF,$F7,$06,$BC
	.dc.b	$0D,$10,$D1,$6B,$F9,$7E,$7F,$C5,$CF,$14,$7F,$7C,$9C,$3F,$C4,$23
	.dc.b	$39,$7E,$AE,$30,$43,$DE,$7F,$1F,$3F,$9F,$10,$FE,$33,$F1,$15,$FD
	.dc.b	$91,$6F,$FA,$0C,$5C,$7F,$C0,$F3,$4F,$E9,$B1,$C5,$FF,$06,$9F,$E1
	.dc.b	$B2,$1F,$E1,$92,$9F,$E1,$42,$9C,$0D,$10,$D1,$67,$FA,$9F,$C0,$FF
	.dc.b	$3B,$3C,$51,$FE,$20,$30,$7F,$34,$73,$7F,$D6,$63,$FB,$1C,$69,$F7
	.dc.b	$9F,$D2,$3E,$45,$DB,$FD,$95,$B1,$98,$FE,$95,$C3,$96,$46,$4F,$E8
	.dc.b	$71,$87,$F9,$E4,$F2,$6C,$F2,$78,$1F,$E3,$98,$D2,$20,$72,$E1,$AE
	.dc.b	$DA,$EE,$3F,$C8,$8E,$5A,$E4,$CC,$70,$24,$69,$B1,$FD,$7B,$E0,$FF
	.dc.b	$B5,$1E,$DC,$59,$F8,$EF,$35,$B3,$DB,$3A,$7C,$46,$4F,$E8,$75,$CE
	.dc.b	$BD,$9E,$17,$D1,$24,$FF,$6E,$F8,$4A,$FC,$C8,$B7,$FE,$16,$9B,$5C
	.dc.b	$AB,$EE,$57,$F8,$06,$78,$C7,$CC,$3E,$64,$7B,$C8,$FC,$46,$BF,$55
	.dc.b	$AF,$1C,$51,$88,$F0,$1B,$39,$1F,$E0,$6E,$7F,$F0,$91,$E6,$7E,$C4
	.dc.b	$61,$D1,$E8,$21,$FD,$A2,$34,$FC,$13,$F3,$1E,$47,$E5,$C8,$FF,$85
	.dc.b	$3E,$16,$BF,$92,$8B,$97,$F8,$F2,$6E,$5F,$8B,$9B,$5F,$E4,$89,$E2
	.dc.b	$FB,$E3,$F0,$84,$7F,$12,$D7,$F3,$4D,$78,$1A,$21,$A2,$D7,$F9,$3B
	.dc.b	$B3,$FC,$6C,$78,$A3,$F4,$13,$87,$FC,$5F,$19,$C9,$39,$9F,$F0,$6C
	.dc.b	$E7,$CC,$FD,$D1,$23,$FC,$6F,$DF,$5F,$DA,$C5,$C3,$FC,$BF,$37,$2F
	.dc.b	$D7,$4D,$AF,$F3,$5C,$F1,$7F,$05,$1F,$A7,$23,$FA,$C6,$BF,$BD,$6B
	.dc.b	$C0,$D1,$0D,$16,$BF,$CE,$3E,$07,$F9,$64,$F1,$47,$F0,$39,$C3,$FE
	.dc.b	$52,$8C,$E4,$9C,$CF,$F8,$C2,$73,$E8,$7F,$1D,$24,$7F,$9A,$3E,$FB
	.dc.b	$F6,$CF,$F1,$88,$CB,$4D,$65,$8C,$77,$8A,$B1,$3F,$C6,$2F,$9D,$6A
	.dc.b	$7F,$86,$12,$FF,$88,$B5,$FE,$44,$D7,$F5,$CD,$7F,$90,$35,$FE,$40
	.dc.b	$F1,$8F,$4C,$BE,$1F,$FD,$3C,$DA,$FE,$7D,$5E,$67,$F9,$74,$65,$FF
	.dc.b	$39,$36,$BD,$11,$C0,$F8,$D5,$D0,$E4,$FF,$35,$F5,$FF,$A4,$11,$F9
	.dc.b	$9E,$35,$7F,$A1,$C5,$AF,$3C,$DC,$7F,$9E,$4D,$AF,$2C,$F1,$7F,$88
	.dc.b	$47,$F3,$52,$3F,$CE,$9A,$FF,$4B,$6B,$C0,$D1,$0D,$16,$BE,$4F,$C3
	.dc.b	$FC,$83,$C5,$1E,$D3,$F0,$CE,$5F,$ED,$13,$AC,$53,$99,$F7,$9F,$DB
	.dc.b	$7E,$4B,$FB,$61,$1F,$C8,$7C,$45,$7F,$B3,$C5,$C3,$ED,$CD,$CB,$FB
	.dc.b	$04,$DA,$FB,$33,$C5,$FE,$69,$1F,$D6,$48,$FF,$5E,$6B,$FD,$BD,$AF
	.dc.b	$03,$44,$34,$5A,$FC,$6F,$99,$F8,$C3,$C5,$1F,$DD,$67,$0F,$EA,$E3
	.dc.b	$3B,$27,$33,$F8,$E9,$E3,$F2,$6F,$7F,$FD,$E8,$D9,$FF,$6D,$B5,$FC
	.dc.b	$AF,$C4,$EA,$64,$69,$FC,$FC,$62,$3F,$D9,$93,$C7,$1F,$DE,$B7,$3F
	.dc.b	$DB,$56,$EB,$FB,$24,$F7,$6D,$FF,$82,$23,$11,$FE,$14,$6B,$FE,$17
	.dc.b	$6A,$69,$23,$91,$23,$FC,$2B,$19,$7F,$89,$7C,$4F,$EE,$0F,$56,$1E
	.dc.b	$8A,$C4,$E8,$F8,$04,$3F,$73,$3A,$7E,$0C,$7C,$CF,$83,$91,$F3,$1F
	.dc.b	$E0,$BE,$9E,$23,$E3,$C5,$BF,$E1,$73,$72,$F8,$13,$6B,$F8,$4C,$F1
	.dc.b	$7E,$4C,$7C,$72,$3E,$E6,$BF,$E1,$13,$E3,$38,$78,$9B,$21,$BC,$51
	.dc.b	$CA,$FF,$8F,$FC,$47,$F1,$E3,$C9,$5D,$0F,$E7,$8E,$5F,$80,$43,$F8
	.dc.b	$FB,$97,$E0,$9F,$90,$F8,$7F,$C9,$91,$FE,$8D,$F1,$15,$FC,$6C,$5A
	.dc.b	$FE,$63,$37,$2F,$C3,$4D,$AF,$E5,$F3,$C5,$FC,$7C,$7E,$34,$8F,$E1
	.dc.b	$1A,$FC,$16,$BC,$0D,$10,$D1,$6B,$FA,$1F,$B3,$FD,$08,$F1,$47,$E5
	.dc.b	$67,$0F,$F6,$18,$CE,$49,$CC,$FF,$44,$9C,$E3,$FB,$62,$47,$F6,$99
	.dc.b	$AF,$EC,$22,$DF,$F5,$49,$B8,$FE,$A6,$6D,$7F,$52,$9E,$2F,$EC,$A3
	.dc.b	$F6,$04,$7F,$4E,$D7,$F0,$8D,$78,$1A,$21,$A2,$D7,$F6,$6F,$87,$FE
	.dc.b	$CA,$78,$A3,$F7,$13,$87,$FC,$09,$19,$C5,$39,$9F,$EC,$F3,$9D,$FF
	.dc.b	$8C,$92,$3F,$C1,$9E,$E5,$7F,$89,$C5,$C3,$FB,$BC,$DC,$BF,$86,$4D
	.dc.b	$AF,$EE,$B3,$C5,$FE,$2B,$1F,$C4,$C8,$FF,$0A,$6B,$FA,$76,$BC,$0D
	.dc.b	$10,$D1,$6B,$FC,$13,$F0,$FF,$E0,$83,$C5,$1F,$C6,$E7,$0F,$F8,$9E
	.dc.b	$33,$92,$73,$3F,$E0,$A9,$CF,$2F,$F2,$92,$47,$F2,$06,$BF,$BD,$6B
	.dc.b	$FC,$C9,$AF,$F2,$F6,$BF,$CB,$DF,$E6,$9C,$D2,$3F,$95,$B5,$FE,$10
	.dc.b	$D7,$F9,$EB,$5F,$E7,$4D,$7F,$9D,$35,$FE,$75,$D3,$FD,$04,$8F,$F3
	.dc.b	$86,$BF,$C5,$5A,$FF,$4A,$6B,$FD,$1D,$AF,$F4,$77,$FA,$67,$54,$8F
	.dc.b	$E8,$6D,$7F,$90,$35,$FE,$AA,$D7,$FA,$93,$5F,$EA,$4D,$5D,$2F,$87
	.dc.b	$FF,$21,$92,$3F,$CC,$1F,$05,$5F,$F2,$0C,$5D,$7F,$C9,$73,$77,$FF
	.dc.b	$2E,$4D,$AF,$F2,$4C,$D7,$BC,$5B,$4D,$8C,$36,$7F,$CE,$11,$FE,$56
	.dc.b	$8F,$F7,$FD,$FF,$9C,$BD,$27,$83,$FE,$F8,$6D,$FF,$BF,$1A,$FF,$4B
	.dc.b	$6B,$FE,$DD,$6A,$23,$86,$C7,$52,$3D,$5A,$F5,$7C,$5E,$C4,$8F,$13
	.dc.b	$5F,$13,$5F,$13,$5F,$10,$94,$66,$3C,$AD,$75,$EC,$7C,$4C,$F8,$D2
	.dc.b	$31,$3E,$C6,$BF,$B3,$F8,$6B,$3B,$8F,$85,$8F,$CD,$AC,$9C,$38,$97
	.dc.b	$8A,$DE,$1E,$0A,$C4,$2F,$FD,$4F,$7C,$23,$67,$8C,$FA,$D3,$F6,$48
	.dc.b	$E0,$87,$D9,$72,$D3,$27,$DB,$72,$D3,$B9,$FC,$47,$C1,$F5,$1E,$6A
	.dc.b	$F7,$46,$A9,$3A,$8E,$38,$FC,$57,$D9,$87,$BD,$E1,$3B,$72,$7B,$4A
	.dc.b	$7D,$BF,$2B,$E7,$91,$F0,$7B,$75,$DD,$A7,$18,$FC,$AA,$B8,$7D,$E9
	.dc.b	$B7,$F9,$9C,$55,$CE,$3F,$3D,$3F,$A5,$23,$FA,$56,$BF,$A5,$6B,$C0
	.dc.b	$D1,$0D,$16,$FE,$EF,$C1,$FE,$14,$F1,$C7,$BC,$F9,$8E,$5F,$97,$3A
	.dc.b	$B4,$E6,$7E,$0C,$FE,$53,$D0,$91,$F4,$9F,$C9,$79,$18,$8F,$8F,$17
	.dc.b	$0F,$C7,$CD,$CB,$F4,$73,$6B,$E3,$4F,$17,$F8,$14,$7F,$09,$23,$FC
	.dc.b	$25,$AF,$F0,$96,$BC,$0D,$10,$D1,$6B,$F4,$3F,$03,$FA,$93,$C5,$1F
	.dc.b	$BD,$9C,$3F,$91,$8C,$E8,$9C,$CF,$E2,$B3,$56,$BD,$6F,$DC,$92,$7F
	.dc.b	$35,$F0,$35,$FC,$5C,$5C,$7F,$67,$37,$3F,$E0,$F3,$6B,$F1,$53,$C5
	.dc.b	$FE,$4D,$1F,$CB,$48,$FF,$2D,$6B,$FC,$B5,$AF,$03,$44,$34,$5A,$FE
	.dc.b	$0B,$E8,$7F,$0D,$3C,$51,$FC,$86,$70,$FE,$D6,$33,$9A,$73,$3F,$16
	.dc.b	$73,$3F,$C7,$09,$1F,$BD,$F7,$EB,$FA,$E8,$B7,$FC,$5A,$6E,$3F,$CA
	.dc.b	$A6,$D7,$EB,$67,$8B,$FD,$26,$3F,$A8,$11,$FE,$A0,$D7,$FA,$83,$5E
	.dc.b	$06,$88,$68,$B5,$FC,$A3,$DF,$FE,$60,$78,$A3,$FA,$24,$E1,$FE,$31
	.dc.b	$19,$8F,$E9,$11,$82,$1E,$F3,$F8,$DF,$BD,$FD,$00,$8F,$F2,$6F,$BF
	.dc.b	$FE,$A2,$FF,$12,$8C,$B4,$D1,$8B,$FE,$E5,$17,$1F,$EB,$35,$CA,$73
	.dc.b	$5F,$DC,$EB,$FB,$B1,$1C,$10,$FE,$EC,$E5,$A6,$4F,$EF,$0E,$5A,$7A
	.dc.b	$5F,$EB,$B2,$7C,$91,$FD,$F1,$C9,$1F,$86,$43,$F9,$BC,$71,$7A,$19
	.dc.b	$95,$7F,$76,$F9,$57,$C5,$7F,$6C,$6D,$7F,$50,$6B,$D3,$E0,$35,$32
	.dc.b	$34,$FA,$23,$11,$FD,$22,$78,$DF,$82,$3F,$C2,$D1,$8B,$F9,$F5,$C2
	.dc.b	$7E,$CC,$F8,$1B,$7F,$EE,$B7,$8B,$54,$6D,$F6,$46,$63,$11,$C0,$E4
	.dc.b	$E1,$E3,$1B,$23,$A7,$E9,$FC,$04,$8F,$79,$0F,$9E,$E5,$F8,$11,$F0
	.dc.b	$B0,$F8,$10,$C5,$5B,$9C,$DF,$C3,$AF,$8B,$AF,$D3,$23,$F9,$DD,$EB
	.dc.b	$F3,$62,$D7,$6C,$3F,$86,$4F,$D9,$9C,$E7,$F7,$DF,$C6,$DF,$29,$F9
	.dc.b	$F1,$F1,$C8,$FC,$66,$BF,$AC,$3E,$7B,$87,$89,$B2,$1B,$C1,$1C,$B0
	.dc.b	$F3,$7B,$BF,$9E,$3C,$B5,$F0,$0F,$DD,$39,$7D,$44,$3F,$3A,$E5,$F8
	.dc.b	$47,$E5,$FE,$13,$F3,$44,$7F,$82,$7B,$B5,$FC,$A4,$5A,$F9,$13,$72
	.dc.b	$FD,$24,$DA,$FF,$36,$66,$AE,$51,$F9,$69,$FC,$71,$1F,$C6,$B5,$F8
	.dc.b	$4D,$78,$1A,$21,$A2,$DF,$DB,$E8,$FE,$06,$78,$E3,$F3,$F3,$87,$F9
	.dc.b	$04,$67,$34,$E6,$7F,$81,$CE,$79,$7F,$7C,$48,$FE,$51,$ED,$57,$F7
	.dc.b	$11,$70,$FC,$64,$DC,$FF,$84,$4D,$AF,$8B,$3C,$5F,$DC,$C7,$EC,$48
	.dc.b	$FE,$C1,$AF,$E1,$5A,$F0,$34,$43,$45,$AF,$CF,$7B,$3F,$C9,$8F,$14
	.dc.b	$7F,$02,$9C,$3F,$D0,$E3,$39,$7F,$84,$23,$04,$3D,$E7,$F1,$FF,$85
	.dc.b	$FE,$44,$47,$FA,$67,$85,$5F,$E3,$91,$70,$FD,$84,$DC,$BF,$96,$4D
	.dc.b	$AF,$C4,$CF,$17,$F8,$EC,$7F,$15,$23,$FC,$51,$AF,$EA,$5A,$F0,$34
	.dc.b	$43,$45,$AF,$E1,$FF,$80,$7F,$10,$6D,$F8,$E7,$F9,$3D,$7F,$AA,$AB
	.dc.b	$53,$FE,$B8,$6C,$FF,$36,$9E,$11,$88,$37,$04,$FF,$49,$EF,$95,$FD
	.dc.b	$25,$ED,$E4,$62,$3F,$A2,$45,$C7,$CF,$37,$1F,$EA,$93,$6B,$FD,$B9
	.dc.b	$8A,$B9,$47,$F4,$69,$FE,$7C,$47,$F9,$EB,$5F,$E2,$4D,$78,$1A,$21
	.dc.b	$A2,$DF,$CA,$F2,$BE,$31,$E3,$8F,$E9,$F3,$87,$FD,$6B,$69,$D6,$5F
	.dc.b	$E9,$A9,$CC,$6D,$F8,$27,$F6,$DF,$8A,$FE,$E0,$47,$9F,$7D,$7F,$B5
	.dc.b	$45,$C7,$EE,$CD,$CF,$FB,$D4,$DA,$F8,$F3,$C5,$FE,$D7,$1F,$D8,$C8
	.dc.b	$FF,$62,$6B,$FC,$DD,$AF,$03,$44,$34,$5A,$FC,$7F,$AB,$F8,$B3,$C5
	.dc.b	$1F,$DD,$27,$0F,$C6,$8C,$E2,$9C,$CF,$FB,$3B,$25,$72,$F9,$77,$93
	.dc.b	$FF,$00,$1B,$5F,$EE,$46,$BC,$7D,$5D,$4C,$8D,$3F,$9F,$8C,$47,$FB
	.dc.b	$72,$78,$E5,$FD,$D9,$BA,$7E,$8E,$B7,$5F,$E4,$C9,$EE,$DB,$FF,$09
	.dc.b	$46,$20,$47,$0D,$8D,$91,$FF,$0E,$46,$93,$49,$1B,$8F,$8D,$EA,$7E
	.dc.b	$E4,$F5,$E1,$E9,$9C,$4F,$E5,$27,$38,$79,$23,$AC,$E9,$E6,$7C,$63
	.dc.b	$E6,$7C,$4F,$CF,$23,$F9,$EF,$4E,$BF,$26,$2E,$1F,$C4,$26,$DF,$7C
	.dc.b	$55,$8B,$FD,$DD,$72,$7E,$64,$7C,$A2,$3F,$11,$AF,$F8,$46,$79,$B9
	.dc.b	$21,$98,$DB,$D0,$CD,$FF,$24,$F4,$BF,$90,$1E,$38,$F9,$F3,$87,$F7
	.dc.b	$71,$9A,$EF,$1F,$95,$8D,$3F,$04,$FC,$95,$FE,$54,$8F,$F0,$4F,$62
	.dc.b	$BF,$8F,$8B,$87,$F3,$69,$B8,$FC,$79,$B5,$FC,$82,$78,$BF,$91,$8F
	.dc.b	$C7,$91,$FC,$4B,$5F,$8C,$D7,$81,$A2,$1A,$2D,$7F,$48,$CB,$FA,$21
	.dc.b	$E2,$8F,$C9,$CE,$1F,$E3,$F1,$9C,$53,$99,$FD,$D4,$E7,$D7,$FD,$C1
	.dc.b	$23,$F4,$B7,$FE,$64,$FC,$05,$E8,$66,$7F,$69,$56,$B5,$37,$3F,$C6
	.dc.b	$4D,$AF,$D7,$DF,$27,$F8,$24,$7E,$D0,$8F,$EB,$9A,$FE,$2D,$AF,$EF
	.dc.b	$5A,$F6,$38,$10,$E0,$7A,$A7,$F5,$92,$F3,$FF,$5D,$87,$B5,$F3,$57
	.dc.b	$EF,$2F,$37,$C2,$6C,$FA,$9D,$E6,$AC,$CB,$0E,$CF,$84,$E0,$7C,$03
	.dc.b	$F9,$27,$0F,$F2,$72,$3B,$39,$73,$E1,$FE,$45,$17,$2F,$A9,$37,$2F
	.dc.b	$DD,$4D,$AF,$9F,$3C,$5F,$E4,$91,$FC,$88,$8F,$F1,$C6,$BF,$B9,$6B
	.dc.b	$C0,$D1,$0D,$16,$BF,$0D,$E7,$FC,$2C,$0F,$8E,$7F,$93,$D6,$62,$47
	.dc.b	$C2,$20,$AE,$6F,$C3,$AE,$6E,$A3,$FA,$17,$3F,$F3,$C2,$47,$F3,$66
	.dc.b	$BF,$C4,$DA,$FF,$45,$6B,$FC,$85,$AF,$F4,$37,$FA,$3F,$B7,$23,$FA
	.dc.b	$0B,$5F,$E3,$CD,$7F,$A8,$35,$FE,$58,$D7,$FA,$73,$5F,$E9,$3D,$3F
	.dc.b	$D5,$08,$FF,$4E,$6B,$FC,$AD,$AF,$F5,$D6,$BF,$CE,$5A,$FF,$5B,$7F
	.dc.b	$AF,$FA,$D2,$3F,$AC,$B5,$FE,$6C,$D7,$FB,$43,$5F,$E8,$8D,$7F,$B3
	.dc.b	$35,$FE,$C3,$D7,$FD,$B0,$8F,$F6,$66,$BF,$D0,$DA,$FF,$75,$6B,$FD
	.dc.b	$45,$AF,$F7,$37,$FB,$BF,$AD,$23,$FB,$8B,$5F,$E9,$CD,$7F,$C0,$0D
	.dc.b	$7F,$AE,$35,$FE,$FC,$D5,$F3,$2F,$80,$FD,$CB,$63,$F6,$4D,$7F,$29
	.dc.b	$C1,$FC,$2B,$D4,$DD,$49,$DB,$FA,$18,$C4,$70,$3F,$C1,$F5,$CF,$55
	.dc.b	$FF,$7F,$8C,$3D,$67,$99,$FE,$FB,$3C,$F7,$5A,$66,$7C,$CF,$F8,$24
	.dc.b	$C9,$4A,$65,$DB,$E1,$8E,$91,$88,$F3,$C6,$08,$F7,$7F,$7D,$E1,$C8
	.dc.b	$F3,$9E,$F3,$C3,$7F,$D2,$63,$51,$9A,$F7,$BD,$8F,$DF,$46,$A7,$E2
	.dc.b	$C6,$4E,$1E,$1F,$D6,$23,$FC,$17,$D3,$AE,$8F,$B3,$1F,$C4,$E3,$33
	.dc.b	$D2,$7F,$03,$38,$C7,$FA,$2D,$72,$9F,$8D,$3F,$10,$8F,$CF,$6B,$F3
	.dc.b	$70,$46,$23,$89,$B2,$1B,$78,$DF,$F6,$DF,$4F,$F9,$41,$E3,$9F,$C3
	.dc.b	$CE,$1F,$E4,$51,$97,$D2,$60,$87,$C1,$3E,$77,$C1,$7E,$70,$8F,$F2
	.dc.b	$BF,$36,$BA,$3F,$3F,$1F,$CE,$63,$37,$F1,$F0,$4E,$32,$FE,$D7,$5C
	.dc.b	$AF,$F1,$93,$F8,$A2,$3F,$95,$6B,$F1,$DA,$F0,$34,$43,$45,$BF,$F0
	.dc.b	$97,$A9,$FD,$34,$F1,$47,$EB,$27,$0F,$F4,$78,$CB,$FB,$E8,$CB,$F9
	.dc.b	$5E,$54,$EB,$17,$99,$FC,$08,$95,$FD,$CF,$E2,$3F,$59,$8F,$C3,$9C
	.dc.b	$CF,$E6,$DF,$F0,$0C,$EB,$3F,$DE,$46,$AB,$85,$73,$38,$9F,$B3,$8F
	.dc.b	$D8,$91,$FD,$E3,$5F,$C7,$35,$E0,$68,$86,$8B,$1F,$9F,$A7,$A9,$E6
	.dc.b	$3C,$97,$E6,$AF,$74,$65,$FE,$43,$7A,$7F,$D4,$F5,$C9,$D9,$A3,$F9
	.dc.b	$3F,$C3,$FF,$2A,$23,$F2,$7D,$5A,$FF,$01,$8B,$8F,$96,$6E,$3F,$C8
	.dc.b	$E6,$D7,$82,$78,$BF,$C6,$63,$F8,$B9,$1F,$E4,$4D,$7F,$68,$D7,$81
	.dc.b	$A2,$1A,$2D,$7C,$7F,$53,$ED,$9E,$28,$FE,$8F,$38,$7F,$09,$19,$CD
	.dc.b	$39,$9F,$E4,$58,$AB,$47,$F4,$E2,$47,$E3,$7D,$4A,$FF,$25,$8B,$87
	.dc.b	$D9,$9B,$97,$F4,$79,$B5,$F5,$27,$8B,$FC,$F6,$3F,$9D,$91,$FE,$8C
	.dc.b	$D7,$F8,$C3,$5E,$06,$88,$68,$B5,$F8,$BF,$53,$F3,$E7,$8A,$3F,$B5
	.dc.b	$CE,$1F,$D4,$46,$72,$F8,$51,$82,$1E,$F3,$FB,$7F,$C4,$7F,$72,$23
	.dc.b	$FB,$1D,$6B,$FD,$26,$2E,$1F,$9D,$9B,$97,$F6,$B9,$B5,$F9,$89,$E2
	.dc.b	$FF,$61,$8F,$EB,$E4,$7F,$B5,$35,$FE,$78,$D7,$81,$A2,$1A,$2D,$7E
	.dc.b	$BF,$D3,$FE,$06,$78,$A3,$FC,$1F,$38,$7F,$86,$46,$70,$4E,$67,$F1
	.dc.b	$33,$97,$FC,$2A,$43,$F8,$A7,$A7,$5F,$ED,$85,$BF,$E0,$11,$71,$FF
	.dc.b	$07,$4D,$3F,$77,$1C,$5F,$F0,$21,$FE,$02,$21,$FE,$0D,$29,$FD,$7C
	.dc.b	$A7,$03,$44,$34,$59,$FE,$25,$CD,$FC,$98,$F1,$47,$F8,$C8,$C1,$FC
	.dc.b	$BD,$CE,$3F,$CA,$E3,$F8,$8C,$69,$F7,$9F,$EA,$7E,$E7,$C3,$FF,$AC
	.dc.b	$9B,$19,$8F,$E6,$BE,$A6,$59,$19,$3F,$9F,$46,$1F,$F0,$D4,$F2,$6C
	.dc.b	$F2,$78,$1F,$E2,$D8,$D3,$FE,$31,$6A,$23,$86,$C7,$F8,$EE,$31,$06
	.dc.b	$92,$34,$48,$FF,$22,$46,$5F,$F6,$47,$E0,$5F,$E1,$DC,$7F,$C3,$CD
	.dc.b	$CF,$C5,$3D,$6F,$DD,$5A,$9F,$F6,$CD,$6E,$BC,$A7,$BA,$74,$F6,$9E
	.dc.b	$27,$4B,$75,$7D,$12,$4F,$93,$C0,$AF,$CC,$8B,$7E,$69,$B7,$CE,$6D
	.dc.b	$7C,$9C,$15,$CA,$3E,$74,$FC,$92,$3F,$11,$AF,$D5,$6B,$C0,$D1,$0D
	.dc.b	$16,$FE,$4E,$1F,$10,$F1,$C7,$D5,$9C,$3F,$42,$32,$7C,$27,$04,$3E
	.dc.b	$01,$F9,$7F,$87,$FC,$B1,$1F,$B3,$E2,$57,$F2,$31,$6F,$ED,$CD,$CB
	.dc.b	$E8,$4D,$AF,$8D,$3C,$5F,$C9,$47,$E3,$48,$FE,$21,$AF,$E6,$5A,$F0
	.dc.b	$34,$43,$45,$AF,$C6,$F8,$DF,$88,$3C,$51,$F9,$99,$C3,$F9,$48,$CE
	.dc.b	$09,$CC,$FF,$A1,$F3,$56,$BC,$2F,$DD,$12,$7F,$3B,$E9,$57,$F6,$B1
	.dc.b	$71,$FC,$FC,$DC,$FF,$29,$36,$BF,$15,$3C,$5F,$DB,$47,$EC,$48,$FE
	.dc.b	$B1,$AF,$EF,$5A,$F0,$34,$43,$45,$AF,$D8,$FA,$5F,$AC,$3C,$51,$FB
	.dc.b	$D9,$C3,$FB,$88,$CE,$69,$CC,$FC,$59,$CF,$C3,$FF,$1D,$24,$7F,$00
	.dc.b	$F4,$2B,$FC,$62,$2E,$1F,$C0,$A6,$E5,$FB,$79,$B5,$FA,$D9,$E2,$FF
	.dc.b	$19,$8F,$E2,$84,$7F,$87,$B5,$FE,$40,$D7,$81,$A2,$1A,$2D,$7F,$14
	.dc.b	$F4,$BF,$87,$9E,$28,$FE,$41,$38,$7F,$8D,$C6,$72,$4E,$67,$F1,$13
	.dc.b	$9F,$3F,$FA,$01,$23,$F9,$1F,$A5,$5F,$E7,$71,$70,$FE,$4B,$37,$2F
	.dc.b	$E3,$53,$6B,$F8,$84,$F1,$7F,$9E,$47,$F3,$72,$3F,$CC,$9A,$FF,$43
	.dc.b	$6B,$C0,$D1,$0D,$16,$BF,$9B,$FA,$5F,$CC,$8F,$14,$7F,$43,$9C,$3F
	.dc.b	$CF,$A3,$39,$27,$33,$FA,$B9,$CF,$C3,$7F,$63,$24,$7F,$53,$BF,$F7
	.dc.b	$17,$E0,$6F,$53,$33,$FD,$82,$AD,$6A,$6E,$7F,$CF,$E6,$D7,$F5,$DB
	.dc.b	$E4,$FF,$73,$8F,$EB,$84,$7F,$AB,$B5,$FE,$D0,$D7,$FB,$43,$5E,$C7
	.dc.b	$02,$1C,$0F,$54,$FF,$5C,$9F,$85,$18,$3C,$98,$F9,$9F,$ED,$95,$FD
	.dc.b	$8B,$0F,$83,$36,$F9,$3B,$C9,$59,$BE,$67,$51,$87,$81,$F9,$06,$7E
	.dc.b	$71,$A7,$18,$46,$1D,$46,$65,$E9,$EE,$B5,$4C,$D9,$FF,$7A,$9C,$C6
	.dc.b	$27,$FC,$1D,$18,$9F,$89,$5C,$F0,$95,$E0,$3D,$58,$78,$1C,$1F,$E0
	.dc.b	$63,$9C,$57,$C0,$D7,$25,$BF,$3B,$C5,$AF,$03,$51,$C5,$F8,$6D,$8D
	.dc.b	$C7,$43,$C0,$E1,$D9,$E8,$F4,$FE,$41,$EC,$7E,$36,$1D,$DA,$7E,$56
	.dc.b	$36,$FB,$B1,$F9,$0F,$43,$67,$DC,$8D,$B8,$AF,$98,$FC,$EF,$87,$FA
	.dc.b	$C4,$7E,$5F,$25,$78,$45,$BF,$D7,$CD,$C3,$ED,$4D,$AF,$D5,$4F,$17
	.dc.b	$E5,$47,$C5,$23,$F3,$DA,$FC,$66,$BC,$0D,$10,$D1,$6B,$F8,$0F,$A1
	.dc.b	$F9,$13,$C5,$1F,$96,$9C,$3F,$86,$8C,$E0,$9C,$CF,$C8,$CD,$5A,$E5
	.dc.b	$FC,$D1,$27,$F1,$FE,$85,$7E,$64,$5C,$7F,$89,$CD,$CF,$F3,$D3,$6B
	.dc.b	$F8,$74,$F1,$7F,$1D,$1F,$89,$23,$F9,$36,$BF,$8A,$6B,$C0,$D1,$0D
	.dc.b	$16,$BF,$92,$F6,$7E,$D4,$F1,$47,$EE,$A7,$0F,$EA,$A3,$39,$A7,$33
	.dc.b	$F1,$67,$3D,$1F,$9F,$24,$7E,$A5,$AF,$E8,$1A,$FE,$AD,$AF,$F0,$16
	.dc.b	$BF,$C0,$1F,$D5,$7C,$3C,$8F,$DA,$35,$FD,$5B,$5F,$DB,$35,$FE,$18
	.dc.b	$D7,$F8,$5B,$5F,$E3,$3D,$5F,$B7,$23,$FC,$05,$AF,$EE,$1A,$FF,$03
	.dc.b	$6B,$FC,$65,$AF,$F1,$87,$F8,$17,$6F,$F2,$12,$47,$F1,$6F,$46,$BF
	.dc.b	$B3,$8B,$9F,$F5,$A9,$BA,$FF,$25,$9B,$5F,$D5,$67,$8B,$FC,$56,$3B
	.dc.b	$BF,$C6,$5A,$FF,$10,$6B,$B3,$44,$34,$53,$47,$E8,$FC,$FF,$E7,$87
	.dc.b	$89,$FE,$83,$18,$7F,$98,$C6,$75,$4E,$67,$91,$FD,$2D,$FE,$95,$F8
	.dc.b	$23,$E3,$A9,$27,$FA,$A4,$3D,$1C,$47,$F1,$C8,$E0,$EA,$79,$CE,$B3
	.dc.b	$FE,$47,$5A,$AE,$15,$A7,$81,$0C,$9E,$0A,$E5,$9F,$FB,$3A,$B9,$5F
	.dc.b	$4B,$E5,$7C,$09,$19,$79,$1E,$AF,$82,$F7,$F9,$D2,$30,$43,$D8,$E5
	.dc.b	$F7,$47,$C6,$31,$3E,$C9,$C4,$F1,$8C,$3F,$22,$32,$F2,$30,$43,$04
	.dc.b	$8E,$2F,$88,$E9,$87,$FB,$BB,$E7,$9E,$B7,$F8,$05,$B3,$FE,$09,$6B
	.dc.b	$F5,$7D,$1D,$4C,$8D,$3F,$80,$8C,$47,$F5,$69,$E3,$7F,$D6,$9B,$7F
	.dc.b	$46,$B7,$5F,$DE,$E7,$BB,$6F,$FB,$94,$62,$3D,$11,$82,$3F,$DF,$E3
	.dc.b	$48,$38,$6C,$70,$7F,$C1,$11,$93,$CA,$F1,$CD,$E8,$FE,$70,$F6,$E3
	.dc.b	$EB,$BC,$57,$E7,$27,$38,$F9,$A3,$04,$3E,$24,$FE,$9B,$E3,$64,$7E
	.dc.b	$91,$FD,$4F,$A1,$88,$FD,$24,$5C,$3F,$43,$36,$FF,$35,$36,$BD,$13
	.dc.b	$56,$29,$C2,$9B,$18,$6C,$F8,$CF,$DB,$BF,$BC,$F8,$6F,$DE,$F9,$F2
	.dc.b	$7F,$77,$19,$7F,$7B,$19,$7F,$79,$19,$7F,$7D,$19,$C7,$F7,$73,$97
	.dc.b	$F7,$B1,$9C,$7F,$6D,$39,$7F,$71,$19,$AF,$AD,$39,$7E,$C4,$65,$F8
	.dc.b	$31,$F9,$DB,$91,$F9,$D7,$F8,$BF,$9F,$88,$FC,$C4,$5C,$3F,$87,$CD
	.dc.b	$C3,$F5,$13,$6B,$F0,$D3,$C5,$FC,$D4,$7E,$60,$8F,$E5,$5A,$FE,$41
	.dc.b	$AF,$03,$44,$34,$5A,$FF,$21,$F8,$3F,$E4,$46,$DF,$8E,$7F,$3F,$5F
	.dc.b	$E7,$BA,$D4,$FE,$8E,$51,$87,$F4,$DB,$7E,$0D,$F8,$87,$FA,$26,$34
	.dc.b	$F0,$9E,$D0,$F6,$BB,$12,$78,$78,$D5,$FF,$2B,$CE,$DE,$93,$E5,$9D
	.dc.b	$63,$EC,$8B,$7E,$7B,$56,$2F,$F7,$91,$CD,$EB,$5C,$DE,$A4,$32,$79
	.dc.b	$0C,$5F,$76,$DF,$7B,$F9,$1F,$19,$F2,$7A,$E5,$7C,$38,$CB,$F1,$63
	.dc.b	$2F,$BA,$32,$FC,$08,$CD,$FD,$89,$CB,$F6,$A3,$37,$EF,$9C,$BF,$06
	.dc.b	$33,$5F,$02,$72,$FC,$28,$CB,$F7,$23,$F9,$D7,$AD,$23,$F9,$D3,$F8
	.dc.b	$3E,$EC,$47,$F3,$58,$B8,$7E,$32,$6E,$1F,$D1,$66,$D7,$D0,$9E,$2F
	.dc.b	$F3,$78,$FE,$6A,$47,$F9,$93,$5F,$E5,$AD,$78,$1A,$21,$A2,$D7,$E7
	.dc.b	$BC,$0F,$CD,$1E,$28,$FE,$77,$38,$7F,$2D,$19,$D5,$39,$9F,$C8,$CF
	.dc.b	$1D,$7F,$D0,$91,$F9,$DF,$13,$F4,$C7,$E9,$FC,$CB,$19,$21,$F9,$A7
	.dc.b	$2F,$E7,$22,$E3,$FB,$49,$B9,$FE,$66,$6D,$7E,$62,$6B,$FB,$28,$B6
	.dc.b	$9B,$18,$7F,$75,$F1,$3F,$BB,$F3,$65,$FE,$D6,$32,$FE,$DE,$32,$FE
	.dc.b	$C2,$32,$FE,$CA,$33,$8F,$F0,$A9,$CB,$FC,$32,$33,$8F,$EC,$67,$2F
	.dc.b	$EC,$E3,$35,$FB,$29,$CB,$FB,$48,$CB,$FC,$3A,$3F,$C1,$8C,$3F,$C1
	.dc.b	$87,$F1,$3A,$FF,$6E,$FE,$0A,$F3,$B5,$3F,$E0,$92,$CE,$A2,$E1,$FE
	.dc.b	$19,$9A,$7F,$7F,$AE,$4F,$63,$FC,$08,$7F,$82,$88,$60,$87,$F8,$08
	.dc.b	$C9,$4C,$9F,$E1,$53,$25,$3F,$C1,$D5,$EA,$F3,$BD,$95,$FC,$BE,$91
	.dc.b	$E6,$AF,$F0,$99,$A3,$D0,$D5,$1C,$1F,$51,$F0,$CE,$F3,$D6,$78,$4E
	.dc.b	$E7,$42,$BE,$6F,$9F,$A2,$47,$23,$E4,$6B,$87,$86,$1E,$99,$C5,$7D
	.dc.b	$6A,$B3,$9A,$E4,$7A,$2B,$C7,$3A,$8F,$A9,$15,$F6,$C5,$B4,$D8,$C3
	.dc.b	$67,$AC,$BE,$73,$DF,$E5,$C9,$CC,$7C,$23,$0F,$A8,$FC,$24,$65,$D9
	.dc.b	$D1,$DE,$3F,$92,$9D,$4F,$13,$E9,$CE,$B1,$FB,$93,$AA,$FC,$CC,$E6
	.dc.b	$BF,$25,$39,$AF,$C1,$4E,$67,$EE,$CF,$DF,$FA,$07,$C4,$52,$4F,$C0
	.dc.b	$F3,$6B,$E3,$9D,$BF,$41,$AF,$9F,$24,$F0,$C9,$97,$F8,$06,$F7,$18
	.dc.b	$BF,$F3,$34,$62,$7F,$31,$5E,$09,$FF,$04,$09,$EE,$E1,$AF,$5C,$36
	.dc.b	$D2,$9C,$5A,$D3,$31,$C9,$F1,$C6,$A3,$11,$B3,$AB,$83,$D0,$FF,$01
	.dc.b	$F3,$7F,$46,$7C,$1B,$F1,$47,$C5,$8C,$BF,$26,$B4,$F1,$3F,$1F,$18
	.dc.b	$7F,$67,$EB,$FE,$B8,$8F,$E5,$3C,$CA,$FD,$18,$B9,$7F,$1C,$9B,$87
	.dc.b	$F0,$39,$B5,$FC,$5E,$6B,$D8,$FC,$FC,$7E,$F2,$33,$9A,$2C,$61,$B3
	.dc.b	$FB,$2F,$C1,$BF,$DB,$77,$FE,$3F,$7F,$DD,$D1,$9F,$2F,$F8,$74,$6A
	.dc.b	$3F,$D4,$CD,$4F,$07,$F4,$4D,$C7,$F1,$11,$AF,$2F,$99,$D2,$A4,$78
	.dc.b	$E3,$2F,$84,$87,$47,$2F,$68,$E2,$E4,$E4,$70,$9F,$0D,$78,$EF,$DD
	.dc.b	$7F,$AB,$E2,$97,$FA,$57,$DF,$EF,$62,$3F,$55,$17,$0E,$73,$6F,$F4
	.dc.b	$D3,$6B,$E2,$CF,$17,$F4,$11,$F9,$52,$3F,$9A,$6B,$F9,$66,$BC,$0D
	.dc.b	$10,$D1,$6B,$DB,$E5,$FD,$63,$C5,$1F,$B1,$9C,$3F,$7E,$33,$C1,$FE
	.dc.b	$7C,$8C,$10,$F7,$9F,$85,$F7,$BF,$02,$47,$F1,$1E,$5D,$7E,$FC,$5C
	.dc.b	$3E,$7C,$DC,$BF,$19,$36,$BF,$13,$35,$FC,$4C,$5D,$3F,$DD,$31,$E1
	.dc.b	$21,$D8,$F0,$98,$BF,$BB,$3C,$6B,$F2,$FE,$4F,$E6,$3C,$69,$7F,$9F
	.dc.b	$8C,$BF,$A1,$8C,$BF,$98,$8C,$BF,$9A,$8C,$DF,$E2,$A7,$2F,$E3,$23
	.dc.b	$38,$FE,$A6,$72,$FE,$AE,$33,$5F,$A8,$9C,$BF,$AA,$8C,$BF,$AA,$8F
	.dc.b	$E7,$DE,$B4,$8F,$E7,$0F,$EC,$BD,$CC,$47,$F3,$D8,$B8,$7E,$7A,$6E
	.dc.b	$1F,$CE,$66,$D7,$EE,$27,$8B,$FC,$CE,$3F,$94,$11,$FE,$5A,$D7,$F9
	.dc.b	$4B,$5E,$06,$88,$68,$B5,$FB,$0F,$2F,$F8,$49,$E2,$8F,$E8,$93,$87
	.dc.b	$F8,$94,$67,$54,$E6,$7F,$85,$CF,$1D,$5E,$6F,$F1,$42,$4F,$F8,$33
	.dc.b	$F0,$7F,$F6,$8C,$EA,$3F,$8F,$4E,$9F,$EE,$D3,$A8,$C6,$3F,$C1,$9F
	.dc.b	$F1,$64,$66,$78,$90,$E3,$9F,$03,$F9,$44,$D8,$09,$E4,$FF,$24,$8D
	.dc.b	$39,$9E,$FE,$5F,$83,$CB,$93,$EB,$8C,$BE,$D8,$CB,$EA,$8C,$BE,$C8
	.dc.b	$CE,$1F,$CC,$30,$FF,$70,$36,$B3,$7F,$EF,$D9,$FF,$73,$DF,$72,$1C
	.dc.b	$CF,$01,$8A,$E4,$E0,$D1,$E3,$3F,$C7,$BF,$2E,$F6,$BF,$CA,$CD,$8C
	.dc.b	$C7,$CE,$F4,$F2,$C8,$C9,$F6,$E3,$0F,$F8,$B6,$79,$36,$79,$3C,$0F
	.dc.b	$F0,$5C,$69,$FF,$09,$B5,$FA,$A1,$C3,$F6,$1A,$E6,$3F,$C2,$31,$82
	.dc.b	$38,$24,$7F,$86,$23,$2F,$E4,$FC,$EF,$C5,$9E,$AB,$FE,$C7,$3E,$A7
	.dc.b	$EB,$46,$A7,$D5,$1A,$AC,$47,$E6,$27,$96,$3F,$67,$E1,$7F,$2F,$E5
	.dc.b	$CA,$FD,$14,$7E,$68,$C3,$FA,$08,$C3,$A3,$F5,$0E,$B0,$FC,$44,$EA
	.dc.b	$78,$1A,$38,$1F,$AA,$78,$57,$E2,$EB,$85,$72,$AE,$07,$23,$81,$FB
	.dc.b	$7F,$2F,$F5,$E4,$7F,$75,$E5,$D7,$F6,$F1,$67,$F3,$93,$6F,$F3,$53
	.dc.b	$6B,$F6,$3A,$FF,$B9,$5E,$C4,$39,$9D,$8C,$5F,$E9,$E7,$94,$7E,$E6
	.dc.b	$ED,$26,$C7,$E7,$3C,$F9,$5F,$A0,$7F,$84,$FB,$58,$8F,$CC,$45,$C3
	.dc.b	$F6,$93,$71,$FC,$9E,$CA,$C5,$FA,$AB,$93,$F9,$A8,$FC,$99,$1F,$C8
	.dc.b	$35,$FD,$0B,$5E,$06,$88,$68,$B7,$FC,$1F,$CD,$FE,$42,$78,$A3,$F4
	.dc.b	$D3,$87,$F9,$74,$66,$FD,$11,$82,$1F,$00,$FF,$5F,$CB,$E2,$7F,$DB
	.dc.b	$0F,$FB,$46,$FF,$C8,$AD,$C7,$F6,$AE,$EB,$8D,$70,$F3,$24,$60,$87
	.dc.b	$FA,$C6,$79,$3F,$CC,$9A,$FC,$D9,$F0,$9C,$9E,$13,$87,$9D,$F9,$23
	.dc.b	$4F,$9A,$75,$3C,$0F,$6C,$E2,$7D,$7E,$67,$B8,$93,$F0,$BC,$1A,$F9
	.dc.b	$E2,$CF,$C2,$9B,$8F,$C2,$9B,$5E,$B9,$AF,$E1,$E2,$E3,$FE,$3B,$8E
	.dc.b	$87,$D0,$8C,$CE,$9B,$1F,$CD,$AB,$F9,$A9,$1F,$AF,$ED,$D7,$F9,$84
	.dc.b	$5C,$3E,$BC,$D8,$FC,$C9,$67,$E8,$47,$17,$F9,$8C,$7F,$2A,$23,$FC
	.dc.b	$9D,$AF,$F3,$86,$BC,$0D,$10,$D1,$67,$F1,$3E,$6F,$E3,$0F,$14,$7F
	.dc.b	$3D,$9C,$3F,$92,$8C,$E9,$FE,$94,$AE,$33,$93,$E2,$1F,$98,$F2,$FF
	.dc.b	$2F,$E5,$C8,$C1,$0F,$C8,$39,$7F,$25,$1F,$9E,$31,$3F,$9E,$9C,$4F
	.dc.b	$18,$C3,$FA,$68,$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$1F,$AA,$F2
	.dc.b	$7F,$5E,$48,$FD,$9F,$95,$5F,$D4,$45,$AF,$D9,$CD,$AE,$55,$FB,$AA
	.dc.b	$FD,$CC,$E6,$3F,$5D,$15,$FE,$35,$17,$6F,$F6,$94,$71,$3C,$25,$3C
	.dc.b	$25,$BF,$0B,$FE,$2A,$7F,$D9,$3F,$46,$F8,$5F,$EF,$31,$86,$E3,$FD
	.dc.b	$F2,$75,$5C,$63,$FB,$FF,$E1,$1F,$E0,$08,$5F,$17,$47,$F8,$E2,$B8
	.dc.b	$BA,$AE,$97,$DA,$3F,$C7,$B7,$CE,$F9,$46,$CE,$53,$B5,$7D,$27,$FD
	.dc.b	$C7,$5C,$6F,$AD,$79,$4E,$25,$38,$DD,$69,$9B,$F3,$BE,$D6,$BD,$0E
	.dc.b	$8E,$1E,$A6,$23,$A4,$66,$3A,$99,$3E,$28,$AF,$F7,$53,$F0,$BC,$DF
	.dc.b	$74,$FB,$6F,$C1,$87,$BA,$7E,$15,$F0,$BE,$B3,$F2,$0D,$3D,$A3,$57
	.dc.b	$CE,$3E,$7F,$B5,$CB,$CC,$91,$B8,$EA,$61,$FA,$47,$9E,$32,$EC,$EC
	.dc.b	$EE,$FE,$9C,$EA,$78,$DA,$B5,$8F,$E2,$AB,$57,$F6,$A7,$37,$E6,$3B
	.dc.b	$1C,$9C,$BF,$26,$7E,$1F,$99,$23,$E0,$3F,$1B,$CC,$C4,$78,$A2,$DF
	.dc.b	$E2,$26,$E1,$F9,$C9,$B5,$F9,$49,$AF,$D7,$8B,$8F,$FB,$B2,$3A,$1E
	.dc.b	$02,$1F,$02,$77,$3C,$1F,$BB,$ED,$49,$FC,$1E,$68,$FB,$BF,$0B,$99
	.dc.b	$FA,$33,$97,$73,$A6,$90,$DB,$96,$9F,$9D,$7F,$2A,$47,$04,$3F,$3A
	.dc.b	$E5,$A6,$4F,$D2,$39,$69,$F8,$FF,$3B,$F2,$A7,$A7,$3F,$E1,$F3,$89
	.dc.b	$D1,$E7,$8C,$B8,$3F,$0F,$3C,$73,$FE,$1F,$ED,$7E,$83,$CB,$95,$F9
	.dc.b	$D8,$FD,$19,$87,$F1,$F1,$87,$47,$E8,$DD,$5F,$F0,$89,$D4,$F0,$34
	.dc.b	$70,$3F,$32,$F0,$AF,$E4,$35,$C2,$B9,$57,$03,$91,$C0,$FD,$97,$95
	.dc.b	$FB,$82,$3F,$C0,$BF,$09,$FE,$D1,$7F,$CB,$27,$33,$FA,$87,$F9,$04
	.dc.b	$EA,$3F,$93,$C6,$AB,$F9,$0B,$AA,$C1,$CE,$7C,$27,$77,$2F,$23,$27
	.dc.b	$EF,$AB,$B3,$FC,$56,$76,$D3,$91,$F9,$8F,$4B,$F8,$B9,$1F,$67,$C2
	.dc.b	$57,$F3,$31,$70,$61,$FE,$C8,$C1,$79,$8F,$EA,$15,$6F,$FE,$00,$D6
	.dc.b	$9F,$01,$FC,$DE,$3F,$97,$91,$C1,$0F,$E6,$EE,$5A,$64,$FE,$7C,$E5
	.dc.b	$A7,$D5,$F3,$7E,$89,$E3,$7F,$A8,$C6,$1D,$1E,$43,$DE,$60,$F9,$91
	.dc.b	$C6,$7F,$CF,$B1,$F3,$FC,$AF,$BA,$7D,$EE,$89,$1F,$85,$8C,$BF,$87
	.dc.b	$8C,$BC,$4D,$19,$21,$92,$65,$F4,$1F,$97,$1C,$A3,$10,$A5,$78,$23
	.dc.b	$F0,$95,$BC,$7E,$C4,$E6,$F8,$DF,$B1,$BE,$57,$E6,$09,$87,$E7,$FC
	.dc.b	$AA,$FE,$3E,$2C,$F2,$9B,$87,$E1,$66,$D7,$E0,$A6,$BE,$78,$B6,$9B
	.dc.b	$18,$6C,$FE,$BE,$3F,$57,$1F,$E0,$BF,$A6,$77,$52,$57,$ED,$7C,$CA
	.dc.b	$FF,$62,$9D,$BF,$DE,$9A,$FE,$A6,$37,$97,$CC,$9C,$B6,$BF,$77,$3D
	.dc.b	$2B,$FC,$55,$5C,$9B,$3F,$E1,$58,$C4,$08,$E1,$B1,$B2,$3F,$E1,$F8
	.dc.b	$D2,$69,$23,$71,$FC,$07,$CA,$FE,$1E,$7A,$B3,$F0,$47,$19,$CB,$FC
	.dc.b	$66,$74,$F8,$8C,$10,$F9,$11,$FC,$6B,$D9,$FE,$37,$E5,$49,$FE,$45
	.dc.b	$19,$7F,$92,$46,$5F,$E2,$91,$97,$F8,$B4,$67,$2F,$C8,$CE,$5F,$C9
	.dc.b	$C6,$70,$FE,$1D,$39,$7F,$88,$46,$6B,$F8,$6C,$E5,$FE,$1F,$19,$7F
	.dc.b	$49,$1F,$E5,$1F,$C2,$BF,$DD,$51,$FE,$24,$6D,$FE,$A9,$B7,$F2,$1F
	.dc.b	$F7,$39,$89,$FB,$4D,$4D,$37,$E5,$F7,$24,$7F,$92,$23,$2D,$8C,$BE
	.dc.b	$3F,$2E,$DF,$8E,$6D,$F8,$E6,$D7,$8A,$6B,$E2,$8F,$D2,$7C,$84,$8F
	.dc.b	$CF,$3F,$07,$E1,$71,$1F,$A3,$8B,$87,$C1,$9B,$87,$E2,$36,$56,$2F
	.dc.b	$E0,$D7,$27,$F4,$31,$F9,$F2,$3F,$99,$6B,$FE,$7B,$9E,$6E,$48,$66
	.dc.b	$36,$F4,$33,$7F,$57,$CD,$FB,$47,$8E,$3F,$49,$38,$7E,$74,$66,$BB
	.dc.b	$C7,$DF,$8D,$3F,$04,$FA,$9E,$DF,$E1,$7C,$C9,$18,$21,$F8,$77,$2F
	.dc.b	$E2,$63,$F1,$86,$30,$FC,$54,$E2,$78,$C6,$1F,$C1,$46,$5E,$46,$08
	.dc.b	$60,$91,$C5,$F1,$1D,$32,$FC,$B7,$97,$F9,$D2,$47,$E3,$F8,$6B,$F9
	.dc.b	$F8,$B5,$F9,$F9,$B5,$CA,$BF,$2D,$5F,$A7,$9C,$C7,$E8,$22,$BF,$A0
	.dc.b	$8B,$BA,$6C,$61,$B3,$FC,$BF,$DC,$FE,$70,$47,$F4,$9F,$07,$5F,$E6
	.dc.b	$D1,$71,$FD,$A4,$DC,$7F,$29,$36,$BF,$67,$3C,$5F,$E6,$71,$FC,$C4
	.dc.b	$8F,$F2,$B6,$BF,$93,$6B,$C0,$D1,$0D,$16,$BF,$83,$79,$9F,$C2,$CF
	.dc.b	$14,$7F,$35,$9C,$3F,$BF,$8C,$EC,$9C,$CF,$F1,$19,$E3,$E1,$BB,$BF
	.dc.b	$DC,$CF,$FB,$14,$8E,$0C,$F9,$72,$BF,$DA,$73,$C5,$B3,$97,$FD,$83
	.dc.b	$1C,$9C,$9C,$27,$58,$FF,$82,$A7,$94,$F0,$AC,$E7,$FD,$2A,$79,$4E
	.dc.b	$4E,$44,$30,$78,$3C,$AF,$59,$1F,$3F,$97,$5D,$1E,$A8,$F8,$31,$9C
	.dc.b	$3C,$91,$6F,$C5,$36,$BF,$B6,$5A,$B1,$1F,$EB,$39,$B3,$98,$EC,$78
	.dc.b	$8A,$78,$89,$7F,$0E,$7F,$B4,$BF,$E2,$EF,$A8,$7A,$78,$C7,$FB,$6C
	.dc.b	$62,$FF,$BF,$56,$EB,$ED,$E1,$B9,$FF,$01,$56,$DF,$F7,$AB,$6B,$84
	.dc.b	$7D,$C9,$E0,$A7,$57,$E0,$BC,$C6,$27,$E1,$D6,$27,$FC,$03,$12,$7C
	.dc.b	$0D,$8C,$BC,$CA,$77,$6B,$4C,$C7,$27,$81,$C9,$C3,$EE,$8E,$31,$88
	.dc.b	$E4,$6E,$3F,$3F,$E5,$C8,$F9,$37,$D3,$0F,$87,$1F,$9A,$9D,$3F,$22
	.dc.b	$BD,$CF,$E8,$A3,$33,$C0,$FC,$E7,$B7,$FA,$CF,$2A,$57,$EA,$23,$F5
	.dc.b	$E6,$1F,$D3,$C6,$1D,$1F,$AE,$75,$3D,$23,$27,$03,$46,$1D,$39,$23
	.dc.b	$82,$61,$FA,$D9,$EA,$FE,$BE,$BF,$81,$F9,$52,$FF,$7E,$FF,$06,$F2
	.dc.b	$B1,$1F,$BE,$8B,$87,$EE,$E6,$DF,$EE,$66,$D7,$EA,$26,$AD,$13,$85
	.dc.b	$36,$30,$D9,$F1,$9F,$C4,$9F,$C9,$F8,$D2,$3E,$D9,$FA,$07,$F8,$EF
	.dc.b	$C1,$66,$3F,$43,$17,$1F,$E2,$53,$71,$FD,$6C,$DA,$FD,$FC,$F1,$7F
	.dc.b	$39,$1F,$8E,$23,$F9,$66,$BF,$95,$6B,$C0,$D1,$0D,$16,$BF,$D5,$BF
	.dc.b	$85,$FF,$BB,$0F,$F7,$6E,$FF,$EA,$E9,$F6,$9F,$B5,$AF,$DB,$CE,$F0
	.dc.b	$F4,$4E,$4D,$C6,$5E,$35,$CD,$5F,$83,$D9,$E7,$E5,$C9,$E3,$1D,$8C
	.dc.b	$3C,$30,$FF,$3F,$E0,$AD,$4F,$03,$C2,$F0,$BE,$D1,$93,$91,$FE,$97
	.dc.b	$9E,$4F,$03,$51,$C1,$D8,$8D,$B8,$6C,$6C,$95,$E6,$AF,$13,$E8,$BF
	.dc.b	$93,$E5,$CC,$3E,$3B,$F3,$32,$C4,$75,$8B,$87,$D6,$9B,$7F,$0A,$6D
	.dc.b	$7A,$E6,$BE,$78,$B8,$FD,$1B,$FB,$F1,$83,$C1,$3E,$22,$1E,$13,$ED
	.dc.b	$D7,$F3,$2F,$8D,$91,$F7,$1F,$C1,$FC,$26,$23,$F9,$FC,$5B,$FC,$6C
	.dc.b	$DC,$7F,$A5,$CD,$AF,$A5,$3C,$5F,$E7,$51,$FC,$B0,$8F,$F3,$46,$BF
	.dc.b	$CC,$DA,$F0,$34,$43,$45,$AF,$CF,$F9,$7F,$9A,$3C,$51,$FD,$46,$70
	.dc.b	$FE,$6A,$33,$AA,$73,$3F,$81,$9E,$39,$FF,$55,$8F,$C3,$79,$5F,$A6
	.dc.b	$3F,$4F,$CD,$23,$F4,$B1,$97,$F4,$F1,$97,$F1,$91,$97,$F1,$D1,$9C
	.dc.b	$7F,$6F,$39,$7F,$73,$19,$CF,$F4,$B3,$97,$F4,$F1,$9A,$FD,$04,$E5
	.dc.b	$FD,$14,$7E,$F3,$C9,$91,$FB,$A7,$F7,$FE,$56,$23,$F4,$11,$70,$FE
	.dc.b	$13,$37,$0F,$DA,$CD,$AF,$D7,$4D,$7F,$55,$17,$14,$D8,$C3,$67,$C4
	.dc.b	$7F,$0C,$7F,$C0,$FF,$00,$F8,$0A,$49,$FE,$2F,$F8,$67,$F9,$4D,$FE
	.dc.b	$33,$3B,$6C,$ED,$FD,$EC,$ED,$FD,$FC,$ED,$B3,$BC,$99,$26,$5F,$19
	.dc.b	$CC,$66,$F9,$3F,$E1,$66,$BC,$43,$87,$85,$FF,$85,$23,$11,$A8,$C1
	.dc.b	$1F,$F0,$DC,$68,$91,$98,$FF,$10,$C7,$AE,$7C,$B0,$A7,$99,$FE,$32
	.dc.b	$6C,$7C,$62,$1F,$1A,$34,$7C,$3C,$3D,$57,$D5,$F6,$57,$2A,$E0,$72
	.dc.b	$BF,$0C,$3D,$BF,$07,$97,$2B,$EC,$C7,$8C,$C3,$F5,$A3,$0E,$8F,$BE
	.dc.b	$EA,$BA,$CE,$A7,$81,$A3,$81,$F8,$07,$85,$FA,$2B,$85,$72,$AE,$07
	.dc.b	$23,$81,$F8,$EF,$2F,$E4,$11,$FC,$8F,$97,$5F,$C6,$C5,$9F,$83,$36
	.dc.b	$FE,$3C,$DA,$FC,$5C,$D7,$EB,$45,$C5,$36,$30,$D9,$F2,$1E,$23,$F4
	.dc.b	$3E,$B7,$EA,$08,$FE,$9B,$E0,$EB,$FA,$08,$B8,$FD,$69,$B8,$FD,$E9
	.dc.b	$B5,$FA,$09,$E2,$FE,$76,$3F,$30,$47,$F2,$AD,$7F,$26,$D7,$81,$A2
	.dc.b	$1A,$2D,$7E,$23,$CC,$FC,$B1,$E2,$8F,$D4,$CE,$1F,$E0,$31,$9D,$53
	.dc.b	$99,$FD,$FC,$F1,$CD,$ED,$7E,$87,$C9,$95,$FC,$2A,$3F,$48,$61,$FE
	.dc.b	$0F,$18,$74,$7F,$10,$75,$8F,$E7,$A7,$53,$C0,$D1,$C0,$FE,$22,$F0
	.dc.b	$CF,$F5,$35,$C2,$B9,$57,$03,$91,$C0,$FF,$1A,$FE,$1B,$FE,$DC,$8F
	.dc.b	$DE,$B6,$BF,$87,$36,$7F,$8C,$2F,$2B,$FD,$62,$E6,$0D,$24,$6B,$CA
	.dc.b	$93,$E0,$78,$C6,$23,$FC,$4F,$3C,$67,$31,$B2,$1A,$72,$D3,$37,$EA
	.dc.b	$8C,$BF,$C6,$35,$F4,$4F,$F3,$CB,$FE,$86,$7A,$A2,$C6,$48,$79,$1C
	.dc.b	$BF,$CF,$A2,$CF,$C8,$9C,$CF,$F3,$79,$FE,$64,$47,$F9,$7B,$5F,$E5
	.dc.b	$AD,$78,$1A,$21,$A2,$CF,$D6,$F2,$7E,$89,$E2,$C7,$F9,$5D,$62,$BE
	.dc.b	$EC,$E5,$FE,$99,$19,$7E,$7C,$65,$FB,$3B,$7D,$CF,$26,$47,$E1,$A3
	.dc.b	$EF,$98,$7E,$BC,$61,$D1,$F8,$C7,$58,$7E,$32,$75,$3C,$0D,$1C,$0F
	.dc.b	$C0,$BC,$30,$FC,$8D,$70,$AE,$55,$C0,$E4,$70,$3F,$37,$E4,$7E,$50
	.dc.b	$8F,$E7,$BC,$9A,$FE,$3A,$2C,$FE,$82,$6D,$FE,$82,$6D,$7E,$82,$6B
	.dc.b	$F9,$E8,$BA,$A6,$C6,$1B,$3E,$43,$C4,$7F,$A9,$3E,$B1,$F4,$1F,$DE
	.dc.b	$F1,$FE,$DB,$18,$9D,$4F,$EB,$BA,$B5,$97,$F8,$06,$B5,$5F,$E8,$E6
	.dc.b	$DF,$2B,$E1,$3C,$88,$6D,$CB,$4F,$F0,$61,$D5,$AD,$28,$9F,$0B,$C1
	.dc.b	$AE,$CE,$2E,$1F,$F1,$03,$5E,$51,$D1,$A7,$68,$D1,$23,$F8,$7F,$93
	.dc.b	$27,$D7,$5E,$F7,$F8,$DC,$65,$F6,$12,$3F,$8E,$C6,$5E,$11,$FC,$43
	.dc.b	$D9,$F9,$AF,$F8,$D7,$D3,$FF,$0F,$CF,$F8,$81,$B8,$FF,$24,$AF,$E4
	.dc.b	$27,$0A,$4E,$9E,$53,$A7,$F9,$7C,$EE,$75,$1C,$27,$4F,$8C,$FF,$74
	.dc.b	$C6,$8D,$EA,$AD,$DF,$96,$FA,$18,$3C,$D7,$C9,$CB,$E2,$3A,$F9,$3E
	.dc.b	$32,$3E,$0F,$2A,$BF,$ED,$AC,$53,$8B,$F8,$95,$6F,$FC,$A9,$5A,$BE
	.dc.b	$A7,$32,$9C,$EF,$83,$F3,$EA,$BF,$2A,$2C,$FF,$B9,$AB,$AC,$60,$F9
	.dc.b	$55,$F4,$74,$7B,$F4,$98,$7C,$AF,$82,$AF,$E9,$E2,$E5,$F7,$66,$E3
	.dc.b	$FA,$D9,$B5,$F8,$39,$E2,$FE,$7E,$3F,$31,$7E,$08,$FC,$C4,$60,$8E
	.dc.b	$08,$7E,$91,$CB,$4C,$9D,$4F,$C0,$F9,$7F,$87,$3C,$6F,$EC,$63,$0E
	.dc.b	$8F,$E1,$8E,$B3,$FF,$7C,$61,$F5,$2B,$1E,$CF,$E4,$49,$7F,$8E,$F2
	.dc.b	$6B,$F7,$E2,$E7,$FA,$19,$B7,$F7,$A6,$D7,$E9,$E6,$BF,$A2,$8B,$82
	.dc.b	$6C,$61,$B3,$F9,$E8,$FD,$24,$7E,$97,$C8,$FD,$37,$93,$2B,$F4,$71
	.dc.b	$97,$F4,$B1,$97,$F2,$D1,$97,$F3,$11,$9C,$7F,$75,$39,$7F,$79,$19
	.dc.b	$C7,$F3,$33,$97,$F3,$71,$9A,$FE,$09,$39,$7F,$83,$46,$5F,$E0,$31
	.dc.b	$F9,$3F,$62,$47,$F3,$F7,$F8,$07,$C0,$E2,$3F,$9C,$45,$C3,$F8,$74
	.dc.b	$DC,$3F,$9F,$4D,$AF,$E2,$93,$C5,$FE,$61,$1F,$92,$23,$FC,$A5,$AF
	.dc.b	$F3,$26,$BC,$0D,$10,$D1,$6B,$F9,$07,$93,$FC,$5C,$F1,$47,$F4,$19
	.dc.b	$C3,$FC,$5A,$33,$BA,$73,$3F,$C2,$67,$8F,$E1,$D9,$8F,$F7,$15,$FF
	.dc.b	$B3,$23,$10,$8C,$57,$F2,$D9,$FF,$62,$46,$27,$8E,$3F,$EE,$39,$E3
	.dc.b	$39,$9E,$53,$9F,$2B,$95,$C9,$E2,$7F,$BC,$EF,$8D,$6E,$32,$23,$2E
	.dc.b	$E7,$93,$D8,$E5,$58,$BF,$11,$2F,$A4,$EC,$E9,$5B,$8F,$53,$EF,$AF
	.dc.b	$81,$E5,$7C,$1F,$2A,$5F,$93,$1F,$F7,$13,$67,$2F,$58,$4F,$28,$D1
	.dc.b	$C5,$DE,$3F,$46,$B5,$3C,$67,$38,$FC,$D9,$CD,$7D,$79,$CD,$7C,$99
	.dc.b	$CD,$7D,$39,$CC,$FC,$99,$FE,$CD,$F5,$CD,$E9,$27,$D9,$C9,$5F,$8D
	.dc.b	$3B,$7E,$9B,$5F,$74,$6F,$16,$29,$95,$B9,$CC,$62,$FF,$19,$7D,$E3
	.dc.b	$FC,$43,$5D,$5B,$3F,$E1,$58,$C4,$75,$0E,$1E,$87,$F8,$99,$C3,$A8
	.dc.b	$E4,$47,$4D,$7F,$C5,$F1,$E7,$BF,$CE,$79,$5F,$A1,$3D,$CF,$86,$35
	.dc.b	$3F,$99,$9D,$3F,$1A,$B6,$E4,$FC,$44,$61,$FC,$E7,$B3,$FA,$D2,$3F
	.dc.b	$8D,$F2,$AB,$F8,$B8,$B9,$7E,$DE,$6E,$3F,$B1,$9B,$5F,$AB,$9A,$F6
	.dc.b	$3F,$31,$1F,$BB,$8C,$EA,$8B,$18,$6C,$FE,$7F,$C9,$FE,$05,$E4,$C8
	.dc.b	$C1,$0F,$D0,$39,$7F,$45,$1F,$C2,$4C,$63,$FC,$3E,$71,$3C,$63,$0F
	.dc.b	$EB,$63,$2F,$23,$04,$30,$48,$E2,$F8,$8E,$98,$FD,$7F,$8D,$FD,$09
	.dc.b	$23,$F8,$6F,$E1,$FF,$E7,$0A,$FC,$AC,$7E,$32,$74,$FE,$89,$FF,$27
	.dc.b	$4E,$EB,$A4,$FE,$92,$77,$5C,$AB,$53,$FA,$29,$FC,$D1,$1F,$CB,$35
	.dc.b	$FC,$8B,$5F,$D1,$B5,$EE,$68,$E2,$43,$F8,$E4,$7F,$BB,$9B,$E4,$F9
	.dc.b	$2F,$8D,$F9,$63,$D6,$66,$7C,$C7,$C4,$72,$FB,$4F,$8F,$EC,$7C,$22
	.dc.b	$3F,$2B,$CA,$AF,$8A,$2D,$7C,$39,$B7,$FB,$9C,$15,$8B,$FE,$11,$0A
	.dc.b	$C4,$23,$6F,$88,$F0,$14,$F0,$16,$FC,$0D,$BF,$05,$F4,$AE,$D3,$F8
	.dc.b	$2F,$2B,$EE,$F9,$52,$30,$43,$E6,$B9,$7E,$7C,$7E,$10,$C6,$1F,$81
	.dc.b	$9C,$4F,$18,$C3,$F9,$08,$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$1F
	.dc.b	$C4,$FE,$03,$F9,$E9,$23,$F3,$BF,$05,$5F,$E7,$71,$6B,$F2,$D3,$6B
	.dc.b	$95,$7F,$3E,$AF,$CF,$CF,$18,$FE,$6B,$1F,$CC,$C8,$FF,$2F,$6B,$FC
	.dc.b	$A9,$AF,$03,$44,$34,$5A,$FD,$47,$97,$FA,$E3,$C9,$1F,$D2,$A7,$0F
	.dc.b	$EE,$23,$2F,$F4,$28,$CB,$FA,$28,$CE,$4F,$67,$F7,$24,$9F,$E0,$3E
	.dc.b	$55,$7F,$53,$17,$0F,$DD,$4D,$CB,$F3,$13,$6B,$F2,$F3,$5F,$CB,$45
	.dc.b	$B4,$D8,$C3,$67,$F8,$8C,$7F,$0C,$8F,$E1,$BE,$4F,$FB,$6F,$F1,$0F
	.dc.b	$F6,$A9,$FE,$D6,$BF,$F6,$85,$FF,$18,$9D,$61,$23,$04,$3F,$C3,$D5
	.dc.b	$CA,$3A,$C6,$71,$FE,$3F,$39,$8C,$4F,$ED,$6B,$A4,$72,$8E,$2F,$F0
	.dc.b	$E8,$CD,$72,$9C,$BF,$CA,$BE,$C7,$E8,$57,$F8,$E9,$AF,$F8,$C5,$B1
	.dc.b	$E2,$D3,$FD,$F5,$18,$AF,$EF,$CD,$AF,$F0,$63,$56,$0A,$E3,$9C,$9F
	.dc.b	$03,$F2,$E3,$11,$FE,$09,$39,$C5,$7B,$B5,$C9,$6B,$D2,$F1,$6B,$C0
	.dc.b	$D4,$71,$51,$A7,$99,$DD,$C3,$A8,$E1,$1E,$DA,$FF,$78,$4F,$F8,$AD
	.dc.b	$AE,$CF,$69,$FE,$66,$72,$6A,$3E,$31,$CF,$CA,$FC,$61,$F0,$B1,$FC
	.dc.b	$0F,$B5,$F8,$F2,$5F,$3F,$2A,$BF,$8A,$8B,$3F,$88,$9B,$9F,$F8,$CF
	.dc.b	$15,$62,$FD,$15,$5F,$AF,$16,$7F,$DC,$B1,$ED,$8C,$1F,$A1,$9B,$1E
	.dc.b	$D8,$C3,$EF,$F2,$7F,$4D,$E4,$C8,$C1,$0F,$D0,$B9,$7F,$47,$1F,$AD
	.dc.b	$31,$8F,$E8,$A7,$13,$C6,$30,$FC,$D8,$CB,$C8,$C1,$0C,$12,$38,$BE
	.dc.b	$23,$A6,$3F,$9C,$E3,$FC,$F9,$23,$ED,$7C,$15,$7F,$4D,$16,$BF,$6F
	.dc.b	$36,$B9,$57,$EB,$6B,$F8,$44,$F1,$8F,$CB,$47,$E3,$08,$FE,$6D,$AF
	.dc.b	$E4,$DA,$F0,$34,$43,$45,$AF,$E1,$3E,$07,$F1,$73,$C9,$1F,$B6,$9C
	.dc.b	$3F,$95,$8C,$BF,$A4,$8C,$BF,$C5,$A3,$3A,$3D,$7F,$E4,$44,$9F,$CF
	.dc.b	$79,$15,$FE,$3B,$17,$0F,$E3,$73,$72,$FC,$A4,$DA,$FD,$54,$D7,$F8
	.dc.b	$4C,$5B,$4D,$8C,$36,$7F,$5D,$1F,$CD,$63,$FD,$25,$7F,$E4,$8D,$BF
	.dc.b	$DA,$B4,$7F,$DC,$F8,$7F,$A5,$DB,$1F,$EA,$57,$64,$F1,$F9,$CE,$DB
	.dc.b	$3B,$F2,$2B,$D5,$B5,$D2,$3C,$13,$89,$91,$FE,$EA,$9D,$12,$33,$1F
	.dc.b	$EE,$38,$CC,$E2,$FD,$95,$DA,$25,$73,$8F,$E6,$FF,$2B,$23,$F9,$F3
	.dc.b	$E4,$F8,$3C,$47,$F3,$F8,$B7,$F1,$E6,$E3,$FC,$FA,$6D,$7C,$59,$E2
	.dc.b	$FF,$2F,$8F,$E4,$A4,$7F,$9A,$35,$FE,$58,$D7,$81,$A2,$1A,$2D,$7D
	.dc.b	$9F,$2B,$E5,$9E,$28,$FE,$73,$38,$7E,$54,$67,$95,$39,$9F,$95,$3C
	.dc.b	$73,$7B,$3F,$58,$93,$F4,$FC,$8A,$FD,$18,$B8,$FE,$2E,$6E,$7F,$89
	.dc.b	$9B,$5F,$89,$9A,$FE,$3E,$2D,$A6,$C6,$1B,$3F,$83,$8F,$C4,$C7,$E2
	.dc.b	$BC,$6F,$C5,$F8,$F2,$BF,$11,$19,$7F,$15,$19,$7F,$11,$19,$7F,$15
	.dc.b	$19,$C7,$F4,$93,$97,$F4,$D1,$9C,$7F,$49,$39,$7F,$4D,$19,$AF,$D3
	.dc.b	$4E,$5F,$D4,$46,$5F,$D7,$47,$E9,$3D,$17,$C1,$52,$4F,$E8,$5A,$B0
	.dc.b	$FF,$5F,$CF,$08,$19,$C2,$A8,$D3,$FA,$7A,$E7,$5C,$9B,$5F,$BF,$7B
	.dc.b	$CF,$F8,$3E,$B8,$B6,$3F,$C0,$D1,$88,$EE,$D7,$C6,$D7,$C5,$1C,$08
	.dc.b	$FF,$85,$E3,$44,$8C,$C7,$EF,$7C,$8F,$E0,$E7,$99,$EB,$18,$7F,$83
	.dc.b	$46,$5E,$E6,$08,$7C,$53,$89,$F2,$2F,$FD,$D3,$F8,$8B,$D9,$FE,$2E
	.dc.b	$FF,$B4,$E3,$3E,$4E,$26,$57,$1C,$BE,$7D,$ED,$B5,$A9,$F0,$09,$EC
	.dc.b	$F1,$3E,$CB,$C5,$B1,$C6,$FC,$B7,$E4,$89,$3C,$4E,$91,$BB,$F0,$C6
	.dc.b	$6F,$A3,$A3,$13,$C9,$F2,$C6,$8F,$2B,$8F,$23,$E1,$91,$E9,$E4,$D7
	.dc.b	$47,$78,$F9,$31,$9C,$7E,$6C,$5C,$3E,$3C,$DA,$F8,$33,$5F,$8D,$16
	.dc.b	$FF,$DD,$31,$D8,$F3,$C6,$67,$4F,$EA,$F8,$24,$FC,$5D,$7F,$8D,$BF
	.dc.b	$0F,$E0,$72,$FD,$48,$CB,$67,$F0,$33,$AC,$BF,$5F,$17,$2F,$BF,$3C
	.dc.b	$AB,$F4,$D3,$FA,$42,$3F,$9D,$6B,$FE,$77,$BE,$EF,$03,$44,$35,$1C
	.dc.b	$1D,$1F,$5B,$B7,$F2,$A7,$8D,$FD,$7C,$61,$FC,$04,$65,$FE,$05,$19
	.dc.b	$79,$1F,$15,$FC,$FF,$B3,$FA,$0F,$26,$4F,$E2,$63,$2F,$E2,$E3,$2F
	.dc.b	$E2,$23,$2F,$E2,$A3,$39,$7E,$B2,$72,$FE,$BA,$33,$97,$EA,$27,$2F
	.dc.b	$EA,$A3,$35,$FA,$49,$CB,$FA,$68,$CB,$FA,$B8,$FD,$BF,$8B,$23,$F4
	.dc.b	$0F,$EE,$BC,$9C,$47,$E7,$A2,$E1,$FC,$02,$6E,$1F,$BA,$9B,$5F,$B4
	.dc.b	$9A,$FE,$E6,$2E,$FF,$EE,$68,$EE,$43,$B1,$FB,$19,$DC,$F0,$7F,$A0
	.dc.b	$76,$C9,$FC,$B3,$FC,$43,$E0,$71,$1F,$95,$8B,$8F,$F1,$69,$B8,$FF
	.dc.b	$30,$9B,$5F,$C2,$A7,$8B,$FC,$E6,$3F,$9B,$91,$FE,$60,$D7,$F2,$8D
	.dc.b	$78,$1A,$21,$A2,$D2,$BF,$A7,$FE,$25,$67,$F8,$A3,$5F,$51,$EE,$33
	.dc.b	$1F,$E4,$48,$D4,$FF,$5C,$AF,$5D,$27,$35,$D8,$CB,$FD,$76,$3B,$F9
	.dc.b	$1C,$CE,$9C,$12,$32,$43,$F9,$9C,$F1,$6C,$71,$75,$87,$A6,$74,$D9
	.dc.b	$D4,$F2,$AE,$CF,$23,$8D,$FC,$48,$CD,$71,$9C,$9F,$DE,$2F,$FD,$CA
	.dc.b	$76,$AF,$97,$E4,$7C,$52,$4F,$CC,$EC,$AF,$AA,$2D,$F9,$E6,$D7,$2A
	.dc.b	$FA,$75,$F3,$E7,$31,$E1,$8A,$FC,$F8,$B7,$FE,$E4,$8E,$27,$84,$A7
	.dc.b	$84,$B7,$E1,$7F,$AA,$7D,$A3,$E7,$3F,$D0,$2D,$8C,$C7,$E0,$7E,$13
	.dc.b	$2C,$8C,$9F,$8A,$8C,$3F,$DA,$A7,$93,$67,$93,$C0,$FF,$10,$9F,$E2
	.dc.b	$48,$DB,$86,$BF,$E2,$66,$BB,$8E,$2E,$5A,$E4,$B1,$C1,$34,$91,$C2
	.dc.b	$32,$D8,$F4,$D7,$E3,$7C,$AF,$D1,$1E,$E7,$F5,$11,$AC,$FF,$5B,$3A
	.dc.b	$BF,$75,$6A,$F1,$1F,$9C,$9E,$55,$FB,$4F,$67,$F5,$DE,$54,$AF,$DA
	.dc.b	$C7,$EC,$4C,$3F,$A4,$8C,$3A,$3F,$76,$EB,$0F,$D2,$4E,$A7,$81,$A3
	.dc.b	$81,$FA,$E7,$85,$7F,$06,$AE,$15,$CA,$B8,$1C,$8E,$07,$F0,$AF,$27
	.dc.b	$F8,$31,$1F,$E1,$BE,$4D,$7F,$77,$16,$7F,$71,$36,$FF,$8A,$CD,$AF
	.dc.b	$D8,$4D,$7F,$87,$45,$D7,$FD,$D7,$1E,$12,$1D,$8F,$09,$8B,$FE,$0D
	.dc.b	$3C,$6B,$F3,$FA,$CA,$FD,$1B,$FE,$E2,$FC,$4D,$F0,$D9,$9F,$CE,$4D
	.dc.b	$C3,$F9,$45,$6A,$78,$90,$D9,$FE,$41,$74,$E0,$FC,$CE,$EA,$FC,$81
	.dc.b	$23,$F3,$2D,$76,$D7,$F4,$CD,$7F,$4C,$D7,$F5,$8F,$AF,$92,$47,$96
	.dc.b	$FF,$33,$38,$8F,$F3,$0D,$FF,$BF,$EF,$C9,$1B,$D3,$FD,$2F,$2C,$38
	.dc.b	$61,$CA,$BD,$DE,$CF,$C9,$23,$F0,$7C,$8A,$F6,$8B,$3F,$2A,$6E,$1D
	.dc.b	$A6,$D7,$FA,$A3,$05,$62,$10,$9C,$3A,$8F,$19,$0F,$01,$E3,$31,$7F
	.dc.b	$6A,$78,$D7,$E0,$7B,$7F,$05,$E4,$4B,$FB,$51,$97,$EE,$46,$5F,$93
	.dc.b	$19,$7E,$5C,$66,$FF,$0F,$39,$7F,$13,$19,$C7,$E6,$CE,$5F,$9F,$19
	.dc.b	$AF,$85,$39,$7E,$24,$65,$F8,$71,$FC,$EB,$A2,$47,$F3,$E7,$F3,$9F
	.dc.b	$07,$88,$FE,$6D,$17,$0F,$CD,$CD,$C3,$F9,$84,$DA,$FA,$53,$C5,$FE
	.dc.b	$6D,$1F,$C9,$C8,$FF,$30,$6B,$F9,$E6,$BC,$0D,$10,$D1,$6B,$F5,$BE
	.dc.b	$4F,$EC,$CF,$14,$7F,$3A,$9C,$3F,$AD,$8C,$ED,$FD,$D6,$3F,$11,$1A
	.dc.b	$7D,$E7,$EC,$FD,$7F,$DF,$11,$FD,$B7,$91,$5F,$D1,$C5,$C3,$F7,$F3
	.dc.b	$72,$FD,$1C,$DA,$FC,$B4,$D7,$F2,$B1,$77,$4D,$8C,$36,$7C,$87,$88
	.dc.b	$FF,$64,$7A,$DF,$EB,$EC,$3F,$C6,$CF,$FB,$18,$9E,$42,$76,$D7,$FD
	.dc.b	$B7,$1B,$F2,$71,$5C,$AA,$E5,$BA,$EC,$DA,$CD,$F5,$9E,$EF,$63,$CB
	.dc.b	$1A,$51,$A7,$C3,$59,$8F,$14,$71,$32,$72,$21,$83,$FD,$E7,$F6,$DF
	.dc.b	$E7,$96,$F8,$5D,$E3,$18,$FC,$4F,$83,$C4,$CA,$FF,$78,$56,$B2,$F7
	.dc.b	$56,$67,$99,$AA,$F3,$46,$63,$FC,$2B,$5E,$16,$CF,$F8,$1E,$31,$1F
	.dc.b	$3D,$AF,$D3,$6A,$69,$23,$91,$23,$FC,$33,$19,$7F,$03,$E5,$7C,$93
	.dc.b	$D1,$87,$F9,$DE,$7D,$93,$F5,$67,$4F,$AC,$87,$E0,$9C,$BF,$00,$FC
	.dc.b	$2E,$3F,$57,$C9,$91,$F9,$08,$FB,$06,$1F,$C5,$C6,$1D,$1F,$96,$75
	.dc.b	$7F,$96,$9D,$4F,$03,$47,$03,$C2,$60,$D1,$82,$47,$E8,$27,$AE,$5F
	.dc.b	$97,$AF,$C6,$79,$12,$FF,$4E,$FE,$3F,$C9,$C4,$7E,$8A,$2E,$1F,$AA
	.dc.b	$9B,$7F,$A1,$9B,$5F,$9F,$9A,$FE,$72,$2D,$A6,$C6,$1B,$3E,$23,$F7
	.dc.b	$2F,$E5,$FE,$8E,$47,$E5,$1F,$D2,$7C,$26,$23,$F3,$F1,$71,$FD,$FC
	.dc.b	$DC,$7F,$1D,$36,$BF,$69,$3C,$5F,$CB,$C7,$E5,$48,$FE,$31,$AF,$F9
	.dc.b	$A3,$44,$62,$38,$9B,$21,$B7,$8D,$FF,$12,$D7,$F7,$87,$8E,$7F,$3D
	.dc.b	$38,$7F,$8C,$C6,$6B,$BC,$60,$87,$C0,$3F,$8F,$FB,$1F,$C1,$C8,$FF
	.dc.b	$1D,$BF,$F6,$B7,$E2,$9F,$EC,$E8,$FE,$29,$3A,$7F,$8F,$BF,$C4,$E3
	.dc.b	$51,$FC,$52,$35,$3F,$CA,$9D,$65,$83,$F9,$13,$67,$F9,$65,$72,$D5
	.dc.b	$39,$9E,$87,$96,$BF,$98,$56,$E7,$55,$C2,$B4,$74,$AF,$E3,$5F,$2D
	.dc.b	$FC,$2C,$91,$FC,$09,$AF,$D7,$6B,$FC,$51,$AF,$F1,$16,$BF,$C4,$5F
	.dc.b	$F4,$E5,$FF,$63,$83,$49,$E5,$FC,$7E,$A9,$3E,$63,$6F,$CB,$8C,$C3
	.dc.b	$2F,$A7,$38,$7E,$AC,$DC,$3E,$1C,$DA,$F9,$73,$5E,$A7,$7B,$F1,$B6
	.dc.b	$75,$18,$8F,$E8,$7C,$7F,$CC,$48,$FF,$28,$6B,$F9,$D6,$BF,$CE,$9A
	.dc.b	$FF,$37,$6B,$FC,$DD,$FE,$79,$E9,$C8,$FE,$66,$D7,$F5,$0D,$7F,$A2
	.dc.b	$B5,$FE,$84,$D7,$FA,$13,$5F,$E8,$5C,$DF,$D2,$48,$FF,$40,$6B,$FB
	.dc.b	$56,$BF,$D4,$9A,$FF,$4F,$6B,$FD,$3D,$FE,$A9,$CD,$23,$FA,$5B,$5F
	.dc.b	$E0,$0D,$7F,$AE,$B5,$FE,$B4,$D7,$FA,$D3,$5F,$EB,$47,$F6,$12,$1F
	.dc.b	$D6,$0A,$7F,$0D,$29,$FD,$A8,$A7,$F6,$72,$9F,$D9,$CF,$F7,$97,$DC
	.dc.b	$3E,$96,$92,$3F,$DF,$B1,$96,$C6,$5F,$F1,$01,$FE,$23,$8D,$38,$0E
	.dc.b	$1D,$B5,$FF,$14,$35,$E5,$1D,$5D,$F4,$F9,$88,$F9,$1A,$F7,$9F,$F1
	.dc.b	$FB,$92,$19,$8D,$BD,$0C,$9E,$4E,$AF,$86,$47,$E0,$B5,$EC,$D7,$81
	.dc.b	$A2,$1A,$29,$F0,$3D,$4F,$9E,$47,$E6,$B5,$F5,$35,$E0,$68,$86,$8A
	.dc.b	$7C,$CF,$88,$FB,$64,$7E,$CB,$5F,$8C,$D7,$81,$A2,$1A,$29,$F6,$3B
	.dc.b	$BF,$0E,$47,$F0,$AD,$7E,$93,$5E,$06,$88,$68,$A7,$E1,$3D,$6F,$C9
	.dc.b	$91,$FC,$8B,$5F,$BC,$D7,$81,$A2,$1A,$29,$F9,$0F,$07,$F3,$E4,$7F
	.dc.b	$3A,$D7,$F1,$4D,$78,$1A,$21,$A2,$9F,$9C,$F9,$EF,$D5,$91,$FD,$4B
	.dc.b	$5F,$CB,$35,$E0,$68,$86,$8A,$7E,$A3,$C4,$FD,$B9,$1F,$DA,$B5,$FD
	.dc.b	$13,$5E,$06,$88,$68,$A7,$ED,$3C,$4F,$E0,$64,$7F,$80,$B5,$FD,$73
	.dc.b	$5E,$06,$88,$68,$A7,$F0,$0F,$84,$FE,$1E,$47,$F8,$6B,$5F,$DD,$35
	.dc.b	$E0,$68,$86,$8A,$7F,$0C,$F8,$7F,$E3,$64,$7F,$8C,$B5,$FE,$0C,$D7
	.dc.b	$81,$A2,$1A,$29,$FC,$63,$C7,$FE,$4E,$47,$F9,$2B,$5F,$E2,$4D,$78
	.dc.b	$1A,$21,$A2,$9F,$C9,$3C,$8F,$E6,$64,$7F,$98,$B5,$FE,$3C,$D7,$81
	.dc.b	$A2,$1A,$29,$FC,$C3,$DB,$FE,$7E,$47,$F9,$EB,$5F,$E5,$4D,$78,$1A
	.dc.b	$21,$A2,$9F,$CF,$3E,$53,$FA,$59,$1F,$E9,$2D,$7F,$9B,$35,$E0,$68
	.dc.b	$86,$8A,$7F,$48,$F2,$FF,$AB,$91,$FE,$AA,$D7,$FA,$13,$5E,$06,$88
	.dc.b	$68,$A7,$F5,$4F,$77,$FB,$19,$1F,$EC,$2D,$7F,$A7,$35,$E0,$68,$86
	.dc.b	$8A,$7F,$60,$F3,$BF,$B7,$91,$FE,$DA,$D7,$FA,$D3,$5E,$06,$88,$68
	.dc.b	$A7,$F6,$CF,$97,$FE,$F6,$47,$FB,$CB,$5F,$EC,$CD,$78,$1A,$21,$A2
	.dc.b	$9F,$DF,$A9,$E8,$7F,$7D,$80,$8C,$B9,$6B,$FE,$03,$6B,$CE,$39,$3C
	.dc.b	$5B,$1A,$8E,$A6,$4E,$91,$A2,$38,$3C,$1E,$8F,$22,$38,$21,$E3,$72
	.dc.b	$D3,$27,$BD,$CB,$4F,$67,$C4,$F9,$88,$E0,$87,$BD,$CB,$4C,$9F,$2D
	.dc.b	$CB,$4F,$91,$EB,$FC,$22,$38,$21,$F2,$DC,$B4,$C9,$F5,$DC,$B4,$FA
	.dc.b	$9E,$AF,$CE,$23,$82,$1F,$5D,$CB,$4C,$9F,$83,$72,$D3,$F0,$1E,$B7
	.dc.b	$DA,$23,$82,$1F,$83,$72,$D3,$27,$E3,$DC,$B4,$FC,$67,$C0,$FE,$18
	.dc.b	$8E,$08,$7E,$3D,$CB,$4C,$9F,$9B,$72,$D3,$F3,$1F,$53,$F9,$22,$38
	.dc.b	$21,$F9,$B7,$2D,$32,$7E,$9D,$CB,$4F,$D2,$7B,$3F,$9E,$23,$82,$1F
	.dc.b	$A7,$72,$D3,$27,$EC,$DC,$B4,$FD,$87,$C1,$FE,$A8,$8E,$08,$7E,$CD
	.dc.b	$CB,$4C,$9F,$BF,$72,$D3,$F7,$9E,$CF,$ED,$88,$E0,$87,$EF,$DC,$B4
	.dc.b	$C9,$FC,$2D,$CB,$4F,$E1,$1F,$0F,$FC,$08,$8E,$08,$7F,$0B,$72,$D3
	.dc.b	$27,$F1,$77,$2D,$3F,$8A,$7B,$9F,$C3,$88,$E0,$87,$F1,$77,$2D,$32
	.dc.b	$7F,$23,$72,$D3,$F9,$07,$BB,$FC,$68,$8E,$08,$7F,$23,$72,$D3,$27
	.dc.b	$F2,$F7,$2D,$3F,$96,$7C,$5F,$F2,$62,$38,$21,$FC,$BD,$CB,$4C,$9F
	.dc.b	$CE,$DC,$B4,$FE,$71,$F4,$5F,$CC,$88,$E0,$87,$F3,$B7,$2D,$32,$7F
	.dc.b	$47,$72,$D3,$FA,$27,$9F,$FC,$F8,$8E,$08,$7F,$47,$72,$D3,$27,$F5
	.dc.b	$37,$2D,$3F,$A8,$79,$FF,$D2,$88,$E0,$87,$F5,$37,$2D,$32,$7F,$5F
	.dc.b	$72,$D3,$FA,$E7,$C3,$FF,$56,$23,$82,$1F,$D7,$DC,$B4,$C9,$FD,$AD
	.dc.b	$CB,$4F,$ED,$1F,$21,$FD,$88,$8E,$08,$7F,$6B,$72,$D3,$27,$F7,$77
	.dc.b	$2D,$3F,$BA,$7A,$5F,$DB,$88,$E0,$87,$F7,$77,$2D,$32,$7F,$80,$DC
	.dc.b	$B4,$FF,$02,$7D,$CB,$D7,$FE,$FA,$D8,$CC,$60,$8E,$08,$7F,$80,$0F
	.dc.b	$F0,$6C,$71,$70,$1C,$3B,$6B,$FE,$08,$6B,$CA,$3C,$7F,$07,$23,$CE
	.dc.b	$D7,$AC,$A3,$11,$D5,$AF,$56,$BE,$66,$BE,$5F,$A8,$F8,$24,$7E,$2B
	.dc.b	$5E,$ED,$7D,$4D,$7D,$4D,$7E,$1B,$F3,$3D,$B9,$1F,$41,$AF,$A9,$AF
	.dc.b	$C5,$6B,$F1,$5A,$FC,$E6,$BF,$37,$E1,$7E,$C9,$1F,$BA,$D7,$E3,$B5
	.dc.b	$FA,$4D,$7E,$93,$5F,$B6,$FE,$13,$DE,$91,$F8,$86,$BF,$49,$AF,$DD
	.dc.b	$6B,$F7,$5A,$FE,$19,$AF,$E1,$7E,$67,$F2,$24,$7F,$2A,$D7,$EF,$B5
	.dc.b	$FC,$53,$5F,$C5,$35,$FC,$9B,$F9,$CF,$80,$91,$FA,$06,$BF,$8A,$6B
	.dc.b	$F9,$56,$BF,$95,$6B,$F9,$E6,$BF,$9D,$F8,$1F,$D4,$91,$FD,$6B,$5F
	.dc.b	$CB,$B5,$FD,$13,$5F,$D1,$35,$FD,$5B,$FB,$4F,$8D,$91,$FB,$86,$BF
	.dc.b	$A2,$6B,$FA,$D6,$BF,$AD,$6B,$FB,$66,$BF,$B5,$F8,$2F,$E0,$24,$7F
	.dc.b	$82,$B5,$FD,$7B,$5F,$DD,$35,$FD,$D3,$5F,$E0,$6F,$F0,$CF,$83,$91
	.dc.b	$FC,$41,$AF,$EE,$9A,$FF,$05,$6B,$FC,$15,$AF,$F0,$E6,$BF,$C3,$7E
	.dc.b	$0B,$F8,$C9,$1F,$E3,$AD,$7F,$83,$B5,$FE,$24,$D7,$F8,$93,$5F,$E3
	.dc.b	$6F,$F2,$4F,$89,$91,$FC,$A1,$AF,$F1,$26,$BF,$C7,$5A,$FF,$1D,$6B
	.dc.b	$FC,$99,$AF,$F2,$5F,$83,$FE,$62,$47,$F9,$AB,$5F,$E3,$ED,$7F,$95
	.dc.b	$35,$FE,$54,$D7,$F9,$9B,$FC,$F3,$E1,$64,$7F,$40,$6B,$FC,$A9,$AF
	.dc.b	$F3,$56,$BF,$CD,$5A,$FF,$3E,$6B,$FC,$F7,$E2,$BF,$A4,$91,$FE,$9A
	.dc.b	$D7,$F9,$BB,$5F,$E8,$4D,$7F,$A1,$35,$FE,$96,$FF,$54,$F8,$39,$1F
	.dc.b	$D6,$1A,$FF,$42,$6B,$FD,$35,$AF,$F4,$D6,$BF,$D5,$9A,$FF,$55,$F8
	.dc.b	$3F,$EC,$24,$7F,$B2,$B5,$FE,$9E,$D7,$FA,$D3,$5F,$EB,$4D,$7F,$B1
	.dc.b	$BC,$7E,$15,$D1,$FD,$7F,$E0,$FF,$DC,$F1,$8A,$FE,$D3,$18,$9E,$71
	.dc.b	$FE,$F8,$9D,$3E,$73,$11,$CD,$A7,$22,$9B,$11,$CB,$A1,$7D,$E7,$81
	.dc.b	$AA,$E8,$D3,$D6,$74,$95,$FF,$82,$09,$1F,$E0,$C6,$BF,$DD,$9A,$FF
	.dc.b	$80,$5A,$F6,$30,$71,$3F,$C2,$A7,$F8,$50,$86,$08,$7F,$80,$8C,$94
	.dc.b	$C9,$FE,$0F,$32,$53,$FC,$2A,$3E,$E9,$F0,$5F,$0E,$30,$D8,$CC,$7C
	.dc.b	$13,$11,$1F,$9F,$1A,$7E,$41,$FE,$21,$34,$61,$D6,$3F,$71,$B8,$7F
	.dc.b	$88,$E7,$67,$F8,$B6,$3A,$36,$BD,$2F,$98,$F5,$C6,$91,$1C,$36,$38
	.dc.b	$C6,$23,$91,$FE,$38,$70,$48,$DF,$C1,$49,$EE,$D7,$83,$5F,$43,$5F
	.dc.b	$19,$F3,$DC,$3E,$56,$BF,$4F,$E0,$BC,$A4,$7D,$AD,$7C,$CD,$7E,$33
	.dc.b	$5E,$F4,$8C,$47,$C3,$7E,$0F,$C1,$48,$F9,$2D,$7E,$13,$5F,$A4,$D7
	.dc.b	$AB,$5F,$9C,$D7,$F1,$9F,$05,$F3,$88,$FD,$66,$BF,$3D,$AF,$DF,$6B
	.dc.b	$EA,$6B,$F6,$DF,$B3,$F0,$52,$3F,$02,$D7,$ED,$35,$FC,$53,$5F,$8A
	.dc.b	$D7,$F0,$CD,$7F,$49,$F0,$5F,$86,$23,$F8,$E6,$BF,$87,$6B,$F9,$76
	.dc.b	$BF,$49,$AF,$E4,$DE,$3A,$7E,$87,$E0,$7F,$5C,$7E,$BE,$24,$FE,$B2
	.dc.b	$32,$FE,$BA,$32,$FE,$AE,$32,$FE,$B6,$2B,$FA,$98,$FC,$F9,$1F,$D3
	.dc.b	$B5,$FD,$03,$5E,$06,$88,$68,$FD,$7F,$C0,$FE,$A8,$8E,$08,$7E,$AD
	.dc.b	$CB,$4C,$9F,$BD,$72,$D3,$F7,$7F,$03,$FB,$62,$38,$21,$FB,$77,$2D
	.dc.b	$32,$7F,$09,$72,$D3,$F8,$3F,$C0,$FF,$02,$23,$82,$1F,$C0,$DC,$B4
	.dc.b	$C9,$FC,$55,$CB,$4F,$E2,$7F,$03,$FC,$38,$8E,$08,$7F,$0F,$72,$D3
	.dc.b	$27,$F2,$17,$2D,$3F,$8F,$FC,$0F,$F1,$A2,$38,$21,$FC,$6D,$CB,$4C
	.dc.b	$9F,$CB,$5C,$B4,$E3,$F8,$73,$DD,$FE,$16,$7F,$1A,$71,$1F,$CB,$63
	.dc.b	$07,$F1,$09,$B1,$FC,$C5,$AF,$F4,$F8,$BC,$3C,$A1,$56,$15,$A3,$75
	.dc.b	$2B,$F9,$ED,$6E,$71,$1C,$A7,$CE,$E9,$AE,$4F,$21,$93,$D4,$53,$CC
	.dc.b	$5A,$F3,$57,$78,$C4,$F8,$23,$D0,$F9,$E3,$27,$F5,$2E,$4F,$EB,$51
	.dc.b	$FD,$09,$B1,$93,$CA,$E1,$FF,$0F,$9E,$22,$D7,$C7,$8D,$12,$38,$90
	.dc.b	$F5,$9C,$3E,$03,$C0,$FF,$44,$6B,$FD,$95,$AF,$F8,$0D,$B2,$69,$23
	.dc.b	$D0,$71,$29,$C4,$B1,$C5,$B1,$C5,$F5,$FC,$07,$62,$3C,$1A,$E8,$C1
	.dc.b	$0C,$12,$BF,$C0,$D1,$C1,$FF,$04,$7C,$07,$A8,$93,$E4,$6B,$F3,$1B
	.dc.b	$1C,$23,$44,$34,$57,$FB,$DA,$FB,$C7,$C0,$FC,$A6,$C7,$C4,$6B,$F7
	.dc.b	$1B,$3C,$9B,$1F,$E0,$66,$BF,$E1,$A6,$BD,$E7,$B1,$1E,$CD,$7B,$7C
	.dc.b	$0F,$8C,$8F,$8D,$AE,$8C,$10,$C1,$2B,$CD,$1C,$1F,$47,$C0,$FB,$C9
	.dc.b	$3E,$F6,$BE,$F6,$C7,$08,$D1,$0D,$15,$D3,$FB,$73,$E7,$7A,$8C,$7F
	.dc.b	$6D,$8A,$31,$3B,$85,$FF,$10,$6C,$70,$9D,$ED,$FC,$46,$79,$5F,$4B
	.dc.b	$E0,$71,$AD,$ED,$F5,$5B,$87,$D5,$9D,$57,$52,$47,$8E,$32,$F9,$67
	.dc.b	$8B,$EC,$9F,$03,$C3,$0F,$24,$EB,$0F,$2C,$F3,$35,$87,$8E,$EC,$78
	.dc.b	$CA,$75,$3C,$0F,$66,$92,$FB,$C6,$5B,$19,$9E,$CF,$84,$C1,$F9,$5B
	.dc.b	$F3,$1E,$7C,$3D,$F5,$C8,$95,$F9,$D8,$CB,$D2,$33,$5E,$93,$A9,$E9
	.dc.b	$72,$FE,$BE,$78,$BE,$1B,$F7,$E1,$F7,$A3,$8B,$F2,$2B,$8B,$96,$E1
	.dc.b	$E8,$AF,$81,$87,$8C,$97,$D4,$F8,$15,$C2,$7B,$E1,$E2,$9F,$35,$70
	.dc.b	$7C,$C7,$92,$FB,$3C,$A3,$58,$79,$5B,$5C,$E3,$13,$E1,$C3,$F2,$D3
	.dc.b	$9B,$E7,$18,$9E,$D8,$79,$1A,$7C,$2A,$E1,$19,$8F,$8B,$5C,$B0,$FA
	.dc.b	$91,$C2,$FE,$3D,$F0,$8E,$46,$67,$CE,$D3,$CC,$F2,$7A,$DF,$7A,$F9
	.dc.b	$11,$C8,$C9,$5C,$9E,$07,$A1,$F3,$0F,$C2,$3A,$9F,$0C,$72,$8E,$0F
	.dc.b	$E7,$0F,$E5,$97,$FC,$11,$B5,$FE,$E0,$C3,$F5,$71,$83,$F9,$D9,$FA
	.dc.b	$73,$04,$3F,$42,$53,$F0,$85,$38,$1A,$21,$A2,$9A,$30,$43,$04,$7C
	.dc.b	$AF,$07,$F0,$6F,$C3,$23,$F9,$BB,$FC,$5D,$7D,$B9,$F8,$F1,$8A,$FA
	.dc.b	$4D,$8F,$A0,$DB,$F9,$27,$CD,$8F,$B6,$72,$29,$E0,$AF,$DF,$3F,$5A
	.dc.b	$FC,$A7,$8D,$DB,$88,$DC,$7D,$7A,$FB,$35,$DE,$3D,$78,$7E,$3E,$3C
	.dc.b	$73,$E5,$36,$59,$EB,$3B,$8C,$C7,$82,$BE,$13,$FB,$23,$C0,$70,$3D
	.dc.b	$0F,$92,$BC,$57,$EA,$BE,$95,$B3,$89,$E6,$38,$E1,$FC,$21,$F0,$3F
	.dc.b	$39,$A7,$CE,$2D,$7B,$1B,$87,$A1,$F9,$93,$C5,$A7,$12,$CF,$16,$CF
	.dc.b	$1B,$F9,$D5,$C5,$B3,$C5,$B5,$C5,$B5,$C7,$0F,$8F,$1F,$06,$BD,$A7
	.dc.b	$F1,$73,$F0,$47,$80,$AF,$DF,$70,$7C,$FA,$FE,$28,$FC,$88,$D1,$FC
	.dc.b	$D7,$1F,$EF,$A4,$8F,$E1,$0D,$7F,$A1,$35,$FB,$4D,$8E,$51,$C0,$87
	.dc.b	$F8,$9F,$2F,$F0,$B3,$FE,$2B,$6B,$93,$FC,$3D,$3F,$E1,$C6,$9B,$29
	.dc.b	$B3,$FC,$44,$7F,$8D,$73,$FF,$17,$47,$F8,$E5,$B1,$93,$FC,$66,$43
	.dc.b	$FC,$80,$E9,$FF,$1E,$35,$FF,$1E,$1F,$E4,$D0,$4D,$BF,$8D,$4F,$F9
	.dc.b	$56,$3F,$CA,$64,$70,$47,$88,$8E,$2E,$1A,$F5,$CF,$FC,$C2,$E5,$FF
	.dc.b	$35,$3B,$D6,$BE,$36,$BD,$5B,$26,$92,$3F,$D0,$24,$3F,$D2,$85,$AF
	.dc.b	$F3,$5C,$FF,$A1,$E3,$24,$8C,$ED,$FE,$89,$6C,$71,$3F,$D3,$87,$FA
	.dc.b	$43,$5F,$F4,$C9,$1C,$10,$FF,$55,$B9,$69,$98,$FF,$56,$3F,$EA,$A6
	.dc.b	$BF,$EB,$3D,$3F,$D7,$EF,$FA,$F8,$FF,$2B,$5F,$FB,$61,$C4,$69,$CD
	.dc.b	$7F,$B1,$5A,$69,$FF,$6A,$C7,$FB,$36,$39,$EF,$FE,$CD,$9C,$9F,$EE
	.dc.b	$D8,$FF,$6A,$46,$5F,$F5,$FC,$F1,$24,$F0,$8C,$BC,$5E,$A7,$F6,$03
	.dc.b	$F0,$D3,$5A,$66,$BD,$8F,$F6,$66,$BF,$62,$38,$46,$23,$B1,$93,$94
	.dc.b	$68,$8F,$5E,$0E,$24,$70,$43,$C0,$E5,$A6,$4F,$C4,$B9,$69,$E9,$E1
	.dc.b	$F2,$91,$C1,$0F,$63,$96,$99,$3F,$2A,$E5,$A7,$C5,$E2,$F8,$24,$70
	.dc.b	$43,$E4,$39,$69,$93,$F4,$2E,$5A,$7D,$1E,$3F,$9A,$47,$04,$3E,$A3
	.dc.b	$96,$99,$3F,$5A,$E5,$A7,$DD,$E4,$FB,$24,$70,$43,$F0,$0E,$5A,$64
	.dc.b	$FD,$CB,$96,$9F,$89,$E5,$FE,$27,$CC,$DB,$F8,$95,$61,$B5,$FC,$82
	.dc.b	$74,$FF,$21,$AD,$34,$D4,$F0,$53,$C1,$FE,$41,$5C,$1F,$E4,$55,$74
	.dc.b	$36,$64,$F4,$3B,$1D,$2F,$F9,$44,$FE,$78,$91,$ED,$21,$FA,$57,$2D
	.dc.b	$32,$7F,$1E,$72,$D3,$F5,$DC,$FF,$AA,$23,$82,$1F,$B1,$72,$D3,$27
	.dc.b	$F2,$A7,$2D,$3F,$75,$D1,$FB,$62,$38,$21,$FB,$D7,$2D,$32,$7F,$36
	.dc.b	$72,$D3,$F8,$37,$4F,$F0,$22,$38,$21,$FC,$25,$CB,$4C,$9F,$D0,$9C
	.dc.b	$B4,$FE,$25,$D5,$FC,$38,$8E,$08,$7F,$15,$72,$D3,$27,$F4,$E7,$2D
	.dc.b	$3F,$8F,$75,$FF,$1A,$23,$82,$1F,$C8,$5C,$B4,$C9,$FD,$69,$CB,$4F
	.dc.b	$E5,$5D,$9F,$C9,$88,$E0,$87,$F2,$D7,$2D,$32,$7F,$66,$72,$D3,$8E
	.dc.b	$BF,$A1,$D5,$3F,$9F,$3F,$45,$DD,$FE,$04,$AF,$CE,$1F,$E0,$D8,$D3
	.dc.b	$FA,$13,$FC,$10,$F0,$72,$70,$32,$6E,$7B,$BC,$67,$B5,$7E,$CD,$FE
	.dc.b	$99,$06,$91,$FE,$A0,$D7,$FA,$A3,$5F,$F5,$AB,$5F,$53,$5F,$F0,$FA
	.dc.b	$EE,$E6,$48,$E6,$D7,$9B,$5E,$6D,$79,$B5,$FF,$19,$C7,$57,$CB,$DF
	.dc.b	$F3,$26,$47,$C1,$8F,$9E,$61,$F9,$13,$C5,$B1,$A7,$E4,$4E,$EB,$E4
	.dc.b	$4E,$67,$E0,$BF,$18,$8F,$88,$86,$8A,$7C,$67,$6E,$48,$7C,$AF,$03
	.dc.b	$E8,$91,$FA,$2D,$7E,$8B,$5E,$06,$88,$68,$A7,$D6,$F0,$7E,$E9,$1F
	.dc.b	$BA,$D7,$EE,$B5,$E0,$68,$86,$8A,$7E,$0B,$C2,$FC,$49,$1F,$C4,$B5
	.dc.b	$FC,$4B,$5E,$06,$88,$68,$A7,$E3,$BC,$3F,$CA,$91,$FC,$AB,$5F,$CA
	.dc.b	$B5,$E0,$68,$86,$8A,$7E,$6B,$C4,$FD,$09,$1F,$D0,$B5,$FD,$0B,$5E
	.dc.b	$06,$88,$68,$A7,$E9,$BC,$5F,$D6,$91,$FD,$6B,$5F,$D6,$B5,$E0,$68
	.dc.b	$86,$8A,$7E,$CB,$C6,$FD,$C9,$1F,$DC,$B5,$FD,$CB,$5E,$06,$88,$68
	.dc.b	$A7,$EF,$BC,$7F,$E0,$A4,$7F,$82,$B5,$FE,$0A,$D7,$81,$A2,$1A,$29
	.dc.b	$FC,$2B,$C8,$FE,$22,$47,$F8,$8B,$5F,$E2,$2D,$78,$1A,$21,$A2,$9F
	.dc.b	$C5,$BC,$9F,$E3,$A4,$7F,$8E,$B5,$FE,$3A,$D7,$81,$A2,$1A,$29,$FC
	.dc.b	$8B,$CA,$FE,$52,$47,$F9,$4B,$5F,$E5,$2D,$78,$1A,$21,$A2,$9F,$CB
	.dc.b	$BC,$BF,$E6,$A4,$7F,$9A,$B5,$FE,$6A,$D7,$81,$A2,$1A,$29,$FC,$EB
	.dc.b	$CC,$FE,$82,$47,$FA,$0B,$5F,$E8,$2D,$78,$1A,$21,$A2,$9F,$D1,$BC
	.dc.b	$DF,$E9,$A4,$7F,$A6,$B5,$FE,$9A,$D7,$81,$A2,$1A,$29,$FD,$4B,$CE
	.dc.b	$FE,$B2,$47,$FA,$CB,$5F,$EB,$2D,$78,$1A,$21,$A2,$9F,$D7,$BC,$FF
	.dc.b	$EC,$A4,$7F,$B2,$B5,$FE,$CA,$D7,$81,$A2,$1A,$29,$FD,$AB,$D0,$FE
	.dc.b	$E2,$47,$FB,$8B,$5F,$EE,$2D,$78,$1A,$21,$A2,$9F,$DD,$BD,$1F,$EF
	.dc.b	$A4,$7F,$BE,$B5,$FE,$FA,$D7,$81,$A2,$1A,$29,$FE,$1B,$F2,$3F,$D0
	.dc.b	$0D,$F0,$BF,$DB,$51,$88,$11,$C3,$63,$64,$7F,$DC,$91,$A4,$D2,$46
	.dc.b	$E3,$FC,$58,$6B,$D2,$F0,$91,$E6,$D7,$72,$8C,$47,$AD,$AF,$A0,$E8
	.dc.b	$72,$21,$CB,$D3,$F7,$91,$F4,$B5,$D1,$82,$18,$21,$F2,$DE,$0F,$CD
	.dc.b	$F5,$3E,$59,$1F,$8A,$D7,$D6,$D7,$81,$A2,$1A,$29,$F6,$7D,$3F,$AE
	.dc.b	$47,$E8,$B5,$F8,$ED,$78,$1A,$21,$A2,$9F,$85,$F5,$7F,$06,$47,$EE
	.dc.b	$B5,$FA,$6D,$78,$1A,$21,$A2,$9F,$91,$F5,$BF,$1E,$47,$F1,$2D,$7E
	.dc.b	$FB,$5E,$06,$88,$68,$A7,$E7,$7D,$7F,$CD,$91,$FC,$AB,$5F,$C5,$B5
	.dc.b	$E0,$68,$86,$8A,$7E,$A7,$D8,$FD,$39,$1F,$D0,$B5,$FC,$BB,$5E,$06
	.dc.b	$88,$68,$A7,$ED,$7D,$9F,$D9,$91,$FD,$6B,$5F,$D1,$B5,$E0,$68,$86
	.dc.b	$8A,$7F,$01,$F6,$BF,$7E,$47,$F7,$2D,$7F,$5E,$D7,$81,$A2,$1A,$29
	.dc.b	$FC,$37,$DB,$FE,$16,$47,$F8,$2B,$5F,$DD,$B5,$E0,$68,$86,$8A,$7F
	.dc.b	$19,$F7,$3F,$8B,$91,$FE,$22,$D7,$F8,$3B,$5E,$06,$88,$68,$A7,$F2
	.dc.b	$5F,$77,$F9,$19,$1F,$E3,$AD,$7F,$89,$B5,$E0,$68,$86,$8A,$7F,$31
	.dc.b	$F7,$BF,$97,$91,$FE,$52,$D7,$F8,$FB,$5E,$06,$88,$68,$A7,$F3,$DF
	.dc.b	$7F,$F9,$D9,$1F,$E6,$AD,$7F,$95,$B5,$E0,$68,$86,$8A,$7F,$49,$F8
	.dc.b	$0F,$E8,$E4,$7F,$A0,$B5,$FE,$6E,$D7,$81,$A2,$1A,$29,$FD,$57,$E0
	.dc.b	$7F,$A9,$91,$FE,$9A,$D7,$FA,$1B,$5E,$06,$88,$68,$A7,$F6,$1F,$81
	.dc.b	$FE,$BE,$47,$FA,$CB,$5F,$E9,$ED,$78,$1A,$21,$A2,$9F,$DB,$7E,$03
	.dc.b	$FB,$59,$1F,$EC,$AD,$7F,$AD,$B5,$E0,$68,$86,$8A,$7F,$79,$F8,$0F
	.dc.b	$EE,$E4,$7F,$B8,$B5,$FE,$CE,$D7,$81,$A2,$1A,$29,$FD,$A4,$7C,$2B
	.dc.b	$E0,$7F,$BB,$C7,$F8,$12,$30,$E1,$AF,$F7,$C6,$BC,$DA,$FF,$75,$6B
	.dc.b	$CE,$3A,$BE,$06,$C6,$A3,$07,$78,$C9,$AF,$81,$91,$E3,$6B,$E3,$6B
	.dc.b	$E3,$6B,$E3,$6B,$ED,$6B,$F0,$BE,$07,$C6,$47,$E0,$35,$F8,$0D,$7E
	.dc.b	$03,$5F,$80,$D7,$E5,$3E,$EE,$B9,$1F,$2D,$AF,$CB,$6B,$F2,$DA,$FC
	.dc.b	$B6,$BF,$55,$AF,$DA,$F8,$3F,$96,$47,$EC,$35,$FB,$0D,$7E,$C3,$5F
	.dc.b	$B0,$D7,$F0,$4F,$D6,$F8,$49,$1F,$83,$6B,$F8,$36,$BF,$83,$6B,$F8
	.dc.b	$36,$BF,$8D,$6B,$F9,$2F,$82,$FC,$19,$1F,$C8,$35,$FC,$83,$5F,$C8
	.dc.b	$35,$FC,$83,$5F,$CD,$3F,$8E,$F8,$29,$1F,$9B,$6B,$F9,$B6,$BF,$9B
	.dc.b	$6B,$F9,$B6,$BF,$A5,$6B,$FA,$AF,$82,$FC,$D9,$1F,$D4,$35,$FD,$43
	.dc.b	$5F,$D4,$35,$FD,$43,$5F,$D9,$3F,$A6,$F8,$29,$1F,$B3,$6B,$FB,$36
	.dc.b	$BF,$B3,$6B,$FB,$36,$BF,$BD,$6B,$FC,$0B,$E0,$BF,$66,$47,$F8,$03
	.dc.b	$5F,$E0,$0D,$7F,$80,$35,$FE,$00,$D7,$F8,$53,$FB,$EF,$82,$91,$FC
	.dc.b	$2D,$AF,$F0,$B6,$BF,$C2,$DA,$FF,$0B,$6B,$FC,$55,$AF,$F1,$AF,$85
	.dc.b	$FE,$16,$47,$F8,$C3,$5F,$E3,$0D,$7F,$8C,$35,$FE,$30,$D7,$F9,$13
	.dc.b	$FC,$5B,$E0,$E4,$7F,$23,$6B,$FC,$8D,$AF,$F2,$36,$BF,$C8,$DA,$FF
	.dc.b	$2D,$6B,$FC,$CB,$E0,$FF,$91,$91,$FE,$60,$D7,$F9,$83,$5F,$E6,$0D
	.dc.b	$7F,$98,$35,$FE,$74,$FF,$2E,$F8,$39,$1F,$CE,$DA,$FF,$3B,$6B,$FC
	.dc.b	$ED,$AF,$F3,$B6,$BF,$D1,$5A,$FF,$4A,$F8,$3F,$E7,$64,$7F,$A4,$35
	.dc.b	$FE,$90,$D7,$FA,$43,$5F,$E9,$0D,$7F,$A9,$3F,$D1,$BE,$0E,$47,$F5
	.dc.b	$36,$BF,$D4,$DA,$FF,$53,$6B,$FD,$4D,$AF,$F5,$D6,$BF,$D8,$BE,$0F
	.dc.b	$FA,$99,$1F,$EC,$0D,$7F,$B0,$35,$FE,$C0,$D7,$FB,$03,$5F,$ED,$4F
	.dc.b	$F5,$EF,$83,$91,$FD,$AD,$AF,$F6,$B6,$BF,$DA,$DA,$FF,$6B,$6B,$FD
	.dc.b	$D5,$AF,$F7,$AF,$86,$FE,$D6,$47,$FB,$C3,$5F,$EF,$0D,$7F,$BC,$35
	.dc.b	$FE,$F0,$D7,$FC,$04,$FF,$78,$F8,$9F,$EF,$30,$69,$1D,$10,$FF,$05
	.dc.b	$B9,$69,$93,$FC,$24,$73,$30,$7F,$85,$57,$C1,$F8,$49,$1D,$1A,$F2
	.dc.b	$8D,$B9,$21,$98,$DB,$D5,$A8,$90,$F8,$8F,$80,$48,$F5,$B5,$F0,$35
	.dc.b	$F4,$35,$E8,$7C,$77,$0F,$67,$E6,$7C,$1C,$8F,$8E,$D7,$D8,$D7,$E2
	.dc.b	$35,$EB,$28,$C4,$7A,$9A,$FD,$7F,$83,$FB,$44,$7E,$AB,$5F,$94,$D7
	.dc.b	$E9,$35,$EE,$D7,$E3,$BF,$E3,$CF,$89,$7A,$1F,$03,$E1,$5E,$A7,$C2
	.dc.b	$9C,$DF,$CF,$9C,$BF,$46,$32,$A3,$2A,$1E,$97,$5A,$E9,$76,$B9,$35
	.dc.b	$E0,$DA,$E3,$1E,$19,$E2,$D3,$89,$4E,$25,$8E,$2F,$96,$38,$B4,$E2
	.dc.b	$53,$89,$63,$8D,$FB,$1F,$7B,$6B,$CC,$79,$EE,$D7,$86,$2B,$CE,$3D
	.dc.b	$D9,$3E,$0F,$F6,$C4,$9F,$D8,$B5,$FD,$33,$5F,$D7,$35,$FC,$5B,$5F
	.dc.b	$D1,$BF,$E7,$8F,$82,$7A,$FF,$35,$B3,$F1,$DA,$FB,$F3,$D4,$C8,$D3
	.dc.b	$F4,$E3,$11,$F1,$E7,$8E,$5F,$80,$6C,$FE,$24,$DC,$66,$AC,$66,$BA
	.dc.b	$90,$E9,$1D,$DC,$4F,$16,$CF,$17,$EC,$60,$F8,$1F,$77,$B7,$27,$F0
	.dc.b	$93,$96,$CE,$5F,$54,$FD,$48,$91,$F1,$63,$57,$F9,$CA,$E5,$3E,$0A
	.dc.b	$FE,$5C,$47,$F9,$4B,$5F,$E3,$CD,$7F,$93,$35,$FE,$06,$D7,$F8,$DB
	.dc.b	$C6,$7C,$0F,$92,$3E,$87,$BB,$67,$31,$EC,$3E,$E4,$67,$0E,$B1,$93
	.dc.b	$89,$F8,$1F,$80,$FC,$21,$1F,$C2,$7C,$05,$7E,$F4,$5B,$FB,$D3,$6B
	.dc.b	$F9,$3E,$8A,$C5,$FF,$35,$85,$62,$11,$B7,$C4,$78,$0A,$78,$0B,$7E
	.dc.b	$06,$DF,$82,$FA,$57,$69,$FC,$E3,$F9,$FF,$82,$91,$82,$1F,$99,$72
	.dc.b	$FE,$6E,$3F,$4C,$62,$FF,$39,$38,$9E,$31,$87,$F5,$51,$97,$91,$82
	.dc.b	$18,$24,$71,$7C,$47,$4B,$FE,$8B,$A7,$ED,$FE,$12,$4F,$E6,$A3,$F8
	.dc.b	$49,$87,$F6,$51,$87,$47,$EF,$5D,$57,$E6,$67,$53,$FD,$C5,$FF,$03
	.dc.b	$11,$FF,$00,$B5,$D1,$82,$18,$21,$FE,$01,$78,$47,$78,$F1,$9F,$C1
	.dc.b	$F8,$FC,$31,$98,$FD,$CC,$FF,$15,$9C,$DF,$C2,$9E,$35,$FC,$36,$7F
	.dc.b	$8D,$73,$7F,$1B,$E6,$95,$FC,$6E,$32,$FF,$1D,$8C,$BF,$C5,$E3,$2F
	.dc.b	$F1,$98,$CD,$7F,$19,$9C,$BF,$C6,$E3,$37,$F9,$E9,$CB,$FA,$08,$CD
	.dc.b	$7E,$7E,$72,$FE,$86,$32,$FE,$82,$3F,$AF,$7C,$53,$AA,$92,$7F,$AD
	.dc.b	$43,$AB,$11,$FD,$56,$38,$3A,$9E,$73,$AC,$7F,$B8,$D6,$AB,$85,$69
	.dc.b	$E0,$43,$27,$82,$B9,$63,$FE,$22,$AE,$57,$D2,$F9,$5F,$02,$46,$5E
	.dc.b	$47,$F7,$3F,$8B,$FE,$CF,$7C,$FB,$BE,$09,$FD,$FE,$33,$3F,$E3,$B6
	.dc.b	$AC,$13,$C2,$3A,$74,$F1,$30,$FC,$58,$C3,$A3,$FC,$AE,$2B,$FC,$2E
	.dc.b	$FF,$8F,$8C,$4E,$9A,$70,$8D,$BF,$E4,$46,$BE,$78,$E1,$18,$8E,$27
	.dc.b	$F9,$99,$C1,$23,$67,$9A,$3F,$C2,$11,$F0,$29,$E0,$67,$1F,$F1,$5C
	.dc.b	$FC,$43,$51,$9C,$7C,$98,$5C,$3F,$CC,$98,$72,$7F,$17,$E1,$FE,$43
	.dc.b	$C3,$91,$82,$1F,$8D,$72,$FE,$3E,$3F,$2E,$63,$2F,$C6,$CE,$27,$8C
	.dc.b	$61,$FC,$CC,$65,$E4,$60,$86,$09,$1C,$5F,$11,$D3,$0F,$C9,$78,$FF
	.dc.b	$A8,$24,$7E,$9B,$C7,$AF,$E9,$62,$D7,$E8,$A6,$D7,$2A,$FD,$6D,$7E
	.dc.b	$CA,$73,$1F,$22,$2B,$F9,$18,$B8,$A6,$C6,$1B,$3F,$9F,$F6,$7F,$7B
	.dc.b	$E5,$C8,$C1,$0F,$DD,$39,$7F,$03,$1F,$A3,$31,$8F,$EE,$A7,$13,$C4
	.dc.b	$FC,$C6,$CA,$FC,$89,$23,$F3,$0D,$7F,$24,$D7,$8B,$5F,$CE,$35,$FD
	.dc.b	$93,$E0,$7C,$A7,$F1,$4F,$37,$17,$F9,$49,$B9,$FF,$07,$9B,$5C,$8F
	.dc.b	$81,$5F,$C5,$23,$FC,$FD,$F1,$8F,$4A,$92,$BF,$CF,$50,$F4,$B1,$1F
	.dc.b	$E7,$58,$E0,$EA,$79,$CE,$AF,$FD,$23,$5A,$AE,$15,$A7,$81,$0C,$9E
	.dc.b	$0A,$E5,$87,$FB,$0A,$B9,$5F,$4B,$E5,$7C,$09,$19,$79,$1E,$AF,$5B
	.dc.b	$DF,$EB,$48,$C1,$0F,$63,$97,$DD,$1F,$18,$C4,$FB,$27,$13,$C6,$30
	.dc.b	$FC,$88,$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$1F,$C7,$FD,$6F,$AF
	.dc.b	$EC,$C9,$FE,$79,$1F,$CE,$4C,$3F,$56,$30,$E8,$FB,$6E,$AB,$F8,$DC
	.dc.b	$EA,$7F,$93,$3F,$CA,$88,$FF,$31,$6B,$A3,$04,$30,$43,$F4,$AF,$08
	.dc.b	$EF,$1E,$33,$F1,$5A,$F8,$63,$31,$FC,$CA,$7F,$2B,$39,$BF,$85,$3C
	.dc.b	$6B,$F1,$B3,$F9,$AF,$76,$4F,$E6,$1F,$CE,$FB,$B8,$8F,$CC,$45,$AF
	.dc.b	$CA,$CD,$BF,$C7,$4D,$A0,$D5,$59,$8C,$4D,$9C,$C7,$84,$87,$80,$FD
	.dc.b	$5C,$EE,$78,$3F,$B2,$E7,$FD,$9F,$95,$2B,$F6,$31,$97,$F6,$71,$97
	.dc.b	$F6,$51,$97,$F6,$91,$9C,$7F,$63,$39,$7F,$67,$19,$C7,$F5,$93,$97
	.dc.b	$F5,$D1,$9A,$F8,$38,$FF,$80,$89,$19,$8F,$89,$5F,$19,$FF,$7C,$2E
	.dc.b	$3F,$E1,$B7,$27,$E5,$E0,$8E,$15,$B9,$CC,$62,$B8,$47,$F0,$F8,$ED
	.dc.b	$5C,$63,$EA,$D7,$17,$C3,$5C,$27,$F8,$8D,$70,$3A,$4F,$63,$8C,$FF
	.dc.b	$83,$EB,$94,$7F,$84,$DA,$F0,$3F,$C2,$C1,$CB,$C8,$FF,$0E,$B8,$75
	.dc.b	$1D,$48,$E9,$AF,$48,$F5,$57,$B6,$38,$9E,$72,$87,$5F,$1A,$F8,$3F
	.dc.b	$74,$F8,$CF,$86,$71,$AE,$B8,$7D,$7A,$E1,$39,$C3,$8B,$DB,$0F,$15
	.dc.b	$F0,$8F,$D1,$D7,$07,$C5,$7C,$FE,$17,$8C,$9A,$63,$C3,$B5,$EB,$9D
	.dc.b	$36,$74,$D2,$1D,$8F,$0B,$A7,$02,$39,$B8,$6E,$1F,$1A,$78,$36,$7A
	.dc.b	$B5,$E1,$5E,$67,$94,$D9,$F9,$F3,$EE,$F1,$FD,$FE,$44,$BF,$64,$65
	.dc.b	$F7,$46,$5F,$24,$65,$F3,$46,$6B,$ED,$CE,$5F,$BB,$19,$C7,$EC,$4E
	.dc.b	$5F,$B5,$19,$AF,$AB,$39,$7E,$BC,$65,$F8,$D1,$F7,$B2,$77,$7D,$8F
	.dc.b	$1E,$57,$E1,$E3,$46,$1F,$99,$18,$74,$7D,$B7,$58,$FE,$22,$75,$3F
	.dc.b	$9D,$7F,$5C,$47,$F4,$ED,$74,$60,$86,$08,$7E,$D5,$E1,$1D,$E3,$C6
	.dc.b	$7E,$07,$E2,$3C,$31,$98,$FE,$2B,$3F,$96,$9C,$E5,$FC,$4A,$2C,$FE
	.dc.b	$7A,$7F,$37,$F1,$32,$3F,$30,$FE,$7B,$C9,$C4,$7E,$3E,$2E,$1F,$B4
	.dc.b	$9B,$87,$EB,$E6,$D7,$EA,$A6,$BF,$99,$8B,$BA,$6C,$61,$B3,$E2,$3F
	.dc.b	$54,$FE,$C3,$C7,$FD,$8F,$91,$27,$F5,$D1,$97,$F6,$11,$97,$F4,$D1
	.dc.b	$97,$F5,$11,$9C,$7F,$86,$CE,$5F,$E1,$F1,$9C,$7F,$85,$4E,$5F,$E1
	.dc.b	$91,$9A,$FE,$0F,$39,$7F,$84,$C6,$5F,$DC,$47,$E4,$FD,$3F,$E2,$3C
	.dc.b	$FF,$EE,$F9,$FE,$40,$D7,$FA,$6C,$6D,$CF,$93,$C0,$C3,$A3,$24,$30
	.dc.b	$43,$F9,$F4,$7F,$41,$23,$FC,$E9,$AF,$F3,$56,$BC,$0D,$10,$D1,$71
	.dc.b	$FE,$53,$1E,$47,$F9,$E5,$23,$18,$FF,$62,$6D,$7F,$84,$27,$4F,$F3
	.dc.b	$98,$CB,$FD,$45,$F8,$E7,$C8,$F5,$7E,$4F,$93,$27,$E4,$C6,$5F,$97
	.dc.b	$19,$7F,$94,$D2,$75,$18,$9D,$61,$E8,$9C,$BE,$A8,$CE,$1F,$E0,$1B
	.dc.b	$E5,$7E,$03,$B9,$0E,$F7,$CA,$78,$57,$29,$F8,$13,$F8,$2E,$09,$1F
	.dc.b	$82,$7F,$0B,$E4,$62,$3D,$31,$71,$F9,$B3,$70,$F2,$4D,$AF,$99,$35
	.dc.b	$FF,$1D,$C5,$C5,$36,$30,$D9,$F1,$1F,$94,$7F,$B3,$FC,$73,$D4,$FC
	.dc.b	$CF,$93,$2B,$FB,$C4,$E5,$B3,$97,$E8,$CF,$28,$91,$FA,$08,$D6,$5F
	.dc.b	$E0,$3A,$E5,$3F,$D8,$1A,$FF,$8C,$43,$87,$FC,$68,$D7,$FC,$B6,$D7
	.dc.b	$75,$FE,$1C,$72,$D7,$26,$63,$81,$23,$4D,$8D,$47,$A9,$F3,$3F,$BC
	.dc.b	$F6,$B1,$5E,$B8,$B9,$FE,$26,$6D,$78,$8F,$8E,$7F,$02,$7F,$82,$FB
	.dc.b	$9F,$C1,$BB,$64,$FF,$06,$8C,$BF,$C2,$23,$2F,$E7,$A3,$2F,$E8,$23
	.dc.b	$38,$7E,$D2,$72,$FE,$DA,$33,$7F,$9D,$9C,$BF,$9F,$8C,$D7,$ED,$E7
	.dc.b	$2F,$EE,$63,$2F,$FA,$22,$3F,$C5,$BF,$1F,$FE,$DA,$AF,$F3,$43,$67
	.dc.b	$FD,$1E,$D4,$FF,$BC,$5F,$F2,$13,$92,$79,$1D,$3C,$9E,$AC,$9D,$B9
	.dc.b	$69,$0F,$F1,$4C,$F3,$9E,$11,$89,$B8,$FF,$94,$EB,$C5,$39,$39,$46
	.dc.b	$67,$B4,$7E,$57,$AF,$DF,$E4,$48,$FD,$1C,$7E,$A0,$C3,$E2,$8C,$3A
	.dc.b	$3D,$EE,$AB,$F4,$53,$A9,$FF,$39,$EC,$8F,$C9,$92,$3F,$30,$D7,$6D
	.dc.b	$7F,$51,$1C,$5D,$90,$D9,$67,$C2,$F9,$4F,$93,$F0,$3E,$38,$CC,$7C
	.dc.b	$13,$EB,$C6,$73,$FD,$9D,$D9,$FB,$F3,$F8,$4F,$83,$FC,$2F,$95,$27
	.dc.b	$EF,$C6,$5F,$C0,$C6,$5F,$B3,$19,$7E,$DC,$67,$0F,$B1,$39,$7E,$D4
	.dc.b	$67,$0F,$C8,$CE,$5F,$C9,$C6,$6B,$F2,$73,$97,$F2,$B1,$97,$F2,$B1
	.dc.b	$F9,$CF,$22,$47,$E5,$9F,$CF,$F3,$62,$3F,$1F,$17,$0F,$C6,$4D,$C3
	.dc.b	$F4,$73,$6B,$F4,$73,$5F,$D1,$C5,$DB,$FD,$DF,$1D,$0F,$01,$0F,$CE
	.dc.b	$CE,$E7,$83,$FC,$E3,$C5,$FD,$9F,$8F,$2B,$F9,$BC,$65,$FD,$7C,$65
	.dc.b	$FE,$61,$19,$7F,$53,$19,$C7,$F9,$B4,$E6,$7F,$1A,$FF,$26,$23,$FC
	.dc.b	$AD,$AF,$E6,$9A,$F0,$34,$43,$45,$8E,$EF,$A0,$FE,$0B,$F0,$FE,$58
	.dc.b	$CC,$7A,$0C,$9B,$30,$47,$F7,$D1,$FE,$C2,$F9,$07,$B3,$FC,$63,$49
	.dc.b	$7F,$ED,$39,$CB,$67,$2F,$F0,$C9,$E5,$12,$3F,$8C,$C6,$AF,$FD,$BD
	.dc.b	$5C,$A7,$FD,$76,$D8,$FE,$21,$3B,$13,$27,$51,$88,$46,$1F,$E5,$71
	.dc.b	$C5,$E1,$5E,$37,$CC,$79,$0F,$51,$ED,$C3,$E1,$11,$F8,$1E,$4D,$7C
	.dc.b	$F1,$70,$F5,$CD,$BF,$26,$1F,$ED,$26,$D7,$CD,$9D,$5F,$AE,$2B,$EA
	.dc.b	$8B,$5C,$63,$47,$88,$F3,$10,$F1,$99,$3F,$DC,$BF,$24,$CB,$EE,$79
	.dc.b	$52,$7E,$5E,$4C,$BE,$55,$EE,$73,$18,$BE,$11,$F5,$AF,$E8,$5F,$18
	.dc.b	$FA,$B7,$C5,$F0,$D7,$09,$FB,$F5,$C2,$FA,$57,$63,$8D,$7F,$84,$6F
	.dc.b	$94,$7F,$87,$5A,$F0,$3F,$C4,$01,$CB,$C8,$FF,$0E,$38,$75,$1D,$48
	.dc.b	$E9,$AF,$48,$F5,$5F,$B6,$38,$9E,$72,$9F,$98,$F8,$DF,$63,$E1,$3E
	.dc.b	$09,$B9,$E7,$7F,$84,$9D,$CF,$2A,$ED,$5A,$8E,$51,$FA,$69,$FD,$A7
	.dc.b	$B1,$FB,$5E,$59,$5F,$B3,$8C,$BF,$B5,$8C,$BF,$AC,$8C,$BF,$AE,$8C
	.dc.b	$D7,$EC,$A7,$2F,$ED,$23,$38,$7E,$6A,$72,$FE,$72,$33,$5F,$B3,$9C
	.dc.b	$BF,$B5,$8C,$BF,$B4,$8F,$E1,$9E,$44,$8F,$E1,$4F,$F0,$FF,$23,$11
	.dc.b	$FC,$0A,$2E,$1F,$C1,$A6,$E1,$FA,$C9,$B5,$FC,$0E,$6B,$FC,$0A,$2E
	.dc.b	$BF,$EE,$B8,$EE,$43,$B1,$FC,$6E,$77,$3C,$1F,$AB,$9B,$E2,$3F,$92
	.dc.b	$79,$32,$FE,$E4,$68,$C3,$FC,$6A,$30,$E8,$FE,$50,$EB,$2F,$BF,$3A
	.dc.b	$9F,$C8,$3F,$A1,$23,$F9,$A6,$BA,$30,$43,$04,$3F,$50,$F0,$8E,$F1
	.dc.b	$E3,$3F,$97,$9E,$17,$2F,$F0,$88,$FE,$84,$E7,$2F,$E0,$B1,$63,$F9
	.dc.b	$91,$F9,$AF,$92,$7B,$37,$CA,$93,$FA,$79,$CB,$67,$2B,$0F,$D7,$4F
	.dc.b	$FB,$FD,$B3,$D4,$DB,$AA,$C5,$73,$3B,$3A,$C3,$FA,$25,$F8,$27,$B4
	.dc.b	$62,$F5,$1F,$D3,$27,$C1,$39,$8E,$87,$03,$55,$C1,$4F,$5B,$F4,$F9
	.dc.b	$1F,$0C,$93,$F0,$7C,$9A,$F3,$38,$C6,$27,$E0,$CD,$C3,$D3,$36,$BC
	.dc.b	$D3,$5E,$27,$CD,$8F,$A5,$19,$BF,$F7,$9C,$F7,$6D,$F6,$8E,$04,$35
	.dc.b	$5C,$9F,$E3,$1E,$07,$DD,$F2,$24,$60,$87,$57,$2F,$F5,$08,$FD,$49
	.dc.b	$8A,$FB,$B3,$89,$E2,$7F,$3D,$9F,$E8,$04,$7F,$9C,$B5,$FE,$68,$D7
	.dc.b	$81,$A2,$1A,$2C,$FF,$2C,$8F,$04,$7C,$3F,$83,$B3,$98,$F5,$BF,$92
	.dc.b	$8C,$E5,$FD,$92,$AD,$7E,$1E,$7F,$41,$EC,$7E,$87,$CA,$93,$F9,$F8
	.dc.b	$CB,$FA,$18,$CB,$F4,$E3,$2F,$D5,$8C,$E1,$FA,$29,$CB,$FA,$48,$CE
	.dc.b	$1F,$9C,$9C,$BF,$9E,$8C,$D7,$E6,$E7,$2F,$E7,$63,$2F,$E6,$23,$F7
	.dc.b	$1E,$3C,$8F,$DB,$3F,$BB,$F2,$31,$1F,$89,$8B,$87,$ED,$A6,$E1,$FA
	.dc.b	$E9,$B5,$FA,$B9,$AF,$E9,$A2,$EA,$9B,$18,$6C,$F8,$8F,$E1,$0F,$F6
	.dc.b	$6F,$93,$4F,$F6,$8F,$85,$FF,$69,$E3,$FD,$D5,$B8,$FF,$90,$EB,$78
	.dc.b	$7F,$77,$6B,$FB,$19,$DB,$FE,$4C,$6A,$8D,$D7,$F8,$05,$B5,$FE,$4F
	.dc.b	$AD,$9C,$23,$35,$FE,$56,$7B,$BF,$E5,$53,$A5,$FF,$86,$1D,$35,$C9
	.dc.b	$98,$E2,$FA,$E3,$51,$88,$E2,$6E,$32,$47,$04,$3F,$88,$47,$99,$EE
	.dc.b	$7C,$3F,$89,$F9,$D3,$23,$E4,$3F,$3E,$31,$1C,$A2,$E1,$F4,$67,$EA
	.dc.b	$7B,$3F,$57,$79,$7F,$E2,$7C,$53,$A8,$C4,$E9,$FA,$91,$97,$EB,$46
	.dc.b	$6F,$EF,$4E,$5F,$C0,$46,$72,$FA,$93,$97,$EB,$46,$6B,$EF,$4E,$5F
	.dc.b	$C0,$46,$55,$F9,$9F,$47,$8F,$F8,$F2,$3E,$BF,$22,$BF,$8B,$8B,$8F
	.dc.b	$E4,$E6,$E3,$F8,$79,$B5,$F8,$F9,$AF,$82,$2D,$A6,$C6,$1B,$3E,$43
	.dc.b	$C4,$7E,$77,$87,$F4,$9E,$34,$8C,$10,$FD,$03,$97,$F0,$71,$FA,$D3
	.dc.b	$18,$FE,$AE,$71,$3C,$4F,$CD,$6C,$AF,$C9,$92,$3F,$30,$D7,$FC,$EB
	.dc.b	$1B,$72,$43,$31,$B7,$AB,$C6,$BF,$60,$F8,$A3,$F7,$1F,$17,$67,$31
	.dc.b	$ED,$7F,$81,$46,$73,$FE,$05,$56,$BF,$7D,$3F,$C2,$7D,$59,$1F,$95
	.dc.b	$7F,$86,$79,$18,$8F,$E0,$D1,$70,$FE,$1D,$37,$0F,$E0,$53,$6B,$F8
	.dc.b	$54,$D7,$F1,$71,$75,$4D,$8C,$36,$7C,$47,$E9,$DF,$F4,$2F,$CA,$7F
	.dc.b	$B6,$E3,$F2,$B7,$FC,$84,$CF,$93,$23,$F5,$D1,$97,$FD,$2D,$3C,$4E
	.dc.b	$58,$7F,$A6,$F1,$FE,$51,$5C,$6B,$FC,$29,$5A,$8E,$51,$8B,$D5,$70
	.dc.b	$8E,$E7,$68,$CC,$FF,$25,$38,$CF,$FA,$A6,$BD,$5A,$3E,$1F,$DB,$E4
	.dc.b	$CC,$7F,$9B,$47,$F3,$F3,$0F,$92,$30,$E8,$F6,$BA,$BF,$E6,$D3,$A9
	.dc.b	$FC,$8B,$FC,$A0,$8F,$F2,$D6,$BA,$30,$43,$04,$3F,$42,$F0,$8E,$F1
	.dc.b	$E3,$3E,$4F,$C1,$78,$63,$31,$FC,$C2,$7E,$DC,$E7,$1F,$EA,$D1,$64
	.dc.b	$31,$57,$E0,$BD,$99,$3F,$71,$FC,$2F,$97,$88,$FB,$31,$71,$FA,$13
	.dc.b	$71,$EB,$5D,$6B,$0F,$83,$05,$62,$34,$77,$3F,$16,$E6,$13,$63,$0D
	.dc.b	$9F,$C8,$F9,$1F,$9A,$F2,$64,$60,$87,$E4,$1C,$BF,$92,$8F,$D0,$98
	.dc.b	$C7,$F1,$53,$89,$E3,$18,$7F,$3F,$19,$79,$18,$21,$82,$47,$17,$C4
	.dc.b	$74,$C7,$E3,$FC,$13,$C6,$FD,$9F,$91,$2B,$EB,$DA,$F7,$5F,$EF,$CB
	.dc.b	$E1,$5F,$B2,$BE,$0F,$FA,$06,$32,$D9,$CD,$FE,$9E,$B9,$44,$9F,$F0
	.dc.b	$45,$70,$7F,$C1,$8D,$7C,$26,$A3,$FC,$19,$19,$8F,$C6,$B5,$34,$91
	.dc.b	$C0,$91,$CC,$F7,$DF,$67,$D0,$7F,$0E,$9F,$33,$82,$5F,$F0,$18,$CB
	.dc.b	$F1,$E3,$C7,$58,$AF,$CA,$E9,$E4,$9F,$E3,$3E,$C7,$F2,$22,$3F,$C7
	.dc.b	$2A,$BF,$C6,$22,$DF,$F0,$F9,$B7,$D2,$6D,$7E,$7E,$6B,$FA,$F8,$B7
	.dc.b	$F5,$2F,$F8,$C4,$60,$FE,$67,$56,$BC,$A7,$8C,$F8,$A7,$EE,$BE,$57
	.dc.b	$FD,$89,$87,$EC,$1F,$F6,$6C,$67,$C7,$FD,$9E,$1F,$0E,$13,$C1,$E5
	.dc.b	$E4,$6A,$F9,$5D,$9E,$47,$80,$E3,$39,$69,$9A,$F0,$BD,$A7,$51,$DE
	.dc.b	$36,$7F,$B5,$5D,$92,$76,$70,$99,$3D,$63,$2F,$78,$CB,$CE,$3F,$49
	.dc.b	$B7,$83,$C8,$93,$E9,$D5,$3A,$8F,$44,$E9,$FD,$44,$65,$F3,$46,$72
	.dc.b	$F9,$55,$CA,$7A,$57,$E8,$88,$FE,$A9,$AF,$F9,$D4,$FD,$1B,$87,$F5
	.dc.b	$AD,$7F,$81,$B5,$FD,$5B,$CA,$7C,$2F,$95,$FB,$98,$BE,$0F,$19,$F2
	.dc.b	$BF,$29,$F5,$63,$2F,$43,$27,$D3,$7E,$BF,$B3,$23,$EE,$3F,$6B,$C9
	.dc.b	$C4,$7D,$98,$B8,$78,$A6,$DF,$E4,$E6,$D7,$E5,$E6,$BF,$92,$8B,$82
	.dc.b	$6C,$61,$B3,$E2,$3F,$1C,$FE,$33,$C8,$FC,$6F,$93,$27,$F2,$51,$97
	.dc.b	$F2,$91,$97,$F2,$11,$97,$F2,$51,$9C,$7E,$5C,$E5,$F9,$B1,$9C,$7F
	.dc.b	$5D,$39,$7F,$61,$19,$AF,$DA,$4E,$5F,$DB,$46,$5F,$D8,$C7,$F3,$9E
	.dc.b	$7F,$D2,$F9,$12,$7F,$2D,$19,$7F,$59,$19,$7F,$9E,$46,$5F,$D5,$46
	.dc.b	$70,$FC,$AC,$E6,$7F,$9C,$BF,$CD,$88,$FF,$3E,$6B,$F9,$5D,$BF,$C0
	.dc.b	$6E,$48,$66,$36,$F4,$33,$3E,$07,$D2,$7F,$03,$F8,$1F,$34,$66,$3D
	.dc.b	$26,$4D,$98,$23,$F9,$C8,$FE,$1D,$F2,$CF,$62,$92,$BF,$8A,$47,$FB
	.dc.b	$B0,$D3,$88,$D2,$AF,$E4,$53,$B9,$CC,$61,$EF,$EC,$E3,$5F,$F0,$DD
	.dc.b	$74,$BE,$51,$8A,$DC,$77,$21,$A3,$FC,$5F,$3C,$6F,$C3,$7D,$67,$35
	.dc.b	$55,$7A,$FC,$AF,$83,$E5,$49,$F7,$C7,$C4,$30,$F7,$38,$C6,$5D,$9F
	.dc.b	$19,$DE,$1F,$0A,$75,$3C,$69,$5A,$C7,$E0,$D6,$AF,$E4,$CE,$6F,$DB
	.dc.b	$39,$AF,$8F,$39,$9F,$5C,$FF,$47,$F9,$77,$C8,$FD,$CF,$2E,$57,$F4
	.dc.b	$D9,$CB,$67,$2F,$96,$79,$44,$8F,$C0,$C6,$B1,$FE,$BD,$5C,$A7,$E2
	.dc.b	$35,$FF,$2A,$87,$0F,$F9,$61,$AF,$F9,$99,$AE,$EB,$F9,$D3,$96,$B9
	.dc.b	$33,$1C,$09,$1A,$6C,$6A,$3D,$4F,$99,$FC,$FF,$C1,$E2,$BD,$71,$72
	.dc.b	$FC,$9C,$DA,$F1,$1F,$1C,$FD,$1B,$FA,$8F,$66,$47,$E9,$9F,$D5,$F9
	.dc.b	$B8,$8F,$C0,$45,$C3,$F4,$B3,$6F,$F4,$33,$6B,$F3,$73,$5F,$CB,$C5
	.dc.b	$D5,$36,$30,$D9,$F1,$1F,$B8,$7F,$7B,$E5,$FE,$FB,$CB,$93,$FB,$C8
	.dc.b	$CB,$FB,$E8,$CB,$F9,$78,$CB,$F9,$98,$CE,$3F,$BE,$9C,$BF,$C0,$23
	.dc.b	$38,$FE,$F2,$72,$FE,$FA,$33,$5F,$B8,$9C,$BF,$BA,$8C,$BF,$B6,$8F
	.dc.b	$CE,$5F,$F1,$6A,$93,$F9,$B8,$CB,$FC,$5A,$32,$FE,$52,$32,$FE,$C6
	.dc.b	$33,$7F,$9A,$9C,$CF,$E6,$34,$F9,$AF,$E3,$A3,$F1,$84,$70,$43,$FC
	.dc.b	$21,$1C,$1B,$1B,$69,$B3,$B3,$E4,$9F,$14,$7F,$90,$7E,$05,$F0,$59
	.dc.b	$C3,$F5,$17,$6B,$FD,$9B,$5A,$7E,$11,$0F,$80,$73,$AF,$F2,$54,$3D
	.dc.b	$BA,$49,$F0,$79,$75,$FE,$3D,$3B,$7C,$93,$CA,$75,$8F,$F9,$7E,$B5
	.dc.b	$5C,$2B,$4F,$02,$19,$3B,$98,$9E,$51,$D3,$1F,$F6,$ED,$2F,$17,$F0
	.dc.b	$2A,$A9,$D3,$F2,$7C,$9F,$7F,$95,$23,$04,$3D,$AE,$5F,$7C,$7C,$63
	.dc.b	$18,$FB,$67,$13,$C6,$30,$FD,$78,$CB,$C8,$C1,$0C,$12,$38,$BE,$23
	.dc.b	$A6,$3F,$C9,$79,$3E,$BF,$95,$27,$F9,$C4,$7F,$3D,$30,$FD,$68,$C3
	.dc.b	$A3,$F1,$4E,$AB,$F9,$BC,$EA,$7F,$2E,$FF,$27,$23,$FC,$B1,$AE,$8C
	.dc.b	$10,$C1,$0F,$D4,$3C,$23,$BC,$78,$CF,$C5,$E3,$E1,$8C,$C7,$F4,$89
	.dc.b	$FC,$AC,$E6,$FE,$14,$F1,$AF,$C7,$4F,$E6,$BD,$89,$3F,$A7,$7F,$3B
	.dc.b	$E5,$E2,$3F,$33,$16,$BF,$2D,$36,$FF,$1B,$36,$BF,$57,$35,$FC,$64
	.dc.b	$5B,$4D,$8C,$36,$7C,$47,$EE,$5F,$D7,$F9,$5F,$B0,$F2,$A4,$FE,$F6
	.dc.b	$32,$FE,$FE,$32,$FE,$C2,$32,$FE,$CA,$33,$8F,$EB,$E7,$2F,$EC,$63
	.dc.b	$38,$FE,$A2,$72,$FE,$AA,$33,$5F,$C2,$27,$2F,$F0,$A8,$CB,$FA,$B8
	.dc.b	$FF,$68,$7C,$C3,$BB,$FD,$A5,$F3,$3F,$ED,$03,$FD,$A3,$19,$AF,$F6
	.dc.b	$C3,$5F,$F1,$34,$6F,$CC,$E7,$52,$7F,$DB,$D3,$97,$8C,$65,$FF,$5D
	.dc.b	$57,$37,$FC,$16,$D4,$47,$0D,$8F,$F0,$3C,$62,$0D,$24,$68,$91,$FE
	.dc.b	$10,$8C,$BE,$42,$1E,$CF,$83,$F3,$67,$FE,$FF,$9F,$80,$DA,$F8,$F3
	.dc.b	$AC,$FC,$F7,$E2,$9C,$3E,$AC,$15,$F3,$BD,$99,$3F,$19,$FA,$3E,$6E
	.dc.b	$23,$E6,$C5,$C5,$8F,$6C,$17,$98,$C5,$5B,$F0,$4D,$54,$9C,$3C,$A3
	.dc.b	$C4,$43,$C2,$7D,$A9,$DC,$F0,$7F,$0F,$E5,$7E,$23,$CA,$95,$F8,$08
	.dc.b	$CB,$F8,$28,$CB,$F8,$88,$CB,$F8,$A8,$CE,$3E,$E8,$F9,$13,$A8,$C4
	.dc.b	$7A,$A7,$78,$FC,$3A,$D3,$F1,$63,$33,$C8,$F2,$1E,$32,$1E,$33,$F3
	.dc.b	$BE,$9F,$E8,$BC,$B9,$18,$21,$FA,$07,$2F,$FA,$1F,$64,$FE,$B0,$CE
	.dc.b	$3E,$8D,$17,$9B,$E6,$7E,$82,$BF,$36,$47,$F2,$CD,$7F,$4E,$D7,$81
	.dc.b	$A2,$1A,$2D,$7E,$66,$3C,$11,$FB,$8F,$82,$B3,$98,$F6,$BE,$78,$CE
	.dc.b	$5F,$9C,$C7,$DB,$5A,$8F,$C9,$4F,$F0,$8F,$66,$4F,$F0,$07,$F8,$5F
	.dc.b	$99,$88,$FE,$0F,$17,$0F,$89,$36,$FF,$39,$36,$BF,$51,$35,$FC,$CC
	.dc.b	$5C,$D3,$63,$0D,$9F,$11,$FC,$49,$FD,$AD,$FE,$2B,$E6,$BF,$D9,$58
	.dc.b	$7E,$E1,$79,$7A,$8C,$36,$FF,$89,$CF,$06,$F9,$79,$51,$C9,$FE,$33
	.dc.b	$7D,$EB,$53,$C6,$77,$58,$99,$1C,$2F,$C9,$3C,$0C,$C7,$FB,$56,$BB
	.dc.b	$36,$34,$F1,$3C,$F5,$C2,$FE,$04,$EA,$79,$4E,$6F,$BC,$7F,$3E,$E1
	.dc.b	$F3,$F9,$72,$30,$43,$C4,$E5,$FD,$14,$7F,$4C,$31,$87,$CF,$9C,$4F
	.dc.b	$13,$F9,$FC,$FF,$3A,$23,$FC,$D5,$AF,$F4,$66,$BC,$0D,$10,$D1,$67
	.dc.b	$F9,$C4,$78,$23,$E4,$7C,$15,$9C,$C7,$AD,$FC,$24,$66,$FF,$41,$56
	.dc.b	$BF,$03,$3F,$6B,$DA,$91,$F3,$CF,$C0,$BF,$7B,$CD,$CC,$7D,$78,$B8
	.dc.b	$7E,$46,$6E,$1F,$91,$9B,$5F,$8E,$9A,$F4,$3B,$68,$9E,$D1,$87,$71
	.dc.b	$98,$DC,$E2,$72,$FE,$4B,$CA,$FC,$9F,$8F,$2B,$AC,$65,$EF,$19,$7F
	.dc.b	$21,$19,$7F,$25,$19,$C7,$F5,$93,$97,$F5,$D1,$9C,$7F,$5D,$39,$7F
	.dc.b	$61,$19,$AF,$D6,$CE,$5F,$D7,$C6,$5F,$D6,$C7,$EB,$FE,$6D,$DB,$FA
	.dc.b	$CF,$2E,$57,$D3,$9C,$B6,$72,$FE,$A6,$79,$44,$8F,$9B,$1A,$C7,$FA
	.dc.b	$7D,$72,$94,$62,$BF,$BF,$C4,$9D,$36,$32,$FF,$81,$23,$2D,$8C,$BF
	.dc.b	$E1,$E8,$CB,$63,$2A,$32,$D8,$CC,$60,$8E,$0F,$64,$7A,$31,$F0,$3F
	.dc.b	$8A,$F8,$2C,$DF,$AE,$3E,$31,$92,$9C,$4F,$21,$87,$F9,$54,$7F,$21
	.dc.b	$F9,$C7,$B7,$49,$5F,$98,$BA,$F3,$13,$C1,$FD,$53,$5F,$E3,$D1,$C2
	.dc.b	$FF,$C0,$33,$96,$D7,$F2,$C9,$D1,$FE,$32,$9D,$36,$35,$5D,$9D,$BF
	.dc.b	$E4,$E9,$DB,$FC,$FA,$72,$F9,$A6,$EF,$FE,$12,$BF,$1F,$95,$F0,$7C
	.dc.b	$B9,$7E,$A8,$F8,$86,$1F,$81,$8F,$FB,$69,$B1,$A7,$E0,$CE,$EB,$AC
	.dc.b	$73,$32,$43,$89,$93,$AC,$71,$C7,$E1,$57,$59,$C9,$C6,$BA,$BF,$4F
	.dc.b	$4F,$F7,$73,$F4,$BC,$AF,$B8,$7E,$BF,$C8,$91,$92,$1C,$5C,$BE,$E8
	.dc.b	$E2,$FE,$CE,$32,$FE,$9D,$FC,$C9,$1F,$D1,$CF,$89,$FD,$4B,$5E,$06
	.dc.b	$A3,$83,$A6,$BA,$30,$78,$48,$7D,$2F,$83,$F3,$DF,$EF,$A6,$E3,$F8
	.dc.b	$99,$B5,$FB,$59,$AF,$E6,$23,$F4,$1E,$C4,$8F,$CC,$BF,$A3,$F3,$31
	.dc.b	$1F,$80,$8B,$87,$E6,$A6,$E1,$F9,$79,$B5,$F9,$39,$AF,$E9,$A2,$EA
	.dc.b	$9B,$18,$6C,$F8,$8F,$D5,$BF,$B5,$F2,$FF,$6D,$E5,$C9,$FD,$7C,$65
	.dc.b	$FD,$8C,$65,$FC,$BC,$65,$FC,$CC,$67,$1F,$D9,$CE,$5F,$DA,$C6,$71
	.dc.b	$FD,$94,$E5,$FD,$A4,$66,$BF,$61,$39,$7F,$65,$19,$7F,$82,$C7,$F3
	.dc.b	$FF,$98,$FE,$1D,$E6,$49,$FE,$7D,$19,$7F,$84,$C6,$5F,$CE,$46,$5F
	.dc.b	$D8,$C6,$70,$FE,$7B,$39,$9F,$E7,$4F,$F2,$C2,$3F,$94,$6B,$F9,$46
	.dc.b	$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$8F,$9E,$57,$2F,$A0,$C9,$B3,$04
	.dc.b	$7F,$91,$BF,$E5,$1E,$77,$B5,$49,$1F,$E1,$58,$79,$B8,$8F,$E6,$51
	.dc.b	$C1,$D4,$F3,$9D,$57,$FA,$C2,$B5,$5C,$2B,$4F,$02,$19,$3C,$15,$CB
	.dc.b	$3F,$F7,$2D,$72,$BE,$97,$CA,$F8,$12,$32,$F2,$3D,$5E,$5F,$BF,$BE
	.dc.b	$46,$08,$7B,$1C,$BE,$E8,$F8,$C6,$27,$D9,$38,$9E,$31,$87,$E4,$46
	.dc.b	$5E,$46,$08,$60,$91,$C5,$F1,$1D,$30,$33,$E4,$FF,$A5,$47,$C4,$F2
	.dc.b	$FE,$D1,$FE,$8B,$8C,$CE,$E7,$2F,$DB,$9D,$44,$8C,$10,$FE,$DD,$3C
	.dc.b	$93,$FE,$EA,$7F,$C3,$2D,$7C,$A6,$A3,$FC,$23,$19,$8F,$11,$FE,$18
	.dc.b	$70,$EC,$C4,$70,$23,$82,$1F,$E2,$48,$F2,$12,$3F,$35,$F0,$1E,$B9
	.dc.b	$F4,$C5,$BF,$C9,$4D,$8F,$3D,$71,$AF,$CF,$4F,$E8,$FD,$A9,$1F,$A1
	.dc.b	$7F,$4D,$E6,$E2,$3F,$43,$17,$0F,$CF,$4D,$C3,$F2,$F3,$6B,$EC,$EB
	.dc.b	$FE,$FA,$7A,$10,$E6,$7E,$BE,$B8,$28,$4E,$1E,$51,$C0,$97,$FA,$F9
	.dc.b	$F3,$C7,$EE,$7C,$9F,$DF,$79,$32,$30,$43,$F7,$6E,$5F,$DE,$C7,$F0
	.dc.b	$53,$15,$FB,$B9,$C4,$F1,$8C,$3F,$C1,$E3,$2F,$23,$04,$30,$48,$E2
	.dc.b	$F8,$8E,$98,$FD,$6C,$3F,$DC,$DF,$3B,$FE,$D8,$A4,$FE,$42,$30,$EE
	.dc.b	$BF,$D5,$9A,$7C,$97,$FD,$5D,$3A,$7F,$56,$D7,$F3,$B7,$FA,$93,$BF
	.dc.b	$93,$23,$9C,$65,$FC,$D3,$F9,$E2,$38,$21,$F9,$77,$2D,$32,$7E,$A9
	.dc.b	$CB,$4E,$E4,$3F,$CC,$12,$F8,$2F,$14,$7E,$DA,$6E,$3F,$12,$3D,$B3
	.dc.b	$8C,$7E,$24,$6C,$F8,$71,$F3,$3D,$39,$1F,$19,$F9,$FE,$5E,$23,$C3
	.dc.b	$16,$FB,$43,$15,$67,$0F,$A9,$76,$BE,$7C,$D7,$84,$5A,$E3,$1E,$12
	.dc.b	$1E,$03,$EC,$CE,$E7,$83,$F8,$4F,$2B,$F0,$BE,$54,$AF,$BF,$19,$7F
	.dc.b	$03,$19,$7E,$5C,$65,$F9,$B1,$9C,$7D,$B8,$7F,$B8,$E7,$67,$22,$18
	.dc.b	$24,$70,$9E,$58,$FE,$56,$FB,$4E,$6B,$8B,$DA,$3F,$9C,$79,$7F,$9E
	.dc.b	$F2,$E5,$7E,$6A,$32,$FE,$6A,$32,$FF,$36,$8C,$BF,$86,$8C,$D7,$D6
	.dc.b	$DD,$5A,$AF,$E7,$2F,$F3,$62,$3F,$CD,$5A,$FF,$2F,$6B,$C0,$D1,$0D
	.dc.b	$16,$7C,$11,$E4,$7E,$5F,$C1,$5C,$7F,$98,$CD,$AF,$D9,$CD,$7F,$A7
	.dc.b	$C5,$CD,$3F,$C0,$BD,$8F,$E0,$7E,$6C,$BF,$DE,$C6,$5F,$DF,$C6,$5F
	.dc.b	$D4,$C6,$5F,$D5,$C6,$71,$FB,$F3,$97,$F0,$31,$9C,$7F,$87,$4E,$5F
	.dc.b	$E2,$11,$9A,$FE,$17,$39,$7F,$86,$C6,$5F,$D3,$47,$F6,$6A,$FD,$4F
	.dc.b	$CF,$7F,$B2,$24,$D3,$18,$FF,$71,$71,$E6,$71,$C7,$FB,$7C,$69,$C5
	.dc.b	$7F,$81,$A3,$6F,$F8,$22,$FB,$D7,$F2,$AA,$DC,$F8,$22,$D7,$08,$FE
	.dc.b	$1F,$38,$99,$1C,$E3,$C3,$1C,$E3,$07,$85,$DB,$FE,$F7,$9F,$33,$A1
	.dc.b	$F3,$EF,$97,$F0,$F9,$B2,$75,$39,$6C,$E5,$E9,$3E,$A9,$D4,$31,$FE
	.dc.b	$2B,$18,$7E,$7C,$F1,$73,$5C,$4F,$F0,$CD,$4A,$E6,$D8,$CB,$FE,$1E
	.dc.b	$8C,$B6,$32,$FF,$8D,$A3,$2D,$8C,$BF,$01,$B1,$A8,$C1,$C9,$C1,$23
	.dc.b	$CD,$1F,$08,$F9,$56,$F7,$F3,$1F,$07,$2F,$23,$B3,$E9,$46,$B0,$F8
	.dc.b	$31,$97,$F2,$71,$F8,$2F,$6B,$F0,$7E,$74,$AE,$F1,$97,$C3,$19,$7E
	.dc.b	$F4,$65,$FC,$04,$66,$BF,$3F,$39,$7F,$43,$19,$BF,$CD,$CE,$5F,$CE
	.dc.b	$C6,$6B,$F3,$93,$97,$F3,$D1,$97,$F5,$31,$F9,$AF,$2A,$47,$C6,$7F
	.dc.b	$3B,$E5,$62,$3F,$29,$17,$0F,$D9,$CD,$C3,$F5,$B3,$6B,$F5,$93,$5F
	.dc.b	$DB,$C5,$CB,$FD,$D9,$1D,$C8,$76,$3E,$BC,$EE,$78,$3F,$99,$F0,$3F
	.dc.b	$63,$E5,$4A,$FD,$0C,$65,$FB,$F1,$97,$EF,$1F,$8F,$8D,$3F,$AF,$8D
	.dc.b	$63,$FA,$29,$CC,$FE,$75,$FC,$89,$1F,$C8,$35,$FC,$FE,$DF,$E4,$E7
	.dc.b	$81,$A2,$1A,$8E,$0E,$8F,$29,$FB,$E7,$CB,$3F,$C8,$3E,$02,$E7,$FA
	.dc.b	$B9,$B5,$FC,$66,$6B,$C4,$F8,$67,$E8,$1F,$E2,$FE,$9F,$F1,$8F,$2E
	.dc.b	$4F,$E9,$23,$2F,$E9,$A3,$2F,$F1,$38,$CB,$FC,$56,$33,$87,$F3,$49
	.dc.b	$CB,$FC,$DA,$33,$87,$F3,$09,$CB,$FC,$CA,$33,$5F,$CC,$67,$2F,$F3
	.dc.b	$38,$CB,$FC,$FA,$3F,$98,$FD,$07,$FB,$6E,$BF,$D5,$2D,$9F,$F5,$FB
	.dc.b	$53,$FE,$F2,$7F,$95,$39,$27,$95,$D3,$CB,$EA,$C9,$DB,$96,$90,$FA
	.dc.b	$F3,$CE,$78,$46,$26,$E3,$FE,$40,$AF,$14,$E4,$E5,$19,$9E,$D1,$F8
	.dc.b	$2D,$1E,$5F,$C0,$F2,$E4,$FF,$3E,$8F,$E7,$46,$1F,$1C,$61,$D1,$F0
	.dc.b	$1D,$5F,$F3,$E9,$D4,$FE,$61,$FE,$5A,$47,$F9,$63,$5D,$18,$21,$82
	.dc.b	$1F,$D0,$DE,$11,$DE,$3C,$67,$C8,$F8,$3F,$0C,$66,$3E,$04,$FD,$E9
	.dc.b	$CE,$7F,$D2,$62,$CF,$DE,$9F,$C1,$FB,$7F,$84,$F3,$64,$FD,$E8,$CB
	.dc.b	$F8,$08,$CB,$F6,$23,$2F,$DA,$8C,$E1,$F5,$E7,$2F,$D9,$8C,$E1,$F9
	.dc.b	$09,$CB,$F9,$28,$CD,$7E,$4A,$72,$FE,$52,$32,$FE,$52,$3F,$37,$E6
	.dc.b	$48,$FC,$AB,$F9,$EE,$2C,$47,$E3,$A2,$E1,$F8,$B9,$B8,$7E,$8A,$6D
	.dc.b	$7E,$8A,$6B,$FA,$28,$BB,$7F,$B8,$AF,$F2,$91,$83,$C1,$3E,$22,$1E
	.dc.b	$13,$F5,$35,$FE,$22,$FA,$1F,$F0,$0D,$7E,$77,$CC,$FD,$A9,$9A,$87
	.dc.b	$EC,$1C,$BF,$E1,$B9,$E3,$8A,$24,$61,$FD,$DC,$F0,$7F,$92,$C6,$53
	.dc.b	$FE,$06,$69,$FE,$08,$70,$47,$04,$3F,$C2,$2E,$5A,$64,$FF,$0C,$B9
	.dc.b	$7C,$27,$B0,$FE,$03,$EF,$5B,$FF,$5E,$CF,$94,$F3,$64,$9D,$61,$EF
	.dc.b	$BF,$39,$F0,$CF,$E3,$B1,$FE,$4E,$C5,$EC,$DF,$3A,$57,$F9,$0E,$72
	.dc.b	$D9,$CB,$FD,$2A,$3F,$DE,$C6,$9C,$46,$B2,$FF,$69,$CE,$A3,$C0,$67
	.dc.b	$0F,$34,$66,$BB,$1E,$1B,$D3,$F6,$63,$B5,$F4,$BC,$CA,$79,$3E,$5F
	.dc.b	$33,$DC,$47,$D9,$E6,$57,$47,$BE,$3E,$24,$66,$FD,$31,$71,$F3,$CD
	.dc.b	$AF,$24,$D7,$C5,$16,$D3,$63,$0D,$9F,$CB,$7C,$B7,$D3,$F3,$64,$60
	.dc.b	$87,$CF,$72,$FE,$5A,$3F,$4C,$63,$1F,$9F,$38,$9E,$27,$F9,$AB,$65
	.dc.b	$7E,$64,$91,$F9,$66,$BF,$94,$6B,$C5,$AF,$E3,$63,$9B,$C4,$87,$94
	.dc.b	$F4,$D7,$E0,$BE,$0F,$C9,$19,$24,$65,$DE,$7D,$6F,$73,$99,$FC,$87
	.dc.b	$A1,$F9,$9F,$3A,$47,$E5,$E3,$F3,$86,$1F,$CB,$46,$1D,$1F,$9D,$75
	.dc.b	$3F,$96,$9D,$4F,$03,$47,$03,$C2,$60,$D1,$82,$47,$E8,$27,$AD,$FE
	.dc.b	$7A,$BF,$61,$E5,$4B,$FD,$73,$FB,$3F,$2F,$11,$FA,$D8,$B8,$7E,$AA
	.dc.b	$6D,$FE,$A2,$6D,$7E,$96,$6B,$FA,$38,$BB,$A6,$C6,$1B,$3E,$23,$F8
	.dc.b	$03,$FC,$C7,$C6,$FE,$0D,$72,$7F,$96,$46,$5F,$E0,$D1,$97,$F9,$4C
	.dc.b	$65,$FE,$0D,$19,$C3,$F8,$CC,$E6,$7F,$28,$FF,$2B,$23,$FC,$A1,$AF
	.dc.b	$F2,$66,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$DB,$DF,$44,$F8,$2F,$34
	.dc.b	$EA,$7D,$26,$A3,$37,$F9,$FB,$B3,$FD,$2E,$F9,$3D,$3D,$BF,$F7,$CC
	.dc.b	$3C,$CA,$62,$24,$F8,$A3,$2F,$12,$1F,$E7,$69,$E4,$EC,$CE,$2C,$3F
	.dc.b	$9F,$4A,$F9,$D7,$1B,$EE,$F3,$8C,$DF,$18,$F5,$4E,$A3,$10,$AE,$8F
	.dc.b	$84,$E8,$78,$AF,$C3,$E5,$7C,$02,$47,$CE,$C6,$BE,$B8,$B7,$DA,$D5
	.dc.b	$8B,$E7,$7F,$5A,$FE,$5C,$E6,$3E,$84,$57,$CD,$16,$93,$63,$0D,$9F
	.dc.b	$F0,$27,$9B,$F5,$BC,$99,$18,$21,$F4,$5C,$BF,$E0,$48,$FF,$0A,$18
	.dc.b	$C7,$D5,$38,$9E,$27,$F2,$F9,$FF,$03,$11,$FF,$01,$B5,$FF,$00,$B5
	.dc.b	$E0,$68,$86,$8B,$3F,$E0,$28,$F0,$47,$E1,$FD,$AB,$39,$8F,$5B,$F3
	.dc.b	$E3,$39,$7F,$39,$AB,$5F,$9B,$9F,$CC,$FB,$1F,$9A,$F2,$E4,$FE,$A2
	.dc.b	$32,$FE,$AA,$32,$FE,$5E,$32,$FE,$66,$33,$87,$DF,$9C,$BF,$81,$8C
	.dc.b	$E1,$F9,$A9,$CB,$F9,$C8,$CD,$7E,$BE,$72,$FE,$C6,$32,$FE,$5E,$3F
	.dc.b	$63,$E4,$C8,$FD,$FB,$FB,$4F,$2B,$11,$FA,$D8,$B8,$7E,$4E,$6E,$1F
	.dc.b	$AC,$9B,$5F,$BF,$9A,$FE,$96,$2E,$C9,$B1,$86,$CF,$88,$FE,$22,$FF
	.dc.b	$8D,$6F,$F5,$FF,$46,$F4,$B1,$8F,$F8,$E7,$E9,$3F,$D7,$75,$FB,$46
	.dc.b	$C7,$F8,$F6,$B8,$61,$FE,$5D,$AE,$77,$CA,$3F,$CC,$17,$C1,$BE,$6E
	.dc.b	$55,$F2,$57,$C9,$E9,$87,$8E,$B8,$5F,$39,$E3,$87,$3C,$39,$47,$F9
	.dc.b	$44,$C9,$4F,$F3,$93,$A7,$D5,$19,$72,$D7,$89,$C6,$38,$3A,$38,$38
	.dc.b	$76,$78,$5C,$12,$3C,$C4,$8F,$69,$F1,$BE,$13,$0F,$C2,$C3,$CA,$E8
	.dc.b	$F1,$E3,$F4,$E7,$6F,$C7,$8E,$07,$94,$A3,$04,$FC,$7E,$BF,$76,$26
	.dc.b	$92,$BD,$AD,$7E,$13,$5F,$73,$5F,$C2,$35,$F9,$FF,$1B,$D8,$8F,$C8
	.dc.b	$6B,$F3,$5A,$FC,$96,$BF,$8F,$6B,$D0,$FA,$3D,$DE,$A2,$38,$21,$F6
	.dc.b	$5C,$B4,$C9,$F5,$DC,$B4,$FB,$BE,$07,$E4,$3C,$D9,$1F,$4F,$04,$FE
	.dc.b	$44,$CC,$7D,$0C,$95,$98,$E1,$3A,$CF,$F5,$93,$99,$FA,$F7,$D1,$F0
	.dc.b	$90,$FC,$95,$7D,$82,$39,$3F,$20,$E1,$D9,$93,$F7,$0E,$6B,$F8,$0B
	.dc.b	$E1,$7F,$7D,$E6,$5D,$77,$5E,$83,$F2,$58,$A7,$51,$FA,$CB,$AF,$E3
	.dc.b	$E3,$F6,$9E,$1F,$EE,$3C,$C9,$3F,$26,$32,$F4,$8C,$BF,$0A,$32,$FC
	.dc.b	$B8,$CE,$3F,$C6,$27,$33,$FA,$F7,$F3,$64,$7F,$6A,$D7,$F7,$4D,$78
	.dc.b	$1A,$21,$A2,$C7,$77,$D0,$7C,$2F,$6F,$CB,$19,$8F,$41,$93,$66,$08
	.dc.b	$FF,$2B,$8F,$E3,$1E,$D7,$F1,$CF,$0E,$57,$E3,$63,$2F,$D0,$8C,$BF
	.dc.b	$86,$8C,$BF,$8F,$8C,$D7,$F3,$C9,$CC,$FF,$13,$7F,$7E,$47,$F8,$CB
	.dc.b	$5F,$E3,$CD,$78,$1A,$21,$A2,$C7,$77,$D0,$7E,$1B,$E0,$FC,$B1,$98
	.dc.b	$F4,$19,$36,$60,$8F,$F4,$F8,$FE,$8D,$E1,$BD,$7F,$E8,$3E,$74,$BF
	.dc.b	$E4,$93,$96,$CE,$5F,$D9,$4F,$28,$91,$F9,$A8,$D5,$FF,$94,$2B,$94
	.dc.b	$FF,$28,$6B,$FD,$2D,$A8,$8E,$1B,$5F,$D2,$63,$11,$FE,$A1,$8C,$11
	.dc.b	$FE,$A1,$1A,$24,$66,$3F,$AC,$47,$92,$7C,$D3,$EA,$9F,$E1,$DE,$6D
	.dc.b	$9C,$C7,$C2,$3F,$6F,$19,$E1,$F4,$47,$8C,$D1,$C8,$F9,$78,$7F,$75
	.dc.b	$FA,$57,$9D,$4F,$A7,$12,$7F,$AA,$4E,$9B,$39,$7E,$54,$F1,$72,$76
	.dc.b	$21,$EE,$7E,$41,$1F,$8A,$D7,$46,$08,$60,$87,$D7,$78,$61,$F7,$5F
	.dc.b	$23,$FE,$8E,$A6,$98,$C3,$E2,$E6,$BC,$61,$F7,$EF,$6F,$A3,$0E,$8D
	.dc.b	$BF,$9B,$3A,$8F,$C5,$7A,$1F,$2F,$CE,$91,$F9,$88,$FC,$E9,$87,$FB
	.dc.b	$BC,$A7,$31,$B3,$E7,$BB,$BF,$CC,$56,$A7,$F0,$EF,$E4,$48,$FE,$39
	.dc.b	$AE,$8C,$10,$C1,$0F,$D0,$3C,$23,$BC,$78,$CF,$3F,$B5,$E1,$8C,$C7
	.dc.b	$C0,$3F,$0B,$19,$C3,$E6,$D5,$9F,$99,$3F,$BE,$E1,$FC,$87,$2C,$9F
	.dc.b	$DE,$C6,$5F,$C7,$C6,$5F,$DB,$46,$5F,$93,$19,$C3,$F7,$93,$99,$FD
	.dc.b	$D3,$FB,$52,$3F,$B2,$6B,$FB,$06,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F
	.dc.b	$4B,$D3,$E5,$8C,$C7,$A0,$C9,$B3,$04,$7F,$2D,$1F,$C8,$BE,$77,$F6
	.dc.b	$9D,$92,$BF,$90,$C6,$5F,$D9,$C6,$5F,$E3,$51,$97,$F1,$B1,$9A,$FE
	.dc.b	$41,$39,$9F,$E3,$CF,$F1,$92,$3F,$C5,$9A,$FF,$14,$6B,$C0,$D1,$0D
	.dc.b	$16,$3B,$BE,$83,$F8,$4F,$77,$96,$33,$1E,$83,$26,$CC,$11,$FD,$D4
	.dc.b	$7F,$46,$EC,$FE,$31,$C3,$2B,$FA,$2C,$65,$FE,$2F,$19,$7F,$9F,$46
	.dc.b	$5F,$D8,$C6,$6B,$FA,$24,$E6,$7F,$A1,$3F,$CF,$48,$FF,$3A,$6B,$FC
	.dc.b	$E1,$AF,$03,$44,$34,$58,$EE,$FA,$0F,$E5,$BF,$09,$E5,$8C,$C7,$A0
	.dc.b	$C9,$B3,$04,$7F,$8F,$47,$F6,$AF,$5F,$F9,$E7,$04,$AF,$ED,$31,$97
	.dc.b	$F9,$DC,$65,$FE,$C5,$19,$7F,$8A,$C6,$6B,$FB,$44,$E6,$7F,$B3,$3F
	.dc.b	$D8,$48,$FF,$5E,$6B,$FD,$71,$AF,$03,$44,$34,$58,$EE,$FA,$0F,$F5
	.dc.b	$27,$A2,$F3,$FC,$D3,$A9,$F4,$9A,$8C,$DF,$F3,$EB,$B3,$FD,$EE,$F9
	.dc.b	$3F,$EA,$FF,$A6,$FF,$65,$60,$BF,$ED,$34,$F3,$FF,$DB,$4D,$BE,$07
	.dc.b	$FB,$76,$F8,$D7,$0A,$D9,$99,$87,$A9,$CB,$C8,$87,$FB,$66,$7F,$DA
	.dc.b	$89,$A4,$7F,$DB,$0D,$7F,$DB,$8D,$79,$1A,$21,$A3,$C6,$7F,$B2,$1F
	.dc.b	$4B,$F1,$7D,$CC,$E1,$F0,$E3,$FC,$45,$3A,$7C,$87,$C4,$9D,$DF,$F8
	.dc.b	$42,$74,$6F,$04,$FC,$7F,$95,$F9,$7E,$8C,$C3,$E9,$46,$5F,$BF,$19
	.dc.b	$7C,$D1,$97,$FD,$BF,$49,$D6,$3E,$79,$F9,$15,$B7,$FC,$46,$60,$F8
	.dc.b	$EF,$CA,$23,$82,$1F,$4D,$CB,$4C,$9D,$30,$46,$A7,$C5,$18,$9F,$69
	.dc.b	$F9,$AF,$84,$F5,$4E,$AA,$E5,$F9,$09,$B5,$F4,$26,$AC,$53,$F9,$1F
	.dc.b	$3B,$F2,$BE,$8C,$AF,$CE,$46,$5F,$D4,$C6,$5F,$C0,$46,$5F,$97,$19
	.dc.b	$C7,$D3,$39,$9F,$73,$F9,$02,$3F,$90,$6B,$F9,$96,$BC,$0D,$10,$D1
	.dc.b	$63,$BB,$E8,$3F,$7F,$E8,$79,$63,$31,$E8,$32,$6C,$C1,$1F,$C1,$60
	.dc.b	$9F,$DB,$7B,$BF,$BA,$B9,$7F,$C0,$63,$2F,$F0,$E8,$CB,$FA,$58,$CB
	.dc.b	$F9,$08,$CD,$FD,$E9,$CC,$FE,$0D,$FD,$A9,$1F,$DA,$B5,$FD,$F3,$5E
	.dc.b	$06,$88,$68,$B1,$DD,$F4,$1F,$C9,$3D,$DF,$2C,$66,$3D,$06,$4D,$98
	.dc.b	$23,$F1,$A3,$F8,$D6,$5F,$C7,$BC,$F9,$5F,$C9,$63,$2F,$F3,$28,$CB
	.dc.b	$FC,$26,$32,$FE,$D2,$33,$5F,$A2,$9C,$CF,$E9,$DF,$E3,$24,$7F,$8C
	.dc.b	$B5,$FE,$44,$D7,$81,$A2,$1A,$2C,$77,$7D,$07,$F4,$83,$CA,$E5,$F4
	.dc.b	$19,$36,$60,$8F,$E2,$9F,$E7,$DF,$07,$FE,$AC,$8F,$E1,$DF,$4E,$F4
	.dc.b	$71,$C1,$FE,$B0,$8C,$5F,$F4,$DB,$E2,$FF,$B2,$EF,$8D,$F0,$8F,$F6
	.dc.b	$75,$70,$7F,$9A,$57,$05,$5C,$1E,$57,$DE,$F5,$3F,$C4,$2B,$FA,$29
	.dc.b	$2B,$FA,$2B,$5F,$E9,$CD,$7F,$8D,$35,$FE,$74,$D7,$FA,$0B,$C6,$3B
	.dc.b	$47,$B4,$F1,$9E,$AF,$47,$C7,$7C,$A3,$47,$A9,$F8,$D1,$A8,$EA,$6A
	.dc.b	$7E,$6C,$7F,$6D,$E5,$F8,$D2,$C5,$D3,$86,$C7,$F7,$B9,$E0,$F9,$0E
	.dc.b	$13,$B3,$2F,$F6,$18,$C3,$B9,$E6,$E4,$E8,$7F,$6E,$8F,$EE,$24,$70
	.dc.b	$43,$FB,$D3,$96,$99,$3F,$A4,$39,$69,$D8,$F0,$1F,$1C,$FC,$2F,$9D
	.dc.b	$E8,$C3,$FA,$6C,$DC,$7F,$21,$36,$BF,$BB,$CD,$7E,$44,$7F,$85,$FD
	.dc.b	$5F,$91,$E7,$C9,$FF,$11,$46,$5F,$85,$19,$7F,$C0,$D1,$97,$C8,$7D
	.dc.b	$38,$D6,$1F,$D8,$27,$53,$FD,$95,$FF,$0B,$91,$FF,$0B,$B5,$FF,$10
	.dc.b	$35,$E0,$68,$86,$8B,$1D,$DF,$49,$F6,$7C,$2F,$34,$66,$3D,$26,$4D
	.dc.b	$98,$23,$FB,$C8,$FC,$7F,$D4,$3E,$3B,$FC,$13,$42,$46,$5C,$87,$0F
	.dc.b	$06,$BF,$E1,$26,$BC,$CE,$11,$C9,$A7,$48,$DB,$E0,$8D,$46,$23,$67
	.dc.b	$78,$CF,$7C,$8E,$4D,$7C,$8D,$7C,$8D,$7C,$92,$8C,$47,$C0,$6B,$EF
	.dc.b	$F6,$BC,$64,$7D,$0D,$7E,$13,$5F,$84,$D7,$BB,$5F,$9A,$FB,$BC,$19
	.dc.b	$1F,$0D,$AF,$CD,$6B,$F3,$5A,$FA,$9A,$FD,$86,$BF,$5F,$E8,$FE,$59
	.dc.b	$1F,$A0,$D7,$ED,$35,$FB,$4D,$7E,$3B,$5F,$C2,$BF,$5B,$C3,$91,$F6
	.dc.b	$DA,$FE,$15,$AF,$E1,$5A,$FD,$26,$BF,$90,$6B,$F8,$FF,$17,$F9,$17
	.dc.b	$A5,$23,$F9,$06,$49,$FC,$D9,$98,$FD,$CF,$1A,$B3,$1C,$27,$5B,$7F
	.dc.b	$5E,$9C,$CF,$F0,$0B,$E8,$FE,$56,$BF,$26,$47,$04,$3F,$44,$E5,$A6
	.dc.b	$4F,$D2,$39,$69,$E3,$21,$E8,$F4,$BC,$77,$FE,$7F,$AF,$1C,$F9,$6B
	.dc.b	$FA,$25,$6B,$93,$F3,$71,$67,$FB,$7C,$FE,$FB,$E1,$BF,$A5,$79,$F2
	.dc.b	$7E,$4C,$65,$FE,$89,$19,$7E,$14,$65,$F9,$71,$9C,$7C,$93,$99,$FE
	.dc.b	$00,$FE,$E0,$8F,$EF,$5A,$FF,$07,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83
	.dc.b	$FA,$D7,$A5,$E5,$8C,$C7,$A0,$C9,$B3,$04,$7F,$C3,$B1,$FC,$8B,$CD
	.dc.b	$FE,$E3,$F5,$2F,$4E,$99,$BF,$C7,$CE,$1F,$EE,$13,$B7,$24,$3F,$A4
	.dc.b	$B9,$6C,$F3,$7F,$1D,$3F,$C7,$88,$FF,$23,$6B,$FC,$B5,$AF,$F2,$D6
	.dc.b	$BF,$C6,$9A,$FF,$38,$78,$DF,$E2,$5F,$03,$FE,$C9,$A7,$A7,$8D,$79
	.dc.b	$5F,$A6,$B5,$39,$8E,$86,$67,$F9,$E5,$72,$8F,$D9,$FC,$E7,$CA,$F4
	.dc.b	$E4,$60,$87,$C4,$72,$FE,$BE,$3F,$A9,$98,$C3,$CD,$38,$9E,$27,$F4
	.dc.b	$D9,$FE,$8A,$47,$FA,$53,$5F,$EA,$8D,$78,$1A,$21,$A2,$CF,$EB,$E3
	.dc.b	$C1,$1F,$6B,$1B,$39,$8F,$5B,$F2,$E3,$38,$7F,$4C,$AB,$5F,$DA,$F0
	.dc.b	$57,$F7,$0C,$BF,$1F,$E9,$4A,$FE,$37,$19,$7F,$1B,$19,$7F,$8A,$46
	.dc.b	$5F,$C6,$C6,$71,$FE,$0D,$39,$9F,$EE,$4F,$F6,$A2,$3F,$DB,$DA,$FF
	.dc.b	$79,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$F4,$7E,$97,$96,$33,$1E,$83
	.dc.b	$26,$CC,$11,$FC,$A4,$7F,$98,$BE,$A9,$F1,$5F,$B4,$E3,$97,$FE,$5F
	.dc.b	$9C,$B6,$72,$FE,$C2,$79,$44,$8F,$DB,$C6,$AD,$AF,$F8,$56,$F9,$D7
	.dc.b	$F9,$01,$AF,$F8,$64,$38,$78,$5F,$F8,$6E,$31,$1F,$E2,$96,$BC,$23
	.dc.b	$44,$7F,$C5,$71,$A2,$46,$63,$FC,$67,$1E,$4A,$F3,$C7,$59,$F6,$3F
	.dc.b	$C3,$B9,$EC,$EA,$3C,$07,$C5,$3F,$8B,$C6,$AF,$D3,$19,$39,$1F,$37
	.dc.b	$93,$F9,$1F,$4C,$8C,$10,$FE,$3A,$E5,$F9,$11,$F6,$4C,$5F,$F0,$F9
	.dc.b	$C4,$F1,$3E,$66,$2A,$F9,$44,$8F,$8A,$D7,$E1,$B5,$E2,$D7,$ED,$35
	.dc.b	$FA,$4F,$81,$F2,$9F,$CE,$3A,$F1,$7F,$1E,$BC,$53,$ED,$8F,$E4,$D3
	.dc.b	$AC,$3A,$CE,$5F,$E4,$31,$F9,$7F,$90,$FE,$93,$12,$7F,$2D,$19,$7F
	.dc.b	$A3,$C6,$5F,$C7,$46,$5F,$E8,$F1,$9A,$F2,$4E,$67,$E3,$BF,$8F,$23
	.dc.b	$F8,$A6,$BF,$88,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$FD,$99,$F5,$6F
	.dc.b	$57,$CD,$3A,$9F,$49,$A8,$CD,$FE,$AE,$EC,$FF,$A2,$EF,$93,$FB,$9F
	.dc.b	$0B,$FD,$BF,$0F,$12,$9F,$C0,$62,$4F,$8A,$32,$F1,$21,$FE,$B6,$9E
	.dc.b	$4E,$CC,$E3,$F7,$6B,$F6,$C4,$8F,$D8,$B5,$FD,$7B,$5F,$CA,$B5,$FE
	.dc.b	$0C,$D7,$F7,$CF,$18,$ED,$1E,$0C,$3F,$D4,$14,$F5,$3D,$73,$FC,$36
	.dc.b	$ED,$FE,$DE,$B5,$3C,$4E,$A5,$0E,$9F,$E2,$DF,$11,$F2,$BD,$59,$1F
	.dc.b	$C9,$63,$F9,$51,$87,$E3,$46,$1D,$1F,$79,$D5,$7E,$8E,$75,$3F,$C4
	.dc.b	$9F,$E3,$C4,$7F,$8C,$B5,$D1,$82,$18,$21,$FC,$01,$E1,$1D,$E3,$C6
	.dc.b	$7D,$BF,$33,$C3,$19,$8F,$79,$F9,$48,$CE,$5F,$CE,$6A,$CF,$E3,$E7
	.dc.b	$FA,$57,$95,$F8,$EF,$4E,$4F,$F4,$98,$CB,$F9,$E8,$CB,$FD,$06,$32
	.dc.b	$FE,$36,$33,$87,$F0,$79,$CC,$FF,$1A,$7F,$A1,$11,$FE,$7A,$D7,$F9
	.dc.b	$DB,$5E,$06,$88,$68,$B1,$DD,$F4,$1F,$B4,$F4,$BC,$B1,$98,$F4,$19
	.dc.b	$36,$60,$8F,$E3,$E3,$FB,$77,$D2,$7E,$CB,$D3,$95,$FD,$B6,$32,$FF
	.dc.b	$02,$8C,$BF,$D9,$63,$2F,$EC,$63,$35,$FC,$AE,$73,$3F,$CF,$9F,$EC
	.dc.b	$C4,$7F,$B0,$B5,$FE,$BE,$D7,$81,$A2,$1A,$2C,$77,$7D,$07,$F1,$8F
	.dc.b	$4B,$CB,$19,$8F,$41,$93,$66,$08,$FE,$CE,$11,$23,$FB,$67,$C1,$3D
	.dc.b	$3F,$E3,$24,$9D,$6D,$C1,$4E,$DC,$9D,$1C,$3F,$C6,$67,$92,$AE,$F1
	.dc.b	$CD,$B1,$FC,$D2,$37,$18,$8E,$15,$E1,$AE,$B1,$FE,$19,$6B,$9A,$FF
	.dc.b	$0E,$34,$A7,$A1,$AD,$33,$1C,$9F,$54,$6A,$31,$1B,$3A,$B8,$24,$7A
	.dc.b	$E7,$C1,$3E,$83,$99,$FD,$40,$D1,$83,$47,$5A,$F8,$07,$F2,$97,$66
	.dc.b	$63,$AC,$69,$FE,$67,$1A,$AF,$AA,$FD,$41,$D4,$F4,$F9,$B6,$35,$3C
	.dc.b	$1F,$1B,$6F,$FC,$61,$3B,$9F,$3B,$63,$F1,$33,$E1,$3A,$D4,$AE,$B1
	.dc.b	$97,$E7,$BF,$34,$8E,$08,$7C,$E7,$2D,$32,$7E,$05,$CB,$4E,$E4,$3C
	.dc.b	$FA,$F8,$5F,$6C,$5A,$F6,$CD,$C3,$F1,$D3,$6B,$F4,$F8,$2B,$E7,$6B
	.dc.b	$F2,$BD,$29,$5F,$2A,$32,$FD,$1A,$4E,$A3,$E6,$CE,$9F,$97,$19,$C7
	.dc.b	$F6,$73,$99,$FC,$F3,$F9,$32,$3F,$98,$6B,$F9,$46,$BC,$0D,$10,$D1
	.dc.b	$63,$BB,$E9,$3D,$1E,$97,$96,$33,$1E,$83,$26,$CC,$11,$FC,$3C,$7E
	.dc.b	$4F,$E8,$BF,$1F,$CD,$2B,$F1,$F1,$97,$E3,$46,$5F,$C8,$C6,$5F,$C7
	.dc.b	$C6,$6B,$F8,$BC,$E6,$7F,$81,$3F,$B7,$23,$FB,$C6,$BF,$B8,$6B,$C0
	.dc.b	$D1,$0D,$16,$3B,$BE,$83,$EE,$7A,$5E,$58,$CC,$7A,$0C,$9B,$30,$47
	.dc.b	$F5,$71,$FB,$7E,$6F,$D9,$F9,$D2,$BF,$67,$19,$7F,$15,$19,$7F,$6B
	.dc.b	$19,$7F,$67,$19,$AF,$E7,$73,$99,$FE,$4C,$FF,$1B,$23,$FC,$81,$AF
	.dc.b	$F1,$C6,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$41,$DF,$E5,$8C,$C7,$A0
	.dc.b	$C9,$B3,$04,$7F,$87,$C7,$F1,$BF,$85,$FE,$2F,$E7,$4A,$FE,$2F,$19
	.dc.b	$7F,$5D,$19,$7F,$8C,$C6,$5F,$E2,$F1,$9A,$FE,$BF,$39,$9F,$E9,$4F
	.dc.b	$F3,$F2,$3F,$D1,$1A,$FF,$40,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$F8
	.dc.b	$27,$9B,$E5,$8C,$C7,$A0,$C9,$B3,$04,$7F,$99,$C7,$F3,$FF,$A7,$FE
	.dc.b	$AC,$49,$FE,$B6,$D7,$FA,$BB,$5F,$ED,$0D,$7F,$B6,$35,$FE,$F4,$FF
	.dc.b	$61,$F0,$64,$7F,$66,$6B,$FD,$89,$AF,$F7,$36,$BF,$DD,$DA,$FF,$81
	.dc.b	$5A,$BE,$B1,$F1,$5F,$D7,$9B,$1F,$D6,$5A,$FF,$66,$5E,$56,$E6,$4E
	.dc.b	$DE,$51,$88,$E1,$17,$17,$47,$F8,$6E,$F9,$D7,$12,$1C,$0F,$15,$F1
	.dc.b	$9E,$17,$67,$FC,$57,$5C,$9E,$04,$BE,$EE,$A7,$CA,$FF,$75,$FA,$D7
	.dc.b	$AF,$FE,$14,$6C,$7F,$86,$1A,$FF,$87,$1A,$FF,$8C,$C3,$87,$FC,$6C
	.dc.b	$D7,$FC,$7E,$D7,$71,$DC,$8F,$76,$BD,$FC,$7F,$21,$1F,$23,$5D,$18
	.dc.b	$21,$82,$1E,$79,$47,$18,$C7,$AF,$F0,$48,$FC,$16,$BF,$05,$AF,$03
	.dc.b	$44,$34,$53,$C1,$E5,$7C,$D2,$3F,$35,$AF,$CD,$6B,$C0,$D1,$0D,$14
	.dc.b	$F6,$75,$FE,$67,$D2,$91,$F7,$63,$FC,$73,$87,$F8,$CA,$3F,$39,$3A
	.dc.b	$7F,$C8,$31,$96,$CE,$73,$FC,$4C,$7E,$9E,$B9,$CC,$8F,$C3,$3F,$89
	.dc.b	$23,$82,$1F,$89,$72,$D3,$27,$A8,$A7,$E2,$9E,$CF,$90,$FD,$A7,$A1
	.dc.b	$8A,$F2,$DF,$A5,$B5,$FB,$B9,$D6,$DC,$EF,$E2,$1C,$8F,$D1,$7C,$37
	.dc.b	$F0,$1F,$A9,$91,$82,$1F,$C0,$5C,$BF,$22,$3E,$69,$8C,$3E,$3F,$02
	.dc.b	$B3,$5C,$8F,$C9,$57,$EB,$C8,$FE,$BD,$AF,$EB,$DA,$F0,$34,$43,$45
	.dc.b	$AF,$87,$1E,$08,$FE,$31,$E9,$59,$CC,$7B,$1F,$3C,$67,$0F,$DC,$D5
	.dc.b	$AF,$E1,$B3,$F8,$A3,$F9,$37,$AD,$23,$F8,$54,$65,$FE,$51,$19,$7F
	.dc.b	$1B,$19,$7F,$95,$46,$6F,$EC,$CE,$67,$F6,$CF,$F1,$32,$3F,$C4,$DA
	.dc.b	$FF,$13,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$F9,$F7,$A7,$E5,$8C,$C7
	.dc.b	$A0,$C9,$B3,$04,$7E,$FC,$7E,$BB,$E1,$7F,$A5,$61,$2B,$F9,$74,$65
	.dc.b	$FE,$99,$19,$7F,$63,$19,$7F,$A7,$46,$6B,$F3,$B3,$99,$FE,$34,$FF
	.dc.b	$37,$23,$FC,$DD,$AF,$F3,$76,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$B1
	.dc.b	$7A,$9E,$58,$CC,$7A,$0C,$9B,$30,$47,$F4,$71,$FC,$4B,$D7,$FE,$DD
	.dc.b	$F1,$72,$BF,$A9,$46,$5F,$EE,$11,$97,$F8,$AC,$65,$FE,$E5,$19,$AF
	.dc.b	$E0,$33,$99,$FE,$7C,$FF,$5B,$23,$FD,$6D,$AF,$F5,$B6,$BC,$0D,$10
	.dc.b	$D1,$63,$BB,$E8,$3F,$C0,$DC,$1E,$58,$CC,$7A,$0C,$9B,$30,$47,$F8
	.dc.b	$3C,$7F,$BB,$BE,$B9,$F0,$FF,$E1,$3F,$76,$5F,$FB,$AE,$72,$D9,$CB
	.dc.b	$FE,$15,$9E,$51,$23,$FC,$35,$1A,$BF,$E7,$75,$CA,$7F,$E0,$0B,$5F
	.dc.b	$EF,$8D,$44,$70,$DA,$FF,$01,$3A,$11,$A6,$C6,$9E,$05,$8D,$12,$35
	.dc.b	$1C,$0D,$38,$8F,$44,$FA,$A3,$DD,$3F,$E3,$5E,$6B,$39,$8F,$8C,$7F
	.dc.b	$8E,$E3,$3C,$3E,$7B,$E6,$DB,$FF,$21,$4F,$D3,$F0,$3F,$C8,$9D,$13
	.dc.b	$0F,$A5,$19,$7F,$C8,$D1,$97,$E5,$C6,$5F,$F2,$44,$67,$1F,$A3,$39
	.dc.b	$9F,$9E,$FB,$48,$FC,$56,$BF,$3B,$14,$62,$38,$9B,$21,$B7,$8C,$F8
	.dc.b	$23,$D2,$7F,$97,$7A,$7C,$D1,$98,$F4,$99,$36,$60,$8F,$F5,$E8,$FC
	.dc.b	$C7,$C3,$7F,$9A,$7A,$E5,$7E,$5E,$32,$FF,$9A,$E3,$2F,$E4,$23,$2F
	.dc.b	$F9,$B6,$33,$5F,$96,$9C,$CF,$E5,$1F,$C1,$11,$FC,$53,$5F,$8C,$D7
	.dc.b	$81,$A2,$1A,$2C,$77,$7D,$07,$FA,$07,$4F,$2C,$66,$3D,$06,$4D,$98
	.dc.b	$23,$FE,$02,$8F,$DE,$78,$BF,$A5,$8F,$F1,$5F,$D8,$3E,$03,$15,$FA
	.dc.b	$98,$C5,$FE,$FE,$F8,$BF,$B2,$BE,$37,$C2,$3F,$6B,$5C,$1F,$DD,$D7
	.dc.b	$05,$5C,$1E,$57,$DE,$F5,$3F,$C1,$2B,$F5,$C4,$AF,$DB,$35,$FC,$93
	.dc.b	$5F,$E0,$ED,$7F,$58,$D7,$F8,$43,$C6,$3B,$47,$B4,$F1,$9E,$AF,$07
	.dc.b	$C7,$7C,$A3,$47,$A9,$F8,$D1,$A8,$EA,$6A,$7E,$6C,$7F,$2B,$F8,$7F
	.dc.b	$8D,$2C,$5E,$D6,$1B,$1F,$CB,$67,$83,$E4,$38,$4E,$CC,$BF,$CA,$23
	.dc.b	$0E,$E7,$9B,$93,$A1,$FC,$7E,$3F,$8D,$11,$C1,$0F,$DF,$39,$69,$93
	.dc.b	$F9,$AB,$96,$9D,$8F,$01,$F1,$CF,$C2,$FB,$5E,$8C,$3F,$A0,$CD,$C7
	.dc.b	$F2,$13,$6B,$F9,$F4,$D7,$E4,$47,$F5,$5F,$3F,$E4,$7C,$14,$9F,$EA
	.dc.b	$91,$97,$E1,$46,$5F,$E9,$B1,$97,$C8,$7D,$38,$D6,$1F,$D5,$27,$53
	.dc.b	$FD,$49,$FE,$80,$47,$FA,$43,$5F,$E4,$8D,$78,$1A,$21,$A2,$C7,$77
	.dc.b	$D2,$7D,$9F,$2F,$CD,$19,$8F,$49,$93,$66,$08,$FE,$F2,$3F,$BD,$7C
	.dc.b	$3F,$E3,$7E,$36,$57,$F7,$98,$CB,$F8,$78,$CB,$FD,$CA,$32,$FC,$6D
	.dc.b	$13,$AB,$FE,$F3,$5A,$AF,$EE,$EF,$F6,$52,$3F,$DB,$5A,$FF,$49,$6B
	.dc.b	$C0,$D1,$0D,$16,$3B,$BE,$93,$F3,$DE,$DF,$9A,$33,$1E,$93,$26,$CC
	.dc.b	$11,$FE,$43,$1F,$B4,$FB,$0F,$F0,$44,$FB,$FD,$1F,$DA,$19,$61,$FA
	.dc.b	$D7,$2F,$ED,$63,$8D,$7F,$B5,$9B,$BF,$F2,$BA,$DD,$7E,$42,$72,$FF
	.dc.b	$87,$1F,$F0,$7A,$69,$1F,$F0,$9B,$5F,$F0,$D3,$5F,$F1,$0B,$5F,$5B
	.dc.b	$5F,$F1,$23,$E0,$7B,$1E,$C3,$EF,$7B,$96,$FD,$51,$6F,$FA,$24,$D8
	.dc.b	$F4,$14,$FE,$38,$FC,$BF,$88,$FE,$39,$F0,$F2,$7E,$34,$65,$FE,$2D
	.dc.b	$19,$7E,$24,$65,$FC,$84,$66,$FE,$64,$E6,$7D,$AF,$C4,$23,$F1,$1A
	.dc.b	$FC,$46,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$96,$7A,$FE,$58,$CC,$7A
	.dc.b	$0C,$9B,$30,$47,$FB,$7C,$7E,$3F,$DE,$FE,$81,$F1,$12,$BF,$15,$19
	.dc.b	$7F,$9D,$46,$5F,$C4,$46,$5F,$DA,$46,$6B,$F2,$13,$99,$FC,$0B,$F8
	.dc.b	$82,$3F,$88,$6B,$F8,$86,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$A8,$47
	.dc.b	$96,$33,$1E,$83,$26,$CC,$11,$FF,$09,$C7,$EC,$F0,$FE,$C9,$ED,$CA
	.dc.b	$FD,$74,$65,$FE,$BD,$19,$7F,$59,$19,$7F,$8C,$46,$6B,$F6,$93,$99
	.dc.b	$FD,$2B,$FA,$C2,$3F,$AC,$6B,$FA,$C6,$BC,$0D,$10,$D1,$63,$BB,$E8
	.dc.b	$3F,$BA,$7B,$DE,$58,$CC,$7A,$0C,$9B,$30,$47,$FC,$6F,$1F,$C5,$F7
	.dc.b	$FF,$04,$67,$2B,$F8,$94,$65,$FF,$01,$46,$5F,$E2,$11,$97,$F9,$E4
	.dc.b	$66,$BF,$8C,$4E,$67,$F8,$4B,$FC,$40,$8F,$F1,$06,$BF,$C4,$1A,$F0
	.dc.b	$34,$43,$45,$8E,$EF,$A0,$FF,$0C,$7B,$DE,$58,$CC,$7A,$0C,$9B,$30
	.dc.b	$47,$FC,$B7,$1F,$CE,$F9,$BF,$C5,$9F,$1B,$2B,$F9,$B4,$65,$FF,$13
	.dc.b	$46,$5F,$E6,$91,$97,$FB,$04,$66,$BF,$9E,$4E,$67,$F9,$6B,$FC,$D0
	.dc.b	$8F,$F3,$46,$BF,$CD,$1A,$F0,$34,$43,$45,$8E,$EF,$A0,$FF,$1E,$1E
	.dc.b	$57,$2F,$A0,$C9,$B3,$04,$7F,$CF,$EF,$F5,$FF,$67,$FA,$59,$1F,$E9
	.dc.b	$6D,$7F,$A5,$B5,$FE,$96,$D7,$FA,$5B,$5F,$EB,$4F,$F5,$6E,$B9,$1F
	.dc.b	$D5,$9A,$FF,$56,$6B,$FD,$59,$AF,$F5,$66,$BF,$D9,$5A,$FF,$65,$F6
	.dc.b	$3F,$B1,$91,$FE,$C6,$D7,$FB,$1B,$5F,$EC,$6D,$7F,$B1,$B5,$FE,$E4
	.dc.b	$FF,$6E,$ED,$91,$FD,$B9,$AF,$F6,$E6,$BF,$DB,$9A,$FF,$6E,$6B,$FD
	.dc.b	$F5,$AF,$F7,$DE,$FF,$EF,$64,$7F,$BD,$B5,$FE,$F6,$D7,$FB,$DB,$5F
	.dc.b	$EF,$6D,$7F,$C1,$4F,$F8,$69,$13,$DA,$73,$66,$30,$D5,$1A,$74,$D7
	.dc.b	$A3,$5F,$F1,$3B,$5E,$E7,$28,$E9,$19,$8E,$E7,$48,$C9,$1E,$FE,$1F
	.dc.b	$FA,$9B,$EC,$5F,$01,$4F,$F5,$4D,$A6,$99,$AF,$F5,$35,$70,$7F,$D5
	.dc.b	$13,$B6,$CE,$DF,$F4,$0C,$ED,$FF,$41,$CE,$DB,$3E,$A2,$3E,$62,$1F
	.dc.b	$11,$CB,$4C,$9F,$19,$CB,$4E,$3A,$FC,$D9,$F0,$36,$7E,$17,$67,$BA
	.dc.b	$BD,$24,$3D,$2E,$5F,$39,$99,$FB,$16,$9F,$93,$E5,$FC,$DF,$80,$97
	.dc.b	$F1,$A3,$2F,$CA,$8C,$BF,$12,$32,$FC,$88,$CD,$FA,$67,$33,$F8,$47
	.dc.b	$F0,$24,$7F,$16,$D7,$F1,$6D,$78,$1A,$21,$A2,$C7,$77,$D0,$7E,$03
	.dc.b	$E8,$3C,$B1,$98,$F4,$19,$36,$60,$8F,$C6,$8F,$C6,$FA,$3F,$91,$F1
	.dc.b	$25,$7E,$2A,$32,$FE,$3A,$32,$FE,$22,$32,$FE,$32,$33,$5F,$76,$73
	.dc.b	$3F,$A8,$7F,$4A,$47,$F5,$ED,$7F,$5E,$D7,$81,$A2,$1A,$2C,$77,$7D
	.dc.b	$07,$E9,$3E,$0F,$CB,$19,$8F,$41,$93,$66,$08,$FE,$2A,$3F,$63,$EB
	.dc.b	$FE,$D7,$E0,$A5,$7E,$BA,$32,$FE,$CA,$32,$FE,$B2,$32,$FE,$C2,$33
	.dc.b	$5F,$A1,$9C,$CF,$F0,$C7,$F8,$49,$1F,$E2,$6D,$7F,$89,$B5,$E0,$68
	.dc.b	$86,$8B,$1D,$DF,$41,$FC,$23,$C9,$F2,$C6,$63,$D0,$64,$D9,$82,$3F
	.dc.b	$AE,$8F,$E2,$BE,$DF,$F1,$9F,$82,$95,$FC,$4A,$32,$FF,$16,$8C,$BF
	.dc.b	$C4,$23,$2F,$F1,$48,$CD,$7F,$05,$9C,$CF,$F3,$07,$F9,$69,$1F,$E6
	.dc.b	$ED,$7F,$9B,$B5,$E0,$68,$86,$8B,$1D,$DF,$41,$FC,$B3,$CC,$F2,$C6
	.dc.b	$63,$D0,$64,$D9,$82,$3F,$C4,$A3,$F9,$CF,$C4,$FF,$3D,$F8,$69,$5F
	.dc.b	$CD,$A3,$2F,$F3,$A8,$CB,$FC,$D2,$32,$FF,$38,$8C,$D7,$F2,$99,$CC
	.dc.b	$FF,$54,$7F,$A8,$91,$FE,$B6,$D7,$FA,$DB,$5E,$06,$88,$68,$B1,$DD
	.dc.b	$F4,$1F,$D4,$3E,$13,$CB,$19,$8F,$41,$93,$66,$08,$FF,$36,$8F,$EB
	.dc.b	$BF,$07,$FD,$87,$D1,$95,$FD,$6A,$32,$FF,$5E,$8C,$BF,$D6,$23,$2F
	.dc.b	$F5,$C8,$CD,$7F,$4D,$9C,$CF,$F7,$87,$FB,$A9,$1F,$EF,$ED,$7F,$BF
	.dc.b	$B5,$E0,$68,$86,$8B,$1D,$DF,$41,$FD,$D3,$A3,$CB,$19,$8F,$41,$93
	.dc.b	$66,$08,$FF,$5A,$8F,$F0,$6B,$FE,$13,$FB,$27,$C4,$FF,$81,$FE,$62
	.dc.b	$5F,$F8,$82,$B5,$3C,$3B,$7F,$C2,$B7,$C2,$BF,$C0,$F7,$C1,$5F,$07
	.dc.b	$95,$70,$BD,$4F,$F6,$CA,$E9,$5F,$E2,$EB,$E8,$F2,$BE,$0F,$F8,$BE
	.dc.b	$32,$FF,$89,$83,$87,$67,$F8,$E5,$C3,$D6,$FF,$C6,$D1,$88,$E3,$18
	.dc.b	$23,$DD,$AF,$37,$CE,$F0,$3D,$A7,$53,$FC,$53,$EB,$6A,$7D,$78,$7C
	.dc.b	$39,$F6,$1B,$3F,$C5,$33,$BA,$91,$D0,$EC,$68,$C5,$7F,$93,$9F,$F1
	.dc.b	$BC,$7F,$8F,$CF,$06,$32,$32,$43,$C2,$7C,$87,$4F,$F8,$EA,$39,$3D
	.dc.b	$A3,$2F,$CF,$9F,$03,$F3,$63,$E5,$91,$FA,$CD,$74,$60,$86,$08,$75
	.dc.b	$7C,$77,$F9,$57,$FC,$B5,$F0,$F8,$C7,$F2,$31,$6B,$FC,$EB,$DA,$AC
	.dc.b	$47,$B4,$F5,$1F,$E5,$C7,$EA,$70,$7D,$AD,$9B,$53,$35,$F3,$A7,$0F
	.dc.b	$D3,$9D,$B9,$21,$F3,$1C,$B6,$79,$BF,$43,$35,$7E,$48,$91,$F9,$D6
	.dc.b	$BF,$99,$6B,$F4,$5A,$FE,$C9,$AF,$EA,$5E,$39,$FE,$F5,$F0,$BF,$E9
	.dc.b	$1B,$7C,$3E,$34,$E5,$7E,$AA,$D4,$E6,$3A,$19,$9F,$F4,$AD,$72,$8F
	.dc.b	$C9,$F2,$7C,$BF,$7A,$46,$08,$7C,$57,$2F,$C3,$8F,$D3,$18,$C3,$CD
	.dc.b	$38,$9E,$27,$E4,$67,$F7,$24,$7F,$82,$35,$FE,$00,$D7,$81,$A2,$1A
	.dc.b	$2C,$FF,$23,$8F,$04,$7D,$BE,$3B,$39,$8F,$5B,$F2,$E3,$38,$7F,$24
	.dc.b	$AB,$5F,$EC,$8C,$15,$FC,$0F,$A7,$F2,$1F,$0F,$2B,$F7,$71,$97,$F1
	.dc.b	$D1,$97,$F1,$11,$97,$F1,$D1,$9C,$7F,$9D,$4E,$67,$F6,$CF,$F1,$E2
	.dc.b	$3F,$CA,$5A,$FF,$25,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$F4,$9F,$03
	.dc.b	$E5,$8C,$C7,$A0,$C9,$B3,$04,$7F,$29,$1F,$C9,$FC,$0F,$DA,$74,$CA
	.dc.b	$FE,$3F,$19,$7F,$65,$19,$7F,$59,$19,$7F,$65,$19,$AF,$EB,$D3,$99
	.dc.b	$FE,$34,$FF,$42,$23,$FD,$35,$AF,$F4,$96,$BC,$0D,$10,$D1,$63,$BB
	.dc.b	$E8,$3F,$84,$7C,$2F,$96,$33,$1E,$83,$26,$CC,$11,$FD,$C4,$7F,$4B
	.dc.b	$F8,$8F,$E3,$1F,$0D,$2B,$FA,$1C,$65,$FE,$2D,$19,$7F,$88,$46,$5F
	.dc.b	$E2,$D1,$9A,$FF,$01,$4E,$67,$F9,$F3,$FD,$98,$8F,$F7,$16,$BF,$DB
	.dc.b	$5A,$F0,$34,$43,$45,$8E,$EF,$A0,$FE,$59,$F0,$DE,$58,$CC,$7A,$0C
	.dc.b	$9B,$30,$47,$F8,$E4,$7F,$74,$FB,$37,$A1,$FC,$F7,$E1,$E5,$FF,$6B
	.dc.b	$9C,$B6,$72,$FF,$39,$9E,$51,$23,$FA,$04,$6A,$DC,$69,$97,$D1,$FE
	.dc.b	$DA,$D7,$FC,$2C,$D7,$80,$70,$F2,$BF,$F0,$E3,$A8,$CB,$B8,$CB,$B2
	.dc.b	$C6,$89,$1A,$8E,$06,$9C,$47,$9E,$BD,$73,$CC,$F4,$CF,$F6,$5F,$13
	.dc.b	$E1,$46,$0F,$90,$7F,$63,$73,$3C,$A3,$B9,$B3,$FB,$24,$7D,$5F,$5B
	.dc.b	$FB,$87,$C3,$C9,$FA,$91,$97,$FB,$74,$65,$F9,$71,$97,$FB,$74,$66
	.dc.b	$BC,$18,$2B,$55,$F4,$1F,$79,$1F,$8E,$D7,$E7,$C7,$17,$24,$33,$1B
	.dc.b	$7A,$19,$9F,$0C,$79,$5F,$ED,$DF,$0F,$72,$FB,$53,$6B,$FA,$24,$D7
	.dc.b	$F0,$51,$79,$93,$F9,$BF,$6F,$FC,$2D,$E5,$CB,$FC,$D4,$65,$FF,$0A
	.dc.b	$46,$5F,$C8,$C6,$5F,$F0,$A4,$67,$1F,$8F,$39,$9F,$CA,$BF,$84,$23
	.dc.b	$F8,$C6,$BF,$19,$AF,$03,$44,$34,$58,$EE,$FA,$0F,$F1,$57,$C3,$79
	.dc.b	$63,$31,$E8,$32,$6C,$C1,$1F,$F0,$D4,$7E,$FF,$DF,$FF,$1D,$7C,$34
	.dc.b	$AF,$DF,$46,$5F,$F1,$C4,$65,$FD,$AC,$65,$FF,$1C,$46,$6B,$F1,$73
	.dc.b	$99,$FD,$CB,$FA,$82,$3F,$B0,$6B,$F8,$A6,$BC,$0D,$10,$D1,$63,$BB
	.dc.b	$E8,$3F,$C9,$DE,$87,$96,$33,$1E,$83,$26,$CC,$11,$FF,$1F,$47,$F2
	.dc.b	$3F,$81,$FD,$F9,$27,$F8,$3B,$5F,$D0,$B5,$FE,$2E,$D7,$F8,$93,$5F
	.dc.b	$E3,$EF,$F0,$AF,$83,$91,$FC,$49,$AF,$EB,$1A,$FF,$22,$6B,$FC,$75
	.dc.b	$AF,$F2,$A6,$BF,$C9,$7D,$9F,$E2,$E4,$7F,$8F,$B5,$FD,$CB,$5F,$E5
	.dc.b	$ED,$7F,$95,$35,$FE,$6E,$FF,$22,$F8,$49,$1F,$CA,$9A,$FF,$04,$6B
	.dc.b	$FC,$E9,$AF,$F3,$56,$BF,$D0,$9A,$FF,$3D,$F6,$BF,$97,$91,$FE,$6E
	.dc.b	$D7,$F8,$8B,$5F,$E8,$ED,$7F,$A1,$35,$FE,$9E,$FF,$3A,$F6,$A4,$7F
	.dc.b	$42,$6B,$FC,$71,$AF,$F5,$26,$BF,$D3,$5A,$FF,$5A,$6B,$FD,$57,$DA
	.dc.b	$FE,$8E,$47,$FA,$7B,$5F,$E5,$2D,$7F,$AF,$B5,$FE,$B4,$D7,$FB,$3B
	.dc.b	$FD,$4B,$DA,$91,$FD,$69,$AF,$F3,$46,$BF,$DA,$9A,$FF,$65,$6B,$FD
	.dc.b	$C9,$AF,$F6,$DF,$6B,$FA,$F9,$1F,$EC,$ED,$7F,$A0,$B5,$FE,$EE,$D7
	.dc.b	$FB,$93,$5F,$EF,$EF,$F6,$DF,$B5,$FE,$C1,$1F,$DF,$03,$87,$FB,$FB
	.dc.b	$5F,$F0,$43,$5D,$F3,$F3,$23,$FD,$FA,$34,$48,$CC,$7F,$81,$A3,$3F
	.dc.b	$0D,$E3,$24,$71,$6B,$B6,$BE,$78,$46,$23,$CE,$D7,$CE,$FB,$F7,$91
	.dc.b	$E5,$6B,$E3,$6B,$D5,$AF,$C2,$6B,$F0,$9A,$FC,$8E,$2F,$98,$47,$E1
	.dc.b	$35,$F8,$0D,$7D,$4D,$7E,$7B,$5F,$9E,$FD,$7E,$39,$1F,$35,$AF,$CB
	.dc.b	$6B,$F1,$5A,$FD,$A6,$BF,$69,$AF,$E0,$39,$7F,$08,$47,$ED,$35,$FB
	.dc.b	$0D,$7E,$93,$5F,$C3,$B5,$FC,$3B,$F8,$FE,$69,$1F,$85,$6B,$F8,$36
	.dc.b	$BF,$75,$AF,$E4,$9A,$FE,$49,$AF,$E6,$3A,$BF,$38,$47,$F2,$4D,$7F
	.dc.b	$20,$D7,$F1,$4D,$7F,$3E,$D7,$F3,$EF,$E9,$FA,$E4,$7E,$75,$AF,$E6
	.dc.b	$DA,$FE,$55,$AF,$EA,$9A,$FE,$A9,$AF,$EC,$3B,$7F,$68,$47,$F5,$4D
	.dc.b	$7F,$50,$D7,$F4,$4D,$7F,$6E,$D7,$F6,$EF,$F9,$AF,$ED,$5B,$FF,$9E
	.dc.b	$BA,$3F,$CC,$11,$8A,$FF,$42,$46,$27,$5C,$1F,$E8,$0E,$BF,$F3,$34
	.dc.b	$E6,$B7,$3F,$E8,$4B,$E5,$09,$D3,$67,$91,$D1,$FF,$49,$4F,$09,$D9
	.dc.b	$27,$73,$D8,$DD,$F1,$55,$C6,$7C,$53,$9B,$F3,$C6,$2B,$91,$C2,$79
	.dc.b	$57,$18,$FF,$52,$57,$07,$B9,$C2,$34,$F1,$39,$C6,$DF,$E5,$3E,$17
	.dc.b	$4E,$59,$1E,$0E,$34,$FA,$5B,$1D,$67,$6F,$C7,$3F,$97,$CE,$CC,$E5
	.dc.b	$F6,$27,$13,$D6,$7F,$9E,$11,$FE,$64,$D7,$F9,$83,$5F,$E5,$4D,$7F
	.dc.b	$9F,$B5,$F0,$9C,$8E,$C4,$3C,$C7,$F4,$8F,$2F,$C6,$E2,$FD,$AD,$9F
	.dc.b	$81,$3A,$BF,$F5,$85,$69,$B3,$AA,$ED,$18,$6E,$7F,$D5,$2F,$93,$E8
	.dc.b	$8C,$9F,$D5,$A3,$2F,$A8,$F1,$1F,$4D,$CB,$F6,$0E,$27,$AC,$B7,$E5
	.dc.b	$9E,$D1,$ED,$9E,$33,$98,$ED,$5E,$C9,$F5,$47,$80,$91,$97,$C4,$76
	.dc.b	$BD,$C6,$63,$77,$E0,$8C,$11,$C1,$23,$E0,$1D,$0E,$27,$91,$F9,$0E
	.dc.b	$63,$E6,$3A,$3E,$01,$ED,$3F,$BD,$57,$F8,$1C,$8E,$08,$7F,$7B,$72
	.dc.b	$D3,$27,$F7,$77,$2D,$3C,$53,$E4,$7F,$17,$C0,$F5,$3E,$1C,$E6,$FE
	.dc.b	$05,$05,$6D,$E6,$7C,$23,$D4,$74,$AE,$6E,$67,$9B,$CA,$32,$FC,$28
	.dc.b	$CC,$F5,$8E,$71,$C3,$16,$1F,$C4,$EB,$52,$C3,$CD,$7D,$4F,$29,$D5
	.dc.b	$CC,$F0,$6E,$1E,$2C,$3A,$D7,$F1,$8A,$D3,$F0,$5A,$F3,$3D,$95,$C9
	.dc.b	$E1,$5C,$F0,$F3,$32,$3C,$77,$D9,$F8,$C7,$03,$33,$B7,$27,$C2,$AF
	.dc.b	$A8,$77,$73,$87,$C5,$9E,$0F,$C3,$8F,$45,$F8,$DE,$C7,$43,$FC,$A5
	.dc.b	$87,$CF,$6C,$7F,$95,$5E,$C6,$0C,$1F,$34,$D4,$71,$76,$43,$C2,$70
	.dc.b	$74,$E0,$E3,$1E,$8B,$F6,$57,$A6,$7F,$1F,$ED,$E5,$E1,$19,$BF,$A6
	.dc.b	$27,$53,$E5,$9D,$5F,$E1,$E7,$35,$C2,$31,$5F,$12,$B6,$DB,$FC,$3D
	.dc.b	$F0,$AE,$2D,$9D,$1F,$85,$70,$FD,$49,$E2,$E8,$F2,$3E,$E3,$2E,$CF
	.dc.b	$6B,$97,$A1,$C6,$FD,$0F,$E5,$67,$BD,$FA,$23,$F7,$71,$8A,$E5,$1F
	.dc.b	$28,$E5,$7E,$A6,$9D,$67,$83,$E2,$3C,$27,$94,$F4,$99,$3A,$19,$BF
	.dc.b	$2B,$E2,$3C,$C7,$63,$04,$BF,$41,$2F,$E1,$BF,$76,$BC,$73,$DA,$30
	.dc.b	$F3,$6B,$CE,$7D,$F5,$F8,$4A,$FB,$B3,$E5,$BB,$7C,$0F,$04,$7A,$EB
	.dc.b	$F1,$4F,$A0,$E0,$58,$E1,$7D,$AB,$9B,$E8,$31,$3B,$70,$F5,$8D,$BF
	.dc.b	$32,$7C,$2F,$CC,$3D,$D3,$C1,$C5,$7A,$DE,$87,$32,$9C,$0C,$1F,$0C
	.dc.b	$FB,$27,$EF,$AF,$F7,$2F,$D8,$30,$7E,$D5,$CC,$7C,$28,$FD,$EC,$60
	.dc.b	$8E,$C8,$68,$B5,$F1,$9F,$75,$7F,$95,$FA,$BE,$0C,$79,$C9,$B2,$33
	.dc.b	$87,$86,$F9,$57,$F9,$42,$7F,$A0,$7B,$BF,$E4,$AE,$D9,$3F,$CD,$E3
	.dc.b	$2F,$F9,$26,$32,$FF,$5E,$8C,$BF,$E2,$AC,$13,$AB,$FE,$19,$5A,$AF
	.dc.b	$E4,$AF,$C4,$23,$FC,$6D,$AF,$F2,$26,$BC,$0D,$10,$D1,$63,$BB,$E9
	.dc.b	$3F,$CC,$FF,$6C,$F8,$6F,$3C,$EA,$7D,$46,$A3,$39,$FF,$5C,$BB,$3F
	.dc.b	$DD,$EF,$93,$D4,$E2,$4B,$FE,$5D,$1A,$8C,$57,$37,$84,$EA,$FA,$C7
	.dc.b	$F8,$0A,$ED,$F1,$AF,$31,$2F,$95,$7F,$96,$27,$51,$89,$D4,$FF,$6D
	.dc.b	$AF,$F4,$75,$6A,$30,$78,$AB,$81,$FD,$FA,$35,$1E,$33,$47,$03,$C1
	.dc.b	$7E,$8C,$D7,$E8,$8C,$36,$78,$9C,$CF,$86,$7C,$B8,$DB,$AA,$F0,$D7
	.dc.b	$48,$E2,$D7,$C2,$7D,$73,$35,$EE,$A9,$18,$34,$70,$3C,$27,$8F,$0F
	.dc.b	$A4,$E9,$F9,$2E,$70,$F8,$93,$D2,$38,$DF,$8E,$B8,$C7,$57,$8C,$66
	.dc.b	$B8,$CF,$57,$E7,$BF,$2E,$78,$92,$33,$1E,$13,$E3,$1C,$4E,$0E,$B0
	.dc.b	$FA,$E7,$C8,$C3,$E8,$4F,$8E,$79,$DF,$8E,$39,$99,$39,$90,$57,$4C
	.dc.b	$3E,$AC,$6C,$CB,$8C,$3D,$11,$D1,$F8,$38,$7C,$77,$31,$EB,$25,$75
	.dc.b	$7E,$B1,$D5,$F5,$3F,$83,$AF,$A0,$FF,$94,$8F,$09,$FE,$28,$8F,$D7
	.dc.b	$39,$6B,$F4,$0F,$F1,$83,$87,$E0,$9B,$8F,$E7,$31,$98,$EA,$70,$8C
	.dc.b	$91,$FA,$46,$A7,$FC,$59,$93,$E0,$7E,$06,$5E,$19,$F8,$33,$9A,$E3
	.dc.b	$87,$C1,$7F,$94,$5F,$07,$E5,$5F,$13,$FC,$E3,$5C,$63,$E3,$61,$B5
	.dc.b	$19,$C3,$E7,$E1,$D5,$F2,$E1,$B8,$C6,$1D,$23,$18,$73,$BE,$53,$E1
	.dc.b	$C3,$C6,$43,$BD,$70,$7E,$1C,$F0,$75,$7D,$9F,$F4,$F4,$65,$F9,$15
	.dc.b	$D5,$A6,$AB,$A3,$F8,$F9,$E1,$5C,$E3,$67,$88,$F9,$77,$E9,$AE,$38
	.dc.b	$7D,$06,$E1,$DF,$0E,$75,$8B,$F0,$1C,$8F,$5D,$65,$EF,$3C,$1D,$1F
	.dc.b	$60,$B7,$DD,$B7,$DE,$B9,$61,$FC,$B2,$F9,$57,$DA,$3F,$D8,$57,$F5
	.dc.b	$9E,$27,$CF,$8D,$C7,$DC,$C3,$FD,$66,$EE,$79,$1F,$1E,$FB,$CE,$B0
	.dc.b	$FB,$D3,$D0,$F5,$CE,$9F,$0C,$69,$E7,$5E,$B6,$C7,$AD,$E5,$87,$AE
	.dc.b	$B9,$47,$C3,$36,$7A,$5F,$13,$F3,$8C,$9F,$64,$E1,$5E,$D7,$C6,$6D
	.dc.b	$F7,$19,$77,$5F,$ED,$4F,$B7,$76,$7F,$8C,$9B,$3F,$EA,$C6,$BF,$EB
	.dc.b	$16,$BF,$C2,$DA,$F3,$8E,$44,$79,$35,$E4,$D7,$97,$67,$84,$8F,$85
	.dc.b	$AF,$85,$AF,$03,$44,$34,$53,$CD,$D9,$ED,$23,$ED,$6B,$ED,$6B,$C0
	.dc.b	$D1,$0D,$14,$F8,$5B,$7C,$92,$3F,$25,$AF,$C9,$6B,$C0,$D1,$0D,$14
	.dc.b	$F9,$DD,$9F,$54,$8F,$D5,$6B,$F5,$5A,$F0,$34,$43,$45,$3E,$D7,$67
	.dc.b	$E0,$48,$FE,$05,$AF,$E0,$5A,$F0,$34,$43,$45,$3F,$0D,$D9,$F8,$D2
	.dc.b	$3F,$8D,$6B,$F8,$D6,$BC,$0D,$10,$D1,$4F,$C9,$76,$7E,$64,$8F,$E6
	.dc.b	$5A,$FE,$65,$AF,$03,$44,$34,$53,$F3,$DE,$47,$E9,$48,$FE,$95,$AF
	.dc.b	$E9,$5A,$F0,$34,$43,$45,$3F,$55,$E6,$7E,$F9,$FE,$A7,$C5,$78,$FF
	.dc.b	$C9,$7B,$AB,$4D,$AF,$DF,$9C,$0B,$3C,$27,$67,$3A,$93,$C2,$B9,$2E
	.dc.b	$17,$0F,$F0,$5B,$DC,$FF,$A5,$6F,$9B,$C6,$33,$5F,$DE,$A7,$33,$CE
	.dc.b	$B9,$3E,$48,$DC,$62,$34,$05,$F9,$2B,$A4,$71,$3C,$04,$3C,$86,$4E
	.dc.b	$67,$2B,$F1,$C7,$28,$D9,$A9,$ED,$5E,$38,$FA,$67,$F2,$1A,$FE,$48
	.dc.b	$47,$04,$3F,$92,$39,$69,$93,$99,$83,$C0,$F9,$23,$FC,$2B,$E3,$59
	.dc.b	$F8,$55,$E4,$6D,$7E,$0E,$75,$1E,$D3,$E1,$5F,$F8,$AA,$33,$5C,$67
	.dc.b	$47,$FB,$36,$33,$5C,$B1,$46,$27,$9E,$1E,$47,$F0,$DD,$5F,$93,$F8
	.dc.b	$69,$18,$21,$FE,$D9,$72,$FC,$D8,$FC,$B9,$8C,$BF,$C7,$F3,$89,$E3
	.dc.b	$18,$7F,$33,$19,$79,$18,$21,$82,$47,$17,$C4,$74,$AF,$EB,$D5,$FD
	.dc.b	$74,$91,$FD,$75,$AF,$F5,$D6,$BF,$D7,$5A,$F2,$38,$10,$E0,$77,$3E
	.dc.b	$CB,$FB,$9F,$85,$F5,$57,$EB,$27,$26,$CF,$F2,$8C,$6A,$BB,$47,$B1
	.dc.b	$FF,$2C,$C7,$F0,$3F,$84,$FE,$09,$E6,$4A,$FD,$FC,$65,$FE,$03,$19
	.dc.b	$7F,$47,$9A,$75,$18,$9D,$61,$FE,$72,$9C,$BF,$E7,$48,$CE,$1F,$E6
	.dc.b	$A9,$CB,$FE,$6C,$8C,$D7,$F9,$5A,$72,$FF,$96,$23,$2F,$F9,$BA,$3F
	.dc.b	$8F,$7D,$C3,$DD,$FF,$4A,$4F,$F0,$11,$19,$72,$62,$2B,$4C,$C6,$DE
	.dc.b	$6D,$7F,$D4,$0D,$7B,$35,$FE,$20,$D7,$A1,$CA,$3B,$C6,$47,$BD,$82
	.dc.b	$4F,$89,$AF,$89,$AF,$89,$AF,$89,$AF,$89,$EF,$E0,$C8,$F6,$B5,$F6
	.dc.b	$B5,$F6,$B5,$F6,$B5,$F6,$B5,$F9,$BE,$F7,$B0,$8F,$CA,$6B,$F2,$9A
	.dc.b	$FC,$A6,$BF,$29,$AF,$CA,$7E,$3F,$8D,$23,$EA,$B5,$FA,$AD,$7E,$AB
	.dc.b	$5F,$AA,$D7,$EA,$B5,$FC,$2B,$F8,$A2,$3F,$D8,$AD,$ED,$62,$3F,$A1
	.dc.b	$C2,$71,$AF,$EC,$E7,$F1,$F5,$8D,$7C,$31,$CE,$BF,$2F,$32,$3F,$1E
	.dc.b	$FE,$10,$8E,$08,$7E,$49,$CB,$4C,$9F,$94,$72,$FF,$BD,$7E,$E5,$F0
	.dc.b	$DF,$E0,$36,$C7,$F7,$E6,$BF,$D7,$97,$87,$B9,$93,$B7,$94,$62,$38
	.dc.b	$45,$C7,$E1,$57,$29,$E0,$43,$67,$86,$B8,$63,$FB,$79,$F1,$D7,$2B
	.dc.b	$EF,$7C,$63,$0D,$AE,$17,$ED,$F1,$BD,$DE,$B4,$AF,$74,$65,$F8,$11
	.dc.b	$97,$E0,$46,$5F,$85,$19,$BF,$5C,$E5,$F6,$C6,$6F,$D1,$39,$7D,$51
	.dc.b	$9A,$FF,$09,$62,$AD,$46,$27,$4A,$74,$FD,$5F,$63,$EC,$7C,$34,$8C
	.dc.b	$10,$FB,$0E,$5F,$B5,$1F,$7C,$C6,$3F,$52,$71,$3C,$63,$0F,$E0,$E3
	.dc.b	$2F,$23,$04,$30,$48,$E2,$F8,$8E,$98,$FD,$BF,$84,$FC,$81,$23,$F2
	.dc.b	$1F,$0D,$5F,$C8,$45,$AF,$C4,$CD,$AE,$55,$F9,$7A,$FC,$D4,$E6,$3E
	.dc.b	$6C,$57,$E5,$45,$C5,$36,$30,$D9,$FD,$17,$CE,$7E,$97,$DF,$91,$82
	.dc.b	$1F,$A5,$72,$FE,$9E,$3F,$58,$63,$1F,$D0,$CE,$27,$8C,$61,$FD,$84
	.dc.b	$65,$E4,$60,$86,$09,$1C,$5F,$11,$D3,$1F,$EA,$3B,$7F,$4C,$24,$7F
	.dc.b	$5A,$6B,$FD,$69,$AF,$F5,$A6,$BC,$8E,$04,$3F,$B0,$FB,$7F,$D6,$08
	.dc.b	$FF,$66,$6B,$A3,$04,$30,$43,$FB,$4B,$C1,$FE,$EC,$3B,$7F,$B9,$B4
	.dc.b	$FE,$EA,$E0,$8E,$08,$7F,$7C,$72,$D3,$27,$F8,$11,$E5,$D3,$D0,$91
	.dc.b	$C5,$AE,$DA,$F9,$A3,$FC,$40,$EC,$86,$CA,$6F,$AB,$D2,$47,$CC,$D7
	.dc.b	$C8,$D7,$81,$A2,$1A,$29,$A3,$FD,$2B,$08,$FF,$60,$36,$3F,$9F,$72
	.dc.b	$7F,$AC,$D7,$DD,$7C,$D7,$E8,$57,$02,$3C,$33,$FF,$55,$4E,$EF,$A4
	.dc.b	$F8,$2B,$6F,$F4,$9B,$ED,$5D,$78,$3F,$B4,$E1,$CB,$0F,$F1,$A1,$CA
	.dc.b	$7A,$61,$FE,$37,$78,$5F,$29,$FE,$88,$DC,$3F,$D7,$91,$C4,$E7,$5C
	.dc.b	$5D,$DF,$79,$FE,$63,$5C,$5F,$79,$92,$9F,$12,$78,$B4,$E3,$3B,$8F
	.dc.b	$04,$78,$AB,$C0,$E8,$E0,$7F,$3B,$AF,$04,$F0,$8E,$F7,$EE,$BF,$16
	.dc.b	$1F,$1A,$78,$D6,$CE,$11,$9A,$E1,$3A,$C3,$C2,$F4,$AF,$4C,$E8,$F9
	.dc.b	$B1,$E9,$9D,$D7,$3A,$D9,$F2,$A7,$47,$93,$0F,$91,$3B,$8F,$80,$68
	.dc.b	$F7,$B9,$7E,$06,$19,$C3,$C3,$1A,$AE,$67,$C4,$C3,$D3,$87,$BA,$33
	.dc.b	$7C,$63,$C9,$5C,$63,$C8,$65,$F1,$99,$25,$F8,$09,$3E,$E8,$FA,$D1
	.dc.b	$87,$8B,$F1,$E3,$51,$F2,$CC,$B8,$C3,$DC,$FD,$5C,$3E,$5E,$1E,$8A
	.dc.b	$FA,$87,$94,$F8,$67,$92,$7D,$77,$E1,$8C,$BE,$17,$B9,$0C,$C7,$90
	.dc.b	$F0,$3F,$08,$E8,$6D,$F9,$58,$66,$7B,$46,$27,$67,$B4,$98,$7D,$93
	.dc.b	$DB,$87,$D4,$3D,$B7,$CE,$BD,$24,$AF,$96,$F9,$67,$84,$6C,$C9,$E5
	.dc.b	$79,$BE,$28,$F5,$BE,$4A,$F0,$BF,$25,$F2,$61,$F8,$78,$C3,$F3,$F0
	.dc.b	$F6,$DF,$47,$E5,$4F,$B6,$FE,$04,$61,$EE,$EA,$FB,$CF,$D8,$3C,$57
	.dc.b	$DD,$C4,$EE,$30,$FD,$99,$F2,$E1,$F7,$EF,$B9,$F0,$EF,$CF,$1F,$0D
	.dc.b	$F2,$61,$E7,$9D,$47,$67,$F1,$D8,$76,$7E,$74,$7E,$08,$F9,$67,$23
	.dc.b	$A5,$FA,$B0,$F2,$57,$C5,$AE,$27,$92,$B8,$BF,$08,$EF,$87,$A7,$0F
	.dc.b	$24,$E9,$FB,$6F,$C4,$8D,$3E,$0A,$F3,$3F,$97,$C3,$E1,$BE,$B3,$11
	.dc.b	$E3,$7F,$04,$71,$9E,$6F,$53,$81,$F9,$48,$E0,$FB,$CD,$61,$F8,$07
	.dc.b	$91,$C3,$0F,$73,$F1,$CE,$11,$F6,$6B,$BB,$63,$86,$1F,$01,$E1,$5F
	.dc.b	$15,$F9,$F8,$7C,$D8,$FA,$6F,$BA,$BC,$53,$E4,$9F,$9F,$5C,$EB,$84
	.dc.b	$FC,$2B,$FC,$15,$75,$8F,$34,$E9,$F8,$13,$AA,$FA,$37,$E7,$3B,$96
	.dc.b	$FB,$D7,$C2,$8C,$C7,$DC,$8F,$69,$97,$6F,$9C,$FA,$65,$3E,$3B,$C6
	.dc.b	$39,$C7,$C2,$C3,$EF,$BE,$63,$F7,$55,$E6,$AE,$07,$6A,$F3,$4F,$66
	.dc.b	$BD,$8E,$97,$F5,$27,$8C,$6D,$AE,$CF,$8F,$7E,$28,$FB,$F8,$7C,$96
	.dc.b	$C6,$70,$FA,$4D,$AF,$84,$75,$3C,$27,$C6,$3E,$14,$FD,$5A,$FB,$CF
	.dc.b	$AC,$F1,$99,$9F,$CA,$B6,$FE,$C5,$7F,$00,$3E,$29,$70,$FE,$25,$87
	.dc.b	$C7,$8F,$0C,$7C,$E7,$DB,$3F,$0D,$CD,$78,$5A,$7A,$4D,$9E,$99,$F9
	.dc.b	$B1,$E3,$8F,$09,$FA,$CC,$3D,$6F,$16,$BF,$35,$B1,$C6,$7C,$F3,$E9
	.dc.b	$8F,$E2,$F1,$FC,$3A,$BD,$F3,$D7,$0F,$87,$87,$B6,$3C,$B5,$ED,$C3
	.dc.b	$AC,$E5,$B3,$D4,$FE,$20,$53,$F6,$71,$BA,$EB,$19,$9F,$C0,$4F,$DE
	.dc.b	$BF,$91,$3E,$0C,$3E,$43,$4F,$81,$7C,$9F,$44,$FA,$B0,$F6,$47,$C8
	.dc.b	$69,$F0,$9D,$99,$AF,$E0,$75,$CE,$77,$1F,$18,$CD,$75,$6D,$71,$3F
	.dc.b	$82,$14,$FD,$64,$FF,$8F,$BE,$13,$FC,$52,$49,$FF,$2A,$B9,$7F,$C4
	.dc.b	$AE,$67,$FC,$A4,$E5,$FF,$12,$B9,$7F,$C2,$6F,$AE,$7C,$F1,$F4,$0F
	.dc.b	$6B,$99,$F3,$B6,$3C,$ED,$7C,$F1,$EF,$72,$FF,$9A,$9F,$F5,$EE,$3F
	.dc.b	$E5,$D6,$E1,$FE,$93,$6E,$1F,$EC,$27,$FC,$D4,$60,$C4,$F1,$7A,$10
	.dc.b	$D9,$4F,$F1,$79,$E0,$F8,$77,$E2,$9D,$61,$29,$CC,$76,$6B,$C6,$78
	.dc.b	$4E,$DB,$3A,$91,$27,$4F,$39,$D9,$1C,$10,$E3,$F7,$82,$00

;--------------------------------------------------------------------------------
;	FDADD.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fdadd_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FDADD.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FDADD.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fdadd_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdadd.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDADD.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdadd.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDADD.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fdadd_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3F,$D9,$BE,$53,$FB,$3B,$4A,$7F,$6C,$6B,$4A,$FF
	.dc.b	$70,$FA,$0F,$4C,$7F,$72,$F9,$BF,$E3,$DF,$05,$9D,$3A,$DD,$AD,$7B
	.dc.b	$5F,$A3,$C3,$8C,$7F,$56,$0E,$30,$E7,$18,$BE,$6D,$39,$E1,$D5,$E5
	.dc.b	$F3,$FC,$70,$FD,$83,$DA,$3D,$4E,$67,$FC,$12,$E6,$18,$F3,$C3,$C3
	.dc.b	$8E,$63,$18,$7A,$71,$F1,$CF,$3B,$F2,$1E,$B7,$07,$17,$37,$D2,$78
	.dc.b	$9F,$E1,$0C,$7D,$53,$E6,$8C,$2C,$79,$1D,$CF,$93,$F9,$B7,$F8,$3A
	.dc.b	$BE,$91,$FE,$23,$71,$2B,$FE,$24,$3F,$51,$7C,$4E,$2F,$87,$C1,$DB
	.dc.b	$FF,$16,$E1,$C9,$FF,$16,$B5,$DC,$FE,$22,$31,$1C,$3E,$0F,$F5,$4F
	.dc.b	$8D,$AE,$67,$84,$62,$12,$8F,$16,$3E,$7F,$6B,$C9,$5A,$9C,$E4,$9C
	.dc.b	$FC,$F7,$EE,$F1,$F8,$77,$FE,$46,$8C,$3F,$0E,$3F,$31,$4F,$6F,$8F
	.dc.b	$7C,$9F,$D1,$DF,$0B,$FF,$26,$B5,$FF,$21,$C6,$A7,$C7,$ED,$F1,$9E
	.dc.b	$C7,$D1,$C1,$1E,$AF,$93,$FF,$1A,$5F,$9A,$BF,$C1,$5E,$37,$B3,$E2
	.dc.b	$F4,$C9,$FD,$AD,$70,$AF,$DC,$B5,$FF,$30,$B5,$FC,$17,$B5,$F3,$78
	.dc.b	$0D,$25,$7D,$9F,$0F,$44,$70,$78,$3C,$0F,$AD,$81,$A4,$8F,$C1,$79
	.dc.b	$3D,$C8,$E0,$FC,$37,$89,$E3,$23,$83,$F8,$8F,$CE,$3E,$1F,$F0,$8D
	.dc.b	$8C,$C6,$0F,$67,$8D,$23,$F2,$4D,$D7,$F8,$EE,$BC,$5B,$5F,$0C,$95
	.dc.b	$FC,$80,$FF,$36,$17,$C0,$FC,$A3,$89,$FE,$4E,$D7,$F4,$7F,$0F,$F3
	.dc.b	$49,$1F,$A3,$7E,$87,$8F,$23,$F4,$CD,$9F,$E4,$BB,$FE,$87,$C1,$FE
	.dc.b	$9D,$5F,$CC,$9B,$3A,$8F,$D7,$6F,$F6,$C8,$E0,$FD,$97,$BB,$F7,$C8
	.dc.b	$E0,$FE,$7B,$F3,$7F,$E9,$F6,$F8,$FF,$B1,$8C,$5F,$F4,$46,$BF,$EA
	.dc.b	$CF,$CE,$3F,$CB,$D5,$FD,$19,$B1,$FD,$29,$AE,$F9,$3F,$79,$F5,$2F
	.dc.b	$4F,$F7,$2D,$BF,$DA,$F8,$3F,$C4,$FE,$3F,$35,$27,$FA,$9D,$6F,$0F
	.dc.b	$EA,$ED,$7F,$D7,$ED,$7F,$D8,$4E,$EF,$F8,$FF,$EA,$5F,$E5,$BA,$FD
	.dc.b	$74,$E5,$FE,$B1,$19,$AF,$DB,$D2,$3F,$D3,$D3,$E6,$6C,$E9,$FF,$45
	.dc.b	$46,$7E,$3F,$F8,$BC,$6A,$BC,$31,$8F,$8F,$CB,$AF,$84,$DF,$C7,$DC
	.dc.b	$3B,$37,$1E,$CD,$C3,$D6,$FF,$22,$F8,$5F,$E0,$51,$FE,$17,$6D,$7B
	.dc.b	$3C,$0D,$11,$FE,$FB,$1F,$C1,$6F,$FD,$D2,$D7,$FC,$91,$3F,$CC,$E3
	.dc.b	$9C,$7F,$49,$CF,$FC,$83,$5F,$C5,$E7,$37,$F1,$FE,$03,$E0,$F9,$52
	.dc.b	$BE,$47,$95,$71,$F9,$13,$6F,$EA,$CD,$AF,$A7,$F0,$FF,$2B,$E1,$FE
	.dc.b	$5B,$63,$E9,$37,$0F,$A4,$DB,$FF,$1C,$7C,$E7,$F9,$46,$BF,$96,$36
	.dc.b	$BE,$C7,$8D,$FD,$2A,$31,$5C,$3E,$0B,$FA,$94,$6E,$87,$FB,$72,$FF
	.dc.b	$C4,$B1,$98,$C1,$F8,$3E,$EE,$95,$F8,$27,$F0,$7E,$3F,$E1,$3C,$7B
	.dc.b	$87,$E2,$DB,$7F,$8B,$6E,$3F,$98,$7F,$17,$E3,$4A,$FC,$97,$8B,$6F
	.dc.b	$F2,$53,$6F,$F3,$F3,$6B,$FD,$C6,$7F,$4D,$F9,$BF,$EA,$2F,$E6,$FC
	.dc.b	$5F,$ED,$71,$8C,$3F,$AB,$87,$0F,$FB,$DF,$CA,$FE,$ED,$18,$8E,$A7
	.dc.b	$F9,$29,$E5,$87,$F8,$3B,$F1,$17,$E8,$9D,$AF,$FB,$DC,$63,$1F,$D6
	.dc.b	$79,$1F,$A8,$3F,$53,$E3,$D8,$C9,$32,$FD,$84,$65,$FD,$97,$8F,$FB
	.dc.b	$03,$FC,$0F,$E9,$7F,$14,$CF,$F8,$0C,$66,$FF,$74,$EB,$C5,$C5,$FC
	.dc.b	$0F,$9C,$FF,$54,$57,$F9,$89,$B1,$FE,$66,$6B,$F8,$43,$F0,$AE,$1F
	.dc.b	$F8,$03,$DB,$FE,$29,$FA,$47,$E2,$4F,$23,$FC,$33,$7D,$33,$FF,$62
	.dc.b	$56,$A7,$FC,$3B,$E4,$FF,$B7,$72,$FF,$02,$37,$2F,$E2,$BA,$FF,$90
	.dc.b	$E7,$3F,$1D,$C0,$FF,$15,$56,$EB,$F9,$C4,$60,$FF,$17,$64,$FD,$3A
	.dc.b	$92,$FE,$03,$6B,$FC,$33,$93,$8B,$FB,$56,$9F,$EA,$5C,$39,$61,$FE
	.dc.b	$2E,$AD,$73,$FB,$BC,$BE,$E4,$9F,$98,$DC,$7F,$1F,$F3,$BF,$AE,$9F
	.dc.b	$F1,$16,$1F,$CD,$67,$F6,$33,$97,$F2,$BA,$FF,$AA,$EB,$E2,$FC,$07
	.dc.b	$F9,$12,$B5,$9F,$FA,$E6,$B5,$7C,$2B,$E2,$92,$FF,$AD,$3F,$E5,$1B
	.dc.b	$FF,$22,$C7,$DD,$94,$61,$FC,$03,$5F,$2F,$AD,$FE,$FD,$9E,$13,$FE
	.dc.b	$4E,$F4,$7F,$D7,$75,$CB,$E6,$D5,$FE,$AD,$E1,$E7,$86,$2B,$FD,$23
	.dc.b	$1D,$AB,$17,$CB,$E6,$9E,$73,$E8,$F9,$61,$FC,$5A,$73,$18,$8F,$06
	.dc.b	$56,$73,$1D,$8E,$15,$CD,$C0,$8D,$DF,$AB,$36,$3E,$4C,$BF,$45,$5A
	.dc.b	$6E,$1C,$EF,$D0,$D8,$F6,$63,$E3,$8C,$63,$E2,$CB,$07,$A7,$2E,$F5
	.dc.b	$89,$F2,$E3,$DD,$E0,$7A,$27,$6E,$AF,$AB,$EB,$6C,$F1,$8F,$8A,$6E
	.dc.b	$FB,$63,$E8,$7C,$59,$7B,$E7,$BC,$E5,$E0,$78,$63,$4F,$79,$E7,$97
	.dc.b	$C4,$79,$E5,$89,$E7,$39,$79,$46,$5E,$47,$6B,$FA,$57,$DB,$0F,$51
	.dc.b	$31,$F7,$E5,$F0,$AB,$CC,$79,$EF,$F3,$95,$F8,$26,$D7,$E0,$5B,$97
	.dc.b	$E8,$5F,$C6,$79,$EE,$AC,$47,$E1,$63,$0E,$ED,$18,$8E,$ED,$7F,$0E
	.dc.b	$F0,$C3,$F3,$8D,$D3,$37,$F9,$7E,$DF,$E5,$D5,$A8,$FC,$D5,$6E,$FC
	.dc.b	$B1,$B7,$2C,$7F,$0F,$1B,$7F,$39,$5D,$2F,$BC,$6A,$B8,$5F,$EB,$EB
	.dc.b	$53,$D8,$D1,$23,$4F,$0A,$E7,$1E,$78,$FC,$BC,$F2,$7E,$45,$6D,$E9
	.dc.b	$3C,$9E,$95,$C8,$8F,$EF,$6B,$96,$1F,$3A,$F9,$4F,$43,$B9,$88,$D3
	.dc.b	$D0,$E4,$67,$0F,$89,$1A,$C3,$C4,$7A,$49,$87,$8A,$BE,$49,$FC,$2F
	.dc.b	$B7,$C7,$8C,$8C,$10,$FF,$80,$37,$FE,$C7,$7F,$9D,$1F,$C5,$E7,$FB
	.dc.b	$56,$DF,$94,$6C,$7F,$6D,$FC,$F3,$F9,$8B,$F9,$06,$BF,$95,$6B,$FC
	.dc.b	$DD,$AF,$1D,$BF,$BB,$71,$FF,$2D,$84,$F2,$8C,$36,$FF,$BE,$7C,$E3
	.dc.b	$E1,$38,$C6,$3D,$8B,$87,$F9,$12,$B4,$FF,$BE,$5A,$FF,$8A,$94,$70
	.dc.b	$84,$61,$FD,$5B,$5F,$D8,$35,$FE,$A4,$D7,$CB,$E1,$F8,$BA,$DF,$8B
	.dc.b	$BC,$2C,$36,$BC,$95,$E1,$BF,$F1,$9C,$78,$23,$F7,$6F,$91,$B1,$A8
	.dc.b	$F9,$DF,$84,$EB,$43,$49,$1F,$55,$F7,$46,$A3,$FC,$D3,$AF,$CE,$9F
	.dc.b	$92,$BC,$1F,$34,$FF,$94,$3E,$79,$F0,$9E,$AF,$8C,$93,$CB,$83,$FC
	.dc.b	$61,$5A,$8C,$1F,$E7,$67,$0F,$F1,$96,$BF,$DE,$DA,$F2,$AF,$83,$F3
	.dc.b	$8F,$9C,$FA,$CD,$9E,$A1,$C3,$FE,$66,$5F,$39,$C2,$64,$F0,$9D,$E7
	.dc.b	$F9,$17,$71,$D3,$F1,$87,$CD,$7E,$15,$B1,$98,$C6,$99,$64,$64,$E5
	.dc.b	$1F,$8D,$F9,$D7,$B3,$CF,$B6,$57,$AA,$72,$D9,$FC,$0C,$62,$3F,$A0
	.dc.b	$35,$FF,$11,$35,$FF,$30,$3B,$CB,$F5,$8D,$DF,$35,$EB,$F8,$BF,$D0
	.dc.b	$73,$3D,$CA,$4C,$7E,$6C,$3D,$CC,$47,$FA,$4A,$38,$3A,$9E,$73,$FA
	.dc.b	$BF,$C6,$3E,$6E,$3F,$E7,$37,$E6,$D7,$FA,$69,$B9,$FF,$01,$77,$1F
	.dc.b	$EB,$C6,$A7,$73,$CF,$8B,$B7,$17,$EE,$7C,$67,$CE,$FE,$EF,$27,$0D
	.dc.b	$35,$13,$2F,$EC,$F5,$C1,$B1,$FB,$08,$C4,$7F,$6F,$6E,$7F,$C5,$DB
	.dc.b	$97,$FA,$3D,$B3,$FE,$D4,$C2,$BF,$27,$B3,$C9,$C5,$23,$F3,$33,$FA
	.dc.b	$EC,$3F,$91,$37,$0F,$F4,$D3,$FE,$56,$8C,$D7,$B6,$FF,$8C,$7E,$31
	.dc.b	$F4,$B0,$FF,$62,$B6,$BF,$96,$37,$0F,$F5,$23,$F6,$E7,$F5,$2E,$5F
	.dc.b	$F7,$5B,$53,$4E,$0F,$FB,$4B,$EE,$9F,$0B,$FC,$A7,$D2,$FA,$B3,$FE
	.dc.b	$F4,$6A,$9D,$BF,$B8,$6B,$FE,$1B,$6B,$FE,$7D,$6B,$FE,$BB,$6B,$FF
	.dc.b	$00,$1E,$38,$7F,$49,$6E,$1B,$BF,$27,$D1,$FF,$3D,$F5,$BF,$73,$87
	.dc.b	$FB,$A9,$FC,$2B,$FC,$29,$AE,$7A,$78,$1C,$78,$64,$6E,$3A,$6B,$23
	.dc.b	$93,$DF,$6C,$47,$78,$BA,$7F,$5F,$C3,$EE,$67,$FE,$D7,$FB,$BF,$E9
	.dc.b	$D3,$FD,$8F,$C6,$F9,$ED,$E1,$FE,$AD,$3A,$AE,$1E,$37,$F4,$68,$FE
	.dc.b	$A5,$90,$38,$6C,$6C,$E6,$47,$F9,$5E,$0F,$84,$FE,$C9,$F8,$43,$E4
	.dc.b	$BF,$A9,$B6,$BF,$C9,$ED,$7F,$CA,$8D,$7F,$D6,$DF,$41,$FE,$00,$7C
	.dc.b	$8D,$4D,$31,$F8,$43,$D9,$FE,$C2,$D9,$FF,$2E,$B5,$FF,$31,$35,$FF
	.dc.b	$07,$7A,$DF,$D8,$C8,$FF,$85,$5F,$ED,$75,$E5,$8F,$8C,$78,$B0,$C9
	.dc.b	$1F,$F0,$E7,$AD,$FD,$B4,$8E,$0F,$F1,$07,$37,$F7,$12,$38,$3F,$CE
	.dc.b	$7F,$41,$E5,$7B,$B5,$FA,$CD,$4D,$F3,$3F,$AD,$C7,$F8,$A7,$BF,$FB
	.dc.b	$F1,$1F,$F1,$87,$81,$FE,$01,$23,$83,$FD,$07,$F4,$1C,$9F,$7B,$5F
	.dc.b	$C3,$35,$36,$9F,$43,$FE,$0D,$24,$FF,$8F,$DF,$F0,$AF,$E1,$1F,$CF
	.dc.b	$23,$FD,$E8,$D7,$E6,$B5,$FC,$A3,$57,$CF,$B7,$FF,$0F,$B6,$3F,$D3
	.dc.b	$6D,$7F,$D4,$0D,$7B,$E9,$E2,$23,$FE,$A2,$A4,$7F,$89,$7C,$F9,$3F
	.dc.b	$E5,$76,$BF,$EA,$BF,$9F,$FE,$1A,$9A,$4F,$07,$FC,$66,$FF,$0E,$8F
	.dc.b	$D7,$35,$E1,$1F,$E2,$A9,$FD,$EC,$6B,$D6,$FF,$27,$61,$FE,$22,$ED
	.dc.b	$FF,$30,$37,$B3,$F2,$78,$FF,$9C,$ED,$5F,$E6,$58,$C4,$7E,$60,$F3
	.dc.b	$5F,$90,$8F,$FB,$0A,$39,$4F,$19,$FF,$2F,$46,$2B,$EA,$FD,$4F,$F9
	.dc.b	$13,$E1,$7D,$1F,$45,$31,$FF,$16,$F0,$A7,$51,$FE,$7F,$7F,$CA,$04
	.dc.b	$7F,$D0,$5F,$33,$FE,$53,$23,$82,$1E,$CC,$3F,$D1,$51,$FE,$54,$3E
	.dc.b	$47,$CB,$FF,$9C,$FE,$83,$FC,$FD,$AF,$F1,$BF,$47,$FC,$93,$EB,$3E
	.dc.b	$AB,$F7,$31,$86,$C7,$EF,$2B,$8E,$4A,$FF,$28,$C6,$3E,$7F,$84,$FF
	.dc.b	$27,$6A,$77,$E2,$7D,$28,$C5,$F5,$8C,$4F,$02,$46,$71,$E8,$DA,$E2
	.dc.b	$7F,$35,$7D,$4E,$8C,$63,$FC,$52,$BC,$35,$99,$61,$EC,$C7,$CE,$4C
	.dc.b	$3B,$3C,$27,$B9,$FE,$78,$C3,$A4,$FF,$20,$3C,$31,$AA,$F8,$91,$E7
	.dc.b	$C3,$D7,$3E,$43,$C5,$3A,$7F,$A7,$3F,$39,$FC,$A3,$5F,$80,$63,$1F
	.dc.b	$94,$EB,$C5,$7B,$BF,$9B,$8E,$71,$87,$15,$FE,$B6,$9D,$47,$C2,$9D
	.dc.b	$38,$F8,$1F,$F5,$3A,$69,$1F,$F5,$5B,$5E,$06,$08,$60,$87,$95,$E0
	.dc.b	$D3,$65,$36,$7E,$D5,$DB,$4D,$94,$D9,$FD,$49,$DB,$4D,$94,$D9,$FD
	.dc.b	$F5,$DB,$4D,$94,$D9,$F0,$CC,$94,$CB,$C9,$C1,$EB,$21,$E7,$29,$E3
	.dc.b	$29,$DC,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C
	.dc.b	$11,$F6,$11,$F8,$2D,$7E,$09,$82,$18,$21,$EC,$78,$34,$D9,$4D,$9E
	.dc.b	$B7,$6D,$36,$53,$67,$B1,$DB,$4D,$94,$D9,$EF,$76,$D3,$65,$36,$7C
	.dc.b	$33,$25,$32,$F2,$70,$7A,$C8,$79,$CA,$78,$CA,$77,$29,$C8,$E0,$43
	.dc.b	$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$7D,$84,$7E,$0B,$5F
	.dc.b	$82,$60,$86,$08,$7B,$1E,$0D,$36,$53,$67,$AD,$DB,$4D,$94,$D9,$EC
	.dc.b	$76,$D3,$65,$36,$7B,$DD,$B4,$D9,$4D,$9F,$0C,$C9,$4C,$BC,$9C,$1E
	.dc.b	$B2,$1E,$72,$9E,$32,$9D,$CA,$72,$38,$10,$E0,$53,$81,$4E,$06,$88
	.dc.b	$68,$A6,$8C,$10,$C1,$1F,$61,$1F,$82,$D7,$E0,$98,$21,$82,$1E,$C7
	.dc.b	$83,$4D,$94,$D9,$EB,$76,$D3,$65,$36,$7B,$1D,$B4,$D9,$4D,$9E,$F7
	.dc.b	$6D,$36,$53,$67,$C3,$32,$53,$2F,$27,$07,$AC,$87,$9C,$A7,$8C,$A7
	.dc.b	$72,$9C,$8E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$47
	.dc.b	$D8,$47,$E0,$B5,$F8,$26,$0F,$9F,$5F,$E0,$F5,$D0,$F5,$FF,$C9,$97
	.dc.b	$FE,$15,$8C,$4F,$F8,$63,$0F,$F7,$84,$62,$3F,$C2,$D1,$8A,$E2,$FF
	.dc.b	$86,$AF,$B4,$FF,$88,$E3,$17,$FE,$1C,$8F,$97,$87,$F9,$07,$67,$D0
	.dc.b	$38,$BF,$C8,$95,$FE,$01,$D7,$FD,$19,$39,$8F,$F2,$09,$9D,$BF,$DB
	.dc.b	$F7,$CA,$F8,$CF,$F8,$16,$F8,$C6,$30,$F0,$61,$CE,$58,$7F,$88,$5D
	.dc.b	$DF,$63,$94,$76,$9E,$0E,$CC,$5F,$59,$FF,$2A,$E1,$E3,$79,$61,$E0
	.dc.b	$9D,$61,$D8,$F0,$CF,$F8,$AD,$C5,$F5,$81,$FE,$A8,$C3,$C2,$EE,$3C
	.dc.b	$2F,$63,$C3,$1A,$7F,$C6,$13,$D1,$E0,$74,$C3,$CD,$1E,$28,$DD,$74
	.dc.b	$8C,$47,$CF,$38,$57,$57,$07,$96,$FD,$B5,$C4,$F5,$46,$4F,$35,$78
	.dc.b	$DE,$E7,$49,$E2,$D3,$73,$CD,$F4,$9C,$E7,$CA,$FD,$63,$83,$8C,$3E
	.dc.b	$84,$72,$33,$3E,$4C,$3E,$0B,$70,$F6,$4F,$33,$B4,$67,$0F,$2C,$60
	.dc.b	$F1,$D7,$26,$D7,$2B,$F0,$E1,$E3,$9F,$AB,$7C,$AF,$84,$7C,$23,$A1
	.dc.b	$5C,$BF,$39,$AF,$95,$B5,$DB,$0F,$34,$FB,$CE,$67,$1C,$3C,$8F,$88
	.dc.b	$D5,$72,$71,$87,$9C,$FA,$CF,$1C,$3D,$15,$CA,$73,$3E,$38,$C5,$FD
	.dc.b	$78,$F5,$46,$1F,$A2,$7C,$6A,$E6,$47,$C7,$1B,$7D,$27,$58,$D5,$79
	.dc.b	$6B,$B4,$F9,$4E,$33,$A8,$F8,$73,$B9,$F0,$9E,$49,$D3,$E3,$38,$92
	.dc.b	$38,$3E,$2C,$3E,$73,$B8,$CE,$1F,$25,$F8,$73,$A9,$EC,$73,$BF,$19
	.dc.b	$1E,$4D,$77,$3D,$30,$F8,$57,$D6,$BD,$F5,$F0,$CE,$87,$22,$1C,$9E
	.dc.b	$8F,$C5,$AE,$98,$7C,$08,$D1,$C0,$D4,$F0,$7B,$4F,$07,$B9,$C0,$87
	.dc.b	$03,$91,$C2,$39,$61,$F0,$E3,$47,$03,$47,$07,$4E,$0D,$38,$23,$F9
	.dc.b	$68,$EA,$D3,$91,$4E,$25,$3C,$24,$3B,$16,$3A,$CF,$DA,$8E,$06,$89
	.dc.b	$1A,$9F,$A6,$D8,$FC,$69,$CC,$E2,$E5,$FC,$51,$82,$4F,$72,$4F,$48
	.dc.b	$CB,$63,$31,$D8,$F1,$9A,$31,$3D,$EF,$F1,$11,$83,$47,$8A,$79,$B6
	.dc.b	$BB,$11,$E1,$1A,$9E,$71,$B3,$33,$B7,$04,$8D,$93,$0F,$CD,$93,$0F
	.dc.b	$C9,$4F,$C7,$BE,$D3,$98,$FC,$79,$9B,$ED,$7C,$EF,$8C,$FC,$2B,$E3
	.dc.b	$18,$C3,$E1,$61,$CE,$7F,$34,$4C,$3F,$12,$F0,$BE,$E7,$38,$EF,$3C
	.dc.b	$5D,$98,$BE,$D3,$F3,$30,$F2,$3C,$B0,$F0,$CE,$B0,$EC,$78,$67,$F2
	.dc.b	$EE,$2F,$AC,$7C,$C2,$1F,$2F,$0F,$13,$C2,$3C,$4F,$73,$C5,$1A,$7F
	.dc.b	$43,$3D,$5E,$07,$4C,$3C,$F1,$E3,$8D,$D7,$48,$C4,$7D,$13,$85,$75
	.dc.b	$70,$79,$6F,$DD,$5C,$4F,$5C,$64,$F3,$57,$8D,$EE,$74,$9E,$2D,$37
	.dc.b	$3C,$DF,$49,$CE,$7C,$AF,$D7,$38,$38,$C3,$E8,$C7,$23,$33,$E4,$C3
	.dc.b	$E0,$B7,$0F,$64,$F3,$3B,$46,$70,$F2,$C6,$0F,$1D,$72,$6D,$72,$BF
	.dc.b	$0E,$1E,$39,$FA,$B7,$CA,$F8,$47,$C2,$3A,$15,$CB,$F3,$9A,$F9,$5F
	.dc.b	$EF,$1C,$1F,$E5,$83,$FC,$1C,$5C,$7D,$13,$F0,$4F,$E2,$31,$97,$FC
	.dc.b	$9A,$7F,$93,$B0,$FF,$27,$C6,$61,$F0,$2F,$83,$7C,$27,$09,$E3,$39
	.dc.b	$8C,$47,$F9,$62,$1E,$BB,$E1,$73,$3D,$2F,$9B,$C3,$85,$C8,$D7,$FD
	.dc.b	$18,$D9,$D3,$5F,$14,$F0,$9F,$F4,$7F,$14,$9E,$9C,$5F,$E9,$17,$FD
	.dc.b	$38,$7F,$A9,$78,$7F,$24,$FC,$DE,$2F,$B0,$FD,$63,$E4,$5F,$D0,$E0
	.dc.b	$FB,$7C,$3F,$32,$FE,$E1,$CA,$BE,$A5,$F2,$8C,$57,$09,$FC,$2F,$17
	.dc.b	$E0,$6B,$57,$F9,$D9,$FC,$04,$61,$FC,$B9,$C0,$F1,$B9,$7F,$0E,$7A
	.dc.b	$67,$A1,$1D,$B6,$FC,$F5,$DE,$78,$1A,$24,$69,$F0,$9E,$49,$E5,$3F
	.dc.b	$04,$91,$E5,$9E,$0F,$90,$F8,$73,$C1,$D4,$F0,$7A,$1E,$07,$0F,$92
	.dc.b	$79,$E1,$F4,$63,$EA,$46,$2F,$E1,$4E,$63,$E7,$CE,$AB,$9B,$E2,$3B
	.dc.b	$1E,$22,$1E,$28,$E9,$7E,$78,$C4,$F2,$C3,$ED,$46,$0E,$27,$9E,$34
	.dc.b	$F8,$4C,$9E,$1A,$EA,$F8,$8E,$44,$39,$1C,$27,$BB,$83,$47,$39,$EF
	.dc.b	$3F,$00,$8F,$DD,$8D,$1C,$CF,$21,$CA,$7A,$BA,$30,$68,$F0,$10,$D1
	.dc.b	$5E,$24,$78,$1A,$23,$99,$F2,$C7,$E3,$0E,$47,$08,$D4,$FA,$5B,$1B
	.dc.b	$39,$9B,$24,$F4,$7F,$21,$1C,$1A,$6A,$7A,$34,$E9,$1B,$BF,$C5,$4E
	.dc.b	$AB,$A4,$62,$B6,$71,$3A,$C6,$B0,$FC,$99,$30,$FC,$54,$FB,$A7,$51
	.dc.b	$88,$F5,$C6,$1F,$C2,$1B,$BF,$11,$1F,$0C,$6C,$FC,$59,$23,$F0,$66
	.dc.b	$E3,$CA,$DB,$F5,$46,$0E,$71,$9B,$F4,$47,$88,$F2,$C6,$6F,$D1,$87
	.dc.b	$C7,$AD,$E1,$E1,$BF,$F4,$8F,$07,$F4,$37,$FA,$1B,$5F,$E9,$0F,$F4
	.dc.b	$4D,$FF,$A1,$B9,$8F,$EA,$7A,$FF,$3B,$5C,$4C,$FF,$AB,$8E,$5F,$EB
	.dc.b	$31,$99,$1B,$8C,$CF,$4E,$0F,$E9,$55,$BA,$5F,$77,$FA,$5C,$FF,$4D
	.dc.b	$78,$6D,$FD,$48,$FE,$98,$FF,$50,$DF,$FA,$63,$C9,$B4,$73,$1F,$DB
	.dc.b	$4F,$EC,$23,$6A,$D3,$2E,$0F,$9D,$BF,$F9,$3B,$1F,$EE,$93,$CD,$AF
	.dc.b	$D6,$3E,$D9,$0F,$BC,$4C,$3F,$02,$4D,$FE,$CE,$7F,$88,$9E,$13,$9B
	.dc.b	$FB,$F1,$F9,$F8,$C3,$F8,$5C,$3C,$AD,$35,$7D,$1F,$E2,$1B,$FE,$5F
	.dc.b	$0E,$D1,$E3,$9D,$3F,$C4,$27,$4F,$09,$FC,$83,$D6,$32,$EE,$33,$1F
	.dc.b	$91,$8C,$1E,$A9,$EA,$FE,$AC,$C6,$1E,$C9,$E7,$19,$3C,$91,$9A,$EA
	.dc.b	$7A,$30,$F8,$35,$DE,$B8,$46,$27,$57,$E0,$9F,$A8,$74,$3D,$51,$D1
	.dc.b	$F7,$1E,$43,$25,$AE,$CF,$2A,$EC,$F0,$3A,$46,$B0,$FA,$6D,$8F,$E4
	.dc.b	$D8,$FE,$1A,$7F,$C7,$B7,$FE,$40,$B6,$0C,$73,$87,$F8,$F6,$EB,$FE
	.dc.b	$33,$9D,$3F,$B4,$BF,$F1,$FC,$EA,$F9,$DF,$0A,$F8,$B7,$F0,$EB,$CB
	.dc.b	$7F,$1A,$BE,$75,$7B,$E3,$0F,$C2,$3D,$AE,$31,$F6,$1F,$62,$3E,$25
	.dc.b	$F5,$7F,$23,$5D,$1E,$27,$C7,$C3,$E4,$BF,$86,$32,$57,$2F,$A2,$39
	.dc.b	$46,$23,$BC,$62,$7C,$15,$E3,$C3,$DB,$5D,$E7,$38,$7D,$F8,$ED,$18
	.dc.b	$7E,$01,$BB,$E8,$47,$E1,$61,$E1,$BE,$B1,$CA,$75,$87,$C4,$C3,$B4
	.dc.b	$FA,$4E,$38,$77,$7D,$E6,$67,$83,$E2,$3A,$1C,$2B,$AD,$7D,$13,$EF
	.dc.b	$C2,$7F,$95,$61,$FE,$92,$9F,$C6,$5F,$E1,$B1,$FF,$49,$C6,$30,$FC
	.dc.b	$26,$1F,$C7,$E7,$F2,$8F,$E2,$CF,$E7,$37,$F8,$A7,$07,$FA,$1F,$8F
	.dc.b	$FA,$3F,$17,$FA,$22,$3F,$A5,$F1,$FF,$A3,$5F,$E9,$F1,$97,$F9,$C9
	.dc.b	$E7,$C7,$FA,$4F,$07,$FA,$83,$4F,$E8,$51,$FD,$2E,$7F,$A4,$C7,$F4
	.dc.b	$5D,$FF,$A9,$37,$83,$FD,$24,$F0,$BE,$57,$D2,$F8,$BE,$09,$E1,$9F
	.dc.b	$F4,$99,$FF,$4C,$B5,$FE,$95,$C7,$FD,$31,$FE,$C3,$C3,$FD,$28,$FE
	.dc.b	$92,$FF,$45,$3F,$D5,$3C,$3F,$D1,$5F,$E8,$E4,$BF,$C9,$3F,$A2,$3E
	.dc.b	$D1,$1F,$BE,$FE,$08,$FC,$D9,$9C,$3E,$F3,$F9,$33,$05,$8E,$38,$7E
	.dc.b	$4E,$3B,$46,$38,$7F,$1D,$39,$8F,$19,$2B,$F2,$B5,$E0,$8E,$1C,$3F
	.dc.b	$8E,$8E,$D3,$E2,$7F,$27,$3A,$7C,$66,$23,$4F,$E6,$DA,$F5,$6E,$1F
	.dc.b	$A8,$8D,$1D,$88,$73,$29,$A3,$04,$30,$43,$B1,$C0,$A6,$CA,$6D,$F1
	.dc.b	$38,$23,$82,$3F,$32,$3D,$2D,$3B,$46,$A7,$E6,$47,$EE,$4E,$C6,$CF
	.dc.b	$14,$6E,$7E,$8B,$63,$81,$E2,$38,$1C,$A7,$AB,$F8,$78,$E0,$D3,$67
	.dc.b	$18,$DD,$FE,$06,$BB,$CE,$AB,$8C,$E6,$3C,$31,$8C,$3F,$12,$4C,$3F
	.dc.b	$09,$3E,$A3,$91,$30,$EF,$3B,$8F,$C2,$9E,$9C,$3C,$33,$AA,$E7,$39
	.dc.b	$3C,$27,$39,$D9,$1C,$CF,$17,$A9,$FE,$92,$E1,$FE,$32,$7F,$22,$BF
	.dc.b	$E4,$15,$FE,$52,$9F,$E3,$07,$F1,$4E,$1F,$F2,$FF,$0F,$F1,$78,$FF
	.dc.b	$31,$F0,$FF,$17,$7F,$CB,$3C,$3F,$E6,$1C,$7F,$87,$47,$F7,$2C,$3F
	.dc.b	$8B,$3F,$C9,$27,$51,$FC,$3C,$FE,$25,$1F,$C7,$CF,$F3,$1B,$C3,$2F
	.dc.b	$E2,$71,$93,$FC,$C3,$C3,$FC,$72,$78,$CF,$82,$75,$5F,$C8,$A7,$F9
	.dc.b	$5C,$7F,$99,$B6,$FE,$34,$E0,$FE,$33,$C1,$FC,$6D,$F5,$F0,$7F,$1B
	.dc.b	$8F,$E3,$11,$FC,$68,$FE,$32,$53,$18,$7C,$F3,$F8,$01,$2F,$F4,$6E
	.dc.b	$5A,$7D,$78,$FC,$A4,$65,$FD,$39,$F7,$4F,$B7,$1B,$BF,$C9,$47,$08
	.dc.b	$C6,$DF,$8B,$AF,$04,$F2,$7A,$3F,$93,$6B,$F8,$B3,$91,$D8,$86,$09
	.dc.b	$5E,$02,$4F,$7A,$F5,$CF,$09,$CE,$FF,$98,$8F,$D7,$4F,$9A,$31,$C1
	.dc.b	$FA,$96,$CF,$BC,$EF,$7D,$5F,$55,$71,$7F,$37,$5C,$5A,$71,$BF,$0B
	.dc.b	$8C,$3D,$EF,$88,$F3,$5F,$57,$E2,$5F,$59,$F9,$31,$83,$78,$79,$A3
	.dc.b	$9E,$1E,$17,$73,$F2,$F0,$F0,$C6,$23,$C0,$7B,$2B,$95,$F7,$3E,$4C
	.dc.b	$64,$F4,$39,$7E,$09,$E2,$32,$59,$ED,$7E,$39,$E7,$7E,$26,$E1,$EF
	.dc.b	$BF,$33,$5F,$2C,$ED,$EE,$78,$EF,$AB,$F3,$67,$85,$F7,$8D,$D7,$17
	.dc.b	$FA,$8F,$17,$FB,$95,$FF,$6E,$B5,$3F,$EE,$B7,$F8,$71,$FE,$0E,$3F
	.dc.b	$C5,$38,$7F,$08,$DB,$F7,$87,$F0,$22,$3F,$C2,$1C,$BF,$E2,$87,$F8
	.dc.b	$3E,$3F,$C0,$C9,$1F,$C3,$E7,$F8,$BE,$DF,$EE,$C8,$E4,$48,$CE,$1F
	.dc.b	$E8,$AD,$BF,$DD,$D1,$FC,$1C,$C9,$E7,$3F,$C4,$67,$F2,$93,$F8,$4E
	.dc.b	$BF,$C2,$DF,$F7,$73,$FC,$29,$FF,$29,$9F,$C2,$4A,$7D,$6D,$7F,$0A
	.dc.b	$FD,$DB,$FF,$06,$5F,$F9,$34,$FF,$4D,$C7,$FA,$76,$7F,$CA,$27,$F9
	.dc.b	$7B,$1F,$E7,$8E,$23,$FC,$AB,$AF,$FA,$8A,$B8,$9C,$E7,$85,$FF,$97
	.dc.b	$E9,$3A,$AE,$15,$98,$F0,$C8,$6B,$FE,$4A,$7B,$CF,$85,$F3,$1D,$1E
	.dc.b	$47,$81,$D9,$92,$CF,$86,$39,$47,$79,$EC,$F4,$9D,$3C,$4C,$4F,$56
	.dc.b	$9D,$1F,$0B,$FE,$C7,$3F,$A6,$47,$C3,$74,$43,$FB,$24,$F8,$23,$51
	.dc.b	$D0,$87,$94,$E7,$5F,$21,$E5,$1C,$1D,$61,$F5,$9F,$44,$74,$34,$FC
	.dc.b	$C3,$9D,$FC,$B8,$D5,$75,$9F,$29,$A3,$9B,$98,$E7,$7E,$98,$C1,$2B
	.dc.b	$E6,$D7,$7B,$F2,$5F,$CC,$24,$FC,$B7,$C2,$F8,$EB,$C5,$5E,$CC,$3E
	.dc.b	$6B,$F1,$6B,$C2,$F9,$5E,$6F,$CD,$38,$1F,$1A,$7C,$8F,$DB,$32,$73
	.dc.b	$73,$1F,$92,$8F,$44,$F8,$23,$CB,$1C,$4C,$9D,$6B,$D0,$7C,$0C,$3E
	.dc.b	$23,$88,$F9,$2E,$5F,$43,$D6,$FD,$F1,$D0,$F2,$47,$1A,$F2,$57,$49
	.dc.b	$E6,$48,$CB,$D0,$F2,$1E,$78,$E1,$1F,$1A,$3F,$1A,$77,$7D,$A6,$5F
	.dc.b	$8A,$E2,$7E,$8C,$62,$BE,$3C,$62,$39,$46,$1F,$A4,$62,$3D,$46,$EB
	.dc.b	$C6,$FC,$13,$E2,$61,$F8,$67,$13,$E9,$3D,$87,$C4,$31,$1B,$7D,$F7
	.dc.b	$F0,$A3,$11,$B8,$C6,$1D,$1F,$74,$79,$4F,$95,$5E,$57,$BC,$7C,$17
	.dc.b	$F7,$95,$D2,$31,$1F,$A6,$8C,$61,$F0,$63,$17,$DE,$FA,$56,$9F,$87
	.dc.b	$3D,$27,$4F,$C8,$C3,$E2,$BE,$C9,$FD,$35,$F5,$C3,$F4,$38,$73,$AF
	.dc.b	$49,$CC,$E0,$73,$3D,$73,$E5,$7E,$D5,$78,$9F,$C1,$DF,$81,$F9,$11
	.dc.b	$E0,$35,$7E,$0C,$3C,$F3,$C0,$EB,$3C,$DE,$A7,$90,$D6,$1E,$7A,$FA
	.dc.b	$37,$DD,$F9,$73,$C1,$FC,$B1,$DD,$F5,$9C,$AF,$C3,$19,$8F,$39,$EF
	.dc.b	$AE,$D7,$F4,$2B,$78,$7E,$BB,$0F,$03,$F0,$EF,$F0,$58,$78,$23,$51
	.dc.b	$EC,$3D,$47,$96,$FE,$0B,$4F,$29,$4F,$19,$F8,$23,$D3,$3E,$73,$F1
	.dc.b	$25,$9F,$B8,$78,$30,$F4,$CF,$AF,$0F,$4D,$FE,$99,$B1,$FA,$26,$DF
	.dc.b	$E7,$A3,$BD,$F3,$9C,$D7,$DB,$9F,$A9,$5E,$18,$FB,$B5,$B9,$E3,$1D
	.dc.b	$DA,$77,$2C,$F7,$7D,$EF,$37,$AC,$F3,$72,$D8,$F8,$2D,$3E,$0C,$F7
	.dc.b	$9D,$B6,$76,$D9,$DC,$F8,$CD,$96,$76,$D8,$DC,$79,$CC,$95,$CC,$FC
	.dc.b	$58,$FC,$13,$DC,$F8,$A5,$9F,$6B,$63,$71,$DE,$36,$D3,$65,$8D,$9E
	.dc.b	$18,$DB,$63,$6D,$37,$1E,$53,$F2,$31,$F9,$86,$E1,$F9,$C7,$F6,$51
	.dc.b	$93,$F2,$A5,$3F,$2A,$7E,$5E,$7F,$62,$E5,$B3,$CE,$3F,$42,$66,$BF
	.dc.b	$51,$19,$9F,$DD,$CF,$E9,$DB,$87,$E6,$A3,$F7,$07,$E2,$2F,$D4,$F9
	.dc.b	$2B,$94,$7E,$DE,$34,$7F,$1B,$3E,$ED,$7C,$38,$F7,$C7,$77,$83,$A7
	.dc.b	$AB,$EE,$3D,$0F,$C4,$9E,$F1,$FA,$D3,$B9,$FB,$73,$91,$E1,$9F,$C3
	.dc.b	$CF,$47,$E2,$C7,$84,$E4,$70,$39,$3C,$1E,$57,$F8,$E6,$CF,$E0,$0E
	.dc.b	$91,$F3,$23,$A1,$F2,$0E,$87,$D7,$75,$7F,$36,$BF,$60,$7E,$1A,$7A
	.dc.b	$9F,$8A,$AE,$A7,$C1,$7A,$C7,$08,$F9,$73,$CE,$BF,$69,$1E,$5A,$E7
	.dc.b	$5D,$5F,$6C,$FA,$DB,$3B,$69,$E1,$38,$BF,$E7,$D5,$F4,$EF,$92,$FF
	.dc.b	$18,$46,$1B,$19,$9F,$F5,$14,$FF,$A1,$63,$0F,$FA,$F8,$FF,$53,$6F
	.dc.b	$FE,$C3,$3F,$D8,$4E,$38,$5C,$7C,$A1,$BB,$83,$FD,$2B,$0D,$7F,$D7
	.dc.b	$AF,$FA,$4F,$77,$27,$A6,$7F,$D8,$71,$FE,$C1,$93,$A7,$FD,$16,$D7
	.dc.b	$A3,$BB,$F0,$6F,$AB,$E9,$3F,$14,$E1,$7D,$AF,$FD,$8A,$F9,$1E,$F3
	.dc.b	$C1,$CC,$F8,$CE,$27,$38,$D4,$7C,$0C,$3D,$F5,$E8,$8F,$F6,$89,$FE
	.dc.b	$CF,$C3,$FD,$AA,$E0,$FF,$6B,$1F,$ED,$03,$E9,$AE,$37,$17,$FB,$56
	.dc.b	$30,$FF,$B5,$67,$FD,$48,$7F,$B5,$4A,$7F,$B5,$4F,$8A,$7F,$C0,$0E
	.dc.b	$17,$07,$FB,$62,$3F,$D7,$F8,$79,$DF,$F4,$AF,$1F,$FA,$7C,$C3,$FE
	.dc.b	$A7,$71,$A7,$E7,$0F,$EC,$31,$F9,$23,$F8,$81,$FA,$2B,$FC,$FB,$F9
	.dc.b	$73,$F3,$3A,$FE,$7D,$FD,$86,$1F,$A9,$C3,$F4,$CF,$F4,$13,$B4,$7F
	.dc.b	$0A,$36,$7F,$28,$E3,$FE,$1E,$FE,$EE,$BF,$82,$3D,$A7,$8B,$FB,$E3
	.dc.b	$F8,$B9,$C4,$F3,$3D,$DF,$4D,$78,$9F,$E9,$91,$CA,$7A,$C7,$89,$FE
	.dc.b	$67,$87,$C2,$8F,$01,$C9,$C3,$B9,$FE,$57,$3C,$A3,$C6,$F2,$7D,$C6
	.dc.b	$23,$78,$7F,$03,$8E,$47,$F2,$49,$E5,$87,$C2,$7F,$88,$47,$84,$E7
	.dc.b	$19,$9F,$A7,$3F,$46,$7B,$1C,$E3,$B3,$83,$E6,$38,$9E,$55,$F1,$E7
	.dc.b	$FA,$14,$61,$F8,$E7,$84,$FA,$0E,$6B,$FA,$74,$65,$F4,$BF,$48,$E2
	.dc.b	$74,$30,$F0,$71,$87,$D5,$8C,$1F,$D2,$1C,$57,$F3,$49,$FB,$D1,$87
	.dc.b	$F2,$C7,$2C,$3E,$AB,$EF,$AF,$92,$E4,$F9,$27,$8A,$BA,$47,$D7,$8F
	.dc.b	$69,$93,$AB,$E3,$9E,$B7,$F0,$6B,$89,$F7,$0F,$E7,$95,$F7,$0F,$9E
	.dc.b	$E1,$F2,$61,$F6,$23,$F2,$4F,$FB,$C2,$3F,$A0,$0B,$47,$F4,$06,$CF
	.dc.b	$F3,$56,$CF,$F4,$06,$BC,$9A,$FF,$4B,$6A,$3B,$4F,$53,$F8,$5B,$E1
	.dc.b	$9E,$1F,$11,$FE,$AB,$23,$FE,$9E,$6C,$7F,$99,$9B,$5F,$E9,$46,$CF
	.dc.b	$86,$BA,$CF,$6A,$F7,$56,$9F,$24,$7B,$0D,$08,$D4,$7A,$2B,$67,$98
	.dc.b	$F2,$11,$F2,$36,$7C,$8D,$3C,$87,$28,$F0,$CF,$69,$F1,$4F,$59,$E6
	.dc.b	$48,$E6,$D8,$E6,$FC,$93,$9B,$DD,$F6,$D7,$C2,$8E,$93,$C8,$8F,$27
	.dc.b	$E0,$D7,$8E,$78,$57,$43,$C0,$74,$AF,$53,$F3,$1F,$03,$F1,$CF,$21
	.dc.b	$93,$89,$9A,$EF,$1C,$1F,$83,$1D,$DF,$2B,$DC,$E6,$FD,$58,$E1,$3E
	.dc.b	$69,$ED,$3C,$1E,$D3,$F0,$0F,$71,$5E,$91,$E2,$7C,$AF,$89,$E7,$1C
	.dc.b	$1F,$23,$5C,$9C,$9E,$8F,$80,$D9,$92,$99,$38,$96,$3D,$8F,$8A,$7D
	.dc.b	$91,$83,$D8,$79,$4C,$4F,$8D,$B3,$DC,$EA,$EE,$BE,$E4,$FD,$88,$DC
	.dc.b	$F8,$67,$2F,$CE,$6C,$7C,$53,$C3,$5E,$48,$CD,$7C,$09,$EC,$FD,$43
	.dc.b	$B4,$6A,$BE,$0C,$E9,$F7,$4E,$8E,$0F,$85,$B1,$E1,$7D,$D3,$CE,$7D
	.dc.b	$B5,$E3,$9D,$9E,$37,$11,$D8,$E8,$5A,$E8,$FC,$77,$E1,$9E,$53,$DF
	.dc.b	$3E,$58,$E0,$F5,$31,$1A,$8E,$AD,$3A,$BF,$36,$3C,$AF,$C2,$9F,$2C
	.dc.b	$EE,$7A,$4F,$27,$DE,$70,$7D,$EF,$80,$CB,$E4,$3A,$39,$7D,$95,$F1
	.dc.b	$E7,$DB,$5E,$0A,$E3,$1E,$1A,$E2,$78,$4E,$31,$E1,$8E,$2F,$58,$F6
	.dc.b	$1E,$13,$CA,$71,$33,$1D,$5C,$BF,$10,$F9,$D1,$E2,$24,$78,$0E,$91
	.dc.b	$E0,$7A,$71,$7F,$A2,$F8,$BF,$D1,$31,$FE,$92,$8F,$F4,$6B,$FE,$9B
	.dc.b	$51,$A8,$FE,$BF,$C6,$F8,$BF,$F4,$84,$7F,$A8,$1B,$19,$8F,$F4,$A3
	.dc.b	$5F,$F3,$6F,$17,$F9,$DB,$5F,$F3,$D4,$7F,$A1,$CF,$EB,$DB,$7F,$A1
	.dc.b	$5F,$F3,$E6,$59,$7F,$CD,$67,$FA,$18,$DC,$FF,$A4,$6B,$A4,$7F,$A7
	.dc.b	$78,$FF,$D0,$B3,$FE,$88,$8F,$F4,$59,$0E,$47,$47,$4F,$80,$EF,$C5
	.dc.b	$FE,$9C,$75,$3D,$0F,$F4,$8C,$67,$0F,$F3,$D3,$FE,$8E,$9E,$47,$FA
	.dc.b	$46,$FF,$D6,$6D,$C3,$FD,$19,$C7,$FE,$8E,$8F,$F4,$8B,$5F,$F4,$C3
	.dc.b	$5F,$F6,$3F,$0F,$FA,$45,$FF,$68,$35,$FF,$03,$3F,$11,$FE,$81,$87
	.dc.b	$FA,$29,$46,$5F,$F4,$8E,$5F,$E9,$79,$D4,$E6,$FF,$C2,$F1,$8A,$FA
	.dc.b	$31,$8A,$F1,$C6,$38,$3F,$C2,$75,$B3,$FD,$39,$5B,$9F,$F2,$84,$FF
	.dc.b	$96,$DE,$1C,$1F,$D8,$BE,$75,$F0,$1F,$D4,$DB,$1F,$DF,$ED,$F3,$EF
	.dc.b	$1B,$57,$2B,$93,$FD,$C2,$B7,$39,$C3,$FB,$5F,$D0,$7F,$B8,$2F,$FB
	.dc.b	$8B,$63,$27,$F6,$37,$15,$FE,$21,$6C,$FF,$BF,$DB,$3F,$E3,$57,$FC
	.dc.b	$6D,$1D,$38,$79,$11,$EF,$34,$F1,$C5,$AE,$D1,$67,$B4,$D8,$EC,$51
	.dc.b	$5E,$26,$C7,$A5,$D4,$64,$8E,$23,$2D,$9F,$41,$E6,$2B,$E6,$6D,$79
	.dc.b	$9A,$79,$8B,$1E,$66,$BE,$46,$D7,$C6,$8E,$84,$9F,$9D,$5E,$9A,$DB
	.dc.b	$E6,$69,$CA,$7D,$07,$17,$EA,$38,$9E,$71,$EB,$8D,$1F,$5E,$7A,$38
	.dc.b	$3C,$C5,$AF,$35,$72,$8C,$4F,$95,$B5,$E3,$3C,$AF,$13,$C4,$FC,$57
	.dc.b	$F1,$AE,$A7,$A1,$F2,$E3,$99,$C6,$34,$71,$8D,$1C,$1C,$12,$3E,$81
	.dc.b	$CA,$34,$E1,$AF,$B9,$B5,$EE,$9E,$53,$96,$C6,$63,$E4,$47,$91,$E8
	.dc.b	$EA,$BE,$5D,$74,$6C,$64,$F8,$D3,$E5,$7A,$BF,$18,$EA,$7C,$63,$46
	.dc.b	$27,$E0,$C7,$38,$F8,$C6,$8F,$41,$E4,$38,$3B,$8E,$31,$82,$1C,$0F
	.dc.b	$D5,$9E,$33,$15,$F0,$5B,$5E,$C7,$8C,$7E,$B6,$3F,$09,$3F,$06,$BD
	.dc.b	$ED,$AF,$3B,$4F,$A2,$F7,$AF,$DA,$CF,$CD,$8F,$09,$9A,$EB,$1E,$B3
	.dc.b	$E4,$9B,$9E,$33,$F0,$DB,$5D,$E7,$73,$9A,$E6,$FA,$9F,$63,$EE,$8F
	.dc.b	$49,$DD,$F7,$9E,$37,$F5,$93,$D6,$33,$1E,$39,$EB,$1C,$CF,$83,$1C
	.dc.b	$DF,$6B,$B3,$99,$9A,$F4,$B6,$79,$C7,$CF,$8F,$2C,$FA,$5A,$F7,$AF
	.dc.b	$9B,$1D,$63,$F1,$91,$E7,$7C,$27,$C3,$8F,$0B,$CC,$F3,$96,$39,$BB
	.dc.b	$39,$C7,$94,$E2,$F8,$5F,$9F,$1D,$E7,$D7,$3C,$E7,$27,$17,$99,$93
	.dc.b	$E5,$9D,$E7,$F3,$2F,$53,$2F,$5A,$F8,$A7,$E2,$E3,$EB,$1D,$0E,$07
	.dc.b	$C0,$8E,$13,$E8,$AF,$99,$3F,$00,$DB,$E8,$6D,$76,$9F,$C8,$47,$6A
	.dc.b	$F2,$D7,$27,$E3,$9D,$A3,$C5,$19,$9F,$83,$1E,$F3,$C8,$FC,$39,$E7
	.dc.b	$5E,$B8,$FB,$51,$87,$C2,$60,$FE,$82,$59,$FE,$04,$7A,$1F,$74,$F3
	.dc.b	$8F,$99,$1C,$1A,$70,$AF,$33,$F3,$A3,$C8,$6E,$BD,$87,$8E,$3F,$60
	.dc.b	$7C,$38,$D1,$C1,$E9,$5E,$19,$E3,$3A,$9F,$85,$5F,$3C,$F0,$BF,$10
	.dc.b	$CD,$7D,$17,$BC,$FB,$DF,$E3,$2D,$9F,$E3,$07,$23,$A4,$F8,$63,$C9
	.dc.b	$3C,$0F,$82,$5A,$F2,$BE,$88,$F9,$EE,$AB,$AD,$73,$8F,$2C,$FC,$E3
	.dc.b	$75,$CA,$3A,$47,$82,$B8,$BA,$31,$1F,$24,$F0,$1E,$E3,$C0,$64,$E6
	.dc.b	$78,$0F,$C6,$D7,$8D,$FC,$01,$88,$EC,$EE,$BE,$C1,$FB,$57,$A5,$79
	.dc.b	$1F,$04,$7A,$63,$0E,$E7,$E7,$4F,$03,$FC,$26,$0F,$F0,$01,$4F,$EB
	.dc.b	$A5,$3F,$C1,$E5,$97,$94,$9D,$4F,$F9,$36,$30,$E2,$77,$3A,$8F,$F2
	.dc.b	$1C,$E9,$ED,$CD,$FE,$BF,$EA,$D3,$FE,$BE,$8F,$33,$FE,$BE,$8E,$5C
	.dc.b	$0C,$BD,$19,$73,$87,$6E,$32,$F6,$46,$31,$F8,$15,$BB,$D4,$7F,$AC
	.dc.b	$CC,$D7,$FB,$01,$4E,$DC,$9F,$EB,$5E,$5F,$F5,$65,$EF,$88,$D3,$FD
	.dc.b	$81,$5F,$EC,$58,$CD,$77,$3F,$D7,$3C,$BF,$EC,$47,$FD,$79,$1A,$7F
	.dc.b	$D8,$2D,$AF,$F6,$0D,$F8,$27,$FD,$8D,$1F,$EC,$43,$35,$CE,$75,$3C
	.dc.b	$A7,$C8,$FA,$67,$FD,$7E,$FE,$06,$7F,$D8,$84,$8C,$C7,$E1,$E1,$CB
	.dc.b	$FE,$BF,$8C,$35,$46,$BA,$FF,$00,$D8,$FF,$79,$F3,$7F,$B0,$DF,$F5
	.dc.b	$F9,$82,$63,$FE,$B4,$E6,$FF,$5B,$E5,$FE,$C1,$24,$FE,$56,$BF,$D8
	.dc.b	$53,$A8,$FF,$5C,$E9,$F5,$9F,$B3,$1F,$94,$D3,$F2,$01,$C4,$FF,$B1
	.dc.b	$B0,$F0,$3F,$EB,$62,$57,$E0,$9F,$C4,$F2,$FF,$AF,$5C,$DF,$E6,$F9
	.dc.b	$BF,$D8,$75,$FA,$46,$BF,$AB,$6B,$FA,$F6,$BF,$CB,$0F,$DB,$4F,$EE
	.dc.b	$08,$FF,$0C,$6C,$FE,$FE,$36,$64,$8E,$63,$67,$F1,$82,$C7,$F0,$D7
	.dc.b	$F9,$E3,$FC,$4A,$3F,$87,$C7,$F8,$76,$3A,$C7,$37,$AC,$73,$69,$CC
	.dc.b	$FE,$47,$1F,$C4,$5C,$9D,$63,$BF,$27,$F3,$FE,$0F,$E8,$D5,$B3,$B1
	.dc.b	$27,$B3,$6B,$B1,$FC,$B0,$E8,$78,$23,$D0,$7B,$9C,$9B,$AF,$04,$7F
	.dc.b	$41,$8F,$E5,$2F,$62,$3D,$A3,$F9,$5C,$74,$8F,$5C,$72,$7F,$9C,$1C
	.dc.b	$8E,$07,$C8,$8E,$6F,$8A,$39,$BF,$7E,$3C,$D1,$CE,$7E,$8C,$E4,$EF
	.dc.b	$5F,$D1,$E7,$55,$F2,$A3,$04,$7F,$A4,$CF,$73,$C4,$79,$1E,$F1,$B8
	.dc.b	$F4,$1B,$8C,$9C,$A7,$BC,$7C,$D3,$C9,$5F,$4D,$E0,$79,$A7,$85,$7E
	.dc.b	$B2,$3F,$A5,$9F,$D3,$88,$60,$87,$F4,$E3,$25,$3F,$D4,$52,$E5,$FD
	.dc.b	$C5,$7F,$BE,$A3,$19,$F7,$AF,$F5,$6C,$62,$75,$5C,$27,$FD,$DB,$39
	.dc.b	$6C,$66,$3F,$DA,$66,$4A,$E6,$7F,$DB,$51,$96,$C6,$67,$CE,$D7,$C6
	.dc.b	$DC,$3F,$AC,$4F,$F4,$AA,$E5,$5C,$2F,$D5,$5A,$8F,$E2,$07,$F5,$2B
	.dc.b	$E5,$87,$CD,$BE,$33,$DA,$F8,$CE,$6B,$78,$77,$6D,$74,$BE,$38,$77
	.dc.b	$7F,$93,$4E,$9B,$87,$4B,$E4,$DB,$E5,$5B,$9F,$ED,$AD,$8F,$44,$EE
	.dc.b	$B9,$D7,$D4,$8C,$3C,$1B,$3C,$23,$44,$7E,$BC,$6A,$7D,$38,$7B,$E3
	.dc.b	$17,$E1,$9C,$DF,$A0,$E4,$FD,$C9,$D9,$23,$27,$DC,$71,$7E,$08,$CB
	.dc.b	$F5,$DA,$FD,$16,$E1,$F2,$DF,$61,$8C,$3C,$CD,$3C,$93,$D4,$97,$E5
	.dc.b	$8D,$47,$CE,$BF,$1C,$E6,$37,$3E,$73,$71,$F3,$A3,$27,$9C,$A7,$9C
	.dc.b	$B7,$E7,$7E,$81,$8B,$F3,$B6,$BC,$F5,$E1,$23,$E1,$8D,$1E,$77,$C6
	.dc.b	$7C,$27,$31,$E5,$38,$BE,$56,$BE,$56,$D7,$95,$FB,$71,$D4,$E7,$3D
	.dc.b	$63,$75,$E8,$6C,$7A,$0F,$33,$89,$E4,$D3,$C9,$1A,$24,$79,$5A,$F9
	.dc.b	$5B,$5E,$5C,$3F,$20,$4C,$3E,$BC,$EB,$0F,$23,$4F,$25,$F5,$21,$D4
	.dc.b	$B7,$D6,$B9,$C6,$2B,$9B,$6B,$9B,$C4,$95,$C6,$3E,$9C,$7D,$EC,$3E
	.dc.b	$04,$7C,$93,$C9,$1F,$11,$B1,$F1,$23,$B1,$92,$B9,$7B,$D7,$91,$B3
	.dc.b	$E4,$AF,$14,$7D,$83,$C5,$1F,$E9,$1E,$4F,$F4,$14,$7F,$A3,$1F,$F4
	.dc.b	$12,$7F,$D0,$4F,$FA,$43,$C6,$78,$FF,$E8,$D8,$46,$1F,$F4,$83,$5E
	.dc.b	$11,$BE,$0F,$F4,$54,$7F,$91,$DF,$F4,$33,$FC,$C0,$FF,$3D,$F2,$FF
	.dc.b	$A1,$9F,$F3,$EE,$56,$78,$C6,$DB,$9F,$FA,$1E,$39,$71,$7F,$9C,$8F
	.dc.b	$F3,$9C,$FF,$9D,$EB,$FD,$61,$C3,$FE,$7F,$9D,$72,$FF,$A0,$EB,$57
	.dc.b	$FE,$B7,$AD,$4F,$F9,$F4,$D5,$F2,$AF,$F3,$ED,$7F,$A2,$9F,$F3,$E4
	.dc.b	$EA,$3F,$D2,$51,$FE,$94,$55,$FE,$8D,$AE,$51,$FE,$95,$8E,$F8,$78
	.dc.b	$6B,$9B,$FE,$7C,$BF,$F4,$2D,$7F,$A0,$1B,$87,$FA,$2B,$97,$FD,$08
	.dc.b	$FF,$A0,$67,$FD,$98,$D7,$FD,$21,$1F,$E8,$97,$FD,$A9,$C9,$FE,$85
	.dc.b	$23,$E6,$C3,$FD,$7F,$C3,$FE,$7F,$8F,$F6,$C4,$FF,$A0,$B8,$3C,$D1
	.dc.b	$FE,$85,$38,$1F,$EC,$09,$E0,$F6,$32,$7F,$A1,$A3,$8B,$D0,$CE,$5F
	.dc.b	$E0,$9C,$FF,$D6,$91,$C5,$FF,$43,$BC,$8C,$61,$E6,$AE,$71,$83,$C4
	.dc.b	$E1,$FF,$50,$CF,$23,$AB,$89,$E4,$F8,$AF,$FC,$3D,$7F,$E0,$D8,$E8
	.dc.b	$7A,$4F,$CC,$3F,$E1,$13,$4F,$33,$C2,$7F,$90,$38,$7F,$C7,$AF,$F9
	.dc.b	$97,$5F,$F1,$57,$1B,$D8,$FF,$7B,$4F,$FA,$5D,$A8,$46,$F3,$FF,$7F
	.dc.b	$36,$3F,$D4,$53,$FE,$17,$E5,$7A,$5D,$A3,$0F,$F8,$E7,$AD,$F2,$34
	.dc.b	$FF,$0D,$77,$DC,$BF,$D7,$93,$A7,$C5,$8C,$11,$97,$06,$2B,$D2,$D9
	.dc.b	$CB,$4F,$F0,$DF,$27,$76,$C7,$FB,$2F,$F3,$D7,$B9,$FE,$4F,$6C,$7F
	.dc.b	$8B,$79,$71,$32,$3A,$CE,$AB,$C6,$D7,$D0,$D7,$D6,$EF,$87,$FC,$67
	.dc.b	$F5,$9C,$E7,$FC,$5E,$F8,$7C,$89,$7F,$E3,$DF,$AE,$B5,$FC,$67,$05
	.dc.b	$7C,$06,$D7,$92,$30,$47,$FC,$8D,$F3,$5C,$B1,$FF,$21,$BE,$FD,$24
	.dc.b	$FF,$92,$F6,$B8,$F7,$9E,$C4,$9F,$82,$D8,$FF,$23,$FC,$7F,$0C,$3F
	.dc.b	$C9,$EF,$CB,$CE,$5F,$F9,$5F,$E5,$2D,$FA,$E7,$D4,$49,$F9,$AD,$8F
	.dc.b	$F2,$9F,$D3,$F0,$C3,$FC,$B6,$FD,$7C,$25,$FF,$98,$FE,$26,$DF,$C7
	.dc.b	$9F,$8C,$49,$FB,$2D,$8F,$F2,$FF,$C2,$F0,$C3,$FC,$CE,$FE,$0E,$E5
	.dc.b	$FF,$9B,$FE,$1A,$DF,$D3,$9F,$A4,$49,$FC,$2B,$63,$FC,$D7,$F0,$3C
	.dc.b	$30,$FF,$39,$BF,$E8,$4F,$26,$5F,$FA,$27,$A9,$F4,$B8,$C3,$FA,$C7
	.dc.b	$72,$F1,$8B,$0F,$DC,$2C,$38,$C6,$0F,$D5,$D7,$17,$17,$FB,$F8,$C3
	.dc.b	$DA,$FF,$79,$87,$6A,$E3,$16,$58,$77,$AC,$42,$F9,$CF,$96,$3B,$57
	.dc.b	$4B,$F0,$C6,$E3,$8D,$70,$9F,$5E,$1E,$AA,$F2,$47,$09,$D4,$74,$8F
	.dc.b	$91,$87,$2B,$F3,$61,$E5,$9E,$11,$97,$66,$09,$5D,$8F,$F5,$56,$1E
	.dc.b	$1A,$F5,$47,$9E,$77,$18,$C3,$C6,$F7,$9E,$4E,$6F,$C7,$7C,$EF,$E5
	.dc.b	$9C,$8B,$5E,$0C,$3C,$CF,$28,$F8,$71,$8B,$E9,$1E,$A9,$D1,$F0,$CB
	.dc.b	$1E,$33,$6E,$8F,$51,$DE,$3C,$4F,$18,$D9,$F6,$4E,$51,$E6,$72,$D3
	.dc.b	$CA,$72,$8E,$B1,$A7,$F0,$F5,$F0,$5E,$D3,$C5,$A6,$AB,$DD,$19,$AF
	.dc.b	$0C,$E8,$EB,$19,$AE,$91,$83,$58,$7D,$48,$CB,$F8,$88,$D1,$BA,$EA
	.dc.b	$E2,$34,$71,$BF,$1C,$79,$A7,$73,$AC,$3E,$ED,$6A,$BA,$5F,$82,$78
	.dc.b	$DF,$06,$C6,$70,$F9,$18,$77,$6C,$F4,$69,$F5,$AB,$93,$8B,$FB,$47
	.dc.b	$DD,$AF,$24,$7E,$16,$7E,$94,$FB,$6B,$F6,$4E,$DE,$CD,$9E,$D7,$E5
	.dc.b	$8C,$1D,$AB,$9D,$F6,$8C,$57,$DB,$7B,$3E,$13,$11,$B9,$FC,$65,$75
	.dc.b	$79,$DF,$91,$B5,$E2,$C3,$E6,$5F,$C5,$C3,$C7,$5F,$8D,$3E,$1B,$8A
	.dc.b	$FC,$9B,$4F,$B1,$87,$78,$F1,$E1,$DD,$A7,$73,$E6,$57,$59,$E1,$5F
	.dc.b	$AD,$3F,$C2,$97,$FC,$A1,$B7,$FE,$12,$8C,$47,$EF,$0C,$DF,$F8,$52
	.dc.b	$33,$1C,$CF,$CD,$7A,$9F,$C1,$89,$3F,$49,$B1,$FB,$CB,$46,$30,$F3
	.dc.b	$36,$B9,$9C,$49,$5C,$63,$4F,$C4,$6B,$F9,$86,$DF,$EA,$5B,$87,$F8
	.dc.b	$FD,$B5,$FD,$5D,$B5,$FC,$A1,$B7,$F7,$9B,$3F,$CF,$1B,$3F,$C1,$1B
	.dc.b	$3F,$D6,$5B,$87,$BA,$BD,$33,$EE,$69,$EE,$AF,$63,$87,$D8,$D7,$D8
	.dc.b	$DC,$3D,$8D,$C3,$D8,$D3,$D8,$59,$F6,$36,$BD,$8D,$3D,$85,$8F,$63
	.dc.b	$63,$D8,$DB,$F6,$4F,$A0,$93,$ED,$6C,$FB,$63,$04,$70,$48,$F7,$9C
	.dc.b	$0B,$1B,$6C,$6D,$F8,$66,$CB,$1B,$6C,$6E,$FE,$3C,$6D,$B3,$B6,$CE
	.dc.b	$EF,$E5,$D6,$DB,$5B,$6C,$EC,$F9,$F3,$B6,$CE,$DA,$6E,$3E,$99,$B2
	.dc.b	$C7,$F7,$6F,$CF,$9E,$07,$F5,$9A,$FF,$45,$C6,$1F,$F1,$0C,$BF,$53
	.dc.b	$FF,$50,$47,$F8,$52,$75,$7F,$E1,$E8,$E3,$7F,$E8,$E9,$CC,$FF,$A5
	.dc.b	$5A,$EA,$7F,$D2,$31,$CA,$FF,$B5,$5B,$C7,$F2,$5F,$FA,$BE,$B5,$7F
	.dc.b	$E3,$AA,$D3,$E3,$FA,$FD,$D2,$91,$CB,$0F,$56,$3A,$9C,$DF,$13,$C7
	.dc.b	$5E,$92,$38,$3E,$1E,$0F,$96,$E5,$3D,$63,$13,$F1,$7E,$BF,$B1,$9F
	.dc.b	$72,$C6,$4F,$83,$1F,$1C,$8E,$0E,$F5,$C3,$3F,$9C,$7F,$93,$2F,$D2
	.dc.b	$E0,$F8,$1F,$0B,$ED,$3C,$9E,$4D,$BF,$B1,$1F,$54,$8F,$D2,$6C,$F6
	.dc.b	$D7,$F0,$1F,$2D,$C6,$31,$82,$3E,$57,$C2,$4C,$3E,$2E,$76,$FC,$9B
	.dc.b	$2B,$F0,$84,$AF,$C0,$B6,$7A,$FC,$AF,$1C,$BF,$06,$FD,$7F,$26,$5F
	.dc.b	$D2,$E3,$B7,$DA,$7F,$20,$49,$FC,$6B,$63,$D3,$AF,$0C,$3F,$1E,$FE
	.dc.b	$23,$91,$F0,$94,$FB,$9F,$07,$71,$FF,$36,$47,$F5,$92,$3F,$D5,$DB
	.dc.b	$87,$C6,$F8,$6E,$0D,$9F,$F3,$DF,$CE,$7F,$1C,$8F,$F7,$C3,$5F,$F7
	.dc.b	$DB,$50,$70,$BD,$67,$B8,$F6,$BF,$DF,$71,$86,$FB,$FF,$AB,$AE,$64
	.dc.b	$9D,$38,$3B,$FD,$5B,$D5,$DF,$07,$ED,$39,$2E,$9F,$B1,$6D,$F0,$AA
	.dc.b	$AF,$93,$D4,$D1,$4D,$45,$69,$9E,$9F,$E2,$AF,$FA,$89,$F5,$3F,$9F
	.dc.b	$6E,$1F,$E9,$07,$71,$9E,$DF,$1C,$62,$75,$1E,$6F,$61,$E5,$FF,$AD
	.dc.b	$AD,$EB,$B0,$F8,$F5,$AC,$3E,$9C,$E6,$BE,$AD,$78,$23,$11,$C7,$0F
	.dc.b	$9F,$18,$BE,$53,$DA,$BA,$E1,$E1,$AE,$31,$CC,$FA,$E5,$9F,$C1,$D7
	.dc.b	$29,$E9,$5C,$6B,$77,$E0,$6D,$FA,$1E,$F8,$58,$E7,$5C,$5F,$5D,$71
	.dc.b	$69,$B1,$87,$86,$F9,$C7,$AC,$D9,$4F,$84,$71,$8E,$0F,$82,$FF,$05
	.dc.b	$1F,$12,$3F,$24,$D7,$A9,$8F,$4D,$EE,$70,$6C,$7E,$36,$51,$88,$E1
	.dc.b	$3F,$8D,$5E,$CD,$23,$F9,$16,$E7,$F6,$B0,$FD,$16,$FF,$06,$BF,$67
	.dc.b	$5F,$0E,$73,$1E,$F3,$27,$C1,$73,$5F,$84,$8C,$BD,$75,$FF,$80,$0E
	.dc.b	$63,$CA,$71,$8C,$E7,$FA,$9B,$F0,$46,$27,$F7,$11,$8C,$3C,$D1,$CA
	.dc.b	$FB,$46,$27,$53,$E7,$35,$7F,$BE,$BE,$90,$C3,$DA,$7F,$BF,$0B,$7E
	.dc.b	$38,$F3,$3F,$C5,$A3,$53,$FD,$C9,$DC,$F9,$8D,$9F,$D5,$FA,$3F,$BD
	.dc.b	$BF,$C7,$9A,$9F,$F0,$7B,$BD,$7F,$9B,$12,$FC,$07,$F1,$C9,$DD,$23
	.dc.b	$13,$FE,$04,$F6,$5E,$CF,$F8,$0F,$C7,$76,$E6,$30,$DB,$FF,$10,$CE
	.dc.b	$9E,$F9,$7F,$34,$BE,$71,$AA,$FF,$08,$FA,$4F,$63,$C5,$7F,$C8,$63
	.dc.b	$15,$FC,$85,$B1,$99,$DB,$FC,$CE,$33,$9F,$21,$EE,$7F,$18,$BF,$1D
	.dc.b	$2B,$71,$F4,$E7,$53,$FE,$05,$9C,$BF,$76,$33,$5E,$09,$CB,$DB,$97
	.dc.b	$FC,$57,$EB,$BD,$9F,$2E,$1F,$D2,$A3,$18,$FF,$3B,$6C,$7F,$3F,$6B
	.dc.b	$FD,$35,$CD,$FA,$F1,$7B,$3D,$B0,$FF,$1B,$46,$30,$F6,$DF,$07,$D9
	.dc.b	$5B,$7F,$CA,$B3,$BC,$3F,$C4,$54,$F6,$7C,$F1,$FD,$AA,$31,$5E,$79
	.dc.b	$CC,$A7,$2F,$A6,$3C,$7E,$CF,$C2,$6D,$79,$1B,$3E,$06,$E5,$F6,$AF
	.dc.b	$E8,$CF,$C1,$CF,$BC,$62,$3E,$54,$60,$F8,$3E,$B7,$D0,$6E,$1F,$09
	.dc.b	$B3,$EE,$6D,$F7,$F6,$3A,$92,$BE,$A3,$6B,$FC,$79,$9B,$D8,$FB,$B3
	.dc.b	$F4,$1B,$5F,$31,$B5,$C1,$FC,$1E,$3E,$A2,$4F,$E0,$5B,$3D,$BD,$8D
	.dc.b	$CF,$D6,$6C,$66,$3E,$EC,$FE,$02,$7F,$1F,$EB,$FC,$62,$4F,$E3,$5B
	.dc.b	$5E,$AF,$5F,$73,$F8,$26,$C6,$63,$F1,$33,$F8,$C9,$FF,$1F,$79,$92
	.dc.b	$7F,$7B,$EB,$DB,$FE,$29,$37,$5F,$D8,$74,$2B,$17,$FC,$3E,$AD,$FF
	.dc.b	$27,$AD,$4F,$ED,$3A,$D5,$AA,$FE,$38,$FF,$11,$BC,$4E,$AB,$84,$E5
	.dc.b	$A4,$AF,$2B,$E8,$31,$87,$F0,$8A,$E5,$3A,$3F,$CF,$58,$79,$1C,$4E
	.dc.b	$EF,$D2,$A7,$4E,$6F,$A9,$EF,$3C,$15,$C9,$C3,$4F,$85,$87,$B2,$FA
	.dc.b	$DF,$4A,$DD,$F5,$BE,$0F,$86,$F8,$4F,$82,$79,$B5,$F9,$2E,$67,$83
	.dc.b	$9C,$3E,$0C,$75,$3E,$CD,$74,$AF,$F2,$94,$FC,$67,$FC,$C0,$D7,$D6
	.dc.b	$63,$0F,$F7,$8B,$FD,$55,$AF,$FC,$00,$31,$7F,$EF,$06,$9F,$EE,$E7
	.dc.b	$FD,$E5,$7F,$8A,$F5,$FE,$19,$23,$D5,$8A,$31,$5F,$EF,$7A,$E1,$5A
	.dc.b	$9C,$C7,$FB,$FE,$7B,$1E,$07,$6F,$19,$FE,$E9,$5D,$5E,$44,$8D,$1F
	.dc.b	$99,$AE,$B1,$E8,$38,$B8,$8E,$D1,$83,$95,$78,$CF,$09,$5D,$47,$68
	.dc.b	$FC,$74,$F3,$8F,$94,$4C,$3E,$CD,$70,$AE,$53,$E5,$8E,$E7,$95,$C1
	.dc.b	$D6,$FD,$CD,$BE,$D7,$D2,$37,$7D,$0F,$1D,$F4,$8F,$69,$A7,$B4,$6A
	.dc.b	$3A,$46,$DF,$46,$1F,$84,$AE,$D5,$B3,$B6,$1E,$46,$E1,$E4,$AE,$04
	.dc.b	$C3,$F1,$58,$74,$6E,$1D,$30,$E5,$1E,$46,$CF,$A3,$0F,$04,$E7,$0E
	.dc.b	$CF,$2C,$3B,$57,$E2,$0E,$C5,$C3,$B5,$70,$8C,$4F,$0B,$FB,$B7,$D4
	.dc.b	$F8,$57,$C1,$B3,$B7,$ED,$1B,$AF,$92,$4A,$F8,$71,$C0,$F3,$12,$3C
	.dc.b	$2D,$8F,$0C,$7B,$F0,$F8,$2F,$AA,$32,$D9,$F0,$BB,$C3,$D9,$3A,$C3
	.dc.b	$D1,$3E,$3A,$F3,$4F,$92,$30,$71,$8F,$33,$E4,$24,$73,$C3,$E0,$47
	.dc.b	$E2,$4E,$07,$23,$C8,$64,$B1,$F6,$1D,$BE,$12,$4F,$90,$91,$A6,$BC
	.dc.b	$B0,$F8,$77,$EC,$AE,$33,$B2,$1B,$C3,$E1,$DF,$C6,$8C,$3E,$E9,$F6
	.dc.b	$3F,$44,$D5,$75,$7C,$2D,$9E,$AD,$79,$9F,$CF,$AF,$EB,$FA,$0E,$DF
	.dc.b	$F1,$8B,$6D,$50,$91,$97,$06,$9C,$38,$7F,$D1,$78,$3D,$37,$AF,$FE
	.dc.b	$69,$BE,$78,$5C,$3F,$D3,$D5,$A9,$FF,$35,$E1,$CA,$31,$3F,$EA,$39
	.dc.b	$E2,$7F,$A0,$B0,$ED,$87,$39,$5B,$D5,$F6,$92,$FD,$33,$E9,$8C,$3D
	.dc.b	$4F,$1F,$AF,$CE,$B9,$57,$7C,$53,$9B,$F5,$63,$EF,$8C,$56,$DE,$5E
	.dc.b	$AF,$CE,$24,$78,$B1,$F3,$C6,$1F,$15,$7C,$9B,$E3,$5E,$16,$CE,$B0
	.dc.b	$F2,$BE,$93,$9C,$EC,$F9,$7E,$A7,$DE,$23,$E1,$C9,$18,$9E,$9E,$B7
	.dc.b	$C1,$8F,$6D,$A3,$17,$F0,$DB,$5C,$23,$EC,$F9,$BF,$8C,$23,$DD,$B1
	.dc.b	$E9,$F5,$BE,$74,$76,$6C,$FC,$F6,$CE,$E3,$F0,$BE,$27,$E6,$08,$FA
	.dc.b	$DB,$1F,$17,$1F,$B5,$1E,$A6,$CF,$DB,$6C,$EE,$3F,$C9,$1E,$AF,$FA
	.dc.b	$53,$8C,$D2,$47,$D9,$8C,$BF,$85,$F5,$7F,$25,$18,$AF,$AD,$18,$AF
	.dc.b	$96,$DA,$E0,$7F,$B2,$BD,$17,$A8,$F5,$3F,$0D,$18,$6F,$A9,$FA,$CC
	.dc.b	$7F,$37,$39,$7F,$3B,$F2,$CF,$82,$FC,$74,$23,$0F,$E4,$5A,$F0,$8D
	.dc.b	$FA,$CF,$43,$F8,$17,$3F,$FB,$87,$95,$C9,$8D,$BF,$83,$CF,$EF,$E3
	.dc.b	$18,$FF,$0A,$C3,$93,$FE,$E6,$6B,$F9,$76,$CF,$FB,$C1,$B5,$FC,$45
	.dc.b	$78,$FD,$29,$E8,$FF,$19,$7F,$DE,$2F,$C3,$7F,$44,$DB,$FF,$7D,$BB
	.dc.b	$8C,$F6,$F3,$8C,$4E,$A3,$B7,$AD,$E0,$8F,$E4,$7E,$A7,$82,$FF,$92
	.dc.b	$CE,$54,$65,$F2,$D7,$F2,$98,$C5,$F8,$5D,$57,$16,$E0,$C3,$BD,$7F
	.dc.b	$31,$38,$16,$FA,$3F,$CD,$6F,$93,$6B,$96,$1E,$48,$E0,$F4,$3F,$9E
	.dc.b	$BD,$1F,$5C,$70,$C3,$C7,$86,$E7,$D6,$5A,$E0,$AF,$C2,$F2,$9D,$61
	.dc.b	$E6,$8E,$ED,$78,$61,$DE,$FD,$15,$FC,$8E,$FE,$65,$7F,$4E,$6B,$A3
	.dc.b	$F9,$CA,$F5,$9E,$B7,$16,$CE,$9A,$F3,$6B,$FD,$0B,$D6,$FD,$2A,$69
	.dc.b	$1F,$D4,$BF,$0E,$FF,$49,$AF,$C5,$9F,$EC,$F3,$9E,$04,$E6,$7A,$1F
	.dc.b	$4C,$FD,$3C,$69,$C1,$FD,$BE,$79,$4F,$16,$C6,$8E,$E7,$0A,$F1,$B9
	.dc.b	$3B,$EB,$F9,$38,$CB,$F9,$58,$E2,$E6,$FC,$4E,$63,$8E,$7F,$C0,$27
	.dc.b	$2F,$C7,$8F,$04,$61,$F5,$E1,$FC,$08,$E0,$57,$FC,$0F,$87,$AB,$0F
	.dc.b	$81,$1F,$9F,$AF,$EF,$11,$EA,$33,$1E,$6B,$E6,$FF,$5D,$E9,$FF,$09
	.dc.b	$4F,$FB,$BF,$6F,$E5,$F3,$FC,$C1,$98,$FF,$C0,$17,$2D,$3C,$90,$F4
	.dc.b	$DE,$A7,$F8,$25,$B3,$FE,$0D,$F3,$DE,$E6,$6A,$4F,$F7,$9A,$DD,$FF
	.dc.b	$83,$1A,$FF,$89,$9A,$FF,$BC,$22,$3F,$EE,$C3,$FD,$F0,$7F,$BB,$0F
	.dc.b	$F1,$B7,$95,$FE,$E7,$7F,$DC,$51,$FD,$3F,$CA,$FE,$A8,$7C,$68,$FF
	.dc.b	$1B,$FE,$80,$F6,$3F,$C4,$4D,$AF,$F0,$C7,$83,$89,$91,$F0,$A7,$55
	.dc.b	$FE,$0B,$AC,$BF,$E0,$B8,$CB,$FE,$40,$6B,$FE,$45,$6B,$E2,$AF,$22
	.dc.b	$87,$D2,$79,$0F,$F0,$5E,$B7,$0F,$14,$5A,$FF,$0E,$4D,$DF,$F0,$F7
	.dc.b	$CA,$FF,$C5,$0D,$AF,$F1,$DB,$5F,$35,$F8,$A1,$F4,$5F,$10,$FF,$0D
	.dc.b	$F8,$B6,$FE,$14,$5B,$FF,$16,$4D,$C3,$C3,$3C,$C9,$5E,$27,$B7,$D0
	.dc.b	$FD,$03,$FC,$4F,$B5,$9F,$9D,$17,$0F,$F1,$C4,$DB,$F6,$CF,$2A,$F6
	.dc.b	$B6,$BD,$BC,$1F,$70,$FF,$19,$F9,$F6,$BE,$D4,$59,$FF,$22,$4D,$BF
	.dc.b	$93,$3F,$0C,$95,$F2,$9F,$AF,$F0,$CF,$6F,$FC,$D4,$D9,$FF,$30,$7C
	.dc.b	$46,$26,$47,$EA,$A7,$58,$FF,$94,$EB,$2F,$F9,$4E,$32,$FF,$A2,$1A
	.dc.b	$FF,$A3,$5A,$FF,$A4,$1A,$B9,$FF,$C5,$58,$FF,$08,$6D,$FF,$84,$3C
	.dc.b	$27,$BF,$FE,$2B,$BF,$E0,$DC,$B7,$0F,$F0,$C4,$DB,$FF,$4E,$61,$CA
	.dc.b	$FF,$D3,$ED,$AF,$F4,$C5,$6E,$73,$76,$73,$7F,$E4,$62,$4F,$81,$FF
	.dc.b	$43,$57,$18,$F5,$57,$1C,$3C,$F5,$C2,$FD,$93,$9B,$E1,$87,$97,$0E
	.dc.b	$37,$D0,$E0,$7A,$AB,$83,$88,$F3,$B6,$FC,$F1,$E5,$23,$E0,$8E,$91
	.dc.b	$87,$A1,$C7,$0F,$3C,$64,$EA,$59,$EA,$73,$74,$7A,$0F,$F4,$B6,$1E
	.dc.b	$CB,$F3,$C6,$B0,$F7,$1F,$46,$FB,$DF,$32,$61,$E4,$AE,$6F,$BD,$ED
	.dc.b	$1A,$7F,$D9,$78,$77,$9F,$53,$63,$D1,$5D,$A7,$A5,$71,$9E,$71,$F8
	.dc.b	$DA,$E1,$3D,$4F,$7C,$70,$24,$FC,$48,$FC,$7C,$79,$6F,$DB,$87,$FA
	.dc.b	$AA,$F9,$CF,$0A,$E7,$7C,$63,$9B,$E2,$3D,$6F,$3A,$F9,$33,$E7,$BE
	.dc.b	$B3,$F6,$A3,$13,$A8,$E7,$3A,$AF,$92,$7D,$19,$D3,$E1,$21,$E1,$AF
	.dc.b	$2B,$E3,$24,$FD,$7B,$F5,$C7,$19,$DC,$F7,$7E,$46,$1F,$3E,$7A,$CE
	.dc.b	$70,$F9,$C7,$B4,$FA,$E5,$8F,$AF,$7D,$9F,$25,$73,$7C,$78,$7A,$9B
	.dc.b	$5C,$9F,$73,$70,$F8,$AD,$3E,$2C,$F1,$C3,$C9,$5E,$BB,$ED,$18,$8F
	.dc.b	$0E,$1E,$38,$F8,$57,$E7,$C3,$DF,$39,$C3,$F0,$13,$93,$F0,$18,$74
	.dc.b	$8F,$93,$5F,$21,$B1,$99,$E1,$87,$E0,$8F,$59,$F1,$1C,$BE,$46,$E1
	.dc.b	$E4,$8D,$DF,$92,$73,$7E,$49,$CC,$79,$30,$F4,$3F,$8B,$3E,$29,$F1
	.dc.b	$E3,$EC,$CF,$C7,$3C,$F3,$C2,$3F,$1B,$5F,$39,$F3,$1D,$8F,$99,$1B
	.dc.b	$8F,$D3,$C6,$4F,$51,$6F,$D5,$1C,$CF,$54,$66,$BD,$51,$9B,$F5,$1E
	.dc.b	$E3,$D5,$3E,$87,$D4,$FE,$18,$F2,$C7,$17,$CB,$1C,$5F,$5B,$F0,$9D
	.dc.b	$B6,$3C,$67,$18,$F4,$B6,$BD,$33,$B3,$CB,$3E,$97,$CA,$F8,$E7,$9C
	.dc.b	$6E,$3C,$66,$CF,$19,$EB,$3C,$65,$8E,$2D,$8E,$F1,$E3,$AF,$8F,$1E
	.dc.b	$B9,$1F,$CB,$4A,$7F,$3E,$29,$FC,$EC,$B5,$FC,$DF,$CA,$FE,$B8,$49
	.dc.b	$FE,$90,$D9,$FE,$5E,$DB,$FE,$06,$D9,$FD,$1B,$67,$F6,$0D,$7F,$7E
	.dc.b	$DA,$F0,$4F,$62,$4F,$66,$D7,$66,$9D,$8B,$1D,$9A,$F6,$6C,$F6,$8E
	.dc.b	$87,$F8,$6F,$A7,$FC,$4A,$7F,$AE,$6A,$64,$FA,$1C,$E1,$FE,$25,$AB
	.dc.b	$7F,$EF,$D8,$FF,$0C,$4E,$2F,$FD,$47,$A7,$FA,$3A,$7E,$7D,$7F,$88
	.dc.b	$EF,$FC,$6A,$48,$CC,$25,$EF,$7F,$8E,$0F,$24,$FD,$CB,$7E,$A9,$FE
	.dc.b	$D4,$8F,$F1,$C4,$ED,$A7,$FA,$DE,$76,$E6,$FF,$C7,$53,$FE,$46,$24
	.dc.b	$78,$AF,$C2,$DE,$CF,$F2,$49,$E2,$A5,$F8,$BE,$A7,$35,$FE,$49,$AB
	.dc.b	$5F,$14,$DE,$2A,$CD,$FC,$6A,$FF,$2B,$92,$FC,$07,$8F,$DC,$FF,$2C
	.dc.b	$9E,$19,$F8,$3F,$17,$89,$FF,$2C,$4D,$C7,$C1,$5A,$9C,$92,$3E,$9E
	.dc.b	$4A,$FF,$33,$92,$7E,$17,$C2,$FF,$9A,$0F,$06,$6F,$3F,$13,$FE,$67
	.dc.b	$9B,$87,$82,$B5,$39,$24,$78,$27,$FC,$EA,$48,$FC,$0F,$E8,$4F,$6B
	.dc.b	$E7,$52,$31,$3F,$6E,$DE,$9E,$A7,$35,$F3,$CE,$17,$93,$D3,$36,$FF
	.dc.b	$DC,$F5,$C2,$BF,$81,$B5,$FE,$0E,$D7,$F8,$5B,$57,$D3,$BE,$13,$FD
	.dc.b	$8A,$DA,$FC,$DB,$67,$F4,$2D,$7F,$88,$FC,$33,$BB,$FD,$9A,$DA,$FD
	.dc.b	$1B,$67,$F5,$2D,$7F,$8C,$FA,$BF,$0B,$D6,$FD,$4B,$63,$F5,$EF,$D7
	.dc.b	$D3,$3A,$2B,$F5,$FF,$15,$FE,$CA,$BF,$D6,$63,$FC,$9B,$E4,$1E,$5F
	.dc.b	$F2,$8E,$54,$E6,$30,$FF,$2C,$BE,$57,$F4,$5B,$5C,$23,$F9,$9D,$F4
	.dc.b	$7B,$98,$BE,$6F,$F3,$8A,$E3,$7F,$CE,$3D,$57,$9F,$AC,$5D,$CF,$5B
	.dc.b	$0D,$C7,$A8,$BE,$2E,$1A,$FE,$1D,$B5,$E7,$8E,$17,$E4,$79,$7A,$D7
	.dc.b	$D1,$7A,$37,$86,$57,$82,$72,$D9,$E0,$47,$CB,$EB,$2F,$F8,$5C,$3D
	.dc.b	$45,$FF,$58,$AD,$65,$FD,$72,$73,$5F,$D7,$AB,$BB,$C5,$FE,$BE,$62
	.dc.b	$37,$0A,$E6,$EC,$FE,$CE,$E2,$37,$3C,$4E,$78,$7F,$6A,$8E,$73,$B9
	.dc.b	$F3,$61,$EA,$9E,$35,$B9,$E3,$7E,$97,$78,$58,$F0,$DF,$19,$DC,$79
	.dc.b	$4F,$87,$5C,$DD,$D7,$17,$D9,$5C,$5F,$89,$3B,$78,$9B,$C3,$D0,$F2
	.dc.b	$3F,$AA,$BF,$11,$FF,$02,$B5,$F3,$9A,$F5,$5E,$B7,$F5,$89,$46,$27
	.dc.b	$53,$98,$C4,$7F,$5E,$F5,$BF,$AF,$BF,$E1,$66,$A6,$91,$F8,$58,$FE
	.dc.b	$8B,$87,$E0,$47,$13,$3B,$7E,$BF,$5F,$BF,$19,$7F,$03,$5C,$63,$E6
	.dc.b	$19,$D7,$F7,$D5,$C5,$FE,$EF,$87,$CD,$AE,$17,$D2,$FC,$C7,$83,$0F
	.dc.b	$1B,$88,$EF,$87,$38,$FF,$17,$DF,$38,$C4,$7F,$8D,$23,$15,$C9,$B1
	.dc.b	$D4,$FE,$79,$7E,$93,$F6,$CF,$A4,$FF,$02,$57,$F4,$88,$D5,$7B,$0F
	.dc.b	$E0,$73,$B3,$FB,$CC,$FF,$4E,$F2,$3F,$77,$3F,$C2,$1B,$3F,$D2,$DF
	.dc.b	$F2,$5C,$67,$5E,$04,$78,$4D,$3F,$A9,$78,$8F,$3B,$CF,$1F,$E5,$48
	.dc.b	$C4,$FF,$88,$1B,$1F,$E2,$66,$BF,$E5,$67,$38,$FF,$8D,$7D,$5F,$E1
	.dc.b	$15,$FC,$7A,$72,$FF,$89,$63,$35,$E9,$E7,$F7,$1F,$E0,$9F,$80,$7A
	.dc.b	$58,$BF,$81,$3E,$6A,$69,$FD,$0A,$7F,$C6,$AD,$8F,$F1,$DB,$5F,$F3
	.dc.b	$4C,$F5,$EF,$E9,$1F,$E1,$2C,$3F,$AD,$67,$FC,$B7,$D5,$7A,$FF,$DF
	.dc.b	$9B,$97,$F8,$DF,$A9,$E8,$66,$A4,$FF,$91,$AB,$78,$FF,$93,$1A,$FF
	.dc.b	$9F,$1A,$FF,$A0,$1D,$E1,$FE,$5D,$F1,$9F,$55,$D2,$BF,$C3,$0F,$8F
	.dc.b	$D8,$98,$78,$F8,$6D,$7F,$83,$71,$57,$C7,$6D,$79,$63,$04,$7F,$C3
	.dc.b	$1F,$31,$CB,$2F,$13,$F0,$7D,$89,$3F,$07,$D9,$B8,$F7,$9E,$C4,$9F
	.dc.b	$84,$D8,$FF,$18,$7C,$77,$0C,$3D,$CF,$CD,$F6,$A5,$FC,$DE,$6B,$7E
	.dc.b	$B9,$F5,$12,$7E,$73,$63,$FC,$79,$C1,$C3,$0F,$94,$FD,$9E,$89,$7F
	.dc.b	$67,$96,$DF,$C7,$9F,$8C,$49,$FB,$4D,$8F,$F2,$47,$D4,$F0,$C3,$EB
	.dc.b	$3F,$E4,$2F,$49,$ED,$D3,$FC,$C5,$CE,$F6,$F1,$97,$F8,$02,$B6,$FF
	.dc.b	$9E,$EB,$F3,$91,$8C,$7F,$CC,$31,$83,$FC,$BB,$EE,$72,$8F,$F2,$D3
	.dc.b	$FC,$17,$A6,$5F,$FB,$73,$B3,$F2,$AF,$F0,$78,$CD,$7F,$B6,$8F,$F7
	.dc.b	$04,$7F,$B5,$8F,$F7,$03,$FE,$EA,$6C,$FF,$BA,$A3,$FD,$B8,$7F,$B9
	.dc.b	$34,$FF,$6E,$4F,$7A,$EB,$1F,$EE,$29,$EA,$6E,$79,$CF,$FB,$8A,$39
	.dc.b	$9D,$CB,$5D,$DE,$26,$4A,$E4,$F6,$BF,$C8,$63,$C3,$3F,$EB,$1C,$DE
	.dc.b	$E6,$2B,$D3,$87,$F8,$D2,$31,$1E,$68,$C5,$7C,$72,$47,$FB,$F8,$F0
	.dc.b	$DF,$A5,$B1,$F0,$4D,$CF,$C1,$C3,$CA,$FF,$8D,$DB,$1F,$3E,$7D,$F5
	.dc.b	$CE,$73,$1E,$6C,$3D,$35,$D0,$46,$70,$F4,$D7,$08,$F8,$CD,$25,$F5
	.dc.b	$36,$5A,$E9,$3F,$54,$F1,$BE,$B7,$E6,$3A,$C3,$E2,$C6,$08,$FB,$E7
	.dc.b	$AC,$F1,$8F,$81,$1F,$45,$E2,$7E,$22,$7F,$25,$5E,$6A,$ED,$87,$C0
	.dc.b	$9E,$B1,$F0,$1F,$AD,$19,$AE,$47,$F9,$E4,$EB,$3F,$8E,$3C,$47,$1A
	.dc.b	$ED,$5E,$0B,$F3,$B6,$7E,$1B,$F2,$A3,$35,$D6,$FC,$15,$D6,$B8,$1E
	.dc.b	$63,$E5,$1F,$84,$33,$3F,$0A,$3D,$B8,$74,$7F,$05,$3B,$9E,$D3,$98
	.dc.b	$E7,$87,$A2,$7E,$75,$79,$A3,$F4,$58,$7E,$F2,$BE,$31,$FD,$91,$FE
	.dc.b	$CD,$1F,$A6,$8C,$B5,$F8,$91,$A2,$4F,$F6,$A6,$C7,$FB,$DB,$07,$B9
	.dc.b	$F6,$DB,$5F,$62,$30,$47,$04,$9F,$8D,$87,$FA,$92,$75,$18,$AE,$6D
	.dc.b	$8C,$CF,$C9,$8F,$14,$62,$BD,$27,$F7,$37,$11,$FA,$6C,$3F,$57,$5D
	.dc.b	$EB,$C4,$F3,$3E,$E5,$70,$AF,$91,$3C,$CD,$9D,$9F,$0B,$87,$B6,$1F
	.dc.b	$CF,$A7,$95,$F8,$30,$F5,$E1,$DF,$0F,$E2,$95,$AB,$E1,$1C,$F0,$F0
	.dc.b	$9D,$8F,$84,$E7,$0E,$95,$C6,$3C,$D3,$A2,$5F,$74,$FC,$17,$31,$EB
	.dc.b	$8F,$82,$F5,$BF,$1C,$E8,$F6,$C7,$69,$E8,$73,$2E,$1E,$E8,$E8,$7B
	.dc.b	$8E,$87,$13,$33,$E4,$8E,$0F,$63,$E0,$4F,$4B,$F7,$46,$2B,$8C,$F8
	.dc.b	$5B,$5C,$A7,$E4,$CF,$B1,$F5,$18,$9F,$23,$8C,$3E,$5C,$79,$A3,$B1
	.dc.b	$E3,$9E,$47,$8E,$34,$E1,$F0,$DF,$9A,$31,$3C,$67,$BC,$71,$8F,$06
	.dc.b	$1F,$3A,$30,$49,$E6,$76,$38,$98,$9E,$2F,$CB,$7B,$1F,$2E,$BC,$51
	.dc.b	$C2,$33,$3F,$1D,$B8,$7C,$7C,$3E,$0C,$74,$3E,$09,$F1,$CE,$95,$E3
	.dc.b	$69,$BA,$F7,$BB,$7B,$36,$BB,$61,$EF,$AF,$55,$7B,$EB,$C7,$87,$EC
	.dc.b	$74,$7B,$7F,$E2,$B3,$FC,$53,$EE,$5C,$3F,$C5,$11,$FE,$26,$23,$FC
	.dc.b	$82,$FF,$B2,$D7,$28,$C4,$23,$0F,$E9,$23,$FB,$5E,$0E,$EF,$F3,$2C
	.dc.b	$E4,$E7,$7F,$D3,$6B,$73,$AA,$47,$F5,$5B,$7B,$9F,$E4,$33,$86,$1F
	.dc.b	$B2,$BD,$D7,$F9,$06,$74,$F9,$AB,$8B,$63,$FC,$86,$47,$CE,$78,$BE
	.dc.b	$A7,$8D,$F9,$F0,$F5,$47,$F9,$4F,$C3,$95,$C2,$BC,$5E,$EE,$1F,$71
	.dc.b	$F1,$67,$FC,$A2,$49,$F2,$36,$3B,$65,$F2,$E7,$FC,$B5,$3E,$EC,$3E
	.dc.b	$56,$5F,$E6,$3F,$72,$57,$86,$7E,$77,$BB,$8C,$7C,$36,$BF,$F3,$09
	.dc.b	$2F,$E2,$36,$7C,$3D,$DC,$70,$FF,$53,$3F,$E7,$1F,$72,$5F,$7C,$DE
	.dc.b	$EE,$30,$EF,$5F,$E6,$E2,$57,$D1,$6C,$7B,$BE,$A7,$86,$3F,$EB,$37
	.dc.b	$ED,$C3,$DE,$7A,$98,$6E,$1E,$1F,$6B,$13,$23,$F2,$73,$AA,$FC,$C3
	.dc.b	$5F,$EF,$2D,$7F,$C4,$0E,$F0,$FF,$62,$7C,$97,$39,$FA,$6F,$93,$DD
	.dc.b	$78,$5F,$EE,$56,$E1,$FA,$26,$D7,$E9,$DA,$FF,$82,$3D,$D6,$6F,$77
	.dc.b	$F5,$31,$86,$EF,$F0,$2B,$F6,$13,$97,$F6,$5F,$00,$F0,$7F,$63,$08
	.dc.b	$C3,$FB,$46,$BC,$23,$7E,$EB,$93,$FC,$81,$BF,$EB,$AB,$FC,$C3,$EF
	.dc.b	$3D,$3C,$EF,$FE,$45,$8C,$4F,$E7,$C6,$1C,$DC,$35,$FE,$10,$D7,$FC
	.dc.b	$52,$D7,$FC,$A0,$D7,$FC,$E3,$EE,$BD,$6F,$0C,$7F,$5B,$F7,$AE,$4F
	.dc.b	$15,$1A,$95,$F1,$7C,$CD,$7F,$66,$DA,$F8,$11,$C2,$F9,$4B,$DE,$E9
	.dc.b	$ED,$BD,$9B,$D5,$30,$F8,$53,$96,$CF,$02,$3F,$1F,$DE,$E9,$97,$FA
	.dc.b	$3B,$DC,$5F,$F9,$C6,$73,$49,$CB,$FE,$94,$9E,$CE,$5F,$F3,$C1,$88
	.dc.b	$DC,$27,$9C,$61,$B3,$C2,$FF,$D0,$13,$C2,$39,$1E,$3C,$7F,$C3,$B7
	.dc.b	$DC,$FF,$52,$DF,$27,$C9,$7C,$AB,$A4,$7B,$5E,$58,$58,$E3,$7D,$30
	.dc.b	$F0,$5F,$19,$E6,$6C,$F5,$4E,$EF,$9C,$78,$CF,$93,$3C,$5A,$71,$C3
	.dc.b	$CC,$F2,$C3,$FA,$BC,$7C,$E8,$FF,$3D,$B5,$EA,$63,$DD,$7B,$DC,$1B
	.dc.b	$1F,$D8,$E5,$18,$8E,$13,$FB,$45,$EF,$52,$3F,$B8,$6E,$7F,$63,$0F
	.dc.b	$D0,$71,$FC,$1A,$FF,$69,$D7,$C3,$9C,$C7,$C1,$32,$7C,$37,$35,$F8
	.dc.b	$28,$CB,$D7,$7F,$DF,$39,$8F,$29,$C6,$33,$A7,$F7,$FB,$F0,$46,$27
	.dc.b	$FD,$43,$18,$C3,$CD,$1C,$AF,$B4,$62,$75,$3E,$73,$57,$FC,$42,$FA
	.dc.b	$47,$F9,$A9,$FE,$83,$87,$B8,$FD,$AC,$7A,$A7,$85,$FF,$11,$9D,$38
	.dc.b	$9F,$ED,$E6,$CF,$F0,$8D,$7F,$4C,$F3,$BF,$75,$3F,$C6,$5B,$3F,$DD
	.dc.b	$9F,$F8,$03,$19,$D7,$81,$1E,$13,$4F,$EF,$BE,$3B,$D3,$FE,$ED,$EE
	.dc.b	$7F,$09,$AF,$E4,$D3,$97,$FB,$B4,$66,$B8,$D7,$0E,$1F,$EC,$37,$B9
	.dc.b	$FF,$08,$56,$F3,$FE,$2F,$97,$F3,$5C,$7F,$C6,$2D,$AF,$F1,$B3,$5F
	.dc.b	$F6,$9F,$47,$FB,$58,$FF,$05,$7B,$DF,$D5,$AF,$9C,$63,$C2,$F8,$CD
	.dc.b	$C3,$E3,$36,$FF,$C3,$7E,$EB,$D7,$FA,$B0,$11,$9A,$FF,$22,$B6,$3F
	.dc.b	$C9,$4D,$7F,$DD,$6E,$70,$FE,$A4,$7F,$91,$FD,$EF,$F0,$1C,$66,$BF
	.dc.b	$C8,$B1,$8B,$FF,$12,$79,$48,$C4,$FF,$85,$5E,$3E,$F7,$9B,$1F,$2B
	.dc.b	$70,$FF,$29,$C7,$8A,$30,$74,$F7,$B8,$DF,$F8,$7D,$B3,$FE,$36,$6E
	.dc.b	$5E,$8C,$51,$E8,$F0,$7B,$12,$FE,$1B,$6F,$FC,$7F,$EE,$EE,$7F,$C5
	.dc.b	$8D,$8C,$C7,$F6,$FC,$BF,$BC,$3A,$31,$EE,$7A,$C8,$FD,$06,$CF,$F9
	.dc.b	$33,$5F,$97,$1F,$21,$B3,$FE,$3B,$6C,$EE,$3C,$1C,$BF,$1C,$8F,$DC
	.dc.b	$6C,$7F,$96,$3D,$CF,$AF,$1F,$51,$B3,$FE,$47,$6C,$EE,$3F,$C1,$5E
	.dc.b	$8B,$DB,$FA,$CD,$8F,$F3,$C4,$60,$8F,$F9,$47,$C7,$FF,$23,$CE,$EB
	.dc.b	$FC,$D3,$18,$AF,$C2,$B6,$B8,$1F,$C0,$7D,$AF,$E0,$87,$F9,$B7,$C6
	.dc.b	$7B,$78,$9D,$13,$0F,$F4,$14,$EA,$31,$C7,$FB,$4B,$E1,$5A,$BB,$39
	.dc.b	$AD,$9F,$EA,$4B,$E5,$5F,$C4,$6F,$73,$FE,$F0,$BB,$3F,$ED,$23,$FD
	.dc.b	$80,$E1,$FF,$65,$18,$8E,$AE,$FA,$55,$ED,$AE,$11,$8A,$DE,$1F,$E2
	.dc.b	$B9,$CD,$78,$23,$13,$D2,$FB,$DF,$17,$DD,$19,$8C,$4F,$BE,$72,$7B
	.dc.b	$67,$55,$E3,$69,$E2,$C3,$E1,$D7,$12,$47,$C3,$37,$5F,$C7,$AB,$F7
	.dc.b	$97,$E3,$8F,$E2,$47,$F1,$37,$0F,$FB,$92,$63,$FE,$E4,$7F,$DE,$15
	.dc.b	$FE,$EC,$9F,$F7,$BB,$FE,$E9,$9F,$F7,$77,$B7,$FF,$00,$1F,$F7,$AC
	.dc.b	$66,$77,$5D,$27,$6F,$4C,$3F,$DE,$AF,$FA,$D5,$B1,$F8,$AB,$E9,$7E
	.dc.b	$1C,$3F,$DF,$53,$E5,$3E,$41,$BB,$E9,$5C,$2F,$9D,$FF,$28,$6C,$FE
	.dc.b	$36,$74,$47,$89,$2B,$AD,$71,$9D,$46,$27,$91,$B9,$F8,$17,$E6,$9F
	.dc.b	$79,$B3,$E2,$CF,$C3,$C3,$E2,$C7,$AE,$B8,$CF,$C3,$BE,$F1,$89,$F7
	.dc.b	$1C,$DF,$89,$3B,$3C,$47,$5B,$F1,$5F,$6A,$F1,$D7,$33,$D9,$5C,$23
	.dc.b	$C4,$F1,$3A,$3D,$F0,$F9,$2F,$A8,$D9,$0E,$B1,$C5,$F3,$CF,$78,$F8
	.dc.b	$F1,$CF,$0F,$6C,$6E,$7C,$73,$98,$F2,$3D,$0E,$D7,$EB,$26,$1E,$E2
	.dc.b	$4F,$A2,$BC,$15,$E7,$39,$12,$B8,$CE,$CF,$11,$E6,$9E,$27,$CE,$9F
	.dc.b	$5C,$69,$E9,$3C,$9C,$9E,$6A,$F1,$B8,$3C,$0E,$2B,$DE,$76,$C3,$E6
	.dc.b	$BD,$CC,$DF,$77,$0F,$94,$C4,$70,$26,$1F,$32,$FC,$53,$ED,$69,$E3
	.dc.b	$C3,$D1,$7D,$A3,$2D,$7C,$F3,$C5,$A7,$1C,$3D,$51,$E3,$8E,$2D,$77
	.dc.b	$3C,$EB,$BC,$E5,$B3,$F2,$AB,$9B,$67,$C1,$3A,$AF,$83,$39,$3C,$53
	.dc.b	$AA,$E3,$3A,$26,$1F,$12,$7F,$17,$7C,$E3,$17,$F8,$F2,$57,$D3,$C3
	.dc.b	$E5,$47,$AA,$FD,$18,$76,$BE,$37,$FE,$E6,$D1,$ED,$7F,$76,$25,$FF
	.dc.b	$30,$C3,$F9,$8C,$EA,$31,$1F,$E4,$58,$C2,$AF,$F1,$47,$3F,$F7,$AB
	.dc.b	$FF,$1D,$B6,$FF,$C9,$93,$A7,$FC,$0F,$3A,$7F,$C1,$3E,$D7,$F8,$20
	.dc.b	$8F,$82,$D8,$27,$31,$E1,$9E,$BD,$0F,$6B,$FC,$17,$5F,$E4,$C6,$CF
	.dc.b	$88,$DB,$8A,$E4,$7A,$EB,$94,$FF,$87,$3B,$FF,$C3,$64,$7C,$98,$23
	.dc.b	$15,$CB,$25,$FA,$2F,$91,$E1,$9E,$D7,$FE,$20,$BF,$F3,$0B,$67,$57
	.dc.b	$CA,$3F,$C4,$BE,$C7,$F8,$B0,$93,$E2,$6D,$F8,$70,$FF,$26,$BE,$C3
	.dc.b	$E4,$C7,$F9,$BB,$AB,$9D,$76,$AD,$E1,$F8,$5A,$D4,$FF,$8F,$3D,$8F
	.dc.b	$F1,$D9,$1F,$82,$DC,$3C,$1E,$C7,$F8,$EA,$3F,$CF,$6D,$9E,$CD,$9D
	.dc.b	$C7,$C3,$5E,$C3,$CD,$FF,$3F,$36,$7F,$04,$D9,$FC,$9B,$5F,$CC,$39
	.dc.b	$C7,$FC,$F5,$EC,$7F,$A4,$AB,$FC,$17,$39,$7F,$0F,$19,$AF,$F5,$04
	.dc.b	$E5,$FC,$14,$65,$F2,$DF,$FA,$5A,$31,$EA,$3D,$77,$AF,$F9,$78,$C3
	.dc.b	$7D,$7F,$66,$1F,$AA,$9C,$BF,$AC,$F5,$DD,$BF,$9F,$84,$61,$FD,$13
	.dc.b	$5E,$11,$BA,$7C,$4D,$24,$FE,$A6,$32,$FE,$AF,$D7,$72,$77,$6C,$66
	.dc.b	$30,$47,$EB,$46,$9F,$C6,$37,$81,$C1,$FE,$38,$F6,$1F,$0F,$FE,$3A
	.dc.b	$9C,$DF,$FB,$37,$D8,$7A,$59,$C3,$FC,$D5,$EC,$DA,$FE,$19,$E0,$23
	.dc.b	$10,$BE,$2F,$E2,$5A,$FE,$7D,$B8,$78,$23,$86,$1C,$A5,$E9,$74,$F5
	.dc.b	$1E,$A5,$E8,$98,$7E,$56,$72,$D9,$E0,$47,$F3,$5E,$CF,$4C,$BF,$DD
	.dc.b	$5E,$BA,$FF,$CB,$73,$9A,$4E,$5F,$F7,$84,$F6,$72,$FF,$98,$8C,$46
	.dc.b	$E1,$3C,$E3,$0D,$9E,$17,$FE,$6A,$9E,$11,$C8,$F1,$E3,$FC,$E6,$FB
	.dc.b	$9F,$F0,$06,$F9,$3E,$4B,$E5,$5D,$23,$DA,$F2,$C2,$C7,$1B,$E9,$87
	.dc.b	$82,$F8,$CF,$33,$67,$AA,$77,$7C,$E3,$C6,$7C,$99,$E2,$D3,$8E,$1E
	.dc.b	$67,$96,$1F,$D1,$E3,$E7,$47,$F5,$B6,$BD,$4C,$7B,$2F,$67,$83,$63
	.dc.b	$FA,$FC,$A3,$11,$C2,$7F,$6C,$BD,$9A,$47,$F7,$4D,$CF,$EC,$61,$FA
	.dc.b	$0E,$1F,$83,$43,$91,$F0,$23,$4F,$C2,$35,$5F,$22,$33,$5F,$83,$9C
	.dc.b	$9C,$77,$FD,$7C,$66,$3A,$BE,$28,$CE,$9F,$DF,$2F,$A4,$62,$BF,$C0
	.dc.b	$51,$8C,$3C,$F1,$CA,$BF,$79,$5B,$8F,$F5,$BD,$6D,$F4,$1B,$3F,$C0
	.dc.b	$31,$BC,$3D,$A7,$ED,$4B,$7E,$39,$F5,$3F,$D1,$63,$31,$CA,$3F,$A4
	.dc.b	$3F,$03,$0F,$64,$7F,$4B,$F2,$FF,$76,$DD,$7F,$9D,$47,$CF,$8C,$12
	.dc.b	$BF,$B3,$47,$06,$9C,$27,$FA,$17,$C6,$BC,$EA,$7F,$49,$C5,$ED,$7E
	.dc.b	$1F,$1F,$EB,$95,$B6,$D7,$F1,$3C,$3F,$DB,$B3,$97,$FC,$5A,$D7,$FC
	.dc.b	$64,$D7,$FD,$C7,$EA,$3D,$3F,$EA,$8D,$BF,$F6,$A7,$9F,$89,$91,$FD
	.dc.b	$B6,$75,$5F,$E1,$5E,$8F,$F7,$EC,$F4,$AD,$CF,$F8,$FA,$B6,$FF,$90
	.dc.b	$DA,$EE,$7F,$C6,$9E,$13,$D7,$FE,$CC,$D8,$FF,$0C,$FB,$4F,$7F,$35
	.dc.b	$27,$FC,$01,$5B,$BF,$F1,$45,$EE,$73,$5B,$9F,$F2,$8D,$6D,$FF,$2B
	.dc.b	$35,$FD,$51,$FD,$CA,$3E,$73,$FA,$DF,$D0,$DE,$B6,$21,$18,$BF,$F2
	.dc.b	$73,$63,$FC,$BE,$D7,$FC,$C8,$BD,$4F,$F0,$23,$67,$FC,$34,$F1,$E2
	.dc.b	$FE,$A0,$E6,$BF,$C6,$4D,$79,$D7,$8E,$31,$1E,$A6,$BE,$9F,$4B,$44
	.dc.b	$BF,$3B,$C7,$C4,$F1,$9E,$46,$C7,$92,$34,$FB,$9B,$3E,$EF,$37,$C4
	.dc.b	$49,$F8,$6F,$97,$D8,$F7,$9F,$01,$B1,$F0,$23,$4F,$CA,$6C,$FC,$AF
	.dc.b	$A8,$F7,$12,$7E,$7B,$F0,$7D,$9F,$96,$7C,$C6,$C7,$CC,$8D,$3F,$59
	.dc.b	$B3,$FE,$94,$FD,$13,$FC,$63,$3F,$E8,$76,$C6,$78,$3F,$01,$18,$8F
	.dc.b	$F4,$F3,$FB,$1D,$BF,$C0,$CD,$AF,$C5,$BC,$A3,$3D,$1F,$E4,$38,$FD
	.dc.b	$A5,$7F,$39,$6A,$69,$A9,$FF,$41,$E5,$FE,$4B,$F9,$2F,$F4,$6D,$7F
	.dc.b	$9D,$E3,$1B,$7F,$0B,$F5,$51,$8C,$BF,$CE,$CD,$AF,$F3,$E5,$F2,$C3
	.dc.b	$FC,$3C,$DA,$FF,$51,$D6,$E7,$38,$59,$C9,$FE,$71,$AE,$17,$FD,$4D
	.dc.b	$A9,$CD,$7A,$A7,$85,$73,$AE,$0F,$A6,$F1,$1C,$63,$55,$C6,$3D,$31
	.dc.b	$AB,$EB,$87,$C0,$C3,$A9,$2B,$6F,$C2,$31,$7E,$C6,$9E,$C8,$D9,$23
	.dc.b	$4F,$CF,$8D,$BE,$86,$C7,$94,$D9,$1F,$E1,$B5,$E6,$C3,$D1,$87,$B3
	.dc.b	$0F,$46,$1D,$EF,$73,$9A,$DC,$7F,$B6,$EF,$F0,$35,$FE,$EF,$6B,$C2
	.dc.b	$35,$5F,$CB,$3E,$57,$E5,$E8,$69,$27,$D7,$5F,$E9,$28,$C3,$E3,$3F
	.dc.b	$C1,$D3,$C6,$76,$D8,$D1,$C6,$33,$3C,$B2,$F8,$53,$CA,$31,$3E,$78
	.dc.b	$FF,$67,$57,$3A,$DC,$73,$3A,$C6,$67,$84,$FF,$9D,$63,$13,$F4,$CE
	.dc.b	$75,$D9,$C4,$EE,$31,$1F,$8D,$AF,$E9,$71,$8C,$3F,$33,$7D,$2B,$47
	.dc.b	$4A,$F6,$91,$F2,$D7,$1C,$3E,$84,$E7,$0F,$5D,$F3,$C3,$AD,$EE,$BC
	.dc.b	$26,$CF,$69,$FD,$52,$78,$5F,$4C,$3B,$CE,$DF,$6B,$BA,$E9,$1E,$73
	.dc.b	$64,$C3,$C7,$1A,$7E,$24,$F3,$9F,$7C,$61,$E1,$1E,$47,$ED,$3B,$C3
	.dc.b	$D6,$DB,$F3,$57,$68,$C1,$EE,$AE,$67,$A0,$E6,$60,$F0,$BE,$B3,$8B
	.dc.b	$EB,$8E,$8E,$EB,$C2,$DA,$F0,$D7,$9E,$3F,$15,$5C,$0F,$61,$23,$58
	.dc.b	$7D,$F2,$61,$F8,$66,$C7,$C5,$6D,$79,$DB,$5D,$67,$D1,$3E,$77,$AD
	.dc.b	$79,$23,$DF,$3B,$6C,$F0,$8C,$47,$A2,$30,$72,$3E,$54,$7A,$63,$E5
	.dc.b	$1B,$39,$90,$F9,$65,$AF,$3C,$F3,$8F,$31,$BC,$3E,$5D,$75,$8E,$87
	.dc.b	$22,$D7,$27,$C2,$D7,$73,$CA,$BD,$31,$D7,$0F,$84,$F2,$C3,$E7,$18
	.dc.b	$AE,$11,$E7,$8C,$D7,$C0,$3D,$18,$79,$70,$FD,$2C,$66,$B8,$3F,$5E
	.dc.b	$BA,$61,$E0,$AF,$36,$1D,$AF,$C5,$7B,$39,$3E,$F9,$F6,$1D,$9F,$8E
	.dc.b	$70,$3F,$05,$7D,$A7,$8C,$F6,$BF,$04,$E6,$B8,$CE,$6B,$73,$9C,$3D
	.dc.b	$8D,$8F,$93,$5D,$E7,$F0,$B5,$C6,$39,$D6,$DF,$82,$7F,$98,$1F,$F2
	.dc.b	$C4,$7F,$37,$8F,$F3,$0C,$60,$57,$FB,$2E,$AC,$FF,$B3,$7E,$E5,$ED
	.dc.b	$7D,$86,$D7,$C6,$6C,$FC,$46,$D7,$E2,$5B,$3E,$A6,$CF,$E5,$5A,$FE
	.dc.b	$39,$B8,$7F,$31,$6E,$1E,$4B,$F2,$D2,$73,$7E,$93,$FC,$F6,$E2,$13
	.dc.b	$C2,$30,$DC,$3D,$87,$5A,$E2,$DA,$F8,$4E,$A3,$B4,$61,$AF,$8E,$34
	.dc.b	$F8,$23,$99,$AA,$F0,$B8,$8D,$1C,$CA,$73,$7E,$74,$F8,$E3,$06,$A7
	.dc.b	$9B,$C6,$BD,$95,$DE,$BC,$71,$89,$FF,$4F,$9D,$CB,$5F,$69,$B3,$FE
	.dc.b	$9F,$8F,$AB,$93,$D9,$F1,$46,$1F,$27,$E8,$AF,$47,$FD,$4C,$DC,$BB
	.dc.b	$E2,$11,$97,$2D,$7C,$AB,$D6,$E6,$49,$E7,$3C,$A3,$BF,$54,$9C,$D7
	.dc.b	$FA,$F6,$BC,$0D,$7C,$3F,$01,$E6,$24,$F0,$7D,$3F,$01,$23,$9B,$67
	.dc.b	$CF,$E0,$7B,$89,$3E,$07,$E0,$F6,$C8,$F1,$B6,$7D,$FE,$B7,$C6,$24
	.dc.b	$FA,$1F,$93,$DD,$23,$D6,$D9,$F8,$FC,$BF,$38,$93,$F0,$1F,$A3,$BC
	.dc.b	$8F,$86,$D9,$F9,$FD,$BF,$58,$93,$F2,$1F,$B3,$DB,$23,$E5,$B6,$7E
	.dc.b	$DF,$13,$DD,$FC,$31,$2B,$EA,$C6,$5F,$AF,$F7,$4F,$97,$FF,$75,$36
	.dc.b	$33,$18,$3F,$DD,$DF,$16,$F8,$CA,$49,$FC,$0C,$65,$FC,$1F,$C3,$BE
	.dc.b	$1F,$FD,$EC,$D8,$CC,$60,$FF,$7D,$7B,$FF,$8D,$EF,$FC,$5B,$75,$FC
	.dc.b	$BB,$F9,$BF,$63,$3D,$6A,$FC,$6F,$8D,$F9,$BB,$FD,$CF,$80,$F0,$D3
	.dc.b	$98,$FE,$31,$18,$9F,$E0,$FD,$B5,$FD,$FB,$67,$75,$C7,$4F,$E6,$97
	.dc.b	$97,$C1,$56,$FF,$3B,$5F,$A4,$7F,$5A,$D8,$ED,$1F,$AC,$8D,$7E,$8E
	.dc.b	$F9,$3E,$D3,$FC,$9B,$0F,$F4,$33,$FC,$7B,$1F,$67,$89,$A8,$5F,$17
	.dc.b	$F9,$0B,$5F,$87,$5C,$9B,$3C,$67,$55,$FC,$EB,$DB,$7A,$BE,$28,$79
	.dc.b	$EF,$73,$0D,$C7,$F9,$7D,$F0,$6D,$F9,$63,$11,$FC,$A9,$B7,$DF,$DC
	.dc.b	$F5,$7B,$2F,$5A,$F2,$4C,$3F,$9D,$CE,$5B,$3C,$08,$FF,$40,$F7,$17
	.dc.b	$FD,$3A,$1E,$CA,$FF,$0D,$5A,$C7,$F9,$14,$E6,$BF,$92,$57,$77,$8B
	.dc.b	$F8,$D3,$11,$B8,$57,$37,$67,$F2,$D7,$11,$B9,$E2,$73,$C3,$F3,$11
	.dc.b	$CE,$77,$3E,$6C,$3D,$53,$C6,$B7,$3C,$6F,$D2,$EF,$0B,$1E,$1B,$E3
	.dc.b	$3B,$8F,$29,$F0,$EB,$9B,$BA,$E2,$FB,$2B,$8B,$F1,$27,$6F,$13,$78
	.dc.b	$7A,$1E,$47,$F4,$E7,$E2,$3F,$D7,$9A,$F9,$CD,$7B,$8F,$73,$FA,$5C
	.dc.b	$A3,$13,$A9,$CC,$62,$3F,$A8,$FB,$9F,$AF,$4D,$23,$FB,$37,$DB,$8F
	.dc.b	$E7,$B9,$FD,$F3,$FD,$8E,$73,$BF,$F8,$C3,$5F,$BD,$39,$3E,$4D,$F1
	.dc.b	$7E,$59,$98,$FC,$24,$65,$E3,$1E,$43,$8C,$F5,$D7,$FB,$9D,$F8,$30
	.dc.b	$F1,$D6,$EF,$9D,$EE,$3F,$C0,$F5,$B7,$C2,$6E,$FF,$8F,$4E,$E3,$F9
	.dc.b	$24,$EE,$B9,$DF,$F8,$9B,$0F,$51,$FE,$3A,$AF,$56,$1C,$A7,$58,$7F
	.dc.b	$87,$EB,$71,$E8,$AF,$5B,$FE,$30,$F7,$3F,$74,$DD,$7F,$C4,$D1,$FE
	.dc.b	$0F,$8C,$12,$7F,$C5,$B1,$C1,$A7,$09,$FF,$17,$7B,$6F,$57,$CD,$3F
	.dc.b	$E1,$98,$C5,$FF,$84,$9B,$1F,$E1,$E6,$BF,$E2,$E7,$38,$7F,$96,$7D
	.dc.b	$CF,$F2,$FC,$62,$BF,$CE,$11,$FE,$62,$F6,$DE,$7F,$F8,$FD,$B3,$FC
	.dc.b	$CF,$99,$EE,$66,$A4,$FF,$89,$2B,$79,$7F,$8C,$1A,$FF,$90,$1A,$FF
	.dc.b	$59,$77,$97,$F5,$6F,$4F,$9C,$FF,$3A,$7F,$CA,$1E,$01,$A7,$F9,$96
	.dc.b	$5E,$DB,$DC,$CE,$3F,$E6,$4B,$FF,$18,$12,$FF,$D2,$64,$8F,$F3,$4F
	.dc.b	$AE,$F4,$B8,$E5,$FE,$62,$8C,$4F,$F9,$BB,$DC,$B3,$FE,$66,$6E,$7F
	.dc.b	$E8,$45,$EA,$FF,$99,$DC,$D7,$F7,$77,$8D,$7F,$98,$9A,$E7,$0F,$BE
	.dc.b	$A3,$2F,$F9,$65,$E6,$FF,$96,$FC,$1F,$F0,$D3,$E3,$9F,$F3,$B4,$63
	.dc.b	$0F,$27,$A5,$E3,$3C,$9E,$A5,$C7,$D9,$1E,$0C,$4D,$24,$7F,$98,$23
	.dc.b	$2F,$F7,$0F,$93,$E5,$97,$FA,$23,$14,$7C,$1F,$6E,$57,$C9,$F5,$6D
	.dc.b	$FC,$99,$F0,$12,$7B,$36,$3A,$76,$F0,$D3,$FD,$63,$8F,$F5,$DB,$FA
	.dc.b	$9E,$EF,$CF,$3E,$87,$97,$63,$27,$A6,$3D,$C4,$70,$7F,$82,$BE,$4B
	.dc.b	$86,$3E,$B7,$FD,$41,$EF,$7E,$9E,$3F,$D1,$4D,$4F,$F8,$D6,$7F,$D6
	.dc.b	$6E,$4F,$D9,$C7,$F8,$DD,$AF,$2E,$EF,$BF,$18,$8E,$71,$F6,$BD,$FF
	.dc.b	$0B,$6B,$F2,$FC,$5F,$EB,$3E,$53,$FA,$3F,$35,$C1,$98,$95,$FE,$98
	.dc.b	$AD,$B6,$3F,$1F,$18,$C7,$FD,$4E,$DD,$FF,$DB,$FD,$0F,$37,$9D,$7F
	.dc.b	$B1,$A3,$18,$79,$0F,$D8,$4E,$AB,$8B,$63,$FC,$B1,$3C,$2B,$F5,$D8
	.dc.b	$7E,$FA,$B5,$1F,$EB,$79,$D5,$FF,$BC,$30,$F2,$61,$CA,$F7,$7F,$E6
	.dc.b	$56,$DF,$09,$FF,$33,$57,$0A,$E9,$5A,$24,$FA,$63,$D9,$18,$7C,$B1
	.dc.b	$EC,$35,$7E,$C6,$9E,$32,$C7,$66,$BE,$56,$DF,$96,$3F,$DA,$97,$E5
	.dc.b	$74,$78,$0F,$1C,$F0,$BF,$25,$FC,$3C,$3E,$44,$F9,$6B,$67,$19,$F3
	.dc.b	$38,$69,$C6,$38,$38,$7D,$51,$C5,$B1,$98,$F6,$19,$2B,$97,$D2,$F5
	.dc.b	$7C,$11,$C8,$8F,$C0,$9F,$4B,$A3,$9C,$7D,$88,$CC,$F5,$C3,$E4,$53
	.dc.b	$0F,$91,$18,$26,$1F,$1B,$0E,$F5,$E4,$7A,$1F,$00,$CC,$F4,$AE,$75
	.dc.b	$E0,$9E,$31,$87,$85,$79,$8F,$53,$87,$B1,$93,$CD,$87,$37,$17,$DE
	.dc.b	$31,$1C,$09,$5D,$DF,$88,$62,$37,$3E,$EA,$FC,$21,$E1,$7A,$36,$FA
	.dc.b	$DF,$3C,$3E,$54,$7C,$5C,$3C,$6D,$7C,$57,$C9,$F0,$BD,$DF,$C0,$1A
	.dc.b	$8E,$B8,$79,$1B,$87,$84,$F8,$98,$78,$5F,$9F,$1D,$EF,$A6,$1F,$87
	.dc.b	$3C,$B3,$C2,$FA,$34,$E9,$87,$C4,$C3,$A3,$EA,$8F,$86,$FA,$30,$EC
	.dc.b	$79,$1C,$47,$36,$9C,$EB,$BB,$FE,$ED,$3F,$3E,$5C,$3F,$8C,$36,$BF
	.dc.b	$BB,$C6,$23,$F5,$26,$6B,$93,$C8,$FC,$9F,$B0,$F7,$FF,$02,$D9,$FC
	.dc.b	$53,$67,$F0,$2D,$CB,$F5,$CD,$9E,$75,$C8,$8F,$26,$D7,$26,$D7,$26
	.dc.b	$CF,$23,$81,$27,$83,$67,$83,$4E,$11,$A2,$46,$9B,$1A,$8C,$11,$C1
	.dc.b	$23,$F5,$15,$ED,$6D,$79,$9B,$3E,$16,$9D,$8B,$5D,$5B,$1D,$5F,$D7
	.dc.b	$57,$56,$D7,$56,$CF,$56,$9D,$4B,$5D,$5B,$1D,$6F,$F8,$AD,$F5,$6D
	.dc.b	$F5,$6C,$F5,$6D,$75,$AF,$21,$C7,$05,$25,$86,$1C,$57,$0C,$31,$39
	.dc.b	$94,$F3,$AE,$53,$98,$E1,$18,$3B,$4F,$0A,$F8,$93,$A8,$F3,$47,$69
	.dc.b	$DC,$F5,$9F,$2C,$62,$75,$1E,$66,$BD,$63,$E7,$4F,$E7,$E3,$F6,$0D
	.dc.b	$79,$47,$EA,$9E,$07,$EB,$23,$ED,$64,$F6,$FA,$C6,$1B,$7F,$E2,$76
	.dc.b	$DF,$FA,$01,$B7,$FE,$7F,$6E,$1F,$E0,$F6,$CF,$F8,$29,$B3,$FD,$E1
	.dc.b	$AF,$F5,$36,$E1,$E3,$9F,$99,$3F,$05,$F0,$18,$7D,$75,$DE,$33,$1E
	.dc.b	$37,$58,$7A,$DC,$CE,$E7,$E7,$9D,$4E,$04,$74,$7D,$FC,$3E,$14,$F2
	.dc.b	$3D,$87,$67,$33,$C6,$32,$F6,$3C,$EE,$27,$AB,$4F,$AF,$1C,$4F,$8C
	.dc.b	$F7,$7C,$D1,$C1,$C9,$FE,$50,$C3,$FC,$A6,$49,$FF,$2C,$3F,$E9,$CE
	.dc.b	$DF,$2D,$7F,$A0,$9F,$C5,$FA,$4F,$4F,$FD,$52,$DB,$FC,$43,$70,$FE
	.dc.b	$66,$D7,$F9,$F3,$FE,$CD,$8D,$F7,$F2,$23,$F9,$1F,$D1,$DF,$0F,$FE
	.dc.b	$B6,$6C,$66,$30,$78,$BD,$F9,$1E,$26,$E1,$E2,$F1,$7C,$E4,$9F,$3B
	.dc.b	$D7,$E5,$64,$78,$5B,$3E,$AF,$80,$F7,$92,$7D,$EF,$C2,$F1,$24,$7C
	.dc.b	$26,$CF,$F5,$4E,$FF,$E1,$93,$FA,$D6,$C6,$4F,$F6,$9B,$87,$E3,$F0
	.dc.b	$FC,$E2,$4F,$CE,$7E,$8F,$89,$23,$E8,$B6,$7C,$1E,$AF,$42,$4F,$85
	.dc.b	$FF,$70,$F2,$BE,$02,$92,$7F,$81,$46,$5F,$E0,$9F,$01,$FE,$E2,$BF
	.dc.b	$EF,$0D,$7F,$8D,$B5,$FE,$54,$D4,$BC,$DF,$F7,$8B,$67,$AC,$60,$FF
	.dc.b	$7A,$FB,$AF,$76,$92,$7F,$8B,$46,$5F,$E3,$1E,$EF,$E5,$7B,$AF,$49
	.dc.b	$CF,$C1,$36,$CC,$7F,$1F,$F8,$67,$27,$F2,$1D,$91,$87,$51,$98,$C4
	.dc.b	$7E,$7B,$C9,$FE,$77,$DD,$FD,$13,$C4,$46,$35,$FE,$87,$18,$AF,$E2
	.dc.b	$9D,$D5,$FE,$00,$D7,$F4,$F5,$E3,$8C,$47,$F4,$86,$BF,$D5,$DE,$57
	.dc.b	$FD,$4D,$B7,$BA,$F1,$57,$F5,$CB,$79,$38,$87,$43,$DF,$DE,$4B,$FF
	.dc.b	$6A,$57,$F2,$E8,$C7,$89,$C6,$F8,$4F,$F2,$F6,$BF,$A1,$BF,$99,$18
	.dc.b	$8F,$F2,$03,$5F,$EC,$7E,$F3,$D8,$F0,$FA,$AF,$7E,$9E,$79,$D5,$7F
	.dc.b	$3D,$6B,$FC,$A5,$B9,$7F,$75,$6E,$DD,$BD,$FF,$81,$E8,$3D,$9B,$CD
	.dc.b	$31,$FE,$91,$39,$6C,$F0,$23,$FD,$37,$DF,$5F,$F8,$22,$1E,$E2,$FF
	.dc.b	$BF,$D6,$B3,$FF,$07,$4E,$6B,$FC,$21,$5D,$DE,$2F,$F8,$18,$C4,$6E
	.dc.b	$15,$CD,$D9,$FE,$19,$71,$1B,$9E,$27,$3C,$3F,$C2,$31,$CE,$77,$3E
	.dc.b	$6C,$3D,$53,$C6,$B7,$3C,$6F,$D2,$EF,$0B,$1E,$1B,$E3,$3B,$8F,$29
	.dc.b	$F0,$EB,$9B,$BA,$E2,$FB,$2B,$8B,$F1,$27,$6F,$13,$78,$7A,$1E,$47
	.dc.b	$FA,$09,$F8,$8F,$FA,$35,$AF,$9C,$D7,$BC,$F7,$FF,$C4,$12,$8C,$4E
	.dc.b	$A7,$31,$88,$FF,$13,$7B,$FF,$C1,$D3,$48,$FF,$0B,$7D,$B8,$FE,$77
	.dc.b	$5F,$7C,$FF,$93,$E7,$3C,$3F,$E4,$FD,$7E,$F4,$E4,$F9,$37,$C5,$F9
	.dc.b	$66,$63,$F0,$91,$97,$8C,$79,$0E,$33,$D7,$5F,$F5,$3D,$F8,$30,$F1
	.dc.b	$D6,$EF,$9D,$EE,$3F,$D6,$D5,$B7,$C2,$6E,$FF,$90,$CE,$E3,$FC,$BB
	.dc.b	$3B,$AE,$77,$FE,$33,$C3,$D4,$7F,$9A,$6B,$D5,$87,$29,$D6,$1F,$E5
	.dc.b	$AA,$DC,$7A,$2B,$D6,$FF,$8F,$7D,$FF,$DD,$37,$5F,$F3,$0C,$7F,$6E
	.dc.b	$AF,$F3,$B3,$AA,$FF,$17,$36,$33,$3C,$27,$8B,$FE,$6A,$F7,$DE,$AF
	.dc.b	$F9,$BA,$BF,$AD,$7A,$59,$8C,$78,$BF,$E9,$FA,$DE,$5F,$C9,$E3,$13
	.dc.b	$AA,$FF,$3A,$FB,$CF,$3B,$E1,$67,$FE,$91,$8C,$61,$FE,$73,$6C,$7F
	.dc.b	$B7,$9A,$FF,$BA,$5C,$E5,$FD,$47,$D2,$EB,$7E,$7E,$DF,$F1,$85,$FF
	.dc.b	$93,$6F,$78,$7F,$54,$BD,$BF,$CF,$6B,$78,$7F,$A7,$7C,$3E,$38,$7F
	.dc.b	$BD,$6F,$FD,$E6,$4B,$FF,$44,$4F,$07,$FD,$3B,$EA,$3C,$FF,$F4,$9C
	.dc.b	$E6,$FF,$D5,$D3,$9A,$FF,$3B,$36,$3F,$DF,$6D,$7F,$D4,$E1,$C3,$DA
	.dc.b	$BE,$4A,$F7,$32,$FF,$AE,$DA,$E6,$37,$18,$3C,$18,$FF,$AE,$FB,$DF
	.dc.b	$05,$FE,$AA,$9C,$E3,$FE,$B8,$9C,$BF,$D7,$E7,$FC,$72,$47,$FC,$7C
	.dc.b	$F6,$F4,$24,$7F,$AB,$69,$18,$C7,$DA,$7A,$5F,$F5,$CD,$BD,$DF,$14
	.dc.b	$61,$97,$FE,$4B,$7F,$C9,$24,$7B,$7B,$3E,$E2,$38,$25,$FF,$A3,$EF
	.dc.b	$FA,$B4,$7F,$A9,$F0,$57,$8B,$A6,$4F,$F9,$65,$FF,$2B,$91,$F6,$78
	.dc.b	$DF,$30,$8E,$09,$87,$7F,$73,$E6,$4F,$B9,$93,$FE,$65,$7F,$CC,$64
	.dc.b	$7F,$DA,$9E,$E7,$EB,$2B,$E3,$B5,$FC,$2B,$5F,$F7,$6B,$5F,$F7,$E3
	.dc.b	$C7,$6F,$BD,$EE,$7F,$B9,$A3,$15,$F7,$63,$E8,$9F,$E7,$42,$3F,$E7
	.dc.b	$97,$F1,$FC,$2F,$73,$F3,$BE,$57,$FB,$C2,$31,$5F,$ED,$1E,$44,$EA
	.dc.b	$30,$EE,$B1,$97,$FB,$06,$B7,$5A,$3F,$4B,$3A,$CB,$F6,$97,$67,$F5
	.dc.b	$F3,$6C,$7F,$98,$67,$0F,$EA,$62,$DF,$68,$F0,$4F,$8E,$BC,$11,$86
	.dc.b	$D9,$EE,$60,$EF,$1D,$DC,$3F,$BE,$BE,$F5,$E2,$C3,$E2,$CF,$52,$5F
	.dc.b	$B2,$FD,$E7,$28,$E8,$69,$F0,$B4,$F7,$E1,$EB,$9E,$51,$FE,$EB,$AE
	.dc.b	$F1,$82,$57,$A2,$B8,$3D,$23,$35,$D2,$FD,$18,$7A,$AB,$C9,$87,$47
	.dc.b	$94,$FD,$AA,$FE,$2F,$87,$CD,$6C,$7F,$16,$8C,$47,$09,$FA,$F8,$7C
	.dc.b	$D8,$FF,$7C,$C6,$54,$65,$7A,$AE,$AF,$BA,$DA,$F9,$0D,$9F,$8C,$70
	.dc.b	$9D,$92,$76,$D8,$DC,$F1,$AF,$87,$1C,$6B,$64,$77,$5E,$99,$F3,$57
	.dc.b	$1A,$FC,$2B,$6B,$E6,$47,$C0,$33,$5C,$5A,$66,$3E,$19,$C0,$F8,$6E
	.dc.b	$9A,$68,$F3,$1E,$47,$93,$8A,$F2,$CF,$BD,$B3,$AA,$F3,$57,$E4,$27
	.dc.b	$2F,$63,$94,$F8,$63,$E2,$CF,$02,$57,$59,$D1,$2B,$D5,$5D,$09,$19
	.dc.b	$8F,$03,$6B,$9B,$F3,$E7,$73,$EF,$BF,$CC,$BF,$06,$B6,$FC,$13,$D0
	.dc.b	$E2,$7C,$F7,$F3,$E7,$83,$E7,$6D,$F7,$3C,$51,$98,$F8,$27,$87,$0E
	.dc.b	$4E,$09,$7F,$2B,$0F,$2C,$62,$B7,$39,$AB,$12,$FE,$93,$DE,$7D,$11
	.dc.b	$FA,$3C,$3E,$0D,$F3,$8F,$33,$E8,$3E,$04,$6A,$7B,$4F,$42,$4F,$C4
	.dc.b	$AE,$C7,$85,$E3,$1F,$00,$EA,$F7,$32,$77,$AF,$97,$1D,$23,$FC,$2F
	.dc.b	$5F,$B1,$6D,$7F,$85,$A3,$11,$F9,$93,$33,$FE,$02,$7F,$32,$60,$FF
	.dc.b	$02,$C7,$E5,$31,$7A,$FE,$7C,$3F,$82,$46,$2F,$DB,$18,$9D,$47,$F8
	.dc.b	$D1,$B7,$FE,$32,$6B,$FE,$2D,$6D,$73,$8F,$8D,$3F,$E3,$69,$C9,$1E
	.dc.b	$E7,$52,$C7,$56,$BD,$5B,$3D,$6B,$ED,$B6,$3F,$44,$D9,$FD,$E3,$5F
	.dc.b	$C7,$36,$FE,$25,$F8,$E7,$CE,$7F,$41,$71,$7D,$9B,$7D,$9B,$3D,$9A
	.dc.b	$F6,$6D,$76,$9F,$4C,$63,$0F,$F0,$03,$63,$E3,$1A,$25,$79,$2F,$EA
	.dc.b	$46,$1F,$23,$4F,$21,$63,$C8,$D7,$C8,$DB,$F2,$47,$58,$FA,$4D,$3E
	.dc.b	$91,$4F,$A4,$5C,$3E,$56,$1E,$A9,$E8,$D9,$D6,$1F,$12,$BE,$91,$C2
	.dc.b	$BD,$F3,$F3,$DB,$87,$B2,$30,$47,$07,$C2,$71,$1F,$01,$A7,$A2,$37
	.dc.b	$1E,$C7,$81,$88,$D4,$64,$E5,$1D,$63,$FD,$13,$3F,$9B,$6D,$7F,$8E
	.dc.b	$89,$5F,$D1,$B8,$DE,$E7,$C1,$C3,$FD,$08,$D9,$FF,$41,$36,$FF,$C5
	.dc.b	$F5,$C2,$7E,$09,$A2,$C6,$A7,$82,$9E,$0F,$C0,$31,$1A,$8C,$D7,$C0
	.dc.b	$AE,$95,$B9,$CC,$7C,$18,$EC,$F8,$1E,$6D,$8D,$C6,$8F,$F4,$15,$77
	.dc.b	$3F,$D2,$24,$9E,$0D,$8D,$C6,$0F,$91,$19,$AF,$34,$F5,$7E,$4C,$E8
	.dc.b	$E1,$1C,$89,$1C,$1F,$39,$93,$FD,$83,$5F,$19,$B5,$FE,$C0,$8C,$47
	.dc.b	$C4,$9F,$F0,$97,$EA,$AD,$7C,$98,$7F,$8B,$A3,$0F,$F8,$13,$D0,$FF
	.dc.b	$6E,$DF,$F9,$69,$AF,$F9,$A9,$A9,$BE,$EB,$DD,$FF,$68,$B6,$BF,$9D
	.dc.b	$36,$BF,$D7,$AD,$7F,$DB,$7F,$30,$F9,$1F,$F6,$CB,$6B,$FA,$13,$67
	.dc.b	$FD,$8A,$D7,$FD,$C7,$83,$C2,$FF,$70,$B6,$BF,$A5,$36,$7F,$D9,$AD
	.dc.b	$7F,$DD,$78,$3D,$EF,$49,$2B,$D2,$F9,$3E,$46,$47,$91,$B5,$D3,$C1
	.dc.b	$F0,$92,$7C,$2F,$D5,$89,$1F,$D7,$78,$3F,$D2,$AD,$8F,$83,$EC,$C8
	.dc.b	$F8,$2D,$8F,$4F,$83,$EE,$24,$FB,$9F,$C1,$7A,$52,$3B,$36,$7E,$4F
	.dc.b	$B5,$F3,$89,$3F,$39,$FF,$36,$47,$FA,$6B,$DE,$FF,$7E,$BF,$EF,$46
	.dc.b	$BF,$E9,$E0,$E1,$E1,$E2,$F2,$7F,$D6,$30,$F2,$B1,$1C,$1B,$19,$BF
	.dc.b	$F0,$03,$67,$EF,$F8,$FF,$84,$24,$FE,$11,$E7,$EE,$31,$FC,$F5,$79
	.dc.b	$44,$67,$A3,$FC,$18,$B5,$69,$8B,$E3,$7E,$DF,$6D,$EE,$7F,$84,$E1
	.dc.b	$18,$7F,$C2,$CD,$7C,$0D,$7E,$35,$BE,$2B,$93,$63,$FC,$3C,$D7,$D0
	.dc.b	$D7,$F3,$BE,$EF,$87,$9C,$D2,$47,$E7,$9F,$F6,$8F,$6F,$FA,$96,$77
	.dc.b	$3F,$BD,$9C,$EF,$FE,$AE,$F3,$5E,$4E,$2B,$F7,$D5,$5F,$DF,$36,$F9
	.dc.b	$C7,$FA,$3B,$05,$F2,$8F,$E9,$1E,$EF,$C7,$E7,$34,$91,$FA,$17,$B4
	.dc.b	$6A,$38,$9F,$E9,$FA,$78,$0F,$9C,$F2,$7B,$8F,$5B,$D1,$4F,$79,$EF
	.dc.b	$61,$B8,$FF,$A8,$2F,$83,$6F,$99,$1D,$7B,$DF,$07,$D1,$7B,$17,$A6
	.dc.b	$57,$FA,$AA,$72,$D9,$E0,$47,$FD,$5F,$EF,$2F,$FD,$B3,$0F,$6D,$68
	.dc.b	$95,$BD,$71,$5A,$9D,$57,$81,$E4,$F1,$79,$E1,$67,$84,$2F,$A3,$C0
	.dc.b	$EF,$87,$92,$77,$7D,$5B,$7C,$E3,$B5,$F2,$9E,$AD,$8F,$11,$C4,$B3
	.dc.b	$C6,$FB,$E1,$EC,$9E,$78,$78,$5D,$CE,$4F,$7D,$71,$7D,$53,$B7,$99
	.dc.b	$EB,$3E,$09,$C4,$A6,$6F,$CC,$F0,$3F,$8D,$3F,$11,$B1,$F5,$9B,$7F
	.dc.b	$E9,$39,$7B,$AF,$7B,$0D,$AF,$E2,$13,$A9,$C9,$FC,$26,$1E,$F5,$23
	.dc.b	$FC,$35,$B9,$FD,$8C,$3F,$3D,$D3,$EB,$AF,$C3,$57,$B6,$73,$1F,$04
	.dc.b	$C9,$F0,$DC,$D7,$E0,$A3,$2F,$5D,$97,$D6,$7C,$A7,$18,$CB,$D4,$CC
	.dc.b	$F1,$9E,$17,$E3,$AD,$CF,$7A,$DE,$BF,$C5,$27,$2F,$F3,$5A,$E1,$7C
	.dc.b	$AB,$C0,$E1,$61,$E2,$6D,$FF,$BB,$9F,$5D,$F5,$AF,$D5,$56,$CF,$8A
	.dc.b	$5A,$E1,$5E,$43,$6F,$FB,$E3,$DE,$FD,$FD,$72,$3F,$58,$E2,$7C,$71
	.dc.b	$87,$86,$D2,$38,$47,$EE,$E7,$85,$FA,$5F,$EF,$F1,$C8,$D5,$7C,$E9
	.dc.b	$F7,$7B,$CF,$9B,$C6,$1F,$BD,$8C,$3F,$C9,$7C,$67,$CD,$EB,$1F,$E1
	.dc.b	$55,$A7,$B0,$70,$FF,$79,$6B,$D6,$FF,$BA,$C6,$23,$F8,$A7,$BE,$F2
	.dc.b	$FB,$42,$31,$8F,$F3,$7E,$7F,$EF,$19,$FD,$79,$CD,$78,$9B,$19,$BF
	.dc.b	$43,$FC,$E7,$3E,$B3,$AC,$3F,$8E,$3C,$AF,$FC,$20,$D7,$37,$FE,$0C
	.dc.b	$8C,$47,$F2,$BF,$79,$F0,$DF,$E1,$06,$DF,$F8,$16,$51,$87,$FC,$10
	.dc.b	$D7,$FC,$3C,$F8,$DE,$97,$F0,$A3,$86,$3F,$95,$75,$3D,$27,$C0,$4A
	.dc.b	$F0,$4F,$9A,$5E,$BB,$DB,$FF,$12,$B6,$BC,$CD,$9F,$F0,$E3,$5F,$F1
	.dc.b	$97,$96,$F6,$FE,$AB,$6B,$F2,$73,$DA,$F3,$52,$7E,$15,$6E,$BF,$C5
	.dc.b	$0D,$7A,$CF,$F9,$D3,$D4,$95,$FE,$74,$6C,$FF,$9D,$34,$FF,$3C,$92
	.dc.b	$7F,$CF,$2F,$F9,$F3,$CB,$91,$FE,$7C,$6C,$FF,$9F,$34,$FF,$40,$92
	.dc.b	$7F,$D0,$2F,$FA,$13,$29,$1F,$E8,$46,$CF,$FA,$13,$D9,$FF,$44,$92
	.dc.b	$7F,$D1,$2F,$FA,$33,$D2,$91,$FE,$8C,$6C,$FF,$A3,$3E,$1B,$FD,$22
	.dc.b	$49,$FF,$48,$BF,$E9,$4F,$8E,$91,$FE,$94,$6C,$FF,$A5,$3E,$07,$FD
	.dc.b	$32,$49,$FF,$4C,$BF,$E9,$CF,$32,$47,$FA,$71,$B3,$FE,$9C,$DF,$FD
	.dc.b	$42,$49,$FF,$50,$BF,$EA,$4F,$37,$F6,$3E,$CF,$F8,$E2,$7F,$C4,$DE
	.dc.b	$53,$E0,$F3,$52,$7F,$D5,$2F,$16,$93,$0F,$E0,$BE,$7B,$5E,$11,$86
	.dc.b	$F4,$7F,$92,$DB,$5F,$EC,$56,$CF,$FB,$15,$AF,$FB,$15,$B5,$D6,$7F
	.dc.b	$27,$97,$A6,$BB,$46,$30,$FE,$1B,$5B,$9F,$E3,$35,$BC,$3C,$57,$B3
	.dc.b	$FD,$B1,$87,$63,$89,$1D,$57,$62,$4F,$27,$AE,$1B,$73,$3E,$D3,$8C
	.dc.b	$EC,$8E,$63,$8B,$E2,$23,$C8,$FF,$1E,$5F,$A6,$38,$4F,$29,$F9,$11
	.dc.b	$D8,$C9,$1C,$9D,$8E,$87,$CD,$7F,$DB,$B7,$FE,$E0,$23,$83,$F8,$0F
	.dc.b	$BE,$F8,$6F,$75,$FF,$6B,$8C,$4F,$A2,$31,$3A,$8F,$F2,$64,$EA,$FC
	.dc.b	$2D,$AF,$54,$23,$13,$E9,$8C,$57,$37,$FC,$BD,$3A,$7D,$B5,$E0,$AF
	.dc.b	$11,$E0,$AE,$13,$CA,$32,$F2,$8C,$D7,$38,$F8,$87,$F6,$07,$04,$8D
	.dc.b	$9F,$D6,$9C,$4F,$49,$EC,$71,$9D,$1C,$23,$E5,$B8,$78,$4F,$29,$E1
	.dc.b	$18,$9D,$46,$0F,$92,$E2,$BE,$04,$62,$7A,$91,$C1,$C6,$36,$7D,$0A
	.dc.b	$F0,$38,$9F,$A1,$18,$7E,$51,$E3,$3A,$46,$E7,$E8,$C6,$9E,$CD,$8E
	.dc.b	$D5,$E7,$8C,$1D,$9C,$4F,$D3,$8E,$CF,$23,$07,$AA,$7C,$0F,$CE,$7A
	.dc.b	$C7,$90,$EB,$1E,$43,$8C,$70,$AF,$BD,$97,$F0,$58,$C5,$F8,$AB,$F0
	.dc.b	$31,$8B,$E9,$1C,$CF,$03,$D9,$A6,$6F,$E4,$3E,$08,$F7,$1C,$9D,$9D
	.dc.b	$9E,$51,$C2,$30,$71,$21,$C2,$FD,$13,$C2,$FD,$11,$89,$DD,$73,$8C
	.dc.b	$47,$82,$31,$87,$BD,$FD,$A4,$7F,$B0,$4F,$CC,$95,$CB,$F7,$4C,$57
	.dc.b	$F9,$AD,$C1,$0F,$AE,$6A,$FF,$C1,$3E,$C3,$D9,$F9,$CD,$8F,$88,$D9
	.dc.b	$FC,$00,$71,$97,$E0,$1B,$3F,$45,$B3,$F1,$9A,$FC,$E6,$E1,$FB,$D6
	.dc.b	$E1,$FE,$52,$6D,$79,$B0,$F2,$91,$F2,$B6,$FC,$B3,$C1,$FF,$32,$B7
	.dc.b	$0F,$F4,$73,$6B,$FD,$D8,$D9,$FF,$7B,$35,$F7,$57,$5A,$F7,$36,$BD
	.dc.b	$CD,$9F,$73,$5F,$73,$70,$F7,$37,$0F,$73,$6B,$DC,$7B,$09,$5E,$C8
	.dc.b	$D3,$EC,$6D,$FB,$1B,$3E,$C6,$CF,$B1,$AF,$B2,$3A,$CF,$B1,$A7,$B0
	.dc.b	$B1,$EC,$6B,$EC,$6D,$FB,$1F,$4C,$79,$5D,$E1,$F8,$03,$E6,$16,$FE
	.dc.b	$0C,$69,$B3,$EF,$8E,$77,$ED,$74,$60,$F3,$47,$72,$47,$AD,$C1,$FA
	.dc.b	$9A,$FC,$D3,$6B,$F4,$CF,$C5,$AF,$F4,$44,$F9,$FE,$53,$C7,$7F,$88
	.dc.b	$6C,$FE,$21,$B5,$F8,$86,$BF,$A5,$6D,$1F,$CE,$12,$7D,$53,$C5,$FD
	.dc.b	$9C,$66,$3D,$71,$DC,$C9,$1C,$9C,$4F,$CF,$C7,$B1,$F1,$47,$B2,$31
	.dc.b	$3E,$38,$C1,$FC,$0B,$BF,$C1,$7F,$A6,$6C,$7E,$99,$B7,$FA,$66,$BF
	.dc.b	$C0,$DB,$5F,$EC,$69,$E0,$75,$3D,$15,$D5,$C5,$7E,$DA,$B8,$9F,$C0
	.dc.b	$49,$1C,$0E,$71,$B7,$EB,$C7,$B0,$8E,$0F,$F5,$2F,$2B,$E7,$B9,$D7
	.dc.b	$FB,$A2,$31,$3F,$EA,$AF,$D2,$9F,$2F,$F2,$5B,$1F,$EC,$B6,$D7,$F3
	.dc.b	$30,$E1,$FE,$6D,$FA,$5A,$7E,$93,$5F,$F6,$A3,$63,$93,$5E,$3F,$05
	.dc.b	$C0,$93,$C1,$E7,$F0,$12,$39,$B6,$79,$FB,$7E,$02,$4F,$81,$F1,$FB
	.dc.b	$D2,$3C,$6D,$9F,$1F,$C0,$7A,$09,$3E,$87,$D7,$F2,$32,$3D,$6D,$9F
	.dc.b	$5F,$C6,$FC,$02,$4F,$C0,$7E,$1F,$8D,$23,$E1,$B6,$7E,$1F,$8D,$F2
	.dc.b	$09,$3F,$21,$F9,$7E,$34,$8F,$96,$D9,$FA,$9F,$20,$F9,$1F,$D1,$36
	.dc.b	$BE,$C4,$A3,$0F,$09,$FD,$47,$C7,$3E,$3E,$92,$F8,$C6,$5E,$7F,$0C
	.dc.b	$DD,$E6,$61,$B8,$FE,$1E,$5C,$2F,$7B,$57,$F8,$4F,$37,$CB,$6E,$64
	.dc.b	$61,$B1,$B3,$D7,$F7,$2F,$37,$8B,$63,$31,$83,$AF,$37,$E8,$E3,$E1
	.dc.b	$B5,$FC,$E3,$5F,$F0,$83,$51,$29,$D7,$88,$77,$7E,$85,$BE,$4B,$EA
	.dc.b	$B5,$39,$11,$99,$EB,$1C,$5F,$85,$5C,$38,$7B,$61,$E0,$9F,$F7,$2E
	.dc.b	$19,$7B,$B5,$E2,$74,$9E,$35,$F2,$C9,$1F,$B4,$7E,$88,$FA,$0E,$98
	.dc.b	$79,$BC,$0B,$97,$C7,$8F,$35,$6E,$FF,$87,$DE,$AE,$46,$0F,$11,$F2
	.dc.b	$BE,$09,$ED,$F6,$F5,$9F,$07,$4F,$54,$EA,$B8,$87,$0F,$D0,$D3,$E9
	.dc.b	$46,$33,$F9,$4D,$8C,$FC,$1F,$C5,$F7,$9E,$ED,$E7,$98,$F6,$9C,$B6
	.dc.b	$78,$11,$F1,$FC,$12,$FF,$03,$0F,$81,$5F,$E1,$6B,$59,$FE,$2A,$73
	.dc.b	$5F,$8C,$AE,$EF,$17,$F1,$86,$23,$70,$AE,$6E,$CF,$CA,$B8,$8D,$CF
	.dc.b	$13,$9E,$1F,$97,$8E,$73,$B9,$F3,$61,$EA,$9E,$35,$B9,$E3,$7E,$97
	.dc.b	$78,$58,$F0,$DF,$19,$DC,$79,$4F,$87,$5C,$DD,$D7,$17,$D9,$5C,$5F
	.dc.b	$89,$3B,$78,$9B,$C3,$D0,$F2,$3F,$AA,$3F,$11,$FF,$44,$35,$F3,$9A
	.dc.b	$F8,$37,$C1,$FF,$24,$94,$62,$75,$39,$8C,$47,$F2,$AF,$83,$FE,$2E
	.dc.b	$9A,$47,$F8,$DB,$ED,$C7,$F3,$FA,$FB,$E7,$F8,$5C,$E7,$87,$F8,$46
	.dc.b	$BF,$7A,$72,$7C,$9B,$E2,$FC,$B3,$31,$F8,$48,$CB,$C6,$3C,$87,$19
	.dc.b	$EB,$AF,$F7,$9B,$F0,$61,$E3,$AD,$DF,$3B,$DC,$7F,$AC,$6B,$6F,$84
	.dc.b	$DD,$FF,$1C,$9D,$C7,$F2,$39,$DD,$73,$BF,$DC,$E1,$EA,$3F,$CA,$33
	.dc.b	$D1,$FE,$40,$66,$3A,$1F,$C0,$2B,$DF,$86,$1F,$E0,$07,$F3,$4F,$83
	.dc.b	$FD,$DC,$FF,$06,$6C,$68,$3F,$CE,$A3,$5A,$FF,$33,$7F,$9D,$47,$0B
	.dc.b	$FE,$64,$D8,$CD,$6C,$F0,$55,$5F,$02,$F8,$2F,$F2,$93,$6B,$FB,$8B
	.dc.b	$67,$FC,$54,$D7,$FD,$BC,$E7,$0F,$E8,$DE,$87,$F7,$67,$FA,$24,$66
	.dc.b	$BF,$A3,$CE,$5F,$EA,$5F,$05,$FD,$A1,$FE,$4E,$DB,$F8,$F6,$E6,$FF
	.dc.b	$2B,$47,$F5,$C9,$D4,$FF,$51,$78,$47,$FB,$C5,$AF,$6F,$03,$FC,$DD
	.dc.b	$18,$8E,$51,$F1,$65,$F0,$EF,$3B,$E3,$5F,$6C,$7F,$AF,$CE,$5F,$E7
	.dc.b	$BF,$07,$9C,$7F,$B7,$B5,$2F,$37,$E8,$39,$3C,$51,$AC,$BF,$D0,$51
	.dc.b	$82,$3F,$5F,$E1,$BF,$DE,$71,$D5,$FF,$7A,$F8,$CF,$7F,$EE,$37,$1F
	.dc.b	$EA,$3F,$08,$F8,$4C,$D4,$9F,$F1,$6D,$6F,$2F,$F3,$33,$5F,$F0,$31
	.dc.b	$FE,$26,$9F,$F4,$B7,$B6,$F8,$EA,$4B,$FF,$1B,$C6,$5F,$F1,$CF,$C0
	.dc.b	$7C,$16,$ED,$FE,$9F,$8C,$61,$FE,$75,$6B,$E0,$5E,$A6,$C9,$5D,$1F
	.dc.b	$F3,$BF,$C2,$48,$FF,$3B,$B6,$BA,$69,$DC,$93,$E4,$7C,$5E,$6C,$8F
	.dc.b	$3B,$67,$FC,$FF,$F0,$9F,$E8,$42,$4F,$FA,$11,$F5,$7C,$0C,$8F,$7B
	.dc.b	$67,$D9,$F0,$BE,$F2,$4F,$C4,$7F,$D1,$FB,$C8,$FF,$47,$B6,$7E,$25
	.dc.b	$FC,$72,$4F,$CC,$7E,$54,$C8,$F9,$ED,$9F,$F4,$BF,$C2,$7F,$A7,$09
	.dc.b	$3F,$E9,$C7,$FD,$A1,$F0,$8C,$7F,$82,$36,$7F,$C3,$3B,$2B,$35,$27
	.dc.b	$FC,$EB,$5B,$BF,$F5,$73,$5F,$C4,$B5,$FF,$28,$3B,$EA,$FF,$24,$70
	.dc.b	$BE,$13,$A5,$7F,$21,$8C,$5F,$F8,$AA,$F8,$3F,$E5,$98,$CC,$FF,$27
	.dc.b	$9C,$BF,$E6,$08,$CD,$FF,$95,$67,$32,$9C,$D7,$93,$0F,$E4,$91,$89
	.dc.b	$FF,$34,$D7,$08,$C4,$23,$18,$7A,$30,$F5,$B7,$0F,$33,$6B,$C8,$D7
	.dc.b	$C4,$DB,$EB,$7D,$2B,$FC,$93,$1E,$03,$38,$79,$AA,$4E,$9E,$46,$2F
	.dc.b	$E5,$38,$8D,$12,$3E,$44,$EF,$0F,$4C,$E5,$F9,$11,$98,$F5,$99,$3E
	.dc.b	$41,$67,$DA,$FF,$02,$AF,$E0,$E4,$70,$7F,$99,$3E,$09,$F0,$9D,$6B
	.dc.b	$FD,$2F,$18,$9E,$90,$8C,$3E,$C8,$E5,$3C,$63,$4D,$9D,$E1,$EC,$9F
	.dc.b	$ED,$91,$8A,$F5,$C6,$2B,$81,$23,$C3,$3B,$9F,$69,$B8,$F0,$91,$F2
	.dc.b	$19,$25,$7F,$A9,$67,$94,$7E,$02,$78,$4E,$6B,$FB,$A5,$6B,$1F,$C7
	.dc.b	$BC,$09,$1E,$5B,$EB,$7F,$EB,$B9,$E1,$87,$2A,$93,$D8,$F7,$E1,$DD
	.dc.b	$C5,$FD,$09,$E9,$39,$8E,$A7,$C0,$70,$49,$F0,$4F,$06,$9D,$27,$DC
	.dc.b	$72,$3E,$AB,$9C,$3D,$B3,$E6,$9D,$46,$27,$96,$1E,$C6,$D6,$EF,$EE
	.dc.b	$1E,$3B,$F5,$4F,$5C,$34,$F9,$23,$AC,$61,$D9,$88,$EA,$71,$2B,$93
	.dc.b	$D5,$7F,$4A,$3F,$0B,$7C,$E7,$DA,$7E,$0A,$FC,$8E,$1B,$87,$B6,$39
	.dc.b	$CF,$C9,$7C,$73,$AC,$3A,$61,$DE,$7A,$61,$EB,$AE,$B3,$F1,$67,$9C
	.dc.b	$FA,$5A,$7A,$67,$77,$E7,$C3,$C5,$18,$AF,$59,$F9,$BA,$FF,$00,$DF
	.dc.b	$EC,$1B,$7F,$EF,$F7,$E6,$18,$9F,$F0,$3B,$F3,$0C,$18,$9E,$4F,$A6
	.dc.b	$3F,$31,$F0,$9E,$E2,$3F,$80,$6C,$7D,$06,$DF,$C0,$6C,$F3,$9F,$E1
	.dc.b	$31,$87,$E5,$C7,$63,$A1,$1E,$8D,$9E,$8D,$3A,$47,$F1,$19,$E8,$FD
	.dc.b	$B8,$E8,$F2,$24,$F2,$6C,$72,$A5,$72,$7F,$78,$DB,$FE,$02,$D9,$FE
	.dc.b	$4C,$DA,$FE,$32,$DB,$FE,$18,$D9,$F1,$C7,$84,$8F,$86,$35,$3F,$1A
	.dc.b	$3D,$CF,$C6,$30,$7A,$0C,$1F,$05,$DB,$F0,$4D,$CF,$67,$77,$EE,$9D
	.dc.b	$B6,$3D,$31,$8A,$F7,$B4,$F7,$96,$3F,$11,$1F,$8F,$7C,$A7,$E5,$1C
	.dc.b	$3E,$53,$F9,$EF,$81,$D1,$B3,$D1,$A7,$88,$B7,$F2,$A3,$FA,$A4,$F1
	.dc.b	$8F,$E7,$0F,$F3,$42,$47,$38,$CB,$C0,$93,$D6,$32,$F6,$8F,$E6,$D3
	.dc.b	$D5,$A7,$08,$FD,$7C,$7B,$2C,$D2,$4A,$7F,$D8,$DE,$93,$D1,$F2,$C6
	.dc.b	$1B,$82,$B8,$4A,$B8,$47,$A2,$75,$3D,$67,$2F,$A4,$C5,$74,$AE,$73
	.dc.b	$AA,$E3,$5D,$6B,$FD,$9D,$39,$AF,$7C,$F9,$5F,$55,$70,$9D,$4F,$69
	.dc.b	$F2,$1D,$AB,$99,$1C,$C6,$CE,$87,$F1,$29,$F4,$BF,$E9,$A6,$BC,$E3
	.dc.b	$F8,$BB,$E7,$7F,$CB,$FE,$B7,$58,$FF,$79,$70,$A3,$18,$7B,$A3,$15
	.dc.b	$EA,$3A,$57,$0A,$DC,$7D,$09,$DC,$F9,$AB,$95,$7F,$BC,$E7,$31,$8A
	.dc.b	$E9,$5E,$D8,$C3,$D6,$3C,$E6,$9F,$51,$92,$47,$03,$E7,$61,$DE,$E5
	.dc.b	$7F,$AD,$9F,$F5,$A1,$1C,$1D,$7F,$4C,$79,$3C,$30,$61,$F8,$09,$CD
	.dc.b	$78,$FD,$7F,$C8,$3F,$39,$B5,$FD,$19,$AF,$FB,$39,$AF,$E5,$97,$C1
	.dc.b	$FE,$F9,$B3,$CE,$30,$7F,$02,$D3,$FD,$07,$5F,$7F,$8B,$04,$AF,$17
	.dc.b	$B5,$CC,$8E,$0F,$37,$B9,$DC,$8E,$0F,$C3,$DB,$7F,$01,$23,$2F,$9F
	.dc.b	$C8,$91,$F0,$5B,$8F,$C1,$F8,$2F,$61,$27,$E3,$3E,$BF,$A2,$91,$E0
	.dc.b	$6C,$FC,$9F,$0B,$E2,$12,$7E,$73,$F1,$FC,$29,$1F,$45,$B3,$FA,$5F
	.dc.b	$9A,$F9,$04,$9F,$6B,$FC,$B3,$E0,$1F,$03,$49,$3F,$A8,$8C,$BF,$AA
	.dc.b	$F8,$16,$EF,$23,$0D,$C7,$FC,$0B,$2E,$17,$B7,$AB,$FD,$9F,$C6,$FF
	.dc.b	$C0,$0B,$BA,$FE,$D9,$46,$5F,$F0,$7B,$5F,$E3,$A3,$E0,$9E,$47,$16
	.dc.b	$C7,$F0,$06,$BF,$E1,$86,$BC,$BB,$3F,$96,$3F,$C1,$9A,$FF,$87,$9A
	.dc.b	$EF,$B7,$CD,$0E,$1A,$49,$E4,$FA,$FA,$BF,$99,$47,$EF,$BB,$1F,$02
	.dc.b	$F8,$2F,$E6,$F1,$FC,$D0,$CB,$8E,$3F,$D3,$75,$DB,$FE,$89,$87,$47
	.dc.b	$F5,$D1,$DC,$8F,$EC,$DE,$0A,$38,$5F,$F3,$A6,$D7,$F3,$6C,$1E,$B7
	.dc.b	$0C,$1F,$05,$E1,$C3,$D3,$AD,$BF,$8F,$1F,$CF,$67,$81,$F6,$EB,$8E
	.dc.b	$5F,$AC,$8C,$1F,$7F,$E0,$5E,$EF,$5F,$61,$F0,$54,$F4,$CE,$AF,$F9
	.dc.b	$13,$5F,$E2,$BA,$FF,$60,$8C,$65,$FD,$59,$B1,$9F,$82,$F8,$7E,$33
	.dc.b	$DC,$BD,$53,$1F,$E5,$D3,$96,$CF,$02,$3F,$CC,$FE,$09,$7F,$C9,$A1
	.dc.b	$F0,$1F,$CA,$9F,$EF,$53,$AC,$FF,$BE,$4E,$67,$F9,$7D,$FF,$78,$8C
	.dc.b	$61,$E1,$9E,$6F,$F7,$CB,$E5,$18,$3A,$E1,$E1,$AE,$37,$CE,$FF,$C0
	.dc.b	$67,$38,$F0,$5F,$26,$D7,$19,$DD,$71,$BF,$F0,$94,$71,$3C,$53,$B8
	.dc.b	$F0,$E1,$EC,$9E,$31,$4E,$35,$CD,$61,$E1,$9E,$75,$B9,$F8,$65,$8F
	.dc.b	$03,$D8,$F8,$07,$2B,$FA,$0F,$F0,$4A,$F8,$AD,$8C,$BD,$8F,$F4,$CC
	.dc.b	$3E,$05,$F0,$5F,$E0,$68,$C3,$63,$33,$C1,$AF,$F8,$33,$E0,$BF,$8C
	.dc.b	$26,$91,$FE,$38,$FB,$71,$FC,$FE,$BE,$F9,$FF,$1F,$4E,$78,$7F,$C6
	.dc.b	$BA,$FE,$16,$72,$7C,$0B,$E2,$FC,$23,$31,$F8,$B8,$CB,$C6,$3C,$87
	.dc.b	$19,$EB,$A7,$FA,$EA,$FC,$18,$78,$EB,$77,$CE,$F7,$1F,$CC,$2B,$6F
	.dc.b	$84,$DD,$FF,$1A,$9D,$C7,$F9,$4A,$77,$5C,$EF,$FC,$35,$87,$A8,$FF
	.dc.b	$40,$4F,$47,$FC,$9A,$66,$3A,$1F,$E2,$1A,$F7,$E1,$87,$F8,$09,$FE
	.dc.b	$67,$F8,$2F,$DE,$4F,$F8,$9D,$B1,$A0,$FF,$9B,$63,$5B,$7F,$31,$7F
	.dc.b	$CD,$31,$C2,$FF,$CB,$ED,$8C,$D6,$CF,$05,$55,$F0,$5F,$CA,$A7,$FA
	.dc.b	$07,$9E,$F8,$3D,$4E,$6B,$D3,$8F,$FA,$0A,$73,$D5,$D6,$31,$87,$F9
	.dc.b	$06,$F8,$47,$F4,$7C,$3A,$4F,$F9,$17,$E0,$DF,$05,$FE,$A8,$6C,$7F
	.dc.b	$98,$9B,$97,$FB,$E9,$AF,$F6,$C7,$37,$FD,$4F,$E0,$DE,$7F,$B6,$73
	.dc.b	$7F,$E7,$C9,$CB,$E1,$C3,$D1,$18,$C3,$E8,$57,$07,$FC,$97,$19,$C3
	.dc.b	$FD,$17,$F0,$7E,$29,$FF,$4C,$B5,$FF,$51,$35,$34,$E8,$BE,$0D,$EC
	.dc.b	$7D,$39,$CC,$63,$2F,$F3,$53,$6B,$FC,$16,$71,$AF,$F6,$7C,$66,$3F
	.dc.b	$C1,$ED,$7F,$D5,$BE,$53,$DD,$FF,$6A,$B6,$FF,$D1,$ED,$9E,$AD,$7F
	.dc.b	$C3,$3E,$D3,$E0,$BF,$DB,$AD,$AF,$F4,$BB,$67,$CA,$D7,$FC,$43,$F0
	.dc.b	$7F,$E2,$09,$FF,$4E,$B6,$3D,$6D,$77,$E6,$7F,$BA,$BD,$77,$9F,$49
	.dc.b	$7F,$EA,$68,$CB,$FE,$7B,$F8,$3F,$F3,$A1,$1C,$1F,$E6,$6F,$90,$FF
	.dc.b	$3A,$11,$C1,$CE,$FC,$04,$70,$7F,$A1,$FE,$0B,$FD,$00,$47,$07,$F9
	.dc.b	$CB,$E3,$FF,$D0,$04,$70,$79,$FE,$13,$D8,$47,$07,$FA,$4F,$83,$FD
	.dc.b	$18,$47,$07,$FB,$13,$DF,$FE,$05,$AF,$E0,$9A,$FF,$93,$9A,$FF,$BB
	.dc.b	$1A,$FF,$BD,$74,$FB,$8F,$FB,$06,$BE,$71,$2B,$FD,$49,$F0,$9F,$E9
	.dc.b	$D2,$38,$3C,$5E,$4F,$FB,$B2,$BB,$35,$0C,$3F,$CD,$73,$C0,$FC,$3F
	.dc.b	$BF,$FC,$47,$88,$FF,$13,$F8,$47,$C2,$66,$25,$70,$26,$1F,$8E,$8F
	.dc.b	$C9,$9F,$C6,$FE,$0F,$F4,$DD,$7F,$F0,$03,$1E,$B7,$FE,$D3,$9C,$BE
	.dc.b	$6B,$4E,$9F,$E6,$13,$A7,$D7,$19,$A5,$71,$AD,$57,$2B,$F1,$75,$7A
	.dc.b	$23,$9B,$6F,$8C,$7C,$6A,$E3,$87,$5C,$3A,$5E,$DA,$7F,$BA,$E3,$55
	.dc.b	$C6,$74,$D9,$D6,$1E,$27,$E0,$4D,$9F,$E9,$95,$C9,$5F,$A8,$93,$D2
	.dc.b	$32,$F1,$C3,$D3,$18,$BE,$E7,$62,$3F,$01,$CB,$49,$1C,$AB,$E0,$E1
	.dc.b	$D6,$BD,$F5,$A7,$EB,$61,$E7,$AF,$2C,$76,$BF,$BD,$7C,$6F,$E1,$1F
	.dc.b	$C9,$AB,$F9,$13,$83,$FD,$66,$E1,$F5,$1F,$80,$F8,$4E,$4F,$C4,$6C
	.dc.b	$1A,$4B,$F1,$1F,$83,$7F,$24,$47,$33,$F7,$E3,$0F,$78,$CB,$DD,$FC
	.dc.b	$B4,$6A,$FE,$36,$36,$7B,$1F,$D0,$AB,$81,$EF,$AF,$D9,$38,$3D,$EF
	.dc.b	$E9,$7C,$1E,$51,$F9,$16,$BF,$82,$6E,$06,$92,$BF,$54,$F3,$7C,$E6
	.dc.b	$09,$3F,$AA,$8E,$0D,$8C,$E1,$F8,$1B,$E5,$39,$24,$66,$76,$7E,$BE
	.dc.b	$B7,$19,$9D,$CE,$4D,$BF,$3D,$AF,$B5,$B7,$F6,$DB,$5D,$1F,$9B,$3D
	.dc.b	$0F,$8C,$F0,$BD,$46,$0D,$1F,$4A,$FF,$39,$5A,$8F,$01,$F7,$CD,$DF
	.dc.b	$CD,$BE,$87,$88,$B1,$A9,$E5,$3E,$78,$C4,$EA,$BB,$BC,$63,$B4,$61
	.dc.b	$D9,$88,$F8,$CE,$23,$E5,$1B,$24,$69,$CB,$49,$1F,$06,$32,$D8,$CB
	.dc.b	$F0,$23,$2D,$8C,$CF,$C1,$8F,$21,$C4,$86,$CF,$B4,$79,$E3,$8D,$FC
	.dc.b	$4B,$F8,$AD,$8F,$88,$76,$8F,$D1,$D7,$7A,$FE,$03,$18,$24,$FE,$AE
	.dc.b	$75,$3F,$18,$DC,$66,$7F,$CD,$1E,$9B,$D5,$FB,$CD,$9F,$C7,$B6,$BF
	.dc.b	$0A,$DC,$7E,$6C,$7F,$66,$8C,$4F,$C8,$6C,$7C,$26,$BF,$0C,$38,$C3
	.dc.b	$E5,$57,$F3,$48,$C6,$1E,$6C,$3C,$A4,$C3,$CA,$DC,$3C,$AD,$BF,$2C
	.dc.b	$76,$8C,$4F,$95,$B1,$E5,$6B,$E5,$BF,$0C,$FF,$80,$AB,$C9,$5A,$C3
	.dc.b	$CF,$3F,$DC,$6B,$77,$FE,$9F,$6D,$F0,$9F,$39,$27,$CE,$D8,$F3,$CF
	.dc.b	$08,$E9,$7D,$F0,$F8,$51,$B3,$D4,$5B,$F5,$35,$F5,$4F,$18,$F5,$1D
	.dc.b	$0D,$61,$EA,$9F,$B5,$3B,$7D,$4D,$C3,$D5,$1E,$82,$38,$25,$79,$23
	.dc.b	$8B,$E0,$7A,$BE,$5C,$3F,$06,$7B,$E3,$27,$03,$FD,$0A,$FE,$45,$FF
	.dc.b	$0F,$1F,$6E,$3F,$A4,$46,$63,$D8,$7F,$B3,$F8,$DE,$CF,$C8,$8C,$36
	.dc.b	$FF,$C0,$33,$AB,$F3,$57,$53,$6E,$2F,$F3,$33,$B9,$F3,$57,$2B,$FF
	.dc.b	$69,$4E,$63,$17,$C7,$0C,$D4,$8E,$F5,$B7,$D3,$5C,$2F,$C2,$47,$E6
	.dc.b	$BF,$E6,$1A,$E4,$FF,$AA,$AB,$C8,$FC,$66,$BF,$E4,$56,$BC,$6B,$DF
	.dc.b	$C2,$F2,$B9,$E1,$FE,$24,$8C,$4F,$9B,$0F,$F7,$74,$62,$FC,$93,$FE
	.dc.b	$75,$BE,$2D,$38,$1D,$9E,$24,$73,$0B,$F0,$4E,$63,$13,$CD,$EE,$7F
	.dc.b	$9D,$2B,$C5,$5C,$27,$B1,$D1,$C4,$F2,$38,$57,$F8,$32,$BD,$51,$FE
	.dc.b	$8F,$8C,$12,$BE,$2F,$CE,$78,$EF,$E2,$36,$3F,$D2,$8D,$AF,$23,$5F
	.dc.b	$F4,$43,$6C,$FC,$A2,$4F,$A2,$7C,$C7,$07,$C2,$47,$33,$F0,$23,$C2
	.dc.b	$66,$BC,$AF,$36,$CF,$0A,$F8,$04,$9F,$F5,$93,$63,$E3,$4B,$D7,$79
	.dc.b	$FE,$69,$CC,$62,$FF,$4F,$EC,$3E,$17,$FC,$6E,$D8,$FF,$1E,$36,$BF
	.dc.b	$DF,$ED,$7F,$77,$BF,$FA,$5D,$BC,$9F,$E4,$08,$C4,$7F,$2D,$8F,$B1
	.dc.b	$C5,$2F,$FD,$94,$D9,$EB,$F0,$1E,$22,$4F,$89,$E9,$C3,$23,$F8,$4B
	.dc.b	$67,$FD,$A5,$F3,$FF,$85,$24,$FF,$B5,$DF,$6F,$C3,$48,$F6,$B6,$7F
	.dc.b	$8B,$FC,$37,$A4,$93,$FC,$59,$FC,$87,$CF,$C8,$FF,$71,$36,$7E,$2F
	.dc.b	$8D,$F2,$89,$3F,$29,$F8,$9F,$01,$23,$F9,$0B,$67,$FD,$D5,$F0,$DF
	.dc.b	$9D,$24,$FF,$BB,$DF,$F3,$5F,$C2,$B5,$79,$9F,$E7,$18,$C3,$7E,$15
	.dc.b	$E1,$E6,$B7,$8F,$F3,$21,$F0,$AD,$3F,$CF,$51,$86,$C6,$63,$FC,$FC
	.dc.b	$D7,$95,$BE,$4B,$8B,$63,$FD,$08,$D7,$71,$DB,$BB,$FC,$31,$1F,$E8
	.dc.b	$96,$BE,$56,$BF,$D2,$1A,$FF,$B9,$D6,$DC,$C9,$1C,$DF,$5F,$CD,$FF
	.dc.b	$A2,$63,$FD,$2C,$D7,$E1,$35,$FE,$AA,$D7,$FD,$E9,$D1,$FE,$1E,$7F
	.dc.b	$47,$D1,$F1,$A3,$11,$FA,$48,$C1,$FE,$FB,$EC,$FF,$6A,$CE,$9B,$19
	.dc.b	$E2,$FF,$0A,$75,$BD,$8C,$60,$18,$71,$AF,$1D,$EA,$72,$F0,$C3,$F8
	.dc.b	$05,$E6,$1F,$0B,$F7,$1B,$5F,$DC,$23,$04,$7F,$B3,$3C,$4E,$98,$BE
	.dc.b	$7B,$C1,$F0,$9F,$C3,$2B,$FC,$09,$8B,$E1,$9F,$0D,$E3,$6C,$FF,$74
	.dc.b	$AE,$0F,$F2,$66,$D7,$FB,$E9,$BA,$7C,$19,$46,$1F,$88,$BE,$1B,$E1
	.dc.b	$78,$CF,$80,$BD,$32,$FF,$C0,$73,$96,$CF,$02,$3F,$E0,$8F,$86,$E9
	.dc.b	$97,$D1,$F8,$4F,$ED,$93,$F4,$E7,$2A,$32,$FF,$71,$AF,$44,$62,$FC
	.dc.b	$13,$B7,$D9,$5C,$A7,$8B,$63,$E0,$1C,$23,$95,$74,$8E,$38,$FF,$6F
	.dc.b	$BE,$87,$95,$C5,$70,$8F,$54,$F4,$AE,$51,$BC,$32,$D3,$99,$EB,$32
	.dc.b	$53,$38,$79,$DE,$17,$DD,$E3,$7E,$93,$CF,$5C,$AB,$E4,$CF,$F9,$09
	.dc.b	$AE,$8F,$E2,$8E,$1F,$A1,$F0,$AF,$86,$FE,$31,$09,$4E,$63,$8B,$5E
	.dc.b	$CD,$7F,$90,$7C,$37,$F1,$E4,$D2,$3F,$C8,$9F,$83,$7F,$D9,$E3,$E2
	.dc.b	$9F,$B5,$8C,$F9,$4D,$FE,$3C,$E4,$F9,$75,$C7,$3F,$C3,$C6,$5F,$C4
	.dc.b	$C6,$73,$F2,$CF,$0B,$E8,$F8,$0F,$39,$9B,$EE,$E2,$F9,$CF,$07,$BC
	.dc.b	$66,$BF,$8A,$4E,$5E,$F3,$C2,$B9,$E1,$EC,$65,$FF,$87,$30,$F8,$47
	.dc.b	$91,$FD,$B4,$EA,$30,$6A,$7B,$9F,$E7,$49,$F8,$17,$87,$FC,$4F,$F0
	.dc.b	$DF,$B9,$9F,$F3,$CB,$63,$27,$C2,$AF,$04,$65,$BA,$E6,$BF,$84,$4E
	.dc.b	$5B,$59,$3F,$96,$46,$5F,$E5,$DF,$0B,$FC,$2E,$7F,$D1,$2D,$8F,$F1
	.dc.b	$5F,$94,$F2,$F9,$42,$75,$8F,$C3,$78,$47,$FA,$79,$AF,$F1,$B6,$BE
	.dc.b	$0E,$1F,$94,$FF,$1F,$F8,$3F,$F7,$5E,$3F,$E9,$A6,$C7,$F8,$F3,$95
	.dc.b	$E5,$F8,$A7,$38,$FD,$36,$DF,$3E,$BF,$81,$3F,$CD,$6B,$84,$49,$FD
	.dc.b	$5D,$76,$C3,$FD,$6A,$D7,$94,$7E,$C3,$E0,$9F,$17,$E8,$9F,$EA,$D1
	.dc.b	$89,$D4,$E6,$53,$97,$FC,$20,$F9,$9E,$37,$F3,$23,$58,$FE,$35,$D4
	.dc.b	$F2,$9E,$91,$E5,$8E,$37,$E5,$9F,$07,$9C,$F8,$DA,$49,$FE,$13,$19
	.dc.b	$6F,$C6,$70,$23,$83,$7E,$4F,$42,$38,$3F,$89,$FC,$1F,$8F,$13,$49
	.dc.b	$8F,$C3,$FD,$37,$F8,$35,$7F,$1A,$6D,$7F,$8C,$DA,$FF,$B9,$9A,$FF
	.dc.b	$63,$FB,$A7,$C0,$7C,$D6,$D7,$F2,$46,$CF,$F8,$F5,$AF,$3F,$57,$C4
	.dc.b	$49,$F1,$3E,$5F,$90,$91,$E5,$6C,$F9,$7C,$EF,$51,$27,$D4,$FB,$7D
	.dc.b	$09,$1E,$D6,$CF,$B7,$CE,$F8,$44,$9F,$84,$FC,$5F,$8B,$91,$F1,$5B
	.dc.b	$3F,$17,$D7,$F9,$44,$9F,$94,$FC,$DE,$39,$1F,$35,$B3,$F3,$78,$BE
	.dc.b	$91,$27,$E9,$3F,$57,$8B,$F7,$BC,$1F,$E6,$39,$FF,$0E,$F9,$7F,$BE
	.dc.b	$7F,$C2,$91,$9B,$FF,$5F,$B5,$FB,$15,$CE,$7E,$F3,$63,$FD,$63,$E2
	.dc.b	$F6,$8F,$DA,$DF,$F5,$AA,$EF,$F1,$18,$33,$F2,$96,$32,$4C,$3F,$D7
	.dc.b	$71,$97,$FD,$7B,$F1,$BD,$3C,$1E,$AD,$AF,$5B,$67,$71,$EE,$9F,$83
	.dc.b	$F1,$52,$3E,$0F,$55,$AF,$7C,$FE,$6C,$93,$F9,$B6,$C7,$FB,$3B,$DB
	.dc.b	$E1,$DB,$FC,$E7,$44,$7C,$0F,$56,$61,$F3,$BD,$6B,$87,$FB,$5A,$6E
	.dc.b	$1C,$66,$D7,$F9,$A3,$D3,$F9,$3F,$15,$F2,$9B,$1F,$61,$B8,$7F,$B8
	.dc.b	$1B,$7F,$D6,$7C,$0F,$D3,$B7,$BF,$E9,$FC,$47,$F3,$E9,$CE,$1F,$E5
	.dc.b	$37,$F5,$78,$5A,$FE,$3D,$18,$8F,$F3,$C3,$5F,$E0,$71,$FC,$CB,$3F
	.dc.b	$F3,$E5,$7F,$9F,$70,$F0,$3F,$E6,$A8,$CD,$7F,$BE,$9F,$F5,$CC,$67
	.dc.b	$D2,$E5,$E6,$73,$64,$64,$E5,$18,$AE,$5F,$2D,$E1,$F8,$4F,$13,$6B
	.dc.b	$C0,$D9,$DC,$78,$E7,$FA,$FF,$E9,$DF,$95,$96,$5F,$E0,$18,$C3,$FD
	.dc.b	$D7,$27,$CE,$F0,$EB,$98,$FF,$8E,$2F,$96,$1F,$E7,$D0,$E1,$FF,$40
	.dc.b	$35,$FF,$51,$3D,$A3,$E4,$5B,$D1,$EB,$5F,$1D,$F9,$DF,$3B,$2B,$E6
	.dc.b	$7C,$E5,$AF,$99,$37,$3F,$44,$DA,$EB,$F4,$9F,$5F,$E2,$7E,$C3,$63
	.dc.b	$EB,$37,$0F,$AC,$DB,$FF,$12,$79,$6F,$8F,$FF,$03,$B6,$FE,$F7,$D2
	.dc.b	$62,$64,$7E,$0C,$FE,$7C,$FE,$1E,$33,$87,$D0,$9C,$9B,$7F,$D8,$D5
	.dc.b	$D5,$FC,$44,$F0,$BF,$1D,$DB,$F9,$33,$6B,$8C,$DA,$FF,$75,$CE,$65
	.dc.b	$7D,$67,$9D,$FA,$8F,$14,$EA,$F8,$56,$E7,$B9,$A7,$13,$E0,$8C,$4F
	.dc.b	$BD,$A4,$8F,$80,$6C,$F6,$B9,$AF,$1C,$68,$F1,$47,$95,$C5,$78,$1C
	.dc.b	$98,$AF,$03,$67,$A1,$F3,$9E,$E6,$49,$5E,$03,$F4,$17,$F0,$2B,$E5
	.dc.b	$6A,$8C,$5F,$4B,$FD,$4C,$E6,$31,$1E,$4B,$E9,$1F,$22,$3B,$BE,$38
	.dc.b	$CC,$72,$AE,$AE,$8C,$4F,$4C,$3E,$01,$F4,$2B,$EE,$C6,$70,$F3,$DF
	.dc.b	$29,$EA,$6A,$FA,$1E,$5A,$F1,$1E,$58,$F1,$C6,$4F,$AC,$FC,$48,$EC
	.dc.b	$7C,$17,$A1,$C2,$FF,$31,$3F,$C0,$DA,$FD,$13,$17,$FE,$1F,$25,$7F
	.dc.b	$87,$DB,$1F,$E1,$F7,$F9,$77,$D2,$7D,$22,$4F,$87,$14,$62,$BE,$46
	.dc.b	$09,$CD,$6E,$B5,$3F,$12,$73,$5C,$1B,$19,$9D,$D7,$C6,$AF,$BF,$18
	.dc.b	$7E,$2C,$7F,$1A,$8C,$47,$C9,$9F,$14,$63,$0F,$B1,$7D,$AF,$A5,$69
	.dc.b	$F3,$12,$FC,$97,$C7,$0F,$87,$39,$AF,$33,$4E,$77,$D6,$FC,$A7,$A2
	.dc.b	$34,$E2,$7C,$0F,$68,$CB,$E9,$9E,$4D,$BF,$14,$FF,$5D,$6C,$7F,$5D
	.dc.b	$6D,$7F,$49,$6C,$FC,$53,$AE,$1F,$0A,$33,$5E,$53,$FB,$37,$C4,$FF
	.dc.b	$6D,$3F,$B7,$7C,$05,$BF,$ED,$91,$67,$FB,$64,$DC,$3F,$AC,$7D,$17
	.dc.b	$F7,$5F,$6B,$FB,$B3,$70,$FE,$E8,$DB,$FE,$E8,$DA,$FF,$71,$4B,$E5
	.dc.b	$FF,$D8,$AD,$3F,$BE,$7B,$18,$89,$1F,$D8,$6F,$FC,$07,$1F,$EA,$78
	.dc.b	$E1,$19,$6B,$FD,$9D,$CD,$7F,$6F,$F8,$BE,$73,$FE,$07,$7F,$C1,$DF
	.dc.b	$17,$2B,$FC,$19,$F4,$16,$BF,$C1,$93,$6F,$FB,$C4,$DA,$FE,$DB,$F3
	.dc.b	$BF,$EC,$8F,$D5,$BF,$D2,$B5,$FD,$F6,$31,$3F,$E1,$B6,$BF,$E1,$96
	.dc.b	$E1,$FC,$CE,$3F,$BC,$3F,$E2,$0E,$BE,$9E,$CF,$C6,$64,$FC,$39,$3F
	.dc.b	$E0,$16,$C7,$C5,$24,$7F,$61,$B6,$0F,$67,$9D,$FF,$48,$9C,$C7,$AF
	.dc.b	$D9,$F6,$7B,$37,$0E,$58,$23,$18,$71,$8C,$4E,$AB,$8B,$C3,$D9,$F8
	.dc.b	$C7,$C7,$F6,$6C,$64,$97,$CE,$32,$FC,$5F,$37,$F0,$E7,$CF,$F6,$6D
	.dc.b	$7F,$6C,$3F,$CC,$E7,$C5,$2C,$FE,$16,$7F,$10,$48,$F5,$FB,$3C,$A7
	.dc.b	$D4,$FD,$BF,$67,$EE,$7B,$16,$BE,$0B,$6B,$E0,$37,$7F,$A7,$1F,$E4
	.dc.b	$0D,$7F,$D4,$11,$83,$E8,$FC,$D7,$E1,$CF,$C4,$7B,$B7,$0F,$9B,$16
	.dc.b	$FE,$64,$DA,$EB,$0F,$C0,$FF,$CC,$31,$86,$DF,$F9,$9F,$07,$CC,$F8
	.dc.b	$61,$72,$B6,$D3,$A9,$A3,$B1,$83,$51,$F9,$18,$FF,$38,$C7,$F9,$55
	.dc.b	$E7,$7F,$8E,$F7,$BC,$13,$C0,$FC,$33,$85,$DB,$FD,$F5,$FD,$F3,$5F
	.dc.b	$F3,$03,$5F,$23,$5F,$5B,$FD,$39,$DF,$83,$E7,$7A,$B6,$35,$1C,$BC
	.dc.b	$1C,$13,$1F,$D6,$4A,$38,$78,$32,$72,$DA,$F3,$D7,$52,$4F,$27,$C1
	.dc.b	$E0,$C8,$EA,$D9,$F7,$F8,$3E,$52,$4F,$89,$F5,$68,$F3,$A9,$27,$E9
	.dc.b	$3A,$C2,$47,$D5,$6D,$FF,$29,$F9,$E7,$B9,$F1,$5B,$5F,$C2,$DB,$3B
	.dc.b	$9F,$95,$F3,$CF,$0E,$92,$FF,$89,$46,$5F,$E2,$9F,$27,$FC,$17,$E4
	.dc.b	$EF,$80,$73,$E0,$3D,$5C,$C7,$D0,$F1,$FF,$8E,$69,$F8,$46,$EC,$69
	.dc.b	$8D,$24,$FD,$86,$BF,$8A,$F1,$F8,$91,$FB,$AF,$4F,$1E,$47,$E0,$1A
	.dc.b	$FF,$4F,$FD,$43,$F1,$49,$A4,$DB,$CE,$BC,$8A,$49,$FE,$6D,$19,$6E
	.dc.b	$DF,$AD,$EB,$FE,$09,$DA,$F9,$9C,$63,$FB,$96,$DF,$EE,$30,$FE,$60
	.dc.b	$FF,$A2,$9A,$FF,$31,$8D,$DF,$F3,$E7,$8C,$7A,$70,$B5,$C2,$3F,$D9
	.dc.b	$8D,$7F,$80,$35,$F0,$47,$D4,$FD,$45,$F1,$3D,$B2,$7B,$54,$EF,$3A
	.dc.b	$BF,$F4,$FB,$5F,$E8,$D1,$B7,$35,$E6,$F9,$17,$C7,$FA,$BE,$2D,$E0
	.dc.b	$53,$51,$2F,$FD,$57,$5C,$1B,$1C,$EF,$E0,$FB,$2F,$76,$F4,$CB,$FF
	.dc.b	$59,$CE,$5B,$3C,$08,$FF,$AE,$3E,$43,$E0,$F1,$7F,$2A,$F9,$0F,$F7
	.dc.b	$64,$FF,$2C,$9C,$A8,$CB,$FE,$F1,$AF,$E5,$91,$8B,$F0,$4E,$DF,$E6
	.dc.b	$15,$CA,$78,$B6,$3F,$9A,$1C,$23,$95,$74,$8E,$38,$B0,$F3,$46,$1B
	.dc.b	$87,$F4,$28,$E4,$74,$AE,$51,$EB,$AE,$4F,$94,$C5,$72,$7D,$85,$AE
	.dc.b	$6E,$8E,$03,$0F,$53,$E1,$BE,$EF,$23,$71,$A7,$85,$F8,$9F,$EA,$86
	.dc.b	$2F,$FC,$45,$3F,$42,$32,$D7,$E5,$FA,$6F,$91,$DB,$63,$FC,$99,$28
	.dc.b	$C4,$72,$6B,$FE,$51,$F9,$1F,$E2,$A9,$A4,$7F,$8C,$BF,$16,$FF,$37
	.dc.b	$CF,$F1,$E7,$FB,$8C,$E7,$6F,$F0,$3E,$BF,$86,$9C,$9F,$22,$F8,$BF
	.dc.b	$28,$CC,$7E,$32,$32,$F1,$8F,$21,$C6,$7A,$EB,$E9,$8C,$61,$E3,$77
	.dc.b	$5C,$EB,$71,$F3,$B0,$F3,$56,$DB,$1E,$33,$8D,$FF,$1B,$AE,$31,$F9
	.dc.b	$17,$FB,$F6,$1E,$A3,$FC,$15,$5C,$9F,$F0,$D9,$9B,$E4,$7F,$85,$9D
	.dc.b	$D7,$A0,$FC,$E4,$7F,$85,$7E,$47,$F7,$8E,$CC,$6D,$FE,$0D,$8F,$F0
	.dc.b	$EC,$9D,$BC,$8F,$F1,$1B,$8A,$FF,$16,$B6,$BB,$4F,$3A,$FF,$0C,$7C
	.dc.b	$8F,$F2,$29,$FF,$39,$36,$7F,$D2,$CD,$7F,$2E,$D7,$FC,$58,$EE,$FF
	.dc.b	$C8,$9E,$4B,$E2,$FA,$F0,$FF,$8A,$25,$39,$C3,$DD,$87,$F8,$EB,$E4
	.dc.b	$5E,$77,$F8,$AD,$B3,$FE,$7F,$E7,$79,$99,$A9,$3F,$E8,$9A,$DD,$FF
	.dc.b	$AB,$5A,$FE,$B9,$AF,$F9,$0D,$DE,$3F,$E5,$BF,$9F,$7A,$FD,$2B,$FD
	.dc.b	$15,$18,$BF,$3C,$3E,$4B,$8C,$7F,$93,$5B,$1F,$E5,$45,$E9,$FF,$91
	.dc.b	$5C,$C7,$1C,$BF,$C7,$91,$E4,$8E,$11,$82,$57,$83,$E7,$DF,$0F,$CE
	.dc.b	$7F,$CA,$11,$89,$F2,$F0,$BD,$8F,$F2,$43,$63,$FD,$5C,$DC,$3F,$DA
	.dc.b	$4D,$7F,$89,$AF,$63,$89,$27,$8B,$FE,$53,$F9,$67,$C8,$52,$4F,$FA
	.dc.b	$EE,$32,$FF,$AF,$7E,$53,$FD,$7A,$D3,$FC,$B8,$E3,$0F,$F7,$23,$5F
	.dc.b	$E4,$BF,$05,$FD,$D2,$7F,$D8,$CD,$8F,$F7,$53,$5D,$F8,$BF,$E6,$7F
	.dc.b	$27,$EC,$47,$F8,$87,$8F,$04,$BF,$27,$C0,$F8,$88,$E0,$F4,$7A,$7E
	.dc.b	$62,$38,$3F,$C4,$7E,$F7,$72,$38,$3E,$07,$C0,$FB,$88,$E0,$F8,$9F
	.dc.b	$03,$F0,$88,$E0,$FF,$17,$F1,$7A,$C8,$E0,$F9,$95,$F2,$88,$E0,$FA
	.dc.b	$11,$F3,$88,$E0,$FF,$1D,$F0,$FC,$72,$38,$3F,$DE,$9F,$19,$F3,$FA
	.dc.b	$BF,$D6,$ED,$8F,$E3,$9F,$22,$F8,$EF,$F5,$14,$E7,$7F,$F5,$AB,$F5
	.dc.b	$5F,$F5,$93,$5C,$F0,$FF,$A7,$63,$10,$79,$60,$F8,$6F,$E4,$31,$87
	.dc.b	$FD,$6A,$7F,$9C,$71,$FE,$17,$3B,$F3,$7F,$88,$4F,$F2,$1B,$FF,$05
	.dc.b	$3B,$7F,$94,$4E,$EB,$8C,$E6,$3F,$DB,$CF,$17,$53,$C2,$3F,$D3,$53
	.dc.b	$C3,$E3,$BB,$FC,$07,$81,$B1,$93,$AC,$65,$FE,$17,$F1,$DF,$DD,$DB
	.dc.b	$E7,$7A,$3B,$FF,$D9,$93,$9C,$7F,$9F,$07,$0F,$F8,$29,$AF,$FB,$7D
	.dc.b	$AF,$FB,$99,$AF,$19,$F8,$BE,$6F,$66,$DF,$C2,$F8,$EF,$92,$7C,$AF
	.dc.b	$8E,$B1,$92,$5F,$C3,$8C,$BF,$17,$E3,$7E,$8F,$C7,$7D,$26,$D7,$CE
	.dc.b	$6C,$EE,$3E,$8C,$FF,$BA,$FB,$9F,$1B,$4F,$B7,$F1,$D6,$BF,$99,$CE
	.dc.b	$9F,$E0,$B3,$FE,$E7,$8C,$5F,$F0,$68,$C1,$F8,$4F,$8A,$E5,$1F,$6D
	.dc.b	$FC,$3F,$C5,$4B,$FC,$1F,$B5,$6B,$F0,$13,$79,$7F,$0B,$36,$BF,$BF
	.dc.b	$72,$BE,$13,$F2,$3E,$8F,$E4,$9B,$3F,$8E,$6E,$3F,$8B,$6E,$09,$4C
	.dc.b	$5F,$1D,$98,$FC,$1C,$60,$FC,$CF,$A3,$9A,$D4,$F3,$AC,$57,$48,$C3
	.dc.b	$FE,$52,$7B,$47,$07,$51,$D5,$D3,$5F,$F2,$CB,$99,$FD,$3F,$C2,$F6
	.dc.b	$9E,$73,$FA,$C8,$FD,$9F,$C2,$C8,$FD,$7F,$C2,$DA,$FD,$64,$DB,$FD
	.dc.b	$8C,$DA,$F9,$BF,$0B,$FE,$0B,$F8,$2F,$E9,$15,$FB,$68,$C4,$FF,$00
	.dc.b	$6B,$FB,$C6,$E1,$F2,$3A,$7C,$CF,$CF,$7A,$7C,$2E,$2B,$F6,$17,$F2
	.dc.b	$9A,$7C,$49,$DB,$FE,$04,$33,$5F,$5E,$39,$1A,$7C,$11,$B8,$CB,$FC
	.dc.b	$63,$0F,$F7,$83,$92,$4F,$03,$F8,$BC,$69,$C5,$79,$6B,$9C,$6A,$EC
	.dc.b	$E6,$F9,$BF,$EF,$68,$CC,$7A,$CC,$9F,$E5,$7B,$ED,$3C,$6B,$BC,$ED
	.dc.b	$EC,$F0,$7E,$03,$61,$3C,$A3,$84,$73,$23,$DE,$BC,$75,$C8,$EE,$7F
	.dc.b	$B0,$6B,$93,$9A,$E2,$DA,$E4,$78,$DF,$B8,$F8,$E7,$8C,$60,$EA,$59
	.dc.b	$F8,$93,$C4,$FB,$71,$97,$AB,$C1,$C9,$F8,$08,$F7,$D7,$AA,$78,$1E
	.dc.b	$53,$D4,$E6,$78,$DF,$E1,$6B,$D7,$5C,$0E,$2F,$49,$F2,$4F,$C6,$8C
	.dc.b	$4F,$8E,$31,$7E,$EA,$EA,$7C,$1B,$F0,$4E,$9B,$1F,$C0,$6B,$E2,$1F
	.dc.b	$BC,$9F,$E0,$4F,$D3,$31,$1B,$3F,$A5,$92,$37,$1E,$8A,$79,$7E,$42
	.dc.b	$57,$D3,$6C,$FD,$EC,$3F,$0F,$18,$7E,$E9,$89,$E0,$49,$E1,$5D,$A3
	.dc.b	$13,$C2,$31,$3A,$AE,$95,$F9,$23,$F3,$CE,$23,$EB,$57,$77,$84,$F4
	.dc.b	$38,$3F,$93,$9E,$0F,$BD,$B1,$F8,$3A,$F0,$D7,$8E,$73,$7F,$C6,$5B
	.dc.b	$5F,$C0,$5B,$7F,$BE,$6D,$7F,$3B,$BE,$75,$E2,$8F,$31,$D5,$F3,$57
	.dc.b	$AA,$76,$DA,$E7,$3F,$00,$DC,$7A,$5A,$7A,$63,$71,$EF,$8E,$B5,$FB
	.dc.b	$CF,$2F,$F9,$97,$C2,$7F,$33,$6C,$FF,$1B,$6D,$7F,$18,$6D,$7A,$FC
	.dc.b	$CF,$E7,$47,$F3,$BF,$32,$DF,$F2,$38,$B7,$FC,$82,$6D,$7F,$33,$F8
	.dc.b	$EF,$E6,$C7,$F4,$3F,$3A,$E1,$FC,$CE,$3F,$26,$9A,$47,$F3,$4F,$F4
	.dc.b	$BF,$3B,$8E,$1F,$CB,$9F,$E9,$DE,$77,$F4,$FF,$3A,$DF,$F3,$76,$D7
	.dc.b	$F3,$46,$E3,$F5,$DF,$EA,$DF,$1D,$2B,$F9,$F7,$B1,$6F,$F9,$DC,$DB
	.dc.b	$FB,$D3,$6B,$FB,$B7,$C6,$FF,$C0,$03,$FC,$05,$7F,$D0,$BE,$3B,$FA
	.dc.b	$CC,$66,$3F,$62,$E9,$FE,$FC,$1C,$3D,$0D,$9F,$DA,$FD,$9E,$6D,$AF
	.dc.b	$EA,$9E,$BF,$F6,$FF,$13,$FB,$83,$63,$27,$F5,$98,$CB,$4F,$EB,$9F
	.dc.b	$3F,$FD,$DC,$FE,$F1,$F3,$F6,$32,$4B,$FE,$BB,$19,$7F,$CB,$FE,$BB
	.dc.b	$E2,$FF,$BB,$B6,$BF,$CF,$FF,$AB,$BE,$37,$35,$27,$F8,$E5,$6E,$FF
	.dc.b	$C3,$CD,$7E,$2B,$5F,$F1,$23,$BE,$0F,$F1,$56,$0F,$A4,$E9,$5F,$E8
	.dc.b	$88,$C5,$F8,$BE,$93,$FC,$41,$69,$FF,$12,$4F,$F8,$EA,$31,$3A,$9C
	.dc.b	$D2,$73,$5E,$CF,$17,$57,$B7,$87,$CF,$C9,$F8,$3F,$39,$70,$F2,$4D
	.dc.b	$C3,$D3,$36,$BF,$D6,$3C,$3F,$EB,$63,$C5,$F1,$96,$3C,$C6,$8F,$F6
	.dc.b	$0C,$7F,$AD,$48,$E0,$FA,$3F,$39,$C7,$1F,$88,$FF,$8B,$3C,$B7,$A1
	.dc.b	$FD,$11,$B5,$FD,$3D,$B5,$FE,$51,$6B,$F9,$85,$E8,$71,$24,$F1,$7F
	.dc.b	$A7,$FC,$97,$F9,$4E,$BF,$CB,$8D,$7F,$CD,$2D,$4D,$AF,$F3,$B5,$78
	.dc.b	$DF,$31,$1F,$37,$9F,$E9,$23,$83,$FB,$44,$B6,$F1,$92,$32,$FE,$D3
	.dc.b	$DB,$91,$FD,$BF,$AF,$FD,$93,$8F,$F9,$C5,$AF,$EE,$1F,$2B,$9C,$7F
	.dc.b	$7A,$49,$FE,$F1,$F1,$CF,$1B,$FB,$8B,$63,$31,$83,$FB,$AF,$CA,$3D
	.dc.b	$0A,$49,$FF,$01,$46,$5F,$F0,$27,$CA,$3C,$BF,$D7,$65,$FE,$87,$8C
	.dc.b	$76,$EB,$CB,$7C,$07,$0B,$DD,$FF,$12,$F9,$E7,$91,$FE,$0F,$84,$61
	.dc.b	$FF,$09,$35,$FF,$4A,$35,$FC,$26,$9C,$77,$F5,$47,$E2,$E3,$06,$7D
	.dc.b	$39,$1F,$8F,$6B,$FE,$9F,$E4,$75,$F8,$DB,$1F,$72,$30,$79,$FC,$1F
	.dc.b	$F0,$EC,$1A,$4E,$0D,$7A,$52,$7F,$88,$B5,$F0,$76,$F7,$23,$F8,$A7
	.dc.b	$B6,$DF,$EB,$5E,$EF,$F0,$07,$87,$70,$FF,$3A,$36,$FF,$D8,$B8,$FD
	.dc.b	$57,$F7,$26,$2F,$9B,$FC,$89,$AF,$F2,$27,$93,$EB,$6C,$73,$8F,$E3
	.dc.b	$DF,$1D,$AD,$0F,$87,$1F,$F3,$EF,$C2,$5F,$4C,$EE,$65,$FE,$99,$F9
	.dc.b	$C4,$7F,$CA,$1F,$2A,$D7,$A7,$C7,$BD,$BA,$79,$A7,$55,$EE,$0E,$1F
	.dc.b	$E5,$5B,$FF,$B7,$63,$19,$FF,$B8,$9B,$19,$C7,$E0,$FC,$13,$E4,$6F
	.dc.b	$4C,$C7,$B4,$E5,$B3,$C0,$8F,$8F,$E5,$3E,$2E,$5F,$EF,$8F,$94,$5F
	.dc.b	$FB,$BE,$73,$49,$CB,$FE,$FE,$9E,$CE,$5F,$F7,$B1,$88,$DC,$27,$9C
	.dc.b	$61,$B3,$C2,$FF,$DF,$D3,$C2,$39,$1E,$3C,$7F,$A8,$5F,$70,$75,$73
	.dc.b	$3C,$6B,$9D,$75,$AF,$73,$CF,$0B,$1C,$B0,$F1,$BE,$DA,$E2,$D3,$67
	.dc.b	$79,$B3,$C2,$3E,$4C,$F0,$69,$AC,$3C,$B7,$E1,$8F,$F0,$F9,$CC,$B5
	.dc.b	$F4,$1F,$F1,$1B,$5F,$A9,$ED,$BE,$57,$FD,$45,$28,$C4,$F2,$6B,$C9
	.dc.b	$AF,$83,$C3,$FE,$38,$9A,$47,$F9,$03,$F0,$32,$FD,$06,$BF,$0A,$7E
	.dc.b	$8C,$E7,$85,$39,$9E,$87,$D7,$3F,$C1,$11,$A7,$07,$23,$55,$D9,$C9
	.dc.b	$CB,$5F,$C5,$C6,$5F,$C6,$C7,$17,$33,$D5,$CC,$71,$CF,$F8,$B4,$E5
	.dc.b	$F8,$51,$E0,$8C,$3F,$E5,$2C,$3D,$98,$7A,$DB,$1B,$8F,$DE,$61,$E7
	.dc.b	$3F,$DA,$33,$A7,$DD,$3F,$02,$77,$5F,$06,$77,$3C,$CF,$E0,$67,$E7
	.dc.b	$3E,$57,$F7,$D5,$FC,$25,$B1,$A0,$FF,$9B,$A7,$BB,$A6,$ED,$DE,$BF
	.dc.b	$4F,$5C,$23,$0F,$72,$99,$8F,$E3,$1F,$2B,$FC,$A9,$FE,$6F,$4F,$25
	.dc.b	$F2,$3C,$2B,$57,$EA,$AD,$4E,$78,$7F,$C9,$F7,$C3,$0E,$B1,$89,$F0
	.dc.b	$FC,$AB,$D7,$FF,$7A,$B6,$BF,$E0,$03,$71,$FD,$4B,$5F,$F4,$4B,$9C
	.dc.b	$7E,$77,$CA,$BD,$FE,$B8,$7F,$98,$23,$18,$7F,$35,$BE,$0F,$F9,$86
	.dc.b	$33,$87,$CF,$9C,$BF,$E6,$18,$CE,$1F,$7B,$E5,$7C,$53,$F3,$9A,$FF
	.dc.b	$16,$6A,$69,$D1,$7C,$AB,$E3,$FE,$9C,$E6,$31,$8F,$F4,$B6,$D7,$F0
	.dc.b	$A6,$BF,$EA,$A6,$BF,$80,$6B,$FC,$A3,$4F,$F0,$EC,$FF,$12,$8F,$B7
	.dc.b	$E4,$FE,$61,$FC,$3C,$E9,$FE,$2E,$D7,$FD,$6F,$F2,$AF,$5B,$84,$7F
	.dc.b	$34,$8C,$5F,$56,$C7,$F2,$06,$BF,$EC,$0D,$9F,$33,$CA,$3F,$04,$D8
	.dc.b	$CD,$79,$DB,$3F,$CA,$BE,$2F,$FB,$DC,$FA,$DB,$19,$F0,$BA,$46,$23
	.dc.b	$FD,$57,$C5,$DF,$C5,$FB,$F1,$8B,$FF,$59,$CF,$8F,$E4,$FF,$D6,$13
	.dc.b	$FD,$41,$A9,$A6,$3E,$59,$E8,$7A,$DB,$7F,$3D,$B5,$FD,$09,$AF,$FA
	.dc.b	$FB,$E1,$BC,$84,$9F,$F6,$2B,$FE,$8A,$F0,$64,$7F,$9F,$1B,$3C,$FE
	.dc.b	$23,$C0,$49,$F0,$3E,$FF,$87,$91,$FE,$D0,$6C,$FF,$93,$FE,$77,$FD
	.dc.b	$2E,$49,$FF,$44,$BE,$B9,$91,$EB,$6C,$FF,$B6,$39,$7E,$61,$27,$FD
	.dc.b	$BA,$FF,$A8,$BE,$76,$47,$FA,$51,$B3,$F6,$FE,$1B,$F8,$9F,$09,$FF
	.dc.b	$6B,$FC,$7B,$E4,$33,$13,$0F,$C0,$3F,$78,$8E,$0F,$E3,$33,$FE,$BA
	.dc.b	$ED,$FE,$6B,$7D,$67,$3F,$03,$DB,$E0,$64,$70,$F3,$EE,$7F,$C7,$71
	.dc.b	$FF,$5F,$CF,$4A,$F2,$D6,$B1,$FE,$7D,$5F,$EA,$46,$F8,$9E,$AF,$67
	.dc.b	$D6,$DC,$3C,$AD,$C7,$B3,$FC,$D2,$3F,$DE,$31,$FD,$BB,$E0,$FE,$1F
	.dc.b	$C8,$65,$93,$FB,$B3,$FD,$94,$47,$F5,$8F,$FB,$EB,$06,$2F,$90,$E3
	.dc.b	$87,$B5,$F9,$FE,$E4,$BF,$93,$EE,$5A,$F8,$53,$70,$F7,$CD,$AF,$F1
	.dc.b	$AC,$76,$F9,$0F,$E9,$EB,$D8,$FF,$03,$CF,$D4,$9C,$C6,$30,$F9,$D1
	.dc.b	$89,$FC,$0B,$5F,$F3,$61,$FE,$21,$2F,$57,$2F,$73,$E4,$32,$3E,$23
	.dc.b	$F0,$88,$E0,$FE,$BB,$6F,$90,$E1,$87,$F5,$B8,$C4,$EB,$BF,$CF,$EE
	.dc.b	$7A,$1B,$19,$39,$C6,$5A,$76,$F8,$3F,$79,$F0,$3B,$EC,$64,$98,$78
	.dc.b	$63,$2F,$F8,$06,$1F,$21,$F8,$53,$E4,$F8,$96,$3E,$2C,$7E,$00,$8F
	.dc.b	$E1,$5B,$3E,$DF,$0B,$83,$67,$E1,$F8,$3F,$53,$C1,$FA,$AD,$8C,$9F
	.dc.b	$22,$32,$D3,$E5,$FC,$87,$DC,$3E,$EF,$C8,$58,$C9,$2F,$E7,$46,$5F
	.dc.b	$F0,$F7,$C3,$FF,$87,$67,$FC,$32,$D8,$72,$FE,$2F,$E4,$3F,$C3,$33
	.dc.b	$99,$34,$E1,$3B,$AF,$F5,$A4,$66,$78,$9F,$92,$AB,$3A,$8A,$7E,$0E
	.dc.b	$2D,$7F,$8F,$AB,$95,$7E,$2A,$72,$FA,$2A,$9E,$8E,$24,$E2,$3C,$E6
	.dc.b	$AF,$CC,$FC,$23,$57,$EA,$AE,$37,$E5,$24,$ED,$F1,$18,$AF,$34,$75
	.dc.b	$9E,$2E,$C9,$1A,$9E,$31,$87,$D4,$73,$3F,$7B,$3C,$A7,$84,$EC,$FF
	.dc.b	$5A,$D7,$9E,$FD,$31,$E1,$6D,$F3,$8F,$89,$18,$79,$57,$C2,$3A,$92
	.dc.b	$7C,$71,$F4,$DE,$4F,$CA,$7A,$47,$17,$6F,$18,$F2,$9F,$7E,$38,$D7
	.dc.b	$DA,$8E,$87,$AC,$E8,$79,$2F,$E0,$4F,$42,$47,$47,$E4,$D7,$78,$E2
	.dc.b	$F8,$4F,$23,$B3,$B5,$7C,$79,$F1,$C7,$F8,$8E,$3F,$DB,$F1,$FE,$37
	.dc.b	$6C,$7F,$B2,$9A,$FF,$9F,$8F,$39,$5C,$BF,$20,$C1,$D4,$FE,$63,$E5
	.dc.b	$FE,$14,$93,$F2,$AF,$FC,$AD,$5F,$91,$9C,$D7,$C3,$6D,$FB,$6B,$73
	.dc.b	$D2,$F9,$56,$9F,$8F,$5F,$D0,$E3,$13,$F0,$E3,$FD,$4D,$5C,$AB,$F9
	.dc.b	$AD,$6A,$BE,$5C,$7E,$4A,$31,$7E,$58,$EE,$49,$F3,$36,$FC,$D5,$C2
	.dc.b	$7C,$C6,$89,$1A,$9E,$07,$A2,$FC,$93,$FD,$16,$B9,$4F,$98,$DC,$7A
	.dc.b	$DF,$31,$87,$D2,$7F,$03,$8E,$2F,$38,$D5,$FA,$E3,$C7,$1E,$19,$DD
	.dc.b	$FC,$67,$E4,$BF,$DB,$7C,$EF,$ED,$7E,$2F,$F6,$C6,$D7,$F5,$E6,$DF
	.dc.b	$F5,$D6,$D7,$D0,$F9,$0F,$EE,$67,$F7,$4F,$90,$B7,$FD,$9A,$2D,$FF
	.dc.b	$65,$9B,$5F,$EE,$0F,$35,$F0,$BF,$EC,$C6,$E1,$FE,$CA,$F8,$F7,$C4
	.dc.b	$E6,$A4,$FF,$A4,$EB,$FC,$DA,$48,$FF,$3B,$1F,$E0,$AA,$7C,$87,$0C
	.dc.b	$BF,$DA,$11,$8B,$ED,$F2,$1B,$F1,$B8,$36,$32,$78,$E3,$2D,$3C,$DF
	.dc.b	$1F,$DC,$F0,$7C,$5D,$8C,$93,$0F,$0C,$65,$FF,$7B,$7A,$9F,$EA,$0D
	.dc.b	$95,$FE,$F2,$8C,$3F,$E2,$6C,$9F,$05,$FE,$F2,$9C,$DF,$FA,$3F,$8B
	.dc.b	$FD,$24,$4B,$FF,$5C,$4F,$FA,$AA,$7A,$4F,$C5,$F9,$0E,$4D,$AF,$A1
	.dc.b	$E2,$FC,$9F,$57,$E5,$36,$32,$7D,$58,$CB,$4F,$B1,$EA,$FD,$13,$E9
	.dc.b	$7C,$CD,$8C,$93,$1F,$A7,$19,$7F,$BB,$C1,$9F,$1B,$FE,$52,$8F,$F2
	.dc.b	$39,$98,$46,$5F,$BB,$F3,$BF,$29,$E6,$72,$F4,$31,$C1,$98,$E3,$1F
	.dc.b	$EC,$BA,$E5,$1F,$C9,$5A,$F7,$6B,$CE,$FF,$2D,$EE,$F4,$75,$5F,$91
	.dc.b	$8F,$C9,$78,$DF,$93,$8B,$3F,$9D,$6D,$7E,$51,$B8,$79,$B8,$BD,$8D
	.dc.b	$F9,$DF,$CE,$1F,$9D,$F9,$DB,$87,$E9,$E2,$E1,$F9,$E9,$B5,$EF,$FD
	.dc.b	$4D,$F1,$1E,$7A,$F9,$D1,$8C,$3F,$61,$F0,$19,$94,$E5,$F9,$66,$2B
	.dc.b	$FD,$E5,$19,$8F,$E9,$8D,$7E,$B3,$5E,$F1,$FB,$5F,$9C,$ED,$3D,$70
	.dc.b	$F8,$11,$E1,$F6,$BF,$D5,$37,$FD,$5C,$38,$7F,$AD,$35,$FC,$23,$5F
	.dc.b	$C4,$F6,$7D,$D2,$65,$CE,$3F,$13,$1E,$AF,$85,$93,$9C,$51,$8A,$F1
	.dc.b	$5B,$DB,$FC,$54,$78,$1A,$FE,$66,$35,$7D,$DB,$1E,$BE,$37,$77,$E4
	.dc.b	$49,$5E,$58,$CB,$E3,$F8,$CF,$9E,$47,$07,$F7,$FB,$FF,$01,$11,$C1
	.dc.b	$EF,$F8,$C7,$A9,$E6,$C5,$18,$9D,$4E,$63,$13,$F4,$74,$7C,$D7,$E0
	.dc.b	$09,$7F,$1D,$E9,$F2,$AF,$5F,$F9,$CD,$FD,$98,$C7,$93,$F4,$57,$B0
	.dc.b	$F9,$2C,$5F,$0B,$FF,$0B,$7C,$AB,$D6,$FB,$10,$8C,$3F,$6D,$AF,$E1
	.dc.b	$DA,$FF,$87,$ED,$D7,$C9,$B1,$F8,$36,$BF,$8F,$6B,$F8,$EF,$7B,$C3
	.dc.b	$C2,$69,$23,$EC,$79,$5E,$6E,$43,$49,$1F,$8D,$F9,$AF,$5F,$11,$A4
	.dc.b	$8F,$EB,$DE,$F7,$C0,$F4,$0D,$24,$7E,$1F,$CD,$F0,$11,$C1,$F9,$BF
	.dc.b	$9A,$EE,$47,$07,$F6,$BE,$1E,$C4,$70,$43,$F8,$0D,$FD,$D7,$F9,$99
	.dc.b	$FC,$DD,$C3,$FD,$BB,$E4,$3F,$9B,$78,$1F,$DF,$EA,$F4,$7F,$39,$BD
	.dc.b	$CE,$6F,$FC,$01,$7A,$7F,$A0,$BF,$2C,$8F,$E9,$5A,$FF,$14,$D9,$1C
	.dc.b	$63,$FA,$2B,$CE,$5F,$AC,$7F,$98,$5F,$E8,$5F,$2F,$FD,$B2,$BD,$70
	.dc.b	$F9,$87,$C4,$71,$6C,$FC,$01,$5C,$5F,$E4,$AD,$AE,$BB,$78,$E3,$1A
	.dc.b	$7F,$59,$7C,$F5,$CF,$BF,$F9,$AB,$7A,$BD,$51,$FA,$58,$D7,$8B,$82
	.dc.b	$38,$3C,$9F,$2F,$F0,$B2,$FE,$D5,$F2,$EB,$FE,$D1,$39,$A4,$E5,$FE
	.dc.b	$E1,$3E,$27,$2F,$F6,$F3,$11,$B8,$4F,$38,$C3,$67,$85,$FF,$77,$9E
	.dc.b	$11,$C8,$F1,$EF,$FE,$F6,$BE,$E7,$F8,$06,$F9,$3E,$4B,$E5,$5D,$23
	.dc.b	$DA,$F2,$C2,$C7,$1B,$E9,$87,$82,$F8,$CF,$33,$67,$AA,$77,$7C,$E3
	.dc.b	$C6,$7C,$99,$E2,$D3,$8E,$1E,$67,$96,$1F,$D3,$63,$E7,$47,$FA,$BD
	.dc.b	$AF,$53,$1E,$93,$E6,$38,$36,$3F,$B0,$4A,$31,$1C,$27,$F9,$1B,$FC
	.dc.b	$E8,$FE,$4B,$E1,$E0,$8E,$0F,$5D,$FF,$92,$B6,$F8,$31,$FE,$2F,$9F
	.dc.b	$87,$39,$8F,$82,$64,$E3,$BF,$F8,$B7,$5F,$C1,$46,$5F,$C2,$4F,$17
	.dc.b	$31,$E5,$73,$3C,$74,$FE,$F9,$7E,$08,$C5,$7F,$B5,$63,$18,$79,$A3
	.dc.b	$95,$7F,$0C,$AD,$C7,$F9,$0E,$B6,$FA,$0D,$9F,$E4,$BC,$3C,$95,$C6
	.dc.b	$FF,$75,$1E,$F3,$FB,$CD,$72,$7F,$C8,$C6,$63,$85,$7F,$88,$8F,$81
	.dc.b	$58,$9F,$F2,$67,$CC,$7F,$86,$EB,$FC,$AC,$DA,$FF,$2E,$3F,$EF,$49
	.dc.b	$DE,$BF,$E2,$F8,$D4,$9A,$6E,$B9,$CF,$F8,$D7,$E6,$1E,$A5,$3F,$C7
	.dc.b	$98,$7F,$85,$7D,$1C,$5F,$F1,$7A,$DB,$FE,$21,$6E,$3F,$E2,$D6,$BF
	.dc.b	$E6,$D0,$E1,$FF,$38,$7C,$C7,$FA,$41,$3F,$E9,$2B,$FE,$51,$F1,$BF
	.dc.b	$E4,$98,$D1,$3D,$8F,$2F,$3A,$75,$13,$1F,$F1,$A5,$7F,$A5,$F0,$FF
	.dc.b	$41,$35,$EF,$3F,$E1,$9F,$97,$FF,$63,$C7,$F9,$15,$AF,$A5,$AF,$FA
	.dc.b	$1D,$AF,$FA,$35,$F5,$13,$DE,$FF,$30,$38,$9E,$C4,$BF,$F2,$44,$78
	.dc.b	$FD,$77,$C3,$FF,$8A,$DB,$3F,$E4,$D6,$E1,$FE,$59,$6B,$F3,$57,$C3
	.dc.b	$F1,$24,$F1,$7A,$7C,$3C,$8E,$8D,$AF,$F3,$27,$CA,$7F,$81,$27,$FC
	.dc.b	$B8,$D8,$CF,$8B,$FE,$3F,$8C,$47,$DD,$5F,$1B,$92,$57,$27,$87,$99
	.dc.b	$23,$AB,$6B,$FC,$E9,$F0,$2F,$95,$FF,$26,$36,$BF,$DA,$31,$83,$FC
	.dc.b	$A1,$F2,$8F,$88,$A4,$9F,$F3,$A4,$65,$FF,$3B,$1F,$C8,$4B,$E0,$FF
	.dc.b	$96,$5C,$61,$FE,$85,$6B,$FE,$92,$F9,$4F,$F4,$A9,$23,$FC,$F8,$FF
	.dc.b	$A6,$3E,$52,$47,$F9,$FD,$B3,$FE,$5F,$F2,$3F,$D3,$A4,$9F,$F4,$23
	.dc.b	$FE,$A0,$F4,$A4,$7F,$A1,$DB,$3F,$E6,$7F,$6F,$FD,$4A,$49,$FF,$46
	.dc.b	$3F,$EA,$8F,$94,$91,$FE,$8F,$6C,$FF,$9B,$FA,$FF,$D4,$A7,$EF,$3C
	.dc.b	$CF,$F3,$35,$7F,$84,$7D,$2F,$D5,$36,$7F,$D6,$64,$7F,$D3,$4D,$AF
	.dc.b	$F7,$1F,$7F,$F8,$8B,$C3,$EB,$5F,$E2,$0C,$3F,$8B,$4F,$F8,$9B,$E6
	.dc.b	$BC,$1F,$3B,$E1,$6D,$6D,$B8,$74,$B4,$62,$FF,$DD,$F1,$FE,$AA,$AE
	.dc.b	$DE,$FF,$AF,$C8,$F6,$32,$32,$78,$E3,$18,$7F,$B4,$3C,$6F,$F4,$3C
	.dc.b	$FF,$8C,$5B,$1E,$1F,$77,$FD,$B1,$18,$C7,$FD,$FC,$EA,$7F,$8E,$35
	.dc.b	$11,$C3,$7A,$BF,$D4,$D1,$CE,$FF,$A5,$65,$FE,$77,$7C,$53,$F5,$7D
	.dc.b	$6F,$AD,$EB,$7D,$7F,$9C,$B5,$F3,$5B,$8F,$C7,$6E,$5F,$19,$FB,$DF
	.dc.b	$31,$2B,$EA,$74,$DB,$F9,$D3,$6F,$E6,$CD,$AF,$F7,$0F,$2F,$FB,$A4
	.dc.b	$FC,$47,$4D,$8F,$F2,$C7,$42,$7F,$D8,$4D,$8F,$F6,$B4,$60,$8F,$E1
	.dc.b	$7A,$79,$65,$F7,$DF,$CA,$57,$E5,$7A,$6D,$7E,$2D,$B5,$F8,$56,$E5
	.dc.b	$F8,$47,$F3,$98,$4A,$FC,$97,$45,$BF,$C6,$4D,$BF,$C5,$CD,$AF,$B3
	.dc.b	$4E,$2F,$E2,$0F,$E1,$FB,$7F,$51,$F3,$1F,$84,$65,$7F,$C0,$1A,$DD
	.dc.b	$7E,$12,$7E,$F9,$27,$F4,$9E,$DF,$09,$FC,$F3,$FB,$3F,$98,$FE,$DF
	.dc.b	$4A,$C1,$FE,$8A,$70,$FE,$B6,$36,$EA,$BF,$67,$3A,$6F,$89,$D0,$EB
	.dc.b	$E1,$DA,$E9,$16,$78,$CD,$D3,$C5,$1F,$9C,$F9,$3F,$D9,$DF,$FA,$5D
	.dc.b	$B1,$E0,$E8,$DD,$70,$84,$EA,$30,$DA,$F9,$ED,$94,$FF,$41,$AE,$75
	.dc.b	$AE,$DB,$5E,$36,$BD,$9F,$93,$38,$BE,$93,$9A,$F4,$56,$9A,$48,$F0
	.dc.b	$BD,$CF,$A1,$5C,$E6,$46,$EF,$DF,$97,$EE,$EF,$C9,$39,$39,$3F,$15
	.dc.b	$C3,$A3,$B9,$67,$9D,$F6,$8C,$9E,$92,$E1,$F1,$E7,$E3,$57,$49,$D4
	.dc.b	$7C,$EC,$3D,$D1,$C1,$E9,$19,$9F,$69,$C6,$BE,$39,$D7,$0F,$65,$F4
	.dc.b	$9F,$5D,$71,$8F,$61,$D6,$39,$5F,$82,$74,$48,$E2,$7F,$9A,$6B,$AD
	.dc.b	$F8,$2B,$CB,$39,$C2,$E1,$F1,$63,$8B,$F8,$6B,$E3,$5A,$8F,$31,$9A
	.dc.b	$C5,$F5,$BF,$7C,$F9,$C9,$87,$95,$E1,$1E,$A3,$EB,$C7,$CE,$8D,$C7
	.dc.b	$89,$EF,$19,$39,$3E,$23,$38,$79,$A7,$A4,$7C,$89,$DD,$7E,$AA,$3F
	.dc.b	$CC,$CD,$7F,$12,$63,$0F,$D6,$38,$21,$F8,$B3,$55,$FE,$6F,$7E,$41
	.dc.b	$F1,$FD,$CF,$97,$7F,$E2,$D6,$E0,$69,$2B,$E0,$C6,$5B,$19,$AF,$86
	.dc.b	$75,$8C,$91,$CD,$77,$8C,$3F,$3C,$C4,$6A,$BB,$36,$BB,$4E,$E7,$2D
	.dc.b	$8C,$D7,$E1,$23,$99,$97,$B9,$99,$E2,$D9,$E3,$5F,$1A,$B8,$CE,$4E
	.dc.b	$31,$9A,$E2,$DA,$E2,$FD,$E6,$E1,$FA,$96,$E1,$FC,$F1,$B7,$FC,$ED
	.dc.b	$B5,$FC,$E1,$B8,$7F,$0C,$6C,$FE,$9D,$B3,$F3,$CF,$8D,$1F,$D1,$7E
	.dc.b	$0F,$FA,$4F,$CC,$7F,$4A,$6C,$64,$FE,$95,$19,$69,$FD,$37,$E6,$3F
	.dc.b	$A8,$9F,$D4,$BE,$62,$C6,$49,$7F,$D3,$A3,$2F,$F1,$4E,$DF,$E6,$07
	.dc.b	$F5,$AF,$98,$B5,$FD,$1E,$3F,$33,$3F,$D2,$DA,$FF,$46,$6E,$A6,$9F
	.dc.b	$D9,$3E,$17,$9C,$FF,$5E,$7F,$B3,$FA,$1F,$DA,$3C,$FB,$87,$F6,$76
	.dc.b	$D7,$F6,$46,$EF,$FC,$E1,$FE,$DF,$E7,$CA,$FE,$DD,$F2,$F6,$FF,$B5
	.dc.b	$CD,$BF,$E7,$F3,$6B,$FA,$CF,$77,$F4,$F3,$FB,$CF,$A9,$63,$FB,$34
	.dc.b	$7B,$08,$FE,$D7,$0F,$EC,$B1,$87,$FB,$CF,$2F,$16,$CF,$F8,$07,$D4
	.dc.b	$FF,$02,$9F,$E0,$6F,$42,$C6,$49,$3F,$E0,$28,$CB,$FE,$04,$F9,$7F
	.dc.b	$F0,$6F,$CB,$FF,$83,$9B,$5F,$E0,$D6,$CE,$E3,$FC,$21,$3F,$DA,$FB
	.dc.b	$64,$7F,$97,$B5,$FE,$FF,$FA,$CB,$E2,$F3,$87,$F5,$6B,$FB,$64,$BF
	.dc.b	$81,$CA,$8C,$4F,$FA,$2A,$DF,$31,$CB,$1E,$AF,$F8,$01,$43,$E6,$30
	.dc.b	$DC,$7C,$87,$95,$7C,$7E,$A6,$4E,$B0,$FF,$07,$D7,$17,$C5,$5A,$7C
	.dc.b	$9F,$1F,$DB,$E3,$3B,$B6,$F9,$63,$E4,$8C,$57,$09,$CE,$0B,$FF,$0B
	.dc.b	$FC,$1B,$C0,$A7,$AB,$E1,$2D,$F9,$70,$56,$E3,$F1,$0D,$8F,$E7,$CD
	.dc.b	$7F,$CB,$27,$0A,$FD,$6C,$6F,$2F,$AF,$F3,$1D,$AB,$E8,$3F,$E2,$AF
	.dc.b	$D5,$9F,$29,$FB,$46,$E3,$F9,$76,$D7,$F4,$D6,$BE,$05,$F2,$9C,$49
	.dc.b	$3C,$5E,$9F,$29,$23,$A3,$6B,$FA,$C7,$EA,$FF,$EA,$A9,$FD,$33,$63
	.dc.b	$27,$F0,$87,$0F,$C0,$7F,$DB,$6F,$2F,$8F,$D1,$27,$9F,$BF,$C4,$8E
	.dc.b	$0F,$DA,$7D,$13,$E6,$BF,$8A,$B6,$33,$18,$3F,$8C,$FD,$03,$D4,$A4
	.dc.b	$9F,$DF,$46,$5F,$E0,$1F,$31,$FD,$BF,$B6,$F9,$FF,$A9,$D3,$FB,$C4
	.dc.b	$BD,$16,$BA,$BF,$87,$E8,$FF,$09,$B6,$FF,$EE,$67,$24,$3F,$C0,$5D
	.dc.b	$2F,$4B,$83,$63,$F8,$93,$5D,$C6,$BC,$99,$1A,$6B,$F5,$BE,$0F,$C5
	.dc.b	$66,$92,$3E,$CF,$C1,$E8,$8E,$0F,$BB,$F0,$7C,$88,$E0,$FC,$0F,$C1
	.dc.b	$F6,$23,$83,$F0,$BF,$05,$E2,$23,$83,$F1,$3E,$9F,$98,$8E,$0F,$C6
	.dc.b	$FC,$07,$FB,$03,$BF,$F8,$DF,$C9,$DE,$6F,$E1,$4D,$AF,$F7,$EE,$AB
	.dc.b	$C4,$7D,$27,$57,$FC,$5A,$AF,$88,$77,$7C,$6F,$F7,$CD,$7F,$43,$7F
	.dc.b	$30,$8F,$F4,$7F,$D6,$1E,$9F,$4F,$87,$93,$FC,$1E,$72,$D9,$CB,$FC
	.dc.b	$B6,$7C,$F3,$A8,$FF,$46,$FC,$AB,$C8,$F4,$FC,$7B,$E1,$29,$EE,$9D
	.dc.b	$57,$F1,$56,$BC,$AB,$46,$F4,$F8,$BC,$6F,$0E,$F6,$4B,$FE,$3D,$39
	.dc.b	$6C,$F0,$23,$FC,$8F,$E5,$7E,$57,$12,$1F,$28,$BF,$F7,$CD,$6A,$73
	.dc.b	$1F,$EF,$C3,$B3,$83,$93,$89,$E5,$1D,$67,$94,$62,$11,$89,$E9,$7E
	.dc.b	$39,$E8,$78,$DC,$57,$1A,$E1,$3D,$71,$FF,$6F,$BC,$2F,$C4,$74,$7D
	.dc.b	$D5,$C6,$76,$DC,$33,$5B,$AE,$58,$78,$23,$E1,$DF,$06,$D7,$0A,$F8
	.dc.b	$35,$B6,$99,$3D,$F5,$C5,$F4,$4F,$F8,$67,$0F,$9D,$1F,$D4,$1A,$F6
	.dc.b	$3F,$D2,$4B,$77,$CA,$F1,$6C,$E9,$AF,$36,$BF,$D7,$8F,$E4,$1E,$67
	.dc.b	$F2,$17,$FB,$29,$82,$67,$F5,$F0,$FC,$E7,$37,$AE,$7F,$11,$5E,$D9
	.dc.b	$CC,$7D,$83,$27,$DA,$73,$5F,$81,$8C,$BD,$76,$5F,$59,$F1,$CE,$63
	.dc.b	$17,$DE,$B7,$5D,$CD,$EB,$FC,$46,$33,$7D,$E7,$27,$F8,$BE,$33,$3F
	.dc.b	$0E,$7B,$4F,$3A,$ED,$72,$18,$78,$1B,$87,$8A,$3F,$CE,$75,$B7,$47
	.dc.b	$C4,$2C,$6E,$BC,$18,$79,$DF,$F0,$B7,$CA,$FE,$F5,$BB,$6F,$1F,$DB
	.dc.b	$4F,$F7,$9C,$3C,$4E,$E3,$CB,$5F,$A9,$8C,$4E,$AB,$95,$27,$95,$6E
	.dc.b	$73,$F2,$8F,$95,$FF,$45,$B6,$7F,$CD,$CD,$BF,$F7,$23,$5F,$F1,$3B
	.dc.b	$9C,$3E,$27,$C5,$3E,$6F,$AE,$1F,$E5,$18,$C6,$1F,$E3,$5F,$96,$7B
	.dc.b	$BE,$66,$C7,$F9,$13,$A1,$F1,$79,$A9,$3F,$E8,$1A,$DE,$9F,$EF,$76
	.dc.b	$BF,$E3,$A6,$BF,$C3,$5D,$E7,$F6,$7E,$85,$F0,$FD,$2B,$E5,$9F,$80
	.dc.b	$8C,$DF,$F9,$2F,$E5,$F8,$BF,$C5,$DB,$3F,$CA,$D7,$C2,$7D,$07,$30
	.dc.b	$7C,$6E,$F2,$FF,$5C,$C6,$23,$51,$83,$EC,$FD,$0F,$F3,$5B,$F1,$46
	.dc.b	$39,$3E,$EF,$CC,$3E,$1F,$FD,$86,$DA,$FF,$53,$37,$0F,$F2,$E8,$70
	.dc.b	$FF,$98,$1F,$EC,$B1,$C3,$C2,$E6,$47,$FD,$61,$F4,$2F,$86,$FF,$68
	.dc.b	$B6,$33,$18,$3F,$DA,$BF,$33,$FE,$29,$8F,$F5,$C3,$70,$F1,$B5,$FF
	.dc.b	$38,$B5,$FE,$88,$FF,$8E,$E3,$8F,$A7,$D0,$8F,$FB,$0F,$E6,$FE,$BD
	.dc.b	$7B,$9A,$FF,$9F,$1A,$FF,$4F,$6B,$FD,$8F,$E6,$7D,$4F,$C6,$6B,$FE
	.dc.b	$84,$6B,$FD,$5D,$AF,$F9,$BF,$C1,$FF,$79,$D7,$F3,$73,$FD,$AC,$58
	.dc.b	$FA,$2D,$7F,$D1,$FF,$0F,$FE,$72,$7F,$B1,$35,$FE,$EC,$D4,$DF,$99
	.dc.b	$7C,$B7,$F1,$59,$FE,$D5,$18,$AF,$F7,$33,$63,$F0,$2D,$7F,$D3,$EF
	.dc.b	$F8,$5E,$37,$E5,$73,$9F,$F0,$34,$E5,$E3,$1F,$E8,$9F,$8D,$95,$FE
	.dc.b	$8E,$6C,$F6,$EE,$E4,$AD,$18,$6D,$79,$A3,$1F,$2D,$23,$D2,$DA,$FF
	.dc.b	$4E,$CF,$FA,$64,$93,$FE,$9E,$7B,$DA,$BF,$90,$B6,$7F,$D9,$FF,$33
	.dc.b	$F0,$FB,$7F,$9E,$4E,$70,$FE,$02,$1C,$3F,$EC,$46,$BF,$E1,$76,$BF
	.dc.b	$E2,$B6,$BC,$6F,$FD,$E1,$5F,$26,$BB,$C6,$32,$CD,$EB,$1B,$8F,$99
	.dc.b	$B8,$FF,$14,$C1,$8F,$C8,$8C,$3E,$7A,$E0,$7B,$B1,$F0,$9E,$2E,$FB
	.dc.b	$19,$25,$72,$8C,$BF,$0B,$E5,$DF,$21,$FD,$29,$B5,$F2,$BE,$07,$13
	.dc.b	$23,$F8,$6D,$2B,$6F,$F1,$06,$BF,$EE,$E6,$BF,$E4,$17,$79,$7F,$6D
	.dc.b	$C9,$F2,$FD,$6B,$DE,$FC,$CF,$8F,$FE,$CD,$5E,$4C,$3E,$EC,$63,$1F
	.dc.b	$F0,$6C,$60,$FF,$30,$4F,$F7,$98,$CD,$FC,$AA,$E1,$5C,$7E,$5E,$3C
	.dc.b	$0E,$35,$FC,$66,$F9,$4E,$AF,$EB,$4A,$3C,$5E,$F6,$6B,$FC,$15,$C8
	.dc.b	$F8,$0F,$EE,$AD,$C7,$B6,$3F,$E1,$0C,$7F,$90,$3F,$94,$9D,$36,$BF
	.dc.b	$5B,$18,$8F,$E1,$71,$82,$57,$4B,$FB,$7F,$07,$E0,$9F,$8C,$FC,$7F
	.dc.b	$8C,$F9,$1F,$2F,$6F,$86,$48,$C6,$3E,$E6,$E3,$F2,$DF,$A1,$F2,$F2
	.dc.b	$F8,$F9,$56,$FE,$2C,$DB,$FA,$53,$6B,$F9,$0F,$BD,$F1,$0F,$B5,$F2
	.dc.b	$F6,$3E,$2E,$1F,$C8,$A7,$E0,$92,$3F,$23,$B2,$31,$3F,$53,$E5,$F9
	.dc.b	$36,$BD,$BF,$2F,$F8,$7F,$23,$F1,$0D,$8C,$9F,$0A,$32,$D3,$E2,$F8
	.dc.b	$FF,$8F,$3F,$21,$F0,$D6,$32,$4B,$FC,$2C,$65,$FB,$9D,$DF,$82,$3F
	.dc.b	$31,$E7,$5A,$F7,$E9,$FB,$DA,$DC,$7D,$49,$D4,$60,$9B,$FE,$76,$32
	.dc.b	$FE,$42,$ED,$7E,$6A,$6D,$7F,$4E,$9B,$5F,$C0,$27,$33,$FD,$42,$73
	.dc.b	$C5,$FE,$1E,$BF,$2C,$6B,$0F,$36,$1E,$19,$C9,$2F,$CA,$F8,$4C,$47
	.dc.b	$95,$A6,$63,$67,$95,$CB,$FA,$5C,$3C,$17,$E4,$6C,$EA,$BD,$4F,$0A
	.dc.b	$F5,$39,$C3,$D5,$1D,$8C,$9E,$09,$E2,$E7,$0F,$6D,$7F,$76,$C3,$EB
	.dc.b	$BD,$63,$33,$E7,$3F,$81,$E1,$E4,$7A,$61,$E0,$C3,$E0,$47,$5C,$39
	.dc.b	$B7,$0F,$35,$F1,$AD,$9F,$36,$74,$74,$8D,$4F,$94,$FC,$31,$E5,$AF
	.dc.b	$8C,$D3,$A9,$67,$FC,$8B,$87,$E3,$DF,$F0,$AC,$FF,$0E,$24,$FE,$B9
	.dc.b	$B1,$C1,$B7,$FC,$B7,$0F,$C3,$92,$BE,$83,$6B,$E1,$B6,$7D,$B3,$FC
	.dc.b	$BA,$30,$FA,$E7,$F9,$6C,$61,$F9,$D3,$DE,$7A,$92,$3A,$B6,$3A,$C7
	.dc.b	$F2,$D9,$E4,$FE,$02,$39,$3C,$08,$F0,$3F,$B1,$C7,$0B,$FB,$F1,$8A
	.dc.b	$EC,$79,$DE,$67,$19,$FE,$BB,$19,$3D,$C7,$0A,$F5,$B8,$7C,$2E,$DF
	.dc.b	$43,$B7,$D0,$EE,$7B,$4F,$96,$FF,$B1,$36,$7F,$9D,$B6,$BE,$5C,$73
	.dc.b	$BF,$5C,$F9,$A7,$C6,$78,$0F,$19,$8B,$F8,$0E,$0F,$19,$B7,$FB,$77
	.dc.b	$99,$FD,$CC,$FE,$E9,$E4,$DC,$3F,$A6,$C5,$AF,$ED,$33,$6B,$FB,$B7
	.dc.b	$CB,$FF,$7B,$F9,$7F,$EF,$8D,$C3,$FA,$AB,$6F,$FB,$8B,$6B,$FB,$7F
	.dc.b	$67,$F6,$E3,$FC,$0F,$F0,$8F,$8A,$C6,$1F,$CB,$A7,$FB,$59,$1F,$E9
	.dc.b	$6D,$C3,$FC,$65,$4F,$97,$E3,$18,$7A,$FC,$BC,$BF,$EE,$14,$F9,$7C
	.dc.b	$5F,$2A,$B9,$2F,$FC,$2B,$7B,$78,$CF,$A3,$E5,$FD,$3F,$1D,$67,$AB
	.dc.b	$70,$F2,$B7,$0E,$90,$58,$27,$31,$A9,$7C,$0B,$C0,$CE,$09,$93,$98
	.dc.b	$F6,$FC,$0E,$A7,$35,$5E,$6F,$FA,$0A,$B8,$47,$FB,$49,$AF,$53,$8D
	.dc.b	$7F,$92,$A3,$77,$E6,$C5,$F2,$FD,$67,$35,$F2,$23,$EE,$FD,$04,$8F
	.dc.b	$A3,$EF,$DC,$7E,$CC,$DC,$7A,$4D,$AF,$A3,$F2,$7F,$86,$FA,$0F,$C3
	.dc.b	$B6,$3E,$DB,$70,$FC,$1B,$6F,$E5,$7C,$73,$93,$E1,$36,$FF,$07,$EE
	.dc.b	$62,$64,$7F,$9A,$27,$58,$7F,$BD,$9A,$FC,$50,$E1,$F9,$0F,$0B,$F8
	.dc.b	$DE,$7F,$49,$FC,$BB,$FA,$0F,$83,$97,$F9,$4F,$9E,$B5,$F9,$C9,$B7
	.dc.b	$F3,$E6,$D7,$E5,$3E,$6B,$F5,$3F,$33,$FA,$A6,$C7,$E7,$9B,$87,$E9
	.dc.b	$9B,$7F,$8D,$F6,$5F,$27,$F8,$56,$DF,$E9,$BE,$67,$13,$23,$FD,$B2
	.dc.b	$FF,$B1,$C8,$E0,$FF,$2E,$7E,$B5,$F9,$57,$FC,$8B,$5D,$31,$FF,$54
	.dc.b	$4E,$5F,$F5,$EF,$CD,$B3,$FF,$6A,$B6,$FF,$D3,$2D,$9F,$A0,$1C,$3F
	.dc.b	$4D,$63,$C8,$93,$C9,$FF,$70,$FC,$A3,$D1,$A4,$9F,$F5,$3C,$65,$FC
	.dc.b	$17,$C9,$FF,$BA,$38,$8D,$24,$7E,$1F,$E3,$B4,$47,$07,$4F,$4F,$C2
	.dc.b	$47,$07,$FA,$DB,$E6,$9E,$EF,$FB,$D9,$B1,$98,$C1,$FE,$FA,$F9,$A7
	.dc.b	$B3,$FC,$A7,$7F,$8B,$18,$F2,$7F,$D8,$2B,$E1,$FD,$33,$AA,$E0,$FE
	.dc.b	$6F,$E8,$1E,$C7,$FB,$22,$11,$87,$FD,$96,$D7,$E9,$B5,$FD,$27,$90
	.dc.b	$F7,$79,$36,$3F,$DA,$6D,$7E,$DB,$5E,$3E,$B7,$02,$3C,$1F,$27,$DA
	.dc.b	$FF,$27,$8F,$F6,$EB,$5F,$C2,$B5,$FD,$B3,$5F,$DA,$FC,$DF,$B7,$90
	.dc.b	$D2,$47,$F0,$1F,$8D,$D1,$1C,$1F,$86,$CF,$E3,$73,$7C,$18,$FE,$50
	.dc.b	$D7,$F9,$B3,$5F,$E0,$FF,$37,$E1,$23,$FC,$3D,$F2,$7B,$D2,$3F,$89
	.dc.b	$35,$FD,$0C,$BD,$EF,$09,$23,$F3,$2F,$B7,$E6,$FF,$9D,$78,$5F,$CD
	.dc.b	$FE,$6A,$F4,$7F,$5C,$6D,$FF,$4B,$D9,$78,$8F,$E3,$AF,$37,$F9,$23
	.dc.b	$5F,$E4,$8E,$E3,$8C,$7F,$B1,$1A,$FF,$00,$EC,$E9,$93,$DF,$F0,$E5
	.dc.b	$FD,$8B,$E3,$2F,$A4,$77,$33,$0F,$DA,$BF,$40,$8F,$F0,$41,$F3,$6F
	.dc.b	$57,$AF,$CB,$3E,$36,$9E,$D9,$D5,$7F,$B6,$9A,$FF,$75,$E0,$FE,$C5
	.dc.b	$18,$CF,$FB,$A3,$63,$3C,$1F,$1F,$95,$E1,$DE,$09,$8F,$FB,$9A,$72
	.dc.b	$D9,$E0,$47,$FD,$D7,$F3,$6B,$FF,$03,$43,$E6,$97,$FD,$D6,$B5,$9F
	.dc.b	$F8,$0E,$73,$5F,$E0,$5A,$EE,$F1,$7F,$BE,$18,$8D,$C2,$B9,$BB,$3F
	.dc.b	$C1,$CE,$23,$73,$C4,$E7,$87,$F8,$1E,$39,$CE,$E7,$CD,$87,$AA,$78
	.dc.b	$D6,$E7,$8D,$FA,$5D,$E1,$63,$C3,$7C,$67,$71,$E5,$3E,$1D,$73,$77
	.dc.b	$5C,$5F,$65,$71,$7E,$24,$ED,$E2,$6F,$0F,$43,$C8,$FE,$9E,$FC,$47
	.dc.b	$FC,$00,$D7,$CE,$6B,$81,$F3,$7F,$93,$9F,$CD,$36,$33,$3B,$8F,$E5
	.dc.b	$CF,$F8,$F8,$FE,$61,$EC,$E0,$8E,$0F,$81,$8F,$E7,$75,$F7,$47,$F9
	.dc.b	$08,$CF,$0F,$F9,$3B,$5F,$BB,$19,$7E,$FD,$71,$8F,$94,$67,$5F,$E0
	.dc.b	$F5,$C5,$FF,$22,$47,$4B,$EE,$F9,$EF,$9D,$6E,$3F,$C4,$98,$7D,$1C
	.dc.b	$3B,$57,$8A,$78,$61,$DB,$0E,$77,$BC,$39,$57,$F9,$46,$31,$32,$3F
	.dc.b	$68,$7F,$98,$70,$F5,$1F,$E6,$77,$D4,$66,$BF,$CC,$11,$A9,$F0,$4F
	.dc.b	$23,$B1,$ED,$31,$F3,$7F,$BA,$8E,$C7,$F9,$8D,$C4,$68,$FF,$36,$38
	.dc.b	$9F,$F3,$C3,$67,$86,$BA,$AD,$BF,$E7,$3F,$9B,$7B,$3F,$E7,$7C,$1F
	.dc.b	$24,$F5,$75,$39,$8C,$63,$E4,$C7,$FD,$05,$87,$28,$C6,$1C,$B0,$E3
	.dc.b	$18,$BE,$37,$F0,$7E,$6D,$F0,$DF,$26,$3F,$D1,$91,$8A,$FE,$58,$D8
	.dc.b	$FF,$18,$B5,$FF,$31,$39,$C3,$FD,$33,$F3,$7F,$E2,$99,$ED,$7E,$D8
	.dc.b	$C4,$FF,$A7,$6B,$85,$FD,$B9,$CB,$FE,$91,$8C,$DF,$FA,$9E,$72,$FF
	.dc.b	$A8,$7E,$6F,$C4,$FF,$A5,$DA,$E7,$0F,$62,$8C,$B6,$3F,$D4,$9F,$37
	.dc.b	$FC,$B3,$0F,$99,$39,$6C,$66,$3F,$10,$D7,$FD,$04,$D7,$FD,$50,$D7
	.dc.b	$FD,$6C,$D5,$EE,$FF,$61,$9F,$EB,$0D,$8F,$F2,$EB,$5F,$F4,$73,$5F
	.dc.b	$F5,$A7,$CD,$BE,$2B,$DA,$DA,$FE,$C4,$D9,$FF,$33,$B5,$FF,$4C,$79
	.dc.b	$5F,$E6,$69,$FE,$D0,$D8,$FF,$36,$B5,$DF,$AD,$F2,$BE,$6F,$FC,$95
	.dc.b	$5F,$DB,$DB,$3F,$E7,$46,$BF,$EA,$56,$BF,$EC,$BF,$9B,$FC,$23,$FE
	.dc.b	$A9,$6B,$9F,$63,$EB,$C6,$23,$FD,$BD,$F4,$3F,$E0,$3B,$FE,$FC,$D8
	.dc.b	$FF,$41,$B5,$FF,$5A,$35,$FF,$6D,$45,$7F,$D2,$9D,$DF,$84,$8C,$47
	.dc.b	$FB,$FE,$3F,$0F,$F1,$FF,$E9,$58,$FF,$5F,$35,$FF,$72,$35,$FF,$79
	.dc.b	$35,$7D,$0F,$E3,$89,$C3,$F9,$27,$F2,$3E,$FF,$F8,$4F,$0F,$F0,$C3
	.dc.b	$6B,$FD,$3A,$D7,$FD,$9C,$D7,$FD,$9B,$F2,$3F,$E9,$32,$4F,$FB,$5D
	.dc.b	$FF,$4E,$EF,$23,$FD,$36,$D9,$EB,$F3,$7F,$A4,$B3,$49,$5F,$EE,$1E
	.dc.b	$9F,$F5,$01,$1C,$1F,$EA,$6F,$93,$FF,$55,$91,$C1,$FE,$33,$D1,$F2
	.dc.b	$E8,$FD,$6C,$61,$FE,$2F,$F2,$9F,$8A,$EC,$D6,$7F,$32,$EC,$62,$17
	.dc.b	$C5,$FF,$61,$B5,$FA,$F3,$F6,$89,$7D,$BD,$BE,$6F,$DA,$8F,$E6,$18
	.dc.b	$7F,$AA,$A3,$C7,$F1,$D8,$F9,$7C,$B6,$32,$7A,$23,$2D,$3D,$7F,$2F
	.dc.b	$D0,$EB,$CF,$63,$24,$CB,$FA,$DE,$8C,$7F,$8E,$B6,$FC,$F3,$F8,$EF
	.dc.b	$5E,$BF,$09,$F4,$FC,$26,$36,$45,$3F,$36,$4C,$7E,$A3,$6B,$C3,$8F
	.dc.b	$FB,$02,$7E,$AC,$FC,$4F,$6B,$E2,$FB,$3F,$1B,$E5,$ED,$7D,$86,$E1
	.dc.b	$C9,$BA,$7D,$8C,$51,$F3,$BD,$69,$7F,$80,$E7,$B8,$78,$A6,$E1,$C6
	.dc.b	$6D,$7E,$C3,$A3,$DC,$7D,$8E,$5B,$1F,$8C,$8F,$DC,$91,$FC,$93,$70
	.dc.b	$F7,$E5,$C1,$B3,$F9,$2E,$AF,$C1,$9F,$84,$F0,$6C,$64,$93,$F2,$23
	.dc.b	$2F,$CA,$F9,$6F,$C5,$FC,$BF,$E3,$1B,$5F,$9E,$6C,$EE,$3F,$43,$3F
	.dc.b	$67,$E3,$E4,$7F,$7A,$F6,$9F,$0D,$8B,$FF,$0C,$72,$AF,$F8,$B9,$2F
	.dc.b	$F6,$0D,$9F,$C6,$DB,$E5,$F9,$70,$FE,$B2,$31,$7D,$FE,$5B,$FC,$25
	.dc.b	$06,$FC,$BE,$19,$3B,$3C,$D1,$8C,$3B,$FC,$BF,$2F,$86,$E6,$DA,$F6
	.dc.b	$36,$77,$1E,$F9,$FC,$EF,$AE,$F0,$29,$E4,$F5,$ED,$7F,$76,$9D,$36
	.dc.b	$74,$D2,$5F,$F8,$FE,$76,$FC,$4A,$E6,$E6,$AE,$9F,$16,$6D,$FF,$94
	.dc.b	$E6,$D2,$9A,$7F,$93,$60,$30,$EF,$3F,$C6,$5B,$7C,$CF,$BD,$87,$4A
	.dc.b	$E1,$87,$6B,$EB,$39,$BE,$35,$D9,$B3,$AB,$F7,$BD,$EF,$96,$1E,$5B
	.dc.b	$EB,$5A,$3C,$4F,$C3,$70,$F5,$9E,$38,$78,$5B,$3A,$9E,$67,$C9,$7D
	.dc.b	$66,$70,$F3,$C6,$B0,$F9,$93,$9A,$E7,$3C,$49,$3E,$99,$E7,$18,$7C
	.dc.b	$F1,$98,$F4,$19,$3D,$11,$EA,$8F,$89,$19,$3C,$51,$C8,$FF,$45,$61
	.dc.b	$F1,$6F,$CE,$4B,$FC,$5D,$71,$8C,$57,$92,$FC,$44,$7A,$4F,$E1,$6F
	.dc.b	$A5,$FC,$99,$E0,$6A,$7C,$67,$A0,$F5,$D7,$37,$07,$FA,$B2,$7F,$CB
	.dc.b	$77,$F9,$A8,$C5,$7E,$41,$A5,$38,$9F,$78,$C9,$AA,$FE,$05,$F2,$FF
	.dc.b	$86,$24,$FC,$9C,$11,$8A,$F9,$D5,$FA,$08,$C3,$E7,$31,$5C,$63,$2F
	.dc.b	$C4,$8C,$CF,$2A,$EB,$3B,$24,$66,$B8,$BF,$16,$BC,$F3,$A7,$C1,$5C
	.dc.b	$A7,$84,$E4,$F0,$46,$67,$75,$D2,$BE,$8C,$79,$0C,$D7,$18,$F0,$3F
	.dc.b	$86,$9E,$17,$FB,$0B,$E9,$3D,$A7,$38,$7F,$2A,$6C,$7F,$3B,$6D,$FF
	.dc.b	$34,$6E,$1E,$C8,$F4,$91,$F1,$B6,$7B,$1E,$57,$D3,$7E,$27,$CE,$7E
	.dc.b	$46,$BE,$4D,$7F,$16,$F9,$7F,$E7,$A7,$F3,$EF,$5E,$E1,$FD,$3E,$2D
	.dc.b	$FF,$48,$9B,$5F,$D1,$FE,$5B,$FA,$2F,$CB,$FF,$46,$6E,$1F,$D5,$DB
	.dc.b	$7F,$D4,$1B,$5F,$E3,$9F,$86,$FE,$B1,$5F,$EA,$E6,$D7,$F5,$6F,$53
	.dc.b	$FC,$3B,$18,$AE,$1C,$BF,$EB,$A8,$DD,$7E,$75,$AF,$E8,$5A,$FF,$32
	.dc.b	$79,$5F,$EF,$BE,$5F,$AD,$7F,$5E,$7F,$AF,$FB,$52,$7F,$B8,$7C,$BD
	.dc.b	$C3,$FB,$4C,$DC,$7F,$B4,$4D,$AF,$E1,$3F,$2F,$FD,$A3,$E5,$7F,$B4
	.dc.b	$B6,$3F,$BD,$37,$0F,$EE,$6D,$BF,$F2,$E7,$A4,$F9,$7F,$F1,$EB,$6F
	.dc.b	$FC,$05,$EC,$E2,$64,$7F,$A2,$1F,$F3,$D1,$1C,$1F,$D4,$AB,$FC,$D3
	.dc.b	$F2,$FF,$D5,$23,$FD,$BB,$7D,$71,$FF,$28,$CE,$5B,$F0,$3F,$E5,$0B
	.dc.b	$4A,$30,$FF,$9C,$63,$AC,$60,$97,$E1,$3C,$53,$B6,$C7,$2F,$9E,$E7
	.dc.b	$F2,$16,$78,$61,$FD,$B6,$D3,$98,$F3,$B7,$0F,$6B,$FE,$64,$F8,$E7
	.dc.b	$8B,$4F,$3F,$C3,$5C,$7F,$DF,$B3,$A7,$F9,$36,$1F,$DE,$23,$0E,$1A
	.dc.b	$FF,$2B,$6B,$FC,$BD,$DE,$1F,$3B,$5E,$D3,$F3,$1F,$8F,$F2,$12,$FD
	.dc.b	$B5,$F3,$3D,$3C,$5F,$D2,$AB,$3F,$62,$6D,$7C,$AF,$A1,$FA,$3F,$43
	.dc.b	$F4,$9B,$1D,$36,$46,$31,$FB,$AD,$C3,$FD,$55,$E8,$7F,$80,$CF,$F0
	.dc.b	$C3,$F7,$BE,$4B,$FD,$1B,$19,$F1,$7F,$D6,$17,$FD,$36,$11,$88,$FF
	.dc.b	$0F,$C6,$31,$F9,$70,$8C,$3F,$39,$EF,$8B,$E7,$FB,$DF,$E3,$1F,$F4
	.dc.b	$EF,$CD,$B9,$FB,$60,$8C,$63,$D9,$B1,$F6,$9A,$FD,$7F,$6F,$8F,$11
	.dc.b	$A4,$AF,$3E,$6F,$82,$EE,$D8,$CC,$60,$FE,$25,$D1,$23,$DF,$E0,$7E
	.dc.b	$95,$B1,$E0,$F6,$E4,$7C,$4C,$51,$8C,$BE,$27,$8D,$E0,$25,$79,$1F
	.dc.b	$83,$F3,$2F,$80,$FF,$41,$D7,$E6,$23,$1E,$9F,$D0,$5F,$32,$F9,$3C
	.dc.b	$5F,$0B,$FF,$02,$7C,$CB,$E0,$3E,$BC,$23,$0F,$DA,$6B,$FA,$66,$BF
	.dc.b	$E0,$DB,$78,$DC,$9B,$1F,$82,$6B,$FA,$E6,$BF,$E1,$5F,$49,$EE,$F8
	.dc.b	$9B,$1F,$8A,$6B,$FB,$66,$BF,$E1,$DF,$98,$47,$A9,$B1,$F9,$26,$BF
	.dc.b	$BE,$6B,$FE,$25,$E5,$7A,$1F,$09,$B1,$F8,$48,$C1,$F1,$B9,$5E,$25
	.dc.b	$24,$FE,$09,$F9,$7C,$BF,$1A,$3F,$83,$6D,$82,$3F,$90,$F3,$FC,$04
	.dc.b	$70,$7E,$33,$C3,$F0,$11,$C1,$F8,$BF,$81,$EE,$47,$07,$E7,$3D,$1F
	.dc.b	$61,$1C,$1F,$98,$F0,$BD,$84,$70,$43,$FD,$3D,$D3,$FD,$B6,$3F,$C9
	.dc.b	$DE,$1F,$F7,$B6,$C7,$F3,$EE,$7B,$5F,$DC,$66,$AD,$91,$86,$C7,$F1
	.dc.b	$99,$DE,$AF,$80,$F9,$64,$9F,$D5,$35,$F2,$9F,$CE,$5E,$27,$8B,$1F
	.dc.b	$EB,$3F,$2F,$E2,$F9,$7F,$E3,$35,$FC,$CB,$E6,$13,$B6,$C7,$F5,$21
	.dc.b	$3C,$1E,$E6,$3C,$6E,$84,$78,$43,$E6,$3E,$0F,$27,$F6,$99,$E8,$FE
	.dc.b	$D5,$BA,$6B,$C2,$91,$A6,$BE,$0F,$98,$E9,$A7,$FB,$3B,$E5,$FF,$27
	.dc.b	$3F,$ED,$09,$CA,$8C,$BF,$98,$AF,$F4,$34,$63,$0F,$24,$ED,$FF,$45
	.dc.b	$57,$29,$E2,$D8,$FF,$47,$1C,$23,$95,$74,$8E,$38,$FF,$B9,$6F,$A1
	.dc.b	$E5,$71,$5C,$23,$D5,$3D,$2B,$94,$6F,$0C,$B4,$E6,$7A,$CC,$94,$CE
	.dc.b	$1E,$77,$85,$F7,$78,$DF,$A4,$F3,$D7,$2A,$F9,$33,$FE,$A9,$6B,$A3
	.dc.b	$FC,$0E,$E1,$FA,$1E,$2B,$E6,$3F,$B0,$CF,$F4,$B6,$C6,$67,$8B,$5F
	.dc.b	$EC,$FE,$EF,$F3,$24,$D2,$3F,$CD,$9F,$7D,$FF,$6C,$C7,$EE,$C8,$E2
	.dc.b	$7C,$68,$D6,$3F,$E1,$EA,$E3,$B3,$97,$E4,$DF,$19,$CC,$62,$39,$E1
	.dc.b	$F8,$AC,$39,$C7,$FB,$36,$B6,$FD,$43,$91,$E1,$BE,$CE,$4E,$78,$7F
	.dc.b	$0A,$8C,$DF,$AA,$72,$7F,$96,$63,$33,$E9,$8E,$98,$78,$30,$EB,$07
	.dc.b	$FC,$1E,$7F,$85,$70,$D4,$F7,$3F,$79,$1A,$7A,$61,$ED,$C3,$CF,$3D
	.dc.b	$3E,$63,$FB,$E6,$1F,$C0,$1B,$19,$8D,$DF,$FB,$C6,$72,$DD,$73,$3F
	.dc.b	$56,$72,$D3,$21,$CA,$F9,$87,$B9,$A6,$DF,$F9,$2B,$E3,$BF,$B8,$BF
	.dc.b	$C5,$E3,$35,$FC,$59,$AF,$E1,$4F,$C3,$38,$70,$D7,$35,$FE,$30,$F9
	.dc.b	$8F,$F2,$ED,$71,$6C,$FC,$FE,$57,$CD,$7A,$A7,$32,$69,$C3,$45,$70
	.dc.b	$BF,$E5,$D3,$AF,$5F,$C0,$48,$FF,$25,$C7,$F9,$C2,$FF,$3D,$F3,$0F
	.dc.b	$8A,$FF,$22,$B6,$BF,$CA,$2D,$9F,$F3,$EB,$5F,$84,$1F,$39,$27,$E9
	.dc.b	$B8,$BE,$A4,$AE,$AF,$F9,$53,$3F,$F0,$CC,$7F,$98,$1B,$7F,$E8,$C6
	.dc.b	$BE,$06,$BF,$4D,$FF,$33,$47,$1F,$17,$A1,$1E,$9F,$19,$DC,$8E,$0F
	.dc.b	$F3,$7F,$CB,$7F,$91,$5B,$E7,$7F,$9A,$23,$18,$7F,$A7,$9A,$FC,$25
	.dc.b	$F0,$DB,$25,$74,$79,$7C,$8C,$8E,$ED,$AF,$F5,$4F,$C0,$7F,$7C,$9F
	.dc.b	$F4,$0B,$63,$3E,$2F,$F9,$EA,$31,$1F,$49,$FF,$41,$4F,$0F,$67,$44
	.dc.b	$79,$F8,$7C,$48,$E0,$FF,$47,$FC,$A7,$DD,$BF,$CF,$35,$FE,$48,$D4
	.dc.b	$DD,$9F,$09,$FE,$80,$25,$7F,$A1,$1F,$F4,$8F,$73,$E4,$69,$27,$F9
	.dc.b	$CB,$FE,$88,$F5,$24,$7F,$A3,$1B,$8F,$FA,$87,$E1,$3F,$D1,$A4,$9F
	.dc.b	$F4,$8B,$FE,$90,$F8,$59,$1F,$E9,$46,$CF,$FA,$A7,$DE,$FF,$5A,$9F
	.dc.b	$EF,$2F,$91,$B5,$FE,$17,$D3,$FD,$43,$5F,$EB,$7A,$FF,$4D,$92,$7F
	.dc.b	$D3,$EF,$FB,$EB,$C0,$E3,$3F,$E2,$AE,$EF,$F7,$FC,$7F,$BF,$BE,$52
	.dc.b	$5F,$FB,$37,$E7,$ED,$EF,$E0,$95,$8C,$3F,$94,$C2,$B1,$7F,$D5,$65
	.dc.b	$3F,$C0,$6F,$97,$CF,$7F,$26,$9C,$F8,$76,$93,$63,$0D,$BF,$F7,$03
	.dc.b	$FC,$AB,$F5,$AF,$F3,$AC,$7F,$8E,$5B,$87,$9F,$E3,$35,$E5,$CA,$FF
	.dc.b	$6B,$D6,$AB,$F8,$2B,$5F,$E5,$0D,$7F,$B7,$3D,$32,$FE,$8F,$FA,$DF
	.dc.b	$F6,$BA,$F0,$5F,$AE,$31,$7F,$35,$7D,$06,$4D,$7C,$FD,$BF,$8C,$7C
	.dc.b	$79,$DB,$8A,$FE,$BD,$18,$7A,$4F,$17,$B7,$C0,$F8,$7C,$3F,$13,$6F
	.dc.b	$96,$4C,$3F,$D8,$F1,$87,$F8,$3C,$7B,$27,$85,$FF,$98,$3B,$3F,$9E
	.dc.b	$9F,$CE,$EB,$F8,$4D,$FF,$99,$E3,$3E,$37,$FB,$E1,$FE,$71,$5F,$B0
	.dc.b	$8C,$12,$7C,$1E,$17,$F6,$2A,$E8,$FC,$9F,$99,$93,$E5,$F8,$8B,$87
	.dc.b	$CB,$9B,$9F,$D2,$9B,$5F,$27,$E2,$3E,$97,$C0,$7D,$36,$C7,$BD,$B8
	.dc.b	$7D,$56,$DF,$F8,$33,$8B,$F2,$CD,$E5,$FB,$9E,$F7,$E3,$D9,$87,$E2
	.dc.b	$9F,$F0,$9E,$1F,$88,$6B,$F8,$86,$AB,$EB,$3F,$7F,$C1,$E7,$3F,$87
	.dc.b	$7F,$15,$ED,$C9,$FA,$DF,$09,$6B,$F1,$93,$72,$FC,$9C,$DA,$FC,$57
	.dc.b	$7F,$E4,$FC,$CF,$CA,$36,$3F,$00,$DC,$3F,$2C,$DB,$FF,$16,$F8,$BF
	.dc.b	$D7,$1B,$87,$E7,$7E,$67,$FC,$33,$39,$C3,$FC,$1C,$DC,$3F,$C9,$95
	.dc.b	$B9,$FF,$56,$4E,$67,$FC,$BF,$3C,$8D,$3F,$EA,$E8,$D1,$23,$37,$E2
	.dc.b	$AE,$71,$C5,$FC,$D5,$DB,$E5,$36,$BF,$6F,$36,$BC,$F3,$6B,$F5,$B3
	.dc.b	$5F,$2C,$F6,$3A,$95,$EA,$DA,$EA,$D3,$A9,$63,$AB,$5E,$A7,$23,$F8
	.dc.b	$34,$F8,$6F,$E0,$4F,$82,$75,$5F,$05,$B1,$FC,$1E,$B8,$5F,$AE,$BF
	.dc.b	$7D,$18,$9F,$14,$61,$F1,$5F,$6B,$F0,$C6,$2F,$97,$33,$0E,$55,$E0
	.dc.b	$3D,$65,$3D,$78,$79,$4F,$86,$E7,$0E,$D1,$83,$58,$7C,$69,$E9,$1E
	.dc.b	$57,$86,$1E,$2B,$F0,$CF,$09,$F0,$B5,$F0,$D7,$38,$EC,$D3,$B4,$7F
	.dc.b	$A6,$6B,$FB,$E9,$23,$FD,$66,$D8,$FF,$54,$35,$E1,$1F,$C9,$FD,$0F
	.dc.b	$8A,$47,$E6,$E3,$FE,$51,$BF,$31,$FC,$BE,$75,$5E,$E9,$D5,$79,$63
	.dc.b	$F9,$EC,$EA,$B8,$D6,$A7,$F5,$75,$E1,$3B,$D6,$DE,$55,$DD,$C9,$27
	.dc.b	$C1,$3B,$79,$9F,$D5,$5C,$3F,$28,$F1,$C7,$19,$F1,$CE,$B1,$FD,$F4
	.dc.b	$77,$AE,$B5,$B3,$C7,$7E,$98,$E9,$3B,$39,$CE,$E3,$C7,$1F,$DB,$2F
	.dc.b	$B6,$1F,$DA,$DB,$5F,$06,$BB,$92,$B9,$B6,$3C,$33,$C2,$3D,$D1,$DD
	.dc.b	$F1,$DF,$97,$0F,$5B,$70,$F0,$57,$97,$0F,$ED,$1F,$33,$FD,$F4,$FE
	.dc.b	$FD,$D1,$67,$FB,$54,$5C,$7F,$C0,$13,$6B,$FD,$9B,$E1,$3E,$27,$FC
	.dc.b	$40,$BE,$62,$C6,$1B,$8F,$F7,$89,$D3,$67,$4F,$F2,$CB,$FE,$66,$74
	.dc.b	$F9,$8B,$19,$3F,$D4,$38,$FF,$BF,$E7,$FC,$BA,$48,$CC,$35,$FF,$5C
	.dc.b	$60,$F9,$8E,$73,$FE,$47,$8C,$57,$97,$E6,$3D,$67,$B3,$E2,$EC,$64
	.dc.b	$95,$C6,$32,$F3,$F7,$7E,$1F,$CB,$FC,$46,$D7,$C1,$6C,$EE,$3E,$24
	.dc.b	$FF,$BA,$BF,$09,$5D,$3E,$6F,$C2,$5A,$FF,$15,$5A,$B7,$1F,$01,$B1
	.dc.b	$F1,$9A,$FF,$B2,$1A,$FF,$BA,$DD,$F0,$FD,$CF,$82,$ED,$3F,$15,$FB
	.dc.b	$BF,$21,$F7,$BE,$36,$E1,$F6,$DB,$5F,$2D,$B7,$F8,$67,$F0,$BF,$1B
	.dc.b	$2B,$F0,$5F,$19,$6F,$E9,$4D,$BF,$C6,$CD,$AF,$35,$FC,$53,$F1,$FE
	.dc.b	$FD,$8F,$87,$9A,$7F,$C2,$AD,$8F,$89,$18,$23,$F9,$3F,$98,$E5,$97
	.dc.b	$E0,$DF,$CD,$FB,$5F,$9C,$F6,$AD,$7E,$65,$B5,$F8,$96,$E7,$FA,$47
	.dc.b	$F4,$7E,$D4,$AF,$D0,$7B,$56,$FF,$21,$36,$FF,$57,$36,$BE,$A5,$FE
	.dc.b	$0C,$FD,$6F,$B3,$63,$DF,$1E,$E2,$3F,$81,$6E,$1F,$EA,$3F,$D7,$0F
	.dc.b	$13,$02,$7F,$17,$F6,$7F,$E5,$E2,$76,$7F,$8F,$9F,$F3,$0E,$FF,$EB
	.dc.b	$4A,$E6,$D7,$F2,$58,$FD,$03,$FC,$8B,$F1,$8F,$96,$A4,$C7,$F6,$B1
	.dc.b	$97,$F6,$FF,$4C,$E5,$FF,$26,$B6,$33,$18,$3F,$CA,$3F,$49,$FE,$B1
	.dc.b	$C7,$DA,$7B,$9C,$38,$6B,$9F,$3B,$FC,$B3,$F2,$CF,$63,$F8,$D4,$FF
	.dc.b	$0C,$8C,$78,$BF,$C2,$97,$B0,$D3,$17,$C2,$FC,$23,$F0,$9C,$38,$6B
	.dc.b	$E6,$6B,$F5,$1A,$F1,$F2,$3F,$8C,$B4,$FE,$37,$1F,$CB,$5B,$1A,$87
	.dc.b	$91,$49,$3C,$1D,$78,$F2,$39,$B5,$FE,$57,$2F,$1B,$A1,$23,$C2,$F7
	.dc.b	$F1,$24,$79,$1A,$F2,$F1,$BC,$84,$7D,$2F,$9F,$C6,$91,$EC,$6B,$E2
	.dc.b	$F1,$BD,$84,$7E,$0B,$EF,$F1,$A4,$7C,$46,$BE,$AF,$1B,$E2,$11,$F9
	.dc.b	$2F,$C7,$F1,$24,$7C,$C6,$BF,$0B,$C5,$FC,$5F,$85,$FA,$58,$2F,$03
	.dc.b	$F0,$51,$FE,$D2,$6B,$AE,$AF,$BD,$5C,$9B,$1B,$E4,$C3,$D2,$B8,$DF
	.dc.b	$78,$FC,$55,$E9,$A7,$E0,$9F,$C0,$11,$FE,$BF,$F0,$23,$AC,$F9,$FE
	.dc.b	$3A,$D9,$E9,$3B,$AF,$E9,$6D,$7F,$84,$47,$03,$71,$E3,$F9,$27,$07
	.dc.b	$6F,$8D,$7C,$35,$35,$13,$0F,$EA,$F5,$C1,$B1,$CF,$87,$DD,$F0,$3E
	.dc.b	$59,$FE,$BA,$FC,$4D,$FF,$7A,$D8,$FE,$36,$ED,$FA,$06,$33,$E6,$47
	.dc.b	$FB,$3F,$C9,$2F,$ED,$C3,$E4,$57,$FC,$2A,$B5,$BF,$F2,$89,$CD,$75
	.dc.b	$31,$5E,$4C,$3F,$88,$CE,$A5,$7C,$9E,$D5,$D7,$0F,$0D,$6E,$3D,$4D
	.dc.b	$BE,$71,$BC,$3B,$DF,$F3,$AC,$3A,$5E,$E3,$27,$AA,$F8,$BB,$C3,$BE
	.dc.b	$1C,$E3,$C6,$7A,$CA,$71,$39,$DF,$68,$E1,$7D,$E7,$71,$E8,$32,$75
	.dc.b	$7F,$91,$CF,$C1,$6D,$7D,$06,$E1,$FE,$36,$97,$23,$E4,$B0,$DA,$FF
	.dc.b	$1F,$CE,$A7,$27,$F2,$68,$FF,$1F,$1F,$CA,$3D,$1C,$11,$C1,$EE,$BF
	.dc.b	$CC,$E9,$F2,$23,$FB,$2C,$FC,$A9,$CC,$7C,$43,$27,$1D,$7F,$82,$69
	.dc.b	$F7,$E3,$2F,$E0,$67,$8B,$98,$F2,$B9,$9E,$3A,$7F,$96,$AF,$C1,$18
	.dc.b	$AF,$E9,$71,$8C,$3C,$D1,$CA,$BF,$8A,$D6,$E3,$FB,$0D,$6D,$F4,$1B
	.dc.b	$3F,$C0,$38,$79,$2B,$8D,$FF,$92,$A3,$DE,$7F,$11,$9F,$5C,$E9,$FD
	.dc.b	$69,$FC,$8A,$FF,$92,$4F,$F1,$73,$8D,$7A,$E7,$E3,$3F,$E5,$6F,$92
	.dc.b	$FE,$20,$DD,$B8,$BF,$DE,$23,$FC,$2A,$E4,$93,$E0,$8E,$2D,$38,$9F
	.dc.b	$E1,$C7,$35,$FC,$A7,$E4,$9C,$BE,$BA,$FF,$10,$C6,$2F,$FD,$0A,$D8
	.dc.b	$FF,$45,$B5,$FE,$F6,$E6,$BD,$FC,$8F,$3F,$C3,$4A,$D5,$FC,$EA,$D3
	.dc.b	$FE,$76,$F9,$27,$9D,$C1,$B1,$F5,$0F,$AB,$CE,$F1,$F5,$52,$7F,$D3
	.dc.b	$75,$C3,$1F,$F5,$0B,$5F,$F0,$8B,$5F,$F1,$1B,$BC,$FE,$BF,$D2,$BE
	.dc.b	$07,$A5,$7F,$28,$94,$E6,$FF,$D1,$BF,$27,$C5,$FF,$16,$B6,$BF,$C9
	.dc.b	$6F,$FB,$8C,$C2,$7C,$4E,$B1,$FA,$4E,$5B,$1A,$8C,$4F,$FA,$73,$E9
	.dc.b	$5F,$3B,$F3,$AB,$AC,$62,$7F,$D4,$3C,$2F,$7F,$AB,$63,$FD,$82,$DC
	.dc.b	$3F,$D8,$ED,$7F,$C6,$EB,$E0,$38,$92,$78,$BE,$8F,$A5,$7B,$B4,$93
	.dc.b	$FE,$CE,$8C,$BF,$ED,$0F,$96,$7C,$B7,$C0,$6C,$66,$30,$67,$E2,$E4
	.dc.b	$71,$6E,$5F,$ED,$EF,$A5,$FA,$8D,$E3,$F9,$71,$88,$FF,$28,$B5,$FF
	.dc.b	$31,$FC,$B3,$E5,$BE,$8B,$70,$FF,$72,$B6,$7F,$DD,$6D,$7F,$CB,$6F
	.dc.b	$F7,$C8,$DF,$8F,$C8,$8F,$FB,$BB,$E6,$5F,$2F,$F7,$9B,$19,$8C,$1F
	.dc.b	$82,$E2,$7C,$15,$24,$FF,$BE,$23,$2F,$FB,$EB,$C4,$7C,$5F,$E2,$9B
	.dc.b	$19,$8C,$1F,$8E,$F4,$DF,$05,$49,$3F,$E6,$A7,$C5,$E1,$C8,$F0,$37
	.dc.b	$6F,$F3,$D9,$F8,$DF,$83,$F4,$12,$7C,$CF,$79,$91,$FE,$76,$6C,$F2
	.dc.b	$F9,$1F,$D2,$9F,$A9,$F0,$1F,$13,$F8,$49,$FE,$5B,$3A,$F0,$3F,$4E
	.dc.b	$FE,$94,$8E,$0F,$E4,$35,$CB,$9F,$F8,$9C,$F5,$B4,$69,$7C,$8D,$31
	.dc.b	$F2,$37,$3F,$E2,$59,$7F,$97,$AB,$71,$8C,$7C,$F7,$C2,$31,$3A,$AF
	.dc.b	$E1,$4F,$E5,$AB,$FA,$2A,$F9,$1F,$17,$57,$8D,$B7,$C9,$B7,$C9,$FE
	.dc.b	$5B,$F2,$1F,$C8,$E7,$CB,$28,$C3,$E0,$F9,$1D,$F9,$12,$BE,$CC,$EE
	.dc.b	$3F,$88,$D7,$FB,$0A,$31,$3A,$8E,$79,$7A,$39,$39,$57,$B1,$F9,$5F
	.dc.b	$23,$2F,$E0,$FC,$7D,$AF,$83,$36,$FE,$0C,$DA,$FE,$E3,$AF,$F1,$39
	.dc.b	$F8,$BE,$27,$D5,$E9,$FA,$CD,$AF,$9A,$DC,$7E,$6B,$FD,$43,$A5,$8D
	.dc.b	$3E,$F7,$C8,$DB,$F8,$F3,$A7,$F5,$53,$FA,$92,$38,$3E,$DF,$C8,$F1
	.dc.b	$C3,$ED,$BF,$89,$F1,$A4,$FE,$03,$A2,$D7,$E0,$26,$E9,$F8,$09,$B5
	.dc.b	$FE,$10,$F9,$17,$C4,$7E,$61,$7C,$2D,$8C,$36,$7F,$21,$3A,$6C,$E9
	.dc.b	$F6,$D7,$C3,$3A,$7C,$2D,$8C,$9F,$C2,$63,$F8,$41,$1C,$1D,$BE,$17
	.dc.b	$86,$7F,$D8,$F0,$47,$9F,$E1,$65,$79,$3E,$16,$DE,$E6,$E9,$FE,$24
	.dc.b	$C1,$58,$BF,$85,$EC,$FC,$1F,$85,$F8,$4D,$8F,$7B,$71,$F1,$37,$0F
	.dc.b	$F6,$27,$C7,$74,$AF,$D1,$B6,$BD,$1F,$03,$FE,$14,$8C,$57,$0C,$70
	.dc.b	$7D,$19,$B7,$CA,$2D,$F2,$8C,$11,$FF,$20,$DF,$48,$FB,$95,$C2,$FB
	.dc.b	$46,$DF,$85,$19,$85,$4A,$E4,$DB,$F2,$4F,$13,$70,$9D,$B6,$76,$F9
	.dc.b	$1E,$CF,$39,$FC,$4D,$62,$38,$47,$57,$FD,$F7,$97,$FB,$0E,$31,$1A
	.dc.b	$3A,$3C,$8C,$DF,$88,$F4,$E0,$AE,$55,$C0,$91,$F1,$1F,$98,$62,$BD
	.dc.b	$B5,$CF,$0F,$5D,$6C,$93,$A8,$CB,$4F,$1B,$8A,$F8,$07,$52,$CF,$D2
	.dc.b	$9D,$9F,$05,$F4,$61,$F1,$6F,$C5,$3E,$FA,$E4,$7C,$3A,$F1,$BF,$22
	.dc.b	$B8,$5F,$D0,$9E,$75,$DA,$BC,$51,$87,$9D,$F6,$7E,$1C,$74,$7B,$4F
	.dc.b	$B5,$CB,$63,$53,$F4,$30,$F9,$F1,$8B,$FF,$26,$CF,$E8,$EF,$F1,$27
	.dc.b	$DB,$2B,$97,$B1,$88,$E2,$4B,$FE,$93,$3D,$FE,$07,$DD,$87,$FB,$5A
	.dc.b	$7E,$53,$6F,$DE,$DC,$3E,$39,$F8,$5A,$E2,$4A,$E2,$DA,$E2,$D9,$E3
	.dc.b	$5C,$EB,$E1,$1F,$CA,$E7,$4E,$2B,$F3,$71,$87,$F0,$31,$B3,$B9,$1E
	.dc.b	$ED,$AE,$F3,$C6,$7E,$99,$D2,$74,$77,$9F,$63,$88,$EE,$6C,$EA,$48
	.dc.b	$EB,$1B,$8E,$B1,$C9,$D4,$75,$AF,$81,$1D,$27,$47,$27,$66,$63,$6F
	.dc.b	$46,$DF,$D9,$AF,$03,$67,$89,$BB,$F8,$11,$9B,$EC,$7F,$3B,$F8,$6F
	.dc.b	$EB,$1F,$0D,$FD,$65,$B8,$7F,$56,$6E,$1F,$D2,$5B,$5F,$D0,$FE,$2F
	.dc.b	$FB,$01,$FD,$87,$E4,$6D,$FF,$5E,$8B,$7F,$D4,$66,$D7,$F4,$0C,$3F
	.dc.b	$9E,$1F,$DA,$7D,$7B,$87,$F4,$28,$FD,$09,$1F,$CC,$B6,$FF,$AE,$7C
	.dc.b	$8F,$06,$CF,$F6,$EF,$3F,$FB,$A7,$9D,$FD,$D5,$B1,$93,$FB,$9C,$65
	.dc.b	$A7,$F7,$6F,$37,$FB,$E1,$FD,$F7,$CE,$B1,$92,$5F,$F6,$D8,$CB,$FD
	.dc.b	$32,$BF,$AC,$1F,$E0,$5F,$90,$B5,$FD,$7A,$3F,$7A,$47,$F6,$AD,$9F
	.dc.b	$F6,$D7,$C4,$FF,$4E,$9E,$07,$F9,$65,$C4,$6B,$E9,$3F,$CD,$30,$FA
	.dc.b	$2B,$18,$6D,$71,$6D,$76,$78,$FD,$14,$9E,$DF,$11,$6B,$C7,$37,$85
	.dc.b	$5F,$D8,$96,$2B,$D4,$FF,$60,$F8,$57,$C4,$70,$C5,$85,$53,$AA,$F7
	.dc.b	$7C,$1E,$EB,$53,$9A,$E3,$5D,$6F,$FD,$E7,$39,$9F,$E4,$ED,$75,$3F
	.dc.b	$E6,$A8,$E5,$5D,$71,$7D,$17,$7A,$DD,$FD,$28,$F9,$BF,$3D,$F3,$BE
	.dc.b	$8A,$D7,$D3,$6D,$7D,$86,$E1,$D5,$FA,$BF,$2D,$2B,$ED,$7C,$F5,$BF
	.dc.b	$BF,$36,$FC,$93,$6B,$FC,$13,$F8,$47,$F9,$41,$FF,$44,$B6,$FF,$0F
	.dc.b	$F0,$7E,$97,$FC,$2B,$19,$BF,$9C,$EE,$3F,$A5,$B5,$EB,$E5,$7D,$36
	.dc.b	$DF,$FA,$42,$79,$3F,$7F,$5F,$F7,$54,$FD,$23,$BD,$7E,$4B,$E0,$BF
	.dc.b	$C7,$1C,$FF,$D6,$89,$7F,$D4,$9E,$D5,$FE,$2A,$F9,$36,$B8,$84,$62
	.dc.b	$BF,$D4,$2D,$8F,$F6,$DB,$5F,$EC,$BF,$88,$3E,$73,$93,$6B,$FD,$52
	.dc.b	$D9,$FF,$71,$B5,$FE,$DB,$F3,$3F,$EA,$B2,$7B,$9E,$37,$FD,$5D,$F3
	.dc.b	$2F,$98,$FF,$40,$D7,$FB,$B2,$31,$B7,$FA,$D9,$7C,$CF,$AA,$75,$5C
	.dc.b	$1F,$EF,$7F,$47,$FE,$CB,$7F,$D7,$AD,$7F,$DE,$ED,$77,$F0,$5F,$EC
	.dc.b	$25,$F1,$1F,$E9,$07,$FD,$8A,$D7,$FD,$FE,$D7,$FC,$12,$D7,$FC,$A6
	.dc.b	$BD,$6E,$8D,$8E,$71,$83,$C1,$D9,$F5,$A3,$F5,$5C,$58,$23,$E2,$F5
	.dc.b	$39,$91,$C1,$E6,$F4,$7B,$91,$C1,$FB,$5E,$BE,$E4,$70,$7B,$BD,$4F
	.dc.b	$39,$1C,$1F,$0B,$D3,$F5,$91,$C1,$FC,$07,$B7,$D6,$47,$07,$CA,$F4
	.dc.b	$3E,$19,$1C,$1F,$3B,$CD,$F8,$E4,$70,$7F,$0D,$EC,$F8,$E4,$70,$7D
	.dc.b	$6F,$3B,$FD,$07,$E0,$7F,$11,$ED,$BD,$9F,$AD,$6D,$7F,$A3,$B8,$D7
	.dc.b	$88,$FC,$0B,$F6,$08,$F5,$21,$FB,$53,$67,$F3,$7A,$FC,$0C,$76,$BF
	.dc.b	$E7,$3F,$13,$D7,$8F,$F9,$9D,$FF,$4F,$F1,$75,$E0,$FF,$3A,$8E,$47
	.dc.b	$FA,$8A,$76,$F7,$6B,$FC,$C9,$FC,$49,$1F,$F2,$D9,$FE,$AA,$F9,$A7
	.dc.b	$AD,$E3,$F9,$56,$F4,$F8,$13,$AA,$F0,$87,$0F,$BD,$B9,$FF,$4D,$6E
	.dc.b	$BF,$EB,$25,$C7,$F2,$FC,$AF,$65,$FC,$77,$F7,$0D,$E1,$D6,$D2,$3C
	.dc.b	$33,$C3,$1F,$EB,$4D,$C7,$FD,$7F,$F3,$5D,$A7,$FD,$8F,$F3,$3F,$D9
	.dc.b	$6F,$FD,$93,$39,$51,$97,$FB,$56,$1F,$EC,$88,$C6,$1E,$67,$55,$C5
	.dc.b	$B9,$7F,$6C,$BE,$75,$FE,$D0,$38,$16,$FA,$3E,$69,$E1,$5C,$B0,$F3
	.dc.b	$CE,$EE,$9F,$EE,$07,$9B,$EA,$AE,$31,$CE,$77,$3E,$A2,$C7,$05,$87
	.dc.b	$A6,$38,$47,$A0,$ED,$87,$99,$A7,$49,$E0,$F8,$6B,$F1,$6F,$4B,$F8
	.dc.b	$4F,$FA,$A2,$7E,$73,$63,$2F,$D6,$F6,$1F,$35,$CA,$91,$89,$FC,$BC
	.dc.b	$62,$75,$39,$9F,$0F,$C0,$FF,$35,$4D,$23,$FC,$E5,$F8,$B8,$7E,$83
	.dc.b	$4F,$8F,$23,$89,$F0,$CF,$F1,$3F,$0F,$F8,$A7,$5F,$C3,$46,$EF,$9C
	.dc.b	$E7,$5F,$D6,$C6,$2F,$AD,$74,$BF,$C7,$61,$E1,$8E,$C7,$33,$C0,$E6
	.dc.b	$3B,$C7,$1B,$FE,$33,$87,$78,$FF,$6A,$E1,$DD,$E6,$6C,$FF,$23,$46
	.dc.b	$DF,$71,$93,$DB,$87,$1C,$3F,$C8,$87,$94,$FF,$03,$D6,$63,$C2,$7F
	.dc.b	$02,$8D,$3F,$02,$7D,$35,$FE,$23,$5F,$35,$FE,$40,$9F,$0B,$63,$E8
	.dc.b	$3F,$E6,$CA,$FF,$2A,$39,$8E,$3B,$73,$8C,$1B,$AE,$71,$DF,$E6,$BF
	.dc.b	$9B,$C7,$F2,$6B,$FF,$2E,$FA,$F9,$E8,$FF,$1B,$46,$27,$F8,$DD,$70
	.dc.b	$C7,$FC,$59,$39,$7F,$9D,$46,$63,$FC,$C9,$CC,$C3,$DD,$F3,$4F,$80
	.dc.b	$FF,$15,$37,$1F,$D2,$B7,$1F,$E6,$2D,$7E,$98,$75,$87,$F9,$C7,$E6
	.dc.b	$9F,$1D,$DB,$0F,$4C,$63,$0F,$E6,$77,$C1,$FF,$42,$46,$70,$FE,$A5
	.dc.b	$39,$7F,$D0,$31,$9C,$3F,$D0,$5F,$35,$E3,$9F,$C2,$B5,$FD,$23,$53
	.dc.b	$4E,$8B,$E6,$9E,$C7,$E0,$67,$31,$8C,$7F,$A4,$B6,$BF,$A9,$B5,$F8
	.dc.b	$6D,$7F,$2E,$D7,$F6,$DB,$BE,$2B,$FC,$B0,$DA,$FE,$26,$D9,$FE,$BE
	.dc.b	$D7,$E7,$FC,$D3,$E2,$3F,$CC,$0D,$AF,$E3,$6D,$9F,$EC,$ED,$7E,$BF
	.dc.b	$D7,$FF,$B0,$9A,$7F,$9A,$5C,$3F,$AE,$6B,$FC,$37,$E6,$DF,$2D,$FE
	.dc.b	$6F,$6D,$FF,$2A,$6C,$FF,$76,$6B,$F8,$6F,$9A,$FB,$8D,$E3,$FF,$3C
	.dc.b	$46,$23,$F7,$AD,$7F,$8E,$7D,$43,$CA,$FF,$3F,$37,$0F,$E7,$2D,$9F
	.dc.b	$F0,$2B,$5F,$CA,$FC,$DF,$FA,$41,$BE,$37,$FA,$1E,$31,$1F,$C3,$1A
	.dc.b	$FF,$2B,$F9,$B7,$95,$FE,$8D,$6E,$1F,$D2,$1B,$3F,$E1,$46,$BF,$A4
	.dc.b	$FA,$47,$C9,$7F,$A5,$5B,$5F,$D4,$1B,$3F,$E1,$C6,$BF,$C9,$3E,$5F
	.dc.b	$FD,$39,$E1,$9A,$4A,$FF,$3C,$FC,$4F,$FA,$87,$CF,$34,$91,$FD,$7B
	.dc.b	$E7,$3F,$05,$87,$F2,$A6,$BF,$D2,$5A,$9B,$F2,$CF,$8C,$FF,$57,$36
	.dc.b	$BF,$B4,$B6,$BF,$C6,$AD,$7F,$D1,$3E,$1F,$94,$93,$FE,$90,$7F,$DE
	.dc.b	$3F,$25,$23,$FD,$DC,$D9,$FE,$A7,$B7,$F3,$3D,$7E,$3F,$83,$FE,$FD
	.dc.b	$F9,$77,$CB,$E6,$A6,$3F,$DB,$1F,$92,$47,$07,$F2,$D9,$7C,$BF,$0C
	.dc.b	$7F,$94,$C6,$27,$5D,$FF,$CC,$23,$F5,$98,$F8,$7C,$7C,$4C,$AD,$9D
	.dc.b	$63,$17,$FC,$D3,$4F,$F7,$1A,$B4,$3E,$5F,$F9,$EB,$A8,$CC,$63,$E1
	.dc.b	$72,$DC,$3F,$A3,$57,$17,$FC,$BB,$B7,$F1,$C3,$F9,$45,$F6,$F8,$0F
	.dc.b	$95,$94,$AF,$F0,$8B,$F8,$E2,$38,$3A,$FC,$BE,$F3,$F5,$3E,$6F,$80
	.dc.b	$93,$E2,$F8,$0B,$5E,$F9,$B8,$AB,$FC,$2C,$FC,$0A,$CC,$63,$3F,$EE
	.dc.b	$77,$F1,$57,$BF,$F1,$3D,$3F,$8A,$DA,$F8,$2D,$AF,$9C,$FF,$79,$F9
	.dc.b	$76,$14,$F9,$BE,$D5,$BF,$F4,$1C,$E9,$FF,$21,$B6,$FF,$D1,$ED,$7F
	.dc.b	$AF,$B5,$FF,$09,$BB,$CB,$DD,$1E,$9F,$63,$B3,$F8,$17,$EE,$F8,$BF
	.dc.b	$7B,$DF,$B8,$7D,$A6,$D7,$E1,$DB,$7D,$0F,$C3,$BF,$86,$F9,$79,$5F
	.dc.b	$82,$F9,$7B,$7F,$8F,$9B,$7C,$77,$56,$2F,$D3,$87,$C7,$3F,$23,$E7
	.dc.b	$D8,$F8,$B1,$EC,$C7,$E0,$B5,$E3,$1F,$22,$31,$1F,$E5,$3D,$11,$F9
	.dc.b	$CF,$97,$E9,$3F,$A3,$7F,$3F,$F2,$FF,$A0,$F0,$EE,$1F,$9C,$6D,$7E
	.dc.b	$A5,$BA,$7A,$DF,$D3,$F8,$12,$BF,$47,$E1,$5B,$FD,$84,$DB,$F8,$53
	.dc.b	$6B,$EC,$DF,$E1,$CF,$D8,$FC,$85,$8F,$C2,$C5,$C3,$F7,$53,$6F,$E2
	.dc.b	$69,$FC,$C2,$F6,$7E,$D2,$74,$F6,$33,$5C,$9B,$5C,$AF,$73,$FA,$1A
	.dc.b	$AA,$FB,$3F,$EF,$56,$D7,$DD,$9D,$C6,$2F,$C1,$5D,$CE,$17,$A3,$DF
	.dc.b	$1E,$23,$77,$E3,$7D,$46,$67,$EE,$4F,$9A,$7F,$8C,$DF,$3A,$D4,$61
	.dc.b	$E1,$5E,$9A,$E1,$BA,$F8,$4F,$92,$BB,$B4,$97,$E3,$7D,$4D,$24,$7A
	.dc.b	$0F,$E1,$53,$D2,$33,$3F,$2E,$7C,$78,$7C,$58,$F5,$C7,$82,$77,$1E
	.dc.b	$78,$F5,$DF,$9D,$E3,$87,$AB,$0E,$D1,$E1,$AE,$C7,$A1,$C3,$CC,$F4
	.dc.b	$B8,$C3,$C5,$19,$C3,$ED,$4F,$63,$C3,$5D,$A7,$CE,$72,$2C,$FC,$57
	.dc.b	$F8,$AD,$FE,$C9,$F0,$18,$AF,$E3,$B1,$9B,$FC,$5C,$E4,$FE,$3F,$19
	.dc.b	$3F,$D1,$FF,$09,$FE,$97,$8E,$F5,$FC,$EA,$31,$1E,$98,$C3,$F8,$58
	.dc.b	$CD,$7C,$FB,$46,$2F,$AD,$6C,$97,$F8,$A9,$D3,$FE,$A3,$6C,$7F,$A7
	.dc.b	$9B,$5F,$C9,$5B,$3E,$23,$94,$F4,$3E,$A1,$FD,$C6,$36,$D3,$53,$F8
	.dc.b	$83,$CE,$E2,$7E,$F4,$61,$EA,$48,$EB,$19,$6C,$F6,$8C,$1D,$5C,$12
	.dc.b	$3B,$1F,$38,$F2,$4F,$59,$E1,$1A,$3E,$81,$A3,$B9,$9B,$FA,$15,$D6
	.dc.b	$B9,$1E,$A7,$13,$C2,$32,$F8,$6B,$CF,$1B,$8E,$97,$F6,$23,$EA,$7C
	.dc.b	$BF,$E1,$BE,$5F,$F0,$ED,$8C,$9F,$66,$32,$D3,$EE,$7C,$BF,$E3,$8F
	.dc.b	$C7,$FC,$0D,$8C,$93,$0F,$C0,$46,$5F,$F1,$54,$19,$EF,$7F,$BC,$4F
	.dc.b	$F7,$B6,$5F,$90,$F7,$3F,$DF,$91,$9F,$80,$FD,$32,$69,$2F,$FD,$BB
	.dc.b	$7D,$23,$F4,$7F,$2F,$C9,$B3,$F9,$5F,$97,$FD,$3F,$CB,$FE,$A1,$B1
	.dc.b	$93,$F3,$51,$96,$9F,$9D,$F8,$DF,$D7,$9F,$B0,$F8,$DB,$19,$26,$3F
	.dc.b	$A1,$8C,$BF,$08,$F8,$5F,$2F,$FE,$4E,$4F,$F9,$2F,$8D,$F1,$DF,$E0
	.dc.b	$BB,$F9,$15,$AF,$33,$0F,$27,$F8,$55,$9A,$49,$F8,$B3,$C5,$FD,$B7
	.dc.b	$B5,$CA,$30,$7E,$DD,$FE,$17,$BF,$F0,$CE,$9B,$5F,$BD,$6D,$7E,$F5
	.dc.b	$BC,$5F,$BD,$7F,$89,$FD,$24,$AF,$E0,$9F,$2D,$6F,$F8,$24,$DB,$FE
	.dc.b	$09,$36,$BD,$FF,$21,$FE,$53,$7F,$CA,$66,$0F,$E3,$FF,$47,$E9,$8D
	.dc.b	$61,$F3,$43,$87,$E4,$98,$9F,$E6,$11,$98,$E9,$1F,$1A,$39,$DF,$F1
	.dc.b	$7F,$A3,$ED,$3F,$CB,$9F,$A3,$0F,$D7,$77,$3F,$E6,$85,$F3,$7A,$73
	.dc.b	$87,$FA,$2D,$CC,$FB,$23,$0E,$E7,$83,$4D,$57,$7F,$A4,$E8,$75,$F9
	.dc.b	$9B,$3A,$C1,$38,$C3,$F2,$17,$E7,$BC,$3F,$E4,$7C,$9F,$41,$F2,$CF
	.dc.b	$47,$CD,$5C,$BC,$D8,$7C,$C9,$F9,$24,$8F,$90,$D8,$EF,$93,$E8,$B9
	.dc.b	$46,$1F,$C8,$7C,$C3,$AB,$FB,$13,$70,$FC,$BB,$6B,$F7,$AD,$7E,$DF
	.dc.b	$CF,$7C,$A7,$F8,$0B,$5C,$EF,$FD,$B6,$31,$1F,$8C,$F9,$BF,$C0,$92
	.dc.b	$FF,$00,$FF,$73,$F9,$D6,$BF,$AB,$AF,$E1,$B1,$8E,$9F,$D5,$AE,$16
	.dc.b	$78,$BE,$17,$F9,$0F,$9D,$63,$FB,$28,$46,$1F,$DA,$B5,$FE,$36,$D7
	.dc.b	$F3,$36,$F2,$39,$36,$3F,$7A,$D7,$F9,$1B,$5F,$D0,$FC,$33,$DD,$F1
	.dc.b	$36,$3F,$82,$B5,$FE,$56,$D7,$F5,$3D,$9F,$5D,$FE,$16,$D7,$F9,$8B
	.dc.b	$5D,$F0,$FB,$F9,$5F,$43,$49,$3F,$AF,$7C,$7E,$CC,$8F,$1B,$5E,$BE
	.dc.b	$1F,$C9,$C0,$D2,$47,$EE,$7E,$23,$C0,$47,$07,$BB,$DA,$F8,$24,$70
	.dc.b	$7B,$3C,$4E,$C4,$70,$7F,$05,$FA,$0F,$61,$1C,$1F,$2B,$D6,$F9,$A4
	.dc.b	$70,$7C,$8F,$17,$D4,$47,$07,$F1,$1F,$86,$F9,$04,$70,$7D,$6F,$63
	.dc.b	$EC,$91,$C1,$0F,$E4,$5C,$28,$FE,$93,$E0,$FF,$B1,$63,$13,$FE,$37
	.dc.b	$F8,$CC,$F6,$FF,$4E,$9C,$CF,$2B,$CB,$FD,$E6,$33,$7F,$E3,$BB,$CC
	.dc.b	$B8,$23,$F8,$67,$EC,$11,$C1,$FC,$31,$DB,$FC,$E5,$B3,$FE,$EB,$F0
	.dc.b	$5F,$25,$DB,$E7,$DF,$03,$E4,$AF,$E7,$FB,$5E,$73,$B9,$99,$7F,$1A
	.dc.b	$BE,$2D,$8E,$9C,$7E,$AF,$4F,$FC,$1D,$3F,$C8,$1F,$DE,$B7,$C0,$D7
	.dc.b	$A3,$23,$4D,$7F,$A9,$7C,$FA,$DA,$7F,$88,$7E,$79,$8F,$F7,$CA,$D4
	.dc.b	$E6,$3F,$BF,$9E,$2B,$57,$07,$FC,$55,$19,$8C,$42,$B9,$BC,$27,$B5
	.dc.b	$F8,$A3,$07,$2C,$3F,$C1,$B7,$CA,$BA,$1C,$30,$F2,$57,$F8,$E7,$0E
	.dc.b	$B5,$BB,$E3,$7E,$86,$D6,$F0,$F0,$57,$9A,$F8,$DE,$EB,$A9,$F1,$2F
	.dc.b	$8B,$EB,$BE,$33,$E3,$8F,$49,$B3,$8B,$E8,$7C,$18,$74,$BF,$87,$1F
	.dc.b	$E0,$26,$BC,$CC,$57,$C5,$75,$F0,$4F,$9F,$FF,$14,$D2,$31,$3F,$E5
	.dc.b	$08,$C4,$EA,$73,$3F,$E3,$0F,$88,$FE,$80,$9A,$47,$FA,$23,$F0,$B0
	.dc.b	$FC,$EF,$3F,$C4,$9F,$F3,$44,$E7,$64,$E6,$7A,$1F,$12,$B8,$BF,$18
	.dc.b	$E2,$66,$78,$B9,$38,$EB,$F8,$A9,$E2,$FE,$32,$38,$B9,$9E,$2E,$63
	.dc.b	$8E,$9F,$C3,$A7,$2F,$F9,$D3,$0F,$46,$1E,$76,$D7,$85,$B7,$D6,$7F
	.dc.b	$C1,$F8,$7A,$8F,$3D,$7F,$98,$AB,$8C,$60,$D4,$F8,$4F,$F0,$BC,$FA
	.dc.b	$2F,$0F,$F9,$17,$E7,$FF,$D1,$53,$FE,$1C,$6C,$64,$FF,$32,$38,$AF
	.dc.b	$0B,$67,$FC,$E6,$D9,$FF,$4E,$3F,$EE,$E3,$0B,$E7,$FF,$BC,$4F,$F8
	.dc.b	$A9,$FF,$77,$73,$3E,$03,$FC,$E5,$1A,$AF,$F2,$D3,$FF,$00,$1F,$F4
	.dc.b	$0B,$5F,$F5,$53,$5E,$DE,$37,$C6,$7F,$D0,$FD,$DF,$DF,$F0,$FF,$1C
	.dc.b	$B6,$3F,$C8,$7C,$EF,$4B,$FD,$05,$39,$D3,$E7,$B6,$F9,$F8,$DF,$E6
	.dc.b	$0B,$E1,$1D,$23,$89,$E1,$C3,$FD,$7D,$5D,$DF,$F3,$8B,$5E,$73,$FE
	.dc.b	$94,$FB,$7F,$F4,$C3,$7C,$DE,$52,$9C,$BF,$EB,$E6,$BF,$E7,$C7,$CC
	.dc.b	$FF,$B3,$5A,$F1,$8F,$F4,$F4,$60,$F1,$46,$EF,$BC,$62,$3C,$11,$E2
	.dc.b	$F0,$1F,$21,$49,$87,$FA,$B2,$32,$DF,$90,$E0,$47,$06,$FE,$3F,$A1
	.dc.b	$1C,$1F,$EB,$6F,$99,$FF,$39,$B7,$D5,$F8,$71,$8C,$BF,$DB,$ED,$7F
	.dc.b	$D3,$AB,$E8,$F6,$4A,$E8,$F2,$F9,$A9,$1D,$DB,$5F,$EE,$AF,$98,$78
	.dc.b	$FF,$49,$B5,$FE,$CA,$6C,$EE,$75,$D1,$2B,$4D,$9E,$9E,$EF,$32,$4F
	.dc.b	$37,$EF,$FC,$AF,$E6,$C9,$D1,$83,$3F,$15,$2B,$FD,$B7,$28,$C5,$FF
	.dc.b	$AF,$84,$7E,$B6,$77,$D7,$D8,$8F,$FB,$93,$E5,$9F,$29,$F9,$06,$C6
	.dc.b	$63,$07,$E5,$3C,$07,$9D,$49,$3F,$CF,$DF,$CC,$FC,$5B,$D4,$A4,$9F
	.dc.b	$F7,$9C,$65,$FF,$68,$74,$FF,$B6,$CF,$F6,$47,$35,$8F,$F5,$8F,$77
	.dc.b	$E3,$B6,$F7,$35,$FF,$6E,$B5,$FF,$74,$D7,$FB,$78,$FF,$69,$73,$73
	.dc.b	$8F,$F5,$AB,$FE,$D5,$E6,$FF,$6B,$7B,$37,$0F,$F6,$53,$6B,$FD,$76
	.dc.b	$DF,$03,$FC,$68,$FF,$B7,$7D,$69,$5F,$ED,$1F,$56,$DF,$FB,$0A,$6D
	.dc.b	$FF,$8E,$E6,$D2,$A7,$D0,$3E,$57,$FA,$AD,$FF,$40,$8C,$57,$FB,$77
	.dc.b	$C4,$CB,$C6,$D3,$A7,$FC,$5F,$19,$8C,$46,$CD,$DF,$F0,$E8,$CC,$6E
	.dc.b	$30,$7F,$BE,$78,$BB,$C7,$9A,$31,$5F,$EE,$FE,$7F,$F7,$F7,$CF,$7F
	.dc.b	$BF,$DB,$19,$3F,$DE,$91,$96,$9F,$EF,$7E,$0E,$B9,$BD,$5F,$BC,$FF
	.dc.b	$98,$7C,$2C,$C6,$2B,$E4,$4E,$EB,$26,$AA,$CF,$FB,$E3,$E9,$7F,$37
	.dc.b	$9F,$C4,$3E,$83,$8B,$FF,$37,$78,$7E,$E9,$4E,$6B,$E7,$3C,$A3,$E5
	.dc.b	$35,$CF,$57,$F3,$88,$C4,$76,$8F,$49,$D6,$3A,$47,$F9,$93,$2B,$83
	.dc.b	$F5,$EF,$E7,$EF,$E0,$97,$CC,$FC,$67,$38,$FC,$27,$2F,$C4,$31,$E4
	.dc.b	$EE,$78,$36,$35,$87,$F9,$D2,$BF,$A4,$DF,$87,$C8,$ED,$C3,$DD,$97
	.dc.b	$C2,$74,$DA,$FC,$84,$79,$E7,$E5,$71,$BE,$4E,$9E,$9F,$99,$B8,$78
	.dc.b	$B2,$FC,$CC,$FE,$4C,$91,$FC,$6E,$32,$DA,$E8,$F8,$7B,$F9,$CF,$DD
	.dc.b	$C7,$E5,$47,$C7,$F4,$7E,$47,$CC,$DB,$F4,$36,$F8,$68,$8C,$61,$F4
	.dc.b	$5F,$A1,$EB,$4B,$F8,$1E,$A5,$C3,$8C,$DB,$FB,$13,$6B,$F6,$DE,$1B
	.dc.b	$E5,$7F,$1C,$D8,$F9,$5F,$33,$89,$91,$F1,$67,$58,$F9,$AB,$2F,$E6
	.dc.b	$C4,$6B,$0F,$F2,$3B,$6B,$F0,$D5,$C2,$73,$76,$72,$77,$71,$7F,$C5
	.dc.b	$6F,$93,$F8,$7B,$CB,$E0,$9D,$DF,$E6,$67,$35,$C2,$B5,$7E,$E3,$B1
	.dc.b	$1F,$54,$76,$AE,$53,$97,$CE,$D3,$CF,$3C,$2F,$B9,$FA,$5B,$B3,$B9
	.dc.b	$F4,$35,$F1,$9C,$5F,$23,$49,$5E,$07,$E8,$47,$1B,$F6,$1E,$7C,$D7
	.dc.b	$E0,$AE,$35,$E6,$BE,$44,$7F,$07,$87,$9D,$E8,$FC,$AB,$E7,$18,$3D
	.dc.b	$53,$F1,$2B,$A9,$F5,$6B,$85,$F8,$67,$B4,$FC,$03,$67,$C0,$72,$73
	.dc.b	$36,$E6,$BC,$A7,$C4,$8C,$E1,$F0,$2F,$D1,$5D,$23,$15,$B3,$8C,$7C
	.dc.b	$B3,$FA,$84,$7E,$69,$B3,$FE,$06,$24,$FF,$81,$1B,$1F,$E1,$B9,$E0
	.dc.b	$F4,$AF,$E4,$9E,$FF,$C7,$25,$FC,$6C,$51,$8A,$F8,$F5,$FC,$B6,$30
	.dc.b	$FC,$29,$DD,$71,$9C,$BD,$CF,$E5,$CE,$2B,$E8,$47,$F2,$18,$C4,$7C
	.dc.b	$93,$F9,$63,$8B,$FC,$33,$70,$FE,$74,$DA,$E4,$78,$49,$5E,$8B,$F2
	.dc.b	$B6,$78,$3F,$C6,$1B,$87,$C4,$AE,$2F,$B6,$33,$1E,$A3,$27,$B6,$7C
	.dc.b	$EE,$1F,$69,$D5,$C3,$ED,$9F,$95,$18,$C3,$DA,$DC,$3D,$AD,$3D,$B5
	.dc.b	$E9,$70,$F5,$3F,$98,$FC,$CF,$F5,$DF,$99,$FE,$BC,$D9,$FE,$9C,$D9
	.dc.b	$DC,$7F,$51,$9F,$EC,$9E,$9C,$8F,$EA,$9F,$33,$6B,$F9,$EC,$DC,$BF
	.dc.b	$B6,$4D,$AF,$F1,$57,$BC,$F9,$7F,$F0,$13,$63,$FA,$F7,$36,$26,$47
	.dc.b	$F5,$59,$D6,$3F,$E7,$16,$BF,$E2,$F6,$BD,$A7,$FB,$D7,$A3,$CA,$3F
	.dc.b	$A8,$BF,$DF,$7D,$1F,$EF,$DE,$8D,$C3,$FB,$73,$6B,$FA,$BB,$73,$FF
	.dc.b	$03,$3F,$E0,$5D,$E5,$7F,$77,$C3,$FC,$9C,$FF,$77,$8C,$DF,$F6,$07
	.dc.b	$FC,$63,$3A,$8C,$4F,$F7,$5C,$9F,$0D,$F1,$0D,$B6,$35,$1F,$ED,$88
	.dc.b	$DC,$7C,$58,$C1,$1F,$27,$C0,$71,$AF,$EF,$16,$8F,$1F,$CC,$CA,$F5
	.dc.b	$7C,$CD,$BD,$CD,$D3,$FC,$CF,$82,$B1,$7F,$E4,$68,$FA,$B3,$AF,$7F
	.dc.b	$DF,$BF,$C0,$6D,$7C,$76,$E3,$E4,$7F,$07,$EB,$C9,$F9,$9F,$33,$6B
	.dc.b	$FB,$96,$CA,$F8,$44,$AF,$C3,$36,$7D,$FF,$33,$C7,$2F,$4B,$F4,$BE
	.dc.b	$66,$5F,$D9,$F9,$9B,$7F,$1A,$6D,$FC,$09,$B5,$EF,$E1,$FB,$7F,$19
	.dc.b	$F7,$1B,$1F,$83,$6E,$1F,$41,$B6,$9F,$F4,$67,$C8,$BE,$4B,$FC,$C8
	.dc.b	$D8,$FC,$67,$C7,$62,$64,$7F,$8D,$E7,$58,$7F,$B4,$DA,$FF,$A4,$5A
	.dc.b	$F6,$9F,$C8,$FC,$67,$28,$FC,$1B,$FE,$73,$FA,$67,$7F,$96,$7F,$DB
	.dc.b	$D1,$8C,$3F,$C8,$CD,$8C,$CE,$DF,$F5,$07,$D2,$BE,$63,$11,$FE,$A9
	.dc.b	$8C,$5F,$F9,$3D,$B1,$FE,$EB,$6B,$FE,$A9,$F9,$57,$CA,$BD,$1F,$F2
	.dc.b	$B4,$61,$BF,$28,$E7,$CD,$6E,$FF,$DE,$DF,$62,$F9,$AF,$F2,$F4,$23
	.dc.b	$0F,$F9,$85,$AF,$08,$DF,$E2,$0F,$96,$A4,$9F,$F3,$44,$65,$BE,$E7
	.dc.b	$02,$38,$3B,$6F,$E2,$C3,$E3,$46,$08,$7A,$3E,$43,$44,$7C,$CF,$1F
	.dc.b	$9B,$91,$E9,$6B,$FE,$D1,$3F,$CF,$27,$F9,$F7,$C9,$FA,$6F,$FB,$6D
	.dc.b	$AF,$FB,$95,$AF,$FA,$2F,$9B,$E3,$C7,$E5,$63,$94,$60,$CF,$CD,$C8
	.dc.b	$EC,$D7,$C5,$2E,$0E,$64,$8F,$23,$DB,$79,$1E,$76,$BC,$BE,$6F,$C6
	.dc.b	$47,$D8,$F9,$B7,$91,$EF,$6B,$E2,$F8,$CF,$59,$1F,$88,$FB,$BE,$2E
	.dc.b	$47,$C7,$6B,$EA,$F8,$93,$FC,$FF,$E4,$2B,$FC,$B1,$BD,$FF,$A3,$D1
	.dc.b	$58,$EA,$FA,$0D,$B5,$3C,$96,$19,$8C,$3B,$9F,$F5,$F5,$5C,$3A,$C7
	.dc.b	$DD,$8F,$AE,$47,$FD,$79,$F6,$3F,$E3,$67,$FA,$B7,$C8,$F5,$C4,$F5
	.dc.b	$9D,$4F,$FA,$7D,$B8,$FF,$40,$9E,$06,$FD,$6F,$27,$CD,$3D,$DF,$47
	.dc.b	$C8,$3C,$4A,$6A,$26,$5F,$ED,$3A,$E0,$D8,$E7,$EE,$FB,$FC,$FF,$EA
	.dc.b	$D3,$FE,$D9,$7F,$7C,$DF,$07,$5E,$9C,$8D,$35,$FF,$5C,$7C,$D2,$FE
	.dc.b	$7C,$3E,$65,$87,$DA,$AD,$6F,$F7,$A7,$35,$F8,$0A,$F1,$BC,$5F,$C0
	.dc.b	$98,$8D,$C2,$B9,$BB,$3F,$12,$E2,$37,$3C,$4E,$78,$7E,$32,$39,$CE
	.dc.b	$E7,$CD,$87,$AA,$78,$D6,$E7,$8D,$FA,$5D,$E1,$63,$C3,$7C,$67,$71
	.dc.b	$E5,$3E,$1D,$73,$77,$5C,$5F,$65,$71,$7E,$24,$ED,$E2,$6F,$0F,$43
	.dc.b	$C8,$FE,$28,$FC,$47,$F5,$27,$4A,$FE,$30,$D3,$7E,$FB,$E6,$B3,$48
	.dc.b	$C4,$EA,$73,$18,$8E,$7F,$29,$FD,$01,$34,$8F,$F4,$47,$DD,$7F,$9B
	.dc.b	$E8,$F8,$13,$FC,$02,$73,$B7,$F6,$5D,$7E,$FC,$E4,$F9,$57,$C5,$F9
	.dc.b	$86,$63,$F0,$B1,$97,$8C,$79,$0E,$33,$D7,$5F,$E7,$57,$E0,$C3,$C7
	.dc.b	$5B,$BE,$77,$B8,$F9,$75,$B7,$C2,$6E,$FF,$8A,$CE,$E3,$F8,$CC,$EE
	.dc.b	$B9,$DF,$F1,$EC,$3F,$BE,$1E,$B3,$FA,$4C,$F5,$73,$7F,$E0,$76,$CF
	.dc.b	$13,$D3,$7E,$C7,$0F,$F2,$AF,$9A,$FF,$29,$46,$21,$18,$7F,$92,$C7
	.dc.b	$F2,$E8,$FF,$34,$4F,$F8,$3D,$D3,$75,$F1,$4E,$DB,$5D,$CA,$76,$6F
	.dc.b	$CD,$7F,$31,$9F,$C8,$B6,$7F,$B5,$35,$FC,$BB,$5F,$E8,$2E,$EB,$E1
	.dc.b	$72,$BD,$BE,$BD,$1E,$29,$FE,$5F,$1E,$2F,$93,$7C,$4D,$35,$D0,$F1
	.dc.b	$F1,$A7,$C5,$AD,$BF,$A6,$6E,$9F,$DF,$DA,$FE,$B1,$AF,$F5,$67,$78
	.dc.b	$FD,$7F,$88,$7C,$AF,$6A,$F9,$E7,$F5,$48,$CD,$FD,$5F,$9A,$E2,$FF
	.dc.b	$60,$6C,$FF,$64,$7F,$C6,$D1,$BC,$FE,$7C,$78,$A7,$51,$FE,$10,$8C
	.dc.b	$41,$A7,$45,$F4,$CF,$73,$9D,$7D,$A8,$C5,$FE,$27,$E1,$5E,$87,$F8
	.dc.b	$71,$B1,$FC,$31,$B8,$7F,$89,$5A,$FF,$12,$5F,$25,$C4,$93,$C5,$FF
	.dc.b	$13,$FD,$33,$E4,$E9,$27,$F8,$CC,$65,$FE,$37,$F3,$8F,$2B,$FC,$5E
	.dc.b	$D8,$CC,$60,$CF,$CD,$48,$E2,$DC,$BF,$C7,$9F,$4B,$F5,$1B,$C5,$FE
	.dc.b	$3A,$8C,$47,$F2,$C6,$BF,$E1,$6F,$9D,$FE,$FF,$7F,$CB,$9B,$1F,$E4
	.dc.b	$76,$BF,$CD,$1A,$F1,$F8,$5F,$F2,$3C,$7F,$AC,$67,$94,$60,$FF,$25
	.dc.b	$FD,$27,$0A,$FF,$13,$B5,$34,$C7,$CD,$BE,$37,$FC,$A6,$DB,$FE,$86
	.dc.b	$DA,$FF,$2E,$35,$FE,$92,$BE,$37,$89,$27,$8B,$FE,$5E,$F9,$FE,$15
	.dc.b	$FE,$38,$6B,$FE,$3C,$6A,$6F,$8A,$F9,$DF,$F3,$3B,$6F,$FA,$6C,$60
	.dc.b	$FF,$35,$FC,$93,$E5,$A9,$27,$FA,$EC,$65,$BE,$1F,$02,$38,$3F,$B1
	.dc.b	$FC,$DF,$E2,$F2,$FD,$A8,$71,$E0,$7F,$31,$F9,$BF,$CC,$D7,$F2,$B9
	.dc.b	$CB,$FE,$56,$79,$57,$F7,$06,$B9,$E9,$FF,$41,$C6,$23,$FC,$BE,$F2
	.dc.b	$7F,$CA,$F5,$5F,$14,$FF,$1D,$8D,$5B,$E5,$BF,$81,$D7,$F0,$C8,$C7
	.dc.b	$CD,$FF,$23,$9C,$C7,$F4,$95,$19,$F8,$7D,$37,$2F,$E9,$35,$C9,$B5
	.dc.b	$FC,$4A,$3C,$1E,$0F,$5F,$9B,$EC,$CA,$DB,$96,$9F,$D4,$7C,$5F,$F5
	.dc.b	$3D,$7F,$B6,$DB,$1F,$D8,$ED,$F3,$7F,$D2,$A7,$39,$FF,$A1,$1F,$F5
	.dc.b	$7B,$FE,$1C,$6B,$FE,$86,$6B,$BF,$73,$FD,$63,$1C,$B0,$F9,$1C,$5D
	.dc.b	$67,$E3,$3F,$1F,$D0,$F9,$1F,$37,$73,$F8,$2D,$C7,$E7,$66,$8C,$63
	.dc.b	$F3,$5F,$A1,$EB,$CB,$F9,$3E,$AD,$C3,$8C,$DB,$FA,$93,$6B,$FD,$3F
	.dc.b	$CD,$FD,$75,$FE,$F0,$DB,$F0,$FC,$DF,$F6,$E8,$C5,$AB,$8F,$A1,$FA
	.dc.b	$62,$3F,$A4,$6D,$FA,$BD,$CE,$51,$F8,$77,$F1,$1F,$1D,$F8,$9F,$9B
	.dc.b	$B7,$F8,$26,$DF,$C2,$6E,$1F,$8D,$7F,$21,$E6,$4A,$FC,$4F,$CD,$DB
	.dc.b	$F9,$33,$6F,$F2,$93,$6B,$FD,$7F,$EE,$BE,$67,$FC,$16,$D8,$FC,$9F
	.dc.b	$2E,$26,$47,$F9,$72,$75,$8F,$FB,$D5,$AF,$F9,$89,$AF,$69,$FD,$37
	.dc.b	$7F,$28,$FB,$AF,$EA,$7C,$5F,$D5,$78,$97,$0F,$D1,$B6,$BF,$06,$DC
	.dc.b	$FF,$5C,$FE,$C7,$82,$57,$EB,$30,$FF,$16,$BF,$9F,$8C,$DF,$E3,$1F
	.dc.b	$F0,$D4,$EA,$3F,$DB,$3C,$DF,$7A,$BE,$2F,$CA,$FC,$43,$7A,$58,$FF
	.dc.b	$1F,$28,$C5,$7F,$05,$B4,$E3,$2F,$C4,$4F,$FB,$C2,$38,$5F,$2B,$EA
	.dc.b	$D2,$4F,$F2,$69,$DB,$D5,$E9,$1E,$63,$75,$FE,$59,$BF,$0C,$60,$D5
	.dc.b	$73,$BE,$ED,$8D,$3E,$58,$CD,$7A,$23,$D7,$1E,$07,$53,$F3,$EF,$11
	.dc.b	$A8,$E7,$4B,$F1,$CF,$72,$47,$5A,$F6,$BE,$97,$99,$8A,$F4,$C6,$5F
	.dc.b	$3C,$F7,$39,$BF,$E7,$FA,$EF,$3D,$6F,$99,$EC,$9E,$64,$AF,$59,$F3
	.dc.b	$5E,$B7,$E1,$8C,$D7,$85,$EE,$62,$79,$47,$BE,$7B,$C7,$C2,$BF,$1B
	.dc.b	$DE,$BD,$15,$E5,$8F,$94,$78,$2F,$E7,$47,$38,$C1,$CE,$FA,$D7,$8A
	.dc.b	$74,$F9,$0E,$B3,$E4,$BF,$0C,$79,$C8,$68,$4F,$23,$C8,$F3,$AF,$85
	.dc.b	$19,$AF,$54,$7E,$52,$7F,$80,$5F,$E1,$DF,$C7,$98,$9F,$E1,$11,$98
	.dc.b	$F9,$E6,$4F,$E1,$4E,$5F,$25,$7E,$27,$DA,$F8,$78,$7F,$A9,$AF,$D9
	.dc.b	$18,$23,$F3,$DB,$5F,$16,$78,$3F,$06,$BA,$56,$C9,$1B,$AE,$33,$F0
	.dc.b	$8F,$D6,$38,$9F,$09,$C2,$34,$FE,$46,$35,$3D,$EB,$C2,$FB,$1B,$87
	.dc.b	$CC,$6E,$1F,$C5,$2B,$8D,$6D,$B5,$B6,$CE,$EF,$C4,$79,$2B,$D4,$E1
	.dc.b	$F1,$4F,$08,$FE,$4C,$D8,$FE,$30,$DC,$3D,$F3,$C6,$7C,$4D,$BF,$13
	.dc.b	$6F,$C5,$3E,$02,$57,$F1,$3F,$9B,$FE,$5E,$7F,$30,$F9,$BB,$7F,$CD
	.dc.b	$E2,$D7,$F2,$99,$B5,$FC,$6F,$7F,$E6,$FE,$FF,$F3,$86,$E1,$FC,$FD
	.dc.b	$B7,$FC,$C5,$B5,$FC,$37,$CC,$FE,$BA,$7F,$40,$F9,$BB,$7F,$CE,$A3
	.dc.b	$F2,$1A,$7F,$A0,$F0,$FE,$B9,$18,$24,$FF,$4D,$F6,$38,$CF,$F3,$F7
	.dc.b	$FA,$5F,$CD,$C9,$FE,$A5,$F3,$76,$BF,$A2,$4D,$CB,$F9,$B4,$DA,$FE
	.dc.b	$31,$BF,$F5,$4F,$9B,$FE,$AA,$D8,$FE,$B8,$DC,$3F,$A7,$36,$FF,$9B
	.dc.b	$7C,$D3,$E4,$BF,$A7,$C6,$1F,$EB,$FF,$37,$8C,$18,$4B,$E1,$18,$BE
	.dc.b	$0E,$A6,$09,$DB,$83,$17,$FB,$98,$CC,$70,$77,$5F,$D5,$7E,$2B,$C5
	.dc.b	$3E,$1A,$FE,$E7,$1F,$DD,$36,$98,$7F,$7B,$F9,$8B,$7F,$DB,$66,$E1
	.dc.b	$FD,$7E,$6D,$7F,$41,$F9,$7F,$EF,$BF,$09,$FD,$F9,$B1,$FE,$05,$6E
	.dc.b	$1F,$DD,$DB,$7F,$D7,$FB,$BF,$DD,$AD,$F8,$1F,$F0,$3F,$CB,$7A,$2D
	.dc.b	$5A,$C3,$E3,$3F,$EF,$28,$B5,$FC,$6E,$31,$1F,$41,$AF,$28,$F8,$76
	.dc.b	$FD,$7F,$FC,$9D,$7F,$4C,$E6,$7F,$A2,$1C,$D7,$FA,$4D,$F9,$B1,$F6
	.dc.b	$5B,$1F,$CF,$17,$CB,$F6,$73,$5D,$F7,$C1,$99,$B7,$DF,$E5,$F8,$7C
	.dc.b	$BF,$16,$C6,$4F,$24,$65,$A7,$A7,$1F,$C7,$FC,$57,$FB,$FF,$1F,$F0
	.dc.b	$63,$67,$D1,$F2,$FF,$36,$31,$5C,$3E,$0F,$FA,$24,$1A,$4A,$FC,$94
	.dc.b	$A7,$8D,$FC,$BF,$97,$E9,$7F,$05,$FB,$5F,$36,$E2,$FE,$A0,$DC,$79
	.dc.b	$B7,$1F,$CF,$35,$FD,$1F,$CD,$B4,$7B,$9E,$28,$C3,$76,$FF,$21,$D7
	.dc.b	$E9,$E7,$2F,$EA,$7E,$87,$D8,$FA,$9A,$E7,$4F,$6A,$8C,$BF,$B4,$F9
	.dc.b	$C6,$BA,$6C,$7C,$46,$BF,$B7,$6B,$C7,$93,$81,$1E,$0F,$83,$F1,$17
	.dc.b	$AD,$49,$3F,$42,$32,$FD,$2F,$9D,$75,$7A,$1B,$19,$8C,$19,$EA,$91
	.dc.b	$C5,$AF,$F0,$DF,$9A,$7B,$7F,$09,$B1,$F8,$06,$BB,$8F,$8F,$D9,$FC
	.dc.b	$8E,$3F,$0A,$D7,$F8,$C3,$5F,$E3,$AD,$7F,$77,$D5,$C8,$8F,$27,$C5
	.dc.b	$ED,$C8,$F1,$35,$EB,$CB,$E2,$23,$E2,$7C,$BD,$72,$3C,$AD,$7E,$07
	.dc.b	$B7,$EF,$23,$EF,$7D,$BF,$3B,$23,$DA,$D7,$DB,$F1,$DF,$08,$8F,$C2
	.dc.b	$7E,$57,$0C,$8F,$94,$D7,$E2,$FC,$E7,$F4,$1E,$EF,$F4,$67,$CE,$DE
	.dc.b	$EF,$F3,$73,$6B,$F8,$E5,$57,$F9,$2B,$F4,$88,$FD,$26,$BF,$E5,$09
	.dc.b	$FE,$53,$1C,$77,$3D,$6F,$F9,$26,$5D,$78,$5E,$EE,$7D,$57,$CB,$D3
	.dc.b	$51,$33,$FC,$33,$F8,$92,$3F,$C2,$3E,$77,$F1,$38,$FF,$68,$F1,$1E
	.dc.b	$4F,$85,$B5,$FC,$6C,$57,$0D,$BF,$B3,$36,$3F,$B6,$B5,$DF,$3F,$C5
	.dc.b	$87,$1F,$C7,$F5,$1E,$BD,$F4,$E6,$1D,$E7,$2D,$9E,$04,$7C,$9F,$3B
	.dc.b	$D3,$5F,$EF,$BF,$38,$BF,$EF,$53,$9A,$4E,$5F,$F0,$1C,$F6,$72,$FF
	.dc.b	$80,$4C,$46,$E1,$3C,$E3,$0D,$9E,$17,$FE,$09,$9E,$11,$C8,$F1,$E3
	.dc.b	$FE,$90,$BE,$E7,$F8,$52,$F9,$3E,$4B,$E5,$5D,$23,$DA,$F2,$C2,$C7
	.dc.b	$1B,$E9,$87,$82,$F8,$CF,$33,$67,$AA,$77,$7C,$E3,$C6,$7C,$99,$E2
	.dc.b	$D3,$8E,$1E,$67,$96,$1F,$D4,$A3,$E7,$47,$F8,$A1,$AF,$53,$1C,$EF
	.dc.b	$9D,$E0,$D8,$FF,$16,$4F,$F3,$09,$C9,$FD,$0A,$1F,$1F,$48,$FF,$46
	.dc.b	$6E,$7F,$5F,$0F,$E5,$3B,$FC,$0A,$FF,$20,$D7,$C2,$9C,$C7,$C0,$32
	.dc.b	$7C,$27,$35,$F8,$18,$CB,$D7,$75,$F5,$9F,$29,$C6,$32,$F5,$33,$3C
	.dc.b	$67,$85,$F8,$EB,$73,$DE,$B7,$A7,$F1,$39,$CB,$FE,$51,$AE,$17,$CA
	.dc.b	$BC,$0E,$1F,$E1,$C7,$F9,$76,$BF,$C7,$4F,$AA,$BA,$4F,$EF,$70,$F8
	.dc.b	$67,$A2,$7F,$DC,$B5,$C1,$F1,$9C,$E7,$F8,$27,$CE,$FF,$AF,$89,$1F
	.dc.b	$E4,$77,$1A,$D3,$11,$67,$F8,$67,$CE,$B9,$BF,$C8,$7C,$CF,$7F,$FC
	.dc.b	$A1,$39,$AF,$F1,$66,$7F,$C8,$27,$37,$C1,$B3,$AB,$E4,$FF,$19,$F9
	.dc.b	$D7,$C6,$7F,$BB,$DB,$1F,$E2,$A6,$E5,$FE,$7D,$6B,$FE,$82,$73,$8F
	.dc.b	$FA,$3B,$E2,$FF,$C6,$F2,$9C,$D7,$FA,$2E,$73,$18,$95,$27,$37,$FE
	.dc.b	$62,$79,$FC,$EF,$82,$BF,$D2,$4D,$7F,$D3,$2D,$4D,$3B,$2F,$9D,$7B
	.dc.b	$FF,$E7,$99,$CC,$63,$2F,$E7,$8D,$AF,$F4,$E3,$5F,$F5,$0B,$5F,$F5
	.dc.b	$23,$5F,$F5,$5F,$C4,$3E,$77,$F1,$11,$FE,$A9,$8C,$57,$F9,$59,$B1
	.dc.b	$FE,$A9,$6B,$FE,$AE,$F9,$D7,$CA,$72,$8F,$F5,$BC,$62,$BF,$CB,$ED
	.dc.b	$8F,$F5,$93,$5F,$F5,$BF,$D4,$3B,$B9,$42,$31,$5F,$E6,$86,$C7,$FA
	.dc.b	$ED,$AF,$FB,$03,$E7,$9F,$3D,$C9,$B5,$FE,$70,$6C,$FF,$B0,$DA,$FF
	.dc.b	$B2,$3E,$7B,$FC,$F1,$3F,$E7,$66,$C7,$FB,$25,$AE,$FD,$8F,$3F,$D4
	.dc.b	$7F,$7E,$AF,$F3,$EB,$67,$FD,$A0,$D7,$FD,$AA,$D7,$FD,$B1,$F3,$FF
	.dc.b	$EF,$B7,$FD,$B2,$D7,$3F,$07,$F1,$23,$11,$FE,$E4,$F9,$FF,$F0,$5D
	.dc.b	$FF,$A3,$1B,$1F,$ED,$F6,$BF,$EE,$46,$BF,$EE,$7F,$A8,$7D,$07,$D0
	.dc.b	$6D,$7F,$A5,$9B,$3F,$EE,$96,$BF,$EE,$E9,$7C,$97,$D8,$6D,$7F,$A7
	.dc.b	$9B,$3F,$EF,$16,$BF,$EF,$6F,$A0,$7C,$A7,$E0,$1B,$5F,$EA,$66,$CF
	.dc.b	$FB,$E5,$AF,$FB,$FB,$BB,$E1,$BF,$E2,$CD,$3F,$13,$18,$33,$F3,$0F
	.dc.b	$9A,$FC,$6B,$70,$FF,$59,$B6,$C4,$2B,$97,$01,$FE,$3C,$E0,$C7,$85
	.dc.b	$D8,$93,$FE,$C0,$FA,$6F,$F5,$8B,$7E,$1F,$F3,$F1,$8A,$E4,$D7,$BD
	.dc.b	$FF,$24,$7F,$A2,$9E,$1F,$13,$FD,$E7,$F2,$EF,$9B,$FF,$80,$15,$BF
	.dc.b	$17,$FA,$23,$C6,$F9,$DF,$C2,$26,$5F,$09,$E7,$EA,$F2,$6C,$7F,$3A
	.dc.b	$C5,$F3,$7E,$0F,$9B,$F0,$B6,$74,$DC,$3F,$AD,$E4,$BE,$35,$CE,$FD
	.dc.b	$3E,$14,$8E,$DF,$37,$70,$E3,$6A,$C6,$1F,$DD,$8E,$95,$8B,$FB,$1E
	.dc.b	$0B,$E6,$75,$18,$7E,$3F,$BB,$8C,$BE,$FD,$6E,$BF,$CD,$0D,$8F,$B8
	.dc.b	$D7,$F0,$2D,$7A,$47,$FB,$F2,$38,$61,$E5,$CB,$F0,$AF,$F7,$58,$CC
	.dc.b	$FE,$3A,$3E,$27,$1F,$DA,$E8,$FB,$7D,$56,$BE,$83,$71,$F8,$6D,$BE
	.dc.b	$CF,$E0,$B8,$25,$7F,$70,$C3,$F3,$6F,$D9,$8C,$DF,$CC,$7E,$15,$FF
	.dc.b	$7B,$8C,$BE,$7F,$95,$F8,$D1,$FC,$D5,$DD,$FF,$32,$9D,$5F,$C8,$9F
	.dc.b	$E7,$39,$1A,$48,$F8,$CF,$2F,$3F,$AC,$EE,$7D,$11,$DB,$E6,$E4,$7A
	.dc.b	$FE,$6E,$E5,$F8,$CC,$95,$8C,$BF,$09,$0A,$C5,$FC,$F8,$FA,$37,$FD
	.dc.b	$F1,$FF,$48,$47,$C7,$E0,$F6,$FA,$9E,$E6,$C6,$4F,$99,$1E,$8E,$D9
	.dc.b	$1F,$3F,$E6,$EE,$1F,$B6,$D7,$F8,$A9,$2B,$E2,$92,$BE,$B7,$9F,$C3
	.dc.b	$4F,$4B,$F3,$BE,$6E,$4F,$DB,$F9,$BB,$8F,$C1,$9B,$7F,$02,$6D,$7B
	.dc.b	$F8,$7E,$BF,$CD,$FD,$86,$C7,$E1,$5B,$87,$CB,$6D,$FF,$92,$FE,$69
	.dc.b	$F3,$1F,$82,$6D,$FE,$37,$E6,$F1,$32,$3F,$DA,$13,$AC,$3E,$B5,$65
	.dc.b	$FE,$33,$19,$BE,$D7,$F8,$CA,$D4,$62,$F9,$CE,$5A,$4B,$E7,$19,$6C
	.dc.b	$66,$79,$47,$8A,$30,$7F,$3D,$BF,$13,$6B,$95,$70,$24,$64,$F8,$07
	.dc.b	$2A,$FE,$8F,$5C,$9B,$5C,$A7,$84,$79,$09,$5F,$96,$BF,$7C,$EA,$3B
	.dc.b	$CF,$E5,$67,$2F,$49,$F0,$4F,$F9,$C6,$73,$2B,$F1,$36,$FC,$4F,$13
	.dc.b	$C3,$7D,$CF,$D2,$57,$2A,$E2,$7A,$EF,$B4,$FB,$EB,$A3,$E2,$E2,$5F
	.dc.b	$1C,$3C,$F7,$C2,$FB,$9E,$17,$DA,$64,$F7,$BD,$0D,$B9,$BF,$34,$71
	.dc.b	$6C,$66,$30,$43,$93,$F3,$63,$D1,$87,$B2,$BE,$1D,$6A,$7A,$B4,$FA
	.dc.b	$0F,$08,$F5,$E1,$EE,$8C,$1F,$DF,$E7,$D0,$7F,$81,$2B,$F4,$0D,$AF
	.dc.b	$F2,$DC,$62,$3F,$06,$66,$79,$46,$63,$D7,$7F,$EA,$DF,$84,$F0,$92
	.dc.b	$BE,$83,$67,$E3,$E0,$8C,$57,$CE,$8C,$57,$02,$57,$0A,$DC,$7C,$C9
	.dc.b	$DC,$E6,$37,$5C,$EB,$73,$E6,$AD,$CE,$6B,$67,$3A,$DD,$7C,$D9,$F6
	.dc.b	$4E,$A3,$94,$EA,$FF,$21,$7D,$2B,$91,$BA,$E5,$19,$78,$C6,$67,$8D
	.dc.b	$73,$C3,$FA,$BB,$67,$FB,$23,$6B,$FA,$53,$70,$FE,$92,$D9,$FE,$90
	.dc.b	$D9,$FE,$90,$D9,$F5,$C7,$C7,$3F,$AA,$6F,$FD,$98,$FE,$CF,$F3,$76
	.dc.b	$FF,$BA,$45,$9F,$EB,$53,$6B,$FA,$CF,$CD,$7F,$6E,$F9,$1F,$ED,$ED
	.dc.b	$C3,$FB,$E3,$6F,$FB,$13,$69,$3F,$E5,$9E,$E7,$CA,$7F,$7A,$6C,$7F
	.dc.b	$81,$7E,$4B,$13,$23,$F3,$93,$AC,$3F,$CB,$8D,$7F,$CC,$2D,$7B,$4F
	.dc.b	$F8,$2B,$E6,$FC,$35,$FE,$1D,$32,$59,$FF,$75,$C6,$9F,$F2,$2B,$FC
	.dc.b	$3C,$9F,$27,$CF,$79,$5C,$72,$56,$A3,$0F,$1F,$9A,$F1,$7C,$9F,$8D
	.dc.b	$B3,$A6,$E3,$E0,$6D,$FC,$88,$FF,$40,$FC,$6F,$FB,$02,$31,$08,$C5
	.dc.b	$7F,$8E,$70,$7C,$9F,$F8,$9A,$73,$5C,$BD,$8F,$F6,$DC,$7F,$B2,$23
	.dc.b	$FD,$18,$D7,$57,$FC,$2E,$38,$E1,$EE,$F9,$8E,$97,$F3,$5F,$9D,$F2
	.dc.b	$92,$FE,$0F,$BB,$71,$F9,$73,$70,$FF,$33,$67,$FE,$E0,$6C,$7D,$7C
	.dc.b	$7E,$C7,$C9,$DC,$7E,$5B,$70,$FA,$AD,$BE,$6F,$F9,$13,$E4,$9C,$74
	.dc.b	$FC,$17,$D2,$DC,$3F,$01,$82,$B7,$1F,$CA,$5B,$1F,$EA,$96,$BF,$EA
	.dc.b	$E6,$BF,$EB,$37,$78,$7B,$3E,$27,$B4,$FE,$29,$FC,$97,$D1,$4B,$FC
	.dc.b	$2F,$B9,$6B,$F1,$F3,$6F,$E2,$CD,$AF,$B9,$F4,$BF,$9B,$FA,$5F,$CE
	.dc.b	$36,$3F,$1E,$DC,$3F,$32,$DB,$FC,$3F,$D9,$BC,$FF,$F3,$4B,$6F,$F2
	.dc.b	$FD,$78,$99,$1F,$D2,$E7,$58,$7F,$B2,$DA,$FF,$B4,$1A,$FF,$B5,$5D
	.dc.b	$DF,$DE,$F8,$9E,$73,$FA,$C7,$FC,$F3,$F3,$EE,$B7,$BD,$FD,$66,$30
	.dc.b	$DE,$C7,$9B,$9A,$DE,$1F,$ED,$DF,$9F,$74,$7F,$61,$84,$61,$FE,$C8
	.dc.b	$D7,$84,$6E,$FF,$D1,$D1,$FE,$E7,$6B,$FE,$EC,$6B,$FE,$F1,$6B,$FE
	.dc.b	$F4,$F9,$FF,$E3,$6F,$F7,$16,$BF,$EF,$16,$BF,$EF,$66,$BF,$EF,$BF
	.dc.b	$9F,$7C,$CF,$A5,$B1,$FD,$ED,$AF,$FB,$ED,$AF,$FC,$00,$FC,$11,$E3
	.dc.b	$7C,$16,$C7,$F8,$0D,$A8,$E1,$1F,$1F,$E8,$1E,$3D,$24,$FF,$82,$E3
	.dc.b	$2F,$E9,$7E,$5F,$E7,$76,$9A,$48,$FC,$47,$AF,$F4,$F9,$BF,$DC,$11
	.dc.b	$82,$1F,$CE,$BC,$9F,$B1,$E0,$9A,$48,$FD,$A7,$D0,$78,$08,$E0,$FC
	.dc.b	$B7,$B1,$DC,$8E,$0F,$E9,$3E,$57,$62,$38,$3F,$80,$7C,$EF,$B0,$8E
	.dc.b	$0F,$D1,$7B,$3E,$B2,$38,$3F,$AA,$F9,$1E,$A2,$38,$3F,$86,$72,$7C
	.dc.b	$82,$38,$3F,$5D,$EC,$FC,$72,$38,$3F,$B0,$F9,$3F,$18,$8E,$08,$3C
	.dc.b	$0F,$E8,$BB,$27,$F0,$13,$9F,$07,$FB,$94,$62,$BF,$99,$FC,$FD,$9C
	.dc.b	$D7,$F2,$6B,$E4,$DB,$D3,$FD,$A2,$AB,$FC,$AF,$E1,$7F,$0A,$47,$F9
	.dc.b	$63,$5F,$11,$F9,$3E,$54,$72,$6C,$7F,$6A,$F9,$4E,$BF,$3F,$F8,$F8
	.dc.b	$FE,$E1,$E8,$FF,$93,$5B,$EB,$6B,$A6,$47,$F3,$EA,$E0,$F8,$03,$8D
	.dc.b	$BF,$B5,$AE,$3F,$5F,$2F,$F2,$7A,$FE,$8E,$FF,$12,$6F,$89,$AF,$22
	.dc.b	$46,$9A,$F7,$F9,$FE,$98,$FE,$C7,$E7,$98,$7D,$49,$CD,$27,$2F,$ED
	.dc.b	$E7,$C4,$E5,$FB,$66,$23,$70,$9E,$71,$86,$CF,$0B,$FC,$1C,$F0,$8E
	.dc.b	$47,$8F,$1F,$C5,$5F,$73,$F8,$45,$F2,$7C,$97,$CA,$BA,$47,$B5,$E5
	.dc.b	$85,$8E,$37,$D3,$0F,$05,$F1,$9E,$66,$CF,$54,$EE,$F9,$C7,$8C,$F9
	.dc.b	$33,$C5,$A7,$1C,$3C,$CF,$2C,$3F,$D4,$91,$F3,$A3,$F4,$66,$4A,$53
	.dc.b	$2F,$F3,$4F,$9F,$E3,$CE,$69,$27,$FA,$87,$CF,$FF,$51,$4D,$23,$A3
	.dc.b	$D5,$7F,$CA,$79,$FE,$04,$7F,$35,$9F,$85,$39,$8F,$80,$64,$E3,$BF
	.dc.b	$F9,$17,$5F,$C0,$C6,$5F,$C1,$CF,$17,$31,$E4,$38,$DF,$81,$C4,$70
	.dc.b	$AE,$9A,$7C,$4C,$3C,$71,$8A,$F9,$31,$8B,$E4,$D8,$F3,$9D,$2B,$F8
	.dc.b	$75,$71,$8F,$E2,$CF,$F9,$77,$0F,$F2,$D9,$EF,$3F,$CB,$B1,$EB,$9E
	.dc.b	$37,$FE,$38,$9D,$38,$9F,$DE,$1B,$3F,$AF,$46,$9F,$EC,$1F,$3F,$FE
	.dc.b	$14,$9F,$E4,$6D,$8C,$9E,$38,$CE,$B6,$38,$B5,$4F,$16,$99,$70,$3E
	.dc.b	$07,$07,$3B,$7D,$33,$DC,$CE,$3F,$E5,$4B,$E0,$DB,$E1,$5B,$8C,$1B
	.dc.b	$FA,$1F,$01,$D7,$99,$F0,$78,$9D,$13,$3F,$E6,$39,$FF,$A2,$67,$51
	.dc.b	$FC,$97,$26,$18,$AF,$AD,$87,$4C,$39,$47,$D9,$F9,$F7,$AD,$F1,$7E
	.dc.b	$A1,$ED,$FB,$E7,$37,$FE,$70,$9C,$B6,$78,$1B,$AB,$19,$3F,$CE,$FE
	.dc.b	$A3,$D6,$CE,$BF,$CE,$E3,$13,$FE,$7C,$F8,$E7,$7F,$D7,$6C,$7F,$AD
	.dc.b	$FC,$D7,$D0,$E6,$BE,$E7,$DC,$BD,$CA,$66,$FF,$D0,$F1,$97,$FD,$13
	.dc.b	$F7,$2F,$92,$FC,$23,$63,$75,$89,$FC,$47,$CF,$FF,$AC,$B5,$3F,$EA
	.dc.b	$CC,$FF,$A8,$5F,$FA,$52,$35,$C2,$F7,$FE,$61,$F8,$9A,$7C,$7E,$2F
	.dc.b	$FD,$37,$F6,$DF,$EE,$5E,$6F,$F6,$2B,$63,$F4,$7E,$73,$E3,$7F,$35
	.dc.b	$39,$D3,$FC,$89,$F3,$DF,$8D,$7F,$B3,$37,$0F,$DB,$7D,$B7,$F9,$CB
	.dc.b	$1F,$F5,$4C,$E5,$FF,$06,$F6,$3E,$33,$F4,$CD,$8F,$F7,$35,$3E,$3F
	.dc.b	$35,$FA,$BF,$B6,$7C,$9D,$3F,$4D,$C2,$F8,$3C,$63,$FE,$B8,$E6,$F8
	.dc.b	$2D,$E2,$FD,$67,$D9,$FE,$62,$30,$7E,$3F,$AB,$F2,$87,$EC,$7E,$C3
	.dc.b	$FD,$CB,$19,$3F,$D8,$7E,$63,$E2,$FF,$7C,$D8,$FF,$7F,$4F,$E6,$B9
	.dc.b	$3F,$C1,$53,$97,$FC,$19,$E2,$E7,$97,$FD,$EA,$DB,$FE,$27,$F5,$CF
	.dc.b	$67,$FD,$E9,$E2,$FF,$0B,$F2,$BF,$2B,$8F,$F8,$55,$B8,$FF,$25,$BF
	.dc.b	$CE,$CE,$75,$FE,$25,$F3,$9F,$0A,$3F,$C3,$8D,$AF,$E5,$9F,$34,$F9
	.dc.b	$DD,$F1,$7F,$19,$FC,$43,$5B,$7F,$91,$1B,$8F,$F9,$27,$4F,$F4,$1C
	.dc.b	$FF,$B9,$A7,$32,$69,$C3,$1F,$E3,$7F,$0F,$D7,$09,$7C,$5F,$E3,$31
	.dc.b	$8B,$FE,$23,$FB,$07,$FB,$02,$BF,$9B,$9F,$CE,$17,$D8,$7F,$27,$75
	.dc.b	$3C,$23,$1E,$66,$5B,$5C,$A7,$FC,$C5,$5C,$A3,$18,$BE,$91,$F3,$1F
	.dc.b	$CC,$DB,$3A,$7F,$D2,$4D,$8D,$11,$D7,$C4,$DC,$3F,$C9,$9F,$8E,$FE
	.dc.b	$76,$7C,$50,$8C,$3F,$03,$CF,$FE,$95,$18,$C5,$5F,$E7,$68,$C3,$7C
	.dc.b	$3F,$85,$5C,$19,$1D,$A7,$77,$F0,$2B,$2F,$F3,$EC,$BF,$55,$5E,$DF
	.dc.b	$93,$F8,$1E,$07,$C1,$6C,$64,$F8,$51,$97,$E4,$EA,$FB,$1F,$A8,$7D
	.dc.b	$5F,$B1,$B3,$A8,$B8,$FD,$E9,$BA,$FB,$A3,$FD,$A5,$9F,$EC,$9F,$F3
	.dc.b	$A3,$63,$FC,$09,$9D,$C3,$FD,$33,$36,$B8,$FC,$67,$F7,$38,$C5,$FD
	.dc.b	$D8,$C1,$2B,$E9,$CE,$AF,$95,$FF,$42,$AF,$C7,$7D,$37,$E3,$FD,$FB
	.dc.b	$87,$C0,$6D,$7E,$61,$B8,$FF,$86,$74,$FF,$07,$3F,$8D,$EF,$FC,$61
	.dc.b	$F8,$DF,$8C,$B8,$7E,$36,$2E,$1F,$56,$6D,$7C,$58,$FE,$AB,$EC,$7B
	.dc.b	$4F,$D1,$7C,$75,$C3,$D7,$16,$7F,$57,$36,$B9,$6D,$FC,$21,$B5,$CA
	.dc.b	$F8,$9F,$9F,$F9,$BF,$D1,$7C,$B7,$E8,$DB,$19,$3F,$49,$19,$69,$F8
	.dc.b	$4F,$88,$FD,$B1,$FB,$7F,$87,$B1,$92,$67,$FC,$02,$32,$FF,$AC,$FE
	.dc.b	$C9,$F4,$DF,$E2,$D6,$D7,$EC,$FE,$43,$13,$23,$F5,$B3,$AA,$FC,$F5
	.dc.b	$65,$E9,$57,$4F,$C5,$CD,$BF,$F7,$1D,$F1,$9F,$A0,$D8,$FF,$24,$F2
	.dc.b	$FF,$BA,$5E,$F7,$C7,$00,$AE,$0F,$FB,$C6,$B8,$3E,$2A,$E2,$47,$CE
	.dc.b	$DB,$F0,$DF,$57,$F8,$B6,$1D,$EB,$B4,$E6,$FA,$D7,$0B,$E4,$7B,$27
	.dc.b	$F8,$E5,$F1,$C3,$CB,$87,$2A,$F9,$87,$07,$D7,$1A,$3D,$77,$E2,$9D
	.dc.b	$4F,$13,$D8,$7A,$8B,$1E,$36,$C7,$CA,$AF,$15,$FA,$5B,$1E,$1A,$ED
	.dc.b	$5C,$0F,$1D,$6E,$7C,$37,$E5,$8F,$6C,$61,$D4,$F1,$3C,$D5,$E3,$3F
	.dc.b	$03,$3B,$3C,$18,$7C,$A7,$E4,$19,$9F,$64,$F1,$8C,$C7,$96,$3C,$E6
	.dc.b	$5F,$3C,$74,$C3,$DB,$1B,$AE,$CE,$8F,$36,$1F,$3A,$33,$1E,$8A,$EF
	.dc.b	$7D,$4E,$EE,$09,$1C,$E3,$DA,$F4,$C3,$E1,$C7,$96,$3A,$1B,$9F,$A8
	.dc.b	$7B,$E3,$B5,$78,$AF,$E7,$4F,$AC,$D1,$E0,$9F,$1D,$73,$21,$D9,$CB
	.dc.b	$E0,$24,$FB,$F0,$F8,$B3,$E0,$6D,$75,$BF,$16,$18,$C3,$F5,$95,$5E
	.dc.b	$D3,$AA,$F2,$C6,$9A,$EA,$78,$61,$E1,$9C,$B6,$7E,$44,$7B,$DA,$7D
	.dc.b	$99,$E3,$7F,$90,$8C,$E1,$E9,$AE,$73,$98,$FA,$17,$F2,$AB,$F1,$75
	.dc.b	$F6,$1F,$87,$5B,$8C,$4E,$EB,$ED,$9F,$81,$E2,$FF,$1D,$4F,$F7,$2F
	.dc.b	$13,$FC,$49,$F2,$1F,$E2,$56,$46,$4F,$EE,$F1,$8C,$7F,$C0,$5F,$39
	.dc.b	$FE,$1F,$F3,$FF,$C4,$0D,$AF,$F1,$03,$67,$71,$FE,$23,$9F,$F7,$07
	.dc.b	$D9,$3E,$5A,$9F,$E3,$5D,$6D,$7F,$88,$27,$4D,$9D,$34,$96,$D3,$FB
	.dc.b	$05,$70,$8F,$F2,$2D,$F4,$7F,$C3,$5E,$9F,$F8,$DC,$FF,$1C,$7A,$97
	.dc.b	$0F,$F1,$C4,$59,$FF,$10,$4D,$C3,$FB,$B7,$C0,$FF,$92,$FE,$13,$FC
	.dc.b	$98,$DC,$3F,$C3,$AD,$BF,$F2,$AB,$6B,$CB,$F2,$FF,$08,$FF,$24,$7A
	.dc.b	$76,$FE,$04,$5B,$FF,$18,$4D,$AE,$35,$A3,$FC,$9D,$19,$7F,$CC,$1F
	.dc.b	$4F,$FE,$5F,$F8,$1F,$F3,$03,$63,$FC,$62,$DB,$FF,$34,$B6,$FF,$C4
	.dc.b	$FF,$3D,$FE,$5B,$3F,$CB,$9D,$B6,$FF,$CB,$91,$6F,$FC,$89,$36,$BF
	.dc.b	$C2,$1F,$8A,$FF,$12,$3F,$8B,$AE,$77,$D2,$7F,$8E,$63,$7D,$57,$2F
	.dc.b	$EC,$78,$B0,$E5,$3F,$E7,$9B,$D4,$FF,$68,$9C,$D7,$F9,$99,$EF,$87
	.dc.b	$68,$C5,$FF,$9D,$7D,$79,$1F,$E7,$AF,$AA,$FF,$41,$E5,$FE,$46,$9C
	.dc.b	$E3,$FE,$5F,$7B,$C2,$77,$5F,$DF,$27,$FD,$19,$F5,$32,$3F,$CB,$BF
	.dc.b	$53,$70,$FF,$4A,$CD,$C3,$FC,$7D,$36,$BD,$FC,$BF,$E6,$37,$E0,$DE
	.dc.b	$3C,$CF,$F1,$D3,$63,$2F,$F8,$22,$78,$38,$FC,$7F,$FD,$7A,$F2,$6B
	.dc.b	$F3,$1A,$FF,$84,$1A,$BE,$91,$E8,$3D,$6E,$F3,$98,$C3,$E2,$9F,$F7
	.dc.b	$67,$CD,$69,$E3,$5C,$30,$7A,$3F,$73,$5F,$F2,$F7,$B3,$74,$FF,$5B
	.dc.b	$37,$2F,$65,$55,$F1,$6F,$95,$D6,$0F,$D8,$7E,$3B,$F8,$B6,$E3,$F8
	.dc.b	$AA,$E1,$7F,$EA,$AF,$A4,$7D,$3F,$5F,$CB,$5A,$D3,$C1,$3A,$AF,$F5
	.dc.b	$8C,$3D,$3F,$0E,$4E,$1A,$79,$E7,$55,$FE,$B8,$FA,$57,$C6,$70,$F9
	.dc.b	$B9,$39,$8F,$77,$D3,$7F,$96,$23,$87,$0F,$C1,$BF,$4C,$7F,$1D,$E0
	.dc.b	$FD,$2D,$6E,$30,$7E,$0F,$D0,$F2,$FC,$6C,$8F,$4C,$7C,$BF,$AF,$4F
	.dc.b	$AB,$EB,$65,$7F,$B3,$A7,$2F,$FB,$43,$DE,$7C,$1F,$D3,$FC,$2D,$E0
	.dc.b	$53,$EB,$4E,$AB,$E3,$F9,$DF,$1B,$D3,$91,$83,$FD,$B7,$F8,$43,$E7
	.dc.b	$FE,$5F,$D0,$C9,$CC,$7E,$07,$EF,$5E,$F7,$E0,$FE,$11,$E5,$53,$51
	.dc.b	$F4,$7D,$3F,$AB,$F5,$53,$1F,$B3,$1F,$8B,$FB,$67,$9F,$F8,$DF,$90
	.dc.b	$FA,$B5,$FE,$EE,$6B,$F3,$04,$F0,$7F,$DE,$7F,$81,$BD,$FF,$CB,$7E
	.dc.b	$05,$F5,$BC,$AF,$F5,$83,$7B,$BF,$8D,$D6,$EF,$F1,$7C,$FF,$8A,$FA
	.dc.b	$D9,$18,$3F,$DF,$F6,$FA,$CF,$C6,$54,$9C,$C7,$E9,$7E,$33,$E9,$C6
	.dc.b	$08,$7A,$5C,$BE,$BF,$C1,$9F,$09,$E8,$FA,$99,$3F,$E9,$2D,$7F,$D9
	.dc.b	$15,$A8,$C6,$BF,$CD,$23,$11,$F1,$73,$5F,$28,$ED,$87,$EE,$6B,$78
	.dc.b	$7C,$4A,$D3,$DA,$75,$EE,$FC,$AE,$A9,$87,$80,$FF,$52,$E0,$9D,$C7
	.dc.b	$C6,$9D,$BE,$D8,$FE,$11,$F5,$AE,$9F,$E0,$FF,$35,$F2,$EB,$A6,$48
	.dc.b	$C4,$7D,$FA,$E0,$F8,$43,$87,$55,$D8,$EF,$D3,$FC,$5F,$17,$97,$7C
	.dc.b	$D9,$7C,$E7,$2D,$9E,$04,$7C,$1F,$5B,$FA,$0C,$FF,$D8,$BF,$58,$BF
	.dc.b	$EC,$B3,$9A,$4E,$5F,$F6,$5C,$F6,$72,$FF,$6C,$31,$1B,$84,$F3,$8C
	.dc.b	$36,$78,$5F,$F7,$49,$E1,$1C,$8F,$1E,$3F,$E3,$1B,$EE,$7F,$B7,$2F
	.dc.b	$93,$E4,$BE,$55,$D2,$3D,$AF,$2C,$2C,$71,$BE,$98,$78,$2F,$8C,$F3
	.dc.b	$36,$7A,$A7,$77,$CE,$3C,$67,$C9,$9E,$2D,$38,$E1,$E6,$79,$61,$FD
	.dc.b	$82,$3E,$74,$7F,$74,$6B,$D4,$C7,$C6,$BE,$B7,$83,$63,$FB,$BC,$A3
	.dc.b	$11,$C2,$7F,$B6,$A1,$F3,$D8,$69,$A7,$04,$7D,$B7,$F9,$FE,$5F,$87
	.dc.b	$1C,$E7,$E2,$CE,$63,$E1,$99,$38,$EC,$AE,$31,$E0,$D7,$F0,$D5,$C5
	.dc.b	$FC,$44,$71,$73,$5C,$63,$A9,$C7,$4F,$F0,$04,$66,$78,$CE,$4F,$E2
	.dc.b	$11,$99,$FF,$0E,$E1,$E8,$8C,$4F,$F8,$92,$31,$87,$99,$B1,$EA,$8F
	.dc.b	$DC,$61,$EF,$3F,$C5,$31,$DC,$FF,$43,$57,$F9,$29,$B3,$C2,$31,$1E
	.dc.b	$DB,$F4,$CF,$F0,$3F,$AD,$FD,$EB,$75,$E5,$1F,$66,$30,$4A,$FE,$95
	.dc.b	$1C,$1A,$70,$9E,$3C,$B2,$BF,$8A,$FC,$EF,$F8,$7D,$F0,$46,$6B,$9B
	.dc.b	$FE,$54,$9D,$46,$27,$CD,$F4,$9E,$43,$FC,$B3,$C9,$FD,$BE,$32,$49
	.dc.b	$F2,$CE,$BC,$6E,$13,$9D,$7E,$E5,$33,$5E,$EB,$47,$F4,$9F,$AD,$7C
	.dc.b	$5F,$CD,$FA,$97,$D6,$FF,$9A,$A7,$35,$FC,$B6,$72,$D9,$FA,$F8,$A7
	.dc.b	$2F,$F9,$81,$AF,$F9,$99,$AF,$FA,$23,$E0,$BF,$A3,$57,$F3,$96,$C7
	.dc.b	$F7,$B9,$E7,$9F,$B2,$72,$FB,$BE,$3F,$FC,$1B,$7F,$E6,$06,$DF,$F9
	.dc.b	$C1,$AF,$F9,$D9,$AF,$FA,$63,$EB,$5F,$23,$EA,$6D,$7F,$9A,$9B,$3F
	.dc.b	$E7,$A6,$BF,$E8,$0F,$AD,$7A,$5F,$09,$B5,$FE,$72,$6C,$FF,$A0,$9A
	.dc.b	$FF,$A2,$3E,$B5,$F3,$3F,$29,$B5,$FE,$7A,$6C,$FF,$A2,$9A,$FC,$BF
	.dc.b	$80,$FA,$3B,$7F,$99,$2B,$FD,$02,$D9,$FA,$53,$F5,$BE,$B1,$F2,$74
	.dc.b	$97,$FE,$88,$8C,$BF,$E8,$AF,$AC,$FC,$BD,$FF,$A7,$9A,$FF,$B0,$1A
	.dc.b	$FF,$23,$6B,$F9,$0F,$83,$FC,$2F,$37,$F9,$B2,$BF,$D2,$AD,$9F,$F5
	.dc.b	$2B,$5F,$F5,$5F,$D5,$FF,$98,$27,$FD,$36,$D8,$FF,$55,$B5,$DF,$C0
	.dc.b	$7E,$4B,$89,$F1,$94,$97,$FE,$A4,$8C,$BF,$1B,$D9,$F9,$24,$70,$7F
	.dc.b	$AA,$BC,$3F,$F1,$8B,$7C,$DF,$CF,$C6,$31,$FE,$43,$1F,$A2,$EF,$FD
	.dc.b	$3C,$7F,$B0,$1A,$FF,$B1,$9A,$FF,$BA,$1A,$FF,$4B,$F0,$3F,$2A,$DC
	.dc.b	$3F,$5B,$18,$83,$7E,$A3,$D7,$1F,$EC,$E6,$BF,$ED,$46,$BF,$EF,$66
	.dc.b	$BF,$D6,$7E,$A3,$8F,$67,$F0,$86,$E7,$FE,$4D,$B7,$D4,$7F,$31,$9C
	.dc.b	$DF,$1D,$BF,$80,$C7,$48,$34,$91,$FE,$F4,$F9,$FF,$EB,$18,$7F,$8D
	.dc.b	$E7,$9E,$29,$CD,$FF,$A9,$A1,$F2,$FF,$EF,$A8,$46,$3E,$A3,$FD,$FF
	.dc.b	$39,$8C,$61,$FE,$48,$8C,$09,$DF,$83,$C1,$F2,$7D,$44,$AF,$F2,$7E
	.dc.b	$3F,$EA,$A6,$C6,$71,$F2,$4A,$3D,$9F,$0D,$9C,$BF,$98,$FD,$3B,$EA
	.dc.b	$3F,$AC,$37,$1F,$27,$D2,$E2,$64,$7F,$BE,$27,$59,$7F,$83,$48,$E1
	.dc.b	$C0,$70,$EA,$7F,$BE,$5F,$FB,$B7,$ED,$3A,$CF,$CB,$F9,$14,$7D,$3F
	.dc.b	$A9,$FA,$9F,$35,$72,$F8,$2D,$BF,$7B,$70,$EB,$B2,$3E,$8F,$CE,$4B
	.dc.b	$FA,$1F,$35,$70,$FB,$5C,$6A,$C6,$3E,$68,$56,$2F,$E0,$6E,$F8,$0F
	.dc.b	$F0,$03,$63,$EA,$C6,$08,$FC,$4C,$7F,$C6,$93,$B7,$0D,$7E,$4B,$5E
	.dc.b	$F3,$F9,$5F,$99,$E7,$1F,$79,$FC,$6F,$C7,$FE,$3B,$D8,$B8,$7E,$35
	.dc.b	$B5,$EE,$6E,$7E,$D7,$F4,$1F,$53,$2B,$F1,$BF,$53,$6F,$F1,$73,$6F
	.dc.b	$E6,$CD,$AF,$87,$F0,$5F,$08,$FC,$FF,$CC,$D8,$FF,$50,$6B,$FD,$A3
	.dc.b	$1F,$6D,$7F,$74,$8C,$47,$29,$F8,$CA,$3F,$B5,$C7,$FB,$2F,$97,$FD
	.dc.b	$8A,$F4,$AF,$C5,$B9,$F6,$BF,$75,$F5,$3F,$BB,$F4,$AC,$FE,$A1,$B8
	.dc.b	$FE,$99,$BA,$7E,$21,$FD,$BF,$C1,$4A,$FD,$B7,$C0,$DB,$FC,$3C,$DB
	.dc.b	$FC,$34,$DA,$FD,$27,$D4,$BE,$9F,$FC,$88,$D8,$FD,$CF,$C8,$62,$64
	.dc.b	$7F,$B1,$67,$58,$FE,$EA,$B2,$FF,$0C,$9E,$B5,$8C,$3F,$3F,$56,$72
	.dc.b	$73,$9F,$E3,$91,$9A,$FF,$42,$5F,$6B,$FF,$19,$B6,$FF,$DF,$1D,$6B
	.dc.b	$E1,$6A,$D3,$CB,$0E,$B7,$DC,$91,$E7,$34,$FC,$03,$38,$78,$27,$C8
	.dc.b	$FF,$2D,$BE,$F7,$DA,$FC,$15,$D9,$E8,$62,$F9,$9F,$01,$F9,$0E,$1F
	.dc.b	$09,$C2,$7D,$37,$DE,$75,$3C,$CD,$57,$79,$F3,$4E,$AB,$8E,$1F,$16
	.dc.b	$31,$5E,$A8,$C5,$F2,$9C,$9F,$05,$C1,$2F,$AB,$E8,$31,$1E,$B8,$C9
	.dc.b	$F7,$9E,$71,$A3,$FA,$7C,$F8,$27,$73,$C4,$93,$AC,$3E,$6D,$F9,$DB
	.dc.b	$3F,$1E,$BE,$25,$75,$3E,$23,$D1,$E5,$1E,$28,$C1,$C0,$D4,$70,$8F
	.dc.b	$0C,$65,$F1,$C7,$CF,$3F,$A3,$E1,$FA,$06,$DF,$F9,$E2,$31,$1F,$74
	.dc.b	$CC,$FF,$06,$8C,$C7,$DE,$32,$7F,$62,$F7,$3E,$A1,$27,$DD,$8F,$F9
	.dc.b	$E1,$F2,$B6,$7F,$0F,$87,$F4,$AA,$E1,$1F,$3E,$3F,$C0,$51,$87,$E0
	.dc.b	$CF,$19,$F8,$07,$D4,$9D,$57,$66,$C7,$68,$FD,$7C,$62,$38,$D7,$69
	.dc.b	$F3,$CE,$6F,$F2,$2D,$C3,$FC,$30,$D9,$FF,$02,$B6,$FE,$14,$72,$23
	.dc.b	$EB,$C3,$D1,$7D,$27,$38,$7B,$2F,$93,$F0,$E3,$27,$AC,$B7,$EB,$3E
	.dc.b	$03,$87,$D7,$7D,$B0,$F3,$CE,$70,$F0,$57,$8F,$0E,$F3,$F1,$0D,$57
	.dc.b	$09,$FE,$EF,$F3,$3F,$E2,$C3,$FC,$5B,$F5,$36,$BF,$C5,$91,$67,$FB
	.dc.b	$EC,$DC,$BF,$C0,$BF,$21,$FE,$3F,$F3,$FF,$C8,$0D,$C3,$FC,$5A,$DB
	.dc.b	$FF,$15,$B6,$BF,$DB,$3F,$52,$F9,$8F,$F8,$00,$DB,$FF,$1E,$67,$89
	.dc.b	$91,$FD,$72,$75,$87,$F8,$0D,$AF,$F8,$A1,$AF,$F8,$B5,$DD,$FF,$8A
	.dc.b	$39,$9F,$5D,$D2,$BF,$C5,$0F,$F9,$67,$D4,$98,$7F,$91,$BD,$CB,$5F
	.dc.b	$E4,$49,$B8,$7F,$89,$A6,$D2,$BF,$CB,$3E,$1F,$FB,$12,$BF,$C4,$5F
	.dc.b	$05,$FE,$5A,$F8,$1F,$F2,$DB,$6F,$FC,$B4,$DB,$FF,$15,$BF,$13,$EA
	.dc.b	$A4,$FF,$96,$3E,$16,$D7,$F3,$AD,$95,$FE,$F6,$6D,$7C,$78,$C1,$1F
	.dc.b	$36,$3F,$E4,$AF,$C3,$D8,$F4,$8C,$3F,$E7,$0F,$AB,$FF,$38,$FC,$CD
	.dc.b	$BF,$F3,$83,$72,$FF,$20,$B6,$FF,$C8,$0F,$FA,$0F,$E0,$E5,$7F,$9C
	.dc.b	$3C,$2B,$7F,$E6,$E9,$B7,$FE,$50,$9B,$5F,$E6,$5F,$AA,$FE,$68,$DE
	.dc.b	$EF,$F4,$1F,$D5,$FF,$86,$31,$FA,$15,$A6,$CA,$BF,$C4,$73,$FD,$32
	.dc.b	$7F,$CD,$AD,$7A,$46,$A7,$A6,$0E,$67,$C6,$77,$B6,$1D,$EB,$53,$FE
	.dc.b	$A1,$F7,$A6,$3F,$38,$F9,$FF,$57,$FE,$99,$F0,$93,$AC,$7F,$D1,$EF
	.dc.b	$82,$74,$FF,$A8,$3E,$AE,$5F,$FA,$7F,$E6,$6D,$FF,$9C,$26,$E5,$FE
	.dc.b	$6F,$9B,$5F,$E8,$DF,$AB,$FF,$66,$13,$0F,$89,$87,$F9,$AE,$3F,$DA
	.dc.b	$1F,$57,$FE,$7F,$F7,$BF,$C9,$2D,$9F,$F3,$82,$F6,$1F,$2F,$AF,$65
	.dc.b	$7F,$E5,$0F,$C5,$1E,$87,$F9,$4B,$14,$61,$D4,$66,$31,$1F,$E7,$DF
	.dc.b	$0B,$FD,$0E,$FF,$96,$9A,$E7,$D9,$E9,$18,$8F,$F5,$3F,$D6,$3E,$AB
	.dc.b	$C0,$D8,$FF,$31,$B5,$FF,$45,$35,$FF,$57,$FD,$5B,$DB,$F4,$36,$3F
	.dc.b	$CD,$6D,$7F,$D2,$4D,$7F,$D6,$F6,$F5,$FC,$A4,$8F,$53,$F0,$FE,$AD
	.dc.b	$F1,$F4,$93,$FE,$78,$8C,$BF,$E7,$AF,$AB,$FF,$31,$37,$CF,$F9,$91
	.dc.b	$88,$FF,$52,$35,$F1,$FA,$7F,$18,$8F,$CB,$7E,$A7,$D5,$3E,$3E,$92
	.dc.b	$7F,$D1,$31,$97,$FD,$17,$F0,$0F,$86,$FB,$8D,$8C,$C6,$0F,$AB,$E8
	.dc.b	$48,$FB,$4D,$7F,$D7,$5F,$53,$FA,$C7,$FD,$2E,$D7,$3E,$27,$E2,$63
	.dc.b	$11,$FE,$E4,$FA,$97,$BB,$F8,$D6,$C7,$FA,$81,$AF,$FB,$1D,$AF,$D6
	.dc.b	$F3,$7F,$12,$47,$F1,$CF,$E5,$FE,$9F,$FC,$F1,$1F,$EA,$A6,$BF,$ED
	.dc.b	$16,$BF,$EF,$66,$BF,$EF,$BF,$A7,$7B,$5F,$A1,$6C,$7F,$AD,$1A,$FF
	.dc.b	$B6,$DA,$FE,$2B,$C6,$FC,$E9,$1F,$D1,$3F,$AB,$FA,$57,$D3,$52,$4F
	.dc.b	$FB,$02,$32,$FF,$B0,$BE,$21,$5F,$B3,$6C,$66,$30,$7E,$DF,$E9,$9E
	.dc.b	$FF,$F5,$B9,$FF,$43,$DF,$ED,$BC,$CF,$E1,$30,$8C,$79,$9B,$BE,$33
	.dc.b	$FC,$84,$FE,$C8,$E1,$E7,$E2,$F5,$BE,$77,$CA,$77,$86,$70,$E8,$D8
	.dc.b	$FE,$DD,$3B,$7C,$57,$95,$87,$7C,$3C,$EB,$D4,$D5,$BE,$8A,$9C,$67
	.dc.b	$57,$E9,$39,$D6,$9C,$57,$17,$53,$C1,$FF,$56,$FD,$33,$E2,$FD,$18
	.dc.b	$BE,$5F,$EA,$D7,$FA,$F6,$31,$C1,$E9,$BE,$13,$E1,$6B,$D2,$F4,$6F
	.dc.b	$9B,$E0,$68,$FA,$3B,$E7,$4B,$F5,$4E,$5B,$3C,$08,$FB,$FE,$9B,$F5
	.dc.b	$B9,$FF,$86,$FE,$97,$DD,$29,$CA,$8C,$DF,$58,$C3,$F8,$93,$11,$CA
	.dc.b	$78,$B7,$15,$F7,$AE,$18,$7E,$46,$BA,$B6,$3F,$29,$7C,$9B,$7C,$B0
	.dc.b	$F0,$9F,$E3,$38,$F0,$9D,$1E,$C7,$F8,$DE,$79,$36,$79,$5F,$82,$36
	.dc.b	$7E,$1E,$FC,$4E,$F0,$CD,$F2,$8D,$1C,$30,$F0,$BB,$3C,$A7,$CB,$C3
	.dc.b	$D0,$FE,$BC,$C5,$57,$81,$9A,$FA,$2D,$F8,$A7,$D3,$6E,$3E,$B4,$61
	.dc.b	$E5,$28,$C4,$71,$9F,$EE,$CB,$E7,$29,$1F,$EF,$4D,$CF,$F0,$78,$7E
	.dc.b	$77,$9B,$E2,$D7,$F9,$86,$BE,$3C,$E6,$3D,$E6,$4F,$82,$E6,$BF,$1B
	.dc.b	$19,$7A,$ED,$FE,$16,$73,$1E,$53,$8C,$67,$4F,$8D,$7E,$08,$C4,$FC
	.dc.b	$B8,$C6,$1E,$68,$E5,$7D,$A3,$13,$A9,$F3,$9A,$BF,$E2,$D7,$D2,$3F
	.dc.b	$2E,$FF,$90,$F0,$F7,$1F,$C9,$23,$D5,$3C,$2F,$F8,$54,$E9,$C4,$FE
	.dc.b	$81,$F4,$BE,$C7,$F8,$87,$D3,$7E,$ED,$BB,$7F,$94,$63,$FD,$1D,$18
	.dc.b	$24,$F3,$3F,$D2,$51,$C5,$C0,$78,$B8,$75,$5F,$CF,$1F,$F1,$CC,$F8
	.dc.b	$B9,$BF,$CC,$7F,$39,$FE,$2D,$8C,$12,$F9,$4E,$BD,$1F,$33,$E7,$F9
	.dc.b	$F9,$5F,$D0,$F1,$FE,$45,$E1,$E3,$0F,$15,$59,$FF,$2A,$9A,$AF,$87
	.dc.b	$19,$7F,$AA,$64,$8F,$F3,$BF,$3F,$F9,$DA,$76,$7F,$39,$FA,$27,$C7
	.dc.b	$FF,$9E,$5B,$1E,$3F,$A6,$7C,$5E,$6A,$4F,$F4,$1F,$67,$FA,$43,$CE
	.dc.b	$3F,$D0,$51,$8D,$3E,$6B,$74,$FE,$CA,$DA,$FE,$D5,$8F,$82,$B7,$DF
	.dc.b	$FE,$8B,$F8,$8F,$F4,$1D,$7F,$4F,$FA,$67,$C7,$67,$E0,$7F,$D2,$11
	.dc.b	$89,$D4,$FF,$A4,$FE,$9B,$FD,$9D,$1F,$DD,$E7,$58,$FE,$02,$72,$FF
	.dc.b	$B5,$5A,$BF,$06,$7D,$67,$3F,$43,$FD,$3D,$18,$C3,$FB,$17,$D4,$3E
	.dc.b	$5F,$FD,$4B,$F0,$1F,$C2,$E7,$31,$8F,$0F,$E7,$7B,$DF,$EA,$AF,$4B
	.dc.b	$F1,$FC,$BF,$E0,$8E,$44,$E7,$6F,$F5,$67,$D4,$3E,$3F,$87,$76,$1B
	.dc.b	$A7,$F7,$3F,$A8,$FC,$95,$FF,$69,$FB,$1F,$CC,$CF,$F8,$5E,$73,$9F
	.dc.b	$E2,$A7,$35,$FE,$F5,$EF,$FF,$38,$DF,$E4,$E7,$FC,$56,$FF,$7F,$8C
	.dc.b	$CF,$5F,$3B,$FD,$81,$F5,$3F,$E6,$7A,$FC,$D6,$AF,$88,$CF,$BD,$FE
	.dc.b	$C8,$8C,$4F,$F8,$27,$AD,$F4,$FF,$EC,$CF,$B1,$7C,$7F,$FB,$2E,$73
	.dc.b	$5F,$A5,$9C,$BF,$99,$EF,$FF,$69,$FD,$8F,$FA,$56,$30,$7E,$AD,$C6
	.dc.b	$DF,$BE,$F3,$FF,$C7,$33,$9E,$4F,$F6,$C4,$62,$7F,$C2,$15,$F3,$EB
	.dc.b	$FC,$3D,$E4,$7E,$89,$FD,$B7,$D8,$BE,$8B,$FD,$C1,$39,$BF,$F0,$C4
	.dc.b	$E6,$BF,$C4,$DE,$77,$E6,$A7,$39,$FF,$B9,$E3,$1C,$FF,$BC,$BB,$7F
	.dc.b	$E4,$DF,$AC,$F8,$38,$7F,$88,$FE,$B9,$F4,$3F,$EE,$F9,$CE,$1F,$C1
	.dc.b	$27,$37,$FE,$35,$FA,$CC,$BF,$E5,$76,$E1,$FE,$0C,$C9,$F5,$BF,$EF
	.dc.b	$99,$CC,$9D,$E0,$F9,$86,$1B,$E6,$FE,$26,$AF,$FC,$2F,$3A,$8F,$14
	.dc.b	$EA,$73,$1C,$4F,$E4,$13,$B8,$FF,$20,$7D,$0F,$FB,$A2,$7F,$92,$2F
	.dc.b	$AD,$E5,$82,$35,$5C,$27,$31,$89,$FE,$4E,$F1,$AE,$7E,$86,$AB,$9E
	.dc.b	$0F,$AD,$A6,$3E,$2E,$DF,$9A,$2E,$5F,$E5,$5A,$FF,$60,$4E,$4F,$F3
	.dc.b	$27,$D6,$BE,$B7,$FB,$05,$7C,$58,$C4,$FF,$84,$DB,$1F,$E1,$B6,$BF
	.dc.b	$E1,$F7,$3A,$7F,$40,$F9,$F7,$D6,$FC,$BA,$56,$AF,$E9,$5A,$F7,$1E
	.dc.b	$4C,$3E,$4C,$62,$FF,$B0,$DF,$17,$FC,$9F,$19,$BF,$E9,$F8,$BE,$73
	.dc.b	$E3,$F8,$7F,$21,$B9,$7D,$A6,$E1,$F1,$5B,$7F,$4B,$EB,$7E,$A9,$F5
	.dc.b	$BE,$72,$DF,$DE,$8B,$7F,$82,$9B,$5F,$EA,$CF,$AC,$7C,$77,$CC,$A4
	.dc.b	$63,$0F,$F1,$F3,$63,$FC,$8C,$D7,$FC,$94,$E6,$BD,$BF,$3B,$F2,$E3
	.dc.b	$15,$F2,$B4,$FE,$13,$5E,$1C,$3F,$0F,$7B,$6D,$7E,$5B,$E5,$7F,$23
	.dc.b	$F0,$DF,$92,$6C,$64,$FC,$DC,$65,$A7,$E8,$BE,$B7,$F2,$67,$E5,$3C
	.dc.b	$3B,$19,$26,$9F,$91,$8C,$BF,$E1,$AD,$3F,$B6,$3F,$6B,$E1,$3D,$9F
	.dc.b	$01,$F8,$06,$DF,$F8,$3F,$8F,$FC,$37,$5C,$6B,$53,$FE,$BD,$BE,$7B
	.dc.b	$7E,$2F,$0E,$37,$FE,$B5,$C3,$8C,$FE,$1B,$EB,$7F,$57,$EC,$FE,$B1
	.dc.b	$B8,$FE,$F9,$B7,$FA,$96,$D7,$ED,$BE,$33,$F7,$27,$EE,$BD,$AB,$7F
	.dc.b	$C0,$A2,$DF,$F0,$69,$B5,$FE,$F0,$FA,$D7,$CE,$7E,$F3,$E1,$7F,$4D
	.dc.b	$18,$C7,$F8,$7C,$62,$75,$1F,$ED,$B9,$D6,$1F,$BD,$AD,$4F,$03,$55
	.dc.b	$C2,$7F,$C9,$35,$C2,$FF,$4F,$7C,$27,$95,$F0,$39,$3C,$2F,$95,$7A
	.dc.b	$5B,$3F,$CA,$E5,$7D,$63,$F2,$97,$D6,$FB,$C7,$5A,$EE,$75,$7B,$C7
	.dc.b	$56,$9D,$6B,$A4,$FF,$38,$AE,$93,$D8,$E9,$5D,$A7,$A1,$D9,$E8,$D8
	.dc.b	$E9,$7E,$A8,$FE,$89,$7D,$AB,$FA,$55,$F6,$9F,$05,$F6,$3C,$0F,$69
	.dc.b	$F0,$1D,$8B,$1D,$8F,$5C,$76,$BF,$03,$F2,$1B,$7E,$1E,$6F,$F2,$75
	.dc.b	$F3,$BE,$ED,$BF,$19,$E1,$8C,$9D,$DF,$71,$DD,$F0,$BD,$DB,$5D,$F0
	.dc.b	$F7,$57,$C0,$9F,$75,$76,$8F,$01,$97,$C0,$66,$3C,$11,$F0,$23,$C0
	.dc.b	$78,$8F,$01,$6B,$C1,$3D,$A7,$A1,$DA,$FE,$0C,$74,$3D,$B5,$D2,$7B
	.dc.b	$3D,$1B,$5D,$23,$CB,$1D,$27,$94,$79,$CE,$4F,$47,$93,$67,$94,$F8
	.dc.b	$EF,$E2,$DF,$92,$F8,$9C,$CE,$25,$AE,$37,$E0,$BF,$14,$F0,$6B,$C1
	.dc.b	$B3,$C2,$35,$5D,$DD,$36,$35,$18,$BF,$29,$2F,$E6,$56,$F0,$FC,$69
	.dc.b	$23,$F0,$2D,$AF,$A8,$DC,$3E,$5B,$67,$E2,$36,$7D,$CD,$9F,$4B,$6B
	.dc.b	$C1,$5D,$89,$5D,$9B,$5D,$9A,$76,$2C,$76,$6C,$76,$7F,$56,$FF,$8B
	.dc.b	$7E,$6F,$FC,$56,$7F,$8B,$3C,$AB,$3F,$E3,$48,$B5,$FE,$38,$9B,$5F
	.dc.b	$E0,$8F,$97,$FF,$16,$FC,$AF,$F8,$B9,$B8,$7F,$90,$1B,$7F,$E2,$B6
	.dc.b	$D7,$F9,$3A,$3F,$C9,$9F,$56,$FA,$DF,$F0,$CF,$D6,$FF,$88,$63,$18
	.dc.b	$7F,$5B,$8C,$4E,$A3,$FC,$46,$7F,$2B,$9D,$D7,$F5,$9A,$DC,$F1,$37
	.dc.b	$1E,$0A,$E4,$FF,$8F,$FD,$7F,$F2,$39,$FE,$48,$F6,$6D,$7F,$96,$A2
	.dc.b	$CF,$F9,$16,$6E,$3F,$E5,$2F,$87,$FF,$2A,$FC,$5F,$F9,$59,$B8,$7F
	.dc.b	$97,$9B,$7F,$E6,$26,$D7,$C3,$86,$6F,$AD,$FF,$31,$FD,$6B,$1F,$89
	.dc.b	$18,$6E,$3E,$CC,$15,$B8,$F7,$E1,$D2,$F8,$46,$30,$EC,$F1,$AE,$14
	.dc.b	$BF,$25,$6A,$31,$5F,$E5,$2A,$DD,$7F,$9C,$BE,$B7,$FC,$EB,$F1,$1F
	.dc.b	$E7,$66,$D7,$F9,$F9,$B7,$FE,$82,$6D,$7F,$94,$BD,$DF,$F3,$B9,$FE
	.dc.b	$78,$F0,$2D,$FF,$A2,$A2,$DF,$F9,$D6,$6D,$7C,$9F,$AB,$7D,$67,$F9
	.dc.b	$25,$B8,$7F,$6F,$BF,$EE,$11,$87,$01,$C3,$B7,$55,$FE,$75,$F8,$8F
	.dc.b	$F3,$A4,$62,$BF,$CE,$7E,$0A,$73,$87,$F9,$C9,$B3,$FE,$33,$7F,$D2
	.dc.b	$7E,$EC,$8F,$B1,$F0,$5F,$E9,$B7,$FC,$A9,$19,$CB,$FD,$24,$FF,$A9
	.dc.b	$27,$51,$FE,$B2,$8F,$F5,$1F,$D6,$4A,$FF,$55,$FC,$FD,$AF,$F5,$7C
	.dc.b	$DD,$3F,$D4,$F3,$6B,$FC,$25,$F5,$7F,$EF,$BF,$67,$E1,$E0,$8C,$3F
	.dc.b	$19,$FF,$43,$7C,$EE,$9C,$C7,$CD,$FA,$BF,$F4,$CB,$C9,$AF,$CE,$6B
	.dc.b	$BF,$63,$BA,$F8,$77,$9F,$86,$C7,$91,$AF,$D8,$6B,$F7,$7B,$5E,$A7
	.dc.b	$F8,$B7,$6F,$D6,$46,$27,$F4,$6D,$7F,$CB,$DE,$CE,$F6,$FF,$1A,$7C
	.dc.b	$83,$E6,$71,$5F,$EE,$18,$C3,$FA,$C6,$BF,$E6,$7F,$8C,$DA,$FA,$B7
	.dc.b	$B3,$86,$C7,$ED,$1A,$FE,$B0,$FF,$77,$7D,$8B,$CA,$E7,$A3,$EB,$29
	.dc.b	$2B,$FC,$FD,$F2,$1E,$0F,$AB,$7D,$65,$24,$FF,$A0,$FB,$7C,$BF,$59
	.dc.b	$FE,$6A,$8F,$E0,$AD,$7C,$41,$C7,$1F,$F1,$5B,$57,$93,$D1,$97,$F3
	.dc.b	$1F,$8F,$5F,$ED,$5C,$3E,$1F,$D6,$3E,$4E,$92,$FF,$D2,$DC,$9F,$23
	.dc.b	$E0,$1F,$0F,$49,$3F,$C6,$BE,$C5,$E1,$FC,$4F,$46,$4F,$DC,$7C,$3E
	.dc.b	$47,$F2,$98,$FE,$43,$F5,$BF,$B5,$6F,$91,$F5,$A3,$11,$FE,$9F,$FB
	.dc.b	$6F,$B3,$F5,$CF,$7E,$92,$7F,$95,$FC,$2B,$CF,$FA,$BE,$94,$9F,$C6
	.dc.b	$BF,$82,$F9,$2F,$A7,$1F,$CC,$BE,$BB,$F8,$AB,$7D,$1F,$C3,$C6,$23
	.dc.b	$E8,$FD,$6F,$E3,$3A,$7F,$A2,$7D,$73,$DE,$FC,$6B,$6B,$F9,$EB,$5F
	.dc.b	$B2,$F9,$5F,$CF,$FA,$7F,$8A,$23,$FD,$12,$DE,$47,$E6,$1B,$19,$8C
	.dc.b	$1F,$9C,$F8,$A5,$52,$4F,$F4,$E8,$CB,$FD,$43,$F0,$B7,$C4,$FE,$93
	.dc.b	$EB,$FF,$39,$3E,$18,$F0,$19,$F2,$64,$64,$87,$E5,$27,$8B,$63,$8B
	.dc.b	$AD,$7F,$B5,$4E,$9B,$C5,$ED,$FA,$B9,$5F,$D9,$2B,$E3,$3D,$67,$2F
	.dc.b	$D2,$9F,$44,$7E,$F3,$EB,$DC,$BF,$9E,$F9,$CF,$8F,$7F,$B9,$C5,$18
	.dc.b	$7F,$BE,$57,$09,$FE,$E6,$D7,$F5,$ED,$8D,$D7,$7E,$7F,$E1,$78,$BE
	.dc.b	$B2,$FB,$53,$1F,$EF,$B3,$96,$CF,$02,$3F,$E0,$2F,$AF,$FC,$E6,$28
	.dc.b	$7D,$72,$FF,$DA,$95,$A9,$CC,$7F,$B5,$CE,$CE,$0E,$4E,$27,$94,$75
	.dc.b	$9E,$51,$88,$46,$27,$A5,$FF,$B8,$EB,$51,$CC,$F1,$E3,$FE,$13,$BF
	.dc.b	$01,$C9,$C4,$F0,$AE,$55,$D2,$BD,$D3,$E1,$BF,$04,$79,$6F,$8D,$6E
	.dc.b	$3A,$BE,$17,$58,$72,$8F,$1B,$63,$E0,$E1,$E0,$69,$93,$CD,$16,$FE
	.dc.b	$64,$7F,$84,$DA,$F4,$3F,$C2,$6E,$1F,$95,$F1,$CF,$AF,$FE,$EB,$28
	.dc.b	$C4,$F0,$6B,$D5,$AF,$F8,$7B,$E8,$BF,$B7,$A6,$91,$FE,$E6,$FA,$6F
	.dc.b	$F3,$FE,$C7,$AE,$7F,$11,$39,$DB,$FC,$F5,$AF,$DD,$9C,$9F,$5A,$F8
	.dc.b	$BF,$60,$CC,$7E,$0E,$32,$F1,$8F,$21,$C6,$7A,$E9,$E9,$8C,$61,$E3
	.dc.b	$77,$5C,$EB,$71,$FE,$3F,$C3,$CD,$5B,$6C,$78,$CE,$37,$FE,$07,$AE
	.dc.b	$31,$FE,$61,$7F,$D2,$D8,$7F,$9D,$4F,$59,$FE,$77,$8F,$5C,$F1,$BF
	.dc.b	$F3,$7C,$E9,$C4,$FF,$6C,$36,$7E,$C6,$34,$FE,$CF,$EB,$FF,$7D,$3F
	.dc.b	$E7,$86,$C7,$F7,$47,$F7,$11,$9D,$7C,$AF,$EE,$E5,$3C,$2B,$FD,$18
	.dc.b	$D9,$D3,$CA,$7A,$57,$27,$C5,$CB,$5E,$6E,$6D,$F4,$1F,$E2,$5C,$3C
	.dc.b	$D3,$BC,$3F,$9F,$C6,$0F,$F4,$A4,$EE,$FD,$55,$BF,$52,$57,$FA,$67
	.dc.b	$93,$FC,$D4,$FC,$18,$CD,$7C,$17,$C5,$3A,$8C,$4E,$C8,$FF,$33,$8D
	.dc.b	$63,$F0,$AB,$D9,$F5,$0F,$92,$A7,$B7,$EB,$DF,$5F,$8C,$BF,$D5,$35
	.dc.b	$B7,$FD,$57,$F0,$DF,$E0,$17,$37,$F1,$E3,$19,$FF,$8A,$9B,$97,$FA
	.dc.b	$F5,$B5,$F5,$EF,$A5,$6F,$AB,$E8,$7C,$73,$E2,$3E,$97,$D7,$BE,$43
	.dc.b	$31,$86,$63,$FE,$BE,$FA,$F7,$A5,$8C,$BF,$C7,$71,$89,$D4,$FD,$DF
	.dc.b	$AE,$7D,$7F,$DF,$AD,$5F,$F9,$0A,$73,$7F,$EC,$DF,$86,$7B,$DF,$83
	.dc.b	$6C,$67,$E5,$B1,$3F,$87,$F4,$9E,$2D,$31,$DB,$6F,$FD,$B1,$2F,$88
	.dc.b	$E0,$71,$F3,$6C,$FF,$B7,$7E,$B3,$FC,$94,$DD,$FF,$29,$18,$DB,$FB
	.dc.b	$93,$E8,$6F,$66,$BE,$0B,$6D,$BF,$F7,$47,$C3,$3C,$FF,$CF,$3F,$E7
	.dc.b	$88,$CC,$63,$E2,$7F,$45,$F1,$7F,$E6,$E9,$FF,$78,$37,$1F,$F0,$13
	.dc.b	$F9,$FE,$EF,$E1,$2E,$5F,$E1,$7F,$57,$FE,$9E,$AF,$F7,$C3,$6B,$FC
	.dc.b	$14,$BE,$7F,$F6,$0E,$6B,$8F,$76,$0C,$F6,$5C,$3F,$86,$1F,$C3,$7E
	.dc.b	$AD,$F3,$DF,$B9,$5F,$3F,$63,$0F,$EF,$7D,$DF,$F6,$6E,$25,$F4,$D9
	.dc.b	$F1,$74,$EF,$E7,$F8,$78,$57,$3F,$0F,$03,$1E,$53,$FC,$27,$E9,$FA
	.dc.b	$9D,$2B,$72,$8C,$E1,$BC,$78,$36,$B7,$3C,$70,$FF,$18,$3C,$23,$35
	.dc.b	$C1,$EB,$F4,$0F,$A7,$FE,$34,$D8,$CF,$C5,$62,$64,$70,$77,$18,$C3
	.dc.b	$FD,$4B,$F4,$FF,$97,$AF,$7C,$23,$0F,$93,$E9,$3F,$95,$46,$34,$57
	.dc.b	$F3,$78,$DD,$7F,$39,$39,$78,$1F,$D3,$5F,$F1,$B4,$69,$E1,$19,$A6
	.dc.b	$4F,$A6,$F6,$FD,$3F,$B9,$B8,$FC,$26,$E1,$FE,$57,$AE,$13,$F1,$AB
	.dc.b	$83,$4E,$05,$9F,$9F,$F4,$FF,$6C,$FB,$9F,$35,$67,$EA,$C5,$9F,$AB
	.dc.b	$37,$1F,$83,$1F,$E5,$1D,$3F,$61,$7F,$29,$B1,$E5,$EE,$FE,$A5,$18
	.dc.b	$AE,$32,$8D,$BF,$25,$AF,$F6,$36,$BD,$E7,$E4,$70,$BE,$47,$F2,$9E
	.dc.b	$D7,$E5,$5B,$8F,$E3,$9B,$7F,$8E,$6E,$3F,$E0,$4B,$E5,$5F,$94,$BE
	.dc.b	$4D,$3F,$23,$F3,$7F,$90,$3F,$23,$DF,$63,$24,$C7,$EA,$46,$5F,$F7
	.dc.b	$9E,$7F,$D2,$1F,$91,$DD,$FA,$BF,$99,$F8,$AC,$BF,$C5,$4E,$9A,$FC
	.dc.b	$78,$FC,$63,$63,$79,$FE,$2B,$B7,$F4,$BD,$BF,$A6,$6C,$FE,$A9,$B5
	.dc.b	$F8,$96,$D7,$F8,$9B,$8F,$F8,$7D,$7E,$C2,$F9,$34,$FD,$CF,$C1,$7F
	.dc.b	$03,$3F,$82,$7C,$05,$8C,$93,$1F,$DE,$C6,$5F,$E1,$91,$FE,$1F,$FA
	.dc.b	$77,$7F,$F8,$5D,$B1,$FC,$03,$E3,$71,$32,$3F,$6F,$3A,$AF,$D1,$56
	.dc.b	$5F,$E1,$7E,$87,$F1,$73,$F8,$C7,$A3,$6F,$F8,$8C,$59,$FE,$23,$37
	.dc.b	$5F,$C2,$4D,$AF,$F1,$F7,$2A,$B1,$7F,$C9,$63,$FB,$4B,$E0,$9F,$E3
	.dc.b	$B3,$AB,$FE,$49,$76,$F9,$57,$88,$97,$C4,$F1,$B8,$AF,$1B,$4F,$1D
	.dc.b	$F3,$BE,$8D,$8D,$57,$96,$30,$F5,$AF,$D0,$46,$1E,$B1,$E0,$35,$3F
	.dc.b	$02,$BA,$BD,$27,$D6,$DB,$F5,$D7,$03,$E2,$57,$07,$15,$C1,$B3,$C2
	.dc.b	$3C,$24,$AF,$01,$F2,$8E,$D5,$CE,$35,$5D,$A7,$C5,$7E,$B7,$8C,$7A
	.dc.b	$E7,$67,$84,$B7,$E1,$9D,$C6,$2F,$F0,$F1,$E0,$3A,$96,$BD,$37,$DD
	.dc.b	$F1,$1D,$E3,$99,$1C,$D7,$D0,$8E,$4F,$8A,$79,$3E,$58,$FB,$A6,$8E
	.dc.b	$4F,$F6,$D8,$DD,$FE,$42,$3A,$57,$CC,$BF,$8D,$7C,$2B,$4D,$BE,$33
	.dc.b	$EE,$AE,$13,$EE,$69,$EE,$7C,$4F,$A9,$F2,$BA,$BF,$54,$69,$F5,$47
	.dc.b	$E4,$63,$17,$E2,$8C,$1C,$63,$C8,$6A,$BC,$95,$F4,$E3,$C8,$D9,$F2
	.dc.b	$1C,$4F,$21,$D1,$D1,$E4,$74,$F9,$0F,$95,$1D,$CF,$C7,$46,$AB,$CD
	.dc.b	$3A,$8C,$57,$92,$BE,$63,$F3,$A3,$C3,$3D,$6B,$89,$F8,$D8,$CE,$CF
	.dc.b	$8C,$FF,$0C,$FC,$47,$F8,$69,$B8,$FF,$87,$9B,$5F,$DF,$5B,$87,$F6
	.dc.b	$8F,$A4,$7D,$37,$FA,$F5,$B7,$FE,$2C,$DF,$13,$23,$FC,$3B,$3A,$C3
	.dc.b	$FC,$5F,$59,$7F,$C1,$5E,$CF,$F8,$B0,$FF,$16,$F9,$D6,$FF,$C6,$31
	.dc.b	$67,$FC,$21,$37,$1F,$E9,$13,$6B,$FD,$C9,$A2,$B1,$7F,$C9,$FC,$8F
	.dc.b	$F2,$57,$B7,$FE,$4B,$6D,$FF,$91,$1B,$3F,$E4,$46,$E3,$F0,$7E,$9B
	.dc.b	$E0,$1F,$E4,$2F,$7E,$DF,$BA,$3F,$B2,$EA,$69,$23,$FB,$3F,$0B,$E9
	.dc.b	$7F,$CB,$9E,$C7,$F9,$75,$B3,$FE,$56,$6E,$7F,$E5,$66,$E5,$FE,$09
	.dc.b	$6C,$FF,$46,$6C,$FF,$9A,$76,$7D,$2F,$F9,$60,$FF,$2C,$F6,$DB,$FF
	.dc.b	$2E,$C5,$AF,$F1,$E4,$DC,$3F,$C4,$FF,$4A,$E9,$FF,$7E,$37,$0F,$F3
	.dc.b	$5F,$D8,$E2,$64,$7F,$82,$27,$58,$7F,$83,$1A,$FF,$95,$DA,$FD,$8F
	.dc.b	$4F,$FC,$DA,$7F,$8F,$B0,$9E,$C7,$FA,$33,$4F,$F6,$E6,$1F,$E7,$4A
	.dc.b	$DB,$FE,$82,$9D,$46,$2F,$FC,$AC,$F3,$94,$7F,$A3,$67,$71,$FE,$84
	.dc.b	$F9,$AF,$F4,$71,$FE,$8F,$F9,$2B,$19,$25,$7F,$A2,$23,$2F,$F8,$30
	.dc.b	$EB,$F6,$3E,$67,$FC,$56,$D7,$FC,$EC,$D7,$C6,$1C,$3E,$6F,$A7,$79
	.dc.b	$EF,$57,$FC,$67,$18,$6F,$A3,$FE,$E6,$AF,$F1,$CC,$E5,$FF,$1D,$FD
	.dc.b	$33,$EB,$3B,$F0,$7F,$98,$A3,$13,$FE,$92,$6B,$EB,$1F,$9E,$FF,$9A
	.dc.b	$27,$85,$72,$D7,$D3,$7C,$27,$37,$FE,$9C,$BE,$15,$AB,$E1,$3F,$4A
	.dc.b	$F8,$4E,$63,$83,$AB,$7C,$FE,$A7,$FD,$89,$DB,$D1,$B5,$B9,$D5,$2B
	.dc.b	$FD,$71,$5F,$EC,$65,$EC,$75,$6C,$67,$D9,$C6,$B9,$DE,$4F,$18,$E1
	.dc.b	$EA,$C8,$E8,$F3,$F7,$FA,$7A,$92,$7C,$0F,$7F,$53,$C1,$E9,$49,$F2
	.dc.b	$3E,$3F,$4B,$C9,$DB,$E5,$F8,$B9,$5E,$98,$F4,$79,$BE,$9F,$3A,$4F
	.dc.b	$B6,$3D,$9E,$67,$B7,$CC,$93,$F0,$63,$E0,$71,$48,$F8,$8F,$C3,$F1
	.dc.b	$7E,$27,$8B,$27,$E4,$3F,$1F,$C2,$FA,$9F,$B1,$7E,$4E,$7F,$DE,$2D
	.dc.b	$7F,$45,$D0,$8C,$3F,$BB,$5E,$EF,$E0,$A7,$F9,$B7,$93,$74,$34,$D7
	.dc.b	$BC,$FA,$5E,$1F,$3B,$FC,$4E,$FF,$DD,$2D,$7F,$30,$DC,$BF,$D7,$FF
	.dc.b	$38,$FA,$3F,$07,$CF,$7F,$63,$BF,$F7,$8B,$5F,$D0,$36,$BF,$D8,$EB
	.dc.b	$7F,$37,$93,$2B,$FD,$F3,$AA,$3D,$98,$BE,$2A,$92,$BF,$DF,$F1,$97
	.dc.b	$FE,$00,$F7,$71,$6D,$7C,$58,$C4,$7E,$CD,$B9,$7F,$B6,$BE,$79,$F2
	.dc.b	$BF,$2F,$E7,$9D,$5E,$28,$98,$7F,$20,$BF,$F7,$0D,$6A,$31,$E3,$FF
	.dc.b	$12,$8C,$47,$F0,$1C,$FA,$BE,$8B,$ED,$84,$8C,$10,$FE,$11,$5C,$B0
	.dc.b	$EF,$B7,$C4,$F9,$99,$5E,$03,$F9,$9D,$7C,$D7,$01,$C3,$E3,$8E,$2F
	.dc.b	$69,$FC,$57,$CF,$3E,$13,$F8,$47,$CC,$7C,$DA,$FC,$9B,$71,$FF,$80
	.dc.b	$35,$B9,$EE,$D7,$F9,$03,$63,$75,$DB,$E0,$FF,$31,$8B,$E4,$6F,$C9
	.dc.b	$4C,$7D,$33,$96,$CF,$02,$3E,$EF,$9D,$5B,$1F,$EA,$BF,$3A,$C3,$FA
	.dc.b	$2D,$6A,$73,$1F,$D2,$0E,$D6,$AE,$0F,$F5,$E8,$CC,$62,$15,$CD,$E1
	.dc.b	$3D,$AF,$C5,$18,$39,$61,$FD,$5A,$F9,$57,$43,$86,$1E,$4A,$FE,$DB
	.dc.b	$87,$5A,$DD,$F1,$BF,$43,$6B,$78,$78,$2B,$CD,$7C,$6F,$75,$D4,$F8
	.dc.b	$97,$C5,$F5,$DF,$19,$F1,$C7,$A4,$D9,$C5,$F4,$3E,$0C,$3A,$5F,$C3
	.dc.b	$8F,$EE,$67,$48,$AE,$CB,$1B,$7E,$AF,$C3,$BE,$7B,$71,$F3,$A3,$0F
	.dc.b	$06,$BD,$A5,$18,$8E,$FF,$3D,$C9,$FF,$0B,$B5,$34,$8F,$BE,$FF,$45
	.dc.b	$CB,$F1,$E3,$89,$9D,$91,$98,$E6,$65,$C5,$7C,$FA,$E7,$3F,$2C,$E7
	.dc.b	$AF,$E2,$E3,$2F,$E3,$63,$39,$FF,$83,$B0,$EF,$1C,$4E,$F3,$C6,$32
	.dc.b	$F1,$8C,$CF,$1C,$3C,$71,$E5,$39,$61,$E5,$BE,$35,$B9,$FF,$0C,$61
	.dc.b	$E8,$3C,$2F,$F8,$BA,$B8,$5F,$F0,$E9,$D3,$C5,$DC,$FA,$8F,$E5,$17
	.dc.b	$FE,$49,$7F,$C9,$3F,$3D,$FC,$06,$BF,$8B,$B6,$3F,$C5,$0F,$F9,$3A
	.dc.b	$33,$B7,$19,$4E,$EF,$CA,$DA,$FF,$21,$BC,$67,$9D,$71,$8F,$E7,$F3
	.dc.b	$E0,$E5,$F2,$5B,$E7,$BF,$AF,$4E,$49,$87,$92,$B7,$3A,$8C,$4F,$AB
	.dc.b	$E5,$BB,$1D,$39,$1F,$45,$89,$D1,$2B,$F9,$6C,$65,$F8,$35,$71,$FF
	.dc.b	$3D,$5F,$2B,$E1,$1F,$53,$E7,$9E,$77,$C5,$6C,$FF,$45,$97,$CF,$FF
	.dc.b	$8E,$E3,$FA,$2C,$EB,$1F,$F2,$B3,$6B,$EF,$FB,$DF,$DD,$EB,$78,$7D
	.dc.b	$18,$C5,$FF,$97,$5B,$5C,$3A,$BC,$55,$B7,$EC,$7C,$57,$F9,$EE,$7F
	.dc.b	$17,$F3,$EF,$43,$3E,$77,$DD,$8C,$67,$FD,$69,$B9,$FF,$9A,$FE,$7D
	.dc.b	$F1,$9B,$C3,$F0,$91,$8B,$FE,$C2,$DA,$FF,$39,$7C,$FB,$E7,$B7,$87
	.dc.b	$E2,$E3,$17,$F1,$FE,$7D,$F2,$5F,$84,$6C,$67,$E4,$F1,$3F,$88,$F4
	.dc.b	$DE,$2D,$31,$DB,$6F,$E9,$4B,$E6,$B8,$1C,$7C,$EB,$3F,$5F,$E7,$DF
	.dc.b	$49,$F9,$47,$CB,$C1,$FE,$44,$73,$3B,$E0,$D7,$6E,$FE,$36,$CF,$37
	.dc.b	$F3,$BE,$93,$D1,$BF,$05,$86,$D7,$E1,$FE,$83,$FC,$8C,$DF,$63,$F4
	.dc.b	$B1,$F9,$FE,$CF,$F6,$63,$AD,$7F,$CA,$FF,$40,$F8,$4F,$D5,$B7,$0F
	.dc.b	$C9,$2F,$84,$CC,$EB,$C1,$95,$AF,$02,$E1,$FE,$62,$FA,$07,$C2,$FE
	.dc.b	$D9,$7C,$2D,$8C,$3F,$BA,$F7,$FF,$D7,$D8,$7E,$85,$FF,$33,$B7,$C7
	.dc.b	$D7,$C0,$6F,$E6,$38,$78,$96,$B8,$3F,$C1,$3E,$5B,$9E,$9F,$E7,$47
	.dc.b	$9B,$70,$D6,$1B,$6D,$FE,$C6,$B8,$F1,$FF,$BC,$DE,$2D,$9D,$DF,$3A
	.dc.b	$E1,$7F,$C5,$3E,$4D,$F1,$74,$C7,$C6,$5A,$FD,$F4,$E9,$B3,$A6,$9F
	.dc.b	$E4,$4F,$A0,$FC,$93,$FD,$B9,$B8,$7F,$76,$6B,$FD,$E9,$AF,$FB,$B9
	.dc.b	$DE,$5F,$D5,$0F,$F6,$6F,$C7,$FF,$35,$8C,$D7,$F9,$9F,$E3,$D3,$EB
	.dc.b	$F8,$EF,$67,$C7,$DA,$F4,$B7,$1F,$F5,$0E,$7F,$EF,$89,$F6,$D7,$06
	.dc.b	$9C,$0B,$3C,$2B,$EC,$7D,$04,$9F,$A5,$F4,$16,$BE,$D4,$DD,$3E,$A4
	.dc.b	$DA,$54,$F9,$FF,$D4,$47,$F8,$51,$B1,$FE,$1C,$6B,$FE,$20,$6B,$C8
	.dc.b	$3C,$2B,$E4,$64,$F0,$7F,$AE,$4E,$6F,$FD,$1D,$39,$7F,$15,$F4,$1F
	.dc.b	$93,$F7,$BF,$28,$D8,$C9,$F8,$F8,$CB,$4C,$E3,$F9,$98,$CB,$63,$F1
	.dc.b	$3D,$DF,$8F,$3F,$21,$DF,$63,$24,$9F,$AB,$19,$7F,$CE,$11,$EF,$C7
	.dc.b	$FA,$1B,$F8,$1F,$77,$DB,$E0,$7F,$6E,$E3,$FE,$E3,$7C,$0F,$9D,$19
	.dc.b	$9D,$D7,$F5,$9A,$DE,$DF,$23,$BB,$F4,$9D,$DF,$A5,$6D,$7E,$89,$B8
	.dc.b	$FE,$29,$B5,$C6,$7F,$4F,$5C,$5A,$7E,$DF,$E1,$7F,$80,$1F,$C0,$7E
	.dc.b	$0E,$C6,$49,$87,$F0,$38,$CB,$EB,$1F,$3E,$F0,$3F,$75,$ED,$FF,$86
	.dc.b	$63,$17,$FB,$88,$C4,$EA,$3F,$D7,$33,$AA,$FD,$1D,$7F,$13,$F5,$25
	.dc.b	$7F,$0D,$F5,$6D,$7F,$15,$9B,$A7,$F0,$F9,$B5,$FE,$D9,$F0,$BF,$AD
	.dc.b	$DE,$E3,$91,$FC,$5E,$F7,$1F,$EA,$B9,$DD,$F3,$AD,$12,$79,$1F,$C9
	.dc.b	$6F,$8C,$AB,$8D,$78,$6F,$F9,$65,$6A,$73,$7D,$E7,$35,$D6,$FF,$94
	.dc.b	$46,$EF,$83,$67,$47,$52,$D7,$A6,$77,$18,$AF,$24,$62,$7D,$31,$ED
	.dc.b	$6C,$F3,$69,$CD,$E5,$7E,$89,$F7,$5F,$A2,$B7,$18,$AD,$DF,$78,$F5
	.dc.b	$C7,$D0,$9D,$DF,$32,$56,$E7,$C5,$5E,$53,$8D,$70,$BF,$04,$60,$93
	.dc.b	$C0,$FC,$05,$73,$AF,$9D,$1E,$19,$DC,$60,$F1,$CE,$E3,$D5,$1F,$76
	.dc.b	$7E,$0D,$7D,$18,$F1,$47,$B8,$C9,$5C,$9D,$1C,$B4,$CC,$F8,$8F,$AB
	.dc.b	$1A,$71,$6B,$F2,$3F,$1A,$74,$7B,$EB,$EB,$1D,$1C,$57,$48,$C1,$EF
	.dc.b	$7E,$64,$6A,$7E,$95,$FC,$C8,$C3,$EF,$9F,$1B,$63,$27,$BE,$32,$D3
	.dc.b	$35,$EF,$8C,$B6,$33,$1E,$EB,$F6,$D7,$86,$BC,$D5,$A3,$ED,$9F,$01
	.dc.b	$C9,$F0,$2B,$A9,$CE,$34,$FD,$A3,$DF,$1C,$1C,$47,$C1,$AF,$0B,$F8
	.dc.b	$53,$C1,$3B,$BF,$A5,$1F,$AB,$E7,$7C,$9F,$F8,$5F,$E2,$BF,$C3,$0D
	.dc.b	$9F,$F0,$AB,$6F,$FB,$EB,$71,$FF,$66,$FC,$F3,$E7,$FF,$A3,$7D,$67
	.dc.b	$FA,$FE,$31,$87,$F4,$38,$C4,$EA,$3F,$BB,$D7,$28,$C6,$1F,$E0,$AF
	.dc.b	$7B,$FC,$54,$7F,$8A,$FD,$1B,$7F,$E2,$88,$B3,$FE,$10,$9B,$7F,$DD
	.dc.b	$A6,$D7,$F7,$49,$B5,$FC,$97,$CA,$FF,$23,$7C,$07,$F9,$1D,$B5,$FE
	.dc.b	$41,$6C,$FF,$92,$5B,$87,$C0,$FA,$0F,$F6,$8F,$D6,$30,$F8,$11,$86
	.dc.b	$E1,$FE,$BB,$9D,$3F,$00,$FF,$07,$6A,$AE,$11,$FE,$51,$FA,$1F,$F2
	.dc.b	$D7,$BB,$FE,$5B,$6C,$64,$FF,$2B,$46,$5A,$67,$3F,$F3,$04,$65,$B1
	.dc.b	$FE,$50,$F9,$8F,$F2,$B1,$FE,$57,$F0,$2C,$64,$93,$FE,$3C,$8C,$BF
	.dc.b	$D9,$FE,$81,$F3,$6A,$FF,$14,$4F,$F8,$A6,$73,$18,$8F,$F1,$AC,$61
	.dc.b	$C3,$5E,$2E,$76,$FF,$2B,$78,$3F,$EE,$99,$F0,$5F,$F8,$4F,$E2,$BF
	.dc.b	$CD,$DE,$B4,$C7,$FD,$D1,$ED,$FF,$9E,$1F,$F1,$9C,$66,$FF,$CA,$AF
	.dc.b	$F9,$86,$3F,$23,$3B,$96,$2E,$AF,$F7,$05,$72,$3F,$CE,$F7,$C6,$B7
	.dc.b	$7C,$6F,$84,$FF,$A4,$7E,$82,$5F,$FA,$2F,$E5,$EE,$1F,$E9,$39,$B7
	.dc.b	$FE,$8D,$9B,$5F,$E5,$1F,$9F,$7A,$2F,$63,$E5,$62,$9C,$C6,$17,$A7
	.dc.b	$FE,$DC,$9F,$A1,$1A,$7E,$97,$D6,$7F,$9F,$1E,$6D,$73,$EA,$F6,$51
	.dc.b	$97,$EE,$C3,$E7,$FF,$CC,$DA,$7E,$06,$7F,$CE,$33,$97,$F3,$C6,$27
	.dc.b	$79,$FF,$9B,$7E,$83,$84,$FF,$8C,$6B,$83,$6B,$6F,$13,$98,$9E,$37
	.dc.b	$CA,$B8,$B6,$B8,$D7,$FA,$69,$B5,$C2,$FF,$D3,$D1,$DA,$9F,$41,$9B
	.dc.b	$FC,$1B,$61,$CB,$DA,$FF,$8C,$CF,$FA,$FA,$BD,$F5,$CB,$DB,$E9,$26
	.dc.b	$FB,$78,$DB,$3C,$13,$0E,$37,$F6,$9B,$F0,$5C,$88,$EB,$D9,$EB,$EB
	.dc.b	$C8,$E2,$F8,$3D,$7F,$0F,$AB,$27,$AB,$E4,$F5,$7C,$BD,$BE,$6F,$46
	.dc.b	$57,$8A,$3D,$3E,$6F,$AB,$CE,$93,$E6,$8F,$6F,$99,$EE,$F3,$24,$FA
	.dc.b	$A3,$E0,$EF,$23,$DA,$FC,$4F,$17,$E2,$F8,$B2,$7E,$0B,$F2,$3C,$1F
	.dc.b	$A9,$F4,$1F,$A1,$9F,$9A,$D7,$FD,$25,$D0,$8C,$3F,$ED,$D5,$EF,$7E
	.dc.b	$0B,$37,$91,$49,$7F,$8C,$F7,$DF,$07,$F8,$8F,$A0,$FE,$3B,$3F,$90
	.dc.b	$6B,$FE,$87,$6E,$5F,$9A,$F9,$D7,$D0,$7E,$43,$E8,$3F,$B6,$DF,$E6
	.dc.b	$1A,$FF,$A3,$DB,$5F,$A2,$5B,$F9,$FC,$99,$5F,$A0,$D5,$1F,$9C,$C5
	.dc.b	$F1,$94,$95,$FA,$68,$CB,$FA,$8E,$EE,$2D,$AF,$D2,$C6,$23,$FD,$44
	.dc.b	$DC,$BF,$6B,$F4,$0F,$80,$FD,$5F,$D0,$3A,$BC,$51,$30,$FD,$AC,$E5
	.dc.b	$B3,$97,$FD,$5D,$97,$37,$24,$33,$5B,$C3,$F8,$2D,$F3,$9F,$F5,$B3
	.dc.b	$F1,$9F,$77,$CE,$7C,$33,$CB,$B4,$8C,$90,$E2,$FF,$04,$FA,$07,$C2
	.dc.b	$FF,$01,$F9,$AF,$8F,$5F,$8C,$6E,$3F,$C5,$AB,$73,$FC,$49,$AF,$FB
	.dc.b	$25,$B1,$BA,$ED,$F0,$BF,$C4,$B1,$7C,$9D,$F9,$49,$8F,$F1,$C9,$CB
	.dc.b	$67,$81,$1F,$E4,$3F,$3E,$BF,$EA,$F0,$F9,$F5,$A2,$56,$F2,$C5,$6A
	.dc.b	$75,$5E,$07,$93,$C5,$E7,$85,$9E,$10,$BE,$8F,$03,$BE,$1E,$49,$DD
	.dc.b	$F5,$6D,$F3,$8E,$D7,$CA,$7A,$B6,$3C,$47,$12,$CF,$1B,$EF,$87,$B2
	.dc.b	$79,$E1,$E1,$77,$39,$3D,$F5,$C5,$F5,$4E,$DE,$67,$AC,$F8,$27,$12
	.dc.b	$99,$BF,$33,$C0,$FF,$10,$BF,$11,$B1,$F5,$9B,$7F,$DE,$A5,$F1,$2F
	.dc.b	$A0,$C3,$6B,$FC,$51,$3A,$9C,$9F,$DD,$61,$F4,$14,$8F,$F7,$96,$E7
	.dc.b	$F6,$30,$FC,$FF,$A5,$EB,$AF,$C3,$57,$B6,$73,$1F,$04,$C9,$F0,$DC
	.dc.b	$D7,$E0,$A3,$2F,$5D,$BF,$C0,$4E,$63,$CA,$71,$8C,$EB,$FE,$43,$BF
	.dc.b	$04,$62,$7F,$C9,$31,$8C,$3C,$D1,$CA,$FB,$46,$27,$53,$E7,$35,$7F
	.dc.b	$E1,$4B,$E9,$1F,$E3,$C7,$FC,$70,$DA,$FF,$05,$BE,$A9,$E1,$1F,$E0
	.dc.b	$EC,$3E,$19,$FE,$0E,$AE,$35,$B3,$FD,$32,$F4,$FA,$0F,$DE,$D7,$F8
	.dc.b	$59,$B3,$FE,$4C,$8F,$D6,$47,$ED,$5D,$6D,$C4,$8F,$19,$A7,$13,$F7
	.dc.b	$93,$FC,$CE,$3B,$72,$7F,$94,$F0,$7D,$07,$F5,$29,$C9,$2F,$90,$BE
	.dc.b	$4F,$F6,$A9,$E0,$EA,$33,$5F,$E3,$65,$F2,$DE,$F3,$C1,$C9,$FE,$75
	.dc.b	$9D,$46,$70,$F8,$4E,$BE,$03,$18,$75,$8B,$5F,$E4,$91,$7C,$EF,$E3
	.dc.b	$CF,$D8,$F9,$F7,$9D,$4F,$81,$F4,$1F,$E3,$8A,$FF,$22,$CE,$9B,$3F
	.dc.b	$80,$F7,$FF,$B6,$C6,$0F,$F2,$5B,$8C,$BE,$2B,$72,$FF,$2A,$B6,$7F
	.dc.b	$92,$5F,$5A,$DF,$57,$E3,$3E,$23,$FD,$0F,$1F,$CC,$7E,$81,$E8,$67
	.dc.b	$CE,$FC,$8C,$63,$1F,$F2,$DB,$6B,$EA,$FD,$03,$E2,$F7,$87,$E6,$23
	.dc.b	$17,$FE,$62,$6D,$7D,$CF,$A0,$7C,$E6,$F0,$FC,$FC,$62,$FF,$CA,$BF
	.dc.b	$40,$F9,$5F,$C4,$3C,$B8,$7E,$93,$99,$DD,$FE,$37,$D2,$7C,$FD,$CB
	.dc.b	$0D,$AF,$F3,$0C,$FE,$19,$BF,$15,$C7,$9F,$93,$F9,$7F,$9F,$7C,$95
	.dc.b	$31,$E0,$DB,$FF,$35,$CF,$F8,$D1,$BE,$AF,$1F,$53,$93,$FA,$1F,$4D
	.dc.b	$F0,$14,$C7,$C0,$5B,$FF,$3A,$FC,$FF,$FB,$BD,$BF,$4B,$FA,$98,$C7
	.dc.b	$3F,$F9,$31,$47,$EB,$7E,$81,$F0,$77,$E0,$F0,$DC,$3F,$D0,$3A,$7F
	.dc.b	$B0,$CF,$F2,$8B,$8F,$0B,$F6,$D1,$8C,$BF,$CB,$3F,$41,$FE,$DF,$AF
	.dc.b	$EF,$1D,$48,$C4,$FF,$B5,$1F,$C8,$47,$E9,$DC,$AF,$79,$F3,$9F,$C0
	.dc.b	$DB,$7F,$E9,$45,$F3,$F9,$9D,$78,$92,$B5,$F2,$B7,$2F,$F3,$6F,$CA
	.dc.b	$3E,$33,$F8,$73,$6B,$FD,$3E,$FF,$A7,$5B,$1A,$23,$AF,$8D,$B5,$D2
	.dc.b	$6E,$9F,$C0,$6F,$94,$E6,$FF,$D7,$FF,$40,$FA,$1F,$E7,$2D,$8F,$F5
	.dc.b	$7F,$CE,$7B,$AB,$F8,$EC,$E5,$B5,$FE,$CE,$7F,$C3,$91,$97,$FC,$3C
	.dc.b	$1C,$3A,$8E,$0B,$06,$1F,$EE,$EC,$3E,$0F,$C7,$FC,$4F,$90,$F8,$AD
	.dc.b	$8C,$9F,$1A,$32,$FA,$38,$3F,$9C,$67,$F5,$23,$D7,$F4,$38,$99,$1B
	.dc.b	$32,$5C,$BF,$D4,$35,$C5,$B3,$D2,$B8,$C7,$F4,$0E,$17,$C3,$7F,$62
	.dc.b	$6E,$1F,$6A,$30,$47,$E2,$CF,$F9,$2E,$76,$E1,$AF,$CA,$6B,$F9,$AC
	.dc.b	$1F,$43,$EE,$F6,$FD,$ED,$BF,$B2,$DA,$F5,$B7,$6F,$C2,$77,$FE,$40
	.dc.b	$FC,$8F,$7D,$BF,$C8,$C5,$BF,$CB,$CD,$AE,$35,$A3,$F2,$F1,$97,$FC
	.dc.b	$BD,$97,$F3,$DA,$F7,$7A,$5F,$53,$E9,$3E,$33,$2B,$F6,$A7,$07,$4E
	.dc.b	$27,$8B,$FA,$4E,$DF,$D3,$F7,$7E,$A1,$B1,$93,$F5,$31,$96,$9F,$9C
	.dc.b	$F8,$2F,$C2,$9F,$86,$F8,$2B,$19,$25,$FE,$0A,$32,$FE,$12,$6E,$9F
	.dc.b	$DB,$78,$3F,$C9,$F7,$C2,$0F,$F9,$FF,$E8,$BF,$A0,$5F,$F8,$19,$B3
	.dc.b	$FB,$5F,$93,$FF,$14,$C6,$2B,$87,$A5,$83,$F8,$2C,$D8,$FD,$AF,$D1
	.dc.b	$7E,$78,$FC,$FF,$D1,$58,$C9,$31,$FC,$DC,$65,$FE,$23,$F2,$BF,$C4
	.dc.b	$3E,$33,$F8,$8B,$6B,$F8,$8B,$67,$71,$FC,$52,$76,$D3,$65,$8D,$9F
	.dc.b	$C7,$A3,$6D,$8D,$B4,$DE,$7F,$C5,$E7,$6D,$9D,$B6,$77,$1F,$CF,$3B
	.dc.b	$7F,$DD,$0E,$8C,$57,$26,$99,$9F,$E2,$13,$C1,$E6,$62,$78,$36,$78
	.dc.b	$57,$F1,$69,$CB,$63,$35,$CE,$3F,$98,$5F,$C9,$6C,$FC,$56,$DF,$C1
	.dc.b	$69,$ED,$2C,$7A,$23,$CB,$1D,$DB,$5D,$E7,$E1,$57,$76,$9D,$CB,$3D
	.dc.b	$DB,$5D,$DA,$77,$7C,$31,$F0,$4E,$F7,$F0,$2F,$C1,$1F,$2A,$FC,$0D
	.dc.b	$BF,$03,$4F,$01,$6F,$C0,$D8,$F0,$1E,$27,$C1,$1E,$23,$E0,$16,$BC
	.dc.b	$2F,$F4,$8A,$F0,$B6,$BC,$2D,$AF,$0B,$4F,$09,$63,$C3,$1E,$33,$C2
	.dc.b	$F8,$E3,$C2,$DA,$F0,$8A,$F0,$B6,$3C,$2D,$AF,$0B,$6B,$C2,$D9,$F0
	.dc.b	$C7,$8E,$7C,$27,$8D,$F0,$B6,$7C,$35,$E0,$25,$78,$1B,$3E,$06,$D7
	.dc.b	$81,$B3,$E0,$AF,$15,$78,$2B,$C4,$78,$0B,$1E,$09,$EC,$49,$EC,$D9
	.dc.b	$EC,$D3,$B5,$78,$27,$B5,$78,$2B,$B3,$4E,$D1,$D0,$91,$D1,$B5,$D0
	.dc.b	$EC,$F4,$AE,$D3,$D1,$B5,$D2,$39,$12,$B9,$57,$43,$93,$F9,$A5,$F4
	.dc.b	$1F,$E6,$C8,$FE,$A9,$18,$F4,$3F,$C1,$FA,$FF,$B8,$AB,$53,$FE,$C4
	.dc.b	$6D,$FF,$44,$81,$C6,$FF,$C5,$F5,$C6,$FF,$C8,$7E,$D4,$9F,$F2,$1F
	.dc.b	$9D,$6F,$FC,$93,$37,$0F,$F1,$F4,$DA,$FF,$03,$7A,$FF,$E2,$2F,$23
	.dc.b	$FC,$46,$D8,$FF,$1E,$B7,$0F,$F0,$FB,$6F,$FA,$E3,$67,$99,$FE,$44
	.dc.b	$AF,$77,$D0,$48,$FF,$2B,$FC,$BD,$C3,$E0,$4D,$C3,$FC,$A5,$36,$BF
	.dc.b	$C3,$5E,$57,$F8,$DB,$D7,$FF,$1B,$B6,$3F,$CA,$2D,$C3,$FC,$66,$DB
	.dc.b	$FF,$2E,$7C,$BF,$F9,$A0,$FF,$34,$F6,$DB,$FF,$34,$C5,$BF,$F3,$7C
	.dc.b	$DA,$E3,$5A,$3F,$CD,$F1,$97,$EA,$CB,$E8,$3F,$D0,$ED,$E8,$FF,$26
	.dc.b	$FD,$27,$FB,$B6,$B7,$5A,$13,$FE,$5E,$AE,$13,$FE,$40,$69,$FE,$42
	.dc.b	$9D,$C7,$F9,$13,$0F,$F3,$1F,$C0,$FF,$A1,$0E,$D5,$FE,$67,$F6,$9E
	.dc.b	$A6,$6F,$FD,$0F,$56,$BF,$CD,$91,$FE,$A0,$AD,$BA,$3F,$CF,$BF,$43
	.dc.b	$FE,$66,$3F,$CC,$FF,$2B,$63,$24,$AF,$F3,$1C,$65,$FF,$32,$CD,$C7
	.dc.b	$FC,$45,$36,$BE,$4F,$D2,$31,$F8,$4D,$8F,$84,$DC,$3F,$CC,$6D,$7F
	.dc.b	$D1,$3F,$44,$F4,$7D,$F1,$FE,$68,$8C,$69,$F2,$D7,$A2,$F9,$3C,$5F
	.dc.b	$0B,$FF,$49,$FD,$0B,$E8,$3E,$EE,$DF,$B3,$94,$E5,$FD,$EB,$5F,$96
	.dc.b	$7E,$E3,$E8,$78,$4F,$19,$D3,$6B,$AE,$BF,$8A,$9C,$BF,$8C,$AE,$2E
	.dc.b	$7D,$6D,$E1,$FC,$12,$BF,$D6,$8F,$8A,$32,$F9,$BE,$6E,$CE,$A3,$CA
	.dc.b	$7A,$A7,$26,$A6,$C7,$32,$1C,$25,$1E,$2C,$5F,$3F,$FE,$E0,$AF,$F3
	.dc.b	$94,$63,$D1,$D4,$E6,$7E,$2C,$F0,$F6,$33,$EC,$49,$E3,$48,$E3,$EB
	.dc.b	$72,$F5,$E5,$69,$EB,$EA,$F6,$F4,$BB,$FA,$72,$B9,$C7,$8B,$D1,$F1
	.dc.b	$FA,$32,$7B,$C7,$9B,$CD,$F3,$F9,$52,$7C,$71,$EA,$F2,$A4,$79,$9F
	.dc.b	$6F,$8F,$EE,$F2,$24,$FA,$9F,$83,$E3,$7C,$2F,$0E,$4F,$B9,$F8,$BE
	.dc.b	$1F,$C6,$FA,$DF,$99,$FB,$1F,$FB,$93,$9F,$FA,$10,$71,$D9,$FD,$12
	.dc.b	$3E,$8C,$F0,$F0,$38,$FC,$B4,$CB,$E9,$E1,$F9,$F6,$FC,$4F,$6F,$8A
	.dc.b	$91,$83,$FA,$BF,$C0,$BE,$0F,$F1,$7E,$93,$86,$9F,$8E,$9D,$57,$F5
	.dc.b	$EF,$9D,$7C,$47,$E4,$BE,$83,$F8,$EC,$FF,$62,$6B,$F9,$29,$DB,$FD
	.dc.b	$9B,$E8,$1E,$9F,$E6,$9F,$F1,$44,$67,$DB,$C4,$F9,$39,$64,$FF,$6E
	.dc.b	$D1,$1F,$A2,$C3,$F2,$F3,$E5,$F1,$F0,$47,$FB,$AF,$C1,$F0,$6D,$7E
	.dc.b	$AA,$31,$1F,$A8,$6F,$37,$F0,$3F,$A0,$7C,$D7,$EC,$3E,$81,$F0,$1E
	.dc.b	$18,$98,$7F,$80,$A7,$2D,$9C,$BF,$0E,$79,$44,$8F,$99,$1A,$BF,$E2
	.dc.b	$55,$CA,$7F,$76,$FC,$87,$E9,$7C,$7F,$C1,$3C,$9B,$48,$C9,$0E,$2F
	.dc.b	$F0,$8F,$A0,$79,$3F,$BD,$F9,$9F,$93,$5F,$B8,$6E,$3F,$C8,$EB,$73
	.dc.b	$FE,$1E,$6B,$FC,$39,$B1,$BA,$ED,$E4,$7F,$16,$C5,$DB,$7B,$A6,$3F
	.dc.b	$E2,$89,$CB,$67,$81,$1F,$F1,$6F,$D0,$2D,$8F,$F8,$A3,$E7,$D8,$7F
	.dc.b	$76,$AD,$4E,$63,$FB,$C9,$DA,$D5,$C1,$FF,$18,$C6,$63,$10,$AE,$6F
	.dc.b	$09,$ED,$7E,$28,$C1,$CB,$0F,$F0,$3D,$F2,$AE,$87,$0C,$3C,$95,$FE
	.dc.b	$40,$C3,$AD,$6E,$F8,$DF,$A1,$B5,$BC,$3C,$15,$E6,$BE,$37,$BA,$EA
	.dc.b	$7C,$4B,$E2,$FA,$EF,$8C,$F8,$E3,$D2,$6C,$E2,$FA,$1F,$06,$1D,$2F
	.dc.b	$E1,$C7,$F8,$90,$E9,$1F,$E1,$B8,$CC,$75,$34,$BC,$B7,$D0,$6D,$B3
	.dc.b	$CC,$FA,$0E,$1E,$2D,$75,$F4,$1F,$E2,$97,$FC,$D2,$D4,$D2,$3E,$EB
	.dc.b	$FD,$0F,$8B,$F1,$A3,$89,$9D,$BF,$CE,$9A,$FE,$0A,$32,$FE,$12,$B8
	.dc.b	$C7,$C8,$33,$9F,$F8,$2A,$B8,$BE,$68,$E9,$7D,$DF,$29,$83,$91,$8B
	.dc.b	$ED,$5C,$A3,$E7,$61,$F7,$30,$F0,$D7,$96,$78,$61,$E1,$C3,$B5,$EF
	.dc.b	$0E,$55,$FA,$87,$FC,$92,$D9,$FF,$4C,$BE,$A3,$FA,$84,$6B,$0F,$79
	.dc.b	$FE,$1E,$78,$C6,$CE,$47,$F8,$43,$E7,$FF,$7F,$5F,$EA,$46,$C6,$4F
	.dc.b	$0B,$8A,$FF,$18,$36,$7F,$CD,$ED,$9F,$F5,$2B,$FE,$A6,$8E,$9B,$72
	.dc.b	$E4,$91,$FE,$14,$FA,$0F,$EA,$AF,$86,$33,$7E,$57,$FC,$D3,$82,$B7
	.dc.b	$39,$AF,$47,$CC,$79,$8D,$F2,$3E,$83,$13,$A2,$57,$F2,$98,$CB,$F0
	.dc.b	$EA,$E1,$FD,$FA,$F9,$5F,$08,$FF,$28,$7C,$F3,$E7,$FE,$AB,$67,$FC
	.dc.b	$5F,$F4,$0F,$89,$CD,$49,$FF,$68,$FC,$1F,$F6,$C7,$FD,$A7,$7C,$73
	.dc.b	$FF,$21,$CE,$5B,$D7,$F8,$28,$C6,$5E,$2C,$31,$3F,$86,$F4,$BF,$CE
	.dc.b	$93,$FE,$65,$FA,$07,$C6,$67,$E1,$FF,$17,$18,$D3,$FD,$C6,$DC,$FF
	.dc.b	$CA,$7F,$40,$F6,$37,$87,$E4,$A3,$17,$FE,$EA,$6D,$7F,$96,$BE,$81
	.dc.b	$E9,$6F,$0F,$CC,$C6,$2F,$FC,$A7,$F4,$0F,$6B,$EE,$BC,$B5,$FF,$16
	.dc.b	$B9,$9D,$DF,$E0,$BD,$37,$AB,$72,$C3,$6B,$FC,$BF,$2E,$8E,$1F,$17
	.dc.b	$C5,$B3,$FE,$64,$FA,$0F,$F7,$EB,$FE,$39,$6E,$DF,$D6,$1F,$EA,$F1
	.dc.b	$BF,$91,$FC,$9C,$7F,$8C,$E3,$97,$B1,$8F,$5F,$2F,$E6,$3D,$47,$CC
	.dc.b	$D3,$1F,$33,$6F,$FC,$EB,$F4,$0F,$94,$FC,$FA,$FA,$0B,$18,$7F,$47
	.dc.b	$F4,$0F,$9C,$BF,$39,$86,$E1,$FE,$80,$D3,$FD,$93,$C1,$FE,$7E,$5F
	.dc.b	$43,$FA,$C7,$3D,$3F,$E5,$C7,$35,$FE,$88,$FA,$1F,$F6,$ED,$FF,$A1
	.dc.b	$9B,$5F,$E6,$27,$F3,$0D,$F9,$9F,$DB,$FC,$1B,$E4,$29,$8F,$90,$B8
	.dc.b	$FF,$A5,$25,$E2,$F0,$38,$FC,$55,$9F,$F4,$CF,$CE,$3E,$83,$F8,$1B
	.dc.b	$63,$3F,$43,$89,$91,$C1,$DC,$63,$6F,$F4,$FD,$71,$6C,$EA,$B8,$1F
	.dc.b	$E7,$B1,$13,$E8,$5F,$2D,$FE,$D1,$8C,$88,$CC,$FF,$65,$8C,$4E,$DE
	.dc.b	$4D,$73,$1F,$EB,$6F,$A3,$FF,$6D,$BE,$18,$DB,$9F,$97,$F8,$3F,$37
	.dc.b	$F0,$BE,$6E,$CF,$C6,$6D,$7B,$9B,$C7,$FE,$98,$D3,$FD,$E9,$5F,$EC
	.dc.b	$4F,$A2,$71,$7F,$2E,$6C,$67,$E6,$71,$32,$38,$3B,$8C,$63,$FE,$CA
	.dc.b	$AE,$2D,$9D,$57,$03,$D9,$1F,$85,$F8,$BF,$85,$CA,$69,$27,$E5,$46
	.dc.b	$5F,$C1,$FD,$1F,$F6,$48,$C5,$7D,$C8,$F5,$7C,$D7,$AF,$E0,$6D,$FB
	.dc.b	$DB,$5E,$76,$EB,$FE,$BC,$CF,$FB,$4D,$7F,$AE,$F5,$FD,$CB,$F9,$2F
	.dc.b	$13,$F2,$3E,$27,$E4,$99,$19,$3F,$2F,$18,$C7,$F1,$F3,$70,$FA,$53
	.dc.b	$6B,$F1,$3A,$7F,$3F,$BF,$F1,$B3,$5C,$FA,$DF,$18,$93,$F8,$7F,$7F
	.dc.b	$E1,$47,$F6,$6E,$5F,$F5,$0C,$FE,$A3,$C1,$FD,$3F,$85,$FA,$86,$C6
	.dc.b	$4F,$D6,$C6,$5A,$7E,$17,$E2,$3F,$08,$7E,$17,$E1,$AC,$64,$99,$7E
	.dc.b	$06,$32,$FE,$0E,$6E,$BF,$36,$6D,$7E,$A3,$E8,$FF,$9F,$5F,$ED,$3D
	.dc.b	$BF,$F0,$FC,$62,$FF,$31,$18,$F5,$BF,$80,$B6,$FF,$82,$BF,$99,$F9
	.dc.b	$99,$1F,$9D,$F9,$AB,$8F,$E5,$66,$E1,$F7,$E6,$D7,$DF,$F9,$4F,$E1
	.dc.b	$BF,$21,$FC,$39,$B1,$FC,$49,$B8,$7F,$0A,$6D,$FE,$B5,$B3,$CC,$FE
	.dc.b	$33,$5C,$DA,$73,$2D,$73,$6D,$73,$69,$CE,$3F,$8C,$CF,$36,$C7,$36
	.dc.b	$9C,$CB,$3C,$DB,$1C,$E3,$F8,$24,$F3,$6C,$73,$6C,$F3,$69,$CC,$B1
	.dc.b	$CC,$FD,$44,$F3,$69,$CC,$B3,$CD,$B3,$CD,$A7,$38,$FF,$64,$F3,$7F
	.dc.b	$BE,$A3,$94,$62,$BB,$36,$33,$4A,$E1,$3D,$CD,$16,$35,$5C,$9B,$5C
	.dc.b	$AB,$8C,$79,$6B,$8B,$4D,$BC,$E3,$73,$E9,$9C,$9B,$8C,$CF,$B2,$FF
	.dc.b	$0E,$DA,$FB,$ED,$3E,$B9,$6B,$E7,$B6,$3E,$3B,$4F,$2B,$E1,$AE,$F3
	.dc.b	$F0,$EB,$BB,$67,$BB,$6B,$BB,$4E,$E5,$8E,$ED,$8E,$E7,$85,$F7,$C7
	.dc.b	$86,$FE,$65,$F8,$85,$F8,$9B,$3E,$26,$DF,$89,$B7,$E2,$6C,$F8,$9B
	.dc.b	$3E,$29,$F2,$1E,$27,$C9,$1E,$1A,$F0,$12,$BC,$0D,$9F,$03,$6B,$C0
	.dc.b	$D9,$F0,$36,$7C,$15,$E2,$AF,$01,$E2,$7C,$15,$D8,$93,$D9,$B3,$D9
	.dc.b	$A7,$62,$C7,$6A,$F0,$4F,$6A,$F0,$1D,$A7,$A1,$23,$A3,$6B,$A3,$4E
	.dc.b	$8F,$68,$E9,$5D,$AB,$A4,$72,$3F,$23,$4F,$A3,$FF,$33,$C3,$89,$F4
	.dc.b	$DF,$ED,$89,$CC,$63,$1F,$F0,$0C,$62,$75,$EA,$FF,$59,$7F,$C7,$BE
	.dc.b	$FC,$AF,$F2,$0F,$A9,$72,$FF,$1C,$CD,$CB,$FC,$39,$36,$BF,$C6,$F0
	.dc.b	$EC,$FF,$46,$46,$1F,$F2,$7B,$FE,$21,$F7,$3F,$C4,$1E,$67,$F8,$85
	.dc.b	$B1,$93,$FC,$4D,$19,$69,$9C,$7F,$C4,$11,$EF,$F5,$3F,$C7,$9A,$FF
	.dc.b	$9E,$EB,$FC,$84,$DA,$FF,$23,$9F,$E4,$C7,$0E,$1D,$62,$9F,$5F,$D3
	.dc.b	$7A,$29,$39,$79,$5F,$FB,$9A,$75,$3F,$E1,$BF,$3B,$FC,$79,$ED,$FF
	.dc.b	$8F,$5B,$87,$F9,$05,$B8,$7F,$8E,$5B,$5F,$E4,$7F,$A5,$FF,$37,$9F
	.dc.b	$E7,$0E,$FB,$7F,$E7,$48,$B7,$FE,$6C,$9B,$5C,$6B,$47,$F9,$EA,$32
	.dc.b	$FF,$9A,$3E,$97,$FD,$1B,$5F,$11,$AE,$7D,$3F,$F3,$34,$62,$3F,$CE
	.dc.b	$6F,$FA,$4A,$7E,$FF,$2F,$FB,$DA,$7E,$2C,$71,$3F,$CD,$9F,$15,$FE
	.dc.b	$88,$F8,$59,$1F,$E6,$DC,$9E,$FF,$FA,$8A,$3F,$C8,$13,$AB,$FF,$43
	.dc.b	$BF,$3E,$A5,$6E,$7F,$AA,$CF,$F9,$A7,$E7,$A5,$FF,$9B,$7E,$62,$D7
	.dc.b	$F9,$96,$6E,$1F,$E4,$F9,$B5,$F9,$59,$B5,$F7,$B9,$95,$88,$7D,$23
	.dc.b	$D2,$7C,$4F,$FA,$3E,$30,$DF,$51,$F2,$D9,$AD,$DF,$E4,$87,$D3,$3D
	.dc.b	$7F,$F4,$C4,$61,$B1,$98,$FF,$4E,$35,$E5,$0F,$A5,$FF,$51,$EF,$F9
	.dc.b	$59,$E9,$18,$DF,$FD,$4F,$5B,$9F,$F4,$1F,$D2,$EE,$F8,$4E,$67,$ED
	.dc.b	$9F,$EB,$27,$11,$AB,$E9,$3F,$EB,$6A,$E5,$7D,$BD,$AC,$CF,$B5,$B8
	.dc.b	$7F,$AE,$A3,$FD,$26,$BE,$77,$33,$27,$2F,$08,$C4,$79,$0E,$86,$8E
	.dc.b	$6F,$93,$17,$D2,$FF,$B8,$E7,$F2,$0D,$7D,$B1,$AF,$53,$84,$FF,$AD
	.dc.b	$27,$7E,$E6,$08,$F2,$F6,$77,$EC,$C8,$EA,$F2,$FA,$5E,$7E,$AF,$4F
	.dc.b	$56,$57,$8A,$3B,$FA,$5E,$0F,$4A,$4F,$9A,$3C,$7E,$A7,$93,$CC,$93
	.dc.b	$EA,$8F,$3F,$99,$23,$DA,$FA,$BC,$8F,$5F,$93,$27,$E0,$BE,$EF,$A5
	.dc.b	$F7,$F8,$52,$7E,$2B,$F0,$BC,$1F,$87,$F4,$BF,$2F,$E9,$7F,$94,$F3
	.dc.b	$FE,$CD,$BD,$7F,$D0,$39,$D5,$BF,$03,$87,$CC,$CC,$FF,$80,$61,$FA
	.dc.b	$46,$FC,$7F,$5F,$90,$91,$83,$FD,$AF,$F0,$EF,$89,$FC,$5F,$AA,$E2
	.dc.b	$A7,$E3,$A7,$55,$FE,$DE,$FA,$37,$C7,$FE,$4B,$E9,$5E,$9D,$3F,$2B
	.dc.b	$3A,$AF,$F7,$2F,$D2,$BD,$5F,$CC,$BF,$E2,$88,$CF,$BD,$89,$F1,$F3
	.dc.b	$C9,$FF,$76,$EB,$FC,$21,$F1,$78,$FD,$88,$E0,$FF,$79,$FC,$46,$DF
	.dc.b	$CE,$37,$A7,$F7,$4D,$E9,$4C,$8C,$38,$FA,$57,$C3,$7E,$C3,$E9,$5F
	.dc.b	$07,$EE,$89,$5F,$DD,$B0,$E6,$AB,$6E,$4E,$EE,$1F,$D8,$70,$76,$7E
	.dc.b	$75,$77,$B9,$18,$21,$FB,$7A,$E5,$97,$F1,$57,$E7,$47,$D7,$FA,$1F
	.dc.b	$A6,$7A,$77,$91,$92,$1C,$DF,$E1,$BF,$4A,$F2,$FF,$80,$7C,$FF,$F0
	.dc.b	$BA,$FE,$9C,$D7,$F8,$44,$ED,$F7,$07,$0E,$A7,$A9,$DB,$CA,$FE,$37
	.dc.b	$83,$BE,$F8,$12,$F9,$CE,$5B,$3C,$08,$F8,$3E,$95,$7F,$E0,$C8,$7D
	.dc.b	$22,$FE,$AD,$6B,$6F,$C0,$CE,$6B,$F0,$75,$DD,$E2,$FD,$C3,$11,$B8
	.dc.b	$57,$37,$67,$E3,$1C,$46,$E7,$89,$CF,$0F,$C3,$C7,$39,$DC,$F9,$B0
	.dc.b	$F5,$4F,$1A,$DC,$F1,$BF,$4B,$BC,$2C,$78,$6F,$8C,$EE,$3C,$A7,$C3
	.dc.b	$AE,$6E,$EB,$8B,$EC,$AE,$2F,$C4,$9D,$BC,$4D,$E1,$E8,$79,$1F,$DC
	.dc.b	$DF,$88,$FF,$91,$9A,$F9,$CD,$79,$AF,$A5,$FE,$F5,$1F,$32,$30,$EA
	.dc.b	$33,$29,$CC,$F0,$FA,$5F,$EF,$E9,$A4,$7F,$C0,$6F,$BB,$1F,$CF,$72
	.dc.b	$FC,$09,$FE,$07,$39,$DB,$FC,$A9,$AF,$DF,$9C,$9F,$2A,$F8,$BF,$30
	.dc.b	$CC,$7E,$16,$32,$F1,$8F,$21,$C6,$7A,$EB,$FE,$19,$BF,$06,$1E,$3A
	.dc.b	$DD,$F3,$BD,$C7,$F9,$C2,$B6,$F8,$4D,$DF,$F8,$62,$77,$1F,$C5,$27
	.dc.b	$75,$CE,$FF,$6F,$87,$A8,$FF,$0F,$57,$AB,$0E,$53,$AC,$3F,$C9,$0E
	.dc.b	$27,$FC,$10,$6C,$FF,$4B,$C6,$9F,$E5,$BF,$4B,$FB,$D9,$FE,$0A,$D8
	.dc.b	$FF,$08,$3F,$CC,$A3,$3A,$F1,$94,$EF,$0F,$F1,$93,$6B,$F9,$7B,$C6
	.dc.b	$79,$D7,$18,$FE,$73,$3E,$0E,$5F,$25,$BE,$97,$FA,$DC,$E4,$98,$79
	.dc.b	$2B,$73,$A8,$C4,$FA,$BE,$7B,$B1,$D3,$93,$FC,$F1,$87,$9A,$33,$3B
	.dc.b	$3F,$94,$FC,$1E,$2F,$E0,$CD,$AF,$F3,$A5,$F2,$BF,$6C,$FC,$3F,$A3
	.dc.b	$7C,$7D,$3D,$7F,$4A,$FA,$5C,$5F,$F3,$2A,$DB,$FC,$D3,$E0,$FF,$B9
	.dc.b	$39,$BF,$9D,$18,$CF,$FC,$84,$DC,$FF,$CC,$0D,$AF,$EB,$B7,$D2,$B7
	.dc.b	$D7,$F5,$FD,$47,$C3,$7D,$9F,$A5,$7C,$7E,$63,$0C,$C7,$FA,$3F,$D2
	.dc.b	$FF,$B7,$31,$FE,$E7,$7C,$27,$3E,$D7,$E0,$EB,$54,$FA,$57,$A7,$9A
	.dc.b	$FC,$44,$62,$FF,$A9,$FD,$A7,$C5,$6F,$17,$E0,$BE,$9B,$FD,$9B,$1F
	.dc.b	$DE,$27,$58,$FF,$98,$FF,$0A,$FE,$1D,$3F,$D7,$1B,$1F,$E5,$EF,$55
	.dc.b	$F2,$DF,$97,$9C,$F2,$FF,$A1,$ED,$F3,$DF,$9B,$69,$FE,$B7,$BF,$F1
	.dc.b	$93,$FD,$A0,$CE,$1F,$E7,$3F,$7D,$F1,$BF,$A3,$6C,$FF,$AF,$FE,$9D
	.dc.b	$F6,$F9,$AD,$7A,$92,$B5,$F2,$F7,$1F,$F3,$F7,$DA,$7F,$A9,$67,$FB
	.dc.b	$BB,$63,$FC,$F3,$4F,$9E,$FD,$94,$E7,$6F,$F5,$17,$E1,$1F,$DA,$F0
	.dc.b	$FE,$FE,$DC,$3F,$D0,$3F,$55,$FE,$A9,$9C,$F9,$7F,$BB,$A6,$74,$FD
	.dc.b	$9F,$D4,$FF,$B8,$70,$FF,$19,$CE,$AF,$FC,$17,$E7,$7E,$31,$BD,$FF
	.dc.b	$BA,$F8,$27,$C7,$FF,$07,$AD,$61,$FE,$9B,$F1,$FE,$5E,$DF,$EB,$96
	.dc.b	$E3,$FE,$40,$87,$D4,$3D,$FF,$E2,$15,$AB,$FF,$22,$BF,$E6,$5B,$FF
	.dc.b	$0E,$B6,$3F,$D3,$FF,$08,$F9,$0E,$3D,$5F,$C6,$7D,$9F,$C2,$E1,$FE
	.dc.b	$24,$6D,$F0,$9D,$77,$7F,$1F,$FA,$67,$C9,$53,$F8,$BF,$CC,$FB,$B1
	.dc.b	$FF,$16,$CE,$9B,$3B,$7F,$C9,$DB,$AB,$86,$3F,$EB,$DB,$E0,$DB,$FE
	.dc.b	$7E,$FF,$98,$A3,$2F,$F9,$68,$38,$7F,$CB,$AD,$54,$FA,$1F,$CA,$C7
	.dc.b	$FB,$09,$FF,$7E,$61,$E4,$73,$28,$FA,$51,$A9,$E1,$DB,$FC,$DA,$31
	.dc.b	$5B,$9E,$F1,$CA,$B8,$FC,$CF,$C6,$F0,$A4,$7C,$2C,$FE,$A3,$63,$37
	.dc.b	$FE,$D6,$9F,$BD,$87,$C7,$DF,$2F,$FB,$FF,$95,$F0,$9F,$D4,$DB,$8F
	.dc.b	$AE,$30,$47,$FC,$C3,$1A,$C3,$E1,$4E,$5F,$F7,$74,$65,$F9,$CD,$7F
	.dc.b	$D1,$2D,$7F,$D2,$AD,$AF,$B9,$F2,$BF,$84,$F6,$FF,$0A,$DC,$3E,$03
	.dc.b	$6F,$DC,$DA,$F9,$47,$D7,$FA,$5F,$9C,$7C,$FF,$A5,$B7,$F5,$E2,$DF
	.dc.b	$DB,$9B,$5F,$17,$D3,$F8,$87,$E5,$BE,$5E,$E1,$F0,$A2,$DF,$CF,$9B
	.dc.b	$5C,$6B,$47,$D7,$8C,$BF,$84,$F1,$7F,$05,$F4,$BF,$83,$6C,$7E,$2D
	.dc.b	$B7,$F8,$F6,$DF,$BB,$83,$FD,$07,$5F,$83,$F8,$4F,$D5,$FC,$17,$EB
	.dc.b	$19,$19,$3F,$11,$18,$C3,$F3,$B1,$FE,$BC,$FA,$57,$CB,$FF,$85,$5B
	.dc.b	$1F,$9D,$F8,$FC,$4C,$8F,$F5,$1C,$EA,$BF,$49,$59,$7A,$9F,$ED,$28
	.dc.b	$D6,$5F,$9F,$84,$62,$BF,$4F,$5C,$6B,$DD,$F3,$1F,$C1,$8F,$E0,$FF
	.dc.b	$33,$6B,$F4,$51,$70,$FD,$04,$DA,$FC,$1F,$57,$F2,$1B,$FD,$B7,$C6
	.dc.b	$7E,$C7,$C4,$FD,$93,$63,$27,$EE,$E3,$2F,$FB,$03,$E9,$3F,$8E,$DF
	.dc.b	$A5,$B1,$FC,$BF,$83,$D5,$97,$FB,$C2,$72,$DF,$07,$FC,$81,$18,$BF
	.dc.b	$E3,$5B,$E2,$66,$1D,$DF,$E4,$B1,$8B,$F9,$62,$FA,$B6,$77,$5C,$49
	.dc.b	$1C,$27,$C1,$87,$9A,$38,$4E,$B0,$F6,$5F,$29,$E0,$F8,$63,$33,$FC
	.dc.b	$47,$0C,$2A,$F7,$3C,$CF,$89,$1A,$8C,$9D,$23,$91,$B7,$33,$D9,$FE
	.dc.b	$35,$3C,$5F,$39,$C4,$CC,$F1,$73,$3C,$4F,$D3,$4F,$1C,$3E,$39,$F0
	.dc.b	$9C,$5F,$96,$7E,$9D,$F3,$AD,$CF,$C3,$9E,$93,$EF,$BE,$ED,$AE,$8F
	.dc.b	$1A,$E9,$87,$CF,$9E,$73,$EF,$BE,$F7,$EB,$7A,$BF,$0A,$3E,$49,$CC
	.dc.b	$E1,$5C,$EF,$D3,$1F,$2A,$79,$3E,$38,$F0,$E1,$A9,$C9,$FD,$72,$B9
	.dc.b	$57,$58,$F9,$66,$8F,$69,$E3,$38,$4F,$3B,$F3,$9E,$A3,$D3,$1C,$0E
	.dc.b	$2E,$67,$C8,$77,$7E,$AB,$4C,$C7,$8A,$39,$E1,$F2,$23,$13,$A9,$CE
	.dc.b	$1F,$56,$30,$49,$F4,$4F,$85,$F5,$1E,$5A,$FA,$EF,$BD,$ED,$1E,$48
	.dc.b	$C5,$7C,$AA,$F3,$47,$4C,$3E,$4B,$A8,$FA,$55,$E4,$7E,$BE,$1F,$1D
	.dc.b	$F2,$47,$8E,$7C,$57,$D5,$E4,$7D,$68,$D6,$1E,$4A,$E3,$87,$82,$7B
	.dc.b	$57,$C4,$3D,$B5,$B7,$EB,$D7,$A1,$FA,$75,$C6,$3E,$34,$63,$0F,$33
	.dc.b	$ED,$6B,$98,$E5,$5F,$22,$FC,$27,$E0,$AF,$A5,$7F,$86,$BD,$DF,$F0
	.dc.b	$F1,$FE,$1F,$F4,$6E,$1F,$E2,$78,$B3,$FE,$2B,$9B,$7F,$D7,$7D,$AF
	.dc.b	$F2,$2F,$BD,$FE,$46,$6E,$1F,$E2,$16,$DF,$F8,$7D,$B5,$FC,$D7,$D0
	.dc.b	$FE,$5E,$7F,$8B,$FD,$AB,$7F,$CE,$E2,$DF,$F8,$EE,$6D,$71,$AD,$1F
	.dc.b	$E3,$D8,$CB,$FC,$FF,$D7,$FF,$2C,$F9,$3F,$E5,$A6,$C7,$F8,$D5,$B7
	.dc.b	$FE,$33,$6D,$FF,$54,$FA,$3F,$F2,$11,$FE,$43,$EE,$B7,$FE,$4B,$8B
	.dc.b	$7F,$E4,$F9,$B5,$FD,$A7,$E8,$FF,$D1,$92,$7F,$D1,$D1,$FE,$65,$F9
	.dc.b	$CF,$F1,$1E,$1F,$C5,$E7,$64,$EA,$FF,$38,$CE,$AF,$F9,$D4,$E6,$BF
	.dc.b	$CC,$D1,$8A,$FF,$35,$BD,$DB,$27,$FC,$DF,$7F,$E6,$7F,$4F,$FC,$BC
	.dc.b	$75,$5F,$44,$F8,$4C,$E3,$FE,$69,$AB,$8F,$F8,$22,$F7,$58,$7F,$C4
	.dc.b	$1F,$3B,$FE,$88,$3F,$D1,$3F,$2D,$70,$FF,$32,$45,$BF,$F3,$14,$DA
	.dc.b	$FF,$08,$7D,$5B,$6F,$5B,$70,$F7,$DA,$30,$FC,$26,$BF,$E8,$97,$FD
	.dc.b	$66,$DC,$B9,$92,$79,$FD,$1B,$D0,$F8,$15,$F3,$23,$1A,$6B,$D1,$7A
	.dc.b	$9C,$2F,$77,$FE,$97,$FA,$07,$D1,$7F,$51,$DB,$FC,$55,$18,$9F,$F6
	.dc.b	$06,$7F,$D5,$63,$11,$FE,$08,$FA,$2F,$F1,$85,$70,$BE,$57,$B6,$D7
	.dc.b	$F7,$BB,$E3,$5F,$E3,$7B,$E3,$7D,$7D,$7D,$DA,$FF,$D7,$AF,$88,$E5
	.dc.b	$1A,$87,$CD,$D8,$93,$B3,$93,$B8,$EB,$F8,$DB,$E7,$F9,$7C,$14,$BE
	.dc.b	$13,$96,$CE,$5F,$F1,$3E,$0A,$E7,$32,$76,$43,$7F,$17,$EC,$F7,$5F
	.dc.b	$2D,$F1,$36,$7D,$0B,$CB,$C3,$70,$F5,$5F,$07,$D9,$F6,$6F,$27,$7A
	.dc.b	$49,$CC,$7C,$AF,$B4,$7C,$9F,$CC,$BF,$D6,$57,$2F,$07,$53,$E5,$CF
	.dc.b	$C3,$F1,$72,$BC,$31,$E9,$8F,$2F,$DC,$C9,$F2,$C7,$D8,$FB,$57,$87
	.dc.b	$F6,$BE,$D7,$EC,$D7,$F9,$39,$BC,$3F,$E7,$7A,$DC,$FB,$B9,$7E,$09
	.dc.b	$1C,$1F,$0B,$EB,$3E,$2F,$CE,$48,$C1,$F7,$FE,$09,$F0,$3F,$8A,$FB
	.dc.b	$47,$9D,$4F,$C6,$CE,$AB,$E6,$F7,$7D,$0C,$64,$60,$FA,$3F,$35,$F5
	.dc.b	$3E,$72,$46,$0F,$C7,$7D,$9B,$D6,$FC,$CF,$C3,$BD,$8A,$7E,$72,$75
	.dc.b	$5F,$73,$E4,$3F,$3F,$2F,$5A,$9F,$87,$8F,$D1,$FD,$2F,$E4,$EB,$FC
	.dc.b	$EF,$F5,$1F,$69,$BD,$BA,$F1,$BF,$53,$F5,$1F,$C4,$E7,$F4,$4B,$E0
	.dc.b	$32,$6B,$D5,$F0,$FA,$D2,$BF,$51,$AA,$3C,$3E,$64,$9F,$D6,$69,$FB
	.dc.b	$06,$C7,$EE,$3E,$37,$84,$7E,$C5,$AF,$FA,$39,$BD,$DF,$EB,$76,$F9
	.dc.b	$DF,$E6,$1F,$A8,$77,$7F,$02,$FA,$87,$C2,$7C,$28,$98,$FE,$F6,$72
	.dc.b	$D9,$CB,$FE,$9C,$CF,$9B,$92,$19,$AD,$E5,$FE,$C9,$BE,$73,$FE,$A4
	.dc.b	$7E,$53,$CF,$CC,$FA,$67,$97,$79,$1F,$12,$32,$F9,$63,$F8,$DF,$D4
	.dc.b	$3A,$7F,$80,$7D,$1F,$C9,$AF,$EA,$8D,$C7,$FD,$B5,$5B,$9F,$E3,$4D
	.dc.b	$7F,$D6,$ED,$8D,$D7,$6F,$8C,$FE,$53,$8F,$F0,$7B,$FF,$5E,$BF,$D0
	.dc.b	$9F,$E4,$8D,$73,$E9,$70,$23,$A7,$F9,$9F,$D3,$A9,$87,$FA,$D7,$E9
	.dc.b	$FF,$91,$5F,$FA,$DE,$B5,$39,$C3,$FD,$38,$78,$EB,$75,$FC,$A7,$0F
	.dc.b	$F5,$04,$62,$F8,$CA,$31,$87,$82,$31,$7D,$2F,$8B,$63,$C0,$78,$E7
	.dc.b	$85,$72,$C3,$33,$E1,$2D,$79,$4F,$F5,$96,$1E,$2A,$E1,$3E,$03,$FD
	.dc.b	$A9,$3B,$6E,$1E,$27,$C1,$7D,$AF,$95,$79,$67,$91,$E1,$2C,$69,$E2
	.dc.b	$7F,$89,$9F,$86,$FF,$B1,$43,$87,$CF,$87,$C5,$75,$0F,$8E,$7D,$47
	.dc.b	$F8,$4A,$30,$D8,$CC,$F7,$6B,$FE,$18,$F9,$AF,$F0,$CB,$FE,$F3,$6A
	.dc.b	$69,$1F,$6E,$3F,$9F,$E5,$F9,$F1,$C4,$CE,$A8,$CC,$73,$32,$E2,$BE
	.dc.b	$BD,$73,$9F,$92,$73,$D7,$F1,$71,$97,$F1,$B1,$9D,$3F,$C3,$98,$77
	.dc.b	$8E,$27,$79,$E3,$19,$78,$C6,$67,$8E,$1E,$38,$F2,$9C,$B0,$F2,$DF
	.dc.b	$1A,$DC,$FE,$9A,$AE,$1E,$18,$FF,$7F,$5F,$09,$FF,$0B,$4F,$B4,$FF
	.dc.b	$15,$DF,$F9,$A6,$B6,$78,$23,$97,$D4,$7F,$82,$AB,$FC,$3C,$D9,$F0
	.dc.b	$00,$D8,$DE,$DD,$63,$85,$74,$8C,$56,$5D,$71,$CB,$FF,$19,$53,$EA
	.dc.b	$3F,$C5,$51,$E2,$9D,$5F,$88,$F0,$BF,$E7,$79,$DC,$67,$0F,$E7,$F5
	.dc.b	$EA,$F1,$BC,$87,$F9,$F7,$8D,$E3,$62,$74,$4B,$FE,$53,$19,$7A,$55
	.dc.b	$CB,$F0,$02,$F9,$DF,$18,$FA,$9F,$51,$FE,$26,$8F,$F2,$43,$6F,$FD
	.dc.b	$11,$F3,$8F,$8B,$FA,$93,$9B,$E3,$EB,$FD,$7F,$84,$FF,$3E,$C6,$31
	.dc.b	$FF,$29,$46,$1F,$A9,$8F,$FA,$5F,$AB,$81,$D2,$EC,$FF,$96,$7E,$A3
	.dc.b	$FC,$7C,$DF,$73,$F0,$9E,$6B,$E1,$FF,$0D,$5A,$C7,$E8,$37,$3F,$C8
	.dc.b	$7D,$43,$E3,$F7,$8F,$E3,$63,$19,$7F,$99,$FE,$A1,$F0,$7F,$91,$6C
	.dc.b	$7F,$9A,$E7,$79,$7D,$D9,$CB,$F7,$FE,$9F,$E3,$57,$FA,$81,$B1,$FE
	.dc.b	$C0,$6B,$FC,$6D,$AF,$C4,$FA,$7F,$E3,$13,$FE,$A8,$6C,$7F,$B2,$1A
	.dc.b	$FF,$23,$6B,$F3,$21,$F4,$1F,$EC,$C6,$9F,$EB,$22,$C7,$FB,$49,$AF
	.dc.b	$F2,$DF,$65,$F1,$9F,$ED,$46,$CF,$FA,$E9,$B3,$FE,$DA,$6B,$C7,$CF
	.dc.b	$E0,$49,$E0,$FF,$B7,$BE,$9B,$FD,$39,$1F,$EC,$46,$CF,$FB,$91,$AF
	.dc.b	$F3,$F6,$BF,$88,$FA,$4F,$F4,$84,$FF,$B3,$1B,$1F,$EE,$C6,$BF,$D1
	.dc.b	$DA,$F1,$F9,$5F,$F7,$7C,$1A,$49,$FE,$81,$C9,$FE,$F4,$F1,$3F,$DA
	.dc.b	$77,$FC,$62,$30,$7F,$BE,$3B,$3E,$4C,$7E,$71,$AF,$F1,$16,$A6,$9F
	.dc.b	$E5,$FA,$FF,$72,$B6,$3F,$AF,$1B,$F6,$7B,$1F,$CA,$5F,$D4,$FC,$7F
	.dc.b	$FA,$9E,$BF,$DD,$AD,$8E,$2D,$7F,$B4,$35,$FD,$8F,$D1,$BD,$EF,$E5
	.dc.b	$CD,$AF,$E3,$BF,$2E,$FA,$2C,$D4,$9F,$F7,$DD,$6E,$FC,$ED,$7F,$BB
	.dc.b	$35,$EF,$3F,$C4,$A7,$F9,$A7,$5B,$EB,$7F,$02,$FF,$45,$BE,$8F,$F1
	.dc.b	$88,$CC,$F0,$F9,$CF,$C8,$CF,$F1,$C5,$F2,$9D,$71,$FF,$1A,$C6,$AB
	.dc.b	$84,$E6,$93,$9A,$DC,$E6,$31,$3A,$8E,$19,$FE,$A2,$3F,$95,$B6,$FD
	.dc.b	$6B,$E9,$37,$D1,$27,$E5,$BC,$A3,$18,$FF,$45,$D5,$ED,$7F,$5F,$6D
	.dc.b	$FF,$4E,$CE,$D6,$EB,$47,$E0,$DF,$EB,$D5,$A7,$FC,$4C,$D7,$F9,$7B
	.dc.b	$5D,$C7,$BF,$EB,$F9,$DF,$77,$DF,$F0,$DF,$03,$D1,$B3,$F8,$46,$DF
	.dc.b	$F6,$ED,$5A,$FF,$79,$AF,$87,$7C,$5F,$57,$D2,$FC,$D3,$E7,$7C,$4D
	.dc.b	$8C,$93,$0E,$9A,$FF,$B6,$A3,$18,$FF,$82,$75,$FE,$95,$5F,$03,$C9
	.dc.b	$FA,$BF,$4B,$F2,$59,$7F,$AA,$9F,$CD,$47,$C7,$6C,$66,$93,$97,$AD
	.dc.b	$FE,$57,$D3,$E7,$7E,$FE,$4F,$EE,$EF,$9B,$E9,$3F,$11,$F4,$BF,$89
	.dc.b	$6E,$5F,$65,$B3,$B8,$FB,$93,$F8,$CF,$72,$47,$EC,$3D,$AB,$5F,$37
	.dc.b	$8B,$FD,$21,$3F,$B4,$AF,$C9,$D6,$B3,$FF,$16,$7D,$2B,$E8,$FF,$C5
	.dc.b	$0D,$BF,$C5,$7C,$4E,$26,$47,$C5,$9D,$65,$FC,$41,$AF,$F9,$BD,$AF
	.dc.b	$F7,$87,$77,$FB,$1F,$4B,$9C,$FD,$87,$F4,$FF,$4B,$2F,$F8,$5F,$D2
	.dc.b	$DA,$F6,$4D,$BF,$D5,$CD,$AF,$DF,$7C,$4F,$EC,$FE,$1B,$F6,$8D,$8F
	.dc.b	$D2,$37,$0F,$C3,$B6,$FF,$C9,$DF,$4B,$FC,$8D,$BF,$31,$FB,$AD,$7F
	.dc.b	$C9,$93,$9C,$3F,$C2,$6D,$C3,$FC,$90,$DB,$FC,$ED,$70,$B9,$7F,$C6
	.dc.b	$8E,$2F,$57,$17,$CA,$36,$FF,$90,$EF,$A7,$0F,$CA,$AE,$27,$09,$4F
	.dc.b	$8D,$B5,$D5,$E4,$F4,$C3,$DB,$5F,$CE,$30,$C3,$D2,$B6,$E4,$ED,$7D
	.dc.b	$49,$7D,$6F,$9D,$70,$23,$FC,$82,$FA,$C7,$B6,$F9,$D7,$17,$C1,$5C
	.dc.b	$88,$FF,$26,$AE,$35,$EC,$8E,$27,$95,$EE,$F2,$23,$FE,$F0,$C3,$E4
	.dc.b	$3E,$13,$65,$AF,$34,$F4,$8C,$9D,$CF,$D6,$57,$8B,$0F,$83,$3E,$1A
	.dc.b	$F3,$D7,$38,$F0,$C7,$E0,$9E,$8E,$A3,$A3,$E0,$7B,$CF,$47,$BC,$F0
	.dc.b	$3D,$6F,$07,$E2,$19,$7C,$04,$9F,$03,$D3,$0F,$A3,$18,$3C,$55,$DD
	.dc.b	$F1,$18,$8D,$E1,$F2,$2B,$C0,$7C,$93,$FD,$D9,$5F,$E0,$57,$07,$E2
	.dc.b	$4C,$CF,$F2,$98,$D4,$7C,$B3,$37,$FC,$BE,$3F,$51,$C8,$F9,$1F,$3B
	.dc.b	$67,$DD,$8F,$F8,$06,$72,$48,$CC,$EE,$FE,$ED,$6E,$73,$5B,$9C,$D6
	.dc.b	$EB,$EC,$9B,$8C,$CE,$E3,$33,$BA,$FB,$AD,$AF,$A0,$D7,$E3,$CF,$33
	.dc.b	$8D,$73,$AE,$2D,$9E,$2D,$78,$D7,$3B,$F8,$B3,$96,$D7,$B0,$95,$E8
	.dc.b	$6D,$F8,$1A,$78,$0B,$1E,$06,$BE,$0A,$ED,$87,$F8,$87,$0F,$13,$63
	.dc.b	$35,$FE,$DC,$9C,$B6,$33,$5F,$D8,$5B,$7F,$E3,$57,$FC,$51,$EB,$4A
	.dc.b	$FF,$0C,$79,$F6,$FF,$BD,$CD,$BF,$F1,$C4,$DA,$FF,$14,$FA,$FF,$E2
	.dc.b	$EF,$67,$FC,$5E,$D8,$FF,$29,$B7,$0F,$EC,$8D,$BF,$E7,$FE,$7F,$F3
	.dc.b	$C3,$FC,$75,$EE,$5B,$FE,$3F,$1F,$99,$23,$F9,$76,$DF,$F8,$6F,$EB
	.dc.b	$D8,$F1,$8C,$3F,$E4,$2F,$53,$FC,$87,$E7,$DC,$3F,$CC,$ED,$AF,$EF
	.dc.b	$0D,$CB,$FC,$8E,$FF,$93,$7E,$9A,$57,$F8,$F7,$C2,$B7,$FE,$20,$9B
	.dc.b	$7F,$E5,$D9,$B5,$FE,$E8,$FA,$57,$6F,$F9,$7F,$04,$62,$7F,$CF,$3F
	.dc.b	$4D,$98,$C4,$B0,$57,$3B,$FF,$12,$56,$AB,$FD,$EA,$74,$8C,$92,$7F
	.dc.b	$0D,$F0,$7D,$63,$F1,$8F,$F9,$A7,$D5,$93,$FD,$8B,$07,$D3,$3E,$07
	.dc.b	$38,$FF,$8F,$AE,$D7,$92,$3C,$D5,$BB,$FF,$31,$57,$F9,$CB,$E6,$E4
	.dc.b	$7F,$A5,$BE,$4A,$DF,$F8,$9E,$6E,$1F,$E7,$69,$B4,$A9,$F4,$DF,$EB
	.dc.b	$C6,$E1,$FE,$45,$8C,$41,$BF,$4D,$FE,$8E,$CF,$FC,$98,$D9,$FF,$5B
	.dc.b	$BF,$EE,$8F,$6B,$3E,$C7,$BE,$7F,$D7,$7F,$5E,$F9,$CF,$F2,$A4,$23
	.dc.b	$0F,$F9,$5D,$AF,$08,$D8,$FA,$7F,$54,$7D,$BD,$2C,$E6,$3F,$C1,$4D
	.dc.b	$74,$7C,$1F,$A7,$F9,$CE,$0F,$25,$74,$DF,$35,$C5,$B1,$9B,$F0,$DF
	.dc.b	$29,$DF,$BD,$9C,$3E,$76,$1F,$EC,$D7,$C7,$19,$7E,$17,$CC,$59,$D4
	.dc.b	$78,$DB,$5D,$23,$D7,$3C,$7E,$9D,$F3,$7C,$7E,$4E,$57,$F8,$A6,$72
	.dc.b	$D9,$F9,$F1,$88,$FF,$76,$B5,$FC,$FB,$5F,$9A,$EE,$FE,$3B,$73,$CD
	.dc.b	$7A,$7E,$33,$D9,$F4,$CF,$94,$FB,$AD,$C7,$FD,$24,$D7,$FD,$FA,$D7
	.dc.b	$F5,$FF,$4A,$F2,$7F,$0A,$D8,$FF,$4D,$35,$FD,$A8,$70,$FE,$E3,$E9
	.dc.b	$9E,$AF,$E3,$9B,$1F,$EA,$36,$BC,$9A,$FF,$00,$B7,$9D,$EA,$24,$7A
	.dc.b	$9F,$6F,$D2,$C8,$F6,$B5,$F2,$FD,$23,$BF,$F3,$ED,$8F,$F5,$A3,$5D
	.dc.b	$C7,$C8,$E4,$91,$F2,$1A,$FC,$8F,$3F,$E6,$91,$F9,$AF,$EB,$3E,$0D
	.dc.b	$F0,$54,$93,$FE,$C3,$8C,$BF,$53,$A3,$EB,$91,$C1,$F6,$3E,$5B,$ED
	.dc.b	$91,$C1,$FE,$CD,$FA,$3F,$E8,$ED,$F3,$BF,$7B,$18,$8F,$AE,$D7,$E7
	.dc.b	$7D,$1F,$E1,$88,$FE,$19,$FE,$0B,$F4,$1F,$93,$8F,$E3,$0D,$4D,$31
	.dc.b	$F4,$5F,$ED,$A8,$FF,$70,$35,$FC,$3B,$5F,$E7,$AD,$7F,$85,$C3,$BB
	.dc.b	$F6,$DE,$B3,$C9,$A7,$EE,$67,$55,$FE,$97,$6B,$FC,$2D,$BD,$FF,$C3
	.dc.b	$9B,$DD,$FC,$63,$E2,$FA,$57,$FA,$7F,$E8,$9F,$0D,$FC,$16,$73,$3F
	.dc.b	$C5,$5B,$87,$F1,$8A,$DC,$FF,$A9,$FE,$89,$F0,$1E,$5C,$24,$E6,$3F
	.dc.b	$88,$4E,$A3,$D1,$3A,$7F,$91,$46,$5B,$39,$BF,$55,$72,$8E,$7E,$A7
	.dc.b	$C3,$F1,$A5,$F8,$E3,$2F,$C5,$8C,$BD,$A3,$F9,$17,$D1,$3E,$37,$F9
	.dc.b	$46,$2F,$9D,$F8,$F7,$FD,$DD,$AA,$B8,$3F,$EC,$76,$BF,$CE,$DB,$3F
	.dc.b	$CF,$9B,$B7,$F4,$85,$F1,$1F,$CC,$F2,$FE,$29,$5F,$D1,$17,$9F,$93
	.dc.b	$5E,$44,$C7,$FD,$AD,$3C,$31,$FE,$A6,$D8,$FE,$DD,$F4,$5D,$B0,$FE
	.dc.b	$E1,$F4,$2A,$FE,$B3,$39,$A4,$E5,$FE,$ED,$3F,$C9,$E3,$11,$C7,$0F
	.dc.b	$EC,$31,$8A,$E4,$DC,$BF,$C3,$D5,$D2,$FC,$B7,$C1,$B1,$E4,$7C,$73
	.dc.b	$C2,$B9,$61,$9A,$F1,$DF,$16,$99,$18,$78,$B0,$E7,$59,$7D,$47,$82
	.dc.b	$7C,$87,$C6,$C3,$C1,$1C,$23,$27,$BC,$E5,$7E,$17,$BC,$F0,$79,$57
	.dc.b	$F4,$47,$2D,$BF,$9D,$1F,$E3,$C6,$BF,$4B,$E2,$9F,$45,$B6,$C7,$F4
	.dc.b	$E9,$46,$23,$93,$5F,$EA,$7F,$2F,$FE,$1C,$4D,$23,$FE,$20,$7E,$45
	.dc.b	$FE,$87,$4F,$95,$3F,$E3,$59,$CF,$07,$F9,$0F,$5F,$C0,$CE,$4F,$8B
	.dc.b	$7C,$5F,$8E,$66,$3F,$0F,$19,$78,$C7,$90,$E3,$3D,$75,$F4,$C6,$30
	.dc.b	$F1,$BB,$AE,$75,$B8,$FF,$32,$E1,$E6,$AD,$B6,$3C,$67,$1B,$FF,$16
	.dc.b	$57,$18,$FF,$39,$3F,$E7,$9C,$3D,$47,$F8,$D2,$3D,$53,$C2,$FF,$CB
	.dc.b	$13,$FE,$2F,$AE,$07,$FA,$06,$B6,$69,$FF,$42,$7D,$0F,$F8,$4E,$B8
	.dc.b	$9F,$E4,$87,$13,$FE,$41,$7F,$CC,$91,$A8,$FF,$33,$46,$AB,$FC,$8E
	.dc.b	$D9,$E5,$B7,$6A,$EE,$F8,$39,$24,$FF,$A0,$7E,$8B,$FC,$6E,$F5,$A5
	.dc.b	$FF,$40,$8C,$36,$3C,$95,$FE,$83,$C2,$5F,$29,$CD,$F0,$F0,$7D,$E6
	.dc.b	$DF,$8F,$C8,$F7,$33,$1B,$25,$7F,$99,$30,$FE,$61,$3A,$8F,$89,$76
	.dc.b	$7C,$57,$CE,$F8,$C7,$F9,$AB,$E8,$BF,$C5,$F1,$ED,$6D,$F5,$F9,$A7
	.dc.b	$C3,$FF,$AD,$E7,$37,$C7,$D1,$FF,$5D,$7B,$8E,$BD,$46,$19,$97,$C9
	.dc.b	$9D,$3D,$E7,$39,$7F,$9E,$FE,$8B,$FC,$95,$3F,$3D,$B1,$FE,$7F,$F2
	.dc.b	$1E,$FF,$FB,$26,$73,$97,$FA,$13,$E8,$9F,$15,$FE,$CD,$6E,$3C,$FD
	.dc.b	$8C,$4F,$FB,$47,$E8,$9E,$D5,$33,$5F,$7E,$32,$FE,$07,$E8,$BE,$1B
	.dc.b	$FE,$C5,$FA,$C7,$C5,$FF,$B6,$E7,$3B,$FD,$59,$CD,$7E,$27,$E8,$FF
	.dc.b	$89,$BF,$EC,$DF,$8C,$FE,$98,$E7,$C3,$FC,$15,$FF,$B7,$E7,$2F,$FA
	.dc.b	$7E,$5F,$3F,$AE,$05,$7E,$35,$FF,$52,$CE,$8E,$DF,$03,$FE,$EF,$D1
	.dc.b	$F4,$7C,$7E,$B1,$F5,$19,$8C,$3B,$F5,$35,$F6,$77,$1F,$F7,$25,$A3
	.dc.b	$FD,$F5,$F4,$DF,$EC,$59,$FF,$74,$53,$E7,$73,$F0,$1E,$78,$C4,$FE
	.dc.b	$AF,$E9,$BD,$BB,$7E,$9F,$EB,$3F,$DF,$B3,$9F,$37,$DD,$18,$E3,$FF
	.dc.b	$60,$5B,$E5,$F3,$E0,$FC,$89,$4E,$70,$FF,$7D,$5F,$E6,$1F,$F6,$8F
	.dc.b	$9F,$F8,$A7,$FD,$FD,$EF,$7C,$7B,$FF,$66,$4E,$5F,$D7,$46,$73,$FF
	.dc.b	$68,$F9,$16,$FF,$8E,$FD,$67,$F8,$E9,$FE,$0D,$19,$B5,$2B,$51,$FE
	.dc.b	$D9,$9F,$B3,$5F,$C6,$69,$F1,$2F,$83,$F6,$56,$AF,$F8,$04,$E6,$FF
	.dc.b	$8A,$7B,$1F,$EA,$07,$94,$3E,$5B,$E2,$4E,$6B,$8C,$7C,$7F,$A8,$7C
	.dc.b	$2D,$3F,$19,$F3,$8F,$A3,$C6,$5F,$C8,$6B,$6F,$9F,$E7,$3F,$01,$8F
	.dc.b	$FB,$CE,$72,$FF,$15,$8C,$ED,$FE,$F8,$9C,$BF,$C9,$63,$34,$A9,$5A
	.dc.b	$8F,$E9,$33,$A9,$F9,$7F,$4B,$F9,$A9,$F9,$0B,$E5,$BF,$48,$E6,$7F
	.dc.b	$93,$39,$9D,$F8,$1A,$AD,$B6,$3E,$96,$7F,$67,$23,$47,$D4,$7C,$6F
	.dc.b	$16,$57,$07,$E4,$C6,$32,$FC,$0C,$FB,$5F,$99,$F0,$DF,$D9,$DF,$81
	.dc.b	$B2,$31,$1A,$9F,$F1,$24,$E5,$B3,$EE,$FA,$8F,$D0,$4E,$6B,$DF,$18
	.dc.b	$9D,$57,$F5,$9A,$D6,$5F,$D5,$75,$5F,$E0,$39,$7F,$03,$EF,$DC,$3E
	.dc.b	$1B,$72,$FE,$A1,$AF,$F8,$02,$7F,$BE,$BF,$93,$8C,$CF,$D2,$FA,$89
	.dc.b	$5F,$5B,$E3,$AD,$7D,$49,$B9,$F8,$E6,$D7,$C7,$D3,$FA,$3D,$FC,$93
	.dc.b	$E3,$FB,$1F,$1A,$33,$19,$6B,$C6,$3F,$93,$71,$FF,$75,$C1,$79,$3F
	.dc.b	$35,$EF,$FC,$AA,$DB,$63,$27,$E3,$BE,$9F,$F1,$BE,$37,$E3,$9B,$19
	.dc.b	$3F,$27,$19,$69,$F8,$CF,$86,$FD,$69,$FA,$EF,$3E,$C6,$49,$9F,$CE
	.dc.b	$8C,$BF,$8A,$3F,$4D,$F4,$EF,$A4,$FC,$FF,$8D,$FC,$86,$31,$7F,$A4
	.dc.b	$8C,$4E,$A3,$FC,$35,$3A,$AF,$D2,$D6,$A7,$81,$AA,$E1,$3F,$80,$FA
	.dc.b	$7F,$E0,$A7,$F0,$6F,$A7,$B5,$FA,$88,$B3,$F8,$69,$BB,$FE,$3F,$E9
	.dc.b	$FF,$6B,$F1,$BF,$B6,$6E,$1F,$BA,$6D,$FE,$E1,$B5,$F6,$23,$F8,$37
	.dc.b	$D3,$FF,$1B,$8F,$E1,$5F,$2D,$FD,$06,$31,$7F,$C2,$23,$1F,$35,$F8
	.dc.b	$26,$C7,$F1,$96,$DF,$F1,$76,$DF,$49,$FF,$20,$5F,$4A,$FE,$77,$CA
	.dc.b	$BD,$D7,$73,$78,$77,$69,$DC,$B5,$DD,$B8,$77,$8F,$BC,$30,$F0,$43
	.dc.b	$0F,$05,$70,$3C,$15,$C2,$B7,$3A,$8C,$5F,$8D,$B5,$E3,$69,$E3,$2D
	.dc.b	$F8,$CF,$A0,$F8,$E7,$94,$62,$FB,$E1,$F0,$AB,$E2,$B6,$7D,$0D,$C3
	.dc.b	$C9,$1C,$6F,$9E,$1F,$C9,$70,$E6,$FB,$70,$E6,$DA,$E6,$DB,$E6,$F1
	.dc.b	$BE,$75,$F8,$66,$BE,$98,$F8,$15,$E1,$69,$C4,$B1,$C5,$AF,$13,$E3
	.dc.b	$47,$16,$C7,$16,$9C,$4A,$71,$8F,$99,$1B,$6C,$6D,$AE,$E3,$EE,$C7
	.dc.b	$73,$37,$F1,$A7,$EB,$CE,$8E,$85,$7A,$35,$E8,$7C,$C3,$89,$67,$B4
	.dc.b	$6E,$7B,$46,$27,$53,$D9,$A7,$62,$9D,$A3,$EB,$BD,$9A,$73,$2C,$73
	.dc.b	$6C,$73,$6B,$CE,$FE,$FB,$96,$B9,$C3,$F1,$C4,$AF,$C3,$36,$3C,$53
	.dc.b	$C2,$FD,$B5,$E1,$8C,$61,$F8,$89,$CD,$77,$BE,$58,$7A,$6F,$78,$7E
	.dc.b	$1A,$F7,$7E,$38,$F0,$91,$F0,$DF,$28,$F0,$36,$3B,$46,$EF,$B4,$E6
	.dc.b	$7A,$1C,$A7,$47,$22,$BC,$8D,$12,$79,$C6,$A3,$9C,$61,$FF,$13,$FB
	.dc.b	$FF,$E2,$A3,$FC,$57,$E8,$DA,$FF,$17,$C5,$9F,$F1,$6C,$DA,$FF,$70
	.dc.b	$76,$7F,$AC,$27,$FC,$0F,$F0,$7F,$E4,$CF,$80,$FF,$26,$B7,$1F,$F1
	.dc.b	$7B,$67,$71,$FC,$C3,$D2,$FF,$4F,$7D,$3A,$BF,$9F,$46,$1B,$3F,$E6
	.dc.b	$99,$D3,$C4,$C4,$6A,$BF,$DD,$13,$A7,$F9,$45,$73,$AF,$F2,$0F,$B5
	.dc.b	$FE,$62,$F7,$3F,$CC,$6D,$8C,$9F,$E4,$68,$CB,$4F,$F2,$9F,$D3,$FF
	.dc.b	$94,$8F,$F2,$9F,$7D,$8C,$93,$0F,$F2,$AC,$65,$FE,$83,$F4,$EF,$A7
	.dc.b	$FF,$0C,$FC,$6F,$FA,$76,$31,$7F,$E7,$18,$C4,$EA,$3F,$CD,$5C,$1F
	.dc.b	$CE,$2B,$77,$FE,$73,$BD,$D7,$13,$67,$F9,$56,$77,$5F,$E7,$59,$FF
	.dc.b	$35,$FC,$44,$AF,$F4,$B7,$D3,$BE,$2B,$18,$FF,$9B,$AA,$E1,$FD,$DC
	.dc.b	$56,$EF,$FC,$EF,$3F,$E8,$EF,$9D,$91,$FE,$70,$F9,$AB,$87,$F9,$3A
	.dc.b	$6E,$1F,$8C,$9B,$5F,$E6,$BF,$A6,$FF,$40,$65,$FE,$45,$C3,$FD,$15
	.dc.b	$18,$70,$FF,$B8,$3D,$8D,$42,$34,$FF,$93,$7E,$99,$F0,$5A,$84,$61
	.dc.b	$E2,$D7,$C0,$D7,$97,$97,$C6,$0D,$24,$78,$A5,$F4,$DF,$E8,$5D,$3F
	.dc.b	$B7,$C6,$27,$EA,$9F,$EA,$57,$0F,$F8,$1B,$E9,$BF,$BA,$CF,$0A,$E5
	.dc.b	$A7,$FA,$3E,$F7,$39,$BE,$62,$F8,$D6,$FD,$BE,$33,$DD,$AF,$C4,$6D
	.dc.b	$BE,$7E,$C4,$9E,$B3,$FE,$CC,$6C,$F2,$FA,$67,$C5,$FC,$D6,$CF,$D4
	.dc.b	$6B,$F6,$DA,$FF,$A0,$5C,$E5,$FE,$14,$6E,$3F,$EA,$D6,$CF,$C7,$F9
	.dc.b	$0F,$05,$C8,$F8,$33,$A7,$EF,$CF,$BB,$E0,$5F,$33,$49,$3F,$EF,$5F
	.dc.b	$AC,$F6,$3F,$E3,$6F,$A7,$78,$FF,$19,$B1,$F3,$9A,$FF,$9E,$74,$57
	.dc.b	$C9,$8F,$A7,$F4,$EF,$4B,$E7,$B6,$33,$1D,$7D,$6F,$37,$9D,$27,$ED
	.dc.b	$E8,$8F,$B3,$8F,$6F,$9A,$F6,$7D,$3C,$BD,$BA,$CF,$F6,$B1,$F8,$4F
	.dc.b	$A7,$77,$FE,$01,$B1,$98,$FC,$1F,$D6,$3C,$3F,$91,$D1,$2B,$C6,$F5
	.dc.b	$F8,$CF,$99,$E8,$C9,$F3,$BE,$1E,$8F,$C6,$FC,$0B,$DF,$A4,$AF,$CB
	.dc.b	$FD,$67,$F4,$17,$DF,$DD,$F5,$C8,$FD,$5F,$3F,$ED,$71,$48,$F8,$AF
	.dc.b	$BF,$E4,$FF,$37,$F5,$3C,$E7,$FC,$DE,$DE,$5F,$F2,$FB,$7C,$0F,$F7
	.dc.b	$C7,$C2,$7D,$A9,$FF,$31,$F4,$B8,$7F,$10,$4C,$3F,$5B,$F5,$9F,$CE
	.dc.b	$5F,$D7,$FD,$4F,$E3,$C9,$C5,$FA,$CF,$AA,$7B,$B7,$D3,$97,$FB,$69
	.dc.b	$47,$EC,$BE,$29,$F1,$7F,$B4,$C7,$F1,$57,$FE,$88,$F7,$5E,$97,$EE
	.dc.b	$1B,$1E,$2F,$2F,$94,$F1,$64,$7F,$04,$C3,$6F,$FA,$4B,$EA,$BF,$8C
	.dc.b	$B9,$AE,$2F,$F0,$A8,$CB,$FE,$98,$FA,$A7,$C1,$79,$7D,$19,$3F,$C4
	.dc.b	$27,$2D,$9C,$BF,$EA,$0C,$F9,$F2,$7F,$A4,$2B,$73,$D6,$B7,$3F,$E9
	.dc.b	$4A,$DF,$32,$A5,$F0,$AE,$5B,$FC,$7F,$98,$95,$E3,$8F,$E3,$B5,$F3
	.dc.b	$A3,$EB,$CE,$9F,$64,$FF,$22,$FA,$A7,$47,$EB,$BE,$97,$E8,$E3,$FD
	.dc.b	$7D,$B9,$78,$F0,$5F,$07,$F9,$8B,$5F,$F5,$FB,$6B,$FD,$5E,$D9,$E4
	.dc.b	$BE,$2F,$F9,$9E,$5F,$C5,$AF,$F9,$DA,$F6,$32,$6B,$D2,$93,$A6,$BF
	.dc.b	$EC,$EF,$A9,$E9,$97,$C9,$FA,$96,$1E,$99,$CD,$27,$2F,$CF,$9F,$13
	.dc.b	$97,$E0,$18,$8D,$C2,$79,$C6,$1B,$3C,$2F,$E4,$4F,$08,$E4,$78,$F1
	.dc.b	$61,$E7,$8C,$37,$0F,$25,$F2,$9E,$91,$93,$C5,$5C,$9F,$41,$F8,$9C
	.dc.b	$3C,$6F,$23,$84,$F8,$9D,$4F,$1C,$3D,$4D,$9E,$F3,$D0,$F9,$77,$E3
	.dc.b	$78,$E1,$E3,$C3,$E6,$C7,$E1,$CF,$2C,$7F,$85,$A3,$31,$EE,$3F,$C3
	.dc.b	$CB,$E3,$5F,$55,$B6,$CE,$9A,$F5,$9E,$13,$FE,$1D,$87,$CD,$D2,$3F
	.dc.b	$E2,$16,$E7,$F6,$B0,$FD,$0E,$DF,$2E,$BF,$61,$5F,$36,$73,$1F,$04
	.dc.b	$C9,$F0,$DC,$D7,$E1,$23,$2F,$5E,$05,$F5,$9F,$29,$C6,$32,$F5,$33
	.dc.b	$3C,$67,$85,$F8,$EB,$73,$DE,$B7,$A7,$F8,$BE,$72,$FF,$92,$6B,$85
	.dc.b	$F2,$AF,$03,$87,$F2,$2F,$13,$FC,$56,$FA,$A7,$0E,$8F,$F3,$9D,$7F
	.dc.b	$98,$E7,$38,$7A,$A7,$F8,$35,$F5,$8C,$7D,$57,$F8,$4A,$BF,$C6,$ED
	.dc.b	$AE,$81,$FE,$41,$1A,$D7,$B4,$FF,$23,$9E,$15,$D2,$73,$3C,$4B,$3A
	.dc.b	$AF,$E3,$0F,$F9,$61,$F0,$F3,$74,$B7,$D5,$7F,$8E,$A7,$24,$BF,$2D
	.dc.b	$F1,$7F,$C4,$13,$B7,$37,$E0,$F1,$A5,$FF,$97,$B0,$72,$7F,$83,$63
	.dc.b	$E0,$CE,$AF,$AB,$E3,$9D,$46,$27,$64,$7F,$C6,$11,$AA,$F9,$B5,$F0
	.dc.b	$BE,$A7,$FC,$61,$7F,$E6,$46,$DF,$F8,$DF,$E7,$5F,$55,$BF,$66,$5F
	.dc.b	$F9,$AB,$E0,$FF,$BA,$BC,$63,$E6,$C6,$2F,$EB,$B7,$2F,$C0,$36,$BF
	.dc.b	$A3,$63,$E0,$9C,$F5,$7D,$6F,$AA,$FF,$26,$D7,$F9,$E1,$B8,$7E,$0B
	.dc.b	$E3,$5F,$0D,$BF,$80,$FB,$BF,$55,$FE,$F1,$CB,$F1,$CD,$C7,$FA,$A6
	.dc.b	$3C,$A7,$3F,$1F,$F8,$4F,$AA,$7C,$25,$39,$C5,$BF,$AD,$F5,$2F,$8F
	.dc.b	$FC,$23,$63,$3F,$45,$89,$FC,$40,$F8,$3A,$63,$EA,$2D,$7E,$06,$5E
	.dc.b	$CF,$03,$8F,$B3,$67,$F0,$DF,$51,$FE,$B6,$6F,$6F,$E5,$63,$1C,$BF
	.dc.b	$E6,$F7,$F2,$5C,$1F,$ED,$37,$FD,$5F,$5F,$8F,$F4,$BF,$87,$BD,$3E
	.dc.b	$33,$F3,$D1,$8B,$FF,$3C,$3F,$EE,$49,$FF,$76,$BF,$5D,$BF,$23,$C3
	.dc.b	$B3,$F4,$DF,$4C,$FB,$1A,$63,$EC,$AD,$FE,$7E,$7F,$C8,$2D,$F6,$B8
	.dc.b	$FA,$9C,$9F,$D8,$7B,$BF,$E4,$09,$FD,$33,$74,$FF,$46,$2F,$83,$DF
	.dc.b	$B7,$FB,$7F,$A8,$FB,$D7,$FE,$25,$E2,$43,$83,$5E,$06,$DF,$DF,$7C
	.dc.b	$07,$F3,$AA,$FF,$7D,$2F,$AE,$CF,$A1,$A7,$7F,$15,$C3,$DD,$B9,$FE
	.dc.b	$EE,$FF,$CA,$6F,$EF,$7E,$95,$F4,$FF,$C2,$D7,$D4,$58,$C3,$67,$83
	.dc.b	$78,$BF,$D5,$55,$C2,$7A,$57,$06,$9C,$0F,$E1,$9F,$3A,$FA,$8F,$E3
	.dc.b	$4D,$8C,$F1,$62,$64,$70,$77,$18,$C3,$F9,$8F,$D3,$3E,$13,$F9,$5B
	.dc.b	$6F,$DD,$18,$23,$FD,$F6,$7F,$CA,$D3,$B7,$0D,$7F,$C0,$4D,$7E,$C6
	.dc.b	$1F,$76,$7D,$7C,$1E,$AE,$5F,$5B,$63,$27,$C0,$8C,$B4,$CD,$7F,$B4
	.dc.b	$EB,$A3,$6B,$EA,$7D,$3C,$AF,$B1,$F1,$B6,$BE,$8C,$DE,$4F,$EE,$63
	.dc.b	$7F,$F5,$D5,$FD,$F9,$F8,$BE,$8C,$8F,$A7,$CB,$70,$F9,$73,$FD,$8D
	.dc.b	$FF,$34,$6B,$FC,$A5,$F8,$71,$97,$E4,$B7,$1F,$07,$D3,$FE,$2F,$E5
	.dc.b	$BF,$18,$DC,$7F,$20,$DB,$FC,$4B,$6B,$D3,$9A,$31,$5F,$A7,$8C,$3F
	.dc.b	$99,$BF,$C4,$FC,$1C,$8F,$C6,$FC,$1D,$C7,$E6,$4D,$C7,$E2,$F0,$7F
	.dc.b	$8D,$EF,$F2,$73,$FD,$CB,$E9,$DF,$4F,$4F,$D2,$78,$D6,$BF,$70,$7F
	.dc.b	$A2,$E7,$6F,$E2,$F1,$46,$1F,$C7,$35,$FF,$26,$35,$FC,$5F,$D3,$FE
	.dc.b	$88,$FD,$1F,$D3,$DA,$FD,$3C,$59,$FC,$34,$DC,$FD,$73,$6B,$F2,$B3
	.dc.b	$6B,$FC,$EB,$B7,$F1,$C7,$F6,$FE,$27,$EF,$FC,$4F,$E0,$0D,$BF,$E0
	.dc.b	$8D,$9D,$C7,$C0,$7F,$D7,$9F,$4C,$F9,$BF,$F0,$CB,$67,$F7,$BB,$62
	.dc.b	$64,$71,$A5,$6D,$FD,$05,$69,$FE,$17,$E2,$7F,$13,$3F,$8A,$78,$96
	.dc.b	$FF,$8C,$45,$9F,$E2,$33,$76,$FC,$4C,$DA,$FC,$44,$DA,$F8,$73,$6B
	.dc.b	$F4,$D3,$74,$FE,$6B,$5A,$BF,$E4,$D3,$9B,$D3,$63,$C6,$68,$B3,$EA
	.dc.b	$6D,$FE,$72,$78,$4E,$4F,$4C,$E8,$FC,$F5,$76,$71,$1E,$33,$85,$73
	.dc.b	$8C,$BF,$BF,$9E,$13,$A9,$E6,$6A,$FF,$A7,$36,$77,$5F,$22,$30,$7A
	.dc.b	$EB,$91,$E1,$8D,$3E,$83,$2F,$86,$3B,$5F,$BA,$36,$79,$E7,$77,$F3
	.dc.b	$A3,$84,$73,$77,$5E,$88,$DD,$73,$8C,$4F,$33,$73,$E5,$8C,$9F,$80
	.dc.b	$3D,$6F,$26,$C7,$03,$F0,$2E,$0E,$31,$FA,$89,$F0,$3D,$4C,$46,$DC
	.dc.b	$4F,$27,$DB,$7F,$32,$79,$CF,$B8,$F0,$5F,$D2,$AD,$CF,$C8,$3A,$4E
	.dc.b	$DF,$54,$68,$F5,$16,$39,$DF,$99,$E1,$1E,$63,$53,$E2,$AF,$04,$F8
	.dc.b	$AF,$CD,$5E,$A9,$FA,$84,$BE,$8D,$BF,$2B,$67,$A9,$C6,$77,$5F,$B8
	.dc.b	$AD,$B6,$77,$7E,$63,$E4,$1E,$39,$F8,$51,$AB,$ED,$39,$8F,$82,$D3
	.dc.b	$E0,$CF,$63,$F3,$0E,$5F,$81,$7E,$47,$27,$D4,$E2,$FF,$39,$52,$90
	.dc.b	$F6,$FF,$C2,$BE,$B7,$F8,$59,$B7,$FE,$1A,$6D,$FF,$79,$7F,$91,$FA
	.dc.b	$32,$7F,$C5,$DE,$F5,$AF,$F6,$3E,$DF,$E6,$C9,$FE,$48,$D8,$FF,$15
	.dc.b	$CE,$F2,$FE,$AF,$F4,$BF,$E2,$6F,$53,$FC,$4E,$DC,$3F,$C5,$6D,$BF
	.dc.b	$F0,$53,$6B,$8C,$FF,$8C,$2B,$8B,$4F,$F2,$17,$D2,$FF,$90,$4F,$F2
	.dc.b	$17,$6D,$8C,$93,$0F,$F1,$FC,$65,$FF,$70,$FD,$2F,$F9,$CD,$07,$18
	.dc.b	$7C,$CF,$67,$FD,$BF,$1B,$9D,$75,$FF,$82,$DB,$87,$F6,$B9,$4E,$5B
	.dc.b	$1F,$E3,$9B,$FF,$0F,$7B,$FF,$E5,$93,$FD,$C9,$F4,$AF,$82,$C5,$7F
	.dc.b	$97,$E6,$E3,$F7,$E3,$FC,$E5,$5B,$74,$48,$FF,$2E,$CF,$4A,$E3,$7F
	.dc.b	$E0,$AF,$43,$FC,$B2,$7F,$96,$BD,$0B,$87,$F9,$7A,$2E,$1F,$E3,$D9
	.dc.b	$B4,$AF,$8F,$2F,$AE,$FF,$64,$C7,$F8,$2D,$B1,$FE,$57,$6B,$C0,$E1
	.dc.b	$8F,$C9,$FA,$6F,$F3,$CE,$5F,$E1,$36,$CF,$F9,$7D,$FF,$3F,$FC,$86
	.dc.b	$7D,$2F,$8D,$3F,$E6,$3F,$9D,$7B,$7E,$4D,$3C,$D3,$F7,$27,$2F,$F7
	.dc.b	$B6,$BF,$2C,$FC,$0F,$D2,$F0,$70,$23,$73,$96,$EB,$F0,$9B,$3C,$A3
	.dc.b	$C4,$66,$EC,$78,$DA,$F5,$6D,$F8,$6F,$B7,$CE,$6B,$03,$4C,$7C,$C5
	.dc.b	$C3,$D9,$15,$F4,$CA,$7B,$FD,$2B,$E6,$FA,$7C,$9C,$BF,$F0,$F4,$E5
	.dc.b	$B3,$FE,$45,$51,$97,$FD,$24,$D7,$EE,$35,$FC,$2F,$E3,$0E,$5F,$7B
	.dc.b	$63,$FC,$58,$D7,$E7,$AE,$5D,$CC,$9D,$BC,$A3,$11,$C2,$2E,$9F,$66
	.dc.b	$B9,$4F,$F9,$06,$1F,$49,$F5,$97,$EC,$9F,$3F,$87,$FD,$B8,$DA,$FE
	.dc.b	$59,$5B,$8F,$F7,$94,$73,$5F,$05,$F7,$2F,$9B,$63,$73,$F8,$D9,$EF
	.dc.b	$EF,$67,$92,$4F,$38,$E1,$EE,$48,$EC,$F3,$FA,$BE,$9E,$D4,$9F,$13
	.dc.b	$DF,$DB,$F0,$7B,$32,$7C,$CF,$8F,$D9,$F2,$78,$5E,$5F,$85,$95,$EB
	.dc.b	$8F,$47,$AB,$E9,$F5,$64,$FB,$E3,$D9,$E9,$FB,$7D,$29,$3F,$0E,$3E
	.dc.b	$07,$9B,$23,$E3,$3F,$0F,$CB,$F8,$9E,$5C,$9F,$94,$FC,$7F,$1F,$E5
	.dc.b	$77,$7E,$F2,$7E,$43,$79,$BF,$D8,$AD,$EA,$FE,$6F,$1F,$9B,$FB,$5F
	.dc.b	$D5,$FC,$13,$E1,$29,$30,$FF,$3D,$7D,$0F,$F3,$47,$FC,$F9,$F3,$FF
	.dc.b	$9B,$C3,$F8,$03,$71,$FE,$3A,$D9,$FE,$8B,$F4,$2E,$2F,$DC,$AD,$A9
	.dc.b	$27,$FD,$11,$D6,$F8,$5F,$DF,$E1,$FB,$39,$FF,$46,$72,$7E,$59,$FE
	.dc.b	$1E,$DC,$FC,$7E,$3F,$F0,$6C,$31,$A7,$92,$39,$99,$F1,$7F,$85,$FD
	.dc.b	$0B,$E3,$31,$97,$8A,$73,$1F,$C4,$3E,$85,$D5,$E4,$89,$5F,$E9,$F9
	.dc.b	$CB,$67,$2F,$F2,$4D,$7B,$39,$21,$9A,$DE,$9F,$CA,$EF,$9C,$FF,$2C
	.dc.b	$7E,$43,$F4,$BC,$AF,$8A,$79,$78,$24,$64,$87,$17,$F9,$4F,$D0,$BE
	.dc.b	$23,$F6,$5F,$35,$F3,$6B,$F6,$0D,$C7,$F9,$E5,$6E,$7F,$D7,$8D,$7F
	.dc.b	$9F,$B6,$37,$5D,$BE,$1B,$F9,$BE,$2F,$9D,$BE,$A4,$C7,$FD,$8D,$39
	.dc.b	$6C,$F0,$23,$FE,$CB,$FA,$15,$FF,$8E,$61,$F4,$0B,$FE,$DB,$5A,$CB
	.dc.b	$FC,$2B,$39,$AF,$F0,$BD,$77,$78,$BF,$DD,$0C,$46,$E1,$5C,$DD,$9F
	.dc.b	$E2,$07,$11,$B9,$E2,$73,$C3,$FB,$FC,$73,$9D,$CF,$9B,$0F,$54,$F1
	.dc.b	$AD,$CF,$1B,$F4,$BB,$C2,$C7,$86,$F8,$CE,$E3,$CA,$7C,$3A,$E6,$EE
	.dc.b	$B8,$BE,$CA,$E2,$FC,$49,$DB,$C4,$DE,$1E,$87,$91,$FE,$13,$39,$55
	.dc.b	$69,$98,$F9,$0F,$F8,$65,$7C,$53,$E8,$78,$46,$1B,$19,$9E,$4D,$7B
	.dc.b	$7D,$0F,$F6,$D4,$D2,$3F,$DC,$5F,$7D,$FE,$7F,$A3,$E0,$CF,$F9,$62
	.dc.b	$73,$B7,$F8,$83,$5F,$C0,$4E,$4F,$97,$7C,$5F,$9A,$66,$3F,$0D,$19
	.dc.b	$78,$C7,$90,$E3,$3D,$75,$F9,$17,$E0,$C3,$C7,$5B,$BE,$77,$B8,$FF
	.dc.b	$20,$56,$DF,$09,$BB,$FF,$1C,$4E,$E3,$FC,$9B,$3B,$AE,$77,$FD,$00
	.dc.b	$D9,$67,$A9,$E6,$70,$FF,$A9,$6B,$FC,$87,$87,$F8,$B5,$B3,$EA,$9E
	.dc.b	$07,$F9,$03,$E8,$7F,$7D,$1C,$9B,$19,$38,$9F,$E8,$87,$35,$FE,$49
	.dc.b	$6C,$FF,$99,$9B,$3B,$38,$87,$2D,$E4,$CC,$7F,$3B,$9F,$F0,$D7,$D0
	.dc.b	$FF,$8E,$5F,$24,$E8,$8E,$BC,$4B,$7E,$BF,$A1,$E6,$7A,$31,$72,$BE
	.dc.b	$97,$38,$7B,$EA,$DF,$F2,$A9,$B7,$ED,$9B,$5F,$DC,$AF,$93,$C4,$FA
	.dc.b	$3F,$42,$F3,$7D,$ED,$8F,$F1,$A7,$D1,$3C,$CC,$D4,$9F,$B3,$F0,$3F
	.dc.b	$DB,$5E,$71,$F2,$23,$1A,$7F,$9D,$9B,$9F,$F9,$E5,$B5,$FE,$96,$C7
	.dc.b	$C1,$5B,$EC,$FA,$3F,$42,$F3,$E9,$F5,$3E,$2F,$FD,$33,$97,$FA,$0E
	.dc.b	$75,$8F,$E3,$3E,$8B,$FD,$E3,$89,$B3,$A8,$FF,$44,$C6,$BE,$89,$F4
	.dc.b	$5F,$80,$6D,$F3,$F8,$FC,$4F,$D5,$FA,$27,$C6,$D3,$1F,$1B,$6F,$FC
	.dc.b	$B9,$0F,$A2,$FB,$8D,$8C,$FD,$16,$27,$3F,$25,$27,$8F,$D1,$5D,$BF
	.dc.b	$CA,$DF,$45,$FE,$9F,$9F,$F3,$43,$F9,$18,$CF,$C5,$FE,$26,$31,$C5
	.dc.b	$FA,$C7,$FD,$63,$5F,$E6,$FF,$53,$FD,$8C,$F2,$F8,$EF,$C8,$46,$30
	.dc.b	$FF,$32,$3F,$EE,$E3,$EA,$DF,$ED,$9C,$75,$7E,$5F,$E8,$5F,$45,$4C
	.dc.b	$7C,$55,$CB,$FC,$FD,$3F,$E3,$56,$FB,$7C,$7D,$5E,$4F,$E8,$7D,$DF
	.dc.b	$F1,$A4,$FF,$A1,$5B,$A7,$F9,$DD,$7D,$06,$FD,$CF,$D4,$7D,$17,$DD
	.dc.b	$DF,$F8,$5A,$DF,$3E,$06,$9F,$D7,$7C,$07,$F2,$FA,$FE,$22,$BE,$87
	.dc.b	$3E,$7E,$9D,$FC,$87,$0F,$76,$EF,$FE,$99,$BF,$F5,$23,$FE,$9B,$F9
	.dc.b	$B7,$C6,$FE,$ED,$7C,$75,$8C,$36,$78,$37,$8B,$FD,$29,$5C,$27,$A5
	.dc.b	$70,$69,$C0,$FF,$55,$7C,$DB,$E8,$BF,$85,$36,$33,$C5,$89,$91,$C1
	.dc.b	$DC,$63,$0F,$F7,$6F,$D1,$3E,$1B,$FD,$E1,$02,$46,$6F,$FB,$A3,$63
	.dc.b	$FB,$C3,$5E,$07,$09,$F8,$9F,$23,$FC,$BA,$78,$71,$7F,$42,$D1,$3E
	.dc.b	$AE,$6F,$5F,$1D,$CB,$DE,$DC,$BF,$D7,$5A,$FD,$B9,$FF,$48,$BF,$86
	.dc.b	$76,$7C,$6A,$E6,$D3,$33,$F4,$3E,$47,$EC,$1F,$67,$E4,$6C,$FD,$D8
	.dc.b	$B3,$F5,$A6,$E6,$AF,$7E,$9F,$C8,$AB,$E0,$87,$0E,$BD,$0F,$9A,$FC
	.dc.b	$C6,$BB,$7F,$A2,$7D,$67,$EF,$2B,$81,$FE,$E9,$71,$2B,$6C,$F9,$2F
	.dc.b	$C9,$7C,$97,$E4,$DB,$87,$E5,$DB,$7F,$90,$6D,$FD,$36,$CF,$FA,$F2
	.dc.b	$FA,$3F,$9F,$BF,$C7,$FC,$34,$8F,$C9,$FC,$1D,$C7,$EA,$CD,$C7,$FB
	.dc.b	$FF,$07,$F8,$DA,$FF,$33,$3F,$DC,$3E,$8D,$F4,$1F,$A9,$F1,$F0,$DA
	.dc.b	$FF,$04,$CE,$9F,$D8,$4E,$9A,$7E,$93,$E8,$FF,$46,$7E,$93,$E8,$EC
	.dc.b	$64,$99,$FE,$2A,$32,$FE,$32,$6E,$7E,$89,$B5,$F0,$FC,$5F,$DF,$78
	.dc.b	$DF,$BF,$6C,$7F,$03,$6E,$1F,$BC,$6D,$FE,$1E,$3F,$59,$F4,$4F,$A3
	.dc.b	$FB,$FF,$4D,$FE,$12,$8C,$61,$F7,$E3,$13,$A8,$FF,$15,$CE,$AB,$F0
	.dc.b	$51,$FE,$F2,$E5,$5F,$F1,$3F,$1E,$4F,$F1,$7F,$1E,$DF,$F1,$09,$B9
	.dc.b	$7E,$CE,$6D,$7E,$06,$6D,$7E,$3F,$9B,$F9,$74,$FF,$21,$9C,$DF,$E4
	.dc.b	$9F,$E4,$57,$8A,$FE,$6D,$5C,$CF,$11,$C4,$97,$E3,$72,$D2,$5F,$F0
	.dc.b	$7B,$EB,$3C,$23,$0D,$3F,$31,$7E,$19,$F1,$36,$35,$5E,$B8,$C3,$DA
	.dc.b	$38,$9A,$AF,$63,$6F,$E0,$CE,$CF,$89,$7D,$A7,$4D,$8D,$57,$26,$D7
	.dc.b	$28,$F4,$C6,$27,$DA,$DA,$F0,$5F,$9A,$B6,$7D,$78,$EE,$68,$B3,$DA
	.dc.b	$78,$47,$D1,$8F,$E9,$F1,$87,$D3,$1F,$42,$7A,$BB,$6C,$7C,$78,$EB
	.dc.b	$18,$3F,$88,$CF,$37,$DC,$79,$67,$D4,$78,$E7,$89,$D4,$F9,$EE,$A3
	.dc.b	$A5,$FC,$59,$D4,$F9,$8E,$46,$DC,$CF,$89,$EB,$1B,$8F,$11,$D2,$FE
	.dc.b	$A3,$67,$E4,$4F,$33,$83,$E4,$8D,$47,$81,$A7,$9A,$FE,$05,$FE,$7A
	.dc.b	$7E,$31,$CE,$7E,$1D,$73,$3C,$F3,$B8,$CC,$F0,$7A,$CE,$AB,$AD,$FE
	.dc.b	$12,$73,$5E,$63,$85,$74,$9F,$6B,$4F,$6C,$F0,$3B,$9F,$1E,$7C,$D7
	.dc.b	$E3,$3E,$83,$D6,$37,$19,$BF,$1D,$7E,$1C,$ED,$5F,$15,$B5,$F1,$5F
	.dc.b	$D5,$72,$3D,$8F,$F0,$BF,$B9,$FE,$18,$6C,$FF,$87,$1B,$7F,$DF,$9B
	.dc.b	$7F,$CA,$61,$AB,$D3,$FF,$55,$7D,$63,$1F,$E5,$51,$86,$E3,$FE,$D5
	.dc.b	$C1,$5B,$8F,$E5,$75,$B7,$FC,$5B,$F4,$BF,$E2,$CF,$63,$FC,$5A,$D8
	.dc.b	$C9,$FE,$32,$8C,$B4,$CE,$5F,$E1,$68,$CB,$63,$FC,$91,$F3,$5F,$E4
	.dc.b	$73,$FC,$91,$DF,$63,$24,$9F,$F2,$2C,$65,$FE,$7E,$7F,$8D,$7E,$95
	.dc.b	$F3,$9F,$E0,$8F,$88,$FF,$6D,$C6,$2F,$FC,$11,$18,$9D,$47,$FB,$D2
	.dc.b	$75,$5D,$35,$73,$7F,$B2,$A3,$FC,$AB,$F3,$9F,$E5,$D3,$FD,$B1,$F4
	.dc.b	$AF,$86,$C6,$3F,$E6,$39,$B5,$CE,$B5,$39,$24,$74,$7E,$A4,$6A,$30
	.dc.b	$7F,$8F,$BD,$3F,$F2,$E9,$FE,$5E,$F4,$ED,$7F,$98,$A2,$CF,$F9,$12
	.dc.b	$6E,$5F,$87,$FA,$4F,$F3,$9E,$5F,$E3,$4A,$FF,$2E,$46,$1C,$3F,$E7
	.dc.b	$AF,$96,$D4,$23,$4F,$F8,$F3,$EB,$1E,$FE,$A1,$18,$78,$B5,$F0,$35
	.dc.b	$FF,$22,$C3,$DE,$FF,$33,$E3,$F9,$D8,$C4,$FF,$9C,$63,$13,$AD,$3F
	.dc.b	$CD,$BF,$4B,$FA,$39,$CD,$6E,$7F,$CE,$37,$CA,$3F,$0E,$7A,$2F,$93
	.dc.b	$FE,$7D,$BE,$55,$C2,$73,$3B,$3A,$1F,$E8,$3A,$E9,$7F,$B3,$BE,$33
	.dc.b	$9F,$A5,$E3,$46,$98,$F9,$EB,$3D,$0F,$97,$18,$7C,$F2,$9F,$1F,$D2
	.dc.b	$BE,$4B,$E8,$36,$BE,$BB,$5F,$BC,$D7,$FC,$E6,$E7,$1F,$E1,$4D,$CF
	.dc.b	$FD,$2E,$D9,$F7,$7C,$57,$C2,$FA,$57,$1D,$24,$FA,$61,$C7,$88,$F9
	.dc.b	$F1,$C1,$D4,$F3,$9D,$63,$FC,$6E,$B5,$5C,$2B,$4F,$F1,$7F,$A5,$F9
	.dc.b	$F9,$7E,$0F,$E9,$7E,$7E,$06,$CE,$AB,$FD,$09,$1F,$EF,$67,$F9,$02
	.dc.b	$F8,$6F,$C0,$D7,$16,$C6,$E7,$F8,$13,$FE,$DC,$78,$FC,$0E,$8D,$FC
	.dc.b	$0F,$0F,$7E,$4F,$67,$9F,$BF,$D3,$DD,$93,$E2,$7B,$FB,$BE,$0F,$72
	.dc.b	$4F,$99,$F1,$FB,$9E,$4F,$0F,$CB,$EC,$4A,$F5,$C7,$A3,$D5,$F4,$FA
	.dc.b	$D2,$7D,$F1,$EC,$F5,$3D,$BE,$9C,$9F,$87,$1F,$03,$E4,$A4,$7C,$67
	.dc.b	$E1,$F9,$9F,$13,$CC,$93,$F2,$9F,$8F,$E4,$FC,$AE,$FF,$E4,$53,$FD
	.dc.b	$35,$AF,$F2,$B6,$F4,$7F,$A0,$4F,$CD,$7D,$2F,$EB,$BE,$11,$F0,$F4
	.dc.b	$97,$FA,$5F,$A5,$FE,$7A,$FE,$9F,$E8,$FF,$2C,$9A,$63,$E9,$5C,$3F
	.dc.b	$BA,$5A,$D2,$57,$EC,$7A,$9F,$0F,$FC,$03,$0F,$D8,$4F,$ED,$78,$FF
	.dc.b	$3A,$DF,$13,$5E,$3F,$F0,$6C,$33,$3F,$C3,$AF,$8B,$67,$55,$FC,$33
	.dc.b	$E9,$5F,$37,$8A,$F1,$4E,$63,$F8,$8F,$D2,$BA,$7C,$91,$2B,$F8,$44
	.dc.b	$E5,$B3,$97,$C5,$3C,$A2,$47,$C9,$8D,$72,$7F,$82,$EB,$94,$FF,$26
	.dc.b	$7E,$3B,$F3,$FE,$87,$E2,$1E,$4D,$E4,$64,$87,$17,$F9,$4F,$D2,$BE
	.dc.b	$33,$F5,$FF,$3B,$F1,$31,$FC,$13,$7C,$A6,$12,$F8,$3F,$C9,$5A,$FF
	.dc.b	$3A,$6D,$7F,$88,$DB,$3C,$A5,$0F,$8A,$FE,$77,$93,$E9,$6F,$AF,$2F
	.dc.b	$F9,$94,$E5,$B3,$C0,$8F,$F3,$7F,$A5,$E9,$97,$A3,$E8,$D7,$E1,$9C
	.dc.b	$D2,$72,$FB,$A7,$B3,$97,$D0,$62,$37,$09,$E7,$18,$6C,$F0,$BF,$81
	.dc.b	$3C,$23,$91,$E3,$C5,$87,$9E,$30,$DC,$3C,$97,$CA,$7A,$46,$4F,$15
	.dc.b	$72,$7D,$07,$D8,$C3,$C6,$F2,$38,$4F,$89,$D4,$F1,$C3,$D4,$D9,$EF
	.dc.b	$3D,$0F,$97,$7E,$37,$8E,$1E,$3C,$3E,$6C,$7D,$F3,$CB,$1F,$E1,$88
	.dc.b	$CC,$7B,$8F,$F1,$02,$F8,$E7,$D2,$ED,$B3,$A6,$BD,$67,$84,$FF,$87
	.dc.b	$E1,$F4,$B4,$8F,$F8,$8D,$B9,$FD,$AC,$3F,$45,$B7,$CB,$AF,$CF,$D7
	.dc.b	$CD,$9C,$C7,$C1,$32,$7C,$37,$35,$F8,$48,$CB,$D7,$81,$7D,$67,$CA
	.dc.b	$71,$8C,$BD,$4C,$CF,$19,$E1,$7E,$3A,$DC,$F7,$AD,$E9,$FE,$36,$9C
	.dc.b	$BF,$E4,$BA,$E1,$7C,$AB,$C0,$E1,$FF,$18,$E1,$EE,$3F,$D3,$4F,$AA
	.dc.b	$BA,$4F,$F8,$8E,$74,$E2,$7F,$C7,$26,$CF,$F2,$25,$FF,$4B,$8F,$E0
	.dc.b	$1F,$4B,$FB,$D7,$FC,$58,$D9,$FF,$20,$BF,$C3,$A3,$3A,$F1,$9F,$E2
	.dc.b	$33,$BB,$FF,$57,$36,$BA,$CE,$C4,$65,$E3,$C9,$98,$FE,$7B,$3F,$EA
	.dc.b	$FF,$A5,$FF,$1C,$3E,$49,$D1,$1D,$78,$96,$FC,$FF,$4B,$CC,$F3,$62
	.dc.b	$E5,$FF,$4A,$3E,$68,$D5,$FC,$07,$73,$A8,$F8,$B1,$AA,$EB,$1F,$EC
	.dc.b	$39,$D4,$7C,$29,$F7,$7D,$23,$CC,$A7,$8B,$E9,$5E,$5E,$31,$FF,$2D
	.dc.b	$56,$DF,$F2,$E7,$C1,$7F,$70,$73,$7F,$26,$31,$8F,$F9,$D1,$B9,$FF
	.dc.b	$B7,$5B,$5F,$CE,$EF,$A5,$6F,$AF,$E9,$7D,$2B,$CE,$A7,$D5,$F8,$C7
	.dc.b	$9D,$8C,$7F,$CD,$DF,$4B,$FE,$F1,$6F,$C7,$E7,$2F,$F4,$2C,$E5,$FF
	.dc.b	$43,$FD,$2B,$E0,$3F,$00,$DB,$E2,$7F,$9E,$5C,$4F,$FB,$27,$E9,$7F
	.dc.b	$CC,$F3,$FE,$EF,$7F,$CC,$AD,$F8,$0F,$B7,$18,$9F,$F6,$87,$C6,$FF
	.dc.b	$AB,$2B,$FD,$EC,$FE,$2A,$33,$F0,$9F,$82,$8C,$71,$FF,$8A,$DF,$F5
	.dc.b	$55,$7F,$BF,$7E,$F3,$F8,$1B,$CB,$E2,$3F,$13,$18,$C3,$FD,$BD,$F4
	.dc.b	$4F,$96,$FC,$6B,$6F,$FD,$3F,$8A,$3F,$0A,$D9,$FC,$9F,$A6,$FA,$2A
	.dc.b	$63,$E1,$AE,$1C,$A7,$FC,$66,$DF,$63,$8F,$A3,$C9,$FC,$EF,$D1,$3E
	.dc.b	$16,$FC,$36,$1B,$7E,$49,$FC,$03,$70,$E3,$D7,$C9,$FD,$3F,$BB,$FC
	.dc.b	$2B,$83,$FB,$0A,$FA,$1C,$F9,$9A,$7F,$5B,$F4,$5F,$9C,$C7,$E0,$B6
	.dc.b	$FF,$B4,$C7,$FB,$4E,$1F,$05,$C3,$C6,$FD,$AF,$C0,$FF,$83,$EB,$E4
	.dc.b	$B6,$F8,$9F,$DC,$9F,$EA,$51,$AF,$5B,$8F,$DA,$C8,$FA,$3F,$4F,$FC
	.dc.b	$5E,$F8,$F4,$30,$5F,$F0,$4F,$9B,$7D,$25,$31,$F0,$77,$2F,$B7,$3A
	.dc.b	$6C,$E9,$A4,$C7,$BC,$ED,$B5,$CD,$FF,$75,$FD,$1B,$E1,$3F,$92,$36
	.dc.b	$FF,$13,$F3,$D6,$B7,$5A,$3F,$A0,$BF,$E4,$0A,$D3,$FD,$ED,$AF,$F8
	.dc.b	$ED,$AF,$F8,$6B,$07,$D2,$7C,$1E,$4F,$84,$DC,$7D,$ED,$AF,$5B,$76
	.dc.b	$FF,$49,$64,$F8,$EF,$E5,$AD,$BF,$4F,$C7,$62,$64,$7B,$1D,$C6,$30
	.dc.b	$FA,$B5,$C5,$B3,$AA,$E0,$7F,$BF,$F3,$FE,$3B,$3F,$07,$D0,$FB,$5C
	.dc.b	$FF,$29,$95,$FE,$1D,$9D,$35,$57,$F2,$A7,$E4,$4F,$CC,$C9,$8B,$E3
	.dc.b	$BD,$BF,$1D,$EE,$6E,$3F,$59,$B8,$7A,$9B,$7F,$EB,$7D,$FF,$34,$DE
	.dc.b	$6F,$C8,$7C,$1F,$E4,$59,$87,$E3,$9C,$B4,$98,$7E,$32,$32,$D8,$FF
	.dc.b	$40,$D7,$F8,$5A,$1F,$46,$FA,$4F,$EE,$0D,$AF,$C6,$FC,$86,$26,$47
	.dc.b	$23,$FD,$05,$3B,$AF,$AB,$5A,$7F,$3B,$F4,$9F,$A9,$3F,$55,$F4,$96
	.dc.b	$FF,$4F,$16,$7F,$47,$37,$1F,$C4,$F8,$DF,$88,$F1,$FF,$12,$DC,$3F
	.dc.b	$42,$DB,$FC,$2B,$6B,$8C,$FE,$9A,$B8,$B4,$F5,$57,$FA,$93,$E8,$DF
	.dc.b	$49,$FE,$11,$6C,$FE,$F3,$5C,$4C,$8E,$34,$AD,$BF,$BC,$AD,$3F,$98
	.dc.b	$F1,$BF,$42,$7E,$8B,$C6,$B7,$FB,$A8,$B3,$F9,$E9,$B9,$7E,$17,$D7
	.dc.b	$FE,$25,$EC,$FF,$13,$6E,$1F,$C4,$1B,$7F,$C3,$1B,$5C,$67,$F8,$AD
	.dc.b	$71,$69,$C4,$B5,$C5,$B3,$C6,$3F,$8B,$CF,$16,$CF,$16,$9C,$4B,$3C
	.dc.b	$4F,$CF,$4F,$16,$CF,$16,$CF,$16,$9C,$63,$F3,$3C,$8A,$E5,$18,$6D
	.dc.b	$7F,$19,$6D,$2B,$85,$77,$38,$16,$B8,$36,$78,$57,$2A,$E3,$5E,$5A
	.dc.b	$E2,$D3,$89,$6B,$8C,$73,$9E,$31,$EB,$9D,$B4,$DB,$C6,$37,$3F,$02
	.dc.b	$FE,$DB,$4F,$AA,$59,$F9,$ED,$AF,$92,$D3,$CE,$F7,$AE,$B3,$FC,$FE
	.dc.b	$BA,$B6,$BA,$B4,$EA,$5A,$EA,$D8,$EA,$77,$7D,$AD,$8F,$69,$E1,$BF
	.dc.b	$98,$48,$F8,$CD,$9F,$84,$D9,$E1,$3A,$24,$E9,$B1,$A8,$C1,$1C,$12
	.dc.b	$3C,$97,$F0,$A3,$F0,$D7,$F0,$9B,$3F,$09,$B5,$F0,$9A,$7C,$22,$C7
	.dc.b	$C2,$8F,$88,$7C,$22,$C7,$C2,$8F,$0D,$FB,$5A,$7A,$8B,$3E,$96,$CF
	.dc.b	$A6,$39,$DF,$A4,$FF,$36,$47,$8D,$B3,$E3,$8F,$2C,$F8,$DA,$78,$DE
	.dc.b	$75,$E3,$6C,$F8,$DA,$78,$CB,$3E,$33,$9D,$78,$EF,$E5,$3E,$29,$F0
	.dc.b	$36,$7C,$07,$0B,$F0,$36,$FC,$0D,$AF,$01,$FC,$BF,$E8,$FF,$9B,$1F
	.dc.b	$E0,$CF,$1A,$D7,$F3,$4D,$53,$8A,$FF,$04,$D5,$BF,$EE,$1F,$47,$FE
	.dc.b	$3D,$F5,$3F,$C7,$CD,$C7,$FC,$76,$DB,$FF,$1B,$B6,$BF,$C3,$BE,$4F
	.dc.b	$F8,$70,$FF,$0E,$F6,$DB,$FF,$1A,$C5,$BF,$F0,$CC,$DA,$E3,$5A,$3F
	.dc.b	$C7,$71,$95,$33,$E8,$9D,$1F,$EE,$56,$CF,$F9,$43,$CA,$C4,$C8,$FE
	.dc.b	$AB,$3A,$AF,$F2,$7D,$65,$FF,$15,$FB,$9F,$E3,$43,$FD,$C3,$F4,$2F
	.dc.b	$80,$C6,$1F,$E4,$E9,$B5,$FE,$66,$AD,$4E,$4F,$F3,$0F,$B7,$FE,$67
	.dc.b	$3F,$CD,$1F,$25,$63,$24,$AF,$F3,$0C,$65,$FF,$31,$CD,$D7,$FC,$49
	.dc.b	$36,$BE,$1D,$3E,$A3,$FD,$A3,$1F,$E3,$76,$C7,$F8,$ED,$AF,$03,$86
	.dc.b	$3F,$E6,$45,$F4,$4F,$5B,$E0,$47,$F9,$0E,$31,$A7,$F9,$0D,$FF,$54
	.dc.b	$7C,$95,$AD,$D7,$5F,$79,$EF,$79,$B5,$FF,$17,$46,$27,$FC,$0E,$D7
	.dc.b	$FD,$2E,$3E,$87,$FC,$69,$3C,$2B,$96,$BF,$02,$F8,$4E,$6F,$FC,$23
	.dc.b	$7C,$2B,$5F,$43,$C2,$7F,$13,$68,$C4,$23,$0F,$93,$E6,$33,$18,$8F
	.dc.b	$2B,$70,$F5,$D6,$AB,$6C,$AF,$6B,$97,$FC,$4B,$F4,$2F,$8D,$EB,$F2
	.dc.b	$52,$73,$1E,$49,$D4,$7F,$9C,$5A,$FF,$9D,$5A,$FD,$B6,$BF,$84,$FA
	.dc.b	$17,$2F,$B9,$B1,$FE,$36,$6B,$F4,$17,$2E,$E6,$4E,$DE,$51,$CF,$CB
	.dc.b	$E9,$E5,$CA,$E9,$19,$7B,$46,$5E,$D1,$97,$C1,$19,$D7,$FC,$E5,$82
	.dc.b	$BE,$C4,$7C,$7F,$4A,$E7,$FE,$92,$C3,$FC,$AF,$39,$6C,$7F,$B5,$1B
	.dc.b	$EC,$F0,$23,$AF,$5F,$9F,$B1,$23,$8B,$DB,$D6,$EF,$E9,$78,$3D,$39
	.dc.b	$5D,$A3,$C7,$A7,$93,$69,$3E,$28,$F3,$F9,$FE,$8F,$36,$4F,$9A,$3D
	.dc.b	$7E,$64,$8F,$4B,$EE,$F2,$3D,$FE,$4C,$9F,$6B,$F0,$BC,$5F,$87,$CD
	.dc.b	$27,$E0,$BF,$1B,$9B,$E3,$FB,$7F,$2B,$F6,$5F,$E0,$1C,$CE,$EF,$F5
	.dc.b	$0C,$61,$FF,$51,$F3,$FF,$BF,$B1,$F9,$B2,$F0,$B9,$7C,$A4,$9C,$C7
	.dc.b	$5F,$B6,$ED,$F1,$72,$78,$C7,$EB,$BE,$7F,$F9,$BC,$7F,$A1,$BE,$05
	.dc.b	$F0,$7F,$B2,$6C,$7F,$A2,$A7,$6F,$F0,$4F,$9F,$7B,$7F,$B8,$F9,$FF
	.dc.b	$D1,$F0,$7F,$AE,$5B,$C1,$FC,$92,$B7,$7F,$E9,$4F,$83,$78,$5F,$C0
	.dc.b	$25,$EA,$53,$F8,$1C,$EA,$BF,$D3,$78,$3C,$9F,$E0,$FE,$D7,$E9,$A7
	.dc.b	$FD,$3E,$D7,$F8,$CC,$ED,$F0,$73,$F9,$EA,$46,$0F,$F5,$3F,$CF,$BE
	.dc.b	$5F,$C3,$52,$73,$1F,$C5,$BE,$7D,$F0,$1F,$06,$24,$FF,$AC,$27,$2D
	.dc.b	$9C,$BE,$49,$E5,$12,$3D,$B1,$AE,$0F,$F7,$15,$72,$9F,$E6,$6F,$C6
	.dc.b	$7E,$1F,$9B,$F4,$4F,$26,$D2,$32,$43,$8B,$FC,$BB,$E7,$DE,$57,$EF
	.dc.b	$3E,$67,$E3,$53,$1F,$CA,$36,$33,$1F,$EF,$5A,$E3,$3F,$EC,$E6,$BF
	.dc.b	$D2,$1B,$1B,$AF,$07,$93,$FC,$FB,$17,$6D,$ED,$98,$FF,$B6,$27,$2D
	.dc.b	$9E,$04,$7F,$DB,$BF,$3E,$BF,$44,$3E,$79,$7F,$DD,$2B,$59,$7F,$86
	.dc.b	$27,$35,$FE,$1A,$AE,$EF,$17,$FB,$D9,$88,$DC,$2B,$9B,$B3,$FC,$46
	.dc.b	$E2,$37,$3C,$4E,$78,$7F,$81,$A3,$9C,$EE,$7C,$D8,$7A,$A7,$8D,$6E
	.dc.b	$78,$DF,$A5,$DE,$16,$3C,$37,$C6,$77,$1E,$53,$E1,$D7,$37,$75,$C5
	.dc.b	$F6,$57,$17,$E2,$4E,$DE,$26,$F0,$F4,$3C,$8F,$F0,$D1,$CA,$AB,$4C
	.dc.b	$C7,$C8,$7F,$C4,$0B,$CC,$7C,$FF,$08,$C3,$63,$33,$C9,$AF,$6F,$9F
	.dc.b	$EE,$FF,$95,$1A,$9A,$47,$E2,$5F,$E8,$BB,$3E,$14,$71,$33,$B7,$E8
	.dc.b	$B5,$FC,$0C,$65,$FC,$1D,$71,$8F,$9C,$67,$5F,$F1,$75,$71,$7E,$2E
	.dc.b	$1F,$3E,$B8,$5F,$4B,$F3,$1E,$0C,$3C,$6E,$23,$BE,$1C,$E3,$FC,$91
	.dc.b	$7C,$E3,$11,$FE,$52,$8C,$57,$26,$C7,$53,$FC,$7B,$7E,$93,$F7,$EF
	.dc.b	$13,$4F,$F9,$02,$BF,$81,$B6,$BF,$D0,$2F,$09,$F8,$06,$27,$73,$FE
	.dc.b	$64,$F9,$FF,$DF,$BC,$0C,$6B,$AA,$FF,$45,$CF,$FA,$32,$75,$18,$9D
	.dc.b	$29,$E6,$FF,$95,$9B,$7E,$A7,$94,$66,$78,$C7,$F3,$DA,$EB,$3C,$F9
	.dc.b	$3C,$56,$F9,$FF,$F2,$25,$6A,$F8,$5F,$7B,$E6,$70,$AD,$C6,$6B,$87
	.dc.b	$CF,$F8,$0E,$BC,$8F,$A6,$C4,$E8,$95,$FC,$C5,$F9,$B3,$B7,$39,$7B
	.dc.b	$6F,$8D,$6C,$FE,$6D,$78,$9F,$B7,$F3,$CF,$9F,$F9,$2D,$C3,$FC,$6F
	.dc.b	$F4,$0F,$88,$CD,$49,$F8,$BF,$05,$FD,$D1,$F8,$F7,$C7,$2F,$F4,$2C
	.dc.b	$E5,$BD,$7F,$5A,$31,$97,$8B,$0C,$4F,$DA,$FA,$07,$C6,$D3,$EE,$FA
	.dc.b	$BF,$EA,$6B,$FC,$9C,$EB,$3F,$AF,$F4,$1F,$EF,$7C,$CD,$9D,$47,$FA
	.dc.b	$62,$35,$F4,$0F,$43,$F1,$2D,$BE,$7E,$CE,$27,$F0,$3F,$41,$FE,$76
	.dc.b	$9F,$F2,$DB,$FE,$72,$6F,$AD,$AF,$1F,$F0,$FE,$C7,$FA,$5E,$7F,$CC
	.dc.b	$2F,$E1,$B5,$FD,$3B,$9D,$3E,$B4,$23,$53,$FA,$A7,$FD,$6A,$FF,$9A
	.dc.b	$BE,$83,$FC,$88,$F4,$F8,$6F,$CA,$C6,$2F,$9F,$D0,$3E,$83,$F3,$2D
	.dc.b	$BF,$F3,$9B,$F8,$99,$DC,$FE,$7B,$D0,$7D,$05,$31,$F4,$17,$0F,$F3
	.dc.b	$DC,$FF,$8F,$9B,$EB,$71,$F4,$39,$3F,$A8,$F9,$F7,$D0,$5F,$8D,$C3
	.dc.b	$6F,$FD,$0B,$3F,$5B,$87,$F8,$5A,$C7,$97,$5F,$37,$F6,$7E,$E7,$F1
	.dc.b	$5B,$FE,$1E,$BE,$1B,$3E,$5E,$9F,$DC,$FD,$07,$E6,$EB,$FD,$26,$DE
	.dc.b	$2F,$A2,$BE,$57,$7E,$2F,$F0,$0F,$7F,$FC,$15,$7F,$E9,$A6,$DF,$D8
	.dc.b	$7F,$AB,$46,$7D,$5E,$1F,$07,$23,$FD,$43,$F0,$9F,$C4,$38,$7F,$93
	.dc.b	$DA,$BF,$86,$FC,$C3,$E8,$29,$8F,$82,$B8,$FF,$AA,$A7,$4D,$9D,$34
	.dc.b	$99,$FE,$32,$76,$DA,$E6,$FF,$BE,$BE,$31,$F0,$5F,$CD,$E0,$23,$37
	.dc.b	$FD,$E1,$B1,$FD,$F1,$AF,$03,$8B,$AC,$FF,$A2,$FC,$F3,$E8,$BF,$9B
	.dc.b	$CE,$6F,$FA,$3C,$E5,$F7,$7C,$77,$C4,$E5,$F8,$AD,$8C,$9F,$0E,$32
	.dc.b	$D3,$FD,$9F,$2F,$A0,$FE,$72,$D8,$CF,$D0,$E2,$64,$70,$77,$18,$CF
	.dc.b	$F6,$B5,$C5,$B3,$AA,$E0,$7F,$48,$C7,$F9,$04,$FC,$33,$E5,$7A,$1F
	.dc.b	$35,$B3,$F2,$E3,$13,$FC,$92,$3F,$B4,$FD,$37,$F0,$2A,$F9,$75,$FD
	.dc.b	$4A,$31,$AB,$17,$D0,$FB,$BE,$8B,$DE,$DC,$7E,$1B,$71,$F5,$B6,$FF
	.dc.b	$BC,$F1,$7E,$71,$BC,$FF,$94,$F8,$2F,$CA,$B3,$0F,$C9,$39,$69,$30
	.dc.b	$FC,$84,$65,$B1,$FD,$B3,$E8,$9F,$25,$F0,$DB,$1F,$E4,$46,$CF,$F9
	.dc.b	$21,$AF,$CB,$73,$87,$E3,$3D,$7F,$F0,$0C,$62,$BF,$17,$1F,$AC,$F9
	.dc.b	$39,$1F,$A9,$F9,$4B,$87,$E9,$66,$E1,$F0,$A6,$D7,$E5,$3C,$8F,$C4
	.dc.b	$F9,$3F,$8A,$6C,$7E,$39,$B8,$7E,$19,$B7,$FA,$06,$CF,$33,$F2,$B5
	.dc.b	$FE,$18,$FA,$17,$D1,$7D,$DF,$3B,$0D,$AF,$F0,$9C,$E9,$FF,$10,$E0
	.dc.b	$E5,$FE,$79,$97,$F7,$F1,$5B,$7F,$A6,$C6,$A7,$F6,$7E,$4F,$E9,$0F
	.dc.b	$D2,$F9,$37,$0F,$D4,$C5,$BF,$D0,$CD,$AF,$C3,$FA,$FF,$C6,$3D,$BF
	.dc.b	$E3,$2D,$C3,$F8,$B3,$6F,$F8,$93,$6B,$8C,$FF,$1B,$AE,$2D,$38,$96
	.dc.b	$B8,$B6,$78,$C7,$F1,$D9,$E2,$D9,$E2,$D3,$89,$67,$89,$FA,$09,$E2
	.dc.b	$D9,$E2,$D9,$E2,$D3,$8C,$7F,$97,$B9,$95,$CA,$30,$DA,$FF,$74,$4E
	.dc.b	$A7,$A9,$A2,$C6,$A7,$85,$2B,$84,$F8,$AB,$83,$4E,$0F,$29,$E1,$5C
	.dc.b	$AB,$8C,$7A,$2B,$67,$17,$75,$C6,$76,$FB,$6F,$EC,$34,$FA,$45,$AF
	.dc.b	$98,$DA,$F8,$CD,$3C,$EF,$84,$EF,$1E,$1A,$EE,$FC,$6A,$EE,$DA,$EE
	.dc.b	$D3,$B9,$6B,$BB,$63,$B9,$E1,$7D,$F1,$E1,$AE,$EF,$CD,$AE,$ED,$AE
	.dc.b	$ED,$AE,$ED,$3B,$96,$3B,$C7,$84,$EE,$F8,$63,$DE,$78,$6F,$EA,$12
	.dc.b	$3E,$83,$67,$E2,$36,$78,$4E,$89,$3A,$6C,$6A,$30,$47,$04,$8F,$25
	.dc.b	$FC,$21,$7F,$09,$B1,$F0,$9B,$5F,$09,$B5,$F0,$9B,$3F,$0A,$3E,$24
	.dc.b	$FC,$23,$E2,$3F,$0A,$3C,$37,$ED,$69,$EA,$2C,$FA,$5B,$3E,$98,$E7
	.dc.b	$5E,$9B,$F4,$12,$BD,$0D,$9F,$43,$6B,$D0,$D9,$F4,$57,$AA,$BD,$15
	.dc.b	$EA,$3D,$0F,$82,$FD,$0D,$AF,$43,$4F,$E6,$59,$BE,$8B,$FD,$93,$F4
	.dc.b	$FF,$9E,$9F,$E6,$B1,$8B,$FE,$84,$DB,$FE,$7E,$D9,$FF,$1D,$7C,$C7
	.dc.b	$F9,$1B,$D7,$FF,$23,$B6,$BF,$C8,$8D,$9D,$C7,$F9,$1E,$7F,$C4,$3E
	.dc.b	$64,$8F,$F1,$2F,$7D,$AF,$F0,$EC,$DD,$3F,$C6,$F3,$6B,$FC,$6B,$36
	.dc.b	$95,$FE,$11,$AC,$43,$E8,$9F,$47,$FE,$1F,$6E,$1F,$E2,$39,$FF,$26
	.dc.b	$C6,$1C,$35,$FF,$14,$39,$B4,$FF,$89,$3E,$13,$FD,$FF,$39,$8C,$1F
	.dc.b	$E4,$0F,$A1,$FF,$1F,$9F,$EF,$8F,$A3,$7C,$26,$27,$44,$C7,$FD,$63
	.dc.b	$3A,$8C,$63,$FE,$24,$F7,$FF,$CE,$A7,$F9,$DB,$DF,$B3,$FE,$73,$8B
	.dc.b	$7F,$E6,$F9,$B5,$C6,$B4,$7F,$9E,$63,$2F,$F9,$9F,$E8,$9E,$DF,$9A
	.dc.b	$FE,$34,$63,$2F,$91,$08,$FF,$55,$FC,$9E,$1D,$47,$CD,$FA,$77,$D1
	.dc.b	$70,$84,$61,$E6,$D7,$84,$6C,$7C,$1F,$F9,$EF,$4F,$A3,$3F,$E8,$49
	.dc.b	$CB,$FE,$86,$8C,$BE,$09,$E1,$9F,$FA,$0F,$E8,$F3,$5A,$BE,$37,$FE
	.dc.b	$17,$6C,$66,$F8,$CF,$FA,$4A,$B8,$DF,$FA,$36,$DF,$47,$AA,$F7,$B6
	.dc.b	$33,$1E,$59,$E1,$F3,$79,$A9,$39,$3D,$8F,$18,$CC,$79,$62,$E3,$E9
	.dc.b	$9F,$07,$D1,$BE,$8B,$E7,$36,$FE,$BB,$5F,$D0,$35,$FF,$42,$39,$AF
	.dc.b	$92,$DC,$3F,$D5,$4D,$9F,$A1,$F1,$5E,$FF,$A3,$72,$52,$4F,$AC,$43
	.dc.b	$93,$31,$FE,$45,$8E,$2E,$DA,$E6,$3B,$79,$9D,$FC,$C9,$3B,$8C,$BC
	.dc.b	$63,$2F,$58,$CB,$DE,$33,$9F,$E1,$AD,$5F,$7A,$3E,$57,$AB,$75,$FF
	.dc.b	$68,$61,$FE,$5E,$6B,$FC,$43,$5F,$C8,$47,$FB,$9E,$35,$EE,$71,$23
	.dc.b	$C7,$DA,$E9,$ED,$48,$E8,$F7,$F8,$3F,$07,$AB,$E1,$F5,$A5,$78,$63
	.dc.b	$C9,$AF,$97,$79,$3E,$58,$F4,$7A,$DE,$9F,$36,$4F,$A6,$3D,$9E,$54
	.dc.b	$8F,$63,$EF,$F2,$7E,$07,$93,$27,$E0,$3F,$0F,$E0,$7E,$27,$3C,$9F
	.dc.b	$88,$FC,$7E,$7F,$91,$F0,$1F,$2F,$E8,$FF,$DA,$1D,$1F,$B8,$0E,$39
	.dc.b	$FE,$4C,$EF,$C3,$E1,$F2,$F3,$2F,$F4,$6D,$A7,$A7,$D1,$F5,$ED,$95
	.dc.b	$A8,$FD,$97,$D1,$FF,$40,$8F,$C9,$7C,$2B,$E2,$3F,$6C,$D8,$FC,$B4
	.dc.b	$ED,$FF,$4C,$7D,$1B,$DD,$FD,$E7,$D1,$FE,$86,$4D,$9D,$47,$E7,$FE
	.dc.b	$1D,$E0,$FF,$03,$97,$A9,$4F,$E0,$D3,$AA,$FD,$35,$BC,$7F,$E1,$5E
	.dc.b	$D7,$E9,$E7,$F5,$4D,$7F,$D5,$93,$B7,$C1,$CD,$E6,$A9,$18,$3F,$65
	.dc.b	$F4,$6F,$9B,$F0,$D4,$9C,$C7,$F1,$9F,$A3,$7C,$17,$C0,$89,$3F,$BB
	.dc.b	$9C,$B6,$72,$F9,$27,$94,$48,$F6,$C6,$BA,$BF,$9A,$D7,$29,$FF,$63
	.dc.b	$3F,$19,$F8,$7F,$41,$F4,$0F,$26,$D2,$32,$43,$8B,$FC,$C7,$E8,$DE
	.dc.b	$4F,$EE,$FE,$77,$E3,$57,$ED,$CF,$F0,$3B,$8C,$7F,$A3,$D7,$09,$FE
	.dc.b	$30,$D7,$FD,$B4,$D8,$DD,$77,$F2,$3F,$A0,$62,$ED,$BD,$93,$1F,$E4
	.dc.b	$53,$96,$CF,$02,$3F,$C9,$FE,$8D,$7F,$E1,$C8,$7D,$12,$D1,$2B,$79
	.dc.b	$62,$B5,$3A,$AF,$03,$C9,$E2,$F3,$C2,$CF,$08,$5F,$47,$81,$DF,$0F
	.dc.b	$24,$EE,$FA,$B6,$F9,$C7,$6B,$E5,$3D,$5B,$1E,$23,$89,$67,$8D,$F7
	.dc.b	$C3,$D9,$3C,$F0,$F0,$BB,$9C,$9E,$FA,$E2,$FA,$A7,$6F,$33,$D6,$7C
	.dc.b	$13,$89,$4C,$DF,$99,$E0,$7C,$83,$A5,$FF,$88,$A3,$E3,$46,$5A,$FC
	.dc.b	$1F,$2D,$F4,$7B,$6C,$7F,$94,$65,$18,$8E,$4D,$7F,$CA,$DF,$47,$FE
	.dc.b	$29,$4D,$23,$FE,$2D,$7D,$77,$FA,$2F,$13,$DB,$3F,$88,$9C,$ED,$FE
	.dc.b	$2A,$D7,$F0,$13,$93,$E2,$DF,$17,$E3,$99,$8F,$C3,$46,$5E,$31,$E4
	.dc.b	$38,$CF,$5D,$7D,$31,$8C,$3C,$6E,$EB,$9D,$6E,$3E,$76,$1E,$6A,$DB
	.dc.b	$63,$C6,$71,$BF,$F2,$25,$71,$8F,$F4,$1B,$FD,$3E,$B2,$7F,$84,$E3
	.dc.b	$D5,$38,$7F,$82,$1C,$23,$FC,$31,$87,$C1,$3F,$C9,$71,$C0,$FF,$27
	.dc.b	$57,$57,$F6,$7F,$47,$FC,$06,$BF,$C4,$0D,$8E,$AF,$EE,$23,$3B,$71
	.dc.b	$94,$EE,$FF,$CB,$2D,$AF,$DB,$BC,$67,$9D,$71,$8F,$E7,$93,$E0,$E4
	.dc.b	$F2,$5B,$E8,$FF,$C8,$93,$92,$61,$E4,$AD,$CE,$A3,$13,$EA,$FA,$3E
	.dc.b	$C7,$4E,$47,$C1,$62,$74,$4A,$FE,$5D,$8A,$B7,$39,$8F,$6D,$F1,$AD
	.dc.b	$D6,$9F,$8F,$1F,$17,$E8,$9F,$1D,$4F,$6F,$D1,$BE,$63,$19,$7F,$93
	.dc.b	$EB,$6F,$F9,$47,$E0,$BF,$B8,$B9,$BF,$A1,$18,$C7,$F0,$2D,$CF,$FD
	.dc.b	$12,$DA,$FE,$4F,$7D,$2B,$7D,$7F,$67,$E8,$DF,$1D,$4F,$B9,$EA,$3E
	.dc.b	$1F,$18,$FF,$98,$FE,$8F,$FD,$F9,$89,$B8,$72,$9D,$7B,$1F,$85,$FA
	.dc.b	$37,$A5,$4F,$C4,$5F,$28,$CD,$7C,$1F,$2B,$E4,$1F,$82,$FA,$37,$AB
	.dc.b	$89,$FF,$3A,$7D,$EB,$E3,$BF,$24,$D8,$FF,$49,$72,$3D,$A7,$CA,$EA
	.dc.b	$FF,$31,$5F,$C6,$0F,$F3,$F1,$78,$3F,$03,$F4,$8F,$A3,$DF,$8F,$F9
	.dc.b	$FE,$DF,$F1,$95,$FF,$A1,$5B,$7F,$88,$FA,$4F,$F5,$5C,$E7,$DA,$FD
	.dc.b	$37,$DA,$3D,$CA,$7E,$7B,$8D,$F4,$98,$BF,$F4,$85,$32,$E2,$6F,$B2
	.dc.b	$CF,$E1,$FA,$FF,$1C,$7E,$A3,$E9,$1E,$CE,$27,$FD,$35,$E6,$FF,$89
	.dc.b	$1B,$EA,$EF,$C5,$FD,$D4,$E7,$87,$F8,$5F,$DE,$3E,$23,$F7,$CD,$C3
	.dc.b	$E6,$F1,$7F,$7E,$7F,$D4,$D1,$9A,$FE,$27,$E5,$FE,$26,$7F,$D5,$4D
	.dc.b	$8F,$D6,$FD,$2F,$E5,$5F,$F0,$1C,$67,$B3,$FD,$C3,$7F,$80,$BE,$BD
	.dc.b	$DF,$C1,$3E,$07,$FD,$3F,$18,$3F,$D6,$DE,$BF,$DE,$6F,$77,$F0,$AF
	.dc.b	$84,$FF,$38,$46,$0F,$F5,$E7,$C5,$FC,$66,$F1,$FF,$10,$FA,$1C,$6D
	.dc.b	$FC,$B2,$AE,$BF,$EF,$46,$DF,$F8,$49,$B2,$95,$62,$13,$4E,$AF,$F7
	.dc.b	$36,$BC,$EF,$F9,$BB,$FE,$26,$8C,$BD,$7E,$73,$E7,$5F,$FA,$4D,$FE
	.dc.b	$7D,$87,$F8,$B2,$11,$A8,$C3,$A9,$E5,$C5,$FC,$BA,$31,$5C,$23,$1E
	.dc.b	$B7,$C3,$FA,$29,$87,$EB,$33,$7D,$1E,$32,$FE,$4F,$0A,$C6,$3F,$EA
	.dc.b	$78,$DD,$F1,$9F,$E9,$19,$FE,$06,$FE,$2F,$9F,$E3,$F8,$EC,$B2,$FE
	.dc.b	$BB,$96,$92,$7F,$C3,$87,$26,$B4,$CD,$7F,$27,$7F,$C9,$38,$FD,$6F
	.dc.b	$A2,$FC,$07,$D1,$FE,$05,$B7,$EF,$6D,$7B,$9B,$5F,$47,$E0,$3E,$51
	.dc.b	$F2,$FC,$7B,$7F,$C4,$B8,$7F,$4D,$87,$F4,$C6,$DF,$D9,$AD,$CF,$F7
	.dc.b	$AF,$A3,$7C,$7F,$F6,$D6,$E1,$F4,$7E,$7B,$13,$23,$F0,$07,$F9,$92
	.dc.b	$77,$5F,$52,$B4,$F5,$AB,$87,$D5,$9B,$7F,$86,$F9,$1F,$C2,$1F,$85
	.dc.b	$F9,$2B,$1E,$B8,$B8,$7A,$E6,$DF,$DF,$F8,$EF,$D5,$78,$9F,$AB,$6E
	.dc.b	$1F,$89,$6D,$FE,$21,$B4,$9F,$F3,$EF,$D0,$BE,$8F,$FC,$1C,$D8,$F9
	.dc.b	$9A,$62,$64,$7C,$2C,$3F,$D1,$B5,$B7,$E8,$56,$9F,$F1,$3C,$66,$FC
	.dc.b	$11,$C2,$FF,$49,$5B,$9F,$CD,$79,$3F,$C1,$4F,$E0,$DE,$6D,$C7,$F4
	.dc.b	$91,$70,$FD,$1C,$DA,$FD,$B7,$B3,$FA,$FF,$6F,$F6,$0D,$C3,$F0,$CD
	.dc.b	$BF,$C3,$36,$BC,$B0,$CD,$F4,$3F,$10,$FE,$31,$E5,$DC,$3E,$15,$6E
	.dc.b	$70,$FE,$DA,$6D,$F4,$BE,$77,$C1,$B5,$B8,$C5,$F0,$7E,$AB,$E0,$72
	.dc.b	$48,$F9,$B1,$96,$C6,$6B,$C1,$39,$6C,$64,$EA,$7F,$0F,$BF,$05,$7D
	.dc.b	$CA,$B5,$FC,$BA,$73,$3E,$FA,$E4,$FF,$33,$ED,$FF,$17,$46,$23,$47
	.dc.b	$90,$F7,$16,$BC,$6D,$8F,$E5,$93,$CA,$30,$78,$E3,$BC,$F8,$DE,$D8
	.dc.b	$7C,$A8,$E6,$7F,$23,$BF,$3C,$66,$7D,$F5,$E2,$AD,$C6,$1B,$1E,$BA
	.dc.b	$E3,$87,$C2,$C3,$C9,$5A,$AF,$6B,$C0,$F2,$3E,$83,$37,$D4,$F6,$DF
	.dc.b	$47,$D7,$5D,$1D,$4F,$BC,$F8,$35,$D0,$F0,$61,$EB,$8D,$47,$BE,$7E
	.dc.b	$1E,$1E,$7C,$3B,$57,$2C,$3B,$3C,$2B,$BB,$E6,$3E,$03,$C8,$F2,$CF
	.dc.b	$2B,$F7,$57,$2A,$E0,$4C,$3C,$93,$E7,$BE,$B1,$E7,$9E,$D3,$97,$C4
	.dc.b	$43,$35,$D6,$78,$13,$0F,$CD,$5F,$47,$DC,$F9,$EB,$DD,$87,$9E,$3E
	.dc.b	$5E,$1D,$E7,$CD,$5C,$1F,$AD,$7D,$6B,$F4,$47,$3A,$F7,$DF,$3A,$F3
	.dc.b	$C7,$3A,$E1,$1E,$0A,$E0,$78,$30,$F5,$61,$C6,$F8,$36,$33,$3C,$2B
	.dc.b	$E8,$9E,$49,$F3,$57,$E0,$1B,$3F,$04,$FA,$A7,$19,$F9,$38,$79,$DB
	.dc.b	$19,$C3,$EA,$E1,$E0,$9E,$D3,$F0,$1A,$7A,$CF,$C8,$F3,$3E,$57,$FC
	.dc.b	$35,$EB,$FF,$86,$DB,$5F,$DC,$5B,$5F,$DC,$5B,$9F,$F8,$57,$D7,$FF
	.dc.b	$20,$1F,$E4,$1E,$EB,$7F,$E1,$F8,$B7,$FE,$1E,$9B,$5F,$CD,$3E,$85
	.dc.b	$D3,$FE,$E3,$6E,$1F,$E0,$1F,$1B,$13,$23,$F9,$AC,$EB,$0F,$F0,$25
	.dc.b	$65,$FF,$7D,$C6,$6F,$B5,$FF,$8D,$EB,$51,$FE,$2D,$F4,$7F,$CB,$07
	.dc.b	$FB,$A3,$E8,$5F,$03,$8C,$BF,$C7,$13,$72,$FF,$6E,$D6,$A7,$27,$F8
	.dc.b	$5F,$E5,$FF,$C8,$87,$F9,$17,$E4,$2C,$64,$95,$FE,$1B,$8C,$BF,$E6
	.dc.b	$18,$33,$91,$5F,$E5,$C8,$FF,$70,$36,$33,$3F,$E3,$86,$CF,$F8,$F3
	.dc.b	$E9,$FF,$D1,$0F,$43,$AB,$87,$0D,$73,$8F,$7F,$A1,$7D,$17,$C4,$9F
	.dc.b	$F2,$3C,$63,$4F,$F2,$3B,$FE,$B2,$F4,$2D,$6E,$BC,$03,$E7,$DF,$43
	.dc.b	$FE,$4A,$D3,$C9,$18,$9F,$EE,$2D,$7E,$21,$F1,$3E,$87,$FB,$DF,$0F
	.dc.b	$F7,$C9,$D3,$6F,$A5,$7C,$A9,$CD,$FC,$1B,$E2,$F8,$3E,$87,$75,$E5
	.dc.b	$6D,$1B,$F3,$38,$64,$F9,$9C,$BF,$E0,$8F,$A1,$7D,$07,$2F,$95,$93
	.dc.b	$98,$EF,$3A,$8F,$7C,$E9,$F7,$88,$D3,$86,$BF,$31,$AF,$DD,$6B,$F8
	.dc.b	$A6,$AF,$A1,$FC,$5B,$FE,$1E,$6B,$E0,$6E,$BF,$E2,$85,$CD,$F1,$9C
	.dc.b	$E9,$F8,$B6,$D7,$E6,$1A,$F7,$9E,$1E,$67,$1F,$32,$4F,$86,$32,$F8
	.dc.b	$E3,$2F,$8E,$3E,$AF,$DF,$BD,$2F,$17,$D3,$CA,$FF,$1B,$CE,$5B,$39
	.dc.b	$7D,$93,$CA,$24,$68,$87,$C1,$FA,$9F,$C1,$FC,$6F,$D0,$D7,$FC,$BA
	.dc.b	$3E,$9F,$F3,$6E,$1B,$B7,$EB,$67,$83,$FA,$FF,$B3,$7D,$4F,$E3,$BE
	.dc.b	$D1,$E8,$53,$F2,$33,$AC,$FF,$CA,$2B,$C8,$E1,$EE,$49,$CC,$7E,$63
	.dc.b	$EC,$DF,$11,$FA,$0D,$BC,$DF,$3D,$77,$3B,$9F,$CF,$FD,$A7,$FA,$9B
	.dc.b	$1F,$F2,$FE,$1E,$87,$F8,$8B,$54,$6D,$FC,$AF,$E2,$8F,$91,$F1,$78
	.dc.b	$12,$7F,$CD,$33,$97,$FC,$D7,$F7,$B5,$FD,$DE,$9F,$2F,$E6,$7C,$C7
	.dc.b	$EC,$A7,$51,$FE,$78,$F8,$77,$D7,$FE,$DB,$ED,$BF,$27,$87,$F9,$D9
	.dc.b	$AF,$EA,$27,$6F,$F9,$E9,$7C,$0F,$1E,$F9,$39,$8F,$E0,$3F,$7A,$FA
	.dc.b	$2F,$91,$E2,$CA,$FF,$41,$CE,$5F,$F4,$2F,$DB,$BE,$1F,$F8,$4F,$C6
	.dc.b	$3D,$2A,$7F,$0C,$9D,$57,$0F,$90,$DF,$D0,$C8,$C1,$FE,$90,$FA,$FF
	.dc.b	$EC,$6F,$FA,$4B,$E4,$BE,$EB,$7A,$FF,$8B,$CE,$A3,$F8,$3F,$D7,$3E
	.dc.b	$2F,$F8,$DF,$D7,$7D,$93,$ED,$46,$67,$FD,$3B,$F2,$6F,$AF,$FE,$43
	.dc.b	$2F,$83,$A7,$F2,$49,$D5,$7F,$A9,$2D,$EB,$7F,$28,$F2,$3E,$8F,$67
	.dc.b	$FA,$B9,$BE,$5F,$F4,$FA,$DD,$F8,$3D,$2F,$2E,$12,$30,$7F,$AC,$3E
	.dc.b	$BD,$F5,$3E,$1A,$93,$98,$FE,$6F,$F5,$FF,$72,$3C,$91,$F0,$0C,$F6
	.dc.b	$C8,$C9,$0F,$E5,$B3,$C5,$B1,$C5,$D6,$9F,$EC,$89,$D3,$76,$F9,$BD
	.dc.b	$92,$BF,$D8,$F5,$F2,$5B,$39,$7F,$9D,$E8,$AF,$4C,$FF,$50,$FA,$F7
	.dc.b	$5F,$EE,$7E,$AF,$E3,$E1,$FA,$66,$E5,$FE,$DA,$AD,$CF,$FB,$59,$AF
	.dc.b	$F4,$86,$C6,$EB,$B7,$6F,$F5,$EC,$7F,$81,$DF,$FB,$79,$7B,$F9,$35
	.dc.b	$E6,$49,$D3,$5F,$EA,$BF,$5D,$FA,$1C,$7F,$C2,$FF,$5D,$ED,$9F,$F0
	.dc.b	$CC,$E5,$46,$5F,$85,$5F,$E1,$38,$C6,$1E,$49,$DB,$FE,$17,$AE,$53
	.dc.b	$C5,$B1,$FE,$1B,$38,$47,$2A,$E9,$1C,$71,$61,$E6,$8C,$37,$0F,$F1
	.dc.b	$7C,$72,$3A,$57,$28,$F5,$D7,$27,$CA,$62,$B9,$3E,$C2,$D7,$37,$47
	.dc.b	$01,$87,$A9,$F0,$DF,$77,$91,$B8,$D3,$C2,$FC,$4F,$F8,$B2,$3E,$73
	.dc.b	$63,$2F,$B8,$FF,$18,$9E,$0F,$97,$7D,$7E,$23,$0D,$A4,$EA,$78,$C6
	.dc.b	$27,$84,$7F,$95,$BE,$86,$A6,$91,$DB,$F1,$B2,$FC,$FF,$47,$C8,$9E
	.dc.b	$47,$C5,$8C,$BF,$1C,$CE,$9F,$89,$8C,$BF,$8B,$8C,$E9,$FE,$5D,$9C
	.dc.b	$BF,$02,$FC,$1B,$30,$F0,$5F,$9C,$F0,$46,$5F,$01,$9C,$3C,$11,$97
	.dc.b	$C1,$19,$BF,$04,$E5,$F0,$47,$4A,$EF,$7E,$26,$47,$F8,$FB,$0F,$79
	.dc.b	$E1,$7F,$CE,$D3,$A8,$FF,$2F,$DF,$F8,$4A,$BB,$9F,$E8,$68,$E1,$5F
	.dc.b	$08,$C4,$ED,$FF,$45,$FD,$7F,$F8,$0A,$B8,$1F,$E8,$37,$11,$FE,$50
	.dc.b	$7F,$D2,$31,$A8,$E3,$B7,$35,$3C,$2F,$D0,$DA,$E9,$3B,$8C,$3E,$2E
	.dc.b	$6C,$1E,$5A,$7D,$0F,$F9,$A2,$FC,$15,$DA,$B7,$1F,$97,$BF,$E8,$AD
	.dc.b	$8F,$F5,$24,$62,$7F,$3E,$BE,$93,$C4,$77,$E4,$FE,$FD,$7D,$A7,$4D
	.dc.b	$9D,$BF,$07,$15,$70,$BF,$8D,$5A,$7B,$BE,$22,$57,$EB,$7E,$B9,$E2
	.dc.b	$FF,$AE,$5B,$19,$8C,$1F,$EB,$DF,$8A,$FE,$E5,$1F,$09,$B7,$FE,$84
	.dc.b	$AE,$2F,$FA,$1A,$33,$8F,$E2,$5D,$C7,$FA,$C1,$AF,$FB,$05,$AF,$FB
	.dc.b	$39,$AB,$EB,$9F,$37,$FE,$CF,$6E,$1F,$C1,$DB,$3F,$C3,$1A,$FF,$AF
	.dc.b	$BE,$B9,$F3,$7F,$ED,$76,$D7,$F0,$F6,$CF,$F1,$46,$BF,$EC,$6F,$AE
	.dc.b	$7A,$DF,$ED,$F6,$D7,$F1,$76,$CF,$F1,$C6,$BF,$EC,$ED,$3F,$CD,$B3
	.dc.b	$F4,$7C,$5F,$A0,$47,$F9,$07,$D7,$7F,$9B,$5B,$EF,$7F,$BB,$63,$17
	.dc.b	$FC,$A1,$AF,$FB,$6A,$7F,$92,$9F,$CA,$4A,$FF,$2E,$6B,$BE,$DF,$F7
	.dc.b	$C7,$17,$F9,$DA,$7F,$98,$36,$7F,$9A,$B5,$FF,$73,$B5,$FF,$7C,$7D
	.dc.b	$57,$F9,$BE,$7F,$9C,$36,$3F,$9E,$B5,$FF,$77,$B5,$E3,$3F,$C1,$90
	.dc.b	$8F,$D0,$BC,$A3,$4E,$5F,$CC,$7A,$52,$BF,$38,$D9,$FE,$5D,$E3,$70
	.dc.b	$EC,$FF,$37,$D7,$F4,$D6,$CE,$A3,$FA,$87,$D4,$3D,$7E,$ED,$8C,$C6
	.dc.b	$0F,$1F,$67,$F9,$B6,$3F,$AC,$36,$BF,$AE,$B5,$F6,$9E,$E7,$0E,$29
	.dc.b	$E9,$F8,$09,$5E,$47,$E0,$F8,$AF,$A8,$A4,$9F,$ED,$11,$97,$FB,$57
	.dc.b	$D3,$3C,$FF,$E4,$CD,$8F,$E1,$5F,$3F,$72,$FF,$7E,$CD,$EE,$FE,$80
	.dc.b	$E6,$3D,$21,$C3,$EC,$6B,$F6,$9A,$EE,$FF,$96,$57,$FB,$42,$FF,$95
	.dc.b	$D7,$77,$FA,$44,$66,$BF,$D6,$7F,$3F,$F4,$5B,$E2,$FF,$35,$8C,$63
	.dc.b	$FE,$E2,$7F,$AB,$DF,$F5,$17,$2D,$9E,$1B,$F1,$7C,$7F,$0B,$27,$E1
	.dc.b	$FD,$45,$C7,$FD,$87,$92,$B1,$97,$FB,$C2,$15,$8B,$F8,$5E,$77,$CE
	.dc.b	$3C,$DF,$51,$63,$DF,$1F,$9A,$D7,$F8,$E9,$1C,$1F,$D4,$7E,$1F,$FA
	.dc.b	$46,$3F,$4A,$31,$E4,$FC,$AF,$A8,$F9,$7F,$1D,$71,$FB,$6D,$DB,$D2
	.dc.b	$DC,$3D,$0F,$C9,$F4,$65,$7F,$4B,$E2,$FE,$0F,$5F,$2E,$73,$7F,$49
	.dc.b	$FA,$1D,$D8,$9F,$F0,$27,$D4,$3E,$9F,$FB,$93,$71,$FC,$6F,$D0,$E2
	.dc.b	$64,$7F,$8F,$67,$58,$FF,$90,$9A,$FE,$39,$AF,$EA,$1D,$DF,$D1,$F3
	.dc.b	$79,$CF,$D6,$7F,$09,$F2,$92,$FD,$9F,$0B,$6B,$F1,$73,$6F,$F1,$13
	.dc.b	$6B,$F2,$3E,$FF,$E5,$BC,$2F,$CB,$B6,$3F,$4E,$DC,$3F,$0A,$DB,$FF
	.dc.b	$13,$7D,$2B,$EA,$3F,$C2,$8D,$BF,$9D,$AE,$26,$47,$F9,$7E,$75,$87
	.dc.b	$F9,$91,$AF,$EC,$1A,$FF,$0A,$77,$7F,$E4,$51,$CD,$FC,$06,$7F,$D1
	.dc.b	$B5,$BB,$FC,$67,$81,$FB,$2F,$46,$4F,$F0,$2F,$2E,$DF,$E8,$26,$E3
	.dc.b	$F9,$F9,$B5,$FA,$DF,$43,$F6,$5F,$51,$FB,$36,$C7,$E2,$DB,$87,$EE
	.dc.b	$9B,$7F,$1B,$EA,$3E,$29,$FC,$13,$E8,$AD,$FF,$8B,$B6,$4E,$30,$FD
	.dc.b	$9D,$5B,$E5,$7B,$AE,$4D,$25,$7C,$23,$FA,$6D,$F4,$8F,$E4,$37,$CD
	.dc.b	$F9,$95,$6B,$FD,$6B,$E0,$AB,$10,$BE,$4F,$53,$15,$E2,$71,$1A,$C3
	.dc.b	$CF,$18,$9D,$46,$09,$87,$88,$F2,$9E,$A7,$33,$EA,$9E,$33,$B8,$E8
	.dc.b	$EC,$FE,$57,$7E,$8B,$F1,$46,$EF,$F8,$D4,$E6,$FE,$04,$30,$F8,$15
	.dc.b	$CF,$0F,$03,$67,$57,$C6,$31,$5C,$2B,$C6,$4A,$F4,$9C,$27,$A4,$7C
	.dc.b	$87,$8D,$73,$8E,$11,$EA,$37,$87,$AA,$3C,$67,$E1,$23,$38,$7A,$B0
	.dc.b	$F1,$61,$F2,$2B,$B1,$D0,$FB,$B8,$77,$8F,$2C,$F0,$BF,$8C,$4A,$EF
	.dc.b	$3F,$22,$3B,$9E,$82,$C7,$58,$F7,$3D,$CF,$09,$FB,$BB,$FF,$0D,$E1
	.dc.b	$F8,$A8,$C5,$7F,$03,$9C,$91,$C1,$23,$F3,$94,$FB,$26,$7F,$69,$B5
	.dc.b	$F5,$DB,$5E,$56,$D7,$D4,$6C,$FB,$E3,$A5,$73,$24,$73,$6D,$73,$6C
	.dc.b	$F3,$6C,$F3,$AE,$B5,$F2,$A3,$FD,$8D,$18,$8F,$AB,$5F,$C6,$E3,$18
	.dc.b	$7D,$F8,$F1,$D7,$84,$95,$E1,$6C,$78,$5A,$78,$6F,$C7,$3E,$11,$3D
	.dc.b	$1E,$44,$79,$36,$79,$1D,$1E,$51,$C4,$D9,$23,$71,$C4,$DD,$FB,$30
	.dc.b	$FF,$67,$36,$7F,$D7,$8D,$9F,$BA,$DA,$F8,$97,$EA,$9F,$8F,$5F,$2A
	.dc.b	$31,$87,$C7,$BE,$13,$93,$FB,$F7,$D4,$7F,$8A,$7E,$A7,$FC,$54,$D8
	.dc.b	$C9,$FE,$03,$8C,$B4,$FF,$23,$FA,$5F,$E3,$A3,$FC,$77,$E1,$58,$C9
	.dc.b	$32,$FF,$14,$C6,$5F,$E4,$DF,$52,$ED,$FF,$7C,$36,$BF,$C2,$5F,$53
	.dc.b	$89,$91,$FE,$50,$AF,$F4,$2D,$6D,$C3,$5F,$F2,$AB,$5F,$F3,$7B,$BD
	.dc.b	$BF,$B3,$73,$FF,$B9,$D5,$6A,$FF,$C6,$1F,$1F,$FE,$56,$F4,$A4,$FF
	.dc.b	$BB,$FE,$A5,$EE,$63,$3F,$F2,$2D,$5C,$7F,$B9,$56,$AF,$FC,$DB,$3F
	.dc.b	$E5,$7F,$A9,$91,$FE,$2F,$F8,$8B,$87,$F9,$7A,$6E,$1F,$E5,$B9,$B5
	.dc.b	$FE,$88,$FC,$73,$A5,$FC,$36,$BF,$E7,$46,$BB,$CB,$FD,$4F,$F5,$2F
	.dc.b	$A9,$F4,$CF,$CB,$8C,$4F,$CD,$84,$3E,$B5,$EB,$66,$37,$3F,$4F,$EC
	.dc.b	$DF,$37,$C6,$11,$87,$A3,$5E,$11,$B1,$F5,$7F,$EA,$4D,$BE,$9C,$FF
	.dc.b	$B0,$A7,$2F,$FB,$12,$32,$F8,$27,$86,$BF,$EA,$CF,$AB,$CD,$6A,$F8
	.dc.b	$DF,$F7,$56,$C6,$6F,$8C,$FF,$B4,$2B,$8D,$FF,$AD,$AD,$F5,$7A,$AF
	.dc.b	$7B,$63,$31,$E5,$9E,$1F,$27,$9A,$93,$93,$D8,$F1,$8E,$9F,$54,$FA
	.dc.b	$BE,$9F,$1D,$2F,$80,$9D,$38,$6B,$FA,$56,$BF,$EB,$66,$BF,$ED,$06
	.dc.b	$BF,$EE,$E7,$8E,$9F,$31,$B8,$EE,$FD,$DF,$13,$F0,$7E,$AD,$CF,$49
	.dc.b	$8F,$9A,$1C,$F8,$8F,$F1,$24,$70,$75,$3C,$E7,$A7,$A5,$D7,$D2,$95
	.dc.b	$CE,$32,$F5,$8C,$BD,$23,$EA,$7D,$5B,$D9,$F3,$7D,$5C,$AF,$93,$39
	.dc.b	$6C,$FE,$5E,$31,$1F,$C3,$5A,$FF,$B7,$9A,$FF,$BE,$1D,$EF,$F9,$26
	.dc.b	$ED,$9A,$F8,$DF,$57,$F8,$7F,$AB,$F9,$F8,$1A,$4E,$0F,$D4,$7D,$4B
	.dc.b	$EA,$E9,$2B,$F6,$91,$96,$F9,$3C,$08,$E0,$FD,$D7,$D5,$3E,$87,$F6
	.dc.b	$8D,$8C,$C6,$0F,$DC,$7D,$57,$FA,$E6,$3F,$81,$35,$FE,$60,$D7,$FE
	.dc.b	$00,$B5,$13,$EA,$5F,$55,$FC,$0D,$B1,$FC,$35,$AF,$F3,$B6,$BC,$8E
	.dc.b	$7F,$55,$87,$0D,$73,$3F,$C3,$E3,$10,$F8,$C7,$D5,$7F,$13,$6C,$7F
	.dc.b	$1D,$6B,$FD,$2D,$AF,$84,$E5,$F0,$3C,$48,$F1,$7F,$8F,$7D,$4B,$CE
	.dc.b	$A4,$9F,$E5,$51,$97,$F9,$67,$D4,$BE,$1F,$F9,$33,$63,$31,$83,$3F
	.dc.b	$23,$23,$8B,$5F,$EC,$5F,$51,$FD,$A5,$FE,$70,$D7,$3E,$9F,$F3,$38
	.dc.b	$C4,$7B,$7E,$9D,$F0,$FF,$CD,$DB,$1F,$D0,$5A,$FF,$6F,$6B,$F1,$7E
	.dc.b	$4D,$F5,$1F,$CF,$DB,$1F,$D2,$5A,$FF,$77,$6B,$F3,$7E,$99,$E5,$FF
	.dc.b	$47,$6C,$7F,$51,$6B,$FD,$FD,$AF,$87,$C6,$F0,$11,$F0,$3F,$D4,$7E
	.dc.b	$A1,$F2,$34,$93,$FD,$76,$32,$FF,$5F,$FA,$87,$27,$F3,$DC,$33,$3F
	.dc.b	$EB,$B6,$F8,$DF,$D4,$2B,$73,$FC,$D1,$AE,$F0,$E4,$7F,$47,$8C,$CF
	.dc.b	$1B,$FE,$CD,$1C,$2B,$F2,$0D,$7F,$44,$D7,$F5,$AB,$DF,$F3,$79,$F2
	.dc.b	$7C,$51,$DC,$C3,$E9,$8C,$3F,$D2,$BE,$A1,$F3,$1F,$D8,$31,$7D,$1F
	.dc.b	$D0,$7F,$B8,$98,$DF,$FB,$05,$70,$BE,$57,$82,$3D,$79,$FF,$B3,$62
	.dc.b	$FA,$5B,$ED,$4A,$FE,$B7,$39,$6C,$F0,$23,$FD,$83,$EA,$3F,$51,$9F
	.dc.b	$F0,$EF,$A7,$5F,$F0,$79,$CD,$27,$2F,$F1,$49,$EC,$E5,$FE,$1C,$62
	.dc.b	$37,$09,$E7,$18,$6C,$F0,$BF,$E2,$D3,$C2,$39,$1E,$3C,$7F,$D1,$77
	.dc.b	$DC,$FE,$4B,$7C,$9F,$25,$F2,$AE,$91,$ED,$79,$61,$63,$8D,$F4,$C3
	.dc.b	$C1,$7C,$67,$99,$B3,$D5,$3B,$BE,$71,$E3,$3E,$4C,$F1,$69,$C7,$0F
	.dc.b	$33,$CB,$0F,$F2,$24,$7C,$E8,$FE,$4C,$D7,$A9,$8F,$96,$7D,$47,$06
	.dc.b	$C7,$F9,$E6,$51,$88,$E1,$3F,$E4,$95,$F4,$14,$8F,$F9,$35,$B9,$FD
	.dc.b	$8C,$3F,$37,$E8,$FC,$1A,$FE,$A7,$5F,$0E,$73,$1F,$04,$C9,$F0,$DC
	.dc.b	$D7,$E0,$A3,$2F,$5D,$BF,$CA,$6E,$63,$CA,$71,$8C,$E9,$FE,$95,$BF
	.dc.b	$04,$62,$7F,$A6,$46,$30,$F3,$47,$2B,$ED,$18,$9D,$4F,$9C,$D5,$FF
	.dc.b	$99,$6F,$A4,$7F,$B6,$1F,$DB,$D6,$4F,$F2,$3C,$7A,$A7,$0F,$FB,$7A
	.dc.b	$2E,$1F,$E6,$6A,$E3,$3E,$63,$D9,$3F,$E5,$0F,$A8,$FD,$D3,$75,$DC
	.dc.b	$A7,$FC,$03,$39,$24,$77,$3F,$C0,$B1,$C9,$C1,$C4,$E6,$E6,$FF,$CF
	.dc.b	$AF,$F3,$A9,$F1,$73,$79,$A5,$F4,$3F,$E6,$79,$C9,$2F,$C1,$3A,$F9
	.dc.b	$0E,$B7,$C6,$D1,$E0,$FA,$29,$87,$F9,$D6,$FF,$91,$F8,$D8,$C7,$CF
	.dc.b	$92,$F1,$5F,$C9,$B9,$FF,$D9,$58,$7F,$86,$6B,$4F,$C9,$AF,$44,$FF
	.dc.b	$A2,$A3,$07,$77,$17,$F1,$3E,$95,$F3,$7F,$3B,$E1,$FF,$A0,$5F,$FA
	.dc.b	$0A,$72,$FF,$A1,$23,$35,$F3,$1B,$97,$FA,$99,$B3,$FE,$2D,$AE,$35
	.dc.b	$BE,$BE,$C7,$DC,$FA,$77,$B9,$8B,$FF,$48,$C6,$1F,$F4,$9F,$D4,$3E
	.dc.b	$6B,$87,$CC,$FE,$12,$31,$3C,$23,$2F,$FA,$6B,$EA,$1F,$09,$8C,$FF
	.dc.b	$07,$18,$9D,$4F,$E3,$A3,$F0,$53,$A8,$C7,$37,$E4,$B5,$7C,$DF,$E1
	.dc.b	$7E,$57,$F2,$77,$F2,$E3,$0F,$E6,$3E,$A1,$EE,$6F,$A3,$0D,$D3,$FD
	.dc.b	$5E,$7F,$16,$3F,$09,$F6,$3F,$E4,$C7,$3E,$17,$E4,$E3,$18,$7E,$73
	.dc.b	$A7,$FC,$E1,$5F,$8B,$FB,$0F,$F0,$FB,$FE,$BA,$8C,$CF,$E5,$A7,$2F
	.dc.b	$E9,$3E,$AB,$FC,$D3,$5F,$93,$D3,$FD,$F0,$FF,$B0,$A3,$33,$F9,$E9
	.dc.b	$CB,$C6,$FF,$4D,$1F,$12,$B8,$46,$23,$E4,$7C,$AF,$EA,$FE,$C3,$FD
	.dc.b	$21,$5F,$8D,$8C,$7C,$1F,$EF,$BC,$5F,$F1,$2C,$FE,$96,$BE,$4B,$FE
	.dc.b	$D4,$8C,$E3,$FB,$09,$CB,$FB,$EF,$AB,$FF,$3A,$57,$EB,$7E,$0B,$E6
	.dc.b	$4E,$7D,$DF,$DE,$46,$3C,$0F,$F7,$EF,$8A,$D3,$F8,$07,$D8,$5C,$7F
	.dc.b	$DC,$73,$71,$FD,$FC,$DB,$57,$FA,$2F,$27,$B9,$FC,$27,$EB,$5F,$1B
	.dc.b	$FC,$6A,$73,$5F,$C0,$67,$2F,$FB,$C3,$D1,$7D,$27,$F1,$1F,$9E,$FE
	.dc.b	$2D,$3F,$E9,$79,$CB,$FC,$26,$33,$A7,$FA,$93,$EA,$9E,$E7,$F2,$86
	.dc.b	$D7,$F0,$FF,$A4,$B5,$FC,$92,$6E,$DE,$A8,$E5,$7F,$C8,$6B,$EB,$C6
	.dc.b	$E3,$1C,$DD,$5B,$87,$FA,$D2,$F9,$53,$0E,$B3,$9C,$39,$D7,$F9,$03
	.dc.b	$EA,$7E,$BD,$FF,$2A,$7F,$B4,$C6,$79,$7F,$99,$46,$2F,$B3,$C2,$79
	.dc.b	$7A,$5A,$AD,$BE,$9F,$AD,$97,$F0,$7E,$B6,$E3,$FC,$E3,$25,$63,$1F
	.dc.b	$F2,$9C,$2B,$17,$FE,$C7,$CF,$F8,$EE,$1F,$3F,$CE,$F2,$FD,$6F,$CF
	.dc.b	$65,$7F,$3D,$FA,$3F,$E1,$79,$7C,$4C,$38,$A8,$CC,$F5,$AF,$EA,$11
	.dc.b	$8C,$3B,$4F,$3F,$27,$EA,$7D,$6F,$D5,$F9,$5B,$3F,$75,$B5,$ED,$6E
	.dc.b	$3E,$C7,$E5,$FB,$B2,$BF,$A9,$71,$7E,$8D,$F9,$B1,$9B,$FA,$8F,$D2
	.dc.b	$F1,$F1,$87,$DE,$FA,$D7,$D6,$FF,$76,$6E,$1F,$8F,$F8,$BC,$4C,$8F
	.dc.b	$C9,$7B,$17,$2F,$A4,$DC,$3E,$8B,$67,$F1,$0F,$E0,$BD,$79,$1E,$8F
	.dc.b	$02,$DF,$E2,$66,$DF,$E3,$66,$D7,$E4,$7E,$2B,$F3,$9F,$0B,$F9,$D6
	.dc.b	$C7,$E9,$5B,$87,$E1,$5B,$7F,$9C,$FA,$A7,$D6,$FE,$7A,$84,$8C,$DF
	.dc.b	$FB,$F6,$31,$3B,$6C,$65,$4F,$07,$FC,$17,$CB,$FB,$E7,$FC,$33,$19
	.dc.b	$BF,$F5,$3C,$E5,$F2,$5F,$E2,$3C,$2F,$F1,$3D,$6F,$0F,$DA,$5E,$DF
	.dc.b	$D9,$7B,$5F,$BD,$3F,$7D,$E6,$D8,$FE,$0D,$17,$2F,$D3,$4D,$CB,$F7
	.dc.b	$1E,$CF,$ED,$3E,$B3,$F6,$AD,$C3,$F1,$6D,$BF,$DE,$36,$BE,$57,$C0
	.dc.b	$7F,$8B,$7E,$B1,$87,$A2,$30,$DC,$3F,$C4,$D3,$A7,$F8,$71,$F4,$30
	.dc.b	$F9,$35,$C2,$33,$1F,$4E,$30,$FC,$A6,$DF,$8A,$7A,$DF,$7A,$FF,$1D
	.dc.b	$DF,$77,$84,$E8,$EA,$F8,$1C,$3A,$38,$CF,$3A,$FF,$6E,$D7,$37,$C1
	.dc.b	$3A,$AE,$7A,$7F,$91,$9E,$73,$A3,$8D,$F3,$8F,$E5,$D7,$CD,$F8,$55
	.dc.b	$C6,$FA,$1E,$3B,$E3,$3D,$B0,$F8,$2F,$06,$CF,$C7,$8F,$09,$92,$B9
	.dc.b	$3A,$CF,$96,$3C,$78,$79,$65,$2C,$3C,$D1,$C6,$3C,$6F,$96,$B8,$9E
	.dc.b	$53,$BB,$F3,$4E,$EF,$8D,$A7,$37,$8B,$CC,$F0,$CF,$81,$E5,$3E,$03
	.dc.b	$D0,$EA,$BD,$11,$E7,$6C,$F1,$C3,$E6,$4F,$1B,$F1,$36,$FC,$47,$38
	.dc.b	$F5,$3E,$63,$C2,$F9,$A3,$C2,$77,$76,$F3,$8D,$9C,$CF,$63,$B6,$D7
	.dc.b	$97,$0F,$B3,$1C,$27,$47,$69,$D5,$7C,$CB,$F5,$BE,$33,$D6,$5A,$F5
	.dc.b	$D7,$28,$CC,$78,$E3,$A6,$1F,$18,$93,$EA,$39,$4F,$8E,$FC,$2F,$C3
	.dc.b	$BF,$0B,$CE,$7A,$DF,$78,$F0,$D7,$08,$F8,$95,$C1,$D4,$F5,$BF,$0B
	.dc.b	$F1,$1B,$1E,$6B,$F2,$4F,$73,$AC,$EA,$3C,$37,$DD,$F4,$DF,$78,$C5
	.dc.b	$71,$3C,$13,$D2,$FC,$75,$BB,$E9,$5D,$AB,$57,$CA,$F7,$87,$D8,$C3
	.dc.b	$85,$6B,$0E,$38,$76,$3D,$91,$92,$38,$25,$FB,$DB,$1E,$C9,$E7,$5D
	.dc.b	$A7,$27,$ED,$BA,$5F,$59,$FE,$2E,$F6,$BF,$C5,$ED,$C3,$FC,$0A,$DA
	.dc.b	$FF,$1B,$37,$1F,$F1,$3F,$B1,$FE,$47,$3F,$C9,$1E,$05,$BF,$F2,$84
	.dc.b	$5B,$FF,$18,$4D,$AF,$F2,$3F,$D6,$3A,$FF,$C4,$3E,$D7,$FC,$00,$8C
	.dc.b	$63,$FE,$2B,$8C,$4E,$A3,$F9,$9C,$EB,$0F,$ED,$35,$A9,$E0,$6A,$B8
	.dc.b	$4F,$F9,$07,$D3,$FF,$2F,$1F,$EF,$FF,$AC,$7C,$16,$2F,$FC,$D1,$36
	.dc.b	$BF,$DF,$F5,$A9,$C9,$FE,$2A,$FA,$1F,$F2,$B9,$FE,$58,$F9,$0B,$19
	.dc.b	$25,$7F,$97,$E3,$2F,$F2,$88,$FE,$B9,$F5,$2F,$AB,$F1,$CF,$F8,$2E
	.dc.b	$31,$97,$F8,$25,$7D,$5B,$D5,$C5,$F0,$BF,$EC,$C3,$EA,$DF,$71,$FE
	.dc.b	$11,$8C,$36,$33,$1F,$E1,$56,$BF,$E1,$BF,$BE,$EC,$FF,$86,$5A,$E6
	.dc.b	$FA,$46,$23,$C3,$F5,$6F,$AD,$FE,$F1,$CB,$FE,$86,$8C,$4F,$F6,$C6
	.dc.b	$BF,$0C,$FF,$16,$FD,$6F,$FA,$32,$78,$57,$2D,$FF,$C0,$57,$C2,$73
	.dc.b	$7F,$DD,$AF,$85,$6B,$EB,$78,$4F,$F8,$E2,$D1,$88,$46,$1F,$27,$CB
	.dc.b	$E6,$31,$1E,$53,$73,$93,$53,$CF,$EB,$5F,$51,$C7,$E5,$65,$FF,$82
	.dc.b	$A7,$2D,$9E,$04,$7E,$27,$C7,$D7,$C1,$3C,$1B,$3E,$59,$F5,$7D,$6B
	.dc.b	$A2,$92,$BF,$C3,$30,$E8,$C4,$7A,$E3,$83,$A9,$E7,$3D,$3D,$0E,$BE
	.dc.b	$84,$AE,$71,$97,$AC,$65,$E9,$1F,$37,$EB,$1E,$BF,$9B,$EB,$65,$7F
	.dc.b	$8B,$27,$2D,$9E,$04,$7D,$3F,$5B,$5F,$A5,$3C,$1B,$3F,$0A,$7E,$E7
	.dc.b	$D6,$FC,$7E,$6F,$F6,$AB,$79,$7F,$C9,$4D,$BF,$F0,$CF,$A2,$E4,$FC
	.dc.b	$27,$D6,$BE,$1A,$92,$BF,$C8,$7F,$62,$F6,$37,$E5,$C9,$FF,$23,$DA
	.dc.b	$3F,$1D,$5C,$E7,$FC,$97,$F5,$CF,$8F,$FC,$93,$63,$31,$F9,$8D,$BF
	.dc.b	$D9,$1E,$47,$E5,$FE,$BB,$FD,$63,$5F,$E5,$46,$BF,$95,$6C,$FF,$BD
	.dc.b	$2D,$5F,$9F,$FA,$E7,$D7,$52,$61,$E4,$F1,$FF,$49,$F5,$DA,$8F,$F3
	.dc.b	$0B,$5F,$D1,$37,$1F,$F4,$AF,$7F,$EA,$FE,$31,$F5,$D4,$97,$FE,$6A
	.dc.b	$FB,$1F,$C7,$BF,$AB,$9F,$94,$FF,$9B,$9A,$E7,$8F,$F6,$D2,$F8,$5F
	.dc.b	$DC,$7D,$7B,$D4,$A4,$BF,$F7,$97,$DC,$7E,$F7,$EC,$1F,$13,$49,$3F
	.dc.b	$E7,$CF,$91,$7B,$5B,$F9,$69,$3F,$E8,$0B,$47,$F0,$7A,$FC,$87,$0B
	.dc.b	$C1,$FF,$4A,$7D,$87,$F5,$BC,$3F,$D0,$ED,$7F,$84,$36,$78,$47,$FB
	.dc.b	$2B,$EB,$7F,$24,$BE,$C1,$F1,$7F,$C5,$5B,$1F,$E9,$26,$BF,$C4,$DF
	.dc.b	$4B,$FE,$96,$F9,$67,$D8,$7F,$1D,$6C,$66,$31,$F0,$7F,$C8,$70,$FD
	.dc.b	$04,$FF,$A7,$9A,$FF,$1F,$6E,$3E,$88,$5F,$F0,$B7,$C1,$E7,$F7,$23
	.dc.b	$FE,$A8,$FB,$07,$C9,$FF,$2F,$6C,$66,$30,$78,$F9,$7F,$9A,$7D,$87
	.dc.b	$E0,$B1,$F0,$65,$C2,$FC,$97,$AE,$A9,$1E,$98,$F6,$18,$7F,$9C,$5F
	.dc.b	$3E,$1F,$F6,$53,$63,$FD,$F6,$D5,$7F,$D8,$B0,$F8,$13,$F3,$A7,$E2
	.dc.b	$7A,$FF,$24,$F3,$F3,$48,$F9,$51,$97,$FD,$93,$3C,$6B,$E5,$7D,$83
	.dc.b	$97,$F7,$DF,$57,$F3,$B3,$61,$FE,$AC,$8C,$38,$AE,$0F,$FB,$51,$AF
	.dc.b	$F5,$56,$DF,$6B,$FE,$BD,$1F,$A2,$9F,$F6,$DE,$2F,$AD,$FE,$C4,$FF
	.dc.b	$7C,$BF,$BC,$DC,$7F,$DE,$0E,$EB,$8D,$70,$23,$FE,$E6,$FB,$0F,$D7
	.dc.b	$DF,$AB,$EB,$D7,$FE,$F9,$9C,$D2,$72,$FC,$09,$EC,$E5,$F0,$DF,$FC
	.dc.b	$01,$9D,$CA,$B9,$C6,$1B,$5C,$27,$C4,$D7,$A1,$F3,$A3,$6D,$EC,$FF
	.dc.b	$6D,$BA,$3D,$37,$E1,$AE,$57,$DE,$B6,$DC,$3B,$DF,$AE,$F8,$34,$C9
	.dc.b	$E9,$AE,$37,$F1,$5B,$1B,$3C,$51,$F1,$2F,$B4,$79,$70,$F9,$6F,$F9
	.dc.b	$01,$AF,$3C,$3F,$C8,$6F,$B5,$AF,$C9,$EF,$7D,$86,$2B,$F9,$73,$63
	.dc.b	$33,$CA,$72,$FF,$92,$FE,$9B,$FC,$98,$9A,$47,$47,$CB,$BF,$D0,$69
	.dc.b	$F0,$A3,$F3,$F3,$F1,$27,$31,$F4,$4C,$9C,$76,$FE,$53,$97,$E0,$A3
	.dc.b	$2F,$E1,$27,$8B,$98,$F2,$1C,$6F,$C0,$E2,$38,$57,$4C,$FE,$3E,$1E
	.dc.b	$38,$C5,$7F,$9B,$23,$17,$C9,$B1,$E7,$3A,$57,$F9,$82,$B8,$C3,$0F
	.dc.b	$69,$FE,$5D,$8F,$E8,$6F,$F2,$D8,$F5,$CF,$1B,$FE,$07,$3B,$69,$99
	.dc.b	$FF,$2A,$1C,$0D,$D7,$48,$FE,$17,$0F,$B0,$FE,$01,$1C,$A3,$F8,$7C
	.dc.b	$61,$FF,$2F,$3F,$C4,$A3,$51,$C7,$6E,$8F,$06,$B9,$8D,$CF,$18,$E7
	.dc.b	$3A,$8F,$E8,$73,$AE,$4E,$78,$3E,$9B,$FC,$D3,$39,$26,$1F,$E8,$59
	.dc.b	$D7,$C9,$F3,$BE,$36,$8E,$DF,$51,$30,$F7,$5B,$93,$FC,$11,$1E,$69
	.dc.b	$D6,$3F,$01,$F1,$CE,$A3,$13,$B3,$9C,$7F,$A2,$9C,$47,$96,$30,$FF
	.dc.b	$A1,$FE,$B5,$F3,$9F,$1B,$E3,$FF,$C1,$F7,$F3,$27,$2F,$CE,$8C,$D7
	.dc.b	$FA,$3D,$B8,$FF,$71,$6C,$FD,$CA,$E3,$5B,$E9,$EC,$7D,$4F,$B0,$7C
	.dc.b	$0E,$2F,$ED,$C6,$1F,$BB,$F6,$0F,$9C,$E1,$F3,$7F,$72,$31,$3C,$23
	.dc.b	$2F,$E1,$3E,$C1,$F1,$38,$CF,$FD,$4D,$18,$9D,$4F,$E1,$A3,$F0,$33
	.dc.b	$A8,$C7,$27,$E0,$BC,$7F,$F6,$E5,$FD,$86,$D7,$E0,$5F,$CB,$70,$7E
	.dc.b	$A1,$CB,$7E,$2F,$91,$CF,$E0,$6D,$FE,$0B,$EB,$FF,$DD,$2D,$F9,$3F
	.dc.b	$C8,$C6,$31,$FC,$73,$FE,$F3,$9F,$C4,$FD,$7B,$A3,$F2,$EF,$F9,$B2
	.dc.b	$33,$18,$F7,$7F,$37,$F3,$FF,$E2,$69,$FC,$8B,$FE,$8B,$6F,$9D,$AF
	.dc.b	$85,$DF,$D9,$49,$DF,$C2,$DD,$3F,$41,$F5,$EF,$AB,$FD,$3B,$6B,$F3
	.dc.b	$AB,$EA,$B3,$3A,$ED,$95,$AF,$9C,$B8,$7E,$AF,$E2,$3F,$04,$DF,$91
	.dc.b	$FD,$9C,$63,$9F,$F8,$AA,$8F,$DB,$FD,$6B,$C3,$BF,$1F,$86,$E5,$FB
	.dc.b	$09,$7D,$3F,$0F,$03,$8B,$67,$F6,$BF,$5B,$FE,$94,$6F,$C5,$7F,$04
	.dc.b	$8F,$C5,$6B,$FC,$05,$D3,$FC,$0F,$E8,$DF,$0F,$FC,$29,$FE,$5B,$7F
	.dc.b	$CA,$1C,$BF,$C0,$9C,$CF,$F3,$37,$2F,$3F,$43,$95,$73,$6D,$71,$6C
	.dc.b	$EE,$3F,$86,$FD,$6F,$D3,$8F,$E2,$CD,$E1,$FE,$7A,$FD,$59,$DF,$27
	.dc.b	$F2,$08,$D4,$7F,$3E,$6D,$7F,$47,$78,$3D,$27,$F9,$7F,$D6,$7F,$02
	.dc.b	$8F,$F2,$3B,$6D,$93,$EB,$77,$CD,$32,$FF,$00,$DF,$F0,$EC,$3F,$A2
	.dc.b	$87,$0F,$F4,$86,$BE,$3F,$AD,$F4,$7C,$BF,$A5,$B5,$F0,$9B,$3F,$09
	.dc.b	$B9,$7C,$26,$CF,$BD,$B3,$F4,$FD,$CF,$B4,$7D,$BF,$6E,$D7,$D9,$8B
	.dc.b	$3F,$81,$9B,$7F,$E9,$5F,$AD,$FC,$7D,$FF,$58,$6D,$7D,$4F,$A6,$FE
	.dc.b	$AF,$18,$AE,$1C,$BF,$DC,$6B,$FC,$E3,$3C,$1F,$8A,$D7,$FB,$5B,$5F
	.dc.b	$C5,$7A,$5F,$91,$3F,$25,$D7,$71,$FC,$84,$59,$FC,$C4,$DC,$3F,$C0
	.dc.b	$3C,$3F,$AF,$BF,$CC,$5F,$27,$89,$F8,$EF,$88,$FC,$47,$C1,$FE,$25
	.dc.b	$B1,$93,$F2,$71,$96,$9F,$DE,$BE,$A5,$F5,$BF,$DE,$1B,$1F,$A5,$D7
	.dc.b	$13,$23,$FC,$8F,$3A,$D3,$FC,$96,$D7,$F1,$2D,$7F,$35,$EA,$FE,$74
	.dc.b	$FC,$F7,$93,$6B,$F4,$F1,$67,$F4,$F3,$75,$FD,$3C,$DA,$FD,$24,$DA
	.dc.b	$FD,$C7,$77,$EF,$FE,$AF,$F8,$03,$6B,$F7,$CD,$9F,$E0,$CD,$C3,$FD
	.dc.b	$B5,$8F,$F8,$1E,$BE,$27,$B7,$FB,$3F,$27,$E2,$32,$32,$7E,$FA,$31
	.dc.b	$87,$F1,$1F,$AB,$FE,$21,$EA,$7F,$11,$6D,$7F,$0F,$6C,$EE,$3F,$89
	.dc.b	$4E,$DA,$6C,$B1,$B3,$F8,$F4,$6D,$B1,$B6,$9B,$C7,$F1,$73,$B6,$CE
	.dc.b	$DB,$3B,$8F,$E4,$D3,$B6,$9B,$2C,$6C,$FE,$51,$1B,$6C,$6D,$A6,$E7
	.dc.b	$F8,$9C,$ED,$B3,$B6,$CE,$F8,$D5,$B6,$D6,$DB,$3B,$9D,$12,$74,$D8
	.dc.b	$D4,$60,$8E,$09,$5F,$B2,$BF,$AE,$D9,$FA,$6D,$9F,$9E,$D3,$E5,$96
	.dc.b	$3E,$3B,$63,$E1,$B4,$F7,$96,$FD,$6D,$BF,$43,$6B,$CE,$DA,$F3,$C7
	.dc.b	$F5,$DB,$F3,$B4,$F3,$96,$3C,$ED,$9F,$3B,$4F,$39,$63,$CE,$D8,$F3
	.dc.b	$B6,$FC,$ED,$BF,$3B,$4F,$39,$6B,$CE,$7F,$6F,$BF,$3B,$63,$CE,$D3
	.dc.b	$CE,$5B,$F3,$B6,$3C,$ED,$3C,$E5,$8F,$3B,$6F,$CE,$DB,$F3,$B6,$BC
	.dc.b	$ED,$3C,$F5,$FD,$AA,$FC,$ED,$9F,$3B,$67,$CE,$D3,$CE,$58,$F3,$B6
	.dc.b	$3C,$ED,$3C,$E5,$BF,$3B,$6F,$CE,$DA,$F3,$B6,$BC,$F4,$BF,$3B,$6B
	.dc.b	$CE,$D9,$F3,$B6,$BC,$ED,$9F,$3B,$67,$CE,$D9,$F3,$B4,$F3,$96,$FC
	.dc.b	$FD,$EF,$2F,$FC,$3C,$7F,$87,$FB,$AE,$3F,$E2,$88,$B8,$FF,$8A,$26
	.dc.b	$E2,$AF,$F6,$CF,$D5,$BA,$BF,$D9,$ED,$8F,$F1,$E7,$89,$89,$91,$FE
	.dc.b	$D8,$9D,$61,$FE,$47,$AC,$BF,$E3,$6F,$2B,$FC,$64,$7F,$B4,$BC,$FF
	.dc.b	$F2,$E5,$FF,$8F,$23,$2D,$3F,$D9,$71,$A9,$E5,$19,$69,$FE,$D7,$8D
	.dc.b	$63,$FE,$42,$AF,$F2,$B7,$D0,$CB,$FF,$2A,$7C,$65,$C3,$FC,$B9,$36
	.dc.b	$FF,$C0,$B3,$6B,$FD,$F7,$F6,$5E,$1B,$FF,$1C,$B5,$CF,$99,$FE,$1B
	.dc.b	$51,$94,$3E,$B1,$EB,$FF,$9E,$E3,$FC,$86,$65,$C3,$FE,$46,$87,$D6
	.dc.b	$3E,$AF,$73,$BC,$2F,$D4,$3E,$AB,$BC,$B5,$FB,$8F,$8A,$74,$F9,$23
	.dc.b	$2A,$32,$F1,$D3,$9F,$D5,$6A,$B7,$5A,$6D,$F8,$1B,$7D,$6F,$89,$93
	.dc.b	$CB,$5C,$5B,$59,$9F,$33,$55,$F8,$67,$8F,$BD,$9B,$34,$C7,$BD,$6F
	.dc.b	$C4,$7C,$18,$C3,$ED,$3E,$24,$77,$FA,$A7,$CF,$75,$F8,$C9,$5C,$84
	.dc.b	$E9,$C3,$5F,$B6,$D7,$FC,$FE,$D7,$EA,$B5,$FC,$2F,$D5,$39,$7E,$03
	.dc.b	$63,$C0,$D7,$D0,$17,$2F,$09,$93,$C1,$E7,$18,$8E,$33,$CF,$7F,$C0
	.dc.b	$B7,$5F,$C4,$BD,$27,$C3,$E6,$79,$3C,$D9,$18,$21,$D1,$CB,$DA,$3E
	.dc.b	$D7,$D4,$BD,$3F,$6F,$D5,$4A,$F9,$93,$96,$CF,$E7,$94,$65,$FD,$3B
	.dc.b	$5F,$F5,$5B,$5F,$D0,$FE,$36,$F3,$7F,$14,$BC,$DA,$62,$24,$FD,$D9
	.dc.b	$D3,$67,$2F,$D0,$8E,$8E,$48,$66,$7F,$35,$AB,$C9,$7D,$AE,$17,$D1
	.dc.b	$D3,$11,$9E,$DD,$77,$CC,$7F,$33,$87,$39,$EB,$F6,$FF,$E8,$C9,$CB
	.dc.b	$5F,$CE,$4A,$30,$FF,$02,$31,$CF,$E3,$F6,$E5,$76,$CD,$1E,$7F,$82
	.dc.b	$F4,$7B,$52,$B4,$FA,$F8,$BC,$7F,$75,$E4,$FA,$79,$5E,$48,$F8,$1E
	.dc.b	$CF,$C1,$F6,$64,$F7,$8F,$89,$C7,$F1,$79,$64,$F8,$E3,$DF,$E1,$48
	.dc.b	$F7,$BF,$0B,$E2,$FE,$67,$9F,$27,$D4,$FC,$FF,$2F,$E9,$FD,$7F,$F2
	.dc.b	$CE,$4F,$84,$DF,$03,$F9,$13,$6F,$F9,$42,$F8,$5F,$9F,$F6,$5F,$43
	.dc.b	$E7,$E5,$7D,$08,$FE,$2D,$F4,$7F,$D9,$E3,$F8,$33,$5F,$DB,$B7,$1F
	.dc.b	$E5,$FF,$41,$FC,$1D,$FE,$63,$F4,$6E,$2F,$E4,$0D,$BF,$E2,$0D,$7F
	.dc.b	$80,$AD,$BF,$93,$7C,$73,$E8,$E9,$2B,$F8,$C4,$7F,$9F,$9F,$E3,$58
	.dc.b	$BC,$DF,$E5,$CD,$8C,$C7,$4F,$2A,$47,$26,$BF,$C4,$BE,$8D,$C9,$FC
	.dc.b	$DD,$B1,$FC,$9D,$AE,$E3,$BF,$07,$F3,$DF,$A3,$74,$F8,$A6,$57,$F2
	.dc.b	$F9,$CB,$67,$2F,$F2,$2C,$BA,$F0,$7F,$5A,$AD,$CF,$F2,$66,$C6,$6B
	.dc.b	$C0,$7C,$0E,$BF,$16,$FF,$0B,$C1,$98,$78,$63,$2F,$C8,$8C,$BD,$63
	.dc.b	$FA,$B7,$D1,$BE,$33,$F8,$0F,$CE,$7C,$BA,$FD,$9B,$72,$FE,$DF,$5B
	.dc.b	$9F,$E9,$8D,$7F,$9E,$B6,$37,$5D,$BE,$2F,$FB,$26,$2F,$9F,$BE,$CC
	.dc.b	$C7,$FA,$B4,$E5,$B3,$C0,$8F,$F5,$BF,$A2,$5F,$F8,$1E,$1F,$44,$BF
	.dc.b	$F0,$5D,$6B,$2F,$F0,$DC,$E6,$BF,$C3,$B5,$DD,$E2,$FF,$84,$8C,$46
	.dc.b	$E1,$5C,$DD,$9F,$E2,$67,$11,$B9,$E2,$73,$C3,$FC,$3B,$1C,$E7,$73
	.dc.b	$E6,$C3,$D5,$3C,$6B,$73,$C6,$FD,$2E,$F0,$B1,$E1,$BE,$33,$B8,$F2
	.dc.b	$9F,$0E,$B9,$BB,$AE,$2F,$B2,$B8,$BF,$12,$76,$F1,$37,$87,$A1,$E4
	.dc.b	$7F,$90,$9F,$88,$FF,$8E,$8E,$95,$FE,$40,$8C,$C7,$D4,$F8,$E7,$D1
	.dc.b	$E5,$B1,$CD,$AF,$39,$46,$23,$FC,$A3,$F4,$7F,$DB,$D3,$48,$FF,$73
	.dc.b	$7D,$F7,$F9,$FF,$0B,$E0,$CF,$F9,$7A,$73,$B7,$F9,$43,$5F,$C0,$4E
	.dc.b	$4F,$97,$7C,$5F,$9A,$66,$3F,$0D,$19,$78,$C7,$90,$E3,$3D,$75,$F4
	.dc.b	$C6,$30,$F1,$BB,$AE,$75,$B8,$F9,$98,$79,$AB,$6D,$8F,$19,$C6,$FF
	.dc.b	$CB,$95,$C6,$3F,$CD,$8F,$F6,$BA,$C9,$FE,$45,$8F,$54,$E1,$FF,$1B
	.dc.b	$E1,$EF,$3F,$CD,$55,$C6,$3C,$67,$A4,$FF,$1E,$FD,$1F,$EF,$6F,$FC
	.dc.b	$A0,$D8,$C9,$C4,$39,$AF,$F3,$DB,$67,$FD,$06,$D9,$FF,$49,$BF,$E9
	.dc.b	$48,$EB,$AF,$38,$FE,$77,$1E,$0E,$7F,$35,$BE,$8F,$FC,$CB,$39,$26
	.dc.b	$1F,$E1,$08,$7C,$67,$3B,$E3,$18,$AF,$5F,$CD,$F7,$3A,$F2,$BA,$F1
	.dc.b	$3A,$26,$3F,$CA,$A3,$2F,$C6,$AE,$3E,$FE,$33,$FF,$0F,$61,$FE,$84
	.dc.b	$C3,$FC,$F1,$F4,$4F,$A3,$F9,$1F,$09,$FD,$A6,$73,$18,$EB,$91,$FE
	.dc.b	$7F,$9D,$67,$FE,$70,$BE,$53,$9A,$DC,$FF,$9D,$3E,$8D,$EF,$63,$3F
	.dc.b	$F5,$6C,$62,$FF,$D1,$5F,$47,$FE,$55,$7F,$D1,$90,$F9,$3F,$B9,$39
	.dc.b	$C3,$A4,$7D,$F9,$D5,$7F,$A4,$FE,$8D,$F4,$79,$AF,$C2,$46,$27,$53
	.dc.b	$8E,$4C,$B6,$3F,$C9,$BF,$25,$FE,$CD,$6F,$83,$F8,$08,$FC,$46,$87
	.dc.b	$59,$FF,$93,$9F,$9C,$DE,$3E,$1E,$4F,$17,$F1,$3F,$47,$FE,$DA,$C7
	.dc.b	$FC,$B6,$FF,$B8,$9B,$EF,$6B,$DF,$FC,$7F,$D1,$BA,$6F,$C7,$E1,$B7
	.dc.b	$FE,$64,$F6,$9E,$87,$E5,$97,$D1,$D8,$C3,$9F,$81,$95,$C7,$E5,$2F
	.dc.b	$07,$F9,$A3,$E8,$FF,$CC,$33,$FE,$70,$6C,$67,$E0,$BF,$47,$18,$E4
	.dc.b	$FF,$36,$3F,$86,$6F,$6E,$BD,$DD,$BF,$AA,$FA,$3F,$C1,$61,$FE,$76
	.dc.b	$5F,$17,$9F,$03,$4F,$EC,$3E,$8D,$E1,$DF,$7F,$0D,$DB,$FD,$0B,$2F
	.dc.b	$9E,$E1,$E1,$71,$6C,$FF,$A2,$BE,$59,$F4,$7F,$BC,$6C,$67,$C2,$C4
	.dc.b	$FF,$00,$F9,$B7,$D1,$DF,$47,$0D,$BF,$F4,$AC,$E9,$B3,$A6,$9A,$9E
	.dc.b	$1C,$3F,$E9,$6A,$E0,$DA,$FE,$1F,$F3,$9F,$36,$D2,$E5,$4E,$63,$0B
	.dc.b	$9F,$8F,$C6,$4B,$FF,$54,$CE,$EF,$8D,$66,$FF,$DD,$7F,$45,$F6,$6F
	.dc.b	$BC,$23,$0F,$87,$27,$D1,$FF,$31,$9C,$CE,$FA,$3F,$93,$FD,$7F,$F2
	.dc.b	$A8,$EB,$18,$F2,$FD,$BF,$1F,$EE,$F8,$FB,$9F,$99,$B7,$FE,$BA,$DB
	.dc.b	$FD,$D1,$3E,$AA,$E0,$D3,$81,$67,$85,$7D,$9F,$7E,$4F,$D6,$F7,$6D
	.dc.b	$7D,$19,$BC,$1F,$EC,$5D,$FF,$C0,$B7,$F7,$27,$FA,$77,$D2,$7E,$36
	.dc.b	$3E,$9D,$1F,$8F,$19,$BF,$47,$D0,$70,$E9,$95,$F4,$BD,$6B,$5E,$D7
	.dc.b	$F1,$DE,$04,$9F,$C5,$F8,$16,$BF,$0D,$37,$2F,$34,$DA,$F1,$E9,$FA
	.dc.b	$C9,$F5,$57,$E6,$6B,$57,$F8,$6F,$8B,$FC,$6F,$C2,$FE,$39,$B8,$FE
	.dc.b	$19,$B3,$F4,$5B,$87,$BE,$3F,$27,$F4,$4F,$A4,$C4,$7F,$8D,$E3,$17
	.dc.b	$FE,$39,$8D,$CE,$9F,$C6,$CE,$9D,$36,$BF,$DE,$DC,$EF,$8F,$FE,$FD
	.dc.b	$39,$C3,$E8,$CE,$5F,$CE,$FC,$97,$EA,$0F,$D4,$F9,$D6,$32,$4C,$7F
	.dc.b	$19,$19,$7F,$1D,$37,$1F,$71,$FA,$E9,$C5,$7C,$EF,$13,$F8,$27,$D2
	.dc.b	$7F,$05,$6C,$7F,$02,$6E,$1F,$BD,$6D,$FC,$6C,$BF,$C0,$53,$FB,$BA
	.dc.b	$FC,$5F,$D9,$30,$F7,$FC,$17,$F8,$42,$31,$3F,$90,$3E,$4C,$66,$77
	.dc.b	$19,$AF,$E0,$D5,$C1,$FE,$1D,$F4,$BF,$C6,$FD,$BF,$E3,$8D,$AF,$E3
	.dc.b	$2D,$9D,$C7,$F1,$C9,$DB,$4D,$96,$36,$7F,$1B,$8D,$B6,$36,$D3,$79
	.dc.b	$7E,$BE,$76,$D9,$DB,$67,$71,$F9,$E9,$DB,$4D,$96,$36,$7F,$17,$8D
	.dc.b	$B6,$36,$D3,$72,$E5,$FE,$FD,$3A,$8C,$57,$26,$C6,$67,$75,$C9,$B3
	.dc.b	$FC,$2E,$FE,$63,$6B,$E4,$36,$BE,$23,$4F,$80,$58,$F6,$36,$3D,$0D
	.dc.b	$3C,$6F,$86,$FC,$D7,$E2,$AF,$E0,$B7,$E2,$69,$E2,$2D,$78,$9B,$7E
	.dc.b	$26,$9E,$22,$C7,$89,$B1,$E2,$3C,$95,$E1,$9F,$E9,$15,$E1,$6D,$78
	.dc.b	$5A,$78,$4B,$5E,$16,$C7,$85,$A7,$84,$B1,$E1,$8F,$1D,$78,$63,$EE
	.dc.b	$57,$85,$B3,$E1,$6D,$78,$5A,$78,$4B,$1E,$16,$C7,$85,$A7,$85,$F1
	.dc.b	$D7,$84,$57,$85,$B5,$E1,$6C,$F8,$5B,$5E,$16,$CF,$85,$B3,$E1,$6C
	.dc.b	$F8,$4F,$1B,$E1,$AF,$01,$2B,$C0,$DA,$F0,$36,$BC,$0D,$9F,$03,$67
	.dc.b	$C0,$D9,$F0,$57,$88,$F0,$57,$62,$57,$66,$CF,$66,$9D,$8B,$1D,$9B
	.dc.b	$1D,$AB,$C1,$5D,$AB,$A1,$27,$A3,$6B,$A3,$4E,$85,$8F,$91,$DC,$F4
	.dc.b	$3F,$C6,$27,$F8,$CB,$C1,$B7,$FE,$2A,$8B,$7F,$E1,$59,$B7,$FD,$8B
	.dc.b	$E9,$5D,$9F,$D7,$EB,$FC,$73,$18,$C3,$FB,$24,$23,$0F,$F6,$76,$BF
	.dc.b	$E3,$23,$53,$FE,$2C,$F8,$3F,$F7,$04,$F0,$BF,$EC,$F1,$FE,$48,$F4
	.dc.b	$66,$1F,$D0,$E5,$EB,$FF,$9A,$A3,$FC,$87,$3A,$C7,$FC,$5C,$FF,$B7
	.dc.b	$A7,$51,$89,$D9,$1F,$EB,$07,$F9,$4E,$37,$5F,$DA,$34,$5F,$F9,$A7
	.dc.b	$E7,$E6,$1F,$E6,$6F,$92,$B8,$FF,$98,$66,$E1,$FE,$31,$9B,$5F,$E5
	.dc.b	$2F,$A4,$7B,$9F,$E8,$2C,$3F,$C1,$91,$8F,$53,$E5,$AF,$75,$F0,$38
	.dc.b	$BE,$17,$F2,$FE,$C9,$F2,$3F,$4A,$11,$87,$EA,$B5,$FF,$0D,$35,$FA
	.dc.b	$9F,$32,$FA,$5F,$F3,$3E,$DF,$E2,$28,$C4,$FF,$73,$0E,$1F,$92,$3E
	.dc.b	$97,$FC,$53,$3C,$6B,$9E,$DF,$E7,$2B,$E1,$39,$BE,$97,$C2,$B5,$7C
	.dc.b	$27,$FD,$0F,$7C,$27,$31,$C1,$D3,$E4,$F8,$5B,$26,$98,$F8,$6B,$3E
	.dc.b	$B9,$D3,$67,$4F,$BF,$E9,$1F,$4B,$D3,$E4,$64,$E6,$3C,$33,$A8,$C1
	.dc.b	$0F,$8B,$7F,$EF,$C7,$E5,$7D,$2B,$9B,$D2,$D8,$CC,$63,$9B,$2C,$8C
	.dc.b	$9C,$A3,$0F,$D2,$9E,$4D,$9E,$4F,$83,$CF,$EF,$E7,$C8,$C1,$0E,$4E
	.dc.b	$5E,$91,$F3,$BE,$91,$EB,$7A,$3E,$96,$57,$E1,$67,$2D,$9E,$04,$7E
	.dc.b	$0F,$C2,$7D,$8F,$A5,$79,$F7,$CF,$95,$F8,$F9,$CB,$67,$2F,$E4,$21
	.dc.b	$3C,$E6,$4F,$18,$DF,$47,$F9,$F2,$B9,$CF,$F0,$08,$7C,$43,$E9,$3F
	.dc.b	$21,$18,$FA,$5B,$26,$CE,$3C,$0C,$F7,$CC,$3F,$8B,$5F,$EE,$5F,$E3
	.dc.b	$1F,$4B,$FE,$8D,$6F,$AD,$DA,$31,$1E,$35,$D5,$E0,$F8,$39,$3D,$1F
	.dc.b	$1F,$C1,$F9,$3E,$02,$4F,$81,$F3,$F1,$F5,$FA,$5E,$DE,$2C,$AF,$1C
	.dc.b	$7B,$3D,$BF,$6F,$B9,$27,$D3,$1F,$03,$9B,$E0,$F3,$49,$F6,$C7,$A7
	.dc.b	$BE,$47,$B9,$F6,$7C,$17,$C8,$F4,$A4,$FC,$47,$E5,$F9,$BF,$3B,$E9
	.dc.b	$7F,$8D,$F5,$7F,$A5,$9B,$E4,$7F,$B5,$1B,$69,$43,$B7,$E5,$7B,$DF
	.dc.b	$2F,$D2,$97,$F4,$23,$F8,$2F,$D2,$FF,$5A,$8F,$E3,$4D,$7F,$8C,$B7
	.dc.b	$0F,$F6,$0F,$D1,$7E,$EE,$7F,$8E,$FD,$2B,$8B,$F8,$7B,$63,$F9,$23
	.dc.b	$5D,$ED,$FC,$53,$E4,$9F,$4B,$49,$3F,$CA,$E1,$EE,$FF,$1A,$C5,$E8
	.dc.b	$52,$56,$FC,$DE,$84,$70,$7F,$35,$FA,$57,$2F,$F2,$46,$C6,$63,$07
	.dc.b	$7E,$1F,$E5,$3F,$4A,$EA,$F1,$4C,$AF,$E8,$13,$96,$CE,$5F,$E8,$39
	.dc.b	$76,$7F,$A1,$CE,$DB,$3D,$88,$FF,$4B,$9F,$EA,$9C,$3F,$EF,$0A,$6F
	.dc.b	$F0,$FC,$19,$7E,$28,$F3,$98,$7C,$31,$87,$FA,$B7,$D2,$BE,$3B,$F7
	.dc.b	$7F,$3D,$F3,$27,$FC,$14,$D7,$FA,$CC,$ED,$E3,$3F,$D3,$23,$11,$A8
	.dc.b	$C7,$C6,$7F,$50,$B7,$D0,$DF,$6E,$5F,$F6,$79,$CB,$67,$81,$1F,$ED
	.dc.b	$9F,$48,$BF,$CF,$43,$E9,$3F,$81,$A1,$3B,$DB,$13,$A9,$FE,$13,$7B
	.dc.b	$79,$61,$4F,$1B,$83,$A5,$F3,$70,$76,$C3,$C7,$7D,$6B,$8B,$6F,$B3
	.dc.b	$E1,$BE,$4D,$AE,$35,$E1,$AD,$B6,$FD,$2F,$9E,$F8,$B6,$B8,$CF,$C1
	.dc.b	$8F,$09,$C5,$F3,$CF,$07,$D7,$3C,$1F,$86,$76,$39,$57,$89,$DE,$19
	.dc.b	$7E,$03,$FE,$5C,$AF,$89,$1F,$E5,$96,$BE,$58,$D7,$C8,$3E,$96,$92
	.dc.b	$7F,$CC,$07,$D8,$8D,$3F,$E3,$7F,$A5,$FF,$1C,$26,$91,$D1,$F5,$2B
	.dc.b	$F3,$9E,$17,$9E,$3F,$09,$3E,$99,$CC,$7D,$A3,$27,$1D,$95,$C6,$3C
	.dc.b	$06,$5C,$57,$39,$DC,$F3,$37,$AF,$E3,$EF,$BB,$F9,$18,$EE,$E6,$78
	.dc.b	$B9,$8E,$3A,$FF,$93,$A7,$2F,$C6,$BF,$06,$1E,$2A,$F2,$13,$0F,$F2
	.dc.b	$6D,$F9,$1F,$F4,$A5,$70,$9F,$F2,$9C,$7C,$23,$75,$FE,$59,$8F,$2D
	.dc.b	$F8,$63,$FC,$B9,$F4,$BF,$BA,$6E,$BD,$5F,$D5,$47,$EC,$5C,$92,$BF
	.dc.b	$CC,$31,$C5,$A7,$13,$F7,$0E,$6F,$E5,$C2,$3F,$9B,$CF,$8B,$9B,$85
	.dc.b	$3E,$97,$FC,$A3,$39,$25,$FF,$86,$61,$F1,$DD,$E3,$A5,$F2,$9C,$D7
	.dc.b	$BB,$E7,$F5,$89,$EF,$CA,$EA,$CC,$4A,$E0,$7F,$28,$8C,$63,$F0,$27
	.dc.b	$8F,$C0,$E3,$4F,$A9,$87,$F9,$C7,$0F,$F4,$47,$D2,$BE,$53,$CD,$F1
	.dc.b	$1F,$CF,$A7,$31,$8E,$B9,$1F,$E8,$F9,$D6,$7F,$E8,$EB,$E5,$39,$AD
	.dc.b	$CF,$FA,$4B,$E9,$1F,$01,$8C,$FE,$7C,$62,$FF,$D3,$5F,$4B,$FE,$50
	.dc.b	$7F,$D3,$90,$F9,$5F,$95,$39,$C3,$A4,$7C,$D9,$D5,$7F,$A8,$FE,$95
	.dc.b	$F3,$D9,$AF,$A5,$18,$9D,$72,$E0,$CD,$58,$FF,$6E,$79,$0F,$9E,$FC
	.dc.b	$0B,$C7,$2F,$F2,$B3,$99,$DD,$6B,$93,$7F,$21,$67,$9B,$F8,$AF,$A5
	.dc.b	$79,$94,$C7,$D2,$5A,$FF,$76,$7D,$2B,$E9,$3F,$22,$D8,$CF,$C7,$E2
	.dc.b	$7F,$2B,$ED,$BE,$06,$98,$F8,$1B,$7F,$EF,$79,$FB,$3C,$FF,$B7,$7F
	.dc.b	$DD,$71,$9E,$9E,$7E,$DF,$E8,$3E,$95,$F0,$57,$E0,$B0,$DB,$FE,$C5
	.dc.b	$8F,$F0,$1D,$3F,$DE,$F1,$87,$FC,$88,$E8,$C7,$7F,$EB,$23,$3C,$FF
	.dc.b	$E8,$15,$F3,$F9,$35,$D9,$6F,$95,$7F,$94,$7E,$95,$F4,$5F,$B6,$6C
	.dc.b	$EB,$24,$3E,$3F,$55,$BF,$9F,$97,$BA,$7C,$F6,$31,$FF,$4A,$7D,$2B
	.dc.b	$E2,$FF,$81,$BF,$E8,$4B,$EC,$E5,$BE,$87,$F0,$8F,$9B,$FA,$91,$E3
	.dc.b	$6D,$FF,$A7,$D7,$46,$FA,$65,$6D,$E5,$18,$AE,$11,$77,$FE,$61,$7C
	.dc.b	$AF,$81,$F1,$AB,$E2,$F4,$7F,$13,$DF,$F0,$33,$F6,$3E,$7D,$F4,$99
	.dc.b	$A9,$3F,$27,$E8,$FF,$91,$E1,$FE,$7F,$8F,$F0,$D6,$1C,$A3,$FD,$05
	.dc.b	$3A,$CB,$C7,$D1,$F0,$7A,$7E,$13,$70,$F8,$4D,$BF,$84,$DA,$FE,$95
	.dc.b	$9F,$F8,$C6,$BE,$D3,$6B,$FD,$8C,$BD,$8F,$E6,$AE,$6B,$8F,$B7,$83
	.dc.b	$33,$6F,$9C,$5B,$E7,$36,$BF,$A1,$FD,$27,$E4,$AF,$E5,$36,$BE,$C7
	.dc.b	$D0,$FF,$3D,$8C,$57,$0E,$BF,$C0,$7A,$39,$6E,$1E,$AF,$43,$CE,$7A
	.dc.b	$3E,$66,$C6,$49,$97,$BA,$32,$FE,$1F,$E2,$7F,$19,$F0,$9F,$8D,$6D
	.dc.b	$7E,$35,$B3,$B8,$FC,$84,$ED,$A6,$CB,$1F,$74,$FF,$2E,$7D,$27,$E6
	.dc.b	$DC,$07,$19,$FD,$6C,$BF,$AF,$46,$2B,$FC,$BC,$F1,$AF,$F3,$01,$C7
	.dc.b	$B3,$FB,$7B,$A6,$BE,$06,$DB,$A3,$FA,$03,$F9,$EF,$90,$FD,$67,$8F
	.dc.b	$FA,$D6,$63,$FA,$D7,$2D,$3F,$13,$E1,$FE,$1B,$C2,$FC,$3B,$63,$27
	.dc.b	$E3,$A3,$2D,$33,$97,$E4,$23,$2D,$8F,$D7,$67,$FD,$F2,$3E,$3B,$71
	.dc.b	$FD,$AF,$CA,$7C,$6F,$76,$5F,$F9,$4A,$FE,$64,$7C,$26,$C6,$6F,$A4
	.dc.b	$7F,$A2,$6D,$D1,$FC,$5A,$BF,$4B,$E9,$7E,$8B,$C4,$FD,$1B,$63,$27
	.dc.b	$EA,$A3,$2F,$F1,$3F,$A3,$FE,$2C,$7F,$17,$F0,$2C,$64,$98,$FF,$17
	.dc.b	$8C,$BF,$C6,$66,$E3,$FC,$4A,$6D,$7F,$44,$E1,$56,$2D,$7F,$76,$F0
	.dc.b	$FD,$F9,$B5,$FC,$36,$31,$7E,$27,$F8,$FC,$F9,$E3,$07,$77,$8D,$72
	.dc.b	$6C,$6A,$FC,$37,$E4,$8C,$5F,$8D,$B1,$99,$F7,$47,$F3,$9B,$E8,$E9
	.dc.b	$B1,$A3,$E3,$47,$2B,$F0,$C6,$6F,$D3,$3C,$4D,$16,$34,$FA,$CC,$47
	.dc.b	$08,$C9,$EA,$9F,$84,$E2,$3E,$3B,$6F,$E5,$1E,$B8,$E4,$4A,$FA,$D7
	.dc.b	$EA,$8C,$57,$C6,$8C,$4F,$8E,$3D,$A7,$5B,$F6,$D7,$03,$C2,$5B,$F0
	.dc.b	$C7,$29,$E9,$5F,$01,$F3,$5E,$6F,$8C,$F9,$0D,$3C,$E3,$57,$EB,$9D
	.dc.b	$57,$8D,$F8,$E7,$9E,$BC,$57,$F1,$9B,$1A,$9F,$94,$F6,$39,$C6,$A3
	.dc.b	$CB,$19,$76,$7C,$28,$D1,$E5,$8D,$3E,$A8,$F7,$9E,$53,$99,$5E,$0D
	.dc.b	$8F,$41,$E2,$2B,$E0,$8D,$9E,$08,$CB,$F9,$98,$E4,$6E,$39,$34,$91
	.dc.b	$E5,$38,$1C,$DC,$D7,$D8,$9F,$5B,$63,$AC,$F0,$9F,$0C,$FD,$39,$E3
	.dc.b	$7F,$5A,$FE,$85,$69,$FC,$3B,$4F,$A5,$7F,$1E,$9F,$46,$EB,$FF,$55
	.dc.b	$36,$BF,$C0,$7E,$36,$26,$47,$F8,$7E,$75,$5F,$E0,$6A,$CB,$FE,$2F
	.dc.b	$F3,$3F,$C6,$87,$FA,$F3,$D2,$FF,$2B,$C7,$F0,$48,$CD,$FF,$8D,$DD
	.dc.b	$79,$98,$9F,$F0,$8F,$CE,$FF,$84,$0F,$F0,$8F,$C6,$5B,$FF,$0B,$C5
	.dc.b	$9F,$F0,$BC,$DD,$3F,$C3,$B3,$6B,$FC,$65,$36,$83,$E9,$98,$7F,$45
	.dc.b	$6D,$7F,$49,$6C,$FF,$6E,$38,$4F,$F4,$BF,$A4,$7C,$D7,$29,$D4,$62
	.dc.b	$7F,$A9,$2F,$9B,$7C,$66,$2F,$85,$FF,$7A,$F9,$87,$D1,$7F,$53,$CF
	.dc.b	$CF,$29,$CB,$FE,$51,$6B,$FE,$0A,$FA,$3F,$F7,$7E,$FF,$EF,$09,$CB
	.dc.b	$6F,$A5,$7C,$29,$CD,$F6,$BE,$2F,$82,$1F,$0D,$C2,$BC,$4D,$8C,$C7
	.dc.b	$8F,$E5,$31,$32,$3C,$CE,$E3,$11,$FE,$20,$9E,$2D,$9D,$3F,$E6,$DF
	.dc.b	$A3,$7B,$7D,$FD,$79,$39,$8F,$24,$EA,$30,$43,$B7,$C0,$FA,$7E,$8D
	.dc.b	$CD,$49,$3F,$E7,$B8,$73,$62,$3F,$C6,$B1,$C1,$D4,$F3,$9E,$9E,$77
	.dc.b	$5F,$3A,$57,$38,$CB,$D6,$32,$F4,$8C,$BD,$A3,$3C,$4E,$0F,$C0,$E3
	.dc.b	$FE,$FE,$AF,$F4,$87,$D1,$BD,$5F,$67,$B9,$27,$31,$F5,$E7,$51,$82
	.dc.b	$1E,$CF,$9C,$FB,$9F,$46,$F3,$AF,$9F,$2B,$FD,$3D,$39,$6C,$E5,$FF
	.dc.b	$2C,$C2,$79,$CF,$4F,$72,$57,$27,$BF,$BB,$88,$E3,$17,$6F,$81,$1F
	.dc.b	$B6,$C3,$FD,$72,$F1,$21,$C0,$F2,$D7,$E7,$25,$F3,$CF,$9B,$C3,$6B
	.dc.b	$FD,$B0,$D7,$FC,$E9,$3C,$3E,$9B,$7F,$2B,$27,$6F,$2F,$81,$ED,$F4
	.dc.b	$5F,$98,$BF,$38,$5D,$FA,$37,$DD,$E2,$E8,$F1,$F4,$4C,$BC,$71,$E6
	.dc.b	$F6,$7C,$FE,$C4,$9F,$3C,$78,$3E,$2A,$47,$FB,$C7,$24,$79,$3C,$CF
	.dc.b	$7F,$A9,$2B,$DE,$FC,$2F,$3B,$E1,$F8,$12,$7E,$1B,$F1,$BC,$0F,$6F
	.dc.b	$C2,$FB,$BC,$49,$5D,$E3,$E7,$7D,$D7,$F1,$A8,$FE,$57,$C7,$F0,$A1
	.dc.b	$F2,$DA,$37,$E6,$70,$F8,$DE,$3E,$3C,$C3,$8F,$8F,$CB,$05,$7C,$EF
	.dc.b	$6E,$57,$BD,$FA,$3F,$55,$FC,$4F,$E7,$BF,$80,$4F,$FA,$AB,$E8,$7F
	.dc.b	$AE,$3F,$EF,$56,$BF,$EA,$E8,$DE,$BF,$8E,$87,$D0,$BD,$BF,$E3,$CB
	.dc.b	$E1,$29,$88,$FE,$45,$F2,$1F,$89,$BE,$4D,$DB,$F9,$BB,$70,$FE,$9E
	.dc.b	$B9,$FD,$1E,$04,$9F,$82,$7F,$54,$CF,$F1,$2E,$05,$7F,$14,$70,$D7
	.dc.b	$FD,$92,$DC,$3C,$7C,$DF,$CD,$FE,$85,$F2,$74,$97,$D6,$32,$F7,$C1
	.dc.b	$F3,$3F,$CF,$FE,$85,$F1,$3F,$3F,$09,$7E,$59,$CB,$67,$2F,$CF,$9E
	.dc.b	$51,$23,$44,$3E,$A7,$D2,$77,$E0,$FB,$9E,$BC,$9E,$D1,$97,$CD,$19
	.dc.b	$7F,$DD,$38,$7D,$97,$E4,$7D,$13,$CA,$FE,$09,$F3,$9F,$32,$7F,$7E
	.dc.b	$DE,$0F,$C1,$56,$E7,$8D,$7F,$5F,$8C,$47,$FB,$D9,$B3,$C3,$E4,$3F
	.dc.b	$B2,$63,$FA,$0A,$FB,$0F,$E9,$5F,$F7,$EB,$67,$F2,$0E,$E3,$3D,$FC
	.dc.b	$88,$FE,$07,$E8,$97,$FE,$BC,$87,$D0,$AF,$FC,$0B,$5A,$C3,$FC,$3B
	.dc.b	$39,$AF,$F0,$FD,$77,$8C,$47,$1C,$3F,$C1,$B1,$8B,$E5,$3D,$AB,$AE
	.dc.b	$1E,$1A,$E3,$1C,$CF,$F0,$A9,$67,$FC,$5B,$5C,$A7,$A5,$71,$AD,$DF
	.dc.b	$81,$B7,$E8,$7B,$E1,$63,$9D,$71,$7D,$75,$C5,$A6,$C6,$1E,$1B,$E7
	.dc.b	$1E,$B3,$65,$3E,$11,$C6,$38,$3E,$0B,$FF,$20,$3D,$2F,$E9,$3F,$E3
	.dc.b	$43,$11,$A9,$F9,$31,$95,$F2,$4F,$A2,$D3,$67,$FC,$99,$3A,$9C,$9E
	.dc.b	$08,$FF,$2B,$FD,$15,$4D,$23,$B7,$E1,$5F,$E8,$7A,$FE,$24,$F2,$3E
	.dc.b	$04,$65,$F8,$46,$74,$FC,$2C,$65,$FC,$3C,$67,$3F,$F2,$9C,$E5,$FF
	.dc.b	$32,$5F,$83,$6F,$EF,$37,$C2,$FC,$E7,$82,$33,$5C,$23,$18,$79,$9D
	.dc.b	$D7,$8A,$B7,$1F,$3E,$FB,$56,$DB,$1C,$86,$1F,$E6,$2A,$FE,$EB,$5D
	.dc.b	$EF,$A3,$FC,$B6,$7E,$11,$B7,$F9,$85,$7B,$CF,$EB,$B5,$C9,$FF,$44
	.dc.b	$FD,$0F,$F5,$E9,$FE,$6C,$D8,$F0,$C7,$F9,$E2,$3F,$D1,$EE,$B6,$E5
	.dc.b	$29,$E1,$7F,$D4,$DB,$5C,$67,$71,$87,$C3,$CB,$83,$C9,$4F,$A2,$FF
	.dc.b	$33,$DF,$92,$74,$FF,$4E,$9D,$78,$58,$BF,$F1,$0D,$A3,$BF,$99,$2F
	.dc.b	$87,$27,$F4,$47,$DF,$19,$BF,$79,$F1,$1F,$1C,$EE,$32,$7B,$E3,$86
	.dc.b	$1F,$E2,$89,$FF,$46,$D7,$43,$BF,$A5,$F1,$BE,$89,$F2,$DF,$23,$E2
	.dc.b	$BF,$BB,$57,$F6,$59,$CB,$FD,$9E,$33,$97,$81,$B8,$FF,$7F,$6C,$FF
	.dc.b	$AB,$2B,$8D,$6F,$AB,$B1,$F5,$BE,$85,$EA,$E2,$FF,$BA,$46,$1F,$EE
	.dc.b	$DF,$45,$FE,$69,$73,$EB,$7D,$E8,$C6,$1F,$0A,$93,$D4,$FC,$1F,$D1
	.dc.b	$3C,$FC,$61,$FE,$01,$8C,$37,$93,$67,$09,$B1,$F7,$7C,$9F,$F1,$4B
	.dc.b	$63,$F0,$B1,$8C,$7F,$CB,$EF,$46,$F1,$EB,$D5,$DB,$6F,$F1,$1F,$45
	.dc.b	$FE,$12,$6F,$95,$F9,$08,$7D,$15,$8F,$C9,$FD,$13,$D5,$A6,$3D,$6B
	.dc.b	$87,$E4,$FE,$11,$F0,$7F,$9A,$6C,$67,$E0,$B1,$39,$E7,$D7,$C9,$5A
	.dc.b	$E4,$FE,$8B,$E8,$9F,$3F,$7E,$2B,$0D,$AF,$D1,$69,$FE,$FC,$EE,$FF
	.dc.b	$3C,$AF,$89,$FD,$5B,$99,$FF,$2E,$B9,$78,$FC,$16,$8D,$F5,$D8,$FD
	.dc.b	$77,$D1,$3E,$8B,$F6,$AD,$8C,$FC,$6E,$27,$3F,$3F,$27,$8F,$CE,$5E
	.dc.b	$CF,$F4,$AF,$C3,$3E,$83,$F7,$EF,$F8,$06,$BF,$78,$E5,$BE,$97,$F0
	.dc.b	$4F,$9D,$F9,$77,$FE,$9E,$6F,$0F,$FA,$61,$74,$6F,$A2,$61,$FC,$1E
	.dc.b	$76,$FF,$9F,$9D,$B9,$8D,$DF,$57,$C1,$58,$BF,$F3,$76,$BF,$B4,$AF
	.dc.b	$ED,$0D,$44,$70,$DE,$8F,$F3,$B3,$A3,$E6,$3F,$C8,$3E,$8B,$FC,$BB
	.dc.b	$8F,$F9,$9E,$BF,$7B,$5F,$D0,$A7,$2F,$C1,$E6,$F8,$9C,$FF,$15,$B7
	.dc.b	$EF,$6C,$EE,$3E,$14,$FF,$2E,$F9,$FF,$F1,$9C,$7F,$2B,$6C,$FF,$B2
	.dc.b	$9F,$EF,$F1,$9F,$67,$81,$1E,$1E,$CD,$AE,$13,$74,$FF,$65,$DF,$29
	.dc.b	$CD,$FF,$A0,$FE,$89,$D3,$F1,$27,$E4,$C6,$27,$FC,$2E,$D8,$FF,$0B
	.dc.b	$07,$0F,$F8,$65,$D5,$FE,$0F,$CF,$FE,$AB,$18,$AF,$C0,$47,$C1,$F3
	.dc.b	$64,$7C,$1F,$8D,$B8,$7C,$19,$B8,$7A,$A6,$D7,$E2,$7E,$17,$F2,$9F
	.dc.b	$01,$F9,$56,$C7,$E3,$DB,$87,$E5,$DB,$7F,$97,$6C,$F3,$3F,$2D,$5F
	.dc.b	$DB,$3E,$87,$F4,$57,$F0,$1A,$FF,$8D,$DA,$FE,$2D,$AF,$E4,$9E,$37
	.dc.b	$FE,$FC,$E2,$7D,$1F,$5E,$BF,$C8,$3F,$AD,$F9,$59,$97,$E9,$FC,$6B
	.dc.b	$97,$EB,$E6,$E3,$FA,$F9,$B5,$FE,$E8,$E2,$FE,$86,$FE,$2F,$BF,$F1
	.dc.b	$DE,$17,$E3,$DB,$19,$3F,$21,$19,$69,$9C,$7F,$25,$19,$7F,$B8,$EB
	.dc.b	$FD,$F6,$7F,$C0,$5F,$50,$BF,$91,$EC,$FF,$82,$A3,$13,$FE,$D1,$AE
	.dc.b	$17,$FD,$FB,$35,$6A,$3A,$DF,$1C,$3F,$4D,$E1,$7E,$9F,$D0,$B9,$7E
	.dc.b	$9D,$B7,$FA,$76,$E3,$FA,$07,$F8,$C7,$D1,$CA,$FE,$25,$DF,$6F,$F8
	.dc.b	$D4,$DB,$FE,$35,$36,$BF,$8B,$4D,$AF,$F2,$1F,$85,$FC,$12,$CA,$75
	.dc.b	$18,$7F,$94,$5E,$27,$57,$FC,$8E,$72,$D2,$5F,$91,$F3,$98,$BF,$E5
	.dc.b	$D3,$CE,$FA,$36,$74,$7A,$0B,$5E,$89,$DC,$62,$BD,$31,$89,$F0,$1F
	.dc.b	$CF,$5C,$47,$82,$7C,$73,$A8,$C3,$63,$F9,$E4,$F0,$7C,$51,$CD,$F0
	.dc.b	$5F,$C2,$8E,$07,$96,$33,$7E,$B8,$CD,$FC,$B9,$E3,$7D,$67,$E9,$D7
	.dc.b	$CA,$77,$7D,$6F,$CA,$7D,$CB,$EA,$72,$2C,$F0,$BE,$93,$DE,$3D,$07
	.dc.b	$4B,$F2,$47,$4A,$F6,$B6,$7D,$46,$E3,$B3,$4E,$CF,$82,$38,$1F,$49
	.dc.b	$C1,$D0,$F9,$66,$23,$DE,$FD,$28,$E3,$7E,$F8,$E1,$3E,$29,$CD,$7A
	.dc.b	$27,$31,$ED,$38,$DF,$C0,$BE,$57,$C6,$B7,$18,$AE,$11,$E4,$9F,$64
	.dc.b	$72,$8F,$33,$A8,$E4,$7A,$0D,$C7,$18,$F2,$BF,$19,$B1,$F1,$8D,$9C
	.dc.b	$5F,$B4,$E1,$EC,$7C,$97,$13,$E7,$3F,$6C,$5B,$F8,$51,$CE,$BD,$B1
	.dc.b	$F8,$9A,$F0,$9E,$B9,$E7,$19,$9D,$9E,$72,$C7,$9E,$78,$C7,$C8,$BF
	.dc.b	$99,$5C,$6B,$C0,$7F,$85,$BE,$89,$F4,$7F,$E1,$1F,$88,$FF,$5D,$C6
	.dc.b	$30,$FF,$57,$73,$FF,$A8,$72,$47,$F8,$66,$75,$87,$F2,$BA,$D4,$E7
	.dc.b	$0E,$8F,$F8,$C7,$CD,$FF,$1D,$1F,$EB,$EF,$53,$FC,$A9,$7F,$C7,$23
	.dc.b	$33,$B3,$FD,$7D,$3A,$C7,$F8,$FC,$FF,$86,$7E,$4E,$57,$F8,$67,$E3
	.dc.b	$2E,$1F,$E1,$99,$B7,$FE,$12,$9B,$5F,$D5,$E6,$D7,$FB,$BB,$9D,$58
	.dc.b	$87,$D1,$3E,$85,$EF,$FF,$84,$A7,$31,$87,$FC,$2B,$2F,$A2,$7C,$B6
	.dc.b	$E3,$95,$F1,$FA,$87,$C7,$F3,$C1,$18,$75,$19,$8C,$47,$F8,$8B,$E8
	.dc.b	$5F,$47,$FE,$4B,$E0,$FA,$B3,$FE,$52,$9C,$BF,$E5,$48,$CA,$8C,$BF
	.dc.b	$ED,$CF,$A3,$CC,$EA,$72,$DA,$ED,$C3,$FE,$59,$9C,$BF,$E5,$BA,$E2
	.dc.b	$E7,$E1,$F7,$87,$8D,$AF,$85,$B4,$F9,$5B,$12,$7D,$15,$CD,$A7,$B6
	.dc.b	$74,$FF,$BD,$29,$58,$C7,$D5,$5E,$0F,$A3,$7B,$DF,$3D,$B8,$7D,$B6
	.dc.b	$BF,$E4,$E6,$BF,$E7,$70,$EA,$BE,$5B,$71,$FF,$3E,$36,$7E,$97,$C2
	.dc.b	$7C,$0F,$A3,$73,$D2,$4F,$B0,$47,$E4,$9C,$3C,$1B,$1C,$16,$1F,$72
	.dc.b	$78,$46,$30,$FF,$47,$47,$FC,$00,$8E,$7E,$77,$8C,$F2,$79,$F2,$32
	.dc.b	$43,$91,$FE,$9C,$9E,$4E,$0F,$3C,$73,$78,$5F,$8E,$3E,$FF,$D1,$BD
	.dc.b	$6F,$CF,$36,$BF,$54,$D7,$FC,$FA,$D7,$E8,$39,$AF,$CC,$B7,$1F,$F5
	.dc.b	$53,$67,$CD,$F0,$1F,$8E,$FA,$37,$9F,$7C,$F9,$5F,$3E,$72,$D9,$CB
	.dc.b	$ED,$84,$F3,$9E,$9E,$E4,$AE,$4F,$7F,$73,$11,$C6,$2E,$7F,$B0,$C3
	.dc.b	$FD,$AB,$3C,$08,$6C,$F2,$57,$EA,$25,$F4,$4F,$A3,$C3,$6B,$F4,$73
	.dc.b	$A9,$CF,$D1,$EB,$E5,$25,$6A,$D1,$CB,$E8,$FB,$7D,$1F,$EA,$2B,$9B
	.dc.b	$FB,$36,$F5,$6B,$AF,$C3,$CB,$E2,$E7,$95,$D6,$3C,$BE,$F7,$9B,$D8
	.dc.b	$93,$E1,$8E,$FF,$1B,$23,$BB,$E2,$F3,$3D,$BE,$A4,$9F,$43,$F0,$3E
	.dc.b	$17,$E0,$F7,$49,$F6,$3F,$13,$CC,$F5,$FC,$47,$B3,$C3,$95,$EC,$8F
	.dc.b	$99,$F4,$7F,$CD,$23,$E4,$AF,$97,$C9,$AF,$37,$7F,$47,$C3,$C6,$97
	.dc.b	$C2,$3E,$4F,$B9,$23,$E4,$BF,$33,$E8,$FF,$8F,$FD,$07,$F0,$39,$7D
	.dc.b	$1F,$F7,$07,$F8,$03,$5F,$D6,$46,$FE,$8D,$EE,$FF,$28,$5F,$0D,$4C
	.dc.b	$47,$F2,$CF,$92,$FC,$25,$FE,$D5,$BE,$27,$F4,$06,$F9,$89,$90,$E7
	.dc.b	$F3,$F7,$C9,$FE,$29,$9F,$E0,$DE,$9C,$DD,$C8,$FF,$19,$BF,$E2,$38
	.dc.b	$FF,$4E,$FA,$37,$CB,$7F,$40,$6D,$7F,$1F,$6B,$FC,$2D,$B3,$FD,$5B
	.dc.b	$07,$C0,$7F,$48,$FA,$37,$C5,$7C,$CC,$66,$3F,$CA,$67,$2D,$9C,$BF
	.dc.b	$32,$79,$44,$8D,$10,$FA,$3E,$CF,$7E,$0F,$B3,$F2,$32,$7B,$46,$5F
	.dc.b	$8B,$19,$7F,$92,$61,$F6,$1F,$E7,$9F,$47,$FB,$07,$F6,$FF,$3B,$F2
	.dc.b	$F5,$FE,$EE,$1C,$71,$FD,$59,$57,$17,$F9,$83,$6B,$FB,$DB,$67,$83
	.dc.b	$67,$EF,$2F,$90,$FE,$E3,$8F,$E7,$E7,$FA,$83,$FA,$37,$F9,$E3,$6F
	.dc.b	$FC,$0E,$EE,$33,$DB,$C8,$8F,$F5,$9F,$A3,$ED,$5F,$8D,$FA,$15,$FA
	.dc.b	$E7,$34,$9C,$BF,$93,$9E,$D1,$88,$E3,$7F,$0E,$31,$5C,$9B,$87,$F6
	.dc.b	$2A,$E9,$7F,$24,$E6,$5B,$E4,$F8,$E7,$85,$72,$C3,$D1,$3B,$BA,$79
	.dc.b	$6F,$8B,$FA,$9C,$3C,$53,$CE,$B7,$3E,$82,$C7,$89,$E1,$87,$82,$38
	.dc.b	$47,$A0,$EC,$70,$BE,$EA,$FB,$BC,$AF,$FC,$A5,$1F,$32,$3F,$C2,$CD
	.dc.b	$7A,$98,$F9,$37,$D1,$F0,$6C,$7F,$87,$4E,$CE,$1E,$11,$FE,$55,$5F
	.dc.b	$47,$48,$FF,$96,$5B,$9F,$D9,$C3,$FA,$7E,$BF,$12,$BF,$88,$57,$C6
	.dc.b	$9C,$C7,$C1,$3F,$41,$19,$F4,$7F,$CB,$77,$F8,$39,$E2,$6E,$B8,$C6
	.dc.b	$63,$CA,$66,$78,$E9,$ED,$BE,$98,$78,$EB,$77,$E7,$BD,$C7,$F8,$FE
	.dc.b	$B6,$F9,$4F,$0C,$EE,$B9,$4E,$A3,$13,$FE,$25,$C3,$DA,$7F,$9C,$1E
	.dc.b	$38,$7F,$9D,$F0,$FE,$03,$85,$3F,$C8,$55,$C1,$F4,$57,$AD,$FF,$23
	.dc.b	$7D,$1F,$F5,$76,$EB,$CA,$7F,$9E,$4F,$F3,$D9,$C9,$23,$FC,$71,$1C
	.dc.b	$5A,$71,$0E,$6B,$FD,$0E,$FF,$3A,$9F,$0F,$2F,$79,$7D,$1F,$F9,$B2
	.dc.b	$72,$4B,$FF,$24,$CE,$BC,$2E,$D8,$73,$9C,$DA,$73,$E6,$4C,$3F,$93
	.dc.b	$F4,$DB,$F4,$CD,$C7,$F9,$4F,$2F,$FB,$42,$7E,$0C,$E6,$FF,$D4,$13
	.dc.b	$FE,$84,$7E,$4C,$7A,$63,$13,$FE,$5B,$FA,$37,$CB,$FC,$3F,$87,$FE
	.dc.b	$E3,$87,$F9,$D6,$72,$FF,$9D,$E3,$35,$FE,$63,$6E,$3F,$EA,$E6,$CF
	.dc.b	$F3,$BA,$E3,$5B,$EA,$EC,$7D,$1F,$A2,$7A,$78,$BF,$F4,$24,$61,$FF
	.dc.b	$43,$7D,$1F,$F9,$71,$CF,$A9,$F6,$63,$18,$7F,$9D,$69,$3D,$4F,$BD
	.dc.b	$F4,$6F,$37,$18,$7F,$A4,$63,$0D,$E5,$D9,$C2,$6C,$7F,$A5,$FE,$F5
	.dc.b	$DB,$F8,$96,$C7,$F9,$53,$C9,$7C,$DE,$6B,$F1,$D2,$F9,$FA,$7E,$2E
	.dc.b	$9E,$9E,$27,$FD,$48,$3E,$63,$89,$F9,$1F,$A4,$7C,$26,$27,$FD,$55
	.dc.b	$3F,$E7,$86,$FC,$37,$E5,$FE,$8F,$F1,$D1,$83,$FD,$61,$F6,$FF,$EE
	.dc.b	$36,$FC,$77,$E7,$69,$F4,$3F,$A5,$AD,$73,$7F,$A8,$5F,$F7,$8D,$F2
	.dc.b	$F9,$1F,$D2,$7E,$22,$ED,$E1,$1F,$AD,$FB,$EF,$0E,$1F,$EA,$96,$F7
	.dc.b	$FF,$80,$BE,$99,$F3,$7B,$F1,$BF,$69,$E6,$BE,$6E,$9F,$B1,$E6,$7D
	.dc.b	$2E,$37,$FF,$65,$7D,$C3,$E2,$7F,$78,$D8,$FF,$43,$D3,$E8,$B3,$5F
	.dc.b	$C0,$7E,$F3,$FC,$85,$5F,$ED,$36,$F0,$7E,$EB,$EA,$3F,$D2,$53,$9E
	.dc.b	$DF,$E1,$1F,$0D,$F7,$A3,$FD,$B4,$DA,$E1,$3A,$F0,$77,$9C,$BF,$F4
	.dc.b	$CF,$C5,$3E,$B3,$18,$7F,$B3,$77,$FD,$4B,$FE,$57,$6F,$2B,$E8,$5F
	.dc.b	$23,$FC,$66,$77,$5F,$13,$CE,$C1,$FC,$5E,$B7,$97,$FB,$BE,$B6,$74
	.dc.b	$FA,$2F,$E2,$98,$A7,$FC,$65,$39,$F2,$BF,$90,$7D,$2F,$E1,$27,$FD
	.dc.b	$62,$FF,$B7,$5B,$E0,$EB,$E7,$36,$D8,$C9,$C2,$3C,$9E,$FC,$8F,$F5
	.dc.b	$DE,$7F,$AA,$D7,$F9,$4B,$FE,$39,$8C,$CE,$E7,$3E,$5D,$D3,$D1,$1F
	.dc.b	$E7,$DF,$A9,$7C,$9F,$F3,$D6,$E1,$E8,$F0,$31,$32,$3F,$D0,$97,$CA
	.dc.b	$D7,$C1,$FE,$77,$AF,$E5,$AB,$67,$F9,$5E,$75,$5C,$5B,$1D,$27,$F9
	.dc.b	$0C,$E6,$FE,$5F,$AD,$F5,$7D,$59,$7F,$0B,$B6,$E5,$F0,$A6,$E1,$FE
	.dc.b	$68,$DB,$F3,$93,$9B,$FA,$7F,$17,$F5,$BE,$1B,$EB,$B6,$32,$7D,$88
	.dc.b	$CB,$4F,$5C,$FF,$A9,$7E,$A3,$F3,$58,$7C,$36,$CF,$C6,$C3,$DB,$5B
	.dc.b	$EC,$FE,$D5,$5A,$8C,$57,$DC,$D7,$35,$23,$FC,$61,$2A,$FD,$43,$FE
	.dc.b	$81,$31,$3F,$E0,$97,$99,$9C,$1B,$7F,$40,$7F,$D1,$6D,$BF,$A1,$E6
	.dc.b	$7E,$63,$E7,$24,$7E,$61,$FC,$DC,$62,$FF,$1F,$EB,$FE,$9F,$C4,$FD
	.dc.b	$43,$6F,$F2,$AD,$9D,$C7,$E6,$27,$E0,$F9,$12,$3F,$4B,$F5,$96,$99
	.dc.b	$FC,$BA,$C3,$E8,$9B,$5F,$EA,$1D,$7F,$92,$B6,$BF,$63,$3A,$BE,$7B
	.dc.b	$FF,$45,$AF,$D4,$F8,$7F,$BF,$F4,$BF,$80,$36,$32,$7E,$BA,$32,$FE
	.dc.b	$EB,$EA,$3F,$7A,$7E,$FB,$C1,$B1,$92,$61,$FA,$28,$CB,$EC,$8F,$F5
	.dc.b	$C7,$D4,$7E,$96,$30,$7F,$BB,$DC,$6B,$F9,$9F,$17,$FC,$61,$18,$AE
	.dc.b	$3D,$98,$3F,$8C,$4D,$8F,$CA,$DA,$B8,$3D,$6B,$73,$A9,$CC,$EE,$79
	.dc.b	$3F,$C3,$2A,$DF,$EF,$EB,$67,$47,$F9,$74,$6A,$FC,$A7,$5B,$EF,$7E
	.dc.b	$E3,$B4,$64,$8F,$76,$D7,$79,$E9,$3F,$6A,$AB,$C8,$E9,$7E,$3D,$13
	.dc.b	$A8,$F3,$D7,$8A,$FA,$61,$D8,$F3,$E1,$E1,$39,$CF,$13,$84,$6A,$78
	.dc.b	$CE,$A3,$F9,$9D,$ED,$E9,$1E,$76,$BE,$79,$E2,$7D,$D8,$CD,$F9,$9E
	.dc.b	$66,$2F,$E0,$CF,$CE,$8C,$1D,$9C,$57,$8C,$8F,$68,$FE,$25,$7E,$59
	.dc.b	$CC,$60,$E6,$E0,$F4,$14,$F4,$3F,$04,$E0,$E6,$FD,$0F,$C9,$AF,$03
	.dc.b	$F8,$F9,$F3,$61,$F5,$1E,$F3,$A3,$D2,$70,$9E,$E7,$13,$A4,$78,$5C
	.dc.b	$47,$19,$E9,$1E,$3C,$3E,$3C,$62,$7A,$5F,$99,$FA,$15,$BB,$F2,$E1
	.dc.b	$F1,$67,$A3,$63,$C1,$1E,$67,$C5,$3E,$63,$CF,$19,$7C,$B1,$AA,$FC
	.dc.b	$A5,$73,$9C,$9F,$3A,$BC,$B8,$7A,$27,$53,$D3,$0F,$56,$1C,$EB,$4F
	.dc.b	$CA,$3C,$B1,$A9,$EA,$79,$70,$F0,$CE,$E3,$C3,$87,$9D,$D5,$F4,$7E
	.dc.b	$21,$F8,$38,$D5,$7C,$9A,$EF,$87,$9A,$3A,$D7,$0A,$F9,$35,$DE,$3D
	.dc.b	$55,$BB,$FD,$5C,$79,$AF,$A3,$F0,$EB,$DC,$7D,$43,$F3,$DC,$9F,$E5
	.dc.b	$D7,$FC,$57,$E5,$FF,$B3,$7E,$63,$FC,$D5,$18,$25,$7F,$84,$A7,$5E
	.dc.b	$8F,$F8,$D1,$FF,$1F,$7C,$C4,$AF,$F1,$5F,$C8,$DB,$FF,$15,$CD,$D7
	.dc.b	$FC,$15,$36,$95,$3E,$F7,$FD,$2D,$1F,$DD,$9B,$1F,$DD,$43,$87,$FB
	.dc.b	$C3,$5E,$63,$EB,$7D,$6F,$16,$BF,$DF,$DA,$F0,$6B,$E4,$FA,$87,$D5
	.dc.b	$7F,$A3,$63,$C5,$18,$D3,$FC,$16,$BE,$A9,$E8,$E2,$F8,$5F,$F8,$37
	.dc.b	$DB,$7B,$3E,$6D,$BF,$C8,$11,$89,$FE,$9E,$D7,$F9,$F7,$D4,$FF,$91
	.dc.b	$67,$75,$C7,$6F,$7D,$EE,$73,$7F,$D6,$2F,$85,$6B,$E7,$F8,$4F,$D0
	.dc.b	$B4,$62,$11,$87,$C7,$F3,$39,$8C,$47,$F5,$16,$E1,$EA,$AD,$57,$3F
	.dc.b	$A9,$7D,$07,$1F,$92,$98,$FF,$6C,$9C,$B6,$72,$F5,$9E,$51,$23,$44
	.dc.b	$3C,$5F,$0F,$E9,$FA,$97,$3D,$24,$FF,$7A,$87,$3E,$23,$FB,$54,$70
	.dc.b	$75,$3C,$E7,$A7,$9D,$D7,$CF,$95,$CE,$32,$F5,$8C,$BD,$23,$E6,$7D
	.dc.b	$4B,$D5,$F3,$7C,$44,$AF,$F0,$6C,$E5,$B3,$97,$CB,$3C,$A2,$46,$88
	.dc.b	$7A,$BE,$A7,$ED,$7D,$4B,$D0,$BE,$84,$AF,$F0,$DC,$E5,$B3,$97,$FC
	.dc.b	$19,$09,$E7,$3D,$3E,$26,$57,$27,$BF,$D4,$E2,$38,$C5,$E7,$FF,$4E
	.dc.b	$60,$AF,$C7,$77,$34,$F3,$F5,$4C,$7F,$C5,$F3,$96,$CE,$5F,$2C,$F2
	.dc.b	$89,$1A,$21,$FE,$25,$FC,$6D,$F4,$9F,$9D,$FB,$9F,$81,$3F,$E4,$6C
	.dc.b	$DF,$4E,$EA,$FD,$2C,$61,$E1,$BF,$E9,$7E,$E9,$EE,$73,$F3,$66,$3F
	.dc.b	$E4,$69,$CB,$FE,$48,$FB,$AF,$53,$FE,$37,$CF,$F8,$33,$FE,$4C,$6B
	.dc.b	$FE,$3B,$8D,$E5,$FE,$C3,$F8,$8F,$F7,$B4,$FF,$A9,$A3,$CF,$F4,$7E
	.dc.b	$6D,$4F,$EE,$6B,$75,$CF,$B7,$F7,$9F,$75,$FB,$6E,$3F,$E4,$AD,$9F
	.dc.b	$E6,$39,$7E,$85,$FF,$26,$7D,$D7,$F0,$27,$FC,$C3,$2F,$7B,$99,$FC
	.dc.b	$1E,$75,$3F,$E5,$3C,$5F,$65,$E5,$F8,$79,$3F,$E6,$A9,$CB,$FE,$6C
	.dc.b	$FA,$FF,$C2,$3F,$E6,$EF,$B9,$7B,$9F,$C4,$DB,$1B,$EC,$FE,$2D,$4F
	.dc.b	$57,$5F,$47,$2B,$A4,$7C,$3F,$9F,$F8,$DF,$41,$27,$C5,$1F,$C7,$FE
	.dc.b	$E1,$F0,$BF,$C8,$70,$FD,$05,$7F,$A0,$1A,$FF,$9B,$E7,$6F,$FA,$13
	.dc.b	$F1,$7F,$88,$EF,$C5,$E4,$7F,$2B,$8C,$C7,$F9,$DF,$EA,$FF,$8F,$BF
	.dc.b	$E7,$9F,$AF,$77,$FF,$31,$6C,$7F,$A4,$27,$7D,$AE,$AF,$EB,$BF,$5E
	.dc.b	$F8,$4F,$E7,$0B,$E3,$69,$88,$FE,$79,$F5,$EF,$B3,$A4,$CB,$FD,$3B
	.dc.b	$0F,$89,$FE,$85,$9B,$D9,$A4,$AD,$FA,$BD,$08,$E0,$FF,$53,$FD,$83
	.dc.b	$BB,$FA,$63,$63,$31,$83,$BF,$47,$F5,$1F,$B0,$7C,$8F,$8A,$65,$7F
	.dc.b	$AC,$A7,$2D,$9C,$BF,$36,$79,$44,$8D,$10,$FA,$5E,$3F,$7D,$FD,$FE
	.dc.b	$6C,$9E,$D1,$97,$E9,$46,$5F,$F5,$86,$1F,$45,$FF,$62,$FD,$83,$AF
	.dc.b	$F8,$4F,$D5,$7C,$99,$FD,$C3,$5F,$F5,$C4,$ED,$E3,$3F,$DC,$63,$11
	.dc.b	$A8,$C7,$CB,$7F,$75,$C1,$F5,$B7,$E1,$A5,$FF,$B5,$A7,$2D,$9E,$04
	.dc.b	$7F,$DB,$7F,$60,$B7,$5F,$F8,$CB,$EB,$FF,$91,$57,$F8,$D2,$B5,$39
	.dc.b	$C3,$FC,$5A,$76,$95,$70,$7F,$95,$61,$FE,$32,$8C,$5F,$29,$DD,$7F
	.dc.b	$8D,$A3,$13,$C6,$3A,$36,$B9,$46,$F0,$F3,$DF,$23,$CC,$E2,$B8,$47
	.dc.b	$9C,$FF,$26,$CF,$43,$96,$19,$86,$1E,$29,$F6,$E1,$E0,$9E,$2D,$77
	.dc.b	$87,$79,$E7,$1E,$33,$E0,$9D,$5F,$03,$E6,$BF,$05,$FC,$38,$FF,$2B
	.dc.b	$9D,$23,$FC,$BC,$D3,$FC,$C1,$1B,$7E,$47,$CC,$BE,$C3,$72,$8C,$4F
	.dc.b	$06,$BD,$A3,$53,$DA,$3F,$CE,$3F,$47,$53,$48,$ED,$F9,$17,$F9,$FE
	.dc.b	$FF,$95,$3C,$8F,$85,$19,$7E,$21,$9D,$3F,$0D,$19,$7F,$11,$19,$CB
	.dc.b	$FC,$C9,$B7,$F1,$F8,$E2,$79,$0E,$AE,$67,$8C,$65,$EB,$19,$9F,$F4
	.dc.b	$46,$1E,$88,$C5,$7D,$08,$C6,$1E,$A8,$E5,$5E,$28,$C4,$EA,$7C,$31
	.dc.b	$FE,$98,$C3,$FC,$E9,$7E,$77,$D1,$3F,$E9,$2A,$E1,$1A,$AE,$07,$59
	.dc.b	$FD,$8B,$E6,$7A,$98,$3F,$D4,$BF,$61,$FC,$02,$38,$C6,$36,$F0,$47
	.dc.b	$FA,$82,$7F,$D5,$6E,$A5,$1C,$5E,$86,$27,$8B,$CE,$78,$4F,$28,$DB
	.dc.b	$C6,$BD,$AF,$F3,$D9,$DF,$37,$29,$7D,$87,$FA,$02,$72,$4B,$FF,$18
	.dc.b	$43,$E6,$7C,$31,$E6,$BE,$53,$9A,$F1,$F9,$3A,$C4,$F7,$E5,$75,$66
	.dc.b	$25,$70,$3F,$98,$C6,$31,$F5,$CF,$1F,$8A,$C6,$9F,$67,$0F,$F5,$1E
	.dc.b	$1F,$27,$EC,$3F,$C7,$EF,$CB,$F9,$1F,$EE,$0E,$7D,$CF,$4C,$63,$2F
	.dc.b	$9A,$DC,$3F,$06,$DA,$FF,$6A,$DF,$4A,$DF,$67,$23,$E1,$FD,$7B,$E2
	.dc.b	$31,$7F,$6E,$30,$FD,$DF,$B0,$FF,$3F,$39,$F9,$EF,$97,$18,$C3,$EF
	.dc.b	$D2,$7E,$7F,$D8,$3E,$3A,$9E,$08,$B8,$7E,$26,$7E,$F5,$FF,$BC,$5A
	.dc.b	$FF,$18,$6B,$FE,$F0,$6B,$FE,$FC,$FB,$0F,$E2,$93,$FC,$71,$B1,$FE
	.dc.b	$FB,$6B,$FC,$89,$AF,$FB,$EA,$3F,$DF,$8F,$E7,$3E,$1F,$F5,$AE,$5A
	.dc.b	$E6,$3F,$94,$9F,$E4,$A2,$FB,$5A,$8F,$F1,$53,$63,$51,$FC,$C8,$FF
	.dc.b	$28,$FC,$56,$E7,$FD,$45,$3F,$CD,$1B,$3B,$6B,$FC,$EB,$EB,$BF,$DD
	.dc.b	$33,$FC,$E9,$B1,$D5,$AE,$FE,$4F,$C9,$0F,$95,$A4,$AE,$4F,$A7,$EB
	.dc.b	$7F,$49,$7E,$76,$BF,$D2,$9A,$88,$E1,$FE,$7F,$EE,$7C,$3F,$07,$FC
	.dc.b	$D5,$7F,$D4,$1B,$3F,$CA,$27,$E4,$7D,$63,$E4,$E9,$2F,$FA,$BC,$65
	.dc.b	$FE,$B3,$F5,$9F,$E4,$B6,$FA,$BF,$46,$31,$87,$CF,$6B,$FD,$8B,$CD
	.dc.b	$FF,$47,$57,$F6,$26,$C7,$D5,$6B,$BF,$86,$FB,$90,$FA,$CA,$4A,$E4
	.dc.b	$FE,$07,$EA,$7F,$65,$7F,$15,$BD,$EF,$A3,$7D,$5F,$F1,$49,$DD,$1F
	.dc.b	$DF,$3F,$DE,$1A,$FC,$F3,$5D,$FD,$C9,$8F,$FA,$E7,$EA,$FF,$57,$8F
	.dc.b	$F4,$6B,$F0,$C6,$27,$FA,$67,$D1,$BE,$AF,$F9,$62,$F9,$AB,$18,$6C
	.dc.b	$FF,$B1,$67,$4D,$9D,$3F,$EC,$9D,$3F,$C8,$0D,$F5,$BF,$9C,$47,$ED
	.dc.b	$32,$FE,$B4,$EB,$4F,$F7,$33,$63,$33,$C3,$D6,$E2,$FF,$3A,$FA,$B7
	.dc.b	$D5,$D3,$D3,$F0,$D7,$2F,$EA,$93,$A7,$FC,$3E,$DC,$BF,$64,$D7,$FC
	.dc.b	$4A,$D7,$E7,$3B,$C7,$FB,$86,$9F,$CA,$DF,$EE,$31,$9A,$F9,$7E,$5F
	.dc.b	$BB,$C9,$91,$F0,$FA,$AE,$3E,$E9,$B8,$7C,$29,$B5,$F5,$3E,$13,$E9
	.dc.b	$FB,$DF,$51,$B1,$F4,$5B,$87,$F7,$9B,$67,$FE,$B5,$AF,$AD,$7F,$D8
	.dc.b	$3E,$AF,$F3,$B5,$FE,$08,$6C,$FF,$83,$78,$7D,$58,$7E,$EE,$72,$DE
	.dc.b	$AF,$E3,$91,$FA,$48,$FC,$93,$5D,$57,$F9,$32,$38,$DB,$97,$FD,$EF
	.dc.b	$3F,$E0,$79,$CC,$62,$BD,$59,$7F,$4B,$BF,$9F,$F3,$1F,$94,$F1,$BF
	.dc.b	$2A,$CC,$7F,$24,$E5,$A7,$E4,$7C,$AF,$C7,$FC,$67,$E4,$1B,$19,$3F
	.dc.b	$2D,$19,$69,$E5,$F6,$3E,$41,$FA,$4F,$AB,$B1,$93,$DD,$1F,$10,$8E
	.dc.b	$0F,$F1,$4E,$2F,$B5,$FF,$80,$31,$C7,$4F,$B1,$18,$7F,$43,$E5,$FE
	.dc.b	$9B,$C4,$FD,$3B,$6B,$F5,$8D,$9D,$C7,$EB,$E7,$F7,$3F,$59,$23,$F6
	.dc.b	$DE,$0D,$AF,$9B,$37,$1F,$CD,$CD,$AF,$E1,$9F,$56,$EC,$FF,$13,$B6
	.dc.b	$3F,$79,$F5,$98,$99,$1F,$E4,$19,$D6,$3F,$BA,$AC,$BF,$E4,$28,$CE
	.dc.b	$1F,$E5,$86,$D7,$E4,$6B,$73,$9B,$B3,$93,$B3,$8B,$FE,$A1,$7C,$9F
	.dc.b	$F6,$6F,$3B,$0D,$42,$74,$DA,$FE,$27,$3A,$9C,$9C,$F0,$FE,$6D,$19
	.dc.b	$9F,$04,$63,$0F,$19,$2F,$DD,$3D,$23,$13,$EB,$69,$EB,$9D,$B8,$7D
	.dc.b	$73,$C5,$C9,$27,$93,$F3,$67,$83,$EB,$79,$47,$96,$38,$47,$DC,$9F
	.dc.b	$4E,$1F,$16,$38,$E1,$F0,$27,$C1,$7D,$89,$3C,$4F,$C0,$9C,$23,$9E
	.dc.b	$1F,$16,$3C,$37,$D7,$0F,$09,$EC,$8C,$CE,$E3,$37,$E7,$3B,$BB,$33
	.dc.b	$5C,$E3,$31,$EF,$33,$3C,$E3,$89,$CE,$BD,$13,$E2,$38,$47,$F5,$09
	.dc.b	$F1,$47,$FB,$93,$0F,$CC,$B7,$0F,$F1,$D4,$62,$39,$99,$BF,$F7,$4C
	.dc.b	$66,$3D,$06,$4F,$E6,$9C,$BF,$E8,$89,$FA,$AD,$8F,$82,$DC,$3D,$2D
	.dc.b	$9F,$4C,$7F,$AF,$E3,$13,$F1,$63,$1E,$D7,$E5,$27,$33,$EE,$AF,$F6
	.dc.b	$44,$62,$FF,$38,$DC,$3F,$D5,$4D,$9F,$F0,$45,$F3,$23,$E8,$6D,$FA
	.dc.b	$1B,$87,$A1,$B3,$E8,$3C,$2E,$27,$D1,$18,$3A,$11,$E8,$D9,$E8,$D3
	.dc.b	$A4,$23,$FA,$64,$F5,$8F,$80,$66,$BD,$8F,$D2,$8D,$61,$F0,$9B,$87
	.dc.b	$C2,$6E,$1F,$08,$F0,$57,$A0,$E9,$1A,$7A,$46,$9F,$0E,$1E,$48,$C5
	.dc.b	$F7,$6D,$F7,$7F,$C5,$5E,$57,$F8,$C0,$FF,$6E,$7D,$63,$D8,$C4,$E8
	.dc.b	$98,$FF,$7F,$9D,$46,$31,$FF,$15,$FC,$DF,$F8,$B8,$FF,$17,$FC,$55
	.dc.b	$9F,$F1,$A4,$5B,$FF,$17,$CD,$A5,$7F,$93,$7E,$B3,$FC,$AA,$FF,$A5
	.dc.b	$E9,$18,$7F,$D3,$4D,$7F,$CA,$86,$BC,$BF,$F5,$17,$D5,$BE,$AF,$D9
	.dc.b	$3F,$EA,$18,$C6,$9D,$57,$D2,$3D,$0C,$5F,$0B,$EB,$F6,$AF,$8E,$F2
	.dc.b	$42,$30,$F9,$DA,$FF,$AB,$DA,$F9,$BD,$9F,$F5,$13,$FE,$99,$DB,$E0
	.dc.b	$BF,$E9,$B8,$CB,$F8,$FD,$BF,$CE,$11,$89,$DC,$BE,$B2,$C6,$27,$55
	.dc.b	$C3,$D9,$FF,$3C,$4E,$5B,$1F,$E7,$BA,$E2,$F6,$F8,$3C,$E1,$E3,$6A
	.dc.b	$C1,$18,$5F,$23,$62,$4F,$B6,$BA,$34,$F7,$4F,$4F,$AC,$7D,$67,$3F
	.dc.b	$8D,$95,$B1,$3A,$70,$D7,$FD,$B8,$D7,$F0,$6D,$7F,$D3,$0D,$7F,$D9
	.dc.b	$0F,$1D,$3E,$5B,$70,$DD,$FB,$7E,$0F,$E0,$FD,$63,$A6,$93,$1F,$34
	.dc.b	$3A,$71,$1F,$64,$FF,$6B,$47,$17,$73,$D2,$7A,$FA,$7D,$BD,$39,$5D
	.dc.b	$23,$2F,$68,$CB,$D4,$5A,$7E,$BF,$D6,$7F,$AA,$E7,$CB,$F0,$7E,$A3
	.dc.b	$3E,$E4,$8F,$D7,$4E,$DF,$D5,$87,$0F,$EB,$9A,$FF,$B0,$DE,$1B,$FE
	.dc.b	$51,$BB,$66,$BE,$4F,$D6,$7E,$2B,$EB,$1E,$95,$F4,$E6,$5F,$7E,$72
	.dc.b	$D9,$CB,$EF,$84,$F3,$9E,$9F,$3D,$2B,$93,$DF,$E8,$71,$1C,$62,$E9
	.dc.b	$FA,$AC,$15,$F9,$EF,$01,$B7,$9F,$59,$8F,$E4,$E7,$2D,$9F,$E0,$71
	.dc.b	$88,$F9,$2D,$7F,$8A,$B5,$FF,$75,$3B,$C3,$F7,$6D,$AC,$D7,$EC,$3E
	.dc.b	$B1,$F5,$9F,$B2,$FA,$BF,$8B,$8F,$D6,$6B,$FC,$89,$AF,$FB,$DD,$A8
	.dc.b	$7D,$63,$DF,$FE,$36,$D8,$FE,$60,$D7,$F0,$6D,$7F,$97,$FD,$5F,$AD
	.dc.b	$E8,$D4,$D3,$1F,$50,$FA,$EF,$E5,$2D,$9F,$E7,$CD,$7F,$22,$D7,$BF
	.dc.b	$5F,$F3,$0A,$FD,$54,$7E,$59,$AE,$8F,$E9,$1F,$57,$FC,$5D,$BE,$1F
	.dc.b	$F3,$88,$C4,$7E,$79,$AF,$F4,$AB,$7D,$5F,$98,$91,$E6,$7F,$A0,$FD
	.dc.b	$57,$E2,$23,$FA,$B3,$5F,$D4,$B5,$FE,$AA,$D7,$E9,$05,$E8,$F4,$24
	.dc.b	$74,$7E,$3F,$CC,$48,$F8,$ED,$7F,$6F,$F5,$4F,$81,$FE,$AA,$D8,$FE
	.dc.b	$D4,$D7,$71,$FD,$6B,$EA,$9F,$0D,$49,$3F,$DC,$63,$2F,$F7,$3F,$A6
	.dc.b	$FE,$54,$9A,$4F,$23,$FB,$2F,$D4,$BD,$FA,$4A,$FE,$F9,$19,$7F,$BF
	.dc.b	$7D,$43,$EA,$7F,$A7,$B6,$3F,$A1,$35,$FE,$40,$BE,$0F,$FA,$AE,$3F
	.dc.b	$D4,$A7,$FA,$43,$5F,$E4,$CD,$F0,$3F,$B4,$D7,$F3,$79,$FE,$9D,$F5
	.dc.b	$2F,$A2,$F0,$F9,$D2,$73,$1F,$D8,$A7,$51,$82,$1D,$BE,$07,$FB,$3F
	.dc.b	$D4,$BE,$3B,$CB,$72,$BF,$AE,$4E,$5B,$39,$7C,$F3,$CA,$24,$68,$87
	.dc.b	$B3,$DA,$EF,$C1,$F0,$BD,$59,$3D,$A3,$2F,$C4,$8C,$BF,$D2,$70,$FA
	.dc.b	$4F,$F4,$BF,$A9,$7D,$37,$F8,$0B,$77,$D1,$FC,$9F,$2B,$FC,$30,$DB
	.dc.b	$FE,$19,$7C,$1F,$EF,$0D,$7A,$5F,$F8,$2A,$31,$7F,$E1,$A5,$F2,$BF
	.dc.b	$E0,$EC,$9F,$4D,$7E,$06,$63,$FE,$03,$9C,$B6,$78,$11,$FF,$04,$7D
	.dc.b	$4F,$4C,$7F,$90,$7D,$42,$FF,$63,$39,$A4,$E5,$FE,$4D,$3D,$9C,$BF
	.dc.b	$B8,$31,$1B,$84,$F3,$8C,$36,$78,$5F,$F0,$09,$E1,$1C,$8F,$1E,$3F
	.dc.b	$E8,$DB,$EE,$7F,$37,$BE,$4F,$92,$F9,$57,$48,$F6,$BC,$B0,$B1,$C6
	.dc.b	$FA,$61,$E0,$BE,$33,$CC,$D9,$EA,$9D,$DF,$38,$F1,$9F,$26,$78,$B4
	.dc.b	$E3,$87,$99,$E5,$87,$F9,$E6,$3E,$74,$7F,$9D,$9A,$F5,$31,$F3,$2F
	.dc.b	$A9,$E0,$D8,$FF,$09,$CA,$31,$1C,$27,$FC,$E6,$BE,$82,$91,$FF,$3B
	.dc.b	$B7,$3F,$B1,$87,$E7,$B8,$3E,$0D,$7F,$65,$AF,$87,$39,$8F,$82,$64
	.dc.b	$F8,$6E,$6B,$F0,$51,$97,$AE,$FF,$E7,$C7,$31,$E5,$38,$C6,$74,$FF
	.dc.b	$17,$DF,$82,$31,$3F,$E9,$C8,$C6,$1E,$68,$E5,$7D,$A3,$13,$A9,$F3
	.dc.b	$9A,$BF,$F4,$0D,$F4,$8F,$F6,$D3,$FE,$90,$C3,$FD,$12,$DB,$F1,$CF
	.dc.b	$99,$FF,$39,$4F,$C3,$36,$FF,$9D,$63,$E0,$57,$68,$FF,$3C,$7D,$47
	.dc.b	$EE,$9B,$AF,$5A,$FF,$09,$4F,$F8,$4E,$72,$48,$FF,$2E,$47,$16,$9C
	.dc.b	$43,$9B,$FF,$51,$BF,$CE,$67,$C3,$CD,$DE,$5F,$53,$FE,$8E,$9C,$92
	.dc.b	$FF,$CE,$D3,$AF,$96,$ED,$87,$39,$CD,$A7,$3E,$34,$C3,$FD,$39,$87
	.dc.b	$F2,$6E,$AC,$61,$EA,$AB,$3F,$3F,$99,$58,$BF,$F0,$1E,$4B,$F4,$57
	.dc.b	$FA,$BA,$7E,$5C,$63,$E1,$7E,$17,$D4,$FF,$8E,$67,$FD,$4D,$F1,$9F
	.dc.b	$DB,$DF,$F5,$44,$67,$3F,$F4,$34,$E5,$BD,$7F,$2E,$31,$A7,$FB,$32
	.dc.b	$B8,$3F,$EC,$E8,$CD,$FF,$86,$3E,$A3,$FD,$1C,$FF,$AD,$5B,$87,$48
	.dc.b	$C7,$C1,$FD,$79,$7C,$D5,$3E,$D7,$D4,$FF,$A0,$AB,$94,$7D,$DF,$A9
	.dc.b	$7C,$4D,$38,$4F,$FB,$0E,$32,$DE,$5E,$1F,$27,$C5,$B1,$FE,$C9,$FA
	.dc.b	$9F,$E1,$0F,$D6,$F2,$BF,$A1,$39,$EF,$FC,$34,$63,$6F,$F5,$64,$FF
	.dc.b	$A0,$8E,$1E,$6F,$E4,$5C,$47,$E2,$27,$F1,$8F,$0F,$9B,$FC,$87,$D9
	.dc.b	$7F,$A3,$2B,$FD,$B4,$23,$3F,$09,$F9,$B7,$FD,$37,$1F,$8A,$FB,$2F
	.dc.b	$F3,$FB,$FE,$E1,$8C,$DF,$E5,$A7,$2F,$FB,$07,$EA,$FF,$DE,$95,$F9
	.dc.b	$2F,$8A,$7C,$67,$E9,$E7,$35,$F9,$E9,$CD,$7F,$BB,$65,$F2,$7F,$A7
	.dc.b	$FB,$27,$C7,$71,$8C,$3F,$B1,$FA,$CF,$D1,$5F,$E8,$3E,$CB,$F0,$EF
	.dc.b	$F9,$72,$33,$97,$EB,$27,$2F,$F2,$DF,$7B,$F6,$9F,$57,$F2,$BB,$D8
	.dc.b	$7F,$99,$3C,$CF,$F3,$1B,$FA,$C9,$D3,$FE,$68,$FA,$D7,$CA,$F2,$E3
	.dc.b	$FE,$09,$18,$9E,$11,$FC,$1F,$EB,$7F,$C8,$D1,$FB,$6F,$85,$FF,$38
	.dc.b	$3F,$EA,$3C,$BC,$3A,$7F,$A9,$6B,$6F,$F0,$4A,$DB,$FE,$EF,$F5,$7F
	.dc.b	$CC,$F7,$CF,$E8,$BF,$8A,$46,$27,$84,$63,$4F,$E2,$4B,$D9,$C5,$FF
	.dc.b	$08,$8C,$3E,$33,$FD,$0B,$F5,$BF,$B0,$BF,$F5,$C3,$73,$7D,$23,$E8
	.dc.b	$3F,$92,$4E,$E8,$EF,$E9,$3F,$8D,$5F,$BA,$73,$E6,$FF,$25,$9C,$E3
	.dc.b	$E0,$8C,$78,$9D,$1B,$1F,$C7,$BE,$A1,$F5,$BF,$CB,$5F,$C5,$E1,$FC
	.dc.b	$A9,$CC,$EF,$BF,$55,$B6,$C6,$4F,$E5,$3A,$FF,$90,$9B,$EE,$FF,$39
	.dc.b	$8F,$DC,$E1,$FC,$D9,$D3,$63,$46,$FC,$6B,$5F,$2E,$7F,$DC,$1F,$58
	.dc.b	$FA,$DF,$E9,$4D,$9F,$67,$D3,$E2,$64,$7B,$FE,$5A,$DF,$B5,$BC,$3F
	.dc.b	$DA,$76,$FC,$EC,$B1,$E8,$EE,$FD,$BE,$97,$AC,$F6,$76,$D8,$C9,$27
	.dc.b	$DD,$19,$7E,$67,$D6,$FC,$FF,$85,$FA,$0D,$AF,$A8,$D9,$DC,$7D,$79
	.dc.b	$FE,$D1,$F5,$BF,$95,$CF,$F0,$4D,$9F,$D1,$70,$FA,$27,$3F,$1D,$C0
	.dc.b	$FC,$37,$C9,$5A,$FE,$BF,$D1,$FE,$E6,$7E,$76,$4C,$7A,$2C,$15,$B8
	.dc.b	$C5,$7F,$B8,$AB,$75,$AB,$FC,$67,$D0,$4A,$FC,$87,$99,$70,$FC,$74
	.dc.b	$DB,$FA,$53,$6B,$D7,$EB,$7E,$27,$E4,$7F,$14,$D8,$FC,$7B,$70,$FC
	.dc.b	$63,$6F,$E2,$FC,$0F,$C8,$3F,$3D,$F5,$B6,$FD,$FA,$7F,$1E,$AF,$F1
	.dc.b	$57,$D6,$FF,$BC,$67,$37,$FE,$11,$8C,$63,$F0,$01,$F1,$E3,$59,$75
	.dc.b	$8F,$F0,$9C,$F0,$BF,$F1,$B7,$07,$F4,$9C,$3F,$4F,$EA,$FE,$97,$C9
	.dc.b	$FD,$33,$63,$27,$EB,$23,$2F,$EE,$7E,$B7,$F7,$07,$EE,$7C,$1B,$19
	.dc.b	$26,$5F,$BB,$8C,$BE,$91,$F5,$AE,$EF,$E1,$5F,$0B,$FE,$35,$8C,$5F
	.dc.b	$E1,$A3,$13,$A8,$FC,$3C,$EA,$BF,$C6,$3C,$DF,$E2,$49,$CD,$6E,$7C
	.dc.b	$15,$BD,$7F,$CB,$17,$B9,$CD,$6E,$7F,$8F,$56,$EF,$F9,$F5,$EE,$73
	.dc.b	$5B,$9F,$A7,$7E,$C8,$C3,$F5,$AF,$93,$F9,$BB,$E5,$5C,$27,$37,$F9
	.dc.b	$EC,$3D,$F3,$F0,$4F,$59,$6F,$CE,$D9,$F1,$B4,$E9,$87,$8A,$7C,$91
	.dc.b	$97,$C9,$19,$BF,$0E,$1E,$49,$CD,$6D,$E2,$7C,$EC,$3C,$AD,$8F,$2B
	.dc.b	$4F,$29,$70,$F2,$B6,$3C,$A7,$5B,$F0,$D7,$4A,$FA,$AD,$9F,$A2,$DA
	.dc.b	$F9,$4D,$9F,$91,$3E,$27,$8B,$4E,$25,$9E,$2D,$8E,$31,$E8,$78,$B6
	.dc.b	$38,$B4,$E2,$58,$E2,$7B,$8E,$25,$8E,$2D,$8E,$2D,$38,$C7,$C4,$8F
	.dc.b	$8C,$7D,$22,$9F,$4A,$3A,$C6,$DB,$1B,$6B,$B8,$F0,$9B,$2B,$B6,$BB
	.dc.b	$3C,$AE,$DA,$6C,$A6,$E3,$D2,$E5,$AE,$6F,$F1,$8D,$AF,$B1,$18,$BF
	.dc.b	$C0,$DF,$D3,$8C,$61,$F9,$52,$47,$E1,$DB,$7F,$12,$B8,$BF,$4E,$B8
	.dc.b	$D6,$C9,$1B,$AE,$38,$7C,$6B,$E2,$FD,$3B,$E3,$5B,$BE,$33,$BB,$E3
	.dc.b	$1F,$22,$F8,$BF,$1E,$7E,$0C,$62,$BE,$25,$F4,$9E,$55,$D0,$E5,$7D
	.dc.b	$2B,$E6,$DF,$17,$E4,$4F,$C5,$C3,$E4,$91,$C1,$23,$E1,$46,$5B,$1F
	.dc.b	$E3,$1F,$3F,$FC,$60,$7F,$B7,$7D,$B7,$C0,$62,$74,$4A,$FF,$6E,$CE
	.dc.b	$A3,$19,$7F,$4C,$F9,$BF,$F1,$71,$FE,$2F,$F9,$1B,$3F,$E3,$48,$B7
	.dc.b	$FE,$31,$9B,$5F,$E4,$BF,$AC,$7D,$67,$8A,$BF,$95,$9F,$D7,$A3,$39
	.dc.b	$62,$31,$F4,$8F,$4B,$10,$AD,$BB,$FA,$D7,$97,$A8,$46,$1E,$2D,$78
	.dc.b	$46,$E1,$E5,$D2,$4F,$07,$A7,$CF,$7F,$A8,$A3,$FC,$36,$D7,$FB,$03
	.dc.b	$74,$FF,$35,$7D,$67,$F9,$AD,$FF,$36,$46,$78,$3F,$A5,$B6,$BF,$BF
	.dc.b	$B5,$57,$19,$DF,$C5,$7F,$52,$BE,$6D,$8C,$D7,$93,$EB,$30,$F4,$9A
	.dc.b	$B0,$46,$1B,$1C,$4F,$F1,$BF,$D6,$3E,$1B,$5F,$25,$27,$31,$D2,$75
	.dc.b	$18,$21,$F1,$3E,$2B,$53,$23,$39,$FC,$49,$F2,$FD,$63,$9E,$92,$BF
	.dc.b	$C9,$30,$E7,$C4,$7F,$83,$E3,$83,$A9,$E7,$3D,$3D,$0E,$BE,$84,$AE
	.dc.b	$71,$97,$AC,$65,$E9,$1F,$23,$EB,$3F,$DC,$F1,$E2,$FA,$AF,$39,$9F
	.dc.b	$5E,$47,$F9,$7A,$76,$FF,$88,$DB,$A7,$FA,$B7,$3F,$F7,$4C,$EE,$B8
	.dc.b	$BB,$7E,$1F,$C2,$FD,$8F,$AB,$7A,$37,$D1,$97,$FE,$6C,$9C,$B6,$72
	.dc.b	$FF,$8C,$E1,$3C,$E7,$A7,$D6,$4A,$E4,$F7,$F8,$CC,$47,$18,$B9,$FF
	.dc.b	$B1,$70,$57,$E3,$3C,$06,$DE,$7E,$A9,$8F,$FA,$02,$72,$D9,$E0,$47
	.dc.b	$FD,$0B,$F5,$8F,$AC,$C4,$7F,$92,$EB,$8B,$6B,$F3,$75,$B8,$FF,$46
	.dc.b	$FD,$67,$C2,$7F,$CA,$4D,$CF,$FD,$B8,$DC,$7F,$DF,$9F,$62,$EE,$FD
	.dc.b	$2F,$D6,$BD,$FA,$4C,$7A,$F2,$7E,$AB,$EB,$3C,$79,$FF,$B9,$DB,$97
	.dc.b	$F9,$19,$B5,$FE,$7A,$CD,$F7,$1F,$B2,$F8,$A7,$D7,$52,$57,$FB,$4F
	.dc.b	$16,$CF,$37,$C1,$DF,$2B,$FD,$57,$9F,$E8,$1F,$F3,$87,$D9,$3D,$0F
	.dc.b	$DF,$FD,$7F,$EF,$E7,$FD,$62,$D7,$FC,$EC,$DC,$BF,$E0,$05,$3C,$3F
	.dc.b	$37,$C0,$4B,$FF,$5C,$EC,$8F,$E1,$B7,$E6,$F1,$7F,$86,$FD,$7F,$E1
	.dc.b	$2F,$FD,$7E,$D7,$FD,$08,$D9,$71,$FF,$9B,$3E,$C9,$EF,$72,$F5,$66
	.dc.b	$1C,$9F,$E3,$32,$C3,$E5,$7C,$FC,$AF,$2B,$E4,$E4,$FC,$24,$7F,$A4
	.dc.b	$FE,$C3,$AB,$FE,$D2,$6B,$9C,$7F,$92,$FD,$2F,$E1,$63,$FD,$63,$F6
	.dc.b	$0F,$87,$FE,$52,$D8,$FF,$6D,$B5,$FF,$50,$4B,$ED,$FF,$97,$FD,$7F
	.dc.b	$F2,$49,$FF,$52,$37,$83,$E0,$C7,$FA,$FA,$31,$C1,$FE,$8E,$FB,$17
	.dc.b	$C6,$7F,$38,$5F,$05,$49,$7F,$EE,$CF,$B0,$7D,$8F,$F3,$E5,$F0,$F6
	.dc.b	$24,$EE,$35,$EB,$79,$BC,$89,$3C,$E3,$FA,$3F,$D8,$BE,$53,$73,$93
	.dc.b	$FD,$F7,$19,$6C,$70,$23,$FF,$00,$36,$FC,$A3,$FE,$DC,$DF,$E0,$AF
	.dc.b	$B1,$7C,$8F,$C8,$B9,$3B,$E1,$FF,$58,$D6,$A3,$07,$FB,$99,$C3,$F3
	.dc.b	$A7,$8B,$FE,$D3,$8C,$DF,$81,$B5,$D2,$FF,$B2,$C6,$2B,$EA,$7C,$CE
	.dc.b	$38,$3E,$DF,$BB,$27,$C2,$7A,$63,$4F,$BE,$34,$F8,$63,$FB,$8F,$D8
	.dc.b	$B9,$BF,$53,$F5,$7F,$C4,$6B,$F8,$03,$5F,$F7,$54,$ED,$EE,$D7,$33
	.dc.b	$D0,$EB,$F0,$BF,$DF,$B0,$7D,$75,$F8,$89,$7E,$89,$CB,$67,$81,$1F
	.dc.b	$6F,$D8,$FF,$8B,$38,$10,$FB,$05,$FF,$B9,$2B,$53,$98,$FF,$73,$9D
	.dc.b	$9C,$1C,$9C,$4F,$28,$EB,$3C,$A3,$10,$8C,$4F,$4B,$FF,$79,$D6,$A3
	.dc.b	$99,$E3,$C5,$87,$A2,$79,$DF,$2C,$3C,$58,$72,$C3,$A4,$6D,$E5,$76
	.dc.b	$3C,$B8,$74,$7D,$67,$C2,$AE,$4F,$43,$81,$E9,$8E,$18,$78,$9B,$7E
	.dc.b	$09,$E6,$FA,$9E,$F7,$CB,$0F,$9B,$1F,$E6,$86,$BE,$63,$FC,$D0,$E1
	.dc.b	$F9,$7F,$10,$FB,$1F,$F3,$4C,$FF,$97,$9B,$19,$9E,$2D,$7F,$CD,$FF
	.dc.b	$47,$FE,$6B,$4D,$23,$FE,$6F,$7D,$F7,$F9,$DE,$4F,$83,$3F,$88,$9C
	.dc.b	$ED,$FC,$83,$4F,$BB,$39,$3E,$5D,$F1,$7E,$69,$98,$FC,$1C,$65,$E3
	.dc.b	$1E,$43,$8C,$F5,$CF,$FD,$15,$7E,$0C,$3C,$75,$BB,$E7,$7B,$8F,$F4
	.dc.b	$9D,$6D,$F0,$9B,$BF,$F3,$84,$EE,$3F,$46,$64,$FE,$38,$E7,$0F,$51
	.dc.b	$FE,$81,$7F,$9F,$BF,$C8,$23,$D9,$3D,$30,$FE,$17,$3C,$1C,$4F,$E6
	.dc.b	$9F,$5B,$EE,$7F,$88,$FD,$87,$EF,$9B,$B7,$67,$CB,$1F,$B8,$72,$49
	.dc.b	$E8,$7E,$F2,$39,$38,$0F,$17,$0E,$AB,$F9,$CB,$FD,$16,$7C,$7C,$DE
	.dc.b	$1A,$FD,$8F,$D8,$FF,$A2,$67,$24,$8F,$0D,$A5,$F3,$1C,$67,$C3,$7C
	.dc.b	$EB,$53,$9C,$1E,$45,$35,$CB,$FE,$21,$CB,$E2,$C6,$5A,$7F,$2E,$F4
	.dc.b	$B1,$97,$B2,$78,$FC,$66,$32,$FA,$F8,$7F,$A9,$B0,$FE,$7F,$F6,$3F
	.dc.b	$E4,$57,$FA,$0F,$C8,$FF,$74,$73,$EE,$FA,$63,$19,$7C,$B6,$E1,$FD
	.dc.b	$81,$B5,$F8,$AB,$E9,$5B,$EB,$F8,$5F,$62,$F8,$CA,$75,$8B,$7F,$D4
	.dc.b	$A5,$F3,$BC,$0F,$95,$F6,$3F,$E8,$89,$D4,$7C,$DF,$B1,$FF,$1F,$47
	.dc.b	$F5,$A6,$E3,$F8,$1B,$E7,$39,$F6,$78,$72,$F1,$F7,$AC,$F1,$7E,$B7
	.dc.b	$95,$FE,$5C,$8F,$F2,$D3,$70,$FB,$EB,$C4,$DF,$BF,$C3,$A6,$5F,$0F
	.dc.b	$87,$B8,$7E,$1B,$EC,$1F,$49,$F8,$35,$F0,$B6,$30,$FE,$1F,$EC,$1F
	.dc.b	$5D,$F8,$9F,$90,$FF,$27,$E5,$FE,$6D,$9C,$BF,$E6,$F8,$CE,$3F,$94
	.dc.b	$5F,$39,$C7,$B3,$93,$6B,$FC,$ED,$F5,$EF,$A8,$FC,$B3,$63,$3F,$1D
	.dc.b	$89,$FC,$D7,$D7,$7F,$8F,$27,$FC,$FC,$DC,$BF,$46,$FF,$88,$A3,$3E
	.dc.b	$DF,$E8,$7E,$B3,$FC,$7B,$1F,$E8,$56,$D7,$EA,$9F,$F3,$24,$67,$DA
	.dc.b	$E1,$CF,$C7,$EB,$2C,$F1,$7F,$59,$F1,$3F,$C3,$F0,$FE,$CE,$BE,$B3
	.dc.b	$3E,$7E,$9D,$FC,$77,$0F,$76,$E3,$FE,$97,$BF,$F7,$A3,$FE,$99,$C5
	.dc.b	$F3,$BF,$B9,$6C,$67,$C0,$C4,$FE,$F7,$EA,$FE,$EC,$FF,$A8,$5F,$BA
	.dc.b	$DE,$BD,$75,$ED,$B1,$93,$84,$65,$A6,$76,$FE,$1F,$5D,$1F,$E2,$3F
	.dc.b	$4E,$F6,$BF,$87,$36,$BF,$D6,$4B,$D8,$CC,$C9,$CB,$C2,$31,$5F,$CA
	.dc.b	$7E,$AD,$F5,$9F,$C6,$DB,$7E,$DE,$CC,$4C,$8F,$81,$DB,$6F,$CE,$F9
	.dc.b	$3C,$D9,$3E,$4F,$A0,$B5,$EA,$9B,$E8,$FF,$4E,$D5,$58,$BF,$F3,$84
	.dc.b	$7F,$0B,$7D,$B5,$F0,$27,$57,$FE,$28,$DB,$F2,$98,$7F,$58,$7F,$29
	.dc.b	$1F,$2B,$B3,$EC,$7C,$0F,$D9,$6E,$3F,$65,$B3,$FD,$2F,$EA,$FF,$D5
	.dc.b	$53,$FE,$68,$6C,$7D,$2D,$7F,$A0,$46,$31,$FF,$42,$B6,$BF,$D1,$1F
	.dc.b	$51,$FE,$D1,$79,$7C,$5F,$F4,$58,$C6,$1F,$7B,$E7,$FF,$18,$7E,$37
	.dc.b	$C6,$B7,$F8,$D8,$B3,$F8,$49,$B8,$FB,$E6,$D7,$FA,$5F,$55,$62,$FF
	.dc.b	$C9,$BC,$1F,$C9,$9F,$C5,$FC,$6F,$E2,$BD,$3F,$C5,$B7,$0F,$C5,$B6
	.dc.b	$77,$1F,$0F,$CD,$F8,$27,$E8,$FE,$B2,$C6,$4F,$F4,$FE,$8F,$AC,$FF
	.dc.b	$73,$C6,$6F,$FB,$5C,$61,$FC,$EF,$D6,$7E,$73,$C6,$FC,$EB,$6B,$F3
	.dc.b	$AD,$9D,$C7,$E8,$27,$6D,$36,$58,$FD,$5F,$D6,$7E,$E4,$FD,$D7,$B3
	.dc.b	$67,$F7,$51,$67,$F6,$33,$71,$FF,$60,$FD,$67,$F2,$EB,$FF,$55,$B6
	.dc.b	$BF,$6B,$E2,$7F,$80,$E3,$15,$C3,$B7,$FC,$55,$5F,$E9,$49,$E0,$FF
	.dc.b	$8D,$03,$87,$FC,$6E,$D7,$F3,$DF,$2F,$FC,$44,$FF,$07,$7A,$EF,$77
	.dc.b	$19,$7F,$13,$9B,$5F,$E1,$8A,$D4,$E4,$91,$FC,$3A,$79,$D7,$0C,$7F
	.dc.b	$55,$16,$BF,$DC,$57,$CA,$F8,$47,$2A,$DB,$F8,$09,$D4,$62,$76,$47
	.dc.b	$FC,$9B,$1A,$9F,$CC,$57,$7A,$F7,$C7,$3B,$C5,$B9,$97,$8C,$39,$61
	.dc.b	$C0,$ED,$87,$43,$91,$E7,$30,$E8,$FF,$58,$C7,$2B,$ED,$5E,$0B,$EF
	.dc.b	$7C,$AF,$85,$6A,$3A,$47,$2C,$38,$5F,$95,$B3,$AA,$61,$DF,$0E,$B8
	.dc.b	$74,$6C,$FC,$6B,$E3,$3E,$0A,$E1,$86,$2B,$58,$75,$C3,$8B,$EE,$8C
	.dc.b	$DF,$5A,$F2,$4E,$DB,$19,$8F,$35,$F6,$74,$7C,$1A,$FB,$84,$7B,$1F
	.dc.b	$86,$C3,$C8,$E2,$BE,$23,$C9,$F5,$46,$9F,$B1,$3E,$63,$A3,$89,$E5
	.dc.b	$1F,$22,$78,$3F,$46,$7F,$3D,$18,$C3,$E0,$C7,$80,$F6,$1E,$0A,$F1
	.dc.b	$57,$63,$A3,$83,$BD,$FC,$19,$E3,$3E,$59,$F1,$D7,$68,$F1,$19,$9F
	.dc.b	$41,$1C,$CF,$C5,$3B,$46,$6B,$B4,$F8,$EB,$A4,$62,$BB,$E1,$FB,$67
	.dc.b	$8D,$74,$BF,$73,$C6,$7E,$DB,$67,$ED,$E1,$F1,$67,$95,$7B,$27,$DF
	.dc.b	$39,$7E,$F1,$E0,$38,$9E,$0C,$3C,$B1,$B9,$FA,$33,$9A,$EB,$7E,$19
	.dc.b	$FA,$07,$73,$E0,$57,$EF,$39,$5E,$57,$F8,$70,$FF,$0E,$F9,$57,$1F
	.dc.b	$F0,$EC,$5B,$FF,$11,$CD,$C1,$5F,$ED,$EF,$AA,$FF,$22,$3F,$CB,$68
	.dc.b	$23,$2E,$5A,$FF,$33,$6B,$FE,$E8,$FA,$B7,$D6,$7F,$2F,$8E,$91,$8C
	.dc.b	$FB,$42,$1F,$48,$F8,$EC,$C6,$E7,$C9,$E6,$FF,$A0,$1F,$2E,$7E,$A7
	.dc.b	$CF,$19,$7F,$DB,$CD,$8D,$43,$EB,$2C,$62,$75,$A7,$FC,$00,$6D,$74
	.dc.b	$F6,$BE,$1B,$FE,$E9,$BE,$35,$B9,$CD,$F3,$6C,$66,$BC,$EF,$2A,$E3
	.dc.b	$1E,$8F,$85,$DB,$4D,$94,$F5,$FC,$A6,$9F,$F3,$8C,$5C,$4D,$31,$3B
	.dc.b	$FA,$C7,$CF,$6F,$E3,$A6,$1D,$84,$E9,$C3,$56,$9F,$2A,$30,$DA,$F1
	.dc.b	$7C,$27,$97,$EA,$FF,$2B,$7D,$5A,$F2,$D7,$E5,$C6,$2F,$EA,$AE,$5E
	.dc.b	$5C,$D2,$F9,$3D,$63,$11,$CC,$FC,$BC,$F3,$BE,$B5,$B9,$F1,$7A,$1E
	.dc.b	$5F,$42,$46,$08,$73,$17,$FE,$6E,$7D,$51,$ED,$9F,$A5,$F5,$9F,$ED
	.dc.b	$69,$F7,$7C,$07,$C1,$33,$EB,$C8,$F9,$D3,$B7,$B3,$74,$F8,$3F,$59
	.dc.b	$F7,$7E,$AD,$E7,$DF,$3E,$61,$F5,$E7,$2D,$9C,$BE,$98,$FF,$2D,$B9
	.dc.b	$21,$99,$DD,$FE,$92,$79,$CF,$1F,$AC,$F1,$11,$F0,$7D,$65,$74,$78
	.dc.b	$E3,$D1,$1F,$95,$EE,$6B,$EB,$E2,$97,$F8,$F9,$CB,$67,$81,$1F,$CA
	.dc.b	$7D,$63,$E2,$7F,$40,$FE,$7A,$33,$D5,$83,$3D,$72,$32,$43,$EE,$4F
	.dc.b	$16,$C7,$17,$5A,$FF,$06,$9D,$3F,$C3,$4F,$E4,$5F,$71,$FB,$67,$F9
	.dc.b	$27,$D5,$DF,$14,$D9,$C6,$99,$F1,$66,$3F,$E7,$EB,$FE,$3E,$FF,$A0
	.dc.b	$3E,$DF,$F8,$7E,$26,$98,$FD,$9B,$F8,$3C,$7F,$33,$E2,$F0,$12,$BA
	.dc.b	$71,$78,$FE,$0A,$47,$77,$CD,$D3,$E9,$C1,$AD,$DA,$5F,$FA,$4A,$72
	.dc.b	$F9,$FD,$DF,$77,$BD,$23,$07,$AF,$A7,$E1,$74,$C8,$C1,$CF,$8B,$C0
	.dc.b	$47,$B7,$17,$C9,$F4,$A4,$7C,$57,$E6,$79,$BF,$3F,$C4,$FE,$95,$3F
	.dc.b	$BF,$E3,$FE,$37,$E6,$64,$D7,$77,$B7,$E8,$BD,$DF,$49,$2F,$D9,$1F
	.dc.b	$CA,$7E,$8B,$F8,$CF,$81,$FC,$F1,$B3,$FD,$81,$AB,$E9,$1C,$7F,$CC
	.dc.b	$56,$D4,$93,$FC,$AF,$07,$C5,$FF,$37,$5E,$CD,$89,$3B,$8D,$79,$5E
	.dc.b	$5F,$0A,$4F,$38,$FE,$81,$F4,$8F,$89,$DC,$E4,$FE,$75,$19,$6C,$70
	.dc.b	$23,$FC,$FF,$57,$CD,$FF,$4B,$FA,$47,$5F,$C5,$B9,$7F,$D1,$67,$2D
	.dc.b	$9C,$BF,$C9,$EF,$C0,$FF,$29,$9D,$B6,$7B,$11,$F1,$7C,$B7,$87,$7F
	.dc.b	$A9,$E9,$49,$F0,$46,$5F,$44,$65,$ED,$1F,$D8,$FE,$91,$F2,$1F,$B5
	.dc.b	$F9,$C7,$05,$35,$12,$FF,$B0,$57,$06,$C7,$3F,$8F,$FE,$DD,$8B,$E8
	.dc.b	$2F,$B9,$2F,$FB,$44,$E5,$B3,$C0,$8F,$F6,$DF,$A4,$5F,$F6,$A8,$7D
	.dc.b	$1A,$FF,$C1,$B5,$AE,$FF,$F1,$8C,$E6,$BF,$C6,$75,$DD,$E2,$FF,$85
	.dc.b	$0C,$46,$E1,$5C,$DD,$9F,$E3,$C7,$11,$B9,$E2,$73,$C3,$FC,$3F,$1C
	.dc.b	$E7,$73,$E6,$C3,$D5,$3C,$6B,$73,$C6,$FD,$2E,$F0,$B1,$E1,$BE,$33
	.dc.b	$B8,$F2,$9F,$0E,$B9,$BB,$AE,$2F,$B2,$B8,$BF,$12,$76,$F1,$37,$87
	.dc.b	$A1,$E4,$7F,$97,$5F,$88,$FF,$8F,$0E,$95,$FE,$54,$8C,$C7,$D4,$F9
	.dc.b	$27,$D2,$65,$B1,$CD,$AF,$39,$46,$23,$FC,$D9,$F4,$9F,$E6,$07,$FC
	.dc.b	$DA,$D4,$D2,$3F,$12,$FF,$3B,$DF,$F0,$A3,$89,$9D,$BF,$CC,$5A,$FE
	.dc.b	$06,$32,$FE,$0E,$B8,$C7,$CE,$33,$AF,$F9,$62,$B8,$BE,$68,$E9,$7D
	.dc.b	$DF,$29,$83,$91,$8B,$ED,$5C,$A3,$E6,$61,$F6,$30,$F0,$D7,$96,$78
	.dc.b	$61,$E1,$C3,$B5,$EF,$0E,$55,$FE,$50,$7F,$B9,$4E,$E3,$FC,$6C,$7A
	.dc.b	$8F,$F4,$1C,$69,$B1,$FE,$5D,$C3,$E0,$9F,$E8,$28,$E0,$75,$8F,$F3
	.dc.b	$17,$D1,$FE,$FD,$BA,$F3,$94,$FF,$A9,$27,$24,$8F,$F2,$44,$71,$69
	.dc.b	$C5,$C5,$7F,$A4,$9F,$E7,$53,$E0,$E8,$F4,$4B,$E9,$3F,$CE,$F3,$92
	.dc.b	$5F,$F6,$18,$7C,$77,$5B,$E3,$18,$AF,$5F,$D2,$78,$4E,$BC,$DF,$D0
	.dc.b	$23,$24,$C3,$D3,$3A,$EC,$E1,$39,$CB,$FC,$45,$7F,$E8,$E8,$C1,$FE
	.dc.b	$B2,$71,$7F,$E6,$9F,$A4,$7C,$BF,$C9,$F2,$3F,$B3,$DF,$FA,$3E,$72
	.dc.b	$FF,$A4,$63,$35,$FE,$71,$6E,$1F,$EB,$86,$CF,$FA,$CA,$B8,$D6,$FA
	.dc.b	$FE,$B7,$D1,$3D,$FA,$78,$62,$DF,$FA,$7E,$5F,$29,$C0,$FB,$DF,$49
	.dc.b	$FE,$70,$9D,$47,$E0,$7E,$91,$F0,$D4,$E1,$3F,$EA,$98,$CB,$78,$F8
	.dc.b	$7C,$77,$16,$C7,$FA,$F3,$7F,$F0,$1F,$D2,$7E,$03,$CA,$37,$C5,$C3
	.dc.b	$9E,$89,$7C,$6B,$FC,$C8,$D8,$FC,$4F,$D2,$3D,$CB,$EE,$61,$B5,$E1
	.dc.b	$FA,$47,$C7,$7E,$43,$DA,$4F,$E4,$A7,$2F,$9E,$32,$DF,$93,$E1,$D7
	.dc.b	$C5,$B1,$ED,$FA,$47,$D2,$7E,$71,$B1,$9F,$89,$C4,$FE,$83,$E9,$3F
	.dc.b	$C5,$B3,$F1,$5B,$DB,$FE,$3F,$7F,$C2,$B1,$9F,$67,$F4,$FF,$49,$FE
	.dc.b	$2E,$8F,$98,$DA,$FF,$23,$2F,$A8,$DF,$B3,$C3,$97,$8F,$CC,$DC,$38
	.dc.b	$BF,$B2,$FA,$4F,$E0,$F0,$C7,$FD,$C3,$0F,$A5,$D7,$9B,$B7,$87,$CA
	.dc.b	$71,$F7,$6E,$1F,$7E,$FF,$DB,$8F,$E0,$71,$7C,$B7,$F0,$06,$C6,$7C
	.dc.b	$0C,$4F,$F0,$4F,$9A,$FB,$93,$F8,$95,$D9,$9E,$CD,$36,$74,$DD,$BF
	.dc.b	$CD,$15,$C2,$79,$57,$06,$9C,$0F,$CB,$7C,$D7,$F8,$61,$BE,$CF,$F1
	.dc.b	$68,$FE,$BB,$18,$25,$FF,$9D,$63,$83,$7D,$7E,$15,$FC,$AF,$E9,$5F
	.dc.b	$4B,$4F,$77,$6D,$C7,$E8,$78,$DF,$94,$C7,$E9,$5F,$09,$F8,$5D,$FB
	.dc.b	$9F,$6B,$E5,$F3,$A4,$FA,$BE,$62,$D7,$96,$6E,$1F,$EB,$DD,$55,$8B
	.dc.b	$F0,$C7,$F0,$B7,$E1,$D7,$C1,$9D,$5A,$DB,$FE,$61,$FF,$64,$3F,$98
	.dc.b	$8F,$A3,$DB,$F6,$FE,$03,$EE,$37,$1F,$B8,$D9,$E9,$1F,$91,$5F,$4A
	.dc.b	$FA,$5F,$EA,$8D,$9F,$AF,$C1,$89,$91,$FC,$3F,$E9,$5F,$55,$8C,$3F
	.dc.b	$D5,$51,$89,$FC,$67,$C8,$7E,$3C,$FC,$87,$91,$63,$24,$AF,$C5,$C6
	.dc.b	$5F,$C6,$CD,$DB,$E0,$CD,$AF,$97,$E9,$7E,$27,$E9,$BF,$14,$D8,$FC
	.dc.b	$7B,$70,$FC,$53,$6F,$F0,$51,$F0,$BC,$EF,$EF,$4F,$ED,$76,$FE,$C0
	.dc.b	$FF,$8E,$CD,$5D,$69,$98,$DD,$7F,$BF,$7E,$95,$8F,$CC,$C3,$FB,$9C
	.dc.b	$63,$0F,$D1,$FD,$2F,$E8,$7C,$6F,$D1,$36,$32,$7E,$A2,$32,$D3,$39
	.dc.b	$7E,$96,$32,$D8,$FD,$E7,$D2,$FE,$FC,$FE,$01,$EC,$D8,$C9,$27,$F7
	.dc.b	$71,$97,$F0,$E7,$F1,$7F,$A5,$76,$FF,$83,$38,$BF,$C0,$13,$A1,$3A
	.dc.b	$BF,$C3,$7B,$DF,$E1,$F8,$C4,$FE,$1A,$31,$7C,$DF,$E2,$9F,$27,$2F
	.dc.b	$FA,$E7,$B0,$F7,$B1,$87,$F1,$0A,$BC,$5F,$BB,$15,$BB,$FE,$39,$3A
	.dc.b	$7F,$80,$F6,$23,$FC,$4B,$3B,$BE,$95,$C0,$8F,$C2,$BE,$B8,$73,$8F
	.dc.b	$DB,$61,$CE,$B7,$5E,$3B,$E8,$FE,$E3,$0E,$0F,$EF,$1F,$0E,$18,$BE
	.dc.b	$58,$78,$8F,$05,$F2,$C3,$D3,$7B,$38,$BE,$03,$25,$9F,$E1,$53,$C5
	.dc.b	$F3,$CF,$19,$F9,$B7,$67,$C1,$87,$A6,$30,$FA,$67,$67,$57,$D4,$DB
	.dc.b	$ED,$18,$AE,$73,$E4,$7D,$75,$D2,$31,$3D,$DE,$98,$7C,$28,$F0,$57
	.dc.b	$09,$CE,$1F,$2B,$0E,$F7,$B9,$E6,$4A,$F8,$38,$77,$7E,$0C,$7A,$9F
	.dc.b	$29,$82,$E1,$F7,$A3,$B9,$ED,$9E,$51,$ED,$3D,$13,$C2,$BC,$4F,$38
	.dc.b	$C9,$EA,$25,$79,$5E,$D1,$C7,$0F,$65,$F6,$BE,$75,$A3,$8D,$75,$BE
	.dc.b	$58,$7C,$2C,$3C,$AD,$8F,$4C,$FC,$93,$B3,$EB,$33,$5D,$0E,$CF,$96
	.dc.b	$FC,$27,$8A,$BA,$57,$23,$A5,$71,$9D,$DF,$E5,$27,$53,$D6,$B8,$BB
	.dc.b	$9D,$36,$FC,$95,$C1,$F6,$E1,$F0,$EB,$71,$E5,$7E,$3B,$EA,$36,$59
	.dc.b	$F8,$C7,$E0,$9C,$57,$D9,$7F,$C0,$77,$E4,$7A,$1F,$85,$8F,$24,$FE
	.dc.b	$12,$74,$F4,$8E,$4F,$53,$D0,$66,$37,$1A,$23,$8C,$3E,$3D,$77,$3C
	.dc.b	$35,$DF,$0F,$E0,$FD,$EF,$2F,$FC,$4C,$7F,$89,$FC,$BB,$7F,$E2,$C8
	.dc.b	$B7,$FE,$27,$9B,$97,$F5,$EF,$A5,$7C,$AF,$F2,$BA,$FE,$B9,$18,$CB
	.dc.b	$FA,$E2,$F9,$67,$C6,$62,$F8,$5F,$F8,$24,$7C,$EB,$C9,$FE,$C9,$18
	.dc.b	$6C,$66,$3F,$B4,$35,$FE,$D9,$E6,$FF,$A2,$1F,$F1,$AE,$DF,$E1,$58
	.dc.b	$C4,$7F,$4D,$6E,$9F,$DE,$3E,$97,$FB,$CB,$FD,$EA,$33,$5C,$7D,$CF
	.dc.b	$F0,$F4,$E5,$B1,$FE,$21,$AE,$2E,$EF,$C2,$7F,$9D,$6F,$9B,$6B,$8B
	.dc.b	$FE,$79,$AE,$2E,$FE,$83,$2F,$F9,$C9,$B4,$6F,$C8,$E1,$93,$E0,$38
	.dc.b	$16,$53,$D3,$E9,$5F,$35,$F3,$9B,$5F,$E1,$36,$BF,$E1,$76,$BF,$5D
	.dc.b	$CE,$7F,$E2,$06,$E7,$FE,$43,$6C,$FC,$1F,$85,$F5,$FD,$2F,$E4,$63
	.dc.b	$CC,$D7,$FC,$00,$DB,$FF,$15,$AE,$5E,$3C,$D2,$B8,$BD,$23,$11,$CA
	.dc.b	$2D,$72,$9F,$0F,$A1,$E2,$F4,$25,$F8,$A3,$2F,$92,$32,$F9,$23,$E9
	.dc.b	$FD,$2F,$FA,$72,$3D,$3F,$03,$EC,$33,$EB,$C8,$FA,$53,$B7,$FC,$38
	.dc.b	$DC,$7D,$BF,$39,$F7,$BE,$95,$E8,$5F,$42,$61,$F6,$E7,$2D,$9C,$BF
	.dc.b	$E2,$78,$4F,$39,$93,$C6,$37,$7F,$E5,$6A,$E7,$3C,$7E,$8B,$C4,$47
	.dc.b	$C1,$F4,$55,$D1,$E3,$8F,$44,$7E,$5B,$BD,$AF,$AF,$96,$5F,$E4,$E7
	.dc.b	$2D,$9E,$04,$7F,$31,$F4,$8F,$A5,$FD,$0B,$F9,$F8,$CF,$4E,$0C,$F5
	.dc.b	$C8,$C9,$0F,$F2,$54,$F1,$6C,$71,$75,$AF,$F9,$E2,$74,$FF,$90,$7B
	.dc.b	$9F,$49,$FB,$68,$FE,$2B,$F4,$96,$4D,$9C,$69,$9F,$0A,$63,$FC,$0B
	.dc.b	$0F,$E3,$B3,$CF,$E9,$7F,$81,$C1,$A6,$3E,$97,$F8,$1C,$7F,$24,$E2
	.dc.b	$F0,$12,$FA,$66,$E2,$F2,$7C,$0C,$9C,$BE,$7E,$9F,$56,$3E,$C9,$F2
	.dc.b	$3E,$A6,$ED,$AD,$BD,$DE,$EF,$BF,$DE,$95,$E0,$8F,$85,$D3,$F0,$FA
	.dc.b	$64,$F9,$23,$BF,$0C,$8E,$EF,$8B,$8B,$E5,$7A,$72,$7D,$6F,$CD,$F3
	.dc.b	$FE,$87,$8B,$FD,$1B,$93,$FC,$E8,$DE,$BF,$F7,$8B,$6F,$FD,$BC,$BB
	.dc.b	$7D,$FF,$47,$F0,$3E,$96,$57,$C0,$8F,$E5,$BF,$47,$FC,$5A,$0D,$31
	.dc.b	$F4,$AE,$1F,$E6,$6B,$5A,$4A,$FE,$61,$6F,$8B,$FE,$72,$BD,$AB,$12
	.dc.b	$77,$1A,$F1,$FC,$7E,$0C,$9E,$71,$FD,$0B,$E9,$5F,$11,$B9,$C9,$FC
	.dc.b	$FE,$32,$D8,$E0,$47,$FA,$26,$AF,$63,$FA,$6F,$D2,$BA,$BE,$1D,$CB
	.dc.b	$FE,$95,$39,$6C,$E5,$FF,$5B,$5F,$81,$FF,$5C,$4E,$DB,$3D,$88,$F8
	.dc.b	$BE,$5B,$C3,$BF,$D2,$FA,$19,$3E,$08,$CB,$E8,$8C,$BD,$A3,$FB,$2F
	.dc.b	$D2,$BE,$43,$F6,$DF,$3B,$FD,$7A,$BF,$60,$D7,$FD,$99,$3B,$7F,$B3
	.dc.b	$35,$CC,$F4,$3A,$FB,$1F,$DC,$B0,$7D,$0D,$F6,$A5,$FF,$6F,$9C,$B6
	.dc.b	$78,$11,$FE,$E9,$F4,$9F,$E1,$DE,$E4,$3E,$91,$78,$85,$6D,$D4,$64
	.dc.b	$EE,$E0,$E6,$E2,$79,$C7,$69,$E5,$18,$8E,$B8,$AF,$C7,$39,$BE,$47
	.dc.b	$7C,$3C,$91,$C7,$0E,$73,$E7,$6E,$1E,$2A,$DC,$F1,$8E,$AE,$F0,$F0
	.dc.b	$47,$A4,$F1,$E1,$A7,$91,$4F,$8B,$5D,$1F,$64,$F2,$C3,$C8,$DB,$F0
	.dc.b	$4F,$57,$D6,$F0,$BE,$98,$7C,$D8,$FF,$29,$35,$F3,$47,$F9,$55,$E0
	.dc.b	$58,$7C,$8B,$E9,$75,$3F,$62,$30,$ED,$AF,$66,$BD,$7E,$97,$FC,$BE
	.dc.b	$9A,$47,$FC,$C6,$FC,$3B,$FD,$07,$85,$F1,$67,$F1,$73,$9D,$93,$99
	.dc.b	$E8,$7C,$FA,$E2,$FD,$13,$89,$99,$E2,$E4,$E3,$A7,$E2,$67,$8B,$F8
	.dc.b	$B8,$E2,$E6,$78,$B9,$8E,$39,$FF,$9A,$A7,$2F,$C1,$8F,$05,$F9,$F0
	.dc.b	$F5,$36,$32,$4C,$3F,$CE,$F7,$E6,$7C,$F1,$FA,$9A,$E2,$FE,$76,$BC
	.dc.b	$07,$F9,$96,$7E,$09,$C2,$3D,$37,$A7,$81,$FB,$6F,$A5,$FE,$03,$5F
	.dc.b	$E6,$D6,$C6,$83,$FB,$C8,$D6,$DE,$57,$F7,$91,$C2,$FC,$ED,$8C,$D6
	.dc.b	$CF,$05,$57,$AF,$3D,$AF,$E8,$75,$FE,$0D,$FA,$5F,$F3,$DB,$E5,$9D
	.dc.b	$1C,$F0,$4B,$E3,$73,$5C,$5E,$9F,$4B,$23,$F9,$4F,$5D,$CB,$DD,$92
	.dc.b	$B1,$8F,$D1,$E6,$FE,$99,$87,$C5,$9C,$BF,$C7,$A5,$3F,$E9,$98,$D1
	.dc.b	$FC,$DB,$D6,$F7,$7D,$2B,$E5,$BE,$07,$C3,$FF,$6F,$AF,$F4,$E4,$E5
	.dc.b	$FF,$4F,$46,$73,$FA,$4D,$C7,$F0,$6D,$9F,$E5,$95,$C6,$B7,$D7,$F3
	.dc.b	$7E,$95,$EF,$D3,$C3,$16,$FF,$D5,$92,$F9,$4E,$07,$D5,$FA,$5F,$F3
	.dc.b	$CC,$EA,$3E,$C7,$D2,$BE,$1A,$9C,$27,$FD,$6D,$19,$6F,$C7,$70,$38
	.dc.b	$F3,$D8,$FF,$25,$79,$4F,$89,$FB,$8F,$E0,$B6,$FC,$BF,$0F,$F8,$3E
	.dc.b	$73,$7A,$F8,$BF,$C1,$7D,$27,$F0,$4C,$79,$C6,$1C,$FC,$84,$7F,$CB
	.dc.b	$13,$F1,$F0,$FD,$72,$9F,$C6,$7B,$5F,$E6,$6A,$FF,$2F,$37,$2F,$1B
	.dc.b	$FE,$66,$8C,$FB,$7C,$3D,$49,$1C,$3C,$CB,$7E,$AF,$A3,$FF,$15,$4F
	.dc.b	$F9,$A5,$7D,$07,$E6,$DC,$CE,$E3,$F3,$DF,$45,$F4,$A3,$E1,$BF,$EA
	.dc.b	$4A,$FF,$39,$39,$D7,$F8,$43,$99,$F9,$3F,$45,$FC,$12,$BF,$30,$BE
	.dc.b	$8B,$F5,$4E,$67,$FB,$43,$99,$FF,$3F,$2F,$8F,$D9,$C3,$D6,$B3,$FE
	.dc.b	$83,$FA,$27,$A3,$FB,$56,$C6,$7C,$FC,$4E,$7E,$EA,$4F,$FA,$2F,$E8
	.dc.b	$FF,$87,$67,$F8,$0C,$15,$FB,$F9,$7D,$05,$F2,$F0,$DC,$3F,$D2,$9F
	.dc.b	$31,$E0,$C7,$F1,$2D,$E1,$FE,$58,$FF,$5E,$9D,$E7,$FC,$2E,$35,$1F
	.dc.b	$BD,$6D,$7F,$82,$DE,$0F,$49,$E0,$49,$F0,$E9,$67,$F9,$57,$D1,$BE
	.dc.b	$8F,$F9,$03,$6B,$FD,$55,$F3,$7F,$91,$7F,$86,$1F,$E2,$78,$D5,$72
	.dc.b	$AF,$E4,$15,$F9,$9A,$D4,$63,$CB,$F5,$79,$32,$7E,$1F,$CB,$DC,$3E
	.dc.b	$1C,$DD,$7E,$1C,$DA,$FE,$39,$8F,$FA,$15,$F2,$F5,$78,$3A,$FC,$2D
	.dc.b	$8C,$9E,$A8,$CB,$4C,$E3,$FB,$AA,$E8,$DA,$FE,$85,$F4,$4F,$A3,$A7
	.dc.b	$D3,$D6,$E1,$F1,$2F,$FC,$9B,$5B,$8F,$F6,$97,$CF,$FE,$11,$D4,$A7
	.dc.b	$4F,$F0,$A9,$D3,$E3,$F8,$DF,$96,$7C,$CF,$06,$DF,$D4,$8B,$3E,$99
	.dc.b	$BB,$7E,$2B,$CB,$FC,$47,$D1,$FE,$25,$B8,$7E,$49,$B7,$F9,$26,$D7
	.dc.b	$19,$FC,$BD,$71,$69,$EA,$CB,$FA,$E3,$7D,$6F,$C0,$FD,$3F,$C8,$64
	.dc.b	$7F,$BC,$FE,$91,$78,$CB,$F9,$4C,$62,$7F,$4B,$F4,$9F,$9F,$F3,$7F
	.dc.b	$40,$D8,$C9,$FA,$A8,$CB,$4F,$CB,$FD,$27,$E4,$8F,$C9,$FA,$96,$32
	.dc.b	$4C,$3F,$01,$19,$7F,$05,$37,$5F,$F0,$F7,$2F,$F5,$B9,$FF,$3C,$29
	.dc.b	$FF,$03,$7D,$27,$E9,$67,$FC,$28,$DA,$FD,$DF,$83,$AF,$32,$65,$F8
	.dc.b	$4F,$0A,$E1,$F8,$17,$F5,$3F,$23,$27,$F1,$98,$3D,$27,$B1,$9C,$3F
	.dc.b	$31,$76,$BF,$C6,$95,$AB,$FD,$CC,$FF,$09,$F2,$64,$7F,$12,$F2,$6E
	.dc.b	$1F,$C4,$A6,$E1,$FC,$4A,$6D,$7F,$0F,$9B,$5F,$C3,$E6,$D3,$C6,$FE
	.dc.b	$65,$7F,$C7,$AF,$F8,$C5,$6A,$73,$7D,$EB,$C4,$6C,$8E,$4F,$E4,$B3
	.dc.b	$C6,$B9,$36,$77,$7E,$76,$9E,$72,$D7,$9D,$B1,$E7,$AE,$71,$8B,$F2
	.dc.b	$CF,$F3,$29,$F4,$CE,$5D,$D7,$96,$75,$3C,$6F,$C9,$39,$8F,$E6,$D1
	.dc.b	$89,$DC,$FB,$27,$A1,$C4,$AF,$13,$D0,$58,$F4,$46,$E3,$17,$E8,$8F
	.dc.b	$81,$1D,$4C,$DF,$A2,$BC,$24,$9D,$BF,$26,$77,$1D,$63,$51,$FD,$36
	.dc.b	$79,$4F,$02,$46,$5F,$B6,$F2,$3C,$B7,$F0,$63,$75,$F8,$78,$C4,$F1
	.dc.b	$3D,$11,$E0,$3D,$EF,$29,$F8,$B1,$9B,$57,$28,$D9,$EB,$35,$7E,$8B
	.dc.b	$ED,$18,$BF,$3B,$63,$B1,$E9,$7B,$57,$5B,$F9,$D3,$F5,$27,$81,$AA
	.dc.b	$F2,$9E,$E9,$F2,$B8,$9F,$0B,$F8,$43,$13,$E8,$71,$1E,$17,$95,$78
	.dc.b	$6F,$F2,$73,$93,$99,$E3,$9F,$04,$7B,$EB,$A4,$E9,$A7,$D1,$9F,$0D
	.dc.b	$FB,$23,$13,$D5,$DC,$7A,$A3,$0E,$CC,$47,$C2,$9F,$39,$D2,$BC,$F1
	.dc.b	$9A,$F1,$57,$0A,$FC,$4C,$3E,$47,$FC,$DD,$1F,$EF,$C6,$C0,$85,$72
	.dc.b	$E0,$E5,$EA,$7F,$B3,$5E,$1F,$48,$F9,$3F,$E5,$31,$B1,$19,$DB,$11
	.dc.b	$8F,$94,$7C,$4E,$2B,$53,$C7,$07,$AF,$E4,$D7,$FD,$C7,$18,$9F,$EF
	.dc.b	$ED,$7F,$DB,$1F,$49,$FE,$E8,$9D,$D7,$1D,$7D,$B7,$B9,$CD,$FF,$82
	.dc.b	$2F,$85,$6B,$DF,$E1,$3F,$16,$D1,$88,$FF,$4C,$36,$1F,$21,$62,$4F
	.dc.b	$99,$D4,$62,$3F,$DD,$D3,$D1,$D1,$EC,$8E,$DF,$49,$F6,$63,$DF,$F0
	.dc.b	$FD,$CC,$F0,$C8,$FF,$10,$CE,$DF,$F7,$D3,$73,$F9,$3E,$FF,$A7,$E9
	.dc.b	$3F,$CF,$95,$FE,$28,$6B,$FF,$00,$5B,$2D,$3F,$0B,$1F,$C1,$27,$87
	.dc.b	$2F,$32,$3C,$F9,$BB,$7A,$1D,$FD,$09,$3D,$E3,$2F,$86,$32,$F8,$63
	.dc.b	$2F,$8E,$33,$8F,$8E,$3F,$31,$87,$FB,$79,$FF,$20,$FD,$23,$E6,$FD
	.dc.b	$DF,$47,$27,$31,$F6,$67,$51,$82,$1F,$07,$E6,$BE,$F7,$D2,$3D,$0B
	.dc.b	$E8,$4A,$FF,$27,$CE,$5B,$39,$59,$F9,$D4,$AB,$AD,$76,$FA,$19,$5D
	.dc.b	$1F,$0F,$D0,$62,$39,$F1,$7E,$DB,$05,$67,$1E,$D0,$BC,$56,$EA,$DF
	.dc.b	$EF,$EB,$4F,$F9,$A3,$B9,$B7,$C0,$B9,$39,$8F,$D0,$4E,$A3,$04,$3C
	.dc.b	$A7,$F0,$BF,$A3,$7D,$27,$E9,$DF,$CF,$46,$83,$FA,$87,$FC,$F2,$D7
	.dc.b	$4F,$0C,$FE,$53,$70,$FE,$2E,$EE,$33,$D5,$DF,$B3,$C1,$EB,$CC,$3A
	.dc.b	$BE,$3E,$CC,$47,$68,$B7,$DA,$6E,$3D,$A6,$D7,$B6,$7F,$82,$D7,$F2
	.dc.b	$DF,$9C,$BD,$A6,$98,$F8,$5C,$FC,$A4,$C3,$FC,$A3,$48,$E3,$F0,$BD
	.dc.b	$3E,$87,$EE,$D7,$F9,$59,$FA,$EF,$CE,$C5,$18,$7F,$48,$63,$93,$C7
	.dc.b	$DB,$E4,$ED,$97,$E0,$8F,$3F,$B9,$E8,$F6,$E4,$F9,$23,$C5,$F4,$52
	.dc.b	$3C,$4F,$97,$BF,$DF,$EB,$49,$F5,$BF,$0B,$D2,$F8,$7E,$24,$9F,$7B
	.dc.b	$F1,$BC,$4F,$77,$A5,$EF,$F8,$09,$5E,$F8,$F9,$BF,$49,$FD,$1A,$3F
	.dc.b	$CE,$EB,$E5,$B2,$6B,$CF,$DF,$DC,$F0,$F2,$E5,$F0,$8F,$99,$F0,$72
	.dc.b	$3E,$63,$F3,$FE,$0F,$F9,$E7,$D1,$23,$F9,$F7,$D0,$7F,$29,$F0,$BF
	.dc.b	$B2,$B6,$FF,$B4,$56,$DF,$F6,$8F,$D1,$7E,$05,$FF,$69,$C3,$DF,$EA
	.dc.b	$7F,$49,$9D,$4F,$E8,$D7,$26,$7C,$29,$3D,$23,$87,$1F,$93,$C6,$93
	.dc.b	$E0,$8F,$EA,$9F,$44,$F9,$3D,$CE,$4F,$F7,$2C,$65,$B1,$C0,$8F,$FB
	.dc.b	$AB,$47,$BD,$FD,$7F,$E8,$9C,$3F,$12,$E5,$FF,$BC,$27,$2D,$9C,$BF
	.dc.b	$BF,$BF,$03,$FC,$06,$76,$D9,$EC,$47,$C5,$F4,$5E,$1D,$FE,$9F,$93
	.dc.b	$27,$C1,$19,$7D,$11,$97,$B4,$7F,$77,$FA,$2F,$D7,$47,$FB,$03,$E6
	.dc.b	$FF,$84,$72,$7F,$88,$7C,$64,$62,$7F,$C4,$95,$C1,$FE,$30,$D9,$FE
	.dc.b	$78,$DA,$FF,$14,$B6,$78,$FC,$7F,$F8,$1F,$1F,$CC,$47,$7A,$FF,$17
	.dc.b	$C7,$E7,$63,$F9,$23,$6F,$FA,$43,$C2,$33,$D7,$CC,$8F,$33,$FC,$75
	.dc.b	$F4,$4C,$3F,$A8,$C3,$E8,$57,$FE,$3E,$AD,$61,$FE,$41,$9C,$D7,$F9
	.dc.b	$0E,$BC,$11,$88,$E3,$87,$F9,$1A,$31,$7C,$A7,$B5,$75,$C3,$C3,$5C
	.dc.b	$63,$99,$F0,$2F,$93,$BC,$3C,$4D,$BE,$67,$8E,$76,$D3,$67,$C0,$BE
	.dc.b	$F3,$C6,$FC,$15,$C2,$BD,$27,$01,$87,$A6,$7D,$77,$CA,$7B,$1E,$23
	.dc.b	$89,$67,$85,$F9,$E3,$FC,$CD,$1F,$0D,$B3,$F4,$9B,$87,$F9,$8A,$5F
	.dc.b	$22,$FA,$2C,$36,$BF,$CD,$53,$A9,$C9,$FE,$6A,$8F,$F3,$C7,$D1,$54
	.dc.b	$D2,$3B,$7E,$05,$FF,$4F,$D3,$E1,$4F,$23,$DD,$19,$7E,$01,$9C,$FF
	.dc.b	$05,$19,$7F,$09,$19,$CF,$FC,$ED,$E7,$FF,$8D,$67,$33,$E4,$3A,$C6
	.dc.b	$67,$FD,$0D,$87,$92,$31,$3F,$E8,$F8,$C6,$1E,$68,$E5,$7E,$2A,$DC
	.dc.b	$7F,$A6,$2B,$6F,$A0,$DD,$F9,$2F,$38,$7F,$A1,$6F,$FB,$3C,$78,$AB
	.dc.b	$CA,$FE,$AD,$FF,$1C,$C6,$9F,$E0,$CD,$7C,$07,$03,$FC,$39,$F4,$3F
	.dc.b	$D5,$67,$93,$63,$27,$17,$15,$FE,$45,$6C,$FF,$AC,$9B,$3F,$EB,$57
	.dc.b	$FD,$6D,$1D,$36,$E5,$CD,$23,$F8,$15,$BE,$8B,$FD,$81,$1E,$29,$D6
	.dc.b	$1F,$E0,$99,$7C,$8E,$2F,$FC,$13,$18,$BF,$4F,$D1,$79,$CF,$85,$3F
	.dc.b	$C9,$BA,$B1,$3A,$26,$1F,$E1,$5E,$5F,$E9,$31,$8A,$F7,$5F,$29,$FF
	.dc.b	$55,$CE,$4F,$E6,$4E,$2B,$FC,$8D,$F4,$4F,$27,$E8,$7C,$4F,$F6,$BB
	.dc.b	$FA,$D3,$97,$EC,$46,$6B,$FC,$9C,$DC,$7F,$CE,$4D,$9F,$F6,$FD,$71
	.dc.b	$AD,$F4,$FD,$DF,$A2,$7A,$54,$F0,$C5,$BF,$C3,$4B,$D2,$E0,$7E,$17
	.dc.b	$E8,$BF,$D1,$13,$A8,$FC,$47,$D1,$3C,$CA,$70,$9F,$C8,$46,$5B,$F2
	.dc.b	$1C,$0E,$3C,$B6,$3E,$FF,$91,$FE,$FA,$E0,$FF,$20,$BF,$E0,$18,$CF
	.dc.b	$C5,$7D,$88,$C6,$FF,$E4,$97,$FD,$DB,$3F,$88,$FA,$2F,$F5,$F3,$D3
	.dc.b	$E3,$7E,$FC,$62,$BF,$CA,$0B,$D3,$C9,$AE,$5B,$7F,$92,$F8,$87,$D1
	.dc.b	$7E,$21,$B1,$9F,$A4,$C4,$E7,$D3,$93,$C7,$CB,$B9,$FF,$97,$FE,$8F
	.dc.b	$FC,$D9,$3F,$9E,$5F,$05,$F9,$47,$33,$B8,$FC,$BF,$D1,$FD,$28,$FF
	.dc.b	$34,$3F,$EB,$1A,$FD,$3B,$9E,$2F,$F3,$4B,$99,$FF,$37,$7D,$1F,$F2
	.dc.b	$39,$FF,$36,$AF,$83,$FD,$1B,$9B,$FF,$05,$39,$7F,$66,$BD,$CD,$9C
	.dc.b	$3D,$5B,$3F,$B8,$FA,$3F,$EF,$0D,$F4,$7F,$5F,$1F,$D1,$A3,$06,$FE
	.dc.b	$8F,$27,$ED,$3D,$FF,$E0,$BA,$7F,$A1,$69,$39,$FA,$3D,$79,$37,$0F
	.dc.b	$E0,$F8,$7E,$DB,$0F,$F4,$67,$CC,$78,$9B,$97,$F0,$08,$F1,$46,$2B
	.dc.b	$F8,$7C,$6F,$7F,$F4,$8B,$63,$FC,$58,$F0,$8C,$D7,$32,$4F,$2F,$56
	.dc.b	$DF,$FB,$17,$E8,$DF,$0B,$FC,$BD,$B5,$FC,$7B,$E7,$7F,$24,$FF,$37
	.dc.b	$8C,$D7,$1A,$FE,$2F,$5F,$BD,$C7,$FD,$E5,$5A,$7F,$D4,$F1,$9B,$F3
	.dc.b	$F9,$5E,$F3,$E0,$7C,$3D,$8F,$7C,$5C,$7E,$3C,$DB,$E7,$D5,$E0,$EA
	.dc.b	$F0,$B7,$0F,$43,$6F,$FD,$75,$8F,$FA,$3E,$7D,$95,$C1,$A7,$02,$CF
	.dc.b	$FB,$A7,$E8,$9F,$19,$FD,$4D,$B3,$F5,$35,$C4,$C8,$FE,$8B,$F4,$6E
	.dc.b	$4C,$61,$FE,$CB,$8C,$4F,$CD,$F9,$5F,$92,$7C,$AF,$06,$C6,$49,$5E
	.dc.b	$A8,$CB,$F8,$AF,$33,$F1,$1F,$45,$F8,$96,$D7,$E2,$1B,$3B,$8F,$C5
	.dc.b	$CE,$DA,$6C,$B1,$F0,$7D,$6F,$80,$7D,$EF,$A3,$B3,$FD,$E7,$6F,$EB
	.dc.b	$55,$FE,$E7,$FA,$36,$1A,$8C,$3F,$9D,$FA,$3F,$CF,$79,$97,$0F,$CE
	.dc.b	$B6,$BF,$4A,$DE,$5F,$CF,$BF,$8E,$FA,$39,$5F,$95,$F5,$6D,$FD,$E9
	.dc.b	$B7,$F9,$09,$B5,$FE,$91,$E6,$56,$2D,$69,$FE,$B1,$18,$77,$5F,$E1
	.dc.b	$7F,$A3,$FD,$1C,$88,$57,$2E,$03,$87,$6D,$78,$79,$BF,$E0,$0E,$4F
	.dc.b	$E0,$D8,$2F,$73,$9F,$07,$F5,$FF,$47,$27,$FC,$27,$E9,$BD,$9C,$65
	.dc.b	$FA,$0A,$B8,$F3,$8F,$7D,$6E,$FF,$7D,$5F,$C4,$7C,$89,$1F,$C4,$3C
	.dc.b	$8B,$7F,$C6,$26,$E1,$FC,$4A,$6D,$7F,$12,$9B,$5F,$33,$C5,$56,$2D
	.dc.b	$1F,$CD,$EF,$F8,$ED,$FF,$21,$AD,$4E,$6F,$C1,$5E,$33,$64,$72,$7F
	.dc.b	$2C,$9E,$35,$C9,$B3,$BB,$F4,$34,$F4,$16,$BD,$13,$C1,$EB,$3C,$1B
	.dc.b	$3C,$1A,$70,$79,$47,$BA,$BC,$51,$8B,$F6,$4F,$F3,$A9,$F8,$13,$97
	.dc.b	$75,$D2,$79,$4F,$1B,$F2,$47,$18,$F8,$93,$BB,$EB,$1E,$59,$F4,$1B
	.dc.b	$9D,$5F,$81,$B3,$D4,$F1,$15,$C9,$D6,$FC,$31,$D4,$FE,$A2,$E2,$3F
	.dc.b	$01,$3D,$6B,$C0,$4A,$E1,$3F,$1C,$EB,$19,$70,$D3,$E0,$BE,$18,$F2
	.dc.b	$B4,$E2,$F4,$8F,$C6,$D7,$8D,$E2,$7C,$28,$D5,$FC,$6A,$F3,$C7,$CA
	.dc.b	$9F,$04,$E5,$E3,$1E,$E3,$C2,$EA,$BB,$4E,$CF,$CF,$CE,$8E,$F1,$AB
	.dc.b	$F8,$91,$E2,$7E,$FD,$78,$1B,$19,$3E,$94,$65,$A6,$6F,$E1,$46,$5B
	.dc.b	$19,$8F,$33,$E3,$8F,$79,$FA,$77,$11,$EE,$8F,$A3,$1C,$4F,$49,$AB
	.dc.b	$F7,$5F,$AE,$3E,$41,$B9,$EB,$5F,$3D,$B1,$98,$FA,$AF,$3B,$F8,$73
	.dc.b	$A3,$F9,$67,$D1,$3E,$8F,$F8,$E4,$7F,$2B,$11,$9D,$31,$18,$F9,$B7
	.dc.b	$C9,$62,$B5,$3C,$7E,$4F,$A3,$C1,$AE,$70,$FF,$80,$11,$88,$FE,$75
	.dc.b	$EC,$3D,$8F,$E7,$FA,$FF,$2B,$94,$E5,$FF,$6E,$35,$FE,$9B,$AF,$F4
	.dc.b	$6F,$A3,$E1,$5C,$6B,$4D,$BE,$B5,$FD,$32,$72,$F1,$33,$7F,$D4,$61
	.dc.b	$F3,$58,$8F,$1B,$63,$31,$E4,$F9,$1C,$4C,$8F,$3B,$B8,$C4,$7A,$0E
	.dc.b	$D1,$8C,$3D,$13,$DF,$E8,$DF,$37,$DF,$E2,$A6,$1F,$EF,$D9,$CB,$67
	.dc.b	$81,$1F,$8F,$F0,$3E,$9F,$A3,$FF,$3D,$47,$F6,$DC,$7E,$74,$62,$3C
	.dc.b	$81,$C4,$6B,$9B,$97,$34,$AE,$5B,$7F,$76,$BD,$B6,$BC,$1E,$8F,$87
	.dc.b	$D1,$97,$C0,$FE,$FD,$7F,$ED,$67,$07,$86,$76,$F9,$21,$3D,$23,$71
	.dc.b	$C5,$D4,$70,$C7,$DB,$5F,$67,$E8,$DF,$35,$F0,$FE,$82,$61,$E0,$9C
	.dc.b	$B6,$78,$11,$EF,$F4,$7F,$83,$FA,$27,$A3,$7D,$19,$5E,$A9,$CB,$67
	.dc.b	$2F,$CE,$84,$F3,$9E,$9F,$3F,$2B,$93,$DF,$E8,$F1,$1C,$62,$EB,$FA
	.dc.b	$7C,$15,$8E,$1F,$F1,$4C,$2B,$17,$FE,$2D,$8D,$BF,$43,$C0,$6D,$ED
	.dc.b	$B9,$39,$8F,$D1,$4E,$A3,$04,$3F,$13,$F4,$4F,$A3,$FD,$43,$FA,$68
	.dc.b	$CF,$56,$0C,$F5,$C8,$FC,$0C,$A7,$51,$89,$E7,$EB,$C8,$DB,$DB,$B3
	.dc.b	$11,$E2,$8B,$BF,$F0,$BC,$15,$8D,$3F,$C9,$B1,$FE,$92,$BE,$67,$8A
	.dc.b	$7F,$81,$D7,$EC,$BE,$7A,$F6,$1A,$63,$E1,$73,$F2,$B3,$0F,$CF,$5F
	.dc.b	$EE,$A3,$8F,$C2,$F4,$FA,$3F,$A9,$5A,$7E,$9B,$FC,$16,$91,$87,$FC
	.dc.b	$9C,$63,$93,$C7,$D7,$E4,$EC,$97,$E0,$D3,$F4,$97,$C1,$FF,$69,$C6
	.dc.b	$7D,$BF,$59,$EC,$F7,$7C,$9F,$47,$27,$C8,$F9,$FB,$BE,$0F,$AF,$27
	.dc.b	$B3,$F1,$3E,$13,$E2,$F8,$92,$7C,$4F,$C8,$F1,$3E,$07,$A7,$F0,$7E
	.dc.b	$32,$57,$C1,$8F,$9F,$E6,$7F,$4E,$8F,$92,$BE,$5B,$26,$BD,$0D,$FD
	.dc.b	$1F,$0F,$2E,$5F,$08,$F9,$DF,$09,$23,$E7,$3F,$47,$E1,$3F,$9E,$7D
	.dc.b	$1A,$3F,$9F,$7D,$0F,$F2,$CA,$FE,$46,$DE,$7F,$EA,$95,$B9,$FE,$2F
	.dc.b	$F4,$7F,$83,$7F,$8C,$C3,$E0,$3A,$9F,$D2,$67,$53,$FC,$B9,$72,$67
	.dc.b	$C1,$93,$D2,$38,$71,$F9,$3C,$79,$3E,$08,$FE,$A9,$F4,$6F,$94,$DC
	.dc.b	$E4,$FE,$57,$19,$6C,$70,$23,$FC,$C3,$57,$BD,$FD,$7F,$E8,$DC,$3F
	.dc.b	$12,$E5,$FF,$36,$9C,$B6,$72,$FC,$89,$E5,$12,$34,$43,$E6,$FD,$1F
	.dc.b	$7D,$FE,$8F,$91,$27,$B4,$65,$F3,$46,$5F,$EA,$F8,$7C,$F7,$FA,$4F
	.dc.b	$D1,$BC,$8F,$DC,$FC,$E7,$F0,$B9,$FD,$9B,$5F,$EB,$F3,$B7,$8C,$FF
	.dc.b	$7F,$8C,$46,$A3,$1F,$21,$FE,$03,$C3,$F2,$93,$FD,$61,$FC,$C3,$FD
	.dc.b	$A9,$BC,$DF,$E1,$B7,$71,$9E,$BE,$44,$7F,$B0,$FD,$1A,$FF,$DD,$30
	.dc.b	$FA,$2F,$E3,$2F,$FB,$96,$75,$87,$FB,$9E,$73,$3F,$C7,$AF,$FD,$D3
	.dc.b	$18,$C3,$C2,$EA,$F8,$D7,$59,$FF,$77,$C6,$27,$85,$72,$9F,$F7,$99
	.dc.b	$C8,$DE,$1E,$69,$E8,$D8,$DC,$F1,$BF,$F7,$E4,$71,$7D,$53,$B6,$E1
	.dc.b	$EB,$9E,$31,$4E,$35,$CD,$61,$E7,$9F,$1D,$6E,$7E,$11,$63,$C0,$F0
	.dc.b	$BF,$03,$67,$FC,$BF,$5F,$12,$07,$43,$FC,$CC,$E5,$F8,$FF,$24,$FA
	.dc.b	$3E,$2D,$8D,$B5,$E6,$7D,$97,$0F,$F9,$C7,$E8,$FF,$CD,$EF,$81,$A9
	.dc.b	$A4,$7E,$C1,$FE,$3B,$D1,$E9,$7C,$18,$E4,$68,$63,$FA,$58,$D4,$75
	.dc.b	$35,$19,$78,$1C,$B5,$FC,$6C,$65,$FC,$7C,$67,$4F,$F3,$CD,$F8,$A3
	.dc.b	$89,$D6,$78,$C6,$5E,$31,$99,$E3,$7E,$4C,$3E,$D6,$1D,$E7,$E3,$57
	.dc.b	$9B,$0E,$51,$8C,$3A,$E1,$C2,$7F,$B3,$BD,$2B,$FC,$BA,$79,$CF,$90
	.dc.b	$FF,$94,$2F,$E0,$1F,$E8,$A9,$E3,$3F,$60,$F3,$1C,$0F,$F3,$3F,$D1
	.dc.b	$FF,$55,$AF,$F2,$CB,$63,$27,$20,$E6,$BF,$D3,$CD,$9F,$C2,$B6,$7F
	.dc.b	$22,$FE,$4A,$3A,$ED,$CE,$3F,$A0,$47,$83,$97,$CF,$6F,$A3,$FF,$44
	.dc.b	$CE,$49,$87,$F8,$36,$1F,$29,$CE,$F8,$C6,$2B,$D7,$F4,$7D,$CE,$BC
	.dc.b	$9F,$D1,$A3,$24,$C3,$E1,$CE,$BA,$78,$4E,$72,$FF,$0C,$DF,$FA,$B2
	.dc.b	$30,$7F,$A4,$DC,$5F,$FA,$4B,$E8,$DE,$4F,$C9,$F8,$AF,$ED,$97,$FE
	.dc.b	$86,$9C,$BF,$E8,$88,$CD,$7F,$A6,$9B,$87,$FB,$2D,$B3,$FB,$2A,$E3
	.dc.b	$5B,$EA,$FA,$DF,$46,$F5,$29,$E1,$8B,$7F,$E9,$69,$7A,$5C,$0F,$BD
	.dc.b	$F4,$7F,$E8,$49,$D4,$7E,$07,$E8,$DE,$65,$38,$4F,$FA,$82,$32,$DF
	.dc.b	$90,$E0,$71,$E4,$B1,$FE,$A6,$FB,$F7,$CA,$FE,$35,$B1,$FE,$50,$F2
	.dc.b	$9F,$37,$9A,$FC,$94,$BC,$0A,$7E,$23,$E9,$1F,$47,$89,$FF,$59,$FD
	.dc.b	$B7,$E4,$79,$3F,$D8,$AD,$E7,$FF,$29,$DF,$F8,$A6,$B7,$5F,$9F,$A7
	.dc.b	$A7,$4F,$CA,$FC,$5B,$E1,$F1,$7D,$BB,$B8,$1A,$F8,$BB,$1F,$EC,$4F
	.dc.b	$1D,$EA,$7E,$AD,$B1,$FE,$6E,$FA,$4F,$DC,$BF,$CB,$A3,$3A,$7F,$B7
	.dc.b	$1F,$F0,$DC,$73,$F4,$3F,$4F,$F4,$8F,$A0,$E9,$1F,$B7,$FB,$57,$C4
	.dc.b	$53,$F5,$BF,$4B,$FD,$D7,$0F,$F6,$BC,$E9,$BF,$03,$C4,$DF,$D2,$58
	.dc.b	$DB,$D4,$FD,$BF,$D2,$FE,$3E,$32,$7F,$B8,$BE,$03,$EC,$B7,$B3,$F7
	.dc.b	$B6,$9F,$E1,$86,$B7,$FF,$7C,$FE,$11,$F6,$AF,$FD,$D6,$D8,$FF,$46
	.dc.b	$57,$EB,$BA,$BF,$9B,$42,$75,$18,$FA,$67,$C4,$7F,$19,$6E,$1F,$E9
	.dc.b	$CF,$A2,$7C,$36,$6A,$4F,$FB,$EA,$B7,$8F,$FA,$6E,$F7,$39,$AD,$CE
	.dc.b	$7C,$FB,$8F,$FA,$7A,$FF,$6A,$FF,$A8,$3E,$7F,$E3,$57,$FA,$95,$B1
	.dc.b	$9F,$8C,$FE,$43,$18,$C7,$FB,$CC,$A3,$15,$FD,$89,$B1,$9A,$E3,$E7
	.dc.b	$FF,$2A,$C3,$F3,$97,$E9,$E6,$C7,$C3,$E5,$B9,$7C,$36,$CE,$E3,$C2
	.dc.b	$7F,$80,$3E,$99,$F2,$5F,$D0,$DB,$1F,$EC,$6D,$3E,$3E,$9F,$52,$73
	.dc.b	$7F,$DE,$2B,$95,$F1,$3D,$FB,$58,$EF,$3F,$E1,$8A,$E3,$7F,$EC,$8F
	.dc.b	$A2,$FC,$33,$D7,$0F,$E8,$F1,$8B,$F8,$FA,$7F,$19,$C3,$EC,$78,$1F
	.dc.b	$0F,$C3,$F8,$8C,$AF,$C1,$39,$69,$EB,$F4,$3E,$9F,$D2,$FD,$46,$C6
	.dc.b	$4F,$81,$19,$69,$EA,$B6,$9F,$D4,$DF,$84,$7B,$7D,$AF,$AD,$E0,$7C
	.dc.b	$B6,$57,$1A,$E1,$B7,$E2,$EB,$37,$C8,$EB,$7F,$DA,$3E,$99,$8E,$AF
	.dc.b	$FD,$DF,$18,$C7,$EB,$7D,$37,$E3,$BC,$FF,$C7,$B6,$32,$7D,$B8,$CB
	.dc.b	$4F,$D4,$79,$FF,$8D,$3F,$1D,$EB,$D8,$C9,$33,$FD,$44,$65,$F4,$7D
	.dc.b	$37,$EA,$9B,$DB,$FA,$0E,$FF,$F0,$74,$E7,$4F,$DE,$B7,$0F,$F2,$8B
	.dc.b	$6F,$F4,$55,$C2,$E5,$7E,$2A,$DC,$BF,$DF,$EF,$BC,$E6,$31,$3F,$B7
	.dc.b	$FA,$6F,$D4,$1F,$E1,$2F,$55,$ED,$E2,$FF,$84,$4D,$C3,$ED,$DA,$F7
	.dc.b	$5A,$3F,$49,$F2,$3F,$BC,$3F,$7B,$F1,$B6,$32,$4B,$FE,$09,$19,$7E
	.dc.b	$F7,$DD,$BE,$AF,$F8,$FC,$7F,$9C,$A3,$15,$FE,$3D,$8F,$F3,$FC,$61
	.dc.b	$C3,$5F,$F2,$13,$9C,$7F,$CB,$DF,$10,$FC,$69,$F4,$DF,$E3,$FA,$D6
	.dc.b	$1F,$E8,$AA,$D4,$E7,$E7,$E4,$FE,$E2,$7F,$C8,$BF,$43,$8C,$BD,$35
	.dc.b	$E8,$7C,$AD,$58,$23,$0D,$BF,$64,$EB,$E5,$72,$48,$CC,$A3,$13,$FA
	.dc.b	$EC,$1D,$5F,$D1,$2F,$E4,$57,$F9,$92,$73,$EE,$FB,$FD,$49,$5E,$BF
	.dc.b	$89,$FE,$72,$FC,$F8,$CE,$3F,$51,$E5,$3A,$8F,$A3,$1A,$DF,$FA,$AE
	.dc.b	$7F,$DB,$67,$C9,$5B,$7E,$0D,$6E,$FB,$1E,$1F,$37,$13,$D2,$3F,$86
	.dc.b	$5F,$3C,$3E,$CE,$1C,$E3,$13,$CA,$BC,$CF,$33,$8E,$1D,$23,$C3,$19
	.dc.b	$76,$4B,$F4,$CF,$BB,$0E,$17,$DA,$BC,$87,$B2,$33,$3E,$8C,$BF,$D0
	.dc.b	$F1,$8C,$3C,$27,$7C,$3A,$B8,$BF,$03,$F8,$FC,$7D,$98,$71,$9E,$D3
	.dc.b	$BA,$E3,$1C,$0F,$F6,$D5,$70,$9D,$1F,$0E,$7E,$3D,$F3,$78,$98,$8D
	.dc.b	$D7,$B2,$33,$1E,$39,$E0,$F3,$30,$70,$9F,$96,$7E,$C2,$FF,$36,$48
	.dc.b	$FC,$BB,$63,$51,$82,$3F,$C0,$3E,$77,$E7,$92,$3B,$64,$8C,$4F,$C1
	.dc.b	$3F,$56,$E3,$0F,$4C,$7F,$04,$8C,$57,$A6,$30,$F4,$8C,$C7,$C1,$8F
	.dc.b	$E1,$D1,$88,$F6,$CF,$E7,$23,$15,$F8,$39,$F0,$D7,$32,$3E,$58,$F4
	.dc.b	$46,$2B,$CB,$3C,$0D,$57,$4B,$FE,$24,$DA,$FD,$0B,$67,$E3,$47,$33
	.dc.b	$D8,$F3,$AE,$B7,$E0,$8E,$B3,$B7,$DC,$7C,$23,$27,$F3,$0F,$37,$F9
	.dc.b	$11,$FC,$8F,$E4,$2C,$FE,$CA,$3F,$8B,$91,$FE,$26,$FF,$1A,$F0,$24
	.dc.b	$7F,$16,$6D,$7F,$37,$F9,$87,$1F,$F2,$36,$D7,$F2,$06,$CE,$E7,$5F
	.dc.b	$35,$2B,$4D,$9F,$E8,$5F,$52,$FA,$9F,$EB,$31,$86,$A2,$08,$FF,$4E
	.dc.b	$7A,$DF,$CC,$1E,$5D,$FF,$CC,$BE,$A3,$FA,$2D,$7F,$87,$5A,$FF,$84
	.dc.b	$DB,$C1,$FE,$74,$FA,$9F,$F3,$AB,$FE,$76,$8C,$E7,$E0,$6D,$78,$DA
	.dc.b	$AB,$8C,$EF,$E8,$FC,$F7,$CD,$B1,$9A,$F2,$7C,$EE,$1E,$93,$56,$08
	.dc.b	$C3,$63,$89,$FE,$39,$FA,$97,$D0,$EB,$E6,$64,$E6,$3A,$4E,$A3,$E2
	.dc.b	$4E,$9F,$F1,$B4,$65,$B3,$C0,$8F,$CB,$F9,$0F,$37,$D4,$BC,$6A,$49
	.dc.b	$FF,$26,$43,$C6,$C4,$7F,$90,$E3,$83,$A9,$E7,$3D,$3A,$3A,$FB,$B2
	.dc.b	$B9,$C6,$5E,$B1,$97,$A4,$7C,$9F,$A9,$7D,$1F,$9B,$C6,$95,$FE,$5F
	.dc.b	$9C,$B6,$72,$F9,$67,$94,$48,$D1,$0F,$57,$D4,$FD,$7F,$A9,$7B,$57
	.dc.b	$DB,$95,$FE,$6D,$9C,$B6,$72,$FF,$99,$61,$3C,$E7,$A7,$D3,$4A,$E4
	.dc.b	$F7,$EF,$C4,$71,$8B,$D3,$FE,$CE,$C3,$F6,$F3,$FE,$7D,$F8,$FF,$F0
	.dc.b	$C3,$E3,$EA,$F4,$1F,$8F,$8C,$C7,$F9,$DD,$B5,$CE,$77,$3C,$F5,$E0
	.dc.b	$47,$07,$FA,$2F,$EA,$1F,$53,$F9,$B7,$F3,$11,$FE,$83,$5D,$99,$35
	.dc.b	$D9,$27,$94,$65,$E9,$1C,$FE,$06,$47,$77,$B7,$66,$23,$B4,$5C,$FF
	.dc.b	$DE,$38,$2B,$E0,$7A,$FF,$AF,$E0,$FC,$B6,$1F,$EA,$36,$BF,$0E,$36
	.dc.b	$E7,$A3,$81,$87,$47,$CF,$8C,$61,$FF,$00,$29,$1F,$BD,$FB,$77,$D6
	.dc.b	$7E,$FF,$ED,$FF,$03,$87,$FA,$CD,$AF,$FA,$A6,$76,$FF,$AD,$D7,$BB
	.dc.b	$C7,$96,$4E,$63,$F8,$57,$DB,$BE,$67,$F8,$67,$DB,$7F,$18,$3F,$8C
	.dc.b	$C6,$6B,$FD,$84,$BE,$23,$7E,$24,$9C,$C7,$F1,$5F,$B7,$7C,$CF,$4F
	.dc.b	$22,$57,$FB,$2E,$72,$DF,$99,$F0,$F9,$B2,$30,$7F,$B4,$7D,$57,$D8
	.dc.b	$7F,$21,$FB,$7F,$A3,$CE,$F0,$3F,$A0,$C6,$1F,$E8,$75,$C2,$F9,$7C
	.dc.b	$7F,$1F,$2A,$46,$0F,$F6,$F7,$DB,$2B,$B7,$C1,$49,$CC,$7F,$31,$97
	.dc.b	$B1,$FC,$CF,$EA,$5F,$0B,$4D,$46,$FE,$5F,$5E,$F4,$C7,$BC,$7F,$3B
	.dc.b	$FB,$67,$CB,$F3,$F3,$25,$7F,$BC,$67,$2F,$FB,$CF,$EC,$9F,$2D,$FD
	.dc.b	$0F,$EC,$7F,$9A,$72,$FC,$3A,$FE,$C9,$18,$E4,$FE,$CB,$7C,$27,$FD
	.dc.b	$F9,$F4,$1F,$E8,$F7,$A7,$C8,$7F,$4D,$93,$FD,$3E,$75,$3F,$EF,$75
	.dc.b	$F0,$79,$F8,$99,$3D,$23,$87,$99,$E5,$F4,$64,$F8,$23,$FA,$D7,$D9
	.dc.b	$3E,$97,$73,$93,$FB,$E6,$DE,$B5,$3B,$73,$AF,$F8,$02,$31,$09,$E1
	.dc.b	$18,$6D,$7C,$7F,$AB,$FE,$D5,$F6,$2E,$EF,$95,$33,$0E,$A7,$F8,$36
	.dc.b	$74,$E0,$38,$7C,$18,$FF,$84,$B8,$7C,$0D,$AE,$F7,$2B,$95,$F1,$C7
	.dc.b	$DC,$DB,$E5,$F6,$5E,$6E,$1F,$BD,$E9,$CC,$38,$C6,$5F,$26,$7F,$C6
	.dc.b	$5B,$39,$7C,$A7,$F8,$9F,$EC,$9D,$BF,$E0,$DC,$DF,$5D,$FC,$C2,$BE
	.dc.b	$0C,$63,$8B,$FC,$21,$7C,$27,$E1,$35,$EA,$1C,$4F,$D9,$6E,$9F,$E3
	.dc.b	$67,$F5,$33,$F2,$F1,$7D,$87,$F8,$75,$78,$D4,$C4,$49,$E0,$D7,$C3
	.dc.b	$F6,$5F,$B4,$C7,$FC,$89,$F6,$5F,$0E,$53,$95,$19,$C3,$C3,$1C,$1F
	.dc.b	$F2,$59,$87,$6F,$F9,$3A,$79,$4F,$56,$C7,$0B,$FF,$29,$CF,$56,$C7
	.dc.b	$1F,$07,$F2,$D7,$D0,$FF,$2D,$5F,$27,$FC,$B9,$3E,$1B,$E5,$1E,$53
	.dc.b	$75,$CA,$3C,$4D,$9F,$14,$EC,$FE,$B9,$7E,$27,$8E,$19,$C3,$C1,$1A
	.dc.b	$38,$57,$84,$F4,$4E,$EF,$D3,$3F,$E7,$03,$E6,$B8,$76,$64,$CC,$57
	.dc.b	$E9,$FA,$2F,$92,$C6,$1F,$D4,$DB,$19,$9E,$53,$97,$90,$FA,$9A,$D3
	.dc.b	$2E,$08,$FC,$F3,$F8,$37,$0F,$F2,$7D,$3F,$0D,$1F,$E8,$B9,$F9,$18
	.dc.b	$FE,$9B,$7F,$95,$39,$39,$46,$5E,$86,$6B,$F2,$11,$97,$AC,$79,$5C
	.dc.b	$DF,$87,$4F,$8D,$87,$82,$31,$5F,$36,$31,$87,$8A,$39,$57,$FA,$2A
	.dc.b	$B7,$1F,$EA,$2A,$DB,$E4,$36,$7F,$2C,$8D,$E1,$E9,$3F,$D3,$AF,$F4
	.dc.b	$97,$F9,$84,$7A,$EB,$B5,$FF,$21,$9E,$0E,$27,$F8,$09,$B3,$B3,$FC
	.dc.b	$9F,$E4,$FF,$7A,$DD,$BB,$3F,$EB,$88,$FF,$5D,$39,$24,$FF,$0C,$8E
	.dc.b	$2D,$38,$9F,$EC,$17,$35,$FD,$1D,$FE,$99,$3E,$1E,$8D,$CB,$EC,$BF
	.dc.b	$D4,$73,$92,$5F,$F2,$69,$D7,$CF,$F6,$86,$1D,$1D,$5A,$75,$2F,$3E
	.dc.b	$9E,$0E,$67,$66,$32,$F8,$93,$71,$F0,$56,$A7,$24,$8C,$FC,$06,$F3
	.dc.b	$FA,$D5,$FE,$C1,$AF,$EA,$9F,$64,$F2,$BD,$1F,$21,$FD,$EA,$73,$18
	.dc.b	$ED,$91,$FD,$6E,$75,$97,$F3,$AB,$E5,$39,$AF,$85,$F6,$4F,$8C,$A7
	.dc.b	$2B,$FE,$83,$19,$AF,$EC,$CB,$E9,$38,$9F,$2F,$EC,$BF,$D4,$73,$A8
	.dc.b	$F9,$DF,$64,$F9,$2A,$70,$9F,$EE,$11,$96,$FC,$FF,$03,$8F,$3D,$8F
	.dc.b	$EE,$DF,$65,$FE,$62,$6F,$C1,$7E,$0B,$CC,$FF,$17,$C6,$0F,$EF,$87
	.dc.b	$F9,$8C,$BE,$0F,$E2,$3E,$C9,$F3,$DF,$83,$9D,$70,$7F,$60,$FB,$2F
	.dc.b	$CC,$B4,$E1,$E9,$FE,$26,$33,$D3,$FE,$1B,$7F,$CC,$B5,$FE,$0B,$6D
	.dc.b	$7E,$47,$E4,$9E,$B7,$1F,$82,$E5,$DD,$30,$FE,$EB,$5F,$2A,$FC,$53
	.dc.b	$F9,$7F,$19,$EA,$D3,$F4,$3F,$65,$FB,$8A,$FE,$3D,$1B,$F4,$E6,$3F
	.dc.b	$E0,$0F,$B2,$7D,$56,$2F,$FB,$EF,$D8,$FF,$99,$6B,$FC,$44,$D8,$FD
	.dc.b	$47,$D9,$7F,$99,$27,$3F,$03,$FA,$9A,$FF,$4E,$C7,$0F,$90,$FD,$9D
	.dc.b	$3E,$03,$51,$B6,$4F,$F8,$4B,$EC,$BF,$19,$8F,$8A,$7F,$69,$EB,$7D
	.dc.b	$A8,$FF,$08,$B7,$9B,$F8,$D6,$1F,$E0,$AC,$3F,$C7,$13,$95,$53,$EC
	.dc.b	$BE,$EC,$7F,$8F,$5B,$1F,$C0,$EB,$F5,$CF,$F8,$7A,$33,$7F,$E5,$5F
	.dc.b	$B2,$FF,$52,$5F,$F9,$15,$B5,$FC,$2F,$EC,$5F,$5B,$FC,$2E,$73,$7C
	.dc.b	$7E,$3B,$07,$F1,$6A,$DE,$1C,$E3,$2F,$F8,$D2,$3F,$6B,$87,$F8,$C2
	.dc.b	$F9,$CF,$F9,$A2,$73,$7F,$CE,$FE,$CB,$E4,$B7,$CB,$FE,$47,$0F,$8B
	.dc.b	$B1,$2F,$FC,$83,$3C,$1F,$F3,$9B,$B7,$31,$E7,$EF,$F4,$73,$D9,$F3
	.dc.b	$37,$3F,$84,$DC,$7F,$A7,$65,$FE,$71,$AF,$F3,$67,$D9,$3E,$6B,$F9
	.dc.b	$83,$63,$C7,$B6,$26,$47,$07,$D7,$18,$C7,$FC,$A7,$5C,$5F,$E7,$9F
	.dc.b	$5F,$F8,$17,$F0,$31,$9C,$7F,$B0,$4E,$56,$3F,$AD,$7C,$FF,$3F,$F5
	.dc.b	$7C,$0F,$AC,$CC,$FE,$A3,$96,$9F,$2B,$CE,$F9,$FF,$65,$F4,$1B,$19
	.dc.b	$3E,$7C,$65,$A7,$C4,$F6,$BE,$19,$F8,$5E,$EB,$19,$26,$BF,$8F,$8C
	.dc.b	$BF,$0A,$7F,$B0,$FD,$72,$F3,$9F,$F7,$F8,$C6,$1F,$80,$FB,$2F,$C2
	.dc.b	$F9,$9F,$86,$6C,$64,$FC,$2C,$65,$A7,$E5,$7C,$DF,$CD,$1F,$9B,$F5
	.dc.b	$6C,$64,$99,$FE,$96,$32,$FF,$A2,$BE,$CB,$F4,$2D,$EC,$FC,$BF,$77
	.dc.b	$F7,$39,$CE,$9F,$EC,$56,$E1,$FB,$CB,$46,$1F,$F6,$27,$5F,$F1,$27
	.dc.b	$A4,$FF,$AE,$E7,$54,$E5,$FE,$E5,$5A,$3F,$13,$EE,$7E,$D7,$E4,$66
	.dc.b	$1F,$EA,$6F,$51,$ED,$62,$FF,$80,$55,$C7,$FC,$59,$5A,$BF,$DD,$4F
	.dc.b	$EC,$BE,$9E,$47,$EC,$3E,$3A,$E1,$FE,$13,$D5,$58,$C7,$F7,$95,$6F
	.dc.b	$E0,$7D,$7B,$E5,$7F,$C4,$BF,$12,$FB,$27,$D9,$7F,$86,$EB,$57,$FE
	.dc.b	$DE,$AD,$4E,$6F,$8D,$F0,$AF,$F7,$05,$F0,$D3,$FD,$97,$87,$09,$CF
	.dc.b	$D3,$7F,$92,$27,$F9,$6D,$23,$10,$8C,$4F,$FA,$0D,$B1,$FE,$87,$6E
	.dc.b	$69,$FE,$6F,$6E,$AC,$47,$80,$F5,$38,$BF,$F6,$9D,$BE,$53,$E1,$CF
	.dc.b	$F3,$8A,$F8,$F1,$8F,$6B,$E2,$36,$3D,$ED,$BE,$7E,$97,$C9,$FB,$2F
	.dc.b	$94,$DA,$F8,$AD,$9D,$C7,$C8,$9D,$B4,$D9,$63,$67,$A2,$36,$D8,$DB
	.dc.b	$4D,$E7,$F1,$5F,$9F,$3C,$23,$11,$F3,$E3,$13,$F8,$66,$C7,$08,$EB
	.dc.b	$3C,$32,$56,$AB,$95,$68,$E4,$57,$91,$E4,$72,$D3,$35,$D6,$FE,$33
	.dc.b	$67,$E1,$36,$7D,$CD,$3C,$95,$E3,$BF,$2D,$78,$6B,$CA,$77,$2D,$71
	.dc.b	$8F,$1B,$63,$C7,$1F,$A8,$BF,$1B,$4F,$19,$63,$C6,$D9,$F1,$BE,$1A
	.dc.b	$F1,$CF,$85,$B3,$DE,$3C,$73,$C4,$F1,$95,$F1,$9F,$A8,$AF,$1B,$63
	.dc.b	$C6,$D3,$C6,$59,$F1,$BE,$1B,$F8,$C4,$9F,$33,$63,$51,$82,$38,$24
	.dc.b	$FA,$6F,$D7,$3E,$79,$F5,$9E,$17,$D6,$D3,$D6,$7E,$FE,$FD,$6D,$8F
	.dc.b	$5B,$4F,$59,$6B,$D6,$78,$6F,$D0,$D9,$F3,$B6,$7C,$F1,$E5,$BF,$3C
	.dc.b	$78,$CF,$39,$D6,$3C,$ED,$3C,$F1,$FE,$5B,$9F,$0C,$F3,$AF,$0B,$4F
	.dc.b	$09,$6B,$C2,$7F,$20,$F2,$FF,$87,$FA,$BF,$CB,$DB,$3F,$CA,$9B,$3F
	.dc.b	$C5,$9B,$7F,$B8,$FB,$27,$D3,$7F,$28,$6D,$FF,$5D,$BF,$EB,$D1,$87
	.dc.b	$01,$C3,$B7,$FC,$5C,$E9,$AE,$7C,$DE,$C7,$F3,$6F,$B0,$7D,$8F,$F5
	.dc.b	$C9,$EB,$18,$EC,$EF,$08,$FF,$4B,$FA,$98,$75,$1D,$FC,$BD,$BE,$56
	.dc.b	$E2,$69,$AF,$A1,$FE,$AF,$3F,$E7,$6D,$7F,$8D,$C6,$21,$18,$EA,$FF
	.dc.b	$00,$3F,$C8,$3E,$C7,$F9,$0B,$99,$E5,$B7,$F8,$16,$F9,$4E,$67,$9C
	.dc.b	$FF,$82,$6B,$8B,$BF,$8D,$EA,$D9,$E6,$7F,$A2,$AD,$3A,$8C,$7C,$BD
	.dc.b	$53,$A7,$89,$89,$D4,$66,$3C,$67,$23,$7F,$62,$FA,$4D,$FC,$A4,$9C
	.dc.b	$C7,$79,$D4,$76,$A5,$6E,$3A,$89,$DB,$86,$E7,$FD,$29,$B9,$7F,$8B
	.dc.b	$5B,$3F,$07,$E3,$3D,$DF,$63,$FE,$16,$8F,$6B,$5E,$ED,$BF,$EA,$AB
	.dc.b	$C3,$E3,$E2,$4A,$E2,$F4,$8E,$BD,$3D,$BD,$B9,$5D,$A3,$2F,$82,$32
	.dc.b	$F8,$23,$E9,$7D,$8B,$EC,$7C,$FE,$2C,$AF,$A7,$39,$6C,$E5,$F4,$CF
	.dc.b	$28,$91,$A2,$1E,$EF,$B1,$FC,$07,$D8,$FE,$76,$3E,$FA,$F6,$B2,$6B
	.dc.b	$A6,$4E,$A3,$2F,$08,$E7,$F5,$92,$39,$BD,$BC,$1C,$47,$DC,$B4,$E3
	.dc.b	$87,$FC,$07,$1F,$C3,$27,$F2,$3F,$19,$FE,$35,$7C,$BC,$BE,$83,$F3
	.dc.b	$11,$98,$FC,$3B,$6B,$9C,$EE,$79,$EB,$C0,$8E,$0F,$CF,$FD,$83,$EC
	.dc.b	$7F,$48,$FE,$7E,$FF,$C2,$EB,$B3,$26,$BB,$25,$F2,$8C,$BD,$23,$9F
	.dc.b	$C1,$48,$EE,$F6,$EB,$C4,$7E,$82,$FF,$D9,$CE,$8F,$2C,$7C,$0F,$5B
	.dc.b	$F7,$1C,$1F,$95,$9F,$DB,$B5,$F8,$71,$B7,$3D,$1C,$0C,$3A,$3E,$7C
	.dc.b	$63,$3F,$F1,$AD,$23,$F8,$37,$D7,$BE,$C7,$F8,$47,$D8,$FE,$1F,$0F
	.dc.b	$E0,$ED,$7F,$73,$3B,$7F,$85,$AF,$73,$8F,$2C,$9C,$C7,$F1,$3F,$B0
	.dc.b	$7C,$AF,$F1,$5F,$B1,$FE,$47,$CD,$FE,$A9,$6D,$85,$70,$7F,$8C,$AF
	.dc.b	$88,$E5,$E3,$49,$CC,$7F,$21,$FB,$17,$CA,$F7,$F2,$25,$7F,$22,$9C
	.dc.b	$B7,$E5,$7C,$9E,$74,$8C,$1F,$CA,$3D,$57,$D8,$FF,$2D,$FB,$1F,$A9
	.dc.b	$D1,$FE,$60,$6F,$07,$FB,$0E,$B7,$7C,$7D,$4E,$1E,$54,$8C,$1F,$CD
	.dc.b	$3E,$C5,$5D,$7D,$E9,$39,$8F,$E7,$92,$FA,$9F,$E7,$DF,$4E,$F8,$2A
	.dc.b	$6A,$37,$F2,$9A,$FA,$39,$8F,$78,$FE,$8D,$F6,$0F,$B1,$E7,$E6,$4A
	.dc.b	$FE,$8F,$39,$7F,$A4,$FD,$8B,$E5,$BF,$A7,$7D,$77,$F3,$49,$36,$75
	.dc.b	$1F,$D4,$3E,$07,$FD,$48,$F2,$F8,$FF,$EA,$D2,$7F,$AC,$4E,$A7,$FA
	.dc.b	$3A,$F9,$0C,$FC,$34,$9E,$91,$C3,$CC,$F1,$FA,$32,$7C,$11,$FD,$8F
	.dc.b	$EC,$5F,$49,$B9,$C9,$FD,$92,$32,$D8,$E0,$47,$FB,$4E,$8F,$8F,$FE
	.dc.b	$DD,$F6,$2E,$CF,$87,$72,$FF,$B7,$CE,$5B,$39,$7F,$B1,$DF,$81,$FE
	.dc.b	$CB,$3B,$6C,$F6,$23,$E2,$FB,$1F,$0E,$FF,$4B,$D3,$93,$E0,$8C,$BE
	.dc.b	$88,$CB,$DA,$3F,$C0,$DF,$62,$EB,$FF,$05,$78,$4F,$AC,$FE,$49,$7F
	.dc.b	$E1,$08,$C7,$0F,$0A,$93,$FE,$0C,$BE,$4D,$8E,$BD,$9F,$E1,$3C,$9F
	.dc.b	$5F,$7C,$69,$7F,$E1,$49,$CB,$67,$81,$1F,$F0,$CF,$D8,$FE,$87,$3F
	.dc.b	$E0,$DF,$63,$FE,$0E,$9F,$E1,$13,$95,$19,$7F,$C2,$75,$FC,$35,$CD
	.dc.b	$D3,$C2,$E0,$E7,$8F,$F9,$01,$C0,$BE,$D7,$D6,$B8,$B6,$FB,$3E,$19
	.dc.b	$DC,$F8,$5B,$5F,$C7,$E3,$89,$E3,$AE,$2E,$EF,$C5,$7C,$DB,$1B,$AF
	.dc.b	$86,$D7,$CF,$7D,$9F,$44,$F0,$7C,$17,$D9,$A7,$23,$E0,$E1,$5E,$2F
	.dc.b	$F9,$CA,$BE,$74,$7F,$A1,$5A,$F9,$E3,$5E,$8B,$E4,$E9,$27,$FD,$13
	.dc.b	$5F,$D1,$23,$FC,$EF,$0F,$A8,$A4,$7F,$CF,$6D,$CF,$EE,$61,$F9,$9F
	.dc.b	$27,$CD,$5F,$D2,$6B,$D1,$39,$8F,$B2,$64,$FB,$6E,$6B,$F0,$D1,$97
	.dc.b	$AE,$CB,$EB,$3E,$53,$8C,$65,$EA,$66,$78,$CF,$0B,$F1,$D6,$E7,$BD
	.dc.b	$6F,$5F,$F4,$5C,$E5,$FF,$56,$D7,$0B,$E5,$5E,$07,$0F,$F9,$B3,$0F
	.dc.b	$71,$E3,$7F,$3D,$3A,$8F,$F4,$AD,$F9,$DB,$5F,$E8,$4A,$F8,$17,$E2
	.dc.b	$8F,$F4,$3F,$D8,$7E,$E1,$BA,$F1,$7F,$55,$1F,$DD,$DC,$92,$BF,$CF
	.dc.b	$F1,$C5,$A7,$13,$FB,$F3,$9B,$F4,$3F,$CD,$A7,$C3,$D1,$B9,$7D,$8F
	.dc.b	$FA,$4E,$72,$4B,$FF,$4C,$CE,$BE,$7F,$B4,$30,$E8,$EA,$D3,$A9,$79
	.dc.b	$F4,$F0,$73,$7F,$3F,$C7,$E1,$C6,$5A,$7F,$25,$8D,$61,$F0,$27,$4F
	.dc.b	$92,$3F,$D6,$F3,$C2,$31,$F0,$1E,$6F,$B1,$79,$5E,$8F,$93,$FE,$C9
	.dc.b	$7F,$EB,$89,$CB,$FE,$BA,$8C,$DF,$FA,$B1,$BA,$FE,$05,$B3,$FC,$7E
	.dc.b	$B8,$D6,$FA,$FE,$2F,$D8,$BE,$32,$9E,$18,$B7,$FE,$C8,$97,$D1,$F0
	.dc.b	$3E,$6F,$D8,$FF,$BA,$E7,$51,$F4,$3E,$C5,$F2,$54,$E1,$3F,$ED,$28
	.dc.b	$CB,$7E,$7F,$81,$C7,$9A,$C7,$FB,$63,$EC,$7F,$C9,$CD,$F8,$2F,$C2
	.dc.b	$79,$9F,$E1,$98,$C1,$FE,$DE,$3F,$C9,$E5,$F0,$7F,$15,$F6,$2F,$9E
	.dc.b	$FC,$2C,$EB,$8B,$FD,$AB,$F6,$3F,$9B,$69,$C3,$D3,$FC,$5C,$67,$9F
	.dc.b	$F5,$CF,$F9,$3A,$BF,$DD,$AD,$AF,$C9,$FC,$93,$D6,$E3,$F0,$5C,$BB
	.dc.b	$A6,$1F,$DB,$6B,$E5,$5F,$8A,$7F,$33,$E2,$BD,$5A,$7E,$8F,$EC,$7F
	.dc.b	$7F,$87,$EF,$A3,$7E,$84,$8F,$EF,$1F,$62,$FA,$AC,$65,$FE,$F2,$FB
	.dc.b	$1F,$F2,$7D,$7F,$9B,$F8,$3F,$84,$46,$27,$F5,$7F,$63,$FE,$4F,$AD
	.dc.b	$1F,$DA,$67,$55,$C3,$E4,$3F,$5F,$1F,$E8,$D9,$FF,$04,$53,$DF,$CD
	.dc.b	$69,$93,$FE,$0D,$FB,$1F,$C9,$3D,$B1,$FF,$41,$61,$C6,$31,$0F,$53
	.dc.b	$F0,$0B,$2F,$E4,$11,$87,$F8,$F6,$3B,$AE,$33,$97,$FC,$05,$F6,$3F
	.dc.b	$83,$7A,$B6,$FF,$85,$D7,$EE,$1E,$D3,$C3,$A3,$F8,$6F,$D8,$FF,$A4
	.dc.b	$E3,$CA,$D9,$FE,$27,$F6,$0F,$AE,$DF,$C7,$4B,$DB,$FC,$76,$75,$18
	.dc.b	$25,$72,$F3,$78,$36,$3F,$C6,$9F,$59,$F2,$23,$F9,$0B,$76,$FE,$7C
	.dc.b	$BE,$2F,$F9,$53,$9A,$FC,$03,$6B,$E5,$36,$8F,$19,$EF,$E6,$F9,$FC
	.dc.b	$0F,$47,$45,$9F,$F2,$36,$7F,$E6,$36,$CE,$9A,$6A,$7F,$9A,$FD,$8B
	.dc.b	$E6,$A9,$E6,$DA,$D7,$E0,$27,$4D,$9D,$34,$9A,$7F,$42,$FA,$EF,$BC
	.dc.b	$FF,$B2,$EB,$8B,$9F,$8A,$FA,$1F,$3F,$2F,$D3,$E1,$5A,$FB,$33,$78
	.dc.b	$3F,$B4,$6C,$AC,$5F,$CA,$F3,$FE,$6F,$D8,$FC,$E6,$C7,$D6,$6E,$3F
	.dc.b	$59,$B8,$7F,$77,$CF,$FA,$9D,$7B,$FD,$9F,$C2,$78,$1F,$19,$91,$93
	.dc.b	$E7,$C6,$30,$FC,$A9,$EF,$9F,$ED,$1F,$5C,$CB,$58,$FF,$80,$63,$18
	.dc.b	$7E,$2B,$EC,$7F,$0B,$E6,$FE,$19,$B1,$93,$F1,$F1,$96,$9F,$80,$F3
	.dc.b	$BF,$34,$7E,$6F,$D6,$B1,$92,$67,$FA,$48,$CB,$E8,$1F,$63,$FA,$27
	.dc.b	$0D,$CB,$F3,$7E,$07,$F7,$C8,$C5,$FF,$9E,$DB,$43,$9F,$77,$E8,$2B
	.dc.b	$83,$2B,$FD,$39,$CD,$FD,$E3,$0F,$F1,$46,$1C,$94,$66,$7F,$93,$72
	.dc.b	$7F,$BB,$DF,$CF,$FC,$87,$F8,$57,$D4,$7B,$5F,$BB,$89,$3A,$3F,$D5
	.dc.b	$53,$A7,$1A,$7E,$E3,$E9,$FF,$6A,$7E,$DB,$E3,$AC,$FE,$FA,$2C,$FE
	.dc.b	$FA,$6E,$7F,$E3,$FF,$B0,$7D,$27,$F8,$81,$B8,$7F,$2B,$F8,$7F,$F1
	.dc.b	$17,$D1,$66,$A4,$FF,$A8,$F1,$FF,$52,$5F,$07,$01,$C3,$B6,$BF,$EA
	.dc.b	$BF,$C2,$B8,$63,$FE,$51,$6D,$7F,$4D,$D1,$F4,$FF,$E3,$49,$CD,$FD
	.dc.b	$76,$A7,$36,$AE,$37,$FD,$0A,$77,$87,$F5,$28,$F4,$FC,$C7,$F9,$37
	.dc.b	$4F,$F2,$07,$95,$F0,$E2,$4F,$F4,$AA,$E1,$97,$FA,$57,$E1,$BF,$C4
	.dc.b	$B9,$FC,$BB,$F9,$5E,$4C,$C3,$87,$C6,$DC,$3F,$31,$8F,$F2,$97,$E3
	.dc.b	$57,$D1,$9D,$61,$F5,$62,$DF,$CE,$9B,$5F,$97,$BE,$44,$8C,$C7,$49
	.dc.b	$F4,$57,$56,$92,$3C,$71,$96,$C6,$6F,$C4,$FA,$8C,$1F,$80,$9E,$F7
	.dc.b	$FD,$FE,$B8,$92,$B8,$BF,$02,$F7,$5C,$1B,$7D,$27,$D9,$18,$79,$47
	.dc.b	$C1,$34,$FB,$4C,$9C,$1C,$DF,$C4,$AF,$65,$F8,$89,$3E,$93,$D9,$5C
	.dc.b	$8E,$91,$A7,$A1,$9A,$E9,$3C,$27,$53,$C2,$3A,$3E,$F7,$C7,$5E,$F6
	.dc.b	$9E,$6B,$F7,$57,$2B,$F9,$4D,$AE,$17,$E1,$6C,$66,$F9,$1D,$27,$CD
	.dc.b	$7D,$23,$C7,$3F,$BF,$8C,$4F,$E2,$6B,$E2,$D6,$89,$19,$7D,$C6,$1F
	.dc.b	$71,$93,$DC,$79,$5F,$1B,$4F,$1B,$E5,$AF,$75,$78,$A7,$99,$FC,$1D
	.dc.b	$C5,$FE,$6E,$B8,$1E,$17,$E2,$B6,$FE,$1B,$4F,$0D,$FC,$5B,$F0,$9F
	.dc.b	$67,$CC,$FE,$3C,$7E,$FF,$2F,$EE,$95,$FB,$D8,$CB,$4F,$E4,$51,$A9
	.dc.b	$FD,$CC,$FF,$B1,$BE,$B9,$F6,$7F,$EE,$99,$FE,$94,$23,$37,$FC,$92
	.dc.b	$31,$3B,$7F,$A8,$4E,$9F,$E5,$3F,$60,$FB,$1F,$E8,$B1,$86,$BF,$CC
	.dc.b	$63,$FD,$0D,$EA,$75,$78,$75,$77,$B7,$81,$E1,$BF,$23,$63,$31,$B8
	.dc.b	$C7,$5C,$8E,$4D,$7F,$9E,$FD,$0B,$EC,$7D,$FC,$0D,$BE,$0C,$E6,$BF
	.dc.b	$CE,$2D,$7F,$BC,$7D,$8E,$DC,$DF,$2A,$E2,$D8,$FE,$E9,$5C,$6F,$84
	.dc.b	$2E,$C6,$25,$F1,$BC,$69,$18,$7C,$71,$98,$7C,$B5,$89,$3A,$3D,$31
	.dc.b	$CF,$EC,$5F,$4B,$C7,$E5,$25,$7F,$A3,$A7,$2D,$9C,$BF,$E1,$7C,$15
	.dc.b	$CE,$64,$EC,$86,$FE,$3B,$D5,$F6,$2F,$16,$92,$7F,$D3,$B0,$F1,$71
	.dc.b	$1F,$E2,$58,$E0,$EA,$79,$CF,$4E,$BE,$BD,$52,$B9,$C6,$5E,$B1,$97
	.dc.b	$A4,$7C,$DF,$B1,$7D,$8F,$9B,$C6,$95,$FE,$B0,$9C,$B6,$72,$F9,$67
	.dc.b	$94,$48,$D1,$0F,$57,$8D,$F6,$FE,$C5,$EC,$5F,$62,$57,$FA,$FA,$72
	.dc.b	$D9,$CB,$FE,$47,$8F,$F3,$93,$93,$87,$D6,$75,$23,$CF,$CB,$AF,$18
	.dc.b	$BD,$1F,$92,$A5,$7E,$3F,$E3,$7F,$C5,$77,$E3,$EC,$F4,$19,$D6,$46
	.dc.b	$48,$7F,$96,$E7,$8B,$63,$6D,$3F,$DB,$7F,$60,$FB,$1F,$CF,$3F,$9D
	.dc.b	$8F,$F3,$23,$F9,$C7,$7D,$7A,$EC,$91,$82,$1C,$1E,$7E,$F4,$8E,$EF
	.dc.b	$6E,$BC,$47,$68,$BB,$7E,$D3,$0F,$CE,$CF,$FB,$C3,$37,$A9,$FB,$3D
	.dc.b	$BF,$2D,$3F,$EF,$46,$BF,$06,$3E,$51,$BE,$89,$1C,$E3,$2D,$21,$FE
	.dc.b	$83,$9E,$EF,$FC,$00,$FA,$F7,$D8,$FF,$03,$FB,$1F,$C0,$CF,$FA,$25
	.dc.b	$B9,$FF,$0C,$AD,$CF,$1F,$63,$87,$2C,$8C,$1F,$E8,$7A,$FE,$27,$F6
	.dc.b	$0F,$96,$FE,$21,$F6,$3F,$C5,$F9,$FF,$A1,$B6,$FF,$D1,$D5,$B7,$94
	.dc.b	$FF,$48,$87,$C2,$F2,$F1,$25,$68,$E3,$1F,$C7,$7E,$C5,$F2,$DE,$0F
	.dc.b	$1E,$56,$C4,$E9,$C2,$F9,$6F,$3F,$A3,$27,$31,$FC,$A7,$D7,$7D,$8F
	.dc.b	$F2,$BF,$B1,$FA,$FC,$3F,$CC,$9B,$BF,$F6,$0A,$DE,$1E,$15,$F1,$7C
	.dc.b	$7C,$A9,$39,$8F,$E6,$DF,$60,$AE,$DF,$63,$2B,$D5,$39,$6F,$B9,$FC
	.dc.b	$F7,$E9,$9E,$AD,$39,$C7,$4F,$96,$D7,$BF,$2F,$9C,$7F,$45,$FB,$17
	.dc.b	$C8,$73,$F1,$65,$7C,$99,$CB,$F2,$FE,$C5,$F2,$5F,$D3,$7E,$BB,$F9
	.dc.b	$FC,$FF,$B4,$DB,$CB,$FD,$5A,$B7,$3F,$4F,$E7,$1F,$63,$FD,$56,$7F
	.dc.b	$B8,$CF,$FB,$6D,$B5,$FD,$7A,$B7,$3C,$7E,$3B,$FA,$E4,$62,$3F,$DC
	.dc.b	$4D,$9E,$13,$AF,$2F,$CD,$E8,$4B,$F1,$47,$F6,$5F,$B1,$7D,$26,$E7
	.dc.b	$27,$E2,$63,$2D,$8E,$04,$7F,$1D,$A3,$EC,$7F,$B8,$7D,$8F,$E0,$27
	.dc.b	$35,$F1,$CC,$F5,$C8,$C9,$0F,$F7,$E4,$F1,$6C,$6D,$A7,$83,$EC,$7B
	.dc.b	$FC,$2F,$D3,$F5,$64,$FE,$86,$FE,$5F,$81,$FE,$34,$6C,$7F,$BF,$F3
	.dc.b	$FE,$2E,$FD,$9A,$E6,$FE,$A7,$EC,$5D,$BF,$E1,$3E,$27,$D6,$FF,$2D
	.dc.b	$A4,$7F,$86,$CC,$F1,$7F,$84,$6F,$8C,$F3,$BC,$11,$ED,$DB,$FE,$17
	.dc.b	$C5,$F5,$F7,$C7,$95,$FB,$B9,$CB,$67,$81,$1F,$E0,$1F,$62,$BF,$2C
	.dc.b	$3E,$C7,$DF,$39,$D1,$39,$C3,$FC,$B2,$76,$70,$69,$C5,$72,$BE,$AD
	.dc.b	$AE,$6F,$CC,$9D,$9D,$30,$FF,$32,$C6,$EF,$A8,$C3,$C8,$E2,$B8,$BE
	.dc.b	$1A,$E3,$1F,$E6,$E9,$DD,$73,$3C,$13,$AA,$E0,$79,$2B,$86,$1E,$86
	.dc.b	$E1,$99,$D9,$E1,$AE,$4F,$96,$B8,$3F,$23,$0F,$3B,$FE,$80,$35,$5F
	.dc.b	$E7,$F8,$CC,$76,$34,$7B,$FD,$17,$C9,$6D,$B1,$A9,$FE,$99,$39,$3F
	.dc.b	$CF,$50,$FA,$9A,$47,$FC,$FC,$DC,$FF,$3B,$87,$E6,$B8,$FD,$D5,$FE
	.dc.b	$96,$AB,$39,$8F,$A6,$7E,$12,$33,$B7,$F9,$DE,$FF,$4F,$39,$3E,$CD
	.dc.b	$71,$8C,$C7,$94,$CC,$F1,$D7,$E0,$DF,$4C,$3C,$75,$CE,$FC,$F7,$B8
	.dc.b	$F9,$95,$B7,$CA,$75,$9D,$D7,$28,$C5,$FF,$A5,$2F,$94,$7E,$D9,$FD
	.dc.b	$B4,$E4,$FE,$CD,$1E,$19,$C3,$FB,$C8,$B8,$7F,$A3,$AB,$8C,$F9,$0F
	.dc.b	$6C,$FF,$6F,$F9,$2F,$DC,$B7,$6D,$CA,$7F,$5F,$39,$24,$77,$3F,$67
	.dc.b	$1C,$9C,$1C,$4E,$6E,$6F,$FD,$46,$FF,$3A,$9F,$17,$47,$9A,$5F,$63
	.dc.b	$FE,$98,$9C,$92,$FC,$13,$AF,$9E,$EB,$7C,$6D,$1E,$0F,$3E,$61,$FE
	.dc.b	$13,$BF,$E4,$7D,$98,$C7,$CF,$97,$F8,$36,$BE,$25,$F8,$F9,$FF,$A3
	.dc.b	$46,$09,$5E,$28,$FF,$5C,$C6,$30,$FE,$59,$18,$9F,$87,$F6,$2F,$27
	.dc.b	$E6,$FC,$97,$FB,$7B,$0F,$F1,$04,$E5,$FF,$11,$46,$6B,$E5,$B7,$0F
	.dc.b	$F6,$5B,$67,$FD,$F3,$5C,$6B,$7D,$BF,$6B,$EC,$5F,$19,$4F,$0C,$5B
	.dc.b	$FF,$18,$4B,$E8,$F8,$1F,$82,$FB,$1F,$F4,$CC,$EA,$3F,$0B,$F6,$2F
	.dc.b	$92,$A7,$09,$FF,$1D,$46,$5B,$F3,$9C,$0E,$3C,$D6,$3F,$C8,$1F,$62
	.dc.b	$F8,$2F,$C8,$B6,$3E,$C7,$97,$FE,$20,$77,$3F,$96,$7F,$CB,$47,$F9
	.dc.b	$24,$BB,$FE,$1B,$EC,$5F,$55,$BF,$0F,$F3,$BF,$63,$F9,$CA,$FC,$BB
	.dc.b	$79,$BF,$B2,$5F,$F8,$A3,$0F,$F2,$A4,$E5,$FE,$A3,$48,$FF,$2D,$C6
	.dc.b	$BE,$03,$F3,$5F,$24,$F5,$BB,$C7,$83,$B2,$57,$E4,$EB,$E6,$E1,$E2
	.dc.b	$9F,$D8,$F8,$AF,$4A,$9F,$A5,$FB,$1F,$E4,$1A,$B1,$FE,$49,$3C,$3D
	.dc.b	$19,$1F,$9F,$FB,$17,$D5,$63,$3F,$DA,$FD,$7F,$F9,$66,$BF,$CE,$0D
	.dc.b	$8F,$D6,$FD,$8F,$F7,$39,$CF,$C0,$FF,$04,$AF,$F5,$2C,$70,$F9,$0F
	.dc.b	$DC,$53,$DF,$D4,$6D,$93,$FB,$1F,$B1,$FC,$9B,$E0,$8F,$E1,$FE,$A7
	.dc.b	$DD,$CB,$F8,$4B,$77,$FF,$0B,$60,$8D,$FC,$07,$F1,$6F,$B1,$FB,$B8
	.dc.b	$FF,$A2,$5B,$87,$F0,$6B,$FD,$96,$1F,$1E,$3F,$C1,$33,$A8,$C4,$3E
	.dc.b	$C7,$FD,$37,$7F,$E9,$36,$C7,$F1,$0F,$A9,$7D,$67,$F2,$69,$CD,$F2
	.dc.b	$6D,$FE,$19,$B7,$FE,$30,$BE,$33,$D7,$E3,$79,$1F,$C8,$63,$F7,$57
	.dc.b	$CE,$31,$E7,$7F,$25,$FB,$1F,$93,$87,$F2,$26,$D7,$F8,$ED,$7C,$56
	.dc.b	$FC,$E9,$97,$F2,$19,$DB,$FE,$C3,$76,$E6,$3C,$5F,$0D,$E3,$E9,$B3
	.dc.b	$FC,$A7,$4F,$F5,$2B,$67,$4D,$35,$3F,$D2,$BE,$C5,$F3,$34,$F5,$6D
	.dc.b	$6B,$FD,$81,$3A,$6C,$E9,$A4,$D7,$FA,$0F,$D6,$FE,$16,$3F,$67,$5C
	.dc.b	$67,$3C,$9F,$2F,$E7,$E5,$7A,$7C,$2B,$5F,$6E,$6F,$0F,$F9,$8F,$65
	.dc.b	$62,$FE,$37,$A1,$F4,$7E,$C7,$E9,$36,$3E,$63,$71,$FA,$8D,$C3,$F1
	.dc.b	$B9,$FF,$54,$AF,$7F,$B5,$F7,$FC,$0F,$8C,$C8,$C9,$F3,$E3,$18,$7F
	.dc.b	$6C,$AE,$4F,$F6,$2F,$B1,$65,$AC,$7F,$CF,$91,$8C,$3F,$01,$F6,$3F
	.dc.b	$89,$F3,$FF,$14,$D8,$C9,$F8,$58,$CB,$4F,$C0,$79,$FF,$95,$3F,$2D
	.dc.b	$EB,$D8,$C9,$32,$FD,$34,$65,$FF,$01,$FD,$8F,$E8,$1B,$DD,$F9,$BF
	.dc.b	$03,$FC,$1F,$39,$D3,$F6,$6D,$C3,$F9,$7B,$6F,$FD,$33,$D9,$FC,$49
	.dc.b	$E7,$3F,$E8,$4A,$E3,$87,$F8,$2F,$9F,$FB,$CD,$68,$FC,$4F,$BB,$FA
	.dc.b	$FF,$90,$98,$7F,$81,$7D,$67,$B9,$8B,$FE,$01,$57,$2F,$F5,$45,$6A
	.dc.b	$FF,$49,$3F,$B5,$FA,$79,$1F,$AF,$F8,$EB,$87,$ED,$E6,$E1,$FB,$79
	.dc.b	$B5,$FD,$33,$EC,$1F,$49,$FC,$53,$0C,$35,$FF,$50,$C6,$A7,$FC,$EB
	.dc.b	$08,$C3,$FE,$78,$6D,$FF,$21,$F8,$87,$3F,$F4,$9B,$F9,$07,$62,$DF
	.dc.b	$F9,$FA,$34,$FF,$8F,$3E,$C7,$9C,$FF,$62,$6C,$7F,$B3,$35,$7D,$3F
	.dc.b	$F9,$2A,$73,$27,$CA,$FF,$A4,$5A,$F3,$CF,$DC,$ED,$B9,$FF,$55,$8F
	.dc.b	$F4,$C3,$5F,$91,$E8,$E8,$98,$FA,$A3,$E2,$7D,$97,$F9,$72,$BF,$AF
	.dc.b	$6B,$FD,$01,$46,$7C,$99,$7C,$67,$FB,$15,$FF,$97,$1F,$E5,$2F,$D2
	.dc.b	$9D,$B9,$F8,$DB,$87,$59,$B7,$F7,$66,$D7,$18,$ED,$3E,$59,$CC,$FD
	.dc.b	$B8,$B7,$E2,$BE,$97,$C0,$95,$E1,$6C,$6A,$3D,$15,$D6,$72,$7E,$3D
	.dc.b	$EB,$7D,$9B,$1A,$9F,$31,$C9,$F8,$93,$B7,$51,$F9,$18,$C3,$B3,$C8
	.dc.b	$5A,$F2,$12,$7E,$21,$CA,$FE,$6C,$FC,$18,$EE,$66,$FC,$B1,$CE,$3D
	.dc.b	$47,$1B,$F0,$DF,$59,$FD,$3C,$62,$77,$1E,$49,$D3,$DC,$FD,$45,$7E
	.dc.b	$40,$93,$FA,$EB,$FE,$E3,$3F,$92,$9C,$BC,$27,$51,$FE,$B0,$DF,$F5
	.dc.b	$FE,$AF,$E0,$1B,$3F,$35,$B5,$EB,$3F,$61,$3B,$E0,$FF,$21,$56,$DB
	.dc.b	$3B,$AE,$38,$FB,$DB,$5E,$E8,$FE,$01,$5C,$1D,$36,$34,$2B,$C2,$F7
	.dc.b	$BF,$11,$1C,$12,$3C,$53,$CD,$FC,$3D,$78,$2B,$84,$69,$B1,$E2,$6D
	.dc.b	$78,$A7,$C2,$6D,$ED,$3B,$69,$BB,$F7,$CF,$82,$3D,$51,$88,$FB,$93
	.dc.b	$E3,$9E,$31,$B3,$BC,$75,$7C,$04,$7F,$83,$79,$9F,$C2,$4F,$B9,$F0
	.dc.b	$FF,$E0,$6A,$FE,$0D,$19,$7F,$0C,$FF,$18,$23,$FC,$DB,$EC,$1F,$63
	.dc.b	$FD,$3B,$0F,$F4,$59,$F4,$63,$3D,$1B,$F7,$7F,$CC,$13,$A1,$3A,$71
	.dc.b	$F6,$2F,$33,$FA,$1C,$23,$0F,$F4,$66,$BC,$23,$70,$EB,$A4,$9E,$0F
	.dc.b	$4F,$C7,$7F,$D8,$D1,$FD,$41,$AF,$95,$AF,$B1,$AF,$FA,$A3,$E8,$1F
	.dc.b	$65,$FE,$AA,$E0,$FF,$38,$C6,$27,$F9,$C3,$54,$7F,$9D,$BE,$CB,$FC
	.dc.b	$EE,$F0,$AE,$5B,$F5,$6D,$7E,$6E,$F9,$57,$1A,$CC,$A5,$F5,$98,$6C
	.dc.b	$78,$A3,$1F,$2F,$96,$46,$4F,$3C,$74,$FB,$27,$D2,$F3,$F9,$69,$5F
	.dc.b	$D4,$E7,$2D,$9C,$BE,$EC,$7E,$35,$73,$99,$3B,$21,$B3,$DD,$F1,$FE
	.dc.b	$CF,$B2,$78,$D4,$93,$FD,$8E,$1E,$36,$23,$90,$8E,$2E,$E7,$A4,$F5
	.dc.b	$EE,$ED,$D5,$2B,$A4,$65,$ED,$19,$7A,$DA,$7E,$8F,$D9,$3E,$CB,$D1
	.dc.b	$E2,$CB,$FE,$F1,$39,$6C,$E5,$E7,$3C,$A2,$46,$88,$78,$3C,$7F,$BF
	.dc.b	$F6,$5F,$A0,$E5,$FF,$5F,$BF,$E7,$E9,$FF,$03,$35,$F8,$6E,$F8,$7F
	.dc.b	$DF,$ED,$8F,$E4,$CE,$E7,$9F,$B3,$DB,$B3,$BF,$D6,$4B,$EA,$F8,$BE
	.dc.b	$C3,$11,$DA,$2D,$F9,$27,$F2,$BE,$07,$F9,$0A,$BC,$B9,$7A,$8C,$EB
	.dc.b	$23,$24,$3E,$DC,$F1,$6C,$6D,$A7,$F8,$97,$EC,$5F,$65,$FA,$47,$F4
	.dc.b	$31,$F8,$45,$D5,$93,$5D,$72,$79,$46,$5E,$91,$CF,$E0,$24,$77,$7B
	.dc.b	$75,$E2,$3B,$45,$D7,$FD,$E7,$82,$BE,$1F,$AB,$FB,$8D,$BF,$33,$87
	.dc.b	$F8,$F5,$AF,$C0,$8D,$B9,$E9,$E0,$61,$D1,$FE,$45,$D9,$39,$C5,$BF
	.dc.b	$F6,$3A,$61,$A7,$FC,$99,$F6,$0F,$B2,$FE,$17,$F6,$5F,$83,$BE,$0D
	.dc.b	$AF,$ED,$95,$B7,$8F,$B5,$C3,$92,$46,$0F,$F2,$C7,$D8,$BE,$5F,$F8
	.dc.b	$BF,$D9,$7F,$1A,$CB,$C8,$DA,$FE,$F1,$5B,$78,$FC,$3F,$0F,$12,$46
	.dc.b	$0F,$F3,$27,$D9,$3E,$5F,$AF,$91,$27,$31,$FC,$9A,$5F,$2F,$E2,$F9
	.dc.b	$09,$5C,$A3,$F9,$5F,$B4,$FB,$2F,$E5,$BF,$65,$F5,$39,$FE,$5D,$FF
	.dc.b	$83,$63,$1C,$1F,$36,$F8,$4F,$F9,$DD,$7C,$77,$2F,$2E,$4E,$63,$F9
	.dc.b	$CF,$D8,$AB,$BF,$D9,$4A,$FF,$40,$4E,$5B,$EF,$7F,$40,$FA,$8F,$EB
	.dc.b	$D1,$FE,$85,$6B,$FC,$8A,$53,$DB,$E5,$B8,$7C,$A4,$AE,$B1,$FD,$2B
	.dc.b	$EC,$9F,$29,$D7,$C3,$95,$FE,$91,$9C,$BF,$E9,$3F,$B1,$7C,$B7,$F5
	.dc.b	$2F,$AE,$FE,$87,$3F,$CC,$9B,$C9,$F2,$21,$5C,$1F,$F4,$E7,$C2,$3E
	.dc.b	$CB,$FA,$DC,$BE,$42,$9F,$D7,$A7,$55,$C3,$E4,$71,$72,$30,$74,$F2
	.dc.b	$FC,$9E,$84,$8C,$1F,$EA,$BF,$B2,$7D,$37,$5A,$93,$98,$FE,$DB,$3A
	.dc.b	$8C,$10,$F7,$7C,$7F,$F7,$2F,$B1,$FB,$F1,$FE,$B7,$DF,$83,$F0,$EF
	.dc.b	$5D,$72,$35,$1F,$24,$C3,$FD,$6A,$B9,$B6,$73,$C1,$F2,$ED,$F6,$5E
	.dc.b	$1E,$0F,$A5,$E9,$4C,$3F,$D9,$18,$FC,$B6,$CE,$5F,$AD,$3C,$5F,$F6
	.dc.b	$77,$D9,$3B,$3F,$C2,$39,$3E,$BB,$F9,$4D,$7F,$BB,$23,$1C,$3F,$E1
	.dc.b	$2B,$E1,$3C,$AF,$04,$7A,$F6,$7F,$86,$31,$7D,$85,$F1,$A5,$7F,$B7
	.dc.b	$E7,$2D,$9E,$04,$7F,$DC,$9F,$65,$F9,$EC,$FF,$D0,$1F,$65,$FC,$BE
	.dc.b	$7F,$D0,$53,$95,$19,$7F,$9A,$57,$F2,$58,$C5,$F8,$27,$6F,$F2,$9A
	.dc.b	$E5,$3C,$5B,$1F,$CB,$4E,$11,$CA,$BA,$47,$1C,$7F,$C8,$37,$D0,$F2
	.dc.b	$B8,$AE,$11,$EA,$9E,$95,$CA,$37,$86,$5A,$73,$3D,$66,$4A,$67,$0F
	.dc.b	$3B,$C2,$FB,$BC,$6F,$D2,$79,$EB,$95,$7C,$99,$FF,$3E,$B5,$D1,$FE
	.dc.b	$7B,$70,$FD,$0F,$41,$F2,$3F,$E7,$F8,$4F,$F4,$38,$C3,$63,$94,$F0
	.dc.b	$7F,$D0,$3F,$53,$FE,$82,$4D,$23,$A3,$E1,$5F,$E6,$7C,$8F,$85,$1F
	.dc.b	$EC,$39,$F8,$93,$98,$F9,$06,$4E,$3B,$7F,$3E,$CF,$F0,$D1,$97,$F1
	.dc.b	$13,$C5,$CC,$79,$0E,$37,$E0,$71,$1C,$2B,$A6,$7F,$EA,$1C,$3C,$71
	.dc.b	$8A,$FF,$54,$C6,$2F,$93,$63,$CE,$74,$AF,$F4,$85,$71,$8F,$F4,$6B
	.dc.b	$FD,$5B,$0F,$F4,$6B,$6F,$C1,$3E,$63,$FC,$CF,$3F,$0C,$DC,$7F,$BA
	.dc.b	$5F,$E7,$11,$F0,$6B,$BC,$E4,$FF,$06,$7C,$97,$EE,$9C,$4F,$FB,$C6
	.dc.b	$30,$FF,$9D,$1F,$F7,$A4,$6A,$11,$AA,$FE,$8C,$D9,$EB,$AF,$A6,$BA
	.dc.b	$14,$E2,$D9,$FE,$77,$3D,$5B,$3F,$E8,$DA,$7D,$97,$FA,$72,$B5,$D3
	.dc.b	$C4,$ED,$3A,$FA,$1C,$4F,$F1,$FA,$47,$7F,$3A,$4F,$F2,$3E,$67,$66
	.dc.b	$32,$F8,$15,$71,$FA,$95,$AB,$F3,$4E,$5F,$A1,$1F,$EB,$A7,$27,$5F
	.dc.b	$7F,$E0,$7D,$93,$CA,$F8,$5F,$29,$FD,$9E,$BE,$64,$E5,$F9,$D1,$9C
	.dc.b	$BF,$D4,$AD,$CB,$FB,$A3,$67,$FC,$8D,$5C,$6B,$7D,$9F,$3F,$EC,$BF
	.dc.b	$D4,$15,$F6,$5B,$3A,$8C,$7C,$5E,$FE,$92,$47,$DF,$FB,$2F,$F7,$E3
	.dc.b	$C6,$3E,$D7,$D9,$3E,$46,$9C,$30,$FF,$5D,$C6,$6B,$F1,$0B,$E7,$B8
	.dc.b	$9C,$B9,$AC,$FE,$37,$EC,$9F,$05,$F8,$76,$C7,$F5,$9F,$37,$FC,$3E
	.dc.b	$EE,$7F,$18,$FF,$95,$CF,$CB,$17,$1F,$C5,$7D,$8B,$EC,$B7,$E1,$7E
	.dc.b	$4F,$EC,$BF,$3B,$5F,$ED,$26,$F3,$7E,$CE,$FF,$C4,$F8,$7E,$86,$72
	.dc.b	$FF,$9D,$5F,$F2,$D3,$9F,$7F,$F2,$7F,$20,$F5,$FB,$47,$7E,$D9,$5F
	.dc.b	$DC,$EB,$E6,$E5,$E1,$9F,$D3,$78,$AF,$52,$9F,$9E,$FB,$2F,$E0,$18
	.dc.b	$7F,$A1,$67,$4F,$ED,$5F,$F2,$CB,$7D,$2F,$D2,$FD,$93,$EB,$3A,$C7
	.dc.b	$EC,$FE,$CB,$FC,$B5,$5F,$BD,$6E,$7F,$B1,$FB,$2F,$F2,$9C,$E7,$E0
	.dc.b	$78,$7C,$84,$8F,$F0,$64,$BE,$03,$17,$FE,$FC,$C3,$15,$D6,$7F,$6F
	.dc.b	$F6,$5F,$92,$E6,$FF,$27,$56,$AD,$1F,$E5,$1F,$57,$EF,$C6,$30,$FF
	.dc.b	$52,$46,$27,$F9,$46,$48,$FE,$1C,$66,$7F,$D5,$5F,$65,$F8,$17,$F8
	.dc.b	$D3,$6F,$F8,$75,$7E,$E5,$EC,$04,$6E,$FF,$CC,$9F,$64,$F8,$CF,$E3
	.dc.b	$8D,$C3,$FC,$5D,$F6,$4F,$AD,$CD,$49,$FE,$55,$5B,$C7,$F8,$F5,$EE
	.dc.b	$73,$5B,$9C,$F9,$B7,$1F,$E8,$D7,$FB,$F7,$FA,$47,$D7,$7C,$BC,$3F
	.dc.b	$A1,$36,$FE,$5A,$F8,$AF,$E4,$EE,$68,$D2,$46,$BC,$BF,$E5,$B8,$7E
	.dc.b	$8A,$FC,$FC,$B8,$F8,$6C,$B7,$2F,$82,$D9,$DC,$7F,$49,$FB,$27,$CC
	.dc.b	$FF,$43,$6C,$7F,$95,$73,$F8,$9A,$7F,$A4,$67,$37,$F5,$03,$88,$D5
	.dc.b	$FB,$36,$E8,$CC,$3F,$CC,$1F,$5F,$F8,$3C,$3E,$E6,$1C,$67,$FB,$44
	.dc.b	$E7,$06,$9F,$B4,$7E,$7F,$CE,$FC,$2F,$07,$E1,$B2,$FE,$FB,$96,$9E
	.dc.b	$BF,$3F,$E7,$FD,$97,$D0,$6C,$64,$F8,$11,$96,$9F,$1F,$D9,$F8,$C7
	.dc.b	$D2,$EF,$B1,$92,$6B,$F8,$A8,$CB,$F1,$F5,$FE,$C3,$3F,$E8,$1F,$AD
	.dc.b	$61,$AD,$3F,$C2,$11,$89,$FA,$5F,$65,$F8,$6F,$3B,$F0,$ED,$8C,$9F
	.dc.b	$5E,$32,$D3,$F4,$3E,$77,$E2,$0F,$C4,$FA,$D6,$32,$4C,$FF,$43,$19
	.dc.b	$7E,$F7,$D9,$7E,$8D,$BD,$BF,$98,$EF,$FF,$05,$4E,$74,$FF,$6B,$37
	.dc.b	$0F,$DC,$5A,$30,$FE,$8A,$B8,$DC,$AF,$C2,$FB,$96,$BF,$C2,$1C,$DF
	.dc.b	$DF,$DF,$DA,$FD,$67,$E8,$8F,$F0,$7F,$AA,$F6,$F1,$87,$F0,$09,$B5
	.dc.b	$FE,$19,$AD,$4E,$4F,$CF,$7D,$47,$EC,$8F,$D9,$FC,$75,$8C,$92,$BF
	.dc.b	$C4,$5D,$C9,$D4,$7E,$8F,$EC,$3F,$DE,$55,$FE,$77,$6C,$7C,$23,$F8
	.dc.b	$D3,$87,$0D,$7E,$3B,$C3,$81,$3F,$C8,$7E,$21,$CD,$FC,$86,$37,$F2
	.dc.b	$1F,$E3,$D9,$CF,$B3,$FE,$3D,$C3,$FA,$4E,$1C,$B1,$F0,$F9,$FB,$F8
	.dc.b	$4F,$E3,$19,$78,$23,$81,$FE,$58,$9D,$C7,$AE,$31,$0B,$FE,$AA,$FA
	.dc.b	$BA,$B3,$06,$92,$BE,$23,$E7,$D1,$E0,$F9,$E3,$0F,$A7,$AE,$57,$F9
	.dc.b	$93,$5F,$BD,$F0,$F8,$CB,$E8,$55,$D7,$DD,$5A,$BF,$F3,$36,$6A,$D4
	.dc.b	$7C,$E7,$E4,$72,$FF,$7A,$C3,$EA,$C6,$9B,$3E,$19,$D4,$7C,$19,$D3
	.dc.b	$67,$58,$E2,$FC,$11,$E2,$AE,$F3,$97,$B6,$1E,$7B,$F5,$D6,$C8,$FF
	.dc.b	$62,$8D,$63,$F6,$E3,$C0,$70,$23,$F0,$B1,$F3,$E3,$E2,$9C,$DF,$C1
	.dc.b	$3C,$E7,$6A,$E9,$3B,$6E,$1E,$89,$D1,$C6,$B9,$9F,$ED,$6B,$EF,$3E
	.dc.b	$D9,$EB,$3A,$3C,$F5,$D2,$32,$D8,$D5,$F8,$67,$F4,$17,$D1,$F0,$CF
	.dc.b	$D4,$3E,$24,$65,$E3,$7E,$76,$CF,$0C,$3E,$4B,$E9,$C3,$C1,$3F,$B2
	.dc.b	$AF,$CC,$92,$7F,$74,$D8,$FD,$D1,$F4,$0E,$11,$A3,$F1,$FC,$7F,$81
	.dc.b	$23,$E8,$8F,$D4,$46,$1F,$7B,$6B,$E1,$1F,$49,$C4,$F9,$CF,$E0,$AE
	.dc.b	$27,$E1,$4E,$E7,$B6,$2A,$D5,$7D,$09,$E2,$6A,$33,$3F,$3E,$7F,$8A
	.dc.b	$C6,$2F,$F8,$53,$6B,$DA,$71,$24,$FA,$9B,$7E,$A7,$BC,$64,$91,$97
	.dc.b	$BC,$70,$3D,$CE,$0D,$B8,$7E,$4B,$B7,$E0,$BD,$A3,$F4,$9F,$6A,$FA
	.dc.b	$7F,$E2,$ED,$9F,$B3,$48,$C3,$F7,$1A,$FF,$37,$FB,$07,$D8,$FF,$4D
	.dc.b	$8F,$C0,$C6,$35,$E6,$FF,$A3,$7D,$1B,$5B,$AF,$E8,$1F,$61,$F4,$DF
	.dc.b	$0B,$5F,$C5,$B5,$DE,$9E,$65,$1B,$D3,$F8,$03,$73,$FE,$DA,$D9,$6E
	.dc.b	$F7,$BF,$CC,$11,$FE,$69,$BE,$3D,$3F,$DD,$1F,$E4,$FF,$54,$F2,$71
	.dc.b	$97,$F4,$F8,$C5,$F1,$8E,$F7,$A8,$FF,$3A,$65,$FD,$8D,$F6,$C7,$A6
	.dc.b	$31,$1E,$E8,$F6,$7C,$FB,$EB,$F9,$4A,$30,$F7,$3F,$D0,$0E,$1F,$F2
	.dc.b	$87,$D8,$7F,$21,$CF,$F9,$14,$E5,$B5,$D2,$31,$8F,$F8,$4A,$B8,$3F
	.dc.b	$E1,$48,$7C,$66,$C9,$87,$84,$F2,$7C,$AE,$1F,$F3,$04,$F0,$74,$D7
	.dc.b	$31,$EC,$FB,$07,$D1,$F8,$FE,$4E,$57,$59,$CB,$67,$2F,$69,$E5,$12
	.dc.b	$34,$43,$C7,$F1,$BF,$2B,$EC,$3F,$05,$06,$93,$7D,$72,$C9,$F7,$4E
	.dc.b	$5B,$ED,$73,$F6,$E4,$60,$87,$FA,$06,$7F,$D3,$4F,$28,$F1,$4F,$DC
	.dc.b	$FB,$07,$D8,$79,$7E,$C2,$5F,$CD,$9C,$B6,$72,$F5,$9E,$51,$23,$44
	.dc.b	$3C,$5E,$07,$E2,$BA,$1E,$CD,$F1,$65,$7D,$B9,$CB,$67,$2F,$FA,$6E
	.dc.b	$3F,$C2,$6E,$4E,$1F,$55,$D4,$8F,$3F,$AE,$AF,$18,$BD,$7F,$DD,$69
	.dc.b	$5F,$9E,$F8,$E6,$9E,$7E,$A9,$8F,$E3,$E7,$2D,$9C,$BE,$59,$E5,$12
	.dc.b	$34,$43,$FD,$69,$F5,$EF,$B0,$FD,$6B,$F9,$5C,$BF,$C1,$0B,$AB,$26
	.dc.b	$BB,$25,$FE,$86,$74,$D9,$CB,$DB,$E0,$3A,$91,$E1,$D7,$5D,$1C,$E3
	.dc.b	$C7,$1F,$0F,$D8,$FE,$03,$B7,$F0,$6C,$BF,$C3,$8D,$D7,$FD,$13,$B2
	.dc.b	$B8,$63,$FB,$5A,$D2,$3E,$7F,$56,$9F,$F6,$B5,$71,$7E,$75,$67,$0E
	.dc.b	$EF,$F1,$2F,$AE,$7D,$87,$F1,$4F,$B0,$FC,$1D,$FF,$B7,$DB,$7F,$E3
	.dc.b	$1A,$DC,$FF,$07,$5E,$F7,$1E,$59,$39,$8F,$E3,$FF,$5E,$F9,$6F,$E4
	.dc.b	$3F,$61,$FC,$5A,$BD,$4D,$53,$B7,$F8,$A2,$F8,$7E,$3E,$24,$9C,$C7
	.dc.b	$F2,$CF,$B0,$7C,$BF,$6F,$22,$57,$F1,$D9,$CB,$7E,$5F,$C7,$E8,$C8
	.dc.b	$C1,$FC,$8F,$DA,$7D,$87,$F3,$8F,$B0,$FA,$32,$6C,$EA,$38,$7C,$6E
	.dc.b	$FC,$A9,$18,$3F,$97,$7D,$7A,$BA,$7D,$84,$9C,$C7,$F4,$69,$7D,$47
	.dc.b	$F4,$8F,$79,$F0,$D4,$D4,$6F,$E5,$F5,$F2,$73,$1E,$F1,$FD,$43,$EB
	.dc.b	$DF,$61,$CF,$C2,$95,$FD,$06,$72,$FF,$43,$FB,$07,$CB,$7F,$58,$FA
	.dc.b	$FF,$E7,$32,$F0,$BF,$CE,$91,$87,$FC,$ED,$5C,$27,$FA,$67,$D0,$3E
	.dc.b	$33,$FB,$0C,$BE,$4E,$9F,$D9,$27,$55,$C3,$E2,$31,$72,$30,$74,$F2
	.dc.b	$BC,$9E,$7C,$8C,$1F,$D6,$FE,$C1,$F4,$9D,$6A,$4E,$63,$FB,$9C,$EA
	.dc.b	$30,$43,$DD,$E7,$FF,$BF,$36,$7C,$9F,$F7,$AF,$AF,$77,$7C,$5C,$66
	.dc.b	$1F,$DA,$E7,$2D,$9C,$BD,$0F,$DD,$5F,$85,$C0,$9E,$2E,$1B,$C3,$FE
	.dc.b	$A3,$6E,$3F,$C0,$E7,$85,$FD,$7F,$B0,$F3,$70,$FD,$9F,$4E,$57,$96
	.dc.b	$32,$FF,$80,$72,$4E,$A3,$C1,$3F,$E1,$8F,$B0,$76,$FF,$88,$F4,$7D
	.dc.b	$67,$F2,$8B,$F6,$46,$38,$B8,$54,$9F,$F0,$75,$F2,$6C,$75,$ED,$FF
	.dc.b	$12,$E4,$FA,$EB,$E3,$4B,$FF,$0B,$4E,$5B,$3C,$08,$FF,$86,$FE,$C3
	.dc.b	$F5,$39,$FF,$27,$FB,$0F,$6C,$FF,$29,$9C,$A8,$CB,$F0,$AB,$E6,$46
	.dc.b	$2F,$C1,$3B,$7E,$85,$72,$9E,$2D,$8F,$A8,$70,$8E,$55,$D2,$38,$E3
	.dc.b	$FA,$BB,$E8,$79,$5C,$57,$08,$F5,$4F,$4A,$E5,$1B,$C3,$2D,$39,$9E
	.dc.b	$B3,$25,$33,$87,$9D,$E1,$7D,$DE,$37,$E9,$3C,$F5,$CA,$BE,$4C,$FE
	.dc.b	$44,$38,$76,$7F,$B2,$9C,$3F,$47,$D1,$FF,$3C,$B9,$6B,$CE,$3F,$A4
	.dc.b	$7C,$77,$FA,$69,$B1,$FE,$87,$FA,$8F,$F4,$42,$69,$1D,$1F,$0A,$FF
	.dc.b	$33,$CB,$F0,$A3,$FB,$54,$FC,$49,$CC,$7C,$83,$27,$1D,$95,$C6,$3C
	.dc.b	$06,$5C,$57,$39,$DC,$F3,$37,$9F,$E3,$AF,$BB,$F9,$08,$EE,$E6,$78
	.dc.b	$B9,$8E,$39,$FF,$A6,$E7,$2F,$FB,$DA,$FC,$18,$78,$AB,$C8,$4C,$3F
	.dc.b	$D1,$57,$E4,$7C,$B1,$FC,$26,$B8,$BF,$C4,$A3,$E1,$9C,$2B,$FD,$0D
	.dc.b	$1E,$6B,$F1,$47,$FA,$27,$E5,$FF,$70,$DD,$7B,$3F,$C5,$63,$FC,$30
	.dc.b	$E4,$95,$FC,$82,$38,$B4,$E2,$7F,$87,$DC,$DF,$A1,$FE,$7B,$3E,$1E
	.dc.b	$8D,$CB,$EC,$3F,$D2,$73,$92,$5F,$F1,$68,$7D,$17,$68,$61,$D1,$D4
	.dc.b	$66,$1E,$7D,$3B,$F3,$7F,$44,$C3,$E1,$46,$5A,$7F,$AF,$71,$FE,$4F
	.dc.b	$D9,$9C,$7E,$05,$72,$F8,$0C,$57,$D5,$C3,$FD,$6F,$87,$FA,$D7,$EC
	.dc.b	$1E,$57,$9F,$E5,$BF,$B5,$4E,$63,$1D,$92,$3F,$D7,$93,$AC,$FF,$A6
	.dc.b	$DF,$29,$CD,$7C,$1F,$B0,$7C,$65,$39,$5F,$F5,$48,$CD,$7F,$B1,$DF
	.dc.b	$F7,$53,$7E,$53,$E5,$7D,$87,$FA,$52,$30,$7F,$B3,$7E,$C1,$F2,$5F
	.dc.b	$3D,$B1,$AC,$3F,$B0,$C6,$1B,$F4,$1B,$38,$73,$DA,$FF,$6C,$7D,$87
	.dc.b	$F9,$49,$BF,$05,$F8,$2F,$37,$FC,$3F,$18,$3F,$DB,$C7,$F9,$4C,$BE
	.dc.b	$17,$E2,$3E,$C1,$F4,$3F,$83,$9D,$67,$FD,$D3,$E6,$FF,$32,$D3,$87
	.dc.b	$A7,$F8,$98,$CF,$4F,$EB,$1F,$F2,$9D,$7F,$BB,$5B,$5F,$D9,$BE,$49
	.dc.b	$EB,$71,$F8,$2E,$5D,$D3,$0F,$C4,$57,$CB,$BF,$14,$FE,$5F,$C5,$7A
	.dc.b	$B4,$FD,$0F,$D8,$7E,$E3,$0F,$DE,$46,$FD,$29,$1F,$93,$FB,$07,$D5
	.dc.b	$63,$2F,$F0,$8F,$D7,$7F,$95,$6B,$FD,$1D,$C1,$FC,$1A,$31,$3F,$DF
	.dc.b	$BE,$C3,$FC,$AB,$5A,$3F,$C2,$53,$AA,$E1,$F2,$1F,$AD,$8F,$F4,$E4
	.dc.b	$FE,$8E,$9E,$FE,$6B,$4C,$9F,$D4,$7D,$87,$E4,$31,$F2,$CF,$ED,$FD
	.dc.b	$5F,$BB,$1F,$E2,$86,$E5,$FC,$77,$04,$6F,$E0,$3F,$7D,$F6,$1F,$7B
	.dc.b	$2F,$46,$1F,$EA,$B8,$C3,$FE,$1A,$BF,$D8,$BF,$E3,$48,$CE,$1F,$CA
	.dc.b	$3E,$C3,$FD,$29,$7C,$A9,$18,$7F,$C4,$7F,$60,$FA,$DF,$E1,$D3,$9B
	.dc.b	$E5,$F1,$D8,$3F,$8C,$D6,$F0,$E9,$5B,$7F,$C7,$CD,$C3,$FC,$E5,$3F
	.dc.b	$BB,$7F,$C9,$31,$9B,$4E,$5F,$E8,$1F,$59,$F2,$DB,$E5,$FF,$28,$87
	.dc.b	$C5,$D8,$98,$FF,$19,$9E,$0F,$E1,$1D,$B9,$8F,$4F,$7F,$AB,$9E,$CF
	.dc.b	$F2,$1C,$7F,$CC,$ED,$9D,$34,$D4,$FF,$33,$FB,$07,$CC,$D3,$CD,$B5
	.dc.b	$AF,$A1,$3A,$6C,$E9,$A4,$DB,$FC,$AB,$F5,$DF,$86,$8F,$F3,$45,$71
	.dc.b	$9C,$F1,$FD,$2F,$9F,$95,$E9,$F0,$2D,$7D,$C9,$BC,$3F,$DA,$F6,$56
	.dc.b	$2F,$E3,$79,$FF,$37,$EC,$3E,$73,$63,$E6,$37,$1F,$F3,$4F,$17,$F2
	.dc.b	$C6,$D7,$C3,$F6,$A5,$FC,$BE,$EB,$5F,$E8,$8D,$BF,$AD,$CF,$C4,$AF
	.dc.b	$C5,$D6,$B4,$FE,$FF,$5A,$BF,$F3,$BF,$D7,$B1,$D6,$3F,$C0,$23,$0F
	.dc.b	$E0,$7E,$BF,$F0,$DE,$67,$E1,$DB,$19,$3F,$0D,$19,$69,$F8,$1F,$37
	.dc.b	$F3,$A7,$E7,$BD,$5B,$19,$26,$3F,$A8,$8C,$BF,$EA,$4F,$B0,$FD,$23
	.dc.b	$7B,$7F,$31,$DF,$FD,$E6,$73,$B7,$F8,$71,$B8,$7F,$A9,$5B,$7F,$ED
	.dc.b	$9E,$1F,$E2,$6F,$39,$FD,$D4,$EA,$9C,$FF,$DF,$6B,$47,$E2,$BD,$CF
	.dc.b	$DB,$FC,$84,$C3,$FB,$F7,$A8,$F6,$B1,$7F,$C0,$AA,$E3,$FE,$31,$AD
	.dc.b	$5F,$E9,$67,$F6,$5F,$51,$23,$F6,$1F,$1D,$70,$FC,$2C,$DC,$3F,$77
	.dc.b	$36,$BF,$D8,$5F,$61,$FE,$EF,$7F,$8E,$FC,$43,$9B,$F8,$CC,$6B,$E4
	.dc.b	$3F,$C5,$B3,$9F,$5B,$FC,$45,$7C,$73,$FF,$6B,$4E,$6F,$F8,$F4,$E7
	.dc.b	$2F,$F2,$EC,$E6,$BF,$DB,$33,$FD,$06,$BC,$7F,$29,$DB,$E1,$3F,$89
	.dc.b	$46,$0E,$71,$E0,$FB,$2F,$0B,$63,$BF,$5F,$9B,$5F,$D6,$C6,$0F,$6B
	.dc.b	$8C,$7C,$33,$BF,$07,$D6,$DA,$F4,$F6,$7B,$B8,$7D,$ED,$8C,$9E,$E8
	.dc.b	$CB,$4C,$DF,$BA,$32,$D8,$CE,$DF,$E6,$CD,$55,$A8,$C3,$67,$EB,$B6
	.dc.b	$82,$B8,$CF,$43,$65,$8D,$C6,$9F,$A3,$7E,$B6,$CF,$A5,$A7,$9C,$B7
	.dc.b	$E4,$8E,$91,$E1,$BF,$23,$63,$37,$DE,$BD,$D7,$DD,$B3,$DD,$B5,$DD
	.dc.b	$A7,$77,$C3,$1E,$C6,$9E,$CB,$F0,$DF,$79,$FC,$ED,$F7,$69,$DC,$B5
	.dc.b	$DD,$B7,$DC,$F0,$BD,$DB,$1D,$E3,$DB,$7E,$02,$38,$24,$79,$8F,$88
	.dc.b	$5B,$F3,$3F,$52,$FC,$CD,$9F,$33,$4F,$31,$6F,$CD,$1E,$83,$CC,$58
	.dc.b	$F3,$1D,$EF,$CA,$D8,$F2,$CF,$B2,$3C,$AD,$BF,$2C,$7E,$36,$FC,$AD
	.dc.b	$AF,$2B,$4F,$29,$6B,$CA,$F9,$E3,$CA,$D3,$CB,$1D,$EF,$CA,$D9,$F2
	.dc.b	$9E,$B9,$F2,$B6,$FC,$A2,$FC,$AD,$8F,$2B,$67,$CA,$DA,$F2,$C7,$9E
	.dc.b	$7C,$AD,$3C,$AF,$7B,$F2,$B4,$F2,$C7,$AC,$F2,$96,$BC,$B7,$F2,$3E
	.dc.b	$C1,$F6,$1F,$CF,$EB,$F7,$24,$8C,$F3,$62,$31,$ED,$7F,$95,$1C,$09
	.dc.b	$DB,$8F,$B1,$70,$6A,$3F,$82,$C6,$1C,$35,$E1,$1B,$5F,$63,$FC,$49
	.dc.b	$FF,$2B,$35,$6C,$D7,$D3,$39,$8F,$E7,$BE,$EF,$F9,$66,$B8,$EF,$EE
	.dc.b	$9E,$11,$FC,$62,$B8,$5F,$2A,$D3,$FC,$73,$95,$87,$3B,$F0,$CF,$83
	.dc.b	$A3,$70,$A7,$D5,$3C,$8C,$C6,$1D,$CF,$BA,$AE,$46,$92,$3F,$CE,$DF
	.dc.b	$7B,$F4,$F4,$F4,$C3,$E8,$1F,$65,$D1,$B1,$FE,$7D,$6B,$E8,$3F,$9B
	.dc.b	$FD,$97,$2A,$E7,$3A,$6D,$73,$3C,$F3,$BD,$7F,$0F,$4F,$8E,$F0,$13
	.dc.b	$0F,$07,$CB,$D7,$C0,$79,$E3,$18,$F8,$27,$E1,$FD,$93,$E9,$3D,$5F
	.dc.b	$29,$30,$FF,$4B,$4E,$5B,$39,$7A,$E2,$AE,$73,$27,$64,$37,$F1,$BF
	.dc.b	$4B,$EC,$BF,$03,$1F,$D6,$9A,$9A,$7D,$57,$FD,$50,$E3,$9B,$91,$CF
	.dc.b	$87,$A7,$B7,$D7,$DC,$97,$C2,$32,$F2,$8C,$BF,$1A,$D3,$F8,$6F,$B2
	.dc.b	$7D,$97,$9F,$EC,$65,$FF,$AE,$67,$2D,$9C,$BC,$E7,$94,$48,$D1,$0F
	.dc.b	$07,$D9,$7E,$53,$AD,$ED,$DF,$1E,$57,$FB,$1E,$72,$D9,$CB,$F7,$A3
	.dc.b	$FD,$16,$E4,$E1,$F5,$9D,$48,$F3,$FB,$0A,$F1,$8B,$CF,$F9,$EA,$57
	.dc.b	$EA,$3C,$46,$9E,$7E,$A9,$8F,$FB,$6E,$72,$D9,$CB,$E5,$9E,$51,$23
	.dc.b	$44,$3F,$2F,$F6,$2F,$B2,$FD,$BB,$F9,$F8,$FF,$74,$2E,$BC,$9A,$EC
	.dc.b	$93,$A8,$CB,$C2,$39,$FC,$0C,$8E,$6F,$6E,$BC,$47,$8A,$2E,$DF,$C0
	.dc.b	$30,$FC,$F4,$FB,$FD,$8F,$E1,$7B,$7E,$56,$3F,$5E,$DA,$FE,$0F,$B2
	.dc.b	$BE,$51,$C3,$A6,$5F,$32,$1F,$B8,$9E,$0E,$2B,$F8,$8D,$E2,$3F,$7B
	.dc.b	$F6,$0F,$B2,$FE,$3B,$F6,$5F,$82,$9F,$E0,$2D,$BF,$E3,$35,$B9,$FF
	.dc.b	$24,$C6,$3D,$EE,$3C,$B2,$73,$C9,$FD,$5A,$D5,$FC,$AB,$EC,$5F,$2F
	.dc.b	$FC,$B3,$EC,$BF,$8D,$5F,$F5,$C6,$E1,$FE,$56,$AD,$BD,$0C,$7C,$47
	.dc.b	$1F,$12,$46,$44,$A7,$F9,$CF,$D9,$3E,$63,$C1,$E3,$CB,$E1,$39,$6F
	.dc.b	$CC,$79,$7E,$42,$46,$0E,$DE,$93,$EC,$BF,$A2,$FD,$97,$D9,$DF,$FB
	.dc.b	$AB,$6B,$FC,$DF,$5B,$78,$FC,$7F,$0F,$2E,$46,$0F,$57,$D8,$FF,$6F
	.dc.b	$E7,$FE,$93,$83,$EC,$BB,$9F,$D5,$27,$59,$7F,$81,$17,$D5,$FF,$58
	.dc.b	$FA,$97,$C3,$53,$84,$71,$F9,$7D,$7D,$34,$D3,$C3,$1F,$D8,$3E,$C9
	.dc.b	$F2,$DC,$FC,$39,$5F,$3E,$72,$FD,$1F,$B2,$7C,$C7,$F6,$8F,$AF,$FE
	.dc.b	$85,$3F,$D3,$1B,$C5,$FD,$92,$B7,$3F,$67,$E8,$1F,$1D,$FD,$C2,$5F
	.dc.b	$2B,$4F,$EE,$73,$AA,$E1,$F1,$38,$B9,$18,$3A,$79,$7E,$3F,$46,$46
	.dc.b	$0F,$C4,$7D,$93,$E9,$BA,$D4,$9C,$C7,$F8,$0A,$75,$18,$21,$ED,$F9
	.dc.b	$3F,$F0,$3F,$D9,$3B,$BE,$15,$CA,$FC,$B4,$E5,$B3,$97,$E3,$4F,$28
	.dc.b	$91,$A2,$1F,$2F,$EC,$BB,$ED,$F3,$FD,$59,$3D,$A3,$2F,$E9,$78,$BF
	.dc.b	$89,$BF,$4E,$72,$FF,$80,$3E,$C9,$D9,$FE,$26,$E4,$7D,$77,$F2,$9A
	.dc.b	$FD,$94,$63,$87,$FC,$4B,$7C,$27,$F6,$4D,$7A,$5E,$8D,$F6,$7F,$8B
	.dc.b	$31,$7D,$85,$F1,$A5,$FE,$EE,$72,$D9,$E0,$47,$F8,$07,$D9,$7E,$83
	.dc.b	$3F,$F6,$EF,$D9,$7B,$A3,$12,$8F,$83,$7D,$23,$0F,$F9,$B0,$C4,$71
	.dc.b	$C7,$FC,$B3,$3A,$3F,$DD,$35,$D2,$79,$E1,$FE,$74,$6C,$F8,$AF,$FC
	.dc.b	$EF,$5B,$6D,$F4,$AD,$D7,$53,$FC,$49,$7C,$EB,$78,$73,$9F,$0D,$F1
	.dc.b	$9F,$45,$6E,$7F,$DF,$B1,$EA,$38,$BD,$70,$F1,$3A,$9E,$0F,$80,$EC
	.dc.b	$23,$6D,$AF,$86,$FF,$A3,$E3,$E6,$B6,$32,$F8,$0D,$7A,$3F,$E6,$F7
	.dc.b	$FD,$27,$3F,$D0,$BE,$3B,$FD,$20,$D8,$CC,$ED,$FF,$4A,$7D,$5F,$FA
	.dc.b	$3D,$34,$8F,$FA,$4D,$F4,$E5,$F9,$AF,$1B,$D7,$23,$89,$F3,$E3,$4F
	.dc.b	$D1,$35,$3C,$B6,$FF,$48,$EB,$FA,$89,$C9,$C6,$33,$AF,$FA,$BA,$B8
	.dc.b	$61,$E0,$AE,$F7,$FB,$0C,$3B,$BC,$E3,$37,$FE,$9F,$C3,$9C,$7F,$AE
	.dc.b	$F0,$E6,$F9,$4D,$9F,$DC,$A3,$6F,$FB,$12,$37,$5D,$E7,$FD,$53,$7F
	.dc.b	$E8,$96,$E1,$E7,$BF,$53,$FD,$A6,$7E,$19,$B7,$FB,$64,$7C,$0A,$E9
	.dc.b	$1F,$DB,$FE,$67,$F7,$2D,$DB,$AD,$7C,$F9,$FA,$13,$92,$47,$FA,$3A
	.dc.b	$38,$B4,$E2,$1C,$DF,$FA,$A9,$FE,$6D,$3E,$1E,$7E,$F2,$FB,$2F,$F4
	.dc.b	$A4,$E4,$97,$FE,$AB,$9D,$7D,$27,$6C,$39,$CE,$6D,$39,$F3,$E6,$1F
	.dc.b	$C8,$39,$9D,$98,$C3,$D5,$57,$2F,$9F,$5A,$BF,$8D,$39,$7F,$C5,$31
	.dc.b	$FE,$B3,$72,$75,$F8,$0F,$83,$F6,$4F,$2B,$E1,$FC,$BF,$F6,$2A,$FF
	.dc.b	$0F,$4E,$5F,$F1,$04,$67,$2F,$F5,$CB,$72,$FF,$68,$B6,$7F,$45,$2A
	.dc.b	$E4,$F0,$ED,$FA,$3F,$65,$FE,$9C,$AF,$F1,$63,$67,$51,$8F,$8C,$DF
	.dc.b	$CA,$C8,$FF,$19,$7D,$97,$FA,$51,$E3,$1F,$73,$EC,$9F,$25,$4E,$18
	.dc.b	$7F,$B4,$23,$35,$FE,$3B,$5F,$43,$C4,$E5,$CF,$67,$FC,$81,$F6,$5F
	.dc.b	$E4,$E6,$FC,$17,$E1,$3C,$DF,$F0,$CC,$60,$FF,$23,$1F,$E5,$22,$F8
	.dc.b	$1F,$8A,$FB,$27,$D1,$FE,$4A,$75,$A7,$FB,$93,$E7,$3F,$36,$D3,$87
	.dc.b	$A7,$F9,$88,$CF,$3F,$FC,$00,$7F,$CA,$15,$FE,$55,$6D,$7E,$4F,$E4
	.dc.b	$9E,$B7,$1F,$82,$E5,$DD,$30,$FE,$DF,$5F,$2E,$FC,$53,$FA,$6F,$19
	.dc.b	$EA,$D3,$F4,$7F,$65,$FB,$AC,$3F,$82,$46,$FD,$29,$1F,$DE,$BE,$C9
	.dc.b	$F5,$98,$CB,$FE,$00,$FD,$93,$E0,$FF,$66,$DB,$FF,$00,$FD,$97,$F9
	.dc.b	$39,$FF,$38,$46,$6B,$8F,$C8,$F0,$3F,$6B,$4F,$80,$C5,$F0,$7A,$9F
	.dc.b	$BA,$FB,$2F,$C7,$72,$A7,$F2,$F8,$F1,$7F,$CF,$DE,$B7,$DC,$C9,$5B
	.dc.b	$7F,$9F,$46,$63,$F9,$0E,$48,$FE,$1A,$66,$BF,$47,$F6,$5F,$80,$7F
	.dc.b	$D1,$4D,$BF,$E1,$15,$FB,$77,$BC,$F0,$E9,$FE,$2F,$F6,$5F,$E8,$F8
	.dc.b	$FF,$49,$36,$7F,$88,$7D,$93,$EC,$37,$F1,$D2,$F6,$FF,$1A,$9D,$46
	.dc.b	$09,$5C,$BC,$EE,$0D,$8F,$F1,$9F,$D8,$7C,$87,$FD,$6A,$D8,$F8,$AB
	.dc.b	$E3,$3F,$93,$B9,$DC,$D2,$46,$BC,$CF,$27,$83,$E5,$E9,$B8,$FF,$90
	.dc.b	$B3,$FF,$2F,$36,$74,$D3,$53,$FC,$F3,$EC,$9F,$35,$4F,$1E,$D6,$BF
	.dc.b	$D7,$53,$A6,$CE,$9A,$4D,$7F,$9E,$7D,$87,$DB,$7F,$BF,$57,$17,$3F
	.dc.b	$17,$F3,$7E,$7E,$5F,$A7,$C3,$B5,$F5,$A6,$F0,$FF,$68,$D9,$58,$BF
	.dc.b	$9B,$E8,$7C,$AF,$B2,$F9,$6D,$8F,$A6,$DC,$7F,$A8,$71,$A3,$18,$FE
	.dc.b	$5E,$5A,$7F,$50,$BF,$83,$ED,$7E,$0F,$C1,$F9,$0C,$8C,$9F,$46,$31
	.dc.b	$5F,$94,$3E,$0C,$FF,$64,$FB,$06,$7A,$C7,$FC,$07,$18,$C3,$F1,$3F
	.dc.b	$61,$F8,$4F,$3B,$F0,$AD,$8C,$9F,$8E,$8C,$B4,$FC,$17,$9F,$F9,$93
	.dc.b	$F3,$5E,$BD,$8C,$93,$3F,$D1,$C6,$5F,$40,$FB,$2F,$D0,$B8,$6E,$5F
	.dc.b	$E1,$AE,$07,$81,$FE,$06,$9C,$E1,$FD,$71,$B7,$FE,$3F,$6D,$FE,$7A
	.dc.b	$B8,$77,$4B,$FF,$4D,$F3,$FF,$77,$C3,$C1,$3F,$E0,$C8,$C5,$FE,$0F
	.dc.b	$E4,$7F,$6C,$7F,$84,$3D,$57,$B7,$8C,$7F,$49,$37,$1F,$F0,$85,$6A
	.dc.b	$72,$7E,$E3,$EA,$3F,$66,$7E,$D3,$E3,$EC,$64,$95,$F8,$A8,$CB,$F2
	.dc.b	$3E,$C7,$F9,$86,$5F,$8B,$6D,$7F,$97,$FE,$27,$FC,$55,$E2,$FF,$8B
	.dc.b	$27,$37,$FE,$63,$6A,$5F,$65,$BF,$3F,$83,$C7,$C6,$98,$FF,$9A,$F2
	.dc.b	$F0,$7C,$5E,$31,$F2,$55,$9C,$D6,$CF,$F0,$8E,$3F,$C5,$6B,$C1,$D1
	.dc.b	$E5,$3F,$C9,$3A,$7C,$CF,$89,$C5,$F1,$AF,$45,$7A,$F7,$CC,$7C,$37
	.dc.b	$DD,$E8,$7E,$B6,$3B,$B5,$F8,$0D,$7E,$03,$71,$FE,$9D,$1F,$94,$F5
	.dc.b	$FF,$BD,$E1,$C8,$F1,$42,$32,$FF,$A3,$5A,$E4,$F1,$C7,$08,$C9,$FE
	.dc.b	$0A,$8F,$F1,$99,$B8,$AF,$33,$CD,$1B,$7F,$C7,$2F,$F5,$B8,$DC,$78
	.dc.b	$FC,$FD,$92,$33,$3B,$9F,$67,$C4,$7F,$9B,$A3,$1A,$FF,$13,$7F,$CD
	.dc.b	$F5,$DA,$74,$6E,$EC,$7A,$E2,$DF,$C7,$9B,$7C,$E6,$D7,$A2,$6D,$7C
	.dc.b	$79,$AF,$A6,$7A,$11,$F3,$BF,$26,$77,$7E,$69,$E5,$39,$38,$5F,$85
	.dc.b	$A7,$84,$B1,$C0,$EA,$7A,$5C,$BF,$76,$FB,$DF,$81,$B3,$AB,$F1,$56
	.dc.b	$E3,$C4,$F4,$BF,$13,$63,$C4,$D3,$C4,$7E,$0A,$7F,$D4,$6D,$AF,$C0
	.dc.b	$92,$7E,$DB,$63,$F0,$F3,$C1,$FB,$13,$FE,$DA,$FA,$5F,$C3,$F9,$5F
	.dc.b	$E2,$5F,$61,$1A,$AF,$F7,$5B,$AA,$E1,$1F,$16,$76,$FE,$76,$B9,$57
	.dc.b	$19,$CC,$FE,$8E,$72,$F1,$9E,$8F,$63,$07,$3A,$EE,$FC,$93,$47,$37
	.dc.b	$37,$FE,$C6,$BE,$EF,$F7,$D6,$E1,$E0,$AE,$07,$29,$E1,$7E,$2A,$DD
	.dc.b	$F4,$C3,$85,$6A,$FC,$87,$1C,$33,$3A,$6D,$79,$1B,$3B,$8E,$86,$4A
	.dc.b	$E4,$F2,$4F,$38,$CD,$F9,$27,$31,$C5,$A7,$18,$F8,$33,$C6,$33,$1C
	.dc.b	$4C,$CF,$4C,$3E,$01,$F5,$CB,$3E,$E8,$FC,$85,$7F,$9A,$5A,$7F,$0F
	.dc.b	$C3,$FC,$54,$FE,$57,$CB,$D5,$7F,$12,$6B,$FC,$85,$AF,$F2,$56,$CF
	.dc.b	$F9,$DF,$0F,$C3,$FD,$8F,$FA,$83,$4F,$BB,$5F,$66,$31,$7F,$82,$6D
	.dc.b	$70,$F6,$FF,$CF,$53,$F7,$6B,$84,$62,$7A,$FD,$97,$F7,$4A,$FE,$57
	.dc.b	$68,$C4,$7F,$30,$8E,$ED,$8C,$BF,$EE,$6F,$B1,$FF,$1C,$3F,$D0,$1F
	.dc.b	$F3,$53,$7D,$3F,$F7,$5F,$07,$F8,$EE,$77,$3F,$E1,$C6,$C6,$67,$7A
	.dc.b	$F5,$6D,$05,$70,$9F,$17,$66,$B9,$DF,$53,$9B,$46,$3C,$EE,$71,$CA
	.dc.b	$AE,$06,$92,$3F,$AD,$FD,$8F,$F5,$2C,$FF,$CA,$8F,$F5,$98,$FF,$1A
	.dc.b	$BF,$41,$BC,$3D,$A3,$53,$C3,$8B,$13,$2B,$AC,$79,$3E,$81,$F6,$34
	.dc.b	$93,$FD,$B3,$EC,$7F,$CC,$98,$7F,$99,$6B,$73,$FD,$C1,$AF,$F9,$21
	.dc.b	$B1,$FE,$6B,$9E,$3B,$FC,$BA,$53,$E3,$7C,$44,$8E,$9F,$2D,$5E,$91
	.dc.b	$70,$F1,$4F,$CC,$FB,$17,$D2,$FA,$FE,$5A,$61,$FE,$04,$9C,$B6,$72
	.dc.b	$F9,$E7,$94,$48,$D1,$0F,$67,$C7,$7D,$AF,$B1,$78,$54,$93,$FE,$12
	.dc.b	$8F,$B8,$D8,$CB,$FE,$6C,$8E,$2E,$F8,$FA,$78,$BD,$7A,$64,$F3,$8C
	.dc.b	$BD,$63,$2F,$38,$FC,$6F,$D8,$BE,$C7,$CD,$E1,$4A,$FF,$12,$CE,$5B
	.dc.b	$39,$7C,$93,$CA,$24,$68,$87,$A7,$EC,$7F,$37,$DA,$F7,$EF,$BD,$2B
	.dc.b	$FC,$67,$39,$6C,$E5,$FF,$46,$C7,$FA,$E9,$C9,$C3,$EB,$3A,$91,$E7
	.dc.b	$F5,$F5,$E3,$17,$9B,$FD,$71,$4A,$FD,$7F,$8B,$FD,$8A,$73,$D5,$E8
	.dc.b	$33,$AC,$C7,$CF,$19,$7F,$C9,$75,$C6,$3F,$30,$D8,$FE,$08,$D4,$DF
	.dc.b	$B0,$7D,$8F,$F0,$07,$F3,$D3,$FE,$56,$5D,$59,$35,$D7,$27,$51,$97
	.dc.b	$84,$73,$F2,$64,$73,$7B,$75,$E2,$3F,$D7,$37,$FE,$FC,$74,$78,$63
	.dc.b	$E2,$FB,$1F,$C5,$36,$FC,$BC,$FF,$B0,$1B,$B7,$FB,$8F,$4F,$F5,$EC
	.dc.b	$FC,$4A,$E1,$D1,$23,$04,$3F,$D9,$33,$C1,$B3,$DC,$FE,$47,$F5,$EF
	.dc.b	$B1,$FE,$4B,$F6,$3F,$83,$AF,$F6,$8B,$71,$FF,$7A,$56,$E7,$FC,$FE
	.dc.b	$BD,$FE,$3C,$B2,$73,$1F,$CC,$3E,$C1,$F2,$FF,$CC,$BE,$C7,$F8,$AF
	.dc.b	$37,$F1,$36,$DF,$F3,$FA,$DB,$FE,$8C,$5F,$15,$C7,$C4,$93,$98,$FE
	.dc.b	$7B,$F6,$3F,$CC,$30,$E9,$E3,$F7,$33,$E5,$6B,$E5,$FC,$5E,$64,$C7
	.dc.b	$FD,$33,$82,$7F,$A3,$FB,$8F,$B1,$FE,$93,$F6,$3F,$53,$3F,$E5,$4D
	.dc.b	$C3,$FA,$B5,$6D,$FF,$51,$AF,$8F,$E3,$E5,$49,$CC,$7F,$54,$FB,$05
	.dc.b	$76,$FB,$19,$5F,$EA,$C9,$CB,$7E,$03,$FA,$DF,$C0,$3E,$12,$9C,$E3
	.dc.b	$A7,$C0,$6B,$E0,$E5,$F3,$8F,$EC,$7F,$62,$F8,$0E,$7E,$1C,$AF,$F5
	.dc.b	$DC,$E5,$FF,$5E,$FD,$8B,$E5,$3F,$B5,$FD,$6F,$F3,$C9,$78,$6E,$5F
	.dc.b	$E9,$D3,$98,$C4,$FF,$7B,$BE,$4F,$FB,$23,$E7,$5F,$63,$FD,$DA,$5F
	.dc.b	$21,$4F,$EF,$33,$AA,$E1,$F1,$38,$B9,$18,$3A,$79,$7E,$6F,$46,$46
	.dc.b	$0F,$F6,$BF,$D8,$BE,$93,$AD,$49,$CC,$7F,$82,$27,$51,$82,$1F,$03
	.dc.b	$6F,$7F,$D5,$7F,$83,$BE,$C7,$F0,$13,$FE,$E4,$E0,$E2,$EC,$FF,$0E
	.dc.b	$38,$E0,$F8,$B0,$9E,$2E,$CC,$76,$FD,$0C,$A4,$7F,$BC,$27,$75,$C9
	.dc.b	$B8,$FF,$7F,$6C,$FF,$8A,$DA,$B0,$7D,$8F,$A7,$87,$EE,$FA,$72,$BB
	.dc.b	$46,$5F,$64,$65,$F3,$5F,$D8,$7F,$C7,$18,$FB,$3E,$C5,$DB,$FE,$3C
	.dc.b	$87,$D6,$FF,$32,$AD,$E9,$C2,$31,$1F,$06,$F8,$BA,$3F,$C8,$8E,$1D
	.dc.b	$5F,$F8,$FA,$31,$7E,$1E,$DF,$F2,$0E,$4F,$8F,$BE,$44,$C7,$A0,$9D
	.dc.b	$38,$6B,$F4,$DB,$7F,$61,$B9,$FF,$8B,$7E,$C5,$8F,$E9,$2D,$F6,$3D
	.dc.b	$5F,$F2,$E4,$EB,$0F,$F2,$F4,$E6,$7C,$59,$7F,$8D,$E3,$18,$F8,$9D
	.dc.b	$5F,$1A,$EB,$3E,$1C,$3C,$15,$C2,$F9,$4F,$F9,$04,$E4,$6F,$0F,$0B
	.dc.b	$6F,$99,$FE,$71,$76,$D9,$E6,$FA,$AF,$BC,$F1,$BF,$04,$F4,$38,$14
	.dc.b	$61,$E9,$8F,$5D,$72,$9D,$9E,$23,$E1,$9C,$67,$85,$79,$E3,$FC,$B8
	.dc.b	$74,$9F,$A0,$E2,$7E,$33,$63,$2F,$D7,$F4,$7F,$D0,$EF,$23,$CA,$E1
	.dc.b	$E4,$D7,$87,$C7,$6D,$FF,$36,$7C,$8D,$7F,$D3,$0D,$4D,$23,$ED,$CF
	.dc.b	$F4,$BA,$7C,$58,$E2,$67,$7F,$E5,$3A,$7E,$16,$32,$FE,$1E,$B8,$C7
	.dc.b	$C7,$33,$A7,$FA,$CA,$B8,$BE,$DC,$3E,$95,$70,$BE,$97,$E6,$3C,$18
	.dc.b	$78,$DC,$47,$7C,$39,$C7,$D0,$BE,$71,$88,$F8,$67,$23,$C8,$72,$C3
	.dc.b	$C9,$7C,$0F,$D8,$E1,$EA,$3F,$91,$CF,$A8,$FE,$17,$5F,$D2,$A3,$FD
	.dc.b	$46,$D9,$E7,$3F,$C9,$CE,$07,$F5,$48,$D3,$FE,$CB,$FB,$1F,$E0,$53
	.dc.b	$FC,$01,$B1,$FC,$BD,$FF,$68,$46,$76,$E3,$29,$DD,$FF,$38,$6D,$7F
	.dc.b	$AB,$5E,$33,$CE,$B8,$C7,$F4,$F9,$F0,$73,$79,$2D,$F6,$3F,$EA,$A9
	.dc.b	$C9,$30,$FE,$91,$3A,$F9,$FE,$77,$C6,$D1,$DB,$CF,$98,$7F,$27,$BF
	.dc.b	$E5,$FD,$78,$C7,$DD,$56,$7E,$67,$32,$B1,$7E,$BA,$F3,$4F,$FB,$1E
	.dc.b	$3E,$44,$63,$E0,$3E,$2F,$D8,$BC,$AF,$8F,$F3,$3F,$DE,$AF,$F9,$6C
	.dc.b	$E5,$FE,$5F,$19,$CF,$FA,$F3,$73,$FE,$C6,$D9,$FF,$6C,$57,$1A,$DF
	.dc.b	$67,$D4,$FB,$17,$C6,$53,$C3,$16,$FF,$9E,$CB,$E5,$78,$1F,$73,$EC
	.dc.b	$7F,$D5,$53,$A8,$FB,$FF,$62,$F9,$2A,$70,$9F,$E8,$F1,$96,$FC,$E7
	.dc.b	$03,$8F,$35,$8F,$E9,$BF,$63,$FE,$67,$6F,$C1,$7E,$07,$CD,$FF,$19
	.dc.b	$46,$0F,$EA,$67,$F9,$A0,$BE,$0F,$E1,$FE,$C5,$F5,$5F,$94,$9D,$71
	.dc.b	$7F,$81,$FE,$C7,$F2,$3C,$DF,$E0,$C6,$CF,$F7,$4B,$FF,$1D,$3F,$D8
	.dc.b	$23,$34,$0B,$E0,$3F,$42,$D9,$FC,$67,$C9,$7F,$AC,$1E,$53,$AE,$B9
	.dc.b	$1A,$F9,$2B,$87,$F8,$6F,$C5,$7A,$9F,$AC,$6D,$7E,$5F,$EC,$7F,$81
	.dc.b	$3F,$CB,$27,$5E,$8C,$8F,$CE,$7D,$8B,$EA,$B1,$8F,$F8,$9F,$EB,$FF
	.dc.b	$CD,$75,$FD,$F1,$B1,$FA,$AF,$B1,$FE,$EB,$39,$F8,$3F,$DE,$D3,$E4
	.dc.b	$69,$FB,$5A,$7C,$0E,$2F,$A3,$C0,$FD,$D7,$D8,$FE,$4A,$32,$7F,$83
	.dc.b	$3D,$5F,$BB,$97,$F8,$ED,$BC,$FF,$C6,$F1,$47,$F0,$E3,$35,$F4,$E9
	.dc.b	$F6,$3F,$80,$8F,$F0,$BB,$63,$F8,$2D,$7E,$D5,$FF,$24,$46,$6F,$FC
	.dc.b	$9B,$F6,$3F,$EA,$AB,$FF,$0F,$B6,$BF,$86,$FD,$8B,$EA,$FF,$8F,$CE
	.dc.b	$6F,$8F,$C7,$E0,$FE,$31,$5B,$C3,$9C,$65,$FE,$21,$1F,$B7,$9E,$31
	.dc.b	$8F,$37,$F9,$07,$D6,$FC,$7C,$BF,$8E,$B6,$BF,$9F,$2F,$8C,$DF,$9B
	.dc.b	$30,$FE,$39,$3B,$7F,$C9,$EE,$DC,$C7,$9B,$E1,$BC,$FD,$16,$7C,$AD
	.dc.b	$C7,$E0,$B7,$1F,$E9,$D9,$FF,$AE,$6B,$FC,$8B,$F6,$2F,$9B,$FE,$86
	.dc.b	$D8,$F1,$ED,$89,$91,$C1,$F7,$C6,$31,$FF,$3E,$57,$17,$F9,$EF,$D6
	.dc.b	$FE,$01,$FD,$94,$67,$1F,$EC,$13,$95,$8F,$F1,$8A,$FB,$3D,$FF,$53
	.dc.b	$C2,$FA,$AC,$CF,$E9,$B9,$69,$F2,$BC,$EF,$A7,$F6,$3F,$51,$B1,$93
	.dc.b	$E7,$C6,$5A,$7C,$4F,$63,$E1,$9F,$84,$EF,$B1,$92,$63,$F8,$E8,$CB
	.dc.b	$F0,$A7,$FB,$17,$D8,$AF,$3A,$7F,$7F,$8C,$61,$F8,$0F,$B1,$FC,$4F
	.dc.b	$99,$F8,$A6,$C6,$4F,$C2,$C6,$5A,$7E,$53,$CD,$FC,$C9,$F9,$AF,$56
	.dc.b	$C6,$49,$9F,$E9,$23,$2F,$FA,$CF,$EC,$7F,$40,$DE,$DF,$CD,$F7,$FF
	.dc.b	$81,$A7,$3A,$7F,$AF,$1B,$87,$EE,$DB,$7F,$EE,$8E,$BF,$E2,$13,$FE
	.dc.b	$92,$9C,$C9,$DD,$39,$7F,$BB,$56,$F1,$FF,$05,$46,$2F,$F7,$7F,$23
	.dc.b	$FB,$63,$FC,$1D,$E9,$BD,$9C,$61,$FB,$69,$B5,$FE,$0E,$AD,$4E,$4F
	.dc.b	$D7,$FD,$47,$EE,$0F,$DC,$FC,$65,$8C,$92,$BF,$77,$19,$7F,$DB,$5F
	.dc.b	$61,$FC,$CB,$4F,$F6,$FB,$6B,$FA,$87,$C3,$7F,$8B,$7C,$4F,$F1,$7C
	.dc.b	$E6,$FF,$AA,$35,$2F,$B1,$DF,$9B,$C1,$E3,$E2,$CC,$7F,$D7,$B5,$FD
	.dc.b	$02,$FC,$73,$96,$CE,$BE,$23,$19,$F9,$BA,$3C,$67,$F9,$2B,$4F,$95
	.dc.b	$F0,$D8,$BE,$B3,$E8,$DF,$18,$FC,$07,$D7,$E6,$FF,$92,$BD,$5F,$ED
	.dc.b	$F3,$EA,$8C,$3A,$BF,$F6,$9C,$EB,$5F,$EA,$31,$89,$DB,$FE,$91,$9F
	.dc.b	$F2,$7C,$7F,$A8,$9A,$FF,$49,$6B,$FD,$F5,$F0,$FD,$A7,$C9,$9E,$87
	.dc.b	$FB,$89,$C4,$B4,$4F,$3F,$3B,$D5,$F0,$FE,$B6,$C7,$A1,$B5,$C6,$38
	.dc.b	$1E,$C9,$CF,$AC,$FA,$6F,$37,$99,$DA,$FF,$D4,$F1,$E9,$AC,$56,$E7
	.dc.b	$D5,$58,$AD,$5F,$FA,$C1,$E6,$D9,$F8,$17,$B7,$8C,$65,$FF,$0F,$46
	.dc.b	$5F,$F1,$05,$5C,$FF,$D0,$B8,$78,$67,$37,$C6,$73,$87,$82,$7D,$27
	.dc.b	$82,$75,$7C,$F0,$E8,$D8,$C9,$CE,$32,$D3,$33,$D0,$F3,$B8,$C3,$C9
	.dc.b	$87,$8A,$7D,$A6,$DC,$3E,$83,$15,$D9,$B5,$CF,$0F,$97,$1C,$70,$F3
	.dc.b	$36,$7C,$8D,$38,$97,$0F,$1D,$F5,$6C,$F5,$9F,$88,$6E,$FB,$CF,$AA
	.dc.b	$FB,$B6,$BB,$B6,$BB,$B4,$EA,$F3,$3A,$B9,$AE,$A7,$D5,$76,$66,$37
	.dc.b	$19,$36,$FC,$A8,$DC,$66,$36,$65,$DB,$F3,$63,$6E,$4D,$99,$8D,$E1
	.dc.b	$FE,$01,$9D,$CE,$4D,$C6,$67,$77,$F1,$1F,$14,$FE,$08,$F8,$05,$AF
	.dc.b	$63,$63,$75,$F8,$23,$A4,$76,$AE,$8F,$9C,$E0,$58,$E1,$1F,$6E,$7F
	.dc.b	$86,$7C,$4D,$AF,$E1,$8D,$7F,$D1,$0D,$A3,$FA,$BF,$96,$FD,$77,$D9
	.dc.b	$3E,$5F,$FC,$C3,$7F,$97,$9C,$C6,$32,$FC,$EC,$63,$5F,$D8,$36,$BF
	.dc.b	$49,$6F,$B2,$FE,$BD,$1F,$E6,$93,$FD,$39,$19,$EF,$C4,$63,$1F,$0B
	.dc.b	$B9,$D3,$A6,$E3,$FC,$05,$BA,$7F,$9D,$5B,$3D,$BC,$6F,$F3,$C5,$7F
	.dc.b	$07,$FB,$2F,$F6,$84,$FF,$AB,$78,$BF,$8F,$28,$FF,$2F,$39,$E1,$FF
	.dc.b	$40,$46,$23,$F9,$51,$FE,$B5,$79,$B8,$39,$79,$BF,$E8,$7A,$EB,$3C
	.dc.b	$5B,$EA,$E6,$FA,$D5,$BE,$9E,$0D,$EC,$7D,$57,$6C,$7F,$9A,$46,$2F
	.dc.b	$85,$FA,$6E,$E4,$69,$23,$D8,$7F,$B3,$7E,$CB,$EA,$61,$F0,$9F,$F6
	.dc.b	$74,$68,$3F,$52,$35,$C9,$E0,$8D,$CF,$C1,$6C,$EA,$39,$47,$1C,$FF
	.dc.b	$D5,$4E,$9F,$F5,$67,$D0,$BE,$CB,$D6,$DA,$F1,$35,$F9,$DF,$65,$F6
	.dc.b	$30,$FB,$33,$96,$D7,$23,$E2,$CE,$EF,$BF,$C7,$F6,$26,$1D,$BE,$6E
	.dc.b	$BF,$6F,$14,$E3,$1E,$F5,$F4,$3E,$C9,$F4,$BE,$8E,$C9,$8F,$C9,$9C
	.dc.b	$B6,$72,$F5,$9E,$51,$23,$44,$3C,$5F,$1F,$F7,$BE,$C9,$E2,$D2,$4F
	.dc.b	$D7,$8F,$C0,$36,$32,$FE,$56,$38,$BB,$E2,$E9,$E4,$F5,$E8,$93,$CE
	.dc.b	$32,$F5,$8C,$BC,$E3,$F2,$3F,$64,$FB,$2F,$37,$74,$AF,$C5,$4E,$5B
	.dc.b	$39,$7C,$93,$CA,$24,$68,$87,$A7,$EC,$BF,$3F,$D8,$F7,$6F,$BD,$2B
	.dc.b	$F3,$13,$96,$CE,$5F,$DB,$42,$79,$CF,$4F,$AC,$95,$C9,$EF,$F6,$18
	.dc.b	$8E,$31,$79,$BF,$84,$E0,$AF,$D9,$F8,$9F,$E1,$59,$CF,$57,$A0,$CE
	.dc.b	$B3,$1F,$3C,$65,$FD,$85,$71,$84,$48,$C3,$67,$93,$FC,$47,$EC,$5F
	.dc.b	$65,$FC,$11,$FC,$CC,$7E,$ED,$75,$64,$D7,$5C,$9D,$46,$5E,$11,$CF
	.dc.b	$CB,$91,$CD,$ED,$D7,$88,$F1,$45,$D3,$F9,$6E,$1F,$D4,$27,$DB,$EE
	.dc.b	$7F,$18,$DB,$F2,$F1,$FC,$A5,$B5,$FC,$D3,$55,$7C,$C3,$87,$44,$BE
	.dc.b	$64,$3F,$99,$4F,$07,$15,$FC,$FA,$F1,$1F,$CD,$FE,$C1,$F6,$5F,$CA
	.dc.b	$BE,$CB,$F0,$93,$FC,$85,$AE,$E3,$F9,$7C,$BC,$AE,$1C,$B2,$BF,$94
	.dc.b	$4E,$5F,$E5,$5F,$62,$F9,$8F,$E6,$FF,$65,$FC,$5F,$9F,$EE,$D2,$30
	.dc.b	$FE,$02,$B8,$3C,$BE,$2F,$8F,$89,$23,$07,$F3,$7F,$B2,$FE,$58,$FE
	.dc.b	$22,$DE,$3F,$80,$FE,$8F,$3A,$92,$F9,$7F,$2F,$99,$27,$A4,$7F,$4E
	.dc.b	$F7,$9F,$65,$FD,$43,$EC,$BE,$BE,$DE,$A6,$DF,$E6,$2B,$6F,$F4,$B5
	.dc.b	$F2,$1C,$7C,$99,$39,$8F,$EB,$7F,$62,$AE,$DF,$65,$2B,$FA,$A4,$E5
	.dc.b	$BF,$05,$FD,$8B,$E0,$9F,$0D,$4E,$71,$D3,$E0,$B5,$F0,$B2,$F9,$C7
	.dc.b	$F6,$AF,$B2,$7C,$07,$3F,$0A,$57,$F6,$39,$CB,$FD,$97,$EC,$9F,$2B
	.dc.b	$FD,$CB,$EC,$3F,$9F,$CF,$F8,$0D,$BE,$2F,$F8,$3A,$B7,$3F,$DB,$7E
	.dc.b	$75,$F6,$5F,$DE,$A5,$F2,$34,$FE,$FB,$3A,$AE,$1F,$17,$8B,$91,$83
	.dc.b	$A7,$95,$E3,$F4,$24,$60,$FE,$F9,$F6,$4F,$A5,$EB,$52,$73,$1F,$E0
	.dc.b	$C9,$D4,$60,$87,$B7,$EA,$FF,$C2,$3F,$63,$F7,$23,$FC,$11,$BF,$07
	.dc.b	$E1,$5E,$BB,$24,$6A,$3E,$41,$87,$FC,$55,$5C,$DB,$39,$E0,$FF,$1B
	.dc.b	$DB,$EC,$BC,$3C,$1F,$47,$D4,$98,$7F,$86,$71,$F9,$6D,$9C,$BF,$56
	.dc.b	$78,$BF,$E1,$FF,$B2,$76,$7F,$8C,$F2,$7D,$77,$F2,$8A,$FF,$14,$C6
	.dc.b	$38,$7F,$C6,$37,$C2,$79,$5E,$08,$F5,$EC,$FF,$1B,$E2,$F9,$0B,$E3
	.dc.b	$CA,$FF,$18,$4E,$5B,$3C,$08,$FF,$8D,$7E,$C9,$6C,$FF,$A3,$FD,$97
	.dc.b	$F2,$FA,$56,$A7,$38,$7F,$10,$3B,$4A,$B8,$46,$1E,$55,$C6,$75,$3F
	.dc.b	$C6,$23,$13,$C3,$0F,$E3,$6D,$AE,$E7,$F5,$A9,$DD,$71,$C3,$D1,$3D
	.dc.b	$0F,$EB,$F1,$E1,$3C,$D7,$D1,$E1,$87,$49,$F6,$D7,$2C,$3C,$11,$CC
	.dc.b	$FF,$4F,$DF,$17,$C5,$3B,$BE,$71,$B7,$C0,$F8,$A7,$83,$DA,$FC,$18
	.dc.b	$7C,$38,$FF,$4E,$35,$E6,$62,$FD,$AE,$BD,$07,$C6,$FF,$A8,$23,$EA
	.dc.b	$C6,$1F,$F5,$1D,$7F,$47,$9D,$36,$38,$0F,$8E,$AD,$32,$E0,$8F,$C3
	.dc.b	$C3,$F3,$BE,$57,$C9,$8F,$F0,$6C,$FC,$B9,$CC,$7C,$93,$27,$1D,$95
	.dc.b	$C6,$3C,$1A,$FE,$3E,$B8,$BF,$91,$8E,$2E,$6B,$8C,$75,$38,$EB,$FE
	.dc.b	$A7,$8C,$CF,$19,$C9,$FE,$B7,$8C,$CF,$FB,$23,$0F,$44,$62,$7F,$D9
	.dc.b	$D1,$8C,$3C,$CD,$8F,$54,$7F,$A7,$70,$F7,$9F,$EC,$28,$F3,$DF,$8E
	.dc.b	$75,$7E,$8A,$FF,$48,$36,$3C,$55,$D0,$FF,$61,$7D,$97,$EF,$63,$FD
	.dc.b	$54,$D8,$C9,$C4,$FF,$1E,$B9,$AF,$F6,$9B,$67,$FC,$0A,$D9,$D9,$C4
	.dc.b	$39,$6F,$46,$63,$F9,$F4,$FF,$18,$FB,$2F,$F4,$BB,$E4,$9D,$1E,$3B
	.dc.b	$4B,$E8,$33,$5C,$5E,$9E,$7C,$8F,$E4,$7C,$CE,$CC,$67,$F0,$AA,$E3
	.dc.b	$F2,$6B,$57,$F1,$27,$2F,$FB,$CE,$3F,$D5,$EE,$4E,$BE,$FF,$BF,$EC
	.dc.b	$9E,$57,$C1,$F9,$4F,$EC,$F5,$FE,$D7,$9C,$BF,$ED,$98,$CE,$5F,$EF
	.dc.b	$56,$E5,$FF,$00,$1B,$3F,$E3,$2A,$E3,$5B,$EC,$F9,$DF,$64,$F8,$BA
	.dc.b	$78,$62,$DF,$FB,$9E,$5F,$29,$C0,$FA,$DF,$65,$FE,$97,$9D,$47,$D9
	.dc.b	$FB,$27,$C8,$D3,$84,$FF,$BC,$63,$2D,$F9,$DE,$07,$1E,$6B,$1F,$EF
	.dc.b	$7F,$B2,$FF,$2A,$37,$E0,$FE,$E7,$9B,$FE,$1C,$8C,$1F,$EF,$D3,$FC
	.dc.b	$A8,$5F,$07,$F0,$5F,$62,$FB,$2F,$C7,$4E,$B8,$BF,$A7,$3F,$E5,$8F
	.dc.b	$B2,$FF,$4A,$BF,$E9,$6E,$6F,$DE,$CA,$72,$F0,$3F,$AB,$5F,$F8,$A2
	.dc.b	$B8,$1C,$E3,$F3,$92,$F9,$EA,$7E,$57,$E4,$9E,$BE,$2F,$8D,$FF,$A9
	.dc.b	$AF,$FC,$B5,$18,$8F,$17,$6F,$1F,$92,$EA,$FE,$A7,$C5,$7A,$B4,$FD
	.dc.b	$17,$D9,$7F,$04,$C3,$FC,$EB,$3A,$7A,$3F,$E5,$A6,$FA,$7F,$A9,$FB
	.dc.b	$27,$D5,$F5,$8F,$DB,$7D,$97,$F9,$6E,$BC,$8D,$D3,$E2,$71,$3E,$CB
	.dc.b	$FB,$DD,$6B,$E0,$F8,$FC,$94,$8F,$F0,$8C,$BE,$07,$18,$7C,$D1,$8E
	.dc.b	$6B,$BC,$FE,$F7,$EC,$BF,$35,$18,$3E,$37,$AD,$F8,$5B,$E4,$DC,$FF
	.dc.b	$90,$64,$8F,$E2,$06,$6B,$FD,$33,$F6,$5F,$87,$7E,$83,$6F,$DB,$5F
	.dc.b	$BC,$7C,$D1,$9B,$FC,$3F,$D9,$3E,$3F,$F8,$E3,$70,$FF,$17,$FD,$8B
	.dc.b	$EB,$33,$52,$7E,$F5,$6F,$1F,$97,$7B,$9C,$D6,$E7,$3E,$75,$C7,$F9
	.dc.b	$E5,$FF,$00,$7F,$9F,$7D,$77,$C6,$AF,$AA,$D8,$CF,$C6,$7F,$27,$8C
	.dc.b	$63,$FA,$16,$D7,$F8,$CD,$B4,$55,$C7,$CE,$FE,$61,$87,$E8,$EF,$D3
	.dc.b	$CB,$8F,$87,$CB,$72,$F8,$6D,$9D,$C7,$E7,$FE,$C9,$F3,$7F,$D1,$5B
	.dc.b	$1F,$E5,$8C,$FE,$36,$9F,$B6,$9C,$DF,$E2,$5B,$19,$BF,$76,$DC,$D9
	.dc.b	$87,$F9,$8B,$EB,$BF,$05,$80,$61,$CA,$7F,$82,$4E,$B0,$69,$FC,$8E
	.dc.b	$FE,$BF,$81,$F0,$BC,$3F,$86,$CC,$3F,$02,$E5,$A7,$AF,$CF,$FA,$5F
	.dc.b	$65,$F4,$DB,$19,$3E,$04,$65,$A7,$C7,$F6,$7E,$31,$F4,$BB,$EC,$64
	.dc.b	$98,$FE,$32,$32,$FC,$79,$FE,$D1,$F6,$4B,$CE,$9F,$E0,$78,$C6,$1F
	.dc.b	$47,$EC,$BF,$11,$E6,$FE,$25,$B1,$93,$EB,$46,$5A,$7E,$8B,$CE,$FC
	.dc.b	$39,$F8,$8F,$5A,$C6,$49,$9F,$E8,$A3,$2F,$F0,$5F,$B2,$FD,$23,$7B
	.dc.b	$7F,$35,$DF,$FE,$0B,$9C,$E9,$FC,$E9,$B8,$7D,$DB,$46,$1F,$CF,$57
	.dc.b	$1B,$95,$F8,$5F,$72,$D7,$F1,$EE,$5F,$F0,$0B,$FB,$6F,$91,$FD,$09
	.dc.b	$FE,$11,$F5,$5E,$DE,$30,$FE,$03,$36,$BF,$93,$D6,$A7,$27,$E7,$7E
	.dc.b	$A3,$F6,$87,$ED,$7E,$3A,$C6,$49,$5F,$BC,$8C,$BF,$CA,$7E,$C7,$F9
	.dc.b	$A6,$FF,$D6,$9B,$5F,$9D,$F8,$7F,$F1,$7F,$8B,$FE,$31,$9C,$DF,$E8
	.dc.b	$5A,$97,$D9,$6F,$CE,$E0,$F1,$F1,$A6,$3F,$D5,$EB,$FA,$0D,$F8,$E7
	.dc.b	$2D,$9D,$7C,$4E,$33,$F3,$74,$78,$CF,$F2,$66,$9F,$2B,$E1,$B1,$7D
	.dc.b	$67,$D1,$BE,$31,$F8,$0F,$F9,$2F,$D8,$FE,$EF,$8F,$F7,$A6,$DF,$E0
	.dc.b	$74,$FE,$99,$7F,$D3,$E7,$2F,$F0,$17,$FB,$FB,$F8,$73,$FB,$D1,$FE
	.dc.b	$6D,$34,$53,$33,$FE,$A3,$FB,$2F,$8E,$FF,$79,$6C,$77,$3F,$CC,$0E
	.dc.b	$2F,$FC,$1E,$D9,$FC,$B3,$67,$BC,$73,$DB,$DD,$F1,$72,$3F,$C0,$D8
	.dc.b	$7F,$5D,$6C,$66,$BD,$CF,$C5,$F3,$F1,$8F,$C5,$F2,$3F,$D3,$7E,$AB
	.dc.b	$E9,$BF,$4D,$97,$F2,$C8,$C6,$3F,$17,$1D,$D6,$AE,$CF,$C9,$6B,$FE
	.dc.b	$21,$6D,$FF,$A2,$6F,$9C,$E7,$2F,$F5,$CD,$F3,$6C,$73,$6B,$FE,$2B
	.dc.b	$C3,$9E,$12,$73,$87,$9E,$FD,$35,$9B,$FF,$67,$47,$56,$9D,$4A,$75
	.dc.b	$AE,$77,$C7,$0F,$3E,$1D,$6B,$E1,$56,$88,$F2,$8D,$3C,$CC,$57,$3B
	.dc.b	$F1,$5F,$4C,$31,$5E,$38,$EF,$5F,$22,$73,$3F,$1C,$E7,$87,$C0,$8C
	.dc.b	$E1,$E0,$23,$8A,$CB,$67,$CF,$7E,$16,$E1,$E4,$26,$1E,$E3,$B9,$27
	.dc.b	$CB,$3E,$13,$8C,$7A,$67,$8B,$F0,$67,$6F,$18,$CE,$1F,$83,$6E,$1E
	.dc.b	$6B,$E7,$5E,$FB,$E2,$79,$DE,$47,$72,$1D,$EB,$9C,$F2,$BE,$D3,$A7
	.dc.b	$E3,$5F,$CD,$9D,$3E,$19,$EA,$78,$6B,$CF,$87,$C9,$8C,$DF,$47,$47
	.dc.b	$5A,$E8,$EB,$E2,$BF,$42,$7D,$62,$E1,$F9,$87,$0F,$E7,$BE,$C7,$FD
	.dc.b	$0B,$A5,$F7,$7F,$CB,$17,$F8,$DF,$97,$FD,$6D,$6A,$3F,$32,$BE,$CB
	.dc.b	$FA,$8C,$7F,$19,$11,$9E,$EC,$46,$2F,$AB,$FC,$7E,$74,$F1,$5E,$2F
	.dc.b	$FA,$2B,$7F,$ED,$51,$AE,$0F,$DD,$B6,$33,$3B,$AF,$F1,$64,$F2,$7D
	.dc.b	$D8,$F9,$AF,$8B,$6F,$8C,$FF,$00,$FB,$2F,$F6,$9B,$FE,$A3,$7F,$CA
	.dc.b	$2D,$F5,$7F,$BE,$71,$FF,$AE,$27,$73,$FE,$8B,$6C,$66,$77,$BF,$F8
	.dc.b	$0E,$79,$D6,$FC,$DC,$BF,$E4,$38,$78,$18,$EC,$7D,$57,$4C,$3F,$B6
	.dc.b	$36,$F8,$57,$25,$7C,$9F,$54,$66,$31,$1F,$EA,$1F,$B2,$FE,$A1,$9F
	.dc.b	$FB,$49,$FE,$A3,$19,$E4,$E9,$1F,$62,$75,$CF,$83,$33,$2B,$FD,$95
	.dc.b	$5C,$AB,$FD,$BB,$F4,$2F,$B2,$F3,$36,$BF,$DA,$0D,$7F,$D6,$9F,$65
	.dc.b	$FE,$E4,$AE,$35,$C0,$93,$FE,$D7,$9D,$6B,$FE,$2F,$B7,$C8,$76,$25
	.dc.b	$76,$F9,$7A,$FF,$B0,$67,$6F,$8A,$7C,$B3,$F2,$FE,$C9,$F4,$DE,$8E
	.dc.b	$D9,$7F,$EE,$79,$CB,$67,$2F,$59,$E5,$12,$34,$43,$C5,$F2,$1F,$67
	.dc.b	$EC,$9E,$2D,$24,$FF,$BD,$63,$EC,$EE,$D7,$FC,$9D,$1C,$27,$55,$C1
	.dc.b	$FF,$6F,$47,$09,$DF,$93,$DF,$A6,$57,$58,$F1,$18,$78,$C6,$1F,$F4
	.dc.b	$C6,$1E,$17,$FD,$35,$F6,$4F,$B2,$F5,$76,$49,$DE,$3F,$E6,$3A,$D4
	.dc.b	$60,$E3,$5B,$7C,$F5,$F9,$B9,$DC,$7F,$BD,$DB,$7E,$18,$46,$2F,$B5
	.dc.b	$FC,$BF,$B2,$FD,$27,$5B,$DC,$BE,$EC,$C7,$C4,$7E,$BA,$36,$65,$65
	.dc.b	$DE,$95,$A9,$E3,$0F,$AD,$EE,$48,$CF,$D8,$57,$9E,$7F,$E8,$5A,$E1
	.dc.b	$1C,$AB,$CF,$5F,$BB,$F1,$BF,$C3,$38,$F8,$70,$F5,$99,$DA,$61,$E2
	.dc.b	$8C,$BE,$CA,$E3,$1F,$E9,$36,$C7,$FB,$45,$A9,$BF,$62,$FB,$2F,$E1
	.dc.b	$8F,$E9,$A7,$D2,$7F,$B6,$57,$4E,$8D,$F5,$49,$F9,$B3,$A6,$CE,$5E
	.dc.b	$62,$3F,$DE,$6F,$0F,$2F,$C2,$78,$BA,$AC,$64,$87,$29,$47,$C9,$F6
	.dc.b	$BF,$92,$6D,$FA,$1A,$FF,$78,$35,$F8,$11,$B7,$F0,$71,$93,$59,$A7
	.dc.b	$53,$E8,$8C,$3F,$3E,$3B,$9F,$CC,$FE,$C1,$F6,$3F,$CD,$7E,$CB,$F1
	.dc.b	$35,$F9,$06,$BF,$0A,$76,$FE,$51,$FE,$B4,$FC,$68,$FC,$3B,$78,$BB
	.dc.b	$79,$BF,$D0,$7E,$C9,$F3,$3F,$D0,$FE,$CB,$F9,$67,$37,$E9,$DB,$CB
	.dc.b	$FE,$D5,$AD,$E3,$FA,$25,$F1,$5C,$7C,$39,$39,$8F,$E9,$DF,$65,$FC
	.dc.b	$F6,$FF,$DB,$D8,$3C,$6F,$01,$AF,$2F,$7F,$31,$E3,$F3,$26,$5F,$B0
	.dc.b	$C5,$3F,$D6,$7D,$D7,$D9,$7F,$5B,$FB,$2F,$AC,$7D,$78,$CD,$FE,$E1
	.dc.b	$7C,$86,$FC,$89,$39,$8F,$EC,$9F,$62,$AE,$9F,$65,$2B,$F8,$0C,$E5
	.dc.b	$BF,$05,$FD,$AF,$E0,$9F,$0B,$4E,$71,$D3,$E5,$F5,$F0,$B2,$F9,$C7
	.dc.b	$F7,$3F,$B2,$7C,$0F,$3F,$0E,$57,$F1,$09,$CB,$FC,$4B,$EC,$9F,$2D
	.dc.b	$FD,$EF,$EC,$3F,$A6,$CF,$E9,$9B,$DF,$FC,$86,$B7,$3F,$C6,$BE,$7D
	.dc.b	$F6,$5F,$E0,$39,$7C,$8D,$3F,$C0,$D3,$AA,$E1,$F1,$98,$B9,$18,$3A
	.dc.b	$7C,$17,$8F,$D0,$91,$83,$F9,$4F,$D9,$3E,$9B,$AD,$49,$CC,$7F,$85
	.dc.b	$67,$51,$82,$1E,$DF,$91,$FF,$0D,$7D,$8F,$DC,$8F,$E6,$BB,$F0,$7E
	.dc.b	$15,$EB,$B2,$46,$A3,$E4,$18,$7F,$86,$D7,$36,$CE,$78,$3F,$A5,$5B
	.dc.b	$EC,$BC,$3C,$1F,$47,$D3,$98,$7F,$46,$C7,$E5,$B6,$72,$FD,$59,$E2
	.dc.b	$FF,$4C,$FB,$27,$67,$F8,$FB,$27,$D7,$7F,$23,$AF,$E5,$71,$8E,$1F
	.dc.b	$F1,$D5,$F0,$9E,$57,$82,$3D,$7B,$3F,$C8,$18,$BE,$42,$F8,$F2,$BF
	.dc.b	$AE,$CE,$5B,$3C,$08,$FF,$62,$FB,$25,$FF,$60,$87,$D9,$7F,$2C,$42
	.dc.b	$77,$A6,$27,$53,$FC,$CA,$F6,$F2,$C2,$9E,$37,$07,$4B,$E6,$E0,$ED
	.dc.b	$87,$8E,$FA,$D7,$16,$DF,$67,$C3,$7C,$9B,$5C,$6B,$C3,$5B,$6D,$FA
	.dc.b	$5F,$3D,$F1,$6D,$71,$9F,$83,$1E,$13,$8B,$E7,$9E,$0F,$AE,$78,$3F
	.dc.b	$0C,$EC,$72,$AF,$13,$BC,$32,$FC,$07,$FD,$53,$5F,$12,$3E,$59,$93
	.dc.b	$FD,$60,$E5,$F4,$FA,$2F,$8D,$E2,$D8,$DB,$5E,$73,$FD,$22,$7F,$D6
	.dc.b	$F1,$F7,$BE,$42,$A6,$91,$DB,$E8,$BF,$CF,$FA,$1E,$A9,$E4,$7D,$B8
	.dc.b	$CB,$F7,$4C,$E9,$F8,$88,$CB,$F8,$A8,$CE,$9F,$EB,$0D,$BF,$CA,$13
	.dc.b	$99,$F2,$1D,$63,$33,$FE,$D3,$C3,$C9,$18,$9F,$93,$18,$C3,$CD,$1C
	.dc.b	$AF,$C5,$5B,$8F,$CC,$D6,$DF,$41,$BB,$F2,$5E,$70,$FF,$4C,$5F,$F9
	.dc.b	$95,$FF,$2B,$57,$A8,$FF,$34,$C6,$67,$8C,$7F,$88,$A7,$E2,$9C,$1E
	.dc.b	$C7,$08,$FF,$2B,$FD,$97,$EF,$DB,$AF,$39,$FD,$AC,$FE,$DA,$72,$48
	.dc.b	$FF,$17,$C7,$16,$9C,$43,$9A,$FF,$37,$3F,$CF,$67,$C3,$D1,$DE,$5F
	.dc.b	$65,$FE,$A6,$9C,$92,$FF,$8E,$C3,$E7,$FB,$61,$CE,$73,$18,$F3,$E5
	.dc.b	$FF,$A1,$70,$FE,$4F,$D9,$8B,$F4,$D5,$9F,$9D,$CE,$AC,$5F,$F1,$0C
	.dc.b	$7F,$C1,$AF,$39,$FF,$62,$BF,$2A,$FD,$51,$8C,$3F,$CA,$7F,$64,$F2
	.dc.b	$7E,$27,$CB,$7F,$6A,$C3,$FD,$0F,$39,$7F,$D1,$31,$9A,$FF,$2D,$B7
	.dc.b	$3F,$F4,$B3,$67,$F9,$1D,$71,$AD,$F5,$FD,$1F,$B2,$7C,$65,$3C,$31
	.dc.b	$6F,$FD,$2F,$2F,$95,$E0,$7D,$9F,$B2,$FF,$4F,$4E,$A3,$EE,$7D,$93
	.dc.b	$E4,$A9,$C2,$7F,$D4,$31,$96,$FC,$EF,$03,$8F,$35,$8F,$F5,$3F,$D9
	.dc.b	$3E,$07,$F1,$0D,$8F,$B1,$E6,$7F,$88,$1D,$CF,$E3,$5F,$F2,$A9,$FE
	.dc.b	$B0,$2E,$FF,$D8,$BE,$C5,$F6,$5B,$F0,$FF,$29,$F6,$5F,$9B,$AF,$F4
	.dc.b	$3B,$79,$BF,$5D,$7F,$E2,$8C,$3F,$D7,$73,$97,$FA,$8A,$F7,$F9,$9F
	.dc.b	$99,$F9,$1F,$F5,$04,$F0,$8E,$FD,$72,$3F,$25,$3F,$33,$1F,$0C,$FE
	.dc.b	$9F,$C3,$7A,$54,$FD,$27,$D9,$7E,$EB,$0F,$E1,$31,$BF,$42,$47,$E7
	.dc.b	$7E,$C9,$F5,$58,$CB,$FD,$47,$F6,$3F,$E5,$7A,$FF,$6A,$B6,$3F,$C0
	.dc.b	$FF,$65,$FD,$DA,$73,$F0,$3F,$B9,$AF,$F4,$B4,$70,$F8,$FF,$DB,$53
	.dc.b	$DF,$D4,$6D,$93,$FA,$FF,$B2,$FC,$8B,$E0,$8F,$E0,$9E,$9F,$DB,$CB
	.dc.b	$FD,$6E,$DE,$6F,$E3,$D8,$23,$7F,$01,$FC,$2F,$EC,$BE,$DE,$3F,$EE
	.dc.b	$E6,$E1,$FE,$20,$BF,$D8,$61,$FE,$FB,$9C,$BF,$EF,$DF,$B2,$FF,$49
	.dc.b	$DF,$FB,$D9,$B7,$FE,$28,$FB,$07,$D6,$7F,$19,$9C,$DF,$1F,$8E,$C1
	.dc.b	$FC,$5E,$B7,$87,$38,$CB,$FC,$3E,$3F,$69,$3C,$63,$1E,$6F,$F1,$FF
	.dc.b	$B2,$F8,$B9,$7F,$8E,$1B,$5F,$CF,$D7,$C5,$6F,$CD,$98,$7F,$1B,$9D
	.dc.b	$BF,$E2,$AA,$47,$08,$D4,$F9,$FE,$17,$D1,$D3,$67,$CC,$DC,$7E,$13
	.dc.b	$70,$FE,$A1,$9F,$FA,$92,$BB,$1F,$E3,$2F,$B2,$7C,$CF,$F3,$F6,$C7
	.dc.b	$97,$6C,$4C,$8E,$38,$FF,$90,$2B,$6F,$C1,$AD,$3F,$EF,$A8,$CE,$1F
	.dc.b	$D8,$7E,$BB,$F0,$D7,$E1,$C3,$FA,$44,$62,$D4,$C7,$F5,$CF,$AB,$E7
	.dc.b	$FE,$C7,$85,$F6,$59,$8F,$D7,$72,$D3,$E7,$79,$FF,$5F,$EC,$BE,$C3
	.dc.b	$63,$27,$D3,$8C,$B4,$F7,$E2,$D3,$FA,$C4,$61,$B1,$F9,$1E,$FF,$95
	.dc.b	$EC,$CA,$D3,$F8,$F8,$C4,$FC,$58,$FE,$D1,$F6,$4C,$33,$A7,$F8,$1E
	.dc.b	$31,$87,$E1,$FE,$CB,$F2,$1E,$77,$E4,$5B,$19,$3F,$19,$19,$69,$F9
	.dc.b	$BF,$3B,$F4,$07,$E8,$7D,$6B,$19,$26,$7F,$AB,$8C,$BF,$52,$3F,$CF
	.dc.b	$3F,$65,$FA,$88,$C1,$FE,$82,$71,$97,$E8,$7B,$FF,$C1,$51,$8A,$E3
	.dc.b	$DD,$83,$F5,$53,$63,$E7,$CE,$EF,$F0,$7C,$BF,$DE,$E7,$37,$F8,$8F
	.dc.b	$73,$F7,$5F,$21,$27,$FC,$0D,$EA,$BD,$BC,$63,$FC,$16,$AE,$9F,$E3
	.dc.b	$3A,$D5,$FE,$FE,$7F,$79,$F5,$12,$3F,$6D,$F1,$B7,$0F,$DF,$4D,$C3
	.dc.b	$F7,$D3,$6B,$FD,$3B,$F6,$5F,$CD,$65,$18,$F3,$7F,$C6,$9F,$0D,$FE
	.dc.b	$2C,$F1,$BF,$C5,$D7,$BA,$D1,$F9,$05,$F6,$3A,$BD,$BC,$3C,$59,$5B
	.dc.b	$F8,$9B,$AF,$D4,$18,$7F,$0E,$7F,$D6,$ED,$73,$1C,$33,$F3,$C6,$BA
	.dc.b	$3C,$A7,$A3,$87,$FA,$D4,$67,$0F,$34,$FA,$B7,$DC,$7A,$5F,$F2,$57
	.dc.b	$AD,$FD,$DE,$BF,$D2,$ED,$7F,$D8,$6D,$7F,$D9,$6D,$7F,$BB,$1F,$E5
	.dc.b	$3F,$B2,$F8,$75,$FB,$49,$FE,$F9,$3F,$ED,$29,$CB,$F0,$5B,$8F,$B2
	.dc.b	$11,$8A,$F0,$C7,$5D,$64,$7F,$AB,$23,$A4,$7F,$9C,$5E,$2D,$9E,$F1
	.dc.b	$EC,$F8,$6D,$61,$FE,$6B,$8C,$5F,$37,$FC,$21,$3A,$3E,$7F,$AE,$FA
	.dc.b	$6F,$87,$E4,$61,$F8,$B3,$6B,$E2,$CD,$7F,$C3,$4D,$C3,$F9,$9D,$F3
	.dc.b	$9F,$97,$59,$A9,$87,$FA,$EF,$0E,$ED,$AD,$B5,$EA,$DC,$3A,$CF,$09
	.dc.b	$FF,$61,$3E,$69,$D5,$76,$9C,$9E,$88,$E5,$79,$C3,$CD,$3D,$8D,$95
	.dc.b	$CE,$1F,$E9,$56,$DF,$1C,$3C,$51,$AB,$C5,$F7,$39,$DF,$A1,$F6,$98
	.dc.b	$BD,$BD,$EF,$D4,$E7,$0F,$59,$1E,$77,$C2,$FA,$47,$13,$AE,$1E,$6A
	.dc.b	$E0,$D8,$E3,$18,$BF,$49,$23,$8B,$5F,$89,$1F,$A8,$C3,$F6,$AD,$3F
	.dc.b	$55,$7F,$DC,$9C,$4F,$FB,$3A,$30,$FD,$78,$C9,$AB,$F9,$BF,$13,$EB
	.dc.b	$24,$79,$67,$F6,$71,$87,$E2,$1F,$31,$C4,$FB,$4E,$73,$C6,$7F,$CD
	.dc.b	$D3,$AC,$3D,$96,$8C,$4F,$D7,$BE,$2F,$66,$9D,$6B,$F8,$3C,$F0,$9F
	.dc.b	$A0,$76,$BF,$05,$F5,$6C,$EE,$3E,$B9,$99,$ED,$5D,$2B,$64,$8F,$73
	.dc.b	$FA,$1A,$E4,$7D,$28,$E0,$7C,$F7,$0F,$C8,$8F,$E3,$DF,$64,$FB,$2F
	.dc.b	$EA,$93,$FC,$36,$31,$C3,$FC,$15,$7C,$BF,$F9,$A6,$75,$5C,$11,$04
	.dc.b	$F2,$7F,$52,$DC,$FF,$5A,$DC,$B8,$3F,$EB,$79,$FD,$F3,$5E,$35,$E6
	.dc.b	$AE,$2D,$AE,$3E,$17,$86,$7A,$BE,$3C,$BC,$97,$E4,$8C,$3C,$0F,$86
	.dc.b	$E1,$F5,$FD,$97,$FB,$65,$FF,$17,$BF,$E6,$86,$FA,$5E,$FE,$3F,$E7
	.dc.b	$53,$B9,$FE,$BC,$D8,$CC,$EE,$FE,$34,$F3,$AD,$F9,$79,$7C,$3D,$B7
	.dc.b	$B5,$F5,$5C,$F0,$F9,$F1,$8B,$E1,$6C,$18,$61,$B1,$EA,$8C,$1C,$3E
	.dc.b	$CB,$E2,$E7,$FD,$4D,$FE,$BB,$3F,$DD,$DF,$C0,$37,$D8,$E1,$CB,$23
	.dc.b	$FC,$A7,$3A,$AE,$55,$99,$FF,$00,$7D,$03,$EC,$BC,$ED,$8F,$F2,$DB
	.dc.b	$5F,$F0,$37,$D9,$7F,$65,$BF,$EC,$D3,$96,$D7,$23,$E0,$CE,$F6,$FC
	.dc.b	$95,$3E,$B3,$B9,$30,$EF,$F2,$F5,$F2,$C5,$C7,$BC,$FC,$CF,$B2,$7D
	.dc.b	$37,$A3,$B2,$61,$FE,$72,$9C,$B6,$72,$FC,$09,$E5,$12,$34,$43,$E2
	.dc.b	$FC,$7F,$DA,$FB,$27,$87,$49,$3F,$E8,$08,$FB,$6F,$F8,$AD,$DB,$98
	.dc.b	$DB,$9E,$3E,$9E,$3F,$5E,$99,$3C,$E3,$2F,$58,$CB,$D6,$3F,$1B,$F6
	.dc.b	$4F,$B2,$F3,$77,$4A,$FF,$49,$CE,$5B,$39,$7C,$D3,$CA,$24,$68,$87
	.dc.b	$AF,$EC,$BF,$37,$DA,$F6,$EF,$B9,$2B,$FD,$43,$39,$6C,$E5,$FF,$27
	.dc.b	$42,$79,$CF,$4F,$AC,$95,$C9,$EF,$F6,$18,$8E,$31,$79,$BF,$C6,$D8
	.dc.b	$2B,$F5,$FE,$37,$F8,$1A,$73,$D5,$E8,$33,$AC,$C7,$CF,$19,$7F,$D7
	.dc.b	$35,$C6,$3F,$94,$36,$3F,$95,$B5,$FE,$63,$3F,$BE,$FB,$17,$D9,$5E
	.dc.b	$BC,$75,$CB,$FF,$62,$D7,$E8,$5B,$39,$7F,$CE,$50,$9E,$51,$C3,$CA
	.dc.b	$EC,$47,$9F,$5D,$78,$C5,$D3,$FC,$B5,$7F,$DA,$67,$DB,$EC,$FF,$14
	.dc.b	$DB,$F2,$F5,$FE,$5F,$6E,$DF,$02,$7E,$59,$BE,$89,$7C,$88,$7F,$A2
	.dc.b	$77,$4F,$08,$FF,$71,$DE,$1F,$F7,$2F,$D8,$BE,$CB,$F9,$2F,$D9,$7E
	.dc.b	$0E,$7F,$DD,$2D,$7F,$D2,$93,$B7,$8F,$95,$C3,$96,$46,$0F,$F7,$8F
	.dc.b	$D9,$3E,$5F,$F9,$8F,$D9,$7F,$15,$DF,$FB,$13,$6B,$FB,$2D,$6D,$E3
	.dc.b	$F1,$1C,$3C,$49,$18,$3F,$DF,$BF,$65,$FC,$CB,$8F,$FD,$0D,$83,$C7
	.dc.b	$EE,$7F,$40,$9D,$65,$FD,$C5,$7C,$BF,$93,$CB,$98,$F4,$8F,$E8,$FE
	.dc.b	$DB,$EC,$BF,$A4,$FD,$97,$D4,$AF,$EF,$6D,$53,$B7,$FB,$FE,$9F,$E9
	.dc.b	$98,$7C,$6F,$2F,$26,$56,$B4,$FF,$02,$56,$A3,$07,$F8,$23,$EC,$7A
	.dc.b	$B8,$18,$BE,$CB,$CA,$7F,$5E,$9D,$CF,$FB,$2D,$7B,$FF,$D8,$BE,$A1
	.dc.b	$F0,$74,$E1,$1C,$7E,$5B,$5F,$09,$30,$F0,$60,$9F,$ED,$7F,$64,$F9
	.dc.b	$7E,$9E,$1C,$BF,$34,$E5,$F4,$7D,$93,$E5,$7F,$B9,$FD,$87,$F4,$49
	.dc.b	$FF,$70,$37,$97,$FD,$81,$5B,$9F,$81,$F3,$CF,$B2,$FE,$F7,$2F,$91
	.dc.b	$A7,$F7,$E9,$D5,$70,$F8,$BC,$5C,$8C,$1D,$3C,$BF,$1F,$9F,$23,$07
	.dc.b	$CF,$FB,$27,$D3,$75,$A9,$39,$8F,$F0,$6C,$EA,$30,$43,$DB,$F2,$1F
	.dc.b	$E1,$2F,$B1,$FB,$D1,$F0,$2B,$E1,$99,$EF,$91,$F7,$F8,$3A,$3B,$3F
	.dc.b	$05,$5C,$B8,$3F,$DC,$CD,$EE,$FF,$27,$36,$7F,$CA,$2D,$54,$FB,$2F
	.dc.b	$27,$0F,$D4,$F5,$25,$76,$8C,$BE,$38,$CA,$C7,$C3,$3F,$E2,$EF,$B2
	.dc.b	$77,$7F,$8D,$E5,$F5,$DF,$CA,$EF,$F3,$B1,$8E,$2E,$15,$27,$F3,$D7
	.dc.b	$C9,$B1,$D7,$BB,$FC,$77,$93,$E4,$2F,$8F,$2F,$F4,$D3,$96,$CF,$02
	.dc.b	$3F,$AB,$FB,$2F,$D2,$E7,$FE,$76,$FB,$2F,$E6,$D2,$9C,$A8,$CD,$F5
	.dc.b	$8C,$3F,$E7,$D3,$11,$CA,$78,$B7,$1F,$91,$5C,$0F,$F4,$2D,$70,$39
	.dc.b	$61,$FE,$87,$AE,$18,$79,$67,$FD,$15,$87,$47,$77,$93,$9D,$F1,$8D
	.dc.b	$E1,$E0,$C3,$9D,$78,$4F,$F5,$E5,$6D,$E2,$73,$3E,$14,$F1,$AE,$78
	.dc.b	$65,$F4,$4F,$03,$A6,$1E,$57,$FD,$42,$FC,$A7,$FD,$1E,$D7,$99,$FE
	.dc.b	$92,$3B,$CD,$7E,$97,$A0,$F8,$EC,$57,$F4,$D6,$C6,$67,$94,$E5,$E4
	.dc.b	$3E,$3E,$B4,$CB,$82,$3F,$16,$FF,$37,$CB,$F0,$E3,$FD,$77,$3F,$16
	.dc.b	$73,$1F,$2C,$C9,$C7,$65,$71,$8F,$06,$BF,$8E,$AE,$2F,$E4,$23,$8B
	.dc.b	$9A,$E3,$1D,$4E,$39,$FF,$AB,$63,$33,$C6,$72,$7F,$52,$8C,$CF,$CD
	.dc.b	$C3,$D1,$18,$9F,$F6,$0C,$63,$0F,$33,$63,$D5,$1F,$D4,$70,$F7,$9E
	.dc.b	$48,$FF,$66,$1A,$3F,$82,$47,$94,$E1,$5E,$A9,$E1,$3E,$EB,$D3,$C0
	.dc.b	$FF,$73,$FD,$8F,$EF,$6F,$FB,$03,$63,$41,$FF,$76,$C6,$B6,$EC,$FF
	.dc.b	$BB,$E3,$85,$F9,$DB,$19,$AD,$9E,$0A,$AF,$5E,$8B,$5F,$D0,$AB,$F8
	.dc.b	$CF,$D9,$7F,$A8,$1F,$2C,$E8,$E7,$82,$5F,$3F,$9A,$E2,$F4,$F3,$E4
	.dc.b	$7F,$24,$E6,$76,$63,$3F,$7E,$7F,$E1,$8A,$F8,$77,$DA,$F7,$18,$25
	.dc.b	$7C,$BC,$BF,$98,$FB,$FC,$DF,$A5,$5F,$EC,$1A,$FF,$06,$7D,$93,$CB
	.dc.b	$F8,$3F,$2D,$FD,$B6,$73,$18,$EC,$91,$FE,$13,$9D,$65,$F5,$2F,$94
	.dc.b	$E6,$BE,$5F,$D9,$3E,$2E,$9C,$AF,$ED,$C6,$6B,$FC,$3C,$BE,$57,$89
	.dc.b	$F5,$3E,$CB,$FD,$41,$3A,$8F,$AF,$F6,$4F,$91,$A7,$09,$FF,$14,$46
	.dc.b	$5B,$F3,$BC,$0E,$3C,$F6,$3F,$C5,$DF,$64,$F8,$2F,$C1,$B6,$3F,$AC
	.dc.b	$F9,$BF,$E2,$17,$73,$F8,$87,$FC,$C0,$7F,$8D,$CB,$AF,$D3,$FB,$17
	.dc.b	$D9,$6F,$C3,$FC,$7F,$D9,$7E,$6A,$BF,$3A,$DE,$8F,$D4,$5F,$F8,$A7
	.dc.b	$0F,$F2,$14,$E5,$54,$87,$CD,$F4,$3F,$27,$F2,$4F,$5F,$17,$CB,$B7
	.dc.b	$C2,$6B,$E4,$AC,$7F,$93,$FC,$67,$AB,$FA,$46,$C7,$F7,$AF,$B2,$FE
	.dc.b	$06,$F8,$B4,$FE,$5D,$5B,$F4,$E4,$FF,$80,$3E,$C9,$F5,$58,$BF,$DD
	.dc.b	$FD,$97,$F9,$AA,$BF,$CB,$CD,$8F,$CD,$7D,$97,$F9,$6A,$73,$F0,$9F
	.dc.b	$B6,$AF,$F4,$FC,$70,$F9,$2F,$D8,$D7,$F9,$5E,$30,$74,$F8,$1E,$07
	.dc.b	$ED,$BE,$CB,$F2,$18,$F8,$A3,$F8,$0F,$AD,$F7,$63,$F8,$7B,$7A,$3F
	.dc.b	$8D,$62,$8D,$FC,$17,$F0,$8F,$B2,$FB,$B9,$7F,$9D,$DB,$87,$EC,$EF
	.dc.b	$F6,$0F,$F1,$A8,$CE,$1F,$CA,$FE,$C9,$F1,$FF,$C4,$9B,$87,$F8,$A3
	.dc.b	$EC,$5F,$57,$9A,$93,$FE,$85,$AD,$E3,$FC,$06,$F7,$39,$AD,$CE,$7C
	.dc.b	$FB,$8F,$F3,$5B,$FD,$CC,$FF,$09,$FB,$2F,$8B,$3F,$CE,$5B,$87,$FA
	.dc.b	$79,$7C,$6F,$F2,$07,$35,$FE,$45,$6D,$1A,$70,$AF,$F2,$1B,$AB,$FC
	.dc.b	$F5,$7A,$79,$FF,$95,$C6,$2F,$CD,$18,$F8,$8C,$B7,$0F,$8C,$FF,$3A
	.dc.b	$FB,$27,$CD,$D3,$CD,$AD,$AF,$F5,$64,$E9,$B3,$A6,$93,$2F,$E5,$9F
	.dc.b	$5D,$F8,$38,$FD,$05,$71,$9C,$F1,$FA,$3B,$E5,$7D,$9F,$0A,$D7,$D5
	.dc.b	$9B,$AF,$D1,$9B,$5F,$37,$CF,$F9,$9F,$65,$F3,$5B,$1F,$21,$B8,$7D
	.dc.b	$76,$DF,$F7,$5C,$BF,$A8,$D7,$BB,$DB,$F9,$7D,$FF,$15,$91,$93,$F1
	.dc.b	$D1,$8C,$3F,$B7,$D7,$27,$FA,$67,$D9,$32,$D6,$3F,$DF,$63,$18,$7D
	.dc.b	$CF,$B2,$FC,$27,$9D,$F8,$56,$C6,$4F,$C0,$C6,$5A,$7E,$7F,$CE,$FC
	.dc.b	$09,$F8,$2F,$5E,$C6,$49,$97,$E7,$E3,$2F,$F8,$1B,$EC,$BF,$42,$DE
	.dc.b	$DF,$CB,$77,$FF,$7E,$9C,$E9,$FE,$1B,$6E,$1F,$C6,$DB,$7F,$9B,$AE
	.dc.b	$17,$2B,$FD,$37,$CB,$FD,$CA,$BB,$46,$3D,$EF,$CC,$FC,$94,$8F,$F5
	.dc.b	$0F,$AA,$F7,$31,$7F,$B7,$AB,$A7,$F8,$9E,$B5,$7F,$C0,$67,$F5,$DF
	.dc.b	$4F,$23,$F5,$3F,$1F,$70,$FD,$C4,$DC,$3F,$71,$36,$BF,$9B,$7D,$97
	.dc.b	$F2,$D9,$FF,$1D,$36,$3F,$A5,$FC,$4F,$F8,$B7,$C5,$FF,$11,$CE,$71
	.dc.b	$FE,$9E,$D4,$BE,$CB,$7E,$77,$07,$8F,$C0,$4C,$7F,$AB,$E5,$E0,$F8
	.dc.b	$BC,$63,$FE,$3C,$C7,$F8,$4B,$D2,$FC,$33,$AA,$F3,$F4,$79,$0F,$F1
	.dc.b	$CE,$9F,$2F,$E2,$71,$97,$63,$DF,$3E,$AD,$F3,$97,$77,$FC,$7F,$EC
	.dc.b	$7F,$3B,$AF,$F4,$83,$67,$FD,$28,$D7,$FA,$F1,$FD,$BB,$CE,$FF,$2D
	.dc.b	$BF,$DC,$A3,$86,$5F,$E6,$57,$FC,$9F,$19,$9E,$FF,$11,$2F,$CD,$3C
	.dc.b	$63,$C7,$39,$8F,$F2,$FD,$7F,$98,$5F,$0B,$5F,$EE,$0D,$79,$57,$AB
	.dc.b	$DF,$FF,$44,$BC,$1B,$1C,$1A,$FA,$5A,$FF,$82,$9F,$F4,$8F,$B0,$FA
	.dc.b	$5E,$FE,$4E,$1E,$57,$FE,$A4,$72,$DB,$F8,$93,$B6,$C6,$5E,$CD,$C3
	.dc.b	$FB,$85,$F7,$9D,$9F,$EB,$9B,$EA,$D3,$D1,$5C,$6F,$B4,$7F,$88,$9B
	.dc.b	$7D,$2B,$CF,$3C,$4D,$57,$12,$39,$9F,$F4,$03,$6B,$99,$27,$D1,$85
	.dc.b	$9D,$47,$2C,$3D,$51,$EE,$32,$47,$37,$E5,$8F,$59,$9C,$3D,$F3,$AC
	.dc.b	$3D,$78,$77,$8C,$61,$DE,$BD,$B1,$8B,$E4,$7C,$A7,$BD,$71,$9E,$A4
	.dc.b	$9F,$0C,$F1,$24,$66,$3D,$51,$87,$EA,$D7,$E5,$2F,$F1,$38,$7F,$8A
	.dc.b	$63,$07,$F9,$76,$76,$7D,$A7,$2F,$FB,$12,$38,$3D,$FE,$C7,$FC,$B1
	.dc.b	$97,$B6,$3F,$CB,$F1,$87,$DA,$1C,$61,$E5,$8C,$7B,$9F,$E3,$77,$73
	.dc.b	$AC,$BB,$4E,$6B,$84,$E6,$BE,$45,$7F,$97,$63,$11,$F8,$83,$F1,$4E
	.dc.b	$30,$FB,$51,$E0,$C3,$D0,$D3,$C3,$87,$7B,$F0,$CE,$AF,$96,$5F,$9A
	.dc.b	$6E,$5F,$A7,$6E,$1E,$0C,$39,$3E,$63,$E2,$B8,$7C,$D9,$7E,$4A,$78
	.dc.b	$B6,$75,$97,$A2,$73,$97,$BE,$73,$8F,$8D,$B8,$F8,$DE,$D5,$F1,$0E
	.dc.b	$38,$F8,$B2,$F4,$7D,$8F,$F5,$F8,$FE,$52,$7D,$77,$0F,$F2,$C9,$FB
	.dc.b	$8D,$8C,$BF,$E8,$DF,$B1,$7C,$D7,$FA,$3F,$1F,$E6,$71,$89,$FC,$70
	.dc.b	$70,$A3,$73,$C6,$3A,$3F,$E9,$9D,$7F,$6F,$5C,$1B,$19,$D3,$C4,$5A
	.dc.b	$FF,$58,$D7,$EF,$AB,$71,$FC,$76,$39,$4F,$08,$C3,$E1,$78,$B5,$F4
	.dc.b	$36,$3F,$84,$DF,$BB,$CC,$FF,$57,$70,$FA,$EF,$C3,$EA,$E2,$BE,$33
	.dc.b	$63,$CB,$55,$F8,$8D,$BF,$F3,$CC,$7F,$52,$3E,$44,$E6,$75,$DB,$CF
	.dc.b	$9D,$F5,$3F,$EF,$5B,$FE,$B3,$18,$AD,$C7,$3C,$17,$88,$65,$E9,$7E
	.dc.b	$B4,$79,$7E,$C7,$FB,$6C,$E5,$FF,$6D,$CF,$95,$FD,$D3,$CB,$3F,$F6
	.dc.b	$84,$62,$78,$72,$73,$3A,$71,$4B,$E5,$1F,$0F,$E8,$1F,$63,$49,$3F
	.dc.b	$0F,$EC,$7C,$D8,$79,$EB,$73,$F1,$DA,$F8,$CF,$F7,$9B,$88,$F7,$4F
	.dc.b	$2C,$BF,$DC,$D2,$F9,$0F,$11,$2F,$A7,$CB,$D7,$A0,$9E,$0F,$86,$7D
	.dc.b	$53,$F7,$7E,$C5,$F4,$BE,$DF,$96,$97,$F7,$67,$2D,$9C,$BD,$B0,$57
	.dc.b	$39,$93,$B2,$1B,$F9,$0F,$C6,$7D,$8B,$C1,$A4,$9F,$C6,$47,$E0,$9B
	.dc.b	$19,$7D,$B1,$C5,$DF,$1F,$4F,$17,$AF,$4C,$9E,$71,$97,$AC,$65,$E7
	.dc.b	$1F,$9F,$FB,$17,$D8,$F9,$BB,$25,$7E,$7E,$72,$D9,$CB,$E4,$9E,$51
	.dc.b	$23,$44,$3D,$3F,$63,$FA,$DE,$D7,$BD,$7D,$E9,$5F,$AD,$9C,$B6,$72
	.dc.b	$FD,$E8,$4F,$39,$E9,$F5,$72,$B9,$3D,$FE,$BF,$11,$C6,$2F,$57,$EB
	.dc.b	$70,$57,$F0,$1F,$1B,$FC,$27,$39,$EA,$F4,$19,$D6,$63,$E7,$8C,$BF
	.dc.b	$C1,$6B,$8C,$22,$46,$1B,$3C,$9F,$CC,$FD,$83,$EC,$7F,$89,$3F,$9A
	.dc.b	$8C,$F5,$E0,$CF,$64,$8F,$E2,$92,$9D,$46,$27,$9F,$C1,$C8,$DB,$DB
	.dc.b	$AF,$11,$E2,$8B,$A7,$F1,$1C,$3F,$3F,$3E,$DF,$6B,$F9,$36,$DF,$97
	.dc.b	$8F,$D8,$B6,$BF,$8C,$6A,$AF,$98,$70,$E8,$97,$CC,$87,$EE,$A7,$83
	.dc.b	$8A,$FE,$45,$78,$8F,$DF,$FD,$7B,$EC,$7F,$9C,$7D,$8F,$E1,$27,$F9
	.dc.b	$B3,$5D,$C7,$F3,$D9,$7C,$07,$0E,$59,$5F,$CF,$67,$2F,$F3,$FF,$B1
	.dc.b	$7C,$CF,$F4,$5F,$B1,$FE,$31,$CF,$F7,$69,$18,$7F,$01,$5C,$1E,$5F
	.dc.b	$15,$C7,$C4,$91,$83,$FA,$67,$D8,$FF,$3A,$7F,$8D,$DF,$F1,$B7,$FA
	.dc.b	$83,$5F,$E3,$F1,$BE,$0F,$E8,$CB,$E6,$7C,$DE,$64,$AE,$B1,$FD,$77
	.dc.b	$DD,$7D,$8F,$F5,$FF,$B1,$FB,$1B,$7A,$DB,$7F,$99,$AD,$BF,$D8,$57
	.dc.b	$C7,$F1,$F2,$24,$E6,$3F,$B5,$7D,$8A,$BB,$7A,$F2,$BF,$B5,$4E,$5B
	.dc.b	$EF,$FF,$71,$F7,$DF,$07,$4E,$71,$D3,$E5,$B5,$F0,$92,$F9,$C7,$F7
	.dc.b	$9F,$B1,$7C,$07,$3F,$06,$57,$F7,$99,$CB,$FD,$EF,$EC,$5F,$29,$FE
	.dc.b	$01,$FA,$EF,$E8,$33,$FD,$31,$BD,$3F,$DC,$EB,$73,$FE,$03,$F9,$E7
	.dc.b	$D8,$FF,$82,$65,$F2,$34,$FF,$06,$4E,$AB,$87,$C4,$E2,$E4,$60,$E9
	.dc.b	$E4,$F8,$FC,$F9,$18,$3F,$C2,$5F,$62,$F8,$9E,$B5,$27,$31,$FE,$1B
	.dc.b	$9D,$46,$08,$7B,$7E,$3F,$FC,$41,$F6,$3F,$72,$3F,$C3,$DB,$F0,$7E
	.dc.b	$15,$EB,$B2,$46,$A3,$E4,$18,$7F,$B9,$D7,$36,$CE,$78,$3F,$C3,$96
	.dc.b	$FB,$1F,$0F,$07,$D1,$F4,$E6,$1F,$E2,$FC,$7E,$5B,$67,$2F,$D5,$9E
	.dc.b	$2F,$F8,$D7,$EC,$5D,$9F,$E3,$DC,$9F,$5B,$FC,$96,$BF,$9A,$47,$F9
	.dc.b	$14,$CF,$0F,$F9,$06,$F8,$CF,$3B,$C1,$1E,$DD,$9F,$E4,$7C,$5F,$1F
	.dc.b	$7C,$79,$5F,$E4,$79,$CB,$67,$81,$1F,$F2,$67,$D8,$AF,$FC,$83,$0F
	.dc.b	$B1,$FE,$62,$A7,$5A,$7F,$4D,$9C,$E1,$DA,$B8,$3F,$C5,$4C,$47,$2A
	.dc.b	$E3,$3A,$3F,$AB,$38,$8E,$78,$7F,$1D,$6C,$F7,$AF,$E4,$15,$B6,$E1
	.dc.b	$E8,$AD,$D7,$53,$FC,$4F,$7C,$EB,$78,$73,$9F,$0D,$F1,$9F,$45,$6E
	.dc.b	$7F,$B5,$47,$A8,$E2,$F5,$C3,$C4,$EA,$78,$3E,$03,$B1,$F2,$6B,$B3
	.dc.b	$E0,$BF,$F5,$7B,$96,$D7,$C6,$8F,$E7,$C6,$4F,$F5,$AB,$FD,$13,$E3
	.dc.b	$B7,$E8,$1A,$49,$EA,$3E,$3E,$B4,$CB,$82,$3F,$26,$FF,$37,$E5,$7A
	.dc.b	$A3,$FC,$19,$3E,$C9,$CC,$7C,$F3,$27,$1D,$BF,$D4,$5A,$FE,$2A,$32
	.dc.b	$FE,$32,$78,$B9,$8F,$23,$99,$E3,$AF,$FB,$52,$FC,$11,$8A,$F8,$B1
	.dc.b	$8C,$3C,$D1,$CA,$BF,$D6,$75,$B8,$FF,$65,$1E,$59,$DD,$71,$8C,$1F
	.dc.b	$E7,$CC,$3D,$A7,$FA,$8D,$F4,$DF,$4A,$FF,$5A,$4E,$9C,$4F,$FA,$10
	.dc.b	$D9,$FC,$02,$FF,$80,$47,$F7,$9F,$B0,$FD,$CB,$FE,$BD,$6C,$FF,$A3
	.dc.b	$1F,$F1,$D4,$67,$6E,$33,$FE,$3D,$9D,$DF,$FA,$DD,$B5,$D6,$76,$23
	.dc.b	$2F,$1E,$8C,$C7,$F3,$B9,$FE,$2A,$7F,$16,$FB,$1F,$F4,$BB,$E5,$9D
	.dc.b	$9E,$2B,$4B,$E8,$33,$5C,$B0,$FF,$65,$CE,$63,$DD,$E7,$F8,$0E,$5C
	.dc.b	$CE,$CC,$4E,$89,$7F,$EB,$9C,$7F,$95,$CE,$A3,$E3,$5F,$2F,$7B,$15
	.dc.b	$FC,$B7,$0F,$F5,$D6,$1F,$EE,$9F,$B1,$79,$5F,$13,$E6,$7F,$B3,$CE
	.dc.b	$63,$1D,$92,$3F,$DE,$13,$AC,$FF,$D7,$F7,$CA,$73,$5F,$3B,$EC,$5F
	.dc.b	$15,$4E,$57,$FE,$C6,$8C,$D7,$FB,$ED,$7C,$AF,$13,$EB,$FD,$8F,$FA
	.dc.b	$5A,$75,$1F,$6B,$EC,$5F,$21,$4E,$13,$F3,$F0,$4E,$A3,$0B,$E7,$39
	.dc.b	$1C,$F9,$AC,$F1,$FB,$1F,$F2,$AB,$7E,$0B,$EF,$79,$BF,$E1,$C8,$C1
	.dc.b	$DC,$FF,$2A,$97,$C1,$FC,$27,$D8,$3E,$C7,$F2,$33,$AD,$3F,$DC,$FF
	.dc.b	$63,$F9,$C6,$9C,$3D,$3F,$CB,$C6,$7A,$55,$FE,$5F,$87,$CD,$68,$FC
	.dc.b	$97,$C8,$BD,$7C,$57,$C3,$AF,$F2,$A3,$7B,$B7,$F2,$3D,$9F,$D2,$F8
	.dc.b	$EF,$5A,$9F,$9D,$FB,$1F,$DE,$6D,$FC,$82,$37,$EA,$4C,$7F,$3D,$B3
	.dc.b	$EC,$5F,$55,$9C,$13,$EA,$7F,$D6,$9F,$63,$FE,$55,$BF,$B0,$D8,$FC
	.dc.b	$F7,$D8,$FF,$73,$9C,$FC,$1F,$EE,$AB,$FD,$39,$1C,$3E,$47,$F6,$54
	.dc.b	$F8,$1D,$46,$D9,$3E,$6F,$B1,$FC,$8E,$3E,$48,$FF,$45,$61,$C5,$FC
	.dc.b	$57,$AF,$F7,$E7,$47,$FB,$2D,$C6,$7F,$E1,$9C,$51,$FC,$38,$CD,$7F
	.dc.b	$A5,$7E,$C7,$F0,$4F,$E5,$5B,$7F,$BB,$AF,$DA,$BD,$84,$6A,$FE,$44
	.dc.b	$3E,$C5,$F1,$DF,$C7,$1B,$87,$D2,$FB,$07,$D5,$E6,$A4,$FE,$96,$B7
	.dc.b	$7F,$C2,$EF,$73,$9A,$DC,$E7,$CF,$B8,$FF,$8D,$EF,$F8,$04,$FF,$13
	.dc.b	$FB,$1F,$97,$3F,$E3,$B6,$E1,$FE,$3E,$5F,$19,$FC,$9D,$CD,$7E,$11
	.dc.b	$B4,$69,$C2,$BF,$1A,$EA,$FF,$D0,$F1,$EF,$E6,$FE,$65,$18,$BF,$64
	.dc.b	$63,$C0,$CB,$6B,$81,$FC,$1F,$EC,$5F,$35,$FD,$19,$B1,$E0,$D7,$13
	.dc.b	$23,$83,$EB,$8C,$61,$F6,$6B,$8B,$FC,$DF,$EB,$7F,$06,$84,$6B,$4F
	.dc.b	$F3,$14,$E9,$67,$FC,$86,$BE,$B7,$7F,$BB,$C2,$F7,$B3,$3F,$C0,$39
	.dc.b	$69,$F2,$3C,$EF,$A1,$F6,$3F,$45,$B1,$93,$E6,$C6,$5A,$7C,$5F,$6B
	.dc.b	$E2,$1F,$3B,$BA,$C6,$49,$8F,$E2,$E3,$2F,$C3,$9F,$EC,$DF,$62,$BC
	.dc.b	$E9,$FE,$81,$8C,$61,$F7,$BE,$C7,$F0,$DE,$67,$E1,$DB,$19,$3F,$07
	.dc.b	$19,$69,$FA,$1F,$37,$F0,$67,$E1,$3D,$6B,$19,$26,$7F,$A1,$8C,$BF
	.dc.b	$B1,$FB,$1F,$D1,$B7,$B3,$F3,$1D,$DF,$E0,$A9,$CE,$9F,$B4,$6E,$1F
	.dc.b	$76,$D1,$87,$F3,$D5,$C6,$E5,$7F,$28,$E4,$FF,$00,$3F,$D6,$6F,$86
	.dc.b	$09,$CE,$1F,$B7,$F9,$1F,$CF,$9F,$E1,$1F,$4D,$ED,$63,$1F,$E0,$53
	.dc.b	$6B,$FC,$11,$5A,$9C,$9F,$AC,$FA,$7F,$DA,$1F,$B5,$F8,$EB,$19,$25
	.dc.b	$7E,$FE,$32,$FF,$17,$FB,$0F,$E7,$19,$FF,$16,$6D,$7E,$7F,$E1,$FF
	.dc.b	$C6,$1E,$1F,$F8,$CA,$73,$7F,$A3,$6A,$5F,$63,$BF,$33,$83,$C7,$DE
	.dc.b	$98,$FF,$7F,$AF,$E8,$D7,$E3,$9C,$B6,$75,$F1,$58,$CF,$CD,$CF,$E3
	.dc.b	$3F,$C9,$7A,$7C,$AF,$87,$C5,$F5,$9F,$46,$F8,$C7,$E0,$3F,$E4,$CF
	.dc.b	$5F,$FA,$0D,$7F,$84,$5F,$F2,$6C,$FF,$76,$F3,$35,$F0,$FB,$6D,$6D
	.dc.b	$B3,$FD,$06,$B8,$92,$33,$9F,$F4,$28,$FF,$32,$FA,$AF,$A4,$FC,$84
	.dc.b	$7F,$10,$8C,$5F,$F0,$F8,$DC,$EB,$C9,$FE,$E5,$5A,$AF,$F0,$5F,$BD
	.dc.b	$8A,$E7,$F2,$B6,$79,$4D,$7F,$C2,$31,$70,$FF,$34,$47,$FA,$B6,$B8
	.dc.b	$E7,$FE,$A6,$7C,$B7,$BB,$93,$9A,$FE,$49,$F6,$3F,$E1,$9D,$31,$1D
	.dc.b	$A6,$CE,$63,$B1,$ED,$AF,$24,$67,$2F,$66,$5F,$DE,$5F,$74,$67,$2F
	.dc.b	$2D,$F2,$11,$97,$A6,$1E,$0F,$73,$FB,$EC,$6F,$0E,$97,$FB,$FB,$E1
	.dc.b	$3E,$73,$A3,$89,$DC,$7B,$AB,$9E,$1E,$6C,$7D,$31,$8C,$3C,$B3,$E8
	.dc.b	$8C,$57,$90,$EA,$E1,$F0,$46,$F2,$F9,$38,$78,$27,$DB,$5B,$8C,$63
	.dc.b	$E5,$C7,$C3,$7C,$1B,$7E,$F6,$E1,$E7,$CB,$E2,$56,$DB,$5B,$6C,$EE
	.dc.b	$3C,$F3,$B6,$9B,$2C,$6C,$F5,$39,$69,$9C,$BE,$86,$5E,$BC,$3C,$B1
	.dc.b	$F2,$0C,$DF,$7A,$E3,$19,$BF,$A5,$39,$8D,$C7,$76,$CF,$78,$D9,$92
	.dc.b	$B9,$7E,$6C,$F3,$73,$3D,$DC,$9C,$4B,$3C,$71,$FA,$58,$FC,$8A,$E1
	.dc.b	$18,$AF,$8B,$18,$C7,$E0,$56,$E7,$F0,$97,$EE,$6D,$7A,$8E,$18,$79
	.dc.b	$63,$C3,$18,$7C,$98,$77,$AE,$51,$FA,$4F,$B1,$FE,$9A,$FF,$5D,$3F
	.dc.b	$6E,$E1,$C4,$FE,$D5,$B1,$97,$F4,$CF,$29,$FD,$9B,$6B,$3D,$BF,$BE
	.dc.b	$F0,$FF,$B5,$57,$F1,$5F,$B1,$7C,$D6,$2F,$FB,$74,$63,$1F,$E0,$ED
	.dc.b	$7F,$78,$FF,$1D,$AE,$55,$C6,$7F,$C6,$53,$B9,$CC,$72,$AD,$29,$E5
	.dc.b	$1E,$0A,$CC,$71,$8F,$19,$82,$4F,$33,$CB,$CD,$C3,$CB,$FF,$03,$72
	.dc.b	$7F,$9F,$2F,$FC,$CB,$5F,$ED,$96,$BF,$E8,$37,$FD,$CA,$E6,$78,$EF
	.dc.b	$F0,$BD,$4B,$3E,$C6,$BE,$99,$B4,$F0,$37,$D4,$FA,$8E,$B7,$FE,$9E
	.dc.b	$8C,$5F,$0A,$93,$E5,$6B,$FE,$97,$FB,$1F,$C2,$67,$FE,$96,$7F,$DA
	.dc.b	$B1,$FE,$F3,$7F,$94,$BB,$E4,$F2,$C6,$A3,$97,$3E,$26,$47,$18,$F5
	.dc.b	$FD,$73,$EC,$69,$27,$FE,$00,$7D,$8F,$FA,$9B,$1F,$F5,$3D,$6E,$7F
	.dc.b	$CE,$11,$8A,$E1,$1F,$EA,$FA,$E1,$C1,$FE,$39,$B7,$C7,$F8,$09,$5C
	.dc.b	$A7,$5F,$2F,$88,$FF,$5C,$57,$07,$47,$F9,$FD,$C3,$A8,$FA,$BF,$62
	.dc.b	$FA,$5F,$6F,$CB,$4A,$E2,$27,$4E,$1A,$F8,$23,$83,$FE,$C9,$8C,$E9
	.dc.b	$FE,$C8,$6E,$3F,$E5,$56,$CF,$FA,$4D,$AB,$17,$C7,$FE,$2B,$EC,$5E
	.dc.b	$2D,$25,$79,$A3,$F1,$8D,$8C,$BF,$ED,$B8,$E2,$EF,$8F,$A7,$91,$D7
	.dc.b	$A6,$4F,$38,$CB,$D6,$32,$F3,$8F,$CE,$FD,$8B,$EC,$7C,$DD,$B2,$BE
	.dc.b	$4C,$E5,$B3,$97,$C9,$3C,$A2,$46,$88,$7A,$7E,$C7,$F5,$7D,$8F,$72
	.dc.b	$FB,$B2,$BE,$BC,$E5,$B3,$97,$FD,$F9,$09,$E7,$3D,$3E,$B2,$57,$27
	.dc.b	$BF,$D7,$E2,$38,$C5,$E5,$72,$FF,$A3,$F1,$5F,$F0,$0F,$27,$FC,$79
	.dc.b	$5A,$C3,$FD,$A1,$5F,$9E,$6D,$EF,$59,$1E,$A8,$F7,$98,$55,$CD,$5F
	.dc.b	$69,$FF,$50,$36,$3F,$DB,$CD,$5F,$60,$FB,$1F,$E2,$AF,$E9,$2F,$A2
	.dc.b	$EB,$C9,$AE,$C9,$7F,$A4,$9D,$36,$72,$F9,$B1,$FF,$59,$4F,$FB,$F2
	.dc.b	$38,$79,$DE,$03,$C3,$D9,$63,$24,$39,$1F,$EF,$79,$47,$CD,$F6,$FF
	.dc.b	$96,$EF,$FA,$0A,$D9,$F4,$BA,$34,$FE,$EB,$65,$70,$99,$39,$8D,$BC
	.dc.b	$A7,$6E,$4F,$E0,$75,$87,$BF,$D7,$BE,$C3,$F9,$FF,$D8,$FE,$36,$7F
	.dc.b	$84,$B5,$DC,$7F,$44,$97,$9B,$C3,$96,$57,$F1,$09,$CB,$FC,$4B,$EC
	.dc.b	$5F,$33,$FD,$33,$EC,$7F,$96,$4F,$C0,$6F,$2F,$E1,$6B,$73,$C7,$E2
	.dc.b	$B8,$78,$D2,$30,$7F,$1E,$FB,$1F,$E8,$57,$F8,$DC,$3F,$89,$4F,$F2
	.dc.b	$26,$BF,$3A,$77,$CD,$FE,$EF,$5F,$31,$E5,$F3,$25,$75,$8F,$EC,$7E
	.dc.b	$F3,$EC,$7F,$B2,$FD,$8F,$D9,$CF,$F3,$8D,$CB,$FD,$F5,$5B,$C3,$F9
	.dc.b	$9A,$F8,$FE,$3E,$44,$9C,$C7,$F7,$0F,$B1,$57,$6F,$5E,$57,$F3,$C9
	.dc.b	$CB,$7E,$03,$FB,$BF,$B0,$F8,$5A,$73,$8E,$9F,$2F,$AF,$85,$97,$CE
	.dc.b	$3F,$BF,$FD,$8B,$E0,$79,$F8,$32,$BF,$A5,$CE,$5F,$E9,$BF,$62,$F9
	.dc.b	$4F,$F0,$3F,$D7,$FF,$42,$3F,$A1,$C6,$67,$FA,$9F,$CE,$BE,$C7,$FC
	.dc.b	$1B,$2F,$90,$A7,$F8,$42,$75,$5C,$3E,$2B,$17,$23,$07,$4F,$2B,$C3
	.dc.b	$E7,$48,$C1,$FD,$8B,$EC,$5F,$15,$D6,$A4,$E6,$3F,$C3,$F3,$A8,$C1
	.dc.b	$0F,$DA,$E8,$FA,$AF,$F1,$2F,$D8,$7D,$B9,$FE,$DF,$3C,$5D,$DE,$BA
	.dc.b	$E4,$6A,$3E,$39,$87,$F8,$0D,$73,$6C,$E7,$BF,$F9,$45,$BE,$C7,$C3
	.dc.b	$C1,$F4,$3D,$19,$87,$F8,$03,$1F,$98,$D9,$CB,$EA,$9E,$2F,$F8,$1B
	.dc.b	$EC,$5D,$9F,$E4,$2C,$9F,$5B,$FC,$9A,$BF,$8D,$C6,$38,$7F,$C8,$B7
	.dc.b	$C2,$79,$5E,$08,$F5,$EC,$FF,$25,$E2,$F9,$0B,$E3,$CA,$FF,$0B,$CE
	.dc.b	$5B,$3C,$08,$FF,$87,$3E,$C5,$7F,$D3,$21,$F6,$3F,$CB,$90,$9D,$E9
	.dc.b	$89,$D4,$FF,$34,$BD,$BC,$B0,$A7,$8D,$C1,$D2,$F9,$B8,$3B,$61,$E3
	.dc.b	$BE,$B5,$C5,$B7,$D9,$F0,$DF,$26,$D7,$1A,$F0,$D6,$DB,$7E,$97,$CF
	.dc.b	$7C,$5B,$5C,$67,$E0,$C7,$84,$E2,$F9,$E7,$83,$EB,$9E,$0F,$C3,$3B
	.dc.b	$1C,$AB,$C4,$EF,$0C,$BF,$01,$FF,$5E,$39,$2D,$7C,$67,$FD,$80,$D7
	.dc.b	$D3,$E7,$BE,$37,$8C,$7D,$78,$C3,$C9,$AF,$26,$BE,$0F,$8F,$E2,$FE
	.dc.b	$01,$A9,$A4,$7E,$C5,$FE,$83,$C9,$F5,$C7,$13,$3B,$7F,$88,$75,$FC
	.dc.b	$3C,$65,$FC,$4D,$71,$8F,$C1,$19,$D7,$FD,$6B,$5C,$5F,$34,$74,$BE
	.dc.b	$EF,$94,$C1,$C8,$C5,$F6,$AE,$51,$FE,$E9,$C3,$E8,$E1,$E1,$AF,$2C
	.dc.b	$F0,$C3,$C3,$87,$6B,$DE,$1C,$AB,$FC,$66,$FF,$8F,$70,$FF,$50,$36
	.dc.b	$BC,$F8,$78,$5F,$F4,$5C,$FB,$CD,$BF,$E8,$EA,$F8,$07,$FA,$9A,$B9
	.dc.b	$3F,$B2,$FB,$1F,$E0,$13,$FE,$94,$6C,$7F,$88,$1F,$DB,$C6,$75,$E3
	.dc.b	$29,$DD,$FF,$92,$5B,$5F,$B8,$78,$CF,$3A,$E3,$1F,$D0,$27,$C1,$D1
	.dc.b	$E4,$B7,$D8,$FF,$AA,$67,$24,$C3,$F8,$FC,$3E,$7B,$9D,$F1,$8C,$57
	.dc.b	$AF,$CF,$EE,$75,$E6,$FE,$8D,$19,$26,$1E,$F9,$D7,$67,$09,$CE,$5F
	.dc.b	$4E,$FF,$D9,$31,$83,$FD,$A2,$E2,$FF,$C8,$FF,$62,$F2,$7E,$3F,$CA
	.dc.b	$7F,$6B,$BF,$F6,$0C,$E5,$FF,$61,$C6,$6B,$FC,$9E,$DC,$3F,$CE,$8D
	.dc.b	$9F,$E4,$75,$C6,$B7,$D7,$F5,$3E,$C5,$F1,$54,$F0,$C5,$BF,$F6,$8C
	.dc.b	$BE,$53,$81,$F7,$3E,$C7,$FD,$45,$3A,$8F,$BF,$F6,$3F,$EC,$88,$FF
	.dc.b	$6C,$B7,$1F,$F3,$2D,$F3,$9C,$FC,$87,$0F,$9C,$91,$C3,$92,$DF,$F9
	.dc.b	$B3,$EC,$7F,$CA,$F3,$FE,$E4,$6C,$7F,$5C,$F2,$FF,$C4,$13,$9F,$81
	.dc.b	$FC,$8B,$FE,$58,$3F,$DD,$65,$9F,$EC,$7F,$62,$FB,$1D,$F8,$3F,$98
	.dc.b	$FB,$1F,$C9,$F3,$7F,$A7,$DB,$87,$F5,$48,$5F,$F8,$A6,$3F,$DE,$F3
	.dc.b	$AC,$3F,$AC,$AF,$9A,$E8,$7E,$67,$E4,$7F,$D5,$35,$C2,$3C,$1D,$92
	.dc.b	$3B,$1D,$E7,$E7,$DF,$FA,$3A,$FF,$CB,$F3,$A8,$FE,$51,$E9,$7E,$C3
	.dc.b	$C4,$3F,$A7,$FB,$1F,$E0,$15,$FD,$9A,$72,$F1,$AF,$E5,$DE,$87,$13
	.dc.b	$F5,$DF,$62,$FA,$9C,$56,$FE,$BF,$FC,$BA,$DF,$80,$FD,$A7,$D8,$FF
	.dc.b	$7C,$8C,$1B,$F8,$FF,$E0,$75,$FE,$A2,$8F,$47,$1A,$9E,$FE,$AF,$6C
	.dc.b	$AD,$FD,$8F,$E6,$1F,$14,$7F,$0E,$F4,$FF,$03,$B7,$FA,$C9,$B8,$FF
	.dc.b	$25,$C1,$1B,$F8,$0F,$E2,$BF,$63,$F8,$1C,$7E,$13,$70,$FF,$13,$DF
	.dc.b	$ED,$F0,$FF,$70,$4E,$5F,$F7,$17,$D8,$BE,$33,$F8,$FB,$70,$F6,$FD
	.dc.b	$7B,$EB,$33,$52,$7E,$85,$6F,$2F,$F1,$AD,$EE,$73,$5B,$9C,$F9,$B7
	.dc.b	$1F,$E8,$17,$FB,$D7,$FA,$17,$D8,$FC,$5A,$FF,$1F,$B6,$33,$F1,$5F
	.dc.b	$CA,$63,$18,$FE,$6D,$B5,$FE,$FE,$6D,$15,$71,$F3,$7F,$99,$61,$FA
	.dc.b	$1B,$F4,$F4,$63,$E1,$72,$DC,$BE,$1B,$67,$71,$F9,$1F,$B1,$7C,$C7
	.dc.b	$F4,$86,$C7,$E1,$B3,$F8,$DA,$23,$FC,$7F,$3A,$BF,$F7,$8B,$67,$51
	.dc.b	$FE,$4A,$8C,$4F,$BF,$6E,$D8,$48,$FC,$9F,$D6,$FE,$21,$EB,$1F,$D3
	.dc.b	$E3,$16,$C1,$A7,$ED,$1F,$A7,$F3,$DF,$13,$C2,$F8,$AC,$C7,$F0,$6E
	.dc.b	$5A,$7B,$7D,$0F,$A9,$F6,$3F,$55,$B1,$93,$E1,$46,$5A,$7C,$9F,$67
	.dc.b	$E4,$1F,$53,$C0,$B1,$92,$67,$F8,$E8,$CB,$F2,$27,$FB,$67,$D8,$AF
	.dc.b	$39,$FF,$83,$63,$18,$7D,$3F,$B1,$FC,$57,$9D,$F8,$B6,$C6,$4F,$B1
	.dc.b	$19,$69,$FA,$4F,$3B,$F1,$27,$E2,$BD,$6B,$19,$26,$7F,$A4,$8C,$BF
	.dc.b	$46,$3F,$CF,$FF,$63,$FA,$78,$C1,$FE,$88,$71,$97,$E9,$3B,$FF,$C2
	.dc.b	$71,$8A,$E3,$DD,$83,$F4,$93,$63,$F9,$24,$EE,$FF,$01,$C9,$FE,$04
	.dc.b	$7B,$88,$D6,$09,$D3,$FB,$6F,$90,$FD,$39,$FE,$19,$F5,$1E,$D6,$31
	.dc.b	$FE,$0F,$37,$1F,$F5,$BD,$6A,$72,$7E,$8F,$E9,$FF,$74,$7E,$EF,$E3
	.dc.b	$6C,$64,$95,$FC,$0A,$32,$FA,$3E,$C3,$F9,$CB,$7C,$CF,$F1,$FF,$C3
	.dc.b	$7F,$8D,$FC,$5F,$F1,$CD,$6B,$5F,$D5,$DA,$31,$8F,$FA,$C1,$7D,$8F
	.dc.b	$0C,$78,$BC,$BC,$59,$5F,$D3,$EB,$FA,$35,$F8,$E7,$2D,$9D,$7C,$4E
	.dc.b	$31,$F0,$DB,$9F,$CA,$7A,$38,$7F,$AE,$E1,$FE,$C8,$9D,$61,$DE,$BD
	.dc.b	$7B,$F0,$8F,$13,$FE,$56,$F5,$BF,$CA,$77,$FE,$D0,$6B,$FD,$E7,$CC
	.dc.b	$D7,$C2,$CB,$FF,$4B,$7D,$27,$FB,$42,$FF,$89,$7A,$88,$C5,$7F,$12
	.dc.b	$8C,$61,$D7,$C6,$C5,$71,$6C,$67,$0E,$AF,$8E,$F2,$FF,$82,$E6,$E1
	.dc.b	$E0,$BE,$57,$C2,$BF,$C2,$17,$AB,$DF,$9F,$2B,$58,$7F,$9B,$6F,$FD
	.dc.b	$5D,$87,$AE,$B5,$18,$3F,$DE,$7F,$62,$F9,$5F,$ED,$71,$93,$FC,$3D
	.dc.b	$3E,$69,$D6,$9E,$36,$CE,$DB,$A7,$C4,$8E,$39,$6F,$0E,$B3,$9A,$DE
	.dc.b	$3F,$EF,$E6,$DF,$D7,$C7,$C1,$7C,$2B,$4D,$9F,$4B,$5E,$E7,$8D,$C5
	.dc.b	$74,$83,$72,$FF,$06,$E1,$E3,$AD,$93,$2F,$97,$8F,$89,$F5,$13,$0F
	.dc.b	$8D,$7C,$B1,$F2,$CE,$70,$E8,$F2,$CB,$13,$AB,$F3,$4E,$5A,$4A,$E2
	.dc.b	$F6,$C3,$C4,$E6,$B9,$4F,$CF,$C3,$C5,$3F,$12,$FB,$B4,$E5,$86,$1F
	.dc.b	$87,$8F,$BD,$C9,$D4,$E1,$7E,$23,$B6,$3E,$27,$18,$F8,$8F,$C7,$5F
	.dc.b	$5C,$BE,$81,$F3,$EB,$84,$EE,$FA,$BC,$1F,$01,$E4,$C7,$CD,$97,$B2
	.dc.b	$30,$6A,$FA,$61,$DF,$2F,$56,$5E,$68,$C6,$1C,$B2,$F0,$E1,$E0,$21
	.dc.b	$F8,$8A,$F5,$B8,$C3,$AB,$DF,$0E,$B9,$79,$EB,$73,$E4,$6E,$5E,$7B
	.dc.b	$F8,$37,$C2,$BF,$39,$97,$C5,$FB,$1F,$E8,$F0,$21,$FB,$47,$2E,$0F
	.dc.b	$82,$5A,$46,$A3,$F5,$5E,$0F,$FA,$0F,$2F,$D6,$4F,$6C,$BF,$60,$D8
	.dc.b	$DD,$7E,$EE,$B8,$3E,$1E,$2F,$DF,$56,$E3,$F8,$3C,$E9,$7D,$8F,$EF
	.dc.b	$F0,$FE,$00,$BE,$67,$4D,$9F,$E3,$4E,$F1,$F7,$B6,$3E,$13,$5E,$B1
	.dc.b	$CB,$E3,$24,$72,$8E,$A6,$1F,$E3,$12,$9C,$C6,$CF,$0B,$E4,$F2,$BE
	.dc.b	$7F,$3F,$F6,$2A,$FF,$31,$D7,$FA,$A5,$AF,$F8,$11,$FF,$74,$39,$9E
	.dc.b	$3B,$7A,$BD,$3B,$3E,$A6,$BF,$1A,$6D,$3B,$F7,$DE,$FA,$8E,$B7,$FE
	.dc.b	$20,$8C,$5F,$0A,$93,$E5,$6B,$FE,$18,$FB,$1F,$C3,$67,$FD,$F1,$FF
	.dc.b	$6B,$47,$FB,$0D,$FE,$7A,$EF,$9F,$CB,$1A,$8E,$5C,$D8,$99,$1C,$63
	.dc.b	$D7,$F5,$CF,$B1,$A4,$9F,$F6,$87,$D8,$FF,$83,$31,$FF,$06,$D6,$E7
	.dc.b	$FD,$A8,$D7,$FC,$5E,$D8,$FF,$09,$CF,$1E,$0F,$D1,$CB,$EA,$FC,$24
	.dc.b	$BE,$7F,$2F,$5E,$71,$70,$F0,$CF,$D2,$FB,$17,$D2,$FA,$BE,$5E,$61
	.dc.b	$FE,$E7,$9C,$B6,$72,$F9,$A7,$94,$48,$D1,$0F,$5F,$C7,$7E,$03,$EC
	.dc.b	$5E,$25,$24,$FF,$BD,$63,$F0,$4D,$8C,$BF,$E5,$28,$E2,$EF,$8F,$A7
	.dc.b	$91,$D7,$A6,$4F,$38,$CB,$D6,$32,$F3,$8F,$C9,$FD,$8B,$EC,$7C,$DD
	.dc.b	$B2,$BF,$89,$70,$7F,$91,$2B,$51,$8E,$3F,$E2,$B1,$88,$F3,$57,$17
	.dc.b	$FC,$D3,$19,$AE,$57,$E0,$24,$FC,$1F,$B1,$FD,$2F,$63,$DB,$BE,$E4
	.dc.b	$AF,$01,$FC,$96,$74,$E0,$FD,$44,$ED,$FF,$3C,$C2,$7A,$CF,$6F,$AC
	.dc.b	$95,$C6,$91,$E2,$FA,$FC,$CF,$28,$B9,$7F,$99,$A9,$5F,$BC,$F2,$7F
	.dc.b	$C3,$D3,$9D,$BD,$66,$76,$98,$F7,$8C,$BE,$9A,$E3,$1F,$D0,$1B,$1F
	.dc.b	$D0,$DA,$FF,$8D,$67,$F8,$5F,$D8,$3E,$C6,$F5,$E3,$AE,$5F,$C5,$AF
	.dc.b	$D4,$36,$72,$FF,$A7,$A1,$3C,$A3,$87,$9B,$D8,$8F,$3E,$BA,$F1,$8B
	.dc.b	$A7,$FA,$2A,$D5,$F1,$7D,$9F,$E4,$3B,$7E,$7B,$0F,$F5,$6B,$6F,$FD
	.dc.b	$23,$AA,$B8,$57,$DD,$AD,$6D,$FD,$91,$B1,$F1,$67,$6F,$FA,$E2,$B9
	.dc.b	$3F,$36,$3C,$27,$F3,$2F,$AF,$7D,$87,$F3,$4F,$B1,$FC,$35,$7E,$2D
	.dc.b	$AF,$FB,$0A,$76,$FE,$3D,$79,$BC,$79,$24,$E6,$3F,$9F,$FD,$8B,$E5
	.dc.b	$FF,$A0,$FD,$8F,$F1,$DD,$BF,$BE,$B6,$FF,$C0,$15,$B7,$F3,$8B,$E2
	.dc.b	$38,$F8,$92,$73,$1F,$D3,$3E,$C7,$F9,$C7,$17,$FA,$EB,$07,$8F,$E0
	.dc.b	$35,$E5,$6F,$E6,$3C,$7E,$64,$CB,$F5,$58,$A7,$FA,$BF,$BA,$FB,$1F
	.dc.b	$EB,$3F,$63,$F4,$2C,$D9,$D4,$7E,$CD,$7C,$7F,$0F,$26,$4E,$63,$FB
	.dc.b	$1F,$D8,$AB,$AF,$AF,$2B,$F7,$93,$96,$FC,$07,$F6,$AF,$61,$F0,$94
	.dc.b	$E7,$1D,$3E,$53,$5F,$09,$2F,$9C,$7F,$72,$FB,$17,$C0,$F3,$F0,$E5
	.dc.b	$7F,$0B,$9C,$BF,$C3,$7E,$C5,$F2,$9F,$DE,$BE,$BF,$FA,$2C,$BA,$FF
	.dc.b	$DD,$31,$87,$FD,$D7,$5C,$27,$F8,$B7,$CE,$BE,$C7,$FC,$07,$2F,$8F
	.dc.b	$A7,$F8,$1A,$75,$5C,$3E,$27,$17,$23,$07,$4F,$6B,$C9,$E8,$48,$C1
	.dc.b	$FC,$97,$EC,$5F,$49,$D6,$A4,$E6,$3F,$C2,$B3,$A8,$C1,$0F,$77,$8F
	.dc.b	$FE,$54,$D9,$F5,$5F,$E1,$CF,$B0,$FB,$F5,$FC,$CE,$79,$3C,$0F,$F2
	.dc.b	$E3,$8E,$05,$FF,$97,$AB,$8D,$6C,$38,$9F,$9B,$8F,$D1,$3C,$1D,$B2
	.dc.b	$BF,$A1,$56,$EF,$A4,$FD,$7F,$B1,$F3,$70,$FD,$9F,$4E,$61,$C2,$32
	.dc.b	$FA,$63,$2F,$8F,$0F,$AC,$FF,$4F,$FB,$1F,$F2,$63,$FD,$53,$37,$D6
	.dc.b	$FF,$2A,$8C,$35,$F6,$4F,$06,$F7,$7F,$91,$23,$15,$F1,$27,$FD,$05
	.dc.b	$18,$D3,$FD,$06,$BB,$7F,$C9,$79,$3E,$42,$F9,$13,$0F,$EC,$93,$96
	.dc.b	$CF,$02,$3F,$DA,$7E,$C7,$A6,$5F,$E9,$1F,$B1,$FE,$73,$3F,$E9,$39
	.dc.b	$CA,$8C,$BF,$CF,$AB,$FD,$2F,$18,$BF,$04,$ED,$FF,$4D,$D7,$29,$E2
	.dc.b	$D8,$FF,$4F,$9C,$23,$95,$74,$8E,$38,$FE,$BE,$FA,$1E,$57,$15,$C2
	.dc.b	$3D,$53,$D2,$B9,$46,$F0,$CB,$4E,$67,$AC,$C9,$4C,$E1,$E7,$78,$5F
	.dc.b	$77,$8D,$FA,$4F,$3D,$72,$AF,$93,$3F,$EB,$96,$BA,$3F,$D7,$81,$CB
	.dc.b	$F4,$7D,$17,$C6,$EA,$35,$18,$78,$CF,$F4,$F9,$D3,$63,$80,$F8,$FA
	.dc.b	$D3,$2E,$08,$FC,$4B,$FF,$55,$6D,$F1,$23,$FD,$AD,$3F,$1A,$73,$1F
	.dc.b	$28,$C9,$C7,$7F,$CE,$67,$F8,$88,$CB,$F8,$A9,$E2,$E6,$3C,$8E,$67
	.dc.b	$8E,$7F,$1A,$FC,$11,$8A,$FF,$72,$46,$30,$F3,$47,$2A,$FF,$5E,$D6
	.dc.b	$E3,$FD,$DB,$5B,$7D,$06,$CF,$F3,$0C,$6F,$0F,$69,$FE,$9B,$7F,$6C
	.dc.b	$FE,$DE,$39,$9E,$FB,$FF,$2A,$47,$07,$C2,$7F,$95,$E3,$E0,$DE,$1F
	.dc.b	$E4,$1F,$63,$FB,$B9,$FF,$2E,$36,$32,$78,$20,$C8,$D5,$78,$DB,$3A
	.dc.b	$94,$61,$E0,$73,$39,$6B,$5D,$F3,$59,$FE,$93,$5F,$C7,$FE,$C7,$FD
	.dc.b	$52,$F9,$A7,$47,$3C,$12,$F9,$EC,$D7,$17,$A7,$9F,$23,$F9,$3F,$5D
	.dc.b	$CB,$E1,$CD,$C3,$E7,$F3,$7F,$50,$BF,$8B,$39,$7F,$D2,$F1,$FE,$C5
	.dc.b	$72,$7F,$32,$F8,$0F,$67,$D8,$BC,$AF,$77,$CA,$FF,$71,$AF,$F4,$54
	.dc.b	$E5,$FF,$46,$46,$72,$FE,$A8,$DC,$BF,$00,$D9,$F9,$D5,$C6,$B7,$D9
	.dc.b	$F2,$FE,$BD,$F1,$94,$F0,$C5,$BF,$F4,$D4,$BE,$53,$81,$F4,$FE,$C7
	.dc.b	$FD,$49,$3A,$8F,$AD,$F6,$2F,$92,$A7,$09,$FF,$52,$46,$5B,$F3,$9C
	.dc.b	$0E,$3C,$D6,$3F,$D5,$5F,$63,$FE,$5C,$6F,$C1,$7D,$BF,$37,$FC,$4F
	.dc.b	$18,$3F,$D6,$07,$F9,$74,$BE,$17,$E0,$7E,$C1,$F6,$3F,$8B,$9D,$71
	.dc.b	$7F,$7D,$FB,$1F,$CC,$C7,$FA,$E5,$B1,$FE,$07,$E6,$5F,$F8,$C2,$B5
	.dc.b	$3F,$95,$A7,$C0,$53,$F2,$1F,$24,$F5,$F1,$7D,$7B,$78,$1A,$F9,$2B
	.dc.b	$1F,$EC,$9F,$19,$EA,$FE,$85,$B1,$FD,$F7,$EC,$7F,$75,$A7,$F2,$0A
	.dc.b	$D7,$A3,$2B,$FC,$05,$F6,$2F,$A9,$C5,$FF,$86,$FE,$C7,$FC,$BD,$5F
	.dc.b	$ED,$76,$C7,$BF,$EC,$7F,$BE,$4E,$7E,$0F,$F6,$55,$FE,$A3,$8E,$1F
	.dc.b	$23,$FA,$DA,$7C,$0E,$A3,$6C,$9F,$F0,$A7,$D8,$FE,$41,$F0,$47,$EF
	.dc.b	$7D,$6F,$B7,$97,$F8,$BD,$BD,$1F,$C7,$30,$46,$FE,$0B,$F8,$27,$D8
	.dc.b	$FD,$CC,$7F,$DE,$0D,$C3,$EC,$5F,$EC,$5F,$F1,$D4,$67,$0F,$E4,$DF
	.dc.b	$63,$FE,$9B,$BF,$F7,$C3,$6B,$F0,$1F,$5E,$FA,$CF,$E2,$53,$9B,$E3
	.dc.b	$F2,$18,$3F,$87,$56,$F0,$E7,$19,$7F,$C6,$91,$FB,$39,$E3,$18,$F3
	.dc.b	$FF,$8B,$7D,$8F,$C5,$CB,$F2,$0D,$AF,$E7,$AB,$E3,$37,$E7,$CC,$3F
	.dc.b	$C8,$13,$B7,$F9,$9D,$23,$84,$6A,$7C,$FF,$11,$E8,$E9,$B3,$E6,$6E
	.dc.b	$3F,$09,$B8,$7F,$4E,$CF,$FD,$57,$5D,$8F,$E7,$9F,$62,$F9,$BF,$E7
	.dc.b	$0D,$8F,$2E,$D8,$99,$1C,$71,$FE,$93,$5B,$7E,$0D,$69,$FF,$39,$46
	.dc.b	$70,$FE,$BF,$F5,$BF,$86,$BF,$0E,$1F,$CF,$E3,$17,$FE,$A3,$C7,$F8
	.dc.b	$BB,$F7,$3C,$0F,$AF,$E1,$FD,$86,$61,$F5,$9C,$B4,$F9,$BE,$87,$D6
	.dc.b	$FB,$1F,$AE,$D8,$C9,$F4,$A3,$2D,$3D,$D6,$D3,$FA,$94,$61,$B1,$F9
	.dc.b	$0E,$FF,$93,$ED,$4A,$D3,$F8,$E8,$C4,$FC,$48,$FE,$BD,$F6,$2C,$33
	.dc.b	$BF,$F8,$12,$31,$87,$E1,$BE,$C7,$F1,$FE,$77,$E4,$1B,$19,$3F,$17
	.dc.b	$19,$69,$F9,$AF,$3F,$F3,$E7,$E8,$3D,$7B,$19,$26,$7F,$AA,$8C,$BF
	.dc.b	$EA,$BF,$B1,$FD,$33,$7B,$7F,$43,$DF,$FD,$FE,$73,$A7,$F8,$6D,$B8
	.dc.b	$7F,$82,$5B,$7F,$ED,$FE,$BF,$E2,$8F,$39,$F9,$F3,$AA,$73,$7F,$77
	.dc.b	$AD,$1F,$CF,$5C,$47,$E7,$7E,$47,$F7,$67,$F8,$27,$D5,$7B,$78,$C7
	.dc.b	$F7,$73,$70,$FE,$8D,$5A,$9C,$9F,$B4,$FA,$8F,$DE,$9F,$BE,$F8,$EB
	.dc.b	$19,$25,$7F,$00,$8C,$BF,$ED,$1F,$B0,$FE,$6B,$97,$F8,$F5,$B5,$FE
	.dc.b	$30,$F8,$9F,$F1,$5F,$89,$FE,$27,$9C,$DF,$F8,$D1,$A9,$7D,$8E,$FC
	.dc.b	$EE,$0F,$1F,$16,$63,$F9,$8A,$FE,$87,$7E,$39,$CB,$67,$5F,$17,$8C
	.dc.b	$FC,$DD,$1E,$33,$FC,$79,$A7,$CA,$F8,$9C,$5F,$59,$F4,$6F,$8C,$7E
	.dc.b	$03,$FE,$41,$F5,$FF,$C8,$D1,$FD,$BB,$E2,$31,$E7,$65,$98,$7F,$A0
	.dc.b	$BE,$97,$FC,$9D,$7F,$C2,$FD,$74,$62,$BF,$DB,$91,$8C,$3F,$CB,$35
	.dc.b	$BA,$E3,$E4,$E9,$B8,$75,$AD,$E1,$97,$8D,$57,$FC,$06,$DB,$E9,$1C
	.dc.b	$2B,$93,$A6,$C7,$F6,$2C,$3C,$93,$FE,$0C,$F4,$39,$7B,$FC,$D9,$7C
	.dc.b	$BE,$C7,$FC,$1D,$A7,$29,$F1,$46,$2F,$FD,$5D,$A7,$F9,$C6,$31,$8F
	.dc.b	$66,$C7,$C8,$AE,$3E,$DF,$F7,$0B,$E2,$7F,$9F,$AB,$6D,$8D,$5F,$A3
	.dc.b	$3F,$56,$1D,$E7,$2D,$C7,$D1,$18,$BF,$A3,$8F,$82,$72,$48,$F0,$9F
	.dc.b	$0F,$2C,$63,$E9,$BE,$52,$CB,$CA,$EC,$E2,$E0,$91,$C3,$2F,$3C,$FA
	.dc.b	$F0,$EB,$97,$9E,$3C,$33,$A7,$B5,$78,$2F,$9D,$FD,$1C,$78,$D7,$33
	.dc.b	$C3,$5D,$F2,$F7,$46,$2B,$6F,$9F,$2F,$64,$63,$0F,$5E,$3E,$17,$BD
	.dc.b	$71,$3C,$B7,$D2,$F8,$C6,$5F,$A9,$19,$CB,$D9,$1A,$33,$97,$7C,$BD
	.dc.b	$87,$9D,$C6,$5E,$6C,$BC,$53,$A8,$C6,$1D,$71,$F8,$18,$75,$C7,$F1
	.dc.b	$98,$F2,$9F,$24,$61,$A7,$07,$CA,$66,$79,$46,$48,$7C,$78,$DB,$8C
	.dc.b	$BE,$3D,$F8,$E3,$EB,$E1,$E6,$C3,$C5,$87,$49,$E4,$71,$FB,$1F,$E7
	.dc.b	$EF,$FB,$5C,$FD,$03,$87,$18,$7E,$B5,$B1,$97,$F5,$CF,$29,$FD,$73
	.dc.b	$71,$CF,$6F,$EE,$7C,$4F,$F8,$03,$1F,$B5,$6D,$F8,$27,$F7,$D5,$C6
	.dc.b	$3F,$65,$3A,$7F,$40,$DC,$13,$B9,$FD,$23,$61,$3C,$7E,$C7,$F8,$7E
	.dc.b	$3F,$C4,$17,$CD,$64,$D6,$53,$1F,$E3,$33,$C2,$3F,$DF,$8D,$80,$5C
	.dc.b	$BD,$CE,$51,$FB,$C8,$E0,$EE,$34,$F6,$9D,$DF,$96,$7D,$1E,$6F,$6A
	.dc.b	$74,$36,$F1,$FA,$99,$8C,$4F,$93,$2F,$F3,$53,$FE,$34,$7F,$DB,$CD
	.dc.b	$BE,$17,$CA,$35,$78,$79,$4D,$7D,$DE,$06,$7A,$9F,$53,$D2,$7C,$B8
	.dc.b	$7F,$77,$8C,$65,$E4,$9E,$27,$8E,$FA,$64,$69,$DF,$EC,$7F,$29,$5C
	.dc.b	$5F,$F6,$8D,$7F,$93,$9F,$E7,$0F,$80,$FA,$6E,$23,$87,$3F,$33,$A7
	.dc.b	$2C,$AE,$51,$ED,$FA,$07,$D8,$D2,$4F,$F9,$73,$EC,$7D,$58,$7A,$EB
	.dc.b	$73,$FE,$60,$6B,$E3,$0E,$23,$E1,$4F,$2D,$7E,$E4,$BE,$3B,$C4,$4B
	.dc.b	$E9,$F3,$75,$E9,$3B,$7C,$13,$E9,$9F,$AD,$F6,$2F,$A5,$F6,$7C,$B4
	.dc.b	$BF,$F3,$AC,$E5,$B3,$97,$AC,$F2,$89,$1A,$21,$E2,$F8,$FF,$C2,$7D
	.dc.b	$8B,$C5,$A4,$9F,$F4,$1C,$7D,$D6,$C6,$5F,$8D,$1C,$5D,$F1,$F4,$F2
	.dc.b	$3A,$F4,$C9,$E7,$19,$7A,$C6,$5E,$71,$F9,$7F,$B1,$7D,$8F,$9B,$BA
	.dc.b	$57,$FA,$5A,$72,$D9,$CB,$E4,$9E,$51,$23,$44,$3D,$3F,$63,$FA,$5E
	.dc.b	$B7,$BD,$7D,$F9,$5F,$EA,$49,$CB,$67,$2F,$E1,$E1,$3C,$E7,$A7,$D5
	.dc.b	$CA,$E4,$F7,$FA,$FC,$47,$18,$BD,$5F,$AF,$C1,$5F,$B9,$F2,$BF,$C5
	.dc.b	$F3,$9E,$AF,$41,$9D,$66,$3E,$78,$CB,$FE,$BC,$AE,$31,$FC,$4D,$B1
	.dc.b	$FC,$89,$A9,$BF,$60,$FB,$1F,$E1,$4F,$E6,$67,$F4,$CB,$AB,$26,$BA
	.dc.b	$E4,$FF,$B3,$27,$4D,$9C,$BC,$BC,$EE,$A4,$78,$75,$D7,$47,$38,$F1
	.dc.b	$C7,$C5,$F6,$FF,$8F,$6D,$F9,$7A,$FD,$D3,$76,$FE,$35,$A2,$B8,$57
	.dc.b	$FB,$82,$B5,$AF,$F3,$76,$C7,$C5,$9D,$BF,$C1,$6B,$93,$F3,$63,$C2
	.dc.b	$7F,$2F,$FA,$F7,$D8,$FF,$31,$FB,$1F,$C2,$D7,$FB,$B9,$AF,$F0,$F9
	.dc.b	$DB,$FE,$F2,$7F,$B0,$BB,$E3,$E4,$75,$F3,$7F,$9E,$7D,$8B,$E6,$7F
	.dc.b	$9F,$7D,$8F,$F1,$7D,$BF,$A8,$37,$0F,$EB,$F5,$B7,$FD,$FE,$BE,$2F
	.dc.b	$8F,$89,$27,$31,$FD,$27,$EC,$7F,$9C,$C7,$F2,$5B,$FE,$4A,$FF,$6A
	.dc.b	$E2,$FE,$8D,$18,$8F,$E5,$53,$C2,$7F,$A5,$2F,$98,$F3,$F9,$92,$B9
	.dc.b	$67,$FD,$9E,$B5,$18,$3F,$BB,$FB,$CF,$B1,$FE,$B9,$F6,$3F,$5A,$4D
	.dc.b	$9D,$47,$10,$BE,$43,$8F,$8F,$27,$53,$97,$87,$D8,$F0,$79,$5F,$F8
	.dc.b	$41,$BE,$C7,$F6,$D9,$D4,$7F,$49,$5F,$01,$FD,$C7,$D8,$7C,$2D,$38
	.dc.b	$47,$1F,$95,$D7,$C2,$CC,$3C,$31,$FD,$E7,$EC,$5F,$03,$CF,$C1,$95
	.dc.b	$ED,$9C,$BE,$FF,$B1,$7C,$AF,$F8,$07,$EB,$FF,$A4,$4F,$F5,$D6,$F4
	.dc.b	$FF,$7D,$AD,$CF,$C7,$F9,$D7,$D8,$FF,$82,$65,$F2,$34,$FF,$06,$4E
	.dc.b	$AB,$87,$C5,$62,$E4,$60,$E9,$EC,$78,$FC,$D9,$18,$3E,$B7,$D8,$BE
	.dc.b	$2B,$AD,$49,$CC,$7F,$86,$E7,$51,$82,$1E,$DF,$AA,$FF,$10,$7D,$87
	.dc.b	$DB,$8F,$81,$5F,$0C,$CF,$74,$8F,$C3,$70,$74,$76,$7E,$26,$B9,$70
	.dc.b	$7F,$88,$DB,$E0,$FF,$8F,$5B,$3F,$E4,$C6,$AA,$7D,$8F,$93,$87,$EA
	.dc.b	$7A,$32,$BB,$46,$5F,$1C,$65,$FF,$0A,$5F,$D2,$7F,$39,$F6,$3F,$E5
	.dc.b	$87,$F4,$D9,$3E,$B7,$F9,$7C,$7F,$8B,$63,$1C,$3F,$E4,$3A,$E1,$3F
	.dc.b	$E1,$F6,$CF,$F8,$FD,$BA,$7F,$96,$5B,$3F,$E6,$87,$F3,$AF,$EB,$71
	.dc.b	$7C,$87,$F9,$3D,$FF,$1D,$57,$F9,$85,$B3,$FE,$70,$76,$F1,$9E,$04
	.dc.b	$7F,$71,$F6,$3F,$AA,$C3,$FC,$EF,$F6,$2B,$FF,$36,$4E,$69,$39,$7B
	.dc.b	$5F,$F9,$C6,$31,$1C,$63,$15,$CB,$BF,$FC,$EB,$3C,$A3,$18,$78,$1F
	.dc.b	$55,$70,$9E,$A6,$4F,$F4,$5E,$1E,$3C,$3B,$B6,$B8,$CE,$CF,$11,$63
	.dc.b	$4F,$2C,$3C,$75,$E8,$AF,$8F,$3E,$53,$C2,$51,$5C,$9B,$5C,$A7,$C3
	.dc.b	$1B,$79,$D7,$9A,$75,$3E,$1C,$3F,$D9,$8D,$2D,$FC,$F7,$FD,$A0,$D7
	.dc.b	$E9,$F9,$EF,$8D,$E3,$1E,$48,$C3,$C9,$AF,$26,$BE,$0F,$8F,$E2,$FF
	.dc.b	$AD,$9A,$9A,$47,$E1,$5F,$FA,$EB,$4F,$9D,$1C,$4C,$F8,$9F,$EA,$EC
	.dc.b	$FF,$07,$19,$7F,$0B,$5C,$63,$E6,$19,$D3,$FD,$6D,$5C,$5F,$34,$74
	.dc.b	$BE,$EF,$94,$C1,$C8,$C5,$F6,$AE,$51,$FE,$F9,$C3,$F0,$98,$78,$6B
	.dc.b	$CB,$3C,$30,$F0,$E1,$DA,$F7,$87,$2A,$FF,$66,$BF,$EC,$AC,$3F,$D3
	.dc.b	$ED,$AF,$3E,$1E,$17,$FB,$24,$FB,$CD,$BF,$D9,$EB,$E0,$1F,$DF,$6B
	.dc.b	$93,$FE,$E8,$FB,$1F,$DF,$CF,$F6,$D6,$C7,$FA,$A5,$FF,$76,$C6,$76
	.dc.b	$E3,$29,$DD,$FF,$B3,$5B,$5F,$EE,$77,$8C,$F3,$AE,$31,$FD,$16,$7C
	.dc.b	$1C,$FE,$4B,$7D,$8F,$FA,$A6,$72,$4C,$3F,$91,$43,$E7,$F9,$DF,$18
	.dc.b	$C5,$7A,$FD,$0E,$E7,$5E,$57,$5E,$27,$44,$C7,$F9,$74,$EA,$31,$97
	.dc.b	$B6,$F9,$7B,$DC,$6F,$6F,$C4,$FB,$17,$93,$F1,$BE,$5B,$FB,$7E,$3F
	.dc.b	$E1,$99,$CB,$FE,$1B,$8C,$DF,$FB,$5D,$B8,$FF,$BD,$DB,$29,$95,$C9
	.dc.b	$E1,$D9,$F5,$3E,$C5,$F1,$54,$F1,$45,$BF,$F1,$54,$BE,$57,$81,$F7
	.dc.b	$3E,$C7,$FD,$53,$3A,$8F,$BF,$F6,$2F,$90,$A7,$09,$FF,$1A,$46,$5B
	.dc.b	$F3,$9C,$0E,$3C,$B6,$3F,$C7,$5F,$62,$F8,$0F,$C5,$B6,$3F,$B0,$F9
	.dc.b	$9F,$E2,$67,$73,$F9,$07,$FC,$CC,$7F,$90,$CB,$BF,$F6,$5F,$B0,$7D
	.dc.b	$8E,$FC,$2F,$CB,$FD,$8F,$E4,$B9,$98,$7C,$BC,$13,$98,$FF,$31,$61
	.dc.b	$FE,$30,$7F,$CA,$11,$99,$FA,$4B,$E6,$BA,$9F,$99,$F9,$1F,$F5,$85
	.dc.b	$70,$8F,$0F,$5C,$8F,$EF,$B3,$F3,$AD,$8F,$91,$EB,$5F,$AE,$F1,$1E
	.dc.b	$95,$3F,$4D,$F6,$3F,$BF,$C7,$D5,$3A,$7F,$CC,$CF,$F9,$95,$BE,$87
	.dc.b	$EB,$7E,$C5,$F5,$5D,$63,$F7,$7F,$63,$FE,$6C,$AF,$F3,$73,$73,$FF
	.dc.b	$08,$7D,$8F,$F9,$8A,$73,$F0,$5C,$3E,$3E,$47,$F8,$62,$5E,$FE,$2F
	.dc.b	$DB,$3F,$3F,$0C,$BD,$A7,$F8,$07,$D8,$FE,$62,$30,$7F,$9F,$FE,$23
	.dc.b	$F0,$57,$C8,$FD,$13,$8E,$2F,$E4,$D9,$23,$F8,$A9,$9A,$FB,$DF,$63
	.dc.b	$F8,$47,$FD,$14,$DB,$FF,$16,$53,$C9,$FE,$3B,$39,$C3,$EA,$53,$EC
	.dc.b	$7F,$D5,$D8,$7F,$A4,$DB,$1F,$E3,$5F,$B1,$7D,$6F,$F2,$69,$CD,$F2
	.dc.b	$6D,$FE,$45,$B7,$F8,$2B,$E3,$3D,$7E,$3B,$91,$FC,$96,$3F,$80,$5F
	.dc.b	$38,$C7,$9B,$FC,$A7,$EB,$7E,$56,$1F,$E4,$76,$D7,$F4,$B5,$F1,$9B
	.dc.b	$F3,$66,$5F,$E5,$09,$DB,$FE,$C3,$76,$E6,$3C,$5F,$0B,$E3,$E7,$B3
	.dc.b	$FE,$58,$D3,$F2,$19,$7C,$3A,$D4,$E5,$B1,$F6,$C7,$D8,$BE,$63,$FA
	.dc.b	$6B,$6B,$C9,$B6,$26,$47,$FB,$0E,$75,$5E,$FA,$CB,$D6,$33,$AF,$F6
	.dc.b	$BF,$AD,$FC,$24,$FF,$2B,$9C,$DF,$FB,$86,$73,$4C,$BF,$6C,$FF,$73
	.dc.b	$6C,$FF,$B9,$D7,$83,$F5,$7E,$7A,$4F,$C3,$7E,$C4,$63,$1F,$A9,$E7
	.dc.b	$FD,$7F,$B1,$FB,$0D,$BF,$A4,$D9,$DC,$7D,$59,$F8,$BE,$CC,$8F,$A1
	.dc.b	$E0,$5A,$F8,$13,$73,$F4,$4D,$AF,$E4,$1A,$FF,$66,$BF,$F4,$67,$D8
	.dc.b	$B1,$D4,$7F,$14,$8C,$57,$E1,$7E,$C7,$F1,$DE,$77,$E3,$DB,$19,$3F
	.dc.b	$15,$19,$69,$F8,$5F,$3B,$F3,$47,$E6,$FD,$6B,$19,$26,$1F,$AB,$8C
	.dc.b	$BF,$87,$FB,$1F,$D2,$B7,$B7,$F4,$1E,$07,$F8,$52,$73,$BF,$EE,$1B
	.dc.b	$87,$F3,$F6,$DF,$F3,$0E,$B5,$F1,$C2,$5F,$FA,$CB,$9B,$FC,$05,$7D
	.dc.b	$E3,$1E,$DF,$EC,$FE,$46,$47,$FA,$E3,$D5,$7B,$78,$BF,$E0,$95,$75
	.dc.b	$FF,$1E,$56,$AF,$F5,$13,$FB,$9F,$A7,$91,$FB,$3F,$8D,$B8,$7E,$EE
	.dc.b	$6E,$1F,$BB,$9B,$5F,$D6,$FE,$C3,$F9,$C4,$FF,$18,$6C,$7F,$82,$7E
	.dc.b	$1B,$FC,$71,$E2,$FF,$8E,$A7,$38,$FF,$83,$5A,$97,$D8,$EF,$CF,$E0
	.dc.b	$F1,$F1,$A6,$3F,$E1,$5C,$BC,$1F,$13,$8C,$7F,$7D,$8F,$F0,$C7,$B5
	.dc.b	$F8,$67,$55,$E7,$E9,$F2,$1F,$E5,$6D,$3E,$5F,$C3,$63,$2E,$C7,$BE
	.dc.b	$7D,$5B,$E7,$2E,$EF,$F9,$5B,$D7,$FF,$29,$C7,$F8,$A5,$AF,$F7,$9F
	.dc.b	$3B,$5F,$0D,$30,$FF,$4B,$7D,$27,$F8,$16,$FF,$86,$FA,$C8,$C5,$7F
	.dc.b	$77,$8C,$61,$D7,$CB,$C5,$71,$6C,$67,$0E,$AF,$8E,$F2,$FF,$82,$E6
	.dc.b	$E1,$E0,$BE,$57,$C2,$BF,$C2,$17,$AB,$DF,$CA,$CA,$D6,$1F,$E6,$DB
	.dc.b	$FF,$57,$61,$EB,$AD,$46,$0F,$F2,$67,$D8,$3E,$5B,$FB,$54,$64,$FF
	.dc.b	$0F,$4F,$9A,$75,$A7,$8D,$B3,$B6,$E9,$F1,$23,$8E,$5B,$C3,$AC,$E6
	.dc.b	$B7,$8F,$F8,$CD,$B7,$F5,$F1,$F0,$5F,$0A,$D3,$67,$D2,$D7,$B9,$E3
	.dc.b	$71,$5D,$20,$DC,$BF,$C1,$B8,$78,$EB,$64,$CB,$E5,$E3,$E2,$7D,$44
	.dc.b	$C3,$E3,$5F,$2C,$7C,$B3,$9C,$3A,$3C,$B2,$C4,$EA,$FC,$D3,$96,$92
	.dc.b	$B8,$BD,$B0,$F1,$39,$AE,$53,$F3,$F0,$F1,$4F,$C4,$BE,$ED,$39,$61
	.dc.b	$87,$E1,$E3,$EF,$72,$75,$38,$5F,$88,$ED,$8F,$89,$C6,$3E,$23,$F1
	.dc.b	$D7,$D7,$2F,$A0,$7C,$FA,$E1,$3B,$BE,$AF,$07,$C0,$79,$31,$F3,$65
	.dc.b	$EC,$8C,$1A,$BE,$98,$77,$CB,$D5,$97,$9A,$31,$87,$2C,$BC,$38,$78
	.dc.b	$08,$7E,$22,$BD,$6E,$30,$EA,$F7,$C3,$AE,$5E,$7A,$DC,$FE,$33,$EC
	.dc.b	$7F,$A0,$BF,$E8,$D3,$F5,$AE,$1C,$5F,$EA,$1B,$19,$7F,$44,$F2,$9F
	.dc.b	$D3,$36,$B3,$D9,$FB,$1F,$13,$FD,$03,$18,$BF,$D9,$56,$DF,$0D,$6D
	.dc.b	$86,$9F,$DD,$46,$E3,$F4,$B5,$D1,$FE,$0A,$D7,$DD,$3C,$0E,$77,$FB
	.dc.b	$FF,$B1,$FE,$00,$DF,$9A,$E1,$06,$E3,$2F,$F8,$E4,$F0,$8F,$F5,$A3
	.dc.b	$63,$FD,$C0,$B9,$39,$9C,$63,$F8,$B4,$6D,$3B,$7A,$C6,$EF,$CD,$5C
	.dc.b	$A7,$C3,$E6,$FF,$B0,$F9,$FF,$CF,$F8,$7F,$99,$A7,$FA,$79,$E2,$F5
	.dc.b	$35,$C5,$E6,$BB,$3E,$66,$BE,$29,$B5,$CE,$3F,$CC,$0F,$FB,$C5,$EA
	.dc.b	$7A,$3B,$F9,$F4,$BE,$A3,$53,$FE,$A1,$8C,$61,$D2,$A4,$F2,$B4,$62
	.dc.b	$3F,$CD,$DF,$63,$F9,$3C,$FF,$D3,$6F,$FB,$66,$3C,$2F,$F4,$07,$7C
	.dc.b	$DE,$A8,$D4,$72,$E7,$C4,$C8,$E3,$1F,$03,$EB,$DF,$63,$49,$3E,$EF
	.dc.b	$B1,$FF,$55,$E3,$FE,$AC,$AD,$CF,$C2,$6B,$FE,$85,$6C,$7F,$AC,$A7
	.dc.b	$8F,$07,$FB,$E6,$7F,$D7,$15,$C7,$DC,$F1,$1E,$3F,$97,$B1,$92,$1D
	.dc.b	$5F,$AF,$F6,$2F,$A5,$F5,$7C,$BC,$AF,$AB,$39,$6C,$E5,$F3,$4F,$28
	.dc.b	$91,$A2,$1E,$BF,$90,$FC,$2F,$D8,$BC,$4A,$49,$FC,$1C,$7D,$A6,$C6
	.dc.b	$5F,$F5,$44,$71,$77,$C7,$D3,$C7,$EB,$D3,$27,$9C,$65,$EB,$19,$79
	.dc.b	$C7,$E6,$3E,$C5,$F6,$3E,$6E,$C9,$5F,$96,$9C,$B6,$72,$F9,$27,$94
	.dc.b	$48,$D1,$0F,$4F,$D8,$FE,$9B,$B5,$EF,$DF,$7E,$57,$E9,$27,$2D,$9C
	.dc.b	$BF,$EC,$98,$FF,$58,$B9,$38,$7D,$5F,$52,$3C,$FE,$BE,$BC,$62,$F5
	.dc.b	$7F,$BF,$A9,$5F,$BA,$F2,$7F,$C3,$53,$9E,$AF,$41,$9D,$66,$3E,$78
	.dc.b	$CB,$FB,$CA,$E3,$08,$91,$86,$CF,$27,$FD,$C9,$F6,$0F,$B1,$FE,$16
	.dc.b	$FE,$7A,$33,$D5,$83,$3D,$72,$3F,$85,$CA,$75,$18,$9E,$7E,$6C,$8D
	.dc.b	$BD,$BA,$F1,$1E,$28,$BA,$34,$F5,$4B,$15,$FC,$7F,$6B,$F9,$0E,$DF
	.dc.b	$99,$AF,$F7,$F3,$70,$E5,$1F,$0A,$78,$3F,$C8,$A7,$58,$76,$6C,$6C
	.dc.b	$F9,$93,$C1,$57,$47,$FC,$B5,$1E,$23,$F9,$9F,$D7,$BE,$C7,$F9,$AF
	.dc.b	$D8,$FE,$22,$B8,$E3,$EC,$8C,$61,$CE,$F8,$57,$F3,$65,$E7,$72,$E4
	.dc.b	$93,$98,$FE,$83,$F6,$2F,$98,$FE,$87,$F6,$3F,$C6,$B3,$F9,$2D,$BF
	.dc.b	$C2,$56,$DF,$E8,$6B,$E2,$F8,$F8,$92,$73,$1F,$D3,$BE,$C7,$F9,$AE
	.dc.b	$BF,$07,$07,$8F,$E0,$35,$E5,$6F,$E6,$3C,$7E,$64,$CB,$FA,$8E,$29
	.dc.b	$FE,$B3,$EE,$BE,$C7,$FA,$DF,$D8,$FD,$2B,$36,$75,$1F,$D6,$97,$C8
	.dc.b	$F0,$F2,$64,$E6,$3F,$B2,$FD,$8A,$BA,$FB,$12,$BF,$B1,$CE,$5B,$EF
	.dc.b	$FF,$6C,$F6,$5F,$07,$4E,$71,$D3,$E5,$B5,$F0,$92,$F9,$C7,$F7,$4F
	.dc.b	$B1,$7C,$0F,$3F,$2E,$57,$F7,$29,$CB,$FD,$D3,$EC,$5F,$2B,$FD,$F3
	.dc.b	$EB,$FF,$A1,$4F,$E2,$0F,$DE,$38,$EB,$FC,$95,$70,$9F,$EF,$BF,$3C
	.dc.b	$FB,$1F,$F0,$2C,$BE,$46,$9F,$E0,$89,$D5,$70,$F8,$9C,$5C,$8C,$1D
	.dc.b	$3D,$AF,$27,$A3,$23,$07,$F8,$33,$EC,$5F,$4B,$D6,$A4,$E6,$3F,$C2
	.dc.b	$F3,$A8,$C1,$0F,$70,$D5,$F1,$FF,$E1,$EF,$B0,$FB,$F3,$FE,$1A,$9E
	.dc.b	$4F,$06,$EF,$FB,$59,$DC,$F4,$6C,$7E,$E6,$77,$E4,$FF,$0E,$6D,$7F
	.dc.b	$29,$AD,$CF,$F1,$26,$C6,$7B,$7C,$87,$D7,$FB,$1F,$37,$0F,$D9,$F4
	.dc.b	$E5,$7F,$8C,$2F,$E9,$36,$72,$FA,$67,$8B,$FE,$36,$FB,$1F,$F2,$63
	.dc.b	$F8,$1C,$DF,$5B,$FC,$BE,$3F,$C8,$F1,$8E,$1F,$F2,$1D,$70,$8F,$E2
	.dc.b	$CD,$D3,$F9,$5B,$6B,$FA,$33,$52,$EC,$FF,$26,$E2,$F9,$0B,$E4,$4B
	.dc.b	$FF,$25,$CE,$5B,$3C,$08,$FF,$94,$3E,$C7,$A6,$5F,$D5,$7E,$C7,$F9
	.dc.b	$DC,$FF,$57,$9C,$A8,$CB,$FD,$02,$BF,$A1,$C6,$2F,$C1,$3B,$7F,$A3
	.dc.b	$D7,$29,$E2,$D8,$FE,$96,$70,$8E,$55,$D2,$38,$E3,$FC,$FA,$FA,$1E
	.dc.b	$57,$15,$C2,$3D,$53,$D2,$B9,$46,$F0,$CB,$4E,$67,$AC,$C9,$4C,$E1
	.dc.b	$E7,$78,$5F,$77,$8D,$FA,$4F,$3D,$72,$AF,$93,$3F,$EC,$36,$BA,$3F
	.dc.b	$D8,$C1,$CB,$F4,$7D,$17,$C7,$6A,$35,$18,$78,$CF,$F4,$F9,$D3,$63
	.dc.b	$80,$F8,$FA,$D3,$2E,$08,$FC,$4B,$FF,$58,$6D,$F1,$23,$FC,$35,$3F
	.dc.b	$1A,$73,$1F,$28,$C9,$C7,$C5,$FE,$FF,$9F,$E2,$23,$2F,$E2,$A7,$8B
	.dc.b	$98,$F2,$39,$9E,$39,$FC,$6B,$F0,$46,$2B,$FD,$DD,$18,$C3,$CD,$1C
	.dc.b	$AB,$FD,$79,$5B,$8F,$F0,$CD,$6D,$F4,$1B,$3F,$C6,$D8,$79,$2B,$8D
	.dc.b	$FF,$AD,$A3,$DE,$7F,$83,$E7,$D7,$3A,$7F,$93,$C6,$A7,$0E,$E3,$B1
	.dc.b	$FE,$18,$BF,$F1,$1C,$7F,$86,$7E,$C7,$F7,$95,$FE,$C4,$6D,$7F,$32
	.dc.b	$7F,$C9,$51,$9D,$78,$BF,$E3,$48,$DD,$73,$9C,$D7,$F9,$46,$78,$BC
	.dc.b	$E3,$2A,$3F,$A3,$D7,$87,$AB,$8D,$BE,$C7,$FD,$4D,$39,$25,$FF,$23
	.dc.b	$87,$D0,$F4,$C3,$9C,$E6,$31,$E7,$CB,$FE,$4F,$CD,$FD,$31,$F8,$11
	.dc.b	$9A,$F8,$0F,$8A,$75,$18,$9D,$9C,$BE,$03,$86,$5F,$32,$91,$EE,$FB
	.dc.b	$17,$95,$F0,$3E,$53,$FB,$96,$1F,$EF,$29,$CB,$FE,$F4,$8C,$E3,$FD
	.dc.b	$89,$B8,$FF,$93,$DB,$3F,$E6,$4A,$E3,$5B,$EB,$F9,$BF,$62,$F8,$CA
	.dc.b	$78,$62,$DF,$E4,$0F,$F3,$6C,$BE,$57,$89,$F5,$BE,$C7,$FD,$4F,$3B
	.dc.b	$C3,$FD,$19,$5B,$8C,$43,$EC,$5F,$25,$F7,$9B,$3A,$AE,$15,$FD,$FA
	.dc.b	$72,$DF,$9C,$E0,$71,$E6,$B7,$DB,$EC,$7F,$CC,$6D,$F8,$2F,$C3,$F9
	.dc.b	$BF,$E2,$E8,$C1,$E4,$3F,$CC,$85,$F0,$7F,$1B,$F6,$0F,$B1,$FC,$9C
	.dc.b	$EB,$3F,$F0,$8F,$D8,$FE,$63,$9B,$FC,$FC,$D9,$FD,$BD,$FF,$8D,$9F
	.dc.b	$83,$19,$AF,$F5,$82,$F8,$0F,$CF,$B6,$7F,$BB,$FC,$93,$D7,$CD,$6B
	.dc.b	$B6,$56,$BE,$4A,$E3,$FE,$24,$F1,$5E,$AF,$EA,$9B,$5F,$E0,$2F,$B1
	.dc.b	$FE,$05,$5F,$CA,$2B,$5E,$8C,$AF,$F0,$3F,$D8,$BE,$AF,$17,$FE,$30
	.dc.b	$FB,$0F,$F3,$35,$7D,$C6,$C7,$F8,$47,$EC,$7F,$C0,$93,$9F,$83,$FD
	.dc.b	$DD,$3E,$46,$9F,$BA,$A7,$C0,$E2,$FA,$3C,$0F,$DF,$FD,$8F,$E5,$E3
	.dc.b	$27,$E3,$3D,$5F,$B9,$97,$F9,$11,$BD,$1F,$C6,$B1,$47,$F0,$C3,$35
	.dc.b	$FE,$C8,$FB,$1F,$BC,$FE,$59,$B7,$FE,$2B,$AF,$DB,$CE,$7A,$BF,$8A
	.dc.b	$FD,$8F,$FA,$8E,$3F,$3A,$DA,$FF,$19,$7D,$8B,$EA,$F7,$F2,$12,$F6
	.dc.b	$FF,$1A,$9D,$46,$09,$5C,$BC,$EE,$0D,$8F,$F1,$DF,$D6,$7C,$38,$FF
	.dc.b	$20,$37,$6F,$E7,$2B,$E3,$3F,$93,$B9,$AF,$E9,$6D,$AF,$F6,$BB,$68
	.dc.b	$F1,$9E,$FE,$77,$9F,$E1,$FD,$1D,$36,$7F,$C9,$B9,$FF,$AB,$9B,$3A
	.dc.b	$69,$A9,$FE,$7B,$F6,$2F,$9B,$A7,$9B,$6B,$5F,$C2,$27,$4D,$9D,$34
	.dc.b	$9A,$7F,$98,$7E,$B7,$EE,$A1,$5C,$9D,$7C,$67,$D1,$F0,$65,$FA,$BC
	.dc.b	$5B,$5F,$6A,$6F,$07,$F6,$4D,$95,$8B,$F9,$7E,$97,$CE,$FB,$1F,$9E
	.dc.b	$D8,$FA,$ED,$C7,$EB,$B7,$04,$F9,$E7,$CF,$EC,$FC,$63,$F0,$DE,$0D
	.dc.b	$9F,$89,$16,$7F,$1F,$36,$B9,$C7,$C1,$D7,$FB,$0D,$FF,$A0,$BE,$C5
	.dc.b	$96,$E3,$FC,$1B,$18,$7F,$1B,$F6,$3F,$88,$F3,$BF,$12,$D8,$C9,$F9
	.dc.b	$28,$CB,$4F,$C1,$F9,$FF,$9D,$3F,$3D,$EB,$D8,$C9,$32,$FD,$3C,$65
	.dc.b	$FF,$5E,$FD,$8F,$E9,$1B,$DB,$F9,$AF,$03,$FC,$0B,$39,$DB,$DF,$48
	.dc.b	$C6,$1F,$2D,$B1,$F9,$EA,$E3,$72,$BF,$D4,$BC,$DF,$DE,$B0,$E7,$87
	.dc.b	$05,$19,$9F,$F1,$9F,$27,$FC,$00,$7F,$43,$F2,$3F,$E1,$4F,$55,$ED
	.dc.b	$FE,$F6,$24,$E8,$FF,$59,$CE,$9C,$69,$FB,$AF,$A8,$FD,$B9,$FB,$8F
	.dc.b	$8F,$B3,$FC,$02,$2C,$FF,$00,$9B,$9F,$C2,$FB,$0F,$E6,$6D,$F3,$7F
	.dc.b	$C5,$FF,$11,$FE,$2B,$F1,$BF,$C5,$B5,$AC,$BF,$20,$DC,$7F,$50,$BE
	.dc.b	$C7,$78,$70,$78,$F8,$B2,$FF,$BE,$D7,$F4,$0B,$FF,$7A,$BF,$C3,$53
	.dc.b	$49,$F1,$5C,$63,$7D,$1E,$43,$CF,$C3,$FD,$5A,$38,$C6,$6F,$BE,$F6
	.dc.b	$7B,$BF,$E4,$8F,$5B,$FC,$9F,$3F,$AB,$6C,$FF,$03,$7F,$A1,$C7,$F8
	.dc.b	$51,$CC,$7F,$73,$F8,$6F,$D4,$FA,$CF,$A5,$E5,$E3,$E1,$E9,$36,$BA
	.dc.b	$4D,$7F,$BE,$36,$FF,$88,$5F,$39,$F1,$56,$6A,$61,$FE,$9A,$C3,$BB
	.dc.b	$6B,$6D,$7A,$B7,$0F,$EC,$38,$75,$9F,$F0,$67,$8F,$E1,$F7,$BC,$4C
	.dc.b	$BE,$5F,$63,$FE,$0D,$F4,$B9,$4F,$A6,$31,$7F,$EB,$5C,$BF,$8A,$46
	.dc.b	$31,$EC,$D8,$F9,$D5,$C7,$DB,$FE,$DF,$7C,$4F,$E6,$55,$B6,$C6,$AF
	.dc.b	$D1,$9F,$AB,$0E,$F3,$96,$E3,$E8,$8C,$5F,$DB,$C7,$C1,$39,$24,$78
	.dc.b	$4F,$87,$96,$31,$F4,$DF,$29,$65,$E5,$76,$71,$70,$48,$E1,$97,$9E
	.dc.b	$7D,$78,$75,$CB,$CF,$1E,$19,$D3,$DA,$BC,$17,$CE,$FE,$8E,$3C,$6B
	.dc.b	$99,$E1,$AE,$F9,$7B,$A3,$15,$B7,$CF,$97,$B2,$31,$87,$AF,$1F,$0B
	.dc.b	$DE,$B8,$9E,$5B,$E9,$7C,$63,$2F,$D4,$8C,$E5,$EC,$8D,$19,$CB,$BE
	.dc.b	$5E,$C3,$CE,$E3,$2F,$36,$5E,$29,$D4,$63,$0E,$B8,$FC,$0C,$3A,$E3
	.dc.b	$F8,$CC,$79,$4F,$92,$30,$D3,$83,$E5,$33,$3C,$A3,$24,$3E,$3C,$6D
	.dc.b	$C6,$5F,$1E,$FC,$71,$F8,$4F,$B1,$FE,$76,$FF,$75,$3F,$36,$E1,$C4
	.dc.b	$FE,$A1,$B1,$97,$F5,$2F,$2B,$FD,$4B,$6F,$3C,$3F,$B4,$F0,$FF,$C0
	.dc.b	$55,$A8,$FD,$BD,$6E,$FC,$31,$B7,$2C,$7F,$31,$1B,$7F,$3B,$5D,$2F
	.dc.b	$F7,$6D,$8E,$67,$0A,$E8,$FE,$86,$34,$FC,$38,$ED,$7F,$C0,$A7,$3F
	.dc.b	$63,$FC,$46,$3F,$89,$2F,$9A,$C9,$AC,$26,$1F,$C6,$E7,$84,$7F,$88
	.dc.b	$5B,$1F,$E3,$95,$F1,$9D,$8E,$31,$FB,$D8,$DA,$76,$F5,$8D,$DF,$92
	.dc.b	$B9,$4F,$87,$CD,$FF,$18,$77,$7F,$82,$B0,$FF,$34,$4F,$FB,$61,$AA
	.dc.b	$35,$3C,$37,$F4,$36,$7D,$1E,$A5,$7C,$8D,$BF,$F7,$54,$75,$F0,$38
	.dc.b	$78,$0F,$A9,$CC,$F7,$C7,$FC,$B5,$39,$C7,$95,$CA,$E4,$43,$81,$FE
	.dc.b	$61,$FB,$1F,$CC,$67,$FE,$26,$7F,$DA,$11,$FE,$F4,$7F,$9B,$BC,$03
	.dc.b	$88,$E1,$CF,$EB,$8E,$33,$B7,$58,$FF,$96,$5B,$1E,$19,$F8,$5F,$40
	.dc.b	$FB,$1A,$49,$FF,$3B,$61,$FE,$3A,$FB,$1F,$F1,$DB,$FE,$3C,$AE,$13
	.dc.b	$AC,$7F,$CC,$D1,$88,$F4,$36,$79,$4F,$83,$E3,$A5,$71,$3F,$D0,$94
	.dc.b	$8F,$2F,$CB,$6A,$79,$C5,$BF,$14,$FD,$BF,$B1,$7D,$2F,$B3,$E5,$66
	.dc.b	$1D,$44,$E9,$C3,$5F,$3C,$70,$7E,$84,$66,$FF,$CB,$6D,$CB,$FD,$08
	.dc.b	$D9,$FF,$4E,$35,$64,$F9,$0F,$C8,$7D,$8B,$C5,$A4,$AF,$34,$38,$B1
	.dc.b	$1F,$81,$8E,$0E,$A7,$9C,$F4,$F2,$7A,$F4,$4A,$E7,$19,$7A,$C6,$5E
	.dc.b	$91,$FA,$3F,$B1,$7D,$8F,$9B,$AE,$57,$C9,$9C,$B6,$72,$F9,$67,$94
	.dc.b	$48,$D1,$0F,$57,$D8,$FE,$C7,$AD,$EF,$DF,$7E,$57,$D7,$9C,$B6,$72
	.dc.b	$FE,$8E,$3F,$D7,$EE,$4E,$1F,$57,$D4,$8F,$3F,$AF,$AF,$18,$BC,$BF
	.dc.b	$E9,$6A,$57,$F0,$5F,$2B,$FC,$71,$39,$E6,$F4,$19,$D6,$63,$E7,$8C
	.dc.b	$BF,$8F,$AE,$31,$FE,$C9,$6C,$7F,$B8,$5A,$9B,$F6,$0F,$B1,$FE,$2C
	.dc.b	$FE,$8E,$7F,$34,$BA,$F2,$6B,$B2,$4E,$A3,$2F,$08,$E7,$E7,$48,$E6
	.dc.b	$F6,$EC,$C4,$7F,$0F,$B4,$E3,$4F,$F5,$DC,$2B,$E3,$7B,$9F,$CA,$F6
	.dc.b	$FC,$C6,$1F,$AF,$6B,$FC,$67,$44,$F0,$8D,$74,$71,$30,$EE,$2E,$3F
	.dc.b	$EC,$AA,$E3,$3D,$67,$F9,$D7,$D7,$BE,$C7,$F9,$E7,$D8,$FE,$12,$BF
	.dc.b	$94,$37,$0F,$F6,$9D,$6E,$7F,$82,$2F,$33,$8F,$2C,$9C,$C7,$F4,$7F
	.dc.b	$B1,$7C,$CF,$F4,$9F,$B1,$FE,$39,$CC,$B9,$18,$71,$5C,$1F,$E2,$4B
	.dc.b	$E2,$B9,$78,$D2,$73,$1F,$D5,$7E,$C7,$F9,$DE,$BF,$EE,$AC,$3F,$85
	.dc.b	$D7,$F1,$B6,$BF,$D1,$67,$78,$F8,$E1,$0F,$99,$F4,$7A,$F2,$7F,$92
	.dc.b	$4E,$5F,$E4,$DE,$F3,$EC,$7F,$B3,$7D,$8F,$DA,$CF,$FD,$F6,$DC,$BB
	.dc.b	$56,$EF,$8F,$C7,$F0,$F2,$24,$60,$FE,$65,$F6,$2A,$EB,$EC,$49,$CC
	.dc.b	$7F,$75,$97,$C0,$7F,$77,$F6,$5F,$0B,$4D,$46,$FE,$5F,$5F,$0B,$31
	.dc.b	$EF,$1F,$DF,$FE,$C5,$F0,$3C,$FC,$19,$5F,$D1,$67,$2F,$F4,$7F,$B1
	.dc.b	$7C,$A7,$F8,$1F,$EB,$BF,$9E,$1F,$CF,$63,$33,$FD,$37,$E7,$9F,$63
	.dc.b	$FE,$0D,$97,$C8,$D3,$FC,$21,$3A,$AE,$1F,$15,$8B,$91,$83,$A7,$93
	.dc.b	$E1,$F3,$E4,$60,$FE,$B9,$F6,$2F,$8A,$EB,$52,$73,$1F,$E1,$F9,$D4
	.dc.b	$60,$87,$F8,$37,$47,$D8,$FF,$89,$7E,$C1,$D9,$F0,$70,$97,$FD,$AE
	.dc.b	$72,$D9,$CB,$CE,$79,$44,$8D,$10,$F0,$7D,$8F,$7D,$FE,$77,$A3,$27
	.dc.b	$B4,$65,$FE,$F5,$97,$F1,$16,$CE,$5F,$F1,$37,$D8,$FF,$93,$9F,$F1
	.dc.b	$7F,$84,$FA,$DF,$E5,$11,$FE,$0F,$8C,$70,$7F,$90,$EB,$84,$3C,$2F
	.dc.b	$D5,$46,$1B,$DD,$B3,$87,$6F,$F9,$2F,$27,$C8,$5F,$1E,$67,$FE,$11
	.dc.b	$9C,$B6,$78,$11,$FF,$0B,$7D,$8F,$E8,$33,$FE,$19,$F6,$3E,$F9,$4E
	.dc.b	$54,$66,$FA,$C6,$1F,$E0,$46,$23,$94,$F1,$6E,$3F,$E6,$9A,$E0,$7F
	.dc.b	$1B,$AE,$07,$2C,$3F,$86,$57,$0C,$3C,$B3,$FC,$3F,$0E,$8E,$EF,$27
	.dc.b	$3B,$E3,$1B,$C3,$C1,$87,$3A,$F0,$9F,$D3,$6B,$6F,$13,$99,$F0,$A7
	.dc.b	$8D,$73,$C3,$2F,$A2,$78,$1D,$30,$F2,$BF,$CD,$DF,$94,$FF,$B2,$5A
	.dc.b	$F3,$3F,$D9,$87,$79,$AF,$D2,$F3,$DF,$1B,$8A,$FE,$8A,$D8,$CC,$F2
	.dc.b	$9C,$BC,$87,$C7,$D6,$99,$70,$47,$E2,$DF,$FA,$C3,$6F,$87,$1F,$D4
	.dc.b	$67,$E2,$CE,$63,$E5,$99,$38,$ED,$F9,$D7,$27,$83,$5F,$C7,$4F,$17
	.dc.b	$F2,$11,$C5,$CC,$F1,$73,$1C,$73,$FF,$5B,$CE,$5F,$93,$7E,$48,$C4
	.dc.b	$FF,$BD,$E3,$18,$7A,$A3,$91,$FE,$90,$9D,$BF,$EF,$03,$D3,$1C,$1F
	.dc.b	$F5,$26,$1E,$F3,$D8,$EF,$0F,$3C,$FF,$A0,$67,$4F,$98,$FF,$42,$4F
	.dc.b	$B7,$0C,$3F,$EA,$FF,$B0,$FD,$DC,$FF,$A2,$5B,$19,$3E,$55,$78,$23
	.dc.b	$2D,$DB,$35,$FD,$CE,$72,$DA,$C9,$FE,$09,$8C,$B7,$9F,$2A,$3F,$9F
	.dc.b	$4F,$F1,$9F,$B1,$FF,$4F,$BE,$5A,$DC,$EA,$BF,$90,$D3,$E8,$73,$5E
	.dc.b	$0C,$53,$9C,$3D,$DE,$87,$23,$5C,$DF,$D1,$23,$24,$C7,$E3,$4E,$BB
	.dc.b	$38,$4E,$70,$F8,$B7,$FE,$BB,$8C,$1F,$EC,$C7,$17,$FE,$DD,$FB,$17
	.dc.b	$93,$F0,$3E,$53,$FB,$4D,$FF,$AE,$67,$2F,$FA,$EE,$33,$5F,$EE,$56
	.dc.b	$E1,$F8,$26,$CF,$F8,$8E,$B8,$D6,$FB,$3E,$6F,$D8,$3E,$2A,$9E,$18
	.dc.b	$B7,$FE,$C9,$97,$CA,$70,$3E,$AF,$D8,$FF,$A6,$A7,$51,$F6,$3E,$C7
	.dc.b	$FD,$81,$1F,$ED,$16,$E3,$FE,$FB,$BE,$73,$9F,$90,$E1,$F3,$92,$38
	.dc.b	$72,$5B,$6F,$F9,$EF,$B1,$FF,$2C,$57,$FB,$75,$B3,$FD,$83,$CB,$FF
	.dc.b	$0F,$CE,$7D,$FF,$C5,$BF,$E5,$A3,$FD,$C8,$59,$FE,$CF,$F6,$2F,$AA
	.dc.b	$DF,$85,$F9,$2F,$B1,$FF,$2F,$5F,$FB,$AD,$B7,$C3,$D2,$D3,$BF,$9A
	.dc.b	$95,$FD,$BF,$E4,$7F,$D3,$CF,$18,$E9,$DB,$23,$5F,$23,$6F,$DF,$1F
	.dc.b	$E5,$BF,$19,$EA,$7E,$91,$B1,$FD,$EF,$EC,$7F,$7B,$5F,$C8,$EB,$5E
	.dc.b	$8C,$AF,$F0,$07,$D8,$BE,$AF,$17,$D8,$FF,$3D,$7D,$8F,$F9,$56,$BF
	.dc.b	$CF,$9C,$FF,$C3,$23,$13,$FE,$0C,$FB,$1F,$F2,$AD,$68,$FF,$38,$4E
	.dc.b	$AB,$87,$C7,$FE,$EA,$3F,$D4,$33,$FE,$11,$A7,$BF,$9A,$D3,$27,$FC
	.dc.b	$2D,$F6,$3F,$95,$7B,$63,$E6,$9F,$E0,$BE,$A7,$DE,$8F,$50,$71,$97
	.dc.b	$F1,$6C,$31,$7C,$AB,$5F,$01,$FC,$3B,$EC,$7F,$01,$39,$6C,$FF,$8A
	.dc.b	$AB,$F6,$F7,$FC,$BA,$72,$FF,$30,$FB,$1F,$F4,$E5,$F5,$6D,$FF,$8C
	.dc.b	$BE,$A5,$F5,$9F,$C7,$27,$37,$C7,$E3,$B0,$7F,$15,$AD,$E1,$CE,$32
	.dc.b	$FF,$8E,$63,$F7,$73,$C6,$31,$E7,$7F,$1D,$FB,$1F,$91,$97,$F9,$09
	.dc.b	$B5,$FC,$E1,$7C,$5E,$FC,$E9,$87,$F9,$16,$76,$FE,$09,$DB,$98,$F3
	.dc.b	$77,$F9,$FA,$2C,$F9,$5B,$8F,$C1,$6E,$3F,$D2,$33,$FF,$36,$D7,$D3
	.dc.b	$FB,$1F,$F7,$7B,$7E,$67,$D5,$B7,$F4,$29,$CE,$3F,$79,$B8,$7F,$60
	.dc.b	$6D,$F9,$EB,$85,$C8,$FF,$30,$7D,$6F,$E1,$1F,$F6,$35,$70,$C3,$F8
	.dc.b	$6C,$E5,$FE,$1F,$8F,$F1,$D7,$ED,$77,$FD,$5F,$0B,$EB,$33,$0F,$A8
	.dc.b	$E5,$A7,$CB,$F3,$FE,$A7,$D8,$FD,$56,$C6,$4F,$A1,$19,$69,$F1,$7D
	.dc.b	$8F,$84,$7E,$17,$BE,$C6,$49,$8F,$E3,$E3,$2F,$C7,$9F,$EC,$1F,$62
	.dc.b	$BC,$F0,$7F,$7A,$8C,$61,$F8,$1F,$B1,$FC,$57,$9B,$F8,$B6,$C6,$4F
	.dc.b	$C3,$46,$5A,$7E,$57,$CD,$FC,$D1,$F9,$BF,$56,$C6,$49,$9F,$E9,$63
	.dc.b	$2F,$EE,$BE,$C7,$F4,$2D,$ED,$FC,$E7,$7F,$F8,$12,$73,$A7,$F8,$81
	.dc.b	$B8,$7F,$C0,$16,$DF,$F4,$8E,$BF,$E2,$2F,$39,$FE,$05,$3A,$A7,$2F
	.dc.b	$F7,$FA,$D1,$FE,$AA,$71,$1F,$99,$F8,$FF,$DB,$1F,$E0,$BF,$4D,$EC
	.dc.b	$E3,$1F,$DB,$4D,$C3,$FD,$67,$5A,$9C,$9F,$AF,$FA,$6F,$DC,$1F,$B9
	.dc.b	$F8,$CB,$19,$25,$7E,$EE,$32,$FF,$20,$FB,$0F,$E6,$12,$89,$18,$7F
	.dc.b	$CE,$FF,$01,$FC,$22,$75,$09,$D3,$7E,$4F,$FC,$63,$0F,$B1,$B1,$87
	.dc.b	$3E,$24,$BF,$E5,$95,$FD,$12,$FC,$93,$96,$CE,$BE,$27,$1B,$FF,$96
	.dc.b	$2D,$CF,$E0,$3F,$C8,$BF,$13,$F3,$3E,$23,$18,$76,$9F,$2E,$F8,$CB
	.dc.b	$E1,$3F,$E4,$8F,$5F,$FC,$97,$06,$9F,$D0,$9F,$E8,$EE,$3E,$1B,$FD
	.dc.b	$09,$EB,$BE,$93,$E2,$CF,$F1,$08,$C5,$7C,$D8,$FF,$31,$C6,$23,$B7
	.dc.b	$91,$CA,$B8,$B6,$38,$35,$F2,$5D,$BE,$D5,$C3,$2F,$F4,$E3,$6B,$FD
	.dc.b	$4E,$F9,$2B,$85,$F1,$F4,$25,$F0,$BF,$37,$BF,$87,$89,$E9,$9F,$17
	.dc.b	$D8,$BE,$57,$FB,$44,$E8,$F9,$13,$A8,$CE,$9E,$FA,$E1,$EA,$77,$B4
	.dc.b	$61,$E5,$AF,$F7,$5A,$E1,$7C,$B5,$FC,$8D,$EE,$73,$87,$3B,$E1,$5A
	.dc.b	$BE,$18,$7E,$77,$0E,$13,$9B,$E1,$8F,$9D,$F2,$4E,$AB,$85,$F2,$7D
	.dc.b	$B9,$EA,$31,$3B,$9F,$5C,$6F,$1F,$1C,$72,$BE,$04,$7D,$F8,$78,$5F
	.dc.b	$95,$7C,$CE,$95,$CA,$3A,$D7,$1C,$BE,$0C,$72,$C7,$19,$7B,$63,$9E
	.dc.b	$5E,$3C,$BB,$5F,$02,$57,$E6,$23,$2F,$18,$CD,$7E,$06,$FA,$BE,$B3
	.dc.b	$AE,$1F,$2A,$33,$1E,$63,$27,$94,$B3,$D1,$D3,$FA,$3C,$7D,$D9,$7C
	.dc.b	$1C,$BE,$69,$E9,$BE,$B1,$9C,$3C,$18,$74,$C7,$D0,$47,$D0,$64,$F4
	.dc.b	$63,$E1,$C7,$BE,$3D,$AF,$D3,$18,$7B,$C7,$86,$B9,$1E,$8B,$EC,$FE
	.dc.b	$10,$CE,$0F,$B1,$FE,$7B,$1F,$CA,$0F,$C9,$38,$71,$1F,$A1,$6C,$65
	.dc.b	$FC,$FB,$CB,$1F,$D5,$36,$B3,$DB,$F9,$8F,$13,$FD,$01,$18,$BF,$C5
	.dc.b	$D6,$DF,$0D,$6D,$FC,$FC,$65,$FD,$9B,$70,$F5,$4E,$E7,$9B,$C0,$8F
	.dc.b	$E6,$2F,$C2,$FE,$BD,$AF,$48,$DE,$1F,$C0,$E7,$34,$6D,$F8,$4F,$13
	.dc.b	$CB,$EC,$7F,$94,$57,$F2,$95,$F3,$39,$35,$94,$CB,$F8,$74,$F0,$8F
	.dc.b	$EB,$2D,$8F,$EC,$0B,$97,$B1,$C6,$3F,$8E,$46,$D3,$B7,$AC,$6E,$FC
	.dc.b	$95,$CA,$7C,$3E,$67,$F6,$8E,$97,$0F,$8B,$D4,$C3,$6B,$FD,$1D,$97
	.dc.b	$FB,$A1,$B7,$A7,$CF,$55,$F1,$B6,$78,$47,$FA,$71,$78,$1C,$7A,$9F
	.dc.b	$51,$AC,$BC,$76,$9C,$E3,$CE,$E4,$F9,$1A,$FF,$AA,$3E,$C7,$F3,$97
	.dc.b	$CA,$11,$FD,$A5,$FF,$54,$3F,$D1,$9E,$1C,$BE,$98,$D4,$73,$E6,$C4
	.dc.b	$C8,$E3,$1E,$FF,$AE,$7D,$8D,$24,$FF,$AD,$7E,$C7,$C5,$8F,$8E,$B7
	.dc.b	$3F,$EB,$96,$BF,$EC,$06,$C7,$A6,$78,$F0,$7F,$87,$A5,$F5,$7E,$12
	.dc.b	$5F,$3F,$97,$AF,$38,$B8,$78,$67,$EB,$7D,$8B,$E9,$7D,$5F,$2D,$30
	.dc.b	$FF,$66,$4E,$5B,$39,$7C,$D3,$CA,$24,$68,$87,$AF,$E3,$FF,$09,$F6
	.dc.b	$2F,$16,$92,$7F,$DB,$51,$F8,$46,$C6,$5F,$F7,$1C,$71,$77,$C7,$D3
	.dc.b	$C9,$EB,$D3,$27,$9C,$65,$EB,$19,$79,$C7,$E5,$FE,$C5,$F6,$3E,$6E
	.dc.b	$C9,$5F,$EE,$D9,$CB,$67,$2F,$92,$79,$44,$8D,$10,$F4,$FD,$8F,$E9
	.dc.b	$7B,$1E,$ED,$F7,$A5,$7F,$BE,$E7,$2D,$9C,$AE,$4F,$CE,$4A,$BA,$57
	.dc.b	$5F,$AB,$97,$CD,$F0,$7D,$7E,$23,$97,$4F,$F9,$A3,$05,$67,$1E,$B7
	.dc.b	$FB,$BF,$27,$FC,$89,$1E,$0B,$F5,$19,$D6,$63,$E9,$8C,$BF,$E7,$99
	.dc.b	$FD,$95,$72,$9F,$F3,$BB,$67,$FC,$FC,$D7,$FC,$75,$5F,$C2,$FE,$C1
	.dc.b	$F6,$37,$AF,$1D,$72,$FB,$63,$FE,$82,$BF,$D2,$C6,$0F,$F4,$53,$87
	.dc.b	$E0,$98,$9E,$4E,$04,$6F,$2F,$E1,$4B,$CE,$F2,$12,$7B,$1C,$3A,$F0
	.dc.b	$F3,$9E,$0E,$9A,$E6,$3E,$77,$B7,$FC,$A7,$6F,$D1,$4F,$8D,$B8,$7C
	.dc.b	$A9,$DC,$F8,$27,$39,$7F,$A8,$9B,$1F,$32,$76,$FB,$76,$46,$2F,$F9
	.dc.b	$35,$62,$BD,$7F,$5E,$FB,$1F,$E7,$9F,$63,$F8,$F9,$F7,$B5,$F9,$13
	.dc.b	$B7,$8F,$9B,$C3,$8E,$46,$0F,$8F,$F6,$2F,$98,$FE,$91,$F6,$3F,$C8
	.dc.b	$37,$FF,$5D,$B6,$BF,$D8,$55,$B7,$8F,$C5,$F0,$F1,$24,$60,$FA,$7F
	.dc.b	$63,$FC,$EB,$83,$FA,$16,$0F,$1F,$B9,$FD,$5E,$75,$97,$FB,$29,$7C
	.dc.b	$CF,$93,$CC,$98,$F4,$8F,$EB,$DE,$F3,$EC,$7F,$B0,$7D,$8F,$D0,$A3
	.dc.b	$67,$51,$F8,$65,$F2,$1C,$3C,$89,$39,$8F,$ED,$5F,$62,$AE,$BE,$C4
	.dc.b	$AF,$C7,$CE,$5B,$F0,$3F,$DC,$7D,$97,$C2,$D3,$9C,$74,$F9,$6D,$7C
	.dc.b	$A4,$BE,$71,$FD,$E7,$EC,$5F,$03,$CF,$C3,$95,$F9,$E9,$CB,$FA,$0F
	.dc.b	$B1,$7C,$B7,$F8,$07,$EB,$7F,$A4,$4F,$EA,$DB,$D7,$FD,$CE,$B7,$3F
	.dc.b	$A8,$F9,$E7,$D8,$FF,$82,$65,$F2,$54,$FF,$06,$4E,$AB,$87,$C5,$62
	.dc.b	$E4,$60,$E9,$EC,$F8,$FC,$F9,$18,$3F,$6F,$F6,$2F,$8A,$EB,$52,$73
	.dc.b	$1F,$E1,$B9,$D4,$60,$87,$B7,$EA,$FF,$C4,$1F,$63,$F7,$23,$F8,$1E
	.dc.b	$FC,$1F,$85,$7A,$EC,$91,$A8,$F9,$06,$1F,$E2,$55,$CD,$B3,$9E,$0F
	.dc.b	$F0,$E5,$BE,$C7,$C3,$C1,$F4,$7D,$29,$87,$F1,$4C,$7E,$5B,$67,$2F
	.dc.b	$D5,$9E,$2F,$F1,$AF,$B1,$FF,$25,$3F,$C9,$B1,$7D,$73,$83,$FC,$81
	.dc.b	$18,$7F,$C8,$15,$C2,$3F,$C6,$0D,$F1,$D7,$23,$0D,$EE,$E3,$DB,$FE
	.dc.b	$49,$C9,$F2,$17,$C7,$97,$FC,$BA,$72,$D9,$E0,$47,$F9,$9F,$D8,$AF
	.dc.b	$FB,$44,$3E,$C5,$7E,$48,$56,$F4,$F3,$D6,$A7,$D3,$5E,$07,$93,$C4
	.dc.b	$C4,$6E,$39,$DF,$82,$31,$7D,$67,$89,$D3,$0F,$14,$F1,$BE,$67,$96
	.dc.b	$B8,$47,$2B,$E9,$3E,$1A,$E2,$DB,$F4,$BC,$67,$71,$F4,$AF,$9C,$EE
	.dc.b	$3C,$67,$C4,$AE,$6E,$EB,$8B,$EC,$C3,$AB,$E9,$29,$C4,$B7,$D9,$E5
	.dc.b	$1F,$EC,$F6,$96,$3E,$23,$FE,$D3,$6B,$F0,$7D,$0F,$F4,$13,$B6,$BB
	.dc.b	$6B,$BF,$8D,$EC,$FD,$DF,$8F,$AF,$E4,$9A,$9A,$47,$E1,$DF,$FA,$C7
	.dc.b	$7F,$6C,$71,$33,$B7,$FA,$5B,$5F,$C0,$46,$5F,$C1,$57,$18,$F9,$A6
	.dc.b	$75,$FF,$58,$57,$17,$E2,$61,$F2,$AB,$85,$F4,$BF,$31,$E0,$C3,$C6
	.dc.b	$E2,$3B,$E1,$CE,$3F,$DF,$97,$CE,$31,$1F,$0C,$E4,$79,$0E,$58,$79
	.dc.b	$2F,$81,$FE,$1C,$C3,$D4,$7F,$AA,$E7,$D4,$7F,$8F,$6B,$FD,$6F,$1F
	.dc.b	$E9,$E6,$CF,$39,$FF,$16,$1C,$0F,$F2,$D4,$69,$FE,$0B,$F2,$BF,$C0
	.dc.b	$27,$FC,$56,$D8,$FF,$1A,$3F,$C2,$A3,$3B,$71,$94,$EE,$FF,$D8,$0D
	.dc.b	$AF,$DF,$3C,$67,$9D,$71,$8F,$E8,$53,$E0,$E8,$F2,$5B,$EC,$7F,$D4
	.dc.b	$D3,$92,$61,$FC,$96,$1F,$3F,$CE,$F8,$C6,$2B,$D7,$E7,$F7,$3A,$F3
	.dc.b	$3B,$31,$3A,$26,$3F,$CC,$67,$51,$8C,$BD,$B7,$CB,$DF,$E3,$7B,$7E
	.dc.b	$27,$D8,$BC,$AF,$8D,$F2,$DF,$DB,$31,$FF,$2B,$CE,$5F,$F2,$CC,$66
	.dc.b	$FF,$CC,$4D,$C7,$FD,$D4,$D9,$FE,$51,$5C,$6B,$7D,$9F,$4F,$EB,$DF
	.dc.b	$17,$4F,$0C,$5B,$FF,$33,$CB,$E5,$38,$1F,$6F,$EC,$7F,$D4,$B3,$A8
	.dc.b	$FB,$DF,$62,$F9,$1A,$70,$9F,$F3,$8C,$65,$BF,$39,$C0,$E3,$CB,$63
	.dc.b	$FC,$EF,$F6,$2F,$82,$FC,$53,$63,$FD,$11,$C2,$F3,$7F,$C4,$71,$C2
	.dc.b	$BF,$20,$FF,$96,$4F,$F4,$09,$78,$3F,$B4,$7D,$8B,$EC,$77,$E1,$FE
	.dc.b	$5F,$EC,$7F,$25,$93,$0F,$EA,$51,$87,$FD,$51,$7F,$E2,$C7,$FD,$1B
	.dc.b	$19,$C3,$FB,$5A,$F9,$AE,$87,$E6,$3E,$49,$EB,$E2,$F8,$F6,$78,$4D
	.dc.b	$7C,$95,$8F,$F4,$CF,$8C,$F5,$7F,$58,$D8,$F8,$7F,$63,$FC,$09,$FE
	.dc.b	$59,$3A,$F4,$E4,$7C,$8F,$B1,$7D,$56,$37,$FF,$56,$7D,$8F,$F9,$6E
	.dc.b	$BF,$D4,$ED,$8F,$F0,$77,$D8,$FF,$77,$9C,$FC,$1F,$EF,$6B,$FD,$2F
	.dc.b	$1C,$3E,$4B,$F6,$B4,$F8,$1D,$46,$D9,$3F,$6B,$EC,$7F,$26,$F8,$23
	.dc.b	$F8,$47,$AD,$F7,$32,$FF,$60,$37,$A3,$F8,$DE,$08,$DF,$C1,$7F,$0F
	.dc.b	$FB,$1F,$B9,$8F,$FB,$01,$B8,$7F,$89,$ED,$E6,$6F,$B7,$F8,$B7,$D8
	.dc.b	$BE,$3E,$9F,$C3,$BE,$C1,$F5,$B8,$CF,$FD,$97,$5B,$6D,$6E,$75,$97
	.dc.b	$FB,$5E,$B4,$FF,$8D,$EF,$F6,$9B,$7E,$36,$3F,$C6,$33,$A8,$FE,$6F
	.dc.b	$3A,$C3,$FD,$ED,$F5,$BF,$1A,$FF,$C8,$0D,$AF,$E7,$AB,$E3,$3F,$92
	.dc.b	$B9,$AE,$3E,$7E,$0C,$CD,$8F,$D5,$E1,$F9,$BC,$3C,$DD,$3F,$07,$E2
	.dc.b	$3E,$13,$6B,$6D,$9D,$C7,$FB,$BB,$EC,$7F,$DE,$4D,$F9,$BF,$1E,$FF
	.dc.b	$D0,$E7,$35,$FE,$FB,$6E,$BF,$0E,$11,$8B,$F8,$35,$C6,$E4,$7F,$02
	.dc.b	$FA,$DF,$C2,$3F,$E4,$7A,$E1,$6C,$15,$A8,$C6,$7F,$C6,$9F,$A9,$E0
	.dc.b	$FD,$FF,$13,$F0,$0C,$C7,$E2,$B9,$69,$FC,$4B,$83,$F9,$33,$7E,$C7
	.dc.b	$E8,$FA,$3F,$49,$B1,$93,$84,$65,$F9,$1E,$DF,$C5,$3F,$21,$E0,$D8
	.dc.b	$C9,$32,$FC,$54,$65,$F9,$53,$FD,$93,$EC,$58,$67,$6F,$F0,$0C,$63
	.dc.b	$0F,$47,$D8,$FE,$1B,$CF,$FC,$3B,$63,$27,$B6,$32,$D3,$F0,$FE,$7F
	.dc.b	$E9,$4F,$D3,$7A,$F6,$32,$4C,$FF,$41,$19,$7E,$A4,$7F,$9F,$7E,$C7
	.dc.b	$F5,$AF,$FA,$4B,$2F,$C1,$46,$27,$F3,$FD,$FF,$E0,$B8,$C4,$EB,$B7
	.dc.b	$07,$E8,$26,$DF,$33,$B4,$7F,$82,$39,$7F,$BB,$E1,$E2,$3F,$D4,$38
	.dc.b	$74,$AF,$14,$EB,$0F,$D9,$FC,$8F,$F0,$33,$FC,$23,$EA,$BD,$BC,$5F
	.dc.b	$EA,$A6,$E3,$FE,$B5,$AD,$4E,$4F,$C4,$7D,$47,$ED,$8F,$DB,$FC,$7D
	.dc.b	$8C,$92,$BF,$7F,$19,$7D,$03,$EC,$3F,$9B,$65,$FA,$96,$D7,$FA,$F3
	.dc.b	$E2,$3F,$C6,$3F,$01,$FE,$33,$9C,$DF,$FB,$01,$AF,$3F,$B1,$D7,$A1
	.dc.b	$B7,$87,$8D,$31,$F9,$95,$C7,$E2,$F0,$F9,$30,$FE,$20,$F5,$AF,$04
	.dc.b	$EB,$3F,$37,$3F,$8C,$FF,$27,$69,$F2,$FE,$23,$19,$76,$C7,$DD,$5E
	.dc.b	$9D,$F3,$1D,$DF,$F2,$77,$B3,$FE,$4B,$8F,$F6,$DB,$5F,$EE,$5E,$7E
	.dc.b	$BE,$22,$61,$FE,$8F,$FA,$5F,$F6,$75,$FF,$14,$F6,$51,$8A,$FE,$6D
	.dc.b	$18,$C3,$AF,$93,$8A,$E2,$D8,$CE,$1D,$5F,$1D,$E5,$FF,$02,$4D,$C3
	.dc.b	$C1,$7C,$AF,$85,$7F,$82,$6F,$57,$BF,$4A,$56,$B0,$FF,$32,$DF,$FA
	.dc.b	$9F,$0F,$5D,$6A,$30,$7F,$BF,$BE,$C7,$FC,$23,$9F,$19,$F2,$46,$27
	.dc.b	$61,$C6,$7F,$06,$31,$EE,$7F,$71,$77,$8F,$F7,$79,$D5,$70,$C7,$FD
	.dc.b	$F4,$DA,$F1,$5F,$1A,$E9,$7C,$6F,$73,$96,$E3,$F1,$30,$E5,$87,$F8
	.dc.b	$B7,$1F,$4D,$EE,$3A,$D6,$E3,$11,$BC,$B1,$32,$3D,$F7,$D3,$1F,$14
	.dc.b	$72,$33,$7C,$A7,$58,$78,$2F,$96,$5F,$02,$BB,$E1,$CB,$0F,$8F,$5A
	.dc.b	$CB,$D7,$1F,$23,$0E,$4F,$69,$D3,$F9,$A8,$CE,$1D,$67,$33,$EB,$38
	.dc.b	$12,$34,$F5,$CB,$79,$7C,$83,$D2,$E3,$2F,$61,$E8,$C3,$BE,$3E,$6C
	.dc.b	$39,$9F,$5A,$33,$87,$C5,$8C,$CF,$E4,$2F,$A6,$3E,$6C,$BE,$16,$5E
	.dc.b	$1C,$BB,$CF,$E4,$6B,$78,$FA,$70,$E4,$D3,$B3,$E3,$33,$8F,$82,$32
	.dc.b	$43,$E2,$39,$7F,$35,$F6,$3F,$CD,$D1,$0F,$D2,$39,$70,$7D,$82,$9C
	.dc.b	$BB,$2B,$4F,$C4,$C7,$59,$FD,$0B,$6B,$F4,$3E,$17,$79,$CD,$FE,$6E
	.dc.b	$B5,$7E,$28,$FC,$F5,$6E,$FF,$53,$5A,$AF,$05,$F2,$7F,$33,$5B,$79
	.dc.b	$1E,$4B,$EC,$74,$69,$92,$4F,$EE,$27,$94,$7F,$01,$6E,$6B,$F1,$36
	.dc.b	$FA,$57,$88,$EB,$7E,$53,$A3,$FC,$73,$EC,$7F,$85,$47,$F0,$B5,$F3
	.dc.b	$5E,$06,$CE,$F2,$C1,$1C,$1C,$F9,$38,$19,$7F,$8C,$CF,$69,$FB,$CA
	.dc.b	$77,$3C,$8D,$CE,$1E,$3E,$57,$E1,$79,$FF,$CA,$97,$FE,$6D,$9F,$EA
	.dc.b	$8D,$7F,$B9,$BF,$EE,$77,$33,$C7,$6F,$0F,$A7,$67,$F2,$35,$C6,$FA
	.dc.b	$DE,$65,$2E,$FE,$1D,$2F,$A9,$ED,$5F,$E1,$98,$C5,$F1,$BE,$13,$E6
	.dc.b	$BC,$11,$FE,$D9,$F6,$3F,$9E,$C7,$FC,$E0,$FF,$B5,$E3,$3C,$9E,$38
	.dc.b	$FE,$7D,$1F,$E1,$07,$4D,$E6,$D4,$C8,$EB,$1E,$BF,$A0,$7D,$8D,$24
	.dc.b	$FF,$7D,$FB,$1F,$F4,$06,$3F,$E8,$1A,$DC,$FF,$80,$5A,$FF,$88,$1B
	.dc.b	$1F,$E8,$79,$E3,$C1,$FB,$B9,$FF,$5E,$D7,$1F,$6F,$C4,$78,$FE,$5E
	.dc.b	$C6,$48,$75,$7E,$97,$D8,$BE,$97,$D5,$F2,$D2,$BF,$C2,$93,$96,$CE
	.dc.b	$5F,$34,$F2,$89,$1A,$21,$EB,$F8,$FF,$C0,$7D,$8B,$C4,$A4,$9F,$F1
	.dc.b	$14,$7D,$C7,$FC,$82,$ED,$CC,$6D,$CF,$1F,$4F,$1F,$AF,$4C,$9E,$71
	.dc.b	$97,$AC,$65,$EB,$1F,$93,$FB,$17,$D8,$F9,$BB,$25,$7F,$8D,$67,$2D
	.dc.b	$9C,$BE,$69,$E5,$12,$34,$43,$D7,$F6,$3F,$A1,$ED,$7B,$97,$DD,$95
	.dc.b	$FE,$43,$9C,$B6,$72,$FF,$99,$21,$3C,$E7,$A7,$D5,$CA,$E4,$F7,$FA
	.dc.b	$FC,$47,$18,$BD,$5F,$ED,$3C,$15,$FB,$5F,$27,$FC,$45,$39,$EA,$F4
	.dc.b	$19,$D6,$63,$E7,$8C,$BF,$E5,$BA,$E3,$1F,$D4,$5B,$1F,$D5,$1A,$9B
	.dc.b	$F6,$0F,$B1,$FE,$0C,$FE,$7A,$7F,$D0,$6B,$AB,$26,$BA,$E4,$FF,$9A
	.dc.b	$E7,$4D,$9C,$BC,$BC,$DE,$A4,$78,$75,$D7,$47,$38,$F1,$C7,$C5,$F6
	.dc.b	$BF,$8D,$6D,$F9,$7D,$BF,$DF,$8D,$D7,$E0,$4E,$F0,$FF,$3F,$4E,$5F
	.dc.b	$EE,$AD,$7E,$14,$6D,$FF,$50,$EA,$8C,$5F,$67,$F9,$57,$D7,$BE,$C7
	.dc.b	$F9,$67,$D8,$FE,$13,$0F,$F4,$63,$5F,$F5,$5C,$ED,$FF,$48,$2F,$43
	.dc.b	$8F,$24,$9C,$C7,$F3,$7F,$B1,$7C,$C7,$F3,$9F,$B1,$FE,$2B,$BF,$F8
	.dc.b	$35,$B7,$FE,$10,$AD,$BF,$E9,$D5,$F1,$5C,$7C,$49,$39,$8F,$E8,$9F
	.dc.b	$63,$FC,$C3,$81,$8F,$5F,$1F,$C0,$6B,$CA,$DF,$CC,$78,$FC,$C9,$97
	.dc.b	$FA,$AF,$14,$FF,$4F,$F7,$5F,$63,$FD,$47,$EC,$7E,$81,$F4,$63,$37
	.dc.b	$FE,$B3,$5F,$1F,$BF,$26,$4E,$63,$FA,$D7,$D8,$AB,$A7,$AF,$2B,$FD
	.dc.b	$79,$39,$6F,$BF,$FD,$87,$D8,$7C,$25,$39,$C7,$4F,$95,$D7,$C2,$4B
	.dc.b	$E7,$1F,$DA,$7E,$C5,$F0,$3C,$FC,$39,$5F,$EC,$B9,$CB,$FE,$CD,$FB
	.dc.b	$17,$CA,$7F,$71,$FA,$FF,$E7,$93,$FE,$EE,$6F,$5F,$D0,$8F,$F2,$4D
	.dc.b	$70,$7F,$DA,$DF,$3A,$FB,$1F,$EF,$52,$F8,$FA,$7F,$7D,$9D,$57,$0F
	.dc.b	$89,$C5,$C8,$C1,$D3,$DA,$F2,$79,$F2,$30,$7F,$B9,$7E,$C5,$F4,$9D
	.dc.b	$6A,$4E,$63,$FC,$19,$3A,$8C,$10,$F7,$7D,$57,$F8,$47,$EC,$3E,$EC
	.dc.b	$7F,$BC,$77,$E0,$FC,$3B,$D7,$64,$8D,$47,$C9,$30,$B8,$3E,$42,$BE
	.dc.b	$AE,$7C,$5F,$F3,$94,$62,$76,$D8,$FF,$3B,$60,$FB,$1F,$37,$0F,$D7
	.dc.b	$F4,$E6,$3F,$E7,$CC,$78,$CE,$A3,$D5,$3A,$7C,$07,$FA,$0B,$1F,$AC
	.dc.b	$F1,$C7,$FD,$09,$F6,$3F,$E4,$78,$C1,$EB,$FA,$DF,$E5,$2F,$2C,$51
	.dc.b	$87,$DF,$3C,$67,$B8,$71,$1A,$EE,$FF,$20,$46,$2F,$C3,$DB,$FE,$43
	.dc.b	$C9,$F2,$17,$C8,$99,$79,$04,$E9,$C3,$5F,$0B,$72,$FA,$6D,$C3,$FD
	.dc.b	$45,$F6,$2C,$7E,$C5,$BE,$C7,$AB,$FE,$A8,$9D,$61,$FE,$AA,$9C,$CF
	.dc.b	$8B,$2F,$F5,$64,$63,$1F,$13,$AB,$E3,$5D,$67,$C3,$87,$82,$B8,$5F
	.dc.b	$29,$FF,$5B,$1C,$8D,$E1,$E1,$6D,$F3,$3F,$D7,$AE,$DB,$3C,$DF,$55
	.dc.b	$F7,$9E,$37,$E0,$9E,$87,$02,$8C,$3D,$31,$EB,$AE,$53,$B3,$C4,$7C
	.dc.b	$33,$8C,$F0,$AF,$3C,$7F,$B3,$A3,$E1,$B6,$7E,$8B,$70,$FF,$67,$CF
	.dc.b	$F4,$EF,$8E,$BE,$91,$A4,$AF,$F6,$A8,$F8,$FA,$D3,$2E,$08,$FA,$2F
	.dc.b	$FD,$4D,$A7,$C0,$8F,$F7,$3C,$FC,$29,$CC,$7B,$8C,$9C,$77,$FE,$4D
	.dc.b	$A7,$E1,$23,$2F,$E1,$A7,$8B,$98,$F2,$39,$9E,$3A,$7F,$BC,$2F,$C1
	.dc.b	$18,$AF,$F7,$AC,$63,$0F,$34,$72,$AF,$F5,$C5,$6E,$3F,$DE,$07,$96
	.dc.b	$77,$5C,$63,$07,$ED,$B0,$F6,$9F,$C6,$1F,$4D,$F4,$AF,$D8,$4E,$9C
	.dc.b	$4F,$EF,$8D,$9F,$C4,$EF,$FA,$44,$FF,$22,$FB,$1F,$DD,$D7,$EE,$5B
	.dc.b	$5F,$C1,$93,$C2,$2A,$8D,$CF,$3D,$B5,$3C,$27,$35,$C1,$E6,$F7,$9D
	.dc.b	$CF,$3E,$6B,$5F,$D2,$AB,$F8,$FD,$BE,$C7,$FD,$45,$1E,$7A,$D9,$E0
	.dc.b	$9D,$7C,$F6,$27,$8D,$23,$AF,$9F,$27,$AF,$2B,$AF,$19,$FC,$1A,$B8
	.dc.b	$FC,$8A,$D5,$FC,$79,$CB,$FC,$FE,$1F,$01,$8B,$FE,$6C,$FB,$BE,$C5
	.dc.b	$E5,$7C,$0F,$94,$FE,$E9,$87,$F2,$99,$CB,$FC,$AE,$33,$7F,$CE,$1B
	.dc.b	$97,$F5,$A6,$CF,$D4,$AE,$35,$BE,$CF,$9B,$F6,$2F,$8C,$A7,$86,$2D
	.dc.b	$FF,$39,$97,$CA,$F0,$3E,$AF,$D8,$FF,$A8,$A7,$51,$F6,$3E,$C5,$F2
	.dc.b	$54,$E1,$3F,$D0,$E3,$2D,$F9,$CE,$07,$1E,$6B,$1F,$D2,$7E,$C7,$FC
	.dc.b	$BC,$DF,$82,$FB,$9E,$6F,$F8,$AA,$30,$7F,$4F,$3F,$CB,$E5,$F0,$7F
	.dc.b	$05,$F6,$2F,$AA,$FC,$6C,$EB,$8B,$FB,$0F,$D8,$FE,$49,$A7,$0F,$4F
	.dc.b	$F2,$71,$9E,$7F,$F0,$3A,$F8,$0C,$9F,$8F,$F9,$2F,$F5,$26,$1F,$D8
	.dc.b	$67,$4D,$ED,$E6,$6F,$E4,$AC,$7F,$67,$F1,$9E,$AF,$E8,$9B,$1F,$A0
	.dc.b	$FB,$1F,$DE,$6B,$FC,$82,$B5,$E9,$4A,$FD,$2F,$D8,$BE,$A7,$17,$FD
	.dc.b	$FB,$EC,$7F,$CC,$15,$FD,$D5,$B1,$FE,$07,$FB,$1F,$ED,$D3,$9F,$83
	.dc.b	$FD,$9D,$7F,$A6,$E3,$87,$C8,$FE,$B6,$9F,$03,$A8,$DB,$27,$F6,$FF
	.dc.b	$63,$F8,$F7,$C1,$1F,$BE,$F5,$BE,$D6,$5F,$E1,$26,$F3,$7F,$14,$C1
	.dc.b	$1B,$F8,$2F,$E0,$BF,$63,$F6,$F1,$FF,$07,$37,$0F,$F1,$05,$FE,$C2
	.dc.b	$73,$D5,$FC,$33,$EC,$7F,$D3,$71,$FE,$15,$6D,$7F,$89,$FE,$C1,$F5
	.dc.b	$9B,$F8,$F9,$7B,$7F,$86,$4E,$A3,$04,$AE,$5E,$7F,$06,$C7,$F1,$8F
	.dc.b	$B0,$F8,$51,$FE,$35,$6E,$FF,$CC,$97,$C6,$7F,$19,$73,$5F,$E4,$96
	.dc.b	$D7,$F9,$65,$B4,$78,$CF,$7F,$3F,$CF,$E1,$7A,$3A,$AC,$FF,$2A,$CF
	.dc.b	$FD,$4E,$D9,$D3,$4D,$4F,$F3,$4F,$B1,$7C,$D5,$3C,$DB,$5A,$FF,$20
	.dc.b	$4E,$9B,$3A,$69,$34,$FF,$2B,$7D,$6F,$E0,$23,$FA,$5D,$71,$9C,$F2
	.dc.b	$FD,$0F,$06,$57,$A7,$C4,$B5,$F6,$66,$F0,$7F,$5C,$D9,$58,$BF,$95
	.dc.b	$E7,$FC,$DF,$B1,$F9,$CD,$8F,$AC,$DC,$7E,$B3,$70,$FF,$43,$E7,$FD
	.dc.b	$26,$BD,$FE,$CF,$E1,$3C,$1F,$8C,$C8,$C9,$F3,$E3,$18,$7F,$9C,$2B
	.dc.b	$93,$FE,$77,$FB,$16,$5A,$C7,$FB,$AC,$63,$0F,$C5,$7D,$8F,$E1,$7C
	.dc.b	$DF,$C3,$36,$32,$7E,$3E,$32,$D3,$F0,$1E,$77,$E6,$8F,$CD,$FA,$D6
	.dc.b	$32,$4C,$BF,$49,$19,$7F,$D0,$FF,$63,$FA,$16,$F6,$FE,$5F,$C0,$FE
	.dc.b	$F3,$39,$D3,$FD,$5A,$DC,$3F,$C0,$ED,$BF,$CD,$57,$0B,$95,$FD,$FB
	.dc.b	$9B,$FB,$6D,$76,$8C,$7B,$9F,$B2,$F9,$19,$1F,$E0,$4F,$55,$ED,$E2
	.dc.b	$FF,$79,$57,$4F,$F0,$FD,$6A,$FF,$45,$3F,$AF,$FA,$79,$1F,$B7,$F8
	.dc.b	$DB,$87,$ED,$E6,$E1,$FB,$79,$B5,$FE,$22,$FB,$0F,$E5,$73,$FE,$CC
	.dc.b	$6C,$7F,$8A,$3E,$1F,$FC,$4D,$E3,$7F,$87,$A7,$38,$FF,$8B,$1A,$97
	.dc.b	$D8,$EF,$D0,$E0,$F1,$F1,$A6,$3F,$EA,$FA,$FE,$73,$7F,$EB,$C7,$F8
	.dc.b	$3B,$CE,$76,$E7,$E2,$AE,$7E,$7E,$8F,$21,$FE,$38,$D3,$E5,$FC,$3E
	.dc.b	$32,$ED,$3E,$9D,$F1,$8F,$77,$FC,$73,$EC,$FF,$8F,$A3,$FC,$8E,$D7
	.dc.b	$FB,$47,$9F,$AF,$87,$98,$7F,$9E,$BE,$97,$FC,$97,$7F,$C0,$7D,$94
	.dc.b	$62,$95,$BC,$3A,$F9,$58,$7F,$D1,$95,$CA,$31,$7C,$A7,$75,$C5,$FE
	.dc.b	$F5,$7A,$BE,$13,$9C,$3C,$51,$8A,$EA,$7B,$2B,$83,$D0,$C9,$FE,$05
	.dc.b	$F2,$B2,$DB,$E9,$EF,$E3,$0F,$EC,$38,$F9,$6B,$47,$2F,$B0,$7C,$B7
	.dc.b	$F6,$28,$EB,$5C,$33,$FF,$38,$D6,$B5,$F6,$36,$B9,$5F,$1C,$3B,$61
	.dc.b	$C6,$7F,$CF,$75,$B9,$F1,$37,$0E,$B7,$C5,$FA,$B9,$FB,$27,$57,$CA
	.dc.b	$B4,$DC,$7E,$55,$6E,$73,$2A,$E2,$D8,$CE,$3E,$58,$C5,$FB,$27,$8D
	.dc.b	$70,$AD,$3C,$89,$1C,$AF,$C1,$5C,$49,$7F,$2A,$32,$FC,$68,$CE,$3F
	.dc.b	$1A,$72,$FC,$4A,$F0,$9D,$73,$DE,$5F,$1A,$EE,$1F,$A0,$C3,$8C,$E6
	.dc.b	$3B,$1E,$1A,$DB,$EE,$33,$87,$86,$33,$8F,$B3,$0F,$97,$18,$C7,$C7
	.dc.b	$18,$BE,$38,$77,$C3,$D1,$39,$6C,$E8,$FA,$C7,$8C,$C1,$E3,$76,$78
	.dc.b	$DC,$E1,$E3,$78,$46,$AB,$C0,$48,$F8,$58,$7A,$BE,$C7,$F9,$AC,$7F
	.dc.b	$B2,$0F,$C8,$B8,$71,$7F,$99,$6C,$65,$FC,$FB,$CA,$FF,$2E,$DB,$CF
	.dc.b	$6F,$E8,$BC,$4F,$F6,$AD,$6A,$3F,$4B,$5B,$BF,$0C,$6D,$CB,$1F,$D9
	.dc.b	$46,$DF,$CC,$57,$4B,$EF,$1A,$AE,$17,$FA,$AA,$D4,$F6,$34,$48,$D3
	.dc.b	$C2,$B9,$C7,$9E,$3F,$81,$4F,$27,$C2,$6C,$E9,$1C,$9D,$99,$3F,$5D
	.dc.b	$7E,$9C,$15,$CF,$EC,$7F,$88,$5F,$F1,$15,$F3,$59,$35,$8C,$C7,$F9
	.dc.b	$44,$F0,$8F,$F8,$02,$D8,$05,$CB,$DC,$E5,$1F,$C9,$A3,$83,$B8,$D3
	.dc.b	$DA,$77,$7E,$59,$F4,$79,$BD,$A9,$D1,$FD,$05,$F0,$7A,$B9,$8C,$63
	.dc.b	$FE,$6C,$9F,$2C,$62,$B5,$BF,$A9,$B3,$EA,$AE,$0F,$69,$F7,$DE,$A3
	.dc.b	$DD,$56,$9E,$07,$2E,$A7,$D4,$FF,$BA,$AF,$C3,$8F,$F8,$0E,$31,$1D
	.dc.b	$4F,$05,$73,$23,$C0,$C9,$47,$D8,$FE,$AF,$3F,$84,$FF,$B5,$63,$4F
	.dc.b	$F4,$97,$B1,$C3,$93,$D5,$1B,$8C,$B6,$79,$CE,$8D,$F3,$FC,$1F,$A0
	.dc.b	$7D,$8D,$26,$1E,$5F,$B1,$FA,$98,$7D,$5A,$DC,$FA,$5A,$F9,$1B,$1F
	.dc.b	$6E,$78,$E9,$F7,$E5,$F1,$FE,$12,$5F,$3F,$9B,$AF,$38,$B8,$78,$67
	.dc.b	$EC,$7D,$8B,$E9,$3D,$5F,$2D,$30,$F9,$93,$96,$CE,$5F,$34,$F2,$89
	.dc.b	$1A,$21,$EB,$F9,$0F,$C3,$7D,$8B,$C5,$A4,$9F,$B5,$1F,$88,$6C,$65
	.dc.b	$FA,$F1,$C5,$DF,$1F,$4F,$23,$AF,$4C,$9E,$71,$97,$AC,$65,$E7,$1F
	.dc.b	$99,$FB,$17,$D8,$F9,$BB,$A5,$7E,$36,$72,$D9,$CB,$E4,$9E,$51,$23
	.dc.b	$44,$3D,$3F,$63,$FA,$7E,$A7,$C0,$5F,$7E,$57,$E6,$E7,$2D,$9C,$BF
	.dc.b	$98,$8F,$F5,$DB,$93,$87,$D5,$F5,$23,$CF,$EB,$EB,$C6,$2F,$37,$F0
	.dc.b	$7A,$57,$EE,$FC,$9F,$F1,$94,$E7,$AB,$D0,$67,$59,$8F,$9E,$32,$FE
	.dc.b	$CE,$B8,$C7,$F1,$56,$C7,$F2,$46,$BF,$E3,$C9,$FE,$15,$F6,$0F,$B1
	.dc.b	$BD,$78,$EB,$97,$FB,$FA,$FD,$0B,$67,$2F,$EF,$21,$3C,$A3,$87,$9F
	.dc.b	$D8,$8F,$3E,$BA,$F1,$8B,$A7,$F2,$BB,$57,$C5,$F6,$FF,$90,$6D,$F9
	.dc.b	$7C,$3F,$85,$B6,$FF,$99,$6A,$AE,$15,$FC,$5E,$B5,$B7,$F3,$96,$C7
	.dc.b	$C5,$9D,$BF,$C5,$AB,$93,$F3,$63,$C2,$7F,$31,$FA,$F7,$D8,$FF,$33
	.dc.b	$FB,$1F,$C2,$D7,$F1,$E6,$E5,$FD,$16,$B7,$3F,$CA,$5F,$EC,$6F,$F2
	.dc.b	$B9,$71,$F3,$3B,$79,$BF,$CF,$FE,$C1,$F3,$5F,$D0,$7E,$C7,$F8,$BE
	.dc.b	$FF,$D4,$DB,$8F,$F6,$2A,$DB,$FC,$DD,$7C,$5F,$1F,$12,$4E,$63,$FA
	.dc.b	$67,$D8,$FF,$31,$C3,$A7,$8F,$DC,$CF,$95,$AF,$9A,$F1,$79,$93,$1F
	.dc.b	$E8,$98,$27,$FA,$B7,$BC,$FB,$1F,$EB,$1F,$63,$F4,$2C,$D9,$D4,$7F
	.dc.b	$4D,$5F,$21,$C3,$C9,$93,$98,$FE,$C5,$F6,$2A,$EB,$EC,$4A,$FE,$AD
	.dc.b	$39,$6F,$C0,$7F,$69,$F6,$1F,$09,$4E,$71,$D3,$E5,$F5,$F0,$B2,$F9
	.dc.b	$C7,$F7,$1F,$B1,$7C,$0F,$3F,$0E,$57,$F6,$59,$CB,$FD,$9F,$EC,$5F
	.dc.b	$2B,$FD,$E7,$EB,$FF,$9F,$4F,$F6,$46,$F8,$9F,$E0,$BA,$DC,$FF,$6F
	.dc.b	$F9,$F7,$D8,$FF,$80,$65,$F2,$34,$FF,$02,$4E,$AB,$87,$C5,$62,$E4
	.dc.b	$60,$E9,$ED,$78,$FD,$09,$18,$3F,$BF,$7D,$8B,$E2,$BA,$D4,$9C,$C7
	.dc.b	$F8,$4E,$75,$18,$21,$ED,$F8,$FF,$F0,$C7,$D8,$7D,$C8,$FF,$05,$6F
	.dc.b	$C1,$F8,$57,$AE,$C9,$1A,$8F,$90,$61,$FF,$05,$D7,$36,$CE,$78,$3F
	.dc.b	$C6,$96,$FB,$1F,$0F,$07,$D1,$F4,$E6,$1F,$E1,$BC,$7E,$5B,$67,$2F
	.dc.b	$D5,$9E,$2F,$F8,$87,$EC,$7F,$C9,$AF,$F8,$97,$17,$D6,$FF,$22,$8C
	.dc.b	$35,$FF,$10,$CF,$06,$F6,$FF,$8E,$E3,$15,$FE,$4B,$6F,$93,$FE,$4F
	.dc.b	$5D,$BF,$E4,$1C,$5F,$21,$7C,$89,$97,$F8,$D6,$72,$D9,$E0,$47,$FC
	.dc.b	$75,$F6,$3D,$32,$FF,$34,$FD,$8F,$F3,$09,$FF,$35,$CE,$54,$65,$FE
	.dc.b	$6B,$5F,$E6,$38,$C5,$F8,$27,$6F,$F9,$9E,$B9,$4F,$16,$C7,$F9,$AC
	.dc.b	$E1,$1C,$AB,$A4,$71,$C7,$FC,$DF,$7D,$0F,$2B,$8A,$E1,$1E,$A9,$E9
	.dc.b	$5C,$A3,$78,$65,$A7,$33,$D6,$64,$A6,$70,$F3,$BC,$2F,$BB,$C6,$FD
	.dc.b	$27,$9E,$B9,$57,$C9,$9F,$F6,$83,$5D,$1F,$ED,$30,$E5,$FA,$3E,$8B
	.dc.b	$E3,$35,$1A,$8C,$3C,$67,$FA,$2C,$E9,$B1,$C0,$7C,$7D,$69,$97,$04
	.dc.b	$7E,$25,$FF,$AA,$B6,$F8,$91,$FE,$AE,$9F,$8D,$39,$8F,$94,$64,$E3
	.dc.b	$BA,$B8,$C7,$83,$3F,$C5,$57,$17,$F1,$91,$C5,$CD,$71,$8E,$A7,$1C
	.dc.b	$FF,$D5,$D1,$99,$E3,$39,$3F,$D5,$F1,$99,$F9,$D8,$7A,$23,$13,$FE
	.dc.b	$BB,$38,$1F,$EB,$C7,$2F,$FB,$E8,$F5,$E1,$C4,$FF,$51,$61,$F0,$4F
	.dc.b	$F6,$14,$7A,$6F,$CB,$3A,$AF,$55,$7F,$A5,$1B,$1E,$4A,$EA,$7F,$A9
	.dc.b	$BE,$C7,$F7,$B1,$FE,$AF,$6C,$64,$E2,$7F,$B8,$1C,$D7,$FB,$4D,$B3
	.dc.b	$FE,$D7,$6C,$EC,$E2,$1C,$B7,$9F,$31,$FD,$06,$7F,$8F,$7D,$8F,$FA
	.dc.b	$65,$F2,$4E,$8F,$1D,$A5,$F4,$19,$AE,$2F,$4F,$3E,$47,$F2,$6E,$BB
	.dc.b	$97,$C1,$9B,$87,$C8,$E6,$56,$2F,$D9,$5E,$09,$FF,$5A,$C7,$C7,$8C
	.dc.b	$7C,$07,$BB,$EC,$5E,$57,$C0,$F9,$6F,$ED,$37,$FE,$DB,$9C,$BF,$ED
	.dc.b	$D8,$CE,$7F,$ED,$86,$E7,$FE,$FF,$6C,$A6,$57,$27,$87,$5F,$CE,$FB
	.dc.b	$17,$C6,$53,$C5,$16,$FF,$DD,$92,$F9,$5E,$07,$D6,$FB,$1F,$F4,$CC
	.dc.b	$EA,$3E,$CF,$D8,$BE,$4A,$9C,$27,$FD,$ED,$19,$6F,$CE,$70,$38,$F3
	.dc.b	$58,$FF,$7E,$7D,$8B,$E0,$BF,$0A,$D8,$FE,$BD,$E6,$FF,$88,$1D,$CF
	.dc.b	$E2,$9F,$F2,$A9,$F0,$F8,$BF,$71,$C1,$FC,$07,$C1,$FC,$2F,$D8,$BE
	.dc.b	$AB,$84,$7E,$4B,$EC,$7E,$56,$1C,$4F,$99,$68,$CA,$C3,$F5,$78,$7F
	.dc.b	$8B,$2B,$84,$FD,$3A,$DD,$78,$97,$C0,$77,$3F,$2F,$F2,$4F,$5F,$18
	.dc.b	$72,$C3,$C1,$D9,$E5,$37,$F2,$56,$75,$E2,$BD,$5F,$D5,$B6,$3F,$C0
	.dc.b	$9F,$63,$FC,$15,$F5,$C6,$70,$F5,$46,$3D,$2E,$27,$EA,$BE,$C5,$F5
	.dc.b	$58,$9F,$27,$D8,$FF,$97,$DB,$F0,$5F,$B0,$FB,$1F,$F0,$2C,$60,$DF
	.dc.b	$C8,$FE,$FA,$BF,$D4,$91,$FE,$17,$97,$C0,$E6,$B4,$C9,$FF,$0E,$7D
	.dc.b	$8F,$E6,$1F,$0C,$7F,$08,$F5,$BF,$05,$AF,$A8,$FD,$13,$8E,$7F,$E3
	.dc.b	$F8,$23,$87,$C1,$7F,$10,$FB,$1F,$C1,$63,$F3,$DB,$87,$3C,$3F,$D7
	.dc.b	$5E,$66,$FB,$7F,$8B,$FD,$8F,$FA,$6B,$0F,$AC,$DB,$E1,$F6,$0F,$AB
	.dc.b	$DF,$C7,$CB,$DB,$FC,$5A,$75,$18,$25,$72,$F4,$38,$36,$3F,$C7,$DF
	.dc.b	$63,$F0,$E3,$CC,$DD,$3F,$9E,$AF,$8C,$FE,$46,$E6,$BF,$48,$DA,$FC
	.dc.b	$D3,$68,$F1,$9E,$FE,$87,$9F,$E2,$3D,$1D,$76,$7F,$CA,$19,$FF,$A9
	.dc.b	$9B,$3A,$69,$A9,$FE,$7F,$F6,$2F,$9B,$A7,$9B,$6B,$5F,$A1,$9D,$36
	.dc.b	$74,$D2,$69,$F5,$FE,$B7,$F0,$11,$FB,$8A,$E3,$39,$E5,$FA,$1E,$14
	.dc.b	$AF,$4F,$8B,$6B,$EC,$CD,$E0,$FE,$CF,$B2,$B1,$7F,$2B,$D1,$F9,$BF
	.dc.b	$63,$F3,$9B,$1F,$59,$B8,$FD,$66,$E1,$FC,$53,$3F,$EA,$B5,$EF,$F6
	.dc.b	$BF,$09,$E1,$7C,$66,$46,$4F,$9F,$18,$C3,$F4,$02,$B9,$BF,$99,$FB
	.dc.b	$16,$5B,$C7,$FC,$03,$18,$C3,$F1,$7F,$63,$F8,$6F,$43,$F0,$ED,$8C
	.dc.b	$9F,$90,$8C,$B4,$FC,$0F,$A1,$F9,$B3,$F3,$9E,$C5,$8C,$93,$2F,$D2
	.dc.b	$C6,$5F,$4F,$D8,$FE,$89,$BD,$BF,$98,$F0,$FF,$C0,$B3,$9D,$3F,$9D
	.dc.b	$B7,$0F,$E6,$AD,$BF,$CD,$D7,$0B,$95,$FE,$A5,$E6,$FE,$F2,$FF,$2D
	.dc.b	$BE,$18,$27,$38,$7E,$EB,$E4,$7F,$6C,$7F,$83,$3D,$67,$B9,$8C,$7F
	.dc.b	$43,$36,$BF,$84,$56,$A7,$27,$EE,$3E,$A3,$F6,$67,$ED,$3E,$3A,$C6
	.dc.b	$49,$5F,$BC,$8C,$BF,$7B,$EC,$3F,$98,$37,$CE,$FF,$18,$FC,$47,$F8
	.dc.b	$A7,$C6,$FF,$15,$D6,$B6,$FE,$C2,$DC,$7F,$AD,$AF,$B1,$DE,$1C,$1E
	.dc.b	$3E,$2C,$AF,$EA,$55,$FC,$F3,$2F,$14,$E5,$B3,$AF,$8B,$C6,$3E,$0E
	.dc.b	$7F,$19,$E6,$E1,$FE,$A5,$7F,$E1,$7C,$53,$AC,$3B,$EF,$6B,$E0,$3F
	.dc.b	$E4,$6F,$63,$FC,$8A,$4F,$3B,$1F,$17,$2F,$FD,$09,$F4,$BF,$DE,$2B
	.dc.b	$F8,$0F,$AC,$8C,$57,$F9,$EE,$31,$8F,$F8,$A6,$B7,$5C,$7C,$8D,$37
	.dc.b	$0E,$B5,$BC,$32,$F1,$AA,$FF,$7E,$6D,$F4,$8E,$15,$C9,$D3,$63,$FA
	.dc.b	$DE,$1E,$49,$FF,$03,$F9,$1C,$BD,$FE,$6C,$BE,$5F,$61,$FE,$09,$D3
	.dc.b	$94,$F8,$A3,$17,$FE,$B0,$D3,$FC,$37,$18,$C7,$B3,$63,$E4,$57,$1F
	.dc.b	$73,$FB,$45,$F1,$3F,$C9,$D5,$B6,$C6,$AF,$D1,$9F,$AB,$0E,$F3,$96
	.dc.b	$E3,$E8,$8C,$5F,$D1,$C7,$C1,$39,$24,$78,$4F,$87,$96,$31,$F4,$DF
	.dc.b	$29,$65,$E5,$76,$71,$70,$48,$E1,$97,$9E,$7D,$78,$75,$CB,$CF,$1E
	.dc.b	$19,$D3,$DA,$BC,$17,$CE,$FE,$8E,$3C,$6B,$99,$E1,$AE,$F9,$7B,$A3
	.dc.b	$15,$B7,$CF,$97,$B2,$31,$87,$AF,$1F,$0B,$DE,$B8,$9E,$5B,$E9,$7C
	.dc.b	$63,$2F,$D4,$8C,$E5,$EC,$8D,$19,$CB,$BE,$5E,$C3,$CE,$E3,$2F,$36
	.dc.b	$5E,$29,$D4,$63,$0E,$B8,$FC,$0C,$3A,$E3,$F8,$CC,$7D,$7F,$63,$FC
	.dc.b	$AA,$BF,$DC,$07,$D8,$70,$FF,$B8,$F3,$FC,$CB,$63,$2F,$E6,$9E,$57
	.dc.b	$F9,$A6,$DE,$78,$BF,$4B,$E1,$7F,$98,$23,$17,$F8,$3A,$DB,$F9,$C6
	.dc.b	$BF,$8A,$9E,$0E,$6B,$F5,$F3,$AC,$0D,$AD,$D7,$AE,$B6,$47,$F2,$51
	.dc.b	$AA,$E3,$39,$C3,$D4,$FA,$2B,$18,$7A,$27,$31,$C8,$C9,$E5,$3D,$93
	.dc.b	$C5,$D1,$82,$47,$F0,$0B,$F1,$1E,$3C,$3D,$87,$AE,$7F,$8B,$7D,$8F
	.dc.b	$F1,$3A,$FE,$28,$BE,$6B,$DE,$D9,$DE,$38,$23,$83,$9F,$C6,$F0,$32
	.dc.b	$FE,$F6,$7B,$4F,$FA,$C1,$4E,$E7,$91,$B9,$C3,$C7,$CC,$FF,$5B,$F4
	.dc.b	$7F,$B0,$2F,$FC,$C5,$3F,$EA,$B6,$BF,$E2,$87,$FD,$BC,$E6,$78,$ED
	.dc.b	$E1,$F5,$2C,$FF,$BB,$6B,$8D,$F5,$BC,$CA,$5E,$07,$0F,$01,$F5,$3D
	.dc.b	$AB,$FD,$2D,$18,$BE,$37,$C2,$7C,$D7,$82,$3F,$EC,$3F,$B1,$FF,$69
	.dc.b	$BF,$E4,$07,$FD,$96,$E7,$97,$C7,$1F,$CE,$23,$5C,$F8,$33,$32,$3A
	.dc.b	$47,$AB,$E8,$1F,$63,$49,$3F,$ED,$3F,$B1,$FF,$75,$ED,$FE,$EC,$AD
	.dc.b	$CF,$FB,$5D,$AF,$F9,$51,$B1,$FE,$F2,$9E,$3C,$1F,$EE,$C9,$7B,$9E
	.dc.b	$12,$5F,$3F,$9B,$AF,$38,$B8,$78,$67,$E8,$FD,$8B,$E9,$3D,$5F,$2D
	.dc.b	$30,$FF,$75,$4E,$5B,$39,$7C,$D3,$CA,$24,$68,$87,$AF,$E3,$FE,$FF
	.dc.b	$D8,$BC,$5A,$49,$FF,$7C,$47,$DB,$7F,$CE,$AE,$DC,$C6,$DC,$F1,$F4
	.dc.b	$F2,$7A,$F4,$C9,$E7,$19,$7A,$C6,$5E,$B1,$F9,$2F,$B1,$7D,$8F,$9B
	.dc.b	$B2,$50,$FF,$43,$CE,$9C,$70,$38,$F5,$09,$D3,$EA,$AE,$6F,$2B,$F0
	.dc.b	$11,$E7,$5C,$AB,$8D,$F5,$9F,$16,$8F,$B1,$FD,$4F,$6B,$DF,$BE,$F4
	.dc.b	$BE,$67,$A6,$74,$E0,$FD,$6C,$ED,$FF,$51,$C7,$FA,$DD,$D3,$C1,$CF
	.dc.b	$D5,$F8,$08,$EB,$EB,$EB,$A3,$8C,$79,$A3,$F7,$FE,$57,$F8,$D7,$3F
	.dc.b	$65,$7E,$FB,$5F,$59,$AD,$E4,$78,$A3,$2F,$A6,$B9,$47,$D0,$6C,$7D
	.dc.b	$76,$A6,$FD,$83,$EC,$7F,$88,$BF,$AA,$9F,$88,$BA,$B2,$6B,$AE,$4E
	.dc.b	$A3,$2F,$08,$E7,$E7,$48,$E6,$F6,$EB,$C4,$7F,$B4,$2F,$FD,$F4,$E8
	.dc.b	$F0,$C7,$C6,$F6,$FF,$92,$ED,$FA,$3D,$BE,$3B,$75,$F8,$13,$BB,$FB
	.dc.b	$93,$97,$F3,$0D,$7E,$14,$6D,$FF,$70,$EA,$8C,$5F,$67,$F9,$AF,$D8
	.dc.b	$3E,$C7,$F9,$BF,$D8,$FE,$17,$0F,$F7,$4B,$6F,$ED,$D6,$E7,$F1,$CB
	.dc.b	$CE,$E3,$C7,$27,$31,$FD,$0B,$EC,$1F,$33,$FD,$13,$EC,$7F,$90,$EF
	.dc.b	$FA,$C6,$DF,$EC,$EB,$6F,$E6,$D7,$C5,$71,$F1,$24,$E6,$3F,$A7,$FD
	.dc.b	$8F,$F4,$08,$FF,$80,$36,$F1,$FC,$06,$BC,$AD,$FC,$CF,$8F,$CC,$97
	.dc.b	$FA,$9C,$13,$FD,$6B,$DE,$7D,$8F,$F5,$CF,$B1,$FA,$07,$D1,$8C,$DF
	.dc.b	$EC,$57,$C7,$EF,$CA,$93,$98,$FE,$CB,$F6,$2A,$E9,$EB,$CA,$FD,$D4
	.dc.b	$E5,$BF,$03,$FD,$B3,$D8,$7C,$2D,$39,$C7,$4F,$96,$D7,$C2,$CB,$E7
	.dc.b	$1F,$DD,$3E,$C5,$F0,$3C,$FC,$39,$5F,$C2,$67,$2F,$F0,$BF,$B1,$7C
	.dc.b	$A7,$F7,$CF,$AF,$FE,$91,$2F,$0F,$F7,$F1,$87,$F8,$0D,$70,$9F,$E2
	.dc.b	$9F,$3E,$FB,$1F,$F0,$2C,$BE,$42,$9F,$E0,$89,$D5,$70,$F8,$AC,$5C
	.dc.b	$8C,$1D,$3D,$BF,$27,$A3,$23,$07,$F2,$3F,$B1,$7C,$57,$5A,$93,$98
	.dc.b	$FF,$0B,$CE,$A3,$04,$3D,$DE,$3F,$F4,$2D,$9F,$55,$FE,$1E,$FB,$0F
	.dc.b	$BF,$5F,$CC,$67,$93,$C0,$FE,$9C,$E3,$83,$F9,$2C,$F0,$9D,$07,$11
	.dc.b	$F3,$31,$FA,$07,$7E,$D9,$5F,$CF,$EB,$78,$FF,$56,$6C,$78,$3E,$C7
	.dc.b	$CD,$C3,$F6,$7D,$39,$3C,$63,$2F,$A6,$32,$F9,$2B,$EB,$3F,$D3,$7E
	.dc.b	$C7,$FC,$A2,$FF,$52,$CD,$F5,$BF,$CB,$63,$0D,$7D,$B3,$C1,$BD,$DF
	.dc.b	$E4,$68,$C5,$7F,$48,$6E,$7F,$DC,$17,$6F,$F9,$33,$27,$C8,$5F,$22
	.dc.b	$65,$FD,$86,$72,$D9,$E0,$47,$FB,$37,$D8,$F4,$CB,$FC,$05,$F6,$2B
	.dc.b	$FE,$FF,$39,$A4,$E5,$FF,$04,$4F,$67,$2F,$F8,$18,$C4,$6E,$13,$CE
	.dc.b	$30,$D9,$E1,$7F,$E0,$E9,$E1,$1C,$8F,$1E,$3F,$AF,$BE,$E7,$F8,$66
	.dc.b	$F9,$3E,$4B,$E5,$5D,$23,$DA,$F2,$C2,$C7,$1B,$E9,$87,$82,$F8,$CF
	.dc.b	$33,$67,$AA,$77,$7C,$E3,$C6,$7C,$99,$E2,$D3,$8E,$1E,$67,$96,$1F
	.dc.b	$ED,$E6,$9F,$EE,$08,$FA,$0F,$F9,$F1,$AF,$D4,$F4,$5F,$1D,$C6,$7F
	.dc.b	$A6,$B6,$33,$3C,$DA,$F7,$F8,$FF,$03,$FE,$41,$6A,$69,$1F,$6D,$FF
	.dc.b	$AC,$36,$F8,$71,$C4,$CE,$FF,$E3,$BD,$7F,$01,$19,$7F,$05,$5C,$63
	.dc.b	$E3,$99,$D3,$FD,$6B,$5C,$5F,$F8,$03,$87,$D0,$AE,$17,$D2,$FC,$C7
	.dc.b	$83,$0F,$1B,$88,$EF,$87,$38,$FF,$52,$DF,$38,$C4,$7F,$97,$23,$15
	.dc.b	$C9,$B1,$D4,$FE,$2B,$7E,$93,$FC,$7A,$FA,$4F,$F3,$3D,$7F,$96,$E3
	.dc.b	$FD,$46,$DA,$F6,$9F,$E5,$59,$E0,$7F,$49,$8D,$3F,$E7,$3F,$B1,$FD
	.dc.b	$EC,$FF,$96,$DB,$1F,$C9,$9F,$F3,$C4,$67,$5E,$32,$9D,$DF,$F9,$49
	.dc.b	$B5,$FE,$7A,$78,$CF,$3A,$E3,$1F,$D1,$A7,$C1,$CD,$E4,$B7,$D8,$FF
	.dc.b	$AA,$E7,$24,$C3,$F9,$F4,$EB,$E8,$39,$DF,$1B,$47,$6F,$3E,$61,$FC
	.dc.b	$9B,$AE,$E3,$ED,$9B,$8F,$CB,$E6,$FE,$9F,$7F,$0E,$72,$FF,$A5,$A3
	.dc.b	$FD,$8E,$E4,$FE,$63,$F0,$1F,$13,$EC,$5E,$57,$C6,$F9,$5F,$EE,$15
	.dc.b	$FE,$87,$9C,$BF,$E8,$98,$CE,$5F,$D5,$1B,$97,$F9,$F1,$B3,$FE,$A4
	.dc.b	$AE,$35,$BE,$CF,$A7,$F6,$2F,$8D,$A7,$86,$2D,$FF,$A5,$E5,$F2,$BC
	.dc.b	$0F,$B7,$F6,$3F,$EA,$19,$D4,$7D,$EF,$B1,$7C,$95,$38,$4F,$FA,$86
	.dc.b	$32,$DF,$9C,$E0,$71,$E6,$B1,$FE,$A7,$FB,$1F,$F2,$D3,$7E,$0B,$F0
	.dc.b	$1E,$6F,$F8,$9A,$30,$7F,$AB,$4F,$F2,$D9,$7C,$0F,$C3,$7D,$8B,$EA
	.dc.b	$BF,$27,$3A,$E2,$FE,$FB,$F6,$3F,$90,$E6,$FF,$52,$B6,$7F,$D9,$F7
	.dc.b	$FE,$2E,$7F,$D7,$51,$9A,$FF,$6B,$2F,$80,$FC,$FB,$67,$FB,$97,$C9
	.dc.b	$7F,$A8,$DE,$33,$AE,$C9,$1A,$F9,$2B,$87,$F8,$43,$C4,$7A,$9F,$AA
	.dc.b	$6D,$7F,$7D,$FB,$1F,$E0,$0F,$F0,$79,$D7,$A1,$23,$FC,$05,$F6,$2F
	.dc.b	$AA,$C6,$3F,$E1,$CF,$B1,$FF,$2F,$57,$FB,$55,$B1,$FA,$8F,$B1,$FE
	.dc.b	$FB,$39,$F8,$3F,$DD,$D7,$FA,$66,$38,$7C,$8F,$EC,$E9,$F0,$3A,$8D
	.dc.b	$B2,$7F,$C2,$9F,$63,$F9,$17,$C1,$1F,$C1,$BD,$5F,$B9,$97,$F8,$C1
	.dc.b	$BC,$FF,$C6,$70,$46,$FE,$0B,$F8,$6F,$D8,$FD,$DC,$7F,$DD,$CD,$C3
	.dc.b	$F8,$0D,$FE,$C9,$FF,$1D,$C6,$70,$FF,$71,$FD,$8F,$FA,$6A,$FF,$DE
	.dc.b	$CD,$AF,$E1,$3F,$60,$FA,$CF,$E3,$73,$9B,$E3,$F2,$18,$3F,$8A,$56
	.dc.b	$F0,$E7,$19,$7F,$C6,$91,$FB,$49,$E3,$18,$F3,$BF,$8E,$7D,$8F,$C5
	.dc.b	$CB,$F8,$CB,$6B,$F9,$D2,$F8,$CD,$F9,$D3,$0F,$F2,$04,$ED,$5F,$FA
	.dc.b	$0E,$38,$46,$A7,$CF,$F0,$FE,$8E,$9B,$3E,$66,$DF,$C2,$6E,$3F,$D2
	.dc.b	$F3,$FF,$55,$D7,$68,$3E,$0A,$7D,$8B,$E6,$FF,$A1,$B6,$7C,$FB,$62
	.dc.b	$64,$73,$39,$4E,$EB,$E1,$D6,$9F,$F3,$AC,$66,$FF,$AF,$7D,$6F,$E2
	.dc.b	$2F,$8E,$1F,$D1,$E3,$17,$FE,$A5,$C7,$F8,$D3,$F7,$BC,$0F,$B3,$E1
	.dc.b	$FD,$A6,$61,$F6,$1C,$B4,$F9,$FE,$7F,$D8,$FB,$1F,$B2,$D8,$C9,$F5
	.dc.b	$23,$2D,$3D,$D7,$F5,$B4,$FE,$A7,$18,$8F,$93,$EC,$FE,$33,$C0,$F9
	.dc.b	$2C,$9D,$57,$0A,$FC,$BD,$65,$F8,$B5,$FD,$9F,$EC,$58,$67,$7F,$F0
	.dc.b	$24,$63,$1F,$C4,$FD,$8F,$E4,$BC,$DF,$C9,$B6,$32,$7E,$3A,$32,$D3
	.dc.b	$F3,$BE,$77,$E8,$8F,$D1,$FA,$D6,$32,$4C,$FF,$5B,$19,$7F,$D5,$FF
	.dc.b	$63,$FA,$96,$F6,$FE,$93,$BF,$FC,$17,$39,$D3,$F1,$F5,$F9,$E8,$C6
	.dc.b	$1F,$91,$6C,$7E,$86,$B8,$DC,$AF,$F5,$37,$2F,$F7,$AA,$EF,$18,$F7
	.dc.b	$3F,$77,$F2,$32,$3F,$D5,$9E,$AB,$DB,$C5,$FF,$06,$AB,$A7,$F8,$BE
	.dc.b	$B5,$7F,$C0,$27,$F7,$BF,$51,$23,$F6,$FF,$1B,$70,$FD,$FC,$DC,$3F
	.dc.b	$7F,$36,$BF,$59,$F6,$1F,$CD,$A7,$E0,$DA,$31,$3F,$C0,$3E,$1F,$FC
	.dc.b	$61,$E2,$FF,$8C,$A7,$35,$FC,$11,$AF,$1F,$B1,$D7,$9F,$B7,$87,$8B
	.dc.b	$31,$FC,$A5,$FF,$44,$BF,$F7,$2B,$FC,$3D,$E9,$3B,$73,$F1,$57,$3F
	.dc.b	$3D,$3A,$3C,$87,$F9,$3B,$4F,$99,$F0,$F8,$CB,$BC,$FA,$77,$C6,$5E
	.dc.b	$07,$FC,$9B,$EB,$FF,$92,$A3,$F8,$FB,$5F,$EE,$9E,$76,$BE,$1E,$61
	.dc.b	$FE,$8D,$FA,$5F,$E2,$D7,$FC,$37,$D8,$46,$2B,$F1,$B1,$8C,$3A,$F9
	.dc.b	$38,$AE,$2D,$8C,$E1,$D5,$F1,$DE,$5F,$F0,$34,$DC,$3C,$17,$CA,$F8
	.dc.b	$57,$F8,$2E,$F5,$7B,$F4,$65,$6B,$0F,$F3,$25,$FF,$A9,$70,$F5,$D6
	.dc.b	$A3,$07,$F4,$2F,$B1,$7C,$B7,$F6,$A8,$C9,$FE,$19,$9F,$34,$EB,$4F
	.dc.b	$1B,$67,$6D,$D3,$E2,$47,$1C,$B7,$87,$59,$CD,$6F,$1F,$E8,$8D,$BF
	.dc.b	$AF,$8F,$82,$F8,$56,$9B,$3E,$96,$BD,$CF,$1B,$8A,$E9,$06,$E5,$FE
	.dc.b	$0C,$C3,$C7,$5B,$26,$5F,$2F,$1F,$13,$EA,$26,$1F,$1A,$F9,$63,$E5
	.dc.b	$9C,$E1,$D1,$E5,$96,$27,$57,$E6,$9C,$B4,$95,$C5,$ED,$87,$89,$CD
	.dc.b	$72,$9F,$9F,$87,$8A,$7E,$25,$F7,$69,$CB,$0C,$3F,$0F,$1F,$7B,$93
	.dc.b	$A9,$C2,$FC,$47,$6C,$7C,$4E,$31,$F1,$1F,$8E,$BE,$B9,$7D,$03,$E7
	.dc.b	$D7,$09,$DD,$F5,$78,$3E,$03,$C9,$8F,$9B,$2F,$64,$60,$FC,$7F,$D8
	.dc.b	$FF,$2F,$7F,$C1,$87,$E6,$DC,$38,$C3,$F2,$CD,$8C,$BF,$90,$79,$65
	.dc.b	$F9,$46,$E1,$9E,$CF,$CF,$F8,$7F,$EF,$18,$C5,$F8,$A7,$F4,$75,$C2
	.dc.b	$3F,$59,$5B,$7C,$93,$A2,$3F,$A8,$8D,$3F,$95,$9E,$4D,$9D,$DF,$EA
	.dc.b	$2B,$78,$2F,$6D,$AD,$E1,$FA,$CA,$DD,$68,$93,$EC,$C3,$67,$98,$F6
	.dc.b	$1E,$53,$27,$8C,$CE,$1E,$37,$91,$8B,$F8,$B1,$98,$E4,$64,$EC,$E4
	.dc.b	$FE,$0F,$F6,$3F,$C2,$1B,$F3,$5F,$12,$0D,$A9,$5F,$C9,$27,$84,$7F
	.dc.b	$93,$5B,$1F,$E5,$E5,$C9,$CC,$E3,$1F,$C7,$A3,$69,$DB,$D6,$37,$7E
	.dc.b	$6A,$E5,$3E,$1F,$33,$FC,$B9,$D1,$FD,$C3,$0F,$F3,$94,$FF,$8D,$DA
	.dc.b	$A3,$53,$C3,$87,$D0,$D9,$F4,$7A,$95,$F2,$36,$FF,$DE,$31,$D7,$BF
	.dc.b	$87,$4B,$EA,$73,$3D,$F1,$FF,$00,$46,$32,$F3,$5F,$32,$57,$33,$FC
	.dc.b	$9D,$F6,$3F,$AC,$9F,$F0,$5B,$FE,$DB,$8C,$F2,$F9,$63,$FA,$44,$77
	.dc.b	$3F,$D1,$AE,$DC,$07,$13,$D2,$79,$1C,$F9,$7E,$07,$D0,$3E,$C6,$93
	.dc.b	$0F,$F3,$17,$D8,$FF,$87,$B0,$FF,$0F,$D6,$E7,$FC,$CC,$D7,$C2,$D8
	.dc.b	$FF,$12,$CF,$1D,$3F,$D2,$92,$F8,$FF,$09,$2F,$9F,$CD,$D7,$9C,$5C
	.dc.b	$3C,$33,$F5,$FE,$C5,$F4,$BE,$AF,$97,$98,$7F,$9E,$67,$2D,$9C,$BE
	.dc.b	$69,$E5,$12,$34,$43,$D7,$F2,$1F,$85,$FA,$F7,$89,$49,$3F,$E8,$78
	.dc.b	$71,$E2,$3E,$9C,$70,$75,$3C,$E7,$A7,$91,$D7,$A2,$57,$38,$CB,$D6
	.dc.b	$32,$F4,$8F,$CC,$7D,$83,$EC,$7C,$DD,$92,$BF,$D3,$53,$96,$CE,$5F
	.dc.b	$2C,$F2,$89,$1A,$21,$EA,$FB,$1F,$D3,$75,$3D,$DB,$EF,$4A,$FF,$54
	.dc.b	$4E,$5B,$39,$7F,$2B,$09,$E7,$3D,$3E,$AE,$57,$27,$BF,$D7,$E2,$38
	.dc.b	$C5,$E6,$FF,$38,$E0,$AF,$DD,$79,$5F,$E2,$49,$CF,$57,$A0,$CE,$B3
	.dc.b	$1F,$3C,$65,$FF,$60,$57,$18,$FF,$75,$36,$3F,$DF,$6D,$7F,$C7,$73
	.dc.b	$FC,$27,$EC,$1F,$63,$7A,$F1,$D7,$2F,$FD,$99,$5F,$A1,$6C,$E5,$FD
	.dc.b	$CC,$27,$94,$70,$F3,$7B,$11,$E7,$D7,$5E,$31,$74,$FF,$4A,$DA,$BE
	.dc.b	$2F,$B5,$FC,$7F,$6F,$CB,$E1,$FE,$DF,$6B,$FC,$2B,$54,$F0,$8D,$74
	.dc.b	$71,$30,$EE,$2E,$3F,$EA,$4A,$E3,$3D,$67,$F9,$77,$D7,$BE,$C7,$F9
	.dc.b	$87,$D8,$FE,$0E,$BF,$DE,$0D,$7F,$8D,$CE,$DF,$F7,$A2,$F3,$F8,$F2
	.dc.b	$C9,$CC,$7F,$3B,$FB,$17,$CC,$FF,$3D,$FB,$1F,$E2,$9C,$EB,$91,$87
	.dc.b	$15,$C1,$D6,$DF,$EB,$E8,$7C,$57,$3F,$12,$56,$B0,$E8,$AB,$6E,$4F
	.dc.b	$07,$D8,$FF,$34,$9D,$E2,$F1,$FC,$C7,$F5,$39,$D5,$70,$F9,$8F,$57
	.dc.b	$99,$23,$A0,$94,$FF,$5B,$F7,$5F,$63,$FD,$77,$EC,$7E,$EE,$FF,$ED
	.dc.b	$86,$E9,$E8,$AD,$E3,$D1,$7C,$7F,$1F,$26,$4E,$63,$FB,$47,$D8,$AB
	.dc.b	$B7,$B1,$2B,$CB,$39,$6F,$C0,$7F,$6F,$F6,$5F,$07,$4E,$71,$D3,$E5
	.dc.b	$75,$F0,$92,$F9,$C7,$F7,$7F,$B1,$7C,$0F,$3F,$0A,$57,$C4,$9C,$BF
	.dc.b	$1B,$EC,$5F,$29,$FD,$FF,$EB,$FF,$A2,$CF,$F6,$86,$F2,$FF,$BF,$2B
	.dc.b	$73,$F3,$BE,$75,$F6,$3F,$E0,$79,$7C,$85,$3F,$C1,$53,$AA,$E1,$F1
	.dc.b	$38,$B9,$18,$3A,$79,$5E,$3F,$3E,$46,$0F,$BB,$F6,$2F,$A5,$EB,$52
	.dc.b	$73,$1F,$E1,$99,$D4,$60,$87,$B7,$E3,$FF,$C3,$DF,$61,$F7,$23,$F1
	.dc.b	$3B,$F0,$7E,$15,$EB,$B2,$46,$A3,$E4,$18,$7F,$C1,$F5,$CD,$B3,$9E
	.dc.b	$07,$77,$E9,$30,$7D,$8F,$8B,$87,$E9,$7A,$53,$1F,$CD,$E5,$F3,$1B
	.dc.b	$39,$7E,$B4,$F1,$7F,$43,$F6,$3F,$E4,$A7,$FC,$49,$93,$EB,$7F,$95
	.dc.b	$47,$EC,$63,$1C,$3F,$E3,$FA,$E1,$1E,$69,$FE,$09,$18,$D3,$F7,$6D
	.dc.b	$8C,$F7,$71,$ED,$FF,$24,$64,$F9,$0B,$E4,$4B,$FD,$AC,$E5,$B3,$C0
	.dc.b	$8F,$EE,$BE,$C5,$7F,$E5,$98,$7D,$8F,$F3,$97,$F8,$C4,$EB,$4F,$E3
	.dc.b	$53,$99,$FE,$7D,$7F,$C4,$A3,$18,$78,$67,$9B,$FC,$5A,$F9,$46,$0E
	.dc.b	$B8,$78,$6B,$8D,$F3,$BF,$E3,$A7,$38,$F0,$5F,$26,$D7,$19,$DD,$71
	.dc.b	$BF,$E6,$11,$C4,$F1,$4E,$E3,$C3,$87,$B2,$78,$C5,$38,$D7,$35,$87
	.dc.b	$86,$79,$D6,$E7,$E1,$96,$3C,$0F,$63,$E0,$1C,$AF,$E8,$3F,$ED,$EA
	.dc.b	$F8,$AD,$8C,$BD,$8F,$F7,$19,$E2,$F4,$1F,$19,$88,$C3,$6B,$FA,$9B
	.dc.b	$67,$73,$B1,$F1,$F5,$A6,$5C,$11,$F4,$E3,$FE,$B0,$DF,$E0,$47,$F5
	.dc.b	$D9,$F8,$53,$98,$F7,$99,$38,$ED,$FA,$1D,$7F,$15,$19,$7F,$19,$3C
	.dc.b	$5C,$C7,$91,$CC,$F1,$D3,$E1,$DF,$82,$31,$5F,$D9,$0E,$13,$FE,$BA
	.dc.b	$8C,$C7,$F6,$89,$E1,$87,$AA,$36,$7F,$72,$9D,$BE,$5A,$E5,$3C,$23
	.dc.b	$FD,$43,$87,$BC,$FD,$E3,$D4,$F7,$C7,$F6,$88,$D3,$CA,$3F,$D5,$27
	.dc.b	$53,$C0,$FF,$6E,$FB,$1F,$DF,$37,$6F,$E1,$15,$DA,$7F,$C1,$93,$92
	.dc.b	$57,$FA,$D2,$38,$B4,$E3,$1F,$E0,$88,$CC,$F3,$E8,$CD,$7F,$85,$E7
	.dc.b	$FA,$15,$7F,$1B,$FB,$1F,$F5,$0B,$E5,$AD,$D2,$B6,$FF,$24,$A7,$CF
	.dc.b	$6A,$73,$87,$AA,$73,$87,$BF,$CF,$D1,$E2,$9F,$E5,$9D,$98,$9D,$13
	.dc.b	$1F,$E6,$3C,$E9,$D4,$7C,$1B,$E5,$3F,$EC,$1A,$E1,$7F,$2A,$31,$5F
	.dc.b	$ED,$AF,$B1,$79,$3E,$CF,$94,$FE,$DF,$87,$F8,$82,$72,$FF,$88,$A3
	.dc.b	$35,$FE,$E2,$6E,$1F,$82,$6C,$FF,$8D,$AB,$8D,$6F,$AF,$E4,$FD,$8B
	.dc.b	$E2,$E9,$E1,$8B,$7F,$E3,$09,$7C,$AF,$03,$E8,$FD,$8F,$FA,$96,$75
	.dc.b	$1F,$53,$EC,$5F,$23,$4E,$13,$FE,$3A,$8C,$B7,$E7,$38,$1C,$79,$AC
	.dc.b	$7F,$90,$3E,$C7,$FC,$C4,$DF,$80,$FA,$DE,$67,$F8,$92,$30,$7F,$91
	.dc.b	$8F,$F3,$19,$7B,$FE,$E7,$D8,$3E,$C7,$F1,$33,$AE,$27,$2F,$EA,$BE
	.dc.b	$C7,$F3,$B1,$89,$FF,$2F,$C6,$27,$F4,$77,$FE,$2E,$9F,$F2,$AC,$E5
	.dc.b	$DF,$CD,$7E,$5A,$0F,$E3,$7E,$49,$EB,$66,$35,$D9,$2F,$FB,$D5,$7C
	.dc.b	$DC,$BC,$47,$F9,$EF,$0F,$F3,$0C,$FF,$99,$3C,$47,$A7,$FA,$46,$C7
	.dc.b	$F8,$0F,$EC,$7F,$80,$DF,$F2,$7A,$D7,$A3,$2B,$FC,$11,$F6,$2F,$A9
	.dc.b	$C5,$F8,$07,$D8,$FF,$98,$6B,$FC,$E6,$D8,$FF,$09,$FD,$8F,$F7,$C9
	.dc.b	$CF,$C0,$FE,$DA,$BF,$D5,$B1,$C3,$E3,$FF,$5D,$4F,$7F,$51,$B6,$4F
	.dc.b	$F8,$6F,$EC,$7F,$2C,$F8,$23,$F8,$0F,$A9,$F8,$1C,$BC,$CD,$E7,$FE
	.dc.b	$2F,$82,$37,$F0,$1F,$C2,$3E,$C7,$F0,$58,$FF,$A3,$1B,$87,$F8,$B2
	.dc.b	$FF,$6B,$87,$F2,$C9,$CB,$FC,$BB,$EC,$7F,$D4,$B7,$FE,$94,$6D,$FF
	.dc.b	$8D,$3E,$C1,$F5,$7F,$C5,$A7,$37,$C7,$E4,$30,$7F,$0F,$AD,$E1,$CE
	.dc.b	$32,$FF,$8E,$E3,$F7,$33,$C6,$31,$E7,$7F,$17,$FB,$1F,$8B,$97,$F9
	.dc.b	$11,$B5,$FC,$E1,$7C,$56,$FC,$E9,$87,$F9,$1E,$76,$FF,$B0,$9D,$B9
	.dc.b	$8F,$37,$C3,$79,$FA,$6C,$F9,$5B,$8F,$C1,$6E,$3F,$D2,$33,$FF,$56
	.dc.b	$D7,$FA,$EB,$EC,$5F,$33,$FC,$E9,$B1,$E3,$DB,$13,$23,$83,$EF,$8C
	.dc.b	$63,$F9,$5A,$E2,$FF,$99,$3E,$B7,$F0,$57,$FE,$D6,$9C,$E3,$FD,$6A
	.dc.b	$73,$5F,$C0,$31,$FD,$74,$F9,$FE,$7F,$EA,$78,$5F,$55,$98,$FD,$37
	.dc.b	$2D,$3E,$57,$A1,$F4,$FE,$C7,$EA,$36,$32,$7C,$F8,$CB,$4F,$89,$ED
	.dc.b	$FC,$33,$F0,$9E,$05,$8C,$93,$3F,$C7,$46,$5F,$85,$3F,$D6,$7E,$C5
	.dc.b	$79,$CF,$FB,$CC,$63,$0F,$C0,$7D,$8F,$E2,$7C,$EF,$C5,$36,$32,$7E
	.dc.b	$16,$32,$D3,$F2,$9E,$7F,$E6,$4F,$CD,$7A,$F6,$32,$4C,$FF,$49,$19
	.dc.b	$7F,$86,$7D,$8F,$E8,$1B,$DB,$F9,$BE,$FF,$EF,$B3,$9D,$3F,$C3,$AD
	.dc.b	$C1,$72,$30,$FF,$3A,$EB,$FE,$1E,$F4,$9D,$D7,$2C,$3F,$BF,$72,$FF
	.dc.b	$79,$9C,$9D,$29,$87,$4B,$56,$E7,$F7,$DF,$21,$FB,$93,$FC,$11,$EA
	.dc.b	$BD,$BC,$63,$FB,$99,$B5,$FE,$B8,$AD,$4E,$4F,$D9,$7D,$3F,$EE,$CF
	.dc.b	$DE,$7C,$75,$8C,$92,$BF,$7D,$19,$7F,$9B,$7D,$87,$F3,$3C,$FF,$C8
	.dc.b	$6D,$AF,$8F,$F0,$FF,$E2,$7F,$13,$FC,$53,$39,$BF,$96,$D4,$BE,$C7
	.dc.b	$7E,$77,$07,$8F,$8B,$31,$F9,$37,$CB,$E2,$B0,$F9,$26,$E7,$E4,$9B
	.dc.b	$5F,$3F,$17,$47,$90,$FF,$21,$E9,$F2,$FE,$23,$15,$DB,$2F,$2D,$7A
	.dc.b	$B7,$CC,$7C,$27,$FC,$8B,$EB,$FF,$93,$AB,$EE,$36,$7F,$12,$FF,$45
	.dc.b	$8E,$8E,$63,$FB,$BF,$C3,$FF,$B9,$FD,$67,$D2,$F2,$F2,$70,$F4,$9B
	.dc.b	$5D,$26,$BF,$E0,$06,$DF,$F1,$6B,$E7,$3E,$2A,$CD,$4C,$3F,$D3,$98
	.dc.b	$77,$6D,$6D,$AF,$56,$E1,$FD,$8F,$0E,$B3,$FE,$10,$F3,$FC,$3E,$FF
	.dc.b	$89,$97,$CB,$EC,$7F,$C2,$7E,$97,$29,$F4,$C6,$2F,$FD,$6F,$97,$E9
	.dc.b	$63,$18,$F6,$6C,$7C,$EA,$E3,$EE,$7F,$71,$BE,$27,$EE,$2B,$6D,$8D
	.dc.b	$5F,$A3,$3F,$56,$1D,$E7,$2D,$C7,$D1,$18,$BF,$B7,$8F,$82,$72,$48
	.dc.b	$F0,$9F,$0F,$2C,$63,$E9,$BE,$52,$CB,$CA,$EC,$E2,$E0,$91,$C3,$2F
	.dc.b	$3C,$FA,$F0,$EB,$97,$9E,$3C,$33,$A7,$B5,$78,$2F,$9D,$FD,$1C,$78
	.dc.b	$D7,$33,$C3,$5D,$F2,$F7,$46,$2B,$6F,$9F,$2F,$64,$63,$0F,$5E,$3E
	.dc.b	$17,$BD,$71,$3C,$B7,$D2,$F8,$C6,$5F,$A9,$19,$CB,$D9,$1A,$33,$97
	.dc.b	$7C,$BD,$87,$9D,$C6,$5E,$6C,$BC,$53,$F0,$7E,$C7,$F9,$46,$5F,$EE
	.dc.b	$93,$E9,$38,$7F,$DD,$99,$FE,$4D,$B1,$97,$F2,$8F,$2C,$7F,$28,$DC
	.dc.b	$B3,$DB,$F9,$FF,$13,$FD,$E5,$18,$BF,$BB,$5B,$7F,$2C,$D7,$F0,$93
	.dc.b	$C1,$CD,$7E,$A6,$75,$81,$B5,$BA,$F5,$D6,$C8,$FE,$32,$35,$5C,$67
	.dc.b	$38,$7A,$9F,$45,$63,$0F,$44,$E6,$39,$19,$3C,$A7,$B2,$78,$B4,$91
	.dc.b	$C4,$FD,$D5,$F8,$B0,$F5,$92,$B8,$CE,$AB,$B1,$EC,$3C,$24,$7A,$5F
	.dc.b	$98,$FE,$11,$7F,$0F,$EC,$7F,$8A,$CF,$F1,$65,$F3,$39,$35,$52,$FF
	.dc.b	$8F,$CF,$08,$FE,$EC,$D8,$FF,$01,$AE,$5E,$C7,$18,$FE,$03,$1B,$4E
	.dc.b	$DE,$B1,$BB,$F2,$57,$29,$F0,$F9,$9F,$E0,$5E,$A7,$0F,$8B,$D4,$C3
	.dc.b	$6B,$FA,$B6,$5F,$EE,$86,$DE,$9F,$3D,$57,$C6,$D9,$E1,$1F,$EB,$15
	.dc.b	$E0,$71,$EA,$7D,$4E,$B2,$F1,$DA,$73,$8F,$3B,$93,$E4,$6B,$FE,$BD
	.dc.b	$FB,$1F,$F1,$03,$79,$7C,$91,$FC,$E6,$35,$CD,$89,$91,$FD,$CE,$75
	.dc.b	$8F,$8A,$11,$8D,$7F,$C5,$8F,$FB,$8A,$3F,$BD,$7C,$FB,$EC,$7E,$13
	.dc.b	$63,$31,$F1,$7E,$C7,$A6,$1D,$6B,$55,$FE,$D5,$6C,$66,$B8,$61,$D3
	.dc.b	$EA,$E5,$F8,$1E,$FF,$2F,$88,$E5,$16,$FB,$4F,$D5,$FB,$17,$D3,$79
	.dc.b	$BE,$5A,$61,$FE,$13,$9C,$B6,$72,$F9,$67,$94,$48,$D1,$0F,$57,$C7
	.dc.b	$FE,$0F,$EC,$5E,$2D,$24,$FF,$88,$63,$F0,$6D,$8C,$BF,$EF,$78,$E2
	.dc.b	$EF,$8F,$A7,$93,$D7,$A6,$4F,$38,$CB,$D6,$32,$F3,$8F,$CB,$7D,$8B
	.dc.b	$EC,$7C,$DD,$92,$BF,$C6,$93,$96,$CE,$5F,$24,$F2,$89,$1A,$21,$E9
	.dc.b	$FB,$1F,$D2,$76,$3D,$DB,$EF,$4A,$FF,$21,$4E,$5B,$39,$5C,$9F,$93
	.dc.b	$95,$74,$AE,$BF,$57,$2F,$9B,$E0,$FA,$FC,$47,$2E,$8F,$F4,$26,$0A
	.dc.b	$CE,$3D,$6F,$F7,$5E,$4F,$F8,$C2,$3C,$17,$EA,$33,$AC,$C7,$D3,$19
	.dc.b	$7F,$CB,$95,$C6,$3F,$D2,$4D,$8F,$F4,$FB,$53,$7E,$C1,$F6,$3F,$C2
	.dc.b	$D7,$5D,$EB,$97,$FE,$69,$9C,$B6,$72,$FB,$CF,$F5,$6C,$27,$A4,$F5
	.dc.b	$F3,$A5,$73,$7C,$1D,$78,$8E,$42,$FF,$DF,$EE,$CF,$14,$7C,$8F,$73
	.dc.b	$F9,$16,$DF,$9C,$9E,$2D,$DB,$F8,$66,$8A,$E1,$5F,$E8,$3A,$D6,$BF
	.dc.b	$EB,$96,$C7,$C8,$9D,$BE,$4A,$E4,$FD,$08,$F0,$9F,$CC,$FE,$BD,$F6
	.dc.b	$3F,$CD,$7E,$C7,$F1,$35,$FE,$91,$6B,$EE,$9D,$BF,$E9,$55,$E6,$F1
	.dc.b	$E3,$93,$98,$FE,$81,$F6,$2F,$99,$FE,$85,$F6,$3F,$C6,$B6,$FF,$69
	.dc.b	$36,$FF,$DB,$F5,$B7,$FD,$44,$BE,$2F,$8F,$89,$27,$31,$FD,$37,$EC
	.dc.b	$7F,$98,$47,$D4,$B7,$91,$E0,$35,$E5,$EF,$E6,$7C,$7E,$6C,$BF,$F5
	.dc.b	$86,$09,$FE,$B1,$EF,$3E,$C7,$FA,$D7,$D8,$FD,$0B,$36,$75,$1F,$EB
	.dc.b	$95,$F2,$1C,$3C,$99,$39,$8F,$EC,$9F,$62,$AE,$BE,$C4,$AF,$F6,$14
	.dc.b	$E5,$BF,$03,$FD,$AF,$D9,$7C,$2D,$39,$C7,$4F,$95,$D7,$C2,$CB,$E7
	.dc.b	$1F,$DC,$FE,$C5,$F0,$3C,$FC,$29,$5F,$ED,$19,$CB,$FE,$D3,$FB,$17
	.dc.b	$CA,$FF,$7B,$FA,$FF,$E8,$73,$FA,$26,$F8,$FF,$D9,$6B,$73,$FE,$DB
	.dc.b	$F9,$D7,$D8,$FF,$80,$E5,$F2,$34,$FF,$03,$4E,$AB,$87,$C5,$62,$E4
	.dc.b	$60,$E9,$ED,$78,$FD,$19,$18,$3F,$DD,$5F,$62,$F8,$AE,$B5,$27,$31
	.dc.b	$FE,$15,$9D,$46,$08,$7B,$7E,$AB,$FC,$35,$F6,$1F,$72,$3F,$DE,$DB
	.dc.b	$F0,$7E,$15,$EB,$B2,$46,$A3,$E4,$18,$7F,$84,$D7,$36,$CE,$78,$3F
	.dc.b	$C1,$D6,$FB,$1F,$0F,$07,$D1,$F4,$E6,$03,$9C,$69,$F2,$C6,$9C,$9B
	.dc.b	$BF,$F0,$D7,$D8,$FF,$91,$63,$F9,$24,$3E,$B7,$F9,$24,$6F,$CD,$E1
	.dc.b	$18,$8F,$E4,$35,$C6,$B9,$F7,$7F,$8F,$E3,$17,$E5,$A4,$61,$F4,$2E
	.dc.b	$DF,$F2,$36,$2F,$90,$BE,$44,$BF,$11,$EF,$9D,$38,$0E,$1F,$E6,$AD
	.dc.b	$CB,$FC,$70,$DC,$7C,$9F,$63,$E1,$9F,$95,$F6,$3C,$AF,$E6,$4E,$54
	.dc.b	$65,$EF,$8F,$B6,$31,$87,$85,$D5,$71,$6E,$5F,$DE,$2F,$9D,$7C,$53
	.dc.b	$81,$6F,$A3,$E6,$9E,$15,$CB,$0F,$3C,$EE,$E9,$F1,$6F,$8B,$EA,$BE
	.dc.b	$33,$CE,$B7,$3E,$A2,$C7,$89,$E1,$1A,$C3,$D5,$3D,$E7,$D2,$77,$35
	.dc.b	$5D,$9E,$27,$FB,$7E,$7E,$53,$FE,$72,$6B,$E3,$35,$E8,$FF,$9F,$5B
	.dc.b	$F1,$BF,$EE,$75,$3F,$D2,$E3,$0D,$8D,$0F,$8F,$AD,$32,$E0,$8F,$A8
	.dc.b	$FF,$15,$EF,$FE,$8E,$D3,$EF,$47,$EB,$67,$E3,$E7,$FE,$B1,$D7,$E4
	.dc.b	$CE,$4E,$51,$97,$A1,$9A,$FC,$44,$65,$EB,$1E,$57,$37,$E1,$D3,$FD
	.dc.b	$2F,$39,$7F,$D6,$91,$98,$FF,$4D,$E1,$E4,$8C,$47,$EB,$E3,$18,$79
	.dc.b	$A3,$91,$FC,$06,$76,$F9,$6B,$AC,$F0,$8F,$F0,$EE,$1E,$B3,$FD,$40
	.dc.b	$EC,$E5,$1F,$B1,$8D,$3C,$A3,$FC,$4A,$6C,$F9,$AF,$ED,$FE,$C7,$F7
	.dc.b	$CD,$DB,$FD,$7B,$5F,$EA,$D9,$FE,$39,$39,$25,$7F,$8C,$23,$8B,$4E
	.dc.b	$31,$FC,$3E,$33,$3C,$F9,$F3,$5F,$C9,$E7,$FA,$25,$7F,$0C,$A7,$D8
	.dc.b	$FF,$A8,$A3,$CD,$7C,$2D,$7C,$1F,$E4,$F4,$F9,$FD,$4E,$70,$F5,$CE
	.dc.b	$70,$F8,$1E,$7E,$8F,$1C,$FF,$2F,$EB,$C4,$E8,$98,$FF,$34,$E6,$4E
	.dc.b	$A3,$E1,$5F,$29,$FF,$31,$D7,$0B,$F9,$71,$8A,$FF,$2B,$FD,$8B,$C9
	.dc.b	$F6,$7C,$A7,$F7,$2C,$3F,$95,$4E,$5F,$E5,$91,$9A,$FF,$2F,$B7,$0F
	.dc.b	$F7,$13,$67,$F9,$F5,$71,$AD,$F5,$7C,$9F,$B1,$7C,$65,$3C,$31,$6F
	.dc.b	$F9,$D4,$BE,$53,$81,$F4,$7E,$C7,$FD,$4D,$3A,$8F,$A9,$F6,$2F,$92
	.dc.b	$A7,$09,$FE,$89,$19,$6F,$CE,$70,$38,$F3,$58,$FE,$95,$F6,$3F,$E5
	.dc.b	$C6,$FC,$0F,$D6,$F3,$3F,$C4,$F1,$83,$FA,$81,$FE,$5D,$2F,$83,$F7
	.dc.b	$3E,$C1,$F6,$3F,$89,$9D,$71,$7F,$A2,$3E,$C7,$F3,$AD,$38,$7A,$5F
	.dc.b	$8F,$8C,$F3,$AB,$FB,$54,$3D,$FD,$1F,$8A,$F9,$2F,$F5,$3C,$FF,$63
	.dc.b	$9D,$37,$B3,$A1,$BF,$92,$B1,$FD,$A7,$C5,$7A,$9F,$9F,$6C,$7E,$97
	.dc.b	$EC,$7F,$7B,$AF,$F2,$3A,$D7,$A1,$2B,$F5,$5F,$62,$FA,$AC,$5F,$FA
	.dc.b	$9B,$EC,$7F,$CB,$D5,$FD,$DD,$B1,$FE,$10,$FB,$1F,$F0,$14,$E7,$E0
	.dc.b	$BF,$61,$5F,$EA,$38,$E1,$F2,$1F,$A8,$A7,$C0,$6A,$36,$C9,$FD,$E7
	.dc.b	$D8,$FE,$41,$F0,$47,$EE,$FD,$5F,$B7,$97,$FA,$E9,$BC,$DF,$C4,$B0
	.dc.b	$46,$FE,$07,$F8,$17,$D8,$FD,$CC,$7F,$C2,$0D,$C3,$FC,$53,$7F,$AF
	.dc.b	$C3,$EC,$47,$F8,$D6,$75,$18,$87,$D8,$FF,$A7,$6F,$FC,$34,$D8,$FF
	.dc.b	$19,$FD,$83,$EB,$3F,$89,$4E,$6F,$93,$6F,$FC,$5C,$DB,$4A,$F9,$57
	.dc.b	$6F,$8F,$E6,$7F,$12,$8F,$DC,$4F,$48,$C7,$9F,$FC,$5B,$EB,$7E,$56
	.dc.b	$1F,$E4,$66,$D7,$F3,$85,$F1,$7B,$F3,$E6,$5F,$CB,$67,$6F,$F9,$45
	.dc.b	$DB,$98,$F1,$78,$3E,$3E,$9B,$3F,$CD,$74,$FF,$55,$B6,$74,$D3,$53
	.dc.b	$FC,$DF,$EC,$5F,$35,$4F,$56,$D6,$BF,$C9,$13,$A6,$CE,$9A,$4D,$7F
	.dc.b	$CC,$BF,$5B,$F8,$68,$FF,$2E,$57,$19,$CF,$27,$CB,$F0,$25,$7A,$7C
	.dc.b	$3B,$5F,$72,$6F,$0F,$F6,$0D,$95,$8B,$F8,$DE,$7F,$D1,$FB,$1F,$A4
	.dc.b	$D8,$F9,$8D,$C7,$EA,$37,$0F,$C7,$67,$FD,$46,$7F,$07,$5E,$8F,$6B
	.dc.b	$F0,$1E,$07,$C7,$64,$7C,$47,$2D,$24,$FF,$A1,$BE,$C5,$97,$08,$C3
	.dc.b	$F8,$7F,$B1,$FC,$47,$9B,$6F,$F0,$4D,$AF,$C5,$37,$3F,$C5,$3F,$93
	.dc.b	$F3,$A5,$7E,$07,$D6,$B7,$FA,$49,$B7,$F2,$66,$D7,$FA,$3F,$EC,$7F
	.dc.b	$3F,$7F,$00,$FF,$53,$38,$AF,$CC,$78,$1F,$DF,$63,$15,$C7,$BF,$07
	.dc.b	$E8,$66,$C7,$FA,$3E,$77,$7F,$AE,$E6,$FE,$EB,$39,$BF,$EF,$B1,$89
	.dc.b	$FD,$DF,$C8,$7E,$BC,$FF,$00,$FA,$8F,$6B,$18,$7E,$7E,$6D,$7F,$AD
	.dc.b	$2B,$53,$93,$F5,$9F,$51,$FB,$53,$F6,$DF,$19,$63,$24,$AF,$DC,$46
	.dc.b	$5F,$88,$3E,$C3,$F9,$9E,$DF,$C5,$5B,$5F,$EC,$DF,$87,$FF,$10,$F8
	.dc.b	$BF,$E2,$39,$CD,$FF,$B4,$5A,$F3,$FB,$1D,$79,$DB,$78,$78,$B3,$1F
	.dc.b	$F5,$C5,$FF,$40,$BF,$1B,$FC,$19,$E9,$3B,$73,$F1,$37,$3F,$37,$3F
	.dc.b	$8C,$FF,$1E,$E9,$F2,$FE,$1F,$19,$76,$9F,$46,$F8,$C7,$BB,$FE,$3D
	.dc.b	$F5,$FF,$C8,$F5,$FE,$D7,$6C,$FF,$BB,$5F,$E8,$71,$FE,$CF,$73,$1F
	.dc.b	$DC,$FE,$1B,$F9,$FF,$AC,$FA,$5E,$5E,$46,$1E,$93,$6B,$A4,$D7,$FB
	.dc.b	$E3,$6F,$F8,$65,$F3,$9F,$15,$66,$A6,$1F,$E9,$5C,$3B,$B6,$B6,$D7
	.dc.b	$AB,$70,$FE,$C3,$87,$59,$FF,$06,$79,$1E,$1F,$7B,$C4,$CB,$E5,$F6
	.dc.b	$2F,$95,$FE,$CB,$7F,$EA,$FC,$BE,$45,$FA,$EB,$5E,$9F,$0B,$DE,$5D
	.dc.b	$E7,$38,$71,$3E,$75,$F1,$AD,$E0,$99,$5C,$2F,$95,$6A,$78,$E3,$DF
	.dc.b	$19,$1D,$45,$F1,$C3,$CF,$5A,$C3,$9D,$D9,$CD,$F1,$BE,$0F,$8B,$0F
	.dc.b	$03,$D1,$B9,$FC,$CA,$F4,$4E,$6F,$A3,$C5,$FA,$59,$7B,$DC,$C6,$CF
	.dc.b	$45,$73,$78,$1E,$3B,$F0,$E1,$E4,$8F,$B3,$97,$28,$C1,$9A,$E5,$3D
	.dc.b	$72,$F7,$BC,$F0,$F1,$DF,$1C,$3A,$9E,$CC,$7C,$D5,$B2,$5F,$E2,$70
	.dc.b	$EB,$8F,$82,$F8,$E3,$D2,$F7,$97,$C4,$AD,$4E,$70,$F1,$92,$75,$7E
	.dc.b	$1C,$BD,$9F,$63,$FC,$9F,$1F,$F7,$2E,$28,$C4,$7D,$03,$E5,$16,$35
	.dc.b	$DB,$FE,$EC,$8F,$C1,$CF,$3C,$7F,$28,$DC,$7C,$CD,$9F,$C8,$F8,$7F
	.dc.b	$EE,$D9,$CD,$FE,$92,$B5,$5D,$23,$F4,$F5,$B8,$FC,$CC,$EB,$0F,$1C
	.dc.b	$7E,$36,$31,$3B,$8E,$46,$6B,$B1,$D1,$A6,$49,$87,$EB,$67,$95,$F6
	.dc.b	$8F,$6D,$70,$75,$57,$0F,$5B,$C9,$FA,$D7,$CB,$0F,$41,$2F,$C5,$5B
	.dc.b	$3D,$AF,$37,$E3,$94,$F5,$DF,$89,$FD,$65,$79,$8E,$D3,$BC,$11,$99
	.dc.b	$C5,$FC,$67,$F8,$E7,$D8,$FF,$19,$C3,$F8,$D2,$F9,$9F,$94,$D9,$DE
	.dc.b	$58,$23,$83,$9F,$2F,$03,$2F,$F0,$99,$ED,$3F,$BC,$53,$B9,$E4,$6E
	.dc.b	$70,$F1,$F3,$3F,$81,$F4,$7F,$A0,$6F,$FC,$CF,$3F,$C7,$63,$1E,$A6
	.dc.b	$B7,$ED,$76,$BF,$91,$5F,$FB,$95,$E1,$3E,$3A,$D4,$27,$AB,$FD,$F9
	.dc.b	$E0,$DF,$03,$A7,$53,$EA,$73,$97,$FA,$62,$31,$97,$4C,$17,$49,$87
	.dc.b	$88,$F2,$BE,$2F,$B1,$FD,$96,$1F,$E9,$F7,$FD,$B1,$3B,$0F,$F4,$A7
	.dc.b	$FC,$26,$D8,$E1,$CD,$CC,$E9,$CD,$2B,$95,$74,$7F,$C3,$7F,$3E,$FB
	.dc.b	$0F,$86,$D8,$F0,$35,$FF,$10,$FD,$8F,$FA,$CF,$1F,$F5,$A4,$E5,$B5
	.dc.b	$C8,$F9,$B3,$BE,$0F,$E7,$B4,$FA,$BE,$E4,$C3,$BF,$CD,$D7,$CB,$17
	.dc.b	$1E,$F3,$F6,$3E,$C5,$F4,$BE,$8F,$96,$98,$7C,$79,$CB,$67,$2F,$C0
	.dc.b	$9E,$51,$23,$44,$3E,$2F,$C7,$FE,$1B,$EC,$5E,$2D,$24,$FD,$58,$FB
	.dc.b	$ED,$8C,$BF,$E4,$88,$E2,$EF,$8F,$A7,$91,$D7,$A6,$4F,$38,$CB,$D6
	.dc.b	$32,$F3,$8F,$CC,$FD,$8B,$EC,$7C,$DD,$92,$BF,$11,$39,$6C,$E5,$F2
	.dc.b	$4F,$28,$91,$A2,$1E,$9F,$B1,$FD,$3F,$6B,$DF,$BE,$F4,$AF,$CB,$4E
	.dc.b	$5B,$39,$7F,$CD,$31,$FE,$B8,$72,$70,$FA,$BE,$A4,$79,$FD,$7D,$78
	.dc.b	$C5,$E6,$74,$FF,$82,$2D,$7F,$BC,$F2,$7F,$C6,$D5,$AC,$3F,$C4,$95
	.dc.b	$F9,$36,$DE,$F5,$91,$EA,$8F,$79,$87,$FD,$03,$7D,$23,$FC,$2E,$D8
	.dc.b	$FF,$18,$35,$7D,$83,$EC,$7F,$87,$BF,$A1,$8F,$DE,$2E,$AC,$9A,$EB
	.dc.b	$93,$A8,$CB,$C2,$39,$F9,$D2,$39,$BD,$BA,$F1,$1F,$E9,$8B,$FF,$7E
	.dc.b	$BA,$3C,$31,$F1,$FD,$BF,$E4,$7B,$FE,$6A,$7F,$D3,$ED,$DF,$E5,$47
	.dc.b	$F9,$6E,$78,$3F,$C5,$67,$5A,$FF,$8E,$9B,$1F,$12,$76,$FC,$A6,$D7
	.dc.b	$49,$47,$F3,$4F,$AF,$7D,$8F,$F3,$6F,$B1,$FC,$46,$1F,$EB,$26,$E1
	.dc.b	$E4,$AD,$CF,$F2,$77,$FB,$1B,$FC,$A6,$5C,$7C,$CE,$DE,$77,$F4,$2F
	.dc.b	$B1,$7C,$CF,$F4,$4F,$B1,$FE,$39,$B7,$F9,$51,$B8,$FF,$9B,$6B,$6F
	.dc.b	$F3,$55,$F1,$5C,$7C,$49,$39,$8F,$E9,$FF,$63,$FC,$C7,$6F,$97,$83
	.dc.b	$C7,$F0,$1A,$F3,$37,$F3,$1E,$3F,$36,$65,$FD,$0F,$14,$FF,$5A,$F7
	.dc.b	$9F,$63,$FD,$73,$EC,$7E,$8D,$9B,$3A,$8F,$E9,$8B,$E3,$F8,$79,$32
	.dc.b	$73,$1F,$D9,$BE,$C5,$5D,$7D,$79,$5F,$D5,$67,$2D,$F8,$0F,$ED,$BE
	.dc.b	$C3,$E1,$69,$CE,$3A,$7C,$AE,$BE,$16,$5F,$38,$FE,$EB,$F6,$2F,$81
	.dc.b	$E7,$E5,$4A,$FE,$C9,$39,$7F,$B3,$7D,$8B,$E5,$3F,$BE,$FD,$77,$F4
	.dc.b	$29,$FF,$7C,$B7,$AF,$F3,$75,$B9,$FE,$DD,$F3,$AF,$B1,$FF,$02,$CB
	.dc.b	$E4,$29,$FE,$08,$9D,$57,$0F,$8A,$C5,$C8,$C1,$D3,$DA,$F1,$FA,$32
	.dc.b	$30,$7F,$7D,$FB,$17,$C5,$75,$A9,$39,$8F,$F0,$BC,$EA,$30,$43,$DB
	.dc.b	$F5,$5F,$E1,$CF,$B1,$FB,$91,$FE,$09,$DF,$83,$F0,$AF,$5D,$92,$35
	.dc.b	$1F,$20,$C2,$E0,$FE,$03,$7D,$23,$1E,$4F,$FA,$E2,$31,$3A,$BF,$A3
	.dc.b	$F6,$3E,$3E,$1F,$A7,$E9,$CC,$7F,$C3,$98,$7C,$D6,$CE,$5F,$AF,$3C
	.dc.b	$5F,$F1,$17,$D8,$FF,$92,$9F,$F1,$56,$2F,$AD,$FE,$55,$18,$6B,$E6
	.dc.b	$3F,$DC,$93,$C5,$C0,$71,$3F,$C7,$9B,$9F,$FB,$51,$B3,$CB,$B3,$FC
	.dc.b	$93,$83,$E4,$2F,$91,$2B,$FC,$71,$39,$6C,$F0,$23,$FE,$3D,$FB,$15
	.dc.b	$FC,$F8,$7D,$8A,$FF,$DC,$B5,$AC,$FF,$DF,$D3,$9A,$FF,$80,$15,$DD
	.dc.b	$E2,$FF,$BB,$0C,$46,$E1,$5C,$DD,$83,$37,$E3,$9E,$15,$C8,$E9,$87
	.dc.b	$FB,$E2,$3A,$5F,$36,$D7,$1A,$DD,$FA,$1A,$F8,$0E,$F8,$65,$F1,$5F
	.dc.b	$39,$DD,$F8,$B0,$F0,$47,$C4,$C3,$B5,$F0,$6B,$F2,$67,$6F,$C6,$6B
	.dc.b	$E0,$3C,$57,$F0,$DF,$F7,$23,$5E,$D8,$7F,$B8,$E3,$11,$E6,$F4,$5F
	.dc.b	$1D,$96,$C7,$FB,$B4,$F9,$EE,$1E,$AD,$75,$F1,$FF,$EE,$F7,$E9,$35
	.dc.b	$34,$8F,$C9,$BF,$F5,$6E,$DF,$26,$3E,$D1,$C0,$D7,$8B,$FE,$23,$D7
	.dc.b	$F0,$51,$D8,$CC,$F1,$71,$1F,$87,$AE,$B1,$F5,$CC,$EB,$FE,$B6,$BE
	.dc.b	$8F,$AA,$39,$9B,$3C,$55,$CE,$77,$1E,$63,$33,$CE,$FC,$51,$83,$FD
	.dc.b	$81,$3C,$1F,$6E,$1E,$FC,$3C,$B8,$75,$BF,$F5,$44,$7A,$0F,$F0,$CD
	.dc.b	$7A,$0E,$71,$AB,$FF,$23,$4F,$FA,$75,$B1,$EB,$AE,$95,$FC,$23,$EC
	.dc.b	$7F,$80,$4F,$FA,$D5,$B1,$93,$89,$FA,$B7,$37,$FE,$2A,$6C,$FD,$66
	.dc.b	$CE,$CE,$21,$CB,$79,$B3,$1F,$D2,$27,$F9,$1F,$D8,$FF,$A9,$1F,$24
	.dc.b	$E8,$F1,$DA,$5F,$3D,$9A,$E2,$F4,$F3,$E4,$7F,$2B,$EB,$B9,$7C,$19
	.dc.b	$B8,$7C,$8E,$65,$62,$FD,$95,$E0,$9F,$F6,$04,$7C,$78,$C7,$C0,$7B
	.dc.b	$BE,$C5,$E5,$7C,$0F,$96,$FE,$E1,$7F,$EE,$19,$CB,$FE,$E3,$8C,$E7
	.dc.b	$FC,$E9,$B9,$FF,$95,$9B,$3F,$A2,$AE,$35,$BE,$BF,$9B,$F6,$2F,$8C
	.dc.b	$A7,$86,$2D,$FF,$BC,$65,$F2,$BC,$0F,$AB,$F6,$3F,$EA,$39,$D4,$7D
	.dc.b	$8F,$B1,$7C,$95,$38,$4F,$FB,$EE,$32,$DF,$9C,$E0,$71,$E6,$B1,$FF
	.dc.b	$00,$7E,$C5,$F0,$5F,$84,$6C,$7F,$62,$F3,$7F,$C4,$CE,$E7,$F1,$2F
	.dc.b	$F9,$70,$FE,$79,$C5,$FE,$7A,$71,$C1,$FD,$96,$1F,$62,$FB,$1E,$3E
	.dc.b	$0F,$E4,$BE,$C7,$FA,$26,$1C,$8F,$E1,$AE,$30,$E5,$7F,$E3,$19,$D4
	.dc.b	$7F,$4B,$9D,$61,$FE,$8A,$5F,$01,$D4,$FC,$AF,$C9,$3D,$7C,$61,$CB
	.dc.b	$05,$FF,$98,$23,$0D,$8F,$55,$7D,$37,$FB,$EC,$67,$0F,$F4,$CF,$88
	.dc.b	$F5,$3F,$5A,$DC,$3F,$C1,$7F,$63,$FC,$1D,$FE,$1B,$3A,$F4,$24,$7F
	.dc.b	$84,$3E,$C5,$F5,$38,$CF,$FC,$17,$F6,$3F,$E6,$6A,$F4,$36,$3E,$77
	.dc.b	$D8,$FF,$7E,$9C,$FC,$17,$EF,$AB,$FD,$43,$1C,$3E,$43,$F6,$F4,$F8
	.dc.b	$0D,$46,$D9,$3F,$E2,$2F,$B1,$FC,$D3,$E0,$8F,$E1,$3E,$AF,$E1,$32
	.dc.b	$FF,$0F,$B7,$87,$F8,$FE,$08,$DF,$C0,$FF,$10,$FB,$1F,$C1,$63,$F4
	.dc.b	$5B,$87,$E1,$EF,$F7,$6F,$F8,$AA,$33,$87,$FB,$1B,$EC,$7F,$D4,$77
	.dc.b	$F6,$5B,$5F,$8F,$FB,$07,$D5,$FF,$1E,$9C,$DF,$1F,$8F,$C1,$FC,$62
	.dc.b	$B7,$87,$38,$CB,$FE,$42,$8F,$DE,$CF,$18,$C7,$9B,$FC,$83,$EC,$7E
	.dc.b	$2E,$5F,$9F,$6D,$7F,$40,$5F,$17,$BF,$36,$61,$FE,$4E,$9D,$BF,$AA
	.dc.b	$76,$E6,$3C,$DF,$0F,$E7,$E9,$B3,$E5,$6E,$3F,$05,$B8,$FF,$50,$CF
	.dc.b	$FD,$65,$5F,$9F,$FB,$17,$CD,$7F,$42,$6C,$78,$F6,$C4,$C8,$E0,$FB
	.dc.b	$E3,$18,$FF,$97,$AB,$8B,$FB,$FF,$AD,$FC,$15,$FF,$81,$A7,$38,$FF
	.dc.b	$62,$9C,$D7,$FA,$07,$1F,$D6,$4F,$9F,$E7,$FE,$A7,$85,$F5,$59,$8F
	.dc.b	$D3,$72,$D3,$E5,$79,$FF,$4F,$EC,$3E,$A3,$63,$27,$CF,$8C,$B4,$F8
	.dc.b	$9E,$D7,$C3,$3F,$09,$E0,$58,$C9,$33,$FC,$74,$65,$F8,$53,$FD,$8B
	.dc.b	$EC,$57,$9C,$FF,$C0,$31,$8C,$3F,$01,$F6,$3F,$89,$F3,$BF,$14,$D8
	.dc.b	$C9,$F8,$58,$CB,$4F,$CA,$79,$DF,$99,$3F,$35,$EB,$58,$C9,$33,$FD
	.dc.b	$24,$65,$FF,$4A,$FD,$8F,$E8,$1B,$DB,$F9,$BE,$FF,$F0,$2C,$E7,$46
	.dc.b	$1F,$E4,$08,$C3,$6B,$83,$FC,$AE,$33,$7F,$E7,$F6,$D3,$93,$FB,$DD
	.dc.b	$6E,$FF,$C1,$31,$8B,$FD,$D7,$C8,$FE,$D4,$FF,$06,$FA,$8F,$6B,$18
	.dc.b	$7E,$D6,$6D,$7F,$4F,$AD,$4E,$4F,$D7,$7D,$47,$ED,$CF,$DC,$7C,$75
	.dc.b	$8C,$92,$BF,$75,$19,$7F,$D7,$7F,$61,$FC,$BF,$4F,$87,$1F,$E9,$B8
	.dc.b	$C5,$7F,$A7,$7E,$1F,$FC,$59,$E2,$7F,$8B,$A7,$35,$BF,$37,$1F,$63
	.dc.b	$96,$E1,$FD,$4E,$BF,$81,$D7,$5F,$13,$97,$C5,$F3,$65,$6D,$CB,$4E
	.dc.b	$5C,$FE,$13,$FC,$8F,$9F,$C9,$F8,$8C,$4E,$8F,$06,$F8,$C7,$DE,$FF
	.dc.b	$91,$BD,$7F,$F2,$29,$3C,$DC,$7C,$5C,$BF,$F4,$5F,$D2,$FF,$A1,$2B
	.dc.b	$F8,$1F,$AE,$8C,$57,$C8,$8C,$63,$FE,$C3,$AD,$D7,$1F,$27,$4D,$C3
	.dc.b	$AD,$6F,$0C,$BC,$6A,$BF,$E0,$06,$DF,$48,$E1,$5C,$9D,$36,$3F,$AD
	.dc.b	$61,$E4,$9F,$F0,$4F,$A1,$CB,$DE,$E6,$CB,$E5,$F6,$3F,$E0,$BD,$39
	.dc.b	$4F,$8A,$31,$7F,$EA,$CD,$3F,$D4,$F1,$8C,$7B,$36,$3E,$45,$71,$F6
	.dc.b	$FF,$B4,$5F,$13,$FD,$D1,$5B,$6C,$6A,$FD,$19,$FA,$B0,$EF,$39,$6E
	.dc.b	$3E,$88,$C5,$FD,$1C,$7C,$13,$92,$47,$84,$F8,$79,$63,$1F,$4D,$F2
	.dc.b	$96,$5E,$57,$67,$17,$04,$8E,$19,$79,$E7,$D7,$87,$5C,$BC,$F1,$E1
	.dc.b	$9D,$3D,$AB,$C1,$7C,$EF,$E8,$E3,$C6,$B9,$9E,$1A,$EF,$97,$BA,$31
	.dc.b	$5B,$7C,$F9,$7B,$23,$18,$7A,$F1,$F0,$BD,$EB,$89,$E5,$BE,$97,$C6
	.dc.b	$32,$FD,$48,$CE,$5E,$C8,$D1,$9C,$BB,$FD,$8F,$F1,$DB,$FF,$7E,$60
	.dc.b	$8C,$47,$FB,$6E,$3F,$1A,$D8,$CB,$F8,$E7,$95,$7E,$39,$B7,$9E,$DF
	.dc.b	$CC,$F8,$7F,$ED,$E8,$C5,$FD,$7A,$DB,$F9,$06,$BF,$7A,$78,$39,$AF
	.dc.b	$D1,$CE,$B0,$F4,$47,$69,$CB,$1E,$CE,$5A,$4C,$3D,$11,$96,$C6,$70
	.dc.b	$F3,$C7,$EA,$B0,$F1,$CF,$E4,$B0,$F1,$12,$F9,$5F,$02,$4F,$A4,$FC
	.dc.b	$B5,$72,$C3,$D3,$1D,$CD,$61,$DD,$F5,$3E,$42,$57,$EF,$AB,$96,$1E
	.dc.b	$1A,$E4,$7C,$93,$C6,$74,$9F,$1B,$93,$C6,$76,$73,$7F,$EB,$68,$FE
	.dc.b	$35,$F6,$3F,$C5,$F0,$FE,$30,$BE,$6B,$E1,$B6,$77,$78,$23,$83,$9F
	.dc.b	$27,$03,$2F,$F0,$19,$ED,$28,$90,$9E,$11,$CC,$E1,$38,$79,$79,$7C
	.dc.b	$57,$77,$F6,$A5,$7F,$E6,$67,$FD,$68,$D7,$7B,$71,$8F,$F7,$3C,$6A
	.dc.b	$BA,$57,$39,$E0,$DF,$53,$37,$C6,$AD,$FA,$BC,$0B,$D2,$FA,$9E,$B3
	.dc.b	$E0,$11,$9A,$E3,$1E,$8A,$B9,$1A,$48,$FF,$67,$7D,$8F,$C1,$7A,$07
	.dc.b	$FD,$AE,$EB,$9B,$AC,$7F,$42,$8D,$BA,$6F,$36,$A6,$47,$68,$F2,$FD
	.dc.b	$03,$EC,$69,$27,$FD,$C1,$F6,$3F,$47,$83,$E9,$56,$E7,$FD,$C8,$D7
	.dc.b	$CE,$D8,$FB,$13,$C7,$7F,$B9,$2F,$8D,$F0,$92,$F9,$FC,$DD,$79,$C5
	.dc.b	$C3,$C3,$3F,$33,$EC,$5F,$4B,$EA,$F9,$79,$87,$FB,$DE,$72,$D9,$CB
	.dc.b	$E6,$9E,$51,$23,$44,$3D,$7F,$1D,$F6,$BE,$C5,$E2,$D2,$4F,$E7,$F2
	.dc.b	$FC,$B4,$FD,$D8,$C6,$BF,$9B,$8E,$31,$83,$F4,$CE,$1F,$BE,$D9,$E7
	.dc.b	$C7,$E1,$F1,$FC,$5D,$32,$F9,$0B,$4E,$E3,$31,$F9,$EC,$3C,$55,$99
	.dc.b	$FC,$6F,$D8,$BE,$C7,$DD,$D9,$27,$AC,$E5,$B3,$97,$CF,$3F,$9E,$9D
	.dc.b	$47,$49,$C1,$1F,$1F,$D8,$FE,$93,$B5,$EE,$5F,$76,$57,$B2,$72,$D9
	.dc.b	$CB,$FA,$48,$4F,$39,$E9,$F5,$72,$B9,$3D,$FE,$BF,$11,$C6,$2F,$0F
	.dc.b	$F0,$DC,$15,$FB,$8F,$2B,$FC,$7D,$39,$DF,$D0,$67,$69,$8F,$9E,$32
	.dc.b	$FD,$4A,$E3,$1F,$C6,$DB,$1F,$C9,$5A,$9B,$F6,$0F,$B1,$FE,$12,$FE
	.dc.b	$9E,$7E,$FA,$EA,$C9,$AE,$B9,$3A,$8C,$BC,$23,$9F,$9D,$23,$9B,$DB
	.dc.b	$AF,$11,$FC,$42,$FF,$E0,$0B,$A3,$C3,$1F,$17,$DB,$FE,$3B,$B7,$E7
	.dc.b	$E7,$F8,$BB,$76,$FE,$71,$A2,$B8,$57,$E6,$2B,$5A,$FF,$3F,$6C,$7C
	.dc.b	$59,$DB,$FC,$8A,$B9,$3F,$36,$3C,$27,$F2,$EF,$B0,$7D,$8F,$F3,$0F
	.dc.b	$B1,$FC,$2D,$7E,$99,$AF,$F2,$C9,$DB,$FA,$A7,$FB,$43,$FC,$C2,$1C
	.dc.b	$7C,$CE,$DE,$7F,$F3,$CF,$B1,$7C,$BF,$F3,$EF,$B1,$FE,$41,$B7,$F5
	.dc.b	$96,$E3,$FD,$8E,$B6,$FE,$E1,$7C,$57,$1F,$12,$4E,$63,$FA,$4F,$D8
	.dc.b	$FF,$3B,$E2,$FE,$C5,$83,$C7,$F0,$1A,$F2,$B7,$F4,$1E,$3F,$32,$65
	.dc.b	$FC,$1B,$14,$FF,$54,$F7,$5F,$63,$FD,$5B,$EC,$7E,$8D,$9B,$3A,$8F
	.dc.b	$E1,$EB,$E3,$F8,$79,$12,$73,$1F,$D8,$3E,$C5,$5D,$7D,$79,$5F,$C6
	.dc.b	$27,$2D,$F8,$0F,$EC,$FE,$C3,$E1,$69,$CE,$3A,$7C,$AE,$BE,$16,$5F
	.dc.b	$38,$FE,$DF,$F6,$2F,$81,$E7,$E1,$CA,$FE,$4F,$39,$7F,$94,$FD,$8B
	.dc.b	$E5,$3F,$BB,$FD,$7F,$F4,$89,$FE,$E2,$DE,$BF,$F0,$95,$6E,$7F,$98
	.dc.b	$FC,$EB,$EC,$7F,$BF,$CB,$E4,$29,$FE,$02,$9D,$57,$0F,$8A,$C5,$C8
	.dc.b	$C1,$D3,$DB,$F1,$FA,$32,$30,$7F,$40,$FB,$17,$C5,$75,$A9,$39,$8F
	.dc.b	$F0,$8C,$EA,$30,$43,$DB,$F5,$5F,$E1,$6F,$B0,$FB,$91,$FD,$33,$7E
	.dc.b	$0F,$C2,$BD,$76,$48,$D4,$7C,$83,$0F,$F8,$62,$B9,$B6,$73,$C1,$FE
	.dc.b	$3B,$B7,$D8,$F8,$78,$3E,$8F,$A7,$30,$FE,$BB,$8F,$CB,$6C,$E5,$FA
	.dc.b	$B3,$C5,$FE,$C7,$F6,$3F,$E4,$F7,$FB,$76,$2F,$AD,$FE,$51,$18,$6B
	.dc.b	$FE,$2E,$9E,$0D,$ED,$FF,$1C,$C6,$2B,$FC,$A4,$DF,$37,$FC,$AC,$BB
	.dc.b	$7F,$C7,$F8,$BE,$42,$F9,$13,$2F,$EE,$F3,$96,$CF,$02,$3F,$DF,$3E
	.dc.b	$C7,$A6,$5F,$E6,$BF,$B1,$5F,$F9,$96,$73,$49,$CB,$FE,$71,$9E,$CE
	.dc.b	$5F,$F3,$61,$88,$DC,$27,$9C,$61,$B3,$C2,$FF,$CE,$93,$C2,$39,$1E
	.dc.b	$3C,$7F,$CE,$F7,$DC,$FF,$41,$5F,$27,$C9,$7C,$AB,$A4,$7B,$5E,$58
	.dc.b	$58,$E3,$7D,$30,$F0,$5F,$19,$E6,$6C,$F5,$4E,$EF,$9C,$78,$CF,$93
	.dc.b	$3C,$5A,$71,$C3,$CC,$F2,$C3,$FC,$E4,$F8,$9A,$E4,$CC,$7D,$27,$FC
	.dc.b	$F0,$FF,$44,$F8,$ED,$B6,$37,$E9,$7F,$A1,$9B,$3D,$1F,$F5,$47,$C7
	.dc.b	$D4,$D2,$3B,$7E,$05,$FF,$AB,$76,$F8,$53,$C8,$F8,$11,$97,$E1,$19
	.dc.b	$D3,$F0,$31,$97,$F0,$71,$9C,$FF,$D4,$FB,$AF,$AC,$F9,$07,$C1,$F2
	.dc.b	$9D,$46,$2B,$E4,$E3,$FE,$C3,$AD,$5D,$9C,$D7,$9A,$F8,$E1,$E2,$8E
	.dc.b	$CF,$07,$D2,$62,$7C,$B5,$E4,$C3,$E1,$DF,$29,$CD,$7F,$AA,$B1,$F8
	.dc.b	$47,$FB,$2D,$F7,$1F,$E7,$99,$F4,$C7,$FA,$7D,$D4,$66,$FF,$03,$3F
	.dc.b	$2E,$3F,$D6,$CD,$7F,$DB,$1F,$63,$FC,$06,$7F,$CE,$AD,$8F,$F5,$DB
	.dc.b	$FE,$E5,$8C,$ED,$E0,$78,$BE,$09,$4F,$17,$FD,$88,$D7,$73,$F9,$48
	.dc.b	$FE,$86,$DE,$8E,$51,$89,$FC,$DB,$5E,$71,$E1,$7F,$DD,$FF,$62,$F9
	.dc.b	$EC,$47,$8A,$75,$8A,$B5,$87,$BB,$D0,$DE,$3F,$E0,$E8,$FE,$5D,$D9
	.dc.b	$89,$91,$E7,$7C,$A4,$70,$7F,$35,$F8,$1D,$E5,$FE,$ED,$C7,$F9,$D4
	.dc.b	$79,$3E,$57,$FB,$75,$7F,$A8,$9B,$5F,$EE,$8F,$B1,$79,$9B,$EC,$97
	.dc.b	$FB,$AB,$DD,$FF,$BB,$DA,$88,$E1,$C7,$D8,$FF,$AB,$A7,$F8,$1B,$63
	.dc.b	$F8,$3B,$5F,$F7,$D3,$5E,$81,$7C,$AF,$DF,$6C,$F4,$8C,$1F,$83,$FB
	.dc.b	$07,$C8,$D2,$4F,$F1,$48,$CB,$7E,$6F,$81,$1C,$1F,$C6,$BE,$C1,$F0
	.dc.b	$3F,$8F,$6C,$66,$30,$7E,$4D,$FF,$2F,$9F,$C8,$8B,$97,$F2,$66,$BE
	.dc.b	$E3,$F8,$47,$77,$E6,$FE,$BB,$F4,$11,$FC,$B1,$AE,$DA,$FC,$66,$BF
	.dc.b	$4D,$7B,$BC,$C9,$5F,$CD,$23,$2D,$E4,$F0,$11,$C1,$FC,$EB,$B5,$E7
	.dc.b	$FE,$BD,$B1,$98,$C1,$9F,$2A,$47,$16,$E3,$FD,$1B,$EB,$3F,$CB,$F3
	.dc.b	$FD,$19,$B1,$9F,$6B,$F7,$91,$88,$F8,$CF,$FA,$96,$78,$7C,$46,$88
	.dc.b	$F3,$F6,$38,$91,$C1,$FD,$53,$C9,$FF,$30,$37,$D8,$FE,$15,$18,$C3
	.dc.b	$FA,$DB,$5F,$AD,$F5,$4F,$B0,$FE,$20,$DB,$FE,$BE,$D9,$FE,$C6,$D7
	.dc.b	$EF,$7D,$5F,$FA,$6A,$7F,$51,$E0,$7F,$42,$94,$E5,$FD,$D7,$D4,$BE
	.dc.b	$93,$F8,$6C,$E6,$B8,$3B,$AE,$BF,$0F,$FC,$8E,$31,$1F,$AA,$6B,$FB
	.dc.b	$77,$96,$1E,$2C,$3F,$80,$BE,$43,$85,$7F,$1A,$8C,$D7,$F0,$DF,$AB
	.dc.b	$FD,$D6,$1F,$D0,$9B,$7E,$C5,$F0,$7F,$C8,$9C,$D2,$91,$86,$FC,$D6
	.dc.b	$EB,$83,$E1,$F7,$FC,$5C,$B7,$1F,$F2,$BE,$5F,$EA,$D6,$CE,$9A,$6A
	.dc.b	$7F,$9A,$FD,$5B,$E4,$E9,$E5,$D6,$D7,$C1,$D7,$F8,$E5,$6E,$3F,$C3
	.dc.b	$6D,$8F,$F0,$FB,$5F,$DD,$35,$FE,$32,$EF,$4F,$EC,$FF,$4D,$F8,$D7
	.dc.b	$F0,$11,$9A,$FC,$1C,$E5,$FA,$9F,$31,$F4,$4F,$A5,$F3,$36,$3E,$0C
	.dc.b	$5C,$3F,$0F,$37,$0F,$A9,$E3,$7D,$1F,$AA,$FA,$4D,$C3,$FC,$FF,$C5
	.dc.b	$FC,$AB,$0F,$B9,$5A,$9C,$BF,$23,$D2,$91,$F4,$3E,$AE,$D7,$DB,$8F
	.dc.b	$7C,$FF,$73,$D7,$FA,$FC,$FC,$68,$C1,$27,$FD,$1F,$F5,$6C,$39,$E7
	.dc.b	$FE,$0C,$8C,$4F,$9B,$EA,$FF,$15,$F0,$FF,$8B,$6C,$64,$F5,$C6,$5A
	.dc.b	$7E,$1B,$C8,$FC,$D1,$F9,$BF,$3E,$C6,$49,$9F,$EC,$23,$2F,$B4,$FE
	.dc.b	$BD,$F5,$7F,$B3,$9F,$C3,$42,$30,$FE,$97,$C2,$FF,$01,$C6,$33,$FD
	.dc.b	$F3,$A9,$FE,$80,$D7,$FA,$FB,$5E,$FD,$BF,$E1,$D8,$E5,$7F,$86,$F5
	.dc.b	$3A,$CF,$F8,$3B,$99,$1F,$B7,$F8,$C9,$5F,$EB,$CF,$39,$E9,$63,$2F
	.dc.b	$E1,$B5,$73,$FF,$1D,$D6,$AF,$F5,$B3,$FB,$9F,$A2,$91,$F8,$8F,$88
	.dc.b	$B8,$7F,$03,$9B,$87,$F0,$39,$B5,$FD,$EB,$EA,$BF,$9D,$5D,$F2,$FF
	.dc.b	$C7,$11,$88,$FF,$05,$35,$3F,$D1,$2F,$F9,$32,$FA,$BF,$F1,$7B,$9B
	.dc.b	$ED,$18,$F1,$24,$7D,$69,$FF,$68,$36,$33,$7F,$E9,$89,$FF,$08,$3E
	.dc.b	$28,$F2,$9E,$6E,$59,$5F,$E9,$CA,$46,$2B,$93,$E4,$CF,$CA,$DA,$FC
	.dc.b	$3F,$A3,$FE,$53,$AF,$F1,$43,$5F,$EE,$FE,$EF,$F9,$6E,$30,$7F,$98
	.dc.b	$FD,$07,$D0,$7F,$89,$23,$F8,$94,$62,$B9,$D7,$FA,$32,$31,$1F,$E0
	.dc.b	$2F,$17,$17,$C5,$B1,$E2,$6B,$E2,$C3,$FD,$5B,$3B,$6C,$6D,$FF,$08
	.dc.b	$E1,$FC,$8D,$FF,$0D,$36,$38,$1F,$E8,$97,$FB,$64,$6B,$CE,$FF,$50
	.dc.b	$B4,$E6,$E2,$7D,$CF,$93,$EA,$9F,$22,$F4,$3C,$75,$FE,$1E,$AD,$4E
	.dc.b	$63,$FD,$87,$8F,$F9,$2A,$F8,$DF,$06,$CE,$8F,$A7,$7C,$A7,$57,$CF
	.dc.b	$3F,$97,$7C,$1F,$9B,$97,$F8,$1D,$CC,$EF,$1F,$03,$71,$FF,$33,$5F
	.dc.b	$19,$F6,$E5,$DB,$19,$5E,$31,$5B,$7C,$B3,$AC,$3B,$D5,$BE,$75,$C6
	.dc.b	$F8,$3C,$B0,$F1,$36,$BE,$4E,$3E,$9B,$F3,$C7,$39,$F3,$D7,$0A,$D4
	.dc.b	$62,$74,$F3,$3D,$11,$A7,$18,$79,$9F,$25,$73,$CB,$E7,$46,$2F,$9D
	.dc.b	$F1,$C7,$D2,$4B,$E1,$97,$82,$31,$5E,$AC,$39,$E3,$DF,$1F,$F3,$65
	.dc.b	$7D,$E6,$E1,$F8,$56,$DF,$F9,$C5,$FA,$BF,$57,$FC,$99,$C5,$23,$11
	.dc.b	$C0,$CB,$87,$F1,$CF,$1B,$FC,$3B,$73,$CF,$47,$E4,$7B,$FF,$D3,$B5
	.dc.b	$A8,$FC,$A5,$6E,$FC,$71,$B7,$2C,$7E,$F4,$6D,$FC,$3D,$74,$BE,$F1
	.dc.b	$AA,$E1,$7F,$9A,$AD,$4F,$63,$44,$8D,$3C,$2B,$9C,$79,$E3,$F1,$D3
	.dc.b	$C9,$FC,$BD,$F8,$F0,$F9,$18,$71,$9F,$D0,$D6,$E7,$31,$DF,$0E,$A7
	.dc.b	$ED,$E7,$57,$CC,$97,$F0,$EB,$81,$E9,$C3,$C8,$7A,$1C,$9E,$48,$F1
	.dc.b	$3F,$18,$B8,$78,$A7,$8E,$1E,$19,$ED,$18,$7B,$46,$5F,$82,$7F,$AC
	.dc.b	$A7,$FD,$8A,$DC,$3F,$DB,$3A,$7E,$49,$B1,$FE,$DC,$FA,$BF,$E6,$B3
	.dc.b	$FE,$C1,$6C,$66,$F6,$FF,$BB,$63,$2D,$8F,$A1,$1F,$CF,$B0,$FE,$7E
	.dc.b	$D8,$ED,$F0,$7A,$36,$CA,$FE,$53,$29,$D4,$7E,$8C,$46,$5C,$35,$FD
	.dc.b	$43,$5F,$57,$93,$FE,$F8,$E8,$FF,$4E,$4F,$FA,$0A,$7F,$C4,$ED,$7F
	.dc.b	$C5,$87,$69,$E1,$8F,$C5,$7C,$51,$93,$C2,$E2,$4D,$EA,$93,$E6,$6B
	.dc.b	$E6,$7A,$BF,$2D,$AF,$B9,$BF,$57,$FE,$F3,$C3,$FD,$68,$FF,$BA,$A7
	.dc.b	$FC,$7E,$FF,$09,$8C,$F0,$70,$E3,$95,$F2,$6B,$83,$FC,$35,$AF,$FB
	.dc.b	$FD,$A8,$9F,$34,$FA,$BF,$2B,$63,$FC,$9E,$D7,$E4,$FD,$5F,$FB,$27
	.dc.b	$1F,$F6,$54,$E5,$B5,$C8,$F7,$CE,$F5,$F0,$5A,$DF,$09,$E0,$24,$78
	.dc.b	$3E,$52,$BF,$62,$76,$F9,$27,$CF,$3F,$33,$EA,$DF,$3F,$EA,$F8,$B9
	.dc.b	$7F,$E6,$89,$CB,$67,$EF,$C6,$23,$F9,$A3,$5F,$38,$70,$FA,$9E,$1A
	.dc.b	$7F,$BA,$DB,$96,$6B,$E5,$FC,$2F,$E0,$3E,$AD,$E0,$D2,$63,$FE,$7C
	.dc.b	$87,$1E,$23,$F2,$F1,$C1,$D4,$F3,$9E,$9E,$1F,$5F,$6E,$57,$38,$CB
	.dc.b	$D6,$32,$F4,$8F,$C9,$FD,$5B,$EA,$FC,$DD,$72,$BF,$D1,$F3,$96,$CF
	.dc.b	$E9,$A3,$11,$FD,$6D,$AF,$C8,$6B,$F8,$07,$7B,$B6,$C5,$78,$8F,$8D
	.dc.b	$F2,$1F,$A6,$FA,$B7,$B7,$7D,$B9,$97,$FA,$8A,$72,$D9,$CB,$FB,$E8
	.dc.b	$4F,$39,$E9,$F4,$B2,$B9,$3D,$FE,$A7,$11,$C6,$2E,$7E,$78,$FD,$46
	.dc.b	$0A,$FD,$DF,$8F,$FE,$60,$AD,$52,$BF,$40,$F0,$D3,$7B,$48,$FE,$41
	.dc.b	$7F,$E6,$87,$F1,$ED,$7F,$4E,$D7,$F7,$0F,$79,$CC,$F3,$9F,$E3,$BF
	.dc.b	$54,$FA,$BF,$E1,$EF,$EA,$23,$3D,$58,$33,$D7,$23,$FD,$9B,$29,$D4
	.dc.b	$62,$79,$F9,$72,$36,$F6,$EC,$C4,$78,$A2,$EB,$F1,$30,$57,$C8,$F5
	.dc.b	$FF,$91,$EF,$FA,$0C,$3F,$DB,$AD,$7F,$9B,$EC,$9E,$11,$F2,$1F,$E6
	.dc.b	$F1,$98,$C1,$0E,$31,$C1,$FE,$83,$3B,$7F,$A1,$FD,$53,$EA,$FF,$A0
	.dc.b	$B6,$3F,$AA,$35,$FF,$1C,$35,$FE,$0A,$FF,$72,$73,$E5,$72,$23,$FD
	.dc.b	$77,$EA,$9F,$1F,$FD,$39,$B1,$98,$C1,$9F,$84,$91,$C5,$AF,$F9,$27
	.dc.b	$EA,$7F,$9D,$BF,$CE,$DA,$9A,$4F,$8F,$FE,$BA,$9A,$48,$FE,$E5,$EC
	.dc.b	$3E,$A7,$FB,$1B,$63,$31,$83,$3F,$11,$23,$8B,$5F,$F2,$F7,$D3,$AB
	.dc.b	$FB,$6B,$63,$FC,$00,$D7,$71,$AF,$9E,$91,$A6,$BD,$3E,$43,$99,$1E
	.dc.b	$6F,$F7,$AF,$A7,$7B,$54,$93,$FE,$0F,$8C,$BF,$E1,$1F,$A7,$7C,$77
	.dc.b	$F8,$09,$B1,$98,$C1,$FE,$06,$F9,$47,$D3,$D2,$4F,$F8,$6E,$32,$DF
	.dc.b	$80,$E0,$47,$06,$FC,$DE,$84,$70,$7F,$89,$7E,$9D,$F3,$FF,$E1,$66
	.dc.b	$C6,$63,$04,$7F,$BF,$78,$1F,$C5,$27,$71,$FE,$22,$6C,$7F,$18,$FA
	.dc.b	$77,$C4,$7F,$85,$BE,$99,$CB,$FE,$18,$9D,$F9,$3F,$DE,$9B,$8F,$F8
	.dc.b	$2E,$B6,$F8,$27,$FB,$FC,$62,$3F,$D4,$0D,$7F,$B6,$B5,$DD,$7F,$8C
	.dc.b	$A3,$86,$1F,$CB,$1B,$5A,$AE,$C7,$F8,$6E,$DF,$4F,$E9,$F8,$EF,$47
	.dc.b	$8B,$2B,$63,$04,$EE,$3F,$CA,$51,$86,$C7,$8C,$FF,$22,$FD,$37,$F9
	.dc.b	$46,$3F,$C3,$6D,$7F,$C5,$8D,$4D,$3F,$CB,$59,$7F,$46,$6E,$5C,$BE
	.dc.b	$0F,$FC,$8B,$0E,$9F,$F2,$3E,$4F,$88,$FF,$2A,$B7,$1F,$F3,$43,$5F
	.dc.b	$F6,$63,$5F,$F0,$CB,$9C,$7F,$AA,$B7,$1F,$F1,$9B,$67,$FC,$89,$F4
	.dc.b	$DE,$1A,$FF,$23,$7D,$2F,$F3,$89,$CE,$29,$CD,$F2,$C7,$FC,$7B,$7B
	.dc.b	$BF,$E7,$F1,$89,$FF,$20,$E1,$D6,$55,$B8,$C3,$C6,$7D,$8D,$AE,$55
	.dc.b	$E2,$BF,$0B,$47,$AB,$B7,$FC,$BC,$D7,$37,$E2,$C7,$D5,$18,$39,$61
	.dc.b	$B2,$4E,$EF,$BC,$B1,$F3,$CF,$C1,$8C,$BE,$83,$83,$EE,$6B,$FE,$4C
	.dc.b	$8F,$84,$DB,$F1,$B6,$BF,$D4,$93,$FD,$23,$E9,$AF,$8A,$69,$2B,$FD
	.dc.b	$33,$F1,$1F,$E9,$A4,$D2,$3A,$3C,$87,$F0,$6D,$FF,$C3,$D9,$FD,$98
	.dc.b	$FF,$43,$4F,$C0,$C7,$FD,$71,$B7,$C2,$9C,$9C,$8F,$C2,$5F,$2D,$3F
	.dc.b	$CD,$31,$E2,$35,$5C,$9E,$07,$2D,$7F,$11,$19,$7C,$51,$C5,$CD,$7F
	.dc.b	$8C,$DC,$E1,$E5,$8F,$04,$FF,$9E,$23,$18,$79,$E3,$C5,$8F,$28,$C4
	.dc.b	$EA,$7C,$B5,$FE,$3F,$C3,$FD,$3C,$DA,$F5,$D6,$E7,$31,$FE,$D9,$7D
	.dc.b	$95,$FE,$BB,$AF,$F2,$1B,$FE,$D5,$6B,$FE,$B1,$FA,$5F,$DE,$57,$F1
	.dc.b	$46,$C7,$FB,$59,$FF,$5B,$46,$77,$F1,$4F,$17,$9A,$9E,$2F,$FA,$61
	.dc.b	$AF,$FA,$C1,$AF,$87,$A3,$8C,$7F,$4C,$AF,$F7,$93,$67,$EC,$B5,$E9
	.dc.b	$3E,$2C,$1F,$4C,$F9,$8F,$2D,$78,$EB,$53,$9C,$3D,$DE,$3E,$B1,$FF
	.dc.b	$09,$47,$F3,$AC,$5F,$4F,$FD,$4A,$36,$4C,$7C,$77,$E8,$24,$7A,$1F
	.dc.b	$24,$FF,$B3,$23,$85,$7D,$B8,$C5,$7F,$19,$9F,$F6,$9F,$C7,$7F,$42
	.dc.b	$7F,$A2,$B6,$34,$7F,$B9,$FE,$9D,$F4,$FE,$E9,$D6,$1F,$E8,$07,$85
	.dc.b	$7F,$B2,$1A,$F6,$E9,$FA,$F1,$88,$FF,$79,$7D,$3B,$E0,$BE,$DB,$72
	.dc.b	$FD,$4C,$A3,$0F,$EB,$1A,$FF,$B5,$97,$C7,$F2,$24,$F2,$8C,$BD,$3E
	.dc.b	$9D,$F1,$1F,$8A,$6C,$66,$30,$67,$E5,$E4,$71,$6E,$1F,$BE,$FA,$77
	.dc.b	$AD,$F9,$46,$D7,$A9,$B3,$B9,$FC,$CC,$FF,$99,$0F,$79,$63,$F8,$4B
	.dc.b	$5F,$F7,$9B,$51,$3E,$9F,$F4,$EF,$F1,$06,$BF,$EF,$A6,$BC,$03,$87
	.dc.b	$97,$C9,$7E,$B2,$0D,$24,$FD,$48,$FC,$CC,$F7,$E4,$E0,$49,$FB,$3D
	.dc.b	$6F,$33,$F6,$ED,$8C,$C6,$0C,$F9,$12,$38,$B6,$FF,$95,$FD,$3F,$F9
	.dc.b	$AE,$7F,$0A,$D8,$CF,$A9,$FC,$12,$31,$1E,$F0,$FF,$AA,$E7,$8F,$BB
	.dc.b	$B2,$3D,$3D,$5E,$44,$70,$7E,$53,$E5,$3F,$CD,$AD,$F5,$BF,$89,$C6
	.dc.b	$30,$FE,$82,$D7,$C5,$F4,$CF,$A7,$FE,$32,$DB,$FC,$FB,$67,$FA,$4B
	.dc.b	$5F,$57,$D3,$BD,$DF,$E1,$ED,$AF,$E1,$5F,$4E,$FA,$0C,$D4,$9F,$D5
	.dc.b	$D6,$EF,$FA,$AB,$5E,$B3,$FC,$B2,$35,$E0,$7F,$43,$C3,$F8,$0D,$FF
	.dc.b	$92,$A7,$2F,$F9,$BA,$33,$7F,$C2,$29,$F4,$FF,$B8,$7F,$A6,$36,$BF
	.dc.b	$CE,$AB,$E4,$BF,$96,$39,$AF,$F1,$73,$6B,$9F,$81,$AA,$DB,$FC,$CF
	.dc.b	$0F,$F4,$CE,$1F,$0B,$93,$1E,$E6,$5B,$19,$3E,$3C,$65,$FF,$32,$FD
	.dc.b	$33,$E2,$7F,$9D,$B6,$3F,$9B,$67,$F0,$34,$FD,$8C,$E5,$FE,$4E,$1C
	.dc.b	$3F,$91,$6B,$FB,$16,$BC,$AB,$FA,$DC,$73,$D3,$FB,$97,$D1,$7E,$2D
	.dc.b	$61,$C6,$BF,$A9,$4E,$63,$E9,$7C,$8F,$C3,$3E,$27,$C9,$58,$FC,$2C
	.dc.b	$5C,$7E,$B6,$AA,$C6,$3F,$5B,$C6,$FA,$7F,$4D,$F5,$1B,$8F,$C2,$6E
	.dc.b	$1F,$7F,$5F,$E5,$AD,$AF,$93,$E8,$4B,$FC,$AF,$D3,$DA,$F8,$FB,$FF
	.dc.b	$65,$7F,$C4,$ED,$7E,$45,$7C,$42,$57,$F6,$7F,$A7,$61,$CF,$3F,$F0
	.dc.b	$C4,$62,$7E,$BF,$D3,$FE,$33,$E1,$3F,$1A,$D8,$C9,$F7,$23,$2D,$3F
	.dc.b	$57,$E3,$FE,$34,$FC,$77,$9D,$63,$24,$CF,$E5,$C6,$5F,$C1,$FD,$3F
	.dc.b	$EC,$E7,$FD,$62,$D8,$F9,$1E,$17,$F8,$22,$31,$9F,$DE,$9F,$E7,$B1
	.dc.b	$B7,$F8,$D3,$5F,$E7,$4D,$77,$DB,$FE,$24,$8E,$77,$FE,$E2,$E6,$FF
	.dc.b	$06,$57,$92,$BF,$13,$F0,$7F,$A9,$FA,$79,$7F,$E1,$3F,$35,$E8,$E3
	.dc.b	$0F,$C3,$D5,$CF,$FC,$8D,$5A,$BF,$E2,$33,$FB,$AF,$9B,$91,$FA,$8F
	.dc.b	$87,$B8,$7E,$1E,$6E,$1F,$C1,$26,$D7,$F5,$5F,$A6,$FE,$73,$77,$CB
	.dc.b	$FF,$1F,$C6,$23,$FB,$0B,$53,$FD,$0E,$FE,$54,$63,$E9,$FF,$C6,$B3
	.dc.b	$9B,$ED,$18,$F0,$E4,$7F,$B9,$A7,$FD,$A4,$D8,$CD,$7F,$9F,$27,$FC
	.dc.b	$22,$F9,$E3,$CA,$79,$B9,$65,$7F,$86,$A9,$18,$AE,$4F,$93,$3F,$2B
	.dc.b	$6B,$FD,$EF,$E8,$FF,$96,$2B,$FB,$9F,$B9,$FE,$60,$8C,$11,$FC,$47
	.dc.b	$A0,$F9,$9F,$EF,$D3,$FC,$56,$31,$5F,$E0,$88,$FF,$37,$C6,$23,$FB
	.dc.b	$FF,$8B,$CA,$B8,$B6,$38,$35,$F1,$E1,$FE,$A6,$9D,$B6,$36,$FF,$84
	.dc.b	$2F,$F9,$33,$FD,$CD,$B1,$C0,$FF,$43,$BF,$DA,$63,$5E,$77,$E8,$DA
	.dc.b	$73,$71,$3E,$E7,$C9,$F4,$CF,$96,$7A,$1E,$3A,$FF,$0D,$D6,$A7,$31
	.dc.b	$FE,$C6,$C7,$FC,$49,$7C,$6F,$83,$67,$47,$C9,$BE,$53,$AB,$E7,$9F
	.dc.b	$CB,$BE,$0F,$CD,$CB,$FC,$08,$E6,$77,$8F,$81,$B8,$FF,$8F,$2F,$8C
	.dc.b	$FB,$72,$ED,$8C,$AF,$18,$AD,$BE,$59,$D6,$1D,$EA,$DF,$3A,$E3,$7C
	.dc.b	$1E,$58,$78,$9B,$5F,$27,$1F,$4D,$F9,$E3,$9C,$F9,$EB,$85,$6A,$31
	.dc.b	$3A,$79,$9E,$88,$D3,$8C,$3C,$CF,$92,$B9,$E5,$F3,$A3,$17,$CE,$F8
	.dc.b	$E3,$E9,$25,$F0,$CB,$C1,$18,$AF,$56,$1C,$F1,$FA,$DF,$4F,$FC,$7B
	.dc.b	$2F,$F7,$BD,$A3,$11,$FE,$49,$8F,$C1,$36,$32,$FE,$21,$AF,$D9,$6B
	.dc.b	$F5,$5B,$1D,$7A,$3F,$D3,$AD,$9F,$A7,$3E,$3A,$F2,$D7,$E4,$BB,$FF
	.dc.b	$CF,$11,$8A,$DC,$FE,$5A,$B8,$6B,$F9,$AB,$DB,$E5,$9D,$12,$BF,$03
	.dc.b	$3A,$7D,$11,$9B,$EC,$7E,$6E,$75,$3E,$B9,$D5,$75,$8E,$66,$0E,$67
	.dc.b	$98,$C9,$6B,$AC,$71,$7B,$47,$13,$B1,$B8,$F6,$39,$C3,$E4,$4F,$28
	.dc.b	$FD,$45,$F9,$A7,$35,$D4,$93,$A3,$96,$1F,$21,$F3,$DF,$89,$F0,$18
	.dc.b	$9E,$2F,$EB,$EF,$B1,$DE,$B9,$47,$A2,$B9,$1C,$F3,$FC,$55,$7F,$AC
	.dc.b	$5B,$3F,$EB,$4C,$3F,$D9,$2F,$FA,$DB,$E9,$FF,$9A,$4F,$FA,$65,$B1
	.dc.b	$FE,$B9,$6B,$FE,$C2,$6A,$7F,$9E,$5F,$F3,$D6,$E3,$C4,$FB,$6B,$E0
	.dc.b	$A9,$88,$95,$FC,$A6,$74,$D9,$FF,$7A,$BD,$DB,$1A,$8F,$F5,$81,$93
	.dc.b	$FD,$F6,$FF,$BD,$9F,$F6,$CF,$3F,$F7,$68,$FF,$41,$46,$30,$F7,$79
	.dc.b	$FF,$ED,$7A,$D7,$95,$67,$E1,$3E,$02,$57,$67,$C7,$D7,$23,$C2,$D7
	.dc.b	$CB,$82,$3A,$47,$CB,$6B,$F2,$DB,$F4,$FF,$EC,$BC,$3F,$C1,$AF,$FB
	.dc.b	$9A,$7D,$07,$FB,$D1,$FE,$9A,$FF,$A1,$1B,$1C,$39,$39,$9F,$2A,$39
	.dc.b	$4F,$C8,$6B,$FE,$D9,$6B,$FE,$F9,$FA,$57,$D3,$F9,$9B,$1E,$20,$E1
	.dc.b	$FF,$4A,$FD,$3F,$F8,$93,$0F,$F1,$2C,$E5,$B5,$CC,$F8,$33,$BD,$3C
	.dc.b	$1F,$3D,$DC,$98,$77,$C1,$F3,$F8,$8F,$F5,$1C,$27,$18,$F8,$2B,$E6
	.dc.b	$FD,$3B,$E9,$FD,$5F,$1B,$31,$F5,$4E,$5B,$3F,$90,$8C,$47,$E3,$9A
	.dc.b	$FF,$BF,$DA,$88,$F0,$BF,$F2,$2B,$72,$D4,$A3,$0F,$1C,$9F,$07,$F8
	.dc.b	$4F,$A7,$78,$54,$95,$F4,$63,$F1,$8D,$8C,$BF,$EC,$68,$E2,$EF,$8B
	.dc.b	$A7,$89,$D7,$9A,$4F,$38,$CB,$D6,$32,$F3,$8F,$CB,$FD,$3B,$E9,$FC
	.dc.b	$DD,$52,$BF,$09,$39,$6C,$FE,$EA,$31,$1F,$B8,$6B,$F1,$03,$87,$E3
	.dc.b	$BC,$37,$FF,$37,$B7,$7C,$D7,$C7,$F9,$0F,$D5,$FD,$33,$DC,$BE,$DC
	.dc.b	$CB,$F3,$13,$96,$CE,$5F,$F7,$6C,$7F,$B5,$DC,$9C,$3E,$9F,$A9,$1E
	.dc.b	$7F,$4B,$5E,$31,$74,$FF,$42,$D2,$BF,$7F,$E4,$7F,$98,$E7,$3A,$7A
	.dc.b	$0C,$ED,$31,$FE,$3A,$7F,$93,$DF,$E4,$75,$FE,$6B,$8F,$90,$D7,$F1
	.dc.b	$AF,$59,$B3,$CE,$55,$FE,$98,$FA,$67,$D3,$FF,$12,$7F,$55,$3F,$BD
	.dc.b	$7F,$5E,$DE,$BD,$76,$48,$C1,$0E,$B8,$FE,$7A,$B9,$C6,$0F,$D3,$2F
	.dc.b	$2F,$C0,$48,$F0,$76,$D7,$80,$9E,$0F,$96,$7D,$13,$F3,$3D,$7F,$E5
	.dc.b	$7B,$7E,$92,$B8,$B7,$4F,$F5,$BE,$CA,$E1,$5F,$3A,$30,$D8,$F9,$E4
	.dc.b	$7F,$90,$46,$9F,$2B,$67,$FD,$87,$57,$2F,$DF,$FD,$33,$E9,$FF,$AD
	.dc.b	$B7,$0F,$BA,$D7,$FA,$DB,$5F,$D2,$3F,$DE,$1C,$F9,$5C,$88,$FE,$1B
	.dc.b	$E9,$DF,$35,$FD,$A1,$B1,$98,$C1,$9F,$88,$91,$C5,$AF,$F6,$CF,$A7
	.dc.b	$FE,$86,$FF,$1D,$6A,$69,$3E,$3F,$FB,$BA,$69,$23,$F3,$7E,$C3,$E9
	.dc.b	$FF,$BE,$B6,$33,$18,$33,$F1,$52,$38,$B5,$FF,$00,$FD,$3A,$BF,$C0
	.dc.b	$ED,$8F,$D4,$B5,$DC,$6B,$E7,$E4,$69,$AF,$4F,$90,$E6,$47,$9B,$FE
	.dc.b	$14,$FA,$77,$B3,$49,$3F,$B8,$8C,$BF,$BA,$FA,$77,$C7,$FF,$87,$1B
	.dc.b	$19,$8C,$1F,$E2,$0F,$94,$7D,$3D,$24,$FF,$04,$8C,$B7,$E0,$38,$11
	.dc.b	$C1,$BF,$37,$A1,$1C,$1F,$C3,$BE,$9D,$F4,$1F,$E3,$06,$C6,$63,$04
	.dc.b	$7F,$C0,$51,$A7,$F6,$ED,$ED,$76,$7F,$60,$FA,$77,$C4,$7F,$88,$3E
	.dc.b	$99,$CD,$FE,$22,$9E,$15,$BC,$B5,$33,$2F,$E4,$77,$C1,$B1,$FE,$46
	.dc.b	$8C,$47,$F9,$11,$B8,$F8,$A3,$FB,$B4,$62,$FF,$B9,$B6,$3F,$BD,$60
	.dc.b	$FA,$7F,$37,$C7,$79,$7C,$59,$5F,$E2,$1B,$4E,$A3,$CA,$F8,$C8,$FF
	.dc.b	$3A,$FA,$7F,$F2,$C3,$FD,$7D,$AF,$F8,$11,$AF,$F8,$2D,$A9,$FF,$30
	.dc.b	$65,$EB,$8C,$7C,$17,$F9,$2A,$73,$9F,$3E,$9F,$F2,$6E,$4F,$88,$FF
	.dc.b	$35,$B7,$0F,$E9,$AD,$7F,$CD,$6D,$7F,$BA,$39,$C7,$C0,$DB,$FF,$0D
	.dc.b	$B6,$7F,$C4,$5F,$4E,$C3,$FA,$7C,$3E,$99,$5F,$E2,$2A,$D6,$1F,$E2
	.dc.b	$99,$CD,$7F,$8A,$EB,$C9,$18,$8E,$38,$FF,$8A,$A3,$17,$CA,$7B,$57
	.dc.b	$5C,$3C,$35,$C6,$39,$9E,$8B,$E4,$EF,$0F,$13,$6D,$EA,$F1,$AF,$F1
	.dc.b	$CB,$63,$FC,$80,$F8,$B0,$F7,$B6,$F9,$CF,$AE,$BB,$B4,$C8,$C7,$D5
	.dc.b	$8F,$18,$F5,$9B,$29,$F0,$8E,$31,$C3,$1F,$44,$7F,$8A,$E3,$E2,$37
	.dc.b	$1F,$A6,$DC,$3F,$C1,$B3,$FD,$23,$E9,$EF,$8A,$69,$2B,$FD,$67,$F0
	.dc.b	$FF,$EB,$44,$D2,$3A,$3C,$F7,$FE,$C1,$D7,$DF,$1F,$E6,$79,$F8,$33
	.dc.b	$98,$F6,$99,$38,$EF,$FB,$9D,$3F,$05,$19,$7F,$09,$3C,$5C,$C7,$90
	.dc.b	$E3,$7E,$07,$11,$C2,$BA,$69,$FE,$CA,$C3,$C7,$18,$AF,$F2,$A4,$62
	.dc.b	$F9,$36,$3C,$E7,$4A,$FF,$65,$57,$18,$FF,$17,$BF,$C4,$70,$FF,$52
	.dc.b	$36,$FC,$13,$E6,$3F,$C3,$F3,$F0,$CF,$F5,$57,$C6,$EE,$3F,$DB,$75
	.dc.b	$5F,$F1,$DC,$5F,$A6,$FD,$F5,$FF,$AD,$5B,$3F,$E2,$A7,$FD,$2D,$19
	.dc.b	$DF,$81,$D7,$07,$C8,$78,$67,$FD,$2F,$5C,$A7,$D5,$39,$BF,$F5,$04
	.dc.b	$FF,$48,$AF,$F6,$AD,$3E,$9D,$F3,$1E,$5E,$AE,$D1,$8A,$F3,$46,$27
	.dc.b	$51,$CA,$75,$87,$A7,$C9,$F7,$1E,$7F,$A7,$FE,$97,$5E,$F8,$CB,$E2
	.dc.b	$8F,$F6,$23,$93,$1E,$B6,$AB,$FC,$A9,$92,$B7,$1D,$EB,$67,$FB,$7F
	.dc.b	$E4,$3F,$9F,$B7,$AB,$E9,$7D,$3B,$E9,$FE,$A5,$6A,$FF,$CB,$2D,$C7
	.dc.b	$F9,$C7,$D3,$BE,$13,$78,$FD,$B8,$C6,$5F,$EE,$D6,$E1,$FE,$62,$CB
	.dc.b	$FC,$13,$39,$BF,$0F,$93,$C7,$E4,$7F,$09,$F4,$FF,$EC,$AA,$E5,$18
	.dc.b	$F8,$A9,$1F,$EF,$99,$FB,$18,$FF,$9B,$9E,$0D,$E4,$E7,$CD,$F8,$FF
	.dc.b	$1B,$FC,$27,$7C,$63,$1E,$67,$E4,$FE,$9D,$EC,$7E,$2F,$CD,$FA,$B9
	.dc.b	$7F,$9E,$A7,$38,$7E,$1A,$73,$C7,$FD,$6E,$3F,$CB,$27,$E2,$BE,$C3
	.dc.b	$FC,$8C,$FF,$BB,$63,$5E,$27,$E5,$E7,$38,$7F,$A1,$BE,$A3,$F3,$AF
	.dc.b	$DC,$8F,$F0,$6D,$F1,$3F,$DE,$33,$AB,$FD,$A4,$EA,$B9,$7C,$A7,$E8
	.dc.b	$FE,$25,$F2,$FF,$AF,$C1,$39,$8C,$77,$6E,$AE,$5F,$A2,$47,$ED,$FC
	.dc.b	$67,$A7,$FA,$FF,$B0,$78,$D9,$8C,$3B,$E5,$FD,$A7,$D8,$7C,$8D,$BF
	.dc.b	$D4,$78,$79,$F8,$3F,$86,$4E,$5F,$F0,$27,$D5,$7F,$98,$F4,$FD,$9F
	.dc.b	$D8,$7C,$CC,$3F,$D5,$73,$97,$F4,$31,$9A,$E3,$F0,$DF,$C0,$FF,$13
	.dc.b	$F8,$D8,$2E,$61,$FE,$A9,$8F,$DF,$CE,$A7,$3E,$EF,$F0,$BF,$B1,$7C
	.dc.b	$9F,$F1,$7C,$13,$98,$FE,$35,$F3,$DF,$E6,$AC,$7F,$84,$E5,$FC,$3D
	.dc.b	$F0,$62,$9D,$3F,$B8,$9D,$3E,$5F,$AC,$7D,$6F,$F1,$B9,$7C,$7F,$F2
	.dc.b	$79,$CE,$1D,$A3,$F9,$57,$D6,$BE,$26,$9F,$C8,$FE,$B5,$F4,$B8,$BF
	.dc.b	$C5,$67,$F0,$EF,$84,$7F,$20,$AF,$E0,$4E,$A5,$3A,$7F,$87,$CE,$9F
	.dc.b	$F6,$EC,$67,$8B,$FA,$34,$E5,$FF,$1F,$46,$6B,$F4,$53,$97,$F4,$9F
	.dc.b	$5B,$FC,$1A,$BF,$A7,$36,$33,$BF,$F3,$D8,$C6,$0A,$46,$1B,$E6,$6E
	.dc.b	$B8,$3F,$0F,$CC,$F8,$9D,$37,$1F,$80,$DC,$7D,$AD,$C3,$D6,$FF,$57
	.dc.b	$FA,$CF,$F8,$01,$5F,$15,$B5,$FC,$CB,$2F,$73,$F8,$F8,$CD,$FF,$97
	.dc.b	$DB,$5F,$C0,$FE,$A3,$F0,$73,$9F,$94,$FE,$B5,$18,$9F,$C4,$D7,$0B
	.dc.b	$F8,$13,$97,$B5,$70,$F8,$BF,$C0,$7C,$D4,$C3,$EE,$78,$56,$FE,$CC
	.dc.b	$DC,$BE,$BC,$DA,$FA,$7E,$8F,$CF,$FA,$CF,$A0,$D8,$FE,$CB,$AA,$31
	.dc.b	$8F,$C1,$6E,$1E,$FF,$6F,$E4,$1F,$93,$FA,$DB,$87,$C9,$8F,$EF,$1F
	.dc.b	$5A,$CB,$38,$FF,$5C,$8C,$56,$CF,$EE,$AE,$23,$F0,$51,$F1,$AB,$9D
	.dc.b	$7E,$1F,$EB,$7F,$11,$E7,$DA,$F4,$B6,$FE,$C3,$72,$FC,$8B,$FA,$9F
	.dc.b	$46,$57,$E9,$7D,$9B,$7F,$A2,$9B,$7F,$A0,$9B,$5F,$C0,$3E,$B7,$F6
	.dc.b	$2F,$EA,$DB,$7F,$87,$F0,$3F,$C4,$51,$8B,$FF,$52,$36,$BF,$0D,$F1
	.dc.b	$BA,$EE,$FC,$2B,$63,$FC,$2F,$CF,$FD,$49,$F0,$4E,$9F,$DE,$7D,$6F
	.dc.b	$F0,$33,$FC,$4D,$EB,$BD,$DC,$4E,$89,$A7,$F8,$32,$75,$18,$CF,$F6
	.dc.b	$BF,$49,$FB,$33,$F6,$9F,$23,$67,$F0,$B1,$6F,$F4,$93,$6B,$FD,$81
	.dc.b	$F5,$BF,$CE,$5B,$E8,$FF,$90,$E3,$18,$7F,$28,$7F,$A6,$BF,$C4,$DC
	.dc.b	$DF,$F3,$B7,$2F,$E0,$8C,$7C,$AF,$3A,$E9,$E1,$CB,$E5,$F1,$B6,$F8
	.dc.b	$0C,$3F,$86,$FC,$63,$AF,$FA,$FD,$F8,$70,$FF,$2D,$4E,$7D,$5F,$47
	.dc.b	$4F,$A7,$7B,$7E,$76,$E5,$DB,$24,$63,$6F,$86,$FF,$98,$FD,$9F,$F2
	.dc.b	$FF,$AC,$FA,$0F,$E9,$31,$FC,$3A,$31,$87,$F5,$B8,$DC,$EB,$C6,$65
	.dc.b	$FE,$95,$8E,$3F,$15,$8B,$FF,$4F,$C6,$E7,$51,$CE,$75,$87,$68,$FF
	.dc.b	$03,$DE,$2F,$84,$72,$3C,$51,$97,$95,$F4,$78,$B7,$0F,$55,$7B,$3E
	.dc.b	$33,$85,$68,$E8,$7F,$85,$9C,$CF,$F6,$8C,$3C,$9F,$5A,$F9,$FF,$ED
	.dc.b	$97,$C2,$B5,$87,$F7,$0A,$D7,$AB,$E1,$F2,$FB,$C6,$2F,$84,$63,$0E
	.dc.b	$78,$F6,$6E,$1C,$DB,$5F,$46,$B8,$61,$EF,$C3,$A6,$1C,$63,$C1,$9E
	.dc.b	$70,$E3,$82,$F7,$8F,$5B,$DC,$FC,$29,$CB,$FE,$19,$6B,$C6,$B9,$E5
	.dc.b	$F0,$23,$51,$DA,$76,$70,$9D,$4F,$1A,$F2,$11,$F8,$78,$5C,$7D,$51
	.dc.b	$E3,$C3,$BE,$5F,$02,$31,$7B,$CB,$F2,$F3,$9B,$E2,$48,$CE,$1B,$BF
	.dc.b	$3D,$D7,$FC,$AB,$7C,$6B,$E2,$47,$63,$CA,$5B,$FC,$1F,$BD,$FC,$6E
	.dc.b	$3F,$C5,$E7,$C7,$70,$E3,$3F,$C0,$36,$32,$FE,$1D,$E5,$3D,$3B,$3E
	.dc.b	$E4,$62,$FF,$19,$E2,$7F,$8F,$2B,$57,$F5,$EB,$4F,$86,$FF,$09,$5B
	.dc.b	$BE,$D1,$D0,$93,$F7,$E7,$4F,$13,$A4,$6A,$FF,$43,$3A,$9F,$C5,$5F
	.dc.b	$81,$B5,$BB,$F3,$C5,$3A,$5F,$A0,$E6,$E4,$E2,$78,$0F,$21,$9C,$3F
	.dc.b	$D1,$75,$70,$F1,$46,$27,$85,$75,$BF,$0C,$61,$F8,$2D,$3F,$39,$5C
	.dc.b	$B0,$F5,$5F,$3B,$E5,$5A,$3C,$53,$A3,$75,$C7,$0C,$9F,$42,$69,$DC
	.dc.b	$FD,$84,$7A,$0E,$0E,$6F,$FD,$82,$F1,$3E,$09,$2F,$F8,$8F,$89,$F4
	.dc.b	$7E,$B7,$F9,$34,$7F,$27,$6E,$5F,$E8,$06,$D7,$FA,$45,$7C,$7F,$19
	.dc.b	$93,$FC,$9E,$76,$9D,$B9,$9D,$BF,$C9,$A3,$66,$9C,$47,$FA,$43,$97
	.dc.b	$FA,$04,$73,$8C,$6F,$E5,$6C,$79,$7B,$6B,$E1,$F2,$EE,$1F,$EA,$28
	.dc.b	$FF,$42,$35,$FE,$D2,$FF,$90,$DC,$C7,$02,$3F,$E9,$0E,$B7,$C0,$F9
	.dc.b	$32,$F2,$DB,$C1,$EE,$4B,$FF,$49,$7D,$6F,$EF,$2B,$73,$99,$F0,$BE
	.dc.b	$89,$D6,$9F,$EB,$07,$FD,$BF,$3C,$B9,$74,$6F,$9E,$47,$66,$BF,$E0
	.dc.b	$BC,$FC,$FE,$87,$C4,$FA,$97,$D6,$D2,$57,$FA,$9F,$EB,$7D,$99,$FF
	.dc.b	$B0,$2B,$73,$FE,$AB,$6B,$FE,$10,$6E,$BF,$5A,$B8,$CF,$FB,$5E,$5F
	.dc.b	$23,$E1,$24,$73,$F7,$2B,$CE,$2D,$FE,$02,$BF,$D9,$33,$FE,$BB,$FA
	.dc.b	$C7,$D6,$FA,$FC,$09,$39,$8F,$C0,$CE,$A3,$04,$3D,$1F,$1F,$A9,$91
	.dc.b	$9C,$3C,$13,$F8,$AF,$AD,$FF,$0F,$CF,$FB,$2D,$AF,$F8,$C1,$B3,$F9
	.dc.b	$67,$F0,$51,$BF,$17,$91,$1E,$5C,$BD,$7C,$9E,$DE,$EC,$9E,$D1,$97
	.dc.b	$C1,$19,$7C,$11,$F9,$EF,$AD,$7D,$6F,$9F,$BE,$57,$FB,$86,$72,$D9
	.dc.b	$E0,$47,$D9,$F3,$75,$FF,$25,$CF,$06,$CF,$C3,$9F,$D6,$FD,$63,$A2
	.dc.b	$FB,$D2,$FF,$DE,$13,$96,$CE,$5F,$F2,$BC,$27,$9C,$F4,$FA,$D9,$5C
	.dc.b	$9E,$FF,$57,$88,$E3,$17,$87,$F8,$2E,$0A,$FE,$03,$E6,$7F,$90,$67
	.dc.b	$3D,$7E,$83,$3B,$4C,$47,$F9,$AA,$B8,$4E,$23,$FC,$6D,$3F,$C3,$3E
	.dc.b	$B1,$F5,$BC,$67,$84,$DC,$3F,$8B,$D6,$A7,$AC,$63,$AF,$97,$64,$9C
	.dc.b	$E3,$DE,$FF,$4F,$18,$3C,$AE,$1F,$F3,$FB,$FA,$46,$CF,$96,$BC,$DE
	.dc.b	$7C,$9E,$42,$91,$EB,$EC,$D4,$F4,$8B,$8F,$AA,$7E,$5F,$BB,$FC,$B3
	.dc.b	$6F,$CE,$DF,$FA,$41,$B5,$FC,$DB,$65,$70,$AF,$0C,$61,$B1,$CC,$D3
	.dc.b	$F4,$5B,$3F,$D0,$2B,$84,$FF,$A7,$BE,$B1,$F5,$BF,$CF,$9B,$1E,$E6
	.dc.b	$BF,$EA,$4F,$B1,$7C,$3E,$FD,$19,$3B,$7F,$A3,$CF,$0C,$7F,$A7,$FD
	.dc.b	$73,$E6,$3F,$A5,$B6,$FE,$5B,$5F,$F5,$8F,$D8,$FD,$F7,$FD,$67,$0F
	.dc.b	$8E,$E2,$48,$FA,$72,$F8,$0F,$EA,$FF,$5F,$FC,$FA,$7F,$D7,$6D,$C7
	.dc.b	$FB,$0B,$79,$FF,$29,$6F,$A9,$E3,$F3,$53,$0F,$BF,$68,$FE,$C7,$5F
	.dc.b	$D3,$27,$F3,$FF,$06,$FA,$FF,$EC,$ED,$8F,$C3,$B5,$FF,$66,$FD,$8F
	.dc.b	$F1,$EC,$BF,$49,$3F,$32,$77,$C5,$C8,$FE,$E3,$5F,$17,$4F,$EF,$3F
	.dc.b	$60,$BF,$EE,$AD,$BF,$CA,$B5,$EF,$E2,$6B,$E1,$A4,$69,$FE,$FB,$ED
	.dc.b	$BD,$DE,$7F,$2F,$2B,$9B,$AF,$27,$FC,$07,$F6,$0F,$83,$A4,$AF,$D2
	.dc.b	$E1,$EF,$D7,$FC,$1F,$F6,$0F,$95,$FF,$05,$B6,$BF,$56,$D7,$FD,$E2
	.dc.b	$F8,$DF,$D7,$FC,$4B,$E8,$BF,$C2,$6D,$8C,$C6,$3E,$2A,$47,$06,$BF
	.dc.b	$EF,$CF,$AF,$FC,$33,$D3,$D5,$E6,$47,$F7,$BF,$60,$FA,$5F,$F1,$0B
	.dc.b	$63,$31,$83,$FC,$4B,$4F,$92,$A6,$67,$67,$F0,$88,$D3,$FC,$2B,$EC
	.dc.b	$3F,$00,$F9,$6B,$99,$FE,$30,$8C,$ED,$FE,$38,$6F,$5F,$F2,$B6,$D7
	.dc.b	$F3,$06,$AB,$FC,$E3,$DF,$AF,$A1,$1F,$27,$81,$DF,$9F,$D7,$E6,$49
	.dc.b	$F3,$46,$5A,$43,$E6,$7B,$9F,$E4,$6F,$B0,$FF,$28,$69,$FE,$4E,$6E
	.dc.b	$9F,$D0,$5B,$5F,$D1,$97,$5F,$1E,$79,$3F,$CB,$A7,$69,$DB,$FC,$CB
	.dc.b	$27,$C7,$FF,$95,$9B,$19,$8F,$F2,$CF,$D8,$76,$CB,$FA,$AF,$D8,$2F
	.dc.b	$FA,$A1,$C2,$D3,$FE,$66,$8C,$36,$F9,$E5,$AA,$5F,$37,$FB,$0E,$1D
	.dc.b	$DC,$E1,$DA,$FF,$B1,$56,$E7,$C5,$18,$9F,$EC,$D5,$C6,$B4,$D8,$E5
	.dc.b	$5E,$A6,$BD,$8F,$0D,$F9,$6B,$8B,$67,$71,$E1,$7E,$03,$AC,$3A,$57
	.dc.b	$26,$BE,$62,$CF,$03,$CF,$5D,$63,$81,$E3,$38,$16,$38,$3A,$AF,$03
	.dc.b	$FE,$F6,$17,$F2,$5F,$F7,$CB,$5E,$F7,$B7,$7E,$53,$E1,$B1,$5F,$D3
	.dc.b	$DB,$19,$9E,$D3,$97,$90,$F8,$EA,$D3,$2E,$08,$FB,$6F,$FD,$73,$B7
	.dc.b	$C6,$8F,$F1,$0C,$FC,$89,$CC,$7C,$33,$27,$1D,$95,$C6,$3C,$19,$7E
	.dc.b	$16,$B8,$BF,$87,$8E,$2E,$6B,$8C,$75,$38,$E7,$FE,$BF,$8C,$CF,$19
	.dc.b	$C9,$FE,$2A,$8C,$CF,$D1,$C3,$D1,$18,$9F,$F1,$C1,$C0,$FD,$2B,$97
	.dc.b	$FC,$79,$1C,$30,$91,$FA,$A7,$C4,$7F,$8A,$1F,$4D,$61,$D1,$E3,$8F
	.dc.b	$F1,$6E,$0F,$8E,$F6,$BF,$EC,$5C,$2C,$FF,$B6,$E6,$D7,$F9,$43,$EC
	.dc.b	$3F,$81,$CF,$F8,$D9,$B1,$D8,$3F,$E5,$68,$D6,$FC,$A7,$2F,$4C,$3F
	.dc.b	$C1,$98,$66,$F9,$C7,$93,$CC,$FB,$AF,$FB,$B2,$3F,$A4,$61,$FE,$4F
	.dc.b	$87,$D8,$3E,$73,$97,$57,$58,$C5,$79,$E3,$13,$A8,$E5,$3A,$CF,$D5
	.dc.b	$87,$F8,$4E,$7D,$DE,$6F,$9F,$EC,$3F,$A9,$46,$09,$5F,$3B,$24,$FF
	.dc.b	$B2,$A7,$31,$F3,$A3,$13,$DA,$BA,$7B,$7F,$05,$B7,$FE,$62,$F9,$4F
	.dc.b	$F0,$43,$7A,$BE,$57,$D7,$3E,$C3,$E6,$56,$B1,$FA,$4D,$CF,$F3,$30
	.dc.b	$FB,$07,$C4,$70,$C7,$EA,$46,$32,$FF,$38,$B6,$BE,$E6,$5F,$E0,$D9
	.dc.b	$CD,$F8,$BC,$AE,$3F,$2D,$F7,$7E,$C1,$E0,$E2,$FF,$CF,$71,$86,$C6
	.dc.b	$F9,$78,$36,$37,$CD,$F8,$7F,$0F,$FC,$29,$18,$3F,$D0,$AE,$34,$FC
	.dc.b	$87,$D8,$3D,$CF,$C6,$79,$7F,$5A,$73,$18,$F2,$7F,$25,$87,$F9,$9C
	.dc.b	$FC,$57,$D8,$7F,$93,$5F,$F4,$89,$9A,$FB,$91,$97,$F3,$DF,$33,$F9
	.dc.b	$D7,$39,$7E,$76,$31,$CB,$FA,$CA,$FF,$3D,$DF,$FA,$72,$5F,$2D,$FA
	.dc.b	$1F,$8D,$7C,$DF,$18,$C3,$CF,$AF,$53,$6F,$F4,$0F,$FB,$4D,$FF,$53
	.dc.b	$F8,$AF,$4B,$F5,$9F,$61,$FB,$D9,$CC,$63,$C6,$FD,$95,$79,$37,$FC
	.dc.b	$C7,$D8,$7C,$9B,$E7,$C1,$FB,$58,$C4,$FE,$F7,$EB,$7F,$CC,$95,$FB
	.dc.b	$1F,$B0,$F8,$DC,$5F,$8E,$9C,$D7,$E8,$E7,$35,$FE,$D5,$9F,$F5,$7B
	.dc.b	$FB,$9F,$B0,$F8,$EE,$7E,$2B,$F8,$24,$62,$FF,$DB,$F4,$F7,$7F,$84
	.dc.b	$7D,$83,$E5,$BF,$86,$4E,$71,$FE,$09,$39,$7F,$D9,$7F,$3D,$FE,$6B
	.dc.b	$7F,$84,$65,$FC,$39,$FF,$74,$46,$71,$FD,$D4,$E5,$FF,$77,$FD,$7B
	.dc.b	$EC,$3F,$8C,$4B,$E4,$FF,$8E,$4E,$70,$EB,$1F,$C7,$FE,$C3,$FD,$59
	.dc.b	$1F,$ED,$C6,$E1,$FC,$1F,$EC,$1F,$53,$BF,$8E,$97,$FC,$76,$7F,$7C
	.dc.b	$F1,$8F,$E5,$11,$8A,$FE,$61,$7D,$38,$FF,$9D,$CE,$5F,$E0,$31,$9B
	.dc.b	$FF,$7E,$CE,$5F,$F7,$FF,$D7,$FE,$FE,$BF,$A1,$B6,$33,$C3,$FC,$DE
	.dc.b	$31,$87,$F8,$E9,$B4,$69,$F4,$63,$F9,$23,$AB,$FE,$15,$1F,$07,$9B
	.dc.b	$D7,$E5,$7B,$1B,$8F,$C1,$6C,$EE,$0F,$F4,$FF,$B0,$7C,$BF,$F4,$D6
	.dc.b	$CF,$F4,$5C,$7E,$05,$70,$9C,$BC,$6B,$FA,$97,$D6,$7E,$07,$3F,$E4
	.dc.b	$F5,$A7,$FB,$2C,$67,$0F,$0D,$AB,$51,$E4,$A5,$6E,$BC,$17,$DB,$E2
	.dc.b	$7E,$CF,$CF,$4C,$3E,$F7,$83,$6F,$F0,$93,$71,$FB,$73,$6B,$EB,$F9
	.dc.b	$FF,$4F,$EB,$FE,$A3,$63,$E0,$37,$0F,$A2,$DB,$F6,$FB,$7F,$1C,$FC
	.dc.b	$5F,$D8,$5B,$F5,$1E,$ED,$BF,$B2,$57,$F8,$1B,$EB,$59,$6F,$2F,$EC
	.dc.b	$11,$8B,$E3,$1F,$3F,$1F,$EF,$D3,$98,$FB,$11,$8A,$FC,$67,$D8,$7E
	.dc.b	$37,$CF,$B8,$7D,$96,$DF,$E2,$9B,$8F,$E5,$9F,$D1,$F9,$F2,$BF,$51
	.dc.b	$EC,$5B,$FD,$74,$DB,$FD,$2C,$DA,$FB,$7F,$61,$FB,$27,$D3,$68,$C3
	.dc.b	$F9,$AE,$FF,$F0,$D4,$62,$FF,$B6,$B6,$BF,$05,$CB,$FD,$46,$73,$DB
	.dc.b	$F8,$EF,$47,$8B,$6B,$BD,$F1,$C7,$FD,$4B,$87,$19,$FD,$BF,$D8,$7F
	.dc.b	$00,$3F,$C5,$1E,$A3,$DB,$C4,$FF,$07,$9B,$7F,$E1,$3A,$D4,$E4,$FD
	.dc.b	$27,$D3,$7E,$EC,$FD,$E7,$C7,$D8,$C9,$2B,$F7,$31,$97,$FC,$7D,$F6
	.dc.b	$1F,$CF,$B6,$FF,$06,$BF,$D3,$E3,$3F,$2F,$FE,$3F,$8C,$4E,$A3,$1E
	.dc.b	$8E,$5B,$87,$D0,$AF,$E1,$B6,$69,$FE,$DA,$8D,$7B,$BD,$4F,$F2,$AF
	.dc.b	$C5,$BA,$BF,$B1,$4E,$8F,$C9,$C6,$7D,$4F,$27,$47,$97,$6B,$5E,$16
	.dc.b	$E5,$EE,$6E,$BE,$E7,$FC,$B5,$EC,$7F,$98,$3D,$57,$D1,$FF,$04,$8F
	.dc.b	$E1,$B1,$8B,$FE,$2F,$1B,$9D,$78,$BF,$E8,$CA,$D5,$7F,$81,$3E,$23
	.dc.b	$15,$AC,$BF,$D3,$EF,$58,$CC,$F4,$BA,$FF,$83,$1B,$87,$8A,$B8,$5F
	.dc.b	$2A,$E0,$D7,$83,$67,$FD,$F0,$DA,$E7,$F1,$3D,$3C,$7E,$FF,$60,$FA
	.dc.b	$37,$A9,$C6,$73,$18,$CB,$FD,$81,$9F,$C5,$9C,$D6,$F2,$F1,$61,$CE
	.dc.b	$2E,$1C,$75,$F8,$B3,$9C,$79,$9F,$33,$0E,$6E,$04,$6E,$78,$D7,$3C
	.dc.b	$7C,$58,$F6,$C6,$61,$C7,$0E,$15,$A8,$C5,$F1,$7F,$A8,$CE,$9B,$19
	.dc.b	$7F,$C7,$B8,$F9,$31,$F0,$E1,$D9,$FB,$FA,$7F,$83,$DC,$D7,$69,$D6
	.dc.b	$3E,$19,$C9,$1F,$EA,$38,$F8,$6B,$BE,$5C,$1F,$1D,$D7,$27,$52,$5F
	.dc.b	$37,$E5,$4E,$EB,$A1,$D7,$0F,$36,$1E,$0C,$BD,$D3,$FE,$58,$EB,$FF
	.dc.b	$6C,$E1,$F1,$3D,$DF,$E2,$8E,$27,$EC,$3F,$ED,$E7,$83,$98,$DE,$5F
	.dc.b	$79,$B8,$7B,$9B,$3F,$43,$C3,$FF,$71,$CE,$70,$FC,$4D,$6A,$BF,$02
	.dc.b	$DC,$3F,$D0,$91,$A9,$8F,$DF,$9D,$5F,$48,$F5,$4E,$B0,$F2,$3F,$96
	.dc.b	$9D,$4F,$68,$D6,$1E,$38,$FC,$54,$EB,$0F,$7D,$F2,$9F,$05,$F3,$78
	.dc.b	$9E,$03,$27,$23,$27,$73,$31,$FA,$4A,$ED,$87,$BC,$95,$D6,$B7,$87
	.dc.b	$96,$FC,$47,$18,$EB,$87,$9A,$74,$7E,$3B,$0E,$B7,$AC,$3E,$3D,$D8
	.dc.b	$E3,$3A,$8E,$E7,$AF,$0E,$CE,$0F,$53,$E2,$BE,$D1,$E7,$8C,$3C,$63
	.dc.b	$2F,$C3,$24,$74,$3F,$82,$F8,$7F,$53,$EC,$3F,$95,$C7,$F2,$C6,$D7
	.dc.b	$F8,$E1,$B5,$FE,$4A,$5F,$19,$C6,$64,$FF,$25,$9D,$A7,$6E,$67,$6F
	.dc.b	$F1,$18,$D9,$A7,$11,$FE,$48,$E5,$FF,$79,$CE,$5B,$19,$DF,$CD,$DB
	.dc.b	$6F,$CD,$E5,$D7,$C5,$E9,$5C,$7F,$D4,$51,$FE,$18,$53,$92,$46,$5E
	.dc.b	$9D,$6F,$80,$63,$CF,$C1,$97,$CD,$F0,$7D,$87,$F9,$82,$1A,$FA,$A7
	.dc.b	$71,$8A,$F9,$D1,$C2,$38,$F3,$72,$E7,$91,$FE,$2E,$C3,$FD,$C6,$D9
	.dc.b	$DB,$4F,$EF,$38,$BD,$0F,$85,$F5,$CF,$B0,$A4,$AF,$F1,$CF,$D8,$7B
	.dc.b	$1F,$F1,$DC,$ED,$B3,$B7,$D0,$7F,$A3,$27,$83,$81,$2F,$8D,$F0,$92
	.dc.b	$3A,$FB,$75,$E1,$3B,$7C,$33,$E8,$9F,$B5,$F6,$0F,$B0,$F5,$F8,$32
	.dc.b	$FF,$CA,$33,$96,$CF,$02,$3E,$CF,$8A,$AF,$9E,$78,$36,$7E,$3C,$FE
	.dc.b	$27,$EC,$1E,$35,$25,$7F,$98,$21,$C7,$88,$F8,$51,$C1,$D4,$F3,$9E
	.dc.b	$9E,$57,$5E,$A9,$5C,$E3,$2F,$58,$CB,$D2,$3F,$39,$F6,$0F,$A1,$F3
	.dc.b	$77,$CA,$FF,$39,$CE,$5B,$3C,$08,$FA,$7E,$8A,$BF,$66,$78,$36,$7E
	.dc.b	$14,$FE,$AF,$EB,$FF,$19,$3F,$E8,$05,$EE,$E4,$D7,$3C,$9D,$46,$5E
	.dc.b	$11,$CF,$EC,$24,$73,$7B,$7D,$76,$23,$F1,$96,$9C,$77,$FE,$42,$15
	.dc.b	$FB,$FF,$2F,$FC,$8B,$39,$F0,$3D,$06,$76,$98,$FF,$A6,$2B,$73,$FE
	.dc.b	$D6,$6C,$7F,$BA,$3E,$BD,$F6,$1C,$23,$F3,$D3,$A7,$FC,$79,$7F,$C3
	.dc.b	$EB,$71,$FE,$A4,$7F,$38,$FE,$95,$75,$6C,$E1,$D7,$27,$84,$65,$E5
	.dc.b	$1D,$7C,$F9,$1D,$5F,$07,$66,$23,$B4,$5D,$3F,$6B,$82,$BE,$37,$B9
	.dc.b	$FC,$97,$6F,$CC,$E1,$FE,$BA,$6B,$EE,$8D,$BF,$1D,$AF,$EF,$E1,$18
	.dc.b	$BF,$E0,$7A,$62,$B8,$1D,$67,$F9,$A7,$D7,$BE,$C2,$92,$7F,$D9,$7F
	.dc.b	$60,$F8,$6D,$FA,$12,$7F,$D9,$F7,$FE,$16,$7F,$88,$57,$37,$FD,$A7
	.dc.b	$F6,$0F,$A2,$FE,$84,$D8,$CC,$7F,$45,$FA,$FF,$AF,$1F,$C6,$65,$F1
	.dc.b	$BC,$09,$1F,$ED,$E9,$7B,$FF,$D3,$7E,$C3,$F9,$9C,$FF,$21,$6F,$07
	.dc.b	$F2,$46,$F4,$A6,$5B,$E9,$F9,$7C,$CC,$9F,$F7,$5D,$A3,$CD,$1F,$D6
	.dc.b	$FE,$09,$F6,$14,$91,$FE,$F2,$FB,$0F,$E2,$0F,$F3,$39,$F9,$4F,$FB
	.dc.b	$D9,$AE,$78,$FF,$B3,$4F,$C5,$8F,$E7,$5F,$60,$AF,$ED,$4D,$8F,$F7
	.dc.b	$FB,$5F,$E8,$12,$F1,$37,$F5,$32,$76,$FF,$73,$F6,$DE,$EF,$4F,$99
	.dc.b	$95,$D1,$D7,$93,$FD,$E7,$EC,$1F,$07,$49,$5F,$2F,$97,$FA,$8E,$3F
	.dc.b	$12,$31,$CB,$F3,$D7,$D8,$3E,$57,$FC,$08,$D8,$D9,$F5,$5C,$3F,$D6
	.dc.b	$DF,$3B,$A1,$F1,$0F,$B0,$FF,$07,$36,$34,$D7,$FB,$22,$F8,$9D,$92
	.dc.b	$36,$FF,$85,$FE,$BB,$F1,$B1,$CB,$D5,$EA,$47,$C5,$F6,$0F,$A5,$FF
	.dc.b	$0F,$36,$33,$18,$3A,$7C,$9C,$8C,$90,$FE,$F1,$68,$FF,$14,$7D,$87
	.dc.b	$E1,$B8,$3F,$BE,$CF,$5A,$DD,$EB,$AE,$47,$C5,$BE,$1C,$1F,$45,$B1
	.dc.b	$FA,$1B,$FD,$03,$BA,$E2,$FF,$82,$67,$6A,$76,$E7,$9B,$E3,$79,$92
	.dc.b	$3C,$11,$E8,$23,$83,$EC,$4B,$DC,$FF,$22,$7D,$87,$F9,$2F,$2F,$C3
	.dc.b	$36,$BF,$6C,$D4,$BA,$B8,$F3,$49,$FC,$24,$EC,$D3,$F8,$7C,$9F,$1F
	.dc.b	$FE,$53,$6C,$66,$3F,$CA,$DF,$61,$D7,$83,$F8,$37,$D8,$7F,$37,$AF
	.dc.b	$E1,$17,$C1,$FF,$17,$36,$7F,$C6,$2D,$CF,$F4,$17,$CE,$97,$CD,$46
	.dc.b	$63,$15,$D5,$B8,$7E,$A5,$E1,$3B,$9E,$B9,$FF,$97,$2F,$AC,$7E,$BC
	.dc.b	$E4,$5B,$E4,$F0,$C7,$DA,$D9,$E2,$7B,$A7,$6D,$36,$7F,$29,$BF,$13
	.dc.b	$6F,$C5,$87,$A2,$36,$7F,$5F,$AE,$6F,$C0,$9D,$DF,$86,$B8,$4F,$88
	.dc.b	$E0,$72,$BE,$EF,$1C,$3F,$DF,$2D,$3F,$DF,$71,$F3,$DF,$F4,$03,$5F
	.dc.b	$A7,$E5,$3E,$2B,$8C,$FF,$55,$6C,$66,$79,$B5,$EF,$F1,$DE,$07,$FA
	.dc.b	$53,$53,$48,$FA,$EF,$FD,$6F,$A7,$C4,$8E,$26,$76,$FF,$40,$65,$F8
	.dc.b	$58,$CB,$F8,$7A,$E3,$1F,$14,$CE,$7F,$EB,$5A,$E2,$FF,$A7,$E3,$C6
	.dc.b	$75,$BF,$03,$8B,$E9,$5C,$23,$FD,$4D,$87,$D8,$C3,$BD,$78,$E7,$86
	.dc.b	$1D,$F0,$E7,$7B,$C3,$95,$7F,$3F,$8C,$4C,$8F,$F4,$A6,$1E,$A3,$FD
	.dc.b	$2E,$FA,$8F,$E9,$55,$FE,$A1,$8F,$F4,$D9,$B9,$FF,$69,$7C,$76,$5F
	.dc.b	$15,$74,$8C,$3F,$DF,$7E,$BF,$F7,$F3,$FD,$1D,$B1,$FE,$A5,$7F,$C0
	.dc.b	$51,$9D,$F8,$CA,$77,$7E,$3F,$33,$C5,$3D,$6B,$BC,$FF,$84,$B1,$E3
	.dc.b	$7F,$D0,$F0,$FE,$B7,$2F,$B0,$7C,$E6,$BA,$78,$C6,$2B,$CD,$18,$9D
	.dc.b	$47,$29,$D6,$5E,$9F,$2F,$B9,$E2,$FB,$0F,$E9,$15,$EF,$8C,$BE,$28
	.dc.b	$FF,$61,$39,$31,$ED,$EA,$BF,$D9,$79,$2B,$71,$DE,$B6,$7F,$71,$F9
	.dc.b	$3F,$F0,$4B,$7A,$FE,$4F,$D7,$3E,$C3,$E5,$D6,$AF,$FD,$A8,$DC,$7F
	.dc.b	$DA,$FF,$60,$F8,$8D,$E3,$F4,$A3,$19,$7F,$7F,$6E,$1F,$ED,$DC,$BF
	.dc.b	$C1,$53,$9B,$F0,$F9,$3C,$7E,$57,$ED,$FD,$87,$F4,$FA,$E5,$18,$F0
	.dc.b	$64,$7F,$82,$E7,$EC,$63,$FE,$E9,$78,$37,$93,$9F,$47,$E1,$BC,$4F
	.dc.b	$F0,$85,$F1,$8C,$79,$5F,$8A,$FB,$07,$BB,$F8,$6F,$2F,$EA,$E5,$FE
	.dc.b	$F5,$9C,$E1,$F8,$69,$CF,$1F,$FB,$EA,$3F,$CD,$07,$E2,$BE,$C3,$FC
	.dc.b	$9E,$FF,$88,$4C,$DF,$D8,$8C,$BF,$F0,$07,$E5,$7F,$3A,$E7,$4F,$CE
	.dc.b	$C6,$2F,$F5,$D5,$FE,$7D,$BF,$F1,$64,$BE,$5B,$F3,$BF,$1A,$F9,$AE
	.dc.b	$31,$87,$9F,$6E,$A6,$DF,$E8,$1F,$F6,$9B,$FE,$37,$F1,$3F,$C6,$EF
	.dc.b	$E9,$3E,$C1,$E2,$FE,$C2,$73,$B7,$E4,$A7,$35,$C7,$6F,$D7,$FD,$87
	.dc.b	$C8,$A6,$EA,$FE,$21,$5A,$E2,$FD,$DF,$D7,$7F,$9A,$67,$F6,$7F,$61
	.dc.b	$F2,$2F,$86,$1F,$E8,$3A,$D3,$FA,$29,$D5,$7F,$96,$67,$FD,$5A,$FE
	.dc.b	$F3,$EC,$3E,$53,$9F,$89,$FE,$07,$18,$BE,$9E,$E7,$F0,$6F,$B0,$7C
	.dc.b	$B7,$92,$30,$FF,$10,$F9,$FF,$F3,$6E,$3F,$C2,$31,$FE,$22,$FF,$9A
	.dc.b	$63,$38,$7E,$DA,$72,$FF,$A6,$FE,$BD,$F6,$1F,$C5,$27,$FD,$BB,$7F
	.dc.b	$E6,$79,$CB,$D7,$E4,$FF,$8F,$FD,$87,$FA,$BA,$3F,$CD,$8D,$C3,$F8
	.dc.b	$1F,$D8,$3E,$AF,$7F,$1D,$2F,$F8,$EC,$FE,$FD,$E3,$1F,$C9,$23,$15
	.dc.b	$F5,$A7,$FD,$05,$7D,$78,$FF,$9F,$56,$9F,$A7,$49,$D5,$7F,$A3,$6B
	.dc.b	$4F,$FA,$3F,$EC,$3F,$81,$DF,$F4,$76,$C6,$77,$FE,$6B,$18,$C3,$F0
	.dc.b	$ED,$A3,$4F,$AB,$5F,$CE,$1D,$5F,$F0,$EA,$F8,$7C,$DE,$DF,$27,$DC
	.dc.b	$DB,$F8,$6D,$9D,$C7,$FA,$5F,$EB,$DF,$31,$FD,$39,$B1,$DF,$5C,$4C
	.dc.b	$8F,$E9,$3F,$59,$F8,$0D,$3E,$2D,$F0,$C3,$F9,$BC,$E6,$BF,$D5,$D3
	.dc.b	$9A,$FE,$D5,$39,$AF,$F7,$04,$E6,$BF,$D7,$13,$9A,$66,$F9,$EF,$B4
	.dc.b	$7D,$BF,$02,$DF,$E0,$22,$E3,$F8,$69,$B7,$F4,$FC,$DF,$A3,$F5,$FF
	.dc.b	$49,$B8,$7C,$D6,$DF,$F5,$0E,$04,$62,$FD,$FE,$C7,$C7,$3F,$19,$F6
	.dc.b	$17,$0F,$99,$1F,$DC,$BE,$B1,$96,$72,$FE,$B7,$18,$AD,$9F,$DD,$1C
	.dc.b	$47,$E5,$4F,$8B,$5C,$EB,$F1,$3F,$61,$F8,$AF,$32,$E1,$F8,$56,$DF
	.dc.b	$A5,$B9,$7E,$4D,$FD,$17,$87,$2B,$F4,$FE,$B5,$BF,$D6,$CD,$BF,$D1
	.dc.b	$CD,$AF,$F7,$47,$D8,$7E,$C5,$FF,$70,$36,$FE,$47,$77,$F8,$62,$31
	.dc.b	$7F,$AA,$6D,$7D,$EF,$89,$D7,$6F,$F8,$63,$9B,$FA,$7C,$F6,$8C,$4E
	.dc.b	$A3,$F2,$2F,$EE,$FE,$C2,$47,$F8,$73,$D3,$7B,$58,$D3,$F8,$3D,$5C
	.dc.b	$FF,$C5,$13,$6F,$F7,$1F,$4D,$FB,$43,$F6,$BF,$1B,$63,$F6,$11,$70
	.dc.b	$FC,$24,$DB,$FE,$0B,$F6,$1F,$CD,$9B,$E8,$7F,$8E,$23,$18,$2E,$47
	.dc.b	$F4,$A7,$EF,$07,$6E,$A7,$97,$CB,$F3,$7A,$7B,$92,$FA,$61,$E2,$F8
	.dc.b	$BC,$61,$D3,$1F,$E1,$DF,$14,$EB,$FE,$B7,$7E,$8C,$7F,$CA,$B3,$9F
	.dc.b	$90,$F4,$74,$FA,$77,$BA,$76,$3D,$59,$F9,$E3,$2F,$76,$E3,$F1,$1F
	.dc.b	$F2,$CF,$B1,$FE,$5E,$F5,$9F,$47,$CA,$3F,$84,$46,$2F,$E3,$46,$E7
	.dc.b	$5E,$3F,$F6,$4A,$D5,$7F,$80,$FE,$27,$15,$AC,$BF,$D4,$2F,$58,$CC
	.dc.b	$F4,$BA,$FF,$82,$DB,$87,$8A,$B8,$5F,$2A,$E0,$D7,$83,$65,$F1,$5C
	.dc.b	$AF,$9B,$D3,$E4,$FB,$FD,$83,$E8,$9E,$AF,$19,$CC,$63,$2F,$F5,$BE
	.dc.b	$7F,$16,$73,$5B,$CF,$C5,$87,$38,$B8,$71,$D7,$E2,$CE,$71,$E6,$7C
	.dc.b	$CC,$39,$B8,$11,$B9,$E3,$5C,$F1,$F1,$63,$DB,$19,$87,$1C,$38,$56
	.dc.b	$A3,$17,$C5,$FD,$04,$E9,$B1,$97,$FC,$79,$8F,$93,$1F,$0E,$1D,$9F
	.dc.b	$BF,$A7,$F8,$39,$CD,$76,$9D,$63,$E1,$9C,$91,$FC,$BE,$3E,$1A,$EF
	.dc.b	$97,$07,$C7,$75,$C9,$D4,$97,$CD,$F9,$53,$BA,$FF,$29,$75,$7F,$C0
	.dc.b	$06,$CF,$FB,$7D,$FA,$4F,$EA,$0F,$9F,$EF,$7F,$16,$74,$DB,$FB,$2D
	.dc.b	$C7,$D2,$D9,$F9,$7E,$27,$FB,$6E,$73,$87,$E1,$2B,$55,$DA,$3F,$0F
	.dc.b	$5B,$8F,$A7,$3A,$C3,$CF,$87,$82,$F7,$5C,$D8,$FD,$FA,$E1,$7C,$E3
	.dc.b	$B1,$99,$DD,$7E,$4A,$75,$7D,$B0,$F3,$DE,$CF,$C5,$4E,$B0,$F2,$13
	.dc.b	$0F,$1C,$EF,$0F,$33,$6F,$EE,$DF,$1A,$F0,$C7,$6A,$F1,$56,$0F,$83
	.dc.b	$35,$FD,$0D,$71,$7A,$1A,$3C,$A6,$F0,$F5,$39,$AE,$AF,$03,$E1,$BE
	.dc.b	$63,$25,$BF,$21,$23,$ED,$9D,$70,$F8,$0F,$80,$E9,$5E,$07,$AC,$FC
	.dc.b	$07,$E2,$D7,$57,$C9,$19,$79,$47,$8A,$30,$FF,$06,$F1,$3E,$D7,$D8
	.dc.b	$7F,$2C,$8F,$E5,$AD,$9F,$E7,$ED,$AF,$EB,$4B,$E3,$B8,$CC,$9F,$E4
	.dc.b	$D3,$B4,$ED,$CC,$ED,$FE,$25,$1B,$34,$E2,$3F,$A7,$F2,$FF,$9C,$A7
	.dc.b	$2D,$8C,$EB,$E6,$ED,$B7,$E6,$F2,$EB,$E2,$F5,$2E,$3F,$EA,$08,$FE
	.dc.b	$B8,$6C,$A7,$A2,$76,$47,$7D,$6F,$81,$FF,$42,$E7,$E5,$F0,$BA,$92
	.dc.b	$B9,$FD,$87,$F0,$19,$D4,$76,$C5,$5B,$8F,$F0,$1D,$6F,$4F,$ED,$EF
	.dc.b	$FB,$86,$39,$F2,$FB,$23,$FA,$44,$70,$E6,$C1,$1F,$0B,$F1,$33,$F4
	.dc.b	$7A,$3F,$13,$EB,$9F,$61,$49,$7F,$0B,$EC,$3D,$B9,$FF,$80,$2B,$73
	.dc.b	$F1,$9A,$FF,$86,$9B,$B7,$FA,$A2,$B8,$CF,$F8,$86,$5F,$21,$E1,$24
	.dc.b	$73,$F7,$2B,$CE,$2D,$FF,$AC,$A9,$5F,$73,$EC,$1F,$61,$EB,$F0,$66
	.dc.b	$3F,$6A,$72,$D9,$E0,$47,$D5,$F1,$D5,$FF,$19,$CF,$06,$CF,$C7,$9F
	.dc.b	$C5,$FD,$83,$C6,$A4,$AF,$C4,$C7,$DC,$6C,$65,$FF,$1F,$C7,$17,$7C
	.dc.b	$BD,$3C,$9E,$BD,$52,$79,$C6,$5E,$B1,$97,$9C,$7E,$7B,$EC,$1F,$41
	.dc.b	$E6,$EF,$95,$F9,$C9,$CB,$67,$81,$1F,$4F,$D0,$D7,$FC,$AB,$3C,$1B
	.dc.b	$3F,$0A,$7F,$5B,$F5,$FF,$EB,$59,$FD,$4A,$F7,$B2,$6B,$A6,$4E,$A3
	.dc.b	$2F,$08,$E7,$F6,$12,$39,$BD,$BE,$BB,$11,$FE,$68,$B4,$E3,$9F,$FD
	.dc.b	$EF,$0A,$FE,$03,$E6,$7F,$92,$27,$3D,$7E,$83,$3B,$4C,$7F,$80,$D6
	.dc.b	$E7,$FC,$BE,$D8,$FF,$3D,$7D,$7B,$EC,$38,$47,$F9,$EA,$74,$FF,$8F
	.dc.b	$EF,$F8,$8D,$6E,$3F,$86,$BF,$A2,$6F,$56,$8D,$F6,$48,$C9,$0F,$F4
	.dc.b	$3C,$F1,$6F,$9F,$D8,$8F,$6E,$CA,$E8,$F0,$C7,$92,$3E,$2F,$BB,$FC
	.dc.b	$9B,$6F,$CC,$57,$FA,$51,$BB,$36,$F8,$75,$C2,$BE,$3B,$5E,$0D,$9E
	.dc.b	$11,$FE,$A1,$D5,$39,$AF,$F4,$5D,$E2,$BF,$98,$FD,$7B,$EC,$3F,$9C
	.dc.b	$B6,$33,$1F,$CF,$3E,$C1,$F1,$3A,$F4,$A5,$72,$75,$5F,$D0,$BE,$C1
	.dc.b	$F4,$74,$95,$FD,$03,$EC,$3E,$B3,$FE,$B4,$97,$C6,$71,$24,$6F,$E0
	.dc.b	$7F,$A6,$FD,$7F,$F3,$58,$FF,$5D,$B7,$8B,$E0,$C2,$31,$C1,$FE,$AA
	.dc.b	$C5,$F6,$1C,$BE,$6A,$5F,$F5,$2B,$47,$F5,$AA,$FE,$89,$3F,$EC,$BF
	.dc.b	$83,$7D,$87,$F5,$E6,$C7,$F5,$A6,$BF,$EC,$EF,$AF,$FE,$20,$FF,$B4
	.dc.b	$23,$E4,$BB,$E2,$E4,$7F,$68,$9F,$8F,$1F,$D9,$3E,$C1,$5F,$DB,$1B
	.dc.b	$19,$8E,$1E,$26,$7E,$1A,$4F,$17,$FB,$9F,$B6,$F7,$79,$7C,$BC,$AE
	.dc.b	$CE,$BC,$9F,$EF,$3F,$60,$F8,$3A,$4A,$FE,$EF,$87,$BB,$97,$E7,$FD
	.dc.b	$83,$E5,$BF,$C0,$2D,$AF,$EF,$AD,$7F,$DE,$2F,$8D,$FF,$00,$FC,$43
	.dc.b	$E8,$BF,$C1,$2D,$8C,$C6,$3E,$26,$47,$06,$BF,$EF,$CF,$AE,$76,$72
	.dc.b	$F5,$64,$F6,$7F,$C2,$BF,$60,$FA,$5A,$49,$FF,$09,$C6,$5F,$F0,$AD
	.dc.b	$BE,$43,$FC,$3A,$D8,$DD,$71,$7C,$53,$FE,$22,$FB,$0F,$BF,$1E,$59
	.dc.b	$E2,$67,$AE,$47,$F8,$8A,$77,$C1,$F9,$A6,$F8,$29,$98,$FE,$69,$E1
	.dc.b	$7C,$95,$F1,$8F,$3F,$83,$E3,$E8,$F4,$F9,$92,$FC,$11,$96,$90,$F9
	.dc.b	$9E,$E7,$F8,$F3,$EC,$3F,$C8,$79,$F8,$E5,$18,$7C,$CD,$73,$D9,$C7
	.dc.b	$A2,$47,$F9,$02,$74,$27,$FC,$93,$8B,$E3,$E9,$2B,$FC,$8F,$F6,$1B
	.dc.b	$DB,$E2,$7D,$82,$FC,$87,$1A,$FF,$26,$B5,$F8,$B9,$EB,$0F,$F2,$BE
	.dc.b	$1C,$E5,$7C,$DF,$A1,$19,$8C,$47,$79,$F8,$51,$8B,$EC,$EA,$B8,$47
	.dc.b	$C7,$6D,$7D,$B8,$D9,$E3,$7D,$6E,$F1,$FB,$71,$E4,$8C,$63,$A8,$E5
	.dc.b	$87,$8A,$79,$46,$CF,$4D,$72,$C3,$C8,$D9,$DC,$7A,$6B,$95,$70,$BE
	.dc.b	$CB,$0F,$4B,$F0,$4C,$DF,$A9,$E0,$78,$4E,$18,$7A,$9F,$F7,$A4,$FC
	.dc.b	$F6,$C6,$5F,$69,$FE,$F9,$3C,$1E,$53,$E1,$B1,$18,$6D,$7F,$4E,$6C
	.dc.b	$EE,$76,$3E,$3A,$B4,$CB,$82,$3E,$9C,$7F,$D6,$1A,$7C,$28,$FD,$94
	.dc.b	$FC,$49,$CC,$7B,$CC,$9C,$76,$FE,$E7,$AF,$E2,$23,$2F,$E2,$A7,$8B
	.dc.b	$98,$F2,$39,$9E,$3A,$7C,$3B,$F0,$46,$2B,$F6,$47,$09,$FF,$5C,$C6
	.dc.b	$63,$F6,$D3,$C3,$0F,$54,$6C,$FE,$19,$87,$8A,$FB,$47,$A6,$78,$3F
	.dc.b	$A8,$37,$7F,$E0,$EA,$FE,$DF,$1E,$CA,$F0,$D7,$FA,$F2,$7E,$29,$FE
	.dc.b	$B8,$F8,$EE,$11,$AA,$E8,$7F,$18,$8C,$BF,$C7,$FE,$C3,$F8,$14,$FF
	.dc.b	$B1,$9B,$1C,$1F,$E4,$91,$9D,$F8,$CA,$77,$7E,$7E,$BF,$1D,$FF,$19
	.dc.b	$9C,$BF,$0A,$33,$5E,$1E,$8F,$1D,$BE,$C1,$F3,$9D,$B1,$FE,$95,$8F
	.dc.b	$FB,$72,$B5,$39,$BE,$37,$C2,$BD,$51,$89,$E5,$E6,$C9,$E5,$F6,$1F
	.dc.b	$D3,$DF,$81,$19,$CB,$E9,$62,$AF,$F5,$ED,$68,$E3,$ED,$EE,$7E,$B4
	.dc.b	$EA,$BE,$45,$FB,$23,$13,$FE,$28,$F9,$5F,$F0,$36,$1F,$EF,$46,$DF
	.dc.b	$D6,$FA,$E7,$D8,$7C,$E9,$CD,$F1,$EC,$FA,$3F,$60,$F8,$8D,$46,$19
	.dc.b	$97,$FC,$00,$C7,$FC,$10,$FE,$48,$FE,$99,$5C,$5D,$79,$5D,$3E,$5B
	.dc.b	$EE,$FD,$83,$C1,$C5,$71,$DB,$FB,$34,$E6,$31,$08,$C3,$F9,$07,$EC
	.dc.b	$B6,$7A,$F1,$F6,$E6,$FC,$5F,$87,$FE,$15,$C3,$9E,$1F,$E0,$E7,$AC
	.dc.b	$67,$0F,$CD,$FD,$83,$DC,$FC,$4F,$97,$F6,$E7,$31,$8F,$27,$F2,$F8
	.dc.b	$7F,$99,$8F,$C5,$7D,$87,$F9,$39,$F4,$99,$AF,$C0,$C6,$5F,$D4,$FC
	.dc.b	$CF,$E8,$9C,$E5,$FA,$28,$C7,$2F,$ED,$EB,$FC,$F3,$7F,$12,$5F,$2D
	.dc.b	$F9,$EF,$8D,$7C,$DF,$18,$C3,$CF,$B3,$53,$6F,$F4,$0F,$FB,$45,$F9
	.dc.b	$FE,$2B,$D2,$FD,$4F,$D8,$3C,$5F,$DA,$CE,$6B,$F5,$13,$97,$8E,$DF
	.dc.b	$B0,$FB,$0F,$95,$18,$3F,$5C,$E3,$83,$F8,$4F,$D7,$FF,$99,$A7,$F6
	.dc.b	$5F,$61,$F1,$F8,$BF,$D0,$73,$9A,$FD,$4C,$E6,$BF,$C3,$53,$FE,$A6
	.dc.b	$7F,$75,$F6,$1F,$21,$CF,$C5,$7F,$02,$8C,$5F,$F8,$CE,$9E,$EF,$F0
	.dc.b	$6F,$B0,$7C,$B7,$F1,$29,$CE,$3F,$C1,$67,$2F,$E5,$3E,$7F,$FC,$DC
	.dc.b	$FF,$09,$CB,$F8,$9E,$3F,$E3,$F9,$CB,$FC,$02,$32,$BE,$BD,$F6,$1F
	.dc.b	$C5,$A5,$F2,$7F,$C8,$A7,$38,$75,$8F,$E4,$BF,$61,$FE,$AE,$8F,$D2
	.dc.b	$B7,$0F,$E1,$BF,$5E,$FA,$9D,$FC,$74,$BF,$E3,$D3,$FC,$09,$E3,$1F
	.dc.b	$C9,$A3,$15,$FC,$F2,$FA,$71,$FF,$47,$9C,$BF,$EB,$68,$CD,$FF,$52
	.dc.b	$9C,$BF,$D5,$3E,$BF,$F8,$15,$FF,$20,$6D,$FF,$51,$7E,$5C,$67,$CD
	.dc.b	$FE,$6F,$1A,$AE,$51,$8E,$2C,$BE,$7F,$2F,$D1,$D1,$6B,$DA,$DC,$3E
	.dc.b	$1B,$71,$F5,$3F,$D3,$BE,$C1,$F2,$F4,$F0,$ED,$70,$FE,$23,$3A,$7F
	.dc.b	$A3,$7D,$67,$DC,$C3,$FD,$AF,$39,$C7,$F9,$EC,$E6,$BF,$CD,$D3,$9A
	.dc.b	$FE,$E1,$39,$AF,$F4,$24,$E6,$BF,$BD,$CE,$69,$B3,$E7,$BE,$B1,$F5
	.dc.b	$FC,$1B,$7F,$72,$2E,$3F,$82,$9B,$7F,$3F,$CF,$F9,$BF,$5F,$F3,$9B
	.dc.b	$87,$F6,$1E,$24,$63,$0F,$B0,$DB,$F7,$FB,$1F,$24,$FC,$47,$D8,$5C
	.dc.b	$3E,$3C,$7F,$70,$FA,$C6,$59,$C7,$FB,$0C,$62,$B6,$7F,$71,$71,$1F
	.dc.b	$E0,$CD,$FF,$B8,$3C,$2B,$F0,$BF,$61,$F8,$6F,$3E,$E1,$E9,$6D,$7E
	.dc.b	$3D,$B9,$FE,$3D,$FC,$F7,$9F,$2B,$F4,$7E,$BD,$BF,$D4,$CD,$BF,$D0
	.dc.b	$4D,$AF,$F5,$47,$D8,$7E,$B5,$FE,$AA,$DB,$FC,$F7,$7F,$F8,$6A,$31
	.dc.b	$6A,$E3,$DB,$FE,$14,$E6,$FE,$99,$39,$C3,$FD,$8D,$7C,$2F,$F0,$7E
	.dc.b	$8F,$3C,$38,$36,$3F,$26,$FE,$E7,$EC,$24,$7F,$86,$FD,$57,$B5,$8C
	.dc.b	$7F,$82,$D5,$CB,$FC,$59,$36,$FF,$71,$F4,$FF,$B2,$3F,$67,$F1,$D6
	.dc.b	$3F,$0D,$17,$0F,$DE,$4D,$BF,$54,$3E,$C3,$F9,$C4,$61,$FE,$8D,$1F
	.dc.b	$E4,$87,$37,$FE,$D3,$73,$3B,$F9,$6F,$F2,$1C,$62,$75,$18,$F4,$B3
	.dc.b	$E2,$49,$FF,$03,$D7,$FB,$4D,$B1,$9C,$3F,$CC,$5F,$12,$EA,$FE,$B5
	.dc.b	$1E,$39,$FF,$2C,$CE,$7D,$4F,$0F,$47,$8B,$6B,$8F,$66,$E1,$EB,$6E
	.dc.b	$BE,$F7,$FC,$B7,$EB,$FF,$98,$3D,$57,$D1,$FF,$BC,$63,$F8,$44,$62
	.dc.b	$C7,$F7,$CF,$86,$FC,$24,$66,$3F,$BF,$F9,$18,$BE,$55,$67,$8D,$F4
	.dc.b	$9C,$E5,$FE,$A0,$7C,$17,$BA,$E7,$39,$78,$C6,$6B,$D1,$5C,$2B,$55
	.dc.b	$C0,$FF,$5A,$47,$AF,$E1,$F8,$5F,$17,$97,$CA,$77,$FB,$07,$D1,$BD
	.dc.b	$4E,$79,$FF,$66,$D7,$DD,$5A,$9C,$DF,$1B,$E1,$9F,$80,$F7,$C6,$67
	.dc.b	$CB,$5C,$67,$41,$C5,$FC,$BC,$3C,$19,$FD,$0A,$D6,$3D,$AB,$C6,$D9
	.dc.b	$D5,$F2,$7E,$D6,$3E,$1B,$EF,$39,$7E,$DB,$5F,$F1,$BC,$6E,$FA,$E3
	.dc.b	$E8,$C3,$B5,$7B,$34,$C4,$C8,$EB,$3A,$79,$CE,$6B,$9C,$6A,$BE,$4C
	.dc.b	$E7,$1F,$4D,$F1,$C7,$C9,$97,$C2,$C3,$CF,$39,$26,$1C,$A3,$A1,$EE
	.dc.b	$F7,$7F,$86,$BF,$EF,$FB,$FF,$28,$F5,$7F,$BF,$9F,$A5,$1F,$ED,$E7
	.dc.b	$2E,$E7,$8E,$7F,$80,$6E,$5F,$39,$B3,$F2,$BC,$3F,$F7,$1C,$E7,$0F
	.dc.b	$BF,$5A,$AF,$BC,$DC,$3F,$CF,$51,$A9,$8F,$A5,$FC,$3D,$F4,$8F,$AF
	.dc.b	$5C,$49,$3A,$8E,$38,$79,$E2,$E1,$F8,$4B,$E5,$5E,$29,$CB,$D1,$E4
	.dc.b	$F8,$49,$5C,$DF,$C4,$CF,$2C,$3E,$3C,$E4,$97,$DE,$BA,$9D,$CE,$6E
	.dc.b	$5A,$78,$4A,$7E,$8E,$BB,$61,$EE,$8C,$DF,$C8,$6C,$66,$FF,$6B,$5C
	.dc.b	$30,$F0,$DF,$9B,$0F,$05,$79,$B0,$F0,$35,$F0,$1C,$30,$E9,$3C,$0F
	.dc.b	$B9,$1E,$0B,$F8,$F5,$C6,$FB,$4F,$4A,$E0,$FE,$14,$E9,$5C,$23,$31
	.dc.b	$E0,$3E,$69,$FC,$1B,$C3,$FA,$FF,$61,$FC,$B7,$0F,$E5,$CD,$AF,$E0
	.dc.b	$8D,$AF,$E6,$AB,$E3,$38,$CC,$9F,$E4,$B3,$B4,$ED,$CC,$ED,$FE,$2B
	.dc.b	$1B,$34,$E2,$3F,$8A,$72,$FF,$6C,$9C,$B6,$33,$C1,$E6,$EC,$B7,$E6
	.dc.b	$F2,$EB,$E2,$F4,$EE,$3F,$E9,$E8,$FF,$76,$BF,$E4,$06,$CE,$88,$FF
	.dc.b	$BD,$7A,$DF,$01,$FD,$FB,$3F,$27,$83,$D0,$95,$FE,$F8,$FB,$0F,$E5
	.dc.b	$8A,$76,$F8,$27,$6E,$74,$FE,$65,$5C,$2B,$FC,$16,$FF,$B7,$A7,$A7
	.dc.b	$2F,$B2,$3F,$A1,$C7,$13,$FB,$4B,$B7,$01,$C4,$F4,$E7,$E4,$7C,$8C
	.dc.b	$BD,$BE,$8F,$C8,$FA,$E7,$D8,$52,$63,$FD,$DB,$0F,$F0,$FF,$D8,$7C
	.dc.b	$3D,$3F,$A5,$5F,$0A,$D4,$62,$F8,$C7,$92,$B8,$3F,$E0,$0A,$7C,$77
	.dc.b	$84,$91,$C8,$D7,$B7,$87,$DB,$3C,$1D,$07,$0E,$A3,$F0,$7F,$60,$FB
	.dc.b	$0F,$77,$83,$2B,$8C,$E5,$B3,$C0,$8F,$6B,$7C,$66,$23,$E7,$D7,$16
	.dc.b	$D7,$CB,$9F,$C9,$FD,$83,$C6,$A4,$AF,$4C,$39,$71,$1F,$6A,$38,$3A
	.dc.b	$9E,$73,$D3,$C9,$EB,$D5,$2B,$9C,$65,$EB,$19,$7A,$47,$E9,$BE,$C1
	.dc.b	$F4,$3E,$6E,$F9,$5F,$32,$72,$D9,$E0,$47,$D3,$F4,$55,$FC,$8C,$F0
	.dc.b	$6C,$FC,$29,$FD,$AF,$D7,$FF,$AF,$67,$EC,$AF,$7B,$26,$BA,$64,$EA
	.dc.b	$32,$F0,$8E,$7F,$61,$23,$9B,$DB,$EB,$B1,$1F,$A4,$B4,$E3,$AF,$FC
	.dc.b	$F3,$0A,$FE,$13,$E5,$FF,$94,$67,$3D,$9E,$83,$3B,$4C,$7F,$23,$5B
	.dc.b	$9F,$F1,$63,$63,$FC,$BB,$F5,$EF,$B0,$E1,$1F,$B6,$9D,$3F,$E4,$6B
	.dc.b	$FE,$33,$5B,$8F,$CE,$BF,$9C,$C3,$FD,$1E,$BA,$F6,$70,$EC,$95,$C2
	.dc.b	$32,$F2,$8E,$BE,$7C,$8E,$AF,$83,$B7,$11,$FC,$2A,$D3,$8C,$3C,$35
	.dc.b	$F1,$BD,$CF,$E5,$FB,$7E,$67,$0F,$D9,$B5,$F7,$46,$DF,$8E,$D7,$F8
	.dc.b	$BC,$23,$15,$FE,$AC,$D3,$15,$C0,$EB,$3F,$CF,$BE,$C1,$F5,$F4,$93
	.dc.b	$FC,$13,$EC,$1F,$11,$BF,$4E,$4F,$F0,$8B,$FF,$00,$65,$FE,$BA,$BE
	.dc.b	$73,$FC,$33,$EC,$1F,$43,$FD,$31,$B1,$98,$FE,$9F,$F5,$FF,$5E,$3F
	.dc.b	$98,$CB,$E2,$F8,$12,$3F,$8B,$CB,$E0,$3F,$AB,$FD,$87,$F3,$6E,$9F
	.dc.b	$F3,$A3,$6B,$FD,$4A,$D5,$4F,$A7,$E3,$F3,$32,$7F,$91,$D2,$3C,$B1
	.dc.b	$FD,$93,$E0,$9F,$61,$49,$1F,$CA,$BE,$C3,$F8,$7B,$FD,$2A,$7E,$4B
	.dc.b	$FC,$B9,$AE,$78,$FF,$B7,$CF,$C4,$8F,$EA,$3F,$60,$AF,$EE,$6D,$8F
	.dc.b	$E6,$ED,$7F,$AB,$CB,$C4,$DF,$D4,$C9,$DB,$FD,$F3,$DB,$7B,$BD,$3E
	.dc.b	$66,$57,$47,$5E,$4F,$F8,$0B,$EC,$1F,$07,$49,$5F,$D2,$30,$F8,$6F
	.dc.b	$F4,$AF,$B0,$7C,$B7,$F8,$29,$B1,$98,$C6,$5F,$E0,$EF,$88,$7D,$85
	.dc.b	$25,$7F,$54,$87,$C4,$EC,$91,$97,$FC,$2F,$F5,$DF,$83,$8E,$1E,$AF
	.dc.b	$52,$3F,$D7,$BE,$C1,$F4,$DF,$E1,$E6,$C6,$63,$07,$4F,$90,$91,$92
	.dc.b	$1F,$E0,$6B,$47,$F8,$A3,$EC,$3E,$EF,$8A,$F0,$3F,$D8,$D5,$DA,$F8
	.dc.b	$61,$BE,$B9,$1F,$DC,$B0,$E3,$7F,$EC,$D6,$C7,$FB,$FA,$FF,$30,$EE
	.dc.b	$B8,$BF,$E1,$89,$DA,$9D,$B9,$DF,$E3,$F9,$92,$3C,$11,$E9,$23,$83
	.dc.b	$FC,$0B,$3F,$DE,$1F,$F0,$3F,$D8,$7F,$93,$31,$FF,$71,$36,$01,$C3
	.dc.b	$7A,$7F,$C9,$B1,$82,$72,$E2,$3F,$CA,$38,$BE,$3E,$92,$FF,$C2,$FF
	.dc.b	$60,$AF,$F2,$34,$3E,$C1,$7F,$EF,$53,$94,$7F,$87,$9A,$FF,$8F,$9B
	.dc.b	$C1,$F5,$E3,$D3,$7D,$63,$D7,$5D,$27,$37,$E0,$8C,$47,$83,$0F,$82
	.dc.b	$F8,$F0,$F1,$46,$27,$78,$75,$AD,$1D,$30,$EF,$7C,$22,$9F,$31,$F4
	.dc.b	$D7,$2C,$3C,$51,$E0,$BE,$93,$E4,$8D,$9C,$EB,$B8,$9D,$C7,$1A,$E7
	.dc.b	$3D,$F0,$F4,$4F,$1B,$F0,$61,$E4,$AE,$57,$C1,$F8,$46,$6B,$C2,$FB
	.dc.b	$4F,$0E,$1F,$05,$FF,$3F,$B5,$F1,$57,$FA,$09,$E0,$6B,$C9,$7C,$36
	.dc.b	$63,$D5,$18,$78,$57,$F5,$09,$D3,$63,$80,$F8,$EA,$D3,$2E,$08,$FC
	.dc.b	$8B,$FF,$5D,$6D,$F0,$63,$F3,$F3,$F0,$E7,$31,$F3,$8C,$9C,$75,$FC
	.dc.b	$76,$9F,$88,$8C,$BF,$8A,$9E,$2E,$63,$C8,$E6,$78,$E7,$F1,$6F,$C1
	.dc.b	$18,$AF,$F5,$4C,$63,$0F,$34,$72,$AF,$F6,$2D,$6E,$3F,$2F,$5B,$7D
	.dc.b	$06,$CF,$F4,$C4,$6F,$0F,$69,$FE,$BB,$7F,$50,$FE,$A6,$39,$9E,$FB
	.dc.b	$FF,$46,$C7,$07,$FD,$AB,$F1,$DC,$1F,$19,$A7,$13,$FC,$2B,$EC,$3F
	.dc.b	$81,$4F,$FA,$59,$B1,$FE,$B6,$7F,$87,$C6,$77,$E3,$29,$DD,$F9,$BC
	.dc.b	$BF,$14,$F5,$AE,$F3,$FE,$0B,$C7,$8D,$FF,$46,$C3,$FD,$55,$2F,$B0
	.dc.b	$7C,$EE,$BA,$38,$C6,$2B,$CD,$18,$9D,$47,$29,$D6,$5E,$9F,$2B,$B9
	.dc.b	$E2,$FB,$0F,$E9,$75,$EF,$8C,$BE,$28,$FF,$5F,$B9,$31,$ED,$6A,$BF
	.dc.b	$DB,$99,$2B,$71,$DE,$B6,$7F,$B0,$3E,$4F,$FC,$06,$DE,$AF,$93,$F5
	.dc.b	$EF,$B0,$F9,$75,$AB,$FF,$72,$B7,$1F,$C0,$7D,$83,$E1,$F7,$8F,$D2
	.dc.b	$8C,$65,$FE,$CE,$6E,$1F,$EE,$EC,$BF,$C0,$F3,$9B,$F0,$F8,$FC,$7E
	.dc.b	$57,$ED,$FD,$87,$FB,$22,$B9,$46,$3E,$36,$47,$FB,$6A,$7E,$C6,$3F
	.dc.b	$EF,$A7,$83,$79,$39,$F4,$7E,$1B,$C4,$FF,$08,$DF,$18,$C7,$95,$F8
	.dc.b	$AF,$B0,$7B,$BF,$86,$F2,$FE,$AE,$4E,$3F,$D0,$56,$B1,$FC,$3D,$6A
	.dc.b	$7F,$49,$1F,$E6,$23,$F1,$7F,$61,$FE,$4A,$7F,$DE,$26,$6F,$EC,$C6
	.dc.b	$5E,$D1,$FD,$33,$E6,$3F,$3F,$1A,$D3,$F3,$F1,$8B,$FD,$9D,$FF,$9E
	.dc.b	$2F,$FD,$FB,$2F,$96,$FC,$FF,$C6,$BE,$6F,$8C,$61,$E7,$DB,$A9,$B7
	.dc.b	$FA,$27,$FD,$9C,$FF,$7E,$C7,$F8,$07,$88,$F4,$7F,$57,$F6,$0F,$1B
	.dc.b	$F6,$73,$9B,$FD,$5C,$E5,$E4,$7F,$12,$DB,$F6,$7F,$61,$F2,$F8,$3E
	.dc.b	$09,$FC,$BA,$33,$C7,$FB,$EF,$AE,$FF,$33,$5F,$ED,$BE,$C3,$E5,$6B
	.dc.b	$FC,$4E,$73,$87,$E9,$27,$35,$FD,$DA,$7F,$D5,$0F,$EF,$BE,$C3,$E5
	.dc.b	$39,$F8,$9F,$E0,$B1,$8B,$FF,$0E,$D3,$DC,$FE,$13,$F6,$0F,$96,$F2
	.dc.b	$46,$1F,$E2,$7F,$3F,$FE,$6A,$CB,$F8,$5E,$5F,$C5,$1F,$F0,$3C,$67
	.dc.b	$1F,$DC,$CE,$5F,$F1,$5F,$D7,$BE,$C3,$F8,$BC,$FF,$B6,$6F,$E4,$D6
	.dc.b	$A7,$3F,$27,$FC,$8B,$EC,$3F,$D5,$55,$AA,$46,$1F,$E0,$FF,$60,$FA
	.dc.b	$9E,$1F,$1D,$2F,$F9,$0C,$FF,$03,$79,$5F,$1E,$3F,$E8,$33,$9C,$3E
	.dc.b	$B9,$DA,$FF,$96,$C6,$67,$FA,$35,$71,$BF,$C5,$4E,$5F,$C6,$7D,$87
	.dc.b	$F0,$6A,$FE,$9A,$D8,$CE,$FF,$CE,$23,$18,$FE,$CD,$B4,$69,$F5,$6B
	.dc.b	$F9,$DB,$AB,$FE,$23,$5F,$0B,$9B,$CF,$E5,$7A,$1B,$7F,$09,$B3,$B8
	.dc.b	$F9,$9F,$5F,$FE,$FE,$6F,$CB,$F9,$75,$FE,$A7,$39,$AF,$8A,$DC,$FF
	.dc.b	$B4,$FD,$6F,$E0,$5F,$ED,$51,$9C,$3F,$5D,$39,$7F,$C9,$B1,$9A,$FE
	.dc.b	$25,$39,$7F,$CE,$31,$9A,$4E,$56,$6F,$9E,$FA,$A7,$D6,$F0,$6E,$3F
	.dc.b	$80,$8B,$8F,$E1,$A6,$DF,$D4,$F3,$BE,$97,$D7,$FD,$36,$E1,$F3,$9B
	.dc.b	$7F,$D5,$38,$11,$8B,$F7,$FB,$3F,$18,$FC,$3F,$D8,$5C,$3E,$6C,$7F
	.dc.b	$76,$FA,$C6,$59,$CB,$FA,$FC,$62,$B6,$7F,$78,$71,$1F,$42,$B9,$57
	.dc.b	$E2,$7E,$C3,$F1,$5E,$6D,$C3,$F0,$AD,$BF,$43,$72,$FC,$9B,$F9,$CF
	.dc.b	$0E,$57,$E9,$BD,$7B,$7F,$AC,$9B,$7F,$A6,$9B,$5F,$E8,$6F,$B0,$FD
	.dc.b	$83,$FB,$56,$DF,$C7,$EE,$FF,$0D,$C6,$2F,$FC,$5C,$DA,$FC,$2F,$C5
	.dc.b	$6B,$B7,$FC,$37,$CD,$FD,$46,$7B,$46,$27,$51,$FE,$1C,$9D,$63,$FD
	.dc.b	$6F,$99,$F6,$1F,$B8,$3F,$C4,$BE,$A3,$DB,$C6,$1F,$C1,$26,$DF,$F8
	.dc.b	$42,$B5,$39,$3F,$5F,$F4,$DF,$B7,$3F,$71,$F1,$D6,$32,$4A,$FC,$34
	.dc.b	$65,$FE,$17,$F6,$1F,$CE,$B7,$FF,$27,$3F,$D2,$63,$3F,$2F,$FE,$40
	.dc.b	$8C,$4E,$A3,$1E,$8E,$5B,$87,$F2,$3A,$FE,$13,$66,$9F,$EC,$D8,$D7
	.dc.b	$BB,$D4,$FF,$2A,$FC,$53,$AF,$FA,$E5,$F9,$A7,$31,$8C,$BC,$FD,$1E
	.dc.b	$5D,$7C,$CD,$C7,$C4,$DC,$3D,$ED,$AF,$F3,$67,$AF,$FE,$60,$F5,$1F
	.dc.b	$47,$FE,$CE,$8F,$E1,$91,$8B,$FF,$78,$C6,$E7,$5E,$37,$FB,$76,$B5
	.dc.b	$5F,$E0,$3F,$88,$C5,$6B,$2F,$F4,$DB,$D6,$33,$3D,$2E,$BF,$E0,$B6
	.dc.b	$E1,$E2,$AE,$17,$CA,$B8,$35,$E0,$D9,$FF,$7B,$B6,$B9,$FC,$4F,$4F
	.dc.b	$93,$EF,$F6,$0F,$A2,$7A,$7C,$67,$31,$8C,$BF,$D7,$99,$FC,$59,$CD
	.dc.b	$6F,$2F,$16,$1C,$E2,$E1,$C7,$5F,$8B,$39,$C7,$99,$F3,$30,$E6,$E0
	.dc.b	$46,$E7,$8D,$73,$C7,$C5,$8F,$6C,$66,$1C,$70,$E1,$5A,$8C,$5F,$14
	.dc.b	$76,$70,$DA,$F6,$3C,$1E,$B8,$78,$9C,$B6,$34,$47,$F0,$7A,$7F,$84
	.dc.b	$5D,$12,$FC,$33,$DC,$F8,$13,$B8,$CE,$1E,$A9,$F3,$BF,$E4,$DE,$AF
	.dc.b	$F7,$B3,$63,$FD,$F8,$FC,$D7,$A9,$F2,$BD,$DF,$E1,$CE,$9B,$87,$D4
	.dc.b	$6D,$F8,$1B,$3E,$0F,$0F,$FD,$B7,$39,$C3,$EF,$D6,$AB,$B4,$7E,$0A
	.dc.b	$B7,$1E,$89,$D6,$1E,$7C,$3C,$17,$BA,$E6,$C7,$ED,$57,$0B,$E7,$1D
	.dc.b	$8C,$CE,$EB,$F1,$B3,$AB,$ED,$87,$9E,$F6,$7E,$16,$75,$87,$90,$98
	.dc.b	$78,$E7,$78,$79,$9B,$7F,$52,$F8,$D7,$86,$3B,$57,$8A,$B0,$7C,$19
	.dc.b	$AF,$E7,$2B,$8B,$D0,$D1,$E5,$37,$87,$A9,$CD,$75,$78,$1F,$0D,$F3
	.dc.b	$19,$2D,$F9,$09,$1F,$6C,$EB,$87,$C0,$7C,$07,$4A,$F0,$3D,$67,$E0
	.dc.b	$3C,$30,$F2,$E1,$EA,$C3,$BC,$F9,$AB,$6F,$48,$CB,$CC,$F1,$19,$8D
	.dc.b	$C7,$A5,$FD,$0F,$87,$F8,$0F,$B0,$FE,$5B,$3F,$CB,$9B,$3F,$AA,$6D
	.dc.b	$7E,$F1,$7C,$6F,$19,$93,$FC,$9E,$76,$9D,$B9,$9D,$BF,$C4,$E3,$66
	.dc.b	$9C,$47,$EE,$B9,$1C,$1E,$38,$C3,$67,$FD,$0D,$5F,$EB,$96,$F9,$5A
	.dc.b	$7D,$1E,$95,$7C,$8D,$9F,$F1,$EC,$70,$21,$FE,$84,$72,$FF,$AF,$3A
	.dc.b	$DF,$01,$E4,$8F,$0F,$81,$CC,$8F,$FA,$67,$EC,$3F,$83,$E9,$B8,$F8
	.dc.b	$93,$A7,$B3,$E5,$9D,$E7,$FC,$69,$FF,$71,$4F,$3E,$5D,$1B,$E5,$91
	.dc.b	$DD,$AF,$9F,$1F,$F3,$43,$FE,$AB,$FA,$E7,$D8,$7C,$36,$C6,$63,$E3
	.dc.b	$7D,$87,$9F,$4F,$45,$6A,$BF,$D4,$ED,$8C,$D7,$0C,$FA,$7C,$7C,$BF
	.dc.b	$03,$DF,$DA,$C4,$72,$8B,$7D,$A7,$EB,$7D,$83,$EC,$3C,$DD,$F3,$0F
	.dc.b	$F6,$24,$E5,$B3,$C0,$8F,$BF,$E3,$6B,$FE,$BE,$9E,$0D,$9F,$85,$3F
	.dc.b	$85,$FB,$07,$8B,$49,$5F,$ED,$38,$72,$62,$3F,$D9,$31,$C1,$D4,$F3
	.dc.b	$9E,$9E,$3F,$5E,$89,$5C,$E3,$2F,$58,$CB,$D2,$3F,$31,$F6,$0F,$A0
	.dc.b	$F3,$76,$CA,$FF,$73,$4E,$5B,$3C,$08,$FA,$7E,$86,$BF,$EE,$09,$E0
	.dc.b	$D9,$F8,$53,$FA,$7F,$AF,$FC,$4C,$FF,$B9,$97,$B9,$93,$5E,$EC,$9F
	.dc.b	$F7,$C4,$E9,$B3,$97,$97,$D8,$75,$23,$C3,$EB,$AB,$A3,$9C,$78,$E3
	.dc.b	$F7,$5E,$57,$F9,$13,$9F,$F5,$98,$7E,$5F,$A7,$FC,$02,$DC,$3F,$C2
	.dc.b	$75,$B9,$FF,$02,$B6,$33,$B7,$43,$F8,$37,$D7,$BE,$C3,$13,$FE,$1C
	.dc.b	$C7,$F7,$57,$C6,$3F,$C7,$D7,$C2,$B8,$F5,$F2,$EB,$95,$C7,$2F,$F0
	.dc.b	$AE,$4A,$DC,$E4,$FF,$14,$C2,$B8,$C3,$2F,$E1,$53,$C6,$38,$05,$E7
	.dc.b	$7A,$89,$1A,$87,$66,$23,$A6,$3E,$4A,$C5,$F6,$AF,$9F,$EE,$7F,$29
	.dc.b	$DF,$F4,$38,$78,$1A,$F2,$3E,$7C,$70,$7C,$BB,$23,$11,$B1,$39,$77
	.dc.b	$1A,$BF,$9E,$F8,$4F,$E7,$5F,$5E,$FB,$0A,$4A,$F8,$7F,$60,$F8,$7D
	.dc.b	$F9,$F2,$7E,$3D,$A3,$FA,$25,$73,$9F,$97,$F6,$0F,$A3,$FE,$90,$D8
	.dc.b	$CC,$7F,$4B,$FB,$0F,$C3,$47,$0F,$8A,$D9,$1F,$AB,$0F,$7B,$FA,$97
	.dc.b	$D7,$FF,$3C,$9F,$84,$DD,$FF,$A3,$37,$93,$FC,$B7,$6F,$B0,$E3,$F3
	.dc.b	$33,$0F,$C0,$DA,$3C,$91,$FD,$7F,$E0,$1F,$61,$49,$1F,$88,$FA,$FF
	.dc.b	$E3,$0F,$D3,$9F,$90,$FE,$31,$AE,$78,$FF,$B5,$4F,$C3,$8F,$B5,$F6
	.dc.b	$0A,$FE,$DC,$D8,$FC,$AB,$5F,$C0,$4B,$C4,$DF,$D5,$49,$DB,$FD,$DF
	.dc.b	$DA,$7B,$9D,$3E,$5A,$57,$47,$5E,$4F,$F7,$DF,$B0,$7C,$15,$25,$7E
	.dc.b	$9B,$0F,$86,$FE,$A3,$EC,$1F,$2B,$FE,$05,$6C,$66,$31,$97,$F8,$27
	.dc.b	$E7,$DF,$45,$49,$5F,$B2,$87,$C4,$6C,$91,$97,$FC,$25,$F5,$DF,$83
	.dc.b	$8E,$1E,$AF,$52,$3F,$BC,$FB,$07,$D2,$FF,$86,$5B,$19,$8C,$1D,$3E
	.dc.b	$4A,$46,$48,$7E,$BE,$D1,$FE,$23,$FB,$0F,$BB,$E0,$7F,$85,$67,$AD
	.dc.b	$6E,$F5,$D5,$23,$F8,$7D,$F0,$F0,$3F,$D5,$ED,$8F,$F6,$0D,$FE,$5D
	.dc.b	$DD,$71,$7F,$80,$4E,$D4,$ED,$CE,$FF,$1B,$CC,$91,$E0,$8F,$41,$1C
	.dc.b	$1F,$C8,$A5,$EE,$7F,$8F,$FE,$C3,$FC,$99,$A7,$F8,$C5,$B9,$FF,$B1
	.dc.b	$9B,$7F,$ED,$85,$D5,$C7,$9A,$4F,$F2,$E9,$DA,$76,$FF,$32,$CD,$F1
	.dc.b	$DF,$E4,$F6,$C6,$63,$FC,$A5,$F6,$1D,$B2,$FF,$73,$FD,$87,$F3,$4A
	.dc.b	$FF,$74,$DF,$07,$F9,$33,$67,$FC,$94,$DC,$FF,$DB,$F7,$CE,$97,$CD
	.dc.b	$46,$63,$15,$D5,$B8,$7F,$BA,$1E,$13,$B9,$EB,$9F,$F5,$1B,$EB,$1F
	.dc.b	$EE,$E3,$91,$6F,$93,$C3,$1F,$6B,$67,$89,$EE,$9D,$B4,$D8,$3D,$35
	.dc.b	$E1,$C3,$CA,$DA,$F0,$4F,$03,$FA,$ED,$74,$78,$DE,$70,$F6,$47,$73
	.dc.b	$C6,$70,$35,$7D,$5F,$14,$7C,$D7,$FE,$00,$B5,$F6,$18,$8F,$8C,$EB
	.dc.b	$C9,$7C,$57,$AE,$30,$D7,$EC,$C6,$9C,$1F,$61,$C3,$81,$F1,$B5,$F8
	.dc.b	$CD,$72,$65,$C1,$F1,$70,$FF,$5C,$E9,$F1,$E7,$89,$F0,$A3,$2F,$C4
	.dc.b	$33,$A7,$E2,$E3,$2F,$E3,$63,$3A,$7F,$AD,$F6,$FF,$52,$4E,$67,$C6
	.dc.b	$F1,$8C,$CF,$BB,$0F,$24,$62,$BE,$24,$63,$0F,$34,$72,$AE,$D1,$89
	.dc.b	$D4,$F9,$CD,$1F,$E3,$7A,$E8,$FF,$8A,$5F,$F5,$C6,$1F,$E9,$D6,$D7
	.dc.b	$8E,$7E,$14,$61,$E5,$1F,$E2,$E9,$7C,$9E,$67,$A4,$E6,$31,$F6,$1F
	.dc.b	$C0,$A3,$FC,$6A,$D9,$FF,$10,$3F,$9D,$8C,$EF,$C0,$F0,$5F,$F8,$1F
	.dc.b	$0C,$4F,$F9,$16,$30,$F4,$E6,$DF,$D8,$3E,$77,$C3,$18,$6C,$F1,$9D
	.dc.b	$3C,$4C,$47,$A3,$CA,$91,$DF,$EC,$3F,$A3,$BE,$A8,$CE,$9F,$E3,$FC
	.dc.b	$55,$FE,$5B,$AD,$1C,$7D,$CD,$CF,$D0,$9D,$57,$C2,$BF,$7C,$62,$7F
	.dc.b	$DC,$DF,$27,$FE,$00,$C3,$FC,$BA,$DB,$FF,$24,$7D,$73,$EC,$3E,$94
	.dc.b	$E6,$F8,$F6,$7D,$5F,$B0,$7C,$46,$A3,$0C,$CB,$FC,$D3,$8F,$F8,$09
	.dc.b	$FF,$35,$D7,$07,$3E,$3F,$3F,$96,$FC,$07,$D8,$7F,$4E,$9D,$C6,$3C
	.dc.b	$09,$1F,$E7,$49,$FE,$F2,$D8,$E3,$C9,$CB,$D7,$FC,$57,$89,$FE,$0C
	.dc.b	$AE,$11,$8F,$2B,$F1,$DF,$61,$FE,$57,$9F,$C1,$79,$5F,$53,$93,$FC
	.dc.b	$D5,$39,$7E,$A4,$67,$89,$5F,$C4,$4F,$F2,$E1,$F8,$BF,$B0,$FF,$22
	.dc.b	$BF,$E8,$A3,$35,$F7,$63,$2F,$F9,$DF,$E5,$7F,$30,$E7,$3F,$D1,$46
	.dc.b	$2F,$F5,$B5,$FE,$71,$BF,$F4,$AC,$BE,$5B,$F3,$BF,$18,$F9,$BE,$31
	.dc.b	$87,$9F,$5E,$A6,$DF,$E8,$5F,$F6,$8B,$FE,$A0,$F0,$DE,$87,$EA,$3E
	.dc.b	$C3,$F6,$B3,$98,$C7,$8B,$FB,$7A,$F2,$6F,$F9,$9F,$B0,$F9,$1B,$FC
	.dc.b	$68,$FE,$57,$C1,$FB,$29,$CE,$3F,$CE,$7E,$BB,$FC,$BC,$FE,$CF,$EC
	.dc.b	$3E,$3E,$1C,$70,$56,$9F,$D3,$4E,$AB,$97,$C4,$7E,$FB,$EC,$3E,$44
	.dc.b	$60,$FD,$F3,$8A,$FE,$C5,$6F,$6F,$F8,$2F,$D8,$3E,$5B,$F8,$94,$E7
	.dc.b	$0F,$E0,$D3,$97,$FD,$8B,$F3,$DF,$E6,$47,$F8,$56,$5F,$C4,$F4,$F4
	.dc.b	$4E,$72,$FD,$E4,$E6,$BF,$AC,$7D,$83,$EC,$3F,$8A,$CF,$FB,$56,$7F
	.dc.b	$DA,$33,$97,$AF,$C9,$FF,$25,$FB,$07,$C5,$D3,$F8,$EF,$D4,$3E,$AB
	.dc.b	$17,$FE,$DA,$AD,$BF,$C7,$E7,$F7,$CE,$6B,$8F,$9D,$C3,$4F,$E2,$55
	.dc.b	$A7,$FD,$99,$7C,$9C,$CF,$F2,$CF,$AF,$FD,$E4,$7F,$19,$6E,$FF,$D1
	.dc.b	$1F,$91,$19,$F3,$78,$13,$2F,$E6,$53,$AA,$E1,$3E,$CF,$2F,$DB,$D3
	.dc.b	$6F,$D4,$DC,$3C,$ED,$BF,$EA,$79,$FE,$51,$FC,$57,$D8,$7F,$BE,$AF
	.dc.b	$FD,$FA,$DB,$F4,$6B,$FD,$46,$31,$5C,$3E,$5F,$FA,$4F,$D6,$7D,$B8
	.dc.b	$C6,$1F,$D2,$63,$0F,$F8,$4E,$B8,$BF,$E2,$F9,$4E,$B0,$FF,$19,$56
	.dc.b	$94,$EB,$E1,$FE,$DF,$CF,$4C,$3E,$C7,$87,$6F,$EA,$CD,$BF,$7C,$DA
	.dc.b	$F0,$70,$3D,$0F,$AD,$F5,$FF,$5D,$B3,$F4,$1B,$8F,$D9,$6E,$1F,$23
	.dc.b	$D8,$F8,$87,$E3,$BE,$C2,$DF,$CA,$8F,$ED,$BF,$58,$C7,$38,$FF,$5B
	.dc.b	$8C,$56,$CF,$EE,$0E,$23,$E0,$C7,$C9,$AE,$75,$F8,$DF,$B0,$FC,$77
	.dc.b	$9F,$6B,$F0,$ED,$BF,$C8,$37,$2F,$C8,$3F,$A4,$F9,$99,$5F,$A1,$F6
	.dc.b	$6D,$FE,$7A,$6D,$FE,$22,$6D,$7F,$93,$7E,$C3,$F5,$8F,$DB,$31,$DB
	.dc.b	$FA,$1F,$03,$FC,$31,$3A,$C7,$D6,$7F,$9B,$DC,$61,$FE,$26,$E6,$FE
	.dc.b	$95,$87,$23,$F1,$5F,$17,$CB,$1F,$C7,$B6,$32,$7E,$E7,$EC,$3F,$7E
	.dc.b	$7F,$87,$3D,$77,$BB,$89,$D1,$33,$FF,$06,$4E,$A3,$19,$7E,$D7,$E9
	.dc.b	$BF,$74,$7E,$EF,$E3,$EC,$FE,$D2,$2D,$FE,$F6,$6D,$7F,$35,$FB,$0F
	.dc.b	$E6,$8D,$F4,$BF,$C7,$51,$8C,$3F,$CE,$CF,$F4,$AB,$F8,$A1,$D3,$FE
	.dc.b	$95,$75,$5D,$7E,$5F,$9D,$F4,$F1,$66,$1F,$4F,$0E,$1F,$19,$87,$AE
	.dc.b	$3F,$C4,$A7,$FD,$41,$F1,$8E,$BF,$EB,$B1,$C6,$BC,$58,$7F,$95,$6B
	.dc.b	$5E,$B7,$AB,$A7,$D7,$BD,$C3,$86,$48,$C6,$3E,$06,$ED,$F1,$5F,$F2
	.dc.b	$E7,$B5,$FE,$57,$F5,$DF,$49,$FE,$99,$8F,$E2,$51,$8B,$FF,$5B,$46
	.dc.b	$E7,$5E,$3F,$F6,$BA,$D5,$7F,$80,$7E,$27,$15,$AC,$FF,$D3,$2F,$58
	.dc.b	$CC,$F4,$BA,$FF,$82,$5B,$87,$8A,$B8,$5F,$2A,$E0,$D7,$83,$65,$F1
	.dc.b	$7C,$AF,$9B,$D3,$C8,$EF,$F6,$0F,$A3,$7A,$BC,$67,$31,$8C,$BF,$D6
	.dc.b	$79,$FC,$59,$CD,$6F,$3F,$16,$1C,$E2,$E1,$C7,$5F,$8B,$39,$C7,$99
	.dc.b	$F3,$30,$E6,$E0,$46,$E7,$8D,$73,$C7,$C5,$8F,$6C,$66,$1C,$70,$E1
	.dc.b	$5A,$8C,$5F,$17,$FD,$E1,$3A,$6C,$65,$FF,$1D,$63,$E4,$C7,$C3,$87
	.dc.b	$67,$EF,$E9,$FE,$0B,$73,$5D,$A7,$58,$F8,$67,$27,$C2,$F7,$FF,$82
	.dc.b	$D7,$FB,$8B,$2F,$F2,$37,$67,$FB,$85,$F8,$D1,$FE,$E9,$72,$EE,$78
	.dc.b	$D7,$3A,$FA,$31,$8B,$FF,$79,$5B,$C4,$FF,$7A,$5E,$F0,$F8,$D7,$B7
	.dc.b	$E1,$B5,$FA,$33,$C2,$F9,$CE,$6F,$C7,$7C,$CF,$BB,$5B,$BE,$78,$79
	.dc.b	$08,$FD,$58,$D3,$F7,$AF,$A1,$27,$F1,$73,$AB,$EF,$87,$8A,$79,$E1
	.dc.b	$E0,$C2,$CE,$5F,$6C,$F2,$72,$79,$0C,$9E,$33,$27,$8D,$B5,$F2,$62
	.dc.b	$E1,$E5,$78,$1C,$AF,$E1,$5F,$85,$F2,$46,$63,$04,$7F,$23,$87,$B6
	.dc.b	$BA,$57,$29,$C9,$E0,$C3,$C7,$5B,$71,$1E,$29,$E0,$FA,$4F,$01,$C8
	.dc.b	$A6,$5D,$E1,$ED,$8C,$BF,$39,$F8,$D1,$EA,$37,$7E,$6C,$3F,$1B,$87
	.dc.b	$D5,$3C,$95,$C9,$C3,$F4,$4F,$34,$7E,$53,$C4,$FC,$0F,$D8,$7F,$28
	.dc.b	$C3,$F9,$4B,$65,$52,$30,$DF,$8D,$E3,$32,$3F,$91,$4E,$92,$9E,$E6
	.dc.b	$E3,$F8,$74,$6C,$D3,$CA,$3C,$2F,$F6,$48,$CB,$63,$3C,$BF,$EA,$39
	.dc.b	$76,$E3,$8B,$8C,$7F,$8F,$EF,$FA,$E3,$FE,$9B,$6F,$B7,$B7,$DD,$E6
	.dc.b	$57,$D2,$D9,$E5,$1E,$38,$E1,$D8,$F8,$2F,$F5,$A6,$5E,$3F,$0B,$81
	.dc.b	$2F,$FB,$67,$D8,$7B,$1F,$64,$65,$A6,$8F,$14,$7F,$B6,$DE,$3C,$B8
	.dc.b	$33,$CD,$23,$AB,$5F,$F1,$7E,$3E,$5F,$47,$D9,$F5,$EF,$AF,$A4,$AF
	.dc.b	$F0,$07,$D8,$7A,$B8,$7E,$8D,$6E,$7F,$C0,$8D,$7F,$C2,$ED,$D7,$FD
	.dc.b	$A9,$5C,$67,$F1,$32,$F9,$0F,$09,$23,$9F,$BB,$5E,$71,$6F,$FD,$BD
	.dc.b	$4A,$FA,$9F,$60,$FB,$0F,$5F,$83,$31,$FF,$0C,$4E,$5B,$3C,$08,$FA
	.dc.b	$BE,$3A,$BF,$E3,$09,$E0,$D9,$F8,$F3,$F8,$3F,$B0,$FF,$0E,$4F,$F8
	.dc.b	$95,$AF,$F8,$E1,$B9,$7F,$BD,$57,$27,$1F,$1A,$57,$17,$A4,$75,$F2
	.dc.b	$BB,$75,$4A,$ED,$19,$7C,$11,$97,$C1,$1F,$97,$FB,$07,$D0,$79,$FB
	.dc.b	$E5,$7F,$8E,$E7,$2D,$9E,$04,$7D,$9F,$43,$5F,$F2,$9C,$F0,$6C,$FC
	.dc.b	$39,$FD,$37,$D7,$BA,$2F,$BD,$2F,$FC,$95,$39,$6C,$E5,$FF,$2F,$C2
	.dc.b	$79,$CF,$4F,$B0,$95,$C9,$EF,$F5,$D8,$8E,$31,$78,$5D,$5F,$C2,$31
	.dc.b	$5F,$EE,$FC,$CF,$F2,$2D,$6B,$0F,$E3,$B5,$F9,$74,$D9,$DD,$FF,$12
	.dc.b	$6C,$FF,$24,$6B,$FE,$39,$DB,$F8,$47,$D7,$BE,$C3,$C1,$12,$7F,$CD
	.dc.b	$95,$C2,$B5,$1F,$E6,$F7,$F4,$CD,$EA,$E6,$74,$EC,$91,$92,$1F,$E8
	.dc.b	$89,$E2,$DF,$3F,$C6,$47,$B7,$65,$74,$78,$63,$D3,$1F,$1F,$DD,$FE
	.dc.b	$45,$BF,$E6,$AB,$FD,$08,$D7,$FD,$2F,$92,$78,$47,$B9,$AF,$FA,$72
	.dc.b	$74,$E4,$87,$28,$EE,$7F,$31,$FA,$F7,$D8,$52,$4F,$FA,$57,$EC,$1F
	.dc.b	$13,$BF,$42,$4F,$FA,$66,$FF,$B6,$72,$7D,$F8,$5F,$47,$FD,$3D,$F6
	.dc.b	$0F,$A3,$FE,$80,$D8,$CC,$7F,$43,$FB,$0F,$AF,$1F,$EB,$69,$7C,$5F
	.dc.b	$02,$47,$FA,$A6,$5F,$01,$FD,$2F,$EB,$FF,$9D,$4F,$FA,$F9,$B9,$79
	.dc.b	$5B,$D3,$FD,$82,$DF,$61,$C7,$E6,$E6,$1F,$EB,$5B,$47,$96,$3F,$AC
	.dc.b	$7C,$1B,$EC,$29,$23,$FD,$79,$F5,$FF,$C5,$B2,$FE,$DB,$5F,$2A,$7F
	.dc.b	$D8,$0D,$7F,$DA,$6D,$CF,$E5,$57,$C5,$9F,$F6,$BF,$D8,$2B,$FB,$43
	.dc.b	$63,$FD,$94,$D7,$FD,$BB,$2F,$13,$7F,$0D,$27,$6F,$F7,$1F,$6D,$EE
	.dc.b	$F4,$F9,$A9,$5D,$1D,$79,$3F,$DD,$FE,$C1,$F0,$74,$95,$FE,$DB,$C3
	.dc.b	$E2,$3F,$ED,$DF,$B0,$7C,$B7,$F7,$F6,$C6,$63,$19,$7F,$80,$FE,$25
	.dc.b	$F4,$74,$95,$FE,$E7,$87,$C5,$6C,$91,$97,$FC,$19,$F5,$DF,$84,$8E
	.dc.b	$1E,$AF,$52,$3F,$EF,$1F,$B0,$7D,$37,$F8,$4D,$B1,$98,$C1,$D3,$E4
	.dc.b	$24,$64,$83,$A3,$F4,$58,$27,$FC,$3D,$F6,$1F,$77,$AB,$FC,$5F,$3D
	.dc.b	$AF,$84,$63,$0D,$F8,$F2,$7F,$C8,$B8,$FE,$B7,$0E,$51,$FE,$48,$6A
	.dc.b	$C3,$F3,$6F,$0C,$3F,$C7,$B3,$A2,$5F,$C6,$EF,$F2,$F0,$7C,$8F,$32
	.dc.b	$61,$C4,$ED,$1A,$70,$2F,$F1,$2E,$1A,$FB,$0F,$F2,$EE,$2F,$73,$FC
	.dc.b	$81,$F6,$1F,$E4,$ED,$3F,$84,$37,$4F,$F2,$9B,$6F,$FC,$CA,$BA,$F8
	.dc.b	$F3,$C9,$F1,$CE,$D3,$B7,$CF,$A3,$E3,$FF,$CA,$0D,$8C,$C7,$F9,$4F
	.dc.b	$EC,$3B,$65,$FE,$74,$FB,$05,$FF,$99,$8E,$17,$ED,$33,$7D,$0E,$59
	.dc.b	$7E,$0E,$15,$C9,$FF,$3F,$E1,$D9,$CD,$F5,$BF,$F3,$9D,$6E,$7A,$9E
	.dc.b	$39,$FF,$3B,$D7,$19,$D3,$63,$94,$EC,$FF,$46,$BE,$89,$DD,$F9,$5B
	.dc.b	$86,$EB,$B9,$E4,$38,$E1,$E4,$BF,$0D,$F1,$6D,$71,$7B,$1E,$2A,$E4
	.dc.b	$30,$F3,$3F,$0A,$B9,$3D,$8E,$57,$DD,$D5,$F9,$6B,$E4,$47,$F9,$F5
	.dc.b	$AF,$88,$FF,$40,$1D,$EB,$E9,$3B,$F2,$9F,$0D,$A8,$F2,$46,$1E,$35
	.dc.b	$FD,$5E,$74,$D8,$E0,$3E,$3A,$B4,$CB,$82,$3F,$0B,$0F,$F5,$B6,$DF
	.dc.b	$3E,$3F,$D8,$33,$F4,$67,$31,$F2,$0C,$9C,$76,$FF,$16,$6B,$F8,$58
	.dc.b	$CB,$F8,$79,$E2,$E6,$3C,$8E,$67,$8E,$BF,$1A,$FC,$11,$8A,$FF,$54
	.dc.b	$46,$30,$F3,$47,$2A,$FF,$61,$D6,$E3,$FD,$93,$5B,$7D,$06,$CF,$E3
	.dc.b	$31,$BC,$3D,$A7,$FA,$E9,$FF,$23,$3F,$E0,$98,$F5,$D7,$6B,$FE,$1F
	.dc.b	$3C,$1F,$F6,$9F,$C7,$70,$75,$5F,$B3,$9F,$F6,$AC,$7F,$16,$FB,$0F
	.dc.b	$E0,$0D,$DF,$85,$7F,$BC,$27,$FD,$E3,$39,$24,$7F,$19,$AF,$F0,$6E
	.dc.b	$3C,$21,$5D,$1D,$79,$9E,$57,$A7,$47,$6B,$7D,$83,$E7,$39,$61,$FD
	.dc.b	$43,$2F,$E5,$D5,$A9,$CD,$F1,$BE,$15,$EB,$8C,$4F,$2F,$32,$4F,$2F
	.dc.b	$B0,$FE,$A0,$FC,$18,$CE,$5E,$B9,$FF,$63,$4E,$4E,$1E,$DE,$B2,$FA
	.dc.b	$73,$AA,$FE,$3D,$9F,$F8,$8A,$F6,$7F,$42,$F9,$5F,$F0,$4B,$7B,$3E
	.dc.b	$37,$D8,$3E,$8F,$E4,$56,$B0,$FE,$50,$DC,$BF,$C9,$1F,$60,$F8,$8D
	.dc.b	$E3,$F3,$E3,$19,$7F,$50,$6E,$1F,$CC,$72,$FF,$06,$CE,$6F,$C3,$E4
	.dc.b	$F1,$F9,$5F,$B1,$F6,$1F,$D4,$6B,$94,$63,$C1,$91,$FD,$72,$7E,$A6
	.dc.b	$3F,$CF,$5E,$0D,$E4,$E7,$CD,$F8,$3F,$17,$FC,$2F,$7C,$63,$1E,$5F
	.dc.b	$E1,$FE,$C1,$EE,$FE,$0F,$CC,$FA,$B9,$7F,$49,$9C,$E1,$F8,$29,$CF
	.dc.b	$1F,$F9,$AE,$3F,$CD,$07,$E1,$BE,$C3,$FC,$A0,$FF,$72,$33,$7F,$97
	.dc.b	$8C,$BF,$D5,$7E,$63,$F3,$AE,$74,$FC,$D4,$62,$FF,$63,$5F,$E7,$EB
	.dc.b	$FE,$FB,$2F,$96,$FC,$D7,$C6,$BE,$6B,$8C,$61,$E7,$DD,$A9,$B7,$F9
	.dc.b	$C7,$FD,$A0,$FF,$82,$3C,$57,$A5,$FA,$5F,$B0,$FD,$BE,$DF,$DA,$E7
	.dc.b	$35,$FA,$49,$CE,$BF,$E8,$D9,$F3,$57,$ED,$7E,$C3,$E4,$BC,$77,$FD
	.dc.b	$84,$62,$BF,$D2,$9F,$5D,$FE,$69,$C7,$F5,$FF,$61,$F2,$27,$3E,$FF
	.dc.b	$ED,$E3,$12,$C6,$5F,$FA,$C6,$3F,$6F,$F6,$1F,$25,$D4,$A7,$4F,$F0
	.dc.b	$99,$D3,$CB,$DD,$FE,$09,$F6,$0F,$96,$FE,$19,$69,$CC,$7F,$0F,$F9
	.dc.b	$EF,$F3,$7E,$3F,$C1,$BE,$33,$F8,$86,$BF,$EA,$F9,$CE,$5F,$C5,$E7
	.dc.b	$35,$E4,$FA,$F7,$D8,$7F,$13,$97,$C9,$FF,$1B,$9C,$D7,$58,$FE,$3D
	.dc.b	$F6,$1F,$EB,$08,$FF,$1F,$B7,$0F,$E4,$9F,$60,$FA,$9D,$FC,$74,$BF
	.dc.b	$E3,$73,$FC,$05,$E3,$1F,$C8,$E3,$15,$FE,$29,$BE,$9A,$FF,$42,$9C
	.dc.b	$BF,$EC,$B8,$CD,$FF,$98,$A7,$2F,$F9,$93,$EB,$FF,$81,$D7,$F4,$A6
	.dc.b	$C6,$77,$FE,$65,$18,$C3,$FC,$EC,$DA,$34,$FA,$95,$FC,$D1,$D5,$FF
	.dc.b	$0E,$AF,$83,$CF,$EB,$F2,$BD,$8D,$BF,$82,$D9,$DC,$7F,$9A,$3E,$C1
	.dc.b	$F2,$FF,$D2,$DB,$1D,$F5,$C4,$C8,$FE,$8B,$F5,$9F,$7B,$4F,$F7,$75
	.dc.b	$F0,$C3,$FA,$B4,$E6,$BF,$27,$39,$AF,$ED,$B3,$9A,$FC,$E4,$E6,$BF
	.dc.b	$D1,$B3,$9A,$66,$F9,$EF,$B0,$7D,$9F,$02,$DF,$DE,$8B,$8F,$E1,$26
	.dc.b	$DF,$D3,$F3,$BE,$8F,$D7,$FD,$26,$E1,$F3,$5B,$7F,$DB,$F8,$11,$8B
	.dc.b	$F7,$FB,$3F,$1C,$FC,$57,$D8,$5C,$3E,$64,$7F,$6F,$FA,$D6,$59,$CB
	.dc.b	$FB,$D4,$62,$B6,$7F,$70,$71,$1F,$E7,$BD,$FF,$B7,$BC,$2B,$F1,$3F
	.dc.b	$61,$F8,$AF,$32,$E1,$F8,$56,$D7,$A5,$B9,$FE,$4D,$FD,$07,$9D,$2B
	.dc.b	$F4,$BE,$BD,$BF,$D5,$CD,$BF,$D1,$4D,$AF,$E1,$BF,$61,$FA,$F7,$FD
	.dc.b	$62,$DB,$F9,$1D,$DF,$E1,$58,$C5,$AB,$8F,$67,$F8,$4F,$9B,$FA,$6C
	.dc.b	$60,$FC,$6F,$C6,$71,$9D,$36,$32,$7E,$F3,$EC,$3F,$74,$7F,$87,$7D
	.dc.b	$47,$B7,$89,$D1,$34,$FF,$12,$C6,$5F,$BD,$F4,$DF,$B4,$3F,$6B,$F2
	.dc.b	$16,$BF,$61,$16,$7F,$09,$37,$4F,$64,$7F,$35,$FB,$0F,$E7,$8F,$FA
	.dc.b	$CD,$FE,$8F,$7F,$CD,$1C,$CF,$09,$46,$9E,$5E,$6F,$F8,$F6,$31,$1A
	.dc.b	$8C,$7C,$36,$7D,$D9,$87,$FB,$8A,$FF,$DA,$8D,$8C,$DF,$FB,$87,$E2
	.dc.b	$9D,$5F,$D7,$23,$CD,$3F,$E5,$29,$CF,$A9,$E1,$E8,$F1,$6D,$71,$EC
	.dc.b	$DC,$3D,$CD,$D7,$D9,$9A,$3F,$CA,$DE,$BF,$F9,$57,$D4,$7D,$1F,$F5
	.dc.b	$68,$FE,$11,$18,$C3,$FA,$F4,$6E,$75,$E3,$7C,$EA,$D5,$7F,$80,$FE
	.dc.b	$23,$15,$AC,$FF,$D3,$2F,$58,$CC,$F4,$BA,$FF,$82,$DB,$87,$8A,$B8
	.dc.b	$5F,$2A,$E0,$D7,$83,$67,$FD,$E6,$DA,$E7,$F1,$3D,$3E,$53,$BF,$D8
	.dc.b	$3E,$8D,$EA,$71,$9C,$C6,$32,$FF,$5D,$67,$F1,$67,$35,$BC,$BC,$58
	.dc.b	$73,$8B,$87,$1D,$7E,$2C,$E7,$1E,$67,$CC,$C3,$9B,$81,$1B,$9E,$35
	.dc.b	$CF,$1F,$16,$3D,$B1,$98,$71,$C3,$85,$6A,$31,$7C,$5F,$F0,$C4,$E9
	.dc.b	$B1,$97,$FC,$79,$8F,$93,$1F,$0E,$1D,$9F,$BF,$A7,$F8,$31,$CD,$78
	.dc.b	$7D,$DF,$DF,$BF,$EE,$EC,$7F,$C8,$9D,$5F,$EE,$B7,$E1,$47,$F8,$90
	.dc.b	$F7,$CF,$07,$11,$C1,$E9,$3F,$3A,$7A,$D7,$CC,$F0,$FF,$D9,$BA,$7F
	.dc.b	$8B,$B0,$E1,$3F,$42,$B7,$39,$6E,$1F,$0A,$F8,$BF,$12,$32,$F9,$30
	.dc.b	$F0,$1F,$7A,$77,$19,$3C,$A4,$BF,$23,$C0,$C5,$F8,$6B,$94,$66,$7C
	.dc.b	$D1,$67,$D5,$5D,$B0,$F5,$C7,$1B,$E6,$7A,$4E,$07,$A6,$AB,$F8,$AA
	.dc.b	$E5,$87,$B2,$F9,$CE,$CF,$C4,$4E,$AB,$8C,$EA,$79,$18,$AE,$47,$3A
	.dc.b	$E1,$1E,$43,$E2,$61,$F0,$9C,$57,$32,$4F,$09,$F8,$33,$A8,$C6,$1E
	.dc.b	$F9,$C9,$CC,$AE,$49,$3F,$0C,$EF,$7E,$6C,$3E,$7C,$61,$F6,$DF,$68
	.dc.b	$F8,$57,$D1,$E4,$D7,$D7,$7D,$9F,$86,$7C,$83,$E2,$57,$69,$F3,$4F
	.dc.b	$DA,$9C,$3F,$95,$F0,$FF,$13,$F6,$1F,$CA,$2B,$F9,$4B,$67,$FD,$0E
	.dc.b	$DA,$FF,$47,$2F,$8D,$E3,$32,$7F,$91,$4E,$D3,$B7,$33,$B7,$F8,$74
	.dc.b	$6C,$D3,$88,$FF,$4B,$F3,$7F,$63,$9C,$B6,$33,$C3,$E6,$EC,$B7,$E6
	.dc.b	$F2,$EB,$E2,$F4,$EE,$3F,$E9,$B8,$FF,$54,$63,$D6,$B2,$48,$CE,$1F
	.dc.b	$0A,$3A,$F6,$3E,$03,$FB,$B5,$F9,$7C,$1E,$A4,$C3,$89,$FE,$B5,$FB
	.dc.b	$0F,$E1,$D3,$B8,$C9,$F1,$A7,$82,$8F,$F0,$14,$F0,$C7,$FD,$72,$FF
	.dc.b	$B7,$E7,$8B,$FD,$15,$E1,$1A,$E6,$F8,$33,$B9,$C9,$E3,$E6,$C9,$1F
	.dc.b	$3E,$5E,$DF,$47,$E4,$FD,$73,$EC,$29,$2B,$D3,$F6,$1F,$0B,$3F,$F6
	.dc.b	$6D,$6E,$7D,$AD,$7C,$21,$C6,$9F,$E1,$FA,$E5,$3F,$ED,$B9,$7C,$87
	.dc.b	$88,$91,$D3,$DB,$AF,$49,$DB,$E0,$9F,$4C,$FE,$0B,$EC,$1F,$61,$EC
	.dc.b	$F0,$65,$FD,$29,$CB,$67,$81,$1F,$5F,$C6,$D7,$D7,$3C,$1B,$3F,$22
	.dc.b	$7F,$23,$F6,$1F,$E1,$F9,$FB,$ED,$7E,$1B,$74,$FF,$20,$2E,$4E,$3E
	.dc.b	$34,$AE,$2F,$48,$EB,$E5,$76,$EA,$95,$DA,$32,$F8,$23,$2F,$82,$3F
	.dc.b	$4B,$F6,$0F,$A1,$F3,$F7,$CA,$FC,$AC,$E5,$B3,$C0,$8F,$B3,$E8,$AB
	.dc.b	$F7,$A7,$83,$67,$E1,$CF,$ED,$3E,$BF,$FD,$71,$3F,$A2,$5E,$F6,$4D
	.dc.b	$74,$C9,$D4,$65,$E1,$1C,$FE,$C2,$47,$37,$B7,$D7,$62,$3F,$23,$69
	.dc.b	$C7,$0F,$F9,$BE,$15,$FC,$23,$CC,$FF,$25,$CE,$7A,$FD,$06,$76,$98
	.dc.b	$FE,$EA,$B7,$28,$91,$87,$1F,$5E,$FB,$0E,$31,$FA,$49,$D3,$FE,$43
	.dc.b	$BF,$E3,$35,$B8,$FE,$0E,$FE,$79,$FD,$5A,$EA,$D9,$C3,$AE,$4F,$08
	.dc.b	$CB,$CA,$3A,$F9,$F2,$3A,$BE,$0E,$CC,$47,$68,$BA,$7F,$A5,$70,$57
	.dc.b	$C7,$F7,$3F,$97,$ED,$F9,$9C,$3F,$8E,$B5,$F7,$46,$DF,$90,$D7,$F8
	.dc.b	$2C,$23,$17,$FE,$A3,$D3,$15,$C0,$EB,$3F,$CF,$BE,$BD,$F6,$14,$93
	.dc.b	$FC,$BB,$EC,$1F,$11,$BF,$46,$4F,$F3,$2B,$47,$F4,$8A,$E7,$3F,$CD
	.dc.b	$BE,$C1,$F4,$7F,$D3,$1B,$19,$8F,$E9,$FF,$61,$F5,$E3,$F8,$FC,$BE
	.dc.b	$2F,$81,$23,$FA,$0C,$BD,$FF,$EA,$FF,$61,$FC,$DF,$83,$FD,$84,$DE
	.dc.b	$9F,$C6,$5A,$30,$FE,$3E,$DF,$55,$CB,$E6,$64,$FF,$4E,$A4,$79,$A3
	.dc.b	$FB,$2F,$C1,$3E,$C2,$92,$3F,$AA,$FD,$7F,$F1,$5C,$BC,$35,$F2,$A7
	.dc.b	$FA,$CB,$5F,$E7,$AD,$CF,$FD,$B5,$5F,$1A,$7F,$A0,$FD,$82,$BF,$BA
	.dc.b	$B6,$3F,$B1,$B5,$FE,$8F,$2F,$13,$7F,$55,$27,$6F,$F7,$EF,$6D,$EE
	.dc.b	$F4,$F9,$79,$5D,$1D,$79,$3F,$E0,$4F,$B0,$7C,$1D,$25,$7F,$74,$C3
	.dc.b	$E2,$3F,$DD,$BE,$C1,$F2,$DF,$E0,$C6,$C6,$63,$19,$7F,$84,$3E,$25
	.dc.b	$F4,$54,$95,$FE,$00,$87,$C5,$6C,$91,$97,$FC,$33,$F5,$DF,$85,$8E
	.dc.b	$1E,$AF,$52,$3F,$E0,$AF,$B0,$7D,$2F,$F8,$81,$B1,$98,$C1,$D3,$E4
	.dc.b	$24,$64,$87,$F7,$6B,$47,$F8,$AB,$EC,$3E,$E7,$57,$E4,$A7,$AD,$6E
	.dc.b	$F5,$E3,$C8,$FF,$0E,$DF,$07,$F9,$23,$55,$7E,$61,$DC,$F1,$7F,$C0
	.dc.b	$F3,$B9,$D4,$63,$6F,$8B,$E6,$48,$EF,$1E,$72,$38,$3F,$C5,$F2,$F7
	.dc.b	$3F,$C8,$BF,$61,$FE,$4C,$CF,$F5,$AD,$AF,$E7,$4D,$4B,$A7,$8F,$2C
	.dc.b	$9F,$F1,$DC,$EC,$D3,$FE,$3E,$C9,$F1,$FF,$E5,$46,$C6,$63,$FC,$AF
	.dc.b	$F6,$1D,$7C,$0F,$E9,$5F,$61,$FC,$BA,$BF,$A6,$5F,$07,$FC,$56,$D9
	.dc.b	$67,$FC,$1A,$30,$FF,$08,$BE,$94,$BE,$8A,$33,$18,$AE,$CD,$C3,$F8
	.dc.b	$8B,$C2,$77,$3D,$73,$FF,$2C,$5F,$58,$F1,$5F,$06,$C7,$B1,$E1,$8D
	.dc.b	$8F,$ED,$17,$D1,$F7,$D7,$16,$9B,$3D,$17,$E0,$C3,$C8,$DC,$3B,$CF
	.dc.b	$8C,$E2,$53,$8E,$1C,$F0,$F0,$47,$F5,$CC,$3C,$0D,$35,$7D,$5F,$3B
	.dc.b	$F0,$1F,$E6,$A7,$42,$94,$CC,$FD,$26,$F9,$4F,$86,$DC,$7D,$38,$C3
	.dc.b	$CA,$7F,$A8,$CE,$9B,$1C,$07,$C7,$56,$99,$70,$47,$E1,$63,$FE,$BB
	.dc.b	$D3,$E6,$47,$F8,$66,$7E,$74,$E6,$3E,$41,$93,$8E,$DF,$90,$72,$78
	.dc.b	$35,$FC,$6C,$F1,$7F,$1F,$1C,$5C,$CF,$17,$31,$C7,$5F,$F6,$0C,$E5
	.dc.b	$F9,$77,$E4,$8C,$4F,$D0,$8C,$61,$EA,$8E,$47,$FA,$FA,$76,$FF,$6A
	.dc.b	$9D,$BE,$A9,$FF,$59,$E1,$EF,$3D,$91,$B3,$47,$FA,$69,$D4,$7F,$B5
	.dc.b	$BE,$3B,$E1,$47,$FB,$12,$EC,$E4,$FF,$5F,$FD,$77,$F0,$26,$EF,$AC
	.dc.b	$3F,$C9,$13,$FE,$49,$9C,$92,$3F,$D7,$37,$FE,$07,$BE,$10,$AE,$8E
	.dc.b	$BC,$CF,$2B,$D3,$9F,$B5,$BE,$C1,$F3,$BC,$B0,$FE,$93,$97,$FB,$62
	.dc.b	$B5,$39,$BE,$37,$C2,$BD,$71,$89,$E5,$E5,$C9,$E5,$F6,$1F,$D2,$9F
	.dc.b	$83,$19,$CB,$D7,$3F,$EB,$E9,$C9,$C3,$DB,$D6,$5F,$32,$75,$5F,$EC
	.dc.b	$AC,$FF,$C3,$77,$B3,$FD,$D7,$F2,$7F,$E0,$26,$F5,$FC,$6F,$AE,$7D
	.dc.b	$87,$C8,$AD,$61,$FE,$D4,$6E,$5F,$80,$FB,$07,$C4,$6F,$1F,$9F,$18
	.dc.b	$CB,$FD,$F2,$DC,$3F,$DB,$B9,$7F,$81,$A7,$37,$E1,$F2,$38,$FC,$AF
	.dc.b	$D8,$FB,$0F,$EA,$55,$CA,$31,$E0,$C8,$FF,$45,$69,$F9,$2A,$FA,$B9
	.dc.b	$7F,$BA,$9E,$2D,$E4,$E9,$CD,$F8,$4F,$13,$FC,$23,$87,$23,$FC,$93
	.dc.b	$7F,$E4,$F7,$17,$FE,$97,$9D,$BF,$EF,$6F,$B0,$FF,$31,$CF,$E0,$FC
	.dc.b	$9F,$B5,$39,$F7,$3F,$17,$18,$CF,$F4,$F5,$FE,$66,$3F,$15,$F6,$1F
	.dc.b	$E4,$D7,$BD,$FF,$56,$F1,$3F,$27,$39,$85,$7E,$C3,$E6,$3F,$45,$1F
	.dc.b	$A0,$9F,$F3,$CD,$F2,$9C,$D7,$EC,$27,$35,$CA,$BF,$CC,$F4,$F9,$6F
	.dc.b	$D0,$FC,$6B,$E6,$BF,$53,$39,$AF,$D0,$4E,$5E,$DD,$DC,$23,$6D,$8F
	.dc.b	$5F,$88,$F4,$3F,$5B,$F6,$0F,$1B,$F6,$B3,$9A,$FD,$64,$E5,$E3,$C7
	.dc.b	$FB,$4F,$B0,$F9,$FB,$7A,$4F,$E6,$51,$9E,$3F,$DF,$FD,$6F,$F9,$A6
	.dc.b	$FF,$6D,$F6,$1F,$3E,$F8,$09,$D3,$FC,$22,$35,$C7,$FC,$72,$7F,$D5
	.dc.b	$D5,$FB,$FF,$B0,$F9,$6F,$FA,$3A,$33,$7F,$C4,$27,$2F,$FB,$76,$1E
	.dc.b	$DF,$F0,$CF,$B0,$7C,$B7,$9A,$30,$FF,$16,$F9,$FF,$F3,$7E,$3F,$C3
	.dc.b	$7E,$33,$F8,$BE,$7F,$CB,$A7,$39,$7F,$1E,$9C,$D7,$FA,$83,$EB,$DF
	.dc.b	$61,$FC,$6A,$7F,$DC,$13,$F8,$C9,$CB,$D7,$E4,$FF,$92,$FD,$87,$FA
	.dc.b	$C6,$3F,$22,$DC,$3F,$95,$FD,$83,$EA,$77,$F1,$D2,$FF,$91,$4F,$F0
	.dc.b	$87,$8C,$7F,$2A,$8C,$57,$D0,$BE,$9A,$FF,$42,$9C,$BF,$D2,$A3,$37
	.dc.b	$FB,$C9,$CB,$FB,$EF,$B0,$FE,$15,$5F,$D2,$9B,$19,$DF,$F9,$CC,$63
	.dc.b	$0F,$D3,$B6,$8D,$3E,$9D,$7F,$3B,$75,$7F,$C5,$6B,$E0,$F3,$7A,$FC
	.dc.b	$AF,$63,$6F,$E0,$B6,$77,$1F,$BA,$FA,$F7,$CC,$7F,$52,$6C,$77,$D7
	.dc.b	$13,$23,$FA,$67,$D6,$7D,$ED,$3F,$B6,$DF,$0C,$3F,$AE,$CE,$6B,$FD
	.dc.b	$BF,$39,$AF,$ED,$B3,$9A,$09,$D5,$7F,$28,$9D,$53,$37,$CF,$7D,$93
	.dc.b	$ED,$78,$56,$FE,$FC,$5C,$7F,$0B,$36,$FE,$A7,$9D,$F4,$BE,$BF,$E9
	.dc.b	$B7,$0F,$9C,$DB,$FE,$EF,$C0,$8C,$5F,$C0,$F6,$7E,$41,$F8,$BF,$B0
	.dc.b	$B8,$7C,$D8,$FE,$EF,$F5,$AC,$B3,$97,$F8,$0A,$31,$5B,$3F,$BC,$38
	.dc.b	$8F,$E0,$7B,$FF,$77,$78,$57,$E2,$BE,$C3,$F1,$7E,$75,$C3,$F0,$CD
	.dc.b	$AF,$4B,$73,$FC,$A3,$FA,$1F,$0E,$57,$E9,$BD,$7B,$7F,$AC,$9B,$7F
	.dc.b	$A3,$9B,$5F,$62,$9F,$61,$FB,$18,$FE,$8A,$D8,$F9,$3D,$DF,$E1,$E8
	.dc.b	$C5,$FF,$8D,$1B,$5F,$7F,$E2,$75,$DB,$FE,$1E,$E7,$FE,$A9,$3D,$EB
	.dc.b	$78,$FF,$4E,$9C,$BF,$8A,$8C,$E0,$9C,$BF,$89,$FB,$0F,$E0,$07,$F8
	.dc.b	$AB,$D5,$7B,$78,$C7,$F8,$3C,$DC,$7F,$C6,$53,$6F,$F7,$9F,$4D,$FB
	.dc.b	$A3,$F7,$7F,$1B,$70,$FD,$B4,$5B,$FC,$4C,$DA,$FF,$29,$FD,$87,$F3
	.dc.b	$C6,$FA,$1F,$E4,$28,$C6,$1F,$35,$FE,$9D,$7F,$D1,$1C,$BF,$A8,$73
	.dc.b	$5D,$3E,$5F,$95,$F3,$F7,$66,$1F,$DE,$2F,$FD,$C0,$D8,$CE,$1F,$D5
	.dc.b	$FE,$29,$D5,$FD,$7E,$3C,$13,$FE,$5D,$9C,$FA,$9E,$7E,$8F,$46,$D7
	.dc.b	$1E,$CD,$C3,$D0,$DD,$7E,$0B,$FE,$60,$F6,$3F,$CB,$FE,$AB,$E8,$FF
	.dc.b	$19,$1F,$C2,$E3,$17,$FC,$1A,$37,$3A,$F1,$BF,$D3,$15,$AA,$FF,$03
	.dc.b	$7C,$46,$2B,$59,$7F,$A8,$DE,$B1,$99,$E9,$75,$FF,$07,$37,$0F,$15
	.dc.b	$70,$BE,$55,$C1,$AF,$06,$CB,$E2,$79,$5F,$37,$A7,$C9,$F7,$FB,$07
	.dc.b	$D1,$3D,$4E,$33,$98,$C6,$5F,$EC,$4C,$FE,$2C,$E6,$B7,$9F,$8B,$0E
	.dc.b	$71,$70,$E3,$AF,$C5,$9C,$E3,$CC,$F9,$98,$73,$70,$23,$73,$C6,$B9
	.dc.b	$E3,$E2,$C7,$B6,$33,$0E,$38,$70,$AD,$46,$2F,$8B,$FD,$02,$74,$D8
	.dc.b	$CB,$FE,$3F,$C7,$C9,$8F,$A7,$DE,$FD,$BC,$FF,$BD,$2F,$FC,$85,$D5
	.dc.b	$FE,$F0,$7D,$B1,$FE,$FA,$72,$EE,$78,$D7,$3A,$F9,$11,$8C,$7F,$8E
	.dc.b	$5B,$C3,$FF,$67,$46,$27,$58,$7A,$5A,$FC,$76,$D7,$18,$EB,$71,$F8
	.dc.b	$B3,$AB,$EB,$1C,$E7,$58,$79,$63,$E5,$CE,$9E,$D3,$AC,$3C,$8F,$E0
	.dc.b	$67,$58,$7C,$0B,$E4,$4A,$F6,$CE,$B0,$F3,$92,$F9,$61,$D4,$F2,$1E
	.dc.b	$87,$2D,$25,$F5,$7E,$E5,$75,$9E,$51,$E8,$3D,$66,$30,$F3,$61,$EA
	.dc.b	$8C,$9D,$1C,$5F,$58,$FC,$A5,$F5,$74,$79,$A3,$F3,$D1,$97,$85,$76
	.dc.b	$9F,$29,$E1,$C3,$E0,$3E,$09,$DC,$61,$E7,$1B,$C3,$D1,$7D,$70,$F2
	.dc.b	$92,$7A,$9E,$43,$E0,$57,$8A,$78,$47,$9E,$78,$3E,$C3,$B1,$C7,$0F
	.dc.b	$C5,$3E,$58,$DC,$7A,$E3,$81,$93,$C3,$5E,$87,$F6,$7E,$1F,$E2,$7E
	.dc.b	$C3,$F9,$34,$FF,$27,$6D,$FF,$7B,$6D,$7F,$8D,$17,$C6,$F1,$99,$3F
	.dc.b	$C8,$27,$69,$DB,$99,$DB,$FC,$32,$36,$69,$C4,$7F,$82,$F9,$5C,$1E
	.dc.b	$38,$C3,$67,$FD,$A3,$5F,$EB,$56,$F9,$7A,$7D,$1E,$9D,$7C,$8D,$9F
	.dc.b	$F1,$D4,$70,$21,$FE,$B0,$72,$FF,$63,$F8,$0F,$1B,$E0,$F0,$79,$11
	.dc.b	$FF,$71,$FD,$87,$F8,$8D,$E5,$19,$69,$A3,$C9,$1F,$ED,$B7,$8F,$2E
	.dc.b	$0C,$F4,$48,$EA,$D7,$C9,$E8,$FB,$7E,$B9,$F6,$14,$93,$FE,$F6,$FB
	.dc.b	$0F,$4B,$FE,$CC,$9D,$BF,$EF,$A6,$B9,$9F,$91,$3C,$38,$78,$D3,$E4
	.dc.b	$3C,$04,$BE,$5E,$DD,$7B,$45,$AE,$33,$F4,$FE,$C1,$F6,$1E,$9E,$F9
	.dc.b	$87,$FA,$0A,$FC,$F5,$A8,$C7,$2F,$F9,$3E,$31,$1F,$EE,$66,$CF,$0B
	.dc.b	$F8,$9F,$1B,$DD,$95,$C8,$FF,$47,$D7,$0B,$FF,$2D,$66,$BF,$C4,$7D
	.dc.b	$83,$C4,$A4,$C3,$90,$87,$26,$63,$FD,$F3,$1C,$5D,$B5,$CC,$76,$F2
	.dc.b	$3B,$F4,$49,$DC,$65,$E3,$19,$7A,$C7,$E7,$7E,$C1,$F4,$1E,$8E,$D9
	.dc.b	$5E,$69,$CB,$67,$81,$1F,$2F,$D0,$D5,$C3,$F8,$8A,$E3,$18,$DF,$FD
	.dc.b	$01,$18,$9F,$C6,$E8,$BF,$D8,$7D,$7F,$FA,$DB,$0F,$8E,$BD,$EC,$9A
	.dc.b	$E9,$93,$A8,$CB,$C2,$39,$FD,$84,$8E,$6F,$6F,$AE,$C4,$79,$4F,$F6
	.dc.b	$5D,$A7,$39,$7E,$8A,$15,$FC,$17,$CA,$FF,$24,$4E,$75,$F4,$99,$DE
	.dc.b	$63,$F8,$0A,$DC,$FF,$A8,$5B,$1F,$ED,$AF,$AF,$7D,$87,$08,$F2,$1F
	.dc.b	$E3,$EB,$C6,$1F,$B4,$AD,$CF,$E3,$1F,$D7,$37,$AF,$F8,$D4,$1B,$D9
	.dc.b	$27,$44,$3B,$CF,$26,$F9,$BC,$88,$F7,$ED,$AE,$8F,$14,$78,$23,$E3
	.dc.b	$7B,$9F,$CA,$F6,$FC,$FD,$7E,$89,$AF,$BB,$54,$F0,$8F,$90,$D7,$E1
	.dc.b	$CE,$9C,$90,$E5,$1D,$CF,$E7,$5F,$5E,$FB,$0A,$49,$FD,$9F,$D8,$3E
	.dc.b	$23,$7E,$7C,$9F,$DB,$DF,$F6,$AE,$0F,$E4,$77,$CE,$7F,$77,$F6,$0F
	.dc.b	$97,$FE,$90,$D8,$CC,$7F,$4B,$FB,$0F,$AF,$1F,$72,$5F,$17,$C0,$91
	.dc.b	$FC,$1A,$5F,$01,$FD,$4F,$EB,$FF,$A0,$4F,$E1,$1B,$97,$F3,$66,$F4
	.dc.b	$FF,$BB,$6D,$F6,$1C,$7E,$66,$61,$FC,$4A,$D1,$E5,$8F,$EC,$1F,$04
	.dc.b	$FB,$0A,$48,$FE,$33,$F5,$FF,$C4,$F3,$FE,$8B,$5F,$26,$7F,$8E,$B5
	.dc.b	$FC,$CB,$63,$FA,$5C,$FC,$59,$FC,$EF,$D8,$2B,$FB,$83,$63,$F9,$3B
	.dc.b	$5F,$D1,$CF,$F0,$77,$3F,$53,$C0,$8F,$F2,$EF,$65,$ED,$F4,$F9,$79
	.dc.b	$3D,$1F,$EF,$D3,$FC,$4F,$81,$33,$EC,$1F,$03,$FE,$03,$6C,$7F,$3A
	.dc.b	$6B,$FB,$6A,$F9,$0F,$F3,$EF,$B0,$7C,$B7,$F8,$2D,$B1,$98,$C6,$5F
	.dc.b	$E0,$FF,$86,$7D,$85,$25,$7F,$48,$87,$C3,$EC,$91,$97,$FC,$31,$F5
	.dc.b	$DF,$88,$8E,$1E,$B7,$52,$3F,$D4,$BE,$C1,$F4,$DF,$E1,$F6,$C6,$63
	.dc.b	$07,$4F,$93,$91,$92,$1F,$C5,$AD,$1F,$E2,$9F,$B0,$FC,$07,$17,$CC
	.dc.b	$A5,$76,$78,$5E,$FC,$59,$1F,$D9,$AF,$8B,$F5,$1A,$F2,$ED,$DB,$FC
	.dc.b	$9A,$32,$6A,$FF,$43,$1C,$F7,$F8,$BE,$54,$8C,$10,$E5,$1E,$93,$E5
	.dc.b	$FB,$9F,$E4,$3F,$B0,$FF,$28,$76,$FF,$89,$1B,$AF,$B5,$B7,$F8,$A5
	.dc.b	$D5,$C7,$9E,$4F,$F8,$0A,$76,$9D,$BF,$E0,$6C,$9F,$1B,$FE,$53,$6C
	.dc.b	$66,$3F,$CA,$DF,$61,$DB,$2F,$CC,$FD,$82,$FE,$81,$C2,$FF,$A5,$B6
	.dc.b	$7F,$C7,$AD,$CF,$EB,$5F,$3A,$5F,$37,$F4,$B1,$98,$C4,$77,$C3,$EF
	.dc.b	$46,$2F,$C0,$EA,$B8,$47,$6C,$FF,$C0,$F7,$DA,$3C,$F7,$C9,$B1,$EC
	.dc.b	$78,$63,$63,$C7,$5C,$B0,$F0,$5F,$19,$FD,$DD,$F8,$8E,$2F,$73,$8D
	.dc.b	$F3,$7D,$45,$8E,$0E,$8E,$03,$0F,$5B,$DA,$BC,$0F,$23,$71,$A7,$86
	.dc.b	$1F,$01,$FF,$40,$47,$CF,$6C,$65,$F3,$9A,$F2,$1F,$21,$FE,$87,$8F
	.dc.b	$A7,$18,$71,$5F,$D4,$E7,$51,$E7,$9D,$38,$F8,$DC,$BF,$C8,$1A,$9A
	.dc.b	$47,$E0,$E5,$FE,$BC,$D3,$E4,$47,$13,$3B,$7F,$A8,$35,$FC,$2C,$65
	.dc.b	$FC,$3D,$71,$8F,$96,$67,$5F,$F5,$D5,$71,$7D,$D8,$7D,$5A,$E1,$7D
	.dc.b	$2F,$CC,$78,$30,$F1,$B8,$8E,$F8,$73,$8F,$F5,$75,$F3,$8C,$47,$F1
	.dc.b	$58,$C5,$72,$6C,$75,$3F,$C6,$77,$E9,$3F,$D4,$AF,$A4,$FF,$49,$57
	.dc.b	$FB,$02,$3F,$D4,$4D,$AF,$F6,$B7,$C9,$DA,$DC,$7F,$A1,$29,$39,$97
	.dc.b	$D7,$7F,$02,$AF,$F4,$4B,$63,$FC,$8A,$FF,$58,$8C,$EF,$C6,$35,$7E
	.dc.b	$3F,$2B,$D1,$5F,$CC,$E7,$2F,$FA,$A6,$33,$7F,$D8,$A7,$FA,$2C,$FF
	.dc.b	$A8,$69,$F6,$0F,$9D,$ED,$D5,$EB,$9C,$C6,$2B,$85,$6D,$F4,$D6,$E3
	.dc.b	$18,$7A,$BC,$AF,$41,$CF,$EC,$3F,$A5,$DF,$C0,$8C,$BE,$38,$FF,$66
	.dc.b	$39,$31,$EC,$EA,$BF,$CF,$99,$2B,$71,$DE,$B6,$7F,$AD,$3E,$53,$FC
	.dc.b	$0E,$DE,$AF,$A5,$F5,$CF,$B0,$FA,$95,$AB,$FF,$3A,$B7,$1F,$F7,$3F
	.dc.b	$D8,$3E,$27,$78,$FD,$B8,$C6,$5F,$EC,$26,$E1,$FE,$7E,$CB,$FC,$0F
	.dc.b	$39,$BF,$0F,$93,$C7,$E5,$BF,$09,$F6,$1F,$EC,$FA,$E5,$18,$F0,$64
	.dc.b	$7F,$B3,$A7,$EC,$E3,$FE,$8A,$78,$37,$8F,$9F,$37,$E3,$FC,$5F,$F0
	.dc.b	$8D,$F1,$8C,$79,$7F,$93,$FB,$07,$B7,$F7,$FC,$CF,$AB,$97,$FA,$5A
	.dc.b	$73,$87,$E1,$E7,$3C,$68,$FF,$01,$1F,$E6,$73,$F1,$9F,$61,$FE,$4F
	.dc.b	$7F,$DC,$A6,$6F,$EC,$C6,$5F,$F5,$1F,$CC,$7E,$7D,$CE,$9F,$A7,$8C
	.dc.b	$5F,$EC,$6B,$FC,$F9,$7F,$EF,$09,$7C,$B7,$E5,$FE,$2D,$F3,$7C,$63
	.dc.b	$0F,$3E,$DD,$4D,$BF,$D1,$3F,$ED,$37,$FD,$F7,$E2,$3D,$2F,$D1,$7D
	.dc.b	$87,$EE,$A7,$31,$8F,$17,$F7,$95,$E4,$DF,$F2,$BF,$61,$F2,$1F,$F5
	.dc.b	$FC,$67,$7F,$E1,$53,$97,$E5,$4F,$F9,$5B,$EB,$BF,$CD,$18,$FE,$D7
	.dc.b	$EC,$3E,$45,$6A,$95,$A7,$F4,$73,$AA,$FF,$1A,$4F,$FA,$C1,$FD,$F7
	.dc.b	$D8,$7C,$A7,$3F,$0F,$FB,$D8,$C5,$F8,$7D,$CF,$E0,$1F,$60,$F9,$7F
	.dc.b	$E3,$18,$A7,$31,$FC,$6F,$E7,$BF,$CD,$79,$7F,$0B,$CB,$F8,$8E,$3F
	.dc.b	$E7,$09,$CB,$FB,$88,$CA,$FA,$F7,$D8,$7F,$0F,$9F,$F6,$E5,$FF,$9E
	.dc.b	$36,$F8,$15,$A8,$ED,$F2,$3F,$CA,$FE,$C3,$FD,$5F,$1B,$3F,$CA,$8E
	.dc.b	$31,$FE,$0F,$F6,$0F,$AB,$E1,$F1,$92,$FF,$91,$4F,$F0,$37,$95,$7F
	.dc.b	$A2,$2F,$96,$5F,$D0,$6B,$53,$FE,$FA,$9C,$D7,$9E,$73,$5F,$5E,$73
	.dc.b	$48,$4E,$AB,$DD,$F6,$1F,$C2,$27,$FA,$73,$63,$3B,$FF,$33,$8C,$5F
	.dc.b	$36,$D1,$A7,$D9,$AF,$E7,$CE,$AF,$F8,$9D,$7C,$4F,$03,$F9,$F4,$62
	.dc.b	$FD,$D1,$8F,$2B,$2D,$C3,$E3,$BF,$D6,$FE,$BD,$F2,$B4,$F3,$6B,$6B
	.dc.b	$E5,$E4,$AD,$C7,$F5,$1F,$AC,$FC,$2E,$5F,$9D,$9C,$D7,$F3,$E9,$CD
	.dc.b	$7F,$A3,$E7,$35,$FD,$CE,$73,$5F,$EB,$E9,$CD,$7E,$26,$73,$5F,$43
	.dc.b	$E7,$3D,$E7,$C0,$F0,$2D,$7E,$0A,$2E,$1F,$88,$9B,$5F,$53,$CE,$FA
	.dc.b	$5F,$5F,$F4,$DB,$87,$CE,$6D,$FF,$56,$E0,$46,$2F,$DD,$EC,$FC,$93
	.dc.b	$E8,$FD,$85,$C3,$E3,$EB,$FD,$AA,$BF,$C2,$1F,$58,$CF,$59,$FF,$62
	.dc.b	$8C,$5F,$0C,$7F,$BC,$46,$27,$F2,$13,$F2,$67,$35,$F8,$BF,$B0,$FC
	.dc.b	$67,$99,$6B,$F0,$ED,$BF,$53,$71,$FC,$AB,$F8,$8F,$3A,$57,$EA,$BD
	.dc.b	$7B,$7F,$B0,$9B,$7F,$A9,$9B,$5F,$EC,$9F,$B0,$FD,$A3,$F8,$66,$DF
	.dc.b	$C9,$EE,$FF,$14,$46,$2F,$F9,$13,$6B,$EF,$7C,$5E,$BB,$7F,$C3,$1C
	.dc.b	$DF,$D5,$A7,$B4,$62,$75,$1F,$91,$7F,$43,$F6,$12,$3F,$C4,$9E,$A3
	.dc.b	$DB,$C6,$9F,$C2,$AA,$E7,$FE,$31,$9B,$7F,$BA,$FA,$7F,$DB,$1F,$B7
	.dc.b	$F8,$FB,$1F,$B2,$8B,$87,$E1,$66,$DF,$F3,$DF,$B0,$FE,$78,$DF,$3F
	.dc.b	$FC,$91,$18,$C3,$FD,$E0,$FF,$53,$BF,$BA,$03,$B7,$0E,$EB,$B7,$CB
	.dc.b	$74,$BE,$BE,$EC,$C3,$F6,$F7,$FE,$DE,$6C,$67,$0E,$FF,$14,$EA,$FE
	.dc.b	$CD,$5A,$3F,$94,$C6,$7D,$3F,$4F,$47,$AB,$6B,$5D,$9B,$87,$A8,$F4
	.dc.b	$B8,$D7,$E2,$3F,$E6,$9F,$5F,$FC,$CD,$EA,$3E,$87,$E0,$5F,$F8,$0B
	.dc.b	$C6,$FE,$D7,$39,$8F,$89,$5C,$2B,$6F,$F0,$E9,$D4,$FF,$82,$BE,$1F
	.dc.b	$8D,$6A,$EC,$F2,$9A,$F7,$CB,$FD,$5D,$5B,$6C,$78,$2B,$78,$78,$63
	.dc.b	$13,$A7,$91,$FE,$B5,$AF,$2F,$C5,$70,$BE,$2F,$2F,$95,$F0,$FD,$83
	.dc.b	$E8,$5E,$A7,$3C,$FF,$B7,$EB,$EC,$AD,$4E,$6F,$8D,$F0,$CF,$C0,$7C
	.dc.b	$68,$CC,$F9,$6B,$8C,$E8,$38,$BF,$93,$87,$83,$3F,$A7,$39,$6E,$3D
	.dc.b	$EB,$C8,$D8,$FA,$78,$F8,$30,$ED,$7C,$5F,$CE,$35,$FF,$1F,$C6,$EF
	.dc.b	$DD,$EE,$FE,$DE,$7F,$DD,$55,$FE,$42,$EA,$FF,$73,$BE,$A8,$FC,$39
	.dc.b	$EA,$9E,$0E,$23,$83,$E2,$8E,$95,$DA,$BE,$27,$87,$FE,$CD,$C7,$DF
	.dc.b	$5A,$6C,$FC,$98,$C5,$FA,$AB,$83,$EC,$8C,$BE,$39,$E0,$76,$69,$E3
	.dc.b	$6D,$FD,$DA,$E7,$3C,$4D,$12,$34,$F0,$BF,$2B,$97,$EB,$CF,$2C,$3E
	.dc.b	$0D,$7B,$EB,$47,$A4,$E4,$7A,$6E,$BF,$84,$AE,$58,$7B,$2F,$9E,$1E
	.dc.b	$0C,$3C,$77,$D0,$F4,$3D,$CC,$96,$38,$12,$3C,$47,$19,$EA,$F8,$CF
	.dc.b	$C4,$DF,$8B,$0F,$2C,$EA,$7C,$C6,$30,$F3,$1E,$7A,$E1,$1D,$63,$0E
	.dc.b	$EB,$9C,$EB,$0F,$34,$F8,$5F,$4E,$1E,$78,$F6,$B7,$0F,$F5,$7C,$73
	.dc.b	$3D,$55,$CE,$75,$87,$82,$7C,$58,$78,$1E,$18,$78,$1F,$59,$F0,$0C
	.dc.b	$D7,$DC,$7F,$14,$64,$91,$FC,$17,$C3,$FC,$4F,$D8,$7F,$22,$8F,$E4
	.dc.b	$6D,$AF,$E5,$AD,$AF,$E9,$CB,$E3,$78,$CC,$9F,$E3,$93,$B4,$ED,$CC
	.dc.b	$ED,$FE,$19,$1B,$34,$E2,$3F,$9E,$F2,$FF,$A8,$67,$2D,$8C,$F1,$F9
	.dc.b	$BB,$2D,$F9,$BC,$BA,$F8,$BD,$3B,$8F,$FA,$76,$3F,$CA,$AF,$F8,$D9
	.dc.b	$B3,$A2,$3F,$D7,$3E,$03,$FD,$65,$97,$8F,$C1,$E6,$49,$FF,$2F,$FD
	.dc.b	$87,$F0,$E9,$D4,$76,$9D,$3F,$D7,$E7,$59,$FF,$5A,$7F,$DB,$71,$CB
	.dc.b	$9F,$D3,$1F,$CD,$E3,$87,$3E,$08,$F8,$1F,$83,$DC,$C9,$E8,$FC,$2F
	.dc.b	$AC,$7D,$85,$26,$1F,$E7,$6F,$B0,$F6,$E9,$FD,$B2,$B7,$3F,$E7,$A6
	.dc.b	$BF,$DD,$DB,$B7,$FB,$56,$B8,$CF,$F8,$4E,$5F,$1F,$E1,$24,$73,$F6
	.dc.b	$EB,$CE,$2D,$FF,$B7,$E9,$5F,$6B,$EB,$DF,$61,$EB,$F0,$66,$3F,$E8
	.dc.b	$E9,$CB,$67,$81,$1F,$57,$C6,$D7,$FC,$21,$3C,$1B,$3F,$1E,$7F,$13
	.dc.b	$F5,$CF,$1A,$92,$BF,$D3,$B0,$E6,$C4,$7F,$86,$E3,$83,$A9,$E7,$3D
	.dc.b	$3C,$AE,$BD,$52,$B9,$C6,$5E,$B1,$97,$A4,$7E,$73,$EB,$DF,$61,$E6
	.dc.b	$EF,$95,$FE,$B0,$9C,$B6,$78,$11,$F4,$FD,$85,$7F,$C6,$53,$C1,$B3
	.dc.b	$F0,$A7,$F5,$7F,$5F,$F8,$C9,$FF,$5E,$AF,$7B,$26,$BA,$64,$EA,$32
	.dc.b	$F0,$8E,$7F,$61,$23,$9B,$DB,$EB,$B1,$1F,$E4,$5B,$4E,$39,$DD,$5F
	.dc.b	$E4,$29,$5F,$F0,$0F,$33,$FC,$7B,$5A,$C3,$FC,$CD,$5F,$96,$4D,$9D
	.dc.b	$DF,$F9,$31,$B3,$FE,$6D,$6B,$FE,$31,$DB,$F8,$6F,$D7,$BE,$C3,$C1
	.dc.b	$12,$7F,$DC,$35,$C2,$B5,$1F,$EE,$47,$F4,$AD,$EA,$E6,$74,$EC,$91
	.dc.b	$92,$1F,$E6,$79,$E2,$DF,$3F,$C6,$47,$B7,$65,$74,$78,$63,$D3,$1F
	.dc.b	$1F,$DC,$FE,$4F,$BF,$E6,$EB,$FD,$F2,$D7,$FC,$ED,$92,$78,$47,$B9
	.dc.b	$AF,$F9,$EE,$74,$E4,$87,$28,$EE,$7F,$36,$FA,$F7,$D8,$52,$4F,$FA
	.dc.b	$77,$87,$ED,$2F,$B0,$7C,$3F,$1F,$4A,$4E,$F8,$FF,$D0,$B8,$FF,$70
	.dc.b	$8F,$F4,$75,$F6,$74,$7F,$AA,$7E,$C1,$F4,$5F,$D2,$1B,$1A,$0E,$1F
	.dc.b	$F4,$C7,$D8,$7D,$C7,$FD,$35,$0F,$8A,$E4,$48,$E5,$2F,$80,$FE,$AB
	.dc.b	$F6,$1F,$D0,$27,$FD,$44,$DD,$BD,$ED,$D3,$FD,$3F,$6F,$AA,$E3,$F3
	.dc.b	$33,$0F,$35,$A3,$FB,$0D,$7F,$4E,$9F,$F6,$37,$C1,$3E,$C3,$FB,$2B
	.dc.b	$63,$DC,$D7,$FD,$69,$F5,$FF,$C6,$9F,$F5,$B4,$7D,$07,$7C,$7C,$8F
	.dc.b	$ED,$F3,$F1,$74,$FA,$FF,$60,$BF,$EE,$6D,$BF,$98,$D7,$BF,$89,$AF
	.dc.b	$AA,$91,$A7,$FB,$DF,$B6,$F7,$79,$FC,$BC,$AE,$6E,$BC,$9F,$F0,$0F
	.dc.b	$D8,$3E,$0E,$92,$BE,$E6,$1E,$FD,$7F,$1B,$F6,$0F,$95,$FF,$04,$B6
	.dc.b	$BF,$04,$D7,$FD,$B0,$F8,$DF,$C3,$7C,$43,$E8,$BF,$C2,$2D,$8C,$C6
	.dc.b	$3E,$26,$47,$06,$BF,$EE,$3F,$AE,$FC,$33,$D3,$D5,$E6,$47,$F2,$9F
	.dc.b	$60,$FA,$5F,$F0,$FB,$63,$31,$83,$FC,$47,$4F,$90,$A6,$67,$67,$E7
	.dc.b	$E3,$4F,$E8,$7E,$C3,$F0,$0F,$96,$B9,$9F,$E2,$E8,$CE,$DF,$B6,$6F
	.dc.b	$81,$FE,$F6,$6D,$03,$85,$7F,$9D,$7C,$1D,$7D,$48,$F9,$7C,$0F,$07
	.dc.b	$3F,$B3,$CC,$93,$E7,$8C,$B4,$87,$CD,$F7,$3F,$C8,$BF,$61,$FE,$4C
	.dc.b	$CF,$C9,$28,$C3,$E7,$6B,$9E,$BE,$3C,$F2,$3F,$83,$4E,$84,$FF,$94
	.dc.b	$B1,$7C,$7D,$25,$7F,$0D,$FB,$0D,$F0,$7C,$5F,$B0,$5F,$90,$E3,$5F
	.dc.b	$C4,$DA,$A7,$F9,$2C,$61,$FE,$4F,$5D,$25,$5D,$1F,$A5,$19,$8C,$47
	.dc.b	$82,$FE,$1C,$62,$FB,$BA,$AE,$11,$F2,$1B,$5F,$76,$36,$7A,$6B,$9B
	.dc.b	$BC,$3D,$13,$E4,$8C,$61,$F0,$5E,$11,$4E,$95,$C9,$F4,$D7,$2C,$3C
	.dc.b	$8D,$AE,$33,$D4,$E2,$73,$18,$7A,$1B,$3C,$5D,$9C,$4F,$88,$71,$8E
	.dc.b	$07,$CD,$7F,$12,$72,$C2,$BE,$66,$B9,$33,$1F,$D4,$3E,$1F,$2D,$8C
	.dc.b	$F9,$69,$CB,$63,$98,$F8,$EA,$D3,$2E,$08,$FA,$EF,$FD,$75,$B7,$C3
	.dc.b	$8F,$DA,$4F,$C5,$9C,$C7,$C1,$32,$71,$DB,$FC,$59,$A7,$E2,$23,$2F
	.dc.b	$E2,$A7,$8B,$98,$F2,$39,$9E,$3A,$7C,$6B,$F0,$46,$2B,$E5,$C6,$30
	.dc.b	$F3,$47,$2A,$FF,$62,$D6,$E3,$F7,$35,$B7,$D0,$6C,$FF,$05,$46,$F0
	.dc.b	$F6,$9F,$EB,$C7,$FC,$8C,$FF,$83,$A3,$D7,$5D,$AF,$FC,$05,$3C,$1F
	.dc.b	$F6,$AF,$C7,$70,$75,$5F,$B3,$72,$7F,$82,$7E,$C3,$F8,$0B,$77,$E1
	.dc.b	$3F,$C8,$E7,$F9,$24,$E4,$91,$CC,$FD,$F7,$99,$E1,$78,$57,$2C,$3F
	.dc.b	$C1,$99,$74,$96,$1F,$D1,$70,$FF,$0F,$D3,$EC,$1F,$39,$BA,$D5,$F7
	.dc.b	$AD,$4E,$7A,$BB,$5F,$0C,$3D,$91,$89,$E5,$E6,$C9,$EF,$F6,$1F,$D3
	.dc.b	$DF,$85,$19,$C7,$B4,$FF,$B1,$A7,$27,$0F,$6F,$59,$7D,$49,$D5,$7C
	.dc.b	$33,$DB,$7F,$06,$33,$3F,$E4,$6F,$93,$FF,$03,$61,$FE,$39,$6D,$F1
	.dc.b	$FA,$E7,$D8,$7C,$79,$CD,$F1,$EC,$F9,$5F,$60,$F8,$8D,$46,$19,$97
	.dc.b	$F9,$17,$1F,$F0,$43,$FE,$47,$AE,$0E,$7C,$AE,$7F,$2B,$F5,$7E,$C1
	.dc.b	$E0,$E2,$BF,$C9,$D1,$86,$C6,$F8,$F8,$36,$37,$D1,$F7,$FC,$3F,$F0
	.dc.b	$84,$60,$FF,$2C,$38,$E1,$F9,$1F,$60,$F7,$7F,$11,$E5,$FD,$59,$CC
	.dc.b	$63,$C9,$FC,$56,$1F,$E6,$43,$F1,$3F,$61,$FE,$49,$7F,$CC,$86,$6B
	.dc.b	$F1,$11,$97,$EB,$7C,$AF,$E7,$1C,$E5,$F9,$58,$C7,$2F,$EC,$2B,$FC
	.dc.b	$F7,$7F,$E6,$F9,$7C,$B7,$E7,$7E,$35,$F3,$5C,$63,$0F,$3E,$CD,$4D
	.dc.b	$BF,$CF,$BF,$ED,$57,$FC,$F5,$E2,$BD,$2F,$D4,$7D,$83,$C6,$FD,$44
	.dc.b	$E6,$BF,$4F,$39,$78,$ED,$FA,$FF,$B0,$F9,$51,$83,$F5,$AE,$38,$3F
	.dc.b	$2D,$F5,$BF,$E6,$19,$FD,$8F,$D8,$7C,$7E,$2F,$F4,$DC,$E6,$BF,$61
	.dc.b	$39,$AF,$F0,$9C,$FF,$AB,$DF,$DC,$FD,$87,$C8,$73,$F1,$3F,$C0,$63
	.dc.b	$17,$FE,$2D,$A7,$BB,$FC,$17,$EC,$1F,$2D,$FC,$16,$73,$8F,$F0,$49
	.dc.b	$CB,$FE,$9F,$F9,$FF,$F3,$43,$FC,$23,$2F,$E3,$38,$FE,$C6,$72,$FF
	.dc.b	$07,$8C,$E9,$FE,$B6,$FA,$F7,$D8,$7F,$15,$97,$C9,$FF,$15,$9C,$E3
	.dc.b	$D6,$3F,$8C,$7D,$87,$FA,$A6,$3F,$D6,$0D,$C7,$F8,$A7,$D8,$3E,$A7
	.dc.b	$7F,$1D,$2F,$F8,$EC,$FE,$F1,$E3,$1F,$C9,$63,$15,$FC,$26,$FA,$69
	.dc.b	$FD,$02,$72,$FF,$B3,$A3,$37,$FE,$D1,$9C,$BF,$ED,$3F,$B0,$FD,$ED
	.dc.b	$FF,$27,$6D,$FF,$4A,$7E,$5C,$67,$CD,$FE,$6D,$1A,$AE,$51,$8E,$2C
	.dc.b	$BE,$7F,$2F,$D1,$D1,$6B,$DA,$DC,$3E,$1B,$71,$F5,$3F,$D0,$7E,$BD
	.dc.b	$F2,$F4,$F0,$ED,$70,$FF,$6E,$4E,$9F,$E8,$BF,$59,$F7,$30,$FF,$77
	.dc.b	$CE,$71,$FE,$89,$39,$AF,$F3,$3C,$E6,$BF,$B4,$CE,$6B,$FC,$FD,$39
	.dc.b	$AF,$F7,$D4,$E6,$9B,$3E,$7B,$EB,$1F,$5F,$C1,$B7,$F7,$22,$E3,$F8
	.dc.b	$29,$B7,$F3,$FC,$FF,$9B,$F5,$FF,$39,$B8,$7F,$60,$E2,$46,$30,$FB
	.dc.b	$0D,$BF,$7F,$B1,$F2,$4F,$C4,$7D,$85,$C3,$E3,$C7,$F6,$FF,$AC,$65
	.dc.b	$9C,$7F,$B4,$46,$2B,$67,$F7,$07,$11,$FD,$67,$7F,$ED,$4F,$0A,$FC
	.dc.b	$2F,$D8,$7E,$1B,$CF,$B8,$7A,$5B,$5F,$8F,$6E,$7F,$8F,$7F,$3D,$E7
	.dc.b	$CA,$FD,$1F,$AF,$6F,$F5,$33,$6F,$F4,$13,$6B,$FD,$43,$F6,$1F,$AD
	.dc.b	$7F,$D6,$D2,$73,$D9,$F9,$AE,$FF,$F0,$75,$6F,$1F,$EE,$8D,$BF,$C1
	.dc.b	$FA,$3A,$BF,$F0,$B7,$3F,$F4,$F9,$E9,$7F,$EC,$3B,$E1,$8F,$E1,$B0
	.dc.b	$E0,$FE,$51,$FD,$D7,$D8,$48,$FF,$0D,$FA,$8F,$6B,$18,$7F,$06,$AB
	.dc.b	$8F,$F8,$F2,$B5,$7F,$C1,$A7,$F6,$7F,$4D,$23,$F0,$DF,$1D,$70,$FD
	.dc.b	$E4,$DC,$3F,$79,$36,$BF,$D8,$FF,$61,$FC,$E6,$7D,$76,$8F,$E9,$0F
	.dc.b	$F8,$5D,$D4,$F0,$F4,$7F,$C6,$D1,$89,$FF,$72,$3C,$2B,$9B,$98,$E9
	.dc.b	$3C,$3C,$39,$7F,$58,$F1,$FC,$56,$2B,$C1,$3F,$E4,$EF,$89,$75,$7F
	.dc.b	$5D,$9C,$DF,$17,$FD,$1D,$19,$CB,$C3,$CF,$E6,$D7,$CE,$DB,$E1,$7E
	.dc.b	$F8,$C3,$EF,$6D,$7C,$4F,$5F,$FA,$4F,$A8,$FA,$3F,$F7,$7C,$7F,$0D
	.dc.b	$8C,$58,$FF,$00,$FC,$3F,$E1,$A3,$31,$FE,$04,$F1,$F1,$7C,$AA,$CF
	.dc.b	$1B,$E9,$39,$CB,$FD,$3A,$F8,$2F,$75,$CE,$72,$F1,$8C,$D7,$A2,$B8
	.dc.b	$56,$AB,$81,$FE,$B1,$8F,$5F,$C4,$70,$BE,$2F,$2F,$93,$EF,$F6,$0F
	.dc.b	$A3,$7A,$7C,$F3,$FE,$D7,$AF,$BA,$B5,$39,$BE,$37,$C3,$3F,$01,$EF
	.dc.b	$8C,$CF,$96,$B8,$CE,$83,$8B,$F9,$78,$78,$33,$FA,$15,$AC,$7B,$57
	.dc.b	$8D,$B3,$AB,$E4,$FD,$AC,$7C,$37,$DE,$7F,$C6,$FD,$5F,$EE,$06,$C7
	.dc.b	$FB,$91,$F2,$BF,$EE,$87,$2D,$F7,$7F,$78,$F3,$6E,$3F,$15,$B8,$78
	.dc.b	$FC,$2F,$F6,$0B,$E5,$8C,$D7,$E2,$EF,$C3,$7C,$70,$F8,$D8,$70,$AD
	.dc.b	$5F,$07,$E3,$B6,$F8,$E1,$DE,$7E,$74,$E4,$97,$D2,$32,$C7,$D4,$E5
	.dc.b	$A4,$C3,$C8,$FC,$9C,$3C,$8F,$8A,$79,$4E,$8F,$C3,$4E,$B0,$F2,$D7
	.dc.b	$AA,$B7,$87,$9E,$3B,$3F,$16,$33,$1F,$16,$B9,$61,$EA,$8D,$99,$36
	.dc.b	$78,$CC,$9F,$8C,$8E,$33,$AB,$F1,$CE,$4C,$57,$5C,$3D,$31,$8A,$E1
	.dc.b	$7C,$A3,$AD,$F2,$3F,$D3,$0E,$CC,$C6,$EB,$E2,$61,$ED,$9F,$94,$6E
	.dc.b	$32,$FA,$CE,$B8,$78,$E7,$E3,$4E,$A7,$DC,$DB,$F0,$C7,$0C,$3C,$EF
	.dc.b	$5A,$F0,$57,$B2,$34,$79,$8F,$C0,$3C,$EF,$F1,$98,$7D,$92,$4E,$AB
	.dc.b	$F5,$DE,$17,$E0,$BE,$C3,$F9,$0E,$1F,$C8,$9B,$3F,$B8,$6D,$7F,$29
	.dc.b	$5F,$19,$C6,$64,$FF,$19,$9D,$A7,$6E,$67,$6F,$F0,$98,$D9,$A7,$11
	.dc.b	$FC,$23,$93,$FB,$0C,$E5,$B1,$9E,$3F,$37,$5D,$BF,$37,$95,$5F,$17
	.dc.b	$A5,$71,$FF,$4B,$C7,$F4,$E7,$FC,$72,$D9,$D1,$1F,$EB,$5D,$6F,$80
	.dc.b	$FE,$E7,$9F,$93,$C0,$E8,$4A,$FE,$B3,$F6,$1F,$C8,$14,$ED,$F0,$4E
	.dc.b	$DC,$E9,$FC,$9A,$B8,$57,$F8,$01,$FF,$6B,$CF,$4E,$8F,$64,$7F,$37
	.dc.b	$8E,$27,$F5,$D7,$6E,$03,$89,$E9,$D1,$C8,$F9,$19,$7B,$7D,$0F,$91
	.dc.b	$F5,$CF,$B0,$A4,$C7,$FB,$AF,$D8,$7C,$2D,$3F,$9E,$D6,$E7,$FB,$CB
	.dc.b	$5F,$13,$72,$FF,$0D,$D7,$19,$FE,$E9,$2F,$8E,$F0,$92,$39,$FB,$55
	.dc.b	$E7,$16,$FF,$C4,$B5,$FE,$C2,$9F,$F0,$4F,$D8,$3E,$C3,$D7,$E0,$C9
	.dc.b	$CC,$7E,$16,$75,$18,$21,$E8,$F8,$AD,$4C,$8C,$E1,$E0,$9F,$C7,$7D
	.dc.b	$83,$C7,$A4,$AF,$F0,$DC,$39,$71,$1F,$56,$38,$3A,$9E,$73,$D3,$C9
	.dc.b	$EB,$D5,$2B,$9C,$65,$EB,$19,$7A,$47,$E8,$7E,$C1,$F4,$1E,$6E,$F9
	.dc.b	$5F,$E2,$E9,$CB,$67,$81,$1F,$4F,$D0,$D7,$F1,$93,$C1,$B3,$F0,$A7
	.dc.b	$F6,$1F,$5F,$FE,$B1,$9F,$F1,$DA,$F7,$72,$6B,$A6,$4E,$A3,$2F,$08
	.dc.b	$E7,$F6,$12,$39,$BD,$BE,$BB,$11,$F9,$FB,$4E,$38,$BF,$CD,$30,$AF
	.dc.b	$E0,$9E,$67,$F9,$02,$73,$D7,$E8,$33,$B4,$C7,$FC,$A7,$5B,$9F,$F0
	.dc.b	$DB,$63,$FC,$99,$F5,$EF,$B0,$E1,$1F,$B1,$9D,$3F,$E3,$5B,$FE,$29
	.dc.b	$5B,$8F,$F2,$FB,$FA,$16,$F6,$68,$DF,$64,$8C,$90,$FD,$F4,$F1,$6F
	.dc.b	$9F,$D8,$8F,$6E,$CA,$E8,$F0,$C7,$92,$3E,$2F,$B7,$FC,$A7,$6F,$CC
	.dc.b	$57,$F9,$D1,$AF,$B6,$36,$FC,$66,$BF,$C4,$61,$18,$DB,$FD,$37,$A6
	.dc.b	$2B,$81,$D6,$7F,$9C,$7D,$7B,$EC,$29,$27,$FD,$0B,$F6,$0F,$87,$DF
	.dc.b	$A1,$27,$FD,$13,$7F,$DB,$32,$FF,$54,$DF,$39,$FF,$46,$FD,$83,$E8
	.dc.b	$7F,$A3,$36,$33,$1F,$D2,$7E,$BF,$EB,$47,$F2,$B9,$7C,$57,$02,$47
	.dc.b	$FA,$62,$5E,$FF,$F5,$1F,$B0,$FE,$61,$D3,$FE,$62,$6D,$7F,$A4,$DA
	.dc.b	$A9,$F4,$DC,$7E,$66,$4F,$FA,$92,$91,$E5,$8F,$EB,$DF,$5C,$FB,$0A
	.dc.b	$48,$FF,$55,$FD,$87,$F0,$E7,$FA,$24,$FC,$97,$FD,$5E,$D7,$3C,$7F
	.dc.b	$DA,$67,$E2,$47,$F4,$BF,$B0,$57,$F6,$D6,$C7,$FA,$E1,$AF,$F5,$19
	.dc.b	$78,$9B,$F8,$49,$3B,$7F,$BB,$7B,$6F,$77,$A7,$CC,$4A,$E8,$EB,$C9
	.dc.b	$FE,$F9,$F6,$0F,$83,$A4,$AF,$F6,$4E,$1F,$0D,$FF,$65,$FD,$83,$E5
	.dc.b	$7F,$C0,$8D,$8C,$C6,$32,$FF,$04,$7C,$FB,$EC,$29,$2B,$FD,$AB,$0F
	.dc.b	$89,$D9,$23,$2F,$F8,$47,$EB,$BF,$07,$1C,$3D,$5E,$A4,$7F,$DB,$FF
	.dc.b	$60,$FA,$5F,$F0,$C3,$63,$31,$83,$A7,$C8,$48,$C9,$0F,$F0,$05,$A3
	.dc.b	$FC,$45,$F6,$1F,$77,$C5,$78,$1F,$EB,$4A,$ED,$7C,$30,$DF,$5C,$8F
	.dc.b	$F7,$9E,$1C,$6F,$FD,$76,$D8,$FF,$76,$5F,$E6,$1D,$D7,$17,$FC,$25
	.dc.b	$3B,$53,$B7,$3B,$FC,$7F,$32,$47,$82,$3D,$24,$70,$7F,$BE,$F2,$F5
	.dc.b	$BE,$2F,$73,$FC,$81,$F6,$1F,$E4,$7C,$FF,$DA,$2D,$90,$70,$DE,$AE
	.dc.b	$3C,$D2,$3B,$57,$28,$EB,$3F,$E4,$DC,$9F,$1F,$49,$7C,$0E,$FF,$60
	.dc.b	$AF,$F2,$1C,$3E,$C1,$7F,$EE,$63,$9C,$71,$33,$7D,$1F,$F1,$E5,$F3
	.dc.b	$D3,$D7,$7C,$A7,$BE,$1D,$A3,$11,$C8,$DE,$1D,$F0,$FF,$7A,$CE,$B0
	.dc.b	$E9,$5E,$5A,$FF,$7C,$C6,$27,$9C,$79,$A7,$47,$CC,$71,$1C,$EF,$C5
	.dc.b	$1C,$9B,$3B,$BF,$0C,$F8,$E3,$CA,$7A,$B0,$AB,$0F,$14,$EE,$3C,$27
	.dc.b	$B4,$B1,$A7,$91,$A8,$F2,$9C,$70,$F3,$57,$6A,$F8,$35,$FE,$79,$6B
	.dc.b	$E9,$8F,$F3,$E3,$D0,$D7,$94,$F8,$6C,$C7,$9E,$30,$F0,$AF,$EA,$13
	.dc.b	$A6,$C7,$01,$F1,$D5,$A6,$5C,$11,$F8,$37,$FE,$B4,$D7,$E3,$C7,$E7
	.dc.b	$A7,$E4,$CE,$63,$E3,$99,$38,$EB,$FA,$AC,$BF,$0F,$19,$7F,$13,$3C
	.dc.b	$5C,$C7,$91,$CC,$F1,$CF,$E2,$DF,$82,$31,$5F,$EA,$38,$C6,$1E,$68
	.dc.b	$E5,$5F,$EB,$FA,$DC,$7E,$D8,$C9,$FB,$F7,$2F,$96,$38,$4F,$48,$FF
	.dc.b	$50,$61,$EF,$3F,$82,$BC,$EF,$51,$FE,$BA,$BF,$E1,$2F,$FA,$F3,$E3
	.dc.b	$AD,$F4,$BE,$71,$87,$F8,$4F,$D8,$7F,$00,$9F,$DB,$36,$3F,$D6,$0F
	.dc.b	$F0,$E8,$CE,$DC,$65,$3B,$BF,$0F,$99,$E7,$BF,$81,$39,$7F,$8F,$46
	.dc.b	$6B,$C3,$CF,$E3,$B7,$D8,$3E,$77,$B6,$3F,$D2,$F1,$FE,$25,$5A,$9C
	.dc.b	$DF,$1B,$E1,$5E,$A8,$C4,$F2,$F2,$E4,$F2,$FB,$0F,$E9,$8F,$C0,$8C
	.dc.b	$E5,$FE,$D2,$C5,$5F,$EB,$BA,$D1,$C7,$DB,$DC,$FD,$09,$D5,$7C,$8B
	.dc.b	$F6,$46,$27,$EC,$FC,$A7,$F8,$17,$0F,$E4,$8D,$BF,$F6,$F7,$D7,$BE
	.dc.b	$C3,$E7,$4E,$6F,$8F,$67,$D1,$FB,$07,$C4,$6A,$30,$CC,$BF,$98,$63
	.dc.b	$FE,$07,$7F,$99,$57,$07,$3E,$47,$3F,$95,$FB,$9F,$60,$F0,$71,$5F
	.dc.b	$CE,$63,$0D,$8D,$F1,$F0,$6C,$6F,$9B,$F0,$DE,$27,$F8,$46,$30,$7F
	.dc.b	$43,$71,$C3,$FE,$FD,$FB,$07,$BB,$F8,$3F,$2B,$EA,$CE,$63,$1E,$47
	.dc.b	$E4,$70,$FF,$30,$1F,$84,$FB,$0F,$F2,$43,$FD,$3C,$CD,$7D,$B8,$CA
	.dc.b	$E4,$FD,$5F,$CC,$7E,$76,$35,$9F,$E7,$63,$13,$FA,$CA,$FF,$3B,$5F
	.dc.b	$F5,$B9,$7C,$B7,$E6,$7E,$35,$F3,$5C,$63,$0F,$3E,$CD,$4D,$BF,$CD
	.dc.b	$BF,$EC,$B7,$FB,$37,$88,$F4,$3F,$49,$F6,$0F,$17,$F6,$13,$9A,$FD
	.dc.b	$1C,$E5,$E3,$B7,$EA,$FE,$C3,$E5,$C6,$0F,$D4,$B8,$E0,$F8,$51,$FC
	.dc.b	$57,$EB,$BF,$CC,$6F,$EB,$BE,$C3,$E4,$F1,$7F,$10,$9C,$D7,$E9,$67
	.dc.b	$35,$C7,$E2,$3F,$75,$F6,$1F,$22,$30,$7E,$E1,$C5,$7F,$86,$2D,$EE
	.dc.b	$7F,$01,$FB,$07,$CB,$7F,$0D,$9C,$E1,$FC,$02,$72,$FF,$82,$FE,$7F
	.dc.b	$FC,$D0,$FF,$04,$CB,$F8,$76,$3F,$E2,$49,$CB,$FB,$B8,$CA,$FA,$F7
	.dc.b	$D8,$7F,$11,$97,$C9,$FF,$1D,$9C,$E1,$D6,$3F,$90,$7D,$87,$FA,$A2
	.dc.b	$3F,$C3,$8D,$C3,$F8,$47,$D8,$3E,$AF,$7F,$1D,$2F,$F8,$CC,$FE,$F5
	.dc.b	$E3,$1F,$C8,$63,$15,$F5,$0F,$F1,$ED,$F5,$E3,$FE,$7B,$3A,$7F,$A1
	.dc.b	$C6,$6F,$FC,$95,$39,$7F,$C9,$9F,$61,$FB,$FB,$FE,$38,$DB,$FE,$8E
	.dc.b	$FC,$C8,$CF,$9B,$FC,$CE,$35,$5C,$A3,$1C,$59,$7D,$1E,$5F,$A7,$A2
	.dc.b	$D7,$B9,$B8,$7C,$46,$E3,$EA,$7F,$A5,$7D,$7B,$E5,$E9,$E1,$DA,$E1
	.dc.b	$FE,$49,$9D,$3F,$D0,$BE,$B7,$EE,$E1,$FD,$92,$73,$8F,$F3,$99,$CD
	.dc.b	$7F,$71,$9C,$D7,$F6,$69,$CD,$7F,$9A,$A7,$35,$FE,$73,$9C,$D3,$67
	.dc.b	$CF,$7D,$73,$EC,$78,$36,$FE,$EC,$5C,$7F,$07,$36,$FE,$7F,$9F,$F3
	.dc.b	$7E,$BF,$E7,$37,$0F,$EB,$BC,$48,$C6,$1F,$61,$B7,$EF,$F6,$3E,$49
	.dc.b	$F8,$9F,$B0,$B8,$7C,$78,$FE,$D9,$F5,$8C,$B3,$8F,$F5,$D8,$C5,$6C
	.dc.b	$FE,$DA,$E2,$3F,$24,$37,$FE,$DA,$F1,$AF,$C3,$7D,$87,$E1,$FC,$FB
	.dc.b	$87,$A9,$B5,$F9,$06,$E7,$F9,$07,$F4,$1E,$7C,$AF,$D2,$FA,$F6,$FF
	.dc.b	$57,$36,$FF,$43,$36,$BF,$C5,$9F,$61,$FA,$F7,$FD,$1C,$DB,$FC,$FF
	.dc.b	$7F,$F8,$5E,$31,$7F,$E2,$A6,$D7,$DF,$F4,$75,$DD,$FE,$13,$E7,$FE
	.dc.b	$A3,$7B,$8C,$4E,$AB,$F2,$4F,$EE,$7E,$C2,$47,$F8,$5B,$D4,$7B,$58
	.dc.b	$D3,$F8,$2D,$5C,$7F,$C4,$F3,$6F,$F6,$FF,$4D,$FB,$13,$F6,$5F,$1B
	.dc.b	$63,$F0,$B1,$70,$FD,$DC,$DB,$FF,$22,$FD,$87,$F3,$66,$FA,$3F,$E3
	.dc.b	$78,$C6,$1F,$E5,$37,$FA,$45,$FD,$E7,$2F,$FB,$31,$CD,$74,$F9,$7E
	.dc.b	$57,$CF,$C4,$98,$7F,$AF,$6F,$FD,$88,$D8,$CE,$1F,$C6,$FE,$25,$D5
	.dc.b	$FD,$66,$3C,$13,$FE,$53,$9C,$FA,$9E,$7E,$8F,$46,$D7,$1E,$CD,$C3
	.dc.b	$D0,$DD,$7E,$0B,$FE,$55,$F5,$FF,$CB,$5E,$AB,$E8,$FF,$CE,$51,$FC
	.dc.b	$1A,$31,$7F,$EE,$A8,$DC,$EB,$C7,$FF,$79,$D6,$AB,$FB,$FF,$C3,$62
	.dc.b	$B5,$97,$FA,$55,$EB,$19,$9E,$97,$5F,$F0,$3B,$70,$F1,$57,$0B,$E5
	.dc.b	$5C,$1A,$F0,$6C,$FF,$BB,$9B,$5C,$FE,$23,$A7,$C9,$F7,$FB,$07,$D1
	.dc.b	$3D,$4E,$33,$98,$C6,$5F,$EB,$6C,$FE,$2C,$E6,$B7,$97,$8B,$0E,$71
	.dc.b	$70,$E3,$AF,$C5,$9C,$E3,$CC,$F9,$98,$73,$70,$23,$FC,$4F,$D3,$FE
	.dc.b	$CE,$6C,$7F,$B5,$1E,$6F,$FA,$A4,$E1,$EE,$FE,$B1,$D3,$6F,$C4,$DC
	.dc.b	$FF,$D6,$15,$D2,$7C,$DE,$1F,$FA,$A6,$73,$5B,$B1,$DB,$0F,$81,$19
	.dc.b	$8F,$45,$71,$77,$87,$5B,$E7,$7C,$0F,$91,$5B,$AE,$87,$C2,$AE,$0F
	.dc.b	$2B,$ED,$1F,$E6,$28,$E4,$6E,$32,$6A,$7A,$4F,$29,$EB,$87,$C2,$BF
	.dc.b	$0C,$72,$8C,$3C,$E3,$2D,$24,$68,$F8,$EF,$1C,$3D,$AE,$9F,$9B,$7E
	.dc.b	$2C,$3C,$B7,$BA,$F8,$B3,$98,$C4,$79,$AB,$8B,$8C,$3D,$44,$AE,$11
	.dc.b	$DC,$F8,$D5,$E8,$C3,$CD,$7E,$46,$CF,$80,$E9,$87,$91,$FA,$17,$D2
	.dc.b	$77,$1E,$33,$81,$DC,$C1,$30,$FA,$8E,$5F,$3D,$7D,$38,$EF,$87,$AE
	.dc.b	$FB,$61,$E5,$24,$F6,$C3,$C3,$87,$2B,$F6,$56,$AB,$C1,$87,$57,$C0
	.dc.b	$FA,$EF,$EA,$C6,$2B,$C0,$F2,$3E,$39,$E7,$C3,$F0,$E4,$74,$7E,$CF
	.dc.b	$C3,$FC,$B7,$D8,$7F,$1F,$BF,$E4,$0D,$C7,$F4,$B4,$8C,$3F,$A8,$5F
	.dc.b	$19,$CA,$64,$FF,$1A,$9D,$BC,$67,$B9,$A7,$F8,$2C,$68,$E6,$F1,$8F
	.dc.b	$0F,$2B,$83,$C7,$5A,$AF,$F0,$BB,$67,$51,$E5,$EC,$E3,$5C,$3C,$BB
	.dc.b	$3E,$9F,$4E,$BE,$D6,$E1,$FE,$33,$8F,$E0,$B8,$74,$7E,$15,$68,$8E
	.dc.b	$AF,$F8,$5F,$5B,$E0,$38,$5F,$1C,$1E,$0F,$62,$61,$C4,$FE,$27,$F6
	.dc.b	$1F,$C8,$6B,$84,$EA,$7D,$EF,$A2,$75,$8F,$D6,$7F,$DA,$B3,$CC,$3F
	.dc.b	$D0,$1F,$F1,$EB,$6B,$D8,$F0,$8C,$F5,$75,$23,$E5,$CB,$D9,$E8,$FC
	.dc.b	$8F,$AE,$7D,$85,$25,$78,$7E,$C3,$E0,$E7,$F8,$AA,$DC,$F9,$5A,$FF
	.dc.b	$94,$1B,$9F,$CF,$AE,$33,$FC,$DA,$5F,$1D,$E1,$24,$73,$F6,$EB,$CE
	.dc.b	$2D,$FD,$9A,$FF,$5F,$CF,$C5,$FB,$07,$D8,$7A,$FC,$19,$39,$8F,$C2
	.dc.b	$CE,$A3,$04,$3D,$1F,$15,$A9,$91,$9C,$3C,$13,$F8,$EF,$B0,$78,$D4
	.dc.b	$95,$F5,$E3,$F0,$AD,$8C,$BF,$E7,$48,$E2,$EF,$93,$A7,$93,$D7,$A6
	.dc.b	$4F,$38,$CB,$D6,$32,$F3,$8F,$D0,$FD,$83,$E8,$3C,$DD,$D2,$BF,$15
	.dc.b	$39,$6C,$F0,$23,$E9,$FA,$1A,$FF,$A2,$A7,$83,$67,$E1,$4F,$EC,$3E
	.dc.b	$BF,$FD,$5B,$3F,$97,$5E,$F6,$4D,$74,$C9,$D4,$65,$E1,$1C,$FE,$C2
	.dc.b	$47,$37,$B7,$D7,$62,$3F,$D3,$B6,$9C,$71,$7E,$F6,$15,$FC,$13,$CB
	.dc.b	$FF,$21,$4E,$7A,$FD,$06,$76,$98,$FE,$BE,$B7,$3F,$CA,$1B,$1F,$E0
	.dc.b	$EF,$AF,$7D,$87,$08,$FF,$57,$CE,$9F,$F1,$CD,$FF,$14,$AD,$C7,$EE
	.dc.b	$DF,$D0,$B7,$B3,$46,$FB,$24,$64,$87,$FA,$F2,$78,$B7,$CF,$EC,$47
	.dc.b	$B7,$65,$74,$78,$63,$C9,$1F,$17,$DC,$FE,$53,$B7,$E6,$2B,$FD,$92
	.dc.b	$DD,$BF,$94,$68,$AE,$15,$F1,$DA,$F0,$6C,$F0,$6E,$DF,$D4,$AF,$95
	.dc.b	$7F,$1D,$FA,$F7,$D8,$7F,$3A,$6C,$66,$3F,$9E,$FD,$83,$E2,$35,$E7
	.dc.b	$CA,$E4,$EA,$BF,$A1,$FD,$83,$E8,$A9,$2B,$F9,$67,$D8,$7D,$47,$FD
	.dc.b	$CF,$2F,$8B,$E2,$48,$DF,$C0,$7F,$4E,$FA,$FF,$E5,$DC,$3F,$D2,$DB
	.dc.b	$C1,$FD,$B9,$B5,$FE,$54,$C5,$F6,$1C,$7E,$66,$4F,$F3,$DA,$47,$F5
	.dc.b	$AA,$FE,$87,$3F,$E5,$CF,$82,$7D,$87,$F5,$E6,$C7,$F4,$56,$BF,$F0
	.dc.b	$03,$EB,$FF,$86,$AC,$FF,$B0,$4F,$C9,$8E,$1C,$5C,$CF,$ED,$35,$F1
	.dc.b	$E3,$FA,$87,$D8,$2B,$FB,$6B,$63,$31,$C3,$C4,$CF,$D4,$C9,$E2,$FF
	.dc.b	$74,$F6,$9E,$E7,$2F,$99,$95,$D9,$D7,$93,$FD,$EB,$EC,$1F,$05,$49
	.dc.b	$5F,$D8,$F0,$F7,$EF,$FD,$EF,$EC,$1F,$2D,$FE,$02,$6D,$7F,$69,$6B
	.dc.b	$F0,$F9,$FF,$C4,$71,$E4,$8F,$ED,$DF,$10,$FB,$0F,$F0,$5B,$63,$31
	.dc.b	$8F,$89,$91,$C1,$AF,$53,$FD,$27,$F5,$CE,$CE,$7E,$AC,$9E,$EF,$F8
	.dc.b	$63,$EC,$1F,$4B,$49,$3F,$E0,$18,$CB,$FE,$03,$B7,$C8,$7F,$88,$1B
	.dc.b	$1B,$AE,$2F,$8C,$46,$9C,$7D,$87,$E0,$DD,$47,$66,$CF,$F8,$B2,$75
	.dc.b	$C1,$FE,$29,$6E,$FF,$E4,$C6,$D7,$FA,$C9,$AF,$83,$C1,$E5,$DB,$23
	.dc.b	$FC,$39,$3A,$C7,$F4,$B1,$8E,$8F,$67,$99,$23,$04,$3D,$31,$E5,$3E
	.dc.b	$17,$B9,$FE,$42,$FB,$0F,$F2,$56,$7F,$E6,$06,$D7,$FB,$19,$A9,$76
	.dc.b	$71,$E8,$93,$FE,$31,$9D,$9A,$7F,$C6,$98,$BE,$3F,$FC,$A2,$D8,$CC
	.dc.b	$7F,$95,$3E,$C3,$AF,$07,$FB,$5B,$EC,$17,$FE,$74,$38,$5F,$D7,$6C
	.dc.b	$FF,$95,$DB,$9F,$F9,$EA,$F9,$D2,$F9,$BF,$EE,$28,$CC,$62,$3B,$E1
	.dc.b	$FE,$82,$8C,$5F,$81,$D5,$70,$8E,$D9,$FF,$09,$BE,$D1,$E7,$BE,$4D
	.dc.b	$8F,$63,$C3,$1B,$1E,$3A,$E5,$87,$82,$F8,$CF,$FB,$E2,$FC,$47,$17
	.dc.b	$B9,$C6,$F9,$BE,$A2,$C7,$07,$47,$01,$87,$AD,$ED,$5E,$07,$91,$B8
	.dc.b	$D3,$C3,$0F,$80,$FF,$AA,$0C,$5D,$7B,$99,$9F,$A6,$DF,$29,$F0,$DB
	.dc.b	$8F,$1C,$61,$E5,$3F,$D5,$67,$4D,$8E,$03,$E3,$6B,$4C,$B8,$23,$F0
	.dc.b	$6F,$FD,$77,$A7,$C8,$81,$C7,$4F,$C2,$4E,$A3,$E4,$1A,$AF,$9F,$19
	.dc.b	$9F,$C5,$4E,$4E,$DB,$7F,$AD,$A3,$31,$D7,$5F,$F5,$FC,$E5,$F9,$17
	.dc.b	$D2,$31,$3F,$36,$31,$87,$9E,$39,$1F,$EB,$89,$DB,$FE,$DA,$9D,$BE
	.dc.b	$83,$67,$F5,$F8,$DE,$1E,$D3,$FD,$72,$FF,$B3,$1F,$F6,$14,$7A,$F0
	.dc.b	$F8,$73,$9B,$F4,$9F,$EB,$1A,$7C,$6E,$6F,$C9,$39,$8C,$43,$EC,$3F
	.dc.b	$82,$4F,$FA,$E5,$B1,$FC,$CD,$FA,$71,$9D,$F8,$C6,$AF,$CD,$D7,$E9
	.dc.b	$AF,$F7,$CC,$E5,$FF,$6D,$C6,$6F,$EE,$CF,$F4,$A9,$FF,$69,$53,$EC
	.dc.b	$1F,$3B,$DB,$AB,$D5,$39,$8C,$57,$0A,$DB,$E9,$AD,$C6,$30,$F5,$79
	.dc.b	$BE,$83,$9F,$D8,$7F,$54,$BF,$81,$19,$7C,$71,$FE,$C0,$72,$63,$DB
	.dc.b	$D5,$7F,$69,$C9,$5B,$8E,$F5,$B3,$FD,$C9,$F2,$5F,$E0,$C6,$F5,$7D
	.dc.b	$2F,$AF,$7D,$87,$D4,$AD,$5F,$F6,$36,$E2,$C3,$F4,$B2,$FB,$07,$C4
	.dc.b	$71,$CB,$EE,$C6,$2B,$FD,$E8,$DA,$FE,$DD,$97,$F8,$4A,$73,$7E,$3F
	.dc.b	$2B,$8F,$CA,$FE,$1B,$EC,$1E,$0E,$2F,$FD,$FD,$18,$6C,$6F,$8F,$83
	.dc.b	$63,$7C,$DF,$90,$F1,$3F,$C3,$11,$83,$F7,$FB,$FB,$75,$FE,$03,$3A
	.dc.b	$7F,$C0,$9F,$60,$F7,$3F,$07,$E5,$FD,$99,$CC,$63,$C9,$FC,$EE,$3F
	.dc.b	$E6,$B3,$F1,$BF,$61,$FE,$54,$7B,$9F,$28,$D5,$7D,$B8,$D3,$FE,$13
	.dc.b	$F9,$8F,$D0,$B9,$CB,$F5,$31,$8C,$FF,$6D,$5F,$E7,$FB,$EB,$7F,$C6
	.dc.b	$A9,$F2,$DF,$9A,$F8,$D7,$CC,$F2,$8C,$3D,$3B,$75,$37,$0F,$D2,$BF
	.dc.b	$ED,$67,$AE,$1F,$CB,$FD,$2F,$D2,$FD,$87,$F0,$09,$CF,$8E,$9C,$F8
	.dc.b	$BF,$C0,$23,$CD,$7F,$99,$FB,$0F,$99,$7C,$38,$3F,$5F,$18,$9F,$F1
	.dc.b	$A7,$D7,$7F,$9B,$6B,$F6,$FF,$61,$F2,$38,$BE,$C5,$2B,$4F,$E9,$67
	.dc.b	$55,$F9,$89,$FF,$5A,$3F,$C0,$3E,$C3,$E5,$39,$F8,$AF,$DF,$C6,$2F
	.dc.b	$FA,$1D,$3D,$DF,$E0,$7F,$60,$F9,$6F,$E3,$73,$9C,$7F,$85,$CE,$5F
	.dc.b	$8F,$F4,$3F,$E7,$67,$F8,$7E,$5F,$C5,$71,$FE,$9F,$39,$7F,$79,$19
	.dc.b	$5F,$5E,$FB,$0F,$E2,$72,$F9,$3F,$E5,$73,$9C,$3A,$C7,$F2,$EF,$B0
	.dc.b	$FF,$5B,$47,$D8,$6E,$1F,$C2,$7E,$C1,$F5,$3B,$F8,$E9,$7F,$C9,$27
	.dc.b	$F8,$2B,$C6,$3F,$91,$C6,$2B,$FC,$D1,$7D,$38,$FF,$A4,$4E,$5F,$AB
	.dc.b	$19,$BF,$D7,$CE,$5F,$D8,$FD,$87,$F0,$5A,$FE,$A0,$D8,$CF,$0F,$F3
	.dc.b	$28,$C6,$1F,$EB,$76,$D1,$A7,$D1,$8F,$E4,$CE,$AF,$F8,$84,$7C,$1E
	.dc.b	$6F,$5F,$95,$EC,$6E,$3F,$05,$B3,$B8,$FD,$0F,$D7,$BE,$5F,$FA,$DB
	.dc.b	$63,$FA,$5E,$3E,$FA,$FF,$05,$CE,$5F,$F4,$73,$72,$FE,$DD,$F5,$9F
	.dc.b	$81,$7F,$37,$19,$BF,$E2,$B3,$97,$F9,$8C,$66,$BF,$DB,$13,$97,$BE
	.dc.b	$1F,$D6,$23,$16,$CD,$F3,$DF,$68,$FB,$7E,$05,$CB,$F0,$11,$71,$FC
	.dc.b	$34,$DB,$FA,$BE,$77,$D3,$FA,$FF,$A8,$DC,$3E,$0B,$6F,$E8,$B6,$BD
	.dc.b	$BE,$CF,$C8,$3F,$17,$F6,$16,$FE,$6C,$7F,$70,$FA,$C6,$39,$C7,$FB
	.dc.b	$04,$62,$B6,$7F,$7A,$71,$1F,$3A,$B9,$57,$E2,$7E,$C3,$F1,$5E,$75
	.dc.b	$C3,$EB,$B6,$FF,$0E,$DC,$BF,$26,$FE,$83,$CE,$95,$FA,$5F,$5E,$DF
	.dc.b	$EA,$E6,$DF,$E8,$A6,$D7,$F6,$8F,$B0,$FD,$7B,$FC,$A9,$B7,$F9,$6E
	.dc.b	$EF,$F1,$2C,$62,$FF,$8E,$B6,$BF,$07,$E8,$6B,$B7,$EF,$36,$3F,$C3
	.dc.b	$7C,$BF,$D5,$6F,$8D,$6A,$7F,$79,$F6,$1F,$BA,$3F,$C3,$BE,$9B,$DA
	.dc.b	$C4,$E8,$9A,$7F,$83,$E7,$51,$8C,$3F,$6F,$F4,$FF,$B5,$3F,$6D,$F1
	.dc.b	$D6,$7F,$47,$16,$FF,$67,$36,$BF,$9C,$FD,$87,$F3,$86,$F9,$FF,$E4
	.dc.b	$68,$C6,$1F,$E1,$07,$FA,$75,$FF,$BD,$DC,$BF,$E7,$87,$35,$D3,$E5
	.dc.b	$B9,$5F,$3F,$72,$61,$FD,$6A,$FF,$DB,$6D,$8C,$E0,$C3,$FC,$09,$F1
	.dc.b	$2E,$AF,$EB,$97,$B3,$FB,$7C,$67,$D3,$F4,$74,$7A,$76,$B5,$DD,$B8
	.dc.b	$FA,$5B,$8F,$C2,$7F,$CC,$BE,$BF,$F9,$8B,$D4,$7D,$1F,$F8,$C2,$3F
	.dc.b	$85,$46,$2F,$FD,$2F,$1B,$9D,$78,$9F,$E1,$BA,$D5,$7F,$80,$BE,$1B
	.dc.b	$15,$AC,$BF,$D3,$4F,$58,$CC,$F4,$BA,$FF,$82,$9B,$87,$8A,$B8,$5F
	.dc.b	$2A,$E0,$D7,$83,$67,$FD,$F6,$DA,$E7,$F1,$3D,$3E,$53,$BF,$D8,$3E
	.dc.b	$8D,$EA,$71,$9C,$C6,$32,$FF,$60,$67,$F1,$67,$35,$BC,$BC,$58,$73
	.dc.b	$8B,$EE,$FE,$7F,$0F,$F6,$B4,$FF,$89,$3A,$FF,$DA,$4E,$E3,$33,$B7
	.dc.b	$FD,$BD,$1B,$7C,$27,$FB,$21,$C6,$FF,$41,$B9,$7F,$9A,$1F,$F2,$54
	.dc.b	$F8,$AE,$3C,$84,$70,$7A,$78,$7F,$EB,$BA,$E5,$87,$A7,$0E,$CD,$8C
	.dc.b	$E1,$F0,$2B,$85,$F9,$6B,$4D,$9F,$0D,$F9,$EB,$73,$E2,$AD,$B6,$B7
	.dc.b	$7E,$BA,$DE,$1E,$7A,$D1,$EB,$70,$7C,$07,$17,$D2,$32,$D9,$E6,$E4
	.dc.b	$97,$EB,$AF,$09,$E9,$39,$B9,$69,$30,$F8,$B1,$96,$C6,$70,$F8,$EF
	.dc.b	$9E,$FC,$35,$DA,$3D,$75,$F1,$5C,$61,$E8,$9E,$97,$CA,$BA,$91,$F8
	.dc.b	$D5,$C3,$0F,$5B,$DC,$C5,$74,$C3,$C3,$1E,$53,$DD,$7E,$17,$07,$92
	.dc.b	$FC,$07,$67,$38,$79,$5E,$C6,$30,$F0,$D7,$33,$CD,$1D,$AB,$9B,$F9
	.dc.b	$03,$CA,$74,$BF,$2C,$F0,$8F,$11,$B3,$B6,$1E,$B7,$8C,$6F,$0F,$C1
	.dc.b	$3E,$1B,$EF,$3C,$8F,$D0,$DF,$E3,$08,$FC,$E3,$F5,$DE,$1F,$E5,$BE
	.dc.b	$C3,$F8,$ED,$FF,$1E,$6D,$FF,$B5,$9B,$40,$BE,$33,$94,$C9,$FE,$31
	.dc.b	$3C,$1D,$CF,$73,$4F,$ED,$E3,$47,$37,$8C,$7F,$B9,$B9,$7F,$D2,$B3
	.dc.b	$99,$FE,$99,$B7,$96,$B4,$F9,$FB,$2C,$F9,$E9,$E5,$62,$3D,$3E,$95
	.dc.b	$9F,$F4,$94,$7F,$AD,$1D,$B6,$74,$47,$FA,$BF,$C0,$7F,$AA,$32,$F2
	.dc.b	$F8,$3C,$C9,$3F,$EB,$EF,$B0,$FF,$7F,$BE,$83,$C9,$1A,$70,$7C,$C7
	.dc.b	$CD,$1C,$27,$FD,$6A,$FF,$B4,$A7,$A7,$26,$8D,$F4,$C8,$F0,$35,$F0
	.dc.b	$77,$31,$FF,$2C,$C7,$FB,$4B,$EB,$9F,$61,$F1,$5B,$19,$8F,$91,$F6
	.dc.b	$1E,$FE,$1F,$C1,$4A,$B6,$FA,$9B,$A7,$FB,$31,$B3,$C2,$B9,$57,$E2
	.dc.b	$E5,$F1,$DE,$12,$47,$97,$DB,$AF,$58,$B8,$75,$9F,$B9,$F6,$0F,$B0
	.dc.b	$F5,$78,$13,$0F,$F7,$7C,$E5,$B3,$C0,$8F,$C3,$F8,$AA,$FD,$49,$E0
	.dc.b	$D9,$F8,$D3,$F8,$BF,$B0,$78,$B4,$95,$FE,$FE,$87,$2E,$23,$EF,$C7
	.dc.b	$07,$53,$CE,$7A,$79,$3D,$7A,$65,$73,$8C,$BD,$63,$2F,$48,$FC,$F7
	.dc.b	$D8,$3E,$87,$CD,$DD,$2B,$F7,$5C,$1F,$F0,$02,$B5,$18,$E4,$FB,$B1
	.dc.b	$88,$FC,$AB,$67,$85,$FC,$1F,$A2,$EE,$CA,$E4,$7F,$07,$AE,$17,$F8
	.dc.b	$8D,$57,$FB,$3F,$AF,$FC,$9D,$FE,$89,$7B,$79,$35,$EE,$49,$EA,$27
	.dc.b	$6E,$1A,$F3,$8E,$DF,$61,$23,$4F,$87,$EB,$B1,$1E,$18,$BA,$B3,$FC
	.dc.b	$FE,$0B,$FE,$15,$E6,$7F,$91,$2B,$58,$7F,$25,$AF,$CE,$BE,$5B,$DE
	.dc.b	$F2,$3D,$37,$C6,$7F,$56,$D8,$FE,$5D,$F5,$EF,$B0,$E1,$1F,$C1,$A7
	.dc.b	$4F,$F8,$F6,$FF,$8E,$56,$E3,$E2,$BF,$AC,$6F,$56,$8D,$F5,$C8,$C9
	.dc.b	$0F,$E2,$73,$C5,$BE,$7F,$62,$3D,$BB,$2B,$A3,$C3,$1E,$48,$F8,$FE
	.dc.b	$E7,$F3,$1D,$FF,$47,$5F,$69,$AF,$B6,$36,$FC,$66,$BF,$C9,$A1,$18
	.dc.b	$DF,$E8,$CF,$F9,$7E,$31,$B7,$F1,$6A,$DC,$FE,$23,$EC,$1F,$5F,$FD
	.dc.b	$0D,$B1,$98,$FE,$8D,$F6,$0F,$87,$D7,$A1,$2B,$93,$AA,$FE,$99,$F6
	.dc.b	$0F,$A2,$A4,$AF,$CB,$FD,$7F,$D5,$7F,$9F,$CB,$E3,$78,$92,$37,$EF
	.dc.b	$FF,$57,$FB,$0F,$E8,$11,$FD,$19,$BC,$3F,$01,$BC,$1F,$CD,$31,$7D
	.dc.b	$4F,$1F,$99,$98,$7E,$A2,$D1,$FD,$8E,$BF,$A7,$5F,$F3,$CF,$82,$7D
	.dc.b	$87,$F6,$76,$E1,$FB,$06,$BF,$D6,$3E,$C3,$F8,$D3,$FD,$6A,$3E,$43
	.dc.b	$BE,$2E,$47,$F7,$19,$F8,$D1,$FB,$AF,$B0,$57,$F7,$56,$C6,$63,$87
	.dc.b	$89,$9F,$A9,$93,$C5,$FE,$F9,$ED,$BD,$DE,$5F,$33,$2B,$B3,$AF,$2F
	.dc.b	$FC,$05,$F6,$0F,$83,$A4,$AF,$E1,$B8,$7B,$B9,$7F,$1B,$F6,$0F,$96
	.dc.b	$FF,$05,$36,$BF,$89,$B5,$FE,$F6,$F8,$DF,$E2,$FF,$10,$FB,$0F,$F0
	.dc.b	$93,$63,$31,$8F,$89,$91,$C1,$AF,$F8,$1B,$EB,$9D,$9C,$BD,$69,$3D
	.dc.b	$9F,$F0,$F7,$D8,$3E,$96,$92,$7F,$94,$C6,$5F,$E5,$76,$F9,$2F,$F1
	.dc.b	$33,$63,$75,$C5,$F1,$4F,$F8,$AF,$EC,$3E,$FC,$79,$67,$89,$9E,$B9
	.dc.b	$1F,$CE,$27,$7C,$1F,$B8,$6F,$6F,$F8,$33,$1F,$CC,$D6,$EF,$8B,$FE
	.dc.b	$26,$AD,$BF,$E5,$39,$DB,$9E,$7F,$4F,$97,$23,$C1,$1E,$72,$38,$3E
	.dc.b	$0F,$B9,$FE,$45,$FB,$0F,$F2,$76,$7F,$C3,$1B,$9F,$F8,$81,$B5,$FE
	.dc.b	$63,$5D,$7C,$79,$E4,$FF,$56,$9D,$A7,$6F,$F5,$AC,$9F,$1F,$FE,$55
	.dc.b	$6C,$66,$3F,$CB,$1F,$61,$DB,$2F,$F3,$97,$D8,$2F,$FC,$66,$70,$BF
	.dc.b	$F2,$5B,$67,$F9,$2B,$73,$FF,$1C,$DF,$3A,$5F,$37,$FC,$FD,$19,$8C
	.dc.b	$47,$7C,$3F,$C8,$31,$8B,$F0,$3A,$AE,$11,$DB,$36,$3E,$88,$F3,$DF
	.dc.b	$26,$E1,$93,$C3,$5C,$1F,$2C,$73,$AD,$47,$8C,$B1,$E2,$34,$F0,$39
	.dc.b	$C6,$AB,$C1,$1F,$E9,$B9,$E0,$E8,$E0,$2F,$D6,$EE,$78,$BE,$13,$71
	.dc.b	$A7,$85,$FA,$5F,$F4,$0C,$7C,$F6,$C6,$5F,$39,$AF,$29,$F0,$FF,$E8
	.dc.b	$88,$FA,$71,$87,$15,$FD,$3E,$75,$1E,$79,$D3,$8F,$8E,$CB,$FE,$B9
	.dc.b	$6A,$69,$1F,$83,$97,$FA,$F7,$4F,$9F,$1C,$4C,$ED,$FE,$20,$D7,$F0
	.dc.b	$B1,$97,$F0,$F5,$C6,$3E,$59,$9D,$7F,$D7,$95,$C5,$F7,$61,$F7,$2B
	.dc.b	$85,$F4,$BF,$31,$E0,$C3,$C6,$E2,$3B,$E1,$CE,$3F,$D5,$F7,$CE,$31
	.dc.b	$1F,$E9,$88,$C5,$72,$6C,$75,$3F,$D6,$77,$E9,$3F,$C1,$AF,$A4,$FF
	.dc.b	$34,$D7,$F9,$02,$3F,$D8,$4D,$AF,$F6,$C7,$C7,$5A,$DC,$7F,$96,$29
	.dc.b	$39,$97,$D8,$7F,$02,$AF,$F2,$EB,$63,$FD,$8A,$FF,$BC,$63,$3B,$F1
	.dc.b	$8D,$5F,$8F,$AB,$D1,$5F,$EB,$69,$CB,$FE,$80,$8C,$DF,$FB,$EE,$7F
	.dc.b	$A3,$4F,$F9,$E2,$9F,$60,$F9,$CE,$DD,$3E,$B9,$CC,$62,$B8,$56,$DF
	.dc.b	$4D,$6E,$31,$87,$AB,$CC,$F4,$1C,$FE,$C3,$FA,$85,$FC,$08,$CB,$E3
	.dc.b	$8F,$F6,$2B,$93,$1E,$D6,$AB,$FD,$D9,$92,$B7,$1D,$EB,$67,$FA,$23
	.dc.b	$E5,$3F,$C0,$CD,$EA,$FA,$5F,$5C,$FB,$0F,$A9,$5A,$BF,$F7,$AB,$71
	.dc.b	$FF,$27,$FD,$83,$E2,$77,$8F,$DB,$8C,$65,$FE,$96,$6E,$1F,$EF,$EC
	.dc.b	$BF,$C1,$53,$9B,$F0,$F9,$3C,$7E,$57,$F0,$9F,$60,$F0,$71,$7F,$EA
	.dc.b	$08,$C3,$63,$7C,$7C,$1B,$1B,$E6,$FC,$77,$89,$FE,$11,$8C,$1F,$EA
	.dc.b	$A7,$1A,$7F,$36,$8F,$F7,$EF,$D8,$3D,$CF,$C0,$79,$5F,$5A,$B4,$04
	.dc.b	$EE,$BF,$15,$3B,$E2,$FF,$3A,$C7,$F9,$94,$FC,$77,$D8,$7F,$93,$9F
	.dc.b	$F5,$D9,$9B,$FB,$71,$97,$B8,$F9,$5F,$D1,$3A,$CB,$F5,$51,$8B,$FD
	.dc.b	$A5,$7F,$9E,$6F,$FD,$8F,$2F,$96,$FC,$D7,$C6,$BE,$6F,$8C,$61,$E7
	.dc.b	$D9,$A9,$B7,$FA,$57,$FD,$AA,$FF,$B4,$BC,$37,$A3,$FA,$5F,$B0,$FD
	.dc.b	$EC,$E6,$31,$E2,$7E,$FE,$BC,$9B,$FE,$63,$EC,$3E,$43,$EE,$8C,$EF
	.dc.b	$FC,$36,$72,$FF,$A7,$BE,$BB,$FC,$CF,$8F,$ED,$FE,$C3,$E4,$4E,$7D
	.dc.b	$CF,$DB,$46,$27,$EB,$62,$BF,$F5,$4C,$7E,$FF,$EC,$3E,$43,$A8,$CC
	.dc.b	$FE,$B2,$72,$FD,$F8,$7B,$9F,$C0,$FE,$C1,$F2,$FF,$C6,$E7,$38,$7F
	.dc.b	$0B,$9C,$BF,$EF,$8F,$9E,$FF,$38,$3F,$C3,$FE,$37,$F8,$AE,$3F,$8B
	.dc.b	$9C,$BF,$BC,$8C,$AF,$AF,$7D,$87,$F1,$39,$7C,$97,$F2,$B9,$CE,$1D
	.dc.b	$63,$F9,$77,$D8,$7F,$AB,$E3,$F3,$1C,$3F,$EC,$88,$C6,$3F,$C2,$BE
	.dc.b	$B9,$F5,$3C,$3E,$3A,$61,$FC,$96,$7F,$83,$3C,$8F,$9B,$7C,$B0,$FE
	.dc.b	$89,$3A,$C3,$F3,$33,$9A,$F0,$CE,$6B,$C5,$7F,$CB,$63,$13,$E6,$FB
	.dc.b	$0F,$E1,$35,$FD,$3D,$B1,$9E,$1F,$E6,$91,$8C,$7F,$C5,$ED,$A3,$4F
	.dc.b	$A7,$1F,$CA,$1D,$5F,$F1,$48,$F8,$5E,$0F,$8F,$CA,$F2,$37,$1F,$84
	.dc.b	$D9,$DC,$7C,$37,$F8,$27,$D8,$3E,$5B,$FA,$F3,$67,$FA,$76,$3F,$0E
	.dc.b	$B7,$5A,$3F,$A7,$7D,$67,$E0,$B3,$FF,$7E,$E1,$C6,$BF,$9F,$CE,$70
	.dc.b	$FC,$EC,$E6,$BF,$BA,$4E,$6B,$AD,$7F,$5D,$8C,$4B,$37,$CF,$7D,$13
	.dc.b	$E9,$78,$36,$FF,$01,$17,$0F,$C3,$4D,$BF,$A9,$E7,$7D,$2F,$AF,$FA
	.dc.b	$6D,$C3,$DE,$DB,$FA,$0D,$AF,$6F,$B3,$F1,$8F,$C1,$7D,$85,$BF,$99
	.dc.b	$6D,$7F,$B5,$DF,$F8,$47,$EB,$19,$6E,$BF,$B1,$C6,$2F,$8D,$F0,$7F
	.dc.b	$43,$19,$9F,$E5,$B3,$97,$FC,$EF,$19,$BF,$C8,$7D,$87,$E3,$BC,$DF
	.dc.b	$C7,$B7,$0F,$B6,$DB,$FC,$63,$6B,$F4,$7E,$7F,$E7,$4F,$CF,$7A,$F6
	.dc.b	$FF,$59,$16,$FF,$65,$36,$95,$FC,$6B,$EC,$3F,$6A,$FC,$46,$D7,$E6
	.dc.b	$FB,$BF,$C5,$91,$8A,$E3,$3F,$D5,$E3,$15,$FC,$0B,$93,$FA,$B3,$D7
	.dc.b	$B3,$F2,$3F,$15,$C9,$B1,$DD,$B7,$FC,$1F,$EC,$3F,$70,$7F,$89,$7D
	.dc.b	$47,$B7,$8B,$FE,$11,$36,$BF,$C2,$F5,$A9,$C9,$FA,$4F,$A7,$FD,$D9
	.dc.b	$FB,$CF,$8E,$B1,$92,$57,$EE,$63,$2F,$F6,$6F,$B0,$FE,$8A,$DF,$43
	.dc.b	$FC,$9D,$18,$DF,$E1,$CA,$3F,$AB,$BC,$7C,$ED,$D7,$06,$E1,$F9,$2A
	.dc.b	$FE,$1D,$5C,$7D,$CE,$BF,$17,$D9,$95,$FC,$0F,$E2,$9D,$3F,$D9,$A3
	.dc.b	$94,$FF,$99,$67,$3E,$A7,$9B,$AB,$CF,$BD,$BF,$03,$72,$F7,$37,$6F
	.dc.b	$5E,$88,$FF,$36,$7B,$1F,$C1,$FD,$57,$D1,$7F,$62,$8C,$37,$E1,$FF
	.dc.b	$0B,$3A,$C7,$FC,$39,$5A,$6F,$8F,$AB,$E5,$F1,$56,$73,$A7,$FA,$B6
	.dc.b	$72,$D8,$E7,$3B,$7F,$C2,$4D,$BE,$D1,$89,$E0,$D3,$FD,$6D,$5D,$1A
	.dc.b	$78,$0A,$6F,$0F,$52,$FB,$07,$D2,$3D,$4E,$B1,$87,$BE,$3F,$DC,$2F
	.dc.b	$8C,$E6,$FA,$BF,$E2,$7E,$BF,$F6,$7B,$63,$FD,$AC,$ED,$FF,$07,$9E
	.dc.b	$1F,$77,$F4,$6E,$BC,$3F,$F5,$96,$7F,$E1,$4A,$E5,$87,$82,$F6,$F8
	.dc.b	$AB,$77,$FE,$4F,$AF,$F2,$A3,$EA,$AD,$DF,$36,$C6,$EB,$8B,$F0,$DB
	.dc.b	$5E,$0A,$E1,$87,$C4,$9C,$9A,$BE,$77,$EB,$9D,$30,$F8,$B1,$AB,$E9
	.dc.b	$1E,$13,$58,$7B,$6B,$D0,$63,$0F,$85,$87,$78,$CD,$7D,$08,$F1,$B8
	.dc.b	$AE,$4E,$4E,$47,$C5,$C3,$C4,$78,$E7,$8C,$79,$AB,$95,$F8,$23,$89
	.dc.b	$ED,$34,$75,$25,$F8,$09,$1C,$63,$2F,$CA,$78,$C7,$D9,$7A,$E1,$F1
	.dc.b	$1E,$D1,$97,$89,$E9,$AF,$2B,$F2,$4C,$9C,$2B,$B4,$F1,$23,$82,$47
	.dc.b	$18,$CB,$D4,$F2,$C7,$A6,$33,$1D,$4D,$38,$AF,$C7,$C7,$03,$33,$C4
	.dc.b	$8E,$08,$7E,$3C,$F8,$38,$7D,$4B,$F8,$78,$7D,$F9,$E8,$6C,$F7,$DF
	.dc.b	$C7,$24,$7D,$AC,$3E,$8C,$F6,$9F,$2C,$79,$EF,$B4,$F2,$F0,$F8,$63
	.dc.b	$E6,$8F,$E0,$91,$FA,$23,$2F,$C7,$8F,$DA,$47,$EA,$3E,$C1,$F1,$96
	.dc.b	$65,$7F,$16,$9C,$B6,$7F,$D1,$FC,$BF,$D3,$A7,$33,$FE,$BA,$6C,$67
	.dc.b	$B7,$7C,$3F,$E9,$69,$46,$2B,$FD,$9D,$1C,$27,$57,$99,$4F,$07,$FC
	.dc.b	$4E,$D7,$C1,$1B,$8E,$51,$D2,$B1,$E5,$E7,$D2,$93,$CA,$32,$D2,$1F
	.dc.b	$EE,$2E,$BF,$F6,$A6,$3D,$7C,$09,$3E,$27,$CD,$F6,$1F,$C9,$34,$F8
	.dc.b	$91,$A9,$F8,$8E,$9D,$F2,$70,$E5,$91,$FE,$42,$BF,$F6,$D3,$67,$6D
	.dc.b	$3F,$B8,$7C,$0F,$F9,$71,$FF,$24,$FD,$73,$EC,$3E,$43,$63,$31,$F2
	.dc.b	$FE,$C3,$FC,$0B,$9F,$F8,$1A,$B5,$5F,$EF,$26,$C6,$6B,$86,$9D,$3E
	.dc.b	$3A,$5F,$81,$EF,$ED,$E2,$39,$45,$BE,$D3,$F6,$FE,$C1,$F6,$1E,$6E
	.dc.b	$F9,$87,$F9,$8A,$72,$D9,$E0,$47,$DF,$F1,$55,$61,$FE,$20,$AE,$31
	.dc.b	$8C,$BF,$DC,$B1,$89,$F8,$15,$F8,$DF,$AF,$FF,$06,$5F,$F9,$C9,$AF
	.dc.b	$10,$1C,$CF,$F8,$B9,$71,$F3,$F1,$25,$73,$7B,$47,$7F,$23,$C1,$D1
	.dc.b	$2B,$C1,$19,$7C,$51,$97,$C4,$78,$63,$F4,$9F,$60,$FA,$0F,$57,$74
	.dc.b	$AF,$F4,$5C,$E5,$B3,$C0,$8F,$BF,$E8,$6B,$DA,$78,$36,$7E,$34,$FE
	.dc.b	$CF,$EB,$FF,$21,$3E,$45,$EE,$64,$D7,$BB,$27,$FD,$3B,$3A,$6C,$E5
	.dc.b	$E5,$F6,$1D,$48,$F0,$FA,$EA,$E8,$E7,$1E,$38,$FE,$0D,$E5,$FF,$89
	.dc.b	$38,$BF,$CC,$38,$7E,$7B,$97,$EC,$E0,$8C,$3F,$72,$B8,$4E,$9B,$19
	.dc.b	$DB,$A9,$FC,$4F,$EB,$DF,$61,$89,$FF,$5C,$D7,$07,$FC,$67,$5B,$9E
	.dc.b	$1D,$7C,$7A,$E5,$7F,$B0,$30,$4E,$A3,$13,$A7,$F0,$33,$A7,$87,$9D
	.dc.b	$E3,$23,$CF,$B2,$BC,$22,$EB,$FE,$84,$A5,$7C,$8F,$73,$F9,$76,$FF
	.dc.b	$9A,$C3,$FD,$A6,$D7,$F2,$3A,$A7,$84,$7C,$06,$BF,$96,$9D,$39,$21
	.dc.b	$CA,$3B,$9F,$CF,$7E,$C1,$F5,$F4,$93,$FE,$E4,$FB,$07,$C3,$EF,$D2
	.dc.b	$93,$FE,$E8,$BF,$EF,$3A,$FF,$A8,$2F,$9C,$FF,$BB,$3E,$C1,$F4,$3F
	.dc.b	$D2,$DB,$19,$8F,$E9,$DF,$5D,$F7,$23,$87,$C6,$6C,$8F,$FB,$DA,$1E
	.dc.b	$FF,$F5,$5F,$AF,$FE,$6F,$3F,$B6,$6E,$5F,$EB,$26,$F4,$FE,$4A,$DF
	.dc.b	$61,$C7,$E6,$66,$1F,$C3,$B0,$E1,$1E,$48,$FE,$C5,$F0,$4F,$B0,$A4
	.dc.b	$8E,$18,$FE,$7B,$EC,$3F,$8A,$BF,$A0,$AF,$93,$3A,$3F,$8D,$B8,$7F
	.dc.b	$86,$B7,$0F,$F6,$4D,$7C,$69,$FE,$23,$F6,$0A,$FE,$E8,$D8,$E8,$1C
	.dc.b	$3F,$C6,$25,$E2,$70,$F8,$59,$3C,$29,$1F,$DF,$FD,$B7,$B9,$DB,$E6
	.dc.b	$25,$ED,$D7,$93,$FE,$05,$FB,$07,$D8,$52,$57,$93,$0F,$90,$F9,$BE
	.dc.b	$C1,$F2,$DF,$E0,$D6,$C6,$63,$19,$7F,$84,$7E,$25,$F6,$14,$95,$F0
	.dc.b	$21,$F1,$5B,$24,$65,$FF,$0D,$7D,$77,$E2,$23,$87,$AB,$D4,$8F,$CA
	.dc.b	$FB,$07,$D2,$FF,$88,$5B,$19,$8C,$1D,$3E,$46,$46,$48,$7F,$48,$B4
	.dc.b	$7F,$8A,$FE,$C3,$EF,$75,$7F,$24,$9E,$B5,$BB,$D7,$8F,$23,$EE,$DF
	.dc.b	$07,$FB,$1B,$55,$7E,$71,$DC,$F1,$7F,$AC,$CE,$E7,$51,$8D,$BE,$2F
	.dc.b	$99,$23,$BC,$79,$C8,$E0,$FC,$7C,$BD,$CF,$F2,$37,$D8,$7F,$90,$24
	.dc.b	$D2,$74,$EB,$96,$4F,$E6,$27,$2F,$E6,$B1,$7C,$7F,$F9,$45,$B1,$98
	.dc.b	$FF,$2A,$7D,$87,$E6,$3C,$0F,$F0,$2F,$D8,$7F,$2C,$AF,$F0,$3D,$F0
	.dc.b	$7F,$BC,$B6,$5C,$3F,$D5,$23,$0F,$F5,$6B,$E9,$4B,$E8,$A3,$31,$8A
	.dc.b	$EC,$DC,$3F,$AF,$BC,$27,$73,$D7,$3F,$F1,$3D,$F5,$8F,$15,$F0,$6C
	.dc.b	$7B,$1E,$18,$D8,$FF,$11,$DF,$47,$DF,$5C,$5A,$6C,$F4,$5F,$83,$0F
	.dc.b	$23,$70,$EF,$3E,$33,$89,$4E,$38,$73,$C3,$C1,$1F,$D6,$B0,$F0,$34
	.dc.b	$D5,$F5,$7C,$EF,$C0,$7F,$C0,$67,$42,$94,$CC,$FD,$26,$F9,$2F,$8A
	.dc.b	$DC,$7D,$38,$C3,$CA,$7F,$A7,$CE,$9B,$1C,$07,$C7,$D6,$99,$70,$47
	.dc.b	$E1,$63,$FE,$B6,$D3,$E6,$47,$F9,$52,$7E,$74,$E6,$3E,$41,$93,$8E
	.dc.b	$DF,$90,$72,$78,$35,$FC,$6C,$F1,$7F,$1F,$1C,$5C,$CF,$17,$31,$C7
	.dc.b	$5F,$F6,$1C,$E5,$F9,$77,$E4,$8C,$4F,$D0,$8C,$61,$EA,$8E,$47,$F6
	.dc.b	$19,$DB,$FE,$38,$9D,$BE,$A9,$FE,$9F,$87,$BC,$F6,$46,$CD,$1F,$E6
	.dc.b	$87,$51,$FE,$D8,$F8,$DF,$85,$1F,$EA,$2B,$B3,$93,$FA,$DF,$D8,$7F
	.dc.b	$02,$6E,$FA,$C3,$FD,$07,$3F,$E8,$49,$C9,$23,$FD,$11,$7F,$E0,$6B
	.dc.b	$E1,$0A,$E8,$EB,$CC,$F2,$BD,$39,$FB,$5B,$EC,$1F,$3D,$CB,$0F,$E8
	.dc.b	$B9,$7F,$7B,$AD,$4E,$6F,$8D,$F0,$AF,$5C,$62,$79,$79,$72,$79,$7D
	.dc.b	$87,$F4,$67,$E0,$C6,$72,$F5,$CF,$F9,$5A,$72,$70,$F7,$35,$97,$CC
	.dc.b	$9D,$57,$FA,$63,$36,$1E,$C9,$CD,$7D,$8F,$95,$FE,$FF,$8F,$F8,$29
	.dc.b	$B8,$7F,$A8,$3E,$B9,$F6,$1F,$2A,$73,$7C,$7B,$7E,$6F,$D8,$3E,$23
	.dc.b	$51,$86,$65,$FE,$16,$C7,$FC,$02,$FF,$86,$2B,$83,$9F,$23,$9F,$CA
	.dc.b	$FD,$8F,$B0,$FE,$9D,$3B,$8C,$78,$32,$3F,$C4,$33,$F5,$36,$FF,$5B
	.dc.b	$BC,$1B,$C9,$CF,$9B,$F0,$7E,$27,$F8,$32,$F8,$C6,$3C,$AF,$C3,$FD
	.dc.b	$83,$DD,$FC,$1F,$97,$F5,$72,$FF,$61,$CE,$70,$FC,$14,$E7,$8F,$F4
	.dc.b	$71,$FE,$60,$3F,$0D,$F6,$1F,$E4,$97,$FC,$6E,$66,$FF,$2F,$19,$7F
	.dc.b	$D9,$FF,$2B,$F9,$D7,$3A,$7E,$6A,$31,$7F,$AA,$AF,$F3,$AD,$FF,$90
	.dc.b	$A5,$F2,$DF,$9A,$F8,$C7,$CD,$71,$8C,$3C,$FB,$35,$36,$FF,$38,$FF
	.dc.b	$B3,$1F,$F2,$6F,$88,$F4,$7F,$4B,$F6,$0F,$1B,$F5,$B3,$9A,$FD,$24
	.dc.b	$E5,$E3,$B7,$EB,$3E,$C3,$E4,$70,$7F,$BB,$2F,$D2,$FE,$E2,$33,$C1
	.dc.b	$FC,$33,$EB,$BF,$CC,$18,$7E,$BF,$EC,$3E,$44,$E7,$DE,$FD,$BC,$62
	.dc.b	$7F,$D9,$F5,$FE,$A7,$7F,$6D,$F6,$1F,$21,$CF,$C4,$FE,$FA,$31,$90
	.dc.b	$53,$DD,$FE,$07,$F6,$0F,$96,$FE,$17,$39,$C7,$F8,$14,$E5,$FF,$39
	.dc.b	$FC,$FF,$F9,$91,$FE,$0D,$97,$F1,$0D,$7F,$93,$CE,$72,$FE,$2F,$39
	.dc.b	$AF,$F7,$3F,$D7,$BE,$C3,$F8,$9C,$BE,$4F,$F8,$DC,$E6,$BA,$C7,$F1
	.dc.b	$EF,$B0,$7C,$65,$3F,$8D,$FD,$7B,$EA,$71,$7F,$E8,$8A,$DB,$FC,$6E
	.dc.b	$7F,$7E,$E6,$B8,$FA,$1C,$34,$FE,$5B,$5A,$5B,$7F,$42,$C3,$9C,$6A
	.dc.b	$BF,$95,$FD,$87,$EF,$A3,$F9,$AB,$6B,$FA,$1B,$F2,$63,$3E,$7F,$02
	.dc.b	$65,$FC,$BE,$75,$5C,$27,$DB,$E6,$FB,$BA,$ED,$FA,$DB,$87,$A1,$B7
	.dc.b	$FD,$4B,$3F,$C8,$BF,$D6,$BE,$B9,$F3,$1F,$D2,$1B,$87,$87,$6C,$4C
	.dc.b	$8F,$E8,$5F,$59,$F6,$EF,$94,$FF,$4A,$8C,$4F,$E3,$72,$61,$D1,$FE
	.dc.b	$AF,$3A,$7F,$25,$19,$BF,$EC,$F3,$95,$8B,$E7,$BE,$E9,$F7,$BC,$3B
	.dc.b	$8F,$DB,$8B,$8F,$D8,$9B,$7F,$43,$D1,$FA,$BF,$5F,$F5,$9B,$87,$CF
	.dc.b	$6D,$FD,$86,$D7,$C7,$F6,$BE,$31,$F8,$EF,$B0,$B7,$FE,$D4,$D7,$FB
	.dc.b	$16,$1F,$DD,$3E,$B5,$96,$A3,$0D,$C7,$FC,$0B,$3A,$BE,$58,$FF,$74
	.dc.b	$8C,$5F,$E2,$BE,$C3,$F1,$BE,$7F,$E3,$9B,$7F,$87,$6C,$EE,$3F,$15
	.dc.b	$3F,$A4,$F4,$64,$7E,$87,$D9,$B5,$F9,$E9,$B9,$7E,$1E,$6D,$7D,$BF
	.dc.b	$B0,$FD,$63,$FE,$F2,$6D,$FE,$7B,$BF,$FC,$23,$18,$BF,$F1,$3B,$6B
	.dc.b	$F8,$84,$39,$3F,$A5,$47,$5E,$DF,$C5,$7C,$5F,$26,$C7,$76,$E3,$FB
	.dc.b	$4F,$B0,$FD,$F9,$FE,$1A,$F5,$DE,$EE,$2F,$F7,$D3,$6B,$FC,$45,$37
	.dc.b	$2F,$D9,$FD,$37,$EE,$0F,$DC,$FC,$7D,$C3,$F6,$51,$6F,$F7,$73,$6B
	.dc.b	$FC,$87,$F6,$1F,$CE,$27,$F8,$2D,$FE,$86,$3F,$A2,$46,$0F,$E6,$0E
	.dc.b	$E7,$8F,$A7,$FE,$38,$8C,$57,$18,$C7,$A1,$97,$5E,$1C,$9F,$96,$77
	.dc.b	$F8,$CC,$63,$DC,$61,$FC,$46,$7F,$A0,$7C,$5B,$A7,$FA,$D5,$6E,$B4
	.dc.b	$FF,$A4,$63,$39,$FB,$BA,$3C,$DB,$79,$DB,$8F,$8A,$FD,$B1,$87,$A3
	.dc.b	$6B,$FC,$D3,$EC,$FF,$03,$AF,$F5,$2B,$6B,$F8,$43,$5F,$EA,$CD,$7F
	.dc.b	$BF,$F8,$7F,$E8,$DF,$55,$F4,$9C,$3E,$2B,$0F,$29,$B5,$CA,$6B,$FE
	.dc.b	$08,$6D,$9B,$38,$66,$1F,$E1,$5A,$E4,$D7,$93,$6B,$93,$9F,$B0,$FF
	.dc.b	$09,$F2,$FF,$AE,$9B,$5F,$EC,$25,$F4,$6F,$4B,$5E,$EF,$E6,$AB,$4D
	.dc.b	$D7,$DF,$1F,$DE,$6F,$9C,$62,$11,$8D,$7F,$99,$3F,$E3,$D8,$FF,$5D
	.dc.b	$46,$3C,$1F,$F5,$03,$AB,$EC,$DC,$B7,$5D,$EB,$6D,$32,$78,$2B,$8B
	.dc.b	$E5,$6B,$E4,$8E,$2D,$38,$DF,$9D,$E1,$58,$9D,$CF,$A6,$7F,$DB,$77
	.dc.b	$E5,$69,$C0,$A7,$52,$C7,$1C,$3E,$33,$DE,$3D,$17,$DD,$87,$0C,$3D
	.dc.b	$33,$B3,$95,$FC,$99,$F4,$CE,$8D,$E1,$E8,$3D,$2E,$6F,$AE,$1E,$29
	.dc.b	$CC,$72,$33,$7D,$6F,$C5,$19,$7D,$73,$D5,$C5,$7B,$2F,$C3,$1C,$8F
	.dc.b	$89,$19,$AC,$47,$C5,$CF,$FB,$B3,$4F,$03,$8C,$7E,$83,$70,$FE,$F8
	.dc.b	$D9,$FC,$F4,$61,$E2,$7A,$6F,$DC,$F8,$29,$E0,$FA,$63,$D6,$65,$C4
	.dc.b	$F0,$8C,$9F,$6A,$33,$97,$52,$47,$E8,$E3,$2F,$E5,$4E,$33,$F8,$08
	.dc.b	$CC,$7E,$16,$3B,$19,$7C,$D7,$AB,$F3,$C6,$0F,$31,$4E,$B7,$FA,$EB
	.dc.b	$E1,$5C,$5B,$3C,$0D,$9E,$5B,$F0,$D7,$18,$FD,$1C,$F0,$7E,$01,$83
	.dc.b	$F8,$63,$F3,$DF,$80,$60,$FE,$29,$F6,$0F,$8C,$FB,$B1,$FE,$33,$8C
	.dc.b	$38,$0E,$1D,$B5,$FC,$2B,$9E,$26,$7C,$EB,$FC,$7D,$18,$AF,$F4,$65
	.dc.b	$6E,$BC,$1D,$5A,$6C,$FF,$47,$3C,$37,$97,$8B,$5F,$2C,$DA,$FF,$59
	.dc.b	$47,$A5,$AF,$A2,$3D,$8F,$03,$5E,$4F,$1F,$62,$47,$86,$93,$D4,$91
	.dc.b	$97,$C9,$D4,$F8,$0F,$74,$FF,$6D,$A7,$7F,$94,$91,$ED,$FB,$0F,$D6
	.dc.b	$EB,$C2,$B7,$3E,$1A,$DB,$FD,$DE,$B7,$AF,$F9,$A5,$FF,$6E,$C7,$3E
	.dc.b	$5F,$89,$1F,$D1,$E3,$B7,$3E,$08,$E0,$FE,$E9,$9F,$A7,$CE,$F9,$9F
	.dc.b	$5E,$FA,$FA,$4A,$FA,$9F,$61,$EE,$DB,$DF,$5B,$9F,$B0,$D7,$FC,$1C
	.dc.b	$D8,$F8,$B3,$C7,$4F,$F4,$14,$BE,$3F,$C2,$4B,$E7,$ED,$57,$9C,$5C
	.dc.b	$3C,$33,$F8,$3F,$AF,$7D,$87,$AB,$C0,$98,$7E,$2E,$72,$D9,$E0,$47
	.dc.b	$D3,$F1,$B5,$FF,$13,$4F,$06,$CF,$C6,$9F,$C9,$7D,$83,$C5,$A4,$AF
	.dc.b	$CD,$47,$DC,$7F,$C6,$0E,$DC,$C6,$DC,$F2,$F4,$F2,$7A,$F5,$49,$E7
	.dc.b	$19,$7A,$C6,$5E,$B1,$FA,$5F,$B0,$7D,$07,$9B,$C0,$95,$FA,$D9,$CB
	.dc.b	$67,$81,$1F,$4F,$D1,$57,$FC,$91,$3C,$1B,$3F,$0A,$7F,$69,$F5,$FF
	.dc.b	$EB,$39,$FD,$D2,$F6,$F2,$6B,$A6,$4E,$A3,$2F,$08,$E7,$F6,$12,$39
	.dc.b	$BD,$BE,$BB,$11,$FE,$5B,$B4,$E3,$9B,$F6,$70,$AF,$E1,$1E,$5F,$F8
	.dc.b	$8E,$73,$D5,$E8,$33,$B4,$C7,$F8,$95,$6E,$7F,$DB,$0D,$80,$FA,$F7
	.dc.b	$D8,$71,$8F,$F3,$74,$ED,$FF,$1D,$DF,$F1,$9A,$DC,$7F,$1F,$7F,$44
	.dc.b	$DE,$AD,$1B,$EB,$91,$92,$1F,$E7,$B9,$E2,$DF,$3B,$B1,$1E,$DD,$95
	.dc.b	$D1,$E1,$8F,$24,$7C,$6F,$6B,$F9,$76,$DF,$99,$DB,$F8,$EB,$76,$F8
	.dc.b	$73,$BC,$3E,$3B,$5F,$F4,$74,$A3,$15,$FC,$9B,$5C,$57,$03,$FA,$05
	.dc.b	$61,$FF,$4B,$7D,$7B,$EC,$3F,$A0,$36,$3F,$A3,$35,$FF,$4E,$7D,$83
	.dc.b	$E1,$77,$E8,$49,$DB,$FD,$26,$78,$65,$FC,$E3,$EC,$1F,$47,$FD,$35
	.dc.b	$B7,$FD,$4D,$AF,$FA,$A7,$EB,$FE,$CB,$FE,$AB,$87,$C4,$71,$24,$7F
	.dc.b	$5B,$97,$BD,$FD,$67,$EC,$3F,$9C,$CF,$FA,$CD,$B8,$FF,$4A,$6F,$17
	.dc.b	$E3,$A5,$6F,$B0,$E5,$F3,$32,$7F,$B4,$5A,$3F,$B3,$57,$F4,$89,$FC
	.dc.b	$BF,$C0,$3E,$C3,$FB,$53,$63,$FB,$83,$5F,$F6,$37,$D8,$7F,$17,$7F
	.dc.b	$D9,$11,$F3,$5D,$F1,$72,$3F,$BA,$4F,$C6,$D3,$FB,$1F,$D8,$2F,$FB
	.dc.b	$C3,$6F,$FB,$F3,$5E,$FE,$26,$BE,$0E,$46,$9F,$F0,$07,$B7,$FE,$30
	.dc.b	$8F,$F6,$D5,$3E,$5F,$A9,$23,$FC,$15,$3F,$C5,$F1,$FE,$EB,$F6,$0F
	.dc.b	$B0,$FF,$05,$36,$BF,$C2,$0D,$7F,$DC,$D5,$F0,$EB,$FB,$EF,$D8,$3E
	.dc.b	$4F,$FC,$24,$DA,$FF,$0C,$35,$FF,$77,$3E,$79,$FF,$02,$FC,$FB,$E8
	.dc.b	$BF,$C3,$4D,$AF,$F1,$03,$5F,$F7,$B2,$F8,$8D,$92,$36,$FF,$88,$FE
	.dc.b	$BB,$F1,$B1,$CB,$D4,$EA,$47,$FC,$59,$F6,$0F,$A5,$FF,$15,$36,$33
	.dc.b	$18,$3A,$7C,$84,$8C,$90,$67,$FE,$20,$C1,$3F,$E3,$5F,$B0,$FC,$35
	.dc.b	$6B,$BB,$F3,$55,$DE,$30,$7F,$2A,$71,$7F,$E2,$AA,$E2,$DE,$9E,$84
	.dc.b	$AF,$F2,$35,$7E,$8D,$FF,$24,$57,$0B,$FD,$43,$63,$F9,$BB,$54,$F1
	.dc.b	$7C,$62,$3E,$C4,$FA,$3C,$AF,$9C,$7B,$7E,$26,$57,$81,$AF,$2B,$7B
	.dc.b	$9F,$E5,$7F,$B0,$FF,$30,$4F,$F4,$86,$A6,$8E,$AE,$1C,$D2,$7F,$CC
	.dc.b	$93,$A9,$FF,$31,$62,$FA,$EA,$4A,$FF,$35,$7D,$87,$3D,$FF,$AB,$FD
	.dc.b	$82,$FF,$86,$1C,$6B,$FC,$E0,$D7,$E1,$B6,$7F,$CB,$B5,$CE,$55,$CD
	.dc.b	$FE,$C7,$19,$8C,$47,$7B,$FE,$35,$18,$BE,$CE,$AB,$84,$7F,$1F,$6D
	.dc.b	$7F,$6E,$8D,$9E,$8A,$E6,$EF,$0F,$C9,$CF,$92,$31,$87,$C0,$78,$45
	.dc.b	$3A,$57,$27,$D3,$5C,$B0,$F2,$36,$B8,$CF,$53,$89,$CC,$61,$E8,$6C
	.dc.b	$F1,$76,$71,$3E,$21,$C6,$38,$1F,$35,$FF,$78,$B5,$F5,$9F,$EB,$87
	.dc.b	$0F,$D0,$F2,$5F,$0D,$FE,$B9,$9F,$EB,$6D,$8C,$CF,$16,$BF,$EB,$FF
	.dc.b	$8E,$FF,$52,$A6,$91,$FF,$54,$BE,$FC,$FF,$D9,$5C,$3F,$06,$7F,$C4
	.dc.b	$73,$9D,$BF,$D8,$1A,$7E,$0A,$72,$7C,$1B,$E2,$FC,$33,$31,$F8,$88
	.dc.b	$CB,$C6,$3C,$87,$19,$EB,$A7,$FB,$6A,$FC,$18,$78,$EB,$77,$CE,$F7
	.dc.b	$1F,$DA,$CC,$DF,$FB,$5A,$33,$1F,$DB,$EF,$9B,$E5,$36,$7F,$90,$30
	.dc.b	$F2,$D7,$1C,$3F,$DB,$11,$EC,$3F,$D8,$73,$EC,$9D,$3F,$C3,$A3,$53
	.dc.b	$FE,$E6,$F8,$EF,$F7,$34,$7F,$AB,$70,$B3,$93,$FD,$C9,$F5,$FF,$BF
	.dc.b	$6E,$FA,$C3,$FC,$A7,$3F,$E5,$49,$C9,$23,$F9,$7D,$FF,$84,$AF,$84
	.dc.b	$2B,$A3,$AF,$2F,$CA,$F4,$E8,$ED,$6F,$B0,$7C,$DF,$2C,$3F,$A6,$65
	.dc.b	$F1,$4F,$8D,$3A,$7D,$37,$CA,$31,$87,$F3,$0A,$E2,$F9,$A3,$38,$7F
	.dc.b	$BB,$27,$FC,$39,$7F,$0B,$CB,$F1,$FD,$87,$F5,$08,$C1,$30,$F7,$C7
	.dc.b	$FB,$36,$31,$7F,$3A,$31,$3E,$1A,$E7,$ED,$7A,$DB,$7E,$B9,$FF,$3F
	.dc.b	$FC,$A3,$AF,$E4,$B6,$75,$2F,$B0,$7D,$1E,$AE,$56,$BE,$C1,$F1,$18
	.dc.b	$C3,$FA,$8C,$62,$75,$3F,$5B,$0F,$F0,$84,$60,$EE,$E2,$B8,$FC,$B7
	.dc.b	$DC,$FB,$0F,$EA,$35,$DE,$31,$E0,$48,$F3,$CF,$D0,$E0,$47,$F9,$3E
	.dc.b	$38,$DF,$F6,$47,$51,$9E,$5F,$C4,$F8,$7F,$E1,$B8,$E7,$18,$F2,$7F
	.dc.b	$1B,$F6,$0F,$6F,$F1,$3E,$57,$D7,$CB,$FB,$7C,$E7,$0F,$9D,$9F,$F8
	.dc.b	$7F,$C8,$FC,$AC,$7F,$9D,$0C,$FD,$87,$F9,$65,$F9,$E6,$6F,$F3,$B1
	.dc.b	$97,$FB,$DF,$CC,$FE,$8D,$CE,$BF,$A0,$8C,$61,$FB,$0A,$FF,$44,$5F
	.dc.b	$DA,$97,$CB,$7E,$83,$E3,$1F,$37,$C6,$30,$F3,$EA,$D4,$DB,$F5,$BF
	.dc.b	$EE,$17,$F0,$BE,$2B,$D0,$FD,$5F,$D8,$3C,$6F,$D9,$CE,$6B,$E1,$CE
	.dc.b	$5E,$3B,$7E,$CB,$EC,$3E,$47,$07,$F8,$62,$FD,$2F,$EF,$A3,$3C,$1F
	.dc.b	$72,$3F,$DA,$FF,$5D,$FE,$73,$C3,$E7,$FD,$87,$C9,$AD,$52,$B4,$FF
	.dc.b	$08,$9D,$57,$FA,$4E,$7F,$D7,$8F,$D8,$FB,$0F,$96,$E7,$E2,$BF,$83
	.dc.b	$C6,$2F,$C3,$ED,$FF,$0A,$FB,$07,$CB,$7F,$12,$C7,$FC,$F5,$8F,$E0
	.dc.b	$A7,$2F,$EB,$3E,$7F,$FC,$F0,$FE,$1B,$2F,$E2,$FA,$7A,$27,$39,$7F
	.dc.b	$1E,$9C,$D7,$FA,$A7,$EB,$DF,$61,$FC,$6A,$5F,$25,$FC,$8A,$73,$5D
	.dc.b	$63,$F9,$2F,$D8,$7F,$AF,$23,$F7,$AD,$C3,$F9,$5F,$D7,$BE,$A7,$7F
	.dc.b	$1B,$2F,$F3,$73,$FC,$09,$E3,$1F,$CA,$A3,$15,$FE,$50,$BE,$9A,$FF
	.dc.b	$42,$9C,$BF,$52,$33,$7F,$D4,$67,$2F,$F5,$3F,$B0,$FE,$11,$7F,$CF
	.dc.b	$9B,$7F,$D3,$1F,$97,$19,$F2,$FF,$9D,$C6,$AB,$94,$63,$8B,$2F,$9F
	.dc.b	$E1,$3D,$1D,$36,$BD,$AD,$C3,$E1,$B7,$1F,$53,$FD,$3B,$EB,$DF,$2D
	.dc.b	$4F,$0E,$D7,$0F,$E5,$33,$A7,$F8,$1F,$D6,$7D,$CC,$3F,$2B,$39,$C7
	.dc.b	$FA,$DC,$E6,$BF,$DC,$D3,$9A,$FE,$D7,$39,$A0,$9D,$57,$F7,$B9,$D5
	.dc.b	$36,$7C,$EF,$D7,$3E,$C7,$83,$6F,$EE,$C5,$C7,$F0,$73,$6F,$E8,$79
	.dc.b	$FF,$3B,$EB,$FE,$7B,$70,$FE,$3F,$C4,$8C,$61,$F6,$5B,$7F,$03,$DB
	.dc.b	$F9,$47,$E2,$7E,$C2,$E1,$F2,$23,$FB,$AF,$D6,$32,$CE,$3F,$E0,$08
	.dc.b	$C5,$6C,$FE,$5E,$E2,$3F,$D4,$3B,$FF,$70,$9F,$F0,$F4,$E7,$1F,$C5
	.dc.b	$7D,$87,$E2,$3C,$FF,$C4,$B6,$BD,$6D,$C3,$F2,$4D,$AF,$CE,$7A,$1F
	.dc.b	$A1,$3F,$45,$EB,$DB,$FD,$3C,$5B,$FD,$6C,$DA,$FE,$2B,$F6,$1F,$AF
	.dc.b	$BF,$ED,$ED,$AF,$D0,$77,$FF,$87,$23,$15,$C3,$B7,$FC,$33,$CD,$FD
	.dc.b	$46,$31,$87,$E3,$BD,$1E,$15,$A6,$C6,$4F,$DD,$FD,$87,$EE,$4F,$F1
	.dc.b	$1F,$AA,$F6,$F1,$3A,$26,$9F,$E2,$B8,$CB,$FB,$5F,$A6,$FD,$89,$FB
	.dc.b	$2F,$8E,$B5,$F8,$48,$B3,$FB,$B9,$B9,$FF,$36,$FB,$0F,$E6,$AD,$F4
	.dc.b	$7F,$C7,$71,$8C,$3F,$3F,$68,$FE,$94,$FF,$B7,$9D,$3F,$A4,$73,$5D
	.dc.b	$7E,$57,$9D,$F4,$F1,$26,$1F,$E1,$EB,$FF,$6B,$36,$33,$87,$FB,$F3
	.dc.b	$E2,$9D,$1F,$D6,$E3,$C1,$3F,$E5,$79,$CF,$A9,$E8,$E9,$F4,$ED,$71
	.dc.b	$EC,$DC,$3D,$0D,$D7,$E1,$3F,$E5,$AF,$5F,$FA,$0D,$7F,$9F,$5B,$5F
	.dc.b	$09,$AF,$F0,$26,$BF,$DF,$BE,$4F,$FB,$37,$AA,$FA,$2F,$93,$F6,$0F
	.dc.b	$A2,$7A,$5F,$E9,$E9,$E9,$7C,$1D,$CE,$B0,$FF,$55,$46,$2B,$84,$7F
	.dc.b	$12,$BE,$91,$88,$F1,$65,$67,$31,$C8,$E1,$ED,$FE,$3D,$FF,$4D,$08
	.dc.b	$FF,$0A,$F3,$E2,$31,$8F,$8D,$FF,$50,$4E,$9E,$3F,$60,$FB,$1F,$C3
	.dc.b	$E3,$FE,$A5,$8C,$4F,$F8,$A4,$FE,$52,$E1,$FF,$4E,$B6,$38,$4F,$FA
	.dc.b	$A7,$3F,$C8,$CE,$F5,$F9,$97,$CA,$B4,$F4,$AA,$AB,$6F,$C1,$31,$3A
	.dc.b	$8E,$93,$D8,$C1,$98,$93,$F3,$67,$9D,$74,$6D,$79,$AB,$75,$DE,$72
	.dc.b	$F9,$4E,$51,$AA,$E5,$5E,$09,$C9,$0E,$4E,$5E,$93,$C6,$3D,$11,$E5
	.dc.b	$70,$FC,$53,$04,$38,$9D,$67,$C2,$E5,$F6,$4F,$28,$F9,$53,$E1,$9C
	.dc.b	$C7,$18,$F0,$BE,$63,$07,$84,$C1,$23,$8B,$E2,$31,$1C,$AF,$F1,$0F
	.dc.b	$E2,$A3,$04,$3E,$51,$F8,$87,$4E,$23,$8B,$F8,$F3,$D2,$7B,$63,$CF
	.dc.b	$3F,$67,$E3,$3D,$91,$82,$4F,$E5,$A3,$2F,$82,$7E,$3B,$F7,$63,$2F
	.dc.b	$63,$D6,$71,$3C,$07,$19,$EA,$FE,$56,$38,$E1,$F8,$68,$E5,$5E,$57
	.dc.b	$89,$8A,$E3,$2A,$E3,$1C,$9B,$3C,$27,$4D,$9D,$34,$D4,$F0,$7D,$27
	.dc.b	$94,$FD,$03,$9A,$F2,$47,$C1,$3F,$83,$57,$67,$F8,$1C,$FF,$09,$79
	.dc.b	$1F,$C3,$BF,$01,$FF,$51,$BF,$C3,$5A,$E7,$83,$84,$62,$3F,$A8,$87
	.dc.b	$7C,$5F,$E8,$E6,$E5,$FD,$59,$B3,$FC,$3B,$E4,$FF,$AC,$D7,$FA,$53
	.dc.b	$14,$62,$F8,$61,$6F,$FC,$7B,$87,$27,$F8,$55,$65,$A7,$C0,$7D,$D7
	.dc.b	$D2,$30,$7C,$4F,$2F,$84,$7F,$A5,$A3,$FD,$FA,$D5,$8F,$96,$76,$48
	.dc.b	$DB,$E5,$E7,$7C,$0F,$FA,$BA,$BD,$9D,$DD,$89,$7F,$E5,$68,$FF,$59
	.dc.b	$FD,$97,$CA,$77,$1E,$29,$DB,$E2,$D7,$FB,$FD,$6D,$BA,$FC,$18,$FE
	.dc.b	$85,$7D,$39,$B1,$32,$FC,$17,$C1,$A7,$F9,$A2,$FE,$AE,$5E,$AF,$43
	.dc.b	$E6,$7D,$83,$EC,$A9,$2F,$88,$FB,$2F,$81,$8F,$F8,$52,$B8,$4E,$9A
	.dc.b	$FA,$CF,$F3,$BB,$8D,$3F,$DA,$B5,$CA,$73,$F1,$DE,$22,$3C,$FD,$CA
	.dc.b	$F3,$16,$9C,$DF,$FB,$82,$3F,$D8,$73,$E8,$FB,$27,$D9,$7B,$BE,$5A
	.dc.b	$4E,$63,$F1,$93,$A8,$C1,$0F,$05,$3E,$3B,$75,$27,$58,$78,$67,$F2
	.dc.b	$DF,$64,$F0,$E9,$2B,$E5,$C3,$97,$11,$ED,$8E,$0E,$A7,$9C,$F4,$F1
	.dc.b	$BA,$FC,$04,$AE,$71,$97,$AC,$65,$E9,$1F,$A9,$FB,$27,$D9,$79,$BC
	.dc.b	$19,$5F,$7A,$72,$D9,$E0,$47,$D3,$F4,$95,$FA,$93,$C1,$B3,$F0,$A7
	.dc.b	$F7,$1F,$63,$FE,$B2,$9F,$C5,$AF,$7F,$26,$BA,$64,$EA,$32,$F0,$8E
	.dc.b	$7F,$65,$23,$9B,$DB,$EC,$31,$1F,$87,$B4,$E3,$89,$D5,$FE,$6C,$95
	.dc.b	$FF,$0D,$F2,$7F,$C2,$F5,$AC,$3F,$D9,$F5,$F9,$74,$D9,$DD,$FF,$9D
	.dc.b	$9B,$3F,$ED,$66,$BF,$E5,$1D,$BF,$8D,$7D,$8B,$EC,$BC,$11,$27,$F5
	.dc.b	$F5,$C2,$B5,$1F,$B2,$7F,$40,$DE,$CE,$67,$4E,$C9,$19,$21,$FA,$99
	.dc.b	$E2,$DF,$86,$F1,$91,$ED,$DB,$5D,$1E,$18,$F4,$C7,$C7,$F7,$3F,$9A
	.dc.b	$6F,$F9,$BA,$FE,$0E,$D7,$F7,$19,$27,$84,$7B,$9A,$FE,$F6,$74,$E4
	.dc.b	$87,$28,$EE,$7F,$43,$FB,$17,$D9,$52,$4F,$F1,$8F,$B2,$7C,$56,$FE
	.dc.b	$0A,$4F,$F1,$CB,$47,$F4,$DA,$E7,$3F,$C8,$3E,$C9,$F4,$BF,$D4,$5B
	.dc.b	$19,$8F,$EA,$9F,$61,$F5,$A3,$F8,$B4,$BE,$37,$81,$23,$F9,$5C,$BE
	.dc.b	$03,$FA,$E7,$D8,$FF,$5F,$27,$9D,$D3,$EC,$B3,$F3,$73,$0F,$E6,$74
	.dc.b	$8F,$0C,$7F,$65,$F8,$37,$D9,$52,$47,$F3,$AF,$B2,$FE,$1A,$FF,$2B
	.dc.b	$9F,$8A,$FF,$3E,$6B,$9E,$2F,$EE,$13,$F0,$23,$F9,$97,$D9,$2B,$FB
	.dc.b	$A3,$63,$FA,$3B,$5F,$E7,$12,$F1,$77,$EC,$C9,$DB,$FD,$F7,$DB,$7C
	.dc.b	$67,$4F,$9A,$95,$D1,$D7,$95,$FE,$03,$FB,$27,$C2,$52,$57,$F5,$8C
	.dc.b	$3E,$1B,$FD,$6B,$EC,$9F,$2F,$FE,$0B,$6C,$66,$31,$97,$F8,$3F,$E2
	.dc.b	$1F,$47,$49,$5F,$D9,$21,$F1,$3B,$24,$65,$FF,$0C,$7D,$87,$DF,$8E
	.dc.b	$1E,$B7,$52,$3F,$DB,$BE,$C9,$F5,$BF,$E1,$F6,$C6,$63,$07,$4F,$91
	.dc.b	$91,$92,$1F,$D9,$ED,$1F,$E2,$9F,$B2,$FB,$BE,$A7,$F1,$DA,$57,$67
	.dc.b	$85,$EF,$AA,$47,$F8,$0A,$F8,$FA,$68,$91,$87,$16,$EF,$E3,$1F,$DE
	.dc.b	$67,$49,$B3,$B8,$ED,$C1,$F2,$3C,$D9,$18,$21,$CE,$3D,$87,$CE,$F7
	.dc.b	$7F,$C8,$DF,$65,$FE,$63,$D7,$DC,$DC,$FD,$16,$8C,$3E,$B7,$FC,$94
	.dc.b	$FF,$83,$DC,$BB,$E6,$EA,$76,$EB,$FF,$2A,$64,$FB,$0A,$4A,$FF,$14
	.dc.b	$FD,$97,$F9,$83,$2F,$91,$F6,$4B,$F1,$9C,$6B,$FC,$5E,$D7,$FC,$40
	.dc.b	$D9,$FB,$15,$CE,$55,$CD,$FA,$91,$98,$C4,$77,$BF,$83,$18,$BE,$CE
	.dc.b	$AB,$84,$7C,$66,$D7,$DF,$8D,$9E,$8A,$E6,$EF,$0F,$F3,$5C,$F9,$23
	.dc.b	$18,$7C,$07,$84,$53,$A5,$72,$7D,$35,$CB,$0F,$23,$6B,$8C,$F5,$38
	.dc.b	$9C,$C6,$1E,$86,$CF,$17,$67,$13,$E2,$1C,$63,$81,$F3,$5F,$F5,$0B
	.dc.b	$5F,$59,$FE,$A4,$3A,$E1,$5F,$0F,$96,$F8,$7C,$57,$F4,$A6,$C6,$67
	.dc.b	$94,$E5,$FF,$5C,$7C,$77,$FA,$E5,$34,$8E,$8F,$55,$FF,$B4,$F6,$F8
	.dc.b	$51,$FB,$59,$F8,$93,$98,$F8,$06,$4E,$3A,$FF,$87,$F4,$FC,$3C,$65
	.dc.b	$FC,$4C,$F1,$73,$1E,$43,$8D,$F8,$1C,$47,$0A,$E9,$A7,$C7,$C3,$C7
	.dc.b	$18,$AF,$F6,$6C,$62,$F9,$36,$3C,$E7,$4A,$FF,$65,$D7,$18,$FF,$5C
	.dc.b	$BF,$E3,$7C,$3F,$D6,$8D,$BF,$04,$F9,$8F,$F4,$54,$FC,$33,$FD,$75
	.dc.b	$F2,$BB,$8F,$F5,$C5,$57,$FD,$8B,$17,$EC,$7F,$7D,$7F,$EA,$46,$CF
	.dc.b	$FA,$55,$FE,$49,$19,$DF,$81,$D7,$0F,$F0,$06,$18,$9F,$F5,$AC,$61
	.dc.b	$E9,$CF,$BF,$B2,$7C,$FF,$86,$30,$D9,$E3,$3A,$78,$98,$8F,$47,$97
	.dc.b	$23,$BF,$D9,$7F,$44,$7D,$51,$9D,$3F,$D7,$18,$AB,$FD,$5D,$5A,$38
	.dc.b	$FB,$9B,$9F,$A1,$3A,$AF,$85,$7E,$F8,$C4,$FF,$4F,$F9,$6F,$EF,$B8
	.dc.b	$7F,$B3,$1B,$7F,$EB,$DF,$B0,$7D,$97,$D1,$9C,$DF,$1E,$BF,$A9,$F6
	.dc.b	$4F,$89,$D4,$61,$99,$7F,$B6,$31,$FE,$FE,$FF,$B6,$AB,$83,$9F,$23
	.dc.b	$9F,$CA,$FD,$FF,$B2,$FF,$5F,$4E,$E3,$1E,$04,$8F,$F7,$1C,$FF,$76
	.dc.b	$6C,$71,$E4,$E5,$EB,$FE,$27,$C5,$FF,$04,$D7,$08,$C7,$97,$F8,$DF
	.dc.b	$B2,$FF,$28,$CF,$E0,$BC,$BF,$A9,$C9,$FE,$DF,$9C,$BF,$52,$33,$C5
	.dc.b	$FD,$E6,$BF,$CA,$A7,$E2,$BE,$CB,$FC,$82,$FF,$BE,$0C,$D7,$DC,$8C
	.dc.b	$BF,$EE,$BF,$98,$FC,$BB,$9C,$FF,$41,$18,$BF,$D6,$57,$F9,$AA,$FF
	.dc.b	$C0,$11,$FE,$09,$97,$B5,$F9,$DF,$8D,$7C,$EF,$28,$C3,$D3,$AF,$53
	.dc.b	$70,$FD,$0B,$FE,$C3,$7A,$9F,$DD,$BC,$47,$A3,$FA,$9F,$B2,$78,$BF
	.dc.b	$B5,$9C,$D7,$EA,$A7,$2F,$2D,$FF,$91,$6F,$FB,$1F,$B2,$F9,$7B,$FF
	.dc.b	$88,$70,$F6,$57,$E8,$27,$37,$FE,$06,$FB,$0F,$F2,$F3,$FB,$6F,$B2
	.dc.b	$F9,$78,$7F,$8A,$67,$37,$FA,$89,$CD,$71,$F8,$8F,$E0,$1F,$65,$F2
	.dc.b	$E3,$07,$F0,$07,$15,$FE,$2B,$B7,$B9,$FC,$1F,$EC,$9F,$2D,$FC,$4E
	.dc.b	$73,$87,$F0,$89,$CB,$EF,$B7,$CF,$FF,$99,$23,$F8,$6E,$5F,$C5,$B4
	.dc.b	$FF,$20,$4E,$6B,$F7,$F3,$9A,$FF,$12,$FD,$8B,$EC,$BF,$8C,$4B,$E4
	.dc.b	$FF,$92,$4E,$6B,$AC,$7F,$27,$FB,$2F,$F5,$04,$78,$DB,$87,$F0,$CF
	.dc.b	$B2,$7D,$5E,$FE,$36,$5F,$F2,$29,$FE,$02,$F1,$8F,$E5,$11,$8A,$6B
	.dc.b	$FE,$59,$C3,$AB,$FD,$06,$74,$E6,$B8,$F9,$DF,$CC,$7E,$CB,$F8,$0E
	.dc.b	$1F,$C7,$9B,$5F,$D2,$5F,$97,$19,$F3,$78,$13,$2F,$E6,$F3,$AA,$E1
	.dc.b	$3E,$6F,$2F,$CF,$D5,$6F,$D8,$DC,$3E,$13,$6F,$D0,$FF,$4F,$FB,$07
	.dc.b	$CB,$D3,$D1,$A5,$C3,$EF,$4E,$9F,$E9,$3F,$5D,$F6,$F0,$FF,$3E,$4E
	.dc.b	$71,$FE,$77,$39,$AF,$F3,$0C,$E6,$BF,$B4,$4E,$6B,$FC,$F9,$39,$AF
	.dc.b	$DF,$CE,$69,$B3,$E7,$BE,$A9,$F5,$BC,$2B,$7F,$6E,$2E,$3F,$81,$9B
	.dc.b	$7F,$3F,$CF,$FA,$9F,$63,$F5,$5B,$87,$CE,$6D,$FD,$76,$D7,$BB,$D7
	.dc.b	$F9,$07,$E1,$BE,$CA,$DF,$C6,$8F,$EE,$1F,$5C,$C7,$38,$FF,$5E,$8C
	.dc.b	$56,$CF,$EE,$6E,$23,$F1,$F1,$F2,$2B,$9D,$7E,$2F,$EC,$BF,$19,$E7
	.dc.b	$5A,$FC,$2B,$6F,$F1,$CD,$CB,$F1,$CF,$E7,$3E,$66,$57,$E8,$7D,$8B
	.dc.b	$7F,$A7,$9B,$7F,$9D,$9B,$5F,$E9,$AF,$B2,$FD,$5B,$FC,$05,$B7,$F9
	.dc.b	$DE,$FF,$F0,$D4,$62,$FF,$D8,$0D,$AF,$BB,$F1,$7A,$EE,$FF,$0B,$72
	.dc.b	$FF,$4E,$9E,$D1,$89,$D4,$7E,$41,$FD,$AF,$D9,$48,$FF,$0D,$FA,$CF
	.dc.b	$73,$1A,$7F,$01,$AB,$9F,$F8,$DE,$B5,$7F,$C0,$67,$F6,$DF,$4D,$23
	.dc.b	$F5,$FF,$1F,$70,$FD,$C4,$DC,$3F,$71,$36,$BF,$D7,$BF,$65,$FC,$D2
	.dc.b	$7F,$DB,$CF,$F4,$4B,$FB,$67,$FB,$65,$D4,$F0,$F4,$BF,$C8,$11,$89
	.dc.b	$D4,$63,$E6,$32,$EB,$C3,$97,$FC,$C6,$DF,$19,$8C,$3B,$88,$FE,$1F
	.dc.b	$3F,$EE,$9F,$8C,$75,$FF,$5A,$AD,$D6,$9F,$F4,$54,$67,$3F,$5F,$47
	.dc.b	$9B,$6F,$3B,$71,$F1,$36,$F9,$B6,$BF,$E0,$07,$B3,$FD,$07,$D5,$7D
	.dc.b	$1F,$FB,$93,$EC,$9F,$47,$FD,$2B,$1F,$F4,$8C,$FF,$BA,$2F,$83,$B9
	.dc.b	$D5,$F2,$A1,$FE,$04,$F9,$66,$7F,$E9,$3B,$E1,$E9,$5A,$DE,$3D,$1E
	.dc.b	$39,$78,$A7,$50,$8F,$F0,$6F,$47,$FA,$BD,$B1,$FE,$B3,$78,$37,$DD
	.dc.b	$FC,$93,$67,$9C,$F0,$CF,$D1,$3D,$6B,$71,$89,$78,$1F,$E1,$3C,$BD
	.dc.b	$B6,$FB,$07,$D9,$66,$3E,$0C,$61,$F4,$B5,$F9,$66,$A7,$83,$CE,$6C
	.dc.b	$EF,$0F,$2B,$6B,$C2,$FA,$23,$31,$C0,$95,$E0,$F8,$CA,$F8,$0F,$44
	.dc.b	$62,$7A,$CD,$8E,$A5,$9F,$5C,$57,$89,$EC,$8F,$3C,$66,$BC,$91,$CA
	.dc.b	$7C,$46,$C8,$78,$E3,$6F,$63,$35,$EC,$6C,$FC,$0A,$EC,$F7,$6C,$68
	.dc.b	$E4,$75,$8C,$BD,$E2,$CF,$58,$C1,$C4,$F6,$9C,$4F,$49,$48,$F1,$3E
	.dc.b	$4B,$C6,$BD,$6F,$28,$CB,$C2,$BE,$23,$4F,$3C,$73,$8C,$B6,$33,$18
	.dc.b	$23,$F8,$93,$CB,$1B,$7C,$11,$A7,$F2,$D5,$F0,$23,$14,$8C,$4F,$33
	.dc.b	$E1,$9E,$53,$99,$9A,$F5,$9D,$E7,$DE,$49,$E7,$3C,$0F,$09,$F1,$6B
	.dc.b	$EC,$34,$F9,$11,$C2,$79,$47,$D0,$E1,$FE,$12,$FD,$C6,$BD,$0F,$E5
	.dc.b	$B1,$FE,$84,$7F,$CC,$4F,$B5,$B3,$FB,$89,$E3,$39,$9E,$3D,$39,$E0
	.dc.b	$FD,$EF,$D9,$7F,$A7,$27,$FC,$D8,$D7,$F9,$DB,$6F,$9F,$4F,$19,$FE
	.dc.b	$43,$39,$AE,$7C,$59,$94,$E5,$FE,$0F,$59,$C7,$FA,$4C,$7F,$0E,$3F
	.dc.b	$7F,$15,$A5,$3F,$89,$35,$A7,$C2,$7A,$1D,$7E,$4F,$B7,$95,$DF,$DB
	.dc.b	$95,$F2,$23,$2D,$21,$FD,$A3,$B1,$F0,$7F,$4B,$5F,$97,$E0,$7A,$49
	.dc.b	$5F,$E9,$3F,$B2,$F5,$4E,$A3,$EA,$CE,$9F,$EC,$B3,$AD,$FF,$96,$BF
	.dc.b	$16,$0F,$F5,$37,$7D,$3F,$2E,$34,$4E,$0C,$1F,$4F,$2F,$3F,$9F,$F4
	.dc.b	$7E,$BD,$F6,$54,$98,$FF,$AA,$7E,$CB,$D9,$9F,$F3,$EA,$DC,$FF,$AB
	.dc.b	$5A,$FF,$7C,$6E,$BF,$94,$AE,$33,$FD,$8A,$5F,$23,$E1,$24,$73,$F7
	.dc.b	$6B,$CE,$2D,$FE,$7A,$BF,$D6,$13,$FE,$BD,$FB,$27,$D9,$7A,$FE,$5E
	.dc.b	$4E,$63,$F1,$D3,$A8,$C1,$0F,$47,$C8,$6A,$64,$67,$0F,$04,$FE,$5F
	.dc.b	$EC,$9E,$45,$25,$7F,$B3,$E1,$CB,$88,$FF,$0F,$C7,$07,$53,$CE,$7A
	.dc.b	$79,$7D,$7A,$65,$73,$8C,$BD,$63,$2F,$48,$FD,$57,$D9,$3E,$8F,$CD
	.dc.b	$E1,$4A,$FF,$71,$4E,$5B,$3C,$08,$FA,$7E,$92,$BF,$E3,$69,$E0,$D9
	.dc.b	$F8,$53,$FB,$9F,$B1,$FF,$50,$4F,$FB,$BD,$7B,$B9,$35,$D3,$27,$51
	.dc.b	$97,$84,$73,$FB,$29,$1C,$DE,$DF,$61,$88,$FF,$24,$DA,$71,$C5,$FC
	.dc.b	$8A,$15,$FC,$37,$CF,$FF,$0D,$CE,$7A,$FD,$06,$76,$98,$FF,$8D,$A7
	.dc.b	$F9,$5D,$70,$AF,$F0,$EB,$67,$FC,$77,$F6,$2F,$B2,$E3,$1F,$E5,$C9
	.dc.b	$D3,$FE,$48,$C3,$F8,$F5,$6E,$3B,$5F,$F8,$A2,$3F,$47,$18,$3A,$F6
	.dc.b	$70,$38,$F6,$48,$D0,$9F,$D1,$BF,$E6,$C6,$E5,$FD,$11,$EB,$19,$F8
	.dc.b	$5F,$19,$1E,$70,$EC,$C4,$6C,$CC,$FA,$A3,$E6,$7B,$BF,$CE,$76,$FC
	.dc.b	$FD,$79,$1A,$FC,$38,$DB,$E8,$D5,$18,$8F,$F4,$1C,$E1,$D4,$67,$1F
	.dc.b	$A6,$F7,$3F,$A4,$7D,$8B,$EC,$A9,$2B,$E2,$FD,$93,$E2,$77,$E9,$C9
	.dc.b	$F9,$37,$FE,$5A,$77,$7F,$D4,$FE,$C9,$E6,$D2,$4F,$D0,$FB,$2F,$B4
	.dc.b	$FF,$A6,$A5,$F1,$7C,$49,$1B,$F8,$0F,$EB,$DF,$65,$FC,$E4,$9E,$6F
	.dc.b	$F6,$1C,$1F,$59,$AF,$9C,$98,$7D,$EB,$46,$E3,$FB,$47,$C1,$3E,$CA
	.dc.b	$92,$3F,$0B,$F6,$3F,$C3,$DF,$F5,$7C,$FC,$37,$F1,$2D,$73,$C5,$FD
	.dc.b	$CE,$7E,$0C,$7E,$3B,$EC,$95,$FD,$DD,$B1,$98,$E1,$E2,$E7,$D8,$93
	.dc.b	$C5,$FE,$FD,$EE,$3E,$2F,$97,$CD,$4A,$EC,$EB,$CA,$FF,$02,$7D,$93
	.dc.b	$E1,$69,$2B,$F4,$38,$7C,$17,$F4,$7F,$64,$F9,$5F,$F0,$63,$63,$31
	.dc.b	$8C,$BF,$C2,$1F,$42,$FA,$4A,$4A,$FD,$6C,$3E,$2B,$64,$8C,$BF,$E1
	.dc.b	$9F,$B0,$75,$F4,$F5,$A5,$72,$7F,$C3,$DF,$64,$FA,$7A,$49,$FD,$DC
	.dc.b	$65,$FD,$ED,$BE,$47,$B1,$B8,$FF,$74,$CE,$DF,$F7,$5C,$ED,$FE,$0B
	.dc.b	$F6,$5F,$77,$C9,$FF,$22,$CE,$DB,$87,$F8,$C6,$75,$E8,$FF,$AE,$9B
	.dc.b	$7F,$ED,$86,$A6,$DF,$E6,$23,$BF,$67,$32,$3E,$AF,$03,$B7,$0F,$C7
	.dc.b	$F3,$64,$F9,$A3,$2D,$21,$EC,$F7,$7F,$C8,$DF,$65,$FE,$55,$D7,$FC
	.dc.b	$BE,$DD,$3F,$DA,$2D,$AF,$F7,$7B,$FE,$48,$53,$C3,$9F,$99,$23,$F9
	.dc.b	$5C,$EE,$7F,$CD,$78,$BE,$BF,$FC,$AC,$DA,$FE,$62,$D7,$AE,$1F,$EE
	.dc.b	$2F,$B2,$77,$78,$25,$F6,$5F,$CB,$51,$FC,$BF,$0E,$A4,$7F,$9E,$D7
	.dc.b	$2C,$77,$5C,$65,$39,$7A,$E3,$FE,$F2,$3C,$93,$B6,$9E,$5C,$3C,$91
	.dc.b	$CE,$B9,$3B,$C3,$FD,$F6,$DB,$F0,$DF,$1C,$2C,$7A,$AF,$8C,$F8,$6F
	.dc.b	$8B,$67,$67,$A3,$0F,$0C,$73,$8F,$49,$F1,$1F,$3B,$67,$9B,$DE,$78
	.dc.b	$B6,$FA,$D7,$08,$F9,$87,$82,$FC,$AE,$CF,$F4,$43,$F2,$5F,$F4,$63
	.dc.b	$5F,$21,$8F,$2D,$F0,$FF,$E8,$D9,$FE,$8A,$D8,$CC,$F4,$6B,$FE,$93
	.dc.b	$F8,$FF,$F4,$92,$69,$1F,$F4,$B3,$F1,$B0,$FF,$69,$69,$F2,$27,$F0
	.dc.b	$F3,$9D,$BF,$D8,$3A,$FE,$0E,$72,$7C,$A3,$F3,$F1,$AC,$FF,$D4,$D5
	.dc.b	$C6,$FA,$57,$8C,$E1,$87,$81,$E4,$74,$C3,$F2,$11,$97,$C1,$1C,$5C
	.dc.b	$E1,$DE,$3F,$D6,$D7,$DD,$E6,$6F,$0F,$F6,$84,$EE,$3F,$46,$64,$FF
	.dc.b	$19,$39,$BF,$21,$FE,$AD,$7F,$DC,$2F,$FB,$82,$3C,$D3,$D3,$0F,$F1
	.dc.b	$2C,$F0,$7F,$DB,$DF,$1F,$C1,$D5,$7E,$E0,$FF,$20,$46,$5F,$DA,$FD
	.dc.b	$97,$F0,$19,$FF,$17,$B6,$38,$07,$F7,$71,$AD,$F9,$4E,$5E,$97,$FD
	.dc.b	$1B,$1C,$DF,$38,$F5,$F5,$F9,$EF,$C3,$D1,$C3,$EC,$9F,$3D,$C2,$30
	.dc.b	$DC,$FF,$C9,$18,$2B,$71,$E8,$AD,$BE,$AF,$33,$C2,$79,$3E,$CB,$FA
	.dc.b	$74,$64,$98,$FC,$5C,$13,$FE,$B8,$9C,$C7,$D2,$8C,$4F,$C4,$CF,$FC
	.dc.b	$2F,$5E,$49,$CD,$23,$EF,$CE,$AB,$B5,$6A,$7F,$07,$F2,$FF,$E0,$77
	.dc.b	$FC,$BE,$D9,$FA,$3F,$64,$FA,$5F,$9F,$39,$C7,$B5,$7F,$2E,$8C,$5F
	.dc.b	$E3,$BE,$C9,$F1,$3C,$3B,$7E,$BC,$62,$7F,$CD,$AD,$BE,$FE,$46,$AF
	.dc.b	$EE,$E7,$FE,$0F,$BD,$FC,$AF,$E0,$3E,$C9,$E0,$74,$8C,$32,$7F,$CF
	.dc.b	$33,$FE,$03,$6C,$71,$E2,$E5,$CB,$F8,$AF,$0B,$FC,$2D,$5C,$23,$1E
	.dc.b	$57,$E3,$BE,$C9,$EE,$FE,$3B,$CB,$FB,$7C,$1F,$96,$9C,$E1,$F8,$69
	.dc.b	$CF,$27,$C6,$AF,$A9,$E0,$FE,$5B,$EC,$BF,$C9,$97,$FE,$8E,$3F,$CC
	.dc.b	$67,$E3,$9C,$CF,$E8,$BE,$63,$F4,$2F,$E1,$E3,$FC,$E1,$3F,$E9,$69
	.dc.b	$CD,$FE,$E2,$73,$5C,$7D,$BF,$D2,$FC,$6B,$E7,$7F,$53,$82,$73,$18
	.dc.b	$EE,$DD,$5C,$BF,$42,$FF,$B4,$1F,$F5,$2F,$8A,$F4,$7F,$63,$F6,$4F
	.dc.b	$1B,$F6,$B3,$9A,$FD,$4C,$E5,$E3,$CB,$FB,$8F,$B2,$F9,$3B,$7F,$00
	.dc.b	$BF,$4F,$07,$F1,$29,$CB,$F2,$FE,$C3,$FC,$CF,$87,$ED,$3E,$CB,$E4
	.dc.b	$E1,$FC,$22,$72,$FE,$B6,$33,$5C,$7E,$1F,$F8,$3F,$D9,$7C,$8C,$3F
	.dc.b	$92,$E1,$FE,$A8,$7F,$7F,$19,$C3,$FC,$17,$2F,$73,$F8,$7F,$D9,$3E
	.dc.b	$5B,$F8,$9C,$E7,$0F,$E0,$F3,$97,$FD,$97,$F3,$FF,$E6,$F7,$F8,$5E
	.dc.b	$5F,$C5,$5F,$E6,$B1,$9C,$7F,$82,$4E,$5F,$F0,$DF,$D8,$BE,$CB,$F8
	.dc.b	$FC,$BE,$4F,$F9,$1C,$E7,$0E,$B1,$FC,$9B,$EC,$BF,$D5,$31,$FE,$DC
	.dc.b	$6E,$1F,$C4,$7E,$C1,$F5,$7B,$F8,$D9,$7F,$C8,$27,$F8,$13,$C6,$3F
	.dc.b	$97,$C6,$2B,$F9,$BD,$F4,$E2,$FE,$93,$39,$7F,$45,$19,$BF,$F1,$AC
	.dc.b	$27,$51,$8F,$B2,$FE,$09,$3F,$D4,$9B,$19,$E1,$FE,$7F,$18,$C3,$FC
	.dc.b	$A2,$DA,$34,$FA,$51,$FC,$BD,$D5,$FF,$0D,$8F,$85,$CF,$EC,$F3,$7D
	.dc.b	$AD,$C7,$E1,$36,$77,$1E,$A3,$FB,$6F,$D9,$3E,$5B,$FA,$AB,$63,$FA
	.dc.b	$4E,$3F,$06,$BF,$CA,$B3,$97,$FA,$FB,$72,$FE,$E3,$F5,$DF,$83,$7F
	.dc.b	$85,$46,$6F,$E4,$CE,$5F,$0E,$1F,$D8,$A3,$17,$DA,$31,$F1,$1F,$63
	.dc.b	$E7,$A5,$FD,$DF,$06,$DF,$E0,$E6,$E5,$F6,$A6,$D7,$D5,$F4,$3E,$8F
	.dc.b	$D8,$FD,$26,$C7,$B9,$B8,$7C,$F6,$DF,$AB,$DB,$F8,$A7,$E2,$BE,$CA
	.dc.b	$DF,$CB,$BF,$8B,$3F,$DF,$3E,$B9,$8E,$A7,$FB,$44,$62,$B8,$43,$7F
	.dc.b	$EE,$AE,$AB,$86,$3F,$E1,$CB,$E1,$3F,$83,$FB,$2F,$C5,$F9,$DF,$8C
	.dc.b	$6E,$3F,$61,$B3,$B8,$FB,$73,$FA,$2F,$42,$47,$E9,$FD,$9B,$5F,$AD
	.dc.b	$9B,$8F,$E9,$26,$D7,$F8,$D3,$EC,$BF,$62,$FA,$9B,$7F,$97,$EF,$FF
	.dc.b	$0F,$C6,$2F,$FA,$63,$6B,$F0,$7F,$19,$AE,$EF,$C1,$36,$3F,$C3,$3C
	.dc.b	$FF,$D5,$1F,$04,$E9,$FD,$F7,$D9,$7E,$F0,$FF,$15,$FA,$CF,$77,$13
	.dc.b	$A2,$69,$FE,$19,$9D,$46,$33,$FD,$C7,$D3,$7E,$D8,$FD,$BF,$C7,$D9
	.dc.b	$FD,$1C,$5B,$FD,$A4,$DA,$FE,$D1,$F6,$5F,$CE,$9B,$E8,$FF,$90,$23
	.dc.b	$18,$7F,$97,$1F,$E9,$97,$FE,$48,$72,$FF,$B0,$9C,$D7,$4F,$95,$E5
	.dc.b	$7C,$FE,$02,$61,$F8,$AB,$FF,$66,$B6,$33,$87,$F9,$6F,$E2,$9D,$1F
	.dc.b	$D7,$63,$C1,$3F,$E5,$D9,$CF,$AB,$E7,$E8,$F4,$6D,$71,$EC,$DC,$3D
	.dc.b	$0D,$D7,$E0,$BF,$E5,$EF,$B2,$7D,$0B,$CF,$FF,$4B,$FD,$27,$F9,$F2
	.dc.b	$73,$18,$C3,$FD,$C3,$ED,$FF,$4B,$F6,$17,$BC,$39,$57,$FA,$9E,$39
	.dc.b	$7B,$7F,$7E,$3F,$D0,$E7,$FB,$AB,$1F,$F0,$47,$3E,$23,$17,$C9,$DC
	.dc.b	$E9,$D3,$FE,$28,$7C,$81,$D6,$1E,$48,$ED,$7E,$18,$C4,$6F,$BF,$FD
	.dc.b	$33,$18,$AE,$93,$E9,$38,$E1,$DA,$3A,$E3,$F1,$6F,$8B,$FE,$A2,$FB
	.dc.b	$0E,$B8,$76,$5F,$65,$AC,$42,$32,$FA,$9E,$35,$CC,$E1,$1B,$9B,$3C
	.dc.b	$23,$CC,$73,$9C,$FC,$66,$99,$1D,$63,$2F,$78,$CB,$CA,$53,$A8,$C4
	.dc.b	$EA,$FD,$93,$97,$DD,$19,$3E,$14,$65,$F5,$4F,$73,$C4,$43,$A9,$93
	.dc.b	$C5,$1C,$5F,$8F,$5F,$11,$B3,$97,$E5,$4F,$85,$C9,$D6,$BC,$EF,$57
	.dc.b	$CE,$E6,$7C,$EE,$5F,$3B,$E2,$34,$74,$3C,$E6,$49,$1D,$1C,$BE,$07
	.dc.b	$31,$D2,$34,$64,$8E,$49,$1F,$02,$7D,$D5,$F0,$63,$EF,$1E,$43,$E0
	.dc.b	$9E,$93,$CD,$5D,$1E,$46,$4E,$27,$90,$C9,$E8,$AF,$0B,$D0,$D1,$C2
	.dc.b	$3D,$31,$F2,$23,$33,$E5,$25,$7A,$27,$EA,$4F,$3A,$F2,$3C,$E3,$BD
	.dc.b	$7E,$42,$3D,$8F,$38,$F8,$35,$CA,$3A,$CE,$A7,$13,$DE,$3D,$D3,$CD
	.dc.b	$FC,$E7,$47,$EC,$1F,$F8,$03,$19,$E1,$FF,$55,$37,$8B,$F7,$55,$B9
	.dc.b	$CB,$63,$E0,$47,$0B,$C5,$FE,$5A,$AB,$FB,$DF,$91,$BF,$65,$FE,$A3
	.dc.b	$8F,$F0,$5B,$5F,$1C,$70,$C4,$FF,$23,$8F,$D0,$B5,$E8,$66,$36,$DF
	.dc.b	$83,$E5,$32,$3F,$C3,$32,$9E,$C6,$63,$C2,$69,$F1,$79,$7E,$3F,$52
	.dc.b	$4F,$28,$CB,$48,$7F,$59,$F8,$2F,$F6,$A6,$9F,$33,$C0,$F6,$12,$7F
	.dc.b	$C5,$7F,$65,$FD,$0A,$75,$1E,$69,$D2,$8F,$F6,$F4,$EF,$7F,$B4,$FC
	.dc.b	$98,$CF,$4F,$D0,$8F,$E9,$11,$C3,$A7,$04,$BF,$0B,$F6,$3B,$59,$3D
	.dc.b	$0F,$B3,$F6,$0F,$B2,$A4,$C3,$FC,$87,$F6,$5E,$ED,$7F,$19,$5B,$9F
	.dc.b	$F2,$3B,$5F,$54,$6A,$BF,$0B,$5C,$67,$FA,$1D,$3E,$AF,$C2,$48,$E7
	.dc.b	$EE,$D7,$99,$C2,$31,$AB,$1F,$1D,$7E,$4B,$EC,$9F,$65,$EC,$F9,$79
	.dc.b	$8F,$F9,$72,$72,$D9,$E0,$47,$D7,$F1,$F5,$F0,$89,$E2,$E1,$B8,$F8
	.dc.b	$E3,$DB,$3F,$A4,$FB,$27,$8D,$49,$3F,$E6,$E8,$73,$62,$3B,$C7,$07
	.dc.b	$53,$CE,$7A,$79,$3D,$7A,$65,$73,$8C,$BD,$63,$2F,$48,$FD,$BF,$D9
	.dc.b	$3E,$8F,$CD,$E1,$4A,$FF,$40,$CE,$5B,$3C,$08,$FA,$7E,$92,$BF,$1A
	.dc.b	$78,$36,$7E,$14,$FF,$06,$FB,$1F,$C8,$4F,$FA,$39,$7B,$D9,$35,$D3
	.dc.b	$27,$51,$97,$84,$73,$FB,$29,$1C,$DE,$DF,$61,$88,$FB,$36,$9C,$71
	.dc.b	$7D,$79,$4A,$BF,$8D,$F9,$DF,$E2,$8A,$D4,$FE,$66,$B7,$D1,$AD,$A4
	.dc.b	$7F,$AA,$2F,$84,$FF,$31,$6C,$7F,$88,$BE,$C5,$F6,$5C,$23,$F1,$B3
	.dc.b	$A7,$FC,$AB,$7F,$CB,$2B,$71,$FE,$B4,$7F,$50,$DE,$AD,$1B,$EB,$91
	.dc.b	$92,$1F,$9A,$9E,$2D,$F8,$4E,$C4,$7B,$76,$57,$47,$86,$3C,$91,$F1
	.dc.b	$BD,$DF,$E8,$3B,$FE,$6A,$BF,$D9,$2D,$7D,$B1,$B7,$E3,$35,$FD,$6C
	.dc.b	$23,$1B,$FD,$5D,$31,$5C,$0E,$B3,$FD,$43,$EC,$5F,$63,$49,$3F,$ED
	.dc.b	$AF,$B2,$7C,$4E,$FD,$39,$3F,$ED,$EB,$FF,$30,$E5,$F8,$3B,$E7,$3F
	.dc.b	$EE,$2F,$B2,$FE,$94,$DF,$37,$FA,$FC,$62,$3F,$82,$FD,$97,$D6,$7F
	.dc.b	$83,$C3,$E3,$F6,$48,$FF,$76,$CB,$DF,$FE,$D3,$F6,$5F,$CF,$7A,$7F
	.dc.b	$BC,$36,$BF,$CD,$2D,$54,$FA,$AE,$3F,$37,$27,$FD,$F3,$48,$FE,$EB
	.dc.b	$5F,$D2,$E7,$FC,$E7,$F0,$4F,$B2,$FE,$F2,$D8,$FF,$80,$2D,$7F,$90
	.dc.b	$FD,$97,$F1,$37,$F9,$1C,$7C,$C7,$7C,$7C,$8F,$F0,$1C,$FC,$5E,$1F
	.dc.b	$D7,$7D,$92,$FF,$C0,$ED,$BF,$F4,$2C,$EA,$78,$F8,$9A,$F6,$65,$6B
	.dc.b	$7F,$F0,$E6,$1F,$E0,$28,$FE,$6B,$EE,$3E,$2F,$AF,$CD,$49,$E0,$7F
	.dc.b	$A4,$A9,$1F,$E1,$8A,$FE,$2F,$9F,$F0,$3F,$B2,$7C,$9F,$F8,$71,$B8
	.dc.b	$72,$0E,$1F,$0F,$07,$F8,$8B,$EC,$9F,$2D,$49,$3C,$63,$CC,$F3,$F8
	.dc.b	$97,$D1,$FF,$8A,$9B,$19,$8C,$7C,$54,$8E,$0D,$7F,$AA,$FD,$87,$E2
	.dc.b	$DE,$9E,$AF,$32,$3E,$9F,$B2,$7D,$37,$F8,$E1,$B1,$98,$C1,$FE,$3C
	.dc.b	$A7,$C9,$D3,$33,$B3,$E2,$46,$9F,$8D,$F6,$5F,$85,$7C,$B5,$CC,$FF
	.dc.b	$23,$C6,$78,$7F,$99,$37,$9B,$FC,$B0,$DA,$FF,$64,$B5,$5F,$E8,$5E
	.dc.b	$FD,$7D,$08,$F9,$3B,$FB,$F4,$7A,$FC,$B9,$3E,$68,$CB,$48,$7C,$CF
	.dc.b	$73,$FC,$BB,$F6,$5F,$E6,$78,$FF,$6D,$B5,$34,$FC,$A6,$9F,$D3,$5C
	.dc.b	$74,$F2,$39,$F6,$7F,$9A,$31,$7D,$85,$26,$1F,$8F,$FB,$2F,$F3,$2E
	.dc.b	$3F,$EE,$AF,$B2,$5F,$F9,$F4,$E3,$5F,$95,$6B,$FE,$18,$6C,$FF,$5F
	.dc.b	$AE,$72,$AE,$6F,$FB,$DA,$33,$18,$8E,$F7,$FE,$8B,$8C,$5F,$67,$55
	.dc.b	$C2,$3F,$D2,$0D,$AF,$F8,$03,$1B,$3D,$15,$CD,$DE,$1F,$E6,$19,$F2
	.dc.b	$46,$30,$F8,$0F,$08,$A7,$4A,$E4,$FA,$6B,$96,$1E,$46,$D7,$19,$EA
	.dc.b	$71,$39,$8C,$3D,$0D,$9E,$2E,$CE,$27,$C4,$38,$C7,$03,$E6,$BF,$EB
	.dc.b	$13,$96,$1F,$E9,$D8,$CC,$7B,$4D,$1E,$1F,$25,$F0,$FB,$6C,$6A,$7F
	.dc.b	$D3,$93,$93,$FD,$4D,$0F,$8F,$A4,$7F,$D5,$4D,$CB,$EE,$60,$D3,$FC
	.dc.b	$B1,$1F,$7E,$87,$33,$DF,$3B,$E4,$F8,$F1,$AB,$F9,$33,$98,$F8,$A6
	.dc.b	$4E,$37,$DB,$4F,$7D,$70,$8C,$4F,$C7,$8C,$5F,$68,$F1,$61,$E8,$AD
	.dc.b	$C7,$FB,$76,$B6,$F8,$0D,$DF,$FB,$6E,$77,$1E,$3A,$E5,$7C,$30,$FF
	.dc.b	$5C,$D7,$F9,$E5,$FF,$22,$56,$EF,$95,$7F,$41,$9F,$71,$FE,$C0,$F8
	.dc.b	$FE,$11,$AA,$FF,$43,$CA,$73,$1F,$33,$EC,$7F,$7F,$5F,$D2,$DB,$1F
	.dc.b	$E8,$E7,$E8,$C6,$77,$E3,$29,$DD,$F9,$FC,$BF,$1D,$FF,$BF,$E7,$2F
	.dc.b	$F6,$C8,$CD,$78,$7A,$3C,$76,$FB,$27,$CF,$F6,$C7,$FA,$2E,$3F,$DA
	.dc.b	$2B,$53,$9B,$E3,$7C,$2B,$D5,$18,$9E,$5E,$64,$9E,$5F,$65,$FD,$29
	.dc.b	$F8,$11,$9C,$BF,$D7,$B8,$AB,$FD,$89,$5A,$38,$FB,$5B,$9F,$AD,$3A
	.dc.b	$AF,$91,$7E,$C8,$C4,$FF,$9F,$7E,$5B,$FC,$09,$87,$F7,$96,$DF,$FA
	.dc.b	$F3,$EC,$1F,$65,$F3,$A7,$37,$C7,$B3,$E8,$FD,$93,$E2,$35,$18,$66
	.dc.b	$5F,$E0,$5C,$7F,$C0,$2F,$F8,$1E,$B8,$39,$F2,$B9,$FC,$AF,$DC,$FB
	.dc.b	$27,$83,$8A,$FF,$07,$46,$1B,$1B,$E3,$E0,$D8,$DF,$37,$E1,$BC,$3F
	.dc.b	$F0,$5C,$60,$FF,$0C,$38,$E1,$FF,$6D,$FD,$93,$DC,$FC,$1F,$97,$F5
	.dc.b	$67,$31,$8F,$27,$F2,$38,$3C,$4F,$C4,$FD,$97,$F9,$26,$3F,$C4,$86
	.dc.b	$4F,$C4,$39,$BF,$F7,$4F,$CA,$FE,$71,$FB,$F1,$FE,$74,$9F,$F1,$64
	.dc.b	$E7,$9B,$F6,$53,$9A,$E3,$F2,$DF,$98,$F8,$C7,$CE,$FE,$93,$04,$E6
	.dc.b	$31,$DB,$BA,$B9,$7E,$6D,$FF,$67,$BF,$E3,$BF,$1D,$E9,$7E,$93,$EC
	.dc.b	$BF,$71,$39,$8C,$78,$DF,$B2,$AF,$26,$DF,$C1,$FE,$CB,$E5,$5F,$3E
	.dc.b	$5F,$D6,$C6,$37,$57,$FB,$3F,$EC,$3F,$CC,$57,$FA,$DF,$B2,$F9,$15
	.dc.b	$A8,$FE,$62,$A7,$6F,$E9,$A7,$75,$CF,$E1,$FF,$77,$F6,$5F,$27,$0E
	.dc.b	$38,$7F,$AA,$DF,$DD,$46,$70,$FE,$7B,$3F,$E6,$29,$FD,$F7,$D9,$3E
	.dc.b	$5B,$F8,$7C,$E6,$BF,$6B,$39,$AF,$F3,$17,$CF,$FF,$99,$9F,$E0,$D8
	.dc.b	$FF,$10,$AF,$E9,$33,$97,$F7,$B1,$95,$F6,$2F,$B2,$FE,$27,$2F,$92
	.dc.b	$FE,$3F,$39,$C3,$AC,$7F,$22,$FB,$2F,$F5,$64,$7F,$9C,$DB,$87,$F0
	.dc.b	$AF,$B2,$7D,$5E,$FE,$36,$5F,$F1,$B9,$FD,$FB,$C6,$3F,$91,$C6,$2B
	.dc.b	$E9,$1F,$D8,$AF,$AF,$0F,$F4,$59,$D2,$BF,$AF,$3A,$7F,$D3,$73,$A7
	.dc.b	$FD,$3B,$F6,$5F,$C0,$AF,$F8,$FB,$6F,$FA,$83,$F3,$63,$3E,$6F,$F3
	.dc.b	$68,$D5,$72,$8C,$71,$65,$F4,$F9,$7E,$AE,$7B,$5E,$F6,$E1,$F1,$5B
	.dc.b	$8F,$AD,$FE,$9B,$F6,$2F,$96,$A7,$87,$6B,$87,$FA,$86,$74,$FF,$45
	.dc.b	$FA,$EF,$BD,$87,$C3,$C7,$FC,$13,$5A,$7F,$9E,$CE,$AB,$F3,$B3,$9A
	.dc.b	$FE,$E1,$39,$AE,$38,$7F,$56,$8C,$4B,$67,$CE,$FD,$83,$EC,$F8,$17
	.dc.b	$2F,$BD,$17,$1F,$C2,$4D,$BF,$9F,$E7,$FC,$DF,$B1,$F9,$CD,$C3,$FB
	.dc.b	$07,$12,$31,$87,$D8,$6D,$FB,$E1,$AF,$F6,$37,$F1,$26,$3D,$8F,$C6
	.dc.b	$FD,$97,$CE,$66,$3F,$DB,$7E,$B9,$9E,$2F,$FB,$1C,$62,$7E,$85,$7F
	.dc.b	$78,$8C,$4F,$6C,$3C,$18,$FF,$86,$67,$2F,$E3,$3E,$CB,$F1,$5E,$7F
	.dc.b	$E2,$DB,$8F,$B9,$B8,$7E,$51,$B5,$F9,$FF,$3F,$F4,$A7,$E9,$BD,$7B
	.dc.b	$7F,$AB,$8B,$7F,$B1,$9B,$5F,$17,$EC,$BF,$67,$7F,$EE,$96,$D7,$E8
	.dc.b	$BB,$FF,$C3,$B1,$8A,$E1,$DB,$F8,$1F,$47,$2D,$C3,$93,$70,$F9,$F7
	.dc.b	$C6,$7F,$CB,$36,$E7,$FE,$B8,$FE,$E7,$EC,$BF,$7E,$7F,$88,$BD,$47
	.dc.b	$B5,$8B,$FE,$0D,$37,$1F,$F0,$AD,$6A,$72,$7E,$1B,$E9,$BF,$6C,$7E
	.dc.b	$DF,$E3,$6C,$64,$95,$FC,$02,$32,$FE,$1B,$EC,$BF,$9E,$E5,$F8,$B7
	.dc.b	$FA,$8C,$67,$E5,$FF,$C8,$31,$89,$D4,$63,$D0,$CB,$70,$F8,$27,$F4
	.dc.b	$AA,$FE,$13,$7F,$E7,$C6,$A5,$F1,$5D,$BC,$59,$5C,$AB,$FD,$27,$D1
	.dc.b	$FD,$9A,$31,$1C,$67,$FC,$BB,$F1,$2A,$D4,$FB,$7A,$7D,$3B,$7A,$9B
	.dc.b	$87,$2C,$91,$8A,$E1,$39,$AF,$F3,$0F,$D9,$7F,$80,$38,$BF,$D4,$F4
	.dc.b	$8C,$3F,$EA,$A5,$F4,$0F,$3B,$71,$A9,$FF,$5A,$3C,$3D,$47,$D1,$FF
	.dc.b	$70,$8C,$37,$D9,$FC,$2D,$7F,$AB,$63,$5D,$DF,$E9,$16,$C7,$FA,$8D
	.dc.b	$E1,$95,$8F,$F5,$5D,$F4,$6D,$71,$69,$FE,$25,$29,$FE,$92,$8F,$F1
	.dc.b	$B3,$93,$E0,$47,$F8,$1B,$4F,$F6,$77,$D8,$7A,$A7,$4D,$7F,$D3,$6D
	.dc.b	$8D,$E1,$C6,$7F,$C1,$F3,$BC,$B8,$36,$72,$79,$E5,$1F,$EC,$97,$DC
	.dc.b	$63,$E3,$35,$1C,$63,$FD,$96,$D7,$47,$07,$93,$F3,$4D,$CF,$23,$9C
	.dc.b	$C8,$E1,$19,$79,$46,$5E,$47,$9A,$34,$E2,$7B,$D7,$BA,$B2,$FC,$0A
	.dc.b	$E4,$E6,$BB,$90,$E5,$3A,$9F,$8B,$19,$9E,$0E,$8C,$46,$9C,$12,$3C
	.dc.b	$0E,$8F,$8F,$3D,$5F,$86,$64,$F9,$27,$19,$E9,$3C,$A3,$07,$23,$85
	.dc.b	$7D,$13,$E0,$CF,$E0,$2B,$F5,$73,$D0,$FC,$5B,$89,$FA,$B1,$83,$F6
	.dc.b	$2E,$1E,$46,$23,$81,$FE,$5A,$E7,$FE,$CD,$3F,$79,$B1,$9C,$3F,$23
	.dc.b	$CF,$6F,$D0,$7F,$6A,$BE,$13,$DA,$31,$3B,$7E,$1D,$F4,$7F,$8A,$D7
	.dc.b	$07,$F2,$33,$B8,$C4,$70,$AE,$95,$E6,$AD,$5C,$AE,$6F,$18,$C4,$79
	.dc.b	$62,$CF,$18,$D9,$0D,$1E,$49,$DD,$78,$09,$1E,$63,$84,$FA,$DE,$27
	.dc.b	$C0,$33,$1C,$1F,$51,$C4,$DC,$71,$24,$7C,$E8,$E2,$7C,$E8,$CB,$C5
	.dc.b	$F5,$3C,$A3,$B5,$AB,$E1,$4F,$E0,$9F,$C3,$11,$FC,$2F,$37,$E2,$48
	.dc.b	$E0,$87,$DD,$7E,$A1,$F5,$7D,$F9,$1F,$6A,$32,$FD,$C8,$CB,$FE,$A6
	.dc.b	$3F,$86,$BF,$97,$23,$F9,$6F,$1F,$F0,$1E,$84,$8F,$C2,$C7,$E0,$C8
	.dc.b	$E0,$FE,$C1,$D0,$F7,$BF,$C7,$DA,$7E,$6B,$B3,$F1,$44,$AF,$EC,$7F
	.dc.b	$61,$E3,$9D,$47,$E4,$A7,$4F,$F9,$12,$75,$B7,$FA,$F9,$FF,$6C,$C7
	.dc.b	$2E,$6F,$CE,$47,$F5,$98,$EB,$EF,$E0,$8E,$0F,$3E,$3F,$E6,$47,$FB
	.dc.b	$AF,$D7,$3E,$C3,$F4,$CD,$8C,$C7,$EA,$7E,$C3,$DB,$97,$FB,$42,$B5
	.dc.b	$5F,$E5,$A6,$E9,$FE,$5F,$6C,$FF,$B5,$AB,$8D,$04,$BE,$33,$C2,$48
	.dc.b	$F2,$FB,$35,$ED,$17,$0E,$D3,$FB,$DF,$AF,$7D,$87,$AB,$E4,$A6,$1F
	.dc.b	$E1,$29,$CB,$67,$81,$1F,$87,$F1,$35,$FF,$3A,$CF,$06,$CF,$C6,$9F
	.dc.b	$E1,$9F,$60,$F0,$E9,$2B,$FC,$43,$0E,$4C,$47,$FA,$02,$38,$3A,$9E
	.dc.b	$73,$D3,$C7,$EB,$D1,$2B,$9C,$65,$EB,$19,$7A,$47,$F1,$FF,$B0,$7D
	.dc.b	$87,$9B,$C0,$95,$FE,$34,$9C,$B6,$78,$11,$F4,$FD,$25,$7F,$D2,$F3
	.dc.b	$C1,$B3,$F0,$A7,$F9,$77,$D7,$FE,$36,$7F,$D3,$AB,$DB,$C9,$AF,$72
	.dc.b	$4F,$F9,$1A,$74,$D9,$CB,$CB,$EC,$3A,$91,$E1,$F5,$D5,$D1,$CE,$3C
	.dc.b	$71,$FD,$07,$CC,$FF,$0F,$73,$FF,$AC,$F0,$FC,$9D,$9B,$3A,$EB,$FC
	.dc.b	$B4,$FF,$91,$B6,$E4,$47,$FC,$BB,$F5,$EF,$B0,$C4,$7F,$AF,$2F,$93
	.dc.b	$F9,$F6,$A9,$E0,$F2,$EB,$FE,$AD,$0E,$BA,$62,$24,$F8,$23,$2F,$13
	.dc.b	$C7,$F0,$5C,$88,$F9,$BB,$2B,$FE,$D0,$BF,$F7,$D3,$C0,$D1,$FE,$78
	.dc.b	$C1,$ED,$7F,$69,$DF,$F3,$53,$FE,$7C,$6B,$EF,$8F,$9C,$47,$93,$C4
	.dc.b	$C3,$B3,$27,$37,$FB,$B7,$D7,$BE,$C2,$92,$7F,$D1,$7F,$60,$F8,$7D
	.dc.b	$FA,$32,$7F,$D1,$F7,$FE,$62,$7F,$DD,$B5,$CD,$FF,$49,$FD,$83,$CD
	.dc.b	$FF,$02,$36,$33,$1F,$E0,$7F,$AF,$FA,$B1,$FE,$F6,$97,$C4,$F0,$24
	.dc.b	$7F,$A7,$A5,$EE,$7F,$83,$FE,$C3,$F9,$BC,$FF,$BF,$9B,$D4,$E9,$FC
	.dc.b	$6C,$61,$FC,$7D,$BE,$97,$97,$CB,$CC,$3F,$D5,$76,$8F,$34,$7F,$87
	.dc.b	$7D,$F7,$D5,$D2,$47,$FA,$CB,$EC,$3F,$87,$AC,$FC,$4F,$F3,$3A,$F9
	.dc.b	$8F,$FA,$E1,$AE,$F3,$FD,$1C,$62,$78,$CA,$BE,$53,$B3,$F9,$F7,$D8
	.dc.b	$2B,$FC,$64,$D8,$FF,$62,$B5,$E0,$73,$F1,$37,$EC,$48,$DB,$FE,$3B
	.dc.b	$F6,$DF,$11,$D3,$EC,$25,$74,$75,$E4,$FF,$90,$7E,$BD,$F6,$14,$95
	.dc.b	$FE,$D6,$C3,$E4,$BF,$ED,$8F,$B0,$7C,$A7,$F9,$25,$B1,$98,$C6,$5F
	.dc.b	$E4,$DF,$9D,$7D,$85,$25,$7F,$B8,$A3,$FD,$B2,$DF,$88,$E0,$47,$FD
	.dc.b	$CF,$F5,$DF,$8A,$78,$7A,$BD,$48,$FF,$BB,$3E,$C1,$F4,$9F,$E5,$D6
	.dc.b	$C6,$63,$07,$F9,$86,$9F,$25,$4C,$CE,$CF,$0B,$FE,$67,$FB,$0F,$BD
	.dc.b	$D5,$FC,$8A,$BA,$D7,$D6,$6F,$57,$E0,$AB,$73,$FD,$F5,$A9,$A6,$27
	.dc.b	$F3,$D1,$FE,$03,$9E,$5E,$3F,$84,$91,$F1,$3B,$FC,$9C,$1E,$BF,$36
	.dc.b	$4F,$1C,$BF,$9B,$E5,$F8,$48,$C1,$E3,$AE,$2E,$0F,$23,$F1,$7D,$CF
	.dc.b	$F4,$37,$D8,$7F,$A0,$09,$CD,$8E,$BC,$B2,$F9,$4E,$E7,$FD,$1D,$93
	.dc.b	$EB,$A9,$2B,$C1,$F6,$0A,$FD,$94,$3E,$C1,$7F,$D4,$0E,$51,$E7,$6B
	.dc.b	$FA,$66,$F0,$7E,$CA,$FA,$3F,$E2,$DA,$E7,$5F,$E2,$F9,$CB,$DE,$BB
	.dc.b	$61,$FD,$7A,$B7,$3E,$28,$C4,$FF,$62,$AE,$35,$A6,$C7,$2A,$F0,$35
	.dc.b	$EC,$78,$70,$F7,$D7,$16,$CE,$E3,$C2,$FB,$9D,$61,$D2,$B9,$35,$F3
	.dc.b	$16,$78,$1E,$7A,$EB,$1C,$0F,$19,$C0,$B1,$C1,$D5,$78,$1F,$F4,$40
	.dc.b	$BF,$92,$FF,$A2,$DA,$F7,$BD,$BB,$F2,$9F,$15,$8A,$FE,$A0,$D8,$CC
	.dc.b	$F6,$9C,$BF,$EE,$EF,$8E,$FF,$77,$A6,$91,$D1,$EC,$B6,$BF,$CF,$31
	.dc.b	$FB,$73,$FE,$6F,$AF,$91,$5A,$E1,$F9,$53,$93,$95,$F1,$CF,$E0,$DF
	.dc.b	$08,$C4,$FF,$AA,$E3,$15,$CA,$3C,$C7,$27,$11,$A8,$FA,$27,$7C,$3F
	.dc.b	$1F,$87,$8E,$32,$72,$8C,$DF,$82,$3F,$C6,$75,$CE,$73,$7F,$BB,$BE
	.dc.b	$47,$F5,$4A,$ED,$87,$F8,$EA,$33,$37,$0F,$EC,$31,$FE,$B8,$F8,$DB
	.dc.b	$39,$3A,$38,$BF,$F4,$C7,$D8,$7E,$F6,$7F,$85,$36,$3F,$AC,$BF,$E9
	.dc.b	$D8,$CE,$FC,$65,$3B,$BF,$0F,$99,$E4,$9E,$B5,$DE,$7F,$C1,$78,$F1
	.dc.b	$BF,$E7,$F8,$7F,$1C,$97,$D8,$3E,$73,$5D,$7C,$63,$15,$E6,$8C,$4E
	.dc.b	$A3,$94,$EB,$2F,$4F,$99,$DC,$F1,$7D,$87,$F4,$8A,$F7,$C6,$5F,$14
	.dc.b	$7F,$AE,$5C,$98,$F7,$35,$5F,$E0,$3C,$95,$B8,$EF,$5B,$3F,$9A,$7C
	.dc.b	$A7,$F7,$C6,$F5,$7C,$9F,$AF,$7D,$87,$CB,$AD,$5F,$F8,$21,$B8,$FF
	.dc.b	$85,$3E,$C1,$F1,$1B,$C7,$E9,$46,$32,$FE,$84,$DC,$3F,$C2,$39,$7F
	.dc.b	$80,$27,$37,$E1,$F2,$78,$FC,$AF,$DB,$FB,$07,$83,$8B,$FE,$99,$18
	.dc.b	$6C,$6F,$93,$83,$63,$7D,$1F,$85,$F1,$3F,$C1,$31,$83,$FA,$B3,$8D
	.dc.b	$3F,$C4,$BF,$60,$F7,$7F,$0D,$E5,$FD,$59,$CC,$63,$C9,$FC,$86,$1F
	.dc.b	$E5,$F3,$F1,$1F,$61,$FE,$46,$7F,$B1,$19,$AF,$AD,$19,$7F,$C6,$7F
	.dc.b	$2B,$F9,$B7,$39,$7E,$6E,$31,$CB,$FA,$AA,$FF,$3A,$5F,$F6,$D9,$7C
	.dc.b	$B7,$E6,$FE,$31,$F3,$7C,$63,$0F,$3E,$DD,$4D,$BF,$CF,$3F,$EC,$97
	.dc.b	$FB,$C7,$8C,$F4,$7F,$4D,$F6,$0F,$1B,$F5,$D3,$9A,$FD,$34,$E5,$E3
	.dc.b	$B7,$EB,$7E,$C3,$E5,$46,$0F,$D6,$38,$E0,$FF,$29,$FD,$77,$F9,$66
	.dc.b	$7F,$61,$F6,$1F,$1F,$8B,$FC,$C3,$39,$AF,$CF,$4E,$6B,$FD,$93,$3F
	.dc.b	$EA,$A7,$F7,$1F,$61,$F2,$1C,$FC,$4F,$EF,$E3,$16,$14,$F7,$7F,$82
	.dc.b	$7D,$83,$E5,$BF,$86,$4E,$71,$FE,$09,$39,$7F,$C3,$DF,$3F,$FE,$60
	.dc.b	$7F,$84,$65,$FC,$45,$FF,$6E,$C6,$71,$FD,$B4,$E5,$F1,$7D,$7B,$EC
	.dc.b	$3F,$8A,$4F,$FB,$56,$FF,$C5,$53,$97,$AF,$C9,$FF,$1F,$FB,$0F,$F5
	.dc.b	$3C,$7F,$8B,$DB,$87,$F0,$3F,$B0,$7D,$5E,$FE,$3A,$5F,$F1,$D9,$FD
	.dc.b	$FB,$C6,$3F,$92,$46,$2B,$FD,$0F,$7D,$38,$FF,$9D,$CE,$5F,$F4,$AC
	.dc.b	$66,$FF,$C9,$D3,$97,$FC,$A1,$F6,$1F,$C0,$6B,$FA,$1B,$63,$3C,$3F
	.dc.b	$CC,$E3,$18,$7F,$AA,$5B,$46,$9F,$46,$3F,$90,$3A,$BF,$E1,$91,$F0
	.dc.b	$79,$BD,$7E,$57,$B1,$B8,$FC,$16,$CE,$E3,$FC,$B1,$F5,$CF,$98,$FE
	.dc.b	$98,$D8,$FE,$89,$8F,$BE,$BF,$15,$39,$7F,$D5,$AD,$CB,$FB,$0F,$D6
	.dc.b	$7E,$05,$FF,$5D,$C6,$6F,$FC,$E9,$39,$7F,$29,$49,$D3,$FE,$CB,$9D
	.dc.b	$3E,$0C,$3F,$AD,$46,$2D,$9B,$E7,$BE,$D9,$F7,$3C,$0B,$97,$E0,$62
	.dc.b	$E3,$F8,$79,$B7,$F5,$BC,$EF,$A9,$F5,$FF,$55,$B8,$7C,$26,$DF,$D2
	.dc.b	$6D,$7B,$BD,$9F,$92,$7E,$33,$EC,$2D,$FC,$E8,$FE,$E5,$F5,$AC,$73
	.dc.b	$8F,$F5,$B8,$C5,$6C,$FE,$E8,$E2,$3E,$7D,$72,$AF,$C5,$7D,$87,$E2
	.dc.b	$FC,$EB,$87,$D8,$6D,$FE,$21,$B9,$7E,$51,$FD,$0F,$9D,$2B,$F4,$DE
	.dc.b	$BD,$BF,$D6,$4D,$BF,$D1,$CD,$AF,$B3,$F6,$1F,$B0,$7F,$D4,$4D,$BF
	.dc.b	$CB,$F7,$7F,$85,$E3,$17,$FE,$FE,$6D,$7E,$13,$D0,$D7,$6F,$DE,$6C
	.dc.b	$7F,$85,$79,$7F,$AA,$5F,$1A,$D4,$FE,$F7,$EC,$3F,$76,$7F,$87,$FD
	.dc.b	$37,$B5,$89,$D1,$34,$FF,$05,$4E,$A3,$18,$7E,$E3,$E9,$BF,$6C,$7E
	.dc.b	$DF,$E3,$AC,$FE,$92,$2D,$FE,$D2,$6D,$2B,$F3,$BF,$61,$FB,$EA,$34
	.dc.b	$FF,$5C,$47,$FB,$25,$C7,$B7,$FE,$45,$F8,$97,$47,$F5,$1B,$61,$D5
	.dc.b	$FE,$79,$7F,$C2,$AB,$7E,$9F,$F9,$16,$3F,$84,$56,$EB,$C3,$E2,$75
	.dc.b	$F8,$3F,$EC,$51,$D3,$3F,$3C,$F1,$23,$FE,$E0,$C7,$FC,$9F,$DD,$8A
	.dc.b	$E0,$7F,$4D,$FB,$57,$CB,$7F,$99,$E3,$0D,$51,$AB,$FF,$4E,$B6,$3F
	.dc.b	$D4,$BF,$63,$FD,$F3,$8F,$FD,$42,$D8,$CF,$D1,$BC,$FC,$FB,$5F,$7A
	.dc.b	$7F,$D5,$BF,$62,$F1,$7F,$CE,$D9,$FF,$0E,$8F,$EB,$D3,$FE,$F7,$6B
	.dc.b	$EA,$0F,$17,$53,$C9,$FE,$D5,$1C,$38,$7E,$63,$6B,$B5,$6E,$B9,$B6
	.dc.b	$75,$1C,$B8,$31,$32,$3C,$91,$B3,$0F,$96,$7C,$8D,$8D,$3E,$62,$19
	.dc.b	$39,$98,$BF,$3D,$70,$8D,$1D,$5E,$F1,$EC,$9D,$3C,$27,$95,$FC,$B9
	.dc.b	$CB,$F0,$23,$27,$B4,$F0,$9D,$4C,$9F,$3A,$38,$BF,$32,$BC,$33,$D6
	.dc.b	$38,$D7,$85,$E4,$7A,$1F,$33,$D2,$FE,$2C,$62,$35,$3E,$57,$8C,$64
	.dc.b	$91,$97,$B1,$88,$F5,$1C,$23,$47,$95,$EA,$7A,$4F,$15,$FB,$1F,$C0
	.dc.b	$98,$3D,$71,$C9,$E0,$72,$3E,$21,$EB,$BE,$57,$F8,$88,$E6,$78,$CF
	.dc.b	$2D,$F9,$DC,$BE,$AA,$F3,$39,$9F,$8B,$1E,$19,$EE,$78,$63,$2F,$92
	.dc.b	$78,$5F,$D4,$8C,$57,$42,$38,$21,$D2,$7C,$17,$C6,$31,$3C,$1F,$3B
	.dc.b	$E3,$BF,$71,$27,$83,$8B,$B1,$F1,$1B,$5F,$30,$F8,$75,$CE,$75,$1F
	.dc.b	$2C,$D1,$E2,$3C,$F5,$F0,$1C,$11,$C1,$2B,$E9,$CF,$91,$B3,$F7,$4F
	.dc.b	$95,$1F,$BF,$F0,$FF,$84,$70,$BE,$A1,$F6,$78,$6D,$FF,$13,$AE,$0F
	.dc.b	$F2,$F8,$CB,$FC,$BA,$B9,$56,$A3,$F8,$C7,$D9,$7F,$A9,$1F,$F3,$F6
	.dc.b	$9F,$CF,$E3,$F9,$34,$63,$4F,$F1,$7C,$62,$38,$FB,$FF,$C5,$E3,$94
	.dc.b	$60,$E0,$7F,$A2,$5C,$D7,$F5,$1C,$1F,$01,$FE,$31,$AD,$8A,$FF,$6D
	.dc.b	$BE,$BF,$13,$E0,$1D,$FE,$0E,$46,$48,$7B,$1F,$E5,$5F,$60,$FB,$2A
	.dc.b	$49,$EF,$F6,$5F,$D7,$73,$FE,$BD,$5B,$9F,$1B,$5F,$E7,$CD,$8F,$EC
	.dc.b	$93,$C7,$5F,$F2,$9C,$BE,$37,$C2,$4B,$E7,$EC,$D7,$9C,$5C,$3C,$33
	.dc.b	$FD,$07,$EC,$9F,$65,$EA,$F9,$39,$87,$C6,$9C,$B6,$78,$11,$F4,$EB
	.dc.b	$5F,$EA,$D3,$C1,$B3,$F1,$A7,$FA,$87,$D9,$3C,$8A,$4A,$FA,$B0,$F2
	.dc.b	$31,$1F,$D8,$63,$83,$A9,$E7,$3D,$3C,$8E,$BD,$72,$B9,$C6,$5E,$B1
	.dc.b	$97,$A4,$7F,$63,$FB,$27,$D1,$F9,$BE,$C6,$57,$E2,$27,$2D,$9E,$04
	.dc.b	$7D,$3E,$35,$7F,$BC,$4F,$06,$CF,$C2,$9F,$EE,$5D,$BF,$95,$9F,$EF
	.dc.b	$CB,$C7,$C9,$AF,$62,$4F,$E6,$A7,$4D,$9C,$BC,$BE,$CB,$A9,$1E,$1F
	.dc.b	$61,$5D,$1C,$E3,$C7,$1F,$E0,$5F,$1D,$A7,$9B,$AA,$5F,$EA,$A7,$2D
	.dc.b	$9E,$04,$7F,$5F,$F6,$2F,$B2,$C4,$7F,$86,$EB,$8B,$6B,$FC,$29,$5B
	.dc.b	$8E,$3D,$78,$EA,$91,$82,$1F,$E2,$59,$46,$A3,$13,$D3,$DE,$91,$E4
	.dc.b	$7B,$F5,$E2,$39,$45,$EC,$FF,$2A,$60,$AF,$89,$E0,$7F,$8A,$76,$FC
	.dc.b	$B6,$1F,$C3,$1A,$FB,$63,$6F,$C7,$6B,$FE,$3B,$84,$62,$FF,$CC,$5A
	.dc.b	$62,$B8,$1D,$67,$FC,$71,$F6,$2F,$B2,$A4,$9F,$E3,$BF,$64,$F8,$4D
	.dc.b	$FB,$B2,$7F,$90,$DA,$3F,$C8,$55,$CE,$7F,$92,$FD,$93,$E8,$BF,$C8
	.dc.b	$CD,$8C,$C7,$F9,$27,$EC,$BE,$B4,$7F,$95,$65,$F1,$5C,$09,$1F,$CC
	.dc.b	$25,$ED,$FF,$94,$7E,$CB,$F9,$BF,$4F,$FB,$B9,$B4,$0E,$15,$3E,$A3
	.dc.b	$97,$CC,$49,$FE,$7B,$48,$F3,$47,$F9,$7F,$EC,$1F,$65,$49,$1F,$D1
	.dc.b	$3E,$CB,$F8,$6B,$FE,$6F,$9F,$94,$FF,$48,$6B,$9E,$2F,$F3,$54,$FC
	.dc.b	$58,$FF,$3B,$7D,$92,$BF,$CD,$CD,$8F,$EA,$2D,$7F,$CF,$93,$FE,$94
	.dc.b	$73,$F5,$5C,$08,$FF,$58,$FA,$77,$C5,$74,$F9,$89,$3D,$1F,$F3,$E4
	.dc.b	$FF,$0F,$EB,$FF,$56,$7D,$93,$EC,$BF,$D0,$0D,$BF,$EC,$6D,$7F,$D1
	.dc.b	$D5,$F1,$9F,$EC,$FF,$64,$F9,$6F,$F4,$43,$63,$31,$8C,$BF,$D1,$9F
	.dc.b	$12,$FB,$2A,$4A,$FE,$E3,$0F,$91,$D9,$23,$2F,$FA,$57,$EC,$7F,$0F
	.dc.b	$1C,$3D,$5E,$A4,$7F,$BD,$FD,$93,$E9,$FF,$D3,$8D,$8C,$C6,$0E,$9F
	.dc.b	$27,$23,$24,$3F,$D5,$D6,$8F,$F5,$37,$D9,$7E,$03,$8B,$FD,$AD,$3D
	.dc.b	$6B,$77,$AE,$89,$1F,$E0,$EB,$E1,$DD,$F9,$B8,$46,$1F,$CF,$5F,$E6
	.dc.b	$DE,$15,$C9,$FF,$60,$4E,$D4,$ED,$CE,$FF,$1F,$CA,$91,$E1,$8F,$49
	.dc.b	$1C,$1F,$E2,$29,$FF,$10,$BF,$EC,$DF,$B2,$72,$7F,$B1,$5B,$1F,$E2
	.dc.b	$86,$BB,$E8,$D4,$C8,$D3,$C6,$3F,$D9,$D8,$3E,$C2,$92,$BF,$C6,$7F
	.dc.b	$64,$A6,$FF,$C0,$7E,$CB,$F9,$85,$7F,$03,$C3,$93,$FE,$3A,$6B,$FE
	.dc.b	$E5,$6C,$B7,$FA,$18,$75,$8E,$93,$8C,$3B,$DB,$0E,$EF,$F3,$FC,$7E
	.dc.b	$C4,$62,$BB,$BA,$C3,$A5,$71,$9E,$98,$76,$8F,$45,$F2,$9F,$C1,$1C
	.dc.b	$8D,$C7,$13,$F9,$05,$74,$BF,$05,$71,$9E,$67,$51,$3B,$AE,$75,$C2
	.dc.b	$75,$1D,$DE,$B8,$7C,$09,$F5,$36,$7C,$B7,$DE,$BE,$11,$DE,$FB,$3C
	.dc.b	$B0,$FC,$E3,$EA,$2C,$7C,$46,$DF,$CE,$73,$E3,$BE,$23,$8C,$7C,$08
	.dc.b	$C3,$8A,$FE,$A5,$3A,$8F,$7C,$E9,$C7,$D4,$E5,$FE,$90,$D4,$D2,$3F
	.dc.b	$3C,$FC,$7E,$BF,$87,$70,$72,$35,$8F,$E9,$35,$FA,$11,$9D,$7F,$0F
	.dc.b	$39,$8E,$27,$80,$E3,$5D,$67,$F1,$B1,$C7,$5F,$E0,$95,$C6,$73,$3C
	.dc.b	$5E,$B5,$C6,$BE,$3E,$1F,$81,$BE,$11,$8B,$EF,$87,$99,$E5,$1D,$AF
	.dc.b	$94,$7D,$8B,$E5,$18,$8F,$87,$5F,$EA,$6B,$F6,$1F,$16,$3F,$8D,$57
	.dc.b	$09,$D5,$7C,$9A,$FF,$4F,$7D,$55,$9F,$01,$CD,$C5,$FF,$7D,$F9,$5F
	.dc.b	$E0,$93,$FE,$B4,$6C,$7F,$02,$8F,$E5,$11,$FE,$03,$75,$BF,$29,$4F
	.dc.b	$0B,$F1,$79,$3E,$8B,$E3,$7E,$09,$FF,$08,$E3,$C6,$FF,$A4,$57,$FB
	.dc.b	$2A,$5F,$64,$F9,$ED,$74,$71,$8C,$57,$9E,$31,$3A,$8E,$53,$AC,$FD
	.dc.b	$47,$BF,$C8,$F0,$1E,$3F,$B2,$FE,$9D,$5F,$06,$32,$F8,$E3,$FD,$7C
	.dc.b	$E4,$C7,$B1,$AA,$EB,$92,$B7,$1E,$03,$E2,$DF,$7A,$F9,$1F,$2F,$FE
	.dc.b	$0A,$C3,$FD,$C6,$DA,$E1,$F6,$0F,$B2,$DF,$64,$BF,$F7,$4F,$D9,$3E
	.dc.b	$1F,$18,$77,$8C,$4E,$A7,$EB,$61,$FE,$09,$8C,$1F,$EF,$27,$15,$C7
	.dc.b	$E5,$BE,$E7,$D9,$3E,$37,$EF,$62,$9C,$C4,$C3,$FD,$F7,$5F,$6A,$FE
	.dc.b	$03,$C6,$33,$C7,$D3,$9F,$F1,$1E,$1F,$F8,$56,$78,$C6,$3C,$9F,$C6
	.dc.b	$7D,$97,$F9,$96,$7F,$09,$E4,$FD,$6C,$BE,$64,$E5,$FA,$31,$9E,$6F
	.dc.b	$D3,$57,$F9,$A4,$FC,$67,$D9,$7F,$94,$1F,$F2,$E4,$6B,$C1,$FC,$A4
	.dc.b	$E6,$FE,$C7,$CB,$7E,$7D,$FB,$B1,$FE,$7B,$BE,$27,$F8,$2E,$75,$7F
	.dc.b	$B1,$9D,$57,$2F,$97,$FD,$07,$C4,$BE,$6F,$F5,$18,$27,$31,$8E,$BD
	.dc.b	$D5,$CB,$F4,$8F,$FB,$4D,$F0,$E1,$FC,$BB,$CF,$FD,$67,$D9,$3C,$4F
	.dc.b	$DA,$78,$A9,$CC,$63,$93,$F6,$5F,$65,$F2,$78,$3F,$86,$5F,$A1,$FE
	.dc.b	$15,$19,$CB,$FC,$51,$F6,$1F,$E6,$B7,$F6,$DF,$65,$F3,$74,$FC,$F4
	.dc.b	$E5,$FD,$1C,$66,$FF,$8C,$4F,$FA,$BA,$BF,$7D,$F6,$5F,$21,$EB,$F0
	.dc.b	$FF,$C1,$63,$18,$7F,$A4,$69,$ED,$7F,$09,$FB,$27,$CB,$7F,$11,$9C
	.dc.b	$E3,$FC,$26,$72,$FC,$6B,$7D,$17,$F9,$C6,$3F,$87,$65,$FC,$5D,$FF
	.dc.b	$22,$46,$67,$F7,$73,$97,$FD,$45,$F6,$2F,$B2,$FE,$35,$3F,$ED,$EB
	.dc.b	$ED,$39,$7A,$FC,$8F,$F2,$5F,$B2,$FF,$59,$C7,$8D,$B8,$7F,$08,$FB
	.dc.b	$07,$D6,$6F,$E2,$E5,$FF,$23,$9F,$E0,$4F,$18,$FE,$55,$18,$AF,$E1
	.dc.b	$D7,$D3,$8B,$FA,$14,$E5,$FE,$91,$19,$BF,$D0,$CE,$5F,$D1,$FD,$97
	.dc.b	$F0,$8B,$FE,$36,$DB,$FE,$98,$FC,$C8,$CF,$99,$FC,$EE,$35,$5C,$A3
	.dc.b	$1C,$59,$7C,$FF,$05,$E8,$E9,$B5,$ED,$6E,$1F,$0D,$B8,$FA,$9F,$E9
	.dc.b	$DF,$64,$F9,$6A,$78,$76,$B8,$7D,$F9,$D3,$FD,$2F,$EB,$FE,$E6,$1F
	.dc.b	$DA,$67,$38,$FF,$38,$9C,$D7,$FA,$16,$73,$5F,$DA,$E7,$35,$FE,$E3
	.dc.b	$9C,$D7,$F0,$A9,$CD,$36,$7C,$E7,$D6,$3E,$BF,$83,$6F,$EE,$45,$C7
	.dc.b	$F0,$53,$6F,$E7,$F9,$FF,$37,$EA,$7E,$73,$70,$FE,$CD,$C5,$FC,$CD
	.dc.b	$B3,$A6,$9F,$0F,$D9,$F9,$27,$E2,$3E,$CA,$C6,$4F,$EE,$1F,$5E,$AC
	.dc.b	$E7,$FD,$72,$31,$3B,$8C,$4E,$B2,$FF,$08,$56,$A7,$F1,$3F,$65,$F8
	.dc.b	$5F,$3B,$F0,$CD,$C3,$D2,$D9,$DC,$7B,$27,$F3,$BF,$2F,$23,$F4,$5E
	.dc.b	$C5,$AF,$D4,$4D,$C7,$F3,$F3,$6B,$FD,$7D,$F6,$5F,$AC,$7F,$81,$36
	.dc.b	$FF,$3B,$DF,$FE,$19,$8C,$5F,$F8,$B5,$B5,$F7,$7E,$2B,$5D,$DF,$E1
	.dc.b	$9E,$6F,$E9,$D3,$DA,$31,$3A,$8F,$C8,$3F,$B6,$FB,$29,$1F,$E1,$DF
	.dc.b	$59,$EE,$E3,$4F,$E0,$55,$73,$FF,$14,$CD,$BF,$DA,$FD,$37,$EB,$CF
	.dc.b	$D8,$7C,$75,$8F,$C1,$C5,$C3,$F7,$33,$6F,$FC,$91,$F6,$5F,$BD,$9F
	.dc.b	$F4,$3B,$5F,$F7,$23,$50,$70,$DF,$0B,$FC,$93,$F1,$2E,$6F,$E9,$D9
	.dc.b	$7F,$2F,$BF,$E0,$77,$FE,$8D,$9C,$BF,$E2,$38,$FE,$15,$87,$93,$C4
	.dc.b	$E7,$F0,$BF,$D7,$6B,$51,$FE,$4D,$AF,$E2,$11,$E8,$8F,$F3,$13,$E8
	.dc.b	$6B,$E8,$8D,$BF,$EF,$FF,$B1,$FE,$E1,$B7,$FA,$65,$B1,$FE,$A0,$5F
	.dc.b	$3F,$FD,$1A,$31,$EC,$7D,$67,$FD,$4B,$F6,$4F,$98,$FF,$4F,$36,$3F
	.dc.b	$D4,$8D,$77,$1F,$EA,$3F,$B2,$FF,$14,$47,$F5,$37,$FA,$CB,$73,$FF
	.dc.b	$48,$EF,$F4,$21,$3C,$1F,$F2,$53,$76,$FF,$41,$3C,$A3,$35,$C9,$FB
	.dc.b	$31,$BA,$E8,$D8,$CC,$ED,$EE,$D8,$D4,$78,$BC,$3C,$4C,$8F,$54,$6C
	.dc.b	$C3,$E2,$32,$43,$04,$BF,$4C,$71,$8F,$84,$43,$DE,$E5,$E1,$7E,$AA
	.dc.b	$C5,$72,$9E,$0F,$C4,$8C,$CF,$C0,$AF,$2C,$E6,$BC,$8F,$17,$A9,$F1
	.dc.b	$0C,$CF,$17,$AB,$E1,$72,$7C,$E9,$EB,$19,$8F,$49,$A3,$E6,$9C,$8E
	.dc.b	$06,$6F,$ED,$57,$96,$38,$91,$C1,$27,$C9,$19,$7C,$11,$98,$EA,$64
	.dc.b	$F6,$1E,$E3,$EE,$1F,$54,$FC,$81,$82,$94,$CC,$FE,$4D,$B5,$F9,$D2
	.dc.b	$3C,$0E,$07,$9A,$7E,$71,$2B,$F8,$6F,$0B,$DD,$FF,$6E,$4E,$6B,$D2
	.dc.b	$D9,$D5,$7D,$C9,$CB,$D4,$E7,$5C,$9C,$46,$89,$3C,$9F,$A1,$7C,$E3
	.dc.b	$11,$F1,$CF,$85,$7E,$1B,$E9,$1C,$CC,$DF,$6A,$E3,$5E,$99,$C9,$E0
	.dc.b	$AE,$2E,$09,$87,$EC,$EB,$8D,$FA,$27,$38,$7E,$EE,$73,$1E,$1A,$FD
	.dc.b	$A9,$C6,$FA,$B8,$AF,$64,$61,$F9,$13,$E4,$AE,$6F,$63,$F3,$6E,$2B
	.dc.b	$94,$62,$B9,$3F,$20,$FF,$48,$CE,$EB,$AD,$78,$EE,$C7,$89,$FE,$19
	.dc.b	$E0,$FA,$FE,$9D,$F6,$54,$CE,$3F,$C9,$67,$F9,$41,$1F,$C6,$61,$FE
	.dc.b	$15,$7F,$1D,$F6,$5F,$EA,$96,$FC,$07,$E4,$54,$65,$FC,$DB,$5F,$CF
	.dc.b	$B9,$D5,$97,$F3,$18,$C3,$6B,$E9,$6B,$FE,$3C,$BF,$F4,$65,$7F,$B9
	.dc.b	$5F,$87,$9B,$C1,$F7,$1F,$CB,$67,$52,$C7,$BB,$FD,$BA,$32,$F9,$3E
	.dc.b	$1F,$D5,$18,$8F,$24,$7B,$7E,$C1,$F6,$5F,$CE,$5B,$3F,$E9,$F6,$BD
	.dc.b	$F4,$FE,$99,$F6,$5F,$92,$9C,$E3,$E2,$A4,$62,$BF,$A9,$36,$3F,$C0
	.dc.b	$75,$C7,$E3,$BC,$04,$8F,$07,$AF,$5F,$01,$FE,$0A,$A4,$E7,$0F,$04
	.dc.b	$2B,$FA,$97,$D9,$3E,$CB,$D9,$F2,$B3,$1F,$F5,$D4,$E5,$B3,$C0,$8F
	.dc.b	$C1,$D6,$BE,$21,$3C,$5C,$37,$1F,$1C,$FF,$63,$FB,$27,$8B,$49,$7F
	.dc.b	$EC,$B8,$78,$B8,$8E,$D1,$C1,$D4,$F3,$9E,$9E,$2F,$5E,$C9,$5C,$E3
	.dc.b	$2F,$58,$CB,$D2,$3F,$BB,$7D,$93,$EC,$BC,$DF,$5F,$2B,$FD,$BD,$39
	.dc.b	$6C,$F0,$23,$E9,$FB,$2A,$FC,$59,$E0,$D9,$F8,$53,$FE,$05,$E9,$FD
	.dc.b	$04,$FC,$B5,$EB,$E4,$D7,$B1,$27,$FD,$E1,$3A,$6C,$E5,$E5,$F6,$5D
	.dc.b	$48,$F0,$FB,$0A,$E8,$E7,$1E,$38,$FF,$0C,$79,$0D,$3C,$DD,$72,$FF
	.dc.b	$E0,$04,$E5,$B3,$C0,$8F,$F9,$7B,$93,$F4,$5F,$62,$FB,$2C,$CF,$E2
	.dc.b	$6F,$93,$6F,$FC,$51,$5B,$8E,$58,$7F,$8E,$63,$F3,$F1,$83,$FC,$D6
	.dc.b	$FE,$7D,$C0,$70,$FE,$5D,$E2,$DE,$CE,$FD,$72,$38,$CE,$B2,$F4,$FC
	.dc.b	$07,$90,$93,$AE,$CA,$F5,$8B,$3F,$02,$7E,$5F,$83,$FE,$43,$DB,$F3
	.dc.b	$B7,$E6,$6B,$F0,$E3,$6F,$CE,$6B,$FA,$F8,$FE,$44,$E4,$87,$13,$A4
	.dc.b	$7F,$94,$7E,$C5,$F6,$54,$93,$F1,$BE,$C9,$F0,$9B,$F7,$24,$FC,$AB
	.dc.b	$47,$F9,$6E,$B9,$CF,$CE,$FB,$27,$CB,$FF,$97,$DB,$19,$8F,$F3,$17
	.dc.b	$D9,$7D,$B8,$FE,$11,$2F,$87,$E0,$48,$FB,$32,$F6,$FF,$CD,$5F,$63
	.dc.b	$FD,$22,$7F,$D5,$2D,$4D,$97,$D9,$70,$F9,$99,$5F,$84,$A4,$79,$23
	.dc.b	$FC,$ED,$EB,$3E,$CA,$92,$3F,$17,$F6,$3F,$C4,$5F,$E4,$13,$F2,$1F
	.dc.b	$C7,$B5,$CF,$0F,$FA,$06,$7E,$1C,$7F,$27,$FB,$2F,$F4,$D7,$3F,$F4
	.dc.b	$56,$F6,$7F,$A5,$5B,$3F,$EC,$6A,$78,$9F,$E8,$B9,$7D,$5D,$25,$7E
	.dc.b	$83,$EA,$1F,$11,$BF,$97,$93,$FA,$4B,$47,$FA,$56,$BF,$8A,$E7,$FD
	.dc.b	$5F,$EC,$5F,$65,$FE,$99,$6E,$1F,$AC,$6B,$FD,$06,$BE,$43,$FB,$0F
	.dc.b	$B2,$7C,$A7,$FA,$85,$B1,$98,$C6,$7F,$EA,$5F,$A0,$7D,$25,$25,$7E
	.dc.b	$EA,$1F,$21,$B2,$46,$5F,$F5,$77,$D8,$7E,$1E,$38,$7A,$BD,$48,$FF
	.dc.b	$04,$FB,$27,$D3,$FF,$AD,$5B,$19,$8C,$1D,$3E,$42,$46,$48,$7F,$61
	.dc.b	$B4,$7F,$AF,$FE,$CB,$F0,$5C,$5F,$E0,$19,$EB,$5B,$BD,$74,$C8,$FE
	.dc.b	$2F,$7C,$3A,$7F,$DA,$CD,$80,$BF,$CE,$BC,$2B,$93,$FD,$CE,$78,$3B
	.dc.b	$EE,$F2,$70,$7C,$6F,$2A,$4F,$82,$32,$D2,$1E,$DF,$73,$FD,$B5,$F6
	.dc.b	$4E,$6A,$49,$FE,$5F,$1F,$94,$6C,$65,$FF,$03,$47,$17,$7D,$7F,$EE
	.dc.b	$5C,$1F,$61,$49,$3F,$CE,$BE,$CB,$F3,$FC,$3F,$0B,$EC,$97,$FE,$FB
	.dc.b	$38,$D7,$F4,$06,$BF,$E1,$56,$CF,$F8,$CE,$B9,$CA,$B9,$BF,$3A,$33
	.dc.b	$18,$8E,$F7,$F4,$9E,$17,$E1,$8C,$4E,$EB,$8C,$ED,$B5,$F6,$A3,$CB
	.dc.b	$3D,$5A,$6F,$0F,$F2,$D4,$F6,$3C,$8E,$27,$85,$F8,$2B,$C5,$5F,$15
	.dc.b	$E5,$85,$8E,$2D,$7D,$65,$8F,$0B,$A3,$C0,$30,$F4,$BC,$0C,$C7,$07
	.dc.b	$58,$79,$63,$C3,$1F,$ED,$E9,$F9,$EF,$FB,$89,$AE,$8C,$79,$2F,$8D
	.dc.b	$FF,$6F,$4F,$F4,$C6,$C6,$67,$A3,$5F,$F7,$27,$D4,$7F,$A5,$53,$48
	.dc.b	$FF,$A6,$5F,$85,$8F,$E6,$F9,$3E,$24,$FE,$BA,$73,$B7,$FA,$2B,$5F
	.dc.b	$C1,$4E,$4F,$8B,$7C,$5F,$8E,$66,$3F,$11,$19,$78,$C7,$90,$E3,$3D
	.dc.b	$73,$FC,$64,$71,$7F,$1D,$5C,$5C,$C7,$1A,$EB,$5C,$63,$8D,$FA,$30
	.dc.b	$F5,$C6,$EF,$9D,$6E,$3F,$35,$5B,$7C,$75,$FE,$28,$C3,$D8,$7F,$A6
	.dc.b	$63,$C8,$7F,$8E,$EB,$FD,$5B,$1F,$E9,$D3,$73,$E8,$FA,$9C,$BE,$03
	.dc.b	$A3,$8A,$FE,$37,$F2,$BF,$C0,$27,$FC,$6E,$D8,$FF,$46,$3F,$C8,$23
	.dc.b	$3B,$F1,$94,$EE,$FC,$7E,$5F,$8A,$7A,$D7,$79,$FF,$05,$63,$C6,$FF
	.dc.b	$A0,$E1,$FE,$49,$97,$D9,$3E,$7B,$5D,$5C,$63,$15,$E6,$8C,$4E,$A3
	.dc.b	$94,$EB,$2F,$4F,$97,$DC,$F1,$7D,$97,$F4,$6A,$F7,$C6,$5F,$14,$7F
	.dc.b	$AE,$9C,$98,$F6,$B5,$5F,$EC,$0C,$95,$B8,$EF,$5B,$3F,$CC,$3F,$2F
	.dc.b	$FE,$02,$6F,$57,$C9,$FB,$17,$D9,$7C,$BA,$D5,$FF,$AE,$DB,$8F,$FB
	.dc.b	$67,$EC,$9F,$11,$BC,$7E,$94,$63,$2F,$F3,$7B,$70,$FF,$62,$65,$FE
	.dc.b	$02,$9C,$DF,$87,$C9,$E3,$F2,$FF,$6F,$EC,$9E,$0E,$2F,$FC,$F3,$18
	.dc.b	$6C,$6F,$93,$83,$63,$7D,$1F,$85,$F1,$3F,$C1,$71,$83,$FD,$06,$E3
	.dc.b	$4F,$F6,$C7,$D9,$7F,$97,$27,$F0,$5E,$57,$D5,$9C,$FB,$7F,$89,$8C
	.dc.b	$71,$7F,$BE,$EB,$FC,$BE,$7E,$27,$EC,$BF,$C9,$0F,$FA,$40,$CD,$FD
	.dc.b	$78,$CB,$FE,$E7,$F9,$6F,$CE,$39,$CB,$F3,$91,$8B,$FD,$6D,$7F,$9D
	.dc.b	$2F,$FD,$37,$2F,$97,$FC,$E7,$C5,$3E,$73,$8C,$61,$E7,$D9,$A9,$B7
	.dc.b	$F9,$F7,$FD,$9A,$FF,$A9,$BC,$37,$A3,$FA,$7F,$B2,$78,$BF,$AF,$9C
	.dc.b	$D7,$E9,$E7,$2F,$1D,$BF,$5D,$F6,$5F,$2F,$81,$C0,$AF,$E1,$F5,$AE
	.dc.b	$2F,$DD,$7D,$83,$DD,$FD,$B7,$D9,$7C,$8C,$38,$63,$FE,$20,$AD,$63
	.dc.b	$FB,$7A,$D3,$FD,$CE,$1F,$0F,$FB,$FF,$B2,$F9,$53,$98,$C5,$F8,$3D
	.dc.b	$DF,$E0,$9F,$64,$F9,$6C,$5F,$F0,$38,$C3,$FE,$CA,$FB,$2F,$F3,$1B
	.dc.b	$FC,$1F,$0F,$E2,$0F,$F8,$0E,$33,$87,$ED,$67,$2F,$F8,$E3,$EC,$5F
	.dc.b	$65,$FC,$4E,$5F,$23,$FC,$6E,$73,$87,$58,$FE,$3D,$F6,$5F,$EA,$58
	.dc.b	$FF,$6D,$B7,$0F,$E0,$5F,$62,$FA,$CD,$FC,$6C,$BF,$E3,$93,$FB,$E7
	.dc.b	$8C,$7F,$23,$8C,$57,$D6,$9F,$F0,$E5,$F5,$E3,$FE,$7B,$5A,$7E,$9D
	.dc.b	$27,$55,$FE,$FD,$AD,$3F,$EF,$FF,$B1,$FE,$05,$7F,$D1,$9B,$19,$E0
	.dc.b	$FE,$69,$18,$C3,$F0,$ED,$A3,$4F,$A7,$1F,$C8,$5D,$5F,$F0,$D8,$F8
	.dc.b	$7C,$FE,$DF,$2B,$DC,$DC,$7E,$1B,$67,$71,$FE,$6C,$C3,$DB,$F6,$4F
	.dc.b	$95,$FE,$9C,$D9,$FE,$8D,$97,$C0,$BF,$F3,$9C,$E5,$F9,$46,$2B,$FA
	.dc.b	$87,$D7,$FE,$13,$0E,$B1,$8F,$8B,$EA,$DC,$7C,$47,$F9,$37,$0E,$95
	.dc.b	$FD,$BA,$F9,$4F,$CF,$F9,$DF,$B4,$7D,$BF,$0E,$C7,$E0,$22,$E5,$F8
	.dc.b	$69,$B7,$F4,$FC,$FF,$A3,$F5,$3F,$49,$B8,$7B,$9B,$7F,$3D,$B5,$F1
	.dc.b	$BD,$AF,$8A,$7E,$2F,$EC,$AD,$FA,$AF,$E2,$CF,$F7,$1F,$AF,$63,$A9
	.dc.b	$FE,$B5,$18,$AE,$10,$D7,$FB,$56,$3F,$E0,$EA,$D4,$FC,$E3,$58,$75
	.dc.b	$C3,$F1,$7F,$65,$F8,$CF,$3E,$E5,$F6,$1B,$7F,$89,$6E,$3F,$95,$7F
	.dc.b	$49,$E2,$CA,$FD,$4F,$B1,$6F,$F5,$F3,$6F,$F4,$B3,$6B,$FD,$09,$F6
	.dc.b	$5F,$B3,$7D,$6D,$BF,$CC,$77,$FF,$86,$63,$17,$FA,$C6,$D7,$E4,$F8
	.dc.b	$FF,$A7,$BD,$3B,$7F,$17,$F1,$9C,$5B,$1D,$9B,$8F,$F0,$4F,$B2,$FD
	.dc.b	$F1,$FE,$21,$F5,$5E,$E6,$2F,$F8,$2C,$DA,$FF,$06,$56,$A7,$27,$E8
	.dc.b	$3E,$9B,$F6,$E7,$EE,$3E,$42,$C6,$49,$5F,$B5,$8C,$BF,$5F,$EC,$BF
	.dc.b	$DF,$33,$FE,$5A,$84,$61,$FF,$2E,$B5,$FF,$31,$B5,$FF,$58,$7D,$B3
	.dc.b	$E8,$3F,$CA,$91,$FE,$E4,$8C,$57,$56,$C7,$F9,$A1,$AF,$F9,$B9,$7D
	.dc.b	$BF,$28,$46,$27,$94,$E6,$31,$3F,$E7,$2F,$B4,$FE,$E1,$AF,$AD,$B1
	.dc.b	$9F,$A5,$79,$B9,$F6,$BE,$8C,$FF,$9E,$BE,$C9,$BF,$C3,$84,$61,$F8
	.dc.b	$CD,$78,$46,$E3,$1C,$DF,$CD,$1B,$AF,$FB,$55,$B5,$FE,$AB,$F7,$7F
	.dc.b	$A3,$3F,$AE,$87,$0F,$FA,$3A,$93,$FE,$B1,$FA,$97,$87,$98,$C3,$5F
	.dc.b	$3D,$5B,$E7,$1F,$C0,$7E,$45,$7C,$A3,$0D,$7F,$75,$3C,$1B,$5D,$EB
	.dc.b	$8D,$DB,$E1,$97,$BE,$BC,$06,$66,$5F,$B6,$32,$F8,$23,$2F,$63,$46
	.dc.b	$48,$64,$93,$FE,$12,$BF,$1C,$6A,$3C,$4F,$23,$FA,$0C,$69,$B7,$E8
	.dc.b	$C7,$EA,$61,$E4,$9C,$E1,$E4,$C3,$9C,$78,$2B,$E9,$E1,$CE,$7C,$58
	.dc.b	$73,$AF,$AF,$8F,$C9,$9E,$26,$B0,$E9,$87,$82,$BC,$D3,$97,$D5,$19
	.dc.b	$7C,$07,$DE,$32,$76,$C7,$D1,$3C,$63,$D2,$74,$9E,$51,$92,$38,$24
	.dc.b	$75,$38,$9E,$03,$F1,$67,$83,$1F,$81,$1C,$23,$27,$09,$EA,$F9,$4F
	.dc.b	$A4,$7F,$BA,$B0,$FC,$63,$6F,$F5,$04,$7F,$5A,$D7,$F6,$8D,$9F,$97
	.dc.b	$5E,$1F,$B2,$FF,$7A,$C6,$2B,$C0,$7A,$63,$33,$E7,$BF,$A8,$F1,$9F
	.dc.b	$92,$E6,$FF,$55,$5C,$27,$9D,$F5,$7D,$EF,$F8,$C2,$35,$18,$AE,$4F
	.dc.b	$F6,$88,$CD,$F8,$23,$9C,$F8,$5F,$AC,$60,$82,$78,$DF,$9A,$33,$18
	.dc.b	$78,$9E,$13,$F8,$3B,$97,$ED,$57,$85,$E8,$73,$38,$12,$38,$5F,$BE
	.dc.b	$78,$46,$2F,$DB,$5F,$DF,$E3,$0F,$A2,$76,$F5,$7C,$06,$1F,$3B,$8B
	.dc.b	$F9,$F1,$B7,$C9,$3F,$30,$FF,$56,$47,$CB,$DB,$EA,$9F,$5B,$E0,$24
	.dc.b	$7F,$59,$9D,$BF,$CA,$08,$E1,$C1,$F3,$63,$F9,$F7,$DA,$BE,$EE,$92
	.dc.b	$7F,$B2,$46,$5B,$1F,$CB,$E3,$11,$FE,$21,$6E,$7F,$EA,$61,$D2,$FB
	.dc.b	$5F,$C6,$F9,$9F,$CA,$78,$E6,$1F,$87,$8C,$BE,$B7,$FA,$96,$7F,$E6
	.dc.b	$18,$C1,$FD,$DF,$EC,$DF,$6B,$FD,$05,$B1,$FE,$38,$6B,$FE,$35,$FB
	.dc.b	$5F,$F4,$3E,$7F,$E8,$89,$CB,$6B,$91,$FD,$2E,$77,$C5,$F2,$E9,$F2
	.dc.b	$3D,$C9,$87,$7F,$66,$BF,$E4,$39,$DB,$E3,$9F,$34,$FF,$57,$FB,$47
	.dc.b	$DA,$FA,$7E,$62,$5F,$F9,$42,$72,$D9,$E0,$47,$DD,$F1,$B5,$FF,$29
	.dc.b	$4F,$06,$CF,$C5,$9F,$EC,$DF,$6B,$FD,$CA,$4D,$27,$46,$B8,$A4,$FF
	.dc.b	$98,$A7,$2D,$F9,$1E,$7E,$EC,$8C,$10,$FF,$32,$4F,$FB,$6D,$E5,$1E
	.dc.b	$29,$FE,$F1,$F6,$AF,$B5,$F2,$FD,$9C,$BF,$F3,$8C,$E5,$B3,$C0,$8F
	.dc.b	$97,$EA,$AB,$FE,$73,$9E,$0D,$9F,$83,$3F,$E0,$7F,$09,$EE,$5F,$72
	.dc.b	$5F,$FA,$02,$72,$D9,$CB,$FE,$7F,$84,$F3,$9E,$9F,$6B,$2B,$93,$DF
	.dc.b	$CD,$C4,$71,$8B,$D7,$FE,$DA,$C3,$FC,$95,$3F,$E9,$0F,$25,$AF,$9F
	.dc.b	$AA,$4E,$63,$FC,$41,$3A,$8C,$10,$FF,$4A,$7D,$A3,$ED,$77,$52,$75
	.dc.b	$8F,$FB,$AA,$B5,$1F,$EA,$07,$F3,$2F,$2E,$AF,$F1,$74,$1B,$D7,$27
	.dc.b	$84,$68,$C3,$E2,$F0,$38,$91,$ED,$D9,$5D,$1E,$18,$EF,$1E,$EF,$5B
	.dc.b	$FC,$7D,$C1,$FC,$EC,$FE,$79,$AA,$35,$7F,$25,$AF,$FA,$D6,$74,$E4
	.dc.b	$87,$FA,$E6,$72,$EE,$7F,$C9,$9F,$68,$FB,$5A,$49,$FF,$62,$FD,$AB
	.dc.b	$E1,$77,$EE,$49,$FF,$64,$DA,$3F,$CA,$D5,$CE,$7F,$D9,$BF,$6A,$FA
	.dc.b	$6F,$F2,$D3,$63,$31,$FE,$5D,$FB,$5F,$AB,$1F,$EC,$F9,$7C,$67,$02
	.dc.b	$47,$FB,$62,$7F,$98,$BF,$ED,$AF,$B5,$FE,$8B,$CA,$E8,$FE,$8D,$18
	.dc.b	$6F,$A1,$DF,$EB,$3F,$CD,$92,$F9,$CA,$4C,$3B,$1F,$E7,$1F,$97,$7D
	.dc.b	$AD,$24,$7F,$BA,$BE,$D7,$F8,$84,$FC,$78,$FF,$76,$B5,$FF,$75,$37
	.dc.b	$3F,$34,$7F,$A0,$A7,$E2,$3F,$EE,$FF,$B5,$57,$FA,$05,$B1,$FE,$F8
	.dc.b	$6B,$FE,$F5,$97,$89,$BF,$AE,$93,$B7,$FD,$15,$F5,$6F,$89,$E9,$F3
	.dc.b	$32,$BA,$3A,$F2,$7F,$D2,$1F,$6A,$F9,$3A,$4A,$FF,$4C,$ED,$EA,$58
	.dc.b	$FC,$67,$3C,$9F,$DD,$FE,$D5,$F2,$DF,$E9,$B6,$CE,$8F,$F5,$1B,$85
	.dc.b	$7D,$63,$D1,$1B,$1F,$46,$FB,$5F,$F5,$33,$63,$4D,$79,$E1,$FE,$09
	.dc.b	$87,$C8,$F0,$24,$F0,$7F,$D6,$1F,$69,$F8,$E8,$E2,$7F,$AE,$27,$FC
	.dc.b	$60,$4F,$5F,$FD,$6D,$F6,$AF,$A9,$A4,$AF,$44,$65,$F5,$5B,$E4,$BB
	.dc.b	$9B,$8E,$E2,$78,$38,$9E,$0F,$C3,$FB,$5F,$C6,$67,$FE,$25,$9D,$ED
	.dc.b	$F5,$5B,$1F,$E2,$99,$DD,$7F,$B4,$9A,$9A,$62,$5D,$DF,$ED,$38,$CF
	.dc.b	$8F,$4F,$8F,$19,$E7,$F9,$3E,$64,$C3,$CD,$19,$69,$0F,$67,$B9,$FE
	.dc.b	$DF,$FB,$5F,$C5,$C7,$DF,$6B,$F1,$1B,$A7,$E4,$57,$5F,$1E,$79,$5C
	.dc.b	$5E,$91,$FE,$EB,$C5,$F6,$74,$95,$F8,$DF,$B5,$E7,$8F,$FB,$DB,$ED
	.dc.b	$57,$FE,$51,$38,$D7,$E4,$DA,$FD,$86,$CF,$E9,$6B,$9C,$AB,$9A,$11
	.dc.b	$A8,$CC,$78,$2F,$FC,$C1,$18,$BE,$EE,$AB,$84,$7F,$99,$5B,$5D,$63
	.dc.b	$67,$A6,$B9,$BB,$C3,$F5,$F3,$E5,$8C,$61,$F0,$5E,$11,$4E,$95,$C9
	.dc.b	$F5,$5F,$81,$F5,$DF,$4C,$3D,$15,$B9,$F6,$1C,$CD,$E1,$6F,$96,$1E
	.dc.b	$89,$EE,$67,$0E,$91,$E6,$8F,$F4,$09,$C8,$B3,$F4,$1F,$F4,$2B,$5F
	.dc.b	$8D,$E5,$3E,$37,$8C,$7D,$28,$C3,$C9,$AF,$26,$BE,$0F,$AB,$E2,$FD
	.dc.b	$B6,$A6,$91,$F9,$C7,$F2,$AD,$3F,$A1,$E9,$F8,$08,$E4,$68,$FE,$47
	.dc.b	$1A,$C7,$C1,$19,$CF,$F9,$8F,$27,$CB,$8E,$37,$F8,$A9,$CC,$73,$33
	.dc.b	$5F,$EA,$B8,$CC,$F8,$67,$D0,$75,$BF,$03,$FE,$AE,$C3,$EC,$57,$0B
	.dc.b	$F1,$DF,$7B,$DD,$FF,$AD,$EF,$71,$88,$F0,$9F,$07,$0C,$9F,$CF,$70
	.dc.b	$F3,$9F,$D3,$63,$67,$C2,$8D,$4F,$FA,$CD,$F5,$7D,$65,$9E,$87,$F2
	.dc.b	$AA,$46,$61,$F6,$BF,$C0,$AB,$FA,$63,$63,$F9,$8B,$FB,$58,$CE,$FC
	.dc.b	$63,$57,$FE,$C3,$F8,$CF,$3D,$7F,$B5,$67,$2F,$F6,$68,$CD,$FE,$FE
	.dc.b	$7F,$A3,$4F,$F6,$1A,$7D,$AB,$E8,$3B,$76,$7A,$E7,$31,$8A,$E1,$5B
	.dc.b	$7D,$35,$B8,$C6,$1E,$AF,$33,$D0,$73,$FB,$5F,$EA,$17,$F0,$23,$2F
	.dc.b	$8E,$3F,$D8,$6E,$4C,$7B,$9A,$AF,$EC,$79,$2B,$71,$DE,$B6,$7F,$77
	.dc.b	$F9,$8F,$F0,$83,$7A,$BE,$97,$D9,$BE,$D7,$EA,$56,$AF,$FB,$0B,$71
	.dc.b	$FF,$07,$7D,$AB,$E2,$77,$8F,$DB,$8C,$65,$FE,$05,$6E,$1F,$DA,$B2
	.dc.b	$FF,$08,$CE,$6F,$C3,$E5,$71,$F9,$8F,$C2,$7D,$AB,$C1,$C5,$FF,$83
	.dc.b	$E3,$0D,$8D,$F1,$F0,$6C,$6F,$9B,$F1,$DE,$1F,$F8,$56,$30,$7F,$86
	.dc.b	$5C,$69,$FD,$FB,$ED,$7F,$CD,$53,$F8,$3F,$2F,$EA,$CE,$7D,$DF,$C1
	.dc.b	$46,$38,$BF,$C6,$15,$FE,$6D,$3F,$15,$F6,$BF,$E5,$67,$FC,$4C,$66
	.dc.b	$FE,$BC,$65,$FF,$06,$FC,$E7,$E7,$5C,$E5,$FA,$58,$C5,$FF,$95,$85
	.dc.b	$7F,$A0,$EF,$FC,$65,$2F,$6F,$F2,$DF,$1A,$F9,$DE,$51,$87,$A7,$66
	.dc.b	$A6,$DF,$E8,$5F,$F6,$7B,$FE,$3D,$F0,$FF,$C7,$AF,$E9,$7E,$D7,$F7
	.dc.b	$2E,$7D,$2F,$D2,$46,$36,$FF,$13,$D7,$95,$FC,$AF,$DA,$FC,$C7,$3B
	.dc.b	$FE,$AE,$31,$BF,$F9,$67,$ED,$3F,$CD,$F9,$7E,$CF,$ED,$7E,$44,$E7
	.dc.b	$DF,$FD,$94,$62,$7F,$BC,$62,$BF,$F5,$B4,$7E,$EF,$ED,$7E,$4B,$A8
	.dc.b	$CC,$FE,$A2,$72,$FF,$80,$61,$EE,$FE,$FF,$ED,$5F,$2D,$FC,$5E,$73
	.dc.b	$87,$F0,$79,$CB,$FE,$63,$FA,$2F,$F3,$B3,$FC,$2F,$E3,$3F,$88,$E3
	.dc.b	$FE,$0D,$9C,$BF,$B8,$8C,$AF,$B4,$7D,$AF,$F0,$F9,$FF,$71,$5F,$F9
	.dc.b	$BE,$72,$F5,$F9,$2F,$E5,$5F,$6B,$FE,$B6,$8F,$F3,$A3,$70,$FE,$0B
	.dc.b	$F6,$AF,$AD,$DF,$C7,$4B,$FE,$41,$3F,$C0,$5E,$31,$FC,$7E,$31,$5F
	.dc.b	$E6,$2B,$E9,$C9,$F9,$39,$CB,$FE,$91,$8C,$DF,$FA,$52,$72,$FF,$A5
	.dc.b	$BE,$D7,$F8,$1D,$7E,$6D,$B1,$9E,$1F,$E5,$D1,$8C,$3F,$D2,$2D,$A3
	.dc.b	$4F,$A3,$1F,$C8,$9D,$5F,$F0,$E8,$F8,$3C,$FE,$BF,$27,$D8,$DC,$7E
	.dc.b	$0B,$67,$71,$FE,$9D,$FB,$47,$CB,$FF,$57,$6C,$7F,$47,$C7,$DF,$5F
	.dc.b	$E4,$99,$CB,$FE,$88,$6E,$5F,$B9,$FB,$0F,$C0,$BF,$EB,$78,$CD,$FF
	.dc.b	$AE,$A7,$2F,$E7,$23,$35,$FE,$B2,$9C,$BD,$F0,$FE,$A9,$18,$B6,$6F
	.dc.b	$9E,$FB,$47,$DB,$F0,$2E,$5F,$80,$8B,$8F,$E1,$A6,$DF,$D5,$F3,$7E
	.dc.b	$9F,$D9,$FD,$46,$E1,$F0,$5B,$7F,$45,$B5,$ED,$F6,$3E,$41,$F8,$BF
	.dc.b	$B5,$B7,$F3,$63,$FB,$67,$DA,$B1,$CE,$3F,$D7,$23,$15,$B3,$FB,$B3
	.dc.b	$88,$F9,$D5,$CA,$BF,$13,$F6,$BF,$8A,$F3,$2E,$1F,$5D,$B7,$F8,$76
	.dc.b	$E5,$F9,$37,$F4,$1E,$6C,$AF,$D2,$FA,$D6,$FF,$57,$36,$FF,$45,$36
	.dc.b	$BF,$89,$FD,$AF,$EB,$DF,$F7,$33,$6F,$F2,$DD,$DF,$E2,$18,$C5,$FF
	.dc.b	$BE,$5B,$5F,$83,$F8,$CD,$76,$FD,$E6,$C7,$F8,$5F,$97,$FA,$93,$E0
	.dc.b	$9D,$3F,$BC,$FB,$5F,$DD,$1F,$E1,$9F,$4D,$ED,$62,$74,$4D,$3F,$C1
	.dc.b	$73,$A8,$C6,$7F,$B7,$FA,$7F,$DA,$9F,$B6,$F8,$FB,$3F,$A3,$8B,$7F
	.dc.b	$B3,$9B,$4A,$FE,$6F,$F6,$6F,$A1,$FF,$26,$5F,$FA,$C6,$31,$3F,$E8
	.dc.b	$26,$C7,$FA,$31,$AF,$F4,$00,$FF,$4F,$70,$D7,$BF,$81,$D9,$DB,$7E
	.dc.b	$CF,$FB,$36,$9F,$E9,$06,$CF,$FA,$69,$7D,$13,$CA,$D7,$AD,$F2,$AB
	.dc.b	$C9,$E0,$7F,$A4,$32,$FD,$84,$7A,$63,$07,$FA,$53,$EC,$DC,$1F,$E7
	.dc.b	$FD,$7F,$AC,$C6,$2B,$FD,$A5,$BF,$C7,$7F,$DA,$91,$95,$19,$F8,$D7
	.dc.b	$B5,$FC,$F6,$78,$DF,$4C,$3B,$47,$26,$AC,$3A,$61,$CE,$F9,$4F,$6B
	.dc.b	$F2,$F1,$70,$B3,$4E,$7F,$61,$F4,$F0,$F4,$5F,$59,$D3,$E4,$3D,$6F
	.dc.b	$07,$27,$73,$26,$CE,$45,$9E,$D7,$E4,$8D,$5F,$91,$B5,$D5,$A4,$3B
	.dc.b	$1D,$0B,$5D,$1F,$09,$A9,$E8,$D9,$E8,$D3,$A5,$FC,$09,$DB,$F0,$AB
	.dc.b	$A4,$F7,$8D,$57,$C0,$75,$16,$FD,$8F,$53,$CA,$64,$F2,$C7,$27,$E2
	.dc.b	$19,$3B,$47,$28,$EB,$1E,$68,$C1,$10,$AF,$3D,$FA,$A7,$53,$D0,$DC
	.dc.b	$EA,$FC,$15,$B9,$C9,$88,$F2,$4E,$E3,$C4,$7A,$AB,$FC,$73,$19,$8F
	.dc.b	$81,$7E,$D7,$E0,$9C,$0F,$11,$23,$84,$6C,$93,$D2,$7C,$47,$19,$DC
	.dc.b	$7B,$2F,$D1,$3F,$04,$D9,$E2,$8E,$4F,$86,$36,$70,$9E,$71,$EB,$21
	.dc.b	$CC,$E1,$7F,$09,$EB,$7E,$67,$E3,$9A,$3C,$C6,$AF,$E8,$D7,$68,$F8
	.dc.b	$37,$E1,$74,$78,$23,$51,$E6,$BF,$CB,$DF,$7A,$E3,$18,$BF,$84,$F9
	.dc.b	$63,$C4,$7C,$13,$CA,$53,$89,$F9,$03,$73,$D6,$FC,$33,$CE,$78,$5F
	.dc.b	$8D,$E0,$7E,$C3,$DB,$FE,$12,$47,$F8,$36,$7F,$C7,$34,$C4,$6C,$FE
	.dc.b	$05,$3F,$C7,$27,$51,$FC,$1F,$EC,$9F,$67,$FC,$79,$B1,$98,$FE,$47
	.dc.b	$9B,$9D,$F6,$7B,$99,$5B,$3F,$8C,$56,$E7,$F8,$FF,$95,$FD,$46,$7F
	.dc.b	$32,$D7,$F4,$AD,$7F,$58,$DD,$BF,$C9,$71,$FC,$6E,$39,$BF,$03,$81
	.dc.b	$E0,$7F,$36,$FB,$17,$D9,$D2,$57,$F2,$5F,$B3,$FE,$5F,$9F,$F3,$0A
	.dc.b	$DC,$FF,$29,$6B,$FC,$81,$B1,$FC,$DA,$78,$F2,$7F,$AA,$A5,$F1,$9E
	.dc.b	$12,$5F,$3F,$56,$BC,$E2,$E1,$E1,$9F,$E9,$BF,$66,$FB,$3F,$57,$CA
	.dc.b	$CC,$3F,$9E,$4E,$5B,$3C,$08,$FA,$7E,$22,$BF,$CE,$27,$83,$67,$E3
	.dc.b	$4F,$F5,$CF,$B3,$FE,$B7,$A7,$FB,$11,$B5,$FC,$79,$A9,$FE,$02,$60
	.dc.b	$9D,$9C,$FB,$3A,$7C,$67,$5F,$66,$5F,$F5,$5B,$4E,$A3,$13,$A7,$FA
	.dc.b	$7C,$E9,$FE,$B9,$F6,$6F,$A4,$F4,$7D,$84,$9C,$C7,$F7,$69,$D4,$60
	.dc.b	$87,$4F,$A6,$D4,$C8,$CF,$17,$F7,$2C,$15,$FE,$00,$FB,$2F,$F0,$AD
	.dc.b	$7F,$62,$5E,$D6,$4D,$78,$72,$7F,$B9,$4E,$9B,$39,$79,$7D,$9F,$52
	.dc.b	$3C,$3C,$DA,$E8,$E7,$1E,$38,$FF,$09,$F9,$4D,$BC,$DC,$72,$FF,$C0
	.dc.b	$53,$96,$CF,$02,$3F,$E0,$7F,$B2,$7D,$9E,$23,$FB,$FD,$71,$6D,$7F
	.dc.b	$88,$AB,$71,$C7,$AF,$1D,$72,$30,$43,$FC,$13,$28,$D4,$62,$7A,$78
	.dc.b	$32,$3C,$8F,$7E,$CC,$47,$28,$BC,$7F,$E2,$8C,$15,$F1,$3D,$5F,$F1
	.dc.b	$BE,$DF,$D5,$8F,$EA,$FA,$23,$58,$FC,$66,$BF,$E1,$E9,$D3,$92,$1F
	.dc.b	$12,$3A,$9F,$E4,$4F,$B2,$7D,$9D,$24,$FF,$8C,$FE,$CD,$E9,$EF,$D9
	.dc.b	$93,$FE,$37,$B4,$7F,$93,$AB,$9C,$FF,$8E,$FE,$CD,$F4,$7F,$E5,$26
	.dc.b	$C6,$63,$FC,$AB,$F6,$7F,$52,$3F,$C6,$F2,$F8,$9E,$04,$8F,$F2,$34
	.dc.b	$FF,$3B,$7F,$C9,$1F,$66,$F4,$FF,$CB,$ED,$8C,$C7,$6F,$A9,$CF,$CC
	.dc.b	$49,$E2,$F8,$9F,$F3,$3F,$CA,$3E,$CE,$92,$3F,$CA,$DF,$65,$FC,$32
	.dc.b	$7E,$2C,$7F,$96,$5A,$FF,$94,$1B,$D1,$FE,$66,$9F,$83,$3F,$E5,$4F
	.dc.b	$B3,$57,$F9,$D9,$B1,$FE,$63,$6B,$FE,$5A,$9F,$DE,$B9,$FA,$BE,$04
	.dc.b	$7F,$CD,$5F,$53,$FD,$31,$E5,$F2,$FD,$48,$EB,$C9,$FF,$42,$FD,$9B
	.dc.b	$EC,$E9,$27,$9E,$BF,$E8,$AF,$B3,$7C,$AD,$24,$FF,$9E,$63,$C0,$FF
	.dc.b	$9C,$3E,$81,$F6,$7F,$E9,$26,$C7,$F9,$FD,$AE,$FE,$3B,$44,$74,$FF
	.dc.b	$A6,$BE,$CB,$F0,$11,$D7,$D5,$E8,$47,$FD,$15,$F6,$6F,$A6,$FF,$50
	.dc.b	$B6,$33,$18,$3A,$7C,$84,$8C,$90,$FF,$44,$DA,$3F,$D5,$7F,$67,$F7
	.dc.b	$3D,$0F,$F3,$5D,$2B,$B3,$C2,$F7,$E2,$C8,$FF,$4F,$5F,$17,$FC,$E6
	.dc.b	$D7,$97,$66,$DF,$F4,$CC,$64,$D5,$A7,$C7,$F0,$3F,$17,$CB,$97,$DA
	.dc.b	$32,$D2,$1E,$CF,$73,$FD,$8B,$F6,$6E,$AA,$49,$FF,$5A,$43,$AF,$11
	.dc.b	$FE,$AC,$8E,$0E,$A7,$9C,$FF,$B4,$30,$7D,$85,$25,$7F,$AF,$FE,$CF
	.dc.b	$A7,$83,$FE,$94,$FB,$3F,$E4,$73,$FE,$96,$BE,$2F,$FB,$15,$AF,$FA
	.dc.b	$F1,$B3,$FE,$D2,$AE,$72,$AE,$6A,$33,$18,$AE,$6D,$BF,$B6,$F0,$9D
	.dc.b	$CF,$DE,$6D,$79,$6B,$73,$F8,$23,$99,$BC,$3F,$DF,$33,$E3,$8C,$61
	.dc.b	$EF,$78,$57,$29,$E8,$64,$FF,$5B,$DF,$81,$F5,$5F,$4C,$3D,$11,$B3
	.dc.b	$FA,$5D,$73,$77,$86,$6F,$96,$1E,$69,$EE,$79,$4E,$F8,$75,$8F,$3C
	.dc.b	$7F,$98,$67,$E6,$BF,$A0,$3A,$99,$9F,$7B,$98,$FE,$83,$F1,$9F,$E6
	.dc.b	$76,$C7,$3F,$29,$39,$6C,$7F,$9A,$FE,$A3,$FC,$D8,$9A,$47,$47,$A4
	.dc.b	$63,$F9,$BE,$5F,$89,$1F,$EE,$09,$F8,$D3,$98,$E0,$64,$E4,$E7,$4F
	.dc.b	$C4,$C6,$5F,$C5,$C6,$73,$FF,$42,$57,$4D,$58,$78,$30,$F4,$1D,$23
	.dc.b	$2F,$80,$CE,$1D,$23,$2F,$82,$33,$7F,$C3,$B0,$F0,$47,$D1,$C3,$C1
	.dc.b	$87,$76,$C7,$78,$FF,$7F,$CF,$FA,$23,$0F,$0B,$E2,$9F,$E3,$35,$C2
	.dc.b	$35,$5C,$0F,$07,$D9,$F0,$7C,$57,$6F,$AC,$D5,$F6,$7F,$C0,$2B,$94
	.dc.b	$0C,$EF,$B9,$C4,$EA,$72,$D8,$E5,$1F,$72,$3D,$7F,$1D,$F6,$71,$F5
	.dc.b	$56,$9F,$6C,$66,$7F,$9A,$CF,$F4,$37,$FD,$7B,$F6,$6F,$9D,$F1,$75
	.dc.b	$7A,$F0,$4E,$6B,$C5,$39,$8C,$63,$E3,$3D,$98,$75,$71,$7C,$B0,$F7
	.dc.b	$79,$73,$1F,$F6,$6F,$D9,$FF,$4B,$7D,$D1,$9B,$E3,$3F,$EB,$49,$C9
	.dc.b	$C3,$DB,$D6,$1F,$6A,$75,$5E,$33,$E8,$DF,$AE,$33,$3F,$7F,$E6,$3F
	.dc.b	$D9,$F8,$7C,$F6,$DF,$1F,$B1,$7D,$9F,$DB,$9C,$DF,$1E,$CF,$BD,$F6
	.dc.b	$7F,$EA,$2C,$79,$46,$3E,$2E,$47,$DB,$C7,$FC,$0E,$FD,$DA,$E1,$97
	.dc.b	$7F,$2B,$9F,$D2,$7E,$27,$EC,$DE,$0E,$30,$FC,$24,$61,$B1,$BE,$5E
	.dc.b	$0D,$8D,$F3,$7E,$4F,$C5,$FF,$08,$C6,$0F,$C7,$38,$CB,$E4,$7D,$9F
	.dc.b	$F9,$6A,$7F,$E0,$0F,$95,$F5,$67,$3E,$E7,$E1,$63,$1C,$5F,$A9,$AF
	.dc.b	$F2,$E9,$F9,$E5,$F6,$7F,$E4,$98,$FC,$E1,$AB,$FC,$2C,$65,$FB,$1F
	.dc.b	$2D,$FA,$27,$39,$7E,$AE,$31,$7F,$B8,$AF,$F3,$9D,$FE,$9E,$7F,$CC
	.dc.b	$0F,$87,$E2,$9F,$3B,$C7,$DB,$FC,$E4,$62,$78,$75,$E2,$3B,$36,$3F
	.dc.b	$65,$E1,$BD,$1F,$D2,$7D,$9B,$C4,$FD,$F4,$E6,$BE,$04,$E5,$E3,$B7
	.dc.b	$EA,$FE,$CF,$E6,$70,$7E,$56,$FD,$2F,$E9,$A3,$3C,$1F,$C4,$FE,$C7
	.dc.b	$FC,$C7,$87,$CB,$FB,$3F,$91,$39,$F7,$FF,$6D,$18,$9F,$F1,$8D,$7F
	.dc.b	$A9,$5F,$A5,$F6,$7F,$21,$CF,$C4,$7E,$F6,$31,$97,$F1,$3A,$7B,$FF
	.dc.b	$C0,$7E,$CD,$F2,$FF,$C6,$E7,$38,$FD,$D9,$CB,$FC,$5F,$E9,$3F,$CD
	.dc.b	$0F,$E0,$72,$FE,$24,$FF,$91,$23,$38,$FF,$04,$9C,$BF,$C8,$FE,$C9
	.dc.b	$F6,$7F,$C4,$67,$FD,$9D,$7F,$C8,$E7,$2F,$5F,$92,$FE,$5D,$F6,$7F
	.dc.b	$EA,$58,$FE,$50,$DC,$3F,$88,$FD,$8B,$EA,$F7,$F1,$B2,$FF,$2D,$3F
	.dc.b	$BE,$78,$C7,$F2,$18,$C5,$7F,$0B,$BE,$9C,$9F,$D2,$27,$2F,$F4,$98
	.dc.b	$CD,$FF,$4E,$9C,$BF,$D4,$3E,$CF,$F7,$F5,$FD,$41,$B1,$9E,$1F,$E6
	.dc.b	$11,$8C,$3F,$BC,$B6,$8D,$3E,$8C,$7F,$2E,$75,$7F,$C3,$A3,$E0,$F4
	.dc.b	$7A,$FC,$AF,$63,$71,$F8,$2D,$9D,$C7,$F5,$1F,$B3,$7C,$BF,$F5,$B6
	.dc.b	$C7,$EF,$71,$F7,$D7,$F5,$B9,$CB,$FC,$C5,$B9,$7F,$6E,$FA,$EF,$C0
	.dc.b	$BF,$DB,$E3,$37,$FD,$D2,$72,$FF,$A2,$A3,$35,$FE,$1A,$9C,$BD,$F0
	.dc.b	$FE,$AD,$18,$B6,$6F,$9D,$FB,$47,$DB,$F0,$AE,$5F,$80,$8B,$8F,$E1
	.dc.b	$A6,$DF,$D5,$F3,$BE,$9F,$D8,$FD,$46,$E1,$F0,$5B,$7F,$45,$B5,$ED
	.dc.b	$F6,$3E,$41,$F8,$BF,$B3,$B7,$F3,$63,$FB,$77,$D9,$B1,$CE,$3F,$DA
	.dc.b	$23,$15,$B3,$F9,$63,$88,$F9,$D5,$CA,$BF,$13,$F6,$7F,$8A,$F3,$6E
	.dc.b	$1F,$5D,$B7,$F8,$76,$E5,$F9,$37,$F4,$1E,$74,$AF,$D2,$FA,$F6,$FF
	.dc.b	$57,$36,$FF,$45,$36,$BF,$D8,$3F,$67,$FA,$F7,$FC,$34,$DB,$FC,$B7
	.dc.b	$77,$F8,$96,$31,$7F,$E3,$16,$D7,$E0,$FE,$2F,$5D,$BF,$79,$B1,$FD
	.dc.b	$37,$97,$FA,$8B,$E0,$9D,$3F,$BC,$FB,$3F,$DD,$1F,$E1,$BF,$51,$ED
	.dc.b	$E2,$74,$4D,$3F,$C2,$F3,$A8,$C6,$7F,$B7,$FA,$7F,$DA,$9F,$B6,$F8
	.dc.b	$FB,$3F,$A3,$8B,$7F,$B3,$9B,$5F,$E4,$DF,$B2,$7D,$A7,$F9,$3E,$BF
	.dc.b	$DC,$F1,$8C,$3F,$D1,$51,$FE,$45,$8C,$38,$6B,$FE,$8F,$FB,$4F,$EB
	.dc.b	$B9,$F0,$84,$61,$E6,$BE,$89,$E5,$6E,$35,$3F,$E9,$9B,$72,$73,$84
	.dc.b	$61,$EC,$D7,$C9,$1B,$86,$F4,$93,$C1,$E9,$F6,$8D,$BF,$CF,$5A,$7F
	.dc.b	$98,$FE,$35,$F5,$0E,$FC,$E3,$FE,$A4,$8C,$3F,$E6,$5C,$39,$37,$0D
	.dc.b	$DF,$2B,$B7,$C2,$BF,$C8,$97,$C1,$FF,$23,$7D,$8B,$DB,$EF,$08,$F8
	.dc.b	$EF,$F9,$CD,$AF,$08,$DC,$71,$8D,$6F,$83,$33,$23,$24,$39,$1D,$5E
	.dc.b	$0E,$78,$FF,$D2,$95,$D6,$36,$7A,$2B,$6E,$27,$75,$87,$8C,$F8,$67
	.dc.b	$5A,$7F,$A4,$5F,$74,$6E,$31,$7D,$EB,$9D,$78,$E7,$35,$CE,$73,$3E
	.dc.b	$33,$47,$FA,$76,$B9,$57,$4B,$ED,$39,$BE,$53,$98,$E2,$79,$0F,$88
	.dc.b	$43,$B1,$2F,$C2,$48,$F2,$C7,$2B,$F0,$BE,$3B,$F0,$BD,$4F,$86,$64
	.dc.b	$E1,$7E,$47,$07,$7A,$EA,$F0,$3F,$0D,$3D,$2B,$F0,$B7,$E3,$9E,$2F
	.dc.b	$D2,$7C,$93,$D5,$F1,$46,$5E,$93,$E8,$73,$5E,$09,$E9,$1E,$03,$CE
	.dc.b	$7C,$57,$33,$E2,$9E,$07,$CD,$9E,$B7,$E7,$BF,$15,$6E,$7E,$A1,$CE
	.dc.b	$B8,$5F,$94,$8E,$09,$1C,$0F,$8C,$7C,$07,$2F,$A8,$D4,$79,$AB,$E4
	.dc.b	$5F,$89,$B3,$E1,$39,$1C,$2B,$D6,$E0,$F8,$47,$80,$EC,$62,$BE,$1D
	.dc.b	$7B,$E3,$A1,$AB,$F8,$0E,$EB,$CF,$18,$25,$FC,$6A,$E5,$5F,$28,$FB
	.dc.b	$67,$49,$E1,$1E,$B8,$FA,$A7,$DC,$F8,$0F,$E2,$7D,$72,$3F,$86,$47
	.dc.b	$F0,$A3,$0F,$FB,$AF,$5F,$E4,$73,$98,$FE,$25,$E3,$7F,$1C,$E7,$91
	.dc.b	$82,$1F,$D2,$B1,$FF,$1C,$BF,$D1,$5A,$E6,$7F,$9F,$43,$E3,$7F,$93
	.dc.b	$F0,$CA,$FE,$41,$1F,$C7,$8C,$3E,$5E,$0F,$E5,$5B,$FF,$16,$9F,$E5
	.dc.b	$3F,$5E,$FB,$1A,$4B,$1F,$EF,$5F,$B1,$71,$66,$3F,$DF,$15,$C5,$DE
	.dc.b	$1A,$78,$C6,$5E,$53,$E0,$F8,$C9,$1C,$23,$F8,$F4,$A7,$CB,$E9,$EA
	.dc.b	$BB,$C5,$BE,$D0,$A5,$7F,$44,$FB,$17,$D8,$FB,$B4,$97,$DC,$F6,$CE
	.dc.b	$9C,$07,$0B,$0E,$F1,$86,$D7,$CA,$E4,$F0,$32,$B9,$D6,$F3,$F6,$D7
	.dc.b	$F5,$8F,$B0,$71,$D2,$5F,$8A,$3E,$BD,$7F,$41,$77,$19,$9D,$BE,$59
	.dc.b	$D7,$0F,$5F,$8D,$ED,$EA,$CA,$E9,$19,$7B,$46,$5E,$47,$E0,$30,$4F
	.dc.b	$F6,$EF,$B1,$7D,$1F,$A7,$EB,$A5,$FC,$89,$CB,$67,$81,$1F,$67,$D8
	.dc.b	$D7,$B8,$9E,$2E,$1B,$C1,$F7,$B1,$57,$F8,$03,$9F,$FD,$11,$7F,$61
	.dc.b	$7A,$79,$35,$EA,$49,$D4,$65,$E1,$1C,$FE,$C6,$47,$37,$B7,$D7,$E2
	.dc.b	$3C,$F6,$9C,$65,$F9,$28,$57,$F8,$53,$C8,$6D,$E7,$E2,$98,$FE,$3A
	.dc.b	$72,$D9,$E0,$47,$F2,$7F,$60,$FB,$1C,$47,$C8,$AE,$2D,$AF,$F1,$1D
	.dc.b	$6E,$38,$F6,$63,$AE,$46,$08,$7D,$29,$FD,$4B,$E0,$8E,$73,$D3,$C0
	.dc.b	$93,$E4,$7B,$F6,$62,$39,$45,$DB,$F6,$B8,$7F,$17,$9F,$81,$E0,$FF
	.dc.b	$8E,$36,$FC,$EC,$7F,$0E,$6B,$FE,$61,$8F,$94,$47,$89,$FB,$68,$46
	.dc.b	$63,$E2,$36,$BF,$82,$CE,$A7,$F7,$BF,$60,$FB,$1F,$F2,$4B,$63,$31
	.dc.b	$FE,$4C,$FB,$17,$BF,$AF,$5A,$57,$27,$55,$FE,$52,$FB,$17,$D0,$D2
	.dc.b	$57,$F0,$CF,$B1,$FA,$8F,$E8,$25,$F0,$BC,$49,$1B,$F5,$BF,$CB,$BF
	.dc.b	$63,$FE,$0D,$27,$9F,$FE,$60,$C1,$F4,$FA,$F9,$59,$87,$F1,$AB,$46
	.dc.b	$E3,$FC,$CF,$F2,$0F,$AD,$A4,$8F,$E4,$3F,$61,$FC,$39,$FD,$AC,$FC
	.dc.b	$27,$F9,$2B,$5C,$F0,$FF,$9C,$A7,$E0,$C7,$F2,$AF,$B1,$57,$F9,$D9
	.dc.b	$B1,$98,$E1,$E2,$E7,$EA,$64,$F1,$7F,$CF,$BF,$4F,$FE,$99,$E9,$FE
	.dc.b	$7F,$83,$E4,$FA,$12,$FF,$9C,$D7,$F0,$8F,$13,$FC,$D1,$F6,$2F,$B1
	.dc.b	$FF,$43,$B6,$BF,$9F,$B5,$FE,$27,$5F,$15,$FE,$87,$F6,$2F,$91,$FF
	.dc.b	$47,$B6,$33,$18,$F8,$1F,$F4,$9F,$CE,$BE,$C6,$92,$BF,$A6,$C3,$E2
	.dc.b	$F6,$48,$CB,$FE,$9C,$FA,$F7,$67,$4F,$5E,$57,$27,$FD,$43,$F6,$2F
	.dc.b	$A5,$A4,$9F,$EB,$71,$97,$FA,$ED,$BE,$37,$B1,$B8,$FE,$67,$3B,$7F
	.dc.b	$9A,$CE,$DF,$EC,$BF,$63,$F8,$1E,$6F,$EE,$33,$B6,$E1,$FE,$B4,$9D
	.dc.b	$72,$7F,$A6,$5B,$7F,$E0,$C6,$A6,$DF,$E7,$23,$BF,$5F,$32,$3E,$AE
	.dc.b	$FE,$DC,$3F,$1F,$CD,$93,$E6,$8C,$B4,$87,$B3,$DD,$FF,$64,$FD,$8B
	.dc.b	$9E,$92,$7F,$C0,$70,$EC,$C4,$7F,$58,$8E,$0E,$A7,$9C,$FF,$B5,$30
	.dc.b	$7D,$7D,$25,$7F,$83,$BE,$C7,$A7,$0F,$F8,$BB,$EC,$7F,$93,$4F,$F8
	.dc.b	$F6,$B6,$FF,$66,$BE,$6F,$1C,$3F,$C2,$8A,$B8,$E1,$D2,$31,$5F,$EC
	.dc.b	$93,$8B,$83,$FC,$7A,$E2,$78,$D7,$46,$E1,$E5,$9E,$58,$7F,$B4,$E7
	.dc.b	$CF,$1C,$4F,$F6,$B6,$1E,$18,$E3,$87,$73,$D8,$73,$3C,$35,$CE,$B8
	.dc.b	$E1,$E1,$6E,$1B,$AE,$27,$38,$F0,$3F,$E5,$6C,$3C,$11,$C2,$35,$7D
	.dc.b	$5E,$E7,$39,$DB,$C6,$FE,$5B,$FE,$EF,$11,$F2,$DC,$35,$EC,$63,$CA
	.dc.b	$7C,$47,$18,$F2,$46,$1C,$4F,$F4,$69,$D4,$76,$9D,$38,$FA,$4C,$BF
	.dc.b	$E7,$26,$A6,$91,$F9,$D9,$7E,$6F,$93,$E4,$47,$13,$3A,$FF,$9D,$75
	.dc.b	$FC,$7C,$65,$FC,$8D,$71,$8F,$B0,$67,$5F,$F4,$1B,$FE,$63,$AE,$51
	.dc.b	$EF,$C3,$EB,$57,$1B,$EB,$7E,$73,$C3,$87,$91,$C4,$78,$30,$E7,$1F
	.dc.b	$46,$F9,$C6,$23,$C3,$18,$AE,$4D,$8E,$A7,$F9,$32,$FD,$47,$FA,$19
	.dc.b	$F5,$1F,$E9,$6A,$FF,$4B,$47,$FA,$28,$DD,$7B,$97,$D3,$6A,$31,$36
	.dc.b	$BF,$D3,$F3,$57,$D8,$7F,$00,$AF,$F4,$5B,$63,$FC,$BE,$FF,$AB,$23
	.dc.b	$3B,$F1,$8D,$61,$E4,$EB,$F1,$DF,$FA,$CA,$73,$7F,$EA,$E9,$CC,$F8
	.dc.b	$3A,$3C,$96,$FB,$17,$CD,$75,$9C,$C6,$30,$F2,$C6,$27,$51,$FE,$A5
	.dc.b	$AE,$51,$E5,$F2,$E4,$E7,$EC,$7F,$A5,$3E,$E8,$CE,$3F,$E8,$AB,$57
	.dc.b	$FA,$EE,$B4,$71,$F6,$B7,$3F,$4A,$75,$5F,$1A,$FD,$71,$89,$FF,$5F
	.dc.b	$FC,$AB,$AB,$E8,$37,$1F,$F5,$CF,$D8,$3E,$C7,$35,$27,$FD,$7B,$F6
	.dc.b	$3F,$E9,$C7,$FD,$7F,$7C,$73,$FF,$49,$CE,$5B,$F1,$5F,$77,$1F,$F0
	.dc.b	$15,$F2,$8C,$79,$3F,$81,$B7,$CA,$FE,$0F,$EC,$5E,$0E,$63,$0C,$C3
	.dc.b	$FD,$9F,$3F,$DE,$5B,$1C,$78,$F9,$72,$FE,$37,$C3,$FF,$06,$57,$08
	.dc.b	$C7,$97,$F9,$1F,$B1,$7B,$9F,$82,$F3,$3E,$AF,$0F,$FA,$BA,$73,$87
	.dc.b	$DD,$9C,$F2,$FF,$BA,$A3,$FC,$B2,$7E,$07,$EC,$7F,$C8,$CF,$FB,$90
	.dc.b	$CD,$FD,$D8,$CB,$FE,$BA,$FB,$1F,$CE,$39,$D3,$F4,$91,$8B,$FD,$7D
	.dc.b	$7F,$9B,$2F,$FD,$DF,$2F,$6F,$F3,$3F,$18,$FB,$1E,$31,$87,$9F,$66
	.dc.b	$A6,$DF,$E5,$DF,$F6,$0B,$FE,$FA,$F1,$1E,$97,$E9,$3E,$C7,$F7,$73
	.dc.b	$98,$C7,$8B,$FB,$9A,$F2,$6F,$F9,$9F,$B1,$F8,$EF,$FB,$5A,$33,$BF
	.dc.b	$F0,$89,$CA,$C3,$F9,$2F,$D7,$FF,$98,$1F,$D5,$7D,$8F,$C8,$9D,$60
	.dc.b	$15,$B7,$F5,$13,$BA,$E7,$F1,$1F,$BC,$FB,$1F,$95,$18,$3F,$6C,$E2
	.dc.b	$BA,$5F,$F9,$96,$7F,$75,$F6,$2F,$97,$FE,$2D,$39,$AF,$DB,$4E,$6B
	.dc.b	$C9,$97,$9A,$5F,$3F,$FE,$69,$9F,$E0,$99,$7F,$12,$9F,$54,$E5,$FE
	.dc.b	$01,$19,$5F,$60,$FB,$1F,$E2,$B2,$F9,$1F,$E5,$13,$9C,$3A,$C7,$F2
	.dc.b	$BF,$B1,$FF,$52,$C7,$90,$FE,$A2,$E3,$0F,$E1,$DF,$5E,$FA,$9E,$1F
	.dc.b	$1B,$2F,$F8,$D4,$FF,$03,$79,$57,$D1,$BE,$5A,$FF,$42,$AD,$4A,$3E
	.dc.b	$14,$EA,$BE,$2D,$6A,$73,$7F,$CC,$23,$13,$F2,$BE,$C7,$F8,$55,$7F
	.dc.b	$4E,$6C,$67,$87,$F9,$BC,$63,$1F,$03,$68,$D3,$EB,$47,$F2,$D7,$57
	.dc.b	$FC,$52,$3E,$1F,$7F,$8F,$CA,$F2,$37,$1F,$86,$D9,$DC,$7C,$45,$F6
	.dc.b	$2F,$96,$FE,$B8,$D9,$FE,$89,$8F,$C3,$AF,$CD,$4E,$5F,$91,$3F,$E0
	.dc.b	$68,$C6,$5F,$DB,$BE,$B3,$F0,$CF,$E3,$63,$33,$F9,$09,$CB,$C7,$0F
	.dc.b	$EB,$71,$8B,$EF,$18,$F8,$9F,$A1,$F3,$B2,$FE,$FF,$81,$6F,$F0,$B3
	.dc.b	$72,$FB,$D3,$6B,$EB,$79,$DF,$4B,$EB,$FE,$9B,$63,$DE,$DC,$3E,$83
	.dc.b	$6F,$D5,$EC,$7C,$63,$F0,$5F,$63,$6F,$E6,$6D,$FD,$BF,$0F,$EC,$9F
	.dc.b	$62,$CB,$51,$86,$E3,$FE,$14,$9D,$5F,$2C,$7F,$BE,$C6,$2F,$F0,$5F
	.dc.b	$63,$F8,$AF,$33,$F1,$6D,$BF,$AE,$D9,$DC,$7D,$A9,$FC,$C7,$9D,$23
	.dc.b	$F4,$FE,$C5,$AF,$D6,$CD,$CB,$F4,$D3,$6B,$FC,$87,$F6,$3F,$B1,$7F
	.dc.b	$14,$DB,$FC,$B7,$7F,$F8,$96,$31,$7F,$C3,$5B,$5F,$81,$F8,$CD,$77
	.dc.b	$7E,$01,$B1,$FE,$13,$E7,$FE,$AE,$F8,$27,$4F,$EF,$BE,$BB,$F6,$47
	.dc.b	$F8,$8B,$D4,$7B,$78,$9D,$13,$4F,$F0,$B4,$EA,$31,$9F,$ED,$FE,$9F
	.dc.b	$F6,$A7,$ED,$BE,$3E,$CF,$E8,$A2,$DF,$EC,$E6,$D7,$F2,$9F,$B0,$FE
	.dc.b	$B5,$8F,$F9,$F9,$B5,$FE,$85,$5F,$45,$FC,$BA,$31,$EA,$7A,$1F,$F4
	.dc.b	$57,$D8,$3B,$BF,$D0,$6D,$8F,$F4,$4B,$5D,$C6,$B6,$91,$A6,$BF,$E9
	.dc.b	$5F,$B0,$FF,$0B,$BF,$E9,$26,$B9,$EE,$FF,$48,$C6,$23,$FD,$48,$B5
	.dc.b	$C9,$23,$93,$5F,$E0,$67,$F5,$F8,$E0,$FF,$0F,$6B,$A1,$F6,$1F,$EB
	.dc.b	$1D,$BF,$A9,$3F,$E0,$58,$FE,$2C,$FC,$77,$83,$63,$7E,$C7,$F9,$E7
	.dc.b	$87,$FD,$2F,$3C,$6B,$7E,$76,$5F,$EE,$51,$C6,$77,$59,$9F,$1B,$67
	.dc.b	$51,$C2,$B1,$32,$3C,$B1,$B3,$0F,$9A,$30,$E8,$F5,$BA,$BF,$7C,$EA
	.dc.b	$78,$1A,$38,$1E,$A3,$06,$8C,$12,$3E,$1B,$A7,$AB,$C9,$CB,$C5,$CB
	.dc.b	$C6,$FE,$1B,$C5,$F9,$73,$E5,$9E,$24,$8C,$CF,$57,$8B,$D5,$CB,$D4
	.dc.b	$FA,$35,$F0,$63,$33,$E2,$3A,$4E,$8E,$87,$D1,$3C,$86,$4E,$87,$E2
	.dc.b	$AB,$A3,$98,$E9,$3D,$8F,$2B,$DA,$38,$1D,$8F,$C0,$9F,$38,$FF,$70
	.dc.b	$98,$3F,$DF,$D7,$F8,$D7,$2D,$7E,$09,$8B,$FF,$7D,$C6,$63,$D0,$64
	.dc.b	$F5,$9D,$67,$FC,$8F,$F1,$AF,$6B,$F7,$D1,$E7,$AF,$CE,$3F,$1A,$75
	.dc.b	$B7,$C0,$BE,$58,$FB,$8F,$F2,$84,$EA,$3B,$DF,$29,$D4,$27,$4F,$D9
	.dc.b	$3E,$54,$6A,$33,$7E,$3C,$7E,$7C,$62,$79,$D7,$88,$E2,$75,$C7,$D7
	.dc.b	$1A,$C3,$C5,$3A,$AE,$4F,$03,$91,$CD,$CB,$88,$E2,$E6,$BA,$BC,$63
	.dc.b	$27,$69,$E2,$E2,$FD,$35,$C2,$BA,$46,$31,$F9,$95,$B9,$FF,$25,$DF
	.dc.b	$2A,$C9,$FC,$35,$C4,$7C,$E8,$C5,$F5,$AE,$53,$F3,$6B,$6F,$CF,$3D
	.dc.b	$57,$FC,$47,$DF,$E9,$F4,$EE,$CA,$66,$BF,$92,$CF,$E3,$88,$FE,$60
	.dc.b	$FE,$3E,$FE,$4B,$C8,$91,$93,$F9,$8B,$87,$F3,$AF,$F2,$12,$3F,$03
	.dc.b	$2F,$F7,$93,$FC,$D3,$E3,$7C,$BC,$32,$30,$43,$F4,$F1,$FA,$22,$3E
	.dc.b	$1F,$AE,$7D,$87,$F3,$16,$C7,$F9,$55,$AF,$F4,$7F,$B0,$FF,$4B,$71
	.dc.b	$7F,$A5,$E7,$2D,$AE,$47,$F3,$B9,$DF,$17,$FA,$C6,$D6,$F8,$BF,$01
	.dc.b	$23,$C1,$EA,$D7,$FA,$B4,$ED,$F2,$4F,$9E,$7F,$A4,$7D,$7B,$EC,$3D
	.dc.b	$5A,$4B,$FF,$37,$4E,$5B,$3C,$08,$FB,$F8,$2B,$FD,$96,$78,$36,$7E
	.dc.b	$34,$FF,$57,$FA,$E7,$83,$49,$5F,$E7,$D8,$FE,$DF,$BF,$FA,$F5,$DC
	.dc.b	$66,$76,$FF,$73,$8D,$EB,$F1,$D7,$C5,$F7,$F5,$E5,$75,$8F,$11,$87
	.dc.b	$94,$61,$FF,$47,$7D,$7B,$EC,$3D,$1F,$5D,$27,$31,$FD,$C6,$75,$18
	.dc.b	$21,$E5,$FA,$3D,$4C,$8C,$E9,$ED,$9F,$EF,$5D,$5F,$BB,$1E,$B6,$0C
	.dc.b	$F6,$4B,$FF,$51,$4E,$9B,$39,$7F,$C2,$D0,$FB,$0E,$A4,$8E,$3F,$5D
	.dc.b	$5D,$C5,$C7,$FD,$D9,$4A,$FF,$07,$79,$0D,$7C,$FC,$33,$1F,$F5,$A4
	.dc.b	$E5,$B3,$C0,$8F,$FA,$E7,$EB,$DF,$61,$88,$FF,$16,$57,$16,$D7,$F8
	.dc.b	$76,$B7,$1C,$7A,$B1,$D5,$23,$04,$3F,$C6,$D3,$FE,$F9,$7C,$11,$CE
	.dc.b	$7A,$78,$12,$7C,$8F,$7E,$BC,$47,$28,$BB,$34,$FD,$A6,$2B,$F8,$BE
	.dc.b	$0F,$F8,$CF,$6F,$C9,$E2,$3E,$57,$81,$A7,$E3,$CD,$7E,$0B,$C4,$C6
	.dc.b	$3D,$E3,$96,$DF,$CA,$DE,$27,$FB,$8E,$F0,$FF,$95,$7E,$C1,$F5,$FF
	.dc.b	$E4,$56,$C7,$FB,$A5,$AF,$F9,$6F,$EC,$1F,$01,$BF,$5E,$4E,$DF,$F2
	.dc.b	$74,$F0,$CB,$C9,$2F,$B0,$7D,$07,$F9,$4D,$B1,$FE,$F7,$6B,$FE,$69
	.dc.b	$FB,$0F,$B6,$FF,$9A,$E1,$F0,$DC,$49,$1F,$EF,$F9,$7A,$DF,$E5,$FF
	.dc.b	$B0,$FE,$D2,$4F,$3F,$FC,$C5,$2F,$A6,$D7,$CA,$CC,$3E,$96,$9E,$8C
	.dc.b	$13,$C2,$7F,$CD,$5F,$5C,$FB,$0A,$48,$E3,$D5,$FC,$CF,$EC,$3F,$89
	.dc.b	$C7,$FA,$0E,$BE,$4B,$A3,$F0,$0E,$1D,$71,$FF,$9D,$E7,$E1,$E5,$F2
	.dc.b	$BE,$C1,$7F,$E7,$B6,$DF,$30,$E1,$FF,$49,$53,$C6,$E1,$F5,$32,$78
	.dc.b	$52,$3F,$D0,$DF,$4F,$FD,$8A,$7F,$D3,$52,$FB,$0F,$01,$23,$C1,$3F
	.dc.b	$C7,$1F,$F5,$07,$D8,$3E,$3B,$FD,$20,$D8,$F3,$35,$FF,$53,$4F,$CB
	.dc.b	$7D,$5F,$60,$F9,$1F,$F4,$C3,$63,$31,$8F,$89,$FF,$4E,$7C,$EB,$E8
	.dc.b	$69,$2B,$E2,$43,$E3,$36,$48,$CB,$FE,$A5,$FA,$EF,$C4,$C7,$0F,$1F
	.dc.b	$A9,$1F,$9D,$F6,$0F,$A4,$FF,$56,$36,$33,$18,$3A,$7C,$74,$8C,$90
	.dc.b	$FF,$63,$5A,$3F,$D6,$DF,$61,$F8,$3E,$8F,$D4,$CF,$5A,$DD,$EB,$A6
	.dc.b	$47,$E0,$6F,$87,$57,$F8,$0D,$B1,$FC,$36,$FF,$40,$EE,$B8,$BF,$ED
	.dc.b	$79,$DA,$9D,$B9,$DF,$E3,$79,$92,$3C,$11,$E8,$23,$83,$F2,$92,$F7
	.dc.b	$7F,$DA,$5F,$60,$E5,$A4,$AF,$CD,$47,$E4,$DB,$19,$7F,$DD,$91,$C5
	.dc.b	$DF,$57,$FB,$77,$17,$D7,$52,$4F,$E9,$7E,$C3,$F3,$DC,$1F,$CB,$FE
	.dc.b	$C3,$F9,$A4,$FF,$31,$BE,$2F,$EA,$DA,$FF,$BE,$DB,$3F,$C7,$2B,$9C
	.dc.b	$AB,$9A,$8C,$C6,$2B,$9B,$6F,$FC,$70,$F0,$9D,$CF,$F8,$ED,$B5,$E5
	.dc.b	$AD,$CF,$F8,$FC,$E6,$6F,$06,$1E,$D8,$C3,$6B,$91,$4F,$EA,$95,$D1
	.dc.b	$FE,$A5,$7E,$07,$D5,$7D,$1F,$75,$F8,$6B,$84,$72,$3A,$57,$13,$94
	.dc.b	$F9,$A3,$C2,$72,$3B,$D7,$5C,$3E,$13,$DE,$3F,$DA,$B5,$F3,$9B,$19
	.dc.b	$7D,$A7,$FA,$0E,$3F,$A6,$FC,$47,$FA,$05,$B1,$9F,$2F,$FD,$84,$D9
	.dc.b	$FF,$43,$3F,$DC,$7E,$96,$A6,$91,$DB,$F2,$73,$FC,$D7,$1F,$CB,$9E
	.dc.b	$47,$B4,$FA,$7B,$7F,$75,$CF,$F0,$B1,$A7,$F0,$F5,$CA,$38,$19,$D3
	.dc.b	$FC,$F9,$19,$9E,$33,$E5,$3B,$46,$67,$FD,$49,$87,$96,$FC,$11,$8A
	.dc.b	$DC,$61,$FF,$80,$33,$BC,$3D,$51,$D7,$0F,$2D,$6E,$3F,$D1,$D5,$B7
	.dc.b	$9D,$7F,$51,$C3,$D4,$7F,$80,$23,$D4,$7F,$6C,$AF,$F0,$B4,$7F,$A7
	.dc.b	$9B,$3E,$EF,$A6,$B5,$B8,$E9,$18,$87,$D8,$7E,$FA,$7F,$B0,$36,$3F
	.dc.b	$43,$1F,$EA,$38,$FF,$16,$3A,$DF,$94,$6F,$0F,$1F,$99,$E8,$AE,$15
	.dc.b	$E2,$9F,$F0,$5E,$3C,$39,$F1,$F6,$0F,$9B,$C6,$1F,$DE,$ED,$39,$8F
	.dc.b	$2D,$70,$AD,$E7,$E6,$3F,$6F,$5B,$70,$7B,$BC,$AF,$39,$E3,$FB,$0F
	.dc.b	$E8,$D8,$7C,$08,$CB,$E1,$8F,$F6,$03,$93,$1E,$DE,$AB,$C1,$92,$B7
	.dc.b	$1E,$0D,$BF,$C3,$F7,$C0,$FF,$06,$7C,$BB,$AF,$E9,$36,$3F,$C2,$1F
	.dc.b	$5E,$FB,$0C,$D4,$9F,$F0,$A7,$D8,$7F,$AB,$9E,$71,$F6,$E3,$18,$FC
	.dc.b	$06,$E7,$FE,$48,$C3,$FC,$05,$7C,$BE,$2B,$F0,$51,$8B,$E1,$F2,$DF
	.dc.b	$85,$FB,$0F,$E9,$D7,$E0,$8C,$78,$12,$3F,$C4,$D3,$FD,$E5,$B1,$C7
	.dc.b	$93,$97,$3F,$E3,$FC,$3F,$F0,$65,$70,$8C,$79,$3F,$85,$F2,$3E,$4E
	.dc.b	$DF,$9C,$6D,$7F,$92,$97,$C1,$EF,$9B,$F1,$9F,$5E,$F8,$1B,$F3,$78
	.dc.b	$6E,$1F,$8B,$97,$D2,$70,$F8,$0E,$2D,$9F,$C8,$7C,$5B,$D0,$FC,$C3
	.dc.b	$FE,$7C,$8C,$C6,$3E,$0F,$35,$27,$8D,$DE,$6F,$C2,$FD,$7F,$D2,$9F
	.dc.b	$F3,$1B,$63,$FA,$6A,$F9,$0F,$D3,$39,$E1,$7D,$7F,$E1,$6B,$F1,$ED
	.dc.b	$8F,$E0,$2B,$E7,$7F,$5C,$E7,$0F,$F3,$8F,$D7,$F3,$3B,$3F,$AD,$71
	.dc.b	$8F,$ED,$7E,$BD,$F0,$B7,$AF,$0D,$BF,$DA,$7D,$77,$FA,$21,$BD,$DF
	.dc.b	$BD,$8C,$6B,$FA,$45,$1F,$C0,$7E,$33,$AD,$7E,$F5,$B5,$FA,$A7,$F1
	.dc.b	$ED,$BE,$19,$CA,$E1,$F3,$16,$BF,$63,$F5,$5F,$1C,$9A,$FF,$0F,$71
	.dc.b	$E1,$D3,$11,$6F,$F8,$74,$E9,$B1,$9A,$FD,$EF,$D5,$FE,$97,$8B,$F9
	.dc.b	$7B,$FE,$B8,$AF,$E2,$CE,$5F,$F4,$FB,$99,$E3,$EF,$7F,$21,$8C,$57
	.dc.b	$28,$C7,$85,$FC,$CF,$EA,$DF,$25,$4F,$57,$C6,$DC,$BF,$D2,$93,$A7
	.dc.b	$F9,$37,$D3,$FF,$A1,$5C,$D7,$F2,$F8,$C6,$7E,$AF,$98,$F1,$1E,$3E
	.dc.b	$FB,$7E,$A8,$B3,$F0,$26,$E9,$F0,$26,$D7,$BA,$6D,$7F,$89,$B7,$FB
	.dc.b	$F3,$F3,$3E,$AF,$E9,$73,$7D,$36,$C6,$4F,$A3,$19,$7D,$BE,$A7,$C3
	.dc.b	$3E,$47,$7D,$8C,$9F,$D1,$7E,$AD,$79,$C7,$F9,$9C,$62,$7E,$EF,$D5
	.dc.b	$FE,$0F,$C9,$FC,$23,$63,$27,$E0,$A3,$2D,$33,$5F,$8E,$8C,$B6,$3F
	.dc.b	$03,$E1,$7D,$D3,$EF,$7A,$16,$32,$49,$FC,$5C,$65,$FE,$AB,$F5,$6F
	.dc.b	$AB,$FE,$E8,$DA,$FC,$AF,$7E,$26,$47,$E0,$3E,$1A,$DF,$F5,$2E,$5F
	.dc.b	$EB,$6F,$E6,$3C,$5F,$C8,$1F,$DC,$7C,$F7,$AB,$8C,$3F,$33,$36,$BF
	.dc.b	$AD,$D6,$A7,$24,$8E,$8F,$F7,$58,$D4,$60,$FD,$9F,$D0,$FE,$C4,$FD
	.dc.b	$97,$C4,$DA,$FD,$7C,$59,$FD,$CC,$DE,$A5,$26,$78,$6F,$A3,$FF,$11
	.dc.b	$46,$1A,$A3,$53,$C6,$21,$99,$DF,$D5,$BD,$1F,$87,$3A,$11,$9D,$71
	.dc.b	$18,$F4,$5F,$53,$8A,$D4,$F1,$C1,$F4,$FF,$E4,$EC,$BC,$51,$89,$F6
	.dc.b	$69,$EE,$7D,$D1,$97,$FC,$2F,$F4,$F9,$7F,$C3,$31,$99,$C3,$6B,$AD
	.dc.b	$F7,$AF,$91,$7C,$6F,$73,$99,$F9,$4D,$BE,$51,$C2,$F1,$32,$3C,$F1
	.dc.b	$D9,$A6,$CA,$74,$2C,$F4,$70,$7B,$23,$A3,$EC,$9B,$5E,$FB,$F4,$57
	.dc.b	$7A,$D1,$D2,$BD,$53,$A8,$F9,$31,$CA,$78,$57,$37,$13,$A3,$91,$E8
	.dc.b	$AE,$4F,$9A,$7E,$3B,$A3,$D0,$57,$2F,$90,$C9,$5C,$92,$7F,$25,$3D
	.dc.b	$AF,$E2,$12,$FD,$73,$E6,$BF,$13,$EA,$38,$99,$9E,$11,$E9,$8C,$4F
	.dc.b	$1A,$EC,$E2,$74,$7A,$67,$A3,$F0,$CF,$4D,$F9,$DC,$35,$57,$47,$91
	.dc.b	$A9,$F2,$B8,$3D,$26,$4F,$C3,$1F,$15,$F0,$BA,$3A,$BE,$77,$47,$09
	.dc.b	$F1,$5F,$C0,$8F,$3C,$F2,$8C,$4F,$DF,$9F,$14,$62,$7C,$B1,$D0,$F0
	.dc.b	$47,$E5,$6B,$D9,$19,$8C,$4F,$8E,$BE,$15,$69,$F1,$DF,$D8,$BE,$EF
	.dc.b	$D1,$AE,$07,$BA,$B8,$3E,$78,$E1,$3D,$DE,$0F,$23,$E5,$4F,$07,$E5
	.dc.b	$1F,$1D,$C3,$EA,$8E,$75,$E4,$8E,$35,$E2,$78,$D7,$0B,$F8,$91,$F3
	.dc.b	$67,$C8,$F0,$3C,$47,$C7,$9E,$17,$E6,$8F,$41,$B2,$9B,$9F,$03,$E7
	.dc.b	$6D,$7C,$C8,$EB,$7F,$32,$73,$5E,$A7,$C7,$3C,$CF,$E1,$7E,$AF,$F1
	.dc.b	$72,$3F,$C3,$BD,$6A,$FF,$26,$8B,$7F,$C9,$A7,$F9,$07,$D3,$FF,$45
	.dc.b	$AF,$E2,$BF,$0F,$FC,$84,$CF,$0C,$8F,$E7,$F3,$B7,$FA,$23,$5F,$E9
	.dc.b	$6D,$7F,$A7,$7D,$3B,$93,$F9,$73,$63,$F9,$2B,$5F,$F6,$9A,$E4,$DC
	.dc.b	$C9,$DB,$CA,$31,$1C,$22,$E5,$FD,$16,$B9,$4F,$6F,$3F,$C3,$E7,$C8
	.dc.b	$C1,$0F,$0B,$97,$C7,$1F,$D1,$7E,$9D,$F4,$FE,$AF,$A4,$95,$FC,$FA
	.dc.b	$72,$D9,$E0,$47,$DF,$F3,$3F,$D3,$FE,$99,$E7,$5F,$3A,$57,$F4,$89
	.dc.b	$CB,$67,$2F,$FB,$E2,$13,$CE,$64,$F1,$8D,$EB,$FD,$9E,$B9,$CF,$1F
	.dc.b	$A7,$F1,$11,$F0,$7D,$2D,$74,$78,$E3,$D1,$1F,$D9,$FB,$5A,$FA,$FA
	.dc.b	$25,$FF,$60,$9C,$B6,$78,$11,$FE,$CB,$F4,$CF,$A7,$FE,$E4,$BA,$AF
	.dc.b	$54,$AD,$C6,$8C,$2D,$7F,$C0,$95,$FE,$E2,$8D,$90,$D6,$5F,$E2,$18
	.dc.b	$D4,$60,$FF,$16,$38,$9E,$75,$E4,$F4,$E5,$7A,$1F,$3F,$5E,$23,$90
	.dc.b	$BF,$F4,$BB,$B3,$DF,$1F,$27,$C8,$FF,$08,$6D,$F9,$69,$E2,$DC,$7F
	.dc.b	$C3,$DB,$2B,$85,$7C,$D6,$BC,$1E,$7C,$D2,$3F,$C2,$36,$8E,$FC,$F9
	.dc.b	$9E,$F1,$70,$FF,$16,$57,$FB,$BE,$78,$CE,$75,$FF,$24,$4F,$39,$EB
	.dc.b	$7F,$E2,$DB,$7A,$0C,$B0,$FF,$92,$EB,$FC,$4B,$3A,$F1,$33,$5C,$7E
	.dc.b	$AF,$E3,$D7,$D3,$A4,$61,$FF,$22,$18,$DF,$86,$DE,$1F,$B6,$FC,$EC
	.dc.b	$E9,$BA,$FF,$93,$1B,$87,$E1,$DF,$D4,$BE,$4C,$DF,$65,$E8,$24,$67
	.dc.b	$AF,$CD,$F6,$9E,$7F,$A7,$93,$EC,$D5,$3E,$3C,$3C,$9E,$8C,$AF,$83
	.dc.b	$97,$C6,$7C,$1D,$BF,$1B,$B6,$46,$0E,$9E,$37,$C5,$23,$DF,$E2,$7E
	.dc.b	$07,$A5,$23,$AB,$F0,$FC,$EF,$A9,$F8,$0F,$F9,$EA,$7E,$63,$78,$FE
	.dc.b	$5B,$78,$FF,$CE,$2B,$C3,$FA,$9F,$45,$F5,$7D,$29,$87,$BE,$3F,$01
	.dc.b	$FB,$3F,$18,$FD,$93,$C5,$FB,$ED,$7E,$E3,$B5,$F4,$4F,$81,$FF,$39
	.dc.b	$36,$3F,$CD,$2D,$7F,$5A,$B5,$FF,$3C,$7C,$5B,$E3,$69,$2B,$FC,$E3
	.dc.b	$0F,$6B,$64,$8C,$BF,$E8,$1F,$A0,$7B,$1D,$3D,$19,$5C,$9F,$F4,$37
	.dc.b	$D1,$39,$69,$27,$FC,$FF,$19,$7F,$D0,$36,$E0,$EC,$6E,$3F,$83,$CE
	.dc.b	$DF,$E1,$33,$B7,$FD,$13,$F4,$4F,$86,$F3,$61,$27,$31,$FE,$99,$9D
	.dc.b	$46,$08,$74,$F0,$35,$32,$33,$CF,$FE,$AB,$C1,$5C,$F8,$3E,$27,$87
	.dc.b	$2F,$C7,$19,$69,$0F,$4F,$C5,$FF,$AA,$FE,$89,$F3,$54,$93,$FE,$A4
	.dc.b	$87,$D1,$62,$3F,$94,$C7,$07,$53,$CE,$7F,$D6,$D8,$3E,$72,$92,$BF
	.dc.b	$D5,$FF,$45,$D3,$5F,$F7,$37,$D0,$2F,$FD,$C0,$71,$AF,$F5,$AB,$5F
	.dc.b	$E7,$2D,$9F,$F6,$75,$73,$95,$73,$7F,$DE,$51,$98,$C4,$77,$BF,$F7
	.dc.b	$74,$62,$FB,$3A,$AE,$11,$FE,$F3,$6D,$7F,$BF,$63,$67,$A2,$B9,$BB
	.dc.b	$C3,$FB,$04,$F9,$23,$18,$7C,$07,$84,$53,$A5,$72,$7D,$35,$CB,$0F
	.dc.b	$23,$6B,$8C,$F5,$38,$9C,$C6,$1E,$86,$CF,$17,$67,$13,$E2,$1C,$63
	.dc.b	$81,$F3,$51,$EA,$29,$4A,$76,$C2,$BF,$4B,$E8,$9F,$45,$88,$F2,$C6
	.dc.b	$1D,$D2,$31,$1D,$23,$13,$C6,$3C,$7F,$45,$53,$48,$ED,$F8,$97,$FA
	.dc.b	$0F,$0B,$E3,$4F,$23,$E1,$46,$5F,$88,$67,$3F,$C4,$C6,$5F,$C5,$C6
	.dc.b	$73,$F8,$73,$97,$E0,$DF,$83,$56,$1E,$0B,$F3,$9E,$08,$CB,$E0,$33
	.dc.b	$87,$82,$32,$F8,$23,$37,$E0,$9C,$BE,$08,$E9,$5D,$EF,$C4,$C8,$FF
	.dc.b	$11,$E1,$EF,$3C,$2F,$DC,$9D,$46,$0D,$4F,$BB,$E8,$BF,$BF,$BE,$13
	.dc.b	$F0,$B1,$B8,$CD,$7E,$23,$E7,$FF,$82,$56,$8F,$C4,$B8,$8F,$F1,$AB
	.dc.b	$F8,$E8,$D4,$71,$DF,$9A,$9E,$17,$FE,$10,$EB,$F2,$CE,$EF,$FC,$37
	.dc.b	$18,$8E,$1D,$1E,$4A,$7D,$13,$E3,$FA,$E3,$FD,$27,$1F,$F1,$1C,$E5
	.dc.b	$F1,$E1,$C9,$CD,$70,$C3,$A5,$6E,$BD,$7F,$45,$D0,$F0,$7D,$17,$F4
	.dc.b	$FC,$39,$E2,$9D,$47,$F8,$CA,$7F,$CD,$CE,$A3,$8F,$B9,$AB,$FF,$19
	.dc.b	$56,$9E,$F5,$C6,$7E,$3F,$C5,$FF,$73,$8F,$4B,$67,$FC,$C7,$F4,$0F
	.dc.b	$A2,$DF,$64,$BF,$7F,$D1,$7F,$95,$5E,$31,$F4,$A3,$17,$FE,$6C,$6F
	.dc.b	$07,$E2,$B0,$FF,$04,$5F,$2F,$7B,$EC,$C6,$2F,$87,$CA,$FD,$CF,$A2
	.dc.b	$78,$1D,$63,$0C,$C3,$E8,$CF,$D5,$AF,$F3,$DB,$C2,$33,$C7,$CF,$97
	.dc.b	$F0,$FE,$1F,$F8,$46,$78,$C6,$3C,$AF,$C1,$79,$3F,$27,$6F,$F2,$CB
	.dc.b	$6B,$F3,$6B,$E0,$F7,$C9,$F8,$9F,$A2,$FF,$35,$57,$F9,$A1,$7C,$06
	.dc.b	$7E,$03,$4E,$FD,$BE,$1E,$DD,$BE,$0F,$E4,$FD,$97,$A3,$7E,$8B,$0D
	.dc.b	$AF,$F3,$AC,$FF,$9E,$9B,$7C,$6F,$93,$F9,$DF,$A2,$FA,$73,$FB,$46
	.dc.b	$F2,$FF,$AC,$D7,$C1,$EF,$AB,$F4,$9F,$45,$F8,$5E,$3F,$F3,$DB,$70
	.dc.b	$FD,$EA,$F7,$37,$DD,$FA,$BF,$9F,$EB,$67,$AD,$FF,$A3,$1C,$98,$CF
	.dc.b	$FD,$15,$F4,$4F,$A1,$FD,$A3,$F6,$E3,$31,$8E,$EF,$DC,$7D,$13,$E8
	.dc.b	$6F,$83,$86,$E5,$FE,$9A,$F8,$BE,$8D,$C3,$F8,$04,$75,$D3,$FD,$2E
	.dc.b	$EB,$2F,$F4,$E3,$E9,$AE,$5A,$70,$F9,$5E,$2F,$F0,$9F,$A0,$F7,$93
	.dc.b	$8F,$1C,$72,$FF,$D5,$73,$96,$C6,$71,$FF,$55,$4E,$5B,$1F,$C5,$7E
	.dc.b	$8B,$F4,$D9,$FF,$AB,$1F,$F2,$75,$FF,$AD,$9C,$BF,$CC,$1C,$D7,$FA
	.dc.b	$D5,$CB,$FE,$F1,$73,$3C,$FE,$07,$B7,$89,$FC,$B3,$E8,$9F,$1D,$4F
	.dc.b	$77,$0D,$CB,$F9,$0C,$E9,$FF,$63,$FC,$EB,$E1,$B5,$A7,$FB,$16,$31
	.dc.b	$3E,$7F,$8A,$F6,$1E,$DE,$FB,$19,$25,$7A,$63,$2F,$AE,$6E,$BF,$0E
	.dc.b	$6D,$7F,$50,$E0,$73,$7D,$5F,$A2,$FA,$CD,$9F,$A6,$DC,$7E,$8B,$70
	.dc.b	$FE,$C3,$A7,$F3,$96,$FA,$BF,$4F,$C0,$F8,$CC,$C3,$FD,$CF,$F4,$4C
	.dc.b	$B1,$97,$FB,$96,$31,$3F,$82,$FA,$2F,$C4,$79,$7F,$89,$6C,$64,$FC
	.dc.b	$3C,$65,$A6,$6B,$F0,$F1,$96,$C7,$E0,$FC,$4F,$C4,$1F,$89,$F0,$6C
	.dc.b	$64,$93,$F8,$58,$CA,$D3,$FC,$03,$F4,$4F,$A2,$FE,$DE,$DB,$FC,$DF
	.dc.b	$7E,$26,$47,$E0,$BE,$26,$E1,$F8,$17,$F3,$5E,$2C,$9F,$EE,$7E,$B5
	.dc.b	$AF,$CB,$4D,$CB,$FB,$D4,$DA,$FC,$E4,$DA,$F3,$7A,$2A,$C5,$B9,$9F
	.dc.b	$3B,$FB,$53,$F6,$DF,$17,$72,$FD,$9C,$5C,$7F,$63,$36,$FF,$C4,$3F
	.dc.b	$43,$FC,$EB,$1F,$F2,$B3,$6B,$FC,$B6,$BD,$07,$D1,$6B,$D1,$FF,$7D
	.dc.b	$57,$F9,$93,$E5,$7F,$A1,$BF,$E5,$F6,$BF,$E6,$46,$BF,$E6,$86,$BF
	.dc.b	$E6,$CF,$93,$7D,$0F,$F9,$33,$6C,$36,$7F,$CE,$73,$A7,$FC,$2B,$F4
	.dc.b	$5F,$E3,$7A,$DC,$7F,$86,$27,$4F,$DF,$8C,$B4,$FF,$3E,$CF,$17,$AD
	.dc.b	$77,$55,$CB,$5F,$F2,$3B,$6F,$FD,$07,$7C,$6B,$77,$C6,$B5,$1D,$0F
	.dc.b	$3D,$66,$64,$79,$DF,$05,$73,$AE,$33,$D0,$D1,$C8,$A7,$AC,$D1,$5E
	.dc.b	$31,$CE,$7B,$CE,$5E,$F5,$8B,$F6,$C5,$8F,$69,$67,$D4,$71,$3C,$45
	.dc.b	$24,$F2,$72,$D7,$31,$F4,$A3,$FC,$E3,$5E,$59,$F5,$C6,$CC,$CF,$86
	.dc.b	$3D,$B1,$98,$F2,$C7,$07,$A1,$1F,$BA,$F1,$BF,$96,$7C,$47,$17,$D6
	.dc.b	$FC,$55,$E6,$3D,$F7,$E2,$7B,$18,$AE,$CA,$78,$CF,$96,$34,$78,$8E
	.dc.b	$27,$AA,$78,$CF,$07,$9D,$F9,$67,$6F,$A2,$FE,$BC,$61,$F4,$4F,$AA
	.dc.b	$3A,$D7,$37,$C2,$77,$3C,$47,$B4,$C1,$1E,$0E,$2B,$C4,$ED,$F1,$3F
	.dc.b	$72,$7C,$F5,$E2,$8F,$21,$E2,$8F,$23,$DE,$BA,$D7,$28,$FC,$C4,$62
	.dc.b	$B9,$47,$76,$DF,$CB,$8F,$24,$7C,$27,$47,$73,$C2,$77,$BF,$0C,$F1
	.dc.b	$AE,$D3,$B3,$89,$B3,$8C,$6E,$78,$9F,$58,$B1,$F3,$27,$9B,$C4,$E7
	.dc.b	$3C,$5E,$73,$C5,$A7,$17,$9B,$F4,$5B,$7F,$42,$7F,$87,$FA,$FF,$C2
	.dc.b	$C9,$5F,$C7,$7D,$8A,$FE,$FE,$2C,$FF,$16,$9F,$E3,$5F,$44,$F4,$3F
	.dc.b	$8A,$7C,$3C,$C3,$FD,$7B,$39,$6C,$F0,$23,$FC,$AF,$D4,$FE,$49,$F4
	.dc.b	$5F,$D4,$23,$FD,$8A,$D7,$F8,$93,$71,$FF,$76,$AE,$5E,$3C,$D2,$B8
	.dc.b	$BD,$23,$11,$CA,$2D,$72,$9F,$0F,$A3,$E2,$F4,$A5,$F8,$A3,$2F,$92
	.dc.b	$32,$F9,$23,$F9,$E7,$D1,$3E,$6F,$D7,$F4,$12,$BF,$DB,$F3,$96,$CF
	.dc.b	$02,$3F,$03,$E6,$7F,$A3,$7D,$13,$D0,$BE,$84,$AF,$F7,$4C,$E5,$B3
	.dc.b	$97,$F9,$C4,$27,$9C,$C9,$E3,$1B,$D5,$CE,$99,$7D,$5E,$7F,$45,$E4
	.dc.b	$23,$E2,$FA,$0A,$E8,$F2,$C7,$AA,$3F,$B0,$77,$B6,$F6,$E1,$2F,$C1
	.dc.b	$8F,$89,$F1,$56,$DC,$CF,$95,$FF,$04,$57,$42,$3A,$3D,$5F,$42,$FA
	.dc.b	$2F,$EE,$0B,$AE,$F5,$CA,$E0,$27,$4E,$1A,$FF,$63,$7F,$DB,$4E,$63
	.dc.b	$AC,$C8,$C1,$0E,$0F,$8B,$D8,$91,$D5,$F2,$F6,$62,$3B,$45,$D3,$E2
	.dc.b	$47,$96,$7E,$5F,$95,$FE,$0B,$DB,$F4,$53,$EA,$6B,$FD,$E7,$64,$F0
	.dc.b	$8F,$7C,$61,$B1,$CC,$E9,$CD,$23,$A3,$DF,$9F,$11,$FE,$05,$B4,$E3
	.dc.b	$1F,$3C,$7F,$B9,$27,$89,$0E,$13,$CD,$FB,$12,$F4,$FF,$C8,$91,$87
	.dc.b	$FC,$7F,$1F,$85,$35,$E1,$FF,$8E,$5E,$19,$5F,$0F,$FD,$63,$7D,$1A
	.dc.b	$FF,$8D,$65,$18,$7A,$2D,$3A,$7D,$07,$E8,$67,$C0,$DE,$4F,$CB,$B6
	.dc.b	$FF,$D0,$0F,$EB,$DF,$47,$D1,$79,$73,$34,$C7,$5F,$9F,$E8,$BD,$1F
	.dc.b	$39,$2F,$D1,$1E,$2B,$F1,$FA,$52,$7E,$3E,$3F,$1D,$F8,$DD,$BD,$FB
	.dc.b	$64,$60,$F8,$5E,$37,$C5,$23,$F1,$7E,$2B,$DF,$E9,$C8,$EA,$FC,$2F
	.dc.b	$43,$E7,$7D,$17,$FA,$1E,$7E,$BB,$5F,$98,$DE,$2F,$AC,$BC,$3F,$A9
	.dc.b	$EE,$FD,$5F,$4E,$5F,$D5,$8F,$B5,$F4,$5C,$79,$BF,$81,$BC,$67,$F0
	.dc.b	$AD,$7E,$D3,$B7,$F8,$C7,$D1,$3E,$0B,$FC,$DC,$D8,$FD,$EB,$5F,$F2
	.dc.b	$E2,$D7,$FC,$E9,$F1,$4F,$8D,$A4,$AF,$E0,$B0,$F6,$F6,$48,$CB,$FE
	.dc.b	$7D,$FA,$07,$B5,$D3,$CF,$95,$C9,$FF,$41,$7D,$13,$96,$92,$7F,$89
	.dc.b	$C6,$5F,$E2,$B6,$E1,$EC,$6E,$3F,$CE,$D3,$B7,$FC,$F1,$3B,$7F,$8E
	.dc.b	$FD,$13,$E1,$FC,$D8,$49,$CC,$7F,$A5,$67,$51,$82,$1D,$3B,$F5,$32
	.dc.b	$33,$CF,$FC,$5B,$05,$73,$E0,$F8,$9E,$14,$BF,$1C,$65,$A4,$3D,$3E
	.dc.b	$AF,$FA,$9F,$E8,$9F,$3B,$49,$3F,$CE,$21,$F4,$58,$8F,$F4,$C4,$70
	.dc.b	$75,$3C,$E7,$FD,$65,$83,$E7,$29,$2B,$FA,$1F,$D0,$F4,$D7,$FB,$27
	.dc.b	$D0,$7F,$05,$9F,$EC,$D7,$C5,$FE,$92,$D7,$FD,$50,$D9,$FE,$73,$5C
	.dc.b	$E5,$5C,$D4,$66,$31,$5C,$DB,$7F,$D8,$5E,$13,$B9,$FE,$C8,$DA,$F2
	.dc.b	$D6,$E7,$FB,$41,$CC,$DE,$0C,$3D,$B1,$86,$D7,$22,$9F,$C9,$6B,$A3
	.dc.b	$FE,$07,$BF,$03,$EA,$BE,$8F,$BA,$FC,$35,$C2,$39,$1D,$2B,$89,$CA
	.dc.b	$7C,$D1,$E1,$39,$1D,$EB,$AE,$1F,$09,$EF,$1F,$E0,$53,$C5,$85,$78
	.dc.b	$35,$C9,$9B,$7D,$13,$E8,$B5,$18,$6C,$66,$78,$35,$E9,$F4,$5D,$5F
	.dc.b	$F1,$53,$53,$48,$FB,$0F,$A5,$A7,$F8,$87,$2F,$C2,$47,$23,$58,$FE
	.dc.b	$8B,$BF,$E9,$C6,$67,$F1,$53,$98,$E8,$67,$4F,$F1,$44,$66,$7A,$CF
	.dc.b	$90,$EB,$7E,$1A,$F9,$18,$7E,$02,$F8,$46,$2B,$78,$78,$9E,$58,$78
	.dc.b	$B0,$E5,$1F,$4F,$0E,$51,$88,$FF,$1D,$46,$2A,$47,$E4,$DF,$3C,$FF
	.dc.b	$7F,$7C,$15,$D5,$FF,$1B,$E1,$EF,$3F,$C4,$A3,$E8,$B8,$5F,$F9,$32
	.dc.b	$EB,$F9,$F8,$BF,$43,$FC,$06,$FF,$C8,$2D,$9F,$11,$FE,$63,$7F,$CC
	.dc.b	$91,$AD,$F8,$9D,$B0,$FE,$7F,$86,$27,$FC,$37,$D7,$E5,$7F,$CB,$17
	.dc.b	$FC,$FD,$FF,$0F,$FD,$13,$E3,$FC,$7D,$5C,$E3,$13,$E4,$8C,$4E,$A3
	.dc.b	$8B,$EA,$FA,$29,$1F,$E2,$AF,$A2,$FE,$8C,$FB,$23,$3A,$FF,$88,$67
	.dc.b	$FC,$C5,$39,$38,$7B,$7A,$D7,$FC,$5D,$3A,$AF,$86,$33,$61,$F0,$67
	.dc.b	$55,$FE,$56,$F8,$BF,$ED,$98,$FF,$9B,$1B,$87,$2F,$A0,$7D,$17,$D4
	.dc.b	$9C,$DF,$1E,$CF,$AF,$F4,$5F,$E4,$CC,$79,$46,$3D,$B9,$1F,$E7,$8C
	.dc.b	$7F,$C0,$0F,$F9,$EA,$B8,$65,$DF,$CA,$E7,$F2,$DF,$82,$FA,$27,$83
	.dc.b	$8C,$3F,$D0,$31,$86,$F2,$6C,$E1,$36,$3F,$D0,$FE,$27,$F8,$1D,$BE
	.dc.b	$5F,$5E,$8F,$C7,$CE,$73,$FF,$28,$79,$5F,$22,$FF,$CB,$EF,$F8,$19
	.dc.b	$BC,$BF,$8B,$8C,$4F,$F9,$5F,$E8,$9E,$67,$E3,$DB,$7F,$E6,$85,$F4
	.dc.b	$59,$9D,$72,$4A,$D7,$A7,$70,$FF,$31,$7A,$AF,$88,$FC,$D3,$6B,$FC
	.dc.b	$E6,$BE,$07,$33,$AC,$37,$85,$BD,$BF,$A3,$FA,$27,$CE,$5E,$6C,$36
	.dc.b	$BF,$CF,$FF,$3C,$ED,$FD,$53,$FE,$47,$8C,$C6,$3D,$8C,$D4,$9E,$37
	.dc.b	$76,$FF,$3E,$7D,$0B,$E2,$3F,$68,$DA,$FF,$46,$BF,$E7,$D6,$C7,$EE
	.dc.b	$7E,$85,$F3,$B4,$C7,$CE,$5A,$FF,$4A,$FC,$5F,$3C,$7F,$D1,$4F,$36
	.dc.b	$E1,$FC,$0E,$31,$E2,$7F,$A5,$DF,$DD,$CF,$0F,$8E,$D7,$95,$B7,$F8
	.dc.b	$57,$D0,$7E,$6A,$7F,$D2,$CF,$F8,$DF,$1F,$F5,$2B,$9C,$3F,$D4,$4E
	.dc.b	$67,$7E,$27,$2E,$FE,$6D,$BE,$2D,$9D,$C7,$49,$FE,$4B,$F3,$FE,$DC
	.dc.b	$3F,$D9,$ED,$9F,$F7,$AF,$CE,$7D,$C7,$FD,$CF,$19,$B0,$71,$4F,$A0
	.dc.b	$7C,$1F,$6D,$FF,$94,$46,$30,$F6,$FC,$37,$C2,$3E,$1F,$7D,$8C,$93
	.dc.b	$2F,$8B,$19,$7F,$D7,$99,$B9,$BF,$9A,$3F,$72,$BF,$D9,$AE,$5B,$C9
	.dc.b	$A9,$DB,$63,$27,$08,$CB,$4C,$D7,$F2,$4D,$7E,$EE,$BF,$23,$8B,$ED
	.dc.b	$FC,$4F,$CC,$65,$FF,$B7,$BE,$83,$C9,$87,$19,$FE,$95,$18,$BF,$F6
	.dc.b	$F6,$6E,$6F,$5F,$8B,$EC,$6E,$3E,$C6,$E3,$F0,$9B,$7F,$86,$F8,$AF
	.dc.b	$C6,$9F,$8E,$F0,$6D,$FE,$32,$2D,$FE,$42,$6D,$71,$AD,$1F,$92,$8C
	.dc.b	$BF,$80,$97,$D0,$3E,$83,$FB,$7B,$67,$E8,$F5,$E2,$64,$7F,$8C,$E7
	.dc.b	$55,$FE,$37,$6B,$FE,$3B,$6B,$F8,$EC,$1D,$4F,$0F,$F5,$27,$F7,$3F
	.dc.b	$2B,$F8,$66,$1F,$A8,$8C,$B4,$FE,$EB,$E1,$63,$0F,$C4,$FC,$CF,$E2
	.dc.b	$0F,$C4,$FC,$45,$AF,$C4,$C5,$9F,$C7,$CD,$CB,$EB,$CD,$AF,$D3,$CD
	.dc.b	$AF,$D2,$FD,$E7,$F8,$56,$7F,$CA,$4D,$8F,$F2,$B3,$5F,$F2,$D3,$5F
	.dc.b	$9D,$F4,$2F,$95,$FF,$78,$46,$1A,$FE,$CA,$1F,$2A,$FA,$1F,$F3,$04
	.dc.b	$71,$D3,$FC,$C5,$F2,$2F,$A1,$FF,$22,$E7,$F6,$E3,$18,$FC,$F8,$C4
	.dc.b	$EA,$3F,$7F,$F4,$3F,$E3,$09,$CC,$6C,$FE,$07,$5C,$9F,$F0,$4B,$77
	.dc.b	$4F,$E2,$6B,$95,$70,$F7,$FB,$2B,$46,$1B,$1D,$CD,$7C,$D6,$66,$46
	.dc.b	$67,$6F,$8A,$72,$F9,$26,$E1,$E4,$8F,$4C,$E3,$0F,$89,$56,$72,$75
	.dc.b	$70,$F9,$DF,$34,$E2,$FC,$37,$D6,$B4,$79,$67,$55,$B6,$C7,$53,$45
	.dc.b	$3D,$05,$9E,$94,$9E,$33,$A3,$E0,$1B,$BF,$5B,$4F,$59,$BA,$E7,$7E
	.dc.b	$38,$D1,$E8,$35,$3F,$87,$AE,$F5,$C2,$7C,$E6,$8F,$B8,$72,$BF,$33
	.dc.b	$6F,$B3,$B3,$B4,$F3,$7D,$26,$67,$E5,$DF,$AA,$33,$3B,$24,$65,$F8
	.dc.b	$B7,$E6,$8C,$1E,$99,$F1,$1E,$58,$D3,$E8,$3A,$C7,$07,$13,$D9,$F9
	.dc.b	$2D,$7D,$31,$B7,$F2,$A7,$63,$D5,$3D,$9D,$9C,$A3,$85,$7B,$A3,$BC
	.dc.b	$79,$DF,$85,$7F,$22,$FC,$15,$B9,$E6,$78,$0D,$9E,$D3,$C5,$1C,$2F
	.dc.b	$C5,$3C,$4F,$AB,$3C,$0F,$39,$C2,$FC,$B3,$C2,$FF,$3F,$39,$3C,$E7
	.dc.b	$AC,$A7,$92,$BB,$4F,$18,$FC,$3D,$FA,$AB,$D9,$7F,$09,$B3,$CA,$FC
	.dc.b	$D1,$E0,$8F,$35,$7A,$DF,$11,$DE,$FB,$3F,$68,$E5,$3D,$1F,$23,$6B
	.dc.b	$AC,$7C,$73,$AD,$FC,$07,$AC,$7A,$9B,$3E,$3B,$FD,$F7,$D0,$BE,$87
	.dc.b	$F8,$C7,$C4,$CB,$FE,$31,$39,$6C,$E5,$FE,$4F,$3F,$C8,$A7,$51,$FE
	.dc.b	$B2,$FA,$17,$2F,$F2,$66,$C7,$F2,$36,$BB,$E5,$D4,$C8,$D3,$C6,$39
	.dc.b	$7C,$0F,$3F,$9D,$95,$CE,$32,$F5,$8C,$BE,$38,$CB,$E5,$8C,$E7,$FC
	.dc.b	$7F,$05,$7F,$3C,$FA,$17,$CE,$7A,$7E,$82,$63,$FC,$EE,$72,$D9,$FF
	.dc.b	$07,$32,$32,$E0,$38,$75,$19,$7F,$CA,$BF,$42,$F3,$BF,$A6,$AF,$3A
	.dc.b	$98,$89,$3F,$D3,$67,$4D,$9C,$BF,$EE,$48,$FF,$55,$3B,$F6,$FA,$9D
	.dc.b	$BD,$CB,$19,$21,$C5,$CB,$CE,$2E,$BF,$D1,$EB,$FD,$65,$3F,$D7,$FB
	.dc.b	$5A,$7A,$B9,$E4,$E6,$3F,$B4,$CE,$A3,$04,$3F,$DF,$7F,$42,$F5,$FF
	.dc.b	$B7,$BF,$C2,$4F,$E0,$38,$FF,$57,$6E,$1F,$E3,$F7,$5D,$1C,$3A,$A6
	.dc.b	$1F,$DC,$E7,$97,$A9,$D0,$91,$9E,$9A,$F8,$A2,$E3,$DE,$6E,$3D,$E6
	.dc.b	$D7,$FA,$3A,$D5,$EE,$F2,$3F,$C0,$FC,$0E,$8A,$66,$FE,$44,$D5,$97
	.dc.b	$F7,$3D,$93,$C8,$E7,$CD,$2B,$FC,$1D,$87,$E6,$9B,$D1,$97,$9E,$BF
	.dc.b	$E6,$39,$D4,$60,$FF,$52,$B8,$C7,$FC,$04,$AB,$D3,$F6,$FE,$AE,$29
	.dc.b	$7E,$08,$CB,$E2,$8C,$BD,$4F,$F7,$B4,$69,$C5,$7E,$05,$B3,$9B,$FF
	.dc.b	$0F,$0F,$9E,$7C,$77,$F8,$F6,$30,$D8,$CD,$7F,$8F,$67,$87,$C8,$6F
	.dc.b	$E8,$BF,$AB,$41,$FC,$3C,$FC,$E6,$C6,$79,$7B,$7C,$FB,$4B,$9C,$C3
	.dc.b	$B4,$E5,$F0,$64,$F8,$0F,$27,$B3,$27,$31,$E3,$F8,$79,$1E,$7D,$11
	.dc.b	$E7,$F3,$BC,$3F,$11,$2B,$BB,$E4,$FA,$9F,$81,$E2,$C9,$F0,$BF,$0F
	.dc.b	$A7,$E0,$7C,$2F,$C1,$F2,$25,$77,$8F,$6F,$2F,$BB,$96,$4F,$B2,$3E
	.dc.b	$77,$DF,$3E,$96,$FC,$64,$AF,$9D,$39,$7D,$5E,$F7,$CF,$23,$ED,$F1
	.dc.b	$3E,$5F,$CD,$C8,$F9,$2F,$CE,$F9,$9F,$F3,$5F,$D1,$3D,$CB,$F1,$12
	.dc.b	$FF,$CD,$53,$97,$FC,$D9,$F1,$AE,$5F,$F3,$9B,$63,$31,$8F,$81,$91
	.dc.b	$C1,$AF,$E5,$7E,$7D,$D1,$CB,$9A,$4F,$67,$FC,$FF,$F4,$2F,$92,$A4
	.dc.b	$9F,$F3,$F4,$65,$FF,$40,$5B,$E8,$BF,$D0,$ED,$8D,$D7,$17,$C5,$3F
	.dc.b	$E8,$CF,$A2,$7C,$77,$0A,$95,$FE,$8B,$9C,$B6,$78,$11,$F0,$FD,$15
	.dc.b	$7F,$65,$3C,$1B,$3E,$49,$F0,$70,$F9,$BC,$F9,$5D,$E3,$2D,$21,$F2
	.dc.b	$7A,$3F,$D4,$7F,$45,$FB,$28,$FE,$00,$DF,$2B,$FD,$18,$DF,$37,$FD
	.dc.b	$E8,$BA,$38,$F6,$CB,$FF,$54,$4E,$D0,$78,$3F,$EA,$DC,$9F,$3D,$FE
	.dc.b	$B4,$6C,$66,$3F,$D6,$FF,$45,$DF,$2E,$7F,$41,$FC,$DA,$BA,$DF,$07
	.dc.b	$F8,$C3,$67,$FD,$48,$DC,$FC,$72,$BE,$8A,$FA,$39,$CF,$F9,$DB,$9B
	.dc.b	$E2,$77,$9D,$B4,$EB,$5D,$23,$9B,$63,$8E,$3E,$29,$E2,$D8,$E3,$7E
	.dc.b	$09,$E2,$D8,$E3,$7F,$09,$AF,$94,$F9,$F7,$DD,$B3,$C5,$F8,$F7,$DD
	.dc.b	$A7,$33,$78,$D7,$E1,$36,$FD,$D1,$C0,$F0,$1A,$2D,$FC,$C7,$FC,$5A
	.dc.b	$D7,$DB,$3F,$E2,$98,$C4,$7A,$3B,$1F,$45,$96,$C7,$F8,$CE,$51,$88
	.dc.b	$EA,$D7,$FC,$71,$F4,$5F,$E3,$57,$F2,$2D,$4D,$23,$EA,$BF,$EA,$FA
	.dc.b	$7C,$18,$E2,$67,$6F,$F7,$D6,$BF,$86,$8C,$BF,$88,$AE,$31,$F1,$0C
	.dc.b	$E7,$F9,$08,$CC,$F1,$9F,$19,$D6,$FC,$0E,$23,$A5,$F8,$2F,$DB,$3B
	.dc.b	$3B,$1B,$7B,$46,$E3,$FC,$BB,$87,$DB,$C3,$C3,$5C,$AF,$AE,$19,$C3
	.dc.b	$FB,$CB,$C8,$FD,$CD,$1E,$15,$9C,$3F,$C0,$D3,$F0,$0D,$3C,$3C,$3E
	.dc.b	$6F,$2B,$AF,$F8,$9A,$2F,$D0,$FF,$02,$C3,$B9,$FB,$67,$13,$FE,$13
	.dc.b	$7F,$85,$C6,$A3,$F8,$6C,$6A,$FC,$30,$8C,$4F,$AB,$AF,$CC,$F5,$DF
	.dc.b	$CB,$7F,$D2,$B1,$F1,$F3,$E3,$E8,$9E,$1E,$30,$FF,$0B,$46,$1F,$16
	.dc.b	$7D,$EB,$85,$6E,$3F,$C6,$D8,$2F,$84,$7A,$FC,$AF,$82,$78,$BE,$8B
	.dc.b	$FA,$86,$3D,$35,$FF,$06,$B6,$75,$E9,$F0,$8F,$F1,$45,$23,$17,$E0
	.dc.b	$8C,$4F,$13,$73,$F4,$BE,$1F,$FB,$C4,$7A,$9B,$5F,$E4,$1F,$A1,$7D
	.dc.b	$16,$FB,$25,$FC,$0F,$A2,$FF,$58,$BC,$63,$EE,$C6,$2F,$FC,$94,$DD
	.dc.b	$BE,$AE,$0F,$27,$7E,$D7,$E1,$23,$18,$F5,$F5,$3F,$0F,$F4,$4F,$07
	.dc.b	$AC,$61,$CF,$2C,$AF,$A7,$5B,$6C,$7E,$43,$C3,$FF,$08,$47,$D8,$6E
	.dc.b	$BF,$E5,$FA,$FF,$05,$CE,$7C,$AF,$C2,$F9,$5F,$4A,$BE,$DB,$A6,$F2
	.dc.b	$EB,$AF,$F1,$7F,$44,$F3,$29,$8F,$A2,$B9,$7F,$97,$E7,$E7,$F1,$FF
	.dc.b	$9A,$1B,$1F,$B0,$5F,$25,$D3,$9B,$F2,$FF,$0C,$F8,$6A,$63,$DF,$B8
	.dc.b	$7F,$9A,$E7,$9E,$DF,$E9,$47,$9C,$66,$F9,$E1,$FA,$2F,$A2,$FF,$31
	.dc.b	$4F,$F9,$CD,$73,$67,$DC,$D3,$FA,$8F,$61,$DD,$7E,$8B,$0D,$C3,$FC
	.dc.b	$F9,$3D,$34,$FD,$1B,$78,$3F,$D0,$EF,$69,$DE,$3F,$B4,$FA,$27,$C3
	.dc.b	$D3,$1E,$55,$AF,$F4,$47,$D1,$3E,$83,$F7,$6D,$8C,$FC,$E6,$27,$F7
	.dc.b	$FF,$11,$D2,$7F,$D2,$0F,$FB,$01,$B8,$6B,$0D,$F9,$7C,$3D,$DB,$5C
	.dc.b	$1F,$E1,$3F,$33,$F9,$8E,$0F,$DF,$BF,$E4,$29,$FF,$4E,$39,$E0,$FF
	.dc.b	$52,$B9,$9D,$F8,$FC,$BC,$3E,$6D,$BE,$2D,$9D,$C7,$49,$FE,$57,$F4
	.dc.b	$1E,$B9,$FF,$5C,$37,$0F,$F5,$9F,$CF,$7C,$FB,$FF,$10,$CE,$5B,$AF
	.dc.b	$F2,$2F,$A1,$7D,$17,$48,$C3,$EE,$F9,$29,$5E,$CE,$EB,$9F,$C0,$9B
	.dc.b	$9F,$BA,$6D,$7F,$4A,$CF,$EF,$5F,$F2,$47,$EB,$37,$AF,$F9,$9C,$62
	.dc.b	$FD,$31,$8E,$4C,$B6,$FF,$D9,$15,$D2,$BD,$F5,$A9,$CB,$F1,$B8,$A4
	.dc.b	$7D,$4F,$A2,$B8,$FC,$89,$FE,$87,$F4,$5E,$1A,$F8,$F1,$8B,$F2,$F3
	.dc.b	$F9,$BC,$8B,$5E,$26,$E1,$ED,$6E,$5F,$49,$FC,$3F,$D1,$4A,$FC,$2F
	.dc.b	$85,$6F,$F1,$33,$6F,$F0,$F3,$6B,$FB,$9F,$3A,$B1,$6B,$4F,$F0,$06
	.dc.b	$CF,$F7,$0F,$A1,$7D,$15,$3E,$7F,$65,$C7,$FA,$DC,$E9,$FF,$7A,$73
	.dc.b	$FF,$54,$CB,$B4,$7F,$9A,$A7,$51,$FD,$7A,$75,$8F,$E7,$BC,$4F,$D2
	.dc.b	$9F,$DB,$BC,$AF,$E0,$F8,$7E,$92,$32,$D3,$FB,$E4,$FF,$AF,$3C,$3C
	.dc.b	$E1,$F8,$7F,$8C,$FC,$29,$F8,$6F,$88,$B7,$F8,$28,$B3,$F8,$D9,$B7
	.dc.b	$FA,$29,$B5,$F9,$39,$B5,$FE,$96,$F9,$F7,$D0,$FF,$1C,$8F,$F4,$39
	.dc.b	$97,$0E,$21,$F2,$CF,$A1,$DC,$EF,$3A,$BE,$8B,$FD,$42,$E9,$AF,$66
	.dc.b	$BF,$E9,$66,$BF,$EA,$AF,$A0,$7D,$17,$F9,$2F,$4F,$F4,$DC,$62,$7F
	.dc.b	$C2,$6D,$7E,$B9,$FE,$B2,$FA,$2F,$F1,$2E,$FF,$E2,$69,$D3,$6F,$A5
	.dc.b	$77,$6A,$AE,$37,$E0,$F7,$B3,$87,$F8,$46,$BC,$31,$8F,$9D,$CB,$23
	.dc.b	$27,$9A,$30,$F9,$84,$F3,$70,$DC,$7D,$13,$C1,$F5,$55,$BF,$87,$36
	.dc.b	$FE,$44,$D7,$AC,$5B,$E9,$5E,$B9,$F2,$1E,$63,$C3,$5B,$AE,$0D,$8D
	.dc.b	$5F,$89,$B5,$E2,$69,$E2,$29,$E2,$2D,$F8,$9D,$1F,$02,$FB,$BA,$6B
	.dc.b	$E4,$3C,$E5,$25,$78,$DE,$47,$6B,$F4,$3A,$BF,$80,$D3,$F1,$D5,$DA
	.dc.b	$B9,$1F,$42,$39,$1E,$32,$9C,$4F,$61,$6F,$D8,$FD,$CB,$FC,$23,$C2
	.dc.b	$7A,$DF,$C7,$8E,$13,$E6,$9C,$C7,$C4,$6B,$D9,$B5,$D8,$97,$E9,$BF
	.dc.b	$6C,$FA,$6B,$A9,$E9,$29,$E7,$9E,$0F,$79,$F2,$4F,$6B,$F2,$47,$42
	.dc.b	$4E,$CF,$92,$7B,$5C,$BE,$76,$DF,$9E,$7E,$1C,$F8,$AB,$D7,$7E,$28
	.dc.b	$F9,$07,$DE,$8F,$13,$E9,$6B,$E9,$3C,$07,$8C,$F0,$16,$BC,$0F,$3B
	.dc.b	$F1,$DF,$B2,$34,$D7,$C2,$79,$63,$E1,$9C,$1F,$13,$5F,$11,$B3,$C1
	.dc.b	$7E,$BA,$FA,$B1,$E6,$3A,$DF,$D5,$3D,$67,$C2,$75,$1C,$1E,$87,$0A
	.dc.b	$F1,$BE,$83,$C6,$07,$D1,$3E,$5B,$F9,$9C,$61,$B1,$98,$FE,$6E,$D7
	.dc.b	$F9,$DB,$6F,$FD,$9A,$DA,$ED,$3C,$CF,$E6,$3E,$EF,$F2,$8F,$A2,$72
	.dc.b	$D2,$4F,$FB,$22,$1C,$B8,$8F,$E4,$51,$C1,$D4,$F3,$9E,$9F,$01,$D7
	.dc.b	$E7,$A5,$73,$8C,$BD,$63,$2F,$48,$CB,$DA,$33,$9F,$FB,$A3,$05,$7F
	.dc.b	$40,$FA,$27,$D1,$7A,$FE,$7E,$63,$FE,$E0,$9C,$B6,$78,$11,$F7,$7B
	.dc.b	$3F,$D2,$BE,$89,$E7,$DF,$3E,$57,$FB,$AA,$72,$D9,$CB,$FD,$16,$13
	.dc.b	$CE,$7A,$7B,$B2,$B9,$3D,$FD,$DC,$47,$18,$BB,$38,$FE,$CE,$2B,$CC
	.dc.b	$62,$AD,$FE,$2E,$B4,$FF,$C0,$0E,$E6,$DE,$DB,$93,$98,$FE,$D3,$3A
	.dc.b	$8C,$10,$FE,$C1,$F4,$4F,$63,$FB,$7B,$FC,$A0,$75,$60,$CF,$5C,$AF
	.dc.b	$CF,$6B,$EB,$7F,$41,$5C,$1C,$E5,$F9,$68,$C4,$78,$2B,$BF,$AB,$23
	.dc.b	$67,$27,$C7,$EC,$E6,$3C,$F1,$70,$F0,$42,$D5,$9C,$3C,$17,$6B,$CB
	.dc.b	$3F,$13,$C9,$FF,$06,$F0,$7E,$8A,$7C,$27,$D3,$E9,$D3,$F3,$E6,$BF
	.dc.b	$E0,$7D,$91,$C8,$E7,$CF,$27,$9D,$FE,$71,$BD,$39,$7F,$C2,$11,$96
	.dc.b	$CF,$92,$7C,$BF,$45,$E6,$E2,$95,$D2,$32,$F6,$8C,$BD,$23,$2F,$68
	.dc.b	$CE,$FF,$23,$05,$7F,$8C,$21,$F4,$2F,$8E,$C3,$70,$FF,$32,$B5,$FF
	.dc.b	$1E,$CF,$0F,$90,$E5,$F4,$3F,$E0,$B8,$3F,$80,$7F,$0C,$63,$97,$7C
	.dc.b	$BE,$1F,$A2,$F1,$E1,$FE,$72,$7F,$30,$F6,$6D,$72,$77,$1E,$2C,$5F
	.dc.b	$05,$E8,$F7,$24,$FA,$23,$CB,$F1,$12,$3F,$CF,$F9,$A3,$D3,$E8,$78
	.dc.b	$FE,$26,$57,$8D,$F3,$7D,$17,$C4,$F1,$A4,$FC,$47,$E3,$F4,$FC,$2F
	.dc.b	$86,$F8,$7E,$54,$AE,$F1,$EF,$E4,$F8,$1C,$92,$7E,$04,$7D,$3F,$A1
	.dc.b	$7D,$15,$F2,$E5,$7F,$A5,$A7,$2F,$AB,$DF,$FA,$24,$7D,$BE,$27,$CD
	.dc.b	$F9,$B9,$1F,$35,$FA,$1F,$33,$FE,$6E,$FA,$27,$BF,$7E,$26,$5F,$E9
	.dc.b	$27,$2F,$E9,$BE,$3D,$CB,$FE,$76,$6C,$66,$31,$F0,$52,$38,$35,$FF
	.dc.b	$38,$7D,$0B,$A3,$97,$34,$9E,$CF,$FA,$0B,$E8,$9F,$27,$49,$3F,$B8
	.dc.b	$8C,$BF,$BA,$B7,$D1,$7F,$A2,$9B,$1B,$AE,$2F,$8A,$7F,$D1,$FF,$44
	.dc.b	$F8,$FE,$15,$2B,$F8,$2C,$E5,$B3,$C0,$8F,$87,$E8,$AB,$FE,$8D,$9E
	.dc.b	$0D,$9F,$24,$F8,$38,$7C,$DE,$7C,$AE,$F1,$96,$90,$F9,$3D,$1F,$EA
	.dc.b	$6F,$A2,$FD,$A7,$8F,$FC,$91,$BE,$97,$F8,$19,$B5,$FE,$15,$7F,$D8
	.dc.b	$73,$BE,$DE,$44,$9F,$E4,$73,$AA,$FF,$07,$E2,$F9,$EF,$F5,$AB,$6B
	.dc.b	$F9,$4B,$5F,$F5,$67,$D0,$BA,$D9,$7F,$89,$7E,$83,$F9,$BD,$7F,$89
	.dc.b	$EF,$84,$FF,$3C,$6E,$3E,$0C,$7F,$D8,$78,$F3,$AE,$55,$FE,$22,$8C
	.dc.b	$5B,$0E,$AF,$F4,$4B,$FF,$13,$C6,$2B,$CC,$EA,$B8,$B7,$1F,$3D,$F3
	.dc.b	$AF,$13,$E0,$3D,$17,$D2,$3C,$53,$C2,$B9,$61,$BB,$F1,$16,$BC,$27
	.dc.b	$08,$D6,$1E,$28,$EF,$3E,$13,$B9,$AB,$ED,$1C,$63,$47,$62,$BD,$9F
	.dc.b	$F2,$A4,$F5,$6C,$FB,$9F,$F2,$58,$8F,$96,$E1,$AF,$73,$1D,$8F,$A2
	.dc.b	$E3,$1D,$A3,$0E,$25,$18,$8E,$33,$FE,$64,$5F,$45,$48,$FF,$9A,$1B
	.dc.b	$97,$DC,$CB,$FA,$B6,$9F,$36,$BF,$CE,$15,$F3,$E7,$31,$F1,$0C,$9F
	.dc.b	$19,$CD,$7E,$1A,$32,$F5,$DB,$FC,$38,$E6,$3C,$A7,$18,$CE,$9F,$02
	.dc.b	$FC,$11,$89,$F9,$71,$8C,$3C,$D1,$CA,$FB,$46,$27,$53,$E7,$35,$7F
	.dc.b	$E7,$EC,$3C,$ED,$C3,$FB,$5D,$D3,$FC,$33,$5E,$A9,$C3,$FB,$08,$B5
	.dc.b	$FE,$28,$1E,$1F,$29,$FF,$1D,$DD,$7F,$6F,$17,$E8,$7F,$80,$5F,$F8
	.dc.b	$99,$B3,$D4,$FF,$50,$3F,$EA,$18,$D6,$FC,$4E,$D8,$7F,$3F,$C3,$13
	.dc.b	$FE,$40,$8C,$3D,$79,$B7,$F4,$4F,$0F,$C5,$18,$6C,$F1,$9D,$3C,$4C
	.dc.b	$47,$A7,$CA,$91,$DF,$E8,$BF,$A2,$BE,$B8,$CE,$9F,$E1,$C9,$7A,$38
	.dc.b	$CB,$FC,$33,$18,$BE,$07,$39,$D3,$FE,$64,$8D,$57,$C5,$1F,$13,$FD
	.dc.b	$BA,$7F,$C9,$ED,$8D,$7D,$03,$E8,$BE,$94,$E7,$1F,$F2,$FB,$6F,$FD
	.dc.b	$83,$F4,$5F,$EA,$87,$9F,$67,$D9,$8C,$63,$E1,$6D,$FF,$9A,$31,$79
	.dc.b	$3B,$F6,$FF,$01,$18,$C7,$FC,$C7,$6F,$53,$F0,$9F,$44,$F0,$3B,$46
	.dc.b	$1C,$F2,$4B,$FF,$35,$D6,$DB,$1F,$8D,$F0,$BF,$C1,$31,$FE,$70,$6E
	.dc.b	$BF,$1E,$BF,$C0,$93,$9F,$2B,$F2,$8F,$FA,$80,$FF,$3B,$96,$7F,$CA
	.dc.b	$7E,$53,$D0,$DF,$B7,$C3,$E5,$E5,$FE,$3F,$E8,$5F,$09,$8C,$7E,$B7
	.dc.b	$E2,$1F,$EC,$6A,$FF,$41,$B6,$3F,$CB,$FF,$86,$FD,$A9,$CF,$C3,$FE
	.dc.b	$9F,$EC,$DF,$1F,$4F,$CF,$FC,$53,$EA,$31,$7F,$E8,$EA,$7C,$97,$13
	.dc.b	$F4,$BF,$4B,$D7,$3F,$C4,$4E,$AB,$FD,$2B,$F8,$2D,$9F,$C6,$37,$BB
	.dc.b	$FC,$DB,$6F,$9F,$FD,$D4,$E6,$D7,$FC,$9B,$E6,$BF,$7D,$E7,$1F,$DA
	.dc.b	$F3,$7F,$A1,$2B,$FD,$47,$39,$7F,$D4,$BF,$74,$F8,$CF,$E0,$CD,$8F
	.dc.b	$DD,$57,$85,$DC,$FF,$0B,$F7,$7F,$83,$EB,$FE,$B8,$6C,$FF,$B2,$BE
	.dc.b	$9D,$F5,$5B,$F4,$3F,$89,$7C,$DB,$E4,$A9,$FC,$2B,$EA,$1E,$06,$31
	.dc.b	$FF,$5C,$D7,$E0,$5B,$AF,$F1,$0F,$92,$EB,$18,$3F,$D7,$F4,$FA,$2F
	.dc.b	$E4,$0D,$8F,$E2,$93,$F2,$78,$FE,$4C,$E5,$FD,$E7,$D4,$79,$6B,$FD
	.dc.b	$92,$D8,$FF,$52,$FD,$3F,$FA,$52,$73,$A7,$F2,$DA,$D7,$0F,$FB,$69
	.dc.b	$B3,$FE,$EC,$AE,$17,$DA,$F8,$1F,$CA,$BE,$89,$F5,$18,$BF,$E5,$51
	.dc.b	$87,$F9,$67,$D3,$7E,$67,$2F,$F5,$8B,$63,$FD,$F2,$BB,$FF,$9A,$B9
	.dc.b	$C7,$FC,$6E,$DA,$34,$E1,$5D,$BC,$1F,$E7,$B9,$FF,$07,$8F,$E7,$8F
	.dc.b	$E0,$EF,$E1,$F9,$BA,$ED,$DB,$6B,$6D,$9F,$9B,$EC,$7C,$73,$CB,$F5
	.dc.b	$16,$BE,$7C,$59,$FF,$69,$E4,$AC,$67,$F3,$6B,$FA,$8F,$D4,$78,$E3
	.dc.b	$18,$7F,$51,$8C,$3F,$47,$CA,$F5,$F1,$FB,$1B,$1E,$C6,$E5,$F7,$9B
	.dc.b	$7F,$3B,$97,$E5,$9F,$33,$D5,$B7,$F5,$22,$DF,$A2,$6D,$2B,$F6,$F2
	.dc.b	$FA,$77,$D4,$7F,$75,$6C,$FC,$EE,$8C,$4C,$8F,$DF,$DF,$3B,$FE,$03
	.dc.b	$7C,$1F,$CA,$DE,$DF,$EA,$D1,$99,$FF,$7B,$F2,$FF,$66,$9C,$E1,$F9
	.dc.b	$1F,$8A,$FC,$3F,$89,$32,$FE,$F5,$E0,$FF,$19,$C7,$F1,$F3,$9C,$3E
	.dc.b	$AB,$FE,$02,$9D,$57,$E4,$A3,$F1,$3F,$29,$23,$F1,$3F,$1D,$6F,$F5
	.dc.b	$73,$70,$F6,$0E,$85,$66,$FE,$4F,$D3,$7F,$45,$C3,$FC,$BA,$D7,$3E
	.dc.b	$97,$C5,$72,$D7,$8F,$C2,$57,$F9,$4C,$FF,$8A,$2B,$FC,$BF,$08,$C3
	.dc.b	$FE,$62,$6B,$FE,$6C,$FA,$77,$D4,$7F,$80,$A3,$FC,$CF,$18,$F7,$F9
	.dc.b	$AF,$A0,$78,$D8,$BE,$17,$FE,$76,$F4,$FF,$A4,$BF,$7B,$2F,$C3,$C6
	.dc.b	$23,$FC,$20,$DD,$FD,$DF,$51,$FE,$41,$AD,$FA,$7F,$E4,$D6,$DF,$33
	.dc.b	$FA,$3D,$F1,$8F,$C9,$D7,$19,$CF,$C2,$71,$69,$C4,$FE,$C6,$E5,$FE
	.dc.b	$CB,$F3,$F6,$FF,$C5,$B1,$B6,$BE,$89,$D4,$F1,$FA,$87,$CD,$F1,$F8
	.dc.b	$C9,$7D,$E7,$2D,$9C,$BF,$0A,$79,$44,$8F,$04,$6B,$4F,$C1,$57,$29
	.dc.b	$FF,$1C,$36,$3F,$0D,$8F,$FA,$47,$3F,$9F,$7C,$EF,$78,$73,$9C,$C7
	.dc.b	$8C,$87,$8C,$F0,$98,$BF,$0C,$F1,$8F,$A3,$5D,$67,$31,$C0,$E4,$78
	.dc.b	$0E,$55,$E0,$73,$1C,$1D,$18,$8D,$46,$48,$E0,$91,$FE,$08,$C3,$E1
	.dc.b	$47,$F1,$46,$C7,$98,$F2,$4F,$27,$85,$FA,$DF,$F1,$2D,$FA,$EF,$94
	.dc.b	$E7,$0F,$99,$5C,$0F,$59,$D4,$F4,$19,$3D,$77,$F5,$2F,$B6,$1E,$49
	.dc.b	$36,$BA,$C7,$88,$EA,$6A,$FC,$55,$E5,$8E,$D7,$D5,$F6,$E1,$E8,$78
	.dc.b	$DF,$43,$AE,$1E,$13,$C1,$7E,$4C,$3A,$57,$84,$F5,$9E,$13,$9C,$6D
	.dc.b	$A7,$80,$F1,$DF,$85,$E3,$7E,$0A,$EA,$F8,$EB,$BD,$F8,$AB,$91,$DC
	.dc.b	$F9,$37,$F7,$9F,$26,$1F,$65,$D3,$F0,$5F,$3C,$FB,$70,$F8,$F1,$F2
	.dc.b	$A3,$B3,$A8,$F1,$DF,$AE,$3B,$9F,$6E,$BA,$1D,$EB,$8B,$8B,$F7,$3E
	.dc.b	$3C,$3D,$95,$DF,$0F,$0B,$D0,$F3,$D7,$AE,$7E,$24,$7C,$A3,$C8,$5A
	.dc.b	$EC,$FC,$C2,$1F,$0A,$7E,$25,$FB,$E3,$E4,$D7,$47,$C0,$F9,$1F,$14
	.dc.b	$7A,$A7,$8D,$FC,$29,$FA,$33,$CD,$FB,$38,$7D,$69,$F5,$E1,$F0,$DB
	.dc.b	$7F,$08,$95,$F2,$5E,$AF,$AD,$AF,$92,$BE,$85,$7F,$5D,$FA,$87,$47
	.dc.b	$F3,$36,$D7,$F3,$76,$BF,$ED,$77,$FB,$5B,$98,$E0,$47,$87,$3F,$3F
	.dc.b	$86,$E9,$EA,$C9,$E9,$19,$7B,$46,$5E,$D1,$FD,$17,$EA,$1F,$51,$E5
	.dc.b	$EB,$95,$FD,$2E,$72,$D9,$CB,$E7,$9E,$51,$23,$44,$3D,$9F,$51,$FD
	.dc.b	$4F,$EA,$1E,$85,$F4,$65,$7F,$59,$9C,$B6,$72,$FF,$BE,$E3,$FC,$D0
	.dc.b	$E4,$E1,$F0,$5D,$48,$F3,$FA,$7A,$F1,$8B,$C3,$FE,$11,$A5,$7F,$67
	.dc.b	$ED,$69,$E7,$E4,$98,$FF,$6D,$9C,$B6,$72,$F9,$67,$94,$48,$D1,$06
	.dc.b	$1F,$E2,$1F,$A8,$7D,$07,$F7,$77,$FB,$9C,$67,$AB,$06,$7A,$E5,$FF
	.dc.b	$7F,$A4,$EA,$31,$3C,$FD,$79,$1B,$7B,$74,$E2,$3C,$79,$FF,$98,$30
	.dc.b	$FF,$60,$46,$CE,$D3,$F1,$7C,$CF,$F0,$66,$DF,$96,$9E,$27,$FB,$79
	.dc.b	$C6,$7F,$E3,$DD,$95,$C6,$BF,$C2,$D5,$AC,$7F,$CD,$ED,$8F,$F7,$3D
	.dc.b	$70,$7E,$85,$71,$7B,$9F,$E3,$77,$C6,$79,$39,$A4,$76,$7C,$FC,$F8
	.dc.b	$8E,$36,$9C,$63,$FE,$53,$8F,$F7,$74,$F9,$BE,$A3,$DF,$C3,$23,$04
	.dc.b	$39,$B9,$7A,$C7,$F8,$C7,$F2,$A7,$C6,$7F,$8C,$FF,$1F,$FC,$CD,$7A
	.dc.b	$9B,$8F,$F9,$8E,$B7,$3F,$E3,$B6,$BB,$E8,$E4,$75,$B9,$1B,$6B,$D7
	.dc.b	$E3,$7F,$C8,$5F,$6E,$F9,$3F,$F2,$27,$DB,$FF,$A7,$2B,$FC,$90,$D7
	.dc.b	$E6,$CE,$DF,$F2,$67,$DB,$FE,$81,$FA,$33,$FE,$67,$78,$F1,$FF,$93
	.dc.b	$E7,$5A,$7F,$9F,$7D,$76,$FF,$E5,$78,$7D,$6F,$D1,$C3,$FC,$C1,$18
	.dc.b	$E3,$FF,$2C,$5F,$09,$F0,$F8,$9F,$E5,$BF,$B7,$FD,$2F,$47,$FA,$D9
	.dc.b	$B2,$19,$FF,$2C,$C7,$FD,$1D,$F6,$EF,$2B,$97,$6C,$BF,$F3,$44,$E5
	.dc.b	$FF,$35,$66,$F9,$3F,$46,$DF,$84,$9F,$CF,$E3,$EA,$EB,$F3,$D7,$F9
	.dc.b	$BF,$EB,$3F,$85,$D7,$F9,$CF,$ED,$9E,$F7,$F9,$CD,$B1,$FE,$77,$9D
	.dc.b	$BF,$9C,$5E,$C6,$FE,$7A,$4E,$E3,$E2,$C7,$93,$EB,$E4,$FA,$23,$FD
	.dc.b	$01,$F6,$CF,$88,$FF,$41,$47,$E3,$B9,$7F,$DB,$AD,$AF,$91,$5B,$7F
	.dc.b	$D1,$5F,$6B,$FC,$09,$FD,$7C,$BC,$CE,$67,$FA,$36,$75,$C1,$F7,$FE
	.dc.b	$9F,$E3,$FC,$54,$C3,$E5,$C7,$FA,$4F,$EB,$9F,$0D,$7E,$82,$57,$FA
	.dc.b	$66,$72,$FF,$A6,$FE,$BB,$E9,$37,$97,$FD,$39,$18,$8F,$E0,$6B,$E2
	.dc.b	$72,$48,$E4,$FF,$A8,$FE,$B5,$F0,$DC,$BD,$A9,$5D,$DF,$F5,$47,$D7
	.dc.b	$7E,$C3,$7F,$F6,$B3,$79,$3F,$1D,$48,$C3,$F9,$16,$AC,$1F,$11,$E1
	.dc.b	$3A,$F9,$92,$7F,$D7,$13,$A7,$FD,$75,$F5,$CF,$37,$D1,$52,$73,$1F
	.dc.b	$EB,$D9,$D4,$60,$87,$F2,$6A,$7D,$76,$EA,$4E,$B3,$FF,$79,$56,$A3
	.dc.b	$C3,$C7,$F1,$FD,$59,$18,$21,$FC,$CB,$1F,$AE,$D9,$F9,$DF,$47,$FE
	.dc.b	$D1,$FA,$E7,$5D,$25,$7F,$B5,$E1,$D1,$88,$FE,$7B,$1C,$1D,$4F,$39
	.dc.b	$FF,$6E,$62,$F9,$8A,$4A,$FF,$71,$7D,$77,$4D,$7F,$82,$7D,$62,$FE
	.dc.b	$C9,$C6,$BF,$DD,$0D,$7F,$A6,$B6,$5C,$9F,$80,$BE,$94,$BE,$8F,$F1
	.dc.b	$18,$CC,$62,$3C,$18,$7E,$22,$31,$87,$77,$55,$C2,$3B,$63,$FE,$06
	.dc.b	$BE,$D1,$E8,$BE,$4D,$8F,$6B,$C3,$1B,$1E,$3A,$E5,$87,$82,$F8,$CF
	.dc.b	$F2,$7B,$F1,$1C,$5E,$E7,$1B,$E6,$FA,$8B,$1C,$1D,$1C,$06,$1E,$B7
	.dc.b	$B5,$78,$1E,$46,$E3,$4F,$0C,$3E,$03,$FA,$E3,$17,$5E,$E6,$67,$E9
	.dc.b	$B7,$EB,$9F,$5D,$B8,$F1,$C6,$1E,$52,$8C,$47,$19,$E0,$FF,$4B,$FA
	.dc.b	$0A,$9A,$47,$6F,$C6,$BF,$CF,$E9,$F2,$27,$91,$F1,$63,$2F,$C7,$33
	.dc.b	$A7,$E1,$E3,$2F,$E2,$63,$3A,$7F,$10,$9C,$BF,$0E,$FC,$1A,$FE,$5A
	.dc.b	$BA,$1E,$73,$C0,$E6,$FA,$46,$5F,$04,$66,$BA,$4F,$0C,$3D,$4E,$E3
	.dc.b	$C5,$3B,$8F,$AD,$7E,$2A,$91,$FE,$36,$C3,$DE,$79,$5F,$E5,$33,$A8
	.dc.b	$C1,$A9,$F7,$7C,$8F,$AE,$33,$76,$72,$7F,$7F,$8E,$9F,$5D,$FC,$0E
	.dc.b	$30,$7F,$36,$71,$87,$F9,$21,$FF,$03,$46,$A3,$FC,$11,$1A,$9E,$7B
	.dc.b	$F7,$9F,$F3,$BF,$5F,$95,$DC,$66,$7F,$A5,$E1,$B9,$5F,$F9,$03,$0F
	.dc.b	$F0,$EF,$D7,$3E,$67,$53,$9A,$FF,$05,$CE,$6B,$C9,$1C,$BA,$BB,$4E
	.dc.b	$63,$D5,$18,$BE,$7F,$5D,$2F,$FC,$57,$F5,$DF,$D4,$1E,$78,$A7,$58
	.dc.b	$7F,$86,$A9,$F0,$B8,$D3,$E8,$46,$30,$E3,$7E,$F8,$C5,$FC,$B8,$C5
	.dc.b	$FC,$1F,$95,$FE,$FB,$7E,$A6,$D7,$F9,$E3,$EB,$BF,$CA,$53,$9E,$BE
	.dc.b	$07,$C9,$FA,$EF,$F3,$A4,$64,$F8,$6E,$2B,$97,$91,$F3,$EC,$F1,$F7
	.dc.b	$30,$FD,$3F,$AE,$77,$D3,$B7,$97,$72,$F9,$F3,$F4,$F1,$FE,$20,$D7
	.dc.b	$FB,$FB,$5F,$F2,$73,$5F,$EE,$DE,$57,$E7,$5F,$F0,$2F,$89,$FE,$49
	.dc.b	$AF,$E2,$4D,$8F,$E3,$4D,$7F,$8F,$47,$FB,$40,$FE,$30,$57,$F9,$03
	.dc.b	$5D,$FC,$37,$77,$F3,$7E,$AD,$7F,$AB,$B6,$7F,$CA,$27,$47,$5F,$5B
	.dc.b	$FE,$F2,$8F,$E4,$CD,$9F,$E5,$CD,$7F,$99,$35,$FF,$10,$FD,$5F,$F9
	.dc.b	$AE,$7F,$97,$36,$3F,$9B,$35,$FE,$74,$D7,$8F,$CB,$F9,$E0,$D2,$4F
	.dc.b	$F3,$9F,$AA,$CB,$63,$DD,$18,$AF,$E8,$8D,$7F,$89,$FA,$FF,$0F,$C2
	.dc.b	$FF,$35,$DF,$F4,$36,$CF,$F2,$99,$F9,$1F,$54,$F6,$E9,$2F,$FA,$54
	.dc.b	$65,$FE,$99,$E1,$FF,$0C,$BF,$EA,$CD,$7F,$C8,$AD,$7F,$D2,$8D,$5F
	.dc.b	$36,$F8,$0F,$B0,$DA,$FE,$B0,$D9,$FE,$C0,$D7,$FB,$22,$FA,$7E,$24
	.dc.b	$9E,$2F,$E0,$FD,$57,$C0,$52,$4F,$F6,$58,$CB,$FD,$9F,$E9,$FC,$1E
	.dc.b	$1F,$F3,$76,$C7,$F2,$FF,$A0,$7B,$FF,$C7,$E7,$39,$FF,$AC,$5F,$33
	.dc.b	$FE,$66,$6B,$9D,$3F,$2D,$18,$83,$CB,$07,$D1,$3E,$9F,$B5,$FF,$2F
	.dc.b	$8C,$65,$FE,$C8,$FA,$7F,$CC,$5F,$F4,$76,$C7,$F4,$55,$F2,$1F,$CD
	.dc.b	$DC,$E4,$A4,$61,$BF,$33,$BA,$E0,$FF,$3F,$D5,$E0,$5F,$84,$C3,$72
	.dc.b	$FF,$6D,$CE,$9B,$3A,$69,$FC,$FB,$6F,$E7,$98,$FC,$FF,$37,$CD,$F4
	.dc.b	$FF,$35,$95,$F5,$5F,$92,$47,$07,$FB,$AB,$E9,$FF,$D8,$0F,$58,$FE
	.dc.b	$AD,$18,$DB,$FD,$B3,$97,$F0,$46,$F1,$FB,$7C,$0F,$73,$70,$F5,$B6
	.dc.b	$77,$1F,$47,$97,$E6,$1F,$37,$CA,$B1,$92,$4F,$FB,$A3,$8B,$FA,$1E
	.dc.b	$BF,$5E,$FF,$B4,$7D,$3B,$E9,$E9,$F3,$FE,$66,$E3,$FD,$92,$74,$FF
	.dc.b	$8F,$5B,$8F,$F9,$15,$AF,$F9,$25,$AF,$FA,$E1,$DE,$1F,$E0,$DE,$6F
	.dc.b	$EC,$D3,$F8,$3A,$D6,$1F,$8F,$F8,$2F,$C4,$78,$12,$7F,$C1,$18,$7F
	.dc.b	$B2,$BC,$FC,$61,$ED,$AB,$3F,$E1,$9E,$F5,$62,$FF,$55,$F2,$3F,$8B
	.dc.b	$3F,$19,$F0,$B6,$3F,$13,$17,$1F,$AF,$37,$0F,$F1,$57,$DF,$F4,$BF
	.dc.b	$F2,$F3,$6B,$FC,$CC,$D7,$FC,$D4,$D7,$FD,$B3,$F4,$EF,$A7,$F2,$C7
	.dc.b	$F9,$B6,$31,$97,$F9,$A5,$7C,$B7,$F1,$89,$D5,$70,$7F,$CE,$DF,$51
	.dc.b	$FE,$BA,$7F,$CE,$0D,$7F,$CF,$0D,$77,$F0,$BF,$E6,$EF,$94,$FE,$83
	.dc.b	$1F,$E2,$06,$BF,$4C,$07,$3A,$7F,$92,$FE,$A3,$FC,$93,$AF,$FB,$D2
	.dc.b	$72,$DF,$3B,$FC,$D0,$DB,$FF,$16,$D7,$07,$AD,$7F,$88,$9F,$17,$B5
	.dc.b	$C1,$CB,$5F,$17,$D0,$59,$E1,$1E,$73,$CD,$39,$35,$3C,$7E,$A1,$F5
	.dc.b	$1C,$7E,$1A,$5F,$F9,$02,$72,$D9,$CB,$D7,$0F,$F4,$36,$5F,$12,$15
	.dc.b	$C1,$F1,$37,$4F,$F3,$85,$70,$9F,$7E,$1E,$3C,$24,$7D,$18,$EF,$5F
	.dc.b	$E7,$7B,$E5,$87,$76,$D7,$23,$81,$E7,$21,$E7,$AE,$73,$C6,$FB,$CE
	.dc.b	$4E,$27,$A1,$CD,$FA,$2B,$8C,$FE,$0A,$FC,$2F,$C0,$AC,$BF,$85,$9E
	.dc.b	$37,$DE,$3A,$DF,$37,$CA,$FB,$6B,$8E,$1F,$16,$7A,$1A,$AF,$C6,$5F
	.dc.b	$6C,$3C,$27,$8A,$B8,$1C,$5C,$DF,$49,$F1,$DF,$A6,$30,$F5,$30,$79
	.dc.b	$AF,$BB,$93,$C6,$7C,$72,$61,$E9,$75,$87,$8A,$FB,$C7,$09,$F8,$0F
	.dc.b	$0A,$F8,$8F,$CF,$3C,$07,$DB,$30,$48,$E2,$48,$E2,$F8,$63,$A1,$C5
	.dc.b	$C1,$23,$D7,$1A,$3C,$8F,$23,$31,$D6,$3E,$19,$E2,$3F,$3C,$7C,$22
	.dc.b	$E1,$F9,$F6,$E1,$FE,$F8,$8C,$47,$B8,$C9,$C8,$EA,$60,$F7,$57,$8D
	.dc.b	$FE,$07,$ED,$7C,$A9,$FB,$71,$FC,$56,$31,$5F,$3B,$1F,$F4,$6C,$FD
	.dc.b	$7A,$DB,$FB,$7A,$DB,$F5,$9A,$F8,$0E,$E7,$1A,$E9,$1A,$C3,$E5,$4E
	.dc.b	$8E,$07,$CA,$3E,$3B,$FB,$E8,$E1,$87,$AE,$7F,$94,$46,$1D,$C7,$F2
	.dc.b	$78,$C4,$73,$36,$78,$CF,$41,$4F,$0E,$1F,$01,$F1,$E1,$F0,$2F,$C5
	.dc.b	$1B,$7D,$2F,$43,$78,$7B,$9D,$61,$E7,$BE,$95,$C2,$FA,$4F,$9E,$B6
	.dc.b	$7A,$8F,$89,$3B,$BE,$91,$D8,$F4,$94,$F4,$E1,$F0,$2F,$F7,$4F,$F3
	.dc.b	$DF,$A8,$73,$D2,$57,$FB,$D2,$1C,$F8,$8F,$E8,$11,$C1,$D4,$F3,$9E
	.dc.b	$9E,$EF,$5F,$52,$57,$38,$CB,$D6,$32,$F4,$8F,$EA,$1F,$50,$FA,$8F
	.dc.b	$37,$B7,$2B,$FB,$2E,$CD,$FF,$A2,$5E,$E7,$31,$8B,$FD,$14,$7F,$6C
	.dc.b	$9D,$3F,$A5,$6B,$FB,$06,$BC,$4E,$67,$18,$F2,$9F,$27,$EA,$3F,$B3
	.dc.b	$7D,$43,$D2,$BE,$9C,$BE,$22,$74,$E1,$AF,$F6,$B7,$FC,$F2,$E6,$3A
	.dc.b	$CF,$6F,$98,$93,$C5,$F0,$FD,$36,$23,$94,$5D,$3E,$34,$7F,$83,$30
	.dc.b	$57,$F7,$DE,$F6,$DE,$AD,$A5,$FA,$27,$2D,$9F,$96,$FF,$89,$9B,$1A
	.dc.b	$8F,$E2,$0D,$7F,$8D,$B9,$C3,$C7,$1C,$4E,$87,$C6,$FA,$77,$D4,$7F
	.dc.b	$84,$5F,$F0,$7C,$F0,$5D,$79,$35,$D9,$2F,$94,$65,$E9,$1C,$FD,$89
	.dc.b	$1D,$DE,$DD,$B8,$8F,$F0,$E5,$FF,$B5,$DD,$1E,$58,$F9,$1E,$67,$F8
	.dc.b	$97,$6F,$D2,$CF,$DE,$6B,$FE,$25,$D1,$3C,$23,$F2,$71,$86,$C7,$CA
	.dc.b	$23,$F4,$5E,$0F,$F8,$BE,$78,$37,$9B,$C4,$47,$C5,$CF,$5D,$1E,$58
	.dc.b	$F4,$47,$A7,$E8,$FD,$5C,$12,$BD,$51,$97,$D9,$19,$7C,$51,$FE,$45
	.dc.b	$FA,$77,$C6,$FF,$91,$FE,$A3,$F3,$35,$FE,$45,$6F,$17,$E0,$2B,$73
	.dc.b	$FA,$46,$BF,$D5,$DA,$FF,$62,$6B,$FE,$41,$6B,$FE,$66,$78,$C7,$5A
	.dc.b	$F2,$1E,$3F,$86,$FF,$2D,$7D,$3B,$EA,$3F,$32,$DC,$3F,$82,$B5,$FE
	.dc.b	$D0,$D7,$FB,$9F,$D3,$7E,$6D,$FF,$29,$35,$FF,$3B,$B5,$37,$D5,$69
	.dc.b	$FA,$86,$CF,$F1,$46,$BF,$DD,$DA,$F8,$3C,$2F,$D7,$D7,$F2,$F8,$FE
	.dc.b	$F8,$D7,$47,$F1,$EF,$A6,$FF,$44,$37,$C8,$FD,$BC,$62,$3F,$C0,$8D
	.dc.b	$7F,$C1,$F6,$F8,$BF,$31,$23,$CC,$FF,$00,$FA,$4F,$E1,$31,$FC,$B1
	.dc.b	$AF,$F8,$3D,$AF,$F8,$69,$AF,$F9,$E5,$7A,$9C,$C9,$1C,$DF,$8D,$F2
	.dc.b	$B2,$3E,$33,$5F,$F0,$E7,$D2,$3E,$13,$F8,$9B,$63,$F9,$FB,$5D,$C7
	.dc.b	$F1,$8F,$A4,$FD,$E4,$7F,$89,$5A,$FF,$8C,$1A,$FF,$A3,$DA,$FF,$AD
	.dc.b	$32,$FF,$5E,$3F,$E9,$36,$B9,$F4,$BE,$F1,$FC,$9B,$E9,$1E,$FD,$24
	.dc.b	$FF,$54,$8C,$BF,$D5,$BE,$93,$E9,$3F,$D2,$1A,$FF,$AB,$DB,$D1,$FC
	.dc.b	$A9,$7C,$0F,$FA,$DF,$E8,$9E,$EF,$FA,$E6,$73,$7C,$EB,$53,$FE,$BD
	.dc.b	$FA,$4F,$D4,$F9,$7F,$EA,$DB,$FE,$33,$8F,$5F,$4F,$5D,$D3,$2F,$EB
	.dc.b	$78,$70,$7F,$DA,$ED,$59,$BD,$FF,$21,$C6,$3F,$D9,$13,$B9,$D4,$7F
	.dc.b	$66,$FA,$37,$81,$EB,$C2,$4E,$63,$FD,$A9,$3A,$8C,$10,$FF,$6A,$53
	.dc.b	$E9,$37,$52,$75,$9F,$F5,$2A,$D4,$78,$78,$7E,$57,$95,$23,$04,$3F
	.dc.b	$DC,$18,$FD,$86,$CF,$D1,$F6,$3F,$DD,$1F,$48,$EA,$A4,$AF,$F0,$1C
	.dc.b	$39,$B1,$1F,$EE,$B8,$E0,$EA,$79,$CF,$FB,$CF,$17,$CF,$D2,$57,$F8
	.dc.b	$3B,$E9,$3A,$6A,$87,$D1,$2F,$FD,$D4,$72,$8F,$F0,$AB,$5F,$F7,$E3
	.dc.b	$67,$FB,$BD,$74,$9E,$73,$FE,$F3,$8C,$57,$59,$F0,$4F,$38,$C4,$78
	.dc.b	$6F,$FD,$F5,$18,$AE,$35,$CF,$06,$1E,$9A,$E3,$3C,$CF,$4D,$F2,$78
	.dc.b	$4E,$DB,$87,$98,$F1,$CE,$DA,$6C,$F8,$D8,$78,$9B,$87,$89,$E4,$7A
	.dc.b	$4E,$05,$33,$1C,$53,$6B,$D1,$1E,$03,$73,$D7,$0F,$84,$F7,$8F,$F6
	.dc.b	$0D,$7C,$F6,$C6,$5F,$69,$AF,$A4,$7D,$27,$F9,$7A,$91,$89,$FF,$61
	.dc.b	$46,$27,$53,$99,$FF,$32,$FC,$C7,$F8,$F1,$34,$8F,$F9,$01,$F8,$39
	.dc.b	$FE,$77,$A7,$E1,$CF,$E3,$E7,$3A,$FF,$94,$33,$FC,$0C,$E4,$F9,$77
	.dc.b	$C5,$F9,$A6,$63,$F0,$F1,$97,$8C,$79,$0E,$33,$D7,$4F,$F6,$F5,$F8
	.dc.b	$30,$F1,$D6,$EF,$9D,$EE,$3F,$DD,$B5,$B7,$C2,$6E,$FF,$50,$FE,$4A
	.dc.b	$78,$38,$9E,$15,$D2,$FF,$CF,$B8,$7A,$CF,$F2,$ED,$7A,$F0,$E7,$19
	.dc.b	$9B,$87,$F9,$D2,$3F,$CC,$63,$E9,$31,$3A,$3A,$B8,$BF,$DE,$7D,$1F
	.dc.b	$F0,$29,$FF,$46,$B6,$3F,$53,$1E,$A8,$FE,$04,$EB,$7E,$52,$9E,$17
	.dc.b	$E3,$EC,$F3,$4F,$6A,$F0,$4F,$F4,$CC,$78,$DF,$F9,$03,$0F,$F5,$2C
	.dc.b	$BE,$91,$F2,$7A,$EA,$E3,$18,$AF,$3C,$62,$75,$1C,$A7,$59,$7A,$8F
	.dc.b	$7F,$D1,$F8,$0F,$1F,$D2,$7F,$4D,$AF,$83,$19,$7C,$70,$F6,$F1,$5C
	.dc.b	$F2,$4E,$6B,$89,$C2,$B7,$3F,$56,$75,$5E,$31,$F1,$BF,$DF,$2F,$FD
	.dc.b	$7A,$D8,$D7,$D1,$FF,$93,$E7,$3D,$7F,$43,$E9,$1E,$CD,$3E,$9E,$1C
	.dc.b	$72,$FF,$52,$CE,$AF,$FD,$93,$83,$E9,$3E,$CB,$63,$FD,$9B,$3B,$CB
	.dc.b	$C3,$5F,$7B,$E9,$1D,$F4,$F1,$7C,$E5,$C3,$FD,$AB,$3F,$6E,$FD,$CF
	.dc.b	$F8,$2A,$33,$C7,$CF,$9F,$F1,$3E,$1F,$F9,$1E,$7F,$DB,$BE,$6B,$C9
	.dc.b	$CF,$A3,$F8,$E8,$C4,$FF,$B8,$8F,$F6,$79,$F8,$8F,$2D,$F1,$BF,$94
	.dc.b	$8C,$E9,$F5,$E7,$31,$C7,$E6,$3F,$27,$F5,$FF,$ED,$A8,$C1,$F9,$27
	.dc.b	$15,$F5,$FE,$97,$FD,$E9,$3F,$96,$FA,$F7,$C6,$FE,$8A,$73,$5F,$85
	.dc.b	$9C,$D7,$FB,$E3,$E9,$9E,$D7,$E8,$BE,$39,$F1,$FF,$AA,$9C,$D7,$E8
	.dc.b	$A7,$2F,$1F,$94,$FD,$4F,$D7,$74,$CB,$F1,$56,$9C,$E9,$FB,$4F,$A7
	.dc.b	$C6,$1F,$AC,$C1,$BE,$6F,$9B,$FB,$AF,$37,$FC,$D7,$5F,$B2,$FA,$5F
	.dc.b	$93,$E0,$7F,$BB,$A7,$3A,$7E,$86,$73,$5F,$C7,$BE,$9D,$F1,$5F,$BE
	.dc.b	$9E,$F3,$FE,$09,$C3,$FD,$EF,$5A,$8E,$D8,$FF,$0A,$EB,$FE,$13,$1B
	.dc.b	$FA,$EF,$F1,$83,$C4,$FE,$57,$1A,$CB,$FB,$EC,$E9,$FD,$67,$D4,$3E
	.dc.b	$47,$F8,$7F,$D6,$FF,$BA,$30,$E5,$5F,$E1,$BA,$D4,$7F,$10,$9D,$4F
	.dc.b	$2D,$3F,$8B,$FC,$87,$5D,$57,$FC,$F3,$05,$6B,$DD,$FE,$47,$D8,$F8
	.dc.b	$2F,$E3,$FF,$88,$70,$C3,$C3,$9F,$F9,$BA,$73,$1F,$CA,$BE,$A3,$C7
	.dc.b	$3F,$C1,$5B,$9F,$F3,$DF,$A6,$7D,$56,$F8,$26,$1F,$CA,$3E,$95,$F5
	.dc.b	$58,$CB,$F8,$B4,$62,$7C,$39,$74,$7E,$A4,$66,$7C,$17,$FC,$DA,$31
	.dc.b	$97,$F4,$DF,$A8,$FC,$EB,$7E,$7B,$F9,$DC,$3C,$0B,$13,$1F,$E7,$73
	.dc.b	$C1,$FC,$13,$B7,$31,$FD,$13,$57,$9B,$7C,$8C,$36,$BF,$A2,$CE,$9B
	.dc.b	$3A,$69,$EE,$F6,$7E,$61,$E6,$FA,$AB,$19,$3F,$A7,$FD,$57,$FB,$05
	.dc.b	$C6,$FF,$CF,$63,$0F,$D3,$AF,$EB,$31,$8D,$7F,$07,$3F,$E6,$5E,$0F
	.dc.b	$E9,$F1,$FD,$CE,$75,$7F,$A1,$CB,$F4,$77,$FD,$25,$FF,$57,$C7,$B3
	.dc.b	$B7,$E1,$73,$7C,$36,$C6,$4F,$83,$1F,$43,$96,$47,$CE,$F5,$EE,$1F
	.dc.b	$D5,$78,$7F,$25,$3F,$4A,$BE,$BD,$6B,$3F,$4D,$BE,$A9,$F5,$5F,$DF
	.dc.b	$1B,$87,$D0,$ED,$C4,$C8,$FC,$87,$C6,$5B,$FC,$7B,$70,$FF,$0A,$73
	.dc.b	$FF,$6D,$9F,$EC,$18,$F4,$77,$7F,$86,$F9,$1F,$C5,$9F,$E0,$6F,$0B
	.dc.b	$FB,$2C,$64,$93,$ED,$9D,$7B,$5C,$27,$F0,$BF,$2B,$F8,$B3,$F1,$9F
	.dc.b	$1F,$63,$24,$BF,$A5,$19,$7F,$BA,$7D,$4B,$EA,$7F,$BA,$C7,$F8,$E8
	.dc.b	$FF,$51,$46,$70,$C4,$63,$E7,$DE,$36,$21,$5B,$77,$F7,$5F,$EC,$66
	.dc.b	$F8,$DB,$51,$97,$C0,$D7,$FC,$97,$F5,$5F,$E7,$47,$B3,$5F,$1B,$5F
	.dc.b	$43,$5F,$F2,$9F,$A9,$FD,$1D,$FE,$26,$D7,$FC,$28,$DE,$3F,$C2,$FD
	.dc.b	$57,$F9,$33,$7E,$3E,$87,$F9,$49,$B7,$CA,$30,$FF,$B1,$E7,$83,$FE
	.dc.b	$C9,$BF,$F1,$53,$E0,$F8,$5D,$BF,$D9,$1B,$0F,$A4,$B1,$27,$81,$E8
	.dc.b	$8E,$3F,$55,$FE,$00,$8F,$E5,$AD,$7F,$C5,$CD,$CB,$F3,$F7,$FE,$1A
	.dc.b	$9D,$FB,$9C,$88,$FC,$4F,$82,$AF,$28,$B5,$CA,$6E,$9F,$E7,$EC,$3B
	.dc.b	$36,$BD,$73,$AB,$FF,$75,$56,$9B,$3A,$7C,$64,$3C,$07,$8C,$C5,$FE
	.dc.b	$E2,$79,$41,$B8,$7B,$63,$89,$27,$85,$77,$8C,$61,$E7,$7A,$BF,$86
	.dc.b	$C3,$D1,$1D,$66,$DF,$5B,$E7,$3C,$0F,$59,$E7,$3A,$98,$29,$FE,$22
	.dc.b	$AE,$B8,$7A,$8F,$4B,$8B,$EE,$D9,$EC,$71,$2C,$F1,$9F,$65,$79,$27
	.dc.b	$BC,$F4,$77,$3D,$1A,$74,$BF,$66,$1F,$0E,$38,$CF,$48,$C5,$F5,$23
	.dc.b	$F1,$E7,$99,$EE,$AE,$6F,$C1,$9D,$B6,$78,$3D,$8F,$16,$1E,$F7,$93
	.dc.b	$67,$8C,$7C,$13,$81,$A3,$C2,$EA,$BC,$31,$F6,$4C,$D7,$82,$79,$E1
	.dc.b	$F2,$67,$EC,$46,$30,$F8,$D1,$DA,$3A,$61,$E3,$BF,$86,$73,$2C,$FB
	.dc.b	$48,$79,$E7,$A3,$EE,$3C,$66,$B0,$F3,$E1,$E3,$8F,$61,$E1,$AE,$AF
	.dc.b	$36,$C6,$EB,$F6,$93,$D0,$F8,$07,$F1,$CB,$F4,$3D,$DA,$EA,$38,$BD
	.dc.b	$CD,$1C,$E7,$F1,$4E,$4E,$15,$F0,$5F,$7C,$78,$E3,$11,$A3,$38,$7E
	.dc.b	$2D,$A7,$C3,$3A,$4F,$6A,$F8,$F1,$F0,$9F,$EE,$FF,$54,$E7,$FE,$52
	.dc.b	$D9,$FF,$26,$B5,$FF,$67,$2E,$7D,$CC,$9D,$BC,$A3,$9F,$C3,$F4,$F5
	.dc.b	$A5,$74,$8C,$BD,$A3,$2F,$68,$FE,$77,$F5,$5F,$E9,$C8,$FF,$2F,$B5
	.dc.b	$FF,$70,$37,$1F,$5F,$B7,$C3,$DD,$95,$C1,$F6,$C6,$23,$8C,$59,$E3
	.dc.b	$1F,$E6,$CF,$A9,$7A,$3F,$D3,$1F,$F2,$A4,$67,$D2,$C4,$C8,$C1,$0F
	.dc.b	$F7,$A4,$F2,$7A,$7D,$57,$32,$3D,$FE,$A2,$BC,$22,$E1,$FE,$2C,$A5
	.dc.b	$7F,$5E,$EF,$7D,$47,$F7,$F6,$E1,$FE,$07,$6B,$FE,$0C,$6B,$FE,$11
	.dc.b	$73,$7F,$E3,$86,$F2,$A6,$46,$1C,$7D,$4B,$EA,$BC,$58,$49,$FF,$49
	.dc.b	$D6,$D5,$6D,$FF,$4B,$BF,$DE,$16,$1F,$E4,$88,$FE,$ED,$1A,$EA,$FE
	.dc.b	$F7,$1B,$9D,$75,$E2,$65,$74,$8D,$FB,$12,$3B,$BC,$BE,$5B,$11,$E0
	.dc.b	$D3,$DC,$F9,$A7,$38,$78,$EB,$E7,$79,$BF,$E0,$CD,$BF,$41,$7F,$EB
	.dc.b	$06,$BC,$CF,$8F,$1C,$1F,$86,$D7,$47,$DA,$71,$5F,$45,$B8,$79,$29
	.dc.b	$1E,$1E,$69,$7E,$17,$C9,$CF,$88,$EB,$49,$C6,$1F,$E7,$48,$FF,$75
	.dc.b	$CF,$A7,$E9,$BD,$9C,$12,$30,$43,$9B,$97,$AC,$7F,$8B,$BE,$A5,$F1
	.dc.b	$BF,$E3,$0F,$AA,$FD,$2D,$7B,$9B,$8F,$FA,$12,$B7,$3F,$ED,$86,$BF
	.dc.b	$11,$B1,$BA,$E9,$F1,$DD,$D9,$5C,$5C,$BF,$EE,$2F,$A9,$7D,$57,$F9
	.dc.b	$09,$B1,$98,$FF,$22,$FD,$7F,$F0,$7E,$1F,$C0,$FD,$5F,$E5,$C9,$B6
	.dc.b	$BE,$EB,$FC,$97,$EC,$B8,$29,$32,$FF,$80,$3B,$3E,$BD,$F0,$7F,$E5
	.dc.b	$2C,$BF,$94,$4F,$FA,$91,$B9,$F8,$3C,$6F,$F2,$BF,$D8,$3D,$AF,$F2
	.dc.b	$CF,$D7,$3C,$EA,$4D,$3F,$DF,$F5,$FD,$CB,$67,$CA,$F9,$89,$3E,$4D
	.dc.b	$9F,$03,$FE,$64,$FA,$EF,$E1,$33,$FB,$DD,$FF,$D6,$F1,$88,$FC,$AB
	.dc.b	$63,$FD,$77,$F6,$1E,$C9,$E1,$BF,$EA,$29,$EE,$75,$24,$E8,$FE,$15
	.dc.b	$49,$F9,$7F,$43,$27,$43,$44,$7F,$A0,$31,$F9,$9F,$09,$FE,$7C,$FA
	.dc.b	$FF,$15,$FE,$A9,$B3,$FE,$D0,$6E,$7F,$BC,$BF,$E6,$15,$FB,$0F,$AF
	.dc.b	$78,$DF,$E8,$66,$C7,$89,$AF,$ED,$A5,$F6,$FF,$82,$FA,$B9,$3E,$8D
	.dc.b	$11,$E7,$E1,$FD,$BC,$FB,$3E,$C1,$F1,$1F,$E9,$46,$C6,$63,$FD,$2F
	.dc.b	$F3,$DF,$6A,$3E,$1A,$E0,$CF,$2C,$9C,$C7,$FA,$7B,$EB,$DF,$0F,$CA
	.dc.b	$24,$FC,$C9,$CB,$F3,$BE,$C3,$F5,$BC,$5F,$C8,$31,$FE,$33,$C9,$FE
	.dc.b	$F8,$6C,$7F,$24,$9D,$D7,$F4,$36,$A6,$98,$BF,$CD,$47,$87,$B3,$FD
	.dc.b	$67,$3A,$F1,$BD,$11,$FE,$B6,$FB,$07,$91,$EB,$99,$7F,$84,$9C,$B6
	.dc.b	$78,$11,$F7,$7C,$2D,$7F,$96,$CF,$06,$CF,$C7,$9F,$07,$07,$C9,$F4
	.dc.b	$25,$77,$8C,$B4,$87,$F3,$8C,$FF,$CC,$3F,$61,$F9,$68,$FC,$CB,$5F
	.dc.b	$E7,$AD,$DD,$B7,$F4,$E8,$FC,$3C,$F0,$E8,$FF,$6B,$C6,$E7,$51,$8E
	.dc.b	$6F,$F6,$DE,$4F,$AB,$A4,$C3,$F5,$3F,$61,$F9,$2C,$BF,$BC,$7D,$72
	.dc.b	$FF,$AE,$9C,$6B,$F5,$ED,$7F,$A9,$B6,$7C,$F1,$FE,$01,$AE,$8A,$BA
	.dc.b	$39,$C3,$FB,$CB,$9B,$EF,$87,$F6,$CA,$DC,$F8,$E3,$13,$FD,$C2,$FA
	.dc.b	$D6,$9B,$1C,$AB,$D6,$D7,$B1,$E1,$C3,$2D,$32,$77,$32,$53,$33,$D9
	.dc.b	$E1,$5E,$6A,$F8,$8D,$BF,$3C,$EA,$7D,$18,$7C,$0A,$D4,$F6,$9E,$26
	.dc.b	$B0,$F0,$E1,$F2,$A7,$FC,$1E,$1C,$3B,$32,$66,$FE,$6B,$C2,$1F,$5E
	.dc.b	$FB,$0D,$CF,$38,$C3,$CA,$77,$3A,$6C,$70,$1F,$43,$5A,$65,$C1,$1F
	.dc.b	$9B,$87,$F3,$6D,$7E,$6C,$7F,$91,$E7,$E7,$CE,$63,$E9,$99,$38,$EB
	.dc.b	$F8,$ED,$3F,$0B,$19,$7F,$0F,$3C,$5C,$C7,$91,$CC,$F1,$D3,$E3,$DF
	.dc.b	$82,$31,$5F,$3A,$31,$87,$9A,$39,$57,$F9,$03,$0E,$CF,$28,$F2,$CE
	.dc.b	$EB,$8C,$60,$FF,$1C,$D7,$B4,$FD,$3B,$FE,$4A,$8D,$C6,$0D,$4F,$7F
	.dc.b	$B0,$FD,$6C,$67,$0B,$39,$3F,$B1,$FD,$87,$EF,$DB,$BE,$B1,$FF,$3A
	.dc.b	$4F,$F9,$D6,$72,$48,$FF,$29,$5F,$F4,$4B,$E1,$0A,$E8,$EB,$AF,$CA
	.dc.b	$F4,$E7,$ED,$6F,$B0,$7C,$DF,$2C,$3F,$C7,$F9,$7F,$88,$27,$2F,$A2
	.dc.b	$F8,$C6,$30,$FF,$2F,$D7,$07,$CB,$19,$C3,$E4,$CB,$EC,$3B,$1E,$2F
	.dc.b	$B0,$FF,$12,$63,$F1,$23,$2F,$F8,$C2,$1E,$F6,$31,$F8,$31,$89,$E0
	.dc.b	$6E,$75,$87,$D2,$9D,$57,$F9,$D3,$E5,$7F,$BB,$DF,$F8,$51,$B1,$FE
	.dc.b	$78,$FB,$0F,$F2,$6C,$E7,$AF,$E6,$FD,$87,$F9,$BE,$3F,$C3,$6D,$AE
	.dc.b	$93,$AF,$7F,$E9,$DB,$EC,$29,$9C,$3F,$C4,$71,$97,$FC,$4B,$F6,$0F
	.dc.b	$93,$FB,$6D,$8E,$FF,$47,$89,$CF,$1E,$B9,$6D,$72,$7F,$07,$E1,$FF
	.dc.b	$92,$23,$FC,$63,$ED,$3C,$8C,$F9,$5F,$88,$8C,$4F,$F9,$7B,$CA,$FB
	.dc.b	$0F,$F9,$25,$BC,$3F,$EB,$A7,$FD,$89,$1B,$EA,$FC,$5F,$D8,$3E,$CE
	.dc.b	$FC,$86,$1B,$5F,$E6,$AF,$B0,$FF,$37,$37,$E0,$7F,$27,$18,$CB,$F3
	.dc.b	$AF,$FB,$82,$7F,$CE,$3F,$22,$FA,$1F,$CC,$BF,$5A,$BF,$46,$E6,$77
	.dc.b	$F0,$BF,$9E,$FA,$EE,$9D,$FF,$B4,$6C,$FF,$99,$5E,$D7,$FA,$C7,$2F
	.dc.b	$FB,$61,$FF,$1B,$47,$FA,$09,$B1,$BF,$8C,$ED,$C1,$FA,$CF,$AB,$FA
	.dc.b	$13,$FB,$66,$E1,$FC,$11,$FF,$33,$61,$FE,$8C,$72,$FF,$9D,$DE,$B1
	.dc.b	$BB,$E9,$7D,$5F,$DC,$FD,$6F,$FA,$26,$FF,$D2,$AF,$EC,$6B,$F8,$8B
	.dc.b	$9C,$3F,$83,$B9,$9F,$F4,$27,$D6,$BE,$1B,$F8,$1A,$F9,$9B,$18,$7F
	.dc.b	$83,$E3,$CF,$1F,$E3,$AD,$BF,$F4,$73,$D7,$1F,$F7,$FB,$97,$FE,$00
	.dc.b	$FD,$6B,$EF,$BF,$88,$B6,$BF,$D5,$AB,$E3,$33,$3F,$C5,$BE,$A1,$F3
	.dc.b	$54,$C7,$1D,$C3,$FD,$6B,$3A,$6C,$E9,$A4,$C7,$F9,$4C,$ED,$B5,$CD
	.dc.b	$FE,$5B,$F5,$2F,$0F,$F9,$3B,$FE,$C0,$BF,$F6,$1B,$96,$F6,$EA,$76
	.dc.b	$D8,$C9,$FE,$A2,$FA,$BF,$D5,$D9,$FD,$6E,$EC,$9F,$59,$FC,$F2,$76
	.dc.b	$7F,$2F,$AF,$F0,$6B,$51,$1C,$37,$BF,$FA,$CB,$FE,$C3,$CF,$FD,$8F
	.dc.b	$8F,$AB,$9B,$DB,$F1,$9E,$E6,$C6,$4F,$7C,$65,$A6,$6B,$E3,$C6,$5F
	.dc.b	$9B,$F0,$BF,$64,$FB,$5E,$65,$AF,$AF,$16,$7E,$9C,$DD,$7F,$B7,$FD
	.dc.b	$57,$EE,$E5,$18,$EE,$FA,$7C,$FF,$D8,$2B,$53,$9C,$7F,$C4,$F1,$C6
	.dc.b	$FF,$C6,$6D,$7E,$38,$70,$F6,$E0,$FF,$6F,$CF,$E3,$3E,$37,$FB,$3F
	.dc.b	$90,$F4,$3F,$29,$12,$74,$7F,$B8,$F1,$FF,$5B,$CE,$70,$FC,$7D,$59
	.dc.b	$FF,$02,$DF,$3B,$E3,$5F,$90,$EC,$FC,$79,$F9,$0E,$CB,$3F,$90,$8B
	.dc.b	$3F,$96,$9B,$7F,$E1,$6F,$B3,$E9,$3F,$E4,$E6,$BF,$E5,$46,$BF,$09
	.dc.b	$AF,$E2,$BE,$A5,$F5,$7E,$F8,$FF,$2B,$C6,$34,$FF,$29,$AF,$9E,$7D
	.dc.b	$5E,$2F,$85,$FF,$98,$BE,$51,$F5,$3F,$E3,$3D,$BE,$8C,$62,$7F,$33
	.dc.b	$A7,$F8,$BA,$31,$1F,$80,$9F,$AF,$F5,$3F,$E2,$F7,$8D,$F3,$9F,$D0
	.dc.b	$D7,$0B,$E5,$39,$95,$F1,$77,$19,$AE,$17,$D2,$7C,$6F,$80,$D7,$B5
	.dc.b	$23,$27,$81,$E5,$F4,$19,$8F,$41,$93,$62,$BF,$C3,$6E,$14,$79,$FE
	.dc.b	$A1,$F5,$3E,$5F,$85,$97,$CA,$72,$D9,$CB,$E0,$3B,$4F,$38,$F8,$5F
	.dc.b	$53,$FC,$FE,$7C,$6D,$78,$37,$5F,$F2,$FB,$FD,$0E,$37,$CB,$C8,$8F
	.dc.b	$2E,$5A,$F2,$8B,$5C,$A6,$EB,$FB,$0B,$EC,$DA,$F2,$CE,$AF,$F8,$DD
	.dc.b	$69,$B3,$A7,$C6,$43,$C0,$78,$CC,$5F,$AA,$3A,$42,$7D,$8F,$02,$4E
	.dc.b	$DE,$46,$2F,$CC,$F4,$7D,$57,$D2,$BC,$D3,$D6,$FC,$CF,$13,$D2,$71
	.dc.b	$3D,$B3,$C6,$39,$36,$34,$79,$DC,$3C,$CC,$1A,$9F,$09,$23,$77,$EB
	.dc.b	$9F,$BB,$18,$BF,$5C,$60,$E5,$7E,$26,$9D,$67,$83,$EA,$38,$1E,$3B
	.dc.b	$F0,$CF,$C0,$BF,$01,$2F,$BB,$F8,$3A,$EA,$71,$BE,$AF,$C9,$3E,$8C
	.dc.b	$71,$7C,$0D,$7E,$09,$E1,$AF,$43,$4F,$41,$DA,$38,$5F,$E0,$1B,$3C
	.dc.b	$1B,$3C,$0E,$47,$94,$E5,$7E,$A3,$F0,$2E,$0F,$2B,$87,$8C,$78,$2F
	.dc.b	$DB,$7F,$56,$78,$1E,$F3,$E5,$CF,$89,$F0,$1C,$AB,$C0,$FE,$20,$E3
	.dc.b	$3C,$DD,$1E,$23,$46,$0E,$D3,$F4,$5C,$47,$18,$F0,$1E,$78,$EB,$7F
	.dc.b	$6E,$BC,$31,$C9,$B5,$F2,$DB,$3F,$2D,$B3,$F2,$EB,$AC,$70,$BF,$87
	.dc.b	$5D,$6B,$D0,$75,$2D,$F5,$7F,$81,$FB,$BF,$C6,$BE,$8E,$47,$F1,$A8
	.dc.b	$FE,$3A,$61,$FE,$37,$18,$7F,$98,$FD,$4B,$D6,$FE,$47,$EA,$C9,$CC
	.dc.b	$7F,$47,$9D,$47,$F2,$69,$D3,$FC,$B6,$32,$FF,$2F,$FA,$7F,$F2,$AB
	.dc.b	$FC,$F9,$79,$9F,$D4,$5B,$19,$F3,$70,$66,$64,$64,$87,$57,$2F,$78
	.dc.b	$E9,$F5,$12,$3C,$0F,$7F,$A6,$C4,$79,$62,$F0,$B8,$7F,$C4,$38,$AF
	.dc.b	$FB,$37,$63,$4F,$45,$4C,$7F,$AC,$CE,$5B,$3F,$E2,$A6,$46,$5C,$1F
	.dc.b	$E0,$58,$AF,$F9,$5A,$34,$F1,$FA,$7F,$EC,$6F,$F5,$97,$FB,$1B,$FD
	.dc.b	$A1,$AE,$7A,$7F,$BE,$47,$0C,$BD,$93,$FE,$62,$8D,$C6,$21,$18,$C3
	.dc.b	$FC,$82,$F1,$6D,$74,$E8,$95,$E1,$8E,$FF,$39,$23,$C8,$F8,$BA,$31
	.dc.b	$1E,$98,$B8,$F8,$71,$57,$CE,$F1,$BF,$C2,$9A,$FE,$77,$0F,$A5,$D1
	.dc.b	$97,$E7,$CF,$C8,$7F,$C5,$CD,$D4,$D3,$54,$E6,$EA,$4A,$FF,$08,$C3
	.dc.b	$9B,$11,$FE,$07,$8E,$0E,$A7,$CB,$3A,$CF,$E5,$56,$AB,$85,$69,$EB
	.dc.b	$F4,$7E,$CE,$19,$18,$21,$DD,$CB,$E1,$8F,$F1,$B7,$D3,$3E,$0B,$FC
	.dc.b	$71,$F4,$FF,$A8,$AF,$F0,$E3,$72,$FB,$95,$B9,$FF,$16,$B5,$FC,$8B
	.dc.b	$5F,$CB,$35,$FF,$40,$B5,$FF,$54,$CB,$9B,$C2,$48,$FF,$1C,$C7,$E9
	.dc.b	$DB,$19,$7C,$AF,$A3,$3C,$5F,$E4,$6B,$A5,$70,$39,$CF,$F9,$66,$5F
	.dc.b	$55,$FE,$6B,$9F,$F4,$53,$FE,$72,$6F,$87,$B8,$D4,$7F,$93,$A5,$F6
	.dc.b	$FF,$E5,$A8,$FF,$53,$B6,$33,$E1,$F7,$FB,$9F,$F3,$14,$3E,$83,$FC
	.dc.b	$5C,$DF,$0B,$C9,$18,$3F,$99,$47,$9B,$E0,$FC,$44,$C7,$CF,$E1,$75
	.dc.b	$F6,$FB,$7B,$92,$7F,$D4,$99,$7F,$C0,$17,$FD,$4D,$8B,$EE,$FE,$16
	.dc.b	$7F,$C2,$F3,$FD,$D9,$FE,$55,$AF,$75,$6D,$F8,$BE,$1F,$C3,$F1,$24
	.dc.b	$60,$EF,$D9,$EF,$23,$E2,$E2,$EF,$E5,$C8,$FF,$5C,$65,$FE,$9A,$6F
	.dc.b	$C9,$FD,$5F,$27,$FD,$37,$A7,$91,$BE,$37,$FA,$61,$B5,$FE,$D6,$5F
	.dc.b	$5B,$F3,$FE,$27,$E8,$7D,$1C,$AF,$87,$1F,$03,$EE,$7E,$0F,$C6,$C9
	.dc.b	$F1,$C7,$FA,$67,$E9,$1E,$DD,$24,$FF,$A3,$2D,$F1,$BF,$E9,$F5,$EF
	.dc.b	$D8,$93,$B8,$FF,$52,$7D,$03,$DA,$E1,$12,$7F,$D2,$F3,$97,$FD,$33
	.dc.b	$F4,$9F,$AD,$E1,$FE,$59,$8F,$F1,$93,$F8,$D4,$67,$87,$FD,$DA,$DA
	.dc.b	$07,$0D,$F1,$FC,$7F,$49,$23,$BC,$7F,$AE,$67,$78,$AB,$FD,$79,$F4
	.dc.b	$8F,$92,$F5,$DC,$C3,$FD,$5F,$39,$6C,$F0,$23,$EE,$FA,$4A,$FF,$AA
	.dc.b	$67,$83,$67,$E3,$CF,$83,$83,$E4,$FA,$52,$BB,$C6,$5A,$43,$FD,$69
	.dc.b	$93,$93,$FD,$A9,$F4,$8F,$9D,$A4,$AF,$F6,$3C,$3E,$93,$11,$FE,$BD
	.dc.b	$8E,$0E,$A7,$9C,$FF,$B7,$F1,$75,$D2,$57,$FB,$4B,$E8,$BA,$72,$7D
	.dc.b	$4F,$A2,$FE,$AB,$3F,$5A,$F8,$BF,$ED,$86,$BF,$EC,$D6,$CF,$F6,$DA
	.dc.b	$E7,$2A,$E6,$A3,$31,$8A,$E6,$DB,$FC,$2D,$71,$9B,$5C,$1B,$83,$0F
	.dc.b	$44,$70,$39,$96,$F9,$E1,$E2,$BE,$58,$7E,$49,$AF,$F7,$49,$DD,$71
	.dc.b	$8E,$A6,$F0,$F0,$3E,$53,$18,$69,$E4,$59,$F1,$BA,$38,$1F,$0E,$79
	.dc.b	$E1,$F0,$63,$85,$F8,$9E,$11,$E9,$9F,$0C,$7F,$8B,$F0,$F9,$ED,$8C
	.dc.b	$BF,$1B,$0F,$F1,$7C,$BE,$91,$F4,$98,$6D,$7F,$8D,$A7,$53,$93,$FC
	.dc.b	$21,$0F,$A4,$A4,$7F,$C2,$8D,$C7,$EF,$E1,$F9,$BE,$7F,$89,$5F,$C0
	.dc.b	$2B,$E3,$4E,$63,$E1,$99,$3E,$2B,$9A,$FC,$4C,$65,$EB,$AF,$F8,$F9
	.dc.b	$CC,$79,$4E,$31,$9D,$7F,$CA,$37,$E0,$8C,$4F,$F9,$5E,$31,$87,$9A
	.dc.b	$39,$5F,$68,$C4,$EA,$7C,$E6,$AF,$F2,$F8,$79,$DB,$87,$F5,$8B,$A7
	.dc.b	$F2,$09,$FF,$29,$57,$AE,$B2,$FE,$6E,$6C,$FF,$95,$C7,$D2,$73,$9E
	.dc.b	$57,$5F,$D1,$45,$FA,$3F,$DF,$DF,$73,$F5,$6E,$27,$B1,$FC,$C9,$FE
	.dc.b	$67,$1B,$8C,$6F,$CC,$F0,$61,$FD,$17,$0C,$56,$E3,$0F,$1A,$5F,$F8
	.dc.b	$E6,$35,$F4,$8F,$92,$F3,$74,$F4,$8C,$4F,$1A,$E1,$7F,$E1,$B9,$CB
	.dc.b	$E8,$3A,$D7,$1B,$F8,$1F,$49,$30,$F2,$7D,$27,$F4,$57,$A4,$66,$FF
	.dc.b	$C4,$D2,$F8,$3C,$67,$FE,$0E,$8C,$5F,$03,$9C,$E9,$FB,$51,$AA,$F2
	.dc.b	$7C,$67,$F7,$49,$F8,$4D,$8F,$47,$D1,$3E,$7F,$EC,$4E,$71,$FF,$4A
	.dc.b	$7D,$27,$F9,$DF,$0F,$94,$DC,$3A,$F6,$FD,$F8,$C5,$7F,$A6,$F1,$7D
	.dc.b	$27,$E0,$DB,$87,$2F,$83,$C4,$FE,$1F,$E9,$3F,$A8,$4F,$D7,$6E,$5F
	.dc.b	$37,$17,$CE,$EF,$C0,$E1,$C7,$C7,$C9,$B8,$F1,$7F,$25,$E2,$7F,$84
	.dc.b	$23,$F0,$5F,$3A,$F2,$73,$E5,$FE,$5E,$31,$3F,$89,$F2,$FE,$EF,$17
	.dc.b	$F8,$F9,$B1,$FA,$75,$F5,$DF,$92,$73,$BF,$F4,$1F,$A5,$7B,$FF,$96
	.dc.b	$7F,$DF,$97,$F9,$57,$2D,$F7,$FF,$37,$F4,$AF,$88,$BF,$13,$86,$D7
	.dc.b	$E7,$7D,$C7,$CD,$FE,$89,$FB,$51,$98,$C7,$77,$E9,$BE,$97,$A4,$FE
	.dc.b	$95,$BA,$FF,$9D,$5F,$F7,$64,$67,$0E,$1B,$48,$E1,$F1,$B6,$FF,$CF
	.dc.b	$9F,$4B,$F3,$5B,$C9,$FB,$48,$7D,$45,$8C,$61,$9B,$B8,$7E,$DE,$FF
	.dc.b	$D8,$EF,$EE,$7E,$99,$F3,$3F,$BE,$7F,$63,$CD,$FE,$70,$73,$3B,$EB
	.dc.b	$FE,$07,$F4,$CF,$81,$BF,$27,$86,$D7,$F0,$5C,$F8,$EF,$FE,$95,$6E
	.dc.b	$5F,$E7,$C7,$BC,$FF,$25,$72,$FF,$27,$FA,$67,$C6,$7F,$12,$6C,$7F
	.dc.b	$11,$5F,$19,$99,$FE,$2F,$F4,$0F,$8A,$A6,$3A,$6E,$1F,$C6,$67,$4D
	.dc.b	$9D,$34,$99,$FF,$AB,$67,$6D,$AE,$6F,$FA,$C7,$E9,$3F,$53,$5F,$C9
	.dc.b	$57,$83,$FC,$A9,$CC,$EE,$31,$E3,$65,$B5,$C5,$FE,$7D,$F4,$BF,$A0
	.dc.b	$A3,$FA,$2D,$D9,$3E,$9B,$7E,$9C,$C7,$F9,$B7,$D8,$F8,$9F,$E6,$D8
	.dc.b	$71,$8C,$63,$E9,$FA,$7F,$37,$5F,$9D,$B7,$EB,$6C,$EE,$3D,$D3,$B6
	.dc.b	$9B,$2C,$7C,$9F,$A7,$FA,$A7,$D6,$F4,$AC,$FD,$38,$B3,$F3,$E6,$EA
	.dc.b	$AF,$F2,$2F,$D3,$FE,$D1,$FE,$A0,$DA,$FA,$5C,$FF,$D9,$63,$15,$C6
	.dc.b	$7F,$D2,$31,$8A,$FE,$E9,$F6,$3F,$D1,$E7,$FA,$74,$E5,$FF,$27,$46
	.dc.b	$6B,$FD,$A7,$C0,$FA,$8F,$C7,$9F,$D9,$BC,$F7,$AD,$8C,$BF,$1D,$37
	.dc.b	$1F,$F7,$B5,$6A,$72,$48,$FC,$74,$F3,$AE,$18,$7F,$B8,$B9,$BF,$61
	.dc.b	$7F,$8D,$EC,$FC,$67,$6F,$E3,$59,$19,$3F,$25,$18,$C7,$FB,$57,$D3
	.dc.b	$BE,$4F,$CF,$1F,$E7,$03,$FD,$6B,$19,$CB,$11,$8F,$94,$FF,$72,$3B
	.dc.b	$9D,$3A,$FB,$17,$C6,$70,$8F,$F5,$FC,$61,$C3,$5E,$11,$B7,$F8,$9F
	.dc.b	$CC,$FF,$8B,$33,$FF,$41,$F8,$E8,$C4,$FF,$B2,$A3,$0F,$F8,$47,$EA
	.dc.b	$3F,$C5,$15,$C2,$B7,$3D,$0F,$F0,$B5,$70,$7F,$C1,$95,$C1,$E9,$3C
	.dc.b	$27,$5C,$1F,$89,$9E,$47,$0B,$D7,$BF,$23,$AB,$C7,$E9,$31,$1F,$E1
	.dc.b	$9B,$FF,$08,$3A,$3C,$71,$E6,$FA,$87,$CC,$52,$4F,$A3,$DC,$AF,$F8
	.dc.b	$8E,$76,$FF,$8C,$5A,$F7,$8F,$83,$F5,$1F,$CD,$E3,$FC,$6C,$D7,$FC
	.dc.b	$58,$DC,$FD,$CB,$93,$8F,$2C,$AE,$2F,$48,$C4,$72,$8B,$5C,$A6,$EB
	.dc.b	$FE,$BC,$BE,$CD,$AE,$57,$D0,$87,$33,$CF,$5C,$90,$AE,$6E,$1A,$E6
	.dc.b	$38,$3B,$9B,$3E,$D8,$EF,$7F,$02,$73,$5A,$6C,$EA,$FD,$0D,$3C,$13
	.dc.b	$B7,$07,$84,$B7,$E4,$24,$6E,$3D,$C7,$D2,$71,$7E,$38,$F8,$27,$2B
	.dc.b	$F1,$BC,$23,$4D,$BF,$9D,$5C,$9F,$3C,$75,$39,$1E,$48,$D3,$F2,$6B
	.dc.b	$E9,$BC,$CC,$1B,$38,$99,$3D,$C5,$BF,$74,$72,$BF,$29,$2F,$C6,$FE
	.dc.b	$16,$B9,$B9,$7E,$59,$88,$D9,$EC,$70,$79,$0F,$61,$E1,$9F,$8B,$3C
	.dc.b	$E3,$F0,$07,$A4,$D1,$5D,$57,$A2,$3E,$79,$A3,$D9,$7F,$19,$E6,$78
	.dc.b	$27,$8D,$78,$2F,$CF,$3E,$59,$EB,$5C,$6F,$F0,$33,$EA,$39,$CF,$C3
	.dc.b	$7C,$0F,$07,$51,$C1,$CD,$76,$7E,$8C,$F8,$9F,$64,$66,$7D,$91,$E7
	.dc.b	$8C,$5F,$9E,$3C,$D1,$87,$E0,$D7,$37,$E0,$9E,$29,$FB,$93,$B8,$C9
	.dc.b	$E4,$39,$9F,$68,$C1,$EE,$7E,$D3,$8A,$F6,$57,$73,$E0,$47,$F1,$4F
	.dc.b	$81,$FE,$33,$F4,$92,$3F,$8C,$C7,$F1,$C3,$0F,$F1,$A8,$C3,$FE,$D3
	.dc.b	$FA,$87,$B7,$FD,$35,$B1,$FD,$55,$AF,$F5,$E6,$BF,$D8,$DC,$DF,$F4
	.dc.b	$16,$E3,$FC,$4A,$51,$87,$F8,$97,$CD,$FF,$50,$FA,$8F,$F2,$94,$7F
	.dc.b	$B9,$57,$9B,$93,$5E,$74,$9D,$46,$5E,$11,$97,$FD,$C5,$3C,$9D,$9A
	.dc.b	$3B,$7D,$2F,$84,$8F,$87,$E9,$2B,$CA,$2E,$7F,$D7,$69,$5F,$DA,$3B
	.dc.b	$5E,$F7,$F7,$56,$E1,$FD,$F5,$AF,$F8,$19,$AF,$F8,$2D,$CD,$FF,$69
	.dc.b	$6E,$BF,$25,$B3,$FC,$F3,$E9,$FF,$D7,$EF,$F3,$EA,$FE,$E1,$1F,$DA
	.dc.b	$21,$18,$E8,$FE,$FB,$26,$F5,$4A,$E5,$87,$D7,$AD,$46,$0F,$05,$6D
	.dc.b	$EB,$3B,$AE,$B3,$27,$80,$AF,$25,$F9,$BE,$A3,$C6,$49,$DC,$BE,$4B
	.dc.b	$11,$D8,$F8,$13,$8A,$F7,$CF,$D1,$F2,$BF,$C3,$3A,$FF,$57,$BE,$07
	.dc.b	$D6,$E9,$D3,$E4,$D5,$58,$8D,$C6,$1F,$4D,$57,$9A,$E7,$EE,$48,$DF
	.dc.b	$3D,$77,$17,$2F,$F1,$25,$2B,$1B,$FE,$96,$3A,$5F,$82,$FD,$75,$EC
	.dc.b	$F7,$BD,$BC,$32,$BC,$B1,$97,$CF,$19,$7C,$F1,$FE,$3C,$FA,$87,$B7
	.dc.b	$FE,$3E,$FA,$8F,$D2,$57,$C6,$6E,$9F,$E3,$9A,$DC,$FD,$A6,$BF,$31
	.dc.b	$B1,$BA,$E9,$F1,$3D,$79,$E5,$FE,$0B,$04,$78,$75,$CC,$F8,$62,$E1
	.dc.b	$FE,$49,$AF,$D6,$CF,$19,$CE,$DF,$C2,$E7,$9C,$F5,$BF,$F2,$FD,$BE
	.dc.b	$A1,$C1,$87,$FC,$F5,$5F,$9B,$9D,$79,$59,$AE,$3F,$51,$FE,$5A,$CB
	.dc.b	$F5,$8D,$BF,$D5,$B6,$DF,$51,$FE,$63,$5F,$4F,$FE,$2E,$6F,$87,$E4
	.dc.b	$8C,$1F,$95,$8F,$F5,$71,$E7,$F8,$BF,$31,$30,$F3,$69,$FC,$9E,$7B
	.dc.b	$FB,$FE,$EC,$7F,$DF,$75,$EB,$6B,$FA,$79,$DB,$EF,$C1,$F5,$1F,$1B
	.dc.b	$2F,$F4,$D5,$78,$9B,$A7,$F0,$BA,$DD,$F8,$BC,$1F,$95,$E1,$48,$C1
	.dc.b	$E1,$DF,$C8,$47,$C9,$C1,$DF,$EA,$24,$7D,$1C,$BF,$D4,$B9,$F8,$7E
	.dc.b	$67,$EC,$F9,$7F,$EA,$EA,$F5,$36,$7E,$3B,$75,$FE,$40,$BC,$6F,$89
	.dc.b	$F2,$1F,$17,$EA,$26,$1F,$16,$3E,$07,$D4,$7C,$1F,$36,$4F,$8E,$3F
	.dc.b	$D4,$3F,$50,$F7,$69,$27,$F8,$DD,$BE,$33,$FD,$4E,$BE,$02,$C4,$9D
	.dc.b	$C7,$FA,$B3,$E9,$5E,$EF,$08,$93,$FC,$9A,$72,$FF,$28,$FA,$8F,$D6
	.dc.b	$F0,$7F,$69,$C7,$F1,$0D,$F2,$FF,$D6,$F3,$AA,$FF,$3F,$B5,$34,$C5
	.dc.b	$BE,$A3,$C8,$72,$F1,$A5,$78,$E3,$FD,$87,$F5,$0F,$1F,$D5,$52,$BF
	.dc.b	$9E,$CE,$5B,$3C,$08,$FB,$7D,$BA,$FF,$34,$9E,$0D,$9F,$8D,$3E,$0E
	.dc.b	$0F,$91,$E7,$CA,$EF,$19,$69,$0F,$E7,$F9,$39,$3F,$DB,$7F,$50,$FA
	.dc.b	$1A,$4A,$FE,$AB,$0F,$A8,$C4,$7F,$48,$8E,$0E,$A7,$9C,$FF,$B9,$B1
	.dc.b	$76,$52,$57,$F6,$0F,$A7,$E9,$DB,$FE,$B2,$FA,$6F,$EB,$93,$FE,$B4
	.dc.b	$BE,$2F,$F6,$66,$BF,$D6,$1B,$3F,$EE,$BA,$E7,$2A,$E6,$A3,$31,$8A
	.dc.b	$E6,$DB,$FF,$19,$3C,$27,$73,$FE,$35,$6D,$79,$6B,$73,$FE,$39,$39
	.dc.b	$9B,$C3,$FC,$1B,$3E,$38,$C6,$1E,$F7,$85,$72,$9E,$86,$4F,$F6,$AD
	.dc.b	$F8,$1F,$55,$F4,$C3,$D1,$1B,$3F,$C0,$95,$CD,$DE,$19,$BE,$58,$79
	.dc.b	$A7,$B9,$E5,$3B,$E1,$D6,$3C,$F1,$FE,$55,$33,$55,$A5,$AF,$A2,$ED
	.dc.b	$7D,$43,$EA,$38,$4F,$D3,$8C,$3C,$5A,$F1,$6B,$DF,$EA,$38,$BF,$EF
	.dc.b	$86,$A6,$91,$F6,$1F,$E3,$ED,$1D,$1F,$1E,$39,$1A,$C7,$F4,$51,$98
	.dc.b	$E8,$66,$78,$BD,$0E,$3A,$FE,$3A,$B8,$C6,$4E,$31,$99,$E2,$FE,$52
	.dc.b	$38,$E7,$FE,$81,$AE,$33,$99,$E3,$39,$AE,$2F,$C3,$8F,$25,$F9,$F0
	.dc.b	$FC,$16,$1C,$AB,$58,$78,$6B,$38,$7F,$6C,$7A,$1F,$E5,$2C,$3C,$67
	.dc.b	$F9,$9E,$34,$D8,$FF,$2B,$DB,$EA,$33,$3F,$36,$73,$18,$FA,$8F,$E0
	.dc.b	$31,$FE,$5E,$6C,$FF,$94,$13,$BD,$E4,$FF,$84,$7B,$7E,$5E,$1F,$42
	.dc.b	$97,$CA,$3F,$C0,$B3,$AA,$FF,$55,$4F,$F8,$E9,$FF,$0D,$7D,$43,$E4
	.dc.b	$FC,$9D,$5E,$CC,$13,$9A,$F3,$4E,$63,$18,$F9,$E3,$17,$ED,$FA,$89
	.dc.b	$7F,$E2,$7F,$A8,$FE,$90,$FB,$E3,$37,$FE,$0F,$97,$C0,$E3,$2F,$F0
	.dc.b	$EC,$62,$F8,$1C,$E7,$4F,$FA,$66,$35,$5F,$E7,$8F,$8C,$FE,$E5,$3F
	.dc.b	$E8,$A6,$C7,$F9,$F3,$EA,$1F,$43,$F5,$A7,$38,$FC,$BF,$A6,$FF,$3B
	.dc.b	$61,$FE,$92,$6E,$1D,$7B,$BE,$EC,$62,$BE,$96,$2F,$A8,$FC,$0B,$70
	.dc.b	$E5,$F0,$58,$9F,$C2,$FD,$47,$F4,$F9,$FF,$4F,$B7,$2F,$F4,$8E,$2F
	.dc.b	$9D,$DF,$83,$C3,$9A,$61,$C3,$92,$E1,$FE,$98,$F9,$D7,$95,$F9,$26
	.dc.b	$D7,$FA,$B3,$C6,$FF,$09,$3F,$EA,$E8,$CD,$7F,$90,$7C,$B7,$C8,$7E
	.dc.b	$0F,$E9,$FF,$82,$DF,$F9,$86,$72,$FF,$98,$E3,$35,$FE,$49,$5F,$4D
	.dc.b	$C4,$E5,$CF,$6B,$FC,$D5,$F4,$FE,$2B,$FF,$28,$37,$9B,$FC,$BA,$BE
	.dc.b	$9B,$F2,$8E,$6B,$FD,$B1,$ED,$FC,$DB,$FF,$2C,$36,$FF,$CC,$AF,$F8
	.dc.b	$46,$7F,$CF,$0E,$5F,$F7,$07,$D3,$F3,$AF,$F6,$FB,$DE,$3F,$CF,$CE
	.dc.b	$68,$E6,$7F,$CC,$CB,$D9,$FD,$42,$F4,$EC,$61,$FD,$67,$D3,$FE,$CE
	.dc.b	$BF,$DD,$AF,$F0,$7C,$3F,$D1,$8E,$70,$39,$9F,$F3,$A3,$DE,$77,$7D
	.dc.b	$2F,$AB,$FB,$9F,$A7,$7D,$95,$31,$EA,$DB,$FF,$4D,$7D,$47,$F9,$F1
	.dc.b	$BF,$07,$FC,$07,$E4,$3F,$88,$C6,$0F,$F5,$03,$D1,$B5,$BB,$E0,$FF
	.dc.b	$09,$FA,$8F,$F5,$4C,$FF,$A9,$D7,$C1,$67,$E0,$F4,$FF,$0F,$F9,$EF
	.dc.b	$CF,$57,$FA,$B5,$75,$E7,$C1,$D3,$67,$4D,$E2,$FF,$4C,$57,$09,$E5
	.dc.b	$5C,$1A,$70,$3F,$96,$F1,$FF,$29,$FA,$6F,$E3,$6D,$F3,$7F,$91,$7D
	.dc.b	$0B,$E9,$FF,$9D,$D6,$AF,$F9,$AE,$48,$C2,$C1,$F4,$FE,$38,$C5,$70
	.dc.b	$AD,$D6,$B4,$F8,$5F,$4F,$F0,$FC,$1B,$7F,$0D,$B8,$7B,$9B,$7F,$EC
	.dc.b	$2D,$BF,$D1,$55,$FE,$D5,$D1,$F4,$FE,$A3,$D7,$E7,$59,$D4,$5C,$7F
	.dc.b	$D8,$15,$FC,$8A,$7E,$2B,$C9,$D9,$23,$75,$C7,$0F,$EA,$5F,$4F,$FC
	.dc.b	$02,$BE,$69,$F1,$9C,$57,$D6,$E7,$FE,$CB,$18,$AE,$3E,$1F,$F5,$1F
	.dc.b	$A7,$FE,$8F,$7F,$20,$FE,$C7,$5B,$C3,$EC,$CE,$AF,$FD,$C7,$93,$E9
	.dc.b	$FE,$C9,$FD,$6F,$CB,$7A,$58,$C3,$E3,$4D,$CB,$FD,$C1,$5A,$9C,$9F
	.dc.b	$93,$F0,$3F,$2C,$7E,$5F,$C1,$B1,$92,$57,$E4,$A3,$2F,$E5,$26,$E5
	.dc.b	$F5,$E6,$D7,$E2,$BE,$AB,$C1,$7F,$E4,$96,$BF,$E4,$D6,$BB,$CB,$FC
	.dc.b	$00,$FE,$4B,$EA,$3F,$5B,$1F,$E4,$A6,$C7,$F9,$4D,$76,$BE,$3F,$5D
	.dc.b	$DF,$EE,$7A,$FC,$DF,$D3,$7F,$40,$7E,$C9,$FE,$12,$70,$FD,$30,$E3
	.dc.b	$4F,$F0,$97,$D3,$FF,$84,$DF,$F0,$A4,$67,$4E,$BE,$77,$F8,$ED,$B7
	.dc.b	$D8,$5F,$18,$F1,$FB,$3B,$9F,$13,$5F,$0B,$6D,$F4,$56,$24,$F5,$9F
	.dc.b	$F0,$6B,$67,$2F,$AE,$32,$D9,$E9,$3D,$FE,$9D,$F2,$7D,$FE,$12,$5F
	.dc.b	$A6,$72,$D9,$FF,$39,$C6,$23,$FD,$08,$D7,$FD,$10,$D7,$F8,$5F,$D3
	.dc.b	$FF,$4B,$7E,$23,$5F,$A0,$DD,$7F,$C8,$0F,$FB,$F2,$3F,$C8,$EE,$73
	.dc.b	$E7,$CB,$F1,$E3,$13,$AE,$5C,$7B,$99,$FA,$29,$3E,$08,$CB,$E2,$8C
	.dc.b	$BE,$28,$CB,$E4,$8C,$E3,$E4,$9C,$BE,$68,$CE,$3E,$E9,$CB,$F0,$23
	.dc.b	$35,$EA,$9C,$BE,$C8,$CA,$C1,$3A,$8C,$4E,$A7,$31,$EB,$3D,$44,$3D
	.dc.b	$46,$4F,$51,$9B,$F5,$39,$7D,$4E,$6F,$D4,$E5,$F5,$39,$3D,$46,$4F
	.dc.b	$51,$93,$D5,$7E,$9A,$5F,$A6,$30,$D9,$E4,$7B,$8E,$53,$EE,$7D,$67
	.dc.b	$A8,$97,$EA,$72,$FA,$9C,$DF,$A9,$CB,$EA,$72,$7A,$8C,$9E,$A3,$27
	.dc.b	$A8,$F2,$CF,$A9,$F2,$DF,$A9,$A6,$AF,$D8,$F4,$3D,$87,$49,$F6,$5F
	.dc.b	$A4,$F4,$12,$FD,$0E,$5F,$43,$99,$F4,$39,$7D,$0E,$5F,$41,$E1,$9F
	.dc.b	$43,$E1,$BF,$43,$4F,$41,$EA,$38,$5F,$AA,$BB,$47,$A9,$ED,$5E,$AB
	.dc.b	$F3,$9E,$62,$4F,$99,$CB,$E6,$72,$F9,$A7,$AC,$F9,$9E,$B7,$E6,$6C
	.dc.b	$F9,$8F,$41,$E6,$3D,$0F,$2B,$F4,$5F,$82,$7D,$0F,$82,$FD,$11,$E5
	.dc.b	$3C,$84,$3C,$91,$C6,$3C,$8F,$1B,$F2,$36,$7C,$97,$E6,$7C,$87,$99
	.dc.b	$F2,$3E,$68,$E9,$7E,$68,$F1,$1E,$63,$C5,$7E,$67,$C6,$62,$3F,$59
	.dc.b	$7F,$C0,$DB,$1F,$A8,$7F,$84,$7D,$33,$E9,$FF,$93,$7D,$24,$AF,$EA
	.dc.b	$93,$96,$CE,$5F,$E6,$73,$FD,$4E,$75,$1F,$D9,$7E,$9D,$F3,$5F,$D5
	.dc.b	$D7,$AF,$4C,$44,$9F,$EC,$53,$A6,$CE,$5E,$9F,$47,$CC,$8F,$0F,$A2
	.dc.b	$AE,$8E,$71,$E1,$8C,$EE,$DF,$FC,$B3,$82,$B3,$18,$AF,$ED,$DD,$8D
	.dc.b	$3D,$53,$31,$FE,$F1,$39,$6C,$FF,$99,$19,$19,$70,$7F,$84,$A2,$A3
	.dc.b	$06,$3E,$9F,$FB,$9E,$5E,$78,$D4,$7F,$7E,$7F,$C1,$0D,$7F,$C2,$8F
	.dc.b	$0B,$FF,$3A,$B6,$38,$F3,$FF,$82,$63,$85,$75,$94,$6D,$BF,$4F,$B2
	.dc.b	$3E,$0E,$8A,$F8,$22,$D7,$3A,$56,$30,$F1,$D5,$BF,$5C,$FC,$EF,$23
	.dc.b	$FC,$39,$A7,$F7,$B9,$37,$44,$EE,$7E,$8B,$5F,$F1,$A4,$E9,$FF,$16
	.dc.b	$3F,$A5,$27,$27,$6E,$4E,$EC,$9D,$B9,$78,$FD,$3F,$93,$79,$18,$21
	.dc.b	$E0,$72,$F8,$A3,$D6,$63,$6F,$9D,$87,$E0,$DB,$1F,$E3,$BF,$A7,$7C
	.dc.b	$27,$F8,$F7,$E9,$FF,$45,$87,$F9,$3D,$B7,$F5,$EB,$73,$FE,$49,$6B
	.dc.b	$FE,$55,$6C,$6E,$BA,$7C,$27,$5E,$49,$7F,$E5,$1C,$3F,$44,$DC,$B2
	.dc.b	$F8,$66,$E1,$F8,$7A,$FA,$B3,$FE,$59,$F5,$7F,$CB,$CF,$E6,$E3,$FC
	.dc.b	$B4,$DE,$3F,$F2,$C4,$EA,$3F,$D8,$BE,$1F,$BE,$64,$73,$8D,$98,$7E
	.dc.b	$14,$61,$FF,$3A,$DF,$F9,$53,$E4,$7F,$CC,$CD,$8C,$F8,$29,$CA,$F9
	.dc.b	$1D,$FD,$1F,$C0,$AF,$F5,$75,$A3,$DB,$18,$6C,$7E,$CC,$D6,$9E,$0F
	.dc.b	$09,$D3,$7A,$66,$1E,$69,$CB,$E4,$FB,$3F,$1F,$B9,$23,$07,$B3,$27
	.dc.b	$D3,$F9,$89,$19,$F0,$FC,$1F,$71,$23,$B3,$E3,$F9,$EF,$7E,$52,$7E
	.dc.b	$1B,$F0,$BC,$7F,$83,$EB,$7C,$2F,$8A,$95,$E1,$8F,$67,$99,$ED,$EC
	.dc.b	$93,$EB,$8F,$9F,$F5,$2F,$A4,$BF,$1F,$2B,$EA,$CE,$5F,$67,$C9,$7D
	.dc.b	$02,$3E,$FF,$03,$E5,$7C,$3C,$8F,$90,$FC,$DF,$87,$FF,$50,$7C,$83
	.dc.b	$E8,$FF,$D4,$5F,$48,$F8,$1A,$6A,$37,$F0,$97,$1F,$F5,$75,$F1,$6C
	.dc.b	$7F,$AA,$FE,$89,$EF,$F0,$9C,$9F,$EB,$58,$CB,$FE,$B7,$FA,$4F,$D6
	.dc.b	$F8,$DF,$CE,$61,$87,$F1,$93,$F8,$D4,$EA,$3F,$9F,$35,$34,$C5,$34
	.dc.b	$F2,$1F,$EB,$F8,$D7,$26,$2B,$FD,$91,$F4,$8E,$0F,$54,$C9,$CC,$7F
	.dc.b	$B1,$E7,$51,$82,$1F,$ED,$BC,$1D,$5B,$A9,$3A,$F1,$3F,$A3,$56,$A3
	.dc.b	$C3,$C1,$F2,$7C,$29,$18,$21,$FE,$E7,$C7,$EB,$38,$BE,$7C,$DF,$ED
	.dc.b	$BF,$A4,$FD,$B4,$FF,$B8,$DA,$FF,$BB,$DB,$3F,$D5,$D7,$D2,$71,$F2
	.dc.b	$65,$71,$7A,$47,$FB,$9F,$17,$CF,$D2,$57,$FB,$C7,$E9,$39,$E3,$FD
	.dc.b	$CF,$E8,$7F,$AC,$CF,$F7,$5B,$E2,$FF,$BD,$DA,$A7,$FB,$54,$61,$FE
	.dc.b	$D9,$5D,$25,$5D,$14,$66,$31,$5D,$1B,$7F,$8D,$78,$4E,$E7,$F2,$0D
	.dc.b	$AF,$35,$6E,$7F,$28,$73,$37,$87,$9E,$7C,$71,$8C,$3E,$03,$C2,$B9
	.dc.b	$4F,$43,$27,$F8,$5E,$FC,$0F,$AA,$FA,$61,$E8,$8D,$9F,$E0,$2A,$E6
	.dc.b	$EF,$0C,$DF,$2C,$3C,$D3,$DC,$F2,$9D,$F0,$EB,$1E,$78,$FF,$15,$CF
	.dc.b	$CD,$7F,$C6,$0D,$74,$63,$E8,$DF,$49,$FE,$30,$94,$62,$79,$B5,$E1
	.dc.b	$3F,$E3,$67,$FC,$6D,$F4,$95,$34,$8E,$DF,$83,$8F,$F5,$1E,$8F,$87
	.dc.b	$3C,$8F,$69,$F8,$9D,$7F,$C7,$DA,$FE,$16,$34,$FE,$1E,$B9,$47,$03
	.dc.b	$39,$FF,$11,$8C,$CF,$19,$F2,$9D,$A3,$33,$FE,$52,$C3,$CB,$7E,$08
	.dc.b	$C5,$6E,$30,$FF,$96,$E7,$78,$7A,$A3,$AE,$1E,$5A,$DC,$7F,$23,$AD
	.dc.b	$BC,$EB,$FC,$BD,$87,$F9,$3C,$F5,$9F,$E5,$18,$F5,$9A,$AF,$ED,$D3
	.dc.b	$ED,$6C,$72,$F3,$AB,$F2,$67,$A9,$F0,$E3,$3F,$49,$FC,$02,$33,$3F
	.dc.b	$E6,$F8,$C3,$FB,$98,$FE,$75,$1F,$E7,$37,$70,$8D,$DF,$F9,$9D,$B3
	.dc.b	$EC,$EB,$F5,$BC,$B8,$3C,$F5,$FD,$33,$1F,$1F,$46,$3E,$91,$F2,$D8
	.dc.b	$C3,$FC,$37,$18,$7C,$39,$7F,$91,$1C,$CE,$F3,$FF,$0B,$D6,$DF,$51
	.dc.b	$ED,$FA,$4F,$82,$78,$BE,$93,$FA,$65,$F4,$8C,$BF,$E2,$A8,$7C,$46
	.dc.b	$2B,$FC,$47,$18,$9E,$06,$E7,$5A,$7F,$88,$E7,$55,$E2,$1F,$1B,$FD
	.dc.b	$D6,$FE,$5B,$63,$5F,$48,$F3,$3E,$DC,$E7,$0F,$F4,$A7,$D1,$BE,$03
	.dc.b	$EF,$B7,$1E,$7D,$F8,$9F,$C1,$E2,$F2,$A9,$9A,$FB,$71,$97,$EE,$FD
	.dc.b	$23,$C0,$FC,$5B,$63,$F0,$38,$BD,$6C,$D6,$BA,$25,$72,$E6,$BC,$1F
	.dc.b	$07,$E6,$7F,$C0,$73,$F8,$A6,$C7,$C5,$F1,$7F,$C3,$53,$9F,$27,$F0
	.dc.b	$FE,$53,$DC,$FC,$4F,$D2,$7F,$B5,$EB,$FC,$BF,$39,$7F,$CC,$31,$9C
	.dc.b	$3F,$CB,$2B,$D3,$E2,$72,$E7,$B5,$FE,$68,$FA,$3F,$0D,$FF,$98,$1B
	.dc.b	$CF,$FE,$5D,$5F,$49,$F9,$A7,$35,$FB,$AF,$8C,$F9,$57,$FE,$68,$6D
	.dc.b	$FF,$99,$5F,$F0,$A4,$FF,$9D,$9C,$BF,$C0,$BE,$93,$9D,$7F,$69,$7B
	.dc.b	$C7,$F9,$F1,$CD,$1C,$CF,$F9,$D9,$7B,$3F,$AE,$5E,$CD,$8C,$3F,$B2
	.dc.b	$FA,$4F,$DA,$D7,$F7,$77,$FD,$05,$87,$FA,$29,$CE,$07,$33,$FE,$84
	.dc.b	$7B,$CE,$EF,$A5,$F5,$7F,$7F,$F4,$8F,$62,$98,$F5,$AD,$FF,$A6,$3E
	.dc.b	$91,$F4,$9F,$C1,$9B,$1F,$E9,$CF,$8D,$FE,$25,$A7,$FA,$52,$72,$FF
	.dc.b	$A5,$A3,$B4,$EE,$FF,$87,$C6,$27,$51,$FC,$4B,$E9,$1F,$1F,$7E,$63
	.dc.b	$0D,$C3,$FD,$57,$F4,$3F,$9F,$6F,$6F,$F1,$A8,$C6,$3F,$EA,$65,$19
	.dc.b	$F0,$74,$DA,$E2,$74,$8C,$B4,$CD,$7F,$07,$FA,$3F,$E3,$4F,$F9,$AC
	.dc.b	$38,$7F,$CD,$CD,$46,$BC,$EF,$E9,$4E,$23,$F9,$27,$2F,$FA,$DF,$E3
	.dc.b	$3C,$F7,$FC,$F2,$72,$FF,$54,$8C,$D7,$B7,$E8,$FE,$07,$7F,$C1,$6D
	.dc.b	$7C,$66,$CE,$E3,$E4,$CF,$F4,$0F,$A0,$7D,$1D,$31,$E6,$DA,$FF,$6C
	.dc.b	$CE,$9B,$3A,$69,$2F,$FD,$AB,$3B,$6D,$73,$7F,$98,$7D,$1F,$F0,$1A
	.dc.b	$F9,$51,$F1,$A3,$0E,$1A,$FC,$F6,$BF,$E9,$B7,$7A,$7F,$B7,$FE,$8F
	.dc.b	$FA,$4C,$F6,$F0,$FF,$AE,$C6,$27,$FD,$C9,$8B,$E8,$FE,$11,$FD,$8B
	.dc.b	$CA,$FE,$87,$97,$C0,$8C,$B4,$FF,$7D,$C6,$B1,$F7,$CF,$E4,$7B,$A5
	.dc.b	$7E,$5B,$BE,$D7,$E4,$E6,$E3,$FE,$F5,$E6,$FE,$07,$7F,$9C,$9C,$BF
	.dc.b	$E1,$28,$FD,$DD,$7E,$6E,$F2,$FE,$33,$E8,$9D,$3F,$D6,$A3,$E5,$C6
	.dc.b	$3D,$2F,$9B,$08,$FD,$6F,$D1,$61,$D4,$7D,$2F,$AA,$FF,$54,$37,$E1
	.dc.b	$FF,$C9,$91,$88,$FD,$03,$5F,$F6,$67,$C6,$FF,$43,$7F,$C2,$6D,$7F
	.dc.b	$B7,$37,$66,$BF,$ED,$3F,$A3,$FF,$17,$3C,$BC,$EF,$F2,$33,$6F,$9C
	.dc.b	$61,$FD,$5C,$F0,$7F,$5B,$7F,$E0,$47,$C3,$EC,$ED,$46,$5F,$1F,$CF
	.dc.b	$66,$31,$8F,$92,$79,$D6,$4D,$55,$8E,$A4,$39,$9F,$E3,$2F,$A3,$7C
	.dc.b	$C7,$85,$B1,$98,$F0,$7A,$D9,$64,$64,$F9,$31,$E9,$FA,$3F,$E9,$31
	.dc.b	$FE,$3E,$6B,$FE,$19,$6E,$3E,$98,$47,$F4,$E7,$87,$37,$32,$3C,$F9
	.dc.b	$7B,$7B,$9D,$FE,$7E,$4F,$78,$CB,$E1,$8C,$BE,$18,$CB,$E3,$8C,$E1
	.dc.b	$E3,$C3,$FC,$49,$3C,$CE,$44,$39,$19,$39,$19,$8E,$55,$D9,$E9,$5D
	.dc.b	$B8,$7F,$96,$D7,$C0,$AD,$46,$2D,$5B,$AF,$E6,$71,$E4,$7C,$27,$80
	.dc.b	$97,$E0,$7C,$8F,$81,$F2,$57,$5B,$EF,$1D,$2F,$B3,$E3,$38,$CF,$8E
	.dc.b	$33,$7E,$1B,$ED,$5E,$0A,$D3,$C0,$F3,$DF,$0B,$EF,$1D,$4E,$07,$84
	.dc.b	$E0,$66,$BD,$0E,$8C,$57,$A5,$F5,$9C,$4F,$87,$1E,$43,$89,$23,$84
	.dc.b	$69,$F3,$BD,$6F,$D9,$1A,$9E,$13,$D5,$E9,$19,$38,$BA,$8F,$29,$F0
	.dc.b	$0F,$29,$F9,$CA,$F8,$35,$DA,$3A,$57,$23,$A3,$C4,$F2,$4F,$0B,$F9
	.dc.b	$51,$F1,$AB,$93,$D4,$EE,$E0,$93,$F6,$27,$C2,$D9,$E4,$D7,$E2,$1E
	.dc.b	$42,$1E,$C2,$4F,$E5,$1E,$F3,$D1,$A7,$42,$9D,$0F,$31,$F3,$DE,$0D
	.dc.b	$38,$14,$E1,$1D,$DC,$5F,$D2,$7C,$B1,$98,$C4,$79,$AB,$D9,$18,$9D
	.dc.b	$5F,$CF,$AF,$9D,$39,$AD,$CF,$D2,$9F,$3C,$FD,$DA,$F9,$97,$F4,$A7
	.dc.b	$CA,$D9,$EC,$DA,$E5,$5F,$4A,$BE,$1D,$77,$6C,$EA,$FC,$31,$E4,$8F
	.dc.b	$33,$E2,$79,$DF,$58,$D1,$0D,$1F,$11,$FE,$95,$F4,$6F,$96,$A4,$9F
	.dc.b	$E6,$9F,$4F,$5F,$F6,$64,$ED,$FF,$7A,$B5,$FE,$71,$1F,$D5,$7E,$8D
	.dc.b	$F4,$97,$D7,$95,$FE,$FC,$9C,$B6,$72,$FF,$B6,$61,$3C,$E7,$A7,$CF
	.dc.b	$4A,$E4,$F7,$FA,$0C,$47,$18,$BB,$7F,$49,$C1,$58,$D7,$F9,$6C,$2B
	.dc.b	$17,$FE,$5C,$8D,$BF,$E5,$EB,$FE,$A5,$E8,$3E,$8F,$FC,$23,$3F,$E7
	.dc.b	$C6,$C6,$63,$FC,$2A,$D7,$FC,$86,$D9,$E7,$3F,$DE,$E3,$18,$7F,$40
	.dc.b	$FA,$47,$D6,$7F,$7F,$7F,$BD,$C6,$7A,$70,$67,$AA,$65,$E7,$9D,$34
	.dc.b	$FF,$3B,$D2,$B7,$0C,$BF,$B7,$4B,$E8,$BC,$04,$9E,$42,$1F,$25,$98
	.dc.b	$E5,$87,$A6,$B1,$7D,$2A,$CE,$6B,$9D,$F7,$6C,$FC,$3F,$63,$E5,$92
	.dc.b	$3E,$5B,$5E,$E7,$FA,$4B,$44,$71,$8F,$3F,$06,$1B,$CD,$FE,$29,$8F
	.dc.b	$E4,$A4,$E7,$C4,$49,$D9,$CA,$38,$7D,$2F,$1E,$19,$3D,$A3,$2F,$82
	.dc.b	$32,$F9,$8F,$C8,$C6,$9C,$5F,$E0,$1F,$F5,$5B,$5D,$9F,$43,$E9,$5F
	.dc.b	$0F,$F9,$06,$C7,$F9,$89,$AF,$FA,$41,$AF,$FA,$ED,$CE,$FF,$E2,$B6
	.dc.b	$EF,$FE,$0E,$6C,$FF,$AF,$3E,$95,$F1,$1F,$E4,$F7,$F4,$D1,$F8,$26
	.dc.b	$BE,$D6,$DF,$F8,$DD,$FD,$4C,$67,$9F,$91,$1E,$59,$7F,$96,$FD,$7F
	.dc.b	$F3,$56,$5F,$E2,$0B,$FF,$31,$51,$B3,$A8,$FF,$68,$B5,$D4,$F9,$CE
	.dc.b	$9E,$34,$8F,$44,$65,$FC,$DC,$F3,$C7,$FC,$BB,$F2,$3F,$E7,$36,$CF
	.dc.b	$EA,$5A,$FE,$41,$FF,$31,$FC,$9E,$FE,$93,$E0,$E1,$FC,$1D,$F8,$33
	.dc.b	$FE,$71,$0E,$23,$5A,$77,$F1,$3F,$45,$87,$57,$F5,$AF,$85,$AE,$7A
	.dc.b	$7C,$DE,$33,$E9,$6F,$BD,$2B,$D1,$39,$7D,$59,$3E,$97,$D0,$48,$CF
	.dc.b	$8D,$E3,$FA,$59,$1E,$17,$CD,$E7,$F8,$34,$93,$D9,$F1,$F9,$5F,$0F
	.dc.b	$D9,$F8,$9F,$19,$2B,$C3,$1E,$FF,$3F,$E0,$76,$49,$F7,$47,$D1,$FA
	.dc.b	$47,$D2,$DF,$42,$57,$D3,$9C,$BE,$CF,$94,$FA,$44,$7D,$FD,$FF,$37
	.dc.b	$E2,$A4,$7C,$B7,$E8,$7C,$47,$FA,$AF,$E4,$5F,$49,$FE,$AD,$FA,$57
	.dc.b	$C2,$53,$51,$BF,$86,$B8,$FF,$24,$BE,$2D,$8F,$F5,$B7,$D2,$3E,$0F
	.dc.b	$84,$E4,$FE,$57,$19,$7F,$96,$FD,$2F,$EC,$95,$FF,$A7,$DF,$E1,$4D
	.dc.b	$F0,$FF,$D4,$93,$B2,$72,$7F,$B1,$5E,$56,$D7,$C6,$66,$66,$1E,$28
	.dc.b	$FF,$66,$7D,$2F,$F0,$08,$F2,$CF,$A8,$CF,$04,$8F,$E8,$73,$B7,$F8
	.dc.b	$F3,$71,$FF,$5C,$61,$FE,$F0,$7F,$D7,$51,$9F,$13,$FD,$3A,$DA,$07
	.dc.b	$0A,$B9,$4F,$0D,$FE,$6F,$A5,$2B,$C7,$1E,$A2,$38,$3E,$7F,$2F,$FB
	.dc.b	$A7,$E9,$5E,$45,$24,$FF,$5E,$87,$D2,$E2,$3F,$9D,$47,$07,$53,$CE
	.dc.b	$7F,$DE,$98,$3E,$82,$92,$BF,$B4,$FD,$2F,$4D,$BD,$FF,$47,$FB,$CB
	.dc.b	$E1,$5F,$DB,$9A,$FF,$49,$6C,$FF,$B8,$EB,$94,$AB,$F8,$2C,$61,$E9
	.dc.b	$7F,$EC,$A6,$CF,$82,$FF,$D9,$B5,$B6,$E1,$FE,$03,$9E,$91,$FE,$D2
	.dc.b	$3C,$75,$CA,$32,$7D,$8A,$E4,$DA,$E5,$3B,$25,$76,$C3,$C5,$5B,$6B
	.dc.b	$FE,$0C,$C3,$D1,$3E,$16,$C6,$CF,$08,$C3,$C2,$F1,$39,$CF,$A0,$F0
	.dc.b	$3C,$23,$E4,$BF,$EE,$E3,$96,$1F,$05,$C4,$69,$E2,$63,$E9,$5F,$4B
	.dc.b	$C6,$3E,$14,$61,$C4,$A3,$11,$C6,$7F,$C9,$2B,$E9,$69,$1F,$F2,$6B
	.dc.b	$73,$FE,$83,$9F,$E6,$38,$3E,$3D,$7E,$A2,$BE,$4C,$E6,$3E,$29,$93
	.dc.b	$E3,$B9,$AF,$E9,$71,$97,$AE,$AB,$EB,$3E,$53,$8C,$65,$EA,$66,$78
	.dc.b	$CF,$0B,$F1,$D6,$E7,$BD,$6F,$4F,$DF,$BF,$E5,$19,$D4,$7C,$BA,$E3
	.dc.b	$7C,$EB,$C2,$E1,$FF,$16,$61,$EF,$3C,$8F,$49,$D4,$60,$D4,$F8,$17
	.dc.b	$9F,$FD,$D6,$31,$85,$BF,$F4,$04,$D5,$F4,$BF,$BD,$AF,$F3,$1B,$63
	.dc.b	$FC,$7C,$FF,$1A,$8C,$EF,$C6,$35,$87,$F8,$0F,$AB,$C7,$7F,$C8,$27
	.dc.b	$37,$FE,$8F,$9C,$CF,$83,$9F,$C9,$6F,$A5,$7C,$B7,$59,$CC,$63,$0F
	.dc.b	$2C,$62,$75,$1F,$E1,$B9,$D5,$FA,$3E,$97,$CC,$76,$FA,$5F,$E9,$15
	.dc.b	$EE,$8C,$BF,$E2,$38,$7C,$16,$2B,$FC,$49,$18,$9E,$06,$E7,$59,$7F
	.dc.b	$88,$E7,$55,$FE,$65,$F8,$FF,$ED,$F7,$FE,$8E,$6C,$7F,$9A,$7E,$95
	.dc.b	$E5,$FD,$39,$CE,$1F,$80,$FA,$47,$C0,$FD,$76,$E3,$CF,$BB,$13,$F6
	.dc.b	$F1,$79,$34,$CD,$7F,$A7,$A3,$2F,$FA,$83,$E9,$5E,$07,$E0,$DB,$1F
	.dc.b	$EA,$4C,$5E,$BE,$6B,$5C,$B2,$B9,$72,$DE,$0F,$F4,$27,$CE,$7F,$7E
	.dc.b	$9F,$F5,$6B,$63,$FD,$11,$E2,$FF,$85,$27,$3E,$47,$E4,$DF,$E1,$87
	.dc.b	$FA,$D0,$B3,$FE,$55,$FA,$57,$D5,$6F,$C3,$FC,$D7,$DB,$FE,$56,$BF
	.dc.b	$D2,$8D,$E5,$FF,$2B,$5F,$C0,$9C,$9F,$1F,$EA,$66,$3F,$E6,$3F,$A7
	.dc.b	$FC,$AF,$22,$7A,$F6,$4A,$FF,$34,$7C,$83,$E4,$31,$8F,$FA,$8F,$C5
	.dc.b	$F7,$69,$FB,$86,$DF,$FB,$BF,$E9,$FF,$43,$18,$AF,$F0,$2F,$E2,$0F
	.dc.b	$77,$F6,$8D,$BF,$F3,$B5,$BE,$9B,$35,$FB,$9F,$B7,$7D,$55,$3F,$57
	.dc.b	$F5,$5F,$DF,$B0,$FF,$6E,$CE,$9B,$F2,$3C,$4D,$E1,$63,$FD,$C7,$F8
	.dc.b	$1B,$C5,$FE,$0C,$D8,$FF,$44,$7D,$5F,$E7,$B8,$3F,$8A,$CE,$77,$FF
	.dc.b	$62,$7D,$03,$DE,$FE,$1C,$DC,$3F,$D2,$19,$A3,$33,$FC,$4F,$F0,$0F
	.dc.b	$BB,$5F,$EF,$46,$DF,$FA,$56,$FC,$13,$FC,$BA,$73,$8F,$F2,$6F,$2D
	.dc.b	$F0,$FF,$C7,$5B,$5F,$E9,$FF,$AB,$FA,$18,$A7,$2F,$FB,$6F,$EA,$FF
	.dc.b	$84,$E1,$FC,$C1,$B7,$FE,$3A,$87,$D3,$FE,$16,$30,$7F,$40,$9D,$D7
	.dc.b	$F9,$01,$A9,$CD,$72,$EC,$FE,$65,$1A,$33,$F2,$2F,$1B,$3E,$9F,$F2
	.dc.b	$68,$C6,$3F,$EB,$9F,$A6,$7D,$5F,$F2,$A5,$F0,$96,$30,$D9,$E0,$DD
	.dc.b	$3F,$D8,$15,$C2,$7F,$D8,$9A,$BE,$AF,$F9,$AB,$63,$3E,$A6,$26,$47
	.dc.b	$07,$71,$8B,$6F,$FD,$6B,$EA,$DF,$19,$FD,$45,$B8,$7A,$F9,$B1,$32
	.dc.b	$3E,$6D,$FF,$9A,$AB,$6F,$A2,$B4,$F6,$9F,$EE,$33,$AB,$FF,$6B,$7D
	.dc.b	$5F,$F4,$0A,$D1,$CF,$0F,$EA,$31,$99,$FF,$6E,$66,$FA,$BF,$82,$7F
	.dc.b	$56,$F4,$BF,$3D,$3A,$8C,$E5,$F2,$1D,$77,$63,$2F,$9B,$F2,$1F,$54
	.dc.b	$FA,$DF,$21,$6F,$ED,$45,$9F,$A7,$37,$0F,$F0,$0F,$42,$BF,$C0,$75
	.dc.b	$FE,$0E,$8C,$5F,$F8,$DE,$3F,$8C,$4E,$9F,$F4,$8C,$E9,$FF,$1E,$7D
	.dc.b	$6F,$ED,$DF,$F2,$33,$5F,$F0,$CB,$5F,$F4,$CB,$57,$D5,$BE,$B3,$F8
	.dc.b	$34,$7F,$91,$E3,$19,$F8,$D7,$D1,$7E,$FE,$75,$5C,$1F,$F2,$B7,$CA
	.dc.b	$3E,$73,$FC,$35,$A7,$F8,$AA,$31,$3F,$82,$0E,$1F,$B6,$3E,$AF,$FC
	.dc.b	$49,$3C,$6B,$9C,$FF,$AC,$AB,$86,$DF,$E2,$C6,$D7,$F8,$E2,$F8,$D6
	.dc.b	$FE,$0F,$8B,$E0,$6D,$7F,$81,$A3,$C8,$BE,$97,$33,$27,$2F,$08,$E7
	.dc.b	$F5,$6F,$9C,$E7,$F1,$92,$FD,$D3,$96,$CE,$5F,$7E,$1F,$E6,$8C,$BF
	.dc.b	$C9,$F5,$B9,$F8,$6D,$CF,$DD,$87,$17,$FC,$97,$19,$AF,$F2,$BF,$D5
	.dc.b	$FE,$36,$FE,$73,$5F,$9A,$DA,$FF,$29,$BF,$EF,$79,$E3,$C9,$F1,$A3
	.dc.b	$51,$8A,$E7,$CB,$9F,$87,$D7,$D3,$49,$EF,$19,$7C,$31,$97,$C3,$1F
	.dc.b	$4B,$EA,$FF,$9E,$C7,$9B,$E9,$FC,$06,$7C,$C9,$19,$21,$F8,$69,$E2
	.dc.b	$D8,$E2,$EB,$4F,$F3,$7C,$E9,$B3,$A9,$F8,$75,$D9,$E4,$71,$CB,$E5
	.dc.b	$C6,$6B,$8C,$E4,$0C,$D3,$BA,$FF,$44,$E1,$E1,$79,$D7,$11,$19,$79
	.dc.b	$B5,$E9,$3E,$7B,$F0,$C7,$98,$F4,$9C,$E7,$81,$C6,$FC,$71,$9C,$3C
	.dc.b	$8F,$A2,$32,$F2,$3A,$57,$63,$F0,$B7,$E1,$C3,$D7,$7C,$AB,$CA,$75
	.dc.b	$3F,$79,$87,$86,$B8,$47,$B2,$78,$61,$E0,$C3,$AC,$F0,$9C,$D7,$08
	.dc.b	$F8,$B1,$87,$B1,$E6,$C3,$E5,$BE,$9A,$F4,$47,$9A,$3E,$75,$F7,$8C
	.dc.b	$47,$13,$E1,$5F,$7C,$3E,$4E,$1E,$1B,$E1,$5E,$93,$CC,$74,$AF,$85
	.dc.b	$5E,$F8,$F2,$E1,$E2,$7A,$98,$C3,$C3,$1F,$02,$3D,$11,$82,$3C,$0C
	.dc.b	$9E,$13,$D0,$7D,$03,$27,$C6,$3C,$A6,$2F,$D2,$E4,$E2,$E0,$CD,$7C
	.dc.b	$08,$F6,$1C,$AB,$D4,$F7,$7D,$06,$23,$E0,$47,$89,$E4,$F8,$6F,$E2
	.dc.b	$E1,$F2,$DB,$1D,$8F,$5C,$F2,$AE,$B8,$7C,$29,$F4,$46,$1F,$88,$F8
	.dc.b	$6F,$A4,$79,$E7,$86,$1E,$DC,$3B,$CF,$8C,$D1,$F2,$EF,$A3,$ED,$AF
	.dc.b	$06,$1E,$37,$F1,$B3,$DE,$7D,$0D,$7A,$1E,$AC,$3C,$6D,$37,$3D,$23
	.dc.b	$33,$C4,$F0,$E1,$F4,$A3,$F3,$D5,$F2,$DB,$3E,$6B,$FE,$09,$F5,$5F
	.dc.b	$E9,$7A,$FE,$D4,$BD,$CF,$EB,$AD,$8C,$FB,$98,$33,$32,$32,$72,$FA
	.dc.b	$4E,$24,$7A,$FD,$35,$7F,$BA,$DA,$71,$8B,$4F,$E4,$74,$A5,$FF,$70
	.dc.b	$EB,$FC,$9C,$7F,$74,$9D,$56,$76,$91,$EC,$8F,$59,$87,$FC,$0B,$5F
	.dc.b	$C6,$1E,$E7,$F8,$52,$7F,$C5,$0D,$3F,$C5,$51,$B7,$CB,$F5,$3F,$DF
	.dc.b	$DF,$F0,$7B,$FD,$FD,$FF,$0A,$35,$CF,$4F,$F8,$22,$38,$65,$EF,$9F
	.dc.b	$E8,$31,$B8,$C4,$23,$18,$78,$2B,$97,$D3,$4A,$EE,$F5,$E7,$C4,$79
	.dc.b	$22,$E1,$D3,$0F,$F0,$34,$FF,$50,$C5,$E4,$7F,$87,$35,$FE,$F5,$3F
	.dc.b	$D5,$1A,$FF,$8E,$23,$E9,$1B,$E6,$91,$FE,$33,$AE,$6D,$9D,$B4,$FF
	.dc.b	$1C,$56,$9F,$F1,$BA,$E5,$F1,$12,$37,$CD,$5D,$C5,$CB,$E3,$D2,$BD
	.dc.b	$5F,$53,$EB,$E0,$98,$78,$23,$2F,$8A,$32,$F8,$A3,$FC,$7B,$F5,$2F
	.dc.b	$8C,$FF,$1F,$FD,$4F,$E8,$EB,$FC,$9A,$DC,$BE,$BD,$6E,$7F,$CA,$8D
	.dc.b	$77,$CD,$C8,$EB,$72,$36,$D7,$AF,$C6,$F8,$39,$E4,$7F,$97,$F1,$FD
	.dc.b	$43,$75,$CB,$E4,$9B,$97,$E2,$AB,$F5,$33,$FE,$67,$F4,$FF,$CB,$AF
	.dc.b	$F8,$42,$3F,$54,$7E,$AE,$33,$06,$C6,$B9,$FF,$CC,$31,$AF,$C7,$DF
	.dc.b	$65,$8F,$0E,$61,$FE,$75,$9C,$B6,$72,$FF,$9D,$AB,$F3,$AF,$F8,$8A
	.dc.b	$76,$FF,$9E,$9B,$8F,$E9,$1B,$87,$F0,$66,$CF,$F8,$AA,$9F,$45,$FE
	.dc.b	$76,$FB,$AF,$F6,$C4,$FF,$A1,$FE,$EB,$E4,$BF,$E8,$66,$E3,$FB,$09
	.dc.b	$DC,$FF,$A2,$BE,$E9,$F2,$3F,$E8,$0F,$B9,$F8,$A7,$C6,$8C,$CF,$FA
	.dc.b	$4E,$1E,$E7,$FA,$3B,$0F,$F5,$B7,$83,$7C,$7F,$E0,$3B,$FF,$1E,$BD
	.dc.b	$E1,$FE,$9B,$FB,$A7,$D6,$77,$F2,$E4,$E6,$3F,$D2,$32,$F9,$0F,$27
	.dc.b	$C3,$4A,$E5,$1F,$E9,$6F,$BA,$7C,$57,$87,$A2,$57,$FA,$A6,$72,$FF
	.dc.b	$AA,$FE,$35,$DB,$C7,$C3,$93,$98,$FF,$50,$FD,$D2,$BB,$78,$72,$BF
	.dc.b	$D6,$93,$97,$FD,$6D,$E6,$FE,$D5,$FF,$7C,$F2,$FF,$B6,$75,$F8,$87
	.dc.b	$FA,$B2,$75,$3F,$EB,$AA,$F2,$BF,$EB,$CF,$A3,$7C,$FF,$FA,$C9,$B1
	.dc.b	$E0,$DF,$FD,$6B,$F7,$4F,$8D,$D7,$99,$2B,$FD,$91,$39,$7F,$D9,$5F
	.dc.b	$81,$FD,$45,$AF,$FB,$B7,$E0,$3E,$C1,$FE,$C0,$9D,$4B,$E7,$1F,$5D
	.dc.b	$FE,$C4,$FA,$F7,$C4,$53,$FD,$8F,$3A,$AF,$F6,$C2,$F8,$BC,$DC,$9C
	.dc.b	$C7,$FB,$3B,$EB,$59,$72,$89,$3F,$EE,$19,$CB,$FE,$E3,$FA,$FF,$E2
	.dc.b	$1A,$7F,$4A,$C7,$F9,$C9,$FC,$EA,$33,$E0,$FF,$6E,$6D,$7F,$A4,$9A
	.dc.b	$9B,$7F,$AC,$8F,$07,$93,$FE,$E2,$9D,$9A,$7F,$DC,$BF,$5F,$F8,$A9
	.dc.b	$F3,$7A,$3E,$B3,$3D,$92,$3F,$DF,$53,$B7,$FD,$F2,$DC,$FF,$B1,$E2
	.dc.b	$FB,$1E,$15,$AB,$B5,$A9,$F0,$F0,$7C,$9F,$6E,$5F,$68,$CB,$48,$7E
	.dc.b	$03,$5F,$F0,$8E,$2F,$A3,$FF,$7F,$FD,$83,$AA,$92,$FF,$D7,$39,$FF
	.dc.b	$78,$97,$D8,$66,$7A,$1F,$EB,$E8,$E4,$F0,$C7,$FC,$3F,$18,$8D,$87
	.dc.b	$9B,$B3,$FD,$8F,$9B,$EB,$3F,$C4,$F8,$7F,$82,$23,$13,$B0,$E1,$EB
	.dc.b	$F6,$1E,$69,$FF,$68,$FD,$7E,$A7,$AD,$F4,$C7,$FC,$67,$87,$1B,$ED
	.dc.b	$39,$AF,$F6,$CE,$1C,$D6,$1C,$D4,$66,$31,$7E,$36,$D7,$F9,$01,$E1
	.dc.b	$3B,$9F,$F2,$1B,$6B,$CA,$7A,$23,$CB,$5C,$5D,$E5,$FB,$A9,$F1,$C6
	.dc.b	$30,$E1,$1C,$8C,$97,$0F,$0D,$71,$6C,$F0,$9F,$97,$7D,$DE,$17,$D5
	.dc.b	$E6,$7C,$0B,$E5,$1E,$68,$E4,$FC,$13,$85,$F6,$8F,$F2,$99,$C6,$EB
	.dc.b	$4C,$E1,$F3,$DF,$F2,$CB,$FD,$5B,$EC,$36,$D8,$DF,$A4,$9C,$B6,$3F
	.dc.b	$CB,$FF,$49,$FE,$60,$4D,$23,$A3,$DB,$7F,$A3,$D3,$E2,$40,$E2,$6B
	.dc.b	$6F,$F2,$F6,$5F,$83,$8D,$5F,$CB,$BE,$51,$F2,$0C,$9C,$6F,$AB,$D2
	.dc.b	$BF,$19,$3C,$1E,$91,$9C,$FF,$47,$87,$81,$F9,$78,$77,$8C,$4F,$FA
	.dc.b	$02,$31,$87,$66,$C7,$8C,$E8,$7E,$FA,$78,$BF,$D1,$9F,$F3,$D6,$1F
	.dc.b	$E7,$83,$D6,$7F,$9E,$A3,$D7,$87,$A2,$77,$7E,$63,$F6,$D4,$FB,$0C
	.dc.b	$DF,$79,$CC,$62,$1F,$61,$FC,$06,$7F,$7C,$D8,$FE,$12,$FD,$38,$CE
	.dc.b	$FC,$63,$57,$E8,$EC,$F3,$D7,$FA,$C2,$72,$FF,$1C,$8C,$DF,$DD,$9F
	.dc.b	$F3,$24,$FF,$7A,$FB,$0F,$F0,$FB,$E2,$E9,$F4,$E0,$9C,$DF,$92,$73
	.dc.b	$18,$9F,$F4,$FD,$F4,$AF,$E4,$F1,$C7,$D5,$F4,$79,$3F,$2E,$7C,$4D
	.dc.b	$7E,$1B,$6B,$FC,$49,$8F,$EA,$1B,$3D,$3E,$79,$EB,$74,$AF,$F0,$9C
	.dc.b	$61,$F9,$77,$EA,$8C,$57,$CC,$8C,$57,$D3,$F9,$2F,$EF,$F7,$FB,$66
	.dc.b	$D7,$EE,$9A,$FF,$B3,$9A,$FF,$B7,$BE,$BB,$FD,$41,$3F,$BE,$6C,$7F
	.dc.b	$02,$6B,$FE,$D6,$6B,$FE,$E6,$FA,$97,$D8,$7C,$36,$D7,$F0,$86,$CF
	.dc.b	$F0,$C6,$BF,$EE,$0F,$B0,$FE,$BD,$3F,$C3,$9B,$1F,$C4,$9A,$EF,$B7
	.dc.b	$E8,$47,$F8,$52,$78,$71,$60,$8F,$F1,$8F,$A1,$FF,$13,$37,$C6,$FB
	.dc.b	$11,$8C,$3F,$8F,$B5,$FF,$79,$CF,$F9,$C0,$FE,$42,$57,$F9,$2B,$5D
	.dc.b	$F5,$FE,$0F,$EB,$BF,$41,$1F,$CA,$9A,$FF,$BF,$9A,$FF,$17,$0E,$1F
	.dc.b	$E3,$59,$BE,$6B,$E6,$12,$FE,$63,$DF,$86,$47,$F3,$6B,$FC,$7D,$F7
	.dc.b	$9F,$CB,$76,$7B,$63,$C0,$D7,$F9,$1E,$7A,$3F,$39,$F5,$8F,$AB,$A4
	.dc.b	$C7,$FA,$1C,$65,$FE,$8B,$F5,$5F,$E7,$36,$FB,$1F,$A5,$8C,$61,$FD
	.dc.b	$31,$AF,$C1,$0B,$E3,$F8,$12,$BA,$BF,$AF,$FA,$97,$3D,$24,$FF,$56
	.dc.b	$8C,$BF,$D6,$3D,$EF,$F3,$93,$7D,$6F,$DC,$46,$31,$FE,$BE,$D7,$C9
	.dc.b	$F5,$2F,$AF,$FD,$F3,$6F,$FB,$1B,$67,$FB,$3B,$5F,$67,$6B,$E6,$3F
	.dc.b	$83,$36,$BF,$B5,$B6,$7F,$B7,$B5,$F8,$9F,$53,$FC,$56,$7F,$0C,$D8
	.dc.b	$FF,$42,$FD,$1F,$E1,$67,$FC,$5D,$39,$93,$FC,$81,$E7,$E1,$7F,$17
	.dc.b	$8C,$47,$48,$C1,$FC,$A3,$E3,$1D,$9F,$CF,$EB,$C1,$E2,$7F,$BC,$E3
	.dc.b	$F9,$6F,$D4,$3E,$A6,$FB,$38,$6E,$7F,$EA,$E9,$D3,$67,$4D,$3F,$D6
	.dc.b	$7A,$3E,$A7,$F9,$CB,$63,$3E,$AE,$26,$47,$07,$71,$8D,$BF,$CB,$3F
	.dc.b	$52,$FA,$9F,$EB,$0D,$BF,$27,$4E,$26,$47,$F8,$7A,$75,$5F,$E2,$26
	.dc.b	$BF,$97,$6B,$FB,$97,$7B,$7F,$72,$FA,$9F,$F8,$03,$5F,$E6,$19,$CB
	.dc.b	$FE,$76,$8C,$D7,$F4,$EC,$DF,$53,$F0,$0F,$EB,$3E,$3F,$E6,$A7,$51
	.dc.b	$9B,$F8,$EE,$BB,$31,$3F,$D6,$38,$1F,$21,$F5,$8F,$AF,$F3,$B7,$0F
	.dc.b	$A9,$16,$B8,$4D,$D5,$5F,$C5,$7F,$0D,$F2,$BF,$E4,$26,$D7,$F9,$19
	.dc.b	$AF,$EE,$DA,$FF,$1D,$FA,$97,$D4,$FE,$E2,$3F,$C9,$B1,$8C,$FF,$C9
	.dc.b	$AB,$E2,$1E,$06,$2F,$85,$FF,$07,$FB,$8F,$F5,$93,$FE,$54,$6B,$FE
	.dc.b	$58,$6B,$BF,$81,$FF,$2C,$AF,$94,$7D,$57,$FA,$26,$30,$D8,$CC,$7F
	.dc.b	$A6,$7E,$AB,$FC,$59,$B7,$F3,$79,$CC,$FF,$A9,$9B,$BF,$C8,$8F,$F2
	.dc.b	$94,$62,$7F,$9F,$57,$29,$CF,$AF,$DC,$8F,$FA,$9E,$9F,$29,$88,$ED
	.dc.b	$37,$0E,$D1,$EB,$9F,$57,$D4,$BE,$AB,$D5,$F0,$52,$BF,$D6,$53,$96
	.dc.b	$CF,$C4,$8C,$47,$FA,$01,$AF,$F3,$56,$BF,$D5,$DD,$DF,$89,$B9,$E6
	.dc.b	$BE,$57,$B3,$F3,$7E,$AB,$F2,$78,$7F,$B0,$DA,$FF,$B3,$9B,$87,$B1
	.dc.b	$73,$F1,$E8,$95,$C5,$E9,$1D,$7E,$9B,$B7,$CB,$CA,$ED,$19,$7C,$11
	.dc.b	$97,$C1,$1F,$81,$FA,$AF,$E6,$D1,$E5,$FA,$7F,$41,$9F,$3A,$46,$48
	.dc.b	$7F,$B9,$A7,$8B,$63,$8B,$AC,$FE,$BC,$E9,$B3,$A9,$F8,$55,$D9,$E4
	.dc.b	$71,$DB,$FC,$0D,$A2,$75,$7C,$AB,$47,$FA,$86,$74,$FF,$BC,$70,$F0
	.dc.b	$5F,$F8,$3D,$B1,$BB,$F1,$5F,$38,$F0,$9E,$43,$A8,$9E,$2D,$C3,$8D
	.dc.b	$F9,$AF,$CA,$F7,$8C,$CF,$73,$C8,$E9,$C1,$DC,$F0,$DF,$8C,$FC,$25
	.dc.b	$72,$8F,$01,$C8,$F2,$E1,$F0,$1E,$37,$E1,$73,$3D,$DE,$27,$7A,$EA
	.dc.b	$F3,$38,$9E,$47,$2E,$0E,$2E,$09,$3D,$E3,$55,$E3,$9C,$C7,$FB,$57
	.dc.b	$0F,$92,$7C,$A9,$D5,$7C,$99,$F1,$57,$C1,$8C,$47,$9D,$ED,$19,$8F
	.dc.b	$34,$6A,$31,$1E,$68,$C1,$1F,$16,$1F,$0E,$BC,$31,$C8,$C9,$E2,$9E
	.dc.b	$0A,$36,$F5,$36,$7F,$8E,$CF,$9A,$7F,$8F,$6F,$F4,$AD,$AF,$E6,$71
	.dc.b	$88,$F9,$06,$4E,$47,$23,$DF,$51,$EF,$7F,$C1,$FE,$E3,$EE,$57,$CC
	.dc.b	$BF,$E2,$11,$87,$DC,$DA,$F6,$1F,$A7,$70,$FD,$86,$D7,$CA,$C1,$18
	.dc.b	$C3,$C5,$5F,$0E,$30,$F9,$23,$F6,$F1,$88,$F8,$C6,$CF,$09,$8C,$3F
	.dc.b	$05,$1B,$6C,$69,$F6,$3F,$CA,$63,$58,$7B,$A3,$4F,$B9,$A7,$6B,$F4
	.dc.b	$B8,$3D,$C5,$C3,$DD,$5D,$23,$B1,$9C,$3D,$F5,$CA,$3F,$45,$87,$91
	.dc.b	$FE,$75,$5C,$5A,$71,$3E,$21,$F3,$CC,$1F,$12,$7A,$5F,$E6,$2B,$A3
	.dc.b	$F7,$64,$1F,$EC,$3F,$56,$F9,$AF,$EF,$4B,$EA,$29,$88,$95,$CA,$B8
	.dc.b	$CE,$A3,$19,$7F,$AC,$EF,$A4,$60,$F0,$AF,$A7,$EE,$48,$D0,$87,$D4
	.dc.b	$66,$3D,$15,$D6,$B1,$87,$AA,$BF,$C2,$5E,$07,$EB,$30,$FF,$03,$61
	.dc.b	$CF,$0C,$EF,$23,$F4,$B5,$FC,$7D,$FF,$55,$35,$FF,$66,$35,$F9,$2F
	.dc.b	$EE,$09,$1F,$E2,$1F,$AA,$7D,$C5,$EB,$C7,$5C,$C3,$DB,$49,$D4,$62
	.dc.b	$74,$FB,$8F,$AF,$3B,$78,$FD,$67,$72,$3D,$3B,$2B,$C4,$4A,$73,$B7
	.dc.b	$F8,$82,$15,$F2,$3C,$EF,$F1,$D6,$FF,$E0,$8C,$3E,$9B,$5E,$9A,$FF
	.dc.b	$B3,$5F,$AE,$D7,$F8,$C3,$5F,$97,$3F,$E4,$88,$C4,$7C,$C8,$F0,$3C
	.dc.b	$9F,$09,$E2,$E7,$91,$E2,$7C,$BD,$18,$8F,$75,$A7,$19,$7F,$8F,$21
	.dc.b	$5E,$CF,$A1,$F6,$F0,$4C,$3D,$B1,$97,$DF,$19,$7A,$C7,$F9,$77,$EB
	.dc.b	$1F,$1D,$FE,$5F,$FA,$CF,$D3,$57,$E6,$1A,$FD,$19,$DB,$F9,$C6,$BF
	.dc.b	$CF,$5A,$FF,$BE,$5A,$FE,$D8,$38,$7F,$72,$F2,$CB,$FC,$B2,$DE,$4E
	.dc.b	$0D,$9F,$DF,$D3,$E6,$7D,$1C,$F2,$7D,$0F,$AF,$5C,$47,$E2,$6B,$F6
	.dc.b	$CE,$8F,$74,$7F,$9F,$FD,$AF,$F3,$A4,$76,$C7,$E2,$19,$F1,$A4,$7C
	.dc.b	$58,$CB,$FB,$E9,$E2,$BE,$AD,$F5,$9D,$7A,$24,$FF,$04,$9C,$B6,$7F
	.dc.b	$D2,$71,$88,$FE,$D0,$D7,$E5,$1D,$63,$67,$FA,$5E,$38,$6D,$FE,$82
	.dc.b	$6E,$1F,$47,$07,$CE,$FF,$A7,$7E,$AF,$FD,$D9,$8F,$F6,$C6,$BF,$DD
	.dc.b	$DA,$F8,$1A,$FD,$BF,$AB,$7C,$87,$FA,$A9,$B1,$FD,$D9,$AF,$F8,$05
	.dc.b	$AF,$AB,$EA,$DF,$05,$FE,$B2,$6C,$7F,$7E,$6B,$FE,$09,$6B,$F0,$BE
	.dc.b	$A9,$F5,$7F,$EB,$A6,$C7,$F8,$19,$AF,$F8,$45,$AF,$1F,$90,$E0,$47
	.dc.b	$83,$FE,$C4,$FA,$B7,$D4,$D2,$4F,$F8,$4A,$32,$FF,$85,$3E,$3B,$F2
	.dc.b	$8D,$E9,$FF,$66,$C6,$23,$FC,$44,$D7,$ED,$FD,$5A,$3F,$DA,$6D,$8F
	.dc.b	$F0,$FB,$5F,$F1,$53,$5F,$C1,$F9,$5F,$C1,$1F,$F1,$2B,$5F,$F1,$83
	.dc.b	$5D,$F8,$1D,$B1,$91,$D9,$AF,$F8,$DF,$EA,$9F,$1D,$FE,$E6,$6C,$7F
	.dc.b	$8C,$9A,$EE,$3F,$DD,$7F,$55,$FC,$F6,$3F,$C8,$0D,$7F,$32,$D7,$F8
	.dc.b	$23,$5F,$EF,$7F,$28,$FA,$9F,$F7,$AB,$63,$FC,$82,$D7,$FC,$98,$D7
	.dc.b	$F4,$AB,$E2,$F8,$92,$38,$BF,$F0,$03,$E9,$DF,$0D,$49,$3F,$E4,$E8
	.dc.b	$CB,$FE,$50,$FA,$9F,$E3,$09,$A4,$F1,$FF,$CA,$B1,$97,$F9,$07,$81
	.dc.b	$FE,$F4,$9F,$E3,$53,$93,$FC,$0B,$5F,$B4,$6F,$6F,$94,$8F,$8D,$EB
	.dc.b	$F5,$3F,$89,$8F,$E5,$8D,$7F,$C4,$6D,$4D,$3F,$25,$6F,$03,$8F,$37
	.dc.b	$99,$B5,$E4,$F8,$0C,$3D,$26,$BD,$1B,$7D,$2B,$F9,$F6,$2B,$CF,$C6
	.dc.b	$FA,$09,$87,$F9,$EA,$D1,$88,$FE,$FD,$18,$7F,$C5,$2D,$7F,$C0,$54
	.dc.b	$F7,$7C,$7F,$52,$FA,$5A,$4A,$FF,$18,$C3,$EA,$71,$1F,$E0,$B8,$E0
	.dc.b	$EA,$79,$CF,$BB,$D4,$7D,$2D,$25,$7F,$8F,$3E,$A7,$A7,$07,$F9,$4B
	.dc.b	$EA,$3F,$C2,$33,$FE,$54,$BE,$2F,$F9,$09,$AF,$F8,$6D,$B3,$F4,$E3
	.dc.b	$FC,$B7,$5D,$15,$74,$51,$98,$C5,$74,$6D,$FF,$6F,$78,$4E,$E7,$FB
	.dc.b	$93,$6B,$CD,$5B,$9F,$EE,$C7,$33,$78,$7F,$9B,$67,$C7,$18,$C3,$E0
	.dc.b	$3C,$2B,$94,$F4,$32,$7B,$2B,$96,$1E,$46,$D7,$19,$D9,$FE,$3A,$AE
	.dc.b	$6F,$C0,$9D,$BC,$EF,$75,$E7,$8F,$01,$E6,$34,$53,$94,$7F,$B2,$B0
	.dc.b	$F9,$D1,$FD,$B1,$AF,$B4,$FF,$6C,$3F,$D7,$BE,$A7,$87,$92,$69,$27
	.dc.b	$F8,$0F,$D0,$7F,$02,$4D,$23,$A3,$C9,$7F,$99,$E3,$F8,$31,$FE,$9C
	.dc.b	$9F,$87,$39,$8F,$51,$93,$8E,$BF,$89,$D7,$F0,$D1,$97,$F1,$13,$C5
	.dc.b	$CC,$79,$0E,$37,$E0,$71,$1C,$2B,$A6,$7F,$EF,$9C,$3C,$71,$8A,$FF
	.dc.b	$09,$C6,$2F,$93,$63,$CE,$74,$AF,$E4,$75,$C6,$3F,$D7,$0F,$E8,$70
	.dc.b	$FD,$14,$62,$FF,$D5,$2F,$AE,$B8,$C7,$FA,$B2,$7E,$21,$C0,$7D,$4F
	.dc.b	$38,$DD,$7F,$A2,$67,$31,$FE,$B4,$FA,$8F,$DE,$B7,$7E,$15,$FE,$CC
	.dc.b	$9F,$F6,$6C,$E4,$91,$FE,$9B,$AF,$E9,$38,$F0,$85,$74,$75,$D5,$E6
	.dc.b	$BF,$F3,$14,$FF,$9C,$1F,$F6,$55,$BE,$A7,$FC,$47,$1E,$5E,$BF,$1C
	.dc.b	$E7,$CC,$ED,$7C,$71,$F1,$61,$C1,$FF,$53,$46,$6B,$E6,$4E,$5F,$9B
	.dc.b	$0F,$27,$AF,$CE,$FF,$AE,$72,$F8,$91,$8F,$80,$EE,$D8,$C9,$E8,$9E
	.dc.b	$18,$F8,$27,$4D,$9D,$57,$C5,$F9,$2F,$E7,$F1,$FE,$E4,$FA,$9F,$F4
	.dc.b	$F3,$FE,$E6,$8C,$D7,$FA,$FE,$72,$DF,$7B,$E7,$7D,$3B,$EA,$7E,$85
	.dc.b	$6B,$5F,$C4,$CE,$71,$FF,$78,$B7,$0F,$F6,$57,$E2,$BF,$ED,$C9,$CE
	.dc.b	$1F,$66,$31,$3F,$EC,$FF,$AA,$FE,$9D,$5F,$EF,$A7,$E6,$3F,$E9,$B7
	.dc.b	$31,$BE,$2F,$C1,$47,$F8,$32,$7F,$E0,$0F,$89,$FE,$1B,$79,$72,$7E
	.dc.b	$1E,$31,$7F,$ED,$DF,$A1,$79,$3F,$8B,$FC,$57,$FD,$B7,$82,$B5,$39
	.dc.b	$EB,$FC,$84,$FF,$9C,$8F,$C5,$FD,$AF,$FA,$41,$E1,$C1,$FA,$C7,$57
	.dc.b	$F9,$99,$D3,$FE,$ED,$FA,$BF,$CF,$B9,$CF,$F3,$91,$8B,$FD,$B5,$7F
	.dc.b	$AC,$6F,$E6,$FD,$67,$E5,$BE,$D7,$F5,$B8,$FF,$BD,$F8,$D3,$9E,$FE
	.dc.b	$7D,$3F,$9C,$F8,$F1,$8C,$7F,$46,$E1,$F4,$9F,$EC,$9F,$07,$E2,$BF
	.dc.b	$A8,$FB,$5E,$FA,$FF,$08,$9C,$E7,$FB,$59,$CD,$7F,$B4,$7E,$B5,$F5
	.dc.b	$DF,$AC,$FC,$4B,$D9,$39,$8C,$5F,$EE,$FE,$BF,$FC,$E9,$3F,$B6,$FB
	.dc.b	$6F,$93,$9F,$F1,$69,$CD,$7F,$03,$9C,$D7,$FB,$7E,$5F,$21,$FB,$AF
	.dc.b	$B6,$EF,$7F,$6E,$73,$5F,$C0,$A7,$32,$FB,$17,$87,$C7,$E2,$FF,$87
	.dc.b	$46,$2F,$84,$7F,$11,$FA,$4F,$F3,$BC,$7F,$0C,$CB,$F8,$96,$9F,$CB
	.dc.b	$27,$38,$FF,$19,$9C,$D7,$FB,$D7,$EC,$1F,$63,$FC,$42,$58,$FF,$1D
	.dc.b	$9C,$D7,$58,$FE,$41,$E4,$BE,$7B,$F8,$BF,$DB,$3E,$F7,$31,$87,$F9
	.dc.b	$37,$D9,$FF,$13,$C3,$F3,$9A,$23,$0B,$67,$D7,$7E,$2A,$BF,$07,$3F
	.dc.b	$A3,$AD,$BC,$7D,$1F,$E4,$FF,$22,$F2,$3B,$C6,$1B,$7F,$CC,$27,$4F
	.dc.b	$4A,$E4,$E6,$7F,$98,$FD,$6B,$EC,$EF,$C2,$61,$B5,$FC,$EA,$74,$D9
	.dc.b	$D3,$4F,$E7,$F9,$BE,$CF,$F9,$EB,$63,$3F,$05,$89,$91,$C1,$DC,$63
	.dc.b	$2F,$8D,$3F,$2B,$EC,$9F,$67,$FD,$AA,$70,$D7,$F8,$54,$6A,$7E,$8C
	.dc.b	$69,$AE,$A7,$FA,$8F,$D9,$BE,$1F,$FA,$F5,$F1,$AF,$EC,$53,$9B,$F9
	.dc.b	$53,$9E,$2F,$EF,$93,$9A,$F3,$9D,$6B,$FB,$0C,$66,$7F,$C0,$59,$BE
	.dc.b	$CF,$E2,$9F,$D8,$BD,$3F,$D0,$61,$FD,$C6,$75,$87,$CC,$AD,$BF,$DB
	.dc.b	$2B,$77,$D2,$7E,$DF,$D5,$CA,$FB,$1F,$57,$71,$FE,$F9,$C2,$AC,$61
	.dc.b	$FE,$15,$85,$62,$FF,$33,$F6,$4F,$B2,$FD,$BE,$3F,$2A,$31,$E1,$FC
	.dc.b	$C8,$43,$E9,$9E,$1E,$63,$73,$F4,$BE,$C9,$F6,$3F,$EC,$C6,$C7,$36
	.dc.b	$BB,$8F,$F6,$8F,$E0,$7D,$23,$C0,$D7,$EE,$B5,$FC,$0B,$5F,$D9,$C3
	.dc.b	$E8,$FF,$D7,$ED,$8F,$D3,$35,$FB,$63,$EC,$FF,$C6,$5A,$F2,$AE,$24
	.dc.b	$9F,$D7,$4E,$9F,$EB,$94,$F8,$4E,$E4,$8E,$FE,$35,$7B,$CE,$DF,$1C
	.dc.b	$F9,$A7,$FD,$AB,$F6,$6F,$B3,$A4,$AF,$4F,$C8,$D7,$D3,$3B,$7F,$81
	.dc.b	$35,$EF,$1B,$E5,$FF,$2A,$CE,$5B,$5F,$1A,$7F,$DC,$BF,$65,$F8,$9B
	.dc.b	$FE,$18,$D7,$E3,$B6,$BF,$CB,$EF,$E3,$23,$7C,$FC,$88,$F2,$E6,$EB
	.dc.b	$F1,$1D,$BE,$72,$4F,$68,$CB,$E0,$8C,$BE,$08,$FF,$7C,$FD,$97,$F3
	.dc.b	$78,$FE,$40,$D7,$ED,$B7,$1F,$2F,$D9,$F0,$F3,$25,$70,$7D,$F1,$88
	.dc.b	$E3,$16,$78,$C6,$C8,$68,$F8,$B3,$BE,$4F,$E8,$77,$E1,$6D,$F6,$AE
	.dc.b	$37,$FE,$86,$AD,$47,$F8,$95,$AB,$46,$1D,$63,$57,$5D,$45,$AE,$27
	.dc.b	$2A,$E3,$3E,$0C,$3C,$6D,$3F,$C6,$91,$A7,$D8,$4C,$3C,$D8,$7A,$63
	.dc.b	$15,$C9,$A7,$2C,$3C,$53,$E5,$BE,$97,$CA,$BC,$03,$0F,$35,$F0,$8C
	.dc.b	$D7,$0C,$3B,$5F,$19,$EE,$7C,$03,$9D,$FC,$0C,$3C,$D3,$B2,$57,$23
	.dc.b	$FB,$BD,$F5,$BF,$96,$72,$AE,$33,$B8,$C9,$E0,$9E,$0F,$A6,$7D,$0F
	.dc.b	$E9,$30,$B7,$E0,$8F,$88,$75,$3D,$B1,$AB,$ED,$1E,$B3,$DB,$87,$AD
	.dc.b	$FC,$69,$C4,$F1,$4F,$17,$CB,$5F,$96,$78,$DF,$D1,$9D,$47,$9C,$E6
	.dc.b	$68,$F1,$9D,$67,$D0,$F8,$DB,$5E,$38,$F7,$1B,$BF,$4E,$1F,$12,$75
	.dc.b	$5C,$A3,$18,$78,$9B,$7C,$4F,$8A,$78,$CF,$69,$93,$D0,$78,$0D,$4F
	.dc.b	$5C,$3D,$B7,$F8,$A9,$E0,$D7,$78,$79,$A7,$78,$7D,$E9,$F9,$33,$F4
	.dc.b	$AB,$85,$FC,$0A,$F9,$55,$D1,$E0,$F9,$23,$4F,$CC,$6B,$EA,$3C,$27
	.dc.b	$63,$DE,$E8,$F5,$3D,$4E,$55,$FE,$4C,$FB,$37,$D0,$7F,$93,$A9,$39
	.dc.b	$F7,$70,$DB,$FF,$21,$5F,$06,$DE,$9F,$F5,$94,$7F,$A6,$1D,$D6,$54
	.dc.b	$71,$7F,$B2,$AF,$AD,$D3,$63,$BC,$63,$EB,$32,$C8,$C9,$E4,$8E,$DD
	.dc.b	$CF,$B0,$A4,$9F,$F2,$D7,$D8,$BE,$CF,$11,$FD,$DA,$B8,$3F,$E5,$F6
	.dc.b	$BF,$DE,$A7,$83,$9A,$E3,$AF,$C8,$8F,$F6,$95,$70,$8C,$42,$B8,$BC
	.dc.b	$FA,$BE,$24,$7F,$A2,$A3,$CB,$D5,$89,$91,$82,$1C,$4F,$5F,$D9,$F2
	.dc.b	$23,$EE,$F8,$CA,$FA,$62,$E1,$E4,$C1,$5F,$23,$DA,$97,$F0,$5F,$B1
	.dc.b	$AF,$FA,$DA,$30,$43,$E4,$46,$8E,$2E,$0D,$5F,$C2,$9E,$51,$FE,$24
	.dc.b	$5C,$FD,$89,$1F,$E8,$F8,$74,$62,$38,$4D,$AE,$93,$E6,$FB,$3F,$3E
	.dc.b	$F2,$F9,$46,$5E,$91,$97,$C5,$1F,$95,$FB,$37,$C9,$7F,$94,$E5,$18
	.dc.b	$7F,$CA,$CD,$7F,$D0,$6D,$7F,$D0,$EE,$70,$FC,$0B,$79,$FF,$D7,$6D
	.dc.b	$90,$FB,$37,$C8,$78,$F0,$93,$FE,$B1,$AE,$0E,$E7,$F5,$95,$FA,$78
	.dc.b	$FF,$27,$B7,$0F,$C8,$B7,$0F,$F6,$5A,$D7,$8F,$34,$BF,$F5,$F4,$ED
	.dc.b	$F2,$BB,$7F,$D8,$5E,$CF,$F9,$77,$0E,$B8,$78,$8F,$DF,$46,$5F,$F6
	.dc.b	$CB,$67,$9C,$6D,$E7,$E3,$7F,$05,$FB,$27,$D9,$FD,$16,$CF,$FA,$15
	.dc.b	$AF,$FA,$C9,$AF,$FA,$D9,$CE,$1F,$B0,$6D,$FF,$BA,$9B,$3F,$43,$0F
	.dc.b	$F5,$C3,$FE,$DD,$8C,$A6,$98,$8D,$C7,$F9,$F3,$EC,$FF,$DA,$CF,$FB
	.dc.b	$95,$AE,$7D,$9F,$E4,$1F,$6F,$FE,$FF,$8F,$F7,$57,$DA,$3E,$8B,$F9
	.dc.b	$23,$63,$31,$FC,$9F,$ED,$3F,$83,$E9,$FC,$29,$7C,$0E,$49,$5F,$F0
	.dc.b	$02,$7F,$D6,$9F,$17,$FC,$BF,$EC,$DF,$69,$49,$5F,$EF,$BC,$1F,$57
	.dc.b	$BF,$94,$93,$FE,$FF,$B4,$78,$E3,$F9,$D7,$DA,$3E,$CA,$92,$3C,$5E
	.dc.b	$8F,$FA,$07,$97,$F3,$B0,$FB,$4F,$CE,$BF,$A0,$DB,$F9,$2C,$62,$3F
	.dc.b	$D4,$ED,$8F,$E5,$38,$7C,$59,$FF,$55,$FD,$A7,$37,$95,$F6,$8C,$46
	.dc.b	$AF,$CF,$E8,$FF,$51,$F3,$FF,$85,$CF,$03,$F6,$2E,$1F,$F5,$BB,$71
	.dc.b	$FE,$78,$7F,$B7,$F8,$BF,$AD,$61,$F0,$6B,$90,$70,$F9,$B4,$FE,$C1
	.dc.b	$F4,$FF,$E7,$0F,$B4,$7D,$A7,$F6,$36,$D7,$26,$BF,$EC,$89,$FC,$9F
	.dc.b	$DA,$7F,$AE,$E3,$FD,$96,$DC,$BF,$A6,$B7,$6F,$99,$C1,$F8,$87,$E8
	.dc.b	$FD,$1B,$ED,$3F,$B9,$37,$0F,$4B,$5F,$F6,$BA,$F8,$ED,$92,$36,$FF
	.dc.b	$7B,$FB,$27,$C4,$D2,$4F,$C4,$8C,$BF,$1B,$F6,$F7,$C8,$FF,$80,$FE
	.dc.b	$D7,$F8,$D9,$3C,$DE,$9F,$21,$2F,$E6,$CE,$5E,$91,$97,$FD,$DB,$87
	.dc.b	$ED,$38,$3F,$B7,$B7,$6F,$87,$E4,$FF,$84,$7E,$D7,$F1,$91,$F9,$46
	.dc.b	$DF,$F3,$26,$AA,$FD,$2B,$F7,$A3,$24,$F3,$BE,$17,$81,$D5,$93,$B7
	.dc.b	$2F,$1F,$93,$F1,$11,$F8,$DF,$07,$56,$7F,$E0,$8C,$15,$98,$F6,$D7
	.dc.b	$F8,$AF,$ED,$5D,$14,$98,$7E,$52,$1F,$6B,$88,$EB,$AF,$ED,$27,$8C
	.dc.b	$6E,$BA,$57,$F8,$DF,$D9,$7D,$95,$25,$7E,$83,$ED,$7A,$E7,$FD,$6B
	.dc.b	$ED,$3F,$C1,$73,$FD,$72,$F8,$BF,$A6,$6B,$E2,$35,$7F,$E4,$5B,$E7
	.dc.b	$0A,$E6,$A3,$31,$8A,$E6,$DA,$FE,$1E,$F0,$9D,$CF,$F1,$26,$D7,$80
	.dc.b	$F4,$47,$F1,$63,$99,$BC,$3E,$9C,$78,$E3,$18,$70,$8E,$46,$4B,$7E
	.dc.b	$1A,$E2,$D9,$E1,$3F,$E3,$0C,$3C,$F7,$E1,$6C,$7F,$81,$30,$F3,$BF
	.dc.b	$07,$0E,$75,$B8,$E6,$FA,$1E,$11,$A6,$DF,$F3,$27,$D2,$53,$A3,$5A
	.dc.b	$66,$BF,$AB,$7D,$AE,$5B,$19,$F3,$53,$96,$C7,$F8,$A7,$EA,$3F,$C5
	.dc.b	$49,$A4,$74,$7C,$93,$F4,$3A,$7F,$9A,$F4,$FC,$0C,$7F,$89,$A7,$E4
	.dc.b	$E3,$FA,$0D,$BE,$5C,$E4,$E5,$19,$7A,$19,$AF,$C5,$C6,$5E,$B1,$E5
	.dc.b	$38,$DF,$87,$4F,$8D,$87,$82,$31,$3F,$2E,$31,$87,$8A,$39,$5F,$6A
	.dc.b	$DC,$7F,$62,$AD,$BE,$43,$77,$FE,$B1,$9D,$E1,$E9,$3F,$D1,$25,$C3
	.dc.b	$C7,$5E,$67,$FA,$8C,$66,$3F,$B6,$7D,$4F,$F5,$89,$E8,$69,$C4,$7F
	.dc.b	$94,$3E,$D3,$F7,$B3,$FD,$69,$B1,$FB,$50,$FF,$95,$A3,$5B,$F2,$9C
	.dc.b	$BD,$2F,$FA,$1E,$19,$BE,$71,$FD,$AB,$B7,$CD,$7D,$3A,$F8,$7D,$AF
	.dc.b	$F8,$73,$2F,$14,$63,$C8,$E4,$D8,$C9,$FE,$1F,$9E,$15,$E7,$C7,$FC
	.dc.b	$2F,$3F,$E2,$3F,$A2,$7C,$2F,$8F,$CF,$F3,$46,$27,$E0,$46,$27,$51
	.dc.b	$C5,$B7,$FC,$5F,$45,$F1,$8F,$05,$F1,$BF,$F0,$1F,$CB,$FF,$38,$7F
	.dc.b	$C0,$BF,$6B,$FE,$9E,$73,$D7,$F2,$63,$17,$F3,$DB,$8F,$FB,$23,$EB
	.dc.b	$9F,$6B,$BF,$86,$FA,$11,$8C,$7F,$C2,$2D,$C3,$EC,$7D,$AF,$FB,$76
	.dc.b	$73,$7F,$5E,$31,$3F,$77,$ED,$5E,$07,$DB,$7F,$D4,$37,$FE,$1C,$72
	.dc.b	$DE,$3D,$72,$FE,$07,$C2,$FF,$06,$4E,$E3,$1E,$67,$E1,$7E,$91,$E6
	.dc.b	$FE,$1F,$ED,$7F,$D9,$F7,$FE,$28,$9C,$BF,$E2,$A8,$CE,$9F,$3F,$ED
	.dc.b	$1F,$21,$F8,$D5,$F4,$16,$30,$E7,$EC,$F5,$F0,$77,$1F,$F3,$25,$A3
	.dc.b	$F2,$DF,$69,$FE,$71,$9F,$F3,$43,$F4,$1F,$F2,$A3,$9C,$7F,$00,$E5
	.dc.b	$FE,$EB,$F2,$CC,$3F,$40,$DA,$FF,$39,$3D,$DB,$1F,$A5,$FB,$4E,$5D
	.dc.b	$26,$9F,$ED,$E9,$FF,$3D,$3A,$FB,$3F,$D4,$C7,$F9,$9D,$B8,$7F,$81
	.dc.b	$9F,$F3,$EC,$7F,$A0,$5C,$D3,$E1,$7B,$9D,$6B,$F3,$4E,$3E,$4B,$F7
	.dc.b	$1F,$63,$CA,$7F,$3C,$FE,$01,$BF,$0F,$AE,$AD,$F3,$F0,$C2,$E7,$FE
	.dc.b	$91,$C3,$C0,$FF,$A4,$FE,$53,$A6,$7F,$E6,$A6,$C7,$FA,$05,$ED,$19
	.dc.b	$C7,$F8,$54,$72,$CB,$F6,$0D,$9E,$5F,$3D,$AC,$7F,$88,$7D,$77,$FB
	.dc.b	$6F,$1F,$F5,$13,$70,$FD,$C2,$FB,$CD,$FC,$A7,$F1,$7F,$A8,$FF,$29
	.dc.b	$DF,$FA,$A5,$FF,$30,$37,$DF,$D7,$C0,$ED,$B1,$93,$84,$65,$A6,$71
	.dc.b	$FE,$11,$5D,$1F,$E1,$5F,$55,$F8,$56,$F4,$FF,$29,$8F,$E1,$11,$8A
	.dc.b	$FF,$5E,$46,$F6,$FF,$49,$B6,$3F,$D5,$4F,$08,$CD,$7F,$35,$FA,$C7
	.dc.b	$D6,$D3,$E0,$70,$DC,$3F,$D3,$33,$A7,$FC,$04,$DC,$7F,$C0,$CD,$7F
	.dc.b	$C1,$4D,$7C,$3F,$59,$E9,$3F,$A1,$72,$BC,$DC,$5F,$AA,$6D,$7F,$2B
	.dc.b	$AD,$4E,$49,$1F,$1A,$79,$D7,$0C,$FF,$D6,$7A,$BE,$1B,$ED,$1F,$6F
	.dc.b	$E9,$6E,$3F,$6E,$2E,$3F,$4A,$6D,$AB,$FD,$AB,$F6,$EC,$B1,$7F,$E2
	.dc.b	$58,$C4,$FF,$89,$5B,$19,$9D,$BF,$E2,$DF,$AC,$FE,$2B,$9F,$F8,$AD
	.dc.b	$B1,$FE,$2F,$7F,$8D,$FD,$6E,$7C,$4F,$8D,$3F,$E3,$3F,$AA,$7C,$47
	.dc.b	$FB,$A3,$4E,$B1,$89,$FF,$67,$B5,$FE,$ED,$F5,$7F,$E0,$6D,$7F,$DD
	.dc.b	$F3,$9A,$E3,$7C,$E7,$2D,$B0,$BE,$4E,$2F,$95,$70,$8F,$0D,$71,$78
	.dc.b	$45,$9F,$F6,$94,$71,$F6,$30,$67,$E9,$6C,$7A,$A2,$BF,$E0,$F9,$4F
	.dc.b	$5F,$AB,$78,$1D,$7E,$06,$5F,$FB,$C2,$72,$D9,$FF,$2D,$28,$CB,$FE
	.dc.b	$5F,$6B,$FE,$63,$6B,$F0,$87,$D5,$FE,$35,$FF,$7E,$B5,$FF,$12,$37
	.dc.b	$5F,$C0,$AE,$4F,$88,$E7,$4E,$8D,$AF,$9E,$D7,$BC,$F0,$E5,$91,$E0
	.dc.b	$79,$C7,$4F,$63,$AF,$8F,$2B,$CB,$19,$7C,$F1,$97,$CF,$1F,$5F,$EA
	.dc.b	$DF,$57,$E6,$FA,$79,$5F,$82,$D3,$F3,$75,$A8,$C6,$9F,$16,$31,$1F
	.dc.b	$0A,$B8,$BF,$E4,$78,$CD,$7E,$93,$EA,$DF,$33,$F8,$75,$E8,$53,$11
	.dc.b	$2F,$B9,$F9,$49,$DB,$80,$E1,$FF,$2B,$3D,$9D,$35,$CC,$F7,$73,$3D
	.dc.b	$DC,$9C,$23,$15,$E9,$9E,$13,$A3,$BB,$AC,$7F,$11,$7E,$3A,$F8,$17
	.dc.b	$9B,$F5,$5F,$29,$CC,$78,$88,$75,$72,$F6,$32,$E0,$97,$E0,$9E,$D3
	.dc.b	$F1,$2B,$B4,$E6,$38,$BE,$68,$CC,$79,$0C,$9F,$3D,$CB,$F3,$A7,$AD
	.dc.b	$FB,$6B,$94,$F0,$34,$7C,$92,$1E,$92,$5F,$A1,$EC,$71,$9F,$AD,$19
	.dc.b	$8F,$51,$9A,$F1,$CF,$86,$3D,$07,$4A,$F1,$BF,$25,$F8,$0E,$4E,$E6
	.dc.b	$4E,$35,$F7,$5F,$3B,$E1,$9F,$1C,$F2,$8C,$5F,$C2,$BF,$1B,$6F,$8D
	.dc.b	$78,$AB,$A4,$78,$4F,$59,$E1,$3E,$39,$9B,$F6,$3A,$8E,$24,$9E,$84
	.dc.b	$8F,$4B,$E5,$30,$F1,$70,$E8,$F8,$31,$ED,$BF,$B7,$7F,$31,$B3,$EA
	.dc.b	$38,$1E,$69,$F8,$2F,$D3,$34,$77,$37,$7F,$21,$EF,$7E,$48,$C4,$73
	.dc.b	$8C,$5F,$C9,$8F,$89,$3F,$08,$D9,$23,$CD,$1F,$87,$6B,$F2,$5B,$5F
	.dc.b	$1E,$3C,$24,$3C,$07,$9A,$37,$5F,$1A,$3D,$44,$AF,$0D,$FD,$E8,$F8
	.dc.b	$07,$9A,$BD,$17,$F1,$23,$ED,$4E,$DA,$7C,$6A,$F6,$47,$CE,$8F,$7B
	.dc.b	$67,$EA,$47,$B4,$E1,$19,$BF,$94,$FF,$29,$FA,$69,$3F,$C7,$9F,$E5
	.dc.b	$9E,$46,$23,$F9,$1C,$5B,$FE,$65,$3F,$DD,$FD,$46,$9F,$CC,$F1,$98
	.dc.b	$7F,$2D,$9C,$B6,$72,$FF,$2E,$9E,$51,$FE,$01,$FA,$9F,$EF,$F2,$7F
	.dc.b	$BC,$BF,$CD,$DA,$B0,$FF,$13,$C7,$02,$73,$F3,$3A,$74,$CA,$E5,$19
	.dc.b	$7A,$47,$87,$EA,$A4,$77,$7C,$9D,$38,$8F,$06,$5F,$DF,$70,$56,$71
	.dc.b	$F1,$DF,$C5,$F1,$BF,$C3,$9A,$FF,$7B,$C7,$FA,$83,$5E,$67,$F8,$A3
	.dc.b	$54,$F1,$8F,$F0,$51,$FE,$0B,$8C,$B8,$38,$47,$F8,$5A,$33,$19,$97
	.dc.b	$27,$88,$91,$E2,$E5,$AF,$88,$4F,$07,$CF,$3E,$A9,$DE,$3F,$E5,$6A
	.dc.b	$D5,$71,$6C,$66,$7E,$0F,$D5,$7C,$2E,$09,$7F,$0A,$32,$FC,$48,$CB
	.dc.b	$CE,$3F,$C8,$9F,$54,$F7,$BF,$C8,$DF,$55,$FC,$78,$FE,$3F,$19,$AF
	.dc.b	$EF,$8D,$7E,$33,$5F,$96,$D7,$FC,$6C,$D7,$FC,$B3,$0E,$4E,$E4,$8F
	.dc.b	$F0,$44,$7C,$66,$C6,$5F,$95,$1C,$5D,$E7,$9C,$FF,$CF,$B3,$99,$DD
	.dc.b	$7B,$2B,$FC,$C5,$F5,$1F,$E6,$69,$F1,$F1,$FC,$03,$3E,$0C,$8F,$1C
	.dc.b	$65,$FF,$0E,$4F,$15,$F5,$2F,$AA,$EB,$D1,$27,$FC,$45,$39,$6C,$FF
	.dc.b	$98,$23,$11,$F8,$F6,$BF,$94,$6B,$FE,$5F,$AF,$D1,$BE,$2E,$6F,$59
	.dc.b	$EC,$E6,$B1,$92,$1F,$91,$94,$6A,$31,$37,$6F,$F5,$2D,$6A,$7F,$C7
	.dc.b	$B0,$FB,$CF,$0C,$7E,$B2,$31,$5F,$E9,$56,$C7,$09,$E4,$EF,$EF,$7F
	.dc.b	$A8,$49,$A4,$ED,$E5,$FB,$4F,$FA,$72,$4D,$27,$DA,$F7,$C1,$F7,$17
	.dc.b	$E7,$A6,$5E,$18,$F2,$64,$F8,$1B,$EB,$4A,$F6,$CE,$5F,$7E,$5F,$53
	.dc.b	$3F,$F6,$69,$E2,$EE,$F2,$9E,$6E,$2F,$17,$97,$2F,$E0,$3E,$5E,$BF
	.dc.b	$0F,$C7,$49,$EE,$F9,$3D,$2F,$7F,$9F,$27,$C2,$FC,$2F,$87,$F7,$7D
	.dc.b	$2F,$BF,$E8,$65,$7C,$D8,$FA,$1F,$80,$FF,$B5,$23,$EA,$2F,$A9,$C9
	.dc.b	$AF,$5F,$E6,$FD,$B7,$CE,$F0,$E5,$FC,$08,$FF,$63,$64,$F8,$DA,$49
	.dc.b	$FF,$3E,$43,$E0,$36,$48,$CB,$FE,$CF,$FA,$47,$BD,$49,$3F,$E8,$58
	.dc.b	$CB,$FE,$87,$FA,$67,$C1,$FF,$B5,$FE,$9B,$F8,$77,$4F,$FB,$E1,$BE
	.dc.b	$47,$FB,$05,$B7,$FE,$E4,$C1,$C7,$E1,$39,$73,$C9,$F0,$46,$5A,$43
	.dc.b	$F3,$9E,$CF,$EF,$7E,$99,$F2,$5F,$EE,$96,$CF,$FA,$75,$AF,$F3,$DC
	.dc.b	$1E,$1E,$EA,$4E,$DF,$89,$18,$8F,$E8,$95,$CB,$46,$9F,$ED,$D8,$C3
	.dc.b	$FE,$DF,$C3,$AD,$78,$7A,$FC,$84,$7E,$67,$C6,$57,$AC,$5C,$3A,$C7
	.dc.b	$C7,$9F,$F7,$54,$62,$1F,$4C,$F9,$EC,$E3,$FE,$EF,$8C,$4F,$FA,$F5
	.dc.b	$AF,$F5,$D5,$F4,$DC,$20,$74,$AC,$34,$E4,$E2,$78,$C7,$FB,$27,$CD
	.dc.b	$7D,$07,$06,$C6,$63,$9F,$D3,$78,$32,$EF,$F4,$9F,$C4,$2F,$77,$FD
	.dc.b	$C5,$B3,$EF,$6E,$7E,$69,$5F,$35,$7F,$C6,$A3,$0F,$5C,$38,$36,$7C
	.dc.b	$37,$CE,$B6,$DC,$FF,$C5,$93,$D2,$3C,$07,$9A,$B9,$46,$4F,$93,$5C
	.dc.b	$9B,$5C,$A7,$64,$AE,$D8,$F8,$AB,$6D,$7F,$A0,$E1,$E8,$9F,$0B,$63
	.dc.b	$67,$84,$F8,$77,$DE,$77,$1C,$DF,$41,$E0,$78,$61,$F2,$5F,$F1,$6B
	.dc.b	$5D,$18,$9F,$86,$EB,$E9,$5F,$4D,$FE,$31,$8F,$A7,$18,$7F,$C6,$94
	.dc.b	$8C,$47,$19,$E0,$FE,$43,$E9,$AA,$69,$1D,$BF,$17,$2F,$D0,$69,$F1
	.dc.b	$E7,$91,$F0,$E3,$2F,$C5,$33,$A7,$F5,$E8,$CB,$FD,$82,$33,$9F,$EB
	.dc.b	$B6,$FD,$8C,$E6,$7C,$87,$58,$CC,$FF,$94,$70,$F2,$46,$27,$FC,$AF
	.dc.b	$18,$C3,$CD,$1C,$AF,$C5,$5B,$8F,$CE,$56,$DF,$41,$BB,$F2,$5E,$70
	.dc.b	$FE,$04,$7B,$8F,$E0,$93,$EA,$3F,$76,$E6,$38,$C7,$F9,$76,$1F,$4D
	.dc.b	$FD,$A2,$73,$1D,$23,$13,$FC,$13,$E9,$3F,$7B,$5D,$0F,$D9,$38,$8F
	.dc.b	$F3,$43,$FC,$2A,$35,$1C,$77,$E6,$A7,$85,$FF,$78,$ED,$F3,$4E,$EF
	.dc.b	$FC,$37,$18,$8E,$1D,$5E,$4A,$7D,$2B,$E9,$BA,$E3,$FE,$65,$C7,$FC
	.dc.b	$47,$39,$7C,$78,$72,$73,$5C,$30,$E9,$5B,$AF,$5F,$D2,$F7,$FA,$67
	.dc.b	$C3,$76,$8C,$37,$1F,$F1,$8C,$E9,$EF,$92,$B7,$18,$6D,$F6,$8E,$37
	.dc.b	$F1,$BC,$87,$57,$C8,$FA,$6F,$F4,$CD,$7A,$A7,$2F,$B2,$33,$7F,$E9
	.dc.b	$96,$ED,$FD,$1F,$E9,$9F,$4D,$CF,$E1,$FE,$9C,$63,$0F,$8A,$DC,$3F
	.dc.b	$D4,$5F,$4C,$FB,$2D,$E1,$F6,$A3,$17,$F3,$3E,$9D,$E0,$FD,$E7,$E9
	.dc.b	$61,$FE,$AB,$73,$3B,$E4,$D7,$37,$E1,$3C,$5F,$F0,$D4,$F3,$8C,$79
	.dc.b	$BF,$88,$FA,$77,$9D,$F8,$AF,$A7,$FF,$32,$DF,$DC,$9C,$BF,$7A,$33
	.dc.b	$9F,$F9,$87,$E9,$DE,$D7,$E2,$57,$AF,$63,$0E,$7D,$ED,$7B,$D7,$1F
	.dc.b	$F3,$6D,$A3,$F2,$5F,$4F,$FC,$76,$7F,$CE,$2F,$D0,$7F,$62,$E7,$1F
	.dc.b	$F3,$93,$97,$FD,$F1,$F0,$7F,$E5,$F9,$FF,$3C,$B6,$3F,$CF,$0F,$78
	.dc.b	$CE,$1F,$A1,$FA,$7E,$5D,$46,$8F,$8F,$D7,$5F,$E9,$FE,$9F,$F5,$F8
	.dc.b	$7F,$04,$6E,$4C,$10,$FA,$6E,$1E,$3F,$EB,$FE,$9B,$B9,$D7,$0F,$F4
	.dc.b	$7B,$8C,$BF,$69,$F4,$FC,$67,$FD,$26,$FE,$03,$2F,$F4,$93,$96,$F6
	.dc.b	$70,$E9,$E3,$85,$BE,$2F,$EF,$FE,$53,$86,$BF,$E7,$E6,$CF,$F1,$E7
	.dc.b	$B4,$67,$4E,$15,$2B,$FD,$4B,$4F,$9D,$C6,$5F,$EA,$5F,$A7,$FF,$7F
	.dc.b	$D7,$FA,$A9,$B1,$9F,$A7,$FE,$21,$18,$CB,$FD,$1F,$F4,$9F,$E6,$2B
	.dc.b	$FF,$58,$BF,$E6,$F6,$FB,$FF,$C5,$E3,$18,$7F,$AC,$DE,$0D,$F8,$1D
	.dc.b	$4E,$DB,$5D,$1B,$3B,$8F,$F5,$FF,$D3,$FE,$16,$3F,$D7,$CD,$C3,$FD
	.dc.b	$46,$FF,$06,$8C,$F5,$7F,$29,$8D,$57,$F4,$66,$CF,$E0,$DE,$15,$CA
	.dc.b	$BF,$9C,$FD,$33,$E9,$FF,$A7,$DF,$F9,$22,$31,$87,$F8,$0D,$B1,$FE
	.dc.b	$37,$35,$3C,$14,$6A,$FF,$D7,$3F,$6B,$FE,$84,$BF,$E6,$F3,$97,$FA
	.dc.b	$DC,$66,$BD,$FF,$17,$EB,$3F,$A2,$72,$3C,$EC,$5F,$C6,$9B,$5F,$EE
	.dc.b	$3A,$D4,$E4,$91,$D1,$FE,$A3,$5C,$31,$FB,$FF,$07,$F6,$CF,$B9,$F4
	.dc.b	$56,$BF,$03,$16,$FE,$F4,$DA,$FF,$3C,$7D,$3F,$F1,$0C,$3F,$C5,$0D
	.dc.b	$AF,$8C,$1F,$E2,$FF,$4F,$AF,$17,$FD,$91,$3F,$E6,$1F,$B5,$7C,$87
	.dc.b	$F8,$A9,$B1,$FE,$33,$6B,$D1,$AF,$D3,$87,$CD,$FF,$9F,$9B,$1F,$D9
	.dc.b	$9A,$FF,$05,$D7,$FC,$E1,$F5,$1F,$E1,$07,$FC,$E5,$3A,$D3,$FD,$28
	.dc.b	$D8,$CD,$FF,$A2,$AB,$85,$FF,$A3,$AB,$4D,$9F,$11,$BA,$F2,$7B,$12
	.dc.b	$7C,$8F,$9F,$E7,$31,$1E,$53,$A4,$63,$1F,$1C,$FF,$A6,$BE,$A3,$F8
	.dc.b	$05,$7F,$80,$9A,$F1,$0E,$27,$D7,$E1,$71,$EF,$95,$C5,$F9,$11,$FE
	.dc.b	$A6,$FA,$87,$25,$25,$7F,$84,$21,$C9,$88,$F0,$47,$07,$53,$CE,$75
	.dc.b	$87,$FA,$FE,$B5,$5C,$2B,$4F,$5F,$23,$C5,$E3,$48,$C1,$0E,$EE,$5F
	.dc.b	$0C,$7F,$AF,$BE,$9D,$F5,$14,$93,$E9,$F9,$5A,$FC,$79,$DB,$FE,$2C
	.dc.b	$6B,$EC,$8F,$F6,$4F,$D4,$3E,$A2,$FA,$12,$BF,$C6,$73,$96,$CE,$5F
	.dc.b	$A9,$09,$E7,$32,$78,$C6,$F5,$FF,$72,$57,$39,$E3,$59,$E4,$FF,$64
	.dc.b	$4F,$39,$EE,$43,$BD,$ED,$FF,$71,$C7,$3B,$EB,$79,$94,$F3,$7B,$9C
	.dc.b	$E7,$C7,$5B,$9E,$57,$E2,$BC,$4F,$AE,$F8,$9E,$23,$DA,$73,$9E,$0E
	.dc.b	$0D,$1C,$2F,$CA,$E5,$F0,$BC,$4F,$5D,$78,$1C,$1E,$A7,$0F,$39,$EE
	.dc.b	$F6,$9E,$6E,$4E,$67,$78,$F9,$B5,$EB,$9F,$1C,$F1,$9C,$BE,$0B,$F8
	.dc.b	$6F,$CE,$AF,$8D,$7C,$6B,$73,$C4,$EE,$78,$4F,$A3,$3C,$1E,$8D,$75
	.dc.b	$5F,$56,$FD,$27,$90,$95,$D9,$CB,$F0,$1C,$CF,$94,$F7,$1E,$27,$2F
	.dc.b	$88,$C9,$F2,$CF,$19,$89,$F7,$B9,$3D,$87,$13,$CF,$7F,$4A,$7A,$CF
	.dc.b	$03,$D5,$5E,$73,$CC,$4A,$F3,$1D,$23,$85,$73,$9D,$4F,$86,$BD,$EF
	.dc.b	$07,$BC,$EE,$3D,$D3,$C2,$BC,$31,$B8,$C4,$EE,$3A,$46,$EF,$E9,$5F
	.dc.b	$A9,$B1,$99,$F8,$95,$F7,$DE,$84,$39,$9E,$23,$17,$E2,$72,$47,$04
	.dc.b	$8E,$91,$F3,$6F,$DE,$EE,$3D,$13,$F1,$A3,$EB,$5F,$08,$FA,$0F,$A2
	.dc.b	$3C,$47,$3B,$FA,$4E,$27,$6F,$B0,$FC,$09,$F4,$EB,$A5,$73,$8F,$43
	.dc.b	$67,$8C,$6C,$FE,$4B,$F2,$FF,$CA,$88,$FF,$28,$FA,$3A,$FF,$26,$8B
	.dc.b	$8F,$F2,$29,$4F,$FB,$83,$D5,$7D,$3E,$23,$FB,$04,$61,$D4,$66,$99
	.dc.b	$EF,$EA,$3F,$D8,$55,$A2,$1C,$AB,$84,$EF,$E0,$6C,$F8,$AB,$FD,$D2
	.dc.b	$FF,$BC,$97,$4E,$A1,$18,$75,$D5,$98,$E3,$1F,$DD,$DD,$36,$75,$32
	.dc.b	$3B,$47,$13,$0F,$1F,$A3,$E8,$47,$C9,$F1,$95,$D1,$E7,$8F,$14,$79
	.dc.b	$F4,$7A,$34,$95,$F4,$35,$F9,$2F,$D2,$DB,$6F,$D3,$9D,$BF,$49,$F8
	.dc.b	$51,$A8,$FF,$0A,$2E,$6C,$92,$3B,$6D,$F6,$E7,$FD,$81,$18,$3C,$13
	.dc.b	$B7,$9B,$DB,$9F,$18,$79,$A5,$5D,$DE,$46,$2B,$83,$88,$7D,$47,$A3
	.dc.b	$86,$4E,$63,$D6,$61,$F5,$C6,$1E,$BF,$50,$F8,$2F,$F3,$0C,$7E,$62
	.dc.b	$30,$E1,$AF,$F9,$99,$AF,$FA,$35,$CE,$3E,$B6,$E7,$F9,$F6,$CF,$E9
	.dc.b	$3E,$A1,$F0,$7F,$91,$5C,$D4,$93,$F0,$A3,$E2,$B6,$32,$FF,$92,$63
	.dc.b	$8B,$BD,$33,$8F,$CD,$9C,$CE,$EB,$C1,$5F,$A1,$FA,$7F,$F3,$1C,$AB
	.dc.b	$FC,$CE,$DC,$B5,$E2,$48,$F2,$47,$A4,$C3,$FE,$5B,$FA,$87,$D1,$FC
	.dc.b	$C8,$46,$1F,$9E,$D7,$FD,$22,$D7,$FD,$70,$E7,$3F,$C3,$36,$BF,$85
	.dc.b	$36,$7F,$87,$E2,$F9,$EF,$67,$3C,$9F,$C5,$CA,$3E,$07,$3E,$67,$FC
	.dc.b	$E5,$39,$77,$3A,$69,$0D,$B9,$7F,$D0,$58,$BE,$9F,$94,$61,$B1,$FB
	.dc.b	$43,$53,$FC,$59,$E1,$77,$E9,$BF,$C2,$B0,$69,$3B,$F8,$7D,$37,$FA
	.dc.b	$72,$4D,$27,$D4,$75,$C5,$F5,$17,$E8,$E6,$9F,$B7,$C3,$D1,$3E,$1F
	.dc.b	$87,$F1,$7B,$72,$B5,$A2,$7C,$3C,$7E,$2F,$02,$57,$3D,$7E,$A6,$5F
	.dc.b	$0F,$D0,$F1,$12,$7C,$1A,$FE,$92,$7D,$DF,$27,$F0,$49,$3E,$0F,$67
	.dc.b	$DF,$EA,$C8,$EC,$FC,$2F,$8D,$F7,$7D,$47,$BF,$A2,$57,$68,$F9,$DF
	.dc.b	$51,$FE,$F5,$8F,$80,$BC,$EC,$9A,$F6,$FE,$6F,$D4,$7C,$EF,$0A,$5F
	.dc.b	$B6,$3F,$A4,$64,$F9,$1A,$49,$FE,$33,$0F,$83,$D9,$23,$2F,$F5,$1F
	.dc.b	$A5,$7C,$05,$24,$FF,$22,$8C,$BF,$C9,$3E,$A1,$F0,$DF,$D6,$7E,$A3
	.dc.b	$F8,$6C,$9A,$62,$BF,$49,$1C,$F8,$78,$79,$92,$30,$43,$FD,$A9,$3C
	.dc.b	$1B,$3F,$97,$F6,$BF,$83,$FD,$47,$F1,$A9,$FE,$6E,$D7,$FD,$B8,$DE
	.dc.b	$8F,$E7,$D6,$F0,$BF,$B8,$4E,$7C,$9F,$F0,$23,$6F,$FC,$12,$D5,$5C
	.dc.b	$5E,$FE,$6F,$22,$3F,$1E,$FA,$F5,$C8,$F2,$BF,$3F,$DE,$C4,$79,$A2
	.dc.b	$E5,$E6,$9F,$F0,$27,$D4,$3E,$8E,$92,$FF,$AA,$C3,$EA,$31,$1F,$EF
	.dc.b	$B8,$E0,$EA,$79,$CF,$F8,$3F,$CB,$7C,$95,$25,$7F,$60,$FA,$8E,$99
	.dc.b	$7F,$8A,$BE,$99,$7F,$E2,$93,$8D,$7F,$66,$6A,$9F,$ED,$91,$87,$FB
	.dc.b	$75,$74,$95,$74,$7F,$C6,$F1,$98,$C4,$78,$2F,$FC,$71,$18,$BE,$EE
	.dc.b	$AB,$84,$7F,$8F,$5B,$5F,$E4,$28,$D9,$E9,$AE,$6E,$F0,$F4,$4F,$92
	.dc.b	$31,$87,$C1,$78,$45,$3A,$57,$27,$D3,$5C,$B0,$F2,$36,$B8,$CF,$53
	.dc.b	$89,$CC,$61,$E8,$6C,$F1,$76,$71,$3E,$21,$C6,$38,$1F,$35,$FF,$2F
	.dc.b	$1C,$B0,$FF,$06,$C6,$63,$DA,$68,$F0,$FD,$43,$EA,$36,$D8,$D4,$A3
	.dc.b	$11,$DA,$31,$3F,$DB,$E1,$F5,$14,$8F,$F7,$36,$E5,$F7,$B0,$FC,$F7
	.dc.b	$27,$C3,$AF,$F3,$CD,$7C,$59,$CC,$7C,$13,$27,$C3,$73,$5F,$88,$8C
	.dc.b	$BD,$75,$FF,$29,$B9,$8F,$29,$C6,$33,$A7,$C3,$BF,$04,$62,$7E,$5C
	.dc.b	$63,$0F,$34,$72,$BE,$D1,$89,$D4,$F9,$CD,$5F,$FA,$36,$FA,$47,$E9
	.dc.b	$DF,$EE,$95,$93,$FC,$A3,$1E,$A9,$C3,$FA,$D8,$B8,$7F,$96,$47,$D4
	.dc.b	$72,$9F,$F3,$95,$D7,$F6,$91,$7E,$97,$F8,$0D,$FF,$97,$9B,$3D,$4F
	.dc.b	$F5,$4B,$FE,$AA,$8D,$6F,$C4,$ED,$87,$F4,$8C,$31,$3F,$DE,$FB,$7E
	.dc.b	$BC,$FF,$9B,$6F,$FC,$BD,$3F,$DF,$FE,$9D,$F5,$1E,$3E,$CE,$71,$89
	.dc.b	$F2,$46,$27,$51,$C5,$F5,$7D,$3F,$8B,$EA,$1F,$0B,$8A,$FF,$06,$46
	.dc.b	$1F,$74,$EE,$75,$B7,$F8,$A2,$B4,$FC,$11,$92,$B8,$5F,$23,$AD,$FC
	.dc.b	$DF,$21,$D5,$F3,$FE,$A3,$FD,$31,$5F,$E8,$A9,$CB,$FE,$8C,$8C,$E3
	.dc.b	$DD,$BB,$7F,$A4,$FE,$A1,$F5,$1C,$FE,$17,$ED,$46,$30,$FF,$4A,$B7
	.dc.b	$0F,$47,$D4,$3E,$2F,$78,$7E,$06,$31,$7F,$E9,$CF,$A8,$FA,$77,$EF
	.dc.b	$6E,$5F,$EA,$35,$E0,$FE,$25,$CD,$7F,$BA,$7C,$3F,$F0,$83,$FE,$A6
	.dc.b	$6D,$F3,$E7,$FC,$84,$63,$0F,$F5,$77,$D3,$FF,$4A,$BF,$F5,$6B,$6F
	.dc.b	$E8,$7D,$43,$CE,$FC,$C4,$E6,$FF,$CB,$1F,$51,$9C,$3F,$C9,$0D,$C3
	.dc.b	$FC,$C2,$BD,$1F,$CA,$B9,$AE,$3C,$B8,$33,$EB,$DC,$3F,$CA,$5F,$50
	.dc.b	$F8,$DF,$CE,$B6,$33,$F8,$06,$27,$F4,$3F,$0B,$F7,$34,$FF,$38,$AC
	.dc.b	$33,$8E,$9F,$D3,$FD,$4B,$0B,$F8,$06,$1B,$97,$F9,$8F,$EA,$BE,$8B
	.dc.b	$7E,$27,$F5,$F1,$8C,$FF,$CF,$EF,$FA,$2E,$7F,$CD,$5F,$51,$D5,$B7
	.dc.b	$FB,$68,$61,$63,$F7,$3F,$55,$C7,$0F,$F4,$4B,$78,$3F,$D0,$CF,$D1
	.dc.b	$C3,$FD,$F8,$E5,$BF,$3B,$C8,$E7,$E3,$58,$FF,$3E,$7C,$AE,$F0,$FF
	.dc.b	$02,$3C,$DB,$AF,$F0,$A8,$C6,$27,$F8,$14,$FF,$A7,$1F,$CC,$B6,$F8
	.dc.b	$77,$FF,$12,$FA,$87,$CE,$5F,$A6,$C3,$73,$FF,$47,$7D,$3B,$E3,$FF
	.dc.b	$8C,$B6,$33,$CD,$89,$91,$C1,$DC,$63,$3F,$F5,$7D,$71,$6C,$EA,$B8
	.dc.b	$08,$FE,$19,$F4,$EC,$7F,$9A,$36,$7F,$D4,$5F,$51,$F0,$5E,$1B,$9E
	.dc.b	$56,$AE,$0F,$4B,$D3,$FD,$EE,$32,$FF,$80,$9A,$F2,$6B,$FE,$37,$C7
	.dc.b	$F3,$15,$F1,$BA,$3F,$9E,$F8,$4F,$2B,$E3,$C4,$9D,$1F,$CF,$E3,$17
	.dc.b	$FE,$D2,$CD,$EF,$FF,$46,$6D,$7B,$BE,$8B,$13,$23,$E0,$BB,$8C,$57
	.dc.b	$38,$BA,$FF,$B6,$6D,$58,$BF,$8B,$F5,$7E,$AC,$3F,$C3,$8D,$7F,$C4
	.dc.b	$6D,$77,$97,$C8,$7F,$C4,$FF,$50,$F8,$DF,$F5,$B4,$61,$AF,$D2,$87
	.dc.b	$C8,$BE,$A3,$FC,$55,$29,$E5,$B6,$3E,$45,$F3,$5F,$6B,$7F,$04,$62
	.dc.b	$FE,$FE,$BF,$7A,$31,$1F,$D4,$FE,$9F,$FB,$F4,$E6,$F9,$36,$78,$9F
	.dc.b	$8F,$AE,$37,$ED,$BE,$33,$9F,$57,$8B,$4E,$22,$BF,$B5,$B8,$5F,$43
	.dc.b	$C2,$35,$36,$77,$1F,$DA,$5B,$7E,$EA,$D5,$4A,$E8,$F1,$8E,$9F,$4E
	.dc.b	$F0,$7A,$7C,$04,$BF,$2C,$E5,$B3,$97,$C1,$3C,$A3,$CD,$F4,$FF,$92
	.dc.b	$9F,$6B,$5F,$EF,$CD,$C7,$F0,$4F,$E3,$23,$7C,$FC,$88,$F2,$E6,$EB
	.dc.b	$EC,$76,$F3,$24,$F6,$8C,$BE,$08,$CB,$E0,$8C,$BE,$28,$CE,$1E,$29
	.dc.b	$F9,$DF,$4F,$FC,$72,$BD,$3F,$47,$EC,$33,$E1,$48,$C9,$0F,$F0,$D4
	.dc.b	$F1,$7F,$C3,$9F,$4C,$FA,$7F,$B4,$BD,$0A,$62,$24,$FE,$12,$74,$D9
	.dc.b	$CB,$D3,$E5,$B9,$91,$E1,$F4,$95,$D1,$CE,$3C,$31,$9D,$3F,$65,$69
	.dc.b	$C7,$47,$F2,$08,$56,$2F,$F9,$64,$6D,$E8,$78,$0A,$78,$09,$7D,$A7
	.dc.b	$94,$27,$93,$AA,$F0,$4E,$A7,$35,$E6,$BF,$4D,$F6,$8C,$36,$F8,$B4
	.dc.b	$F3,$CF,$3B,$E8,$D8,$D1,$E3,$70,$F8,$CE,$26,$AF,$E1,$B6,$37,$7E
	.dc.b	$28,$E0,$73,$3E,$21,$89,$F5,$DF,$CD,$AF,$5C,$E5,$E7,$19,$8E,$C7
	.dc.b	$2A,$EC,$48,$D4,$7E,$3A,$FD,$AF,$9D,$AF,$85,$B3,$C6,$3E,$DC,$62
	.dc.b	$79,$CF,$D7,$9F,$09,$C4,$AF,$17,$ED,$4F,$17,$E3,$9E,$33,$73,$D0
	.dc.b	$FA,$AF,$BC,$F3,$1A,$38,$CF,$85,$D1,$E1,$2B,$E1,$BF,$A7,$18,$9E
	.dc.b	$CF,$B0,$E8,$7D,$33,$C8,$6C,$EE,$F9,$27,$A5,$FD,$29,$C9,$0E,$51
	.dc.b	$E5,$36,$FA,$8E,$86,$A7,$A1,$E4,$70,$73,$AF,$4C,$F2,$8F,$73,$4F
	.dc.b	$75,$FC,$D8,$E2,$7D,$03,$DB,$3F,$8C,$7E,$01,$E2,$70,$EC,$C5,$FE
	.dc.b	$42,$BB,$5F,$96,$BC,$6D,$9F,$85,$5E,$09,$E6,$4B,$F2,$36,$3C,$73
	.dc.b	$D6,$B9,$47,$C5,$BF,$F0,$07,$6B,$5F,$E4,$75,$2B,$FB,$1C,$E5,$B3
	.dc.b	$97,$F9,$5C,$FF,$63,$9D,$47,$FB,$27,$E9,$FF,$B0,$CF,$F8,$E1,$FE
	.dc.b	$D3,$1F,$DB,$9A,$EF,$A3,$FB,$64,$70,$8F,$F2,$43,$AC,$3F,$C2,$CD
	.dc.b	$AE,$95,$C3,$CD,$97,$D5,$E7,$CF,$88,$F2,$45,$C3,$9C,$DB,$F3,$4D
	.dc.b	$AF,$0C,$FB,$7C,$DF,$F0,$06,$DF,$D9,$67,$FD,$CC,$DC,$3E,$5C,$EE
	.dc.b	$7F,$C1,$73,$9C,$3F,$CB,$2D,$8F,$F3,$26,$7F,$E7,$F7,$7C,$7D,$88
	.dc.b	$F2,$E4,$AF,$82,$2E,$3F,$E3,$9B,$FF,$5C,$4F,$83,$E9,$FC,$FB,$C8
	.dc.b	$C1,$0F,$4B,$97,$D7,$1E,$E3,$15,$D6,$71,$3F,$E2,$6F,$A7,$7C,$37
	.dc.b	$F8,$8F,$E9,$FF,$B2,$CA,$FF,$C9,$91,$87,$FC,$9D,$5C,$27,$95,$7F
	.dc.b	$92,$14,$65,$FF,$3F,$9F,$E8,$97,$0E,$1A,$FF,$A1,$2B,$F5,$A4,$E3
	.dc.b	$F2,$65,$E5,$64,$FF,$8F,$A7,$73,$FE,$3B,$FA,$6F,$F2,$3E,$FC,$CF
	.dc.b	$61,$9F,$0A,$5F,$18,$CB,$FE,$48,$9E,$2F,$FA,$7D,$AA,$FF,$CA,$2D
	.dc.b	$9E,$6F,$C7,$7F,$D3,$BF,$4E,$FA,$3F,$27,$85,$2B,$FC,$AB,$39,$6C
	.dc.b	$FC,$98,$C4,$7C,$80,$E1,$F9,$6D,$7F,$D5,$78,$39,$BE,$11,$23,$D9
	.dc.b	$CD,$5F,$31,$ED,$8C,$70,$7F,$A2,$69,$5F,$E6,$5F,$C2,$FF,$C9,$99
	.dc.b	$FF,$AE,$1F,$D0,$60,$69,$A7,$FD,$8C,$D7,$5D,$FC,$CE,$9E,$04,$8F
	.dc.b	$F3,$BC,$F2,$9C,$47,$9B,$17,$D1,$FF,$0B,$6E,$DF,$E8,$06,$CF,$FA
	.dc.b	$21,$DF,$5E,$7E,$89,$D1,$7A,$25,$EE,$72,$F1,$FA,$4F,$E6,$EF,$E8
	.dc.b	$57,$D5,$F7,$6C,$6F,$9F,$C5,$A3,$E3,$69,$27,$C9,$0F,$4B,$C9,$F4
	.dc.b	$72,$7C,$4F,$9F,$EC,$7B,$79,$12,$7C,$8F,$87,$D7,$EF,$F0,$1E,$0F
	.dc.b	$32,$57,$68,$F8,$1F,$27,$F0,$7E,$56,$4F,$BE,$3D,$5F,$80,$7A,$FE
	.dc.b	$42,$4F,$C1,$8F,$5F,$C0,$C8,$F4,$BE,$EF,$81,$F9,$9F,$39,$27,$E5
	.dc.b	$3F,$3F,$E7,$3E,$37,$C7,$C9,$F9,$8F,$C9,$F4,$FF,$D6,$1F,$1C,$E3
	.dc.b	$A4,$AF,$F5,$CC,$3E,$97,$64,$8C,$BF,$EB,$9F,$A0,$F9,$91,$FE,$C0
	.dc.b	$6B,$F9,$E0,$E3,$C0,$FF,$73,$37,$C5,$4C,$FA,$57,$D2,$FF,$B1,$7E
	.dc.b	$97,$F8,$D6,$1F,$EE,$D6,$DF,$0B,$46,$1E,$78,$3E,$97,$CC,$73,$F4
	.dc.b	$E4,$F9,$E3,$2D,$21,$FA,$1F,$2B,$F8,$8F,$D2,$B8,$7F,$DB,$0D,$9F
	.dc.b	$F6,$FB,$5F,$81,$83,$E9,$77,$52,$76,$FC,$98,$C4,$7C,$7A,$E5,$A3
	.dc.b	$4F,$74,$61,$F8,$18,$75,$AF,$0F,$C6,$79,$08,$FD,$3F,$99,$AF,$58
	.dc.b	$B8,$7D,$29,$FF,$79,$7D,$2B,$B2,$92,$FF,$DF,$70,$E8,$C4,$7D,$D8
	.dc.b	$E0,$EA,$79,$CF,$FB,$F7,$D0,$7D,$0D,$25,$7E,$03,$1F,$73,$F8,$2F
	.dc.b	$A5,$ED,$1F,$84,$FA,$4F,$E4,$73,$F8,$6B,$E6,$F1,$DB,$F0,$51,$88
	.dc.b	$FC,$83,$6B,$AC,$2F,$AA,$BE,$AE,$71,$FE,$E0,$E6,$FA,$E3,$F8,$FA
	.dc.b	$DC,$77,$78,$DF,$46,$DF,$27,$75,$C6,$B9,$DF,$B5,$FC,$D5,$F1,$6D
	.dc.b	$F1,$77,$3E,$33,$E1,$57,$37,$E4,$4E,$DB,$87,$57,$65,$3F,$A6,$CF
	.dc.b	$17,$E0,$46,$6F,$C5,$1C,$0F,$01,$C0,$D5,$77,$78,$9F,$D9,$E3,$E6
	.dc.b	$BF,$E1,$06,$BE,$C9,$FE,$D8,$FD,$3E,$C7,$D2,$E2,$91,$89,$E8,$D7
	.dc.b	$A3,$5F,$EE,$DF,$4B,$FE,$21,$4D,$23,$FE,$25,$7E,$15,$FE,$7F,$4F
	.dc.b	$89,$3F,$C4,$A7,$3A,$FE,$53,$5F,$C2,$CE,$4F,$87,$7C,$5F,$8A,$66
	.dc.b	$3F,$17,$19,$78,$C7,$90,$E3,$3D,$75,$F4,$C6,$30,$F1,$BB,$AE,$75
	.dc.b	$B8,$FF,$1D,$E1,$E6,$AD,$B6,$3C,$67,$1B,$FE,$4D,$87,$8D,$B7,$FB
	.dc.b	$08,$FD,$29,$EB,$3F,$4F,$3E,$B9,$DD,$FE,$BE,$3F,$71,$5C,$63,$AF
	.dc.b	$87,$67,$31,$B7,$FC,$B5,$19,$5F,$47,$FC,$12,$74,$D8,$FE,$90,$FF
	.dc.b	$44,$8C,$EF,$C6,$35,$8F,$A7,$CA,$F4,$57,$F4,$39,$CB,$FC,$36,$33
	.dc.b	$7F,$D3,$A7,$FC,$DF,$3F,$E1,$7F,$A5,$78,$9D,$7B,$3D,$58,$A7,$35
	.dc.b	$E6,$9C,$C6,$30,$F3,$C6,$2F,$CD,$E6,$F2,$FA,$57,$AB,$8C,$7F,$C4
	.dc.b	$D1,$87,$E0,$CE,$E7,$58,$7F,$8C,$6B,$4F,$C4,$9D,$57,$03,$9D,$7C
	.dc.b	$DF,$8D,$75,$7C,$FF,$A5,$7C,$1E,$63,$0C,$BF,$AD,$F4,$AF,$37,$19
	.dc.b	$FF,$79,$8C,$4E,$A7,$ED,$7D,$2B,$CD,$FB,$95,$AB,$FF,$00,$4E,$6F
	.dc.b	$F0,$5F,$4B,$F4,$27,$FC,$08,$DE,$3F,$E6,$AB,$C1,$FC,$33,$9A,$FE
	.dc.b	$FD,$E2,$FF,$85,$DF,$C6,$36,$F9,$F3,$FE,$36,$31,$87,$F4,$0F,$A4
	.dc.b	$FF,$35,$DF,$E4,$DB,$7F,$E1,$5F,$A5,$79,$DF,$95,$9C,$DF,$E9,$7E
	.dc.b	$93,$38,$7F,$97,$5B,$87,$F9,$95,$7D,$2F,$E2,$1C,$D7,$1E,$3C,$19
	.dc.b	$F9,$9B,$87,$F9,$9F,$E9,$5F,$1B,$F9,$06,$C6,$7E,$8B,$13,$F9,$4F
	.dc.b	$93,$FB,$B3,$FE,$6F,$57,$9C,$34,$FE,$6B,$E9,$58,$5F,$A4,$C3,$70
	.dc.b	$FF,$3C,$7D,$2F,$D1,$6F,$C6,$7E,$8A,$31,$C1,$FE,$83,$7F,$D3,$D3
	.dc.b	$FE,$7F,$F7,$7A,$B6,$FF,$53,$0C,$2C,$7E,$B3,$E9,$78,$E1,$FE,$8D
	.dc.b	$6F,$07,$F1,$37,$E8,$E1,$FE,$05,$72,$DF,$7F,$91,$CF,$C5,$B1,$FE
	.dc.b	$91,$F9,$5D,$E1,$FE,$E9,$79,$B7,$5F,$DE,$C6,$31,$3F,$BB,$7F,$92
	.dc.b	$3F,$99,$9F,$F5,$13,$9A,$FF,$0C,$39,$9F,$E5,$5F,$4A,$F8,$3F,$E1
	.dc.b	$2B,$E0,$EC,$61,$FE,$1B,$F4,$6F,$A5,$A6,$39,$6E,$5F,$EA,$B9,$D3
	.dc.b	$67,$4D,$26,$7F,$EB,$79,$DB,$6B,$9B,$FE,$73,$FA,$5F,$25,$7F,$72
	.dc.b	$6C,$7F,$B7,$DA,$FF,$9B,$43,$87,$FC,$E0,$F0,$CF,$FA,$57,$D6,$FF
	.dc.b	$9B,$EB,$FA,$04,$E5,$FF,$1F,$46,$6B,$CD,$CF,$F0,$8F,$E5,$FE,$0B
	.dc.b	$CA,$C5,$FC,$19,$B5,$FE,$D3,$AD,$4E,$4F,$F6,$87,$D3,$3E,$03,$F9
	.dc.b	$CB,$63,$3F,$25,$89,$91,$C1,$DC,$62,$FF,$DC,$15,$C5,$B3,$AA,$E0
	.dc.b	$7F,$5D,$FA,$37,$D2,$FA,$6B,$E4,$08,$CE,$7F,$E2,$05,$F3,$AF,$A5
	.dc.b	$C5,$F1,$BF,$A1,$F5,$AF,$A5,$FF,$13,$C2,$30,$FF,$8A,$9A,$F8,$1A
	.dc.b	$FD,$98,$7C,$F7,$F8,$0B,$3F,$F5,$44,$62,$7F,$BB,$35,$FF,$5C,$7D
	.dc.b	$37,$F8,$12,$77,$5C,$67,$FD,$6F,$5B,$E4,$FF,$36,$B6,$BF,$D6,$97
	.dc.b	$C6,$B7,$EB,$71,$7D,$8D,$AF,$EC,$B1,$E2,$F9,$AC,$4C,$8F,$2B,$B8
	.dc.b	$C4,$7F,$B3,$67,$8B,$67,$4F,$F8,$47,$E9,$9E,$0F,$CA,$6C,$7F,$94
	.dc.b	$9A,$FD,$86,$BF,$A1,$73,$97,$FA,$29,$B9,$7F,$B4,$5B,$3F,$13,$D2
	.dc.b	$F7,$7D,$37,$E4,$E3,$FC,$44,$D7,$FD,$C8,$DB,$FF,$4A,$AE,$6E,$3C
	.dc.b	$B2,$B8,$BD,$23,$AF,$B1,$DB,$CC,$95,$DA,$32,$F8,$23,$2F,$82,$32
	.dc.b	$F8,$A3,$38,$78,$A7,$EB,$7D,$37,$F1,$BA,$FF,$1D,$35,$FF,$7E,$36
	.dc.b	$7D,$3F,$51,$C3,$C3,$95,$C1,$F8,$71,$F8,$1F,$A7,$7C,$B5,$F4,$25
	.dc.b	$FF,$91,$E7,$2D,$9C,$AC,$3F,$D7,$72,$AE,$95,$D7,$E9,$65,$F3,$7C
	.dc.b	$1F,$37,$88,$E5,$C6,$66,$0A,$D6,$3D,$AE,$DE,$ED,$7B,$7B,$1E,$99
	.dc.b	$DB,$D2,$EB,$C2,$2C,$F1,$8F,$19,$B3,$91,$D0,$C9,$E0,$9E,$33,$C9
	.dc.b	$B1,$BB,$F8,$B1,$88,$E3,$5D,$E7,$31,$BB,$F6,$4A,$FA,$4E,$4F,$8A
	.dc.b	$E1,$E7,$7F,$31,$D1,$E5,$2C,$F3,$6C,$78,$EF,$C3,$1C,$8D,$D7,$58
	.dc.b	$C9,$F4,$63,$89,$A2,$9C,$4F,$6C,$71,$BF,$5D,$79,$67,$55,$D2,$73
	.dc.b	$1F,$38,$CB,$86,$9F,$00,$EF,$1C,$5C,$1A,$3E,$01,$6F,$E0,$36,$BE
	.dc.b	$05,$75,$AE,$53,$C1,$F7,$3B,$3D,$C5,$3A,$5F,$B5,$FC,$31,$F0,$0F
	.dc.b	$2C,$F5,$3E,$DC,$6E,$3B,$46,$1D,$9C,$A7,$C2,$79,$2F,$F0,$D3,$CE
	.dc.b	$7D,$F3,$9B,$F5,$C7,$19,$D3,$D8,$DB,$8A,$E7,$1B,$32,$53,$81,$1F
	.dc.b	$93,$5D,$E3,$E1,$B5,$F8,$67,$A8,$E1,$7E,$F9,$EF,$1F,$AF,$34,$E0
	.dc.b	$EB,$5E,$5B,$F5,$CF,$85,$B3,$B6,$9D,$6B,$B3,$F8,$DB,$F2,$1F,$26
	.dc.b	$33,$5F,$8D,$8C,$DF,$C3,$9C,$CF,$D9,$3E,$F4,$F1,$7F,$DD,$1D,$AF
	.dc.b	$9F,$FE,$D0,$D8,$CC,$7F,$6B,$FA,$7F,$F5,$94,$60,$87,$F0,$89,$E0
	.dc.b	$FF,$0A,$8D,$B7,$A3,$97,$54,$8F,$F7,$C4,$FF,$6C,$C1,$97,$F2,$18
	.dc.b	$C3,$6B,$85,$73,$38,$F9,$BE,$52,$3D,$3E,$2A,$BE,$38,$B8,$76,$8F
	.dc.b	$E6,$98,$2B,$38,$78,$6E,$D6,$A7,$E2,$F9,$DF,$E2,$6C,$FF,$CF,$73
	.dc.b	$FE,$32,$3F,$C5,$F1,$92,$3F,$39,$E2,$FF,$43,$C7,$C4,$FF,$23,$8F
	.dc.b	$F0,$C6,$BF,$EE,$A8,$D6,$3F,$D0,$E3,$11,$FC,$D9,$B5,$C9,$FF,$5E
	.dc.b	$C7,$03,$FA,$83,$A9,$E5,$CB,$D8,$EF,$CD,$E0,$FA,$1F,$0F,$04,$C3
	.dc.b	$91,$FC,$DE,$34,$E3,$0F,$BF,$8F,$86,$B3,$3E,$5A,$E4,$F9,$E3,$33
	.dc.b	$F0,$3E,$9D,$F0,$FF,$E5,$06,$D7,$CB,$6B,$FE,$5E,$6B,$FE,$87,$73
	.dc.b	$3C,$E7,$FC,$81,$18,$CF,$FB,$57,$D3,$BE,$03,$FC,$88,$FE,$E0,$9C
	.dc.b	$B8,$D3,$2C,$CB,$E0,$CE,$E7,$FC,$99,$F4,$BF,$E5,$19,$FE,$F9,$3F
	.dc.b	$E5,$14,$D8,$D4,$1B,$1B,$35,$E2,$F4,$3C,$B8,$CA,$FA,$35,$CB,$E9
	.dc.b	$9F,$4F,$FE,$76,$6C,$FE,$45,$AF,$FA,$11,$AF,$FA,$B5,$CE,$DF,$90
	.dc.b	$6D,$FF,$83,$DB,$3F,$E1,$7B,$7C,$DF,$C0,$E7,$93,$E4,$7E,$1F,$3E
	.dc.b	$23,$FC,$13,$49,$C6,$3F,$9E,$8F,$D7,$4F,$E3,$3E,$9F,$FC,$A9,$97
	.dc.b	$F8,$7D,$FF,$2A,$46,$7C,$2F,$F3,$D4,$6A,$3F,$C5,$4D,$4F,$07,$FC
	.dc.b	$32,$DB,$FD,$6B,$B2,$4F,$43,$AF,$87,$FE,$8F,$97,$D2,$7F,$10,$BF
	.dc.b	$F2,$12,$ED,$D3,$63,$F5,$D3,$C3,$BB,$97,$D3,$BA,$AF,$54,$BD,$CE
	.dc.b	$5E,$3F,$4D,$FD,$19,$FF,$28,$2F,$A7,$F2,$36,$37,$D1,$E2,$F9,$09
	.dc.b	$1E,$2D,$11,$E6,$F5,$7C,$7F,$4B,$2B,$C2,$F9,$BE,$9F,$AF,$95,$27
	.dc.b	$C6,$F8,$3D,$EF,$7F,$C2,$7C,$0F,$46,$57,$78,$F7,$FC,$C7,$C0,$F9
	.dc.b	$89,$3E,$E8,$F4,$FD,$3F,$AB,$E5,$24,$FC,$08,$F9,$7F,$0B,$23,$D4
	.dc.b	$FC,$EF,$94,$F9,$7F,$3F,$27,$E4,$BF,$3B,$E7,$7E,$2F,$95,$27,$E5
	.dc.b	$BF,$23,$D3,$FF,$61,$7C,$93,$8E,$92,$BF,$8E,$C3,$E9,$F6,$48,$CB
	.dc.b	$FE,$CB,$FA,$27,$35,$24,$FF,$26,$8C,$BF,$CA,$3E,$9D,$F4,$FF,$ED
	.dc.b	$3F,$A7,$FE,$29,$3F,$E8,$36,$A6,$D7,$EC,$23,$A7,$99,$C7,$D3,$91
	.dc.b	$82,$1F,$E7,$89,$E0,$D9,$FC,$E7,$8F,$FC,$4B,$E9,$BF,$91,$4F,$F3
	.dc.b	$C6,$BF,$E8,$06,$F8,$FF,$D5,$2D,$F4,$FF,$EE,$B9,$CF,$99,$FE,$95
	.dc.b	$6D,$FF,$A7,$5A,$AB,$8B,$DF,$8F,$91,$1F,$91,$7D,$7E,$36,$47,$95
	.dc.b	$FA,$1F,$03,$88,$F3,$45,$CB,$CD,$3F,$EF,$FF,$A7,$76,$D2,$5F,$F5
	.dc.b	$A8,$74,$E2,$3F,$D3,$F1,$C1,$D4,$F3,$9F,$F5,$95,$FF,$76,$F4,$9F
	.dc.b	$47,$8D,$FF,$D6,$D1,$8C,$3F,$B3,$B5,$FF,$55,$FD,$37,$82,$FF,$D8
	.dc.b	$5F,$4D,$FC,$8A,$7F,$C2,$18,$73,$7F,$D8,$D7,$C6,$BC,$13,$99,$5F
	.dc.b	$35,$7C,$E3,$0F,$5B,$F1,$36,$BF,$D9,$8E,$F1,$FF,$6F,$5F,$2A,$FF
	.dc.b	$68,$5F,$27,$DB,$7C,$1B,$3E,$A7,$75,$CA,$FA,$61,$9A,$F5,$DF,$16
	.dc.b	$99,$3D,$75,$C6,$F9,$D6,$5F,$49,$E0,$9F,$31,$FE,$EB,$C3,$CC,$DA
	.dc.b	$E4,$F6,$36,$7C,$E3,$C1,$1C,$1F,$0D,$7F,$8E,$63,$E6,$47,$FB,$C4
	.dc.b	$C9,$4A,$65,$76,$BE,$9F,$93,$67,$84,$EA,$72,$7F,$91,$E0,$3E,$9F
	.dc.b	$0D,$34,$E0,$8F,$B2,$FF,$3F,$A7,$CB,$8E,$73,$F3,$67,$31,$F0,$8C
	.dc.b	$9C,$75,$FF,$31,$6B,$F9,$38,$CB,$F9,$59,$E2,$E6,$3C,$8E,$67,$8E
	.dc.b	$9F,$16,$FC,$11,$8A,$F9,$51,$8C,$3C,$D1,$CA,$BF,$5B,$5B,$8F,$86
	.dc.b	$64,$FF,$33,$39,$7C,$B1,$C2,$7A,$43,$0F,$F3,$59,$F0,$0F,$F3,$6B
	.dc.b	$C0,$E7,$1F,$E5,$C8,$D3,$FB,$EF,$13,$FD,$09,$1C,$AE,$CE,$4F,$F3
	.dc.b	$1F,$D3,$FF,$02,$6E,$FA,$C3,$ED,$CF,$DC,$9C,$92,$3C,$CF,$FA,$07
	.dc.b	$B7,$C3,$1C,$2B,$96,$1F,$D3,$B0,$E9,$2C,$3F,$CC,$F8,$7F,$84,$3E
	.dc.b	$9D,$E2,$6A,$73,$5D,$8F,$89,$3A,$AF,$4F,$67,$68,$C4,$EC,$ED,$19
	.dc.b	$BE,$9E,$67,$93,$E9,$DE,$AE,$31,$FF,$10,$C6,$1F,$8B,$3B,$9D,$61
	.dc.b	$E3,$D1,$7B,$8F,$06,$7F,$E1,$8C,$39,$DF,$0B,$FA,$77,$B7,$FD,$4F
	.dc.b	$F2,$4E,$AF,$8B,$F4,$EF,$84,$F8,$F3,$9C,$7F,$D3,$33,$96,$CF,$CC
	.dc.b	$FA,$77,$9B,$CA,$30,$C9,$FF,$50,$FD,$3B,$CD,$C6,$FF,$1E,$31,$3F
	.dc.b	$EA,$6F,$A7,$FA,$3A,$FF,$AF,$1B,$1F,$9C,$5E,$0F,$DC,$73,$7F,$C0
	.dc.b	$BC,$3F,$F0,$B5,$7F,$AC,$1B,$1F,$56,$7F,$C2,$33,$9F,$37,$F0,$BF
	.dc.b	$4D,$FE,$D9,$8F,$F5,$BB,$6B,$EE,$7D,$3B,$CF,$DF,$7F,$E3,$2F,$FC
	.dc.b	$2A,$7F,$AF,$8B,$5F,$5F,$E9,$9F,$11,$BF,$3F,$F2,$7F,$82,$BD,$3A
	.dc.b	$7E,$63,$E9,$DD,$38,$CB,$FD,$93,$F8,$BF,$F8,$B5,$BE,$AE,$FE,$9B
	.dc.b	$F3,$B3,$9D,$7F,$25,$F8,$0B,$E3,$3F,$42,$DC,$3F,$CD,$FF,$53,$D9
	.dc.b	$FF,$6A,$C6,$6B,$F3,$5F,$83,$5E,$AF,$F7,$6B,$6B,$FD,$E1,$6E,$5F
	.dc.b	$D6,$CE,$6F,$F9,$9F,$E2,$EF,$A3,$FD,$8B,$70,$FF,$3F,$FD,$5F,$CD
	.dc.b	$CB,$FD,$F3,$39,$BF,$D5,$7E,$04,$F9,$5F,$DC,$B7,$0F,$F4,$3F,$C0
	.dc.b	$F8,$5F,$F7,$6C,$66,$BF,$65,$E8,$FF,$0D,$6F,$A3,$FC,$13,$EA,$DF
	.dc.b	$65,$FC,$12,$B5,$A3,$1F,$E5,$3E,$65,$7F,$DF,$2D,$C7,$F6,$58,$7E
	.dc.b	$61,$FE,$01,$19,$AE,$51,$FD,$53,$E9,$5A,$7F,$12,$6E,$1F,$EA,$0F
	.dc.b	$99,$F1,$BF,$D6,$AF,$8D,$AB,$6F,$F0,$DE,$97,$D5,$FF,$1C,$6D,$F3
	.dc.b	$E8,$C4,$FF,$20,$F0,$7F,$09,$8F,$46,$D7,$7F,$AC,$7C,$3E,$F1,$FE
	.dc.b	$4F,$F5,$6F,$8A,$A7,$F2,$CF,$AA,$6F,$8C,$7D,$07,$F5,$0A,$E0,$E3
	.dc.b	$6F,$EE,$B1,$87,$F8,$D5,$72,$9D,$42,$75,$A7,$C2,$78,$47,$F8,$2D
	.dc.b	$AF,$7C,$3F,$AA,$BD,$9B,$2F,$AC,$FC,$63,$FD,$51,$B7,$FD,$27,$C2
	.dc.b	$FE,$8F,$18,$BF,$EC,$4D,$AE,$5D,$9F,$D3,$39,$BF,$51,$5F,$CC,$EB
	.dc.b	$59,$BE,$23,$E3,$1F,$1F,$E5,$EE,$7F,$ED,$E9,$D3,$67,$4D,$3E,$1F
	.dc.b	$D7,$FF,$31,$C3,$FC,$50,$D7,$FC,$58,$D7,$FB,$FB,$5F,$F1,$1B,$FA
	.dc.b	$B7,$79,$7C,$65,$19,$78,$C6,$7E,$B1,$F5,$9F,$C4,$A7,$FC,$71,$18
	.dc.b	$DB,$FC,$75,$08,$7D,$17,$E5,$E3,$75,$C5,$FF,$21,$7D,$57,$F1,$37
	.dc.b	$EF,$B5,$FE,$D0,$DE,$0F,$D0,$FD,$5F,$F8,$4B,$4F,$F0,$DC,$E5,$BD
	.dc.b	$5F,$E7,$B8,$C5,$FF,$90,$AB,$83,$FE,$44,$8F,$ED,$EF,$83,$DE,$E0
	.dc.b	$71,$F9,$DB,$19,$21,$E4,$7F,$D1,$1F,$56,$F9,$5F,$07,$A9,$2B,$F3
	.dc.b	$33,$96,$CE,$5F,$6C,$F2,$89,$1F,$0A,$35,$A7,$F0,$0A,$E5,$3F,$E0
	.dc.b	$BF,$AB,$FC,$63,$FA,$66,$B9,$E4,$FF,$4E,$43,$9A,$C4,$9E,$0E,$A3
	.dc.b	$7E,$FF,$0F,$26,$57,$58,$CB,$DE,$32,$F8,$E3,$FD,$55,$F5,$6F,$98
	.dc.b	$F0,$FC,$FC,$AF,$DD,$CE,$5B,$39,$7D,$93,$CA,$24,$7B,$63,$5A,$7F
	.dc.b	$22,$AE,$53,$FE,$2D,$FA,$9F,$ED,$0F,$F0,$85,$F5,$5F,$EB,$D6,$C6
	.dc.b	$7D,$8C,$19,$99,$19,$39,$7B,$7C,$48,$F5,$F9,$EA,$F8,$62,$E3,$FC
	.dc.b	$DB,$05,$7F,$B3,$7A,$D9,$F8,$FA,$A6,$3F,$C7,$A7,$2D,$9C,$BE,$D9
	.dc.b	$E5,$12,$3D,$D1,$AC,$3F,$A2,$D7,$29,$FF,$28,$36,$3C,$93,$B3,$FD
	.dc.b	$07,$19,$6D,$66,$BF,$D5,$38,$7E,$A1,$FD,$DC,$78,$5F,$21,$E2,$3D
	.dc.b	$04,$32,$4B,$F2,$C7,$17,$C7,$1C,$4F,$3C,$66,$BF,$D2,$F5,$E1,$3C
	.dc.b	$57,$E7,$AF,$F5,$EB,$54,$6E,$79,$46,$AB,$D8,$EA,$32,$FB,$0F,$15
	.dc.b	$F9,$1F,$19,$EB,$70,$48,$F8,$F1,$97,$CE,$76,$BF,$21,$E5,$BF,$24
	.dc.b	$EC,$F2,$1E,$09,$E2,$E4,$F2,$1F,$25,$CB,$E4,$25,$78,$DF,$AF,$1E
	.dc.b	$CA,$F0,$3E,$3B,$F7,$DF,$8A,$B8,$CF,$A9,$EC,$72,$3C,$55,$C9,$F3
	.dc.b	$1E,$83,$89,$B3,$89,$E8,$BF,$BE,$F0,$3E,$81,$D2,$FE,$9D,$FD,$C7
	.dc.b	$CB,$7E,$79,$F8,$53,$E7,$AE,$71,$E0,$39,$9E,$0B,$F9,$F7,$D2,$FE
	.dc.b	$13,$67,$C0,$78,$EF,$C0,$F6,$9F,$03,$D6,$3C,$11,$F1,$6F,$C2,$7C
	.dc.b	$F9,$F0,$B8,$8D,$1F,$59,$D5,$F9,$27,$C1,$3E,$46,$9E,$43,$E4,$D7
	.dc.b	$91,$F1,$47,$92,$BB,$47,$CF,$7A,$DF,$C8,$8F,$9D,$7D,$DE,$75,$DD
	.dc.b	$A7,$7A,$F5,$5F,$77,$AC,$7B,$27,$D4,$7B,$1F,$C0,$47,$08,$E8,$D8
	.dc.b	$E9,$1C,$1F,$33,$C5,$FC,$23,$97,$B1,$D1,$D4,$FD,$08,$C0,$1F,$55
	.dc.b	$FD,$E1,$FF,$76,$2E,$7D,$46,$1B,$7F,$E6,$89,$E0,$DE,$8C,$E1,$C2
	.dc.b	$79,$5F,$F9,$41,$AF,$69,$DB,$E3,$87,$D3,$62,$3C,$AD,$8C,$C7,$3E
	.dc.b	$8C,$4C,$8E,$CE,$E3,$FA,$37,$8F,$CF,$6F,$EF,$15,$FE,$62,$6B,$FC
	.dc.b	$F2,$36,$E7,$9B,$81,$87,$46,$48,$60,$67,$FE,$8C,$AF,$F7,$F4,$77
	.dc.b	$E5,$F1,$1E,$3E,$6B,$19,$21,$CB,$5F,$A3,$5F,$EB,$58,$F4,$7D,$27
	.dc.b	$AF,$8A,$46,$08,$71,$3E,$F4,$FF,$3B,$24,$FC,$DF,$AA,$7C,$87,$CF
	.dc.b	$FA,$AF,$EF,$77,$C4,$38,$D3,$FD,$4F,$5C,$27,$FD,$16,$D7,$84,$EF
	.dc.b	$2F,$C4,$B6,$3F,$21,$3B,$AE,$77,$FE,$5E,$94,$F3,$7F,$22,$62,$35
	.dc.b	$19,$8E,$BC,$DE,$92,$3F,$E9,$EB,$FF,$03,$B6,$32,$F9,$9F,$81,$97
	.dc.b	$E3,$FE,$A3,$FC,$BB,$3F,$03,$2F,$88,$67,$C2,$91,$D6,$32,$FF,$AB
	.dc.b	$27,$8F,$D4,$BE,$AB,$A7,$44,$AF,$F5,$8C,$E5,$B3,$97,$E8,$57,$E9
	.dc.b	$B6,$FF,$6A,$D6,$E7,$EA,$35,$F8,$AD,$7F,$6E,$D7,$F7,$CD,$53,$FA
	.dc.b	$A7,$BF,$37,$C1,$3E,$17,$35,$8F,$64,$57,$CB,$1F,$B5,$FA,$AF,$F3
	.dc.b	$2E,$1F,$EE,$A7,$F5,$6D,$E6,$D7,$81,$2F,$FD,$A3,$5A,$D7,$F8,$6B
	.dc.b	$63,$F8,$A3,$D2,$3D,$7D,$BD,$F9,$E5,$7C,$A8,$FE,$0D,$AF,$E3,$61
	.dc.b	$7F,$A7,$7F,$34,$DC,$BE,$5D,$70,$24,$F3,$3C,$BF,$3B,$FC,$47,$EF
	.dc.b	$3F,$C4,$5F,$7B,$FE,$B0,$9F,$F7,$4B,$5F,$D3,$47,$F8,$9F,$D7,$75
	.dc.b	$7F,$1C,$C5,$F5,$9F,$C0,$AF,$F6,$91,$8E,$DF,$E3,$D7,$C2,$7F,$DE
	.dc.b	$BF,$32,$F9,$1E,$5F,$57,$FA,$9C,$0F,$BE,$75,$5F,$EF,$CF,$BC,$79
	.dc.b	$BF,$CA,$7E,$F7,$EF,$60,$E5,$FE,$77,$18,$7F,$9E,$D7,$0B,$FE,$89
	.dc.b	$5B,$7F,$A3,$66,$F9,$2F,$66,$DF,$2F,$0D,$B6,$B7,$3F,$CE,$73,$74
	.dc.b	$FA,$BE,$D6,$57,$BA,$3F,$9F,$7D,$6B,$D6,$FE,$81,$F7,$AF,$A1,$A6
	.dc.b	$A3,$E3,$BE,$2F,$B4,$98,$7A,$35,$FE,$AF,$B7,$18,$F9,$7F,$55,$F2
	.dc.b	$A2,$46,$0E,$07,$F6,$3F,$BB,$F8,$8F,$F6,$5C,$1E,$47,$F5,$36,$C7
	.dc.b	$01,$3C,$15,$3E,$AB,$E3,$FC,$AC,$9E,$13,$97,$EF,$F9,$BF,$7B,$E8
	.dc.b	$A4,$60,$ED,$F3,$5F,$45,$BC,$BF,$D9,$63,$11,$FC,$DD,$7C,$36,$49
	.dc.b	$1C,$9F,$ED,$9F,$58,$F8,$5A,$49,$F6,$46,$5F,$77,$D8,$3E,$73,$FB
	.dc.b	$A7,$D8,$7F,$39,$E4,$FF,$02,$B6,$BF,$C1,$8D,$56,$FB,$0F,$09,$CB
	.dc.b	$69,$3E,$08,$CB,$48,$7E,$A3,$CD,$FE,$6B,$F6,$0F,$A2,$FF,$03,$36
	.dc.b	$7E,$A3,$5F,$EB,$58,$3A,$37,$52,$76,$FC,$48,$C4,$7F,$61,$AE,$5E
	.dc.b	$0F,$E8,$E1,$18,$D3,$FC,$4B,$87,$59,$F0,$F1,$F9,$08,$FC,$CF,$1A
	.dc.b	$BD,$62,$E1,$D6,$BF,$77,$3F,$DC,$7E,$C3,$F3,$4D,$EA,$FF,$12,$46
	.dc.b	$2B,$C8,$FE,$2E,$74,$7E,$55,$D6,$3F,$E3,$A6,$C7,$F9,$05,$E1,$3E
	.dc.b	$DF,$01,$F2,$FF,$E3,$16,$D7,$E7,$DA,$FF,$81,$7E,$C3,$F3,$D5,$FE
	.dc.b	$4B,$FA,$FF,$DF,$CF,$F9,$36,$B7,$3D,$6F,$A1,$FE,$3D,$BE,$34,$BE
	.dc.b	$2A,$33,$18,$BF,$0B,$6B,$FC,$AA,$F0,$9D,$CF,$F9,$61,$B5,$E3,$AD
	.dc.b	$CF,$F9,$70,$E6,$6F,$B7,$FC,$73,$3E,$38,$C6,$1E,$D7,$85,$72,$9E
	.dc.b	$86,$4F,$F3,$65,$F8,$1F,$55,$F4,$C3,$D1,$1B,$3F,$92,$D7,$37,$78
	.dc.b	$66,$F9,$61,$E6,$9E,$E7,$94,$EF,$87,$58,$F3,$C7,$F8,$A0,$E6,$59
	.dc.b	$F9,$EF,$F8,$B1,$AF,$C7,$F3,$DF,$61,$C6,$3E,$94,$61,$E4,$D7,$93
	.dc.b	$5F,$07,$D8,$7F,$8E,$D3,$48,$FF,$8F,$DF,$85,$8F,$F3,$FD,$7E,$24
	.dc.b	$FF,$A5,$67,$3B,$7F,$53,$D7,$F0,$B3,$93,$E1,$1F,$E7,$08,$D6,$7E
	.dc.b	$28,$CE,$1D,$5D,$9C,$6B,$F1,$F1,$97,$AC,$7A,$0E,$35,$D1,$FF,$29
	.dc.b	$DF,$93,$0F,$3C,$6E,$B9,$D6,$E3,$FD,$53,$5B,$7C,$26,$F0,$FF,$55
	.dc.b	$CE,$F0,$F3,$1F,$C3,$CB,$87,$AF,$0E,$AF,$F5,$18,$CD,$FF,$78,$FB
	.dc.b	$0F,$EF,$11,$C6,$ED,$7F,$AE,$26,$D7,$FB,$07,$EB,$BF,$7F,$3F,$D6
	.dc.b	$DB,$1F,$E8,$17,$FD,$8D,$19,$DF,$8C,$A7,$78,$78,$3B,$7C,$73,$9A
	.dc.b	$EF,$3F,$D3,$F1,$E3,$7F,$E9,$0C,$3F,$B8,$CB,$EC,$1E,$5E,$BB,$38
	.dc.b	$C6,$2B,$CD,$18,$9D,$47,$29,$D6,$5E,$9F,$B0,$F0,$FC,$EB,$E1,$FC
	.dc.b	$31,$86,$DF,$8A,$74,$FC,$39,$D3,$4D,$4F,$0A,$FF,$1D,$D7,$07,$FD
	.dc.b	$51,$F2,$6E,$AF,$8F,$F6,$0F,$5B,$E4,$CE,$71,$FF,$05,$4E,$5F,$F0
	.dc.b	$67,$D8,$3E,$C3,$E7,$B6,$37,$78,$99,$1F,$E1,$3F,$AF,$7D,$86,$33
	.dc.b	$FF,$5D,$46,$27,$FC,$31,$F6,$1F,$46,$7F,$83,$35,$FF,$57,$B5,$FF
	.dc.b	$80,$2D,$47,$F7,$1F,$37,$FB,$23,$83,$13,$FC,$27,$C6,$FF,$24,$36
	.dc.b	$77,$3C,$BE,$C3,$FC,$8B,$3F,$C3,$DB,$1F,$C6,$5A,$FF,$B0,$9A,$F7
	.dc.b	$8F,$F0,$A1,$FC,$5C,$AF,$F2,$16,$BF,$EC,$A6,$BE,$7F,$B0,$7C,$9F
	.dc.b	$92,$11,$89,$FE,$44,$D8,$FE,$58,$D7,$FD,$AB,$F5,$FF,$E3,$09,$FE
	.dc.b	$50,$D8,$FE,$64,$D7,$7E,$87,$7F,$AF,$7C,$6D,$25,$FF,$31,$8C,$BF
	.dc.b	$CC,$FE,$B6,$DF,$F3,$D6,$BF,$EE,$66,$BF,$4D,$AF,$D8,$FA,$CF,$F3
	.dc.b	$74,$FF,$3D,$6C,$7F,$47,$6B,$FE,$F0,$6B,$F7,$7E,$B1,$F1,$7F,$1D
	.dc.b	$B5,$FD,$1D,$B3,$FD,$45,$AF,$FB,$EB,$C4,$FE,$1A,$DF,$13,$E8,$46
	.dc.b	$23,$F1,$4D,$7F,$1F,$D9,$FE,$99,$BF,$EA,$8D,$8F,$EB,$CD,$7F,$24
	.dc.b	$1C,$3F,$95,$FA,$97,$D5,$7D,$F6,$D7,$F5,$F6,$CF,$F6,$96,$BD,$37
	.dc.b	$FF,$3E,$CF,$F6,$56,$C7,$F6,$F6,$BB,$F6,$BF,$19,$F5,$5F,$81,$8F
	.dc.b	$D3,$35,$FD,$5B,$53,$7E,$A1,$F5,$1F,$CC,$1B,$87,$F2,$8F,$A7,$7C
	.dc.b	$DE,$6A,$4F,$F7,$9A,$DE,$1F,$E0,$36,$BF,$19,$AF,$EE,$5D,$F9,$5F
	.dc.b	$D2,$BE,$D7,$FC,$F1,$5F,$E9,$F9,$CB,$FE,$C6,$8C,$D7,$FB,$73,$EA
	.dc.b	$7F,$17,$7F,$D4,$9B,$5F,$EC,$CE,$17,$8D,$FD,$12,$73,$6B,$46,$1B
	.dc.b	$F0,$DF,$D2,$AF,$8E,$1F,$D2,$2B,$5A,$BD,$DF,$E9,$ED,$9F,$8D,$F2
	.dc.b	$B8,$99,$1F,$25,$DC,$63,$2F,$14,$FF,$1B,$FA,$9F,$D0,$C7,$F2,$06
	.dc.b	$A6,$98,$FA,$97,$D4,$FB,$A7,$FC,$65,$18,$DB,$FC,$58,$BE,$8D,$CF
	.dc.b	$8B,$E1,$7F,$9A,$FB,$5F,$F7,$A3,$FE,$3B,$6B,$9F,$93,$FF,$1B,$28
	.dc.b	$CB,$FC,$CE,$3F,$86,$BF,$ED,$76,$BE,$E3,$F9,$DB,$8E,$2F,$F4,$F7
	.dc.b	$D5,$7F,$84,$75,$E5,$E1,$FF,$3D,$9F,$F5,$15,$6E,$73,$7F,$EE,$3B
	.dc.b	$E0,$F4,$F4,$B8,$4F,$8D,$B1,$93,$C9,$F3,$F8,$57,$FD,$DD,$F4,$4E
	.dc.b	$63,$11,$E9,$37,$3C,$7E,$A5,$F5,$5B,$F7,$A5,$7F,$BD,$E7,$2D,$9F
	.dc.b	$93,$18,$8F,$F3,$43,$5F,$E1,$6D,$7F,$AF,$3B,$CF,$FD,$88,$DC,$B3
	.dc.b	$5E,$7F,$86,$F6,$7D,$57,$E5,$70,$FE,$D3,$87,$FB,$2E,$31,$1E,$B6
	.dc.b	$C6,$79,$B8,$F3,$CA,$E3,$8F,$F6,$FB,$DB,$6B,$BF,$AB,$E0,$F2,$25
	.dc.b	$F0,$3F,$BC,$DF,$F8,$4D,$C0,$70,$F8,$E3,$8B,$E5,$9F,$AD,$F5,$2F
	.dc.b	$AA,$F6,$FD,$44,$AE,$73,$96,$CF,$E5,$A3,$11,$FE,$95,$6B,$FC,$E1
	.dc.b	$AF,$F8,$2D,$DE,$9F,$EF,$16,$E7,$9A,$F9,$BF,$6B,$F8,$9F,$AB,$FE
	.dc.b	$F7,$87,$B1,$7D,$16,$4D,$7A,$92,$75,$19,$78,$47,$3F,$97,$91,$CD
	.dc.b	$ED,$F5,$18,$8F,$C9,$5A,$71,$A3,$3F,$F1,$24,$AF,$F3,$DD,$ED,$BD
	.dc.b	$15,$31,$FA,$73,$96,$CE,$5E,$F3,$CA,$24,$7B,$E3,$58,$78,$63,$FC
	.dc.b	$6F,$5C,$DF,$D4,$36,$35,$0C,$3F,$72,$DA,$EA,$F0,$BE,$B1,$EC,$33
	.dc.b	$5C,$AF,$A4,$66,$BC,$87,$A8,$87,$A8,$C9,$D4,$CC,$F9,$2F,$C6,$79
	.dc.b	$2B,$AC,$FB,$2F,$C6,$FF,$95,$27,$51,$E4,$BF,$14,$F7,$3B,$12,$FB
	.dc.b	$47,$28,$FE,$69,$3C,$5E,$91,$9B,$F0,$C6,$8F,$4D,$72,$7D,$33,$FE
	.dc.b	$63,$9D,$DF,$AA,$3A,$4F,$13,$C9,$5C,$5F,$24,$71,$BF,$25,$7B,$27
	.dc.b	$A9,$E4,$3C,$33,$B7,$27,$D9,$32,$7C,$73,$31,$E3,$77,$7E,$69,$F7
	.dc.b	$57,$9A,$B5,$5E,$43,$91,$E8,$9F,$9B,$7F,$81,$AE,$11,$CA,$7B,$B5
	.dc.b	$FA,$27,$F8,$DA,$FF,$28,$D9,$FE,$B7,$18,$8F,$61,$FE,$3C,$3E,$41
	.dc.b	$B3,$06,$0D,$D7,$8D,$F8,$7F,$57,$FE,$A4,$7D,$F5,$F4,$A3,$11,$ED
	.dc.b	$3E,$A4,$F1,$7F,$A7,$46,$A3,$04,$BE,$E7,$FA,$86,$34,$E2,$7E,$3C
	.dc.b	$ED,$FF,$54,$CE,$9C,$CE,$DF,$0B,$CF,$DD,$FC,$3E,$1F,$90,$3F,$23
	.dc.b	$5C,$6B,$67,$22,$07,$51,$C7,$0F,$81,$5D,$1B,$5F,$10,$F8,$F1,$96
	.dc.b	$99,$3E,$35,$74,$9D,$47,$C6,$9B,$87,$BE,$3C,$51,$8C,$3D,$D7,$C6
	.dc.b	$7A,$57,$F1,$28,$C4,$7D,$3C,$3D,$F7,$C6,$3D,$B3,$AA,$ED,$5A,$9E
	.dc.b	$77,$DA,$F8,$47,$C6,$3C,$66,$8E,$6E,$A7,$94,$7E,$03,$B3,$FC,$29
	.dc.b	$83,$AE,$98,$89,$3F,$DA,$A7,$4D,$9C,$BF,$8A,$FA,$7E,$64,$78,$76
	.dc.b	$57,$47,$38,$F0,$C7,$F4,$8F,$37,$FC,$4D,$BB,$A6,$99,$BF,$E9,$33
	.dc.b	$FA,$A2,$3F,$A6,$CF,$FD,$0E,$D8,$E0,$FF,$44,$7A,$9D,$B9,$E4,$7F
	.dc.b	$81,$AF,$FC,$0B,$BF,$F4,$87,$FD,$87,$19,$9D,$BF,$E3,$58,$DE,$E8
	.dc.b	$91,$C1,$D7,$D1,$FB,$38,$64,$77,$8F,$79,$87,$A4,$61,$FF,$0C,$FD
	.dc.b	$5B,$E2,$3F,$C9,$5F,$57,$FE,$00,$9F,$F0,$EB,$5F,$F2,$44,$ED,$E3
	.dc.b	$3F,$E6,$A8,$C4,$7F,$0F,$6B,$FE,$CC,$6B,$EB,$0F,$0C,$FF,$B5,$B7
	.dc.b	$2D,$36,$7E,$15,$BE,$03,$CF,$CF,$27,$CE,$FA,$B4,$C4,$7F,$97,$EB
	.dc.b	$FC,$1C,$E8,$F6,$C7,$F9,$A7,$EA,$BF,$CE,$91,$DB,$1F,$86,$67,$C6
	.dc.b	$91,$F1,$23,$2F,$F9,$0A,$78,$AF,$AA,$7D,$5F,$5E,$79,$3F,$E4,$89
	.dc.b	$CB,$67,$FD,$27,$18,$8F,$E7,$0D,$7F,$DE,$AD,$7E,$53,$BD,$BF,$C2
	.dc.b	$6D,$C3,$35,$ED,$F9,$AF,$77,$44,$C3,$D0,$FC,$1E,$8C,$47,$FA,$17
	.dc.b	$34,$E3,$2F,$F0,$FC,$2B,$FD,$21,$F5,$7F,$E6,$AB,$FC,$B3,$FA,$D7
	.dc.b	$FC,$CA,$D7,$FD,$20,$ED,$34,$D5,$7F,$8B,$1B,$97,$E1,$DE,$19,$FE
	.dc.b	$4E,$75,$E1,$F8,$8F,$F5,$14,$66,$BE,$2F,$81,$F3,$09,$87,$F6,$09
	.dc.b	$FC,$E3,$F3,$23,$89,$1F,$4F,$CA,$7F,$AB,$7E,$AB,$FB,$64,$7E,$CD
	.dc.b	$A9,$A6,$3D,$B7,$57,$FB,$39,$B3,$FD,$C1,$AF,$F7,$56,$BF,$A5,$1F
	.dc.b	$20,$F2,$BF,$DA,$ED,$8F,$EF,$2D,$7F,$BF,$35,$E5,$F5,$1F,$ED,$C7
	.dc.b	$F8,$53,$53,$4C,$53,$E2,$7D,$04,$9F,$43,$E2,$E1,$91,$E2,$6B,$FE
	.dc.b	$0F,$FA,$67,$B7,$FE,$EE,$6C,$7F,$84,$1A,$EE,$3E,$37,$CA,$48,$F8
	.dc.b	$CD,$7D,$BF,$3D,$F0,$88,$FC,$27,$FD,$F7,$F4,$9F,$98,$8F,$F0,$EB
	.dc.b	$5F,$F1,$13,$5F,$9C,$D7,$F8,$A6,$4F,$9C,$FB,$04,$8F,$B0,$FD,$2F
	.dc.b	$0E,$47,$D2,$6B,$FE,$2F,$FA,$3F,$94,$FF,$B1,$9A,$FF,$BC,$5B,$D1
	.dc.b	$FE,$8A,$5F,$0B,$FE,$FB,$F9,$F7,$C1,$7F,$BF,$67,$37,$CE,$B5,$32
	.dc.b	$3F,$DA,$75,$C1,$B1,$FC,$BE,$FF,$D2,$FF,$46,$FA,$3C,$76,$7F,$33
	.dc.b	$FA,$3F,$E5,$D8,$1A,$63,$17,$85,$E5,$38,$CF,$F5,$59,$DE,$B9,$8F
	.dc.b	$1C,$D5,$97,$FA,$A2,$F9,$C7,$FB,$B6,$1E,$CF,$F2,$AF,$A3,$FE,$7D
	.dc.b	$3F,$EE,$E6,$BC,$B2,$FE,$95,$18,$AE,$9C,$7E,$4A,$4E,$6B,$7E,$5E
	.dc.b	$63,$13,$FD,$B7,$15,$E6,$A4,$F8,$5F,$A9,$1C,$7E,$6A,$4F,$95,$FB
	.dc.b	$1D,$B8,$8F,$31,$FD,$E7,$04,$E7,$2F,$85,$5E,$7F,$A3,$74,$52,$61
	.dc.b	$FE,$01,$AF,$F6,$7C,$FE,$56,$31,$A7,$F6,$B8,$E3,$18,$3F,$C1,$0E
	.dc.b	$1F,$18,$71,$3D,$39,$BE,$47,$5B,$E3,$E9,$2F,$A0,$FA,$3F,$D3,$5F
	.dc.b	$F8,$57,$E8,$7F,$A2,$4F,$F8,$5E,$F9,$3C,$1A,$F8,$9B,$3F,$EE,$AA
	.dc.b	$E7,$2A,$E6,$F7,$C3,$F1,$2E,$6F,$FC,$14,$74,$9D,$B4,$FF,$07,$57
	.dc.b	$48,$E6,$D8,$E3,$87,$F8,$4E,$78,$B6,$38,$DF,$82,$78,$B6,$38,$DF
	.dc.b	$C1,$6B,$E5,$3F,$C7,$17,$DD,$B3,$C5,$F8,$D7,$DD,$A7,$33,$78,$57
	.dc.b	$E1,$36,$FD,$D1,$C0,$F0,$1A,$2D,$FC,$C7,$FA,$63,$5F,$6C,$FF,$B0
	.dc.b	$A3,$11,$E8,$F1,$1F,$47,$96,$C7,$FB,$2E,$51,$88,$EA,$D7,$FD,$A1
	.dc.b	$F4,$7F,$EC,$E7,$FC,$B4,$D4,$D2,$3E,$AB,$FD,$1E,$BF,$06,$38,$99
	.dc.b	$D9,$19,$8E,$66,$5C,$57,$C9,$AE,$73,$F1,$8E,$7A,$FE,$3A,$32,$FE
	.dc.b	$42,$33,$9F,$F2,$9C,$3B,$C7,$13,$BC,$F1,$8C,$BC,$63,$33,$C7,$0F
	.dc.b	$1C,$79,$4E,$58,$79,$6F,$8D,$6E,$7F,$6F,$87,$A0,$F0,$BF,$E6,$9A
	.dc.b	$E1,$7F,$B4,$9D,$3F,$CF,$FE,$8F,$CF,$1C,$0C,$D7,$F0,$B8,$CC,$FF
	.dc.b	$9C,$FE,$8F,$F8,$03,$FB,$E6,$CF,$F0,$87,$FD,$25,$19,$DB,$8C,$FF
	.dc.b	$A5,$67,$77,$E0,$ED,$EF,$29,$CD,$FF,$1B,$9C,$C7,$FA,$46,$3F,$A1
	.dc.b	$7D,$1F,$F8,$81,$F1,$75,$F6,$C1,$39,$C3,$C9,$39,$8C,$4F,$C6,$BE
	.dc.b	$95,$FC,$CA,$38,$F8,$FE,$88,$F2,$7D,$1F,$5F,$97,$7B,$F9,$C1,$39
	.dc.b	$8C,$37,$3F,$8C,$DB,$F8,$CD,$C7,$C4,$D9,$FF,$50,$7D,$1B,$DC,$F1
	.dc.b	$36,$3F,$96,$FC,$87,$F4,$67,$F9,$7C,$67,$0F,$E7,$3F,$46,$FA,$3F
	.dc.b	$53,$6B,$78,$62,$7D,$DF,$46,$F8,$5A,$66,$BF,$9D,$C6,$5F,$E7,$BF
	.dc.b	$47,$F2,$35,$FC,$7B,$FE,$12,$8C,$F1,$FC,$98,$C6,$7F,$D3,$3C,$4F
	.dc.b	$F2,$45,$7F,$46,$FB,$17,$9B,$F3,$E7,$35,$FD,$4A,$73,$5F,$D3,$3E
	.dc.b	$93,$FD,$18,$DE,$3F,$AD,$18,$AF,$EB,$2F,$D6,$9F,$EA,$67,$F8,$6C
	.dc.b	$FF,$05,$7C,$EB,$E0,$7E,$F4,$66,$BF,$05,$39,$8F,$EB,$7F,$48,$F8
	.dc.b	$CF,$C7,$7D,$9B,$E3,$BF,$0D,$39,$AF,$F0,$B4,$E5,$FE,$C7,$F4,$BF
	.dc.b	$E3,$87,$3A,$FE,$36,$31,$97,$E4,$69,$3F,$91,$FA,$57,$C3,$FE,$77
	.dc.b	$EC,$FA,$5F,$F7,$69,$CE,$3F,$E2,$59,$CF,$2F,$F0,$2F,$A6,$CD,$FE
	.dc.b	$76,$7D,$D3,$9C,$3A,$7A,$FF,$9F,$FA,$6F,$F3,$9E,$3F,$E2,$FF,$B3
	.dc.b	$7D,$9E,$7D,$FF,$D7,$46,$27,$FB,$FF,$D4,$3E,$23,$F6,$5F,$51,$DA
	.dc.b	$7F,$DF,$53,$97,$FC,$75,$19,$D8,$7F,$96,$BC,$CF,$DD,$7D,$9F,$F8
	.dc.b	$BA,$BF,$C1,$7E,$6F,$F0,$4A,$FF,$22,$CE,$5F,$F0,$BF,$97,$FC,$FA
	.dc.b	$BF,$C8,$D3,$F0,$38,$7F,$91,$CE,$6B,$F7,$D3,$9A,$F0,$C3,$E8,$99
	.dc.b	$7F,$0B,$F9,$5E,$D8,$FF,$2C,$9C,$DF,$F9,$4E,$72,$F2,$F8,$37,$C2
	.dc.b	$7F,$13,$FB,$3F,$06,$1F,$CD,$27,$35,$FE,$5B,$9C,$BE,$2F,$03,$F0
	.dc.b	$75,$FC,$D2,$7F,$71,$9F,$F8,$D2,$73,$5F,$E3,$F9,$CD,$7F,$40,$FA
	.dc.b	$97,$D5,$7F,$19,$6D,$7F,$9A,$7E,$A1,$A6,$6A,$4F,$F9,$B7,$EA,$BF
	.dc.b	$CF,$2F,$F8,$EA,$F8,$E7,$FC,$96,$72,$FF,$91,$A3,$38,$FF,$4A,$9C
	.dc.b	$BE,$3B,$FE,$69,$18,$BF,$F2,$B7,$D5,$7E,$3B,$0F,$E6,$0D,$BF,$E8
	.dc.b	$9E,$9F,$F3,$B8,$C5,$70,$E9,$C1,$FC,$FF,$95,$58,$9F,$77,$C3,$FF
	.dc.b	$43,$DB,$FD,$0F,$1F,$07,$E7,$71,$32,$36,$64,$B9,$FF,$98,$BE,$A5
	.dc.b	$F5,$3D,$A3,$FD,$E0,$23,$38,$E2,$31,$F3,$FF,$8B,$77,$3A,$78,$FD
	.dc.b	$4F,$FB,$B4,$9F,$51,$FE,$21,$71,$EA,$52,$4F,$58,$47,$F8,$9E,$7F
	.dc.b	$82,$CF,$F9,$C1,$AF,$F9,$E9,$BC,$3F,$E0,$4F,$A9,$FE,$FD,$A7,$E8
	.dc.b	$27,$2D,$F1,$BF,$D4,$91,$8B,$FD,$AD,$70,$7F,$6F,$1F,$D8,$9F,$07
	.dc.b	$A5,$C0,$E3,$F3,$D6,$32,$43,$C8,$FF,$AC,$7E,$A5,$F5,$3F,$E4,$16
	.dc.b	$CF,$CB,$6B,$F4,$1A,$FE,$E0,$3A,$CF,$FC,$42,$DC,$FF,$7A,$D9,$F4
	.dc.b	$FC,$6F,$86,$E4,$7F,$A6,$27,$4B,$25,$7F,$B1,$FE,$A7,$FD,$A3,$3F
	.dc.b	$E9,$E6,$BF,$EA,$86,$E3,$FE,$32,$5C,$BC,$79,$65,$71,$7A,$47,$5F
	.dc.b	$7B,$B7,$93,$2B,$B4,$65,$F0,$46,$5F,$04,$7F,$B6,$FE,$A5,$F5,$34
	.dc.b	$93,$E7,$F5,$EB,$FE,$BE,$9D,$BF,$EB,$86,$BE,$B8,$DE,$7F,$E4,$C9
	.dc.b	$CB,$6B,$E2,$4F,$FB,$AF,$EA,$7F,$B3,$5F,$FB,$09,$7C,$FE,$4D,$7B
	.dc.b	$52,$75,$19,$78,$47,$3F,$A6,$91,$CD,$ED,$EE,$62,$3F,$DA,$F6,$9C
	.dc.b	$65,$FE,$62,$8F,$EF,$D3,$FE,$D4,$ED,$7D,$47,$FC,$01,$6C,$66,$3F
	.dc.b	$A1,$D7,$F9,$AB,$EA,$7F,$C4,$93,$92,$1F,$EE,$6A,$E2,$FF,$BA,$23
	.dc.b	$6E,$6B,$B1,$1F,$F7,$2C,$6A,$BF,$CE,$F5,$C9,$B5,$BA,$E9,$F0,$96
	.dc.b	$73,$1E,$27,$CD,$19,$E9,$FE,$B7,$7C,$DB,$7E,$C7,$FA,$FE,$1E,$17
	.dc.b	$38,$73,$6B,$E0,$30,$7F,$76,$AE,$8D,$AE,$8F,$F7,$AA,$E8,$E4,$E2
	.dc.b	$27,$8A,$8D,$B9,$9F,$53,$C4,$F1,$1F,$2C,$E7,$3C,$A7,$8B,$D0,$93
	.dc.b	$E4,$9E,$D7,$D1,$EC,$79,$2B,$8C,$E6,$BA,$55,$8E,$13,$AA,$F0,$C7
	.dc.b	$58,$C4,$79,$49,$87,$D4,$76,$77,$AE,$8D,$32,$70,$C3,$EA,$C6,$A3
	.dc.b	$07,$CD,$71,$3F,$82,$8F,$8A,$D2,$4F,$DA,$7C,$51,$C8,$F8,$25,$AF
	.dc.b	$4C,$70,$8F,$A0,$F0,$30,$F0,$6C,$70,$9E,$C4,$9F,$31,$EC,$9D,$D7
	.dc.b	$A2,$75,$3F,$50,$D3,$8A,$E9,$1E,$28,$C4,$F1,$8C,$1D,$8F,$66,$1F
	.dc.b	$80,$7F,$5B,$1B,$7D,$13,$F3,$1E,$E7,$23,$C4,$72,$C3,$EF,$CF,$1A
	.dc.b	$FC,$1C,$F0,$39,$9E,$1C,$3E,$2C,$78,$63,$E9,$DF,$BA,$31,$3F,$2E
	.dc.b	$31,$7D,$AB,$A5,$76,$7C,$47,$E1,$4A,$7D,$7B,$F2,$4F,$E4,$5B,$1F
	.dc.b	$00,$E7,$19,$69,$93,$9D,$FD,$47,$27,$A0,$FD,$D7,$5F,$E7,$7A,$E4
	.dc.b	$7F,$18,$8F,$D0,$18,$7F,$90,$46,$1F,$E3,$BF,$4D,$FA,$52,$3F,$C8
	.dc.b	$3E,$3E,$BF,$C9,$E2,$DF,$F2,$F9,$FE,$69,$E6,$4A,$FE,$70,$FE,$EF
	.dc.b	$83,$FC,$B3,$18,$21,$FC,$CE,$34,$71,$70,$6A,$9A,$FF,$A0,$2B,$9C
	.dc.b	$A7,$2F,$FA,$52,$3F,$DE,$4F,$FA,$5A,$32,$FF,$B2,$79,$BC,$87,$97
	.dc.b	$9E,$C6,$EF,$B4,$E2,$3B,$47,$86,$7D,$9F,$45,$ED,$DA,$57,$13,$C5
	.dc.b	$1A,$70,$7F,$B6,$63,$6F,$43,$66,$BE,$A7,$FC,$8E,$DF,$0F,$F9,$24
	.dc.b	$62,$38,$0F,$A9,$7C,$36,$AA,$4F,$13,$C3,$3A,$70,$72,$AE,$9A,$7C
	.dc.b	$66,$E7,$FE,$AE,$AE,$15,$E0,$7F,$BC,$3D,$39,$FF,$9C,$46,$23,$5B
	.dc.b	$62,$64,$6C,$F1,$C7,$F3,$FF,$A8,$FF,$2C,$47,$AE,$F9,$99,$F1,$A4
	.dc.b	$7A,$A3,$2F,$CC,$9E,$3F,$4E,$FA,$9F,$F3,$E4,$A3,$0F,$F9,$FD,$AF
	.dc.b	$FA,$11,$AF,$F9,$55,$CE,$9F,$88,$6D,$FF,$B4,$1B,$3F,$EE,$2B,$7D
	.dc.b	$0F,$F5,$99,$CC,$F8,$70,$C1,$EF,$9F,$81,$CD,$27,$D4,$FC,$3E,$6C
	.dc.b	$47,$DE,$C1,$38,$CB,$F3,$70,$AF,$ED,$1F,$53,$FE,$5A,$CF,$FD,$CA
	.dc.b	$FF,$96,$B0,$FC,$8B,$5F,$C4,$BB,$7F,$DF,$6D,$4E,$EF,$F5,$2D,$9E
	.dc.b	$BE,$27,$BF,$C0,$F9,$FA,$FF,$92,$30,$F9,$D1,$82,$3F,$A2,$8D,$3F
	.dc.b	$9A,$6E,$1D,$BE,$6B,$A4,$EE,$AB,$E8,$8F,$F0,$37,$D4,$FF,$85,$09
	.dc.b	$CF,$FE,$09,$FB,$2F,$A3,$EE,$3A,$BF,$C1,$8D,$C3,$F6,$6D,$7F,$65
	.dc.b	$93,$EC,$1F,$17,$FE,$13,$F9,$8F,$F4,$5D,$7E,$E9,$AF,$EC,$9B,$5F
	.dc.b	$C4,$87,$D9,$3D,$9F,$F0,$E7,$D6,$7F,$A5,$6F,$F8,$1B,$5F,$DD,$B6
	.dc.b	$BF,$8B,$FD,$97,$4E,$45,$4D,$1F,$5B,$EC,$24,$7F,$11,$BF,$57,$C6
	.dc.b	$FA,$7A,$24,$FA,$5F,$F2,$0C,$BF,$6A,$FF,$73,$4F,$F0,$CF,$AE,$F4
	.dc.b	$3F,$C6,$1A,$E7,$5F,$F1,$B7,$D9,$27,$E6,$FD,$1C,$AF,$9A,$F6,$E3
	.dc.b	$7C,$47,$CB,$FA,$B9,$5F,$2D,$F1,$E4,$F5,$3F,$C8,$5F,$5F,$FA,$0A
	.dc.b	$FE,$52,$D7,$F9,$0B,$75,$FE,$7B,$93,$ED,$FE,$FF,$D5,$4B,$FB,$EF
	.dc.b	$8F,$D0,$FB,$9E,$6C,$9F,$B8,$F8,$BB,$3F,$CA,$5F,$3B,$E8,$9F,$E7
	.dc.b	$2F,$D0,$6F,$36,$B8,$3F,$CB,$1F,$5C,$F8,$8C,$57,$09,$CC,$7F,$97
	.dc.b	$7E,$C1,$D3,$8A,$FE,$8B,$7C,$5F,$E8,$F3,$A7,$FA,$4F,$D8,$7F,$34
	.dc.b	$7F,$06,$D4,$D9,$FE,$27,$1D,$3C,$CF,$F3,$64,$EC,$D3,$E8,$24,$6C
	.dc.b	$FC,$57,$93,$FC,$D3,$EC,$3F,$A0,$F2,$7F,$6A,$6F,$7F,$E5,$29,$18
	.dc.b	$7F,$2F,$8B,$A7,$FC,$F9,$39,$9D,$FA,$98,$3E,$4C,$D8,$FE,$CF,$3B
	.dc.b	$38,$57,$38,$CF,$D8,$74,$23,$F2,$BE,$0E,$BF,$D9,$A9,$38,$CB,$FC
	.dc.b	$07,$7F,$AD,$AF,$D6,$FD,$83,$AB,$FD,$24,$D8,$FE,$F2,$D7,$FB,$8A
	.dc.b	$E9,$DC,$C9,$DB,$CA,$3F,$D3,$9E,$13,$E6,$29,$2B,$FC,$0B,$F5,$FC
	.dc.b	$F2,$FD,$FF,$D7,$FF,$3D,$9F,$E0,$37,$C5,$FF,$05,$B5,$FF,$01,$B6
	.dc.b	$7F,$C3,$F5,$CE,$55,$CD,$46,$63,$15,$CD,$B7,$F7,$5E,$13,$B9,$FC
	.dc.b	$03,$6B,$CB,$5B,$9F,$C2,$1C,$CD,$E1,$FE,$37,$9F,$1C,$63,$0F,$7B
	.dc.b	$C2,$B9,$4F,$43,$27,$F2,$2B,$F0,$3E,$AB,$E9,$87,$A2,$36,$7F,$65
	.dc.b	$AE,$6E,$F0,$CD,$F2,$C3,$CD,$3D,$CF,$29,$DF,$0E,$B1,$E7,$8F,$D0
	.dc.b	$19,$AA,$D2,$D7,$D1,$76,$BC,$F7,$D8,$70,$9F,$A7,$18,$78,$B5,$E2
	.dc.b	$D7,$BF,$D8,$7F,$87,$93,$48,$FF,$88,$9F,$87,$7F,$9D,$E4,$F8,$B3
	.dc.b	$FD,$32,$73,$AF,$F9,$AB,$5F,$C3,$4E,$4F,$86,$7F,$9B,$E3,$59,$F8
	.dc.b	$A3,$38,$75,$76,$71,$AF,$C8,$46,$5E,$B1,$E8,$38,$D7,$47,$E6,$DF
	.dc.b	$93,$0F,$3C,$6E,$B9,$D6,$E3,$F8,$95,$6D,$F0,$9B,$C3,$F8,$B6,$1E
	.dc.b	$16,$D7,$F7,$7C,$29,$FE,$76,$AF,$44,$E1,$FF,$37,$D7,$B4,$FF,$45
	.dc.b	$8F,$B0,$E5,$1F,$E8,$FB,$AF,$FA,$56,$2F,$D7,$FF,$01,$C3,$FD,$06
	.dc.b	$D9,$FF,$3D,$BF,$DF,$A3,$3B,$F0,$3A,$E1,$FD,$37,$0C,$4F,$F7,$FE
	.dc.b	$CF,$23,$FC,$CE,$DE,$0F,$0F,$B0,$7A,$1E,$28,$C3,$6B,$FC,$11,$3A
	.dc.b	$7C,$B3,$A7,$CF,$F6,$1E,$2F,$A0,$7C,$2F,$B2,$30,$D9,$FF,$09,$4E
	.dc.b	$9E,$26,$23,$5B,$7F,$8C,$27,$4F,$C2,$AE,$75,$FE,$A8,$FB,$07,$C2
	.dc.b	$FC,$C6,$C7,$FA,$B3,$E5,$BF,$9E,$BB,$9F,$A3,$F6,$0F,$B0,$A7,$18
	.dc.b	$B3,$FE,$B4,$FB,$07,$C5,$FD,$86,$C7,$FA,$E2,$F9,$3B,$9F,$BB,$F6
	.dc.b	$1F,$3A,$3F,$D7,$AF,$F8,$05,$BB,$EB,$8F,$F0,$7E,$2F,$F8,$E2,$7F
	.dc.b	$D8,$7E,$E3,$CE,$CF,$9D,$F8,$88,$C4,$FF,$B2,$3E,$C3,$FC,$FC,$DE
	.dc.b	$5F,$C6,$C6,$39,$3F,$D8,$6F,$CA,$EA,$FF,$29,$7D,$1B,$ED,$FF,$12
	.dc.b	$BE,$52,$C6,$1F,$C6,$FD,$7F,$F9,$9F,$2F,$F2,$52,$F8,$3C,$FA,$BA
	.dc.b	$77,$EA,$70,$F5,$6D,$F0,$7F,$2F,$F5,$BC,$F6,$FF,$36,$B7,$5F,$F3
	.dc.b	$33,$D6,$BF,$CB,$0E,$5F,$F7,$C3,$FE,$48,$73,$5D,$3D,$7E,$AF,$E9
	.dc.b	$3E,$AF,$FC,$1D,$3F,$E6,$16,$E1,$FE,$7C,$5F,$37,$BF,$57,$F5,$7F
	.dc.b	$53,$D6,$BF,$D0,$2B,$1C,$E3,$A7,$F6,$3F,$55,$CB,$2F,$F3,$73,$F8
	.dc.b	$1B,$FF,$44,$B9,$6F,$5F,$EE,$3F,$15,$55,$FB,$BC,$34,$FF,$3C,$61
	.dc.b	$C7,$7F,$F4,$53,$63,$FD,$03,$1E,$5B,$FE,$07,$39,$C1,$2F,$AB,$61
	.dc.b	$FC,$21,$B8,$7F,$A1,$5F,$F7,$33,$63,$5D,$B2,$35,$F3,$B7,$0F,$F5
	.dc.b	$2F,$D4,$7F,$A6,$E7,$FD,$1C,$D8,$FF,$54,$F8,$0F,$1F,$F8,$BC,$E7
	.dc.b	$3F,$F5,$23,$6F,$D6,$DB,$F4,$DF,$19,$EB,$F1,$1C,$88,$FF,$1F,$C2
	.dc.b	$D7,$FA,$E3,$EA,$1E,$FF,$F2,$66,$D7,$FA,$8D,$7D,$0E,$66,$4E,$5E
	.dc.b	$11,$8A,$F9,$73,$FE,$5D,$FB,$4F,$74,$7F,$80,$1A,$E7,$1F,$BA,$FF
	.dc.b	$80,$A3,$2F,$FA,$E3,$EA,$1F,$53,$FE,$AE,$8F,$F3,$44,$63,$3F,$F0
	.dc.b	$12,$FA,$17,$D6,$E2,$F8,$5F,$FB,$0B,$B1,$F3,$9F,$D7,$34,$FF,$39
	.dc.b	$46,$27,$F0,$F1,$89,$D4,$7F,$B2,$BE,$A7,$FA,$CC,$E6,$36,$75,$9D
	.dc.b	$BF,$A8,$6C,$F0,$8D,$1C,$0D,$16,$BC,$4F,$CA,$3C,$15,$D2,$3F,$A4
	.dc.b	$C6,$FC,$9C,$19,$FA,$3B,$19,$21,$CC,$FD,$DB,$E6,$FA,$87,$D4,$F5
	.dc.b	$F7,$E5,$79,$67,$2D,$9F,$F1,$82,$8C,$BF,$E3,$A6,$BF,$E3,$E6,$BF
	.dc.b	$E9,$FF,$A9,$FC,$7B,$F0,$5A,$F9,$5B,$D3,$F8,$48,$47,$E4,$9E,$1C
	.dc.b	$DF,$3A,$37,$A7,$58,$C7,$36,$59,$5C,$0F,$04,$71,$F5,$B9,$7D,$34
	.dc.b	$9F,$24,$65,$F3,$46,$5F,$34,$7D,$EF,$A9,$FE,$89,$1E,$BF,$2F,$C8
	.dc.b	$67,$CE,$91,$FE,$57,$9D,$BF,$E6,$06,$BF,$E6,$46,$BF,$EC,$5F,$A9
	.dc.b	$78,$1F,$8C,$5F,$3F,$4C,$44,$9F,$C6,$4E,$9B,$39,$7F,$05,$1D,$1C
	.dc.b	$90,$CC,$F7,$FA,$69,$1D,$1F,$17,$D3,$62,$39,$C5,$E1,$FB,$16,$AF
	.dc.b	$D1,$F6,$7F,$63,$BF,$2F,$0F,$AC,$CE,$B2,$32,$43,$F3,$13,$C5,$FC
	.dc.b	$D7,$D4,$FE,$89,$BC,$FF,$B0,$83,$7A,$24,$FE,$BE,$72,$D9,$CB,$D2
	.dc.b	$79,$44,$8E,$11,$AC,$FF,$29,$5C,$A7,$C2,$68,$86,$09,$1C,$27,$96
	.dc.b	$3F,$F0,$06,$B9,$4F,$4B,$CD,$84,$FB,$E7,$51,$E9,$3B,$BD,$5C,$1E
	.dc.b	$0A,$ED,$7D,$23,$9D,$FB,$5F,$80,$72,$9F,$24,$66,$BA,$4E,$8E,$87
	.dc.b	$96,$FD,$35,$DA,$39,$19,$AE,$51,$98,$E5,$5E,$A7,$99,$EE,$BF,$35
	.dc.b	$F4,$8E,$47,$43,$47,$BA,$79,$1E,$62,$57,$AA,$34,$F8,$63,$31,$D0
	.dc.b	$F0,$CF,$CD,$9D,$1C,$63,$35,$E4,$BF,$5C,$7C,$18,$C3,$D8,$C1,$E7
	.dc.b	$AF,$53,$F4,$0C,$47,$27,$07,$C4,$AF,$0B,$E5,$BF,$6C,$62,$3E,$09
	.dc.b	$E8,$BF,$8F,$5C,$EF,$B9,$EB,$9F,$19,$B3,$C2,$79,$CF,$0D,$FC,$47
	.dc.b	$AB,$E0,$7C,$57,$E1,$9E,$91,$C2,$3E,$34,$F1,$9C,$9E,$2B,$F2,$D7
	.dc.b	$78,$F2,$DF,$C8,$8F,$9A,$7C,$79,$F5,$46,$6B,$C5,$3E,$4B,$F1,$56
	.dc.b	$A3,$81,$D6,$B8,$4F,$90,$F4,$3F,$16,$7C,$73,$F0,$CF,$A0,$7C,$6A
	.dc.b	$EF,$7E,$68,$EA,$79,$8D,$9F,$04,$F8,$A7,$1B,$F2,$BE,$33,$8C,$F4
	.dc.b	$BF,$2C,$70,$3D,$87,$13,$15,$E1,$7B,$BF,$2D,$FB,$27,$A6,$7E,$84
	.dc.b	$FB,$63,$31,$EA,$BF,$D5,$FD,$37,$F2,$52,$4F,$F2,$3E,$9A,$FF,$2C
	.dc.b	$8B,$5F,$C7,$A7,$F9,$CF,$93,$FE,$1C,$D7,$FB,$95,$FF,$82,$DB,$5F
	.dc.b	$DD,$F4,$FF,$41,$4F,$F8,$75,$AE,$F9,$B9,$1D,$1F,$F1,$15,$7F,$B5
	.dc.b	$9B,$CB,$DC,$F0,$73,$58,$C9,$0F,$13,$97,$C9,$17,$0F,$F0,$55,$2B
	.dc.b	$D5,$E9,$FA,$F8,$66,$1E,$28,$CB,$E4,$8C,$BE,$F8,$FF,$1E,$FD,$4B
	.dc.b	$CC,$FF,$1F,$FD,$4F,$F6,$FA,$FF,$17,$37,$2F,$F0,$D5,$6E,$7F,$C8
	.dc.b	$0D,$7F,$9F,$11,$C3,$80,$E1,$D5,$7F,$8C,$A9,$CD,$E2,$24,$FF,$93
	.dc.b	$63,$FB,$93,$63,$2F,$99,$F4,$E7,$8C,$3F,$C6,$15,$D2,$B8,$1C,$E7
	.dc.b	$FC,$BB,$F5,$1F,$E5,$6A,$DF,$37,$C2,$33,$E1,$4B,$F1,$46,$5F,$F3
	.dc.b	$05,$71,$8F,$83,$F4,$EF,$A9,$EB,$CF,$2F,$FC,$CD,$39,$6C,$FF,$94
	.dc.b	$A3,$11,$F1,$C3,$87,$E5,$35,$F9,$F6,$E7,$F5,$92,$3C,$BC,$F5,$FF
	.dc.b	$34,$4E,$DF,$4C,$FC,$29,$D6,$7F,$E5,$7A,$D5,$70,$AD,$3F,$E7,$FF
	.dc.b	$A9,$FF,$2B,$BF,$DF,$5F,$D4,$E7,$F8,$26,$C6,$7C,$0F,$F4,$5C,$70
	.dc.b	$83,$A6,$F3,$EA,$7F,$D2,$39,$FE,$46,$FA,$73,$72,$33,$DD,$30,$F9
	.dc.b	$07,$C8,$7E,$61,$1C,$E2,$E6,$F1,$92,$3D,$9C,$F5,$FF,$47,$4E,$DF
	.dc.b	$81,$3E,$A9,$D6,$DF,$E7,$DA,$D5,$70,$AD,$3F,$EA,$C9,$7D,$CF,$E1
	.dc.b	$E3,$FD,$EE,$D8,$CF,$1F,$FA,$D9,$7D,$DF,$FA,$BD,$7D,$1F,$D0,$6E
	.dc.b	$FB,$8C,$1F,$42,$1F,$47,$FE,$4F,$7F,$DA,$A6,$34,$ED,$19,$DF,$FC
	.dc.b	$6D,$3C,$3E,$0F,$C5,$8B,$D5,$BC,$F3,$6F,$64,$E5,$F7,$66,$FB,$EF
	.dc.b	$56,$BF,$C0,$72,$F0,$B7,$3E,$F5,$BB,$F8,$DA,$3E,$97,$CB,$F2,$52
	.dc.b	$73,$1E,$4F,$0A,$47,$89,$F3,$F9,$FE,$0F,$42,$4F,$D1,$CF,$FD,$D6
	.dc.b	$DF,$93,$F8,$3F,$29,$23,$C2,$FC,$4F,$9A,$F8,$3F,$15,$F0,$BE,$8E
	.dc.b	$57,$C0,$8F,$6F,$DE,$FB,$BE,$86,$4F,$78,$F9,$BF,$37,$F3,$BC,$C9
	.dc.b	$3E,$E8,$FF,$76,$DB,$E1,$2F,$5C,$AF,$F7,$64,$E5,$FF,$77,$7C,$FB
	.dc.b	$E0,$38,$DC,$9C,$C7,$FB,$E3,$E9,$1F,$17,$FE,$FA,$BE,$15,$A3,$FD
	.dc.b	$F1,$3A,$7F,$DF,$5F,$49,$FC,$EF,$B3,$FD,$B0,$DE,$CF,$E7,$AD,$9F
	.dc.b	$F3,$E5,$BE,$93,$B1,$FD,$0A,$F8,$FA,$57,$0E,$2E,$AF,$D7,$55,$FF
	.dc.b	$7E,$5F,$37,$EC,$78,$9C,$5F,$F4,$57,$D2,$3E,$1F,$10,$8C,$3F,$9A
	.dc.b	$C7,$FA,$9C,$62,$3A,$FC,$67,$08,$C6,$7F,$EF,$56,$D7,$0B,$F9,$15
	.dc.b	$B6,$5F,$53,$C6,$E9,$C1,$E3,$FA,$4E,$C4,$7E,$CF,$3D,$58,$F8,$67
	.dc.b	$0B,$0F,$ED,$79,$AF,$DF,$F4,$8F,$0A,$93,$1F,$0C,$3A,$F1,$1D,$B0
	.dc.b	$FE,$EB,$3C,$63,$75,$D2,$BE,$5F,$A6,$EE,$A4,$AF,$6F,$D2,$75,$D3
	.dc.b	$FD,$79,$F4,$6B,$FF,$00,$9C,$6B,$E1,$B5,$F1,$1F,$EC,$37,$13,$EF
	.dc.b	$8F,$F6,$35,$75,$55,$D5,$CE,$1F,$21,$CD,$F8,$30,$FF,$08,$D6,$E7
	.dc.b	$C9,$18,$9F,$F0,$AD,$F5,$AD,$36,$39,$57,$B1,$AF,$63,$C3,$86,$5A
	.dc.b	$64,$EE,$64,$A6,$67,$B3,$C2,$BC,$D5,$F1,$5B,$7E,$79,$D4,$FA,$30
	.dc.b	$F8,$15,$A9,$ED,$3C,$4D,$61,$E1,$C3,$E5,$4F,$F8,$C5,$AE,$8F,$F0
	.dc.b	$73,$87,$E8,$F6,$3E,$93,$FB,$5C,$A3,$13,$C1,$AF,$56,$BF,$E1,$9F
	.dc.b	$A4,$FE,$E6,$FF,$C0,$16,$A6,$91,$F9,$57,$F9,$DD,$BE,$54,$71,$33
	.dc.b	$AF,$E1,$F3,$FB,$D1,$97,$F0,$15,$C6,$3E,$A1,$9C,$FF,$8A,$57,$17
	.dc.b	$CD,$1D,$2F,$BB,$E5,$30,$72,$31,$7D,$AB,$94,$7F,$97,$70,$FC,$06
	.dc.b	$1E,$1A,$F2,$CF,$0C,$3C,$38,$76,$BD,$E1,$CA,$BF,$81,$BF,$DD,$A7
	.dc.b	$71,$FA,$83,$FC,$DB,$3C,$23,$F8,$7E,$1E,$F3,$F5,$53,$FC,$33,$C2
	.dc.b	$E4,$FF,$7C,$BA,$E6,$BF,$8B,$FD,$27,$EF,$DB,$BE,$A1,$36,$24,$FF
	.dc.b	$86,$FB,$7C,$0F,$FA,$02,$B9,$57,$EF,$B1,$FE,$AD,$87,$3E,$AF,$1C
	.dc.b	$3E,$8D,$F4,$99,$C7,$FC,$4D,$18,$7C,$B5,$C2,$31,$87,$C3,$C7,$FD
	.dc.b	$31,$5C,$2F,$D1,$F4,$7D,$BE,$91,$F0,$78,$AF,$F1,$94,$61,$F8,$33
	.dc.b	$B9,$D6,$1E,$9A,$D3,$DE,$75,$5C,$0E,$75,$F3,$3E,$91,$E3,$D3,$E7
	.dc.b	$FC,$83,$B7,$17,$FD,$1B,$E8,$9F,$49,$F5,$1B,$1F,$D2,$B0,$E6,$EE
	.dc.b	$7E,$CF,$D2,$3E,$92,$9C,$62,$CF,$F5,$2F,$A4,$F8,$BB,$7E,$25,$FF
	.dc.b	$04,$46,$78,$FF,$07,$18,$9F,$C7,$78,$6F,$37,$F0,$FF,$48,$F3,$BF
	.dc.b	$13,$39,$C7,$FA,$FC,$E5,$FE,$C3,$F4,$9F,$D0,$DB,$C7,$F9,$08,$C6
	.dc.b	$3F,$9B,$7E,$6F,$4F,$F9,$53,$E9,$1F,$49,$F8,$65,$F2,$76,$30,$FE
	.dc.b	$2B,$E9,$3F,$CF,$59,$7F,$9B,$57,$C0,$E7,$E0,$F4,$EF,$D4,$E1,$EA
	.dc.b	$DB,$E0,$FE,$53,$E9,$39,$ED,$F9,$96,$EB,$FE,$68,$7A,$D7,$F9,$ED
	.dc.b	$CB,$FE,$68,$7F,$C7,$4E,$6B,$A7,$C8,$F5,$7F,$43,$F4,$8F,$4E,$98
	.dc.b	$F9,$AB,$7F,$E8,$7F,$A4,$E7,$9F,$EA,$D6,$3F,$AA,$73,$9F,$F9,$F1
	.dc.b	$CD,$7F,$A3,$FE,$91,$87,$EC,$1F,$C1,$D7,$FA,$09,$CC,$EF,$AF,$F6
	.dc.b	$BF,$49,$FE,$11,$8F,$F4,$C3,$FE,$81,$69,$AF,$89,$FD,$DE,$5C,$B3
	.dc.b	$FE,$00,$DC,$7F,$D1,$EF,$82,$BF,$D1,$AE,$5F,$F4,$7F,$D2,$75,$AF
	.dc.b	$E1,$0D,$AF,$F4,$BB,$FE,$EB,$9F,$F5,$53,$97,$8F,$5F,$F0,$C8,$FF
	.dc.b	$22,$46,$BE,$6B,$13,$FC,$43,$E9,$1F,$47,$4F,$E2,$7E,$83,$C9,$C6
	.dc.b	$1F,$EB,$7A,$DB,$6B,$73,$AC,$FF,$8F,$56,$9B,$3A,$AE,$07,$FB,$07
	.dc.b	$E9,$1F,$0B,$FC,$89,$B1,$9F,$94,$C4,$C8,$E0,$EE,$31,$7F,$D0,$FE
	.dc.b	$8D,$F1,$3E,$98,$FE,$6E,$48,$CE,$7F,$EE,$45,$F1,$5F,$EA,$29,$D8
	.dc.b	$8D,$B8,$1F,$5D,$F8,$A7,$FD,$DA,$D7,$4D,$79,$B5,$F1,$F4,$7F,$B7
	.dc.b	$DF,$C9,$FC,$2F,$F9,$E2,$31,$F4,$9F,$D6,$39,$7F,$AC,$C6,$C9,$3F
	.dc.b	$E7,$E9,$D6,$DD,$23,$91,$2F,$FC,$E3,$3A,$7F,$AC,$D7,$F4,$27,$FC
	.dc.b	$EA,$D7,$FB,$63,$6F,$C3,$F3,$5E,$5F,$26,$57,$07,$D1,$1F,$E9,$0F
	.dc.b	$A3,$7D,$27,$F8,$79,$B5,$F2,$43,$87,$E6,$35,$FA,$2E,$6F,$FB,$CB
	.dc.b	$73,$FB,$6D,$9F,$5F,$C3,$FF,$A8,$3E,$93,$F1,$71,$FE,$90,$6B,$FE
	.dc.b	$07,$6D,$FF,$81,$17,$1F,$1E,$49,$5C,$5E,$91,$88,$E5,$16,$B9,$4F
	.dc.b	$87,$D8,$F1,$7C,$D4,$BF,$14,$65,$F2,$46,$5F,$24,$7F,$AE,$3E,$93
	.dc.b	$F9,$EC,$7F,$AA,$1A,$FF,$87,$DB,$7E,$9F,$A7,$E1,$E7,$4A,$E0,$FC
	.dc.b	$38,$FF,$61,$FD,$2B,$C0,$BF,$3B,$2F,$FD,$6B,$39,$6C,$E5,$FF,$17
	.dc.b	$42,$79,$CC,$9E,$31,$BC,$3F,$C5,$F5,$CE,$78,$FD,$1F,$88,$8F,$83
	.dc.b	$E8,$AB,$A3,$C7,$1E,$88,$FF,$6D,$F6,$7F,$5A,$9F,$F6,$5B,$5F,$F2
	.dc.b	$13,$6F,$FC,$7F,$F4,$BF,$E0,$39,$DF,$A3,$C8,$8F,$FB,$52,$34,$FF
	.dc.b	$92,$DF,$D0,$73,$7F,$08,$5D,$1F,$EE,$C6,$CF,$1E,$9C,$19,$99,$19
	.dc.b	$21,$D5,$CB,$DE,$38,$BE,$38,$CE,$1E,$A9,$CC,$F1,$34,$74,$8D,$38
	.dc.b	$34,$75,$35,$7E,$BA,$EA,$FB,$6B,$AD,$7C,$89,$CD,$7F,$97,$AF,$E1
	.dc.b	$B6,$B2,$A3,$E1,$BD,$C9,$5C,$DF,$23,$E6,$73,$7E,$37,$2F,$8D,$F3
	.dc.b	$9E,$1B,$EF,$5E,$28,$F0,$5E,$27,$D3,$5C,$67,$E0,$DF,$85,$EE,$7A
	.dc.b	$4E,$E7,$53,$B9,$D0,$C9,$ED,$9E,$AF,$13,$AC,$FA,$5E,$A7,$13,$27
	.dc.b	$53,$37,$E9,$3E,$3D,$79,$4F,$88,$49,$E8,$E5,$F6,$39,$9E,$8E,$5F
	.dc.b	$23,$97,$D0,$7A,$A7,$C1,$7F,$1E,$3D,$F1,$E3,$BF,$9B,$3E,$E9,$E2
	.dc.b	$7A,$8F,$65,$75,$AF,$4D,$F9,$8F,$29,$27,$CA,$E5,$F2,$B9,$7A,$CF
	.dc.b	$A5,$EB,$19,$9E,$B1,$C8,$E2,$43,$8C,$75,$78,$99,$8E,$2E,$8C,$57
	.dc.b	$C1,$7E,$9B,$6B,$D8,$DA,$F1,$36,$78,$9F,$2E,$FE,$35,$FC,$98,$F9
	.dc.b	$D3,$89,$F9,$71,$F3,$9A,$FE,$09,$B7,$F2,$DB,$7F,$05,$B3,$EB,$6B
	.dc.b	$EE,$3D,$84,$BF,$BA,$F9,$4F,$BB,$7E,$97,$84,$79,$DF,$B8,$F0,$6C
	.dc.b	$FB,$27,$84,$E6,$FE,$B3,$F3,$0F,$CD,$7D,$17,$F2,$32,$5F,$F2,$BF
	.dc.b	$8B,$AF,$F2,$B8,$B5,$FC,$D2,$7F,$9D,$F9,$32,$BF,$9D,$BF,$E3,$69
	.dc.b	$FF,$7B,$EB,$FE,$7C,$9C,$90,$FE,$7F,$3B,$CF,$FC,$7B,$3B,$95,$72
	.dc.b	$E5,$D4,$7F,$BE,$27,$FD,$A6,$FF,$BE,$A7,$4D,$AE,$7C,$DC,$21,$5C
	.dc.b	$1E,$B3,$E0,$BD,$4E,$6F,$D3,$78,$99,$7D,$4F,$13,$C2,$38,$FA,$9C
	.dc.b	$B7,$95,$B3,$C3,$1A,$71,$87,$F5,$FC,$7E,$0D,$66,$7C,$A0,$7D,$2B
	.dc.b	$CE,$F6,$36,$3B,$35,$E2,$3E,$95,$E6,$70,$A9,$3C,$1F,$8D,$3A,$8F
	.dc.b	$42,$E5,$C9,$23,$39,$57,$BC,$ED,$E3,$3D,$27,$5B,$33,$FF,$3D,$DE
	.dc.b	$EF,$8D,$ED,$F9,$1F,$4B,$FE,$6D,$9F,$F3,$04,$7A,$EB,$EE,$4F,$F9
	.dc.b	$C4,$D3,$EA,$33,$E2,$FE,$07,$E9,$5F,$47,$FE,$59,$B4,$61,$FF,$2E
	.dc.b	$35,$FF,$43,$35,$FF,$45,$B9,$CB,$D5,$1F,$56,$31,$7F,$E9,$66,$C7
	.dc.b	$D9,$C1,$F3,$3E,$8E,$69,$3E,$E7,$D7,$CD,$88,$FB,$34,$9C,$63,$E0
	.dc.b	$94,$AB,$38,$FF,$A9,$A3,$84,$F5,$3E,$5C,$FE,$8F,$E9,$7F,$CD,$35
	.dc.b	$FE,$AC,$7F,$CD,$33,$F9,$76,$BF,$8A,$76,$FE,$4D,$A9,$DD,$F9,$5B
	.dc.b	$3D,$7C,$3F,$A5,$DD,$F4,$74,$FF,$29,$61,$F4,$23,$04,$7F,$51,$1A
	.dc.b	$7F,$36,$DC,$3B,$7C,$A7,$AF,$9E,$57,$A5,$F7,$73,$E2,$3D,$51,$6B
	.dc.b	$CD,$37,$2F,$85,$36,$BD,$D3,$FC,$42,$9F,$4A,$E2,$C3,$6B,$F8,$24
	.dc.b	$EA,$78,$7D,$2F,$FA,$DE,$1F,$49,$F2,$9B,$B7,$58,$C1,$F2,$2F,$F9
	.dc.b	$0F,$D2,$7F,$05,$6E,$5E,$38,$C5,$97,$C3,$F8,$3D,$9F,$0F,$3C,$C3
	.dc.b	$CB,$9A,$7C,$1F,$4B,$ED,$D3,$FD,$CF,$97,$E7,$0F,$F5,$7D,$FF,$12
	.dc.b	$7D,$33,$E1,$FA,$5F,$17,$CB,$4A,$F8,$3A,$27,$E0,$F8,$F2,$7C,$2F
	.dc.b	$C4,$F5,$FC,$1F,$4B,$27,$E4,$E7,$FE,$F9,$CF,$C4,$BE,$63,$E0,$7C
	.dc.b	$CC,$AE,$EF,$C3,$F9,$EF,$9F,$F1,$DF,$43,$E9,$65,$7C,$08,$F7,$7D
	.dc.b	$2F,$BF,$D7,$93,$E0,$8F,$99,$DB,$F3,$7C,$C9,$3E,$C8,$FE,$95,$6F
	.dc.b	$89,$BD,$32,$BF,$9F,$4E,$5F,$E8,$1F,$44,$F8,$4E,$37,$27,$31,$FD
	.dc.b	$57,$E9,$5F,$1B,$FD,$5E,$F8,$56,$8F,$E9,$53,$A7,$FA,$67,$D2,$FF
	.dc.b	$3F,$E4,$FE,$08,$DF,$13,$FA,$44,$A3,$0F,$F4,$BB,$FE,$26,$F9,$BC
	.dc.b	$BF,$EC,$D3,$A8,$E1,$E7,$F8,$E6,$46,$C8,$7E,$17,$9F,$F4,$4F,$F5
	.dc.b	$0F,$A5,$FE,$94,$DF,$47,$FB,$94,$63,$3F,$E3,$59,$3E,$37,$35,$2F
	.dc.b	$FB,$45,$EF,$4F,$5D,$EE,$73,$5B,$7B,$1F,$42,$FB,$FC,$DC,$8E,$6F
	.dc.b	$CD,$E7,$C4,$79,$E2,$E5,$E4,$9F,$F0,$47,$D2,$BC,$2A,$4B,$FE,$F9
	.dc.b	$0E,$BC,$47,$F7,$08,$E0,$EA,$79,$CF,$F8,$4F,$CF,$FC,$E4,$FF,$81
	.dc.b	$5A,$FF,$78,$6E,$1F,$CF,$FE,$93,$B3,$FE,$03,$FA,$4F,$E8,$15,$FE
	.dc.b	$05,$AD,$EF,$F8,$06,$C6,$7C,$0F,$F1,$16,$1C,$A1,$7C,$94,$66,$31
	.dc.b	$5E,$46,$D7,$E3,$1E,$13,$B9,$FC,$7B,$6B,$C0,$79,$E3,$C9,$5C,$5D
	.dc.b	$E3,$FE,$3E,$9F,$1C,$63,$1E,$11,$C8,$C9,$70,$F0,$D7,$16,$CF,$09
	.dc.b	$FE,$C9,$7D,$DE,$17,$D5,$E6,$7C,$0B,$E5,$1E,$68,$E4,$FC,$13,$85
	.dc.b	$F6,$8F,$D6,$9F,$E3,$5C,$3E,$63,$86,$BD,$CC,$76,$3E,$97,$8C,$71
	.dc.b	$8C,$38,$94,$62,$38,$CF,$F8,$65,$7D,$2D,$23,$FE,$1D,$6D,$FD,$DC
	.dc.b	$FF,$35,$A7,$C5,$AF,$F2,$65,$7C,$79,$CC,$7C,$43,$27,$C6,$73,$5F
	.dc.b	$87,$8C,$BD,$75,$FF,$11,$B9,$8F,$29,$C6,$33,$9F,$C0,$BF,$04,$62
	.dc.b	$7E,$5C,$63,$0F,$34,$72,$BE,$D1,$89,$D4,$F9,$CD,$5F,$FA,$46,$FA
	.dc.b	$47,$FA,$29,$FE,$B9,$59,$3F,$CE,$91,$EA,$9C,$3F,$E6,$8C,$3E,$19
	.dc.b	$FE,$2F,$1E,$17,$28,$FF,$48,$5D,$7F,$D2,$F1,$7E,$93,$FA,$C6,$1F
	.dc.b	$E8,$16,$CF,$F9,$D1,$FF,$3E,$46,$77,$E0,$75,$C3,$FA,$46,$18,$9F
	.dc.b	$F0,$6F,$6F,$91,$FE,$7B,$7F,$E8,$37,$FC,$23,$F4,$AF,$A5,$F1,$76
	.dc.b	$73,$8C,$4F,$8E,$31,$3A,$8E,$2F,$A7,$E9,$3C,$5F,$4A,$F8,$2C,$57
	.dc.b	$F8,$7A,$30,$FB,$67,$73,$AD,$BF,$C1,$15,$A7,$E0,$4E,$AB,$81,$CE
	.dc.b	$BE,$57,$D2,$3C,$7A,$7C,$DF,$92,$76,$62,$FF,$D5,$9F,$4A,$FA,$3F
	.dc.b	$A4,$D8,$FF,$58,$61,$CD,$DC,$FD,$7F,$A5,$7D,$2D,$38,$C5,$9F,$F5
	.dc.b	$C7,$D2,$FF,$79,$6F,$27,$DF,$8F,$A1,$BF,$FA,$B9,$D3,$FE,$B0,$FA
	.dc.b	$57,$9B,$F8,$56,$DF,$FB,$0F,$C5,$FF,$1B,$BF,$EC,$58,$CD,$7F,$AE
	.dc.b	$3E,$97,$E4,$75,$7F,$9F,$DB,$5F,$DF,$1F,$F4,$03,$FE,$CD,$73,$3F
	.dc.b	$EE,$8F,$A5,$7D,$87,$E2,$DB,$5F,$E5,$65,$F7,$99,$9F,$C8,$7D,$37
	.dc.b	$92,$FF,$C8,$0D,$C7,$FC,$88,$FF,$8F,$70,$FF,$2F,$B9,$7F,$DE,$4B
	.dc.b	$E3,$B9,$9D,$3C,$EB,$3F,$E6,$6F,$A6,$E1,$87,$FB,$F9,$E8,$FF,$93
	.dc.b	$DC,$E0,$72,$FF,$95,$55,$7E,$9D,$7C,$BD,$8C,$3F,$AB,$FA,$6F,$AF
	.dc.b	$5F,$E5,$D5,$F7,$B9,$F0,$74,$FE,$C7,$E9,$FA,$E7,$FE,$7C,$6E,$7F
	.dc.b	$E6,$37,$BC,$67,$2F,$DC,$7D,$3F,$48,$FF,$41,$B6,$BF,$CD,$4F,$FB
	.dc.b	$0A,$33,$7F,$BF,$FA,$7F,$F6,$AC,$7F,$A2,$9B,$5F,$E7,$15,$F0,$1B
	.dc.b	$F8,$6F,$E0,$DC,$1D,$76,$FF,$3C,$3D,$6F,$FC,$F0,$E7,$04,$27,$EB
	.dc.b	$7B,$1F,$C3,$FE,$9F,$AD,$7F,$A6,$97,$C1,$67,$2D,$3B,$F8,$EE,$1F
	.dc.b	$7B,$6F,$83,$FC,$67,$E9,$BE,$3E,$5F,$E8,$B6,$CF,$8A,$11,$F2,$B3
	.dc.b	$FF,$47,$3A,$7C,$CD,$8D,$F0,$F5,$E5,$EC,$CB,$FF,$58,$D6,$CE,$95
	.dc.b	$E1,$9F,$E6,$DF,$54,$FA,$8F,$F7,$56,$3F,$E6,$68,$C6,$1F,$DD,$21
	.dc.b	$18,$7F,$BB,$B5,$FF,$00,$FD,$47,$C6,$F2,$B8,$36,$3F,$BE,$BF,$63
	.dc.b	$E2,$33,$BA,$BB,$7D,$37,$F9,$A3,$4F,$F7,$E3,$72,$FF,$3C,$B7,$1F
	.dc.b	$B3,$F4,$FF,$D4,$E7,$3F,$25,$FE,$7E,$BE,$2D,$8C,$D7,$CF,$7F,$D0
	.dc.b	$75,$C6,$3F,$07,$7F,$CF,$5D,$18,$F4,$39,$34,$C8,$87,$CF,$EA,$31
	.dc.b	$36,$7A,$47,$E3,$EF,$D7,$39,$AD,$32,$7A,$D7,$0C,$7E,$1D,$75,$FA
	.dc.b	$67,$D3,$F5,$F7,$26,$1E,$49,$CB,$67,$ED,$28,$CB,$FE,$30,$6B,$FE
	.dc.b	$3A,$6B,$FE,$A8,$FA,$7F,$E1,$AF,$C0,$6B,$EB,$3F,$56,$E3,$06,$7C
	.dc.b	$79,$BE,$24,$E6,$3A,$72,$E0,$CC,$EB,$D5,$DF,$D0,$CC,$7C,$11,$97
	.dc.b	$C5,$19,$7C,$43,$0F,$ED,$0F,$03,$64,$36,$E9,$E3,$F4,$EF,$A7,$F3
	.dc.b	$FC,$C4,$9F,$BF,$39,$6C,$E5,$E7,$3F,$81,$9D,$47,$93,$E9,$5F,$4D
	.dc.b	$F8,$55,$DF,$4C,$44,$9F,$C6,$CE,$9B,$39,$7A,$7D,$1F,$32,$3C,$3E
	.dc.b	$92,$BA,$39,$C7,$86,$33,$C3,$F5,$A7,$F9,$1D,$AB,$3C,$3F,$EE,$88
	.dc.b	$5F,$E7,$7C,$26,$9E,$BA,$98,$7E,$92,$72,$D9,$CB,$F0,$E7,$94,$7E
	.dc.b	$9F,$E9,$BF,$3D,$1F,$EF,$67,$F5,$73,$FA,$E6,$BF,$79,$DB,$FC,$E9
	.dc.b	$A9,$DE,$5E,$D8,$C7,$3F,$7E,$9F,$07,$D2,$49,$EC,$F8,$FA,$71,$1D
	.dc.b	$E2,$DF,$3C,$3F,$D1,$58,$F9,$E7,$2D,$AF,$45,$E1,$F3,$4D,$7C,$D1
	.dc.b	$69,$36,$30,$D9,$F2,$90,$F2,$1D,$8C,$5F,$96,$38,$46,$89,$19,$78
	.dc.b	$18,$BF,$43,$4F,$99,$5E,$27,$AC,$7D,$58,$C4,$6C,$F5,$9A,$3D,$E5
	.dc.b	$32,$6E,$FD,$EE,$5F,$96,$F0,$35,$7F,$4D,$D3,$EC,$BF,$54,$68,$F5
	.dc.b	$B8,$25,$FB,$1F,$B1,$7E,$87,$33,$EE,$3A,$9E,$38,$D4,$F1,$3E,$04
	.dc.b	$68,$F5,$46,$9F,$44,$F6,$74,$F6,$7E,$0B,$6F,$E0,$C7,$23,$75,$DD
	.dc.b	$F2,$1E,$D3,$45,$9E,$55,$E1,$8F,$D0,$5F,$C7,$75,$5C,$DF,$0B,$6F
	.dc.b	$C3,$7C,$A7,$C2,$7A,$A7,$DE,$F2,$30,$79,$23,$D5,$5F,$4E,$7A,$36
	.dc.b	$76,$79,$0F,$98,$76,$30,$7E,$EE,$FD,$57,$E4,$AF,$0D,$F6,$AE,$27
	.dc.b	$8C,$E2,$FC,$F7,$13,$C5,$F3,$4F,$91,$E3,$3C,$DF,$96,$D3,$E1,$3C
	.dc.b	$AF,$E5,$46,$67,$E4,$5F,$9A,$73,$1D,$6B,$9B,$6B,$99,$C6,$3E,$1D
	.dc.b	$FE,$79,$FF,$2C,$F8,$FF,$E0,$7D,$7F,$CE,$F1,$FE,$0C,$6B,$FD,$73
	.dc.b	$5F,$F4,$53,$FC,$B6,$78,$73,$C8,$C1,$C7,$9F,$A1,$1E,$9C,$F5,$FE
	.dc.b	$CF,$87,$F6,$1C,$BF,$93,$37,$3F,$F3,$2C,$FF,$40,$AF,$37,$D0,$F9
	.dc.b	$F8,$64,$F9,$E3,$2F,$A6,$32,$F6,$8C,$BE,$08,$CE,$1E,$09,$FF,$14
	.dc.b	$7D,$33,$CF,$FF,$15,$7D,$37,$F6,$1C,$3F,$BE,$B6,$FF,$9E,$56,$E7
	.dc.b	$FC,$66,$D7,$7C,$FC,$8E,$B7,$D3,$9E,$4F,$F8,$EB,$0F,$EC,$0D,$D7
	.dc.b	$2F,$82,$6E,$3F,$D2,$EB,$FB,$1C,$FF,$90,$BE,$9B,$FC,$A8,$A3,$FC
	.dc.b	$98,$DD,$BF,$C8,$93,$A3,$73,$A3,$6E,$7C,$2F,$83,$32,$BB,$47,$F9
	.dc.b	$37,$E9,$5E,$77,$4E,$A9,$5F,$E5,$49,$CB,$67,$FC,$83,$18,$8F,$BA
	.dc.b	$7F,$A8,$9C,$38,$6B,$F8,$2C,$3F,$58,$9A,$4E,$9F,$6F,$4C,$9F,$88
	.dc.b	$FC,$08,$FF,$32,$7D,$37,$F9,$06,$7F,$D6,$2F,$E7,$B5,$FE,$EC,$DD
	.dc.b	$3F,$22,$EE,$4D,$34,$F2,$E7,$EB,$E0,$C8,$FF,$3A,$CE,$9E,$33,$97
	.dc.b	$FC,$8F,$9F,$E2,$EF,$BF,$3F,$98,$F9,$DC,$92,$BF,$D4,$55,$F9,$37
	.dc.b	$E3,$47,$CC,$3F,$CD,$B8,$39,$BD,$E4,$9F,$1F,$3D,$7F,$CA,$F3,$B7
	.dc.b	$CF,$3F,$1A,$7F,$D1,$5F,$27,$FE,$48,$9F,$F4,$7B,$E7,$7F,$DB,$2D
	.dc.b	$4E,$DF,$F3,$1B,$75,$FF,$11,$3B,$27,$7F,$63,$BE,$32,$BB,$C7,$FA
	.dc.b	$82,$9F,$49,$FA,$69,$FF,$40,$46,$3A,$35,$3F,$A4,$9C,$BB,$9F,$F2
	.dc.b	$70,$E0,$FE,$6C,$DE,$1E,$D1,$FE,$02,$8D,$71,$78,$7E,$89,$C3,$7E
	.dc.b	$F6,$65,$FE,$B6,$B5,$6A,$31,$93,$EA,$FC,$A4,$8C,$F8,$BE,$0F,$B7
	.dc.b	$91,$CF,$44,$79,$3E,$6F,$C5,$F7,$B2,$BC,$19,$23,$CD,$F0,$7F,$0B
	.dc.b	$E0,$3E,$1F,$C0,$CB,$F1,$C7,$BB,$CF,$F7,$F8,$92,$7C,$11,$EF,$FA
	.dc.b	$1F,$81,$F4,$52,$7B,$C7,$CE,$F8,$A9,$1E,$E7,$E8,$FC,$3F,$CA,$F8
	.dc.b	$09,$3E,$B7,$E6,$FB,$FF,$2F,$CD,$93,$EA,$7E,$77,$1F,$FB,$67,$E2
	.dc.b	$19,$DF,$A3,$97,$FE,$E0,$9C,$BF,$EE,$2F,$A0,$7C,$67,$1B,$93,$98
	.dc.b	$FF,$72,$7D,$1B,$7F,$F7,$35,$F0,$AD,$1F,$EE,$C9,$D3,$FE,$EE,$FA
	.dc.b	$3F,$E7,$9D,$BF,$E7,$B6,$F9,$5F,$D5,$9B,$21,$6D,$FB,$9F,$EF,$59
	.dc.b	$DC,$9B,$3B,$8F,$64,$F0,$69,$F8,$5F,$3F,$FD,$FD,$F4,$6E,$3A,$49
	.dc.b	$F8,$9D,$95,$FF,$6A,$CE,$DF,$36,$7F,$E9,$58,$C4,$7F,$B6,$B1,$57
	.dc.b	$1A,$D5,$DC,$FF,$BA,$D6,$AF,$9C,$7F,$78,$C1,$F2,$9E,$62,$4F,$D7
	.dc.b	$FA,$3A,$F9,$22,$E3,$FE,$A7,$C1,$5F,$E0,$28,$FF,$55,$7D,$1B,$E7
	.dc.b	$31,$97,$F8,$1A,$31,$8F,$88,$FA,$4E,$1F,$F7,$AB,$FC,$59,$D0,$75
	.dc.b	$5D,$AB,$AC,$F2,$6B,$BF,$32,$F9,$2F,$9D,$A4,$9E,$BF,$47,$FC,$93
	.dc.b	$0F,$C2,$7D,$0A,$FF,$C3,$A7,$1A,$F1,$B5,$4F,$FB,$26,$30,$FF,$B2
	.dc.b	$EB,$A4,$AB,$A3,$F9,$18,$CC,$62,$3C,$17,$FE,$2C,$8C,$5F,$77,$55
	.dc.b	$C2,$3F,$C6,$2D,$AF,$CE,$46,$CF,$4D,$73,$77,$87,$A2,$7C,$91,$8C
	.dc.b	$3E,$0B,$C2,$29,$D2,$B9,$3E,$9A,$E5,$87,$91,$B5,$C6,$7A,$9C,$4E
	.dc.b	$63,$0F,$43,$67,$8B,$B3,$89,$F1,$0E,$31,$C0,$F9,$AF,$F8,$59,$AF
	.dc.b	$AC,$FF,$0C,$9D,$70,$AF,$87,$E8,$DF,$47,$8A,$46,$27,$8C,$62,$75
	.dc.b	$39,$9E,$7F,$47,$D1,$FE,$18,$D4,$D2,$3E,$CB,$FD,$36,$DF,$12,$38
	.dc.b	$99,$D5,$19,$8E,$66,$5C,$57,$CC,$AE,$73,$F2,$0E,$7A,$7E,$36,$32
	.dc.b	$FE,$3E,$33,$A7,$EE,$70,$EF,$1C,$4E,$F3,$C6,$32,$F1,$8C,$CF,$1C
	.dc.b	$3C,$71,$E5,$39,$61,$E5,$BE,$35,$B9,$FD,$4D,$5C,$3C,$31,$FE,$8C
	.dc.b	$BE,$13,$FC,$0A,$7D,$A6,$87,$D1,$FF,$82,$23,$91,$FC,$4D,$C5,$7F
	.dc.b	$3F,$FA,$3F,$E0,$53,$FC,$31,$B1,$E2,$34,$70,$0E,$AB,$CA,$D9,$FE
	.dc.b	$6B,$D9,$E4,$7F,$D4,$71,$9A,$FE,$0B,$39,$7C,$5E,$2F,$92,$DF,$46
	.dc.b	$FA,$3F,$06,$1F,$E9,$DC,$7F,$C3,$13,$97,$D1,$7C,$63,$18,$7A,$4F
	.dc.b	$3C,$66,$F9,$FD,$17,$7F,$A3,$7C,$3E,$31,$FF,$12,$46,$1F,$89,$3B
	.dc.b	$9D,$61,$FE,$3A,$AD,$3D,$F2,$56,$EF,$89,$D2,$FE,$57,$D1,$3E,$1A
	.dc.b	$9F,$37,$E3,$FF,$A8,$D7,$F3,$E9,$D3,$FD,$03,$E8,$DE,$6F,$D3,$6C
	.dc.b	$6E,$F1,$3F,$5F,$E8,$DF,$47,$4C,$D7,$F4,$A8,$CB,$FD,$33,$E8,$FF
	.dc.b	$C2,$6D,$E4,$FC,$04,$7D,$1D,$FE,$63,$A7,$E7,$7A,$8F,$37,$F0,$CD
	.dc.b	$BF,$EA,$FE,$33,$CE,$CD,$7E,$2F,$E8,$FE,$4F,$4F,$E7,$9B,$7F,$EE
	.dc.b	$A7,$FA,$83,$FD,$81,$CE,$3F,$D3,$BE,$8D,$EA,$FE,$25,$B5,$FE,$5D
	.dc.b	$5E,$B6,$67,$F1,$DF,$47,$E3,$AF,$F3,$13,$6F,$FC,$EA,$BE,$9B,$7A
	.dc.b	$70,$F5,$A5,$F0,$F5,$AE,$1F,$E7,$CF,$A4,$E5,$B7,$F9,$E1,$E5,$5F
	.dc.b	$E6,$F7,$2F,$F8,$69,$CC,$F1,$AF,$D1,$43,$E2,$AC,$61,$FD,$37,$D2
	.dc.b	$7D,$3C,$3F,$D0,$AB,$DA,$CF,$87,$A7,$F5,$9F,$49,$FE,$89,$CF,$FD
	.dc.b	$00,$DC,$FF,$D2,$6F,$78,$CE,$5F,$B3,$FA,$4E,$91,$FE,$87,$6D,$7F
	.dc.b	$A6,$9F,$F2,$34,$66,$FF,$75,$F4,$9F,$EE,$E8,$FF,$47,$36,$BF,$D4
	.dc.b	$2B,$E5,$77,$F2,$3F,$C0,$78,$3A,$ED,$FE,$A0,$7A,$DF,$FA,$A1,$CE
	.dc.b	$1F,$E4,$27,$33,$FE,$A6,$FA,$4E,$93,$FE,$9D,$5F,$3D,$FC,$35,$CD
	.dc.b	$FF,$AC,$9C,$CE,$FE,$3B,$5F,$49,$B6,$C7,$FA,$9F,$E9,$3E,$46,$3F
	.dc.b	$EB,$46,$C7,$F9,$51,$F9,$58,$7F,$AF,$9C,$BF,$E2,$46,$C9,$A7,$07
	.dc.b	$A7,$07,$F2,$38,$CB,$79,$30,$FF,$33,$FA,$2F,$6F,$93,$FD,$E9,$B8
	.dc.b	$7F,$7E,$7E,$77,$D1,$67,$65,$7F,$80,$3E,$B1,$F1,$1F,$DD,$DB,$86
	.dc.b	$63,$07,$F7,$B1,$F0,$7F,$E4,$C6,$C7,$F9,$51,$AF,$F3,$1F,$A3,$FE
	.dc.b	$A9,$BF,$FA,$36,$3B,$D7,$12,$3F,$E5,$C9,$DB,$FE,$93,$AF,$E6,$AF
	.dc.b	$5F,$3F,$C0,$78,$7E,$6A,$C6,$48,$7F,$41,$94,$6A,$31,$37,$87,$FB
	.dc.b	$6D,$6A,$7F,$CD,$BF,$44,$FA,$3F,$F0,$D3,$63,$FC,$4E,$D7,$FC,$56
	.dc.b	$D7,$DC,$1D,$57,$F7,$A6,$E7,$F0,$DB,$3E,$FF,$83,$FF,$3B,$FD,$1F
	.dc.b	$FB,$FE,$3F,$CF,$ED,$7F,$AE,$36,$FF,$C0,$AF,$F0,$78,$DF,$1F,$22
	.dc.b	$3C,$B9,$3A,$FA,$5D,$BE,$56,$4F,$68,$CB,$E0,$8C,$BE,$08,$CB,$E2
	.dc.b	$8C,$E1,$E2,$9F,$F4,$7F,$D1,$BE,$4F,$FC,$84,$DB,$FF,$27,$B5,$FF
	.dc.b	$29,$B5,$F9,$2E,$6B,$FC,$42,$DD,$7F,$14,$D9,$FF,$6C,$66,$FA,$CF
	.dc.b	$F4,$FF,$D2,$3E,$6E,$F6,$CB,$FF,$54,$CE,$5B,$39,$7F,$C1,$10,$9E
	.dc.b	$73,$D3,$E8,$A5,$72,$7B,$FC,$E6,$23,$8C,$5D,$3F,$C7,$38,$2B,$1A
	.dc.b	$7E,$8E,$15,$8B,$FF,$79,$46,$DF,$F5,$FF,$88,$F9,$DF,$F5,$DB,$63
	.dc.b	$31,$FE,$BE,$FA,$4F,$EF,$F1,$82,$1F,$E2,$39,$E0,$FF,$89,$63,$6D
	.dc.b	$E9,$E5,$D5,$23,$FD,$9F,$3F,$B2,$74,$62,$34,$E0,$E9,$F3,$DE,$32
	.dc.b	$3C,$BE,$1E,$BA,$3A,$C7,$A6,$33,$B7,$F9,$82,$D3,$8D,$7F,$85,$47
	.dc.b	$F9,$E5,$B1,$9B,$09,$E2,$F8,$4E,$E4,$3C,$47,$C3,$BE,$4F,$19,$B3
	.dc.b	$DE,$3A,$BD,$1E,$37,$E7,$9F,$54,$75,$8C,$36,$FD,$37,$D2,$BA,$B6
	.dc.b	$74,$76,$70,$F6,$9E,$2E,$8F,$31,$6F,$CC,$E0,$E0,$7A,$2B,$CA,$E0
	.dc.b	$D5,$F9,$9F,$84,$73,$8F,$2B,$4E,$D7,$E4,$9E,$07,$D7,$71,$7F,$52
	.dc.b	$7E,$81,$DE,$FE,$45,$70,$9F,$5C,$E6,$FB,$B4,$EE,$78,$4E,$57,$D9
	.dc.b	$F6,$1D,$AB,$A3,$83,$67,$17,$2F,$E1,$63,$93,$F0,$AB,$C7,$3A,$BF
	.dc.b	$3C,$7A,$E7,$E2,$9E,$53,$A9,$67,$85,$74,$AE,$33,$F6,$CF,$15,$73
	.dc.b	$AF,$91,$18,$3D,$0E,$1E,$91,$B3,$CA,$57,$6D,$7D,$87,$33,$D8,$78
	.dc.b	$4E,$67,$03,$11,$EF,$BF,$A2,$D8,$FA,$27,$03,$99,$D8,$C5,$7B,$9C
	.dc.b	$47,$17,$E2,$B4,$F8,$AF,$96,$38,$1F,$98,$2C,$F8,$AB,$D9,$18,$9F
	.dc.b	$BD,$18,$9D,$47,$AC,$F2,$CF,$4B,$FA,$53,$FA,$0B,$F1,$5F,$49,$CD
	.dc.b	$F4,$3F,$84,$E4,$F1,$FF,$97,$92,$3F,$86,$DF,$FB,$D3,$7D,$47,$F3
	.dc.b	$69,$D3,$FC,$BD,$FF,$10,$4E,$DB,$87,$F8,$8E,$BF,$D8,$55,$D2,$31
	.dc.b	$CD,$A8,$FE,$37,$3A,$7F,$DB,$8B,$9F,$51,$FD,$36,$6C,$EF,$0F,$34
	.dc.b	$62,$B3,$F3,$FA,$E0,$98,$78,$4E,$F1,$A7,$07,$F5,$98,$DB,$E2,$8D
	.dc.b	$3E,$48,$CD,$FC,$59,$F2,$FD,$23,$CF,$FF,$12,$DF,$F9,$0E,$30,$E0
	.dc.b	$38,$76,$D7,$FC,$5A,$E6,$99,$F3,$AE,$E7,$C4,$FA,$47,$B7,$79,$A6
	.dc.b	$3E,$C9,$FE,$E2,$DD,$B2,$B1,$EA,$FF,$4C,$9D,$B9,$C7,$FC,$B9,$18
	.dc.b	$9D,$CA,$3F,$BD,$3B,$3F,$AA,$FD,$27,$F9,$9A,$77,$3F,$E6,$09,$E1
	.dc.b	$5F,$72,$B4,$FF,$9A,$5B,$1E,$38,$F7,$1B,$F1,$64,$7B,$63,$2F,$DA
	.dc.b	$9E,$2B,$E9,$1E,$7F,$D1,$8C,$36,$33,$1F,$55,$AF,$DA,$6E,$3F,$01
	.dc.b	$B7,$FE,$7F,$6C,$FC,$0F,$99,$F6,$73,$C8,$F2,$3E,$FE,$7C,$47,$C7
	.dc.b	$19,$7F,$0D,$24,$7E,$83,$E9,$3F,$CC,$58,$FF,$A3,$5F,$F3,$15,$71
	.dc.b	$6E,$5F,$51,$DC,$7F,$81,$9A,$9D,$CF,$3F,$0F,$B7,$89,$2B,$F4,$75
	.dc.b	$A7,$94,$E5,$F4,$69,$FE,$49,$7E,$47,$81,$F2,$C8,$E0,$FD,$6C,$6D
	.dc.b	$E3,$1E,$DF,$91,$F7,$73,$C9,$F0,$BF,$07,$A3,$11,$EF,$8B,$A7,$BE
	.dc.b	$7F,$82,$FC,$A7,$F9,$7E,$BE,$63,$DE,$7F,$2E,$DD,$7F,$D6,$CE,$E3
	.dc.b	$3E,$17,$2C,$A5,$FF,$04,$AD,$5F,$0A,$CC,$F6,$AF,$0C,$7F,$1B,$B7
	.dc.b	$D1,$FE,$9E,$FF,$95,$BF,$A6,$8F,$D7,$9A,$E7,$E1,$1B,$C3,$F6,$9C
	.dc.b	$1F,$DB,$28,$FF,$8D,$5B,$C3,$DE,$35,$F4,$3F,$C9,$EB,$F8,$F5,$21
	.dc.b	$F4,$9E,$56,$CE,$9E,$0F,$EE,$67,$D3,$F4,$9E,$42,$47,$4C,$7F,$D2
	.dc.b	$93,$E4,$FA,$4F,$11,$27,$C3,$8B,$E7,$FC,$1F,$49,$27,$C0,$F8,$FE
	.dc.b	$27,$E0,$FC,$2F,$C2,$F8,$69,$5D,$E3,$DD,$EA,$7B,$FC,$49,$3D,$A3
	.dc.b	$D9,$F4,$7E,$DF,$A4,$93,$ED,$8F,$9B,$F2,$12,$3D,$4F,$D0,$F1,$FE
	.dc.b	$57,$C1,$49,$F4,$BF,$37,$C4,$F9,$1E,$64,$9F,$90,$FC,$BE,$1F,$EA
	.dc.b	$BF,$18,$CE,$FC,$44,$BF,$E9,$B3,$97,$FA,$7F,$D0,$3E,$3F,$8D,$C9
	.dc.b	$CC,$7F,$60,$FA,$46,$FF,$D8,$AF,$85,$68,$FE,$B3,$3A,$7F,$AD,$FD
	.dc.b	$27,$F4,$9F,$03,$FD,$0F,$08,$C3,$FE,$8A,$6A,$A6,$DD,$CF,$EE,$33
	.dc.b	$A3,$71,$A9,$F5,$CE,$DA,$7E,$0B,$CD,$FE,$F3,$F4,$9F,$D5,$A3,$FB
	.dc.b	$7B,$5F,$E4,$4D,$EC,$FE,$9B,$7F,$89,$9D,$F1,$F2,$23,$F2,$BA,$EB
	.dc.b	$CA,$2D,$72,$9F,$0F,$D2,$4A,$F0,$BF,$2F,$DE,$C4,$78,$A2,$D7,$8A
	.dc.b	$7F,$C2,$3F,$48,$FA,$0A,$4B,$FF,$04,$C3,$C9,$C4,$7F,$3D,$8E,$0E
	.dc.b	$A7,$9C,$FF,$87,$3C,$67,$D2,$52,$57,$F8,$53,$E9,$3A,$65,$FE,$C0
	.dc.b	$FA,$35,$FD,$C3,$8D,$7F,$86,$9A,$FF,$4D,$6C,$FF,$78,$AE,$72,$AE
	.dc.b	$6F,$FB,$32,$33,$18,$8E,$F7,$F8,$A8,$C5,$F6,$75,$5C,$23,$F1,$ED
	.dc.b	$AF,$F6,$BC,$6C,$F4,$57,$37,$78,$7F,$84,$E7,$C9,$18,$C3,$E0,$3C
	.dc.b	$22,$9D,$2B,$93,$E9,$AE,$58,$79,$1B,$5C,$67,$A9,$C4,$E6,$30,$F4
	.dc.b	$36,$78,$BB,$38,$9F,$10,$E3,$1C,$0F,$9A,$FE,$C0,$E5,$87,$F9,$06
	.dc.b	$33,$1E,$D3,$FC,$8E,$BC,$F7,$D2,$6D,$B3,$A6,$BD,$67,$84,$FF,$87
	.dc.b	$61,$F4,$94,$8F,$F8,$85,$B9,$7D,$DC,$3F,$39,$CF,$F0,$A8,$72,$3D
	.dc.b	$F1,$A7,$E0,$9A,$AF,$93,$19,$AF,$C4,$4E,$4E,$3A,$FF,$93,$A3,$31
	.dc.b	$D5,$F1,$46,$74,$FF,$34,$DF,$48,$C5,$7C,$B8,$C6,$1E,$78,$E5,$5F
	.dc.b	$E9,$2A,$DC,$7F,$20,$AD,$BE,$83,$67,$FA,$7A,$37,$87,$B4,$FF,$2E
	.dc.b	$96,$FC,$73,$EA,$7F,$D0,$F1,$98,$FF,$53,$FD,$27,$F7,$C9,$F1,$1A
	.dc.b	$71,$1F,$33,$E8,$7F,$7F,$3F,$E9,$06,$C7,$F9,$0D,$FA,$31,$9D,$F8
	.dc.b	$CA,$77,$7F,$E0,$AE,$CF,$14,$E6,$BB,$CF,$F4,$DC,$78,$DF,$FA,$57
	.dc.b	$0F,$F0,$2F,$D2,$3E,$93,$3E,$3F,$08,$C4,$F9,$63,$13,$A8,$E2,$F4
	.dc.b	$FA,$3E,$FF,$48,$F8,$6C,$57,$F8,$4A,$30,$FB,$E7,$73,$AD,$7F,$C2
	.dc.b	$15,$A7,$E1,$4E,$AB,$81,$CE,$BE,$1F,$D2,$3E,$12,$9F,$1B,$E4,$5D
	.dc.b	$98,$BF,$F5,$EF,$D2,$3C,$DF,$96,$D8,$FF,$60,$E1,$CD,$DC,$FD,$0F
	.dc.b	$A3,$7D,$25,$38,$C5,$9F,$F6,$4F,$D2,$7F,$7B,$6F,$3F,$D8,$8F,$A1
	.dc.b	$BF,$F9,$FD,$D3,$FE,$81,$F5,$1E,$6F,$DE,$6D,$FF,$B5,$3C,$57,$9D
	.dc.b	$9A,$FC,$1F,$D2,$7C,$7E,$BF,$F5,$9B,$6F,$FB,$FB,$FD,$19,$FF,$6E
	.dc.b	$B9,$C7,$F7,$BF,$77,$F8,$E7,$FD,$24,$DA,$FF,$2A,$FD,$2F,$C5,$9C
	.dc.b	$9F,$23,$E7,$66,$1F,$E4,$EF,$A5,$FC,$8F,$22,$7F,$2F,$F8,$2F,$E5
	.dc.b	$AB,$FC,$22,$DB,$FF,$68,$5B,$E1,$F7,$E1,$7E,$7B,$C4,$F8,$35,$FE
	.dc.b	$A5,$6E,$9F,$E6,$8F,$A5,$FC,$FC,$63,$2E,$BE,$F7,$80,$FD,$0C,$BE
	.dc.b	$AF,$13,$FE,$FE,$F8,$5F,$D3,$9F,$B9,$C1,$F4,$DF,$E2,$09,$D7,$91
	.dc.b	$FB,$0F,$BA,$7B,$B4,$FD,$6F,$D3,$F5,$AF,$E4,$F8,$7F,$B9,$2B,$71
	.dc.b	$83,$FC,$63,$F8,$2B,$C4,$FD,$DB,$63,$FC,$EF,$F5,$1F,$57,$85,$39
	.dc.b	$7F,$D8,$DE,$4B,$E5,$3F,$81,$B6,$FF,$CF,$FF,$50,$FA,$4C,$D7,$F0
	.dc.b	$8F,$C1,$5F,$5F,$4F,$E0,$BC,$1D,$E3,$38,$79,$F2,$FE,$83,$93,$E8
	.dc.b	$BB,$9F,$C3,$3E,$AB,$E8,$4E,$A3,$F8,$AF,$67,$DD,$83,$F7,$F8,$94
	.dc.b	$FB,$5D,$F6,$7F,$1C,$F9,$37,$CA,$34,$FE,$D1,$97,$F5,$29,$CC,$62
	.dc.b	$11,$87,$6D,$7F,$B8,$7D,$6B,$EB,$BA,$47,$48,$C5,$73,$6C,$66,$76
	.dc.b	$FF,$79,$FA,$CF,$99,$D1,$E1,$6C,$79,$97,$D0,$BE,$B3,$5C,$FF,$7E
	.dc.b	$BD,$7F,$22,$F8,$AF,$F3,$9F,$17,$68,$C4,$FD,$10,$E1,$47,$93,$EA
	.dc.b	$FF,$AA,$ED,$FE,$7C,$AD,$4F,$3B,$EB,$39,$6D,$FF,$7C,$BE,$2F,$F7
	.dc.b	$EB,$FE,$70,$F8,$5A,$F8,$2B,$B7,$B5,$CA,$3C,$BF,$4B,$6B,$78,$FA
	.dc.b	$EB,$06,$67,$87,$D5,$3E,$AF,$87,$C2,$4C,$7E,$04,$E5,$B3,$97,$C4
	.dc.b	$79,$E7,$9C,$48,$FC,$24,$F1,$7F,$12,$D7,$FC,$64,$D7,$FD,$47,$F5
	.dc.b	$5F,$EE,$C7,$E8,$35,$EE,$1C,$63,$F8,$47,$F8,$C4,$70,$F9,$5F,$89
	.dc.b	$1A,$D3,$AC,$63,$CE,$CF,$C0,$6B,$E9,$E6,$1E,$08,$CB,$E2,$8C,$BE
	.dc.b	$2C,$3F,$AF,$3F,$85,$FA,$BF,$DA,$3E,$8E,$BF,$09,$F5,$63,$31,$EA
	.dc.b	$6E,$7E,$99,$DC,$F3,$F0,$38,$18,$7C,$11,$87,$F2,$9F,$54,$F3,$BF
	.dc.b	$04,$FE,$CA,$33,$DD,$89,$91,$82,$1F,$32,$79,$3D,$3C,$7E,$64,$7B
	.dc.b	$FD,$0D,$78,$45,$CB,$F6,$B4,$AF,$CA,$76,$7F,$0A,$BF,$3F,$17,$98
	.dc.b	$CE,$72,$32,$43,$EF,$CF,$16,$C7,$17,$58,$7F,$04,$9D,$3F,$C1,$BE
	.dc.b	$AB,$F4,$37,$FE,$EE,$5C,$FF,$A7,$6D,$FE,$F5,$DB,$FE,$6E,$6A,$76
	.dc.b	$FE,$49,$B5,$DF,$A3,$A7,$D3,$CA,$E8,$F7,$E9,$C4,$71,$8B,$5E,$79
	.dc.b	$F8,$FE,$4F,$EE,$F5,$FC,$DD,$FE,$81,$B5,$EF,$9D,$CF,$F1,$59,$CE
	.dc.b	$7F,$E8,$68,$FA,$5C,$FA,$7E,$5D,$75,$89,$1F,$C7,$75,$FE,$D6,$D9
	.dc.b	$CB,$E1,$9E,$58,$78,$63,$C1,$1A,$8C,$47,$63,$24,$30,$47,$F9,$DE
	.dc.b	$1E,$88,$FF,$4E,$B5,$FE,$73,$5E,$0C,$3D,$93,$93,$58,$7A,$2B,$A4
	.dc.b	$EB,$0F,$23,$F0,$E7,$58,$79,$8F,$0C,$EA,$BF,$D5,$B3,$AA,$E1,$86
	.dc.b	$67,$E5,$9F,$14,$87,$C5,$3D,$D7,$D6,$78,$C6,$1F,$23,$C2,$34,$FC
	.dc.b	$0B,$ED,$7E,$18,$F1,$DF,$87,$0F,$7B,$F0,$E3,$81,$D4,$D9,$D7,$0E
	.dc.b	$EF,$94,$F8,$A6,$8C,$61,$F0,$9E,$06,$5C,$3A,$8F,$5B,$C8,$F8,$66
	.dc.b	$4E,$0E,$A3,$89,$1C,$12,$3D,$27,$13,$EF,$9E,$23,$AC,$71,$7A,$90
	.dc.b	$CC,$7B,$A3,$E5,$C7,$23,$A5,$7C,$87,$A3,$F8,$9C,$3E,$54,$6E,$7B
	.dc.b	$3D,$0E,$D8,$7D,$37,$04,$70,$66,$BA,$47,$A6,$31,$5F,$73,$0E,$F5
	.dc.b	$AA,$F9,$55,$A7,$ED,$1E,$3C,$3E,$33,$E0,$9F,$53,$F4,$5E,$58,$7C
	.dc.b	$0A,$E5,$3F,$2D,$F2,$DF,$E2,$70,$F7,$C6,$AB,$DE,$EA,$BC,$37,$DA
	.dc.b	$7E,$5B,$DB,$0F,$1D,$71,$7E,$81,$8C,$3B,$5F,$C1,$6B,$EE,$6D,$F9
	.dc.b	$DB,$3C,$67,$C3,$87,$9A,$73,$1F,$49,$AF,$C5,$3F,$05,$87,$95,$D1
	.dc.b	$C8,$F1,$14,$F9,$1C,$FF,$CB,$48,$FF,$29,$E7,$AF,$F2,$F8,$B8,$FF
	.dc.b	$16,$9F,$E6,$9F,$49,$FC,$D7,$86,$5F,$F3,$18,$CB,$FC,$CE,$32,$FF
	.dc.b	$39,$8F,$F1,$B7,$D5,$3E,$2F,$FC,$71,$F5,$5F,$DD,$EB,$FC,$3A,$DC
	.dc.b	$7F,$CB,$F5,$B9,$FF,$29,$35,$DF,$47,$23,$AD,$C8,$FF,$2C,$5F,$3D
	.dc.b	$FF,$A2,$36,$02,$FA,$61,$FE,$6E,$9F,$F4,$E4,$F2,$E7,$F3,$9E,$8D
	.dc.b	$AC,$64,$87,$A9,$FF,$2A,$7D,$4F,$F9,$8B,$1F,$09,$F0,$0C,$F8,$52
	.dc.b	$BC,$F1,$97,$FC,$E1,$3C,$5F,$3F,$D4,$3E,$AB,$AF,$54,$9F,$F3,$B4
	.dc.b	$E5,$B3,$97,$FC,$93,$5F,$DE,$2F,$FB,$35,$6E,$7F,$C9,$AD,$7F,$B5
	.dc.b	$07,$0F,$0A,$FF,$50,$39,$6D,$3A,$3E,$09,$27,$D7,$D1,$5F,$1C,$5D
	.dc.b	$BF,$D3,$B4,$AF,$F3,$E7,$D5,$7F,$96,$6F,$EB,$BF,$D9,$49,$E0,$E8
	.dc.b	$DF,$4C,$9F,$F4,$C4,$F1,$7D,$5C,$9D,$39,$E4,$7C,$88,$FF,$45,$6B
	.dc.b	$F8,$58,$52,$B7,$1F,$1A,$B6,$7C,$2A,$FC,$D4,$7C,$86,$BF,$3A,$3E
	.dc.b	$3D,$39,$BD,$A4,$9F,$3F,$3D,$7F,$CF,$F3,$B7,$D7,$3F,$12,$7F,$D4
	.dc.b	$1F,$31,$FE,$58,$9F,$F4,$2B,$E9,$27,$7E,$8D,$E1,$27,$FD,$79,$3C
	.dc.b	$77,$FF,$66,$FE,$44,$F9,$9E,$BC,$F2,$FF,$D8,$53,$96,$CE,$5F,$0C
	.dc.b	$FF,$AC,$E7,$51,$FE,$98,$6D,$76,$AD,$1E,$AF,$B1,$FF,$5D,$7D,$DE
	.dc.b	$B9,$FF,$C7,$5F,$71,$FA,$1C,$4D,$9D,$47,$EE,$A1,$F2,$DF,$EC,$CC
	.dc.b	$5C,$BF,$E5,$C6,$F8,$7F,$EC,$7A,$DD,$FF,$B7,$7E,$91,$F3,$5C,$7E
	.dc.b	$B7,$FC,$69,$81,$FF,$1A,$4E,$AB,$FD,$C9,$F7,$2F,$9A,$E9,$E9,$C9
	.dc.b	$CC,$7F,$B5,$7E,$E5,$F0,$FE,$5D,$BF,$65,$5F,$EE,$A6,$A9,$DB,$E5
	.dc.b	$E3,$F2,$54,$8C,$1F,$EF,$5F,$C0,$DF,$6B,$DF,$B2,$4E,$63,$FD,$CD
	.dc.b	$D7,$F9,$B8,$FE,$43,$C0,$E6,$BA,$7A,$E7,$31,$8D,$78,$F7,$FF,$BB
	.dc.b	$F8,$79,$57,$FB,$8B,$07,$ED,$7F,$C2,$1E,$90,$72,$F1,$AF,$F7,$C7
	.dc.b	$DD,$BE,$BB,$FD,$F5,$F1,$6F,$46,$9A,$8F,$F7,$F7,$E0,$9F,$6F,$67
	.dc.b	$CE,$7C,$B3,$3D,$9E,$5F,$B2,$FC,$16,$CE,$9A,$67,$6F,$69,$FD,$0B
	.dc.b	$E6,$5F,$61,$FE,$89,$E1,$F4,$CA,$5F,$37,$9E,$5F,$F4,$74,$62,$79
	.dc.b	$9B,$BD,$3F,$D3,$BE,$B3,$CE,$EA,$7B,$B6,$76,$A7,$6F,$FC,$00,$E5
	.dc.b	$7D,$86,$A7,$39,$79,$9B,$8F,$FA,$86,$B7,$7E,$4F,$B0,$FE,$BB,$3F
	.dc.b	$EA,$56,$C7,$F6,$26,$BE,$1F,$B0,$EC,$79,$23,$26,$AD,$1A,$9F,$83
	.dc.b	$18,$6C,$7E,$7F,$D5,$F6,$FD,$83,$E8,$69,$2B,$8F,$D8,$57,$F1,$19
	.dc.b	$7F,$AE,$EB,$84,$7C,$E6,$BA,$3E,$74,$F1,$D3,$EC,$CE,$9B,$7F,$26
	.dc.b	$7C,$7F,$53,$2F,$A3,$F3,$7E,$C3,$11,$D4,$60,$9C,$E1,$F4,$6B,$EC
	.dc.b	$FD,$83,$B6,$93,$0F,$C2,$C3,$D1,$C4,$75,$8E,$0E,$A7,$9C,$FE,$13
	.dc.b	$D7,$7C,$85,$25,$7E,$43,$EC,$3A,$65,$FE,$1C,$FA,$FF,$ED,$53,$FE
	.dc.b	$1E,$BE,$2F,$E5,$9A,$FB,$5B,$3F,$A0,$AE,$72,$AE,$6A,$33,$18,$AE
	.dc.b	$6D,$BF,$F7,$93,$C2,$77,$3F,$EF,$56,$D7,$96,$B7,$3F,$EF,$93,$99
	.dc.b	$BC,$3F,$85,$4F,$8E,$31,$87,$BD,$E1,$5C,$A7,$A1,$93,$FC,$7F,$7E
	.dc.b	$07,$D5,$7D,$30,$F4,$46,$CF,$F0,$95,$73,$77,$86,$6F,$96,$1E,$69
	.dc.b	$EE,$79,$4E,$F8,$75,$8F,$3C,$7F,$8B,$0E,$65,$9F,$9E,$FF,$8C,$1A
	.dc.b	$FC,$7F,$2D,$F6,$1C,$63,$E9,$46,$1E,$4D,$79,$35,$F0,$7D,$17,$F8
	.dc.b	$E1,$34,$8F,$F8,$F1,$F8,$58,$FE,$6F,$8F,$E2,$4F,$F9,$A2,$73,$AF
	.dc.b	$F3,$ED,$7F,$0B,$39,$3E,$11,$FD,$06,$35,$9F,$8A,$33,$87,$57,$67
	.dc.b	$1A,$FC,$7C,$65,$EB,$1E,$83,$8D,$74,$7F,$CA,$F7,$E4,$C3,$CF,$1B
	.dc.b	$AE,$75,$B8,$FF,$40,$99,$C3,$FA,$04,$66,$3C,$51,$8A,$E9,$1F,$C0
	.dc.b	$F0,$F4,$1F,$D1,$9D,$9F,$DB,$2B,$FB,$0C,$7F,$4A,$37,$3F,$D3,$7D
	.dc.b	$1C,$BD,$6B,$A4,$61,$FF,$4A,$FD,$87,$F0,$09,$FE,$9A,$D8,$FE,$28
	.dc.b	$FF,$A6,$A3,$3B,$F1,$94,$EE,$FC,$7E,$1F,$8A,$7A,$D7,$79,$FE,$D7
	.dc.b	$8F,$1B,$FF,$5B,$61,$FD,$9A,$5F,$61,$FE,$2E,$8F,$1F,$C3,$71,$8C
	.dc.b	$63,$E4,$8C,$7A,$DC,$DB,$5E,$AF,$37,$C1,$F3,$4F,$23,$C1,$18,$6C
	.dc.b	$F8,$67,$4F,$13,$11,$AC,$3F,$C2,$53,$A7,$E2,$D7,$3A,$FE,$EB,$F6
	.dc.b	$0F,$89,$FB,$15,$FE,$C2,$8C,$4F,$EB,$5B,$1F,$B7,$6B,$FE,$A4,$FA
	.dc.b	$87,$D7,$72,$84,$62,$BF,$6C,$D8,$FE,$00,$D7,$FD,$5B,$F6,$0F,$B0
	.dc.b	$E4,$DA,$FD,$F3,$67,$F8,$43,$5F,$F5,$AF,$D8,$7F,$88,$A7,$F8,$23
	.dc.b	$63,$F8,$73,$5D,$F2,$F9,$FE,$C3,$FC,$39,$5F,$C2,$DB,$3F,$C5,$5A
	.dc.b	$FF,$B0,$9A,$FE,$A3,$EC,$3E,$73,$FE,$C6,$6B,$9E,$0F,$89,$18,$8F
	.dc.b	$F7,$7F,$D8,$7E,$5D,$FE,$44,$D7,$FD,$9E,$D7,$F6,$AD,$7F,$DE,$FA
	.dc.b	$3E,$BF,$E8,$37,$1F,$9F,$18,$3E,$A7,$D7,$FE,$9E,$BF,$DB,$6D,$7F
	.dc.b	$80,$B5,$07,$0F,$F2,$6F,$43,$EF,$77,$FC,$28,$FE,$45,$18,$3B,$FD
	.dc.b	$7C,$C3,$F9,$B5,$A3,$15,$FD,$07,$DC,$FE,$40,$FF,$BB,$9A,$FF,$14
	.dc.b	$6B,$E4,$6A,$FA,$C7,$CB,$7E,$45,$B8,$7F,$45,$6C,$FF,$4F,$6B,$FE
	.dc.b	$FC,$FA,$B7,$D1,$FE,$65,$B5,$FD,$35,$B3,$FD,$5D,$AF,$F2,$81,$D6
	.dc.b	$F9,$6F,$D1,$36,$BF,$AB,$36,$7F,$B0,$35,$E5,$F5,$2F,$AA,$FD,$53
	.dc.b	$6B,$FA,$F3,$67,$FB,$43,$5F,$15,$3E,$77,$C2,$4A,$F0,$BF,$B5,$FA
	.dc.b	$AF,$B6,$4F,$1F,$07,$EE,$7E,$A3,$FC,$79,$5F,$DB,$5B,$87,$F7,$76
	.dc.b	$BF,$05,$AF,$F4,$AF,$A7,$7D,$3F,$CC,$8F,$EF,$D1,$8D,$FF,$BC,$3F
	.dc.b	$E9,$DF,$80,$B5,$BA,$F9,$9F,$6E,$FA,$7F,$EF,$D0,$8C,$3F,$E0,$16
	.dc.b	$BF,$E0,$B6,$BF,$56,$3F,$DD,$04,$F6,$FF,$D1,$BF,$51,$FD,$4F,$4F
	.dc.b	$EC,$33,$9A,$FF,$4C,$BF,$EE,$F9,$D4,$7E,$92,$3F,$9B,$3A,$F0,$3F
	.dc.b	$D2,$F1,$88,$FF,$33,$3F,$ED,$A3,$1E,$6F,$FA,$A9,$B7,$E0,$AE,$3F
	.dc.b	$41,$FE,$A0,$FA,$77,$D4,$74,$F5,$26,$5F,$EA,$C9,$CB,$67,$E2,$C6
	.dc.b	$23,$FC,$58,$D7,$F2,$CD,$7F,$BF,$BB,$CB,$FD,$7A,$DC,$73,$5E,$CF
	.dc.b	$6F,$FD,$67,$F5,$0F,$8A,$A4,$C7,$FD,$7F,$0F,$A0,$C4,$7F,$A1,$E3
	.dc.b	$83,$A9,$E7,$3D,$3D,$1E,$BF,$21,$2B,$9C,$65,$EB,$19,$7A,$47,$FB
	.dc.b	$1B,$EA,$3F,$41,$1E,$2F,$9F,$F3,$99,$ED,$91,$FE,$47,$9D,$BF,$E5
	.dc.b	$16,$BF,$BB,$6B,$FE,$24,$77,$BF,$FB,$89,$BB,$66,$BE,$2F,$DB,$FF
	.dc.b	$B6,$3E,$A5,$E3,$DF,$A1,$99,$7F,$B9,$27,$2D,$9C,$BF,$EA,$D8,$4F
	.dc.b	$39,$E9,$E2,$CA,$E4,$F7,$FA,$1C,$47,$18,$BA,$7F,$BE,$30,$57,$FB
	.dc.b	$B3,$E5,$5B,$79,$F5,$98,$FF,$BE,$27,$2D,$9F,$D8,$C6,$23,$FC,$F0
	.dc.b	$D7,$F9,$13,$5F,$F2,$5B,$BC,$15,$8A,$F1,$1F,$EF,$EF,$A9,$FF,$00
	.dc.b	$5F,$F5,$0A,$DB,$FD,$4A,$3F,$4E,$DE,$9E,$07,$1E,$A9,$1C,$70,$FF
	.dc.b	$2D,$D2,$B7,$39,$3A,$57,$87,$E9,$A4,$64,$52,$3C,$DD,$7A,$9F,$F7
	.dc.b	$14,$E5,$E0,$D7,$31,$F3,$FC,$CF,$F3,$6E,$5E,$C9,$FA,$7C,$F8,$CF
	.dc.b	$FC,$E3,$B7,$EA,$67,$E9,$5F,$17,$C9,$19,$7F,$DE,$4D,$BE,$9B,$7F
	.dc.b	$62,$9E,$27,$0C,$3B,$3F,$DF,$2B,$86,$1F,$5B,$0E,$D3,$D2,$B6,$F9
	.dc.b	$8E,$67,$FC,$01,$BE,$CE,$CE,$C6,$B0,$F1,$D7,$2C,$3B,$CF,$86,$FB
	.dc.b	$D6,$A3,$89,$CC,$ED,$5A,$C3,$E2,$D7,$33,$13,$FE,$16,$C3,$C7,$7F
	.dc.b	$37,$0E,$F3,$E9,$39,$9E,$5B,$E6,$F8,$27,$C4,$E8,$F1,$3C,$EF,$BC
	.dc.b	$2F,$BD,$72,$8D,$BF,$2D,$EC,$7B,$CE,$CE,$5E,$D5,$D5,$C9,$EB,$C3
	.dc.b	$BE,$1E,$A3,$D7,$7E,$09,$D1,$C0,$F7,$57,$A8,$97,$E5,$7B,$47,$82
	.dc.b	$3C,$51,$82,$38,$24,$7A,$4F,$69,$E0,$39,$19,$8E,$2E,$5F,$0B,$98
	.dc.b	$F1,$1D,$1F,$95,$7F,$6D,$F5,$57,$B5,$C9,$ED,$BF,$83,$3F,$13,$0E
	.dc.b	$13,$E2,$36,$7F,$14,$8F,$E2,$E4,$7F,$C6,$C7,$A4,$FE,$2E,$E5,$E0
	.dc.b	$68,$E7,$87,$A9,$B8,$7E,$DA,$7F,$07,$87,$9C,$F4,$61,$63,$27,$E8
	.dc.b	$4B,$5F,$84,$3F,$99,$B8,$7D,$51,$C5,$D4,$F4,$9D,$5F,$8E,$B5,$5F
	.dc.b	$06,$B4,$F0,$AF,$C4,$9C,$2F,$E8,$57,$0B,$F6,$5F,$0B,$E5,$7C,$1E
	.dc.b	$47,$9E,$B8,$5F,$D9,$AE,$17,$CA,$F8,$5F,$23,$83,$E1,$7F,$09,$5F
	.dc.b	$25,$EA,$7C,$08,$D3,$F3,$8F,$E2,$CE,$2B,$BC,$F8,$63,$BD,$78,$6B
	.dc.b	$B9,$D1,$F8,$D1,$C4,$FA,$51,$FC,$AB,$9B,$F8,$A1,$1F,$E5,$BC,$F5
	.dc.b	$FE,$5B,$16,$BF,$86,$4F,$F1,$BF,$9A,$FE,$39,$C3,$2F,$F9,$AC,$65
	.dc.b	$FE,$6F,$19,$7F,$9B,$C7,$F2,$CF,$A9,$7B,$FF,$CB,$BE,$A7,$FB,$FD
	.dc.b	$7F,$51,$6A,$C7,$FA,$45,$70,$8F,$EA,$CD,$7F,$3A,$D7,$FD,$28,$D7
	.dc.b	$FD,$76,$1C,$3F,$EB,$E7,$94,$76,$AF,$31,$E5,$F8,$FF,$37,$3C,$BF
	.dc.b	$33,$E9,$E0,$C4,$79,$F7,$EF,$97,$FA,$8E,$36,$7B,$67,$FA,$6F,$D4
	.dc.b	$3C,$5F,$85,$72,$BF,$B9,$4E,$5B,$39,$7B,$1F,$ED,$6B,$FF,$3A,$A9
	.dc.b	$E0,$E8,$7D,$43,$EA,$7C,$3D,$52,$7F,$C0,$13,$96,$CF,$FA,$42,$31
	.dc.b	$1F,$C2,$9A,$FF,$B0,$5A,$FC,$E7,$7B,$7F,$7F,$6E,$59,$AF,$83,$F2
	.dc.b	$7F,$0B,$AA,$61,$F0,$1F,$8B,$D3,$88,$F5,$67,$FE,$0D,$CB,$FC,$14
	.dc.b	$DC,$7E,$EA,$AF,$EF,$3F,$53,$FE,$70,$BF,$84,$FF,$64,$27,$87,$A3
	.dc.b	$7D,$72,$7F,$C4,$53,$C7,$1F,$57,$81,$D3,$A6,$57,$C8,$8F,$F0,$3E
	.dc.b	$FF,$98,$8F,$89,$1F,$24,$8F,$F2,$D8,$FC,$ED,$74,$8E,$47,$83,$E3
	.dc.b	$7D,$3D,$32,$3C,$6F,$B3,$AB,$11,$EC,$8B,$5F,$E2,$DC,$BF,$CA,$1B
	.dc.b	$FE,$83,$E4,$3F,$CD,$F7,$F8,$37,$C3,$19,$CF,$FC,$3F,$1A,$BF,$F1
	.dc.b	$B3,$67,$55,$FE,$45,$8E,$28,$3B,$71,$F5,$2C,$FA,$F4,$49,$FF,$27
	.dc.b	$4E,$5B,$3F,$ED,$E8,$C4,$7F,$58,$6B,$FB,$A0,$E1,$FD,$EB,$C3,$0F
	.dc.b	$F2,$73,$70,$CD,$7B,$BE,$A7,$FC,$7B,$F5,$3C,$70,$F5,$35,$34,$C7
	.dc.b	$D4,$BE,$37,$FD,$E6,$D9,$FE,$D0,$D7,$FB,$7B,$5F,$47,$CD,$BE,$4B
	.dc.b	$FD,$F6,$D8,$FE,$E0,$D7,$FB,$BB,$5F,$81,$F5,$0F,$90,$FE,$ED,$08
	.dc.b	$C3,$FD,$E5,$AF,$F8,$01,$AF,$C9,$FA,$87,$D4,$72,$6C,$7F,$80,$5A
	.dc.b	$FF,$82,$1A,$FA,$B5,$F4,$91,$F4,$BE,$4F,$A8,$7D,$45,$24,$FF,$83
	.dc.b	$63,$2F,$F8,$3F,$CF,$75,$FB,$1B,$19,$8C,$1F,$20,$8F,$C8,$6B,$FE
	.dc.b	$1E,$FA,$87,$D4,$7C,$66,$C7,$F8,$75,$AE,$E3,$E5,$FD,$43,$D1,$A4
	.dc.b	$9F,$F1,$24,$65,$FA,$FF,$51,$F6,$88,$E0,$FF,$15,$7C,$63,$E9,$BE
	.dc.b	$BB,$63,$31,$83,$ED,$FD,$23,$E5,$69,$27,$FC,$69,$19,$7F,$0F,$F4
	.dc.b	$DF,$8A,$23,$83,$FC,$75,$F4,$DF,$D7,$3B,$7F,$B0,$B6,$3F,$70,$D5
	.dc.b	$4F,$33,$FD,$42,$D8,$F0,$C7,$98,$8F,$F9,$33,$C0,$79,$3F,$D4,$1F
	.dc.b	$F5,$F5,$71,$9F,$EA,$4D,$8F,$F6,$E9,$C3,$D0,$FF,$57,$FD,$2B,$D3
	.dc.b	$A4,$BF,$F6,$A5,$BE,$9B,$11,$FD,$76,$B8,$3F,$ED,$96,$BE,$08,$DE
	.dc.b	$1F,$ED,$19,$CB,$6F,$DF,$3E,$2F,$A6,$97,$CD,$F8,$7F,$03,$88,$F0
	.dc.b	$45,$BF,$7C,$FF,$B2,$3E,$99,$F4,$74,$97,$FE,$ED,$8F,$E4,$ED,$8C
	.dc.b	$BF,$DD,$E3,$8B,$BF,$2F,$FD,$A7,$E9,$3E,$3A,$92,$7F,$DF,$1F,$4D
	.dc.b	$FC,$CB,$2F,$E6,$9F,$4B,$FD,$8E,$7F,$9B,$5F,$17,$FD,$FC,$D7,$FC
	.dc.b	$0C,$D9,$FF,$7B,$57,$39,$57,$35,$19,$8C,$57,$36,$DF,$F8,$C1,$E1
	.dc.b	$3B,$9F,$F1,$9B,$6B,$CB,$5B,$9F,$F1,$B9,$CC,$DE,$1F,$E4,$39,$F1
	.dc.b	$C6,$30,$F7,$BC,$2B,$94,$F4,$32,$7F,$57,$BF,$03,$EA,$BE,$98,$7A
	.dc.b	$23,$67,$F8,$2E,$B9,$BB,$C3,$37,$CB,$0F,$34,$F7,$3C,$A7,$7C,$3A
	.dc.b	$C7,$9E,$3F,$D6,$73,$F3,$5F,$EB,$8D,$74,$63,$C3,$7D,$37,$FA,$D6
	.dc.b	$7F,$B6,$36,$33,$3D,$1A,$FF,$AE,$FE,$6B,$FA,$3A,$69,$1F,$E9,$6F
	.dc.b	$C0,$C7,$F3,$3D,$1F,$0A,$7F,$BF,$4E,$76,$4E,$67,$A1,$F0,$0F,$F3
	.dc.b	$C4,$69,$C1,$C8,$D5,$76,$72,$72,$D7,$F1,$D1,$97,$F2,$11,$C5,$CC
	.dc.b	$F5,$73,$1C,$73,$FF,$3F,$D7,$17,$FD,$C3,$87,$A7,$0F,$43,$6B,$C2
	.dc.b	$DB,$EB,$3F,$E1,$DB,$F5,$57,$B2,$30,$74,$7D,$78,$7F,$1C,$8F,$0B
	.dc.b	$63,$97,$97,$5E,$47,$F9,$D6,$BF,$C0,$95,$D2,$3F,$C5,$9F,$4D,$FB
	.dc.b	$F8,$C4,$23,$0F,$F9,$F5,$FF,$19,$46,$9C,$CF,$A1,$B1,$EB,$EC,$FA
	.dc.b	$13,$9E,$0F,$33,$FD,$47,$1F,$35,$FF,$AD,$AF,$C9,$2F,$A6,$FF,$11
	.dc.b	$47,$7F,$80,$E3,$18,$CB,$C9,$18,$F4,$39,$B6,$BC,$DE,$77,$83,$E5
	.dc.b	$BF,$B8,$46,$0E,$EE,$2F,$DD,$3B,$F1,$BC,$AD,$AE,$13,$FE,$A4,$C9
	.dc.b	$7C,$5F,$85,$5B,$9F,$F5,$27,$D3,$3E,$13,$EA,$7C,$FB,$E9,$BE,$B4
	.dc.b	$E7,$2F,$A3,$39,$6C,$FD,$BF,$A5,$7D,$37,$28,$C3,$27,$E4,$FE,$2E
	.dc.b	$F9,$AC,$67,$E7,$8C,$4F,$DE,$FA,$7F,$F0,$13,$F8,$0F,$11,$F5,$1F
	.dc.b	$89,$9C,$DF,$FA,$F6,$73,$5F,$86,$FA,$87,$9F,$F8,$FF,$A8,$FF,$52
	.dc.b	$4E,$63,$1E,$06,$79,$BF,$03,$A7,$E1,$2B,$FD,$CF,$7F,$25,$FD,$14
	.dc.b	$67,$2F,$F4,$07,$D4,$7E,$41,$FA,$1F,$67,$FD,$E6,$BF,$2F,$39,$E7
	.dc.b	$FE,$FF,$39,$AF,$F6,$A7,$23,$EA,$7F,$23,$F4,$7F,$A2,$9C,$C6,$3C
	.dc.b	$2F,$D4,$7D,$4F,$E6,$A5,$8A,$FF,$CB,$79,$FE,$9A,$B5,$39,$F7,$FF
	.dc.b	$61,$E3,$7C,$4A,$FC,$DF,$D9,$F7,$AF,$DE,$CE,$71,$FC,$54,$E6,$BF
	.dc.b	$D3,$D0,$C1,$F5,$5F,$A8,$FB,$3F,$75,$6A,$72,$FE,$A2,$33,$5F,$EA
	.dc.b	$7F,$7F,$E5,$64,$9F,$9D,$39,$F1,$FA,$72,$FF,$07,$FA,$B7,$CC,$7E
	.dc.b	$DB,$EC,$B9,$DA,$73,$9F,$ED,$A7,$3F,$59,$F8,$CA,$FD,$24,$3E,$5F
	.dc.b	$F8,$94,$E7,$4F,$F1,$9C,$E6,$FF,$87,$F9,$0F,$99,$FE,$0B,$F6,$5F
	.dc.b	$16,$FF,$91,$CE,$6B,$F8,$2C,$E5,$F8,$FF,$58,$FA,$DF,$E1,$BB,$75
	.dc.b	$D7,$EC,$D6,$A7,$3A,$FF,$25,$C5,$F4,$5F,$C5,$3E,$C9,$F2,$59,$8C
	.dc.b	$3F,$CA,$FA,$FE,$FC,$BD,$DC,$7A,$BC,$5F,$E6,$1F,$5A,$FA,$CF,$D9
	.dc.b	$6D,$FE,$CB,$11,$9F,$37,$11,$8F,$A2,$7D,$6E,$2B,$53,$C7,$EB,$5E
	.dc.b	$BF,$FA,$D1,$B1,$C9,$AE,$E3,$FD,$73,$DD,$F3,$63,$BB,$5F,$23,$5F
	.dc.b	$F7,$03,$5F,$C9,$FD,$67,$F9,$BF,$96,$FC,$3F,$F6,$06,$CF,$F4,$8F
	.dc.b	$AD,$FE,$B3,$D7,$FD,$6A,$F8,$4F,$FB,$E5,$A9,$B3,$C2,$BB,$F8,$1D
	.dc.b	$6B,$83,$5E,$0D,$AF,$C7,$47,$F5,$87,$FD,$50,$FE,$00,$E2,$66,$38
	.dc.b	$74,$73,$FA,$D7,$D6,$D2,$57,$3F,$87,$AF,$31,$3C,$1F,$EC,$ED,$75
	.dc.b	$82,$78,$E7,$FD,$F2,$B4,$DB,$F5,$4F,$B7,$EB,$7F,$D3,$B8,$7F,$72
	.dc.b	$6B,$E1,$6D,$7F,$81,$97,$D2,$71,$F8,$D9,$5C,$5E,$91,$D7,$DF,$ED
	.dc.b	$F2,$D2,$BB,$46,$5F,$04,$65,$F0,$47,$D1,$FA,$DF,$D3,$C7,$F8,$19
	.dc.b	$AF,$C5,$6E,$3E,$5F,$AD,$E1,$E0,$CA,$E0,$FB,$E3,$11,$C6,$2C,$F1
	.dc.b	$8F,$F0,$9F,$D6,$BC,$DF,$C1,$2F,$A3,$A6,$22,$4F,$08,$CB,$CA,$38
	.dc.b	$F9,$52,$3A,$BD,$3E,$8F,$11,$F7,$EF,$F4,$D8,$7F,$8D,$1B,$CF,$FC
	.dc.b	$D9,$57,$E5,$7B,$DF,$4D,$49,$87,$F8,$B3,$EB,$7F,$BC,$BF,$E2,$E9
	.dc.b	$DB,$67,$6F,$E3,$E7,$6F,$E4,$67,$6D,$9D,$E3,$FE,$48,$AD,$B7,$AB
	.dc.b	$F5,$11,$FA,$6A,$E9,$D5,$89,$91,$E5,$8D,$98,$7C,$5F,$55,$C8,$8F
	.dc.b	$AF,$DF,$AE,$8F,$74,$78,$23,$E2,$79,$B2,$7E,$03,$FB,$BD,$3F,$A7
	.dc.b	$C6,$08,$7B,$E3,$47,$17,$06,$B2,$FF,$33,$CF,$28,$C9,$23,$FC,$D7
	.dc.b	$A6,$5D,$D7,$37,$2D,$DB,$FB,$8D,$ED,$FE,$E7,$56,$C1,$C3,$6B,$CE
	.dc.b	$F9,$67,$8C,$75,$8E,$35,$C2,$A4,$E6,$7F,$16,$D8,$CC,$7F,$9D,$26
	.dc.b	$BE,$DA,$EA,$E6,$B8,$1C,$E7,$77,$EE,$9F,$21,$E2,$9D,$B8,$3C,$25
	.dc.b	$BE,$37,$F1,$27,$73,$E1,$39,$1B,$7C,$06,$AB,$C1,$1F,$CA,$6F,$9B
	.dc.b	$F2,$AE,$BE,$03,$A1,$F2,$CF,$24,$6A,$36,$DA,$EF,$3E,$13,$6F,$86
	.dc.b	$FE,$2C,$78,$63,$0F,$85,$FA,$EF,$82,$BA,$C6,$A3,$C4,$71,$3B,$1E
	.dc.b	$6A,$F1,$3E,$03,$37,$F0,$0F,$C7,$B8,$BF,$1C,$62,$3C,$15,$EA,$69
	.dc.b	$CC,$A7,$77,$C9,$7F,$90,$BF,$0C,$E4,$EC,$7B,$4F,$C2,$18,$3D,$75
	.dc.b	$E1,$3C,$D7,$E4,$24,$7D,$C3,$C3,$1A,$70,$D7,$9C,$70,$7E,$D4,$66
	.dc.b	$BE,$5C,$78,$5F,$45,$FA,$8F,$A3,$3B,$7E,$13,$67,$E1,$35,$F7,$1E
	.dc.b	$4A,$F0,$5F,$C2,$9F,$23,$63,$35,$DE,$BE,$19,$F9,$3E,$5F,$E2,$44
	.dc.b	$7F,$5D,$CD,$5D,$1F,$B1,$8F,$E3,$11,$FC,$67,$EA,$3F,$8D,$70,$4A
	.dc.b	$FD,$B4,$65,$FD,$C4,$65,$FD,$14,$7F,$86,$FE,$B7,$FC,$6F,$9F,$FC
	.dc.b	$00,$6E,$5F,$E8,$26,$D7,$F1,$0F,$AD,$7C,$2F,$1F,$16,$4F,$FB,$E2
	.dc.b	$B6,$AB,$6E,$67,$6F,$F8,$6E,$77,$3A,$8F,$F0,$F3,$FD,$79,$F0,$73
	.dc.b	$7F,$8C,$63,$11,$D7,$7C,$4C,$8F,$E4,$B9,$32,$FF,$4D,$DF,$03,$FB
	.dc.b	$3C,$EA,$3F,$D3,$FF,$59,$FE,$44,$9F,$F2,$14,$FB,$E7,$FC,$81,$3F
	.dc.b	$E4,$83,$4F,$8C,$CF,$8B,$FE,$45,$FA,$C7,$D6,$D2,$5F,$73,$F9,$CD
	.dc.b	$BE,$77,$31,$FE,$3D,$AE,$2E,$C3,$87,$FC,$85,$3C,$5D,$46,$72,$F8
	.dc.b	$11,$FD,$1A,$31,$8F,$FA,$F1,$B1,$E2,$BF,$6F,$3C,$8F,$06,$08,$F8
	.dc.b	$3C,$F9,$9F,$0C,$5C,$7B,$C2,$95,$FE,$67,$FA,$DF,$F2,$3E,$5F,$EC
	.dc.b	$C7,$FC,$8F,$5E,$D6,$BF,$E5,$D7,$6F,$F5,$E6,$A7,$77,$E0,$6C,$F5
	.dc.b	$F0,$FE,$97,$7C,$BF,$8A,$F8,$FE,$5F,$11,$F3,$E2,$E1,$CE,$6D,$FC
	.dc.b	$79,$B5,$E3,$9F,$67,$34,$8F,$53,$EF,$E7,$C4,$7A,$E2,$E1,$E7,$9F
	.dc.b	$F4,$57,$CD,$7F,$92,$EB,$EF,$3F,$E4,$B6,$E1,$AF,$06,$46,$08,$7F
	.dc.b	$A0,$E7,$93,$FE,$85,$FA,$D6,$5F,$E9,$76,$C7,$E2,$DA,$F1,$F5,$F5
	.dc.b	$32,$34,$F9,$A3,$11,$D6,$2E,$5F,$80,$BF,$F6,$BC,$FE,$5F,$EB,$7A
	.dc.b	$CF,$E1,$1B,$AF,$F8,$62,$51,$87,$FC,$37,$F5,$BF,$E4,$89,$D2,$F9
	.dc.b	$7F,$F5,$83,$67,$F4,$8D,$7F,$D5,$D6,$E0,$7D,$9F,$FA,$DF,$EB,$9F
	.dc.b	$15,$49,$7F,$B3,$B7,$D8,$3D,$CF,$F5,$F7,$D7,$BE,$6E,$92,$FC,$27
	.dc.b	$FB,$0F,$EB,$DF,$1D,$49,$1D,$75,$FF,$57,$F0,$7C,$1D,$7E,$04,$9A
	.dc.b	$7C,$3D,$3C,$7F,$75,$FE,$CF,$FA,$F7,$D8,$52,$5F,$F0,$5C,$DE,$A7
	.dc.b	$FB,$57,$A9,$E6,$52,$57,$8F,$9B,$E2,$44,$8F,$E1,$DA,$70,$7E,$27
	.dc.b	$1F,$FB,$25,$FE,$25,$F6,$0F,$B0,$FF,$71,$36,$33,$1F,$EE,$5F,$8F
	.dc.b	$FE,$1B,$0F,$B0,$FF,$71,$F0,$FF,$06,$6F,$17,$E0,$DB,$29,$FB,$5F
	.dc.b	$3F,$F6,$7E,$C2,$4F,$99,$FF,$7A,$4F,$D8,$F2,$3F,$DE,$9F,$3A,$FB
	.dc.b	$0A,$4B,$FE,$55,$F5,$CF,$A8,$FF,$7D,$36,$33,$18,$3F,$11,$F6,$12
	.dc.b	$3E,$2B,$5F,$C7,$F0,$FF,$1E,$7A,$F9,$DF,$9D,$CF,$F9,$1F,$D8,$7F
	.dc.b	$62,$3F,$B1,$E3,$FC,$DE,$BF,$DF,$4D,$9F,$E5,$2D,$E0,$FD,$33,$67
	.dc.b	$E2,$76,$BD,$2F,$82,$D8,$F0,$90,$FD,$1B,$97,$E6,$57,$EC,$7E,$C3
	.dc.b	$FB,$76,$1F,$CE,$1B,$19,$F6,$7E,$2F,$D8,$7A,$E3,$17,$FD,$4E,$B6
	.dc.b	$FF,$A4,$9A,$B2,$57,$18,$D5,$D5,$5C,$0E,$35,$C2,$3B,$7C,$8E,$E4
	.dc.b	$D3,$EC,$FC,$3D,$9D,$90,$F9,$AF,$7F,$B0,$74,$D2,$4F,$F6,$D8,$79
	.dc.b	$B8,$8F,$8E,$7F,$AB,$A3,$8B,$B9,$E9,$3E,$BF,$57,$F8,$44,$FF,$77
	.dc.b	$6B,$C8,$FE,$C7,$E5,$7C,$1F,$B0,$EF,$C1,$FE,$BA,$FA,$FF,$C5,$56
	.dc.b	$8E,$55,$D4,$E1,$87,$E9,$61,$3C,$2F,$9C,$63,$0F,$E5,$C7,$8A,$7A
	.dc.b	$B4,$FE,$65,$3C,$23,$95,$75,$77,$7F,$CD,$DB,$7E,$7B,$E3,$5B,$6E
	.dc.b	$19,$9F,$F6,$AC,$77,$6D,$F3,$8D,$C7,$9C,$FF,$6D,$E1,$E3,$8F,$4C
	.dc.b	$EE,$78,$9D,$44,$ED,$E3,$7E,$37,$67,$53,$E6,$16,$38,$3C,$A3,$E6
	.dc.b	$3F,$D5,$4C,$5D,$7A,$B5,$C9,$99,$FE,$DF,$F6,$19,$6C,$67,$D2,$4E
	.dc.b	$5B,$1C,$C7,$D1,$56,$99,$70,$47,$D9,$7F,$9C,$E4,$F8,$90,$38,$E9
	.dc.b	$F8,$C9,$D4,$7C,$33,$55,$F2,$E3,$33,$F9,$29,$C9,$DB,$5F,$F2,$5C
	.dc.b	$66,$3A,$EB,$FC,$7E,$72,$FC,$7B,$E9,$18,$9F,$99,$18,$C3,$CF,$1C
	.dc.b	$8F,$F3,$BC,$ED,$FF,$03,$4E,$DF,$41,$B3,$FA,$34,$6F,$0F,$69,$FE
	.dc.b	$52,$2D,$78,$E7,$D4,$FF,$9A,$E3,$31,$FE,$86,$F4,$3F,$CC,$F1,$C6
	.dc.b	$AD,$7F,$3B,$9B,$5F,$37,$EC,$3F,$C0,$D3,$FE,$75,$6C,$7E,$FD,$FA
	.dc.b	$51,$9D,$F8,$CA,$77,$87,$83,$B7,$C7,$39,$AE,$F3,$FD,$5F,$1E,$3F
	.dc.b	$91,$F0,$C7,$FC,$54,$DB,$FE,$D1,$F3,$7E,$07,$F8,$84,$65,$3E,$67
	.dc.b	$F3,$CD,$79,$E1,$FE,$7A,$6D,$7E,$7F,$EB,$BF,$C5,$2F,$FA,$4D,$B1
	.dc.b	$FE,$30,$BF,$EA,$93,$9D,$3D,$F5,$AC,$BF,$D0,$AD,$9F,$F5,$AD,$70
	.dc.b	$BF,$E4,$D3,$97,$F3,$4F,$67,$F6,$8D,$7F,$C9,$4D,$77,$7E,$97,$F4
	.dc.b	$5F,$89,$BE,$27,$10,$8C,$61,$FB,$B6,$C7,$EF,$DA,$FF,$03,$5F,$37
	.dc.b	$C9,$B3,$C6,$30,$67,$C5,$91,$D5,$B5,$FC,$2F,$F2,$1F,$80,$FF,$0F
	.dc.b	$6B,$FE,$65,$6B,$FB,$B6,$AF,$C6,$7F,$DB,$95,$FC,$4D,$B1,$FC,$5D
	.dc.b	$AF,$F1,$B6,$BF,$E7,$15,$F4,$1F,$11,$B3,$CE,$30,$7C,$8F,$67,$FC
	.dc.b	$81,$1F,$C8,$5B,$5F,$C9,$5A,$FF,$29,$6B,$FE,$7F,$7F,$DB,$D1,$C7
	.dc.b	$C6,$E8,$47,$A7,$C7,$F7,$23,$83,$F9,$87,$E4,$3F,$E7,$D6,$F8,$DF
	.dc.b	$72,$31,$87,$F3,$76,$BF,$CE,$D7,$CA,$EC,$95,$D1,$E5,$EE,$48,$EE
	.dc.b	$DA,$FE,$85,$F6,$AF,$B8,$FC,$5B,$6B,$FA,$23,$67,$73,$F9,$0F,$A8
	.dc.b	$FF,$42,$D7,$F4,$96,$CF,$F4,$D6,$BF,$D4,$5A,$FF,$A9,$FE,$87,$E5
	.dc.b	$BF,$CF,$5A,$9A,$63,$F1,$57,$DF,$FE,$81,$B7,$FD,$65,$B5,$FD,$75
	.dc.b	$AF,$F6,$1F,$C3,$BF,$D6,$13,$FD,$7D,$B1,$FD,$8D,$AE,$FD,$DF,$D6
	.dc.b	$FD,$27,$E9,$BE,$97,$FB,$33,$67,$FB,$53,$FB,$3C,$B3,$D2,$AF,$ED
	.dc.b	$9F,$18,$FB,$7F,$F3,$EF,$E0,$3F,$EA,$CC,$7F,$D6,$CD,$7F,$C5,$93
	.dc.b	$B7,$FB,$07,$84,$D7,$FD,$9D,$39,$9F,$EF,$4D,$7F,$BF,$35,$DD,$FF
	.dc.b	$73,$8E,$15,$CE,$7F,$B6,$46,$2B,$71,$FE,$C9,$FC,$1F,$FD,$33,$47
	.dc.b	$FC,$E0,$FF,$4C,$FB,$2F,$EA,$95,$E6,$6D,$7F,$A7,$EB,$73,$EA,$6C
	.dc.b	$7F,$A8,$E7,$74,$6C,$EA,$1F,$87,$3E,$4F,$FD,$4A,$DB,$FF,$6F,$35
	.dc.b	$FF,$73,$AF,$93,$DC,$C9,$DB,$CA,$3F,$CF,$1F,$88,$FF,$97,$2F,$FB
	.dc.b	$3F,$E2,$6F,$82,$FF,$59,$B7,$0F,$A7,$A7,$FB,$FE,$F8,$37,$CB,$FF
	.dc.b	$10,$47,$F9,$9A,$7F,$C6,$0D,$74,$7F,$8B,$DE,$2F,$F8,$75,$A9,$CE
	.dc.b	$0B,$FA,$67,$C5,$FC,$4F,$E4,$9F,$29,$F9,$06,$70,$FF,$65,$C6,$7D
	.dc.b	$FF,$F6,$6F,$D4,$7C,$7B,$FF,$02,$B6,$BF,$1E,$DC,$7F,$DB,$0F,$CB
	.dc.b	$7E,$09,$8D,$79,$91,$CD,$27,$84,$7F,$A1,$FE,$EB,$0F,$E3,$4F,$F6
	.dc.b	$FF,$07,$F9,$56,$BF,$29,$09,$DB,$C6,$BF,$C9,$91,$88,$FF,$2E,$35
	.dc.b	$FF,$23,$1C,$63,$67,$FA,$D6,$39,$CF,$66,$E1,$F9,$5F,$A4,$FF,$78
	.dc.b	$F1,$7F,$4A,$C3,$D6,$BE,$77,$26,$BC,$99,$3A,$8C,$BC,$23,$F2,$FC
	.dc.b	$D2,$3F,$2E,$FE,$6B,$E4,$B1,$1F,$98,$8B,$9F,$F8,$E3,$99,$5F,$EC
	.dc.b	$DF,$C2,$DF,$23,$F9,$4F,$6A,$63,$F5,$04,$E9,$C3,$5F,$F4,$03,$5F
	.dc.b	$F4,$23,$5F,$92,$D7,$FC,$E0,$F1,$C3,$EA,$B7,$6D,$DF,$E9,$B6,$FC
	.dc.b	$7E,$9F,$90,$D6,$63,$F8,$4E,$4F,$98,$FF,$B4,$B2,$46,$1F,$F2,$BB
	.dc.b	$F3,$E3,$84,$F4,$7F,$1B,$C3,$FC,$5D,$FC,$AE,$FF,$9C,$3F,$3B,$B5
	.dc.b	$8D,$C5,$7C,$13,$F3,$1F,$91,$F8,$8B,$E6,$FE,$77,$64,$9C,$C7,$FB
	.dc.b	$F6,$75,$1F,$32,$74,$FF,$B8,$E3,$2D,$9C,$E7,$F2,$2B,$94,$49,$E4
	.dc.b	$43,$81,$E0,$31,$3C,$5C,$1A,$3F,$DE,$5C,$0A,$EF,$C1,$FA,$2C,$3C
	.dc.b	$77,$FA,$C6,$C7,$DE,$9D,$DF,$FA,$B6,$72,$DC,$32,$FE,$32,$7D,$0F
	.dc.b	$86,$FC,$8F,$8A,$33,$1D,$E7,$8E,$1E,$5A,$F4,$CE,$5F,$6E,$1E,$37
	.dc.b	$2F,$A6,$B8,$DF,$48,$C4,$F0,$3F,$2B,$19,$7D,$06,$4F,$41,$D0,$C9
	.dc.b	$E7,$35,$87,$C0,$4D,$C3,$D1,$1B,$3E,$05,$78,$9C,$61,$F0,$8F,$63
	.dc.b	$A3,$CA,$76,$C3,$CC,$F9,$0E,$33,$CE,$78,$9C,$DF,$89,$1C,$CF,$2C
	.dc.b	$FC,$47,$D8,$60,$F8,$15,$E8,$AE,$75,$D4,$E6,$F5,$9F,$86,$E9,$F3
	.dc.b	$9E,$43,$31,$EF,$7F,$96,$9D,$CA,$53,$37,$FA,$26,$DE,$9F,$A2,$60
	.dc.b	$C4,$6D,$F3,$1E,$02,$1F,$C7,$3E,$1F,$E4,$C7,$D7,$3E,$B3,$89,$F9
	.dc.b	$B8,$23,$15,$F1,$A3,$0F,$C8,$AE,$55,$FC,$B6,$B7,$3F,$3E,$B6,$FD
	.dc.b	$1A,$DB,$FC,$B2,$77,$87,$C9,$8F,$31,$D6,$FC,$73,$AC,$3D,$B3,$A3
	.dc.b	$A4,$E9,$E0,$79,$E7,$83,$F0,$B0,$F1,$DF,$7C,$3C,$73,$FC,$0A,$B6
	.dc.b	$FD,$59,$FE,$6F,$18,$9F,$8F,$3D,$8F,$74,$66,$FE,$14,$66,$BD,$97
	.dc.b	$E0,$C3,$D3,$87,$82,$7A,$D6,$E7,$DB,$5B,$8F,$B3,$5B,$9F,$E9,$B8
	.dc.b	$77,$9F,$94,$71,$3D,$CF,$F7,$2F,$57,$FB,$9F,$83,$27,$FB,$9C,$65
	.dc.b	$FE,$EB,$19,$7F,$A6,$73,$7F,$B9,$5F,$F2,$2F,$B5,$FE,$48,$23,$FE
	.dc.b	$48,$F5,$EB,$FE,$48,$8B,$A7,$F7,$19,$FE,$81,$F8,$37,$F7,$1F,$C1
	.dc.b	$A5,$FF,$48,$8C,$BF,$9A,$7F,$3C,$FF,$0A,$6E,$A6,$9C,$23,$FA,$43
	.dc.b	$FD,$F7,$EA,$FF,$C6,$FA,$7F,$8E,$34,$95,$FD,$A6,$32,$FF,$6B,$8C
	.dc.b	$BF,$DA,$F9,$BE,$03,$FD,$7B,$4F,$F1,$F9,$1F,$ED,$39,$D7,$47,$F6
	.dc.b	$C8,$FF,$22,$C7,$F5,$FF,$C1,$BF,$C2,$1F,$4B,$2B,$FB,$3C,$65,$FD
	.dc.b	$CB,$EE,$23,$83,$64,$3F,$AB,$F1,$FF,$8D,$B7,$FF,$49,$0F,$C2,$BF
	.dc.b	$CD,$D3,$FE,$05,$6C,$69,$BC,$7F,$EA,$45,$DB,$CE,$31,$E3,$5B,$5F
	.dc.b	$58,$C3,$63,$FC,$AF,$7C,$9B,$EB,$EA,$1D,$D4,$C4,$49,$E3,$19,$78
	.dc.b	$9F,$E3,$3F,$A1,$FF,$30,$DF,$F8,$A3,$8B,$FC,$74,$79,$65,$3A,$CB
	.dc.b	$FC,$B2,$FF,$9A,$23,$F9,$7B,$5C,$F9,$3F,$8E,$8C,$40,$E7,$1D,$63
	.dc.b	$A4,$65,$AF,$F9,$1B,$BF,$DD,$E3,$48,$F8,$6F,$C6,$FC,$23,$FD,$6B
	.dc.b	$AF,$FA,$CC,$91,$9C,$75,$C5,$BF,$A8,$98,$7F,$A0,$B0,$4E,$A3,$13
	.dc.b	$A7,$FC,$85,$3A,$7E,$47,$C4,$7C,$5F,$A3,$FF,$63,$C1,$BA,$27,$DC
	.dc.b	$7C,$32,$4F,$F4,$F8,$D3,$63,$6F,$63,$FC,$D1,$2F,$98,$F0,$7D,$1F
	.dc.b	$EE,$A0,$D2,$4F,$5F,$5F,$E9,$11,$C1,$FD,$7F,$E8,$DF,$47,$FC,$11
	.dc.b	$B1,$98,$C1,$FC,$23,$F7,$0F,$C9,$47,$F6,$C6,$BF,$DC,$1A,$FF,$A7
	.dc.b	$9A,$FE,$05,$7D,$F7,$32,$47,$F7,$18,$CB,$7E,$63,$C0,$47,$06,$FE
	.dc.b	$D3,$C8,$47,$07,$3F,$AD,$F4,$11,$C1,$DF,$E8,$7D,$84,$70,$78,$FE
	.dc.b	$1F,$E0,$11,$C1,$E7,$ED,$F8,$84,$70,$7A,$FB,$7E,$41,$1C,$1E,$FF
	.dc.b	$88,$F9,$84,$70,$7C,$3F,$90,$FA,$04,$70,$7C,$7E,$1F,$A8,$47,$07
	.dc.b	$CB,$F4,$3E,$C1,$1C,$1F,$3F,$93,$F7,$DB,$FF,$B3,$A3,$EB,$4E,$26
	.dc.b	$46,$C8,$7E,$05,$FE,$97,$F0,$FF,$EA,$5E,$BF,$DF,$78,$2F,$9D,$FB
	.dc.b	$77,$FE,$AB,$6C,$7F,$BA,$AB,$84,$FF,$05,$6C,$71,$E4,$FE,$47,$5C
	.dc.b	$2F,$FD,$D3,$5C,$2C,$FF,$AE,$1F,$F2,$0B,$5D,$FA,$BE,$63,$CF,$E1
	.dc.b	$F4,$E6,$93,$FE,$D5,$AF,$EA,$B9,$79,$6B,$33,$E9,$AF,$E4,$95,$AC
	.dc.b	$3F,$DF,$7F,$30,$F2,$BF,$94,$36,$FF,$DC,$0D,$7F,$9F,$CF,$0F,$98
	.dc.b	$DD,$49,$DB,$CA,$3F,$99,$79,$0F,$7E,$92,$BF,$DD,$9F,$30,$AF,$E9
	.dc.b	$70,$F9,$7F,$DB,$DF,$18,$FF,$79,$35,$F1,$9F,$CD,$5C,$6C,$CB,$F9
	.dc.b	$C4,$B0,$ED,$3F,$CE,$EF,$F8,$1C,$62,$7C,$17,$C0,$ED,$3E,$2C,$79
	.dc.b	$BC,$31,$F3,$B7,$0E,$95,$C6,$B6,$DB,$C9,$FD,$2F,$0F,$05,$F1,$8F
	.dc.b	$09,$EB,$AE,$2D,$38,$9E,$BC,$3C,$37,$EE,$BE,$6F,$79,$E6,$7F,$7A
	.dc.b	$AE,$6F,$1C,$39,$D7,$08,$EC,$70,$AF,$82,$75,$BE,$F1,$E9,$7E,$BB
	.dc.b	$FE,$98,$9F,$94,$D8,$CB,$C6,$BF,$D2,$D3,$FD,$37,$E6,$2F,$86,$69
	.dc.b	$2B,$FC,$7F,$F3,$1F,$E4,$04,$D2,$3A,$3D,$F7,$F9,$BD,$3E,$3C,$7F
	.dc.b	$79,$9F,$93,$39,$8F,$8A,$64,$E3,$AF,$E2,$5C,$9E,$03,$35,$D1,$C6
	.dc.b	$BF,$D3,$2B,$A3,$FD,$2A,$FB,$DF,$38,$C5,$72,$3A,$CE,$AF,$A3,$63
	.dc.b	$26,$E7,$B5,$7F,$8F,$2B,$FC,$7C,$D9,$FF,$64,$D7,$1B,$F5,$47,$F8
	.dc.b	$DA,$74,$FF,$8E,$3D,$0B,$3F,$0C,$D8,$77,$F3,$1F,$B9,$C3,$F9,$29
	.dc.b	$FA,$C7,$15,$FB,$20,$FF,$89,$E3,$71,$FE,$29,$8D,$D7,$91,$B3,$FC
	.dc.b	$C3,$AF,$C4,$F5,$E0,$F6,$52,$73,$7F,$E3,$3F,$98,$F0,$BF,$EE,$86
	.dc.b	$BF,$C0,$1A,$EE,$BE,$1B,$97,$AF,$CC,$3E,$27,$E2,$D7,$92,$BF,$CB
	.dc.b	$B3,$9C,$7D,$70,$F9,$8F,$8D,$18,$6C,$7F,$14,$C9,$4F,$3F,$E8,$D6
	.dc.b	$A7,$EA,$BC,$2B,$FE,$00,$B5,$FE,$32,$D7,$FC,$A4,$D7,$C5,$AF,$C6
	.dc.b	$FC,$65,$EE,$76,$8C,$3F,$80,$F9,$97,$B5,$F0,$AF,$FC,$B9,$18,$D3
	.dc.b	$F9,$3C,$7F,$28,$8C,$38,$0E,$1D,$AF,$99,$E9,$1A,$6C,$66,$79,$4E
	.dc.b	$5B,$F9,$1F,$F6,$26,$F6,$F2,$8C,$4F,$85,$AF,$9B,$E6,$7F,$9F,$57
	.dc.b	$46,$C7,$9D,$AE,$E3,$CB,$0F,$06,$92,$39,$3E,$AF,$82,$7D,$15,$24
	.dc.b	$FA,$A3,$2F,$B3,$CA,$79,$BF,$09,$B1,$98,$C1,$9F,$36,$47,$16,$E5
	.dc.b	$F3,$FF,$13,$7B,$9F,$31,$B5,$F2,$9B,$3B,$9F,$A3,$F3,$5F,$1A,$3F
	.dc.b	$B0,$B5,$FF,$4D,$35,$34,$7C,$D7,$22,$5F,$D8,$8C,$BF,$6B,$F1,$27
	.dc.b	$B9,$F8,$06,$C6,$63,$07,$E1,$3E,$67,$8E,$1F,$8B,$6B,$F9,$06,$BF
	.dc.b	$DE,$1A,$FF,$AC,$7B,$DE,$AF,$E3,$9B,$7F,$8C,$6C,$FE,$5D,$AF,$E7
	.dc.b	$3E,$FB,$F0,$13,$F9,$26,$C7,$E7,$5A,$EF,$8F,$F3,$7F,$32,$F9,$9A
	.dc.b	$4B,$FC,$DC,$65,$FC,$EF,$CC,$FE,$BB,$E6,$6F,$69,$CF,$A7,$FC,$3F
	.dc.b	$2F,$D6,$56,$DF,$D7,$7D,$BB,$E2,$FF,$4F,$08,$C3,$FA,$A6,$BC,$23
	.dc.b	$7F,$2B,$FE,$63,$F1,$9F,$2D,$FB,$C7,$FA,$E1,$8A,$FF,$73,$4F,$F9
	.dc.b	$A5,$2F,$6F,$99,$23,$F7,$31,$96,$C7,$22,$3F,$D0,$27,$C3,$EF,$7F
	.dc.b	$8C,$9F,$0F,$C8,$7F,$99,$87,$C5,$67,$83,$FC,$E3,$18,$BE,$4E,$F8
	.dc.b	$F5,$0B,$4E,$E3,$F9,$83,$63,$BC,$F0,$5F,$2C,$F9,$4C,$36,$FD,$0D
	.dc.b	$7F,$9D,$2F,$95,$DC,$C9,$DB,$CA,$3C,$B0,$F9,$47,$CB,$78,$7E,$63
	.dc.b	$FC,$0D,$A7,$FA,$39,$B4,$38,$4F,$2A,$E3,$1F,$CA,$1A,$FF,$33,$6B
	.dc.b	$FC,$E1,$AF,$F9,$81,$DE,$5F,$E0,$26,$DE,$6B,$C9,$B7,$C5,$F9,$4F
	.dc.b	$F4,$C9,$F2,$67,$FD,$46,$DC,$3E,$83,$AD,$78,$6F,$27,$F0,$13,$FE
	.dc.b	$A2,$E0,$F9,$59,$FE,$41,$B8,$7F,$AB,$9B,$5F,$6D,$F9,$CF,$E2,$5C
	.dc.b	$CF,$E4,$F6,$FB,$51,$83,$73,$FE,$11,$F9,$67,$CA,$7C,$9F,$97,$93
	.dc.b	$FD,$E2,$72,$D9,$FD,$6C,$62,$3F,$AF,$35,$FE,$C6,$D7,$FD,$16,$EF
	.dc.b	$0F,$F1,$6B,$71,$CD,$7D,$9F,$2F,$F2,$1C,$9F,$EA,$FC,$3F,$30,$FF
	.dc.b	$72,$7F,$3B,$D9,$AE,$B9,$18,$21,$F9,$C7,$F2,$FE,$E4,$8F,$CB,$BF
	.dc.b	$9A,$F5,$31,$1F,$98,$8B,$A7,$F0,$0E,$8F,$F1,$E4,$FC,$3F,$97,$FC
	.dc.b	$77,$AB,$23,$04,$3E,$1C,$7C,$62,$3F,$DE,$B7,$47,$E8,$1B,$1C,$6F
	.dc.b	$F2,$1D,$BF,$8D,$CF,$F1,$DA,$4A,$FC,$3C,$65,$FC,$4C,$65,$FF,$15
	.dc.b	$72,$7C,$9B,$FE,$29,$A7,$E0,$89,$3F,$87,$D6,$BF,$E3,$1E,$3F,$F6
	.dc.b	$C5,$7E,$46,$BF,$13,$5E,$7F,$97,$FD,$77,$CA,$CC,$3F,$0D,$19,$7F
	.dc.b	$C7,$FD,$69,$D4,$7E,$3E,$74,$FF,$98,$67,$4F,$09,$CE,$1F,$86,$AE
	.dc.b	$2F,$EC,$E3,$3C,$BF,$8C,$9C,$BF,$B9,$8C,$D7,$F0,$5D,$D5,$A8,$FE
	.dc.b	$11,$87,$81,$F2,$1E,$A3,$C4,$43,$24,$C3,$C4,$E5,$F3,$3C,$4F,$65
	.dc.b	$F7,$57,$E2,$75,$3C,$9D,$98,$C3,$D8,$F5,$7D,$8E,$4F,$66,$1D,$70
	.dc.b	$F5,$57,$8A,$7C,$D5,$C1,$EA,$60,$E2,$75,$3E,$1E,$1F,$21,$F3,$DF
	.dc.b	$58,$C6,$1E,$99,$E7,$7E,$93,$C0,$4B,$F0,$39,$7D,$71,$E1,$3F,$9F
	.dc.b	$57,$C5,$AE,$95,$C6,$7C,$67,$94,$F1,$9F,$04,$E2,$66,$38,$D7,$69
	.dc.b	$F0,$4F,$23,$C1,$3A,$71,$1A,$7A,$9E,$5C,$3E,$5D,$7A,$8E,$84,$8E
	.dc.b	$2E,$5F,$96,$7E,$18,$FE,$43,$87,$F2,$62,$3F,$C0,$5A,$FF,$28,$6B
	.dc.b	$C0,$F7,$5F,$9D,$FE,$C9,$CB,$F4,$A7,$EB,$5F,$F2,$99,$F0,$E1,$F2
	.dc.b	$5F,$5B,$FD,$66,$B8,$47,$C9,$9E,$0E,$B0,$F6,$34,$FD,$2C,$F0,$7E
	.dc.b	$23,$6B,$B5,$72,$C3,$CC,$4C,$3F,$B9,$56,$F0,$EB,$5B,$9C,$61,$D8
	.dc.b	$F7,$BE,$CA,$CC,$78,$CF,$82,$68,$F5,$3A,$BF,$8D,$5D,$67,$47,$F0
	.dc.b	$D9,$D5,$79,$23,$38,$7A,$A7,$31,$FC,$61,$AF,$13,$78,$7A,$A3,$33
	.dc.b	$F3,$67,$27,$D2,$9F,$F0,$54,$E9,$B1,$A3,$D9,$7F,$66,$9F,$2C,$F9
	.dc.b	$7F,$F1,$83,$6B,$FC,$88,$D7,$FD,$CC,$FF,$88,$A7,$8C,$F0,$24,$F0
	.dc.b	$F8,$8E,$7E,$97,$4E,$E9,$3D,$23,$2F,$68,$CB,$DA,$3F,$C6,$BF,$2D
	.dc.b	$FE,$AC,$8F,$E7,$BC,$BF,$E2,$F3,$3E,$94,$8F,$F6,$BC,$ED,$FF,$71
	.dc.b	$35,$FF,$73,$B5,$FF,$1D,$07,$86,$4D,$73,$18,$6D,$7F,$8F,$FB,$FE
	.dc.b	$3F,$BD,$30,$F8,$EF,$F9,$67,$E5,$F9,$46,$08,$7C,$A8,$D5,$EF,$93
	.dc.b	$87,$A3,$2B,$FC,$E9,$7F,$C1,$1F,$F3,$53,$73,$3F,$EB,$C8,$DC,$F3
	.dc.b	$7F,$CA,$D3,$E2,$7D,$DF,$11,$FE,$60,$F9,$76,$54,$F9,$F1,$27,$E5
	.dc.b	$BF,$E9,$CD,$4D,$24,$76,$CF,$E3,$C7,$18,$FF,$40,$47,$CE,$B7,$C8
	.dc.b	$7C,$5F,$1E,$57,$B9,$F9,$1E,$2E,$23,$89,$F0,$23,$19,$FF,$A9,$35
	.dc.b	$57,$FA,$1F,$F0,$F7,$CC,$DE,$29,$8F,$FA,$76,$72,$D9,$CB,$DB,$04
	.dc.b	$FD,$F8,$FF,$4C,$3C,$1F,$19,$FD,$F2,$7C,$1F,$6B,$23,$E1,$1F,$E8
	.dc.b	$5E,$CE,$17,$F4,$5E,$51,$F9,$03,$91,$4E,$15,$FE,$CA,$8D,$C7,$E5
	.dc.b	$9E,$7B,$59,$FF,$4C,$C6,$27,$CF,$F8,$4F,$C3,$FA,$19,$1F,$05,$F8
	.dc.b	$DF,$07,$88,$F8,$19,$A7,$19,$7F,$B4,$21,$5F,$EB,$39,$FE,$05,$3F
	.dc.b	$ED,$47,$F8,$11,$3B,$34,$6F,$B2,$61,$FE,$CD,$9E,$3A,$7E,$0F,$D6
	.dc.b	$E9,$D9,$2B,$E4,$8E,$5F,$F7,$D3,$86,$BF,$1D,$AF,$CB,$6B,$D4,$F9
	.dc.b	$87,$39,$FD,$01,$1F,$27,$CB,$FF,$0E,$F9,$CF,$CF,$12,$3F,$85,$BF
	.dc.b	$EF,$CF,$9C,$7C,$E5,$24,$F6,$8C,$B7,$E1,$38,$11,$C1,$E4,$F9,$A7
	.dc.b	$CD,$79,$A1,$18,$75,$19,$8C,$47,$E1,$D7,$CC,$E4,$91,$D5,$E9,$C7
	.dc.b	$FE,$99,$8F,$C7,$35,$FC,$93,$5F,$E4,$6D,$7F,$DB,$D2,$F9,$BF,$11
	.dc.b	$23,$D4,$F9,$7E,$0A,$47,$B5,$AF,$83,$E3,$7C,$64,$7E,$AD,$A3,$DD
	.dc.b	$C3,$27,$E3,$35,$F8,$1E,$47,$C3,$23,$F2,$DF,$7F,$97,$23,$AB,$5F
	.dc.b	$91,$F2,$7F,$2C,$8F,$D3,$7E,$77,$14,$8F,$AC,$D7,$E4,$F9,$1F,$30
	.dc.b	$8F,$A9,$FA,$DE,$27,$F4,$CE,$BF,$E6,$1B,$BD,$0C,$74,$FF,$50,$8C
	.dc.b	$61,$FA,$07,$F9,$2B,$6F,$9B,$F8,$76,$BF,$81,$79,$BF,$59,$AF,$E2
	.dc.b	$9B,$BB,$D8,$BE,$D7,$0C,$9F,$37,$F8,$C7,$FA,$B9,$8E,$6D,$E3,$FB
	.dc.b	$4C,$17,$F8,$C2,$47,$DA,$6B,$FC,$9A,$36,$F1,$8F,$05,$BE,$73,$CB
	.dc.b	$E7,$E3,$A2,$5F,$78,$CB,$F8,$67,$F3,$64,$7F,$7B,$F3,$9F,$EC,$8D
	.dc.b	$FF,$BD,$B6,$01,$C3,$7E,$73,$E3,$C6,$09,$EB,$62,$3E,$5F,$86,$F9
	.dc.b	$7A,$4B,$FE,$2F,$F3,$8A,$70,$F8,$3E,$6D,$87,$F8,$38,$E5,$7F,$C7
	.dc.b	$9A,$FF,$4C,$6C,$FF,$77,$BE,$8F,$38,$FF,$0B,$C6,$25,$5D,$9F,$7C
	.dc.b	$F4,$8C,$47,$8B,$0F,$F0,$FC,$62,$B9,$37,$0F,$4D,$74,$BF,$45,$F0
	.dc.b	$6C,$78,$CF,$9B,$7C,$A3,$CB,$5C,$AF,$C3,$1E,$33,$8D,$75,$38,$CE
	.dc.b	$E3,$8E,$1D,$67,$D0,$58,$E4,$F0,$8D,$5F,$86,$3E,$0E,$1E,$76,$9A
	.dc.b	$BE,$CF,$13,$FB,$84,$FC,$B7,$FB,$A3,$5F,$19,$AF,$45,$F3,$9F,$D3
	.dc.b	$27,$FC,$46,$D8,$CC,$EE,$3F,$9E,$2F,$9C,$A4,$7F,$A0,$37,$3F,$AF
	.dc.b	$93,$4F,$F2,$4C,$7D,$AA,$FC,$DD,$7C,$5A,$D6,$DF,$1E,$72,$7C,$D3
	.dc.b	$F2,$77,$CB,$5F,$F0,$75,$F1,$8C,$4F,$F8,$4E,$31,$5C,$E3,$D0,$73
	.dc.b	$BE,$EF,$53,$BE,$1E,$8B,$E7,$1F,$E6,$CB,$E6,$EC,$E7,$3B,$BE,$D3
	.dc.b	$A8,$C4,$FF,$8B,$AF,$B9,$FE,$59,$7B,$E1,$FC,$B7,$0F,$E2,$F8,$53
	.dc.b	$FC,$13,$4F,$6F,$37,$FE,$62,$AA,$FE,$DA,$2F,$CE,$7E,$F6,$FF,$C2
	.dc.b	$2D,$9F,$F1,$FB,$FC,$2A,$33,$BF,$03,$AE,$0E,$EF,$3C,$FF,$A3,$2B
	.dc.b	$94,$FF,$2C,$9C,$DF,$F1,$4F,$9C,$CD,$FF,$9D,$BE,$71,$F2,$1E,$69
	.dc.b	$CD,$7F,$0C,$9C,$D7,$F8,$9A,$73,$82,$73,$5F,$E5,$69,$CC,$62,$7F
	.dc.b	$D0,$5F,$39,$F1,$6F,$FC,$4A,$DB,$FF,$29,$35,$FF,$2B,$35,$FF,$53
	.dc.b	$BB,$9F,$E8,$FF,$38,$F8,$6F,$9B,$39,$BF,$F4,$8C,$E5,$FF,$1E,$C6
	.dc.b	$6F,$F9,$3C,$E5,$FF,$36,$C6,$6D,$39,$7F,$D3,$5F,$39,$FE,$07,$BF
	.dc.b	$F1,$EB,$6F,$FC,$D4,$D7,$FC,$DC,$D7,$FD,$7F,$F3,$9F,$E5,$C9,$FF
	.dc.b	$22,$B6,$3F,$CE,$4D,$7F,$CE,$CD,$7F,$D8,$FE,$B7,$F6,$29,$FF,$26
	.dc.b	$B6,$3F,$CF,$4D,$7F,$CF,$CD,$7F,$D9,$FF,$39,$5F,$EA,$AD,$4D,$31
	.dc.b	$E1,$BE,$73,$F4,$53,$FD,$72,$31,$7F,$E5,$E6,$C7,$FA,$2D,$AF,$FA
	.dc.b	$3F,$F0,$9F,$CB,$4F,$F9,$91,$B1,$FE,$91,$6B,$BE,$2E,$EB,$D1,$79
	.dc.b	$D8,$6D,$FF,$9B,$9B,$3F,$E9,$B6,$BF,$E9,$FF,$9C,$7B,$DD,$1B,$5F
	.dc.b	$E7,$66,$CF,$FA,$8D,$AF,$FA,$9F,$F0,$57,$B7,$E4,$6D,$7F,$9F,$9B
	.dc.b	$3F,$EA,$B6,$BF,$EA,$FF,$9C,$7C,$EF,$B1,$B5,$FE,$86,$6C,$FF,$AC
	.dc.b	$DA,$FF,$AD,$FE,$75,$F1,$1F,$11,$B5,$FE,$8E,$6C,$FF,$AE,$DA,$FF
	.dc.b	$AF,$FE,$1B,$FD,$D3,$3F,$E9,$46,$C7,$FB,$05,$AE,$FC,$6F,$A3,$DC
	.dc.b	$F6,$69,$2F,$FD,$3B,$19,$7F,$D3,$FF,$39,$F8,$B2,$7A,$1F,$69,$FC
	.dc.b	$77,$80,$F9,$0F,$BA,$DA,$FF,$53,$B7,$2F,$F6,$A3,$5F,$F6,$C7,$CD
	.dc.b	$3E,$45,$F3,$7F,$EA,$E8,$C3,$7D,$1F,$E3,$75,$FE,$DD,$9C,$BF,$ED
	.dc.b	$FF,$9B,$F1,$3F,$EB,$66,$B9,$CB,$FD,$49,$18,$8F,$F7,$45,$3E,$E7
	.dc.b	$FD,$50,$D8,$FF,$5F,$B5,$FF,$76,$35,$FF,$78,$7C,$9B,$DF,$FF,$33
	.dc.b	$FC,$9B,$E3,$A9,$FE,$6A,$9D,$57,$FB,$51,$AF,$FB,$DD,$AF,$4A,$FF
	.dc.b	$5D,$46,$23,$FC,$A2,$1E,$3E,$43,$83,$31,$86,$D7,$93,$C3,$FF,$3B
	.dc.b	$7C,$BB,$3B,$C3,$32,$FF,$2E,$E5,$FE,$78,$AD,$46,$31,$CC,$62,$3F
	.dc.b	$CA,$B1,$8A,$E5,$18,$3F,$CB,$33,$BA,$FF,$41,$FC,$BF,$F9,$AD,$34
	.dc.b	$9E,$8F,$FA,$2D,$C7,$A5,$4C,$47,$CB,$F9,$7F,$9F,$E8,$4C,$7E,$64
	.dc.b	$65,$FF,$7B,$EE,$8F,$96,$48,$CB,$B7,$D0,$7C,$D7,$ED,$F2,$7F,$A7
	.dc.b	$7E,$5F,$D5,$B7,$C6,$7F,$DC,$5A,$7F,$A7,$9B,$19,$DF,$81,$C7,$79
	.dc.b	$8F,$CB,$3F,$CF,$B5,$D1,$FF,$3F,$FC,$1F,$C5,$26,$DF,$EA,$D7,$FC
	.dc.b	$25,$1C,$71,$FF,$61,$4E,$49,$19,$9D,$BE,$CD,$1F,$21,$F9,$6F,$85
	.dc.b	$93,$F9,$38,$FC,$B1,$F9,$02,$3F,$29,$AF,$37,$6F,$AC,$73,$7E,$45
	.dc.b	$FF,$4C,$FC,$BF,$F2,$E7,$D0,$D7,$F0,$D9,$23,$0F,$FB,$21,$FF,$2C
	.dc.b	$ED,$F9,$47,$2D,$F2,$3F,$DA,$11,$8A,$EF,$EC,$E3,$CD,$CF,$C4,$49
	.dc.b	$F0,$C6,$5F,$1C,$65,$F1,$D7,$F5,$47,$D9,$F2,$BF,$D3,$9F,$76,$7F
	.dc.b	$5D,$F7,$F1,$FF,$B9,$A7,$51,$F8,$A6,$BF,$B8,$0E,$1F,$DD,$B5,$FF
	.dc.b	$66,$B5,$DF,$9D,$D4,$8F,$8B,$E5,$FE,$0F,$CB,$C8,$F9,$AF,$FB,$CF
	.dc.b	$E5,$FF,$82,$46,$08,$7B,$A3,$5B,$EF,$8B,$87,$9B,$2B,$F1,$D6,$9D
	.dc.b	$46,$27,$4F,$DA,$9D,$3F,$23,$CB,$FE,$2F,$C7,$E2,$87,$C9,$B2,$CE
	.dc.b	$5F,$C7,$23,$13,$C4,$87,$ED,$1C,$B6,$7E,$85,$C8,$FC,$94,$EA,$FC
	.dc.b	$0D,$BF,$9F,$5B,$AE,$47,$0B,$C5,$F9,$F5,$57,$9A,$BF,$E0,$0E,$1C
	.dc.b	$DB,$7C,$30,$EB,$1C,$C8,$7E,$0E,$B4,$F0,$33,$87,$78,$CC,$7C,$6A
	.dc.b	$F0,$34,$F8,$73,$B7,$8D,$7C,$99,$E3,$87,$A2,$72,$77,$3C,$47,$29
	.dc.b	$E0,$E8,$FC,$AD,$F8,$B0,$F7,$47,$84,$E2,$4C,$3E,$6E,$1E,$69,$DE
	.dc.b	$1E,$1A,$D6,$1D,$9F,$9E,$60,$FC,$FE,$1C,$6E,$47,$94,$C9,$E5,$BE
	.dc.b	$B3,$E5,$9D,$CF,$33,$D2,$6D,$F8,$71,$AC,$3C,$F3,$D4,$D9,$E4,$72
	.dc.b	$F3,$C3,$D5,$18,$26,$1F,$7A,$7C,$18,$7C,$63,$CA,$F5,$71,$87,$8A
	.dc.b	$3C,$87,$86,$FE,$33,$FC,$15,$AF,$53,$18,$7F,$09,$70,$43,$E2,$99
	.dc.b	$8E,$2E,$30,$FC,$33,$E6,$F8,$BF,$13,$4F,$15,$7D,$F8,$CC,$F0,$3F
	.dc.b	$52,$E3,$1F,$81,$18,$AE,$27,$BE,$74,$F8,$67,$F7,$91,$88,$F9,$F3
	.dc.b	$C5,$F9,$D5,$E1,$79,$9E,$53,$27,$53,$31,$E5,$AE,$B1,$87,$CA,$61
	.dc.b	$E2,$73,$8E,$0F,$94,$87,$F1,$4A,$F3,$5F,$E0,$2B,$95,$70,$AF,$4C
	.dc.b	$76,$7F,$81,$D7,$0B,$FD,$13,$63,$F8,$9B,$5F,$E3,$4D,$AF,$84,$75
	.dc.b	$BF,$04,$F9,$31,$56,$F0,$F1,$9E,$19,$F1,$9B,$8F,$E8,$1E,$74,$8F
	.dc.b	$E7,$2F,$F4,$3F,$17,$11,$FC,$DE,$2E,$3F,$D1,$86,$9F,$E5,$99,$FF
	.dc.b	$07,$7C,$AB,$E5,$BF,$91,$52,$3F,$AF,$BC,$5B,$97,$F9,$29,$E1,$5C
	.dc.b	$AB,$8C,$6A,$15,$CF,$8B,$35,$FC,$83,$AB,$8D,$7F,$81,$AD,$F2,$FC
	.dc.b	$8D,$54,$9F,$D3,$1F,$E1,$E6,$B4,$CD,$74,$7F,$BA,$3D,$B0,$FE,$AC
	.dc.b	$DC,$7F,$CD,$ED,$9F,$3E,$0F,$73,$D7,$F2,$F2,$7E,$3B,$EE,$F8,$EC
	.dc.b	$47,$A6,$2E,$3F,$1E,$7E,$5C,$7F,$86,$60,$FF,$88,$27,$E2,$B6,$33
	.dc.b	$DD,$C0,$E3,$DD,$2B,$FC,$89,$5F,$E2,$37,$FB,$7F,$CB,$F6,$EF,$91
	.dc.b	$FD,$9A,$3F,$9C,$1F,$1B,$94,$D2,$47,$CC,$8C,$BC,$48,$7F,$8E,$A9
	.dc.b	$F2,$5F,$86,$FC,$47,$FC,$01,$9F,$E3,$5B,$C1,$FE,$8E,$6D,$FE,$09
	.dc.b	$7A,$BC,$66,$4F,$F9,$7E,$77,$47,$71,$9E,$CF,$07,$E2,$12,$30,$43
	.dc.b	$FC,$9F,$28,$D4,$62,$7F,$B7,$E8,$FA,$8F,$57,$54,$AF,$F2,$F4,$E5
	.dc.b	$B3,$F8,$E8,$C4,$7F,$37,$6B,$FE,$41,$6B,$FE,$BA,$77,$8F,$F8,$95
	.dc.b	$B9,$66,$BE,$57,$D3,$7E,$C3,$F1,$0F,$D4,$59,$A4,$F9,$DD,$75,$CC
	.dc.b	$BF,$D0,$D3,$96,$FD,$9F,$3F,$9E,$91,$82,$1F,$E7,$49,$46,$A3,$13
	.dc.b	$FE,$14,$FB,$3F,$F4,$6E,$3F,$C2,$6D,$D3,$E7,$35,$E9,$CC,$3F,$6B
	.dc.b	$7F,$E9,$47,$FC,$C8,$D7,$FD,$BA,$D4,$1E,$FB,$6A,$7A,$4F,$FA,$2B
	.dc.b	$E6,$5F,$23,$FE,$28,$F9,$E7,$CF,$FF,$02,$6D,$7F,$06,$6B,$FD,$B5
	.dc.b	$AF,$F9,$DD,$7C,$4F,$12,$47,$17,$F8,$6F,$DA,$BE,$22,$92,$7F,$89
	.dc.b	$C6,$5F,$E2,$BF,$3C,$F9,$CF,$E2,$AD,$8C,$C6,$0F,$E3,$3B,$BD,$9A
	.dc.b	$49,$FE,$3F,$19,$6F,$B5,$C0,$8E,$0F,$0F,$C4,$78,$08,$E0,$E7,$F3
	.dc.b	$FE,$02,$38,$3B,$FB,$DE,$42,$38,$3D,$BE,$CF,$B0,$8E,$0F,$3F,$C1
	.dc.b	$7B,$08,$E0,$F5,$FB,$FF,$00,$8E,$0F,$93,$E3,$7C,$82,$38,$3E,$1F
	.dc.b	$AF,$F2,$08,$E0,$F8,$FF,$2B,$F3,$08,$E0,$FA,$BE,$AF,$D4,$23,$82
	.dc.b	$1F,$D6,$3A,$BF,$2A,$B8,$BF,$BB,$BF,$EE,$36,$BB,$3F,$37,$C0,$F8
	.dc.b	$3F,$F3,$EB,$63,$ED,$C7,$E0,$08,$FD,$EF,$2B,$F0,$64,$70,$43,$B1
	.dc.b	$FB,$27,$A9,$FD,$E3,$AD,$F3,$D8,$9F,$F7,$C5,$FF,$2A,$78,$B7,$C0
	.dc.b	$FF,$02,$57,$09,$DF,$B9,$F9,$42,$3F,$92,$6B,$C9,$F6,$9E,$1F,$A0
	.dc.b	$FF,$4D,$FC,$FB,$E0,$E9,$2B,$F7,$F8,$7F,$B1,$E5,$F3,$F9,$9F,$6C
	.dc.b	$71,$77,$BF,$D2,$8C,$46,$A7,$FD,$53,$EA,$7F,$04,$8D,$9F,$C3,$1C
	.dc.b	$3F,$21,$B7,$FE,$D7,$F9,$FF,$0B,$FC,$4B,$E7,$53,$F8,$03,$78,$DF
	.dc.b	$17,$B9,$FE,$B8,$BE,$38,$2F,$8B,$FC,$72,$33,$18,$8E,$B3,$F8,$C8
	.dc.b	$C5,$F9,$9D,$57,$08,$FC,$8B,$6B,$D9,$5C,$67,$C9,$1E,$97,$78,$FF
	.dc.b	$8C,$23,$C9,$18,$C7,$51,$C8,$A7,$90,$B7,$DD,$36,$7C,$C7,$82,$3A
	.dc.b	$9E,$FA,$F0,$61,$EA,$6C,$EE,$3D,$F8,$7A,$6B,$86,$1E,$0B,$F4,$47
	.dc.b	$EC,$0F,$0B,$87,$D4,$7F,$A9,$F0,$FA,$2E,$DF,$F7,$DF,$CF,$FF,$91
	.dc.b	$9B,$1B,$F8,$14,$E5,$B1,$FE,$4A,$F9,$FF,$F2,$5A,$69,$1D,$1F,$00
	.dc.b	$FF,$2E,$EC,$D7,$E5,$C7,$F5,$C9,$F9,$B9,$7E,$8E,$73,$3D,$4E,$51
	.dc.b	$97,$A1,$C4,$CD,$75,$72,$71,$D3,$F1,$93,$C5,$FC,$74,$71,$73,$3C
	.dc.b	$5C,$C7,$1D,$7F,$8E,$D7,$17,$FD,$A5,$87,$A2,$F8,$46,$2B,$C7,$87
	.dc.b	$AA,$B4,$C8,$FE,$2B,$7E,$B3,$D0,$FF,$2A,$9D,$47,$F9,$22,$FF,$D2
	.dc.b	$95,$E2,$F2,$6D,$FF,$06,$9A,$FF,$A2,$A2,$FC,$FF,$F8,$0A,$BF,$7F
	.dc.b	$28,$C3,$FC,$99,$FF,$06,$C6,$9C,$DF,$91,$B1,$FC,$53,$E6,$3D,$0F
	.dc.b	$F4,$08,$CD,$7F,$A6,$A7,$2F,$F8,$63,$E7,$DE,$EF,$99,$B7,$FC,$4B
	.dc.b	$07,$C9,$F2,$B5,$69,$FE,$76,$D7,$F4,$0D,$7F,$A8,$B5,$F0,$5F,$CC
	.dc.b	$7F,$AA,$F7,$7F,$AA,$30,$F8,$6D,$8F,$EA,$98,$3E,$7F,$F2,$D5,$AF
	.dc.b	$84,$F9,$B7,$BA,$FA,$ED,$9F,$E6,$F5,$C2,$FF,$A8,$CE,$5F,$F1,$CB
	.dc.b	$DD,$FD,$C3,$5F,$2D,$F9,$DD,$B5,$7C,$FB,$C5,$C3,$70,$FF,$5F,$B6
	.dc.b	$7F,$82,$1C,$27,$7E,$1C,$AD,$B6,$77,$4F,$73,$A9,$2B,$AB,$E3,$F9
	.dc.b	$E7,$C5,$D2,$4F,$FB,$42,$32,$DE,$FE,$04,$70,$6F,$CE,$E8,$47,$07
	.dc.b	$FB,$6B,$D2,$79,$7F,$0D,$B1,$98,$C1,$98,$91,$C5,$B9,$FC,$45,$CB
	.dc.b	$CC,$93,$DD,$FA,$1F,$89,$FE,$92,$BE,$5B,$5F,$E6,$0D,$7F,$C4,$2D
	.dc.b	$7F,$CE,$EB,$E6,$79,$92,$BF,$DE,$B1,$96,$F8,$5E,$02,$38,$3F,$DF
	.dc.b	$7F,$57,$FE,$92,$6F,$B7,$F8,$58,$C6,$1F,$80,$6B,$FD,$17,$E9,$FF
	.dc.b	$DB,$75,$F8,$48,$46,$1F,$C3,$B5,$E1,$F0,$3F,$91,$F9,$9F,$F3,$87
	.dc.b	$CD,$71,$6C,$FE,$39,$70,$FB,$E3,$1C,$2A,$F0,$7D,$33,$E5,$BC,$54
	.dc.b	$8C,$3A,$8C,$C6,$23,$F3,$7F,$2D,$FA,$17,$D4,$D7,$3D,$5D,$23,$11
	.dc.b	$FD,$95,$7C,$7E,$49,$1C,$9F,$1F,$C4,$7F,$AB,$21,$E0,$3E,$9F,$D9
	.dc.b	$1F,$C1,$9B,$1F,$CE,$27,$84,$9F,$F5,$C3,$7C,$6E,$A4,$7E,$94,$70
	.dc.b	$24,$7F,$4E,$E2,$FF,$20,$DF,$FA,$C2,$BF,$93,$7C,$D3,$DC,$D4,$79
	.dc.b	$A3,$0E,$A3,$38,$A7,$E3,$3F,$D9,$4C,$46,$A3,$8C,$FF,$22,$6C,$ED
	.dc.b	$E3,$7F,$E5,$4F,$9A,$FF,$29,$D7,$F3,$95,$E7,$78,$1B,$19,$F4,$70
	.dc.b	$66,$64,$64,$FE,$AB,$F5,$5F,$D1,$9F,$E8,$9F,$4E,$FA,$7F,$5B,$63
	.dc.b	$7E,$5F,$C2,$55,$C5,$FE,$42,$D7,$FC,$4C,$D7,$FD,$2E,$D7,$FD,$DA
	.dc.b	$F1,$D3,$FB,$9B,$73,$DD,$78,$BC,$2F,$9B,$F3,$BF,$1A,$7F,$CE,$ED
	.dc.b	$C3,$F1,$2D,$C0,$FC,$A8,$CE,$FC,$88,$FF,$63,$8D,$4F,$E9,$F5,$FB
	.dc.b	$7C,$FF,$32,$BF,$AB,$B6,$7E,$E3,$F5,$23,$33,$C3,$69,$1F,$DB,$E7
	.dc.b	$55,$FE,$10,$F9,$37,$C0,$7E,$23,$8F,$FC,$F7,$7F,$A2,$6A,$9D,$BF
	.dc.b	$E0,$C6,$BF,$D4,$1A,$FF,$95,$5A,$FF,$B1,$9A,$8E,$93,$F6,$E3,$9D
	.dc.b	$79,$63,$C9,$E2,$FE,$5F,$CB,$97,$FA,$47,$F3,$5E,$56,$23,$F3,$51
	.dc.b	$6F,$F3,$53,$F9,$5F,$33,$F2,$DF,$19,$2F,$F3,$51,$97,$F3,$91,$97
	.dc.b	$FC,$0B,$CE,$9F,$C8,$72,$FE,$4E,$7F,$C1,$3C,$CF,$86,$FC,$B9,$AE
	.dc.b	$89,$3F,$CC,$6B,$83,$F9,$48,$FE,$47,$18,$8F,$CA,$C7,$07,$FD,$B7
	.dc.b	$1D,$F3,$FD,$D5,$FF,$5A,$9F,$C4,$69,$F9,$42,$3F,$8B,$D2,$BF,$E1
	.dc.b	$EE,$44,$E2,$FF,$82,$47,$F8,$06,$7F,$0D,$9F,$E4,$35,$91,$82,$1F
	.dc.b	$92,$72,$FE,$52,$3F,$69,$F3,$4F,$23,$EE,$72,$4A,$FF,$29,$CE,$5B
	.dc.b	$39,$7F,$DF,$3C,$7F,$D0,$72,$FF,$32,$D6,$E7,$FD,$FA,$D8,$FC,$D5
	.dc.b	$F1,$77,$7E,$23,$1D,$1F,$91,$9C,$5F,$80,$FC,$D3,$A7,$F2,$97,$D6
	.dc.b	$7F,$2D,$08,$C3,$F9,$9C,$3C,$D3,$D6,$3A,$1E,$6B,$E8,$FE,$A6,$FC
	.dc.b	$2E,$4F,$09,$0C,$13,$0F,$6B,$97,$DA,$F6,$72,$FC,$88,$EB,$1A,$8C
	.dc.b	$1C,$6B,$D7,$87,$AA,$3D,$97,$E6,$C3,$D7,$1F,$0A,$39,$D7,$B6,$39
	.dc.b	$9E,$57,$9C,$71,$C3,$73,$C4,$F0,$1F,$1D,$CB,$E0,$32,$7C,$53,$D9
	.dc.b	$3D,$1C,$D7,$91,$E0,$78,$4E,$87,$2B,$F6,$1E,$D8,$E0,$71,$73,$7E
	.dc.b	$79,$EE,$79,$CF,$6E,$1F,$30,$F9,$44,$BF,$84,$E5,$F3,$47,$8C,$F5
	.dc.b	$1C,$9C,$C7,$18,$F3,$99,$3C,$0F,$D7,$7F,$DF,$95,$F9,$F6,$D7,$F8
	.dc.b	$A2,$31,$1E,$03,$27,$F8,$F5,$CB,$EB,$32,$79,$5F,$89,$C7,$F4,$9F
	.dc.b	$87,$1F,$CE,$E3,$13,$F5,$B0,$FE,$1D,$3C,$63,$13,$A8,$F9,$27,$F4
	.dc.b	$57,$0F,$C3,$9E,$37,$F0,$63,$EC,$B6,$33,$87,$C6,$24,$FC,$47,$FA
	.dc.b	$45,$74,$AE,$0D,$8D,$CF,$84,$F1,$E1,$EC,$7F,$85,$CE,$DE,$87,$EE
	.dc.b	$5C,$3E,$49,$EE,$F2,$C3,$DE,$D3,$DF,$87,$A1,$E6,$71,$2B,$97,$04
	.dc.b	$BF,$A2,$F3,$3E,$35,$F8,$A3,$31,$F3,$CE,$F1,$9A,$FE,$CC,$D8,$FB
	.dc.b	$07,$3C,$3E,$2B,$70,$F8,$B8,$79,$E7,$FC,$89,$F2,$5F,$E4,$12,$57
	.dc.b	$F9,$3B,$E2,$2B,$FE,$43,$8B,$3F,$E4,$39,$FF,$1F,$63,$FE,$3F,$FA
	.dc.b	$69,$7F,$E4,$38,$CB,$FE,$45,$8C,$BF,$DA,$A3,$FA,$2F,$CA,$FF,$7F
	.dc.b	$F9,$69,$3F,$D1,$23,$2F,$E6,$DF,$D0,$11,$C1,$B2,$1F,$D2,$3E,$AB
	.dc.b	$FC,$6B,$94,$8F,$F1,$63,$FE,$38,$C7,$11,$FD,$8E,$2E,$5F,$D8,$A7
	.dc.b	$FC,$5F,$8F,$F8,$C3,$29,$7F,$E2,$98,$CB,$FE,$2B,$8C,$BF,$E3,$48
	.dc.b	$FF,$2A,$7C,$7B,$E4,$3F,$C3,$9F,$0F,$2B,$FC,$41,$8B,$7F,$E9,$B7
	.dc.b	$B9,$CC,$62,$FF,$D4,$F3,$99,$FF,$7E,$B5,$FE,$AA,$1C,$3C,$4E,$67
	.dc.b	$F8,$BD,$E5,$5E,$18,$C4,$FF,$8C,$7A,$FF,$D1,$3F,$3F,$FE,$0F,$C3
	.dc.b	$CE,$FF,$87,$6B,$9D,$A3,$11,$FE,$A8,$9D,$F6,$72,$39,$DC,$9E,$35
	.dc.b	$D5,$FF,$57,$BF,$E9,$F7,$91,$8F,$0F,$C6,$79,$39,$2C,$72,$8A,$F0
	.dc.b	$C7,$C9,$3F,$4B,$4F,$F2,$67,$89,$2B,$E1,$9F,$1E,$34,$FF,$B9,$75
	.dc.b	$47,$D0,$24,$65,$DB,$E6,$39,$1F,$E6,$3F,$97,$FF,$58,$FC,$E3,$E2
	.dc.b	$EF,$B7,$2B,$E4,$CE,$5B,$39,$7D,$50,$9E,$73,$D3,$96,$57,$27,$BF
	.dc.b	$2E,$23,$8C,$5D,$7F,$D8,$18,$2B,$F2,$BB,$7F,$9B,$FD,$99,$87,$C4
	.dc.b	$D7,$E9,$61,$F8,$58,$FC,$65,$7E,$48,$8F,$C7,$6B,$F0,$E7,$8F,$1F
	.dc.b	$E7,$27,$FC,$F9,$3F,$EB,$7F,$8D,$7C,$2F,$E6,$23,$FC,$5B,$18,$70
	.dc.b	$1C,$3B,$6B,$FA,$07,$FC,$D0,$FF,$A2,$DA,$F1,$E5,$E6,$D3,$AC,$7E
	.dc.b	$61,$E5,$CB,$A8,$C4,$79,$23,$8F,$DA,$3E,$2E,$92,$7C,$F1,$97,$D3
	.dc.b	$F1,$BF,$EF,$16,$F9,$FE,$28,$C4,$7C,$16,$BF,$BA,$D7,$FD,$4A,$FE
	.dc.b	$EF,$CB,$FE,$02,$47,$AF,$A5,$CC,$8E,$0F,$0F,$D7,$F7,$23,$83,$CB
	.dc.b	$F3,$5E,$32,$38,$3E,$8F,$EE,$2F,$67,$E4,$B6,$33,$18,$33,$F6,$F2
	.dc.b	$38,$B5,$FB,$D2,$FA,$0E,$64,$8E,$EF,$6F,$9C,$91,$E2,$6B,$CB,$E7
	.dc.b	$7C,$64,$7C,$EF,$9B,$EB,$64,$7A,$9A,$F8,$BE,$3F,$D6,$47,$DE,$FB
	.dc.b	$BD,$9F,$F3,$FF,$5F,$E8,$72,$4E,$3A,$FF,$D0,$71,$8B,$FE,$16,$FF
	.dc.b	$BB,$23,$37,$CE,$7E,$5B,$5F,$87,$E8,$F3,$58,$3E,$55,$F3,$59,$8C
	.dc.b	$3B,$96,$2B,$C3,$63,$F9,$AD,$F2,$6D,$6A,$A4,$7D,$E7,$EE,$91,$FC
	.dc.b	$0F,$99,$E1,$A7,$9B,$4E,$F1,$27,$BC,$65,$F8,$71,$E7,$F9,$6F,$8D
	.dc.b	$1F,$C9,$DB,$97,$F0,$D6,$E9,$FE,$AE,$7E,$54,$83,$A7,$0D,$73,$AF
	.dc.b	$7D,$7E,$37,$80,$F8,$9A,$4A,$FE,$83,$F2,$CA,$FA,$30,$F9,$5F,$E1
	.dc.b	$F7,$C6,$3F,$A3,$B5,$FE,$80,$DC,$FF,$93,$5F,$37,$D1,$5F,$C6,$23
	.dc.b	$15,$D6,$BF,$D9,$67,$39,$EF,$87,$FB,$39,$E1,$87,$91,$B8,$74,$AE
	.dc.b	$35,$B6,$DE,$4F,$91,$87,$82,$F8,$C7,$84,$F9,$93,$E1,$38,$B9,$13
	.dc.b	$C6,$3F,$DC,$2D,$8E,$35,$E1,$69,$E1,$38,$E1,$E9,$AE,$2D,$8D,$9D
	.dc.b	$E3,$C4,$FA,$23,$4F,$2C,$3F,$A1,$DF,$CA,$8F,$F0,$FB,$5E,$B1,$AF
	.dc.b	$7D,$F2,$D4,$93,$FD,$3E,$93,$A8,$FE,$8B,$F2,$DF,$D1,$93,$48,$E8
	.dc.b	$F8,$D7,$FC,$13,$6F,$81,$1F,$97,$9F,$85,$39,$8F,$98,$64,$E3,$AF
	.dc.b	$F8,$C7,$5F,$EB,$71,$E0,$33,$7D,$1C,$47,$F6,$2A,$E8,$F0,$BE,$F7
	.dc.b	$CF,$4F,$ED,$91,$8C,$39,$37,$0F,$19,$D2,$7F,$8F,$D7,$18,$FF,$25
	.dc.b	$CE,$EF,$C1,$3F,$E4,$7C,$3F,$C0,$87,$B4,$FF,$03,$57,$B6,$77,$87
	.dc.b	$F8,$BA,$74,$FF,$2C,$F8,$3E,$4F,$80,$D3,$8A,$FE,$01,$F2,$DF,$B9
	.dc.b	$9F,$F1,$BB,$63,$FC,$60,$FF,$05,$8C,$EF,$C6,$53,$BB,$F4,$7A,$7E
	.dc.b	$29,$EB,$5D,$E5,$39,$97,$CA,$BD,$4F,$6C,$7F,$13,$8C,$63,$FE,$16
	.dc.b	$6C,$7F,$86,$DA,$FF,$9D,$1C,$DF,$C9,$F9,$6F,$F4,$25,$7F,$94,$A7
	.dc.b	$2F,$FA,$12,$33,$5F,$E4,$DF,$95,$7B,$FE,$D6,$D7,$F9,$CB,$E5,$BF
	.dc.b	$D1,$4F,$F9,$72,$33,$5C,$6B,$D2,$A3,$2F,$FA,$21,$AF,$F2,$F3,$51
	.dc.b	$C2,$39,$E3,$F8,$1B,$7C,$CF,$8E,$BE,$54,$63,$0F,$F3,$6F,$CB,$BE
	.dc.b	$13,$C2,$D8,$FF,$1E,$B7,$2F,$F2,$13,$5F,$F4,$DA,$F8,$5E,$24,$9E
	.dc.b	$2F,$B3,$EB,$5F,$3B,$49,$3F,$E4,$A8,$CB,$FE,$4C,$F9,$97,$C9,$FC
	.dc.b	$46,$C6,$63,$06,$7E,$4E,$47,$16,$E5,$FE,$59,$F9,$A7,$D0,$FC,$E6
	.dc.b	$D7,$F9,$69,$B3,$B9,$FA,$7E,$D8,$A4,$AF,$F3,$0B,$96,$F3,$F0,$23
	.dc.b	$83,$FC,$CF,$F5,$8F,$33,$EF,$B6,$33,$18,$3F,$07,$F3,$6F,$7A,$92
	.dc.b	$7F,$CE,$31,$96,$F6,$70,$23,$83,$FC,$EF,$F5,$4F,$82,$FC,$7B,$63
	.dc.b	$31,$83,$F2,$7F,$36,$F9,$BA,$49,$FF,$40,$C6,$5F,$F4,$1F,$CD,$B8
	.dc.b	$BF,$CA,$5C,$3F,$EE,$08,$C7,$AF,$AE,$4F,$64,$FF,$A3,$67,$2F,$FA
	.dc.b	$3F,$EA,$7F,$D4,$4F,$FA,$39,$AE,$7E,$17,$FD,$0F,$18,$8F,$F7,$6F
	.dc.b	$CD,$3E,$6F,$FD,$16,$D8,$FF,$4B,$B5,$FF,$4E,$35,$E3,$EF,$70,$23
	.dc.b	$C1,$FF,$4A,$FC,$9F,$F5,$48,$FF,$50,$B5,$FF,$53,$35,$FF,$7F,$B5
	.dc.b	$FD,$C7,$C9,$BE,$07,$FC,$91,$F3,$3F,$E1,$5E,$0F,$F1,$57,$8E,$8C
	.dc.b	$5F,$F8,$B6,$F8,$3F,$EE,$47,$FD,$22,$4B,$FF,$54,$B6,$35,$1F,$E5
	.dc.b	$38,$E3,$1F,$EF,$19,$FE,$56,$F8,$2F,$C3,$F0,$BF,$C0,$E3,$07,$F9
	.dc.b	$81,$C3,$FE,$F9,$6E,$1F,$D3,$DF,$E0,$51,$C2,$11,$A9,$EB,$E2,$FF
	.dc.b	$99,$E3,$15,$D7,$C4,$FF,$35,$7C,$CF,$F2,$A8,$E4,$FF,$99,$DF,$F3
	.dc.b	$8D,$FC,$68,$C6,$1F,$D7,$DE,$13,$A3,$FC,$EE,$E2,$35,$EE,$76,$3B
	.dc.b	$F9,$3F,$E7,$7F,$99,$7C,$67,$F9,$EF,$E6,$5F,$19,$4D,$44,$CB,$FD
	.dc.b	$C3,$5F,$ED,$A2,$47,$FB,$69,$E7,$1F,$28,$38,$7C,$C6,$23,$C1,$E2
	.dc.b	$FF,$A2,$FE,$67,$E2,$E6,$69,$36,$D7,$04,$C3,$E8,$4E,$5F,$A5,$F3
	.dc.b	$3F,$07,$83,$F1,$AF,$C8,$6E,$FF,$E9,$98,$D4,$FF,$A7,$9B,$89,$E1
	.dc.b	$19,$9E,$1B,$FF,$A7,$F9,$5F,$2F,$F8,$FB,$97,$F0,$67,$2D,$9F,$D5
	.dc.b	$91,$FD,$5B,$5F,$8B,$1C,$1B,$1F,$9B,$D7,$F1,$5E,$14,$9F,$C8,$72
	.dc.b	$7F,$1D,$6F,$66,$5F,$CC,$4D,$D3,$F3,$13,$F9,$0C,$3F,$23,$F0,$32
	.dc.b	$FF,$23,$19,$7F,$27,$19,$7F,$1D,$4E,$6F,$32,$DF,$FB,$B6,$8F,$98
	.dc.b	$FD,$97,$CC,$CC,$3E,$24,$7C,$93,$F8,$69,$1F,$E1,$AD,$7E,$1B,$B7
	.dc.b	$F8,$17,$C2,$7E,$1F,$59,$1F,$82,$7F,$15,$AE,$23,$CF,$C7,$F1,$5B
	.dc.b	$3A,$7F,$0F,$B7,$E0,$F6,$91,$82,$1F,$03,$85,$1A,$8C,$4F,$FB,$87
	.dc.b	$E3,$9F,$33,$FC,$47,$E6,$25,$7E,$8E,$72,$D9,$CB,$FA,$4E,$E5,$73
	.dc.b	$99,$3F,$C8,$63,$7C,$5F,$3E,$3F,$DA,$57,$D1,$FD,$6B,$63,$50,$BE
	.dc.b	$AF,$33,$C4,$43,$C5,$5D,$B8,$BF,$C6,$17,$CD,$B8,$66,$BF,$9B,$57
	.dc.b	$99,$F0,$C6,$5F,$04,$67,$0F,$1D,$F3,$9E,$27,$8C,$E2,$78,$E3,$D3
	.dc.b	$87,$A2,$3C,$A6,$70,$F1,$BC,$67,$77,$DE,$73,$3C,$09,$3E,$2C,$35
	.dc.b	$1E,$23,$33,$C2,$75,$1D,$63,$FA,$DD,$70,$86,$1E,$B7,$DE,$6C,$FA
	.dc.b	$B7,$BC,$3D,$58,$79,$A7,$D4,$68,$E0,$72,$3E,$19,$1E,$0F,$98,$E0
	.dc.b	$67,$0F,$2B,$CA,$FC,$87,$96,$BC,$33,$A9,$E0,$72,$AE,$A4,$8F,$84
	.dc.b	$48,$F8,$D1,$97,$84,$65,$F4,$46,$5F,$8B,$7F,$07,$0F,$C2,$61,$E4
	.dc.b	$8F,$C3,$61,$E4,$AE,$67,$B0,$FF,$40,$1C,$8A,$64,$FB,$A6,$A7,$F4
	.dc.b	$4D,$9D,$BB,$21,$CC,$F8,$25,$BF,$C7,$F6,$7D,$DA,$F3,$4F,$F8,$46
	.dc.b	$31,$5F,$6D,$B1,$F6,$8F,$D7,$38,$7A,$DF,$F6,$C7,$55,$D1,$FE,$01
	.dc.b	$3B,$BF,$C7,$C6,$0D,$DF,$76,$9D,$EF,$A3,$EA,$BE,$95,$EA,$8D,$9E
	.dc.b	$2B,$ED,$8F,$D7,$BE,$35,$E0,$24,$66,$77,$5F,$50,$F6,$9D,$27,$6F
	.dc.b	$F8,$DE,$36,$FC,$B9,$FF,$20,$C6,$2F,$F2,$4D,$C3,$FC,$6C,$DA,$F9
	.dc.b	$C7,$5A,$F1,$34,$F1,$61,$E8,$8F,$C8,$5F,$C0,$7B,$CE,$CF,$11,$E5
	.dc.b	$31,$1E,$27,$FC,$71,$F3,$3F,$E1,$5C,$7F,$96,$07,$B4,$23,$18,$7F
	.dc.b	$B4,$DD,$CF,$1B,$FF,$2D,$B6,$B8,$3C,$A7,$8C,$78,$3B,$39,$57,$0F
	.dc.b	$46,$CF,$06,$D7,$8D,$B7,$FE,$5A,$9F,$EE,$6F,$FA,$DD,$F5,$C6,$8D
	.dc.b	$F4,$FF,$55,$D7,$FC,$C9,$EE,$CA,$FE,$A9,$19,$7F,$4F,$A7,$F2,$F7
	.dc.b	$F5,$91,$FA,$A7,$F8,$FB,$63,$F7,$11,$83,$99,$FE,$77,$F9,$9F,$4F
	.dc.b	$CB,$BE,$62,$FC,$CC,$AF,$F3,$44,$E5,$B3,$97,$FC,$1F,$09,$E7,$3D
	.dc.b	$39,$A5,$72,$7B,$F4,$62,$38,$C5,$DF,$EE,$E1,$FE,$AC,$7F,$A9,$ED
	.dc.b	$FE,$87,$F5,$25,$FF,$77,$8F,$EF,$87,$BC,$8F,$C6,$6B,$FD,$C1,$DB
	.dc.b	$FE,$97,$3E,$0F,$7F,$C2,$F2,$24,$7C,$08,$CB,$F0,$A3,$2F,$C0,$8F
	.dc.b	$C0,$FD,$2F,$FB,$A6,$3F,$D1,$CD,$7F,$C6,$6D,$C3,$F3,$2B,$E4,$78
	.dc.b	$FA,$F2,$B8,$BD,$23,$F1,$F0,$C3,$97,$37,$FA,$D2,$FF,$CC,$38,$1F
	.dc.b	$F3,$1D,$6F,$0E,$37,$E3,$8F,$E6,$ED,$44,$70,$D8,$FC,$64,$78,$23
	.dc.b	$07,$FA,$25,$C5,$78,$7E,$D7,$E1,$FB,$DF,$A4,$F0,$BF,$A2,$CF,$EE
	.dc.b	$1B,$97,$A6,$77,$87,$FA,$D6,$72,$FE,$ED,$B5,$ED,$9E,$9D,$52,$B8
	.dc.b	$3D,$FA,$F1,$1F,$E6,$2B,$FD,$F5,$FF,$97,$EB,$53,$FE,$A1,$FA,$36
	.dc.b	$1E,$7F,$A1,$93,$98,$FA,$33,$A8,$FE,$AE,$D7,$FA,$FB,$5F,$EC,$6D
	.dc.b	$7E,$68,$78,$5F,$F9,$01,$B9,$69,$B3,$F0,$7C,$2F,$C0,$FC,$EF,$FA
	.dc.b	$D6,$3F,$B7,$B5,$EA,$D7,$ED,$B5,$3F,$D4,$9E,$3A,$7F,$07,$8C,$47
	.dc.b	$28,$FE,$15,$F3,$6F,$62,$92,$7F,$BB,$C6,$5B,$C9,$C0,$8E,$0F,$E2
	.dc.b	$FE,$97,$F1,$32,$38,$3F,$C0,$3F,$2C,$F8,$0F,$E3,$4D,$8C,$C6,$0F
	.dc.b	$2F,$CD,$C8,$F1,$B5,$FF,$08,$E0,$F9,$BF,$59,$23,$D2,$F6,$E5,$91
	.dc.b	$E2,$6B,$FE,$1A,$C1,$F3,$7E,$42,$47,$A1,$F3,$FC,$DC,$8F,$5B,$5E
	.dc.b	$5F,$6D,$EC,$23,$F0,$1F,$7F,$54,$8F,$86,$D7,$C5,$D9,$F1,$08,$FC
	.dc.b	$87,$E3,$FD,$5C,$8F,$96,$D7,$D5,$F3,$3F,$C2,$B2,$7C,$A5,$31,$12
	.dc.b	$7E,$9B,$F4,$88,$E0,$FE,$0F,$D3,$FE,$98,$A4,$F1,$9F,$EE,$5D,$3F
	.dc.b	$A0,$FD,$C9,$F0,$FE,$1F,$67,$FB,$3F,$CE,$3E,$76,$99,$C3,$AE,$1F
	.dc.b	$A8,$C9,$7F,$84,$24,$FE,$21,$AF,$80,$FE,$37,$1F,$C8,$8F,$E4,$B6
	.dc.b	$F6,$3E,$17,$AD,$E2,$F6,$25,$F8,$0E,$31,$A7,$F2,$AF,$E5,$08,$FF
	.dc.b	$96,$3E,$77,$E5,$BC,$43,$87,$C8,$63,$83,$F9,$DC,$7C,$88,$E0,$EF
	.dc.b	$83,$FD,$86,$DD,$3F,$84,$B6,$7F,$9B,$BC,$5E,$BF,$35,$F3,$9F,$13
	.dc.b	$5C,$EF,$FD,$2B,$E7,$3E,$BE,$1F,$D0,$3E,$69,$7F,$C5,$8E,$11,$FE
	.dc.b	$D4,$6E,$39,$BE,$5C,$7F,$E2,$FC,$58,$72,$7F,$A5,$E3,$D9,$CE,$1C
	.dc.b	$5F,$03,$5C,$9C,$2B,$A5,$73,$85,$73,$8C,$57,$B5,$B7,$FC,$B9,$E1
	.dc.b	$3B,$9F,$E6,$2D,$AF,$EB,$F1,$C4,$F1,$4F,$0B,$F0,$C7,$06,$CF,$0B
	.dc.b	$ED,$3C,$0D,$63,$EC,$9F,$53,$5F,$9B,$3C,$1E,$46,$70,$F2,$3E,$EB
	.dc.b	$F9,$8F,$FA,$B2,$3E,$53,$63,$2F,$80,$FF,$57,$C3,$E7,$1F,$3B,$FE
	.dc.b	$AC,$8C,$36,$33,$3C,$1A,$FF,$AC,$BE,$77,$FC,$94,$9A,$47,$FC,$9C
	.dc.b	$FB,$71,$FE,$17,$A7,$BE,$7F,$C1,$73,$9D,$93,$99,$E8,$7E,$00,$FF
	.dc.b	$3C,$46,$9C,$1C,$8D,$57,$67,$27,$2D,$7F,$0D,$19,$7F,$11,$1C,$5C
	.dc.b	$CF,$57,$31,$C7,$5F,$F4,$34,$E5,$FF,$6C,$E1,$E9,$C3,$D0,$DA,$F0
	.dc.b	$B6,$FA,$CF,$F0,$1C,$3C,$C7,$AE,$BF,$C0,$35,$C6,$30,$6A,$7F,$D2
	.dc.b	$DF,$3B,$FE,$9C,$7C,$67,$37,$15,$FE,$36,$F9,$DF,$E5,$73,$FC,$3D
	.dc.b	$B1,$FC,$39,$FF,$1D,$C6,$77,$E3,$29,$DD,$FF,$AA,$BD,$CF,$14,$F5
	.dc.b	$AE,$F2,$9C,$CB,$E7,$5E,$EE,$BE,$75,$F0,$3D,$A7,$39,$FF,$2B,$C5
	.dc.b	$5A,$8F,$E4,$D3,$AC,$3F,$9C,$56,$AF,$96,$1D,$5F,$4F,$CE,$BE,$1F
	.dc.b	$E6,$5F,$F8,$B2,$31,$8F,$F8,$BA,$31,$3A,$7F,$CB,$B3,$A7,$41,$C6
	.dc.b	$1F,$EC,$CF,$9C,$7C,$EF,$82,$FE,$1C,$63,$0F,$E7,$D7,$C1,$FE,$C7
	.dc.b	$19,$C3,$F9,$FC,$E5,$FF,$6A,$46,$70,$FC,$27,$CE,$BE,$33,$CD,$48
	.dc.b	$C6,$1E,$68,$C4,$EA,$73,$1E,$AF,$9D,$FF,$2C,$D6,$DB,$19,$F5,$FA
	.dc.b	$C6,$23,$E3,$7A,$1F,$EB,$7A,$EC,$D8,$F8,$AD,$77,$F0,$7E,$6F,$9D
	.dc.b	$FF,$22,$57,$95,$B3,$F3,$1A,$FD,$36,$BF,$E6,$2F,$9D,$FF,$1F,$4F
	.dc.b	$B5,$B1,$F5,$1A,$FD,$B6,$BF,$E6,$AF,$B0,$7A,$BF,$21,$B5,$F1,$DB
	.dc.b	$3F,$79,$AF,$E1,$7E,$07,$F3,$93,$F3,$5B,$1F,$84,$6B,$BE,$8F,$AB
	.dc.b	$F3,$CF,$42,$92,$FE,$B4,$65,$FB,$1F,$61,$FD,$B1,$BE,$0F,$DE,$8C
	.dc.b	$61,$F9,$26,$BF,$99,$F9,$FF,$DD,$B7,$BB,$F0,$B1,$88,$FF,$45,$B5
	.dc.b	$FF,$5C,$FC,$FB,$E2,$3F,$16,$DC,$7F,$14,$D9,$FD,$0B,$5F,$D4,$7D
	.dc.b	$67,$E9,$5B,$D9,$F9,$48,$C4,$7F,$A6,$9A,$FF,$B2,$3E,$73,$FD,$0D
	.dc.b	$F3,$DF,$97,$6C,$7E,$B9,$7C,$F3,$E2,$F5,$F0,$4B,$FC,$F7,$CF,$3E
	.dc.b	$3B,$F4,$18,$A3,$0E,$A3,$31,$88,$FD,$DE,$0F,$6F,$8B,$63,$F5,$4D
	.dc.b	$77,$1D,$BE,$77,$FD,$D9,$1F,$AF,$6B,$FC,$19,$AF,$F0,$D6,$BF,$EB
	.dc.b	$CF,$98,$FE,$E0,$FE,$DD,$AF,$F0,$E6,$BF,$C5,$5A,$FF,$B1,$3E,$59
	.dc.b	$F4,$1F,$01,$B1,$FC,$05,$AF,$F1,$86,$BC,$FE,$63,$FD,$9B,$E3,$FF
	.dc.b	$4F,$97,$C6,$63,$97,$B3,$6F,$FD,$FF,$8B,$0C,$42,$30,$FA,$E3,$E9
	.dc.b	$C6,$23,$D9,$18,$38,$C6,$70,$FE,$C0,$DB,$F7,$76,$F8,$65,$F1,$5F
	.dc.b	$97,$C3,$FA,$33,$5F,$E8,$6D,$8F,$F0,$21,$C3,$83,$94,$66,$FF,$D2
	.dc.b	$6D,$A5,$76,$E8,$DD,$F8,$7E,$6B,$C9,$5A,$5E,$E6,$4D,$69,$27,$FA
	.dc.b	$D4,$E9,$B3,$97,$FC,$FB,$A3,$E4,$7E,$5E,$FF,$E8,$F9,$FF,$42,$35
	.dc.b	$DC,$7C,$83,$E6,$91,$FC,$53,$5D,$9F,$E8,$D9,$EC,$F4,$F7,$BE,$67
	.dc.b	$CD,$7C,$1D,$4D,$26,$9A,$D6,$5F,$F7,$59,$CB,$FD,$DF,$87,$E2,$DF
	.dc.b	$E2,$B4,$FB,$11,$F0,$67,$FB,$BB,$A2,$46,$8D,$ED,$F8,$CF,$98,$FF
	.dc.b	$02,$D7,$C2,$F9,$9F,$8A,$67,$D4,$91,$FD,$7E,$76,$FF,$76,$6B,$FD
	.dc.b	$F5,$A8,$8F,$0C,$7F,$C6,$0D,$E3,$D4,$A3,$0F,$1E,$17,$C8,$7E,$57
	.dc.b	$E6,$9F,$2B,$62,$57,$F8,$6E,$72,$D9,$CB,$FE,$1A,$8E,$6E,$4E,$1E
	.dc.b	$57,$52,$3C,$FA,$2B,$C6,$2E,$7F,$A0,$AF,$F1,$94,$FF,$B3,$72,$6B
	.dc.b	$F8,$EF,$62,$4E,$63,$E8,$1F,$CA,$A7,$71,$F3,$63,$04,$78,$10,$FF
	.dc.b	$6B,$E8,$F9,$AF,$DC,$7C,$CB,$E6,$AF,$A7,$2F,$FC,$7D,$39,$6C,$E5
	.dc.b	$FF,$1E,$C2,$79,$CF,$4E,$79,$5C,$9E,$FD,$18,$8E,$31,$76,$FE,$07
	.dc.b	$87,$F4,$17,$EF,$EF,$FB,$0F,$16,$5F,$C7,$8F,$C9,$18,$7D,$B1,$87
	.dc.b	$47,$ED,$9D,$4F,$48,$C9,$C0,$D1,$87,$4E,$48,$E0,$9C,$9F,$CF,$E7
	.dc.b	$AB,$F9,$4D,$D7,$E3,$BE,$33,$B7,$B3,$B8,$E3,$1F,$99,$AE,$17,$DA
	.dc.b	$30,$79,$8F,$E4,$75,$C5,$E6,$68,$EE,$7E,$AA,$B9,$C7,$86,$32,$7E
	.dc.b	$A6,$32,$F1,$9F,$EA,$15,$88,$D1,$E1,$8C,$BE,$03,$27,$4D,$BF,$C0
	.dc.b	$8E,$70,$F8,$6F,$1C,$3C,$55,$ED,$8E,$2F,$47,$CB,$3C,$1F,$79,$EB
	.dc.b	$39,$1E,$79,$F8,$4E,$8E,$6F,$62,$3C,$4D,$47,$59,$F0,$12,$79,$BE
	.dc.b	$5A,$F5,$D7,$93,$0F,$89,$18,$9E,$13,$D5,$E1,$87,$DC,$8F,$19,$9B
	.dc.b	$F1,$AA,$E5,$1E,$02,$47,$B4,$87,$23,$E7,$9F,$CA,$6F,$F8,$21,$1F
	.dc.b	$E6,$0D,$7F,$98,$35,$E0,$7D,$57,$FA,$EB,$F7,$F0,$F8,$93,$E2,$8F
	.dc.b	$31,$99,$F8,$2D,$3F,$05,$5F,$EB,$7C,$3E,$E5,$F2,$9F,$80,$7F,$35
	.dc.b	$38,$1F,$4A,$BC,$2F,$8C,$E6,$6A,$7E,$93,$6F,$C9,$7D,$DF,$35,$F7
	.dc.b	$9F,$8D,$39,$7C,$27,$C9,$8D,$46,$63,$17,$EC,$7C,$47,$94,$FE,$FE
	.dc.b	$78,$4E,$71,$C6,$7C,$04,$9F,$5C,$F1,$25,$FA,$9E,$AD,$73,$7E,$A9
	.dc.b	$E3,$3F,$29,$F2,$4E,$8F,$5D,$7A,$9F,$21,$F0,$0A,$E5,$EF,$5F,$21
	.dc.b	$CC,$F6,$8F,$EE,$14,$C7,$FC,$77,$5F,$D0,$BE,$67,$FA,$39,$FD,$23
	.dc.b	$BE,$46,$48,$7F,$44,$7F,$C8,$9F,$35,$FE,$AD,$8E,$B8,$76,$8C,$47
	.dc.b	$EF,$30,$46,$1F,$F2,$2A,$F9,$0E,$7E,$94,$AE,$67,$F9,$12,$78,$5F
	.dc.b	$5A,$FF,$30,$E3,$8A,$D7,$C7,$79,$0C,$E1,$23,$FD,$E9,$3E,$47,$FC
	.dc.b	$1A,$1C,$3F,$E1,$06,$BF,$E1,$77,$B3,$B8,$E8,$FB,$70,$7C,$A7,$C4
	.dc.b	$F6,$FF,$CD,$1E,$27,$E5,$74,$FF,$20,$DF,$B1,$F8,$73,$B8,$C4,$EC
	.dc.b	$E1,$69,$E7,$D9,$2B,$6F,$6E,$DC,$47,$CF,$BF,$F0,$FB,$67,$4F,$CC
	.dc.b	$1F,$33,$D5,$F0,$FC,$D7,$A0,$FF,$46,$C6,$63,$F9,$83,$5F,$92,$D7
	.dc.b	$FC,$6E,$D7,$FC,$7C,$D7,$BE,$1D,$08,$FC,$0F,$0F,$FD,$1D,$F5,$9F
	.dc.b	$E0,$F8,$F8,$CF,$FB,$49,$BE,$66,$BB,$64,$60,$87,$E3,$E7,$93,$7B
	.dc.b	$3A,$91,$EB,$F5,$55,$D1,$E0,$8F,$1C,$7F,$AB,$BC,$B7,$B1,$E6,$E9
	.dc.b	$97,$F2,$E7,$2D,$9F,$F7,$1C,$62,$3F,$1E,$7F,$B2,$23,$24,$78,$10
	.dc.b	$FD,$47,$D6,$7F,$BB,$DF,$C4,$3F,$A7,$6F,$C6,$7F,$AD,$60,$DE,$A9
	.dc.b	$3A,$21,$D1,$E1,$E5,$48,$EA,$F3,$F8,$0C,$47,$8A,$2F,$47,$FA,$F3
	.dc.b	$05,$7F,$B4,$BE,$B3,$FD,$D3,$39,$E3,$F2,$19,$F8,$E9,$8F,$C2,$15
	.dc.b	$FE,$EC,$7F,$D0,$8D,$73,$1B,$69,$F9,$98,$EF,$DF,$FE,$86,$6C,$7F
	.dc.b	$B2,$BE,$53,$FD,$0E,$BE,$7B,$FC,$FA,$FD,$26,$BF,$C0,$9A,$FF,$A4
	.dc.b	$1A,$FF,$A5,$FE,$75,$EC,$F8,$DB,$1E,$B6,$BF,$79,$AF,$F0,$E5,$CB
	.dc.b	$C4,$91,$C5,$FF,$40,$7C,$1C,$8F,$F5,$63,$5F,$C4,$FC,$BF,$D0,$7E
	.dc.b	$53,$5C,$EF,$F1,$E3,$11,$FC,$7B,$E9,$9F,$33,$F2,$DB,$1F,$4D,$AF
	.dc.b	$E5,$9A,$FA,$7E,$7B,$E0,$11,$F8,$0F,$7F,$16,$47,$91,$AF,$A3,$8B
	.dc.b	$C8,$47,$D2,$F9,$FE,$76,$47,$B1,$AF,$C0,$F9,$DF,$61,$1F,$82,$FE
	.dc.b	$17,$E7,$7F,$0D,$06,$91,$F8,$9D,$31,$F9,$2D,$7E,$6F,$BF,$F2,$48
	.dc.b	$FC,$E7,$B7,$A3,$23,$F3,$55,$FE,$48,$7F,$A5,$B5,$FF,$6C,$EC,$F9
	.dc.b	$BF,$A8,$48,$FB,$2D,$7F,$CB,$91,$B7,$F9,$B3,$5F,$F3,$04,$BE,$23
	.dc.b	$FD,$CB,$D6,$EA,$FF,$73,$CE,$6B,$CB,$5F,$14,$8F,$E2,$3C,$FF,$C3
	.dc.b	$11,$C1,$0F,$1B,$DA,$BD,$6D,$8F,$D7,$7C,$ED,$58,$A7,$0D,$8F,$5C
	.dc.b	$71,$C5,$D6,$BD,$F7,$FE,$64,$97,$F9,$E6,$BF,$19,$DB,$FD,$AE,$5E
	.dc.b	$57,$8F,$E7,$7F,$91,$CF,$F5,$F6,$BF,$E8,$96,$E1,$F2,$60,$23,$E8
	.dc.b	$46,$63,$1E,$17,$52,$3D,$77,$F8,$5E,$1F,$D1,$8F,$F4,$B3,$6B,$93
	.dc.b	$76,$EF,$2F,$9D,$FC,$35,$78,$BE,$73,$FA,$45,$EE,$BC,$AD,$7F,$A9
	.dc.b	$F0,$FD,$1B,$E3,$0A,$FE,$A1,$18,$79,$CF,$81,$B3,$DE,$BC,$75,$B6
	.dc.b	$E2,$C3,$55,$E2,$C3,$95,$7A,$8F,$81,$87,$82,$F8,$57,$8E,$79,$DF
	.dc.b	$28,$F4,$96,$38,$9A,$78,$1F,$53,$0F,$23,$E1,$38,$61,$E2,$75,$7E
	.dc.b	$0C,$3C,$58,$6A,$61,$D0,$FA,$13,$BC,$3C,$11,$FE,$44,$C3,$E5,$B6
	.dc.b	$7D,$6D,$AF,$EB,$52,$F1,$1F,$3B,$86,$D7,$F8,$02,$75,$39,$3F,$8B
	.dc.b	$FC,$EC,$9F,$EB,$CD,$7F,$B3,$3F,$1E,$FF,$BE,$E9,$F2,$67,$F2,$F3
	.dc.b	$9D,$7F,$35,$A7,$F8,$3E,$72,$7B,$EF,$8B,$F0,$4C,$C7,$F8,$56,$32
	.dc.b	$F1,$85,$75,$BF,$04,$60,$E9,$7E,$0D,$3F,$C3,$B5,$B9,$EC,$6E,$FB
	.dc.b	$56,$E3,$FC,$E3,$3B,$9F,$03,$6B,$F8,$ED,$7E,$D7,$0F,$D9,$1E,$A3
	.dc.b	$F6,$97,$EA,$9D,$E1,$FE,$5D,$9D,$3F,$C9,$BE,$77,$93,$AA,$FC,$E3
	.dc.b	$93,$FC,$C9,$F3,$BF,$CB,$5B,$BF,$09,$FE,$0B,$3F,$C1,$A7,$24,$8F
	.dc.b	$F4,$F7,$C1,$78,$1D,$D7,$1C,$13,$99,$7C,$E3,$D6,$ED,$18,$78,$7C
	.dc.b	$EB,$E0,$73,$5B,$CB,$FC,$F4,$DB,$FF,$26,$B5,$FF,$4E,$B5,$FE,$1E
	.dc.b	$D7,$F8,$FF,$CE,$3A,$7C,$2D,$AF,$F4,$B7,$CE,$3E,$0F,$35,$27,$FD
	.dc.b	$15,$5B,$BF,$F5,$F5,$EE,$73,$5B,$9F,$F3,$0D,$6D,$FF,$58,$B5,$FE
	.dc.b	$46,$D7,$F9,$9F,$CD,$BE,$13,$EF,$C6,$1A,$A3,$57,$FE,$4D,$6C,$7F
	.dc.b	$9C,$1F,$F1,$8C,$67,$CE,$E3,$18,$9D,$46,$3D,$69,$5C,$DB,$3F,$E7
	.dc.b	$AF,$9A,$7D,$47,$89,$B5,$FE,$5D,$6C,$EE,$75,$F4,$B2,$B4,$D9,$E9
	.dc.b	$F4,$5C,$C9,$3C,$DF,$7F,$C3,$3D,$4A,$49,$FF,$36,$46,$5F,$F3,$77
	.dc.b	$E1,$EF,$8D,$F8,$ED,$8C,$C6,$0C,$CC,$8E,$2D,$C3,$FD,$33,$F5,$8F
	.dc.b	$8F,$FA,$2D,$AF,$F3,$E3,$67,$73,$F5,$BE,$A1,$EF,$D2,$5F,$FA,$0E
	.dc.b	$32,$FF,$A1,$7E,$63,$FC,$DF,$F3,$17,$DE,$FF,$51,$63,$FE,$F8,$97
	.dc.b	$C2,$FF,$8B,$E1,$1C,$1F,$F5,$7F,$E2,$9C,$1F,$F7,$FB,$5F,$EE,$2D
	.dc.b	$7F,$C0,$2D,$4B,$E9,$7F,$D2,$4D,$8F,$F4,$D3,$5E,$F1,$AF,$6A,$46
	.dc.b	$9A,$FF,$B0,$BE,$B7,$E1,$BF,$EA,$36,$B9,$D7,$FD,$41,$18,$8F,$F0
	.dc.b	$38,$5F,$5D,$A2,$47,$37,$8F,$D7,$C8,$EC,$D7,$FD,$A7,$F5,$EF,$A0
	.dc.b	$FF,$1A,$7D,$6F,$E1,$E7,$FC,$47,$E5,$7F,$31,$8F,$F5,$5D,$6E,$73
	.dc.b	$C5,$FE,$1A,$8C,$47,$4B,$FF,$5C,$46,$23,$D2,$D7,$84,$7F,$92,$63
	.dc.b	$8D,$FF,$8B,$DB,$5E,$2B,$FD,$05,$F8,$0F,$F0,$A7,$C5,$BB,$FF,$C9
	.dc.b	$8B,$EA,$A9,$88,$93,$C7,$15,$6E,$30,$D7,$FA,$29,$DA,$7F,$CB,$1F
	.dc.b	$3F,$FE,$AF,$AF,$EA,$8D,$C7,$E3,$B6,$BF,$C5,$0B,$E1,$78,$FC,$4C
	.dc.b	$9F,$14,$EE,$CE,$E3,$DC,$3E,$B1,$F3,$7F,$E6,$9F,$AD,$B3,$EB,$6B
	.dc.b	$E8,$38,$47,$F9,$D6,$73,$1F,$8A,$6B,$FE,$52,$6B,$FE,$62,$6B,$C8
	.dc.b	$F1,$BD,$AF,$FC,$C8,$DC,$BF,$0F,$7F,$5A,$BE,$4D,$3E,$7F,$EB,$47
	.dc.b	$D1,$7E,$74,$7F,$7D,$6D,$7E,$41,$DC,$F2,$6C,$EF,$6F,$F4,$54,$70
	.dc.b	$9F,$F2,$B3,$AA,$FC,$CF,$CA,$7C,$D2,$69,$FE,$91,$7E,$75,$79,$9C
	.dc.b	$BF,$85,$78,$6B,$FE,$9D,$FA,$6F,$D2,$4F,$B3,$87,$FB,$A3,$F8,$B6
	.dc.b	$BF,$46,$37,$9F,$FA,$25,$E3,$1F,$BA,$6B,$FE,$82,$6B,$FE,$92,$6A
	.dc.b	$79,$E3,$BF,$0F,$C8,$7E,$E3,$1F,$D0,$F9,$BF,$A2,$FA,$A9,$87,$E7
	.dc.b	$23,$2F,$E7,$A3,$2F,$D6,$E4,$4F,$E7,$BE,$32,$4F,$E5,$9F,$D0,$FC
	.dc.b	$5E,$23,$EE,$F2,$A7,$1A,$FE,$92,$BF,$21,$F2,$1F,$97,$FA,$F9,$87
	.dc.b	$E2,$A3,$2F,$F1,$C3,$F2,$0D,$69,$79,$91,$98,$FD,$23,$C9,$FD,$04
	.dc.b	$69,$FD,$26,$BF,$86,$CF,$F0,$FA,$C9,$FC,$94,$65,$FC,$A4,$65,$FC
	.dc.b	$84,$7E,$1F,$39,$1F,$85,$7F,$15,$96,$23,$F1,$51,$73,$FF,$2C,$60
	.dc.b	$AF,$F7,$37,$C7,$3E,$5F,$FD,$D1,$F3,$0F,$94,$A6,$A2,$65,$FC,$0A
	.dc.b	$74,$D9,$CB,$FB,$FA,$E8,$E4,$86,$6B,$78,$FF,$B6,$EF,$9C,$FF,$07
	.dc.b	$6C,$7F,$B7,$EB,$9C,$F1,$39,$C7,$1B,$E7,$C7,$FE,$DA,$AD,$CE,$6B
	.dc.b	$73,$9C,$37,$1F,$EE,$EC,$3B,$CF,$FB,$89,$B1,$FE,$F2,$BE,$73,$C6
	.dc.b	$39,$9C,$70,$E6,$A7,$6E,$4D,$99,$BD,$DF,$68,$36,$78,$C6,$CE,$31
	.dc.b	$99,$E3,$5F,$5A,$7F,$9A,$D7,$03,$8D,$75,$38,$46,$AB,$83,$F3,$EB
	.dc.b	$D3,$5C,$1C,$E1,$F4,$6B,$BD,$7A,$AB,$B4,$6B,$0F,$91,$7C,$A3,$15
	.dc.b	$F0,$4F,$B0,$4B,$FA,$8F,$84,$F2,$10,$C1,$1F,$13,$97,$89,$2B,$EA
	.dc.b	$47,$57,$C6,$F9,$E7,$D2,$F3,$8D,$CF,$48,$C1,$E7,$70,$65,$FF,$0F
	.dc.b	$DF,$E4,$9B,$5F,$C0,$63,$11,$E2,$32,$7F,$09,$72,$FE,$2E,$38,$1F
	.dc.b	$1D,$FD,$67,$CD,$57,$D4,$7E,$55,$C5,$79,$2F,$F2,$D1,$87,$D6,$62
	.dc.b	$7E,$33,$87,$C0,$76,$32,$60,$EE,$7E,$4A,$3A,$39,$9F,$DB,$46,$23
	.dc.b	$F2,$87,$ED,$DC,$5F,$E4,$1B,$7F,$C3,$1B,$3E,$13,$BC,$FE,$29,$C5
	.dc.b	$7B,$5F,$05,$FA,$5F,$21,$E8,$3C,$31,$B7,$27,$4B,$FD,$6D,$F8,$23
	.dc.b	$CC,$71,$BF,$03,$83,$81,$D7,$14,$F1,$BF,$83,$3C,$6F,$AB,$4E,$13
	.dc.b	$BC,$3E,$94,$74,$3E,$6D,$F4,$78,$18,$9F,$F2,$3F,$D3,$FF,$92,$3B
	.dc.b	$E4,$FF,$90,$A3,$2F,$F9,$12,$32,$FF,$67,$8F,$F2,$1C,$C8,$FF,$1D
	.dc.b	$BF,$E4,$6C,$F1,$1F,$D9,$A2,$E5,$FE,$48,$9F,$E8,$7F,$5B,$FD,$0B
	.dc.b	$AA,$5F,$F0,$C8,$CB,$F9,$B7,$F4,$04,$70,$6C,$87,$F3,$BE,$BF,$F1
	.dc.b	$2E,$3F,$E2,$6C,$A4,$FF,$8C,$63,$2F,$F8,$CE,$32,$FF,$8B,$E3,$FC
	.dc.b	$4D,$94,$8F,$F1,$23,$FE,$29,$C7,$11,$FE,$29,$8B,$97,$F6,$09,$FE
	.dc.b	$BF,$C7,$FD,$7B,$E4,$65,$FF,$28,$8C,$BF,$B7,$7F,$78,$47,$06,$C8
	.dc.b	$7F,$57,$F8,$C3,$FE,$7B,$FA,$4F,$F2,$92,$C1,$18,$6F,$85,$C2,$7F
	.dc.b	$CA,$B2,$EA,$C5,$38,$BF,$D4,$33,$C4,$97,$A8,$CD,$5B,$F0,$57,$F9
	.dc.b	$86,$F9,$46,$39,$B5,$1E,$46,$CE,$EF,$BF,$AB,$89,$97,$ED,$8F,$F0
	.dc.b	$4F,$CB,$FF,$82,$3E,$2A,$4F,$F5,$28,$CB,$FC,$85,$FE,$4C,$47,$06
	.dc.b	$C8,$7F,$80,$7A,$3C,$BF,$43,$FE,$EC,$8F,$F5,$33,$5F,$F3,$DB,$73
	.dc.b	$F6,$85,$E9,$72,$F4,$A5,$72,$7A,$C7,$6F,$27,$BF,$91,$2B,$BC,$65
	.dc.b	$F0,$C6,$5F,$0C,$7C,$F5,$F7,$7F,$E0,$17,$FC,$4D,$E8,$7F,$8B,$0D
	.dc.b	$46,$63,$FD,$48,$D7,$FD,$D6,$D7,$F0,$8D,$7F,$1A,$D7,$BE,$7D,$08
	.dc.b	$FF,$8D,$7D,$1F,$27,$CD,$FF,$AA,$09,$CB,$FE,$57,$71,$CB,$49,$5F
	.dc.b	$E5,$F8,$FC,$93,$7C,$6E,$04,$70,$7F,$99,$7A,$9F,$45,$FE,$63,$6C
	.dc.b	$66,$30,$67,$EB,$E4,$71,$6B,$FE,$71,$97,$27,$32,$47,$77,$FC,$E1
	.dc.b	$F3,$AF,$A7,$A4,$9F,$F3,$C4,$65,$FF,$3D,$6E,$F9,$6F,$F3,$C3,$63
	.dc.b	$31,$83,$3F,$3D,$23,$8B,$5F,$F4,$37,$CE,$BE,$9F,$FD,$04,$D8,$FF
	.dc.b	$44,$35,$DC,$6B,$E5,$64,$69,$AF,$4F,$A3,$E6,$47,$9B,$DB,$E4,$A4
	.dc.b	$76,$6B,$E4,$E4,$F1,$91,$F1,$BE,$6E,$49,$1E,$66,$BE,$CE,$3F,$59
	.dc.b	$1F,$5B,$EE,$F9,$EF,$D0,$67,$FE,$9D,$9F,$F7,$F3,$7B,$BF,$87,$C2
	.dc.b	$13,$C0,$E5,$1F,$19,$AF,$CC,$F9,$4F,$E2,$BF,$3C,$F8,$7A,$78,$22
	.dc.b	$4F,$F3,$0C,$3F,$80,$D6,$E3,$1E,$2E,$63,$11,$E6,$AE,$2F,$73,$EB
	.dc.b	$BF,$68,$FB,$7E,$2F,$89,$91,$D0,$FE,$7F,$92,$7E,$19,$A8,$FE,$81
	.dc.b	$7E,$1A,$CC,$71,$1F,$3A,$F1,$7F,$85,$36,$34,$D7,$99,$FD,$39,$FC
	.dc.b	$C3,$83,$9C,$72,$23,$CB,$97,$F8,$B7,$1B,$E5,$E9,$27,$C9,$F3,$BF
	.dc.b	$A5,$DB,$FA,$DF,$CE,$2F,$FA,$B1,$C6,$BD,$4D,$7C,$4D,$9F,$C0,$D7
	.dc.b	$39,$57,$37,$FB,$3C,$66,$31,$1D,$EF,$FB,$24,$62,$FB,$3A,$AE,$11
	.dc.b	$FD,$A5,$B5,$FD,$CA,$36,$7A,$2B,$9B,$BC,$3F,$BA,$4F,$92,$31,$87
	.dc.b	$C0,$78,$45,$3A,$57,$27,$D3,$5C,$B0,$F2,$36,$B8,$CF,$53,$89,$CC
	.dc.b	$61,$E8,$6C,$F1,$76,$71,$3E,$21,$C6,$38,$1F,$35,$FF,$05,$1C,$B0
	.dc.b	$AF,$99,$AE,$4C,$C3,$CB,$7C,$EE,$A3,$0D,$8C,$CF,$06,$BD,$3E,$77
	.dc.b	$AB,$FE,$22,$6A,$69,$1F,$65,$FE,$87,$7F,$89,$1C,$4C,$EB,$FE,$18
	.dc.b	$D3,$F0,$B1,$97,$F0,$F5,$C6,$3E,$31,$9D,$3F,$C5,$35,$C5,$F8,$78
	.dc.b	$7D,$1A,$E1,$7D,$2F,$CC,$78,$30,$F1,$B8,$8E,$F8,$73,$8F,$9F,$7C
	.dc.b	$E3,$11,$FE,$3B,$8C,$57,$26,$C7,$53,$F7,$17,$E9,$3F,$93,$BE,$93
	.dc.b	$F8,$E5,$7F,$87,$2B,$D6,$7F,$09,$97,$CE,$E6,$B7,$39,$8C,$43,$E7
	.dc.b	$7F,$80,$4F,$F1,$16,$C7,$F0,$97,$FC,$B3,$19,$DF,$8C,$6A,$FF,$C5
	.dc.b	$BC,$3E,$7A,$FF,$2C,$4E,$5F,$E6,$51,$9B,$FF,$31,$FC,$DB,$AF,$DB
	.dc.b	$18,$6A,$8D,$57,$E5,$9B,$1F,$9A,$6E,$1F,$E3,$EF,$9D,$FE,$E1,$5F
	.dc.b	$CA,$E7,$2F,$F4,$38,$CD,$7F,$25,$F7,$9F,$35,$F1,$1B,$5F,$CE,$F0
	.dc.b	$7C,$DB,$E7,$35,$72,$BF,$53,$7C,$30,$FD,$6B,$5F,$D8,$B5,$FF,$3E
	.dc.b	$3B,$C3,$FC,$AB,$82,$3A,$4F,$F3,$C9,$F0,$C7,$D8,$F9,$B7,$CD,$FC
	.dc.b	$66,$E1,$FB,$A6,$DF,$EF,$9A,$FF,$02,$5E,$8F,$12,$4F,$17,$E8,$FE
	.dc.b	$25,$F8,$0A,$FE,$0E,$D7,$F8,$5B,$5F,$F4,$B3,$5F,$F4,$FF,$CE,$3E
	.dc.b	$6F,$ED,$B6,$FF,$87,$B6,$7F,$89,$B5,$FE,$2E,$BE,$6F,$89,$27,$8B
	.dc.b	$F8,$6F,$71,$F0,$74,$93,$FC,$72,$32,$FF,$1E,$F9,$B7,$C6,$7E,$39
	.dc.b	$B1,$98,$C1,$9F,$1E,$47,$16,$E5,$FC,$23,$E7,$3F,$2F,$DC,$69,$2B
	.dc.b	$F9,$6F,$CD,$3E,$5F,$F3,$AD,$8C,$C6,$0F,$D0,$FC,$CF,$FA,$5B,$E6
	.dc.b	$BF,$9A,$B7,$1F,$E7,$2B,$E0,$5F,$43,$AF,$82,$5F,$F3,$CF,$AB,$FC
	.dc.b	$3E,$17,$A7,$F9,$9C,$62,$3F,$A1,$B5,$FE,$8F,$F3,$8F,$81,$FE,$72
	.dc.b	$D8,$FE,$8E,$D7,$FA,$5B,$5E,$3F,$07,$C0,$8F,$07,$FA,$1F,$8B,$FE
	.dc.b	$45,$8F,$EA,$AD,$7F,$DC,$0D,$7F,$DD,$2D,$5F,$38,$F9,$CF,$E9,$AD
	.dc.b	$8F,$EB,$6D,$7F,$AF,$B5,$FE,$C6,$BE,$73,$89,$23,$8B,$FD,$63,$F1
	.dc.b	$1F,$F1,$74,$7F,$6A,$6B,$FE,$F5,$6B,$FE,$FE,$6A,$F9,$9F,$C7,$BF
	.dc.b	$C2,$FE,$69,$F4,$3F,$EE,$E6,$C7,$F9,$06,$76,$FF,$8E,$5A,$FF,$7A
	.dc.b	$6A,$23,$86,$F5,$FF,$70,$8E,$7E,$37,$F9,$C1,$BD,$1C,$6B,$CD,$5E
	.dc.b	$3F,$63,$FE,$00,$FC,$DF,$F1,$39,$FE,$44,$BC,$0C,$9A,$F2,$24,$EA
	.dc.b	$32,$F0,$8E,$D9,$7F,$BC,$BE,$6D,$F3,$58,$CF,$FC,$ED,$18,$BF,$E6
	.dc.b	$0D,$7F,$91,$2F,$9A,$E1,$0B,$4E,$DE,$D7,$A7,$2D,$8E,$53,$FE,$6F
	.dc.b	$F9,$BF,$F4,$A6,$1F,$E8,$7F,$9D,$FF,$6E,$CF,$F3,$16,$E1,$FE,$98
	.dc.b	$AD,$CF,$D1,$7F,$28,$4F,$82,$FF,$0F,$B6,$3B,$47,$A1,$47,$29,$57
	.dc.b	$FA,$1E,$7F,$A8,$7C,$EF,$D1,$7F,$A1,$3F,$45,$CE,$BE,$D8,$D5,$31
	.dc.b	$FF,$50,$46,$1B,$3B,$38,$6B,$F1,$77,$F9,$13,$FE,$A7,$6E,$1F,$35
	.dc.b	$B8,$7F,$A6,$F6,$E1,$B4,$AF,$EC,$D5,$A3,$E6,$BF,$E9,$1F,$91,$7C
	.dc.b	$EF,$C1,$F9,$D7,$C8,$53,$E2,$4E,$AB,$FD,$44,$D7,$FC,$AE,$D7,$FC
	.dc.b	$B6,$D7,$F3,$4D,$7F,$48,$F1,$CB,$FD,$9E,$DC,$F7,$87,$8F,$E6,$7F
	.dc.b	$3F,$F3,$5F,$A0,$F9,$B9,$8F,$E8,$23,$2F,$E8,$A3,$2F,$F7,$5E,$64
	.dc.b	$FE,$36,$A4,$FE,$69,$FC,$86,$78,$8F,$CD,$C5,$CF,$F4,$93,$F0,$BE
	.dc.b	$77,$E3,$7A,$32,$FE,$1C,$65,$FF,$3B,$EA,$8F,$F3,$6B,$63,$31,$83
	.dc.b	$83,$F4,$DF,$A5,$F3,$BF,$89,$CB,$F1,$5A,$4A,$FC,$8C,$65,$FC,$9C
	.dc.b	$65,$FC,$9C,$7E,$27,$39,$1F,$82,$7F,$19,$96,$23,$FC,$35,$C0,$9C
	.dc.b	$69,$F9,$79,$FC,$75,$7E,$0B,$93,$F0,$DF,$3B,$27,$F0,$71,$97,$FD
	.dc.b	$CD,$C4,$9D,$47,$E3,$E7,$4F,$FB,$86,$74,$F0,$9C,$DF,$E1,$2B,$8B
	.dc.b	$E3,$95,$70,$7A,$1E,$23,$91,$0C,$92,$39,$47,$16,$D7,$83,$8B,$F2
	.dc.b	$F7,$E0,$7F,$99,$56,$DC,$DF,$73,$D5,$5B,$7F,$3B,$5C,$DF,$E8,$55
	.dc.b	$C5,$B3,$95,$BF,$F0,$97,$99,$E9,$C3,$E1,$3D,$DF,$3B,$9C,$3C,$6F
	.dc.b	$B5,$F3,$D7,$8E,$3A,$E1,$DF,$0F,$14,$7C,$A9,$F1,$61,$E1,$C3,$AC
	.dc.b	$F8,$8E,$87,$B0,$F0,$1E,$A3,$C9,$87,$8A,$38,$91,$C1,$23,$81,$D6
	.dc.b	$3D,$AF,$63,$C8,$76,$3D,$27,$63,$8B,$D5,$F2,$1C,$6F,$DA,$7B,$AB
	.dc.b	$D9,$5E,$19,$EA,$78,$5C,$46,$A3,$AC,$7C,$A7,$C0,$78,$8E,$06,$8F
	.dc.b	$0D,$7A,$8F,$23,$D9,$C4,$7D,$68,$FE,$4B,$87,$F0,$42,$3F,$C2,$1A
	.dc.b	$FF,$2C,$6B,$C0,$F1,$D7,$9D,$FD,$1F,$C3,$7D,$A9,$FB,$CD,$8F,$4E
	.dc.b	$08,$C4,$7C,$76,$E1,$E6,$8F,$CB,$C6,$30,$F9,$15,$C6,$7C,$37,$D9
	.dc.b	$A7,$6A,$E0,$FC,$4A,$FD,$4C,$62,$3E,$55,$FE,$CA,$31,$87,$AF,$0E
	.dc.b	$D5,$F8,$38,$F7,$19,$9D,$E1,$FD,$5D,$B3,$FC,$41,$AF,$F0,$D6,$DF
	.dc.b	$78,$F3,$9F,$04,$AF,$C1,$7E,$59,$8C,$3E,$0C,$F2,$7D,$71,$93,$DC
	.dc.b	$7C,$83,$07,$B9,$DB,$EB,$AF,$90,$D3,$E4,$1F,$ED,$BF,$9D,$7C,$87
	.dc.b	$F8,$95,$B1,$98,$C7,$C8,$65,$91,$93,$94,$71,$F9,$DE,$5C,$12,$7B
	.dc.b	$46,$5F,$04,$65,$FF,$6E,$5A,$7F,$C0,$3F,$31,$FE,$05,$F8,$99,$5F
	.dc.b	$E0,$28,$CB,$FD,$81,$FE,$BE,$47,$06,$C8,$7F,$41,$DB,$FC,$85,$F3
	.dc.b	$BF,$EA,$78,$FF,$7E,$35,$FF,$77,$37,$7F,$EA,$81,$7A,$5C,$BD,$49
	.dc.b	$5C,$9E,$B1,$DB,$E0,$3B,$FC,$EC,$AE,$F1,$97,$C3,$19,$7C,$31,$FE
	.dc.b	$45,$F9,$8F,$E6,$F1,$FD,$7B,$D1,$FF,$14,$19,$CE,$47,$F9,$E4,$FF
	.dc.b	$3D,$4F,$07,$01,$C3,$B6,$BC,$23,$9B,$FE,$6B,$E4,$C3,$FD,$BB,$D3
	.dc.b	$F9,$3F,$37,$23,$E4,$BF,$33,$7C,$47,$CA,$8B,$BF,$CA,$DD,$5F,$E7
	.dc.b	$3F,$A7,$70,$DF,$8E,$98,$FF,$A6,$F0,$FA,$35,$A8,$C6,$79,$8C,$47
	.dc.b	$9C,$FF,$41,$2A,$E7,$19,$A4,$F0,$7F,$BF,$7C,$ED,$BE,$9F,$41,$E5
	.dc.b	$3F,$CE,$D1,$9B,$FF,$57,$3F,$EB,$C2,$63,$F3,$A3,$2D,$51,$C5,$B1
	.dc.b	$D0,$F8,$5E,$3F,$FA,$57,$C0,$FF,$7C,$CF,$B8,$FF,$4E,$38,$7D,$4D
	.dc.b	$C7,$F2,$AB,$E4,$39,$7B,$B2,$B9,$57,$F8,$BE,$3B,$4F,$7E,$DF,$07
	.dc.b	$6C,$BD,$E1,$FC,$09,$EF,$39,$7C,$33,$C5,$F3,$1F,$E6,$0F,$07,$D5
	.dc.b	$F0,$72,$33,$E2,$7F,$80,$8F,$F1,$53,$F9,$76,$BF,$ED,$16,$BF,$ED
	.dc.b	$B6,$BD,$4E,$73,$FE,$A9,$8F,$F0,$F6,$7F,$E4,$BF,$80,$FF,$56,$7C
	.dc.b	$FF,$FA,$E2,$FF,$48,$D7,$FD,$BC,$D7,$FD,$D0,$D4,$BD,$2F,$B4,$D8
	.dc.b	$FB,$CD,$7F,$5A,$D7,$F6,$5E,$3B,$E0,$FF,$04,$D8,$FC,$33,$5F,$DA
	.dc.b	$B5,$E3,$D5,$C0,$8F,$07,$9F,$4C,$8E,$6D,$7F,$7F,$F4,$8F,$37,$F2
	.dc.b	$4D,$8F,$CB,$35,$DC,$7E,$5F,$BD,$F2,$F4,$93,$F9,$D8,$CB,$7E,$7B
	.dc.b	$81,$1C,$1F,$A3,$FB,$DF,$88,$88,$E1,$BB,$7E,$9E,$31,$0F,$9B,$7C
	.dc.b	$C7,$EA,$DB,$1F,$AF,$6B,$FC,$79,$AF,$F2,$35,$F3,$DC,$49,$1C,$5E
	.dc.b	$9C,$52,$3A,$35,$E9,$D1,$E1,$23,$E1,$7C,$9D,$12,$3C,$8D,$7C,$9F
	.dc.b	$07,$E9,$23,$E9,$7D,$9F,$3D,$F1,$63,$F8,$D1,$F0,$74,$C7,$3F,$FA
	.dc.b	$4F,$3F,$C8,$36,$B3,$E1,$71,$94,$74,$BF,$8C,$D7,$E3,$3C,$1F,$A9
	.dc.b	$1C,$1F,$DD,$B5,$F1,$FE,$E7,$FD,$8E,$BB,$7C,$37,$1C,$1F,$3E,$EE
	.dc.b	$C3,$70,$E7,$5D,$DF,$B8,$FD,$B2,$3F,$6D,$AF,$C2,$C5,$1C,$5B,$1F
	.dc.b	$C7,$27,$FC,$10,$FA,$3D,$2F,$0D,$48,$C1,$0F,$C3,$47,$E2,$C8,$FF
	.dc.b	$39,$F9,$FF,$E6,$C4,$F1,$7D,$0F,$E8,$23,$EA,$B9,$7F,$90,$BC,$3A
	.dc.b	$3D,$DC,$BF,$CC,$E7,$F9,$2B,$5F,$BA,$DD,$3F,$D9,$DF,$3F,$FB,$07
	.dc.b	$F4,$DF,$3D,$F9,$C9,$D6,$FF,$85,$6D,$FE,$AD,$AA,$F9,$2A,$FD,$2C
	.dc.b	$61,$E9,$E2,$FE,$3D,$B3,$E0,$AF,$C9,$56,$DB,$8B,$1D,$57,$87,$0E
	.dc.b	$55,$E5,$6C,$7E,$FE,$F9,$36,$F9,$3E,$5A,$E0,$F4,$3F,$7F,$5C,$9F
	.dc.b	$64,$F0,$C3,$D6,$DA,$E3,$3C,$CF,$21,$CE,$FC,$2F,$2C,$3B,$4F,$D3
	.dc.b	$BF,$83,$3C,$4D,$16,$BE,$5B,$FE,$2D,$6B,$F1,$2B,$FC,$01,$18,$8F
	.dc.b	$81,$EB,$BE,$7F,$2D,$8F,$F1,$9C,$FF,$A9,$67,$27,$F8,$1E,$1F,$25
	.dc.b	$48,$FF,$82,$DB,$87,$F2,$0C,$3F,$C5,$3A,$7C,$9A,$FE,$5D,$5F,$2E
	.dc.b	$73,$1E,$D3,$27,$BD,$CD,$7F,$28,$8F,$09,$9D,$97,$D6,$33,$3E,$19
	.dc.b	$CC,$75,$33,$9F,$F1,$88,$CC,$F5,$C3,$C9,$18,$9F,$F2,$AC,$62,$FB
	.dc.b	$36,$FA,$36,$FC,$13,$FE,$10,$C3,$DA,$78,$AB,$F9,$55,$71,$8C,$1A
	.dc.b	$9F,$E5,$7E,$77,$F9,$F1,$F3,$1C,$DC,$57,$F5,$7F,$9F,$FF,$16,$4F
	.dc.b	$F8,$79,$B1,$FC,$D9,$FE,$B9,$19,$DF,$8C,$A7,$77,$FE,$8A,$E2,$F1
	.dc.b	$4F,$5A,$EF,$3F,$C7,$B4,$E3,$F3,$CF,$9F,$CF,$CF,$BE,$83,$31,$86
	.dc.b	$E9,$FE,$3E,$9D,$3F,$E3,$4A,$E4,$E6,$7C,$BF,$0C,$F9,$FF,$91,$A7
	.dc.b	$F5,$18,$C5,$7F,$96,$DB,$1F,$E8,$A6,$BF,$E8,$F7,$39,$FF,$A9,$BE
	.dc.b	$7D,$1D,$AF,$DB,$18,$BF,$F2,$3C,$F0,$7F,$CF,$D1,$9B,$FE,$C3,$39
	.dc.b	$7F,$88,$46,$6F,$EF,$7C,$FB,$DC,$F4,$D7,$F8,$36,$31,$87,$F9,$CD
	.dc.b	$B1,$FE,$A2,$6B,$FE,$A7,$F9,$FF,$B2,$FF,$A9,$9A,$E4,$EA,$1C,$BF
	.dc.b	$DD,$FE,$19,$F0,$3A,$6D,$FF,$A0,$DB,$3F,$EB,$26,$BF,$EB,$7F,$9E
	.dc.b	$7C,$F7,$66,$D7,$FA,$2D,$B3,$FE,$BA,$6B,$FE,$BF,$F8,$6F,$F1,$F4
	.dc.b	$FF,$A4,$5B,$1F,$EC,$06,$BB,$F4,$3D,$7F,$64,$F0,$E9,$2F,$FD,$35
	.dc.b	$19,$7F,$D3,$9F,$3F,$FD,$F5,$BE,$3F,$C4,$8C,$61,$FE,$CE,$6B,$FE
	.dc.b	$D3,$EE,$FF,$31,$57,$FA,$95,$B1,$FC,$03,$D7,$F9,$B1,$88,$FF,$6D
	.dc.b	$7D,$8F,$F8,$E2,$BF,$D5,$8D,$8F,$F6,$DB,$5D,$FA,$3F,$5B,$E7,$FF
	.dc.b	$DE,$BF,$3F,$FE,$B3,$6C,$FF,$B8,$9F,$AF,$EF,$E7,$E2,$55,$FE,$B9
	.dc.b	$F9,$E7,$CF,$FF,$AE,$5B,$86,$63,$07,$FA,$F7,$9D,$F2,$74,$93,$FE
	.dc.b	$C3,$8C,$BF,$EC,$5F,$9F,$7C,$67,$FB,$15,$B1,$98,$C1,$FE,$C9,$F2
	.dc.b	$3F,$CC,$64,$F5,$B0,$7F,$B3,$7C,$6F,$D1,$57,$FB,$45,$AF,$FC,$00
	.dc.b	$6B,$FE,$4D,$0E,$1F,$F2,$87,$CF,$BE,$7F,$FD,$B0,$D8,$FF,$6E,$35
	.dc.b	$E4,$1C,$3D,$7E,$7F,$FD,$E6,$DF,$2F,$FD,$C5,$18,$8F,$F2,$F3,$5F
	.dc.b	$F4,$27,$E1,$BF,$C7,$5F,$F7,$53,$5E,$ED,$7D,$4D,$7F,$CD,$1F,$20
	.dc.b	$F9,$8F,$F2,$77,$CA,$BE,$22,$9F,$E5,$19,$D5,$7F,$B6,$1A,$FB,$1A
	.dc.b	$FC,$76,$BD,$AB,$FD,$F9,$1C,$3C,$AF,$EA,$0D,$EC,$D5,$7F,$B9,$30
	.dc.b	$7C,$67,$F9,$9F,$E6,$1F,$19,$7E,$5E,$5F,$FB,$DA,$72,$D9,$CB,$FE
	.dc.b	$31,$C3,$FD,$55,$92,$97,$D6,$FF,$CE,$9F,$39,$E0,$C3,$FE,$00,$B5
	.dc.b	$E4,$7F,$8E,$DC,$5F,$F7,$A0,$FC,$98,$E3,$8F,$42,$3D,$35,$F9,$9F
	.dc.b	$33,$F4,$3C,$99,$3F,$36,$32,$FE,$20,$6C,$8F,$96,$48,$D4,$65,$E4
	.dc.b	$43,$FD,$4B,$B3,$CD,$FF,$49,$7C,$D7,$C8,$9F,$6E,$88,$F8,$B1,$F2
	.dc.b	$38,$3E,$4C,$62,$36,$F1,$9F,$84,$D9,$DE,$FD,$8E,$FB,$FF,$A8,$7C
	.dc.b	$BF,$93,$3F,$51,$B8,$7F,$99,$DB,$87,$F2,$FD,$B8,$6F,$2B,$C8,$6D
	.dc.b	$E9,$1F,$97,$F3,$3F,$37,$F2,$32,$BF,$31,$19,$7E,$3B,$F9,$42,$38
	.dc.b	$36,$43,$E4,$FC,$07,$FA,$E7,$E6,$9F,$15,$49,$3E,$4C,$92,$F8,$AC
	.dc.b	$CF,$CE,$8E,$2E,$DA,$E6,$3B,$7C,$D7,$7D,$64,$EE,$32,$F1,$8C,$BD
	.dc.b	$63,$F7,$1F,$23,$FB,$CE,$D9,$3F,$B9,$8C,$BF,$88,$7F,$6A,$47,$06
	.dc.b	$C8,$7E,$2B,$9B,$FD,$B5,$F3,$1F,$ED,$D8,$FF,$4F,$37,$33,$47,$97
	.dc.b	$C3,$C7,$95,$F1,$E7,$53,$D3,$D9,$EB,$F3,$52,$B5,$19,$78,$46,$5F
	.dc.b	$CC,$D7,$F3,$AC,$3F,$CB,$9F,$25,$FA,$89,$FE,$19,$E4,$FF,$1E,$3F
	.dc.b	$DD,$91,$98,$FD,$E8,$71,$3F,$E6,$5E,$DF,$E9,$F3,$FA,$AA,$E3,$A4
	.dc.b	$8E,$A4,$39,$DF,$2E,$94,$48,$C3,$8B,$EB,$87,$7C,$3A,$4F,$2C,$3D
	.dc.b	$0F,$59,$F6,$C6,$9E,$11,$9C,$3B,$BE,$B9,$D4,$62,$FB,$B6,$F3,$5E
	.dc.b	$5A,$EE,$FB,$CF,$D7,$9C,$CB,$87,$A9,$5F,$47,$D9,$5D,$1E,$07,$32
	.dc.b	$61,$EA,$79,$DF,$82,$3D,$71,$F0,$4C,$9E,$D3,$A9,$83,$47,$4C,$3D
	.dc.b	$8E,$63,$B3,$93,$89,$C1,$C9,$E0,$24,$7C,$AC,$3D,$C7,$BF,$0F,$75
	.dc.b	$7C,$0C,$3D,$D3,$97,$D5,$5C,$70,$F8,$57,$C0,$E3,$5D,$23,$77,$E4
	.dc.b	$C3,$C7,$7C,$63,$E0,$1D,$CF,$63,$97,$D8,$64,$F6,$39,$8F,$5C,$75
	.dc.b	$8C,$11,$C1,$0F,$AF,$1E,$27,$17,$E8,$AF,$CA,$C7,$F8,$96,$FF,$54
	.dc.b	$D8,$FE,$B0,$FE,$00,$C1,$FD,$6E,$36,$FA,$4C,$9F,$E2,$67,$E2,$3E
	.dc.b	$E3,$F4,$DF,$35,$F3,$A7,$E5,$B6,$34,$7E,$49,$C3,$F0,$8E,$26,$A7
	.dc.b	$AB,$FD,$92,$37,$7E,$F8,$E4,$6E,$BB,$B4,$EF,$6A,$E3,$87,$CB,$9E
	.dc.b	$27,$84,$F9,$6E,$67,$84,$73,$34,$E1,$E1,$87,$9B,$0F,$E2,$57,$DA
	.dc.b	$7F,$6C,$7F,$89,$DC,$3D,$8F,$79,$CB,$0F,$4B,$4F,$4C,$70,$AF,$CE
	.dc.b	$36,$7F,$C8,$6D,$9E,$C7,$9E,$BE,$35,$F9,$DF,$91,$7E,$7C,$3F,$87
	.dc.b	$5F,$1B,$F9,$C7,$18,$F8,$51,$9A,$F6,$CF,$F4,$3F,$95,$FE,$9A,$47
	.dc.b	$FA,$37,$05,$7F,$9F,$C5,$9F,$F1,$96,$8A,$FE,$43,$1F,$E3,$5F,$9A
	.dc.b	$71,$63,$2F,$F2,$1C,$3E,$1F,$19,$2D,$39,$8F,$EF,$8F,$26,$DF,$09
	.dc.b	$DD,$5A,$E1,$5F,$E2,$C7,$83,$FD,$3B,$E6,$6D,$79,$7E,$6B,$81,$E2
	.dc.b	$94,$EB,$1F,$E7,$6F,$FA,$06,$33,$8F,$EB,$A3,$04,$79,$3C,$63,$FA
	.dc.b	$8B,$6F,$DB,$F3,$5E,$9F,$01,$EB,$52,$5F,$F8,$6E,$1F,$13,$88,$FF
	.dc.b	$23,$C7,$07,$53,$CE,$7A,$77,$75,$ED,$95,$CE,$32,$F5,$8C,$BD,$23
	.dc.b	$E1,$78,$6F,$0B,$CD,$EF,$4A,$FE,$A7,$39,$6C,$FF,$10,$8C,$47,$FA
	.dc.b	$C1,$AF,$FA,$ED,$AF,$F6,$67,$7A,$37,$C5,$78,$8F,$8D,$F2,$DF,$7F
	.dc.b	$EB,$3F,$C3,$78,$7F,$8F,$DA,$FF,$9A,$9B,$7F,$E8,$55,$EF,$F1,$F0
	.dc.b	$25,$71,$7A,$47,$5E,$FE,$DF,$0F,$2B,$B4,$65,$F0,$46,$5F,$04,$7E
	.dc.b	$27,$17,$C5,$F9,$F9,$65,$7F,$7A,$9C,$B6,$7F,$9B,$C6,$23,$FD,$AE
	.dc.b	$D7,$FD,$C4,$D7,$FC,$1A,$EF,$3F,$91,$1F,$E1,$E8,$C6,$BA,$AF,$D3
	.dc.b	$7D,$67,$F0,$78,$FF,$59,$BF,$A6,$BF,$F5,$53,$63,$3D,$5C,$3E,$4B
	.dc.b	$8B,$31,$FF,$37,$56,$DF,$DC,$29,$EF,$E8,$78,$3A,$65,$70,$8C,$BC
	.dc.b	$A3,$2F,$FA,$8A,$93,$FB,$3F,$AC,$FC,$2E,$3F,$EC,$AF,$95,$FF,$16
	.dc.b	$DF,$FA,$A1,$B9,$FC,$28,$57,$07,$FC,$82,$FE,$41,$BC,$9F,$D9,$A3
	.dc.b	$11,$C2,$30,$78,$48,$FF,$A3,$69,$F4,$7F,$C1,$FE,$79,$EE,$FF,$70
	.dc.b	$6D,$7F,$B2,$5A,$FA,$03,$87,$D7,$E8,$FE,$59,$FF,$68,$35,$E0,$D7
	.dc.b	$7C,$DF,$DF,$A1,$E5,$D2,$4F,$27,$5E,$24,$8E,$AD,$7C,$9F,$3D,$F4
	.dc.b	$1F,$F7,$03,$5C,$F0,$7F,$83,$23,$11,$F4,$97,$CF,$64,$91,$C9,$FF
	.dc.b	$09,$F9,$5A,$8F,$F0,$94,$60,$8F,$FB,$BF,$E7,$BE,$23,$FE,$6A,$6B
	.dc.b	$FE,$7A,$6A,$6F,$D3,$BE,$7B,$FC,$42,$D9,$FF,$7D,$35,$FA,$4D,$7F
	.dc.b	$17,$F3,$AF,$8B,$FF,$14,$B6,$3E,$C0,$70,$FD,$B6,$BC,$BB,$78,$91
	.dc.b	$E3,$08,$FF,$1A,$FC,$F3,$E9,$E9,$2B,$BC,$65,$E3,$E4,$79,$08,$E0
	.dc.b	$EB,$F4,$1E,$82,$38,$39,$FC,$17,$80,$8E,$0F,$2E,$9F,$00,$8E,$08
	.dc.b	$7F,$AD,$B2,$F9,$5D,$9F,$AF,$6F,$6F,$FA,$83,$2F,$F6,$A5,$17,$AF
	.dc.b	$F0,$09,$1E,$E7,$E7,$FC,$44,$8F,$8E,$D7,$E7,$3F,$DD,$6C,$D3,$FC
	.dc.b	$05,$DA,$F6,$71,$5F,$EA,$8A,$DB,$6B,$73,$CF,$DA,$FB,$E4,$9F,$AE
	.dc.b	$D7,$E0,$3B,$7F,$81,$3F,$E0,$7F,$A0,$FE,$FD,$06,$93,$C5,$D7,$A7
	.dc.b	$27,$FD,$77,$39,$7F,$D7,$BF,$3D,$FD,$F9,$BE,$97,$F8,$4E,$31,$1F
	.dc.b	$EE,$6F,$A0,$FE,$5D,$C3,$FE,$DC,$F9,$F5,$7F,$B4,$0D,$C7,$F2,$36
	.dc.b	$F1,$67,$0F,$F0,$F5,$F2,$C1,$87,$27,$FD,$CF,$19,$8C,$47,$69,$FF
	.dc.b	$6E,$C6,$2F,$CA,$EA,$B8,$47,$FB,$89,$B5,$FE,$F2,$8D,$9E,$37,$D4
	.dc.b	$EF,$3F,$F2,$14,$79,$23,$18,$EA,$39,$61,$E2,$9E,$51,$B3,$D3,$5C
	.dc.b	$B0,$F2,$36,$77,$1E,$9A,$E5,$5C,$2F,$B2,$C3,$D2,$FC,$13,$37,$EA
	.dc.b	$78,$1E,$13,$86,$1E,$A7,$FA,$BC,$FC,$F6,$C6,$5F,$69,$FD,$16,$1F
	.dc.b	$00,$FA,$0F,$E2,$31,$86,$C6,$67,$83,$5F,$E2,$FF,$2B,$FE,$C2,$4D
	.dc.b	$23,$FE,$C6,$7D,$F8,$FF,$9D,$B4,$F8,$32,$38,$9E,$D8,$D3,$EF,$35
	.dc.b	$3C,$B6,$FF,$09,$EB,$F8,$89,$C9,$C6,$33,$A7,$F6,$3A,$E1,$87,$82
	.dc.b	$BB,$DF,$E4,$30,$EE,$F3,$8C,$DF,$FA,$3B,$0E,$71,$FE,$04,$C3,$9B
	.dc.b	$E5,$36,$7A,$EF,$9D,$78,$63,$C9,$87,$09,$FF,$05,$1E,$E3,$FC,$19
	.dc.b	$1E,$AC,$3C,$98,$7F,$8C,$27,$6D,$AF,$F5,$0F,$9D,$6F,$C1,$35,$FF
	.dc.b	$4F,$45,$FA,$0F,$F3,$ED,$79,$4F,$A1,$28,$CB,$FE,$A5,$8C,$C7,$D5
	.dc.b	$78,$47,$D7,$8D,$DF,$F8,$8D,$B3,$F0,$78,$FD,$8F,$2E,$0F,$35,$7F
	.dc.b	$2C,$C7,$CD,$F4,$1F,$E1,$CA,$EB,$8A,$30,$F8,$CF,$F5,$C7,$C8,$7F
	.dc.b	$12,$9D,$CE,$AF,$FC,$68,$F0,$AF,$F1,$CB,$5F,$7B,$5F,$C8,$35,$7A
	.dc.b	$AF,$4B,$E5,$36,$FE,$0D,$BE,$79,$DD,$AB,$95,$EA,$BE,$18,$7A,$70
	.dc.b	$E1,$7A,$8F,$61,$E0,$BF,$C0,$BF,$A3,$AE,$2F,$F9,$49,$AF,$D8,$6B
	.dc.b	$FA,$9F,$9E,$7A,$FC,$A1,$18,$C7,$FC,$1A,$D8,$FF,$23,$35,$FF,$2F
	.dc.b	$BF,$0A,$37,$5C,$E3,$13,$C2,$31,$F1,$12,$BA,$B6,$7F,$CA,$5F,$38
	.dc.b	$F8,$DF,$23,$6B,$FC,$3E,$D9,$DC,$EB,$EA,$E5,$69,$B3,$D3,$E5,$F9
	.dc.b	$92,$79,$BF,$07,$EA,$DF,$0B,$49,$3F,$E2,$E8,$CB,$FE,$30,$FA,$8F
	.dc.b	$F1,$43,$7C,$CF,$95,$18,$C3,$FC,$D8,$D7,$FD,$0B,$F3,$8F,$81,$7C
	.dc.b	$E7,$F8,$EA,$30,$DF,$2B,$FD,$D5,$5C,$2B,$6F,$F9,$DF,$E9,$DF,$4F
	.dc.b	$FE,$42,$84,$61,$FF,$22,$B5,$E1,$1B,$F9,$0F,$C9,$C7,$F9,$29,$AF
	.dc.b	$FA,$0D,$AF,$FA,$71,$AF,$F0,$25,$C7,$CC,$91,$CD,$F2,$7C,$E3,$ED
	.dc.b	$69,$27,$FC,$B3,$19,$6F,$D7,$70,$23,$83,$77,$D0,$8E,$0F,$F3,$1F
	.dc.b	$57,$E6,$DF,$CF,$9F,$D1,$DC,$13,$D5,$F9,$0F,$F9,$6E,$3F,$D4,$4D
	.dc.b	$7F,$D7,$6D,$7F,$90,$BF,$09,$E9,$D9,$C8,$8F,$F9,$D3,$E5,$9F,$4D
	.dc.b	$FC,$A3,$BF,$F9,$A7,$D5,$3E,$A3,$1D,$1F,$D0,$63,$18,$7F,$1C,$C3
	.dc.b	$8B,$70,$FB,$D1,$88,$FF,$5C,$35,$FF,$68,$B5,$FE,$6E,$EF,$0F,$F6
	.dc.b	$93,$6F,$35,$E2,$A5,$2F,$FD,$09,$1F,$C6,$BE,$74,$6E,$6B,$4C,$C6
	.dc.b	$FE,$37,$11,$23,$93,$B8,$E5,$F3,$BF,$EB,$19,$FE,$41,$F5,$3F,$EE
	.dc.b	$8C,$BF,$B2,$36,$3C,$B3,$BC,$3F,$AF,$36,$B7,$3E,$5A,$E1,$4A,$D4
	.dc.b	$FF,$82,$7E,$AB,$FD,$0A,$FC,$0A,$FF,$71,$D7,$F6,$96,$DF,$F6,$0A
	.dc.b	$DD,$1F,$F5,$B3,$9F,$73,$F6,$51,$88,$ED,$18,$3E,$73,$C5,$B1,$E2
	.dc.b	$B5,$FD,$0A,$FE,$85,$F3,$DF,$31,$B9,$FD,$18,$C5,$7F,$7F,$59,$E6
	.dc.b	$65,$7F,$52,$9D,$E5,$FE,$0F,$77,$1F,$D5,$BD,$9F,$8B,$83,$E7,$BE
	.dc.b	$49,$F2,$A3,$26,$9C,$CF,$E0,$E3,$E6,$AF,$82,$7D,$67,$E1,$F8,$3F
	.dc.b	$C5,$19,$7E,$7F,$D6,$F8,$75,$AB,$FE,$F6,$D7,$FD,$C7,$7F,$EE,$82
	.dc.b	$3F,$EE,$87,$89,$1F,$ED,$F6,$F8,$1F,$CA,$7C,$E4,$9F,$CD,$3F,$97
	.dc.b	$F8,$AC,$47,$F6,$FD,$D3,$8D,$BF,$C5,$50,$AF,$C6,$7C,$CF,$E3,$7D
	.dc.b	$69,$87,$F7,$EE,$3F,$E6,$F8,$7E,$26,$B3,$3F,$9C,$AE,$2F,$F8,$63
	.dc.b	$91,$F3,$1F,$97,$DA,$4E,$63,$F0,$A7,$E3,$08,$FF,$C0,$09,$E2,$D8
	.dc.b	$DB,$C4,$F2,$6B,$F8,$7D,$64,$7E,$0D,$FC,$56,$98,$8F,$C7,$C5,$CF
	.dc.b	$F1,$93,$FB,$FE,$3F,$6F,$17,$F9,$5D,$FF,$5C,$DF,$F0,$86,$D7,$E2
	.dc.b	$75,$E0,$71,$D6,$57,$F8,$92,$B9,$BF,$E7,$BF,$8A,$7C,$0F,$E9,$BD
	.dc.b	$9F,$A1,$3F,$E3,$E6,$BF,$E2,$29,$DB,$FE,$42,$6B,$FE,$26,$8D,$E2
	.dc.b	$CB,$FC,$B3,$18,$7F,$CB,$75,$C2,$75,$5E,$3B,$F1,$67,$23,$B4,$7E
	.dc.b	$EA,$B8,$5F,$4A,$D5,$79,$6B,$46,$2F,$AC,$30,$F1,$3D,$CE,$A7,$BE
	.dc.b	$B8,$5F,$9A,$F9,$CE,$48,$77,$38,$9F,$06,$78,$BD,$4D,$19,$38,$DF
	.dc.b	$93,$0F,$5C,$6E,$3C,$F8,$79,$A3,$D1,$5E,$27,$9C,$EA,$3B,$C2,$B9
	.dc.b	$B6,$F3,$7C,$E7,$26,$9E,$C6,$63,$B5,$F5,$AF,$3E,$1F,$0D,$F7,$C6
	.dc.b	$5F,$61,$E4,$8D,$38,$8D,$1E,$F9,$EB,$3F,$66,$7A,$9F,$05,$E3,$3C
	.dc.b	$1D,$47,$53,$47,$FB,$4E,$BD,$38,$74,$8C,$9F,$20,$EA,$7C,$33,$D3
	.dc.b	$3D,$2B,$B4,$70,$C3,$E2,$57,$CC,$7E,$3E,$1F,$2A,$3B,$4F,$33,$DD
	.dc.b	$7C,$CF,$49,$F0,$9F,$F5,$C9,$B2,$D7,$EA,$1B,$5F,$E4,$B8,$C4,$71
	.dc.b	$21,$F8,$B3,$67,$A8,$F7,$E1,$FC,$A9,$C1,$E5,$C3,$E8,$3F,$BB,$D3
	.dc.b	$D3,$5E,$C8,$C1,$C1,$C3,$FB,$B8,$DD,$FC,$0B,$ED,$3F,$4C,$ED,$87
	.dc.b	$C0,$7E,$A5,$70,$7C,$73,$C3,$0F,$1C,$71,$C3,$F2,$55,$B8,$E9,$5B
	.dc.b	$BE,$91,$D8,$F7,$CE,$EF,$99,$FA,$87,$13,$C2,$B9,$57,$DD,$C3,$D5
	.dc.b	$5A,$8E,$67,$C1,$8C,$B4,$F4,$C6,$DF,$11,$FC,$2D,$C6,$1E,$A8,$E0
	.dc.b	$7A,$A3,$33,$FC,$B5,$B1,$F4,$0F,$85,$1E,$06,$9E,$03,$9E,$1E,$EC
	.dc.b	$3F,$95,$D7,$F8,$EB,$CA,$95,$FE,$E0,$EB,$FF,$4B,$BF,$ED,$B5,$F3
	.dc.b	$BA,$8C,$4D,$C3,$FC,$77,$3F,$E3,$9F,$23,$FC,$75,$F2,$72,$FF,$B7
	.dc.b	$46,$5F,$EE,$11,$97,$FC,$93,$1F,$CD,$7E,$83,$FA,$57,$83,$27,$FA
	.dc.b	$34,$65,$FD,$19,$FD,$45,$FD,$01,$1C,$9F,$C6,$DE,$0F,$F5,$17,$FA
	.dc.b	$B7,$CD,$7F,$88,$72,$93,$FE,$2B,$7F,$C4,$98,$E2,$3F,$DF,$9D,$DF
	.dc.b	$13,$3F,$E9,$E1,$C6,$C9,$FF,$39,$C7,$C8,$7F,$B4,$67,$FD,$CF,$49
	.dc.b	$18,$21,$FE,$7A,$AF,$EE,$D1,$A8,$C1,$FD,$DE,$7F,$AD,$7C,$77,$F6
	.dc.b	$FF,$6E,$4F,$F6,$B8,$CB,$F0,$B8,$BF,$CE,$CF,$F0,$88,$FE,$0C,$47
	.dc.b	$81,$0F,$EE,$5E,$B7,$F8,$27,$E5,$7F,$C1,$5D,$B2,$7F,$C0,$51,$97
	.dc.b	$FC,$09,$19,$7C,$FA,$FF,$94,$E7,$FC,$25,$EA,$C9,$FF,$05,$3F,$E1
	.dc.b	$5E,$CC,$47,$0E,$AF,$F3,$55,$FC,$E6,$EF,$FE,$19,$AF,$F3,$9F,$C7
	.dc.b	$FF,$AC,$2F,$FC,$11,$F5,$5F,$E1,$13,$3E,$54,$8F,$C1,$CE,$DF,$E5
	.dc.b	$6D,$7E,$C0,$70,$FD,$B7,$B6,$3F,$E0,$26,$E1,$C1,$B3,$FD,$2F,$EB
	.dc.b	$3F,$D5,$DF,$40,$F9,$7B,$12,$7F,$0D,$87,$E2,$AB,$51,$8C,$33,$18
	.dc.b	$8F,$96,$AB,$94,$62,$FB,$FC,$04,$8E,$47,$FB,$12,$91,$E4,$F5,$B5
	.dc.b	$3D,$22,$E3,$E3,$9F,$9B,$C1,$FE,$3B,$F6,$65,$F4,$3E,$34,$69,$F9
	.dc.b	$AF,$EC,$1C,$98,$21,$F9,$A3,$25,$D3,$FD,$6E,$FF,$73,$F4,$7F,$D5
	.dc.b	$FF,$31,$FE,$8B,$8F,$EC,$2D,$7E,$BB,$5F,$D8,$35,$3F,$EB,$B6,$FA
	.dc.b	$3E,$58,$C1,$3E,$BB,$04,$9F,$E1,$30,$8F,$67,$CB,$7F,$24,$9F,$EE
	.dc.b	$6D,$7F,$12,$D7,$F8,$13,$53,$FC,$0D,$FE,$32,$D7,$9F,$7F,$63,$BF
	.dc.b	$D1,$C8,$FE,$37,$28,$C4,$7F,$22,$F9,$3F,$CB,$A6,$93,$97,$EA,$2F
	.dc.b	$4E,$92,$7F,$93,$C6,$5B,$F2,$FC,$08,$E0,$FE,$5B,$EA,$7C,$56,$E7
	.dc.b	$F8,$18,$C4,$7F,$36,$6B,$FE,$1A,$F9,$6F,$CD,$3F,$CD,$9A,$FF,$3C
	.dc.b	$6B,$BE,$8F,$C6,$70,$FD,$B2,$70,$60,$FC,$85,$3A,$A9,$2F,$B3,$AF
	.dc.b	$86,$91,$E1,$6B,$E1,$F8,$6E,$64,$7C,$CF,$6F,$06,$47,$A5,$AF,$A7
	.dc.b	$C4,$FE,$17,$D5,$FE,$88,$C5,$58,$EE,$FE,$1F,$18,$BF,$D6,$2B,$D4
	.dc.b	$7F,$4B,$8C,$3F,$1E,$3D,$E4,$7E,$57,$93,$C3,$07,$CB,$7F,$28,$94
	.dc.b	$63,$C3,$EB,$0C,$17,$86,$C7,$F0,$AA,$E0,$DB,$E8,$F1,$AE,$07,$D2
	.dc.b	$23,$E4,$3E,$F7,$AF,$E5,$9F,$F7,$F4,$73,$F0,$FC,$13,$23,$04,$3F
	.dc.b	$84,$60,$8F,$4F,$CA,$FF,$AC,$E7,$F8,$A3,$5F,$E1,$CD,$C7,$FD,$72
	.dc.b	$FF,$A7,$0C,$69,$FD,$01,$B3,$FD,$E5,$A9,$FF,$51,$F4,$7F,$BA,$23
	.dc.b	$F5,$5E,$A7,$CA,$8C,$7C,$AF,$FB,$8B,$0F,$F0,$17,$C9,$7E,$AE,$F6
	.dc.b	$7F,$29,$AE,$47,$0F,$53,$EA,$C2,$78,$5F,$38,$C5,$7F,$56,$18,$7F
	.dc.b	$58,$AE,$D1,$89,$F0,$BF,$D7,$2B,$94,$F4,$BE,$EF,$0B,$E5,$5D,$5B
	.dc.b	$7F,$E1,$BB,$E3,$3C,$EF,$8B,$B9,$EE,$7F,$88,$5F,$13,$E8,$39,$47
	.dc.b	$06,$E1,$C1,$F8,$A6,$CB,$87,$9E,$35,$7E,$29,$ED,$87,$99,$DE,$1E
	.dc.b	$5A,$DE,$1F,$32,$7F,$D5,$37,$F2,$DB,$19,$78,$1F,$E4,$C8,$7C,$A3
	.dc.b	$E5,$7F,$C9,$71,$86,$C6,$67,$83,$5F,$F2,$8F,$CA,$FF,$63,$4D,$23
	.dc.b	$FD,$9D,$F8,$58,$3D,$86,$DF,$16,$BF,$C9,$95,$A9,$CC,$7C,$73,$27
	.dc.b	$C9,$AE,$55,$C6,$3C,$26,$67,$8E,$BF,$95,$8E,$2F,$E5,$EB,$8C,$75
	.dc.b	$38,$C6,$67,$8E,$9F,$E7,$2C,$3C,$98,$77,$6D,$F1,$6D,$F7,$8E,$75
	.dc.b	$FC,$9A,$F9,$C7,$F8,$91,$FF,$36,$DE,$4F,$F0,$84,$79,$A7,$0F,$F8
	.dc.b	$2F,$0F,$39,$F1,$1F,$E3,$FF,$2B,$C9,$E9,$5F,$E1,$99,$4E,$65,$F2
	.dc.b	$BF,$AC,$AF,$F0,$F3,$63,$FC,$30,$FF,$9F,$63,$3B,$F1,$8D,$5F,$FA
	.dc.b	$13,$E5,$7C,$75,$FE,$4C,$9C,$BF,$E4,$08,$CD,$FF,$A1,$FE,$51,$E5
	.dc.b	$FA,$63,$FC,$A7,$18,$AF,$F4,$DB,$63,$FD,$42,$D7,$FC,$B2,$1D,$61
	.dc.b	$FE,$96,$F9,$57,$8B,$EF,$9C,$DF,$F9,$4A,$72,$FF,$8F,$7C,$F7,$C3
	.dc.b	$FC,$56,$C7,$F4,$3F,$95,$79,$F9,$A9,$3F,$EB,$0A,$DE,$7F,$EB,$66
	.dc.b	$BE,$66,$BF,$E7,$07,$79,$FF,$AB,$30,$7A,$3D,$2B,$FD,$17,$18,$BF
	.dc.b	$F2,$97,$CA,$FF,$97,$1F,$F5,$FB,$67,$FD,$8A,$D7,$E2,$B5,$FF,$3F
	.dc.b	$BC,$DD,$D7,$D1,$7F,$D7,$51,$97,$8C,$65,$7D,$6B,$E0,$F0,$DB,$FF
	.dc.b	$68,$B6,$7F,$DA,$ED,$7E,$BF,$CB,$3C,$7E,$8D,$AF,$F6,$CB,$67,$FD
	.dc.b	$BE,$D7,$8F,$8F,$C0,$93,$C1,$F3,$63,$F3,$A7,$FD,$A8,$D7,$E6,$ED
	.dc.b	$A3,$D9,$F2,$8F,$1E,$93,$0F,$F7,$5C,$65,$FF,$76,$FC,$A3,$E5,$3F
	.dc.b	$CC,$0D,$F8,$2C,$FB,$3F,$17,$2F,$F7,$BC,$E5,$FF,$7C,$FD,$57,$F8
	.dc.b	$85,$BA,$FF,$A2,$63,$11,$F9,$C6,$BF,$EB,$6F,$91,$FC,$8B,$FE,$FE
	.dc.b	$6B,$FA,$20,$E1,$FD,$2B,$5E,$5B,$7F,$A5,$E3,$F1,$F1,$CE,$30,$7F
	.dc.b	$A6,$FE,$07,$9C,$7F,$BA,$E7,$04,$7C,$03,$E4,$DF,$27,$FE,$A3,$6C
	.dc.b	$69,$AF,$9D,$AF,$EE,$9E,$EE,$67,$91,$1E,$FD,$FF,$E9,$F2,$7C,$17
	.dc.b	$FA,$C5,$FF,$51,$71,$BE,$4F,$FD,$68,$DA,$F3,$B5,$F9,$0D,$7F,$85
	.dc.b	$BF,$1D,$1F,$C1,$3B,$79,$91,$F8,$3F,$4F,$FC,$B1,$BE,$2F,$FB,$0E
	.dc.b	$31,$1F,$55,$AF,$F1,$8F,$69,$97,$F9,$67,$E4,$7F,$A9,$73,$FF,$B9
	.dc.b	$9B,$5E,$2A,$DB,$FE,$D3,$7F,$D8,$04,$AF,$F4,$91,$23,$4F,$F9,$82
	.dc.b	$38,$C7,$F9,$92,$BB,$7A,$7F,$E6,$9F,$49,$F1,$B4,$95,$FE,$63,$87
	.dc.b	$C7,$62,$39,$1F,$E6,$E8,$E2,$EE,$7A,$4F,$F9,$D7,$E4,$7F,$49,$3C
	.dc.b	$3C,$54,$7F,$8A,$1F,$EF,$A6,$3C,$CE,$1E,$24,$AF,$F3,$A5,$6F,$3F
	.dc.b	$AB,$7B,$9C,$D7,$F9,$FB,$E4,$5F,$23,$E1,$DE,$57,$FB,$EE,$72,$D9
	.dc.b	$FF,$7D,$46,$23,$F5,$8D,$7F,$A7,$B5,$FC,$28,$78,$5F,$F7,$36,$E1
	.dc.b	$A6,$CF,$FA,$53,$1F,$F4,$1B,$F0,$FE,$4B,$E4,$B7,$5F,$F4,$E4,$63
	.dc.b	$0F,$8A,$FD,$29,$FF,$4B,$39,$C3,$E1,$B6,$3F,$D4,$2D,$93,$C6,$3E
	.dc.b	$97,$C5,$FD,$49,$FA,$2F,$CD,$8F,$F5,$4B,$60,$EE,$BE,$8B,$67,$7A
	.dc.b	$FF,$AC,$E3,$81,$98,$76,$3D,$8F,$F5,$9F,$A7,$E3,$C7,$FC,$34,$DC
	.dc.b	$7E,$6D,$6E,$FE,$FC,$62,$FF,$8B,$12,$3F,$89,$CE,$DF,$E4,$8D,$BF
	.dc.b	$F1,$33,$6B,$94,$F4,$3F,$7F,$7F,$1F,$E4,$9E,$2F,$FB,$35,$B3,$FE
	.dc.b	$C5,$6B,$F9,$B5,$F1,$3B,$99,$3B,$79,$47,$E8,$FE,$3B,$F4,$9F,$1F
	.dc.b	$2B,$F3,$B1,$97,$F3,$F1,$97,$F5,$1C,$29,$FF,$6D,$E6,$F9,$1F,$97
	.dc.b	$F2,$52,$FD,$87,$D0,$8D,$9F,$CF,$48,$FC,$86,$BF,$BC,$38,$1C,$A7
	.dc.b	$B3,$D3,$4F,$F7,$3B,$FB,$BF,$92,$F8,$7B,$FE,$E9,$BB,$FE,$D9,$B3
	.dc.b	$FE,$F4,$7E,$3B,$FE,$EF,$73,$26,$93,$5F,$C6,$67,$F8,$DD,$65,$FF
	.dc.b	$BD,$F8,$7E,$23,$67,$2F,$E7,$A7,$8B,$EB,$F9,$2F,$C8,$FC,$2C,$8C
	.dc.b	$10,$F8,$5A,$23,$51,$F1,$75,$57,$D2,$33,$A7,$E6,$AB,$15,$CE,$30
	.dc.b	$FC,$98,$CB,$CE,$78,$EB,$FC,$33,$0F,$04,$F1,$8F,$D1,$CE,$A3,$F2
	.dc.b	$F5,$CA,$FF,$3B,$39,$9F,$F5,$64,$E6,$B8,$C7,$84,$87,$84,$EC,$74
	.dc.b	$33,$5D,$1C,$9D,$0C,$C7,$47,$91,$F2,$27,$C5,$87,$C6,$8E,$38,$7B
	.dc.b	$1D,$1D,$23,$4E,$09,$7E,$17,$AB,$F2,$9C,$DF,$86,$34,$64,$8E,$49
	.dc.b	$1E,$D3,$89,$FD,$FA,$7C,$33,$CA,$7E,$95,$FA,$B0,$F9,$57,$C6,$BA
	.dc.b	$BC,$63,$2F,$C5,$3A,$9F,$C8,$1C,$E1,$F0,$67,$8C,$E4,$F8,$B5,$DD
	.dc.b	$A7,$CE,$3B,$18,$3A,$E1,$E8,$C3,$DB,$1E,$1A,$EB,$39,$8E,$27,$F2
	.dc.b	$4C,$3F,$C7,$C4,$7F,$94,$B5,$D1,$82,$1F,$4A,$BB,$E1,$FE,$5B,$BF
	.dc.b	$91,$3D,$9A,$FF,$BF,$7C,$BF,$8B,$87,$8E,$3E,$01,$FA,$E8,$CE,$1F
	.dc.b	$19,$B1,$F8,$86,$BE,$13,$F9,$BB,$8A,$EF,$7E,$3C,$3E,$1E,$1D,$CF
	.dc.b	$04,$66,$FC,$D3,$AC,$3F,$A9,$E1,$CE,$3D,$AD,$9D,$9E,$03,$F3,$13
	.dc.b	$E7,$69,$E7,$29,$E7,$3E,$11,$8C,$3C,$F3,$D6,$3E,$1C,$7F,$62,$8C
	.dc.b	$61,$E8,$6D,$79,$AB,$95,$F9,$9A,$79,$8E,$E7,$86,$BE,$14,$F1,$8C
	.dc.b	$3E,$96,$9E,$92,$9E,$9C,$3E,$46,$1F,$E0,$2F,$6B,$FB,$F7,$5C,$AF
	.dc.b	$E6,$51,$FE,$01,$30,$FF,$83,$A3,$0F,$F3,$7F,$47,$FC,$0C,$47,$F9
	.dc.b	$D7,$81,$5F,$F0,$AC,$5B,$FF,$0A,$CF,$F8,$C1,$67,$FE,$32,$F8,$97
	.dc.b	$C6,$FF,$83,$3E,$26,$63,$FD,$8A,$72,$D9,$FF,$4C,$A8,$CB,$FE,$AB
	.dc.b	$6B,$FE,$3E,$35,$1C,$23,$77,$E2,$9F,$F4,$3C,$62,$FC,$95,$FE,$1E
	.dc.b	$F8,$BF,$44,$7F,$96,$3E,$45,$F2,$58,$C1,$8A,$57,$98,$C3,$63,$8C
	.dc.b	$FF,$86,$67,$85,$66,$A4,$E6,$B6,$FF,$9C,$57,$BB,$CC,$91,$BF,$4E
	.dc.b	$BB,$8B,$87,$F9,$DA,$BF,$D9,$33,$FE,$07,$CD,$C1,$FE,$3C,$F5,$A4
	.dc.b	$FD,$18,$F7,$3F,$3D,$F9,$44,$7F,$D9,$73,$C9,$B1,$C2,$31,$1D,$E3
	.dc.b	$FC,$93,$E7,$7C,$5F,$A5,$FF,$04,$9F,$E0,$C9,$FF,$47,$37,$1F,$F4
	.dc.b	$6B,$AF,$07,$87,$6C,$AF,$95,$3C,$BC,$3E,$84,$8C,$FC,$35,$7A,$C5
	.dc.b	$C3,$AC,$FB,$BE,$53,$FC,$9D,$5F,$72,$DF,$2B,$E7,$35,$84,$9F,$F7
	.dc.b	$5D,$7F,$96,$5F,$C8,$35,$FF,$47,$B5,$FF,$6A,$BD,$93,$4F,$0E,$0B
	.dc.b	$7D,$57,$E2,$FE,$9B,$FC,$4F,$7F,$EB,$35,$E0,$E4,$D7,$85,$27,$F1
	.dc.b	$D3,$A6,$CE,$5E,$5C,$1D,$48,$F0,$ED,$AE,$8E,$71,$E3,$8F,$F6,$B7
	.dc.b	$C9,$79,$7D,$89,$5F,$ED,$F8,$CB,$FB,$33,$FD,$C8,$FE,$EB,$D3,$34
	.dc.b	$91,$C4,$87,$BF,$BF,$F2,$7F,$2F,$C2,$3A,$07,$0F,$76,$BF,$C0,$DA
	.dc.b	$FF,$B0,$97,$CA,$7F,$0C,$6C,$74,$6B,$CE,$3F,$88,$FD,$07,$EF,$A3
	.dc.b	$C2,$D7,$DC,$D7,$F8,$93,$5F,$F6,$8F,$CC,$7E,$7D,$FE,$2C,$D7,$3C
	.dc.b	$FF,$C7,$A3,$10,$0B,$E6,$34,$48,$F8,$71,$96,$FB,$9C,$C8,$E0,$F9
	.dc.b	$3F,$30,$F8,$FF,$E5,$AD,$8C,$C6,$0C,$FC,$74,$8E,$2D,$7E,$F4,$BC
	.dc.b	$9E,$64,$8E,$EF,$F3,$AF,$95,$F8,$91,$ED,$6A,$69,$8F,$07,$EC,$C7
	.dc.b	$F3,$8E,$1F,$E7,$84,$7F,$07,$3F,$AC,$6F,$67,$72,$3A,$F9,$8F,$11
	.dc.b	$1C,$1C,$BE,$2F,$CC,$47,$07,$6F,$6B,$D4,$47,$07,$8B,$D2,$FE,$53
	.dc.b	$87,$F8,$66,$3D,$9E,$96,$26,$46,$C8,$7E,$9E,$7F,$CE,$0F,$F5,$B7
	.dc.b	$FC,$A5,$E6,$FE,$0E,$3E,$34,$F2,$7F,$D8,$57,$F8,$17,$FD,$89,$5F
	.dc.b	$EC,$0F,$0B,$FB,$00,$F9,$7F,$AD,$3F,$D4,$DA,$F0,$F9,$8F,$49,$F5
	.dc.b	$C8,$FC,$E6,$BB,$3F,$DB,$35,$DF,$1F,$CC,$5F,$7F,$AA,$F8,$7F,$05
	.dc.b	$FD,$97,$17,$6D,$33,$8F,$29,$FC,$31,$1F,$BA,$FF,$80,$BE,$65,$EA
	.dc.b	$FF,$56,$6C,$FF,$BA,$9A,$FF,$B9,$5F,$E6,$06,$31,$FF,$01,$42,$31
	.dc.b	$97,$CD,$6C,$7E,$C5,$FE,$71,$D0,$F7,$BF,$B2,$B6,$7F,$DF,$2D,$7F
	.dc.b	$DE,$9F,$32,$F2,$FF,$C0,$F0,$F9,$7F,$D7,$3F,$B1,$BE,$64,$7E,$4E
	.dc.b	$3F,$7E,$F9,$65,$FC,$0E,$F8,$C6,$2B,$A6,$3F,$85,$3C,$53,$D9,$A7
	.dc.b	$3C,$7C,$71,$C9,$B1,$C7,$1F,$C7,$4F,$16,$C7,$1C,$3C,$4D,$BF,$09
	.dc.b	$FC,$5D,$EE,$D9,$E2,$FF,$1B,$C3,$D1,$7E,$1B,$E7,$3B,$29,$F1,$4B
	.dc.b	$3C,$2F,$C1,$5E,$59,$DC,$78,$14,$ED,$F0,$61,$F4,$9F,$F2,$34,$7C
	.dc.b	$96,$C6,$5F,$61,$AF,$39,$F3,$3F,$D4,$29,$18,$9F,$EB,$D1,$89,$D4
	.dc.b	$E6,$7F,$AB,$FC,$CF,$F9,$69,$34,$8F,$F9,$79,$F8,$39,$7E,$7B,$4F
	.dc.b	$87,$3F,$D0,$E7,$3B,$7F,$96,$F4,$FC,$5C,$E4,$FA,$67,$F9,$5A,$35
	.dc.b	$A7,$F7,$3A,$E3,$7D,$2B,$C6,$70,$C3,$C0,$F2,$3A,$61,$F9,$88,$CB
	.dc.b	$E0,$8E,$2E,$70,$EF,$1F,$E7,$3B,$EE,$F3,$37,$87,$FA,$16,$77,$1F
	.dc.b	$C7,$67,$75,$CF,$0F,$D1,$CF,$F0,$CA,$FE,$37,$1E,$B8,$F4,$46,$1D
	.dc.b	$47,$07,$FC,$EF,$F1,$59,$7E,$24,$67,$14,$E7,$E6,$7F,$7D,$3C,$5B
	.dc.b	$5F,$E8,$08,$FE,$65,$1F,$DD,$5D,$6F,$C4,$F1,$5B,$CA,$F9,$6F,$F1
	.dc.b	$0A,$E6,$FF,$34,$8C,$D7,$F3,$DF,$99,$FF,$3A,$BF,$E8,$7F,$99,$78
	.dc.b	$9E,$99,$CE,$5E,$33,$FD,$27,$3A,$AF,$81,$3A,$C3,$FA,$0C,$E6,$B8
	.dc.b	$09,$D5,$7F,$38,$9D,$57,$F2,$6F,$65,$E3,$7F,$28,$C3,$FA,$94,$62
	.dc.b	$FF,$C6,$6D,$8F,$F2,$03,$5F,$F5,$A3,$9A,$FE,$A7,$F3,$2F,$6F,$EA
	.dc.b	$4E,$6F,$FA,$4C,$E5,$FB,$51,$9B,$FF,$58,$CE,$5F,$5C,$66,$D3,$97
	.dc.b	$F9,$FF,$CC,$EE,$BF,$D8,$CD,$7F,$B6,$35,$FF,$18,$35,$7C,$CB,$E3
	.dc.b	$7E,$14,$7F,$8D,$63,$18,$FF,$95,$1B,$1F,$E6,$16,$BF,$ED,$5F,$87
	.dc.b	$7C,$BF,$28,$46,$2B,$FC,$BA,$D8,$FF,$35,$35,$FF,$6F,$7C,$CB,$C7
	.dc.b	$E4,$DA,$FF,$32,$B6,$7F,$CE,$4D,$7F,$DC,$DB,$7D,$57,$FC,$18,$D7
	.dc.b	$FC,$A0,$D7,$EB,$6F,$E7,$FA,$DF,$F1,$F5,$FF,$9C,$9B,$3F,$E7,$F6
	.dc.b	$BF,$EF,$36,$BF,$E1,$9F,$99,$F8,$3F,$01,$FE,$7A,$6C,$7F,$A1,$D7
	.dc.b	$C1,$7C,$F8,$C6,$AA,$FF,$40,$FC,$CF,$C1,$9B,$AF,$FA,$11,$46,$5F
	.dc.b	$F4,$93,$5F,$F1,$51,$DB,$E0,$38,$61,$FE,$B0,$8F,$F4,$6B,$5F,$F4
	.dc.b	$D3,$5E,$3F,$2F,$FE,$EC,$6E,$DE,$18,$C4,$7F,$8D,$DA,$FF,$9E,$FE
	.dc.b	$2D,$1E,$76,$C7,$FA,$71,$AF,$FA,$A5,$AF,$A3,$CD,$7D,$1F,$BD,$B1
	.dc.b	$FE,$A4,$6B,$FE,$B1,$6B,$F0,$3E,$67,$9B,$FE,$4B,$6B,$FE,$8F,$6A
	.dc.b	$6F,$73,$E0,$3E,$73,$67,$FD,$66,$D7,$FD,$80,$D7,$E7,$F6,$FF,$97
	.dc.b	$1F,$F2,$D3,$5F,$F4,$F3,$53,$79,$5F,$0F,$F7,$5B,$3F,$EC,$26,$BF
	.dc.b	$EC,$F6,$BF,$7B,$E6,$3F,$9E,$BF,$EC,$86,$BF,$ED,$56,$BB,$F1,$3F
	.dc.b	$15,$F5,$5F,$D7,$63,$FD,$64,$D4,$D3,$1B,$7F,$BF,$7C,$67,$CC,$7F
	.dc.b	$9D,$FC,$1F,$F2,$7C,$62,$FF,$AA,$57,$F9,$15,$BE,$77,$E0,$C8,$FE
	.dc.b	$0D,$AF,$F3,$88,$E0,$D8,$ED,$DB,$FE,$85,$A7,$AB,$8A,$60,$F9,$5F
	.dc.b	$D5,$47,$F6,$C8,$C7,$66,$DB,$19,$97,$C0,$E5,$E3,$1F,$D4,$67,$6D
	.dc.b	$8C,$C7,$FA,$41,$E3,$1D,$FE,$55,$F1,$5D,$D7,$C4,$D3,$11,$27,$84
	.dc.b	$65,$E5,$1E,$D5,$B7,$F2,$A7,$FC,$01,$A3,$E5,$74,$D8,$FA,$3E,$37
	.dc.b	$1A,$E1,$1F,$AF,$8F,$DA,$11,$FD,$A3,$B2,$3D,$BA,$FE,$47,$CA,$FC
	.dc.b	$0E,$1F,$F5,$0B,$67,$E8,$B7,$63,$F1,$23,$E5,$39,$27,$07,$4E,$0F
	.dc.b	$A9,$F2,$BF,$22,$BE,$C3,$F2,$1B,$BE,$B7,$91,$82,$1F,$5E,$79,$3F
	.dc.b	$8B,$F9,$5F,$8B,$EF,$C8,$FC,$AC,$7C,$53,$F8,$B1,$1F,$E2,$CD,$79
	.dc.b	$1C,$0F,$A3,$E3,$7A,$3E,$37,$D3,$E0,$49,$F4,$C6,$5F,$5C,$65,$F3
	.dc.b	$47,$C0,$F8,$69,$1F,$01,$F8,$7E,$F6,$23,$D9,$17,$83,$FD,$BF,$9F
	.dc.b	$FB,$3A,$7F,$C8,$5F,$0C,$F9,$4F,$C1,$FC,$AC,$9C,$C7,$E2,$E7,$51
	.dc.b	$FD,$18,$07,$2E,$5A,$FF,$4C,$6B,$FE,$F7,$77,$7F,$ED,$06,$F2,$66
	.dc.b	$BF,$2B,$F2,$9F,$B1,$F9,$47,$CA,$D8,$98,$FE,$DE,$72,$D9,$CB,$FB
	.dc.b	$38,$4F,$39,$E9,$EC,$4A,$E4,$F7,$F3,$71,$1C,$62,$E6,$CB,$FD,$F7
	.dc.b	$8F,$F8,$DE,$BE,$5F,$07,$EB,$3C,$E9,$18,$21,$FE,$4E,$EE,$FE,$F6
	.dc.b	$FE,$D2,$3E,$74,$C8,$C1,$0F,$F2,$AE,$C8,$D4,$73,$9E,$EE,$A3,$38
	.dc.b	$FF,$2E,$9C,$BF,$3E,$33,$8F,$F2,$59,$CB,$F5,$63,$35,$F7,$B0,$F3
	.dc.b	$BD,$CF,$5D,$F3,$7C,$CF,$77,$35,$CD,$CB,$DD,$C9,$F8,$AC,$3D,$8E
	.dc.b	$8F,$06,$1E,$27,$F1,$D8,$79,$9C,$9C,$8C,$9C,$46,$1E,$8C,$3C,$D1
	.dc.b	$F1,$4D,$46,$70,$F3,$46,$4F,$71,$27,$89,$23,$D5,$19,$78,$C6,$63
	.dc.b	$CA,$64,$F7,$57,$AE,$BC,$11,$A9,$E0,$FA,$8E,$07,$AA,$FE,$16,$1F
	.dc.b	$1A,$33,$5C,$9E,$15,$CA,$30,$66,$78,$46,$6B,$E5,$57,$0C,$3D,$27
	.dc.b	$B2,$7E,$D4,$7C,$37,$83,$F3,$6B,$9C,$F0,$8F,$53,$C4,$C3,$EB,$3F
	.dc.b	$81,$61,$FE,$53,$9F,$54,$60,$FE,$18,$6C,$FA,$66,$4F,$E1,$91,$C4
	.dc.b	$F8,$67,$F5,$AF,$26,$C7,$E0,$F1,$FF,$10,$E1,$F3,$2F,$CF,$3F,$48
	.dc.b	$FE,$C7,$5C,$1F,$76,$1F,$E0,$78,$C5,$7C,$B6,$C7,$B2,$FA,$BE,$EB
	.dc.b	$EB,$7F,$2C,$FD,$45,$70,$7C,$07,$F0,$17,$0E,$CF,$CB,$B8,$8E,$75
	.dc.b	$DA,$78,$DF,$AF,$0F,$81,$1F,$A1,$AF,$09,$C2,$F9,$4E,$A7,$DA,$72
	.dc.b	$C3,$D5,$3A,$8F,$86,$73,$9F,$CB,$46,$63,$D0,$64,$F9,$CE,$6F,$FC
	.dc.b	$76,$DA,$F9,$53,$B3,$C2,$7B,$8C,$4F,$CD,$6C,$79,$1A,$79,$CF,$96
	.dc.b	$7F,$8D,$7E,$53,$FC,$10,$7F,$81,$E7,$F9,$A4,$7F,$8F,$23,$13,$FD
	.dc.b	$5A,$51,$C3,$B3,$FC,$78,$EA,$B8,$76,$62,$73,$E2,$C9,$F0,$3C,$3E
	.dc.b	$07,$11,$FE,$3F,$BF,$F2,$8E,$1E,$23,$BD,$6E,$7C,$58,$FF,$68,$5F
	.dc.b	$17,$C2,$3C,$D6,$F9,$5F,$29,$FE,$3B,$9D,$57,$F7,$07,$FC,$BE,$FF
	.dc.b	$AE,$1A,$F5,$6B,$BB,$FF,$61,$47,$28,$F0,$C7,$53,$E1,$FA,$7F,$E5
	.dc.b	$FF,$95,$FF,$11,$E5,$EB,$7F,$C5,$95,$FE,$5D,$6C,$67,$C1,$E0,$71
	.dc.b	$F0,$A5,$7F,$9A,$EB,$9E,$9F,$4D,$FD,$9D,$9A,$4E,$FF,$17,$14,$BE
	.dc.b	$4F,$96,$3F,$CB,$DF,$2A,$F8,$FF,$47,$A7,$2F,$E7,$9F,$E1,$69,$D3
	.dc.b	$80,$E1,$FF,$0E,$35,$FF,$71,$B5,$FB,$0D,$7A,$4F,$88,$8F,$BF,$C2
	.dc.b	$FF,$44,$7D,$47,$F8,$42,$3F,$D1,$8B,$C1,$C9,$AF,$02,$4E,$A3,$2F
	.dc.b	$08,$E7,$DF,$23,$9B,$DB,$C8,$C4,$7F,$A4,$6F,$FC,$28,$D9,$D3,$F0
	.dc.b	$7C,$47,$D1,$F9,$BC,$F9,$39,$8F,$F4,$54,$EA,$3F,$9A,$B5,$FF,$21
	.dc.b	$35,$FC,$E0,$70,$FE,$7D,$E1,$C5,$FB,$68,$D4,$F6,$3F,$D5,$DF,$4E
	.dc.b	$F9,$5B,$DB,$8F,$AE,$97,$FE,$B5,$A4,$EA,$31,$3A,$7F,$D6,$53,$A7
	.dc.b	$87,$57,$62,$3C,$F9,$EB,$C2,$2E,$5F,$BB,$AF,$CB,$4F,$E2,$7E,$A3
	.dc.b	$F6,$CF,$FB,$1F,$E4,$5F,$09,$FE,$C7,$6C,$6F,$C0,$FF,$5A,$57,$08
	.dc.b	$FF,$35,$35,$F9,$2D,$7F,$83,$35,$FF,$37,$BC,$70,$F9,$4D,$EC,$DD
	.dc.b	$78,$7C,$BF,$F6,$17,$CD,$3E,$6B,$FB,$6E,$28,$C3,$FD,$C1,$AF,$F9
	.dc.b	$F5,$AF,$DC,$F9,$AF,$CD,$3F,$E8,$16,$B9,$E9,$EB,$18,$8F,$E3,$CB
	.dc.b	$D2,$C9,$23,$FB,$FC,$65,$FF,$00,$FC,$D7,$E0,$5B,$CB,$E8,$8C,$47
	.dc.b	$FA,$45,$AF,$E4,$17,$BD,$B2,$47,$47,$E0,$FC,$D7,$C2,$83,$48,$EB
	.dc.b	$E3,$63,$FE,$14,$84,$62,$3F,$D4,$51,$F9,$F7,$5D,$5D,$88,$FF,$87
	.dc.b	$3E,$61,$F3,$FF,$49,$B1,$98,$C1,$FE,$25,$F0,$E4,$7F,$89,$3D,$0F
	.dc.b	$E2,$0F,$FA,$BE,$39,$79,$92,$3B,$35,$FE,$37,$F3,$DF,$80,$F1,$CD
	.dc.b	$24,$79,$7E,$6F,$C8,$47,$07,$A7,$C3,$F4,$11,$C1,$FE,$38,$7F,$DF
	.dc.b	$0F,$F0,$26,$BF,$D7,$DA,$FF,$B2,$1A,$B3,$FC,$13,$FE,$CD,$6B,$FC
	.dc.b	$31,$AF,$F6,$96,$BF,$ED,$27,$8F,$5F,$F0,$D6,$F5,$6F,$97,$FB,$84
	.dc.b	$FF,$11,$BF,$C5,$B7,$93,$F4,$B1,$F8,$CC,$3D,$11,$83,$EC,$78,$D2
	.dc.b	$3E,$E3,$5F,$43,$DB,$0F,$33,$63,$CC,$FF,$05,$F9,$BC,$47,$C1,$8B
	.dc.b	$91,$A6,$E3,$FA,$C4,$3E,$6F,$F1,$44,$8F,$C7,$35,$65,$F0,$E3,$A9
	.dc.b	$E3,$7F,$C2,$38,$78,$BE,$6F,$FD,$4F,$5F,$D9,$5A,$F2,$8F,$F0,$C4
	.dc.b	$62,$7F,$9D,$3F,$E0,$B3,$11,$AF,$5B,$A1,$D7,$D2,$F7,$FC,$CB,$E5
	.dc.b	$E9,$31,$FE,$E9,$F3,$6A,$F4,$C3,$E6,$57,$FE,$29,$39,$47,$F7,$B6
	.dc.b	$BE,$70,$70,$BF,$9F,$7D,$73,$CD,$F4,$AD,$CE,$5F,$05,$F7,$BF,$F1
	.dc.b	$C5,$6E,$7A,$1C,$A7,$FC,$79,$5C,$67,$4D,$8E,$53,$B3,$CA,$FA,$27
	.dc.b	$78,$7C,$06,$E1,$BA,$EE,$79,$0E,$38,$79,$2F,$C3,$7C,$5B,$5C,$5E
	.dc.b	$C7,$8A,$B9,$0C,$3C,$CF,$C2,$AE,$4F,$63,$95,$F7,$75,$7E,$5A,$F9
	.dc.b	$11,$FE,$02,$6B,$E2,$3F,$C0,$A7,$7A,$FB,$AE,$FC,$C7,$CD,$FF,$B0
	.dc.b	$AB,$FB,$EB,$63,$33,$CA,$72,$FF,$5C,$F9,$BF,$E7,$84,$70,$4C,$FE
	.dc.b	$C6,$2D,$3F,$D0,$F1,$F6,$E7,$F0,$F5,$F3,$EB,$5A,$FD,$19,$C9,$F2
	.dc.b	$23,$2F,$CA,$33,$5D,$AF,$AE,$BF,$03,$0E,$91,$8A,$FF,$13,$C6,$30
	.dc.b	$ED,$08,$C3,$D4,$C1,$E1,$9E,$4F,$FA,$2E,$37,$7E,$B8,$DB,$E3,$AE
	.dc.b	$78,$7F,$8C,$4F,$49,$FE,$33,$9F,$4E,$1F,$C9,$B0,$FF,$4D,$55,$3F
	.dc.b	$CD,$14,$F8,$CC,$DF,$FA,$B6,$AB,$FC,$42,$2F,$CD,$FE,$FE,$FF,$CE
	.dc.b	$0D,$9F,$DB,$07,$F6,$F1,$AD,$F8,$9D,$B0,$7A,$DE,$89,$FF,$55,$57
	.dc.b	$39,$FF,$58,$4E,$6F,$F8,$0F,$CD,$FF,$88,$6F,$FD,$00,$D8,$FE,$0D
	.dc.b	$4F,$90,$7D,$4F,$AA,$B5,$3F,$E5,$77,$85,$7F,$97,$9A,$FF,$AE,$DA
	.dc.b	$F8,$3C,$BF,$98,$FF,$12,$F8,$47,$95,$F1,$9B,$97,$FA,$5B,$07,$CE
	.dc.b	$3E,$13,$57,$2B,$FD,$31,$7C,$30,$F4,$DE,$E7,$35,$B9,$FD,$7D,$6D
	.dc.b	$FF,$3A,$B5,$FF,$69,$35,$FE,$4F,$F3,$9F,$D4,$6B,$F6,$F0,$8C,$3F
	.dc.b	$BA,$6B,$FE,$80,$6B,$FE,$DD,$FC,$53,$FD,$B5,$3C,$9B,$1F,$C0,$9A
	.dc.b	$FF,$A2,$1A,$FF,$B9,$7F,$14,$7C,$17,$CA,$84,$62,$BC,$AD,$8F,$E1
	.dc.b	$AD,$7F,$D2,$7F,$4E,$FA,$8E,$4D,$AF,$6B,$67,$F8,$AB,$5F,$F4,$DF
	.dc.b	$D3,$7F,$A6,$E7,$E1,$B6,$3F,$8D,$B5,$DF,$9F,$E7,$F9,$E7,$CF,$3E
	.dc.b	$2B,$E5,$C6,$1B,$C5,$FE,$86,$BF,$F5,$4C,$E5,$FE,$4F,$E6,$BE,$4B
	.dc.b	$FC,$F2,$D8,$FA,$AD,$77,$1A,$C6,$46,$9A,$F4,$F8,$7E,$64,$79,$BF
	.dc.b	$E8,$6F,$A8,$FF,$7F,$C7,$FA,$F1,$AF,$F6,$C0,$E1,$FE,$DE,$D7,$3E
	.dc.b	$E7,$32,$3F,$8C,$84,$77,$F0,$A4,$E5,$AF,$F4,$8F,$9C,$FF,$09,$BF
	.dc.b	$EC,$E6,$BE,$36,$BF,$E0,$06,$A7,$F8,$C3,$F9,$C6,$BD,$7C,$2F,$F5
	.dc.b	$2C,$62,$3B,$79,$38,$23,$C5,$FF,$55,$FB,$5F,$B7,$8F,$D3,$B5,$FE
	.dc.b	$BE,$D7,$FD,$CA,$D7,$E5,$3F,$86,$78,$F0,$F4,$23,$D3,$E4,$7B,$91
	.dc.b	$C1,$FB,$6F,$97,$FF,$50,$BF,$C2,$FE,$9B,$FD,$2E,$DF,$77,$FC,$9F
	.dc.b	$3A,$8F,$E6,$2D,$7F,$DF,$2D,$7E,$F3,$5F,$F1,$5B,$53,$CF,$C1,$FF
	.dc.b	$7E,$37,$A3,$8E,$5F,$E5,$EF,$A6,$7D,$4F,$8A,$26,$1F,$E3,$6C,$38
	.dc.b	$C6,$E7,$33,$E6,$8B,$2B,$FC,$95,$F4,$AE,$51,$8B,$DB,$E2,$8C,$9F
	.dc.b	$E4,$49,$E0,$FF,$52,$6D,$78,$A3,$67,$BA,$FC,$8E,$27,$75,$D6,$3C
	.dc.b	$C6,$6F,$6F,$F8,$82,$7B,$BF,$A9,$0E,$1F,$41,$FC,$C1,$EA,$F9,$E3
	.dc.b	$9D,$FC,$18,$FF,$2E,$FD,$23,$EA,$F3,$87,$29,$F6,$5F,$FA,$3A,$B5
	.dc.b	$1E,$52,$1A,$3D,$EE,$9F,$EE,$AD,$BF,$04,$7A,$27,$BB,$E2,$BE,$C6
	.dc.b	$1D,$1D,$1C,$BE,$09,$E7,$87,$BA,$17,$DF,$0F,$E1,$CF,$8E,$3E,$43
	.dc.b	$5C,$DF,$F5,$67,$8B,$63,$94,$79,$4F,$11,$F7,$3F,$00,$FE,$51,$87
	.dc.b	$B2,$77,$3C,$8D,$9F,$E1,$69,$DD,$F8,$23,$35,$C6,$FF,$D7,$75,$A9
	.dc.b	$FF,$48,$B5,$E8,$7C,$C8,$DF,$44,$8E,$77,$D5,$F0,$46,$5E,$06,$8E
	.dc.b	$AE,$5F,$C6,$57,$F4,$36,$C7,$C5,$8C,$47,$F3,$97,$C8,$48,$FF,$69
	.dc.b	$FE,$30,$F1,$7C,$18,$FE,$02,$B5,$7F,$A4,$AD,$3F,$2E,$F7,$87,$57
	.dc.b	$9E,$12,$74,$43,$FC,$83,$5C,$9B,$1C,$9F,$69,$DD,$E4,$FC,$43,$96
	.dc.b	$1F,$94,$9E,$4D,$9E,$52,$C3,$E9,$56,$B0,$FE,$B4,$D9,$EE,$75,$3C
	.dc.b	$67,$84,$EF,$1E,$D7,$06,$8E,$17,$F1,$F0,$F3,$BE,$38,$E2,$76,$70
	.dc.b	$F1,$78,$3A,$15,$E5,$6E,$1E,$CC,$3D,$6F,$6C,$3C,$93,$E2,$7D,$B1
	.dc.b	$98,$F0,$46,$AF,$C2,$47,$04,$8F,$F7,$4D,$F8,$4F,$15,$FE,$0A,$3D
	.dc.b	$13,$D9,$F4,$47,$91,$F8,$B5,$EC,$7C,$87,$4B,$F2,$3F,$57,$0F,$AF
	.dc.b	$87,$79,$F7,$46,$A3,$C2,$EA,$38,$DF,$8E,$FD,$57,$D4,$F1,$D7,$1A
	.dc.b	$E7,$1E,$4C,$3D,$0F,$90,$F1,$19,$39,$19,$BF,$63,$F0,$0F,$96,$43
	.dc.b	$EA,$9C,$4E,$67,$5C,$3D,$8F,$F9,$B1,$FD,$03,$FD,$0D,$AF,$90,$C5
	.dc.b	$7F,$1D,$8C,$C7,$90,$C9,$CE,$3D,$CF,$ED,$7D,$BA,$F9,$CF,$C5,$B8
	.dc.b	$C7,$E3,$C6,$0F,$45,$79,$9F,$DA,$CE,$F0,$F8,$F5,$FA,$F8,$C4,$FC
	.dc.b	$68,$C3,$CA,$FE,$C4,$EA,$3E,$29,$D0,$D1,$E0,$75,$7F,$8F,$6C,$FF
	.dc.b	$14,$6C,$FB,$4E,$27,$B6,$51,$A8,$EC,$66,$7C,$6F,$0A,$F8,$35,$C2
	.dc.b	$31,$7E,$E3,$F8,$2D,$F3,$AF,$6C,$F3,$3D,$07,$03,$E1,$4F,$07,$07
	.dc.b	$C0,$8E,$27,$82,$FC,$71,$CC,$F5,$57,$7A,$E5,$19,$AF,$3C,$E6,$3F
	.dc.b	$6A,$FC,$57,$BD,$FA,$DF,$83,$3D,$DF,$90,$62,$BD,$CF,$A0,$E6,$68
	.dc.b	$FC,$A5,$7E,$44,$8F,$E6,$5F,$03,$CC,$0E,$27,$D5,$EA,$FE,$0D,$FB
	.dc.b	$1F,$F0,$5C,$EA,$BC,$B1,$A7,$C6,$1D,$D7,$FA,$45,$B1,$C2,$BA,$1F
	.dc.b	$E9,$5A,$E9,$1C,$63,$15,$E3,$77,$1E,$89,$DD,$74,$8F,$61,$FA,$53
	.dc.b	$91,$BA,$F1,$57,$48,$FF,$73,$D7,$81,$FF,$1F,$35,$D1,$FE,$41,$3E
	.dc.b	$34,$F2,$7A,$1F,$E5,$08,$F1,$CF,$AF,$83,$53,$C0,$D5,$62,$B7,$1E
	.dc.b	$B9,$C4,$C8,$D9,$D6,$30,$FB,$0E,$85,$9E,$6F,$DB,$8D,$DF,$C5,$25
	.dc.b	$F9,$67,$A9,$C6,$F7,$19,$8C,$1A,$9E,$2F,$9C,$F2,$19,$8F,$8F,$1F
	.dc.b	$86,$AF,$71,$D2,$7E,$39,$E7,$72,$F1,$33,$1C,$5E,$A7,$88,$8E,$5E
	.dc.b	$47,$13,$E1,$B9,$24,$F9,$09,$1E,$93,$89,$F7,$6F,$FD,$4C,$4A,$FF
	.dc.b	$2B,$3F,$EA,$58,$E1,$5D,$1F,$F2,$3F,$9D,$F7,$BE,$1D,$DF,$FE,$5D
	.dc.b	$6C,$FE,$24,$38,$7F,$18,$D7,$FD,$1C,$A7,$91,$23,$94,$65,$BC,$7D
	.dc.b	$C8,$E0,$F0,$7C,$33,$EB,$3F,$CE,$0D,$8C,$C6,$0C,$EB,$23,$8B,$5F
	.dc.b	$D1,$4B,$E9,$F9,$92,$3B,$BF,$E7,$DF,$88,$7A,$54,$93,$F0,$E3,$2D
	.dc.b	$FA,$EE,$04,$70,$6F,$E4,$FA,$11,$C1,$F3,$7E,$01,$F2,$DF,$E8,$D6
	.dc.b	$C6,$63,$06,$7E,$B2,$47,$16,$BF,$BF,$97,$C3,$F3,$24,$77,$7F,$CE
	.dc.b	$DE,$BC,$8F,$BD,$D6,$8C,$47,$E3,$7D,$6F,$C5,$91,$FE,$22,$FF,$A8
	.dc.b	$FF,$02,$67,$49,$3F,$89,$8C,$BF,$8B,$EB,$FA,$4D,$E0,$FF,$56,$46
	.dc.b	$23,$F8,$C3,$5F,$F6,$D2,$F2,$3F,$8A,$BB,$E9,$FF,$63,$B7,$B3,$FA
	.dc.b	$2D,$7F,$76,$AD,$9D,$88,$F8,$DF,$0F,$CB,$7F,$25,$B7,$C5,$BE,$3B
	.dc.b	$0D,$BF,$F7,$BE,$2B,$F1,$9A,$A9,$3A,$8C,$BF,$05,$EE,$D3,$E2,$7A
	.dc.b	$DD,$BE,$7A,$47,$68,$F7,$98,$7F,$A2,$E3,$E6,$9C,$C7,$F6,$48,$FF
	.dc.b	$74,$FC,$3B,$E0,$BF,$9D,$B6,$75,$AF,$C6,$8C,$47,$F3,$55,$E7,$70
	.dc.b	$A9,$3C,$0F,$EE,$33,$C2,$FB,$D7,$F4,$8F,$41,$F0,$54,$97,$C4,$7C
	.dc.b	$2F,$6B,$FE,$F7,$F0,$8B,$FB,$27,$2A,$E0,$D7,$FA,$53,$65,$9F,$E0
	.dc.b	$2F,$A5,$2F,$A3,$FE,$08,$8C,$C6,$23,$C1,$87,$E2,$23,$18,$77,$75
	.dc.b	$5C,$23,$B6,$3F,$E0,$7B,$ED,$1E,$8B,$E4,$D8,$F6,$BC,$31,$B1,$E3
	.dc.b	$AE,$58,$78,$2F,$8C,$FF,$87,$AF,$C4,$71,$7B,$9C,$6F,$9B,$EA,$2C
	.dc.b	$70,$74,$70,$3E,$8D,$F8,$23,$CD,$19,$78,$3A,$9E,$B8,$7C,$C8,$FD
	.dc.b	$69,$93,$FC,$1C,$E5,$F3,$1A,$5F,$0C,$F8,$7F,$E9,$91,$86,$C6,$67
	.dc.b	$9B,$5F,$EA,$5F,$0F,$FD,$4D,$FF,$20,$B5,$34,$8F,$BE,$FF,$45,$A7
	.dc.b	$C6,$8E,$26,$76,$FC,$4B,$F1,$4C,$EB,$F8,$5A,$E3,$19,$3B,$9C,$6B
	.dc.b	$A3,$89,$E1,$5D,$1F,$C7,$47,$4C,$FF,$C9,$93,$97,$FC,$6D,$87,$DA
	.dc.b	$C3,$BD,$F0,$BE,$95,$C3,$0E,$F7,$BC,$39,$57,$D4,$8C,$4C,$8F,$F1
	.dc.b	$86,$1E,$A3,$DE,$FF,$26,$9D,$46,$0D,$4F,$F9,$87,$97,$E2,$C6,$6E
	.dc.b	$CE,$4F,$DE,$7C,$3F,$F0,$06,$EF,$AC,$7F,$CE,$F3,$FE,$78,$9C,$92
	.dc.b	$3F,$8D,$F9,$3E,$07,$75,$C7,$0F,$F0,$66,$1C,$E5,$F0,$DF,$D5,$F0
	.dc.b	$FF,$0B,$36,$3F,$CC,$0D,$7F,$2E,$D7,$FA,$2B,$BA,$F6,$FC,$3B,$C4
	.dc.b	$ED,$39,$BF,$E5,$93,$97,$F8,$D7,$A0,$EF,$F2,$36,$3F,$93,$60,$F8
	.dc.b	$77,$C0,$6A,$E5,$7F,$8B,$6F,$86,$7F,$E7,$76,$BF,$AE,$6B,$FD,$71
	.dc.b	$DE,$1F,$53,$07,$C4,$F4,$AF,$E8,$B1,$8B,$FE,$61,$F0,$EF,$91,$FC
	.dc.b	$0D,$7F,$66,$8C,$4F,$F8,$FD,$B1,$FE,$89,$6B,$FC,$03,$E1,$DF,$11
	.dc.b	$CA,$3F,$D6,$71,$8A,$FF,$24,$36,$3F,$D2,$4D,$79,$71,$F1,$51,$FE
	.dc.b	$49,$9E,$33,$98,$C4,$FF,$A6,$FE,$1D,$F0,$BC,$5B,$7F,$E5,$76,$CE
	.dc.b	$E7,$B7,$C3,$3E,$09,$F0,$FF,$E5,$D8,$C3,$78,$DF,$29,$9A,$DE,$1F
	.dc.b	$EA,$BF,$AB,$7D,$17,$F1,$36,$C7,$F9,$9D,$AE,$E3,$F8,$C7,$B9,$E8
	.dc.b	$8F,$F3,$63,$5F,$F5,$B3,$5F,$E5,$AD,$7F,$C3,$6B,$E2,$B9,$92,$39
	.dc.b	$BF,$C9,$7E,$E7,$FC,$51,$1F,$E2,$26,$BF,$EE,$86,$A6,$FC,$5B,$E2
	.dc.b	$FF,$97,$36,$7F,$D0,$4D,$7F,$D9,$8D,$7F,$A1,$BF,$C5,$DF,$F7,$AB
	.dc.b	$53,$4C,$7D,$CF,$C6,$8F,$E9,$6D,$7F,$C7,$2D,$7F,$DF,$ED,$5F,$19
	.dc.b	$FE,$43,$7F,$D2,$CD,$7F,$DC,$0D,$7F,$AB,$B5,$FF,$22,$BF,$E1,$E7
	.dc.b	$8F,$97,$FD,$36,$31,$1C,$A3,$FA,$87,$33,$E5,$29,$27,$FD,$4D,$19
	.dc.b	$7F,$D5,$1F,$0B,$F8,$F6,$F0,$7F,$59,$8C,$47,$FB,$D5,$AF,$F6,$C5
	.dc.b	$F0,$BB,$24,$74,$7F,$B1,$7D,$67,$9A,$3F,$B9,$B5,$FF,$30,$B5,$07
	.dc.b	$0B,$E2,$9F,$15,$FF,$00,$3E,$2D,$F1,$14,$FF,$35,$F8,$9F,$D2,$2B
	.dc.b	$77,$FE,$3B,$6B,$FE,$04,$3A,$5F,$F7,$78,$CC,$7F,$82,$1A,$FF,$9D
	.dc.b	$9E,$33,$D6,$7B,$F7,$65,$B5,$EA,$F7,$DF,$17,$88,$EE,$78,$E7,$18
	.dc.b	$74,$BF,$F4,$1D,$F0,$AD,$10,$61,$E1,$7F,$CA,$D1,$B7,$33,$CB,$0E
	.dc.b	$B8,$49,$DC,$78,$EB,$85,$F3,$9E,$27,$D2,$C3,$C8,$FA,$EB,$81,$CD
	.dc.b	$D9,$EC,$2D,$74,$C3,$C5,$1F,$EB,$89,$FF,$12,$4F,$9E,$3F,$C6,$AD
	.dc.b	$76,$72,$8E,$EF,$F9,$CF,$E2,$DE,$46,$23,$47,$B0,$D3,$9C,$7F,$C1
	.dc.b	$37,$E3,$9F,$51,$A2,$1C,$89,$1F,$E1,$0A,$EF,$1A,$8E,$77,$D7,$0F
	.dc.b	$F5,$CD,$6A,$BE,$1B,$6B,$71,$E5,$33,$87,$E0,$E7,$84,$7F,$95,$9A
	.dc.b	$F4,$8F,$F2,$54,$62,$3F,$4A,$F5,$AD,$47,$95,$F1,$FB,$9D,$6A,$4E
	.dc.b	$4E,$31,$87,$C6,$64,$52,$3C,$C6,$EF,$E2,$56,$2B,$EA,$4E,$6F,$C7
	.dc.b	$1F,$0A,$34,$F6,$25,$7C,$D8,$CC,$7C,$17,$DE,$47,$F6,$C7,$91,$E7
	.dc.b	$1C,$63,$FD,$6D,$BB,$E2,$73,$7F,$0A,$7C,$A6,$A3,$2F,$C2,$32,$7E
	.dc.b	$6A,$79,$3B,$33,$5D,$DD,$4F,$EE,$E7,$71,$EF,$37,$5F,$EF,$C8,$CD
	.dc.b	$7F,$96,$9B,$1E,$59,$E0,$74,$3C,$06,$27,$C0,$71,$AF,$01,$EA,$8F
	.dc.b	$88,$70,$8D,$3E,$23,$DC,$75,$23,$D2,$7C,$4F,$08,$CC,$60,$8F,$F3
	.dc.b	$4A,$F4,$B6,$BC,$58,$FE,$1E,$3C,$55,$F0,$67,$38,$7A,$27,$27,$93
	.dc.b	$0F,$13,$CC,$C9,$EC,$3E,$15,$73,$73,$7D,$DC,$9C,$4F,$49,$D6,$3D
	.dc.b	$2F,$F5,$6B,$F1,$3C,$0E,$D5,$F0,$49,$3C,$2B,$B4,$62,$7B,$3E,$87
	.dc.b	$8C,$F9,$5F,$09,$83,$C4,$70,$75,$1C,$67,$AC,$60,$93,$C0,$E4,$7A
	.dc.b	$4F,$83,$7F,$81,$C3,$E9,$C6,$4F,$65,$74,$7C,$04,$39,$B9,$78,$5F
	.dc.b	$B5,$CD,$7C,$68,$C4,$EC,$E1,$5E,$03,$C2,$E6,$78,$5F,$C3,$BF,$F7
	.dc.b	$8D,$7E,$99,$B8,$7F,$B9,$63,$11,$B2,$1F,$6C,$D9,$FC,$09,$F0,$47
	.dc.b	$68,$FC,$A7,$0D,$7D,$75,$F2,$9F,$F0,$AC,$66,$B8,$C7,$F0,$88,$C5
	.dc.b	$65,$B1,$F9,$29,$46,$23,$EE,$C6,$1D,$4F,$16,$9C,$67,$86,$3F,$36
	.dc.b	$B8,$4F,$2A,$E0,$D3,$81,$C8,$EF,$5C,$AF,$D9,$5C,$AF,$A5,$F2,$6D
	.dc.b	$F2,$3A,$1C,$8E,$8F,$8E,$FA,$47,$E1,$6B,$A5,$F6,$BE,$8D,$9E,$97
	.dc.b	$D9,$E8,$76,$7A,$3D,$A3,$CF,$7D,$A3,$EA,$9F,$60,$D1,$E1,$75,$7F
	.dc.b	$8D,$6D,$FF,$2F,$6C,$FD,$93,$89,$D0,$FB,$E6,$27,$93,$B3,$CD,$5F
	.dc.b	$CD,$67,$A4,$6D,$AE,$8F,$31,$FD,$15,$F2,$9A,$BF,$84,$7F,$3D,$71
	.dc.b	$3F,$1E,$78,$1E,$13,$F4,$71,$FA,$12,$3F,$9D,$78,$BE,$09,$FF,$0C
	.dc.b	$F1,$E2,$3D,$A7,$43,$FC,$5E,$E6,$FE,$CC,$71,$BF,$F2,$5C,$E6,$F9
	.dc.b	$6A,$8C,$57,$C5,$C3,$BB,$6F,$E3,$34,$FF,$2E,$4F,$F9,$24,$E2,$F8
	.dc.b	$8F,$F2,$9B,$89,$E7,$19,$3D,$0F,$C7,$76,$43,$33,$E0,$78,$47,$23
	.dc.b	$83,$A3,$C6,$71,$72,$48,$F8,$31,$97,$A4,$66,$38,$9E,$53,$FC,$A7
	.dc.b	$3F,$E5,$62,$3F,$E1,$2A,$F9,$0F,$33,$8C,$AB,$FD,$33,$0F,$98,$7D
	.dc.b	$8F,$FA,$07,$E3,$78,$CF,$FA,$45,$AF,$00,$E1,$E6,$D7,$FD,$4A,$FF
	.dc.b	$A8,$1F,$F1,$2B,$5E,$9C,$7F,$E2,$87,$9B,$7C,$E7,$C4,$D2,$4F,$F8
	.dc.b	$BA,$32,$FF,$8C,$3E,$33,$F9,$AB,$FE,$B4,$6A,$69,$3E,$13,$C0,$FF
	.dc.b	$BE,$08,$E3,$E7,$24,$F5,$6B,$FE,$B3,$F8,$C7,$AD,$E9,$6C,$7F,$90
	.dc.b	$9A,$EE,$35,$F1,$12,$34,$D7,$A7,$8D,$CC,$8F,$37,$E3,$7C,$5B,$E3
	.dc.b	$29,$27,$FC,$A3,$19,$6F,$C6,$70,$23,$83,$7D,$5D,$08,$E0,$FE,$B5
	.dc.b	$B7,$F6,$12,$38,$3F,$CB,$FF,$16,$CB,$ED,$36,$33,$18,$23,$FE,$E8
	.dc.b	$E6,$FC,$93,$FE,$EA,$9E,$0F,$FA,$D9,$B1,$FE,$A4,$FA,$0F,$F6,$A6
	.dc.b	$0F,$80,$BC,$33,$0E,$9D,$1F,$9C,$AD,$46,$0F,$F7,$BB,$87,$FD,$D8
	.dc.b	$D7,$F3,$8D,$79,$57,$E5,$63,$86,$1E,$89,$FF,$48,$4F,$90,$F2,$F3
	.dc.b	$F9,$BE,$2B,$FD,$DB,$4F,$8C,$7C,$66,$1B,$87,$94,$FF,$4D,$57,$17
	.dc.b	$FD,$17,$8A,$3C,$44,$8C,$BB,$7F,$DB,$B1,$B7,$E0,$5F,$87,$D4,$F9
	.dc.b	$3F,$19,$E1,$F8,$C9,$5E,$1B,$4E,$A3,$C2,$FC,$12,$3F,$E9,$EF,$8C
	.dc.b	$7C,$27,$F0,$AC,$7E,$24,$62,$7D,$8D,$7F,$DE,$2B,$CF,$E1,$C1,$FC
	.dc.b	$4A,$35,$5C,$AF,$36,$13,$AA,$E1,$7A,$97,$C4,$3E,$1B,$0D,$8F,$9C
	.dc.b	$D5,$8F,$DB,$F8,$CF,$23,$DF,$E2,$B9,$CF,$86,$B8,$4F,$E0,$5B,$8F
	.dc.b	$F2,$A6,$CF,$9E,$F9,$AB,$E6,$A3,$31,$89,$F2,$37,$0F,$E6,$AF,$09
	.dc.b	$DC,$FF,$38,$6D,$79,$6B,$73,$E3,$6C,$78,$F1,$F7,$BE,$39,$E5,$18
	.dc.b	$BF,$04,$FC,$19,$E4,$D3,$96,$35,$F5,$57,$19,$DC,$75,$3C,$86,$F0
	.dc.b	$F3,$BC,$2B,$94,$FC,$03,$8D,$FB,$EF,$A5,$78,$0F,$EA,$07,$12,$DF
	.dc.b	$CE,$7F,$AA,$35,$F8,$DF,$16,$F8,$CF,$F7,$AC,$A3,$13,$C9,$AF,$26
	.dc.b	$BE,$0F,$8C,$FE,$AC,$9A,$47,$FA,$D3,$F1,$72,$FD,$2E,$9F,$1E,$7F
	.dc.b	$35,$39,$DB,$F9,$0E,$7F,$84,$9C,$9F,$24,$FC,$AC,$6B,$4F,$14,$67
	.dc.b	$0E,$AE,$CE,$35,$F8,$E8,$CB,$D6,$3D,$07,$1A,$E8,$FF,$7B,$BF,$26
	.dc.b	$1E,$78,$DD,$73,$AD,$C7,$F8,$22,$B6,$F8,$4D,$E1,$FC,$0E,$77,$87
	.dc.b	$98,$FE,$50,$5C,$3D,$78,$75,$7F,$91,$46,$6F,$FA,$37,$3F,$F4,$68
	.dc.b	$E3,$76,$BF,$88,$4D,$AF,$E0,$DF,$19,$FC,$06,$7F,$96,$36,$3F,$7F
	.dc.b	$1F,$E2,$08,$FE,$18,$EB,$7E,$52,$9E,$18,$78,$7C,$BF,$24,$E6,$BC
	.dc.b	$13,$FE,$55,$C7,$8F,$C6,$3E,$37,$3F,$1B,$FE,$06,$CF,$F9,$F4,$E5
	.dc.b	$FE,$6B,$19,$C7,$CA,$7F,$1F,$9D,$38,$C1,$3B,$7F,$A9,$4E,$9F,$8D
	.dc.b	$19,$AF,$F1,$EF,$A6,$F0,$BF,$AC,$61,$86,$BF,$CB,$63,$53,$FA,$E8
	.dc.b	$46,$1F,$D8,$B6,$FE,$A7,$C6,$3E,$37,$E4,$CE,$6F,$FB,$F4,$E5,$FE
	.dc.b	$BD,$19,$BF,$E6,$33,$97,$E0,$9E,$0C,$3E,$6C,$67,$0F,$EE,$1F,$14
	.dc.b	$F8,$BF,$4D,$A3,$18,$FA,$5B,$1F,$C2,$1A,$FF,$96,$5F,$F4,$64,$6F
	.dc.b	$83,$9C,$62,$78,$47,$6F,$A8,$7A,$D4,$97,$F1,$E3,$2F,$C9,$F8,$A7
	.dc.b	$BE,$F8,$AC,$37,$1F,$E3,$72,$F0,$DF,$15,$AB,$FA,$5F,$14,$F8,$AF
	.dc.b	$F3,$33,$70,$CC,$60,$FF,$35,$5B,$D5,$A4,$9F,$B9,$19,$7E,$F7,$C5
	.dc.b	$3D,$4F,$F3,$93,$63,$31,$83,$FC,$ED,$F1,$4F,$3E,$92,$7F,$11,$19
	.dc.b	$7F,$15,$F6,$8F,$90,$FF,$3F,$36,$33,$18,$3F,$D0,$5F,$17,$FC,$5E
	.dc.b	$3F,$C0,$0D,$4D,$31,$F1,$7F,$0E,$3F,$A2,$B5,$FF,$4B,$35,$FF,$4F
	.dc.b	$B5,$FF,$06,$76,$FF,$88,$1F,$D0,$35,$FE,$9E,$D7,$FD,$40,$D7,$FD
	.dc.b	$53,$F1,$7F,$E1,$B7,$F5,$0D,$7F,$AB,$B5,$FF,$54,$35,$FF,$58,$F4
	.dc.b	$3E,$2F,$FD,$44,$D8,$FD,$93,$5F,$EC,$2D,$7F,$D6,$5E,$07,$ED,$9F
	.dc.b	$DC,$35,$FE,$CE,$D7,$7E,$07,$FA,$BF,$E1,$3F,$37,$1F,$BF,$6B,$FD
	.dc.b	$B9,$AF,$FA,$FD,$AF,$FB,$23,$E2,$FD,$4F,$F8,$D5,$A9,$A6,$3E,$B9
	.dc.b	$F2,$9F,$EB,$F6,$CF,$F0,$F6,$BF,$DF,$9A,$FF,$B3,$FE,$27,$D9,$E1
	.dc.b	$FF,$9C,$7E,$21,$F1,$5F,$E4,$16,$D7,$F9,$AE,$B7,$E0,$FF,$B7,$9F
	.dc.b	$85,$17,$4F,$F6,$84,$62,$3F,$C9,$CD,$4F,$27,$31,$FE,$4F,$8C,$46
	.dc.b	$BD,$7C,$4C,$8F,$13,$B8,$C4,$7F,$1E,$9E,$24,$8D,$47,$02,$1B,$72
	.dc.b	$F1,$9E,$78,$FF,$A4,$E9,$5A,$8C,$4E,$AF,$E1,$4F,$2B,$F9,$8F,$C2
	.dc.b	$23,$FD,$06,$F9,$BF,$2E,$3A,$4F,$F9,$CE,$3F,$91,$FC,$56,$1F,$93
	.dc.b	$35,$C9,$C4,$87,$AC,$FF,$50,$78,$A9,$FF,$3D,$36,$37,$8F,$B6,$B3
	.dc.b	$52,$73,$7E,$C7,$9C,$65,$EC,$4A,$FF,$5C,$5F,$58,$FE,$9E,$D7,$D8
	.dc.b	$7A,$0F,$F2,$7B,$97,$E0,$35,$FF,$4A,$3C,$E3,$AC,$F8,$5E,$FB,$DB
	.dc.b	$E3,$36,$F8,$CD,$7C,$15,$C9,$F8,$B1,$9B,$ED,$7C,$A7,$27,$C0,$21
	.dc.b	$EC,$BF,$97,$5C,$5E,$4D,$8D,$DF,$99,$E6,$71,$33,$5F,$02,$79,$C6
	.dc.b	$0F,$83,$58,$7E,$5C,$75,$69,$D2,$BD,$8F,$53,$24,$72,$4A,$F0,$47
	.dc.b	$17,$E0,$3C,$63,$D0,$F1,$8C,$C7,$18,$F9,$A7,$13,$CE,$7A,$4F,$EF
	.dc.b	$73,$E9,$8E,$11,$F4,$A7,$C2,$F5,$38,$9E,$17,$31,$E2,$38,$DF,$D6
	.dc.b	$9F,$09,$C4,$E8,$72,$8E,$13,$C8,$F4,$92,$38,$C6,$9E,$11,$97,$D2
	.dc.b	$7B,$4E,$A4,$95,$7B,$67,$84,$79,$67,$89,$E5,$70,$49,$E1,$1E,$02
	.dc.b	$1C,$2B,$C8,$F7,$BD,$B9,$BE,$F1,$8B,$F1,$CF,$D1,$7B,$3E,$97,$C3
	.dc.b	$1B,$BF,$C5,$5F,$69,$E3,$5F,$44,$F1,$B8,$8D,$57,$E7,$A7,$A3,$5F
	.dc.b	$73,$63,$71,$E3,$AF,$34,$E6,$FF,$03,$3F,$C1,$2F,$C2,$DA,$FF,$33
	.dc.b	$3F,$28,$C1,$FE,$69,$8D,$BC,$88,$7D,$C9,$FC,$03,$40,$CD,$C3,$F0
	.dc.b	$E3,$EF,$46,$1F,$91,$7F,$42,$31,$1F,$6E,$31,$3C,$63,$F2,$53,$AC
	.dc.b	$7E,$65,$6A,$79,$9D,$AB,$84,$FD,$EA,$E1,$7E,$8A,$FB,$C7,$A4,$F4
	.dc.b	$57,$08,$D4,$F0,$9D,$61,$F5,$E7,$35,$B9,$E5,$5B,$7A,$92,$B8,$3F
	.dc.b	$1A,$35,$18,$8D,$61,$FE,$03,$AE,$91,$CA,$B7,$38,$6C,$7C,$03,$FD
	.dc.b	$22,$E1,$F9,$47,$D4,$C3,$FC,$21,$5D,$2F,$C1,$5B,$9F,$97,$5B,$C3
	.dc.b	$FB,$34,$E5,$F9,$15,$C2,$F9,$61,$E7,$BF,$16,$1F,$3E,$73,$1C,$CF
	.dc.b	$80,$E1,$F3,$B6,$3C,$B1,$CA,$FE,$6B,$4F,$9B,$3B,$8F,$2B,$E1,$8F
	.dc.b	$79,$D0,$F0,$1F,$70,$E4,$53,$47,$09,$F0,$B8,$7E,$1B,$67,$BB,$CB
	.dc.b	$0F,$BC,$7C,$57,$CA,$78,$1C,$3F,$66,$BC,$CF,$F7,$68,$FC,$FB,$63
	.dc.b	$F6,$71,$88,$EE,$78,$4F,$92,$7B,$3E,$29,$EF,$7C,$1C,$3D,$24,$C3
	.dc.b	$FC,$3D,$3A,$AF,$26,$1B,$9E,$B4,$77,$79,$AE,$E7,$84,$5F,$37,$18
	.dc.b	$7C,$2C,$3F,$C5,$98,$73,$9F,$E0,$75,$B8,$F2,$47,$46,$BD,$2F,$C0
	.dc.b	$FD,$58,$CC,$F1,$C3,$CB,$3E,$A2,$1F,$EF,$08,$D3,$E0,$AE,$78,$7E
	.dc.b	$CE,$73,$3D,$CE,$37,$DA,$BD,$37,$DA,$7D,$B8,$61,$E8,$7D,$19,$EB
	.dc.b	$19,$8E,$87,$8E,$32,$EC,$87,$F8,$D8,$D9,$FE,$38,$BF,$98,$D8,$FE
	.dc.b	$6B,$1E,$82,$3E,$3B,$7C,$FB,$D4,$F1,$7C,$5B,$E2,$FF,$9E,$B7,$1F
	.dc.b	$00,$70,$F8,$DA,$FF,$92,$BC,$F7,$C5,$FF,$46,$6C,$72,$6B,$EB,$6B
	.dc.b	$C7,$93,$81,$1E,$0F,$F5,$0F,$8B,$FF,$77,$47,$C3,$6B,$FE,$5A,$6B
	.dc.b	$FE,$77,$6B,$FD,$39,$78,$FC,$C9,$1E,$E8,$CB,$F0,$3E,$2D,$E6,$7F
	.dc.b	$60,$6C,$66,$30,$67,$E1,$E4,$71,$6B,$F6,$65,$E4,$73,$24,$77,$7F
	.dc.b	$B6,$FC,$4B,$E0,$E9,$27,$E9,$C6,$5B,$F1,$3C,$08,$E0,$DF,$67,$42
	.dc.b	$38,$3E,$EA,$FA,$7F,$EF,$AD,$8C,$C6,$0F,$F0,$0F,$C5,$33,$A4,$9F
	.dc.b	$C3,$C6,$5B,$1F,$9F,$8C,$43,$9B,$F8,$02,$8C,$C7,$FB,$7F,$E2,$BF
	.dc.b	$DB,$C7,$F1,$E2,$4F,$E5,$3B,$11,$E8,$F7,$E4,$FB,$1A,$FE,$A7,$24
	.dc.b	$7E,$C5,$B1,$FD,$01,$AF,$97,$B7,$FD,$81,$1F,$C0,$8F,$D7,$C6,$3A
	.dc.b	$FE,$39,$1F,$97,$F1,$1F,$28,$8E,$08,$75,$3F,$C4,$32,$8F,$21,$FB
	.dc.b	$7F,$89,$7A,$78,$AF,$34,$D7,$96,$8A,$DC,$7A,$FE,$27,$ED,$91,$FB
	.dc.b	$ED,$7C,$47,$FB,$A9,$F8,$87,$87,$D9,$FE,$2D,$F1,$2F,$81,$A4,$AF
	.dc.b	$EC,$F0,$F8,$9C,$47,$48,$E0,$EA,$79,$CF,$F2,$1F,$09,$F0,$74,$95
	.dc.b	$FD,$CB,$E2,$3A,$69,$FE,$46,$F8,$6F,$E3,$13,$FE,$48,$BE,$2F,$F7
	.dc.b	$86,$BE,$C6,$CF,$E3,$AB,$9C,$AB,$9A,$8C,$C6,$2B,$9B,$6F,$FC,$AF
	.dc.b	$5C,$66,$D7,$06,$E1,$F9,$D9,$D1,$E7,$AE,$37,$CE,$3D,$95,$C3,$0B
	.dc.b	$1F,$E6,$4C,$3A,$3C,$4C,$9F,$CE,$6F,$9D,$71,$8E,$E7,$AF,$0F,$16
	.dc.b	$1D,$E3,$33,$EC,$38,$47,$C2,$BE,$F3,$D4,$EF,$3B,$7C,$95,$D2,$3C
	.dc.b	$11,$F1,$0F,$06,$1F,$D2,$2B,$E7,$C6,$5A,$FC,$8F,$88,$7C,$4E,$DB
	.dc.b	$1F,$D7,$E5,$18,$8E,$4D,$7F,$B2,$7C,$4F,$FA,$51,$34,$8F,$FA,$61
	.dc.b	$F9,$37,$F9,$BE,$5F,$97,$3F,$E9,$29,$CE,$BF,$E1,$3D,$3F,$0B,$39
	.dc.b	$3E,$6D,$F1,$7E,$79,$98,$FC,$5C,$65,$E3,$1E,$43,$8C,$F5,$CF,$D3
	.dc.b	$18,$C3,$C6,$EE,$B9,$D6,$E3,$E7,$61,$E6,$AD,$B6,$3C,$67,$1B,$FF
	.dc.b	$16,$D7,$18,$FE,$60,$FF,$8B,$F0,$FE,$62,$7A,$CF,$E6,$71,$EB,$9E
	.dc.b	$37,$FE,$64,$9D,$3F,$E3,$8E,$6E,$4E,$AB,$F8,$5C,$FF,$2E,$8F,$F3
	.dc.b	$5F,$C4,$FE,$F9,$BB,$F0,$AF,$F6,$0C,$FF,$B0,$A7,$24,$8F,$E3,$BE
	.dc.b	$67,$81,$DD,$71,$BF,$F6,$3E,$5C,$E5,$F1,$1F,$E1,$BC,$3F,$CF,$6D
	.dc.b	$8F,$E4,$FE,$B7,$F5,$99,$CF,$99,$C6,$B5,$5F,$9B,$6B,$FC,$01,$AF
	.dc.b	$FB,$61,$AB,$F2,$7F,$F0,$3D,$7F,$A2,$9B,$1F,$CD,$BE,$25,$CF,$E7
	.dc.b	$9C,$E7,$FD,$89,$B7,$FC,$39,$B7,$FD,$9E,$F8,$CF,$5F,$23,$91,$F6
	.dc.b	$27,$FD,$D7,$5E,$07,$F6,$8D,$7A,$46,$D7,$CC,$BA,$B0,$DB,$FF,$20
	.dc.b	$B6,$7F,$CA,$4D,$7F,$80,$7C,$A3,$E5,$3A,$36,$BF,$C9,$2D,$9F,$F2
	.dc.b	$D3,$5F,$E1,$1F,$13,$F1,$3E,$27,$FC,$9E,$D8,$FF,$30,$3F,$D8,$B2
	.dc.b	$CE,$CA,$FF,$2A,$FC,$8B,$E4,$BF,$CA,$F8,$23,$0E,$A3,$31,$88,$FF
	.dc.b	$36,$7B,$6F,$4F,$8B,$63,$FC,$C2,$D7,$71,$DB,$DA,$7B,$74,$93,$FE
	.dc.b	$68,$8C,$BF,$E6,$AF,$89,$7C,$E7,$99,$B1,$98,$C1,$EA,$FC,$67,$EB
	.dc.b	$41,$A4,$E1,$D7,$C9,$49,$FF,$3B,$CA,$31,$1F,$E8,$9F,$88,$7C,$1F
	.dc.b	$C6,$6C,$7F,$9F,$9A,$EE,$35,$F1,$32,$34,$D7,$A7,$CA,$73,$23,$CD
	.dc.b	$FA,$7F,$11,$FC,$EA,$3F,$D1,$6D,$7F,$D3,$8D,$7F,$A5,$35,$FE,$DE
	.dc.b	$FF,$89,$5E,$3E,$7F,$42,$3D,$3E,$C7,$B9,$1C,$1F,$E9,$AF,$85,$7D
	.dc.b	$1F,$E1,$9B,$19,$8C,$19,$F8,$39,$1C,$5A,$FF,$AC,$E5,$F2,$DC,$C9
	.dc.b	$1D,$DF,$C9,$FC,$E3,$E1,$7F,$D6,$5F,$3A,$DA,$9F,$EC,$39,$D5,$8F
	.dc.b	$F1,$AB,$87,$FD,$84,$D7,$FB,$83,$5F,$F0,$AB,$5E,$4F,$1F,$1B,$FC
	.dc.b	$DE,$DE,$4D,$E1,$FE,$D3,$F7,$1F,$31,$E3,$89,$87,$8F,$0E,$31,$B9
	.dc.b	$CC,$F9,$A2,$CF,$F9,$F2,$B9,$CF,$1B,$EE,$D3,$CB,$5C,$23,$FC,$5B
	.dc.b	$3C,$27,$A1,$B1,$87,$A5,$E2,$66,$38,$5F,$9A,$38,$4E,$DE,$37,$E9
	.dc.b	$78,$61,$E1,$79,$9F,$E6,$4B,$E7,$87,$92,$75,$3C,$8E,$24,$3C,$E4
	.dc.b	$C3,$E1,$47,$78,$C4,$FC,$27,$B4,$66,$7D,$11,$A3,$F0,$55,$E2,$6C
	.dc.b	$F8,$B0,$F6,$4E,$64,$D8,$F8,$06,$A7,$C5,$3F,$E3,$79,$F9,$CD,$8F
	.dc.b	$E5,$31,$88,$F8,$07,$B6,$BD,$6F,$D3,$E9,$E8,$E7,$0D,$BF,$E8,$C8
	.dc.b	$CB,$EE,$34,$7E,$75,$CC,$F4,$8D,$1C,$71,$F8,$95,$D2,$74,$7F,$B2
	.dc.b	$2B,$8C,$64,$95,$D6,$B8,$BF,$20,$F0,$89,$FC,$21,$C9,$D1,$E1,$3D
	.dc.b	$D1,$C9,$F9,$B8,$7F,$A0,$AF,$C5,$36,$FF,$D0,$B7,$CE,$F6,$72,$21
	.dc.b	$EE,$AF,$5C,$F0,$AE,$53,$E9,$3C,$F8,$7D,$DB,$F3,$CF,$8E,$30,$F0
	.dc.b	$9F,$63,$9C,$31,$5C,$27,$9C,$62,$64,$60,$F2,$E1,$EC,$70,$7C,$32
	.dc.b	$3E,$87,$E3,$BB,$73,$87,$D2,$7E,$09,$83,$D8,$7C,$17,$57,$E5,$8C
	.dc.b	$1D,$23,$E2,$3E,$47,$A6,$1F,$02,$30,$6C,$87,$C2,$32,$78,$AB,$D0
	.dc.b	$E4,$F2,$19,$9F,$1D,$FC,$48,$F1,$99,$3D,$33,$EF,$39,$D7,$8E,$39
	.dc.b	$BF,$B3,$31,$1B,$AF,$4E,$1D,$63,$15,$E8,$AD,$D7,$C0,$8F,$0C,$61
	.dc.b	$F0,$C7,$27,$DB,$87,$E3,$E7,$51,$ED,$AF,$2D,$F7,$BE,$91,$89,$E3
	.dc.b	$87,$E1,$DA,$7F,$86,$5F,$F0,$F1,$1C,$10,$FF,$60,$1F,$2C,$FF,$61
	.dc.b	$61,$FD,$45,$F1,$57,$80,$8F,$F5,$DF,$79,$EF,$F9,$63,$DA,$41,$1A
	.dc.b	$7C,$A7,$D8,$70,$FD,$AA,$FC,$56,$0B,$FF,$7B,$D6,$A3,$E7,$D7,$90
	.dc.b	$ED,$19,$9F,$2C,$64,$91,$97,$C9,$5D,$A7,$AD,$FA,$67,$84,$FF,$4F
	.dc.b	$AE,$11,$8C,$35,$18,$9E,$53,$E2,$AE,$51,$88,$F8,$31,$8A,$E2,$D8
	.dc.b	$FC,$ED,$73,$9C,$43,$0F,$84,$FC,$FA,$CC,$7C,$2A,$E1,$3F,$32,$36
	.dc.b	$F8,$0C,$61,$F1,$27,$71,$F3,$EF,$F8,$84,$62,$BE,$6E,$1F,$07,$0E
	.dc.b	$F8,$71,$6E,$1E,$F3,$A9,$4E,$87,$B1,$DB,$F4,$9B,$3D,$63,$89,$92
	.dc.b	$B9,$79,$9F,$87,$8E,$38,$7A,$19,$1C,$B0,$FA,$E7,$AE,$FD,$11,$FE
	.dc.b	$11,$8C,$3E,$7C,$3E,$3C,$62,$3C,$47,$D1,$9F,$C2,$BD,$6B,$BB,$E4
	.dc.b	$AE,$31,$87,$E4,$DD,$AE,$35,$FA,$EB,$FC,$FB,$FE,$45,$3E,$E1,$FB
	.dc.b	$23,$EC,$1B,$BE,$4E,$8F,$F0,$E7,$BF,$87,$78,$78,$9F,$F1,$54,$66
	.dc.b	$BE,$1D,$7F,$91,$63,$0F,$1C,$D1,$8A,$F8,$D3,$B6,$90,$F8,$87,$8E
	.dc.b	$FF,$C8,$D3,$C6,$32,$79,$30,$FF,$21,$C6,$67,$CD,$3E,$A3,$A3,$82
	.dc.b	$4F,$08,$CB,$F2,$A3,$31,$FE,$2C,$9E,$93,$93,$84,$78,$B0,$F8,$2F
	.dc.b	$23,$81,$4F,$31,$EB,$8E,$35,$DA,$32,$4C,$3E,$3B,$F3,$8E,$55,$FC
	.dc.b	$1A,$33,$1F,$0B,$E2,$1F,$17,$FE,$DC,$6C,$7D,$F6,$BF,$CC,$9A,$FE
	.dc.b	$14,$2F,$86,$FF,$03,$3C,$30,$FD,$FB,$71,$FC,$63,$BA,$FF,$38,$35
	.dc.b	$ED,$5E,$32,$3F,$91,$F8,$87,$A7,$FE,$F2,$6C,$66,$30,$67,$DE,$FF
	.dc.b	$0C,$4E,$A7,$F0,$2D,$C7,$F8,$9A,$9D,$9C,$67,$C0,$D7,$FA,$9F,$C4
	.dc.b	$3E,$A3,$F4,$90,$8C,$3F,$A7,$6B,$C2,$37,$0F,$8F,$FF,$16,$3A,$C3
	.dc.b	$F9,$2B,$74,$FD,$93,$BB,$FF,$4B,$35,$F1,$D7,$42,$3F,$B7,$F8,$97
	.dc.b	$CA,$F9,$9B,$19,$8C,$19,$FA,$2F,$F1,$F4,$EA,$7F,$4E,$DC,$7F,$9C
	.dc.b	$A9,$D9,$C6,$7C,$0D,$7F,$BC,$FC,$43,$E5,$BE,$3B,$63,$F8,$63,$5D
	.dc.b	$C6,$BE,$03,$FC,$A1,$39,$C3,$FA,$4B,$74,$FE,$2C,$A7,$67,$1B,$EA
	.dc.b	$D7,$FC,$17,$F4,$6F,$80,$FB,$2D,$8F,$E3,$CD,$77,$1F,$7B,$F0,$47
	.dc.b	$C7,$FF,$98,$EB,$59,$7F,$5C,$6E,$5F,$CA,$6B,$78,$7F,$B3,$5A,$FF
	.dc.b	$87,$1A,$FE,$AD,$AF,$1B,$FC,$6F,$EE,$9F,$AF,$C3,$FB,$78,$F8,$CF
	.dc.b	$F4,$75,$7F,$B7,$E3,$1D,$FF,$E7,$7A,$E1,$8F,$F3,$B6,$D7,$F7,$9B
	.dc.b	$E2,$FF,$28,$BE,$2F,$FB,$8E,$76,$9F,$CF,$B4,$F2,$90,$FE,$8C,$BE
	.dc.b	$13,$FD,$1F,$3E,$7F,$22,$D9,$F3,$4C,$BF,$54,$E5,$B3,$97,$F9,$CF
	.dc.b	$C6,$7E,$E7,$1F,$EA,$CD,$80,$70,$DE,$8F,$F4,$D4,$60,$9D,$58,$8F
	.dc.b	$F4,$F7,$C3,$FA,$6B,$9E,$7F,$E1,$88,$C4,$7F,$4A,$6C,$7F,$87,$3E
	.dc.b	$2F,$F8,$53,$E2,$F8,$A4,$FF,$6B,$37,$95,$D7,$B9,$FE,$AD,$BE,$38
	.dc.b	$2F,$8B,$EC,$8C,$C6,$23,$AC,$FF,$77,$8C,$5F,$99,$D5,$70,$8F,$EF
	.dc.b	$8D,$AF,$8F,$1B,$3C,$8F,$A5,$DE,$3F,$E3,$F8,$F2,$46,$31,$D4,$72
	.dc.b	$C3,$C5,$3C,$A3,$67,$A6,$B9,$61,$E4,$6C,$EE,$3D,$35,$CA,$B8,$5F
	.dc.b	$65,$87,$A5,$F8,$26,$6F,$D4,$F0,$3C,$27,$0C,$3D,$4F,$F5,$99,$F9
	.dc.b	$ED,$8C,$BE,$D3,$FD,$23,$0F,$8B,$7C,$67,$F8,$8E,$30,$D8,$CC,$F0
	.dc.b	$6B,$FE,$29,$F8,$CF,$D4,$91,$FE,$92,$DC,$FE,$E6,$5F,$D7,$B4,$F8
	.dc.b	$15,$F9,$CA,$F8,$53,$98,$F6,$99,$3D,$EE,$6B,$F0,$D1,$97,$AE,$BF
	.dc.b	$E7,$87,$30,$AE,$B3,$9D,$3F,$D6,$31,$8C,$3C,$75,$B3,$BD,$6D,$FF
	.dc.b	$02,$4E,$DF,$29,$BB,$F0,$D7,$29,$D4,$62,$7F,$C6,$B8,$7B,$4F,$F4
	.dc.b	$3B,$E9,$C3,$FC,$A9,$19,$BB,$87,$F2,$47,$3C,$F6,$FF,$73,$35,$FF
	.dc.b	$4C,$45,$F8,$CF,$E2,$D5,$FB,$E3,$FC,$C0,$E2,$7F,$C8,$EF,$F0,$C8
	.dc.b	$D4,$7F,$0E,$8D,$5F,$FA,$71,$B3,$FC,$23,$A3,$CD,$5B,$E0,$F2,$5F
	.dc.b	$F5,$1C,$7C,$9F,$14,$F8,$BF,$7D,$FF,$1A,$8C,$63,$FE,$71,$6C,$7F
	.dc.b	$B9,$1A,$FF,$8C,$DC,$DF,$F4,$6F,$8C,$FF,$0B,$57,$A2,$95,$A8,$FE
	.dc.b	$3B,$3A,$9F,$F3,$2F,$C5,$BE,$1B,$E4,$36,$BF,$A7,$7C,$63,$E1,$F3
	.dc.b	$52,$7F,$D0,$D5,$BB,$FF,$7C,$B5,$FF,$24,$35,$FF,$47,$3B,$CB,$FA
	.dc.b	$EE,$0F,$4B,$A5,$7F,$55,$8C,$5F,$AF,$E2,$5F,$15,$F2,$29,$18,$9F
	.dc.b	$F4,$C3,$63,$FC,$B0,$6A,$78,$7C,$57,$AF,$E2,$BF,$D3,$CD,$9D,$AF
	.dc.b	$8D,$F8,$91,$8D,$95,$FE,$A5,$F9,$AF,$F1,$44,$FF,$99,$5A,$FF,$AA
	.dc.b	$DA,$FF,$56,$6A,$F8,$A7,$C5,$7F,$93,$1B,$1F,$EB,$16,$BE,$A6,$BF
	.dc.b	$E7,$3F,$15,$F1,$5F,$E5,$46,$C7,$FA,$E5,$AF,$C2,$6B,$C7,$DF,$E0
	.dc.b	$47,$83,$FE,$5E,$F8,$8F,$9D,$06,$93,$83,$5F,$11,$27,$FD,$91,$28
	.dc.b	$C4,$7D,$2F,$9A,$FF,$3A,$3F,$E8,$B6,$BF,$ED,$16,$BF,$DF,$1A,$BE
	.dc.b	$25,$F0,$3F,$E7,$26,$C7,$FB,$5D,$AF,$E0,$1A,$FF,$A6,$57,$C4,$71
	.dc.b	$24,$71,$7F,$CF,$BF,$32,$F8,$EA,$49,$FF,$72,$46,$5F,$F7,$37,$C5
	.dc.b	$3E,$2B,$FD,$0A,$D8,$CC,$60,$CF,$C5,$48,$E2,$D7,$F2,$FE,$2B,$E1
	.dc.b	$FF,$D1,$ED,$8F,$F7,$A3,$5D,$C7,$FA,$53,$85,$F1,$54,$93,$FE,$FB
	.dc.b	$8C,$B7,$E2,$B8,$11,$C1,$FF,$00,$7E,$63,$FB,$13,$7C,$FF,$F5,$0C
	.dc.b	$62,$3F,$5A,$D7,$FD,$8D,$F0,$DE,$5F,$11,$CF,$FE,$36,$9F,$DF,$7C
	.dc.b	$57,$F3,$C8,$C7,$8A,$8C,$4F,$F8,$F2,$F9,$57,$17,$E0,$47,$6D,$7F
	.dc.b	$D6,$71,$88,$D4,$60,$E6,$A3,$9B,$D6,$33,$D5,$E1,$8F,$E5,$13,$A7
	.dc.b	$84,$F2,$8C,$4E,$E5,$8F,$B9,$E8,$72,$AE,$07,$6C,$2C,$64,$87,$2C
	.dc.b	$3E,$35,$F5,$BE,$93,$E2,$36,$E3,$0F,$1D,$F4,$9E,$47,$88,$A7,$12
	.dc.b	$D7,$5B,$F8,$F1,$D1,$AE,$A7,$D0,$E4,$E9,$7E,$4A,$E5,$1F,$E7,$26
	.dc.b	$BD,$E3,$CD,$87,$28,$E8,$73,$35,$7F,$02,$F5,$5C,$27,$57,$D0,$EB
	.dc.b	$87,$D5,$3E,$0C,$7F,$5F,$8C,$E1,$F0,$EB,$CD,$7E,$1A,$FE,$50,$FF
	.dc.b	$27,$23,$FC,$9D,$F0,$E1,$E3,$8F,$ED,$5F,$10,$F8,$AC,$C7,$BE,$AA
	.dc.b	$9D,$3E,$38,$CE,$3F,$12,$73,$3F,$08,$98,$79,$9E,$F7,$C2,$FA,$D7
	.dc.b	$0A,$D5,$70,$C3,$CC,$FD,$E9,$D4,$7E,$BD,$B1,$E7,$3C,$87,$62,$1C
	.dc.b	$CD,$1C,$EF,$F5,$31,$CA,$18,$7D,$3C,$33,$52,$7A,$9C,$CF,$88,$E5
	.dc.b	$C4,$FC,$38,$C6,$1E,$C2,$61,$E9,$8C,$BE,$E8,$CD,$7F,$B0,$30,$F2
	.dc.b	$1E,$5B,$F2,$47,$B2,$7F,$53,$39,$C3,$E5,$E1,$D2,$BD,$F5,$D1,$F3
	.dc.b	$9C,$70,$F0,$11,$F0,$4F,$17,$E1,$D7,$47,$81,$D2,$78,$3D,$0F,$98
	.dc.b	$7C,$6A,$F0,$5F,$AA,$FF,$01,$3C,$8E,$13,$E5,$C3,$E4,$46,$5F,$7B
	.dc.b	$D0,$91,$BC,$3C,$D7,$F0,$67,$C1,$87,$CD,$79,$4F,$A5,$E3,$1E,$8B
	.dc.b	$EF,$18,$3D,$84,$3C,$B1,$CD,$FB,$B3,$E0,$C3,$D7,$1C,$4D,$9C,$6B
	.dc.b	$AE,$1F,$5A,$76,$D9,$F2,$9E,$83,$E8,$98,$3F,$1D,$87,$C9,$74,$7D
	.dc.b	$43,$E5,$3F,$05,$DC,$FE,$4D,$FE,$A8,$7D,$52,$C7,$ED,$9B,$1F,$E4
	.dc.b	$A7,$C2,$7F,$93,$23,$4F,$C1,$36,$6D,$F2,$3F,$07,$27,$C5,$7C,$58
	.dc.b	$FB,$D1,$87,$CF,$5F,$8D,$8C,$47,$C9,$AE,$77,$FD,$06,$F9,$4F,$56
	.dc.b	$C7,$C9,$8F,$E8,$35,$CA,$30,$4C,$BE,$C3,$CE,$7C,$27,$92,$B9,$CE
	.dc.b	$8F,$EE,$4E,$23,$E1,$CF,$17,$CD,$87,$A6,$75,$1F,$80,$3F,$B9,$38
	.dc.b	$C3,$F8,$66,$1E,$5C,$3B,$57,$46,$BC,$0F,$29,$EF,$2E,$1E,$B7,$FC
	.dc.b	$11,$7D,$70,$FE,$F0,$D8,$FE,$FA,$D7,$E2,$B6,$7D,$47,$38,$F4,$61
	.dc.b	$E6,$9D,$5F,$C2,$9F,$0B,$E6,$C3,$D5,$1E,$92,$1F,$05,$CB,$D3,$0F
	.dc.b	$8C,$F2,$37,$1A,$BF,$BF,$5D,$6F,$C3,$5E,$E3,$81,$D5,$CB,$F1,$CF
	.dc.b	$A6,$7D,$B3,$73,$EB,$6D,$FC,$66,$9F,$18,$A7,$C6,$2E,$1F,$1A,$FE
	.dc.b	$AD,$79,$A7,$C8,$6C,$F0,$96,$78,$3F,$12,$3C,$4E,$67,$9D,$78,$E3
	.dc.b	$06,$DC,$57,$28,$F9,$47,$9C,$D9,$82,$E1,$F2,$E7,$E7,$C6,$AF,$D0
	.dc.b	$FC,$7A,$E9,$18,$AF,$AE,$71,$9F,$94,$71,$AE,$75,$C1,$B7,$F8,$59
	.dc.b	$F8,$33,$E7,$AF,$DD,$C7,$F9,$9B,$0F,$F2,$C9,$1F,$E1,$D1,$E0,$8F
	.dc.b	$7D,$27,$FC,$9F,$F5,$0F,$8E,$FF,$2E,$FC,$5B,$E2,$EF,$C6,$4B,$E1
	.dc.b	$9F,$F9,$2A,$B5,$18,$C7,$FC,$E9,$18,$8F,$F7,$3B,$5F,$E4,$AD,$77
	.dc.b	$7F,$C8,$D4,$71,$AE,$D3,$E4,$3F,$CE,$3B,$BE,$A7,$FC,$E7,$F1,$AF
	.dc.b	$7A,$FA,$D3,$2F,$21,$EF,$9D,$38,$0E,$1F,$F7,$EB,$5F,$E7,$6D,$7F
	.dc.b	$D2,$0D,$7A,$4F,$88,$8F,$5F,$8C,$7B,$1F,$E8,$85,$F1,$97,$E3,$65
	.dc.b	$6E,$34,$7C,$39,$DC,$7F,$47,$6B,$FD,$38,$38,$7F,$A8,$BA,$D7,$FC
	.dc.b	$FE,$DD,$FF,$D5,$0D,$9F,$C1,$7C,$63,$D2,$FF,$4E,$AF,$8C,$BF,$1B
	.dc.b	$2B,$E4,$4E,$5B,$3F,$66,$31,$1E,$56,$BF,$D9,$1A,$FF,$AE,$5D,$E3
	.dc.b	$FE,$94,$6E,$39,$AF,$F5,$7F,$C6,$3E,$9E,$FC,$0E,$3E,$3A,$65,$F6
	.dc.b	$E9,$3A,$8F,$ED,$D1,$86,$C7,$E3,$E3,$11,$F2,$DB,$97,$FA,$99,$B8
	.dc.b	$FF,$B3,$DB,$3F,$A9,$5F,$2B,$FE,$C4,$F8,$E7,$C7,$7F,$B1,$BE,$3D
	.dc.b	$E5,$D3,$51,$30,$FF,$01,$57,$06,$C7,$E8,$E3,$11,$F7,$DB,$97,$FA
	.dc.b	$E9,$B9,$7F,$B8,$DB,$3F,$EC,$AB,$7D,$2F,$FB,$63,$E4,$3F,$C2,$B8
	.dc.b	$7F,$B3,$47,$C8,$7F,$9C,$AF,$F4,$F1,$8F,$1F,$84,$43,$FC,$31,$5F
	.dc.b	$B7,$4D,$23,$B3,$FD,$8E,$F8,$32,$FE,$13,$6F,$3B,$C9,$D5,$FE,$F6
	.dc.b	$C7,$C7,$F2,$16,$8E,$B1,$FD,$C6,$3F,$AC,$7C,$77,$FC,$26,$31,$1F
	.dc.b	$A5,$6B,$FE,$34,$F9,$07,$7F,$FB,$E9,$B1,$FC,$0D,$AF,$FB,$79,$FE
	.dc.b	$09,$8F,$FB,$A1,$CB,$67,$91,$CF,$BB,$F9,$65,$7F,$B9,$BE,$33,$C7
	.dc.b	$5F,$EF,$06,$EB,$A6,$CF,$F3,$3F,$90,$FE,$35,$5F,$CD,$7E,$39,$5B
	.dc.b	$D3,$AC,$F3,$6E,$19,$C3,$F8,$E6,$1C,$96,$1C,$9E,$D8,$B0,$ED,$8F
	.dc.b	$19,$DB,$6B,$84,$2F,$A4,$61,$B3,$B2,$4F,$6A,$EA,$D3,$A9,$FD,$46
	.dc.b	$3E,$BD,$76,$69,$F2,$AB,$AC,$F1,$9D,$63,$F3,$9B,$1D,$8F,$36,$1E
	.dc.b	$D7,$B9,$8C,$35,$87,$68,$F1,$5F,$38,$F0,$57,$1B,$F7,$3C,$EB,$FC
	.dc.b	$FB,$87,$CD,$6C,$65,$E3,$87,$F9,$F6,$5E,$8B,$E4,$30,$DA,$FE,$75
	.dc.b	$3A,$9C,$9F,$AD,$F9,$09,$3F,$D9,$5A,$FF,$A2,$5F,$81,$7F,$9C,$D3
	.dc.b	$E1,$4F,$F7,$D9,$CE,$A9,$CC,$A7,$31,$86,$C7,$C8,$39,$1F,$2A,$BB
	.dc.b	$39,$39,$19,$AE,$CE,$67,$9E,$7F,$8A,$8E,$6F,$E3,$2B,$9B,$98,$E6
	.dc.b	$E6,$B9,$B9,$AE,$7A,$7E,$8A,$7F,$8D,$61,$4F,$EF,$97,$DA,$70,$FE
	.dc.b	$7E,$2E,$1E,$68,$FD,$17,$5F,$27,$FC,$3B,$75,$CD,$7F,$81,$FE,$43
	.dc.b	$FB,$DB,$77,$D4,$7F,$8A,$E3,$04,$9F,$01,$FC,$CE,$B7,$5D,$F0,$FE
	.dc.b	$8B,$1D,$FE,$3D,$F2,$19,$F9,$07,$91,$98,$C3,$72,$FE,$6F,$3A,$7F
	.dc.b	$C3,$95,$C9,$CC,$FA,$FE,$3D,$F2,$1F,$07,$4F,$E2,$F1,$8A,$FF,$6A
	.dc.b	$36,$3F,$C1,$CD,$7F,$DB,$AE,$73,$FE,$95,$F2,$0F,$67,$B6,$1E,$D8
	.dc.b	$C6,$1F,$E2,$FB,$E0,$FF,$92,$A3,$38,$7C,$61,$3A,$7F,$8B,$46,$B0
	.dc.b	$FB,$1F,$1C,$F8,$E7,$C7,$7F,$BA,$E3,$0D,$F3,$7D,$F8,$7F,$89,$E7
	.dc.b	$2F,$F8,$A7,$E3,$9E,$77,$F9,$45,$B1,$FE,$F5,$6B,$B8,$FF,$2B,$53
	.dc.b	$E6,$69,$27,$FD,$F9,$19,$7F,$DF,$DE,$23,$E3,$7F,$CB,$CD,$8C,$C6
	.dc.b	$0C,$FC,$14,$8E,$2D,$7F,$C8,$1E,$EF,$F8,$31,$FF,$3E,$F9,$3F,$E6
	.dc.b	$98,$C4,$6A,$3F,$CD,$9F,$1A,$F8,$DA,$49,$FF,$25,$08,$D3,$8F,$8D
	.dc.b	$FE,$C6,$FF,$5C,$0E,$1F,$EB,$ED,$7F,$D1,$8D,$5F,$3C,$F8,$CF,$F3
	.dc.b	$F3,$63,$C2,$D7,$FC,$B8,$D7,$C1,$F1,$B5,$FE,$DA,$D7,$FD,$34,$D4
	.dc.b	$DF,$8C,$7C,$6F,$FA,$35,$B3,$F0,$1A,$FF,$9A,$DA,$FB,$FE,$79,$F4
	.dc.b	$BF,$E9,$56,$C7,$C8,$6B,$FE,$73,$6B,$F1,$FE,$3D,$F1,$DF,$E9,$D6
	.dc.b	$C7,$D0,$6B,$FE,$7B,$6B,$F3,$F7,$7C,$67,$FA,$95,$B1,$F6,$1A,$FF
	.dc.b	$A0,$DA,$FD,$7F,$13,$FD,$4C,$FD,$E6,$BF,$E8,$96,$BB,$F8,$3F,$F5
	.dc.b	$97,$23,$E3,$A9,$27,$F0,$F1,$97,$F1,$3F,$1D,$FD,$79,$BE,$97,$FA
	.dc.b	$EE,$31,$1F,$E9,$76,$BF,$8B,$F9,$F7,$C8,$FF,$B0,$5B,$1F,$94,$6B
	.dc.b	$FE,$9F,$6B,$F9,$3E,$B7,$C4,$7F,$88,$A3,$FD,$B9,$D7,$F1,$3C,$4F
	.dc.b	$E7,$CD,$F1,$FF,$C6,$B7,$C2,$BF,$D8,$4D,$7F,$D5,$8D,$7B,$5F,$EE
	.dc.b	$08,$F0,$7C,$0A,$39,$5F,$F4,$DF,$90,$7C,$86,$A7,$38,$78,$B0,$C1
	.dc.b	$CA,$B7,$87,$52,$1F,$C9,$6B,$8B,$FE,$F9,$8D,$B9,$9E,$6F,$0C,$3C
	.dc.b	$58,$62,$34,$67,$0F,$F2,$BD,$F2,$AF,$EC,$98,$76,$AF,$66,$1C,$67
	.dc.b	$9C,$7F,$BA,$6B,$85,$FA,$E3,$A1,$E5,$C3,$D8,$71,$8D,$47,$F3,$F9
	.dc.b	$E2,$EC,$E2,$78,$0E,$D3,$C9,$CE,$1E,$59,$E3,$06,$C7,$A4,$D4,$F1
	.dc.b	$9F,$82,$4B,$F5,$BE,$38,$F3,$23,$C4,$64,$F1,$9F,$CA,$2B,$E8,$BF
	.dc.b	$C9,$5A,$FC,$C8,$FE,$5A,$F9,$5F,$14,$F8,$0F,$CA,$FC,$86,$30,$E1
	.dc.b	$5D,$8F,$EC,$99,$27,$51,$F1,$CF,$21,$A7,$2F,$73,$95,$76,$6D,$76
	.dc.b	$78,$C6,$5B,$19,$8F,$33,$FE,$9C,$AF,$14,$F2,$3C,$53,$A7,$F6,$35
	.dc.b	$E2,$C3,$EA,$9F,$5A,$B8,$CF,$63,$8D,$F7,$3B,$57,$1C,$3D,$75,$C3
	.dc.b	$0E,$D7,$ED,$39,$08,$D5,$F3,$8E,$A4,$35,$7D,$5F,$51,$82,$E1,$E8
	.dc.b	$C3,$F7,$B7,$D2,$79,$4E,$5F,$C3,$61,$F2,$4F,$2C,$66,$3E,$A1,$D2
	.dc.b	$BD,$77,$E1,$8F,$4D,$F5,$7D,$97,$CA,$BC,$35,$B8,$F8,$35,$BB,$ED
	.dc.b	$18,$9D,$61,$F0,$70,$F6,$1F,$1A,$B7,$1F,$3A,$3C,$87,$C2,$BE,$98
	.dc.b	$72,$8F,$B6,$68,$F3,$DE,$5F,$3D,$73,$7B,$4F,$B9,$E1,$87,$95,$B3
	.dc.b	$E5,$BE,$F1,$83,$B9,$C8,$F2,$5F,$7C,$3F,$14,$DC,$3E,$2C,$F1,$3C
	.dc.b	$67,$2C,$3E,$53,$C8,$D5,$72,$6C,$F2,$7E,$B9,$E1,$2C,$FD,$DC,$3E
	.dc.b	$65,$71,$9C,$C6,$2B,$95,$6E,$7F,$0F,$18,$9F,$8D,$3E,$49,$E2,$79
	.dc.b	$E3,$93,$F8,$C7,$64,$79,$99,$38,$10,$F8,$24,$C3,$F1,$87,$CE,$7E
	.dc.b	$C9,$C4,$FC,$2C,$F8,$CF,$43,$89,$F1,$BE,$63,$07,$88,$C4,$F5,$9F
	.dc.b	$BD,$1B,$BF,$C7,$BE,$59,$D5,$7E,$07,$0F,$89,$3C,$EF,$C3,$5A,$8F
	.dc.b	$DC,$3E,$A9,$FC,$83,$63,$F0,$0D,$7E,$C4,$F2,$9D,$9D,$A3,$E2,$DF
	.dc.b	$D9,$9F,$7B,$E6,$8F,$69,$1C,$57,$59,$F8,$D7,$F4,$EB,$58,$7E,$7D
	.dc.b	$F8,$0D,$9E,$AD,$9E,$17,$D3,$0F,$23,$6F,$EA,$61,$F0,$5F,$B5,$3A
	.dc.b	$BE,$51,$F7,$EB,$83,$E0,$3F,$3C,$53,$EF,$97,$0F,$AA,$7F,$3F,$7F
	.dc.b	$CF,$97,$FC,$D1,$B7,$FE,$8B,$23,$E4,$7E,$BC,$BE,$41,$F2,$1F,$36
	.dc.b	$FE,$36,$5F,$C9,$27,$38,$71,$21,$FA,$0B,$E5,$7C,$70,$EB,$5C,$2B
	.dc.b	$F1,$B1,$89,$D4,$72,$A5,$F7,$9C,$CF,$5B,$E5,$5F,$83,$8E,$15,$B6
	.dc.b	$9F,$95,$BF,$F2,$54,$F9,$23,$27,$E1,$30,$FE,$85,$19,$8F,$29,$C9
	.dc.b	$C1,$DD,$F1,$61,$E4,$8F,$E0,$86,$AF,$C5,$3C,$84,$7B,$1B,$87,$AA
	.dc.b	$77,$3D,$4D,$1C,$E7,$FC,$31,$F1,$EF,$8F,$FF,$08,$B6,$3F,$D7,$ED
	.dc.b	$7F,$D7,$81,$C3,$FE,$BF,$75,$3E,$38,$D4,$FF,$98,$27,$35,$FE,$27
	.dc.b	$9F,$F3,$9F,$2F,$F9,$93,$7F,$F4,$6C,$65,$AF,$F8,$B5,$AF,$03,$FD
	.dc.b	$A7,$F1,$EF,$7B,$FC,$52,$D8,$CC,$60,$8F,$F1,$B8,$D3,$F2,$AF,$C9
	.dc.b	$18,$CB,$D5,$F1,$CF,$63,$FC,$F2,$DB,$FF,$71,$B5,$F8,$AD,$7F,$DC
	.dc.b	$4D,$7E,$02,$F8,$FF,$F1,$FB,$63,$9C,$60,$8F,$F3,$18,$D3,$E7,$3F
	.dc.b	$CA,$0E,$2B,$E8,$7C,$7B,$CC,$FF,$47,$B6,$BF,$DF,$2D,$7E,$DB,$5F
	.dc.b	$F7,$C3,$5F,$AC,$BE,$3F,$FC,$AA,$D8,$E7,$18,$23,$FD,$1E,$34,$F9
	.dc.b	$C3,$8A,$FC,$37,$C6,$BE,$3F,$FD,$4A,$DA,$FC,$78,$70,$FE,$49,$AF
	.dc.b	$E2,$CD,$47,$F9,$AA,$1E,$FD,$24,$F4,$8C,$B6,$3C,$24,$7F,$B0,$43
	.dc.b	$E3,$BC,$A7,$52,$47,$37,$FC,$F3,$F1,$8F,$8D,$A4,$9F,$5C,$65,$B1
	.dc.b	$FE,$C4,$8C,$47,$C7,$6E,$3F,$B8,$F1,$5F,$1B,$BB,$97,$FD,$D6,$F7
	.dc.b	$C7,$FE,$4E,$BD,$CF,$F7,$9F,$8A,$FF,$46,$A6,$91,$FF,$48,$BF,$2F
	.dc.b	$1F,$DD,$36,$BF,$70,$D9,$FE,$23,$F1,$8F,$8D,$FF,$6E,$CE,$67,$B1
	.dc.b	$83,$F9,$3C,$ED,$F2,$72,$F8,$08,$FE,$02,$7F,$50,$FF,$AB,$5F,$D7
	.dc.b	$61,$E4,$9E,$E7,$4F,$23,$FD,$DB,$F1,$8F,$86,$A4,$C3,$FC,$35,$1F
	.dc.b	$BD,$CF,$F8,$93,$B8,$CC,$ED,$FC,$C4,$6D,$FF,$5F,$7C,$47,$8A,$BF
	.dc.b	$8D,$36,$33,$87,$FB,$FB,$E3,$3F,$8A,$65,$FE,$C5,$F8,$B5,$F0,$AF
	.dc.b	$E4,$53,$FC,$9E,$39,$5E,$A7,$97,$6F,$F3,$E8,$5F,$27,$FD,$A1,$87
	.dc.b	$67,$37,$D5,$F0,$35,$C9,$C2,$BA,$57,$38,$57,$38,$C5,$7F,$1D,$6D
	.dc.b	$FF,$39,$78,$4E,$E7,$F9,$E3,$6B,$E1,$9E,$22,$9E,$23,$BD,$F8,$A7
	.dc.b	$8B,$63,$75,$E0,$79,$1C,$30,$B7,$DE,$3E,$41,$AB,$F4,$1D,$27,$77
	.dc.b	$F2,$E7,$F8,$45,$7C,$96,$C6,$5E,$07,$F6,$08,$79,$8F,$8C,$FE,$2B
	.dc.b	$18,$6C,$66,$78,$35,$FE,$37,$F1,$9F,$B2,$23,$FF,$00,$1B,$97,$D3
	.dc.b	$C7,$F3,$5A,$7C,$0A,$0A,$F8,$73,$A8,$F9,$86,$A3,$3B,$7F,$81,$6F
	.dc.b	$EF,$4E,$4F,$A5,$5C,$63,$31,$D4,$CC,$F1,$D3,$FC,$03,$7D,$30,$F2
	.dc.b	$56,$EF,$D1,$7B,$8F,$E6,$95,$B7,$CA,$6E,$FF,$43,$3B,$8F,$F0,$24
	.dc.b	$EE,$B9,$DF,$FA,$47,$0F,$71,$FE,$0D,$AF,$54,$E9,$FE,$9D,$1A,$9F
	.dc.b	$D6,$F4,$FF,$88,$1E,$E7,$37,$13,$F3,$3E,$33,$FB,$F4,$FF,$A7,$DB
	.dc.b	$1F,$D6,$5F,$A3,$19,$DF,$8C,$A7,$77,$FC,$DF,$2C,$C6,$2F,$F9,$A4
	.dc.b	$63,$E6,$24,$7F,$AD,$69,$E9,$3C,$BC,$E3,$E1,$BF,$80,$F9,$DA,$F8
	.dc.b	$5B,$89,$CF,$CD,$3E,$6E,$93,$2F,$ED,$B1,$97,$FB,$7F,$CC,$3E,$63
	.dc.b	$AB,$63,$31,$83,$C3,$F1,$0F,$8A,$7C,$57,$F7,$88,$C3,$7C,$87,$C6
	.dc.b	$E6,$B7,$B7,$F9,$33,$E9,$9E,$07,$F9,$B3,$E3,$6C,$FC,$56,$BF,$E0
	.dc.b	$B9,$BC,$9F,$E4,$F6,$EB,$FA,$68,$56,$2B,$FC,$CE,$D5,$54,$EA,$FF
	.dc.b	$82,$DA,$FF,$95,$DA,$FE,$C9,$AB,$E7,$1E,$8F,$FA,$13,$EE,$3F,$B8
	.dc.b	$CF,$F8,$DF,$04,$61,$FF,$33,$57,$09,$FF,$77,$35,$E1,$3B,$C7,$D8
	.dc.b	$D8,$FE,$07,$3B,$AE,$77,$FE,$28,$9F,$23,$67,$FC,$D2,$7F,$9A,$BE
	.dc.b	$2B,$EC,$4F,$E4,$7D,$E7,$C1,$FF,$A7,$1B,$1F,$F0,$06,$76,$FF,$8E
	.dc.b	$DB,$87,$FA,$32,$77,$3E,$F6,$C7,$FA,$0E,$77,$97,$FA,$05,$E7,$5F
	.dc.b	$E3,$96,$B9,$DB,$D5,$18,$8F,$E4,$3F,$0F,$FE,$AC,$FC,$5F,$FB,$A6
	.dc.b	$07,$FB,$B4,$A7,$F5,$4F,$E0,$F0,$FF,$48,$47,$1F,$1B,$FD,$73,$5C
	.dc.b	$23,$75,$FE,$57,$7F,$DB,$CF,$0E,$CF,$99,$18,$8F,$1C,$7C,$FF,$84
	.dc.b	$7D,$1F,$FB,$13,$F1,$6F,$CF,$D7,$EF,$5A,$FF,$98,$E7,$6F,$94,$FE
	.dc.b	$7A,$E1,$FC,$44,$70,$CF,$FD,$72,$DC,$FF,$D5,$55,$B9,$FF,$58,$3C
	.dc.b	$6B,$FC,$C8,$D7,$77,$E7,$3F,$0D,$F0,$AF,$8A,$FF,$6D,$FC,$87,$F0
	.dc.b	$9A,$FE,$28,$D7,$FC,$F5,$3B,$7D,$01,$C3,$F9,$F8,$E1,$87,$FB,$41
	.dc.b	$B8,$FF,$B0,$6B,$73,$FE,$C7,$78,$D7,$F9,$ED,$AE,$EF,$CE,$7F,$9B
	.dc.b	$B9,$3F,$C5,$37,$FE,$CB,$F8,$27,$C2,$62,$FF,$B3,$C6,$1F,$F3,$35
	.dc.b	$7F,$9B,$CD,$28,$D3,$FE,$97,$6E,$3F,$ED,$BF,$53,$F5,$D8,$7F,$B6
	.dc.b	$5B,$5D,$B0,$E7,$7C,$2B,$58,$77,$AF,$8F,$39,$78,$1E,$42,$1E,$1A
	.dc.b	$E6,$71,$7F,$D0,$93,$87,$9C,$67,$0F,$34,$E6,$79,$1D,$23,$2F,$02
	.dc.b	$1D,$0D,$5F,$CB,$AE,$53,$C2,$74,$74,$23,$92,$4F,$18,$EB,$18,$7F
	.dc.b	$C3,$38,$7C,$87,$E8,$57,$6A,$F8,$F1,$E2,$3E,$29,$D9,$CB,$E2,$A4
	.dc.b	$77,$8D,$1D,$AF,$E2,$46,$67,$73,$D1,$F2,$2B,$F2,$61,$F5,$63,$CA
	.dc.b	$78,$CD,$5F,$9D,$E5,$7E,$4B,$E7,$5C,$4E,$57,$EF,$3A,$E1,$C6,$F9
	.dc.b	$9D,$67,$84,$7A,$23,$D4,$7C,$AB,$F5,$1D,$C8,$F8,$88,$74,$3E,$33
	.dc.b	$C1,$CD,$78,$F0,$F1,$5F,$99,$F0,$4F,$C0,$3E,$59,$CF,$0F,$6E,$1E
	.dc.b	$9B,$E5,$3E,$53,$AD,$72,$BF,$5C,$7C,$99,$F8,$2E,$CE,$98,$7A,$9F
	.dc.b	$06,$1F,$59,$B1,$99,$E3,$1E,$47,$C2,$F9,$AB,$D4,$FC,$BC,$3C,$D8
	.dc.b	$78,$DB,$19,$3C,$55,$CB,$0F,$05,$FA,$27,$E4,$9F,$48,$ED,$5F,$1A
	.dc.b	$BC,$75,$C2,$31,$7F,$31,$B5,$EF,$BF,$2B,$D0,$F6,$9F,$20,$F2,$1E
	.dc.b	$D9,$F4,$3E,$66,$B9,$37,$5E,$17,$E7,$47,$85,$A7,$86,$BC,$6F,$A8
	.dc.b	$C9,$4F,$25,$FB,$30,$FC,$44,$7D,$63,$E1,$5F,$77,$C4,$7A,$F0,$F9
	.dc.b	$11,$FA,$CA,$E5,$3E,$D3,$57,$F4,$9A,$7B,$4F,$5D,$F8,$27,$EA,$3E
	.dc.b	$2B,$F9,$37,$F9,$2A,$E0,$72,$3C,$6E,$5B,$3F,$8D,$9F,$2B,$F8,$BC
	.dc.b	$3E,$34,$FC,$06,$D7,$A5,$B3,$B9,$F9,$47,$C8,$9F,$84,$48,$D4,$7C
	.dc.b	$B8,$E4,$F0,$3B,$15,$CB,$B8,$E3,$1C,$E3,$E4,$3E,$B3,$0F,$D3,$25
	.dc.b	$7B,$1F,$AB,$3E,$FC,$3E,$75,$71,$8F,$99,$7F,$8A,$8E,$78,$7C,$EB
	.dc.b	$FE,$13,$87,$CE,$BE,$51,$F3,$AB,$93,$E4,$8C,$BF,$38,$F0,$9C,$B0
	.dc.b	$FC,$7C,$FA,$2B,$A4,$E8,$FD,$09,$2F,$C4,$E2,$36,$FB,$AB,$D7,$7E
	.dc.b	$0B,$F4,$E1,$EB,$8F,$6D,$7E,$28,$F3,$1D,$EF,$E3,$3F,$27,$0F,$CA
	.dc.b	$BE,$1A,$FE,$DC,$FF,$BB,$8F,$E0,$05,$3F,$96,$16,$3F,$CE,$CD,$8F
	.dc.b	$7D,$7F,$84,$FE,$1E,$CF,$85,$AF,$D8,$6D,$7F,$50,$6D,$7F,$82,$9B
	.dc.b	$3F,$E0,$86,$BF,$E1,$56,$E1,$FC,$61,$B7,$E1,$9F,$CB,$3F,$A1,$8C
	.dc.b	$C7,$F8,$91,$AF,$F8,$85,$B8,$1A,$48,$FF,$1F,$FC,$03,$E3,$FF,$C5
	.dc.b	$D8,$3E,$8D,$F4,$38,$6D,$F0,$AF,$F1,$7D,$F1,$8F,$F7,$04,$E9,$FF
	.dc.b	$53,$36,$FF,$A1,$4E,$E5,$18,$AF,$F4,$F4,$E6,$7E,$B3,$5F,$B4,$D4
	.dc.b	$3E,$0D,$F1,$9E,$7F,$97,$93,$FE,$5C,$9C,$B6,$72,$F9,$A7,$FC,$93
	.dc.b	$3A,$8F,$2E,$BF,$89,$8C,$47,$E2,$5A,$E6,$BF,$D6,$D1,$88,$F2,$FC
	.dc.b	$23,$DD,$F9,$5E,$74,$9F,$F3,$7C,$E5,$B3,$97,$E4,$CF,$F9,$72,$75
	.dc.b	$1E,$43,$AC,$FF,$B1,$63,$31,$F9,$D6,$BF,$A1,$6B,$F2,$3E,$0F,$FD
	.dc.b	$77,$8F,$FB,$B7,$67,$99,$F7,$CF,$F3,$5C,$EA,$B8,$78,$79,$7E,$FC
	.dc.b	$EA,$24,$7E,$B4,$D9,$44,$FF,$B3,$1F,$D9,$35,$FD,$B7,$C1,$3E,$7B
	.dc.b	$F2,$7E,$57,$E1,$F8,$FF,$C7,$E1,$C6,$7C,$27,$FD,$03,$5C,$6F,$FD
	.dc.b	$65,$7B,$E3,$E6,$6A,$A4,$F9,$6B,$83,$E9,$6B,$FC,$1D,$AF,$F0,$BF
	.dc.b	$7F,$F8,$55,$F9,$31,$7C,$67,$A0,$FF,$49,$CE,$B2,$E1,$79,$3C,$B5
	.dc.b	$C1,$FF,$58,$CE,$DF,$CE,$BF,$F0,$05,$34,$8F,$CD,$3F,$CC,$18,$7F
	.dc.b	$A6,$3D,$B9,$1F,$EB,$CE,$B4,$FF,$A9,$9B,$1F,$E9,$A9,$DB,$F8,$D3
	.dc.b	$99,$C6,$7B,$4F,$08,$FB,$CD,$7F,$95,$47,$8E,$BF,$88,$47,$BB,$E8
	.dc.b	$FE,$D9,$F1,$3A,$A4,$64,$87,$86,$78,$BE,$3F,$80,$7B,$3F,$EB,$E6
	.dc.b	$C7,$FB,$61,$AE,$FC,$9D,$4C,$8D,$3C,$63,$FD,$8F,$9B,$DA,$A4,$AF
	.dc.b	$F7,$1F,$C0,$2B,$F8,$BC,$3D,$D5,$FF,$4D,$39,$47,$FB,$A9,$AF,$CB
	.dc.b	$6F,$17,$F9,$8E,$FA,$3F,$C6,$AB,$9D,$7F,$1C,$9C,$BD,$EB,$B6,$1F
	.dc.b	$D6,$EB,$73,$E2,$8C,$4F,$F5,$FA,$E3,$5A,$6C,$72,$AF,$03,$5E,$C7
	.dc.b	$87,$0F,$7D,$71,$6C,$EE,$3C,$2F,$B9,$D6,$1D,$2B,$93,$5F,$31,$67
	.dc.b	$81,$E7,$AE,$B1,$C0,$F1,$9C,$0B,$1C,$1D,$57,$81,$FE,$21,$7F,$21
	.dc.b	$FE,$FC,$D7,$B5,$FF,$13,$87,$BE,$F8,$0C,$36,$B9,$B5,$E6,$D7,$F8
	.dc.b	$F7,$C0,$7E,$D4,$8F,$EC,$DB,$87,$D6,$C3,$F3,$3C,$9F,$0E,$BF,$AC
	.dc.b	$D7,$C5,$9C,$C7,$C0,$32,$7C,$27,$35,$F8,$08,$CB,$D7,$5F,$F7,$1B
	.dc.b	$98,$57,$59,$CE,$7E,$89,$E9,$18,$7F,$C4,$93,$BC,$3C,$D1,$C8,$DB
	.dc.b	$88,$D4,$79,$CD,$5F,$8E,$73,$87,$FB,$04,$F6,$9F,$EC,$39,$F4,$E1
	.dc.b	$FE,$35,$8C,$CD,$C3,$F9,$FB,$9F,$80,$B7,$FE,$00,$AD,$B6,$32,$BD
	.dc.b	$FF,$DE,$56,$CF,$F2,$13,$88,$FE,$3C,$FF,$82,$A3,$51,$FE,$0C,$8D
	.dc.b	$57,$FB,$65,$B3,$E1,$EF,$F0,$D7,$7E,$0F,$2D,$FF,$7F,$C7,$C7,$83
	.dc.b	$C0,$F8,$17,$D5,$AF,$9D,$B1,$BC,$7F,$B4,$36,$7F,$C0,$16,$F8,$07
	.dc.b	$C4,$76,$C3,$C3,$18,$C3,$D7,$6F,$7F,$FB,$EC,$7F,$73,$6C,$7F,$77
	.dc.b	$6B,$FE,$65,$6B,$FE,$69,$F7,$DE,$BB,$DF,$FE,$F9,$18,$6F,$BE,$F7
	.dc.b	$F3,$5B,$BF,$F0,$27,$CB,$BE,$5F,$FA,$C3,$63,$FC,$0C,$D7,$71,$FD
	.dc.b	$77,$E0,$3F,$C0,$11,$FD,$F3,$E0,$6B,$F1,$1B,$C2,$DB,$FD,$01,$58
	.dc.b	$6F,$89,$CC,$97,$FE,$15,$9E,$6F,$FA,$15,$E0,$FF,$93,$9A,$9C,$D2
	.dc.b	$DF,$6D,$F9,$BC,$3F,$D2,$1F,$02,$F5,$7F,$CD,$BF,$04,$F8,$8A,$6F
	.dc.b	$3F,$F2,$DD,$6D,$FB,$2F,$FA,$51,$E7,$5F,$DF,$E3,$11,$C2,$30,$73
	.dc.b	$51,$CD,$FE,$F8,$DC,$BF,$C5,$76,$F9,$FE,$1F,$07,$F7,$AF,$FC,$53
	.dc.b	$F0,$9F,$E4,$1B,$3F,$E4,$38,$4E,$DF,$F0,$43,$5F,$F5,$73,$5F,$F5
	.dc.b	$93,$5F,$F3,$CB,$5E,$7C,$5E,$22,$3F,$E3,$AF,$92,$FE,$E4,$FE,$73
	.dc.b	$E1,$1F,$09,$FE,$98,$F8,$5C,$CF,$E9,$9B,$97,$E0,$E3,$FD,$0F,$5C
	.dc.b	$1E,$55,$FE,$29,$8C,$47,$F9,$65,$AF,$FB,$1D,$AF,$FB,$2D,$DC,$F5
	.dc.b	$84,$63,$83,$55,$FE,$AC,$FC,$4F,$F4,$95,$FB,$4F,$85,$7C,$2F,$FA
	.dc.b	$C3,$E1,$B3,$3F,$BE,$6E,$5D,$6B,$73,$C6,$BF,$C8,$11,$88,$FF,$38
	.dc.b	$35,$FF,$6F,$35,$FF,$71,$3B,$9E,$93,$F8,$78,$C6,$3F,$CB,$6F,$F8
	.dc.b	$C5,$FF,$0A,$F8,$57,$C2,$FF,$B1,$BE,$1B,$33,$FC,$51,$B9,$7C,$3A
	.dc.b	$DC,$F1,$AF,$F2,$B4,$62,$3F,$D0,$CD,$7F,$DE,$8D,$7F,$DF,$0E,$E7
	.dc.b	$A4,$FE,$76,$31,$87,$F4,$3F,$C5,$5F,$17,$FE,$5D,$E8,$FD,$17,$C3
	.dc.b	$BE,$22,$9F,$EF,$87,$19,$FF,$26,$6B,$FE,$61,$F6,$FF,$4D,$8F,$FB
	.dc.b	$3A,$11,$87,$FD,$A3,$3C,$34,$FC,$13,$6B,$85,$6E,$B8,$E1,$E2,$7C
	.dc.b	$18,$78,$CC,$3E,$18,$F5,$3A,$21,$93,$FC,$F8,$F0,$C3,$C3,$3C,$E7
	.dc.b	$71,$F9,$5A,$F1,$1E,$3C,$3C,$D3,$EA,$AE,$0F,$FA,$06,$32,$FA,$0E
	.dc.b	$E7,$4C,$3E,$24,$6A,$BA,$E1,$E3,$9C,$DF,$4C,$3C,$B8,$72,$3D,$47
	.dc.b	$78,$E0,$FF,$BF,$2F,$9B,$DC,$F3,$1C,$67,$D9,$7C,$21,$87,$8E,$3E
	.dc.b	$4C,$75,$39,$1E,$B3,$C3,$3E,$47,$46,$30,$FB,$33,$D6,$BA,$4F,$86
	.dc.b	$7D,$AF,$4B,$F3,$BC,$AF,$A1,$E6,$78,$1E,$A3,$53,$D9,$C9,$EA,$C3
	.dc.b	$D9,$5D,$2B,$9D,$78,$2B,$CC,$7C,$7C,$3C,$D7,$F3,$E7,$55,$E1,$AD
	.dc.b	$9D,$4F,$45,$7C,$68,$EB,$1E,$57,$81,$27,$99,$23,$C0,$7A,$30,$FC
	.dc.b	$39,$F1,$9E,$EE,$2F,$C9,$1D,$67,$E0,$3A,$8F,$7D,$F8,$E7,$B4,$F9
	.dc.b	$CF,$4E,$1F,$99,$AD,$C7,$85,$AF,$33,$DA,$64,$A6,$4F,$A2,$7A,$8C
	.dc.b	$61,$F0,$9F,$C3,$3F,$05,$F4,$10,$CD,$7A,$A3,$86,$1F,$05,$E1,$3E
	.dc.b	$97,$A5,$FD,$BA,$F6,$D6,$DE,$46,$2F,$E3,$DF,$18,$CB,$E7,$BE,$F3
	.dc.b	$AA,$E5,$39,$AE,$4F,$E3,$2F,$E4,$3F,$0A,$FC,$B5,$F5,$63,$58,$7E
	.dc.b	$29,$A7,$F2,$13,$EB,$96,$BF,$C2,$2F,$EA,$CC,$46,$DD,$90,$F9,$C6
	.dc.b	$8E,$A7,$F0,$FF,$67,$F0,$0F,$A4,$87,$EF,$1C,$BF,$40,$CD,$23,$35
	.dc.b	$F3,$67,$F9,$AC,$62,$B8,$D6,$E7,$C3,$7F,$0A,$B5,$1F,$9B,$6B,$F9
	.dc.b	$A6,$C6,$E3,$8E,$1F,$51,$F3,$18,$3C,$B3,$C2,$35,$5E,$19,$FC,$DD
	.dc.b	$F9,$2B,$51,$C0,$E4,$7B,$CE,$A7,$03,$CC,$6B,$0F,$E7,$91,$CA,$38
	.dc.b	$57,$7C,$3F,$27,$5F,$16,$79,$C6,$2F,$C5,$1C,$DF,$ED,$37,$CE,$BA
	.dc.b	$CE,$4F,$ED,$11,$99,$DC,$64,$FB,$85,$AF,$EC,$B3,$B8,$E2,$67,$0F
	.dc.b	$E9,$71,$98,$E2,$7B,$CF,$A4,$6C,$FB,$A5,$3D,$71,$F4,$EF,$DD,$87
	.dc.b	$C1,$8C,$61,$EE,$AE,$98,$7D,$B6,$CF,$D6,$C3,$C5,$3E,$89,$ED,$7C
	.dc.b	$DF,$CA,$E1,$DA,$B7,$7E,$A9,$DC,$61,$F3,$9D,$8F,$BC,$53,$EB,$BB
	.dc.b	$7D,$67,$6C,$3D,$18,$7D,$78,$FC,$09,$C6,$FE,$AB,$67,$E9,$B4,$F9
	.dc.b	$E5,$3E,$6C,$F3,$7E,$68,$7F,$CB,$9F,$0E,$F7,$FF,$D2,$D0,$F8,$37
	.dc.b	$C4,$61,$B7,$FE,$D0,$BE,$0D,$BF,$E0,$88,$46,$9F,$E1,$0D,$7F,$88
	.dc.b	$35,$FF,$19,$3B,$BF,$2C,$FF,$A7,$E3,$19,$F5,$9F,$37,$CC,$FF,$A8
	.dc.b	$FE,$1D,$F1,$1E,$7F,$89,$99,$7F,$B9,$A7,$2D,$9F,$3A,$8C,$BF,$C7
	.dc.b	$5A,$FF,$25,$6B,$FE,$47,$77,$7F,$E8,$56,$E1,$9A,$F9,$1E,$9F,$FA
	.dc.b	$DF,$E2,$5E,$C7,$CB,$F5,$66,$5F,$EF,$A9,$CB,$67,$CB,$18,$8F,$E6
	.dc.b	$0D,$7F,$9C,$35,$FF,$2D,$3B,$CB,$FD,$30,$DC,$73,$5F,$6F,$E7,$FF
	.dc.b	$D9,$5F,$11,$FE,$C1,$E1,$7C,$57,$DF,$33,$F1,$53,$3F,$BB,$48,$F9
	.dc.b	$2D,$8C,$C6,$0E,$0F,$E3,$F1,$FF,$37,$4F,$7F,$95,$FF,$6D,$7C,$4B
	.dc.b	$E2,$FF,$23,$EA,$7D,$EC,$3F,$2C,$D5,$3B,$7F,$05,$87,$F9,$E6,$31
	.dc.b	$1F,$D5,$5A,$FF,$5D,$6B,$BA,$FC,$0C,$70,$DF,$FD,$7C,$DA,$D5,$76
	.dc.b	$3C,$7E,$C7,$FB,$C7,$E2,$7F,$B4,$5F,$FB,$44,$62,$F8,$BF,$59,$BB
	.dc.b	$87,$E9,$AB,$F2,$29,$A4,$76,$7C,$93,$F5,$8F,$83,$0F,$D1,$57,$FA
	.dc.b	$92,$7C,$C7,$FA,$6B,$1F,$F5,$2F,$C1,$C8,$CF,$77,$CA,$6C,$FE,$8A
	.dc.b	$31,$1F,$DF,$5A,$FF,$80,$E3,$47,$FA,$9F,$2F,$F4,$83,$70,$FF,$51
	.dc.b	$BF,$EB,$4F,$A9,$F8,$DE,$07,$DC,$F8,$C9,$87,$9B,$4F,$E5,$CF,$C1
	.dc.b	$9F,$D9,$C7,$F6,$98,$FF,$0A,$35,$FF,$5B,$35,$FE,$1D,$F1,$8F,$84
	.dc.b	$F1,$46,$1A,$A3,$4F,$C6,$5F,$19,$A7,$31,$F3,$CE,$13,$93,$53,$D3
	.dc.b	$67,$C3,$52,$57,$D2,$F8,$CE,$BB,$FF,$23,$F8,$95,$FF,$B3,$CE,$35
	.dc.b	$F6,$1A,$F8,$5B,$2E,$0F,$F6,$B5,$F4,$A5,$F4,$7F,$98,$46,$63,$11
	.dc.b	$E0,$C3,$FD,$BD,$18,$C3,$BB,$AA,$E1,$1D,$B1,$FE,$8D,$7D,$A3,$D1
	.dc.b	$7C,$9B,$1E,$D7,$86,$36,$3C,$75,$CB,$0F,$05,$F1,$9F,$E8,$D7,$E2
	.dc.b	$38,$BD,$CE,$37,$CD,$F5,$16,$38,$3A,$38,$0C,$3D,$6F,$6A,$F0,$3C
	.dc.b	$8D,$C6,$9E,$18,$7C,$07,$FC,$C9,$1F,$3D,$B1,$97,$CE,$6B,$E2,$DF
	.dc.b	$19,$FD,$8A,$91,$89,$FF,$34,$46,$27,$53,$99,$FE,$D3,$F1,$9F,$C1
	.dc.b	$08,$FF,$56,$6E,$7F,$73,$3F,$D0,$69,$F1,$6B,$FB,$85,$7C,$79,$CC
	.dc.b	$7C,$43,$27,$C6,$73,$5F,$86,$8C,$BD,$75,$FC,$84,$71,$15,$D6,$33
	.dc.b	$3C,$67,$27,$58,$CE,$BF,$E8,$7A,$E1,$87,$9E,$B7,$7D,$E3,$13,$A9
	.dc.b	$F0,$36,$75,$1F,$B5,$C3,$DA,$79,$9F,$F0,$81,$D2,$FF,$8B,$61,$FC
	.dc.b	$D7,$0A,$7F,$26,$A7,$C6,$66,$FF,$C1,$95,$5F,$F1,$34,$5F,$8C,$FD
	.dc.b	$FD,$FF,$2E,$6C,$FF,$07,$7F,$C4,$11,$9D,$F8,$1D,$70,$FF,$00,$E1
	.dc.b	$89,$FE,$99,$18,$7A,$78,$7B,$F8,$C7,$8B,$E1,$8C,$36,$78,$CE,$9E
	.dc.b	$26,$23,$D5,$F1,$6F,$79,$F1,$7F,$E3,$68,$C3,$7B,$DF,$17,$9A,$DE
	.dc.b	$FF,$EB,$FF,$8B,$7C,$5F,$F8,$A1,$B1,$FE,$40,$6B,$B8,$FF,$16,$DB
	.dc.b	$DA,$A4,$9F,$F2,$3C,$65,$FF,$24,$FC,$5F,$F6,$A6,$F8,$7F,$E3,$68
	.dc.b	$C4,$7F,$B4,$5A,$FF,$B6,$5C,$F0,$FE,$0D,$BB,$FE,$11,$B3,$FE,$61
	.dc.b	$F8,$B7,$C5,$FF,$46,$9C,$C9,$A7,$0B,$57,$09,$FE,$DF,$F1,$5F,$EE
	.dc.b	$62,$7B,$1F,$E4,$C6,$C6,$A3,$FA,$9C,$1F,$F3,$7F,$C5,$3E,$2B,$31
	.dc.b	$2B,$81,$F8,$69,$D3,$FD,$FB,$E2,$7E,$E3,$FE,$70,$6B,$FE,$F5,$6B
	.dc.b	$FE,$FD,$6B,$FA,$67,$78,$FE,$A5,$BA,$67,$8B,$FD,$07,$F1,$3F,$D9
	.dc.b	$AF,$C1,$1F,$DB,$2B,$71,$FE,$7F,$6D,$AB,$84,$FF,$6B,$F8,$87,$C4
	.dc.b	$7F,$9C,$1B,$1F,$E8,$86,$BF,$BA,$3F,$D1,$4E,$1C,$3A,$AF,$E0,$0D
	.dc.b	$C7,$F8,$0B,$67,$FD,$3D,$F1,$1F,$9C,$7F,$C4,$B1,$94,$D3,$85,$27
	.dc.b	$84,$FF,$80,$FE,$1D,$F1,$1F,$E8,$86,$C7,$FA,$81,$AF,$A4,$38,$7D
	.dc.b	$8E,$AB,$F8,$C3,$70,$FE,$32,$D9,$FF,$5B,$7C,$43,$BF,$FC,$37,$39
	.dc.b	$9E,$B8,$60,$FF,$5C,$D6,$E3,$FC,$87,$F0,$EF,$88,$FF,$4F,$B6,$3F
	.dc.b	$D6,$ED,$7D,$0D,$7E,$93,$9C,$BF,$97,$37,$2F,$E5,$ED,$9F,$F6,$67
	.dc.b	$C4,$7F,$94,$DF,$F1,$54,$65,$34,$E1,$49,$E0,$FF,$3B,$F5,$3F,$4D
	.dc.b	$5F,$E5,$CF,$86,$7C,$3F,$F8,$F1,$B1,$FE,$5C,$E6,$FF,$1F,$D7,$08
	.dc.b	$FF,$1D,$4E,$9B,$5A,$A9,$1D,$63,$A1,$87,$FC,$85,$5C,$EF,$FA,$75
	.dc.b	$F3,$6D,$6D,$F4,$5F,$2B,$E8,$79,$48,$78,$8C,$9D,$9E,$59,$7F,$5A
	.dc.b	$BE,$CD,$BC,$D7,$FB,$B7,$0F,$24,$E6,$39,$E1,$FE,$FF,$3C,$AE,$23
	.dc.b	$FB,$2C,$F0,$8E,$58,$79,$DF,$21,$E3,$3C,$F5,$C9,$CB,$D6,$BC,$11
	.dc.b	$87,$AC,$F9,$A1,$87,$A9,$18,$C3,$58,$78,$E3,$07,$9C,$F2,$9E,$2B
	.dc.b	$EF,$87,$8C,$F9,$F1,$F3,$0F,$E1,$13,$F8,$16,$E1,$FE,$9D,$23,$CC
	.dc.b	$EA,$7E,$1B,$D3,$F1,$C7,$A7,$1F,$F4,$AD,$79,$49,$1F,$91,$9D,$D7
	.dc.b	$33,$DE,$E1,$F9,$47,$53,$53,$C9,$FC,$E4,$6E,$FD,$71,$B3,$8B,$F3
	.dc.b	$DF,$93,$7E,$4B,$E1,$87,$E1,$EE,$D7,$B2,$FF,$2F,$18,$86,$1F,$3F
	.dc.b	$0E,$51,$E4,$C3,$D8,$D3,$D9,$1A,$7C,$78,$7D,$E7,$D2,$6C,$C9,$E6
	.dc.b	$BF,$05,$F7,$6D,$F6,$9D,$9E,$47,$E8,$46,$A3,$A9,$A9,$F9,$0F,$C5
	.dc.b	$C3,$E2,$3C,$4D,$B8,$8F,$04,$79,$63,$E6,$9F,$11,$C5,$FA,$DE,$8D
	.dc.b	$8E,$B8,$7C,$16,$DF,$96,$B6,$7C,$FA,$E0,$47,$E1,$1D,$8D,$18,$3D
	.dc.b	$98,$7C,$07,$47,$AC,$EA,$5A,$F0,$BA,$C3,$DA,$F2,$3E,$BD,$F9,$67
	.dc.b	$DD,$19,$3E,$7E,$1E,$BC,$3C,$D1,$D5,$FC,$53,$D7,$0F,$AD,$5F,$2D
	.dc.b	$B3,$C3,$0F,$A7,$3D,$8D,$57,$D2,$9D,$57,$B6,$3C,$53,$F4,$A7,$BC
	.dc.b	$61,$F1,$DF,$96,$3E,$39,$E5,$3E,$44,$F6,$7E,$2D,$F8,$AB,$8E,$1F
	.dc.b	$05,$EB,$5B,$3C,$CF,$92,$BE,$39,$EF,$3D,$07,$82,$BA,$3E,$08,$E0
	.dc.b	$FC,$AA,$EC,$E4,$E1,$87,$BE,$FC,$15,$E2,$38,$BF,$06,$78,$1E,$D2
	.dc.b	$5F,$DC,$8E,$31,$A8,$F1,$BB,$C3,$E3,$3C,$2F,$EA,$BE,$23,$DA,$7C
	.dc.b	$D3,$E5,$C7,$09,$E6,$D9,$F4,$5F,$98,$F9,$75,$F1,$4F,$9F,$1F,$C2
	.dc.b	$E7,$84,$76,$BF,$3B,$FC,$88,$E0,$FD,$07,$C8,$7C,$93,$35,$F1,$5C
	.dc.b	$C7,$C5,$79,$1E,$93,$E0,$61,$F4,$1C,$C7,$82,$BC,$71,$D4,$E1,$87
	.dc.b	$CB,$BF,$64,$7C,$2B,$E5,$5C,$2F,$C0,$DA,$F8,$35,$CB,$0F,$9A,$FC
	.dc.b	$F3,$E0,$47,$47,$CA,$7B,$0F,$B1,$1F,$1E,$FB,$92,$3F,$B1,$DF,$FB
	.dc.b	$67,$E1,$5F,$0D,$FE,$49,$9F,$EC,$F1,$87,$0D,$7F,$C9,$CD,$7F,$CD
	.dc.b	$0E,$67,$A4,$FF,$B9,$1B,$3F,$ED,$0F,$7D,$F0,$D9,$CB,$FD,$D3,$87
	.dc.b	$16,$E1,$9B,$F1,$B7,$06,$1C,$A7,$D7,$5F,$EF,$2F,$85,$7C,$2F,$AE
	.dc.b	$51,$87,$DC,$D7,$FC,$D0,$D7,$FD,$04,$E6,$FF,$DF,$CD,$AE,$D5,$E3
	.dc.b	$BF,$F7,$3D,$F8,$8F,$F0,$7D,$6D,$B3,$EB,$7C,$6F,$1F,$2F,$1F,$0A
	.dc.b	$F8,$5F,$43,$6B,$EA,$35,$FF,$3F,$35,$FF,$4C,$39,$C9,$97,$F8,$82
	.dc.b	$53,$98,$C3,$61,$3F,$33,$E1,$7C,$77,$27,$D8,$72,$7E,$04,$7A,$FE
	.dc.b	$0D,$F0,$9F,$3D,$B5,$F8,$B6,$BF,$E9,$56,$BF,$EA,$F7,$37,$E8,$8F
	.dc.b	$F1,$BC,$63,$1F,$53,$63,$CF,$AB,$E1,$3C,$58,$49,$F3,$D6,$F0,$FC
	.dc.b	$3D,$7D,$1F,$82,$7C,$27,$E1,$5B,$7F,$A1,$6B,$FE,$AB,$6B,$FE,$C5
	.dc.b	$73,$5E,$78,$46,$32,$FA,$0D,$8F,$9B,$E6,$3C,$DF,$16,$12,$7E,$6D
	.dc.b	$6F,$05,$EE,$7E,$8F,$C0,$FE,$6D,$34,$8F,$E7,$DF,$F6,$D4,$F9,$5B
	.dc.b	$8F,$E0,$DB,$3E,$4F,$88,$FC,$A4,$62,$3A,$98,$3F,$61,$7F,$EA,$6A
	.dc.b	$FF,$3D,$E1,$FA,$DE,$BF,$83,$5F,$C2,$5A,$FF,$B4,$DA,$FF,$99,$A3
	.dc.b	$FD,$4E,$47,$FD,$5D,$F0,$95,$FF,$50,$C5,$CB,$FD,$43,$3E,$6E,$AF
	.dc.b	$71,$2B,$F8,$C6,$48,$C4,$7E,$73,$1F,$81,$1C,$63,$F6,$D7,$F3,$BD
	.dc.b	$2F,$DA,$7C,$1B,$D9,$A4,$AF,$D0,$C3,$E0,$F1,$1D,$4F,$F4,$6C,$71
	.dc.b	$77,$3D,$27,$F8,$07,$BC,$F7,$29,$2B,$F5,$BF,$07,$D7,$2F,$F4,$BF
	.dc.b	$C0,$2F,$F7,$47,$1A,$FD,$9B,$5F,$11,$BB,$FE,$1D,$7C,$E1,$5C,$DF
	.dc.b	$F5,$1C,$66,$31,$1C,$4F,$0D,$78,$2B,$F8,$54,$EA,$B8,$4F,$F0,$D6
	.dc.b	$D7,$FA,$BA,$3C,$93,$D1,$A6,$F0,$FE,$1F,$3E,$48,$C6,$1E,$08,$E5
	.dc.b	$87,$82,$7C,$51,$91,$87,$81,$B7,$D3,$0F,$05,$F1,$9E,$67,$1A,$E0
	.dc.b	$71,$BF,$2C,$68,$F2,$1F,$1A,$FA,$47,$9A,$3F,$96,$19,$C3,$FB,$45
	.dc.b	$7C,$F8,$CB,$5F,$8D,$F0,$4F,$83,$DB,$63,$99,$EC,$70,$F2,$6B,$AF
	.dc.b	$83,$FE,$04,$47,$FA,$F3,$73,$FB,$78,$7E,$6F,$83,$E4,$57,$FB,$8A
	.dc.b	$BE,$54,$E6,$3D,$E7,$F4,$78,$CE,$BF,$CC,$EF,$F0,$B3,$C4,$DD,$71
	.dc.b	$8C,$C2,$B8,$DF,$82,$30,$70,$BE,$99,$FA,$E3,$13,$BA,$E3,$1E,$43
	.dc.b	$C1,$7E,$CB,$E3,$1F,$4A,$77,$39,$AF,$EA,$F8,$7B,$4F,$E6,$D5,$CB
	.dc.b	$0F,$F0,$A6,$1F,$D3,$E7,$F9,$D1,$BA,$FF,$0C,$7C,$1E,$5D,$9D,$1C
	.dc.b	$58,$7C,$1F,$EF,$67,$FB,$3B,$63,$F6,$AE,$E3,$3B,$F2,$9E,$2F,$4B
	.dc.b	$FF,$01,$61,$9B,$E5,$1E,$7F,$03,$CD,$7E,$1F,$8C,$7C,$3F,$BA,$30
	.dc.b	$D5,$1A,$CB,$FC,$7C,$D8,$FC,$8F,$C2,$3D,$A7,$C2,$7F,$90,$E3,$0D
	.dc.b	$F0,$1F,$09,$9A,$DD,$FF,$3F,$F3,$BF,$C3,$DE,$7B,$D0,$A7,$F9,$E4
	.dc.b	$B3,$FE,$F9,$3F,$57,$E0,$AB,$FA,$61,$23,$FC,$A9,$F7,$4F,$B2,$FF
	.dc.b	$21,$7C,$17,$F4,$EE,$3F,$B4,$DE,$1F,$F1,$05,$6E,$FF,$C3,$2D,$7F
	.dc.b	$9C,$52,$B8,$47,$F8,$89,$AA,$78,$4F,$F7,$F7,$94,$7F,$02,$6B,$E0
	.dc.b	$BF,$F2,$84,$62,$3F,$0D,$83,$E6,$3F,$CB,$5F,$02,$F9,$FA,$7F,$91
	.dc.b	$67,$55,$FE,$30,$6B,$E8,$9E,$0F,$F5,$A6,$BF,$D0,$A3,$78,$FF,$85
	.dc.b	$5E,$31,$FC,$61,$AF,$F9,$15,$AF,$9A,$FF,$CD,$1F,$3F,$FD,$7E,$3F
	.dc.b	$B3,$FC,$83,$E8,$7F,$CE,$4D,$8F,$F2,$2C,$ED,$FE,$A6,$DC,$3F,$5D
	.dc.b	$3B,$9F,$A2,$D8,$FF,$15,$4E,$F6,$FF,$2C,$3C,$EB,$F9,$7B,$5C,$F8
	.dc.b	$9E,$63,$FC,$FD,$F3,$EF,$A7,$FF,$43,$7C,$9B,$C5,$A7,$F9,$C6,$75
	.dc.b	$7F,$E5,$C6,$BF,$84,$9E,$0F,$F8,$19,$AF,$F6,$E8,$DE,$5F,$E4,$17
	.dc.b	$8C,$7F,$44,$6B,$FE,$71,$6B,$FC,$1D,$AF,$F9,$FF,$BB,$FD,$39,$F0
	.dc.b	$1F,$DE,$EB,$F8,$A3,$74,$FF,$3D,$4F,$F9,$DE,$B8,$1F,$E1,$A9,$D3
	.dc.b	$FE,$7A,$6B,$A3,$FC,$0B,$1F,$E8,$77,$9B,$FE,$A5,$30,$FF,$1B,$6A
	.dc.b	$D7,$FD,$35,$6F,$8D,$FF,$44,$36,$BF,$D3,$93,$FE,$9F,$AE,$1E,$D7
	.dc.b	$FA,$DE,$F6,$7A,$67,$47,$FA,$BE,$7F,$D2,$CD,$7F,$96,$35,$E6,$7F
	.dc.b	$93,$7D,$0F,$E1,$BE,$3C,$8F,$F4,$AF,$57,$EE,$0F,$DC,$F6,$7F,$62
	.dc.b	$D7,$FD,$44,$D9,$FE,$71,$3B,$BE,$51,$FE,$2A,$9D,$29,$D3,$FE,$A2
	.dc.b	$8C,$9A,$8C,$C7,$1B,$E7,$A7,$FA,$16,$72,$FF,$A1,$EB,$8C,$FF,$97
	.dc.b	$70,$FF,$35,$DF,$58,$E3,$18,$8F,$F3,$19,$92,$18,$25,$75,$9E,$31
	.dc.b	$EE,$C3,$B3,$FD,$5E,$FC,$2F,$C8,$8C,$A8,$CB,$D2,$FC,$57,$D6,$3C
	.dc.b	$2F,$B6,$B8,$D7,$08,$C4,$F5,$37,$5F,$4A,$FB,$E1,$F5,$9F,$3C,$FC
	.dc.b	$FC,$39,$C6,$27,$E4,$D7,$1A,$E6,$7B,$4E,$C7,$D1,$32,$73,$33,$7E
	.dc.b	$97,$47,$B4,$E4,$61,$D1,$EA,$74,$7C,$5A,$F2,$5F,$9A,$3B,$DF,$99
	.dc.b	$F9,$D3,$A7,$C3,$7E,$97,$B3,$F3,$08,$F0,$39,$D7,$57,$67,$9A,$FC
	.dc.b	$87,$9A,$7A,$38,$79,$C7,$9C,$E6,$78,$8E,$C7,$43,$E3,$1E,$EC,$3E
	.dc.b	$D4,$72,$C3,$EB,$4F,$23,$58,$7B,$5A,$66,$FC,$4F,$67,$CC,$FA,$6B
	.dc.b	$CB,$7D,$A3,$04,$3D,$07,$96,$FE,$63,$8A,$E6,$FA,$1E,$EE,$6B,$E1
	.dc.b	$9E,$9C,$3F,$1B,$7F,$B7,$7E,$65,$72,$BE,$B3,$E3,$BE,$97,$E6,$C3
	.dc.b	$DD,$7D,$AB,$57,$C2,$73,$5B,$BF,$84,$D7,$D5,$87,$82,$3C,$B1,$EF
	.dc.b	$AF,$9A,$E8,$FA,$C7,$AC,$F9,$51,$EB,$7F,$3B,$7E,$B7,$6F,$92,$BB
	.dc.b	$47,$03,$B3,$F0,$AF,$F6,$2D,$8E,$F7,$E2,$C3,$E2,$4F,$29,$F1,$1F
	.dc.b	$10,$FA,$D7,$DD,$F0,$DF,$C2,$72,$D7,$81,$CA,$3D,$0F,$E2,$0F,$80
	.dc.b	$5A,$F0,$BD,$0F,$96,$74,$29,$D2,$3C,$4F,$1C,$3E,$6B,$E0,$C3,$D9
	.dc.b	$18,$7E,$C4,$72,$72,$FC,$73,$86,$1F,$15,$FE,$03,$1A,$7E,$1B,$C9
	.dc.b	$F4,$DF,$E1,$70,$F8,$13,$BC,$3D,$F3,$9A,$F0,$5F,$58,$FC,$CD,$6E
	.dc.b	$FE,$AD,$7D,$E6,$BF,$B7,$BE,$78,$7E,$36,$FE,$3C,$FC,$13,$D5,$7E
	.dc.b	$58,$C9,$E1,$8F,$3C,$F1,$3E,$26,$1F,$7E,$FA,$CF,$CF,$36,$5A,$E7
	.dc.b	$5C,$63,$BD,$71,$C3,$CF,$5F,$B6,$BE,$75,$FC,$38,$FF,$3F,$D7,$FB
	.dc.b	$81,$FE,$62,$60,$FE,$FC,$7E,$51,$FF,$45,$FA,$1D,$30,$FE,$12,$D9
	.dc.b	$FC,$9B,$5F,$C2,$37,$1F,$F0,$1B,$6B,$9C,$F5,$3F,$53,$19,$7F,$BF
	.dc.b	$35,$FF,$02,$37,$0F,$EF,$4D,$BF,$F0,$6B,$67,$FB,$73,$5F,$ED,$8D
	.dc.b	$C0,$D3,$FD,$95,$0A,$56,$3D,$D7,$83,$A9,$FF,$6C,$7B,$0F,$47,$17
	.dc.b	$FE,$D3,$94,$E7,$0C,$4F,$3B,$DC,$7F,$B6,$70,$EA,$E8,$F1,$57,$17
	.dc.b	$FD,$4E,$DB,$F2,$C7,$1A,$E6,$AB,$9B,$EC,$34,$7D,$F7,$2E,$5A,$FC
	.dc.b	$2F,$71,$F4,$1E,$6B,$F5,$FC,$45,$C4,$F8,$2B,$77,$E9,$0E,$1F,$4C
	.dc.b	$71,$38,$5C,$8F,$C9,$4E,$DF,$43,$5F,$24,$7E,$5D,$7B,$AF,$A6,$F2
	.dc.b	$73,$49,$F1,$4E,$5B,$39,$7E,$5C,$FC,$89,$D4,$7E,$07,$5F,$C1,$C6
	.dc.b	$23,$E5,$35,$CD,$78,$E3,$10,$1E,$E3,$EB,$FE,$47,$D0,$49,$F8,$D3
	.dc.b	$96,$CE,$5F,$BF,$3F,$72,$75,$1E,$53,$B3,$87,$EF,$35,$FC,$DB,$5D
	.dc.b	$CF,$C9,$E4,$7C,$D7,$DD,$F7,$1F,$39,$4F,$C5,$CE,$AF,$EE,$35,$FF
	.dc.b	$7C,$CF,$07,$F3,$1C,$48,$C4,$79,$C4,$F1,$9F,$CC,$C6,$2F,$C8,$47
	.dc.b	$F8,$A4,$F7,$7F,$35,$EE,$FF,$AB,$AB,$B7,$D3,$7A,$4F,$D0,$C7,$FA
	.dc.b	$C0,$D6,$08,$D5,$7E,$56,$75,$EB,$FE,$2D,$34,$91,$FC,$96,$3F,$61
	.dc.b	$C1,$FB,$DB,$7C,$BF,$C9,$3C,$3D,$32,$FE,$1C,$E9,$B3,$97,$E2,$5F
	.dc.b	$EF,$5F,$6D,$78,$C8,$FF,$35,$FA,$77,$AF,$FB,$3E,$57,$9D,$4F,$E0
	.dc.b	$33,$AA,$FD,$93,$5F,$C4,$6C,$AE,$31,$F6,$1A,$FD,$79,$E1,$3F,$5B
	.dc.b	$DF,$7B,$5F,$C3,$9B,$5F,$BF,$6B,$F7,$D7,$97,$B9,$93,$B7,$94,$7F
	.dc.b	$18,$F0,$FF,$DC,$E4,$E1,$FE,$37,$EF,$F2,$C7,$F5,$9E,$EA,$BF,$9B
	.dc.b	$9B,$C7,$F8,$83,$FE,$F0,$AE,$10,$9E,$0F,$ED,$63,$31,$88,$E9,$7F
	.dc.b	$D0,$A3,$15,$F8,$49,$F0,$BF,$D1,$AB,$8B,$70,$FE,$91,$3D,$E3,$CB
	.dc.b	$7C,$DB,$1C,$8F,$14,$F0,$BF,$15,$F4,$9F,$E1,$55,$E1,$7A,$18,$BE
	.dc.b	$8F,$2C,$3A,$2A,$E4,$FB,$0C,$D7,$3B,$EF,$87,$86,$3D,$57,$AC,$3A
	.dc.b	$DF,$13,$C2,$75,$AE,$0F,$84,$FE,$9C,$66,$7F,$C1,$98,$7D,$28,$FE
	.dc.b	$A4,$D7,$E5,$79,$8F,$7F,$6D,$8E,$72,$8C,$47,$26,$BE,$0F,$7F,$FA
	.dc.b	$22,$69,$1F,$E9,$0F,$C6,$BF,$CD,$F3,$7C,$89,$FE,$71,$39,$D7,$FB
	.dc.b	$6E,$BF,$85,$9C,$9F,$0A,$F8,$BF,$10,$CC,$7E,$2E,$32,$F1,$8F,$21
	.dc.b	$C6,$7A,$EB,$E9,$8C,$61,$E3,$77,$5C,$EB,$71,$FD,$EF,$0F,$35,$6D
	.dc.b	$B1,$E3,$38,$DF,$F8,$5E,$B8,$C7,$F8,$75,$FF,$04,$E1,$FC,$C8,$F5
	.dc.b	$9F,$CD,$23,$D7,$3C,$6F,$FB,$E4,$E9,$FF,$11,$7B,$FC,$9D,$57,$F3
	.dc.b	$F9,$4E,$63,$FB,$27,$BF,$FC,$BA,$BF,$CA,$0D,$8F,$C4,$B5,$FF,$0C
	.dc.b	$35,$FF,$1A,$BB,$AF,$4F,$17,$8E,$7A,$F0,$77,$9F,$E3,$D9,$78,$65
	.dc.b	$EF,$3D,$C7,$BD,$FE,$61,$8C,$37,$CA,$7B,$D9,$AD,$DF,$E7,$BE,$5D
	.dc.b	$F0,$3F,$E1,$36,$C7,$F9,$AD,$AE,$E3,$FC,$31,$E8,$3D,$BF,$F1,$BF
	.dc.b	$A2,$F8,$0A,$7F,$8E,$A7,$57,$FE,$2B,$6B,$FA,$E6,$BF,$E3,$D6,$BF
	.dc.b	$E5,$A6,$BF,$E1,$37,$8E,$FF,$DC,$1B,$C7,$BC,$3C,$7E,$07,$F9,$2B
	.dc.b	$DE,$7B,$FF,$E4,$1F,$80,$FE,$F3,$97,$F7,$E6,$BF,$DF,$67,$6F,$F9
	.dc.b	$05,$AF,$F0,$46,$BF,$E5,$36,$BF,$E7,$06,$BF,$E2,$97,$8E,$5F,$E0
	.dc.b	$66,$E3,$BB,$FF,$2A,$7C,$03,$EA,$BF,$CC,$39,$BE,$07,$3E,$2C,$CF
	.dc.b	$FC,$A9,$39,$6C,$FF,$92,$E3,$11,$FC,$69,$AF,$F9,$A5,$AF,$FA,$19
	.dc.b	$DE,$7F,$E1,$A6,$E5,$9A,$F2,$F6,$7F,$9D,$3E,$03,$FB,$1E,$1F,$E2
	.dc.b	$6F,$69,$F0,$3F,$E6,$D6,$C7,$F9,$A6,$76,$FF,$8A,$5A,$FF,$9E,$DA
	.dc.b	$FF,$A6,$1A,$FF,$94,$5A,$9E,$79,$7F,$8C,$1B,$8F,$1F,$1B,$FC,$FD
	.dc.b	$F8,$FF,$FA,$1E,$BF,$C6,$BF,$00,$F8,$2F,$F4,$17,$C1,$7F,$33,$CB
	.dc.b	$FC,$7E,$D7,$FC,$7D,$3B,$7F,$D0,$4D,$7F,$A2,$B5,$FF,$4A,$35,$FF
	.dc.b	$58,$B5,$FF,$35,$3C,$6B,$C7,$5F,$E9,$E8,$C6,$3F,$EA,$FF,$92,$7D
	.dc.b	$37,$FA,$8B,$C5,$FF,$0F,$63,$FE,$54,$F8,$4F,$F2,$41,$FD,$AF,$2F
	.dc.b	$F2,$8C,$EA,$7F,$CA,$CD,$7F,$D5,$0D,$7F,$D8,$AD,$7F,$CF,$CD,$7F
	.dc.b	$C5,$8F,$84,$8F,$F9,$93,$F1,$7F,$F2,$93,$FE,$A5,$E0,$7C,$2F,$FA
	.dc.b	$B7,$E1,$7C,$58,$7F,$34,$6E,$9F,$ED,$4A,$DD,$F1,$BF,$F5,$14,$62
	.dc.b	$3F,$B8,$B5,$FF,$60,$35,$E5,$1F,$EC,$68,$E1,$7F,$E7,$26,$D7,$FB
	.dc.b	$67,$E3,$DF,$37,$FC,$4F,$DA,$FE,$2F,$E7,$4C,$BF,$DF,$07,$F9,$F7
	.dc.b	$C4,$FE,$C1,$87,$FB,$A9,$B1,$FC,$46,$B9,$3F,$EC,$58,$CA,$6F,$87
	.dc.b	$FC,$33,$3F,$E2,$D5,$DE,$BB,$61,$C9,$FF,$46,$56,$F0,$F0,$4E,$EF
	.dc.b	$8D,$7F,$29,$BC,$56,$A3,$D3,$3F,$E9,$2C,$3C,$67,$8A,$BB,$46,$6F
	.dc.b	$8D,$7A,$CE,$87,$AD,$CA,$C3,$C8,$EC,$F4,$9B,$3C,$38,$78,$DC,$9E
	.dc.b	$CB,$F0,$C6,$6F,$C5,$3C,$0F,$F6,$76,$1E,$8C,$37,$1D,$E7,$6F,$47
	.dc.b	$E1,$BD,$4F,$11,$9A,$FE,$91,$7C,$A3,$EA,$57,$B3,$0F,$5C,$7B,$67
	.dc.b	$DF,$87,$C7,$8F,$86,$73,$3C,$B5,$CE,$B8,$CF,$9E,$7D,$47,$63,$C7
	.dc.b	$87,$98,$98,$7C,$12,$47,$B5,$F3,$BC,$1C,$BE,$53,$C7,$1F,$08,$91
	.dc.b	$C2,$3E,$11,$E7,$3C,$46,$23,$C2,$FA,$8E,$C7,$84,$87,$07,$AB,$8A
	.dc.b	$F7,$DF,$C9,$C3,$E6,$61,$F0,$6B,$B6,$1E,$3C,$38,$C7,$38,$C1,$1F
	.dc.b	$11,$E7,$C3,$E9,$9D,$9F,$29,$F0,$4E,$A7,$09,$EB,$1C,$2F,$EC,$E1
	.dc.b	$F5,$5B,$5F,$53,$0F,$35,$FC,$28,$E8,$F9,$A3,$F0,$11,$D8,$F4,$1D
	.dc.b	$F0,$F4,$3C,$1F,$AB,$1C,$27,$A3,$DA,$37,$87,$AE,$BD,$D7,$F2,$30
	.dc.b	$F9,$51,$F7,$2B,$9C,$7A,$63,$E0,$DF,$59,$F0,$BD,$63,$EA,$57,$B9
	.dc.b	$F4,$5F,$7C,$3E,$2C,$F7,$9F,$97,$3F,$1D,$EB,$87,$77,$F1,$17,$DD
	.dc.b	$FC,$1B,$DD,$F8,$06,$23,$D2,$FC,$78,$D4,$F5,$8F,$F1,$E5,$7F,$91
	.dc.b	$88,$FF,$AB,$0F,$B8,$7F,$8F,$DC,$BC,$0D,$47,$3B,$F9,$BF,$0F,$E7
	.dc.b	$6C,$FE,$C2,$7F,$67,$39,$C7,$E3,$B6,$7D,$6D,$7D,$67,$F3,$77,$13
	.dc.b	$CA,$3A,$5F,$F9,$72,$72,$D8,$C9,$FD,$06,$32,$D3,$27,$F4,$B7,$2D
	.dc.b	$33,$7F,$CD,$E3,$2D,$8C,$DF,$F5,$19,$CB,$63,$27,$F2,$88,$CB,$4C
	.dc.b	$9E,$D2,$E0,$8D,$C6,$4F,$C4,$DF,$EF,$23,$33,$B9,$E3,$18,$23,$82
	.dc.b	$1C,$8F,$CE,$1E,$33,$07,$E1,$DE,$2E,$2F,$ED,$B6,$FE,$CB,$D2,$3F
	.dc.b	$71,$87,$E0,$E7,$E1,$4E,$CF,$BA,$6D,$FD,$8E,$1E,$B8,$C1,$FA,$83
	.dc.b	$CA,$7C,$1A,$EB,$1B,$8E,$87,$E1,$B0,$F8,$4E,$2F,$F0,$CD,$9F,$C2
	.dc.b	$34,$FC,$11,$4F,$C0,$16,$BE,$F3,$6F,$D7,$1E,$28,$F2,$DF,$82,$31
	.dc.b	$1F,$24,$F5,$8F,$87,$79,$68,$93,$FE,$E4,$F8,$77,$8F,$99,$4E,$63
	.dc.b	$0B,$1D,$61,$D2,$B9,$C6,$30,$FD,$D4,$78,$5B,$3B,$8E,$C7,$8D,$C3
	.dc.b	$87,$55,$DE,$35,$3E,$2A,$ED,$1E,$03,$CF,$87,$FB,$35,$C0,$F8,$67
	.dc.b	$C3,$71,$8D,$FC,$3D,$C5,$68,$D4,$F1,$AB,$7C,$A3,$F8,$B3,$5F,$59
	.dc.b	$B3,$31,$F0,$A3,$57,$EB,$9D,$1F,$3B,$0F,$A1,$3A,$7E,$AE,$1F,$E7
	.dc.b	$E9,$F5,$FC,$33,$E1,$FD,$13,$FE,$68,$9F,$83,$D3,$E8,$AD,$47,$0A
	.dc.b	$FA,$71,$88,$F3,$B5,$FA,$6D,$7F,$04,$EF,$1F,$43,$70,$CD,$78,$8F
	.dc.b	$8F,$F0,$CF,$86,$F2,$FC,$3C,$C7,$E9,$CE,$5B,$3F,$86,$8C,$47,$CA
	.dc.b	$6B,$F8,$66,$BF,$95,$77,$8F,$CB,$6E,$19,$AF,$B9,$F0,$9F,$3E,$BE
	.dc.b	$87,$23,$E1,$B3,$F0,$F3,$2F,$C3,$CE,$5B,$3F,$9A,$8C,$47,$DE,$6B
	.dc.b	$F9,$A6,$BF,$A9,$77,$5E,$09,$FC,$5C,$63,$1F,$D8,$E0,$F5,$BF,$1F
	.dc.b	$F0,$BF,$EB,$79,$CF,$C3,$FA,$8C,$FB,$B3,$3F,$EC,$D4,$8F,$D7,$B6
	.dc.b	$33,$18,$38,$3F,$8F,$3F,$81,$47,$77,$F7,$F8,$3E,$8B,$C7,$D5,$F2
	.dc.b	$BE,$22,$57,$EB,$2F,$F8,$3B,$67,$C2,$47,$CA,$D7,$C0,$23,$8B,$86
	.dc.b	$ED,$FB,$F1,$83,$DA,$FD,$9F,$C3,$BE,$23,$F7,$1D,$0F,$5A,$9A,$89
	.dc.b	$97,$F8,$2E,$B8,$36,$3F,$8A,$C6,$23,$F7,$3F,$10,$F8,$2F,$E0,$2D
	.dc.b	$8F,$E1,$6D,$7D,$6B,$D3,$DC,$C9,$DB,$CA,$3F,$85,$F8,$DF,$3E,$4D
	.dc.b	$31,$F0,$FC,$F6,$FE,$6D,$F0,$DA,$AF,$E7,$15,$BE,$4E,$67,$F1,$6B
	.dc.b	$E1,$2A,$E0,$A3,$31,$8A,$FC,$34,$ED,$FE,$80,$62,$38,$CF,$08,$FE
	.dc.b	$88,$D9,$F1,$D7,$F4,$7A,$E8,$DA,$DE,$1F,$D2,$AB,$B1,$FD,$46,$F9
	.dc.b	$3F,$0A,$B8,$3D,$0F,$83,$5C,$B0,$B1,$C6,$FA,$61,$E0,$8C,$9C,$CB
	.dc.b	$87,$63,$BD,$FB,$4F,$7E,$1E,$09,$F8,$D7,$D2,$7C,$11,$FC,$F2,$BE
	.dc.b	$6B,$FD,$61,$AE,$8C,$7C,$3B,$E2,$3F,$CB,$52,$8C,$4F,$36,$BC,$27
	.dc.b	$FB,$3A,$F8,$8A,$47,$FB,$5B,$72,$FB,$B9,$FE,$5F,$A7,$E0,$D7,$F7
	.dc.b	$DA,$F8,$73,$98,$FA,$27,$F7,$D8,$CE,$BF,$DB,$6F,$F0,$F3,$C4,$DD
	.dc.b	$71,$8C,$C7,$94,$CC,$F1,$CF,$FB,$0D,$F4,$C3,$C7,$5B,$BF,$3D,$EE
	.dc.b	$3F,$C0,$35,$B7,$CA,$78,$67,$75,$CA,$75,$18,$9F,$E1,$78,$7B,$4F
	.dc.b	$ED,$8F,$1C,$3F,$C2,$F8,$7F,$0B,$C2,$9F,$E0,$4A,$7B,$79,$BF,$F1
	.dc.b	$1D,$57,$FC,$43,$17,$E2,$1F,$09,$DA,$7F,$C5,$51,$8C,$3F,$D3,$0D
	.dc.b	$8F,$F0,$1B,$5F,$F0,$FF,$C4,$7F,$3F,$F8,$8F,$F4,$F3,$63,$FC,$12
	.dc.b	$FF,$8D,$BE,$23,$3E,$02,$BF,$D4,$9C,$BF,$C4,$75,$FE,$4D,$F3,$BF
	.dc.b	$E0,$5C,$7F,$9E,$42,$30,$FF,$83,$E7,$86,$7F,$E3,$D6,$D7,$F8,$FA
	.dc.b	$B7,$3F,$E4,$46,$C7,$F9,$1E,$76,$9E,$35,$DF,$DA,$FF,$15,$57,$0A
	.dc.b	$D6,$FF,$C8,$DF,$2C,$74,$AE,$11,$FC,$98,$86,$AB,$99,$C6,$3D,$55
	.dc.b	$89,$FF,$20,$47,$27,$F9,$9C,$6F,$0F,$55,$6E,$72,$76,$AF,$93,$19
	.dc.b	$7A,$C6,$67,$8B,$DA,$FD,$98,$7C,$23,$E0,$BE,$B3,$38,$78,$67,$E6
	.dc.b	$61,$D6,$FE,$84,$E7,$0F,$41,$E2,$AF,$01,$F3,$E3,$83,$C8,$C9,$F3
	.dc.b	$0F,$25,$69,$C1,$EE,$3E,$51,$A3,$8E,$1E,$88,$E0,$FA,$E7,$BD,$79
	.dc.b	$2F,$CF,$7E,$48,$C3,$E3,$3B,$3F,$21,$EE,$47,$04,$8F,$97,$19,$7A
	.dc.b	$BC,$9C,$C7,$C3,$8C,$1F,$0C,$F5,$1C,$4C,$1A,$7C,$A7,$99,$F8,$6F
	.dc.b	$85,$F4,$3D,$4C,$3A,$8F,$09,$1E,$B1,$C4,$EB,$87,$DE,$8F,$88,$FE
	.dc.b	$07,$0F,$C0,$47,$33,$BD,$73,$70,$66,$B9,$CF,$0C,$3E,$A4,$FB,$A7
	.dc.b	$CB,$18,$8F,$80,$72,$71,$1E,$BA,$F9,$6F,$9E,$FC,$95,$F8,$0C,$3E
	.dc.b	$EC,$7A,$63,$E4,$3A,$7E,$79,$E5,$71,$3E,$29,$E3,$87,$9D,$B5,$D4
	.dc.b	$F4,$5F,$BF,$0F,$C6,$46,$0F,$86,$7C,$C8,$F3,$9F,$3E,$FE,$C4,$F0
	.dc.b	$9F,$BD,$18,$AF,$D4,$61,$E4,$BE,$35,$CC,$EB,$5B,$7E,$AE,$1F,$9E
	.dc.b	$BE,$2D,$8C,$9E,$53,$E6,$47,$B1,$F8,$78,$7C,$EB,$F6,$4F,$29,$E1
	.dc.b	$3E,$7B,$FA,$51,$EE,$39,$9F,$05,$F2,$3F,$42,$FC,$4D,$39,$D7,$57
	.dc.b	$EF,$D7,$5B,$F8,$D8,$7D,$9C,$3E,$F6,$1C,$9B,$7C,$8F,$69,$CB,$0E
	.dc.b	$91,$E6,$8F,$29,$F3,$AB,$A5,$FC,$C9,$FC,$AC,$FD,$D3,$57,$F8,$17
	.dc.b	$D9,$5C,$1F,$A5,$3C,$1F,$93,$7F,$2E,$3F,$15,$1C,$2B,$93,$D0,$ED
	.dc.b	$1F,$28,$F8,$E7,$B1,$F5,$1E,$C7,$D5,$87,$C5,$8F,$C0,$DF,$8D,$FC
	.dc.b	$D5,$7C,$5C,$3D,$93,$F8,$99,$FC,$41,$A3,$E4,$1F,$CA,$41,$FC,$58
	.dc.b	$A7,$F1,$02,$9F,$CC,$0B,$5F,$E8,$56,$E1,$A7,$91,$0F,$0C,$7A,$5F
	.dc.b	$ED,$9C,$DF,$13,$1F,$E0,$2D,$C3,$F3,$F3,$A3,$E3,$B8,$AF,$F4,$94
	.dc.b	$60,$EB,$3C,$5A,$64,$FE,$14,$E5,$A6,$63,$FA,$F4,$65,$B1,$9B,$FF
	.dc.b	$4E,$34,$FF,$52,$CE,$E3,$31,$F8,$23,$FA,$8C,$65,$D9,$C4,$FF,$55
	.dc.b	$B9,$69,$9B,$FF,$0D,$46,$5B,$19,$BF,$F0,$E4,$E5,$B1,$93,$FC,$35
	.dc.b	$19,$69,$93,$E6,$1E,$73,$F8,$0B,$B7,$33,$F9,$A7,$27,$1B,$FA,$2F
	.dc.b	$18,$FB,$2E,$67,$94,$EA,$79,$1A,$9E,$15,$F6,$EB,$64,$8E,$44,$36
	.dc.b	$7D,$C2,$9F,$68,$B5,$F6,$1B,$7F,$52,$FC,$37,$DE,$7A,$46,$1E,$8D
	.dc.b	$7A,$36,$BA,$36,$BA,$57,$E4,$DF,$EB,$97,$E9,$BF,$09,$F8,$02,$9F
	.dc.b	$78,$B5,$F7,$1B,$5F,$69,$A7,$D8,$29,$F5,$1F,$03,$F8,$E3,$D6,$E1
	.dc.b	$FA,$E7,$0B,$FA,$51,$99,$FA,$47,$AE,$7C,$CD,$7C,$CD,$3C,$C5,$8F
	.dc.b	$33,$63,$CC,$D7,$CC,$7F,$00,$BF,$8A,$F9,$8C,$1E,$93,$CD,$7F,$26
	.dc.b	$79,$5F,$A2,$76,$7B,$23,$33,$BB,$FE,$2F,$D0,$F2,$3F,$89,$5F,$F1
	.dc.b	$C8,$C3,$86,$BF,$C5,$DA,$FF,$1E,$73,$6A,$EB,$5F,$C9,$23,$15,$D6
	.dc.b	$31,$3F,$CA,$A5,$D8,$F9,$BC,$DF,$FB,$FC,$4F,$27,$FD,$F4,$62,$BF
	.dc.b	$DE,$EF,$37,$FD,$EB,$CD,$FF,$00,$63,$47,$39,$E5,$CC,$EB,$EF,$3F
	.dc.b	$EF,$E9,$FE,$7B,$4A,$E6,$76,$9F,$F7,$3B,$C5,$DB,$A7,$FA,$2F,$5B
	.dc.b	$5F,$4C,$70,$C3,$FA,$3D,$F9,$4E,$D1,$C2,$77,$18,$8F,$21,$C4,$EE
	.dc.b	$7C,$57,$55,$DF,$0C,$CF,$09,$ED,$87,$A5,$DD,$F6,$8F,$44,$FC,$C9
	.dc.b	$DB,$E3,$AE,$77,$D6,$3F,$AC,$35,$FE,$BE,$D7,$FB,$1B,$5F,$87,$D6
	.dc.b	$EB,$FC,$7C,$A3,$13,$D1,$AF,$46,$BF,$D9,$FA,$FF,$2E,$FF,$72,$6A
	.dc.b	$69,$16,$1F,$2F,$07,$17,$E5,$23,$E7,$4F,$33,$79,$7E,$66,$74,$F8
	.dc.b	$63,$27,$E6,$63,$2F,$F8,$02,$FE,$8D,$F6,$D7,$E9,$CF,$5C,$3C,$57
	.dc.b	$C2,$3F,$C1,$37,$C2,$31,$1B,$C3,$C5,$7C,$A3,$DC,$78,$8F,$D2,$CF
	.dc.b	$07,$37,$DA,$78,$5F,$56,$4F,$E9,$0E,$2F,$EA,$5F,$51,$D7,$0F,$53
	.dc.b	$B6,$9F,$B3,$1D,$7F,$BE,$71,$85,$AF,$E0,$53,$6B,$FC,$53,$D2,$E8
	.dc.b	$74,$7F,$12,$8C,$37,$D0,$75,$66,$B7,$7F,$E2,$9F,$4D,$E3,$FE,$11
	.dc.b	$B1,$FC,$71,$AE,$E3,$F1,$3D,$5F,$E3,$38,$FF,$18,$1F,$E3,$72,$BF
	.dc.b	$E3,$13,$FA,$DE,$9F,$D0,$CF,$EC,$91,$FC,$C8,$FE,$6C,$7F,$9A,$36
	.dc.b	$FE,$8F,$1B,$8F,$EC,$ED,$7F,$B3,$B6,$7A,$C7,$F3,$89,$E2,$D9,$EF
	.dc.b	$1F,$CE,$A3,$BC,$F9,$76,$FE,$DB,$1B,$3B,$94,$EE,$58,$EE,$F9,$EB
	.dc.b	$A9,$E3,$9E,$71,$93,$99,$4E,$65,$9E,$67,$8D,$FE,$CB,$3C,$DA,$78
	.dc.b	$67,$AC,$EE,$3F,$B4,$CE,$DA,$6C,$E2,$77,$9E,$27,$F9,$47,$97,$FA
	.dc.b	$E4,$6C,$E2,$6C,$E3,$5D,$EF,$FB,$AD,$F7,$6C,$F1,$6B,$C4,$E7,$3C
	.dc.b	$67,$FB,$D9,$C4,$B1,$C5,$AF,$19,$E7,$3C,$63,$FC,$03,$3C,$5A,$71
	.dc.b	$29,$C5,$E7,$1C,$63,$FC,$0E,$64,$DB,$93,$FC,$04,$E4,$DB,$9A,$F8
	.dc.b	$F1,$98,$D9,$97,$F1,$B7,$F3,$5A,$7F,$A0,$67,$8C,$EE,$3F,$44,$4B
	.dc.b	$FC,$B3,$6F,$DC,$F8,$AF,$A3,$93,$E0,$19,$3B,$1E,$03,$FC,$AF,$7E
	.dc.b	$08,$E3,$5D,$89,$5D,$AF,$D7,$3A,$8C,$57,$1B,$EA,$78,$4F,$84,$E5
	.dc.b	$EE,$78,$6F,$FC,$B9,$19,$6C,$66,$7F,$75,$39,$6C,$64,$FD,$84,$65
	.dc.b	$A6,$4F,$D5,$39,$69,$9B,$FE,$25,$19,$6C,$66,$FF,$C8,$EF,$AE,$3E
	.dc.b	$A5,$FC,$48,$F9,$B1,$F1,$1F,$83,$19,$7E,$09,$CC,$FF,$27,$39,$8E
	.dc.b	$77,$F2,$A7,$47,$F9,$35,$C3,$E1,$AF,$A0,$F8,$2F,$D1,$1E,$6B,$E8
	.dc.b	$EA,$7C,$27,$13,$CB,$5D,$27,$71,$83,$A1,$E2,$7F,$93,$BC,$9D,$BF
	.dc.b	$68,$E8,$57,$27,$DA,$72,$D3,$31,$F6,$A3,$2D,$8C,$C7,$DA,$3F,$15
	.dc.b	$D5,$FE,$62,$9F,$F3,$0C,$7F,$6B,$11,$97,$FC,$B7,$D3,$FE,$5F,$75
	.dc.b	$19,$77,$D5,$FE,$54,$8E,$17,$FE,$66,$EA,$FF,$2B,$F3,$7F,$94,$6F
	.dc.b	$FB,$D3,$67,$FC,$BB,$D5,$FE,$5C,$E9,$FF,$1C,$74,$7F,$98,$E7,$37
	.dc.b	$E5,$BE,$1A,$7F,$90,$AF,$71,$FE,$47,$24,$FF,$90,$CF,$F0,$84,$69
	.dc.b	$FF,$29,$46,$9F,$F2,$83,$FE,$4C,$3C,$D7,$FE,$4C,$8C,$9C,$DC,$46
	.dc.b	$8F,$16,$DF,$E5,$59,$EA,$43,$89,$FE,$64,$3F,$CA,$2F,$F9,$B9,$B8
	.dc.b	$7F,$93,$A3,$11,$FE,$54,$E9,$FF,$25,$BF,$E5,$4E,$AF,$F2,$9C,$62
	.dc.b	$18,$FF,$95,$67,$FC,$95,$1F,$EA,$6E,$AA,$FF,$90,$CC,$1F,$E2,$7E
	.dc.b	$1F,$F1,$94,$7F,$8B,$B3,$FF,$26,$CF,$F8,$D3,$3F,$F1,$DB,$FE,$39
	.dc.b	$3F,$C7,$C7,$F9,$2F,$4F,$F1,$BB,$CD,$B3,$FE,$3B,$8E,$07,$33,$81
	.dc.b	$8A,$F0,$57,$F9,$5E,$78,$1E,$28,$CC,$87,$A0,$F7,$BF,$D5,$D1,$93
	.dc.b	$2E,$1E,$11,$BC,$3F,$D9,$2D,$8E,$07,$F8,$FB,$9F,$FC,$6E,$E3,$AF
	.dc.b	$FC,$7A,$E8,$2E,$AF,$F1,$FC,$66,$36,$F4,$CF,$F2,$EF,$A5,$DB,$FA
	.dc.b	$2E,$AF,$D2,$F1,$C8,$FC,$F4,$7E,$A0,$C3,$FA,$88,$C3,$A3,$F4,$8E
	.dc.b	$B8,$FF,$D8,$F3,$A9,$E0,$68,$E0,$78,$4C,$1A,$30,$48,$FD,$8B,$97
	.dc.b	$AB,$C9,$C9,$33,$FA,$57,$EE,$8D,$CF,$19,$FF,$6A,$46,$08,$E0,$91
	.dc.b	$F1,$EB,$D5,$3D,$63,$FD,$C9,$3D,$A7,$2F,$08,$FF,$77,$CF,$47,$91
	.dc.b	$EE,$79,$3F,$EE,$79,$E2,$F4,$21,$93,$DD,$3D,$6B,$ED,$C6,$4F,$8D
	.dc.b	$19,$BE,$71,$9B,$EF,$5C,$0F,$89,$5E,$99,$F2,$3D,$6F,$8C,$EE,$EC
	.dc.b	$7D,$D9,$D5,$79,$0F,$3D,$75,$9E,$35,$B9,$F8,$27,$98,$DB,$D1,$F0
	.dc.b	$9C,$A3,$8B,$83,$D5,$5E,$77,$ED,$1B,$3C,$47,$D0,$7C,$AF,$73,$06
	.dc.b	$CC,$57,$B5,$E2,$FC,$47,$CA,$62,$7D,$4F,$23,$C6,$7B,$2F,$E6,$DF
	.dc.b	$AE,$3C,$A6,$EB,$A3,$C0,$C4,$6E,$78,$C7,$73,$E0,$4F,$CB,$7C,$B3
	.dc.b	$E2,$8E,$64,$71,$5F,$22,$7C,$33,$A8,$E1,$3D,$EF,$F0,$D5,$DC,$DC
	.dc.b	$65,$F8,$27,$C6,$9E,$2E,$EF,$CE,$E4,$F3,$CF,$17,$DA,$6E,$BB,$BB
	.dc.b	$3B,$99,$3B,$CF,$E9,$DF,$8D,$7F,$55,$E1,$1E,$03,$51,$F0,$63,$E3
	.dc.b	$DF,$C1,$9F,$CF,$B6,$BF,$07,$1C,$2F,$E3,$D7,$CC,$38,$94,$F9,$35
	.dc.b	$C5,$FC,$44,$F1,$76,$7E,$2A,$79,$B4,$CC,$6D,$E6,$75,$7C,$AD,$AF
	.dc.b	$29,$F9,$D8,$F7,$34,$91,$F1,$9A,$FC,$69,$F0,$CF,$73,$AC,$78,$CF
	.dc.b	$09,$5C,$9B,$9F,$0C,$7E,$6D,$FA,$F1,$F3,$AB,$F0,$33,$F4,$A7,$27
	.dc.b	$8A,$FE,$94,$76,$AE,$8F,$C8,$9D,$5F,$F4,$A8,$FE,$A8,$FF,$5B,$7F
	.dc.b	$AA,$35,$57,$F8,$DD,$D9,$B7,$0F,$F3,$E8,$E1,$18,$8F,$8C,$7F,$2A
	.dc.b	$E1,$FF,$18,$3F,$94,$3F,$6A,$E2,$3F,$1D,$7F,$CD,$23,$15,$F0,$21
	.dc.b	$7F,$E3,$58,$C3,$6F,$F4,$27,$F9,$66,$B8,$B8,$15,$C9,$FF,$1D,$5F
	.dc.b	$CA,$7C,$DC,$7F,$E3,$97,$CD,$3E,$38,$C4,$F9,$AF,$FA,$75,$F6,$BF
	.dc.b	$D5,$9D,$A3,$C8,$E2,$FC,$0F,$F9,$0A,$B9,$3E,$08,$FF,$0E,$70,$7F
	.dc.b	$97,$EB,$C9,$3B,$C7,$DE,$FA,$0F,$9A,$7F,$23,$C3,$C5,$7F,$98,$8D
	.dc.b	$1E,$AA,$F3,$E3,$EA,$BE,$73,$D4,$F8,$D5,$E6,$BE,$77,$E2,$6C,$F8
	.dc.b	$1A,$F5,$3B,$E3,$EC,$7C,$98,$77,$7A,$D7,$C4,$BF,$34,$78,$4F,$8D
	.dc.b	$19,$3C,$13,$C6,$32,$76,$C3,$D3,$87,$8B,$0F,$03,$63,$FA,$3B,$67
	.dc.b	$FA,$D3,$5F,$C8,$D7,$5C,$3C,$73,$97,$F0,$07,$D4,$BF,$A6,$7F,$80
	.dc.b	$23,$58,$7C,$C9,$D5,$FD,$4A,$D5,$F5,$AF,$5D,$6A,$BE,$41,$F1,$9C
	.dc.b	$57,$C9,$38,$B8,$C3,$DF,$7C,$EB,$E9,$35,$E7,$87,$91,$FC,$60,$C7
	.dc.b	$F1,$8F,$0B,$E8,$FD,$43,$E6,$14,$FA,$8F,$C4,$AE,$D8,$7C,$E6,$E1
	.dc.b	$F3,$9B,$5F,$39,$AF,$CE,$AE,$EF,$DA,$3A,$E3,$F1,$1F,$35,$FC,$EB
	.dc.b	$FF,$61,$5F,$2B,$FA,$75,$A3,$E7,$4E,$A3,$A9,$E7,$2D,$F5,$C3,$E0
	.dc.b	$D7,$C6,$BE,$93,$F6,$63,$A1,$F5,$30,$FA,$2E,$1F,$A9,$87,$78,$C3
	.dc.b	$F5,$30,$EF,$1F,$34,$C9,$F2,$F1,$FA,$8F,$DF,$32,$7C,$37,$35,$E0
	.dc.b	$C7,$CE,$FC,$FA,$EB,$5E,$96,$C6,$63,$9C,$7C,$16,$E1,$F0,$67,$E6
	.dc.b	$47,$F2,$5C,$7F,$8F,$60,$7F,$92,$C3,$E1,$7F,$A0,$BF,$D1,$0F,$F5
	.dc.b	$83,$87,$FA,$2E,$5F,$EA,$F8,$C5,$73,$C7,$78,$7F,$22,$CB,$FD,$D5
	.dc.b	$3F,$D0,$8F,$E5,$0E,$31,$F1,$9F,$BF,$71,$5F,$A7,$9D,$D7,$FB,$DD
	.dc.b	$E5,$5B,$8C,$3C,$A7,$F1,$F5,$CB,$1F,$F8,$03,$3F,$EE,$C7,$FD,$95
	.dc.b	$7E,$77,$FD,$78,$D7,$FD,$E6,$63,$93,$FD,$DD,$1F,$EE,$D8,$C7,$27
	.dc.b	$FB,$C8,$FF,$76,$BF,$EC,$AE,$3A,$FF,$B4,$DF,$ED,$8F,$FB,$84,$FE
	.dc.b	$A7,$C9,$FE,$F0,$7F,$DD,$07,$FB,$8C,$B8,$FF,$B8,$5B,$9F,$FB,$94
	.dc.b	$FF,$79,$C6,$4F,$F7,$3B,$98,$FF,$76,$C6,$63,$66,$63,$C3,$B7,$FB
	.dc.b	$DA,$33,$1F,$EE,$1E,$CF,$F0,$27,$0F,$FB,$6A,$3F,$DD,$8F,$FB,$7F
	.dc.b	$87,$3C,$7F,$EE,$69,$FF,$78,$72,$3E,$0F,$FC,$14,$D8,$FF,$77,$C6
	.dc.b	$0F,$F0,$3F,$27,$EA,$63,$F2,$E7,$EA,$74,$FD,$0C,$7F,$0D,$E3,$FC
	.dc.b	$FB,$67,$67,$E8,$CB,$A7,$E8,$EB,$8C,$7F,$13,$AE,$2D,$37,$5C,$E7
	.dc.b	$71,$FC,$5C,$C9,$B7,$33,$FC,$49,$BA,$FF,$09,$6C,$FF,$30,$6B,$FB
	.dc.b	$F6,$DF,$EF,$4F,$D9,$C7,$F8,$1C,$FE,$62,$7F,$03,$E3,$FE,$2A,$F3
	.dc.b	$8D,$DF,$F7,$59,$CC,$F5,$9C,$DF,$F6,$98,$F9,$51,$87,$AD,$7F,$05
	.dc.b	$BF,$1E,$1F,$C6,$DB,$3F,$C1,$67,$E3,$BF,$0E,$32,$F9,$CF,$E7,$27
	.dc.b	$63,$F9,$99,$C4,$F9,$27,$1A,$F1,$3C,$0F,$B1,$1D,$27,$D8,$78,$0F
	.dc.b	$9D,$3C,$38,$BF,$92,$DF,$9E,$31,$8F,$C2,$7E,$66,$1E,$23,$CD,$5C
	.dc.b	$1F,$26,$1D,$F1,$F8,$D1,$E2,$AE,$78,$F8,$67,$EE,$46,$27,$D7,$3E
	.dc.b	$47,$B3,$E8,$73,$3C,$1C,$BD,$A7,$F2,$87,$3A,$F1,$4F,$D4,$9F,$E6
	.dc.b	$CF,$03,$07,$CD,$3F,$AA,$63,$F0,$5D,$1F,$2E,$BA,$B8,$C3,$E0,$3F
	.dc.b	$0B,$1F,$8D,$5C,$63,$99,$F1,$2B,$C8,$47,$94,$F0,$74,$63,$1F,$8C
	.dc.b	$FE,$54,$C1,$F0,$8C,$1D,$8F,$13,$C1,$FC,$C1,$C3,$1F,$86,$FA,$EF
	.dc.b	$F6,$97,$E3,$8E,$55,$F3,$EB,$78,$7B,$A3,$15,$B3,$FB,$BD,$7A,$0E
	.dc.b	$38,$FD,$3C,$3A,$C7,$86,$78,$47,$F1,$3C,$3A,$E3,$F0,$6B,$D2,$7B
	.dc.b	$4F,$49,$D0,$FA,$E7,$E2,$CE,$84,$34,$7B,$4F,$E1,$95,$E0,$7F,$C0
	.dc.b	$D1,$DF,$0F,$5C,$71,$9F,$5B,$E2,$AF,$2C,$FC,$B9,$FD,$0C,$F6,$C7
	.dc.b	$E2,$63,$EF,$8E,$E7,$3C,$7D,$D5,$E1,$8F,$E2,$17,$CE,$7F,$0D,$3E
	.dc.b	$08,$F1,$CF,$5C,$7C,$D7,$D6,$3F,$7C,$75,$3D,$73,$CD,$FA,$55,$CE
	.dc.b	$78,$4F,$EA,$E7,$A3,$B7,$DD,$3E,$58,$FE,$30,$D9,$FD,$3C,$F0,$3E
	.dc.b	$A4,$7C,$23,$B3,$F5,$1F,$D2,$BB,$C7,$E4,$BC,$30,$F7,$9F,$11,$F5
	.dc.b	$1F,$10,$D1,$EA,$74,$F3,$9F,$23,$FB,$77,$C2,$FD,$F3,$E2,$9B,$3A
	.dc.b	$DF,$B1,$B7,$EC,$9E,$38,$FD,$A9,$F5,$5F,$E6,$1E,$44,$9D,$E3,$F5
	.dc.b	$23,$FD,$83,$5F,$E2,$87,$FC,$74,$7F,$88,$B0,$FF,$16,$4F,$F8,$A6
	.dc.b	$33,$8F,$F8,$EA,$3F,$C5,$51,$FE,$56,$8F,$F1,$51,$99,$FF,$16,$DF
	.dc.b	$F8,$C1,$E6,$62,$79,$C1,$F1,$D7,$77,$FC,$92,$70,$3C,$87,$78,$E8
	.dc.b	$E0,$FF,$19,$90,$E8,$2F,$FC,$70,$E1,$FF,$1B,$3A,$35,$C7,$FE,$3A
	.dc.b	$27,$1F,$F8,$DA,$3C,$DC,$75,$3F,$E3,$83,$FC,$B6,$FA,$F1,$71,$B5
	.dc.b	$CC,$FF,$8F,$CF,$F3,$56,$5E,$6A,$DC,$EA,$7B,$9F,$E3,$C3,$FC,$90
	.dc.b	$53,$97,$27,$F9,$BE,$78,$C7,$43,$71,$9B,$F1,$46,$4E,$37,$E1,$8C
	.dc.b	$CF,$F9,$2C,$E1,$7E,$8E,$3F,$F1,$F5,$FF,$90,$4F,$90,$1F,$F2,$03
	.dc.b	$87,$6F,$FB,$E7,$6F,$C2,$E1,$FE,$7F,$3E,$CB,$F8,$78,$FB,$87,$E1
	.dc.b	$0F,$C1,$E3,$F8,$77,$F0,$67,$E0,$A3,$81,$FE,$8F,$96,$FF,$80,$8E
	.dc.b	$F3,$FE,$28,$6B,$E3,$9F,$09,$E3,$7C,$07,$E1,$F7,$FC,$EC,$F5,$AE
	.dc.b	$95,$C2,$BA,$35,$E8,$72,$DF,$FD,$23,$87,$A6,$7C,$B3,$F9,$BB,$F4
	.dc.b	$C7,$E5,$8C,$C7,$C2,$3C,$53,$B8,$D5,$F8,$A7,$B6,$1E,$69,$CC,$7C
	.dc.b	$7C,$3C,$2F,$38,$EB,$87,$82,$3D,$27,$2C,$3B,$3F,$0A,$76,$7E,$9E
	.dc.b	$FB,$46,$70,$F1,$3E,$63,$EA,$4F,$89,$F5,$61,$E2,$C3,$E5,$47,$96
	.dc.b	$F9,$BF,$72,$3E,$AC,$FF,$13,$8C,$57,$87,$0F,$2C,$60,$F4,$1E,$03
	.dc.b	$C8,$6A,$79,$4F,$AA,$3F,$16,$7C,$3C,$3D,$33,$E5,$9F,$45,$F8,$B0
	.dc.b	$FE,$25,$1C,$6F,$DD,$3F,$25,$F1,$12,$B9,$3E,$23,$C8,$7E,$10,$D5
	.dc.b	$7C,$0A,$E8,$75,$9E,$33,$D2,$31,$3B,$BF,$9F,$18,$C3,$E1,$BC,$A3
	.dc.b	$D7,$3B,$AF,$9D,$5A,$9E,$26,$9C,$46,$8F,$3D,$73,$7E,$55,$7C,$2A
	.dc.b	$E1,$1A,$79,$BA,$AF,$CF,$46,$B0,$F8,$D7,$CE,$BE,$1E,$1F,$52,$B5
	.dc.b	$3F,$B1,$9F,$8B,$7E,$47,$F1,$38,$7A,$DB,$19,$C3,$EB,$3E,$08,$F1
	.dc.b	$1F,$74,$B1,$F9,$07,$89,$E9,$3E,$03,$E6,$BF,$5C,$79,$AB,$E0,$3C
	.dc.b	$AF,$F0,$A7,$EF,$F0,$F5,$46,$4F,$C0,$5F,$E1,$63,$31,$CF,$0F,$33
	.dc.b	$4F,$31,$67,$CC,$7B,$1C,$BC,$EF,$C7,$1B,$BF,$1B,$FA,$29,$F5,$3E
	.dc.b	$C8,$F9,$87,$E1,$DF,$80,$7C,$28,$DB,$63,$78,$7D,$48,$FE,$0A,$7D
	.dc.b	$43,$8C,$F9,$4F,$3C,$FC,$3A,$F2,$C7,$9E,$FD,$ED,$AF,$5D,$7E,$ED
	.dc.b	$E6,$F1,$9D,$B9,$9F,$29,$DD,$F7,$CF,$8E,$33,$1C,$63,$A9,$FB,$E8
	.dc.b	$F2,$BF,$16,$3D,$AF,$D7,$3F,$5D,$7E,$F7,$F6,$D1,$F1,$E7,$AC,$71
	.dc.b	$3E,$1C,$64,$E6,$FD,$33,$F0,$2F,$A4,$CC,$7C,$13,$C7,$1E,$98,$C9
	.dc.b	$B7,$AC,$F8,$F0,$FC,$4D,$71,$AF,$82,$78,$6B,$67,$AD,$FD,$74,$6E
	.dc.b	$3C,$67,$C3,$8E,$35,$D7,$0F,$AF,$1E,$19,$F8,$33,$9C,$3E,$9C,$71
	.dc.b	$39,$9F,$C8,$A7,$D2,$E4,$F9,$67,$8E,$7F,$57,$7E,$38,$C9,$F1,$CA
	.dc.b	$E6,$7D,$27,$8F,$0F,$82,$DC,$3C,$F5,$DE,$36,$DB,$EF,$1B,$C3,$D3
	.dc.b	$3E,$73,$E3,$9E,$18,$E2,$E4,$F8,$27,$A4,$FD,$51,$C3,$0F,$35,$76
	.dc.b	$8F,$01,$A3,$89,$D5,$D3,$D7,$0F,$5B,$C0,$FD,$81,$D6,$BD,$0F,$E6
	.dc.b	$8E,$18,$7C,$06,$E1,$EC,$AF,$E8,$F5,$EA,$6D,$F8,$1B,$5E,$57,$C2
	.dc.b	$F2,$3B,$3A,$7C,$67,$AB,$0F,$7B,$C7,$0F,$14,$EE,$7F,$69,$3F,$84
	.dc.b	$8C,$5F,$EC,$E7,$F3,$EF,$E5,$E7,$A3,$EE,$8C,$DF,$E3,$E7,$C7,$7F
	.dc.b	$CF,$E7,$AC,$7E,$1A,$B8,$1F,$9A,$7E,$B1,$E8,$BF,$69,$E2,$7F,$70
	.dc.b	$78,$0F,$A1,$3D,$27,$E4,$57,$77,$D0,$75,$9F,$B4,$7C,$1C,$3E,$84
	.dc.b	$EE,$3A,$DF,$82,$31,$1F,$BD,$3E,$E1,$E0,$AE,$AF,$C7,$8F,$F5,$4B
	.dc.b	$FE,$97,$3F,$D4,$53,$FE,$96,$6C,$2B,$FC,$DF,$1F,$E9,$6D,$AD,$7F
	.dc.b	$A6,$A3,$FD,$39,$BF,$FA,$61,$D9,$FE,$96,$2E,$FF,$E9,$93,$FD,$30
	.dc.b	$FF,$A2,$CF,$F4,$7E,$5F,$E8,$04,$B2,$FF,$3F,$CF,$FA,$DE,$34,$66
	.dc.b	$7F,$D2,$AE,$27,$71,$FE,$8F,$C7,$FD,$2B,$B7,$FA,$36,$7F,$D2,$1B
	.dc.b	$7F,$A3,$78,$9C,$9E,$EE,$0F,$F4,$49,$FE,$8C,$C7,$FD,$6C,$FF,$A3
	.dc.b	$35,$6F,$F1,$67,$FD,$39,$8F,$F9,$AB,$2F,$F0,$01,$FE,$8D,$3F,$9D
	.dc.b	$C7,$D5,$3F,$08,$7E,$8F,$4F,$C9,$BF,$8C,$7F,$0D,$C1,$FA,$BB,$FB
	.dc.b	$B7,$F9,$BC,$3F,$01,$AF,$F8,$4E,$3E,$EE,$1F,$97,$3F,$00,$7E,$1A
	.dc.b	$BF,$52,$72,$BF,$D0,$61,$E0,$9F,$C7,$9F,$9B,$9D,$DF,$58,$E4,$FE
	.dc.b	$48,$C3,$EB,$C3,$C6,$FF,$6C,$36,$7E,$B2,$37,$87,$A6,$BA,$CE,$CE
	.dc.b	$4E,$2F,$E3,$CE,$EB,$C1,$5B,$8C,$3D,$4C,$61,$EC,$8F,$D5,$57,$F0
	.dc.b	$A9,$F6,$57,$28,$EB,$5B,$9F,$D2,$4F,$9A,$31,$1F,$2C,$FA,$D5,$D0
	.dc.b	$FE,$06,$F0,$3C,$27,$D9,$7E,$99,$FB,$E9,$E9,$19,$3E,$2D,$7D,$13
	.dc.b	$CC,$7E,$01,$FE,$00,$7B,$5F,$3B,$C4,$F8,$F1,$EB,$9F,$EC,$71,$E6
	.dc.b	$3C,$93,$C1,$FB,$AF,$57,$E1,$9D,$63,$8E,$1F,$C6,$63,$A9,$93,$F8
	.dc.b	$59,$D4,$F8,$8E,$E3,$E9,$CF,$AD,$F8,$B3,$AC,$3F,$20,$FA,$63,$67
	.dc.b	$CB,$7F,$2E,$78,$27,$93,$88,$DB,$EC,$7C,$F3,$D2,$3D,$B1,$F4,$30
	.dc.b	$FD,$46,$1F,$42,$72,$7D,$09,$F0,$3E,$B7,$EB,$3F,$68,$D9,$E3,$77
	.dc.b	$87,$BE,$BF,$0B,$1C,$30,$F9,$C7,$DA,$7C,$55,$F6,$A3,$D6,$F7,$9F
	.dc.b	$2B,$DE,$38,$3E,$53,$DC,$7D,$0A,$F0,$3F,$26,$7C,$27,$29,$F0,$D7
	.dc.b	$D0,$3C,$87,$C6,$BF,$85,$1D,$AB,$D8,$76,$3F,$89,$DF,$67,$E0,$D7
	.dc.b	$D0,$7E,$D5,$72,$8E,$D7,$E1,$7B,$4F,$AD,$CB,$F0,$5F,$93,$87,$D7
	.dc.b	$C3,$C5,$1D,$4F,$3C,$FB,$E7,$51,$CA,$3F,$2F,$3F,$2B,$0F,$33,$CA
	.dc.b	$FE,$1C,$F7,$3D,$93,$FC,$52,$7E,$2C,$7E,$2E,$76,$F4,$AF,$2E,$1E
	.dc.b	$D8,$D1,$F3,$C8,$7C,$49,$E3,$5F,$AB,$25,$7E,$02,$7A,$1E,$2C,$3F
	.dc.b	$8E,$47,$C6,$9F,$1C,$FE,$0C,$F4,$1F,$38,$F9,$47,$88,$F0,$1E,$27
	.dc.b	$E0,$36,$3E,$34,$78,$A3,$F1,$47,$37,$BE,$1F,$22,$76,$79,$E3,$84
	.dc.b	$FE,$10,$E1,$1E,$72,$3F,$57,$0F,$8C,$FA,$4F,$79,$C0,$FF,$04,$BF
	.dc.b	$E0,$6B,$FE,$CD,$1F,$E0,$F1,$7F,$E0,$F7,$FC,$18,$08,$7F,$85,$37
	.dc.b	$FF,$06,$38,$36,$7F,$85,$37,$FF,$06,$3F,$E0,$E2,$63,$FE,$03,$7F
	.dc.b	$C0,$B8,$FF,$80,$1F,$F0,$6B,$FE,$06,$8D,$47,$82,$7F,$C1,$D3,$B9
	.dc.b	$FF,$07,$4B,$4F,$F0,$2E,$DF,$E0,$88,$F2,$19,$E5,$72,$FF,$82,$E7
	.dc.b	$53,$E9,$CF,$FC,$13,$AF,$BF,$7F,$F0,$3E,$7F,$00,$FF,$05,$1F,$ED
	.dc.b	$77,$FC,$14,$E7,$7F,$8B,$1C,$2F,$E7,$87,$EF,$EF,$FA,$A8,$FC,$56
	.dc.b	$BF,$88,$7F,$16,$47,$04,$BF,$C2,$C7,$E2,$9F,$DA,$BF,$C1,$A7,$A3
	.dc.b	$F8,$19,$FC,$4C,$FE,$EE,$FF,$82,$4F,$E1,$CF,$CD,$9B,$8F,$C8,$11
	.dc.b	$ED,$1C,$1E,$D8,$7E,$3B,$1F,$CC,$47,$AA,$BC,$AF,$E5,$DF,$2C,$7E
	.dc.b	$FD,$FC,$BC,$72,$3C,$31,$C9,$F1,$1E,$03,$F6,$31,$CB,$0F,$DB,$C7
	.dc.b	$A7,$0F,$97,$7E,$39,$F8,$37,$CC,$8E,$09,$1B,$25,$78,$A3,$2D,$8C
	.dc.b	$D7,$68,$C1,$1C,$12,$FD,$B7,$EF,$9F,$34,$62,$FD,$31,$F0,$6B,$83
	.dc.b	$A2,$19,$9E,$17,$FE,$0B,$8C,$DF,$57,$46,$2F,$AC,$64,$8E,$09,$3C
	.dc.b	$E7,$89,$27,$F4,$2F,$F9,$A3,$0F,$CD,$DF,$E7,$63,$F4,$8F,$E8,$48
	.dc.b	$FF,$05,$7F,$4A,$D8,$E2,$FE,$8B,$1F,$7B,$DF,$0F,$1B,$FE,$1B,$31
	.dc.b	$87,$97,$0F,$4B,$67,$78,$FF,$4F,$7E,$BB,$D2,$77,$5C,$DB,$1B,$AF
	.dc.b	$B7,$8F,$CC,$AD,$5F,$48,$FE,$9F,$1F,$75,$D5,$71,$8C,$1F,$3E,$BC
	.dc.b	$84,$79,$B6,$38,$1E,$5A,$ED,$3F,$D4,$67,$55,$E0,$AD,$46,$27,$47
	.dc.b	$42,$47,$49,$F2,$9C,$4B,$1C,$23,$4F,$9C,$E3,$1B,$71,$1A,$7E,$89
	.dc.b	$A3,$94,$6D,$CC,$6E,$FE,$6C,$E5,$B3,$C0,$87,$F5,$13,$8C,$6E,$FE
	.dc.b	$1C,$66,$78,$B6,$33,$3F,$C0,$6F,$E2,$4E,$5B,$19,$8F,$41,$92,$B9
	.dc.b	$BF,$45,$7F,$B1,$70,$FE,$6D,$5F,$ED,$BC,$7F,$2E,$FF,$62,$9D,$5F
	.dc.b	$CC,$8E,$87,$F6,$7C,$7F,$B3,$13,$7F,$EC,$71,$FE,$F0,$E2,$FF,$79
	.dc.b	$C1,$FE,$C7,$1F,$D9,$4F,$EC,$85,$C7,$FD,$DC,$7F,$B8,$DF,$EE,$19
	.dc.b	$FF,$B9,$B0,$FE,$E3,$BF,$F5,$EE,$1F,$F7,$27,$07,$F5,$CB,$FE,$BF
	.dc.b	$3F,$EE,$3D,$3E,$4E,$BF,$CE,$75,$F9,$B8,$FC,$33,$E9,$65,$F2,$A7
	.dc.b	$55,$C3,$0F,$E6,$31,$89,$D4,$60,$87,$F8,$3A,$BA,$3F,$80,$6B,$F4
	.dc.b	$F6,$FA,$33,$CA,$38,$57,$57,$EB,$B6,$77,$18,$3B,$BF,$D7,$23,$FC
	.dc.b	$7F,$5F,$D5,$E7,$FA,$84,$71,$8F,$6C,$71,$3B,$9C,$9E,$07,$F6,$57
	.dc.b	$07,$F5,$13,$64,$9F,$03,$FE,$3D,$3F,$B0,$10,$FE,$B2,$78,$8E,$65
	.dc.b	$39,$47,$82,$7D,$31,$DE,$7A,$BF,$E0,$19,$FE,$D6,$79,$5E,$26,$DE
	.dc.b	$4F,$8C,$F0,$BB,$8E,$67,$F8,$26,$3C,$D8,$7F,$64,$BF,$E4,$71,$FC
	.dc.b	$0A,$BE,$3D,$7D,$02,$38,$24,$7D,$DB,$F9,$CD,$3D,$F1,$E1,$3C,$AF
	.dc.b	$AC,$F0,$E1,$FC,$E2,$3F,$58,$7E,$18,$93,$D2,$7F,$11,$1E,$06,$9D
	.dc.b	$CF,$86,$7C,$DA,$F7,$C7,$C8,$9F,$9B,$39,$8F,$1C,$FC,$38,$F8,$F3
	.dc.b	$A9,$F4,$47,$CD,$21,$C4,$AE,$9C,$11,$C1,$23,$A1,$23,$EB,$61,$F4
	.dc.b	$5F,$E9,$67,$F2,$43,$D9,$5E,$A7,$E3,$E1,$F5,$9D,$1F,$08,$97,$F2
	.dc.b	$1F,$A3,$7D,$DF,$A0,$7B,$8F,$0E,$1F,$45,$FC,$01,$E0,$9F,$8A,$7C
	.dc.b	$97,$75,$E1,$C3,$F8,$4D,$FD,$5A,$F0,$57,$D2,$7E,$03,$C0,$F6,$CF
	.dc.b	$07,$CA,$67,$0F,$EB,$EF,$ED,$5F,$80,$F9,$CF,$4C,$7C,$78,$D9,$F9
	.dc.b	$48,$F4,$1F,$52,$BF,$D2,$B1,$FE,$84,$9F,$F0,$DB,$70,$FF,$4F,$D7
	.dc.b	$FA,$6A,$7F,$9C,$C6,$23,$FC,$EE,$7F,$A0,$F6,$FE,$76,$FF,$A7,$F6
	.dc.b	$FE,$76,$7F,$3A,$7F,$D1,$BC,$3F,$E8,$E7,$FD,$2C,$7F,$32,$E0,$FF
	.dc.b	$46,$BF,$CE,$1A,$FF,$AA,$F5,$6D,$FC,$E6,$3F,$D5,$51,$FC,$E7,$6F
	.dc.b	$F1,$7C,$7F,$8A,$CA,$7F,$57,$2E,$7F,$E3,$8E,$0F,$8B,$BF,$F2,$07
	.dc.b	$FA,$EB,$FE,$34,$9F,$EB,$95,$C3,$5F,$E7,$D1,$F6,$77,$FC,$46,$FF
	.dc.b	$A5,$C7,$E8,$61,$FB,$69,$FC,$01,$2B,$ED,$9F,$5A,$FF,$18,$FE,$5A
	.dc.b	$7F,$0B,$7F,$72,$F9,$CF,$E4,$0F,$CC,$DF,$37,$C4,$7E,$16,$BC,$0F
	.dc.b	$6A,$F2,$5F,$6B,$FC,$64,$F0,$3C,$58,$7E,$5E,$7D,$B5,$E1,$78,$18
	.dc.b	$BF,$CF,$D7,$EB,$8F,$CB,$C6,$5F,$C6,$1E,$33,$D1,$1E,$C3,$8B,$F9
	.dc.b	$79,$F2,$D7,$84,$FE,$BC,$F8,$5C,$47,$E7,$E7,$8D,$FC,$48,$F9,$18
	.dc.b	$7C,$3B,$E1,$1E,$4B,$E7,$7C,$30,$F5,$C6,$27,$CE,$7C,$23,$81,$DC
	.dc.b	$CD,$79,$9C,$BF,$AB,$7E,$99,$F3,$EB,$F8,$13,$CE,$3F,$1A,$FE,$C0
	.dc.b	$FE,$41,$7F,$88,$9D,$CF,$53,$73,$D0,$FA,$CE,$63,$EE,$C6,$1E,$24
	.dc.b	$7E,$A4,$72,$8F,$A1,$1E,$A7,$4F,$D1,$6E,$1F,$2C,$FC,$2C,$FC,$A3
	.dc.b	$95,$7C,$D9,$F9,$57,$F5,$27,$E1,$47,$6B,$F6,$9F,$81,$3E,$A9,$F1
	.dc.b	$4F,$2B,$EA,$8E,$CD,$3A,$C7,$CE,$7F,$AD,$1D,$4F,$E4,$EF,$CD,$78
	.dc.b	$3E,$D3,$A9,$F9,$0A,$F4,$39,$3E,$01,$A3,$CA,$7C,$B3,$17,$F0,$A7
	.dc.b	$C7,$5F,$0A,$7D,$47,$22,$4F,$C9,$C3,$F9,$24,$7F,$08,$07,$F9,$85
	.dc.b	$CC,$37,$FE,$4F,$28,$C6,$FF,$CA,$23,$89,$FE,$65,$1E,$6B,$C6,$C3
	.dc.b	$63,$F9,$81,$A8,$C9,$FC,$E7,$0F,$09,$34,$FE,$65,$87,$F3,$C6,$C6
	.dc.b	$6B,$C2,$7F,$3E,$39,$46,$88,$E8,$F8,$07,$BD,$CB,$4D,$4F,$76,$CF
	.dc.b	$77,$44,$3B,$9A,$9E,$6E,$0D,$1E,$52,$1E,$23,$44,$33,$3E,$17,$89
	.dc.b	$B3,$33,$E3,$78,$19,$24,$71,$72,$D7,$8C,$F9,$9A,$7A,$48,$78,$23
	.dc.b	$83,$D2,$7C,$0F,$03,$4E,$1D,$4F,$56,$9D,$0F,$1C,$6E,$7A,$91,$C1
	.dc.b	$27,$94,$65,$B1,$99,$E8,$78,$CE,$B1,$A2,$3A,$3C,$67,$A1,$CB,$4D
	.dc.b	$4F,$76,$CF,$77,$44,$3B,$9A,$9E,$6E,$0D,$1E,$52,$1E,$23,$44,$33
	.dc.b	$3E,$17,$89,$B3,$33,$E3,$78,$19,$24,$71,$72,$D7,$8C,$F9,$9A,$7A
	.dc.b	$48,$78,$23,$83,$D2,$7C,$0F,$03,$4E,$1D,$4F,$56,$9D,$0F,$1C,$6E
	.dc.b	$7A,$91,$C1,$27,$94,$65,$B1,$99,$E8,$78,$CE,$B1,$A2,$3A,$3C,$67
	.dc.b	$A1,$CB,$4D,$4F,$76,$CF,$77,$44,$3B,$9A,$9E,$6E,$0D,$1E,$52,$1E
	.dc.b	$23,$44,$33,$3E,$17,$89,$B3,$33,$E3,$78,$19,$24,$71,$72,$D7,$8C
	.dc.b	$F9,$9A,$7A,$48,$78,$23,$83,$D2,$7C,$0F,$03,$4E,$1D,$4F,$56,$9D
	.dc.b	$0F,$1C,$6E,$7A,$91,$C1,$27,$94,$65,$B1,$99,$E8,$78,$CE,$B1,$A2
	.dc.b	$3A,$3C,$67,$A1,$CB,$4D,$4F,$76,$CF,$77,$44,$3B,$9A,$9E,$6E,$0D
	.dc.b	$1E,$52,$1E,$23,$44,$33,$3E,$17,$89,$B3,$33,$E3,$78,$19,$24,$71
	.dc.b	$72,$D7,$8C,$F9,$9A,$7A,$48,$78,$23,$83,$D2,$7C,$0F,$03,$F8,$36
	.dc.b	$3F,$E0,$69,$FF,$02,$C7,$F6,$A8,$C1,$31,$FF,$04,$47,$FA,$CD,$AF
	.dc.b	$F8,$24,$C1,$C5,$CB,$FE,$08,$32,$43,$81,$0D,$14,$FF,$0C,$47,$85
	.dc.b	$E8,$43,$91,$C0,$87,$08,$F0,$B4,$E0,$75,$9F,$7B,$C8,$D1,$5C,$BB
	.dc.b	$9F,$13,$96,$B9,$9E,$F3,$C2,$3C,$6D,$7A,$9B,$3C,$66,$A7,$9B,$67
	.dc.b	$9C,$62,$38,$99,$38,$98,$24,$F7,$24,$71,$6C,$F4,$24,$F2,$8C,$B6
	.dc.b	$78,$12,$77,$1F,$05,$AF,$AC,$C1,$C5,$CB,$E7,$32,$43,$81,$0D,$14
	.dc.b	$F5,$C7,$85,$E8,$43,$91,$C0,$87,$08,$F0,$B4,$E0,$75,$9F,$7B,$C8
	.dc.b	$D1,$5C,$BB,$9F,$13,$96,$B9,$9E,$F3,$C2,$3C,$6D,$7A,$9B,$3C,$66
	.dc.b	$A7,$9B,$67,$9C,$62,$38,$99,$38,$98,$24,$F7,$24,$71,$6C,$F4,$24
	.dc.b	$F2,$8C,$B6,$78,$12,$77,$1F,$05,$AF,$AC,$C1,$C5,$CB,$E7,$32,$43
	.dc.b	$81,$0D,$14,$F5,$C7,$85,$E8,$43,$91,$C0,$87,$08,$F0,$B4,$E0,$75
	.dc.b	$9F,$7B,$C8,$D1,$5C,$BB,$9F,$13,$96,$B9,$9E,$F3,$C2,$3C,$6D,$7A
	.dc.b	$9B,$3C,$66,$A7,$9B,$67,$9C,$62,$38,$99,$38,$98,$24,$F7,$24,$71
	.dc.b	$6C,$F4,$24,$F2,$8C,$B6,$78,$12,$77,$1F,$05,$AF,$AC,$C1,$C5,$CB
	.dc.b	$E7,$32,$43,$81,$0D,$14,$F5,$C7,$85,$E8,$43,$91,$C0,$87,$08,$F0
	.dc.b	$B4,$E0,$75,$9F,$7B,$C8,$D1,$5C,$BB,$9F,$13,$96,$B9,$9E,$F3,$C2
	.dc.b	$3C,$6D,$7A,$9B,$3C,$66,$A7,$9B,$67,$9C,$62,$38,$99,$38,$98,$24
	.dc.b	$F7,$3F,$33,$F9,$48,$80

;--------------------------------------------------------------------------------
;	FDBcc Dr,<label>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fdbcc060_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#FPSP060,-(sp)
	peamsg	'FDBCC DR,<LABEL>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FDBcc Dr,<label>',13,10
;------------------------------------------------
;	d1	actual result
;	d2	actual count
;	d3	actual status
;	d4	source count
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result,expected count,expected status,...
;	a1	expected result
;	a2	expected count
;	a3	expected status
;	a4	source count,...
;------------------------------------------------
	lea.l	push_decompressed,a3
;decompress data
	move.l	a3,-(sp)
	pea.l	fdbcc060_data_compressed
	jbsr	decompress
	addq.l	#8,sp
@@:
	addq.l	#4,a3			;source count
	tst.l	(a3)
	bne	@b
	addq.l	#4,a3			;0
	movea.l	a3,a0			;expected result,expected count,expected status,...
;
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
	lea.l	push_decompressed,a4	;source count,...
44:
;FDBcc Dr=count,<forward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	move.l	(a4),d4			;source count
	move.l	d4,d2			;actual count
	jsr	([fdbcc060_execute_forward,za0,d5.l*4])	;EXECUTE
					;actual result
					;actual count
	fmove.l	fpsr,d3			;actual status
;
	movem.l	(a0),a1-a3		;expected result,expected count,expected status
;
	movem.l	d1-d3/a1-a3,-(sp)	;actual result,actual count,actual status,expected result,expected count,expected status
	jbsr	test_double
	lea.l	24(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	' Dr='
	puthex8	d4			;source count
	putmsg	',<forward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d3/a1-a3,-(sp)	;actual result,actual count,actual status,expected result,expected count,expected status
	jbsr	output_double
	lea.l	28(sp),sp
@@:
;
;FDBcc Dr=count,<backward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	move.l	(a4),d4			;source count
	move.l	d4,d2			;actual count
	jsr	([fdbcc060_execute_backward,za0,d5.l*4])	;EXECUTE
					;actual result
					;actual count
	fmove.l	fpsr,d3			;actual status
;
	movem.l	(a0),a1-a3		;expected result,expected count,expected status
;
	movem.l	d1-d3/a1-a3,-(sp)	;actual result,actual count,actual status,expected result,expected count,expected status
	jbsr	test_double
	lea.l	24(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	' Dr='
	puthex8	d4			;source count
	putmsg	',<backward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d3/a1-a3,-(sp)	;actual result,actual count,actual status,expected result,expected count,expected status
	jbsr	output_double
	lea.l	28(sp),sp
@@:
;
	lea.l	(12,a0),a0		;expected result,expected count,expected status,...
;
	addq.l	#4,a4			;source count
	tst.l	(a4)			;source count,...
	bne	44b
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
fdbcc060_execute_forward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fdbcc060_execute_forward_&cc
  .endm
fdbcc060_execute_backward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fdbcc060_execute_backward_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
fdbcc060_execute_forward_&cc::
	moveq.l	#2,d1			;actual result. 2=too long
	FDB&cc	d2,@f			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts

	subq.l	#2,d1			;actual result. -1=too short(forward)/too long(backward)
@@:
	subq.l	#1,d1			;actual result. 1=taken
	rts

fdbcc060_execute_backward_&cc::
	moveq.l	#2,d1			;actual result. 2=too short
	FDB&cc	d2,@b			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts
  .endm
	.cpu	68000

	.align	4
fdbcc060_data_compressed::
	.dc.b	$00,$03,$60,$98,$10,$86,$1C,$7E,$02,$FE,$01,$D7,$E0,$7A,$13,$C6
	.dc.b	$35,$58,$3A,$47,$37,$75,$A8,$56,$8E,$0F,$7A,$4F,$4A,$B1,$E5,$8F
	.dc.b	$21,$1E,$71,$AA,$EF,$1C,$DF,$44,$7B,$5F,$83,$3E,$D7,$E1,$92,$3C
	.dc.b	$B5,$D2,$B4,$FC,$23,$91,$DC,$F0,$CF,$C4,$23,$EB,$8F,$51,$F0,$4F
	.dc.b	$41,$E1,$9F,$34,$6E,$3E,$9C,$43,$EB,$BF,$0D,$CC,$65,$D3,$F6,$3F
	.dc.b	$80,$71,$1F,$4C,$F9,$26,$E3,$ED,$BA,$7E,$D4,$70,$8F,$C1,$4F,$77
	.dc.b	$F0,$B3,$F6,$DE,$07,$DC,$9B,$1D,$A3,$C6,$7C,$33,$32,$9E,$D7,$F1
	.dc.b	$67,$C8,$F7,$7D,$AF,$DE,$8F,$33,$F7,$E3,$73,$F8,$06,$BD,$5F,$83
	.dc.b	$7F,$22,$3D,$CF,$B6,$3F,$04,$7E,$3A,$B2,$FA,$4F,$9B,$1F,$83,$7E
	.dc.b	$43,$F4,$A3,$E3,$3F,$A2,$9F,$44,$FE,$10,$F9,$57,$F8,$AB,$F2,$47
	.dc.b	$D6,$AF,$CA,$BD,$27,$D4,$E5,$FD,$85,$FD,$3B,$F8,$D3,$F7,$9E,$EF
	.dc.b	$EB,$9F,$8C,$FD,$D7,$E3,$3B,$7A,$BF,$88,$7F,$5D,$1F,$0E,$3F,$60
	.dc.b	$7E,$14,$C4,$FE,$2D,$C1,$F1,$4F,$C1,$9F,$14,$FB,$E7,$A0,$F8,$33
	.dc.b	$E6,$8F,$C4,$47,$57,$E8,$36,$F4,$FE,$1C,$D4,$FC,$87,$F4,$51,$F0
	.dc.b	$9F,$C2,$C7,$DB,$9F,$E0,$B3,$B9,$FC,$BB,$F0,$9D,$DF,$D4,$8F,$83
	.dc.b	$1F,$86,$3E,$09,$F5,$4E,$91,$FC,$1A,$FF,$0B,$3E,$B8,$F8,$13,$F8
	.dc.b	$63,$E0,$1F,$B3,$7F,$09,$7F,$90,$BE,$67,$C0,$9F,$BD,$5C,$E3,$EA
	.dc.b	$C7,$EB,$AB,$F1,$97,$F8,$1B,$F2,$CF,$D9,$7E,$01,$F8,$07,$EF,$5F
	.dc.b	$CF,$8F,$E1,$91,$F8,$37,$CA,$FD,$F3,$E4,$1F,$A3,$8F,$C5,$BF,$C1
	.dc.b	$0F,$D3,$57,$F0,$97,$F8,$5C,$E9,$C1,$FC,$00,$F9,$8F,$C8,$7D,$E7
	.dc.b	$E0,$9F,$DF,$9F,$8F,$9F,$E8,$51,$FC,$59,$FA,$CF,$F2,$07,$D2,$FF
	.dc.b	$1D,$7F,$98,$BB,$8F,$86,$FF,$26,$9F,$61,$FD,$41,$FB,$93,$EC,$8F
	.dc.b	$AF,$1B,$3F,$9E,$9F,$81,$9D,$5F,$E8,$E7,$FA,$9C,$FF,$0A,$8F,$D0
	.dc.b	$3F,$D8,$5F,$AE,$D8,$F8,$6E,$67,$E5,$9F,$1D,$FD,$6C,$FD,$BA,$CC
	.dc.b	$F2,$9F,$ED,$4F,$91,$FE,$78,$FE,$7D,$FB,$07,$F3,$03,$E5,$9F,$58
	.dc.b	$F4,$4F,$EE,$E3,$EB,$3F,$7C,$FD,$8C,$7F,$46,$9C,$D7,$EF,$8F,$4C
	.dc.b	$7F,$3E,$AF,$C1,$CF,$A4,$FA,$F1,$F4,$E3,$E1,$BF,$E0,$67,$D4,$FF
	.dc.b	$6C,$7D,$4F,$EE,$1F,$AE,$FF,$72,$3D,$07,$80,$FE,$E1,$1F,$5D,$E6
	.dc.b	$7F,$6C,$AF,$E5,$EF,$23,$FB,$91,$F9,$67,$DA,$FD,$58,$FB,$75,$93
	.dc.b	$B4,$7F,$7C,$AF,$A2,$FF,$43,$7D,$AF,$F6,$F7,$F2,$6F,$F8,$91,$FC
	.dc.b	$9B,$97,$DA,$FF,$76,$7E,$B3,$93,$F9,$39,$FD,$08,$FE,$14,$7F,$2A
	.dc.b	$9F,$E3,$EF,$C8,$AF,$EE,$2F,$E9,$CF,$E4,$C7,$17,$F8,$43,$FE,$00
	.dc.b	$3F,$A4,$9F,$DB,$9E,$4F,$F0,$83,$F4,$D3,$FC,$C6,$7F,$06,$7F,$07
	.dc.b	$8F,$E5,$2E,$63,$C4,$FF,$24,$8F,$E3,$47,$CF,$3F,$9A,$1F,$C8,$AF
	.dc.b	$F9,$1C,$FF,$68,$BF,$EB,$E7,$F6,$59,$FD,$44,$7F,$1D,$73,$7E,$A7
	.dc.b	$F8,$F3,$F8,$A7,$FC,$4F,$3A,$9F,$E7,$51,$F4,$5F,$14,$F8,$E3,$FC
	.dc.b	$2B,$1C,$DF,$EC,$4F,$38,$FF,$20,$47,$F8,$76,$3D,$8F,$F1,$A7,$F5
	.dc.b	$4F,$D5,$39,$1F,$E1,$DB,$FA,$B5,$99,$F1,$CF,$E9,$5F,$E4,$0F,$F8
	.dc.b	$82,$72,$FF,$96,$9F,$EB,$75,$F5,$0F,$E6,$31,$FC,$76,$7D,$73,$F5
	.dc.b	$1F,$0C,$7F,$96,$23,$EA,$57,$F2,$C9,$EB,$3F,$E5,$37,$FC,$2E,$FD
	.dc.b	$03,$8D,$7F,$98,$AB,$FA,$B4,$7E,$36,$BE,$A3,$FD,$56,$7E,$5C,$7E
	.dc.b	$92,$3F,$CA,$F3,$EA,$9F,$F0,$C5,$7F,$59,$AF,$A9,$7F,$D4,$E3,$FA
	.dc.b	$71,$FD,$48,$F0,$9F,$A8,$BF,$AB,$3F,$D0,$63,$D3,$1F,$C6,$8F,$DB
	.dc.b	$9F,$0C,$FE,$06,$7F,$8B,$2B,$F8,$D4,$7F,$56,$8F,$EA,$F1,$FD,$FE
	.dc.b	$FF,$8C,$C7,$F4,$E3,$EF,$4F,$F2,$47,$F9,$2C,$FF,$50,$9F,$E3,$73
	.dc.b	$FC,$31,$FE,$97,$3F,$C6,$E3,$F8,$83,$FC,$52,$7F,$87,$CF,$F1,$3A
	.dc.b	$FE,$5C,$7F,$0A,$3F,$C9,$93,$FD,$14,$FE,$24,$66,$7F,$38,$FF,$A2
	.dc.b	$A3,$F8,$94,$7B,$6B,$F8,$8C,$FF,$7D,$3F,$8A,$BF,$25,$FF,$2C,$4F
	.dc.b	$F1,$58,$FE,$1F,$1F,$A3,$7F,$A6,$3F,$D1,$27,$FB,$64,$7F,$59,$8F
	.dc.b	$E8,$67,$13,$F3,$C7,$F0,$B3,$FB,$B9,$F8,$93,$F6,$A7,$F1,$13,$F9
	.dc.b	$F1,$FE,$05,$7E,$BB,$FC,$F6,$BE,$5D,$64,$FE,$5A,$FF,$22,$8F,$F0
	.dc.b	$51,$FC,$E0,$F9,$07,$C4,$8F,$EE,$6F,$FA,$D1,$CC,$FF,$72,$8F,$97
	.dc.b	$5F,$83,$AF,$E7,$71,$F8,$5A,$F9,$D5,$F2,$A3,$F7,$87,$CA,$7E,$89
	.dc.b	$FC,$E1,$FE,$6C,$FC,$A7,$FB,$7B,$FD,$D4,$FE,$66,$FF,$03,$9F,$D1
	.dc.b	$3F,$D6,$CF,$F3,$0C,$FF,$03,$BF,$D0,$47,$F3,$63,$F8,$13,$FC,$0E
	.dc.b	$BF,$D4,$73,$FC,$E5,$FD,$CD,$7C,$5A,$FE,$72,$7F,$A6,$9F,$D8,$1F
	.dc.b	$81,$BF,$F6,$D4,$7F,$A8,$27,$F8,$4B,$FE,$DA,$7F,$7A,$FF,$01,$9F
	.dc.b	$DC,$57,$9E,$3F,$CB,$4F,$F0,$59,$FE,$12,$FF,$95,$23,$FD,$20,$7F
	.dc.b	$34,$9E,$35,$FC,$A6,$7F,$9F,$CF,$FA,$8A,$3C,$2F,$C8,$AF,$D7,$BF
	.dc.b	$E3,$DA,$FE,$11,$1E,$93,$E8,$47,$F3,$47,$F8,$0B,$FC,$E8,$FF,$46
	.dc.b	$9F,$D9,$0F,$F4,$4C,$7F,$30,$7F,$C3,$D1,$FC,$22,$78,$1F,$C0,$A7
	.dc.b	$EA,$BF,$CF,$E7,$CE,$7F,$A5,$1F,$3B,$FE,$13,$9F,$EB,$CF,$F9,$5A
	.dc.b	$7E,$7C,$7F,$47,$7C,$F0,$12,$8F,$E9,$31,$FE,$49,$9F,$A2,$FF,$63
	.dc.b	$3F,$D8,$A7,$C9,$3F,$98,$1E,$23,$FC,$20,$FF,$A0,$DF,$A1,$7F,$E8
	.dc.b	$59,$FF,$41,$5F,$F2,$E9,$FE,$AD,$5F,$CB,$9F,$9D,$5F,$C7,$AB,$F9
	.dc.b	$8D,$7F,$2D,$8F,$E5,$93,$FC,$9C,$FF,$33,$1F,$CB,$0F,$E2,$C7,$F0
	.dc.b	$F3,$FB,$09,$FD,$68,$FF,$75,$9F,$C0,$CF,$31,$F7,$8F,$E0,$67,$F2
	.dc.b	$C3,$FC,$E5,$3F,$C0,$9F,$F3,$EC,$FF,$02,$AF,$EE,$CF,$F9,$D2,$3F
	.dc.b	$84,$9F,$D5,$63,$FC,$77,$1F,$CB,$2F,$FC,$7B,$1F,$C0,$67,$FA,$11
	.dc.b	$FC,$AE,$BF,$80,$BF,$E7,$3A,$FE,$03,$3E,$B8,$FE,$03,$5F,$E6,$D8
	.dc.b	$FF,$54,$4F,$F2,$93,$F8,$23,$FC,$A2,$7F,$CC,$87,$F1,$D3,$E3,$1F
	.dc.b	$D9,$4F,$E5,$07,$C6,$3F,$7C,$7C,$63,$FB,$89,$F1,$8F,$F5,$71,$F4
	.dc.b	$0F,$8C,$7F,$99,$EB,$E3,$3F,$E9,$99,$F8,$D3,$FD,$42,$3F,$93,$57
	.dc.b	$F5,$08,$F8,$B1,$F1,$A3,$E2,$C7,$C4,$8F,$8B,$1F,$C9,$63,$F9,$24
	.dc.b	$7F,$25,$8F,$E9,$D1,$F1,$63,$FD,$16,$FD,$23,$FA,$73,$FE,$26,$3F
	.dc.b	$C4,$93,$F1,$A3,$FB,$8C,$7F,$96,$8F,$F4,$85,$7F,$24,$7F,$60,$7F
	.dc.b	$94,$0F,$CF,$1F,$C3,$4F,$F2,$A1,$FE,$59,$7F,$6E,$7F,$89,$27,$F9
	.dc.b	$25,$FF,$B1,$2B,$F6,$B5,$FC,$7E,$BF,$6A,$7F,$88,$9F,$E4,$51,$FC
	.dc.b	$69,$FF,$0C,$9F,$EC,$F8,$FE,$42,$FE,$C6,$BF,$90,$57,$C2,$8F,$E3
	.dc.b	$E7,$EB,$DF,$D8,$57,$F8,$76,$3F,$DE,$CF,$F8,$B8,$FE,$6D,$7F,$E4
	.dc.b	$F8,$FF,$7D,$DF,$F1,$E9,$FF,$27,$1F,$C2,$23,$FD,$FE,$FF,$1E,$3F
	.dc.b	$A5,$BF,$22,$BE,$4C,$7F,$AE,$6F,$FA,$54,$FF,$6F,$3F,$CF,$97,$FC
	.dc.b	$76,$BF,$A5,$47,$F1,$33,$33,$F2,$1F,$F3,$ED,$7F,$45,$7F,$8E,$4F
	.dc.b	$C9,$8F,$E9,$13,$FE,$49,$8F,$C4,$CF,$F1,$63,$FA,$43,$FE,$7A,$7F
	.dc.b	$8C,$C7,$C8,$7F,$CF,$33,$F9,$D7,$C4,$78,$E9,$3F,$C9,$AF,$FA,$49
	.dc.b	$FC,$6A,$7E,$5D,$FF,$91,$EF,$F8,$54,$72,$3F,$CE,$F3,$F2,$CF,$E3
	.dc.b	$6F,$F8,$EA,$3F,$A3,$BF,$E1,$D3,$F6,$C7,$F8,$2C,$FD,$11,$FE,$00
	.dc.b	$3F,$6C,$7E,$40,$F8,$87,$ED,$8F,$C0,$9F,$B6,$3F,$8D,$9F,$C9,$AF
	.dc.b	$FC,$8D,$1F,$C6,$A3,$F8,$DC,$7E,$56,$3C,$11,$E9,$3E,$A5,$FC,$37
	.dc.b	$07,$FB,$6A,$3F,$72,$E0,$FF,$6D,$16,$3F,$C9,$2F,$F0,$F3,$F7,$66
	.dc.b	$67,$FC,$4C,$7F,$B8,$A3,$FA,$99,$FC,$82,$3E,$01,$FC,$80,$FF,$4E
	.dc.b	$52,$BF,$90,$C7,$F2,$2A,$FA,$4F,$F4,$63,$E3,$9F,$72,$78,$3F,$C8
	.dc.b	$8F,$75,$7F,$A1,$5F,$D1,$4F,$F2,$83,$FC,$E5,$5E,$D8,$FE,$D1,$1F
	.dc.b	$EF,$87,$CE,$FF,$89,$9F,$9E,$7F,$93,$1F,$9E,$FF,$23,$3E,$79,$F9
	.dc.b	$C3,$E3,$1F,$E5,$B3,$FC,$93,$5F,$E7,$37,$F9,$21,$A9,$FE,$90,$FF
	.dc.b	$9C,$E7,$F9,$23,$FD,$1C,$FE,$48,$7E,$70,$F9,$27,$F2,$43,$FC,$91
	.dc.b	$3F,$09,$FE,$4F,$5F,$D2,$67,$E4,$C7,$F2,$B3,$F0,$07,$FA,$56,$7F
	.dc.b	$92,$1F,$E1,$97,$F0,$71,$FE,$FC,$7F,$42,$7F,$85,$4F,$E4,$97,$FB
	.dc.b	$E7,$FD,$F5,$5F,$BD,$7F,$D2,$97,$F9,$A7,$FC,$8D,$3F,$BD,$7A,$DF
	.dc.b	$F1,$1B,$FF,$0B,$57,$F8,$3E,$FF,$1B,$5F,$AF,$9F,$A8,$7F,$B1,$6F
	.dc.b	$F8,$04,$7F,$91,$8F,$CB,$47,$EF,$1E,$4F,$FB,$18,$FB,$CF,$FB,$1A
	.dc.b	$3F,$94,$BE,$E9,$FF,$18,$BF,$C1,$5F,$F1,$4D,$7F,$2B,$3F,$DF,$27
	.dc.b	$CD,$8F,$8A,$FE,$5E,$FF,$81,$3F,$EB,$A9,$FF,$0E,$4F,$F7,$23,$D4
	.dc.b	$7F,$92,$0F,$F6,$F9,$FE,$5F,$32,$7F,$53,$8F,$F2,$41,$FE,$F2,$7E
	.dc.b	$79,$F5,$0F,$F3,$89,$FE,$47,$8F,$A0,$FE,$F6,$7F,$A0,$9F,$EF,$53
	.dc.b	$FA,$9C,$FF,$2C,$8F,$EA,$75,$FC,$B1,$FE,$09,$7F,$3A,$FF,$CB,$91
	.dc.b	$F4,$5F,$33,$FC,$AA,$3E,$74,$7F,$2F,$8F,$F7,$A5,$7C,$57,$51,$FE
	.dc.b	$F5,$9F,$E5,$E7,$FB,$F9,$F4,$C7,$CF,$8F,$F7,$F1,$FE,$4F,$9F,$EB
	.dc.b	$27,$38,$FC,$13,$F3,$EB,$FC,$94,$FC,$F8,$FE,$61,$1F,$C1,$5F,$E0
	.dc.b	$A7,$F5,$1B,$E4,$FF,$39,$BF,$C1,$BF,$E6,$77,$F8,$35,$6A,$3F,$4A
	.dc.b	$7F,$B3,$CF,$CD,$1F,$1A,$38,$DF,$FA,$CE,$3F,$A6,$4F,$FB,$D9,$FF
	.dc.b	$1A,$47,$F0,$67,$F9,$6D,$7F,$6F,$7F,$3F,$5F,$E2,$AA,$FE,$6B,$1F
	.dc.b	$CC,$A3,$C2,$F9,$CF,$F6,$81,$FC,$AC,$F0,$1F,$9F,$8F,$E1,$2F,$F9
	.dc.b	$06,$BF,$84,$BC,$DF,$E6,$47,$09,$FF,$0F,$47,$F4,$39,$FF,$65,$1F
	.dc.b	$E2,$D3,$FD,$A2,$7E,$AD,$FF,$16,$CF,$F0,$A7,$E7,$1F,$83,$3F,$AE
	.dc.b	$9F,$C8,$0F,$D7,$CF,$CC,$9F,$F3,$E4,$7F,$9D,$CF,$C7,$D7,$CF,$7F
	.dc.b	$98,$CF,$CC,$3F,$11,$3F,$2E,$3F,$CA,$67,$E7,$5F,$E0,$EF,$F3,$13
	.dc.b	$FC,$A9,$5F,$D7,$DF,$11,$FE,$54,$3F,$99,$D7,$F3,$17,$F9,$AD,$7F
	.dc.b	$BE,$2B,$FB,$9C,$7C,$93,$FA,$D4,$FF,$9E,$1F,$95,$3F,$D6,$5F,$E6
	.dc.b	$07,$F4,$99,$FE,$A3,$5F,$E7,$98,$FF,$29,$9F,$D6,$4F,$F5,$1C,$7F
	.dc.b	$95,$1F,$D7,$9F,$C0,$EB,$F9,$8C,$7F,$8B,$63,$FD,$93,$7F,$D4,$5F
	.dc.b	$F1,$69,$FC,$08,$E3,$5F,$EC,$93,$EB,$38,$9F,$F7,$B9,$FD,$AC,$FC
	.dc.b	$CB,$FC,$AE,$BF,$54,$FF,$BE,$63,$F8,$23,$FC,$D1,$FE,$64,$FF,$95
	.dc.b	$8F,$F1,$79,$FE,$C6,$8F,$D3,$4F,$F0,$28,$FE,$63,$7F,$ED,$1A,$FF
	.dc.b	$7B,$CF,$F3,$19,$FF,$63,$9F,$CB,$DF,$F7,$B1,$FC,$02,$3F,$97,$1F
	.dc.b	$EC,$78,$FF,$63,$5F,$EF,$9F,$E5,$F1,$EE,$7F,$81,$BF,$E7,$87,$FD
	.dc.b	$E8,$7F,$5B,$8F,$F2,$BB,$FC,$9C,$FE,$58,$FC,$B7,$FA,$D1,$FD,$20
	.dc.b	$FF,$21,$3F,$D6,$DF,$F2,$B9,$99,$FF,$71,$5F,$F2,$CA,$FF,$04,$BF
	.dc.b	$CB,$27,$E4,$5F,$F9,$5D,$FE,$F9,$3F,$E3,$97,$F8,$41,$FC,$84,$FF
	.dc.b	$3F,$1F,$20,$FE,$AE,$FC,$C8,$FE,$50,$79,$23,$E4,$4F,$F9,$55,$F9
	.dc.b	$53,$F2,$0F,$EE,$AF,$C4,$8F,$EA,$6F,$C8,$9F,$EA,$6F,$EE,$0F,$DC
	.dc.b	$1F,$E2,$18,$FD,$93,$FC,$8C,$FE,$4B,$1F,$B9,$7F,$C4,$D5,$FE,$29
	.dc.b	$6C,$F0,$8F,$E9,$07,$CD,$8F,$F1,$A5,$7F,$2E,$9F,$D2,$C7,$FB,$C2
	.dc.b	$3F,$CA,$CF,$EE,$0F,$F1,$41,$FC,$84,$FF,$14,$4F,$FB,$2D,$FD,$B9
	.dc.b	$FE,$28,$35,$1F,$C8,$1F,$F1,$FC,$7F,$BD,$AB,$FA,$6C,$7F,$97,$A3
	.dc.b	$F9,$0C,$7E,$E6,$3F,$90,$4F,$FB,$1A,$7F,$71,$1F,$E2,$83,$F7,$07
	.dc.b	$FB,$42,$3F,$DE,$0F,$F7,$93,$FD,$CC,$7F,$0E,$3F,$CC,$07,$F7,$F3
	.dc.b	$F7,$A7,$F8,$42,$3F,$C9,$AF,$C8,$8C,$4F,$F1,$F9,$F9,$13,$FD,$5E
	.dc.b	$7E,$41,$C1,$FF,$79,$3F,$21,$FE,$96,$FF,$42,$AF,$DF,$3F,$D2,$A3
	.dc.b	$FC,$C2,$7C,$33,$FA,$4D,$7F,$34,$7F,$A1,$1F,$C1,$4F,$49,$FE,$7A
	.dc.b	$3F,$A4,$4F,$F2,$17,$E6,$C7,$C9,$7F,$D4,$53,$FD,$21,$FF,$28,$BF
	.dc.b	$E7,$87,$FA,$39,$FC,$80,$F8,$87,$F5,$C3,$F6,$93,$FC,$81,$FD,$B4
	.dc.b	$FF,$B3,$1F,$CD,$9F,$22,$BF,$D9,$8F,$F8,$85,$FE,$41,$1F,$E2,$08
	.dc.b	$FF,$5A,$4F,$ED,$5F,$F6,$59,$FC,$74,$FD,$A9,$FE,$20,$3F,$6A,$7F
	.dc.b	$B2,$CF,$D9,$3F,$EF,$23,$F6,$4F,$F0,$F3,$FC,$36,$7E,$DA,$BF,$D9
	.dc.b	$8F,$F8,$96,$BF,$C3,$87,$F1,$CA,$FD,$39,$FE,$4A,$BF,$ED,$13,$FE
	.dc.b	$29,$AF,$D4,$1F,$CC,$9F,$F1,$8B,$FE,$4B,$3F,$C3,$8F,$F2,$07,$FD
	.dc.b	$E8,$7F,$87,$1F,$F3,$CD,$FC,$97,$F1,$F5,$F2,$5F,$F6,$31,$FD,$20
	.dc.b	$FE,$35,$0A,$FE,$3D,$1F,$E4,$93,$F8,$F5,$7F,$48,$3F,$9E,$CF,$F9
	.dc.b	$26,$7F,$CF,$13,$FD,$E2,$7E,$23,$FE,$EB,$9F,$E9,$D3,$FE,$4B,$8F
	.dc.b	$F2,$41,$F9,$E3,$F9,$04,$7C,$CA,$FF,$24,$47,$EE,$2B,$F9,$05,$FE
	.dc.b	$90,$F9,$75,$FC,$7E,$7F,$C6,$11,$FE,$45,$9F,$F3,$99,$FA,$D3,$E5
	.dc.b	$47,$F2,$17,$E4,$BF,$31,$FE,$8C,$FF,$BC,$CF,$E4,$0F,$FA,$D5,$FF
	.dc.b	$00,$9F,$C1,$0F,$E3,$E7,$F8,$F8,$FF,$2D,$4F,$F1,$FA,$FD,$C4,$7F
	.dc.b	$24,$9F,$F0,$B3,$FE,$F3,$3F,$8F,$1F,$E1,$53,$F6,$F1,$FE,$3F,$7F
	.dc.b	$84,$1F,$C7,$CF,$F6,$2C,$7F,$BC,$1F,$E4,$51,$FE,$15,$7F,$6A,$7F
	.dc.b	$B1,$4F,$F0,$AC,$7F,$B6,$63,$FD,$33,$1F,$C7,$DF,$DB,$22,$1F,$BF
	.dc.b	$3F,$C3,$11,$FE,$3F,$9F,$F1,$93,$FE,$55,$3F,$DE,$52,$AF,$CC,$D7
	.dc.b	$F4,$97,$F8,$AB,$FE,$EF,$3F,$CD,$8F,$F8,$75,$FE,$E0,$7F,$9A,$4F
	.dc.b	$F0,$61,$FA,$43,$EF,$1F,$26,$7E,$9B,$FE,$68,$3E,$49,$F2,$0F,$F7
	.dc.b	$69,$FE,$CC,$3E,$49,$FC,$78,$F9,$27,$F4,$63,$FC,$7B,$1F,$93,$7C
	.dc.b	$A7,$F9,$51,$FE,$8C,$7F,$92,$9F,$F1,$E9,$FE,$EE,$3F,$D3,$71,$FE
	.dc.b	$E7,$7F,$8F,$BF,$1C,$FE,$3C,$FF,$BB,$A3,$E2,$BA,$3E,$09,$FE,$0D
	.dc.b	$3F,$D7,$E7,$73,$FD,$7F,$3F,$B7,$7F,$2C,$7F,$1D,$3F,$C7,$C7,$FB
	.dc.b	$22,$3F,$8F,$BF,$B6,$BF,$F2,$F3,$FB,$33,$F8,$F9,$FB,$5B,$FD,$CD
	.dc.b	$FE,$DA,$3F,$05,$5F,$EE,$E8,$FE,$05,$1F,$C8,$8F,$F0,$A9,$FE,$3E
	.dc.b	$3F,$C2,$F3,$FB,$77,$FC,$2D,$1F,$E3,$C3,$F6,$EF,$F8,$F0,$FD,$AC
	.dc.b	$7E,$DD,$FE,$49,$1F,$E1,$57,$F8,$99,$FB,$38,$FF,$0A,$BF,$B5,$BF
	.dc.b	$F0,$AB,$F9,$63,$E5,$1F,$DA,$8F,$D9,$9F,$2A,$3F,$CA,$2F,$F8,$5C
	.dc.b	$FF,$1F,$4F,$FB,$A0,$FF,$47,$3F,$26,$3D,$EF,$F9,$A1,$E3,$5F,$16
	.dc.b	$3F,$3C,$7F,$37,$3F,$C2,$67,$F9,$9E,$7F,$C7,$F7,$FC,$8A,$31,$7F
	.dc.b	$E3,$F7,$FC,$27,$3F,$E3,$F7,$F9,$0D,$FF,$4B,$8F,$F1,$F4,$FF,$20
	.dc.b	$9F,$D3,$C7,$CA,$7F,$CD,$37,$FC,$82,$2C,$FF,$4A,$8F,$E4,$11,$FE
	.dc.b	$65,$9F,$F3,$1D,$7C,$A7,$FC,$C7,$5F,$29,$FF,$1F,$1E,$C3,$F5,$A7
	.dc.b	$F8,$14,$FD,$19,$F9,$D3,$FB,$A1,$FE,$8C,$3F,$C2,$C7,$ED,$CF,$11
	.dc.b	$FB,$73,$FC,$42,$7F,$84,$8E,$C7,$FA,$D6,$3F,$D7,$87,$FB,$7C,$FE
	.dc.b	$17,$5F,$98,$8F,$DC,$1F,$EB,$53,$FC,$81,$1F,$EB,$57,$F9,$0C,$7E
	.dc.b	$E5,$FC,$24,$F6,$7F,$D6,$C7,$F8,$5A,$BF,$91,$4F,$F8,$F8,$AF,$F1
	.dc.b	$C3,$FC,$7D,$1F,$8D,$3F,$1C,$FF,$6F,$7F,$D6,$E7,$F8,$60,$FF,$20
	.dc.b	$3F,$29,$FF,$33,$9F,$00,$F4,$9F,$CD,$0F,$F6,$19,$FC,$00,$FD,$D1
	.dc.b	$FE,$B0,$3F,$C1,$27,$CA,$3F,$C3,$A7,$F3,$5A,$FF,$38,$BF,$ED,$C9
	.dc.b	$FE,$AB,$1F,$2A,$7F,$2D,$28,$F9,$6F,$E5,$8F,$65,$7B,$5F,$E9,$95
	.dc.b	$FE,$69,$7F,$A5,$9F,$E6,$93,$FC,$8B,$5F,$E6,$97,$F9,$2C,$7C,$AA
	.dc.b	$FF,$20,$47,$F3,$CA,$FE,$96,$FF,$4A,$AF,$F2,$05,$7F,$9A,$23,$FC
	.dc.b	$D7,$3E,$17,$E6,$CA,$3E,$C4,$62,$3D,$E0,$CB,$F7,$67,$FC,$8B,$78
	.dc.b	$7F,$CC,$87,$F9,$1A,$30,$7F,$B3,$E3,$8B,$FC,$7A,$7A,$4F,$D2,$3F
	.dc.b	$C9,$17,$FE,$21,$6C,$72,$7C,$D3,$63,$4F,$F9,$B9,$E6,$FF,$2F,$3C
	.dc.b	$93,$E9,$36,$FF,$07,$76,$FF,$06,$37,$1E,$FA,$F3,$DF,$FB,$9E,$3D
	.dc.b	$CF,$E3,$A7,$FB,$BD,$F9,$2F,$AC,$7F,$94,$9E,$AF,$F9,$52,$7F,$D0
	.dc.b	$CF,$F9,$5A,$B6,$FE,$A2,$BE,$74,$76,$7F,$7E,$7D,$53,$F1,$97,$9B
	.dc.b	$EB,$5E,$2B,$FF,$01,$1D,$9C,$57,$AC,$FC,$03,$FE,$92,$BF,$B0,$7D
	.dc.b	$D7,$81,$F5,$5E,$F1,$FC,$69,$D9,$F4,$9F,$F3,$05,$FD,$C9,$F8,$D1
	.dc.b	$FD,$B0,$FC,$29,$FE,$8A,$9F,$C1,$C7,$13,$FD,$62,$FC,$27,$FC,$E0
	.dc.b	$FD,$F7,$C6,$7F,$40,$8F,$1B,$FD,$16,$3F,$09,$5A,$9F,$C6,$57,$E3
	.dc.b	$AF,$FD,$5F,$1F,$E1,$CB,$FB,$B1,$DD,$FF,$14,$3F,$E7,$B8,$FF,$48
	.dc.b	$9F,$A5,$3F,$B9,$4F,$E0,$A7,$EA,$C7,$FA,$4C,$D5,$7F,$05,$9F,$C2
	.dc.b	$CF,$F0,$1B,$B3,$E5,$3F,$45,$5F,$49,$FF,$1C,$47,$DA,$9F,$F0,$0C
	.dc.b	$7F,$43,$9E,$A7,$C0,$8F,$F4,$D5,$ED,$FB,$F1,$F2,$DF,$EF,$77,$FA
	.dc.b	$EB,$F8,$31,$FE,$18,$3E,$CD,$7C,$1A,$FC,$7B,$CD,$FC,$8D,$FC,$BA
	.dc.b	$FF,$43,$C7,$E2,$2B,$EE,$9F,$DF,$CF,$B7,$5C,$DF,$E4,$D3,$F7,$AB
	.dc.b	$EA,$47,$C0,$3F,$23,$5F,$29,$FE,$B4,$7E,$F6,$B9,$47,$C9,$3F,$D9
	.dc.b	$53,$C9,$FC,$D4,$FE,$FE,$BF,$6E,$FE,$06,$BF,$7D,$7F,$7A,$7F,$77
	.dc.b	$79,$78,$D7,$F2,$C7,$EB,$BF,$54,$F9,$C7,$EE,$5F,$B2,$FD,$C7,$F1
	.dc.b	$0F,$EC,$DF,$91,$1F,$5E,$3F,$C5,$B3,$FB,$59,$FD,$D4,$FE,$06,$FC
	.dc.b	$35,$FE,$56,$9F,$A0,$7E,$C5,$FA,$B5,$F5,$A3,$F5,$F5,$F4,$DE,$33
	.dc.b	$FD,$25,$FB,$0F,$F1,$33,$FC,$92,$7E,$4E,$BF,$60,$7F,$33,$3F,$C6
	.dc.b	$35,$FB,$77,$E9,$47,$FB,$45,$FE,$63,$1F,$4E,$7F,$5B,$3F,$CB,$E3
	.dc.b	$F5,$91,$C2,$3F,$C8,$53,$FE,$75,$3F,$51,$7F,$49,$B3,$F6,$E7,$FD
	.dc.b	$A9,$7F,$4E,$3E,$5B,$F8,$C8,$FF,$69,$9F,$4E,$BF,$2D,$1F,$8C,$AF
	.dc.b	$D7,$CF,$E7,$CF,$E5,$57,$FC,$DA,$3C,$91,$FC,$80,$FA,$15,$CA,$3C
	.dc.b	$51,$FE,$0D,$7F,$96,$57,$D5,$BF,$BA,$7E,$A2,$74,$FD,$59,$FF,$1D
	.dc.b	$DF,$D5,$7F,$83,$47,$FA,$CA,$7F,$92,$BF,$86,$7D,$87,$F4,$78,$FE
	.dc.b	$46,$FF,$80,$5F,$EC,$ED,$3F,$D8,$55,$FA,$87,$EF,$9F,$C9,$23,$FB
	.dc.b	$33,$FE,$65,$AF,$AD,$1F,$A6,$8F,$AC,$FF,$3E,$BF,$54,$7D,$67,$D4
	.dc.b	$FF,$A4,$67,$F4,$D3,$FC,$82,$3C,$93,$FC,$01,$FF,$33,$C7,$F1,$F7
	.dc.b	$FD,$1F,$3F,$C7,$A7,$F4,$31,$F4,$CF,$EA,$47,$A0,$F5,$CF,$D4,$7F
	.dc.b	$8E,$BF,$D7,$2B,$F9,$8D,$7E,$9E,$3F,$8F,$C7,$D3,$8F,$E3,$57,$C6
	.dc.b	$BF,$93,$BF,$D9,$0F,$C7,$3F,$E9,$77,$F4,$EF,$D2,$8F,$EA,$8F,$1A
	.dc.b	$FC,$03,$F3,$CF,$DF,$9F,$DE,$8F,$C7,$19,$AF,$A5,$5F,$6E,$79,$4F
	.dc.b	$EF,$CF,$F4,$C5,$7D,$27,$ED,$47,$C3,$7F,$66,$FC,$37,$FD,$79,$3F
	.dc.b	$C6,$1F,$EA,$73,$F4,$9F,$D2,$9F,$E1,$28,$FD,$2B,$E9,$AF,$DF,$3F
	.dc.b	$DF,$8F,$49,$FD,$20,$FC,$E9,$FE,$0F,$3F,$8C,$4F,$E9,$67,$F8,$BD
	.dc.b	$7D,$13,$8C,$FF,$4A,$BC,$DF,$D1,$9F,$F0,$3D,$7E,$2A,$BF,$C3,$EF
	.dc.b	$E2,$9F,$D2,$BC,$A3,$FA,$5B,$F4,$9F,$F0,$29,$FD,$26,$3E,$49,$FE
	.dc.b	$1C,$3F,$8B,$1F,$4E,$BF,$8D,$BF,$24,$FF,$52,$4F,$D1,$3F,$8F,$CF
	.dc.b	$13,$C1,$1F,$DE,$A3,$F8,$BC,$7F,$79,$8F,$ED,$B7,$FC,$62,$7F,$AF
	.dc.b	$57,$A6,$7F,$49,$1E,$98,$FF,$01,$47,$D2,$AF,$E3,$33,$F4,$8F,$CF
	.dc.b	$C7,$F1,$5A,$FD,$14,$7F,$46,$9F,$E2,$AF,$F1,$C7,$EC,$1F,$3E,$BF
	.dc.b	$89,$1F,$3A,$BF,$C7,$D3,$FC,$40,$FC,$B1,$E5,$7E,$7B,$F3,$4F,$2B
	.dc.b	$FE,$CA,$3F,$3D,$3F,$36,$3E,$4C,$7F,$09,$9F,$EE,$8F,$E1,$0F,$CF
	.dc.b	$1F,$36,$7F,$D0,$D5,$FD,$0E,$BE,$6C,$FF,$6F,$AF,$EE,$67,$D0,$BF
	.dc.b	$EE,$53,$FD,$8A,$3F,$09,$23,$E9,$1F,$DB,$DF,$F0,$71,$E8,$7E,$7C
	.dc.b	$FF,$41,$AF,$F1,$65,$7C,$C9,$FE,$19,$1F,$32,$FF,$9F,$D7,$F7,$5A
	.dc.b	$F9,$87,$E6,$AB,$F9,$ED,$7F,$09,$7F,$96,$4F,$E9,$6B,$FC,$53,$5F
	.dc.b	$34,$FE,$DE,$7F,$09,$3E,$61,$F1,$67,$FC,$C1,$3F,$DB,$E3,$F9,$D4
	.dc.b	$7F,$BC,$0F,$EB,$F3,$F9,$BB,$FF,$19,$47,$C1,$3F,$30,$5A,$4F,$F1
	.dc.b	$2A,$DC,$FF,$09,$3F,$83,$C7,$F8,$B4,$F1,$3F,$CA,$9C,$4F,$27,$53
	.dc.b	$E0,$AF,$C3,$57,$85,$FF,$3D,$4F,$F7,$67,$53,$E4,$7F,$DF,$B1,$FD
	.dc.b	$16,$3F,$96,$4F,$FA,$B0,$FF,$37,$96,$3F,$CC,$13,$EE,$AF,$EA,$B3
	.dc.b	$FE,$63,$7C,$8F,$F4,$63,$FD,$98,$FB,$5F,$F4,$E1,$A3,$FD,$C0,$FE
	.dc.b	$DE,$7F,$CC,$47,$C8,$7F,$CB,$6E,$2B,$E6,$9E,$E9,$FF,$59,$1C,$CF
	.dc.b	$A0,$7F,$AD,$6B,$71,$F4,$8F,$7C,$7F,$B7,$DF,$CB,$BF,$E5,$93,$FB
	.dc.b	$94,$FD,$B8,$FE,$8F,$5F,$0A,$7F,$AC,$CF,$EB,$67,$FC,$51,$19,$8F
	.dc.b	$D0,$1F,$E2,$BA,$FB,$AF,$E0,$A3,$D6,$7F,$98,$27,$E9,$BF,$3A,$7E
	.dc.b	$BC,$FD,$37,$EB,$D6,$6B,$FD,$81,$3F,$5D,$FA,$55,$F9,$39,$FF,$68
	.dc.b	$CF,$FA,$2E,$3F,$12,$7F,$B9,$A3,$ED,$9E,$A7,$EB,$1F,$4A,$39,$1E
	.dc.b	$A3,$E9,$D7,$DF,$7E,$A1,$E2,$8F,$BA,$FE,$22,$BE,$A4,$7F,$B6,$0F
	.dc.b	$A8,$7D,$47,$F8,$03,$5F,$A8,$73,$7F,$CB,$4F,$E4,$9F,$EB,$A7,$EB
	.dc.b	$8F,$F4,$84,$FE,$A1,$F9,$87,$13,$EE,$47,$D4,$7E,$AC,$FD,$38,$FF
	.dc.b	$2F,$57,$D7,$3C,$B3,$FE,$2B,$3F,$12,$7E,$A6,$3F,$7D,$5F,$52,$32
	.dc.b	$7F,$A7,$1F,$AA,$FF,$98,$A7,$9B,$FE,$A0,$9F,$AA,$FD,$03,$E6,$1F
	.dc.b	$A4,$3E,$AC,$7E,$AA,$BE,$C3,$CA,$0F,$F9,$21,$FD,$49,$FE,$5D,$34
	.dc.b	$FE,$A5,$FA,$6F,$DE,$8F,$89,$1F,$A4,$8F,$F1,$3C,$7F,$7A,$9F,$E3
	.dc.b	$47,$B6,$7E,$8D,$7E,$A2,$BF,$80,$CF,$EA,$27,$EA,$BF,$45,$FD,$21
	.dc.b	$EC,$3F,$85,$57,$F7,$D9,$C9,$EB,$8F,$51,$FE,$86,$3F,$50,$66,$3F
	.dc.b	$2E,$FF,$37,$3F,$D8,$B1,$99,$FD,$3C,$7F,$A4,$0F,$8D,$5F,$8C,$7E
	.dc.b	$C4,$FE,$9C,$CC,$FF,$1C,$7F,$16,$7E,$9A,$3E,$94,$FE,$AA,$3E,$B1
	.dc.b	$F3,$23,$85,$7E,$91,$F9,$EF,$D8,$3F,$47,$18,$AF,$E1,$F3,$F2,$1F
	.dc.b	$88,$FE,$31,$F3,$3F,$D9,$5F,$F4,$34,$FD,$39,$FC,$04,$7F,$14,$9F
	.dc.b	$F2,$A9,$FD,$0E,$7F,$8A,$1F,$35,$FE,$2A,$FF,$4B,$8F,$D0,$3F,$D2
	.dc.b	$63,$E6,$57,$F1,$2A,$FD,$04,$FF,$7D,$AF,$95,$5F,$2E,$3F,$A3,$47
	.dc.b	$E6,$E3,$F3,$91,$FE,$27,$8F,$EB,$31,$F9,$93,$F3,$87,$C7,$AF,$94
	.dc.b	$FF,$43,$9F,$EF,$73,$F9,$78,$FC,$C4,$7F,$AD,$67,$F8,$65,$7E,$5C
	.dc.b	$F9,$55,$FE,$7E,$AF,$94,$7C,$47,$F8,$D9,$F2,$1F,$DF,$C6,$63,$F2
	.dc.b	$91,$FE,$7E,$3F,$20,$7E,$30,$FE,$0F,$18,$8F,$E6,$B1,$F9,$2A,$FE
	.dc.b	$0C,$7F,$6E,$7F,$83,$BF,$1C,$FE,$CC,$FF,$33,$3F,$1B,$5F,$C0,$A3
	.dc.b	$FB,$55,$7E,$42,$BF,$80,$D7,$C8,$3F,$AF,$9F,$84,$3F,$68,$7F,$02
	.dc.b	$3F,$B0,$1F,$B8,$3F,$C1,$07,$E4,$8F,$DE,$1F,$92,$AF,$92,$FF,$B7
	.dc.b	$E7,$E4,$D7,$EF,$63,$E4,$BF,$EA,$89,$FD,$E4,$7F,$AA,$A7,$E4,$57
	.dc.b	$C7,$8F,$91,$5F,$EA,$6A,$FE,$5D,$5F,$BA,$8F,$C7,$1F,$8D,$9F,$E5
	.dc.b	$AF,$F8,$5E,$7F,$1D,$1F,$22,$3F,$C2,$B5,$F8,$E7,$E4,$9F,$8E,$3E
	.dc.b	$3C,$7E,$39,$FE,$B9,$1F,$1E,$3F,$99,$D7,$EE,$1F,$90,$7C,$4A,$FC
	.dc.b	$7D,$7F,$28,$3F,$1E,$7F,$82,$A3,$E3,$BF,$8F,$AF,$8D,$1F,$1E,$3F
	.dc.b	$D4,$51,$F8,$F8,$FC,$84,$7C,$78,$F8,$D1,$F1,$E3,$FD,$A3,$1F,$8E
	.dc.b	$7F,$D3,$51,$F8,$E9,$F8,$EF,$ED,$23,$FC,$0C,$7C,$73,$F6,$87,$E3
	.dc.b	$9F,$E3,$A7,$C7,$3F,$C0,$93,$F1,$DF,$E9,$E2,$7F,$9B,$3F,$D5,$9F
	.dc.b	$E1,$6E,$63,$FC,$78,$7E,$25,$FE,$A4,$FC,$B7,$F1,$E7,$EE,$27,$FC
	.dc.b	$77,$1F,$E3,$98,$FD,$C4,$7F,$82,$63,$FC,$01,$1F,$2A,$3F,$C0,$71
	.dc.b	$F2,$9F,$F1,$BD,$7E,$42,$7F,$B8,$9F,$90,$8F,$E9,$D5,$F9,$07,$F8
	.dc.b	$F1,$F2,$AB,$F1,$73,$FD,$7A,$BF,$A2,$3F,$EF,$17,$FC,$A4,$FE,$A6
	.dc.b	$97,$FC,$4A,$73,$7F,$DD,$CF,$F4,$83,$8A,$FE,$D8,$75,$3F,$D3,$F7
	.dc.b	$CA,$77,$3F,$DF,$14,$47,$33,$FE,$7B,$7F,$C7,$AC,$9E,$D1,$F5,$6B
	.dc.b	$C9,$1F,$D7,$63,$EA,$1C,$A7,$FE,$00,$C7,$13,$55,$E3,$8F,$31,$FE
	.dc.b	$11,$7C,$4E,$0F,$F5,$59,$D8,$F2,$5F,$A0,$E5,$3F,$11,$FF,$6C,$B6
	.dc.b	$7F,$C1,$73,$C4,$FF,$2B,$CF,$C0,$8E,$33,$FB,$09,$C5,$7A,$5F,$F4
	.dc.b	$CC,$7F,$79,$3E,$89,$FE,$F8,$2B,$F0,$0F,$44,$7F,$94,$5F,$01,$C8
	.dc.b	$F9,$F7,$87,$FC,$3F,$1E,$07,$8C,$FF,$29,$8F,$11,$AB,$F3,$CF,$D5
	.dc.b	$9F,$F1,$11,$E4,$9F,$79,$C1,$FF,$80,$2F,$AE,$3E,$05,$FC,$E9,$FC
	.dc.b	$5B,$F8,$28,$FA,$C7,$FB,$C2,$BB,$BF,$E7,$E8,$F9,$B3,$FE,$90,$AF
	.dc.b	$94,$F0,$9D,$4F,$BA,$7C,$11,$FE,$01,$8F,$87,$1F,$E5,$03,$DB,$1F
	.dc.b	$2E,$7F,$47,$5E,$57,$F4,$A7,$58,$FE,$99,$5E,$47,$F2,$CE,$DF,$CC
	.dc.b	$1F,$24,$FA,$67,$FB,$4A,$78,$57,$E0,$A7,$C7,$3F,$E3,$1A,$FC,$EC
	.dc.b	$75,$8F,$E1,$CF,$57,$81,$E3,$9F,$E2,$55,$D0,$F8,$B1,$F9,$69,$FA
	.dc.b	$6F,$E1,$23,$F2,$17,$F8,$1B,$F8,$97,$F5,$EB,$F8,$14,$7C,$D3,$E2
	.dc.b	$3F,$7C,$FC,$73,$F9,$03,$C4,$7F,$8B,$67,$A4,$FE,$2E,$3F,$09,$3A
	.dc.b	$8F,$DC,$1F,$85,$39,$C7,$77,$CE,$FE,$2E,$FE,$11,$F8,$88,$FC,$9B
	.dc.b	$F8,$79,$FB,$6F,$EC,$EA,$BD,$23,$EA,$BD,$2B,$F1,$0F,$C2,$AF,$C2
	.dc.b	$CF,$9E,$7F,$91,$BF,$34,$F8,$27,$F1,$23,$A1,$5D,$3F,$8F,$3F,$84
	.dc.b	$1F,$92,$8F,$C6,$9F,$31,$FC,$2B,$F0,$9F,$C2,$BD,$1F,$F1,$FB,$E7
	.dc.b	$77,$3F,$83,$BF,$85,$1F,$6E,$3A,$3F,$8B,$8F,$C1,$47,$39,$FE,$AD
	.dc.b	$5F,$87,$9F,$81,$5E,$68,$FC,$0B,$FC,$00,$E6,$7F,$07,$8F,$CB,$C2
	.dc.b	$3A,$3F,$CF,$8F,$E2,$35,$F9,$77,$FD,$3F,$1F,$C2,$AB,$99,$FC,$26
	.dc.b	$FF,$63,$3F,$C2,$0F,$C0,$1F,$85,$AF,$BE,$7D,$3A,$FB,$27,$E7,$DF
	.dc.b	$BD,$5F,$82,$7D,$C7,$F9,$56,$7F,$85,$BF,$4D,$FE,$B7,$3E,$93,$A1
	.dc.b	$F6,$CF,$B3,$7F,$B4,$7D,$CF,$EB,$DF,$F2,$54,$7D,$E9,$FD,$A4,$7B
	.dc.b	$A3,$FA,$9B,$F8,$7A,$FB,$91,$FD,$5A,$B9,$47,$C9,$3E,$E4,$FF,$5C
	.dc.b	$7C,$D3,$FE,$0D,$AF,$2C,$FF,$50,$9F,$E6,$35,$F0,$23,$F0,$0F,$43
	.dc.b	$E0,$57,$F2,$37,$EE,$BF,$00,$FB,$A7,$F4,$68,$FD,$B9,$E6,$7F,$7C
	.dc.b	$FD,$C3,$CA,$7F,$47,$AF,$B9,$5E,$43,$FA,$3C,$7E,$EE,$BF,$0B,$1F
	.dc.b	$84,$38,$9F,$D3,$27,$ED,$9F,$08,$FD,$C3,$F6,$63,$F9,$1C,$7F,$18
	.dc.b	$9F,$E7,$35,$F7,$E3,$E2,$C7,$D9,$3E,$D3,$F3,$1F,$EC,$91,$FC,$E2
	.dc.b	$FF,$95,$C7,$E8,$4F,$CB,$47,$F8,$7A,$7E,$CD,$FA,$E7,$F9,$B4,$FD
	.dc.b	$7A,$FE,$67,$3C,$4F,$5C,$FF,$5A,$3F,$7B,$7F,$B2,$AF,$EF,$D1,$FA
	.dc.b	$DB,$F9,$8F,$F3,$19,$FA,$D1,$F1,$1F,$AB,$3F,$DF,$E3,$F0,$AF,$AA
	.dc.b	$7F,$97,$4F,$EC,$1F,$A9,$1F,$62,$7E,$99,$C2,$BF,$5B,$1F,$D3,$6F
	.dc.b	$F5,$B1,$F4,$E3,$F5,$91,$FD,$FE,$7D,$11,$FC,$89,$FD,$A9,$F8,$D7
	.dc.b	$F5,$37,$F5,$E3,$ED,$9E,$8B,$FA,$47,$E3,$27,$F9,$3B,$F4,$27,$F5
	.dc.b	$4F,$D1,$3E,$49,$F6,$8F,$05,$FF,$24,$7F,$51,$1F,$B1,$8F,$A1,$3F
	.dc.b	$AB,$AF,$E4,$07,$E4,$67,$F1,$35,$FA,$2B,$FE,$F7,$1F,$89,$8F,$ED
	.dc.b	$D3,$F3,$CF,$03,$FE,$56,$3F,$35,$5F,$3D,$FD,$A3,$FC,$F9,$FE,$41
	.dc.b	$1F,$C5,$9F,$CD,$C7,$F5,$DB,$FF,$0D,$3F,$C0,$4F,$EB,$11,$FA,$3B
	.dc.b	$FA,$37,$F5,$67,$E8,$D7,$F8,$86,$FE,$8D,$FF,$1C,$3F,$AB,$D7,$E8
	.dc.b	$A3,$F8,$D9,$F4,$0F,$4C,$7E,$81,$FC,$EB,$FC,$62,$FE,$84,$7E,$82
	.dc.b	$FF,$8B,$47,$F8,$F6,$7C,$C7,$F5,$03,$E7,$BF,$9E,$AF,$D5,$5F,$E7
	.dc.b	$8E,$EF,$9A,$7F,$C1,$E7,$CE,$BF,$CE,$BF,$D3,$DC,$5F,$E8,$5F,$E8
	.dc.b	$51,$F3,$8F,$CA,$9C,$0F,$9B,$5F,$E0,$F9,$FC,$E5,$FF,$82,$2B,$F3
	.dc.b	$95,$F2,$EB,$FC,$0D,$7F,$2E,$3F,$3B,$3F,$C3,$0F,$EF,$EF,$F0,$78
	.dc.b	$F9,$47,$E6,$6F,$F8,$54,$FF,$38,$9F,$E0,$07,$F5,$43,$F0,$15,$FE
	.dc.b	$9B,$7F,$C6,$E7,$E6,$23,$F8,$39,$F1,$9F,$93,$1F,$DF,$2B,$E4,$D7
	.dc.b	$E5,$EF,$FC,$D9,$1F,$26,$BF,$9D,$1F,$97,$AF,$E7,$77,$FD,$D6,$7F
	.dc.b	$9C,$CF,$F8,$DE,$3E,$49,$FC,$11,$F9,$31,$FC,$0C,$C5,$FC,$8A,$FC
	.dc.b	$99,$F9,$C3,$F8,$19,$F3,$CF,$F3,$C5,$7F,$04,$8F,$C9,$D7,$F8,$5A
	.dc.b	$3C,$D3,$FE,$93,$8F,$95,$7F,$94,$9F,$F1,$5D,$7C,$A8,$FC,$64,$FF
	.dc.b	$81,$8F,$F3,$95,$7F,$8A,$AF,$FB,$44,$7F,$6F,$AF,$F3,$43,$67,$8D
	.dc.b	$EA,$3F,$98,$CF,$FB,$0D,$4F,$E6,$4F,$F1,$23,$FB,$63,$11,$A8,$FF
	.dc.b	$3E,$9F,$DA,$23,$FD,$E8,$78,$2B,$FD,$19,$1F,$52,$B9,$C7,$D0,$9F
	.dc.b	$F0,$23,$FE,$D2,$AE,$91,$FC,$FA,$7A,$47,$F0,$78,$FE,$14,$69,$FE
	.dc.b	$DC,$F9,$9F,$E9,$EF,$F5,$77,$FD,$CB,$5F,$E6,$F8,$E6,$FC,$98,$FF
	.dc.b	$1F,$BF,$6A,$BE,$4C,$7F,$A0,$CF,$E5,$E7,$FB,$0A,$3F,$4D,$5F,$CE
	.dc.b	$CF,$A4,$7B,$9F,$E1,$CF,$C8,$3D,$D5,$FE,$A1,$9F,$41,$FC,$EA,$3F
	.dc.b	$2D,$1F,$EA,$33,$FC,$B2,$7F,$BC,$67,$EE,$47,$E0,$27,$ED,$CF,$B2
	.dc.b	$7F,$CD,$33,$F6,$DF,$C2,$47,$B0,$E6,$FD,$A7,$D8,$7F,$AC,$67,$C6
	.dc.b	$FF,$AC,$E7,$F1,$35,$FC,$FA,$7F,$C7,$2F,$B6,$7E,$A3,$A9,$FF,$0B
	.dc.b	$C7,$F9,$7A,$BE,$F4,$7C,$A8,$F9,$C7,$F8,$C6,$3F,$9C,$BF,$5C,$E3
	.dc.b	$5E,$B7,$FC,$C8,$7D,$63,$F3,$2F,$AC,$FF,$16,$47,$DF,$38,$9F,$EE
	.dc.b	$A3,$F6,$73,$FA,$23,$D5,$1F,$3D,$F1,$4F,$EC,$5F,$AC,$7A,$8F,$C9
	.dc.b	$4F,$DD,$7F,$4D,$1F,$30,$E3,$3F,$E5,$C7,$FC,$AF,$1F,$55,$FA,$D3
	.dc.b	$FA,$E8,$FD,$54,$7D,$88,$E3,$3F,$DE,$A3,$FC,$E4,$7D,$33,$F5,$B1
	.dc.b	$FB,$19,$F9,$D1,$FA,$B7,$8B,$FE,$33,$AF,$EE,$0F,$FA,$38,$FD,$75
	.dc.b	$7D,$C7,$F5,$B3,$FA,$A7,$E9,$99,$3F,$BA,$4F,$D7,$9F,$9F,$1F,$2C
	.dc.b	$FA,$73,$FA,$4A,$FD,$41,$E2,$8F,$E1,$95,$9A,$FA,$F3,$C5,$FA,$F5
	.dc.b	$F4,$CF,$D3,$9C,$6B,$FD,$38,$FA,$8F,$D3,$A7,$E9,$CF,$EB,$1F,$A7
	.dc.b	$1F,$A2,$39,$1F,$D2,$67,$EA,$3F,$94,$3F,$18,$E2,$3F,$4D,$1F,$4C
	.dc.b	$F8,$47,$FA,$16,$BF,$55,$3F,$5C,$CC,$7F,$7A,$3E,$59,$F4,$CE,$31
	.dc.b	$F4,$8F,$49,$E8,$7F,$8D,$4F,$E8,$DF,$9A,$78,$27,$F1,$35,$FA,$47
	.dc.b	$E8,$99,$8F,$D1,$1F,$EF,$C3,$84,$7E,$86,$3E,$79,$89,$F9,$D3,$FD
	.dc.b	$01,$FD,$C4,$7F,$AC,$9F,$9C,$7E,$7C,$FA,$CF,$E9,$5F,$9B,$1F,$D3
	.dc.b	$8F,$82,$7F,$09,$3E,$61,$C0,$FC,$D3,$F3,$1F,$CF,$CF,$F4,$9A,$F9
	.dc.b	$71,$F9,$C8,$FE,$8F,$3F,$9C,$8F,$E8,$87,$CA,$7F,$A2,$9F,$C3,$27
	.dc.b	$E5,$57,$F1,$28,$FC,$D4,$FC,$AA,$FE,$79,$5F,$99,$8F,$E7,$D5,$F2
	.dc.b	$6B,$F3,$11,$F9,$43,$E4,$CF,$E5,$DF,$EB,$E7,$FB,$CE,$BD,$8F,$F6
	.dc.b	$79,$F9,$6F,$F3,$F3,$F0,$EF,$CA,$AF,$9C,$F0,$8F,$CD,$BF,$D8,$DF
	.dc.b	$EB,$07,$F7,$63,$F9,$E9,$F2,$E3,$F3,$47,$CE,$8F,$95,$3F,$DB,$27
	.dc.b	$F2,$2F,$F3,$93,$F3,$87,$77,$FD,$B9,$1F,$2E,$11,$FD,$A4,$F9,$95
	.dc.b	$FC,$DE,$BF,$85,$1F,$3E,$BF,$29,$3F,$D9,$4F,$C8,$4F,$CB,$AF,$CA
	.dc.b	$1F,$DA,$67,$E5,$CF,$E4,$E3,$E4,$D7,$F3,$4A,$F9,$91,$FE,$24,$8F
	.dc.b	$CA,$57,$C9,$AF,$95,$3F,$DA,$6B,$F8,$25,$7C,$A3,$E6,$D7,$C9,$AF
	.dc.b	$E6,$07,$E4,$A7,$F9,$9C,$FC,$8A,$FE,$5C,$7C,$79,$FC,$8B,$FE,$14
	.dc.b	$9F,$C8,$BF,$E1,$69,$F9,$75,$F1,$EB,$E3,$4F,$F8,$56,$BE,$34,$7F
	.dc.b	$2A,$8F,$EB,$73,$F1,$63,$E3,$47,$E4,$23,$F1,$F1,$F9,$08,$F8,$D1
	.dc.b	$FB,$78,$F8,$D1,$FE,$0D,$8F,$8F,$5F,$E0,$C8,$FE,$B2,$FC,$67,$FC
	.dc.b	$13,$3F,$8C,$3E,$1D,$7C,$67,$FC,$21,$3F,$1A,$BF,$6F,$5F,$80,$8F
	.dc.b	$E4,$33,$F1,$E7,$FA,$A4,$E6,$7F,$A8,$4F,$E2,$4F,$F6,$51,$FE,$11
	.dc.b	$3E,$49,$FE,$08,$AF,$92,$FF,$82,$A3,$E4,$D7,$EE,$23,$F7,$35,$F2
	.dc.b	$63,$FA,$8C,$7C,$98,$F0,$3F,$D5,$E3,$F7,$2F,$CA,$8F,$F0,$1D,$7E
	.dc.b	$3E,$3F,$C0,$35,$FB,$9A,$F9,$55,$FE,$6E,$9F,$DC,$C7,$CA,$7F,$C7
	.dc.b	$D1,$F2,$8F,$F1,$A9,$F2,$9F,$C7,$BF,$E3,$4A,$F9,$4F,$F8,$DA,$BE
	.dc.b	$55,$7F,$77,$C3,$E2,$CF,$FB,$B9,$B7,$FA,$CB,$60,$9C,$CA,$7F,$AC
	.dc.b	$46,$30,$FD,$9B,$AC,$38,$BF,$4A,$BF,$73,$1C,$DF,$F3,$96,$1F,$BA
	.dc.b	$70,$7F,$91,$5E,$D8,$62,$34,$7F,$1A,$8F,$14,$E6,$39,$47,$F9,$E6
	.dc.b	$BD,$6F,$F6,$DA,$D4,$FF,$81,$AB,$D2,$FF,$85,$67,$FC,$F6,$E8,$FF
	.dc.b	$7E,$E1,$EA,$74,$FF,$77,$38,$BE,$43,$FD,$16,$FF,$9F,$5E,$51,$E6
	.dc.b	$9E,$51,$FE,$23,$C3,$F9,$4E,$1F,$49,$F6,$B6,$3E,$0B,$6F,$BC,$7C
	.dc.b	$2B,$57,$E8,$E3,$BB,$FE,$0E,$C3,$D5,$1A,$9F,$F0,$76,$1A,$7E,$86
	.dc.b	$1E,$17,$BC,$FC,$3B,$F5,$D6,$DF,$BB,$19,$9F,$99,$5F,$76,$BA,$9F
	.dc.b	$00,$FF,$0E,$4F,$03,$FC,$E0,$F6,$6E,$1E,$C7,$EC,$C7,$C9,$9F,$CA
	.dc.b	$BE,$8C,$3D,$73,$F2,$0F,$F5,$D1,$EA,$2D,$7C,$E7,$E3,$1F,$E5,$8A
	.dc.b	$FC,$B3,$F5,$0F,$8C,$FE,$55,$FF,$61,$D7,$8D,$FE,$9B,$71,$F5,$1D
	.dc.b	$1F,$53,$6B,$83,$F9,$5C,$36,$FE,$50,$EB,$87,$D4,$9F,$8F,$87,$EC
	.dc.b	$9F,$8F,$3F,$88,$7F,$1D,$37,$0F,$CF,$D6,$DF,$C0,$D6,$5F,$80,$FF
	.dc.b	$B0,$CF,$BF,$1F,$8B,$7C,$67,$FB,$76,$B6,$D9,$FC,$4B,$D2,$7F,$25
	.dc.b	$1A,$8F,$F6,$6B,$EE,$C3,$F1,$91,$D0,$FF,$64,$56,$E3,$F2,$0E,$F0
	.dc.b	$FF,$77,$9B,$8F,$CA,$57,$5A,$FC,$4B,$F8,$93,$EE,$3F,$86,$C3,$17
	.dc.b	$E1,$9F,$C3,$4F,$EE,$5F,$C3,$61,$F6,$E3,$89,$FC,$46,$FF,$12,$FC
	.dc.b	$A3,$F0,$18,$7A,$5F,$E2,$4A,$37,$5F,$72,$7F,$04,$7E,$4A,$3F,$0B
	.dc.b	$5F,$89,$7A,$47,$EE,$63,$15,$CD,$D5,$73,$BF,$BF,$1F,$C9,$5F,$D3
	.dc.b	$DF,$E1,$67,$9D,$FF,$0D,$9F,$7B,$FE,$AB,$7F,$D7,$35,$C6,$7E,$FC
	.dc.b	$FF,$60,$BF,$E1,$6D,$3C,$E5,$AF,$64,$7F,$09,$AF,$E6,$CF,$F6,$DB
	.dc.b	$FE,$12,$FE,$BE,$BF,$8C,$47,$CC,$7F,$A9,$9E,$D3,$C0,$FA,$1F,$EA
	.dc.b	$F1,$F3,$4F,$C0,$C7,$43,$CF,$1F,$E6,$CB,$E8,$FF,$43,$7E,$1D,$7F
	.dc.b	$0D,$7A,$46,$5F,$CB,$3F,$DF,$9F,$86,$D7,$AB,$FB,$D8,$FC,$23,$E7
	.dc.b	$3E,$43,$F8,$7A,$FC,$14,$7C,$18,$FC,$54,$FE,$06,$3F,$CD,$27,$E0
	.dc.b	$4E,$47,$F4,$47,$F0,$93,$CA,$BF,$98,$57,$E0,$27,$EE,$57,$E3,$EF
	.dc.b	$DA,$FF,$85,$9F,$E9,$91,$F8,$1B,$D3,$CA,$74,$FB,$1F,$DF,$1F,$34
	.dc.b	$FB,$A7,$F4,$D3,$E3,$9F,$D1,$63,$47,$A1,$FE,$0D,$1F,$82,$BF,$A3
	.dc.b	$1F,$81,$34,$7F,$B8,$23,$EC,$4F,$DB,$9F,$24,$FE,$12,$7F,$4C,$7E
	.dc.b	$71,$FA,$47,$23,$FD,$0F,$5A,$8F,$B3,$1F,$81,$AD,$3F,$C0,$AF,$E8
	.dc.b	$CF,$F3,$78,$F2,$DF,$F8,$1E,$FF,$87,$47,$F3,$D3,$FA,$1D,$FF,$0D
	.dc.b	$7F,$9E,$9F,$C0,$6B,$F8,$C3,$F6,$AF,$FC,$0B,$5C,$AF,$E3,$9F,$E0
	.dc.b	$99,$FF,$0A,$D7,$F1,$59,$FE,$6F,$1F,$7C,$FE,$34,$FF,$36,$AF,$E6
	.dc.b	$8F,$F2,$53,$F6,$F1,$FD,$DD,$FE,$2A,$FE,$CE,$7F,$92,$3E,$CB,$FF
	.dc.b	$18,$61,$FC,$CA,$B3,$1F,$C9,$CF,$5C,$FC,$FA,$FE,$F4,$E7,$0F,$EA
	.dc.b	$6F,$F8,$8A,$7F,$BC,$E1,$FD,$F5,$D5,$7F,$2F,$3D,$73,$FD,$D0,$FE
	.dc.b	$68,$FA,$DF,$E5,$8F,$C4,$3D,$6F,$F7,$67,$FC,$02,$FF,$69,$7E,$6B
	.dc.b	$F5,$8F,$5C,$FF,$92,$2F,$35,$F5,$67,$89,$FD,$FA,$BF,$BA,$5F,$13
	.dc.b	$D7,$1F,$DC,$5F,$AA,$FF,$4A,$37,$1F,$54,$FD,$23,$FD,$E2,$BF,$24
	.dc.b	$FF,$8D,$67,$FC,$8F,$1F,$D5,$4F,$D7,$DF,$F6,$BB,$FB,$77,$E4,$BF
	.dc.b	$F1,$E4,$7F,$21,$7F,$C6,$0F,$F2,$C7,$A3,$EE,$7F,$C5,$07,$E2,$9F
	.dc.b	$E5,$8E,$23,$FA,$F4,$7F,$83,$9F,$B3,$1F,$E3,$3A,$FE,$D1,$3F,$62
	.dc.b	$3F,$C5,$CF,$D8,$C3,$FC,$77,$5F,$62,$7F,$C8,$4F,$B9,$AF,$C8,$7B
	.dc.b	$57,$F6,$78,$FD,$7C,$69,$FE,$E0,$FA,$CF,$EC,$D7,$FD,$A6,$7F,$90
	.dc.b	$47,$03,$FA,$0C,$7E,$D0,$FE,$C0,$FF,$1F,$3F,$72,$7A,$4F,$EC,$47
	.dc.b	$F7,$1B,$CE,$1F,$5A,$3F,$D1,$D7,$FC,$41,$FF,$3D,$9F,$C0,$AF,$EA
	.dc.b	$BF,$C8,$2B,$F9,$19,$99,$E5,$1F,$E2,$23,$F9,$11,$FC,$2A,$3F,$82
	.dc.b	$3F,$C8,$DF,$E4,$35,$FD,$72,$38,$C7,$F0,$37,$FA,$C9,$C6,$3F,$94
	.dc.b	$BF,$E6,$98,$FE,$40,$66,$7F,$8D,$47,$F8,$25,$FE,$76,$FF,$5B,$AF
	.dc.b	$E3,$D3,$99,$FE,$4F,$3F,$E8,$77,$ED,$3F,$EA,$5B,$FE,$AB,$7F,$E1
	.dc.b	$6B,$FA,$37,$FD,$52,$B8,$9F,$D9,$EB,$E8,$BF,$D4,$EB,$FC,$29,$5F
	.dc.b	$D5,$DF,$B1,$1F,$98,$7F,$C3,$17,$FE,$11,$3E,$84,$FA,$27,$FC,$1D
	.dc.b	$1F,$02,$3F,$7F,$1F,$CB,$E3,$C2,$FF,$AA,$5F,$94,$FD,$09,$F9,$93
	.dc.b	$FE,$0C,$8F,$09,$FD,$46,$3E,$84,$FF,$4F,$8F,$E9,$47,$CF,$9F,$EA
	.dc.b	$57,$F3,$AB,$FC,$0D,$5F,$3A,$BF,$42,$FF,$5E,$7E,$53,$F5,$8F,$3A
	.dc.b	$3F,$D3,$F3,$FD,$32,$3E,$7C,$70,$BF,$F6,$DB,$FC,$4E,$FF,$A5,$BF
	.dc.b	$C3,$8F,$4D,$FF,$84,$F0,$F9,$37,$FC,$5A,$7F,$A7,$4F,$EB,$CF,$E2
	.dc.b	$11,$FC,$6B,$0C,$BF,$C5,$4F,$E8,$87,$F5,$AB,$FE,$5D,$5F,$C5,$5F
	.dc.b	$CF,$B6,$3C,$31,$FD,$22,$3F,$BE,$47,$F4,$87,$FA,$21,$E1,$9F,$CF
	.dc.b	$BF,$C5,$8F,$4B,$FD,$21,$FE,$37,$5F,$B1,$3F,$CF,$2F,$F1,$1B,$FE
	.dc.b	$25,$5F,$EB,$F8,$FF,$63,$3F,$D1,$CE,$07,$F0,$F7,$FA,$15,$FF,$9F
	.dc.b	$E3,$D3,$1F,$C4,$AF,$FA,$5C,$7F,$2B,$8F,$CC,$1F,$EB,$83,$E8,$57
	.dc.b	$F7,$DA,$FE,$8B,$1F,$DF,$E7,$F8,$AD,$F9,$E3,$F6,$C7,$F3,$B3,$FD
	.dc.b	$DF,$5F,$D2,$2B,$FB,$EC,$FD,$89,$FF,$57,$E1,$FE,$65,$BF,$A3,$1F
	.dc.b	$CF,$AB,$FA,$14,$7F,$80,$1F,$C4,$D7,$FB,$DE,$FF,$2F,$3F,$EA,$4A
	.dc.b	$F9,$F5,$FC,$D9,$FE,$7C,$7C,$E3,$DC,$7F,$A2,$AB,$E7,$61,$E8,$8F
	.dc.b	$D9,$4F,$F3,$E9,$FE,$8C,$7D,$63,$8C,$7E,$66,$BF,$2E,$FF,$47,$3E
	.dc.b	$5C,$FC,$F8,$FE,$E8,$7E,$58,$FC,$34,$7F,$97,$CF,$F5,$14,$FF,$0B
	.dc.b	$8F,$F5,$0D,$7F,$0B,$3F,$9B,$CF,$F2,$A3,$CC,$7C,$D9,$FE,$17,$3F
	.dc.b	$E5,$C8,$FF,$2E,$C7,$E8,$B0,$FF,$48,$47,$F9,$58,$FE,$69,$7F,$C3
	.dc.b	$6F,$F9,$BD,$7F,$A6,$EF,$F8,$25,$7F,$07,$BF,$F4,$7B,$FC,$BA,$BF
	.dc.b	$B3,$BF,$CD,$EB,$E8,$CF,$FA,$51,$FE,$8D,$3F,$EB,$59,$FE,$11,$5F
	.dc.b	$E5,$47,$FC,$AB,$3F,$C2,$23,$F8,$3C,$7F,$31,$3F,$81,$3F,$C1,$CF
	.dc.b	$E2,$AF,$F0,$68,$FE,$1B,$1F,$D8,$23,$F9,$81,$F1,$AF,$F9,$9B,$F2
	.dc.b	$A7,$FD,$8B,$3F,$E8,$38,$FE,$65,$7F,$EC,$3A,$FE,$65,$7F,$E8,$33
	.dc.b	$FB,$0D,$7F,$31,$AF,$91,$7F,$EB,$FA,$F9,$11,$FE,$81,$8F,$91,$3F
	.dc.b	$CB,$A3,$F9,$7C,$7F,$2E,$8F,$EB,$55,$FD,$72,$FE,$5C,$7F,$2D,$AF
	.dc.b	$F5,$71,$F1,$A3,$E5,$4F,$F2,$D7,$E6,$4F,$F2,$D7,$FD,$5D,$3F,$1A
	.dc.b	$FF,$AC,$BF,$CC,$2B,$E4,$D7,$F8,$10,$FF,$1F,$C7,$EE,$AB,$FC,$7D
	.dc.b	$1F,$E7,$79,$FF,$1F,$3F,$BA,$3F,$6B,$3F,$CA,$2B,$FC,$E0,$7E,$E2
	.dc.b	$7F,$93,$E1,$F0,$2F,$F9,$D5,$7F,$23,$3F,$71,$08,$FF,$37,$DD,$9F
	.dc.b	$F5,$95,$7F,$8E,$A3,$FC,$5F,$1F,$B2,$3F,$CD,$C7,$F2,$73,$F7,$31
	.dc.b	$FC,$AA,$3F,$C6,$B8,$7E,$E0,$FF,$3A,$1F,$CB,$4F,$DB,$C7,$F8,$CE
	.dc.b	$BF,$DA,$93,$FE,$2D,$AF,$DA,$DF,$8D,$FF,$28,$47,$F2,$47,$FC,$D1
	.dc.b	$5F,$C9,$23,$FC,$63,$1F,$C7,$4F,$F3,$41,$FD,$2C,$FE,$76,$7C,$2B
	.dc.b	$F8,$73,$F3,$2B,$F6,$53,$F1,$8D,$1F,$BA,$3F,$CD,$18,$7C,$EB,$FE
	.dc.b	$A9,$1F,$E6,$89,$F9,$CF,$FA,$6C,$F4,$1F,$E9,$78,$FE,$B9,$3E,$98
	.dc.b	$FE,$BB,$5F,$36,$BF,$6B,$1F,$E6,$03,$F9,$2B,$FE,$99,$AE,$71,$FE
	.dc.b	$93,$3F,$93,$3F,$38,$1F,$E6,$2C,$3D,$51,$FE,$99,$C3,$E8,$61,$A9
	.dc.b	$FF,$4D,$4F,$F5,$4B,$FD,$BD,$F6,$8F,$F0,$05,$F4,$8F,$F6,$39,$FD
	.dc.b	$BA,$3F,$CC,$B7,$FE,$63,$C3,$E9,$5F,$19,$F0,$94,$F9,$97,$FC,$A7
	.dc.b	$0C,$C7,$CF,$AF,$E5,$53,$FC,$0A,$7F,$C6,$67,$F9,$91,$FE,$52,$F0
	.dc.b	$9F,$F6,$61,$CA,$3F,$C5,$6F,$FA,$AE,$3F,$95,$BE,$1B,$FF,$1A,$3F
	.dc.b	$75,$FA,$4F,$F9,$A2,$BF,$96,$CF,$F5,$DC,$3F,$83,$3F,$CB,$27,$FC
	.dc.b	$C4,$7F,$2D,$3F,$8F,$CF,$F8,$94,$CC,$FF,$07,$9F,$E5,$D3,$FC,$1C
	.dc.b	$F7,$CF,$FA,$BE,$3F,$84,$DF,$F2,$D7,$F8,$3C,$FF,$2E,$8F,$F1,$AD
	.dc.b	$FF,$08,$9F,$81,$7F,$7A,$FF,$84,$BF,$36,$FF,$D2,$07,$C3,$9F,$F4
	.dc.b	$84,$7F,$56,$8F,$E5,$D1,$FE,$8F,$7E,$55,$FF,$A1,$DF,$E8,$4F,$CE
	.dc.b	$7F,$D1,$C7,$F5,$68,$FA,$0F,$FA,$BD,$F9,$A7,$F5,$88,$FE,$61,$3F
	.dc.b	$3C,$FE,$87,$7E,$78,$FD,$2D,$7E,$12,$FC,$EF,$EC,$8F,$F2,$91,$F8
	.dc.b	$33,$E7,$3F,$CC,$DF,$F7,$F5,$7F,$32,$7F,$99,$CF,$E3,$67,$FD,$4D
	.dc.b	$1F,$E3,$47,$F9,$A3,$F4,$23,$FD,$3B,$3F,$E2,$F8,$FE,$46,$FE,$11
	.dc.b	$AF,$F5,$B9,$FF,$47,$1F,$3A,$FE,$44,$FF,$89,$CF,$E1,$B7,$FE,$5E
	.dc.b	$3F,$C5,$F1,$F9,$38,$FF,$23,$1F,$C6,$2B,$FC,$BC,$FF,$8B,$70,$FF
	.dc.b	$17,$47,$F0,$A9,$E1,$87,$F8,$62,$97,$FE,$18,$8F,$F5,$DC,$7E,$B5
	.dc.b	$FF,$78,$DD,$9F,$CD,$C7,$0B,$FF,$2F,$9C,$F0,$FF,$78,$5F,$F0,$93
	.dc.b	$F8,$6E,$1F,$A9,$8F,$E4,$B1,$FC,$01,$FF,$04,$BF,$E6,$19,$FE,$1C
	.dc.b	$7F,$0D,$7F,$7B,$49,$FF,$24,$D7,$38,$FE,$7B,$3F,$C3,$63,$F9,$ED
	.dc.b	$7A,$4F,$A5,$1F,$EE,$CC,$31,$1F,$E3,$2B,$FE,$17,$5F,$42,$BF,$84
	.dc.b	$9F,$88,$3F,$C3,$07,$F3,$D3,$FD,$1D,$1F,$CE,$E3,$F2,$F7,$FE,$5F
	.dc.b	$3F,$B0,$BF,$3B,$0F,$E7,$4F,$D0,$9F,$EC,$EF,$F0,$13,$F8,$64,$79
	.dc.b	$A7,$E7,$C7,$F6,$63,$37,$F3,$8F,$EC,$A7,$FA,$2E,$FE,$B5,$FF,$AA
	.dc.b	$6F,$FC,$BC,$7F,$31,$8F,$93,$1F,$39,$FE,$79,$3F,$3A,$BE,$54,$7F
	.dc.b	$AC,$5F,$99,$5F,$E5,$F3,$94,$7F,$3E,$BF,$F7,$63,$F3,$CF,$E5,$B5
	.dc.b	$F8,$93,$D4,$7F,$39,$7F,$D1,$EF,$D0,$BF,$9F,$1F,$BE,$3F,$C6,$F3
	.dc.b	$FC,$DD,$FE,$03,$1F,$CD,$AF,$EA,$47,$E8,$DF,$E6,$C7,$F8,$DE,$BF
	.dc.b	$CB,$CF,$D2,$9F,$F1,$BC,$7F,$05,$9F,$F1,$BC,$FF,$3B,$BF,$E6,$71
	.dc.b	$F9,$9A,$FE,$66,$FF,$46,$8F,$F3,$CB,$F8,$F3,$FC,$B2,$7F,$08,$3F
	.dc.b	$85,$3F,$EE,$78,$FE,$11,$1F,$EB,$EA,$FE,$11,$5F,$CD,$2F,$F8,$41
	.dc.b	$FE,$67,$3F,$9D,$9F,$D2,$A7,$F3,$8F,$F0,$89,$FC,$4C,$7C,$C7,$C0
	.dc.b	$FF,$A3,$5F,$E1,$0F,$F8,$F6,$3F,$B5,$CF,$CB,$8F,$96,$FF,$34,$3E
	.dc.b	$75,$7C,$B8,$FE,$5B,$5F,$71,$FD,$69,$F8,$13,$F9,$B8,$AF,$E6,$CF
	.dc.b	$EE,$4F,$E6,$C7,$F3,$C8,$FF,$32,$3F,$3E,$3F,$9A,$D7,$CF,$9F,$E6
	.dc.b	$A7,$BE,$FE,$6D,$FE,$B6,$FE,$71,$FD,$82,$3E,$63,$FE,$66,$BF,$F0
	.dc.b	$BB,$FC,$ED,$FF,$45,$9F,$11,$F9,$31,$EB,$7F,$BF,$47,$CE,$9F,$55
	.dc.b	$FF,$9A,$2F,$FD,$1B,$5F,$D8,$63,$FA,$FC,$7C,$C6,$CF,$F3,$39,$F8
	.dc.b	$F3,$FC,$8D,$FE,$09,$5F,$E5,$EB,$FE,$69,$3F,$E6,$99,$FF,$71,$61
	.dc.b	$FE,$3E,$7F,$D6,$87,$F9,$A5,$FE,$67,$87,$F0,$38,$FF,$1E,$3F,$C0
	.dc.b	$DF,$F1,$E3,$FC,$02,$BF,$DB,$E7,$B6,$3F,$C7,$67,$F3,$3A,$FE,$60
	.dc.b	$FF,$99,$CF,$E5,$E7,$F0,$03,$FC,$CE,$7F,$06,$3F,$9B,$5F,$F9,$9D
	.dc.b	$FC,$01,$FB,$DC,$3F,$96,$BF,$C2,$23,$F9,$6D,$FF,$2F,$A6,$1F,$90
	.dc.b	$AF,$E6,$B1,$FC,$00,$FE,$97,$1F,$CB,$B0,$FF,$1C,$4F,$F3,$09,$FF
	.dc.b	$34,$9F,$E9,$27,$E6,$1F,$CB,$63,$FA,$EB,$F9,$43,$FA,$B1,$FE,$03
	.dc.b	$3F,$2F,$7F,$A6,$7F,$DC,$11,$FC,$B9,$CE,$1F,$CB,$9F,$DE,$57,$F9
	.dc.b	$37,$0F,$E5,$71,$FE,$94,$7F,$97,$3F,$E9,$23,$E6,$47,$F2,$E3,$E4
	.dc.b	$3F,$30,$F8,$C7,$CC,$8F,$D2,$47,$F2,$B8,$F9,$07,$CC,$3F,$CA,$73
	.dc.b	$F3,$1F,$9D,$7F,$31,$F9,$D3,$F3,$2B,$FA,$C9,$FE,$4F,$7E,$65,$E6
	.dc.b	$7E,$7C,$7F,$59,$7E,$66,$1F,$E3,$57,$FC,$5C,$7E,$94,$FE,$DE,$7F
	.dc.b	$87,$EB,$F8,$1D,$FF,$98,$63,$F7,$E7,$FB,$88,$FF,$1B,$DF,$FB,$89
	.dc.b	$FD,$F9,$E3,$BF,$E5,$8F,$EF,$EB,$F7,$D3,$FE,$E2,$7F,$80,$61,$FB
	.dc.b	$09,$FA,$35,$F9,$97,$FC,$B4,$7F,$2C,$3F,$80,$3F,$EE,$47,$F8,$04
	.dc.b	$FF,$B9,$23,$F9,$5C,$7F,$8B,$4F,$F2,$F1,$FC,$0E,$7F,$C6,$2F,$F0
	.dc.b	$58,$FF,$72,$3F,$CA,$4F,$F7,$23,$FE,$E3,$7F,$94,$4F,$CE,$7F,$94
	.dc.b	$3F,$CA,$6F,$FA,$94,$7F,$A0,$6B,$F9,$43,$FC,$1A,$7F,$D0,$EF,$F2
	.dc.b	$69,$FE,$4A,$FE,$BE,$3F,$A9,$47,$F0,$18,$FF,$52,$DF,$F8,$49,$FF
	.dc.b	$14,$1F,$14,$FF,$4C,$BF,$D5,$0F,$F3,$83,$FE,$5D,$AD,$3C,$AF,$F3
	.dc.b	$4F,$F9,$F8,$F9,$55,$FE,$55,$7F,$CB,$67,$CA,$3F,$CB,$77,$FE,$86
	.dc.b	$7F,$CB,$6F,$B6,$3F,$AA,$57,$CA,$9F,$F5,$DB,$FA,$F7,$FD,$5B,$5E
	.dc.b	$93,$FC,$B4,$FF,$53,$BF,$F4,$43,$FD,$5E,$7F,$94,$3F,$1C,$FD,$D1
	.dc.b	$FA,$33,$FD,$BE,$7E,$FE,$7F,$94,$5F,$F9,$6E,$3F,$7F,$7F,$E2,$B7
	.dc.b	$F9,$44,$7F,$8A,$E7,$FC,$B5,$1F,$C0,$9F,$D2,$3F,$29,$F8,$AF,$F8
	.dc.b	$5E,$FF,$C5,$87,$F8,$4D,$FF,$67,$BF,$E5,$A3,$E7,$CF,$F8,$BA,$7F
	.dc.b	$80,$3F,$D7,$E7,$F8,$01,$FB,$A3,$F8,$01,$FC,$B6,$3F,$80,$3F,$BC
	.dc.b	$3D,$C7,$F9,$76,$BF,$DC,$71,$FE,$36,$7F,$C1,$93,$FC,$AF,$0F,$E5
	.dc.b	$55,$FE,$79,$38,$9F,$E5,$78,$FD,$85,$FF,$2A,$7F,$65,$7F,$CA,$8F
	.dc.b	$E5,$33,$C2,$FF,$AA,$5F,$F2,$CB,$FE,$57,$3F,$E8,$BA,$F8,$EF,$F0
	.dc.b	$AB,$FF,$41,$BF,$38,$FE,$5B,$3F,$B7,$BF,$F7,$0C,$7F,$99,$1F,$F4
	.dc.b	$44,$7F,$A1,$CF,$9B,$5F,$D0,$2B,$F9,$5D,$7D,$18,$FF,$2C,$5F,$F7
	.dc.b	$A8,$F9,$47,$CC,$07,$FA,$0C,$F9,$A7,$F2,$E8,$F9,$AF,$F2,$AB,$FE
	.dc.b	$80,$DB,$FF,$2B,$4F,$CD,$C3,$E1,$1F,$E1,$97,$DF,$3D,$8F,$F1,$53
	.dc.b	$ED,$3F,$96,$DF,$F5,$77,$FC,$AD,$1F,$CB,$30,$FF,$16,$3F,$CB,$4F
	.dc.b	$D9,$9F,$8F,$7F,$C5,$AF,$FB,$80,$FF,$03,$BF,$C1,$9F,$F5,$7B,$FE
	.dc.b	$57,$3F,$CA,$CF,$F2,$C3,$E9,$1F,$D5,$8F,$F2,$B9,$FD,$62,$FF,$96
	.dc.b	$3F,$D6,$AF,$C2,$FF,$84,$6B,$F3,$EF,$EE,$0F,$E0,$75,$FC,$65,$FE
	.dc.b	$06,$D9,$E3,$1F,$CC,$A7,$F9,$A3,$FE,$52,$3F,$22,$5C,$3F,$CB,$37
	.dc.b	$AA,$FF,$22,$1F,$C1,$E7,$FC,$5F,$5F,$CC,$1F,$E0,$EF,$F8,$D4,$FF
	.dc.b	$37,$9F,$AE,$7F,$83,$57,$E8,$1F,$F7,$FD,$FF,$31,$3F,$CB,$0F,$FA
	.dc.b	$08,$F9,$87,$F0,$B3,$F8,$33,$FE,$82,$7F,$AB,$CF,$FA,$09,$FB,$C7
	.dc.b	$F6,$E8,$FF,$5A,$C7,$F9,$59,$FF,$2A,$47,$C2,$3F,$DB,$57,$F3,$9F
	.dc.b	$E1,$57,$98,$FF,$2B,$C7,$CC,$3F,$98,$3F,$E5,$69,$F9,$97,$F9,$D9
	.dc.b	$FF,$28,$9F,$2A,$BF,$CD,$4F,$F5,$83,$F9,$6C,$7F,$95,$4F,$F4,$2D
	.dc.b	$FC,$98,$F4,$87,$F9,$7B,$5F,$F2,$B3,$F2,$AB,$FA,$D3,$F1,$4F,$F6
	.dc.b	$C9,$FA,$63,$F9,$7D,$FF,$95,$A3,$F9,$75,$FF,$B6,$67,$FC,$F5,$5F
	.dc.b	$E5,$DB,$E1,$1F,$B4,$7F,$93,$E1,$FE,$58,$8F,$F1,$69,$FE,$58,$3F
	.dc.b	$BE,$CF,$F9,$4D,$FF,$16,$1F,$EF,$8C,$3F,$97,$47,$F0,$2C,$3F,$BE
	.dc.b	$1F,$CB,$67,$FC,$A9,$3F,$ED,$89,$FF,$2B,$61,$FB,$A9,$FF,$2B,$C7
	.dc.b	$F1,$38,$FF,$79,$D7,$FB,$5A,$FF,$96,$5E,$70,$FE,$5B,$87,$F8,$B2
	.dc.b	$7F,$1C,$FF,$01,$8F,$94,$FF,$27,$C3,$FA,$23,$FC,$1D,$FF,$16,$5F
	.dc.b	$F2,$EA,$FF,$13,$4F,$FB,$84,$FF,$40,$9F,$EE,$F7,$F9,$74,$FF,$5A
	.dc.b	$8F,$F2,$C3,$FC,$41,$FC,$9D,$7F,$5B,$7E,$75,$FF,$95,$DE,$91,$E8
	.dc.b	$6C,$F1,$7F,$DF,$8F,$EF,$9F,$54,$FF,$30,$3F,$D0,$B1,$FE,$83,$C3
	.dc.b	$FA,$14,$7E,$1E,$FF,$29,$5F,$CB,$4F,$E6,$07,$F2,$F8,$F9,$8F,$84
	.dc.b	$FE,$BD,$5F,$CC,$1F,$91,$5F,$41,$FF,$42,$4F,$F3,$0C,$3F,$2F,$1F
	.dc.b	$D7,$E7,$F9,$8E,$1F,$99,$C3,$E8,$CF,$E2,$F0,$FE,$5F,$7F,$D7,$1F
	.dc.b	$E5,$EF,$CF,$3F,$97,$9F,$C0,$E3,$A9,$FC,$7D,$FE,$5E,$7F,$96,$8C
	.dc.b	$BF,$CF,$63,$F8,$DB,$E7,$3F,$98,$9F,$C1,$A3,$F9,$89,$FE,$36,$3F
	.dc.b	$82,$C7,$F9,$71,$FE,$0B,$5F,$ED,$A9,$FE,$60,$7F,$8C,$D4,$7F,$AD
	.dc.b	$E3,$E0,$DF,$FB,$6C,$FC,$74,$7F,$31,$C3,$F8,$34,$FF,$8D,$E7,$F8
	.dc.b	$81,$FE,$FB,$9F,$F1,$A1,$FE,$33,$8F,$E0,$C7,$FB,$6A,$3F,$97,$3F
	.dc.b	$D6,$5F,$F1,$A3,$FC,$1A,$FF,$C6,$87,$F2,$F3,$F8,$2B,$FC,$CA,$3F
	.dc.b	$C6,$6F,$F2,$E3,$E3,$9F,$E8,$51,$5F,$21,$FC,$11,$FE,$EF,$AD,$BF
	.dc.b	$91,$3F,$86,$BF,$CC,$5F,$AB,$7C,$AF,$F9,$94,$62,$FF,$97,$3E,$83
	.dc.b	$F9,$AD,$FF,$98,$CF,$E3,$A7,$F9,$86,$7C,$31,$FD,$76,$7F,$99,$47
	.dc.b	$86,$3F,$D1,$27,$F2,$D3,$F8,$B5,$FC,$F8,$FF,$45,$3F,$CD,$0F,$E1
	.dc.b	$F1,$FE,$89,$3F,$CC,$27,$E3,$25,$7F,$E6,$28,$FB,$A6,$54,$7D,$87
	.dc.b	$C5,$1F,$C2,$8D,$9F,$D8,$0F,$F4,$7C,$FE,$52,$BF,$B0,$47,$D3,$9F
	.dc.b	$D1,$BF,$D5,$A3,$F9,$F6,$3F,$E6,$18,$FF,$11,$BF,$E6,$3C,$3F,$DD
	.dc.b	$ED,$93,$FC,$EA,$69,$FC,$43,$1A,$FF,$AD,$E3,$FC,$7A,$E0,$FF,$1D
	.dc.b	$57,$47,$FC,$7F,$59,$7F,$93,$1F,$DF,$8C,$1F,$D0,$CF,$F1,$65,$61
	.dc.b	$FE,$2B,$1E,$A7,$FA,$03,$FE,$2F,$AF,$C2,$D7,$F5,$B3,$C1,$3F,$E4
	.dc.b	$57,$83,$FE,$47,$C7,$E0,$BC,$AB,$6F,$F8,$75,$DB,$FE,$4A,$3C,$58
	.dc.b	$7D,$67,$F3,$78,$62,$BF,$94,$E1,$F9,$39,$D3,$FE,$F8,$9F,$54,$7D
	.dc.b	$EA,$FE,$DD,$3F,$99,$7D,$51,$FC,$82,$7A,$47,$E7,$E7,$55,$E8,$AF
	.dc.b	$F7,$ED,$FA,$1F,$BC,$F4,$7F,$DF,$26,$CF,$E8,$AF,$C5,$BF,$5B,$F1
	.dc.b	$CA,$7A,$23,$FE,$00,$BD,$5F,$E0,$4F,$C2,$7F,$C4,$35,$F8,$99,$C4
	.dc.b	$79,$CF,$E6,$F7,$FC,$2D,$CD,$73,$7F,$B3,$3F,$39,$FC,$1C,$7C,$27
	.dc.b	$FA,$BE,$1B,$8F,$F5,$5C,$F8,$5F,$34,$FF,$5A,$AF,$9B,$5F,$E1,$9A
	.dc.b	$FA,$EF,$F8,$76,$7E,$11,$FD,$72,$3E,$CD,$FE,$21,$F4,$47,$E1,$4E
	.dc.b	$2F,$F4,$D8,$FF,$2F,$3E,$78,$FE,$BC,$7E,$04,$E6,$7C,$0C,$3F,$06
	.dc.b	$D3,$CE,$7F,$60,$BF,$C1,$C7,$F8,$0E,$BF,$05,$3E,$63,$FD,$79,$7F
	.dc.b	$06,$7F,$84,$BE,$18,$FF,$32,$E1,$D1,$FE,$D2,$F9,$DF,$EB,$78,$7C
	.dc.b	$3B,$FE,$52,$FC,$38,$FB,$F5,$F3,$63,$F0,$53,$A7,$A4,$FF,$6B,$7A
	.dc.b	$3F,$C3,$DF,$BE,$F9,$CF,$F2,$89,$F0,$8F,$F0,$D1,$F3,$0F,$F2,$3D
	.dc.b	$EE,$FF,$B5,$61,$FA,$49,$F8,$47,$E2,$0E,$77,$FD,$01,$F8,$31,$F8
	.dc.b	$1C,$3E,$15,$7F,$AE,$EB,$4F,$9A,$BF,$CB,$B3,$F0,$5F,$E0,$E7,$85
	.dc.b	$FD,$14,$FE,$05,$E6,$7F,$98,$EF,$58,$73,$AF,$F3,$FC,$7F,$B2,$A3
	.dc.b	$C2,$FE,$8A,$3E,$F4,$FD,$77,$E0,$1F,$E1,$C3,$E5,$9F,$C7,$E7,$F4
	.dc.b	$4F,$F0,$37,$9D,$FE,$86,$3F,$DA,$18,$7F,$02,$AC,$DF,$C9,$3F,$92
	.dc.b	$E1,$F4,$F0,$F8,$67,$DD,$7F,$DF,$75,$F7,$A3,$F7,$B3,$F4,$5F,$F1
	.dc.b	$C4,$FF,$70,$7E,$5D,$FA,$CF,$B3,$7F,$B1,$7F,$36,$7D,$CA,$FB,$4F
	.dc.b	$C3,$BF,$B5,$87,$F1,$18,$FF,$27,$9A,$7F,$08,$FF,$44,$BF,$DD,$56
	.dc.b	$5F,$87,$7F,$5E,$7F,$DF,$87,$D2,$7F,$81,$1F,$96,$7F,$A0,$CF,$D6
	.dc.b	$9F,$96,$FD,$57,$F6,$91,$F1,$CF,$09,$F4,$E7,$EC,$BF,$52,$BF,$9E
	.dc.b	$4F,$27,$D5,$1F,$E6,$CA,$FC,$9B,$FC,$D7,$0F,$E7,$17,$F5,$23,$F9
	.dc.b	$BD,$FD,$C3,$D4,$FF,$30,$7E,$E3,$FC,$81,$F1,$BF,$AD,$8F,$B3,$1E
	.dc.b	$33,$EA,$C7,$F8,$A9,$FA,$AF,$23,$FB,$31,$C8,$FD,$8C,$6A,$BE,$AC
	.dc.b	$7E,$64,$F1,$57,$F7,$63,$F4,$F3,$F7,$6F,$F2,$F3,$FD,$C5,$F8,$8F
	.dc.b	$E3,$8F,$F5,$8C,$FF,$27,$7E,$A6,$1F,$C9,$AB,$F5,$38,$7C,$B7,$D3
	.dc.b	$3F,$E5,$83,$E5,$9F,$A3,$3E,$A3,$FA,$73,$E9,$E1,$F5,$E7,$E9,$1F
	.dc.b	$A4,$9F,$D0,$61,$EF,$3F,$B5,$57,$D1,$C3,$FB,$55,$7D,$13,$BD,$7F
	.dc.b	$19,$3F,$40,$F1,$3F,$14,$FF,$B8,$AB,$FD,$C1,$7A,$9F,$DA,$CF,$E2
	.dc.b	$EB,$FC,$50,$FF,$5E,$BF,$D2,$3F,$20,$FE,$43,$5F,$42,$B8,$DF,$8A
	.dc.b	$7F,$DE,$AF,$D9,$6D,$7D,$13,$F6,$31,$F4,$0F,$B1,$1F,$E2,$09,$FE
	.dc.b	$A6,$FD,$0B,$FE,$31,$1F,$D5,$AF,$F4,$58,$7F,$4C,$AF,$05,$FF,$85
	.dc.b	$DF,$9F,$1F,$D5,$F0,$F9,$F1,$FD,$66,$FE,$79,$FE,$10,$7E,$7C,$7F
	.dc.b	$51,$AF,$71,$FE,$19,$BF,$A7,$5F,$3C,$FC,$CD,$7C,$EA,$FE,$99,$1C
	.dc.b	$0F,$E5,$91,$FE,$0E,$7E,$6B,$E6,$3F,$88,$1F,$ED,$A3,$FD,$C1,$87
	.dc.b	$F9,$4C,$FF,$29,$1F,$D0,$DF,$8A,$7F,$55,$7F,$38,$FF,$49,$BF,$C3
	.dc.b	$3F,$DA,$27,$F8,$7B,$FE,$09,$7F,$87,$9F,$1A,$BF,$94,$47,$F9,$31
	.dc.b	$FE,$1F,$1F,$01,$F4,$D7,$F8,$1E,$3F,$87,$CF,$CE,$BF,$4B,$FC,$37
	.dc.b	$0F,$9D,$3F,$E0,$59,$FC,$E4,$7F,$80,$30,$E0,$7E,$0C,$FC,$DB,$FC
	.dc.b	$FE,$7F,$C8,$33,$F3,$A3,$FC,$93,$1F,$C3,$2B,$F8,$51,$F3,$E3,$F9
	.dc.b	$F5,$FF,$79,$AF,$CD,$47,$F6,$D9,$F9,$C7,$73,$E6,$9F,$DB,$CF,$BA
	.dc.b	$0F,$E0,$27,$CE,$3F,$BA,$9F,$CF,$0F,$EE,$4F,$F7,$3A,$F9,$B5,$FD
	.dc.b	$C9,$F9,$A7,$F0,$49,$FC,$C1,$F3,$5F,$CC,$3F,$E2,$E7,$E5,$9F,$CD
	.dc.b	$1F,$CB,$C7,$F9,$AF,$0F,$CB,$BF,$E6,$B3,$FD,$E7,$7E,$49,$FE,$6D
	.dc.b	$87,$CB,$8F,$F0,$3C,$FF,$36,$BF,$97,$5F,$CD,$B0,$F9,$71,$FA,$63
	.dc.b	$F0,$31,$FC,$0D,$FF,$02,$BF,$E5,$A7,$F8,$3B,$FC,$EE,$BE,$6D,$7F
	.dc.b	$B6,$23,$FC,$44,$7C,$93,$E5,$5F,$D1,$C3,$FC,$71,$1F,$E6,$39,$FF
	.dc.b	$6F,$D7,$FB,$11,$FE,$E5,$7F,$DA,$A7,$99,$FC,$22,$3F,$DD,$77,$FC
	.dc.b	$3E,$3F,$C4,$07,$73,$F8,$D5,$73,$70,$7F,$40,$29,$F7,$6F,$DE,$FF
	.dc.b	$89,$AF,$CF,$7A,$8D,$47,$79,$FD,$0C,$65,$FF,$13,$5F,$F9,$C2,$7F
	.dc.b	$8F,$4F,$F0,$F3,$F8,$91,$E3,$3F,$A1,$4F,$C2,$BF,$C7,$1F,$20,$B5
	.dc.b	$FC,$66,$B4,$FF,$27,$AF,$F5,$41,$CC,$F6,$CF,$EF,$6F,$F0,$73,$FE
	.dc.b	$2E,$9F,$E2,$CF,$EB,$EB,$E8,$9F,$02,$3F,$D4,$E7,$33,$FD,$53,$7A
	.dc.b	$8E,$47,$BA,$BF,$C5,$CF,$EA,$E7,$EA,$9F,$20,$F9,$07,$F8,$98,$FF
	.dc.b	$80,$07,$EC,$6F,$B3,$FE,$AA,$9F,$B6,$FD,$EB,$E4,$FF,$A4,$E3,$EF
	.dc.b	$CE,$4E,$31,$FE,$09,$8E,$F7,$EB,$8F,$C8,$3F,$E9,$FA,$FA,$8F,$E4
	.dc.b	$0F,$B3,$1F,$82,$7D,$8F,$F7,$C3,$CE,$7D,$33,$C6,$7C,$E3,$F2,$15
	.dc.b	$F6,$1F,$E9,$CF,$91,$FF,$14,$DF,$F1,$D7,$EE,$57,$27,$FC,$D2,$F2
	.dc.b	$3E,$B4,$FF,$A6,$AF,$EC,$D7,$D1,$AF,$D5,$BE,$B8,$FF,$48,$5F,$D5
	.dc.b	$AC,$CF,$FB,$AE,$7E,$A9,$FB,$3A,$FA,$D7,$FB,$19,$FA,$51,$F4,$E3
	.dc.b	$F9,$24,$7E,$D2,$3F,$C7,$A7,$FA,$2A,$7F,$5F,$79,$AF,$C6,$4F,$9C
	.dc.b	$FB,$EF,$E8,$23,$E9,$47,$EA,$0F,$88,$FF,$08,$7F,$6C,$FD,$27,$F6
	.dc.b	$17,$AB,$FD,$4B,$FD,$2A,$3F,$DD,$07,$CD,$7F,$18,$FF,$A4,$8F,$92
	.dc.b	$7C,$13,$F1,$F1,$FA,$87,$E8,$9C,$6B,$E8,$4F,$D4,$31,$3F,$3E,$7F
	.dc.b	$91,$9F,$6D,$F4,$3F,$C6,$DF,$43,$FC,$3D,$FC,$4B,$FC,$31,$FE,$A8
	.dc.b	$FF,$09,$BF,$E4,$95,$F4,$5F,$E3,$F1,$F4,$5F,$E3,$E7,$E8,$9F,$E4
	.dc.b	$11,$FE,$8A,$BF,$A1,$7F,$A1,$8F,$F4,$6D,$FE,$6C,$FE,$09,$3F,$85
	.dc.b	$AF,$E2,$91,$FC,$5C,$FE,$05,$1E,$68,$FE,$3D,$1F,$C4,$5F,$E1,$95
	.dc.b	$FC,$DE,$BF,$C5,$15,$98,$FD,$11,$F4,$67,$EA,$CF,$03,$F3,$8F,$F3
	.dc.b	$F8,$FE,$24,$FF,$29,$51,$F9,$E3,$EB,$47,$F3,$28,$FE,$B0,$FF,$12
	.dc.b	$3E,$8C,$7F,$54,$9F,$CF,$47,$F5,$69,$FA,$95,$FC,$BA,$3F,$0C,$7F
	.dc.b	$96,$6F,$F8,$8B,$F8,$63,$F7,$0F,$F4,$EB,$FC,$FC,$FF,$4D,$9F,$9F
	.dc.b	$3F,$9E,$8F,$F5,$CD,$7C,$F8,$FE,$9F,$1F,$3E,$7F,$AE,$DF,$CF,$7F
	.dc.b	$85,$47,$CF,$AC,$BE,$E8,$F9,$EF,$F4,$27,$FA,$1D,$FC,$E8,$FE,$13
	.dc.b	$1F,$C2,$A3,$FA,$04,$7F,$09,$8F,$E1,$51,$FE,$27,$AF,$E8,$75,$F9
	.dc.b	$98,$FE,$79,$5F,$98,$AF,$93,$7F,$C1,$A3,$E4,$47,$C9,$9F,$EF,$F1
	.dc.b	$FB,$F8,$FE,$98,$7E,$06,$7E,$4B,$FD,$B4,$FE,$3C,$FC,$93,$F2,$EF
	.dc.b	$F0,$03,$F9,$C1,$FC,$84,$FC,$09,$FD,$A6,$7F,$83,$9F,$0D,$FF,$06
	.dc.b	$4F,$F6,$77,$E3,$9F,$C0,$6F,$F3,$97,$F2,$8F,$E0,$93,$F9,$79,$FE
	.dc.b	$68,$7E,$4E,$7F,$2C,$FC,$9B,$FF,$11,$BF,$96,$3F,$80,$DF,$C9,$AF
	.dc.b	$F2,$85,$FC,$9A,$FE,$5F,$1F,$25,$FE,$B2,$FD,$DA,$FF,$04,$BF,$C0
	.dc.b	$9F,$F7,$64,$FF,$30,$3F,$27,$1F,$2E,$3F,$B1,$9F,$0C,$F9,$47,$F2
	.dc.b	$83,$E1,$1F,$D8,$4F,$94,$7E,$48,$FE,$57,$7F,$C0,$1F,$C9,$DF,$F0
	.dc.b	$09,$F9,$11,$F2,$63,$FA,$EC,$7F,$84,$6B,$E4,$47,$F8,$4E,$3E,$44
	.dc.b	$FE,$42,$FE,$2D,$FC,$87,$E3,$DF,$EF,$6B,$F9,$45,$7E,$42,$7F,$C1
	.dc.b	$75,$F9,$08,$F9,$37,$F9,$09,$F8,$F5,$FE,$15,$3E,$44,$FF,$82,$9F
	.dc.b	$C8,$C6,$AF,$E3,$C7,$F8,$26,$7F,$AA,$CF,$E4,$2F,$F2,$95,$FE,$0C
	.dc.b	$3E,$55,$7F,$51,$AF,$C7,$DF,$F2,$58,$FD,$BC,$7C,$6B,$FD,$BC,$7C
	.dc.b	$67,$FC,$0A,$FF,$16,$8F,$F1,$DD,$FE,$39,$F9,$15,$F1,$E7,$F6,$D5
	.dc.b	$F8,$A9,$FC,$74,$7E,$DA,$3E,$3C,$FE,$D9,$FF,$17,$CF,$E3,$9F,$8F
	.dc.b	$3F,$8E,$8F,$8F,$1F,$8E,$3E,$3B,$FD,$E0,$FE,$37,$5F,$C7,$A3,$E3
	.dc.b	$57,$ED,$AB,$F8,$BC,$7F,$91,$AF,$FD,$89,$7F,$E6,$D8,$C3,$FC,$E4
	.dc.b	$FF,$4D,$C7,$F9,$46,$BF,$C4,$74,$AD,$D7,$F9,$D0,$FD,$E8,$AE,$0F
	.dc.b	$FA,$86,$58,$71,$8D,$4F,$FA,$0E,$70,$F0,$BF,$24,$E8,$F2,$2C,$39
	.dc.b	$E1,$E2,$9F,$F8,$03,$7C,$AF,$FD,$C1,$7E,$39,$C1,$CA,$78,$BD,$4F
	.dc.b	$6A,$36,$FF,$A4,$2F,$FD,$B7,$5F,$0D,$FF,$4D,$B6,$7C,$87,$09,$FF
	.dc.b	$73,$47,$FB,$06,$FC,$D7,$AA,$F0,$CF,$F9,$F6,$BF,$C7,$F5,$A8,$FF
	.dc.b	$49,$36,$3E,$06,$1D,$6F,$D9,$87,$9E,$B5,$7D,$EF,$D8,$7F,$68,$7E
	.dc.b	$95,$F9,$63,$EB,$5F,$BF,$0E,$95,$FD,$BA,$79,$3F,$81,$3F,$D8,$2F
	.dc.b	$49,$E3,$87,$86,$31,$3D,$DE,$73,$F2,$EF,$C1,$7D,$CF,$51,$F3,$2B
	.dc.b	$FD,$61,$3E,$3B,$F8,$B1,$89,$FA,$91,$DE,$3F,$29,$1F,$EA,$A6,$BD
	.dc.b	$F0,$FC,$26,$1F,$E2,$F8,$F9,$0F,$CC,$3D,$47,$E7,$A7,$F4,$75,$D5
	.dc.b	$FD,$29,$F0,$8F,$D3,$CF,$92,$7F,$A8,$D7,$DC,$7E,$41,$DA,$FE,$BC
	.dc.b	$65,$FC,$2B,$6B,$A4,$FE,$4F,$0F,$A9,$87,$C5,$AF,$C3,$BF,$16,$7F
	.dc.b	$C5,$8F,$E4,$2B,$E3,$BF,$13,$0F,$C2,$3F,$0E,$33,$1F,$0E,$7F,$06
	.dc.b	$7E,$38,$FA,$E7,$EB,$A3,$EF,$3E,$83,$33,$F0,$DF,$F1,$E4,$7E,$F8
	.dc.b	$DC,$FC,$38,$FF,$1C,$61,$F8,$BB,$FC,$D3,$D6,$BE,$E9,$F8,$A3,$F1
	.dc.b	$8F,$C4,$9F,$C4,$BF,$B7,$9E,$35,$D6,$7F,$85,$3D,$5F,$C9,$3F,$15
	.dc.b	$1F,$E4,$67,$E3,$3F,$E5,$37,$E3,$47,$EC,$9F,$8D,$19,$7E,$35,$7E
	.dc.b	$CE,$3B,$CF,$DD,$7D,$47,$F2,$53,$E3,$47,$F1,$E7,$D4,$D3,$EE,$BF
	.dc.b	$C9,$4E,$AF,$F1,$F8,$FC,$24,$68,$FD,$85,$74,$9F,$E3,$37,$F4,$67
	.dc.b	$F9,$74,$FC,$57,$F3,$2F,$47,$FD,$47,$7F,$16,$BF,$8A,$47,$59,$FE
	.dc.b	$2B,$5F,$2D,$B5,$F8,$83,$A5,$7F,$1B,$7E,$85,$7E,$1E,$3F,$0A,$F3
	.dc.b	$7F,$50,$FE,$10,$F8,$2F,$DE,$79,$B6,$39,$BF,$55,$FE,$0B,$1C,$E7
	.dc.b	$E0,$3F,$8E,$79,$B6,$75,$3E,$43,$F4,$D1,$F8,$27,$EA,$BF,$82,$9F
	.dc.b	$E1,$58,$7F,$0E,$9E,$73,$97,$F0,$2D,$7E,$53,$FE,$2B,$AF,$B8,$FC
	.dc.b	$93,$DE,$7D,$BB,$FA,$8F,$F0,$08,$D3,$C8,$FC,$66,$1F,$79,$FE,$09
	.dc.b	$3F,$45,$FF,$4E,$CF,$F5,$F3,$B6,$1C,$63,$FA,$35,$FD,$57,$ED,$57
	.dc.b	$E7,$1F,$EA,$91,$EC,$7F,$A7,$9F,$6D,$F9,$EF,$E0,$EB,$EA,$3F,$64
	.dc.b	$F9,$E7,$C7,$7F,$53,$3E,$CA,$FE,$D1,$1B,$AF,$B3,$1C,$EB,$EB,$1F
	.dc.b	$66,$BF,$19,$5E,$C3,$F7,$F3,$F8,$29,$FB,$95,$F6,$4E,$4F,$D1,$3F
	.dc.b	$2C,$FB,$5F,$C3,$57,$8E,$3E,$BC,$7E,$25,$FC,$AB,$FB,$63,$F9,$D9
	.dc.b	$F8,$23,$F9,$9C,$7C,$F7,$E3,$4F,$FB,$35,$FE,$86,$7C,$F3,$F2,$97
	.dc.b	$F5,$CF,$E7,$13,$F5,$E7,$8C,$FF,$5D,$9F,$E7,$95,$FA,$6A,$FC,$9D
	.dc.b	$FE,$C2,$BF,$59,$1F,$AF,$3F,$88,$BF,$DA,$5F,$5D,$FF,$80,$F0,$FE
	.dc.b	$E9,$5E,$B3,$F5,$F5,$FD,$D0,$FA,$E7,$F6,$DA,$FA,$D1,$FC,$CC,$FD
	.dc.b	$74,$F2,$AF,$9B,$5F,$E0,$17,$E2,$1F,$DE,$23,$FB,$AD,$72,$8F,$AD
	.dc.b	$3F,$83,$7F,$25,$7E,$CB,$FE,$E4,$F2,$3E,$B4,$7F,$77,$7F,$5D,$1F
	.dc.b	$69,$F9,$93,$E2,$9F,$E6,$AF,$88,$FD,$59,$F5,$A3,$E9,$CF,$EA,$9F
	.dc.b	$BB,$3C,$67,$FB,$6E,$1F,$53,$0F,$D5,$D7,$F6,$FB,$FD,$4B,$F4,$CF
	.dc.b	$06,$1F,$D1,$6D,$87,$F2,$2B,$FA,$B5,$FC,$86,$FF,$5A,$FE,$C5,$FA
	.dc.b	$47,$EA,$2B,$E9,$5F,$E7,$9F,$C5,$C7,$D1,$7F,$91,$4F,$E9,$8F,$98
	.dc.b	$FE,$96,$BE,$7C,$7F,$89,$27,$F8,$DC,$7F,$89,$23,$E3,$1E,$7B,$FE
	.dc.b	$33,$87,$F5,$78,$FC,$C9,$F8,$83,$FC,$71,$3F,$E3,$BB,$F3,$9F,$D5
	.dc.b	$4F,$E2,$78,$7E,$84,$FD,$E3,$E7,$3F,$AA,$4F,$CE,$33,$7F,$54,$FC
	.dc.b	$79,$FC,$3A,$FF,$C3,$13,$FE,$1D,$9F,$9C,$7F,$07,$AF,$41,$F8,$43
	.dc.b	$F3,$F8,$7F,$99,$4F,$9B,$1F,$C5,$0F,$CC,$CF,$E5,$23,$F0,$77,$FE
	.dc.b	$5C,$9F,$D0,$47,$F1,$1C,$3F,$39,$1F,$32,$BF,$33,$1F,$33,$0F,$F0
	.dc.b	$7D,$7C,$C7,$FC,$83,$5F,$99,$AF,$EF,$C7,$E5,$70,$FE,$C5,$1E,$6A
	.dc.b	$FC,$E9,$F9,$97,$F8,$6C,$FE,$44,$FC,$C1,$F1,$0F,$E4,$EF,$E6,$1F
	.dc.b	$E1,$6F,$CB,$3F,$C8,$4F,$F3,$C9,$FE,$14,$7C,$E9,$FC,$BD,$FF,$90
	.dc.b	$2B,$F2,$E7,$ED,$A7,$E5,$3F,$C6,$1F,$E0,$E7,$E0,$2F,$FC,$BB,$5F
	.dc.b	$80,$BF,$E7,$47,$CB,$3E,$CB,$E6,$8F,$F4,$2C,$7F,$0B,$AF,$CB,$1F
	.dc.b	$DC,$9F,$97,$1F,$E3,$8A,$FE,$09,$5F,$2E,$7F,$C9,$35,$F2,$DF,$E7
	.dc.b	$07,$F6,$D9,$F9,$71,$F9,$63,$F8,$5C,$7F,$37,$BF,$CB,$3F,$CD,$4F
	.dc.b	$F3,$FD,$FF,$9D,$9F,$31,$FE,$56,$3F,$2C,$7F,$08,$3F,$D2,$EE,$DF
	.dc.b	$E3,$A7,$F8,$B5,$E5,$87,$F8,$05,$B7,$FD,$DA,$7E,$2C,$6B,$0F,$E0
	.dc.b	$6F,$0C,$35,$5F,$C1,$EB,$FD,$76,$68,$FF,$34,$BF,$87,$7F,$C5,$F5
	.dc.b	$FE,$39,$7F,$D8,$2E,$CF,$F5,$E4,$7C,$78,$FF,$2A,$C7,$FA,$44,$FF
	.dc.b	$60,$1F,$C9,$2F,$FA,$0C,$7F,$7B,$3F,$B8,$98,$8E,$87,$C3,$7F,$C6
	.dc.b	$8D,$3F,$92,$BF,$E7,$E7,$FC,$97,$7F,$86,$9E,$91,$FD,$66,$75,$1F
	.dc.b	$31,$B7,$F1,$CF,$84,$7F,$B8,$DF,$09,$F4,$DF,$84,$FF,$AB,$67,$E1
	.dc.b	$47,$FB,$A5,$DC,$7F,$AC,$63,$E9,$3F,$06,$3D,$4F,$C1,$AF,$B9,$7A
	.dc.b	$8F,$C4,$9F,$E8,$C8,$FC,$05,$72,$BF,$C5,$4F,$B8,$FC,$43,$F2,$5F
	.dc.b	$75,$FF,$9E,$E3,$DC,$D9,$FB,$E7,$E4,$EF,$DD,$3F,$B3,$73,$3F,$75
	.dc.b	$F6,$47,$ED,$E7,$F1,$93,$EB,$AF,$CB,$3E,$B7,$E9,$1F,$40,$F5,$BF
	.dc.b	$82,$9F,$BF,$3C,$9F,$D2,$47,$AC,$FC,$2B,$F6,$0E,$27,$A9,$FC,$03
	.dc.b	$FA,$28,$FA,$E7,$E8,$67,$F6,$EF,$E2,$CF,$F1,$4B,$FE,$FF,$9F,$AA
	.dc.b	$FF,$AB,$23,$E8,$3F,$9B,$3D,$51,$FE,$A3,$7E,$61,$F5,$0F,$55,$FD
	.dc.b	$49,$FE,$A5,$5F,$50,$F1,$57,$D1,$BF,$B0,$F5,$8F,$55,$FF,$BA,$8F
	.dc.b	$D6,$D7,$D3,$38,$CF,$E7,$EB,$F1,$F3,$FC,$C2,$3E,$A4,$F2,$8F,$F0
	.dc.b	$6C,$FE,$BD,$4F,$D5,$7F,$C8,$07,$C0,$3F,$A7,$5F,$F2,$A7,$EA,$9C
	.dc.b	$CF,$DB,$9F,$C3,$A3,$F5,$A7,$D4,$79,$3F,$E5,$F7,$F0,$37,$F5,$A3
	.dc.b	$2F,$D5,$BF,$E4,$07,$E3,$AB,$FA,$AB,$F5,$4F,$CF,$9F,$8E,$72,$FD
	.dc.b	$4A,$FE,$52,$7A,$4F,$E4,$F5,$F5,$4F,$B0,$FA,$4F,$F1,$7C,$7E,$9D
	.dc.b	$FA,$07,$1A,$FD,$31,$F6,$CE,$0F,$F9,$88,$FA,$67,$CE,$3E,$45,$FD
	.dc.b	$27,$F8,$F9,$FD,$66,$B3,$5F,$1E,$3F,$36,$7D,$A3,$F4,$A7,$E8,$E3
	.dc.b	$F9,$41,$F8,$A7,$F4,$F1,$FA,$37,$35,$C5,$FF,$24,$CF,$CF,$6C,$7C
	.dc.b	$F3,$FC,$0B,$3F,$E2,$B8,$FE,$26,$F1,$47,$F9,$5E,$FE,$8D,$68,$FC
	.dc.b	$24,$FE,$26,$7F,$10,$7F,$2F,$AF,$EB,$37,$FC,$56,$BE,$79,$F4,$E7
	.dc.b	$FA,$99,$F8,$67,$F8,$EC,$F8,$63,$FC,$54,$7F,$16,$8F,$D1,$1F,$4E
	.dc.b	$BF,$89,$C7,$CF,$9F,$E2,$51,$F3,$63,$E7,$47,$F1,$18,$FE,$8F,$1F
	.dc.b	$28,$ED,$3F,$84,$3F,$A2,$3F,$9C,$7F,$2E,$FE,$6E,$7F,$87,$9C,$0F
	.dc.b	$E1,$D1,$F4,$0F,$CC,$CF,$F4,$33,$F3,$27,$AC,$FC,$11,$FC,$30,$FC
	.dc.b	$C4,$FF,$10,$7F,$31,$7F,$DD,$E3,$E3,$5F,$C9,$7F,$C0,$75,$F2,$6B
	.dc.b	$F2,$D1,$FC,$28,$FC,$B4,$FF,$08,$7F,$86,$CF,$C7,$9F,$F1,$44,$7C
	.dc.b	$7A,$FE,$DB,$3F,$C1,$E7,$FB,$55,$7F,$20,$3C,$B1,$FD,$D2,$3F,$2D
	.dc.b	$5F,$24,$F8,$B3,$FC,$F1,$F9,$17,$FC,$DA,$7F,$D5,$91,$F9,$43,$F7
	.dc.b	$B3,$FC,$E6,$BE,$45,$7E,$4A,$FF,$7B,$5F,$92,$3F,$1B,$7F,$BF,$AF
	.dc.b	$C8,$57,$FA,$06,$3E,$3C,$FF,$88,$4F,$C4,$1F,$6C,$FE,$CF,$3F,$90
	.dc.b	$9F,$F1,$5C,$FC,$8A,$FF,$10,$D7,$C8,$8F,$8E,$FC,$89,$FE,$6B,$1F
	.dc.b	$8E,$8F,$93,$1F,$8E,$9F,$93,$1F,$E1,$C3,$F1,$EE,$1F,$E5,$B3,$F2
	.dc.b	$1F,$C8,$CF,$F8,$62,$FE,$45,$FE,$4C,$FC,$7D,$7E,$EE,$BE,$4D,$7F
	.dc.b	$B5,$6B,$E4,$47,$FB,$5E,$BF,$1F,$5F,$B9,$8F,$DD,$57,$FA,$96,$BF
	.dc.b	$25,$5F,$22,$FF,$92,$DF,$F2,$4B,$FE,$4B,$7F,$B6,$8F,$C6,$C7,$E3
	.dc.b	$A3,$E3,$C7,$E3,$63,$F1,$D1,$FD,$62,$FF,$1D,$5F,$E4,$58,$F8,$F5
	.dc.b	$F8,$EA,$F8,$F3,$F0,$DF,$F4,$0D,$7F,$81,$9F,$8E,$7F,$5A,$9F,$C5
	.dc.b	$3F,$8D,$AF,$8D,$1F,$E3,$B8,$FF,$1E,$47,$E3,$6B,$FC,$6D,$1F,$8D
	.dc.b	$8F,$8D,$3F,$16,$3E,$34,$7E,$35,$FE,$FA,$7E,$35,$F8,$D1,$FE,$CB
	.dc.b	$BF,$8D,$3F,$B3,$8F,$83,$97,$F8,$C2,$3F,$DC,$67,$F9,$B5,$C5,$7F
	.dc.b	$6A,$C5,$18,$9F,$F4,$FB,$5F,$ED,$A6,$8F,$E3,$E7,$F2,$EA,$4E,$63
	.dc.b	$AC,$0C,$3F,$E7,$DA,$FE,$6A,$7F,$36,$72,$7F,$99,$4F,$F7,$CE,$3D
	.dc.b	$2F,$C9,$28,$E9,$18,$C3,$F9,$FB,$AB,$E1,$7F,$06,$BC,$51,$C2,$3F
	.dc.b	$9B,$C7,$F9,$7D,$F0,$9E,$47,$D0,$72,$7F,$A2,$38,$8D,$C7,$99,$F8
	.dc.b	$B3,$F3,$27,$FD,$EE,$FD,$19,$F9,$33,$E1,$7F,$A7,$57,$D3,$7F,$A4
	.dc.b	$BF,$D6,$2B,$C1,$1F,$1E,$3E,$9B,$A3,$B9,$67,$6F,$F8,$51,$EC,$F1
	.dc.b	$BF,$9D,$7F,$80,$AF,$9B,$87,$AA,$7F,$C5,$27,$CD,$BC,$4A,$3B,$DF
	.dc.b	$A1,$FB,$95,$F0,$A3,$E6,$D7,$81,$FF,$78,$9D,$5F,$97,$15,$ED,$7F
	.dc.b	$86,$7A,$CF,$E1,$5F,$92,$FC,$93,$D2,$53,$C9,$17,$0F,$04,$7F,$62
	.dc.b	$8D,$61,$A8,$E5,$5A,$7F,$1C,$7A,$4F,$D2,$E1,$4F,$2B,$F6,$5F,$2B
	.dc.b	$F0,$6B,$C9,$1F,$8C,$7E,$F5,$FE,$2D,$B5,$F9,$E6,$C7,$86,$BF,$C5
	.dc.b	$71,$AB,$F3,$47,$FA,$9A,$7E,$5C,$FE,$26,$7F,$0D,$3F,$D1,$0E,$87
	.dc.b	$EF,$5F,$C2,$B9,$7F,$46,$E4,$FB,$B1,$CD,$F4,$1C,$6F,$E6,$BF,$D4
	.dc.b	$1E,$AD,$AF,$45,$7E,$0C,$EB,$1B,$7D,$2F,$FA,$5A,$FD,$C7,$C3,$7F
	.dc.b	$89,$3F,$EB,$0B,$FD,$23,$FA,$B6,$BF,$3A,$3F,$D9,$D7,$F8,$D8,$F8
	.dc.b	$71,$F6,$9F,$CC,$3F,$10,$FE,$18,$7E,$A7,$0C,$BF,$68,$FF,$1E,$1F
	.dc.b	$E7,$EA,$FE,$39,$3F,$A8,$8F,$C8,$3D,$2B,$E1,$C7,$47,$F8,$85,$67
	.dc.b	$0F,$C5,$47,$F1,$AC,$3E,$1D,$6E,$FA,$3F,$BB,$7E,$7B,$FC,$29,$EA
	.dc.b	$FF,$21,$9F,$D8,$57,$06,$CF,$ED,$63,$FD,$DD,$36,$3F,$63,$87,$DB
	.dc.b	$7C,$77,$F0,$E7,$A3,$F0,$E7,$F8,$04,$7C,$33,$F8,$61,$F3,$CF,$B8
	.dc.b	$74,$3F,$09,$3D,$2F,$CF,$1E,$F7,$E1,$CF,$F0,$B7,$A4,$78,$1E,$87
	.dc.b	$F0,$C9,$F8,$71,$FC,$19,$FE,$04,$FC,$83,$E1,$3F,$C1,$A7,$F8,$25
	.dc.b	$FC,$E8,$F8,$31,$FC,$1D,$E7,$3E,$F3,$FC,$E3,$1F,$C0,$E3,$F7,$91
	.dc.b	$ED,$3F,$D1,$31,$FC,$26,$3F,$65,$1F,$D9,$0F,$E0,$07,$E6,$8F,$AD
	.dc.b	$1F,$BE,$BF,$34,$7F,$2B,$8F,$E0,$33,$FB,$BA,$FC,$F4,$FE,$E6,$3F
	.dc.b	$52,$FF,$95,$63,$51,$E3,$7F,$B2,$57,$79,$FF,$17,$CF,$EB,$9F,$9E
	.dc.b	$7F,$B5,$0D,$1F,$ED,$69,$D4,$F9,$23,$FD,$B2,$E9,$FE,$9F,$3F,$CD
	.dc.b	$63,$DD,$5F,$C2,$EF,$CB,$5F,$ED,$AB,$FD,$EC,$7C,$7A,$FD,$F3,$C4
	.dc.b	$FC,$CC,$FF,$77,$3F,$C0,$17,$E6,$7F,$B6,$9E,$17,$F7,$8F,$F0,$07
	.dc.b	$E9,$1F,$02,$3F,$78,$F4,$8E,$87,$4B,$FF,$08,$3E,$D9,$FF,$0B,$47
	.dc.b	$F8,$4E,$FF,$63,$3F,$BB,$3F,$C2,$67,$E6,$4F,$E9,$91,$FE,$08,$79
	.dc.b	$1F,$D8,$1F,$F0,$5C,$72,$9F,$CB,$CF,$F5,$FA,$FC,$BC,$FF,$10,$8F
	.dc.b	$DC,$1E,$37,$F2,$CF,$F8,$2A,$3F,$42,$FF,$82,$5F,$D0,$C7,$F5,$A8
	.dc.b	$FE,$71,$1F,$B6,$7E,$EC,$FF,$8B,$27,$F8,$05,$FF,$3A,$9F,$DA,$BF
	.dc.b	$04,$FF,$17,$61,$F4,$EF,$F6,$93,$9B,$FD,$9B,$FE,$0A,$7C,$6F,$DA
	.dc.b	$AD,$4F,$EC,$8F,$64,$7F,$20,$3E,$73,$EC,$7F,$DC,$4F,$DA,$6B,$FB
	.dc.b	$C3,$D8,$FF,$75,$C3,$FB,$A3,$F9,$13,$D4,$7F,$57,$7F,$B9,$D7,$ED
	.dc.b	$27,$F6,$2F,$E2,$67,$FB,$8C,$FF,$6C,$C3,$37,$F9,$D9,$FE,$21,$3F
	.dc.b	$B0,$BF,$F0,$85,$7E,$C2,$7F,$D8,$47,$E8,$5F,$43,$FE,$C5,$AE,$51
	.dc.b	$F9,$1A,$FE,$EB,$5F,$AF,$AE,$D5,$FC,$CE,$7F,$B3,$57,$EB,$A7,$F5
	.dc.b	$47,$F6,$A3,$FB,$65,$7E,$A1,$E3,$7F,$AC,$7F,$CB,$97,$FE,$45,$BF
	.dc.b	$D8,$3A,$3F,$6C,$7E,$3E,$BF,$47,$3E,$F7,$FB,$33,$FA,$9B,$FE,$D7
	.dc.b	$1F,$AA,$3C,$51,$FA,$83,$F9,$65,$7F,$1D,$8F,$EC,$31,$FA,$03,$F0
	.dc.b	$A7,$F6,$23,$F0,$B7,$E6,$7F,$B3,$C7,$F2,$C8,$FD,$A4,$7F,$9D,$8F
	.dc.b	$E1,$27,$FA,$01,$E6,$FF,$65,$AF,$D5,$5F,$EB,$A7,$F6,$53,$FE,$F5
	.dc.b	$AF,$EC,$F5,$F5,$E3,$FB,$03,$F8,$F7,$FD,$08,$FD,$87,$31,$FA,$A3
	.dc.b	$9B,$FC,$46,$BF,$83,$3F,$E6,$33,$F4,$C7,$90,$FE,$C2,$7B,$9F,$E0
	.dc.b	$CF,$F9,$C0,$E6,$7D,$6A,$F7,$BF,$C8,$0F,$D6,$57,$ED,$23,$91,$FD
	.dc.b	$72,$7F,$83,$61,$FC,$6A,$FF,$AB,$1F,$BC,$3D,$97,$FD,$71,$F9,$E7
	.dc.b	$F5,$A7,$EC,$57,$F9,$A2,$7F,$AB,$C7,$E7,$CF,$C9,$1F,$E6,$1B,$FE
	.dc.b	$AE,$FF,$5B,$07,$E0,$8F,$E5,$D8,$7F,$95,$1E,$55,$F9,$37,$E5,$C7
	.dc.b	$F5,$B7,$FC,$BD,$87,$ED,$E7,$E0,$C7,$FA,$04,$FD,$3E,$1C,$5F,$F0
	.dc.b	$9D,$7F,$95,$DF,$D8,$C7,$85,$FE,$89,$7F,$C3,$1F,$AC,$7F,$45,$BF
	.dc.b	$5B,$FE,$59,$8F,$D6,$57,$F9,$4E,$7F,$47,$3E,$27,$F1,$EF,$F9,$4C
	.dc.b	$FD,$3D,$FF,$1C,$7F,$D2,$F1,$FA,$B8,$FE,$49,$7F,$AA,$9F,$D5,$C7
	.dc.b	$F4,$EB,$FD,$4D,$7E,$B6,$BD,$15,$FE,$D1,$7D,$0F,$F5,$27,$FA,$6B
	.dc.b	$FC,$D8,$FF,$15,$1F,$D3,$A7,$33,$FA,$D9,$FD,$41,$FD,$2E,$7F,$45
	.dc.b	$3F,$E3,$B7,$F8,$11,$FA,$99,$FD,$34,$FF,$4E,$3F,$23,$5F,$46,$7F
	.dc.b	$C7,$D5,$FE,$43,$3F,$48,$FF,$4D,$AF,$EA,$55,$F9,$A8,$FC,$1C,$7E
	.dc.b	$91,$FC,$F1,$FA,$49,$FE,$9A,$FE,$8C,$FC,$D1,$C4,$FF,$1B,$4F,$FB
	.dc.b	$02,$BF,$AA,$D7,$F6,$79,$FE,$99,$5F,$E4,$38,$FE,$97,$87,$E8,$6F
	.dc.b	$FC,$77,$5F,$EB,$DA,$FF,$2D,$47,$A2,$3F,$8F,$57,$F4,$E8,$F0,$47
	.dc.b	$F8,$E4,$FF,$00,$DF,$F8,$E2,$BF,$3F,$5F,$A3,$9F,$AE,$7F,$84,$6B
	.dc.b	$FA,$4B,$F9,$E3,$F3,$53,$FD,$21,$FC,$ED,$7F,$AE,$E7,$FC,$63,$1F
	.dc.b	$9C,$3F,$A3,$3F,$8B,$9F,$DA,$47,$F8,$E6,$B3,$3F,$D1,$63,$FA,$09
	.dc.b	$F9,$69,$FC,$CB,$FE,$AD,$9F,$CC,$CF,$F4,$38,$FF,$14,$1F,$98,$9F
	.dc.b	$F1,$6C,$FE,$5F,$0F,$F1,$95,$7F,$3C,$AF,$CA,$1F,$E2,$97,$F3,$C7
	.dc.b	$91,$FC,$A3,$F0,$9F,$E7,$E7,$FA,$8E,$3F,$C4,$EF,$FA,$82,$BF,$20
	.dc.b	$FF,$3E,$3F,$90,$9F,$EA,$63,$E8,$DF,$E4,$1F,$D7,$9F,$CE,$F0,$FF
	.dc.b	$4F,$5F,$FA,$8F,$0F,$CB,$C7,$F0,$08,$FE,$76,$7F,$86,$AF,$F8,$FD
	.dc.b	$7F,$88,$5C,$D7,$E5,$A7,$FD,$02,$FE,$56,$7F,$D2,$27,$E5,$69,$87
	.dc.b	$FA,$5A,$FF,$C4,$D5,$FC,$F2,$F8,$E1,$FA,$68,$F3,$E1,$FE,$8F,$AF
	.dc.b	$E6,$C7,$E5,$5F,$CD,$57,$F8,$7C,$FD,$69,$FD,$45,$FE,$53,$48,$F4
	.dc.b	$C7,$F7,$39,$FE,$77,$5F,$97,$9F,$F4,$7C,$FF,$85,$CF,$E4,$87,$F8
	.dc.b	$78,$FF,$2C,$C7,$E5,$CF,$F1,$13,$FC,$B0,$F3,$9F,$EC,$BB,$FF,$0F
	.dc.b	$4F,$F3,$28,$FC,$BC,$7F,$87,$4F,$EF,$8F,$FB,$24,$FC,$C5,$FF,$30
	.dc.b	$BF,$F4,$5D,$FF,$36,$3F,$2F,$7F,$96,$AF,$CB,$D7,$F8,$62,$BF,$2F
	.dc.b	$1F,$E0,$7A,$FE,$0F,$3F,$E1,$CA,$FC,$C5,$FF,$A1,$6B,$F3,$11,$FE
	.dc.b	$81,$AF,$CC,$47,$FB,$12,$3F,$97,$4F,$FB,$E2,$3F,$31,$7F,$CB,$9F
	.dc.b	$F3,$ED,$7E,$6E,$BF,$AC,$9F,$CB,$23,$FC,$29,$3F,$96,$3F,$95,$BF
	.dc.b	$E1,$2C,$3F,$71,$3F,$EF,$38,$FE,$57,$1F,$94,$BF,$8B,$87,$FB,$C6
	.dc.b	$7F,$DF,$15,$FC,$AC,$FC,$AB,$FC,$9A,$FF,$95,$47,$F8,$81,$FD,$64
	.dc.b	$7F,$83,$A3,$F2,$AF,$E3,$4F,$E4,$27,$F8,$2C,$ED,$1F,$94,$7F,$2E
	.dc.b	$7F,$7A,$C3,$FC,$EB,$7F,$BC,$AF,$E5,$53,$FE,$0A,$C3,$F9,$65,$FE
	.dc.b	$56,$BF,$DD,$15,$F9,$53,$FC,$13,$5F,$C9,$67,$F9,$4C,$7F,$82,$0F
	.dc.b	$F3,$81,$FC,$A2,$3F,$24,$7E,$3A,$3F,$93,$BF,$EE,$0B,$FC,$8C,$FF
	.dc.b	$26,$8F,$F6,$EE,$1E,$87,$E4,$D7,$F2,$7B,$FF,$6F,$4F,$D0,$7F,$23
	.dc.b	$3F,$D4,$DB,$3F,$E5,$48,$FF,$6F,$3F,$92,$7F,$C0,$77,$F9,$29,$E2
	.dc.b	$FF,$3D,$83,$FE,$02,$7F,$4E,$7F,$7B,$7F,$8A,$BF,$E0,$87,$F9,$5D
	.dc.b	$FF,$2A,$8F,$E5,$77,$FE,$D3,$BF,$E5,$71,$F9,$58,$FF,$00,$DF,$E4
	.dc.b	$E3,$FC,$02,$FA,$AB,$FA,$94,$7E,$FD,$FF,$35,$5F,$E5,$5F,$F4,$A1
	.dc.b	$FE,$68,$8F,$E5,$58,$7F,$2B,$8F,$91,$1F,$CA,$67,$FC,$01,$3F,$1E
	.dc.b	$3F,$CC,$87,$E5,$AB,$FD,$9B,$5F,$CA,$63,$F2,$B3,$FC,$A6,$39,$C7
	.dc.b	$F9,$78,$FC,$B4,$7F,$80,$23,$31,$FD,$FC,$A2,$D8,$7F,$AC,$A3,$FC
	.dc.b	$5F,$3F,$EC,$D7,$F9,$65,$FF,$2B,$8E,$35,$F4,$A7,$FC,$03,$1F,$CB
	.dc.b	$0F,$F6,$5B,$F9,$6B,$FC,$BC,$7F,$80,$F0,$FE,$FF,$7F,$2E,$3F,$95
	.dc.b	$9F,$E5,$F3,$F4,$D5,$FC,$A5,$EE,$7E,$B9,$FE,$5E,$F9,$DF,$E3,$B1
	.dc.b	$FE,$C8,$BF,$F0,$04,$FE,$66,$7F,$35,$1F,$52,$7F,$CB,$91,$F9,$A7
	.dc.b	$F9,$49,$FE,$5B,$3F,$32,$7F,$29,$7F,$94,$E1,$FC,$FD,$FE,$55,$3F
	.dc.b	$CA,$6B,$F2,$F1,$F9,$88,$FA,$95,$F9,$88,$FF,$2C,$1F,$E6,$1B,$FE
	.dc.b	$7C,$FF,$8B,$8F,$95,$1F,$E5,$99,$FE,$55,$5F,$97,$8F,$E5,$51,$F9
	.dc.b	$78,$FE,$55,$3F,$97,$9F,$E0,$C7,$E5,$CF,$E5,$53,$FE,$57,$9F,$E5
	.dc.b	$33,$FC,$AA,$3F,$C5,$D3,$FE,$BD,$3F,$BF,$9F,$C8,$A3,$FC,$AA,$7F
	.dc.b	$AF,$6F,$F9,$5C,$66,$3F,$BF,$D7,$F9,$55,$FC,$19,$FE,$3D,$3F,$C6
	.dc.b	$27,$F9,$58,$B5,$EE,$7F,$D8,$11,$FC,$B4,$FC,$5D,$7C,$D9,$FC,$9C
	.dc.b	$7E,$09,$FF,$43,$CF,$FA,$7A,$7F,$04,$FE,$65,$FC,$13,$F2,$E7,$F3
	.dc.b	$0F,$E5,$AF,$D6,$FF,$A7,$8F,$E6,$31,$97,$F3,$38,$7F,$00,$7F,$99
	.dc.b	$5F,$F9,$60,$F1,$57,$AE,$39,$1F,$A9,$63,$FE,$06,$9F,$9A,$FF,$AD
	.dc.b	$2B,$D8,$7F,$81,$2B,$F8,$45,$FF,$81,$27,$F3,$B3,$FC,$CD,$FF,$2B
	.dc.b	$1F,$CC,$DF,$CD,$61,$FC,$AA,$3F,$B7,$BF,$65,$FE,$5B,$5E,$D8,$FF
	.dc.b	$5C,$3F,$BE,$8F,$F0,$AD,$7F,$35,$77,$87,$FA,$DE,$FF,$CB,$71,$FE
	.dc.b	$B2,$3F,$CB,$73,$FC,$CE,$FF,$9B,$CF,$F9,$1E,$3F,$D3,$B3,$F9,$93
	.dc.b	$FD,$93,$3F,$9F,$7F,$CB,$93,$E1,$AF,$E6,$71,$F4,$5F,$4B,$FE,$14
	.dc.b	$7F,$C7,$4F,$FB,$CE,$FF,$D7,$C7,$E9,$CF,$E7,$87,$E2,$5F,$F0,$83
	.dc.b	$FE,$59,$8F,$E3,$53,$FE,$61,$9F,$E6,$E7,$F8,$81,$FE,$6E,$7E,$21
	.dc.b	$FE,$71,$1F,$CD,$EB,$E5,$47,$E6,$0F,$F5,$B1,$F9,$13,$F6,$F5,$E9
	.dc.b	$7F,$C1,$91,$FC,$15,$FC,$E9,$FC,$DC,$FF,$79,$9F,$87,$3F,$21,$1F
	.dc.b	$E0,$D8,$FD,$73,$FE,$A7,$BF,$96,$FF,$65,$9F,$CF,$5F,$E8,$67,$FD
	.dc.b	$41,$3F,$A4,$BF,$F3,$16,$1F,$C1,$B0,$FF,$0C,$C7,$FA,$86,$7F,$9A
	.dc.b	$57,$FA,$D8,$FF,$07,$4E,$B0,$FF,$5B,$57,$C8,$C3,$FC,$1B,$1F,$CD
	.dc.b	$E7,$9D,$FE,$C9,$B5,$FC,$E2,$7E,$65,$7F,$38,$7F,$C1,$D8,$7F,$AD
	.dc.b	$E3,$F3,$73,$FE,$5F,$8F,$C9,$4F,$F7,$93,$FB,$C1,$FA,$A3,$F9,$19
	.dc.b	$F3,$6B,$F3,$CF,$FA,$54,$FC,$29,$FE,$8F,$7F,$3A,$7F,$38,$7F,$36
	.dc.b	$7F,$98,$0F,$D1,$D7,$E7,$23,$F3,$31,$F9,$C8,$FE,$6F,$1F,$9C,$AF
	.dc.b	$F0,$74,$FE,$4C,$FE,$6E,$7E,$6A,$3F,$9D,$47,$E7,$A3,$FD,$6D,$7F
	.dc.b	$45,$FC,$D3,$F9,$87,$FC,$BF,$3F,$E1,$88,$FC,$C4,$7F,$9F,$1F,$CC
	.dc.b	$1F,$E0,$98,$FE,$71,$1F,$30,$FF,$30,$D7,$F3,$88,$FE,$CD,$5F,$97
	.dc.b	$9F,$F5,$B1,$F8,$E3,$F2,$F8,$7F,$98,$6B,$F9,$9C,$78,$0F,$C8,$9F
	.dc.b	$E4,$C3,$E2,$BF,$E6,$37,$F9,$CD,$FE,$7C,$FF,$5F,$1C,$5F,$F4,$E9
	.dc.b	$FA,$EC,$3F,$AE,$BF,$96,$3F,$C1,$F8,$7F,$90,$AB,$FC,$C8,$71,$8F
	.dc.b	$E6,$CF,$E4,$1F,$F3,$21,$FE,$0C,$3F,$CC,$07,$E3,$CB,$87,$FA,$DE
	.dc.b	$BF,$9A,$47,$FB,$F2,$3F,$D4,$2F,$F3,$1B,$FE,$6A,$71,$74,$F8,$4F
	.dc.b	$F2,$39,$FE,$B5,$9F,$E6,$B1,$F9,$18,$FF,$1D,$3F,$9A,$BF,$F4,$1D
	.dc.b	$FE,$C0,$FE,$9F,$3F,$E3,$F7,$FD,$73,$3F,$CD,$A3,$FC,$1F,$3F,$98
	.dc.b	$BF,$F5,$C4,$7F,$83,$E7,$F9,$A9,$F8,$93,$F8,$0C,$FE,$65,$FD,$A3
	.dc.b	$FE,$10,$7E,$AB,$F9,$87,$F9,$A4,$7F,$83,$E3,$F3,$11,$FE,$0F,$7F
	.dc.b	$9A,$D7,$FB,$45,$FE,$65,$1F,$EB,$87,$FC,$25,$87,$FA,$E2,$3F,$CC
	.dc.b	$F7,$FE,$64,$9F,$F5,$B4,$7F,$99,$1F,$CC,$4F,$F9,$91,$FC,$D4,$FF
	.dc.b	$91,$DF,$F0,$7C,$FE,$5C,$FE,$5E,$7E,$EC,$FE,$BB,$5F,$99,$7F,$83
	.dc.b	$CF,$E6,$5F,$F2,$04,$FF,$84,$1F,$CD,$57,$E6,$63,$FC,$1D,$3F,$E6
	.dc.b	$47,$F9,$74,$FF,$70,$BF,$C9,$BF,$98,$A6,$1F,$EB,$B8,$FE,$63,$3F
	.dc.b	$81,$7F,$31,$5F,$1B,$0F,$F0,$79,$FD,$9E,$3F,$97,$DF,$FA,$F0,$F9
	.dc.b	$4F,$F3,$2C,$3F,$97,$D7,$F8,$EE,$BF,$C1,$AF,$F2,$BA,$FE,$5D,$1F
	.dc.b	$E0,$B9,$FC,$C4,$FC,$6A,$FC,$99,$FE,$B9,$7F,$96,$D7,$FA,$E6,$3F
	.dc.b	$C1,$F5,$FE,$B9,$3F,$C0,$6F,$E4,$CF,$F0,$53,$FE,$B9,$AF,$E5,$95
	.dc.b	$FC,$9A,$7F,$24,$FF,$28,$AF,$F5,$CB,$F9,$3B,$FC,$94,$FF,$8E,$63
	.dc.b	$F9,$4B,$FE,$08,$9F,$F0,$23,$FD,$02,$3E,$7B,$FE,$1E,$8F,$F3,$03
	.dc.b	$FE,$34,$3F,$95,$47,$E4,$8F,$E5,$53,$FC,$12,$3F,$05,$5F,$EB,$73
	.dc.b	$F9,$FB,$FD,$A6,$BF,$24,$FF,$AD,$67,$F9,$53,$FD,$F0,$FD,$31,$FC
	.dc.b	$9C,$FE,$56,$7F,$2E,$9C,$BF,$E0,$77,$F2,$8F,$F3,$2B,$FF,$01,$5F
	.dc.b	$F9,$88,$FF,$47,$5F,$DC,$9F,$CB,$4F,$F8,$0C,$1F,$C1,$A7,$FD,$28
	.dc.b	$EE,$3F,$AF,$C7,$E5,$DF,$F3,$26,$1F,$EB,$B8,$FC,$BC,$7F,$81,$0F
	.dc.b	$DC,$9F,$EB,$A3,$F2,$D1,$F9,$57,$FC,$C7,$7E,$A7,$F9,$6E,$1F,$E0
	.dc.b	$3B,$FE,$89,$1F,$CB,$4F,$CB,$E1,$F9,$58,$FC,$CD,$7E,$52,$7F,$AC
	.dc.b	$47,$E5,$1F,$E5,$73,$F9,$4C,$3F,$C0,$15,$F9,$49,$F8,$D5,$FE,$00
	.dc.b	$9F,$C9,$E1,$FE,$B6,$7F,$BF,$CF,$F9,$6D,$F8,$73,$F9,$48,$FF,$2B
	.dc.b	$DF,$EC,$27,$F9,$3C,$FE,$56,$7F,$CB,$6F,$F2,$67,$FC,$02,$0F,$F2
	.dc.b	$E4,$7F,$27,$7F,$D6,$AF,$F8,$C9,$FE,$B8,$FE,$59,$FE,$98,$FF,$AD
	.dc.b	$5F,$F2,$CE,$1F,$EE,$77,$F8,$85,$7E,$58,$FB,$6F,$E5,$8F,$F5,$CD
	.dc.b	$7F,$53,$7F,$C5,$53,$F9,$77,$FC,$25,$87,$E5,$DF,$F5,$B1,$FD,$2C
	.dc.b	$FE,$4E,$7F,$96,$CF,$F2,$81,$F5,$8F,$CC,$1F,$CA,$4F,$F0,$1D,$7F
	.dc.b	$29,$7F,$31,$3F,$D6,$9F,$F1,$05,$7E,$B0,$F3,$C7,$FA,$DE,$FF,$BF
	.dc.b	$CF,$FA,$DD,$FA,$27,$C5,$7F,$BD,$9F,$E0,$0C,$3F,$24,$FF,$80,$F0
	.dc.b	$FE,$50,$FE,$48,$FF,$2A,$9F,$92,$9F,$F2,$8E,$1F,$9D,$8F,$F0,$2D
	.dc.b	$FF,$27,$3F,$25,$86,$63,$FD,$12,$FF,$27,$7F,$B4,$4F,$FA,$6C,$FF
	.dc.b	$2A,$9F,$D9,$8F,$F5,$9A,$BF,$F1,$CB,$FE,$A6,$3F,$C0,$07,$F7,$B3
	.dc.b	$E7,$1F,$E6,$67,$FD,$0E,$FF,$AB,$27,$F1,$CF,$E6,$2F,$FC,$AA,$FF
	.dc.b	$2B,$3F,$D6,$F5,$FC,$AD,$FF,$00,$CF,$F2,$B9,$FC,$AC,$7F,$2B,$3F
	.dc.b	$C0,$11,$FD,$F4,$FF,$03,$1E,$87,$FC,$AC,$7E,$5E,$FF,$BE,$BF,$EB
	.dc.b	$1B,$FE,$FB,$5F,$4E,$3F,$CA,$87,$F7,$E9,$FF,$2F,$9E,$C7,$FA,$C4
	.dc.b	$FE,$FD,$F7,$47,$E6,$5F,$F3,$15,$FE,$66,$FC,$17,$FE,$B0,$8F,$F5
	.dc.b	$94,$FE,$6A,$FF,$CA,$EF,$F2,$E3,$FC,$AF,$7F,$95,$9F,$E5,$B1,$FE
	.dc.b	$97,$9F,$EF,$CF,$F9,$58,$F1,$1F,$E2,$43,$F9,$69,$FE,$AC,$72,$7F
	.dc.b	$8B,$6B,$FA,$D4,$7E,$62,$FF,$04,$D3,$FC,$AF,$1F,$98,$3F,$CA,$F7
	.dc.b	$FE,$2E,$8F,$E5,$27,$F9,$5C,$F8,$4F,$FB,$CE,$FF,$C0,$11,$F9,$93
	.dc.b	$F9,$51,$FC,$32,$7F,$D4,$EF,$FA,$B4,$FF,$1C,$1F,$EA,$D3,$F9,$4D
	.dc.b	$F4,$7F,$BF,$9F,$C0,$2F,$F3,$4F,$F9,$48,$FF,$2B,$9F,$E0,$07,$F9
	.dc.b	$5B,$FE,$04,$7F,$1E,$FC,$D7,$F9,$63,$F9,$7A,$FE,$58,$FF,$95,$85
	.dc.b	$7F,$AB,$67,$8C,$7F,$9F,$DF,$0D,$FC,$F8,$FF,$57,$1F,$98,$8E,$2F
	.dc.b	$F8,$A6,$3F,$CA,$F7,$FE,$AF,$8F,$EA,$F3,$FE,$AF,$7F,$98,$E1,$FE
	.dc.b	$AF,$BF,$E6,$35,$FE,$04,$9F,$F0,$24,$7F,$81,$6F,$F3,$31,$FE,$35
	.dc.b	$3F,$5A,$7F,$BC,$DF,$F0,$2B,$FC,$C0,$FC,$D4,$7F,$95,$5F,$F3,$AB
	.dc.b	$FE,$55,$7E,$A1,$FC,$B0,$FC,$5E,$1F,$C6,$A3,$F3,$18,$7F,$AB,$A7
	.dc.b	$F3,$15,$FC,$BD,$FE,$84,$7F,$43,$BF,$E6,$0F,$FA,$BC,$FF,$59,$3F
	.dc.b	$92,$7F,$97,$CF,$DA,$7E,$55,$7F,$AB,$5F,$CC,$9F,$EA,$D9,$FC,$CD
	.dc.b	$7F,$AB,$AB,$FC,$07,$3F,$EA,$D8,$FE,$5C,$FF,$95,$0F,$E5,$C7,$FB
	.dc.b	$72,$3B,$3F,$C6,$23,$FD,$AD,$5F,$25,$FC,$C6,$1F,$CC,$27,$F2,$67
	.dc.b	$F9,$3C,$FF,$47,$4F,$F8,$A9,$FE,$FA,$7E,$64,$FF,$2D,$1F,$99,$13
	.dc.b	$FC,$3E,$FF,$34,$FF,$32,$BF,$F2,$BC,$FE,$60,$FE,$62,$FF,$AB,$A7
	.dc.b	$FC,$05,$1F,$97,$8F,$C5,$BF,$CB,$4F,$E5,$E7,$F1,$23,$F8,$A1,$E0
	.dc.b	$8F,$E5,$F3,$FC,$DE,$3D,$27,$F2,$F3,$F2,$91,$FE,$B0,$8F,$F5,$5E
	.dc.b	$1F,$9E,$9F,$E3,$6F,$F3,$17,$F8,$FC,$72,$7F,$33,$5F,$A2,$3F,$BA
	.dc.b	$BF,$9A,$BF,$F0,$4D,$66,$7F,$99,$9C,$67,$F3,$07,$F9,$78,$A7,$A8
	.dc.b	$FC,$5D,$FF,$34,$7F,$DC,$A7,$F9,$7C,$FC,$D9,$FE,$5F,$8F,$E6,$B1
	.dc.b	$FC,$8A,$3F,$8F,$1F,$D9,$CF,$E6,$71,$FC,$79,$FF,$31,$3F,$CD,$1F
	.dc.b	$F0,$5C,$FF,$34,$7F,$C2,$31,$FE,$49,$7F,$88,$3F,$E0,$77,$F3,$47
	.dc.b	$E7,$4F,$E7,$13,$F9,$D3,$F9,$AB,$FE,$0A,$AF,$CE,$5F,$DA,$8F,$F0
	.dc.b	$AB,$EF,$9F,$B1,$19,$3F,$CC,$51,$FE,$0D,$7F,$5A,$7F,$B7,$0F,$F5
	.dc.b	$A9,$FC,$E6,$FF,$55,$7F,$E6,$13,$D1,$1F,$59,$FF,$1B,$9F,$E6,$4C
	.dc.b	$FF,$41,$1D,$1A,$A7,$FD,$63,$18,$78,$47,$F3,$B8,$DC,$EA,$3E,$DD
	.dc.b	$7F,$9C,$21,$3F,$CF,$A6,$9F,$EB,$AA,$E0,$FF,$40,$AF,$05,$6A,$3F
	.dc.b	$CD,$65,$3F,$5B,$1F,$EF,$FA,$FF,$60,$1F,$B9,$AF,$C8,$B6,$BD,$58
	.dc.b	$F1,$BF,$E2,$B7,$DA,$FD,$71,$FB,$0C,$3F,$CB,$58,$71,$C3,$CB,$3F
	.dc.b	$E4,$A3,$D9,$1F,$C1,$E7,$E4,$47,$F9,$D3,$0E,$CF,$EF,$CF,$5C,$FF
	.dc.b	$9D,$9C,$9F,$E2,$18,$AF,$F8,$AF,$0E,$AF,$ED,$1F,$E5,$73,$FE,$7A
	.dc.b	$3F,$D9,$B7,$CA,$3F,$A5,$CD,$BF,$E1,$F3,$E8,$8F,$C3,$BF,$4C,$E0
	.dc.b	$7C,$67,$FD,$A3,$86,$DF,$F1,$24,$7F,$AA,$23,$E2,$C7,$FB,$D6,$3F
	.dc.b	$D5,$73,$F1,$A3,$FD,$B3,$1C,$5F,$CA,$99,$3F,$77,$5E,$E7,$FD,$0D
	.dc.b	$5D,$1F,$F0,$F5,$FF,$B6,$CF,$E0,$2F,$C2,$3F,$26,$F2,$7F,$95,$BD
	.dc.b	$23,$94,$7F,$21,$8E,$F7,$A8,$FC,$9D,$E9,$F0,$BF,$31,$FE,$BD,$7F
	.dc.b	$4A,$FE,$04,$D3,$F9,$45,$FF,$B8,$8F,$C7,$D7,$F6,$FB,$FC,$0B,$FE
	.dc.b	$37,$3C,$07,$F8,$E6,$BC,$0F,$F3,$18,$F8,$8F,$E2,$DD,$BF,$8D,$AD
	.dc.b	$C7,$F4,$88,$D9,$FC,$01,$C5,$FE,$21,$FE,$9B,$5C,$23,$CF,$3F,$E9
	.dc.b	$4B,$FE,$5B,$5F,$84,$BF,$C7,$9F,$81,$9F,$F0,$89,$E2,$3F,$0E,$FE
	.dc.b	$61,$FE,$D5,$3F,$E4,$03,$FD,$8E,$7D,$68,$F9,$95,$F8,$08,$FE,$0D
	.dc.b	$87,$E5,$6B,$E7,$C7,$FB,$A8,$FE,$0F,$1F,$83,$9F,$BD,$3F,$BC,$3D
	.dc.b	$E7,$FB,$A3,$0F,$C6,$47,$FA,$5A,$BF,$8B,$BE,$E9,$FE,$11,$7F,$91
	.dc.b	$7F,$84,$BE,$4A,$F7,$47,$F2,$89,$FE,$01,$7F,$A9,$9F,$BF,$7F,$C1
	.dc.b	$F0,$F0,$DF,$CC,$C3,$F8,$73,$FE,$F2,$7F,$5C,$FE,$24,$FF,$24,$5E
	.dc.b	$2F,$E1,$D7,$F8,$86,$77,$5E,$58,$FA,$57,$FE,$9E,$9F,$DC,$61,$FC
	.dc.b	$5A,$3F,$7A,$7F,$BD,$8F,$E1,$CF,$8A,$3F,$9D,$47,$FA,$81,$FD,$DD
	.dc.b	$71,$7F,$76,$E2,$BD,$75,$FC,$07,$0F,$D9,$57,$A4,$FE,$6F,$7F,$B0
	.dc.b	$C3,$F2,$F1,$E9,$8F,$DA,$3F,$DD,$30,$CD,$F1,$8F,$E7,$13,$FB,$39
	.dc.b	$FB,$98,$7C,$A9,$FF,$10,$57,$DF,$9F,$D1,$DF,$D4,$9F,$EE,$27,$E4
	.dc.b	$4F,$F4,$44,$7E,$45,$FE,$E3,$5C,$E7,$F8,$8D,$7E,$D6,$BF,$4F,$5E
	.dc.b	$33,$FC,$0E,$7F,$AF,$6F,$F4,$2F,$33,$FD,$FE,$E2,$FC,$11,$F5,$E7
	.dc.b	$F5,$4F,$A6,$3F,$5F,$3F,$AA,$AE,$53,$FB,$37,$F2,$4F,$F9,$55,$FD
	.dc.b	$A1,$F9,$C7,$F5,$67,$C3,$A1,$E5,$7F,$CF,$0F,$E4,$9F,$D7,$98,$7F
	.dc.b	$A2,$4F,$EC,$23,$E9,$47,$EB,$5F,$80,$7F,$07,$3F,$04,$7F,$96,$2F
	.dc.b	$F6,$AF,$DB,$3F,$D2,$40,$FC,$61,$FE,$95,$8D,$D7,$43,$FD,$CD,$3F
	.dc.b	$B2,$9F,$D2,$57,$EC,$AF,$F1,$B1,$FD,$43,$0F,$A9,$87,$EB,$4E,$57
	.dc.b	$FD,$2E,$30,$FE,$D6,$3E,$4C,$FF,$74,$8F,$F0,$44,$FF,$1D,$8F,$94
	.dc.b	$7F,$6B,$3F,$37,$5F,$D2,$A3,$F6,$15,$FB,$48,$CC,$FB,$0F,$AD,$1F
	.dc.b	$B4,$C3,$D9,$3F,$CA,$2B,$F6,$37,$98,$EB,$3C,$6F,$FD,$37,$7F,$DA
	.dc.b	$70,$FD,$81,$FE,$5C,$C3,$F6,$38,$7F,$22,$9F,$EC,$B5,$FB,$47,$F8
	.dc.b	$A3,$FD,$18,$FD,$79,$FE,$5E,$8F,$D7,$BF,$08,$FF,$21,$D7,$F2,$D9
	.dc.b	$FE,$4D,$1F,$AB,$3F,$26,$7F,$97,$E3,$FA,$FB,$FB,$0B,$FD,$74,$FF
	.dc.b	$59,$8F,$D4,$3F,$8F,$7F,$CB,$75,$FB,$28,$F0,$47,$F6,$4C,$3F,$46
	.dc.b	$FF,$2A,$8F,$A4,$FE,$8C,$E4,$7F,$88,$E7,$F6,$33,$93,$89,$F8,$D9
	.dc.b	$FD,$5C,$FE,$90,$CD,$7E,$A9,$FF,$6B,$4F,$EB,$E3,$F3,$F1,$FD,$B5
	.dc.b	$FC,$0B,$CE,$7F,$3E,$FF,$93,$4F,$3B,$FA,$97,$FD,$D7,$3F,$AF,$8E
	.dc.b	$71,$FC,$A0,$FD,$4D,$FF,$97,$9F,$CC,$9F,$A8,$3F,$D7,$27,$94,$FD
	.dc.b	$64,$EA,$7E,$DD,$79,$0F,$D5,$1F,$A7,$9F,$C6,$57,$F5,$6B,$FD,$5D
	.dc.b	$FE,$9A,$BF,$19,$5F,$E1,$2B,$FD,$13,$FC,$DA,$BF,$4D,$7F,$E5,$97
	.dc.b	$F4,$D7,$9A,$E3,$5F,$EB,$B7,$FA,$7B,$F9,$C3,$D4,$7F,$97,$A3,$FB
	.dc.b	$8B,$FE,$BE,$3E,$FD,$F1,$C3,$F8,$BD,$7E,$96,$7F,$43,$7F,$DC,$2F
	.dc.b	$F7,$F1,$FE,$53,$7F,$4B,$3F,$DB,$CF,$D2,$47,$F0,$97,$F8,$41,$EA
	.dc.b	$7F,$A9,$E1,$FE,$ED,$8F,$D2,$B5,$F5,$C7,$DD,$8F,$A8,$FF,$1D,$AF
	.dc.b	$E2,$0F,$D3,$C3,$F9,$EC,$FF,$82,$8F,$F2,$6D,$7F,$18,$9F,$9D,$5F
	.dc.b	$C5,$EB,$84,$7F,$AB,$2B,$FC,$13,$5F,$E0,$7A,$F4,$CF,$FB,$C0,$F9
	.dc.b	$CF,$F8,$40,$FC,$63,$FE,$09,$7D,$2F,$F9,$78,$FF,$03,$47,$F7,$D7
	.dc.b	$F0,$97,$FB,$F7,$E9,$1F,$50,$FF,$42,$4F,$D8,$8F,$E2,$F3,$FE,$05
	.dc.b	$BE,$27,$F1,$63,$85,$7F,$BE,$DF,$85,$3E,$B8,$FF,$43,$4F,$D3,$3F
	.dc.b	$28,$7F,$6D,$7F,$CF,$6F,$29,$FF,$5A,$57,$F1,$78,$FC,$19,$FC,$C2
	.dc.b	$FF,$CE,$35,$FC,$3C,$FD,$F9,$FE,$E2,$7F,$CE,$B7,$93,$F8,$3C,$7D
	.dc.b	$93,$FD,$82,$7F,$17,$BE,$2F,$C9,$AF,$F0,$2B,$FC,$5E,$BE,$CC,$7F
	.dc.b	$17,$8F,$E3,$13,$FE,$71,$BF,$E2,$D7,$FC,$68,$FE,$2A,$FD,$89,$FF
	.dc.b	$5A,$1F,$C5,$4F,$E2,$27,$EE,$4F,$F3,$81,$FD,$2E,$33,$5F,$C5,$2F
	.dc.b	$FD,$E6,$7F,$7D,$AF,$F3,$54,$FF,$13,$3F,$8B,$CF,$F9,$E8,$F9,$95
	.dc.b	$FC,$4C,$FE,$22,$F7,$3F,$0A,$FF,$12,$BF,$F3,$7C,$7F,$1A,$8F,$F3
	.dc.b	$31,$FE,$9E,$8F,$F1,$9D,$7F,$42,$AF,$99,$1F,$E3,$6A,$FF,$17,$D7
	.dc.b	$CB,$3E,$7C,$7B,$5F,$C1,$1F,$E4,$6B,$F9,$67,$E5,$5F,$F4,$F9,$FD
	.dc.b	$08,$F9,$33,$F1,$6F,$CF,$1F,$EB,$87,$FC,$5B,$3F,$E9,$E8,$FF,$4E
	.dc.b	$1F,$D0,$AB,$FB,$A4,$FF,$40,$8F,$99,$1F,$DD,$EB,$E6,$47,$F3,$93
	.dc.b	$F9,$FC,$FC,$C7,$FC,$4B,$7F,$32,$7E,$2C,$7F,$37,$3E,$5C,$FC,$F8
	.dc.b	$FE,$AB,$1F,$E2,$87,$F9,$DB,$FE,$CE,$7F,$04,$FB,$CF,$F2,$E9,$F9
	.dc.b	$43,$E3,$1C,$8F,$F1,$04,$E6,$3E,$6C,$7F,$6E,$AF,$F4,$A5,$7F,$97
	.dc.b	$AB,$F3,$47,$D0,$AF,$ED,$15,$F2,$EB,$F2,$D1,$FE,$C7,$9F,$CB,$47
	.dc.b	$F8,$8A,$FF,$CA,$35,$F9,$58,$F9,$35,$F9,$38,$F9,$33,$FE,$56,$8F
	.dc.b	$93,$1F,$C3,$6B,$FB,$FB,$6B,$F3,$91,$F8,$08,$FF,$28,$57,$F9,$56
	.dc.b	$BE,$44,$7C,$BA,$FC,$8D,$FC,$BB,$FE,$C8,$7E,$62,$BF,$C2,$35,$FB
	.dc.b	$F3,$E4,$4F,$E4,$DF,$F4,$3C,$FC,$79,$FF,$09,$C7,$C7,$BF,$F4,$01
	.dc.b	$FE,$12,$3F,$DE,$37,$FE,$12,$7E,$A5,$FE,$A2,$4F,$F9,$89,$FC,$9B
	.dc.b	$FB,$E7,$E5,$4F,$C9,$7F,$DD,$C7,$F8,$02,$FF,$CB,$0F,$EF,$A7,$F2
	.dc.b	$55,$FE,$7B,$AF,$C9,$5F,$F8,$3E,$BF,$C2,$11,$FE,$74,$AF,$C9,$57
	.dc.b	$F9,$EE,$FF,$24,$FE,$26,$FC,$C7,$96,$BE,$4B,$FB,$A9,$F9,$31,$FB
	.dc.b	$C8,$F9,$37,$F9,$08,$FD,$DC,$7C,$88,$F9,$31,$FB,$B8,$F9,$35,$F9
	.dc.b	$79,$FB,$A7,$F4,$B9,$FD,$E1,$F9,$73,$F3,$EF,$C0,$3F,$9D,$C7,$CC
	.dc.b	$9F,$F1,$1D,$F9,$E3,$0F,$CB,$75,$7F,$C0,$27,$FD,$4F,$5F,$93,$9F
	.dc.b	$F5,$2D,$7C,$A8,$F9,$F5,$FC,$9A,$BE,$55,$7F,$26,$9F,$F1,$A5,$7C
	.dc.b	$A7,$FC,$6D,$5F,$29,$FF,$4C,$1F,$91,$9F,$99,$5F,$C9,$4F,$DD,$9F
	.dc.b	$C3,$A3,$FD,$3C,$FC,$A8,$FE,$A3,$7F,$29,$FE,$9F,$3F,$29,$FF,$32
	.dc.b	$1F,$26,$3F,$C8,$71,$F2,$8F,$8C,$7F,$86,$0F,$E4,$4F,$CB,$AF,$DD
	.dc.b	$57,$F4,$FA,$F9,$55,$F9,$28,$FD,$D3,$F9,$28,$FE,$44,$FF,$4E,$3F
	.dc.b	$91,$3F,$22,$3F,$74,$FE,$44,$FD,$D3,$F2,$23,$F9,$15,$7E,$46,$3F
	.dc.b	$90,$C7,$F2,$28,$FF,$18,$C7,$F2,$28,$FE,$47,$1F,$90,$9F,$93,$5F
	.dc.b	$90,$7F,$D8,$53,$F3,$1F,$F1,$39,$EE,$7F,$D4,$8D,$7F,$C4,$27,$F9
	.dc.b	$00,$FF,$3A,$61,$FB,$57,$FD,$87,$5F,$E5,$93,$FC,$0C,$FC,$93,$F2
	.dc.b	$0F,$F9,$7A,$BF,$CB,$57,$FE,$29,$3F,$CC,$58,$7F,$64,$9F,$DE,$C7
	.dc.b	$F8,$3E,$3F,$C5,$35,$FE,$BD,$8F,$F2,$8C,$7F,$83,$E3,$F7,$B1,$FE
	.dc.b	$5A,$7F,$D7,$31,$FE,$51,$7F,$C3,$E7,$F9,$62,$BD,$38,$7F,$1F,$A3
	.dc.b	$F7,$A3,$FD,$C4,$DB,$FF,$10,$47,$AE,$BF,$CC,$91,$CC,$FF,$35,$D7
	.dc.b	$F4,$A8,$FF,$13,$9F,$EE,$5A,$FF,$5F,$1F,$E5,$7B,$FF,$14,$5E,$67
	.dc.b	$FC,$C1,$5F,$EB,$E7,$F2,$67,$F8,$80,$FD,$C4,$FF,$2B,$AF,$F1,$4D
	.dc.b	$FF,$03,$7F,$35,$3F,$DF,$63,$FA,$8D,$7F,$89,$A7,$F9,$59,$FD,$F2
	.dc.b	$7F,$BE,$57,$F2,$B3,$F8,$03,$89,$FE,$F8,$7F,$68,$8F,$D7,$47,$F9
	.dc.b	$5D,$F7,$3F,$D8,$8F,$E5,$87,$9C,$5F,$FA,$21,$FE,$A4,$7F,$72,$7F
	.dc.b	$B4,$DF,$F2,$EB,$FE,$E1,$1F,$42,$7F,$3F,$1F,$E2,$BB,$FC,$FB,$FC
	.dc.b	$BE,$BF,$3E,$FF,$04,$9F,$E0,$C7,$F8,$5E,$7F,$0E,$FF,$BC,$1F,$E0
	.dc.b	$E7,$F2,$27,$0F,$FB,$A0,$1F,$82,$7D,$35,$FE,$BA,$7E,$B3,$EC,$7F
	.dc.b	$A4,$D7,$F4,$AB,$61,$C6,$B5,$5F,$EB,$C3,$FD,$52,$E1,$FF,$61,$9B
	.dc.b	$7F,$D8,$F7,$CB,$0F,$F0,$6C,$7F,$A8,$DB,$3E,$03,$E3,$47,$3C,$3F
	.dc.b	$D9,$38,$71,$9F,$49,$B8,$FD,$5D,$72,$8F,$24,$7A,$E3,$10,$8F,$F6
	.dc.b	$9D,$F3,$BF,$44,$FF,$BE,$EB,$75,$FB,$68,$F8,$B3,$FE,$75,$BE,$0D
	.dc.b	$BF,$74,$FE,$E6,$36,$FE,$EE,$36,$FF,$55,$8F,$F4,$11,$ED,$3F,$AC
	.dc.b	$5F,$81,$FE,$2B,$1E,$E3,$FD,$47,$3D,$9F,$E1,$C7,$B4,$D1,$ED,$8F
	.dc.b	$E0,$2E,$0F,$CB,$DF,$F9,$16,$BB,$3F,$E0,$66,$DF,$9A,$7F,$8D,$4F
	.dc.b	$CC,$7A,$BC,$AB,$DB,$1A,$BE,$08,$FF,$29,$3F,$90,$BF,$96,$FF,$86
	.dc.b	$E3,$D4,$7D,$37,$FD,$C5,$3F,$7D,$FC,$BD,$F1,$9F,$0C,$7F,$B3,$27
	.dc.b	$4F,$E7,$30,$E7,$3F,$E8,$89,$FB,$6F,$E6,$AF,$F0,$2F,$E6,$EB,$6F
	.dc.b	$EF,$63,$2F,$EE,$1A,$FC,$73,$E5,$1F,$94,$9F,$E8,$71,$C3,$0F,$8D
	.dc.b	$1E,$07,$EA,$34,$FC,$84,$6E,$BD,$91,$F9,$0C,$3E,$24,$F9,$9F,$E2
	.dc.b	$F1,$99,$EA,$FF,$A4,$5E,$AD,$3A,$DF,$DF,$9F,$A3,$5F,$02,$7F,$1B
	.dc.b	$3F,$83,$7E,$23,$4E,$A7,$E5,$4E,$A5,$7A,$9F,$A6,$3F,$32,$7F,$82
	.dc.b	$8F,$88,$FE,$76,$FF,$04,$7D,$48,$FD,$85,$7C,$07,$84,$76,$9F,$F2
	.dc.b	$BC,$7F,$BE,$6B,$D1,$3F,$46,$3F,$6B,$7F,$1A,$BF,$38,$FC,$68,$FF
	.dc.b	$4F,$3E,$A8,$FE,$49,$1F,$96,$7F,$29,$1F,$E2,$B8,$E6,$F5,$7F,$96
	.dc.b	$3F,$94,$9F,$C5,$47,$F7,$C9,$FD,$9B,$FC,$4D,$FE,$3E,$E5,$F5,$1F
	.dc.b	$C3,$4F,$D8,$BF,$83,$9F,$81,$1F,$C5,$CF,$24,$FF,$1C,$C3,$C1,$1F
	.dc.b	$1A,$6C,$F4,$3F,$8B,$CF,$E7,$B0,$FB,$B5,$FB,$47,$E2,$5F,$F2,$77
	.dc.b	$F5,$6F,$F1,$5C,$3F,$8D,$C7,$EA,$63,$FC,$2E,$FF,$83,$5F,$B9,$3F
	.dc.b	$0C,$FE,$82,$FE,$20,$FE,$2C,$FF,$A5,$2B,$E3,$47,$C2,$8D,$9E,$77
	.dc.b	$E0,$9F,$7B,$0F,$83,$1F,$C3,$0F,$83,$1F,$84,$9E,$78,$7D,$99,$FE
	.dc.b	$34,$F3,$3F,$7D,$87,$EE,$AF,$E0,$DF,$F0,$2A,$F9,$71,$F0,$1F,$EB
	.dc.b	$2F,$F0,$A3,$EE,$1F,$05,$CB,$D1,$FD,$E9,$FC,$8C,$A7,$E8,$23,$FC
	.dc.b	$11,$3F,$C1,$1F,$C6,$47,$E6,$2B,$F8,$5B,$F2,$4F,$C2,$4F,$F0,$57
	.dc.b	$91,$FE,$70,$8F,$73,$FC,$EA,$3F,$85,$4E,$A3,$FC,$97,$5F,$21,$FE
	.dc.b	$59,$87,$DE,$9F,$E0,$51,$B3,$33,$D2,$7D,$D1,$FC,$7E,$74,$FF,$5D
	.dc.b	$3F,$80,$1E,$A7,$F8,$19,$F8,$28,$F0,$47,$F8,$0E,$34,$FF,$4A,$AF
	.dc.b	$E1,$8F,$BC,$FF,$0F,$9C,$EF,$F8,$0C,$7E,$EA,$B9,$D7,$F5,$5A,$FE
	.dc.b	$09,$5C,$DF,$E2,$8F,$37,$FC,$2F,$7F,$4E,$7F,$6F,$1E,$A8,$FB,$97
	.dc.b	$F8,$37,$F6,$31,$C2,$75,$1C,$E3,$F8,$B9,$FE,$0D,$3A,$47,$EE,$6F
	.dc.b	$FA,$CD,$FF,$45,$BF,$F0,$0D,$7F,$45,$3F,$B0,$9F,$9C,$3E,$D5,$FD
	.dc.b	$59,$FF,$11,$47,$EE,$2B,$9B,$FB,$BA,$E7,$3F,$BB,$8F,$EC,$18,$7E
	.dc.b	$EE,$FE,$54,$FF,$83,$E7,$F8,$95,$FE,$6D,$FF,$10,$CF,$E9,$67,$FD
	.dc.b	$13,$87,$F0,$AA,$FD,$D3,$FC,$CE,$B9,$BF,$B9,$3F,$D2,$A7,$F1,$A2
	.dc.b	$BC,$DF,$71,$FB,$EA,$FE,$F7,$1F,$E0,$87,$DD,$1F,$CF,$A3,$FC,$0B
	.dc.b	$3F,$22,$3D,$D1,$FE,$06,$7E,$F6,$1F,$E0,$EA,$E6,$7E,$E2,$3F,$3A
	.dc.b	$FD,$29,$FD,$9D,$7D,$28,$FB,$B1,$FE,$00,$7F,$91,$1E,$E7,$FD,$B2
	.dc.b	$71,$2D,$FE,$CA,$7F,$33,$1E,$CA,$DB,$F8,$F7,$FC,$0A,$FE,$16,$79
	.dc.b	$3F,$E4,$BB,$FE,$F9,$5C,$A7,$F1,$B3,$FD,$54,$FF,$01,$BF,$97,$8F
	.dc.b	$DA,$47,$FA,$6A,$FF,$42,$FF,$73,$7F,$68,$7E,$5A,$3F,$A9,$5F,$F7
	.dc.b	$B8,$FD,$2C,$F8,$8F,$F3,$6D,$7F,$69,$7F,$83,$CF,$AD,$FE,$E9,$19
	.dc.b	$9F,$D6,$BC,$67,$FA,$93,$FA,$EA,$CC,$7E,$D4,$FE,$06,$7F,$8C,$67
	.dc.b	$FC,$52,$7A,$AB,$FD,$18,$7E,$AA,$7F,$81,$C7,$F0,$C8,$FE,$D6,$FC
	.dc.b	$B9,$FE,$E2,$FD,$08,$FC,$D9,$F9,$13,$F6,$13,$FA,$47,$F3,$47,$AC
	.dc.b	$FD,$4C,$7E,$41,$F2,$5F,$EA,$1F,$EC,$E7,$F4,$77,$F5,$73,$FB,$18
	.dc.b	$FC,$79,$FB,$6B,$FE,$79,$7F,$E2,$9B,$FD,$B5,$7B,$6B,$FB,$73,$FC
	.dc.b	$00,$FC,$7C,$FF,$69,$9F,$EE,$11,$F8,$F9,$FF,$0A,$9F,$AC,$3F,$0F
	.dc.b	$5F,$EA,$78,$FD,$14,$7E,$AC,$FE,$C4,$78,$AB,$FB,$15,$7F,$5E,$8F
	.dc.b	$EC,$35,$FD,$AE,$FF,$CE,$33,$FD,$8A,$B2,$FE,$A0,$FE,$0A,$7A,$67
	.dc.b	$F9,$2D,$FE,$9E,$3F,$82,$1F,$56,$3F,$CE,$38,$7F,$54,$8F,$8D,$87
	.dc.b	$A0,$FE,$04,$7D,$47,$FC,$22,$FF,$15,$3D,$07,$F1,$E7,$FC,$26,$FD
	.dc.b	$29,$F4,$3F,$3E,$BE,$94,$7F,$1D,$38,$C7,$F9,$BE,$7E,$89,$F2,$27
	.dc.b	$F8,$E4,$7D,$28,$FE,$3E,$7D,$03,$8B,$FE,$F1,$7F,$88,$BF,$42,$FD
	.dc.b	$2F,$FA,$93,$0F,$F5,$43,$F4,$27,$F8,$CC,$FE,$2A,$3F,$8A,$5F,$13
	.dc.b	$D3,$1F,$EB,$17,$F9,$11,$FE,$0F,$9F,$A0,$7F,$3D,$3E,$81,$E8,$7E
	.dc.b	$8B,$FE,$0C,$3F,$8A,$9E,$89,$FF,$24,$9F,$C4,$23,$F8,$A4,$7F,$81
	.dc.b	$A7,$FC,$13,$3F,$E8,$F3,$FC,$23,$1F,$C3,$DC,$CF,$F0,$A9,$FF,$46
	.dc.b	$5F,$F1,$37,$27,$F7,$C9,$F7,$3F,$E8,$77,$FC,$2B,$1F,$E6,$57,$F8
	.dc.b	$A3,$FD,$F0,$FD,$D3,$FD,$2A,$7F,$A9,$57,$F1,$28,$FE,$FE,$7F,$06
	.dc.b	$3F,$88,$9F,$DB,$AB,$D0,$7F,$92,$23,$F8,$8C,$7F,$80,$4F,$E1,$B3
	.dc.b	$FC,$60,$E0,$7F,$07,$7F,$D0,$D1,$FC,$30,$FE,$40,$79,$DF,$F6,$51
	.dc.b	$F3,$AF,$FB,$F5,$7F,$8C,$27,$F9,$4C,$E8,$FF,$67,$BF,$E3,$27,$E6
	.dc.b	$9F,$26,$3F,$C7,$2F,$F8,$F9,$C9,$F2,$5F,$E8,$47,$D0,$3E,$65,$7F
	.dc.b	$8D,$9F,$E1,$B1,$FE,$BB,$8F,$F6,$49,$FD,$06,$7F,$9C,$5F,$A6,$BF
	.dc.b	$C6,$6F,$CF,$8F,$E8,$07,$C1,$8F,$D2,$3F,$CF,$5F,$9F,$3F,$CF,$E7
	.dc.b	$FA,$1D,$7F,$8C,$0F,$A5,$5F,$3A,$BF,$C5,$27,$CE,$9F,$E7,$AF,$F8
	.dc.b	$B2,$7E,$8C,$FC,$D8,$F9,$0F,$C3,$7E,$43,$FC,$82,$3F,$D5,$26,$8F
	.dc.b	$F1,$4B,$FC,$EE,$3E,$59,$F1,$E7,$F9,$F4,$FC,$AA,$FE,$1D,$5F,$12
	.dc.b	$BE,$54,$7C,$6A,$FC,$C4,$7C,$8A,$F9,$B5,$F1,$63,$F2,$EF,$95,$FF
	.dc.b	$7B,$4F,$C9,$9C,$9F,$E5,$A7,$FC,$15,$1F,$ED,$73,$E4,$DF,$F9,$52
	.dc.b	$3E,$4C,$7F,$69,$9F,$93,$5F,$22,$3E,$4D,$FF,$65,$8F,$91,$1F,$26
	.dc.b	$3F,$CA,$B1,$F2,$63,$FB,$2B,$E5,$7F,$B0,$D7,$E6,$6B,$E4,$56,$AB
	.dc.b	$E4,$57,$E4,$E3,$F8,$0D,$FF,$A2,$63,$F8,$14,$7C,$88,$FF,$43,$47
	.dc.b	$C9,$8F,$91,$3F,$1E,$3E,$44,$7E,$42,$3E,$44,$7F,$92,$4F,$88,$FF
	.dc.b	$83,$AB,$E3,$CF,$EF,$23,$F1,$D3,$F8,$F8,$FC,$74,$7C,$58,$F8,$D1
	.dc.b	$FE,$0B,$8F,$8D,$5F,$EF,$29,$FC,$74,$7C,$68,$FF,$7A,$D7,$E3,$A3
	.dc.b	$E2,$CF,$E3,$8F,$F0,$43,$F2,$23,$E2,$CF,$F9,$DE,$3E,$2A,$9F,$F5
	.dc.b	$7D,$7C,$67,$F6,$C7,$C6,$8D,$57,$F8,$E2,$BF,$93,$4F,$C7,$9F,$F1
	.dc.b	$BC,$7C,$77,$FD,$57,$5F,$1E,$3E,$34,$FC,$78,$FF,$38,$5F,$E3,$63
	.dc.b	$FC,$E5,$7F,$2A,$BF,$91,$47,$C7,$BF,$C6,$47,$C7,$7F,$D4,$33,$F8
	.dc.b	$C7,$E3,$4F,$E3,$1F,$F5,$04,$FC,$78,$FF,$4E,$CF,$C7,$9F,$F4,$C4
	.dc.b	$7C,$7B,$FC,$6D,$FE,$2E,$3E,$44,$7E,$2E,$FF,$8E,$47,$E2,$EB,$E1
	.dc.b	$DF,$F1,$C3,$FC,$B6,$7E,$CD,$CC,$7F,$99,$A7,$E3,$C7,$FB,$4E,$BF
	.dc.b	$1A,$FF,$B4,$27,$F1,$AF,$07,$FC,$EA,$D7,$FD,$9A,$7E,$D8,$FC,$7D
	.dc.b	$FC,$98,$D4,$7F,$4B,$6C,$7E,$DA,$3F,$C5,$27,$ED,$85,$FF,$B3,$EF
	.dc.b	$E6,$46,$23,$FD,$34,$7C,$D7,$FD,$8A,$7F,$4D,$BF,$DC,$5F,$E4,$A3
	.dc.b	$F7,$0F,$8A,$7F,$B1,$9F,$D1,$63,$FD,$96,$7C,$F8,$FF,$64,$5F,$CF
	.dc.b	$7F,$A6,$D7,$CF,$7D,$73,$FE,$C8,$3F,$D7,$A7,$F9,$48,$FF,$04,$D7
	.dc.b	$FB,$26,$FF,$C5,$2E,$5F,$F2,$D5,$7F,$98,$8F,$F1,$81,$C2,$7F,$AE
	.dc.b	$BF,$EB,$CA,$FF,$62,$D7,$F8,$B6,$BF,$D6,$37,$F2,$6D,$7F,$EC,$5B
	.dc.b	$FA,$8A,$7F,$C2,$87,$F3,$B9,$EA,$FF,$90,$CF,$F2,$D3,$FE,$2B,$A5
	.dc.b	$7F,$8A,$23,$F6,$4F,$F8,$AA,$7F,$CB,$67,$F8,$C2,$F3,$3F,$EB,$D3
	.dc.b	$FC,$3C,$FC,$E7,$D6,$FF,$B1,$5F,$3B,$FE,$77,$7F,$C6,$31,$FE,$CA
	.dc.b	$BF,$F1,$76,$1F,$3E,$7F,$D7,$C7,$6A,$FC,$2B,$FE,$C4,$8F,$F1,$75
	.dc.b	$7F,$8C,$F0,$FF,$17,$9F,$E8,$79,$FF,$54,$1F,$33,$0F,$B2,$FF,$79
	.dc.b	$3F,$D7,$27,$F7,$13,$FB,$0C,$7F,$57,$7F,$85,$1F,$EB,$C9,$FE,$AF
	.dc.b	$3F,$C2,$8F,$C1,$E1,$FD,$FE,$7F,$44,$FF,$0A,$AF,$E1,$31,$FC,$2A
	.dc.b	$3F,$98,$4F,$E4,$E3,$FD,$68,$7F,$49,$3F,$D1,$13,$FA,$0C,$3D,$F1
	.dc.b	$FE,$8E,$7F,$40,$61,$FE,$43,$5F,$EB,$D3,$94,$7F,$BC,$0D,$47,$EB
	.dc.b	$23,$FC,$5F,$1F,$C3,$E3,$F5,$87,$C0,$3F,$44,$60,$FF,$17,$CF,$E2
	.dc.b	$5A,$FD,$2A,$FD,$29,$FC,$E5,$FE,$FF,$A7,$F9,$76,$7F,$C6,$4D,$7F
	.dc.b	$29,$8A,$31,$8A,$B1,$5F,$EF,$83,$F9,$E9,$83,$F7,$18,$F9,$2F,$11
	.dc.b	$FE,$FE,$9E,$57,$E1,$AF,$F2,$3B,$6D,$7C,$2F,$F5,$E7,$EA,$A3,$8B
	.dc.b	$E1,$AF,$C7,$D7,$3B,$E2,$FF,$09,$9F,$F1,$BB,$FE,$7B,$AF,$54,$7F
	.dc.b	$59,$8C,$4F,$C0,$6C,$F6,$9F,$5B,$F0,$71,$E1,$1D,$6F,$F7,$57,$F1
	.dc.b	$70,$F4,$4F,$FB,$42,$39,$3F,$E5,$F9,$FC,$26,$1F,$E8,$DB,$FE,$D5
	.dc.b	$87,$A2,$BA,$46,$2F,$9C,$7E,$D0,$F7,$9A,$3C,$15,$EF,$7F,$1F,$1E
	.dc.b	$07,$F7,$B7,$F3,$9E,$AF,$BD,$B7,$B7,$FB,$E4,$7E,$4E,$FD,$F1,$F7
	.dc.b	$1F,$E3,$75,$E7,$8F,$EA,$CF,$E1,$4F,$3C,$FF,$38,$8B,$1F,$89,$AE
	.dc.b	$11,$F0,$E3,$06,$8F,$6B,$FE,$D9,$76,$FB,$4F,$86,$7C,$BB,$C3,$F1
	.dc.b	$DA,$F2,$3E,$25,$FF,$32,$8C,$CF,$BE,$7F,$1A,$74,$3F,$9B,$9B,$3C
	.dc.b	$17,$F7,$A7,$BB,$D6,$3E,$23,$C4,$F8,$51,$E7,$BF,$C7,$CF,$A1,$FC
	.dc.b	$8C,$DA,$E9,$3F,$66,$7F,$1C,$E6,$FC,$4F,$E6,$E3,$E7,$4F,$C2,$9F
	.dc.b	$DC,$CF,$B8,$FB,$37,$F1,$E3,$E2,$46,$5F,$B6,$FF,$76,$BE,$4F,$A0
	.dc.b	$F8,$EF,$E1,$DF,$D5,$47,$C4,$8F,$01,$E8,$8C,$BF,$0D,$FF,$60,$C7
	.dc.b	$F4,$A8,$DD,$7C,$F9,$FC,$25,$7D,$B8,$F8,$6D,$7A,$47,$FA,$70,$FB
	.dc.b	$EF,$EC,$66,$BE,$78,$E1,$3F,$C6,$5F,$69,$F7,$2F,$F1,$2F,$E7,$0F
	.dc.b	$86,$7F,$16,$3F,$8C,$C7,$F2,$19,$DC,$7F,$7E,$7F,$0D,$5C,$63,$F3
	.dc.b	$A7,$F1,$3C,$2C,$7A,$23,$F8,$BB,$F0,$DF,$91,$3D,$1F,$E0,$EF,$4B
	.dc.b	$FC,$2D,$FF,$10,$BF,$D4,$C7,$E9,$23,$F8,$AE,$1E,$CB,$FD,$AC,$FE
	.dc.b	$BE,$BE,$CD,$F9,$AB,$F0,$75,$FB,$73,$F7,$C5,$AD,$3E,$5C,$3F,$BB
	.dc.b	$D7,$96,$BF,$7A,$FE,$E9,$FD,$1C,$D9,$E6,$FF,$1B,$8F,$E6,$93,$97
	.dc.b	$CD,$1F,$81,$8F,$80,$7E,$63,$0F,$E2,$D5,$F8,$29,$FE,$89,$3D,$1F
	.dc.b	$A0,$FC,$B7,$F8,$9B,$FA,$49,$E2,$FC,$17,$F7,$F1,$E0,$3F,$82,$1F
	.dc.b	$C1,$2F,$F2,$0F,$DE,$3F,$83,$DF,$EB,$1F,$E5,$15,$AB,$FD,$04,$7F
	.dc.b	$4D,$BF,$DF,$DE,$6F,$B4,$FE,$76,$7F,$81,$DF,$E7,$6B,$ED,$CF,$EE
	.dc.b	$E3,$F6,$6D,$8D,$47,$E1,$63,$FA,$4D,$78,$DF,$F5,$61,$FD,$06,$3F
	.dc.b	$51,$1F,$72,$3F,$CC,$8F,$F9,$4A,$BF,$7D,$1F,$1C,$FE,$E7,$1F,$0C
	.dc.b	$FF,$08,$C6,$A7,$9D,$FB,$AB,$B9,$FE,$11,$3C,$47,$EE,$9E,$8F,$F4
	.dc.b	$33,$F8,$0D,$69,$FE,$9F,$3F,$B4,$7E,$93,$FB,$C7,$E6,$BC,$8F,$B4
	.dc.b	$E2,$39,$3F,$C7,$23,$F9,$93,$FB,$D3,$FB,$41,$F0,$4F,$DE,$61,$FB
	.dc.b	$27,$E0,$57,$F1,$07,$13,$FA,$F7,$F3,$77,$FE,$0F,$7F,$2B,$3F,$AA
	.dc.b	$AF,$DC,$E1,$FB,$E9,$FE,$DD,$5F,$E0,$B9,$FE,$6D,$7F,$B7,$3E,$44
	.dc.b	$FE,$C9,$FC,$C3,$6B,$58,$76,$7F,$2D,$1F,$CD,$67,$F6,$2F,$8A,$7F
	.dc.b	$C0,$35,$FA,$F8,$F5,$C7,$E5,$DF,$DD,$4F,$ED,$70,$F8,$91,$FD,$F0
	.dc.b	$E4,$F8,$9F,$E6,$17,$FC,$0E,$3F,$7D,$5F,$DE,$2F,$F9,$4D,$FF,$1C
	.dc.b	$3F,$2F,$7F,$E3,$87,$FB,$B4,$7F,$73,$BD,$47,$F0,$73,$D7,$3F,$DB
	.dc.b	$B0,$FD,$8E,$1F,$62,$BF,$5F,$3F,$9D,$3F,$22,$7F,$76,$AF,$D7,$3F
	.dc.b	$AA,$9F,$D2,$BF,$E3,$2A,$FE,$DB,$86,$4F,$C2,$C7,$DA,$3F,$58,$7E
	.dc.b	$FE,$FF,$1C,$FF,$65,$BF,$CB,$C7,$E3,$A3,$13,$F8,$E8,$FE,$25,$1F
	.dc.b	$D9,$CF,$D3,$1E,$23,$F5,$A7,$E8,$23,$E9,$3F,$D9,$EB,$F5,$55,$FD
	.dc.b	$8A,$3F,$5D,$7C,$5F,$D5,$D7,$F2,$4A,$FE,$E5,$3F,$BF,$8F,$E1,$13
	.dc.b	$FD,$BA,$BF,$B2,$1F,$49,$FF,$70,$3E,$C6,$BF,$A9,$3F,$84,$1F,$D8
	.dc.b	$EB,$FB,$43,$F4,$0F,$CD,$47,$E5,$8F,$E7,$37,$FD,$72,$FF,$86,$BA
	.dc.b	$BF,$D1,$C8,$FF,$25,$9F,$E2,$0A,$FD,$25,$73,$3F,$51,$7F,$E7,$78
	.dc.b	$FE,$C3,$5C,$A7,$8D,$7D,$39,$FD,$43,$FA,$37,$F2,$D1,$FE,$68,$BF
	.dc.b	$D3,$CF,$F6,$09,$E2,$FD,$27,$F4,$EF,$E9,$63,$D5,$5F,$DE,$A3,$FA
	.dc.b	$91,$F9,$E3,$89,$F4,$AB,$F4,$EF,$F9,$72,$3F,$AD,$DF,$F9,$C6,$7C
	.dc.b	$71,$F5,$E7,$FC,$CA,$FE,$B2,$FF,$AC,$9C,$A3,$FB,$29,$FC,$29,$FD
	.dc.b	$3D,$7F,$54,$7F,$89,$9F,$91,$8F,$F2,$F3,$E4,$7F,$CE,$D1,$FA,$83
	.dc.b	$FC,$AF,$7F,$A8,$AF,$E9,$B3,$FE,$C3,$9F,$F2,$BD,$7F,$93,$63,$F5
	.dc.b	$07,$E9,$2F,$F9,$AB,$FC,$60,$FF,$2B,$D6,$67,$FA,$7C,$7E,$9D,$FC
	.dc.b	$B3,$F8,$27,$F4,$E7,$F5,$3B,$FC,$64,$7F,$59,$8F,$EA,$2F,$E2,$CF
	.dc.b	$F6,$FD,$7F,$92,$27,$35,$FA,$A3,$F9,$D9,$F8,$BA,$FE,$9F,$5F,$A6
	.dc.b	$33,$5F,$AD,$3F,$48,$7B,$DF,$D2,$47,$F4,$B7,$FA,$64,$7F,$90,$2B
	.dc.b	$F4,$F3,$FA,$23,$84,$FF,$A4,$9F,$D1,$1F,$D1,$5F,$CE,$9F,$90,$3F
	.dc.b	$10,$7F,$A0,$2B,$31,$FD,$1E,$BE,$F1,$FE,$3F,$BF,$F6,$A6,$1F,$D1
	.dc.b	$A7,$F3,$F7,$FE,$40,$7F,$C7,$98,$7F,$0C,$8F,$F6,$3C,$7E,$9E,$3F
	.dc.b	$0B,$1F,$EC,$77,$FC,$78,$D8,$FF,$19,$47,$F2,$C8,$FD,$04,$7F,$43
	.dc.b	$7F,$A3,$1F,$96,$3F,$A2,$9F,$9F,$7F,$A2,$1F,$CF,$DF,$E8,$71,$FA
	.dc.b	$6B,$FE,$96,$7F,$42,$3F,$25,$1F,$56,$7D,$E7,$F1,$17,$FC,$63,$3F
	.dc.b	$C1,$E3,$CF,$1F,$D1,$5F,$E8,$47,$27,$FC,$12,$7F,$78,$3F,$9B,$9F
	.dc.b	$94,$3F,$A1,$1F,$58,$FF,$01,$DF,$FA,$A9,$FC,$DD,$7F,$AA,$6B,$FC
	.dc.b	$5D,$1F,$EA,$BC,$3F,$35,$7F,$EA,$78,$FC,$D1,$FC,$ED,$FC,$CC,$7E
	.dc.b	$74,$C3,$F9,$88,$FE,$7A,$7E,$50,$FE,$78,$7F,$33,$3C,$87,$F3,$83
	.dc.b	$F7,$C7,$F8,$8C,$FE,$7D,$5F,$EA,$0A,$FF,$14,$D7,$F4,$03,$F3,$75
	.dc.b	$F9,$4A,$FF,$10,$BF,$AE,$7E,$74,$FE,$62,$72,$7F,$88,$27,$F1,$13
	.dc.b	$F1,$0F,$96,$7F,$9C,$0F,$CC,$CF,$F3,$97,$F3,$27,$FA,$4E,$7F,$33
	.dc.b	$87,$FB,$46,$BF,$9B,$CF,$F3,$3A,$E1,$7F,$E2,$68,$FE,$0C,$7F,$A8
	.dc.b	$5F,$B5,$1E,$28,$FF,$28,$C2,$3F,$09,$1F,$E1,$DA,$FE,$6D,$1F,$CE
	.dc.b	$30,$FF,$3D,$BF,$99,$9F,$95,$7F,$CE,$27,$FC,$FF,$87,$F8,$79,$B3
	.dc.b	$F9,$A8,$FC,$EC,$FE,$68,$FF,$47,$47,$E6,$AF,$FD,$21,$7F,$E1,$BA
	.dc.b	$FE,$69,$1F,$E1,$AA,$5F,$F3,$1B,$FE,$69,$1F,$D3,$8F,$CC,$BF,$CC
	.dc.b	$AB,$31,$FE,$45,$8F,$E3,$87,$F9,$AC,$FF,$44,$3F,$E5,$28,$FE,$66
	.dc.b	$7E,$71,$FF,$0C,$C7,$F3,$3C,$3F,$31,$1F,$97,$8F,$CC,$47,$D3,$8F
	.dc.b	$CC,$47,$E6,$63,$FE,$00,$61,$FE,$82,$AF,$CC,$47,$F9,$3A,$BF,$DE
	.dc.b	$75,$F9,$88,$FE,$5D,$1F,$EF,$69,$FE,$5D,$1F,$EC,$08,$FF,$09,$4F
	.dc.b	$E5,$AF,$FC,$1F,$7F,$96,$8F,$F3,$BD,$7F,$BC,$AB,$F9,$65,$7F,$BB
	.dc.b	$E7,$F2,$51,$FE,$0D,$BF,$CA,$D7,$CD,$AF,$EA,$27,$FB,$B6,$BF,$CE
	.dc.b	$17,$F9,$37,$F1,$87,$FB,$2C,$FE,$81,$87,$C1,$9F,$F3,$84,$7F,$9B
	.dc.b	$A3,$F9,$54,$7F,$B0,$A3,$F2,$67,$F9,$FA,$BF,$27,$87,$F3,$2B,$FC
	.dc.b	$8C,$FF,$B3,$9F,$F0,$D1,$F7,$1F,$CB,$BF,$DB,$DB,$1F,$C5,$70,$FE
	.dc.b	$3D,$7F,$90,$9F,$E5,$75,$FE,$0A,$7F,$24,$7F,$9C,$6B,$F2,$53,$FE
	.dc.b	$05,$8F,$C9,$3F,$E0,$79,$FC,$94,$7F,$B8,$E7,$FC,$0A,$7F,$99,$9F
	.dc.b	$C8,$C7,$F2,$98,$FF,$01,$9F,$CA,$0F,$C8,$47,$F7,$E3,$F9,$3C,$7F
	.dc.b	$80,$5F,$C8,$5F,$FA,$8A,$7F,$1F,$5F,$C9,$E3,$F1,$EF,$F9,$9A,$7F
	.dc.b	$1E,$FF,$B4,$A7,$F1,$EF,$F0,$A8,$FF,$1E,$57,$F9,$2D,$FF,$5F,$BF
	.dc.b	$25,$FF,$01,$5F,$E4,$A7,$84,$7F,$01,$AF,$E4,$73,$FE,$B1,$8F,$F5
	.dc.b	$71,$FC,$AE,$BF,$C0,$2F,$E5,$6F,$F2,$97,$FE,$60,$8F,$F0,$04,$7F
	.dc.b	$29,$BF,$F6,$4C,$7E,$52,$7F,$CC,$B1,$FC,$A6,$3F,$BF,$C7,$8A,$3C
	.dc.b	$D1,$FD,$5E,$FF,$2B,$3F,$E6,$09,$F8,$C7,$F7,$F9,$FC,$AE,$1F,$96
	.dc.b	$8F,$F6,$35,$7F,$26,$AF,$8C,$7E,$50,$FE,$FA,$7F,$96,$CF,$CA,$5F
	.dc.b	$E6,$CF,$C9,$CF,$F9,$6A,$FF,$90,$DF,$F2,$28,$FF,$2C,$5F,$F9,$75
	.dc.b	$47,$C2,$9F,$E7,$74,$AF,$95,$08,$C3,$FD,$F7,$05,$FF,$B0,$EB,$F9
	.dc.b	$2D,$FE,$E7,$0F,$CA,$5F,$F2,$48,$FE,$4B,$5F,$C8,$A3,$FC,$B2,$7E
	.dc.b	$56,$3F,$74,$F8,$A3,$F5,$78,$76,$7F,$93,$1F,$96,$BF,$E4,$B5,$FD
	.dc.b	$EE,$BF,$BE,$57,$FA,$F6,$BF,$92,$47,$E5,$AF,$F8,$2B,$F8,$89,$FE
	.dc.b	$56,$7F,$61,$7F,$DC,$D1,$F9,$73,$FD,$7B,$1F,$C1,$63,$FD,$76,$7F
	.dc.b	$25,$8F,$C8,$E1,$F8,$08,$FC,$96,$1F,$95,$7F,$CA,$97,$FC,$99,$FE
	.dc.b	$4F,$7F,$C9,$AF,$F3,$35,$FE,$B8,$9F,$EF,$8F,$F8,$06,$BF,$CA,$71
	.dc.b	$FC,$9C,$FD,$34,$7F,$A0,$5F,$EE,$2F,$F1,$47,$FD,$14,$FF,$7E,$8F
	.dc.b	$E5,$31,$FA,$D7,$F5,$8F,$FA,$A4,$F9,$97,$FD,$FD,$FA,$87,$F7,$C8
	.dc.b	$FF,$36,$BF,$E5,$77,$F3,$35,$FC,$B4,$FE,$49,$1F,$E5,$73,$F9,$6D
	.dc.b	$7E,$65,$FF,$5A,$4F,$47,$F9,$41,$FD,$FF,$0F,$E5,$AF,$E4,$0F,$51
	.dc.b	$FE,$BD,$C2,$4F,$F9,$60,$FE,$5C,$7F,$95,$CF,$F5,$AE,$1F,$B0,$8F
	.dc.b	$A3,$3F,$EB,$59,$FC,$EE,$1F,$EE,$23,$F9,$64,$7F,$AD,$6F,$FD,$6D
	.dc.b	$3F,$99,$8F,$E5,$91,$F9,$97,$F4,$78,$7F,$AC,$9B,$7F,$EB,$23,$FC
	.dc.b	$AF,$1F,$DF,$AF,$F7,$C7,$F7,$D9,$F8,$07,$F9,$5A,$AB,$EC,$BF,$CD
	.dc.b	$38,$3F,$5C,$FF,$85,$5F,$F5,$89,$FC,$78,$FC,$B1,$FE,$2C,$3F,$C8
	.dc.b	$53,$97,$FC,$8A,$7F,$BC,$9F,$F2,$C3,$FE,$01,$79,$5F,$FA,$C8,$FD
	.dc.b	$DB,$F9,$E3,$F8,$7C,$7E,$79,$FC,$EC,$FE,$97,$0F,$CF,$1F,$D6,$67
	.dc.b	$F9,$9D,$7F,$80,$EB,$F3,$71,$FE,$B1,$C3,$FC,$2C,$FE,$58,$FF,$2D
	.dc.b	$BF,$E0,$48,$FA,$91,$F9,$C9,$E0,$7E,$CE,$7F,$3D,$1F,$E7,$87,$F0
	.dc.b	$AF,$FC,$01,$7F,$8A,$1F,$9F,$9F,$F0,$34,$7F,$34,$3F,$41,$7F,$E0
	.dc.b	$77,$E7,$1F,$C2,$E7,$FC,$17,$1F,$EB,$29,$FC,$E4,$FE,$4C,$FC,$E5
	.dc.b	$FF,$1E,$8F,$CF,$D7,$E7,$67,$FD,$D5,$1F,$CD,$5F,$CD,$5F,$FA,$E4
	.dc.b	$FF,$2D,$1F,$CD,$6F,$F3,$47,$F3,$59,$FF,$29,$BF,$9A,$3F,$AD,$38
	.dc.b	$AF,$E6,$D8,$7F,$35,$BF,$F4,$CD,$FF,$32,$7F,$C0,$B1,$F9,$AA,$FF
	.dc.b	$2B,$D7,$F3,$27,$FA,$BD,$FF,$95,$AB,$F9,$94,$7F,$81,$63,$F2,$17
	.dc.b	$FC,$FC,$FF,$58,$3F,$E0,$63,$27,$F3,$23,$FD,$8B,$3F,$EA,$DC,$3E
	.dc.b	$23,$FC,$DE,$3F,$C1,$18,$7F,$33,$BF,$CA,$4F,$F8,$22,$7F,$99,$C7
	.dc.b	$E5,$F0,$FF,$2B,$47,$D4,$9F,$D7,$1F,$D7,$27,$F2,$F7,$FE,$AE,$7F
	.dc.b	$2D,$1F,$C3,$1F,$CA,$BF,$EA,$EC,$3F,$CA,$F5,$F7,$AB,$F2,$D3,$FE
	.dc.b	$0C,$8F,$D2,$B5,$FF,$58,$3F,$CD,$2B,$F3,$33,$FE,$06,$7F,$37,$7F
	.dc.b	$CC,$EB,$FC,$B3,$1F,$E0,$7B,$FF,$03,$3F,$CE,$9F,$D5,$DF,$E6,$EB
	.dc.b	$F2,$8F,$E4,$8F,$F5,$79,$FE,$5A,$BF,$E6,$15,$FE,$09,$BF,$2C,$FF
	.dc.b	$5F,$7F,$28,$FE,$4C,$FF,$56,$E1,$FE,$07,$9F,$C9,$C7,$DA,$3F,$CB
	.dc.b	$6F,$EE,$9F,$F5,$41,$FC,$BD,$FF,$04,$3F,$96,$B3,$C7,$0F,$F0,$4D
	.dc.b	$FE,$13,$0F,$EE,$27,$E0,$27,$FA,$41,$FE,$5D,$8F,$F0,$5B,$FC,$BF
	.dc.b	$0F,$CB,$3A,$AF,$CB,$3F,$CB,$F0,$FF,$55,$DF,$E2,$23,$F3,$78,$7F
	.dc.b	$61,$7F,$A0,$3E,$A7,$F8,$26,$1B,$7A,$47,$E1,$9B,$7F,$CC,$EB,$FC
	.dc.b	$BF,$7F,$EB,$A8,$FF,$80,$17,$FE,$C7,$AF,$E6,$07,$F9,$83,$0F,$CE
	.dc.b	$C7,$F3,$2A,$FF,$29,$57,$D1,$9F,$CD,$57,$F2,$D8,$FF,$06,$4F,$E6
	.dc.b	$63,$FC,$18,$7F,$35,$AF,$F3,$0B,$FE,$0D,$3F,$C5,$11,$F9,$A8,$FF
	.dc.b	$58,$3F,$9A,$8F,$E1,$27,$F2,$F3,$FC,$C5,$7F,$DD,$9F,$CC,$3F,$EB
	.dc.b	$17,$F3,$93,$F8,$68,$F6,$47,$F2,$33,$F9,$81,$FC,$A4,$FE,$5E,$7F
	.dc.b	$97,$E7,$F9,$7B,$FE,$08,$AF,$E5,$EF,$98,$FE,$65,$3F,$EB,$39,$FC
	.dc.b	$E6,$1F,$9A,$9F,$F0,$25,$FF,$82,$67,$FD,$69,$7F,$CB,$23,$FC,$44
	.dc.b	$7E,$C2,$7F,$97,$47,$F8,$F1,$FC,$F4,$FF,$2E,$7F,$8D,$9F,$EE,$F3
	.dc.b	$F9,$85,$FD,$F7,$FD,$FA,$7E,$14,$FF,$29,$1F,$85,$3E,$49,$F9,$E3
	.dc.b	$FD,$71,$1F,$9E,$7A,$CF,$F3,$1C,$3F,$AB,$47,$F9,$7A,$BF,$3B,$7F
	.dc.b	$CC,$5F,$04,$7F,$84,$5F,$E1,$35,$FE,$8D,$9F,$F0,$5C,$FF,$8C,$8F
	.dc.b	$F2,$A5,$FF,$83,$0F,$CE,$DF,$F3,$38,$FF,$04,$9F,$C1,$0F,$CF,$9F
	.dc.b	$EB,$FA,$E2,$FF,$83,$0F,$F5,$C9,$FE,$0E,$BF,$CD,$BF,$E1,$0B,$FE
	.dc.b	$69,$7F,$E0,$D3,$F2,$EF,$ED,$63,$FC,$79,$5F,$E3,$C8,$FC,$EE,$1F
	.dc.b	$E6,$6A,$FE,$67,$5F,$E6,$67,$FC,$11,$7F,$EB,$B7,$F9,$A5,$FF,$8F
	.dc.b	$A7,$F3,$47,$F8,$24,$FF,$34,$4F,$F3,$37,$FC,$D5,$5F,$89,$7E,$99
	.dc.b	$FC,$18,$FF,$5A,$1F,$CD,$6B,$FC,$17,$3F,$9B,$7F,$C1,$B3,$C6,$78
	.dc.b	$4F,$E7,$1F,$E1,$33,$FE,$0C,$7F,$D7,$57,$FD,$EA,$BF,$09,$1F,$E9
	.dc.b	$47,$F9,$B4,$FF,$AE,$E3,$FC,$D0,$7E,$7E,$BF,$38,$FF,$35,$BF,$F1
	.dc.b	$F3,$F9,$C3,$F9,$AB,$F9,$BB,$FF,$04,$57,$F3,$5B,$FF,$03,$C7,$C5
	.dc.b	$AF,$CC,$9F,$A3,$3F,$33,$87,$F3,$57,$FC,$15,$3F,$E0,$77,$FB,$35
	.dc.b	$7F,$81,$DF,$CA,$E1,$F8,$C6,$DF,$F3,$3A,$FE,$6B,$5F,$EB,$E8,$FF
	.dc.b	$04,$9F,$CD,$4F,$F3,$15,$71,$AF,$E6,$0F,$F9,$8C,$FC,$C9,$FE,$4B
	.dc.b	$7F,$83,$47,$E7,$9F,$F1,$49,$E9,$3F,$2B,$1C,$DF,$C5,$D7,$F9,$3F
	.dc.b	$0F,$9D,$1F,$E6,$6B,$FF,$06,$9F,$EB,$E3,$F9,$85,$71,$8F,$E7,$73
	.dc.b	$FE,$66,$7F,$D7,$D3,$FE,$66,$7F,$C1,$07,$F3,$37,$F3,$77,$FC,$CE
	.dc.b	$3F,$C1,$8F,$37,$F1,$87,$E6,$4F,$F3,$25,$7E,$65,$FF,$33,$5F,$F8
	.dc.b	$32,$BF,$CC,$27,$F9,$82,$3F,$1C,$7F,$8A,$A7,$F3,$4F,$F4,$78,$FE
	.dc.b	$D0,$FE,$4A,$BE,$23,$4E,$27,$FA,$A6,$7F,$CC,$71,$FB,$D8,$E1,$5A
	.dc.b	$7F,$C1,$D1,$FE,$C1,$AF,$CD,$5F,$FB,$06,$7F,$85,$57,$F9,$8D,$FE
	.dc.b	$19,$5F,$EC,$0A,$FF,$07,$1F,$E0,$DB,$FC,$CC,$FF,$39,$AF,$E6,$8F
	.dc.b	$EB,$EB,$FD,$E6,$FE,$6A,$BF,$C1,$95,$FE,$0E,$8F,$F3,$1D,$7E,$6A
	.dc.b	$7F,$DC,$71,$F9,$A3,$FC,$1A,$FF,$AF,$27,$F0,$D7,$FE,$C7,$AF,$CD
	.dc.b	$57,$F9,$4A,$F2,$FF,$73,$3F,$9B,$CF,$F9,$86,$B2,$FF,$31,$BF,$F4
	.dc.b	$5D,$7F,$98,$6B,$F9,$93,$FC,$C4,$FF,$05,$C7,$F9,$91,$FC,$04,$2F
	.dc.b	$FC,$2F,$1E,$47,$FC,$27,$3F,$CC,$0F,$F0,$6C,$7F,$AE,$DF,$EF,$71
	.dc.b	$FD,$1E,$7F,$D1,$55,$FC,$C1,$FF,$08,$BF,$79,$FC,$FB,$FE,$0B,$AF
	.dc.b	$F3,$2C,$FF,$A4,$23,$FC,$CD,$1F,$CC,$A7,$F3,$27,$F8,$2A,$FF,$2D
	.dc.b	$3F,$CC,$6F,$F3,$77,$FC,$C6,$BF,$CC,$71,$FE,$64,$9F,$F0,$5C,$7E
	.dc.b	$60,$FF,$05,$9F,$D9,$8F,$E6,$31,$F9,$98,$FF,$33,$1F,$EB,$99,$FE
	.dc.b	$0B,$3F,$CC,$A3,$37,$FB,$D8,$FE,$62,$7F,$82,$0F,$CA,$0B,$FC,$CD
	.dc.b	$72,$8F,$D6,$C7,$FC,$00,$7F,$45,$3F,$E6,$6B,$FC,$C9,$FE,$66,$7F
	.dc.b	$99,$47,$E8,$4F,$F2,$2D,$7F,$AD,$6F,$FC,$87,$18,$7F,$C1,$E7,$F9
	.dc.b	$A1,$FF,$33,$1F,$CE,$DF,$F3,$0C,$7F,$B8,$A7,$F9,$94,$FE,$6C,$FF
	.dc.b	$5D,$DF,$E6,$DF,$F0,$9C,$FE,$6D,$FF,$19,$4F,$F9,$8D,$FB,$87,$F9
	.dc.b	$7A,$BF,$71,$87,$F3,$58,$FE,$73,$5F,$9C,$7F,$CB,$15,$F9,$87,$FC
	.dc.b	$D0,$7F,$AF,$2F,$F3,$73,$F1,$E3,$FC,$BC,$7F,$BE,$2B,$FC,$1A,$FF
	.dc.b	$0D,$AF,$E6,$B1,$FC,$DE,$BF,$56,$7F,$34,$81,$F9,$EA,$FC,$EB,$F9
	.dc.b	$0A,$F4,$9F,$CD,$A7,$F1,$B3,$FC,$CC,$FF,$12,$5F,$E9,$67,$FC,$19
	.dc.b	$5F,$A5,$8E,$27,$FA,$F6,$B8,$BF,$CC,$F0,$FF,$33,$4F,$E6,$EB,$F7
	.dc.b	$8F,$03,$FC,$18,$FC,$97,$CF,$5F,$C1,$1F,$F1,$29,$FD,$36,$3F,$C0
	.dc.b	$13,$F9,$D3,$FC,$1B,$1F,$EC,$08,$FE,$6B,$5F,$DA,$5F,$F7,$74,$FE
	.dc.b	$74,$FF,$33,$1F,$CF,$9F,$EA,$27,$E8,$2F,$F3,$2F,$E1,$63,$FD,$82
	.dc.b	$FE,$76,$7F,$D8,$31,$FE,$0C,$8F,$F3,$3C,$FE,$82,$FF,$C1,$8F,$F0
	.dc.b	$88,$FF,$1D,$1F,$E0,$C9,$FF,$5F,$BF,$E6,$73,$F9,$9D,$FF,$B0,$1F
	.dc.b	$24,$FF,$94,$AF,$F3,$51,$FC,$CC,$FC,$D4,$7F,$34,$7F,$D4,$F3,$FC
	.dc.b	$D1,$FC,$DC,$FF,$83,$6F,$F3,$51,$FC,$E6,$7F,$D8,$27,$F8,$85,$FC
	.dc.b	$1D,$FF,$99,$DF,$C7,$9F,$B9,$C3,$F3,$4E,$30,$FC,$D3,$FE,$42,$AF
	.dc.b	$CD,$3F,$E6,$43,$F3,$4F,$F8,$35,$FF,$60,$3F,$E0,$CA,$FF,$50,$DF
	.dc.b	$F9,$90,$FE,$36,$7F,$82,$E3,$F7,$D3,$FE,$65,$7F,$33,$7F,$BE,$8F
	.dc.b	$F5,$C9,$F9,$0B,$FF,$3E,$3F,$E0,$C9,$FE,$68,$FF,$99,$1F,$D0,$61
	.dc.b	$FC,$CE,$3F,$D7,$67,$F9,$88,$FE,$6B,$7F,$E6,$27,$F9,$BC,$FE,$5D
	.dc.b	$FD,$BB,$F9,$73,$E2,$61,$FE,$0A,$9F,$CB,$47,$A2,$7F,$86,$3F,$D3
	.dc.b	$1F,$C1,$1F,$8F,$3F,$83,$4F,$E6,$0F,$F5,$F1,$FE,$64,$3F,$B3,$BF
	.dc.b	$98,$3F,$D7,$D5,$FC,$D2,$BF,$C1,$71,$FE,$16,$8F,$F7,$DB,$F9,$73
	.dc.b	$FD,$7C,$7F,$2E,$AE,$EF,$FB,$01,$FE,$68,$7E,$61,$FC,$CC,$7F,$0C
	.dc.b	$7F,$CD,$97,$E8,$69,$FE,$BF,$AF,$E8,$11,$FE,$64,$3F,$99,$5F,$FB
	.dc.b	$02,$7F,$B0,$9F,$C6,$4F,$E6,$87,$F4,$48,$57,$8A,$7F,$D8,$33,$F9
	.dc.b	$C8,$FF,$33,$BF,$97,$3F,$9A,$1F,$EC,$4A,$FC,$DB,$FA,$0B,$FC,$D9
	.dc.b	$F7,$5F,$F6,$4E,$BF,$EE,$E4,$40,$FF,$84,$71,$B1,$FE,$C4,$9A,$FF
	.dc.b	$38,$29,$FA,$45,$18,$8F,$E3,$73,$DE,$FB,$36,$F6,$DB,$E1,$5F,$D1
	.dc.b	$CF,$F3,$2C,$7F,$9C,$6F,$31,$FE,$3C,$8F,$D3,$CF,$48,$FF,$0D,$D7
	.dc.b	$46,$C7,$0B,$FB,$B8,$FC,$0B,$FF,$0F,$61,$D2,$FF,$9D,$5F,$F4,$49
	.dc.b	$FF,$0C,$C7,$53,$FC,$EB,$3C,$EB,$E4,$E1,$B9,$FF,$67,$C7,$57,$F5
	.dc.b	$D3,$D6,$3F,$B6,$57,$F8,$93,$0E,$95,$F4,$4F,$C0,$CF,$E0,$AF,$D5
	.dc.b	$1F,$B2,$C3,$C1,$5F,$D3,$EF,$FA,$4E,$1E,$99,$FE,$B6,$78,$EB,$FA
	.dc.b	$59,$E3,$3F,$A6,$DF,$47,$F7,$2E,$DF,$DD,$E1,$B7,$F7,$B3,$B7,$F6
	.dc.b	$D7,$FE,$29,$3F,$A6,$9F,$B9,$7F,$17,$3C,$5F,$F4,$3B,$F9,$C8,$D4
	.dc.b	$79,$9F,$F4,$23,$F4,$4F,$E7,$8F,$3C,$3F,$66,$D9,$E9,$3F,$1E,$7F
	.dc.b	$CA,$87,$13,$A3,$FB,$B3,$67,$D6,$AF,$F0,$95,$7F,$B8,$9F,$C0,$5F
	.dc.b	$F8,$0A,$F7,$5F,$86,$BC,$D7,$48,$F5,$9F,$60,$FF,$17,$9D,$0F,$EA
	.dc.b	$A7,$A0,$FF,$46,$9F,$81,$3C,$F7,$FC,$7E,$BF,$C1,$EF,$FB,$9C,$C5
	.dc.b	$73,$9F,$F7,$45,$7E,$26,$3F,$01,$1F,$EE,$99,$DC,$FD,$98,$FE,$00
	.dc.b	$7A,$9F,$F4,$9C,$FE,$15,$F2,$47,$F8,$E0,$C5,$7D,$57,$FC,$01,$1F
	.dc.b	$8A,$3E,$A3,$F0,$4F,$EC,$27,$CC,$9F,$F1,$DD,$7F,$A5,$E7,$F5,$D3
	.dc.b	$F8,$22,$9F,$D5,$A3,$EE,$47,$9E,$3F,$7D,$3F,$2D,$B7,$F8,$F6,$CF
	.dc.b	$49,$F8,$31,$FA,$73,$FC,$05,$1B,$7F,$82,$1F,$E9,$C3,$CD,$3F,$E2
	.dc.b	$8C,$3F,$81,$3F,$E2,$A9,$A7,$49,$FF,$7C,$BE,$77,$FC,$23,$7F,$C0
	.dc.b	$9F,$81,$87,$F0,$0B,$FE,$05,$1F,$30,$FC,$A4,$7C,$19,$FE,$13,$5E
	.dc.b	$59,$FF,$4C,$BF,$BA,$79,$E1,$FE,$0B,$9F,$F7,$CC,$FE,$FA,$B2,$F1
	.dc.b	$3E,$A4,$7F,$BD,$AB,$E9,$47,$B6,$3F,$B8,$3F,$C7,$CF,$CE,$45,$7D
	.dc.b	$8F,$EC,$A3,$F7,$4F,$FB,$EA,$3E,$FC,$FE,$8A,$3F,$D1,$53,$FB,$89
	.dc.b	$FB,$11,$FD,$1E,$BE,$C3,$FE,$D0,$9F,$DC,$61,$A8,$FF,$09,$C7,$EE
	.dc.b	$67,$F6,$D5,$9A,$FF,$1A,$4F,$F8,$05,$FC,$BB,$FB,$53,$FC,$94,$66
	.dc.b	$3F,$53,$5F,$DD,$DE,$D7,$FC,$A6,$7F,$DF,$A7,$FA,$94,$FF,$25,$9F
	.dc.b	$AF,$9E,$27,$EE,$A7,$8D,$FA,$A3,$F8,$59,$FA,$33,$F6,$33,$FD,$C8
	.dc.b	$FD,$AD,$66,$3F,$DB,$75,$EF,$3D,$6F,$EA,$9F,$DD,$3F,$E4,$DB,$D4
	.dc.b	$7E,$3E,$3F,$BB,$4F,$37,$FB,$C4,$7C,$97,$F8,$29,$FE,$1B,$3D,$C5
	.dc.b	$3F,$10,$7F,$75,$AF,$EF,$8F,$53,$FB,$51,$F8,$03,$FC,$49,$1F,$6A
	.dc.b	$F3,$3F,$DC,$F0,$FD,$76,$19,$BF,$EE,$27,$86,$B9,$3F,$D5,$E3,$F2
	.dc.b	$6F,$EB,$CF,$D5,$BF,$B3,$8F,$F1,$5B,$DC,$FC,$9D,$7F,$70,$3E,$A5
	.dc.b	$F8,$F0,$F6,$DF,$F6,$E9,$FE,$E3,$59,$8F,$61,$FD,$98,$F1,$9F,$E1
	.dc.b	$99,$FE,$DA,$FF,$90,$5F,$F1,$5B,$FA,$E8,$F2,$47,$F6,$DA,$FD,$74
	.dc.b	$FF,$6C,$7F,$5C,$7F,$1F,$7F,$CF,$4F,$E7,$AF,$FB,$6D,$7B,$63,$F8
	.dc.b	$55,$FF,$A9,$E7,$FB,$8D,$FE,$D2,$7C,$B7,$FC,$5C,$FE,$56,$7D,$F3
	.dc.b	$F8,$D5,$FD,$F7,$DA,$7A,$EB,$F9,$8C,$FE,$0D,$FC,$B4,$7F,$66,$3F
	.dc.b	$CD,$0F,$F1,$17,$F9,$23,$C9,$FF,$05,$BF,$B0,$8F,$EB,$71,$FD,$7A
	.dc.b	$7F,$5E,$FF,$98,$5F,$DA,$47,$89,$FE,$C9,$3F,$D8,$A3,$F5,$F3,$FA
	.dc.b	$7A,$FE,$05,$7F,$AF,$8F,$EB,$EF,$EB,$E3,$DD,$1F,$D0,$E7,$F5,$EF
	.dc.b	$F6,$E8,$F5,$3F,$AE,$3F,$D7,$B5,$F0,$A3,$D3,$5F,$D7,$AB,$FA,$D5
	.dc.b	$7F,$5D,$7F,$5A,$FF,$97,$EF,$F5,$B5,$93,$C3,$1F,$D6,$E7,$E5,$9F
	.dc.b	$D5,$1F,$EB,$47,$F3,$BA,$FE,$B4,$7E,$AA,$73,$1F,$68,$FF,$79,$4F
	.dc.b	$F5,$37,$8D,$FE,$C9,$FA,$27,$E3,$6B,$FC,$C1,$5A,$AF,$E8,$57,$FA
	.dc.b	$8A,$FA,$EF,$F0,$86,$C7,$37,$FA,$9C,$7E,$A6,$3F,$43,$7F,$E9,$67
	.dc.b	$FA,$65,$7F,$06,$AF,$C0,$BF,$AA,$7F,$A9,$9C,$CF,$F7,$DD,$7E,$AE
	.dc.b	$3F,$AD,$C7,$B5,$FD,$4B,$CA,$BF,$41,$1F,$E5,$3C,$3F,$CA,$35,$FB
	.dc.b	$79,$F1,$CF,$FA,$62,$FF,$52,$7F,$0D,$3D,$97,$FE,$4E,$BF,$EA,$57
	.dc.b	$99,$F5,$9F,$E5,$88,$FA,$B8,$7F,$1A,$AF,$A3,$5F,$C2,$DF,$F1,$D4
	.dc.b	$7D,$57,$FC,$1B,$5F,$DB,$5F,$49,$FE,$45,$3F,$80,$9F,$C7,$2B,$F8
	.dc.b	$EC,$7F,$82,$EF,$EA,$47,$E7,$63,$FC,$07,$3F,$E0,$87,$E4,$3F,$EE
	.dc.b	$FB,$FA,$55,$99,$FA,$33,$FC,$1C,$FF,$78,$9F,$32,$7F,$87,$BF,$B5
	.dc.b	$3F,$DE,$B3,$F4,$27,$EC,$E1,$C6,$7F,$6E,$7F,$17,$BF,$9F,$3F,$C5
	.dc.b	$B0,$FF,$3F,$9F,$EE,$77,$F1,$53,$FE,$F2,$3F,$88,$61,$FC,$56,$3C
	.dc.b	$11,$F5,$5F,$9E,$7E,$6D,$FC,$31,$FC,$66,$7F,$CF,$6F,$F9,$EF,$0F
	.dc.b	$F0,$13,$FD,$A0,$FE,$B9,$1F,$E7,$83,$FC,$EC,$FE,$B8,$FD,$01,$FC
	.dc.b	$52,$3F,$C1,$58,$71,$8F,$F6,$19,$FE,$75,$3F,$89,$BF,$E6,$EB,$FE
	.dc.b	$23,$3F,$C5,$6F,$85,$7F,$B5,$67,$DE,$FF,$9C,$0F,$E1,$47,$F0,$23
	.dc.b	$FA,$11,$F0,$04,$FF,$8C,$DF,$F3,$AB,$FC,$10,$FD,$43,$FE,$F4,$3F
	.dc.b	$CD,$C6,$6B,$94,$7F,$62,$8F,$F4,$DC,$79,$1F,$F4,$FB,$F6,$5F,$F5
	.dc.b	$89,$FC,$54,$FE,$23,$5F,$D1,$9F,$F0,$24,$FF,$15,$9F,$F3,$A4,$FF
	.dc.b	$41,$9F,$9E,$7C,$B3,$D4,$7A,$4F,$F4,$2D,$FF,$8C,$E3,$FD,$41,$1F
	.dc.b	$38,$FA,$53,$C0,$FE,$84,$79,$DF,$E3,$33,$FE,$2B,$7E,$71,$FC,$9A
	.dc.b	$FC,$CF,$F5,$77,$FB,$BB,$FD,$64,$FF,$52,$47,$F8,$CE,$BF,$A1,$C7
	.dc.b	$D2,$3E,$61,$F2,$0F,$E7,$47,$86,$FE,$7B,$FE,$DF,$7F,$6B,$3F,$EA
	.dc.b	$08,$FA,$15,$FE,$34,$9F,$E8,$D7,$C4,$F9,$D7,$FC,$32,$BF,$D3,$F1
	.dc.b	$FE,$9D,$AF,$E8,$31,$FE,$9C,$8F,$9B,$3F,$DC,$23,$FB,$8C,$7F,$A7
	.dc.b	$27,$FB,$85,$7E,$78,$FA,$15,$FD,$BE,$BE,$65,$7F,$B3,$6B,$FB,$95
	.dc.b	$7E,$70,$F9,$F5,$FD,$A2,$BE,$5D,$7E,$6E,$3F,$D9,$33,$F9,$B3,$E7
	.dc.b	$3F,$87,$7F,$63,$3F,$98,$3E,$4D,$7F,$6E,$8F,$CE,$9F,$32,$7F,$B4
	.dc.b	$CF,$E5,$AB,$FD,$89,$5F,$1E,$3F,$D8,$D5,$F9,$9A,$FF,$27,$C7,$E4
	.dc.b	$EB,$E3,$47,$FB,$02,$3E,$35,$FF,$0B,$8F,$CD,$47,$90,$FF,$05,$C7
	.dc.b	$F0,$47,$FC,$D4,$FF,$BE,$AB,$E3,$CF,$E4,$A7,$F2,$75,$F1,$E3,$FD
	.dc.b	$E9,$3F,$1E,$3F,$C2,$55,$F9,$1A,$FE,$02,$FF,$9F,$DF,$C7,$5F,$E4
	.dc.b	$1F,$C9,$5F,$E6,$63,$F8,$0C,$7C,$73,$FD,$D8,$7C,$88,$FF,$77,$5F
	.dc.b	$C8,$BF,$F7,$84,$7C,$8B,$FC,$74,$7C,$88,$F8,$F5,$F2,$23,$FD,$DD
	.dc.b	$1F,$22,$3F,$C1,$F1,$F2,$27,$E3,$57,$EE,$5F,$91,$5F,$B9,$AF,$C6
	.dc.b	$C7,$C7,$7F,$51,$3F,$8D,$AF,$EB,$B1,$FE,$38,$3F,$1B,$3F,$EA,$8A
	.dc.b	$F8,$F1,$F2,$27,$E3,$BF,$C9,$E7,$F7,$11,$F8,$D7,$FD,$4B,$1F,$22
	.dc.b	$7E,$31,$F8,$D7,$FC,$6C,$7F,$25,$3F,$DE,$97,$F8,$D7,$E3,$D7,$C6
	.dc.b	$9F,$E4,$31,$F8,$DA,$FE,$00,$FF,$2A,$9F,$E4,$67,$F8,$D4,$FD,$A4
	.dc.b	$FC,$79,$FC,$54,$FF,$20,$8F,$E6,$47,$DB,$7F,$DF,$0F,$F9,$32,$3E
	.dc.b	$4B,$FB,$57,$FD,$A7,$1F,$EB,$F7,$F1,$CF,$FB,$52,$7E,$4B,$FE,$9E
	.dc.b	$8F,$EA,$31,$F8,$F7,$CF,$1F,$CD,$CF,$F4,$14,$7C,$A7,$F1,$67,$ED
	.dc.b	$A3,$F2,$0F,$E3,$AF,$F6,$D3,$FE,$03,$9F,$F6,$44,$FF,$1D,$84,$7F
	.dc.b	$97,$E3,$FC,$F9,$3F,$CD,$51,$FC,$74,$FE,$40,$7E,$FE,$7F,$CD,$07
	.dc.b	$D3,$7E,$65,$FE,$E6,$BF,$D9,$B3,$F9,$2B,$FF,$2F,$D7,$E4,$A7,$F6
	.dc.b	$D5,$FC,$79,$FC,$9D,$7F,$1E,$7F,$A2,$1F,$E2,$99,$F3,$9F,$EC,$79
	.dc.b	$FF,$15,$3F,$E0,$49,$FE,$CF,$3F,$EC,$8A,$FF,$15,$47,$F8,$4C,$FF
	.dc.b	$62,$CF,$FB,$06,$BF,$C6,$51,$C6,$3E,$71,$F0,$0F,$F5,$7B,$FE,$67
	.dc.b	$8F,$DB,$C7,$FA,$BC,$FF,$16,$3F,$EC,$53,$FA,$7B,$C4,$FF,$5A,$5F
	.dc.b	$F8,$B7,$0F,$CA,$61,$FE,$AD,$8F,$E2,$97,$FD,$F2,$BF,$C6,$A7,$FB
	.dc.b	$0E,$FF,$CB,$8F,$C0,$3F,$CC,$67,$F9,$76,$3D,$93,$FD,$42,$3F,$0B
	.dc.b	$3F,$EC,$49,$F8,$93,$EC,$9F,$B7,$1F,$A1,$C3,$FC,$BA,$75,$8F,$F1
	.dc.b	$81,$FE,$5C,$C3,$FC,$BB,$19,$8F,$F6,$4D,$7F,$B0,$E3,$FC,$AD,$1F
	.dc.b	$EB,$A3,$F0,$47,$F9,$70,$FF,$18,$1F,$9F,$C3,$ED,$BF,$C1,$CF,$E5
	.dc.b	$AF,$F0,$73,$F3,$F7,$FD,$FA,$BF,$A9,$D7,$FA,$FA,$BF,$96,$9F,$EB
	.dc.b	$D7,$FA,$9B,$FD,$52,$7F,$3C,$FF,$16,$3F,$83,$9F,$CA,$EB,$FD,$74
	.dc.b	$FF,$06,$7F,$DE,$F3,$FC,$19,$FE,$56,$7F,$06,$AF,$F2,$0C,$FF,$89
	.dc.b	$9F,$EF,$27,$F5,$39,$FC,$DC,$FF,$06,$AF,$E5,$95,$FE,$24,$3F,$A9
	.dc.b	$9F,$E0,$F9,$FC,$4B,$F8,$D4,$7E,$72,$BF,$05,$1F,$58,$FC,$55,$7E
	.dc.b	$CA,$33,$1F,$E8,$2C,$18,$71,$9F,$F4,$34,$A7,$AD,$71,$9C,$9F,$BC
	.dc.b	$71,$3F,$9F,$C3,$C9,$7C,$E3,$FC,$BB,$5E,$09,$FD,$4B,$63,$FD,$55
	.dc.b	$5C,$9F,$44,$F8,$70,$E5,$18,$AF,$03,$FE,$5A,$7F,$D3,$93,$F0,$AF
	.dc.b	$73,$FC,$43,$0F,$8B,$7E,$A8,$E0,$79,$E3,$E4,$BD,$2F,$CD,$1F,$30
	.dc.b	$F8,$B8,$79,$65,$87,$96,$7A,$57,$0B,$FD,$95,$EE,$BF,$DA,$98,$79
	.dc.b	$30,$FD,$B5,$70,$BF,$F6,$9C,$78,$5F,$F6,$AE,$1E,$08,$F3,$9F,$B0
	.dc.b	$8F,$D2,$1F,$2C,$FF,$3D,$15,$F8,$58,$79,$DF,$DB,$57,$C4,$C3,$FA
	.dc.b	$C4,$77,$3E,$C5,$7D,$27,$B9,$EE,$7F,$D3,$27,$78,$F6,$C6,$9F,$E0
	.dc.b	$58,$7B,$6B,$E7,$9D,$CF,$9B,$35,$FC,$64,$5C,$3E,$14,$FF,$04,$35
	.dc.b	$3F,$06,$FF,$2F,$1F,$C8,$CF,$94,$7E,$3F,$0F,$73,$F1,$23,$F7,$11
	.dc.b	$88,$FE,$EC,$D9,$EF,$1F,$9E,$8E,$38,$7F,$1C,$6D,$EA,$7E,$29,$E0
	.dc.b	$9E,$6F,$C6,$3F,$D1,$25,$8F,$7C,$FD,$17,$F1,$B3,$F8,$17,$FD,$41
	.dc.b	$7E,$C7,$FD,$16,$DB,$EA,$D7,$C5,$7F,$B8,$BF,$B9,$3F,$06,$3C,$4F
	.dc.b	$C1,$7F,$B7,$3F,$75,$FF,$60,$BF,$98,$8E,$31,$FA,$47,$9D,$FF,$57
	.dc.b	$6C,$6C,$EB,$1F,$B3,$8D,$CF,$59,$FF,$5F,$BB,$3E,$B9,$FE,$ED,$8F
	.dc.b	$C9,$57,$C4,$9F,$90,$FC,$48,$CB,$D6,$32,$FE,$3A,$7F,$02,$FF,$89
	.dc.b	$63,$2F,$C4,$7F,$C1,$AF,$D7,$AF,$C3,$57,$EC,$67,$F8,$16,$1F,$A1
	.dc.b	$7E,$31,$FC,$94,$ED,$7F,$C8,$2F,$F2,$0F,$E3,$E3,$D4,$7E,$7C,$FE
	.dc.b	$CC,$FF,$19,$BF,$CD,$DF,$77,$D5,$19,$3E,$33,$FD,$A2,$BA,$D7,$F7
	.dc.b	$DC,$3F,$4F,$7F,$C1,$DF,$8D,$7F,$C2,$DD,$CF,$F7,$07,$F1,$31,$FE
	.dc.b	$05,$BF,$E4,$CF,$68,$FE,$3F,$87,$F0,$F8,$F8,$B8,$7F,$01,$7F,$A1
	.dc.b	$DF,$EE,$5F,$8B,$87,$F1,$77,$E2,$C7,$F0,$13,$E2,$9F,$A2,$9F,$1B
	.dc.b	$FE,$FE,$7E,$84,$7C,$43,$C8,$7F,$0B,$3F,$84,$5E,$B0,$E7,$5F,$C6
	.dc.b	$DF,$E2,$EF,$DC,$AF,$E0,$CF,$C8,$7F,$99,$61,$FC,$1D,$FE,$15,$79
	.dc.b	$78,$BF,$C1,$A7,$DA,$D7,$F7,$87,$ED,$CF,$69,$FE,$3E,$30,$7E,$C1
	.dc.b	$FE,$DD,$1F,$BF,$8F,$CF,$1F,$D5,$30,$FE,$0B,$1F,$C0,$E7,$33,$FC
	.dc.b	$9E,$FF,$83,$4F,$EE,$30,$FB,$75,$FC,$65,$FD,$05,$7F,$07,$9F,$C3
	.dc.b	$CE,$5F,$8B,$1F,$D6,$5F,$23,$F8,$77,$F7,$0F,$F0,$03,$F7,$05,$3F
	.dc.b	$65,$1F,$C9,$CF,$79,$FD,$18,$C3,$D0,$FF,$2B,$3F,$86,$BF,$E0,$31
	.dc.b	$F8,$9B,$F9,$33,$FB,$C7,$CF,$1F,$E5,$58,$FC,$34,$FF,$04,$9E,$15
	.dc.b	$FC,$60,$F8,$35,$FE,$5B,$7E,$0D,$7E,$F1,$E7,$1F,$E1,$B7,$E0,$BF
	.dc.b	$DE,$23,$D6,$F1,$8F,$83,$1F,$75,$FE,$01,$1F,$80,$8F,$F0,$CD,$FE
	.dc.b	$5A,$BF,$CE,$4E,$A7,$E6,$C7,$F8,$5A,$FF,$A0,$DF,$28,$F9,$77,$FB
	.dc.b	$DB,$E5,$7F,$D8,$A3,$FC,$1E,$7E,$F1,$FF,$01,$D7,$E9,$1F,$F0,$C3
	.dc.b	$FC,$4E,$7E,$0C,$7E,$E6,$3F,$83,$BF,$DA,$63,$F7,$6F,$F8,$4E,$3C
	.dc.b	$B1,$FB,$AB,$FE,$6F,$1F,$AE,$9F,$2D,$7F,$81,$B0,$FD,$E4,$7C,$A3
	.dc.b	$F7,$11,$FC,$40,$FE,$4B,$1F,$52,$7F,$81,$61,$CA,$BF,$C0,$2F,$92
	.dc.b	$BF,$6F,$3A,$79,$47,$EB,$9F,$F6,$BC,$7D,$27,$FC,$FF,$7F,$E1,$F8
	.dc.b	$FD,$BB,$5F,$24,$FD,$BA,$FE,$FB,$7F,$DE,$9F,$11,$FC,$BE,$31,$87
	.dc.b	$F3,$D8,$FC,$19,$FE,$59,$8E,$15,$FD,$05,$E3,$1F,$E5,$13,$F0,$E7
	.dc.b	$F2,$59,$D6,$1F,$B6,$9F,$D0,$4F,$F9,$5E,$BF,$6C,$72,$7F,$CA,$07
	.dc.b	$23,$F6,$55,$AA,$FE,$06,$FF,$53,$7F,$BB,$4F,$F2,$F9,$E5,$1F,$DD
	.dc.b	$4F,$EC,$6F,$D5,$3F,$D6,$CF,$F0,$C3,$D9,$87,$DE,$9F,$D8,$4E,$6F
	.dc.b	$F6,$35,$FB,$B7,$93,$FB,$08,$FE,$EA,$F9,$DF,$F1,$FD,$7F,$6C,$7F
	.dc.b	$8C,$D7,$EC,$27,$EC,$DF,$ED,$63,$F5,$E7,$AE,$3E,$EB,$F4,$4F,$D4
	.dc.b	$4F,$E7,$CF,$C3,$47,$EB,$A3,$FB,$64,$7E,$B0,$FC,$7D,$7E,$B2,$3F
	.dc.b	$B8,$5F,$D6,$3F,$AD,$61,$8C,$3F,$61,$87,$EA,$E7,$F1,$D8,$7F,$1F
	.dc.b	$BF,$D5,$1C,$6B,$F1,$B1,$FA,$98,$FC,$6C,$7F,$8A,$AB,$F0,$11,$CA
	.dc.b	$7F,$8E,$9F,$97,$3F,$C7,$67,$E9,$4F,$59,$FE,$3E,$9F,$D7,$4E,$67
	.dc.b	$F7,$8F,$F8,$08,$FF,$31,$4F,$E9,$27,$F8,$74,$7F,$05,$8F,$D5,$BF
	.dc.b	$A8,$9F,$E0,$87,$F7,$E8,$FF,$34,$3F,$C1,$1F,$E1,$2F,$89,$FF,$03
	.dc.b	$3F,$A8,$3D,$8F,$E5,$DF,$63,$FC,$8F,$0F,$F0,$F3,$F3,$4F,$EF,$67
	.dc.b	$F8,$5C,$FD,$89,$FD,$04,$FD,$31,$C8,$FF,$2C,$CF,$F8,$6D,$FA,$55
	.dc.b	$FC,$82,$F8,$9F,$74,$E0,$FA,$E7,$FD,$41,$3F,$E1,$1B,$F5,$3F,$EA
	.dc.b	$69,$FE,$3A,$FF,$21,$9D,$4F,$CC,$8F,$A8,$7F,$A7,$4F,$C3,$47,$A8
	.dc.b	$FF,$4E,$4F,$D4,$8F,$E3,$F3,$F5,$2B,$8C,$7F,$71,$3E,$A6,$1F,$C7
	.dc.b	$AB,$FC,$21,$7F,$4E,$3F,$D2,$D7,$C2,$FF,$D8,$B7,$FE,$0B,$BF,$E9
	.dc.b	$13,$E8,$8F,$AB,$7F,$E0,$F8,$FE,$7D,$1F,$D3,$CF,$F2,$31,$FB,$13
	.dc.b	$FC,$1A,$7C,$F3,$F7,$18,$7F,$A4,$63,$FC,$19,$1F,$E0,$DB,$FF,$46
	.dc.b	$5F,$F1,$68,$FF,$06,$47,$F8,$2A,$7F,$D1,$07,$F1,$59,$FD,$D4,$FF
	.dc.b	$80,$DF,$E2,$8E,$AF,$F8,$14,$7D,$57,$FB,$FC,$7F,$0F,$3F,$85,$9F
	.dc.b	$DE,$4F,$E1,$67,$F7,$07,$F8,$9B,$FE,$0D,$9F,$F4,$8C,$7F,$8F,$63
	.dc.b	$FC,$01,$5C,$9F,$F0,$4B,$FC,$50,$E4,$7F,$2E,$7F,$89,$9F,$E7,$C3
	.dc.b	$F8,$9C,$F1,$7F,$CA,$B5,$FC,$4E,$B2,$7F,$1B,$9F,$54,$FC,$E3,$FA
	.dc.b	$41,$F3,$CF,$E5,$87,$CF,$3C,$B7,$F3,$CF,$F5,$EB,$FD,$04,$FA,$D7
	.dc.b	$C0,$FE,$9A,$79,$DF,$DF,$57,$F8,$CD,$FE,$7F,$5E,$77,$FE,$00,$3F
	.dc.b	$3D,$FA,$13,$F5,$A3,$E3,$C7,$F4,$23,$84,$FF,$8C,$9F,$9C,$7C,$57
	.dc.b	$FC,$60,$FF,$AB,$8F,$E8,$07,$F3,$B7,$F8,$63,$F3,$9F,$E8,$86,$67
	.dc.b	$F4,$31,$F9,$A3,$F3,$A7,$C9,$9F,$EA,$07,$F0,$A7,$FC,$5D,$39,$8F
	.dc.b	$ED,$D1,$FE,$5A,$3D,$C7,$E6,$8F,$EE,$27,$C5,$3B,$9F,$83,$3F,$D2
	.dc.b	$47,$CC,$BF,$CD,$5F,$FB,$B2,$BE,$7D,$7F,$6E,$AF,$96,$7E,$59,$FE
	.dc.b	$C9,$5E,$58,$FF,$30,$4E,$67,$E5,$C7,$E6,$4F,$88,$7E,$62,$3F,$38
	.dc.b	$62,$3F,$2F,$7F,$DC,$8F,$A2,$7E,$56,$3E,$4B,$FE,$23,$BF,$EC,$D5
	.dc.b	$F2,$23,$F8,$2D,$FE,$4E,$3F,$83,$D7,$C7,$BF,$C8,$C7,$E4,$AB,$FC
	.dc.b	$2F,$1F,$E1,$B9,$F9,$47,$E5,$5F,$F0,$B4,$FE,$46,$7F,$7F,$1F,$E8
	.dc.b	$98,$FD,$FD,$FF,$67,$0F,$FA,$14,$F8,$F3,$FC,$0A,$FF,$7F,$3A,$B1
	.dc.b	$FE,$89,$3E,$45,$FF,$84,$EF,$E4,$5F,$F8,$56,$FF,$2B,$5F,$22,$3F
	.dc.b	$C3,$15,$F2,$2B,$F2,$31,$FE,$81,$AF,$97,$5F,$22,$FC,$0F,$F8,$4E
	.dc.b	$FF,$C1,$57,$FB,$C3,$F2,$15,$F2,$6B,$FD,$FB,$5F,$25,$E3,$3F,$BB
	.dc.b	$6B,$FE,$41,$6B,$FE,$9F,$6B,$FE,$B7,$8F,$F5,$E1,$F3,$6B,$F9,$5C
	.dc.b	$FE,$FA,$FE,$6B,$F9,$1A,$F9,$A7,$FB,$06,$3F,$4F,$7F,$DC,$63,$FC
	.dc.b	$7F,$1F,$36,$BF,$79,$5E,$0A,$FF,$12,$57,$F8,$4E,$7E,$63,$F3,$6B
	.dc.b	$E6,$47,$F9,$06,$FE,$65,$FF,$8E,$CF,$F5,$59,$EA,$53,$F9,$4B,$F9
	.dc.b	$B7,$FE,$AF,$9F,$DE,$5F,$CC,$8F,$CA,$5F,$C2,$8F,$EA,$91,$FE,$6B
	.dc.b	$8F,$9B,$7F,$D7,$27,$F7,$73,$F9,$59,$F9,$2F,$EE,$EB,$F2,$B5,$FE
	.dc.b	$6A,$AF,$93,$3F,$14,$FD,$D3,$F9,$5A,$FD,$D4,$FF,$53,$7E,$24,$7F
	.dc.b	$53,$3F,$92,$5F,$90,$FD,$D5,$1F,$92,$FF,$AC,$A3,$F8,$BD,$FF,$69
	.dc.b	$8F,$F1,$A9,$FB,$C9,$FC,$AB,$F2,$EF,$F2,$B5,$FE,$64,$8F,$F3,$2C
	.dc.b	$7E,$56,$3E,$44,$7C,$98,$FD,$E3,$FD,$3C,$F9,$27,$E9,$AB,$FC,$C3
	.dc.b	$3F,$BC,$8F,$DE,$CF,$C9,$8F,$89,$7F,$E2,$C9,$FF,$30,$5F,$FB,$09
	.dc.b	$FF,$15,$9F,$E5,$F3,$FC,$54,$7F,$48,$3F,$D2,$55,$FE,$33,$3F,$C4
	.dc.b	$67,$FB,$06,$7F,$3A,$FF,$87,$DB,$1F,$EC,$AC,$3F,$A5,$5A,$77,$7F
	.dc.b	$E4,$7B,$F9,$93,$D2,$3F,$C2,$D1,$8C,$3F,$C5,$B7,$F2,$2F,$BC,$7C
	.dc.b	$FC,$3F,$22,$FF,$8B,$4F,$E1,$EF,$F8,$C6,$FF,$C5,$D3,$C2,$3F,$C8
	.dc.b	$D3,$ED,$7F,$8A,$DF,$F1,$03,$FC,$6E,$7B,$A3,$FA,$BC,$76,$8F,$F7
	.dc.b	$7C,$7B,$67,$F3,$58,$7F,$8C,$4F,$61,$FE,$7C,$8F,$E3,$75,$FE,$5F
	.dc.b	$B4,$FC,$38,$FF,$34,$BF,$EF,$A3,$FD,$CC,$FD,$D3,$F8,$61,$4F,$E1
	.dc.b	$85,$3F,$86,$14,$FE,$19,$3F,$E3,$A3,$F8,$65,$7F,$82,$23,$FC,$6C
	.dc.b	$FE,$AA,$FF,$C1,$13,$FE,$CB,$3F,$AE,$D7,$F8,$DA,$BF,$9B,$D7,$E9
	.dc.b	$AF,$F9,$BC,$77,$7F,$46,$FF,$3E,$7F,$C3,$AF,$E8,$8F,$E1,$91,$FC
	.dc.b	$1D,$FE,$BC,$7E,$8D,$FE,$18,$FF,$81,$CF,$D1,$BF,$C3,$6F,$FA,$F4
	.dc.b	$7E,$8E,$31,$5F,$C2,$DF,$F6,$24,$FF,$5E,$8F,$D1,$47,$F0,$A3,$EA
	.dc.b	$BF,$EC,$11,$B7,$F9,$7A,$7F,$0F,$5F,$D5,$A3,$0F,$AF,$1D,$BF,$EF
	.dc.b	$B3,$DD,$58,$78,$9F,$C8,$8F,$EB,$46,$4F,$F0,$59,$F8,$08,$C3,$C2
	.dc.b	$BF,$07,$38,$8F,$C9,$9C,$CC,$1D,$63,$F8,$94,$E1,$FC,$CE,$1F,$8C
	.dc.b	$BF,$4C,$7E,$41,$B6,$AE,$77,$FB,$08,$E0,$FA,$8F,$F4,$F3,$FA,$A9
	.dc.b	$F8,$D7,$F9,$83,$93,$88,$FD,$CB,$71,$F9,$55,$CF,$0E,$D8,$78,$27
	.dc.b	$FC,$07,$7E,$0A,$F8,$51,$F4,$AF,$EA,$3F,$45,$B1,$D8,$E9,$1F,$3A
	.dc.b	$33,$3E,$49,$FD,$05,$FC,$C9,$E6,$FA,$E3,$8B,$F2,$A3,$38,$78,$E7
	.dc.b	$F0,$73,$FC,$0E,$FD,$71,$E6,$8F,$B5,$5F,$E7,$9B,$FB,$27,$D3,$3F
	.dc.b	$7B,$3C,$0E,$6F,$C9,$C3,$4F,$AE,$3F,$04,$FF,$04,$BE,$11,$FE,$E5
	.dc.b	$8F,$63,$DD,$FC,$03,$F1,$CF,$CA,$1E,$B2,$E1,$EB,$7F,$DC,$B3,$D1
	.dc.b	$FC,$69,$EA,$3F,$D2,$58,$7C,$77,$F3,$CF,$C7,$7F,$2C,$7F,$B8,$6F
	.dc.b	$C3,$1E,$98,$FF,$33,$BF,$8F,$78,$DF,$A1,$FD,$2D,$FC,$6C,$3F,$11
	.dc.b	$1E,$33,$F1,$F1,$B7,$F9,$CD,$7F,$A5,$27,$F1,$91,$F9,$F9,$FD,$3D
	.dc.b	$FE,$2E,$3D,$07,$E2,$8F,$41,$F6,$B0,$FC,$53,$F8,$83,$D0,$76,$7F
	.dc.b	$14,$E7,$0F,$13,$FE,$C2,$7E,$53,$FD,$1A,$3F,$69,$7F,$88,$AF,$D9
	.dc.b	$47,$EA,$23,$F1,$EF,$B7,$0F,$E9,$97,$D2,$FF,$42,$FA,$1F,$C2,$BE
	.dc.b	$87,$FA,$35,$FE,$1B,$0F,$0C,$7F,$BE,$0F,$C4,$CF,$37,$FA,$FB,$AA
	.dc.b	$FE,$27,$87,$F1,$A9,$FC,$21,$F6,$8F,$82,$7E,$0B,$0F,$33,$FB,$4C
	.dc.b	$3F,$82,$BF,$E3,$CB,$E8,$FC,$03,$F0,$B8,$7A,$F0,$FE,$23,$3F,$84
	.dc.b	$BE,$71,$F8,$DA,$FE,$19,$5F,$81,$7F,$82,$3F,$A6,$7E,$E5,$79,$9F
	.dc.b	$D8,$C7,$F0,$F8,$CD,$7C,$B8,$FC,$0E,$1F,$EE,$98,$F9,$6F,$F0,$6A
	.dc.b	$FA,$11,$F8,$07,$F0,$93,$F1,$63,$EF,$C7,$E1,$EB,$E5,$47,$BD,$FB
	.dc.b	$77,$F4,$E7,$F0,$0F,$95,$FC,$43,$F2,$9F,$F0,$24,$F8,$23,$E0,$47
	.dc.b	$F4,$49,$F8,$0F,$EF,$9E,$8F,$7A,$F9,$51,$EF,$BF,$EA,$8F,$F0,$39
	.dc.b	$FB,$F8,$7D,$47,$F1,$07,$E9,$8F,$E0,$47,$DD,$AF,$A8,$FF,$56,$AF
	.dc.b	$C1,$61,$F7,$A3,$CA,$78,$E3,$FA,$3B,$F6,$27,$8E,$1E,$D3,$F4,$11
	.dc.b	$FC,$03,$0F,$C8,$4F,$B2,$3E,$BC,$7B,$27,$FC,$77,$3F,$85,$3D,$93
	.dc.b	$F8,$09,$FE,$0C,$FF,$61,$8F,$B6,$FF,$85,$CF,$69,$E0,$3E,$D9,$F4
	.dc.b	$4F,$F1,$E4,$7E,$C1,$E6,$7E,$FC,$F6,$9F,$B9,$7E,$DB,$F5,$63,$F3
	.dc.b	$17,$FB,$78,$FE,$86,$FD,$99,$FD,$61,$ED,$8F,$E2,$6F,$E5,$EF,$EA
	.dc.b	$BE,$38,$FB,$4F,$F3,$C9,$FE,$00,$FB,$23,$FB,$61,$F5,$8F,$E5,$F3
	.dc.b	$87,$FD,$14,$D7,$F2,$A7,$F5,$A7,$0F,$C4,$3F,$00,$7E,$DA,$FF,$CB
	.dc.b	$31,$C5,$FC,$A4,$7D,$88,$C5,$7F,$AD,$9D,$57,$F0,$C9,$FF,$38,$9F
	.dc.b	$CC,$CF,$F6,$4C,$79,$6F,$EC,$C7,$EE,$A7,$E7,$BF,$E2,$88,$FE,$B1
	.dc.b	$7F,$E0,$C7,$E4,$47,$F9,$06,$BF,$6C,$F9,$8F,$BD,$1E,$1B,$F7,$C7
	.dc.b	$F8,$C9,$3F,$E7,$39,$E3,$7B,$BF,$D4,$CF,$F9,$DC,$F6,$1F,$E0,$0B
	.dc.b	$D5,$FA,$CF,$F3,$C5,$FF,$07,$AF,$95,$5F,$E3,$03,$E0,$5F,$F7,$88
	.dc.b	$FF,$12,$BF,$02,$3F,$74,$FC,$07,$F9,$7B,$E5,$7E,$EB,$F0,$1F,$F1
	.dc.b	$54,$FD,$87,$CA,$7C,$B8,$FD,$D5,$7C,$F8,$F7,$BA,$79,$C7,$DD,$75
	.dc.b	$5F,$25,$FF,$18,$57,$DC,$7F,$05,$5C,$5F,$DE,$1F,$90,$BF,$CC,$36
	.dc.b	$F9,$3F,$DD,$1F,$C3,$C7,$F0,$FA,$F3,$57,$3A,$FF,$6F,$57,$D7,$7E
	.dc.b	$8C,$7F,$76,$9F,$BB,$59,$8F,$A1,$3F,$14,$FF,$26,$3F,$CD,$9F,$F1
	.dc.b	$A9,$F5,$6F,$C6,$FE,$05,$FC,$CC,$7F,$AF,$23,$FC,$91,$1C,$DF,$F3
	.dc.b	$7C,$7F,$37,$AF,$F1,$24,$7D,$AB,$E6,$FD,$6B,$FC,$F4,$7B,$63,$E6
	.dc.b	$D7,$B5,$FE,$E7,$1F,$A3,$7F,$6C,$7F,$A1,$6B,$F6,$35,$98,$FC,$11
	.dc.b	$E7,$8F,$DA,$CF,$1B,$FC,$AC,$FF,$9D,$E7,$F6,$95,$9B,$FB,$67,$86
	.dc.b	$7F,$DA,$37,$FB,$39,$FA,$F1,$FA,$F3,$4F,$E4,$DF,$D8,$47,$89,$FF
	.dc.b	$73,$5F,$EC,$63,$F3,$B5,$EA,$9F,$F7,$1B,$F5,$DF,$D8,$C7,$F8,$A6
	.dc.b	$3E,$B9,$E3,$3E,$B4,$7D,$89,$FB,$E7,$CC,$7F,$2E,$FD,$63,$C5,$1F
	.dc.b	$AE,$8F,$F3,$8C,$FD,$03,$F6,$E7,$EB,$6F,$FC,$B5,$79,$AF,$B0,$7C
	.dc.b	$39,$FD,$55,$7F,$86,$8F,$E4,$37,$FD,$86,$BE,$95,$7E,$A6,$7F,$08
	.dc.b	$7F,$A6,$6B,$E9,$3F,$A8,$3E,$71,$FC,$FC,$FF,$45,$DF,$F2,$4B,$FE
	.dc.b	$55,$3F,$44,$FE,$1B,$7E,$7A,$FD,$43,$9B,$FE,$4C,$FF,$20,$BF,$DE
	.dc.b	$57,$E9,$6B,$F4,$27,$CB,$3F,$36,$7C,$63,$CE,$7D,$43,$EB,$DF,$CE
	.dc.b	$7E,$93,$F3,$6F,$F4,$C6,$27,$E6,$4F,$E7,$CF,$90,$FF,$40,$3F,$3C
	.dc.b	$FC,$37,$BB,$F9,$E7,$FC,$14,$7F,$8F,$1F,$EF,$07,$F8,$9C,$FF,$6B
	.dc.b	$DE,$67,$FC,$11,$3F,$E4,$F9,$FD,$81,$FE,$86,$3F,$C1,$18,$7F,$A1
	.dc.b	$2B,$FC,$8F,$7F,$A6,$3F,$3B,$1F,$95,$AF,$F2,$B4,$FE,$7A,$BF,$C1
	.dc.b	$2F,$F8,$1E,$3F,$C2,$0F,$AA,$3F,$C1,$B1,$FE,$36,$7F,$B1,$52,$FF
	.dc.b	$3D,$1F,$C1,$EB,$FB,$EB,$FE,$19,$8F,$E9,$AF,$F8,$6A,$BF,$C7,$F5
	.dc.b	$CA,$7F,$CB,$93,$FE,$BD,$AF,$F0,$2C,$F1,$3D,$53,$FE,$63,$BF,$E8
	.dc.b	$B7,$E2,$3D,$57,$FE,$78,$C3,$FC,$0B,$87,$F9,$CE,$7D,$53,$FE,$BB
	.dc.b	$AF,$A3,$1F,$EB,$A8,$FA,$27,$F8,$2F,$0F,$F0,$1B,$FE,$BC,$C3,$E8
	.dc.b	$D7,$E5,$AB,$FD,$63,$3F,$43,$0F,$F7,$FE,$1F,$42,$FF,$CE,$35,$FE
	.dc.b	$36,$3E,$3C,$79,$AB,$FC,$EE,$FC,$F3,$E9,$C7,$D7,$7F,$C3,$11,$FC
	.dc.b	$B8,$F8,$C7,$F8,$F6,$7F,$0C,$FF,$AA,$4F,$9C,$7C,$63,$FC,$62,$7F
	.dc.b	$06,$9F,$F3,$86,$1C,$4F,$74,$FE,$CE,$7F,$D6,$D1,$F3,$8E,$2F,$FB
	.dc.b	$B2,$7F,$DC,$EF,$F8,$B1,$F8,$C7,$F2,$37,$F9,$4B,$F8,$F3,$D2,$FE
	.dc.b	$86,$FF,$D4,$4F,$F9,$86,$FF,$CC,$AF,$F9,$74,$FC,$28,$AF,$F5,$5E
	.dc.b	$1F,$3E,$BF,$CC,$B3,$F9,$D8,$FC,$F5,$FF,$A7,$E3,$FC,$C9,$1F,$9D
	.dc.b	$8F,$F4,$DC,$7F,$A5,$DF,$B9,$7E,$78,$EC,$FE,$11,$FF,$54,$C7,$E7
	.dc.b	$4F,$F4,$B3,$F3,$CF,$8E,$7C,$EC,$3F,$CC,$71,$F9,$A3,$E6,$C7,$F9
	.dc.b	$76,$BF,$D3,$35,$F4,$2B,$FC,$B9,$1F,$98,$AF,$CA,$1F,$E7,$70,$7C
	.dc.b	$C3,$FD,$00,$79,$6B,$FD,$48,$FF,$01,$7F,$62,$7E,$C3,$0F,$F1,$BB
	.dc.b	$F9,$93,$FD,$20,$7E,$64,$FC,$F4,$7F,$09,$8E,$27,$F1,$03,$81,$F4
	.dc.b	$CF,$E1,$11,$FA,$FC,$3F,$96,$CF,$CC,$8F,$F6,$3E,$1F,$99,$8F,$F6
	.dc.b	$2C,$7F,$B2,$9F,$94,$7F,$A1,$5F,$F4,$4C,$7F,$04,$3E,$6E,$1F,$2B
	.dc.b	$0F,$F0,$6D,$7C,$AA,$F9,$31,$F2,$A3,$FD,$87,$3F,$2A,$FF,$CF,$11
	.dc.b	$FE,$3F,$C3,$FC,$95,$7F,$EF,$7B,$FF,$7C,$4F,$FA,$0E,$3F,$D7,$35
	.dc.b	$FE,$43,$7F,$D7,$47,$F5,$43,$FC,$EE,$7E,$48,$FE,$EF,$5F,$C1,$A3
	.dc.b	$FC,$F9,$1F,$D6,$E3,$FD,$01,$87,$F8,$2E,$3F,$C8,$31,$F9,$77,$FC
	.dc.b	$7C,$7F,$56,$AF,$EF,$0F,$FA,$D1,$FF,$03,$61,$93,$FC,$1A,$FF,$82
	.dc.b	$23,$FC,$7A,$70,$8F,$B6,$FF,$8E,$8F,$EB,$4F,$F5,$D8,$FF,$3A,$47
	.dc.b	$F8,$EC,$FE,$B3,$5F,$E7,$27,$FC,$73,$3F,$D6,$67,$F9,$4C,$7F,$59
	.dc.b	$AF,$E5,$75,$FE,$03,$8F,$E5,$B5,$FD,$72,$BF,$C6,$11,$FD,$FC,$F9
	.dc.b	$13,$FE,$02,$7F,$CD,$75,$F1,$DF,$F5,$63,$FE,$32,$3E,$0B,$FC,$3C
	.dc.b	$FC,$B9,$FE,$A5,$37,$5F,$E0,$68,$FF,$42,$CF,$F2,$79,$FF,$49,$BF
	.dc.b	$E0,$37,$FD,$44,$FF,$80,$DF,$97,$1F,$C0,$6F,$FD,$3F,$1F,$E0,$28
	.dc.b	$47,$F9,$9A,$FF,$C6,$13,$FE,$31,$9C,$BF,$EE,$18,$F9,$6F,$F9,$80
	.dc.b	$F9,$6F,$CA,$7E,$5C,$7E,$62,$3E,$5C,$7C,$C8,$F9,$71,$F2,$A3,$E5
	.dc.b	$C7,$F0,$18,$FC,$8C,$7E,$4A,$3F,$C5,$91,$FE,$94,$8F,$97,$3F,$E5
	.dc.b	$C8,$FF,$49,$47,$FA,$4E,$7F,$23,$1F,$E9,$08,$FC,$8C,$FF,$B3,$23
	.dc.b	$FC,$A5,$1F,$91,$3F,$1B,$5F,$1E,$7F,$D1,$B5,$F1,$E7,$FE,$00,$1F
	.dc.b	$1E,$3E,$31,$FE,$C8,$BF,$F2,$AC,$FC,$69,$FC,$95,$FC,$6A,$FF,$5E
	.dc.b	$47,$C6,$7F,$D7,$D5,$FE,$28,$AF,$8D,$1F,$E8,$69,$F8,$CF,$F9,$4A
	.dc.b	$BE,$33,$FB,$C7,$FC,$4F,$1F,$D4,$63,$F7,$07,$FA,$EE,$7E,$3B,$FE
	.dc.b	$F4,$AF,$F2,$91,$FB,$BA,$FF,$28,$D7,$C7,$9F,$F4,$0C,$7C,$77,$FC
	.dc.b	$A0,$7C,$73,$FD,$89,$1F,$1E,$BF,$CF,$57,$F1,$E3,$FD,$71,$7F,$B6
	.dc.b	$BF,$F2,$74,$7C,$33,$FB,$B1,$FE,$4F,$9F,$F5,$93,$F8,$DB,$FF,$74
	.dc.b	$CF,$E2,$E3,$F1,$97,$FE,$AC,$AF,$DB,$57,$C8,$8F,$8F,$3F,$ED,$23
	.dc.b	$F6,$D1,$F1,$CF,$8D,$7F,$E7,$89,$FC,$64,$7E,$1C,$FE,$DF,$3F,$E1
	.dc.b	$DA,$FE,$DE,$7F,$92,$DF,$E8,$78,$7F,$17,$7F,$66,$7F,$71,$3F,$C6
	.dc.b	$67,$ED,$A8,$FF,$68,$8F,$EE,$35,$C2,$33,$3F,$BB,$79,$BF,$E5,$08
	.dc.b	$FE,$92,$6E,$3F,$B8,$E1,$FD,$62,$31,$2B,$FE,$C6,$7F,$4D,$8C,$BF
	.dc.b	$D8,$E3,$53,$FE,$16,$C3,$FD,$4F,$7E,$77,$FB,$21,$FE,$26,$7C,$AF
	.dc.b	$F2,$A7,$47,$F8,$9E,$3F,$BD,$1F,$6D,$FE,$D7,$3F,$E5,$D3,$FC,$51
	.dc.b	$5F,$E5,$43,$FB,$E4,$6C,$EB,$1F,$E2,$89,$FE,$A8,$7F,$19,$7F,$BE
	.dc.b	$BF,$DF,$67,$FC,$03,$5F,$E2,$B8,$FE,$F8,$7F,$A0,$EB,$FC,$4D,$87
	.dc.b	$D9,$8F,$F2,$B4,$FF,$94,$B0,$FC,$C5,$7F,$A0,$4F,$F1,$23,$FE,$52
	.dc.b	$9F,$F1,$23,$F3,$23,$FB,$D1,$F4,$6F,$FB,$41,$E5,$3E,$7D,$FF,$7C
	.dc.b	$AF,$EF,$75,$F4,$1F,$9D,$5F,$E5,$23,$FB,$DC,$FC,$17,$FD,$01,$17
	.dc.b	$0F,$F4,$24,$7C,$D9,$FC,$DB,$F8,$4C,$3F,$1A,$FC,$E7,$F1,$97,$F8
	.dc.b	$27,$FA,$34,$7C,$C7,$FC,$00,$FE,$26,$FE,$69,$FE,$29,$3E,$61,$FD
	.dc.b	$FE,$7F,$16,$FF,$94,$A3,$F3,$AF,$CD,$7F,$49,$5F,$36,$7E,$34,$7F
	.dc.b	$88,$0F,$F3,$09,$FE,$AE,$3F,$36,$7A,$0F,$CF,$9E,$83,$F4,$57,$FA
	.dc.b	$2C,$3F,$D0,$51,$FE,$54,$9F,$D0,$BA,$8F,$A1,$7F,$E5,$43,$FD,$5E
	.dc.b	$7F,$A0,$4F,$9E,$7F,$94,$A3,$E7,$CF,$C5,$7F,$CA,$47,$CF,$C3,$F6
	.dc.b	$B3,$F9,$A9,$F9,$57,$F9,$8C,$3F,$C5,$51,$8A,$F9,$77,$FE,$D5,$BF
	.dc.b	$F2,$9B,$FE,$7D,$C1,$7F,$E2,$AB,$F9,$75,$FE,$7D,$3E,$2D,$7F,$94
	.dc.b	$8F,$F5,$4D,$7F,$0E,$8F,$CB,$C7,$F0,$E9,$FC,$BB,$F2,$B0,$FF,$28
	.dc.b	$57,$69,$FC,$C1,$F9,$78,$FF,$28,$9F,$E9,$28,$FC,$BD,$7F,$07,$AF
	.dc.b	$94,$FC,$98,$FF,$54,$E1,$FE,$7F,$8F,$F2,$8D,$FC,$7A,$F9,$27,$D6
	.dc.b	$9F,$E0,$4F,$FA,$A5,$FF,$29,$47,$F1,$58,$FE,$01,$3F,$E5,$38,$F9
	.dc.b	$75,$FC,$06,$3F,$CA,$73,$F4,$6F,$F8,$04,$7F,$B4,$CF,$CA,$4F,$B6
	.dc.b	$7F,$0B,$1F,$E9,$19,$FE,$07,$7F,$C0,$70,$FF,$01,$E1,$FD,$74,$F1
	.dc.b	$5F,$98,$FF,$61,$3F,$D7,$EB,$FC,$5D,$1F,$98,$8F,$F2,$AC,$7F,$81
	.dc.b	$1E,$2F,$EC,$9F,$EB,$F1,$FE,$34,$BF,$F4,$7C,$7F,$60,$8F,$F0,$3C
	.dc.b	$FF,$60,$9F,$F2,$AC,$7F,$8B,$27,$F3,$2F,$F9,$44,$FF,$03,$1F,$CE
	.dc.b	$EF,$CC,$FF,$46,$7F,$CD,$45,$AF,$E0,$D1,$FC,$12,$BF,$C1,$0F,$18
	.dc.b	$FF,$2F,$1F,$CF,$2B,$FC,$13,$1F,$E2,$E7,$EA,$57,$EC,$E3,$F7,$75
	.dc.b	$F8,$37,$FA,$93,$FE,$45,$73,$87,$F9,$6E,$BF,$B3,$DF,$F6,$69,$F9
	.dc.b	$A7,$F9,$6D,$FF,$05,$BF,$42,$FF,$C1,$71,$F3,$A3,$F7,$A7,$FA,$06
	.dc.b	$7F,$C1,$67,$CE,$9F,$EC,$AF,$CE,$8F,$D3,$BF,$34,$FF,$06,$CF,$CD
	.dc.b	$7F,$1D,$5F,$EB,$D9,$FE,$B7,$5F,$E0,$DB,$F8,$EF,$F9,$9A,$BF,$CB
	.dc.b	$77,$8A,$F8,$C7,$D1,$3F,$D0,$6E,$6B,$E6,$C7,$F7,$5A,$FE,$F9,$3F
	.dc.b	$E0,$F9,$FF,$2D,$57,$F8,$3C,$F9,$B1,$FD,$59,$FF,$1B,$1F,$E3,$3A
	.dc.b	$FF,$7A,$3F,$E1,$17,$F9,$94,$FE,$35,$B5,$FE,$87,$8F,$CD,$C7,$CE
	.dc.b	$AF,$9A,$FD,$A3,$FC,$BA,$7E,$7A,$3F,$C6,$F5,$F9,$A3,$E6,$C7,$FA
	.dc.b	$10,$FF,$2E,$9F,$99,$8F,$F2,$EB,$F9,$88,$F9,$71,$F1,$23,$FC,$29
	.dc.b	$7F,$34,$FF,$76,$DF,$CD,$AF,$F4,$2C,$FC,$B3,$FC,$71,$5F,$E8,$48
	.dc.b	$FF,$30,$5F,$CB,$8F,$F1,$C4,$7F,$97,$A3,$E5,$3F,$96,$3E,$54,$FE
	.dc.b	$6E,$7F,$5A,$FF,$8E,$4F,$F2,$F9,$F2,$2F,$F2,$8F,$C8,$AF,$CA,$3F
	.dc.b	$94,$AF,$E0,$4F,$F0,$29,$FF,$57,$47,$F0,$13,$FC,$7D,$1F,$EA,$E3
	.dc.b	$F2,$87,$FA,$C2,$3F,$DA,$91,$F9,$38,$FE,$23,$5F,$22,$3F,$26,$7C
	.dc.b	$89,$FF,$2F,$4F,$C8,$9F,$F3,$8C,$7F,$86,$1F,$8D,$1F,$1E,$BF,$25
	.dc.b	$3F,$E5,$EA,$F8,$E7,$F9,$78,$FF,$41,$CF,$F9,$79,$FC,$6C,$7F,$97
	.dc.b	$A3,$F8,$05,$FF,$B4,$E7,$F1,$F1,$FE,$5F,$AF,$F5,$4C,$7F,$97,$A3
	.dc.b	$FD,$52,$FF,$A1,$6F,$E6,$C7,$FA,$1E,$3F,$CB,$E7,$F7,$AB,$FE,$A4
	.dc.b	$FF,$7C,$BF,$F1,$44,$FE,$26,$3F,$A7,$3F,$B8,$3F,$7B,$7F,$E5,$1C
	.dc.b	$3F,$CB,$B1,$8A,$D3,$FD,$67,$0F,$E7,$8F,$C7,$7C,$6F,$F9,$9F,$0F
	.dc.b	$EF,$A7,$F9,$65,$FE,$AF,$3F,$D5,$6B,$F1,$98,$7F,$55,$9F,$EF,$47
	.dc.b	$85,$FE,$FD,$5D,$23,$FA,$B1,$AB,$FE,$FC,$FF,$5B,$C3,$FB,$F5,$7F
	.dc.b	$56,$9F,$EA,$AF,$F8,$09,$F3,$CF,$F2,$87,$FB,$FB,$FE,$37,$7F,$C8
	.dc.b	$31,$FD,$FA,$BF,$CA,$B3,$99,$FE,$FE,$7F,$8A,$67,$E7,$5F,$F8,$A6
	.dc.b	$7F,$92,$9F,$E2,$99,$EE,$FF,$A0,$9F,$53,$FE,$28,$BF,$F0,$0C,$FC
	.dc.b	$E8,$FF,$02,$61,$F3,$9F,$F4,$81,$FE,$58,$3E,$75,$7F,$93,$70,$F9
	.dc.b	$B5,$FE,$57,$C3,$FC,$AD,$7F,$35,$FC,$95,$FF,$7F,$3F,$A3,$1F,$E8
	.dc.b	$0A,$F2,$BF,$E8,$43,$FC,$57,$1F,$E2,$DB,$F9,$B1,$FE,$57,$C3,$E6
	.dc.b	$BF,$3E,$BF,$D0,$35,$F3,$70,$FD,$CC,$FF,$A9,$EF,$FD,$00,$7C,$C7
	.dc.b	$F6,$35,$F3,$2B,$FC,$57,$3F,$E8,$03,$FC,$D1,$3F,$CF,$CF,$97,$3F
	.dc.b	$9A,$7E,$6E,$1F,$9A,$7F,$A6,$9F,$E8,$1A,$FF,$14,$4F,$B2,$3F,$9F
	.dc.b	$BF,$E5,$28,$FF,$16,$56,$67,$FD,$52,$FD,$47,$FD,$66,$7D,$13,$FD
	.dc.b	$43,$3E,$97,$FD,$76,$FF,$86,$4F,$F5,$91,$FE,$6D,$7F,$C0,$07,$FA
	.dc.b	$14,$FC,$DB,$E3,$C3,$FD,$06,$F7,$7F,$D0,$B7,$FE,$30,$9F,$E5,$15
	.dc.b	$FE,$2B,$7C,$47,$FA,$B2,$BF,$C6,$17,$FE,$59,$8F,$F1,$84,$FA,$CF
	.dc.b	$F4,$1B,$FE,$59,$C3,$FC,$B0,$FF,$20,$BF,$F1,$75,$7F,$28,$8F,$F3
	.dc.b	$D9,$FE,$58,$9F,$CD,$38,$9F,$9C,$FF,$28,$9F,$E0,$EF,$FA,$B8,$FE
	.dc.b	$0F,$7F,$35,$FE,$13,$7F,$36,$3F,$CA,$F1,$EA,$3F,$DB,$EF,$F0,$37
	.dc.b	$E5,$C7,$F9,$FC,$FF,$2B,$BF,$97,$3F,$CA,$E7,$FA,$9C,$FC,$A9,$F9
	.dc.b	$B7,$F2,$AF,$FB,$7F,$0F,$CA,$BF,$97,$9F,$CA,$BF,$C1,$67,$E3,$1F
	.dc.b	$E5,$57,$E3,$57,$EF,$E3,$FD,$51,$1F,$D1,$9F,$C3,$3F,$93,$3F,$C9
	.dc.b	$71,$FE,$28,$9F,$EA,$71,$FE,$27,$AF,$F1,$33,$F3,$27,$FB,$D3,$FD
	.dc.b	$4E,$BF,$C8,$2A,$7F,$C1,$0F,$FA,$5A,$7F,$D0,$A8,$85,$A5,$7E,$EA
	.dc.b	$32,$23,$FC,$80,$E1,$D9,$FD,$71,$D9,$FE,$2E,$AF,$F2,$9C,$7F,$5D
	.dc.b	$AF,$F0,$19,$FE,$2E,$3F,$C2,$CF,$F5,$D9,$FF,$17,$5F,$F8,$89,$FF
	.dc.b	$18,$5F,$F8,$B9,$EB,$1F,$E2,$EA,$FE,$BF,$3F,$E2,$E9,$FE,$BF,$7F
	.dc.b	$EE,$27,$FC,$B2,$7E,$14,$F6,$C7,$C1,$8F,$EC,$31,$FE,$2C,$3F,$B0
	.dc.b	$9F,$E3,$08,$FD,$C3,$EC,$9F,$F0,$3B,$FE,$22,$7F,$CB,$B5,$FE,$5A
	.dc.b	$7F,$41,$1F,$CD,$9F,$AA,$FF,$BF,$5F,$F2,$D3,$FE,$2E,$3F,$86,$9F
	.dc.b	$E0,$33,$CD,$7F,$39,$FA,$AF,$CE,$AF,$F4,$3C,$FF,$8C,$AB,$FC,$5D
	.dc.b	$5F,$E5,$53,$FD,$0E,$7F,$94,$CF,$F1,$B1,$F2,$0F,$F2,$CC,$7F,$38
	.dc.b	$7F,$CB,$17,$F7,$DF,$C1,$CF,$CD,$7F,$D1,$17,$FE,$0B,$77,$1F,$E0
	.dc.b	$B7,$FC,$61,$5F,$E5,$9A,$F9,$71,$FE,$86,$BF,$C8,$BF,$E8,$AA,$FF
	.dc.b	$2C,$57,$FA,$12,$3F,$D0,$8F,$E7,$E7,$F3,$31,$C0,$F2,$C7,$CC,$7F
	.dc.b	$40,$7F,$8C,$EB,$FD,$64,$FF,$96,$63,$FC,$64,$FF,$63,$9F,$F1,$93
	.dc.b	$F9,$43,$F3,$CF,$E0,$63,$FC,$68,$FC,$C8,$D1,$F9,$98,$FF,$59,$BF
	.dc.b	$24,$F9,$A7,$CB,$BF,$CC,$3F,$3E,$FF,$30,$FD,$1A,$FC,$C3,$FA,$A8
	.dc.b	$F9,$8F,$D0,$9F,$98,$FE,$72,$D3,$F3,$63,$E5,$1F,$99,$AF,$C2,$3F
	.dc.b	$B9,$7F,$09,$7F,$E3,$93,$FD,$62,$7F,$8E,$5F,$F4,$2C,$7F,$0C,$3F
	.dc.b	$CB,$CF,$CE,$9F,$CC,$47,$FA,$16,$3F,$84,$D7,$E0,$E3,$F8,$2B,$F1
	.dc.b	$DF,$E1,$D5,$F9,$A3,$E7,$BF,$DC,$E7,$E7,$BF,$9B,$9F,$F5,$93,$F3
	.dc.b	$2F,$FD,$0D,$3F,$E6,$03,$FD,$45,$1F,$99,$3F,$CC,$15,$FC,$12,$3F
	.dc.b	$D0,$D1,$FC,$12,$BF,$C7,$6F,$FB,$6C,$FD,$E5,$FC,$C7,$E5,$5F,$CC
	.dc.b	$7E,$84,$FC,$C3,$F8,$23,$FC,$72,$BF,$8F,$BF,$35,$F7,$3F,$99,$9F
	.dc.b	$F4,$4B,$F9,$BA,$FC,$CB,$F1,$F7,$FC,$3D,$FF,$20,$7F,$C6,$E7,$F9
	.dc.b	$96,$CF,$F9,$96,$69,$FE,$3A,$86,$38,$8E,$71,$FC,$CA,$7F,$B1,$9F
	.dc.b	$E1,$09,$D4,$FF,$8E,$2B,$F9,$A5,$62,$3E,$B9,$FD,$06,$FF,$9F,$5F
	.dc.b	$F8,$52,$F5,$3F,$C3,$2B,$F9,$1C,$FA,$B0,$F8,$73,$A3,$FB,$5D,$5C
	.dc.b	$3B,$CF,$96,$7F,$CD,$E6,$5F,$E5,$B3,$C6,$3F,$36,$76,$3F,$B8,$1C
	.dc.b	$F0,$FE,$A1,$16,$3D,$77,$F3,$27,$FC,$3E,$FF,$91,$E7,$A4,$7F,$5D
	.dc.b	$C3,$C5,$5F,$E4,$93,$DE,$7F,$92,$67,$B6,$1D,$23,$FC,$F5,$5A,$7F
	.dc.b	$CA,$15,$F4,$AF,$F5,$07,$D2,$7F,$B2,$BF,$E7,$7B,$F8,$51,$FC,$CE
	.dc.b	$35,$5C,$EB,$F4,$53,$CD,$FC,$31,$FE,$78,$34,$7D,$63,$FD,$45,$3F
	.dc.b	$87,$3F,$0C,$7E,$31,$FD,$F5,$7C,$B7,$E0,$4F,$D3,$7F,$BF,$1D,$0F
	.dc.b	$F4,$06,$1E,$17,$FB,$9E,$1F,$D7,$E3,$E3,$C7,$E2,$70,$E0,$F2,$9F
	.dc.b	$F0,$0D,$FD,$9A,$FF,$6A,$61,$F8,$5C,$3E,$1D,$FF,$2B,$BF,$C1,$DF
	.dc.b	$E0,$67,$E7,$4F,$F8,$C3,$0E,$B5,$FE,$62,$AF,$B3,$3F,$83,$3F,$C6
	.dc.b	$06,$A7,$F0,$77,$FE,$B4,$BF,$C2,$46,$67,$CB,$3F,$CD,$9F,$DE,$4F
	.dc.b	$B1,$FE,$2F,$3F,$11,$FC,$0E,$1F,$0D,$FF,$1D,$3B,$7F,$C0,$51,$FA
	.dc.b	$77,$F0,$F3,$B7,$F8,$33,$E7,$7F,$8F,$1E,$C7,$FC,$43,$7F,$84,$7E
	.dc.b	$0C,$7F,$23,$9D,$47,$A2,$3F,$06,$7F,$99,$0F,$8A,$7C,$03,$FD,$26
	.dc.b	$FD,$D7,$F5,$37,$F8,$37,$91,$F9,$57,$F8,$2B,$FE,$66,$AF,$C2,$61
	.dc.b	$E2,$8F,$E0,$6F,$27,$FC,$D3,$1F,$76,$BD,$91,$FE,$92,$7E,$EB,$FE
	.dc.b	$68,$C3,$E1,$3F,$89,$8E,$8D,$C3,$EE,$CF,$F0,$0C,$3E,$A4,$7F,$55
	.dc.b	$AF,$DE,$57,$D9,$AE,$6F,$CF,$7E,$93,$F4,$67,$EF,$4E,$A3,$EE,$BF
	.dc.b	$BA,$7F,$0D,$87,$EA,$1F,$F6,$34,$7E,$FA,$72,$F1,$3F,$33,$1F,$97
	.dc.b	$7F,$8E,$DF,$EE,$CF,$F3,$A1,$FB,$A8,$FB,$51,$E1,$7D,$51,$F4,$5F
	.dc.b	$AE,$78,$4F,$F4,$ED,$FA,$AF,$FC,$53,$1F,$66,$B9,$47,$CE,$8F,$F7
	.dc.b	$A1,$F8,$07,$93,$FC,$EE,$79,$3F,$93,$AF,$E7,$07,$E7,$E3,$D7,$1F
	.dc.b	$E9,$C3,$E8,$D7,$27,$ED,$47,$8C,$FF,$3D,$DE,$A7,$E6,$3E,$C9,$FE
	.dc.b	$67,$5F,$CB,$DE,$47,$FB,$DE,$BD,$93,$FC,$CA,$3F,$D4,$98,$7F,$32
	.dc.b	$C3,$D9,$1F,$D3,$DF,$C4,$BF,$AE,$8F,$D8,$E1,$F5,$63,$F5,$B1,$FD
	.dc.b	$16,$3D,$71,$FA,$C3,$F5,$57,$F8,$6A,$FF,$3B,$BF,$36,$7D,$73,$F5
	.dc.b	$AB,$F9,$23,$E2,$AF,$EE,$4F,$FB,$F2,$3F,$55,$5F,$32,$3F,$44,$7F
	.dc.b	$B5,$2B,$EA,$9F,$42,$BD,$31,$FA,$89,$F1,$3F,$EA,$B7,$57,$F6,$6B
	.dc.b	$F2,$D1,$EA,$9F,$D5,$1F,$51,$F5,$1F,$D1,$DF,$EC,$F8,$7F,$2B,$9F
	.dc.b	$A7,$1E,$9A,$FF,$57,$CF,$D3,$AF,$85,$1F,$E5,$88,$FA,$CF,$E9,$B0
	.dc.b	$FA,$57,$F5,$B0,$F4,$E1,$FE,$FB,$9F,$A3,$7F,$EF,$B3,$FB,$64,$FF
	.dc.b	$0A,$9F,$F5,$BD,$7A,$5F,$F5,$23,$FA,$67,$F5,$0F,$E1,$1F,$EA,$38
	.dc.b	$7D,$38,$F9,$71,$FD,$82,$FE,$9C,$7C,$F8,$FD,$3C,$FC,$A7,$F8,$49
	.dc.b	$FC,$D7,$0F,$D5,$CF,$D2,$3F,$8C,$C7,$E9,$67,$E8,$DF,$03,$F9,$BC
	.dc.b	$7F,$1E,$9F,$A2,$7E,$06,$7D,$11,$FC,$66,$FF,$AA,$47,$E8,$E7,$CF
	.dc.b	$3F,$C7,$DF,$A2,$7E,$79,$F9,$27,$F1,$59,$FF,$0A,$4E,$6F,$E8,$61
	.dc.b	$FC,$7B,$0F,$B0,$70,$3F,$C9,$D3,$F6,$1F,$44,$FE,$EE,$FF,$B4,$3F
	.dc.b	$E1,$9B,$FD,$16,$1F,$99,$3D,$33,$FE,$6B,$BF,$D1,$47,$E2,$63,$FD
	.dc.b	$39,$5F,$D3,$23,$F8,$B4,$7E,$84,$FC,$49,$FA,$1C,$3F,$C1,$8F,$F7
	.dc.b	$73,$F7,$47,$E8,$6F,$FC,$19,$87,$13,$D5,$1F,$9F,$8F,$E9,$78,$7E
	.dc.b	$82,$3E,$84,$7F,$49,$BF,$99,$1F,$9F,$8F,$9F,$1F,$E0,$3C,$3F,$88
	.dc.b	$47,$F8,$02,$3E,$74,$FF,$43,$8F,$9B,$1F,$DF,$A3,$FC,$84,$FF,$16
	.dc.b	$BF,$9B,$1F,$D0,$2F,$E6,$1F,$9F,$BF,$E7,$F7,$F2,$AF,$F2,$95,$FD
	.dc.b	$C2,$BE,$4C,$7E,$24,$FE,$13,$87,$F8,$F5,$F9,$27,$E5,$E7,$F8,$3D
	.dc.b	$7F,$08,$8F,$EE,$57,$F9,$68,$F8,$F1,$FC,$1E,$7F,$2D,$1F,$1A,$3F
	.dc.b	$B8,$CF,$F0,$58,$F8,$D1,$FE,$25,$9F,$8D,$5F,$E2,$BB,$FC,$A5,$FC
	.dc.b	$68,$FE,$D3,$3F,$C0,$CF,$C6,$47,$F8,$15,$FF,$1D,$BF,$C0,$DF,$C5
	.dc.b	$9F,$9C,$3F,$4B,$5F,$24,$FE,$08,$FC,$9B,$FE,$06,$FF,$98,$8F,$E6
	.dc.b	$75,$FE,$60,$7F,$7E,$7C,$98,$FE,$67,$3F,$E1,$DA,$FC,$84,$7F,$30
	.dc.b	$9F,$F2,$48,$3F,$D8,$F5,$FE,$FB,$3F,$C6,$CF,$E4,$0F,$F7,$6C,$B0
	.dc.b	$FE,$0F,$7F,$92,$9E,$67,$F9,$4E,$B8,$CF,$F0,$88,$FE,$03,$18,$AE
	.dc.b	$4E,$AB,$FC,$17,$3F,$EA,$43,$CB,$87,$FA,$FA,$FF,$C4,$0E,$8F,$C1
	.dc.b	$C7,$FB,$56,$7F,$D8,$D5,$FD,$BD,$FE,$BB,$1F,$E9,$A3,$DF,$7E,$E7
	.dc.b	$D6,$79,$5F,$73,$FE,$FF,$39,$9A,$8F,$2C,$7F,$95,$67,$F8,$54,$FA
	.dc.b	$8F,$8B,$7C,$CF,$B8,$FC,$D9,$FE,$DD,$3F,$C4,$6B,$DE,$7F,$08,$9F
	.dc.b	$F2,$B9,$CC,$FF,$6D,$3F,$C3,$A7,$EB,$BF,$E0,$4A,$FD,$43,$F6,$E7
	.dc.b	$EE,$47,$F6,$CB,$F9,$71,$EC,$3E,$F1,$2B,$B3,$83,$FC,$CF,$7F,$61
	.dc.b	$FB,$75,$F8,$53,$F8,$7D,$74,$7D,$B1,$FE,$72,$7F,$5B,$7F,$E5,$58
	.dc.b	$FB,$2F,$1A,$FA,$95,$F6,$63,$8B,$EB,$8F,$B1,$1F,$3A,$7F,$9D,$9F
	.dc.b	$66,$BF,$D2,$35,$9B,$FB,$73,$F5,$8F,$51,$F3,$CF,$C0,$BF,$5D,$FA
	.dc.b	$A7,$C0,$3E,$03,$FE,$05,$AF,$F1,$23,$F8,$6A,$E9,$3D,$23,$FD,$73
	.dc.b	$7F,$E9,$9A,$FB,$13,$F3,$1F,$CB,$46,$5F,$D0,$1F,$96,$8F,$28,$7E
	.dc.b	$03,$83,$0F,$E6,$1F,$E8,$67,$EA,$A3,$F6,$8F,$ED,$27,$F5,$EF,$98
	.dc.b	$EB,$3F,$55,$FB,$27,$BC,$FB,$D5,$FB,$37,$FA,$E1,$FE,$0C,$AF,$D9
	.dc.b	$4F,$D9,$8F,$B5,$3F,$B0,$8F,$F4,$85,$7D,$C3,$A4,$FF,$09,$BF,$F0
	.dc.b	$93,$F4,$4F,$AA,$68,$47,$F2,$17,$E1,$BF,$95,$7E,$C3,$B3,$FC,$BC
	.dc.b	$EA,$BF,$07,$1F,$C3,$63,$F5,$F3,$FB,$49,$FC,$9C,$FE,$BA,$7F,$67
	.dc.b	$5F,$AE,$3F,$16,$7D,$67,$E8,$3F,$DD,$5F,$F1,$5B,$F9,$C3,$EA,$BF
	.dc.b	$31,$FA,$8F,$E5,$0F,$E1,$47,$F0,$FB,$FE,$4E,$FE,$90,$FE,$01,$5F
	.dc.b	$C9,$A3,$F5,$91,$FC,$9E,$3F,$57,$3F,$DE,$AB,$F5,$E7,$C7,$9F,$C7
	.dc.b	$9F,$C7,$4F,$D5,$9F,$2E,$FD,$0F,$EE,$5F,$C7,$3F,$A2,$7D,$2D,$9F
	.dc.b	$D1,$BF,$28,$FD,$A1,$FC,$8A,$3F,$58,$FF,$BF,$A7,$E9,$3F,$AA,$AF
	.dc.b	$A2,$70,$7E,$A4,$FD,$19,$FD,$41,$F6,$67,$F9,$8D,$7E,$9A,$BF,$D9
	.dc.b	$87,$E0,$DF,$4B,$FA,$97,$E8,$1F,$9B,$70,$FF,$1A,$7E,$81,$E1,$3F
	.dc.b	$A0,$5F,$D0,$AF,$D1,$9F,$26,$3F,$55,$5F,$5C,$E0,$FF,$06,$7F,$44
	.dc.b	$7D,$73,$C0,$FF,$26,$7F,$A4,$DF,$A6,$3F,$AA,$47,$E8,$EB,$FA,$75
	.dc.b	$7E,$8E,$3F,$C0,$F1,$FE,$02,$9F,$D0,$C7,$F1,$8B,$F9,$91,$F3,$63
	.dc.b	$FC,$59,$1F,$EB,$38,$FE,$20,$7C,$9B,$F9,$8F,$F1,$49,$FF,$43,$CF
	.dc.b	$F9,$42,$3F,$A1,$BE,$C7,$E5,$4F,$E7,$EF,$27,$CF,$9F,$CC,$CF,$F0
	.dc.b	$83,$E3,$4F,$CD,$9F,$93,$5F,$97,$9F,$E2,$87,$F2,$93,$F8,$B3,$DD
	.dc.b	$F8,$D3,$FD,$0E,$7F,$33,$3F,$D0,$AF,$FB,$95,$7E,$62,$3E,$43,$FD
	.dc.b	$DE,$7E,$44,$FF,$40,$B3,$F2,$66,$9F,$DD,$8F,$C9,$DF,$E5,$A3,$FC
	.dc.b	$5F,$5F,$96,$AF,$E0,$55,$F2,$AF,$FD,$BD,$7F,$2A,$FF,$D0,$17,$F9
	.dc.b	$5B,$FE,$09,$7F,$28,$F8,$D5,$F2,$5F,$EE,$17,$FD,$BA,$FF,$7B,$1F
	.dc.b	$26,$3F,$B7,$57,$FB,$6A,$FE,$4D,$7F,$02,$8F,$E6,$55,$FB,$F8,$FF
	.dc.b	$11,$DF,$C7,$8F,$E0,$17,$F1,$DF,$DF,$1F,$CC,$27,$FB,$3D,$7F,$30
	.dc.b	$8F,$E6,$33,$F8,$E7,$E2,$1F,$18,$FE,$56,$7D,$53,$F9,$79,$FE,$1D
	.dc.b	$3F,$92,$9F,$9E,$33,$1F,$21,$FC,$94,$7C,$88,$FE,$C7,$1F,$91,$9F
	.dc.b	$F0,$C4,$7E,$46,$3F,$25,$3F,$EA,$2A,$F9,$15,$FD,$86,$FF,$21,$7F
	.dc.b	$B9,$BF,$91,$3F,$D6,$9F,$F3,$3B,$FB,$9B,$FC,$84,$7E,$38,$FE,$B2
	.dc.b	$FF,$26,$8F,$F2,$44,$7C,$87,$FA,$B9,$F8,$E7,$E4,$1F,$E4,$3B,$FF
	.dc.b	$07,$5F,$E3,$A3,$FC,$8D,$7F,$E0,$D7,$F6,$B5,$FE,$78,$AF,$C6,$47
	.dc.b	$F9,$12,$BF,$19,$5F,$B5,$8F,$8B,$5F,$EC,$9A,$FC,$64,$7F,$90,$2B
	.dc.b	$F1,$95,$F1,$6B,$FC,$05,$3F,$8C,$8F,$F1,$D3,$F8,$C8,$FF,$3C,$3F
	.dc.b	$E0,$77,$FC,$8D,$3F,$DD,$A3,$FD,$86,$FF,$96,$5F,$F3,$81,$FD,$BF
	.dc.b	$0F,$D2,$C6,$70,$FF,$1E,$E1,$FE,$0D,$BE,$77,$FC,$6C,$FF,$12,$38
	.dc.b	$7F,$C7,$87,$FA,$07,$06,$1D,$30,$FF,$0B,$E1,$FE,$14,$BE,$17,$E2
	.dc.b	$8F,$F2,$5C,$61,$FF,$23,$61,$CA,$31,$3F,$E2,$D6,$CF,$E5,$1B,$86
	.dc.b	$E7,$FC,$0B,$7F,$D5,$E3,$D5,$18,$3F,$87,$BE,$07,$07,$81,$5F,$39
	.dc.b	$F8,$8F,$F8,$BC,$E6,$E2,$3C,$53,$C8,$FF,$21,$3C,$63,$FC,$81,$1F
	.dc.b	$04,$D5,$FD,$3B,$F8,$13,$F4,$DB,$3E,$2A,$F9,$4F,$F4,$B9,$FA,$2F
	.dc.b	$27,$E8,$61,$A8,$FF,$3A,$5F,$79,$FB,$87,$F9,$64,$AF,$BA,$78,$4F
	.dc.b	$FA,$CE,$7E,$73,$D6,$7D,$B1,$C2,$BD,$93,$FE,$7F,$7F,$C9,$E7,$29
	.dc.b	$FF,$28,$C7,$FB,$0E,$FE,$E4,$FC,$9B,$D3,$F2,$CA,$76,$BF,$F2,$6B
	.dc.b	$FE,$5A,$8F,$2D,$7F,$7D,$AF,$C1,$BF,$66,$BF,$30,$FC,$93,$E6,$9F
	.dc.b	$EE,$57,$2F,$53,$F3,$47,$F9,$A6,$7A,$BF,$E2,$68,$E7,$7D,$5F,$D5
	.dc.b	$3F,$1E,$BF,$23,$87,$E0,$A7,$27,$BE,$BC,$35,$D6,$BF,$3F,$1F,$6F
	.dc.b	$0E,$0F,$E4,$8E,$15,$F8,$BB,$F9,$B1,$F9,$37,$6F,$E5,$63,$67,$F9
	.dc.b	$CA,$3E,$35,$76,$8F,$D9,$57,$D3,$7F,$80,$9F,$E7,$83,$E3,$1F,$84
	.dc.b	$AF,$C7,$4F,$E4,$23,$AC,$7E,$3C,$F5,$1F,$9D,$3F,$CE,$37,$F8,$D8
	.dc.b	$E8,$FC,$78,$FC,$B3,$F1,$DF,$C5,$5F,$A9,$FC,$74,$7D,$37,$FD,$81
	.dc.b	$7F,$01,$F4,$BB,$8F,$C5,$C7,$E1,$A7,$51,$FE,$9E,$75,$3F,$EF,$AA
	.dc.b	$FE,$34,$FC,$4A,$F4,$1F,$12,$FF,$12,$FB,$8F,$E3,$27,$9E,$7F,$0F
	.dc.b	$5E,$18,$FD,$7C,$7E,$2A,$BC,$D1,$FE,$DB,$75,$7F,$06,$3F,$91,$BC
	.dc.b	$E7,$F1,$35,$A7,$C1,$3F,$C8,$6B,$9C,$FE,$AA,$FF,$91,$3F,$EB,$D9
	.dc.b	$DD,$FF,$03,$8F,$82,$7E,$A6,$77,$5F,$2E,$BF,$88,$E1,$FA,$8C,$3E
	.dc.b	$04,$7F,$07,$3C,$C7,$EF,$CF,$31,$F7,$EF,$E8,$47,$C0,$3F,$11,$1F
	.dc.b	$82,$7F,$A5,$3F,$82,$AD,$E1,$D0,$FC,$2D,$7D,$6A,$FE,$16,$7F,$10
	.dc.b	$3D,$62,$3A,$3F,$EA,$89,$FB,$67,$33,$F8,$46,$1F,$C2,$9F,$CB,$DF
	.dc.b	$F2,$A7,$F0,$13,$FC,$21,$E7,$1F,$EA,$F9,$FB,$EF,$F1,$53,$F4,$33
	.dc.b	$FD,$95,$E6,$FF,$06,$3E,$7B,$FC,$B2,$3F,$72,$FF,$05,$8F,$E0,$6E
	.dc.b	$AB,$E4,$BF,$80,$BF,$BC,$F7,$8F,$1B,$FD,$B2,$FF,$50,$FF,$01,$3F
	.dc.b	$94,$4F,$B9,$FE,$8A,$FD,$D6,$C7,$DD,$9F,$E1,$58,$7D,$D9,$E6,$0F
	.dc.b	$B8,$7C,$E3,$CA,$7D,$FA,$FB,$AF,$3B,$F7,$3F,$BD,$7D,$CF,$EE,$8F
	.dc.b	$AF,$5C,$8F,$BB,$7F,$C3,$5E,$47,$BA,$3F,$73,$1F,$CF,$1F,$CC,$BE
	.dc.b	$D9,$FD,$AB,$F4,$1F,$B0,$69,$E5,$1F,$CF,$CF,$E7,$67,$E3,$4F,$B2
	.dc.b	$7F,$3A,$AF,$DB,$3F,$14,$F5,$CF,$E5,$9F,$59,$FB,$43,$EC,$47,$F0
	.dc.b	$A7,$F2,$B1,$FC,$E9,$FB,$13,$FB,$43,$E6,$BF,$97,$9F,$DA,$3F,$62
	.dc.b	$3E,$24,$7D,$79,$FF,$02,$1F,$B8,$7F,$89,$3F,$A8,$3F,$66,$7F,$31
	.dc.b	$9F,$AD,$1F,$A7,$7E,$21,$FC,$C6,$3E,$D4,$FF,$10,$7D,$71,$FC,$A8
	.dc.b	$FA,$B3,$FA,$FA,$E3,$1F,$83,$7F,$9B,$CF,$DA,$AF,$E3,$27,$F0,$F9
	.dc.b	$FD,$53,$FA,$F3,$EF,$1E,$B8,$FA,$B1,$FD,$DA,$3F,$90,$BE,$A8,$FD
	.dc.b	$53,$FA,$83,$F0,$C7,$F7,$17,$F5,$C6,$63,$8C,$FC,$C3,$F9,$8E,$1F
	.dc.b	$DB,$30,$FD,$46,$1F,$AA,$8F,$F1,$EC,$7F,$80,$A3,$D1,$7F,$E3,$1C
	.dc.b	$3E,$76,$1F,$48,$F8,$2F,$D7,$7F,$51,$3F,$4A,$7F,$C6,$71,$FA,$29
	.dc.b	$FA,$35,$FD,$9D,$E3,$87,$FA,$82,$BE,$8C,$FF,$2B,$8F,$44,$7F,$8A
	.dc.b	$CF,$EC,$37,$F4,$4F,$E4,$6F,$CF,$9F,$F0,$CD,$FA,$23,$F8,$B5,$FE
	.dc.b	$7E,$BF,$AD,$CF,$CF,$8F,$DD,$1F,$C9,$5F,$E3,$AE,$2F,$FC,$5F,$7F
	.dc.b	$E1,$88,$F9,$47,$F7,$E3,$F4,$8F,$F1,$F7,$D6,$7F,$3B,$3F,$C4,$33
	.dc.b	$F4,$8F,$EB,$2F,$E9,$23,$E8,$C7,$F1,$9B,$FE,$3C,$FE,$8C,$FE,$E9
	.dc.b	$7F,$A2,$C3,$FA,$8B,$FD,$52,$3E,$4B,$E8,$BF,$F3,$64,$7E,$65,$F9
	.dc.b	$27,$F4,$7A,$FD,$DB,$F9,$77,$F1,$0F,$E8,$4F,$A9,$1F,$A1,$C3,$E8
	.dc.b	$9C,$4F,$E2,$93,$FB,$A9,$FF,$08,$BF,$46,$FF,$C6,$35,$F4,$67,$E5
	.dc.b	$3E,$97,$F4,$37,$F4,$A3,$F0,$27,$A6,$3E,$93,$FC,$51,$FD,$01,$FE
	.dc.b	$4A,$3F,$41,$1F,$C2,$DF,$9E,$7F,$13,$38,$5F,$D6,$3F,$8D,$1F,$A6
	.dc.b	$3E,$A9,$FC,$64,$FE,$43,$1E,$78,$FF,$4A,$1F,$9E,$7E,$71,$F4,$A7
	.dc.b	$F3,$A7,$F2,$8B,$F2,$CF,$CB,$3E,$71,$F1,$8F,$CE,$BF,$83,$75,$1F
	.dc.b	$34,$FE,$F6,$FF,$0C,$C3,$F9,$F5,$7E,$72,$3F,$A0,$4F,$F0,$A3,$F7
	.dc.b	$47,$CB,$AF,$E7,$B3,$F2,$CF,$EF,$6F,$F3,$D8,$FE,$13,$3F,$E1,$23
	.dc.b	$F3,$E7,$F8,$0F,$0F,$ED,$EF,$EE,$8F,$D1,$CF,$F8,$A1,$FF,$4E,$C6
	.dc.b	$A7,$FC,$99,$1F,$C2,$8F,$F1,$FE,$1F,$C8,$67,$FB,$CB,$CE,$3F,$C4
	.dc.b	$11,$8A,$F0,$47,$0A,$E7,$7E,$B9,$F2,$CF,$FA,$CC,$DB,$FE,$22,$8F
	.dc.b	$DC,$47,$F7,$D8,$F3,$42,$7F,$CC,$F5,$FE,$82,$8F,$B9,$5F,$DD,$EB
	.dc.b	$F7,$0F,$B8,$F4,$1F,$DC,$CF,$84,$FD,$77,$81,$D0,$D5,$7C,$28,$FE
	.dc.b	$29,$1F,$31,$FF,$77,$57,$FA,$16,$F9,$9F,$E4,$79,$FE,$8A,$FF,$1D
	.dc.b	$AF,$96,$7C,$18,$FF,$5F,$3E,$4A,$FF,$20,$BF,$02,$3F,$D1,$F1,$F5
	.dc.b	$63,$E0,$3F,$18,$FF,$1E,$3F,$C5,$E3,$ED,$1F,$C6,$EB,$EF,$DF,$DA
	.dc.b	$8F,$71,$F7,$8F,$F3,$C9,$FD,$C8,$FB,$D5,$FD,$F6,$FC,$8F,$D6,$8F
	.dc.b	$E0,$47,$F9,$62,$BE,$F4,$FF,$AF,$8E,$55,$ED,$9F,$F1,$A1,$E7,$3F
	.dc.b	$26,$FB,$5F,$F7,$95,$7D,$CB,$FA,$0F,$E0,$5F,$23,$F6,$A3,$EA,$9C
	.dc.b	$CF,$B9,$3F,$55,$FA,$11,$F8,$1A,$FF,$1D,$57,$D9,$7F,$49,$1F,$62
	.dc.b	$7F,$C6,$91,$F6,$23,$F4,$47,$FA,$AA,$FE,$E5,$7D,$83,$F5,$87,$FA
	.dc.b	$26,$7F,$A8,$BF,$AB,$3F,$8E,$57,$E5,$5F,$5D,$7D,$6A,$FD,$8C,$FD
	.dc.b	$63,$E6,$9F,$36,$FE,$B3,$F8,$93,$D5,$1F,$B3,$AF,$9D,$1F,$7A,$F8
	.dc.b	$BF,$ED,$78,$FA,$57,$93,$E3,$47,$F3,$48,$FE,$FC,$7F,$05,$BF,$55
	.dc.b	$FE,$A6,$3F,$82,$47,$A8,$FC,$7D,$F0,$8F,$9D,$5F,$18,$F0,$9F,$A8
	.dc.b	$9F,$A5,$79,$3C,$30,$7F,$94,$BF,$54,$F9,$F5,$C9,$FF,$30,$DF,$EA
	.dc.b	$E7,$27,$23,$FC,$6B,$1F,$ED,$88,$FE,$A3,$7C,$4F,$C6,$1F,$48,$F4
	.dc.b	$BA,$7E,$6C,$7D,$13,$CA,$7F,$64,$3F,$89,$CF,$D5,$32,$FE,$94,$FE
	.dc.b	$02,$7F,$2C,$AF,$43,$FA,$43,$FD,$F9,$5F,$A4,$3F,$92,$D7,$F2,$03
	.dc.b	$E8,$1E,$FB,$FD,$11,$F3,$DF,$C3,$CF,$E9,$9F,$D0,$BF,$98,$3E,$35
	.dc.b	$F9,$DF,$E2,$EF,$CF,$BF,$CF,$CF,$F8,$AE,$7E,$69,$F2,$63,$F3,$CF
	.dc.b	$F8,$BE,$71,$3F,$9D,$8F,$E1,$CF,$CB,$3F,$38,$7F,$1B,$3E,$5B,$F6
	.dc.b	$E3,$F3,$8F,$F0,$A3,$E6,$4F,$F8,$A6,$BF,$37,$3F,$2E,$3F,$C5,$95
	.dc.b	$FC,$4D,$FE,$AD,$5F,$B3,$9F,$E2,$15,$97,$F3,$53,$FC,$36,$BE,$55
	.dc.b	$7F,$07,$3F,$1B,$5F,$2A,$3F,$98,$3E,$6A,$FE,$15,$1F,$98,$9F,$F5
	.dc.b	$E1,$F8,$B3,$FA,$09,$FC,$20,$FA,$27,$F4,$77,$F8,$43,$F2,$AB,$FC
	.dc.b	$AD,$5F,$C2,$EB,$F2,$B1,$FC,$F6,$BF,$81,$D7,$CA,$3F,$42,$FE,$50
	.dc.b	$F9,$47,$67,$FD,$77,$7F,$C1,$1F,$E7,$53,$F2,$67,$FC,$A5,$5F,$26
	.dc.b	$BF,$81,$1F,$DC,$2B,$F8,$11,$F9,$AA,$FE,$E4,$7F,$09,$3B,$3F,$E9
	.dc.b	$7A,$F9,$2F,$F0,$23,$E4,$BF,$C0,$0F,$92,$70,$BF,$ED,$71,$FB,$93
	.dc.b	$F2,$47,$C9,$7F,$7E,$FF,$B9,$EF,$E6,$4F,$C8,$8F,$C9,$57,$C8,$8F
	.dc.b	$C8,$47,$E4,$63,$E4,$5F,$EE,$A3,$E4,$47,$F2,$CB,$FE,$61,$7F,$94
	.dc.b	$3E,$5D,$FE,$CE,$BF,$12,$7E,$C8,$FF,$09,$1F,$CA,$4F,$82,$7F,$10
	.dc.b	$3F,$AE,$C7,$F5,$BA,$FD,$C9,$F2,$1F,$C7,$DF,$EE,$2B,$E3,$57,$C7
	.dc.b	$AF,$E5,$37,$FB,$BB,$FE,$B0,$7C,$AA,$FC,$75,$FF,$84,$63,$F8,$5D
	.dc.b	$FE,$E0,$FC,$7B,$FB,$78,$D4,$7F,$82,$CF,$EB,$F7,$FD,$58,$FD,$F9
	.dc.b	$FB,$88,$FE,$48,$FC,$88,$FD,$91,$FE,$0A,$3E,$4C,$7F,$24,$BF,$93
	.dc.b	$1F,$D5,$AB,$E4,$DF,$F2,$28,$FC,$7D,$7F,$81,$A3,$F1,$F1,$F9,$0B
	.dc.b	$FD,$BC,$7F,$50,$8F,$EA,$31,$FD,$FE,$FF,$6D,$1F,$D4,$23,$F6,$D3
	.dc.b	$F2,$2B,$FA,$64,$FE,$3A,$BE,$43,$FD,$4A,$FE,$45,$7E,$D2,$3F,$6B
	.dc.b	$6B,$F8,$B1,$FD,$FA,$7F,$1C,$FE,$3E,$7F,$8F,$CF,$F8,$C6,$3F,$6C
	.dc.b	$FF,$8C,$EF,$F6,$D1,$F8,$F8,$FE,$F7,$5F,$B6,$8F,$DB,$D7,$E3,$E3
	.dc.b	$E3,$47,$C7,$8F,$EF,$B1,$FD,$D0,$FF,$28,$70,$7F,$88,$E5,$3F,$E5
	.dc.b	$A8,$C3,$FE,$CE,$B6,$29,$CD,$04,$F4,$7F,$28,$7F,$94,$A3,$6F,$F9
	.dc.b	$7E,$79,$3E,$13,$FD,$53,$1D,$2B,$FD,$BA,$DA,$F0,$B6,$F8,$D7,$F9
	.dc.b	$C2,$BF,$D3,$53,$C6,$77,$3E,$DA,$FF,$0D,$47,$08,$FF,$07,$1E,$99
	.dc.b	$F7,$CF,$F9,$CC,$E3,$1D,$5C,$4F,$18,$FF,$77,$47,$F9,$69,$C1,$FE
	.dc.b	$DA,$BA,$FF,$9C,$A7,$E2,$3E,$7C,$7F,$D3,$77,$F3,$9F,$95,$7E,$09
	.dc.b	$FE,$E9,$18,$AE,$78,$7F,$19,$8F,$44,$FF,$01,$7E,$94,$F1,$8F,$F3
	.dc.b	$E4,$FF,$8E,$6F,$B3,$F3,$23,$83,$F6,$30,$EB,$7F,$82,$C3,$CF,$5E
	.dc.b	$2A,$FE,$51,$7C,$B0,$ED,$3F,$81,$3E,$6B,$E6,$3F,$D3,$8F,$98,$FA
	.dc.b	$CF,$CD,$C3,$4F,$E2,$0B,$86,$E7,$FC,$29,$1F,$E5,$19,$F9,$71,$DE
	.dc.b	$3C,$AF,$FB,$B4,$F8,$A7,$CA,$8A,$F5,$AF,$C8,$D7,$58,$F4,$5F,$3A
	.dc.b	$EB,$1F,$92,$3F,$33,$1F,$1D,$FC,$71,$F1,$63,$BC,$75,$8D,$CF,$57
	.dc.b	$F4,$91,$FD,$5E,$7D,$38,$7E,$4A,$FF,$1F,$87,$5B,$FF,$30,$4F,$CC
	.dc.b	$7E,$94,$75,$7F,$24,$FD,$27,$FA,$04,$7C,$59,$FD,$56,$1F,$82,$AF
	.dc.b	$B7,$5D,$4F,$7C,$7C,$59,$FC,$84,$75,$3F,$1E,$75,$3F,$2C,$7E,$DD
	.dc.b	$EB,$87,$E2,$E3,$F1,$27,$BC,$FE,$13,$3F,$16,$7F,$1F,$7F,$5A,$7E
	.dc.b	$1C,$E9,$FC,$7D,$75,$7F,$89,$5F,$C5,$7F,$AE,$3F,$8A,$7F,$0F,$5F
	.dc.b	$16,$7F,$D2,$8F,$58,$FE,$33,$3F,$B3,$BF,$E3,$2F,$C5,$8F,$E1,$C7
	.dc.b	$A4,$FB,$67,$C5,$3E,$B9,$F1,$4F,$E3,$27,$C5,$3F,$0E,$7F,$1B,$32
	.dc.b	$FD,$77,$FC,$28,$FE,$11,$F8,$95,$F8,$2A,$F4,$47,$DC,$AF,$E3,$D7
	.dc.b	$A8,$F4,$E1,$F8,$6C,$3C,$D1,$FC,$58,$FE,$03,$7F,$A9,$BF,$C4,$BF
	.dc.b	$07,$0F,$F6,$8C,$EA,$BE,$63,$FE,$DF,$C3,$F0,$97,$F8,$08,$F2,$C7
	.dc.b	$BD,$FB,$F7,$CE,$3D,$D1,$F8,$2A,$F7,$47,$E9,$63,$F8,$D5,$FC,$97
	.dc.b	$F0,$2F,$F0,$BB,$FC,$0D,$7D,$EA,$EF,$3E,$4A,$FF,$7F,$D7,$D6,$AE
	.dc.b	$53,$F7,$30,$FB,$F1,$FC,$12,$BF,$83,$BF,$3D,$F5,$9F,$58,$F9,$47
	.dc.b	$84,$FB,$38,$7C,$57,$D6,$7F,$A7,$CF,$E4,$F3,$FC,$02,$BE,$7C,$7D
	.dc.b	$E7,$C6,$FE,$6C,$FE,$07,$3C,$E3,$F4,$D1,$CD,$F6,$3F,$E9,$FA,$FA
	.dc.b	$93,$F9,$9B,$FA,$F7,$F5,$63,$33,$CA,$BF,$A5,$BC,$9F,$F3,$C3,$FB
	.dc.b	$A7,$EB,$9E,$D3,$EB,$BF,$BE,$C3,$2F,$D6,$C3,$D7,$3F,$B0,$7E,$6C
	.dc.b	$FA,$E3,$EC,$D7,$DF,$9F,$14,$6E,$BF,$38,$7E,$CD,$FE,$72,$FD,$63
	.dc.b	$F2,$4F,$ED,$9F,$D3,$3F,$A2,$AF,$D0,$C7,$EC,$63,$F6,$27,$F2,$E3
	.dc.b	$F9,$2E,$18,$BF,$DD,$47,$F4,$BA,$FE,$18,$DB,$E1,$7A,$8F,$F0,$B9
	.dc.b	$A2,$C7,$DB,$AD,$3E,$C8,$FE,$42,$FD,$F9,$EF,$3F,$56,$7F,$C0,$D1
	.dc.b	$E4,$6C,$6A,$39,$47,$F5,$7A,$FF,$13,$DF,$F8,$66,$BE,$34,$FF,$99
	.dc.b	$8F,$4C,$FF,$4E,$9F,$E6,$B1,$F1,$8F,$F6,$F9,$FD,$72,$3F,$C5,$47
	.dc.b	$47,$FC,$68,$FE,$79,$FD,$D5,$FE,$BD,$47,$C0,$9F,$E6,$B5,$F2,$A3
	.dc.b	$F9,$74,$7F,$38,$7A,$4F,$2B,$F9,$E7,$C1,$7F,$BD,$1F,$DB,$5F,$84
	.dc.b	$E0,$FE,$99,$5C,$23,$E1,$83,$C7,$7F,$DD,$9F,$87,$3F,$E3,$DB,$FC
	.dc.b	$0B,$E2,$3E,$7B,$F8,$4B,$FB,$51,$CA,$3F,$0F,$3F,$6A,$3E,$F5,$79
	.dc.b	$1F,$EC,$31,$FB,$9A,$C9,$D6,$3F,$2F,$5F,$56,$7E,$45,$FF,$85,$2F
	.dc.b	$F9,$73,$F7,$6F,$F1,$EF,$AD,$E5,$7F,$E6,$1A,$FD,$5B,$F5,$8F,$A5
	.dc.b	$1F,$75,$FA,$E7,$DD,$3F,$C6,$C7,$DD,$3D,$86,$A3,$EB,$57,$23,$D9
	.dc.b	$1F,$E3,$8A,$CD,$7E,$2A,$78,$CF,$D9,$AF,$9C,$7E,$0C,$F6,$5F,$F9
	.dc.b	$AD,$FF,$2B,$B5,$FA,$E7,$28,$FE,$62,$7D,$48,$FA,$B3,$FB,$5B,$FD
	.dc.b	$74,$FE,$80,$FC,$AC,$7E,$CA,$BF,$94,$47,$D4,$3E,$39,$FC,$6A,$BF
	.dc.b	$B7,$CF,$13,$C1,$1F,$50,$FD,$79,$F4,$27,$E0,$BF,$3E,$BE,$A3,$FD
	.dc.b	$8E,$FA,$BF,$E4,$79,$FC,$F3,$F4,$8E,$27,$D5,$9C,$CF,$EC,$A3,$C2
	.dc.b	$7F,$2C,$BF,$F0,$DB,$F3,$4E,$B1,$FD,$D5,$FD,$8D,$FF,$00,$8F,$EB
	.dc.b	$B3,$FE,$22,$BF,$D1,$57,$F1,$67,$E8,$BF,$E1,$A7,$F0,$47,$F8,$4D
	.dc.b	$C4,$FA,$A3,$F4,$B1,$FE,$B1,$9F,$A2,$7F,$23,$8F,$E4,$D3,$FA,$43
	.dc.b	$E0,$CF,$F0,$33,$FC,$2B,$3F,$39,$CD,$7E,$45,$FE,$02,$7F,$AB,$5F
	.dc.b	$82,$FF,$98,$2B,$F8,$E3,$F4,$4F,$9B,$7F,$88,$7F,$91,$DF,$D5,$33
	.dc.b	$3F,$A2,$8F,$E2,$27,$D1,$3E,$49,$FA,$18,$FE,$DE,$FF,$0C,$7F,$13
	.dc.b	$7F,$C8,$A7,$F4,$33,$F4,$6F,$F9,$11,$C4,$FA,$07,$A1,$FE,$1B,$1F
	.dc.b	$41,$FD,$09,$F2,$DF,$9E,$7E,$7A,$B0,$FE,$68,$FF,$01,$47,$F7,$C3
	.dc.b	$FB,$51,$FE,$03,$3F,$9A,$4F,$F5,$17,$FD,$13,$1F,$E0,$F8,$FE,$02
	.dc.b	$7F,$91,$A3,$FD,$03,$7F,$E3,$C8,$E3,$1F,$40,$FF,$43,$BF,$DE,$DF
	.dc.b	$F5,$D4,$FF,$80,$1F,$F2,$2C,$FF,$6A,$7F,$CF,$AF,$F8,$12,$B3,$3C
	.dc.b	$67,$E7,$BF,$E3,$57,$FB,$D9,$EA,$3F,$CF,$A7,$AA,$3E,$8B,$FE,$70
	.dc.b	$39,$3F,$E3,$97,$FD,$78,$7F,$8D,$27,$FC,$D9,$1F,$A1,$8F,$F1,$EB
	.dc.b	$FE,$38,$9F,$F0,$19,$F0,$8F,$E9,$11,$FE,$B6,$7C,$4F,$F9,$8A,$7F
	.dc.b	$CA,$91,$FA,$18,$FF,$04,$BF,$E3,$97,$E5,$9F,$EE,$23,$FC,$6F,$87
	.dc.b	$FA,$E2,$7E,$BD,$7F,$8F,$2B,$E9,$1F,$E3,$1A,$FF,$34,$5F,$81,$FF
	.dc.b	$00,$D7,$F8,$D6,$B3,$5F,$E7,$58,$FD,$7C,$FF,$AB,$0F,$9A,$FF,$A2
	.dc.b	$CF,$D0,$3F,$EA,$E9,$FA,$33,$FE,$60,$8F,$C0,$C7,$9D,$FF,$33,$DF
	.dc.b	$FA,$BA,$BE,$A4,$FF,$8C,$A3,$F3,$75,$FE,$2D,$AF,$9D,$1F,$EA,$0A
	.dc.b	$FF,$18,$57,$CD,$8F,$E7,$75,$F2,$0F,$F1,$71,$ED,$BF,$F4,$FB,$FE
	.dc.b	$63,$C3,$FC,$D0,$FC,$B3,$F3,$B8,$7F,$98,$B0,$FF,$47,$BF,$94,$3F
	.dc.b	$C4,$55,$FE,$DF,$AF,$E6,$B5,$F2,$E7,$F3,$77,$F9,$C8,$FF,$6E,$E1
	.dc.b	$F2,$A3,$FD,$29,$1F,$99,$9F,$F2,$D4,$7F,$96,$E3,$FD,$21,$3F,$22
	.dc.b	$3F,$B0,$47,$E5,$E7,$FC,$B3,$87,$E3,$9F,$CB,$1E,$4A,$FF,$39,$DF
	.dc.b	$F2,$DA,$FF,$62,$CF,$C8,$8F,$F4,$59,$FE,$13,$7F,$CB,$93,$F2,$63
	.dc.b	$FA,$F4,$FC,$99,$FF,$28,$BF,$24,$FB,$AF,$E5,$E7,$31,$F2,$44,$F5
	.dc.b	$9F,$F2,$69,$FC,$1E,$BE,$59,$F9,$47,$F8,$2D,$FF,$92,$EB,$FD,$81
	.dc.b	$1F,$2A,$B9,$61,$FE,$C1,$9F,$97,$3D,$8F,$54,$1F,$F6,$0C,$FF,$85
	.dc.b	$E7,$E7,$3F,$E8,$83,$FD,$7E,$FC,$E8,$FF,$5F,$3F,$95,$3E,$73,$F9
	.dc.b	$3A,$FC,$A4,$7F,$A1,$63,$C0,$7F,$AE,$E3,$F2,$B1,$F3,$A3,$F2,$91
	.dc.b	$F9,$58,$FF,$5C,$47,$FB,$D9,$FE,$02,$7C,$D7,$FD,$E7,$1F,$EB,$69
	.dc.b	$FF,$5D,$61,$F3,$23,$FC,$FD,$3F,$E3,$C3,$FA,$C9,$FD,$54,$FF,$06
	.dc.b	$1F,$CF,$4F,$EB,$87,$FB,$D2,$7F,$AE,$5F,$F9,$EE,$FF,$C7,$D3,$FE
	.dc.b	$3F,$8F,$A1,$3F,$E3,$D8,$FE,$B5,$1F,$E3,$C9,$FF,$04,$47,$F8,$26
	.dc.b	$3F,$33,$1F,$E3,$38,$FF,$02,$1F,$D6,$30,$FF,$03,$BF,$E3,$7C,$3F
	.dc.b	$AD,$CF,$F5,$68,$FE,$AA,$7F,$81,$AB,$FA,$B4,$FF,$8D,$4C,$4F,$F5
	.dc.b	$5B,$FF,$18,$57,$F5,$5B,$FF,$34,$57,$F8,$1E,$BF,$AE,$1F,$C5,$27
	.dc.b	$FC,$DD,$5F,$E0,$27,$F2,$47,$E1,$6B,$FC,$67,$5F,$E0,$3B,$FF,$19
	.dc.b	$57,$F8,$06,$BF,$C6,$71,$F1,$E3,$FC,$63,$3F,$E6,$48,$F8,$F3,$FE
	.dc.b	$30,$3F,$06,$7F,$9F,$6B,$FC,$32,$FE,$B5,$FD,$3B,$FE,$84,$7F,$9D
	.dc.b	$3B,$BF,$EE,$91,$E9,$AF,$E4,$93,$FE,$A1,$AF,$F3,$04,$FC,$CB,$FF
	.dc.b	$50,$D7,$CC,$BF,$97,$7F,$32,$FF,$C0,$37,$FE,$60,$8F,$F0,$1D,$7C
	.dc.b	$CA,$FE,$FF,$1F,$E0,$0A,$FF,$30,$D7,$CC,$8F,$F3,$1D,$7F,$80,$2B
	.dc.b	$F2,$31,$F9,$2A,$FF,$2E,$47,$F8,$B2,$3F,$C5,$B1,$F2,$9F,$F1,$31
	.dc.b	$FE,$5B,$7E,$53,$FE,$29,$9F,$95,$7F,$CA,$2B,$E4,$CF,$C5,$7F,$D1
	.dc.b	$A7,$F9,$49,$5F,$CA,$7F,$D1,$55,$FE,$2B,$BF,$C9,$DF,$FB,$12,$7F
	.dc.b	$27,$1F,$22,$FF,$25,$1F,$93,$8F,$E5,$71,$F2,$27,$FC,$B3,$1F,$22
	.dc.b	$3E,$3C,$FC,$88,$FD,$EC,$FE,$4A,$3E,$43,$FE,$83,$BF,$90,$FF,$A1
	.dc.b	$6F,$E4,$3F,$EB,$69,$F9,$0F,$F9,$4C,$FF,$5B,$3F,$21,$FF,$5A,$DF
	.dc.b	$FA,$0A,$3F,$23,$1F,$E8,$0A,$FF,$78,$D7,$C8,$9F,$F2,$95,$7F,$94
	.dc.b	$1F,$92,$7F,$AB,$4F,$F2,$6C,$FF,$A0,$2B,$FD,$71,$5F,$E5,$0A,$FF
	.dc.b	$3F,$D7,$C6,$9F,$F7,$5D,$7E,$36,$3F,$D6,$35,$F8,$D9,$FF,$3E,$C7
	.dc.b	$E3,$5F,$F7,$2D,$7C,$67,$FD,$C9,$7F,$1A,$7F,$CF,$07,$ED,$4F,$DC
	.dc.b	$4F,$E3,$1F,$C6,$C7,$F7,$68,$FC,$44,$7F,$B9,$63,$F7,$2F,$F8,$6A
	.dc.b	$7F,$A0,$BF,$B9,$9F,$7B,$FD,$32,$BF,$C3,$98,$7E,$D2,$3F,$C8,$E7
	.dc.b	$F8,$73,$0F,$F2,$61,$FE,$1B,$78,$E1,$FD,$1C,$FF,$22,$47,$F8,$71
	.dc.b	$FC,$7D,$FF,$4B,$9F,$EE,$B8,$7F,$48,$3F,$A9,$D7,$EE,$A3,$55,$FE
	.dc.b	$AB,$BF,$C6,$D7,$F5,$8C,$3F,$77,$5F,$E4,$CB,$FE,$F0,$7F,$87,$9F
	.dc.b	$1C,$7F,$78,$7F,$D4,$F8,$7F,$87,$DF,$E9,$75,$FD,$E1,$FE,$BF,$3F
	.dc.b	$E6,$57,$CC,$30,$FF,$3E,$E1,$F3,$6B,$FB,$4D,$FF,$93,$AB,$FD,$30
	.dc.b	$FF,$7B,$3F,$CE,$E7,$AD,$F8,$88,$FF,$34,$3F,$D3,$0F,$F3,$C1,$FE
	.dc.b	$25,$9F,$F2,$74,$7C,$C8,$FF,$33,$9F,$E0,$07,$FC,$9E,$7F,$7C,$9F
	.dc.b	$B7,$5F,$E3,$39,$FF,$33,$1E,$A7,$FC,$DD,$1A,$9F,$F2,$81,$FE,$88
	.dc.b	$C3,$D5,$5F,$E0,$37,$FC,$A1,$1F,$E9,$07,$D5,$1F,$0D,$F5,$47,$F7
	.dc.b	$D3,$E7,$C7,$FA,$02,$3E,$9C,$7F,$7D,$9F,$B1,$3F,$E7,$EA,$FF,$2B
	.dc.b	$9F,$83,$8F,$B8,$FF,$92,$F0,$F4,$CF,$E0,$9F,$A8,$D8,$FF,$48,$57
	.dc.b	$E3,$4F,$4C,$7D,$77,$FD,$30,$FF,$A0,$6F,$FC,$B1,$5F,$E0,$1B,$FE
	.dc.b	$0A,$7E,$8A,$7F,$C5,$58,$7D,$08,$C4,$FE,$89,$FE,$58,$7C,$E9,$FD
	.dc.b	$14,$FF,$94,$E3,$E8,$CF,$F9,$4D,$FF,$13,$5F,$E9,$2F,$CB,$87,$E8
	.dc.b	$9F,$9A,$7F,$AA,$F0,$FF,$13,$E1,$F9,$E7,$F2,$67,$CB,$7F,$C4,$F7
	.dc.b	$F2,$B0,$FF,$3D,$3F,$E5,$33,$E4,$CF,$F9,$4C,$FC,$DD,$FF,$AB,$CF
	.dc.b	$F2,$D1,$F2,$30,$FF,$53,$1F,$E7,$AC,$18,$7F,$89,$EB,$FC,$83,$3F
	.dc.b	$80,$AF,$BE,$FF,$94,$EF,$E4,$4F,$E5,$DF,$95,$7F,$E5,$17,$E6,$4F
	.dc.b	$E5,$9F,$91,$3F,$E5,$13,$FC,$9A,$7D,$E3,$F8,$41,$F7,$8B,$5F,$C2
	.dc.b	$A7,$F2,$C7,$E1,$E3,$FC,$AB,$5F,$79,$FE,$0E,$FF,$5B,$7F,$2D,$1F
	.dc.b	$E8,$C8,$FC,$BD,$FC,$D9,$FE,$49,$87,$F0,$A7,$FD,$A5,$3F,$98,$C3
	.dc.b	$E7,$47,$E7,$2F,$E2,$C7,$CF,$8F,$9A,$7C,$F3,$9E,$1F,$3E,$FF,$D0
	.dc.b	$11,$FE,$77,$AF,$CB,$1F,$3F,$0F,$9D,$1F,$CD,$63,$E7,$1E,$6B,$FE
	.dc.b	$BC,$7F,$8B,$E3,$FC,$5F,$87,$F8,$B9,$FE,$BC,$7F,$8C,$8F,$D9,$C7
	.dc.b	$F8,$BC,$FF,$2A,$9F,$D7,$E7,$FC,$AA,$FF,$8C,$30,$FE,$79,$3F,$E2
	.dc.b	$73,$FC,$AB,$1D,$A7,$F9,$BB,$FD,$7A,$3F,$CA,$CF,$E7,$0F,$F0,$23
	.dc.b	$FE,$57,$9F,$F0,$24,$7F,$8A,$A3,$FC,$08,$FF,$91,$B0,$FE,$3D,$1F
	.dc.b	$E8,$B7,$FD,$33,$1F,$E0,$63,$FB,$09,$B9,$47,$F9,$65,$FE,$C5,$3F
	.dc.b	$EA,$DA,$FD,$FB,$FC,$48,$FE,$C7,$5F,$E0,$77,$FB,$1B,$FE,$B0,$8F
	.dc.b	$F1,$7B,$F2,$E3,$FD,$55,$7F,$2E,$3E,$54,$7F,$9F,$CF,$96,$7F,$8C
	.dc.b	$0F,$F0,$49,$FE,$60,$AF,$8A,$FF,$8C,$4F,$F4,$04,$7F,$06,$7F,$CB
	.dc.b	$4F,$FA,$06,$3F,$D0,$0F,$CB,$3F,$C1,$2F,$F8,$BA,$3F,$D7,$97,$F9
	.dc.b	$37,$E3,$1F,$E3,$77,$FB,$B4,$FC,$D3,$FC,$AF,$3F,$9B,$9F,$EA,$E7
	.dc.b	$C8,$7F,$4C,$FF,$B4,$E7,$FA,$D3,$FE,$85,$AF,$F2,$D5,$7F,$83,$8F
	.dc.b	$CE,$47,$CD,$BE,$2F,$47,$FC,$E2,$D7,$E0,$BB,$AF,$F4,$41,$FE,$36
	.dc.b	$7E,$59,$FE,$36,$3F,$2D,$5F,$3D,$FC,$C4,$7F,$AC,$27,$FB,$25,$7E
	.dc.b	$66,$7F,$CB,$F7,$FE,$5E,$7F,$A4,$C7,$E6,$4F,$9C,$FF,$A0,$E7,$FC
	.dc.b	$BA,$FF,$8D,$9F,$F5,$6C,$7C,$E3,$F3,$11,$F3,$8F,$69,$FD,$FA,$BF
	.dc.b	$C1,$F1,$FE,$38,$9F,$F2,$EB,$F3,$8F,$CD,$4F,$E1,$0F,$99,$5C,$9F
	.dc.b	$F0,$E5,$7C,$E8,$FF,$30,$4F,$E6,$E3,$F3,$4F,$CE,$9F,$94,$7F,$8E
	.dc.b	$E7,$FD,$0F,$5F,$E3,$B9,$FF,$30,$47,$F8,$3E,$7E,$5C,$7E,$65,$F8
	.dc.b	$0F,$F9,$80,$FC,$CD,$7C,$BA,$FF,$30,$47,$F0,$29,$FC,$C5,$FF,$98
	.dc.b	$1F,$CB,$9F,$E6,$09,$FF,$20,$BF,$E3,$63,$F8,$19,$FA,$67,$EF,$BF
	.dc.b	$EA,$DA,$FF,$43,$57,$CC,$8F,$C7,$1F,$0C,$FC,$C9,$FE,$D9,$9F,$96
	.dc.b	$FF,$98,$AF,$F3,$0F,$D1,$78,$3F,$99,$3F,$2D,$1F,$E6,$37,$F4,$B1
	.dc.b	$F9,$D9,$F9,$51,$FE,$36,$9F,$E0,$67,$E5,$9C,$3E,$86,$D7,$E5,$AB
	.dc.b	$E6,$47,$FB,$6A,$7F,$22,$FF,$54,$3F,$C0,$75,$FD,$72,$3F,$C6,$51
	.dc.b	$FC,$50,$AF,$F5,$D7,$FA,$DC,$FC,$13,$FC,$C5,$7F,$E6,$69,$FF,$1D
	.dc.b	$47,$1A,$FF,$19,$61,$FC,$06,$FF,$CE,$D7,$FE,$06,$3F,$AE,$3F,$E3
	.dc.b	$3B,$FF,$05,$1F,$D6,$CF,$F2,$54,$FC,$A8,$FF,$38,$61,$FC,$27,$0F
	.dc.b	$35,$FF,$5B,$9F,$F0,$5C,$FF,$82,$1F,$F3,$1D,$FF,$99,$AB,$D1,$3F
	.dc.b	$E3,$DB,$C5,$FC,$4B,$FE,$BB,$3F,$D7,$1F,$E0,$D7,$97,$FC,$9C,$FF
	.dc.b	$87,$8F,$4B,$FE,$33,$7F,$CC,$67,$C1,$7F,$C1,$AF,$F9,$89,$FD,$61
	.dc.b	$FE,$57,$3D,$40,$FF,$2E,$1F,$6C,$FF,$03,$9C,$CF,$F4,$F1,$FE,$7F
	.dc.b	$3F,$CE,$AF,$F9,$95,$FA,$B3,$F4,$F0,$FF,$1B,$C7,$D4,$9F,$F3,$19
	.dc.b	$EB,$7F,$C1,$6F,$DB,$9F,$5E,$1F,$E3,$98,$FA,$77,$E5,$BF,$F4,$CC
	.dc.b	$7D,$47,$F8,$3B,$CE,$FE,$A9,$FE,$3B,$8F,$F3,$1D,$FF,$A5,$DF,$F3
	.dc.b	$2C,$FF,$83,$A7,$EA,$C7,$F8,$F1,$FF,$4B,$9F,$E3,$C9,$E8,$FF,$A5
	.dc.b	$6B,$FC,$50,$DB,$FF,$1F,$5E,$9F,$EC,$D1,$FE,$B6,$8F,$59,$FE,$27
	.dc.b	$8F,$F3,$DC,$FF,$99,$67,$FC,$19,$3C,$5F,$A6,$7D,$2A,$FF,$1F,$3F
	.dc.b	$A7,$8F,$D1,$3F,$EC,$63,$FC,$7C,$7D,$4A,$FB,$0F,$F2,$5A,$FF,$33
	.dc.b	$3F,$C3,$67,$CD,$5F,$A1,$4F,$F9,$99,$F9,$F1,$FE,$66,$BF,$9B,$1F
	.dc.b	$E9,$08,$FA,$47,$F9,$A2,$7F,$40,$FF,$8F,$8F,$C3,$1F,$EA,$3A,$FC
	.dc.b	$33,$FD,$A1,$FC,$B5,$7C,$F3,$FD,$27,$3F,$4E,$7F,$C7,$AF,$CB,$9F
	.dc.b	$9F,$7F,$EC,$4A,$FF,$07,$1F,$90,$AF,$19,$FE,$3D,$BF,$F6,$25,$FF
	.dc.b	$A4,$A3,$E9,$47,$E8,$4F,$9C,$FF,$10,$3F,$89,$47,$CE,$7F,$3A,$7F
	.dc.b	$B2,$2B,$F4,$EF,$FB,$98,$FF,$15,$9F,$9E,$C3,$FD,$28,$FF,$14,$C3
	.dc.b	$E7,$BF,$D9,$EF,$E7,$9F,$9D,$3E,$7C,$FF,$0E,$BF,$F4,$A6,$1F,$9C
	.dc.b	$9F,$F5,$F3,$FE,$BA,$AF,$69,$F2,$CF,$F5,$C1,$F7,$B0,$FE,$4D,$87
	.dc.b	$F0,$F8,$FE,$CF,$5F,$EF,$1A,$F9,$B8,$7F,$0E,$8F,$E2,$55,$FC,$36
	.dc.b	$BF,$C1,$D1,$F9,$93,$F8,$61,$F2,$8F,$93,$1F,$97,$8F,$F1,$CD,$7F
	.dc.b	$65,$7F,$C4,$78,$7F,$83,$DF,$F2,$64,$7F,$B0,$1F,$F1,$E3,$FD,$95
	.dc.b	$FD,$CB,$FE,$6B,$C3,$F0,$F1,$FE,$7C,$8F,$F0,$7B,$FE,$67,$3F,$B3
	.dc.b	$CF,$F4,$F7,$F8,$B1,$FC,$16,$3F,$C7,$CF,$F6,$9A,$57,$F3,$F8,$FC
	.dc.b	$81,$FD,$82,$3F,$C2,$2F,$F8,$56,$BF,$7A,$FC,$A3,$FC,$D6,$7F,$86
	.dc.b	$6F,$33,$F8,$08,$FF,$09,$9F,$DA,$2F,$FC,$20,$FF,$7C,$8F,$E0,$91
	.dc.b	$FD,$A2,$3C,$F3,$FD,$A0,$FE,$5D,$7F,$E1,$29,$FF,$35,$1F,$E1,$23
	.dc.b	$F3,$B3,$FE,$12,$3F,$CC,$F1,$FE,$08,$7F,$CA,$A7,$F8,$F0,$FF,$4A
	.dc.b	$9F,$DB,$67,$FC,$26,$FF,$8F,$27,$FD,$28,$7C,$FA,$FC,$5C,$7C,$F3
	.dc.b	$FC,$29,$1F,$E9,$37,$FC,$26,$7F,$8F,$EF,$E2,$8A,$FA,$F1,$F4,$A3
	.dc.b	$15,$F4,$2B,$FC,$2D,$3F,$40,$EE,$7F,$8E,$0F,$F3,$71,$FC,$58,$FA
	.dc.b	$0F,$F9,$05,$FA,$15,$F2,$EF,$EC,$1F,$E7,$53,$85,$7F,$32,$9F,$F3
	.dc.b	$93,$E1,$7F,$9D,$9F,$EC,$39,$FF,$0F,$9F,$89,$7F,$AF,$3F,$ED,$97
	.dc.b	$F9,$99,$FE,$9A,$38,$3F,$84,$7F,$D0,$C7,$E4,$0E,$C7,$FA,$E6,$BF
	.dc.b	$C8,$8F,$F9,$BA,$BE,$B3,$FC,$20,$FE,$DC,$7F,$A7,$0F,$D0,$9F,$E9
	.dc.b	$B8,$FF,$37,$3F,$48,$FF,$4D,$9F,$E4,$33,$FC,$DC,$FE,$79,$FF,$39
	.dc.b	$57,$89,$FD,$8D,$77,$7F,$C8,$67,$FB,$1E,$7F,$CD,$C7,$FA,$6D,$FF
	.dc.b	$22,$3F,$CF,$0F,$F6,$13,$F5,$8F,$F6,$45,$7C,$FB,$FF,$4C,$C7,$F9
	.dc.b	$12,$FC,$0F,$F9,$2E,$FF,$52,$7E,$75,$FF,$62,$9F,$46,$3F,$C8,$57
	.dc.b	$F9,$C8,$FE,$12,$7E,$71,$FF,$61,$D7,$F0,$98,$FF,$28,$C7,$F0,$97
	.dc.b	$FD,$5E,$7F,$84,$9F,$E1,$4E,$2F,$F1,$4F,$E0,$CF,$D8,$9F,$2C,$FF
	.dc.b	$4C,$1F,$9C,$3E,$3C,$FE,$72,$FF,$37,$1F,$E2,$18,$FC,$DC,$FF,$09
	.dc.b	$8F,$F4,$C1,$FC,$24,$FC,$CC,$7F,$04,$78,$9F,$CC,$A3,$FC,$8C,$FC
	.dc.b	$A3,$E6,$1F,$E6,$F7,$FD,$2E,$FC,$B3,$F3,$D7,$F3,$1F,$F4,$EC,$7D
	.dc.b	$B7,$FC,$17,$7F,$BD,$7E,$69,$FE,$C3,$3F,$30,$7F,$91,$AF,$F2,$27
	.dc.b	$ED,$F8,$7E,$4C,$7F,$B5,$0F,$E5,$2E,$2B,$F7,$58,$A3,$FC,$E4,$66
	.dc.b	$7F,$DC,$AD,$7F,$CC,$6F,$F3,$9A,$ED,$1F,$E7,$43,$FC,$95,$1C,$6B
	.dc.b	$FA,$FB,$71,$DC,$FF,$85,$F1,$DB,$FE,$7C,$B6,$1E,$38,$FF,$0C,$45
	.dc.b	$7F,$CF,$2F,$A9,$FC,$15,$7F,$15,$38,$DF,$F9,$19,$EF,$1F,$B9,$C3
	.dc.b	$FC,$3F,$5F,$E8,$DA,$FE,$E1,$7D,$6B,$FA,$3C,$F0,$7F,$A7,$D7,$FA
	.dc.b	$4D,$F3,$9F,$E1,$FB,$EE,$FF,$88,$E7,$93,$C3,$0F,$92,$FE,$5D,$FE
	.dc.b	$23,$7A,$8F,$BF,$3D,$DF,$F4,$6D,$FB,$5F,$F1,$1C,$ED,$FF,$28,$5E
	.dc.b	$DF,$F2,$95,$F9,$1F,$F1,$85,$7D,$67,$F8,$DD,$F1,$8F,$F2,$AE,$1F
	.dc.b	$E7,$8B,$F4,$D7,$F8,$A5,$E9,$3F,$C9,$27,$FC,$FB,$5A,$AF,$E5,$91
	.dc.b	$F8,$37,$C0,$7F,$89,$4F,$BE,$7F,$8B,$23,$C0,$FC,$23,$FC,$5A,$F3
	.dc.b	$7F,$C7,$11,$F8,$77,$99,$FD,$98,$FF,$53,$3F,$89,$8E,$73,$F6,$67
	.dc.b	$F1,$18,$7C,$57,$EE,$CF,$AA,$FF,$AC,$47,$43,$FC,$0A,$FE,$1E,$7E
	.dc.b	$23,$F8,$99,$A7,$59,$FC,$7B,$E7,$7F,$AD,$3F,$1B,$0F,$E8,$98,$7F
	.dc.b	$A3,$AF,$E7,$47,$F0,$B9,$B5,$CC,$FF,$1D,$3C,$DF,$E5,$EF,$3C,$3F
	.dc.b	$01,$1F,$6B,$0D,$42,$BD,$35,$DA,$3A,$57,$E3,$DF,$C0,$3F,$C0,$F0
	.dc.b	$FF,$33,$3F,$81,$9F,$81,$1F,$81,$70,$72,$8F,$F0,$AB,$FD,$FC,$F3
	.dc.b	$3F,$E1,$A7,$ED,$BE,$63,$F0,$17,$E6,$7E,$04,$FF,$07,$BF,$A1,$7F
	.dc.b	$E2,$2C,$3E,$B5,$7F,$B5,$EB,$F0,$0F,$DC,$BE,$55,$FB,$69,$FF,$64
	.dc.b	$9F,$EF,$A3,$F8,$DC,$FD,$EC,$3E,$BB,$DA,$BD,$AF,$F0,$48,$E4,$DC
	.dc.b	$35,$1F,$1A,$3F,$7E,$FD,$67,$CF,$1F,$D0,$F0,$FB,$35,$F7,$E7,$E3
	.dc.b	$1F,$EC,$33,$91,$FD,$8E,$BE,$A3,$CB,$0F,$D1,$BF,$E9,$C8,$FB,$17
	.dc.b	$97,$E6,$9F,$3C,$FA,$71,$F2,$1F,$5B,$FB,$57,$EB,$C6,$5F,$0C,$7F
	.dc.b	$85,$4F,$B1,$87,$ED,$5F,$AF,$7F,$55,$F8,$67,$FA,$7A,$BE,$AB,$FC
	.dc.b	$52,$7F,$9E,$57,$EB,$A3,$37,$EA,$7F,$95,$1F,$92,$9F,$AA,$FE,$C9
	.dc.b	$FA,$B7,$EB,$7F,$DE,$F1,$F4,$23,$F3,$67,$EB,$E3,$FD,$F7,$7F,$52
	.dc.b	$FF,$1E,$FF,$8D,$9F,$A9,$5F,$4B,$0F,$A7,$1F,$AC,$9F,$AF,$3F,$08
	.dc.b	$FD,$5C,$7D,$27,$F4,$D7,$FA,$A9,$F4,$47,$F3,$3C,$3E,$D9,$FA,$D3
	.dc.b	$93,$F4,$EB,$E9,$BF,$A6,$AF,$D9,$C7,$A6,$3F,$C2,$4F,$F9,$89,$FD
	.dc.b	$5B,$FA,$87,$D5,$87,$D3,$3F,$B5,$4F,$D3,$3F,$97,$C7,$F2,$53,$FC
	.dc.b	$8F,$3C,$6B,$ED,$C7,$F6,$D3,$F4,$D8,$7E,$9E,$7F,$91,$E1,$F5,$9F
	.dc.b	$DB,$DF,$D1,$AF,$D1,$1F,$46,$BF,$8E,$C7,$F2,$59,$F8,$2F,$EF,$E3
	.dc.b	$F9,$55,$7A,$2B,$FA,$C1,$FB,$F3,$FD,$32,$FE,$97,$0F,$F0,$53,$FC
	.dc.b	$02,$36,$FD,$4B,$FB,$78,$7A,$A7,$F2,$CF,$CF,$7F,$4D,$87,$D5,$8F
	.dc.b	$CA,$C7,$F5,$AA,$FD,$36,$1F,$CF,$6F,$D9,$3F,$EF,$F3,$F4,$57,$F5
	.dc.b	$23,$8C,$7F,$78,$7E,$A6,$1F,$D6,$AB,$F8,$8C,$F8,$CF,$F3,$A9,$FA
	.dc.b	$23,$F3,$C7,$F1,$C9,$FE,$3F,$79,$7E,$A1,$F6,$30,$F8,$93,$FE,$D0
	.dc.b	$3F,$AB,$9F,$E5,$53,$F1,$E7,$F8,$7C,$FE,$B3,$1F,$58,$F1,$3F,$E0
	.dc.b	$EC,$3F,$AB,$BF,$5A,$FE,$BC,$FF,$54,$BF,$D3,$47,$D6,$9E,$27,$D4
	.dc.b	$3C,$77,$FD,$1D,$F5,$99,$BF,$E3,$55,$F5,$6B,$EF,$D7,$2A,$FF,$74
	.dc.b	$CF,$F9,$1F,$0F,$BB,$3C,$A7,$FC,$23,$87,$F1,$A7,$FA,$93,$FC,$68
	.dc.b	$E5,$1F,$E0,$E8,$FD,$2C,$71,$8F,$E9,$98,$7F,$83,$A3,$EA,$CE,$9F
	.dc.b	$D4,$3C,$8F,$ED,$A7,$CB,$39,$1F,$8F,$3F,$8D,$1F,$D5,$1F,$AA,$7F
	.dc.b	$19,$7F,$4F,$5F,$E0,$93,$F8,$BB,$F9,$21,$5F,$C0,$2B,$F1,$B1,$FE
	.dc.b	$47,$7E,$9C,$FF,$92,$F0,$FA,$6F,$F1,$89,$FF,$03,$DF,$F1,$87,$E0
	.dc.b	$E1,$E8,$7F,$C8,$D5,$FC,$62,$BF,$4F,$1F,$C4,$27,$F9,$83,$FE,$76
	.dc.b	$7F,$B2,$3F,$C5,$4F,$D2,$CF,$F1,$AA,$FA,$15,$F5,$1F,$A0,$7C,$B3
	.dc.b	$F0,$27,$F4,$89,$FF,$30,$3F,$C5,$DF,$D8,$BF,$41,$AF,$F2,$07,$F8
	.dc.b	$01,$F1,$4F,$C5,$9F,$BF,$3F,$C7,$91,$FD,$F5,$FA,$13,$FD,$F3,$0F
	.dc.b	$B0,$70,$3E,$B9,$FB,$48,$FF,$1E,$3F,$CC,$1F,$A3,$5F,$A3,$3F,$8A
	.dc.b	$3F,$C5,$EB,$FC,$F5,$1F,$A1,$8F,$9F,$5F,$A1,$8F,$EE,$91,$FD,$0B
	.dc.b	$0F,$E7,$CF,$73,$FB,$B6,$1F,$9F,$BF,$F0,$4B,$F3,$A3,$F8,$9B,$F9
	.dc.b	$F7,$FA,$24,$FF,$13,$3F,$C6,$91,$FC,$4A,$7F,$C2,$35,$F8,$6A,$FE
	.dc.b	$1C,$FD,$03,$F3,$C7,$CF,$7E,$8B,$FD,$38,$FE,$5D,$1F,$E1,$A8,$FC
	.dc.b	$31,$FE,$1C,$C3,$F8,$AD,$FF,$9F,$60,$7F,$45,$3F,$DC,$77,$F9,$8A
	.dc.b	$E5,$1F,$B9,$3F,$CD,$07,$D3,$34,$75,$9E,$8F,$F8,$66,$3C,$55,$FD
	.dc.b	$85,$FF,$00,$9F,$DE,$AF,$F8,$EC,$FF,$AA,$0E,$67,$FB,$D2,$BD,$CF
	.dc.b	$98,$FF,$30,$5F,$F9,$BD,$FF,$57,$BC,$1F,$F0,$B4,$7C,$97,$84,$7C
	.dc.b	$27,$FD,$DE,$6C,$FF,$78,$C7,$C3,$8F,$85,$1C,$0F,$41,$FA,$93,$E1
	.dc.b	$3F,$6D,$F8,$F7,$F2,$8F,$E3,$F3,$CD,$FF,$1D,$C7,$F8,$06,$7C,$0F
	.dc.b	$FA,$D2,$BF,$A8,$1F,$08,$FF,$45,$CE,$2B,$EB,$47,$EA,$E7,$FC,$64
	.dc.b	$8F,$2B,$FE,$73,$3E,$B9,$FD,$48,$EE,$7C,$93,$F3,$07,$B6,$FE,$ED
	.dc.b	$EA,$3F,$CE,$51,$F6,$1F,$9C,$7F,$51,$BF,$D9,$5F,$DE,$7D,$B1,$FC
	.dc.b	$F1,$FD,$B5,$FE,$36,$7F,$8E,$CF,$F9,$24,$FF,$4C,$1D,$DF,$BA,$FD
	.dc.b	$D7,$DC,$F9,$DF,$75,$EE,$7E,$8C,$FE,$22,$3F,$DD,$37,$FB,$08,$F6
	.dc.b	$C7,$DA,$7F,$3A,$CB,$FD,$E5,$7F,$54,$9F,$C5,$4F,$B6,$BF,$73,$5C
	.dc.b	$8F,$B3,$1E,$D8,$CD,$7D,$43,$EC,$1F,$81,$7F,$72,$FC,$E3,$D7,$5F
	.dc.b	$E9,$D9,$F9,$D1,$F9,$58,$F8,$93,$F6,$2B,$F6,$B3,$FE,$0A,$AF,$A3
	.dc.b	$5F,$78,$FF,$72,$9F,$BB,$3E,$AC,$F8,$5F,$D9,$4F,$F2,$79,$C9,$F7
	.dc.b	$63,$F8,$A1,$FA,$68,$FD,$8D,$FE,$42,$FF,$68,$FD,$98,$FC,$CB,$F4
	.dc.b	$AF,$FD,$C4,$FA,$6B,$FC,$04,$7C,$A3,$C4,$7F,$0A,$3E,$99,$F2,$8F
	.dc.b	$51,$F9,$3B,$FE,$86,$FF,$A3,$5F,$D6,$36,$79,$B6,$7F,$59,$3F,$62
	.dc.b	$79,$CF,$EA,$EF,$E8,$BF,$C1,$6F,$D8,$FE,$44,$FF,$02,$BF,$B0,$AD
	.dc.b	$3C,$A3,$FD,$08,$FF,$1A,$7F,$8A,$9F,$8E,$3F,$96,$5E,$67,$EA,$9F
	.dc.b	$36,$3F,$93,$1E,$A3,$FC,$05,$7F,$A8,$AF,$A9,$1F,$56,$3F,$4F,$1F
	.dc.b	$30,$FA,$67,$C0,$8F,$D1,$3F,$B8,$AF,$C5,$C7,$F2,$29,$E1,$3F,$C6
	.dc.b	$63,$F4,$87,$13,$F8,$D4,$E6,$BE,$B1,$FB,$F3,$E8,$C7,$D3,$AF,$D6
	.dc.b	$1C,$5F,$D4,$4F,$D0,$7F,$C7,$73,$C6,$7F,$A2,$57,$EF,$A7,$F7,$CF
	.dc.b	$D0,$6B,$FA,$33,$F9,$19,$F8,$BB,$FD,$3B,$F3,$E3,$F3,$E7,$CB,$BF
	.dc.b	$9C,$FE,$8E,$BF,$CD,$B3,$FC,$46,$3F,$42,$7E,$76,$3F,$8E,$BF,$D4
	.dc.b	$4F,$EC,$17,$99,$FC,$E4,$7F,$49,$AF,$33,$FA,$37,$F4,$13,$99,$FC
	.dc.b	$24,$A7,$F4,$13,$FD,$2E,$3F,$A3,$1F,$2C,$FC,$FB,$F3,$AB,$F8,$7D
	.dc.b	$FF,$0C,$9F,$CC,$1F,$26,$BE,$64,$7F,$10,$AF,$99,$7F,$C3,$6B,$F2
	.dc.b	$A7,$F0,$87,$E5,$47,$E6,$0C,$3F,$26,$3F,$A0,$5F,$F0,$AA,$FE,$7B
	.dc.b	$1F,$25,$FE,$F7,$3F,$26,$3F,$BD,$4F,$E5,$67,$FB,$BD,$7F,$08,$8F
	.dc.b	$CA,$C7,$F7,$79,$F8,$87,$F3,$3B,$F8,$EF,$E6,$6B,$F9,$B5,$FE,$4E
	.dc.b	$3F,$82,$1F,$16,$7F,$24,$FF,$A3,$67,$E4,$4F,$F6,$78,$F8,$B7,$FD
	.dc.b	$AA,$3F,$1F,$1F,$90,$8F,$F4,$5C,$7C,$58,$FE,$D2,$7C,$5A,$FE,$69
	.dc.b	$5F,$16,$BF,$B6,$D7,$C5,$BF,$E5,$C7,$F8,$72,$BF,$1A,$FF,$62,$9F
	.dc.b	$C6,$CF,$F8,$76,$FE,$2C,$FF,$86,$AB,$FB,$05,$FC,$58,$FF,$0D,$CF
	.dc.b	$F0,$D9,$FD,$13,$FC,$B4,$F8,$D5,$F1,$5F,$8D,$7F,$B8,$8F,$F5,$74
	.dc.b	$7C,$68,$FF,$58,$4F,$C6,$AF,$EB,$97,$F1,$AF,$FA,$C5,$FF,$59,$BF
	.dc.b	$8D,$5F,$89,$8F,$8C,$FF,$56,$9F,$8B,$1F,$1A,$7F,$C1,$91,$FC,$86
	.dc.b	$3F,$91,$47,$C6,$8F,$F0,$44,$7E,$2A,$7E,$34,$7F,$53,$8F,$C5,$1F
	.dc.b	$E4,$D7,$F1,$51,$FE,$07,$9F,$C5,$4F,$EC,$0F,$E3,$D1,$F1,$9F,$F1
	.dc.b	$E4,$FF,$1E,$7E,$2C,$7F,$90,$EB,$F6,$07,$E2,$A7,$E2,$CF,$C6,$AF
	.dc.b	$8B,$1F,$89,$8F,$C5,$47,$EB,$9F,$F4,$F3,$FC,$69,$FD,$74,$7E,$2A
	.dc.b	$BF,$5D,$3F,$8A,$8F,$8B,$1F,$D2,$23,$E2,$BF,$DE,$AB,$F1,$4F,$F1
	.dc.b	$69,$FC,$2B,$FE,$0C,$8F,$F7,$ED,$23,$0F,$F9,$5E,$BF,$C3,$98,$66
	.dc.b	$15,$FD,$A1,$FF,$7E,$1F,$C3,$6F,$51,$9A,$FF,$1B,$4F,$08,$FF,$1D
	.dc.b	$47,$F9,$8E,$55,$C2,$7F,$DC,$77,$E2,$BF,$F5,$19,$FE,$7E,$70,$FF
	.dc.b	$8E,$F0,$E3,$18,$C3,$FC,$B5,$1F,$E4,$D8,$C4,$FF,$2B,$AF,$23,$FE
	.dc.b	$AC,$BF,$03,$FD,$5B,$0E,$0F,$9C,$F8,$57,$FE,$42,$AF,$63,$FE,$59
	.dc.b	$3C,$31,$B3,$C3,$3F,$0E,$FC,$93,$FE,$FA,$AE,$33,$FE,$61,$BF,$0D
	.dc.b	$FF,$9D,$E3,$C3,$3F,$24,$FF,$7E,$96,$3F,$9D,$4F,$07,$FD,$EB,$1F
	.dc.b	$EA,$2A,$E2,$FC,$E3,$FC,$93,$1B,$C3,$EC,$BE,$1C,$3E,$5B,$F7,$23
	.dc.b	$FC,$E9,$86,$EB,$FD,$80,$78,$E3,$C3,$5F,$3E,$11,$EF,$AF,$84,$7F
	.dc.b	$9A,$E3,$EC,$38,$8E,$D5,$F2,$27,$F1,$53,$E0,$8F,$4B,$FE,$C8,$38
	.dc.b	$C7,$E2,$E3,$E8,$61,$EB,$7E,$91,$4F,$6D,$FE,$0D,$FF,$65,$D7,$90
	.dc.b	$FF,$42,$9E,$33,$E6,$D7,$E5,$1F,$F7,$D3,$F8,$67,$D8,$D3,$C6,$6D
	.dc.b	$F9,$47,$53,$FD,$EF,$1F,$3D,$F9,$31,$D6,$FE,$4C,$78,$5E,$B1,$93
	.dc.b	$AC,$7E,$C4,$FC,$B5,$7D,$6C,$3E,$33,$F8,$CB,$FF,$66,$36,$3E,$65
	.dc.b	$7E,$B9,$F8,$B1,$B8,$EB,$5F,$96,$AD,$BF,$E5,$B3,$F0,$33,$F7,$5E
	.dc.b	$A7,$D9,$9F,$8A,$7E,$35,$F4,$9F,$C1,$DF,$C7,$C7,$19,$EB,$3F,$E5
	.dc.b	$D3,$95,$F4,$AF,$C6,$5E,$A7,$FD,$64,$FE,$09,$FF,$6A,$47,$EE,$4F
	.dc.b	$CE,$4F,$D0,$7F,$13,$5F,$12,$7F,$1C,$FE,$DE,$3F,$87,$BF,$88,$BF
	.dc.b	$C2,$3E,$89,$FF,$70,$3F,$B5,$9D,$BD,$5F,$F4,$B9,$F8,$38,$F9,$27
	.dc.b	$F2,$0B,$F4,$5F,$FB,$06,$BF,$01,$3F,$92,$3F,$1C,$73,$3F,$DB,$47
	.dc.b	$E1,$AF,$F8,$AC,$FC,$3B,$FC,$35,$F4,$AF,$D6,$BF,$0D,$FF,$6D,$47
	.dc.b	$8A,$7D,$38,$7F,$03,$9F,$E6,$58,$6E,$7F,$9A,$46,$EF,$E1,$3F,$B3
	.dc.b	$7F,$0A,$D9,$F4,$5F,$F1,$67,$EA,$47,$F0,$87,$A4,$7E,$A2,$3F,$00
	.dc.b	$7E,$71,$FE,$2D,$3A,$8E,$71,$F8,$79,$E6,$FC,$18,$FC,$19,$EB,$7D
	.dc.b	$F7,$FC,$46,$FE,$B4,$FF,$0B,$31,$5E,$43,$F0,$15,$FC,$15,$F9,$51
	.dc.b	$FD,$65,$F7,$C6,$6F,$E2,$47,$DF,$9D,$3F,$55,$FA,$AF,$C9,$7D,$C0
	.dc.b	$FC,$05,$FD,$E9,$D4,$7B,$8F,$DC,$1C,$CE,$B8,$7D,$E7,$EC,$1E,$E7
	.dc.b	$F0,$31,$DF,$0E,$53,$DD,$FB,$B3,$FC,$06,$FF,$6B,$3F,$5C,$FB,$8F
	.dc.b	$D1,$C3,$F8,$03,$F2,$2B,$FB,$A4,$6A,$3E,$8C,$F2,$8F,$E9,$8F,$D1
	.dc.b	$39,$1F,$72,$FE,$D4,$66,$7E,$CD,$71,$8F,$D9,$D7,$D9,$8E,$2F,$EE
	.dc.b	$1F,$14,$7E,$E6,$3F,$83,$47,$C5,$7F,$A0,$4F,$F3,$97,$F3,$8F,$D8
	.dc.b	$9F,$D9,$D7,$D8,$8F,$8B,$3F,$CB,$9F,$CA,$BF,$55,$FA,$6F,$CE,$3C
	.dc.b	$67,$ED,$DF,$DA,$CE,$5E,$35,$EB,$9F,$E6,$0F,$E2,$9F,$A4,$7D,$73
	.dc.b	$F9,$C5,$7D,$7A,$FB,$73,$EB,$8F,$C9,$C7,$D2,$9F,$AB,$1F,$C1,$8F
	.dc.b	$D7,$C7,$F8,$08,$FA,$98,$7E,$E5,$F5,$5F,$F7,$1A,$F1,$1F,$DF,$8F
	.dc.b	$5D,$FF,$7A,$BF,$AB,$1B,$9F,$AA,$7E,$DA,$78,$9F,$DF,$67,$F8,$BD
	.dc.b	$7F,$0C,$9F,$D7,$3F,$65,$FD,$61,$F0,$EB,$F9,$2B,$C8,$FA,$8F,$EA
	.dc.b	$27,$EA,$4F,$D8,$9F,$B9,$3E,$AC,$3F,$C6,$2F,$88,$FE,$92,$7E,$9A
	.dc.b	$BF,$53,$5F,$4F,$0F,$E3,$53,$FC,$16,$3F,$C6,$F8,$71,$3E,$0C,$7A
	.dc.b	$6F,$F9,$4C,$FD,$2C,$3F,$8B,$E1,$F4,$AB,$FC,$73,$1F,$0A,$3D,$31
	.dc.b	$FE,$36,$7F,$93,$1F,$C0,$DF,$EC,$98,$7D,$37,$E1,$1F,$44,$FD,$34
	.dc.b	$7C,$A7,$F3,$A7,$E5,$30,$FE,$4D,$3F,$6D,$E2,$7F,$8B,$63,$F4,$D3
	.dc.b	$F4,$A3,$F4,$A7,$1B,$FE,$FF,$3F,$E1,$7B,$F5,$3F,$48,$F9,$D5,$FD
	.dc.b	$42,$3F,$17,$3F,$D6,$0F,$A4,$FE,$92,$7C,$27,$F1,$A3,$F8,$CE,$1F
	.dc.b	$52,$BE,$8C,$7F,$58,$9F,$D1,$9F,$C4,$DF,$CE,$C7,$F5,$3B,$FE,$19
	.dc.b	$1F,$BB,$3F,$4A,$FE,$86,$7D,$F3,$FD,$C6,$7E,$8B,$FA,$18,$FE,$28
	.dc.b	$FC,$F8,$FC,$94,$7F,$14,$3F,$A4,$3F,$3B,$0F,$D1,$4F,$E7,$AF,$FC
	.dc.b	$19,$1F,$36,$BF,$3D,$5F,$32,$3F,$A4,$57,$E8,$70,$FC,$E4,$7F,$65
	.dc.b	$AF,$97,$5F,$E3,$63,$E5,$C7,$F9,$26,$BF,$37,$1F,$E4,$B9,$F9,$47
	.dc.b	$F0,$C7,$FA,$04,$FF,$41,$8F,$E1,$71,$F2,$63,$F3,$38,$7D,$03,$F3
	.dc.b	$15,$F2,$4F,$E1,$13,$FE,$45,$8F,$91,$5F,$92,$3F,$2C,$7E,$FA,$7E
	.dc.b	$EC,$FE,$6A,$FF,$83,$E1,$FE,$78,$7F,$5D,$1F,$E3,$53,$F7,$EF,$EB
	.dc.b	$9F,$F1,$49,$F9,$67,$F9,$D6,$1F,$26,$FF,$B7,$4F,$C9,$AF,$F3,$E6
	.dc.b	$1F,$94,$BF,$93,$1F,$E8,$6A,$F9,$33,$FE,$86,$AF,$93,$3F,$E2,$CC
	.dc.b	$3E,$4B,$FE,$2E,$AF,$C9,$4F,$F8,$A0,$FF,$60,$C7,$FA,$41,$FF,$4C
	.dc.b	$47,$F0,$07,$F9,$E9,$FD,$BC,$FF,$4A,$46,$DF,$D9,$9F,$CD,$0F,$9E
	.dc.b	$7F,$A4,$9E,$95,$FE,$08,$8F,$83,$18,$BD,$47,$F9,$DE,$79,$DF,$E5
	.dc.b	$E3,$FC,$74,$6A,$7F,$D2,$B1,$F4,$A3,$FC,$7D,$1E,$B8,$5F,$9A,$7F
	.dc.b	$D1,$D1,$F8,$E7,$FD,$CD,$7F,$5D,$FE,$F8,$D8,$FF,$1E,$47,$FA,$D2
	.dc.b	$BE,$31,$FE,$0F,$3C,$47,$F9,$20,$FB,$E7,$FA,$A0,$FF,$46,$BF,$1C
	.dc.b	$FE,$1D,$7F,$C4,$67,$AC,$FE,$DE,$7A,$4F,$9E,$39,$3F,$32,$FE,$14
	.dc.b	$FF,$52,$3C,$27,$57,$E1,$3F,$D5,$A7,$EE,$47,$C2,$77,$1F,$EF,$D8
	.dc.b	$FA,$EF,$F1,$13,$F0,$27,$FA,$9C,$FA,$C7,$FA,$20,$C4,$72,$7C,$11
	.dc.b	$F7,$9F,$BF,$3F,$AE,$AF,$93,$1E,$D3,$FC,$21,$7F,$82,$AF,$9B,$3E
	.dc.b	$C9,$FF,$31,$D7,$F9,$9E,$3D,$71,$FE,$26,$31,$3F,$16,$FF,$CE,$F3
	.dc.b	$F6,$DF,$51,$F9,$5A,$FE,$8D,$1F,$3A,$BF,$33,$3F,$3A,$FF,$01,$7F
	.dc.b	$4A,$FF,$BD,$D7,$D6,$9F,$DD,$47,$E5,$EB,$D4,$7F,$BF,$1F,$DB,$47
	.dc.b	$EC,$6B,$89,$F9,$0A,$F8,$C7,$FB,$E2,$FF,$17,$1F,$CB,$63,$FD,$F7
	.dc.b	$7A,$8F,$CA,$47,$D3,$3D,$55,$FA,$DA,$F9,$71,$FB,$23,$D4,$7D,$79
	.dc.b	$F9,$F7,$EA,$8F,$F8,$02,$78,$8F,$EE,$87,$EB,$6F,$F8,$03,$EB,$BF
	.dc.b	$A9,$5F,$3A,$3F,$54,$FE,$91,$FD,$4B,$FB,$9B,$F5,$1F,$8D,$BF,$AB
	.dc.b	$1F,$3E,$7F,$4E,$71,$BF,$D2,$3F,$2C,$FA,$67,$DB,$3F,$DB,$D1,$F4
	.dc.b	$DF,$D7,$CF,$19,$FD,$73,$F4,$CF,$B6,$71,$8F,$F6,$D4,$7F,$B4,$27
	.dc.b	$E9,$9C,$63,$F2,$F5,$F4,$A3,$F8,$D9,$E9,$BF,$B4,$FC,$E9,$F8,$A7
	.dc.b	$13,$F1,$EF,$F2,$3A,$C9,$F3,$63,$F5,$77,$F4,$4E,$0F,$F8,$7A,$BF
	.dc.b	$8F,$3F,$D4,$A7,$EA,$4F,$F1,$EA,$FC,$8D,$FE,$26,$3F,$3E,$FE,$9A
	.dc.b	$3F,$98,$BE,$7A,$FA,$71,$F3,$DF,$F1,$21,$F9,$0A,$FE,$73,$7F,$C7
	.dc.b	$5F,$A6,$66,$3F,$9B,$BF,$C5,$9F,$E7,$8F,$E2,$2F,$F8,$B5,$7F,$89
	.dc.b	$E3,$E9,$9D,$63,$F8,$A4,$7C,$E9,$FC,$D3,$FC,$F4,$FC,$F1,$E7,$9F
	.dc.b	$A8,$FC,$D8,$FE,$98,$7C,$9B,$FC,$EC,$FC,$CA,$FE,$A1,$5F,$9C,$AF
	.dc.b	$CB,$C7,$CC,$7F,$D6,$55,$FA,$0A,$C9,$F3,$2B,$E4,$4F,$E6,$8F,$D9
	.dc.b	$CF,$F0,$C3,$F3,$31,$F8,$98,$FE,$BE,$FC,$BB,$FE,$89,$1F,$1E,$7F
	.dc.b	$85,$57,$CA,$BF,$F2,$95,$FE,$5E,$BF,$C9,$87,$C5,$AF,$E1,$95,$F2
	.dc.b	$2B,$F8,$3C,$FF,$92,$6B,$E4,$47,$F8,$85,$F9,$07,$E5,$0F,$24,$7F
	.dc.b	$AC,$AF,$FB,$73,$F9,$43,$E4,$47,$E4,$DF,$CA,$4F,$C8,$8F,$E6,$93
	.dc.b	$FD,$C2,$7F,$1F,$7F,$1D,$FC,$95,$7F,$89,$2B,$E3,$C7,$F0,$18,$FC
	.dc.b	$8C,$7F,$30,$BF,$E0,$67,$CA,$7F,$13,$3F,$CA,$23,$F8,$BA,$7F,$7F
	.dc.b	$5F,$22,$3F,$CF,$11,$F9,$19,$F9,$17,$FC,$BA,$7E,$44,$7E,$3D,$FC
	.dc.b	$85,$FC,$8A,$FD,$E4,$7F,$5C,$9F,$C1,$1F,$D1,$8F,$C7,$9F,$20,$FE
	.dc.b	$56,$FF,$23,$3F,$10,$7E,$90,$FE,$B6,$7C,$8A,$FF,$0D,$47,$E3,$E3
	.dc.b	$FA,$DC,$7F,$86,$2F,$E4,$3F,$E1,$7B,$F9,$15,$FC,$9E,$BE,$44,$FF
	.dc.b	$81,$6F,$F8,$3D,$FE,$E5,$F9,$37,$FC,$9A,$FF,$6F,$7F,$B8,$8F,$EA
	.dc.b	$F1,$F1,$A3,$E3,$C7,$E3,$A3,$F1,$B1,$F8,$E8,$FD,$AD,$FF,$56,$AF
	.dc.b	$C7,$3F,$D6,$2B,$F8,$F5,$7C,$73,$F6,$B7,$FC,$86,$BE,$51,$FB,$89
	.dc.b	$FD,$A5,$7E,$3A,$FF,$C8,$15,$FB,$4B,$F8,$C7,$F4,$FA,$FC,$6B,$FE
	.dc.b	$37,$3F,$1A,$FF,$1D,$7F,$1B,$5F,$D3,$0F,$88,$FF,$88,$E7,$FC,$00
	.dc.b	$7F,$1F,$8C,$C7,$F9,$36,$3F,$6D,$3F,$C8,$0F,$D1,$BF,$DF,$E3,$F2
	.dc.b	$07,$F0,$3B,$FB,$87,$F4,$E3,$F6,$D5,$FD,$FA,$BE,$4B,$F8,$FB,$F9
	.dc.b	$35,$F9,$7A,$FE,$9A,$7C,$A7,$07,$F8,$E5,$FF,$70,$1F,$E3,$B7,$FC
	.dc.b	$DC,$7E,$E4,$F9,$B5,$FB,$B8,$F9,$B3,$FB,$D8,$FE,$F2,$7F,$95,$38
	.dc.b	$B9,$57,$F9,$9C,$07,$2F,$F7,$A1,$83,$1D,$47,$2B,$F9,$17,$88,$F3
	.dc.b	$47,$F3,$C7,$07,$F9,$4D,$FF,$3D,$47,$3B,$EF,$3E,$29,$FF,$34,$CF
	.dc.b	$08,$FF,$16,$1E,$39,$F2,$CF,$F3,$2A,$D4,$FF,$B4,$2F,$DC,$F9,$44
	.dc.b	$FF,$05,$76,$7F,$4F,$BF,$13,$F0,$27,$C3,$1A,$9F,$82,$FF,$A4,$6F
	.dc.b	$C3,$3F,$E7,$E3,$8D,$7C,$07,$13,$C5,$FF,$31,$45,$8F,$F4,$2D,$F8
	.dc.b	$A3,$C9,$1E,$87,$E4,$E1,$FE,$79,$BF,$7B,$F5,$AF,$BC,$FC,$B8,$FF
	.dc.b	$53,$CE,$A3,$DC,$DA,$DC,$7E,$2F,$0F,$9B,$3D,$1F,$64,$7D,$BA,$F8
	.dc.b	$F7,$C2,$3E,$A4,$62,$FE,$23,$F3,$0F,$C4,$DF,$38,$E1,$3F,$8C,$7E
	.dc.b	$5C,$7D,$87,$FC,$AD,$1F,$35,$F9,$58,$57,$E7,$CE,$DF,$5C,$FE,$3D
	.dc.b	$FC,$2C,$FC,$98,$F8,$C7,$E2,$8F,$C5,$4D,$79,$99,$AF,$3B,$70,$E4
	.dc.b	$FC,$43,$F2,$6F,$C4,$7F,$1E,$7E,$2A,$7B,$47,$48,$FA,$11,$89,$D4
	.dc.b	$72,$BF,$B8,$FF,$8E,$8F,$09,$FB,$2C,$3F,$C6,$15,$FA,$27,$F1,$2D
	.dc.b	$9F,$83,$3B,$7F,$AF,$5F,$E2,$6B,$E6,$CF,$EE,$2F,$E1,$BF,$BE,$7F
	.dc.b	$D2,$97,$FA,$E7,$F0,$F1,$B3,$A9,$FE,$CB,$3A,$C6,$DF,$88,$73,$AF
	.dc.b	$88,$FF,$99,$0F,$79,$9A,$F7,$BF,$8F,$AF,$7B,$FC,$7A,$7F,$41,$1F
	.dc.b	$B4,$BD,$CF,$FB,$14,$D9,$FE,$BE,$3A,$9F,$C8,$5F,$8A,$FF,$16,$7F
	.dc.b	$5B,$7F,$C2,$9F,$F4,$23,$5F,$4C,$6D,$FD,$E3,$F1,$8F,$E2,$87,$EB
	.dc.b	$70,$CB,$F7,$4C,$9F,$18,$FE,$26,$7E,$EA,$7F,$07,$3E,$A8,$FF,$2D
	.dc.b	$1E,$A3,$F8,$A4,$F8,$2F,$A4,$7F,$1C,$7A,$3F,$C8,$4F,$89,$1F,$76
	.dc.b	$7D,$31,$FE,$CD,$C3,$F5,$37,$FC,$4A,$7F,$9A,$D7,$F0,$D9,$F1,$5F
	.dc.b	$FB,$66,$B5,$5F,$B4,$BF,$E1,$EF,$EE,$9F,$B6,$7F,$B6,$70,$F8,$53
	.dc.b	$FC,$2E,$F9,$C7,$F4,$63,$D9,$3F,$C0,$4F,$84,$7E,$46,$FC,$EF,$F0
	.dc.b	$87,$AB,$67,$F8,$2C,$7E,$69,$EA,$FC,$28,$FF,$6F,$5F,$9C,$FC,$EB
	.dc.b	$FD,$31,$F8,$2F,$F0,$37,$E0,$DF,$F0,$7A,$F8,$31,$F8,$49,$D3,$F0
	.dc.b	$5F,$C0,$1F,$D2,$A7,$F8,$34,$FD,$98,$FE,$B1,$1F,$C2,$6F,$F2,$31
	.dc.b	$FD,$A2,$B4,$FF,$53,$3F,$83,$9A,$3F,$AF,$56,$AF,$D9,$1F,$0E,$FF
	.dc.b	$0F,$5F,$76,$3E,$59,$FD,$38,$E4,$7F,$3F,$7F,$0E,$FF,$6E,$BD,$DF
	.dc.b	$E7,$A3,$FA,$C4,$FD,$07,$EB,$1F,$40,$A7,$43,$FB,$14,$7B,$E7,$75
	.dc.b	$FB,$F7,$47,$F3,$9A,$FD,$BD,$F3,$BF,$E2,$26,$9F,$2B,$F8,$57,$DF
	.dc.b	$3F,$D1,$5E,$33,$F3,$4F,$B3,$3F,$2D,$FE,$6E,$FE,$D6,$F5,$1F,$20
	.dc.b	$F7,$46,$DF,$A2,$FB,$E7,$FC,$49,$7F,$CD,$A3,$F0,$67,$F8,$04,$F2
	.dc.b	$C7,$DC,$71,$3F,$83,$8E,$51,$F6,$CF,$F0,$C9,$FD,$8B,$0F,$DC,$5F
	.dc.b	$EB,$23,$F4,$17,$FD,$12,$BF,$43,$1F,$CD,$6F,$F6,$35,$EE,$8F,$C1
	.dc.b	$C7,$91,$FF,$03,$3E,$48,$FC,$B4,$62,$3F,$6D,$87,$D1,$BF,$EF,$93
	.dc.b	$FB,$0B,$FB,$38,$7E,$E7,$0F,$DA,$CF,$EA,$67,$FC,$17,$3F,$43,$0F
	.dc.b	$EE,$F1,$F8,$33,$F6,$11,$8A,$FD,$8E,$1D,$4F,$E2,$38,$7F,$29,$BF
	.dc.b	$D5,$9C,$4F,$EE,$EF,$EB,$DF,$EF,$11,$E5,$7F,$BA,$C7,$F7,$1C,$3F
	.dc.b	$3A,$7F,$52,$3F,$BB,$57,$DA,$AC,$C7,$F5,$03,$FB,$D5,$7E,$A5,$E3
	.dc.b	$87,$E1,$9F,$E4,$17,$FD,$BD,$FF,$01,$47,$17,$FC,$11,$1F,$DA,$5F
	.dc.b	$D3,$CF,$17,$FD,$4A,$FF,$2D,$AC,$9F,$A5,$BF,$E1,$F3,$FD,$8C,$FD
	.dc.b	$AC,$7A,$5F,$ED,$87,$E8,$0F,$F2,$63,$A7,$C4,$7F,$6B,$9F,$D7,$5F
	.dc.b	$EA,$67,$FC,$A7,$3F,$A9,$AC,$C7,$69,$E3,$3F,$E3,$7A,$FD,$1B,$FC
	.dc.b	$19,$FF,$03,$BF,$A7,$7F,$C9,$47,$E2,$9F,$EC,$CF,$27,$FC,$DE,$FF
	.dc.b	$05,$6D,$FE,$9D,$FC,$61,$FD,$00,$FD,$45,$FE,$7D,$FC,$61,$FE,$3B
	.dc.b	$C3,$E7,$D7,$E8,$EF,$F1,$93,$FA,$7B,$FD,$25,$E7,$0E,$B5,$FA,$6A
	.dc.b	$FF,$4D,$4F,$CE,$7F,$AE,$5F,$E5,$63,$F1,$B1,$FE,$8D,$7F,$BF,$BF
	.dc.b	$A8,$AF,$D1,$47,$E3,$A3,$FC,$35,$3F,$A2,$7F,$34,$7E,$BC,$FF,$33
	.dc.b	$1F,$C0,$A7,$F8,$FB,$FE,$5F,$3C,$87,$CF,$8F,$D3,$BF,$A3,$AE,$4F
	.dc.b	$F6,$57,$91,$FA,$F9,$FD,$3C,$7F,$83,$67,$F4,$EF,$E8,$DE,$2F,$FA
	.dc.b	$75,$F5,$BF,$EA,$37,$EC,$3F,$EB,$23,$FC,$48,$F9,$1F,$F5,$9D,$FE
	.dc.b	$AE,$BF,$CA,$F7,$FA,$99,$FF,$2C,$DF,$F3,$D7,$F8,$41,$FD,$50,$FD
	.dc.b	$29,$99,$FC,$C3,$FC,$68,$FF,$2F,$C6,$6F,$FB,$15,$7F,$19,$9F,$F7
	.dc.b	$44,$7E,$A5,$B1,$A8,$FE,$BA,$FF,$36,$8F,$BB,$1F,$8F,$9F,$ED,$F1
	.dc.b	$FE,$C3,$3F,$A6,$9F,$A9,$7E,$9B,$FA,$C7,$31,$FA,$88,$FD,$21,$FA
	.dc.b	$77,$FA,$9C,$7E,$9C,$FE,$A4,$FE,$9A,$3F,$46,$F9,$DF,$D0,$CF,$F4
	.dc.b	$C9,$FD,$54,$FE,$96,$7F,$AF,$3F,$D3,$0F,$D0,$3F,$A4,$3E,$C9,$FE
	.dc.b	$E3,$9F,$D2,$4E,$67,$FA,$51,$FA,$39,$FC,$FC,$FF,$87,$9F,$F5,$45
	.dc.b	$FF,$47,$AF,$D1,$D7,$F8,$CE,$BF,$45,$5F,$EE,$17,$E8,$36,$3C,$37
	.dc.b	$FE,$D2,$9F,$E9,$B1,$FC,$FD,$FD,$14,$FF,$14,$8F,$D1,$3F,$D1,$70
	.dc.b	$FE,$8D,$1F,$E3,$FB,$FD,$2B,$FD,$3A,$FF,$B5,$47,$F6,$1A,$FF,$66
	.dc.b	$47,$82,$3D,$31,$FE,$C5,$8F,$D0,$1F,$95,$7E,$4C,$FF,$AE,$5F,$D0
	.dc.b	$5F,$F8,$FA,$3F,$3F,$1F,$9D,$9F,$F7,$ED,$7E,$96,$7F,$C6,$93,$FE
	.dc.b	$0D,$3F,$97,$1F,$E1,$CC,$3F,$C6,$4F,$E7,$AB,$FD,$6F,$3F,$D1,$63
	.dc.b	$F3,$B1,$FD,$0A,$FF,$C6,$17,$FD,$20,$FC,$E5,$7F,$8B,$EB,$FC,$61
	.dc.b	$5F,$E6,$DA,$FC,$D5,$7F,$42,$8F,$F3,$85,$7F,$89,$6B,$F9,$C1,$F8
	.dc.b	$B3,$FC,$5B,$59,$AF,$E8,$07,$E6,$27,$FC,$4F,$3F,$98,$AF,$F1,$4D
	.dc.b	$7F,$3E,$8F,$CC,$C7,$F7,$5B,$FF,$4B,$47,$E5,$0F,$F1,$B5,$7C,$87
	.dc.b	$F9,$CD,$FE,$60,$CC,$7F,$3B,$8F,$DF,$9F,$E1,$83,$F9,$99,$F8,$83
	.dc.b	$FD,$7E,$66,$3F,$31,$87,$F3,$A9,$FF,$4A,$4F,$E6,$23,$FC,$3F,$1F
	.dc.b	$E9,$79,$FC,$BC,$7C,$48,$57,$FA,$46,$BF,$99,$C7,$E5,$B0,$F4,$43
	.dc.b	$0F,$F4,$93,$FC,$EE,$7F,$10,$FF,$A4,$AB,$FC,$3A,$FE,$5D,$FF,$48
	.dc.b	$47,$E5,$E3,$FD,$2C,$FE,$5C,$FF,$43,$BF,$CD,$6B,$FD,$17,$87,$E5
	.dc.b	$A3,$F9,$51,$FE,$91,$7F,$9D,$4F,$FA,$2C,$FF,$24,$1F,$AE,$8F,$E5
	.dc.b	$F3,$F9,$78,$FF,$23,$9F,$CD,$23,$F9,$C3,$E8,$7F,$8D,$9D,$8F,$E6
	.dc.b	$07,$F2,$C3,$FC,$FE,$7F,$98,$EB,$4F,$F7,$46,$BF,$EC,$A6,$BF,$EC
	.dc.b	$1B,$FC,$D4,$FF,$87,$2F,$F3,$51,$F5,$EB,$FC,$39,$5F,$9A,$78,$DF
	.dc.b	$D0,$8F,$9E,$78,$1F,$E6,$8F,$F3,$AB,$FC,$E3,$FE,$17,$3F,$99,$C7
	.dc.b	$E6,$DF,$F7,$A9,$FC,$94,$FA,$C7,$FA,$F4,$FC,$E4,$FA,$2F,$E8,$47
	.dc.b	$C3,$7F,$A6,$3F,$01,$FF,$41,$46,$EF,$FC,$25,$7F,$46,$8F,$F8,$71
	.dc.b	$FF,$42,$5F,$F3,$77,$F9,$DD,$FE,$82,$7F,$9B,$1F,$EF,$68,$FE,$6D
	.dc.b	$3F,$CB,$DF,$F3,$F9,$FD,$30,$FF,$7C,$C7,$F3,$63,$F3,$EF,$F9,$F3
	.dc.b	$0F,$CF,$E1,$FC,$D6,$3F,$DE,$D3,$FC,$D4,$FF,$47,$57,$E1,$F0,$FE
	.dc.b	$65,$7F,$EE,$F9,$FD,$1D,$7E,$77,$0F,$D1,$C7,$F9,$EA,$3F,$96,$9F
	.dc.b	$E7,$18,$FF,$17,$3F,$89,$AF,$F5,$9C,$FE,$7F,$0F,$C2,$C7,$F9,$E9
	.dc.b	$F9,$11,$FA,$09,$FF,$3C,$5F,$F9,$E6,$3F,$CF,$B7,$F9,$E8,$FE,$C3
	.dc.b	$1F,$E1,$29,$F2,$9F,$EE,$97,$FA,$59,$FB,$6B,$FE,$67,$1A,$3F,$C2
	.dc.b	$91,$F9,$F7,$F9,$A1,$FC,$84,$FF,$0B,$61,$FE,$11,$C3,$FD,$CD,$5F
	.dc.b	$CD,$2B,$FC,$1F,$5F,$CD,$27,$F3,$AF,$E1,$8F,$81,$87,$E7,$23,$FD
	.dc.b	$C6,$7E,$70,$FF,$3C,$57,$F3,$5A,$FF,$3B,$57,$F9,$DE,$3F,$89,$4F
	.dc.b	$E4,$63,$F9,$A9,$F9,$A9,$FC,$CD,$7F,$34,$AF,$CD,$C7,$F8,$2E,$7F
	.dc.b	$9A,$C7,$F3,$48,$FF,$05,$C7,$F3,$4A,$F9,$F1,$FE,$0A,$8F,$E6,$93
	.dc.b	$FE,$0C,$BF,$CA,$57,$EE,$2B,$F5,$B5,$FE,$6E,$7F,$2B,$1F,$E6,$D8
	.dc.b	$F9,$87,$F6,$13,$F0,$27,$F1,$6A,$F9,$73,$A8,$1F,$E8,$63,$FC,$E2
	.dc.b	$7F,$11,$8F,$F6,$B9,$FE,$70,$3F,$9F,$1F,$A0,$3F,$CE,$E7,$F2,$EA
	.dc.b	$FF,$06,$C7,$F3,$37,$F3,$47,$F9,$B9,$FF,$69,$BF,$E0,$CB,$FC,$CE
	.dc.b	$1F,$22,$FC,$95,$FC,$EE,$7F,$90,$DF,$FB,$4B,$0F,$E6,$E7,$E6,$23
	.dc.b	$F6,$13,$FD,$FD,$FF,$44,$47,$CC,$35,$3F,$A3,$3F,$C1,$C7,$F9,$B2
	.dc.b	$3F,$35,$1F,$ED,$1A,$FF,$06,$57,$8A,$BF,$83,$C7,$E6,$8F,$E6,$E7
	.dc.b	$FB,$84,$FF,$11,$C7,$E4,$27,$FC,$D0,$7E,$49,$FE,$6D,$5F,$ED,$03
	.dc.b	$F9,$CC,$FF,$82,$E7,$F2,$F5,$F9,$FA,$FC,$C1,$FE,$8A,$8C,$D7,$F8
	.dc.b	$31,$FC,$C4,$7F,$3C,$C3,$F8,$D3,$FE,$CE,$8F,$E6,$B7,$FE,$66,$3F
	.dc.b	$96,$4F,$D4,$95,$7F,$4C,$69,$FA,$F7,$FB,$CB,$FC,$B4,$CD,$7E,$86
	.dc.b	$3F,$C2,$0F,$F2,$D8,$FE,$6A,$7F,$4E,$8F,$4B,$F8,$C7,$F9,$AB,$FC
	.dc.b	$DE,$FF,$9F,$47,$FA,$42,$7F,$C2,$73,$FA,$19,$FD,$05,$FF,$32,$8F
	.dc.b	$E6,$71,$FC,$3A,$3F,$80,$4F,$E7,$CF,$E3,$33,$F9,$E3,$F4,$07,$F3
	.dc.b	$43,$F3,$71,$FC,$CE,$37,$3F,$9B,$BF,$F3,$43,$F9,$BB,$FF,$08,$CF
	.dc.b	$CA,$AF,$F3,$0D,$FF,$84,$2D,$87,$F3,$3A,$FE,$53,$1F,$DE,$67,$F4
	.dc.b	$15,$FE,$10,$BF,$F1,$3C,$7F,$B1,$0F,$F6,$3C,$FF,$82,$9F,$F6,$31
	.dc.b	$FE,$BD,$38,$61,$FE,$66,$3F,$31,$1F,$E1,$07,$F3,$11,$F9,$59,$FF
	.dc.b	$61,$BF,$E0,$0A,$FD,$8C,$FE,$62,$3F,$CC,$B8,$7E,$6E,$7F,$99,$1E
	.dc.b	$79,$FC,$EB,$FE,$64,$9F,$E0,$78,$7E,$69,$FE,$57,$87,$F3,$17,$F9
	.dc.b	$6E,$1F,$9A,$AF,$E6,$35,$F9,$A9,$FF,$60,$57,$F9,$91,$FF,$06,$5F
	.dc.b	$F8,$2D,$FC,$C5,$7F,$B4,$6B,$F9,$89,$F9,$8A,$FF,$32,$CF,$F3,$18
	.dc.b	$FA,$53,$FE,$06,$8F,$F0,$55,$7F,$82,$2B,$FC,$15,$5F,$95,$8F,$E1
	.dc.b	$4F,$E5,$67,$F9,$8B,$FE,$03,$9F,$F3,$23,$FE,$05,$BF,$F5,$DD,$FF
	.dc.b	$31,$AF,$C9,$4F,$F2,$F8,$57,$F3,$29,$FF,$15,$C7,$F3,$27,$FD,$27
	.dc.b	$3F,$BF,$8F,$F5,$DC,$7F,$09,$3F,$28,$7F,$2E,$BF,$E6,$33,$F9,$88
	.dc.b	$FC,$AE,$1F,$15,$FE,$EC,$7F,$A9,$4F,$E7,$53,$FC,$C2,$FF,$C3,$35
	.dc.b	$FE,$09,$3F,$37,$7F,$E0,$93,$F2,$F5,$FC,$66,$7F,$98,$3F,$D0,$5F
	.dc.b	$EE,$47,$E0,$4F,$E3,$2F,$F9,$6B,$0F,$CB,$BF,$CA,$CF,$F0,$51,$FE
	.dc.b	$B5,$3F,$2E,$FF,$AF,$30,$FF,$46,$47,$FA,$DC,$FF,$33,$3F,$E0,$7B
	.dc.b	$FE,$63,$3F,$CB,$A7,$E4,$DF,$F9,$90,$FC,$13,$FE,$58,$C3,$F9,$7E
	.dc.b	$1F,$E6,$53,$DF,$19,$9F,$CC,$1E,$87,$FC,$BB,$7F,$97,$7F,$90,$BF
	.dc.b	$EB,$A3,$FC,$0E,$7D,$58,$FC,$BC,$7F,$98,$AB,$F9,$81,$F7,$A3,$FC
	.dc.b	$C5,$1F,$98,$9F,$F0,$3B,$F9,$83,$F4,$A5,$9F,$F0,$44,$7F,$30,$BF
	.dc.b	$F5,$E3,$F9,$83,$FC,$11,$3F,$98,$7F,$C1,$57,$F9,$88,$FE,$50,$7F
	.dc.b	$24,$3F,$A7,$C7,$E6,$4F,$E5,$D5,$FE,$09,$7F,$31,$1F,$E0,$93,$F9
	.dc.b	$84,$7F,$97,$AF,$E5,$C7,$F0,$37,$FC,$54,$7F,$9D,$CF,$E2,$CF,$FB
	.dc.b	$92,$3F,$96,$4F,$C5,$3F,$96,$47,$D3,$3F,$32,$7F,$81,$6B,$E9,$BF
	.dc.b	$EB,$D3,$F3,$C7,$F8,$1E,$B2,$FF,$83,$CF,$CD,$1F,$E8,$F8,$FE,$5E
	.dc.b	$69,$FC,$ED,$FF,$30,$7F,$82,$CF,$E6,$AB,$F9,$6B,$FE,$BF,$3E,$B3
	.dc.b	$89,$FF,$04,$BF,$E3,$58,$E9,$3F,$AF,$9F,$F0,$49,$EA,$BF,$F3,$0D
	.dc.b	$7F,$82,$67,$F9,$85,$7F,$AF,$CF,$E5,$93,$FC,$C3,$0F,$CD,$4F,$E5
	.dc.b	$6F,$FA,$DB,$FE,$0F,$9F,$F6,$6D,$7F,$30,$3F,$CF,$07,$EC,$23,$FC
	.dc.b	$5A,$7F,$B0,$1F,$E6,$47,$F8,$24,$FF,$30,$3F,$8C,$3F,$C0,$77,$FE
	.dc.b	$09,$C3,$F9,$84,$7F,$A6,$E3,$EB,$3F,$99,$38,$1F,$5E,$BF,$D7,$71
	.dc.b	$FE,$08,$8F,$CD,$1F,$9F,$AF,$8A,$FF,$90,$2B,$FD,$48,$7F,$82,$9F
	.dc.b	$F6,$DB,$FE,$BD,$3F,$0E,$7F,$1F,$7F,$4D,$1F,$EB,$C7,$F9,$A5,$7F
	.dc.b	$33,$8F,$F3,$1C,$7F,$99,$23,$F9,$9B,$FE,$BB,$3F,$99,$9F,$98,$9F
	.dc.b	$E6,$6F,$E6,$AF,$F9,$9D,$FF,$81,$0F,$E6,$51,$FE,$FD,$7F,$9A,$4F
	.dc.b	$F3,$38,$FC,$F4,$FF,$AD,$EF,$FC,$0B,$1F,$99,$3F,$46,$7E,$4C,$FE
	.dc.b	$77,$87,$E6,$2B,$E1,$E1,$F5,$AF,$E5,$D7,$F8,$2A,$FF,$C1,$0F,$CD
	.dc.b	$32,$7E,$6E,$FF,$D7,$6F,$D1,$9F,$CD,$3F,$EE,$39,$FF,$5D,$9E,$27
	.dc.b	$F0,$75,$FE,$BC,$C3,$F3,$B8,$7F,$98,$E7,$F9,$C5,$7F,$82,$5F,$F5
	.dc.b	$D9,$FE,$08,$C3,$FC,$00,$FF,$32,$2C,$7E,$66,$BF,$CB,$F1,$FC,$C9
	.dc.b	$FF,$30,$D7,$BB,$0F,$CC,$CF,$E9,$70,$FC,$CD,$FF,$33,$8F,$F0,$3D
	.dc.b	$FF,$B1,$23,$FC,$0E,$7E,$5C,$FF,$03,$47,$F9,$D8,$FF,$46,$47,$E5
	.dc.b	$E3,$11,$FE,$74,$3F,$C8,$11,$F9,$87,$FB,$03,$FE,$07,$9F,$CB,$C7
	.dc.b	$F3,$3A,$FF,$5B,$3F,$97,$AE,$D7,$F9,$67,$FD,$88,$FE,$55,$FF,$2F
	.dc.b	$4F,$E1,$67,$F8,$4C,$7F,$81,$DF,$CA,$DF,$F3,$08,$FC,$AB,$FE,$07
	.dc.b	$3F,$84,$3E,$B9,$FE,$2F,$3F,$E5,$9B,$FF,$32,$1F,$C4,$1F,$CB,$1F
	.dc.b	$CC,$4F,$DE,$9F,$89,$7F,$99,$3F,$EB,$AA,$FC,$BD,$7F,$96,$DF,$F0
	.dc.b	$65,$7E,$5D,$F8,$93,$F9,$53,$F1,$A7,$F2,$FC,$3F,$2D,$5F,$CB,$AB
	.dc.b	$37,$FC,$32,$3F,$2B,$5F,$E0,$88,$FF,$2E,$57,$E5,$23,$15,$F9,$47
	.dc.b	$CE,$7E,$EA,$BF,$1E,$FF,$97,$67,$FC,$B8,$FF,$73,$C3,$FC,$0D,$3F
	.dc.b	$93,$3F,$15,$1F,$E0,$48,$FF,$2D,$18,$AF,$F0,$26,$1F,$EB,$19,$FF
	.dc.b	$02,$1F,$CB,$27,$FC,$B4,$7E,$4F,$0F,$F4,$6C,$7F,$2D,$7F,$27,$7F
	.dc.b	$B4,$3F,$95,$9F,$EB,$19,$C3,$F9,$3C,$3F,$1F,$5F,$AD,$3F,$C1,$57
	.dc.b	$F9,$33,$FC,$04,$FF,$96,$2F,$F8,$2B,$F9,$23,$F9,$53,$F9,$27,$FA
	.dc.b	$24,$7F,$AC,$23,$FC,$69,$1F,$22,$FF,$24,$FE,$3C,$FD,$F3,$FE,$58
	.dc.b	$9F,$F0,$14,$FE,$62,$FF,$BD,$D7,$F7,$E3,$FC,$04,$7F,$80,$2B,$FC
	.dc.b	$04,$FD,$43,$FD,$67,$87,$FA,$B9,$FE,$51,$7F,$E0,$23,$FC,$B5,$5F
	.dc.b	$97,$BF,$93,$3C,$6B,$FD,$D1,$1F,$E3,$48,$FF,$01,$1F,$93,$3F,$C6
	.dc.b	$37,$FC,$A5,$FC,$1D,$7F,$3D,$3F,$D8,$CF,$F9,$08,$FF,$2C,$1F,$EF
	.dc.b	$23,$FB,$51,$FE,$C6,$3F,$D5,$F5,$F9,$47,$FD,$E2,$7E,$E6,$FF,$D5
	.dc.b	$EE,$6F,$FC,$B5,$5F,$E6,$57,$D3,$1F,$E0,$75,$1F,$E5,$B3,$F2,$B1
	.dc.b	$FC,$92,$FF,$D6,$35,$FC,$A2,$7F,$93,$DF,$F8,$0A,$BF,$C5,$71,$F9
	.dc.b	$67,$F9,$44,$FF,$7F,$AF,$CB,$47,$F2,$8C,$3E,$33,$FB,$AC,$3F,$DC
	.dc.b	$A7,$F8,$06,$BF,$94,$DF,$F8,$13,$0F,$F2,$D5,$7F,$8B,$63,$F2,$6F
	.dc.b	$F2,$83,$FC,$AC,$FF,$07,$7F,$5F,$5E,$C8,$FC,$C5,$7D,$F9,$FC,$03
	.dc.b	$FE,$59,$7F,$D4,$AF,$E6,$1F,$F0,$01,$B3,$EA,$1F,$ED,$E7,$F9,$51
	.dc.b	$97,$FD,$E2,$7F,$88,$4F,$F0,$09,$FC,$EE,$3F,$28,$FF,$0E,$3F,$CC
	.dc.b	$07,$E6,$CF,$CF,$3E,$96,$D7,$F9,$71,$FE,$5B,$3F,$E5,$C7,$F9,$69
	.dc.b	$F9,$B9,$FE,$5A,$FF,$97,$27,$F3,$4F,$E5,$6B,$F6,$CF,$F2,$D3,$C0
	.dc.b	$7F,$A4,$5F,$03,$63,$CC,$7E,$72,$BF,$86,$47,$E7,$0F,$F5,$B9,$F9
	.dc.b	$CA,$FE,$5C,$FF,$A3,$A7,$F9,$73,$F1,$63,$FD,$6B,$3F,$EB,$1C,$3F
	.dc.b	$17,$3F,$E0,$77,$FC,$06,$FF,$59,$17,$FC,$BF,$0F,$E0,$71,$FE,$B4
	.dc.b	$3F,$39,$3F,$E5,$CB,$FC,$E6,$1F,$CB,$E7,$FC,$57,$1F,$E0,$27,$F9
	.dc.b	$76,$1F,$99,$3F,$CB,$4F,$F8,$02,$3C,$6F,$EE,$CF,$F0,$09,$FE,$00
	.dc.b	$AF,$E5,$F1,$FE,$02,$AF,$E5,$F3,$FE,$82,$8F,$E5,$8F,$F8,$04,$FE
	.dc.b	$59,$3F,$CB,$E7,$F3,$31,$FC,$02,$7F,$C0,$0F,$F9,$52,$7F,$D6,$11
	.dc.b	$FE,$56,$9F,$E9,$EF,$FA,$BE,$7F,$2E,$7E,$3C,$FB,$27,$E2,$2B,$F3
	.dc.b	$0F,$FA,$51,$FE,$BA,$FF,$B7,$0F,$F0,$2E,$1F,$EB,$0B,$FF,$29,$C7
	.dc.b	$E5,$E7,$F9,$81,$FE,$B0,$7F,$C5,$47,$E5,$E3,$FC,$AC,$FF,$9B,$1F
	.dc.b	$F0,$15,$FE,$56,$3F,$C6,$17,$FE,$55,$AF,$F5,$8D,$FF,$95,$5F,$D5
	.dc.b	$57,$DD,$AF,$E5,$F3,$FE,$56,$7F,$2C,$7E,$41,$FF,$00,$DF,$F9,$5A
	.dc.b	$3F,$35,$3F,$E5,$67,$FD,$B6,$7F,$2D,$3F,$D6,$11,$FC,$B5,$FF,$02
	.dc.b	$DF,$F2,$D3,$FA,$1B,$FB,$B7,$F2,$6F,$E4,$70,$FE,$FF,$87,$F0,$5C
	.dc.b	$3F,$CA,$93,$FE,$AE,$AF,$EF,$EF,$F2,$73,$F2,$35,$FD,$FD,$FE,$55
	.dc.b	$87,$E4,$5F,$8D,$3F,$91,$7F,$95,$4F,$F7,$E7,$F8,$AE,$1F,$DF,$A7
	.dc.b	$F2,$31,$FE,$06,$AF,$E1,$F6,$7F,$D5,$71,$FE,$6A,$C3,$F9,$65,$7E
	.dc.b	$4E,$3F,$96,$47,$ED,$9F,$E5,$93,$FE,$2C,$AF,$E5,$B7,$FE,$02,$BF
	.dc.b	$E5,$95,$FE,$B0,$7C,$2F,$C8,$7F,$2B,$7F,$C6,$67,$F2,$AF,$F1,$BB
	.dc.b	$FC,$AC,$FF,$AA,$AB,$FC,$06,$7E,$50,$FD,$8C,$FE,$51,$FE,$24,$FE
	.dc.b	$51,$FF,$42,$3F,$CA,$9F,$E5,$D7,$FC,$49,$F4,$9F,$CA,$8F,$E7,$B1
	.dc.b	$FC,$CE,$FF,$C0,$73,$FD,$FD,$FF,$01,$4F,$F3,$D7,$FD,$50,$7F,$41
	.dc.b	$AF,$F2,$A5,$7E,$50,$FF,$2A,$1F,$94,$9F,$EC,$67,$F9,$32,$FF,$CA
	.dc.b	$51,$FE,$56,$3F,$33,$7F,$CA,$A7,$FA,$B9,$EF,$3E,$6D,$7F,$81,$6B
	.dc.b	$F3,$47,$E5,$E3,$F9,$45,$7F,$AB,$2B,$FC,$F5,$5F,$DC,$2F,$F3,$15
	.dc.b	$FE,$4D,$8F,$EC,$75,$F9,$77,$F3,$31,$FD,$65,$FE,$EF,$C7,$F9,$78
	.dc.b	$FF,$31,$BF,$CB,$71,$3F,$CB,$E6,$CF,$FA,$D7,$13,$83,$F8,$34,$2B
	.dc.b	$11,$FD,$B2,$BF,$C1,$6E,$0F,$F5,$59,$FC,$DB,$0E,$11,$E3,$C5,$1E
	.dc.b	$69,$FF,$5B,$1C,$4F,$F0,$EC,$7F,$84,$2B,$11,$FE,$E5,$BF,$EF,$17
	.dc.b	$FB,$3B,$FF,$0A,$CF,$A2,$3F,$D7,$35,$C2,$FB,$DC,$9F,$E7,$A7,$17
	.dc.b	$B3,$FE,$90,$3F,$7D,$87,$C8,$BF,$F2,$15,$7E,$BD,$FF,$6B,$CF,$EB
	.dc.b	$8D,$B9,$6D,$7C,$08,$FE,$84,$F8,$EB,$FA,$CC,$79,$CF,$E8,$77,$9B
	.dc.b	$F3,$47,$F9,$C6,$74,$FF,$86,$1F,$EC,$35,$D6,$3F,$C4,$0F,$BC,$F8
	.dc.b	$6D,$7C,$EF,$F8,$8B,$0F,$A0,$FF,$88,$A3,$FC,$7B,$3D,$A3,$FC,$61
	.dc.b	$1F,$E2,$5B,$FA,$31,$FE,$96,$C3,$FA,$73,$FD,$C9,$FF,$67,$1F,$D4
	.dc.b	$6B,$94,$7F,$83,$63,$FA,$A5,$78,$23,$E5,$9A,$3F,$C9,$D1,$FE,$27
	.dc.b	$C3,$E0,$C7,$CC,$8F,$F4,$01,$99,$E5,$1F,$DE,$8E,$47,$E1,$4F,$21
	.dc.b	$FB,$F7,$9B,$FE,$28,$3D,$78,$7C,$0A,$B3,$F8,$C9,$FF,$42,$1F,$02
	.dc.b	$FF,$46,$F4,$3F,$C0,$55,$E2,$7F,$A1,$C7,$89,$FE,$B3,$7F,$8B,$7F
	.dc.b	$C9,$91,$F8,$07,$A5,$FE,$B8,$B5,$CE,$3F,$DB,$D3,$CC,$FF,$44,$D7
	.dc.b	$F0,$EB,$D4,$FF,$96,$4F,$03,$FA,$2B,$FA,$EF,$F0,$92,$9F,$95,$8F
	.dc.b	$F1,$94,$62,$39,$C7,$F0,$07,$C5,$1F,$7E,$3F,$B1,$4F,$F0,$73,$F8
	.dc.b	$63,$CC,$FE,$0B,$87,$E1,$9F,$CA,$D7,$EC,$5F,$81,$3F,$C0,$70,$FC
	.dc.b	$C4,$7F,$A3,$5E,$EF,$F8,$F3,$0F,$B7,$87,$C5,$BD,$CF,$F8,$E1,$F2
	.dc.b	$D7,$E7,$AF,$F7,$E7,$FA,$44,$FD,$EC,$FF,$B9,$E7,$F7,$AF,$FA,$49
	.dc.b	$FD,$93,$F7,$23,$FD,$25,$86,$2B,$E0,$E1,$FC,$88,$F1,$9F,$16,$3E
	.dc.b	$CD,$EB,$0E,$58,$7E,$EB,$0F,$F4,$A9,$FC,$2A,$3E,$34,$7F,$70,$7C
	.dc.b	$6F,$EE,$23,$FD,$04,$FC,$09,$FE,$81,$5F,$B9,$C3,$E9,$C7,$ED,$A7
	.dc.b	$F7,$8F,$C4,$7F,$D3,$15,$FB,$4B,$FA,$51,$FA,$03,$FC,$B1,$1F,$B5
	.dc.b	$8C,$CF,$E0,$1F,$4C,$FF,$35,$9F,$D9,$56,$70,$FC,$E4,$FE,$48,$FE
	.dc.b	$E7,$7F,$92,$9F,$A5,$1F,$AF,$3F,$4C,$7F,$A6,$1D,$47,$C2,$8F,$F3
	.dc.b	$8C,$E9,$FB,$91,$FA,$F7,$C6,$7F,$87,$CF,$D3,$57,$F7,$3A,$FD,$0D
	.dc.b	$7D,$59,$E0,$F3,$7F,$C6,$B5,$AB,$E7,$1F,$E9,$B3,$47,$DD,$7E,$33
	.dc.b	$FB,$E9,$FC,$93,$FB,$67,$6F,$FB,$4A,$7F,$62,$7C,$07,$FD,$F7,$3F
	.dc.b	$AE,$AC,$CF,$BE,$7F,$BC,$57,$28,$FF,$08,$5F,$F2,$47,$F4,$D3,$E6
	.dc.b	$3F,$DF,$A7,$F3,$83,$FC,$ED,$5F,$A6,$7A,$1F,$C2,$E3,$F0,$4F,$F1
	.dc.b	$6B,$FE,$34,$FF,$B7,$8F,$D8,$9F,$CC,$1E,$8F,$EC,$63,$F6,$D1,$D0
	.dc.b	$FD,$86,$1F,$C9,$23,$F5,$E7,$96,$7F,$BE,$3C,$9F,$05,$7F,$0F,$BF
	.dc.b	$D8,$C7,$F9,$6A,$7F,$3E,$FD,$B7,$F5,$AF,$3B,$FE,$D9,$5E,$A6,$CF
	.dc.b	$ED,$B0,$FD,$8D,$72,$3F,$DA,$91,$FB,$17,$F6,$B3,$E3,$7F,$96,$4F
	.dc.b	$9A,$FF,$64,$EA,$BE,$3C,$FF,$90,$5F,$EC,$CF,$EC,$5E,$4F,$F1,$AB
	.dc.b	$FD,$8C,$7C,$53,$FB,$25,$7F,$62,$9F,$D7,$9F,$86,$3F,$C2,$CF,$F6
	.dc.b	$37,$F6,$75,$97,$EA,$9F,$D8,$8F,$E5,$17,$F9,$E7,$FC,$BF,$7F,$B1
	.dc.b	$73,$3D,$67,$8C,$7F,$61,$3F,$AC,$9F,$96,$9F,$A4,$FF,$61,$9F,$EA
	.dc.b	$91,$FD,$7E,$BF,$CC,$D7,$CA,$BF,$D3,$67,$F4,$F3,$C4,$7F,$5E,$7F
	.dc.b	$5A,$FE,$BA,$FF,$6D,$1F,$AD,$8F,$EA,$D1,$C1,$FF,$4B,$3F,$AC,$7F
	.dc.b	$AD,$9F,$AB,$7F,$94,$C7,$F5,$48,$FE,$B2,$F1,$7F,$D1,$D5,$FA,$AB
	.dc.b	$C9,$FA,$99,$F4,$4F,$FA,$4E,$FF,$A9,$BF,$D2,$0F,$F2,$CD,$7E,$2D
	.dc.b	$FA,$67,$F9,$74,$FD,$7C,$FE,$09,$FE,$AF,$3F,$E5,$A3,$FA,$B6,$1F
	.dc.b	$A4,$AF,$EA,$B8,$7E,$96,$BF,$C6,$33,$FA,$5B,$FF,$27,$57,$F0,$39
	.dc.b	$FF,$03,$9F,$A6,$7F,$38,$7D,$47,$F4,$98,$7F,$5E,$8F,$C5,$C7,$F5
	.dc.b	$5C,$3F,$C9,$6F,$E8,$23,$F4,$93,$EA,$8F,$F2,$11,$FB,$77,$FA,$A4
	.dc.b	$FF,$81,$6B,$FC,$95,$3F,$A4,$AE,$38,$7D,$33,$FD,$90,$71,$3F,$59
	.dc.b	$1F,$E8,$D9,$FC,$FB,$FA,$39,$FD,$A4,$7A,$63,$FC,$8A,$FE,$8C,$E2
	.dc.b	$FD,$33,$F3,$87,$D3,$3F,$57,$3F,$EE,$0A,$FF,$04,$DE,$6B,$E9,$4F
	.dc.b	$DC,$3F,$90,$CF,$D2,$3F,$C1,$11,$FC,$5A,$7E,$F4,$F8,$2B,$FC,$FB
	.dc.b	$3F,$C5,$9F,$98,$F8,$2F,$D1,$1F,$C5,$A3,$F8,$F4,$7D,$08,$FA,$E7
	.dc.b	$F4,$B8,$FE,$2A,$FD,$1C,$3F,$C0,$91,$FC,$4E,$3F,$C1,$2F,$F0,$AB
	.dc.b	$FC,$4C,$7F,$A1,$9F,$E5,$EF,$F9,$DE,$FE,$0C,$7F,$13,$8F,$F3,$E1
	.dc.b	$FE,$7F,$9F,$E2,$67,$F4,$47,$FD,$1E,$FE,$2D,$FE,$78,$FF,$81,$30
	.dc.b	$FF,$78,$3F,$E7,$D9,$FE,$30,$FE,$6E,$7F,$9B,$1F,$92,$AF,$F0,$1E
	.dc.b	$1F,$E3,$E8,$FE,$92,$7F,$5F,$8F,$F0,$1C,$FF,$12,$8F,$EF,$F1,$FB
	.dc.b	$37,$94,$FF,$13,$BF,$F3,$9C,$7F,$80,$AB,$F8,$8B,$FC,$4A,$BF,$CD
	.dc.b	$75,$FE,$00,$BF,$E2,$35,$F9,$0B,$FE,$22,$FF,$7A,$BF,$3C,$FF,$59
	.dc.b	$9F,$EF,$D3,$FC,$41,$FF,$79,$47,$F7,$23,$FC,$AD,$1F,$E6,$A9,$FE
	.dc.b	$22,$66,$FF,$88,$3F,$9E,$9F,$E2,$A6,$67,$FB,$AC,$7F,$10,$3F,$86
	.dc.b	$5F,$F0,$F7,$FC,$CF,$3F,$C3,$E7,$E4,$47,$F8,$88,$F8,$87,$EF,$A7
	.dc.b	$E6,$4E,$4F,$E7,$D3,$FE,$2D,$9F,$96,$7F,$39,$9F,$94,$FE,$40,$FE
	.dc.b	$F1,$3F,$CF,$67,$FC,$4F,$1F,$20,$FE,$78,$FB,$27,$FC,$85,$7F,$CE
	.dc.b	$AB,$33,$F2,$EB,$F9,$D1,$F1,$CF,$87,$1F,$E2,$79,$F9,$77,$FC,$EA
	.dc.b	$BE,$55,$7C,$B8,$FF,$49,$C7,$F8,$86,$3F,$9C,$4F,$F0,$D3,$D0,$FD
	.dc.b	$23,$E5,$4F,$F8,$86,$7E,$4C,$7C,$A8,$FF,$47,$47,$F6,$68,$FE,$CF
	.dc.b	$1F,$94,$8F,$EC,$B1,$FD,$9A,$3F,$27,$3F,$EC,$38,$F9,$31,$FD,$9A
	.dc.b	$FE,$4C,$7F,$9B,$67,$FD,$89,$5F,$92,$7F,$AE,$C7,$C8,$9F,$C9,$3F
	.dc.b	$EB,$EB,$F8,$F5,$FE,$4F,$8F,$8F,$7F,$E5,$1A,$FC,$A5,$7F,$63,$3E
	.dc.b	$55,$7F,$5F,$AF,$8D,$7F,$09,$FE,$57,$5F,$E4,$FA,$FC,$7D,$FF,$5E
	.dc.b	$AD,$5F,$EE,$EF,$FD,$96,$FF,$AD,$E3,$2F,$E6,$0F,$BB,$1F,$EB,$68
	.dc.b	$CC,$7E,$4A,$3E,$4C,$7F,$BC,$1F,$11,$FB,$C3,$FC,$B8,$FF,$93,$4F
	.dc.b	$94,$FE,$F6,$7E,$54,$FF,$83,$23,$FC,$1B,$1F,$E1,$28,$FC,$8C,$7E
	.dc.b	$EE,$7F,$C0,$F7,$F2,$9F,$C8,$57,$F6,$17,$E1,$5F,$F6,$06,$D7,$E9
	.dc.b	$8F,$F1,$DB,$FB,$A7,$FC,$14,$FF,$B9,$5F,$D1,$47,$F2,$87,$FC,$12
	.dc.b	$7E,$EE,$BE,$63,$F9,$23,$FD,$D7,$5F,$E0,$27,$FD,$C4,$7F,$82,$DF
	.dc.b	$C9,$57,$CC,$AF,$F0,$4C,$7E,$4A,$3F,$C7,$07,$FA,$9C,$F9,$53,$E8
	.dc.b	$7F,$D4,$C7,$C9,$7E,$F4,$7F,$2C,$4F,$B5,$FE,$78,$7E,$91,$FA,$47
	.dc.b	$F1,$F7,$F9,$34,$7E,$51,$F9,$87,$EA,$AB,$FA,$AA,$3F,$C3,$37,$FC
	.dc.b	$DA,$3F,$C7,$11,$B8,$F8,$17,$F3,$E3,$F6,$6F,$F2,$96,$9F,$C0,$0F
	.dc.b	$E7,$EF,$E1,$1F,$F0,$71,$F2,$8F,$E5,$37,$FC,$02,$FE,$DB,$C0,$F5
	.dc.b	$47,$D6,$7E,$85,$7E,$6E,$3F,$81,$47,$D0,$7F,$82,$C7,$E6,$9F,$E3
	.dc.b	$CF,$A9,$FF,$70,$1F,$98,$3F,$96,$1F,$E6,$5B,$FE,$56,$FF,$03,$AF
	.dc.b	$A4,$FC,$D3,$F8,$1B,$FE,$D6,$9F,$CD,$3F,$D5,$23,$E9,$47,$E6,$A3
	.dc.b	$F8,$1D,$7F,$54,$8F,$E0,$53,$FC,$0E,$3F,$34,$FC,$83,$F8,$1C,$FF
	.dc.b	$52,$3F,$30,$7F,$28,$3F,$30,$FE,$58,$FE,$06,$7E,$52,$BF,$2F,$1F
	.dc.b	$2A,$FF,$A8,$5F,$F9,$7E,$BF,$2D,$3F,$C0,$2B,$FA,$83,$F9,$7B,$FC
	.dc.b	$B4,$FE,$F2,$BF,$D8,$0F,$F8,$AE,$BF,$CB,$8F,$F8,$EC,$FE,$78,$7F
	.dc.b	$A2,$E3,$FC,$56,$7F,$96,$1F,$F1,$3C,$7F,$98,$E7,$81,$FE,$5B,$3F
	.dc.b	$C5,$4F,$F8,$46,$3F,$D7,$07,$F8,$A9,$FF,$2C,$CF,$F9,$25,$FB,$91
	.dc.b	$FE,$04,$8F,$F1,$95,$FF,$74,$BF,$F2,$CC,$9F,$D2,$47,$F9,$D3,$0F
	.dc.b	$F1,$5C,$62,$7F,$CB,$47,$2C,$3F,$C8,$55,$FC,$0E,$3F,$B2,$C7,$F8
	.dc.b	$A8,$FF,$2D,$47,$FB,$DA,$7A,$C7,$F9,$0E,$FD,$55,$FE,$56,$3F,$CB
	.dc.b	$57,$E3,$7F,$C5,$B1,$FE,$63,$3D,$6F,$F8,$B9,$FF,$26,$3F,$AD,$3F
	.dc.b	$96,$CF,$F5,$7A,$FF,$5E,$1F,$C1,$A3,$EB,$4F,$F8,$A5,$FE,$5D,$1F
	.dc.b	$95,$38,$DF,$F8,$B0,$FE,$FD,$3F,$D5,$E7,$F7,$87,$F5,$77,$F2,$B3
	.dc.b	$FC,$B8,$FF,$13,$BF,$D5,$E3,$FA,$C5,$7F,$57,$8F,$F1,$34,$7F,$00
	.dc.b	$AF,$EA,$F1,$FD,$F2,$BF,$77,$3F,$CA,$E3,$FB,$E4,$7E,$56,$BF,$AB
	.dc.b	$4F,$EE,$27,$FC,$92,$7F,$55,$71,$87,$E5,$9F,$EA,$86,$E3,$FC,$98
	.dc.b	$0F,$9D,$63,$FC,$A5,$1F,$C5,$DC,$3F,$C7,$CF,$E5,$67,$E1,$6F,$F9
	.dc.b	$14,$6A,$3F,$92,$5F,$FA,$25,$E9,$1D,$1E,$73,$C2,$31,$4C,$3C,$71
	.dc.b	$F8,$09,$D5,$78,$6F,$FD,$1E,$F4,$30,$7F,$3E,$C3,$94,$F0,$8F,$19
	.dc.b	$FE,$7B,$AF,$33,$EA,$AF,$1B,$CC,$DB,$F1,$6F,$C7,$1F,$E4,$9C,$3E
	.dc.b	$41,$87,$DF,$87,$CB,$7F,$A2,$61,$E1,$7F,$D9,$95,$B7,$FD,$9D,$56
	.dc.b	$D7,$FE,$0A,$C3,$CB,$1F,$EA,$6B,$E1,$7F,$4A,$FE,$C6,$1D,$6B,$D6
	.dc.b	$DB,$FA,$95,$EB,$3F,$B0,$BF,$EE,$C3,$A9,$FC,$48,$AF,$BE,$3D,$0F
	.dc.b	$FB,$42,$BB,$BC,$EB,$FD,$A6,$FB,$8C,$57,$D1,$BF,$63,$F3,$CF,$F4
	.dc.b	$E1,$FC,$06,$37,$3D,$DF,$31,$F3,$2B,$BC,$FC,$27,$E6,$47,$C7,$7F
	.dc.b	$D5,$8F,$8D,$FC,$D6,$1C,$27,$FD,$BE,$79,$4E,$51,$DD,$FC,$0C,$FF
	.dc.b	$B2,$B0,$F7,$BF,$DC,$E3,$2F,$C7,$7F,$BA,$C6,$A3,$FC,$0D,$1A,$7F
	.dc.b	$07,$1C,$AF,$F4,$CD,$8F,$F0,$8B,$9B,$F3,$47,$FB,$6A,$7F,$23,$5E
	.dc.b	$07,$AB,$FA,$18,$F8,$B3,$F0,$AB,$EA,$BF,$BF,$9F,$5B,$F0,$AF,$F0
	.dc.b	$EF,$E2,$E3,$FD,$CE,$FF,$85,$1D,$C7,$A2,$BE,$8C,$7E,$FC,$FF,$76
	.dc.b	$9F,$0C,$AF,$E5,$A3,$F2,$71,$F8,$93,$27,$BA,$39,$9D,$5F,$9E,$7F
	.dc.b	$08,$78,$DF,$F0,$6B,$FC,$7C,$FE,$1E,$7F,$87,$1F,$EC,$33,$6F,$53
	.dc.b	$9C,$F5,$77,$1F,$E4,$C9,$FC,$CD,$7C,$59,$CB,$F1,$63,$2F,$D0,$8C
	.dc.b	$BF,$42,$7C,$2F,$FB,$F6,$3E,$F3,$F6,$23,$F8,$59,$F7,$8F,$D2,$1F
	.dc.b	$B0,$8F,$DC,$BF,$89,$3F,$8A,$9D,$CB,$1F,$C5,$9F,$D0,$9F,$E5,$67
	.dc.b	$B3,$FC,$54,$FC,$D4,$7F,$34,$AF,$E0,$F8,$7F,$09,$9F,$90,$7F,$18
	.dc.b	$9D,$BF,$C9,$27,$F4,$11,$F8,$2A,$F9,$13,$FC,$AF,$0F,$C6,$3F,$BD
	.dc.b	$7F,$5B,$3F,$C4,$9F,$91,$19,$9F,$53,$FC,$81,$EA,$FD,$89,$FF,$06
	.dc.b	$3F,$C1,$0F,$8C,$FF,$0E,$9F,$80,$D9,$FE,$0F,$3D,$0F,$E2,$E7,$43
	.dc.b	$FA,$84,$7C,$58,$FE,$1B,$5D,$23,$F8,$6B,$F1,$30,$FE,$44,$FE,$5D
	.dc.b	$B1,$F1,$2B,$FA,$AB,$F3,$F0,$FE,$5B,$7F,$7E,$3F,$C2,$53,$FC,$3E
	.dc.b	$3A,$3F,$C3,$DE,$8D,$8E,$8F,$96,$3D,$B1,$FE,$BE,$AF,$E1,$13,$F9
	.dc.b	$98,$FF,$42,$47,$F0,$B7,$8C,$7B,$AF,$F9,$FB,$F2,$A3,$F8,$23,$A3
	.dc.b	$FA,$C4,$FF,$0B,$AD,$D7,$C9,$9F,$6B,$88,$FC,$5C,$7E,$E4,$EF,$87
	.dc.b	$E9,$5F,$EB,$8E,$8F,$E1,$0F,$F5,$FB,$F4,$3F,$D8,$67,$9C,$7F,$07
	.dc.b	$8F,$C0,$1F,$BB,$3F,$80,$CE,$9F,$24,$7F,$A6,$CF,$E0,$6F,$E4,$8F
	.dc.b	$80,$7C,$27,$DA,$FF,$14,$AF,$9F,$1F,$C0,$0D,$3F,$B7,$8F,$91,$5E
	.dc.b	$E3,$F3,$73,$FC,$F9,$F7,$47,$FB,$E9,$E7,$1F,$C0,$0F,$A2,$FF,$25
	.dc.b	$7F,$81,$BF,$22,$3E,$94,$7F,$86,$5F,$BF,$3F,$E2,$3A,$FB,$F1,$FB
	.dc.b	$E7,$F7,$31,$F9,$88,$FD,$B1,$F9,$B3,$FC,$B1,$7F,$DE,$5F,$E4,$8D
	.dc.b	$8F,$C7,$D7,$F0,$A7,$CF,$3F,$BB,$8F,$0B,$F9,$99,$FE,$87,$7F,$2E
	.dc.b	$3F,$C4,$4F,$E9,$5F,$DF,$3D,$27,$FD,$05,$7E,$63,$EF,$CF,$EE,$E7
	.dc.b	$D6,$FA,$5F,$5C,$FD,$F8,$FD,$49,$CC,$FF,$1A,$3F,$C7,$DC,$BF,$18
	.dc.b	$FF,$73,$47,$48,$FF,$23,$DF,$9C,$FF,$05,$BF,$BC,$AD,$47,$F9,$4C
	.dc.b	$F8,$31,$F9,$CA,$E9,$1F,$2E,$3F,$C1,$15,$FB,$77,$EA,$BF,$C6,$A7
	.dc.b	$FA,$5C,$FF,$8E,$2B,$EA,$CF,$C0,$AF,$DC,$61,$FB,$07,$E4,$1F,$D3
	.dc.b	$0F,$D0,$9F,$CA,$70,$FD,$C3,$BC,$32,$FF,$48,$AF,$D9,$BF,$B7,$C3
	.dc.b	$F4,$93,$FD,$F7,$0F,$C6,$47,$ED,$E7,$FC,$6A,$FE,$DE,$BF,$BC,$9E
	.dc.b	$28,$1F,$B9,$8C,$CF,$ED,$E7,$F8,$03,$F9,$99,$FD,$B4,$7F,$76,$79
	.dc.b	$1F,$E7,$59,$C4,$FE,$D6,$BF,$61,$3F,$AB,$3F,$CF,$67,$E2,$CF,$AB
	.dc.b	$7F,$DD,$AB,$D7,$3F,$E0,$67,$EA,$CF,$DC,$8F,$DA,$9F,$11,$FF,$25
	.dc.b	$5F,$F2,$E3,$C6,$7E,$D1,$FC,$2B,$C9,$FF,$36,$C7,$ED,$23,$C9,$1F
	.dc.b	$7A,$BF,$C6,$51,$FA,$E8,$FC,$A1,$FE,$0D,$BF,$C4,$BF,$93,$3F,$C2
	.dc.b	$15,$FA,$87,$F1,$07,$F4,$BA,$FD,$83,$FA,$CB,$FC,$E3,$FE,$82,$AE
	.dc.b	$D1,$FD,$54,$FF,$4E,$DF,$E6,$E7,$FC,$48,$FE,$AC,$FB,$11,$FE,$27
	.dc.b	$9F,$CD,$9F,$DA,$AF,$FB,$64,$F3,$8F,$F4,$5E,$1F,$5E,$3F,$5F,$7C
	.dc.b	$AB,$FC,$18,$7F,$9A,$9F,$9F,$7F,$EB,$48,$FD,$74,$7E,$8D,$F6,$3F
	.dc.b	$E6,$E3,$F4,$F7,$FA,$C8,$FF,$03,$C7,$E4,$1F,$D5,$D7,$AC,$FE,$C5
	.dc.b	$1F,$EE,$D3,$F5,$6F,$C4,$7F,$B9,$4F,$EA,$A3,$E2,$1E,$AC,$3E,$AB
	.dc.b	$FC,$82,$7E,$84,$FD,$47,$D2,$7F,$92,$4F,$F1,$1E,$1F,$A1,$AF,$A7
	.dc.b	$1F,$E1,$DC,$3C,$07,$F8,$3E,$3F,$93,$61,$F5,$F0,$FA,$58,$7A,$2F
	.dc.b	$FC,$28,$7F,$36,$7F,$81,$3F,$C8,$9F,$AD,$39,$9F,$91,$3F,$C4,$E3
	.dc.b	$F1,$77,$FE,$E0,$AF,$91,$3F,$42,$7D,$0F,$F2,$2A,$FC,$14,$7D,$03
	.dc.b	$FD,$5B,$5F,$64,$F4,$BF,$41,$FF,$2E,$5F,$F8,$5A,$7E,$81,$FC,$8C
	.dc.b	$FE,$B3,$3F,$C6,$1F,$4D,$7E,$41,$FC,$09,$FC,$66,$BF,$D5,$18,$7F
	.dc.b	$19,$9F,$F0,$31,$FB,$63,$F8,$D6,$1F,$D6,$DF,$E2,$07,$EB,$0F,$EB
	.dc.b	$A7,$FA,$4C,$FE,$30,$7E,$E4,$FC,$39,$FD,$03,$0F,$F4,$F6,$19,$BF
	.dc.b	$E0,$D5,$E8,$AF,$F7,$31,$FC,$05,$FD,$04,$FF,$94,$4F,$F0,$8E,$1F
	.dc.b	$BC,$8F,$B1,$3F,$C4,$0F,$E2,$67,$F0,$99,$FD,$D3,$F9,$F3,$F8,$E3
	.dc.b	$97,$F8,$B3,$FD,$A9,$FF,$04,$1C,$6B,$F8,$64,$7F,$B4,$4F,$F0,$53
	.dc.b	$FE,$92,$AF,$F0,$41,$FC,$50,$F9,$13,$FE,$07,$AC,$D7,$F1,$33,$F8
	.dc.b	$E4,$FA,$27,$FD,$94,$7F,$9E,$8E,$27,$FA,$6A,$7E,$84,$FD,$43,$E6
	.dc.b	$47,$F2,$D3,$D0,$FF,$8F,$AB,$FC,$77,$3F,$D1,$A3,$FA,$51,$F3,$EB
	.dc.b	$FA,$05,$7C,$A3,$E7,$3F,$CA,$67,$FA,$E4,$7F,$AD,$E7,$33,$FE,$33
	.dc.b	$8F,$9B,$5F,$D0,$A3,$E5,$C7,$CC,$8F,$E2,$51,$FE,$B6,$8F,$F1,$79
	.dc.b	$F2,$EB,$FC,$83,$3F,$E3,$C3,$F9,$AD,$7E,$84,$FF,$55,$CF,$F3,$93
	.dc.b	$E3,$D7,$F3,$D3,$CA,$7F,$57,$7F,$C6,$8F,$E5,$5F,$CC,$9F,$1A,$BE
	.dc.b	$4B,$FD,$BE,$BF,$BB,$4F,$F6,$A8,$FC,$BC,$7C,$88,$FF,$31,$D7,$C8
	.dc.b	$8F,$E1,$93,$FE,$EA,$9F,$EC,$75,$FD,$46,$3F,$9B,$47,$F7,$0A,$FC
	.dc.b	$A9,$F2,$1F,$ED,$51,$F2,$2F,$FD,$AE,$FE,$30,$FF,$2E,$9F,$08,$FC
	.dc.b	$9D,$FF,$70,$AF,$F2,$D9,$FE,$D8,$BF,$92,$FF,$66,$8F,$C9,$5F,$F6
	.dc.b	$5A,$F9,$11,$FD,$9E,$BF,$23,$1F,$DA,$6B,$E5,$D7,$C7,$8F,$E0,$15
	.dc.b	$FE,$19,$AF,$8F,$7F,$E8,$D8,$F8,$F5,$FE,$14,$AF,$F0,$AC,$7C,$79
	.dc.b	$FD,$DC,$7C,$88,$F8,$F6,$3F,$D9,$11,$F8,$F8,$FF,$08,$D7,$C7,$8F
	.dc.b	$91,$1F,$E1,$88,$FC,$74,$7C,$88,$FC,$75,$7C,$8A,$FF,$07,$9F,$B9
	.dc.b	$AF,$C5,$D7,$E3,$A9,$5F,$E0,$FB,$F8,$F3,$83,$FD,$09,$7F,$BD,$BF
	.dc.b	$DC,$C7,$C8,$AF,$F7,$ED,$7C,$87,$FD,$FF,$5F,$22,$BF,$93,$57,$C8
	.dc.b	$9F,$F1,$6C,$FB,$E3,$F9,$3C,$FF,$8F,$5F,$C8,$4F,$F5,$C3,$73,$FE
	.dc.b	$45,$BF,$F1,$6C,$FD,$E8,$C5,$FE,$E6,$3E,$65,$70,$7F,$C7,$67,$FA
	.dc.b	$08,$FC,$A1,$F3,$5F,$F3,$DC,$7F,$8F,$67,$F2,$75,$F3,$6F,$F2,$77
	.dc.b	$F3,$6F,$F2,$77,$FE,$3B,$8F,$DD,$47,$E5,$A7,$E6,$C7,$EE,$EB,$E5
	.dc.b	$5F,$E4,$E6,$BF,$E3,$B3,$F2,$95,$FE,$3F,$8F,$9B,$5F,$95,$AF,$CA
	.dc.b	$47,$E4,$E3,$F2,$91,$FE,$3D,$8F,$CA,$4F,$FB,$82,$3F,$29,$1F,$2E
	.dc.b	$3F,$28,$7C,$B9,$FC,$75,$7C,$B7,$F7,$71,$FE,$DD,$3F,$2A,$7F,$50
	.dc.b	$AF,$CA,$CF,$E5,$27,$FC,$CE,$7F,$22,$BF,$90,$FE,$52,$BF,$D4,$B7
	.dc.b	$F9,$47,$F7,$31,$FD,$26,$7E,$44,$7E,$56,$FF,$90,$BF,$E5,$F9,$FC
	.dc.b	$A3,$F9,$3A,$F8,$F1,$F6,$8F,$EF,$67,$E5,$27,$F7,$73,$F7,$E3,$FD
	.dc.b	$A0,$7E,$54,$F7,$9F,$E8,$7A,$FC,$B3,$83,$FD,$AD,$7F,$97,$9F,$8C
	.dc.b	$FE,$56,$3F,$D0,$75,$FD,$42,$31,$1F,$E6,$1C,$3F,$D9,$F7,$FE,$65
	.dc.b	$8F,$F2,$C1,$FE,$B2,$8F,$92,$FE,$4A,$B9,$BF,$E6,$19,$FF,$34,$C7
	.dc.b	$F8,$D8,$E2,$7F,$AB,$DF,$EE,$E9,$FF,$00,$47,$91,$FF,$0B,$BF,$ED
	.dc.b	$27,$FC,$5E,$FC,$83,$FD,$9D,$3F,$39,$F8,$02,$7D,$77,$F2,$63,$17
	.dc.b	$FE,$62,$7F,$D9,$E7,$F8,$CC,$FF,$3A,$3E,$D8,$FC,$C6,$1D,$E7,$FD
	.dc.b	$5A,$7F,$98,$AF,$FD,$9D,$19,$BF,$B3,$7F,$E3,$88,$FF,$19,$9F,$E5
	.dc.b	$DA,$FF,$30,$E1,$FE,$D4,$8F,$F0,$DC,$7C,$97,$FC,$64,$7F,$98,$5F
	.dc.b	$F0,$11,$FD,$6A,$BF,$43,$87,$FB,$20,$FD,$0C,$7F,$81,$E7,$FA,$D3
	.dc.b	$FC,$D2,$3F,$40,$FC,$43,$FC,$60,$53,$FD,$8E,$FC,$57,$FC,$88,$DA
	.dc.b	$FA,$71,$FE,$C8,$7F,$AF,$1E,$87,$F8,$8B,$FE,$06,$3F,$B9,$3B,$9F
	.dc.b	$F1,$5B,$FD,$64,$FF,$0D,$9F,$E6,$37,$FC,$0F,$3F,$A7,$3F,$87,$3F
	.dc.b	$E0,$78,$FE,$1D,$1F,$A7,$3F,$D9,$17,$FC,$32,$3D,$8F,$F8,$1D,$FF
	.dc.b	$74,$1F,$A5,$7F,$36,$FF,$BD,$E7,$F8,$74,$FF,$3A,$BF,$E1,$D1,$FD
	.dc.b	$73,$93,$FD,$ED,$5F,$89,$3F,$68,$E2,$7D,$F8,$A3,$13,$EF,$6B,$FD
	.dc.b	$05,$B4,$9E,$13,$D8,$FC,$B4,$0C,$BF,$E6,$9A,$FE,$7B,$5D,$9F,$D3
	.dc.b	$56,$A7,$F8,$44,$7E,$9E,$B1,$1F,$B3,$3F,$3E,$60,$FD,$AC,$77,$9C
	.dc.b	$47,$D7,$9F,$15,$7B,$F1,$F2,$48,$DD,$FF,$B4,$EE,$CF,$A4,$E3,$3F
	.dc.b	$D6,$2F,$E1,$35,$F8,$57,$F0,$71,$FF,$7B,$BF,$35,$F9,$F3,$FD,$8D
	.dc.b	$FC,$85,$F8,$1F,$9D,$19,$3A,$9F,$CA,$4F,$7D,$76,$7E,$DC,$7C,$E8
	.dc.b	$E5,$87,$CC,$9E,$CF,$CC,$BF,$13,$FC,$4E,$77,$5F,$06,$73,$1D,$6B
	.dc.b	$C0,$7D,$96,$47,$E1,$DF,$93,$1F,$21,$F8,$D7,$5E,$D3,$B7,$D7,$3F
	.dc.b	$EB,$77,$EB,$47,$E1,$E3,$E2,$96,$FE,$CB,$FB,$29,$AF,$F8,$48,$A7
	.dc.b	$F8,$C6,$2C,$7D,$68,$FC,$BC,$F3,$7F,$94,$3D,$70,$D4,$7C,$F8,$F0
	.dc.b	$61,$A8,$FC,$D5,$FC,$07,$F5,$35,$F0,$2B,$F3,$CF,$E2,$A7,$CE,$FE
	.dc.b	$1E,$E9,$F0,$2B,$FD,$3C,$F9,$EB,$8B,$D6,$3E,$7D,$FC,$27,$F0,$0F
	.dc.b	$F5,$19,$FC,$64,$7C,$C3,$F0,$27,$53,$FC,$4F,$1F,$E3,$6B,$F9,$2F
	.dc.b	$E4,$A3,$F0,$F1,$F8,$57,$BB,$FD,$9A,$7F,$30,$F5,$8C,$CF,$E1,$5E
	.dc.b	$2F,$C5,$78,$D7,$CA,$7F,$8A,$9F,$61,$F8,$95,$FC,$0D,$F8,$8D,$7D
	.dc.b	$11,$FB,$17,$F8,$34,$F5,$8F,$D0,$47,$E4,$4F,$BA,$7E,$D9,$F8,$C7
	.dc.b	$F0,$33,$F3,$C7,$F8,$94,$F5,$1F,$C2,$8F,$C4,$4F,$F1,$8C,$3F,$17
	.dc.b	$7F,$46,$74,$FC,$56,$E1,$FA,$36,$DF,$86,$7D,$27,$F8,$39,$D4,$7D
	.dc.b	$E7,$E2,$3F,$C6,$4F,$94,$7E,$EE,$3F,$87,$5F,$F1,$39,$F3,$CF,$F8
	.dc.b	$3E,$BE,$1B,$FC,$4A,$3F,$4E,$FE,$01,$EA,$FF,$22,$3A,$9F,$CA,$30
	.dc.b	$FE,$1B,$5F,$C4,$63,$A9,$FE,$24,$8F,$86,$FF,$0F,$9F,$4B,$F8,$AB
	.dc.b	$F8,$8F,$F8,$86,$7F,$A4,$3F,$C3,$AF,$6F,$F8,$42,$7A,$B6,$7C,$6F
	.dc.b	$18,$F4,$BF,$C3,$2B,$EE,$5F,$F0,$58,$D9,$F2,$E3,$ED,$BF,$1D,$FC
	.dc.b	$4B,$FE,$8B,$7E,$2C,$7F,$0D,$7C,$71,$FC,$9E,$3F,$86,$3F,$9D,$3F
	.dc.b	$87,$5F,$C8,$7F,$4E,$7F,$0A,$BF,$C8,$3E,$5A,$FC,$2B,$E4,$3F,$A8
	.dc.b	$5F,$E2,$2B,$94,$7F,$52,$BF,$E1,$33,$F8,$CA,$ED,$5F,$C1,$EB,$F8
	.dc.b	$4C,$78,$DF,$F5,$4C,$6A,$3F,$47,$1F,$CD,$6F,$C6,$FF,$21,$8F,$44
	.dc.b	$7E,$82,$3D,$26,$8F,$E4,$93,$A7,$9C,$7F,$03,$BF,$E1,$35,$CA,$FF
	.dc.b	$3D,$5F,$E7,$5A,$FF,$02,$47,$F0,$83,$ED,$C7,$F2,$D8,$F8,$4D,$8D
	.dc.b	$4F,$CB,$3D,$D1,$F2,$4F,$21,$FE,$15,$AF,$E0,$15,$CA,$7F,$7F,$5F
	.dc.b	$C1,$DF,$C7,$D7,$F0,$07,$F7,$CF,$98,$FF,$38,$9F,$84,$3F,$78,$7E
	.dc.b	$42,$3E,$DD,$7E,$F4,$FE,$95,$3F,$A6,$8E,$31,$FE,$0F,$BF,$CC,$C7
	.dc.b	$F8,$36,$3F,$C2,$57,$F7,$E3,$F7,$98,$7F,$3A,$BE,$57,$EE,$C3,$FC
	.dc.b	$2A,$FF,$60,$C3,$F7,$4E,$A7,$F8,$2B,$FD,$A6,$BF,$99,$D7,$B8,$FD
	.dc.b	$C1,$FE,$01,$3D,$87,$F9,$E2,$3D,$18,$7E,$DA,$3F,$84,$3F,$E9,$5B
	.dc.b	$FE,$6F,$7A,$C3,$E3,$57,$EC,$EF,$DB,$7F,$E8,$77,$4F,$B4,$FF,$1A
	.dc.b	$E1,$F4,$6F,$E3,$1E,$DA,$FF,$3C,$3E,$47,$F6,$95,$A8,$FE,$42,$7B
	.dc.b	$70,$FD,$7C,$FE,$18,$FE,$B0,$7E,$C2,$3F,$5C,$FE,$BF,$0F,$E7,$73
	.dc.b	$FD,$D6,$3F,$5F,$1F,$AD,$9F,$E6,$B8,$73,$8F,$ED,$13,$FD,$D2,$7F
	.dc.b	$50,$FE,$52,$FF,$A7,$BF,$AE,$7F,$59,$59,$9E,$33,$FE,$1C,$9F,$E2
	.dc.b	$2F,$F6,$A3,$F3,$B3,$E8,$7F,$B9,$9F,$A2,$7E,$93,$FD,$D0,$E6,$59
	.dc.b	$FD,$63,$CA,$BF,$7C,$FF,$7A,$79,$1F,$B4,$38,$BF,$B1,$7F,$08,$FF
	.dc.b	$8B,$CF,$D3,$47,$27,$FD,$C4,$FE,$9E,$72,$7E,$B9,$F5,$C7,$F6,$DB
	.dc.b	$FF,$3B,$BF,$9B,$3F,$A1,$9F,$D9,$6F,$F4,$8F,$B0,$FE,$D5,$3F,$CA
	.dc.b	$9F,$F2,$64,$FF,$9D,$5E,$57,$F8,$37,$F8,$3D,$72,$8F,$EC,$71,$F8
	.dc.b	$C9,$FF,$0E,$D7,$E8,$E3,$F5,$95,$FA,$77,$FC,$D4,$FE,$9E,$F3,$3F
	.dc.b	$D1,$A7,$E6,$3F,$A5,$7F,$30,$78,$6F,$FD,$5D,$7F,$D7,$6B,$35,$FD
	.dc.b	$09,$F4,$D7,$F9,$88,$FD,$05,$71,$8F,$D2,$4F,$F9,$8A,$7F,$CD,$2F
	.dc.b	$F4,$A3,$E5,$BF,$E5,$73,$F8,$23,$E9,$7F,$AA,$3F,$C4,$4F,$EA,$6F
	.dc.b	$F5,$57,$D4,$FF,$A8,$EB,$F4,$B5,$FE,$58,$BF,$E9,$33,$FD,$00,$FE
	.dc.b	$6C,$7F,$98,$EB,$FA,$6C,$FE,$F2,$3F,$D3,$B5,$FE,$59,$9F,$D2,$C7
	.dc.b	$E9,$A7,$FA,$84,$7F,$94,$23,$F4,$07,$F3,$2A,$FE,$D2,$FE,$86,$3F
	.dc.b	$A7,$C7,$F3,$2A,$FA,$F1,$FC,$5A,$7F,$46,$7E,$44,$FE,$CF,$3F,$D3
	.dc.b	$6F,$27,$F3,$E3,$FC,$9D,$1F,$E4,$C3,$F4,$46,$6F,$F4,$31,$FA,$53
	.dc.b	$15,$FD,$24,$F7,$1F,$BA,$9F,$EE,$55,$99,$FC,$FD,$7F,$17,$AF,$E1
	.dc.b	$35,$FE,$3F,$3F,$3C,$7E,$14,$FF,$55,$9F,$E2,$F3,$F3,$C7,$F4,$A7
	.dc.b	$F3,$D1,$FC,$CC,$FE,$8B,$1F,$EA,$07,$FD,$88,$73,$3F,$A7,$3F,$DF
	.dc.b	$CD,$1F,$9F,$3F,$BE,$C7,$E7,$DF,$F1,$C5,$7F,$4E,$3E,$8B,$FD,$22
	.dc.b	$3F,$A2,$9F,$86,$7F,$C2,$6F,$F4,$07,$F9,$DC,$AB,$F7,$31,$FD,$08
	.dc.b	$FE,$C0,$FB,$1F,$F6,$2E,$1F,$DD,$2B,$FD,$87,$1F,$A0,$BF,$E8,$77
	.dc.b	$C0,$FF,$2C,$DF,$A6,$7F,$C4,$15,$FC,$F5,$FC,$FB,$FA,$F3,$D2,$FF
	.dc.b	$44,$8C,$C7,$F0,$78,$FC,$B1,$FC,$A4,$FF,$0F,$E1,$FE,$32,$9F,$E8
	.dc.b	$91,$FC,$80,$FE,$FD,$1F,$D0,$4F,$F5,$B5,$7F,$41,$7F,$D6,$71,$F9
	.dc.b	$D7,$FB,$DD,$7E,$76,$3F,$3D,$3F,$EA,$EB,$FE,$7F,$1F,$EB,$0B,$FD
	.dc.b	$05,$7F,$2C,$7F,$C5,$11,$FE,$C9,$3F,$30,$7F,$3D,$3F,$C4,$E7,$94
	.dc.b	$FF,$67,$47,$F3,$C7,$FC,$48,$7E,$6E,$7F,$C5,$53,$F9,$BA,$FA,$17
	.dc.b	$FE,$9D,$AF,$E7,$27,$F3,$5B,$FF,$15,$9F,$C1,$9F,$F2,$75,$66,$3F
	.dc.b	$81,$9F,$CD,$23,$F0,$27,$E0,$0F,$F0,$FD,$E6,$7F,$3F,$3F,$9A,$3F
	.dc.b	$99,$4F,$F3,$69,$FF,$0E,$BF,$C7,$AB,$FC,$3A,$FF,$A6,$67,$FC,$CF
	.dc.b	$3F,$C7,$A3,$FD,$21,$1F,$90,$8F,$E3,$D1,$F9,$77,$F9,$B3,$FE,$93
	.dc.b	$8F,$E5,$CF,$EA,$6B,$E6,$DF,$E5,$E3,$F9,$71,$F9,$63,$F1,$E7,$E7
	.dc.b	$CF,$F6,$B4,$FF,$86,$5F,$CB,$61,$FE,$8D,$8F,$E6,$98,$27,$F6,$97
	.dc.b	$F9,$77,$FC,$8D,$3F,$97,$53,$B7,$F9,$AE,$1F,$E3,$E9,$FC,$CC,$7F
	.dc.b	$A3,$6B,$F3,$37,$C2,$3F,$C3,$37,$FC,$D8,$FF,$44,$DF,$F3,$6A,$FC
	.dc.b	$C9,$FE,$5C,$9F,$E3,$E7,$FB,$3E,$7F,$32,$7F,$35,$3F,$32,$7F,$A1
	.dc.b	$0F,$CC,$9F,$F0,$01,$FE,$6B,$87,$E6,$70,$FC,$C4,$7E,$66,$3F,$CA
	.dc.b	$11,$FC,$CE,$3F,$9A,$47,$F9,$2A,$FF,$99,$47,$F9,$32,$FF,$37,$5F
	.dc.b	$97,$BF,$F7,$DC,$FE,$5E,$3F,$CF,$93,$F9,$79,$FF,$24,$C7,$F9,$C6
	.dc.b	$7F,$2F,$7F,$E7,$3B,$FE,$61,$1F,$E0,$F8,$FE,$61,$5F,$EA,$CA,$FC
	.dc.b	$B4,$FF,$30,$AF,$C7,$BF,$C9,$8F,$F0,$19,$F7,$CF,$F5,$E4,$FF,$82
	.dc.b	$DF,$F7,$54,$7E,$5A,$BF,$96,$9F,$E0,$A3,$83,$F9,$58,$FF,$75,$1F
	.dc.b	$CB,$30,$FE,$5B,$7F,$EE,$5C,$3F,$C1,$2F,$E5,$0F,$E5,$53,$FE,$08
	.dc.b	$AF,$F3,$1B,$F9,$39,$FF,$03,$D7,$EE,$23,$FC,$0C,$7F,$B7,$5F,$C9
	.dc.b	$61,$F8,$B8,$FF,$76,$61,$FC,$A0,$F6,$61,$E5,$84,$7F,$8E,$5F,$C6
	.dc.b	$1F,$92,$AA,$F3,$8F,$F5,$23,$FE,$07,$BF,$C9,$DF,$F9,$AA,$36,$78
	.dc.b	$4F,$E5,$87,$F7,$97,$F9,$63,$FE,$6B,$7E,$A5,$FF,$B6,$23,$FC,$D7
	.dc.b	$1F,$93,$79,$1F,$EF,$B3,$F9,$61,$FE,$6B,$8F,$E5,$91,$F9,$5A,$FC
	.dc.b	$A3,$C2,$7F,$65,$87,$F1,$97,$C8,$FF,$82,$5F,$EC,$35,$F9,$79,$E1
	.dc.b	$1F,$C0,$CF,$E5,$F3,$F9,$88,$CD,$FF,$88,$27,$FC,$DF,$5F,$99,$7C
	.dc.b	$11,$FE,$D4,$9D,$DF,$F8,$2A,$7F,$36,$FD,$97,$FD,$A0,$7E,$6D,$FC
	.dc.b	$D5,$7F,$99,$E3,$F3,$53,$E4,$8F,$B1,$1F,$C1,$DF,$CD,$BF,$ED,$0B
	.dc.b	$FC,$D9,$FC,$BD,$FF,$69,$C7,$F2,$F3,$EA,$BF,$E0,$63,$F0,$27,$F3
	.dc.b	$1C,$3F,$3B,$1A,$8F,$F7,$6E,$1F,$E6,$A8,$FE,$62,$7F,$84,$6B,$FD
	.dc.b	$61,$3F,$E1,$07,$F9,$8B,$FC,$D1,$FE,$63,$7F,$4F,$04,$7F,$AC,$67
	.dc.b	$E3,$9F,$CC,$CF,$F0,$4D,$FF,$49,$7F,$D9,$97,$FE,$CF,$8F,$D0,$DF
	.dc.b	$F0,$23,$A3,$EA,$9F,$D1,$46,$2F,$FC,$11,$3F,$E6,$AA,$FF,$33,$1F
	.dc.b	$E2,$93,$D3,$5F,$A2,$7D,$2F,$FB,$21,$F8,$52,$7F,$15,$1F,$EC,$E7
	.dc.b	$FC,$14,$FF,$81,$8F,$E5,$C7,$B9,$FF,$64,$D7,$F8,$2E,$7F,$C8,$33
	.dc.b	$F0,$CF,$F6,$31,$FC,$D5,$E1,$1F,$CC,$A3,$F2,$71,$FD,$F5,$FE,$26
	.dc.b	$7E,$76,$FD,$2F,$73,$FC,$12,$FF,$83,$27,$F3,$6F,$E7,$E3,$FC,$D5
	.dc.b	$5E,$B9,$FC,$DC,$7F,$9D,$EF,$F6,$EF,$F9,$A2,$7F,$9B,$3F,$CD,$EF
	.dc.b	$F0,$8F,$F6,$78,$FD,$03,$FE,$69,$8C,$CF,$F7,$98,$FD,$1C,$7F,$B2
	.dc.b	$63,$FC,$D3,$3F,$CE,$67,$F4,$15,$F9,$D8,$FF,$07,$9F,$DC,$DC,$4F
	.dc.b	$F8,$3E,$3E,$4C,$FF,$13,$3F,$BC,$9F,$87,$3F,$38,$7F,$3B,$8F,$F0
	.dc.b	$7E,$1F,$E8,$EA,$FE,$74,$7D,$1B,$FF,$33,$61,$FE,$C3,$AF,$CF,$57
	.dc.b	$F5,$AA,$FC,$C4,$FE,$1A,$BE,$E3,$FE,$65,$C3,$EB,$E1,$FE,$71,$8F
	.dc.b	$C2,$BF,$A3,$8F,$E9,$58,$7F,$70,$8F,$E3,$47,$F3,$DC,$3F,$1B,$1F
	.dc.b	$E8,$08,$FE,$5C,$FF,$99,$6F,$E7,$4F,$F0,$79,$FE,$76,$7F,$84,$30
	.dc.b	$FC,$F6,$1F,$6E,$3F,$C1,$F5,$FC,$FF,$0F,$CF,$9F,$E0,$FA,$FD,$93
	.dc.b	$F9,$E9,$FD,$03,$FE,$C3,$3E,$83,$FE,$C3,$7F,$3A,$FE,$00,$FC,$29
	.dc.b	$FD,$36,$BF,$C1,$CF,$F3,$B3,$FD,$85,$1F,$CE,$E7,$FC,$1B,$1F,$9A
	.dc.b	$9F,$F6,$2E,$1F,$E2,$59,$FF,$07,$4F,$F3,$B7,$FC,$1C,$FF,$99,$9F
	.dc.b	$CC,$D7,$CE,$8F,$F6,$14,$F1,$7F,$45,$1F,$A5,$AF,$E4,$B3,$FE,$68
	.dc.b	$9F,$E7,$57,$FE,$11,$AF,$CD,$DF,$E8,$2B,$FC,$23,$1F,$E6,$6A,$FF
	.dc.b	$33,$D7,$E6,$8F,$E7,$15,$FC,$EA,$BF,$32,$FF,$99,$E7,$F3,$27,$F4
	.dc.b	$FA,$FF,$1B,$BF,$E4,$17,$F0,$6D,$7F,$74,$7F,$3A,$7F,$08,$7E,$60
	.dc.b	$FF,$0A,$61,$FC,$EE,$3F,$88,$C7,$E0,$CE,$91,$F9,$B7,$FD,$77,$7F
	.dc.b	$E7,$03,$F3,$73,$C1,$FF,$0A,$3F,$CE,$0F,$F5,$F5,$7E,$71,$FE,$67
	.dc.b	$5F,$E1,$48,$C5,$FF,$85,$23,$F7,$77,$FC,$D4,$FC,$C4,$FF,$30,$7F
	.dc.b	$C2,$87,$E6,$8F,$CF,$DF,$F9,$B4,$FE,$76,$76,$8F,$CB,$DF,$F9,$9A
	.dc.b	$7F,$2E,$FE,$58,$FE,$5E,$7E,$BC,$FC,$B4,$FF,$84,$23,$FC,$18,$FF
	.dc.b	$33,$8F,$CB,$C6,$1F,$E9,$8F,$F7,$A3,$F9,$A4,$7E,$65,$FC,$51,$FD
	.dc.b	$42,$7F,$9A,$1F,$E1,$23,$FD,$78,$FF,$30,$3F,$9A,$61,$F9,$87,$FC
	.dc.b	$22,$7E,$AA,$BF,$C2,$2F,$E4,$8F,$E6,$2F,$F9,$92,$7F,$33,$87,$FA
	.dc.b	$F2,$3F,$6D,$3F,$E1,$C8,$FC,$B1,$FA,$C7,$FD,$25,$5F,$E0,$F7,$FB
	.dc.b	$2C,$7E,$5A,$BF,$CC,$B1,$F9,$67,$FC,$1B,$3F,$EB,$C8,$FF,$06,$C7
	.dc.b	$F6,$57,$F2,$95,$FE,$8D,$8F,$F5,$F5,$FF,$83,$9F,$E5,$B3,$FE,$0E
	.dc.b	$37,$3F,$EB,$C3,$F2,$2F,$FA,$F0,$FB,$47,$F8,$D2,$7F,$D3,$15,$FE
	.dc.b	$0F,$AF,$EB,$93,$FE,$69,$3F,$86,$BF,$CC,$1F,$CB,$C7,$F9,$A4,$FE
	.dc.b	$60,$FF,$AF,$70,$FE,$60,$FF,$2B,$BF,$F5,$E1,$E0,$7F,$03,$3F,$AE
	.dc.b	$BF,$ED,$D8,$7E,$5F,$0F,$EB,$77,$FE,$6A,$BF,$CA,$5F,$F9,$A8,$FA
	.dc.b	$2F,$F2,$E3,$FD,$7B,$0C,$3F,$97,$47,$FA,$CF,$06,$1F,$E0,$DC,$3F
	.dc.b	$97,$D7,$F9,$BA,$BF,$97,$CF,$F8,$36,$3F,$D1,$F1,$F1,$6F,$F3,$47
	.dc.b	$A5,$FF,$35,$3E,$93,$F8,$2D,$FE,$6D,$FE,$5F,$7F,$EB,$B9,$FC,$D6
	.dc.b	$1F,$CB,$A3,$F2,$F1,$FC,$B9,$FC,$BB,$FD,$6D,$FF,$34,$3F,$B5,$7F
	.dc.b	$96,$BF,$97,$7F,$96,$9F,$E6,$73,$F9,$6C,$7F,$81,$E3,$F9,$5E,$1F
	.dc.b	$97,$AF,$F3,$24,$7F,$8D,$8F,$F3,$1B,$FC,$A9,$FC,$AD,$FF,$06,$8F
	.dc.b	$F0,$1C,$7F,$A0,$67,$FC,$C4,$FF,$80,$AB,$F2,$D7,$FC,$AE,$FE,$7C
	.dc.b	$FF,$96,$EF,$F2,$CF,$E6,$0F,$F4,$AC,$FF,$81,$6B,$FD,$7A,$FF,$81
	.dc.b	$2B,$E7,$CF,$F8,$26,$3F,$22,$FF,$2A,$7F,$27,$3F,$95,$8F,$CC,$1F
	.dc.b	$CA,$A3,$FD,$BA,$FF,$2A,$3F,$CC,$4F,$F8,$16,$3F,$2A,$FF,$81,$EB
	.dc.b	$F9,$51,$FE,$44,$8F,$F0,$21,$FC,$58,$FC,$A1,$F7,$0F,$E2,$73,$F0
	.dc.b	$CF,$E5,$47,$E5,$30,$FC,$AC,$FF,$97,$63,$F9,$53,$F9,$3A,$FD,$D4
	.dc.b	$7F,$20,$C3,$F2,$4F,$F7,$EB,$D3,$6B,$F9,$63,$BC,$3F,$D7,$F1,$FE
	.dc.b	$5F,$3E,$6C,$7F,$AF,$6F,$F8,$D5,$7E,$92,$3F,$41,$3F,$94,$3D,$2F
	.dc.b	$FA,$92,$7F,$28,$FF,$74,$7F,$D7,$67,$E5,$1F,$A4,$7F,$29,$3F,$D7
	.dc.b	$67,$F8,$06,$3F,$C1,$55,$FC,$E2,$3F,$C2,$F3,$F4,$9F,$F5,$FD,$7F
	.dc.b	$97,$0F,$F3,$83,$FC,$B2,$3F,$95,$BF,$E5,$28,$FF,$2E,$3F,$EB,$CA
	.dc.b	$FE,$89,$3F,$E5,$C8,$FF,$03,$1F,$CB,$C1,$F9,$A7,$FC,$04,$FF,$97
	.dc.b	$5F,$E1,$31,$FE,$03,$7F,$D7,$31,$F9,$23,$F3,$51,$FE,$05,$BE,$B1
	.dc.b	$F9,$FB,$FF,$3B,$47,$F2,$08,$FC,$D1,$FE,$06,$3F,$DB,$73,$FE,$06
	.dc.b	$7F,$AD,$47,$27,$FD,$F1,$1E,$C6,$BE,$C7,$F8,$01,$F9,$A3,$F9,$A5
	.dc.b	$FF,$B0,$23,$F9,$BC,$FF,$98,$2F,$E6,$57,$F9,$9E,$FF,$C0,$D7,$FC
	.dc.b	$C6,$FF,$C0,$D7,$FE,$60,$BF,$F1,$7C,$7F,$81,$AF,$F9,$95,$7F,$97
	.dc.b	$CF,$F5,$DB,$FC,$D6,$3F,$22,$FF,$B8,$9F,$F0,$13,$FE,$06,$3F,$A3
	.dc.b	$57,$F9,$75,$FF,$7D,$3F,$EB,$A7,$FD,$0B,$1F,$91,$7E,$AD,$FF,$30
	.dc.b	$3F,$C1,$35,$DA,$3F,$CC,$B1,$FE,$BC,$7F,$CD,$2F,$E7,$A7,$FC,$BD
	.dc.b	$1F,$CC,$5F,$E6,$0F,$E7,$1F,$F2,$F5,$7C,$F8,$EF,$1F,$9E,$7F,$0A
	.dc.b	$43,$FC,$13,$7F,$E0,$B7,$FC,$12,$7E,$2D,$FD,$03,$FD,$B9,$FE,$66
	.dc.b	$FF,$8D,$1F,$E6,$6F,$C2,$3F,$C1,$27,$95,$FE,$66,$FF,$82,$6B,$F3
	.dc.b	$93,$DA,$3F,$B6,$BF,$A7,$7F,$CE,$0F,$F9,$7C,$FC,$23,$F9,$D3,$D8
	.dc.b	$7E,$7A,$BF,$C1,$87,$D1,$7D,$AE,$5F,$D0,$1F,$EF,$28,$FE,$72,$FC
	.dc.b	$18,$FE,$72,$FF,$85,$E7,$FC,$C6,$FF,$AF,$23,$FC,$C6,$FE,$7A,$BF
	.dc.b	$C1,$A7,$F3,$7A,$FF,$63,$9F,$E6,$28,$FC,$FD,$7F,$38,$7E,$41,$F9
	.dc.b	$B3,$F9,$C1,$F9,$DB,$FE,$70,$FF,$90,$0F,$E7,$07,$FA,$1C,$FE,$70
	.dc.b	$7F,$97,$DF,$CE,$57,$FA,$F6,$FF,$39,$3F,$E8,$6B,$FC,$E3,$FE,$0A
	.dc.b	$7F,$0A,$FF,$AD,$EB,$F3,$AE,$63,$F2,$0F,$C3,$3F,$9C,$D7,$F9,$86
	.dc.b	$7F,$C1,$93,$FC,$E6,$3F,$CC,$8F,$09,$F9,$F7,$FE,$63,$7F,$D7,$07
	.dc.b	$F3,$79,$FC,$73,$FE,$0F,$B4,$FF,$AE,$70,$F8,$D3,$F9,$93,$EE,$C7
	.dc.b	$F9,$87,$0E,$D1,$FE,$BA,$7F,$54,$FF,$AF,$2B,$78,$7E,$76,$3B,$9F
	.dc.b	$E6,$33,$F8,$E1,$FC,$EC,$FF,$34,$47,$E2,$9F,$4E,$1F,$E6,$57,$F9
	.dc.b	$E3,$FE,$BC,$7F,$D1,$4F,$F9,$BE,$FF,$2D,$1F,$B6,$8F,$E2,$2F,$E6
	.dc.b	$0E,$97,$FC,$F2,$7F,$3E,$FF,$84,$63,$FD,$7B,$1F,$E4,$27,$F3,$87
	.dc.b	$FA,$48,$FE,$70,$7F,$99,$0F,$CC,$9F,$C4,$67,$FC,$CA,$7F,$38,$8F
	.dc.b	$F5,$D4,$FF,$84,$27,$F3,$11,$FE,$10,$7F,$D6,$87,$BE,$FC,$95,$F9
	.dc.b	$93,$FC,$CA,$7F,$83,$CF,$93,$1F,$EB,$B8,$FF,$33,$47,$F8,$42,$3F
	.dc.b	$CC,$F5,$FC,$CE,$3F,$CC,$77,$FC,$CE,$7F,$30,$7E,$6E,$FF,$99,$E1
	.dc.b	$F9,$9C,$3F,$99,$CF,$F9,$8E,$FF,$99,$C7,$F8,$3C,$FF,$31,$E1,$FC
	.dc.b	$CD,$C6,$1F,$E6,$1A,$FD,$D4,$7F,$32,$7F,$C1,$D5,$F9,$78,$FC,$B4
	.dc.b	$7F,$13,$7F,$12,$FF,$2F,$7F,$13,$87,$F3,$18,$FC,$B9,$FE,$0B,$BC
	.dc.b	$DF,$E5,$8F,$F3,$05,$FC,$49,$FF,$2D,$5F,$FA,$E8,$FF,$07,$BF,$E6
	.dc.b	$07,$F5,$EF,$E5,$DD,$4F,$E5,$9F,$F4,$44,$70,$3E,$7D,$FF,$AE,$6F
	.dc.b	$FC,$C5,$1F,$96,$8F,$F5,$E4,$FE,$82,$3F,$A4,$9F,$99,$3F,$8C,$9F
	.dc.b	$E8,$BB,$FC,$B5,$E6,$BF,$A7,$47,$E5,$A7,$F9,$79,$F9,$6A,$FE,$A7
	.dc.b	$1F,$E6,$67,$FD,$79,$1F,$E6,$69,$FC,$B3,$FE,$3D,$3F,$51,$5F,$CC
	.dc.b	$1F,$E7,$47,$F2,$F7,$2F,$F8,$18,$FE,$72,$7F,$82,$4E,$31,$E0,$7F
	.dc.b	$C1,$F5,$F3,$67,$E8,$47,$FA,$F8,$FE,$D3,$5F,$CB,$8F,$90,$FF,$9A
	.dc.b	$CF,$DE,$9F,$C3,$9F,$F0,$81,$FC,$CC,$FF,$07,$3F,$E7,$17,$F3,$6F
	.dc.b	$F3,$33,$F3,$78,$7E,$72,$3F,$C1,$B1,$FE,$C4,$7F,$12,$FF,$87,$CF
	.dc.b	$CE,$1F,$E6,$A3,$FD,$7E,$FE,$72,$BF,$CD,$53,$F2,$27,$FD,$5D,$87
	.dc.b	$F9,$AA,$BF,$0B,$82,$3F,$CD,$71,$FC,$32,$7F,$92,$C7,$F1,$B7,$FD
	.dc.b	$6A,$7F,$34,$3F,$B7,$3F,$CD,$2F,$F1,$38,$7F,$5F,$7F,$9A,$E1,$FC
	.dc.b	$07,$0F,$CF,$5F,$F8,$2A,$FF,$95,$C7,$F8,$45,$FC,$FB,$FC,$CE,$BF
	.dc.b	$3D,$87,$FA,$DD,$FC,$AB,$FE,$BE,$9E,$2F,$FA,$F4,$F4,$4F,$FA,$F6
	.dc.b	$FF,$98,$5F,$F3,$18,$FC,$E4,$AF,$FC,$D7,$3F,$9D,$7F,$98,$3F,$EE
	.dc.b	$C7,$F9,$83,$F9,$B3,$F6,$13,$F1,$5F,$F1,$4E,$1F,$80,$7F,$A6,$C7
	.dc.b	$E3,$9F,$E6,$47,$F8,$38,$FF,$36,$1F,$EC,$03,$FC,$23,$1F,$EC,$19
	.dc.b	$FF,$35,$47,$F3,$19,$F3,$3F,$C8,$A7,$FC,$18,$FE,$74,$FE,$66,$7F
	.dc.b	$A7,$EB,$FC,$16,$FF,$62,$3C,$87,$F6,$4C,$3F,$9A,$3F,$E0,$D9,$FF
	.dc.b	$05,$BF,$E6,$69,$FC,$DB,$FC,$DA,$BF,$3D,$5F,$EC,$28,$FE,$67,$3F
	.dc.b	$9A,$7F,$9A,$3F,$C1,$07,$2F,$F3,$53,$FD,$87,$8E,$27,$FC,$1D,$35
	.dc.b	$FF,$07,$1F,$8C,$9F,$E2,$2F,$FA,$22,$13,$B8,$FF,$27,$4E,$AF,$FC
	.dc.b	$92,$FF,$B2,$67,$53,$FE,$4C,$9F,$F2,$ED,$AF,$33,$FC,$62,$7C,$0F
	.dc.b	$FB,$59,$D5,$6A,$3F,$D3,$35,$C6,$B9,$5F,$F2,$68,$FF,$53,$CE,$EF
	.dc.b	$F5,$55,$FA,$BB,$C5,$78,$4E,$35,$FE,$04,$AD,$E1,$FE,$A6,$C3,$A6
	.dc.b	$1E,$27,$FD,$4A,$7E,$B2,$BD,$E7,$EC,$30,$F0,$57,$9A,$AC,$FB,$0F
	.dc.b	$F6,$D1,$CC,$FE,$33,$5F,$B6,$BE,$71,$FD,$3E,$F9,$C7,$AA,$75,$7E
	.dc.b	$5A,$FF,$46,$CF,$37,$F1,$17,$EC,$7F,$42,$FC,$4A,$FF,$6F,$DF,$C5
	.dc.b	$8F,$84,$53,$CD,$3F,$D5,$DF,$BF,$3F,$E8,$18,$F0,$1F,$E1,$F7,$FD
	.dc.b	$09,$87,$78,$FD,$21,$F5,$CF,$F2,$0C,$7F,$A2,$67,$51,$F8,$E8,$FF
	.dc.b	$14,$3F,$E4,$07,$FC,$55,$3A,$8F,$F4,$95,$F9,$1F,$F7,$E4,$FF,$8B
	.dc.b	$EB,$51,$F3,$CF,$29,$FE,$E3,$3A,$1F,$E1,$8A,$F7,$CF,$FB,$A8,$FE
	.dc.b	$F0,$FE,$4F,$0D,$CF,$FB,$CA,$BF,$1D,$5F,$C0,$8F,$F7,$65,$FB,$6B
	.dc.b	$EF,$47,$FB,$A8,$FE,$AD,$5F,$EB,$97,$F0,$EF,$F0,$79,$FB,$AF,$F5
	.dc.b	$F7,$91,$F7,$5F,$F4,$61,$89,$F1,$BF,$EE,$DA,$F7,$BF,$A1,$3D,$E7
	.dc.b	$FB,$98,$D9,$FE,$E8,$BD,$C7,$B6,$3E,$A3,$EF,$AA,$FB,$A3,$FC,$34
	.dc.b	$53,$F8,$19,$D8,$FF,$06,$9F,$AB,$BF,$E0,$AF,$A4,$FE,$13,$7B,$7F
	.dc.b	$CF,$67,$F0,$9C,$3F,$08,$FF,$9F,$AB,$EC,$DF,$4B,$FE,$10,$7F,$0B
	.dc.b	$C3,$35,$FE,$00,$8F,$ED,$27,$F9,$10,$FD,$AB,$DE,$7F,$7F,$3F,$98
	.dc.b	$8F,$F7,$CD,$7F,$00,$3F,$14,$FF,$12,$7F,$50,$FF,$37,$7F,$6E,$F3
	.dc.b	$C3,$F8,$7E,$1F,$BA,$AF,$3C,$7F,$B2,$70,$FE,$D6,$7F,$85,$70,$F9
	.dc.b	$17,$FE,$F6,$39,$3F,$1D,$FD,$BD,$72,$8F,$1C,$7F,$92,$DF,$1B,$FE
	.dc.b	$F1,$C3,$F8,$F1,$FB,$57,$DC,$FF,$19,$C3,$F7,$15,$F2,$A3,$EF,$5F
	.dc.b	$F7,$47,$9D,$7F,$9E,$23,$F7,$57,$F4,$A3,$EC,$1F,$B8,$8F,$F7,$F6
	.dc.b	$1F,$8A,$8F,$F1,$45,$7D,$8A,$FC,$CC,$7F,$9E,$6B,$F6,$F3,$9A,$FC
	.dc.b	$04,$FF,$72,$AE,$13,$F9,$8B,$FE,$25,$3F,$EF,$FB,$FF,$26,$1F,$B3
	.dc.b	$7F,$C9,$8F,$DD,$7F,$85,$9D,$2F,$FA,$54,$7E,$BA,$7F,$47,$1F,$4E
	.dc.b	$7F,$88,$BE,$97,$FD,$FD,$5A,$BF,$4C,$7F,$73,$7F,$63,$59,$AE,$D1
	.dc.b	$FB,$99,$FE,$6A,$7F,$6E,$AE,$2F,$F7,$3B,$FF,$7F,$1F,$E2,$48,$FB
	.dc.b	$71,$FE,$FE,$8C,$47,$EE,$23,$FB,$53,$FD,$55,$1D,$26,$BF,$87,$3F
	.dc.b	$B9,$DF,$E8,$9F,$C3,$C6,$A7,$F3,$D1,$F7,$EB,$FB,$93,$FA,$33,$FC
	.dc.b	$8F,$87,$7A,$FB,$F1,$FD,$CC,$FD,$61,$FD,$BE,$B9,$BF,$EF,$4B,$FC
	.dc.b	$8D,$FE,$BA,$3F,$55,$1A,$7F,$CB,$33,$D1,$FD,$66,$1F,$B3,$9F,$EE
	.dc.b	$D8,$7F,$27,$9F,$D2,$9F,$EF,$F3,$E0,$4F,$EB,$AF,$FD,$6C,$7F,$6D
	.dc.b	$7F,$65,$1F,$9A,$3E,$3B,$FC,$26,$FF,$B2,$C7,$EB,$AF,$C8,$FE,$45
	.dc.b	$FD,$74,$FE,$CA,$73,$3F,$E2,$6A,$FD,$33,$FA,$C8,$FA,$0F,$FA,$DE
	.dc.b	$BF,$59,$1F,$40,$F5,$CF,$F6,$39,$FD,$51,$F8,$F3,$EC,$CF,$DF,$7F
	.dc.b	$06,$7A,$30,$FD,$85,$7E,$0D,$FC,$5C,$7F,$88,$6B,$F5,$67,$FA,$D9
	.dc.b	$FD,$64,$7E,$AA,$7F,$B6,$57,$EA,$A7,$FA,$B1,$E1,$7F,$AE,$BF,$E5
	.dc.b	$BA,$FE,$BB,$3F,$9B,$8F,$EB,$8F,$EA,$1F,$0D,$FE,$B1,$FE,$F3,$59
	.dc.b	$8F,$D8,$47,$13,$F9,$35,$7F,$1F,$7F,$4E,$7C,$18,$FF,$4A,$D7,$F5
	.dc.b	$98,$FE,$B5,$7F,$D7,$A3,$FA,$99,$F0,$0F,$F2,$14,$7C,$97,$D2,$FF
	.dc.b	$98,$67,$F8,$5C,$7F,$57,$AF,$D5,$BF,$28,$FD,$54,$FE,$A4,$FC,$51
	.dc.b	$FE,$5C,$8F,$EA,$B3,$FA,$38,$FC,$51,$FB,$0A,$FC,$9B,$FE,$51,$3F
	.dc.b	$50,$FE,$09,$FD,$33,$F9,$D7,$FA,$DB,$FB,$27,$FA,$D6,$1F,$52,$F9
	.dc.b	$C7,$C2,$3F,$44,$F3,$7F,$A8,$61,$FD,$80,$FE,$6E,$FF,$94,$5F,$C6
	.dc.b	$BF,$E2,$A8,$FE,$9E,$FE,$98,$E4,$7F,$B8,$A3,$37,$FA,$69,$FC,$EC
	.dc.b	$A3,$F4,$D8,$7F,$93,$63,$F4,$D7,$FE,$4C,$BF,$D2,$C7,$F9,$3A,$F8
	.dc.b	$9F,$C0,$1F,$F3,$23,$F8,$F7,$F4,$93,$FD,$42,$3F,$C9,$17,$FA,$97
	.dc.b	$C3,$3F,$C7,$AF,$FD,$20,$FF,$9B,$4F,$AA,$7F,$1C,$8F,$B3,$3F,$55
	.dc.b	$E3,$87,$CA,$3F,$C0,$93,$FE,$83,$C3,$F7,$31,$FD,$E6,$FF,$D2,$17
	.dc.b	$F6,$2F,$E0,$D7,$A6,$FF,$D8,$2F,$A5,$FE,$2E,$F2,$3F,$1E,$FF,$93
	.dc.b	$9F,$A6,$F0,$8F,$C3,$CF,$AE,$3F,$D0,$CF,$F3,$67,$FC,$09,$87,$F8
	.dc.b	$16,$7F,$CF,$F3,$F4,$E3,$F4,$11,$FC,$5A,$7E,$84,$7D,$18,$FF,$3E
	.dc.b	$CF,$F1,$58,$FF,$19,$C7,$F1,$09,$FE,$29,$5F,$02,$3C,$F1,$FE,$AB
	.dc.b	$7C,$EF,$F9,$DE,$7F,$89,$BF,$C5,$30,$FE,$FD,$5F,$C4,$EF,$FB,$AC
	.dc.b	$FF,$13,$C3,$FC,$E9,$87,$F7,$D8,$CD,$FE,$96,$7F,$89,$C7,$FB,$BA
	.dc.b	$B2,$7F,$1F,$7F,$1A,$7F,$9B,$6B,$F8,$DD,$FE,$F2,$FF,$48,$FF,$80
	.dc.b	$63,$FB,$2B,$FB,$B9,$E1,$1F,$C4,$DF,$E1,$E7,$F0,$B3,$F6,$07,$E6
	.dc.b	$CF,$ED,$B5,$E9,$7F,$4D,$1F,$E6,$A7,$F8,$9D,$7F,$7C,$8F,$E2,$57
	.dc.b	$FE,$35,$AF,$E2,$D3,$FB,$6A,$FF,$35,$1F,$E6,$23,$F8,$95,$78,$0F
	.dc.b	$9A,$FF,$8C,$DF,$E4,$4D,$7F,$9F,$3F,$30,$F9,$07,$F9,$AA,$FE,$83
	.dc.b	$FE,$2F,$AF,$90,$F8,$4F,$F3,$29,$F4,$6B,$FC,$5D,$3F,$D1,$67,$E7
	.dc.b	$C7,$F4,$1B,$FE,$77,$5F,$E2,$EA,$F9,$D1,$FE,$2A,$8F,$E8,$07,$F8
	.dc.b	$A8,$F9,$B7,$FD,$C5,$F9,$B5,$FD,$CE,$BF,$9F,$47,$F7,$5B,$FA,$15
	.dc.b	$FE,$27,$8F,$97,$7F,$E5,$DA,$FF,$13,$47,$F3,$C8,$FF,$30,$4F,$F3
	.dc.b	$C8,$FE,$6E,$FC,$27,$F4,$C7,$F8,$82,$7F,$A0,$D6,$8F,$9F,$3F,$2A
	.dc.b	$7F,$31,$5F,$EC,$C8,$F9,$11,$F2,$6F,$FA,$04,$FF,$B2,$23,$E4,$3F
	.dc.b	$96,$7F,$DE,$8F,$FA,$6A,$3E,$BC,$7F,$B2,$A7,$F2,$4F,$C9,$7C,$CF
	.dc.b	$F8,$7D,$FC,$7D,$FE,$02,$3F,$BD,$C7,$27,$FD,$F8,$FE,$5D,$F9,$93
	.dc.b	$F2,$E3,$E5,$47,$CB,$8F,$F1,$04,$7C,$B8,$FC,$B4,$7E,$56,$BF,$29
	.dc.b	$1F,$2C,$FF,$27,$9F,$20,$FC,$AB,$FE,$4D,$7F,$2B,$5F,$29,$FF,$5C
	.dc.b	$DF,$CA,$BF,$E0,$35,$FE,$7E,$3E,$51,$FE,$7C,$9F,$94,$FF,$00,$3E
	.dc.b	$51,$FE,$F2,$BF,$C8,$CF,$F9,$EE,$7F,$C2,$33,$FE,$EF,$8F,$EB,$D1
	.dc.b	$F2,$4F,$F0,$83,$F2,$0F,$DE,$9F,$1D,$5F,$F3,$37,$F0,$77,$FC,$A6
	.dc.b	$BF,$DD,$67,$F9,$EE,$BE,$4B,$FE,$EA,$8F,$C9,$C7,$EF,$67,$E4,$CF
	.dc.b	$E4,$E3,$F7,$B1,$FE,$ED,$9F,$DE,$C7,$E4,$E3,$F7,$97,$FE,$DE,$BF
	.dc.b	$91,$7F,$E7,$7B,$F9,$0F,$F8,$EC,$F9,$15,$FB,$9B,$FC,$8C,$7F,$A9
	.dc.b	$2F,$F2,$33,$F5,$4F,$EA,$A5,$7F,$16,$FF,$49,$8F,$F1,$D1,$F9,$3A
	.dc.b	$FE,$51,$3F,$C4,$6F,$FA,$A5,$7E,$38,$F9,$6F,$F2,$79,$FC,$B1,$F3
	.dc.b	$AF,$F2,$17,$F2,$2B,$FC,$6F,$1F,$21,$FF,$50,$DF,$E4,$23,$E4,$4F
	.dc.b	$F8,$DE,$3F,$21,$3F,$1C,$FC,$81,$F8,$FA,$FE,$76,$FE,$46,$BC,$AF
	.dc.b	$FA,$7C,$FF,$51,$5F,$CA,$7F,$D4,$93,$F2,$9F,$88,$7D,$D2,$9F,$10
	.dc.b	$FD,$75,$FE,$4C,$F9,$87,$E4,$DE,$37,$FE,$64,$7F,$C5,$F1,$FB,$D7
	.dc.b	$FB,$89,$F9,$43,$77,$FB,$C7,$F2,$B3,$FD,$42,$3F,$C6,$95,$F2,$A3
	.dc.b	$FC,$6D,$1F,$2A,$7F,$D9,$B1,$FB,$9A,$FC,$84,$7E,$E5,$F5,$D7,$F8
	.dc.b	$06,$3E,$34,$FF,$9A,$0F,$63,$FE,$7D,$BF,$E4,$B5,$FD,$3E,$7F,$C0
	.dc.b	$2F,$EE,$67,$F2,$6F,$FB,$21,$F9,$27,$EE,$A3,$FC,$03,$7F,$8E,$8F
	.dc.b	$E5,$31,$FC,$AA,$7F,$1D,$1F,$E2,$A3,$FC,$25,$1F,$C5,$A3,$FC,$BF
	.dc.b	$1F,$E6,$49,$FF,$15,$83,$FD,$84,$FF,$97,$AB,$FC,$56,$FF,$9E,$27
	.dc.b	$FC,$56,$FF,$B1,$E7,$FC,$56,$7F,$8C,$E7,$F0,$75,$FB,$C8,$FF,$29
	.dc.b	$B6,$3F,$CB,$91,$FE,$D6,$7F,$CB,$73,$F8,$D3,$FC,$79,$7F,$85,$7D
	.dc.b	$AF,$AD,$FF,$62,$47,$F9,$71,$5F,$EB,$67,$FD,$CF,$5F,$E5,$E7,$F2
	.dc.b	$F3,$F0,$9F,$ED,$73,$BB,$F2,$C7,$B6,$7F,$C7,$31,$FE,$EA,$8F,$F3
	.dc.b	$CC,$FF,$A3,$0F,$F6,$1C,$F8,$0F,$F3,$0D,$7B,$8F,$F1,$8B,$FE,$89
	.dc.b	$7F,$A3,$3F,$ED,$37,$FC,$C2,$FE,$64,$FC,$29,$F5,$4F,$D1,$CF,$F8
	.dc.b	$1C,$FF,$01,$C7,$FB,$0E,$3F,$C6,$27,$F0,$6B,$FF,$01,$3F,$9F,$3F
	.dc.b	$DF,$C7,$F8,$BC,$FE,$65,$7F,$9F,$BF,$F5,$FD,$FE,$7D,$FF,$60,$DF
	.dc.b	$E7,$EB,$E9,$1F,$E2,$DB,$FF,$00,$CF,$F2,$FB,$FE,$10,$FF,$AF,$67
	.dc.b	$F9,$79,$FC,$7E,$FE,$53,$FA,$24,$FF,$0A,$3F,$9E,$CF,$F8,$B6,$FF
	.dc.b	$84,$C7,$E4,$5F,$86,$7C,$A7,$ED,$BF,$E3,$DA,$FE,$69,$3F,$A2,$7F
	.dc.b	$82,$57,$F9,$06,$BF,$D7,$07,$F0,$D7,$0F,$F8,$AA,$D8,$7D,$99,$CD
	.dc.b	$FD,$16,$D7,$A1,$F4,$DD,$7F,$57,$1F,$8E,$AE,$51,$F8,$29,$D4,$FF
	.dc.b	$B1,$0E,$27,$FA,$BE,$38,$D7,$F6,$0A,$E3,$3F,$EC,$AB,$FD,$95,$F2
	.dc.b	$8F,$5D,$7C,$17,$91,$FE,$D0,$C3,$D7,$3B,$8F,$F0,$8C,$74,$7F,$DB
	.dc.b	$55,$FE,$EA,$70,$7F,$BC,$9E,$2F,$43,$FD,$A3,$1A,$AF,$DD,$61,$EA
	.dc.b	$3E,$3D,$7F,$B4,$9F,$9B,$5F,$E7,$1A,$F4,$BC,$C5,$7D,$2B,$F3,$3F
	.dc.b	$E7,$7B,$F0,$DF,$D2,$3F,$0E,$F2,$71,$5F,$05,$B7,$C1,$FC,$5B,$B7
	.dc.b	$F7,$D5,$F9,$09,$F5,$47,$8A,$B8,$46,$67,$FD,$C4,$7F,$17,$BF,$2D
	.dc.b	$7A,$0F,$80,$74,$BE,$35,$D5,$FF,$71,$47,$E2,$9F,$9D,$3E,$E7,$F9
	.dc.b	$65,$F0,$AF,$87,$18,$3E,$71,$FE,$FE,$AE,$4F,$F3,$3C,$3E,$BC,$FF
	.dc.b	$3D,$3E,$61,$FC,$F2,$FF,$00,$FE,$66,$BF,$64,$F0,$3F,$D4,$4F,$D6
	.dc.b	$7F,$DA,$CD,$C3,$E3,$4F,$29,$FE,$D6,$76,$3F,$92,$4E,$CF,$86,$FF
	.dc.b	$34,$9E,$C7,$EE,$67,$C7,$87,$C2,$9F,$3B,$FE,$92,$6E,$1F,$3D,$B7
	.dc.b	$D1,$FF,$35,$BF,$88,$AE,$73,$D6,$7E,$0C,$FE,$F4,$FF,$23,$BF,$EA
	.dc.b	$C7,$F2,$67,$E2,$A3,$E4,$BF,$BB,$8A,$7E,$02,$7E,$4C,$7C,$43,$F8
	.dc.b	$AC,$78,$CF,$E0,$57,$D6,$3F,$8C,$DF,$E3,$E7,$FD,$FD,$1F,$87,$7B
	.dc.b	$CF,$57,$CE,$FA,$63,$F4,$F7,$F1,$5A,$F5,$9F,$F6,$55,$75,$7F,$1D
	.dc.b	$1F,$A1,$9F,$F5,$3B,$FE,$E5,$78,$BB,$8F,$F6,$95,$FE,$B9,$F8,$C7
	.dc.b	$E7,$CF,$D6,$E1,$FB,$D7,$E3,$46,$DE,$C6,$CF,$C9,$3F,$85,$C3,$D5
	.dc.b	$7F,$C6,$A7,$E3,$47,$F0,$EA,$E8,$FD,$D9,$FE,$56,$F4,$9F,$E3,$77
	.dc.b	$FA,$37,$FC,$0C,$FA,$63,$FC,$11,$1D,$30,$FD,$4D,$FF,$13,$7F,$51
	.dc.b	$87,$F0,$A7,$F5,$18,$7F,$87,$9F,$E2,$8F,$E1,$30,$E8,$FC,$2B,$FE
	.dc.b	$A7,$87,$F1,$1C,$3F,$82,$C7,$9E,$7F,$C0,$CF,$9C,$FB,$95,$E7,$8F
	.dc.b	$DC,$BF,$C3,$2B,$F5,$33,$FC,$7D,$FE,$17,$7F,$BF,$6C,$EE,$BE,$39
	.dc.b	$FE,$0C,$AF,$DD,$C7,$F0,$A8,$FF,$69,$3D,$1F,$E0,$75,$CE,$3E,$0C
	.dc.b	$F7,$7E,$0B,$FC,$6D,$F8,$2F,$F9,$BB,$0F,$DE,$3F,$8D,$3F,$CD,$A7
	.dc.b	$EF,$4F,$71,$FC,$9E,$BF,$60,$E6,$7F,$18,$FF,$6C,$3F,$7E,$6A,$3F
	.dc.b	$85,$56,$A3,$F4,$55,$F8,$79,$D1,$FD,$96,$3F,$80,$9F,$AD,$3E,$FD
	.dc.b	$FF,$1E,$AF,$6C,$FF,$4C,$7D,$B3,$FC,$12,$75,$7F,$92,$8F,$D8,$9F
	.dc.b	$1E,$BF,$81,$BE,$03,$35,$EA,$8F,$E0,$57,$FC,$02,$7F,$6D,$1F,$C2
	.dc.b	$1E,$87,$C3,$BE,$8F,$EF,$9F,$94,$FF,$9E,$EF,$F7,$F3,$EB,$3F,$A4
	.dc.b	$4F,$F0,$28,$F9,$27,$EF,$E3,$F7,$2F,$29,$FF,$10,$5E,$A3,$E9,$9F
	.dc.b	$BD,$9D,$BF,$36,$7C,$A7,$F6,$C3,$F3,$87,$F8,$77,$0F,$DE,$3F,$9B
	.dc.b	$3F,$75,$3F,$B6,$7F,$C3,$18,$7E,$EE,$BF,$61,$1F,$E6,$43,$F3,$47
	.dc.b	$F6,$7A,$FD,$DD,$EB,$0F,$D2,$1F,$E5,$D8,$FA,$75,$FB,$99,$F7,$57
	.dc.b	$F6,$D7,$F7,$37,$FB,$79,$F2,$4F,$F8,$51,$E7,$3F,$98,$8F,$C9,$1E
	.dc.b	$EB,$FF,$04,$3F,$AD,$C3,$F8,$CD,$FF,$76,$8E,$67,$F3,$18,$FD,$6B
	.dc.b	$FC,$60,$FE,$63,$7F,$C0,$AB,$ED,$C7,$F0,$29,$F7,$57,$F8,$02,$79
	.dc.b	$9F,$E0,$48,$FD,$BB,$CD,$FE,$5F,$5F,$DE,$0E,$75,$FC,$12,$FF,$6D
	.dc.b	$1C,$8F,$CC,$C7,$FA,$07,$0F,$EE,$8F,$03,$FB,$D1,$FB,$39,$FB,$D5
	.dc.b	$FA,$58,$FE,$F2,$FF,$86,$DF,$E7,$A7,$F7,$5B,$E4,$FF,$78,$8F,$D9
	.dc.b	$DF,$F7,$47,$94,$FE,$79,$FD,$91,$FA,$D8,$F1,$57,$F7,$58,$C3,$FC
	.dc.b	$00,$F4,$D7,$F9,$0A,$BE,$FD,$7F,$6A,$8F,$D0,$1F,$9B,$AF,$E9,$6F
	.dc.b	$F7,$63,$F8,$6B,$EA,$7F,$6F,$1F,$B3,$9F,$85,$3E,$93,$FB,$6C,$7E
	.dc.b	$B2,$7F,$5B,$3F,$AB,$8F,$F1,$19,$F8,$E8,$FE,$CA,$7E,$FC,$FE,$CF
	.dc.b	$5F,$AB,$AF,$E1,$D3,$FA,$B9,$F9,$4F,$F2,$F7,$E6,$1F,$E0,$13,$FB
	.dc.b	$01,$FA,$73,$FC,$ED,$87,$F6,$5A,$FD,$3C,$7E,$88,$FD,$59,$F4,$E7
	.dc.b	$F4,$CF,$EB,$EF,$8B,$FA,$C7,$F1,$AF,$F2,$63,$11,$FD,$8E,$7C,$91
	.dc.b	$FA,$53,$FB,$14,$72,$BF,$F3,$8B,$F9,$69,$FC,$6D,$7F,$6A,$C3,$FC
	.dc.b	$3B,$3F,$D8,$2F,$F2,$B3,$FD,$7F,$0F,$D4,$5E,$70,$FD,$3D,$FF,$5D
	.dc.b	$AF,$D8,$47,$F0,$33,$FC,$E8,$F8,$5F,$F3,$6C,$FE,$AD,$FD,$2C,$FF
	.dc.b	$7B,$3D,$37,$FD,$98,$FF,$02,$BF,$C0,$9F,$EB,$0F,$E9,$4F,$E1,$67
	.dc.b	$F0,$17,$F3,$B1,$FE,$B2,$7D,$27,$F0,$03,$F8,$A9,$F2,$8F,$A4,$7E
	.dc.b	$D5,$FC,$E9,$FE,$15,$7F,$8C,$E1,$FC,$6A,$7F,$C2,$B3,$F5,$5F,$A0
	.dc.b	$7F,$0F,$AF,$C6,$57,$FA,$96,$BF,$C2,$13,$F3,$CF,$E2,$E7,$78,$FC
	.dc.b	$DB,$FE,$0A,$7F,$88,$CF,$F8,$06,$BF,$46,$7F,$A6,$8F,$9F,$5F,$9B
	.dc.b	$C3,$F8,$85,$7F,$0E,$AE,$18,$7F,$12,$BF,$C3,$1F,$4B,$0F,$9C,$7F
	.dc.b	$25,$3C,$F1,$FD,$2B,$0F,$E2,$95,$FE,$2B,$9F,$C2,$9F,$36,$FF,$D1
	.dc.b	$57,$FE,$83,$7F,$A9,$1E,$67,$FD,$13,$87,$F8,$21,$FD,$64,$7D,$47
	.dc.b	$F4,$EF,$E1,$5B,$87,$23,$F9,$9C,$7F,$2D,$3F,$89,$3F,$E8,$6A,$FE
	.dc.b	$18,$7F,$10,$3F,$B1,$5F,$F8,$02,$BF,$66,$FF,$10,$3F,$71,$5F,$EB
	.dc.b	$B7,$F1,$AF,$F4,$48,$FE,$F0,$F9,$8F,$F6,$64,$FF,$9F,$A3,$F8,$A9
	.dc.b	$93,$F7,$D1,$FD,$F4,$FE,$F1,$5F,$C6,$4F,$4C,$7D,$03,$F8,$29,$FD
	.dc.b	$F2,$FF,$D0,$F1,$FD,$CA,$55,$F8,$86,$BF,$DF,$4F,$63,$FC,$41,$FE
	.dc.b	$FA,$7F,$33,$70,$FF,$7D,$7F,$A0,$1F,$50,$F9,$67,$F8,$60,$FF,$1C
	.dc.b	$57,$BD,$F5,$9F,$E7,$7A,$FF,$3B,$39,$AE,$27,$F8,$F9,$F5,$1F,$C5
	.dc.b	$EB,$FC,$56,$FF,$8C,$EF,$FA,$15,$FD,$29,$FB,$71,$F4,$DF,$44,$FD
	.dc.b	$37,$E1,$4F,$F8,$4A,$0F,$FB,$B6,$FF,$C5,$D5,$FE,$31,$9F,$E8,$67
	.dc.b	$23,$FD,$69,$5F,$D0,$DF,$A6,$7D,$7B,$FA,$67,$F4,$A3,$83,$FE,$F8
	.dc.b	$8F,$F1,$43,$FC,$F9,$FA,$07,$E4,$8F,$EB,$27,$E2,$91,$F5,$9F,$A0
	.dc.b	$FF,$74,$3E,$61,$F4,$0F,$90,$78,$0F,$9E,$7F,$98,$EB,$E7,$CF,$FB
	.dc.b	$8A,$BF,$CC,$4F,$CF,$38,$1F,$87,$7F,$0C,$7C,$E9,$FE,$F8,$FE,$19
	.dc.b	$FF,$70,$9F,$E5,$EB,$FC,$E4,$7E,$81,$F9,$C7,$CD,$8F,$ED,$D1,$F3
	.dc.b	$63,$F3,$F3,$FD,$0A,$FF,$DA,$75,$F3,$23,$FD,$AB,$5F,$DC,$6B,$F4
	.dc.b	$27,$CF,$AF,$97,$5F,$DB,$9F,$97,$3F,$DA,$E3,$13,$FA,$F9,$FD,$9C
	.dc.b	$7E,$04,$F2,$47,$F0,$D9,$FF,$0D,$C7,$F8,$82,$3E,$6C,$FC,$68,$FF
	.dc.b	$0A,$9F,$97,$9F,$93,$3F,$E1,$83,$F3,$53,$FC,$22,$7F,$2C,$FF,$08
	.dc.b	$3F,$2A,$7F,$A6,$A7,$EF,$47,$FB,$32,$3F,$C2,$CF,$F8,$5E,$BF,$E0
	.dc.b	$0C,$FC,$78,$FE,$0D,$1F,$1E,$FF,$7B,$1F,$1D,$FD,$FC,$FC,$78,$FF
	.dc.b	$09,$C7,$F0,$2A,$F8,$EF,$E3,$2F,$FD,$F7,$3F,$16,$7F,$24,$7F,$00
	.dc.b	$8C,$3F,$90,$9F,$F5,$AC,$FE,$42,$BE,$3C,$7F,$31,$BF,$F2,$3D,$7C
	.dc.b	$7A,$FF,$58,$C7,$E3,$E3,$FC,$7D,$1F,$CB,$A7,$FC,$F9,$1F,$BC,$8F
	.dc.b	$DE,$CF,$FA,$A2,$BF,$98,$DF,$CA,$3F,$1B,$5F,$BA,$AF,$F1,$99,$FE
	.dc.b	$3B,$3F,$91,$9F,$B8,$8F,$E5,$4F,$F0,$0A,$FF,$1C,$9F,$CA,$6B,$E3
	.dc.b	$5F,$F2,$6A,$FC,$65,$FE,$2E,$FF,$6C,$7C,$5A,$FC,$65,$7C,$9A,$FF
	.dc.b	$52,$9F,$1A,$BF,$69,$5F,$B5,$BF,$F3,$4D,$FC,$68,$FE,$47,$1F,$1A
	.dc.b	$7F,$DA,$B1,$F1,$AF,$A3,$FE,$3E,$9F,$E4,$77,$F8,$D7,$FA,$A5,$7C
	.dc.b	$7A,$DD,$7F,$1D,$3E,$2D,$FC,$87,$F6,$95,$FE,$62,$9F,$F6,$54,$7F
	.dc.b	$98,$8F,$8B,$3F,$8D,$8F,$F6,$75,$FC,$8B,$FE,$3D,$5F,$B4,$BF,$D9
	.dc.b	$C7,$F9,$7E,$BE,$44,$7F,$81,$67,$F9,$5D,$7F,$B2,$1F,$93,$1F,$B5
	.dc.b	$BF,$93,$1A,$3F,$A7,$4F,$C9,$AF,$F8,$03,$7F,$E5,$89,$FF,$2E,$53
	.dc.b	$0F,$F6,$3B,$FE,$5C,$BF,$F3,$05,$66,$FF,$16,$7C,$4B,$FF,$68,$61
	.dc.b	$FE,$34,$3F,$C8,$8F,$F2,$C3,$37,$FE,$C7,$BF,$F1,$76,$1F,$1E,$7C
	.dc.b	$C7,$F9,$74,$FF,$16,$CF,$77,$22,$91,$FC,$96,$30,$FF,$98,$0F,$F0
	.dc.b	$E5,$7F,$62,$9F,$F0,$23,$FE,$A7,$7F,$CC,$13,$FE,$C9,$3F,$C5,$87
	.dc.b	$FB,$6A,$3F,$CF,$27,$F5,$87,$FD,$FD,$87,$FB,$1D,$FF,$1A,$5B,$0F
	.dc.b	$F6,$36,$1F,$E6,$19,$C1,$FE,$D5,$C3,$FC,$88,$FC,$3B,$FF,$31,$C7
	.dc.b	$F8,$12,$3F,$CC,$71,$F2,$1F,$EA,$71,$FE,$08,$9F,$F1,$C4,$FF,$99
	.dc.b	$0E,$73,$FE,$30,$AF,$E0,$F5,$FA,$29,$FE,$59,$3F,$CB,$63,$F4,$53
	.dc.b	$FE,$02,$8F,$D1,$1F,$EC,$7A,$FE,$5A,$FF,$07,$3F,$C0,$51,$F7,$5F
	.dc.b	$9F,$5F,$E3,$57,$F4,$2F,$F5,$37,$F4,$31,$B9,$FE,$A9,$1F,$C2,$0F
	.dc.b	$D0,$CF,$F8,$C5,$FE,$5B,$3D,$70,$FE,$D2,$7F,$92,$63,$FD,$B8,$FF
	.dc.b	$B0,$5B,$7F,$E0,$08,$FE,$AA,$FE,$8D,$FE,$5B,$1F,$C2,$E3,$F4,$55
	.dc.b	$FD,$52,$3F,$AA,$C7,$F5,$98,$FF,$1A,$D7,$F0,$D7,$FC,$6A,$00

;--------------------------------------------------------------------------------
;	FDBcc Dr,<label>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fdbcc88x_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040,-(sp)
	peamsg	'FDBCC DR,<LABEL>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FDBcc Dr,<label>',13,10
;------------------------------------------------
;	d1	actual result
;	d2	actual count
;	d3	actual status
;	d4	source count
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result,expected count,expected status,...
;	a1	expected result
;	a2	expected count
;	a3	expected status
;	a4	source count,...
;------------------------------------------------
	lea.l	push_decompressed,a3
;decompress data
	move.l	a3,-(sp)
	pea.l	fdbcc88x_data_compressed
	jbsr	decompress
	addq.l	#8,sp
@@:
	addq.l	#4,a3			;source count
	tst.l	(a3)
	bne	@b
	addq.l	#4,a3			;0
	movea.l	a3,a0			;expected result,expected count,expected status,...
;
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
	lea.l	push_decompressed,a4	;source count,...
44:
;FDBcc Dr,<forward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	move.l	(a4),d4			;source count
	move.l	d4,d2			;actual count
	jsr	([fdbcc88x_execute_forward,za0,d5.l*4])	;EXECUTE
					;actual result
					;actual count
	fmove.l	fpsr,d3			;actual status
;
	movem.l	(a0),a1-a3		;expected result,expected count,expected status
;
	movem.l	d1-d3/a1-a3,-(sp)	;actual result,actual count,actual status,expected result,expected count,expected status
	jbsr	test_double
	lea.l	24(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	' Dr='
	puthex8	d4			;source count
	putmsg	',<forward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d3/a1-a3,-(sp)	;actual result,actual count,actual status,expected result,expected count,expected status
	jbsr	output_double
	lea.l	28(sp),sp
@@:
;
;FDBcc Dr,<backward-label>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	move.l	(a4),d4			;source count
	move.l	d4,d2			;actual count
	jsr	([fdbcc88x_execute_backward,za0,d5.l*4])	;EXECUTE
					;actual result
					;actual count
	fmove.l	fpsr,d3			;actual status
;
	movem.l	(a0),a1-a3		;expected result,expected count,expected status
;
	movem.l	d1-d3/a1-a3,-(sp)	;actual result,actual count,actual status,expected result,expected count,expected status
	jbsr	test_double
	lea.l	24(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDB'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	' Dr='
	puthex8	d4			;source count
	putmsg	',<backward-label> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d3/a1-a3,-(sp)	;actual result,actual count,actual status,expected result,expected count,expected status
	jbsr	output_double
	lea.l	28(sp),sp
@@:
;
	lea.l	(12,a0),a0		;expected result,expected count,expected status,...
;
	addq.l	#4,a4			;source count
	tst.l	(a4)			;source count,...
	bne	44b
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
fdbcc88x_execute_forward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fdbcc88x_execute_forward_&cc
  .endm
fdbcc88x_execute_backward::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fdbcc88x_execute_backward_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
fdbcc88x_execute_forward_&cc::
	moveq.l	#2,d1			;actual result. 2=too long
	FDB&cc	d2,@f			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts

	subq.l	#2,d1			;actual result. -1=too short(forward)/too long(backward)
@@:
	subq.l	#1,d1			;actual result. 1=taken
	rts

fdbcc88x_execute_backward_&cc::
	moveq.l	#2,d1			;actual result. 2=too short
	FDB&cc	d2,@b			;EXECUTE
	moveq.l	#0,d1			;actual result. 0=not taken
	rts
  .endm
	.cpu	68000

	.align	4
fdbcc88x_data_compressed::
	.dc.b	$00,$03,$60,$98,$10,$86,$1C,$7E,$02,$FE,$01,$D7,$E0,$7A,$13,$C6
	.dc.b	$35,$58,$3A,$47,$37,$75,$A8,$56,$8E,$0F,$7A,$4F,$4A,$B1,$E5,$8F
	.dc.b	$21,$1E,$71,$AA,$EF,$1C,$DF,$44,$7B,$5F,$83,$3E,$D7,$E1,$92,$3C
	.dc.b	$B5,$D2,$B4,$FC,$23,$91,$DC,$F0,$CF,$C4,$23,$EB,$8F,$51,$F0,$4F
	.dc.b	$41,$E1,$9F,$34,$6E,$3E,$9C,$43,$EB,$BF,$0D,$CC,$65,$D3,$F6,$3F
	.dc.b	$80,$71,$1F,$4C,$F9,$26,$E3,$ED,$BA,$7E,$D4,$70,$8F,$C1,$4F,$77
	.dc.b	$F0,$B3,$F6,$DE,$07,$DC,$9B,$1D,$A3,$C6,$7C,$33,$32,$9E,$D7,$F1
	.dc.b	$67,$C8,$F7,$7D,$AF,$DE,$8F,$33,$F7,$E3,$73,$F8,$06,$BD,$5F,$83
	.dc.b	$7F,$22,$3D,$CF,$B6,$3F,$04,$7E,$3A,$B2,$FA,$4F,$9B,$1F,$83,$7E
	.dc.b	$43,$F4,$A3,$E3,$3F,$A2,$9F,$44,$FE,$10,$F9,$57,$F8,$AB,$F2,$47
	.dc.b	$D6,$AF,$CA,$BD,$27,$D4,$E5,$FD,$85,$FD,$3B,$F8,$D3,$F7,$9E,$EF
	.dc.b	$EB,$9F,$8C,$FD,$D7,$E3,$3B,$7A,$BF,$88,$7F,$5D,$1F,$0E,$3F,$60
	.dc.b	$7E,$14,$C4,$FE,$2D,$C1,$F1,$4F,$C1,$9F,$14,$FB,$E7,$A0,$F8,$33
	.dc.b	$E6,$8F,$C4,$47,$57,$E8,$36,$F4,$FE,$1C,$D4,$FC,$87,$F4,$51,$F0
	.dc.b	$9F,$C2,$C7,$DB,$9F,$E0,$B3,$B9,$FC,$BB,$F0,$9D,$DF,$D4,$8F,$83
	.dc.b	$1F,$86,$3E,$09,$F5,$4E,$91,$FC,$1A,$FF,$0B,$3E,$B8,$F8,$13,$F8
	.dc.b	$63,$E0,$1F,$B3,$7F,$09,$7F,$90,$BE,$67,$C0,$9F,$BD,$5C,$E3,$EA
	.dc.b	$C7,$EB,$AB,$F1,$97,$F8,$1B,$F2,$CF,$D9,$7E,$01,$F8,$07,$EF,$5F
	.dc.b	$CF,$8F,$E1,$91,$F8,$37,$CA,$FD,$F3,$E4,$1F,$A3,$8F,$C5,$BF,$C1
	.dc.b	$0F,$D3,$57,$F0,$97,$F8,$5C,$E9,$C1,$FC,$00,$F9,$8F,$C8,$7D,$E7
	.dc.b	$E0,$9F,$DF,$9F,$8F,$9F,$E8,$51,$FC,$59,$FA,$CF,$F2,$07,$D2,$FF
	.dc.b	$1D,$7F,$98,$BB,$8F,$86,$FF,$26,$9F,$61,$FD,$41,$FB,$93,$EC,$8F
	.dc.b	$AF,$1B,$3F,$9E,$9F,$81,$9D,$5F,$E8,$E7,$FA,$9C,$FF,$0A,$8F,$D0
	.dc.b	$3F,$D8,$5F,$AE,$D8,$F8,$6E,$67,$E5,$9F,$1D,$FD,$6C,$FD,$BA,$CC
	.dc.b	$F2,$9F,$ED,$4F,$91,$FE,$78,$FE,$7D,$FB,$07,$F3,$03,$E5,$9F,$58
	.dc.b	$F4,$4F,$EE,$E3,$EB,$3F,$7C,$FD,$8C,$7F,$46,$9C,$D7,$EF,$8F,$4C
	.dc.b	$7F,$3E,$AF,$C1,$CF,$A4,$FA,$F1,$F4,$E3,$E1,$BF,$E0,$67,$D4,$FF
	.dc.b	$6C,$7D,$4F,$EE,$1F,$AE,$FF,$72,$3D,$07,$80,$FE,$E1,$1F,$5D,$E6
	.dc.b	$7F,$6C,$AF,$E5,$EF,$23,$FB,$91,$F9,$67,$DA,$FD,$58,$FB,$75,$93
	.dc.b	$B4,$7F,$7C,$AF,$A2,$FF,$43,$7D,$AF,$F6,$F7,$F2,$6F,$F8,$91,$FC
	.dc.b	$9B,$97,$DA,$FF,$76,$7E,$B3,$93,$F9,$39,$FD,$08,$FE,$14,$7F,$2A
	.dc.b	$9F,$E3,$EF,$C8,$AF,$EE,$2F,$E9,$CF,$E4,$C7,$17,$F8,$43,$FE,$00
	.dc.b	$3F,$A4,$9F,$DB,$9E,$4F,$F0,$83,$F4,$D3,$FC,$C6,$7F,$06,$7F,$07
	.dc.b	$8F,$E5,$2E,$63,$C4,$FF,$24,$8F,$E3,$47,$CF,$3F,$9A,$1F,$C8,$AF
	.dc.b	$F9,$1C,$FF,$68,$BF,$EB,$E7,$F6,$59,$FD,$44,$7F,$1D,$73,$7E,$A7
	.dc.b	$F8,$F3,$F8,$A7,$FC,$4F,$3A,$9F,$E7,$51,$F4,$5F,$14,$F8,$E3,$FC
	.dc.b	$2B,$1C,$DF,$EC,$4F,$38,$FF,$20,$47,$F8,$76,$3D,$8F,$F1,$A7,$F5
	.dc.b	$4F,$D5,$39,$1F,$E1,$DB,$FA,$B5,$99,$F1,$CF,$E9,$5F,$E4,$0F,$F8
	.dc.b	$82,$72,$FF,$96,$9F,$EB,$75,$F5,$0F,$E6,$31,$FC,$76,$7D,$73,$F5
	.dc.b	$1F,$0C,$7F,$96,$23,$EA,$57,$F2,$C9,$EB,$3F,$E5,$37,$FC,$2E,$FD
	.dc.b	$03,$8D,$7F,$98,$AB,$FA,$B4,$7E,$36,$BE,$A3,$FD,$56,$7E,$5C,$7E
	.dc.b	$92,$3F,$CA,$F3,$EA,$9F,$F0,$C5,$7F,$59,$AF,$A9,$7F,$D4,$E3,$FA
	.dc.b	$71,$FD,$48,$F0,$9F,$A8,$BF,$AB,$3F,$D0,$63,$D3,$1F,$C6,$8F,$DB
	.dc.b	$9F,$0C,$FE,$06,$7F,$8B,$2B,$F8,$D4,$7F,$56,$8F,$EA,$F1,$FD,$FE
	.dc.b	$FF,$8C,$C7,$F4,$E3,$EF,$4F,$F2,$47,$F9,$2C,$FF,$50,$9F,$E3,$73
	.dc.b	$FC,$31,$FE,$97,$3F,$C6,$E3,$F8,$83,$FC,$52,$7F,$87,$CF,$F1,$3A
	.dc.b	$FE,$5C,$7F,$0A,$3F,$C9,$93,$FD,$14,$FE,$24,$66,$7F,$38,$FF,$A2
	.dc.b	$A3,$F8,$94,$7B,$6B,$F8,$8C,$FF,$7D,$3F,$8A,$BF,$25,$FF,$2C,$4F
	.dc.b	$F1,$58,$FE,$1F,$1F,$A3,$7F,$A6,$3F,$D1,$27,$FB,$64,$7F,$59,$8F
	.dc.b	$E8,$67,$13,$F3,$C7,$F0,$B3,$FB,$B9,$F8,$93,$F6,$A7,$F1,$13,$F9
	.dc.b	$F1,$FE,$05,$7E,$BB,$FC,$F6,$BE,$5D,$64,$FE,$5A,$FF,$22,$8F,$F0
	.dc.b	$51,$FC,$E0,$F9,$07,$C4,$8F,$EE,$6F,$FA,$D1,$CC,$FF,$72,$8F,$97
	.dc.b	$5F,$83,$AF,$E7,$71,$F8,$5A,$F9,$D5,$F2,$A3,$F7,$87,$CA,$7E,$89
	.dc.b	$FC,$E1,$FE,$6C,$FC,$A7,$FB,$7B,$FD,$D4,$FE,$66,$FF,$03,$9F,$D1
	.dc.b	$3F,$D6,$CF,$F3,$0C,$FF,$03,$BF,$D0,$47,$F3,$63,$F8,$13,$FC,$0E
	.dc.b	$BF,$D4,$73,$FC,$E5,$FD,$CD,$7C,$5A,$FE,$72,$7F,$A6,$9F,$D8,$1F
	.dc.b	$81,$BF,$F6,$D4,$7F,$A8,$27,$F8,$4B,$FE,$DA,$7F,$7A,$FF,$01,$9F
	.dc.b	$DC,$57,$9E,$3F,$CB,$4F,$F0,$59,$FE,$12,$FF,$95,$23,$FD,$20,$7F
	.dc.b	$34,$9E,$35,$FC,$A6,$7F,$9F,$CF,$FA,$8A,$3C,$2F,$C8,$AF,$D7,$BF
	.dc.b	$E3,$DA,$FE,$11,$1E,$93,$E8,$47,$F3,$47,$F8,$0B,$FC,$E8,$FF,$46
	.dc.b	$9F,$D9,$0F,$F4,$4C,$7F,$30,$7F,$C3,$D1,$FC,$22,$78,$1F,$C0,$A7
	.dc.b	$EA,$BF,$CF,$E7,$CE,$7F,$A5,$1F,$3B,$FE,$13,$9F,$EB,$CF,$F9,$5A
	.dc.b	$7E,$7C,$7F,$47,$7C,$F0,$12,$8F,$E9,$31,$FE,$49,$9F,$A2,$FF,$63
	.dc.b	$3F,$D8,$A7,$C9,$3F,$98,$1E,$23,$FC,$20,$FF,$A0,$DF,$A1,$7F,$E8
	.dc.b	$59,$FF,$41,$5F,$F2,$E9,$FE,$AD,$5F,$CB,$9F,$9D,$5F,$C7,$AB,$F9
	.dc.b	$8D,$7F,$2D,$8F,$E5,$93,$FC,$9C,$FF,$33,$1F,$CB,$0F,$E2,$C7,$F0
	.dc.b	$F3,$FB,$09,$FD,$68,$FF,$75,$9F,$C0,$CF,$31,$F7,$8F,$E0,$67,$F2
	.dc.b	$C3,$FC,$E5,$3F,$C0,$9F,$F3,$EC,$FF,$02,$AF,$EE,$CF,$F9,$D2,$3F
	.dc.b	$84,$9F,$D5,$63,$FC,$77,$1F,$CB,$2F,$FC,$7B,$1F,$C0,$67,$FA,$11
	.dc.b	$FC,$AE,$BF,$80,$BF,$E7,$3A,$FE,$03,$3E,$B8,$FE,$03,$5F,$E6,$D8
	.dc.b	$FF,$54,$4F,$F2,$93,$F8,$23,$FC,$A2,$7F,$CC,$87,$F1,$D3,$E3,$1F
	.dc.b	$D9,$4F,$E5,$07,$C6,$3F,$7C,$7C,$63,$FB,$89,$F1,$8F,$F5,$71,$F4
	.dc.b	$0F,$8C,$7F,$99,$EB,$E3,$3F,$E9,$99,$F8,$D3,$FD,$42,$3F,$93,$57
	.dc.b	$F5,$08,$F8,$B1,$F1,$A3,$E2,$C7,$C4,$8F,$8B,$1F,$C9,$63,$F9,$24
	.dc.b	$7F,$25,$8F,$E9,$D1,$F1,$63,$FD,$16,$FD,$23,$FA,$73,$FE,$26,$3F
	.dc.b	$C4,$93,$F1,$A3,$FB,$8C,$7F,$96,$8F,$F4,$85,$7F,$24,$7F,$60,$7F
	.dc.b	$94,$0F,$CF,$1F,$C3,$4F,$F2,$A1,$FE,$59,$7F,$6E,$7F,$89,$27,$F9
	.dc.b	$25,$FF,$B1,$2B,$F6,$B5,$FC,$7E,$BF,$6A,$7F,$88,$9F,$E4,$51,$FC
	.dc.b	$69,$FF,$0C,$9F,$EC,$F8,$FE,$42,$FE,$C6,$BF,$90,$57,$C2,$8F,$E3
	.dc.b	$E7,$EB,$DF,$D8,$57,$F8,$76,$3F,$DE,$CF,$F8,$B8,$FE,$6D,$7F,$E4
	.dc.b	$F8,$FF,$7D,$DF,$F1,$E9,$FF,$27,$1F,$C2,$23,$FD,$FE,$FF,$1E,$3F
	.dc.b	$A5,$BF,$22,$BE,$4C,$7F,$AE,$6F,$FA,$54,$FF,$6F,$3F,$CF,$97,$FC
	.dc.b	$76,$BF,$A5,$47,$F1,$33,$33,$F2,$1F,$F3,$ED,$7F,$45,$7F,$8E,$4F
	.dc.b	$C9,$8F,$E9,$13,$FE,$49,$8F,$C4,$CF,$F1,$63,$FA,$43,$FE,$7A,$7F
	.dc.b	$8C,$C7,$C8,$7F,$CF,$33,$F9,$D7,$C4,$78,$E9,$3F,$C9,$AF,$FA,$49
	.dc.b	$FC,$6A,$7E,$5D,$FF,$91,$EF,$F8,$54,$72,$3F,$CE,$F3,$F2,$CF,$E3
	.dc.b	$6F,$F8,$EA,$3F,$A3,$BF,$E1,$D3,$F6,$C7,$F8,$2C,$FD,$11,$FE,$00
	.dc.b	$3F,$6C,$7E,$40,$F8,$87,$ED,$8F,$C0,$9F,$B6,$3F,$8D,$9F,$C9,$AF
	.dc.b	$FC,$8D,$1F,$C6,$A3,$F8,$DC,$7E,$56,$3C,$11,$E9,$3E,$A5,$FC,$37
	.dc.b	$07,$FB,$6A,$3F,$72,$E0,$FF,$6D,$16,$3F,$C9,$2F,$F0,$F3,$F7,$66
	.dc.b	$67,$FC,$4C,$7F,$B8,$A3,$FA,$99,$FC,$82,$3E,$01,$FC,$80,$FF,$4E
	.dc.b	$52,$BF,$90,$C7,$F2,$2A,$FA,$4F,$F4,$63,$E3,$9F,$72,$78,$3F,$C8
	.dc.b	$8F,$75,$7F,$A1,$5F,$D1,$4F,$F2,$83,$FC,$E5,$5E,$D8,$FE,$D1,$1F
	.dc.b	$EF,$87,$CE,$FF,$89,$9F,$9E,$7F,$93,$1F,$9E,$FF,$23,$3E,$79,$F9
	.dc.b	$C3,$E3,$1F,$E5,$B3,$FC,$93,$5F,$E7,$37,$F9,$21,$A9,$FE,$90,$FF
	.dc.b	$9C,$E7,$F9,$23,$FD,$1C,$FE,$48,$7E,$70,$F9,$27,$F2,$43,$FC,$91
	.dc.b	$3F,$09,$FE,$4F,$5F,$D2,$67,$E4,$C7,$F2,$B3,$F0,$07,$FA,$56,$7F
	.dc.b	$92,$1F,$E1,$97,$F0,$71,$FE,$FC,$7F,$42,$7F,$85,$4F,$E4,$97,$FB
	.dc.b	$E7,$FD,$F5,$5F,$BD,$7F,$D2,$97,$F9,$A7,$FC,$8D,$3F,$BD,$7A,$DF
	.dc.b	$F1,$1B,$FF,$0B,$57,$F8,$3E,$FF,$1B,$5F,$AF,$9F,$A8,$7F,$B1,$6F
	.dc.b	$F8,$04,$7F,$91,$8F,$CB,$47,$EF,$1E,$4F,$FB,$18,$FB,$CF,$FB,$1A
	.dc.b	$3F,$94,$BE,$E9,$FF,$18,$BF,$C1,$5F,$F1,$4D,$7F,$2B,$3F,$DF,$27
	.dc.b	$CD,$8F,$8A,$FE,$5E,$FF,$81,$3F,$EB,$A9,$FF,$0E,$4F,$F7,$23,$D4
	.dc.b	$7F,$92,$0F,$F6,$F9,$FE,$5F,$32,$7F,$53,$8F,$F2,$41,$FE,$F2,$7E
	.dc.b	$79,$F5,$0F,$F3,$89,$FE,$47,$8F,$A0,$FE,$F6,$7F,$A0,$9F,$EF,$53
	.dc.b	$FA,$9C,$FF,$2C,$8F,$EA,$75,$FC,$B1,$FE,$09,$7F,$3A,$FF,$CB,$91
	.dc.b	$F4,$5F,$33,$FC,$AA,$3E,$74,$7F,$2F,$8F,$F7,$A5,$7C,$57,$51,$FE
	.dc.b	$F5,$9F,$E5,$E7,$FB,$F9,$F4,$C7,$CF,$8F,$F7,$F1,$FE,$4F,$9F,$EB
	.dc.b	$27,$38,$FC,$13,$F3,$EB,$FC,$94,$FC,$F8,$FE,$61,$1F,$C1,$5F,$E0
	.dc.b	$A7,$F5,$1B,$E4,$FF,$39,$BF,$C1,$BF,$E6,$77,$F8,$35,$6A,$3F,$4A
	.dc.b	$7F,$B3,$CF,$CD,$1F,$1A,$38,$DF,$FA,$CE,$3F,$A6,$4F,$FB,$D9,$FF
	.dc.b	$1A,$47,$F0,$67,$F9,$6D,$7F,$6F,$7F,$3F,$5F,$E2,$AA,$FE,$6B,$1F
	.dc.b	$CC,$A3,$C2,$F9,$CF,$F6,$81,$FC,$AC,$F0,$1F,$9F,$8F,$E1,$2F,$F9
	.dc.b	$06,$BF,$84,$BC,$DF,$E6,$47,$09,$FF,$0F,$47,$F4,$39,$FF,$65,$1F
	.dc.b	$E2,$D3,$FD,$A2,$7E,$AD,$FF,$16,$CF,$F0,$A7,$E7,$1F,$83,$3F,$AE
	.dc.b	$9F,$C8,$0F,$D7,$CF,$CC,$9F,$F3,$E4,$7F,$9D,$CF,$C7,$D7,$CF,$7F
	.dc.b	$98,$CF,$CC,$3F,$11,$3F,$2E,$3F,$CA,$67,$E7,$5F,$E0,$EF,$F3,$13
	.dc.b	$FC,$A9,$5F,$D7,$DF,$11,$FE,$54,$3F,$99,$D7,$F3,$17,$F9,$AD,$7F
	.dc.b	$BE,$2B,$FB,$9C,$7C,$93,$FA,$D4,$FF,$9E,$1F,$95,$3F,$D6,$5F,$E6
	.dc.b	$07,$F4,$99,$FE,$A3,$5F,$E7,$98,$FF,$29,$9F,$D6,$4F,$F5,$1C,$7F
	.dc.b	$95,$1F,$D7,$9F,$C0,$EB,$F9,$8C,$7F,$8B,$63,$FD,$93,$7F,$D4,$5F
	.dc.b	$F1,$69,$FC,$08,$E3,$5F,$EC,$93,$EB,$38,$9F,$F7,$B9,$FD,$AC,$FC
	.dc.b	$CB,$FC,$AE,$BF,$54,$FF,$BE,$63,$F8,$23,$FC,$D1,$FE,$64,$FF,$95
	.dc.b	$8F,$F1,$79,$FE,$C6,$8F,$D3,$4F,$F0,$28,$FE,$63,$7F,$ED,$1A,$FF
	.dc.b	$7B,$CF,$F3,$19,$FF,$63,$9F,$CB,$DF,$F7,$B1,$FC,$02,$3F,$97,$1F
	.dc.b	$EC,$78,$FF,$63,$5F,$EF,$9F,$E5,$F1,$EE,$7F,$81,$BF,$E7,$87,$FD
	.dc.b	$E8,$7F,$5B,$8F,$F2,$BB,$FC,$9C,$FE,$58,$FC,$B7,$FA,$D1,$FD,$20
	.dc.b	$FF,$21,$3F,$D6,$DF,$F2,$B9,$99,$FF,$71,$5F,$F2,$CA,$FF,$04,$BF
	.dc.b	$CB,$27,$E4,$5F,$F9,$5D,$FE,$F9,$3F,$E3,$97,$F8,$41,$FC,$84,$FF
	.dc.b	$3F,$1F,$20,$FE,$AE,$FC,$C8,$FE,$50,$79,$23,$E4,$4F,$F9,$55,$F9
	.dc.b	$53,$F2,$0F,$EE,$AF,$C4,$8F,$EA,$6F,$C8,$9F,$EA,$6F,$EE,$0F,$DC
	.dc.b	$1F,$E2,$18,$FD,$93,$FC,$8C,$FE,$4B,$1F,$B9,$7F,$C4,$D5,$FE,$29
	.dc.b	$6C,$F0,$8F,$E9,$07,$CD,$8F,$F1,$A5,$7F,$2E,$9F,$D2,$C7,$FB,$C2
	.dc.b	$3F,$CA,$CF,$EE,$0F,$F1,$41,$FC,$84,$FF,$14,$4F,$FB,$2D,$FD,$B9
	.dc.b	$FE,$28,$35,$1F,$C8,$1F,$F1,$FC,$7F,$BD,$AB,$FA,$6C,$7F,$97,$A3
	.dc.b	$F9,$0C,$7E,$E6,$3F,$90,$4F,$FB,$1A,$7F,$71,$1F,$E2,$83,$F7,$07
	.dc.b	$FB,$42,$3F,$DE,$0F,$F7,$93,$FD,$CC,$7F,$0E,$3F,$CC,$07,$F7,$F3
	.dc.b	$F7,$A7,$F8,$42,$3F,$C9,$AF,$C8,$8C,$4F,$F1,$F9,$F9,$13,$FD,$5E
	.dc.b	$7E,$41,$C1,$FF,$79,$3F,$21,$FE,$96,$FF,$42,$AF,$DF,$3F,$D2,$A3
	.dc.b	$FC,$C2,$7C,$33,$FA,$4D,$7F,$34,$7F,$A1,$1F,$C1,$4F,$49,$FE,$7A
	.dc.b	$3F,$A4,$4F,$F2,$17,$E6,$C7,$C9,$7F,$D4,$53,$FD,$21,$FF,$28,$BF
	.dc.b	$E7,$87,$FA,$39,$FC,$80,$F8,$87,$F5,$C3,$F6,$93,$FC,$81,$FD,$B4
	.dc.b	$FF,$B3,$1F,$CD,$9F,$22,$BF,$D9,$8F,$F8,$85,$FE,$41,$1F,$E2,$08
	.dc.b	$FF,$5A,$4F,$ED,$5F,$F6,$59,$FC,$74,$FD,$A9,$FE,$20,$3F,$6A,$7F
	.dc.b	$B2,$CF,$D9,$3F,$EF,$23,$F6,$4F,$F0,$F3,$FC,$36,$7E,$DA,$BF,$D9
	.dc.b	$8F,$F8,$96,$BF,$C3,$87,$F1,$CA,$FD,$39,$FE,$4A,$BF,$ED,$13,$FE
	.dc.b	$29,$AF,$D4,$1F,$CC,$9F,$F1,$8B,$FE,$4B,$3F,$C3,$8F,$F2,$07,$FD
	.dc.b	$E8,$7F,$87,$1F,$F3,$CD,$FC,$97,$F1,$F5,$F2,$5F,$F6,$31,$FD,$20
	.dc.b	$FE,$35,$0A,$FE,$3D,$1F,$E4,$93,$F8,$F5,$7F,$48,$3F,$9E,$CF,$F9
	.dc.b	$26,$7F,$CF,$13,$FD,$E2,$7E,$23,$FE,$EB,$9F,$E9,$D3,$FE,$4B,$8F
	.dc.b	$F2,$41,$F9,$E3,$F9,$04,$7C,$CA,$FF,$24,$47,$EE,$2B,$F9,$05,$FE
	.dc.b	$90,$F9,$75,$FC,$7E,$7F,$C6,$11,$FE,$45,$9F,$F3,$99,$FA,$D3,$E5
	.dc.b	$47,$F2,$17,$E4,$BF,$31,$FE,$8C,$FF,$BC,$CF,$E4,$0F,$FA,$D5,$FF
	.dc.b	$00,$9F,$C1,$0F,$E3,$E7,$F8,$F8,$FF,$2D,$4F,$F1,$FA,$FD,$C4,$7F
	.dc.b	$24,$9F,$F0,$B3,$FE,$F3,$3F,$8F,$1F,$E1,$53,$F6,$F1,$FE,$3F,$7F
	.dc.b	$84,$1F,$C7,$CF,$F6,$2C,$7F,$BC,$1F,$E4,$51,$FE,$15,$7F,$6A,$7F
	.dc.b	$B1,$4F,$F0,$AC,$7F,$B6,$63,$FD,$33,$1F,$C7,$DF,$DB,$22,$1F,$BF
	.dc.b	$3F,$C3,$11,$FE,$3F,$9F,$F1,$93,$FE,$55,$3F,$DE,$52,$AF,$CC,$D7
	.dc.b	$F4,$97,$F8,$AB,$FE,$EF,$3F,$CD,$8F,$F8,$75,$FE,$E0,$7F,$9A,$4F
	.dc.b	$F0,$61,$FA,$43,$EF,$1F,$26,$7E,$9B,$FE,$68,$3E,$49,$F2,$0F,$F7
	.dc.b	$69,$FE,$CC,$3E,$49,$FC,$78,$F9,$27,$F4,$63,$FC,$7B,$1F,$93,$7C
	.dc.b	$A7,$F9,$51,$FE,$8C,$7F,$92,$9F,$F1,$E9,$FE,$EE,$3F,$D3,$71,$FE
	.dc.b	$E7,$7F,$8F,$BF,$1C,$FE,$3C,$FF,$BB,$A3,$E2,$BA,$3E,$09,$FE,$0D
	.dc.b	$3F,$D7,$E7,$73,$FD,$7F,$3F,$B7,$7F,$2C,$7F,$1D,$3F,$C7,$C7,$FB
	.dc.b	$22,$3F,$8F,$BF,$B6,$BF,$F2,$F3,$FB,$33,$F8,$F9,$FB,$5B,$FD,$CD
	.dc.b	$FE,$DA,$3F,$05,$5F,$EE,$E8,$FE,$05,$1F,$C8,$8F,$F0,$A9,$FE,$3E
	.dc.b	$3F,$C2,$F3,$FB,$77,$FC,$2D,$1F,$E3,$C3,$F6,$EF,$F8,$F0,$FD,$AC
	.dc.b	$7E,$DD,$FE,$49,$1F,$E1,$57,$F8,$99,$FB,$38,$FF,$0A,$BF,$B5,$BF
	.dc.b	$F0,$AB,$F9,$63,$E5,$1F,$DA,$8F,$D9,$9F,$2A,$3F,$CA,$2F,$F8,$5C
	.dc.b	$FF,$1F,$4F,$FB,$A0,$FF,$47,$3F,$26,$3D,$EF,$F9,$A1,$E3,$5F,$16
	.dc.b	$3F,$3C,$7F,$37,$3F,$C2,$67,$F9,$9E,$7F,$C7,$F7,$FC,$8A,$31,$7F
	.dc.b	$E3,$F7,$FC,$27,$3F,$E3,$F7,$F9,$0D,$FF,$4B,$8F,$F1,$F4,$FF,$20
	.dc.b	$9F,$D3,$C7,$CA,$7F,$CD,$37,$FC,$82,$2C,$FF,$4A,$8F,$E4,$11,$FE
	.dc.b	$65,$9F,$F3,$1D,$7C,$A7,$FC,$C7,$5F,$29,$FF,$1F,$1E,$C3,$F5,$A7
	.dc.b	$F8,$14,$FD,$19,$F9,$D3,$FB,$A1,$FE,$8C,$3F,$C2,$C7,$ED,$CF,$11
	.dc.b	$FB,$73,$FC,$42,$7F,$84,$8E,$C7,$FA,$D6,$3F,$D7,$87,$FB,$7C,$FE
	.dc.b	$17,$5F,$98,$8F,$DC,$1F,$EB,$53,$FC,$81,$1F,$EB,$57,$F9,$0C,$7E
	.dc.b	$E5,$FC,$24,$F6,$7F,$D6,$C7,$F8,$5A,$BF,$91,$4F,$F8,$F8,$AF,$F1
	.dc.b	$C3,$FC,$7D,$1F,$8D,$3F,$1C,$FF,$6F,$7F,$D6,$E7,$F8,$60,$FF,$20
	.dc.b	$3F,$29,$FF,$33,$9F,$00,$F4,$9F,$CD,$0F,$F6,$19,$FC,$00,$FD,$D1
	.dc.b	$FE,$B0,$3F,$C1,$27,$CA,$3F,$C3,$A7,$F3,$5A,$FF,$38,$BF,$ED,$C9
	.dc.b	$FE,$AB,$1F,$2A,$7F,$2D,$28,$F9,$6F,$E5,$8F,$65,$7B,$5F,$E9,$95
	.dc.b	$FE,$69,$7F,$A5,$9F,$E6,$93,$FC,$8B,$5F,$E6,$97,$F9,$2C,$7C,$AA
	.dc.b	$FF,$20,$47,$F3,$CA,$FE,$96,$FF,$4A,$AF,$F2,$05,$7F,$9A,$23,$FC
	.dc.b	$D7,$3E,$17,$E6,$CA,$3E,$C4,$62,$3D,$E0,$CB,$F7,$67,$FC,$8B,$78
	.dc.b	$7F,$CC,$87,$F9,$1A,$30,$7F,$B3,$E3,$8B,$FC,$7A,$7A,$4F,$D2,$3F
	.dc.b	$C9,$17,$FE,$21,$6C,$72,$7C,$D3,$63,$4F,$F9,$B9,$E6,$FF,$2F,$3C
	.dc.b	$93,$E9,$36,$FF,$07,$76,$FF,$06,$37,$1E,$FA,$F3,$DF,$FB,$9E,$3D
	.dc.b	$CF,$E3,$A7,$FB,$BD,$F9,$2F,$AC,$7F,$94,$9E,$AF,$F9,$52,$7F,$D0
	.dc.b	$CF,$F9,$5A,$B6,$FE,$A2,$BE,$74,$76,$7F,$7E,$7D,$53,$F1,$97,$9B
	.dc.b	$EB,$5E,$2B,$FF,$01,$1D,$9C,$57,$AC,$FC,$03,$FE,$92,$BF,$B0,$7D
	.dc.b	$D7,$81,$F5,$5E,$F1,$FC,$69,$D9,$F4,$9F,$F3,$05,$FD,$C9,$F8,$D1
	.dc.b	$FD,$B0,$FC,$29,$FE,$8A,$9F,$C1,$C7,$13,$FD,$62,$FC,$27,$FC,$E0
	.dc.b	$FD,$F7,$C6,$7F,$40,$8F,$1B,$FD,$16,$3F,$09,$5A,$9F,$C6,$57,$E3
	.dc.b	$AF,$FD,$5F,$1F,$E1,$CB,$FB,$B1,$DD,$FF,$14,$3F,$E7,$B8,$FF,$48
	.dc.b	$9F,$A5,$3F,$B9,$4F,$E0,$A7,$EA,$C7,$FA,$4C,$D5,$7F,$05,$9F,$C2
	.dc.b	$CF,$F0,$1B,$B3,$E5,$3F,$45,$5F,$49,$FF,$1C,$47,$DA,$9F,$F0,$0C
	.dc.b	$7F,$43,$9E,$A7,$C0,$8F,$F4,$D5,$ED,$FB,$F1,$F2,$DF,$EF,$77,$FA
	.dc.b	$EB,$F8,$31,$FE,$18,$3E,$CD,$7C,$1A,$FC,$7B,$CD,$FC,$8D,$FC,$BA
	.dc.b	$FF,$43,$C7,$E2,$2B,$EE,$9F,$DF,$CF,$B7,$5C,$DF,$E4,$D3,$F7,$AB
	.dc.b	$EA,$47,$C0,$3F,$23,$5F,$29,$FE,$B4,$7E,$F6,$B9,$47,$C9,$3F,$D9
	.dc.b	$53,$C9,$FC,$D4,$FE,$FE,$BF,$6E,$FE,$06,$BF,$7D,$7F,$7A,$7F,$77
	.dc.b	$79,$78,$D7,$F2,$C7,$EB,$BF,$54,$F9,$C7,$EE,$5F,$B2,$FD,$C7,$F1
	.dc.b	$0F,$EC,$DF,$91,$1F,$5E,$3F,$C5,$B3,$FB,$59,$FD,$D4,$FE,$06,$FC
	.dc.b	$35,$FE,$56,$9F,$A0,$7E,$C5,$FA,$B5,$F5,$A3,$F5,$F5,$F4,$DE,$33
	.dc.b	$FD,$25,$FB,$0F,$F1,$33,$FC,$92,$7E,$4E,$BF,$60,$7F,$33,$3F,$C6
	.dc.b	$35,$FB,$77,$E9,$47,$FB,$45,$FE,$63,$1F,$4E,$7F,$5B,$3F,$CB,$E3
	.dc.b	$F5,$91,$C2,$3F,$C8,$53,$FE,$75,$3F,$51,$7F,$49,$B3,$F6,$E7,$FD
	.dc.b	$A9,$7F,$4E,$3E,$5B,$F8,$C8,$FF,$69,$9F,$4E,$BF,$2D,$1F,$8C,$AF
	.dc.b	$D7,$CF,$E7,$CF,$E5,$57,$FC,$DA,$3C,$91,$FC,$80,$FA,$15,$CA,$3C
	.dc.b	$51,$FE,$0D,$7F,$96,$57,$D5,$BF,$BA,$7E,$A2,$74,$FD,$59,$FF,$1D
	.dc.b	$DF,$D5,$7F,$83,$47,$FA,$CA,$7F,$92,$BF,$86,$7D,$87,$F4,$78,$FE
	.dc.b	$46,$FF,$80,$5F,$EC,$ED,$3F,$D8,$55,$FA,$87,$EF,$9F,$C9,$23,$FB
	.dc.b	$33,$FE,$65,$AF,$AD,$1F,$A6,$8F,$AC,$FF,$3E,$BF,$54,$7D,$67,$D4
	.dc.b	$FF,$A4,$67,$F4,$D3,$FC,$82,$3C,$93,$FC,$01,$FF,$33,$C7,$F1,$F7
	.dc.b	$FD,$1F,$3F,$C7,$A7,$F4,$31,$F4,$CF,$EA,$47,$A0,$F5,$CF,$D4,$7F
	.dc.b	$8E,$BF,$D7,$2B,$F9,$8D,$7E,$9E,$3F,$8F,$C7,$D3,$8F,$E3,$57,$C6
	.dc.b	$BF,$93,$BF,$D9,$0F,$C7,$3F,$E9,$77,$F4,$EF,$D2,$8F,$EA,$8F,$1A
	.dc.b	$FC,$03,$F3,$CF,$DF,$9F,$DE,$8F,$C7,$19,$AF,$A5,$5F,$6E,$79,$4F
	.dc.b	$EF,$CF,$F4,$C5,$7D,$27,$ED,$47,$C3,$7F,$66,$FC,$37,$FD,$79,$3F
	.dc.b	$C6,$1F,$EA,$73,$F4,$9F,$D2,$9F,$E1,$28,$FD,$2B,$E9,$AF,$DF,$3F
	.dc.b	$DF,$8F,$49,$FD,$20,$FC,$E9,$FE,$0F,$3F,$8C,$4F,$E9,$67,$F8,$BD
	.dc.b	$7D,$13,$8C,$FF,$4A,$BC,$DF,$D1,$9F,$F0,$3D,$7E,$2A,$BF,$C3,$EF
	.dc.b	$E2,$9F,$D2,$BC,$A3,$FA,$5B,$F4,$9F,$F0,$29,$FD,$26,$3E,$49,$FE
	.dc.b	$1C,$3F,$8B,$1F,$4E,$BF,$8D,$BF,$24,$FF,$52,$4F,$D1,$3F,$8F,$CF
	.dc.b	$13,$C1,$1F,$DE,$A3,$F8,$BC,$7F,$79,$8F,$ED,$B7,$FC,$62,$7F,$AF
	.dc.b	$57,$A6,$7F,$49,$1E,$98,$FF,$01,$47,$D2,$AF,$E3,$33,$F4,$8F,$CF
	.dc.b	$C7,$F1,$5A,$FD,$14,$7F,$46,$9F,$E2,$AF,$F1,$C7,$EC,$1F,$3E,$BF
	.dc.b	$89,$1F,$3A,$BF,$C7,$D3,$FC,$40,$FC,$B1,$E5,$7E,$7B,$F3,$4F,$2B
	.dc.b	$FE,$CA,$3F,$3D,$3F,$36,$3E,$4C,$7F,$09,$9F,$EE,$8F,$E1,$0F,$CF
	.dc.b	$1F,$36,$7F,$D0,$D5,$FD,$0E,$BE,$6C,$FF,$6F,$AF,$EE,$67,$D0,$BF
	.dc.b	$EE,$53,$FD,$8A,$3F,$09,$23,$E9,$1F,$DB,$DF,$F0,$71,$E8,$7E,$7C
	.dc.b	$FF,$41,$AF,$F1,$65,$7C,$C9,$FE,$19,$1F,$32,$FF,$9F,$D7,$F7,$5A
	.dc.b	$F9,$87,$E6,$AB,$F9,$ED,$7F,$09,$7F,$96,$4F,$E9,$6B,$FC,$53,$5F
	.dc.b	$34,$FE,$DE,$7F,$09,$3E,$61,$F1,$67,$FC,$C1,$3F,$DB,$E3,$F9,$D4
	.dc.b	$7F,$BC,$0F,$EB,$F3,$F9,$BB,$FF,$19,$47,$C1,$3F,$30,$5A,$4F,$F1
	.dc.b	$2A,$DC,$FF,$09,$3F,$83,$C7,$F8,$B4,$F1,$3F,$CA,$9C,$4F,$27,$53
	.dc.b	$E0,$AF,$C3,$57,$85,$FF,$3D,$4F,$F7,$67,$53,$E4,$7F,$DF,$B1,$FD
	.dc.b	$16,$3F,$96,$4F,$FA,$B0,$FF,$37,$96,$3F,$CC,$13,$EE,$AF,$EA,$B3
	.dc.b	$FE,$63,$7C,$8F,$F4,$63,$FD,$98,$FB,$5F,$F4,$E1,$A3,$FD,$C0,$FE
	.dc.b	$DE,$7F,$CC,$47,$C8,$7F,$CB,$6E,$2B,$E6,$9E,$E9,$FF,$59,$1C,$CF
	.dc.b	$A0,$7F,$AD,$6B,$71,$F4,$8F,$7C,$7F,$B7,$DF,$CB,$BF,$E5,$93,$FB
	.dc.b	$94,$FD,$B8,$FE,$8F,$5F,$0A,$7F,$AC,$CF,$EB,$67,$FC,$51,$19,$8F
	.dc.b	$D0,$1F,$E2,$BA,$FB,$AF,$E0,$A3,$D6,$7F,$98,$27,$E9,$BF,$3A,$7E
	.dc.b	$BC,$FD,$37,$EB,$D6,$6B,$FD,$81,$3F,$5D,$FA,$55,$F9,$39,$FF,$68
	.dc.b	$CF,$FA,$2E,$3F,$12,$7F,$B9,$A3,$ED,$9E,$A7,$EB,$1F,$4A,$39,$1E
	.dc.b	$A3,$E9,$D7,$DF,$7E,$A1,$E2,$8F,$BA,$FE,$22,$BE,$A4,$7F,$B6,$0F
	.dc.b	$A8,$7D,$47,$F8,$03,$5F,$A8,$73,$7F,$CB,$4F,$E4,$9F,$EB,$A7,$EB
	.dc.b	$8F,$F4,$84,$FE,$A1,$F9,$87,$13,$EE,$47,$D4,$7E,$AC,$FD,$38,$FF
	.dc.b	$2F,$57,$D7,$3C,$B3,$FE,$2B,$3F,$12,$7E,$A6,$3F,$7D,$5F,$52,$32
	.dc.b	$7F,$A7,$1F,$AA,$FF,$98,$A7,$9B,$FE,$A0,$9F,$AA,$FD,$03,$E6,$1F
	.dc.b	$A4,$3E,$AC,$7E,$AA,$BE,$C3,$CA,$0F,$F9,$21,$FD,$49,$FE,$5D,$34
	.dc.b	$FE,$A5,$FA,$6F,$DE,$8F,$89,$1F,$A4,$8F,$F1,$3C,$7F,$7A,$9F,$E3
	.dc.b	$47,$B6,$7E,$8D,$7E,$A2,$BF,$80,$CF,$EA,$27,$EA,$BF,$45,$FD,$21
	.dc.b	$EC,$3F,$85,$57,$F7,$D9,$C9,$EB,$8F,$51,$FE,$86,$3F,$50,$66,$3F
	.dc.b	$2E,$FF,$37,$3F,$D8,$B1,$99,$FD,$3C,$7F,$A4,$0F,$8D,$5F,$8C,$7E
	.dc.b	$C4,$FE,$9C,$CC,$FF,$1C,$7F,$16,$7E,$9A,$3E,$94,$FE,$AA,$3E,$B1
	.dc.b	$F3,$23,$85,$7E,$91,$F9,$EF,$D8,$3F,$47,$18,$AF,$E1,$F3,$F2,$1F
	.dc.b	$88,$FE,$31,$F3,$3F,$D9,$5F,$F4,$34,$FD,$39,$FC,$04,$7F,$14,$9F
	.dc.b	$F2,$A9,$FD,$0E,$7F,$8A,$1F,$35,$FE,$2A,$FF,$4B,$8F,$D0,$3F,$D2
	.dc.b	$63,$E6,$57,$F1,$2A,$FD,$04,$FF,$7D,$AF,$95,$5F,$2E,$3F,$A3,$47
	.dc.b	$E6,$E3,$F3,$91,$FE,$27,$8F,$EB,$31,$F9,$93,$F3,$87,$C7,$AF,$94
	.dc.b	$FF,$43,$9F,$EF,$73,$F9,$78,$FC,$C4,$7F,$AD,$67,$F8,$65,$7E,$5C
	.dc.b	$F9,$55,$FE,$7E,$AF,$94,$7C,$47,$F8,$D9,$F2,$1F,$DF,$C6,$63,$F2
	.dc.b	$91,$FE,$7E,$3F,$20,$7E,$30,$FE,$0F,$18,$8F,$E6,$B1,$F9,$2A,$FE
	.dc.b	$0C,$7F,$6E,$7F,$83,$BF,$1C,$FE,$CC,$FF,$33,$3F,$1B,$5F,$C0,$A3
	.dc.b	$FB,$55,$7E,$42,$BF,$80,$D7,$C8,$3F,$AF,$9F,$84,$3F,$68,$7F,$02
	.dc.b	$3F,$B0,$1F,$B8,$3F,$C1,$07,$E4,$8F,$DE,$1F,$92,$AF,$92,$FF,$B7
	.dc.b	$E7,$E4,$D7,$EF,$63,$E4,$BF,$EA,$89,$FD,$E4,$7F,$AA,$A7,$E4,$57
	.dc.b	$C7,$8F,$91,$5F,$EA,$6A,$FE,$5D,$5F,$BA,$8F,$C7,$1F,$8D,$9F,$E5
	.dc.b	$AF,$F8,$5E,$7F,$1D,$1F,$22,$3F,$C2,$B5,$F8,$E7,$E4,$9F,$8E,$3E
	.dc.b	$3C,$7E,$39,$FE,$B9,$1F,$1E,$3F,$99,$D7,$EE,$1F,$90,$7C,$4A,$FC
	.dc.b	$7D,$7F,$28,$3F,$1E,$7F,$82,$A3,$E3,$BF,$8F,$AF,$8D,$1F,$1E,$3F
	.dc.b	$D4,$51,$F8,$F8,$FC,$84,$7C,$78,$F8,$D1,$F1,$E3,$FD,$A3,$1F,$8E
	.dc.b	$7F,$D3,$51,$F8,$E9,$F8,$EF,$ED,$23,$FC,$0C,$7C,$73,$F6,$87,$E3
	.dc.b	$9F,$E3,$A7,$C7,$3F,$C0,$93,$F1,$DF,$E9,$E2,$7F,$9B,$3F,$D5,$9F
	.dc.b	$E1,$6E,$63,$FC,$78,$7E,$25,$FE,$A4,$FC,$B7,$F1,$E7,$EE,$27,$FC
	.dc.b	$77,$1F,$E3,$98,$FD,$C4,$7F,$82,$63,$FC,$01,$1F,$2A,$3F,$C0,$71
	.dc.b	$F2,$9F,$F1,$BD,$7E,$42,$7F,$B8,$9F,$90,$8F,$E9,$D5,$F9,$07,$F8
	.dc.b	$F1,$F2,$AB,$F1,$73,$FD,$7A,$BF,$A2,$3F,$EF,$17,$FC,$A4,$FE,$A6
	.dc.b	$97,$FC,$4A,$73,$7F,$DD,$CF,$F4,$83,$8A,$FE,$D8,$75,$3F,$D3,$F7
	.dc.b	$CA,$77,$3F,$DF,$14,$47,$33,$FE,$7B,$7F,$C7,$AC,$9E,$D1,$F5,$6B
	.dc.b	$C9,$1F,$D7,$63,$EA,$1C,$A7,$FE,$00,$C7,$13,$55,$E3,$8F,$31,$FE
	.dc.b	$11,$7C,$4E,$0F,$F5,$59,$D8,$F2,$5F,$A0,$E5,$3F,$11,$FF,$6C,$B6
	.dc.b	$7F,$C1,$73,$C4,$FF,$2B,$CF,$C0,$8E,$33,$FB,$09,$C5,$7A,$5F,$F4
	.dc.b	$CC,$7F,$79,$3E,$89,$FE,$F8,$2B,$F0,$0F,$44,$7F,$94,$5F,$01,$C8
	.dc.b	$F9,$F7,$87,$FC,$3F,$1E,$07,$8C,$FF,$29,$8F,$11,$AB,$F3,$CF,$D5
	.dc.b	$9F,$F1,$11,$E4,$9F,$79,$C1,$FF,$80,$2F,$AE,$3E,$05,$FC,$E9,$FC
	.dc.b	$5B,$F8,$28,$FA,$C7,$FB,$C2,$BB,$BF,$E7,$E8,$F9,$B3,$FE,$90,$AF
	.dc.b	$94,$F0,$9D,$4F,$BA,$7C,$11,$FE,$01,$8F,$87,$1F,$E5,$03,$DB,$1F
	.dc.b	$2E,$7F,$47,$5E,$57,$F4,$A7,$58,$FE,$99,$5E,$47,$F2,$CE,$DF,$CC
	.dc.b	$1F,$24,$FA,$67,$FB,$4A,$78,$57,$E0,$A7,$C7,$3F,$E3,$1A,$FC,$EC
	.dc.b	$75,$8F,$E1,$CF,$57,$81,$E3,$9F,$E2,$55,$D0,$F8,$B1,$F9,$69,$FA
	.dc.b	$6F,$E1,$23,$F2,$17,$F8,$1B,$F8,$97,$F5,$EB,$F8,$14,$7C,$D3,$E2
	.dc.b	$3F,$7C,$FC,$73,$F9,$03,$C4,$7F,$8B,$67,$A4,$FE,$2E,$3F,$09,$3A
	.dc.b	$8F,$DC,$1F,$85,$39,$C7,$77,$CE,$FE,$2E,$FE,$11,$F8,$88,$FC,$9B
	.dc.b	$F8,$79,$FB,$6F,$EC,$EA,$BD,$23,$EA,$BD,$2B,$F1,$0F,$C2,$AF,$C2
	.dc.b	$CF,$9E,$7F,$91,$BF,$34,$F8,$27,$F1,$23,$A1,$5D,$3F,$8F,$3F,$84
	.dc.b	$1F,$92,$8F,$C6,$9F,$31,$FC,$2B,$F0,$9F,$C2,$BD,$1F,$F1,$FB,$E7
	.dc.b	$77,$3F,$83,$BF,$85,$1F,$6E,$3A,$3F,$8B,$8F,$C1,$47,$39,$FE,$AD
	.dc.b	$5F,$87,$9F,$81,$5E,$68,$FC,$0B,$FC,$00,$E6,$7F,$07,$8F,$CB,$C2
	.dc.b	$3A,$3F,$CF,$8F,$E2,$35,$F9,$77,$FD,$3F,$1F,$C2,$AB,$99,$FC,$26
	.dc.b	$FF,$63,$3F,$C2,$0F,$C0,$1F,$85,$AF,$BE,$7D,$3A,$FB,$27,$E7,$DF
	.dc.b	$BD,$5F,$82,$7D,$C7,$F9,$56,$7F,$85,$BF,$4D,$FE,$B7,$3E,$93,$A1
	.dc.b	$F6,$CF,$B3,$7F,$B4,$7D,$CF,$EB,$DF,$F2,$54,$7D,$E9,$FD,$A4,$7B
	.dc.b	$A3,$FA,$9B,$F8,$7A,$FB,$91,$FD,$5A,$B9,$47,$C9,$3E,$E4,$FF,$5C
	.dc.b	$7C,$D3,$FE,$0D,$AF,$2C,$FF,$50,$9F,$E6,$35,$F0,$23,$F0,$0F,$43
	.dc.b	$E0,$57,$F2,$37,$EE,$BF,$00,$FB,$A7,$F4,$68,$FD,$B9,$E6,$7F,$7C
	.dc.b	$FD,$C3,$CA,$7F,$47,$AF,$B9,$5E,$43,$FA,$3C,$7E,$EE,$BF,$0B,$1F
	.dc.b	$84,$38,$9F,$D3,$27,$ED,$9F,$08,$FD,$C3,$F6,$63,$F9,$1C,$7F,$18
	.dc.b	$9F,$E7,$35,$F7,$E3,$E2,$C7,$D9,$3E,$D3,$F3,$1F,$EC,$91,$FC,$E2
	.dc.b	$FF,$95,$C7,$E8,$4F,$CB,$47,$F8,$7A,$7E,$CD,$FA,$E7,$F9,$B4,$FD
	.dc.b	$7A,$FE,$67,$3C,$4F,$5C,$FF,$5A,$3F,$7B,$7F,$B2,$AF,$EF,$D1,$FA
	.dc.b	$DB,$F9,$8F,$F3,$19,$FA,$D1,$F1,$1F,$AB,$3F,$DF,$E3,$F0,$AF,$AA
	.dc.b	$7F,$97,$4F,$EC,$1F,$A9,$1F,$62,$7E,$99,$C2,$BF,$5B,$1F,$D3,$6F
	.dc.b	$F5,$B1,$F4,$E3,$F5,$91,$FD,$FE,$7D,$11,$FC,$89,$FD,$A9,$F8,$D7
	.dc.b	$F5,$37,$F5,$E3,$ED,$9E,$8B,$FA,$47,$E3,$27,$F9,$3B,$F4,$27,$F5
	.dc.b	$4F,$D1,$3E,$49,$F6,$8F,$05,$FF,$24,$7F,$51,$1F,$B1,$8F,$A1,$3F
	.dc.b	$AB,$AF,$E4,$07,$E4,$67,$F1,$35,$FA,$2B,$FE,$F7,$1F,$89,$8F,$ED
	.dc.b	$D3,$F3,$CF,$03,$FE,$56,$3F,$35,$5F,$3D,$FD,$A3,$FC,$F9,$FE,$41
	.dc.b	$1F,$C5,$9F,$CD,$C7,$F5,$DB,$FF,$0D,$3F,$C0,$4F,$EB,$11,$FA,$3B
	.dc.b	$FA,$37,$F5,$67,$E8,$D7,$F8,$86,$FE,$8D,$FF,$1C,$3F,$AB,$D7,$E8
	.dc.b	$A3,$F8,$D9,$F4,$0F,$4C,$7E,$81,$FC,$EB,$FC,$62,$FE,$84,$7E,$82
	.dc.b	$FF,$8B,$47,$F8,$F6,$7C,$C7,$F5,$03,$E7,$BF,$9E,$AF,$D5,$5F,$E7
	.dc.b	$8E,$EF,$9A,$7F,$C1,$E7,$CE,$BF,$CE,$BF,$D3,$DC,$5F,$E8,$5F,$E8
	.dc.b	$51,$F3,$8F,$CA,$9C,$0F,$9B,$5F,$E0,$F9,$FC,$E5,$FF,$82,$2B,$F3
	.dc.b	$95,$F2,$EB,$FC,$0D,$7F,$2E,$3F,$3B,$3F,$C3,$0F,$EF,$EF,$F0,$78
	.dc.b	$F9,$47,$E6,$6F,$F8,$54,$FF,$38,$9F,$E0,$07,$F5,$43,$F0,$15,$FE
	.dc.b	$9B,$7F,$C6,$E7,$E6,$23,$F8,$39,$F1,$9F,$93,$1F,$DF,$2B,$E4,$D7
	.dc.b	$E5,$EF,$FC,$D9,$1F,$26,$BF,$9D,$1F,$97,$AF,$E7,$77,$FD,$D6,$7F
	.dc.b	$9C,$CF,$F8,$DE,$3E,$49,$FC,$11,$F9,$31,$FC,$0C,$C5,$FC,$8A,$FC
	.dc.b	$99,$F9,$C3,$F8,$19,$F3,$CF,$F3,$C5,$7F,$04,$8F,$C9,$D7,$F8,$5A
	.dc.b	$3C,$D3,$FE,$93,$8F,$95,$7F,$94,$9F,$F1,$5D,$7C,$A8,$FC,$64,$FF
	.dc.b	$81,$8F,$F3,$95,$7F,$8A,$AF,$FB,$44,$7F,$6F,$AF,$F3,$43,$67,$8D
	.dc.b	$EA,$3F,$98,$CF,$FB,$0D,$4F,$E6,$4F,$F1,$23,$FB,$63,$11,$A8,$FF
	.dc.b	$3E,$9F,$DA,$23,$FD,$E8,$78,$2B,$FD,$19,$1F,$52,$B9,$C7,$D0,$9F
	.dc.b	$F0,$23,$FE,$D2,$AE,$91,$FC,$FA,$7A,$47,$F0,$78,$FE,$14,$69,$FE
	.dc.b	$DC,$F9,$9F,$E9,$EF,$F5,$77,$FD,$CB,$5F,$E6,$F8,$E6,$FC,$98,$FF
	.dc.b	$1F,$BF,$6A,$BE,$4C,$7F,$A0,$CF,$E5,$E7,$FB,$0A,$3F,$4D,$5F,$CE
	.dc.b	$CF,$A4,$7B,$9F,$E1,$CF,$C8,$3D,$D5,$FE,$A1,$9F,$41,$FC,$EA,$3F
	.dc.b	$2D,$1F,$EA,$33,$FC,$B2,$7F,$BC,$67,$EE,$47,$E0,$27,$ED,$CF,$B2
	.dc.b	$7F,$CD,$33,$F6,$DF,$C2,$47,$B0,$E6,$FD,$A7,$D8,$7F,$AC,$67,$C6
	.dc.b	$FF,$AC,$E7,$F1,$35,$FC,$FA,$7F,$C7,$2F,$B6,$7E,$A3,$A9,$FF,$0B
	.dc.b	$C7,$F9,$7A,$BE,$F4,$7C,$A8,$F9,$C7,$F8,$C6,$3F,$9C,$BF,$5C,$E3
	.dc.b	$5E,$B7,$FC,$C8,$7D,$63,$F3,$2F,$AC,$FF,$16,$47,$DF,$38,$9F,$EE
	.dc.b	$A3,$F6,$73,$FA,$23,$D5,$1F,$3D,$F1,$4F,$EC,$5F,$AC,$7A,$8F,$C9
	.dc.b	$4F,$DD,$7F,$4D,$1F,$30,$E3,$3F,$E5,$C7,$FC,$AF,$1F,$55,$FA,$D3
	.dc.b	$FA,$E8,$FD,$54,$7D,$88,$E3,$3F,$DE,$A3,$FC,$E4,$7D,$33,$F5,$B1
	.dc.b	$FB,$19,$F9,$D1,$FA,$B7,$8B,$FE,$33,$AF,$EE,$0F,$FA,$38,$FD,$75
	.dc.b	$7D,$C7,$F5,$B3,$FA,$A7,$E9,$99,$3F,$BA,$4F,$D7,$9F,$9F,$1F,$2C
	.dc.b	$FA,$73,$FA,$4A,$FD,$41,$E2,$8F,$E1,$95,$9A,$FA,$F3,$C5,$FA,$F5
	.dc.b	$F4,$CF,$D3,$9C,$6B,$FD,$38,$FA,$8F,$D3,$A7,$E9,$CF,$EB,$1F,$A7
	.dc.b	$1F,$A2,$39,$1F,$D2,$67,$EA,$3F,$94,$3F,$18,$E2,$3F,$4D,$1F,$4C
	.dc.b	$F8,$47,$FA,$16,$BF,$55,$3F,$5C,$CC,$7F,$7A,$3E,$59,$F4,$CE,$31
	.dc.b	$F4,$8F,$49,$E8,$7F,$8D,$4F,$E8,$DF,$9A,$78,$27,$F1,$35,$FA,$47
	.dc.b	$E8,$99,$8F,$D1,$1F,$EF,$C3,$84,$7E,$86,$3E,$79,$89,$F9,$D3,$FD
	.dc.b	$01,$FD,$C4,$7F,$AC,$9F,$9C,$7E,$7C,$FA,$CF,$E9,$5F,$9B,$1F,$D3
	.dc.b	$8F,$82,$7F,$09,$3E,$61,$C0,$FC,$D3,$F3,$1F,$CF,$CF,$F4,$9A,$F9
	.dc.b	$71,$F9,$C8,$FE,$8F,$3F,$9C,$8F,$E8,$87,$CA,$7F,$A2,$9F,$C3,$27
	.dc.b	$E5,$57,$F1,$28,$FC,$D4,$FC,$AA,$FE,$79,$5F,$99,$8F,$E7,$D5,$F2
	.dc.b	$6B,$F3,$11,$F9,$43,$E4,$CF,$E5,$DF,$EB,$E7,$FB,$CE,$BD,$8F,$F6
	.dc.b	$79,$F9,$6F,$F3,$F3,$F0,$EF,$CA,$AF,$9C,$F0,$8F,$CD,$BF,$D8,$DF
	.dc.b	$EB,$07,$F7,$63,$F9,$E9,$F2,$E3,$F3,$47,$CE,$8F,$95,$3F,$DB,$27
	.dc.b	$F2,$2F,$F3,$93,$F3,$87,$77,$FD,$B9,$1F,$2E,$11,$FD,$A4,$F9,$95
	.dc.b	$FC,$DE,$BF,$85,$1F,$3E,$BF,$29,$3F,$D9,$4F,$C8,$4F,$CB,$AF,$CA
	.dc.b	$1F,$DA,$67,$E5,$CF,$E4,$E3,$E4,$D7,$F3,$4A,$F9,$91,$FE,$24,$8F
	.dc.b	$CA,$57,$C9,$AF,$95,$3F,$DA,$6B,$F8,$25,$7C,$A3,$E6,$D7,$C9,$AF
	.dc.b	$E6,$07,$E4,$A7,$F9,$9C,$FC,$8A,$FE,$5C,$7C,$79,$FC,$8B,$FE,$14
	.dc.b	$9F,$C8,$BF,$E1,$69,$F9,$75,$F1,$EB,$E3,$4F,$F8,$56,$BE,$34,$7F
	.dc.b	$2A,$8F,$EB,$73,$F1,$63,$E3,$47,$E4,$23,$F1,$F1,$F9,$08,$F8,$D1
	.dc.b	$FB,$78,$F8,$D1,$FE,$0D,$8F,$8F,$5F,$E0,$C8,$FE,$B2,$FC,$67,$FC
	.dc.b	$13,$3F,$8C,$3E,$1D,$7C,$67,$FC,$21,$3F,$1A,$BF,$6F,$5F,$80,$8F
	.dc.b	$E4,$33,$F1,$E7,$FA,$A4,$E6,$7F,$A8,$4F,$E2,$4F,$F6,$51,$FE,$11
	.dc.b	$3E,$49,$FE,$08,$AF,$92,$FF,$82,$A3,$E4,$D7,$EE,$23,$F7,$35,$F2
	.dc.b	$63,$FA,$8C,$7C,$98,$F0,$3F,$D5,$E3,$F7,$2F,$CA,$8F,$F0,$1D,$7E
	.dc.b	$3E,$3F,$C0,$35,$FB,$9A,$F9,$55,$FE,$6E,$9F,$DC,$C7,$CA,$7F,$C7
	.dc.b	$D1,$F2,$8F,$F1,$A9,$F2,$9F,$C7,$BF,$E3,$4A,$F9,$4F,$F8,$DA,$BE
	.dc.b	$55,$7F,$77,$C3,$E2,$CF,$FB,$B9,$B7,$FA,$CB,$60,$9C,$CA,$7F,$AC
	.dc.b	$46,$30,$FD,$9B,$AC,$38,$BF,$4A,$BF,$73,$1C,$DF,$F3,$96,$1F,$BA
	.dc.b	$70,$7F,$91,$5E,$D8,$62,$34,$7F,$1A,$8F,$14,$E6,$39,$47,$F9,$E6
	.dc.b	$BD,$6F,$F6,$DA,$D4,$FF,$81,$AB,$D2,$FF,$85,$67,$FC,$F6,$E8,$FF
	.dc.b	$7E,$E1,$EA,$74,$FF,$77,$38,$BE,$43,$FD,$16,$FF,$9F,$5E,$51,$E6
	.dc.b	$9E,$51,$FE,$23,$C3,$F9,$4E,$1F,$49,$F6,$B6,$3E,$0B,$6F,$BC,$7C
	.dc.b	$2B,$57,$E8,$E3,$BB,$FE,$0E,$C3,$D5,$1A,$9F,$F0,$76,$1A,$7E,$86
	.dc.b	$1E,$17,$BC,$FC,$3B,$F5,$D6,$DF,$BB,$19,$9F,$99,$5F,$76,$BA,$9F
	.dc.b	$00,$FF,$0E,$4F,$03,$FC,$E0,$F6,$6E,$1E,$C7,$EC,$C7,$C9,$9F,$CA
	.dc.b	$BE,$8C,$3D,$73,$F2,$0F,$F5,$D1,$EA,$2D,$7C,$E7,$E3,$1F,$E5,$8A
	.dc.b	$FC,$B3,$F5,$0F,$8C,$FE,$55,$FF,$61,$D7,$8D,$FE,$9B,$71,$F5,$1D
	.dc.b	$1F,$53,$6B,$83,$F9,$5C,$36,$FE,$50,$EB,$87,$D4,$9F,$8F,$87,$EC
	.dc.b	$9F,$8F,$3F,$88,$7F,$1D,$37,$0F,$CF,$D6,$DF,$C0,$D6,$5F,$80,$FF
	.dc.b	$B0,$CF,$BF,$1F,$8B,$7C,$67,$FB,$76,$B6,$D9,$FC,$4B,$D2,$7F,$25
	.dc.b	$1A,$8F,$F6,$6B,$EE,$C3,$F1,$91,$D0,$FF,$64,$56,$E3,$F2,$0E,$F0
	.dc.b	$FF,$77,$9B,$8F,$CA,$57,$5A,$FC,$4B,$F8,$93,$EE,$3F,$86,$C3,$17
	.dc.b	$E1,$9F,$C3,$4F,$EE,$5F,$C3,$61,$F6,$E3,$89,$FC,$46,$FF,$12,$FC
	.dc.b	$A3,$F0,$18,$7A,$5F,$E2,$4A,$37,$5F,$72,$7F,$04,$7E,$4A,$3F,$0B
	.dc.b	$5F,$89,$7A,$47,$EE,$63,$15,$CD,$D5,$73,$BF,$BF,$1F,$C9,$5F,$D3
	.dc.b	$DF,$E1,$67,$9D,$FF,$0D,$9F,$7B,$FE,$AB,$7F,$D7,$35,$C6,$7E,$FC
	.dc.b	$FF,$60,$BF,$E1,$6D,$3C,$E5,$AF,$64,$7F,$09,$AF,$E6,$CF,$F6,$DB
	.dc.b	$FE,$12,$FE,$BE,$BF,$8C,$47,$CC,$7F,$A9,$9E,$D3,$C0,$FA,$1F,$EA
	.dc.b	$F1,$F3,$4F,$C0,$C7,$43,$CF,$1F,$E6,$CB,$E8,$FF,$43,$7E,$1D,$7F
	.dc.b	$0D,$7A,$46,$5F,$CB,$3F,$DF,$9F,$86,$D7,$AB,$FB,$D8,$FC,$23,$E7
	.dc.b	$3E,$43,$F8,$7A,$FC,$14,$7C,$18,$FC,$54,$FE,$06,$3F,$CD,$27,$E0
	.dc.b	$4E,$47,$F4,$47,$F0,$93,$CA,$BF,$98,$57,$E0,$27,$EE,$57,$E3,$EF
	.dc.b	$DA,$FF,$85,$9F,$E9,$91,$F8,$1B,$D3,$CA,$74,$FB,$1F,$DF,$1F,$34
	.dc.b	$FB,$A7,$F4,$D3,$E3,$9F,$D1,$63,$47,$A1,$FE,$0D,$1F,$82,$BF,$A3
	.dc.b	$1F,$81,$34,$7F,$B8,$23,$EC,$4F,$DB,$9F,$24,$FE,$12,$7F,$4C,$7E
	.dc.b	$71,$FA,$47,$23,$FD,$0F,$5A,$8F,$B3,$1F,$81,$AD,$3F,$C0,$AF,$E8
	.dc.b	$CF,$F3,$78,$F2,$DF,$F8,$1E,$FF,$87,$47,$F3,$D3,$FA,$1D,$FF,$0D
	.dc.b	$7F,$9E,$9F,$C0,$6B,$F8,$C3,$F6,$AF,$FC,$0B,$5C,$AF,$E3,$9F,$E0
	.dc.b	$99,$FF,$0A,$D7,$F1,$59,$FE,$6F,$1F,$7C,$FE,$34,$FF,$36,$AF,$E6
	.dc.b	$8F,$F2,$53,$F6,$F1,$FD,$DD,$FE,$2A,$FE,$CE,$7F,$92,$3E,$CB,$FF
	.dc.b	$18,$61,$FC,$CA,$B3,$1F,$C9,$CF,$5C,$FC,$FA,$FE,$F4,$E7,$0F,$EA
	.dc.b	$6F,$F8,$8A,$7F,$BC,$E1,$FD,$F5,$D5,$7F,$2F,$3D,$73,$FD,$D0,$FE
	.dc.b	$68,$FA,$DF,$E5,$8F,$C4,$3D,$6F,$F7,$67,$FC,$02,$FF,$69,$7E,$6B
	.dc.b	$F5,$8F,$5C,$FF,$92,$2F,$35,$F5,$67,$89,$FD,$FA,$BF,$BA,$5F,$13
	.dc.b	$D7,$1F,$DC,$5F,$AA,$FF,$4A,$37,$1F,$54,$FD,$23,$FD,$E2,$BF,$24
	.dc.b	$FF,$8D,$67,$FC,$8F,$1F,$D5,$4F,$D7,$DF,$F6,$BB,$FB,$77,$E4,$BF
	.dc.b	$F1,$E4,$7F,$21,$7F,$C6,$0F,$F2,$C7,$A3,$EE,$7F,$C5,$07,$E2,$9F
	.dc.b	$E5,$8E,$23,$FA,$F4,$7F,$83,$9F,$B3,$1F,$E3,$3A,$FE,$D1,$3F,$62
	.dc.b	$3F,$C5,$CF,$D8,$C3,$FC,$77,$5F,$62,$7F,$C8,$4F,$B9,$AF,$C8,$7B
	.dc.b	$57,$F6,$78,$FD,$7C,$69,$FE,$E0,$FA,$CF,$EC,$D7,$FD,$A6,$7F,$90
	.dc.b	$47,$03,$FA,$0C,$7E,$D0,$FE,$C0,$FF,$1F,$3F,$72,$7A,$4F,$EC,$47
	.dc.b	$F7,$1B,$CE,$1F,$5A,$3F,$D1,$D7,$FC,$41,$FF,$3D,$9F,$C0,$AF,$EA
	.dc.b	$BF,$C8,$2B,$F9,$19,$99,$E5,$1F,$E2,$23,$F9,$11,$FC,$2A,$3F,$82
	.dc.b	$3F,$C8,$DF,$E4,$35,$FD,$72,$38,$C7,$F0,$37,$FA,$C9,$C6,$3F,$94
	.dc.b	$BF,$E6,$98,$FE,$40,$66,$7F,$8D,$47,$F8,$25,$FE,$76,$FF,$5B,$AF
	.dc.b	$E3,$D3,$99,$FE,$4F,$3F,$E8,$77,$ED,$3F,$EA,$5B,$FE,$AB,$7F,$E1
	.dc.b	$6B,$FA,$37,$FD,$52,$B8,$9F,$D9,$EB,$E8,$BF,$D4,$EB,$FC,$29,$5F
	.dc.b	$D5,$DF,$B1,$1F,$98,$7F,$C3,$17,$FE,$11,$3E,$84,$FA,$27,$FC,$1D
	.dc.b	$1F,$02,$3F,$7F,$1F,$CB,$E3,$C2,$FF,$AA,$5F,$94,$FD,$09,$F9,$93
	.dc.b	$FE,$0C,$8F,$09,$FD,$46,$3E,$84,$FF,$4F,$8F,$E9,$47,$CF,$9F,$EA
	.dc.b	$57,$F3,$AB,$FC,$0D,$5F,$3A,$BF,$42,$FF,$5E,$7E,$53,$F5,$8F,$3A
	.dc.b	$3F,$D3,$F3,$FD,$32,$3E,$7C,$70,$BF,$F6,$DB,$FC,$4E,$FF,$A5,$BF
	.dc.b	$C3,$8F,$4D,$FF,$84,$F0,$F9,$37,$FC,$5A,$7F,$A7,$4F,$EB,$CF,$E2
	.dc.b	$11,$FC,$6B,$0C,$BF,$C5,$4F,$E8,$87,$F5,$AB,$FE,$5D,$5F,$C5,$5F
	.dc.b	$CF,$B6,$3C,$31,$FD,$22,$3F,$BE,$47,$F4,$87,$FA,$21,$E1,$9F,$CF
	.dc.b	$BF,$C5,$8F,$4B,$FD,$21,$FE,$37,$5F,$B1,$3F,$CF,$2F,$F1,$1B,$FE
	.dc.b	$25,$5F,$EB,$F8,$FF,$63,$3F,$D1,$CE,$07,$F0,$F7,$FA,$15,$FF,$9F
	.dc.b	$E3,$D3,$1F,$C4,$AF,$FA,$5C,$7F,$2B,$8F,$CC,$1F,$EB,$83,$E8,$57
	.dc.b	$F7,$DA,$FE,$8B,$1F,$DF,$E7,$F8,$AD,$F9,$E3,$F6,$C7,$F3,$B3,$FD
	.dc.b	$DF,$5F,$D2,$2B,$FB,$EC,$FD,$89,$FF,$57,$E1,$FE,$65,$BF,$A3,$1F
	.dc.b	$CF,$AB,$FA,$14,$7F,$80,$1F,$C4,$D7,$FB,$DE,$FF,$2F,$3F,$EA,$4A
	.dc.b	$F9,$F5,$FC,$D9,$FE,$7C,$7C,$E3,$DC,$7F,$A2,$AB,$E7,$61,$E8,$8F
	.dc.b	$D9,$4F,$F3,$E9,$FE,$8C,$7D,$63,$8C,$7E,$66,$BF,$2E,$FF,$47,$3E
	.dc.b	$5C,$FC,$F8,$FE,$E8,$7E,$58,$FC,$34,$7F,$97,$CF,$F5,$14,$FF,$0B
	.dc.b	$8F,$F5,$0D,$7F,$0B,$3F,$9B,$CF,$F2,$A3,$CC,$7C,$D9,$FE,$17,$3F
	.dc.b	$E5,$C8,$FF,$2E,$C7,$E8,$B0,$FF,$48,$47,$F9,$58,$FE,$69,$7F,$C3
	.dc.b	$6F,$F9,$BD,$7F,$A6,$EF,$F8,$25,$7F,$07,$BF,$F4,$7B,$FC,$BA,$BF
	.dc.b	$B3,$BF,$CD,$EB,$E8,$CF,$FA,$51,$FE,$8D,$3F,$EB,$59,$FE,$11,$5F
	.dc.b	$E5,$47,$FC,$AB,$3F,$C2,$23,$F8,$3C,$7F,$31,$3F,$81,$3F,$C1,$CF
	.dc.b	$E2,$AF,$F0,$68,$FE,$1B,$1F,$D8,$23,$F9,$81,$F1,$AF,$F9,$9B,$F2
	.dc.b	$A7,$FD,$8B,$3F,$E8,$38,$FE,$65,$7F,$EC,$3A,$FE,$65,$7F,$E8,$33
	.dc.b	$FB,$0D,$7F,$31,$AF,$91,$7F,$EB,$FA,$F9,$11,$FE,$81,$8F,$91,$3F
	.dc.b	$CB,$A3,$F9,$7C,$7F,$2E,$8F,$EB,$55,$FD,$72,$FE,$5C,$7F,$2D,$AF
	.dc.b	$F5,$71,$F1,$A3,$E5,$4F,$F2,$D7,$E6,$4F,$F2,$D7,$FD,$5D,$3F,$1A
	.dc.b	$FF,$AC,$BF,$CC,$2B,$E4,$D7,$F8,$10,$FF,$1F,$C7,$EE,$AB,$FC,$7D
	.dc.b	$1F,$E7,$79,$FF,$1F,$3F,$BA,$3F,$6B,$3F,$CA,$2B,$FC,$E0,$7E,$E2
	.dc.b	$7F,$93,$E1,$F0,$2F,$F9,$D5,$7F,$23,$3F,$71,$08,$FF,$37,$DD,$9F
	.dc.b	$F5,$95,$7F,$8E,$A3,$FC,$5F,$1F,$B2,$3F,$CD,$C7,$F2,$73,$F7,$31
	.dc.b	$FC,$AA,$3F,$C6,$B8,$7E,$E0,$FF,$3A,$1F,$CB,$4F,$DB,$C7,$F8,$CE
	.dc.b	$BF,$DA,$93,$FE,$2D,$AF,$DA,$DF,$8D,$FF,$28,$47,$F2,$47,$FC,$D1
	.dc.b	$5F,$C9,$23,$FC,$63,$1F,$C7,$4F,$F3,$41,$FD,$2C,$FE,$76,$7C,$2B
	.dc.b	$F8,$73,$F3,$2B,$F6,$53,$F1,$8D,$1F,$BA,$3F,$CD,$18,$7C,$EB,$FE
	.dc.b	$A9,$1F,$E6,$89,$F9,$CF,$FA,$6C,$F4,$1F,$E9,$78,$FE,$B9,$3E,$98
	.dc.b	$FE,$BB,$5F,$36,$BF,$6B,$1F,$E6,$03,$F9,$2B,$FE,$99,$AE,$71,$FE
	.dc.b	$93,$3F,$93,$3F,$38,$1F,$E6,$2C,$3D,$51,$FE,$99,$C3,$E8,$61,$A9
	.dc.b	$FF,$4D,$4F,$F5,$4B,$FD,$BD,$F6,$8F,$F0,$05,$F4,$8F,$F6,$39,$FD
	.dc.b	$BA,$3F,$CC,$B7,$FE,$63,$C3,$E9,$5F,$19,$F0,$94,$F9,$97,$FC,$A7
	.dc.b	$0C,$C7,$CF,$AF,$E5,$53,$FC,$0A,$7F,$C6,$67,$F9,$91,$FE,$52,$F0
	.dc.b	$9F,$F6,$61,$CA,$3F,$C5,$6F,$FA,$AE,$3F,$95,$BE,$1B,$FF,$1A,$3F
	.dc.b	$75,$FA,$4F,$F9,$A2,$BF,$96,$CF,$F5,$DC,$3F,$83,$3F,$CB,$27,$FC
	.dc.b	$C4,$7F,$2D,$3F,$8F,$CF,$F8,$94,$CC,$FF,$07,$9F,$E5,$D3,$FC,$1C
	.dc.b	$F7,$CF,$FA,$BE,$3F,$84,$DF,$F2,$D7,$F8,$3C,$FF,$2E,$8F,$F1,$AD
	.dc.b	$FF,$08,$9F,$81,$7F,$7A,$FF,$84,$BF,$36,$FF,$D2,$07,$C3,$9F,$F4
	.dc.b	$84,$7F,$56,$8F,$E5,$D1,$FE,$8F,$7E,$55,$FF,$A1,$DF,$E8,$4F,$CE
	.dc.b	$7F,$D1,$C7,$F5,$68,$FA,$0F,$FA,$BD,$F9,$A7,$F5,$88,$FE,$61,$3F
	.dc.b	$3C,$FE,$87,$7E,$78,$FD,$2D,$7E,$12,$FC,$EF,$EC,$8F,$F2,$91,$F8
	.dc.b	$33,$E7,$3F,$CC,$DF,$F7,$F5,$7F,$32,$7F,$99,$CF,$E3,$67,$FD,$4D
	.dc.b	$1F,$E3,$47,$F9,$A3,$F4,$23,$FD,$3B,$3F,$E2,$F8,$FE,$46,$FE,$11
	.dc.b	$AF,$F5,$B9,$FF,$47,$1F,$3A,$FE,$44,$FF,$89,$CF,$E1,$B7,$FE,$5E
	.dc.b	$3F,$C5,$F1,$F9,$38,$FF,$23,$1F,$C6,$2B,$FC,$BC,$FF,$8B,$70,$FF
	.dc.b	$17,$47,$F0,$A9,$E1,$87,$F8,$62,$97,$FE,$18,$8F,$F5,$DC,$7E,$B5
	.dc.b	$FF,$78,$DD,$9F,$CD,$C7,$0B,$FF,$2F,$9C,$F0,$FF,$78,$5F,$F0,$93
	.dc.b	$F8,$6E,$1F,$A9,$8F,$E4,$B1,$FC,$01,$FF,$04,$BF,$E6,$19,$FE,$1C
	.dc.b	$7F,$0D,$7F,$7B,$49,$FF,$24,$D7,$38,$FE,$7B,$3F,$C3,$63,$F9,$ED
	.dc.b	$7A,$4F,$A5,$1F,$EE,$CC,$31,$1F,$E3,$2B,$FE,$17,$5F,$42,$BF,$84
	.dc.b	$9F,$88,$3F,$C3,$07,$F3,$D3,$FD,$1D,$1F,$CE,$E3,$F2,$F7,$FE,$5F
	.dc.b	$3F,$B0,$BF,$3B,$0F,$E7,$4F,$D0,$9F,$EC,$EF,$F0,$13,$F8,$64,$79
	.dc.b	$A7,$E7,$C7,$F6,$63,$37,$F3,$8F,$EC,$A7,$FA,$2E,$FE,$B5,$FF,$AA
	.dc.b	$6F,$FC,$BC,$7F,$31,$8F,$93,$1F,$39,$FE,$79,$3F,$3A,$BE,$54,$7F
	.dc.b	$AC,$5F,$99,$5F,$E5,$F3,$94,$7F,$3E,$BF,$F7,$63,$F3,$CF,$E5,$B5
	.dc.b	$F8,$93,$D4,$7F,$39,$7F,$D1,$EF,$D0,$BF,$9F,$1F,$BE,$3F,$C6,$F3
	.dc.b	$FC,$DD,$FE,$03,$1F,$CD,$AF,$EA,$47,$E8,$DF,$E6,$C7,$F8,$DE,$BF
	.dc.b	$CB,$CF,$D2,$9F,$F1,$BC,$7F,$05,$9F,$F1,$BC,$FF,$3B,$BF,$E6,$71
	.dc.b	$F9,$9A,$FE,$66,$FF,$46,$8F,$F3,$CB,$F8,$F3,$FC,$B2,$7F,$08,$3F
	.dc.b	$85,$3F,$EE,$78,$FE,$11,$1F,$EB,$EA,$FE,$11,$5F,$CD,$2F,$F8,$41
	.dc.b	$FE,$67,$3F,$9D,$9F,$D2,$A7,$F3,$8F,$F0,$89,$FC,$4C,$7C,$C7,$C0
	.dc.b	$FF,$A3,$5F,$E1,$0F,$F8,$F6,$3F,$B5,$CF,$CB,$8F,$96,$FF,$34,$3E
	.dc.b	$75,$7C,$B8,$FE,$5B,$5F,$71,$FD,$69,$F8,$13,$F9,$B8,$AF,$E6,$CF
	.dc.b	$EE,$4F,$E6,$C7,$F3,$C8,$FF,$32,$3F,$3E,$3F,$9A,$D7,$CF,$9F,$E6
	.dc.b	$A7,$BE,$FE,$6D,$FE,$B6,$FE,$71,$FD,$82,$3E,$63,$FE,$66,$BF,$F0
	.dc.b	$BB,$FC,$ED,$FF,$45,$9F,$11,$F9,$31,$EB,$7F,$BF,$47,$CE,$9F,$55
	.dc.b	$FF,$9A,$2F,$FD,$1B,$5F,$D8,$63,$FA,$FC,$7C,$C6,$CF,$F3,$39,$F8
	.dc.b	$F3,$FC,$8D,$FE,$09,$5F,$E5,$EB,$FE,$69,$3F,$E6,$99,$FF,$71,$61
	.dc.b	$FE,$3E,$7F,$D6,$87,$F9,$A5,$FE,$67,$87,$F0,$38,$FF,$1E,$3F,$C0
	.dc.b	$DF,$F1,$E3,$FC,$02,$BF,$DB,$E7,$B6,$3F,$C7,$67,$F3,$3A,$FE,$60
	.dc.b	$FF,$99,$CF,$E5,$E7,$F0,$03,$FC,$CE,$7F,$06,$3F,$9B,$5F,$F9,$9D
	.dc.b	$FC,$01,$FB,$DC,$3F,$96,$BF,$C2,$23,$F9,$6D,$FF,$2F,$A6,$1F,$90
	.dc.b	$AF,$E6,$B1,$FC,$00,$FE,$97,$1F,$CB,$B0,$FF,$1C,$4F,$F3,$09,$FF
	.dc.b	$34,$9F,$E9,$27,$E6,$1F,$CB,$63,$FA,$EB,$F9,$43,$FA,$B1,$FE,$03
	.dc.b	$3F,$2F,$7F,$A6,$7F,$DC,$11,$FC,$B9,$CE,$1F,$CB,$9F,$DE,$57,$F9
	.dc.b	$37,$0F,$E5,$71,$FE,$94,$7F,$97,$3F,$E9,$23,$E6,$47,$F2,$E3,$E4
	.dc.b	$3F,$30,$F8,$C7,$CC,$8F,$D2,$47,$F2,$B8,$F9,$07,$CC,$3F,$CA,$73
	.dc.b	$F3,$1F,$9D,$7F,$31,$F9,$D3,$F3,$2B,$FA,$C9,$FE,$4F,$7E,$65,$E6
	.dc.b	$7E,$7C,$7F,$59,$7E,$66,$1F,$E3,$57,$FC,$5C,$7E,$94,$FE,$DE,$7F
	.dc.b	$87,$EB,$F8,$1D,$FF,$98,$63,$F7,$E7,$FB,$88,$FF,$1B,$DF,$FB,$89
	.dc.b	$FD,$F9,$E3,$BF,$E5,$8F,$EF,$EB,$F7,$D3,$FE,$E2,$7F,$80,$61,$FB
	.dc.b	$09,$FA,$35,$F9,$97,$FC,$B4,$7F,$2C,$3F,$80,$3F,$EE,$47,$F8,$04
	.dc.b	$FF,$B9,$23,$F9,$5C,$7F,$8B,$4F,$F2,$F1,$FC,$0E,$7F,$C6,$2F,$F0
	.dc.b	$58,$FF,$72,$3F,$CA,$4F,$F7,$23,$FE,$E3,$7F,$94,$4F,$CE,$7F,$94
	.dc.b	$3F,$CA,$6F,$FA,$94,$7F,$A0,$6B,$F9,$43,$FC,$1A,$7F,$D0,$EF,$F2
	.dc.b	$69,$FE,$4A,$FE,$BE,$3F,$A9,$47,$F0,$18,$FF,$52,$DF,$F8,$49,$FF
	.dc.b	$14,$1F,$14,$FF,$4C,$BF,$D5,$0F,$F3,$83,$FE,$5D,$AD,$3C,$AF,$F3
	.dc.b	$4F,$F9,$F8,$F9,$55,$FE,$55,$7F,$CB,$67,$CA,$3F,$CB,$77,$FE,$86
	.dc.b	$7F,$CB,$6F,$B6,$3F,$AA,$57,$CA,$9F,$F5,$DB,$FA,$F7,$FD,$5B,$5E
	.dc.b	$93,$FC,$B4,$FF,$53,$BF,$F4,$43,$FD,$5E,$7F,$94,$3F,$1C,$FD,$D1
	.dc.b	$FA,$33,$FD,$BE,$7E,$FE,$7F,$94,$5F,$F9,$6E,$3F,$7F,$7F,$E2,$B7
	.dc.b	$F9,$44,$7F,$8A,$E7,$FC,$B5,$1F,$C0,$9F,$D2,$3F,$29,$F8,$AF,$F8
	.dc.b	$5E,$FF,$C5,$87,$F8,$4D,$FF,$67,$BF,$E5,$A3,$E7,$CF,$F8,$BA,$7F
	.dc.b	$80,$3F,$D7,$E7,$F8,$01,$FB,$A3,$F8,$01,$FC,$B6,$3F,$80,$3F,$BC
	.dc.b	$3D,$C7,$F9,$76,$BF,$DC,$71,$FE,$36,$7F,$C1,$93,$FC,$AF,$0F,$E5
	.dc.b	$55,$FE,$79,$38,$9F,$E5,$78,$FD,$85,$FF,$2A,$7F,$65,$7F,$CA,$8F
	.dc.b	$E5,$33,$C2,$FF,$AA,$5F,$F2,$CB,$FE,$57,$3F,$E8,$BA,$F8,$EF,$F0
	.dc.b	$AB,$FF,$41,$BF,$38,$FE,$5B,$3F,$B7,$BF,$F7,$0C,$7F,$99,$1F,$F4
	.dc.b	$44,$7F,$A1,$CF,$9B,$5F,$D0,$2B,$F9,$5D,$7D,$18,$FF,$2C,$5F,$F7
	.dc.b	$A8,$F9,$47,$CC,$07,$FA,$0C,$F9,$A7,$F2,$E8,$F9,$AF,$F2,$AB,$FE
	.dc.b	$80,$DB,$FF,$2B,$4F,$CD,$C3,$E1,$1F,$E1,$97,$DF,$3D,$8F,$F1,$53
	.dc.b	$ED,$3F,$96,$DF,$F5,$77,$FC,$AD,$1F,$CB,$30,$FF,$16,$3F,$CB,$4F
	.dc.b	$D9,$9F,$8F,$7F,$C5,$AF,$FB,$80,$FF,$03,$BF,$C1,$9F,$F5,$7B,$FE
	.dc.b	$57,$3F,$CA,$CF,$F2,$C3,$E9,$1F,$D5,$8F,$F2,$B9,$FD,$62,$FF,$96
	.dc.b	$3F,$D6,$AF,$C2,$FF,$84,$6B,$F3,$EF,$EE,$0F,$E0,$75,$FC,$65,$FE
	.dc.b	$06,$D9,$E3,$1F,$CC,$A7,$F9,$A3,$FE,$52,$3F,$22,$5C,$3F,$CB,$37
	.dc.b	$AA,$FF,$22,$1F,$C1,$E7,$FC,$5F,$5F,$CC,$1F,$E0,$EF,$F8,$D4,$FF
	.dc.b	$37,$9F,$AE,$7F,$83,$57,$E8,$1F,$F7,$FD,$FF,$31,$3F,$CB,$0F,$FA
	.dc.b	$08,$F9,$87,$F0,$B3,$F8,$33,$FE,$82,$7F,$AB,$CF,$FA,$09,$FB,$C7
	.dc.b	$F6,$E8,$FF,$5A,$C7,$F9,$59,$FF,$2A,$47,$C2,$3F,$DB,$57,$F3,$9F
	.dc.b	$E1,$57,$98,$FF,$2B,$C7,$CC,$3F,$98,$3F,$E5,$69,$F9,$97,$F9,$D9
	.dc.b	$FF,$28,$9F,$2A,$BF,$CD,$4F,$F5,$83,$F9,$6C,$7F,$95,$4F,$F4,$2D
	.dc.b	$FC,$98,$F4,$87,$F9,$7B,$5F,$F2,$B3,$F2,$AB,$FA,$D3,$F1,$4F,$F6
	.dc.b	$C9,$FA,$63,$F9,$7D,$FF,$95,$A3,$F9,$75,$FF,$B6,$67,$FC,$F5,$5F
	.dc.b	$E5,$DB,$E1,$1F,$B4,$7F,$93,$E1,$FE,$58,$8F,$F1,$69,$FE,$58,$3F
	.dc.b	$BE,$CF,$F9,$4D,$FF,$16,$1F,$EF,$8C,$3F,$97,$47,$F0,$2C,$3F,$BE
	.dc.b	$1F,$CB,$67,$FC,$A9,$3F,$ED,$89,$FF,$2B,$61,$FB,$A9,$FF,$2B,$C7
	.dc.b	$F1,$38,$FF,$79,$D7,$FB,$5A,$FF,$96,$5E,$70,$FE,$5B,$87,$F8,$B2
	.dc.b	$7F,$1C,$FF,$01,$8F,$94,$FF,$27,$C3,$FA,$23,$FC,$1D,$FF,$16,$5F
	.dc.b	$F2,$EA,$FF,$13,$4F,$FB,$84,$FF,$40,$9F,$EE,$F7,$F9,$74,$FF,$5A
	.dc.b	$8F,$F2,$C3,$FC,$41,$FC,$9D,$7F,$5B,$7E,$75,$FF,$95,$DE,$91,$E8
	.dc.b	$6C,$F1,$7F,$DF,$8F,$EF,$9F,$54,$FF,$30,$3F,$D0,$B1,$FE,$83,$C3
	.dc.b	$FA,$14,$7E,$1E,$FF,$29,$5F,$CB,$4F,$E6,$07,$F2,$F8,$F9,$8F,$84
	.dc.b	$FE,$BD,$5F,$CC,$1F,$91,$5F,$41,$FF,$42,$4F,$F3,$0C,$3F,$2F,$1F
	.dc.b	$D7,$E7,$F9,$8E,$1F,$99,$C3,$E8,$CF,$E2,$F0,$FE,$5F,$7F,$D7,$1F
	.dc.b	$E5,$EF,$CF,$3F,$97,$9F,$C0,$E3,$A9,$FC,$7D,$FE,$5E,$7F,$96,$8C
	.dc.b	$BF,$CF,$63,$F8,$DB,$E7,$3F,$98,$9F,$C1,$A3,$F9,$89,$FE,$36,$3F
	.dc.b	$82,$C7,$F9,$71,$FE,$0B,$5F,$ED,$A9,$FE,$60,$7F,$8C,$D4,$7F,$AD
	.dc.b	$E3,$E0,$DF,$FB,$6C,$FC,$74,$7F,$31,$C3,$F8,$34,$FF,$8D,$E7,$F8
	.dc.b	$81,$FE,$FB,$9F,$F1,$A1,$FE,$33,$8F,$E0,$C7,$FB,$6A,$3F,$97,$3F
	.dc.b	$D6,$5F,$F1,$A3,$FC,$1A,$FF,$C6,$87,$F2,$F3,$F8,$2B,$FC,$CA,$3F
	.dc.b	$C6,$6F,$F2,$E3,$E3,$9F,$E8,$51,$5F,$21,$FC,$11,$FE,$EF,$AD,$BF
	.dc.b	$91,$3F,$86,$BF,$CC,$5F,$AB,$7C,$AF,$F9,$94,$62,$FF,$97,$3E,$83
	.dc.b	$F9,$AD,$FF,$98,$CF,$E3,$A7,$F9,$86,$7C,$31,$FD,$76,$7F,$99,$47
	.dc.b	$86,$3F,$D1,$27,$F2,$D3,$F8,$B5,$FC,$F8,$FF,$45,$3F,$CD,$0F,$E1
	.dc.b	$F1,$FE,$89,$3F,$CC,$27,$E3,$25,$7F,$E6,$28,$FB,$A6,$54,$7D,$87
	.dc.b	$C5,$1F,$C2,$8D,$9F,$D8,$0F,$F4,$7C,$FE,$52,$BF,$B0,$47,$D3,$9F
	.dc.b	$D1,$BF,$D5,$A3,$F9,$F6,$3F,$E6,$18,$FF,$11,$BF,$E6,$3C,$3F,$DD
	.dc.b	$ED,$93,$FC,$EA,$69,$FC,$43,$1A,$FF,$AD,$E3,$FC,$7A,$E0,$FF,$1D
	.dc.b	$57,$47,$FC,$7F,$59,$7F,$93,$1F,$DF,$8C,$1F,$D0,$CF,$F1,$65,$61
	.dc.b	$FE,$2B,$1E,$A7,$FA,$03,$FE,$2F,$AF,$C2,$D7,$F5,$B3,$C1,$3F,$E4
	.dc.b	$57,$83,$FE,$47,$C7,$E0,$BC,$AB,$6F,$F8,$75,$DB,$FE,$4A,$3C,$58
	.dc.b	$7D,$67,$F3,$78,$62,$BF,$94,$E1,$F9,$39,$D3,$FE,$F8,$9F,$54,$7D
	.dc.b	$EA,$FE,$DD,$3F,$99,$7D,$51,$FC,$82,$7A,$47,$E7,$E7,$55,$E8,$AF
	.dc.b	$F7,$ED,$FA,$1F,$BC,$F4,$7F,$DF,$26,$CF,$E8,$AF,$C5,$BF,$5B,$F1
	.dc.b	$CA,$7A,$23,$FE,$00,$BD,$5F,$E0,$4F,$C2,$7F,$C4,$35,$F8,$99,$C4
	.dc.b	$79,$CF,$E6,$F7,$FC,$2D,$CD,$73,$7F,$B3,$3F,$39,$FC,$1C,$7C,$27
	.dc.b	$FA,$BE,$1B,$8F,$F5,$5C,$F8,$5F,$34,$FF,$5A,$AF,$9B,$5F,$E1,$9A
	.dc.b	$FA,$EF,$F8,$76,$7E,$11,$FD,$72,$3E,$CD,$FE,$21,$F4,$47,$E1,$4E
	.dc.b	$2F,$F4,$D8,$FF,$2F,$3E,$78,$FE,$BC,$7E,$04,$E6,$7C,$0C,$3F,$06
	.dc.b	$D3,$CE,$7F,$60,$BF,$C1,$C7,$F8,$0E,$BF,$05,$3E,$63,$FD,$79,$7F
	.dc.b	$06,$7F,$84,$BE,$18,$FF,$32,$E1,$D1,$FE,$D2,$F9,$DF,$EB,$78,$7C
	.dc.b	$3B,$FE,$52,$FC,$38,$FB,$F5,$F3,$63,$F0,$53,$A7,$A4,$FF,$6B,$7A
	.dc.b	$3F,$C3,$DF,$BE,$F9,$CF,$F2,$89,$F0,$8F,$F0,$D1,$F3,$0F,$F2,$3D
	.dc.b	$EE,$FF,$B5,$61,$FA,$49,$F8,$47,$E2,$0E,$77,$FD,$01,$F8,$31,$F8
	.dc.b	$1C,$3E,$15,$7F,$AE,$EB,$4F,$9A,$BF,$CB,$B3,$F0,$5F,$E0,$E7,$85
	.dc.b	$FD,$14,$FE,$05,$E6,$7F,$98,$EF,$58,$73,$AF,$F3,$FC,$7F,$B2,$A3
	.dc.b	$C2,$FE,$8A,$3E,$F4,$FD,$77,$E0,$1F,$E1,$C3,$E5,$9F,$C7,$E7,$F4
	.dc.b	$4F,$F0,$37,$9D,$FE,$86,$3F,$DA,$18,$7F,$02,$AC,$DF,$C9,$3F,$92
	.dc.b	$E1,$F4,$F0,$F8,$67,$DD,$7F,$DF,$75,$F7,$A3,$F7,$B3,$F4,$5F,$F1
	.dc.b	$C4,$FF,$70,$7E,$5D,$FA,$CF,$B3,$7F,$B1,$7F,$36,$7D,$CA,$FB,$4F
	.dc.b	$C3,$BF,$B5,$87,$F1,$18,$FF,$27,$9A,$7F,$08,$FF,$44,$BF,$DD,$56
	.dc.b	$5F,$87,$7F,$5E,$7F,$DF,$87,$D2,$7F,$81,$1F,$96,$7F,$A0,$CF,$D6
	.dc.b	$9F,$96,$FD,$57,$F6,$91,$F1,$CF,$09,$F4,$E7,$EC,$BF,$52,$BF,$9E
	.dc.b	$4F,$27,$D5,$1F,$E6,$CA,$FC,$9B,$FC,$D7,$0F,$E7,$17,$F5,$23,$F9
	.dc.b	$BD,$FD,$C3,$D4,$FF,$30,$7E,$E3,$FC,$81,$F1,$BF,$AD,$8F,$B3,$1E
	.dc.b	$33,$EA,$C7,$F8,$A9,$FA,$AF,$23,$FB,$31,$C8,$FD,$8C,$6A,$BE,$AC
	.dc.b	$7E,$64,$F1,$57,$F7,$63,$F4,$F3,$F7,$6F,$F2,$F3,$FD,$C5,$F8,$8F
	.dc.b	$E3,$8F,$F5,$8C,$FF,$27,$7E,$A6,$1F,$C9,$AB,$F5,$38,$7C,$B7,$D3
	.dc.b	$3F,$E5,$83,$E5,$9F,$A3,$3E,$A3,$FA,$73,$E9,$E1,$F5,$E7,$E9,$1F
	.dc.b	$A4,$9F,$D0,$61,$EF,$3F,$B5,$57,$D1,$C3,$FB,$55,$7D,$13,$BD,$7F
	.dc.b	$19,$3F,$40,$F1,$3F,$14,$FF,$B8,$AB,$FD,$C1,$7A,$9F,$DA,$CF,$E2
	.dc.b	$EB,$FC,$50,$FF,$5E,$BF,$D2,$3F,$20,$FE,$43,$5F,$42,$B8,$DF,$8A
	.dc.b	$7F,$DE,$AF,$D9,$6D,$7D,$13,$F6,$31,$F4,$0F,$B1,$1F,$E2,$09,$FE
	.dc.b	$A6,$FD,$0B,$FE,$31,$1F,$D5,$AF,$F4,$58,$7F,$4C,$AF,$05,$FF,$85
	.dc.b	$DF,$9F,$1F,$D5,$F0,$F9,$F1,$FD,$66,$FE,$79,$FE,$10,$7E,$7C,$7F
	.dc.b	$51,$AF,$71,$FE,$19,$BF,$A7,$5F,$3C,$FC,$CD,$7C,$EA,$FE,$99,$1C
	.dc.b	$0F,$E5,$91,$FE,$0E,$7E,$6B,$E6,$3F,$88,$1F,$ED,$A3,$FD,$C1,$87
	.dc.b	$F9,$4C,$FF,$29,$1F,$D0,$DF,$8A,$7F,$55,$7F,$38,$FF,$49,$BF,$C3
	.dc.b	$3F,$DA,$27,$F8,$7B,$FE,$09,$7F,$87,$9F,$1A,$BF,$94,$47,$F9,$31
	.dc.b	$FE,$1F,$1F,$01,$F4,$D7,$F8,$1E,$3F,$87,$CF,$CE,$BF,$4B,$FC,$37
	.dc.b	$0F,$9D,$3F,$E0,$59,$FC,$E4,$7F,$80,$30,$E0,$7E,$0C,$FC,$DB,$FC
	.dc.b	$FE,$7F,$C8,$33,$F3,$A3,$FC,$93,$1F,$C3,$2B,$F8,$51,$F3,$E3,$F9
	.dc.b	$F5,$FF,$79,$AF,$CD,$47,$F6,$D9,$F9,$C7,$73,$E6,$9F,$DB,$CF,$BA
	.dc.b	$0F,$E0,$27,$CE,$3F,$BA,$9F,$CF,$0F,$EE,$4F,$F7,$3A,$F9,$B5,$FD
	.dc.b	$C9,$F9,$A7,$F0,$49,$FC,$C1,$F3,$5F,$CC,$3F,$E2,$E7,$E5,$9F,$CD
	.dc.b	$1F,$CB,$C7,$F9,$AF,$0F,$CB,$BF,$E6,$B3,$FD,$E7,$7E,$49,$FE,$6D
	.dc.b	$87,$CB,$8F,$F0,$3C,$FF,$36,$BF,$97,$5F,$CD,$B0,$F9,$71,$FA,$63
	.dc.b	$F0,$31,$FC,$0D,$FF,$02,$BF,$E5,$A7,$F8,$3B,$FC,$EE,$BE,$6D,$7F
	.dc.b	$B6,$23,$FC,$44,$7C,$93,$E5,$5F,$D1,$C3,$FC,$71,$1F,$E6,$39,$FF
	.dc.b	$6F,$D7,$FB,$11,$FE,$E5,$7F,$DA,$A7,$99,$FC,$22,$3F,$DD,$77,$FC
	.dc.b	$3E,$3F,$C4,$07,$73,$F8,$D5,$73,$70,$7F,$40,$29,$F7,$6F,$DE,$FF
	.dc.b	$89,$AF,$CF,$7A,$8D,$47,$79,$FD,$0C,$65,$FF,$13,$5F,$F9,$C2,$7F
	.dc.b	$8F,$4F,$F0,$F3,$F8,$91,$E3,$3F,$A1,$4F,$C2,$BF,$C7,$1F,$20,$B5
	.dc.b	$FC,$66,$B4,$FF,$27,$AF,$F5,$41,$CC,$F6,$CF,$EF,$6F,$F0,$73,$FE
	.dc.b	$2E,$9F,$E2,$CF,$EB,$EB,$E8,$9F,$02,$3F,$D4,$E7,$33,$FD,$53,$7A
	.dc.b	$8E,$47,$BA,$BF,$C5,$CF,$EA,$E7,$EA,$9F,$20,$F9,$07,$F8,$98,$FF
	.dc.b	$80,$07,$EC,$6F,$B3,$FE,$AA,$9F,$B6,$FD,$EB,$E4,$FF,$A4,$E3,$EF
	.dc.b	$CE,$4E,$31,$FE,$09,$8E,$F7,$EB,$8F,$C8,$3F,$E9,$FA,$FA,$8F,$E4
	.dc.b	$0F,$B3,$1F,$82,$7D,$8F,$F7,$C3,$CE,$7D,$33,$C6,$7C,$E3,$F2,$15
	.dc.b	$F6,$1F,$E9,$CF,$91,$FF,$14,$DF,$F1,$D7,$EE,$57,$27,$FC,$D2,$F2
	.dc.b	$3E,$B4,$FF,$A6,$AF,$EC,$D7,$D1,$AF,$D5,$BE,$B8,$FF,$48,$5F,$D5
	.dc.b	$AC,$CF,$FB,$AE,$7E,$A9,$FB,$3A,$FA,$D7,$FB,$19,$FA,$51,$F4,$E3
	.dc.b	$F9,$24,$7E,$D2,$3F,$C7,$A7,$FA,$2A,$7F,$5F,$79,$AF,$C6,$4F,$9C
	.dc.b	$FB,$EF,$E8,$23,$E9,$47,$EA,$0F,$88,$FF,$08,$7F,$6C,$FD,$27,$F6
	.dc.b	$17,$AB,$FD,$4B,$FD,$2A,$3F,$DD,$07,$CD,$7F,$18,$FF,$A4,$8F,$92
	.dc.b	$7C,$13,$F1,$F1,$FA,$87,$E8,$9C,$6B,$E8,$4F,$D4,$31,$3F,$3E,$7F
	.dc.b	$91,$9F,$6D,$F4,$3F,$C6,$DF,$43,$FC,$3D,$FC,$4B,$FC,$31,$FE,$A8
	.dc.b	$FF,$09,$BF,$E4,$95,$F4,$5F,$E3,$F1,$F4,$5F,$E3,$E7,$E8,$9F,$E4
	.dc.b	$11,$FE,$8A,$BF,$A1,$7F,$A1,$8F,$F4,$6D,$FE,$6C,$FE,$09,$3F,$85
	.dc.b	$AF,$E2,$91,$FC,$5C,$FE,$05,$1E,$68,$FE,$3D,$1F,$C4,$5F,$E1,$95
	.dc.b	$FC,$DE,$BF,$C5,$15,$98,$FD,$11,$F4,$67,$EA,$CF,$03,$F3,$8F,$F3
	.dc.b	$F8,$FE,$24,$FF,$29,$51,$F9,$E3,$EB,$47,$F3,$28,$FE,$B0,$FF,$12
	.dc.b	$3E,$8C,$7F,$54,$9F,$CF,$47,$F5,$69,$FA,$95,$FC,$BA,$3F,$0C,$7F
	.dc.b	$96,$6F,$F8,$8B,$F8,$63,$F7,$0F,$F4,$EB,$FC,$FC,$FF,$4D,$9F,$9F
	.dc.b	$3F,$9E,$8F,$F5,$CD,$7C,$F8,$FE,$9F,$1F,$3E,$7F,$AE,$DF,$CF,$7F
	.dc.b	$85,$47,$CF,$AC,$BE,$E8,$F9,$EF,$F4,$27,$FA,$1D,$FC,$E8,$FE,$13
	.dc.b	$1F,$C2,$A3,$FA,$04,$7F,$09,$8F,$E1,$51,$FE,$27,$AF,$E8,$75,$F9
	.dc.b	$98,$FE,$79,$5F,$98,$AF,$93,$7F,$C1,$A3,$E4,$47,$C9,$9F,$EF,$F1
	.dc.b	$FB,$F8,$FE,$98,$7E,$06,$7E,$4B,$FD,$B4,$FE,$3C,$FC,$93,$F2,$EF
	.dc.b	$F0,$03,$F9,$C1,$FC,$84,$FC,$09,$FD,$A6,$7F,$83,$9F,$0D,$FF,$06
	.dc.b	$4F,$F6,$77,$E3,$9F,$C0,$6F,$F3,$97,$F2,$8F,$E0,$93,$F9,$79,$FE
	.dc.b	$68,$7E,$4E,$7F,$2C,$FC,$9B,$FF,$11,$BF,$96,$3F,$80,$DF,$C9,$AF
	.dc.b	$F2,$85,$FC,$9A,$FE,$5F,$1F,$25,$FE,$B2,$FD,$DA,$FF,$04,$BF,$C0
	.dc.b	$9F,$F7,$64,$FF,$30,$3F,$27,$1F,$2E,$3F,$B1,$9F,$0C,$F9,$47,$F2
	.dc.b	$83,$E1,$1F,$D8,$4F,$94,$7E,$48,$FE,$57,$7F,$C0,$1F,$C9,$DF,$F0
	.dc.b	$09,$F9,$11,$F2,$63,$FA,$EC,$7F,$84,$6B,$E4,$47,$F8,$4E,$3E,$44
	.dc.b	$FE,$42,$FE,$2D,$FC,$87,$E3,$DF,$EF,$6B,$F9,$45,$7E,$42,$7F,$C1
	.dc.b	$75,$F9,$08,$F9,$37,$F9,$09,$F8,$F5,$FE,$15,$3E,$44,$FF,$82,$9F
	.dc.b	$C8,$C6,$AF,$E3,$C7,$F8,$26,$7F,$AA,$CF,$E4,$2F,$F2,$95,$FE,$0C
	.dc.b	$3E,$55,$7F,$51,$AF,$C7,$DF,$F2,$58,$FD,$BC,$7C,$6B,$FD,$BC,$7C
	.dc.b	$67,$FC,$0A,$FF,$16,$8F,$F1,$DD,$FE,$39,$F9,$15,$F1,$E7,$F6,$D5
	.dc.b	$F8,$A9,$FC,$74,$7E,$DA,$3E,$3C,$FE,$D9,$FF,$17,$CF,$E3,$9F,$8F
	.dc.b	$3F,$8E,$8F,$8F,$1F,$8E,$3E,$3B,$FD,$E0,$FE,$37,$5F,$C7,$A3,$E3
	.dc.b	$57,$ED,$AB,$F8,$BC,$7F,$91,$AF,$FD,$89,$7F,$E6,$D8,$C3,$FC,$E4
	.dc.b	$FF,$4D,$C7,$F9,$46,$BF,$C4,$74,$AD,$D7,$F9,$D0,$FD,$E8,$AE,$0F
	.dc.b	$FA,$86,$58,$71,$8D,$4F,$FA,$0E,$70,$F0,$BF,$24,$E8,$F2,$2C,$39
	.dc.b	$E1,$E2,$9F,$F8,$03,$7C,$AF,$FD,$C1,$7E,$39,$C1,$CA,$78,$BD,$4F
	.dc.b	$6A,$36,$FF,$A4,$2F,$FD,$B7,$5F,$0D,$FF,$4D,$B6,$7C,$87,$09,$FF
	.dc.b	$73,$47,$FB,$06,$FC,$D7,$AA,$F0,$CF,$F9,$F6,$BF,$C7,$F5,$A8,$FF
	.dc.b	$49,$36,$3E,$06,$1D,$6F,$D9,$87,$9E,$B5,$7D,$EF,$D8,$7F,$68,$7E
	.dc.b	$95,$F9,$63,$EB,$5F,$BF,$0E,$95,$FD,$BA,$79,$3F,$81,$3F,$D8,$2F
	.dc.b	$49,$E3,$87,$86,$31,$3D,$DE,$73,$F2,$EF,$C1,$7D,$CF,$51,$F3,$2B
	.dc.b	$FD,$61,$3E,$3B,$F8,$B1,$89,$FA,$91,$DE,$3F,$29,$1F,$EA,$A6,$BD
	.dc.b	$F0,$FC,$26,$1F,$E2,$F8,$F9,$0F,$CC,$3D,$47,$E7,$A7,$F4,$75,$D5
	.dc.b	$FD,$29,$F0,$8F,$D3,$CF,$92,$7F,$A8,$D7,$DC,$7E,$41,$DA,$FE,$BC
	.dc.b	$65,$FC,$2B,$6B,$A4,$FE,$4F,$0F,$A9,$87,$C5,$AF,$C3,$BF,$16,$7F
	.dc.b	$C5,$8F,$E4,$2B,$E3,$BF,$13,$0F,$C2,$3F,$0E,$33,$1F,$0E,$7F,$06
	.dc.b	$7E,$38,$FA,$E7,$EB,$A3,$EF,$3E,$83,$33,$F0,$DF,$F1,$E4,$7E,$F8
	.dc.b	$DC,$FC,$38,$FF,$1C,$61,$F8,$BB,$FC,$D3,$D6,$BE,$E9,$F8,$A3,$F1
	.dc.b	$8F,$C4,$9F,$C4,$BF,$B7,$9E,$35,$D6,$7F,$85,$3D,$5F,$C9,$3F,$15
	.dc.b	$1F,$E4,$67,$E3,$3F,$E5,$37,$E3,$47,$EC,$9F,$8D,$19,$7E,$35,$7E
	.dc.b	$CE,$3B,$CF,$DD,$7D,$47,$F2,$53,$E3,$47,$F1,$E7,$D4,$D3,$EE,$BF
	.dc.b	$C9,$4E,$AF,$F1,$F8,$FC,$24,$68,$FD,$85,$74,$9F,$E3,$37,$F4,$67
	.dc.b	$F9,$74,$FC,$57,$F3,$2F,$47,$FD,$47,$7F,$16,$BF,$8A,$47,$59,$FE
	.dc.b	$2B,$5F,$2D,$B5,$F8,$83,$A5,$7F,$1B,$7E,$85,$7E,$1E,$3F,$0A,$F3
	.dc.b	$7F,$50,$FE,$10,$F8,$2F,$DE,$79,$B6,$39,$BF,$55,$FE,$0B,$1C,$E7
	.dc.b	$E0,$3F,$8E,$79,$B6,$75,$3E,$43,$F4,$D1,$F8,$27,$EA,$BF,$82,$9F
	.dc.b	$E1,$58,$7F,$0E,$9E,$73,$97,$F0,$2D,$7E,$53,$FE,$2B,$AF,$B8,$FC
	.dc.b	$93,$DE,$7D,$BB,$FA,$8F,$F0,$08,$D3,$C8,$FC,$66,$1F,$79,$FE,$09
	.dc.b	$3F,$45,$FF,$4E,$CF,$F5,$F3,$B6,$1C,$63,$FA,$35,$FD,$57,$ED,$57
	.dc.b	$E7,$1F,$EA,$91,$EC,$7F,$A7,$9F,$6D,$F9,$EF,$E0,$EB,$EA,$3F,$64
	.dc.b	$F9,$E7,$C7,$7F,$53,$3E,$CA,$FE,$D1,$1B,$AF,$B3,$1C,$EB,$EB,$1F
	.dc.b	$66,$BF,$19,$5E,$C3,$F7,$F3,$F8,$29,$FB,$95,$F6,$4E,$4F,$D1,$3F
	.dc.b	$2C,$FB,$5F,$C3,$57,$8E,$3E,$BC,$7E,$25,$FC,$AB,$FB,$63,$F9,$D9
	.dc.b	$F8,$23,$F9,$9C,$7C,$F7,$E3,$4F,$FB,$35,$FE,$86,$7C,$F3,$F2,$97
	.dc.b	$F5,$CF,$E7,$13,$F5,$E7,$8C,$FF,$5D,$9F,$E7,$95,$FA,$6A,$FC,$9D
	.dc.b	$FE,$C2,$BF,$59,$1F,$AF,$3F,$88,$BF,$DA,$5F,$5D,$FF,$80,$F0,$FE
	.dc.b	$E9,$5E,$B3,$F5,$F5,$FD,$D0,$FA,$E7,$F6,$DA,$FA,$D1,$FC,$CC,$FD
	.dc.b	$74,$F2,$AF,$9B,$5F,$E0,$17,$E2,$1F,$DE,$23,$FB,$AD,$72,$8F,$AD
	.dc.b	$3F,$83,$7F,$25,$7E,$CB,$FE,$E4,$F2,$3E,$B4,$7F,$77,$7F,$5D,$1F
	.dc.b	$69,$F9,$93,$E2,$9F,$E6,$AF,$88,$FD,$59,$F5,$A3,$E9,$CF,$EA,$9F
	.dc.b	$BB,$3C,$67,$FB,$6E,$1F,$53,$0F,$D5,$D7,$F6,$FB,$FD,$4B,$F4,$CF
	.dc.b	$06,$1F,$D1,$6D,$87,$F2,$2B,$FA,$B5,$FC,$86,$FF,$5A,$FE,$C5,$FA
	.dc.b	$47,$EA,$2B,$E9,$5F,$E7,$9F,$C5,$C7,$D1,$7F,$91,$4F,$E9,$8F,$98
	.dc.b	$FE,$96,$BE,$7C,$7F,$89,$27,$F8,$DC,$7F,$89,$23,$E3,$1E,$7B,$FE
	.dc.b	$33,$87,$F5,$78,$FC,$C9,$F8,$83,$FC,$71,$3F,$E3,$BB,$F3,$9F,$D5
	.dc.b	$4F,$E2,$78,$7E,$84,$FD,$E3,$E7,$3F,$AA,$4F,$CE,$33,$7F,$54,$FC
	.dc.b	$79,$FC,$3A,$FF,$C3,$13,$FE,$1D,$9F,$9C,$7F,$07,$AF,$41,$F8,$43
	.dc.b	$F3,$F8,$7F,$99,$4F,$9B,$1F,$C5,$0F,$CC,$CF,$E5,$23,$F0,$77,$FE
	.dc.b	$5C,$9F,$D0,$47,$F1,$1C,$3F,$39,$1F,$32,$BF,$33,$1F,$33,$0F,$F0
	.dc.b	$7D,$7C,$C7,$FC,$83,$5F,$99,$AF,$EF,$C7,$E5,$70,$FE,$C5,$1E,$6A
	.dc.b	$FC,$E9,$F9,$97,$F8,$6C,$FE,$44,$FC,$C1,$F1,$0F,$E4,$EF,$E6,$1F
	.dc.b	$E1,$6F,$CB,$3F,$C8,$4F,$F3,$C9,$FE,$14,$7C,$E9,$FC,$BD,$FF,$90
	.dc.b	$2B,$F2,$E7,$ED,$A7,$E5,$3F,$C6,$1F,$E0,$E7,$E0,$2F,$FC,$BB,$5F
	.dc.b	$80,$BF,$E7,$47,$CB,$3E,$CB,$E6,$8F,$F4,$2C,$7F,$0B,$AF,$CB,$1F
	.dc.b	$DC,$9F,$97,$1F,$E3,$8A,$FE,$09,$5F,$2E,$7F,$C9,$35,$F2,$DF,$E7
	.dc.b	$07,$F6,$D9,$F9,$71,$F9,$63,$F8,$5C,$7F,$37,$BF,$CB,$3F,$CD,$4F
	.dc.b	$F3,$FD,$FF,$9D,$9F,$31,$FE,$56,$3F,$2C,$7F,$08,$3F,$D2,$EE,$DF
	.dc.b	$E3,$A7,$F8,$B5,$E5,$87,$F8,$05,$B7,$FD,$DA,$7E,$2C,$6B,$0F,$E0
	.dc.b	$6F,$0C,$35,$5F,$C1,$EB,$FD,$76,$68,$FF,$34,$BF,$87,$7F,$C5,$F5
	.dc.b	$FE,$39,$7F,$D8,$2E,$CF,$F5,$E4,$7C,$78,$FF,$2A,$C7,$FA,$44,$FF
	.dc.b	$60,$1F,$C9,$2F,$FA,$0C,$7F,$7B,$3F,$B8,$98,$8E,$87,$C3,$7F,$C6
	.dc.b	$8D,$3F,$92,$BF,$E7,$E7,$FC,$97,$7F,$86,$9E,$91,$FD,$66,$75,$1F
	.dc.b	$31,$B7,$F1,$CF,$84,$7F,$B8,$DF,$09,$F4,$DF,$84,$FF,$AB,$67,$E1
	.dc.b	$47,$FB,$A5,$DC,$7F,$AC,$63,$E9,$3F,$06,$3D,$4F,$C1,$AF,$B9,$7A
	.dc.b	$8F,$C4,$9F,$E8,$C8,$FC,$05,$72,$BF,$C5,$4F,$B8,$FC,$43,$F2,$5F
	.dc.b	$75,$FF,$9E,$E3,$DC,$D9,$FB,$E7,$E4,$EF,$DD,$3F,$B3,$73,$3F,$75
	.dc.b	$F6,$47,$ED,$E7,$F1,$93,$EB,$AF,$CB,$3E,$B7,$E9,$1F,$40,$F5,$BF
	.dc.b	$82,$9F,$BF,$3C,$9F,$D2,$47,$AC,$FC,$2B,$F6,$0E,$27,$A9,$FC,$03
	.dc.b	$FA,$28,$FA,$E7,$E8,$67,$F6,$EF,$E2,$CF,$F1,$4B,$FE,$FF,$9F,$AA
	.dc.b	$FF,$AB,$23,$E8,$3F,$9B,$3D,$51,$FE,$A3,$7E,$61,$F5,$0F,$55,$FD
	.dc.b	$49,$FE,$A5,$5F,$50,$F1,$57,$D1,$BF,$B0,$F5,$8F,$55,$FF,$BA,$8F
	.dc.b	$D6,$D7,$D3,$38,$CF,$E7,$EB,$F1,$F3,$FC,$C2,$3E,$A4,$F2,$8F,$F0
	.dc.b	$6C,$FE,$BD,$4F,$D5,$7F,$C8,$07,$C0,$3F,$A7,$5F,$F2,$A7,$EA,$9C
	.dc.b	$CF,$DB,$9F,$C3,$A3,$F5,$A7,$D4,$79,$3F,$E5,$F7,$F0,$37,$F5,$A3
	.dc.b	$2F,$D5,$BF,$E4,$07,$E3,$AB,$FA,$AB,$F5,$4F,$CF,$9F,$8E,$72,$FD
	.dc.b	$4A,$FE,$52,$7A,$4F,$E4,$F5,$F5,$4F,$B0,$FA,$4F,$F1,$7C,$7E,$9D
	.dc.b	$FA,$07,$1A,$FD,$31,$F6,$CE,$0F,$F9,$88,$FA,$67,$CE,$3E,$45,$FD
	.dc.b	$27,$F8,$F9,$FD,$66,$B3,$5F,$1E,$3F,$36,$7D,$A3,$F4,$A7,$E8,$E3
	.dc.b	$F9,$41,$F8,$A7,$F4,$F1,$FA,$37,$35,$C5,$FF,$24,$CF,$CF,$6C,$7C
	.dc.b	$F3,$FC,$0B,$3F,$E2,$B8,$FE,$26,$F1,$47,$F9,$5E,$FE,$8D,$68,$FC
	.dc.b	$24,$FE,$26,$7F,$10,$7F,$2F,$AF,$EB,$37,$FC,$56,$BE,$79,$F4,$E7
	.dc.b	$FA,$99,$F8,$67,$F8,$EC,$F8,$63,$FC,$54,$7F,$16,$8F,$D1,$1F,$4E
	.dc.b	$BF,$89,$C7,$CF,$9F,$E2,$51,$F3,$63,$E7,$47,$F1,$18,$FE,$8F,$1F
	.dc.b	$28,$ED,$3F,$84,$3F,$A2,$3F,$9C,$7F,$2E,$FE,$6E,$7F,$87,$9C,$0F
	.dc.b	$E1,$D1,$F4,$0F,$CC,$CF,$F4,$33,$F3,$27,$AC,$FC,$11,$FC,$30,$FC
	.dc.b	$C4,$FF,$10,$7F,$31,$7F,$DD,$E3,$E3,$5F,$C9,$7F,$C0,$75,$F2,$6B
	.dc.b	$F2,$D1,$FC,$28,$FC,$B4,$FF,$08,$7F,$86,$CF,$C7,$9F,$F1,$44,$7C
	.dc.b	$7A,$FE,$DB,$3F,$C1,$E7,$FB,$55,$7F,$20,$3C,$B1,$FD,$D2,$3F,$2D
	.dc.b	$5F,$24,$F8,$B3,$FC,$F1,$F9,$17,$FC,$DA,$7F,$D5,$91,$F9,$43,$F7
	.dc.b	$B3,$FC,$E6,$BE,$45,$7E,$4A,$FF,$7B,$5F,$92,$3F,$1B,$7F,$BF,$AF
	.dc.b	$C8,$57,$FA,$06,$3E,$3C,$FF,$88,$4F,$C4,$1F,$6C,$FE,$CF,$3F,$90
	.dc.b	$9F,$F1,$5C,$FC,$8A,$FF,$10,$D7,$C8,$8F,$8E,$FC,$89,$FE,$6B,$1F
	.dc.b	$8E,$8F,$93,$1F,$8E,$9F,$93,$1F,$E1,$C3,$F1,$EE,$1F,$E5,$B3,$F2
	.dc.b	$1F,$C8,$CF,$F8,$62,$FE,$45,$FE,$4C,$FC,$7D,$7E,$EE,$BE,$4D,$7F
	.dc.b	$B5,$6B,$E4,$47,$FB,$5E,$BF,$1F,$5F,$B9,$8F,$DD,$57,$FA,$96,$BF
	.dc.b	$25,$5F,$22,$FF,$92,$DF,$F2,$4B,$FE,$4B,$7F,$B6,$8F,$C6,$C7,$E3
	.dc.b	$A3,$E3,$C7,$E3,$63,$F1,$D1,$FD,$62,$FF,$1D,$5F,$E4,$58,$F8,$F5
	.dc.b	$F8,$EA,$F8,$F3,$F0,$DF,$F4,$0D,$7F,$81,$9F,$8E,$7F,$5A,$9F,$C5
	.dc.b	$3F,$8D,$AF,$8D,$1F,$E3,$B8,$FF,$1E,$47,$E3,$6B,$FC,$6D,$1F,$8D
	.dc.b	$8F,$8D,$3F,$16,$3E,$34,$7E,$35,$FE,$FA,$7E,$35,$F8,$D1,$FE,$CB
	.dc.b	$BF,$8D,$3F,$B3,$8F,$83,$97,$F8,$C2,$3F,$DC,$67,$F9,$B5,$C5,$7F
	.dc.b	$6A,$C5,$18,$9F,$F4,$FB,$5F,$ED,$A6,$8F,$E3,$E7,$F2,$EA,$4E,$63
	.dc.b	$AC,$0C,$3F,$E7,$DA,$FE,$6A,$7F,$36,$72,$7F,$99,$4F,$F7,$CE,$3D
	.dc.b	$2F,$C9,$28,$E9,$18,$C3,$F9,$FB,$AB,$E1,$7F,$06,$BC,$51,$C2,$3F
	.dc.b	$9B,$C7,$F9,$7D,$F0,$9E,$47,$D0,$72,$7F,$A2,$38,$8D,$C7,$99,$F8
	.dc.b	$B3,$F3,$27,$FD,$EE,$FD,$19,$F9,$33,$E1,$7F,$A7,$57,$D3,$7F,$A4
	.dc.b	$BF,$D6,$2B,$C1,$1F,$1E,$3E,$9B,$A3,$B9,$67,$6F,$F8,$51,$EC,$F1
	.dc.b	$BF,$9D,$7F,$80,$AF,$9B,$87,$AA,$7F,$C5,$27,$CD,$BC,$4A,$3B,$DF
	.dc.b	$A1,$FB,$95,$F0,$A3,$E6,$D7,$81,$FF,$78,$9D,$5F,$97,$15,$ED,$7F
	.dc.b	$86,$7A,$CF,$E1,$5F,$92,$FC,$93,$D2,$53,$C9,$17,$0F,$04,$7F,$62
	.dc.b	$8D,$61,$A8,$E5,$5A,$7F,$1C,$7A,$4F,$D2,$E1,$4F,$2B,$F6,$5F,$2B
	.dc.b	$F0,$6B,$C9,$1F,$8C,$7E,$F5,$FE,$2D,$B5,$F9,$E6,$C7,$86,$BF,$C5
	.dc.b	$71,$AB,$F3,$47,$FA,$9A,$7E,$5C,$FE,$26,$7F,$0D,$3F,$D1,$0E,$87
	.dc.b	$EF,$5F,$C2,$B9,$7F,$46,$E4,$FB,$B1,$CD,$F4,$1C,$6F,$E6,$BF,$D4
	.dc.b	$1E,$AD,$AF,$45,$7E,$0C,$EB,$1B,$7D,$2F,$FA,$5A,$FD,$C7,$C3,$7F
	.dc.b	$89,$3F,$EB,$0B,$FD,$23,$FA,$B6,$BF,$3A,$3F,$D9,$D7,$F8,$D8,$F8
	.dc.b	$71,$F6,$9F,$CC,$3F,$10,$FE,$18,$7E,$A7,$0C,$BF,$68,$FF,$1E,$1F
	.dc.b	$E7,$EA,$FE,$39,$3F,$A8,$8F,$C8,$3D,$2B,$E1,$C7,$47,$F8,$85,$67
	.dc.b	$0F,$C5,$47,$F1,$AC,$3E,$1D,$6E,$FA,$3F,$BB,$7E,$7B,$FC,$29,$EA
	.dc.b	$FF,$21,$9F,$D8,$57,$06,$CF,$ED,$63,$FD,$DD,$36,$3F,$63,$87,$DB
	.dc.b	$7C,$77,$F0,$E7,$A3,$F0,$E7,$F8,$04,$7C,$33,$F8,$61,$F3,$CF,$B8
	.dc.b	$74,$3F,$09,$3D,$2F,$CF,$1E,$F7,$E1,$CF,$F0,$B7,$A4,$78,$1E,$87
	.dc.b	$F0,$C9,$F8,$71,$FC,$19,$FE,$04,$FC,$83,$E1,$3F,$C1,$A7,$F8,$25
	.dc.b	$FC,$E8,$F8,$31,$FC,$1D,$E7,$3E,$F3,$FC,$E3,$1F,$C0,$E3,$F7,$91
	.dc.b	$ED,$3F,$D1,$31,$FC,$26,$3F,$65,$1F,$D9,$0F,$E0,$07,$E6,$8F,$AD
	.dc.b	$1F,$BE,$BF,$34,$7F,$2B,$8F,$E0,$33,$FB,$BA,$FC,$F4,$FE,$E6,$3F
	.dc.b	$52,$FF,$95,$63,$51,$E3,$7F,$B2,$57,$79,$FF,$17,$CF,$EB,$9F,$9E
	.dc.b	$7F,$B5,$0D,$1F,$ED,$69,$D4,$F9,$23,$FD,$B2,$E9,$FE,$9F,$3F,$CD
	.dc.b	$63,$DD,$5F,$C2,$EF,$CB,$5F,$ED,$AB,$FD,$EC,$7C,$7A,$FD,$F3,$C4
	.dc.b	$FC,$CC,$FF,$77,$3F,$C0,$17,$E6,$7F,$B6,$9E,$17,$F7,$8F,$F0,$07
	.dc.b	$E9,$1F,$02,$3F,$78,$F4,$8E,$87,$4B,$FF,$08,$3E,$D9,$FF,$0B,$47
	.dc.b	$F8,$4E,$FF,$63,$3F,$BB,$3F,$C2,$67,$E6,$4F,$E9,$91,$FE,$08,$79
	.dc.b	$1F,$D8,$1F,$F0,$5C,$72,$9F,$CB,$CF,$F5,$FA,$FC,$BC,$FF,$10,$8F
	.dc.b	$DC,$1E,$37,$F2,$CF,$F8,$2A,$3F,$42,$FF,$82,$5F,$D0,$C7,$F5,$A8
	.dc.b	$FE,$71,$1F,$B6,$7E,$EC,$FF,$8B,$27,$F8,$05,$FF,$3A,$9F,$DA,$BF
	.dc.b	$04,$FF,$17,$61,$F4,$EF,$F6,$93,$9B,$FD,$9B,$FE,$0A,$7C,$6F,$DA
	.dc.b	$AD,$4F,$EC,$8F,$64,$7F,$20,$3E,$73,$EC,$7F,$DC,$4F,$DA,$6B,$FB
	.dc.b	$C3,$D8,$FF,$75,$C3,$FB,$A3,$F9,$13,$D4,$7F,$57,$7F,$B9,$D7,$ED
	.dc.b	$27,$F6,$2F,$E2,$67,$FB,$8C,$FF,$6C,$C3,$37,$F9,$D9,$FE,$21,$3F
	.dc.b	$B0,$BF,$F0,$85,$7E,$C2,$7F,$D8,$47,$E8,$5F,$43,$FE,$C5,$AE,$51
	.dc.b	$F9,$1A,$FE,$EB,$5F,$AF,$AE,$D5,$FC,$CE,$7F,$B3,$57,$EB,$A7,$F5
	.dc.b	$47,$F6,$A3,$FB,$65,$7E,$A1,$E3,$7F,$AC,$7F,$CB,$97,$FE,$45,$BF
	.dc.b	$D8,$3A,$3F,$6C,$7E,$3E,$BF,$47,$3E,$F7,$FB,$33,$FA,$9B,$FE,$D7
	.dc.b	$1F,$AA,$3C,$51,$FA,$83,$F9,$65,$7F,$1D,$8F,$EC,$31,$FA,$03,$F0
	.dc.b	$A7,$F6,$23,$F0,$B7,$E6,$7F,$B3,$C7,$F2,$C8,$FD,$A4,$7F,$9D,$8F
	.dc.b	$E1,$27,$FA,$01,$E6,$FF,$65,$AF,$D5,$5F,$EB,$A7,$F6,$53,$FE,$F5
	.dc.b	$AF,$EC,$F5,$F5,$E3,$FB,$03,$F8,$F7,$FD,$08,$FD,$87,$31,$FA,$A3
	.dc.b	$9B,$FC,$46,$BF,$83,$3F,$E6,$33,$F4,$C7,$90,$FE,$C2,$7B,$9F,$E0
	.dc.b	$CF,$F9,$C0,$E6,$7D,$6A,$F7,$BF,$C8,$0F,$D6,$57,$ED,$23,$91,$FD
	.dc.b	$72,$7F,$83,$61,$FC,$6A,$FF,$AB,$1F,$BC,$3D,$97,$FD,$71,$F9,$E7
	.dc.b	$F5,$A7,$EC,$57,$F9,$A2,$7F,$AB,$C7,$E7,$CF,$C9,$1F,$E6,$1B,$FE
	.dc.b	$AE,$FF,$5B,$07,$E0,$8F,$E5,$D8,$7F,$95,$1E,$55,$F9,$37,$E5,$C7
	.dc.b	$F5,$B7,$FC,$BD,$87,$ED,$E7,$E0,$C7,$FA,$04,$FD,$3E,$1C,$5F,$F0
	.dc.b	$9D,$7F,$95,$DF,$D8,$C7,$85,$FE,$89,$7F,$C3,$1F,$AC,$7F,$45,$BF
	.dc.b	$5B,$FE,$59,$8F,$D6,$57,$F9,$4E,$7F,$47,$3E,$27,$F1,$EF,$F9,$4C
	.dc.b	$FD,$3D,$FF,$1C,$7F,$D2,$F1,$FA,$B8,$FE,$49,$7F,$AA,$9F,$D5,$C7
	.dc.b	$F4,$EB,$FD,$4D,$7E,$B6,$BD,$15,$FE,$D1,$7D,$0F,$F5,$27,$FA,$6B
	.dc.b	$FC,$D8,$FF,$15,$1F,$D3,$A7,$33,$FA,$D9,$FD,$41,$FD,$2E,$7F,$45
	.dc.b	$3F,$E3,$B7,$F8,$11,$FA,$99,$FD,$34,$FF,$4E,$3F,$23,$5F,$46,$7F
	.dc.b	$C7,$D5,$FE,$43,$3F,$48,$FF,$4D,$AF,$EA,$55,$F9,$A8,$FC,$1C,$7E
	.dc.b	$91,$FC,$F1,$FA,$49,$FE,$9A,$FE,$8C,$FC,$D1,$C4,$FF,$1B,$4F,$FB
	.dc.b	$02,$BF,$AA,$D7,$F6,$79,$FE,$99,$5F,$E4,$38,$FE,$97,$87,$E8,$6F
	.dc.b	$FC,$77,$5F,$EB,$DA,$FF,$2D,$47,$A2,$3F,$8F,$57,$F4,$E8,$F0,$47
	.dc.b	$F8,$E4,$FF,$00,$DF,$F8,$E2,$BF,$3F,$5F,$A3,$9F,$AE,$7F,$84,$6B
	.dc.b	$FA,$4B,$F9,$E3,$F3,$53,$FD,$21,$FC,$ED,$7F,$AE,$E7,$FC,$63,$1F
	.dc.b	$9C,$3F,$A3,$3F,$8B,$9F,$DA,$47,$F8,$E6,$B3,$3F,$D1,$63,$FA,$09
	.dc.b	$F9,$69,$FC,$CB,$FE,$AD,$9F,$CC,$CF,$F4,$38,$FF,$14,$1F,$98,$9F
	.dc.b	$F1,$6C,$FE,$5F,$0F,$F1,$95,$7F,$3C,$AF,$CA,$1F,$E2,$97,$F3,$C7
	.dc.b	$91,$FC,$A3,$F0,$9F,$E7,$E7,$FA,$8E,$3F,$C4,$EF,$FA,$82,$BF,$20
	.dc.b	$FF,$3E,$3F,$90,$9F,$EA,$63,$E8,$DF,$E4,$1F,$D7,$9F,$CE,$F0,$FF
	.dc.b	$4F,$5F,$FA,$8F,$0F,$CB,$C7,$F0,$08,$FE,$76,$7F,$86,$AF,$F8,$FD
	.dc.b	$7F,$88,$5C,$D7,$E5,$A7,$FD,$02,$FE,$56,$7F,$D2,$27,$E5,$69,$87
	.dc.b	$FA,$5A,$FF,$C4,$D5,$FC,$F2,$F8,$E1,$FA,$68,$F3,$E1,$FE,$8F,$AF
	.dc.b	$E6,$C7,$E5,$5F,$CD,$57,$F8,$7C,$FD,$69,$FD,$45,$FE,$53,$48,$F4
	.dc.b	$C7,$F7,$39,$FE,$77,$5F,$97,$9F,$F4,$7C,$FF,$85,$CF,$E4,$87,$F8
	.dc.b	$78,$FF,$2C,$C7,$E5,$CF,$F1,$13,$FC,$B0,$F3,$9F,$EC,$BB,$FF,$0F
	.dc.b	$4F,$F3,$28,$FC,$BC,$7F,$87,$4F,$EF,$8F,$FB,$24,$FC,$C5,$FF,$30
	.dc.b	$BF,$F4,$5D,$FF,$36,$3F,$2F,$7F,$96,$AF,$CB,$D7,$F8,$62,$BF,$2F
	.dc.b	$1F,$E0,$7A,$FE,$0F,$3F,$E1,$CA,$FC,$C5,$FF,$A1,$6B,$F3,$11,$FE
	.dc.b	$81,$AF,$CC,$47,$FB,$12,$3F,$97,$4F,$FB,$E2,$3F,$31,$7F,$CB,$9F
	.dc.b	$F3,$ED,$7E,$6E,$BF,$AC,$9F,$CB,$23,$FC,$29,$3F,$96,$3F,$95,$BF
	.dc.b	$E1,$2C,$3F,$71,$3F,$EF,$38,$FE,$57,$1F,$94,$BF,$8B,$87,$FB,$C6
	.dc.b	$7F,$DF,$15,$FC,$AC,$FC,$AB,$FC,$9A,$FF,$95,$47,$F8,$81,$FD,$64
	.dc.b	$7F,$83,$A3,$F2,$AF,$E3,$4F,$E4,$27,$F8,$2C,$ED,$1F,$94,$7F,$2E
	.dc.b	$7F,$7A,$C3,$FC,$EB,$7F,$BC,$AF,$E5,$53,$FE,$0A,$C3,$F9,$65,$FE
	.dc.b	$56,$BF,$DD,$15,$F9,$53,$FC,$13,$5F,$C9,$67,$F9,$4C,$7F,$82,$0F
	.dc.b	$F3,$81,$FC,$A2,$3F,$24,$7E,$3A,$3F,$93,$BF,$EE,$0B,$FC,$8C,$FF
	.dc.b	$26,$8F,$F6,$EE,$1E,$87,$E4,$D7,$F2,$7B,$FF,$6F,$4F,$D0,$7F,$23
	.dc.b	$3F,$D4,$DB,$3F,$E5,$48,$FF,$6F,$3F,$92,$7F,$C0,$77,$F9,$29,$E2
	.dc.b	$FF,$3D,$83,$FE,$02,$7F,$4E,$7F,$7B,$7F,$8A,$BF,$E0,$87,$F9,$5D
	.dc.b	$FF,$2A,$8F,$E5,$77,$FE,$D3,$BF,$E5,$71,$F9,$58,$FF,$00,$DF,$E4
	.dc.b	$E3,$FC,$02,$FA,$AB,$FA,$94,$7E,$FD,$FF,$35,$5F,$E5,$5F,$F4,$A1
	.dc.b	$FE,$68,$8F,$E5,$58,$7F,$2B,$8F,$91,$1F,$CA,$67,$FC,$01,$3F,$1E
	.dc.b	$3F,$CC,$87,$E5,$AB,$FD,$9B,$5F,$CA,$63,$F2,$B3,$FC,$A6,$39,$C7
	.dc.b	$F9,$78,$FC,$B4,$7F,$80,$23,$31,$FD,$FC,$A2,$D8,$7F,$AC,$A3,$FC
	.dc.b	$5F,$3F,$EC,$D7,$F9,$65,$FF,$2B,$8E,$35,$F4,$A7,$FC,$03,$1F,$CB
	.dc.b	$0F,$F6,$5B,$F9,$6B,$FC,$BC,$7F,$80,$F0,$FE,$FF,$7F,$2E,$3F,$95
	.dc.b	$9F,$E5,$F3,$F4,$D5,$FC,$A5,$EE,$7E,$B9,$FE,$5E,$F9,$DF,$E3,$B1
	.dc.b	$FE,$C8,$BF,$F0,$04,$FE,$66,$7F,$35,$1F,$52,$7F,$CB,$91,$F9,$A7
	.dc.b	$F9,$49,$FE,$5B,$3F,$32,$7F,$29,$7F,$94,$E1,$FC,$FD,$FE,$55,$3F
	.dc.b	$CA,$6B,$F2,$F1,$F9,$88,$FA,$95,$F9,$88,$FF,$2C,$1F,$E6,$1B,$FE
	.dc.b	$7C,$FF,$8B,$8F,$95,$1F,$E5,$99,$FE,$55,$5F,$97,$8F,$E5,$51,$F9
	.dc.b	$78,$FE,$55,$3F,$97,$9F,$E0,$C7,$E5,$CF,$E5,$53,$FE,$57,$9F,$E5
	.dc.b	$33,$FC,$AA,$3F,$C5,$D3,$FE,$BD,$3F,$BF,$9F,$C8,$A3,$FC,$AA,$7F
	.dc.b	$AF,$6F,$F9,$5C,$66,$3F,$BF,$D7,$F9,$55,$FC,$19,$FE,$3D,$3F,$C6
	.dc.b	$27,$F9,$58,$B5,$EE,$7F,$D8,$11,$FC,$B4,$FC,$5D,$7C,$D9,$FC,$9C
	.dc.b	$7E,$09,$FF,$43,$CF,$FA,$7A,$7F,$04,$FE,$65,$FC,$13,$F2,$E7,$F3
	.dc.b	$0F,$E5,$AF,$D6,$FF,$A7,$8F,$E6,$31,$97,$F3,$38,$7F,$00,$7F,$99
	.dc.b	$5F,$F9,$60,$F1,$57,$AE,$39,$1F,$A9,$63,$FE,$06,$9F,$9A,$FF,$AD
	.dc.b	$2B,$D8,$7F,$81,$2B,$F8,$45,$FF,$81,$27,$F3,$B3,$FC,$CD,$FF,$2B
	.dc.b	$1F,$CC,$DF,$CD,$61,$FC,$AA,$3F,$B7,$BF,$65,$FE,$5B,$5E,$D8,$FF
	.dc.b	$5C,$3F,$BE,$8F,$F0,$AD,$7F,$35,$77,$87,$FA,$DE,$FF,$CB,$71,$FE
	.dc.b	$B2,$3F,$CB,$73,$FC,$CE,$FF,$9B,$CF,$F9,$1E,$3F,$D3,$B3,$F9,$93
	.dc.b	$FD,$93,$3F,$9F,$7F,$CB,$93,$E1,$AF,$E6,$71,$F4,$5F,$4B,$FE,$14
	.dc.b	$7F,$C7,$4F,$FB,$CE,$FF,$D7,$C7,$E9,$CF,$E7,$87,$E2,$5F,$F0,$83
	.dc.b	$FE,$59,$8F,$E3,$53,$FE,$61,$9F,$E6,$E7,$F8,$81,$FE,$6E,$7E,$21
	.dc.b	$FE,$71,$1F,$CD,$EB,$E5,$47,$E6,$0F,$F5,$B1,$F9,$13,$F6,$F5,$E9
	.dc.b	$7F,$C1,$91,$FC,$15,$FC,$E9,$FC,$DC,$FF,$79,$9F,$87,$3F,$21,$1F
	.dc.b	$E0,$D8,$FD,$73,$FE,$A7,$BF,$96,$FF,$65,$9F,$CF,$5F,$E8,$67,$FD
	.dc.b	$41,$3F,$A4,$BF,$F3,$16,$1F,$C1,$B0,$FF,$0C,$C7,$FA,$86,$7F,$9A
	.dc.b	$57,$FA,$D8,$FF,$07,$4E,$B0,$FF,$5B,$57,$C8,$C3,$FC,$1B,$1F,$CD
	.dc.b	$E7,$9D,$FE,$C9,$B5,$FC,$E2,$7E,$65,$7F,$38,$7F,$C1,$D8,$7F,$AD
	.dc.b	$E3,$F3,$73,$FE,$5F,$8F,$C9,$4F,$F7,$93,$FB,$C1,$FA,$A3,$F9,$19
	.dc.b	$F3,$6B,$F3,$CF,$FA,$54,$FC,$29,$FE,$8F,$7F,$3A,$7F,$38,$7F,$36
	.dc.b	$7F,$98,$0F,$D1,$D7,$E7,$23,$F3,$31,$F9,$C8,$FE,$6F,$1F,$9C,$AF
	.dc.b	$F0,$74,$FE,$4C,$FE,$6E,$7E,$6A,$3F,$9D,$47,$E7,$A3,$FD,$6D,$7F
	.dc.b	$45,$FC,$D3,$F9,$87,$FC,$BF,$3F,$E1,$88,$FC,$C4,$7F,$9F,$1F,$CC
	.dc.b	$1F,$E0,$98,$FE,$71,$1F,$30,$FF,$30,$D7,$F3,$88,$FE,$CD,$5F,$97
	.dc.b	$9F,$F5,$B1,$F8,$E3,$F2,$F8,$7F,$98,$6B,$F9,$9C,$78,$0F,$C8,$9F
	.dc.b	$E4,$C3,$E2,$BF,$E6,$37,$F9,$CD,$FE,$7C,$FF,$5F,$1C,$5F,$F4,$E9
	.dc.b	$FA,$EC,$3F,$AE,$BF,$96,$3F,$C1,$F8,$7F,$90,$AB,$FC,$C8,$71,$8F
	.dc.b	$E6,$CF,$E4,$1F,$F3,$21,$FE,$0C,$3F,$CC,$07,$E3,$CB,$87,$FA,$DE
	.dc.b	$BF,$9A,$47,$FB,$F2,$3F,$D4,$2F,$F3,$1B,$FE,$6A,$71,$74,$F8,$4F
	.dc.b	$F2,$39,$FE,$B5,$9F,$E6,$B1,$F9,$18,$FF,$1D,$3F,$9A,$BF,$F4,$1D
	.dc.b	$FE,$C0,$FE,$9F,$3F,$E3,$F7,$FD,$73,$3F,$CD,$A3,$FC,$1F,$3F,$98
	.dc.b	$BF,$F5,$C4,$7F,$83,$E7,$F9,$A9,$F8,$93,$F8,$0C,$FE,$65,$FD,$A3
	.dc.b	$FE,$10,$7E,$AB,$F9,$87,$F9,$A4,$7F,$83,$E3,$F3,$11,$FE,$0F,$7F
	.dc.b	$9A,$D7,$FB,$45,$FE,$65,$1F,$EB,$87,$FC,$25,$87,$FA,$E2,$3F,$CC
	.dc.b	$F7,$FE,$64,$9F,$F5,$B4,$7F,$99,$1F,$CC,$4F,$F9,$91,$FC,$D4,$FF
	.dc.b	$91,$DF,$F0,$7C,$FE,$5C,$FE,$5E,$7E,$EC,$FE,$BB,$5F,$99,$7F,$83
	.dc.b	$CF,$E6,$5F,$F2,$04,$FF,$84,$1F,$CD,$57,$E6,$63,$FC,$1D,$3F,$E6
	.dc.b	$47,$F9,$74,$FF,$70,$BF,$C9,$BF,$98,$A6,$1F,$EB,$B8,$FE,$63,$3F
	.dc.b	$81,$7F,$31,$5F,$1B,$0F,$F0,$79,$FD,$9E,$3F,$97,$DF,$FA,$F0,$F9
	.dc.b	$4F,$F3,$2C,$3F,$97,$D7,$F8,$EE,$BF,$C1,$AF,$F2,$BA,$FE,$5D,$1F
	.dc.b	$E0,$B9,$FC,$C4,$FC,$6A,$FC,$99,$FE,$B9,$7F,$96,$D7,$FA,$E6,$3F
	.dc.b	$C1,$F5,$FE,$B9,$3F,$C0,$6F,$E4,$CF,$F0,$53,$FE,$B9,$AF,$E5,$95
	.dc.b	$FC,$9A,$7F,$24,$FF,$28,$AF,$F5,$CB,$F9,$3B,$FC,$94,$FF,$8E,$63
	.dc.b	$F9,$4B,$FE,$08,$9F,$F0,$23,$FD,$02,$3E,$7B,$FE,$1E,$8F,$F3,$03
	.dc.b	$FE,$34,$3F,$95,$47,$E4,$8F,$E5,$53,$FC,$12,$3F,$05,$5F,$EB,$73
	.dc.b	$F9,$FB,$FD,$A6,$BF,$24,$FF,$AD,$67,$F9,$53,$FD,$F0,$FD,$31,$FC
	.dc.b	$9C,$FE,$56,$7F,$2E,$9C,$BF,$E0,$77,$F2,$8F,$F3,$2B,$FF,$01,$5F
	.dc.b	$F9,$88,$FF,$47,$5F,$DC,$9F,$CB,$4F,$F8,$0C,$1F,$C1,$A7,$FD,$28
	.dc.b	$EE,$3F,$AF,$C7,$E5,$DF,$F3,$26,$1F,$EB,$B8,$FC,$BC,$7F,$81,$0F
	.dc.b	$DC,$9F,$EB,$A3,$F2,$D1,$F9,$57,$FC,$C7,$7E,$A7,$F9,$6E,$1F,$E0
	.dc.b	$3B,$FE,$89,$1F,$CB,$4F,$CB,$E1,$F9,$58,$FC,$CD,$7E,$52,$7F,$AC
	.dc.b	$47,$E5,$1F,$E5,$73,$F9,$4C,$3F,$C0,$15,$F9,$49,$F8,$D5,$FE,$00
	.dc.b	$9F,$C9,$E1,$FE,$B6,$7F,$BF,$CF,$F9,$6D,$F8,$73,$F9,$48,$FF,$2B
	.dc.b	$DF,$EC,$27,$F9,$3C,$FE,$56,$7F,$CB,$6F,$F2,$67,$FC,$02,$0F,$F2
	.dc.b	$E4,$7F,$27,$7F,$D6,$AF,$F8,$C9,$FE,$B8,$FE,$59,$FE,$98,$FF,$AD
	.dc.b	$5F,$F2,$CE,$1F,$EE,$77,$F8,$85,$7E,$58,$FB,$6F,$E5,$8F,$F5,$CD
	.dc.b	$7F,$53,$7F,$C5,$53,$F9,$77,$FC,$25,$87,$E5,$DF,$F5,$B1,$FD,$2C
	.dc.b	$FE,$4E,$7F,$96,$CF,$F2,$81,$F5,$8F,$CC,$1F,$CA,$4F,$F0,$1D,$7F
	.dc.b	$29,$7F,$31,$3F,$D6,$9F,$F1,$05,$7E,$B0,$F3,$C7,$FA,$DE,$FF,$BF
	.dc.b	$CF,$FA,$DD,$FA,$27,$C5,$7F,$BD,$9F,$E0,$0C,$3F,$24,$FF,$80,$F0
	.dc.b	$FE,$50,$FE,$48,$FF,$2A,$9F,$92,$9F,$F2,$8E,$1F,$9D,$8F,$F0,$2D
	.dc.b	$FF,$27,$3F,$25,$86,$63,$FD,$12,$FF,$27,$7F,$B4,$4F,$FA,$6C,$FF
	.dc.b	$2A,$9F,$D9,$8F,$F5,$9A,$BF,$F1,$CB,$FE,$A6,$3F,$C0,$07,$F7,$B3
	.dc.b	$E7,$1F,$E6,$67,$FD,$0E,$FF,$AB,$27,$F1,$CF,$E6,$2F,$FC,$AA,$FF
	.dc.b	$2B,$3F,$D6,$F5,$FC,$AD,$FF,$00,$CF,$F2,$B9,$FC,$AC,$7F,$2B,$3F
	.dc.b	$C0,$11,$FD,$F4,$FF,$03,$1E,$87,$FC,$AC,$7E,$5E,$FF,$BE,$BF,$EB
	.dc.b	$1B,$FE,$FB,$5F,$4E,$3F,$CA,$87,$F7,$E9,$FF,$2F,$9E,$C7,$FA,$C4
	.dc.b	$FE,$FD,$F7,$47,$E6,$5F,$F3,$15,$FE,$66,$FC,$17,$FE,$B0,$8F,$F5
	.dc.b	$94,$FE,$6A,$FF,$CA,$EF,$F2,$E3,$FC,$AF,$7F,$95,$9F,$E5,$B1,$FE
	.dc.b	$97,$9F,$EF,$CF,$F9,$58,$F1,$1F,$E2,$43,$F9,$69,$FE,$AC,$72,$7F
	.dc.b	$8B,$6B,$FA,$D4,$7E,$62,$FF,$04,$D3,$FC,$AF,$1F,$98,$3F,$CA,$F7
	.dc.b	$FE,$2E,$8F,$E5,$27,$F9,$5C,$F8,$4F,$FB,$CE,$FF,$C0,$11,$F9,$93
	.dc.b	$F9,$51,$FC,$32,$7F,$D4,$EF,$FA,$B4,$FF,$1C,$1F,$EA,$D3,$F9,$4D
	.dc.b	$F4,$7F,$BF,$9F,$C0,$2F,$F3,$4F,$F9,$48,$FF,$2B,$9F,$E0,$07,$F9
	.dc.b	$5B,$FE,$04,$7F,$1E,$FC,$D7,$F9,$63,$F9,$7A,$FE,$58,$FF,$95,$85
	.dc.b	$7F,$AB,$67,$8C,$7F,$9F,$DF,$0D,$FC,$F8,$FF,$57,$1F,$98,$8E,$2F
	.dc.b	$F8,$A6,$3F,$CA,$F7,$FE,$AF,$8F,$EA,$F3,$FE,$AF,$7F,$98,$E1,$FE
	.dc.b	$AF,$BF,$E6,$35,$FE,$04,$9F,$F0,$24,$7F,$81,$6F,$F3,$31,$FE,$35
	.dc.b	$3F,$5A,$7F,$BC,$DF,$F0,$2B,$FC,$C0,$FC,$D4,$7F,$95,$5F,$F3,$AB
	.dc.b	$FE,$55,$7E,$A1,$FC,$B0,$FC,$5E,$1F,$C6,$A3,$F3,$18,$7F,$AB,$A7
	.dc.b	$F3,$15,$FC,$BD,$FE,$84,$7F,$43,$BF,$E6,$0F,$FA,$BC,$FF,$59,$3F
	.dc.b	$92,$7F,$97,$CF,$DA,$7E,$55,$7F,$AB,$5F,$CC,$9F,$EA,$D9,$FC,$CD
	.dc.b	$7F,$AB,$AB,$FC,$07,$3F,$EA,$D8,$FE,$5C,$FF,$95,$0F,$E5,$C7,$FB
	.dc.b	$72,$3B,$3F,$C6,$23,$FD,$AD,$5F,$25,$FC,$C6,$1F,$CC,$27,$F2,$67
	.dc.b	$F9,$3C,$FF,$47,$4F,$F8,$A9,$FE,$FA,$7E,$64,$FF,$2D,$1F,$99,$13
	.dc.b	$FC,$3E,$FF,$34,$FF,$32,$BF,$F2,$BC,$FE,$60,$FE,$62,$FF,$AB,$A7
	.dc.b	$FC,$05,$1F,$97,$8F,$C5,$BF,$CB,$4F,$E5,$E7,$F1,$23,$F8,$A1,$E0
	.dc.b	$8F,$E5,$F3,$FC,$DE,$3D,$27,$F2,$F3,$F2,$91,$FE,$B0,$8F,$F5,$5E
	.dc.b	$1F,$9E,$9F,$E3,$6F,$F3,$17,$F8,$FC,$72,$7F,$33,$5F,$A2,$3F,$BA
	.dc.b	$BF,$9A,$BF,$F0,$4D,$66,$7F,$99,$9C,$67,$F3,$07,$F9,$78,$A7,$A8
	.dc.b	$FC,$5D,$FF,$34,$7F,$DC,$A7,$F9,$7C,$FC,$D9,$FE,$5F,$8F,$E6,$B1
	.dc.b	$FC,$8A,$3F,$8F,$1F,$D9,$CF,$E6,$71,$FC,$79,$FF,$31,$3F,$CD,$1F
	.dc.b	$F0,$5C,$FF,$34,$7F,$C2,$31,$FE,$49,$7F,$88,$3F,$E0,$77,$F3,$47
	.dc.b	$E7,$4F,$E7,$13,$F9,$D3,$F9,$AB,$FE,$0A,$AF,$CE,$5F,$DA,$8F,$F0
	.dc.b	$AB,$EF,$9F,$B1,$19,$3F,$CC,$51,$FE,$0D,$7F,$5A,$7F,$B7,$0F,$F5
	.dc.b	$A9,$FC,$E6,$FF,$55,$7F,$E6,$13,$D1,$1F,$59,$FF,$1B,$9F,$E6,$4C
	.dc.b	$FF,$41,$1D,$1A,$A7,$FD,$63,$18,$78,$47,$F3,$B8,$DC,$EA,$3E,$DD
	.dc.b	$7F,$9C,$21,$3F,$CF,$A6,$9F,$EB,$AA,$E0,$FF,$40,$AF,$05,$6A,$3F
	.dc.b	$CD,$65,$3F,$5B,$1F,$EF,$FA,$FF,$60,$1F,$B9,$AF,$C8,$B6,$BD,$58
	.dc.b	$F1,$BF,$E2,$B7,$DA,$FD,$71,$FB,$0C,$3F,$CB,$58,$71,$C3,$CB,$3F
	.dc.b	$E4,$A3,$D9,$1F,$C1,$E7,$E4,$47,$F9,$D3,$0E,$CF,$EF,$CF,$5C,$FF
	.dc.b	$9D,$9C,$9F,$E2,$18,$AF,$F8,$AF,$0E,$AF,$ED,$1F,$E5,$73,$FE,$7A
	.dc.b	$3F,$D9,$B7,$CA,$3F,$A5,$CD,$BF,$E1,$F3,$E8,$8F,$C3,$BF,$4C,$E0
	.dc.b	$7C,$67,$FD,$A3,$86,$DF,$F1,$24,$7F,$AA,$23,$E2,$C7,$FB,$D6,$3F
	.dc.b	$D5,$73,$F1,$A3,$FD,$B3,$1C,$5F,$CA,$99,$3F,$77,$5E,$E7,$FD,$0D
	.dc.b	$5D,$1F,$F0,$F5,$FF,$B6,$CF,$E0,$2F,$C2,$3F,$26,$F2,$7F,$95,$BD
	.dc.b	$23,$94,$7F,$21,$8E,$F7,$A8,$FC,$9D,$E9,$F0,$BF,$31,$FE,$BD,$7F
	.dc.b	$4A,$FE,$04,$D3,$F9,$45,$FF,$B8,$8F,$C7,$D7,$F6,$FB,$FC,$0B,$FE
	.dc.b	$37,$3C,$07,$F8,$E6,$BC,$0F,$F3,$18,$F8,$8F,$E2,$DD,$BF,$8D,$AD
	.dc.b	$C7,$F4,$88,$D9,$FC,$01,$C5,$FE,$21,$FE,$9B,$5C,$23,$CF,$3F,$E9
	.dc.b	$4B,$FE,$5B,$5F,$84,$BF,$C7,$9F,$81,$9F,$F0,$89,$E2,$3F,$0E,$FE
	.dc.b	$61,$FE,$D5,$3F,$E4,$03,$FD,$8E,$7D,$68,$F9,$95,$F8,$08,$FE,$0D
	.dc.b	$87,$E5,$6B,$E7,$C7,$FB,$A8,$FE,$0F,$1F,$83,$9F,$BD,$3F,$BC,$3D
	.dc.b	$E7,$FB,$A3,$0F,$C6,$47,$FA,$5A,$BF,$8B,$BE,$E9,$FE,$11,$7F,$91
	.dc.b	$7F,$84,$BE,$4A,$F7,$47,$F2,$89,$FE,$01,$7F,$A9,$9F,$BF,$7F,$C1
	.dc.b	$F0,$F0,$DF,$CC,$C3,$F8,$73,$FE,$F2,$7F,$5C,$FE,$24,$FF,$24,$5E
	.dc.b	$2F,$E1,$D7,$F8,$86,$77,$5E,$58,$FA,$57,$FE,$9E,$9F,$DC,$61,$FC
	.dc.b	$5A,$3F,$7A,$7F,$BD,$8F,$E1,$CF,$8A,$3F,$9D,$47,$FA,$81,$FD,$DD
	.dc.b	$71,$7F,$76,$E2,$BD,$75,$FC,$07,$0F,$D9,$57,$A4,$FE,$6F,$7F,$B0
	.dc.b	$C3,$F2,$F1,$E9,$8F,$DA,$3F,$DD,$30,$CD,$F1,$8F,$E7,$13,$FB,$39
	.dc.b	$FB,$98,$7C,$A9,$FF,$10,$57,$DF,$9F,$D1,$DF,$D4,$9F,$EE,$27,$E4
	.dc.b	$4F,$F4,$44,$7E,$45,$FE,$E3,$5C,$E7,$F8,$8D,$7E,$D6,$BF,$4F,$5E
	.dc.b	$33,$FC,$0E,$7F,$AF,$6F,$F4,$2F,$33,$FD,$FE,$E2,$FC,$11,$F5,$E7
	.dc.b	$F5,$4F,$A6,$3F,$5F,$3F,$AA,$AE,$53,$FB,$37,$F2,$4F,$F9,$55,$FD
	.dc.b	$A1,$F9,$C7,$F5,$67,$C3,$A1,$E5,$7F,$CF,$0F,$E4,$9F,$D7,$98,$7F
	.dc.b	$A2,$4F,$EC,$23,$E9,$47,$EB,$5F,$80,$7F,$07,$3F,$04,$7F,$96,$2F
	.dc.b	$F6,$AF,$DB,$3F,$D2,$40,$FC,$61,$FE,$95,$8D,$D7,$43,$FD,$CD,$3F
	.dc.b	$B2,$9F,$D2,$57,$EC,$AF,$F1,$B1,$FD,$43,$0F,$A9,$87,$EB,$4E,$57
	.dc.b	$FD,$2E,$30,$FE,$D6,$3E,$4C,$FF,$74,$8F,$F0,$44,$FF,$1D,$8F,$94
	.dc.b	$7F,$6B,$3F,$37,$5F,$D2,$A3,$F6,$15,$FB,$48,$CC,$FB,$0F,$AD,$1F
	.dc.b	$B4,$C3,$D9,$3F,$CA,$2B,$F6,$37,$98,$EB,$3C,$6F,$FD,$37,$7F,$DA
	.dc.b	$70,$FD,$81,$FE,$5C,$C3,$F6,$38,$7F,$22,$9F,$EC,$B5,$FB,$47,$F8
	.dc.b	$A3,$FD,$18,$FD,$79,$FE,$5E,$8F,$D7,$BF,$08,$FF,$21,$D7,$F2,$D9
	.dc.b	$FE,$4D,$1F,$AB,$3F,$26,$7F,$97,$E3,$FA,$FB,$FB,$0B,$FD,$74,$FF
	.dc.b	$59,$8F,$D4,$3F,$8F,$7F,$CB,$75,$FB,$28,$F0,$47,$F6,$4C,$3F,$46
	.dc.b	$FF,$2A,$8F,$A4,$FE,$8C,$E4,$7F,$88,$E7,$F6,$33,$93,$89,$F8,$D9
	.dc.b	$FD,$5C,$FE,$90,$CD,$7E,$A9,$FF,$6B,$4F,$EB,$E3,$F3,$F1,$FD,$B5
	.dc.b	$FC,$0B,$CE,$7F,$3E,$FF,$93,$4F,$3B,$FA,$97,$FD,$D7,$3F,$AF,$8E
	.dc.b	$71,$FC,$A0,$FD,$4D,$FF,$97,$9F,$CC,$9F,$A8,$3F,$D7,$27,$94,$FD
	.dc.b	$64,$EA,$7E,$DD,$79,$0F,$D5,$1F,$A7,$9F,$C6,$57,$F5,$6B,$FD,$5D
	.dc.b	$FE,$9A,$BF,$19,$5F,$E1,$2B,$FD,$13,$FC,$DA,$BF,$4D,$7F,$E5,$97
	.dc.b	$F4,$D7,$9A,$E3,$5F,$EB,$B7,$FA,$7B,$F9,$C3,$D4,$7F,$97,$A3,$FB
	.dc.b	$8B,$FE,$BE,$3E,$FD,$F1,$C3,$F8,$BD,$7E,$96,$7F,$43,$7F,$DC,$2F
	.dc.b	$F7,$F1,$FE,$53,$7F,$4B,$3F,$DB,$CF,$D2,$47,$F0,$97,$F8,$41,$EA
	.dc.b	$7F,$A9,$E1,$FE,$ED,$8F,$D2,$B5,$F5,$C7,$E9,$4F,$F7,$94,$7F,$53
	.dc.b	$8F,$EA,$F3,$FA,$19,$FB,$67,$F2,$59,$FF,$22,$3F,$A6,$C3,$FC,$19
	.dc.b	$7F,$D5,$AB,$F5,$11,$C2,$7F,$CA,$AF,$E3,$1F,$F4,$79,$FA,$63,$F4
	.dc.b	$33,$F8,$B7,$FD,$8F,$3F,$AA,$33,$7C,$5F,$44,$7F,$50,$3F,$42,$7E
	.dc.b	$24,$FF,$20,$1F,$C6,$DC,$C7,$E9,$2F,$FD,$86,$7F,$4D,$AF,$F1,$D4
	.dc.b	$FE,$8C,$FD,$3C,$FF,$92,$8F,$ED,$15,$FC,$51,$FD,$9B,$FE,$38,$9D
	.dc.b	$47,$EF,$23,$FD,$65,$3F,$A6,$8F,$81,$1E,$83,$E1,$D7,$CF,$3F,$CE
	.dc.b	$33,$F3,$CF,$E0,$71,$FB,$18,$F9,$E7,$F9,$CE,$FE,$7D,$7F,$80,$27
	.dc.b	$F8,$94,$FF,$A0,$23,$F8,$74,$7C,$D9,$FF,$36,$5F,$F1,$1B,$F9,$91
	.dc.b	$FD,$22,$FF,$BB,$D7,$F0,$F3,$E5,$D7,$F0,$E7,$E5,$4F,$F4,$5B,$F9
	.dc.b	$51,$FC,$1C,$FE,$1B,$3F,$29,$FE,$E7,$7F,$2A,$7F,$CC,$27,$F0,$A3
	.dc.b	$FB,$9C,$7C,$59,$F9,$33,$FC,$24,$F9,$13,$FC,$DA,$7F,$24,$7C,$43
	.dc.b	$EF,$1F,$DB,$4F,$F0,$35,$FF,$39,$3E,$3B,$FE,$E4,$3E,$5D,$F0,$3E
	.dc.b	$B2,$7F,$2C,$FF,$AB,$E7,$F9,$D3,$FC,$EE,$3E,$4C,$7F,$A5,$6B,$E4
	.dc.b	$D7,$FA,$52,$BF,$D2,$47,$F3,$8A,$FE,$D5,$7F,$26,$7F,$C4,$11,$F2
	.dc.b	$6F,$F9,$A5,$7F,$35,$8F,$F0,$CD,$7C,$9A,$FF,$10,$5F,$C9,$7F,$D7
	.dc.b	$B5,$F9,$13,$F9,$71,$F2,$6F,$E6,$CF,$C9,$BF,$9F,$1F,$26,$7F,$CA
	.dc.b	$27,$C9,$BF,$EC,$35,$F8,$FA,$FF,$2A,$D7,$F9,$3A,$BF,$72,$7C,$8A
	.dc.b	$FD,$E3,$F8,$E3,$FD,$7B,$1F,$1A,$FC,$B1,$FE,$BE,$3F,$6F,$3F,$8F
	.dc.b	$7E,$45,$7E,$3D,$FF,$5D,$1F,$D3,$2F,$FC,$48,$FE,$42,$FF,$70,$FE
	.dc.b	$42,$FF,$C9,$71,$F9,$08,$F8,$73,$FD,$D1,$B5,$FC,$06,$3E,$E1,$FB
	.dc.b	$B7,$F9,$7C,$FE,$4A,$7E,$4C,$7E,$52,$FF,$23,$1F,$2A,$3E,$4C,$FC
	.dc.b	$88,$F9,$31,$FE,$7A,$9F,$93,$1F,$C9,$A3,$E4,$CF,$F8,$22,$3F,$92
	.dc.b	$5F,$F9,$D4,$FF,$03,$9F,$EE,$69,$FF,$03,$BF,$22,$3F,$91,$9F,$8F
	.dc.b	$71,$5F,$C8,$E3,$E2,$87,$F9,$DB,$F4,$1F,$F3,$AC,$7E,$4E,$36,$2F
	.dc.b	$F8,$83,$4F,$E3,$27,$F9,$E2,$7F,$2C,$FD,$48,$FE,$4F,$7F,$96,$AF
	.dc.b	$E4,$F3,$FE,$05,$AF,$CB,$3F,$ED,$F3,$FC,$11,$3F,$C8,$9F,$9B,$3E
	.dc.b	$D8,$F5,$47,$F2,$9A,$FC,$B4,$FF,$9B,$63,$E1,$1F,$BE,$BF,$9D,$5F
	.dc.b	$D4,$A3,$E7,$4F,$EF,$5F,$DE,$1F,$38,$FF,$1A,$1F,$2E,$7F,$7A,$7E
	.dc.b	$F1,$FC,$6B,$FD,$46,$FF,$9D,$C7,$E5,$E3,$F7,$B3,$F4,$2B,$F9,$D5
	.dc.b	$7C,$CA,$F9,$D1,$FD,$45,$FC,$C1,$F2,$8F,$F6,$84,$7C,$A7,$F2,$A7
	.dc.b	$84,$FF,$04,$47,$CB,$7F,$94,$57,$F0,$08,$FE,$4F,$1F,$CA,$23,$F2
	.dc.b	$F5,$F2,$67,$F9,$41,$DE,$7F,$D3,$6F,$F5,$F3,$F0,$6F,$F8,$CA,$7F
	.dc.b	$33,$3F,$32,$3F,$C7,$27,$C2,$7F,$A8,$3F,$42,$FF,$81,$47,$CD,$8F
	.dc.b	$DD,$CF,$03,$C0,$FF,$85,$0C,$CF,$F7,$E3,$F9,$6D,$7F,$50,$8F,$CD
	.dc.b	$1F,$22,$FF,$81,$5F,$CD,$AF,$EF,$75,$FE,$CA,$AF,$EA,$17,$FC,$A2
	.dc.b	$3F,$80,$D7,$F2,$83,$F8,$0B,$F2,$67,$F7,$E7,$F5,$07,$E4,$C7,$F2
	.dc.b	$89,$F9,$33,$FE,$BB,$3F,$26,$7F,$AF,$6F,$F7,$F3,$F3,$6F,$FA,$7D
	.dc.b	$FE,$FE,$3E,$45,$7F,$27,$8F,$C9,$47,$FB,$D2,$3E,$A3,$FD,$39,$F8
	.dc.b	$E7,$F2,$78,$FE,$E9,$1F,$BF,$3F,$BA,$3F,$C9,$E3,$F1,$67,$F2,$67
	.dc.b	$FA,$63,$FC,$9A,$3F,$74,$FF,$00,$9F,$DC,$9F,$65,$FE,$53,$5F,$8F
	.dc.b	$7F,$7F,$1F,$EC,$D3,$F7,$33,$FE,$E9,$8F,$F3,$F1,$F1,$DF,$E9,$AF
	.dc.b	$E3,$DF,$F7,$75,$7E,$E6,$3F,$93,$5F,$F2,$58,$FE,$9B,$1F,$C9,$67
	.dc.b	$F7,$0F,$EE,$1F,$8F,$3F,$B8,$7F,$C3,$71,$FB,$83,$FD,$02,$7E,$E1
	.dc.b	$FF,$0D,$1F,$C9,$5F,$DC,$57,$E2,$E7,$F9,$2C,$7C,$63,$F1,$67,$FA
	.dc.b	$AA,$BE,$3D,$7E,$2E,$7F,$A6,$4F,$F7,$28,$FF,$0D,$C7,$E2,$DF,$E5
	.dc.b	$D5,$FE,$07,$9F,$8E,$7E,$D5,$FF,$17,$3F,$E1,$B3,$FB,$A4,$7F,$91
	.dc.b	$6B,$F1,$B1,$FC,$7C,$FE,$9E,$FC,$78,$FD,$B4,$7F,$73,$7F,$18,$FC
	.dc.b	$73,$FB,$2C,$7F,$9E,$1F,$C4,$9F,$CD,$67,$FD,$42,$84,$FD,$B8,$FF
	.dc.b	$52,$DB,$0C,$D7,$13,$F8,$E9,$FE,$90,$9F,$E7,$CF,$FA,$98,$FC,$21
	.dc.b	$C1,$CB,$FE,$21,$3F,$D9,$11,$98,$FF,$46,$57,$86,$7A,$E1,$FD,$DA
	.dc.b	$31,$1D,$63,$FA,$A5,$71,$8D,$DB,$0C,$46,$F0,$F7,$61,$CA,$3C,$58
	.dc.b	$6F,$0D,$C7,$FA,$DE,$EC,$FF,$66,$AE,$0F,$F6,$38,$F0,$C6,$30,$F0
	.dc.b	$1E,$17,$11,$FD,$85,$AF,$CB,$7C,$55,$FC,$D9,$EB,$87,$F7,$06,$C7
	.dc.b	$63,$E8,$1F,$ED,$28,$FA,$07,$F7,$6B,$FF,$48,$CF,$0A,$ED,$5C,$0F
	.dc.b	$DB,$4F,$F3,$A9,$EE,$FC,$D8,$D3,$FD,$E4,$F0,$57,$D8,$71,$7E,$D9
	.dc.b	$DC,$FF,$88,$1F,$8E,$FF,$97,$6F,$C2,$FE,$02,$3F,$A4,$C7,$18,$E8
	.dc.b	$FF,$9D,$23,$E7,$BF,$EC,$B8,$CB,$F3,$EF,$E8,$BF,$E1,$28,$CB,$D9
	.dc.b	$FE,$09,$3E,$F3,$E2,$4F,$CE,$7E,$39,$FD,$99,$C3,$EF,$6C,$F9,$67
	.dc.b	$FC,$C3,$7F,$15,$F5,$57,$E2,$2B,$A4,$F9,$63,$C0,$7C,$38,$CC,$FF
	.dc.b	$9F,$DD,$D7,$59,$FF,$16,$DF,$C1,$9F,$C1,$BF,$E9,$6B,$FC,$9C,$FE
	.dc.b	$3E,$38,$3F,$E5,$87,$E1,$5F,$FB,$62,$77,$1F,$EF,$68,$FA,$07,$E2
	.dc.b	$E7,$E2,$3F,$CE,$CC,$BE,$96,$9F,$A6,$76,$7C,$52,$DF,$89,$FC,$59
	.dc.b	$B8,$EF,$5F,$C2,$CF,$4B,$FE,$9F,$AE,$13,$F8,$D8,$E0,$7F,$4A,$9E
	.dc.b	$CF,$2A,$D9,$F1,$67,$EE,$3F,$16,$FF,$52,$FE,$C0,$F8,$E7,$C4,$8F
	.dc.b	$F2,$CB,$FE,$2B,$9F,$C9,$3F,$15,$FE,$4A,$FE,$FA,$3E,$34,$7E,$7D
	.dc.b	$FE,$49,$7F,$8D,$7F,$8F,$9F,$1E,$3F,$60,$FF,$1E,$AF,$E0,$AF,$F1
	.dc.b	$C7,$EF,$D7,$5B,$FC,$55,$70,$7F,$5A,$DB,$F9,$91,$FB,$DB,$E1,$5E
	.dc.b	$17,$F0,$0F,$EE,$5F,$DB,$CF,$C5,$AF,$DB,$3F,$A9,$7F,$82,$3F,$16
	.dc.b	$FF,$89,$C7,$D0,$9F,$F2,$AC,$FF,$8F,$30,$FA,$F1,$D6,$7F,$80,$1F
	.dc.b	$10,$FE,$1C,$7C,$48,$FD,$FC,$7C,$33,$FC,$BF,$1F,$0D,$FE,$0A,$7F
	.dc.b	$2D,$AF,$E3,$F1,$FC,$16,$FF,$83,$E1,$F1,$2B,$8B,$F1,$1B,$87,$C4
	.dc.b	$AF,$D6,$47,$F1,$09,$E8,$FE,$12,$BD,$13,$FE,$EB,$8F,$71,$F9,$38
	.dc.b	$F8,$8F,$F9,$D1,$F8,$8F,$F2,$07,$F1,$0E,$DE,$B1,$FC,$02,$3C,$C7
	.dc.b	$F0,$4B,$FE,$13,$7F,$9C,$8F,$81,$3B,$3F,$80,$3F,$C1,$0F,$29,$FE
	.dc.b	$A6,$BE,$6F,$F5,$EA,$FE,$17,$5C,$E7,$FA,$44,$FF,$0C,$7F,$4B,$5F
	.dc.b	$D9,$A3,$F8,$F4,$FE,$06,$3F,$D6,$2F,$E9,$63,$37,$F5,$5F,$C4,$CF
	.dc.b	$C4,$7F,$3B,$1D,$1F,$E2,$6F,$99,$FF,$72,$C7,$E0,$5F,$F2,$21,$FB
	.dc.b	$DB,$FB,$6F,$53,$FB,$FC,$75,$7F,$D8,$0F,$C2,$3F,$87,$CF,$F0,$27
	.dc.b	$FC,$05,$1F,$C1,$AF,$A3,$FC,$59,$E8,$FF,$07,$35,$1C,$CF,$E1,$13
	.dc.b	$FC,$18,$F9,$71,$FE,$96,$34,$7F,$02,$79,$9F,$79,$FF,$10,$3F,$C1
	.dc.b	$63,$FD,$B2,$FF,$05,$8F,$F6,$D9,$F0,$CF,$39,$F2,$67,$F9,$96,$1F
	.dc.b	$C2,$63,$E1,$47,$C8,$71,$3F,$B9,$7F,$20,$FF,$03,$79,$1F,$C9,$67
	.dc.b	$F8,$36,$1F,$C0,$A3,$F3,$F7,$FC,$4E,$FF,$26,$FD,$EB,$F2,$47,$F8
	.dc.b	$56,$7F,$81,$C7,$EA,$5F,$E6,$55,$F9,$EA,$FE,$0B,$3F,$8C,$6B,$E5
	.dc.b	$9F,$F0,$91,$FB,$83,$DD,$1F,$BF,$C3,$FC,$27,$5F,$B3,$9E,$53,$FC
	.dc.b	$11,$F2,$9F,$E8,$9B,$F2,$C7,$DE,$AF,$F0,$7D,$7C,$39,$FF,$05,$DF
	.dc.b	$F0,$2A,$EF,$1C,$AF,$FC,$A8,$7F,$82,$27,$FC,$0A,$7E,$A2,$3F,$DA
	.dc.b	$75,$CE,$3D,$D5,$FE,$23,$7E,$F4,$FF,$84,$30,$FE,$D7,$7F,$9B,$3F
	.dc.b	$74,$7E,$D6,$FF,$74,$FB,$0F,$C0,$DF,$E2,$9F,$C2,$DF,$2B,$FC,$7C
	.dc.b	$7F,$80,$CF,$DB,$57,$EE,$6B,$F4,$50,$9F,$D7,$3E,$5A,$FE,$EE,$7E
	.dc.b	$62,$BF,$71,$5F,$E2,$77,$D7,$3F,$B4,$8F,$D6,$47,$E5,$E3,$F6,$F3
	.dc.b	$FA,$B7,$F4,$33,$FD,$F4,$FF,$47,$CE,$5F,$DA,$1F,$E4,$2C,$3F,$73
	.dc.b	$5F,$B3,$BE,$27,$EC,$A7,$FB,$DD,$65,$E3,$1E,$AB,$FF,$25,$D7,$EC
	.dc.b	$5F,$E0,$0F,$AA,$7E,$3D,$7E,$D5,$CC,$71,$9F,$B9,$7F,$E3,$08,$FD
	.dc.b	$54,$7E,$90,$FC,$D1,$FC,$1A,$BF,$48,$FD,$53,$FD,$1C,$7E,$1C,$FD
	.dc.b	$6C,$FF,$72,$BF,$C3,$47,$EB,$63,$13,$FA,$D9,$E5,$3F,$E8,$2A,$FD
	.dc.b	$14,$FF,$01,$8F,$D4,$C7,$E3,$E3,$FA,$4C,$61,$FE,$EE,$7E,$38,$FE
	.dc.b	$D5,$5F,$32,$7F,$B2,$CF,$EA,$8F,$5D,$2F,$FC,$A7,$1F,$DF,$DD,$4F
	.dc.b	$EC,$DF,$EC,$D5,$FE,$76,$9F,$D2,$DF,$EA,$67,$FC,$59,$5F,$A9,$9F
	.dc.b	$ED,$6F,$EE,$CF,$53,$FD,$BC,$FE,$30,$FA,$9F,$EE,$53,$FE,$10,$8F
	.dc.b	$D3,$9E,$A8,$FE,$B0,$7E,$68,$FC,$24,$FF,$24,$7F,$9A,$9F,$D8,$1C
	.dc.b	$1F,$A7,$7F,$98,$57,$F9,$99,$FD,$1D,$F2,$7F,$B1,$5F,$E9,$DF,$D6
	.dc.b	$57,$17,$FB,$95,$7F,$99,$2F,$F4,$C7,$18,$FF,$11,$D7,$F9,$30,$FE
	.dc.b	$78,$FF,$AD,$6B,$94,$7C,$28,$FF,$68,$D7,$F5,$87,$B4,$FD,$27,$FA
	.dc.b	$C9,$FE,$22,$72,$FF,$41,$3F,$D4,$87,$E8,$8E,$37,$FD,$5A,$3F,$9B
	.dc.b	$9F,$E5,$B3,$37,$FA,$59,$FF,$2A,$BE,$97,$FA,$8B,$FD,$50,$FC,$9B
	.dc.b	$FE,$D9,$7F,$4C,$7A,$5F,$E3,$07,$F8,$26,$BF,$C2,$31,$FC,$72,$38
	.dc.b	$9F,$BD,$3F,$C2,$0F,$F8,$2D,$FA,$27,$CD,$3F,$13,$1F,$E9,$67,$E8
	.dc.b	$D7,$F1,$D9,$F9,$87,$D0,$8F,$96,$FF,$74,$BF,$9E,$7F,$13,$8F,$F2
	.dc.b	$AC,$7F,$13,$7F,$87,$9E,$F3,$F4,$07,$C7,$3D,$F7,$E7,$3F,$C0,$B1
	.dc.b	$FE,$03,$8F,$9D,$5F,$E0,$5A,$FE,$25,$1F,$E0,$28,$FF,$04,$4F,$CD
	.dc.b	$BF,$F4,$1B,$DC,$FE,$53,$3F,$D2,$9F,$C2,$9E,$77,$E8,$BF,$D0,$1F
	.dc.b	$CF,$1F,$14,$FC,$E9,$F3,$0F,$E3,$07,$D8,$8F,$CE,$3F,$D0,$CF,$99
	.dc.b	$3F,$A4,$3E,$99,$C0,$FC,$55,$7F,$8D,$E3,$FC,$5E,$7C,$C7,$E7,$DF
	.dc.b	$F3,$EB,$FE,$4D,$1F,$99,$8F,$C0,$CF,$CA,$3F,$2B,$1F,$CF,$5F,$CC
	.dc.b	$57,$F9,$EA,$BF,$9E,$47,$CA,$BF,$F3,$8C,$FE,$5C,$FA,$27,$F3,$3B
	.dc.b	$FE,$1F,$3F,$E6,$B3,$F9,$B9,$FB,$3A,$F9,$6F,$E6,$23,$F9,$D9,$98
	.dc.b	$FD,$34,$79,$8F,$EA,$47,$F8,$AA,$FF,$B6,$CF,$F0,$B3,$E7,$4F,$E5
	.dc.b	$67,$E5,$57,$F6,$DA,$FE,$09,$7F,$EE,$A8,$FC,$A5,$7F,$68,$AF,$8C
	.dc.b	$7C,$97,$FB,$5D,$FC,$93,$F2,$EF,$E4,$E3,$FD,$3A,$7C,$87,$F7,$CF
	.dc.b	$C8,$AF,$E7,$07,$C6,$7E,$41,$FE,$53,$BF,$C8,$C7,$CA,$AF,$DE,$D7
	.dc.b	$EF,$2B,$E3,$DF,$EF,$AF,$FB,$04,$7F,$00,$AF,$8D,$5F,$8F,$8F,$DC
	.dc.b	$1F,$8F,$9F,$8A,$FE,$EE,$7E,$2C,$FE,$36,$3E,$2D,$FF,$A1,$A3,$F9
	.dc.b	$3C,$7E,$E2,$3F,$D9,$E7,$F2,$77,$E2,$DF,$F8,$3A,$FF,$1B,$1F,$1A
	.dc.b	$3F,$C2,$35,$F1,$6B,$F9,$24,$7F,$B0,$AB,$F1,$91,$FE,$7E,$BF,$E4
	.dc.b	$75,$F8,$CB,$FD,$B5,$FE,$D6,$3F,$CE,$B1,$FB,$5A,$FC,$7C,$7E,$D6
	.dc.b	$3F,$19,$3F,$16,$3E,$24,$7C,$58,$FE,$4B,$1F,$16,$53,$FD,$4A,$7F
	.dc.b	$8F,$5F,$C6,$9C,$DF,$FA,$BE,$7F,$91,$5F,$E2,$E7,$FA,$84,$7E,$2E
	.dc.b	$3E,$3C,$7E,$2C,$F8,$F7,$FE,$A6,$3E,$3B,$FD,$3E,$FE,$3B,$FD,$46
	.dc.b	$FF,$17,$5F,$8A,$AF,$C5,$DF,$C7,$8F,$C5,$D7,$C7,$8F,$D9,$47,$F1
	.dc.b	$B9,$F8,$F3,$FB,$2B,$FC,$5C,$7F,$A7,$EF,$F1,$73,$FC,$6A,$3F,$16
	.dc.b	$FF,$1C,$9F,$C5,$BF,$C7,$A7,$F1,$75,$F8,$AA,$FC,$5D,$FC,$68,$FC
	.dc.b	$54,$7E,$2E,$3E,$2C,$7E,$2E,$BE,$29,$F8,$B3,$FA,$11,$F1,$67,$FD
	.dc.b	$A1,$1F,$D0,$9F,$8A,$7F,$B3,$2F,$E2,$BF,$EC,$EB,$F8,$B5,$FD,$B6
	.dc.b	$3E,$2B,$FE,$C6,$9F,$C5,$C7,$F8,$86,$7F,$CF,$AF,$FB,$22,$FE,$3D
	.dc.b	$7E,$C6,$3E,$39,$FB,$1B,$FC,$6C,$7E,$C6,$7F,$A0,$C7,$EC,$67,$F6
	.dc.b	$51,$F8,$D9,$F8,$D1,$F8,$D7,$FA,$2C,$FE,$36,$3F,$19,$1F,$DA,$8F
	.dc.b	$C5,$D7,$F8,$6A,$3E,$44,$7F,$95,$4F,$E7,$E7,$C8,$7F,$CF,$F3,$F8
	.dc.b	$C7,$FD,$ED,$3F,$D0,$A3,$E4,$47,$F1,$7A,$FD,$94,$7F,$16,$8F,$D9
	.dc.b	$57,$F8,$76,$3F,$65,$1F,$14,$FF,$77,$9F,$C5,$E3,$F6,$4F,$F1,$73
	.dc.b	$47,$FB,$3E,$3F,$69,$1F,$C6,$67,$F6,$8F,$F1,$79,$FD,$A1,$FE,$ED
	.dc.b	$9F,$DA,$57,$F4,$B8,$FF,$09,$9F,$CE,$EF,$F1,$EF,$63,$FA,$05,$FE
	.dc.b	$DA,$BF,$21,$7F,$22,$7F,$C8,$93,$F9,$18,$C4,$7D,$E3,$FB,$55,$FF
	.dc.b	$1C,$8F,$97,$1D,$4F,$DF,$CF,$F8,$1D,$FF,$0A,$9F,$30,$FE,$3C,$7E
	.dc.b	$52,$BF,$A1,$3F,$36,$7F,$71,$1F,$35,$FD,$C4,$FF,$1E,$8F,$DC,$47
	.dc.b	$F1,$E3,$F2,$8F,$EE,$27,$C5,$5F,$E1,$33,$D0,$7F,$7F,$73,$5F,$EA
	.dc.b	$67,$FC,$83,$7F,$BA,$9F,$95,$1F,$E7,$D9,$FE,$D8,$7F,$92,$6B,$FC
	.dc.b	$01,$1F,$E1,$37,$FC,$2C,$7F,$90,$67,$FB,$5C,$FC,$DB,$FF,$52,$69
	.dc.b	$F3,$2F,$F4,$ED,$58,$31,$4E,$61,$8E,$2B,$E9,$1F,$ED,$33,$F9,$49
	.dc.b	$93,$EB,$61,$FE,$27,$BC,$47,$08,$FF,$2A,$5F,$8D,$FF,$7A,$B6,$FC
	.dc.b	$4D,$BE,$35,$FE,$FA,$AF,$BB,$8F,$F3,$CA,$E5,$5C,$2B,$FD,$19,$7F
	.dc.b	$E6,$18,$E3,$1F,$EB,$23,$81,$E3,$8F,$69,$C5,$FF,$45,$C7,$C4,$38
	.dc.b	$BF,$EA,$6A,$E2,$EA,$38,$3F,$78,$F6,$9E,$F8,$F9,$ED,$3E,$7E,$1D
	.dc.b	$30,$FF,$2D,$DF,$CE,$C3,$F4,$37,$F3,$1B,$7F,$4A,$FC,$6F,$E1,$70
	.dc.b	$E0,$F7,$C3,$E5,$B7,$0F,$2B,$F2,$70,$FF,$5F,$E1,$DA,$BF,$CD,$6D
	.dc.b	$BD,$D7,$E3,$5E,$AF,$DC,$7F,$06,$FC,$23,$F9,$8C,$FF,$A3,$EF,$8C
	.dc.b	$7F,$B1,$27,$DC,$7E,$2A,$F8,$47,$DB,$8F,$F4,$B6,$1E,$C8,$F8,$E7
	.dc.b	$FB,$14,$FD,$DE,$15,$FE,$F7,$7E,$47,$F9,$74,$65,$FD,$43,$6B,$51
	.dc.b	$F9,$43,$C2,$7A,$06,$1E,$13,$F0,$13,$DD,$F8,$B1,$E3,$7A,$B6,$7A
	.dc.b	$BF,$E0,$E3,$F1,$07,$63,$CE,$FA,$9F,$CA,$BD,$9B,$86,$DF,$F6,$84
	.dc.b	$FD,$EB,$FC,$D5,$FC,$3A,$F8,$B5,$F8,$F3,$D3,$19,$7A,$1F,$B6,$8F
	.dc.b	$C5,$3F,$D4,$DD,$BF,$D5,$6F,$D2,$FF,$88,$AF,$F1,$2F,$17,$D0,$DB
	.dc.b	$DB,$FC,$56,$BD,$07,$CA,$3A,$BC,$6B,$67,$E2,$1F,$CC,$BF,$8A,$AA
	.dc.b	$FA,$23,$F9,$F3,$F8,$77,$33,$D1,$CC,$7E,$B9,$FC,$33,$EA,$9F,$E8
	.dc.b	$D7,$FC,$67,$0F,$D7,$57,$D2,$8F,$E4,$17,$F5,$6B,$F8,$B4,$7D,$FB
	.dc.b	$E5,$5F,$C7,$4E,$33,$FB,$89,$FE,$28,$7F,$19,$AE,$53,$F8,$AA,$FD
	.dc.b	$0B,$FC,$25,$F2,$61,$F0,$6B,$FA,$EB,$FC,$43,$0F,$E0,$4F,$F1,$29
	.dc.b	$A7,$E4,$2B,$F8,$74,$FF,$00,$BF,$F7,$FD,$FF,$04,$9F,$54,$7E,$92
	.dc.b	$7F,$86,$3E,$F8,$FE,$20,$E9,$FE,$15,$3F,$8B,$C3,$E5,$47,$EE,$A7
	.dc.b	$E0,$4F,$F0,$5C,$3F,$77,$5F,$D4,$DF,$9D,$3F,$7E,$BB,$CF,$47,$FA
	.dc.b	$29,$FC,$C7,$0F,$E0,$27,$E3,$1F,$E0,$EF,$C1,$BF,$EF,$EF,$C1,$6B
	.dc.b	$F0,$63,$6F,$57,$F7,$D1,$FB,$F7,$F0,$87,$FB,$16,$3F,$85,$3C,$4F
	.dc.b	$CD,$47,$F0,$38,$FD,$F4,$7E,$BC,$FE,$B5,$3F,$C0,$6F,$F3,$B1,$ED
	.dc.b	$AF,$E0,$57,$DA,$3C,$71,$FC,$56,$7E,$E4,$7E,$F0,$D1,$FC,$01,$FD
	.dc.b	$39,$F9,$B8,$E1,$7F,$C2,$A7,$D0,$FE,$B2,$BF,$7B,$1F,$E3,$8B,$FC
	.dc.b	$03,$FA,$53,$FC,$BF,$1F,$BC,$7F,$7C,$73,$3F,$A9,$4F,$39,$FA,$11
	.dc.b	$C1,$FD,$DB,$FE,$14,$7A,$3F,$C6,$67,$E0,$3F,$69,$FD,$E4,$7E,$44
	.dc.b	$FC,$3C,$7F,$9C,$27,$E0,$BF,$E0,$47,$F7,$95,$FB,$47,$AC,$7F,$BF
	.dc.b	$CF,$DD,$CF,$F0,$08,$F9,$F1,$D0,$FF,$1C,$9F,$08,$A7,$EC,$4F,$F0
	.dc.b	$51,$F9,$D3,$F9,$E1,$F7,$CD,$BF,$B2,$7F,$68,$FE,$E9,$FD,$BC,$73
	.dc.b	$7F,$A2,$3E,$63,$FC,$1A,$FF,$00,$8F,$DD,$DF,$F3,$28,$FF,$04,$47
	.dc.b	$CC,$7F,$C9,$F1,$F9,$B7,$2F,$F0,$08,$F2,$47,$F4,$E8,$FE,$BF,$5F
	.dc.b	$CD,$E7,$F5,$73,$CC,$FD,$64,$FE,$C2,$FD,$D1,$FB,$53,$9D,$A3,$FD
	.dc.b	$BB,$5F,$C4,$1F,$7D,$FF,$7D,$AF,$E0,$B5,$EF,$7F,$7C,$FB,$E7,$EF
	.dc.b	$C7,$F8,$E1,$FE,$35,$87,$D5,$BF,$DC,$CF,$F2,$77,$DC,$FF,$80,$DF
	.dc.b	$90,$7E,$75,$FF,$1F,$3E,$E7,$F6,$D7,$FB,$7B,$FD,$AE,$1F,$80,$C3
	.dc.b	$F6,$55,$F9,$93,$F1,$57,$FB,$67,$F6,$F5,$FB,$58,$FE,$ED,$3F,$B9
	.dc.b	$8E,$13,$FB,$43,$F6,$47,$F6,$99,$FC,$FB,$FE,$C6,$AF,$E1,$CF,$F7
	.dc.b	$03,$55,$FB,$87,$F9,$C4,$FE,$C1,$FE,$BA,$72,$3F,$4C,$7E,$70,$FA
	.dc.b	$47,$DE,$2C,$73,$3F,$9B,$BF,$AB,$AF,$BC,$FF,$9F,$4D,$3C,$E3,$E9
	.dc.b	$C7,$ED,$5C,$CF,$F6,$A3,$89,$FA,$E8,$FE,$5E,$FB,$1F,$93,$1F,$E0
	.dc.b	$B8,$FF,$18,$4F,$DB,$BF,$E4,$47,$FA,$06,$BF,$AC,$9F,$85,$3F,$CD
	.dc.b	$C7,$E4,$0A,$7E,$40,$FC,$7E,$1F,$E2,$7A,$FE,$CF,$5F,$99,$3F,$47
	.dc.b	$1F,$DA,$30,$CE,$1F,$B0,$AF,$D5,$9F,$BF,$BF,$D2,$3C,$67,$F1,$B1
	.dc.b	$FE,$4C,$7D,$8D,$8F,$D0,$C7,$E3,$63,$F9,$0C,$7E,$A9,$FE,$BD,$1F
	.dc.b	$A9,$7F,$3C,$7E,$A2,$7D,$38,$7F,$61,$AF,$EB,$78,$7A,$67,$F6,$D8
	.dc.b	$7F,$5C,$C3,$31,$FA,$AA,$FE,$FA,$7F,$03,$C3,$F4,$47,$E9,$5F,$94
	.dc.b	$FF,$B4,$9F,$4B,$FE,$5F,$7F,$72,$7A,$5F,$F2,$E6,$1F,$A5,$3F,$CB
	.dc.b	$55,$FA,$58,$FE,$B7,$5F,$95,$7F,$16,$7F,$AE,$67,$FC,$AC,$E7,$0F
	.dc.b	$CA,$3F,$8A,$3F,$78,$FD,$7B,$FD,$24,$7F,$93,$EF,$F1,$4F,$F6,$97
	.dc.b	$FA,$BC,$7E,$6D,$FC,$4D,$7F,$58,$8F,$EA,$D3,$F9,$AB,$FF,$5A,$4F
	.dc.b	$F9,$46,$B3,$1F,$A8,$3F,$45,$3F,$7C,$FD,$01,$FD,$2C,$FD,$0C,$7F
	.dc.b	$57,$8F,$F2,$5D,$FE,$F2,$7F,$CA,$97,$FA,$17,$F3,$8F,$EC,$DF,$E4
	.dc.b	$2F,$EC,$CF,$E2,$4F,$E8,$A3,$F4,$0F,$FA,$26,$3F,$C7,$AF,$E5,$CF
	.dc.b	$F1,$45,$FF,$B7,$E3,$FC,$8F,$1F,$9F,$3F,$C2,$07,$F1,$C7,$FA,$53
	.dc.b	$F9,$DC,$3F,$3D,$1F,$E4,$6B,$FC,$F4,$7F,$11,$8F,$E8,$53,$FE,$4F
	.dc.b	$C3,$F3,$AF,$FB,$6D,$FC,$2B,$FD,$2A,$3F,$3A,$78,$1F,$EF,$E7,$E9
	.dc.b	$F0,$FE,$1E,$FE,$7A,$BF,$C7,$EF,$E7,$AF,$FC,$7D,$5F,$D0,$EB,$84
	.dc.b	$7F,$8F,$27,$FD,$89,$5F,$9D,$AF,$F1,$CC,$FE,$76,$3E,$3C,$FE,$70
	.dc.b	$FF,$1C,$DF,$F9,$0A,$BC,$F1,$FD,$B2,$BF,$41,$1F,$DB,$E3,$F8,$E9
	.dc.b	$F8,$43,$FD,$71,$3F,$9B,$C3,$FD,$6D,$5F,$9B,$9F,$C2,$1E,$53,$FC
	.dc.b	$93,$87,$DD,$7E,$24,$FF,$8C,$9F,$E8,$31,$F9,$B8,$FF,$15,$C7,$E6
	.dc.b	$EF,$85,$F9,$A3,$FD,$F5,$79,$3F,$9F,$47,$EE,$63,$FD,$54,$7F,$3F
	.dc.b	$BF,$F1,$7C,$7E,$6A,$3F,$C6,$33,$F9,$AA,$FF,$16,$47,$E5,$8F,$CC
	.dc.b	$BF,$E2,$58,$FC,$C9,$FC,$E1,$FF,$4E,$C7,$E6,$0F,$E9,$CF,$94,$FC
	.dc.b	$BB,$FE,$9C,$8C,$C7,$C4,$8F,$C0,$1F,$E2,$33,$FD,$29,$3F,$E7,$47
	.dc.b	$F9,$B1,$FE,$26,$9F,$E7,$73,$FE,$21,$AE,$27,$E7,$CF,$E2,$B3,$FE
	.dc.b	$60,$AF,$E6,$AF,$F8,$7F,$0F,$CC,$57,$FA,$4A,$3F,$30,$F0,$A5,$7E
	.dc.b	$09,$F9,$B3,$FE,$98,$9F,$F0,$93,$F9,$03,$F9,$61,$C6,$FC,$EF,$FA
	.dc.b	$25,$FC,$BD,$FF,$93,$67,$F9,$9C,$FF,$86,$2F,$FC,$45,$7F,$E1,$73
	.dc.b	$F9,$D1,$FA,$58,$FF,$78,$4F,$DB,$7F,$97,$1F,$CC,$E7,$F3,$11,$FE
	.dc.b	$16,$3F,$98,$9C,$63,$FD,$F0,$7E,$60,$FC,$81,$E6,$3F,$BF,$E1,$F9
	.dc.b	$8A,$FF,$46,$CF,$FA,$19,$FC,$F9,$C0,$FE,$62,$7F,$80,$E7,$FC,$D3
	.dc.b	$3F,$EC,$D9,$FE,$63,$1F,$CC,$A7,$FC,$2D,$5F,$CC,$30,$FC,$B4,$7D
	.dc.b	$2B,$FF,$7E,$57,$F8,$48,$FE,$57,$5F,$95,$8F,$F8,$01,$5F,$95,$BF
	.dc.b	$E5,$95,$FC,$B6,$FF,$CF,$17,$FC,$B6,$3F,$2D,$1F,$3E,$7F,$2B,$1F
	.dc.b	$96,$8F,$F7,$9C,$FF,$83,$EB,$F9,$79,$F9,$4A,$FF,$3B,$9F,$CA,$AB
	.dc.b	$FC,$13,$3F,$93,$7F,$CD,$47,$FA,$54,$F9,$B5,$FE,$EE,$7F,$D8,$A7
	.dc.b	$F9,$58,$FF,$3D,$BF,$95,$8F,$CE,$3C,$8F,$F0,$81,$F5,$E3,$FC,$6B
	.dc.b	$7F,$EE,$AA,$FF,$04,$1F,$95,$AF,$CA,$1F,$AC,$9F,$E5,$D1,$C6,$FF
	.dc.b	$B4,$C7,$F9,$AC,$FE,$55,$5F,$C4,$23,$F2,$B7,$FE,$3A,$9F,$E5,$B5
	.dc.b	$FC,$AC,$FC,$AB,$FE,$05,$C3,$F9,$55,$F1,$AF,$CA,$D7,$F8,$18,$FE
	.dc.b	$70,$7F,$A9,$AB,$FD,$B7,$87,$E5,$04,$7F,$B9,$B0,$FE,$57,$87,$FB
	.dc.b	$6A,$BF,$CC,$A7,$F2,$7A,$FE,$54,$FC,$A3,$FB,$31,$FA,$2A,$FF,$34
	.dc.b	$61,$FE,$04,$C3,$FC,$D0,$FF,$29,$3F,$DB,$6F,$F2,$99,$FF,$33,$C7
	.dc.b	$F2,$A8,$FC,$9C,$FF,$99,$27,$FC,$CB,$1F,$93,$7F,$D8,$87,$F8,$04
	.dc.b	$FE,$6D,$87,$E9,$CF,$C0,$3F,$D6,$1F,$CA,$57,$E0,$1F,$E7,$D1,$FE
	.dc.b	$D1,$3F,$94,$C6,$67,$FC,$D5,$3F,$A8,$9F,$F3,$45,$7F,$80,$5F,$CB
	.dc.b	$57,$FB,$36,$FF,$2D,$1F,$EC,$C8,$FF,$68,$4F,$E5,$8F,$EF,$71,$F9
	.dc.b	$6C,$3F,$BE,$D7,$F8,$0A,$FF,$93,$C7,$C3,$9F,$F6,$19,$FE,$C6,$9F
	.dc.b	$C9,$9F,$C9,$AF,$FA,$AC,$7F,$A9,$9F,$F1,$25,$FF,$27,$84,$7E,$56
	.dc.b	$55,$89,$FF,$62,$D5,$7F,$88,$9F,$E6,$2B,$FE,$5F,$1F,$CA,$8F,$F2
	.dc.b	$E4,$7F,$7E,$AF,$F2,$E1,$FE,$C8,$9F,$E5,$37,$FE,$5B,$AE,$95,$FE
	.dc.b	$BE,$3F,$95,$5F,$E5,$9E,$E7,$FB,$F2,$BF,$C0,$15,$F2,$2B,$FD,$7F
	.dc.b	$3F,$E0,$0A,$FF,$00,$C7,$E5,$63,$A9,$FE,$5B,$7F,$C0,$0F,$E5,$B0
	.dc.b	$FF,$5E,$C7,$F9,$62,$3F,$CB,$31,$FC,$A2,$3F,$2B,$1F,$C0,$A3,$F9
	.dc.b	$46,$1F,$CA,$F0,$FF,$5E,$1F,$E5,$BA,$FF,$29,$D7,$F7,$D3,$FB,$CB
	.dc.b	$FE,$B1,$9F,$DF,$40,$8F,$F9,$61,$F7,$9F,$C2,$0F,$CC,$9F,$CA,$EB
	.dc.b	$FB,$FC,$FF,$7E,$8F,$EF,$F1,$EE,$7F,$D7,$F1,$FE,$00,$7F,$96,$61
	.dc.b	$F9,$7A,$FE,$58,$FE,$66,$FF,$C0,$0F,$F1,$EA,$FE,$01,$5F,$D3,$9F
	.dc.b	$F3,$7B,$F8,$07,$FC,$04,$7F,$B2,$4F,$43,$FE,$E2,$7F,$35,$1F,$EB
	.dc.b	$78,$FC,$CE,$1F,$9A,$9F,$F2,$BD,$7E,$6A,$BF,$A7,$1F,$96,$C3,$F9
	.dc.b	$65,$7F,$95,$A3,$13,$FE,$56,$7F,$32,$7F,$2B,$BF,$EF,$F8,$7E,$62
	.dc.b	$3F,$35,$87,$C3,$9F,$CB,$1F,$DF,$EF,$0F,$FB,$8A,$3F,$2F,$3F,$CA
	.dc.b	$E3,$FC,$00,$7F,$95,$4F,$E5,$73,$FD,$46,$3F,$94,$47,$E5,$9F,$E5
	.dc.b	$55,$FE,$B0,$8F,$06,$0C,$3F,$61,$1F,$EE,$0C,$3F,$96,$9F,$ED,$13
	.dc.b	$FC,$DA,$FF,$AC,$5F,$F0,$04,$7F,$2C,$BF,$E5,$13,$FD,$FC,$FE,$59
	.dc.b	$3F,$97,$8F,$EF,$C7,$FB,$9A,$3F,$2C,$7E,$35,$FE,$58,$7F,$31,$AF
	.dc.b	$EF,$F5,$F9,$87,$F8,$54,$FE,$F2,$FF,$02,$FE,$54,$FF,$14,$DF,$F9
	.dc.b	$58,$FC,$D1,$FE,$AF,$3F,$9D,$C7,$F8,$6E,$38,$57,$F9,$05,$FF,$40
	.dc.b	$D7,$E6,$2F,$FC,$03,$1F,$98,$7F,$97,$3F,$EA,$F9,$FE,$5E,$FF,$87
	.dc.b	$94,$FF,$38,$3C,$73,$F9,$99,$FF,$02,$BF,$C2,$CF,$CD,$5F,$A5,$B1
	.dc.b	$FE,$56,$BF,$E6,$31,$AB,$FF,$03,$57,$EF,$E3,$F3,$07,$F2,$8A,$FF
	.dc.b	$25,$1F,$DB,$8F,$E8,$E7,$EE,$6F,$FC,$0C,$FF,$A0,$A7,$FB,$2C,$FE
	.dc.b	$6E,$3F,$9A,$61,$F9,$B7,$FC,$13,$7F,$9B,$7F,$7F,$87,$C9,$7D,$EF
	.dc.b	$D2,$6D,$FF,$81,$CF,$82,$FF,$24,$C3,$FC,$B7,$87,$F8,$36,$7F,$99
	.dc.b	$1F,$EA,$E9,$FF,$03,$9B,$BF,$F0,$39,$EA,$7E,$6B,$FC,$4E,$3F,$C1
	.dc.b	$07,$E6,$9F,$A7,$1E,$F7,$CB,$87,$F3,$58,$FF,$7B,$C7,$F3,$47,$F3
	.dc.b	$D3,$F0,$9F,$F6,$91,$FE,$5D,$9F,$E6,$B5,$F1,$E7,$FC,$BA,$7F,$34
	.dc.b	$9F,$F0,$55,$FF,$34,$9F,$F5,$8C,$7F,$34,$7F,$3C,$FD,$57,$F3,$AF
	.dc.b	$F3,$5C,$3F,$9A,$3F,$EB,$08,$FE,$1A,$FF,$82,$0F,$9F,$5F,$9B,$9F
	.dc.b	$F2,$D4,$7F,$82,$27,$F8,$6B,$FC,$C0,$FF,$57,$3F,$96,$8F,$F2,$D4
	.dc.b	$7F,$83,$30,$FF,$03,$BF,$CE,$1F,$8E,$FF,$3D,$7F,$32,$FF,$82,$0D
	.dc.b	$5F,$F8,$20,$FF,$11,$C7,$F9,$36,$3F,$AF,$DF,$E0,$5F,$F1,$7B,$FE
	.dc.b	$CD,$3F,$98,$C7,$F9,$7A,$7F,$DC,$33,$FC,$36,$FF,$32,$FE,$14,$FE
	.dc.b	$9B,$87,$F3,$18,$FC,$E5,$7F,$31,$9F,$EB,$97,$F5,$DF,$F0,$59,$FC
	.dc.b	$68,$FB,$F3,$FC,$D7,$0F,$F0,$64,$7F,$98,$A7,$F9,$83,$F3,$1E,$4F
	.dc.b	$F9,$AA,$3F,$35,$87,$F0,$A7,$71,$FE,$0A,$8C,$4F,$F2,$89,$FE,$CB
	.dc.b	$3F,$A0,$8F,$CD,$D7,$F3,$58,$FC,$D9,$C6,$FF,$36,$FF,$2F,$3F,$9B
	.dc.b	$C7,$F0,$37,$F4,$17,$FC,$0D,$B7,$FE,$B6,$3F,$9C,$4F,$F9,$8E,$FF
	.dc.b	$CC,$47,$AA,$7F,$D6,$A7,$F8,$36,$3F,$31,$39,$8F,$E7,$17,$FC,$DD
	.dc.b	$FE,$6E,$7E,$6A,$7F,$9C,$CF,$F9,$8A,$3F,$C9,$11,$F9,$A7,$EB,$1F
	.dc.b	$D5,$0F,$C3,$DF,$F9,$89,$FF,$08,$57,$F3,$79,$FC,$DC,$7F,$89,$2B
	.dc.b	$F0,$C7,$F7,$67,$FD,$8A,$FF,$9D,$8F,$CE,$C7,$FA,$69,$FE,$EF,$1F
	.dc.b	$CF,$EF,$F9,$83,$FE,$65,$3F,$D6,$EF,$F8,$36,$7F,$D6,$F7,$FD,$A2
	.dc.b	$FF,$D2,$75,$FC,$D6,$3E,$63,$FE,$36,$8F,$4C,$FF,$83,$EB,$85,$7E
	.dc.b	$7E,$7F,$C1,$F5,$FD,$8C,$FE,$AC,$58,$FF,$5D,$CF,$F7,$63,$F0,$71
	.dc.b	$FA,$29,$FE,$6F,$5F,$26,$3F,$AB,$57,$F9,$52,$3F,$9C,$5F,$FA,$F4
	.dc.b	$FF,$34,$4F,$E6,$CF,$F5,$A1,$FA,$F3,$FD,$7E,$7E,$E2,$7F,$9E,$47
	.dc.b	$F9,$AA,$FF,$45,$1F,$EB,$C7,$F3,$33,$F2,$5F,$C1,$3F,$CD,$8F,$F0
	.dc.b	$9E,$1F,$9B,$8F,$E6,$F1,$F9,$FA,$FC,$ED,$7F,$6C,$9F,$CE,$CF,$F3
	.dc.b	$53,$E5,$9F,$ED,$2C,$3F,$73,$87,$F8,$56,$FF,$CD,$53,$F9,$E3,$FD
	.dc.b	$1F,$5F,$CD,$EF,$8C,$7C,$4B,$FD,$CB,$FA,$09,$D4,$7F,$37,$8F,$F1
	.dc.b	$C4,$7F,$9A,$AF,$F3,$B8,$7F,$36,$9F,$E9,$D3,$E8,$3F,$89,$C7,$FA
	.dc.b	$FE,$3F,$3C,$7F,$AF,$70,$FC,$F3,$FC,$7A,$BE,$B3,$F3,$8F,$CF,$61
	.dc.b	$FE,$6A,$BB,$87,$E7,$67,$F8,$9C,$7F,$8D,$AF,$F3,$AF,$F3,$73,$FD
	.dc.b	$7D,$7F,$9B,$7F,$C2,$27,$E6,$A7,$FC,$CD,$1F,$CD,$E3,$F3,$51,$FC
	.dc.b	$DE,$FF,$34,$FF,$83,$DF,$8E,$FF,$86,$1F,$F0,$75,$7F,$99,$A3,$F9
	.dc.b	$BD,$FE,$61,$FE,$6B,$5C,$1F,$ED,$0F,$F3,$53,$EF,$BF,$E6,$77,$FC
	.dc.b	$CC,$7F,$B8,$0F,$EF,$CF,$FB,$86,$FF,$9B,$BF,$C8,$CF,$BE,$7F,$04
	.dc.b	$AF,$CC,$C7,$F8,$4E,$7F,$23,$3F,$CA,$5F,$C8,$DF,$F8,$4E,$3F,$C1
	.dc.b	$F3,$FC,$DE,$7F,$33,$7F,$E0,$F9,$FE,$11,$1F,$D3,$27,$FC,$D5,$1F
	.dc.b	$97,$BE,$AF,$E8,$A7,$D6,$FE,$BA,$3F,$32,$7E,$0A,$7F,$C7,$33,$FE
	.dc.b	$6A,$9F,$CD,$4F,$FB,$02,$3F,$C1,$51,$FC,$D6,$7F,$CD,$D1,$FE,$0E
	.dc.b	$7F,$CD,$71,$F9,$99,$FE,$69,$3F,$96,$8F,$CE,$3F,$CD,$0F,$CB,$3F
	.dc.b	$94,$51,$E8,$8F,$F2,$3C,$7F,$21,$3F,$D8,$53,$FE,$0F,$7F,$C1,$CF
	.dc.b	$F3,$6B,$FF,$07,$4F,$FA,$C0,$FF,$12,$1F,$83,$8F,$CC,$BF,$C3,$8F
	.dc.b	$E2,$E7,$FA,$5D,$FF,$36,$57,$F8,$41,$FF,$37,$5F,$FB,$0E,$FF,$34
	.dc.b	$7E,$1E,$BF,$CD,$67,$FB,$12,$BF,$9A,$BF,$E1,$49,$FF,$63,$CF,$F0
	.dc.b	$C8,$FE,$B8,$7E,$70,$FF,$07,$1F,$50,$FF,$37,$9F,$34,$FC,$F5,$7D
	.dc.b	$48,$FE,$63,$1E,$83,$FD,$F3,$3F,$CD,$9F,$F0,$9D,$7E,$7D,$FE,$75
	.dc.b	$7F,$E1,$43,$F0,$CF,$F3,$73,$F9,$A5,$7E,$7E,$7F,$CD,$71,$FE,$9D
	.dc.b	$AF,$E6,$AF,$CB,$38,$9F,$9D,$3F,$51,$5F,$E8,$D9,$FE,$6D,$5F,$EC
	.dc.b	$49,$FE,$6C,$FD,$B7,$FD,$49,$7F,$E6,$B7,$F9,$9C,$FF,$35,$8F,$F6
	.dc.b	$2C,$7F,$36,$9E,$13,$FE,$6B,$AF,$9F,$3F,$9D,$C3,$F3,$D5,$FC,$D7
	.dc.b	$0F,$89,$3F,$9D,$81,$FC,$D1,$FF,$07,$1E,$88,$FD,$AB,$8B,$FE,$67
	.dc.b	$87,$E7,$2F,$FD,$8D,$3F,$E1,$03,$FD,$8C,$7F,$31,$8F,$F7,$EE,$1F
	.dc.b	$EC,$53,$EC,$47,$F8,$6D,$F0,$CF,$F8,$F2,$3F,$C2,$33,$FC,$D0,$FF
	.dc.b	$08,$E1,$F9,$C8,$FD,$13,$F9,$C7,$FD,$8F,$5F,$9C,$7F,$D5,$B3,$FE
	.dc.b	$C6,$7E,$A4,$FE,$35,$FF,$2B,$9F,$C2,$67,$FC,$20,$FF,$B8,$E7,$F3
	.dc.b	$71,$FE,$C6,$8F,$F0,$6C,$FE,$6E,$FF,$C2,$11,$F6,$E3,$FD,$8B,$87
	.dc.b	$CC,$8C,$3F,$9C,$8F,$E6,$31,$F9,$B3,$FC,$10,$7E,$6C,$FF,$34,$1F
	.dc.b	$10,$FC,$B9,$FE,$66,$3F,$9C,$1D,$4F,$F3,$75,$7F,$38,$8F,$F0,$7C
	.dc.b	$E6,$7E,$73,$FC,$E5,$FE,$68,$66,$3D,$2F,$F3,$AA,$FC,$DB,$FC,$FA
	.dc.b	$7F,$36,$FF,$B3,$2F,$F9,$BB,$FD,$6E,$7F,$33,$1F,$E6,$C8,$FC,$CE
	.dc.b	$1F,$1D,$F9,$87,$F3,$43,$FC,$E9,$87,$BC,$FF,$56,$3F,$99,$8F,$E0
	.dc.b	$F8,$7F,$3C,$BF,$F3,$49,$FE,$3B,$8F,$CC,$DF,$E2,$A3,$F9,$9C,$FE
	.dc.b	$22,$3F,$80,$1F,$E6,$7A,$FE,$63,$3F,$CC,$E3,$F2,$38,$7F,$B0,$70
	.dc.b	$FF,$34,$5F,$F3,$3B,$FC,$C1,$FC,$CF,$0F,$F5,$F3,$FE,$13,$9F,$A8
	.dc.b	$FF,$83,$5F,$03,$FE,$C0,$9F,$E6,$AF,$F5,$57,$F9,$AC,$FF,$82,$E3
	.dc.b	$15,$FE,$67,$7F,$83,$1F,$E0,$A3,$FB,$E5,$FF,$B0,$0F,$CC,$3F,$44
	.dc.b	$FE,$6D,$7F,$CA,$DF,$E3,$47,$C3,$BF,$F0,$5B,$FE,$C1,$3F,$C1,$73
	.dc.b	$FE,$07,$7F,$09,$87,$E0,$1F,$F4,$8C,$7F,$82,$EF,$35,$FE,$0C,$3F
	.dc.b	$99,$29,$FF,$33,$5F,$F8,$2A,$FF,$D8,$71,$F9,$A9,$FF,$61,$BF,$CA
	.dc.b	$8F,$C9,$3E,$67,$F9,$8C,$FE,$7D,$FC,$A9,$F9,$87,$F8,$4B,$FE,$0E
	.dc.b	$3F,$2D,$5F,$CB,$CF,$E6,$33,$FE,$C1,$BF,$EB,$11,$FE,$0C,$9F,$F3
	.dc.b	$A9,$FE,$03,$3F,$97,$9F,$C6,$0F,$CC,$3F,$CB,$E3,$FC,$19,$87,$E5
	.dc.b	$E7,$FD,$22,$7E,$5C,$FF,$0A,$5F,$E5,$D3,$FE,$63,$8F,$CB,$D7,$F9
	.dc.b	$7E,$7F,$CC,$35,$F4,$F0,$FC,$D4,$FE,$22,$7F,$C1,$8F,$F2,$D7,$FC
	.dc.b	$BC,$7F,$2D,$3F,$0F,$3F,$E5,$F8,$FF,$04,$E1,$FB,$FC,$3F,$2F,$87
	.dc.b	$C6,$BF,$CA,$1F,$CB,$8F,$E5,$58,$7F,$AF,$1F,$F6,$0D,$FF,$82,$70
	.dc.b	$FE,$62,$FF,$B5,$0F,$D6,$3F,$CB,$0C,$BE,$78,$FF,$05,$BF,$E8,$87
	.dc.b	$F2,$07,$E0,$EF,$F9,$73,$FD,$D2,$7F,$2E,$FF,$BC,$CF,$CB,$9F,$C4
	.dc.b	$CF,$C1,$1F,$E8,$39,$FC,$BC,$7F,$2E,$3F,$D7,$B1,$FE,$B5,$3F,$26
	.dc.b	$7F,$81,$E3,$FC,$C2,$FF,$81,$E7,$E8,$CF,$F8,$67,$0F,$CB,$4F,$F2
	.dc.b	$D3,$F9,$4D,$FE,$59,$FE,$57,$1F,$96,$7F,$26,$7E,$5A,$BF,$95,$57
	.dc.b	$37,$FD,$70,$7C,$F8,$EA,$7F,$66,$BF,$F0,$E1,$FA,$B3,$FC,$61,$3F
	.dc.b	$94,$7F,$CC,$35,$FC,$AA,$78,$9E,$68,$FE,$DF,$AF,$F6,$48,$EA,$FF
	.dc.b	$AE,$B1,$3F,$EB,$C9,$B3,$FE,$E8,$AF,$DA,$4F,$E6,$8F,$F0,$86,$1E
	.dc.b	$57,$FC,$10,$75,$7F,$89,$BF,$E6,$53,$FC,$12,$66,$97,$DA,$FF,$CC
	.dc.b	$F1,$E6,$7D,$27,$FA,$FC,$D1,$F1,$2B,$FC,$A7,$1D,$A3,$71,$8C,$79
	.dc.b	$47,$E0,$70,$F4,$BA,$8F,$F5,$6E,$1F,$E3,$93,$F9,$D1,$FE,$C1,$C3
	.dc.b	$BC,$FF,$AF,$4E,$F1,$FC,$F1,$57,$47,$FC,$33,$87,$B2,$7E,$E3,$E2
	.dc.b	$8F,$D4,$9D,$0F,$CE,$4F,$EF,$AF,$84,$7F,$B2,$6B,$FA,$4E,$1D,$A3
	.dc.b	$81,$FC,$12,$BF,$CE,$38,$62,$3F,$0D,$87,$77,$FD,$EE,$FF,$8B,$9F
	.dc.b	$31,$F2,$5F,$EB,$D7,$FE,$1E,$3F,$67,$5A,$9E,$33,$FE,$26,$9F,$E9
	.dc.b	$31,$93,$E7,$BF,$D3,$30,$F0,$47,$F8,$2A,$3F,$A7,$D7,$F1,$A7,$FA
	.dc.b	$8D,$6A,$7F,$CF,$C7,$A4,$FE,$3E,$6C,$FE,$A3,$5B,$7F,$BC,$CF,$F8
	.dc.b	$98,$FC,$8C,$E6,$7D,$D3,$F0,$DF,$81,$85,$9F,$B3,$3F,$E8,$13,$17
	.dc.b	$DD,$FC,$8C,$7C,$B8,$FD,$2C,$7F,$A0,$8D,$1F,$BE,$8D,$3F,$56,$FF
	.dc.b	$82,$3F,$E4,$F7,$F1,$8F,$E6,$8F,$F2,$91,$F2,$4F,$F1,$59,$D0,$FD
	.dc.b	$FC,$FE,$0E,$7C,$55,$D6,$7F,$DC,$AE,$23,$E2,$1F,$02,$BF,$81,$4F
	.dc.b	$33,$F8,$04,$7C,$07,$FC,$5C,$7E,$1D,$FF,$38,$D6,$E3,$FD,$DB,$1F
	.dc.b	$3D,$FF,$77,$CF,$F0,$77,$F8,$FB,$FC,$1C,$FE,$0C,$7F,$8C,$CF,$E0
	.dc.b	$8F,$F8,$22,$3E,$F5,$7F,$65,$9F,$E0,$97,$F9,$EA,$FE,$BE,$E9,$FC
	.dc.b	$E1,$F9,$E9,$FE,$81,$5B,$8F,$E3,$73,$F9,$EA,$FF,$75,$9F,$D8,$70
	.dc.b	$E1,$1F,$ED,$38,$FF,$0B,$61,$F3,$F0,$FE,$0F,$87,$3C,$3E,$79,$FC
	.dc.b	$64,$FE,$13,$3F,$80,$7F,$D2,$18,$78,$E3,$FC,$29,$87,$B1,$FF,$34
	.dc.b	$E1,$FE,$93,$3F,$7F,$7B,$C3,$E3,$E1,$D1,$FE,$2C,$FF,$68,$8D,$57
	.dc.b	$94,$FD,$DD,$7F,$3B,$8F,$92,$7D,$D9,$FF,$79,$5F,$A5,$F8,$B1,$F7
	.dc.b	$5F,$DE,$3F,$87,$7F,$C8,$8E,$DF,$E5,$73,$FE,$F3,$3F,$71,$1F,$B1
	.dc.b	$8E,$07,$E5,$EB,$F0,$8F,$E2,$A7,$F9,$FB,$F9,$93,$FD,$34,$FF,$BD
	.dc.b	$27,$F7,$57,$CA,$7F,$D5,$D3,$FB,$69,$FD,$94,$71,$BF,$DA,$CF,$E9
	.dc.b	$63,$F9,$C4,$FE,$D1,$FD,$AD,$7E,$E6,$3F,$81,$5F,$D5,$9F,$F4,$B9
	.dc.b	$E6,$7E,$AC,$7E,$D1,$F3,$1F,$E9,$8A,$FD,$9C,$FF,$91,$CF,$D6,$9C
	.dc.b	$6B,$E5,$3F,$B2,$7D,$4F,$F7,$E9,$FE,$61,$1F,$B1,$AF,$D5,$1F,$7A
	.dc.b	$7F,$C9,$51,$8B,$FD,$AC,$F2,$9F,$F2,$E3,$FB,$09,$E2,$FE,$BE,$3F
	.dc.b	$24,$7F,$BE,$4F,$CC,$1F,$90,$C3,$F4,$8F,$BB,$0F,$D8,$3F,$D8,$0E
	.dc.b	$B1,$F4,$A7,$F5,$95,$E5,$9F,$ED,$07,$F4,$F9,$E7,$1F,$02,$3F,$DA
	.dc.b	$98,$78,$23,$FA,$04,$7F,$19,$C3,$F6,$0F,$F6,$9B,$E8,$FF,$BE,$E3
	.dc.b	$F1,$D3,$F6,$B0,$FD,$8C,$FE,$D2,$B9,$BF,$ED,$27,$F5,$4F,$23,$FD
	.dc.b	$6B,$5C,$AF,$DD,$1F,$B4,$3E,$55,$FF,$95,$2F,$F5,$E7,$F6,$68,$FD
	.dc.b	$54,$7E,$B9,$FF,$55,$57,$ED,$E3,$8D,$7E,$BA,$7F,$D5,$55,$FE,$52
	.dc.b	$7F,$83,$5F,$F6,$47,$F5,$EF,$F1,$33,$F9,$3B,$5F,$5B,$F9,$93,$F5
	.dc.b	$47,$EA,$0F,$49,$FD,$12,$3F,$B1,$D7,$EB,$E7,$F9,$73,$F0,$67,$FC
	.dc.b	$96,$FF,$97,$E7,$F5,$87,$F8,$BA,$BF,$61,$3F,$92,$8F,$D6,$1F,$EA
	.dc.b	$F7,$F5,$B3,$FD,$7E,$FF,$B2,$5F,$E4,$67,$FC,$4F,$7F,$BC,$9F,$B7
	.dc.b	$1F,$E2,$E8,$FD,$63,$FC,$90,$FD,$61,$F8,$53,$FD,$68,$FF,$20,$7F
	.dc.b	$83,$1F,$D7,$CB,$3F,$B6,$AF,$64,$7F,$60,$7F,$AD,$D6,$6B,$F6,$B1
	.dc.b	$C6,$7C,$4F,$E8,$9F,$F2,$CD,$7F,$21,$7F,$5B,$7F,$E5,$FB,$FD,$6D
	.dc.b	$FF,$13,$AF,$D6,$4F,$F2,$97,$E7,$1F,$A8,$AF,$9D,$7F,$E7,$8B,$FC
	.dc.b	$BC,$7D,$63,$F4,$A7,$E3,$8F,$D5,$D7,$CE,$7F,$5D,$3C,$9F,$D5,$47
	.dc.b	$F9,$75,$CD,$FE,$A6,$FF,$98,$47,$F5,$07,$F4,$4F,$F5,$3B,$FC,$73
	.dc.b	$FE,$95,$7F,$B3,$9F,$8E,$AF,$F7,$74,$F2,$AF,$EE,$75,$FA,$33,$FA
	.dc.b	$41,$F9,$F3,$FC,$AD,$5F,$AC,$8C,$CF,$1B,$F9,$95,$F8,$E9,$FF,$17
	.dc.b	$BF,$D5,$EB,$F6,$13,$F7,$CF,$F7,$3D,$FF,$94,$AF,$D6,$7E,$22,$BF
	.dc.b	$A7,$9F,$A8,$7C,$4F,$F5,$4A,$FE,$B3,$1F,$B8,$3F,$CD,$F1,$FE,$4A
	.dc.b	$9F,$EA,$4F,$EA,$67,$F9,$15,$FE,$7A,$3F,$08,$7F,$8A,$6B,$F4,$EF
	.dc.b	$E8,$AF,$F4,$CF,$F9,$1A,$3F,$A8,$CE,$4F,$D9,$4F,$C0,$9F,$C0,$4F
	.dc.b	$E5,$2B,$FA,$8C,$7F,$1D,$8F,$CF,$3F,$A6,$3F,$A8,$DF,$F5,$59,$F8
	.dc.b	$11,$FA,$53,$FC,$82,$7E,$99,$FE,$AA,$FE,$BE,$FF,$A8,$3F,$A4,$AF
	.dc.b	$E9,$EF,$F5,$0B,$FD,$1C,$7F,$8F,$A7,$F4,$67,$F5,$37,$FA,$4C,$FE
	.dc.b	$75,$F7,$BF,$EA,$47,$FC,$3C,$FC,$57,$F7,$C7,$E8,$9F,$E3,$A7,$CF
	.dc.b	$3F,$8B,$1F,$C5,$4F,$79,$FB,$B3,$FD,$A5,$7F,$E7,$68,$FE,$F7,$1F
	.dc.b	$3C,$F9,$77,$FC,$62,$BE,$75,$FF,$4F,$AF,$EF,$B5,$FC,$4A,$FE,$44
	.dc.b	$7C,$CB,$FC,$E4,$FF,$BA,$E7,$33,$F2,$23,$FB,$B9,$EE,$3F,$CD,$D1
	.dc.b	$FE,$6C,$BC,$9F,$C5,$27,$F8,$74,$FF,$0A,$3F,$BD,$57,$F0,$D7,$E6
	.dc.b	$5F,$F7,$7B,$FF,$32,$C7,$F7,$83,$FC,$3B,$5F,$29,$FF,$32,$C7,$FB
	.dc.b	$AE,$F3,$3F,$89,$9F,$F1,$49,$F2,$AB,$FC,$51,$7F,$98,$8F,$F1,$55
	.dc.b	$7C,$DA,$FC,$BC,$7C,$AA,$FF,$12,$D7,$C9,$8F,$CB,$C7,$C9,$BF,$E7
	.dc.b	$47,$C8,$8F,$E6,$D5,$F2,$4F,$E7,$55,$F7,$63,$FD,$2F,$3F,$E9,$07
	.dc.b	$E4,$9F,$93,$9F,$F1,$13,$FE,$23,$AF,$F4,$BC,$7C,$88,$FC,$94,$FF
	.dc.b	$8C,$6F,$FD,$20,$DA,$FC,$D1,$FB,$29,$FF,$12,$57,$E5,$1F,$95,$1F
	.dc.b	$E1,$F8,$FF,$11,$D7,$E5,$CF,$92,$FF,$87,$67,$F7,$87,$E4,$6F,$E4
	.dc.b	$5F,$F6,$43,$F2,$07,$F9,$3A,$3E,$43,$FD,$89,$FF,$23,$5F,$F5,$B7
	.dc.b	$FB,$11,$FE,$8F,$AF,$C8,$CF,$C9,$9F,$EC,$51,$FE,$4B,$8F,$C9,$DF
	.dc.b	$CC,$8F,$C8,$57,$EE,$AB,$FD,$77,$7F,$BA,$AF,$C7,$C7,$C8,$AF,$DC
	.dc.b	$DF,$C5,$BF,$C7,$C7,$EE,$5F,$C7,$C7,$F8,$51,$FD,$A1,$FB,$93,$F1
	.dc.b	$77,$FD,$61,$F8,$F5,$FE,$42,$AF,$F0,$4C,$7F,$82,$A3,$F1,$B1,$F8
	.dc.b	$C8,$FC,$6C,$7F,$5C,$AF,$8C,$7F,$9D,$1F,$C6,$CF,$F2,$29,$FC,$6C
	.dc.b	$FF,$59,$47,$FA,$AA,$3F,$91,$1F,$AF,$72,$7F,$9A,$AB,$FC,$13,$7F
	.dc.b	$E7,$79,$FC,$14,$FF,$2A,$7E,$4D,$7E,$46,$FE,$4C,$7E,$42,$3F,$71
	.dc.b	$1F,$25,$F9,$13,$F9,$03,$FD,$1D,$1F,$CC,$2B,$FD,$83,$3F,$91,$7F
	.dc.b	$91,$57,$ED,$EB,$F1,$F3,$FE,$01,$8F,$C7,$C7,$CA,$8F,$F0,$01,$FC
	.dc.b	$79,$FC,$7D,$7C,$A8,$FE,$43,$5F,$D3,$4F,$DB,$57,$F1,$EA,$FC,$79
	.dc.b	$FD,$2E,$7F,$C6,$0F,$CA,$3F,$1C,$FE,$D9,$FE,$90,$7E,$DA,$FE,$3B
	.dc.b	$FD,$20,$FE,$99,$3F,$19,$FF,$13,$9F,$B5,$9F,$E3,$8F,$ED,$5F,$F4
	.dc.b	$AB,$F4,$8F,$F1,$7D,$FF,$8A,$AC,$FD,$78,$FF,$4A,$3F,$24,$FD,$BC
	.dc.b	$7C,$89,$FE,$3E,$7F,$AB,$6B,$FA,$FD,$7F,$49,$7F,$AF,$9F,$CD,$1C
	.dc.b	$4F,$EE,$23,$F2,$73,$FD,$DA,$7F,$9D,$BF,$29,$FC,$9B,$FB,$87,$F6
	.dc.b	$A7,$F4,$7A,$FC,$9C,$7C,$A3,$FC,$B4,$FF,$1F,$8F,$92,$7F,$47,$3F
	.dc.b	$B9,$5F,$CA,$9F,$DB,$CF,$FB,$06,$BF,$B7,$D7,$ED,$E7,$F8,$C3,$F8
	.dc.b	$93,$FB,$81,$F9,$63,$FC,$AE,$7F,$39,$17,$FE,$20,$8F,$DD,$57,$F7
	.dc.b	$27,$F2,$13,$FB,$A3,$F2,$B5,$FC,$7E,$3E,$53,$F4,$63,$FD,$76,$FC
	.dc.b	$A3,$FB,$89,$E1,$4F,$F8,$21,$FB,$87,$F9,$90,$FF,$3F,$C7,$FB,$C8
	.dc.b	$F9,$87,$F2,$37,$FD,$19,$1F,$D2,$A3,$F7,$C7,$F2,$38,$E6,$7E,$F2
	.dc.b	$7F,$92,$1E,$17,$F0,$F3,$FE,$82,$7F,$74,$7E,$5D,$FE,$48,$FF,$73
	.dc.b	$3F,$2E,$FC,$E9,$FD,$E1,$F3,$AF,$F9,$24,$FE,$5C,$FF,$0E,$47,$EF
	.dc.b	$0F,$96,$7F,$49,$8F,$EE,$4F,$EE,$EB,$FB,$94,$FF,$24,$3F,$76,$7F
	.dc.b	$86,$67,$F9,$24,$FF,$71,$9F,$E4,$8F,$EE,$E7,$F9,$23,$F1,$DF,$F5
	.dc.b	$85,$FE,$EE,$7F,$92,$47,$E4,$8F,$DD,$9F,$18,$FE,$BA,$7F,$4A,$8F
	.dc.b	$F0,$CC,$FF,$25,$8F,$E4,$87,$C6,$9F,$E9,$31,$FD,$22,$3F,$78,$7F
	.dc.b	$70,$3F,$76,$7E,$4C,$FF,$51,$D7,$C6,$8F,$E4,$27,$EF,$23,$F7,$95
	.dc.b	$F9,$17,$F9,$09,$FD,$2A,$3F,$22,$7C,$73,$FA,$53,$FB,$73,$FA,$55
	.dc.b	$7F,$6F,$8F,$F0,$B4,$7F,$21,$3F,$C7,$E7,$E5,$9F,$F5,$BC,$FF,$A7
	.dc.b	$23,$FA,$0C,$61,$FF,$67,$9F,$D1,$29,$82,$75,$3F,$8D,$3F,$DE,$75
	.dc.b	$FA,$A7,$89,$63,$6F,$FA,$33,$0E,$71,$88,$FC,$B4,$7E,$92,$37,$87
	.dc.b	$59,$F0,$1F,$B6,$AF,$E4,$51,$F9,$DC,$39,$CE,$4F,$4C,$23,$A4,$64
	.dc.b	$F4,$47,$95,$C1,$FC,$A8,$FD,$2E,$1E,$0A,$F6,$47,$F5,$DC,$31,$1E
	.dc.b	$2B,$FD,$DC,$FC,$A3,$9D,$F8,$6B,$C5,$7C,$27,$FD,$F7,$5F,$E7,$6B
	.dc.b	$EB,$1F,$CE,$A3,$FC,$C7,$18,$BE,$57,$EE,$8D,$D7,$CF,$8F,$6C,$6C
	.dc.b	$FE,$8D,$87,$CA,$7E,$D6,$1E,$1A,$FE,$09,$1F,$60,$F9,$67,$FB,$E4
	.dc.b	$E4,$FF,$16,$8E,$0F,$F5,$0A,$EF,$3C,$2F,$FC,$0B,$7D,$DF,$7B,$70
	.dc.b	$DD,$FF,$B3,$67,$E8,$57,$9D,$ED,$5F,$83,$BF,$E4,$A7,$0A,$FC,$2B
	.dc.b	$D9,$F9,$AF,$F2,$AB,$F8,$97,$F8,$48,$F9,$8D,$7B,$B5,$DC,$FF,$01
	.dc.b	$3D,$A7,$E3,$2B,$F2,$8F,$C5,$3F,$95,$CF,$DD,$3E,$8B,$FE,$BA,$6B
	.dc.b	$E8,$8E,$37,$E7,$9D,$BF,$ED,$46,$DF,$B6,$BF,$DE,$07,$88,$FF,$2C
	.dc.b	$C7,$0B,$D4,$7E,$2C,$F1,$47,$F1,$4A,$F4,$D7,$C0,$9F,$F0,$33,$F8
	.dc.b	$A9,$FF,$1E,$1F,$E0,$6B,$FD,$84,$7C,$EB,$F8,$33,$F8,$8A,$FD,$19
	.dc.b	$FA,$0B,$CB,$F9,$BB,$FD,$C4,$FE,$1C,$FE,$B2,$7C,$32,$C7,$C3,$39
	.dc.b	$BF,$C7,$4F,$CC,$9F,$AF,$9F,$09,$D6,$36,$FD,$87,$F9,$2B,$FD,$D8
	.dc.b	$DD,$7C,$43,$F6,$31,$F6,$9F,$F7,$CC,$FE,$99,$F8,$91,$FE,$5A,$7E
	.dc.b	$23,$FE,$2B,$7F,$DC,$91,$B7,$D0,$FD,$C7,$EB,$4F,$F7,$6A,$FE,$2B
	.dc.b	$5F,$AC,$6B,$E9,$AF,$DC,$3F,$14,$FE,$47,$5F,$90,$7F,$29,$3C,$AB
	.dc.b	$F0,$D5,$FC,$72,$7F,$81,$61,$FB,$09,$FE,$3D,$76,$7F,$CC,$67,$F0
	.dc.b	$E3,$F8,$BD,$FD,$C8,$F4,$DF,$F1,$57,$AC,$FF,$10,$7F,$6A,$FF,$16
	.dc.b	$C3,$E2,$BF,$C4,$63,$E8,$3E,$07,$D0,$FF,$8A,$1F,$87,$3F,$87,$3F
	.dc.b	$83,$CE,$AF,$FA,$13,$AA,$E6,$7F,$95,$A7,$FD,$1F,$5D,$23,$EE,$3D
	.dc.b	$2B,$CF,$1F,$C5,$9F,$3C,$FF,$07,$AD,$3F,$A2,$9F,$E1,$95,$CC,$FC
	.dc.b	$E5,$FF,$4A,$7F,$0A,$D7,$9C,$FF,$05,$AF,$BB,$1F,$84,$AF,$E1,$EF
	.dc.b	$95,$F7,$4F,$FB,$3A,$BF,$82,$C7,$DF,$8F,$F0,$51,$E3,$7F,$47,$7F
	.dc.b	$BB,$7E,$99,$FE,$3E,$7D,$AF,$F5,$69,$FE,$0F,$7F,$BE,$8F,$BB,$5F
	.dc.b	$C2,$0F,$DC,$3F,$B4,$3F,$C7,$37,$FC,$00,$E6,$FF,$05,$BC,$57,$F0
	.dc.b	$83,$D9,$1F,$DE,$E7,$F5,$57,$F8,$77,$F8,$6D,$FB,$A7,$FC,$26,$7F
	.dc.b	$17,$7F,$77,$5A,$8F,$B6,$7F,$7F,$8F,$29,$FB,$C7,$DE,$7F,$A4,$8F
	.dc.b	$88,$7F,$A6,$0D,$BF,$1A,$3C,$37,$D2,$3F,$90,$9F,$2A,$BF,$7C,$7B
	.dc.b	$8F,$DE,$BF,$E6,$87,$F3,$35,$FD,$22,$3B,$B6,$7F,$82,$4F,$92,$3F
	.dc.b	$80,$3F,$AB,$7E,$9C,$7F,$B7,$AB,$F3,$51,$FD,$04,$FF,$09,$57,$AA
	.dc.b	$7F,$85,$61,$FC,$DD,$E8,$7F,$85,$F0,$FD,$E5,$F2,$8F,$C3,$5F,$F8
	.dc.b	$52,$F9,$3F,$B8,$3F,$97,$CF,$E6,$1F,$DA,$61,$E4,$8F,$EB,$D8,$7F
	.dc.b	$03,$7F,$6F,$7F,$E0,$DB,$F2,$47,$F6,$89,$FE,$6F,$5F,$B1,$BF,$45
	.dc.b	$7F,$17,$7F,$C2,$AF,$B5,$FD,$A1,$FB,$6B,$FF,$06,$57,$F2,$EA,$FE
	.dc.b	$25,$5F,$CE,$4F,$F0,$05,$FE,$FD,$FF,$34,$9F,$DE,$CE,$F5,$FD,$F2
	.dc.b	$FF,$A5,$D7,$B6,$7F,$C3,$8F,$ED,$23,$FB,$DB,$FB,$44,$7F,$80,$2B
	.dc.b	$F6,$B1,$E5,$3D,$D8,$7E,$CA,$7E,$F9,$FE,$A8,$3F,$4C,$7D,$03,$FA
	.dc.b	$9B,$FB,$8A,$FC,$8C,$7F,$B7,$E3,$FB,$BB,$CA,$FF,$40,$FE,$DE,$FF
	.dc.b	$69,$1F,$A0,$8F,$CA,$9F,$BB,$BF,$D7,$1F,$B2,$9F,$CA,$C7,$F7,$28
	.dc.b	$F1,$3F,$B1,$BF,$CA,$CF,$F5,$07,$FD,$82,$FC,$57,$FB,$4C,$7E,$69
	.dc.b	$F6,$1F,$AD,$8F,$D9,$39,$8F,$DD,$9F,$DB,$8F,$CF,$1F,$D9,$CE,$55
	.dc.b	$EC,$7E,$69,$FD,$AE,$7F,$63,$3A,$3E,$2C,$69,$FE,$1F,$7F,$E3,$5B
	.dc.b	$FD,$91,$CE,$7F,$DB,$F5,$FB,$6A,$FD,$8C,$FB,$67,$F5,$EF,$F1,$D7
	.dc.b	$F4,$E7,$F0,$F3,$FC,$FD,$3F,$B0,$7E,$3D,$F8,$5F,$DB,$4F,$F3,$38
	.dc.b	$E2,$7E,$E2,$3F,$B3,$DF,$EC,$27,$FB,$34,$FE,$6E,$7F,$5D,$7F,$E6
	.dc.b	$8B,$F9,$D1,$FD,$8C,$FD,$4D,$7E,$45,$FE,$DF,$1F,$E0,$03,$F9,$4D
	.dc.b	$7F,$9C,$B0,$FD,$6D,$7E,$92,$BF,$84,$DF,$EB,$4F,$EC,$71,$97,$F6
	.dc.b	$31,$FA,$37,$FC,$43,$5F,$C2,$1F,$D6,$47,$F9,$96,$FF,$59,$3F,$0A
	.dc.b	$7F,$A1,$9F,$E6,$FA,$FE,$BC,$7F,$10,$3F,$54,$FE,$B1,$F5,$57,$F4
	.dc.b	$6A,$FD,$4C,$FF,$19,$9F,$D4,$CF,$E9,$8F,$D4,$9F,$6C,$FE,$44,$FF
	.dc.b	$59,$8C,$9F,$84,$8F,$E5,$38,$7E,$A1,$FE,$44,$64,$FF,$0B,$BF,$E6
	.dc.b	$28,$FD,$61,$98,$FC,$59,$FC,$82,$7F,$50,$78,$6B,$F9,$D9,$FD,$62
	.dc.b	$3F,$8F,$5F,$F5,$F9,$FE,$A5,$1F,$A1,$3C,$15,$FC,$6D,$FE,$A8,$7F
	.dc.b	$96,$1C,$BF,$A4,$7E,$7B,$FD,$56,$7F,$CA,$35,$FA,$28,$FF,$29,$CF
	.dc.b	$49,$FD,$5D,$7C,$E8,$FE,$A0,$FC,$73,$FA,$89,$E7,$3F,$41,$3F,$E4
	.dc.b	$F7,$F6,$6F,$E7,$CF,$F0,$15,$FF,$82,$AF,$F4,$95,$F3,$AF,$FC,$11
	.dc.b	$7F,$C4,$2F,$F8,$A1,$F2,$63,$F8,$5B,$FE,$BA,$3F,$A9,$4E,$67,$F8
	.dc.b	$9C,$7F,$0E,$7E,$61,$C2,$BE,$61,$E6,$8F,$F0,$24,$7C,$C7,$FC,$FE
	.dc.b	$7C,$88,$FE,$1E,$FC,$BA,$FD,$0D,$FF,$A0,$9F,$F0,$8D,$FF,$9E,$E3
	.dc.b	$15,$F2,$DD,$47,$CB,$BF,$EF,$77,$FC,$19,$F9,$47,$B0,$F9,$47,$F1
	.dc.b	$C8,$FF,$3A,$3F,$C2,$63,$F2,$E7,$E2,$CF,$86,$7F,$9B,$CF,$F1,$8C
	.dc.b	$FF,$8D,$A3,$E5,$47,$E5,$AF,$FB,$C5,$FF,$3C,$BF,$CA,$C7,$F7,$BA
	.dc.b	$F9,$35,$FE,$2D,$3F,$39,$5F,$94,$8F,$F1,$0B,$F2,$63,$FD,$EB,$1F
	.dc.b	$21,$F9,$33,$FE,$A8,$9F,$93,$7F,$EA,$89,$F6,$5F,$FA,$9A,$FF,$C4
	.dc.b	$31,$FC,$D8,$FC,$85,$7C,$97,$FD,$47,$3F,$26,$7F,$C3,$71,$F2,$4F
	.dc.b	$D8,$3F,$92,$3F,$C6,$A7,$94,$FE,$3E,$FF,$40,$3F,$24,$7F,$B7,$CF
	.dc.b	$EC,$F3,$F9,$27,$F8,$05,$FE,$4A,$7F,$7B,$1F,$26,$FF,$CA,$F1,$F9
	.dc.b	$08,$FC,$8C,$FC,$98,$F9,$11,$F2,$63,$FC,$A7,$1F,$E5,$7B,$FC,$A1
	.dc.b	$F3,$1F,$DA,$4F,$E2,$8F,$E2,$51,$FB,$B3,$EE,$3F,$E0,$47,$FD,$9C
	.dc.b	$FF,$95,$A3,$E4,$BF,$EC,$A3,$E4,$BF,$BA,$7F,$77,$3F,$E1,$BA,$FE
	.dc.b	$B5,$1F,$26,$3F,$C8,$B7,$FD,$66,$3F,$C8,$F7,$F3,$2F,$F7,$31,$FE
	.dc.b	$6E,$3F,$B9,$C7,$CB,$7F,$93,$DF,$F8,$35,$F9,$2F,$E4,$67,$E4,$BF
	.dc.b	$E0,$EB,$FC,$8C,$7C,$98,$FE,$4D,$1F,$25,$FF,$5D,$4F,$EE,$4F,$8D
	.dc.b	$7F,$E0,$EB,$F9,$35,$FC,$8E,$FF,$21,$7F,$8D,$BF,$DC,$47,$F2,$17
	.dc.b	$F7,$11,$F9,$18,$F8,$EF,$EE,$63,$E3,$BF,$B6,$7E,$3B,$FB,$D8,$FD
	.dc.b	$BC,$7E,$42,$7E,$3C,$7C,$68,$F8,$F1,$FE,$6E,$8F,$DB,$47,$FA,$84
	.dc.b	$FD,$B5,$FE,$36,$BF,$C6,$55,$F8,$D7,$FA,$51,$F1,$4F,$C6,$9F,$D2
	.dc.b	$67,$F1,$AF,$E0,$6F,$E8,$47,$F4,$A7,$FB,$E5,$FF,$4A,$8F,$8E,$FE
	.dc.b	$3D,$FC,$85,$7C,$7A,$F8,$D1,$FB,$67,$F6,$97,$FB,$68,$FF,$15,$5F
	.dc.b	$F4,$8B,$F8,$F0,$8F,$EF,$2F,$C8,$3F,$6A,$7C,$83,$4F,$FA,$66,$3F
	.dc.b	$6F,$1F,$26,$BF,$C5,$55,$F8,$F7,$FC,$59,$5F,$8F,$AF,$93,$1F,$E5
	.dc.b	$DB,$FF,$65,$57,$C9,$BF,$F2,$ED,$FF,$21,$BF,$E3,$B7,$F2,$67,$FD
	.dc.b	$8D,$7F,$C9,$0F,$F2,$C1,$FC,$41,$FF,$2C,$4F,$ED,$A7,$F1,$F3,$FB
	.dc.b	$67,$E7,$4F,$F9,$66,$BF,$1F,$1F,$EC,$7A,$FC,$7C,$FF,$93,$CF,$C7
	.dc.b	$C7,$F1,$D9,$FF,$2A,$1F,$26,$7F,$69,$3F,$B5,$AF,$ED,$F1,$FB,$5A
	.dc.b	$FE,$8B,$1F,$B5,$9F,$F4,$3C,$FF,$93,$9F,$D9,$5F,$C7,$73,$7F,$1E
	.dc.b	$7F,$1D,$1F,$8F,$AF,$8F,$1F,$E1,$98,$FD,$99,$FE,$84,$7E,$3B,$F1
	.dc.b	$8F,$ED,$95,$FD,$0D,$FC,$79,$F2,$A3,$FA,$1D,$FC,$63,$E2,$C7,$F8
	.dc.b	$9A,$7F,$68,$FE,$3A,$3E,$31,$F8,$E3,$F6,$95,$FD,$0E,$BF,$69,$1F
	.dc.b	$1E,$3F,$B5,$CF,$E2,$0F,$E6,$D7,$FE,$16,$7F,$DD,$11,$F2,$27,$F6
	.dc.b	$C7,$C8,$3F,$6D,$5F,$8E,$7F,$8E,$C7,$ED,$9F,$C6,$CF,$ED,$9F,$F3
	.dc.b	$D4,$FE,$3A,$3F,$C8,$53,$F2,$27,$F8,$D1,$FB,$63,$F8,$D1,$FB,$68
	.dc.b	$F9,$11,$FC,$6A,$3F,$6C,$7F,$1A,$3E,$2C,$7F,$A3,$A3,$FD,$27,$7F
	.dc.b	$E2,$98,$FF,$52,$1F,$B6,$9D,$BF,$ED,$B3,$E5,$1F,$E4,$5A,$FF,$0A
	.dc.b	$47,$CA,$3F,$6E,$FC,$A8,$FD,$BD,$7C,$AB,$FD,$AC,$7E,$3E,$BF,$C2
	.dc.b	$75,$F8,$F8,$FE,$D7,$5F,$8F,$AF,$DE,$47,$F6,$59,$F3,$3F,$E3,$93
	.dc.b	$F2,$2F,$F1,$FA,$FD,$BD,$7F,$1F,$8F,$EC,$DB,$7F,$14,$A4,$FF,$45
	.dc.b	$8C,$3F,$EB,$0C,$18,$A7,$35,$FE,$E0,$95,$E6,$3F,$8F,$DF,$FA,$8E
	.dc.b	$B7,$1A,$BD,$3D,$AB,$8C,$F4,$AF,$F5,$FB,$6F,$C2,$D9,$E3,$5F,$E0
	.dc.b	$5A,$FF,$26,$4F,$19,$DC,$FB,$6B,$CB,$1D,$A3,$CB,$1E,$87,$FD,$F3
	.dc.b	$3F,$02,$3D,$67,$28,$FF,$76,$4A,$39,$46,$31,$E2,$F4,$9F,$75,$7C
	.dc.b	$D8,$FA,$F1,$C1,$F4,$4F,$F8,$76,$7E,$7B,$F2,$F0,$F6,$CF,$E4,$E3
	.dc.b	$E7,$56,$23,$FC,$D5,$1E,$07,$FB,$E6,$1D,$AB,$51,$FE,$11,$8F,$83
	.dc.b	$7D,$9F,$9B,$1F,$5A,$BF,$03,$87,$BE,$3C,$B3,$EF,$C3,$9B,$FD,$2C
	.dc.b	$FC,$8D,$F3,$8E,$13,$FE,$E8,$76,$7C,$D6,$BF,$27,$0B,$7F,$0E,$7F
	.dc.b	$61,$3F,$8A,$C3,$D9,$1F,$76,$3E,$CD,$7B,$23,$FD,$2D,$7F,$8C,$76
	.dc.b	$7E,$E4,$F9,$4F,$E4,$2F,$B4,$0F,$CF,$9F,$E3,$53,$DA,$66,$BD,$6F
	.dc.b	$85,$CE,$1F,$22,$3E,$79,$F2,$2F,$35,$DA,$3A,$CF,$C6,$7F,$46,$7E
	.dc.b	$BA,$BA,$BF,$E9,$A7,$E6,$CF,$C7,$BF,$C9,$47,$C7,$6C,$75,$AF,$E2
	.dc.b	$11,$A9,$E7,$5F,$8E,$7E,$BE,$1F,$8D,$8F,$BC,$73,$2D,$F3,$8F,$85
	.dc.b	$5F,$92,$9F,$CC,$9F,$C1,$63,$D1,$5D,$2F,$E1,$4F,$E1,$AF,$D7,$1F
	.dc.b	$85,$7A,$3F,$06,$3F,$2B,$5F,$07,$0F,$B5,$7F,$B4,$3F,$0B,$3F,$68
	.dc.b	$FA,$D5,$FB,$8B,$EB,$3F,$B4,$3F,$17,$7B,$AF,$2C,$7C,$27,$ED,$3F
	.dc.b	$71,$F3,$9F,$85,$3E,$11,$F6,$8F,$CF,$61,$FC,$5A,$7A,$5F,$E1,$E7
	.dc.b	$E1,$9F,$C5,$EB,$F0,$6E,$8F,$C1,$57,$F1,$53,$F8,$D5,$66,$B8,$CF
	.dc.b	$F9,$C6,$FE,$05,$7D,$E8,$EF,$1F,$C4,$30,$FC,$1E,$1E,$49,$FE,$19
	.dc.b	$86,$A3,$9C,$7F,$A2,$DF,$BE,$FB,$8F,$F7,$75,$7D,$49,$FE,$13,$7F
	.dc.b	$75,$F6,$9F,$9F,$BF,$D6,$3F,$7A,$3F,$83,$56,$B0,$F8,$EF,$DF,$C3
	.dc.b	$F0,$55,$97,$F2,$27,$B2,$3F,$A0,$BF,$9B,$6B,$C9,$FD,$FB,$ED,$7F
	.dc.b	$C7,$B1,$F4,$23,$F7,$F5,$FC,$A6,$7E,$DB,$FB,$F7,$EE,$61,$F9,$07
	.dc.b	$DA,$7E,$B8,$FC,$81,$FD,$14,$F7,$47,$3B,$FE,$47,$1F,$B4,$7E,$43
	.dc.b	$FB,$89,$E9,$1E,$58,$E4,$F4,$7F,$87,$3F,$49,$FE,$CD,$5D,$1F,$E8
	.dc.b	$E7,$D7,$9F,$D2,$9F,$99,$AF,$DE,$DF,$DB,$AF,$DC,$BE,$EA,$FD,$B9
	.dc.b	$F5,$8F,$90,$7D,$C8,$FB,$F1,$97,$8D,$F8,$67,$F9,$C9,$F6,$6F,$F0
	.dc.b	$51,$EB,$8F,$E9,$8F,$AD,$FD,$7D,$7E,$D6,$FF,$AC,$5F,$DB,$9F,$F3
	.dc.b	$89,$F5,$5F,$1B,$FC,$12,$FF,$8A,$47,$F3,$98,$FE,$7E,$FE,$CD,$F9
	.dc.b	$A7,$F7,$79,$FB,$8F,$8C,$FF,$6A,$57,$EE,$DF,$D8,$3F,$E1,$48,$DC
	.dc.b	$FF,$2D,$9F,$E0,$A6,$CF,$E5,$F1,$F7,$6B,$51,$FE,$1C,$AD,$5F,$F5
	.dc.b	$9A,$FC,$15,$F3,$BF,$CB,$5F,$FA,$D9,$FF,$02,$57,$F4,$F7,$EC,$47
	.dc.b	$F8,$2E,$7E,$43,$FD,$D2,$FC,$CE,$63,$CD,$1F,$CC,$63,$FD,$82,$FF
	.dc.b	$93,$6F,$FC,$08,$FE,$62,$32,$74,$3F,$C7,$AF,$47,$33,$D2,$FE,$5C
	.dc.b	$FC,$09,$FE,$6D,$1F,$2D,$FF,$00,$47,$CB,$3E,$14,$7E,$F8,$FC,$34
	.dc.b	$7F,$92,$67,$EA,$9F,$02,$BF,$02,$F3,$8F,$F1,$4D,$7D,$58,$FE,$ED
	.dc.b	$5F,$CE,$9E,$73,$FE,$4B,$AF,$BA,$FA,$4F,$90,$FE,$02,$BE,$C4,$7D
	.dc.b	$33,$F7,$D3,$F6,$2B,$F0,$F1,$ED,$3F,$B1,$D7,$EE,$6B,$ED,$9D,$63
	.dc.b	$FD,$8B,$1F,$E2,$FB,$FB,$71,$FC,$BE,$FF,$A5,$CE,$5F,$64,$FE,$5E
	.dc.b	$31,$5C,$EB,$F7,$75,$F6,$9F,$9E,$72,$3F,$B2,$1F,$16,$3F,$B4,$B8
	.dc.b	$79,$3F,$E0,$D8,$FB,$0F,$D3,$BE,$4F,$E1,$1E,$47,$E0,$67,$EB,$D7
	.dc.b	$E1,$A3,$E6,$CF,$F5,$C3,$E9,$57,$27,$FA,$B1,$FD,$82,$FF,$4D,$1E
	.dc.b	$D8,$FF,$3F,$3F,$62,$BE,$EB,$FA,$68,$F9,$E7,$33,$F4,$4F,$F8,$BD
	.dc.b	$F6,$9E,$38,$FD,$9C,$66,$BE,$B9,$C6,$7F,$CE,$95,$F6,$EF,$EB,$CE
	.dc.b	$EB,$F6,$07,$F0,$49,$FC,$85,$7E,$B5,$CD,$7F,$03,$7D,$11,$FA,$D9
	.dc.b	$FE,$AE,$FD,$43,$E7,$9F,$EA,$48,$FB,$13,$FA,$18,$FC,$71,$FA,$A9
	.dc.b	$FD,$13,$F4,$CE,$31,$FE,$62,$8F,$D3,$3F,$5C,$CD,$7D,$29,$FF,$45
	.dc.b	$9F,$48,$F0,$CF,$F0,$67,$E9,$9F,$E1,$29,$FA,$E6,$A3,$94,$7F,$05
	.dc.b	$3F,$D0,$D3,$F3,$DF,$D8,$1F,$40,$FD,$ED,$FD,$03,$FA,$D3,$FC,$70
	.dc.b	$F8,$27,$A5,$FD,$33,$F5,$6B,$33,$E9,$BF,$D2,$CF,$FB,$72,$3F,$8F
	.dc.b	$5F,$E9,$0F,$E2,$53,$F4,$27,$F0,$71,$FA,$37,$E6,$1F,$20,$FD,$11
	.dc.b	$F9,$93,$E7,$9F,$CF,$0F,$91,$5F,$E7,$07,$E8,$D7,$F1,$3A,$FE,$39
	.dc.b	$3F,$3E,$3F,$8B,$57,$E6,$8F,$E2,$A7,$E3,$6B,$F3,$CF,$F1,$19,$F9
	.dc.b	$B3,$FC,$4E,$3F,$38,$7F,$99,$E7,$E7,$1F,$E4,$1A,$FF,$04,$3F,$C9
	.dc.b	$E7,$FD,$AC,$FF,$82,$67,$FC,$95,$1F,$B0,$3F,$A7,$C7,$F9,$FB,$0F
	.dc.b	$F3,$FC,$FF,$A1,$2F,$FC,$7C,$FF,$9C,$AF,$FB,$45,$7E,$8A,$3F,$C1
	.dc.b	$B7,$FE,$AB,$9F,$F0,$1C,$78,$5F,$DB,$4F,$FB,$3A,$3F,$C0,$6F,$86
	.dc.b	$BD,$2F,$F8,$F2,$FF,$C7,$F5,$FE,$79,$8F,$EA,$77,$FE,$C1,$C3,$F8
	.dc.b	$46,$1F,$E7,$29,$FF,$01,$5F,$F9,$E6,$FF,$CE,$B8,$7F,$8E,$63,$E4
	.dc.b	$57,$F9,$CA,$BF,$D6,$B1,$FE,$B6,$8F,$F3,$7C,$7F,$7C,$8F,$CB,$47
	.dc.b	$CD,$8F,$69,$FE,$B0,$3F,$C6,$55,$9A,$FF,$33,$47,$F8,$9C,$FF,$20
	.dc.b	$DF,$FB,$FD,$FF,$33,$4F,$F8,$C0,$FB,$18,$7B,$23,$FC,$28,$7F,$01
	.dc.b	$9F,$F3,$43,$FE,$66,$9F,$96,$7F,$A9,$4F,$96,$FD,$1B,$F9,$67,$C7
	.dc.b	$7F,$C5,$91,$FE,$65,$7F,$05,$7F,$AF,$7F,$DE,$07,$9C,$FF,$70,$4F
	.dc.b	$CB,$7E,$56,$1F,$2E,$3F,$CC,$71,$FE,$A5,$7F,$CC,$67,$E2,$4F,$D7
	.dc.b	$BF,$E9,$C7,$FC,$FD,$7F,$E8,$BB,$FF,$14,$C7,$DA,$3F,$CC,$33,$FD
	.dc.b	$91,$FF,$4D,$3F,$2A,$3F,$32,$7C,$D7,$FD,$3D,$1F,$98,$3E,$4B,$F3
	.dc.b	$2B,$F2,$F5,$FE,$94,$8F,$F6,$DC,$FE,$5C,$FC,$F5,$79,$A3,$F9,$13
	.dc.b	$F3,$23,$FD,$27,$7F,$E5,$7B,$FE,$63,$1F,$2E,$BF,$27,$1F,$94,$8F
	.dc.b	$F2,$AC,$7F,$A5,$1F,$F2,$A9,$F2,$6F,$F9,$2C,$7F,$94,$30,$FF,$25
	.dc.b	$9F,$EC,$99,$F9,$73,$A9,$FF,$0C,$C7,$CC,$3E,$3B,$FE,$53,$C3,$FD
	.dc.b	$8B,$1F,$BB,$3F,$03,$1F,$E8,$E7,$FC,$A4,$7E,$5D,$F9,$57,$FE,$8A
	.dc.b	$8F,$F6,$36,$1F,$96,$7E,$51,$FE,$BE,$9F,$E0,$17,$FE,$BD,$7F,$D1
	.dc.b	$13,$FC,$02,$BF,$94,$D7,$E4,$4F,$DF,$DF,$DE,$AF,$7E,$1F,$C0,$67
	.dc.b	$E5,$C7,$FA,$F9,$FC,$A9,$FE,$BA,$7F,$2B,$5F,$2F,$0F,$F5,$C4,$7C
	.dc.b	$B8,$FE,$4B,$3F,$EF,$38,$FD,$EC,$FC,$7A,$F9,$4D,$C3,$FC,$C5,$3F
	.dc.b	$95,$BF,$DF,$CF,$FB,$6E,$3E,$54,$7E,$FE,$3F,$DB,$D3,$FE,$7F,$BF
	.dc.b	$F3,$D5,$7F,$8E,$0F,$E9,$E7,$F5,$63,$FC,$B0,$0F,$E8,$27,$DE,$7F
	.dc.b	$C7,$8F,$F5,$7B,$FF,$04,$61,$FE,$05,$8F,$EA,$D1,$FE,$05,$9F,$EA
	.dc.b	$71,$FE,$01,$7F,$A0,$3F,$26,$FF,$C6,$EF,$F5,$8B,$FF,$00,$DF,$F3
	.dc.b	$08,$FF,$11,$C7,$F2,$87,$C8,$FF,$9C,$5F,$F1,$CC,$7F,$81,$9F,$F0
	.dc.b	$5C,$7F,$5B,$AF,$F1,$BC,$7F,$56,$AF,$F1,$95,$7F,$56,$8F,$F0,$2D
	.dc.b	$FD,$29,$FF,$33,$C7,$F8,$0E,$3F,$CC,$93,$FD,$66,$FF,$AA,$D7,$F8
	.dc.b	$C2,$FF,$BF,$5F,$CB,$BF,$F0,$05,$FF,$8C,$EF,$FC,$CA,$7C,$EA,$F9
	.dc.b	$35,$FE,$2F,$3F,$BD,$CF,$C9,$BF,$F3,$25,$FC,$97,$FC,$BD,$3F,$24
	.dc.b	$F8,$B5,$FD,$F6,$FE,$61,$87,$FB,$E4,$7F,$8B,$0F,$EF,$97,$F1,$6B
	.dc.b	$FD,$2F,$1F,$70,$FF,$5A,$9F,$DF,$2F,$FD,$26,$7D,$6B,$FF,$09,$47
	.dc.b	$FA,$51,$FE,$F9,$3F,$D5,$2F,$FD,$11,$3F,$1E,$3F,$D1,$F7,$F1,$E3
	.dc.b	$FB,$E5,$FE,$59,$FF,$45,$CF,$FA,$32,$3E,$3D,$7F,$89,$EF,$E3,$C7
	.dc.b	$F5,$48,$F8,$D5,$F1,$E3,$F1,$F1,$FE,$23,$8F,$C7,$D7,$FB,$0A,$3F
	.dc.b	$1F,$1F,$90,$9F,$F1,$1D,$FF,$94,$A3,$F1,$D1,$FB,$E9,$F8,$D1,$F1
	.dc.b	$E3,$FC,$A7,$3F,$1E,$3F,$D0,$51,$FC,$9A,$7F,$C9,$F1,$F8,$B3,$FD
	.dc.b	$6F,$3F,$8B,$AF,$F3,$F5,$FF,$93,$AB,$FC,$41,$5F,$14,$FF,$3C,$BF
	.dc.b	$EB,$33,$F8,$EB,$FA,$F3,$FD,$6B,$7F,$1D,$FF,$76,$D7,$C7,$3F,$DD
	.dc.b	$F7,$F8,$D8,$F9,$13,$F1,$E7,$FC,$97,$1F,$1D,$FF,$75,$57,$C7,$7F
	.dc.b	$1A,$7C,$73,$F1,$B1,$F1,$DF,$F5,$65,$7C,$77,$FD,$57,$5F,$1D,$FC
	.dc.b	$4C,$FF,$92,$23,$E3,$D7,$EC,$23,$F6,$31,$F1,$E7,$FD,$41,$1F,$E7
	.dc.b	$58,$FF,$23,$4F,$F9,$D2,$3F,$63,$3F,$8B,$3F,$62,$FE,$32,$7F,$63
	.dc.b	$3F,$E4,$58,$FC,$5C,$FE,$C4,$FF,$35,$3F,$CD,$0F,$E0,$D3,$F2,$EB
	.dc.b	$FB,$6D,$7F,$44,$9F,$ED,$B8,$7E,$B5,$E1,$5F,$E2,$03,$FA,$CB,$FA
	.dc.b	$03,$FD,$02,$FF,$23,$3F,$C3,$47,$F8,$2E,$BF,$8B,$46,$5F,$EE,$11
	.dc.b	$FE,$45,$9F,$F0,$E5,$7F,$46,$BF,$E7,$95,$FE,$04,$BF,$F4,$D4,$7F
	.dc.b	$3D,$7F,$C9,$B8,$7F,$8C,$5F,$0B,$FC,$CE,$7F,$C8,$CF,$F3,$59,$E1
	.dc.b	$5F,$DD,$0F,$E9,$55,$FE,$1C,$8F,$E9,$18,$7F,$89,$0F,$F2,$6D,$7F
	.dc.b	$8C,$A7,$F9,$E5,$7F,$87,$4F,$F2,$E9,$FD,$7D,$F8,$2F,$FA,$8D,$FF
	.dc.b	$6A,$CA,$3F,$33,$08,$C4,$7F,$8B,$E3,$FA,$53,$FD,$DD,$57,$F5,$D7
	.dc.b	$DC,$74,$3D,$D1,$F5,$E3,$51,$CE,$7F,$CF,$4F,$FA,$46,$39,$47,$D8
	.dc.b	$3F,$D0,$51,$FD,$9A,$33,$3F,$E2,$53,$FB,$E4,$7C,$93,$FB,$DD,$FB
	.dc.b	$DF,$F3,$F3,$FE,$01,$BF,$F1,$35,$7F,$28,$AF,$F1,$33,$F5,$E7,$FC
	.dc.b	$A0,$7D,$4A,$FE,$F9,$1F,$9B,$7F,$C0,$F5,$E7,$6D,$FF,$89,$A3,$F1
	.dc.b	$D1,$F3,$1F,$43,$AB,$FF,$13,$1F,$E7,$BA,$FF,$27,$61,$FD,$F6,$7F
	.dc.b	$C2,$73,$F4,$AF,$E8,$BF,$E2,$63,$FC,$9F,$1F,$40,$FF,$3D,$CF,$F8
	.dc.b	$9E,$BE,$6D,$FF,$89,$E3,$F4,$93,$FA,$03,$D0,$5B,$F9,$D8,$7F,$9F
	.dc.b	$2B,$FC,$51,$7F,$E3,$DA,$4F,$E9,$1F,$F7,$91,$F3,$CF,$D5,$BF,$3D
	.dc.b	$FE,$41,$5F,$E0,$37,$E9,$E1,$FE,$A2,$9F,$D3,$47,$FB,$46,$3E,$8B
	.dc.b	$F3,$0F,$94,$7F,$A0,$0F,$CE,$9F,$E7,$FB,$FC,$A4,$7F,$95,$E3,$E8
	.dc.b	$CB,$0F,$F4,$06,$1F,$A6,$BF,$F5,$53,$FC,$36,$BE,$6D,$FE,$90,$FF
	.dc.b	$2E,$9F,$A3,$7D,$2F,$D1,$3C,$EF,$EB,$23,$F3,$2F,$F9,$59,$FC,$DD
	.dc.b	$FF,$9F,$DF,$CE,$CF,$F9,$59,$FF,$55,$4F,$F9,$58,$FE,$2B,$3F,$3D
	.dc.b	$FF,$3F,$DF,$F3,$37,$E3,$9F,$34,$FC,$CB,$FC,$42,$7F,$87,$D7,$CB
	.dc.b	$3E,$24,$FF,$B4,$CF,$E1,$E7,$CA,$AF,$CD,$4F,$F0,$CA,$FF,$12,$4F
	.dc.b	$E6,$0F,$F2,$A9,$F2,$AB,$FC,$FA,$FE,$5E,$3F,$AC,$3F,$C1,$A3,$F9
	.dc.b	$19,$F7,$DF,$F5,$43,$FE,$7D,$3F,$D4,$C7,$F0,$77,$FB,$14,$FF,$5F
	.dc.b	$7F,$18,$7F,$95,$6F,$FC,$0B,$1F,$D0,$23,$F8,$2B,$AC,$3F,$C9,$B1
	.dc.b	$FA,$C3,$F9,$49,$FE,$06,$3F,$2C,$FF,$16,$8F,$EC,$27,$F8,$B9,$FF
	.dc.b	$03,$D7,$F9,$4E,$7F,$C0,$F3,$F9,$DC,$3F,$39,$0B,$FF,$2A,$34,$FF
	.dc.b	$14,$1F,$A2,$AF,$EC,$95,$FE,$30,$9F,$F2,$A4,$7F,$82,$A3,$FB,$2D
	.dc.b	$FF,$64,$7F,$C1,$35,$FD,$90,$FF,$54,$57,$E6,$AB,$FC,$5F,$1F,$D9
	.dc.b	$67,$FC,$5F,$1F,$D9,$5F,$D9,$3F,$CD,$67,$F9,$A4,$7E,$6E,$7F,$C1
	.dc.b	$67,$CC,$3F,$D0,$37,$FE,$0B,$8C,$4F,$F9,$5E,$BF,$C6,$31,$F3,$6B
	.dc.b	$FC,$AF,$1F,$E0,$B7,$E5,$CF,$67,$E6,$C7,$F4,$59,$F9,$B1,$FE,$C1
	.dc.b	$AF,$E9,$0F,$F9,$69,$FF,$07,$39,$9F,$F2,$A3,$FD,$09,$FF,$2D,$CF
	.dc.b	$F8,$42,$FF,$C1,$EF,$CB,$9F,$F0,$7B,$FC,$D6,$7E,$64,$7C,$89,$FF
	.dc.b	$2D,$1F,$4E,$7E,$69,$FE,$36,$7F,$3B,$1F,$D8,$8F,$F1,$B1,$FD,$92
	.dc.b	$BF,$CB,$53,$F3,$2B,$CB,$3F,$CB,$9F,$F4,$0B,$FE,$82,$AF,$F2,$DC
	.dc.b	$FC,$A8,$FF,$1A,$BF,$97,$7F,$C2,$C7,$F8,$D4,$FC,$9D,$7F,$A0,$5F
	.dc.b	$97,$3F,$CC,$9F,$F2,$21,$FD,$90,$FF,$41,$3F,$96,$9F,$F5,$49,$FE
	.dc.b	$36,$3F,$2D,$5F,$9D,$9F,$C4,$9F,$AC,$29,$FE,$84,$7F,$2A,$FF,$97
	.dc.b	$A3,$E5,$1F,$E3,$87,$FC,$BC,$7F,$BF,$AF,$F5,$8F,$F9,$5D,$FE,$AE
	.dc.b	$E6,$BE,$2C,$7E,$5E,$7F,$C7,$67,$FA,$B2,$3F,$C7,$6F,$CB,$8F,$F5
	.dc.b	$5C,$FF,$99,$2B,$F1,$0F,$FA,$18,$FC,$B5,$7C,$C7,$F8,$1C,$FF,$02
	.dc.b	$7F,$82,$CF,$FA,$B2,$BF,$2D,$5F,$E8,$67,$F8,$25,$7F,$25,$AF,$F1
	.dc.b	$5C,$FC,$D7,$E6,$DF,$E5,$1F,$F3,$11,$F9,$37,$F2,$AF,$D2,$3F,$C2
	.dc.b	$B3,$FE,$AF,$BF,$F4,$4D,$FE,$54,$FF,$56,$D7,$FA,$26,$BE,$41,$F1
	.dc.b	$0F,$F5,$71,$FE,$63,$8F,$F6,$9C,$FF,$98,$DF,$CA,$1F,$E0,$98,$FC
	.dc.b	$A1,$FB,$F3,$FD,$13,$3F,$94,$7F,$D1,$75,$FB,$17,$FC,$77,$5F,$29
	.dc.b	$FC,$C5,$7F,$52,$7F,$C0,$47,$F7,$D3,$FC,$BE,$7F,$8A,$0F,$E3,$4F
	.dc.b	$F5,$4C,$3E,$34,$7F,$98,$4F,$F0,$19,$FE,$37,$C3,$FC,$07,$5D,$23
	.dc.b	$17,$FE,$04,$7F,$79,$1F,$E3,$07,$47,$F8,$03,$0F,$E8,$8F,$F9,$25
	.dc.b	$FE,$D8,$FF,$9A,$E3,$FA,$C1,$FD,$DF,$0F,$CA,$C7,$F8,$04,$FB,$87
	.dc.b	$F9,$D6,$BF,$C9,$6F,$F8,$31,$FF,$25,$C7,$F9,$B5,$DC,$7F,$95,$DF
	.dc.b	$F2,$AE,$1F,$C0,$A3,$FC,$65,$87,$F9,$C8,$FF,$1B,$1F,$E6,$99,$FF
	.dc.b	$02,$C7,$E5,$EF,$FC,$68,$FF,$81,$B0,$F3,$36,$3D,$EF,$F6,$09,$F0
	.dc.b	$D7,$F8,$DA,$FF,$D7,$13,$FC,$10,$FF,$03,$1F,$C1,$AF,$FC,$0C,$FF
	.dc.b	$08,$BF,$F0,$33,$FE,$98,$39,$9F,$26,$7B,$BF,$E7,$33,$F6,$E5,$AF
	.dc.b	$F4,$CC,$FF,$98,$CD,$BF,$D5,$E3,$CB,$5F,$52,$3F,$C6,$91,$FE,$38
	.dc.b	$9F,$A7,$5F,$EB,$D8,$FA,$47,$2B,$FA,$2F,$D3,$3F,$C6,$8F,$C0,$3E
	.dc.b	$7E,$1F,$9D,$8F,$9E,$7F,$83,$0F,$DA,$E1,$F9,$97,$E9,$CF,$D0,$7F
	.dc.b	$3B,$7F,$41,$FF,$4A,$BF,$E3,$43,$FC,$19,$1F,$39,$FF,$19,$C7,$EE
	.dc.b	$5F,$F4,$95,$7F,$8C,$AF,$FC,$C5,$3F,$CD,$2B,$FC,$C5,$3F,$EE,$58
	.dc.b	$F9,$C7,$F8,$CC,$FF,$31,$5F,$F8,$CE,$3F,$D1,$B5,$FE,$33,$9F,$8C
	.dc.b	$7C,$AB,$FF,$31,$C7,$F8,$CE,$3F,$CB,$B7,$FE,$8C,$7F,$9A,$CF,$F9
	.dc.b	$71,$FF,$0B,$DF,$FA,$32,$3E,$64,$7F,$97,$1F,$F5,$F5,$FF,$AF,$23
	.dc.b	$E4,$1F,$98,$9F,$90,$FE,$87,$0F,$F5,$F9,$FE,$B7,$C3,$FC,$6B,$1F
	.dc.b	$E5,$8B,$FF,$2E,$57,$E6,$23,$E5,$BC,$9F,$CC,$1F,$86,$8F,$96,$FF
	.dc.b	$54,$7E,$5B,$FE,$8C,$BF,$F2,$EC,$7F,$AE,$63,$F3,$15,$FC,$29,$FC
	.dc.b	$91,$F9,$8B,$FE,$13,$5F,$EB,$AC,$3F,$84,$4F,$F2,$F3,$FC,$55,$7F
	.dc.b	$C2,$67,$FC,$B7,$87,$9E,$7F,$D6,$F3,$FE,$ED,$7F,$D1,$2E,$5F,$ED
	.dc.b	$35,$FE,$D2,$8F,$97,$1F,$93,$3F,$CC,$0F,$CE,$8C,$3F,$3A,$3F,$CC
	.dc.b	$07,$FB,$AC,$FF,$30,$57,$FA,$1D,$FD,$05,$FC,$D8,$FF,$17,$47,$F9
	.dc.b	$86,$7E,$6E,$1F,$EB,$78,$FC,$BD,$FE,$56,$7F,$CC,$11,$FC,$50,$FF
	.dc.b	$76,$D7,$F8,$D7,$0F,$F1,$A4,$7F,$60,$C3,$F8,$3B,$FE,$60,$3F,$CB
	.dc.b	$67,$F8,$D2,$7F,$AF,$BF,$E5,$F3,$E5,$1F,$DC,$CF,$F1,$71,$FE,$0A
	.dc.b	$7F,$C0,$F3,$FD,$7A,$3F,$AF,$3F,$E4,$99,$FC,$E4,$7F,$61,$BF,$F2
	.dc.b	$F9,$FE,$B8,$AF,$EB,$F7,$FE,$0A,$9F,$EB,$F3,$FE,$5E,$9F,$F0,$51
	.dc.b	$FE,$5E,$3F,$AF,$9F,$E3,$28,$FF,$1B,$47,$F5,$97,$B5,$7F,$86,$5F
	.dc.b	$EC,$33,$FE,$09,$3F,$CB,$E7,$F8,$25,$FE,$3E,$2B,$D1,$1F,$4A,$3F
	.dc.b	$DA,$8F,$FA,$2D,$FF,$12,$BF,$E3,$53,$FC,$BF,$3F,$32,$3F,$C6,$B1
	.dc.b	$FE,$0A,$3F,$D1,$97,$FE,$35,$8F,$97,$3F,$E4,$27,$E6,$1F,$97,$8F
	.dc.b	$F0,$54,$FF,$A2,$CF,$9B,$5F,$E8,$B7,$FC,$6D,$3F,$D9,$23,$FC,$12
	.dc.b	$FF,$8D,$2B,$FC,$BA,$FF,$8E,$2F,$F0,$81,$FF,$05,$4F,$ED,$5C,$47
	.dc.b	$F7,$47,$F3,$F3,$FE,$60,$BD,$47,$FA,$5E,$3F,$BE,$BF,$41,$FE,$C0
	.dc.b	$FE,$EE,$BF,$C8,$CF,$F9,$8E,$3F,$C6,$EF,$F8,$38,$F9,$F1,$F3,$9F
	.dc.b	$EC,$57,$FE,$8B,$7F,$34,$7E,$14,$FF,$5C,$D7,$E6,$9F,$C9,$9F,$24
	.dc.b	$FB,$C7,$D0,$3F,$37,$5F,$E3,$97,$FD,$1A,$7E,$6C,$FF,$79,$CF,$FA
	.dc.b	$31,$FA,$15,$F1,$67,$FC,$C9,$1F,$E8,$BB,$FD,$65,$F9,$E3,$75,$FE
	.dc.b	$8C,$7E,$85,$FE,$68,$FF,$46,$1F,$EB,$7B,$FE,$BD,$1F,$E5,$D3,$FC
	.dc.b	$70,$7F,$A3,$6B,$E6,$DF,$F9,$8A,$3F,$32,$FE,$6D,$FA,$77,$F4,$8F
	.dc.b	$C2,$BF,$B4,$7F,$D4,$75,$F3,$A3,$F9,$04,$7F,$AE,$5F,$9E,$FF,$A4
	.dc.b	$0F,$F6,$8B,$FE,$D0,$3E,$71,$F9,$C3,$FC,$CA,$7F,$A4,$0F,$F3,$2B
	.dc.b	$FE,$BA,$7E,$6D,$FE,$72,$3F,$D7,$51,$F9,$CA,$F4,$3F,$C6,$67,$E7
	.dc.b	$3F,$C1,$27,$E7,$47,$E7,$0F,$F4,$8C,$7C,$FA,$FE,$0E,$FF,$30,$3F
	.dc.b	$16,$7F,$09,$3F,$DC,$B7,$FE,$65,$71,$7F,$C1,$DF,$F5,$C1,$F3,$8F
	.dc.b	$F4,$64,$7C,$E9,$FC,$CC,$7F,$A3,$E7,$CE,$FE,$8A,$3F,$16,$FE,$6C
	.dc.b	$FF,$5D,$9F,$E0,$1B,$FF,$61,$3F,$9B,$BF,$F4,$94,$FF,$64,$8D,$6F
	.dc.b	$FE,$67,$3F,$C0,$C7,$F5,$C3,$F8,$0B,$FB,$00,$FF,$99,$CF,$CE,$BF
	.dc.b	$EF,$77,$71,$FB,$98,$D9,$FD,$75,$FC,$C6,$0C,$73,$3F,$E0,$EC,$2B
	.dc.b	$FE,$6C,$29,$CE,$BF,$41,$1C,$6F,$FD,$82,$FF,$23,$7F,$8B,$DF,$FA
	.dc.b	$75,$FE,$D3,$7C,$AF,$FC,$DD,$3B,$7F,$C2,$A6,$4F,$F3,$A5,$7F,$82
	.dc.b	$6B,$FD,$DC,$7E,$1E,$3D,$8D,$9F,$14,$7B,$71,$FE,$31,$1D,$63,$FD
	.dc.b	$55,$82,$7E,$55,$73,$BE,$35,$FE,$CB,$78,$3F,$E4,$63,$C9,$5F,$C6
	.dc.b	$EF,$71,$FE,$21,$9C,$DF,$F9,$F6,$2D,$7A,$5F,$F3,$C5,$FF,$3A,$AF
	.dc.b	$89,$1F,$E7,$9C,$3A,$C7,$F4,$BC,$3C,$33,$FE,$27,$C3,$F9,$95,$F1
	.dc.b	$AE,$93,$FC,$71,$DB,$FE,$A7,$3E,$29,$FE,$7B,$BF,$4B,$F1,$22,$BF
	.dc.b	$0C,$AF,$B1,$FF,$2E,$CF,$E2,$9F,$09,$F8,$93,$FD,$00,$FE,$3A,$FC
	.dc.b	$31,$FE,$7F,$BD,$BF,$40,$61,$E8,$7F,$79,$7C,$23,$FD,$05,$1F,$3A
	.dc.b	$BE,$16,$16,$91,$B3,$F8,$B6,$1F,$8E,$8E,$95,$FC,$05,$E9,$87,$E8
	.dc.b	$63,$EA,$9F,$E6,$49,$FC,$44,$E2,$7C,$F3,$BC,$3E,$24,$FC,$38,$FD
	.dc.b	$54,$FE,$B5,$FF,$1A,$D7,$F1,$9A,$E9,$1F,$91,$7C,$F5,$F8,$2A,$EB
	.dc.b	$3F,$E6,$58,$E8,$FF,$91,$5F,$84,$7E,$11,$FF,$31,$E1,$F8,$48,$E7
	.dc.b	$7F,$81,$AE,$75,$FB,$29,$F8,$2F,$F0,$A3,$FD,$78,$7D,$EC,$3B,$E1
	.dc.b	$FC,$0D,$FA,$11,$FA,$23,$FC,$CF,$7F,$C2,$A7,$FC,$39,$5F,$88,$7F
	.dc.b	$17,$3F,$E9,$2C,$31,$7C,$A7,$F1,$57,$FE,$94,$3F,$81,$DF,$D9,$BF
	.dc.b	$39,$FB,$CA,$F1,$9F,$9F,$8F,$BD,$86,$AF,$FC,$E3,$87,$F0,$3B,$FA
	.dc.b	$75,$F6,$9D,$57,$EF,$A7,$EE,$DF,$E0,$27,$C6,$7F,$03,$3E,$29,$FC
	.dc.b	$DE,$BF,$0B,$7E,$C8,$FF,$4D,$57,$EE,$2B,$37,$DA,$B8,$D7,$C4,$BF
	.dc.b	$AE,$71,$7F,$D3,$95,$F3,$A3,$D9,$87,$D7,$9F,$B5,$1F,$80,$7F,$9A
	.dc.b	$3F,$49,$E4,$7E,$E6,$7E,$BB,$FB,$79,$FE,$82,$FB,$0F,$D6,$1E,$43
	.dc.b	$F2,$87,$EB,$EB,$9B,$FE,$F1,$9E,$71,$F8,$A8,$FD,$D3,$CD,$FD,$01
	.dc.b	$FB,$6A,$F3,$4F,$BE,$BF,$5D,$3F,$56,$3F,$6B,$1F,$EA,$49,$FB,$33
	.dc.b	$FB,$77,$DD,$7F,$EA,$19,$FE,$71,$3F,$60,$FD,$39,$FE,$D0,$9F,$D9
	.dc.b	$3F,$B6,$78,$DF,$E4,$A7,$FD,$A1,$3F,$57,$0F,$AD,$5F,$AD,$8F,$E5
	.dc.b	$EF,$CD,$3D,$0F,$F5,$F7,$F5,$8F,$FA,$97,$0F,$AD,$1F,$0E,$3F,$D4
	.dc.b	$A7,$D6,$9F,$C8,$3E,$93,$F5,$75,$FD,$D1,$FE,$44,$71,$3E,$49,$F2
	.dc.b	$CF,$EC,$C7,$F9,$A6,$7F,$00,$FD,$63,$E7,$DF,$D6,$3F,$B9,$D7,$EB
	.dc.b	$27,$F6,$18,$7E,$8E,$BE,$A4,$FE,$8F,$0F,$6C,$7F,$C0,$03,$F9,$33
	.dc.b	$FA,$CA,$F2,$47,$F7,$78,$FE,$4E,$FF,$B5,$A7,$F5,$95,$ED,$8F,$EC
	.dc.b	$AF,$F2,$7C,$3F,$B6,$4F,$EB,$2F,$EA,$C7,$F6,$78,$E3,$87,$D5,$BF
	.dc.b	$E4,$67,$F2,$1C,$3F,$87,$E1,$FD,$EA,$FF,$DB,$07,$F2,$7C,$3E,$B1
	.dc.b	$C6,$BF,$BA,$BF,$59,$FD,$5E,$1F,$C8,$6F,$F5,$F7,$FC,$16,$BF,$8A
	.dc.b	$CF,$D6,$AE,$47,$D2,$3E,$41,$F8,$E3,$EA,$D7,$EB,$EB,$35,$F5,$30
	.dc.b	$E3,$5F,$61,$FA,$8F,$F5,$E3,$F4,$F5,$FC,$7E,$BE,$A4,$FF,$04,$8F
	.dc.b	$A7,$3F,$D2,$5F,$53,$FE,$1E,$BF,$AB,$7F,$A6,$C3,$FA,$C3,$FA,$69
	.dc.b	$F0,$9F,$E8,$BC,$3E,$93,$FC,$66,$FE,$5B,$F7,$0D,$1F,$8F,$2C,$FE
	.dc.b	$A0,$FE,$3A,$FF,$19,$8F,$A7,$87,$18,$FE,$E7,$5F,$E1,$7B,$FA,$98
	.dc.b	$7F,$2F,$C3,$D7,$5F,$E3,$F7,$D6,$FF,$05,$7E,$63,$FC,$81,$FE,$AD
	.dc.b	$5F,$55,$FE,$29,$1E,$C3,$FC,$02,$7B,$0F,$E3,$38,$7F,$1E,$7C,$98
	.dc.b	$7F,$6E,$7F,$C2,$77,$FB,$28,$E4,$7F,$87,$70,$FF,$08,$47,$E9,$DF
	.dc.b	$F2,$FE,$1F,$53,$0F,$E3,$27,$0C,$3D,$37,$FD,$D1,$FD,$31,$FE,$0B
	.dc.b	$8F,$95,$87,$F9,$41,$FA,$87,$F4,$B7,$F8,$C4,$7D,$23,$FA,$75,$7F
	.dc.b	$1B,$9F,$A0,$7F,$90,$27,$E8,$61,$FC,$5B,$0F,$D0,$C7,$F8,$16,$7F
	.dc.b	$42,$7F,$4A,$7E,$45,$FF,$14,$3F,$3E,$FF,$13,$9F,$F3,$EB,$F0,$8F
	.dc.b	$94,$7E,$72,$BF,$BF,$57,$F8,$F2,$FF,$3B,$1F,$26,$3F,$BB,$4F,$E6
	.dc.b	$2F,$FA,$19,$F9,$B7,$FB,$F4,$7C,$83,$E2,$3F,$CF,$DF,$BE,$FF,$9E
	.dc.b	$1F,$CD,$3F,$C2,$EF,$FC,$EA,$FF,$6F,$3E,$3D,$7F,$09,$8F,$F3,$C5
	.dc.b	$FC,$78,$FE,$13,$5F,$96,$AF,$E7,$95,$FE,$2B,$BF,$EE,$31,$FC,$F2
	.dc.b	$BF,$29,$5F,$82,$3E,$D9,$FD,$4A,$3F,$9F,$3F,$8A,$3F,$8F,$BF,$C3
	.dc.b	$1F,$C9,$BF,$E1,$23,$FC,$7D,$5F,$32,$BF,$CB,$31,$F9,$18,$F9,$F7
	.dc.b	$FE,$58,$7E,$8D,$6A,$7F,$36,$7E,$54,$D1,$FE,$4F,$3F,$81,$3F,$99
	.dc.b	$9F,$04,$7E,$C3,$0F,$A9,$3F,$CE,$CF,$EE,$F1,$FE,$48,$39,$9C,$CF
	.dc.b	$CC,$3B,$9F,$55,$A3,$FD,$39,$5C,$DF,$5C,$FE,$1A,$BF,$DC,$87,$43
	.dc.b	$FD,$C9,$5A,$7F,$42,$7F,$A0,$E3,$4F,$F9,$AE,$7F,$9F,$BF,$26,$3F
	.dc.b	$77,$5F,$D0,$AB,$BB,$CA,$3D,$D7,$E9,$AF,$F1,$85,$FD,$A7,$E9,$3E
	.dc.b	$97,$B9,$FE,$24,$7F,$8B,$3F,$E5,$29,$DC,$7B,$A7,$FD,$40,$FF,$0D
	.dc.b	$9F,$A1,$1F,$A6,$79,$9F,$E8,$1B,$FB,$0F,$B8,$FF,$7C,$47,$DC,$9F
	.dc.b	$B5,$1F,$6E,$FF,$88,$57,$DF,$78,$1E,$18,$F6,$1F,$DE,$CE,$D3,$F6
	.dc.b	$A7,$D9,$3F,$75,$FC,$69,$F9,$49,$FC,$33,$F6,$6F,$CF,$1F,$E8,$8A
	.dc.b	$FA,$8F,$CF,$06,$A3,$FC,$88,$FC,$D7,$FC,$F4,$FE,$1D,$F6,$B9,$3E
	.dc.b	$33,$F9,$8B,$F4,$BF,$69,$B7,$F4,$23,$FA,$24,$F5,$8F,$80,$FF,$B9
	.dc.b	$6B,$ED,$3F,$6E,$3F,$6B,$7F,$AA,$9F,$74,$FF,$AC,$5F,$74,$FE,$E9
	.dc.b	$FA,$71,$FA,$43,$DB,$1F,$DE,$E7,$EB,$D6,$4F,$8B,$3F,$E1,$73,$FC
	.dc.b	$63,$7F,$E9,$77,$35,$F8,$87,$D2,$7F,$AD,$63,$FD,$39,$7F,$48,$FA
	.dc.b	$C7,$81,$FB,$77,$FE,$78,$9F,$CC,$3E,$A7,$FD,$51,$3F,$09,$FF,$6C
	.dc.b	$BF,$E2,$A7,$F7,$37,$F5,$9F,$AC,$7C,$B3,$F4,$87,$13,$F5,$95,$F4
	.dc.b	$1F,$D6,$5F,$8D,$FA,$17,$FA,$CA,$C9,$F8,$19,$FD,$05,$FF,$74,$AF
	.dc.b	$F5,$B9,$F8,$A7,$D4,$FF,$09,$7D,$4F,$F9,$4D,$F1,$3F,$AF,$7E,$9B
	.dc.b	$FE,$2B,$3E,$79,$F2,$CF,$D4,$1C,$0F,$B8,$FE,$A2,$FE,$94,$FE,$A2
	.dc.b	$3F,$90,$4F,$F2,$18,$FE,$3F,$3F,$3A,$7E,$8C,$FE,$82,$FC,$EF,$E9
	.dc.b	$23,$F8,$9C,$7E,$22,$7F,$D9,$67,$E6,$27,$E7,$9F,$1C,$FD,$01,$F3
	.dc.b	$EF,$F5,$07,$D2,$AF,$E2,$B1,$F3,$8F,$95,$5F,$35,$FF,$2D,$CF,$CD
	.dc.b	$9F,$CD,$1F,$CD,$0F,$DD,$D7,$F1,$0A,$CC,$FF,$14,$3E,$84,$FF,$13
	.dc.b	$AF,$E6,$07,$F3,$9B,$FC,$E4,$7F,$95,$E7,$F3,$91,$FE,$59,$8F,$E1
	.dc.b	$D3,$FD,$1A,$3D,$8F,$1A,$FE,$25,$59,$3F,$3F,$5F,$9B,$9F,$E8,$6F
	.dc.b	$07,$F8,$9C,$FE,$7E,$72,$7C,$F9,$FE,$07,$3F,$28,$FC,$BC,$FF,$43
	.dc.b	$9F,$93,$5F,$9D,$AF,$93,$5F,$D0,$A3,$FA,$04,$7C,$88,$F9,$31,$FD
	.dc.b	$F2,$3F,$82,$DF,$F3,$69,$F8,$B7,$F2,$1F,$EF,$77,$F2,$2B,$F2,$35
	.dc.b	$FC,$E6,$FE,$3B,$FD,$CA,$BE,$44,$7C,$79,$FC,$84,$7F,$32,$8F,$80
	.dc.b	$FF,$70,$9F,$E6,$AF,$F0,$23,$E9,$57,$C8,$AF,$EE,$11,$F2,$0F,$89
	.dc.b	$7F,$1D,$FE,$DD,$7F,$DA,$E7,$F7,$D1,$FC,$CC,$F8,$EF,$F0,$1A,$F8
	.dc.b	$F7,$FE,$22,$9F,$E6,$55,$FC,$10,$F2,$BF,$CE,$A7,$FD,$48,$FF,$2D
	.dc.b	$7F,$7B,$1F,$26,$7E,$44,$FF,$2D,$7F,$21,$5F,$22,$3E,$4C,$FF,$2B
	.dc.b	$8F,$DC,$47,$EE,$63,$E4,$BF,$0A,$BF,$1F,$5F,$D7,$CF,$E5,$11,$FD
	.dc.b	$7D,$F8,$F3,$FC,$A4,$FE,$BF,$3F,$1E,$3F,$71,$3F,$1E,$7F,$AE,$57
	.dc.b	$F8,$4E,$3E,$3C,$FE,$D9,$FE,$A0,$7C,$79,$FE,$4D,$7F,$8D,$8F,$91
	.dc.b	$1F,$8D,$AF,$F0,$6C,$7E,$36,$BF,$A9,$D7,$ED,$5F,$EA,$B7,$FB,$5A
	.dc.b	$FC,$6D,$7F,$83,$5F,$DA,$B5,$FF,$04,$9F,$C8,$1F,$F0,$55,$FC,$7A
	.dc.b	$FF,$03,$47,$C7,$7F,$C8,$53,$FB,$58,$FC,$7C,$7C,$77,$F1,$C7,$ED
	.dc.b	$4F,$E9,$F7,$FB,$47,$F8,$F4,$7F,$4E,$9F,$C6,$C7,$E3,$AB,$E3,$CF
	.dc.b	$F2,$1B,$FD,$9C,$7E,$36,$3F,$8E,$4F,$C7,$7F,$8F,$5F,$C7,$7F,$90
	.dc.b	$57,$F1,$BB,$F8,$F1,$FD,$26,$FF,$66,$FC,$6B,$FD,$9B,$FD,$F2,$7F
	.dc.b	$8D,$C7,$F8,$A8,$FD,$9B,$F8,$B8,$FF,$15,$D7,$EC,$E3,$F5,$D1,$F9
	.dc.b	$78,$FF,$76,$4A,$30,$FF,$5F,$AF,$EC,$16,$C3,$52,$B8,$EC,$FF,$0C
	.dc.b	$B8,$7F,$C0,$27,$E9,$4E,$2E,$5F,$F3,$7E,$1D,$70,$F0,$C7,$F6,$F3
	.dc.b	$FD,$A8,$E1,$FF,$2A,$9F,$E2,$87,$18,$7F,$83,$B0,$E5,$3E,$5B,$E5
	.dc.b	$39,$BF,$0C,$62,$7F,$C4,$6F,$8E,$7F,$C2,$F3,$CA,$3C,$D3,$FE,$39
	.dc.b	$8E,$2F,$C5,$AF,$F2,$F3,$83,$B3,$C6,$7F,$D0,$EF,$F9,$EE,$FC,$EF
	.dc.b	$F9,$16,$78,$1F,$ED,$5C,$2B,$F3,$2F,$83,$EB,$3E,$A5,$FC,$5B,$F8
	.dc.b	$2F,$F9,$2A,$FC,$13,$A0,$F2,$8F,$71,$F4,$E7,$8C,$FF,$A3,$EF,$C3
	.dc.b	$3F,$42,$3C,$07,$F8,$E9,$FF,$27,$38,$3F,$D4,$15,$E0,$75,$87,$83
	.dc.b	$0F,$6B,$FE,$9A,$AF,$B1,$3C,$63,$FD,$3B,$3C,$63,$FC,$34,$F9,$E7
	.dc.b	$EC,$BF,$3B,$0F,$BE,$FF,$B1,$63,$EC,$3F,$E3,$D6,$BF,$94,$8D,$BE
	.dc.b	$D7,$F1,$53,$FE,$93,$7F,$12,$FC,$B3,$C8,$7C,$BC,$3D,$2F,$E7,$2B
	.dc.b	$F1,$0F,$53,$67,$53,$E2,$1F,$8A,$AE,$37,$FE,$49,$BE,$EF,$D4,$BC
	.dc.b	$E1,$EF,$BF,$D5,$4F,$5C,$3F,$29,$08,$ED,$1F,$94,$9F,$F6,$36,$1C
	.dc.b	$6B,$F3,$C7,$CE,$9F,$CE,$4F,$C7,$7F,$27,$3F,$52,$3F,$25,$87,$E5
	.dc.b	$23,$AC,$FF,$B7,$CF,$F2,$64,$7E,$1A,$BA,$47,$EB,$0F,$8D,$1F,$8A
	.dc.b	$AF,$14,$F9,$E7,$F1,$CF,$DF,$7F,$D3,$F5,$F7,$E3,$F0,$2F,$E1,$A3
	.dc.b	$F1,$A7,$E5,$AF,$E3,$5F,$E7,$9F,$C6,$36,$3C,$75,$FA,$49,$F4,$BF
	.dc.b	$E5,$AA,$F1,$C7,$F9,$AA,$3F,$27,$1F,$0D,$FE,$3C,$FE,$26,$7F,$1E
	.dc.b	$FD,$16,$BE,$78,$F8,$67,$E0,$0F,$09,$FC,$78,$F3,$9F,$84,$BF,$0B
	.dc.b	$F0,$A7,$ED,$BE,$63,$BD,$FE,$11,$E6,$7F,$BC,$A7,$F4,$97,$FC,$66
	.dc.b	$7F,$13,$5A,$8F,$E2,$F8,$7F,$08,$AF,$C1,$E1,$F7,$4F,$C8,$61,$FA
	.dc.b	$F8,$F8,$0F,$F1,$68,$D5,$F3,$9F,$C5,$D7,$F0,$2A,$FC,$6E,$1F,$81
	.dc.b	$7E,$B1,$EE,$AF,$E1,$6F,$D4,$3E,$F1,$F2,$4F,$E2,$57,$F7,$E3,$F8
	.dc.b	$61,$F7,$63,$D4,$FE,$76,$FD,$EF,$F8,$A8,$FC,$11,$AC,$3E,$44,$FF
	.dc.b	$06,$AF,$B6,$7B,$A7,$F8,$0D,$7C,$88,$FB,$93,$F7,$9F,$23,$FB,$57
	.dc.b	$EE,$3E,$E3,$EE,$1F,$C0,$0F,$74,$FF,$00,$9F,$B7,$1C,$8F,$EB,$38
	.dc.b	$7F,$38,$7D,$B3,$FD,$36,$FC,$8D,$AF,$B7,$3E,$D8,$D1,$F6,$DF,$DF
	.dc.b	$5F,$E9,$67,$ED,$E1,$FB,$59,$FA,$D3,$E4,$7F,$A9,$3F,$46,$BF,$7E
	.dc.b	$F9,$6F,$F9,$93,$F4,$67,$DD,$3F,$61,$FA,$27,$33,$F9,$F1,$CD,$FD
	.dc.b	$F3,$CD,$FD,$CE,$1F,$D1,$AB,$9C,$7F,$03,$7D,$F3,$FB,$3C,$3E,$DC
	.dc.b	$73,$8F,$DB,$4F,$DB,$BF,$E0,$47,$BA,$3F,$A2,$3F,$CF,$DA,$F9,$5F
	.dc.b	$8A,$F3,$7F,$97,$47,$E8,$E3,$F3,$38,$7F,$3E,$9F,$DB,$CF,$EC,$1F
	.dc.b	$9C,$73,$AF,$F8,$03,$3F,$D0,$A3,$E4,$1E,$E7,$94,$7B,$A3,$F9,$A9
	.dc.b	$F5,$DE,$4F,$DB,$3F,$9C,$CE,$6B,$ED,$1C,$63,$F9,$A9,$FB,$5B,$FD
	.dc.b	$E9,$F9,$67,$FC,$2D,$3F,$7A,$7E,$C1,$FC,$69,$FE,$AC,$7E,$C2,$7F
	.dc.b	$4F,$3F,$0C,$FE,$60,$7D,$73,$F4,$E7,$F2,$EB,$FE,$54,$FE,$AE,$3E
	.dc.b	$A9,$D6,$FE,$F1,$C2,$FE,$A5,$FF,$6C,$AF,$9D,$1F,$24,$F4,$3F,$52
	.dc.b	$33,$1F,$AC,$3F,$4F,$1F,$42,$BF,$42,$F9,$E8,$FF,$25,$AF,$ED,$B5
	.dc.b	$FC,$BE,$7E,$44,$FF,$90,$0F,$EC,$F8,$7F,$68,$9F,$ED,$B5,$F3,$E3
	.dc.b	$F9,$14,$70,$AF,$F0,$95,$7E,$A2,$FE,$7C,$F8,$5F,$EC,$31,$FC,$96
	.dc.b	$3E,$3D,$FE,$20,$FD,$0C,$7E,$9A,$7E,$79,$F4,$DF,$9D,$3F,$C7,$6B
	.dc.b	$C1,$5F,$D6,$8F,$D2,$57,$CE,$9F,$EB,$D3,$F3,$9F,$D0,$9E,$73,$F4
	.dc.b	$13,$FC,$6A,$FF,$8E,$D7,$F5,$A9,$F3,$CF,$F8,$79,$F9,$C7,$F1,$7B
	.dc.b	$F9,$D5,$FC,$66,$BF,$8B,$D7,$E8,$2B,$E6,$1F,$D5,$2F,$F6,$F7,$FE
	.dc.b	$76,$7F,$36,$7D,$57,$FC,$CB,$3F,$C6,$E7,$F3,$67,$F6,$CA,$F3,$C7
	.dc.b	$F9,$7A,$FF,$39,$3F,$D3,$2B,$FA,$93,$F9,$A3,$FA,$4B,$F2,$E3,$FA
	.dc.b	$5D,$7C,$BA,$FE,$27,$5F,$2C,$FC,$A5,$7E,$5E,$BE,$4D,$7E,$5E,$BF
	.dc.b	$85,$47,$F9,$A2,$BF,$2F,$1F,$DE,$AB,$F2,$F5,$FD,$0E,$7F,$20,$7C
	.dc.b	$A9,$FE,$FB,$5F,$CE,$6B,$F1,$07,$EB,$A3,$FA,$04,$7C,$B9,$FE,$89
	.dc.b	$1F,$2E,$3F,$A0,$5F,$03,$E5,$D7,$F7,$98,$FC,$B6,$1F,$DE,$23,$FC
	.dc.b	$87,$7F,$28,$FE,$EC,$FE,$56,$7E,$4C,$FF,$8C,$EF,$F8,$14,$7F,$73
	.dc.b	$AF,$C8,$57,$C9,$AF,$F1,$84,$7F,$74,$AF,$93,$5F,$CD,$63,$E4,$D7
	.dc.b	$C8,$8F,$93,$1F,$92,$9F,$93,$5F,$E6,$48,$F9,$35,$FD,$9C,$F9,$35
	.dc.b	$F9,$28,$FE,$DF,$5F,$BE,$8F,$91,$3F,$E1,$83,$FD,$5F,$5F,$CC,$5F
	.dc.b	$E2,$87,$EF,$8F,$ED,$05,$9F,$F2,$E4,$6E,$BF,$C9,$33,$FE,$2B,$8F
	.dc.b	$E6,$2E,$8F,$F3,$69,$FC,$BC,$FE,$AA,$FF,$36,$9D,$1F,$EB,$88,$FE
	.dc.b	$E6,$7F,$B2,$F0,$FC,$BB,$6B,$F8,$8E,$1F,$C2,$27,$C2,$7E,$16,$75
	.dc.b	$5D,$0F,$F6,$A3,$FE,$C3,$BF,$F4,$5C,$7E,$1E,$3F,$DB,$0F,$FB,$A6
	.dc.b	$33,$1F,$0A,$3F,$8F,$3A,$3F,$DD,$55,$F7,$6F,$FB,$64,$7C,$B9,$FF
	.dc.b	$40,$3E,$13,$AC,$F3,$3E,$5C,$7F,$9B,$DF,$C0,$4F,$E8,$CF,$C9,$CF
	.dc.b	$FA,$18,$E4,$7D,$37,$F9,$CB,$FE,$8A,$9F,$B5,$03,$CB,$5F,$85,$7F
	.dc.b	$99,$CF,$C2,$8F,$BC,$F4,$3E,$63,$F8,$19,$FC,$24,$7F,$3E,$74,$FF
	.dc.b	$AC,$4F,$B7,$5E,$67,$DE,$FE,$B2,$FF,$07,$1F,$C4,$8F,$A3,$1E,$47
	.dc.b	$8D,$FD,$83,$DF,$3E,$98,$EE,$FB,$8F,$B8,$FE,$4D,$F7,$56,$A7,$DA
	.dc.b	$FD,$B9,$FC,$B1,$F8,$6B,$FB,$75,$F9,$69,$FB,$73,$A7,$DA,$7F,$49
	.dc.b	$7E,$B9,$F6,$EF,$93,$FE,$C6,$8F,$CD,$3F,$16,$3E,$D9,$ED,$81,$FE
	.dc.b	$B3,$7E,$8D,$FE,$6A,$7E,$D3,$D2,$77,$1F,$AF,$3F,$D8,$B5,$FA,$C9
	.dc.b	$F8,$31,$FE,$B2,$7B,$4F,$E7,$1F,$E0,$31,$F9,$E3,$EC,$9D,$63,$0F
	.dc.b	$EF,$0F,$82,$7E,$52,$FF,$75,$3F,$C9,$23,$F0,$37,$C5,$FB,$37,$F7
	.dc.b	$23,$F8,$1C,$7E,$66,$BF,$C5,$F5,$F6,$E3,$F8,$65,$7D,$BA,$FD,$B4
	.dc.b	$FF,$7E,$BC,$BE,$AA,$FD,$E3,$F5,$5F,$D0,$C7,$EC,$DF,$A0,$7E,$CC
	.dc.b	$FD,$49,$C4,$FB,$F1,$F7,$67,$89,$E9,$8F,$E3,$0F,$EC,$67,$EE,$1F
	.dc.b	$80,$7E,$AC,$FE,$C0,$F4,$5F,$EA,$6F,$E9,$47,$F2,$23,$F2,$33,$FB
	.dc.b	$77,$F3,$F7,$FA,$33,$C0,$7E,$B0,$FE,$42,$7C,$83,$D0,$7D,$F7,$E6
	.dc.b	$3F,$40,$FD,$79,$F8,$E3,$FC,$FB,$5F,$C9,$E7,$FA,$CC,$FC,$F9,$FF
	.dc.b	$14,$9F,$A3,$AF,$D6,$5F,$CE,$3F,$4D,$3F,$36,$7F,$8B,$1E,$63,$F2
	.dc.b	$33,$F4,$67,$27,$E9,$23,$F5,$B5,$F3,$1F,$91,$3F,$3E,$7E,$5C,$FF
	.dc.b	$1A,$BF,$E2,$15,$FC,$52,$FB,$1F,$C6,$0F,$CD,$BF,$C6,$67,$E5,$47
	.dc.b	$F5,$48,$FE,$AB,$1F,$DF,$6F,$E6,$DF,$E7,$0F,$8B,$3F,$C7,$0F,$F1
	.dc.b	$C1,$FC,$38,$F9,$33,$F9,$F3,$F2,$E7,$D0,$7F,$A4,$D7,$C9,$BF,$E3
	.dc.b	$D5,$FD,$1A,$FF,$03,$5F,$BA,$8C,$57,$F6,$49,$F4,$9F,$82,$AE,$35
	.dc.b	$F9,$A9,$FF,$79,$3F,$38,$F9,$53,$F3,$5F,$F7,$B5,$7C,$D9,$FC,$C4
	.dc.b	$7F,$0F,$3E,$6B,$FE,$B7,$8F,$9B,$5F,$D6,$2F,$F2,$F5,$F3,$4F,$E6
	.dc.b	$B5,$F3,$4F,$CB,$1E,$F3,$F8,$2B,$FC,$16,$7F,$A0,$CF,$CD,$9F,$F1
	.dc.b	$55,$7C,$D8,$FF,$16,$47,$E5,$2B,$F8,$24,$7F,$3D,$8F,$F1,$4C,$FF
	.dc.b	$06,$3E,$7B,$F2,$E7,$E4,$1F,$DD,$CF,$CA,$3F,$C0,$5F,$E0,$53,$FE
	.dc.b	$80,$8F,$CA,$47,$C8,$85,$7F,$38,$9F,$8F,$3F,$22,$7F,$B8,$4F,$C3
	.dc.b	$8F,$91,$1F,$ED,$DB,$FC,$A4,$7C,$98,$FC,$A3,$E4,$3F,$98,$9F,$25
	.dc.b	$FE,$E1,$1F,$BE,$BF,$CC,$1F,$32,$FF,$1D,$5F,$BD,$3F,$23,$3F,$94
	.dc.b	$7E,$4D,$7C,$89,$FE,$55,$1F,$22,$FF,$DA,$75,$FC,$A6,$FE,$45,$FE
	.dc.b	$EA,$FF,$73,$7F,$BA,$8F,$E4,$F7,$F8,$C8,$FE,$C1,$3F,$E3,$2B,$FC
	.dc.b	$84,$7A,$CF,$F5,$9B,$F7,$9F,$E9,$67,$E4,$6B,$E5,$BF,$CA,$A3,$F2
	.dc.b	$55,$FE,$1E,$BF,$C9,$5F,$F8,$6A,$BF,$C3,$77,$F2,$E3,$F2,$51,$F2
	.dc.b	$EB,$F2,$51,$FC,$92,$3E,$54,$7F,$27,$8F,$F0,$D5,$7C,$AA,$FD,$EC
	.dc.b	$FF,$84,$2B,$F7,$AF,$F9,$3E,$FE,$53,$FE,$11,$9F,$94,$FF,$92,$A7
	.dc.b	$E5,$1F,$E1,$09,$F9,$4F,$E4,$A7,$E5,$4F,$C7,$8F,$87,$7F,$E0,$C3
	.dc.b	$F8,$F5,$7F,$82,$DF,$94,$7F,$03,$7E,$45,$FF,$82,$67,$FC,$40,$7F
	.dc.b	$25,$79,$D7,$C7,$8F,$C9,$3F,$E4,$6A,$FD,$CC,$7C,$88,$FD,$B5,$7E
	.dc.b	$49,$FD,$C5,$7F,$4F,$AF,$91,$5F,$E3,$F3,$F1,$D5,$F9,$17,$E4,$DF
	.dc.b	$C8,$7F,$A4,$CF,$F8,$F7,$83,$FB,$DC,$FF,$97,$1B,$3F,$82,$C5,$18
	.dc.b	$85,$F1,$7F,$C0,$D5,$FE,$26,$9D,$5A,$F8,$D7,$F8,$20,$57,$F9,$36
	.dc.b	$73,$1F,$CB,$E3,$FD,$0D,$5E,$27,$BD,$6A,$7F,$C3,$91,$FE,$78,$BF
	.dc.b	$14,$71,$C0,$EC,$FF,$48,$4F,$F3,$6A,$C4,$7E,$8B,$0E,$2F,$4B,$FF
	.dc.b	$24,$3A,$3C,$EF,$F9,$8C,$F6,$61,$F2,$1C,$4E,$EF,$E2,$BF,$E3,$29
	.dc.b	$D3,$FE,$6B,$BE,$97,$FE,$36,$8F,$0B,$E6,$C3,$FC,$71,$5F,$51,$FA
	.dc.b	$4D,$8E,$AD,$8F,$7B,$99,$F2,$CF,$F8,$DA,$7E,$7C,$76,$8F,$F4,$FC
	.dc.b	$7F,$A3,$1F,$43,$EE,$AB,$7D,$A7,$EE,$CF,$D8,$3B,$B8,$8F,$65,$FD
	.dc.b	$DB,$FB,$87,$C3,$3E,$5B,$F8,$E7,$E5,$5E,$5E,$D5,$FE,$7C,$7F,$CE
	.dc.b	$6E,$5F,$5B,$D5,$F9,$CF,$C6,$3F,$22,$7C,$62,$9D,$0F,$C3,$9F,$B2
	.dc.b	$9F,$CA,$CF,$D2,$8F,$8B,$3D,$2B,$F2,$31,$D1,$FC,$8D,$7E,$4F,$0F
	.dc.b	$95,$3F,$E9,$6C,$3E,$1B,$F9,$9B,$F3,$BF,$DF,$9D,$DF,$E1,$23,$67
	.dc.b	$E3,$E3,$AB,$FA,$AA,$FC,$5D,$FE,$22,$F7,$3F,$9A,$9F,$D4,$61,$F8
	.dc.b	$A9,$FF,$3D,$57,$E1,$6B,$67,$DE,$3A,$9F,$EA,$88,$FE,$37,$1F,$89
	.dc.b	$8F,$C0,$9F,$68,$FC,$46,$1F,$B5,$8F,$C9,$3F,$12,$39,$3D,$1F,$D9
	.dc.b	$CF,$E9,$B0,$FC,$4B,$E2,$8F,$F6,$A4,$7C,$47,$FC,$AC,$F5,$7F,$84
	.dc.b	$BF,$86,$6B,$D5,$FD,$CB,$F1,$1C,$C7,$59,$FC,$7D,$FD,$B9,$E8,$7F
	.dc.b	$8E,$4F,$41,$F8,$23,$A1,$FF,$00,$0F,$86,$7E,$22,$7A,$3F,$C2,$AB
	.dc.b	$D0,$FF,$10,$7E,$DB,$FA,$9B,$FC,$25,$7E,$27,$0C,$E1,$FC,$06,$BF
	.dc.b	$15,$7C,$9F,$E9,$6F,$E1,$27,$E0,$47,$FA,$92,$3C,$AF,$CB,$8D,$47
	.dc.b	$38,$D3,$CD,$FE,$14,$FD,$A7,$E2,$61,$FA,$37,$F0,$07,$E8,$CF,$E1
	.dc.b	$C7,$33,$EC,$9F,$81,$3F,$D8,$0F,$E9,$30,$FC,$0C,$74,$BF,$29,$FE
	.dc.b	$D4,$3A,$3F,$C3,$1E,$8F,$F1,$B7,$A3,$FE,$CD,$AF,$AA,$FC,$B7,$EA
	.dc.b	$BF,$79,$EA,$7D,$F7,$F8,$23,$D1,$FF,$37,$BD,$1F,$F6,$B3,$F5,$1E
	.dc.b	$67,$DD,$3C,$E7,$DE,$3F,$09,$5F,$79,$E4,$78,$EB,$FA,$83,$F8,$2A
	.dc.b	$E2,$7B,$9F,$E0,$07,$F0,$7A,$FA,$57,$FC,$82,$BF,$9F,$CF,$E5,$63
	.dc.b	$F8,$FC,$7F,$42,$7E,$DD,$E6,$BC,$35,$FE,$06,$7F,$83,$C7,$F1,$D8
	.dc.b	$FB,$33,$F8,$9B,$FF,$56,$D7,$EE,$2B,$EF,$D7,$23,$EA,$47,$29,$F8
	.dc.b	$71,$FC,$E5,$FB,$33,$FB,$7B,$FD,$EB,$F7,$6B,$F1,$87,$AE,$BF,$66
	.dc.b	$7E,$BD,$F5,$BF,$B6,$BF,$E3,$B1,$FC,$E2,$BF,$9F,$3E,$39,$F5,$1F
	.dc.b	$E0,$1B,$FD,$B5,$7F,$2A,$8F,$F0,$41,$FE,$05,$3F,$19,$55,$FC,$39
	.dc.b	$FB,$B7,$53,$FD,$F8,$FE,$74,$7D,$58,$FE,$41,$1A,$BF,$F0,$7B,$C2
	.dc.b	$FC,$17,$FB,$0A,$FF,$59,$9A,$3F,$C1,$27,$FA,$D9,$DC,$7E,$46,$3F
	.dc.b	$C8,$75,$B7,$67,$F8,$2E,$3F,$9A,$3F,$40,$FF,$34,$DE,$2B,$E0,$57
	.dc.b	$F5,$39,$FE,$EC,$7C,$02,$D6,$AB,$94,$FF,$8A,$27,$C0,$78,$0F,$D9
	.dc.b	$56,$E3,$F9,$84,$7E,$D5,$DB,$E6,$7F,$78,$F1,$3F,$8F,$57,$9C,$1F
	.dc.b	$36,$FF,$B8,$57,$D7,$8F,$BD,$3C,$4F,$F7,$84,$FF,$75,$8F,$B9,$1F
	.dc.b	$80,$3F,$C9,$B1,$F6,$E7,$31,$F2,$CF,$9F,$5E,$67,$F8,$34,$FD,$E7
	.dc.b	$D6,$7C,$1B,$FB,$F5,$D9,$E5,$1F,$9D,$9F,$BD,$7F,$E2,$EA,$FB,$D3
	.dc.b	$FE,$C9,$AC,$4F,$E1,$E7,$D7,$3F,$0D,$F8,$37,$E3,$8F,$E6,$F3,$F7
	.dc.b	$1F,$9B,$1F,$E4,$CB,$FF,$2E,$3F,$C3,$63,$EA,$3F,$75,$FE,$56,$7D
	.dc.b	$D2,$CF,$DD,$79,$DF,$CF,$7F,$B8,$DF,$DB,$7F,$BF,$47,$17,$AB,$FA
	.dc.b	$47,$FD,$A2,$7F,$93,$27,$EC,$BE,$93,$FC,$DD,$3F,$66,$3E,$81,$EE
	.dc.b	$9E,$0F,$39,$FC,$25,$FE,$C5,$E4,$7B,$4F,$5B,$FB,$5B,$CC,$71,$3F
	.dc.b	$98,$47,$D9,$AF,$AF,$5F,$B6,$AF,$B7,$1E,$BA,$FE,$7B,$5F,$E8,$5B
	.dc.b	$FA,$F1,$FE,$D6,$3D,$4F,$F4,$37,$F4,$0F,$D8,$6D,$F3,$7F,$65,$3F
	.dc.b	$51,$FD,$49,$F3,$4F,$A7,$3F,$B2,$7F,$25,$7F,$E8,$83,$E4,$1F,$E2
	.dc.b	$C7,$FC,$46,$7D,$63,$37,$FC,$3E,$3F,$CC,$91,$F5,$9C,$DF,$CF,$8F
	.dc.b	$D9,$BF,$52,$7E,$7D,$7D,$3A,$CC,$FA,$6B,$F6,$6F,$E9,$9F,$D0,$9E
	.dc.b	$93,$F6,$13,$FA,$A9,$C9,$F4,$A7,$F4,$93,$F3,$8F,$B3,$5F,$C9,$5F
	.dc.b	$3D,$FF,$25,$8F,$E5,$33,$FC,$7A,$7F,$84,$DF,$9E,$BE,$A4,$FE,$9C
	.dc.b	$FD,$C4,$7E,$26,$33,$1F,$2E,$7F,$4A,$79,$CF,$E2,$07,$C7,$3E,$79
	.dc.b	$EF,$3F,$C3,$31,$FA,$AA,$FE,$3E,$71,$8F,$E0,$AF,$CF,$3E,$B9,$C6
	.dc.b	$FF,$8D,$C7,$F8,$0A,$BF,$8D,$1C,$5F,$E3,$B7,$F3,$DF,$CF,$CF,$E9
	.dc.b	$67,$FA,$F4,$7F,$A3,$6F,$FD,$BC,$FF,$81,$6B,$FB,$83,$FA,$13,$FC
	.dc.b	$19,$5F,$E0,$9A,$FF,$19,$47,$88,$FF,$47,$DF,$F7,$03,$FD,$AD,$87
	.dc.b	$F9,$16,$7F,$D1,$D5,$FE,$0D,$7F,$C1,$27,$EC,$67,$FD,$0D,$5F,$EC
	.dc.b	$BB,$FE,$DE,$FF,$91,$5F,$F0,$4B,$99,$FF,$00,$57,$E5,$E3,$FC,$7E
	.dc.b	$7F,$81,$27,$FC,$11,$1E,$98,$FC,$BB,$FE,$08,$9F,$F2,$53,$E9,$8F
	.dc.b	$F1,$C1,$FE,$3F,$9E,$35,$FE,$21,$3F,$A6,$47,$FB,$1A,$BF,$C0,$E7
	.dc.b	$A4,$FD,$CB,$FE,$AC,$9F,$F0,$C4,$7E,$E6,$3F,$C7,$D5,$FE,$0B,$8F
	.dc.b	$F0,$34,$7C,$13,$FA,$41,$FE,$39,$9F,$A2,$F8,$4F,$E5,$87,$F9,$EE
	.dc.b	$3E,$8D,$71,$3E,$84,$7F,$B6,$F0,$FF,$5C,$4F,$D0,$8F,$F1,$C6,$1F
	.dc.b	$D0,$A3,$FC,$DB,$1F,$E3,$3B,$FF,$17,$1E,$77,$FC,$1C,$FF,$0F,$7F
	.dc.b	$D6,$51,$F4,$1F,$B1,$87,$A1,$FE,$07,$3F,$41,$F9,$11,$E8,$7F,$CE
	.dc.b	$33,$F4,$30,$E2,$7E,$1E,$7D,$13,$FE,$ED,$AF,$F1,$D4,$7F,$8D,$6F
	.dc.b	$F9,$BD,$FF,$8C,$0F,$C7,$1F,$86,$AF,$F3,$65,$66,$BE,$74,$FE,$5A
	.dc.b	$7E,$71,$F1,$8F,$F3,$29,$F3,$4F,$F0,$F5,$FF,$88,$A3,$E6,$9F,$E9
	.dc.b	$C8,$FF,$32,$57,$E9,$27,$EA,$1C,$20,$FF,$94,$E3,$F5,$B1,$FE,$DF
	.dc.b	$3E,$75,$FE,$66,$3F,$37,$87,$F8,$2E,$91,$FE,$DE,$8F,$E7,$57,$FE
	.dc.b	$64,$BF,$8C,$7F,$96,$8F,$9A,$7A,$30,$FF,$5C,$61,$FD,$A2,$7F,$CC
	.dc.b	$53,$F9,$B8,$57,$FB,$72,$FE,$6C,$FE,$6E,$7E,$6D,$6A,$3F,$CC,$25
	.dc.b	$3F,$D2,$73,$F9,$C8,$FD,$21,$F8,$58,$FF,$4A,$BF,$3C,$FF,$4A,$3F
	.dc.b	$C2,$E7,$93,$FE,$D3,$9F,$F4,$C5,$FF,$0B,$9F,$E6,$75,$AA,$FF,$29
	.dc.b	$E1,$FD,$B1,$FF,$1C,$3F,$C3,$30,$FC,$E9,$F2,$63,$F9,$E9,$FE,$D0
	.dc.b	$AF,$CF,$57,$F0,$78,$FA,$33,$FE,$93,$8F,$E1,$57,$FE,$93,$7F,$81
	.dc.b	$CF,$E7,$0F,$9F,$87,$E6,$CF,$A5,$3F,$C2,$6F,$E7,$57,$9E,$BF,$D8
	.dc.b	$47,$CF,$BF,$9D,$19,$8F,$F4,$95,$7E,$6E,$7F,$C9,$13,$9C,$3F,$CA
	.dc.b	$2F,$E7,$0F,$F6,$19,$F3,$8F,$CA,$9F,$85,$AF,$F0,$B4,$FF,$85,$E7
	.dc.b	$FD,$1B,$3D,$E7,$FD,$1A,$7F,$4D,$9F,$F4,$64,$7F,$66,$8F,$F4,$49
	.dc.b	$FD,$9B,$0F,$F2,$65,$FF,$85,$8F,$ED,$33,$FD,$96,$7F,$C9,$27,$F8
	.dc.b	$51,$FE,$C9,$1F,$DA,$1F,$EB,$93,$FA,$09,$FE,$C2,$7F,$64,$3F,$C1
	.dc.b	$A7,$F6,$39,$FE,$CE,$62,$7F,$C8,$91,$FE,$40,$3F,$C1,$F3,$FE,$43
	.dc.b	$AF,$EC,$D5,$FE,$0E,$AF,$F1,$CD,$7F,$82,$8F,$EB,$8F,$FA,$8C,$FC
	.dc.b	$A3,$FE,$7E,$8C,$C7,$F8,$36,$3E,$31,$FE,$5B,$3F,$D6,$55,$FE,$0D
	.dc.b	$72,$7F,$77,$3F,$A7,$1F,$92,$3F,$C1,$D7,$FE,$3E,$BF,$F5,$84,$7F
	.dc.b	$8F,$E3,$F3,$71,$FE,$0A,$8F,$8C,$7C,$99,$FF,$56,$DF,$C9,$3E,$65
	.dc.b	$FF,$82,$E7,$FC,$E1,$7F,$EA,$98,$FF,$56,$CF,$C8,$8F,$F5,$7C,$7C
	.dc.b	$8B,$FF,$05,$57,$C8,$8F,$F1,$A3,$FE,$35,$9F,$CB,$CF,$FA,$66,$3F
	.dc.b	$1E,$7C,$98,$FF,$1A,$CF,$F8,$1A,$3F,$C0,$F1,$FC,$12,$7F,$CC,$91
	.dc.b	$F1,$0F,$F3,$0D,$FC,$78,$FF,$32,$4F,$C8,$AF,$C7,$47,$F9,$8C,$FF
	.dc.b	$6C,$CF,$E3,$AB,$E3,$47,$FA,$60,$F8,$CF,$E3,$AB,$FD,$2F,$3F,$1A
	.dc.b	$3F,$D3,$37,$F1,$AF,$FD,$29,$1F,$1A,$FF,$CC,$15,$FE,$5D,$AF,$8D
	.dc.b	$1F,$CA,$AB,$F1,$95,$F8,$DB,$FF,$46,$DF,$E3,$63,$FD,$79,$7F,$EA
	.dc.b	$B6,$CF,$F9,$F6,$BF,$BE,$C7,$C8,$8F,$8F,$7F,$2E,$7F,$1F,$3F,$EC
	.dc.b	$43,$FD,$8C,$FE,$8D,$FF,$64,$47,$ED,$CF,$F6,$24,$7E,$3A,$BF,$CB
	.dc.b	$75,$F2,$63,$F6,$B5,$FB,$68,$FC,$73,$F6,$5F,$F6,$09,$F9,$53,$FD
	.dc.b	$25,$7E,$63,$FC,$B5,$1F,$E6,$89,$FD,$BB,$F2,$A7,$F9,$2D,$FF,$A1
	.dc.b	$E3,$F6,$D1,$F8,$F8,$FF,$45,$5F,$CA,$9F,$F7,$AD,$7C,$A9,$FC,$7D
	.dc.b	$7C,$A8,$FC,$7C,$7E,$DA,$3F,$1F,$3F,$0A,$7F,$CA,$AF,$ED,$AF,$FC
	.dc.b	$A9,$5F,$26,$7F,$CC,$87,$C7,$91,$FD,$CD,$FE,$4C,$FF,$83,$EB,$2F
	.dc.b	$F7,$AA,$FE,$96,$E8,$FF,$7D,$E1,$FB,$27,$F7,$13,$FE,$26,$3F,$B7
	.dc.b	$C7,$F9,$56,$FF,$A5,$47,$F7,$B3,$FB,$D5,$FF,$95,$6F,$FC,$7B,$1F
	.dc.b	$DD,$A3,$8C,$FF,$4B,$3F,$BD,$4F,$F1,$09,$D4,$FF,$4C,$3F,$BD,$5F
	.dc.b	$F9,$56,$FF,$A0,$61,$FE,$55,$A3,$FE,$27,$9F,$83,$18,$7F,$CB,$EF
	.dc.b	$F8,$A1,$FE,$9C,$7F,$89,$EB,$9C,$7F,$80,$CF,$83,$87,$F4,$E3,$FC
	.dc.b	$1D,$1E,$A3,$FC,$51,$3F,$DB,$9F,$F2,$03,$FE,$1A,$7D,$47,$F4,$F7
	.dc.b	$FD,$0C,$7F,$21,$3F,$C5,$58,$7F,$95,$8F,$A3,$3F,$DC,$E3,$C2,$7E
	.dc.b	$5E,$7F,$C5,$4F,$DA,$AF,$92,$FD,$4A,$4F,$F9,$5E,$BE,$9B,$FE,$86
	.dc.b	$3F,$CA,$F3,$FE,$29,$C3,$FC,$55,$1F,$DE,$A3,$FD,$08,$FC,$F3,$E6
	.dc.b	$BF,$DE,$A2,$D7,$F7,$A8,$CC,$FF,$94,$CF,$F1,$3B,$F2,$A3,$FD,$02
	.dc.b	$7F,$89,$AB,$FC,$6D,$87,$C5,$BF,$E4,$95,$FE,$26,$9F,$9B,$7F,$DE
	.dc.b	$2F,$FC,$A5,$3F,$DF,$63,$E5,$3F,$A3,$9F,$F5,$7B,$F9,$79,$F9,$53
	.dc.b	$F9,$9B,$FF,$12,$57,$F2,$5A,$FC,$C9,$FE,$52,$BF,$CC,$3F,$9C,$9F
	.dc.b	$F5,$64,$7E,$F2,$3F,$C4,$F7,$F9,$83,$EF,$BF,$E7,$CB,$FF,$4E,$47
	.dc.b	$E5,$CF,$2C,$7F,$B6,$23,$F2,$EF,$C8,$9F,$93,$1F,$32,$BF,$C4,$57
	.dc.b	$FE,$D9,$8F,$92,$FF,$93,$67,$E5,$47,$E5,$27,$FC,$41,$5F,$C0,$E3
	.dc.b	$FC,$F4,$FF,$A8,$EF,$FC,$98,$7C,$8B,$F9,$73,$FE,$4B,$3F,$79,$3F
	.dc.b	$E4,$B8,$FC,$74,$FE,$42,$3F,$29,$3F,$E7,$83,$F1,$CF,$C7,$9F,$F2
	.dc.b	$3B,$F1,$CF,$F5,$1B,$FE,$FA,$8F,$F2,$34,$7E,$36,$3F,$78,$FC,$A8
	.dc.b	$FF,$42,$CF,$C6,$8F,$E2,$35,$FE,$04,$7F,$CA,$EF,$F1,$53,$F8,$CC
	.dc.b	$7F,$92,$67,$E4,$4F,$E4,$2F,$F9,$2B,$FC,$68,$FF,$24,$9F,$91,$7E
	.dc.b	$4D,$FE,$43,$0F,$DE,$57,$FB,$40,$F9,$35,$F8,$D7,$F2,$05,$C3,$FD
	.dc.b	$43,$7F,$DD,$E7,$FB,$C5,$7F,$91,$AF,$FC,$05,$1F,$C3,$CF,$15,$7F
	.dc.b	$A6,$A3,$FC,$8F,$7F,$D4,$9F,$EF,$53,$FD,$47,$0F,$EF,$51,$CA,$FF
	.dc.b	$A9,$57,$F9,$1E,$3F,$C4,$11,$FD,$EA,$3F,$BC,$D7,$F5,$19,$FA,$07
	.dc.b	$F5,$17,$FB,$D1,$FD,$C8,$FF,$23,$1F,$E2,$09,$FF,$0F,$CF,$F5,$18
	.dc.b	$FE,$EA,$7F,$87,$E7,$FC,$8A,$FF,$87,$DF,$CA,$47,$F9,$1A,$3F,$C8
	.dc.b	$F3,$FD,$E1,$FE,$A9,$7F,$D3,$AB,$FA,$93,$FE,$46,$AF,$91,$1F,$DD
	.dc.b	$E7,$FC,$8B,$1F,$DD,$DF,$EF,$15,$FD,$DE,$7E,$14,$7F,$91,$1F,$F0
	.dc.b	$FC,$FF,$93,$A7,$FB,$BD,$7F,$9B,$23,$FC,$D7,$5F,$E1,$FA,$FF,$1A
	.dc.b	$5F,$F3,$28,$FE,$9D,$5F,$E4,$5B,$FD,$E4,$7E,$1C,$FE,$CB,$1F,$EF
	.dc.b	$17,$FA,$31,$FE,$6E,$9F,$F1,$13,$F2,$6B,$FC,$44,$7F,$91,$9F,$F1
	.dc.b	$13,$FE,$43,$9F,$F6,$CC,$FF,$88,$E7,$B1,$FD,$52,$A5,$7C,$C3,$FB
	.dc.b	$C4,$7F,$92,$5F,$EF,$53,$FD,$F2,$3E,$6C,$FF,$26,$7F,$1E,$7E,$4A
	.dc.b	$7F,$C4,$26,$2B,$FC,$42,$FE,$40,$F9,$71,$FE,$21,$7E,$59,$FE,$6E
	.dc.b	$3F,$21,$1F,$2D,$FF,$24,$9F,$90,$3E,$5D,$FF,$88,$1B,$1F,$E4,$97
	.dc.b	$F2,$07,$F8,$86,$BF,$C4,$15,$F9,$19,$FF,$10,$3F,$09,$FD,$F3,$FC
	.dc.b	$02,$3F,$C2,$33,$FE,$22,$3E,$53,$FE,$05,$7F,$26,$7F,$88,$CF,$91
	.dc.b	$1F,$E2,$33,$E4,$1F,$E2,$5B,$F9,$0F,$D0,$BF,$EC,$77,$F9,$3A,$FF
	.dc.b	$38,$9F,$93,$9F,$6B,$F2,$5F,$C9,$BF,$25,$FF,$23,$1F,$93,$3F,$71
	.dc.b	$3F,$93,$9F,$F5,$F5,$FF,$89,$67,$EF,$DF,$F9,$26,$BF,$C2,$37,$F2
	.dc.b	$A3,$FC,$E4,$FC,$AA,$FC,$9D,$7E,$4E,$FF,$24,$7F,$AA,$DF,$90,$7E
	.dc.b	$EA,$7F,$C9,$73,$FB,$A7,$F2,$47,$C9,$8F,$F3,$7D,$7E,$45,$E9,$3F
	.dc.b	$90,$7F,$CE,$77,$FE,$06,$AF,$C8,$C7,$EF,$1F,$C8,$C7,$FA,$65,$FD
	.dc.b	$DC,$FE,$46,$7F,$B8,$1F,$E1,$93,$FA,$51,$FE,$1B,$AF,$E9,$4F,$EB
	.dc.b	$63,$FC,$9B,$5F,$D2,$EB,$FC,$8B,$3F,$D2,$CF,$F0,$C4,$7F,$85,$DF
	.dc.b	$F2,$6B,$FE,$17,$3F,$A5,$C7,$F9,$11,$FF,$0F,$C7,$5A,$FF,$17,$D7
	.dc.b	$EF,$A3,$FC,$33,$87,$C5,$7F,$63,$1F,$BC,$7F,$B9,$5F,$F8,$49,$FC
	.dc.b	$71,$AC,$3F,$9D,$4F,$F9,$1F,$0F,$F0,$E1,$FE,$44,$9F,$F2,$39,$ED
	.dc.b	$9F,$DF,$C7,$F7,$58,$FF,$24,$BF,$D3,$4F,$E9,$8F,$F7,$5B,$FE,$9D
	.dc.b	$87,$F7,$58,$F2,$CF,$F6,$37,$FA,$D3,$F5,$5F,$9C,$7F,$87,$67,$FC
	.dc.b	$3D,$1F,$E1,$D8,$FF,$3B,$3F,$35,$FF,$0E,$9F,$E4,$59,$4F,$F8,$7A
	.dc.b	$3E,$A0,$BF,$F3,$B5,$7C,$FC,$3E,$74,$FF,$3B,$3F,$BC,$5F,$28,$FE
	.dc.b	$F7,$1F,$46,$7F,$BD,$9F,$E5,$67,$F2,$EF,$F8,$82,$3E,$74,$FF,$79
	.dc.b	$7E,$86,$1F,$E2,$0A,$FC,$9D,$7C,$D7,$FB,$C4,$FF,$77,$AF,$F2,$44
	.dc.b	$7D,$5B,$EC,$FF,$39,$AF,$F4,$F4,$7F,$BC,$A3,$E6,$E1,$F3,$A7,$FB
	.dc.b	$D4,$7F,$7D,$7F,$C4,$57,$FE,$48,$3F,$C4,$33,$F3,$1F,$F3,$86,$1F
	.dc.b	$E4,$8A,$FF,$38,$57,$F8,$89,$F9,$95,$F2,$DF,$F0,$85,$FF,$A3,$AB
	.dc.b	$FC,$93,$7F,$DD,$27,$FC,$40,$7C,$B3,$E4,$61,$F2,$DF,$9B,$87,$F9
	.dc.b	$C6,$7F,$3C,$FF,$9C,$5F,$CB,$47,$FA,$7D,$F9,$71,$FE,$22,$7F,$34
	.dc.b	$FE,$59,$FF,$0F,$47,$E5,$8F,$2D,$7E,$5A,$3F,$C4,$4F,$F9,$2A,$3F
	.dc.b	$93,$4F,$FA,$20,$FC,$BB,$F2,$A7,$FC,$E7,$5F,$97,$8F,$DE,$4F,$E5
	.dc.b	$DF,$F6,$64,$FC,$B7,$F8,$16,$1F,$2D,$FF,$4F,$5F,$CB,$7F,$CF,$B3
	.dc.b	$FC,$09,$F9,$6F,$F9,$31,$FC,$AE,$1F,$94,$AF,$E4,$B1,$FE,$D6,$8F
	.dc.b	$F6,$43,$F2,$CF,$A3,$87,$F9,$D1,$FD,$9B,$FA,$97,$F2,$D7,$B7,$94
	.dc.b	$7A,$63,$FA,$9C,$7E,$8D,$FF,$28,$3F,$C3,$8F,$92,$7F,$06,$3F,$80
	.dc.b	$1F,$EC,$BB,$FA,$37,$FE,$52,$8F,$9F,$7F,$D7,$1F,$CC,$5F,$FA,$8B
	.dc.b	$0F,$E0,$D7,$FE,$53,$7F,$83,$3F,$EA,$09,$FE,$0C,$FD,$08,$FC,$CB
	.dc.b	$FE,$7A,$BF,$F1,$5C,$FF,$80,$AF,$FC,$A4,$FF,$94,$B0,$FF,$00,$BF
	.dc.b	$E5,$4B,$F8,$4F,$FB,$3E,$FF,$CA,$2F,$F2,$B7,$FC,$04,$FF,$85,$8F
	.dc.b	$CD,$57,$F5,$B7,$FC,$4D,$87,$F5,$B7,$FC,$B8,$6D,$FF,$17,$1F,$DB
	.dc.b	$9F,$F2,$A1,$9A,$FC,$65,$7F,$81,$A3,$FC,$A8,$7F,$8B,$61,$1F,$E2
	.dc.b	$98,$F9,$87,$F8,$26,$BF,$9D,$3F,$E0,$97,$E6,$34,$FF,$2C,$1E,$9B
	.dc.b	$FB,$4F,$F9,$64,$FE,$1B,$3F,$D7,$8F,$F1,$7C,$FF,$96,$63,$F4,$33
	.dc.b	$F8,$57,$FA,$BD,$7F,$67,$7F,$4A,$FC,$A7,$E3,$1F,$E3,$13,$FC,$B7
	.dc.b	$5F,$E0,$93,$FC,$65,$5F,$E5,$B8,$FF,$04,$9F,$E3,$13,$FC,$13,$1F
	.dc.b	$41,$FF,$4A,$3F,$ED,$57,$FC,$13,$3E,$E7,$E8,$BF,$E0,$A7,$E6,$5F
	.dc.b	$F8,$29,$F9,$D7,$FE,$33,$7E,$69,$FE,$82,$3F,$C1,$31,$FE,$5A,$AF
	.dc.b	$F0,$49,$FE,$5A,$8F,$F4,$09,$FE,$5A,$3F,$C1,$35,$F5,$1F,$99,$7F
	.dc.b	$51,$FA,$57,$F8,$27,$EA,$4F,$C5,$7E,$49,$FE,$34,$3F,$86,$DF,$F8
	.dc.b	$D1,$FE,$71,$1F,$35,$AF,$17,$FC,$CF,$7F,$E3,$57,$F1,$4F,$CD,$7F
	.dc.b	$3D,$7F,$35,$FC,$C3,$F1,$DF,$E1,$31,$FE,$5E,$7E,$BC,$7F,$A2,$0F
	.dc.b	$CE,$3F,$EC,$7A,$FC,$E4,$7F,$AB,$67,$FC,$BB,$3F,$9A,$3E,$9D,$79
	.dc.b	$5F,$F0,$54,$7F,$97,$CF,$CC,$57,$CD,$8F,$CC,$47,$CD,$7F,$2C,$7C
	.dc.b	$D3,$F3,$51,$FE,$35,$7F,$D8,$F5,$F0,$DB,$5D,$C1,$ED,$7E,$6C,$7E
	.dc.b	$65,$F9,$77,$F9,$94,$7F,$3B,$7F,$CC,$0F,$FA,$20,$FC,$C1,$FE,$60
	.dc.b	$3F,$84,$1F,$E6,$03,$F8,$44,$FF,$B5,$DF,$F3,$0C,$FF,$AC,$6B,$FD
	.dc.b	$AF,$1F,$32,$BF,$D1,$0F,$EF,$CF,$11,$FC,$DA,$7E,$63,$F9,$63,$E6
	.dc.b	$1F,$9C,$AF,$99,$3F,$E6,$08,$FC,$F1,$F9,$88,$FC,$93,$FE,$62,$8F
	.dc.b	$F4,$5B,$FE,$67,$3F,$86,$9F,$9A,$BF,$F3,$13,$F3,$A7,$8B,$FE,$62
	.dc.b	$9F,$29,$FC,$FE,$BF,$9C,$47,$F8,$C2,$7E,$73,$F2,$9F,$A2,$FE,$7C
	.dc.b	$FC,$CD,$FE,$6D,$E1,$5F,$39,$FE,$0B,$3F,$63,$87,$F2,$B1,$F5,$0F
	.dc.b	$AA,$FF,$84,$0C,$AA,$E0,$FF,$68,$11,$A9,$FF,$1D,$BF,$C4,$A3,$FC
	.dc.b	$75,$1C,$1F,$ED,$27,$63,$FC,$C2,$7F,$B5,$31,$E6,$FF,$9C,$27,$CB
	.dc.b	$83,$0E,$13,$FD,$CA,$FB,$BF,$CE,$DE,$CF,$F7,$07,$D9,$5F,$E7,$5C
	.dc.b	$37,$3F,$E1,$E9,$EE,$FF,$72,$29,$C5,$DB,$F6,$DE,$AF,$F7,$FA,$F9
	.dc.b	$0F,$03,$07,$C7,$C3,$0F,$F8,$C4,$F9,$53,$87,$FA,$1C,$FB,$1F,$EF
	.dc.b	$71,$E0,$7F,$5E,$EF,$0F,$F7,$B4,$6C,$F1,$3D,$F0,$FF,$2A,$4E,$EB
	.dc.b	$FA,$7C,$7F,$7B,$BF,$05,$FF,$8A,$EB,$D6,$FE,$D2,$FF,$CA,$D5,$FE
	.dc.b	$2C,$AF,$05,$7E,$FB,$0D,$DF,$DC,$8F,$A8,$E9,$FE,$07,$3D,$A3,$F2
	.dc.b	$D1,$E9,$7E,$3D,$7E,$35,$FF,$5A,$46,$5E,$71,$F8,$B9,$E5,$7F,$E4
	.dc.b	$18,$FF,$1A,$1F,$84,$BF,$8D,$5E,$93,$FD,$3F,$1F,$88,$C3,$E3,$46
	.dc.b	$6B,$CB,$3F,$D9,$AB,$E1,$3F,$E8,$C9,$F9,$D1,$EF,$BB,$87,$D8,$AF
	.dc.b	$F5,$F6,$1D,$2F,$FD,$5B,$3E,$27,$F8,$EC,$7C,$09,$FF,$32,$C7,$D3
	.dc.b	$3E,$B1,$FE,$BB,$3A,$16,$3A,$61,$FE,$BD,$BF,$AF,$1F,$3E,$BF,$CD
	.dc.b	$35,$F7,$6B,$9B,$FE,$BF,$79,$B6,$79,$9F,$B2,$AF,$C0,$DF,$C0,$9F
	.dc.b	$F5,$EB,$FA,$66,$E1,$EB,$BF,$F3,$83,$F8,$07,$F9,$C4,$7D,$F9,$F8
	.dc.b	$0F,$F8,$53,$0F,$E0,$EF,$A8,$FE,$7F,$5C,$DB,$3F,$7F,$0D,$47,$D4
	.dc.b	$3F,$D3,$75,$F7,$9F,$E0,$B7,$EE,$7F,$D3,$67,$DE,$33,$3D,$E3,$94
	.dc.b	$7F,$64,$AF,$B2,$FC,$CB,$FB,$97,$FC,$0E,$7E,$ED,$E5,$E3,$5F,$BF
	.dc.b	$7E,$C3,$F2,$8F,$11,$FB,$EA,$F8,$93,$F8,$08,$FB,$D3,$F1,$23,$EC
	.dc.b	$47,$FB,$14,$FA,$77,$C9,$FF,$2D,$D7,$28,$FD,$0C,$7F,$B2,$EB,$E6
	.dc.b	$9F,$5D,$FF,$4E,$BF,$5D,$B3,$F5,$E3,$F6,$C7,$EC,$EB,$F1,$6F,$FA
	.dc.b	$A7,$0F,$DA,$57,$EA,$CF,$F7,$CE,$1F,$B3,$7F,$DE,$27,$E2,$8F,$9E
	.dc.b	$79,$0C,$CF,$E0,$1F,$69,$FC,$F0,$FD,$F6,$1E,$D7,$FD,$F4,$7D,$73
	.dc.b	$FD,$53,$7F,$5E,$7F,$13,$87,$B6,$7F,$9C,$1F,$4D,$FC,$91,$EC,$BF
	.dc.b	$D8,$C7,$ED,$27,$E7,$3E,$B3,$FD,$42,$7D,$79,$FC,$D5,$7E,$42,$BF
	.dc.b	$DF,$B3,$F5,$B0,$FD,$6E,$1C,$63,$FD,$4A,$7D,$6B,$E3,$1F,$0E,$3F
	.dc.b	$97,$DF,$DA,$C3,$EA,$9C,$63,$FD,$84,$FF,$74,$7F,$92,$57,$D5,$8E
	.dc.b	$51,$FC,$BD,$FF,$4A,$61,$FE,$A9,$3F,$B4,$47,$33,$F8,$E1,$F7,$CF
	.dc.b	$A7,$3F,$AB,$79,$9F,$38,$FA,$C7,$D8,$9F,$AD,$7C,$E3,$E9,$BC,$DF
	.dc.b	$F6,$C1,$FD,$C6,$BF,$96,$4F,$DA,$BF,$AD,$87,$F4,$8C,$3E,$B5,$FF
	.dc.b	$B5,$AB,$93,$FD,$26,$3F,$5A,$FF,$64,$31,$1E,$6C,$3F,$B0,$DF,$EB
	.dc.b	$E3,$EC,$9F,$AA,$9E,$67,$FA,$9A,$BE,$CB,$C6,$7F,$5D,$87,$FB,$70
	.dc.b	$FF,$4F,$BF,$92,$8F,$B2,$7F,$25,$9F,$B2,$7E,$B6,$3E,$C6,$1F,$C8
	.dc.b	$AF,$E2,$61,$FD,$82,$FF,$90,$CF,$D8,$9F,$D5,$47,$F1,$F8,$FD,$4C
	.dc.b	$7F,$23,$8F,$EB,$67,$F8,$68,$FD,$04,$FF,$8A,$EF,$51,$F5,$0F,$9E
	.dc.b	$7F,$0E,$3F,$77,$87,$F1,$F7,$E4,$4F,$CB,$BF,$D5,$D7,$78,$E5,$3F
	.dc.b	$D6,$EF,$F5,$27,$AA,$7F,$21,$7F,$E1,$FB,$FE,$B3,$1F,$CE,$4F,$C7
	.dc.b	$D7,$D5,$3F,$AC,$C7,$F4,$E7,$E1,$E1,$EC,$7C,$6F,$F3,$A8,$FE,$B5
	.dc.b	$3F,$5E,$3F,$92,$5F,$F1,$FC,$38,$9F,$68,$FF,$1C,$1F,$12,$3F,$81
	.dc.b	$5F,$F8,$A2,$33,$1F,$53,$0F,$F2,$AD,$7D,$43,$D7,$5F,$52,$3F,$D5
	.dc.b	$71,$FC,$0E,$3F,$AC,$CF,$F1,$E8,$F9,$91,$FE,$BA,$9F,$EA,$6F,$F8
	.dc.b	$36,$FF,$50,$7D,$4A,$FE,$A3,$5F,$A6,$3E,$74,$FF,$4B,$9F,$EB,$17
	.dc.b	$F9,$98,$FF,$15,$9F,$D5,$27,$E8,$9F,$34,$F8,$13,$FA,$28,$FF,$27
	.dc.b	$5F,$F1,$83,$FD,$87,$1F,$CC,$DF,$CC,$C7,$D0,$3E,$45,$FC,$F8,$FF
	.dc.b	$64,$BF,$C6,$63,$FC,$0C,$7F,$19,$8F,$D0,$DF,$D0,$8F,$E3,$18,$7E
	.dc.b	$87,$0F,$F0,$35,$FC,$D9,$FC,$01,$FE,$41,$8F,$F1,$C3,$F2,$70,$FE
	.dc.b	$2C,$7F,$7F,$8F,$C3,$C7,$F5,$07,$FA,$03,$FA,$83,$F0,$C7,$F1,$4A
	.dc.b	$F9,$D1,$FD,$F6,$3F,$C7,$8F,$CD,$3F,$A2,$BF,$36,$3F,$3B,$1F,$C3
	.dc.b	$E3,$FA,$4B,$F3,$0F,$EF,$2F,$E6,$A3,$F3,$71,$FD,$E2,$3F,$A1,$47
	.dc.b	$E6,$63,$F8,$54,$7F,$41,$8F,$CB,$C7,$F8,$C6,$BF,$2F,$1F,$D0,$23
	.dc.b	$FB,$84,$FC,$8C,$3F,$CD,$27,$D0,$C3,$F5,$F1,$FE,$B3,$9F,$E7,$E7
	.dc.b	$C5,$3F,$3C,$7D,$F7,$F3,$8F,$E5,$5F,$EE,$F7,$F9,$58,$F9,$31,$FE
	.dc.b	$49,$3F,$45,$5F,$D0,$30,$FF,$36,$C7,$FA,$CE,$B4,$FF,$93,$2F,$F8
	.dc.b	$8D,$FF,$99,$23,$31,$FE,$79,$BF,$F5,$83,$FD,$12,$3F,$D2,$11,$8A
	.dc.b	$4F,$F9,$B2,$BB,$DF,$EB,$AF,$F8,$23,$FE,$CC,$9F,$E8,$13,$CE,$3F
	.dc.b	$07,$3F,$CF,$5F,$F8,$03,$5D,$2F,$FC,$D9,$5F,$D3,$EB,$FD,$0A,$7B
	.dc.b	$E7,$FD,$9D,$7F,$C3,$EF,$DF,$1F,$CF,$27,$F2,$AF,$BE,$FF,$4B,$7A
	.dc.b	$BF,$E8,$B5,$FE,$A2,$7F,$BC,$DE,$6F,$B4,$7F,$10,$BF,$D2,$DF,$CC
	.dc.b	$3D,$D7,$FD,$55,$F7,$47,$F1,$0B,$D4,$7E,$02,$4F,$BA,$BF,$A4,$3F
	.dc.b	$5A,$3F,$42,$7E,$B2,$3F,$C5,$B3,$98,$E4,$7A,$CF,$F7,$24,$F1,$9F
	.dc.b	$5C,$7F,$74,$AF,$B3,$59,$BF,$BA,$7C,$D9,$FF,$32,$9F,$BA,$8F,$5B
	.dc.b	$FE,$D2,$7D,$6E,$6F,$EC,$1E,$B8,$F5,$57,$D6,$8F,$A3,$3F,$32,$BE
	.dc.b	$AC,$FF,$14,$AC,$C7,$D9,$9F,$98,$7D,$63,$EF,$3F,$E1,$19,$FE,$27
	.dc.b	$1E,$73,$FA,$74,$7F,$40,$7F,$DC,$EF,$F2,$A7,$37,$D1,$B7,$F4,$9F
	.dc.b	$D3,$3E,$C7,$FA,$C3,$FA,$D7,$ED,$3F,$5C,$F6,$BF,$C7,$9F,$F0,$BB
	.dc.b	$FC,$82,$BF,$C5,$AF,$94,$F8,$CF,$F8,$C2,$BF,$5F,$1E,$33,$94,$FE
	.dc.b	$BD,$FD,$85,$7F,$B5,$2F,$EB,$47,$FB,$5A,$BA,$3F,$DF,$0F,$A4,$7E
	.dc.b	$FE,$3D,$4F,$EA,$6B,$EA,$CE,$5F,$04,$7D,$B7,$E6,$BF,$54,$FD,$F1
	.dc.b	$F4,$67,$EA,$BF,$DE,$67,$C2,$FF,$09,$3E,$AD,$7D,$D3,$8B,$F4,$27
	.dc.b	$F5,$0F,$F9,$8A,$78,$BF,$C3,$EB,$E9,$3F,$E2,$73,$27,$FB,$11,$FF
	.dc.b	$15,$47,$D6,$7D,$30,$3F,$E0,$0B,$FA,$37,$E6,$9C,$8F,$EB,$87,$F6
	.dc.b	$2B,$FA,$57,$FC,$79,$FE,$37,$1F,$C7,$8E,$11,$FC,$75,$FC,$51,$FE
	.dc.b	$1B,$7C,$2F,$F9,$CA,$7F,$8C,$BF,$34,$F4,$DF,$E2,$5F,$9D,$7F,$44
	.dc.b	$CC,$FD,$08,$FD,$01,$F0,$23,$CE,$FF,$97,$5F,$E3,$0F,$E7,$8F,$08
	.dc.b	$7E,$14,$7F,$19,$8F,$E2,$A7,$3A,$FE,$6D,$3E,$A8,$FD,$2C,$72,$8F
	.dc.b	$F2,$E4,$7C,$B3,$9C,$FF,$22,$3F,$16,$E4,$FA,$97,$FE,$2A,$9F,$A8
	.dc.b	$7D,$89,$FD,$1C,$FF,$90,$67,$F4,$75,$FC,$62,$7F,$D8,$B5,$FA,$23
	.dc.b	$EB,$57,$F3,$B3,$F7,$C7,$F8,$28,$FE,$30,$59,$FA,$87,$F2,$03,$E1
	.dc.b	$BE,$A3,$FC,$C9,$3F,$CE,$AF,$D3,$18,$7F,$A9,$C7,$F1,$73,$E5,$BE
	.dc.b	$88,$FE,$96,$7F,$95,$CF,$A8,$7C,$A3,$F9,$C5,$7F,$2F,$7E,$A4,$FD
	.dc.b	$17,$FD,$7E,$7D,$19,$FD,$0D,$FF,$8A,$AB,$EA,$57,$E8,$23,$FC,$A7
	.dc.b	$5F,$26,$3F,$B6,$1F,$D8,$9F,$A0,$FF,$10,$7F,$8A,$BF,$1D,$FE,$7A
	.dc.b	$7E,$C0,$FE,$1E,$7C,$93,$FA,$29,$F1,$8F,$9D,$7F,$D1,$0E,$0F,$FA
	.dc.b	$12,$FF,$87,$D7,$F7,$B9,$FE,$F5,$5F,$9E,$AF,$CF,$C7,$F7,$8A,$F9
	.dc.b	$77,$F9,$D8,$FE,$F7,$5F,$D0,$2B,$F3,$B1,$FD,$CE,$3F,$85,$47,$F0
	.dc.b	$B8,$FF,$16,$47,$C8,$8F,$93,$1F,$DC,$23,$F8,$44,$7F,$09,$8F,$F1
	.dc.b	$64,$7F,$07,$8F,$E1,$11,$FE,$AF,$AF,$ED,$B5,$F9,$48,$FE,$D7,$4B
	.dc.b	$FE,$6B,$74,$FF,$71,$CF,$E5,$63,$F9,$B4,$7E,$5A,$3F,$83,$3F,$91
	.dc.b	$3F,$83,$3F,$CD,$67,$F9,$A4,$7F,$35,$8F,$8F,$3F,$E1,$68,$FE,$09
	.dc.b	$1F,$1E,$7F,$C3,$F5,$F1,$EB,$FB,$34,$7E,$3E,$BF,$21,$5F,$1E,$7F
	.dc.b	$CF,$35,$FD,$94,$FC,$9D,$7E,$3E,$3F,$B1,$CF,$C7,$7F,$1D,$5F,$1D
	.dc.b	$FF,$0A,$CF,$C7,$7F,$1F,$3F,$1E,$3F,$96,$CF,$F2,$E8,$F8,$F3,$F1
	.dc.b	$8F,$DC,$9F,$E4,$99,$FD,$CC,$FC,$58,$FC,$73,$F8,$C3,$E2,$BF,$E0
	.dc.b	$E8,$F8,$AF,$E2,$CF,$F2,$4B,$FB,$28,$F8,$B1,$FE,$3C,$BF,$C5,$C7
	.dc.b	$C6,$AF,$C5,$D7,$C5,$AF,$F4,$DC,$FE,$D2,$BF,$AA,$C7,$C6,$AF,$C5
	.dc.b	$BF,$E0,$69,$FC,$5B,$FE,$9A,$9F,$D9,$C7,$F9,$0A,$7F,$91,$1F,$E4
	.dc.b	$BA,$FF,$0F,$C7,$F2,$37,$F8,$4B,$FC,$92,$7F,$1A,$FC,$7A,$FF,$65
	.dc.b	$CF,$F8,$CC,$FF,$1D,$BF,$D5,$6F,$F8,$5D,$FF,$29,$AF,$C8,$46,$1F
	.dc.b	$F5,$91,$FC,$ED,$FF,$6C,$3F,$EB,$37,$4F,$F3,$F3,$8B,$A7,$FC,$93
	.dc.b	$86,$1F,$F0,$24,$75,$9F,$E7,$95,$E3,$8E,$B0,$C1,$7C,$6F,$FB,$64
	.dc.b	$C9,$F5,$C7,$94,$FF,$36,$9F,$ED,$6A,$ED,$87,$2C,$3B,$4F,$F8,$36
	.dc.b	$38,$CF,$C2,$AF,$F7,$7B,$63,$C5,$87,$86,$BF,$D1,$35,$FE,$F1,$AD
	.dc.b	$1F,$DF,$EF,$AC,$F9,$6F,$E0,$C6,$23,$E8,$1F,$E6,$C8,$E9,$87,$F3
	.dc.b	$C9,$D3,$F2,$EF,$D1,$5E,$D3,$E0,$DF,$D1,$3F,$B0,$4F,$D1,$7C,$51
	.dc.b	$B3,$FD,$3E,$FF,$72,$C3,$D5,$1F,$DD,$67,$C5,$7E,$0A,$FE,$BA,$7E
	.dc.b	$10,$F7,$D7,$99,$F8,$38,$5A,$FC,$05,$71,$7D,$F7,$E1,$3F,$DD,$D1
	.dc.b	$EC,$9F,$9B,$5F,$EC,$67,$E6,$C6,$5F,$C0,$1F,$10,$FF,$56,$CD,$7A
	.dc.b	$CF,$F6,$28,$FD,$14,$7C,$A7,$FC,$05,$3F,$29,$FC,$CC,$FE,$5B,$0F
	.dc.b	$BE,$FE,$54,$F8,$E7,$E6,$AB,$E7,$BD,$67,$FD,$8F,$1C,$1F,$C6,$DF
	.dc.b	$E6,$1F,$9C,$FE,$4F,$0F,$13,$EA,$C3,$C4,$FC,$69,$D9,$D2,$7E,$9E
	.dc.b	$1F,$83,$BF,$C1,$BF,$1A,$7F,$D8,$EF,$FA,$C8,$F2,$E1,$F8,$07,$F2
	.dc.b	$4D,$9E,$CF,$E6,$5D,$CF,$FA,$40,$FB,$E7,$A2,$3E,$3B,$4F,$A6,$59
	.dc.b	$F8,$0F,$E4,$0E,$B1,$F9,$03,$F1,$98,$7E,$76,$3F,$19,$5C,$1F,$C8
	.dc.b	$E1,$F1,$EB,$F1,$8F,$66,$DE,$E3,$F2,$91,$F5,$0E,$CF,$E0,$1F,$F4
	.dc.b	$FD,$6C,$F8,$D5,$FC,$11,$F8,$D8,$7E,$1E,$F5,$1E,$89,$FF,$32,$3F
	.dc.b	$11,$FF,$54,$E1,$F8,$77,$CD,$7F,$C6,$6B,$50,$C3,$FC,$73,$5B,$AF
	.dc.b	$34,$FF,$10,$3C,$C7,$F1,$E7,$F8,$DB,$A3,$F8,$C9,$FA,$C8,$FC,$2D
	.dc.b	$7C,$18,$FE,$31,$3F,$C8,$AF,$C3,$1F,$A4,$7E,$34,$FC,$08,$FD,$9B
	.dc.b	$E6,$BF,$D4,$BA,$9F,$81,$3F,$C4,$DE,$75,$FE,$6A,$BF,$E0,$13,$F0
	.dc.b	$0F,$F3,$75,$FF,$11,$C3,$F8,$04,$7E,$6C,$FC,$35,$7A,$DF,$E1,$A7
	.dc.b	$C0,$3F,$CE,$CF,$EA,$2F,$F8,$34,$6E,$BF,$88,$61,$E7,$AF,$E1,$78
	.dc.b	$7F,$09,$74,$7E,$95,$FC,$65,$FE,$95,$FE,$0D,$7C,$E3,$EF,$C7,$F0
	.dc.b	$9A,$FB,$F5,$99,$F2,$57,$F9,$59,$FE,$0C,$FB,$CF,$A6,$7E,$4E,$3F
	.dc.b	$01,$1A,$7D,$11,$F9,$E3,$F8,$E1,$ED,$AF,$DE,$BA,$BE,$4F,$F0,$6B
	.dc.b	$FB,$91,$F4,$A3,$F1,$47,$B4,$FE,$4D,$1E,$D7,$F8,$5B,$F7,$63,$ED
	.dc.b	$E1,$FC,$13,$0F,$A5,$87,$F5,$0B,$C4,$7D,$C7,$F7,$B1,$FC,$E5,$F7
	.dc.b	$1F,$5A,$FE,$83,$CA,$3F,$74,$F9,$6F,$FB,$D5,$7E,$19,$E7,$3F,$B2
	.dc.b	$9F,$AB,$1E,$E7,$FA,$9B,$CD,$FC,$4C,$7E,$69,$B1,$F4,$9F,$E8,$67
	.dc.b	$F2,$89,$F7,$1F,$69,$C3,$CD,$FD,$E3,$E5,$3F,$B4,$5F,$F5,$A8,$FC
	.dc.b	$D1,$FC,$B0,$F7,$CF,$FB,$1E,$FF,$49,$1F,$D9,$CF,$DB,$5F,$DC,$7C
	.dc.b	$87,$DB,$3F,$9A,$CF,$27,$F6,$C6,$27,$ED,$CF,$DF,$BF,$DA,$C7,$D9
	.dc.b	$8F,$F0,$44,$FE,$7E,$3D,$91,$FE,$0B,$7F,$68,$E6,$7E,$23,$EA,$3E
	.dc.b	$C4,$FE,$88,$F9,$C7,$D9,$8F,$D9,$C6,$5F,$AC,$7F,$29,$7F,$BC,$9F
	.dc.b	$CA,$27,$F5,$D5,$F2,$EB,$F9,$9C,$FD,$AA,$FE,$1A,$7A,$5F,$EE,$91
	.dc.b	$FA,$9A,$FE,$19,$3F,$D3,$A3,$F8,$2B,$FD,$D3,$0F,$EF,$0F,$F1,$68
	.dc.b	$F5,$57,$F9,$05,$F9,$D5,$F4,$9F,$15,$FF,$06,$7F,$67,$5C,$63,$F1
	.dc.b	$F1,$FD,$B8,$FD,$64,$FD,$A8,$E3,$1F,$8D,$3F,$95,$BF,$8D,$8F,$54
	.dc.b	$7F,$68,$9F,$C1,$C7,$D2,$AF,$EC,$CF,$D2,$9F,$D4,$E1,$FE,$24,$9F
	.dc.b	$4C,$7E,$68,$FC,$5B,$FE,$0A,$3F,$4F,$3F,$3E,$3F,$51,$1F,$C1,$4F
	.dc.b	$F0,$A1,$F8,$B3,$FA,$FB,$FA,$88,$FA,$67,$D7,$BE,$27,$E9,$6F,$D3
	.dc.b	$3F,$A0,$3E,$95,$FF,$5B,$8F,$F1,$15,$FF,$1D,$7E,$84,$FC,$83,$F8
	.dc.b	$D9,$F4,$27,$F8,$FD,$FF,$4D,$8F,$DF,$1E,$7A,$FF,$40,$3E,$77,$FA
	.dc.b	$BB,$FE,$1A,$8F,$A0,$7E,$81,$FF,$0E,$D7,$E8,$0F,$03,$FE,$44,$3F
	.dc.b	$39,$1F,$D4,$CF,$C4,$3F,$E6,$39,$FD,$04,$FF,$17,$3E,$54,$7E,$7E
	.dc.b	$7F,$8A,$9F,$36,$7F,$3D,$5F,$56,$FB,$9F,$E5,$9A,$F9,$91,$FE,$53
	.dc.b	$AF,$99,$5F,$9D,$8F,$F2,$FD,$FC,$CA,$FE,$A8,$7C,$CC,$3F,$A4,$47
	.dc.b	$CC,$8F,$F2,$73,$F3,$0F,$90,$FC,$B8,$FF,$03,$5F,$CB,$BF,$E1,$55
	.dc.b	$F9,$93,$F9,$FD,$7E,$62,$FE,$54,$7F,$91,$27,$F8,$44,$7F,$90,$23
	.dc.b	$D7,$5F,$CF,$CF,$CB,$57,$E6,$5F,$CB,$5F,$FA,$11,$FE,$42,$FF,$8B
	.dc.b	$EF,$F2,$F3,$FE,$35,$3E,$55,$FF,$78,$AF,$95,$1F,$DE,$AF,$F8,$39
	.dc.b	$FD,$D1,$F9,$13,$F2,$63,$FC,$F9,$5F,$26,$FF,$B8,$C7,$D0,$3F,$9B
	.dc.b	$9F,$C5,$30,$FF,$4D,$61,$FD,$C3,$0F,$93,$87,$F3,$57,$FA,$89,$FE
	.dc.b	$96,$BF,$F2,$33,$FC,$CC,$FF,$68,$61,$FE,$0B,$7F,$C1,$37,$C7,$0F
	.dc.b	$F2,$54,$7F,$76,$AE,$33,$FE,$C2,$6E,$1A,$9F,$E0,$87,$F5,$08,$DD
	.dc.b	$72,$9F,$F1,$44,$FF,$A3,$8F,$F1,$E4,$FF,$96,$63,$F2,$F0,$3F,$D2
	.dc.b	$77,$E5,$85,$7F,$AA,$6B,$83,$FA,$2A,$FF,$5D,$1F,$EA,$1A,$FF,$15
	.dc.b	$CF,$D7,$8F,$E6,$47,$FA,$DC,$FE,$20,$73,$38,$1C,$E3,$F9,$DB,$E3
	.dc.b	$3F,$DF,$17,$F0,$5E,$CF,$48,$F8,$31,$FE,$6A,$8F,$D4,$57,$CA,$3E
	.dc.b	$09,$EA,$8F,$82,$7F,$AD,$DF,$82,$FF,$B5,$67,$EA,$47,$C0,$76,$7F
	.dc.b	$AC,$5F,$D6,$47,$94,$FB,$47,$C0,$3F,$C4,$E7,$C0,$3F,$DC,$35,$F8
	.dc.b	$27,$E0,$4F,$FA,$BD,$F9,$47,$C0,$3F,$DB,$A7,$81,$FE,$2F,$5C,$CF
	.dc.b	$F1,$64,$7F,$84,$A7,$EF,$D7,$EB,$63,$EC,$3F,$1E,$BF,$D9,$17,$EE
	.dc.b	$3F,$AD,$4F,$F8,$CA,$7C,$73,$EB,$8F,$F3,$84,$7D,$67,$F2,$D1,$ED
	.dc.b	$3F,$9F,$BF,$44,$F1,$BF,$5A,$BF,$01,$3F,$9A,$8F,$64,$FD,$A3,$D8
	.dc.b	$7C,$97,$FC,$C6,$FD,$C7,$BD,$F9,$23,$4F,$E3,$E7,$ED,$9E,$D9,$F4
	.dc.b	$3F,$68,$F6,$BF,$5D,$E5,$3F,$EE,$87,$F5,$0F,$ED,$8F,$64,$7F,$44
	.dc.b	$7E,$C1,$F3,$8F,$61,$F7,$23,$F6,$13,$C4,$F1,$47,$C9,$7E,$B4,$FE
	.dc.b	$11,$F0,$C7,$F8,$2A,$7E,$B3,$F5,$E3,$F9,$25,$FE,$DE,$7E,$EB,$F9
	.dc.b	$18,$FE,$51,$1F,$2E,$7F,$86,$D7,$A1,$FF,$6F,$56,$63,$EB,$C7,$80
	.dc.b	$FF,$2E,$CF,$F4,$B7,$E6,$9F,$48,$FE,$23,$3F,$5E,$BE,$D9,$C6,$BF
	.dc.b	$B3,$4F,$FA,$2C,$FD,$2D,$7C,$07,$F6,$CF,$D1,$3F,$2A,$7A,$6F,$F9
	.dc.b	$5B,$F4,$23,$F4,$E7,$13,$F5,$57,$F5,$2B,$27,$96,$3F,$D3,$55,$FC
	.dc.b	$89,$F9,$86,$4F,$E3,$17,$F4,$1F,$F3,$33,$F0,$5F,$49,$FE,$65,$BF
	.dc.b	$E6,$B7,$FC,$0E,$7E,$8C,$7F,$55,$BF,$D4,$9D,$67,$F7,$EF,$EF,$EF
	.dc.b	$F8,$E5,$F8,$63,$F9,$E3,$FC,$A5,$FD,$FB,$FC,$71,$D3,$FC,$76,$FE
	.dc.b	$C9,$F0,$5F,$F6,$2D,$FF,$7F,$9F,$D0,$9D,$67,$F4,$11,$FD,$30,$F4
	.dc.b	$57,$F5,$6B,$FE,$3D,$1F,$40,$FC,$01,$FC,$5A,$FD,$07,$F1,$43,$E4
	.dc.b	$1E,$83,$FC,$A9,$3F,$C5,$6F,$31,$FC,$52,$FF,$3C,$F0,$7E,$74,$FF
	.dc.b	$54,$9C,$57,$E6,$6B,$F8,$89,$F2,$5F,$E8,$37,$FB,$49,$FE,$1C,$FE
	.dc.b	$16,$7F,$89,$4F,$E7,$5F,$CC,$9F,$22,$3F,$38,$FF,$48,$9F,$97,$3F
	.dc.b	$97,$8F,$CD,$DF,$F8,$9E,$BF,$35,$5F,$2A,$3F,$A1,$D7,$F8,$90,$F9
	.dc.b	$47,$D1,$3F,$31,$1F,$9C,$3F,$18,$FC,$9A,$FE,$0C,$7D,$73,$E4,$1F
	.dc.b	$7A,$3F,$9D,$1F,$22,$3F,$82,$C7,$E5,$A7,$F2,$75,$FC,$22,$BF,$82
	.dc.b	$1F,$22,$BF,$81,$DF,$F6,$CB,$F9,$15,$FC,$0E,$BF,$27,$7F,$C0,$AF
	.dc.b	$F9,$A5,$7F,$35,$8F,$ED,$75,$F1,$AF,$FB,$85,$FC,$6B,$FE,$01,$7F
	.dc.b	$C0,$6F,$E3,$47,$E3,$E3,$F2,$11,$F1,$AB,$FC,$43,$1F,$BE,$3E,$1C
	.dc.b	$FF,$7A,$3E,$D3,$FD,$85,$FB,$55,$FC,$BD,$FD,$F4,$7F,$88,$CF,$C5
	.dc.b	$DF,$EF,$5F,$F1,$34,$FC,$79,$FD,$BC,$7C,$7A,$FF,$0E,$C7,$EE,$A7
	.dc.b	$E3,$DF,$EF,$23,$F1,$D5,$FC,$A6,$BF,$C2,$77,$F1,$E3,$F7,$55,$F1
	.dc.b	$E7,$F6,$F1,$F1,$DF,$F0,$A4,$FC,$78,$FD,$B4,$FF,$85,$2B,$F1,$91
	.dc.b	$F1,$EB,$FC,$A1,$3F,$E0,$F3,$F9,$2D,$FE,$D6,$BF,$63,$5F,$1E,$7F
	.dc.b	$1D,$7F,$B4,$AF,$C6,$47,$E4,$6B,$E2,$DF,$F8,$0E,$FE,$2C,$7C,$47
	.dc.b	$E2,$CF,$F9,$0E,$3E,$2C,$7C,$68,$FC,$5C,$7C,$58,$F8,$D1,$F1,$63
	.dc.b	$F1,$71,$FD,$F4,$FC,$5B,$FC,$75,$F8,$B3,$FE,$3A,$8F,$D8,$47,$C5
	.dc.b	$3F,$C6,$77,$F1,$6F,$FB,$F9,$F8,$BA,$F8,$73,$FB,$07,$E2,$CF,$EC
	.dc.b	$23,$F1,$71,$F2,$2B,$E2,$C7,$EC,$4F,$C5,$CF,$C5,$7F,$69,$5F,$8E
	.dc.b	$9F,$EE,$A7,$F9,$43,$8B,$FB,$A6,$1F,$E1,$56,$D7,$F9,$5E,$D8,$A7
	.dc.b	$32,$9E,$4F,$F8,$A8,$FE,$6C,$6C,$C1,$FE,$5E,$9C,$3E,$0B,$E2,$FF
	.dc.b	$7D,$17,$FE,$55,$BF,$1B,$DE,$B8,$CF,$F8,$EA,$3C,$55,$88,$FF,$2E
	.dc.b	$47,$F5,$8C,$3F,$CE,$17,$CA,$BD,$6F,$F8,$4E,$BF,$C6,$6F,$C2,$C3
	.dc.b	$FC,$B5,$1C,$CF,$EC,$2F,$F9,$D8,$E2,$FF,$92,$1C,$4E,$E3,$D3,$1F
	.dc.b	$E5,$D7,$17,$E6,$9E,$31,$FE,$E6,$BF,$F4,$24,$7C,$28,$FF,$30,$4A
	.dc.b	$7F,$CC,$57,$F5,$9F,$A9,$87,$47,$EB,$5D,$BE,$05,$7B,$BF,$EA,$4B
	.dc.b	$FF,$47,$C7,$17,$E7,$47,$F9,$8A,$7D,$35,$9C,$3C,$93,$F3,$27,$EA
	.dc.b	$9F,$31,$E1,$5F,$77,$0F,$99,$87,$63,$E0,$9F,$31,$FC,$03,$F6,$DB
	.dc.b	$7D,$AB,$FC,$1F,$3F,$26,$76,$FA,$A3,$EC,$47,$D9,$7E,$FB,$5F,$C1
	.dc.b	$B6,$FC,$D1,$F7,$DF,$C6,$C6,$E7,$C1,$3F,$88,$36,$7E,$6E,$FE,$E3
	.dc.b	$F8,$28,$AF,$C2,$2C,$7D,$72,$C7,$48,$FF,$62,$1C,$AF,$FC,$49,$5F
	.dc.b	$91,$7F,$12,$D9,$FB,$2D,$BF,$3D,$7F,$9A,$6B,$E6,$4F,$AE,$3E,$C9
	.dc.b	$F8,$B3,$FC,$6E,$E6,$3C,$53,$FB,$48,$E0,$F4,$9E,$15,$AB,$E7,$1F
	.dc.b	$87,$34,$7E,$26,$F5,$5F,$1A,$3F,$D1,$2E,$9F,$F4,$5D,$7C,$13,$F0
	.dc.b	$0F,$FA,$05,$F8,$2D,$7E,$BB,$63,$A3,$FE,$60,$8F,$99,$7F,$8C,$7F
	.dc.b	$32,$FD,$97,$F5,$B1,$FA,$63,$F8,$7D,$7E,$1D,$FF,$2C,$9F,$A6,$9F
	.dc.b	$E3,$35,$DA,$3F,$87,$9F,$B2,$C3,$EC,$DF,$4C,$3E,$1C,$FE,$15,$FC
	.dc.b	$C3,$F8,$5C,$3F,$13,$1F,$C3,$E3,$84,$F8,$5F,$DA,$BA,$9F,$87,$1F
	.dc.b	$E9,$B7,$F8,$65,$7F,$06,$7F,$07,$1F,$C3,$67,$F0,$6D,$7A,$3F,$E9
	.dc.b	$0C,$3F,$82,$3E,$63,$C5,$3E,$1B,$FE,$06,$F9,$B0,$FC,$14,$FE,$FC
	.dc.b	$FD,$C1,$FE,$B5,$7F,$98,$DE,$EB,$A4,$FF,$00,$7E,$0B,$FE,$B3,$9F
	.dc.b	$E3,$F3,$FC,$2D,$F9,$93,$FA,$38,$F3,$9F,$98,$9F,$F7,$7B,$F0,$9F
	.dc.b	$E3,$EF,$C2,$BF,$E0,$6F,$F0,$83,$E0,$C7,$FB,$05,$FE,$33,$87,$EC
	.dc.b	$A3,$CA,$7F,$28,$9F,$E0,$58,$7E,$C6,$3D,$E7,$F0,$1B,$E1,$79,$78
	.dc.b	$C7,$E0,$9F,$DD,$CE,$70,$FC,$34,$FF,$4B,$9F,$C2,$DF,$ED,$E3,$35
	.dc.b	$F9,$C7,$FD,$FC,$7F,$54,$7F,$81,$1F,$C6,$DF,$DD,$5F,$E7,$1A,$FE
	.dc.b	$D9,$F5,$DF,$F1,$C7,$EA,$1F,$A6,$3F,$7C,$F9,$0F,$F5,$25,$FF,$16
	.dc.b	$7F,$A5,$CD,$8E,$67,$F0,$49,$FB,$CF,$37,$FA,$24,$F3,$8F,$24,$7F
	.dc.b	$55,$7F,$5E,$FE,$DC,$F8,$73,$EF,$47,$F9,$5D,$E8,$D9,$DC,$74,$8D
	.dc.b	$9F,$BD,$7E,$B3,$D8,$FF,$01,$1F,$90,$9D,$D7,$E6,$E3,$F7,$47,$F8
	.dc.b	$32,$BF,$1E,$FE,$E6,$FF,$A2,$57,$38,$FF,$03,$1A,$7F,$32,$FF,$82
	.dc.b	$6B,$F7,$2F,$F3,$B9,$FF,$03,$61,$FC,$DF,$0E,$53,$FE,$0C,$AF,$CC
	.dc.b	$36,$39,$B6,$3F,$7D,$1F,$98,$8F,$EF,$71,$F9,$8A,$FE,$71,$3F,$B3
	.dc.b	$9F,$76,$1F,$DF,$27,$F7,$75,$98,$FD,$59,$ED,$9F,$B3,$5F,$B5,$73
	.dc.b	$87,$89,$FC,$B5,$7E,$E6,$B3,$5F,$B2,$3D,$73,$FD,$B9,$F9,$C7,$EC
	.dc.b	$5F,$DA,$BF,$69,$D4,$7E,$F2,$3D,$51,$FD,$BC,$FD,$91,$FD,$04,$FF
	.dc.b	$68,$BF,$B5,$8F,$D4,$C7,$E5,$AF,$F9,$C1,$F5,$9C,$DF,$E5,$5F,$F1
	.dc.b	$95,$7F,$6C,$AC,$9D,$63,$FC,$7B,$1F,$AC,$3F,$80,$DF,$E4,$E3,$F1
	.dc.b	$A7,$EC,$63,$ED,$4F,$ED,$9F,$C9,$46,$A7,$F8,$A3,$FD,$A4,$FD,$61
	.dc.b	$E3,$39,$9F,$CA,$A3,$F5,$CF,$DD,$3B,$CF,$B5,$FD,$BC,$7F,$6B,$7F
	.dc.b	$58,$7A,$4F,$F0,$34,$FF,$6A,$7F,$6D,$39,$3F,$98,$CF,$F8,$40,$FC
	.dc.b	$7C,$7F,$63,$39,$3F,$D1,$9F,$E1,$EF,$F0,$D6,$BE,$E7,$F5,$CF,$F2
	.dc.b	$03,$FB,$3B,$FA,$D3,$F2,$07,$F6,$87,$31,$FA,$C8,$FE,$8A,$7F,$8E
	.dc.b	$2F,$FC,$E6,$FE,$AE,$BF,$69,$7F,$87,$8F,$E5,$55,$FA,$FB,$FE,$AF
	.dc.b	$5F,$3A,$7F,$47,$1F,$BD,$8F,$C8,$47,$F6,$38,$E5,$5F,$E7,$CA,$FE
	.dc.b	$97,$5F,$90,$AF,$F0,$BD,$7F,$97,$DF,$D2,$9F,$A1,$3F,$C6,$91,$FA
	.dc.b	$97,$FC,$C5,$87,$F2,$39,$FD,$44,$7F,$97,$63,$FC,$BD,$1F,$D6,$63
	.dc.b	$FC,$C8,$FE,$9C,$FE,$FA,$7D,$27,$EA,$CF,$F6,$D8,$FA,$51,$FD,$69
	.dc.b	$FC,$23,$FA,$D7,$FB,$BB,$FE,$15,$3F,$8A,$E1,$FD,$5A,$3F,$4D,$1F
	.dc.b	$A7,$C3,$EB,$4F,$E9,$63,$FC,$9A,$7E,$24,$FE,$DE,$FF,$1E,$C3,$F4
	.dc.b	$B3,$FD,$52,$7F,$A6,$9F,$9F,$3F,$21,$7E,$67,$F4,$17,$E6,$7F,$D1
	.dc.b	$35,$A8,$FE,$0B,$1F,$EB,$A3,$F4,$67,$E5,$4F,$BC,$7F,$A3,$A3,$F4
	.dc.b	$E6,$63,$F4,$11,$FD,$44,$FC,$E5,$70,$9F,$E5,$D1,$FD,$49,$CC,$FF
	.dc.b	$91,$CF,$EF,$67,$EE,$E3,$FC,$85,$59,$9F,$E9,$67,$E8,$A7,$CF,$3F
	.dc.b	$E4,$7B,$FF,$6A,$BF,$99,$AF,$E9,$0F,$E3,$0F,$F0,$19,$FE,$AE,$3F
	.dc.b	$96,$1F,$9D,$3F,$D0,$07,$E7,$70,$FF,$21,$47,$F4,$6C,$3E,$41,$FE
	.dc.b	$96,$8F,$E8,$D7,$FD,$02,$BF,$2E,$FE,$4E,$7F,$89,$5F,$F1,$CC,$33
	.dc.b	$3F,$E3,$A7,$FB,$BC,$7F,$8E,$AB,$F3,$78,$7F,$8E,$67,$FA,$FC,$7F
	.dc.b	$8B,$CF,$0B,$E8,$7F,$D7,$95,$F9,$CA,$FF,$63,$5F,$86,$BF,$0F,$5F
	.dc.b	$EB,$67,$FA,$1E,$1F,$C3,$6F,$EF,$D7,$FB,$12,$BF,$9F,$47,$A4,$FC
	.dc.b	$43,$FD,$08,$FF,$1A,$3F,$9C,$7F,$C5,$27,$F2,$73,$FC,$43,$87,$F7
	.dc.b	$CB,$FE,$83,$1F,$C4,$CF,$F0,$36,$1F,$9D,$3F,$A1,$5F,$F8,$C2,$FF
	.dc.b	$A0,$4F,$F4,$18,$FC,$E3,$D2,$3F,$9F,$9D,$DF,$F0,$2B,$F9,$4B,$FE
	.dc.b	$45,$1F,$D4,$AF,$F3,$93,$FE,$2D,$8F,$CD,$D7,$F3,$AA,$FF,$18,$57
	.dc.b	$E6,$A3,$F9,$ED,$7E,$6A,$FF,$C9,$17,$F9,$AB,$FE,$71,$1F,$E4,$BB
	.dc.b	$FC,$CD,$7C,$BA,$FF,$0F,$9F,$95,$AF,$CC,$3F,$3E,$7F,$31,$3F,$E2
	.dc.b	$03,$FC,$23,$1F,$CC,$CF,$C4,$1F,$80,$3F,$2C,$7E,$40,$FC,$01,$FE
	.dc.b	$63,$7F,$D2,$CF,$E5,$EB,$FC,$A2,$FF,$96,$EB,$F3,$15,$FC,$CF,$0F
	.dc.b	$CB,$DF,$FA,$3C,$FE,$64,$7E,$5E,$7E,$FD,$7F,$9F,$2B,$F2,$F5,$F1
	.dc.b	$2B,$F9,$84,$AF,$FC,$2F,$36,$3F,$D1,$97,$FC,$C6,$7F,$D2,$B5,$F9
	.dc.b	$FC,$3F,$1F,$87,$FA,$C2,$FF,$99,$57,$E6,$2F,$FC,$31,$1F,$CC,$B0
	.dc.b	$FC,$B6,$1F,$CC,$8F,$F0,$B9,$F9,$53,$FA,$39,$F8,$08,$FF,$02,$BF
	.dc.b	$D6,$1F,$86,$7E,$48,$FE,$54,$7F,$1A,$3F,$81,$9F,$E5,$FA,$FF,$0A
	.dc.b	$BC,$30,$FF,$42,$9F,$8F,$7F,$C8,$11,$FC,$C5,$FC,$B3,$FE,$50,$9F
	.dc.b	$F0,$A4,$7F,$31,$9F,$F2,$55,$FF,$85,$F0,$FF,$09,$1F,$94,$AF,$F0
	.dc.b	$5D,$7E,$52,$FF,$DF,$11,$FE,$0E,$9F,$CA,$47,$F9,$22,$3F,$96,$4F
	.dc.b	$F2,$DB,$FF,$3F,$5F,$E5,$23,$F9,$64,$7F,$5F,$9F,$CA,$47,$E5,$63
	.dc.b	$F8,$04,$FE,$52,$BF,$DD,$97,$F9,$4B,$FE,$A3,$7F,$E9,$AB,$FF,$10
	.dc.b	$D7,$E3,$1F,$F3,$9E,$1F,$E0,$AC,$3F,$33,$87,$F8,$1F,$0F,$EB,$F5
	.dc.b	$FC,$33,$0F,$BD,$87,$15,$1F,$9E,$3F,$9A,$1F,$9D,$3E,$C3,$FD,$5A
	.dc.b	$7F,$96,$3F,$97,$BF,$F6,$E1,$F9,$69,$FF,$03,$47,$E6,$E7,$F9,$63
	.dc.b	$F8,$E3,$FC,$DA,$7F,$23,$3F,$C1,$B7,$F8,$07,$FD,$F1,$1F,$95,$3F
	.dc.b	$95,$3F,$06,$3C,$F5,$FC,$CE,$B8,$9F,$CC,$8F,$3C,$FF,$8B,$AF,$E3
	.dc.b	$5F,$F8,$1E,$4F,$F8,$7A,$7F,$C0,$D5,$F9,$BB,$FE,$5F,$3F,$CA,$E7
	.dc.b	$F2,$D5,$FE,$DD,$AF,$E5,$51,$FE,$05,$BF,$CA,$BF,$2C,$FF,$02,$CF
	.dc.b	$F2,$98,$FC,$A1,$FE,$05,$7F,$94,$C7,$F9,$D0,$FC,$A4,$7E,$92,$7F
	.dc.b	$7F,$1F,$2A,$31,$5F,$CD,$E1,$3F,$E3,$97,$FC,$BE,$7F,$B5,$8F,$D2
	.dc.b	$1F,$E6,$59,$FD,$7D,$FC,$D8,$FF,$62,$BF,$E6,$57,$FC,$73,$1F,$CB
	.dc.b	$1F,$CD,$3F,$EF,$59,$FC,$D4,$7F,$B4,$67,$F9,$6C,$FE,$EA,$BF,$C6
	.dc.b	$53,$FD,$5E,$FF,$37,$1F,$CB,$63,$F9,$74,$7F,$08,$8F,$CD,$C7,$FB
	.dc.b	$30,$FF,$02,$61,$F9,$98,$FF,$2F,$C7,$F9,$91,$FC,$C9,$FD,$FD,$FE
	.dc.b	$58,$7E,$8C,$FE,$D6,$7F,$A1,$27,$FD,$C2,$7E,$66,$3F,$CB,$E7,$F6
	.dc.b	$CC,$3F,$04,$7E,$06,$7F,$D8,$D1,$FC,$C7,$0F,$E5,$B8,$7E,$F4,$FF
	.dc.b	$01,$4F,$F2,$B3,$FC,$01,$5F,$E0,$37,$F3,$27,$F8,$08,$FC,$89,$F3
	.dc.b	$AF,$F9,$69,$FE,$CC,$3F,$C0,$47,$E6,$2B,$FC,$03,$5F,$2E,$FF,$C0
	.dc.b	$27,$E5,$CF,$35,$7F,$B1,$6F,$F2,$CF,$F8,$03,$0F,$F5,$FB,$FE,$C0
	.dc.b	$C3,$F2,$D5,$FE,$5A,$BF,$E5,$F3,$F7,$CF,$EA,$71,$FA,$28,$FF,$00
	.dc.b	$47,$F2,$D3,$FC,$B9,$3F,$CC,$CF,$F2,$A3,$F3,$30,$CD,$7F,$2C,$8F
	.dc.b	$F7,$A1,$FE,$57,$C3,$F9,$63,$FE,$BE,$AF,$EB,$B7,$F9,$28,$FC,$B1
	.dc.b	$FE,$59,$9F,$EF,$EF,$F2,$D9,$FE,$B9,$3F,$93,$3F,$BF,$BF,$CB,$2B
	.dc.b	$FD,$6B,$3F,$EB,$68,$FE,$59,$3F,$26,$3F,$BF,$C7,$E4,$A3,$F9,$5C
	.dc.b	$7E,$49,$F9,$D3,$FD,$FC,$FF,$58,$C7,$E4,$A3,$F7,$55,$FC,$AA,$3F
	.dc.b	$25,$87,$EE,$11,$F8,$E3,$F9,$59,$F9,$3B,$FE,$55,$3F,$DF,$A3,$F1
	.dc.b	$07,$F9,$76,$3F,$0A,$FE,$EA,$FF,$28,$7F,$80,$27,$F9,$5B,$FE,$B0
	.dc.b	$7F,$B5,$47,$E7,$1F,$F5,$83,$FE,$E2,$C3,$FC,$AE,$FF,$25,$3F,$96
	.dc.b	$C7,$E5,$4F,$F0,$03,$99,$FE,$58,$79,$1F,$59,$FD,$0A,$7D,$6F,$EA
	.dc.b	$DF,$E5,$F7,$FE,$B4,$C3,$FC,$B7,$1F,$98,$9F,$C5,$D7,$E6,$0F,$F0
	.dc.b	$1D,$FF,$AC,$DF,$F0,$19,$FE,$76,$3D,$31,$FE,$02,$AF,$CA,$BF,$C6
	.dc.b	$0F,$F1,$71,$FE,$FE,$3F,$22,$7E,$64,$FF,$75,$BF,$98,$7F,$AD,$61
	.dc.b	$FC,$BD,$FF,$03,$5F,$FA,$CA,$7F,$C0,$55,$F9,$69,$FC,$BD,$7F,$80
	.dc.b	$A3,$E1,$47,$F9,$62,$BF,$C0,$73,$F9,$88,$FF,$1C,$C7,$FB,$1D,$FA
	.dc.b	$07,$F2,$E7,$CE,$FC,$A7,$F9,$73,$FE,$B3,$8F,$CD,$BC,$9F,$9F,$5F
	.dc.b	$9C,$3F,$97,$57,$E7,$1F,$F2,$D4,$7E,$72,$7E,$69,$F9,$C8,$FE,$0D
	.dc.b	$7F,$E0,$78,$FA,$37,$FE,$07,$9F,$EA,$71,$FE,$56,$3F,$38,$7F,$27
	.dc.b	$3F,$D6,$07,$F9,$64,$FC,$99,$FE,$02,$BF,$E5,$CF,$FA,$BB,$0F,$F5
	.dc.b	$7C,$7C,$68,$FF,$03,$9F,$EC,$83,$F9,$81,$C6,$3E,$43,$FC,$CE,$FF
	.dc.b	$97,$CF,$F9,$6C,$FC,$E1,$FE,$B8,$7F,$07,$1F,$24,$FF,$03,$1F,$EB
	.dc.b	$09,$F0,$61,$FD,$6E,$3F,$3D,$87,$EC,$B0,$FF,$2D,$D7,$C7,$9F,$CC
	.dc.b	$57,$F2,$FB,$FF,$03,$1F,$E5,$AA,$FF,$59,$E0,$BF,$EA,$91,$FE,$06
	.dc.b	$8F,$CC,$C7,$81,$FF,$0E,$1F,$B2,$BF,$8C,$FC,$87,$FD,$5E,$7F,$96
	.dc.b	$AB,$FC,$0B,$5F,$9C,$C3,$FC,$BB,$5F,$CB,$EB,$F9,$71,$FE,$06,$9F
	.dc.b	$E6,$B8,$7F,$2F,$AF,$CC,$4F,$F9,$7E,$BF,$97,$57,$F8,$0A,$3F,$97
	.dc.b	$47,$D3,$7F,$C0,$C7,$FA,$3A,$FF,$C3,$B5,$FE,$5B,$AF,$E6,$13,$FE
	.dc.b	$5B,$3F,$36,$7F,$C0,$13,$F9,$EC,$7E,$B5,$FA,$55,$FE,$AF,$7F,$30
	.dc.b	$7F,$AB,$DF,$CB,$3F,$E5,$C9,$FF,$06,$DF,$FC,$00,$9F,$F0,$19,$FE
	.dc.b	$2E,$9F,$F5,$83,$FE,$03,$9F,$F2,$EC,$FD,$09,$FE,$AB,$3F,$EB,$07
	.dc.b	$F2,$E7,$E3,$0F,$CB,$D7,$F9,$6F,$0F,$E6,$51,$FC,$C6,$BF,$56,$FF
	.dc.b	$31,$C3,$EE,$61,$FC,$11,$FE,$64,$FF,$06,$C3,$FC,$06,$7F,$96,$9F
	.dc.b	$F3,$54,$7F,$6E,$9F,$F2,$6C,$FE,$5B,$0F,$F0,$1C,$7C,$88,$FB,$4F
	.dc.b	$FC,$01,$3F,$2D,$5F,$E7,$39,$FE,$07,$1F,$CA,$CF,$F0,$24,$FF,$30
	.dc.b	$9D,$5F,$E5,$70,$FE,$59,$5F,$E5,$D8,$FF,$2D,$57,$AA,$3F,$96,$BF
	.dc.b	$EA,$F7,$F4,$07,$FA,$0D,$F4,$47,$F9,$71,$FE,$5F,$7F,$E5,$AA,$FF
	.dc.b	$02,$DF,$F2,$EB,$FC,$B5,$7F,$81,$5F,$F0,$2D,$7E,$5A,$3F,$19,$1F
	.dc.b	$D0,$8F,$EB,$93,$F9,$6C,$3E,$0C,$7F,$4F,$8F,$F2,$E1,$FC,$B8,$FC
	.dc.b	$B3,$FC,$B9,$FF,$2C,$C7,$F8,$11,$FF,$2D,$D7,$E5,$67,$F2,$67,$E5
	.dc.b	$4F,$E8,$0F,$CA,$4F,$F2,$C3,$F2,$CF,$F9,$68,$F8,$11,$FE,$07,$7F
	.dc.b	$2F,$87,$F8,$1E,$3F,$2B,$3F,$22,$7F,$CB,$53,$FC,$1D,$FE,$C2,$7F
	.dc.b	$7C,$BF,$CC,$9F,$E6,$A3,$FC,$E6,$7F,$01,$7E,$74,$FF,$2F,$7F,$33
	.dc.b	$1F,$E0,$A7,$FD,$65,$87,$F9,$6E,$FF,$D0,$95,$F9,$97,$E3,$4F,$F8
	.dc.b	$1C,$FF,$2D,$9F,$CA,$AF,$FD,$63,$3E,$78,$FF,$04,$5F,$E6,$0F,$E4
	.dc.b	$78,$7F,$2C,$3F,$C0,$E7,$D4,$8F,$EF,$75,$FC,$11,$FE,$65,$3F,$E1
	.dc.b	$0A,$FF,$03,$C7,$E6,$4F,$E5,$AF,$F8,$72,$BF,$C1,$13,$89,$FC,$CE
	.dc.b	$1F,$1D,$EB,$5F,$85,$6D,$FF,$97,$2F,$F9,$74,$7E,$6E,$FF,$34,$FF
	.dc.b	$7B,$7F,$D3,$0F,$F9,$D9,$FF,$2F,$1F,$E9,$CB,$FE,$60,$FE,$6A,$3F
	.dc.b	$CB,$D8,$7E,$65,$FF,$2F,$3F,$E0,$8A,$FE,$5F,$1F,$9A,$BF,$F4,$6B
	.dc.b	$FD,$7E,$7F,$96,$BF,$E5,$C9,$F8,$CF,$E0,$27,$FC,$1A,$7F,$AB,$5F
	.dc.b	$E7,$B3,$F9,$A3,$F9,$84,$7F,$81,$DF,$CC,$C7,$F9,$75,$FF,$5A,$D7
	.dc.b	$48,$FC,$A1,$F9,$B3,$FC,$0F,$5F,$E5,$E9,$FF,$31,$61,$FE,$F0,$7F
	.dc.b	$D1,$D3,$FC,$B8,$FF,$58,$9F,$E0,$79,$FF,$58,$BF,$E3,$FB,$FF,$61
	.dc.b	$57,$9A,$3F,$C0,$A7,$F4,$3B,$FE,$61,$3F,$E6,$18,$FF,$02,$9F,$E6
	.dc.b	$13,$F8,$5C,$7C,$67,$F0,$E7,$F8,$74,$A7,$D3,$BF,$F3,$1C,$7F,$AD
	.dc.b	$1F,$F6,$F3,$FE,$B4,$3F,$BA,$C7,$E0,$CF,$E6,$97,$F8,$07,$FC,$C0
	.dc.b	$FF,$32,$7F,$CC,$D1,$F9,$B9,$FE,$62,$FF,$98,$70,$FC,$EB,$FC,$C6
	.dc.b	$6B,$CD,$FF,$6F,$BF,$3E,$FF,$82,$C7,$F9,$82,$7F,$31,$7F,$E6,$08
	.dc.b	$FF,$2D,$1F,$A1,$9F,$F5,$9B,$FA,$D3,$F3,$31,$FE,$06,$AF,$D1,$CF
	.dc.b	$F3,$2A,$FE,$AD,$5F,$EC,$D9,$FC,$BD,$7F,$97,$4F,$F5,$9D,$FF,$82
	.dc.b	$63,$CD,$3F,$E5,$93,$F2,$E7,$FA,$CA,$7F,$2E,$FF,$68,$9F,$F1,$BB
	.dc.b	$FD,$7D,$FC,$BD,$FF,$96,$E3,$FA,$61,$FC,$D1,$FC,$B1,$FC,$81,$FF
	.dc.b	$00,$57,$F6,$27,$FC,$D1,$5F,$82,$AF,$E6,$33,$FE,$07,$8F,$E3,$47
	.dc.b	$F8,$1A,$BF,$C6,$F5,$FE,$5C,$3F,$30,$7F,$24,$C3,$F9,$91,$FC,$C5
	.dc.b	$FC,$89,$E1,$3F,$9B,$C7,$85,$FC,$B1,$FE,$07,$BF,$E6,$2F,$FA,$DD
	.dc.b	$FF,$03,$DF,$E9,$E3,$FA,$C1,$FE,$5D,$3F,$D1,$77,$FE,$5D,$7F,$84
	.dc.b	$47,$F4,$2A,$FA,$71,$FE,$90,$AF,$9D,$1F,$CE,$A3,$F9,$8C,$7E,$60
	.dc.b	$FE,$62,$FF,$AE,$9F,$E6,$2F,$F2,$F3,$F9,$89,$F9,$69,$FC,$DC,$FE
	.dc.b	$32,$7F,$C1,$18,$7C,$6A,$FE,$5B,$7F,$CC,$63,$F2,$EF,$E4,$CF,$F2
	.dc.b	$35,$7F,$AF,$6F,$F0,$75,$FD,$5D,$FC,$C6,$1F,$AF,$8F,$F7,$61,$FA
	.dc.b	$F3,$FD,$75,$5F,$99,$7F,$8B,$47,$F5,$5B,$FE,$65,$1F,$EF,$3B,$FF
	.dc.b	$05,$61,$FE,$0C,$BF,$E6,$33,$FE,$62,$C3,$C2,$FF,$81,$5F,$F0,$53
	.dc.b	$FC,$BA,$3F,$35,$1F,$CC,$30,$FF,$04,$57,$F2,$F3,$FC,$11,$7F,$E0
	.dc.b	$4B,$FF,$6C,$47,$F9,$75,$C4,$FF,$82,$63,$FD,$69,$1F,$E0,$83,$F4
	.dc.b	$47,$F8,$B1,$FF,$2F,$C7,$F9,$7A,$3F,$CE,$11,$FC,$BD,$FC,$D9,$FE
	.dc.b	$5E,$7F,$97,$9F,$E0,$89,$FA,$51,$FE,$87,$3F,$2E,$7F,$A2,$6F,$F9
	.dc.b	$73,$FE,$5E,$AF,$E5,$CF,$F8,$E6,$FF,$CD,$11,$FC,$5E,$3F,$DF,$6F
	.dc.b	$FA,$72,$BF,$C1,$11,$FE,$B7,$8F,$E0,$CF,$FA,$F6,$3F,$98,$9F,$E5
	.dc.b	$F9,$FD,$04,$7F,$82,$9F,$D8,$3F,$9A,$7F,$CC,$0F,$FA,$DA,$3F,$32
	.dc.b	$FF,$B6,$8F,$F5,$A1,$FE,$60,$3F,$AA,$1F,$E4,$A7,$FD,$49,$7F,$CC
	.dc.b	$30,$FF,$30,$CF,$9C,$FC,$C9,$FC,$EA,$7F,$C1,$2F,$E6,$AB,$F9,$73
	.dc.b	$FE,$EB,$86,$1F,$99,$BF,$E6,$2F,$E6,$6B,$FD,$68,$7F,$81,$EB,$FD
	.dc.b	$66,$7F,$2C,$BF,$E5,$B1,$FE,$05,$C3,$E2,$CF,$F9,$7E,$3F,$CB,$CF
	.dc.b	$F9,$6C,$CE,$1F,$B9,$BF,$E5,$D8,$7F,$98,$AF,$27,$FB,$25,$FC,$3D
	.dc.b	$FF,$81,$AF,$FD,$5E,$72,$7F,$CB,$E7,$FA,$D0,$FD,$B3,$FE,$63,$7F
	.dc.b	$D6,$8F,$90,$FE,$FD,$1F,$E0,$78,$FF,$30,$47,$F3,$13,$FC,$73,$1F
	.dc.b	$98,$7F,$C6,$4F,$F9,$7E,$3F,$98,$4F,$E5,$E7,$B4,$FE,$5D,$FE,$5F
	.dc.b	$27,$F3,$07,$F3,$49,$F1,$57,$F2,$FA,$FF,$03,$D7,$E4,$AB,$F1,$27
	.dc.b	$E6,$EF,$15,$F9,$A8,$FF,$31,$9F,$CC,$27,$FD,$68,$FF,$30,$3F,$97
	.dc.b	$1F,$97,$3F,$97,$1F,$D4,$6F,$FC,$0C,$FF,$2A,$3F,$97,$1F,$CB,$27
	.dc.b	$F9,$73,$FC,$A6,$7C,$CF,$F8,$1A,$FF,$AA,$BE,$DB,$FE,$5D,$1F,$CC
	.dc.b	$23,$FC,$91,$87,$F9,$7D,$FC,$71,$FE,$75,$8F,$CC,$1F,$CC,$1F,$CE
	.dc.b	$CF,$C0,$7F,$61,$7F,$C9,$A3,$B3,$FE,$07,$8F,$F3,$25,$FE,$69,$FF
	.dc.b	$2D,$9F,$8C,$8F,$F5,$08,$8F,$F0,$45,$7F,$83,$1F,$E4,$38,$7F,$98
	.dc.b	$2B,$F9,$59,$FE,$07,$9F,$ED,$B5,$FC,$BE,$3F,$CB,$A7,$F8,$1E,$BF
	.dc.b	$9B,$DF,$F2,$E8,$FF,$5B,$4F,$F2,$E7,$F9,$AC,$78,$70,$FD,$EC,$7F
	.dc.b	$97,$6B,$DD,$3F,$EB,$5A,$FE,$5F,$7F,$D6,$27,$FA,$39,$FE,$B9,$BF
	.dc.b	$F5,$BD,$FF,$28,$3E,$84,$FE,$64,$FE,$60,$65,$FF,$2C,$1F,$DC,$CF
	.dc.b	$F3,$01,$FC,$C2,$F8,$B6,$FF,$B8,$3F,$ED,$59,$FF,$5C,$BB,$BF,$EE
	.dc.b	$93,$FE,$F9,$8F,$F5,$D1,$FE,$0E,$7F,$D7,$4F,$E7,$6B,$F9,$9B,$FE
	.dc.b	$60,$3F,$38,$7F,$30,$3F,$33,$1F,$9B,$7F,$BB,$61,$FE,$BA,$7F,$C7
	.dc.b	$35,$F9,$C7,$FC,$13,$5F,$CC,$9F,$F0,$6C,$FE,$EC,$47,$F3,$29,$FF
	.dc.b	$05,$1F,$EB,$89,$F0,$1F,$9B,$7F,$AB,$BF,$E6,$2B,$FE,$1B,$1F,$E0
	.dc.b	$B6,$BF,$12,$3F,$DD,$5C,$7F,$9D,$7F,$78,$1C,$60,$C7,$13,$FA,$03
	.dc.b	$F9,$6D,$61,$FF,$0A,$61,$CB,$1F,$E0,$D8,$7F,$3B,$3E,$F4,$78,$27
	.dc.b	$FD,$88,$71,$3F,$CD,$37,$5F,$E7,$D8,$7E,$7E,$BF,$65,$5F,$E6,$27
	.dc.b	$FC,$2E,$7F,$0E,$3F,$CD,$B3,$E0,$3F,$A1,$4F,$E9,$CF,$54,$7F,$42
	.dc.b	$C3,$B4,$FE,$C9,$E8,$F0,$7F,$3D,$1F,$DF,$30,$FF,$74,$47,$09,$FA
	.dc.b	$B1,$FE,$EC,$9F,$D6,$E1,$9C,$3F,$CE,$F8,$7A,$6F,$FA,$A5,$FF,$BE
	.dc.b	$EF,$FC,$BF,$87,$92,$7F,$8D,$5F,$C5,$7E,$5D,$7E,$09,$F6,$18,$8F
	.dc.b	$F6,$55,$7F,$88,$EB,$C1,$5E,$4A,$B5,$EA,$7F,$D4,$75,$FB,$3A,$FB
	.dc.b	$ED,$BF,$14,$FF,$A0,$67,$EE,$CE,$EF,$C3,$3F,$EC,$B9,$FD,$D4,$FF
	.dc.b	$B5,$CF,$F1,$49,$FD,$44,$FF,$14,$1F,$B9,$9F,$CD,$BF,$BB,$8F,$89
	.dc.b	$3F,$91,$8F,$F4,$19,$FD,$4C,$FF,$14,$BF,$3C,$FF,$21,$3F,$8A,$8E
	.dc.b	$31,$FE,$63,$AF,$DD,$4F,$FB,$68,$FF,$41,$9F,$ED,$A3,$FC,$58,$7E
	.dc.b	$1E,$3F,$A4,$1F,$08,$A7,$98,$CC,$76,$7F,$D0,$C7,$E1,$8F,$EA,$F5
	.dc.b	$B7,$FA,$CD,$6D,$FF,$15,$1F,$20,$FF,$47,$5F,$F4,$DB,$FC,$63,$F3
	.dc.b	$A3,$FA,$E4,$62,$BA,$47,$FB,$8E,$BA,$3F,$92,$3A,$1F,$B8,$7F,$DC
	.dc.b	$73,$FC,$02,$F7,$3F,$D1,$E7,$D0,$FF,$07,$AE,$AF,$F8,$EE,$3F,$20
	.dc.b	$FF,$AE,$DE,$F7,$FE,$8F,$BF,$E0,$78,$7E,$76,$BF,$C0,$58,$69,$FC
	.dc.b	$E1,$FC,$16,$FF,$16,$FF,$85,$EB,$F7,$2F,$A0,$FE,$C7,$5C,$EF,$E1
	.dc.b	$47,$F0,$A3,$F3,$53,$FB,$E7,$FD,$83,$5E,$4B,$FB,$F1,$FB,$77,$BB
	.dc.b	$FE,$98,$8E,$6F,$F1,$C3,$F9,$5C,$7F,$69,$C3,$C5,$7F,$E9,$9C,$3F
	.dc.b	$1D,$1F,$BE,$C3,$FD,$A7,$7A,$8F,$F0,$0C,$7F,$6A,$C3,$F5,$F3,$F3
	.dc.b	$0F,$ED,$6F,$89,$FF,$35,$3F,$2A,$7F,$75,$1F,$B4,$C3,$4F,$F0,$AC
	.dc.b	$3C,$EF,$F3,$27,$FD,$85,$3E,$17,$F7,$51,$FE,$C6,$AF,$AA,$7C,$27
	.dc.b	$F7,$15,$D2,$7F,$B9,$4F,$F1,$33,$E0,$4D,$79,$BF,$DC,$6F,$F7,$2F
	.dc.b	$FB,$E6,$3F,$60,$FD,$C3,$F2,$F1,$FB,$98,$CC,$FE,$A6,$7F,$9A,$CF
	.dc.b	$EE,$5F,$EE,$97,$F7,$DF,$DC,$9F,$DC,$6B,$F7,$97,$E9,$8F,$DD,$47
	.dc.b	$E9,$5F,$EE,$4E,$8F,$9C,$FF,$56,$8F,$F7,$E5,$7E,$EE,$3F,$C9,$27
	.dc.b	$7A,$F9,$51,$FC,$E6,$3E,$69,$FD,$54,$E6,$7F,$A9,$0F,$F7,$83,$BA
	.dc.b	$FD,$4C,$FD,$88,$FD,$BC,$FE,$A0,$FF,$53,$56,$AF,$9C,$7D,$87,$F1
	.dc.b	$67,$EB,$1F,$D7,$1F,$0E,$FF,$6E,$FF,$AA,$6B,$FB,$BD,$E6,$7C,$73
	.dc.b	$FC,$E0,$FF,$13,$3F,$AF,$7E,$81,$FE,$AA,$3F,$A1,$9F,$DD,$6F,$53
	.dc.b	$F9,$E8,$FF,$76,$57,$F2,$9A,$FD,$84,$66,$F8,$BF,$B1,$7E,$3B,$FE
	.dc.b	$C7,$AD,$4F,$E6,$23,$FE,$00,$3F,$D7,$EF,$ED,$47,$EC,$5F,$C9,$1F
	.dc.b	$DC,$63,$47,$D0,$AF,$D8,$5F,$CF,$8F,$D5,$D7,$EB,$CF,$19,$FE,$5A
	.dc.b	$AF,$D5,$3C,$6F,$D6,$FF,$68,$9F,$D7,$5F,$C4,$8F,$D5,$4F,$F0,$0C
	.dc.b	$3F,$61,$5F,$D9,$23,$F5,$8F,$AC,$FE,$4C,$7E,$96,$7F,$82,$BF,$E0
	.dc.b	$99,$FD,$45,$FE,$C2,$3D,$47,$F6,$29,$FF,$2B,$1F,$A8,$3F,$34,$7F
	.dc.b	$11,$3F,$89,$47,$F3,$47,$F5,$87,$F8,$03,$0F,$D6,$1F,$E2,$67,$F5
	.dc.b	$07,$FB,$DD,$FD,$63,$CC,$FF,$4B,$5E,$B0,$FC,$3C,$7F,$7D,$9D,$3F
	.dc.b	$CD,$AF,$E2,$3F,$A9,$7C,$AF,$EC,$2F,$FA,$39,$FC,$1A,$7F,$B3,$C7
	.dc.b	$F3,$09,$FE,$4A,$F2,$3F,$50,$7E,$8C,$FD,$54,$FE,$A1,$FD,$1D,$FF
	.dc.b	$B1,$CD,$3F,$E6,$29,$FD,$2B,$FB,$7B,$F1,$BF,$D9,$AB,$F5,$95,$93
	.dc.b	$89,$FE,$64,$AF,$EB,$A7,$E4,$27,$FB,$13,$C9,$B3,$A8,$FD,$A9,$FC
	.dc.b	$22,$3F,$58,$7F,$99,$AF,$FC,$C4,$FE,$9C,$FC,$89,$FC,$CC,$F9,$4F
	.dc.b	$F9,$9E,$B5,$3F,$A3,$9F,$23,$FA,$C7,$FA,$E5,$7F,$33,$7F,$A8,$C7
	.dc.b	$F5,$F9,$FD,$5E,$1F,$C7,$63,$FA,$E1,$F3,$70,$FF,$36,$DF,$EA,$A3
	.dc.b	$F6,$71,$FE,$70,$AF,$EB,$33,$F9,$1B,$FF,$08,$D7,$EA,$9F,$EA,$B1
	.dc.b	$FD,$0E,$7D,$31,$FE,$52,$3F,$54,$7F,$42,$3F,$C2,$B3,$FA,$A7,$FA
	.dc.b	$84,$F1,$AF,$0D,$7F,$74,$9F,$F2,$DC,$71,$7E,$8D,$7E,$95,$FD,$4B
	.dc.b	$F2,$9F,$F6,$35,$7E,$A6,$B3,$3F,$26,$7F,$18,$7F,$77,$AF,$D5,$57
	.dc.b	$E9,$E3,$F9,$69,$F8,$B8,$FD,$74,$FE,$AD,$E3,$1F,$A7,$3F,$D8,$51
	.dc.b	$FD,$5A,$BF,$4C,$7E,$4A,$3E,$BE,$1F,$A6,$7F,$D3,$67,$F5,$7B,$FE
	.dc.b	$6A,$FE,$BD,$FE,$35,$1F,$E4,$B8,$FF,$62,$3F,$D4,$5F,$F5,$F9,$EA
	.dc.b	$9F,$EA,$2F,$F9,$2A,$BF,$53,$1F,$D5,$63,$F6,$E7,$08,$FE,$94,$FE
	.dc.b	$8C,$FC,$FD,$FF,$32,$7F,$C0,$67,$F4,$C9,$CC,$FE,$76,$3F,$A4,$1F
	.dc.b	$AE,$7C,$EF,$F8,$ED,$FD,$09,$E0,$7F,$C8,$0F,$CD,$7F,$87,$9F,$A9
	.dc.b	$3F,$8B,$BF,$C6,$E3,$27,$C9,$8F,$EB,$87,$F5,$1A,$FF,$3C,$BF,$41
	.dc.b	$FF,$3C,$9F,$EC,$08,$FA,$15,$C2,$7F,$8A,$1F,$3E,$BC,$0F,$E8,$1F
	.dc.b	$F0,$04,$E4,$FE,$13,$7F,$CD,$AF,$FB,$E5,$7C,$B3,$F8,$29,$F1,$E7
	.dc.b	$F9,$6B,$FE,$78,$73,$5F,$3A,$3F,$88,$DF,$F7,$6B,$FE,$1F,$1F,$36
	.dc.b	$3F,$BC,$CF,$CD,$8F,$F4,$93,$F8,$47,$FD,$E1,$3F,$C3,$6B,$FB,$C1
	.dc.b	$F4,$67,$DD,$3F,$E6,$93,$F1,$C7,$B8,$FD,$99,$FE,$15,$BF,$F5,$35
	.dc.b	$FE,$72,$7F,$C5,$B3,$FC,$F4,$F9,$37,$F9,$FB,$FC,$DD,$7F,$8C,$AB
	.dc.b	$F9,$C9,$FA,$A9,$F3,$57,$DA,$73,$5F,$2E,$BF,$88,$57,$F3,$CA,$F9
	.dc.b	$51,$FC,$FA,$BF,$9D,$57,$E5,$CF,$93,$5F,$CE,$2B,$E4,$DF,$FA,$6A
	.dc.b	$BE,$4B,$FE,$99,$BF,$93,$5F,$20,$FD,$E9,$E4,$AF,$E6,$AE,$6F,$E4
	.dc.b	$CE,$E3,$F2,$93,$FD,$AC,$FC,$85,$FF,$03,$AF,$C9,$D7,$C7,$BF,$ED
	.dc.b	$15,$F2,$23,$FC,$AB,$5F,$1E,$3F,$81,$47,$C7,$BF,$F2,$64,$7E,$44
	.dc.b	$FC,$6B,$FA,$F9,$FE,$C5,$1F,$BE,$3F,$B1,$57,$E4,$1F,$91,$5F,$1E
	.dc.b	$BF,$7B,$1F,$BE,$8F,$EC,$F1,$F8,$F9,$FF,$24,$C7,$E3,$E3,$F7,$53
	.dc.b	$F8,$F9,$FF,$61,$C7,$E3,$DF,$E2,$8F,$C7,$81,$FE,$7C,$8F,$F2,$3B
	.dc.b	$F2,$67,$E4,$4F,$C7,$8F,$91,$1F,$CA,$A3,$F7,$51,$FB,$B8,$F9,$0F
	.dc.b	$F2,$17,$F9,$49,$F9,$53,$F8,$83,$FC,$FC,$FF,$05,$BF,$C2,$5F,$E6
	.dc.b	$D1,$F9,$07,$E5,$47,$F9,$DE,$BF,$1F,$1F,$BA,$8F,$C7,$D7,$C9,$3F
	.dc.b	$72,$FC,$97,$FC,$ED,$7F,$8F,$8F,$93,$7F,$8F,$9F,$92,$7E,$3D,$F9
	.dc.b	$2F,$E3,$CF,$92,$FE,$3E,$BF,$91,$CF,$E4,$6F,$F6,$D1,$F1,$EB,$E6
	.dc.b	$D7,$ED,$A3,$FD,$8D,$1F,$EA,$27,$FA,$24,$FE,$3D,$F9,$11,$FE,$36
	.dc.b	$7E,$41,$FB,$5A,$FC,$7D,$7F,$4F,$8F,$C7,$CF,$F9,$BC,$E7,$1F,$B5
	.dc.b	$3F,$8F,$D7,$F3,$69,$FE,$9D,$1F,$E8,$A8,$FF,$0B,$46,$23,$FC,$58
	.dc.b	$FE,$DA,$7F,$D3,$93,$FB,$68,$FC,$84,$7F,$9A,$CF,$F4,$BB,$FB,$63
	.dc.b	$FC,$87,$7F,$E1,$28,$FE,$C0,$FD,$D8,$FF,$2B,$3F,$ED,$38,$F9,$67
	.dc.b	$C6,$8F,$96,$FC,$89,$F9,$6F,$FA,$62,$FF,$22,$FF,$98,$A7,$F2,$2F
	.dc.b	$F2,$07,$F1,$67,$F7,$A3,$F1,$47,$FB,$D6,$FF,$BD,$3F,$2E,$3F,$CB
	.dc.b	$EF,$F8,$5D,$FE,$95,$3F,$B9,$7E,$5C,$7E,$41,$F9,$B5,$F9,$07,$FC
	.dc.b	$B3,$1F,$26,$BF,$8F,$3F,$B8,$8F,$93,$1F,$B8,$7F,$23,$5F,$90,$7E
	.dc.b	$49,$F9,$09,$F9,$31,$FD,$34,$FF,$2E,$9E,$A5,$7F,$E0,$C7,$FB,$81
	.dc.b	$F2,$0F,$F5,$F1,$FD,$CD,$FE,$3F,$3F,$92,$AF,$E3,$EF,$E4,$67,$FC
	.dc.b	$F3,$7F,$91,$8F,$E8,$A7,$E4,$63,$F6,$67,$C7,$3F,$B8,$1F,$EF,$03
	.dc.b	$F2,$27,$FB,$BD,$F8,$EF,$F2,$0B,$FD,$B5,$7F,$6E,$8F,$DB,$4F,$F8
	.dc.b	$6D,$FE,$1E,$7F,$9F,$8F,$90,$7F,$86,$DF,$DB,$BF,$90,$3F,$1A,$7F
	.dc.b	$60,$BF,$F7,$43,$F2,$0F,$F0,$DD,$FF,$46,$7E,$5C,$7F,$86,$AB,$F7
	.dc.b	$11,$FB,$9B,$FD,$C4,$7F,$45,$8F,$DC,$1F,$E7,$83,$FB,$5D,$FC,$77
	.dc.b	$F5,$2F,$E3,$EB,$FA,$2B,$F1,$E3,$FD,$51,$5F,$C7,$9F,$ED,$75,$FB
	.dc.b	$79,$FC,$69,$FB,$78,$FA,$31,$FB,$77,$FA,$2D,$FE,$35,$FE,$88,$7E
	.dc.b	$36,$7F,$8E,$3F,$08,$FF,$09,$C7,$ED,$63,$F1,$A7,$ED,$63,$E2,$1F
	.dc.b	$B5,$3F,$D7,$97,$FB,$57,$E2,$9F,$D1,$4F,$F6,$F5,$7C,$69,$FD,$A5
	.dc.b	$FC,$67,$FA,$19,$FB,$33,$F1,$91,$FC,$FA,$BF,$2B,$5F,$E9,$B3,$EE
	.dc.b	$CF,$FA,$D6,$3F,$CE,$11,$FD,$3A,$D8,$2B,$53,$C6,$57,$98,$FD,$0D
	.dc.b	$F2,$7B,$D7,$07,$FA,$CE,$1F,$8C,$AE,$B3,$99,$FC,$7C,$F4,$9E,$95
	.dc.b	$E8,$9E,$27,$E3,$1C,$29,$EB,$7D,$23,$C6,$7F,$AE,$61,$1E,$8B,$F0
	.dc.b	$47,$FB,$AA,$7B,$D7,$C2,$8F,$6E,$18,$9F,$F3,$1E,$1C,$5E,$A7,$81
	.dc.b	$F9,$6E,$0F,$F5,$FE,$1E,$3B,$F3,$DF,$F9,$77,$0F,$0C,$FF,$83,$A7
	.dc.b	$2F,$C1,$C3,$C1,$7F,$B2,$9F,$F0,$0E,$1E,$17,$99,$FD,$62,$3C,$18
	.dc.b	$7D,$EB,$F0,$4F,$FA,$AC,$FF,$60,$9C,$27,$EA,$9C,$1E,$B7,$E0,$8F
	.dc.b	$F5,$75,$F7,$9E,$57,$EE,$7C,$2F,$CC,$AE,$0F,$F2,$23,$E7,$CE,$27
	.dc.b	$FD,$90,$70,$7C,$B5,$FE,$92,$78,$9F,$8C,$7F,$C3,$F8,$7D,$68,$FC
	.dc.b	$7E,$1C,$27,$FD,$97,$1C,$0F,$5D,$7C,$E7,$E5,$57,$E3,$67,$FC,$E5
	.dc.b	$3F,$88,$7F,$D9,$D1,$97,$E2,$C7,$F1,$A8,$C4,$7A,$DB,$3F,$0E,$3F
	.dc.b	$92,$9F,$94,$3E,$41,$F9,$6B,$AF,$EE,$EB,$C0,$FC,$27,$FD,$D3,$5E
	.dc.b	$19,$FD,$03,$E7,$BF,$85,$5D,$1F,$89,$36,$FE,$6D,$6D,$F8,$37,$E4
	.dc.b	$BF,$A2,$FE,$96,$75,$3E,$99,$FC,$61,$F6,$E7,$FD,$B2,$EC,$FC,$D1
	.dc.b	$F8,$B8,$F6,$C7,$F9,$48,$FA,$B1,$A7,$FD,$E9,$3F,$3D,$FC,$BC,$F0
	.dc.b	$AE,$D5,$F2,$27,$FD,$EC,$7D,$33,$E3,$1F,$E1,$28,$F5,$C6,$DF,$A4
	.dc.b	$F7,$7E,$3B,$E6,$9F,$8F,$5F,$7D,$F8,$ED,$7E,$3C,$6D,$EA,$FF,$13
	.dc.b	$7A,$D7,$E1,$5F,$F6,$74,$7C,$23,$D7,$3F,$C2,$1F,$D8,$4F,$EF,$1F
	.dc.b	$E5,$18,$7E,$89,$FC,$59,$FC,$BE,$3E,$39,$B7,$F8,$6C,$FE,$06,$BE
	.dc.b	$A4,$7F,$1B,$3F,$8F,$4F,$F1,$3C,$3F,$41,$19,$C3,$D7,$3E,$17,$D7
	.dc.b	$19,$AD,$BF,$B2,$7A,$C7,$ED,$23,$C7,$3F,$9A,$8F,$DB,$D7,$67,$F8
	.dc.b	$B4,$FF,$0C,$8F,$E2,$D3,$B7,$F5,$B1,$F8,$13,$E9,$C7,$F0,$D7,$A9
	.dc.b	$FC,$F1,$F8,$B7,$FC,$C1,$F8,$B1,$FC,$C5,$DB,$FC,$CE,$FD,$31,$FE
	.dc.b	$2B,$3F,$87,$C6,$8F,$E8,$0F,$F4,$1A,$F4,$3F,$8A,$8F,$05,$FC,$43
	.dc.b	$F8,$A5,$7C,$43,$EE,$CF,$F0,$77,$F2,$B3,$FD,$59,$FE,$1F,$87,$F0
	.dc.b	$DB,$F9,$31,$F0,$A7,$FA,$BB,$F0,$AB,$F7,$B1,$C8,$FD,$95,$FF,$02
	.dc.b	$7E,$BC,$7F,$0D,$7F,$18,$7E,$41,$FD,$C3,$F8,$E3,$F2,$11,$CD,$FE
	.dc.b	$19,$87,$F0,$7B,$FC,$3C,$7F,$59,$8F,$E1,$55,$FB,$93,$F8,$7D,$7D
	.dc.b	$47,$F8,$25,$7E,$9A,$BF,$7A,$FE,$99,$B7,$E6,$7F,$8E,$C7,$BD,$FF
	.dc.b	$21,$BF,$E0,$09,$EC,$74,$9E,$D3,$FB,$D3,$D1,$1F,$DF,$2F,$F8,$24
	.dc.b	$75,$3F,$88,$C7,$DB,$7F,$C2,$AF,$47,$94,$FB,$A3,$F4,$71,$FC,$0C
	.dc.b	$D1,$FD,$96,$FA,$3F,$C5,$5F,$D1,$BF,$D4,$E3,$6F,$F5,$C9,$FE,$6E
	.dc.b	$7C,$38,$FE,$00,$FF,$85,$AB,$F8,$45,$FC,$E8,$F8,$55,$FC,$72,$B9
	.dc.b	$BF,$E1,$78,$FA,$6F,$D9,$8F,$DF,$3F,$97,$BF,$93,$87,$F0,$58,$F3
	.dc.b	$C7,$F3,$2C,$3F,$86,$47,$F0,$43,$F0,$37,$FB,$87,$E0,$9F,$D3,$70
	.dc.b	$FD,$B3,$F9,$F3,$FA,$86,$1D,$DF,$F1,$24,$FF,$02,$9F,$CF,$CF,$EE
	.dc.b	$63,$F2,$D7,$FC,$03,$0F,$EF,$35,$FD,$BA,$69,$FA,$97,$F3,$D5,$FE
	.dc.b	$C6,$BF,$BE,$7F,$8A,$4F,$73,$FB,$E7,$FC,$18,$7B,$9F,$E9,$0F,$F8
	.dc.b	$0C,$F7,$3F,$D0,$6F,$9C,$7E,$F1,$F2,$C7,$EF,$6F,$FC,$15,$1C,$CF
	.dc.b	$DE,$47,$BA,$3E,$DD,$F2,$9F,$F2,$55,$7D,$18,$FD,$DE,$1F,$E0,$B7
	.dc.b	$E8,$9F,$72,$7E,$D5,$7E,$DE,$BF,$28,$7E,$64,$FD,$BE,$1F,$79,$FC
	.dc.b	$69,$FD,$80,$FD,$DD,$7E,$0A,$BF,$71,$1F,$D7,$B0,$FF,$48,$D6,$5E
	.dc.b	$B3,$FC,$06,$7F,$83,$D7,$EF,$2B,$F2,$95,$FA,$D9,$FC,$7C,$7D,$5C
	.dc.b	$3F,$C0,$13,$FA,$98,$FD,$A5,$FF,$81,$6F,$F5,$17,$FB,$23,$F3,$87
	.dc.b	$F3,$79,$CD,$7E,$C4,$F5,$1F,$DC,$E3,$E9,$9F,$E3,$B3,$55,$CA,$3E
	.dc.b	$C4,$FE,$D6,$3F,$5F,$1F,$66,$FF,$C0,$51,$CA,$FD,$77,$FE,$FF,$79
	.dc.b	$57,$F8,$5D,$F6,$47,$F7,$58,$FE,$09,$1E,$CA,$FE,$F4,$FD,$13,$F4
	.dc.b	$93,$F9,$A3,$F2,$13,$FD,$B6,$F2,$FE,$8E,$3F,$1E,$FF,$B0,$A7,$F5
	.dc.b	$97,$C4,$FF,$2D,$5F,$F7,$D7,$F6,$B5,$FE,$37,$3F,$D1,$53,$FA,$B3
	.dc.b	$F1,$D7,$EC,$8F,$F4,$13,$FD,$A4,$F5,$DF,$F9,$F2,$7F,$54,$78,$A7
	.dc.b	$F8,$EB,$FA,$37,$F5,$27,$88,$FD,$4C,$FF,$78,$9F,$D4,$DF,$F6,$99
	.dc.b	$E4,$7F,$21,$AD,$5F,$EA,$0E,$55,$FE,$76,$BC,$BF,$B6,$8E,$35,$FE
	.dc.b	$36,$AF,$EB,$A7,$F0,$43,$F4,$EF,$FA,$31,$F5,$CF,$F8,$1D,$FD,$41
	.dc.b	$FC,$12,$3F,$AC,$9F,$A8,$3E,$E9,$FE,$F4,$3F,$AF,$C6,$4F,$E5,$2F
	.dc.b	$F2,$77,$FD,$EA,$FE,$90,$CC,$FF,$3C,$7F,$AF,$1F,$D9,$23,$33,$FA
	.dc.b	$73,$FB,$35,$7F,$5B,$8F,$F3,$F3,$E8,$9F,$EB,$73,$FD,$12,$3F,$19
	.dc.b	$1F,$D8,$23,$F4,$EF,$E8,$CF,$49,$FE,$66,$8F,$D3,$BF,$E6,$23,$F7
	.dc.b	$07,$F9,$AA,$3F,$44,$FE,$98,$FE,$8B,$3F,$CD,$5F,$EA,$8F,$EA,$8C
	.dc.b	$D7,$E9,$67,$FB,$F1,$E9,$8F,$D2,$3F,$04,$FE,$A3,$3F,$EE,$58,$FE
	.dc.b	$A3,$1F,$CF,$23,$E8,$9F,$C6,$67,$FC,$21,$5F,$C5,$EB,$E8,$47,$F8
	.dc.b	$4E,$BF,$8B,$57,$CF,$8F,$F4,$B9,$F3,$DF,$D0,$1F,$EC,$0B,$FF,$2B
	.dc.b	$4F,$FA,$46,$3F,$88,$3F,$19,$1F,$E4,$8A,$FA,$31,$F3,$CF,$8F,$1A
	.dc.b	$BF,$43,$FE,$92,$3E,$81,$F2,$0F,$C0,$4F,$E8,$1F,$F2,$73,$99,$F9
	.dc.b	$F1,$FE,$8D,$3E,$5C,$FF,$13,$9F,$9D,$7F,$E9,$1A,$FE,$F4,$7C,$E3
	.dc.b	$D1,$47,$F0,$CF,$FA,$0E,$7F,$88,$C7,$A2,$3F,$BF,$3F,$D1,$5F,$9D
	.dc.b	$7F,$EC,$6B,$F9,$D1,$FE,$3F,$AF,$CE,$BF,$CA,$8F,$EF,$27,$C6,$3F
	.dc.b	$A2,$1F,$E4,$68,$FE,$88,$FD,$28,$FC,$ED,$7C,$F8,$FA,$17,$FE,$C1
	.dc.b	$AF,$D0,$57,$D2,$3F,$A2,$D7,$F3,$87,$FD,$2A,$7F,$AF,$27,$F3,$13
	.dc.b	$F9,$AB,$FF,$19,$CF,$CB,$AF,$95,$7F,$2E,$FF,$D7,$17,$F2,$E3,$FD
	.dc.b	$75,$5F,$98,$AF,$97,$1F,$18,$FC,$9B,$F9,$73,$FC,$62,$7E,$BA,$3F
	.dc.b	$C5,$AF,$F6,$98,$FE,$0C,$7C,$69,$F9,$37,$FE,$F0,$AF,$93,$5F,$96
	.dc.b	$AF,$CA,$9F,$C1,$2B,$E3,$D7,$F9,$82,$FF,$33,$5F,$C0,$EB,$F2,$A7
	.dc.b	$C4,$AF,$8E,$FF,$BA,$6F,$E3,$CF,$F0,$08,$FD,$E1,$F9,$19,$F8,$CF
	.dc.b	$E4,$2B,$F8,$04,$FE,$42,$FF,$AE,$CF,$E4,$23,$E2,$BF,$1A,$7F,$21
	.dc.b	$1F,$E1,$48,$F8,$D1,$FE,$16,$AF,$C9,$DF,$E3,$8F,$93,$5F,$16,$BF
	.dc.b	$1A,$7F,$84,$1F,$C6,$3F,$8D,$AF,$DD,$47,$C4,$8F,$8B,$1F,$BA,$AF
	.dc.b	$8B,$1F,$C9,$E3,$E2,$CF,$E2,$EF,$FD,$F5,$7F,$E0,$D9,$FC,$64,$FC
	.dc.b	$5A,$F8,$D1,$FE,$0B,$8F,$8B,$1F,$1A,$3F,$17,$1F,$C8,$E7,$F9,$41
	.dc.b	$F8,$BA,$F8,$B5,$F8,$B7,$FC,$07,$7F,$8B,$7F,$C0,$B3,$F8,$BA,$F8
	.dc.b	$D5,$F1,$EF,$E3,$47,$F8,$F2,$7E,$34,$7F,$8F,$A3,$E3,$57,$AA,$BF
	.dc.b	$A8,$CF,$F2,$09,$F8,$F3,$F8,$B8,$F8,$EE,$E7,$FD,$C0,$7F,$21,$7F
	.dc.b	$D5,$93,$FB,$18,$FD,$94,$FC,$89,$E1,$1F,$C6,$67,$FA,$79,$FC,$2A
	.dc.b	$FF,$84,$C6,$6F,$F6,$97,$F2,$6F,$FD,$43,$3F,$C7,$A3,$E4,$C7,$EC
	.dc.b	$A3,$E5,$47,$E3,$AF,$F6,$73,$F8,$E8,$FD,$9C,$7E,$3A,$3E,$4C,$7E
	.dc.b	$39,$F2,$CF,$F9,$80,$F9,$4F,$FA,$6A,$7E,$54,$7C,$B9,$F9,$51,$FC
	.dc.b	$6E,$78,$CF,$F9,$7E,$BF,$DB,$77,$FB,$48,$FE,$F3,$7F,$29,$FF,$6A
	.dc.b	$9F,$24,$FF,$6B,$C7,$E4,$27,$FD,$A2,$7F,$1A,$AF,$2C,$7F,$89,$E7
	.dc.b	$E5,$1F,$B5,$BF,$C8,$4F,$F9,$72,$FE,$44,$AF,$FC,$A8,$FC,$99,$FF
	.dc.b	$65,$DF,$ED,$A7,$FC,$B0,$7E,$46,$78,$3F,$CB,$DF,$F6,$21,$FD,$D6
	.dc.b	$BE,$5D,$7E,$4A,$3F,$70,$FF,$96,$2F,$F8,$EC,$7C,$A3,$99,$FE,$59
	.dc.b	$BF,$DC,$BF,$2E,$7F,$CA,$B3,$F2,$E3,$FD,$FA,$7F,$1D,$7F,$26,$FC
	.dc.b	$A8,$F9,$EF,$F4,$5A,$FF,$43,$9F,$D5,$1F,$F0,$41,$FE,$20,$BF,$97
	.dc.b	$1F,$93,$8F,$F7,$B9,$FC,$79,$FD,$CC,$FF,$A1,$23,$F7,$2F,$D0,$AF
	.dc.b	$E8,$91,$FE,$83,$3F,$8F,$3F,$B9,$7F,$23,$1F,$EF,$68,$FE,$3D,$1F
	.dc.b	$00,$FF,$77,$1F,$C7,$E3,$E4,$1F,$92,$3E,$44,$7E,$E9,$FC,$9B,$FE
	.dc.b	$1B,$8F,$90,$FE,$4A,$32,$FF,$6D,$3F,$C0,$71,$F5,$CF,$E8,$B3,$FB
	.dc.b	$B9,$FE,$89,$1F,$C8,$5F,$93,$3F,$D1,$5F,$CC,$CF,$C9,$3F,$C3,$47
	.dc.b	$FA,$D8,$FE,$F5,$3F,$EE,$49,$FE,$44,$FC,$A8,$F9,$2F,$BC,$FD,$D1
	.dc.b	$F2,$8F,$F2,$34,$7F,$22,$9F,$DD,$CF,$EE,$A7,$FC,$2A,$7C,$A3,$91
	.dc.b	$F9,$93,$FB,$A3,$FE,$01,$8F,$DE,$1F,$C9,$27,$F7,$93,$C2,$BF,$D5
	.dc.b	$4D,$AF,$F2,$33,$FD,$2A,$BF,$78,$7C,$A9,$FE,$4A,$FE,$48,$F8,$B4
	.dc.b	$E4,$FE,$4B,$5F,$A3,$6A,$C1,$8E,$E3,$FC,$E6,$66,$B0,$FD,$5C,$7F
	.dc.b	$C3,$58,$FF,$90,$30,$E7,$5C,$9F,$F2,$9D,$7C,$F7,$C3,$5F,$E2,$16
	.dc.b	$E1,$B9,$FF,$40,$CF,$79,$DD,$F5,$9C,$DE,$DF,$C3,$1E,$57,$FC,$75
	.dc.b	$3F,$D0,$AF,$FC,$79,$3C,$4F,$F1,$74,$FB,$F0,$F5,$C7,$F0,$5C,$3D
	.dc.b	$D5,$CA,$3F,$DE,$F5,$F6,$6F,$A3,$FE,$47,$BE,$33,$FE,$BA,$9F,$EA
	.dc.b	$38,$7C,$D8,$F2,$E1,$F8,$1C,$3D,$11,$DE,$7F,$A5,$57,$08,$EC,$70
	.dc.b	$8F,$F3,$5E,$1C,$23,$F8,$D4,$7C,$B7,$57,$F2,$DB,$7B,$7F,$CE,$71
	.dc.b	$FC,$A4,$F9,$8F,$09,$F8,$8F,$8E,$3D,$97,$FE,$8A,$BF,$F3,$21,$F0
	.dc.b	$E7,$F0,$0F,$CC,$7E,$09,$D9,$FF,$35,$5F,$86,$3F,$5C,$F1,$7F,$D9
	.dc.b	$11,$EE,$79,$BE,$07,$F0,$E7,$90,$FE,$F7,$38,$9E,$51,$F8,$0B,$D3
	.dc.b	$FC,$1D,$5F,$92,$3F,$A0,$5F,$07,$FB,$F9,$93,$E2,$9C,$2F,$F2,$F3
	.dc.b	$F1,$A3,$EA,$46,$2B,$51,$FE,$AE,$9F,$8B,$3F,$EB,$0B,$F8,$8D,$BF
	.dc.b	$A2,$D8,$F1,$57,$F0,$C8,$D5,$F4,$8F,$CD,$CF,$D0,$AF,$34,$FE,$2E
	.dc.b	$7E,$01,$D0,$FF,$77,$BF,$8A,$69,$F3,$4F,$88,$7E,$06,$BF,$D8,$B1
	.dc.b	$F8,$C9,$EB,$7F,$89,$AF,$98,$F5,$6D,$FA,$1F,$C5,$57,$D0,$7F,$DB
	.dc.b	$73,$CA,$B4,$7C,$28,$FD,$A9,$F8,$A3,$DC,$7F,$B7,$2B,$71,$E7,$8F
	.dc.b	$AE,$F9,$DE,$2F,$C3,$BF,$C2,$BF,$2A,$BF,$A3,$47,$F1,$AA,$FC,$84
	.dc.b	$7F,$AB,$5D,$57,$FB,$52,$FD,$13,$FE,$B1,$BF,$CB,$4F,$E6,$5F,$C4
	.dc.b	$47,$E1,$22,$9F,$8C,$AE,$AD,$9C,$CF,$A6,$3F,$16,$FE,$6D,$FB,$71
	.dc.b	$F8,$67,$F8,$59,$E2,$9F,$D3,$E1,$E0,$7F,$53,$34,$E9,$3D,$DF,$3D
	.dc.b	$FD,$9C,$3F,$84,$4F,$B2,$3F,$DD,$97,$EA,$7F,$09,$1F,$CD,$DF,$84
	.dc.b	$E4,$F3,$1F,$A9,$C3,$CC,$ED,$FC,$23,$FC,$17,$0F,$9B,$3F,$06,$FF
	.dc.b	$8F,$D7,$98,$FE,$1D,$87,$CC,$3E,$C5,$FF,$22,$7F,$08,$FF,$00,$BF
	.dc.b	$AF,$1F,$02,$BF,$CF,$2F,$FB,$CE,$FE,$BB,$FC,$4B,$0F,$DC,$5F,$E8
	.dc.b	$CF,$D9,$57,$F0,$59,$CC,$7C,$83,$FA,$AE,$18,$8F,$CD,$47,$F0,$63
	.dc.b	$F3,$47,$D8,$9F,$E0,$38,$6A,$B9,$4F,$F0,$CC,$3F,$32,$FF,$18,$AF
	.dc.b	$DF,$3C,$AB,$FA,$7D,$7F,$0D,$7D,$91,$FA,$27,$D8,$E6,$FF,$3C,$FE
	.dc.b	$19,$FD,$75,$FB,$63,$F9,$F9,$FB,$33,$F7,$73,$A7,$DD,$1F,$9D,$AF
	.dc.b	$E0,$F1,$F6,$8F,$B8,$7B,$A3,$E0,$47,$F0,$09,$EE,$F2,$3C,$73,$F8
	.dc.b	$3A,$FD,$AB,$F1,$9F,$6D,$FF,$45,$7F,$93,$47,$F8,$DD,$F6,$C6,$6F
	.dc.b	$E3,$CF,$EB,$E7,$C8,$FF,$78,$7F,$6A,$FE,$38,$1F,$B1,$BF,$D7,$BD
	.dc.b	$8C,$9F,$64,$FC,$81,$FD,$9B,$0F,$C2,$E1,$DA,$3F,$66,$7F,$00,$3F
	.dc.b	$75,$1F,$B8,$9F,$E0,$B3,$C8,$FD,$1C,$FF,$A1,$E3,$F3,$2F,$F8,$20
	.dc.b	$FD,$C5,$7F,$82,$70,$FF,$01,$CF,$F1,$33,$E0,$CF,$F8,$22,$7F,$6E
	.dc.b	$F8,$CF,$C5,$1E,$F3,$FB,$AE,$1F,$B8,$AF,$F0,$15,$7F,$1D,$8F,$DB
	.dc.b	$C7,$F8,$16,$BF,$80,$DF,$33,$FB,$E5,$FF,$01,$7C,$8F,$F8,$0A,$3F
	.dc.b	$81,$47,$ED,$6B,$FB,$2C,$7E,$D4,$E4,$7E,$D2,$7F,$BA,$DF,$ED,$23
	.dc.b	$E8,$47,$F2,$59,$FD,$9C,$7C,$68,$FE,$7B,$7F,$DE,$1F,$BE,$7C,$98
	.dc.b	$FD,$9C,$FE,$BA,$3F,$2D,$3F,$CE,$A7,$EB,$C6,$6F,$F3,$CF,$F8,$DA
	.dc.b	$7F,$B8,$3F,$72,$BA,$BF,$CD,$E3,$F6,$2F,$EA,$CC,$DF,$EC,$1F,$E6
	.dc.b	$33,$FB,$5B,$FC,$EC,$79,$5F,$D5,$3F,$A5,$3C,$27,$F9,$70,$F8,$27
	.dc.b	$F2,$88,$D1,$E2,$3F,$99,$3F,$DC,$AF,$EE,$3F,$EC,$E7,$F5,$D3,$FD
	.dc.b	$BA,$7F,$65,$1E,$29,$FE,$2B,$3F,$AE,$7F,$C2,$8F,$EB,$9F,$EC,$2F
	.dc.b	$89,$FF,$19,$CF,$EC,$E3,$94,$7F,$7B,$7F,$8E,$BF,$85,$3F,$85,$1F
	.dc.b	$AB,$C3,$FA,$DB,$EC,$AF,$76,$1F,$E7,$89,$FD,$4D,$7E,$92,$BF,$51
	.dc.b	$5F,$8C,$7F,$D0,$07,$F4,$13,$FC,$EF,$7A,$9F,$ED,$78,$7E,$A2,$7F
	.dc.b	$CD,$8F,$E3,$5F,$DC,$D7,$F9,$BA,$FF,$AF,$4F,$EA,$6B,$FA,$53,$FC
	.dc.b	$FF,$0F,$F0,$BD,$FE,$7F,$0F,$5B,$FA,$A9,$FE,$A6,$FA,$E3,$F4,$EF
	.dc.b	$F9,$A4,$FC,$B1,$FC,$96,$BF,$59,$87,$D6,$3F,$CC,$95,$FD,$71,$E4
	.dc.b	$7E,$9C,$E3,$3F,$E3,$39,$FF,$02,$9E,$B7,$FC,$11,$3F,$A7,$3F,$C2
	.dc.b	$91,$FC,$0A,$3F,$CB,$AF,$F8,$39,$F4,$9F,$E8,$FB,$FE,$B0,$7F,$98
	.dc.b	$2F,$F4,$D1,$F7,$0F,$E5,$11,$FE,$54,$7F,$4C,$7E,$88,$FA,$47,$F9
	.dc.b	$58,$FF,$2A,$46,$6F,$F8,$84,$7F,$AE,$4F,$D1,$1F,$A6,$AF,$C6,$5F
	.dc.b	$9E,$3F,$DE,$0F,$F7,$19,$FE,$AD,$3F,$A6,$9F,$E9,$A7,$E8,$4F,$CA
	.dc.b	$57,$D0,$7F,$CA,$AE,$63,$FA,$6C,$7E,$80,$FD,$44,$F0,$3F,$A8,$4E
	.dc.b	$27,$F4,$57,$FD,$36,$BF,$7F,$1F,$BB,$7F,$A7,$3F,$DF,$4F,$34,$7F
	.dc.b	$53,$7F,$42,$7F,$4C,$8F,$E8,$07,$E6,$4F,$F4,$0D,$FF,$50,$AC,$CF
	.dc.b	$D5,$BF,$D4,$1F,$85,$AF,$E8,$F7,$FE,$D6,$9F,$F1,$FD,$7E,$7F,$0F
	.dc.b	$8B,$1F,$E3,$7A,$FE,$8B,$87,$F9,$48,$F4,$3F,$A0,$7E,$73,$63,$F1
	.dc.b	$B5,$FC,$2A,$3F,$C6,$C7,$E5,$5F,$E8,$8F,$E6,$E7,$FA,$23,$FC,$EC
	.dc.b	$E3,$87,$E9,$4E,$07,$BD,$FD,$7C,$FE,$19,$FF,$5F,$3F,$D1,$EB,$FA
	.dc.b	$11,$F9,$83,$F4,$87,$F9,$53,$0F,$F1,$9B,$FC,$1C,$F4,$CF,$F9,$9D
	.dc.b	$FC,$E3,$FC,$80,$FC,$E1,$FF,$00,$2F,$F8,$9D,$7E,$70,$FC,$C4,$7E
	.dc.b	$50,$FE,$F5,$5F,$9D,$BF,$EF,$35,$FE,$31,$AF,$E8,$87,$E6,$EF,$FD
	.dc.b	$25,$1F,$B1,$BF,$F0,$3B,$FC,$9D,$FF,$37,$BA,$8F,$F2,$13,$F8,$68
	.dc.b	$FF,$19,$47,$E9,$27,$F6,$67,$F1,$4B,$FE,$1F,$5F,$E2,$D8,$FE,$7F
	.dc.b	$7F,$E2,$CA,$FC,$DD,$FF,$AA,$0F,$CD,$D7,$F8,$AA,$BF,$9E,$9F,$9B
	.dc.b	$9F,$E7,$67,$C5,$3C,$D3,$FE,$A1,$73,$1F,$CE,$CF,$CF,$BF,$16,$3F
	.dc.b	$9D,$1F,$93,$3E,$E1,$FE,$25,$7F,$9D,$C7,$E6,$B0,$FE,$73,$1F,$E9
	.dc.b	$49,$FE,$73,$5F,$E9,$8B,$FE,$81,$1F,$E1,$2C,$3F,$D6,$B1,$FE,$F9
	.dc.b	$BF,$CD,$C6,$2B,$E7,$4F,$F3,$78,$FF,$0F,$DF,$FA,$57,$0F,$F1,$1B
	.dc.b	$FE,$1E,$3F,$2B,$87,$FA,$4C,$F3,$9F,$CD,$DF,$F4,$84,$65,$FF,$0D
	.dc.b	$9F,$CD,$23,$FC,$35,$7F,$CC,$CF,$E6,$F3,$F9,$7C,$3E,$5D,$7E,$5E
	.dc.b	$BF,$C2,$75,$FC,$C6,$3F,$9A,$9F,$E9,$33,$FA,$54,$7F,$85,$6F,$FC
	.dc.b	$2D,$87,$FA,$1A,$BF,$2B,$3F,$E1,$68,$FF,$43,$57,$F8,$52,$3F,$C2
	.dc.b	$B1,$F9,$48,$FF,$09,$47,$F8,$4E,$3E,$7D,$7F,$84,$8F,$F4,$1B,$FE
	.dc.b	$7D,$AF,$C9,$61,$FE,$11,$3F,$98,$5F,$F2,$BA,$FE,$BB,$7F,$CA,$EB
	.dc.b	$F2,$0F,$F9,$EA,$BF,$95,$57,$E3,$E3,$F2,$17,$FE,$0C,$BF,$F7,$DB
	.dc.b	$FC,$A8,$FF,$79,$4F,$E4,$27,$FC,$0C,$7E,$32,$FF,$DD,$F7,$F8,$FA
	.dc.b	$F9,$75,$F0,$2F,$F0,$77,$FD,$BA,$7F,$DD,$58,$7F,$82,$30,$FE,$51
	.dc.b	$3F,$E0,$AB,$FE,$50,$7F,$81,$DF,$F3,$64,$FF,$9B,$63,$F9,$3C,$7F
	.dc.b	$B9,$9F,$F0,$29,$FE,$59,$9F,$C6,$9F,$C3,$E3,$FC,$FA,$7F,$25,$7F
	.dc.b	$C0,$98,$7E,$3A,$BF,$65,$5F,$ED,$9A,$FC,$75,$7F,$80,$8F,$E4,$B5
	.dc.b	$F8,$DA,$F8,$78,$7F,$B5,$AB,$F1,$B1,$FE,$64,$AF,$C6,$C7,$F9,$9A
	.dc.b	$FF,$91,$9F,$8A,$7F,$1B,$3F,$8C,$9F,$E4,$D7,$FE,$62,$AF,$E4,$35
	.dc.b	$FD,$E0,$FE,$3B,$3F,$8C,$BF,$EF,$98,$7C,$D9,$F6,$9E,$E7,$F9,$43
	.dc.b	$FE,$B4,$8F,$F6,$8D,$7E,$39,$DC,$6E,$FF,$91,$4F,$F9,$7E,$3F,$CC
	.dc.b	$15,$F8,$F8,$FF,$66,$4F,$E3,$E3,$F7,$73,$C8,$47,$F2,$07,$FC,$BE
	.dc.b	$7F,$AC,$1F,$E5,$73,$F9,$18,$FC,$94,$FD,$0A,$FC,$7C,$7F,$7B,$8F
	.dc.b	$C9,$3F,$E6,$1A,$FE,$43,$1F,$42,$3F,$CB,$27,$E4,$9F,$E4,$77,$F9
	.dc.b	$28,$FE,$43,$1F,$92,$AF,$F4,$21,$F9,$28,$FE,$41,$3F,$EC,$23,$F2
	.dc.b	$31,$FE,$57,$3D,$91,$FE,$83,$3E,$3C,$FF,$B2,$8F,$EF,$47,$F1,$23
	.dc.b	$13,$FE,$C0,$AF,$EF,$55,$CA,$3F,$CB,$15,$FD,$D4,$FE,$F7,$87,$F9
	.dc.b	$5B,$0F,$E4,$10,$AF,$E4,$D1,$FD,$EA,$3F,$CA,$C7,$F7,$93,$C1,$7F
	.dc.b	$32,$7F,$97,$BF,$E2,$78,$FB,$87,$F7,$A9,$FE,$44,$FE,$56,$BF,$2D
	.dc.b	$1F,$EB,$A9,$FE,$45,$5F,$E1,$F8,$FC,$B5,$7E,$56,$BF,$2D,$1F,$C8
	.dc.b	$AB,$F2,$D1,$E7,$9F,$E4,$47,$C8,$3F,$02,$7F,$BB,$2F,$F3,$0E,$67
	.dc.b	$FB,$BD,$0F,$E4,$97,$FD,$DC,$FC,$B9,$FE,$B6,$3A,$61,$FE,$59,$AF
	.dc.b	$F6,$7C,$FE,$5E,$7F,$92,$9F,$3D,$FF,$28,$9F,$DD,$8F,$F3,$05,$FD
	.dc.b	$29,$FF,$59,$1F,$EB,$53,$F9,$34,$7E,$56,$7F,$D0,$B7,$FC,$99,$FC
	.dc.b	$AC,$FF,$26,$7F,$92,$D7,$F7,$58,$FF,$59,$5F,$E5,$5F,$D6,$DF,$F8
	.dc.b	$2D,$F7,$3F,$10,$FF,$1C,$A2,$3F,$E2,$26,$BF,$E7,$73,$D0,$FF,$1A
	.dc.b	$7F,$CA,$4F,$F2,$9C,$3F,$95,$47,$F7,$88,$FE,$55,$1F,$D2,$9F,$CC
	.dc.b	$9F,$E2,$42,$CF,$E6,$5F,$F1,$D1,$FE,$AA,$3F,$BD,$1A,$9F,$E5,$A7
	.dc.b	$FA,$AD,$FC,$CD,$FF,$78,$3F,$2F,$87,$F2,$B7,$FD,$5B,$3F,$DE,$27
	.dc.b	$F0,$C7,$F9,$4B,$0F,$E5,$91,$FD,$E2,$32,$7E,$61,$FE,$F2,$61,$FF
	.dc.b	$27,$DF,$F9,$4A,$7F,$BD,$4F,$E6,$67,$F9,$76,$1F,$BF,$7F,$6F,$1F
	.dc.b	$E0,$E3,$FA,$13,$FC,$9A,$BE,$9D,$7F,$30,$8F,$EF,$B3,$FC,$C1,$F0
	.dc.b	$1F,$C0,$2F,$FD,$56,$FE,$64,$FE,$F8,$FF,$0C,$7F,$83,$D7,$F3,$03
	.dc.b	$F4,$2F,$CC,$3F,$67,$1F,$9A,$C3,$F9,$7B,$FD,$F6,$3F,$C9,$27,$FA
	.dc.b	$A6,$FF,$BE,$BF,$CC,$2B,$F9,$7B,$FC,$E2,$BF,$97,$3F,$98,$AF,$CB
	.dc.b	$BF,$CB,$1F,$A3,$7F,$DF,$5F,$E5,$B5,$FC,$B2,$79,$3F,$DF,$63,$FB
	.dc.b	$FC,$FF,$A9,$A7,$F9,$6C,$FF,$2C,$8F,$F2,$5C,$7F,$2C,$9F,$E0,$C7
	.dc.b	$C1,$7F,$DF,$27,$F2,$CC,$3F,$27,$1F,$CB,$EF,$FC,$94,$FE,$EA,$BF
	.dc.b	$BE,$3F,$E0,$4B,$FC,$91,$FE,$4A,$3F,$A5,$DF,$F7,$DB,$F8,$2F,$FC
	.dc.b	$00,$AF,$F7,$76,$1F,$72,$30,$FF,$7C,$7E,$C8,$A4,$7F,$8E,$23,$E5
	.dc.b	$1F,$CB,$0F,$EF,$F5,$C2,$7C,$33,$FE,$00,$9C,$61,$FB,$6A,$FD,$F4
	.dc.b	$7F,$69,$AF,$CA,$47,$F2,$CC,$3F,$99,$C7,$FB,$FF,$0F,$F0,$0C,$7F
	.dc.b	$A8,$2B,$FC,$02,$FF,$A8,$B0,$FC,$93,$FE,$A4,$9F,$C9,$3F,$E4,$C3
	.dc.b	$F2,$33,$FD,$53,$0F,$F5,$05,$FF,$80,$63,$FA,$E4,$FF,$29,$7F,$27
	.dc.b	$7F,$E9,$F7,$FC,$99,$1F,$46,$FF,$94,$47,$E4,$2B,$FB,$F4,$FF,$29
	.dc.b	$BF,$EF,$CF,$F2,$BA,$FF,$26,$57,$F7,$F9,$FF,$4E,$34,$FF,$7B,$9F
	.dc.b	$62,$7F,$94,$D7,$F7,$D8,$E0,$FF,$AE,$E7,$FC,$29,$18,$BF,$E5,$55
	.dc.b	$FE,$50,$9F,$F4,$F3,$FE,$F7,$9F,$F2,$85,$7E,$4E,$3F,$BA,$DF,$F2
	.dc.b	$79,$FE,$5B,$5F,$E9,$E8,$FF,$7B,$9F,$6E,$3F,$CA,$07,$FB,$80,$FC
	.dc.b	$BD,$FF,$80,$1F,$E4,$A7,$F5,$3C,$3F,$94,$3F,$C1,$CF,$E5,$05,$C3
	.dc.b	$F2,$F3,$FD,$FE,$3F,$2D,$87,$C7,$BF,$F7,$C4,$7E,$2E,$7F,$2C,$FF
	.dc.b	$94,$63,$FB,$FC,$FF,$7E,$7F,$D3,$C7,$F9,$46,$3F,$1E,$7F,$7E,$3F
	.dc.b	$32,$FF,$27,$8F,$EF,$F1,$FD,$F6,$3F,$2B,$5F,$92,$9F,$C8,$D7,$F9
	.dc.b	$49,$FC,$9C,$FE,$4A,$BF,$92,$BF,$EF,$43,$FC,$34,$7F,$54,$7F,$B6
	.dc.b	$E1,$FD,$FA,$7F,$BE,$47,$F9,$30,$FE,$3A,$79,$1C,$C7,$E5,$1F,$E0
	.dc.b	$91,$F3,$1F,$CA,$3F,$E1,$C7,$FB,$E3,$FC,$A2,$3F,$91,$BF,$C8,$8F
	.dc.b	$F2,$8C,$7F,$7B,$AF,$E4,$6F,$E5,$6F,$F8,$1C,$FF,$9F,$2B,$F9,$1C
	.dc.b	$7F,$00,$8F,$DC,$DF,$F7,$B7,$F2,$D3,$FD,$ED,$FC,$99,$FE,$4F,$AF
	.dc.b	$EF,$73,$C6,$7F,$B0,$CF,$F2,$67,$F2,$6F,$F9,$40,$FC,$9C,$FE,$4A
	.dc.b	$7E,$14,$7F,$98,$DF,$CA,$1F,$E1,$D9,$FC,$A3,$FB,$78,$FF,$3B,$1F
	.dc.b	$94,$3F,$B9,$1F,$EE,$A3,$F5,$47,$E5,$1F,$F5,$19,$FE,$F1,$7F,$D4
	.dc.b	$73,$FC,$A6,$FF,$24,$FF,$2B,$9F,$EE,$E7,$E4,$9F,$F2,$5C,$30,$FF
	.dc.b	$28,$47,$F7,$8A,$FC,$B5,$7F,$93,$CF,$EF,$11,$F8,$EA,$FE,$EE,$FE
	.dc.b	$42,$FF,$D4,$4F,$F4,$89,$FC,$8D,$7F,$A8,$A3,$F2,$31,$FD,$D5,$FC
	.dc.b	$8D,$FF,$75,$9F,$EE,$53,$FD,$D4,$FC,$89,$F2,$9C,$BF,$EE,$F3,$F6
	.dc.b	$93,$F9,$28,$FF,$52,$61,$FE,$EF,$9F,$C0,$47,$F9,$84,$FE,$BB,$3F
	.dc.b	$92,$7F,$AA,$BF,$92,$7F,$CF,$78,$7F,$25,$8F,$EE,$95,$FC,$02,$7F
	.dc.b	$D5,$73,$FE,$39,$3F,$CA,$CF,$F7,$5A,$FF,$51,$BF,$DD,$6B,$F8,$01
	.dc.b	$FE,$F0,$7F,$14,$7F,$98,$B0,$FF,$3B,$3F,$92,$3F,$BB,$1F,$CA,$67
	.dc.b	$FC,$C5,$3F,$E4,$E3,$FB,$B9,$F9,$43,$F9,$24,$7F,$93,$9F,$A0,$7F
	.dc.b	$25,$3F,$24,$FF,$BC,$E3,$FB,$BC,$7E,$56,$7F,$22,$FC,$39,$FF,$4D
	.dc.b	$1F,$91,$AF,$F5,$15,$7F,$92,$67,$FD,$45,$3F,$E1,$C8,$FE,$EE,$FF
	.dc.b	$92,$27,$F1,$53,$F9,$19,$FF,$79,$BF,$91,$C3,$F6,$B5,$FE,$A4,$9F
	.dc.b	$EE,$E7,$FA,$9E,$BF,$87,$61,$FC,$73,$0F,$F2,$45,$7C,$3A,$FE,$45
	.dc.b	$87,$E4,$70,$FE,$47,$87,$F7,$38,$FE,$43,$3F,$E4,$67,$FD,$E5,$5F
	.dc.b	$E9,$D7,$F2,$27,$F7,$2B,$FF,$4E,$C7,$ED,$67,$FC,$E7,$3F,$DC,$E3
	.dc.b	$FD,$3C,$FF,$73,$8F,$F4,$F4,$7F,$91,$2F,$F1,$AF,$F2,$13,$FD,$E7
	.dc.b	$18,$AF,$E4,$11,$FE,$A1,$8F,$B6,$FF,$91,$0F,$F3,$F5,$7E,$45,$FF
	.dc.b	$29,$5F,$E4,$5F,$E4,$58,$7E,$45,$FF,$42,$1F,$EE,$83,$89,$FD,$DE
	.dc.b	$7F,$92,$57,$E4,$9F,$E4,$87,$E4,$A3,$FB,$A1,$FE,$9D,$9F,$C9,$D7
	.dc.b	$92,$3F,$C4,$37,$D1,$B3,$FD,$DD,$FF,$22,$9F,$C9,$4F,$EE,$D7,$FC
	.dc.b	$96,$3F,$C9,$11,$FD,$F2,$3F,$C9,$0F,$FA,$81,$FF,$4F,$1F,$BB,$7D
	.dc.b	$67,$FA,$72,$7C,$0E,$0F,$EF,$33,$FC,$15,$FA,$D7,$FE,$F7,$76,$7F
	.dc.b	$A5,$4F,$E4,$F7,$FD,$E5,$FF,$28,$DF,$F2,$69,$FF,$26,$47,$03,$F3
	.dc.b	$F5,$FD,$E2,$7F,$2F,$1F,$E9,$DC,$3F,$2E,$FF,$91,$EF,$F7,$EF,$F2
	.dc.b	$47,$F3,$EF,$F0,$89,$F6,$3F,$5A,$BF,$34,$FF,$2A,$8F,$EF,$77,$FD
	.dc.b	$DA,$3F,$C9,$AF,$F2,$B8,$FE,$F5,$1F,$04,$FF,$2B,$61,$FE,$4D,$9F
	.dc.b	$DE,$C6,$69,$3F,$E1,$28,$FF,$28,$5F,$FA,$8A,$BF,$3D,$1F,$EA,$27
	.dc.b	$FC,$EA,$FF,$7E,$3F,$C9,$E7,$F8,$02,$3F,$C9,$EF,$F8,$0A,$FF,$C9
	.dc.b	$EF,$F7,$EB,$FE,$59,$5F,$DF,$23,$FD,$3E,$7E,$76,$7F,$DE,$EF,$E7
	.dc.b	$67,$F9,$89,$F1,$1F,$F4,$F9,$F9,$C9,$FF,$28,$1F,$E9,$F7,$E0,$57
	.dc.b	$F2,$C9,$FF,$28,$C7,$C3,$3E,$5C,$FE,$71,$C4,$FF,$B8,$9F,$F4,$29
	.dc.b	$F9,$D8,$FF,$32,$3F,$E5,$2B,$FC,$03,$FE,$ED,$65,$7E,$76,$3F,$21
	.dc.b	$1F,$2D,$FF,$05,$47,$F3,$73,$D4,$F0,$8E,$6F,$D2,$89,$3F,$5A,$3F
	.dc.b	$40,$6A,$FD,$91,$FE,$05,$BF,$D8,$4F,$E7,$E7,$FC,$0A,$7F,$BF,$4F
	.dc.b	$E2,$6F,$CD,$9F,$D0,$3F,$EF,$D7,$E2,$57,$F0,$4A,$FD,$04,$FF,$82
	.dc.b	$8F,$EC,$A7,$F3,$7A,$FF,$03,$57,$F8,$17,$0F,$F3,$15,$FF,$AB,$70
	.dc.b	$F8,$AA,$FF,$D4,$D7,$FE,$08,$BE,$F8,$7E,$89,$FF,$04,$E1,$FE,$08
	.dc.b	$AF,$D1,$61,$FC,$E1,$FF,$55,$DF,$F8,$21,$FC,$C4,$7F,$9B,$DF,$E0
	.dc.b	$38,$7F,$8D,$2F,$FC,$15,$1F,$E5,$EA,$FE,$74,$F0,$9F,$F2,$D3,$70
	.dc.b	$FF,$2D,$39,$9F,$F2,$D3,$FE,$06,$3F,$3A,$7F,$82,$23,$F3,$AF,$F8
	.dc.b	$2A,$3F,$3A,$FE,$34,$FF,$2C,$C7,$D4,$7F,$D8,$75,$FE,$5A,$AF,$F1
	.dc.b	$B4,$7F,$82,$0F,$CF,$47,$17,$FD,$F8,$7F,$01,$BF,$C2,$BF,$BA,$9F
	.dc.b	$E7,$4F,$D7,$BF,$E7,$75,$E1,$BF,$C2,$CF,$E7,$27,$F1,$EF,$E8,$0C
	.dc.b	$5F,$FB,$66,$BF,$9E,$47,$FB,$60,$FD,$01,$FE,$B1,$AF,$53,$FD,$22
	.dc.b	$7F,$C4,$4D,$3F,$2E,$7F,$6D,$9F,$F0,$73,$FE,$62,$BF,$E7,$95,$FE
	.dc.b	$B2,$8F,$F5,$BD,$7F,$98,$1F,$CE,$9F,$E6,$03,$FB,$54,$78,$1F,$E7
	.dc.b	$87,$F9,$C6,$FF,$CC,$15,$FC,$EE,$3F,$A2,$5F,$E7,$0F,$CD,$BF,$85
	.dc.b	$3F,$9E,$57,$E7,$1F,$E7,$D5,$FC,$ED,$FE,$44,$7E,$DC,$FE,$77,$1F
	.dc.b	$CD,$1F,$F5,$A3,$FE,$62,$7F,$9E,$4F,$C8,$7F,$23,$1F,$9B,$7F,$9D
	.dc.b	$3C,$5C,$CF,$F0,$0A,$FA,$51,$FD,$DA,$3F,$CC,$87,$F8,$51,$FC,$E9
	.dc.b	$FE,$B5,$9F,$F0,$A3,$88,$FC,$DB,$FD,$76,$7F,$84,$3F,$E6,$53,$F3
	.dc.b	$91,$FE,$65,$7F,$39,$87,$C7,$7F,$69,$29,$FF,$33,$57,$9E,$7F,$A8
	.dc.b	$47,$F8,$A8,$FF,$47,$3F,$E1,$7B,$FF,$33,$3F,$E1,$83,$FC,$2E,$FF
	.dc.b	$3B,$3F,$3C,$7F,$90,$DF,$ED,$8F,$F8,$55,$F8,$6F,$FA,$D9,$FE,$23
	.dc.b	$87,$FA,$D9,$FE,$73,$3F,$9C,$7F,$C2,$87,$E6,$EB,$FD,$6E,$FE,$6A
	.dc.b	$FF,$B6,$BF,$CE,$0F,$F3,$29,$F9,$D3,$F3,$53,$FE,$B7,$7F,$C2,$77
	.dc.b	$FC,$F2,$3F,$1D,$1F,$28,$FF,$09,$1C,$4F,$ED,$57,$FB,$47,$F9,$B3
	.dc.b	$FC,$E0,$FD,$A4,$FF,$37,$8F,$E6,$53,$E9,$7F,$9C,$D7,$E6,$63,$F6
	.dc.b	$67,$F3,$6A,$FF,$34,$CF,$F6,$1E,$5F,$E6,$A7,$E6,$96,$1F,$EA,$D8
	.dc.b	$C3,$F6,$5D,$4F,$F2,$F7,$85,$6E,$BF,$9C,$47,$F8,$61,$3F,$CE,$4F
	.dc.b	$F3,$71,$FA,$8B,$FF,$01,$C6,$8F,$F0,$D4,$7F,$0D,$9F,$C4,$19,$8F
	.dc.b	$CD,$1E,$13,$F4,$38,$E5,$FD,$0E,$1E,$77,$F9,$E9,$93,$FC,$3C,$F4
	.dc.b	$8C,$47,$F0,$D9,$E0,$FE,$7F,$0F,$F1,$FD,$FF,$9E,$4F,$F1,$01,$FD
	.dc.b	$00,$FF,$3C,$D7,$29,$FE,$EB,$5C,$9F,$E8,$F4,$C3,$FC,$C7,$7F,$92
	.dc.b	$7E,$7D,$F1,$7F,$BA,$61,$C6,$FF,$9A,$47,$C6,$AF,$F6,$89,$FE,$F7
	.dc.b	$C3,$53,$FA,$E7,$FC,$7F,$1F,$DB,$9F,$D8,$57,$D5,$9F,$E9,$91,$DC
	.dc.b	$FD,$7C,$ED,$FD,$B5,$FA,$E7,$FC,$B9,$3F,$ED,$7B,$C4,$7F,$6F,$BF
	.dc.b	$0B,$FE,$64,$C3,$EA,$3F,$D5,$5F,$81,$5F,$D3,$E3,$ED,$BF,$D4,$6B
	.dc.b	$E1,$C7,$F8,$C6,$31,$5D,$23,$F8,$4D,$74,$7F,$B6,$4F,$1C,$35,$5F
	.dc.b	$02,$7F,$A9,$1E,$03,$F7,$51,$EB,$8F,$7C,$7E,$EE,$7F,$1D,$3D,$EB
	.dc.b	$EF,$CF,$FA,$42,$FE,$14,$7B,$E6,$CF,$D3,$3F,$C7,$D1,$FE,$34,$AF
	.dc.b	$DF,$4F,$0A,$F6,$D7,$FA,$49,$C5,$F7,$79,$47,$F0,$07,$FC,$69,$5F
	.dc.b	$25,$F8,$F1,$FE,$BF,$3E,$11,$FE,$92,$AF,$F2,$84,$FF,$8D,$5F,$B0
	.dc.b	$7B,$A3,$FD,$35,$7F,$0D,$FF,$4E,$47,$CD,$7F,$C5,$0F,$DA,$AF,$BF
	.dc.b	$1F,$D5,$8F,$CA,$D7,$F0,$29,$FC,$B4,$7F,$94,$CA,$7E,$22,$3E,$A4
	.dc.b	$6A,$7E,$0C,$FF,$8A,$0F,$F4,$A6,$19,$C3,$C1,$1F,$E8,$83,$85,$7E
	.dc.b	$5D,$FF,$56,$47,$F0,$28,$FD,$33,$F7,$E3,$FA,$E3,$B7,$F7,$B1,$FC
	.dc.b	$68,$D9,$F3,$9F,$EC,$AE,$DF,$E0,$0F,$EC,$70,$FC,$0C,$7F,$01,$3E
	.dc.b	$09,$FD,$9A,$7E,$34,$7C,$1C,$3F,$80,$9F,$5E,$B3,$87,$C0,$BF,$DC
	.dc.b	$DF,$77,$FC,$58,$EA,$FE,$DC,$7E,$7A,$7E,$53,$FE,$FD,$AF,$A8,$7B
	.dc.b	$DF,$F2,$8B,$FB,$CC,$3F,$74,$FB,$F0,$FD,$D5,$F4,$AF,$F3,$1C,$F4
	.dc.b	$7F,$65,$3F,$BB,$3F,$69,$87,$E0,$E3,$E3,$1F,$93,$8F,$DC,$4F,$E1
	.dc.b	$F0,$F6,$4F,$DB,$7F,$81,$9E,$17,$FC,$A8,$F2,$7F,$90,$BA,$8F,$DB
	.dc.b	$9C,$A3,$F5,$87,$E5,$4F,$BE,$F3,$7F,$BA,$DF,$AD,$FB,$8E,$DF,$F7
	.dc.b	$9D,$FE,$16,$FF,$D5,$31,$B9,$FF,$6E,$47,$FA,$B2,$7F,$7D,$3E,$53
	.dc.b	$F3,$53,$CD,$FC,$1C,$FE,$C5,$E5,$7F,$E1,$A8,$FD,$A5,$7E,$E6,$B2
	.dc.b	$71,$AF,$E8,$37,$FB,$23,$DC,$FF,$5A,$34,$7F,$5C,$9F,$E3,$D1,$FB
	.dc.b	$78,$FE,$F7,$3A,$8F,$64,$7F,$77,$3F,$68,$7D,$DB,$FF,$55,$61,$FB
	.dc.b	$2B,$FD,$5C,$FB,$67,$FD,$48,$FE,$04,$FF,$79,$57,$EC,$23,$E9,$1F
	.dc.b	$E3,$69,$FC,$B3,$63,$F7,$57,$DE,$BF,$D5,$B7,$FC,$AE,$7F,$8A,$C7
	.dc.b	$F7,$43,$F5,$F7,$C8,$FE,$A3,$1F,$B2,$AF,$44,$7F,$71,$C3,$F9,$39
	.dc.b	$FE,$34,$AF,$1C,$FF,$4D,$BF,$DA,$47,$E0,$CF,$ED,$67,$B6,$3F,$8C
	.dc.b	$C7,$F2,$FB,$FD,$8C,$FC,$59,$FE,$1F,$5F,$DB,$EF,$F6,$13,$FB,$FC
	.dc.b	$3E,$23,$FC,$AE,$FD,$6F,$EB,$9F,$8A,$7E,$BE,$BF,$B3,$57,$ED,$23
	.dc.b	$E7,$1E,$B9,$FE,$BC,$7F,$66,$3F,$10,$7F,$AB,$E3,$F5,$D7,$F9,$B8
	.dc.b	$FE,$90,$7D,$D3,$FC,$65,$5F,$D9,$5F,$C3,$9F,$E0,$99,$FD,$94,$F1
	.dc.b	$BF,$E8,$F3,$FE,$30,$3F,$B0,$9F,$54,$FC,$7C,$FD,$2A,$FF,$16,$57
	.dc.b	$EA,$E3,$FD,$3B,$1F,$D8,$AF,$FA,$ED,$7E,$A4,$F4,$9F,$E5,$43,$E9
	.dc.b	$1F,$DF,$CF,$D5,$1F,$E3,$AA,$FC,$E9,$FD,$32,$3F,$B1,$47,$EA,$8C
	.dc.b	$DF,$27,$FD,$7C,$F2,$9F,$F6,$85,$7F,$2B,$AF,$D5,$57,$F0,$17,$FA
	.dc.b	$FB,$FE,$B8,$7F,$5D,$39,$7E,$1C,$7F,$5E,$7E,$91,$FA,$47,$F3,$47
	.dc.b	$F3,$C3,$FD,$38,$7F,$7D,$9F,$F2,$EB,$FD,$72,$3F,$9C,$3F,$E2,$D7
	.dc.b	$FA,$E3,$FA,$3A,$E4,$FF,$23,$BF,$D6,$5E,$6B,$8C,$7E,$EA,$7F,$4D
	.dc.b	$7F,$B4,$7F,$CA,$F8,$7E,$A2,$BF,$4D,$1F,$D6,$1F,$0D,$FC,$BA,$FD
	.dc.b	$34,$FE,$74,$F0,$9F,$EF,$93,$E6,$1F,$DA,$E3,$53,$CB,$0F,$CE,$CF
	.dc.b	$F9,$30,$FD,$33,$C9,$FE,$A5,$87,$E9,$AF,$FC,$9F,$1F,$D4,$EB,$FA
	.dc.b	$7B,$FC,$45,$FD,$23,$FC,$31,$F0,$D7,$F9,$E2,$F5,$1F,$AE,$9F,$80
	.dc.b	$7F,$24,$3F,$9A,$9F,$A5,$7F,$51,$1F,$A5,$8F,$B6,$70,$3E,$11,$FD
	.dc.b	$38,$FD,$11,$FA,$69,$F8,$2F,$F9,$20,$FE,$AB,$19,$9F,$EA,$31,$FA
	.dc.b	$17,$F3,$AF,$F5,$0B,$FE,$AB,$39,$3F,$53,$3F,$22,$7F,$95,$C7,$F8
	.dc.b	$19,$F5,$BE,$A7,$F4,$D3,$FD,$3A,$3F,$54,$7E,$92,$39,$03,$FD,$85
	.dc.b	$7F,$D3,$EF,$FC,$8B,$7F,$E4,$9A,$FE,$9B,$1F,$C7,$63,$F4,$75,$FA
	.dc.b	$48,$FE,$9D,$7F,$D2,$A7,$F6,$67,$EF,$27,$EC,$5F,$FB,$A8,$FF,$00
	.dc.b	$3F,$3D,$FF,$3A,$5F,$CF,$3F,$87,$C7,$98,$FF,$7C,$9F,$3A,$FF,$BF
	.dc.b	$DF,$F1,$28,$F9,$B1,$FD,$F6,$FF,$88,$47,$F1,$18,$FE,$8D,$3F,$32
	.dc.b	$3F,$A4,$57,$CC,$AF,$EE,$71,$FC,$3C,$F8,$D3,$F2,$5F,$F3,$7C,$FC
	.dc.b	$99,$FF,$34,$47,$F0,$BA,$F8,$F7,$FE,$67,$AF,$8E,$FF,$09,$3F,$11
	.dc.b	$3F,$C4,$CF,$E0,$C7,$F6,$DA,$FC,$39,$FD,$C5,$FE,$85,$39,$8F,$F1
	.dc.b	$24,$7E,$4C,$F8,$B7,$FC,$E9,$FC,$BD,$FF,$0E,$9F,$F1,$1C,$7C,$7B
	.dc.b	$FF,$4D,$C7,$E4,$A3,$E3,$C7,$F9,$98,$FF,$6F,$3F,$93,$3F,$1F,$5F
	.dc.b	$1E,$3F,$9B,$D7,$F6,$FA,$F8,$F5,$FC,$DA,$3F,$D2,$75,$F1,$EB,$FD
	.dc.b	$61,$3F,$A8,$3F,$9B,$57,$C8,$7F,$99,$D7,$C8,$7F,$D3,$2F,$F3,$13
	.dc.b	$83,$FC,$D2,$3F,$18,$FC,$88,$FC,$7C,$7F,$30,$8F,$E6,$31,$FE,$1F
	.dc.b	$8F,$8F,$1F,$D8,$4F,$8E,$FE,$F1,$F8,$F3,$FE,$4C,$8F,$8F,$1F,$22
	.dc.b	$7F,$1B,$1F,$1E,$7F,$D8,$D7,$FB,$98,$FC,$6C,$7C,$89,$FC,$6C,$7E
	.dc.b	$32,$BF,$1B,$1F,$1E,$3F,$AD,$C7,$C7,$9F,$F2,$14,$7E,$32,$3F,$1B
	.dc.b	$3F,$EB,$C8,$F8,$D3,$FB,$33,$FC,$7E,$FE,$CE,$7F,$69,$3F,$B3,$9F
	.dc.b	$C6,$C7,$F8,$F4,$FD,$A3,$FE,$B3,$33,$3F,$8D,$8F,$F1,$9B,$FB,$57
	.dc.b	$FA,$B4,$7E,$D5,$FF,$73,$4F,$F8,$26,$FF,$1D,$3F,$B6,$8F,$C7,$3F
	.dc.b	$E0,$59,$F9,$0F,$E3,$8F,$8E,$FE,$38,$FF,$02,$1F,$CA,$67,$FD,$14
	.dc.b	$FF,$9A,$2B,$F9,$0B,$63,$F2,$0F,$FB,$94,$FD,$4B,$FD,$65,$FE,$3E
	.dc.b	$7F,$97,$2F,$F9,$03,$88,$FC,$93,$FE,$E4,$3F,$16,$7F,$B1,$E7,$E6
	.dc.b	$BF,$90,$3F,$C0,$71,$FE,$6B,$7E,$6D,$7E,$E5,$FF,$70,$1F,$30,$FF
	.dc.b	$00,$3F,$32,$7F,$8F,$9D,$A3,$FD,$8C,$FE,$45,$FC,$95,$FC,$B8,$FC
	.dc.b	$94,$FC,$B3,$F2,$47,$CB,$3F,$C6,$11,$F9,$27,$E5,$9F,$92,$7F,$8F
	.dc.b	$3F,$B8,$9F,$E3,$CF,$E0,$AB,$FC,$5D,$1F,$22,$BF,$C4,$B1,$F7,$E3
	.dc.b	$FD,$3B,$1F,$BA,$9F,$F4,$BD,$7E,$EA,$3F,$29,$1F,$BA,$BF,$E4,$11
	.dc.b	$FE,$2A,$AF,$DC,$CF,$E5,$0F,$E4,$0F,$C8,$3F,$77,$1F,$B8,$7E,$41
	.dc.b	$FD,$A5,$F2,$B5,$F9,$33,$FC,$7D,$F9,$13,$FD,$1A,$7D,$91,$F2,$1F
	.dc.b	$CF,$9F,$3C,$FE,$E6,$7F,$96,$0F,$EE,$95,$F2,$5F,$F6,$1D,$7F,$95
	.dc.b	$CF,$DD,$57,$C9,$3F,$74,$76,$BF,$C8,$CF,$FB,$02,$BF,$75,$3F,$EB
	.dc.b	$C8,$FD,$D1,$F8,$D9,$FC,$7B,$FE,$55,$AF,$8F,$3F,$EB,$98,$F8,$E7
	.dc.b	$F1,$E9,$FC,$74,$7F,$47,$AF,$DD,$4F,$E3,$A3,$F1,$8F,$F4,$59,$FE
	.dc.b	$D9,$5F,$E4,$D3,$E3,$4F,$ED,$E7,$E1,$D7,$E3,$63,$FB,$5C,$7F,$1E
	.dc.b	$3F,$CF,$87,$F8,$66,$3F,$C4,$07,$EC,$1F,$F2,$13,$FE,$18,$4F,$F9
	.dc.b	$B1,$FF,$31,$9F,$BA,$3F,$DC,$A7,$F4,$77,$F9,$0C,$7F,$9F,$23,$F9
	.dc.b	$0B,$FE,$7B,$3E,$41,$F8,$E3,$F9,$09,$F8,$D7,$F9,$0B,$FE,$B2,$7F
	.dc.b	$68,$7C,$83,$FC,$61,$3F,$E9,$1B,$FE,$45,$3F,$24,$FD,$B4,$FE,$2C
	.dc.b	$FE,$8C,$FF,$21,$7F,$B8,$C7,$F2,$17,$FD,$09,$3F,$29,$FF,$53,$CF
	.dc.b	$E3,$DF,$D4,$3F,$28,$FD,$B3,$F2,$8F,$F3,$B9,$FC,$80,$FE,$63,$5F
	.dc.b	$24,$F9,$07,$E5,$4F,$F0,$D5,$FF,$20,$8F,$F7,$1B,$FE,$8A,$7F,$92
	.dc.b	$5F,$F9,$22,$3C,$D0,$7F,$70,$FF,$27,$8F,$DC,$3F,$C7,$4F,$F0,$BB
	.dc.b	$FB,$87,$F9,$05,$FE,$49,$F9,$77,$FD,$2D,$FC,$94,$FC,$A7,$FB,$8C
	.dc.b	$FF,$A9,$23,$F7,$13,$F9,$2B,$FE,$DE,$FF,$1F,$9F,$C9,$47,$F6,$F9
	.dc.b	$FE,$87,$49,$8F,$FC,$00,$9F,$F3,$7D,$7F,$93,$E7,$33,$FE,$3C,$3D
	.dc.b	$D8,$27,$53,$F9,$49,$FF,$52,$1D,$5C,$3D,$A3,$6F,$73,$FD,$E6,$E1
	.dc.b	$FF,$75,$35,$FC,$E9,$0F,$01,$D4,$E5,$5E,$48,$E0,$FF,$99,$CF,$A9
	.dc.b	$7C,$2B,$D7,$5D,$27,$C8,$F3,$3F,$CB,$2A,$57,$C6,$BF,$99,$C5,$AF
	.dc.b	$91,$86,$DF,$F3,$B6,$1E,$28,$FF,$4F,$D7,$83,$0F,$1C,$75,$9D,$5F
	.dc.b	$8A,$B5,$3E,$18,$F0,$CF,$C3,$BF,$AF,$1C,$5F,$41,$FC,$0A,$BF,$D8
	.dc.b	$D3,$F0,$9F,$F4,$45,$72,$9D,$DF,$E0,$23,$94,$F9,$4F,$C0,$4F,$B4
	.dc.b	$FF,$52,$5F,$18,$F9,$31,$F8,$9A,$F3,$5F,$77,$FD,$2D,$1C,$4F,$23
	.dc.b	$F5,$5F,$0C,$FF,$05,$9E,$0F,$C8,$8C,$CF,$D1,$C3,$94,$7F,$76,$AF
	.dc.b	$BB,$7F,$3E,$3B,$57,$C7,$C3,$CD,$87,$F2,$6A,$FE,$DA,$FD,$17,$EE
	.dc.b	$3D,$CC,$3D,$EF,$E7,$5F,$CA,$3E,$69,$FC,$C1,$CB,$F8,$C6,$CF,$33
	.dc.b	$FC,$CA,$EA,$7E,$6C,$7F,$3B,$9D,$1F,$ED,$CB,$FC,$B5,$78,$1E,$91
	.dc.b	$E0,$3E,$E1,$F0,$8F,$E7,$B3,$F4,$9F,$E8,$35,$D9,$FF,$1B,$1D,$8F
	.dc.b	$E3,$2F,$67,$D7,$7F,$12,$7F,$DD,$AF,$C4,$8F,$9E,$FF,$B0,$E3,$EF
	.dc.b	$BF,$93,$9F,$A7,$1F,$14,$FC,$19,$FE,$F4,$2C,$FA,$9A,$78,$DF,$C6
	.dc.b	$BF,$D6,$2F,$F3,$8F,$FB,$D8,$E7,$7F,$8A,$8C,$BE,$98,$CD,$FB,$EB
	.dc.b	$FD,$DB,$3F,$A9,$9F,$F7,$6B,$F9,$5A,$E3,$7B,$AF,$F5,$E1,$F8,$C7
	.dc.b	$F7,$13,$5F,$45,$7E,$6A,$7F,$A8,$99,$7D,$0F,$E1,$23,$E2,$57,$DF
	.dc.b	$7F,$27,$1F,$16,$7F,$90,$DF,$E5,$23,$F9,$01,$D8,$B1,$F1,$6F,$FC
	.dc.b	$FF,$87,$DF,$7F,$5B,$3F,$B8,$3F,$91,$61,$FC,$79,$FF,$00,$DF,$57
	.dc.b	$F8,$CB,$EF,$9F,$E4,$57,$D6,$7F,$0F,$7F,$7B,$0E,$B3,$FC,$41,$FA
	.dc.b	$13,$FC,$1A,$FF,$8C,$CF,$F8,$41,$FC,$44,$7F,$20,$9F,$A0,$D7,$C4
	.dc.b	$FF,$9E,$5F,$DA,$CF,$C3,$8F,$E9,$CF,$F1,$4A,$CC,$7C,$33,$CE,$74
	.dc.b	$3F,$88,$BF,$90,$BF,$3C,$FF,$88,$8F,$87,$3F,$C2,$2F,$C3,$3F,$C5
	.dc.b	$5F,$85,$3F,$AA,$C3,$F8,$B6,$1E,$6B,$FE,$18,$7F,$24,$7F,$86,$47
	.dc.b	$F0,$E7,$F8,$C4,$FF,$06,$AF,$99,$1F,$A7,$9D,$3C,$EB,$DC,$7F,$8E
	.dc.b	$63,$F8,$3C,$FE,$46,$3F,$92,$D7,$E8,$DF,$CC,$CF,$E2,$4F,$E0,$4E
	.dc.b	$4F,$EA,$CF,$F0,$59,$E5,$1E,$B3,$F8,$35,$FE,$7D,$B7,$DE,$7D,$71
	.dc.b	$FC,$02,$3F,$55,$7F,$C0,$1F,$A1,$1F,$D1,$9F,$A0,$D8,$FC,$F9,$FC
	.dc.b	$0A,$3E,$8C,$7E,$96,$3D,$EF,$F4,$83,$6F,$FA,$F6,$3F,$83,$C7,$43
	.dc.b	$FC,$1F,$1A,$79,$1D,$0F,$DD,$9F,$15,$F8,$2F,$F3,$27,$FA,$43,$6B
	.dc.b	$A4,$7F,$24,$3F,$80,$CF,$CF,$3E,$0B,$FD,$61,$FD,$FD,$FF,$6B,$7F
	.dc.b	$37,$5F,$EE,$E7,$CC,$7F,$21,$9F,$E0,$51,$F8,$67,$F9,$83,$E7,$3F
	.dc.b	$C2,$87,$EB,$4F,$C3,$BF,$C0,$23,$F8,$1C,$7C,$CA,$FD,$ED,$7F,$3F
	.dc.b	$9F,$DB,$4F,$E6,$AB,$FC,$0C,$FE,$F3,$0F,$E1,$77,$F5,$8F,$D2,$DF
	.dc.b	$F0,$17,$FA,$C4,$FD,$DB,$FE,$01,$87,$C0,$C3,$F7,$11,$E4,$9F,$DC
	.dc.b	$4E,$A3,$C8,$FF,$83,$F0,$FA,$F1,$FA,$89,$FE,$07,$5F,$E0,$57,$91
	.dc.b	$FB,$6A,$D4,$7F,$19,$3F,$C7,$71,$F5,$67,$F6,$37,$CA,$7F,$70,$F2
	.dc.b	$9F,$F0,$2D,$6A,$79,$4F,$B1,$FB,$15,$F3,$E3,$FC,$49,$5F,$C9,$E3
	.dc.b	$F8,$AB,$FE,$59,$7E,$DB,$FD,$91,$FA,$07,$B7,$0F,$EF,$31,$FA,$D3
	.dc.b	$F4,$67,$F9,$F1,$FE,$F6,$FF,$6F,$3F,$59,$5F,$80,$9F,$D8,$D7,$AE
	.dc.b	$7F,$BD,$1F,$9B,$3E,$7D,$7F,$7B,$AF,$DD,$4F,$27,$FD,$79,$5F,$B5
	.dc.b	$39,$57,$F7,$63,$91,$FB,$3B,$D4,$FF,$07,$7F,$29,$1F,$E8,$D9,$FD
	.dc.b	$6B,$FA,$73,$D6,$7F,$0E,$3F,$7A,$7E,$C0,$FD,$9C,$F6,$7F,$C9,$71
	.dc.b	$FC,$B5,$FD,$85,$72,$7F,$D0,$4F,$29,$FE,$20,$7F,$76,$39,$1F,$3E
	.dc.b	$FF,$59,$5E,$CA,$1F,$62,$BF,$B7,$5F,$ED,$67,$F5,$E7,$C5,$9F,$EE
	.dc.b	$0F,$F6,$C7,$FC,$44,$7E,$04,$FE,$1F,$3F,$EB,$4C,$3F,$5B,$5F,$AA
	.dc.b	$9F,$D7,$1F,$92,$C3,$FC,$4D,$5C,$4F,$21,$FE,$85,$9F,$E8,$F3,$FC
	.dc.b	$32,$7F,$63,$3C,$8F,$C9,$3F,$E8,$0C,$3F,$CE,$F3,$FD,$DE,$BF,$22
	.dc.b	$F3,$C3,$F6,$D1,$FD,$16,$3F,$B5,$47,$27,$F8,$FB,$FA,$D7,$F5,$07
	.dc.b	$E8,$EF,$FC,$E9,$1F,$D9,$2F,$2F,$F5,$77,$E7,$4F,$F9,$98,$FD,$3D
	.dc.b	$7F,$60,$8F,$51,$FA,$E3,$F5,$73,$FD,$7C,$F9,$8F,$F9,$82,$7F,$55
	.dc.b	$1E,$13,$F5,$35,$FE,$91,$9C,$C7,$F5,$D9,$FB,$87,$F9,$8C,$FA,$B5
	.dc.b	$FD,$51,$FE,$7C,$7E,$22,$FF,$AE,$3F,$D5,$8F,$D1,$9F,$91,$3C,$E7
	.dc.b	$F9,$8E,$BF,$4E,$66,$FF,$97,$47,$FA,$E8,$FD,$15,$FE,$55,$FF,$55
	.dc.b	$BF,$E1,$B9,$C9,$F9,$DB,$F7,$BF,$CE,$8F,$EA,$F3,$FE,$DC,$AF,$EA
	.dc.b	$B3,$FD,$D6,$7F,$D4,$37,$FE,$54,$AF,$CA,$1F,$AB,$BD,$5F,$C2,$AF
	.dc.b	$EA,$A7,$F9,$52,$3F,$C2,$CF,$EA,$8F,$D2,$C7,$AA,$3F,$9A,$4F,$E7
	.dc.b	$9F,$51,$FC,$61,$FA,$07,$F9,$3E,$3F,$CA,$91,$F3,$63,$F8,$C1,$F2
	.dc.b	$4F,$DE,$9E,$83,$F4,$F3,$98,$FE,$2F,$1F,$42,$7F,$C9,$97,$FC,$6A
	.dc.b	$BE,$7C,$7F,$93,$EB,$F8,$A5,$7C,$E3,$E5,$3F,$C2,$AB,$F3,$95,$FE
	.dc.b	$03,$33,$1F,$26,$7E,$21,$F5,$63,$FC,$19,$7F,$2D,$FE,$2C,$7F,$08
	.dc.b	$8F,$BA,$FF,$61,$3F,$CF,$2F,$C8,$AF,$9F,$1F,$2E,$FF,$C8,$57,$FC
	.dc.b	$26,$BF,$C5,$4F,$F7,$C7,$FC,$E8,$FF,$20,$3C,$F3,$FD,$E9,$FE,$53
	.dc.b	$1F,$DF,$DF,$98,$FF,$3F,$3E,$65,$7E,$6E,$7F,$A3,$1F,$2E,$7F,$CE
	.dc.b	$93,$F2,$EF,$FC,$69,$1F,$E7,$79,$F9,$71,$FD,$F6,$3F,$CD,$D3,$F2
	.dc.b	$EF,$FC,$E1,$7F,$CF,$23,$F3,$11,$FE,$73,$9F,$95,$1F,$CE,$CF,$8D
	.dc.b	$3F,$36,$FE,$4D,$7F,$76,$AF,$E7,$15,$FC,$E6,$3F,$CC,$D5,$F2,$23
	.dc.b	$F2,$71,$F2,$27,$FD,$3D,$1F,$21,$FC,$89,$F1,$CF,$E6,$75,$FD,$15
	.dc.b	$FD,$4C,$7F,$88,$8F,$E0,$27,$FA,$FE,$30,$FE,$48,$FE,$02,$7E,$A0
	.dc.b	$FC,$F5,$FF,$97,$1F,$93,$5F,$92,$AF,$91,$1F,$26,$3F,$25,$1F,$22
	.dc.b	$3E,$4C,$7C,$88,$FD,$F4,$7C,$89,$FE,$9B,$5F,$E5,$59,$F9,$33,$FD
	.dc.b	$92,$BF,$7B,$1F,$D8,$23,$FB,$0C,$7C,$9A,$FF,$66,$C7,$F6,$03,$F2
	.dc.b	$13,$FB,$A7,$E4,$C7,$FB,$2A,$3E,$4C,$7C,$A9,$FE,$B7,$1F,$EC,$58
	.dc.b	$FD,$C4,$FF,$62,$8F,$C6,$BF,$E1,$09,$FF,$06,$5F,$E3,$6B,$FC,$1D
	.dc.b	$7F,$8D,$AF,$8C,$7F,$A0,$AB,$E3,$3F,$E0,$B9,$F8,$CF,$E3,$4F,$83
	.dc.b	$3F,$E7,$B8,$FC,$74,$61,$FF,$6D,$57,$F0,$D3,$F4,$6E,$2F,$EB,$28
	.dc.b	$FB,$EF,$F8,$24,$FC,$95,$FC,$AA,$FD,$D4,$7C,$A7,$F9,$35,$7F,$24
	.dc.b	$8F,$95,$3F,$92,$8F,$95,$3F,$EF,$AB,$FE,$47,$1F,$B9,$AF,$95,$3F
	.dc.b	$E0,$8B,$FF,$02,$57,$E5,$4F,$C9,$57,$F4,$2A,$FE,$72,$7E,$44,$FF
	.dc.b	$1C,$1F,$C8,$0F,$97,$7F,$E3,$98,$FE,$43,$1F,$90,$BF,$DB,$9F,$2C
	.dc.b	$FF,$1B,$4F,$E4,$1F,$91,$1F,$E3,$33,$F2,$2E,$1F,$F1,$A4,$FE,$4A
	.dc.b	$31,$1F,$26,$3E,$5D,$FC,$A8,$FF,$52,$1F,$93,$9F,$E9,$91,$FB,$B8
	.dc.b	$F9,$47,$FA,$74,$FE,$47,$5F,$D3,$4F,$E5,$47,$F8,$C4,$FF,$4F,$9F
	.dc.b	$62,$FF,$C0,$67,$F9,$9A,$3F,$78,$7C,$89,$F9,$6F,$E5,$0F,$E4,$B7
	.dc.b	$F9,$47,$FC,$5A,$7F,$20,$7F,$26,$FC,$AB,$FE,$F3,$5F,$E2,$A8,$FE
	.dc.b	$40,$7F,$B4,$2B,$FA,$53,$F9,$43,$FD,$A3,$5F,$B9,$AF,$8B,$3F,$CD
	.dc.b	$6B,$FD,$94,$FF,$21,$8C,$CF,$C8,$3F,$A1,$D7,$F4,$A3,$F9,$0B,$FB
	.dc.b	$98,$FE,$95,$7F,$B9,$9F,$F6,$11,$F9,$08,$FA,$77,$FE,$0C,$8F,$F7
	.dc.b	$7C,$FC,$88,$FC,$8D,$7E,$EA,$BF,$CA,$F5,$F1,$E3,$F7,$51,$FE,$5A
	.dc.b	$9F,$C8,$C7,$C7,$7F,$D8,$33,$F1,$E7,$FD,$0E,$7E,$09,$FF,$10,$1F
	.dc.b	$E5,$89,$FC,$8C,$7E,$EC,$FC,$8C,$7F,$BE,$27,$FA,$3B,$F2,$23,$E1
	.dc.b	$BF,$BA,$3F,$DF,$17,$F1,$DF,$F7,$D4,$FF,$70,$AF,$F4,$0C,$7E,$34
	.dc.b	$FC,$54,$FE,$36,$BE,$39,$FB,$69,$FE,$8E,$FF,$20,$BF,$DB,$3F,$B5
	.dc.b	$AF,$E3,$E7,$FA,$E6,$7E,$33,$FD,$B0,$F8,$D1,$F1,$5F,$C6,$5F,$FA
	.dc.b	$DA,$7E,$33,$FD,$12,$3E,$31,$FB,$59,$F8,$CF,$F1,$03,$FD,$0B,$3F
	.dc.b	$19,$FF,$23,$57,$F2,$07,$E3,$1F,$8D,$AF,$F3,$DC,$FF,$85,$A3,$F6
	.dc.b	$A7,$C6,$7F,$68,$FC,$69,$FE,$D9,$1F,$1A,$BF,$C7,$93,$F8,$C7,$F8
	.dc.b	$D9,$F1,$9F,$D9,$47,$F1,$B7,$FB,$64,$7E,$32,$7F,$65,$5F,$E1,$38
	.dc.b	$FE,$88,$7E,$2A,$7F,$8D,$9F,$E1,$2B,$FD,$8C,$7E,$2E,$BF,$C7,$76
	.dc.b	$8F,$E7,$28,$80

;--------------------------------------------------------------------------------
;	FDDIV.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fddiv_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FDDIV.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FDDIV.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fddiv_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fddiv.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDDIV.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fddiv.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDDIV.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fddiv_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3D,$61,$F3,$7F,$DA,$BD,$E7,$99,$C6,$30,$DA,$FE
	.dc.b	$50,$D9,$DC,$7E,$72,$3F,$B8,$BF,$98,$29,$F9,$E6,$B4,$CF,$9D,$C9
	.dc.b	$E2,$D7,$FB,$DB,$5E,$86,$A3,$9C,$66,$3C,$47,$48,$C9,$1C,$10,$E4
	.dc.b	$53,$BB,$B7,$24,$32,$6C,$F1,$98,$34,$60,$87,$90,$C9,$4E,$04,$36
	.dc.b	$7C,$53,$65,$34,$71,$37,$1F,$05,$D1,$88,$F5,$3C,$0C,$9C,$48,$60
	.dc.b	$91,$CC,$91,$CA,$3A,$1F,$04,$E3,$1A,$70,$72,$8E,$0E,$9A,$E4,$DC
	.dc.b	$71,$7D,$ED,$7D,$0D,$7C,$26,$A3,$9C,$66,$3B,$1D,$23,$24,$70,$43
	.dc.b	$91,$4E,$EE,$DC,$90,$C9,$B3,$C6,$60,$D1,$82,$1E,$43,$25,$38,$10
	.dc.b	$D9,$F1,$4D,$94,$D1,$C4,$DC,$7C,$17,$46,$23,$D4,$F0,$32,$71,$21
	.dc.b	$82,$47,$32,$47,$28,$E8,$7C,$13,$8C,$69,$C1,$CA,$38,$3A,$6B,$93
	.dc.b	$71,$C5,$F7,$B5,$F4,$35,$F0,$9A,$8E,$71,$98,$EC,$74,$8C,$91,$C1
	.dc.b	$0E,$45,$3B,$BB,$72,$43,$26,$CF,$19,$83,$46,$08,$79,$0C,$94,$E0
	.dc.b	$43,$67,$C5,$36,$53,$47,$13,$71,$F0,$5D,$18,$8F,$53,$F8,$1C,$BF
	.dc.b	$D1,$17,$FC,$F2,$3F,$A1,$35,$FE,$97,$87,$F4,$47,$F9,$F9,$A2,$9A
	.dc.b	$8E,$13,$D2,$32,$DC,$3F,$9E,$36,$FC,$93,$F2,$4F,$51,$0F,$49,$70
	.dc.b	$F6,$12,$3D,$71,$C8,$F8,$B1,$97,$A1,$C8,$8E,$5D,$C7,$27,$B3,$5D
	.dc.b	$B6,$38,$47,$D0,$32,$76,$72,$F8,$CD,$38,$7A,$9B,$8E,$84,$70,$43
	.dc.b	$CA,$53,$9B,$B7,$24,$32,$6C,$F1,$10,$F0,$9B,$30,$7A,$8C,$1A,$39
	.dc.b	$98,$34,$60,$87,$28,$F7,$3A,$31,$1E,$77,$67,$53,$27,$9C,$C1,$0E
	.dc.b	$45,$8E,$A4,$8E,$91,$C8,$F3,$46,$5E,$87,$22,$39,$77,$1C,$9E,$CD
	.dc.b	$76,$D9,$FC,$C4,$7E,$86,$BF,$37,$87,$E9,$5B,$87,$E8,$A3,$F5,$12
	.dc.b	$3F,$C5,$05,$3F,$54,$53,$F5,$47,$F8,$A2,$BB,$B8,$AE,$AF,$EB,$27
	.dc.b	$C1,$7F,$E4,$2C,$9F,$37,$FE,$2E,$C3,$FC,$CA,$D8,$CD,$F5,$6B,$E1
	.dc.b	$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7C,$51,$B7,$2D,$72,$71,$29,$A3
	.dc.b	$07,$F9,$61,$E0,$D3,$27,$F9,$7A,$3C,$0E,$4E,$44,$3A,$90,$D1,$4E
	.dc.b	$51,$D9,$A7,$69,$F8,$6F,$33,$89,$0D,$99,$29,$D5,$E2,$D9,$F3,$BA
	.dc.b	$39,$46,$9B,$1A,$8C,$46,$C8,$E0,$91,$B2,$4F,$AE,$32,$D8,$CC,$FB
	.dc.b	$5A,$FA,$67,$6E,$67,$81,$1C,$10,$E1,$3C,$9F,$74,$6D,$CB,$5C,$9C
	.dc.b	$4A,$68,$C1,$F0,$9E,$0D,$32,$7C,$28,$F0,$39,$39,$10,$EA,$43,$45
	.dc.b	$39,$47,$66,$9D,$A7,$E1,$BC,$CE,$24,$36,$64,$A7,$57,$8B,$67,$CE
	.dc.b	$E8,$E5,$1A,$6C,$6A,$31,$1B,$23,$82,$46,$C9,$3E,$B8,$CB,$63,$33
	.dc.b	$ED,$6B,$E9,$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7D,$D1,$B7,$2D,$72
	.dc.b	$71,$29,$A3,$07,$C2,$78,$34,$C9,$F0,$A3,$C0,$E4,$E4,$43,$A9,$0D
	.dc.b	$14,$E5,$1D,$9A,$76,$9F,$86,$F3,$38,$90,$D9,$92,$9D,$5E,$2D,$9F
	.dc.b	$3B,$A3,$94,$7E,$F9,$FE,$8D,$F2,$DF,$EE,$44,$D2,$3F,$EE,$86,$BF
	.dc.b	$EE,$C6,$BF,$EF,$06,$BF,$D3,$5A,$FF,$51,$6B,$E0,$3A,$10,$E8,$53
	.dc.b	$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10
	.dc.b	$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$A6,$7B,$4A,$79
	.dc.b	$8A,$78,$4A,$76,$29,$D4,$A7,$57,$E9,$B8,$23,$82,$3E,$16,$BE,$16
	.dc.b	$BE,$16,$BE,$16,$BE,$16,$BE,$13,$A1,$0E,$85,$3A,$14,$E8,$53,$A1
	.dc.b	$4E,$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$10,$FA,$67,$B4,$A7,$98,$A7,$84,$A7,$62
	.dc.b	$9D,$4A,$75,$7E,$9B,$82,$38,$23,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$6B
	.dc.b	$E1,$6B,$E1,$3A,$10,$E8,$53,$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8
	.dc.b	$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$FE,$09,$1F,$E7,$22,$61,$FE,$70,$8F,$F4,$16,$1F,$E7,$C6,$1F
	.dc.b	$E7,$C2,$8F,$83,$7C,$0B,$E3,$FF,$D0,$13,$FE,$85,$9C,$C6,$23,$84
	.dc.b	$ED,$DC,$E9,$FF,$4C,$9F,$E8,$DC,$DF,$09,$FE,$8D,$DD,$8F,$FA,$3E
	.dc.b	$73,$5F,$E9,$39,$CB,$C6,$BF,$D2,$9F,$88,$BE,$7F,$84,$7F,$A5,$FD
	.dc.b	$37,$A1,$8B,$FF,$4E,$7C,$23,$C2,$C4,$23,$0F,$53,$FD,$45,$5C,$9C
	.dc.b	$07,$0F,$C4,$6B,$FE,$A5,$6B,$E3,$AF,$8C,$F1,$7C,$31,$CB,$3F,$7C
	.dc.b	$E4,$EF,$59,$C3,$FD,$27,$F4,$2E,$9F,$F5,$A0,$F9,$0F,$F5,$B4,$66
	.dc.b	$FF,$D7,$53,$97,$FD,$79,$19,$7F,$D1,$2D,$E4,$FF,$55,$B9,$9E,$12
	.dc.b	$8C,$57,$69,$F1,$92,$7B,$C6,$5A,$78,$23,$4F,$81,$F4,$C6,$E3,$D8
	.dc.b	$DA,$F1,$CF,$03,$C7,$5E,$2A,$E8,$7F,$AB,$DF,$57,$37,$FB,$23,$6D
	.dc.b	$4F,$E1,$6B,$BB,$FE,$DD,$8D,$E7,$67,$B6,$3E,$37,$B9,$E8,$C7,$C7
	.dc.b	$7E,$C3,$E2,$47,$46,$9E,$37,$9B,$E6,$3D,$A5,$8E,$6F,$C3,$3A,$63
	.dc.b	$E8,$3C,$AF,$43,$47,$23,$13,$E6,$7C,$04,$35,$1C,$E3,$D5,$8F,$FB
	.dc.b	$B2,$31,$3E,$09,$CE,$1E,$77,$A4,$72,$72,$F8,$30,$F8,$C7,$D7,$BE
	.dc.b	$93,$FE,$E8,$C7,$E8,$91,$C1,$F1,$2B,$91,$F0,$E7,$8B,$FE,$E9,$9E
	.dc.b	$0F,$CF,$7E,$19,$F1,$30,$F4,$C7,$94,$F8,$18,$FC,$CB,$F4,$9E,$A7
	.dc.b	$E7,$99,$C3,$D3,$3E,$1C,$63,$E4,$25,$F7,$BF,$43,$E0,$BE,$78,$FA
	.dc.b	$2B,$C4,$48,$CD,$FC,$BC,$3B,$CF,$DB,$9C,$BD,$B0,$FC,$56,$1D,$E3
	.dc.b	$ED,$47,$AA,$BE,$75,$F9,$AB,$E0,$CF,$B7,$0F,$16,$1E,$DC,$3C,$06
	.dc.b	$E7,$E8,$BF,$2D,$FC,$C6,$3F,$31,$F6,$9E,$23,$EF,$DF,$6C,$7C,$35
	.dc.b	$E9,$C3,$E0,$57,$96,$3B,$1E,$89,$F7,$D7,$AE,$3F,$33,$3B,$79,$63
	.dc.b	$C1,$F9,$31,$EB,$8C,$3B,$9E,$58,$FB,$6B,$E4,$1F,$80,$C3,$F8,$05
	.dc.b	$7B,$09,$1F,$8B,$7D,$AF,$E7,$CE,$17,$F1,$9E,$67,$EB,$AB,$B5,$FD
	.dc.b	$FB,$F9,$A7,$C0,$C3,$DB,$87,$E9,$31,$FC,$BE,$1F,$B5,$7E,$63,$F0
	.dc.b	$CE,$D7,$EA,$BE,$51,$8A,$F9,$E7,$D1,$7C,$C7,$8E,$3D,$AF,$D9,$20
	.dc.b	$C1,$1A,$8F,$9F,$87,$99,$F7,$9E,$EA,$F9,$CF,$D2,$3F,$93,$1D,$1F
	.dc.b	$E8,$38,$7A,$A3,$99,$E7,$39,$BF,$02,$36,$7E,$3A,$36,$7C,$CB,$FC
	.dc.b	$FC,$EE,$93,$FC,$C4,$FA,$C7,$C7,$38,$5F,$D0,$C7,$51,$F3,$70,$FB
	.dc.b	$AF,$C1,$9F,$81,$87,$D8,$9F,$95,$3F,$26,$BD,$0F,$ED,$6B,$A3,$F0
	.dc.b	$EF,$F3,$B8,$FB,$EF,$D7,$3F,$3B,$0F,$9B,$87,$89,$ED,$1F,$59,$F6
	.dc.b	$CF,$90,$E3,$1E,$59,$E1,$87,$E0,$23,$DB,$1F,$9E,$C3,$EB,$47,$9C
	.dc.b	$F2,$1F,$A5,$BF,$0B,$F5,$8F,$84,$79,$6B,$E0,$9D,$70,$FC,$C5,$F5
	.dc.b	$9F,$83,$5E,$67,$EF,$5F,$D2,$7A,$47,$6C,$7F,$71,$8F,$EC,$4F,$CA
	.dc.b	$CF,$E5,$1F,$C4,$E1,$AB,$F4,$DF,$F1,$C8,$FE,$11,$5E,$9C,$7C,$97
	.dc.b	$F5,$2B,$4F,$D6,$39,$9F,$B6,$9F,$AD,$8F,$A5,$F8,$E7,$E5,$EF,$AB
	.dc.b	$67,$D5,$5F,$1B,$93,$FC,$89,$3F,$8E,$8C,$65,$AC,$58,$FE,$4A,$F7
	.dc.b	$7F,$5C,$FF,$30,$57,$F9,$83,$9B,$4D,$8C,$9F,$24,$FF,$77,$CF,$27
	.dc.b	$F3,$B3,$C1,$D3,$FE,$03,$A7,$C1,$FF,$97,$B3,$FF,$1A,$57,$03,$FC
	.dc.b	$01,$3A,$D7,$FC,$15,$19,$9F,$F1,$D4,$E6,$3F,$C0,$C6,$74,$FF,$35
	.dc.b	$5F,$57,$D1,$5C,$23,$13,$E4,$3B,$B8,$7F,$09,$87,$9E,$31,$1D,$4F
	.dc.b	$F3,$D3,$85,$1B,$C3,$CF,$18,$3A,$B8,$8D,$47,$A0,$D5,$79,$67,$51
	.dc.b	$C0,$D4,$F0,$9F,$8F,$1C,$2F,$DF,$5C,$27,$94,$F0,$39,$3C,$2B,$95
	.dc.b	$7D,$BC,$B6,$66,$BF,$5D,$8F,$CD,$8C,$5F,$42,$D7,$82,$BB,$46,$63
	.dc.b	$E7,$99,$7A,$4F,$68,$F0,$1C,$5E,$73,$F3,$30,$F5,$4F,$13,$9C,$64
	.dc.b	$EE,$E5,$FB,$55,$E1,$6C,$65,$FB,$67,$B3,$1F,$F5,$6D,$F9,$27,$C4
	.dc.b	$76,$9F,$D2,$CF,$DD,$C7,$E3,$C7,$E1,$EB,$D4,$DA,$E2,$F8,$0F,$35
	.dc.b	$FB,$1E,$78,$7B,$1C,$CF,$7C,$7F,$35,$5B,$6C,$67,$1F,$15,$7A,$6F
	.dc.b	$FA,$1B,$FA,$7C,$3C,$6F,$BB,$0F,$84,$DB,$F8,$4E,$DF,$3B,$F1,$27
	.dc.b	$8B,$F9,$D3,$CE,$7C,$E3,$CE,$7B,$CF,$A4,$FC,$96,$C7,$9E,$BD,$D1
	.dc.b	$CA,$B4,$F4,$AB,$87,$E3,$6F,$84,$F7,$38,$38,$AF,$89,$87,$C2,$BF
	.dc.b	$3D,$7E,$B5,$B8,$FD,$68,$F8,$07,$A0,$A7,$43,$EB,$14,$CE,$1F,$41
	.dc.b	$CB,$5F,$8D,$87,$C8,$7E,$51,$F0,$0D,$DF,$D6,$C3,$E8,$C7,$CB,$AE
	.dc.b	$4D,$9D,$57,$D4,$8F,$DD,$1E,$73,$58,$FC,$AA,$E3,$1E,$F9,$F1,$34
	.dc.b	$F1,$1E,$AA,$EA,$F3,$3A,$9E,$F3,$A9,$F3,$E7,$AE,$1F,$26,$33,$87
	.dc.b	$D0,$C3,$C5,$7D,$27,$91,$D2,$F8,$BF,$C3,$B0,$FB,$04,$7E,$2D,$72
	.dc.b	$3F,$8E,$E3,$F6,$23,$F3,$87,$F7,$D3,$C3,$3C,$5F,$64,$F1,$7D,$17
	.dc.b	$E3,$BF,$0E,$1F,$42,$7C,$93,$92,$57,$19,$F6,$4F,$28,$C5,$F7,$8F
	.dc.b	$09,$E8,$34,$49,$E2,$48,$E1,$1F,$4C,$FC,$F3,$FC,$79,$FA,$A7,$43
	.dc.b	$D6,$7F,$35,$C3,$E1,$1F,$0E,$7D,$58,$FD,$39,$F5,$DF,$99,$F2,$4F
	.dc.b	$85,$B7,$E2,$AE,$04,$8F,$A8,$68,$87,$8A,$79,$36,$79,$3C,$0F,$D0
	.dc.b	$9E,$A3,$DD,$8F,$EE,$5E,$67,$B4,$87,$A0,$F9,$11,$D2,$7D,$B1,$D4
	.dc.b	$DD,$7E,$E7,$E5,$9E,$8B,$1F,$F3,$3E,$36,$FF,$CD,$35,$A9,$C9,$AF
	.dc.b	$D4,$9E,$DE,$EF,$85,$AF,$83,$FE,$22,$BF,$F2,$14,$62,$3F,$C3,$8D
	.dc.b	$73,$87,$26,$CF,$F8,$85,$F1,$C7,$78,$F1,$78,$D6,$73,$1E,$67,$FC
	.dc.b	$63,$8A,$75,$0F,$96,$7B,$D8,$6D,$F1,$C9,$83,$E2,$35,$72,$BC,$98
	.dc.b	$73,$8F,$43,$FE,$90,$C3,$9B,$E9,$6B,$FE,$3B,$ED,$74,$3E,$67,$9C
	.dc.b	$61,$B1,$99,$F8,$53,$97,$E2,$73,$BC,$07,$85,$E4,$8C,$36,$33,$3C
	.dc.b	$1A,$F8,$FC,$17,$4F,$89,$B1,$E0,$6B,$B8,$F8,$3E,$03,$C1,$FF,$3F
	.dc.b	$71,$7E,$0B,$C1,$7B,$B8,$E2,$FF,$41,$46,$31,$FA,$C3,$0E,$4E,$1A
	.dc.b	$FE,$05,$AF,$F9,$8D,$AF,$FA,$35,$AF,$FB,$03,$77,$83,$E1,$8F,$F6
	.dc.b	$1F,$82,$F0,$B1,$08,$C3,$E4,$AE,$2D,$AD,$CE,$A3,$F3,$87,$4A,$E7
	.dc.b	$5C,$23,$27,$DE,$AF,$C3,$B4,$FC,$4C,$6D,$FF,$3F,$35,$FF,$4B,$77
	.dc.b	$7F,$A6,$1F,$F6,$C7,$77,$FA,$C1,$FF,$4A,$1F,$EA,$7F,$0B,$FD,$9B
	.dc.b	$3C,$1B,$1C,$23,$FD,$49,$E1,$7F,$B5,$2B,$FD,$3A,$23,$FD,$63,$36
	.dc.b	$BF,$D5,$F5,$B9,$FF,$55,$47,$49,$FF,$4D,$78,$3F,$EE,$F7,$FD,$A3
	.dc.b	$C1,$1F,$F6,$69,$E6,$3F,$DE,$F8,$7A,$4F,$F5,$E3,$FE,$C8,$C3,$B3
	.dc.b	$EB,$3F,$D7,$93,$FE,$EC,$C3,$E1,$47,$27,$D7,$1F,$EB,$9A,$E9,$1F
	.dc.b	$06,$78,$9E,$28,$F7,$3F,$EB,$7A,$EE,$FB,$CF,$15,$73,$AE,$B1,$E9
	.dc.b	$7F,$DA,$CF,$18,$D3,$FE,$F5,$9E,$73,$A8,$FF,$77,$57,$CE,$8F,$9A
	.dc.b	$67,$9B,$FD,$AB,$3E,$F3,$E3,$3F,$F0,$01,$FF,$6A,$9E,$1C,$7E,$9E
	.dc.b	$1E,$08,$CD,$FC,$A7,$FD,$B2,$D3,$E1,$BF,$81,$3C,$D1,$C6,$FE,$5D
	.dc.b	$F7,$6E,$1F,$EE,$AA,$D9,$E1,$3E,$29,$4E,$F1,$F3,$DF,$4C,$71,$7C
	.dc.b	$B1,$9B,$FA,$31,$F1,$23,$F1,$6F,$D6,$8F,$8C,$FC,$D3,$B3,$EA,$7F
	.dc.b	$00,$79,$1F,$74,$7D,$66,$C6,$63,$C0,$7D,$0C,$3E,$4C,$74,$8E,$06
	.dc.b	$9F,$04,$F2,$7D,$27,$C9,$29,$E2,$2D,$FC,$9A,$FD,$23,$63,$D9,$5E
	.dc.b	$E7,$E2,$E1,$F9,$BA,$DE,$1F,$1E,$BE,$8C,$E6,$B8,$9F,$85,$C3,$E4
	.dc.b	$E1,$F8,$88,$FC,$C1,$F4,$8F,$35,$72,$7F,$40,$FC,$17,$EC,$CF,$B6
	.dc.b	$34,$FD,$08,$FA,$87,$78,$E0,$F5,$3E,$03,$C0,$F6,$47,$92,$FF,$43
	.dc.b	$87,$CE,$8E,$CF,$BB,$05,$B1,$FE,$10,$DC,$32,$FF,$0C,$7E,$A3,$E4
	.dc.b	$3D,$F7,$F8,$3C,$7E,$74,$7C,$C3,$DE,$7D,$A9,$E3,$1F,$1D,$FD,$86
	.dc.b	$2C,$3C,$77,$E9,$7C,$BC,$7E,$5C,$3F,$24,$D5,$82,$F9,$47,$F6,$8C
	.dc.b	$7D,$33,$9B,$E7,$1F,$E1,$9F,$06,$D7,$B8,$38,$7F,$DD,$38,$77,$7E
	.dc.b	$3C,$65,$FF,$1C,$9F,$E3,$B3,$FC,$81,$8F,$F9,$31,$CB,$FE,$40,$26
	.dc.b	$5F,$EA,$19,$FF,$2B,$46,$27,$FD,$31,$1F,$E5,$9B,$FF,$32,$D7,$FA
	.dc.b	$82,$36,$F1,$9F,$04,$61,$E2,$60,$95,$D5,$FF,$4C,$C6,$DA,$6A,$7A
	.dc.b	$4E,$F0,$FF,$46,$C6,$09,$3D,$A7,$5E,$17,$FA,$3A,$3F,$CB,$E7,$F9
	.dc.b	$90,$F8,$73,$FE,$61,$38,$BE,$09,$E3,$C7,$FE,$7D,$8F,$F3,$91,$C7
	.dc.b	$0F,$36,$1E,$28,$F0,$5F,$28,$C4,$73,$8C,$61,$E3,$C3,$E9,$4F,$88
	.dc.b	$E8,$7B,$CE,$64,$39,$16,$BA,$92,$7A,$57,$AE,$75,$18,$7A,$D7,$B1
	.dc.b	$F8,$67,$B0,$87,$E1,$27,$CF,$87,$C1,$8C,$1D,$1C,$3E,$93,$89,$4D
	.dc.b	$9E,$82,$5F,$9E,$38,$E1,$F2,$B0,$F3,$57,$28,$C4,$EA,$3E,$21,$CA
	.dc.b	$76,$F6,$AE,$8F,$B0,$C4,$69,$F2,$12,$BA,$BB,$30,$76,$9E,$78,$7D
	.dc.b	$69,$DB,$67,$55,$D6,$73,$87,$D1,$9F,$36,$1E,$9B,$FA,$07,$12,$CF
	.dc.b	$FA,$F3,$1F,$D5,$D7,$FA,$EE,$F9,$BF,$62,$7E,$9B,$E7,$8C,$BF,$1C
	.dc.b	$F1,$DF,$79,$DD,$F7,$7F,$CD,$F8,$7F,$AD,$E3,$FB,$6F,$83,$FE,$AB
	.dc.b	$BF,$F5,$19,$FE,$A9,$70,$FF,$AD,$7C,$1F,$F4,$BB,$70,$FF,$50,$78
	.dc.b	$1F,$EA,$77,$FC,$EB,$5F,$EB,$AD,$3F,$D7,$D3,$9A,$FF,$2E,$CE,$6B
	.dc.b	$FD,$66,$70,$BF,$F5,$A5,$FF,$AA,$BC,$1F,$F5,$45,$FF,$A8,$FC,$0F
	.dc.b	$0F,$81,$FE,$A4,$8F,$E8,$2D,$7F,$D5,$51,$CB,$F0,$F7,$E8,$7C,$B0
	.dc.b	$FF,$68,$46,$32,$FF,$45,$F8,$3F,$EA,$5C,$3F,$D5,$DC,$FF,$EA,$96
	.dc.b	$F8,$5F,$EA,$C3,$FD,$49,$E1,$48,$FF,$52,$35,$FF,$4F,$1F,$EA,$1E
	.dc.b	$6F,$F5,$21,$1F,$F6,$A7,$C4,$3E,$13,$FD,$16,$D8,$FF,$4E,$B5,$FF
	.dc.b	$41,$FC,$43,$E5,$77,$52,$76,$FF,$A5,$27,$51,$FE,$A3,$F1,$3F,$D1
	.dc.b	$3E,$17,$FA,$7A,$3F,$D5,$11,$82,$3B,$21,$FE,$92,$F0,$FF,$AD,$70
	.dc.b	$FF,$BD,$AB,$FD,$5A,$DA,$FF,$51,$C6,$38,$3F,$D6,$FD,$BF,$EA,$6E
	.dc.b	$EF,$ED,$FD,$7F,$D9,$F2,$FF,$15,$1F,$E2,$F3,$FC,$23,$E0,$7F,$7F
	.dc.b	$7F,$B9,$9F,$DA,$3C,$0F,$EC,$F3,$C6,$FF,$C0,$25,$EF,$FE,$EF,$3F
	.dc.b	$E0,$AA,$FF,$0B,$C7,$F6,$C8,$FE,$DF,$3C,$23,$1A,$7F,$A3,$EF,$FC
	.dc.b	$35,$8F,$F7,$FC,$7F,$C2,$55,$AC,$7F,$C1,$C7,$F8,$49,$47,$F8,$43
	.dc.b	$0F,$F2,$AC,$62,$7F,$D3,$51,$FE,$1A,$7F,$C6,$67,$F8,$4A,$3F,$C9
	.dc.b	$2F,$F9,$5E,$3C,$8D,$8F,$5E,$1F,$E6,$FA,$F8,$95,$AA,$FF,$0F,$1F
	.dc.b	$20,$EB,$5F,$E1,$6A,$FF,$0A,$BE,$99,$E1,$7F,$E0,$DA,$F5,$9F,$E3
	.dc.b	$4B,$FA,$18,$7C,$2C,$7E,$7E,$1C,$4F,$F1,$69,$ED,$AE,$2D,$C3,$E9
	.dc.b	$47,$8C,$FF,$16,$57,$3C,$7C,$71,$ED,$3E,$56,$1F,$26,$FA,$B6,$34
	.dc.b	$79,$4F,$6B,$F5,$1C,$C7,$84,$E2,$FD,$07,$E4,$BF,$8E,$3F,$CD,$F8
	.dc.b	$7F,$99,$E7,$DF,$3F,$E5,$23,$FC,$AF,$7F,$E4,$59,$D5,$78,$0F,$C5
	.dc.b	$9E,$D8,$FC,$D1,$F7,$5F,$7C,$F3,$7F,$C9,$46,$DE,$A7,$E0,$1F,$61
	.dc.b	$FE,$51,$C3,$EA,$D7,$E5,$A7,$D3,$5D,$4E,$EE,$09,$5D,$4E,$E7,$48
	.dc.b	$DB,$F1,$AF,$E6,$3F,$E8,$0B,$FA,$EE,$EB,$CA,$FE,$CC,$C1,$D3,$1F
	.dc.b	$DC,$C6,$A3,$E5,$E3,$FE,$74,$AF,$4B,$FA,$B9,$F8,$53,$FE,$71,$7F
	.dc.b	$28,$FD,$37,$FD,$1D,$3C,$23,$F2,$75,$D9,$F0,$63,$F0,$DF,$C3,$9D
	.dc.b	$EB,$EF,$BC,$23,$4F,$A0,$F0,$61,$F8,$77,$15,$FA,$1C,$7E,$0D,$FE
	.dc.b	$26,$7A,$9C,$F1,$F5,$3F,$E9,$B3,$C6,$FF,$AB,$5F,$B1,$1E,$2C,$7F
	.dc.b	$80,$E1,$F3,$B0,$FD,$74,$7D,$99,$F9,$07,$C0,$38,$12,$BF,$3F,$3F
	.dc.b	$05,$F4,$9F,$90,$3D,$38,$FE,$D1,$FB,$87,$C3,$3D,$35,$F7,$27,$91
	.dc.b	$EF,$AF,$DD,$4E,$9F,$03,$FC,$28,$8F,$C3,$7F,$0C,$F0,$7F,$99,$61
	.dc.b	$FE,$8F,$9F,$E1,$5C,$5F,$E9,$D6,$DF,$F2,$78,$FF,$49,$57,$FA,$58
	.dc.b	$FF,$4F,$1F,$E9,$9F,$03,$FD,$28,$E6,$37,$18,$3D,$91,$FE,$AF,$F0
	.dc.b	$1A,$E2,$3F,$D3,$A7,$F1,$AC,$53,$A8,$FE,$3A,$6B,$0F,$F1,$5B,$CE
	.dc.b	$B9,$3F,$E3,$ED,$59,$FF,$A5,$EB,$78,$74,$BD,$E3,$FE,$2E,$C3,$FC
	.dc.b	$97,$1F,$E2,$D7,$FD,$37,$5C,$31,$F1,$3F,$E2,$F8,$FF,$33,$1F,$E3
	.dc.b	$73,$D1,$1A,$C3,$FD,$7D,$18,$AE,$EF,$F8,$E8,$E1,$8F,$99,$EF,$3F
	.dc.b	$EA,$F7,$81,$F0,$0F,$63,$FE,$40,$3F,$CA,$58,$FF,$96,$9E,$F8,$79
	.dc.b	$EF,$BD,$FF,$94,$AF,$EB,$1D,$27,$C0,$FF,$BB,$B1,$FA,$B8,$7C,$58
	.dc.b	$F0,$47,$C5,$7C,$27,$C0,$AF,$F3,$35,$77,$8D,$9E,$D8,$F8,$CF,$F9
	.dc.b	$98,$F7,$E1,$E9,$7F,$CC,$07,$C0,$3F,$CC,$87,$95,$F6,$E3,$FE,$71
	.dc.b	$8E,$91,$8C,$7E,$B4,$75,$AE,$F8,$FA,$5F,$6C,$7E,$1E,$BE,$66,$1F
	.dc.b	$02,$BD,$CF,$0C,$3B,$5F,$C7,$BF,$F3,$83,$5F,$C2,$9E,$A3,$E1,$5F
	.dc.b	$8D,$C1,$E3,$71,$1D,$70,$F3,$E1,$F5,$EB,$FD,$87,$87,$C9,$34,$F2
	.dc.b	$8F,$55,$FD,$C3,$D2,$7E,$16,$FC,$93,$F5,$0F,$F7,$5D,$73,$9F,$09
	.dc.b	$23,$F0,$31,$97,$E8,$9E,$73,$C7,$1D,$9F,$A7,$3D,$B1,$FC,$11,$F8
	.dc.b	$6C,$7F,$3D,$3D,$2B,$C8,$7D,$D7,$E4,$BF,$B1,$3A,$3F,$B7,$3F,$0F
	.dc.b	$5D,$5F,$BA,$73,$3F,$07,$7F,$5E,$BB,$C7,$43,$E8,$9E,$83,$75,$E2
	.dc.b	$72,$6C,$F7,$17,$0F,$A4,$BB,$1D,$9F,$E8,$BC,$D2,$A7,$80,$F1,$F5
	.dc.b	$3F,$E9,$4A,$DD,$7F,$17,$8E,$97,$FE,$A0,$BE,$15,$A3,$FD,$3B,$1F
	.dc.b	$EA,$29,$EE,$73,$3F,$D4,$91,$87,$6D,$7C,$BF,$32,$F2,$7F,$93,$61
	.dc.b	$E1,$8C,$57,$7A,$F5,$7E,$C0,$F1,$77,$3E,$0A,$DD,$7B,$E3,$18,$6A
	.dc.b	$3B,$BF,$EB,$AB,$E5,$3D,$65,$5C,$B0,$F8,$5E,$07,$FA,$C6,$7D,$AD
	.dc.b	$7E,$6B,$53,$79,$1E,$53,$BF,$FD,$65,$5F,$1A,$73,$18,$8E,$7E,$53
	.dc.b	$9F,$FD,$69,$1F,$EB,$82,$1F,$3B,$C0,$FF,$59,$3F,$35,$AE,$7B,$FE
	.dc.b	$73,$F8,$28,$CB,$63,$FD,$6C,$FC,$73,$07,$C3,$EF,$FF,$5B,$B9,$3F
	.dc.b	$D6,$A7,$C7,$EF,$79,$5F,$80,$DB,$F1,$11,$89,$FF,$5B,$C7,$FA,$CD
	.dc.b	$F1,$F7,$FF,$AF,$C9,$E0,$7F,$AF,$A3,$B4,$7F,$AD,$DF,$CC,$B5,$57
	.dc.b	$FB,$03,$8B,$FC,$9F,$1F,$E1,$2B,$FF,$4A,$3F,$E6,$51,$DB,$FE,$69
	.dc.b	$8F,$F1,$33,$FE,$52,$8F,$F0,$6F,$81,$FE,$52,$EC,$FF,$07,$71,$7F
	.dc.b	$89,$9E,$AF,$F8,$63,$C0,$FF,$5D,$57,$49,$FF,$34,$4E,$70,$F0,$CF
	.dc.b	$F9,$C0,$FF,$1B,$1F,$E2,$83,$81,$FE,$5B,$8E,$0F,$F9,$46,$FF,$D5
	.dc.b	$75,$FE,$22,$3C,$77,$FE,$4A,$AD,$E1,$FE,$46,$AF,$41,$FE,$3D,$9D
	.dc.b	$1F,$E8,$6B,$F1,$1F,$EB,$D3,$C0,$E6,$7E,$15,$74,$BF,$7C,$E5,$FF
	.dc.b	$22,$DF,$FA,$52,$FD,$75,$FE,$D3,$7F,$C8,$4F,$17,$FD,$00,$7F,$8D
	.dc.b	$0F,$79,$F3,$8F,$F1,$4C,$FB,$30,$FF,$2D,$C7,$94,$FF,$3C,$9F,$EE
	.dc.b	$E9,$EC,$F2,$9F,$F5,$F6,$1E,$8C,$3C,$D3,$FE,$62,$38,$3F,$5A,$3F
	.dc.b	$01,$18,$BF,$AB,$1C,$5F,$1E,$1F,$2A,$7F,$D6,$C7,$AB,$0F,$0D,$72
	.dc.b	$3F,$D9,$11,$F5,$CF,$F6,$49,$EF,$7E,$A3,$D0,$E6,$7E,$54,$E0,$7D
	.dc.b	$4C,$3F,$D2,$31,$C2,$3F,$DB,$35,$F6,$EF,$E2,$D6,$8F,$AA,$F5,$9D
	.dc.b	$9D,$AF,$E9,$3F,$3C,$C9,$FE,$89,$25,$FA,$8F,$07,$7F,$FA,$1A,$39
	.dc.b	$63,$F3,$1F,$66,$1F,$E6,$F3,$FC,$E4,$7D,$97,$9B,$FE,$AC,$BF,$8E
	.dc.b	$FE,$92,$79,$47,$F9,$FE,$7F,$D0,$A7,$92,$BF,$D0,$93,$A3,$FD,$15
	.dc.b	$1E,$98,$F2,$3E,$97,$FD,$30,$FD,$FB,$F5,$5F,$C4,$8E,$8F,$FA,$2A
	.dc.b	$FF,$75,$1F,$6E,$FF,$09,$3E,$5C,$3E,$56,$1D,$8F,$1B,$FC,$1A,$3F
	.dc.b	$43,$5F,$C2,$F0,$FC,$14,$7B,$EF,$9C,$FE,$B2,$B8,$3E,$59,$D4,$7E
	.dc.b	$41,$E7,$1F,$A4,$C3,$E2,$3F,$A7,$E2,$FF,$4C,$5F,$F7,$78,$FF,$4A
	.dc.b	$99,$C7,$FD,$34,$7F,$9D,$27,$FD,$2B,$1A,$3F,$D2,$2E,$0C,$E3,$FE
	.dc.b	$15,$26,$1F,$E1,$D9,$FF,$21,$C7,$F8,$E0,$91,$FE,$2F,$8F,$F0,$77
	.dc.b	$7F,$F9,$32,$3F,$C4,$51,$FE,$16,$3F,$C5,$AF,$37,$FC,$75,$97,$F9
	.dc.b	$01,$FF,$11,$1F,$E2,$7C,$3F,$CB,$6F,$FA,$C5,$FF,$12,$69,$FE,$7E
	.dc.b	$AE,$5E,$07,$F8,$DA,$BF,$C5,$67,$F9,$F4,$FF,$36,$1D,$2F,$FC,$C3
	.dc.b	$3E,$78,$FF,$33,$1E,$53,$C0,$E5,$F2,$E1,$EB,$7F,$CF,$06,$8E,$8F
	.dc.b	$F8,$D9,$F8,$B1,$FE,$59,$7A,$4F,$81,$FF,$28,$BE,$F7,$CD,$87,$F9
	.dc.b	$55,$F9,$38,$7F,$06,$7C,$A7,$73,$75,$FE,$3F,$7F,$C7,$A7,$80,$F1
	.dc.b	$C7,$FA,$C2,$7E,$3C,$7F,$94,$27,$E3,$BF,$18,$E0,$74,$3C,$F7,$EB
	.dc.b	$7E,$0D,$7F,$9E,$A3,$E1,$47,$80,$FF,$72,$63,$D8,$FA,$F8,$FF,$99
	.dc.b	$EB,$C5,$7E,$7C,$7F,$CA,$37,$E0,$21,$D8,$FF,$41,$61,$DA,$37,$19
	.dc.b	$C3,$FD,$63,$5F,$E6,$C9,$F9,$93,$F2,$0E,$37,$E7,$23,$82,$57,$E6
	.dc.b	$1F,$A7,$7D,$C9,$3C,$27,$D7,$3F,$8A,$3E,$31,$F4,$CF,$4E,$1E,$F7
	.dc.b	$FD,$87,$8F,$D0,$3B,$3E,$59,$AF,$03,$E9,$1E,$DC,$7F,$21,$5F,$9B
	.dc.b	$9F,$F4,$35,$7E,$16,$B8,$47,$ED,$1F,$F3,$F9,$8A,$FF,$49,$78,$1F
	.dc.b	$E7,$CC,$3F,$CA,$CD,$9F,$F1,$C6,$5F,$E7,$CF,$03,$FC,$F7,$1F,$E7
	.dc.b	$8A,$FF,$20,$D6,$E3,$FC,$E4,$D7,$FC,$E6,$7F,$9D,$8E,$7E,$23,$C5
	.dc.b	$FF,$42,$E0,$8C,$5F,$F9,$9F,$B7,$FC,$CC,$FF,$9B,$8F,$F3,$B4,$BB
	.dc.b	$BF,$CD,$B8,$7F,$92,$6B,$6F,$F9,$CA,$BF,$CA,$B1,$FE,$78,$8F,$F4
	.dc.b	$33,$5E,$87,$FA,$1F,$BB,$FC,$DC,$4F,$15,$E2,$FF,$97,$A3,$FC,$E1
	.dc.b	$E0,$7F,$9B,$EB,$FD,$1D,$CD,$FE,$6E,$6C,$65,$FF,$4A,$46,$5F,$F3
	.dc.b	$69,$FE,$72,$E5,$FF,$38,$BF,$E8,$0E,$FF,$F3,$A3,$FE,$99,$EF,$77
	.dc.b	$FF,$98,$75,$FF,$50,$F7,$FF,$BD,$5B,$CB,$FE,$6E,$E1,$FF,$56,$5F
	.dc.b	$F9,$CB,$03,$49,$87,$F9,$8E,$7F,$CD,$7D,$FD,$6B,$FD,$77,$DB,$FE
	.dc.b	$6D,$7F,$CC,$FD,$F9,$7F,$CE,$A7,$F9,$DE,$35,$5C,$A3,$FC,$E7,$97
	.dc.b	$FA,$DB,$BF,$FC,$E3,$1F,$E7,$4B,$FF,$39,$3F,$13,$BB,$FB,$A5,$FF
	.dc.b	$59,$8F,$EB,$87,$F9,$0A,$3F,$B0,$BF,$E3,$8E,$BF,$EE,$6F,$F6,$F9
	.dc.b	$E0,$FF,$98,$63,$FC,$67,$D9,$FD,$BF,$6F,$EE,$D7,$FE,$58,$BF,$ED
	.dc.b	$37,$B9,$FE,$EC,$FF,$8E,$4F,$F1,$4D,$FF,$8E,$5C,$3F,$E0,$47,$81
	.dc.b	$FD,$CC,$CC,$7F,$81,$63,$CB,$1F,$E3,$37,$87,$17,$F8,$62,$38,$46
	.dc.b	$A3,$FC,$56,$4B,$F8,$73,$E2,$AF,$EF,$75,$E3,$38,$4E,$9E,$AF,$C6
	.dc.b	$8F,$F2,$C9,$C0,$93,$F3,$0F,$F3,$13,$A7,$0F,$F9,$05,$F9,$87,$84
	.dc.b	$CF,$7F,$FB,$C2,$BF,$C2,$B1,$F5,$70,$FA,$38,$7B,$63,$EB,$5F,$F8
	.dc.b	$51,$E4,$7C,$83,$27,$13,$31,$C6,$3C,$91,$B6,$C6,$6F,$FD,$0D,$3E
	.dc.b	$A6,$CF,$A6,$3E,$83,$F4,$CF,$15,$FC,$6B,$F3,$E1,$F1,$EF,$89,$ED
	.dc.b	$AF,$C1,$D7,$CE,$AF,$C3,$E1,$E1,$C3,$FC,$6D,$5A,$3A,$9D,$DF,$65
	.dc.b	$77,$7D,$B8,$77,$78,$9F,$82,$9E,$2F,$D0,$9F,$F3,$3C,$7E,$62,$FF
	.dc.b	$CE,$B8,$79,$6B,$81,$DA,$7C,$0F,$C3,$3A,$9F,$38,$C1,$FE,$43,$BF
	.dc.b	$F3,$0E,$07,$FC,$85,$C7,$FE,$44,$C1,$5F,$E4,$4A,$D4,$62,$75,$3F
	.dc.b	$E4,$31,$DD,$9C,$FF,$C9,$04,$8E,$67,$F9,$21,$FE,$AE,$FF,$5A,$8F
	.dc.b	$EB,$18,$FF,$83,$B3,$FE,$AD,$3F,$E0,$79,$FE,$CD,$5F,$DB,$0F,$EC
	.dc.b	$D5,$CA,$B7,$87,$F8,$02,$B9,$DF,$F5,$E3,$FB,$31,$FD,$E3,$6F,$EE
	.dc.b	$D7,$C4,$F0,$12,$38,$4F,$27,$C4,$79,$C9,$7E,$0E,$FF,$F0,$0C,$7F
	.dc.b	$81,$AB,$A3,$C8,$DB,$D5,$E7,$3C,$6F,$C7,$1E,$53,$91,$63,$FC,$02
	.dc.b	$FA,$8F,$F0,$24,$F6,$79,$BD,$70,$F8,$51,$E0,$3C,$53,$DD,$EA,$72
	.dc.b	$3B,$4F,$39,$E3,$3B,$3C,$38,$7C,$38,$CC,$F7,$C3,$E0,$D6,$AF,$C0
	.dc.b	$79,$0F,$0C,$EF,$0F,$7C,$78,$48,$7F,$84,$CF,$91,$8F,$F8,$65,$F8
	.dc.b	$06,$4E,$84,$C3,$C6,$F6,$20,$7F,$CD,$0F,$E7,$31,$FF,$75,$C7,$F9
	.dc.b	$AF,$BB,$FC,$2D,$3F,$A0,$6B,$FE,$1D,$3F,$C3,$07,$F8,$73,$C3,$61
	.dc.b	$FE,$12,$EE,$FF,$0B,$CF,$78,$FF,$0F,$DE,$A3,$C5,$3A,$8C,$47,$57
	.dc.b	$47,$F8,$87,$BF,$FC,$47,$3B,$AF,$F0,$AD,$7F,$86,$3C,$2E,$2E,$08
	.dc.b	$7F,$87,$CF,$F1,$BF,$EC,$2F,$96,$EB,$5F,$E1,$EC,$7B,$CE,$63,$F7
	.dc.b	$F1,$E4,$EE,$FF,$10,$77,$FF,$87,$FC,$0F,$F1,$0E,$3F,$E2,$56,$BF
	.dc.b	$E1,$CF,$03,$FC,$3C,$FE,$FB,$9B,$FC,$3C,$D8,$CB,$FC,$32,$32,$FF
	.dc.b	$88,$4F,$F1,$87,$81,$FE,$20,$7F,$C4,$A3,$04,$66,$3F,$C4,$DC,$FF
	.dc.b	$6F,$B6,$C7,$F1,$26,$F2,$7F,$86,$6D,$E9,$FF,$12,$8C,$61,$FE,$2E
	.dc.b	$6E,$1C,$F0,$F2,$47,$1A,$E1,$87,$42,$4E,$9F,$E3,$B8,$76,$8E,$47
	.dc.b	$A7,$0E,$D3,$C2,$31,$86,$AB,$84,$7A,$4E,$84,$36,$64,$F8,$97,$E4
	.dc.b	$7F,$9B,$F6,$FF,$8B,$9F,$E5,$E6,$3B,$7F,$C5,$EF,$F8,$CD,$FF,$1D
	.dc.b	$41,$A7,$F8,$F7,$B9,$D3,$FE,$43,$E1,$F9,$3C,$B7,$17,$93,$4C,$44
	.dc.b	$CB,$FC,$EB,$3A,$6C,$E5,$FF,$45,$C3,$DC,$E8,$48,$FC,$9F,$2F,$F4
	.dc.b	$8F,$87,$7C,$8E,$67,$8D,$5C,$BF,$57,$5A,$9F,$0F,$B1,$E7,$F6,$24
	.dc.b	$60,$87,$8A,$3D,$87,$B7,$0B,$1A,$BF,$DA,$E4,$AD,$4F,$BB,$CA,$F8
	.dc.b	$84,$7E,$0C,$FC,$72,$38,$3E,$2F,$B1,$F2,$C8,$E0,$F9,$3F,$15,$F4
	.dc.b	$7F,$6B,$FD,$26,$3F,$C1,$9B,$B7,$EE,$1B,$5F,$C7,$DF,$B2,$FF,$A9
	.dc.b	$97,$C5,$F3,$24,$67,$A3,$B7,$C0,$77,$64,$71,$8F,$17,$B5,$23,$A3
	.dc.b	$E5,$F2,$7C,$DE,$DC,$9F,$03,$E9,$F2,$FD,$5E,$0F,$AF,$E6,$E5,$79
	.dc.b	$63,$DD,$F3,$B2,$3D,$0F,$C1,$F6,$3E,$17,$44,$9F,$63,$F1,$7C,$AF
	.dc.b	$8D,$FA,$87,$C7,$F8,$39,$5F,$06,$3E,$57,$B1,$23,$E2,$3F,$37,$E3
	.dc.b	$BE,$77,$BB,$27,$E4,$3F,$47,$E2,$3F,$45,$FB,$67,$EB,$3A,$7F,$9B
	.dc.b	$BF,$C8,$9B,$E7,$6A,$F7,$EB,$47,$FD,$AF,$48,$E7,$F3,$DD,$3D,$F9
	.dc.b	$5A,$7B,$FC,$F7,$83,$D8,$F0,$FE,$11,$2B,$9C,$7F,$AC,$FE,$19,$AD
	.dc.b	$DA,$57,$EA,$8F,$F1,$DC,$E9,$C0,$70,$FF,$59,$37,$3D,$4F,$F6,$05
	.dc.b	$7C,$1F,$07,$FC,$77,$5C,$37,$FE,$B6,$D8,$CD,$70,$BF,$93,$F2,$FF
	.dc.b	$2B,$DF,$91,$F1,$23,$2F,$C6,$8C,$BE,$3C,$1C,$0A,$FA,$7F,$13,$F5
	.dc.b	$3E,$7E,$5F,$CE,$8C,$BF,$42,$32,$F3,$A4,$FD,$CF,$6F,$EE,$FC,$C4
	.dc.b	$AF,$AF,$19,$7E,$CC,$65,$E5,$19,$7A,$46,$71,$FC,$2C,$FE,$23,$E6
	.dc.b	$25,$7E,$09,$FC,$5F,$C9,$62,$3C,$91,$6B,$F1,$B3,$F9,$0F,$9F,$FC
	.dc.b	$8F,$CC,$CB,$FC,$5C,$65,$FC,$6C,$65,$F7,$47,$D8,$E8,$FB,$3D,$32
	.dc.b	$7F,$03,$C9,$F9,$36,$CE,$DA,$7E,$1B,$DE,$FF,$40,$AF,$84,$57,$F9
	.dc.b	$F0,$E5,$1F,$89,$33,$5C,$DC,$57,$FA,$A5,$FF,$49,$67,$FE,$B2,$23
	.dc.b	$FE,$A6,$6D,$FF,$AA,$49,$7F,$E9,$E3,$FD,$7D,$5F,$EE,$79,$FF,$5A
	.dc.b	$F8,$DF,$EA,$DB,$EA,$73,$25,$F2,$7F,$D5,$D7,$CF,$0F,$F5,$64,$7F
	.dc.b	$A9,$E3,$AD,$7F,$BB,$5D,$47,$FA,$92,$BF,$DC,$2F,$FA,$B1,$FF,$7A
	.dc.b	$47,$FA,$EA,$3F,$DA,$27,$62,$18,$26,$1E,$97,$E6,$47,$18,$C4,$FF
	.dc.b	$B4,$67,$27,$FA,$FC,$E7,$1F,$53,$DD,$FF,$59,$93,$D9,$FF,$19,$61
	.dc.b	$FE,$BC,$7F,$D6,$55,$FE,$21,$3F,$D7,$1E,$FF,$FA,$DD,$FF,$03,$B5
	.dc.b	$FF,$39,$9F,$E7,$02,$3F,$E8,$1F,$2F,$FC,$FF,$97,$FA,$1D,$B3,$FE
	.dc.b	$8A,$21,$CA,$78,$75,$7F,$A0,$23,$FC,$F5,$AF,$F9,$E1,$FF,$3D,$57
	.dc.b	$27,$81,$FE,$8A,$8D,$12,$35,$1C,$9A,$F9,$8E,$0E,$2B,$C3,$1F,$EA
	.dc.b	$03,$85,$78,$DF,$F6,$C9,$F3,$8F,$F4,$44,$FF,$9E,$74,$FC,$FC,$7F
	.dc.b	$99,$A3,$FC,$D6,$D9,$FE,$56,$FF,$A3,$4E,$56,$9D,$C7,$49,$FE,$63
	.dc.b	$39,$78,$C7,$8A,$BB,$BB,$3F,$59,$3B,$6F,$97,$D2,$30,$7F,$A3,$03
	.dc.b	$E6,$7F,$94,$FC,$2D,$E8,$7C,$25,$8C,$37,$1F,$EA,$ED,$C3,$F9,$43
	.dc.b	$C7,$E7,$E4,$FF,$5C,$F9,$FB,$5F,$CB,$66,$E3,$FD,$9A,$6D,$7F,$A3
	.dc.b	$BF,$6D,$FF,$54,$BF,$D5,$65,$0F,$A0,$F4,$3A,$C3,$FA,$13,$9A,$E0
	.dc.b	$E6,$BF,$D5,$2E,$67,$73,$CA,$B9,$B4,$E2,$58,$EB,$F2,$3E,$0F,$87
	.dc.b	$F0,$B6,$7B,$B6,$77,$1E,$29,$F3,$FC,$3F,$A3,$E5,$6C,$F9,$9B,$5E
	.dc.b	$A6,$E5,$E8,$7E,$D7,$A9,$2B,$F0,$BC,$F6,$FF,$03,$36,$FE,$F4,$DA
	.dc.b	$57,$FA,$EA,$BF,$09,$EA,$E6,$3F,$D7,$AF,$CC,$F9,$BF,$9B,$F4,$16
	.dc.b	$FE,$5B,$F4,$3E,$52,$47,$CE,$F9,$EB,$87,$D2,$F8,$AF,$AA,$7D,$6F
	.dc.b	$89,$B5,$F5,$3E,$27,$EC,$FC,$4F,$DA,$6C,$7D,$8F,$87,$FB,$A7,$DE
	.dc.b	$F8,$9B,$1F,$73,$E2,$7F,$02,$7E,$0B,$E2,$6C,$7E,$03,$E2,$7F,$0B
	.dc.b	$F1,$3F,$86,$6C,$7E,$13,$E2,$7F,$12,$7E,$87,$BD,$E4,$62,$7F,$21
	.dc.b	$1F,$E9,$E6,$F9,$1B,$F8,$8E,$0F,$1F,$88,$E5,$F1,$16,$B8,$BD,$7E
	.dc.b	$22,$47,$4F,$2A,$F3,$7F,$86,$72,$FD,$9C,$F8,$3C,$9F,$27,$77,$95
	.dc.b	$F3,$77,$79,$FB,$AD,$79,$9F,$57,$C4,$48,$C1,$EA,$7D,$DF,$11,$23
	.dc.b	$DB,$F1,$17,$3F,$17,$C4,$7C,$33,$E2,$7C,$3D,$AF,$87,$F1,$1F,$1C
	.dc.b	$C5,$7A,$5F,$95,$F1,$12,$76,$7C,$A7,$E7,$7C,$44,$8F,$9B,$F0,$96
	.dc.b	$FE,$14,$DF,$21,$5F,$98,$7F,$89,$D6,$63,$1C,$4A,$FB,$5F,$07,$9C
	.dc.b	$18,$7F,$1A,$8F,$E3,$DF,$B7,$7E,$8B,$0F,$E1,$8D,$AF,$F0,$A6,$DF
	.dc.b	$EC,$F9,$CC,$70,$23,$FE,$64,$8D,$47,$F9,$46,$79,$5F,$A2,$3C,$30
	.dc.b	$F6,$BC,$3E,$D7,$89,$B8,$78,$9B,$5C,$DB,$7C,$E1,$F2,$1F,$CD,$2B
	.dc.b	$FC,$EA,$D8,$FF,$31,$7C,$AB,$C3,$FE,$97,$39,$BE,$4E,$EB,$FE,$00
	.dc.b	$B5,$FE,$0A,$76,$DB,$C0,$E2,$34,$FF,$0A,$F8,$F7,$A5,$ED,$87,$B6
	.dc.b	$F0,$39,$D6,$B5,$FE,$23,$39,$AF,$E2,$7F,$18,$F8,$6F,$91,$39,$95
	.dc.b	$F1,$7A,$E1,$D3,$1F,$05,$6E,$B4,$F1,$C3,$A6,$FE,$0B,$E1,$95,$C7
	.dc.b	$D5,$8F,$29,$CC,$76,$3C,$B9,$79,$23,$38,$F7,$AE,$15,$CA,$FA,$2C
	.dc.b	$7C,$2F,$33,$D7,$3C,$1B,$3C,$32,$F4,$E1,$F4,$AF,$8E,$1D,$EB,$58
	.dc.b	$7C,$FC,$3E,$B9,$C8,$A5,$33,$87,$C9,$77,$EF,$BB,$9F,$2B,$C2,$53
	.dc.b	$98,$C3,$63,$AB,$5E,$BE,$BF,$F2,$87,$C3,$68,$C4,$6B,$C8,$ED,$39
	.dc.b	$6C,$F1,$CF,$ED,$37,$5F,$C4,$B6,$FF,$21,$E5,$7E,$5A,$B7,$7F,$26
	.dc.b	$B5,$5F,$88,$6E,$0B,$E1,$87,$4B,$E0,$4B,$FA,$D5,$B7,$18,$76,$73
	.dc.b	$87,$4C,$39,$C7,$E8,$E3,$B9,$2F,$EE,$4E,$A5,$F2,$0F,$4F,$F5,$FC
	.dc.b	$BF,$BD,$F9,$0F,$D0,$37,$4C,$F2,$7E,$D2,$F8,$61,$FE,$F4,$6B,$F8
	.dc.b	$06,$BF,$84,$6B,$C7,$0F,$BD,$7F,$0F,$D2,$EB,$3F,$C9,$B2,$46,$27
	.dc.b	$53,$FE,$31,$97,$A9,$96,$C6,$6B,$06,$67,$5E,$A7,$82,$8D,$25,$7E
	.dc.b	$62,$FE,$B4,$70,$8F,$03,$E6,$5D,$CF,$4B,$C5,$EA,$78,$FD,$49,$5E
	.dc.b	$78,$CB,$D7,$14,$78,$C9,$19,$76,$FF,$48,$DB,$FA,$84,$F0,$8F,$85
	.dc.b	$7F,$07,$BA,$47,$C5,$7E,$24,$C8,$F0,$3F,$04,$8F,$8F,$D4,$F8,$84
	.dc.b	$70,$7C,$3F,$90,$FD,$4E,$BF,$D2,$C3,$8C,$7F,$B2,$36,$7F,$AF,$AF
	.dc.b	$4F,$ED,$2F,$53,$F0,$2D,$9E,$D1,$83,$3E,$A7,$0E,$59,$3F,$E7,$1A
	.dc.b	$4F,$4F,$52,$4E,$5E,$FE,$A7,$83,$D3,$EE,$49,$F6,$3E,$2F,$03,$CB
	.dc.b	$E9,$C9,$ED,$1E,$8F,$4E,$47,$85,$F5,$FA,$7E,$CE,$7F,$59,$27,$E4
	.dc.b	$3E,$EF,$4B,$E0,$FA,$52,$7D,$51,$F1,$3D,$2F,$CF,$79,$FF,$CD,$E7
	.dc.b	$FC,$07,$E5,$65,$F8,$1E,$97,$C8,$23,$F5,$5F,$D5,$FC,$0F,$F3,$2E
	.dc.b	$8F,$AC,$FF,$35,$6F,$99,$FA,$B8,$D7,$4F,$F8,$6D,$7A,$3E,$2F,$90
	.dc.b	$98,$FF,$A7,$E7,$2D,$8E,$E4,$7F,$51,$08,$F1,$FC,$27,$7F,$8F,$95
	.dc.b	$CE,$3F,$55,$F1,$EF,$5A,$EB,$2B,$FC,$5D,$9F,$F8,$0A,$B5,$18,$3F
	.dc.b	$D1,$2E,$1F,$E9,$ED,$7F,$A9,$B5,$E5,$5F,$D3,$54,$71,$D7,$FC,$38
	.dc.b	$DA,$DC,$F7,$3F,$DB,$A7,$EC,$CF,$F1,$DE,$6F,$8E,$F9,$9F,$1F,$27
	.dc.b	$E4,$47,$CF,$30,$F1,$C1,$4A,$D4,$FC,$FF,$8F,$FA,$DE,$B4,$8C,$10
	.dc.b	$E5,$5F,$E1,$77,$EB,$47,$DD,$9F,$BD,$F1,$DF,$7F,$E3,$E5,$7D,$B8
	.dc.b	$CB,$F7,$63,$2F,$48,$CB,$DA,$33,$8F,$E3,$27,$F1,$5F,$1F,$2B,$F0
	.dc.b	$CF,$E3,$7E,$3F,$11,$E6,$8B,$5F,$94,$9F,$C9,$7C,$77,$E4,$FE,$3E
	.dc.b	$5F,$E3,$E3,$2F,$E4,$63,$2F,$C0,$8F,$E6,$5C,$7F,$94,$8F,$A8,$1F
	.dc.b	$D0,$70,$7F,$92,$9B,$1F,$ED,$07,$86,$1F,$E5,$B7,$94,$7F,$86,$9A
	.dc.b	$F1,$E6,$FC,$44,$62,$39,$47,$88,$8F,$E1,$FD,$A7,$C1,$7F,$A0,$21
	.dc.b	$F1,$FE,$57,$FD,$C3,$3A,$C3,$FD,$C7,$39,$9F,$55,$FF,$98,$A3,$18
	.dc.b	$7B,$DD,$5F,$1A,$EB,$3F,$E6,$88,$C4,$F0,$AE,$53,$EB,$AE,$EE,$F0
	.dc.b	$F3,$57,$46,$C6,$E7,$A9,$F0,$0E,$27,$AA,$36,$DC,$32,$FA,$8B,$3E
	.dc.b	$93,$C1,$1A,$C3,$CF,$1D,$EF,$D1,$59,$BE,$AF,$F9,$11,$F9,$9E,$66
	.dc.b	$A3,$FA,$75,$79,$E3,$D5,$82,$72,$79,$4F,$59,$E5,$2C,$F9,$5D,$5F
	.dc.b	$27,$38,$FC,$A7,$E1,$1C,$B1,$F4,$BC,$2B,$D4,$47,$04,$3C,$15,$E9
	.dc.b	$3B,$12,$38,$46,$9F,$8E,$F7,$C7,$E0,$C7,$BC,$91,$ED,$C7,$93,$E6
	.dc.b	$AE,$D3,$C9,$CC,$75,$35,$19,$23,$82,$1F,$47,$1F,$A8,$7C,$DA,$FA
	.dc.b	$44,$9F,$3B,$D9,$F9,$93,$DC,$E4,$7A,$E7,$94,$7A,$A3,$EA,$4F,$D1
	.dc.b	$39,$63,$F4,$63,$96,$3F,$41,$E7,$87,$B9,$F3,$C6,$4D,$5F,$88,$F1
	.dc.b	$B9,$C7,$E5,$CF,$DC,$C3,$D4,$FC,$BA,$F8,$6F,$A5,$E3,$3F,$94,$AE
	.dc.b	$B8,$79,$67,$38,$FC,$6B,$FC,$89,$EC,$8F,$75,$7E,$E7,$0F,$6C,$FE
	.dc.b	$9A,$FD,$D3,$E5,$BC,$4C,$3E,$41,$2B,$CE,$F5,$3D,$07,$E3,$EF,$E0
	.dc.b	$5F,$4C,$3D,$07,$EB,$23,$33,$E7,$C7,$C6,$FE,$80,$C4,$7D,$C3,$EB
	.dc.b	$1F,$CC,$6F,$EB,$3F,$18,$E7,$1E,$69,$CB,$F3,$63,$E0,$D7,$D0,$8F
	.dc.b	$C1,$D7,$AA,$7A,$1D,$E3,$F7,$1E,$CF,$F9,$2D,$B3,$FA,$48,$C4,$7E
	.dc.b	$31,$AF,$FB,$F8,$38,$5E,$DF,$F0,$58,$C3,$67,$FD,$B2,$D9,$FF,$6E
	.dc.b	$BF,$96,$7F,$86,$64,$F2,$5A,$E6,$7F,$D8,$87,$E0,$27,$51,$98,$F2
	.dc.b	$5F,$F3,$F9,$CC,$7E,$18,$CE,$3F,$EB,$4B,$EB,$19,$3D,$27,$F1,$A9
	.dc.b	$E3,$9F,$F8,$5A,$32,$F5,$9E,$2E,$71,$F3,$57,$83,$0C,$5F,$9E,$FB
	.dc.b	$C6,$2F,$96,$3F,$CF,$2B,$E6,$46,$1F,$E5,$66,$31,$F4,$4E,$E3,$C7
	.dc.b	$18,$76,$60,$F9,$77,$E6,$8F,$A3,$1E,$D8,$C3,$F4,$8F,$34,$FC,$B7
	.dc.b	$E1,$DF,$8C,$F4,$57,$7B,$B3,$DA,$BC,$98,$FF,$56,$C7,$E4,$DF,$82
	.dc.b	$FD,$87,$C4,$C3,$A4,$79,$1B,$87,$8E,$77,$3C,$4F,$01,$63,$C0,$7C
	.dc.b	$0B,$F0,$63,$EC,$8F,$C3,$DF,$46,$D7,$C7,$8F,$4C,$30,$F3,$BE,$19
	.dc.b	$EE,$D3,$BE,$1C,$27,$71,$F9,$48,$F3,$BB,$3A,$5F,$8D,$FC,$89,$E8
	.dc.b	$BF,$85,$7E,$33,$D9,$1C,$31,$F9,$6F,$D7,$6C,$FA,$E3,$99,$B3,$F3
	.dc.b	$58,$FC,$29,$E3,$87,$92,$7C,$D7,$EC,$69,$EC,$7C,$38,$FB,$23,$F2
	.dc.b	$06,$4F,$C3,$61,$E1,$7F,$0B,$3E,$97,$E0,$4F,$17,$F2,$47,$96,$BE
	.dc.b	$46,$3F,$0E,$33,$3B,$C7,$D1,$1F,$6C,$F8,$05,$AF,$53,$E4,$3A,$5F
	.dc.b	$A2,$FA,$57,$2A,$F9,$B5,$F3,$EB,$F1,$C7,$9D,$F8,$27,$23,$D3,$5C
	.dc.b	$9F,$0C,$FA,$31,$F9,$75,$F2,$67,$C5,$87,$A5,$DC,$78,$E7,$71,$F2
	.dc.b	$4E,$F3,$F8,$83,$F8,$A9,$FC,$49,$FB,$4D,$7A,$DF,$F1,$68,$D1,$FC
	.dc.b	$72,$3F,$DA,$91,$B7,$F4,$66,$27,$FC,$6C,$D3,$FC,$F7,$2F,$27,$A4
	.dc.b	$FF,$09,$8E,$B1,$89,$FF,$59,$FC,$7D,$9F,$F5,$6F,$C0,$DE,$94,$D8
	.dc.b	$C3,$70,$FF,$6F,$36,$FF,$D5,$BF,$1F,$C4,$E5,$F1,$F6,$BF,$DC,$91
	.dc.b	$67,$FD,$67,$37,$1F,$85,$F1,$FF,$D1,$2F,$FD,$B1,$08,$7C,$7F,$9D
	.dc.b	$D5,$FF,$B1,$5C,$D7,$07,$35,$D2,$B8,$4F,$16,$BB,$6D,$73,$7B,$7C
	.dc.b	$7F,$7F,$85,$B5,$D5,$B5,$E1,$6D,$FA,$9F,$37,$C1,$F9,$FE,$3E,$DF
	.dc.b	$95,$B7,$E9,$6D,$FC,$27,$EB,$47,$F1,$9F,$4A,$C9,$BD,$18,$65,$FE
	.dc.b	$12,$BF,$0B,$5B,$8C,$47,$F7,$29,$FC,$63,$D6,$32,$D3,$F1,$7E,$9F
	.dc.b	$46,$C7,$CD,$F8,$FF,$A5,$F1,$FF,$4D,$B1,$F4,$7E,$3F,$F3,$18,$FD
	.dc.b	$96,$F0,$1A,$63,$E1,$7F,$25,$E8,$BE,$43,$EF,$46,$1F,$C0,$7C,$37
	.dc.b	$E0,$7E,$16,$E5,$F7,$DF,$C2,$7C,$27,$62,$47,$C0,$CF,$F1,$51,$E2
	.dc.b	$F8,$4E,$FD,$FF,$8B,$6D,$7E,$27,$E0,$FF,$1D,$F0,$7F,$8F,$6C,$7E
	.dc.b	$37,$E3,$FD,$64,$7B,$70,$23,$E0,$F7,$BC,$5F,$D7,$36,$BF,$3A,$BE
	.dc.b	$13,$33,$AF,$84,$DF,$C1,$DB,$DB,$CB,$E0,$FA,$92,$3D,$D8,$23,$C3
	.dc.b	$E3,$78,$0F,$0F,$6D,$AF,$0F,$6F,$93,$B7,$CA,$D8,$F2,$FC,$1F,$A8
	.dc.b	$8F,$67,$DB,$F0,$7E,$C3,$DB,$F0,$76,$7D,$BF,$1F,$F0,$88,$F9,$9F
	.dc.b	$8B,$F0,$92,$3D,$2D,$F0,$BF,$35,$D4,$DD,$F0,$7F,$28,$F9,$75,$6F
	.dc.b	$E5,$FC,$7F,$CE,$3E,$7F,$C0,$58,$C9,$F1,$A3,$EA,$91,$C1,$F5,$3E
	.dc.b	$03,$86,$5F,$E0,$98,$FF,$2C,$E2,$9F,$CB,$CC,$9F,$F2,$E9,$FC,$D3
	.dc.b	$47,$7E,$67,$91,$FE,$13,$F0,$FF,$DD,$D5,$F9,$56,$C7,$C0,$6B,$FB
	.dc.b	$A6,$BF,$E6,$A5,$F0,$7F,$A8,$F5,$7F,$52,$DA,$FE,$53,$29,$C5,$78
	.dc.b	$B7,$4F,$2A,$DD,$7F,$10,$F8,$EF,$A1,$1F,$C1,$1B,$7F,$C2,$9A,$FF
	.dc.b	$0F,$6B,$FD,$51,$DD,$FF,$5D,$F9,$27,$17,$5D,$BF,$A8,$46,$30,$F9
	.dc.b	$38,$07,$C7,$3C,$EF,$7C,$7F,$A4,$BB,$7F,$B0,$CE,$76,$E1,$EB,$A7
	.dc.b	$FB,$2F,$C3,$BE,$0B,$FD,$07,$39,$C3,$B6,$1C,$5E,$86,$31,$EF,$3B
	.dc.b	$8F,$1C,$63,$0E,$B7,$D1,$F1,$C6,$67,$96,$3E,$49,$CB,$63,$DD,$8F
	.dc.b	$97,$2C,$47,$69,$B9,$79,$72,$F1,$37,$0F,$0D,$EB,$1F,$06,$1C,$4F
	.dc.b	$59,$6F,$D5,$8F,$78,$CE,$1C,$DF,$39,$FD,$19,$F8,$8F,$F4,$B6,$BE
	.dc.b	$A3,$FA,$5F,$B7,$FD,$4F,$E3,$FF,$A4,$B6,$33,$DA,$9C,$B6,$3B,$7A
	.dc.b	$7F,$EC,$E6,$FB,$9F,$D4,$B0,$4E,$67,$C4,$D7,$F0,$2F,$1C,$BE,$2B
	.dc.b	$75,$CD,$FC,$AF,$27,$FC,$4F,$5A,$8F,$9B,$5B,$BF,$C1,$B5,$FC,$8D
	.dc.b	$70,$BE,$71,$F2,$27,$58,$79,$AF,$B4,$F4,$63,$B7,$2D,$26,$1E,$37
	.dc.b	$E9,$E1,$DD,$FE,$0D,$F0,$FF,$DE,$B7,$FE,$01,$F0,$D6,$7F,$85,$37
	.dc.b	$9B,$F7,$75,$7A,$4D,$25,$7F,$04,$F8,$6E,$4B,$E3,$9E,$A7,$F8,$0A
	.dc.b	$30,$D8,$CC,$7F,$81,$9A,$FF,$83,$1B,$87,$F9,$7D,$B9,$7F,$95,$9B
	.dc.b	$3F,$BD,$F8,$FF,$EC,$8F,$F0,$01,$F8,$27,$F8,$05,$C3,$5F,$E5,$26
	.dc.b	$BD,$6F,$77,$C4,$FA,$A7,$58,$FF,$9B,$DB,$8F,$F9,$99,$B7,$FC,$C6
	.dc.b	$BE,$37,$5F,$C7,$F8,$89,$3F,$CA,$23,$2D,$21,$FC,$A7,$77,$C4,$EA
	.dc.b	$31,$37,$2F,$F3,$E5,$6A,$7F,$99,$FC,$4F,$D5,$23,$FC,$DB,$E2,$7E
	.dc.b	$C1,$1C,$1F,$CE,$BE,$23,$EE,$11,$C1,$FC,$FB,$E2,$7F,$01,$F0,$D2
	.dc.b	$3F,$A0,$BF,$83,$F1,$3F,$09,$F0,$D2,$7F,$A2,$BF,$87,$F8,$6F,$C5
	.dc.b	$7E,$DF,$FE,$92,$E3,$FE,$65,$B2,$31,$7F,$EE,$B6,$C6,$7E,$1F,$8F
	.dc.b	$C3,$CA,$FF,$3C,$D2,$3A,$F8,$FD,$BC,$29,$5A,$7C,$3F,$0F,$E2,$F8
	.dc.b	$6F,$1F,$C3,$CA,$E7,$1E,$6F,$87,$91,$D9,$F4,$FC,$9F,$AB,$C0,$93
	.dc.b	$E2,$7D,$BF,$25,$EE,$F8,$9F,$7F,$A9,$2B,$CF,$1F,$0B,$E2,$64,$7A
	.dc.b	$9F,$8B,$F2,$5F,$1B,$E2,$64,$FB,$9F,$93,$F2,$5F,$2B,$CB,$93,$F0
	.dc.b	$9F,$9B,$F1,$5F,$3B,$E2,$64,$FC,$67,$F7,$5F,$B8,$7F,$1F,$E9,$FF
	.dc.b	$62,$3F,$A6,$9F,$EF,$0D,$8C,$FC,$4F,$0F,$8A,$E3,$EA,$4B,$FF,$54
	.dc.b	$DF,$F5,$9B,$E7,$F1,$5F,$C0,$3E,$43,$1D,$FF,$7A,$BF,$C9,$17,$C3
	.dc.b	$E2,$F5,$8C,$8F,$DB,$5F,$07,$FB,$B3,$71,$5B,$E3,$F8,$57,$EE,$2B
	.dc.b	$82,$78,$4E,$E3,$CF,$F1,$9F,$08,$8F,$AB,$E3,$2B,$F0,$A2,$E1,$F1
	.dc.b	$A7,$E4,$FB,$72,$BE,$0B,$F3,$3E,$33,$11,$F2,$A2,$D7,$CE,$9F,$A3
	.dc.b	$F1,$9F,$4B,$E3,$25,$FC,$A8,$CB,$48,$7D,$4F,$87,$CB,$F5,$A2,$DF
	.dc.b	$DA,$9F,$BB,$F0,$B2,$BE,$AB,$F8,$0F,$92,$C4,$7D,$E8,$B5,$F8,$29
	.dc.b	$FC,$2F,$C6,$7E,$1B,$E3,$65,$FD,$E8,$CB,$48,$7E,$23,$E3,$3F,$01
	.dc.b	$C5,$FD,$1F,$F5,$B7,$AD,$4C,$D7,$DE,$9A,$FD,$97,$F1,$1C,$38,$CD
	.dc.b	$CF,$FE,$37,$AF,$B9,$59,$8C,$72,$AF,$F3,$BF,$1D,$2F,$F2,$AF,$E8
	.dc.b	$3E,$3B,$11,$F9,$E8,$B8,$FE,$8A,$7F,$AF,$47,$FA,$13,$0F,$E2,$AD
	.dc.b	$A5,$23,$0D,$F8,$BE,$33,$0F,$F2,$7C,$6D,$EB,$1F,$D9,$3F,$44,$7A
	.dc.b	$BE,$0F,$8F,$95,$FD,$5E,$72,$D9,$E0,$47,$AF,$B5,$FA,$C7,$FA,$CC
	.dc.b	$F0,$6C,$ED,$FF,$3A,$4E,$F0,$F9,$B1,$EC,$C5,$F1,$BF,$08,$91,$EA
	.dc.b	$87,$CA,$62,$3F,$CF,$D1,$C1,$D4,$FC,$A9,$F9,$7F,$09,$27,$93,$F3
	.dc.b	$BE,$4B,$11,$CA,$2E,$1E,$43,$E7,$4F,$D4,$F8,$99,$5E,$37,$EB,$FC
	.dc.b	$56,$23,$C7,$16,$B9,$64,$AF,$BB,$F1,$52,$FD,$8F,$E0,$3E,$1F,$11
	.dc.b	$EC,$8B,$7C,$A7,$F0,$BF,$0B,$2B,$E2,$3F,$88,$F8,$7C,$47,$C4,$8B
	.dc.b	$5E,$29,$FF,$11,$FC,$7F,$E5,$E3,$FD,$25,$F2,$8D,$7F,$3C,$7F,$AA
	.dc.b	$F0,$CB,$8E,$4F,$D1,$4E,$E7,$FD,$35,$7C,$D4,$65,$46,$63,$15,$D6
	.dc.b	$76,$F5,$7B,$9D,$3A,$3F,$3E,$E8,$C6,$1A,$7D,$18,$79,$2B,$94,$6C
	.dc.b	$F0,$D7,$2B,$F2,$61,$E3,$AD,$C7,$92,$F9,$5F,$63,$95,$6D,$FF,$5C
	.dc.b	$57,$A6,$B8,$47,$8E,$74,$F8,$30,$F3,$56,$CE,$E7,$F0,$B9,$D9,$C1
	.dc.b	$D4,$F6,$BF,$1B,$CC,$F1,$E1,$EB,$9E,$71,$B3,$C5,$19,$7D,$58,$79
	.dc.b	$9F,$89,$3D,$9E,$91,$F1,$27,$84,$7C,$C3,$D3,$5E,$2B,$F3,$CE,$AF
	.dc.b	$91,$CE,$33,$87,$0B,$87,$A1,$F2,$98,$C3,$C8,$78,$5C,$C7,$9B,$0F
	.dc.b	$83,$7D,$70,$F2,$E1,$F1,$E7,$27,$96,$32,$79,$5C,$D7,$4C,$3B,$3D
	.dc.b	$4F,$15,$FD,$3A,$E3,$86,$27,$E6,$1E,$92,$7C,$77,$F8,$5F,$1F,$D0
	.dc.b	$36,$BF,$2E,$DA,$FF,$7F,$3F,$E1,$DA,$FF,$7C,$87,$0B,$E3,$B7,$E7
	.dc.b	$FF,$4F,$8D,$CE,$A7,$31,$97,$F4,$55,$FE,$13,$EA,$FD,$24,$F6,$CF
	.dc.b	$F6,$53,$98,$FD,$21,$9C,$FF,$AE,$46,$67,$F7,$13,$93,$FC,$CB,$DB
	.dc.b	$FC,$DA,$33,$1E,$73,$F5,$D3,$C6,$BD,$11,$99,$F0,$D7,$19,$CE,$1E
	.dc.b	$88,$C5,$7A,$CE,$6E,$1F,$F6,$D9,$8C,$3D,$AF,$5A,$F4,$CE,$E3,$D0
	.dc.b	$7F,$66,$70,$EC,$F5,$CF,$2A,$F1,$36,$3B,$1F,$18,$E0,$73,$75,$7E
	.dc.b	$C6,$DF,$89,$B3,$D8,$E5,$5D,$9A,$76,$AF,$97,$19,$6C,$66,$BE,$3B
	.dc.b	$4F,$8D,$1E,$17,$9B,$4E,$25,$8E,$2D,$8E,$31,$E7,$78,$B6,$38,$B4
	.dc.b	$E2,$58,$E2,$7B,$63,$6D,$8D,$B4,$D9,$F2,$4E,$E6,$70,$FF,$0D,$7C
	.dc.b	$5F,$F6,$FC,$3F,$C3,$ED,$9F,$F0,$27,$7D,$AD,$D7,$F8,$0E,$B1,$CB
	.dc.b	$FE,$D6,$8F,$F0,$35,$62,$73,$F1,$52,$7F,$C1,$5D,$77,$1F,$F0,$4C
	.dc.b	$DC,$7F,$BB,$CD,$AF,$F1,$8F,$C5,$76,$3B,$FC,$55,$8F,$F0,$A4,$5C
	.dc.b	$3F,$C2,$73,$6F,$FC,$03,$F1,$5E,$63,$CF,$DF,$70,$FF,$0E,$45,$BF
	.dc.b	$F0,$DC,$DA,$FF,$04,$FC,$57,$B8,$F8,$5F,$B8,$FF,$95,$30,$FF,$12
	.dc.b	$46,$5A,$7F,$90,$23,$57,$FE,$26,$9F,$98,$BF,$44,$7C,$36,$1B,$7F
	.dc.b	$E4,$8F,$D7,$1B,$66,$A4,$FF,$BA,$0F,$F3,$25,$70,$BF,$F2,$AC,$7F
	.dc.b	$BB,$2F,$8B,$B8,$1E,$1A,$E6,$F0,$3F,$CB,$B8,$BE,$37,$39,$FF,$8C
	.dc.b	$E3,$17,$ED,$F8,$CF,$86,$7C,$4F,$8C,$B1,$92,$61,$D6,$D6,$AD,$CE
	.dc.b	$61,$F1,$9F,$38,$F9,$FF,$19,$70,$F9,$91,$6B,$84,$DC,$3F,$CB,$70
	.dc.b	$FB,$17,$83,$F4,$1B,$87,$C3,$9D,$19,$99,$19,$9D,$CF,$C5,$37,$19
	.dc.b	$9D,$C7,$F9,$B3,$B6,$DF,$F9,$67,$87,$FA,$03,$F2,$7E,$37,$F1,$27
	.dc.b	$E2,$B8,$2E,$3F,$87,$8F,$C6,$FC,$6E,$31,$F8,$8F,$E4,$3E,$36,$57
	.dc.b	$E3,$7E,$36,$D7,$C9,$9B,$8F,$D7,$9B,$5F,$80,$8F,$A3,$F1,$BF,$9A
	.dc.b	$3F,$37,$F1,$96,$7F,$31,$17,$0F,$A9,$36,$BF,$05,$3F,$A2,$F8,$DC
	.dc.b	$5F,$5E,$34,$7E,$9B,$E3,$64,$FE,$8F,$E3,$6E,$5F,$7E,$6E,$3F,$8D
	.dc.b	$9B,$5C,$FE,$37,$F5,$E7,$EC,$3E,$36,$C7,$EB,$62,$E1,$F8,$99,$B7
	.dc.b	$F8,$2E,$57,$A7,$F8,$13,$F7,$3F,$1B,$71,$FB,$D1,$70,$FC,$94,$DB
	.dc.b	$E5,$3F,$C0,$3E,$37,$17,$EF,$7F,$81,$FC,$6C,$8F,$F5,$FF,$EE,$5F
	.dc.b	$E0,$D7,$F8,$14,$7F,$29,$87,$9F,$4C,$7C,$75,$AE,$3F,$19,$B3,$87
	.dc.b	$C7,$58,$EB,$F1,$DC,$CE,$9D,$D6,$3C,$3F,$1B,$DC,$F0,$7A,$B6,$3C
	.dc.b	$BF,$1B,$E3,$3C,$9F,$1B,$63,$D3,$F1,$9E,$7F,$8D,$F4,$36,$3D,$BF
	.dc.b	$1B,$EB,$3D,$9F,$1B,$63,$E0,$FC,$6F,$BC,$F8,$1F,$19,$63,$E2,$FC
	.dc.b	$6F,$C3,$F8,$DF,$88,$D8,$F9,$3E,$DF,$C7,$3E,$47,$C5,$D8,$FF,$75
	.dc.b	$FE,$E7,$FC,$69,$BD,$DF,$D5,$61,$F1,$96,$31,$E7,$CA,$E3,$F1,$76
	.dc.b	$7F,$8A,$5F,$FB,$72,$7A,$FC,$5F,$3F,$03,$A3,$D7,$E0,$64,$F8,$BE
	.dc.b	$2E,$E1,$CB,$E2,$FC,$47,$8F,$E2,$ED,$7A,$3E,$2F,$CC,$62,$BC,$2F
	.dc.b	$A7,$E2,$E4,$EC,$F7,$BE,$DF,$8B,$91,$F0,$BE,$1A,$DF,$AA,$6F,$93
	.dc.b	$FE,$E9,$DF,$FB,$A4,$E6,$FE,$4D,$7C,$7F,$87,$DC,$7F,$05,$A4,$7C
	.dc.b	$BF,$87,$93,$F4,$3D,$BB,$87,$C8,$9B,$8F,$79,$B5,$FD,$3A,$9F,$30
	.dc.b	$F8,$6F,$EC,$2D,$9F,$EA,$96,$F9,$C7,$A5,$AB,$95,$FE,$5C,$C3,$9C
	.dc.b	$7F,$3E,$7B,$61,$CD,$FF,$30,$B5,$FF,$36,$E8,$F8,$9F,$C3,$1F,$87
	.dc.b	$F8,$9B,$87,$E3,$23,$F1,$58,$62,$FF,$0A,$FF,$78,$F9,$7F,$EB,$1D
	.dc.b	$9F,$DA,$BE,$31,$E7,$FB,$E7,$37,$FE,$8B,$8F,$85,$DA,$E6,$67,$C6
	.dc.b	$78,$7C,$2F,$E6,$2F,$81,$FD,$DB,$0E,$78,$F6,$8C,$4E,$E3,$10,$C7
	.dc.b	$B6,$1D,$63,$79,$55,$87,$4C,$7C,$33,$9B,$E3,$5E,$23,$78,$F7,$AD
	.dc.b	$C6,$6F,$93,$FE,$0C,$AE,$38,$F9,$27,$33,$CB,$0F,$14,$62,$B7,$7D
	.dc.b	$2F,$86,$1F,$0F,$0F,$F5,$95,$FC,$C6,$C6,$5F,$19,$FE,$B7,$F8,$E7
	.dc.b	$98,$F9,$8F,$C4,$46,$1B,$19,$9F,$F5,$3C,$E5,$FF,$54,$DF,$F8,$4B
	.dc.b	$E3,$FF,$20,$D8,$CF,$C4,$7F,$B2,$5B,$3A,$8C,$65,$F0,$5B,$9F,$E1
	.dc.b	$5B,$5F,$77,$DA,$FD,$35,$6E,$FE,$2D,$6A,$BE,$BB,$70,$5F,$0C,$3A
	.dc.b	$5F,$02,$5F,$D5,$9D,$3F,$8B,$BE,$44,$9F,$A7,$3A,$9E,$F1,$CB,$0F
	.dc.b	$24,$73,$3F,$19,$D0,$F9,$57,$85,$FE,$DA,$E2,$FF,$07,$43,$C2,$C5
	.dc.b	$AF,$84,$61,$E5,$76,$B9,$D7,$F0,$47,$93,$6B,$53,$D2,$1F,$30,$F8
	.dc.b	$5C,$7C,$C7,$E9,$31,$FF,$21,$B7,$2F,$E2,$95,$BB,$FF,$21,$35,$FF
	.dc.b	$74,$B5,$FE,$34,$76,$BF,$E2,$4D,$3F,$8A,$7C,$C7,$F8,$39,$BE,$4E
	.dc.b	$A3,$11,$FE,$F4,$6B,$D1,$CE,$3F,$E5,$86,$E1,$FC,$A5,$B3,$F2,$74
	.dc.b	$78,$DF,$1F,$AB,$E4,$78,$B2,$BE,$44,$65,$F4,$9C,$5F,$E5,$58,$FD
	.dc.b	$96,$DF,$F1,$59,$DC,$E5,$E6,$43,$FC,$F5,$0F,$1B,$EC,$78,$D2,$7E
	.dc.b	$C3,$F6,$FC,$69,$1D,$29,$1E,$02,$46,$7C,$6D,$BD,$A3,$F0,$91,$F8
	.dc.b	$5F,$0A,$4F,$E1,$5F,$C4,$78,$3F,$89,$F9,$2F,$59,$27,$C0,$FC,$0F
	.dc.b	$98,$FF,$42,$E9,$FC,$CA,$7E,$67,$C9,$FE,$44,$FC,$A4,$EB,$8B,$FC
	.dc.b	$D0,$BE,$43,$24,$AF,$EB,$31,$FE,$9B,$6F,$6F,$72,$3F,$32,$3F,$D3
	.dc.b	$AD,$F8,$EE,$FF,$1F,$23,$9C,$78,$BE,$3E,$47,$67,$CB,$F1,$FE,$6E
	.dc.b	$AF,$61,$27,$B3,$EF,$F8,$FF,$5F,$C6,$C9,$F3,$C7,$BB,$E3,$7F,$65
	.dc.b	$E7,$38,$A9,$9A,$F6,$FC,$67,$C9,$23,$EB,$7E,$67,$8D,$23,$DC,$D7
	.dc.b	$F5,$3C,$8E,$07,$C5,$78,$39,$65,$7C,$A8,$FE,$09,$F3,$1F,$E6,$68
	.dc.b	$FC,$9B,$5E,$6D,$7F,$C1,$D6,$8C,$47,$FA,$7D,$FE,$00,$FF,$11,$6C
	.dc.b	$6F,$C9,$EF,$F1,$1F,$C2,$FE,$5F,$34,$F8,$9F,$19,$9A,$98,$7F,$84
	.dc.b	$EB,$FB,$8B,$FC,$19,$AF,$F8,$85,$AF,$FB,$C1,$F1,$A6,$9E,$0F,$89
	.dc.b	$F8,$3F,$13,$27,$FC,$4B,$82,$3E,$2F,$C4,$E6,$7F,$DD,$93,$97,$73
	.dc.b	$A7,$8F,$A5,$FC,$BF,$E3,$7F,$39,$1F,$CC,$DB,$AF,$C9,$9F,$A2,$49
	.dc.b	$E2,$78,$BE,$2B,$EA,$FC,$4F,$D6,$F8,$A9,$3E,$68,$CB,$F6,$1F,$E1
	.dc.b	$F9,$1A,$48,$E0,$43,$D7,$F1,$5F,$82,$F8,$AF,$47,$C9,$C9,$F8,$71
	.dc.b	$96,$90,$F8,$7F,$17,$F8,$AF,$8B,$FC,$5F,$C5,$49,$F9,$71,$96,$90
	.dc.b	$F9,$7F,$17,$F8,$3E,$3F,$EA,$DF,$31,$FE,$64,$9F,$F2,$BB,$5F,$EB
	.dc.b	$9E,$57,$DF,$9E,$11,$F8,$57,$FA,$3F,$31,$A4,$8F,$F7,$A4,$F1,$69
	.dc.b	$F8,$78,$FC,$A7,$37,$E9,$3E,$33,$ED,$79,$72,$BF,$0B,$19,$69,$0F
	.dc.b	$C2,$FC,$67,$F6,$F8,$FD,$9E,$09,$FC,$17,$C1,$7E,$14,$D3,$0F,$F3
	.dc.b	$AC,$F0,$7F,$BD,$37,$A0,$FF,$8D,$DF,$E5,$90,$F9,$8F,$E1,$8F,$2F
	.dc.b	$8D,$F1,$1F,$DE,$63,$31,$F2,$8F,$23,$87,$FB,$8B,$5F,$F4,$63,$5D
	.dc.b	$F8,$BF,$DD,$23,$9D,$78,$9B,$8F,$09,$FF,$08,$7A,$5F,$BD,$BF,$6F
	.dc.b	$C2,$FE,$97,$D1,$91,$82,$1E,$5C,$7D,$F3,$F2,$CF,$99,$F1,$BF,$37
	.dc.b	$DD,$FD,$77,$CA,$CB,$F9,$31,$96,$90,$E9,$B3,$E3,$BE,$B7,$C1,$7F
	.dc.b	$76,$8F,$F1,$1C,$FF,$2A,$6B,$FC,$81,$AF,$DB,$CB,$E9,$CF,$28,$FB
	.dc.b	$2F,$E0,$7E,$0F,$F0,$5F,$1D,$FC,$0F,$D7,$95,$F7,$A3,$2D,$21,$E6
	.dc.b	$F8,$1F,$C5,$7C,$07,$94,$91,$FE,$08,$CB,$FC,$14,$F9,$E3,$6F,$E2
	.dc.b	$9F,$C9,$7C,$0F,$F8,$E7,$E2,$1D,$7F,$E4,$5A,$7C,$87,$C3,$7F,$2D
	.dc.b	$BB,$A7,$FD,$99,$39,$9F,$91,$18,$9F,$F6,$85,$F4,$E0,$FC,$F5,$EE
	.dc.b	$31,$6F,$98,$E5,$7D,$63,$F4,$91,$8C,$3F,$DB,$4F,$1C,$7C,$CD,$C7
	.dc.b	$77,$FE,$DF,$39,$5F,$57,$17,$C9,$F2,$DF,$26,$D7,$2A,$F7,$D6,$DA
	.dc.b	$64,$61,$EA,$C3,$C5,$1F,$22,$FB,$BF,$01,$AE,$4F,$E0,$13,$F0,$B1
	.dc.b	$E3,$5E,$F9,$ED,$5E,$A8,$CB,$E6,$8F,$61,$E4,$29,$F4,$8F,$35,$79
	.dc.b	$63,$85,$75,$9F,$11,$B2,$CF,$8E,$7B,$56,$C8,$E0,$93,$E1,$8C,$BE
	.dc.b	$CC,$7E,$63,$71,$F8,$4F,$9F,$0F,$11,$F2,$0F,$6B,$CE,$FE,$3D,$E2
	.dc.b	$3E,$7D,$F4,$7D,$11,$E0,$73,$7F,$02,$75,$5F,$24,$E6,$7A,$27,$67
	.dc.b	$C2,$7F,$0F,$7E,$47,$E3,$9B,$39,$47,$94,$F6,$1F,$43,$0F,$90,$FB
	.dc.b	$63,$EF,$CF,$B5,$E7,$19,$21,$E2,$8F,$C4,$FC,$4F,$F8,$B6,$BF,$4A
	.dc.b	$DB,$FD,$A3,$70,$FE,$90,$FF,$8B,$BE,$26,$DF,$FA,$39,$B8,$7F,$AA
	.dc.b	$1A,$FF,$B7,$84,$F1,$7F,$07,$5F,$D9,$7A,$FE,$BC,$6E,$FF,$7D,$39
	.dc.b	$8F,$C4,$9F,$D6,$6F,$71,$99,$FE,$09,$39,$3A,$F8,$1F,$E7,$F8,$CC
	.dc.b	$2B,$F5,$95,$D7,$4C,$C6,$2F,$FD,$0D,$18,$AE,$78,$79,$9E,$D8,$79
	.dc.b	$AB,$B6,$9F,$0E,$7F,$4B,$8F,$58,$C6,$1B,$C3,$F7,$95,$E7,$8C,$3F
	.dc.b	$AC,$31,$3F,$E5,$EB,$F1,$BC,$8C,$57,$29,$F6,$1E,$09,$DB,$F4,$CF
	.dc.b	$49,$F2,$8D,$E1,$E0,$3D,$D7,$E2,$2D,$F6,$9F,$21,$DA,$35,$1E,$38
	.dc.b	$FE,$A5,$5E,$07,$ED,$47,$1A,$F9,$47,$77,$84,$30,$F8,$CF,$9E,$FA
	.dc.b	$B6,$7C,$51,$E4,$8E,$F8,$7C,$1C,$7F,$07,$8F,$D2,$9F,$81,$8F,$77
	.dc.b	$91,$C0,$F5,$CF,$DF,$AE,$51,$F2,$5B,$7E,$28,$FC,$D3,$5F,$AA,$73
	.dc.b	$7D,$B8,$7B,$CF,$C1,$5F,$87,$1F,$F1,$3F,$C3,$7F,$B4,$59,$C3,$FE
	.dc.b	$EB,$C7,$FC,$5F,$F0,$D9,$7F,$BE,$46,$72,$FF,$03,$DF,$F8,$B9,$CB
	.dc.b	$5F,$F1,$AB,$6F,$BD,$7F,$75,$F8,$5E,$95,$FE,$3A,$6C,$FF,$8E,$BB
	.dc.b	$AE,$1F,$EF,$6B,$FF,$20,$D6,$BD,$27,$C1,$EA,$30,$F0,$F8,$D9,$5F
	.dc.b	$E4,$8F,$86,$B9,$FF,$87,$A7,$C9,$89,$A4,$AF,$F0,$57,$C3,$78,$8F
	.dc.b	$1F,$89,$63,$FC,$A7,$1E,$3F,$86,$C6,$1F,$E3,$5D,$11,$EA,$F7,$65
	.dc.b	$FF,$9B,$37,$7C,$AA,$33,$87,$F8,$CA,$BD,$DF,$0F,$8B,$FA,$D8,$A3
	.dc.b	$E4,$D3,$E6,$3E,$1E,$1F,$E7,$88,$C7,$07,$CC,$F9,$87,$1F,$0B,$DD
	.dc.b	$7F,$97,$CD,$DF,$F9,$CD,$86,$0A,$53,$31,$C9,$B7,$F6,$B4,$7A,$DD
	.dc.b	$E3,$0E,$7C,$F9,$87,$C3,$A7,$9D,$8C,$3F,$CB,$D0,$AF,$C0,$7C,$4E
	.dc.b	$77,$F8,$6F,$7F,$3A,$63,$CF,$CC,$B8,$F2,$9F,$C4,$7C,$C6,$30,$F9
	.dc.b	$6F,$CA,$F9,$7F,$F1,$E6,$1F,$25,$B7,$F1,$5F,$F4,$B4,$67,$C6,$E0
	.dc.b	$7E,$7E,$53,$C1,$F9,$2D,$7E,$63,$5F,$F5,$03,$DA,$FF,$65,$C3,$FD
	.dc.b	$35,$FF,$4C,$C6,$6F,$FC,$D9,$39,$7F,$CD,$DA,$BC,$BF,$A0,$7D,$1E
	.dc.b	$1B,$3F,$36,$3E,$A7,$3E,$32,$F9,$AF,$D7,$E5,$95,$F4,$F8,$2D,$7D
	.dc.b	$29,$FD,$87,$C5,$63,$5F,$C9,$BF,$7F,$79,$87,$DB,$8B,$7F,$6A,$7F
	.dc.b	$81,$68,$69,$2B,$F1,$B7,$FA,$48,$FC,$57,$C5,$F2,$8F,$E2,$D4,$8F
	.dc.b	$C7,$7C,$5C,$BF,$C5,$7C,$C5,$C7,$F1,$33,$FC,$1E,$B1,$87,$EA,$9F
	.dc.b	$CB,$7C,$5C,$C3,$F2,$5F,$2B,$70,$FC,$8C,$DB,$F0,$CD,$AF,$B7,$EA
	.dc.b	$7E,$24,$FD,$07,$C6,$58,$FC,$54,$7E,$17,$94,$D2,$47,$EB,$31,$FC
	.dc.b	$4C,$7E,$A7,$BB,$94,$7E,$89,$FD,$67,$C5,$CA,$FE,$15,$F3,$1F,$E1
	.dc.b	$A7,$F5,$51,$FC,$AA,$3F,$D3,$31,$F3,$1B,$5F,$BC,$CD,$18,$8F,$E0
	.dc.b	$11,$8F,$2B,$F9,$C4,$BE,$2B,$F9,$87,$07,$56,$E1,$E2,$F8,$BD,$91
	.dc.b	$F0,$3C,$BE,$2E,$47,$8D,$B9,$78,$F0,$7C,$5F,$F3,$DE,$4F,$53,$6F
	.dc.b	$DF,$F1,$1E,$DF,$8A,$F7,$36,$3E,$1F,$C4,$FA,$08,$FC,$17,$D7,$F1
	.dc.b	$1F,$14,$F8,$DF,$0F,$67,$E5,$FC,$47,$C9,$F8,$8F,$94,$D8,$F9,$FE
	.dc.b	$8F,$C4,$23,$F3,$5F,$EB,$DF,$1D,$FC,$1B,$7F,$F7,$B3,$79,$7F,$B7
	.dc.b	$2F,$95,$FE,$BE,$E6,$FF,$DD,$0F,$EF,$EB,$FC,$24,$FE,$22,$3F,$A4
	.dc.b	$39,$A3,$99,$EF,$F0,$BE,$22,$57,$EB,$25,$1E,$6F,$67,$C2,$78,$BE
	.dc.b	$16,$D7,$6F,$85,$F2,$FC,$27,$99,$B1,$E2,$8F,$79,$1E,$CF,$C2,$F8
	.dc.b	$4F,$69,$C3,$2F,$86,$FC,$0F,$84,$95,$EC,$F7,$6D,$7C,$89,$F9,$64
	.dc.b	$9F,$53,$63,$E6,$FB,$DC,$36,$F9,$AF,$CC,$F7,$A5,$FC,$8F,$46,$DF
	.dc.b	$D4,$9F,$AE,$49,$F8,$CF,$F3,$BF,$96,$7B,$94,$93,$FB,$78,$CB,$63
	.dc.b	$FC,$33,$F3,$2F,$8D,$FE,$EB,$5A,$C3,$FC,$99,$67,$FC,$9D,$5B,$8F
	.dc.b	$BB,$F0,$1F,$84,$C3,$F0,$AD,$8C,$9F,$82,$8C,$B4,$FF,$00,$7C,$B7
	.dc.b	$F6,$AE,$7F,$F4,$A7,$8B,$FD,$EE,$31,$8F,$FA,$3E,$3F,$BF,$7C,$53
	.dc.b	$8F,$FC,$25,$39,$3C,$D5,$C3,$E1,$7F,$81,$46,$3D,$EF,$CD,$3B,$13
	.dc.b	$BC,$3A,$DF,$4B,$E3,$5A,$BE,$31,$8B,$E1,$3D,$71,$F4,$BF,$1F,$1F
	.dc.b	$36,$58,$8E,$53,$6B,$8C,$60,$F2,$57,$07,$CD,$1A,$26,$1E,$2C,$BC
	.dc.b	$F8,$73,$8C,$57,$8C,$FF,$10,$D7,$CA,$7F,$C4,$AD,$7E,$74,$7C,$47
	.dc.b	$FC,$53,$F1,$3F,$C7,$3E,$5F,$F9,$F3,$63,$FC,$56,$D7,$7E,$32,$7F
	.dc.b	$A2,$5B,$DE,$FF,$6B,$D7,$89,$AF,$89,$AF,$DF,$6B,$F6,$9E,$3A,$FC
	.dc.b	$F6,$E7,$BF,$89,$F9,$5E,$9F,$F9,$06,$B5,$1F,$36,$B7,$7C,$E3,$6E
	.dc.b	$58,$FC,$E8,$DB,$F8,$4A,$E9,$7D,$E3,$55,$C2,$FF,$05,$5A,$9E,$C6
	.dc.b	$89,$1A,$7F,$0B,$D2,$F4,$5E,$AF,$F5,$AE,$0F,$D7,$4E,$63,$15,$FE
	.dc.b	$BD,$AD,$BF,$E7,$35,$EA,$EA,$BA,$36,$37,$3F,$E7,$DA,$C4,$E7,$1D
	.dc.b	$54,$AE,$91,$96,$90,$FF,$43,$7A,$7E,$45,$F2,$CF,$73,$FC,$09,$18
	.dc.b	$6C,$66,$3F,$C1,$0D,$7F,$CE,$AD,$CB,$F7,$71,$C4,$FF,$42,$46,$73
	.dc.b	$FF,$3F,$7C,$BB,$D6,$F1,$73,$3C,$BF,$27,$BC,$F3,$29,$A8,$99,$7F
	.dc.b	$86,$6B,$83,$63,$D3,$18,$8F,$F4,$4B,$73,$F2,$B7,$2F,$F4,$83,$67
	.dc.b	$F9,$36,$16,$9F,$BA,$7E,$81,$FF,$6C,$36,$FE,$FB,$6B,$F9,$77,$E1
	.dc.b	$9F,$E8,$C7,$FD,$48,$D7,$FD,$36,$D7,$FD,$3E,$D7,$FD,$44,$D5,$5E
	.dc.b	$2C,$FE,$C4,$F3,$8F,$9E,$DB,$F1,$CF,$E2,$ED,$F3,$74,$98,$7F,$B0
	.dc.b	$E3,$F9,$2E,$1C,$BA,$3F,$03,$8F,$F0,$1A,$FF,$23,$35,$F4,$C7,$F9
	.dc.b	$26,$7E,$2F,$CB,$F7,$24,$F5,$F6,$BF,$1D,$F5,$2F,$91,$A4,$9F,$F6
	.dc.b	$47,$C8,$FE,$AD,$FF,$65,$4B,$E5,$3F,$26,$D8,$FF,$2C,$35,$FF,$62
	.dc.b	$B5,$FF,$64,$FD,$4F,$F4,$D7,$FD,$A8,$D7,$FD,$A0,$DC,$BF,$2F,$7F
	.dc.b	$EB,$F9,$F8,$8F,$C9,$23,$F1,$FD,$FF,$D4,$53,$D1,$FD,$3B,$67,$FC
	.dc.b	$D8,$D7,$FD,$B2,$D7,$FC,$BF,$BF,$C3,$7E,$96,$7F,$5C,$FB,$1E,$27
	.dc.b	$D9,$F9,$9F,$B5,$EE,$CA,$FB,$11,$E2,$F9,$2F,$09,$23,$C2,$FD,$5C
	.dc.b	$1E,$FF,$E0,$C9,$3F,$81,$F7,$FF,$0D,$E9,$7E,$1F,$DF,$93,$F8,$58
	.dc.b	$F7,$FC,$97,$B8,$91,$EE,$7B,$7C,$07,$E3,$C9,$1F,$8D,$EF,$FE,$1D
	.dc.b	$F4,$EF,$84,$FE,$21,$F3,$AF,$76,$9A,$89,$7F,$E9,$5A,$E0,$D8,$FE
	.dc.b	$2B,$18,$8F,$F0,$E1,$1C,$74,$7F,$12,$6F,$83,$9A,$F0,$FB,$DF,$A5
	.dc.b	$F8,$39,$87,$E8,$DF,$D4,$78,$D8,$8F,$F1,$49,$FB,$2E,$34,$E7,$2F
	.dc.b	$E0,$75,$FA,$FE,$E9,$7F,$AD,$7F,$9E,$7E,$EB,$FE,$A1,$CF,$F8,$3C
	.dc.b	$6A,$7A,$DF,$F0,$E8,$DC,$70,$F5,$E3,$FD,$32,$BF,$A6,$DF,$F7,$FB
	.dc.b	$FF,$14,$85,$F2,$1C,$2B,$AC,$6E,$B8,$DF,$92,$B5,$7C,$AB,$57,$FC
	.dc.b	$41,$B1,$D6,$7D,$7E,$C7,$B3,$E1,$25,$F9,$4E,$11,$A7,$15,$E5,$8C
	.dc.b	$47,$93,$E2,$3E,$2F,$C2,$C8,$DE,$28,$F9,$3F,$0B,$99,$F5,$C5,$C3
	.dc.b	$D1,$3F,$8E,$F5,$3F,$1B,$32,$BC,$83,$93,$F8,$73,$F7,$0C,$4F,$E3
	.dc.b	$08,$FE,$45,$AE,$E3,$8D,$FE,$36,$7F,$37,$9F,$DE,$E3,$97,$EA,$7F
	.dc.b	$03,$F3,$18,$8F,$A7,$16,$FE,$84,$FF,$73,$E4,$4F,$F8,$0F,$D5,$7C
	.dc.b	$7D,$35,$13,$0F,$6D,$70,$6C,$71,$F5,$7F,$BF,$7C,$CF,$FA,$32,$7E
	.dc.b	$9F,$C3,$3C,$FF,$F0,$83,$63,$E3,$CE,$DF,$8C,$D7,$F5,$E1,$C3,$FB
	.dc.b	$26,$BF,$E8,$F6,$BB,$F8,$2F,$19,$1F,$BF,$DB,$FA,$6F,$82,$FD,$3F
	.dc.b	$C4,$49,$FC,$1C,$65,$A4,$3F,$3B,$E7,$7E,$BB,$E2,$3F,$5F,$F1,$12
	.dc.b	$7F,$17,$19,$69,$0F,$D2,$E7,$FB,$6F,$86,$FE,$B9,$D4,$69,$27,$E0
	.dc.b	$9F,$A5,$78,$3F,$98,$7F,$7F,$F3,$1F,$C0,$35,$FE,$03,$E6,$CA,$FC
	.dc.b	$EC,$65,$A4,$3F,$6F,$E4,$FF,$90,$FF,$76,$FF,$15,$C7,$F1,$EE,$94
	.dc.b	$62,$3F,$C9,$ED,$7F,$C0,$ED,$8D,$F8,$7C,$C9,$3F,$B4,$97,$CB,$FF
	.dc.b	$90,$FE,$27,$FB,$B5,$FF,$B1,$FD,$DF,$D9,$5F,$FB,$27,$C1,$FE,$1A
	.dc.b	$E5,$FE,$62,$61,$FF,$1E,$3F,$CD,$2B,$9C,$60,$FF,$69,$38,$F9,$AE
	.dc.b	$B1,$FB,$FF,$1B,$F9,$DC,$62,$BA,$3A,$AE,$11,$D7,$D6,$FF,$12,$DA
	.dc.b	$B8,$5F,$2A,$F2,$B6,$7C,$35,$C2,$BA,$36,$B7,$3F,$EE,$79,$E6,$D8
	.dc.b	$F0,$63,$63,$9B,$F0,$DF,$8D,$7F,$D5,$31,$E3,$1F,$04,$63,$F1,$23
	.dc.b	$C3,$5C,$E7,$E1,$16,$FE,$09,$E0,$7D,$CD,$C3,$A9,$1C,$1B,$AF,$09
	.dc.b	$1C,$1A,$9F,$39,$DF,$0F,$8C,$F9,$B1,$F8,$8F,$B8,$F0,$9C,$0A,$7A
	.dc.b	$0B,$8F,$D6,$78,$5F,$85,$E0,$FC,$07,$E2,$46,$DE,$58,$7A,$DF,$03
	.dc.b	$B3,$0E,$B0,$F5,$3E,$7C,$7D,$4C,$3C,$78,$7A,$9D,$1F,$4E,$7F,$31
	.dc.b	$87,$96,$32,$F7,$8F,$C1,$C7,$B9,$FA,$6F,$91,$F9,$6F,$B1,$DC,$7B
	.dc.b	$48,$E0,$91,$F1,$1F,$B9,$F0,$FF,$EA,$38,$F9,$CD,$7C,$ED,$C7,$F5
	.dc.b	$6E,$60,$3E,$1E,$C6,$63,$01,$C4,$6A,$FF,$8D,$B6,$7F,$9E,$BF,$9B
	.dc.b	$7F,$D2,$3E,$DB,$7C,$47,$82,$75,$9A,$B5,$3F,$9F,$34,$7E,$86,$35
	.dc.b	$5C,$A3,$CB,$7F,$C0,$A7,$31,$C4,$F4,$E9,$9B,$EA,$FF,$08,$8E,$AE
	.dc.b	$6F,$F5,$51,$E0,$BC,$3C,$F0,$F3,$DF,$9A,$79,$4F,$6C,$FF,$97,$4F
	.dc.b	$67,$E2,$C7,$E7,$B1,$F2,$E3,$C3,$0F,$DC,$46,$2B,$73,$E0,$35,$5C
	.dc.b	$6F,$A3,$F3,$8C,$63,$E8,$7E,$1D,$78,$A3,$E3,$61,$E8,$6C,$F0,$3F
	.dc.b	$62,$7C,$29,$E0,$E0,$F6,$1F,$86,$C3,$CC,$F8,$70,$F3,$4F,$BB,$1F
	.dc.b	$8E,$7A,$0A,$FB,$0E,$B1,$EC,$38,$CF,$38,$C9,$B7,$C1,$8F,$C0,$7F
	.dc.b	$CE,$F8,$FA,$DF,$A8,$7E,$16,$3E,$A4,$F1,$8F,$C8,$E3,$F1,$6B,$E7
	.dc.b	$D7,$39,$F8,$07,$69,$D3,$F1,$6F,$CD,$7E,$48,$F3,$1B,$3E,$35,$F8
	.dc.b	$F0,$F8,$13,$F2,$9F,$EC,$38,$FF,$70,$F9,$8F,$F3,$7E,$3F,$DC,$1B
	.dc.b	$87,$F7,$0F,$83,$B5,$BA,$FE,$DD,$5F,$DD,$E3,$15,$FE,$2A,$6B,$FD
	.dc.b	$F7,$85,$3B,$5F,$01,$D1,$B7,$FD,$E7,$E6,$31,$32,$3F,$BE,$7C,$BD
	.dc.b	$C3,$FD,$1B,$68,$E5,$F3,$12,$BF,$C0,$5F,$0D,$6F,$73,$E1,$24,$FF
	.dc.b	$88,$33,$FF,$08,$C7,$97,$C9,$F2,$1E,$5F,$1E,$D7,$F8,$36,$3D,$1F
	.dc.b	$2F,$8C,$3C,$6F,$AF,$E6,$25,$7F,$84,$FE,$5E,$D7,$A2,$7E,$07,$B3
	.dc.b	$8C,$FF,$D3,$FB,$7F,$B2,$F0,$FF,$3E,$FC,$CB,$6F,$A7,$8A,$31,$3F
	.dc.b	$E1,$FF,$96,$CC,$62,$7F,$D0,$4D,$7F,$CB,$AD,$7F,$D9,$AD,$7F,$97
	.dc.b	$B5,$ED,$5F,$80,$F9,$AF,$F2,$45,$F9,$23,$1E,$27,$93,$E0,$9F,$1F
	.dc.b	$4F,$C3,$F9,$0F,$89,$C6,$1F,$E9,$8A,$DB,$FE,$78,$F0,$B3,$9F,$FB
	.dc.b	$22,$D1,$F8,$5F,$91,$95,$FE,$3D,$F9,$1B,$7F,$82,$9B,$C3,$F2,$A6
	.dc.b	$D7,$FB,$AB,$E6,$7F,$DE,$8F,$FA,$A1,$B7,$FE,$D6,$DD,$E5,$FC,$E9
	.dc.b	$CE,$1F,$D0,$29,$1C,$1F,$F4,$3B,$5E,$DE,$97,$F9,$5E,$31,$1F,$D6
	.dc.b	$9E,$78,$7F,$17,$C1,$E5,$F7,$BF,$9C,$FE,$8F,$D7,$95,$FE,$5B,$F6
	.dc.b	$AE,$3F,$A0,$9B,$97,$E2,$26,$D7,$F9,$A3,$8B,$F5,$87,$EE,$3C,$EF
	.dc.b	$EC,$D1,$FE,$65,$8C,$BF,$ED,$A8,$78,$B8,$C7,$FD,$F3,$18,$9D,$7C
	.dc.b	$87,$33,$A7,$C9,$D8,$C9,$2B,$8C,$65,$E7,$37,$2F,$F7,$ED,$AB,$17
	.dc.b	$FC,$EE,$3F,$D7,$4F,$97,$D8,$F5,$7C,$87,$AD,$91,$93,$D5,$18,$C7
	.dc.b	$D3,$E6,$FC,$13,$E1,$74,$DA,$F8,$11,$67,$DF,$37,$2F,$C0,$71,$BD
	.dc.b	$5F,$C2,$1F,$C8,$BE,$71,$EC,$E3,$2F,$C6,$4F,$E0,$48,$FF,$AB,$31
	.dc.b	$FC,$4C,$61,$61,$FE,$0C,$16,$F6,$3A,$4E,$AF,$FA,$F4,$E5,$FB,$1D
	.dc.b	$7F,$74,$FB,$DF,$1F,$63,$FB,$21,$FE,$2A,$A7,$DA,$FE,$F4,$9E,$0F
	.dc.b	$FA,$B5,$FE,$05,$93,$89,$F1,$FF,$89,$F9,$0F,$C5,$37,$1F,$C4,$7C
	.dc.b	$87,$E3,$4F,$C7,$78,$76,$3C,$39,$FF,$8E,$B2,$FF,$7F,$C7,$C3,$6C
	.dc.b	$7F,$B4,$1A,$FF,$BC,$9A,$FE,$0F,$E3,$FF,$35,$F2,$1F,$9B,$6D,$7E
	.dc.b	$67,$E4,$3F,$3C,$7E,$7F,$E4,$2C,$78,$AF,$E0,$F8,$DE,$F2,$47,$8D
	.dc.b	$FE,$B5,$B3,$E3,$FF,$50,$7E,$A7,$E3,$ED,$73,$D1,$E1,$FF,$34,$6C
	.dc.b	$7C,$0E,$26,$8F,$84,$D5,$FF,$3A,$A7,$2F,$FB,$75,$A7,$E1,$CB,$AF
	.dc.b	$FB,$D1,$AF,$DB,$1F,$6B,$FE,$66,$8F,$73,$63,$F1,$07,$E3,$BF,$64
	.dc.b	$7A,$7F,$D2,$A7,$59,$FE,$CB,$E4,$7F,$83,$1F,$DF,$BF,$77,$CE,$3F
	.dc.b	$B3,$8F,$EA,$3E,$3B,$DE,$FF,$7A,$36,$7F,$3C,$DC,$FC,$C7,$F0,$C9
	.dc.b	$FF,$03,$5B,$E4,$BB,$1D,$FE,$16,$E1,$F2,$65,$F2,$5E,$0F,$7F,$C8
	.dc.b	$D9,$FB,$99,$7F,$82,$32,$EE,$D7,$E3,$35,$FF,$6C,$57,$8B,$E3,$FD
	.dc.b	$C7,$38,$F8,$FA,$B1,$4F,$F7,$0F,$AC,$77,$D3,$FC,$47,$F4,$AF,$7B
	.dc.b	$1A,$7E,$3E,$B6,$FF,$01,$6E,$5F,$11,$AF,$DB,$6B,$FB,$77,$79,$7F
	.dc.b	$05,$B7,$BD,$DA,$BC,$F1,$8B,$FC,$F7,$CF,$7F,$4C,$6F,$AB,$E6,$F6
	.dc.b	$7F,$C1,$33,$9B,$FC,$28,$70,$FE,$35,$AF,$F0,$96,$BF,$EB,$46,$E3
	.dc.b	$F1,$FE,$57,$F1,$07,$E2,$7E,$1E,$E3,$F6,$E3,$F1,$9F,$2B,$8C,$3E
	.dc.b	$7F,$0A,$3F,$21,$F2,$5F,$91,$F9,$4B,$87,$E0,$DC,$34,$FC,$34,$62
	.dc.b	$FD,$7F,$39,$F9,$5E,$9F,$F5,$D7,$B3,$FC,$86,$31,$3F,$EC,$B8,$FE
	.dc.b	$49,$D3,$FB,$5C,$37,$3F,$ED,$0A,$DF,$C6,$7F,$55,$8C,$76,$7E,$65
	.dc.b	$E4,$63,$1E,$2F,$E6,$3C,$3F,$F6,$3D,$71,$85,$F4,$CB,$C9,$87,$3A
	.dc.b	$D4,$E6,$BF,$DC,$39,$79,$5E,$E7,$2A,$E1,$7C,$B1,$F0,$DF,$1A,$F1
	.dc.b	$65,$8B,$DC,$79,$CE,$73,$BC,$3B,$63,$E2,$AD,$D7,$7C,$3F,$AD,$5F
	.dc.b	$CA,$8F,$E2,$4D,$7C,$47,$F0,$0F,$8F,$FE,$3B,$F3,$BF,$D8,$5B,$19
	.dc.b	$F1,$13,$96,$C7,$F0,$8A,$FF,$42,$FC,$87,$83,$BF,$FD,$FD,$39,$6B
	.dc.b	$E2,$6B,$F7,$DD,$E7,$F3,$1B,$B6,$6B,$E3,$F8,$DF,$EB,$88,$C5,$FC
	.dc.b	$4A,$DB,$F8,$36,$BF,$6A,$78,$39,$AF,$A3,$3A,$C3,$D3,$3F,$3A,$B6
	.dc.b	$FE,$0A,$74,$4A,$F1,$C7,$26,$3E,$37,$2F,$E4,$39,$DF,$00,$F3,$7F
	.dc.b	$BF,$47,$F8,$2E,$31,$3F,$D4,$DA,$FF,$1E,$5D,$7C,$38,$FF,$DD,$51
	.dc.b	$AA,$E5,$79,$BF,$F0,$BC,$E6,$B7,$79,$BF,$1E,$56,$BC,$B7,$E1,$71
	.dc.b	$12,$30,$43,$C8,$F7,$5F,$3A,$F8,$CC,$7C,$EF,$EA,$F0,$FF,$3A,$37
	.dc.b	$2F,$F0,$F5,$6E,$FF,$C0,$0D,$7F,$A9,$B5,$FE,$D4,$D7,$8D,$FF,$8D
	.dc.b	$DF,$EB,$DE,$8F,$8F,$D7,$7A,$14,$F3,$74,$26,$99,$AF,$F0,$85,$70
	.dc.b	$7F,$C2,$53,$B6,$CF,$06,$9C,$1F,$99,$F3,$1F,$9E,$D3,$FC,$8D,$DE
	.dc.b	$F8,$5F,$A0,$DB,$FF,$08,$56,$F5,$FF,$20,$7C,$EF,$F7,$79,$CF,$27
	.dc.b	$43,$E7,$46,$63,$FA,$1B,$71,$FF,$50,$36,$B9,$E7,$E4,$F3,$3C,$BF
	.dc.b	$0B,$E2,$F0,$26,$3E,$28,$EB,$D1,$EA,$EA,$93,$FE,$33,$CB,$FE,$00
	.dc.b	$BE,$8F,$85,$DB,$9A,$F8,$31,$F8,$3F,$9D,$FF,$61,$CF,$F9,$AD,$AF
	.dc.b	$F8,$39,$AF,$F8,$95,$AF,$F9,$85,$7A,$3F,$8D,$F8,$5F,$ED,$73,$FE
	.dc.b	$45,$6B,$FD,$92,$76,$FF,$92,$7E,$77,$FC,$74,$FF,$9E,$DA,$FF,$87
	.dc.b	$9A,$FF,$8D,$5A,$8F,$F3,$FF,$12,$BE,$E9,$23,$EC,$FC,$37,$E8,$6B
	.dc.b	$FD,$05,$1F,$E2,$3F,$85,$FB,$E6,$7E,$13,$F0,$BF,$07,$27,$F0,$2F
	.dc.b	$E2,$3C,$5F,$C4,$FC,$17,$E2,$BE,$0E,$57,$E1,$A3,$B7,$C0,$7E,$2B
	.dc.b	$09,$3E,$F8,$FC,$8F,$B1,$F8,$EF,$9A,$93,$F8,$E8,$FC,$AF,$C1,$7E
	.dc.b	$5B,$E0,$E4,$FE,$4A,$3D,$5F,$33,$F9,$6F,$87,$93,$F2,$E3,$F3,$BF
	.dc.b	$07,$F9,$AF,$8D,$93,$F9,$A8,$FE,$05,$F3,$8F,$93,$FE,$09,$F3,$BF
	.dc.b	$9F,$AF,$F0,$CB,$7C,$4F,$F7,$B5,$6E,$7F,$D0,$0D,$77,$E3,$F2,$3A
	.dc.b	$DC,$8F,$F5,$B4,$ED,$FF,$37,$35,$FF,$55,$35,$FB,$7F,$13,$FB,$62
	.dc.b	$3F,$B2,$F2,$EB,$FE,$4C,$EC,$FF,$46,$3F,$B6,$9F,$DE,$CF,$EF,$BB
	.dc.b	$A4,$FE,$E9,$FE,$59,$F3,$9F,$E7,$FC,$FF,$D8,$D8,$7E,$52,$BA,$57
	.dc.b	$07,$FC,$B3,$5B,$F2,$3F,$21,$68,$FF,$0C,$4E,$BD,$8E,$A7,$6F,$86
	.dc.b	$B1,$FE,$99,$9D,$D7,$FB,$55,$B7,$D1,$57,$0A,$E8,$71,$BE,$8F,$3F
	.dc.b	$6B,$D7,$F0,$D2,$30,$43,$DD,$7F,$E8,$07,$FD,$13,$8F,$A6,$BF,$DD
	.dc.b	$4D,$8C,$CB,$E0,$3E,$39,$23,$D5,$EF,$D7,$E2,$C5,$AF,$64,$FE,$3B
	.dc.b	$D7,$FC,$6C,$CA,$FC,$4C,$65,$FF,$65,$72,$77,$6C,$E5,$FF,$35,$F1
	.dc.b	$7F,$60,$7F,$D4,$2D,$7F,$17,$5F,$D3,$63,$11,$F2,$FD,$EF,$BE,$47
	.dc.b	$E9,$7B,$F5,$FB,$B1,$72,$FA,$93,$FD,$9F,$89,$3F,$DC,$FD,$97,$CB
	.dc.b	$53,$51,$30,$FF,$5A,$8A,$E2,$E1,$AF,$FA,$E9,$AF,$FB,$47,$E7,$5F
	.dc.b	$09,$FE,$01,$EC,$79,$D4,$FE,$F5,$3A,$AF,$F7,$93,$5F,$F5,$7C,$F0
	.dc.b	$7F,$DD,$0D,$7F,$D2,$B1,$BD,$1C,$FF,$CA,$E5,$3C,$DF,$26,$1F,$DE
	.dc.b	$63,$11,$B6,$AA,$FD,$27,$A7,$FA,$EF,$2E,$4F,$E8,$63,$F6,$27,$E4
	.dc.b	$39,$7F,$9D,$C7,$16,$D7,$28,$FC,$9F,$8B,$FB,$8F,$0A,$47,$EA,$E3
	.dc.b	$F7,$67,$32,$3C,$0F,$CD,$FC,$A7,$F7,$C2,$3F,$B3,$7F,$81,$9E,$32
	.dc.b	$3C,$0F,$D1,$E7,$FC,$23,$C7,$91,$FB,$B8,$FE,$16,$61,$FE,$0F,$18
	.dc.b	$7F,$83,$EB,$FC,$23,$E5,$FE,$7C,$7F,$BC,$FE,$57,$FC,$1E,$7F,$8F
	.dc.b	$E3,$31,$FE,$E8,$6B,$FE,$52,$6B,$CE,$7F,$73,$18,$8F,$EC,$9F,$0A
	.dc.b	$F3,$FE,$A4,$3E,$73,$F5,$CF,$EF,$7C,$2F,$C0,$46,$3B,$7F,$88,$CF
	.dc.b	$F6,$D8,$DD,$F3,$E0,$C5,$7E,$1E,$31,$4F,$67,$96,$1F,$8B,$6C,$7F
	.dc.b	$77,$79,$E3,$E4,$6E,$1D,$27,$FB,$E6,$1E,$27,$D0,$FE,$54,$C6,$3C
	.dc.b	$23,$27,$C0,$BF,$7C,$78,$23,$27,$BC,$EC,$31,$F4,$C7,$33,$81,$2B
	.dc.b	$DD,$5D,$1B,$7E,$B7,$C9,$8F,$BD,$B8,$64,$F2,$4D,$7C,$15,$CD,$F7
	.dc.b	$D7,$D2,$AE,$31,$E8,$C3,$CB,$1E,$08,$F5,$BF,$08,$DE,$1E,$97,$58
	.dc.b	$FC,$12,$38,$39,$47,$B6,$7C,$27,$C4,$AE,$EF,$A0,$EF,$7E,$57,$67
	.dc.b	$97,$0F,$53,$EC,$C7,$E6,$E1,$E5,$8D,$E1,$E5,$7A,$3E,$17,$EA,$1C
	.dc.b	$E7,$C5,$5D,$2B,$94,$EA,$B9,$3D,$CF,$29,$E0,$BF,$33,$98,$DC,$F2
	.dc.b	$39,$E1,$EE,$9D,$47,$CA,$23,$E6,$7E,$39,$87,$C4,$E1,$F8,$DE,$DF
	.dc.b	$E8,$5B,$97,$D7,$8C,$47,$9D,$AF,$F9,$67,$E1,$7F,$CB,$51,$88,$46
	.dc.b	$27,$F9,$A4,$6C,$4E,$E7,$51,$A7,$F3,$B9,$FF,$98,$F5,$FC,$FC,$F1
	.dc.b	$CB,$F8,$0C,$E6,$3F,$4A,$67,$1F,$F6,$FF,$AD,$FC,$5A,$BF,$83,$CE
	.dc.b	$4F,$F1,$0C,$66,$7C,$B1,$E8,$3F,$5F,$5D,$6B,$8C,$65,$F0,$D7,$18
	.dc.b	$F4,$1C,$1C,$3F,$D4,$F0,$F5,$C6,$23,$C4,$7F,$4F,$70,$F5,$BF,$4E
	.dc.b	$1F,$02,$30,$7A,$1C,$57,$46,$C7,$85,$B7,$E6,$6D,$74,$9F,$8E,$74
	.dc.b	$3C,$51,$A7,$B1,$AA,$EC,$D3,$F9,$34,$F3,$CB,$E2,$CF,$F0,$5C,$7E
	.dc.b	$A4,$62,$BC,$87,$1C,$3C,$8D,$BF,$24,$6C,$C9,$C6,$76,$58,$F0,$1F
	.dc.b	$62,$FC,$B3,$C6,$39,$99,$AF,$0C,$65,$F6,$B5,$F8,$27,$8C,$CC,$71
	.dc.b	$AF,$0B,$4F,$0C,$7E,$02,$7C,$F1,$DD,$B1,$FE,$0B,$D9,$F3,$9F,$D8
	.dc.b	$31,$46,$32,$FF,$2D,$B6,$BF,$CF,$8D,$7F,$B4,$57,$F7,$0A,$FF,$7F
	.dc.b	$F1,$5B,$FE,$E9,$36,$BF,$C3,$5E,$F7,$F8,$CA,$7F,$BC,$C7,$8F,$DA
	.dc.b	$D6,$1E,$2F,$9B,$B9,$26,$C6,$1B,$7C,$AB,$D3,$18,$BF,$EF,$DF,$37
	.dc.b	$ED,$C0,$D2,$5F,$B9,$F0,$E2,$E6,$F0,$C6,$1F,$1F,$85,$CC,$95,$F1
	.dc.b	$9B,$87,$F8,$37,$E3,$F7,$3E,$5C,$51,$EC,$F8,$EF,$21,$27,$E7,$B6
	.dc.b	$BF,$C2,$DE,$8E,$E7,$93,$F2,$E1,$F3,$4F,$85,$FF,$0B,$36,$BF,$CC
	.dc.b	$ED,$9F,$E1,$6D,$7F,$90,$B9,$DF,$ED,$7C,$D3,$BB,$B4,$E6,$FF,$D6
	.dc.b	$B3,$97,$EC,$DF,$8B,$0F,$B3,$39,$8F,$F2,$C4,$FF,$98,$6B,$94,$7F
	.dc.b	$96,$E7,$58,$7F,$8D,$3E,$59,$F3,$5F,$E2,$F6,$E1,$FE,$80,$6D,$FF
	.dc.b	$24,$6B,$FC,$E9,$CD,$7E,$4F,$C7,$FF,$1B,$57,$9A,$72,$FF,$B3,$A3
	.dc.b	$35,$FE,$43,$F8,$DF,$C7,$9F,$90,$F8,$DB,$5F,$66,$2C,$FF,$92,$66
	.dc.b	$E1,$FE,$4C,$F9,$97,$B5,$FE,$47,$6E,$1F,$E9,$86,$DF,$F3,$C6,$BF
	.dc.b	$D4,$9C,$D7,$D2,$EB,$FB,$D1,$8A,$FA,$31,$8B,$F9,$ED,$8F,$D1,$6B
	.dc.b	$FA,$DC,$3F,$CB,$DE,$7F,$EA,$8F,$D5,$FA,$56,$BF,$29,$16,$7F,$CC
	.dc.b	$93,$72,$FD,$E7,$37,$CE,$9F,$F3,$77,$07,$EA,$C9,$87,$F0,$67,$F7
	.dc.b	$3D,$1F,$E5,$78,$FD,$04,$63,$C0,$CF,$C4,$C9,$FD,$C5,$3E,$27,$18
	.dc.b	$7E,$20,$FF,$50,$42,$B3,$97,$F0,$F8,$E1,$7C,$A7,$2A,$B8,$C7,$92
	.dc.b	$F3,$1D,$3D,$0F,$39,$E8,$F9,$AB,$7E,$48,$B3,$E3,$C7,$F8,$25,$61
	.dc.b	$F1,$78,$DF,$00,$F8,$3F,$11,$70,$F6,$C5,$BE,$59,$AB,$17,$F8,$9F
	.dc.b	$3B,$F1,$04,$C3,$F1,$0D,$BF,$17,$CD,$3E,$17,$F0,$B3,$FC,$A2,$31
	.dc.b	$5F,$B4,$E7,$4E,$6B,$73,$FC,$9E,$B7,$7E,$4F,$3B,$F9,$0F,$C1,$7D
	.dc.b	$A6,$D7,$EF,$BE,$6B,$F7,$39,$BF,$78,$FF,$6E,$CE,$7C,$1F,$E7,$31
	.dc.b	$8C,$7D,$FF,$11,$F8,$4F,$89,$FC,$2B,$6F,$F0,$3F,$13,$C3,$E7,$BF
	.dc.b	$12,$D8,$F0,$79,$5D,$3C,$FF,$E4,$1A,$A3,$13,$F8,$AF,$89,$FC,$8F
	.dc.b	$C5,$7E,$49,$B1,$F8,$EF,$8A,$FC,$A9,$F9,$6F,$90,$B1,$ED,$F2,$3D
	.dc.b	$3F,$21,$D9,$B1,$F9,$6F,$8C,$FC,$E9,$F9,$EF,$8C,$B1,$F1,$7C,$7F
	.dc.b	$F6,$36,$A6,$92,$3F,$C7,$7C,$65,$F1,$BF,$D8,$3E,$73,$FB,$77,$47
	.dc.b	$FB,$5A,$72,$7F,$5D,$8C,$BF,$9D,$A7,$CD,$FE,$7E,$3A,$B6,$3F,$C9
	.dc.b	$2D,$7F,$CB,$ED,$7F,$4D,$5F,$D7,$FE,$3F,$FA,$67,$90,$DF,$F7,$71
	.dc.b	$86,$61,$FD,$BF,$D8,$7C,$06,$31,$FD,$34,$62,$75,$F1,$FF,$84,$F6
	.dc.b	$FF,$B6,$B6,$3F,$A4,$E5,$FD,$F5,$B3,$FD,$CB,$E3,$BC,$6C,$8F,$07
	.dc.b	$C7,$DC,$7B,$BF,$DE,$0F,$C8,$FC,$F5,$9F,$F0,$27,$82,$8F,$ED,$1F
	.dc.b	$38,$EF,$A7,$F8,$43,$E7,$3F,$BA,$DF,$F5,$C9,$D3,$67,$5E,$F7,$08
	.dc.b	$FF,$08,$52,$31,$97,$F0,$78,$C4,$F1,$37,$3F,$DE,$7E,$73,$F9,$B5
	.dc.b	$7F,$A4,$9A,$FD,$D6,$BF,$C6,$DA,$9F,$E7,$DC,$9F,$E3,$1F,$9F,$78
	.dc.b	$1D,$7D,$9F,$F1,$5C,$63,$1F,$B1,$F3,$BF,$E1,$EF,$95,$FF,$0F,$B6
	.dc.b	$32,$7D,$D8,$CB,$4F,$C0,$7C,$9F,$F8,$98,$FF,$13,$FC,$9D,$8C,$93
	.dc.b	$4F,$C1,$C6,$5F,$E7,$83,$E7,$3F,$1D,$D1,$FE,$B7,$F6,$7F,$39,$18
	.dc.b	$9F,$F5,$BC,$7E,$7F,$A1,$F2,$3F,$A9,$9C,$9F,$CF,$EB,$87,$C6,$7F
	.dc.b	$4B,$8C,$77,$31,$ED,$18,$6A,$C3,$AE,$78,$C3,$9E,$5E,$4C,$39,$5F
	.dc.b	$80,$F5,$38,$C7,$C1,$1C,$F2,$F3,$46,$2F,$9C,$F8,$4E,$4E,$1E,$18
	.dc.b	$78,$6F,$A6,$1E,$1C,$B0,$7A,$32,$F3,$DF,$2A,$DB,$D4,$FF,$6E,$E1
	.dc.b	$F2,$DB,$1E,$A6,$D7,$F1,$4F,$8E,$FE,$27,$F3,$B7,$C4,$34,$95,$FE
	.dc.b	$D4,$C3,$FC,$E9,$F1,$FF,$C7,$1B,$19,$EF,$FF,$75,$B6,$75,$18,$DB
	.dc.b	$E2,$37,$5F,$C0,$B6,$BE,$E7,$8D,$FE,$D7,$9C,$DF,$D5,$AD,$57,$E2
	.dc.b	$DB,$87,$F3,$18,$D4,$C7,$E7,$CE,$AF,$A4,$7E,$26,$38,$92,$7F,$05
	.dc.b	$3A,$9E,$D1,$CF,$0F,$4B,$A8,$FC,$6F,$47,$F6,$5F,$8D,$FE,$6C,$FF
	.dc.b	$A9,$7C,$0F,$F4,$8C,$FF,$14,$F8,$DC,$F1,$FF,$32,$95,$E2,$97,$CD
	.dc.b	$E0,$43,$FC,$29,$3C,$1F,$E3,$FE,$AF,$19,$FE,$EC,$D9,$FF,$43,$4F
	.dc.b	$F2,$4F,$47,$13,$23,$89,$93,$FC,$FB,$F5,$6F,$91,$F2,$42,$30,$EA
	.dc.b	$33,$D4,$F1,$B7,$5A,$AF,$E0,$2D,$CB,$F7,$51,$CE,$F8,$D7,$6F,$9D
	.dc.b	$7C,$EF,$F6,$8A,$46,$1F,$ED,$6D,$7F,$B2,$1F,$E4,$17,$0E,$1D,$5F
	.dc.b	$81,$BB,$7F,$6B,$6C,$FF,$92,$7E,$79,$E5,$78,$F0,$93,$FE,$A0,$AD
	.dc.b	$AA,$DB,$FD,$42,$7E,$81,$1F,$F5,$3F,$CF,$7F,$BC,$1B,$EC,$7C,$C8
	.dc.b	$C4,$7F,$26,$6E,$5F,$89,$F9,$3F,$B4,$6E,$A5,$7F,$5E,$8F,$C0,$79
	.dc.b	$DF,$E9,$28,$FE,$ED,$F1,$B8,$3F,$09,$F3,$5F,$B4,$8E,$DE,$17,$84
	.dc.b	$8F,$FB,$AE,$3D,$6F,$F9,$85,$B8,$FF,$BF,$DB,$BF,$F8,$29,$B3,$F9
	.dc.b	$8E,$6F,$EC,$8F,$FB,$1B,$E2,$3F,$22,$FF,$42,$6D,$FE,$09,$B7,$FE
	.dc.b	$1D,$FC,$53,$FA,$DD,$71,$F9,$5F,$C8,$D1,$A7,$E5,$27,$FC,$57,$3D
	.dc.b	$FE,$37,$64,$7F,$DA,$F0,$AF,$CD,$FC,$DB,$E6,$E9,$2B,$FC,$4B,$8F
	.dc.b	$F8,$9B,$EB,$1F,$1B,$FA,$2F,$A0,$7B,$74,$97,$FE,$E4,$F8,$FF,$C2
	.dc.b	$E5,$FE,$77,$8F,$F7,$3B,$5F,$EC,$6D,$DF,$FA,$E0,$F8,$1E,$F3,$FE
	.dc.b	$1A,$FA,$1F,$DD,$11,$FF,$0E,$F0,$FE,$D3,$E8,$5E,$BD,$24,$FF,$88
	.dc.b	$BE,$8B,$F8,$17,$C2,$C8,$FF,$12,$BF,$C1,$7E,$63,$F8,$37,$9B,$27
	.dc.b	$FC,$52,$FF,$09,$F9,$7E,$DC,$52,$7F,$E0,$0E,$5E,$67,$F8,$9F,$47
	.dc.b	$F1,$1F,$DE,$7F,$1D,$C1,$EC,$6F,$37,$F9,$41,$B5,$FC,$79,$FC,$4A
	.dc.b	$FA,$2F,$DC,$BF,$E5,$F6,$BF,$EB,$D6,$BB,$F0,$FF,$83,$47,$ED,$08
	.dc.b	$F8,$FE,$63,$F8,$CC,$62,$7E,$E9,$CE,$74,$E3,$05,$FF,$7C,$7D,$5E
	.dc.b	$A7,$B4,$F7,$7C,$C5,$AE,$15,$BC,$BD,$D5,$A9,$CF,$6F,$C5,$23,$C3
	.dc.b	$A6,$BF,$EC,$FB,$FC,$B3,$C0,$D1,$DB,$E6,$3E,$71,$1F,$0F,$C9,$D7
	.dc.b	$94,$5C,$FE,$94,$DC,$BE,$94,$DA,$FA,$53,$F6,$7D,$99,$1E,$C7,$EE
	.dc.b	$7C,$96,$23,$CD,$17,$0F,$BD,$3F,$81,$F9,$39,$5F,$11,$FC,$27,$C4
	.dc.b	$62,$3D,$D1,$6B,$F0,$D3,$F5,$BB,$3E,$D9,$2B,$FD,$53,$C5,$FC,$05
	.dc.b	$B1,$B7,$ED,$9F,$89,$7F,$AB,$FD,$1B,$E0,$B5,$3F,$E5,$2A,$E4,$DA
	.dc.b	$DC,$FF,$26,$F0,$A4,$FF,$23,$7F,$B6,$7D,$1F,$F4,$EC,$BF,$91,$46
	.dc.b	$A7,$F9,$33,$A7,$7F,$05,$1F,$C6,$57,$F9,$3D,$BF,$0B,$97,$FB,$FF
	.dc.b	$17,$FA,$D2,$75,$18,$3F,$C0,$6E,$30,$FE,$0E,$AB,$F3,$BD,$BF,$A2
	.dc.b	$25,$7C,$86,$BF,$05,$DB,$F2,$35,$FE,$F0,$FC,$AD,$53,$A8,$E0,$6F
	.dc.b	$D5,$FF,$01,$F9,$9F,$36,$3E,$8B,$5F,$F3,$FB,$72,$FE,$3B,$E0,$7F
	.dc.b	$68,$8C,$69,$FE,$DB,$6D,$03,$8A,$FF,$76,$76,$7D,$2F,$84,$95,$F3
	.dc.b	$1F,$AB,$F3,$18,$8F,$91,$4D,$FF,$D1,$31,$B3,$E9,$CF,$F1,$7F,$9E
	.dc.b	$91,$FB,$A7,$F8,$D7,$C4,$E2,$3F,$69,$16,$FF,$8E,$4F,$F1,$FF,$96
	.dc.b	$95,$FC,$09,$FE,$45,$F2,$D8,$8F,$DE,$45,$AF,$E4,$93,$FC,$9F,$E5
	.dc.b	$A5,$7F,$0A,$7F,$95,$79,$B8,$8F,$E0,$91,$6B,$F9,$64,$FF,$8D,$3E
	.dc.b	$25,$DF,$FE,$48,$B7,$D0,$BE,$8B,$E6,$45,$F9,$4B,$39,$8F,$A1,$3F
	.dc.b	$EF,$C7,$E9,$46,$55,$F6,$C7,$35,$D2,$31,$3E,$08,$EE,$73,$70,$76
	.dc.b	$BF,$0C,$71,$C3,$D1,$87,$7B,$E6,$F8,$67,$93,$6B,$8D,$FE,$1E,$B7
	.dc.b	$5C,$F0,$F2,$CF,$72,$9F,$04,$B3,$C1,$E5,$87,$8A,$BD,$37,$DA,$78
	.dc.b	$1E,$A3,$E1,$CE,$DE,$37,$DA,$BC,$E7,$28,$EF,$3B,$78,$CF,$E7,$FE
	.dc.b	$03,$D9,$16,$B8,$3F,$E5,$27,$D8,$D3,$B9,$E3,$7B,$BE,$F3,$CD,$87
	.dc.b	$B5,$EE,$6F,$0F,$64,$F0,$8C,$E3,$F4,$71,$FC,$56,$1E,$49,$F0,$1F
	.dc.b	$09,$D6,$3E,$78,$F1,$D6,$C9,$7E,$87,$71,$E1,$3E,$36,$1F,$37,$1F
	.dc.b	$31,$E7,$AF,$21,$E5,$73,$8F,$C9,$8C,$CF,$18,$F9,$93,$C6,$3D,$A6
	.dc.b	$70,$FA,$27,$82,$E1,$F1,$AB,$89,$D6,$78,$C7,$03,$53,$C3,$0F,$6C
	.dc.b	$7C,$0C,$3E,$AF,$C4,$7F,$9A,$31,$FD,$1B,$71,$FD,$A3,$70,$FE,$3C
	.dc.b	$FF,$9B,$AF,$F9,$07,$C4,$E3,$3F,$F4,$DC,$23,$0F,$FA,$79,$AF,$EF
	.dc.b	$1A,$F2,$3F,$2D,$4F,$21,$D1,$FA,$29,$D6,$9F,$BD,$AD,$4E,$67,$C0
	.dc.b	$72,$AE,$2F,$E8,$4E,$3A,$7F,$B7,$A3,$37,$FC,$1E,$72,$7F,$3B,$BE
	.dc.b	$AF,$13,$AC,$78,$67,$8D,$7C,$0C,$3F,$6F,$7C,$2B,$A5,$F0,$AF,$06
	.dc.b	$1E,$7B,$DE,$1C,$AB,$E3,$1E,$99,$D3,$4F,$14,$F0,$8C,$DF,$8D,$B7
	.dc.b	$DA,$B6,$71,$9F,$1D,$71,$8C,$E3,$C5,$FD,$39,$F4,$DC,$1C,$1E,$55
	.dc.b	$EF,$69,$E8,$AE,$46,$CB,$1B,$9E,$6D,$3C,$CE,$DF,$44,$59,$FA,$58
	.dc.b	$7C,$57,$F1,$07,$80,$B1,$D9,$F1,$4F,$23,$CD,$19,$8D,$C6,$5D,$9D
	.dc.b	$B0,$F9,$2D,$3E,$4C,$F1,$7F,$0A,$31,$FB,$D5,$D8,$E8,$FF,$9C,$BE
	.dc.b	$63,$FD,$4B,$7F,$E8,$96,$D7,$FA,$27,$C1,$B5,$BA,$FF,$39,$D6,$3C
	.dc.b	$93,$6B,$0F,$07,$FC,$21,$3F,$E8,$16,$DF,$F8,$29,$7C,$76,$E2,$4E
	.dc.b	$DE,$51,$87,$87,$0F,$53,$B7,$87,$6B,$B4,$59,$ED,$37,$2F,$F6,$1F
	.dc.b	$CF,$3E,$4F,$DA,$DC,$3F,$D2,$CB,$D8,$CC,$C9,$CB,$C2,$31,$87,$F9
	.dc.b	$2C,$75,$BD,$2F,$97,$18,$7E,$5F,$D0,$3C,$FC,$D6,$F1,$FF,$37,$B6
	.dc.b	$FA,$37,$3F,$F6,$DD,$F1,$AD,$E1,$DB,$E7,$DF,$05,$4F,$B3,$F4,$0F
	.dc.b	$8A,$C5,$FF,$9F,$2B,$6D,$AD,$CE,$B3,$F6,$56,$9F,$E5,$2D,$7F,$C4
	.dc.b	$21,$C3,$FE,$24,$6B,$F9,$3D,$FD,$57,$FC,$65,$B1,$F9,$7F,$65,$D9
	.dc.b	$F5,$33,$97,$F5,$1F,$C9,$CE,$A3,$07,$FB,$35,$7A,$FA,$37,$EB,$59
	.dc.b	$D4,$5C,$3F,$CB,$74,$AC,$6F,$FC,$D2,$7D,$9F,$41,$F3,$4F,$D1,$FD
	.dc.b	$03,$E0,$F1,$97,$CB,$9B,$5F,$E6,$6A,$D4,$E4,$91,$F2,$A7,$B3,$F1
	.dc.b	$DA,$FF,$9D,$F6,$46,$30,$75,$3C,$FF,$D7,$1F,$AF,$F4,$EE,$3F,$AF
	.dc.b	$8B,$8F,$EB,$E6,$DF,$F5,$03,$9F,$7F,$ED,$CF,$DC,$7C,$B5,$C3,$F7
	.dc.b	$11,$6F,$F7,$13,$6B,$96,$3F,$EC,$3B,$FD,$F7,$AB,$FC,$07,$E3,$FF
	.dc.b	$81,$32,$32,$7F,$03,$8C,$61,$FC,$0E,$6D,$F4,$9B,$5F,$D7,$EF,$F2
	.dc.b	$53,$E0,$E7,$FE,$1D,$E8,$7F,$0F,$65,$7F,$0F,$72,$D3,$F8,$97,$0F
	.dc.b	$F1,$53,$F8,$B7,$9F,$63,$24,$CF,$F8,$B4,$65,$FD,$1F,$C1,$7E,$BC
	.dc.b	$FE,$3D,$F4,$76,$BF,$59,$16,$7F,$8F,$4D,$CB,$F1,$9C,$6A,$FB,$F5
	.dc.b	$FA,$96,$C6,$6F,$FC,$2F,$39,$7F,$C3,$35,$F8,$2E,$3F,$E5,$C7,$F2
	.dc.b	$FE,$DB,$3F,$EF,$8F,$9F,$7C,$9F,$F1,$26,$C7,$F3,$3F,$A3,$C4,$FF
	.dc.b	$37,$EF,$93,$FC,$DF,$4B,$9F,$F3,$7E,$BF,$E7,$87,$F3,$DF,$A2,$B5
	.dc.b	$DF,$D7,$EC,$7F,$48,$FD,$EB,$FC,$4B,$19,$3F,$C3,$07,$F8,$65,$7D
	.dc.b	$1E,$CE,$1F,$25,$67,$78,$FF,$B6,$7E,$91,$F1,$5F,$CB,$1B,$3A,$E3
	.dc.b	$7E,$D1,$FA,$68,$E3,$6A,$DB,$FC,$DE,$1F,$0F,$FC,$D9,$A7,$3F,$A5
	.dc.b	$F2,$BC,$8C,$CF,$F3,$FA,$7C,$67,$8C,$FE,$CF,$FB,$4F,$F8,$D3,$2F
	.dc.b	$E1,$53,$AB,$F1,$FD,$33,$C1,$FE,$8A,$D8,$F5,$6B,$F5,$B2,$FE,$99
	.dc.b	$39,$6F,$57,$03,$FB,$AC,$BE,$3F,$17,$DB,$93,$A9,$FD,$EB,$F6,$BF
	.dc.b	$E5,$18,$74,$8F,$EF,$BF,$2E,$F5,$69,$F6,$7E,$5E,$E1,$F2,$BD,$D7
	.dc.b	$D0,$7F,$5C,$6C,$7D,$1B,$E4,$EE,$7F,$C1,$3F,$30,$FA,$7A,$78,$3C
	.dc.b	$FB,$5F,$67,$B3,$9E,$BF,$DA,$F2,$E1,$39,$C7,$C1,$84,$8F,$23,$F8
	.dc.b	$F8,$C5,$7F,$88,$67,$F1,$D3,$F9,$33,$3C,$BF,$C8,$70,$EB,$5F,$DE
	.dc.b	$7E,$97,$F6,$0D,$F0,$3F,$C4,$3F,$4D,$F6,$63,$18,$FE,$32,$B7,$87
	.dc.b	$F4,$76,$C7,$F9,$D6,$B8,$4E,$6F,$FC,$07,$5D,$AF,$FB,$4B,$51,$1C
	.dc.b	$36,$3F,$83,$FD,$93,$E7,$7F,$C1,$AD,$C7,$F4,$5F,$66,$F8,$AC,$D4
	.dc.b	$9F,$D1,$D6,$F0,$F8,$C7,$F9,$1E,$F8,$4E,$AB,$F7,$5F,$49,$23,$EB
	.dc.b	$7D,$2D,$BE,$3A,$36,$5E,$63,$E7,$55,$BF,$E0,$1F,$4B,$FC,$17,$CA
	.dc.b	$FE,$0C,$D8,$FC,$13,$71,$EB,$8A,$31,$87,$F7,$5F,$A6,$74,$7F,$92
	.dc.b	$AF,$FD,$79,$F3,$9F,$29,$4E,$BE,$1F,$FD,$87,$39,$EB,$7A,$DF,$EB
	.dc.b	$2F,$8C,$4F,$1B,$E1,$87,$43,$EA,$B8,$9D,$C6,$31,$F1,$46,$23,$F5
	.dc.b	$97,$D6,$FA,$56,$9B,$8F,$92,$B9,$35,$E3,$AD,$C7,$C1,$5C,$27,$C8
	.dc.b	$7B,$AF,$A3,$F8,$AC,$BD,$B7,$C6,$73,$3B,$C3,$07,$7A,$DC,$74,$C3
	.dc.b	$BB,$E4,$3F,$DD,$2F,$C4,$7F,$C8,$2D,$7C,$87,$F9,$13,$E3,$BF,$96
	.dc.b	$FD,$3F,$F5,$A6,$C7,$FB,$C9,$AE,$FC,$B4,$FF,$60,$AF,$E1,$FF,$07
	.dc.b	$7B,$7F,$C8,$B7,$E2,$6B,$E2,$6B,$F6,$5D,$EF,$F4,$EB,$9C,$FB,$67
	.dc.b	$3B,$7D,$FF,$96,$FF,$80,$0F,$1C,$3E,$84,$EA,$BF,$02,$D7,$F1,$B5
	.dc.b	$C1,$CC,$F0,$23,$F2,$E3,$4F,$E5,$27,$95,$F9,$30,$F3,$46,$2B,$73
	.dc.b	$9C,$3C,$97,$E9,$C3,$F0,$95,$E1,$E1,$78,$2E,$6F,$D1,$47,$F8,$FE
	.dc.b	$31,$7F,$D9,$DA,$FE,$E9,$FD,$CB,$9E,$7E,$3C,$7F,$C5,$2B,$AD,$E2
	.dc.b	$BF,$D9,$D7,$CA,$7B,$5E,$5F,$1F,$2D,$E6,$7A,$34,$D5,$FE,$F7,$0F
	.dc.b	$25,$60,$8F,$9F,$A3,$C0,$BE,$9F,$FD,$38,$FF,$89,$9A,$FF,$A5,$0D
	.dc.b	$7A,$9F,$E2,$A6,$9B,$79,$C7,$0C,$BF,$D0,$0D,$D7,$F8,$57,$99,$FE
	.dc.b	$6F,$BF,$F0,$5F,$D3,$FA,$9F,$F0,$6F,$D3,$FE,$FF,$0F,$F4,$43,$6F
	.dc.b	$F8,$95,$6E,$BF,$D8,$CD,$7F,$D5,$2D,$77,$AF,$9E,$31,$1F,$D1,$1E
	.dc.b	$57,$FE,$95,$6D,$70,$9F,$1F,$8D,$F3,$2E,$47,$DF,$7E,$4F,$D3,$BD
	.dc.b	$5A,$7C,$57,$13,$FE,$50,$6B,$F2,$DA,$FE,$33,$83,$FA,$93,$FE,$2E
	.dc.b	$8D,$E5,$FE,$A9,$6E,$3F,$89,$AD,$D7,$F4,$8F,$2B,$FD,$AE,$FF,$82
	.dc.b	$7D,$FF,$C4,$46,$23,$FA,$6F,$CD,$7E,$FB,$1E,$9E,$17,$9C,$93,$E7
	.dc.b	$7F,$CC,$F0,$96,$BF,$93,$6D,$7F,$91,$DA,$FE,$87,$E3,$FF,$1B,$F1
	.dc.b	$1F,$93,$8F,$F7,$E3,$6F,$FD,$88,$DD,$3F,$A7,$7D,$3B,$E6,$B8,$FC
	.dc.b	$FF,$E3,$E8,$D2,$61,$FE,$7C,$87,$C6,$F8,$09,$1F,$E5,$8B,$FF,$6A
	.dc.b	$3F,$E5,$AF,$A5,$FF,$68,$37,$E1,$FF,$47,$18,$8F,$EE,$7F,$4E,$F8
	.dc.b	$2F,$D3,$FD,$3F,$FB,$22,$7F,$D1,$6D,$7F,$4C,$D7,$FD,$F6,$D7,$FB
	.dc.b	$55,$3E,$27,$B9,$23,$FC,$E1,$F3,$9D,$DF,$E1,$9F,$2D,$FB,$FF,$4A
	.dc.b	$46,$0F,$F3,$9F,$D3,$FF,$A8,$5F,$F4,$EB,$5F,$DC,$35,$FE,$F4,$1C
	.dc.b	$3F,$8B,$F7,$FF,$84,$91,$FE,$39,$F0,$9F,$C3,$3C,$AF,$E2,$1C,$F2
	.dc.b	$7F,$90,$47,$F0,$AB,$F4,$47,$83,$6F,$E0,$D1,$88,$F0,$B5,$FE,$45
	.dc.b	$F3,$5F,$93,$E7,$BE,$67,$F1,$F8,$C4,$BA,$51,$F8,$E5,$5F,$E1,$5F
	.dc.b	$A2,$FD,$CD,$6F,$CB,$70,$7F,$27,$3F,$8F,$CE,$A4,$FF,$04,$6F,$89
	.dc.b	$FC,$76,$30,$47,$89,$23,$FD,$4D,$8B,$E9,$FD,$1E,$7C,$9C,$BE,$BF
	.dc.b	$96,$C4,$7F,$8F,$AB,$F8,$E3,$ED,$6E,$9E,$E9,$F8,$5D,$B2,$BB,$BF
	.dc.b	$17,$E7,$B1,$1C,$E2,$D7,$C4,$8E,$93,$F2,$FE,$6E,$47,$A1,$F9,$DF
	.dc.b	$3F,$88,$F2,$45,$AE,$47,$D0,$9C,$70,$7D,$89,$B5,$E9,$9F,$B1,$EC
	.dc.b	$7F,$55,$F0,$A4,$FC,$68,$CB,$48,$7C,$3F,$A7,$FC,$07,$D3,$C8,$F9
	.dc.b	$AF,$E0,$FE,$9F,$11,$F2,$62,$E7,$E6,$DD,$5F,$7B,$1F,$EB,$BF,$4F
	.dc.b	$FC,$86,$BE,$26,$08,$C6,$BF,$89,$AE,$13,$FD,$7C,$FC,$2B,$F8,$32
	.dc.b	$39,$E4,$FE,$26,$FF,$B9,$1A,$FF,$A0,$8E,$27,$FB,$A2,$9F,$33,$FC
	.dc.b	$A7,$E6,$64,$FF,$29,$7F,$BF,$7D,$37,$ED,$B2,$FA,$3E,$5F,$F0,$D6
	.dc.b	$CF,$0A,$FF,$43,$D6,$F3,$FF,$06,$3F,$ED,$18,$FF,$7D,$3F,$DB,$DD
	.dc.b	$FC,$5F,$6F,$82,$91,$A6,$BD,$3E,$77,$F4,$FF,$4F,$F2,$BE,$7A,$4F
	.dc.b	$BE,$32,$FC,$37,$F5,$04,$70,$6C,$87,$C5,$DB,$C9,$A7,$D4,$D7,$E2
	.dc.b	$CF,$F8,$97,$4F,$F1,$37,$86,$F7,$7F,$C4,$4D,$9D,$1F,$EB,$EA,$E2
	.dc.b	$FD,$16,$DF,$5E,$0F,$F0,$FC,$63,$0E,$BD,$FF,$43,$E1,$FE,$8F,$9F
	.dc.b	$31,$F9,$31,$96,$90,$FA,$3E,$67,$F1,$7F,$A2,$FE,$CF,$D2,$69,$27
	.dc.b	$FC,$55,$A7,$F1,$A8,$E1,$1F,$C2,$9F,$E4,$1E,$9F,$F2,$1F,$A7,$FF
	.dc.b	$1F,$7C,$9C,$AF,$E1,$F1,$96,$90,$FE,$19,$F4,$FF,$CA,$7E,$3B,$C6
	.dc.b	$48,$F0,$E5,$FE,$DB,$6C,$6D,$FE,$36,$FF,$31,$F4,$BF,$C8,$BF,$16
	.dc.b	$F1,$3F,$99,$57,$F8,$0B,$A7,$FD,$88,$3E,$97,$FD,$8D,$ED,$23,$31
	.dc.b	$87,$FD,$95,$7C,$F8,$3F,$DE,$B5,$BB,$FF,$7B,$CE,$57,$D3,$E5,$E9
	.dc.b	$87,$FB,$4E,$31,$87,$86,$7F,$DF,$F1,$BA,$E4,$DC,$7D,$2F,$0B,$E5
	.dc.b	$7E,$6B,$E2,$F9,$AB,$6D,$C7,$32,$3C,$66,$0A,$7A,$23,$83,$4F,$6C
	.dc.b	$FC,$AC,$7C,$51,$F0,$A7,$C4,$FC,$D3,$75,$96,$9C,$CE,$F8,$72,$71
	.dc.b	$8F,$C0,$AF,$0C,$E4,$8E,$0F,$2D,$F1,$23,$83,$55,$EF,$7E,$73,$E1
	.dc.b	$6C,$FC,$DB,$F8,$57,$E1,$AE,$18,$FC,$E6,$9A,$C7,$E6,$BE,$CC,$7C
	.dc.b	$2E,$8F,$84,$2B,$F0,$91,$DE,$3C,$55,$D5,$F8,$78,$FA,$1E,$17,$D9
	.dc.b	$C9,$D7,$1E,$44,$9F,$A2,$FC,$68,$F9,$EF,$7B,$F2,$C7,$DE,$71,$3D
	.dc.b	$CF,$86,$72,$3C,$58,$7A,$1F,$BB,$1C,$A3,$13,$F0,$67,$24,$70,$7E
	.dc.b	$03,$D0,$FF,$36,$CF,$E9,$5B,$87,$ED,$1B,$97,$E5,$5F,$F3,$9D,$7E
	.dc.b	$5F,$E3,$31,$8F,$FA,$ED,$B1,$FE,$44,$6B,$F9,$D0,$E1,$E4,$7D,$4A
	.dc.b	$79,$7F,$E0,$FA,$FD,$B4,$E9,$FD,$C4,$EB,$A7,$F3,$53,$D6,$EC,$E6
	.dc.b	$3E,$F9,$C7,$3F,$F7,$7E,$1D,$5F,$44,$7E,$86,$78,$69,$E0,$8F,$39
	.dc.b	$9A,$FF,$11,$46,$6F,$BC,$73,$9F,$80,$7A,$DC,$3E,$53,$96,$1F,$BD
	.dc.b	$7E,$0D,$7B,$27,$AE,$1E,$48,$C5,$6E,$BC,$71,$87,$F7,$07,$29,$F6
	.dc.b	$CE,$EF,$F8,$56,$1E,$C9,$F7,$E3,$C4,$F5,$5F,$9E,$FE,$5B,$CC,$FE
	.dc.b	$29,$5C,$5F,$6E,$1E,$13,$89,$6F,$D0,$78,$A3,$CA,$7C,$AB,$F0,$47
	.dc.b	$88,$D3,$C9,$E1,$36,$7C,$38,$FC,$47,$C3,$3A,$3B,$BA,$C3,$E0,$3E
	.dc.b	$B3,$F0,$27,$86,$3B,$9C,$1F,$F4,$14,$F5,$7E,$B5,$7A,$DE,$E7,$53
	.dc.b	$D7,$8F,$CC,$9E,$11,$F1,$CE,$77,$E2,$AE,$51,$F8,$1C,$3E,$0B,$F8
	.dc.b	$66,$E3,$FE,$92,$FA,$7F,$F3,$2D,$FF,$A6,$9B,$5F,$E7,$BF,$16,$D6
	.dc.b	$EB,$FD,$0F,$5F,$E5,$2A,$E1,$1F,$DD,$23,$11,$FE,$83,$6C,$FF,$B3
	.dc.b	$E7,$FC,$3E,$D9,$FF,$52,$47,$E0,$29,$F1,$FE,$07,$FA,$B4,$70,$6C
	.dc.b	$F0,$3F,$D4,$DE,$C6,$DF,$3F,$A8,$8E,$5C,$52,$F9,$78,$D6,$F9,$64
	.dc.b	$C7,$FC,$05,$5F,$E1,$E6,$CF,$F9,$6D,$AF,$F0,$D9,$E1,$7E,$EF,$A6
	.dc.b	$7B,$1F,$21,$B8,$7C,$07,$FD,$76,$D8,$D1,$BF,$32,$C6,$DF,$A7,$87
	.dc.b	$F8,$CF,$DF,$BD,$E7,$EA,$FD,$13,$D3,$CC,$4C,$3E,$7C,$FF,$88,$5F
	.dc.b	$F2,$7D,$F3,$73,$D3,$DB,$E8,$DF,$09,$4F,$C0,$FD,$3B,$E3,$71,$7C
	.dc.b	$AB,$6F,$FA,$3D,$B8,$7F,$BD,$83,$87,$FD,$F2,$D7,$F9,$B3,$C3,$87
	.dc.b	$FD,$6F,$6F,$96,$EF,$5F,$5A,$31,$7F,$2F,$4F,$9C,$7E,$5B,$DF,$77
	.dc.b	$E2,$74,$7E,$42,$1D,$79,$C3,$F0,$31,$89,$D7,$B7,$CC,$E9,$ED,$D8
	.dc.b	$C9,$2B,$8C,$65,$FF,$54,$6A,$F9,$CF,$98,$7E,$AB,$E9,$DE,$76,$30
	.dc.b	$F9,$53,$FA,$8B,$F8,$ED,$BE,$53,$F3,$63,$07,$3F,$37,$95,$7E,$6B
	.dc.b	$04,$7E,$A7,$D7,$97,$FA,$9F,$5E,$DF,$E8,$3B,$95,$8E,$4F,$F7,$3C
	.dc.b	$2B,$17,$FB,$5F,$1F,$F6,$87,$ED,$7E,$9E,$C7,$ED,$63,$F6,$3F,$07
	.dc.b	$8C,$BF,$5B,$E2,$23,$F7,$9E,$AC,$BF,$DE,$7C,$85,$BF,$24,$DD,$3C
	.dc.b	$73,$6B,$F8,$34,$DA,$57,$EE,$6B,$0F,$EE,$FB,$3F,$85,$1F,$C2,$FD
	.dc.b	$4B,$87,$F0,$B8,$B7,$F0,$E6,$D7,$C2,$E4,$FE,$24,$7F,$13,$E1,$B8
	.dc.b	$7F,$13,$8B,$7F,$2E,$6D,$7F,$BD,$B9,$3F,$2F,$3F,$84,$EC,$FE,$39
	.dc.b	$F4,$FF,$AF,$66,$3F,$A9,$7F,$00,$47,$07,$F2,$8D,$DE,$C7,$0B,$F9
	.dc.b	$AF,$F2,$7E,$79,$5F,$C9,$FC,$1B,$5F,$EF,$BF,$A7,$FF,$69,$4F,$ED
	.dc.b	$61,$18,$7F,$70,$D7,$FC,$54,$6B,$DA,$FE,$77,$D1,$FC,$0E,$BC,$9E
	.dc.b	$E7,$F3,$7F,$0F,$F9,$8B,$FC,$E7,$BE,$4F,$F3,$9F,$9F,$B8,$F8,$FD
	.dc.b	$BF,$D0,$66,$69,$2B,$FD,$9F,$BB,$E9,$BF,$9E,$F8,$BF,$CF,$9B,$3F
	.dc.b	$C7,$3E,$8F,$FA,$03,$7D,$0F,$EB,$3F,$4D,$FC,$F6,$30,$7F,$22,$1F
	.dc.b	$0D,$FB,$16,$9F,$14,$BA,$7F,$1B,$6B,$F1,$2F,$F3,$91,$E1,$F0,$3F
	.dc.b	$B3,$FD,$2F,$F9,$0F,$AD,$3F,$E5,$3A,$DD,$FF,$59,$1F,$4A,$F0,$7F
	.dc.b	$9F,$B6,$BF,$B3,$E9,$F5,$70,$E7,$7F,$EB,$9A,$D3,$CD,$FE,$0E,$FF
	.dc.b	$31,$6B,$F6,$9A,$FF,$99,$5A,$F3,$E2,$F0,$9F,$E0,$0F,$A5,$FE,$57
	.dc.b	$87,$F4,$8A,$4F,$F8,$13,$E5,$1E,$B5,$3F,$B8,$FC,$A5,$BF,$E9,$FE
	.dc.b	$73,$E7,$FF,$8F,$36,$3F,$24,$DC,$FF,$A4,$35,$FE,$DD,$F2,$CF,$A5
	.dc.b	$FF,$09,$B6,$BF,$C1,$79,$7F,$1A,$7F,$C1,$B1,$9A,$FE,$23,$87,$F6
	.dc.b	$AC,$31,$B7,$CA,$1D,$1E,$17,$E6,$D7,$F4,$6C,$B8,$1E,$82,$63,$E8
	.dc.b	$7F,$C3,$B5,$D2,$3F,$CB,$2D,$9F,$F7,$75,$DB,$EF,$7F,$DF,$3E,$93
	.dc.b	$F8,$2C,$78,$1B,$87,$F9,$8B,$E9,$7F,$0F,$39,$F1,$78,$1F,$D1,$EB
	.dc.b	$F8,$6E,$3F,$B6,$6B,$FE,$A1,$6B,$FE,$FF,$7A,$BF,$CA,$3E,$A1,$F1
	.dc.b	$5D,$31,$FF,$1A,$C6,$1F,$F0,$3F,$D3,$3E,$5F,$FC,$26,$D8,$FF,$1D
	.dc.b	$FD,$33,$E2,$73,$52,$7F,$87,$56,$F1,$FF,$02,$35,$FE,$1A,$D7,$E4
	.dc.b	$FD,$1F,$F8,$E8,$FF,$1D,$FD,$25,$AF,$E4,$51,$FE,$54,$FD,$EF,$F2
	.dc.b	$B5,$FD,$3A,$31,$EE,$FF,$90,$BE,$97,$FC,$87,$E6,$DC,$3F,$95,$38
	.dc.b	$69,$FC,$B6,$31,$7F,$B5,$FA,$67,$5F,$F9,$52,$3F,$D9,$7F,$35,$FE
	.dc.b	$CD,$9C,$FC,$4F,$FB,$56,$3F,$DA,$1D,$CF,$63,$F6,$FF,$1A,$9E,$18
	.dc.b	$F3,$8C,$62,$18,$F4,$7B,$99,$71,$1C,$23,$F6,$37,$DA,$FA,$E3,$E3
	.dc.b	$9E,$18,$77,$9C,$C7,$43,$8E,$7B,$7C,$33,$B7,$8D,$67,$1F,$0D,$FB
	.dc.b	$B0,$F0,$56,$8E,$93,$A7,$18,$F9,$F2,$F8,$99,$73,$CB,$B5,$78,$32
	.dc.b	$FE,$A9,$87,$C1,$8F,$EA,$2D,$7D,$27,$F5,$4F,$8E,$FE,$73,$F4,$FF
	.dc.b	$D6,$DB,$1F,$E5,$16,$BB,$F2,$D3,$DB,$C1,$FF,$23,$CF,$F9,$5E,$FF
	.dc.b	$8B,$70,$E4,$91,$99,$DE,$BF,$46,$B7,$3F,$02,$73,$BF,$E1,$7D,$7F
	.dc.b	$B9,$39,$C3,$E5,$56,$A3,$EE,$35,$FC,$0C,$F0,$C3,$B4,$E7,$0E,$F1
	.dc.b	$8B,$FA,$95,$BC,$3F,$29,$7B,$24,$EB,$0F,$0B,$1F,$21,$F6,$2B,$93
	.dc.b	$FE,$2C,$F0,$DC,$DF,$E9,$FE,$0F,$F7,$AC,$62,$7F,$97,$CE,$5E,$1C
	.dc.b	$B7,$BD,$5B,$9F,$F5,$7B,$63,$F9,$93,$6F,$AD,$71,$87,$2F,$3F,$0F
	.dc.b	$FC,$0D,$87,$F8,$8F,$D0,$D4,$C8,$C1,$0E,$EF,$95,$7D,$3F,$FA,$6D
	.dc.b	$FF,$51,$F9,$6F,$7F,$4D,$8F,$F5,$3C,$ED,$FE,$F0,$DD,$3F,$A3,$CE
	.dc.b	$E7,$FB,$E3,$63,$FA,$5C,$EF,$4F,$C1,$FD,$37,$AA,$FF,$C0,$5F,$4F
	.dc.b	$FB,$BB,$FE,$A2,$DA,$DE,$7E,$9B,$E1,$5A,$E3,$F6,$5E,$F0,$FE,$15
	.dc.b	$5A,$C3,$8C,$7F,$8F,$E3,$C0,$FF,$8F,$1A,$E8,$FF,$1F,$DF,$D8,$C3
	.dc.b	$ED,$77,$7F,$AA,$27,$FD,$F2,$D7,$ED,$C7,$CD,$FA,$67,$D3,$FD,$98
	.dc.b	$93,$FE,$D0,$9C,$B6,$78,$11,$FE,$F3,$E1,$7E,$27,$CC,$FF,$69,$C7
	.dc.b	$83,$E8,$7C,$66,$7E,$0F,$C1,$E0,$FE,$3F,$E6,$FF,$77,$5F,$EF,$C6
	.dc.b	$EB,$FD,$F3,$3F,$E7,$73,$FC,$BE,$3F,$C0,$53,$EB,$75,$5D,$27,$5B
	.dc.b	$EE,$B4,$FE,$F3,$E2,$BF,$2B,$3F,$EE,$CF,$A7,$7C,$DF,$E3,$DB,$1F
	.dc.b	$EF,$09,$DB,$FE,$41,$5F,$25,$BF,$A1,$93,$B8,$FC,$C4,$FF,$BF,$E3
	.dc.b	$B7,$CC,$F5,$33,$EF,$FE,$77,$E9,$7F,$D9,$B3,$FE,$FE,$6B,$FE,$4F
	.dc.b	$6E,$DF,$E2,$1F,$A7,$7C,$1F,$E9,$7E,$9D,$F0,$FF,$A7,$C9,$F3,$14
	.dc.b	$D4,$77,$F8,$D9,$9F,$77,$F8,$07,$CE,$F4,$8F,$F1,$9F,$D3,$7E,$E7
	.dc.b	$AE,$46,$0F,$9A,$7F,$61,$FA,$7F,$F8,$00,$FE,$3B,$C2,$FC,$89,$FB
	.dc.b	$98,$CC,$7F,$0A,$F8,$4F,$CA,$FC,$BC,$8F,$F2,$0C,$7F,$09,$F3,$7F
	.dc.b	$84,$6B,$27,$FC,$8B,$1C,$F8,$3F,$1D,$AC,$9F,$18,$C7,$D4,$FF,$B3
	.dc.b	$BA,$BF,$11,$F4,$1F,$8C,$7F,$15,$E2,$3A,$3F,$4E,$7F,$21,$9D,$57
	.dc.b	$F7,$DF,$A5,$FD,$A3,$FE,$D8,$F3,$1C,$5F,$C7,$4F,$E2,$D3,$A9,$3C
	.dc.b	$3A,$3F,$CA,$F3,$B9,$F3,$CE,$CD,$57,$48,$FF,$64,$4F,$4F,$1B,$C9
	.dc.b	$C7,$27,$E8,$E0,$C7,$F8,$53,$78,$F5,$3F,$04,$FF,$77,$CE,$89,$1E
	.dc.b	$FC,$E4,$F4,$9F,$88,$DE,$7C,$BC,$EF,$C1,$58,$7F,$CE,$94,$AF,$95
	.dc.b	$F3,$92,$FA,$3F,$37,$E7,$31,$1C,$E9,$38,$BE,$75,$78,$3E,$CC,$DB
	.dc.b	$F5,$CF,$D7,$F5,$FF,$A7,$77,$C9,$F5,$46,$5F,$F7,$94,$65,$F0,$47
	.dc.b	$DF,$F0,$24,$7C,$17,$F0,$5F,$4D,$88,$F3,$45,$CB,$DF,$3F,$80,$BF
	.dc.b	$EA,$DF,$4B,$FC,$82,$FF,$D8,$CD,$AF,$F4,$AF,$27,$F0,$89,$FC,$0D
	.dc.b	$70,$F2,$24,$60,$87,$27,$2F,$FB,$46,$79,$3F,$C9,$FB,$FF,$95,$11
	.dc.b	$FC,$2F,$D3,$7F,$44,$DF,$F9,$14,$E6,$78,$1A,$3F,$C0,$13,$FE,$C4
	.dc.b	$78,$FC,$EE,$0C,$FB,$D2,$32,$43,$F0,$93,$C5,$F3,$7C,$A7,$C4,$F5
	.dc.b	$24,$7E,$82,$3F,$4E,$61,$F8,$F1,$87,$47,$EF,$36,$FE,$E9,$97,$F8
	.dc.b	$D1,$B1,$FB,$5A,$FF,$03,$79,$6F,$9B,$FF,$0D,$78,$4F,$82,$A6,$A2
	.dc.b	$5F,$EA,$A7,$4D,$9C,$BE,$CF,$82,$F8,$BF,$37,$23,$E6,$C7,$C8,$30
	.dc.b	$FC,$28,$C3,$FA,$BF,$A0,$FE,$E7,$D9,$23,$04,$3F,$59,$E1,$FE,$A5
	.dc.b	$FE,$EF,$1F,$C4,$E7,$F8,$A7,$A1,$FE,$2B,$FA,$69,$5F,$B7,$8C,$BF
	.dc.b	$C1,$3C,$54,$EA,$3F,$5D,$3F,$C7,$7E,$1F,$FC,$0B,$E0,$49,$FE,$01
	.dc.b	$19,$7F,$C0,$F1,$97,$F6,$D1,$FE,$33,$F8,$57,$82,$EE,$FF,$51,$7D
	.dc.b	$2F,$F8,$AF,$A1,$F4,$9F,$E6,$A8,$C3,$63,$FC,$DB,$7C,$AB,$8C,$62
	.dc.b	$FA,$62,$C3,$8B,$A3,$FD,$61,$3D,$9C,$0B,$E8,$F1,$8E,$97,$A7,$84
	.dc.b	$64,$E7,$87,$95,$FF,$5E,$57,$2B,$FF,$23,$36,$BF,$D0,$A7,$28,$F2
	.dc.b	$CF,$33,$C5,$87,$AA,$38,$B6,$78,$DF,$7A,$FF,$2D,$36,$FB,$47,$8A
	.dc.b	$B8,$3E,$63,$D5,$3B,$8C,$C7,$83,$0C,$B0,$F2,$57,$47,$47,$9E,$FD
	.dc.b	$47,$91,$F9,$71,$B3,$84,$6A,$3C,$47,$4A,$F0,$CE,$AF,$D9,$5B,$AE
	.dc.b	$F8,$7C,$2C,$3C,$8F,$C8,$31,$1C,$E3,$FD,$D9,$F2,$3F,$30,$95,$F6
	.dc.b	$A5,$1C,$2B,$94,$FA,$67,$91,$23,$2F,$A1,$E1,$1C,$DC,$C7,$D2,$3E
	.dc.b	$0E,$3E,$C3,$E6,$CF,$AC,$F6,$63,$F1,$63,$38,$FC,$03,$EA,$63,$C8
	.dc.b	$F7,$D7,$97,$0F,$0E,$1F,$76,$4B,$E1,$3F,$CE,$D8,$FE,$69,$B8,$FE
	.dc.b	$91,$B5,$C1,$FF,$3E,$57,$2F,$86,$C6,$7F,$ED,$A8,$C1,$D5,$C2,$8D
	.dc.b	$BB,$6B,$FE,$DB,$3F,$39,$5F,$DF,$7B,$3F,$5F,$1D,$32,$FF,$54,$71
	.dc.b	$A7,$33,$D4,$E3,$19,$7F,$4C,$71,$D3,$FD,$DF,$7D,$63,$27,$1B,$EB
	.dc.b	$7D,$30,$FD,$FD,$6E,$7F,$D7,$55,$CC,$FF,$43,$61,$DE,$33,$1C,$F0
	.dc.b	$EF,$39,$7E,$1D,$F4,$8C,$4F,$C3,$3D,$38,$78,$5A,$79,$A7,$84,$67
	.dc.b	$1F,$3C,$FE,$5A,$FA,$B4,$F2,$CF,$E7,$A3,$38,$70,$3C,$F7,$D2,$BB
	.dc.b	$1E,$08,$CB,$67,$89,$EC,$AF,$34,$F3,$8C,$CF,$80,$E7,$19,$7D,$CD
	.dc.b	$7F,$00,$D7,$E6,$1F,$CF,$F0,$F8,$AE,$A7,$C0,$F4,$6C,$F9,$8E,$51
	.dc.b	$F4,$4F,$13,$98,$D9,$E6,$8E,$D5,$EA,$8C,$C7,$99,$AF,$84,$E6,$F1
	.dc.b	$C7,$F0,$8D,$BF,$B7,$1F,$E6,$7F,$A5,$FC,$93,$39,$7F,$AD,$E1,$FE
	.dc.b	$61,$9C,$BF,$E8,$6D,$6E,$3F,$E6,$A9,$B6,$BE,$33,$FE,$5B,$9C,$C6
	.dc.b	$27,$BF,$83,$FB,$EC,$3F,$BB,$7A,$5C,$A3,$18,$7F,$9B,$79,$93,$98
	.dc.b	$EF,$18,$AF,$F4,$0D,$70,$BF,$5C,$E6,$BF,$CE,$D3,$FD,$CA,$BF,$E0
	.dc.b	$0F,$91,$DF,$8F,$CA,$D8,$C9,$D2,$32,$FF,$A3,$BE,$97,$FD,$96,$DF
	.dc.b	$1F,$E0,$C6,$30,$FF,$42,$3F,$EE,$49,$E1,$E9,$EA,$B6,$D8,$FF,$5B
	.dc.b	$F5,$BD,$9F,$97,$F3,$DF,$1E,$7F,$91,$E2,$AD,$47,$FA,$82,$75,$3F
	.dc.b	$CA,$27,$2F,$FA,$56,$32,$F1,$AE,$BC,$5F,$6A,$1F,$42,$F2,$FE,$DF
	.dc.b	$D2,$BE,$53,$53,$9C,$FF,$D5,$F3,$97,$FD,$9F,$19,$BF,$F5,$24,$E5
	.dc.b	$FF,$58,$1D,$AF,$AD,$F4,$6D,$F2,$6C,$FC,$AC,$7F,$1B,$EC,$7F,$5E
	.dc.b	$AF,$A5,$18,$EC,$FA,$2D,$8C,$9F,$EC,$49,$E1,$5F,$EB,$75,$E9,$F4
	.dc.b	$3A,$FA,$76,$B8,$45,$AE,$13,$79,$FF,$9C,$47,$C8,$FA,$1F,$D0,$FD
	.dc.b	$2B,$DA,$F8,$F1,$86,$E3,$FE,$D6,$9D,$3F,$22,$74,$D3,$53,$C2,$BF
	.dc.b	$D9,$D5,$C1,$FC,$DF,$A5,$FA,$73,$F5,$1F,$4B,$6F,$F4,$71,$67,$F4
	.dc.b	$33,$71,$FD,$8F,$83,$FA,$AF,$9F,$FD,$83,$70,$FD,$5B,$6F,$F5,$2D
	.dc.b	$AF,$DC,$F9,$FF,$B7,$3F,$71,$F1,$76,$FF,$67,$16,$FF,$63,$36,$B8
	.dc.b	$D6,$8F,$DC,$C6,$5F,$F0,$0C,$3B,$FF,$18,$FF,$08,$E4,$FE,$0D,$E7
	.dc.b	$7F,$07,$65,$7F,$02,$72,$D3,$F8,$2F,$9D,$FC,$34,$FE,$1D,$C1,$63
	.dc.b	$24,$CB,$F8,$3C,$65,$FC,$CF,$4F,$E6,$F1,$FE,$1B,$E8,$7E,$93,$A7
	.dc.b	$F8,$9B,$70,$FF,$48,$36,$FF,$8D,$36,$B8,$CF,$EA,$2B,$8B,$FC,$6B
	.dc.b	$93,$F9,$3F,$D3,$FF,$23,$6C,$7C,$43,$F8,$1F,$D1,$FF,$AE,$DC,$63
	.dc.b	$FE,$91,$F8,$F7,$C5,$FE,$F6,$B5,$3F,$E9,$5A,$F9,$B5,$A7,$F9,$4F
	.dc.b	$D2,$7F,$2C,$F5,$FF,$99,$36,$3F,$96,$F6,$FF,$3A,$F7,$7F,$9B,$36
	.dc.b	$3C,$38,$31,$7C,$07,$F4,$0F,$7B,$C4,$DA,$FF,$5A,$7D,$27,$F3,$AF
	.dc.b	$0B,$F9,$DB,$63,$9F,$D1,$7F,$39,$E3,$FD,$7B,$63,$F0,$BF,$49,$F8
	.dc.b	$57,$7E,$87,$F1,$B5,$E7,$FF,$54,$FA,$8F,$76,$DF,$C4,$A7,$27,$C8
	.dc.b	$C5,$3A,$7B,$F8,$5F,$C9,$72,$3F,$D7,$BE,$97,$FC,$69,$3A,$BF,$83
	.dc.b	$F4,$8F,$0B,$FB,$1E,$9F,$0A,$73,$18,$E0,$FE,$61,$4E,$9D,$DE,$1B
	.dc.b	$7D,$3E,$93,$FB,$6F,$D2,$FF,$1A,$CB,$F8,$8E,$09,$CF,$99,$FD,$CB
	.dc.b	$E5,$1F,$0D,$FC,$F6,$1F,$37,$98,$C3,$32,$F8,$54,$F3,$31,$C9,$F8
	.dc.b	$73,$FD,$B9,$19,$9F,$91,$DE,$FA,$5F,$EF,$BE,$A7,$F8,$0A,$31,$3F
	.dc.b	$37,$DA,$FE,$B0,$FE,$1B,$4C,$E3,$FD,$11,$BB,$FF,$78,$D3,$D1,$94
	.dc.b	$BF,$F7,$D3,$6F,$FC,$08,$FF,$80,$8F,$F0,$6F,$D1,$D9,$F0,$3F,$D7
	.dc.b	$7E,$93,$FA,$95,$7F,$A0,$5A,$FF,$6C,$6B,$FE,$90,$6A,$7F,$80,$4F
	.dc.b	$F1,$4F,$A4,$57,$5F,$23,$FC,$43,$18,$9F,$D1,$D7,$0B,$64,$AD,$47
	.dc.b	$F6,$49,$D4,$E7,$0F,$EE,$9F,$46,$F9,$2F,$F1,$6F,$CA,$BC,$FC,$C6
	.dc.b	$1B,$9F,$E6,$E7,$4F,$ED,$2B,$93,$99,$FF,$14,$FC,$FF,$F8,$D3,$E1
	.dc.b	$AC,$FD,$76,$E9,$F8,$1C,$FF,$C7,$91,$87,$FC,$8C,$FF,$8C,$BE,$32
	.dc.b	$57,$EA,$FE,$0D,$F3,$F8,$C3,$FC,$75,$57,$0F,$F1,$AC,$DB,$FF,$0E
	.dc.b	$FD,$0B,$E6,$DF,$47,$FE,$B9,$F9,$6C,$C6,$32,$FF,$5B,$47,$F8,$93
	.dc.b	$B9,$ED,$7F,$AA,$FE,$31,$3C,$30,$E5,$7F,$EC,$CC,$7B,$46,$21,$18
	.dc.b	$7A,$9F,$ED,$2C,$3A,$E5,$B9,$D6,$1D,$AB,$A6,$5E,$58,$C4,$74,$8C
	.dc.b	$63,$E7,$CB,$CC,$F9,$67,$B9,$BC,$BB,$57,$08,$F1,$57,$07,$BC,$E8
	.dc.b	$8F,$2C,$BC,$EF,$4C,$7C,$F8,$77,$8F,$F7,$14,$7B,$DB,$87,$CD,$6D
	.dc.b	$FF,$B9,$17,$C6,$FF,$2C,$FA,$3C,$36,$3F,$DD,$2D,$7F,$DD,$CD,$7E
	.dc.b	$EF,$67,$FB,$ED,$FF,$77,$E1,$FC,$53,$8B,$B3,$63,$33,$B7,$CE,$F1
	.dc.b	$D7,$E6,$CE,$DB,$19,$CF,$F1,$7E,$D7,$F9,$A2,$73,$87,$D9,$AD,$61
	.dc.b	$F8,$36,$BF,$82,$AE,$17,$E8,$AE,$04,$9F,$99,$3A,$79,$46,$6F,$BC
	.dc.b	$68,$C5,$73,$7E,$A5,$F8,$30,$61,$CF,$87,$FD,$55,$F2,$3F,$E8,$07
	.dc.b	$FD,$69,$F5,$17,$8B,$FC,$4B,$E0,$A3,$19,$7F,$0E,$85,$E2,$F9,$CF
	.dc.b	$F3,$38,$C4,$EA,$3F,$9B,$7C,$8E,$3E,$6E,$47,$F3,$97,$7F,$21,$23
	.dc.b	$04,$3F,$83,$52,$3C,$70,$FC,$39,$58,$6C,$7F,$82,$1A,$F2,$F8,$4F
	.dc.b	$29,$B9,$91,$FD,$2E,$39,$54,$8E,$4F,$BF,$F7,$CF,$F5,$3C,$7F,$A4
	.dc.b	$2F,$C8,$E8,$87,$F1,$D8,$DE,$7F,$EE,$26,$EB,$FE,$B0,$AE,$35,$CE
	.dc.b	$FB,$3E,$EF,$A3,$FF,$0F,$BF,$A7,$6B,$E2,$0E,$1F,$2B,$5F,$F5,$EB
	.dc.b	$C3,$0F,$F4,$6B,$6F,$3E,$97,$CC,$FA,$47,$95,$49,$87,$F8,$CA,$32
	.dc.b	$FF,$AD,$BB,$3E,$9F,$77,$AB,$2F,$F4,$CB,$71,$FF,$66,$B6,$87,$F3
	.dc.b	$3E,$1F,$B7,$DA,$69,$23,$FC,$83,$39,$7F,$C5,$9B,$FF,$0B,$C3,$BD
	.dc.b	$78,$8F,$AD,$DB,$F5,$FE,$52,$5F,$07,$ED,$67,$F8,$AF,$BC,$FE,$89
	.dc.b	$3F,$E4,$E6,$BF,$EC,$C6,$E1,$F6,$7D,$0F,$C1,$7C,$67,$E0,$FE,$32
	.dc.b	$5F,$92,$3F,$0D,$F5,$72,$3C,$EF,$E2,$7E,$37,$C1,$DF,$27,$C0,$FE
	.dc.b	$37,$DA,$FC,$77,$C6,$FE,$3F,$C7,$95,$F0,$23,$F2,$5E,$5F,$E4,$FE
	.dc.b	$36,$4F,$C4,$8F,$D2,$FD,$DF,$F8,$E6,$3F,$CD,$2D,$7F,$DC,$ED,$DF
	.dc.b	$F2,$9F,$59,$F9,$DF,$77,$F3,$DF,$2B,$2F,$E7,$C7,$E8,$7E,$B7,$F4
	.dc.b	$5F,$1B,$27,$E9,$C7,$ED,$3E,$4F,$FD,$B3,$8F,$FB,$69,$BD,$DF,$80
	.dc.b	$A4,$61,$FF,$0C,$F6,$3D,$AF,$D7,$7C,$6F,$EB,$FE,$02,$57,$E0,$23
	.dc.b	$F8,$0F,$BB,$FE,$EE,$8F,$F4,$2B,$5F,$F2,$61,$FE,$49,$F9,$4D,$F9
	.dc.b	$DC,$19,$1C,$1C,$BC,$A3,$2F,$29,$E9,$A7,$FB,$D1,$FE,$49,$1F,$E7
	.dc.b	$DF,$8C,$F5,$49,$7C,$67,$F2,$56,$CE,$F6,$C1,$99,$91,$93,$A7,$C8
	.dc.b	$F1,$F1,$E4,$60,$87,$F8,$A6,$FF,$2B,$7F,$ED,$C6,$C7,$F9,$11,$FF
	.dc.b	$5B,$BE,$7E,$CF,$26,$52,$30,$43,$F8,$BC,$FF,$9F,$1F,$E1,$D9,$7B
	.dc.b	$6B,$FB,$C3,$63,$33,$B7,$C9,$3F,$04,$F8,$5E,$DF,$C3,$FA,$BF,$89
	.dc.b	$E4,$CA,$F9,$51,$97,$E6,$46,$5F,$74,$7C,$BF,$6E,$47,$D1,$7E,$77
	.dc.b	$9B,$88,$F8,$91,$72,$69,$F1,$6B,$FA,$4F,$DD,$3C,$0A,$4C,$3F,$D9
	.dc.b	$FF,$43,$FE,$63,$7F,$DA,$33,$B6,$CE,$DF,$C7,$CE,$DB,$E0,$67,$1F
	.dc.b	$0C,$7F,$20,$BE,$4E,$8F,$C5,$7C,$0F,$E1,$FE,$6A,$46,$08,$7D,$F7
	.dc.b	$2F,$E0,$63,$FB,$1F,$CA,$FF,$6F,$8F,$D0,$AF,$9F,$C9,$AF,$6A,$4F
	.dc.b	$FB,$B2,$74,$D9,$CB,$FA,$AF,$A3,$FC,$4B,$7A,$3F,$B8,$C6,$33,$F6
	.dc.b	$8F,$13,$FC,$AA,$FC,$0A,$4E,$F8,$7F,$81,$36,$3F,$A1,$B6,$79,$7B
	.dc.b	$DF,$E0,$DB,$E7,$7C,$30,$F0,$99,$9F,$D7,$7D,$2F,$EB,$FE,$7A,$65
	.dc.b	$FB,$58,$CB,$FB,$78,$CB,$FA,$98,$FD,$B7,$D1,$7E,$DF,$E4,$64,$FE
	.dc.b	$F6,$32,$FE,$FE,$32,$FE,$C6,$3F,$7D,$F2,$72,$3F,$82,$3F,$C0,$7E
	.dc.b	$0A,$46,$08,$7E,$E5,$FE,$0D,$C1,$23,$F8,$63,$FC,$27,$BA,$46,$08
	.dc.b	$7F,$01,$7F,$C6,$5F,$BE,$FF,$96,$F8,$3E,$9B,$77,$FE,$DF,$08,$C5
	.dc.b	$7F,$B2,$1F,$EF,$6F,$FA,$CD,$B1,$CB,$E1,$BA,$9D,$BD,$EF,$F1,$8F
	.dc.b	$C3,$BE,$83,$FD,$A3,$83,$E8,$DE,$D7,$F3,$F1,$EF,$7F,$41,$8C,$CF
	.dc.b	$F8,$0A,$72,$FF,$81,$23,$2F,$FA,$12,$BA,$C6,$2F,$C1,$96,$EF,$B6
	.dc.b	$1D,$5B,$86,$4F,$F0,$7D,$70,$7C,$D1,$FD,$4E,$31,$7E,$08,$F1,$CE
	.dc.b	$8F,$35,$75,$79,$E1,$E4,$9E,$4D,$8D,$9E,$13,$E1,$CF,$17,$D9,$86
	.dc.b	$70,$F0,$DF,$A2,$7C,$07,$03,$57,$D5,$E2,$6A,$3C,$F8,$7A,$70,$ED
	.dc.b	$7D,$2B,$CF,$87,$8E,$F6,$77,$BF,$99,$87,$5B,$E1,$7E,$13,$B6,$1D
	.dc.b	$27,$47,$8E,$F9,$CF,$C0,$3D,$35,$C8,$E7,$1A,$8E,$97,$E0,$C3,$DA
	.dc.b	$79,$70,$F8,$D5,$BB,$E5,$87,$79,$C9,$E3,$AE,$38,$79,$8E,$46,$30
	.dc.b	$EC,$E4,$8E,$09,$87,$C3,$24,$7C,$83,$D7,$7E,$17,$A9,$8A,$ED,$3C
	.dc.b	$23,$DA,$FE,$12,$33,$1F,$16,$BE,$B3,$83,$ED,$E5,$FB,$2C,$7F,$2C
	.dc.b	$DC,$3F,$5A,$DB,$FF,$34,$BF,$B8,$AF,$F3,$5F,$C4,$E3,$3F,$F4,$23
	.dc.b	$63,$FD,$2E,$D7,$FD,$B0,$27,$92,$8F,$CC,$53,$E8,$FF,$AC,$E5,$FB
	.dc.b	$4A,$EE,$7E,$5A,$75,$B7,$E7,$E3,$33,$FB,$A9,$CC,$7E,$70,$CE,$9F
	.dc.b	$ED,$1B,$EA,$FA,$23,$A4,$61,$F4,$98,$BE,$CF,$9C,$C1,$D4,$C5,$F8
	.dc.b	$67,$94,$7C,$08,$C4,$F2,$C3,$DF,$5A,$AF,$29,$FC,$A1,$C3,$D2,$FD
	.dc.b	$15,$C6,$3C,$37,$F1,$5D,$1C,$63,$51,$E0,$8F,$35,$F8,$27,$35,$D2
	.dc.b	$72,$74,$7E,$25,$71,$3B,$57,$19,$DC,$66,$76,$73,$77,$1C,$CF,$96
	.dc.b	$ED,$B1,$B3,$8B,$E9,$3A,$B9,$76,$66,$3A,$9C,$DF,$9D,$1B,$69,$B7
	.dc.b	$AA,$C3,$F0,$B3,$CA,$78,$4F,$D2,$35,$87,$D2,$BF,$81,$5B,$9E,$87
	.dc.b	$F5,$BF,$AC,$FF,$09,$E3,$FD,$83,$77,$46,$2B,$53,$FD,$90,$D4,$67
	.dc.b	$3F,$EC,$CD,$AE,$33,$BA,$E2,$7F,$67,$9D,$79,$DB,$F4,$ED,$6F,$0F
	.dc.b	$EE,$11,$98,$C6,$1F,$EA,$26,$D6,$70,$FE,$E5,$A7,$87,$BB,$C4,$DA
	.dc.b	$E3,$16,$77,$18,$3A,$F0,$FA,$7E,$6F,$D4,$D8,$C9,$E2,$8C,$B4,$F2
	.dc.b	$FD,$07,$C1,$FA,$0F,$84,$D8,$C9,$EA,$8C,$B4,$F8,$1F,$46,$F5,$3E
	.dc.b	$86,$3F,$C3,$FD,$1F,$9D,$39,$AF,$99,$39,$6F,$32,$B8,$B6,$3A,$DE
	.dc.b	$26,$47,$07,$AC,$62,$BF,$D8,$7F,$AD,$BC,$EF,$1A,$F2,$2C,$61,$B8
	.dc.b	$79,$27,$4D,$9D,$3F,$6B,$E8,$7F,$0D,$F3,$5F,$87,$6C,$64,$FB,$F1
	.dc.b	$96,$9F,$03,$E7,$DE,$67,$BB,$C2,$F8,$F1,$89,$F8,$71,$96,$9C,$67
	.dc.b	$19,$FF,$B8,$36,$56,$3A,$7F,$CE,$58,$73,$AD,$4F,$E7,$7E,$62,$63
	.dc.b	$F9,$4F,$36,$D7,$F9,$0E,$6E,$1F,$E4,$C9,$B5,$FA,$DF,$67,$F4,$DF
	.dc.b	$19,$FA,$76,$C7,$E7,$9B,$87,$F9,$35,$B7,$FE,$56,$F9,$2F,$D7,$1F
	.dc.b	$AF,$F9,$0B,$7F,$A6,$8F,$F4,$55,$FE,$ED,$AF,$F9,$65,$B5,$FE,$5D
	.dc.b	$6F,$71,$FD,$D7,$C9,$74,$AF,$F2,$F3,$FB,$FF,$DF,$91,$7E,$43,$0D
	.dc.b	$C7,$F7,$53,$A6,$CE,$9A,$6B,$CA,$E9,$F2,$1D,$5B,$19,$39,$C6,$5A
	.dc.b	$7E,$8B,$F0,$67,$37,$EA,$3F,$06,$7C,$8F,$EA,$A7,$35,$FA,$89,$CB
	.dc.b	$67,$81,$BA,$B1,$92,$67,$FC,$C5,$F8,$2A,$FA,$B9,$3E,$2D,$F6,$71
	.dc.b	$BB,$4C,$3F,$EC,$25,$F2,$3F,$37,$2F,$9C,$FF,$02,$F3,$1E,$FD,$3E
	.dc.b	$97,$9D,$74,$F9,$FF,$23,$F5,$7E,$47,$EB,$36,$3E,$9F,$C8,$FD,$9F
	.dc.b	$81,$FB,$4D,$8F,$E1,$78,$BE,$0F,$C0,$7D,$FF,$91,$B3,$F7,$3E,$0B
	.dc.b	$F0,$5F,$23,$F8,$36,$C7,$E0,$3E,$43,$F0,$DF,$21,$F8,$76,$C7,$F1
	.dc.b	$8F,$C1,$5F,$17,$FC,$79,$B1,$F8,$8F,$85,$C4,$FE,$3F,$E4,$24,$FE
	.dc.b	$33,$E4,$6F,$4F,$CC,$E8,$7C,$87,$E5,$7D,$1F,$CB,$36,$FC,$1F,$1D
	.dc.b	$DC,$FC,$D7,$C8,$58,$FE,$A1,$FB,$F9,$FE,$B8,$D7,$3F,$23,$87,$3F
	.dc.b	$19,$FD,$87,$ED,$7F,$8F,$E9,$F7,$AD,$39,$F5,$FF,$9D,$FD,$B3,$D6
	.dc.b	$A6,$71,$FE,$77,$19,$7F,$B1,$63,$FC,$47,$C4,$F2,$65,$FE,$01,$8C
	.dc.b	$4F,$E4,$AB,$84,$FE,$3A,$72,$EF,$E2,$BC,$C7,$B3,$92,$DF,$F4,$AF
	.dc.b	$A0,$F9,$2D,$E1,$CD,$FE,$6C,$4E,$9C,$35,$FD,$19,$A8,$FE,$1E,$D9
	.dc.b	$CB,$FC,$22,$78,$3F,$EE,$ED,$9C,$3F,$D8,$37,$7B,$DF,$4A,$30,$CC
	.dc.b	$3F,$B0,$CE,$9F,$B3,$39,$BF,$44,$E5,$B5,$CE,$B8,$BF,$7B,$DC,$F0
	.dc.b	$7B,$D6,$7A,$B8,$69,$F8,$68,$C5,$7D,$FF,$94,$FC,$3F,$91,$F8,$86
	.dc.b	$D7,$E3,$22,$CE,$E3,$07,$E4,$3C,$2F,$C7,$9F,$90,$F7,$6C,$64,$FC
	.dc.b	$AF,$CA,$62,$BF,$1C,$FF,$81,$BE,$73,$F5,$7E,$9F,$FA,$73,$E2,$3F
	.dc.b	$8A,$C6,$2B,$FD,$2D,$1F,$C6,$3D,$37,$91,$F8,$CF,$27,$FC,$7F,$5F
	.dc.b	$CF,$2F,$84,$62,$11,$8B,$EB,$CE,$C7,$C1,$5F,$D1,$31,$EB,$9E,$23
	.dc.b	$8D,$6D,$F1,$9D,$DC,$63,$DF,$1E,$D3,$CF,$2F,$36,$5E,$0C,$BA,$56
	.dc.b	$AB,$AC,$77,$CB,$C3,$1D,$8C,$96,$78,$1C,$63,$78,$78,$71,$F1,$E1
	.dc.b	$DE,$FF,$33,$7F,$2D,$FC,$FB,$5D,$1F,$A2,$F8,$C7,$C2,$3E,$77,$F4
	.dc.b	$32,$9C,$C6,$1B,$1F,$A6,$6B,$DB,$9F,$FD,$A0,$F8,$9A,$E7,$E0,$BA
	.dc.b	$E1,$FA,$A6,$CF,$0C,$FE,$63,$72,$FC,$2B,$6F,$F0,$1F,$1F,$FD,$72
	.dc.b	$B7,$7F,$3A,$B5,$5F,$82,$6E,$0B,$E1,$87,$4B,$E0,$4B,$FA,$13,$A7
	.dc.b	$F1,$D7,$C8,$93,$F7,$27,$53,$DE,$39,$61,$E4,$8E,$67,$E4,$3C,$07
	.dc.b	$CE,$BD,$FF,$EE,$B1,$FC,$A0,$CB,$8E,$3F,$F0,$CC,$62,$BF,$D2,$14
	.dc.b	$F8,$1D,$47,$FA,$BE,$30,$EA,$F3,$5C,$63,$FB,$EC,$E9,$BE,$F6,$BB
	.dc.b	$FF,$CD,$11,$FD,$3B,$C6,$E3,$5F,$E2,$F8,$F0,$D6,$08,$F9,$3E,$03
	.dc.b	$C6,$BE,$77,$83,$FC,$B1,$AF,$F3,$26,$BF,$E1,$66,$BF,$E3,$87,$8E
	.dc.b	$3F,$D3,$23,$A5,$76,$9C,$E9,$F1,$B0,$7C,$06,$71,$F6,$7C,$34,$C7
	.dc.b	$FD,$27,$82,$75,$1F,$A0,$8C,$36,$3F,$46,$FA,$DB,$1A,$8F,$EA,$0D
	.dc.b	$CB,$CC,$DC,$7E,$CF,$CC,$FF,$50,$9F,$57,$CE,$7F,$17,$6F,$2F,$C8
	.dc.b	$8C,$57,$F8,$E3,$E7,$5D,$FF,$31,$B9,$7E,$D9,$AF,$F5,$86,$BF,$D7
	.dc.b	$9A,$FF,$90,$A9,$F2,$7C,$EA,$4F,$FA,$DA,$72,$FF,$83,$BE,$5F,$F0
	.dc.b	$84,$7F,$D7,$7E,$1F,$B5,$FF,$36,$F7,$BE,$23,$E3,$B6,$3F,$85,$B5
	.dc.b	$FE,$DA,$D7,$FB,$9C,$E7,$5F,$BC,$DD,$3F,$CB,$2D,$94,$B6,$FE,$6D
	.dc.b	$1F,$E1,$BF,$98,$FC,$D1,$1F,$F7,$B4,$3E,$23,$F2,$BF,$3B,$FE,$3C
	.dc.b	$9F,$E3,$ED,$7F,$C0,$2D,$7F,$C0,$ED,$7E,$DF,$C2,$7E,$B2,$8D,$31
	.dc.b	$F0,$BF,$AB,$F9,$89,$5F,$E3,$87,$87,$C2,$F8,$BC,$3F,$C6,$E3,$FE
	.dc.b	$0C,$AF,$1C,$60,$FE,$67,$A7,$BF,$CA,$F0,$12,$7D,$BF,$0B,$E3,$F8
	.dc.b	$4F,$DD,$7C,$CC,$9F,$83,$1F,$B4,$F9,$D7,$BB,$F9,$C6,$CF,$F4,$06
	.dc.b	$BF,$E2,$26,$BF,$C5,$B8,$1F,$05,$F0,$49,$1F,$3B,$E1,$7E,$27,$C1
	.dc.b	$7F,$0B,$F7,$24,$FD,$28,$FE,$0B,$F2,$FF,$E3,$08,$34,$9D,$3F,$03
	.dc.b	$1F,$1F,$07,$F2,$56,$C7,$EB,$34,$FA,$5F,$05,$FC,$43,$D9,$7C,$BF
	.dc.b	$F1,$2F,$C0,$BF,$DC,$F9,$FF,$2A,$C3,$FD,$47,$18,$8F,$F2,$45,$70
	.dc.b	$74,$7F,$B2,$1C,$3F,$E4,$46,$BF,$E4,$A6,$BF,$EA,$E6,$BC,$6B,$C8
	.dc.b	$47,$A8,$8F,$E8,$4F,$83,$27,$CD,$BE,$17,$F9,$CB,$67,$4D,$7F,$D6
	.dc.b	$EF,$F2,$59,$FE,$94,$E5,$B3,$C8,$E7,$8F,$4F,$02,$5F,$47,$BF,$C0
	.dc.b	$62,$3A,$C5,$BE,$B3,$E4,$C6,$57,$91,$F3,$F4,$E2,$3C,$B1,$6B,$CB
	.dc.b	$37,$0F,$EB,$B9,$7C,$3B,$E4,$7B,$A7,$E0,$F8,$3F,$D5,$7D,$59,$3F
	.dc.b	$0A,$32,$D2,$1F,$13,$DD,$FE,$B9,$F2,$72,$3F,$D5,$2F,$CC,$F7,$B1
	.dc.b	$1F,$EA,$38,$B9,$FC,$A9,$FE,$B7,$F3,$AE,$EF,$C7,$79,$F3,$0F,$B7
	.dc.b	$39,$6C,$FF,$48,$8C,$47,$FA,$31,$AF,$FA,$49,$AF,$FB,$D9,$DD,$FF
	.dc.b	$5F,$6E,$D9,$AF,$ED,$BF,$3C,$F8,$DF,$C2,$FA,$53,$1F,$C2,$BF,$88
	.dc.b	$F5,$B1,$1F,$86,$8B,$97,$E1,$A7,$FB,$BE,$AF,$9A,$BD,$B3,$0F,$CB
	.dc.b	$CE,$5B,$3C,$08,$EB,$DA,$FF,$00,$FC,$F7,$FB,$32,$3F,$3F,$F6,$7F
	.dc.b	$C2,$1B,$E2,$6B,$AA,$47,$F9,$42,$B8,$3F,$EB,$26,$BF,$EB,$86,$A2
	.dc.b	$3C,$34,$FF,$03,$37,$87,$52,$8C,$3F,$D2,$AF,$FB,$7B,$FD,$33,$A5
	.dc.b	$E7,$FF,$88,$D9,$1F,$B0,$8F,$D9,$91,$C1,$FB,$4F,$47,$F6,$FE,$4F
	.dc.b	$F8,$C3,$C3,$FF,$30,$E7,$FE,$3D,$6E,$7F,$A0,$9D,$DF,$EF,$1A,$FE
	.dc.b	$F3,$CD,$FE,$03,$E3,$FF,$42,$CC,$D2,$4F,$FB,$63,$1F,$E1,$31,$C2
	.dc.b	$3F,$83,$BF,$C2,$BB,$BF,$85,$F1,$F9,$7D,$49,$5F,$C3,$23,$2D,$21
	.dc.b	$FC,$3B,$0F,$F2,$5F,$CB,$FF,$9F,$23,$FC,$B2,$D7,$FC,$AE,$D7,$FB
	.dc.b	$BB,$71,$3F,$0A,$33,$EF,$F4,$3A,$E1,$FE,$4A,$F8,$17,$CE,$7F,$80
	.dc.b	$2B,$FC,$33,$EA,$7D,$C3,$FB,$E3,$F8,$3A,$FE,$E6,$FE,$16,$32,$F0
	.dc.b	$3F,$11,$5D,$31,$E1,$3F,$86,$8C,$4B,$E7,$F9,$BF,$DF,$BD,$6F,$C5
	.dc.b	$C6,$2B,$AB,$AA,$E2,$DC,$7D,$0E,$EB,$8D,$7A,$5B,$7E,$4A,$DC,$F8
	.dc.b	$DB,$1E,$B9,$DB,$E3,$3F,$45,$5D,$F0,$B7,$C5,$F6,$CF,$12,$9C,$4F
	.dc.b	$88,$7F,$AF,$EF,$E7,$63,$CE,$3E,$15,$79,$DE,$A7,$33,$27,$A6,$B2
	.dc.b	$F3,$6C,$7C,$63,$94,$70,$34,$5C,$7E,$8D,$7C,$78,$F3,$CF,$DA,$8F
	.dc.b	$21,$E5,$C3,$53,$CE,$7B,$3E,$33,$A6,$1F,$29,$F8,$13,$C0,$E5,$3F
	.dc.b	$1E,$3E,$1B,$DC,$F3,$D7,$81,$D4,$F1,$7D,$6F,$29,$F0,$39,$7E,$33
	.dc.b	$97,$DE,$E5,$F1,$C7,$43,$C5,$3E,$C2,$38,$35,$1D,$1F,$09,$87,$E5
	.dc.b	$C7,$07,$EB,$3E,$97,$E2,$39,$7F,$04,$E5,$FA,$79,$7F,$A0,$9B,$1F
	.dc.b	$16,$31,$87,$E9,$9D,$4F,$F3,$3F,$82,$FE,$69,$18,$84,$62,$BF,$99
	.dc.b	$C6,$C4,$EE,$75,$1A,$7F,$2F,$93,$E7,$9C,$9F,$9A,$AE,$58,$7E,$DA
	.dc.b	$B5,$39,$9F,$09,$F9,$E8,$CB,$FA,$03,$8E,$5F,$EF,$2B,$EB,$19,$38
	.dc.b	$E7,$D6,$32,$FF,$04,$8F,$53,$9B,$FD,$0C,$74,$3F,$DF,$33,$D3,$0E
	.dc.b	$11,$E4,$78,$3D,$F1,$FD,$4E,$3E,$6C,$37,$18,$BF,$0E,$3F,$2A,$FC
	.dc.b	$AF,$A7,$1F,$0C,$60,$F0,$B8,$AE,$4D,$8F,$D9,$61,$DE,$BA,$57,$D0
	.dc.b	$BF,$44,$62,$BE,$54,$61,$FA,$C7,$27,$CA,$E3,$0F,$69,$EA,$75,$87
	.dc.b	$86,$33,$5D,$71,$F8,$66,$CA,$F8,$63,$75,$D6,$3E,$03,$FC,$2C,$FA
	.dc.b	$6E,$23,$C0,$70,$8F,$F0,$1E,$1E,$DB,$FC,$33,$63,$E9,$1E,$63,$A1
	.dc.b	$F2,$CF,$F6,$8E,$1E,$E7,$99,$BA,$EF,$7E,$13,$D0,$58,$F4,$1F,$50
	.dc.b	$EB,$5E,$1C,$3E,$04,$FB,$1F,$EB,$78,$FF,$68,$F7,$7F,$2A,$CD,$D1
	.dc.b	$8C,$FF,$B8,$FC,$FD,$CB,$FB,$8C,$DB,$E5,$5F,$E0,$7B,$E2,$D9,$FD
	.dc.b	$0D,$7F,$7C,$6B,$FD,$F3,$C0,$E5,$5F,$EB,$8A,$FF,$4C,$3F,$EB,$06
	.dc.b	$C6,$8F,$07,$43,$1C,$4F,$FA,$7E,$30,$F2,$F0,$3A,$9D,$B7,$B1,$93
	.dc.b	$D1,$1E,$4C,$7F,$C1,$AD,$CF,$FC,$1A,$DC,$9F,$0F,$E8,$3D,$3C,$77
	.dc.b	$0F,$4C,$79,$B5,$C6,$2C,$7F,$C7,$11,$F0,$3E,$2E,$5F,$C0,$F8,$CB
	.dc.b	$7B,$8F,$F1,$0C,$FC,$0F,$4B,$38,$7F,$90,$5F,$A9,$5F,$C8,$3D,$0B
	.dc.b	$CD,$FE,$C3,$C3,$FD,$85,$F3,$EF,$57,$35,$27,$F8,$05,$F1,$6D,$F1
	.dc.b	$8F,$E0,$8D,$7F,$97,$B9,$C7,$EE,$FC,$F3,$C9,$E9,$5E,$93,$FC,$D9
	.dc.b	$19,$BF,$F6,$BF,$CF,$D9,$D0,$71,$7F,$88,$5E,$8F,$B9,$CD,$7F,$9C
	.dc.b	$9B,$5F,$E7,$9F,$D8,$5F,$39,$BC,$BE,$2C,$62,$FF,$1B,$F3,$FF,$90
	.dc.b	$3F,$23,$F2,$F6,$32,$4C,$3D,$59,$FF,$93,$A7,$51,$E8,$D5,$E2,$7D
	.dc.b	$03,$E9,$EC,$F4,$31,$8F,$FB,$DA,$B6,$FC,$EA,$FC,$85,$FF,$58,$6C
	.dc.b	$EA,$3E,$8C,$7D,$EF,$9F,$73,$74,$9F,$93,$18,$BF,$D1,$77,$FE,$98
	.dc.b	$FD,$6F,$D0,$58,$C9,$2B,$DB,$D8,$9D,$46,$3E,$1F,$F6,$A7,$ED,$BE
	.dc.b	$42,$DF,$ED,$A3,$F6,$7E,$06,$2B,$E3,$F8,$3F,$D9,$38,$BF,$86,$7C
	.dc.b	$6F,$EF,$8F,$DF,$FC,$7D,$C7,$F7,$F1,$6F,$C7,$36,$B8,$D6,$8F,$E0
	.dc.b	$B1,$97,$FD,$71,$0F,$9A,$7C,$AF,$F1,$56,$CE,$8D,$BF,$EB,$F7,$17
	.dc.b	$FB,$58,$DB,$F8,$06,$D7,$F2,$66,$CF,$29,$ED,$5D,$FE,$56,$4F,$4B
	.dc.b	$F2,$79,$B8,$BE,$95,$67,$A4,$DA,$FF,$66,$FD,$07,$E5,$EF,$F4,$AD
	.dc.b	$AF,$E5,$7C,$6E,$AF,$E0,$33,$9B,$E3,$DB,$F8,$C8,$E5,$87,$F8,$99
	.dc.b	$AE,$A7,$F8,$0C,$71,$8F,$F7,$12,$F7,$7A,$D7,$67,$E6,$7C,$74,$BF
	.dc.b	$76,$DC,$73,$FA,$53,$FC,$53,$CE,$7C,$0F,$E7,$9B,$7F,$C4,$1B,$9F
	.dc.b	$F8,$1A,$7F,$98,$60,$FA,$0F,$E7,$FE,$6F,$DB,$6E,$1E,$0C,$FE,$F1
	.dc.b	$F7,$FE,$82,$C7,$DC,$C7,$FC,$8F,$F0,$9E,$12,$47,$85,$EF,$F1,$BF
	.dc.b	$D3,$3D,$0F,$C4,$36,$7D,$BF,$19,$F8,$B3,$F1,$9F,$41,$63,$B7,$D0
	.dc.b	$3E,$2B,$F7,$6D,$8F,$E5,$AD,$CB,$FC,$5C,$D4,$7E,$6B,$B7,$FA,$6C
	.dc.b	$FF,$5B,$F5,$3F,$35,$F4,$1F,$D8,$1F,$CE,$7A,$32,$7E,$8F,$CF,$5C
	.dc.b	$3F,$95,$E2,$F8,$CF,$19,$2B,$C6,$FF,$5D,$CB,$F8,$45,$7E,$E3,$C9
	.dc.b	$FE,$F9,$F3,$5F,$E6,$4B,$FE,$D9,$1F,$DF,$FE,$7F,$F9,$7F,$61,$B4
	.dc.b	$F8,$8D,$7A,$7F,$D5,$3E,$81,$E9,$FF,$1E,$6E,$5F,$D4,$1B,$9F,$F9
	.dc.b	$45,$AF,$C4,$3F,$B4,$68,$F6,$BF,$C0,$27,$8F,$D3,$B7,$FA,$FD,$1F
	.dc.b	$07,$FD,$93,$E8,$1C,$7E,$98,$C3,$2B,$FB,$0C,$E9,$B1,$FC,$D2,$31
	.dc.b	$97,$FA,$3B,$81,$F4,$9F,$C9,$EF,$F7,$CD,$9F,$EB,$75,$C2,$BF,$C5
	.dc.b	$B3,$97,$FC,$47,$C1,$F4,$1B,$C5,$F4,$38,$FF,$BD,$CE,$71,$F8,$4F
	.dc.b	$D6,$C7,$FD,$3E,$D7,$35,$FD,$2A,$31,$1F,$8C,$79,$3F,$2B,$87,$AC
	.dc.b	$FF,$8A,$9E,$FB,$CB,$FE,$31,$C7,$FD,$F5,$F0,$C6,$6F,$FD,$0D,$87
	.dc.b	$D3,$9F,$F0,$4B,$63,$33,$FE,$29,$AF,$DE,$E5,$FE,$4E,$6D,$FF,$38
	.dc.b	$5F,$43,$F9,$0F,$12,$4F,$F8,$EF,$C9,$B8,$FF,$4C,$7F,$2B,$E0,$7E
	.dc.b	$5B,$E7,$2D,$7A,$1C,$34,$F5,$C6,$2B,$FC,$1F,$F4,$1F,$B7,$F3,$FF
	.dc.b	$D9,$DF,$03,$F8,$F8,$C4,$FF,$B3,$A3,$F2,$5E,$83,$E6,$BF,$D7,$7E
	.dc.b	$67,$F9,$9A,$BF,$73,$7C,$23,$10,$8C,$5F,$5F,$0F,$FA,$AC,$F5,$3F
	.dc.b	$80,$E1,$D7,$3F,$1C,$74,$9D,$E1,$D3,$0E,$73,$9C,$B6,$B0,$E4,$F5
	.dc.b	$3C,$57,$CB,$1F,$15,$70,$BE,$75,$C7,$2F,$6D,$6A,$39,$4E,$8D,$E1
	.dc.b	$DB,$1F,$26,$3D,$6B,$C8,$7F,$B2,$EB,$E5,$BF,$EF,$66,$BD,$4F,$F7
	.dc.b	$CF,$C3,$FF,$9B,$3E,$87,$FC,$A6,$D8,$FF,$6A,$35,$DF,$88,$9F,$F2
	.dc.b	$CD,$7F,$1F,$F7,$6F,$51,$A4,$BF,$24,$65,$F6,$B7,$2F,$9C,$DD,$3F
	.dc.b	$09,$F1,$5F,$EB,$E7,$F0,$D1,$9A,$FC,$0B,$5F,$C0,$37,$05,$71,$9D
	.dc.b	$E1,$F4,$6B,$46,$E7,$53,$CA,$74,$47,$E9,$C6,$AB,$E0,$47,$AA,$75
	.dc.b	$87,$E1,$EB,$C1,$CE,$FA,$17,$B7,$FC,$B6,$3F,$D8,$3E,$27,$F9,$07
	.dc.b	$1F,$F5,$57,$B9,$9E,$5F,$F1,$6E,$1C,$27,$FC,$38,$D7,$F9,$23,$5E
	.dc.b	$8D,$78,$E1,$FE,$8E,$C1,$D7,$FE,$87,$9F,$E1,$FE,$3F,$77,$FD,$4D
	.dc.b	$19,$7F,$CE,$36,$F6,$75,$7E,$B6,$D6,$E3,$F8,$75,$EA,$B8,$5F,$FB
	.dc.b	$17,$E4,$FC,$5A,$1A,$48,$FE,$A0,$7F,$73,$7B,$C6,$E3,$91,$FE,$0D
	.dc.b	$FA,$1E,$AF,$FA,$1B,$1F,$F4,$03,$FE,$D3,$CF,$89,$E6,$F2,$64,$76
	.dc.b	$3D,$EF,$C5,$9E,$31,$FD,$3D,$AF,$D7,$73,$9F,$F2,$B9,$E7,$3F,$EA
	.dc.b	$AB,$7C,$C7,$A3,$DF,$93,$FE,$E5,$7E,$72,$FA,$0F,$E2,$F8,$7F,$BD
	.dc.b	$9F,$E3,$31,$F0,$9A,$FF,$73,$6B,$FE,$18,$6A,$FA,$17,$83,$FE,$CC
	.dc.b	$8C,$36,$33,$1F,$ED,$06,$BF,$DB,$75,$F9,$77,$F9,$4E,$1F,$A2,$7D
	.dc.b	$EF,$93,$F4,$F8,$5E,$EB,$FF,$6C,$B5,$F9,$8D,$7E,$DB,$5F,$F0,$62
	.dc.b	$CF,$E1,$F7,$D9,$FF,$24,$35,$FF,$6A,$CD,$C3,$F4,$FA,$FE,$10,$93
	.dc.b	$FD,$A3,$E5,$3F,$0F,$F0,$5F,$8E,$24,$7D,$FA,$68,$9F,$D2,$7C,$EB
	.dc.b	$E4,$7F,$3B,$FC,$01,$FE,$0F,$EA,$FF,$4E,$B7,$16,$2B,$E1,$1F,$E8
	.dc.b	$F9,$FE,$9A,$FF,$B6,$1B,$2A,$E7,$E3,$FE,$A3,$E5,$7B,$FC,$37,$ED
	.dc.b	$6C,$D2,$61,$E9,$C7,$FB,$13,$DF,$C8,$F3,$9F,$B4,$8C,$C7,$E8,$FE
	.dc.b	$5B,$D7,$F0,$B2,$3F,$C1,$B1,$EE,$F8,$5F,$01,$23,$E3,$62,$8F,$8B
	.dc.b	$F2,$FF,$13,$79,$5F,$EA,$DC,$53,$FC,$0B,$D8,$F9,$3F,$45,$2B,$FC
	.dc.b	$37,$1F,$33,$E1,$7D,$C4,$8E,$EF,$D2,$E4,$FA,$3F,$09,$27,$B4,$7E
	.dc.b	$A3,$E7,$FF,$DF,$F1,$FE,$8D,$5D,$99,$35,$F1,$D2,$7F,$7C,$27,$8B
	.dc.b	$FC,$05,$AF,$F1,$06,$BF,$E6,$17,$7C,$FF,$66,$3F,$D0,$F1,$8E,$7D
	.dc.b	$57,$EF,$3E,$4D,$F2,$BF,$C9,$BE,$65,$F3,$FF,$8C,$F8,$B9,$97,$FB
	.dc.b	$56,$72,$D9,$CB,$F8,$99,$E5,$1C,$FC,$1F,$CA,$FB,$12,$B9,$46,$5F
	.dc.b	$C9,$46,$5F,$CB,$E7,$FE,$F6,$C3,$FD,$4D,$1E,$1C,$BF,$3F,$F3,$B2
	.dc.b	$30,$43,$F3,$4E,$5E,$67,$E8,$A5,$3E,$F3,$5B,$7C,$71,$D2,$BD,$5F
	.dc.b	$F9,$7B,$D1,$FE,$92,$47,$E2,$78,$95,$FE,$79,$16,$FA,$62,$AF,$EA
	.dc.b	$3C,$1F,$B8,$F8,$B9,$87,$F9,$AA,$32,$FE,$D6,$32,$FF,$47,$8F,$B1
	.dc.b	$E2,$FF,$34,$F9,$E7,$C5,$D3,$35,$FE,$DB,$DB,$F8,$D3,$FE,$C8,$6B
	.dc.b	$F0,$AB,$F9,$84,$60,$FE,$49,$9F,$E2,$67,$47,$39,$FE,$87,$F3,$FF
	.dc.b	$EF,$F8,$FC,$27,$9F,$FC,$4B,$09,$18,$21,$FC,$35,$CB,$F5,$2D,$AF
	.dc.b	$F5,$6E,$1F,$F5,$D6,$8F,$95,$FE,$A2,$DB,$FF,$7B,$35,$FB,$A1,$7C
	.dc.b	$87,$09,$93,$C1,$E7,$1F,$D7,$3E,$7F,$F8,$1C,$FF,$23,$F9,$0F,$E5
	.dc.b	$66,$7C,$29,$1F,$B4,$9D,$BF,$DC,$DA,$FF,$80,$9A,$FF,$B4,$DD,$E7
	.dc.b	$F2,$DB,$96,$6B,$FB,$6F,$CF,$FF,$B8,$EF,$FD,$57,$A7,$F8,$58,$8F
	.dc.b	$EC,$7C,$EA,$FF,$82,$62,$E3,$F8,$E9,$FF,$08,$FB,$92,$87,$CF,$F0
	.dc.b	$7F,$CB,$4F,$D1,$B7,$93,$B8,$D5,$5C,$3F,$33,$3F,$D3,$B2,$FE,$80
	.dc.b	$4A,$FE,$62,$D7,$FA,$53,$B7,$F8,$47,$8B,$FE,$24,$F9,$E9,$1E,$4B
	.dc.b	$FF,$25,$C7,$F1,$1E,$29,$39,$DB,$5A,$2B,$8C,$FA,$6B,$FB,$0F,$C9
	.dc.b	$7F,$53,$24,$7F,$44,$6B,$FD,$75,$DB,$D4,$FE,$CB,$F3,$1F,$13,$1F
	.dc.b	$D0,$7C,$2B,$E5,$3F,$B4,$D7,$87,$CD,$F9,$6F,$F7,$1F,$46,$E7,$FD
	.dc.b	$B2,$30,$7D,$07,$17,$CF,$0C,$5F,$28,$C5,$3E,$7F,$94,$FF,$6C,$8F
	.dc.b	$B3,$18,$C3,$C2,$F4,$C7,$CA,$DC,$37,$7E,$63,$95,$F5,$71,$7C,$9F
	.dc.b	$15,$F2,$6D,$72,$AF,$C4,$C7,$2B,$F1,$46,$46,$1E,$A8,$F1,$47,$C7
	.dc.b	$38,$E1,$F1,$AB,$C3,$1F,$08,$FC,$AD,$7C,$2C,$78,$DF,$69,$F4,$1C
	.dc.b	$8F,$04,$F7,$AF,$3C,$68,$FA,$25,$7D,$06,$A7,$CB,$5D,$49,$1C,$9F
	.dc.b	$0B,$E6,$9E,$57,$F1,$DF,$3E,$3F,$1D,$B8,$FC,$3A,$EF,$1E,$73,$C9
	.dc.b	$8F,$9D,$D9,$F1,$4F,$3E,$1E,$27,$2F,$89,$E5,$8F,$C3,$30,$FD,$0C
	.dc.b	$3C,$4F,$C0,$8D,$5F,$BC,$91,$C9,$F6,$47,$2B,$FA,$11,$EF,$BF,$15
	.dc.b	$75,$C3,$EB,$34,$F8,$67,$BF,$1F,$88,$E7,$0F,$9B,$7C,$C9,$1E,$C7
	.dc.b	$E7,$18,$7E,$0B,$87,$E2,$F1,$7E,$FD,$B8,$7D,$E8,$C4,$78,$DA,$FF
	.dc.b	$1E,$F8,$2F,$E3,$F1,$88,$46,$27,$FD,$54,$D8,$FF,$5C,$B5,$FD,$EB
	.dc.b	$F0,$31,$7C,$CF,$F8,$AE,$BF,$7D,$5D,$CF,$B5,$3A,$E1,$F8,$F1,$99
	.dc.b	$FE,$09,$39,$8F,$C0,$19,$C7,$FD,$D9,$7D,$5F,$2C,$7E,$6E,$78,$65
	.dc.b	$E0,$8F,$59,$9A,$FF,$1F,$46,$71,$F4,$47,$39,$F6,$9E,$67,$0F,$B8
	.dc.b	$E5,$8F,$EE,$DF,$65,$7A,$27,$AE,$1E,$48,$C5,$6E,$BC,$71,$87,$F6
	.dc.b	$47,$29,$F6,$CE,$EF,$F8,$46,$1E,$C9,$F7,$E3,$C4,$F5,$5F,$9E,$FE
	.dc.b	$5B,$DA,$79,$BD,$8C,$1F,$D2,$F1,$F3,$BF,$03,$1F,$21,$D0,$B3,$E0
	.dc.b	$39,$DF,$80,$F1,$3E,$F3,$BC,$71,$69,$99,$F0,$5F,$B5,$F9,$67,$97
	.dc.b	$1F,$A7,$1E,$A6,$C7,$E0,$8F,$49,$FE,$B6,$C3,$D2,$FD,$5B,$F3,$BE
	.dc.b	$BB,$F0,$B4,$F0,$9D,$CF,$0C,$79,$DC,$BE,$D8,$F7,$9F,$06,$38,$E1
	.dc.b	$FE,$C4,$C7,$FC,$C3,$F4,$1F,$E6,$3C,$BF,$D9,$6D,$AF,$EC,$FF,$0F
	.dc.b	$6B,$75,$FE,$CB,$AF,$F2,$45,$70,$8F,$F5,$AC,$62,$3F,$DA,$7E,$42
	.dc.b	$31,$4E,$1F,$F2,$05,$FF,$75,$8F,$F7,$0F,$D0,$77,$9C,$E1,$FE,$1B
	.dc.b	$8F,$17,$29,$A6,$FB,$EE,$1F,$ED,$FF,$94,$F4,$66,$7A,$78,$18,$64
	.dc.b	$FF,$B3,$1F,$2C,$EE,$30,$7F,$BD,$70,$6B,$E2,$3C,$DF,$17,$67,$D5
	.dc.b	$3A,$6C,$E9,$A6,$BD,$0F,$59,$EF,$F8,$FB,$19,$26,$BF,$E6,$9D,$93
	.dc.b	$A8,$FF,$31,$FC,$E3,$E2,$7E,$C4,$23,$17,$FE,$21,$87,$47,$FA,$92
	.dc.b	$31,$3A,$9E,$75,$F6,$5E,$4D,$8D,$47,$F8,$90,$CC,$FF,$B6,$A3,$97
	.dc.b	$17,$FA,$B7,$67,$CC,$F9,$27,$35,$FE,$43,$9C,$BE,$5F,$9D,$7B,$BE
	.dc.b	$A6,$C7,$A7,$E7,$B8,$E5,$FE,$5E,$9C,$BF,$E8,$46,$E5,$FE,$9B,$6D
	.dc.b	$7F,$8E,$AB,$8D,$27,$2F,$F9,$13,$E4,$7F,$0F,$C5,$F9,$16,$DF,$E5
	.dc.b	$78,$11,$8A,$FA,$8D,$C7,$FD,$55,$F3,$CF,$97,$FA,$6D,$C3,$E9,$70
	.dc.b	$FF,$A6,$5F,$A9,$19,$C3,$FD,$3A,$D7,$E7,$35,$FF,$29,$B5,$FF,$3F
	.dc.b	$B5,$3C,$EF,$FD,$79,$BB,$DC,$F1,$E3,$F4,$DF,$D5,$F4,$FE,$F7,$F8
	.dc.b	$05,$96,$0F,$DA,$38,$7F,$3B,$1B,$75,$97,$FA,$F6,$74,$FF,$A4,$7C
	.dc.b	$0F,$DB,$9F,$BB,$F8,$7B,$5F,$A6,$8B,$3F,$9D,$9B,$9F,$FB,$D7,$77
	.dc.b	$C3,$F9,$7D,$DF,$E0,$8D,$C7,$EA,$37,$0F,$D3,$36,$F9,$36,$7F,$D2
	.dc.b	$57,$D1,$F6,$DF,$F1,$DF,$9E,$7C,$47,$C3,$F9,$8C,$36,$BF,$8B,$4E
	.dc.b	$9F,$8B,$3A,$69,$AF,$88,$F9,$9F,$03,$D5,$B1,$93,$E6,$C6,$5A,$7E
	.dc.b	$DF,$E7,$DF,$07,$FB,$C6,$C6,$7B,$31,$32,$3F,$69,$BF,$FB,$0A,$B6
	.dc.b	$E1,$AF,$EE,$5A,$FF,$A7,$9D,$EF,$F8,$28,$E5,$E9,$75,$7F,$DE,$32
	.dc.b	$8F,$91,$F1,$3F,$89,$F9,$6B,$87,$E2,$1F,$C6,$7B,$FF,$D0,$39,$BF
	.dc.b	$88,$63,$FC,$27,$24,$62,$3F,$D5,$F1,$89,$FE,$8F,$DD,$FE,$FE,$C3
	.dc.b	$F2,$7F,$33,$FD,$0B,$CA,$FC,$C3,$F6,$76,$FC,$D7,$CC,$DA,$FC,$CB
	.dc.b	$F9,$DF,$84,$F1,$12,$3E,$3E,$5E,$28,$EF,$EA,$FE,$8F,$E8,$3F,$A7
	.dc.b	$B6,$BF,$03,$EF,$FE,$9F,$E4,$7F,$10,$D8,$FD,$3F,$97,$EF,$23,$DB
	.dc.b	$87,$F9,$4F,$0F,$FB,$8D,$7C,$17,$EC,$4F,$EB,$BE,$85,$AF,$C7,$F9
	.dc.b	$BF,$B5,$F1,$7F,$2C,$D8,$FD,$AF,$BF,$F3,$48,$F7,$7E,$87,$C7,$FE
	.dc.b	$F4,$FE,$D3,$E4,$59,$FE,$E1,$F3,$DF,$C9,$3E,$7D,$F3,$BF,$DF,$E3
	.dc.b	$0D,$D3,$FB,$CC,$7A,$2B,$E4,$B7,$6F,$B8,$D7,$ED,$EC,$F8,$2F,$E1
	.dc.b	$C7,$F7,$AF,$52,$DF,$F5,$5F,$9F,$FE,$E1,$7F,$9A,$6B,$FD,$3D,$AF
	.dc.b	$FB,$D1,$AB,$A1,$F3,$EF,$9D,$FF,$07,$B6,$FE,$FB,$6B,$F2,$4D,$7C
	.dc.b	$FF,$0B,$FC,$8C,$FF,$10,$46,$EF,$F4,$87,$E6,$23,$FC,$27,$B4,$BF
	.dc.b	$ED,$DE,$4F,$F3,$15,$87,$97,$C1,$F9,$B1,$FE,$06,$AF,$CE,$91,$FF
	.dc.b	$07,$36,$7F,$96,$7C,$77,$56,$CF,$F3,$4F,$86,$FE,$70,$7F,$80,$30
	.dc.b	$FF,$02,$C6,$49,$3F,$AF,$E0,$FE,$A0,$DA,$FD,$A1,$FE,$34,$F9,$A7
	.dc.b	$83,$FE,$2E,$F6,$3F,$C8,$11,$8C,$3F,$C3,$B1,$96,$9F,$E3,$98,$D5
	.dc.b	$FD,$38,$F8,$3B,$35,$FE,$D3,$1B,$EF,$F0,$11,$FD,$9D,$A3,$17,$FD
	.dc.b	$37,$E2,$3C,$71,$FD,$5D,$FF,$0F,$FC,$F3,$CE,$FE,$73,$5F,$EC,$2F
	.dc.b	$7F,$E8,$A9,$D7,$A5,$FE,$C9,$9C,$FA,$2F,$92,$FE,$41,$F0,$FF,$E3
	.dc.b	$EC,$3E,$CE,$1C,$1E,$86,$31,$EF,$3B,$8E,$D1,$8C,$3A,$DF,$47,$C7
	.dc.b	$7C,$EF,$83,$63,$35,$C6,$FF,$13,$87,$97,$5F,$44,$78,$72,$AF,$81
	.dc.b	$B8,$73,$8E,$39,$78,$5E,$07,$A8,$E0,$58,$E0,$AF,$C0,$E7,$0F,$23
	.dc.b	$C0,$FF,$21,$4F,$C5,$6C,$7D,$26,$DF,$F5,$45,$F1,$2F,$19,$F3,$F9
	.dc.b	$8C,$36,$33,$3F,$D7,$27,$2F,$F5,$EC,$1E,$8F,$F9,$72,$30,$D7,$CA
	.dc.b	$D7,$E8,$B5,$FB,$AF,$1D,$BE,$9B,$76,$DF,$BF,$F3,$3A,$BF,$17,$5A
	.dc.b	$8F,$9F,$5B,$BE,$71,$B7,$2C,$7E,$74,$6D,$FC,$A5,$74,$BE,$F1,$AA
	.dc.b	$E1,$7F,$86,$AD,$4F,$63,$44,$8D,$3F,$86,$E5,$7E,$0B,$F0,$DF,$E3
	.dc.b	$2D,$E2,$FD,$AB,$67,$F9,$52,$F8,$8F,$EE,$73,$A6,$CE,$B1,$B7,$D9
	.dc.b	$B7,$C6,$7F,$D7,$8F,$FA,$DD,$B9,$68,$DF,$C3,$C8,$DB,$5F,$E1,$47
	.dc.b	$93,$E1,$BA,$FC,$0F,$E2,$F1,$F6,$B7,$03,$49,$5E,$27,$8B,$E4,$F8
	.dc.b	$1F,$4F,$D0,$7F,$93,$A3,$FD,$4D,$F3,$CF,$9C,$F7,$C2,$30,$FF,$B5
	.dc.b	$E7,$83,$E6,$C5,$18,$8F,$F0,$DB,$5F,$E4,$6D,$7F,$AC,$36,$38,$FC
	.dc.b	$1F,$90,$93,$F2,$BD,$BF,$1F,$DD,$7F,$BA,$E3,$FD,$78,$D7,$E3,$87
	.dc.b	$15,$FE,$3B,$87,$CD,$7F,$AB,$9F,$F6,$23,$5F,$F2,$4B,$5F,$F5,$F3
	.dc.b	$5F,$E5,$6D,$7F,$0F,$EC,$FC,$68,$E5,$3D,$DB,$C9,$F8,$E9,$F9,$7F
	.dc.b	$3D,$FC,$F6,$FF,$82,$35,$FF,$2C,$B5,$FF,$67,$35,$FE,$77,$F3,$FF
	.dc.b	$8D,$7F,$CB,$ED,$7F,$DA,$8D,$77,$CD,$F8,$7F,$A3,$FE,$39,$1F,$D5
	.dc.b	$30,$FE,$AB,$18,$8F,$B4,$DC,$BF,$BF,$AF,$82,$FC,$EF,$74,$AF,$CE
	.dc.b	$BF,$A0,$F5,$FF,$43,$E9,$FE,$4C,$93,$FE,$4E,$E7,$FF,$7C,$B7,$CF
	.dc.b	$FD,$39,$1F,$D3,$F9,$7F,$AA,$F4,$A4,$7E,$A9,$FD,$6F,$D1,$7E,$C3
	.dc.b	$E8,$7E,$96,$C6,$93,$E1,$FF,$49,$F4,$5F,$B4,$26,$1F,$B4,$F8,$CF
	.dc.b	$DB,$F6,$FE,$E3,$E7,$24,$FE,$E2,$3F,$77,$D9,$FA,$D2,$47,$AD,$FE
	.dc.b	$79,$F6,$3F,$E6,$AE,$6F,$F0,$FB,$64,$D3,$FC,$47,$E6,$7F,$0A,$7F
	.dc.b	$C4,$BF,$31,$67,$E8,$57,$26,$9C,$84,$7C,$1F,$9C,$FE,$05,$B1,$A4
	.dc.b	$9F,$F5,$1D,$FF,$8C,$B6,$EC,$48,$FC,$11,$D7,$A3,$C5,$F3,$1F,$C0
	.dc.b	$3E,$85,$F1,$34,$CD,$70,$C3,$FC,$7B,$5F,$0A,$0D,$23,$E2,$32,$71
	.dc.b	$29,$FC,$7A,$39,$B9,$FA,$3F,$89,$F4,$72,$38,$3F,$1F,$C3,$C4,$71
	.dc.b	$11,$8E,$6F,$F5,$D5,$FF,$94,$AB,$E7,$7C,$6C,$BF,$13,$F4,$7D,$7C
	.dc.b	$47,$38,$B7,$CA,$3F,$C4,$C7,$A7,$E8,$FE,$C7,$D1,$C8,$C1,$0F,$23
	.dc.b	$97,$CD,$1F,$7C,$C4,$F5,$9C,$4F,$C5,$E1,$FC,$21,$1F,$91,$F1,$DF
	.dc.b	$87,$30,$FC,$18,$C3,$F3,$7E,$4B,$F1,$84,$7E,$87,$CF,$D7,$E4,$45
	.dc.b	$C3,$DD,$3F,$D0,$3E,$4D,$DD,$7D,$29,$87,$E2,$04,$E9,$C3,$5F,$F7
	.dc.b	$9B,$5F,$F2,$99,$C2,$7F,$C4,$8D,$3F,$C4,$D1,$C6,$FF,$9C,$37,$AB
	.dc.b	$FC,$B6,$D9,$FF,$46,$FD,$1F,$FB,$CD,$FC,$DD,$3D,$F7,$C3,$7F,$76
	.dc.b	$6C,$E9,$AF,$E2,$7E,$7F,$53,$23,$4F,$ED,$63,$97,$C2,$C9,$E4,$FE
	.dc.b	$E7,$E7,$E4,$60,$87,$E5,$5F,$ED,$78,$B8,$FF,$B6,$FD,$1B,$96,$9A
	.dc.b	$89,$7F,$E4,$9A,$E0,$D8,$F9,$31,$88,$F9,$A1,$C6,$9F,$D8,$5B,$B7
	.dc.b	$FA,$11,$B3,$FE,$C9,$F1,$3F,$D2,$CF,$DD,$F8,$87,$91,$FE,$02,$6C
	.dc.b	$66,$3F,$8C,$7C,$9C,$8C,$10,$FD,$A3,$F4,$7E,$52,$47,$D1,$7F,$90
	.dc.b	$7C,$54,$8C,$10,$FD,$E3,$FC,$9F,$E1,$BF,$BD,$47,$FA,$73,$BF,$FC
	.dc.b	$D3,$A7,$F3,$29,$CC,$FF,$32,$6B,$FC,$C9,$FF,$01,$6B,$FC,$11,$B3
	.dc.b	$B3,$B7,$CD,$77,$E5,$F0,$7C,$44,$C3,$F9,$05,$FE,$23,$0E,$55,$B9
	.dc.b	$CB,$E7,$F8,$6F,$47,$4F,$94,$93,$FC,$D2,$FF,$52,$D8,$DB,$D9,$F8
	.dc.b	$1F,$31,$FE,$32,$F8,$97,$BB,$FE,$FF,$AF,$F2,$B7,$91,$FE,$BE,$3F
	.dc.b	$A2,$BF,$EF,$0F,$6F,$F9,$6C,$62,$36,$7F,$BD,$2B,$9E,$9B,$9F,$F6
	.dc.b	$3C,$62,$5F,$43,$CB,$D0,$FF,$65,$B6,$3F,$DF,$AF,$F3,$A9,$E0,$DC
	.dc.b	$7C,$CE,$E7,$9D,$F9,$EF,$9B,$E0,$AD,$B7,$0E,$AE,$EA,$9F,$D4,$2B
	.dc.b	$8B,$EB,$AE,$2D,$AF,$63,$B2,$9F,$D5,$A7,$FD,$CB,$8F,$9E,$73,$8F
	.dc.b	$AB,$0F,$05,$FA,$0E,$45,$3F,$D7,$78,$78,$1B,$5E,$67,$B1,$67,$E9
	.dc.b	$63,$DE,$39,$5F,$CF,$C3,$E3,$12,$BE,$14,$FD,$83,$A9,$E1,$79,$9E
	.dc.b	$6A,$E8,$47,$04,$8F,$88,$78,$30,$FC,$55,$77,$8F,$6B,$E5,$39,$3F
	.dc.b	$02,$BE,$D1,$87,$C6,$E2,$BB,$46,$1F,$4C,$66,$39,$46,$0F,$6C,$F8
	.dc.b	$63,$E3,$9E,$FB,$F8,$93,$B8,$E2,$43,$B1,$EC,$9E,$51,$F9,$F7,$D4
	.dc.b	$F6,$3E,$DB,$DE,$7A,$CF,$C4,$F8,$7F,$F3,$64,$7E,$85,$B8,$7E,$BD
	.dc.b	$B8,$A9,$1F,$E6,$FF,$88,$C3,$63,$38,$7F,$34,$9C,$B6,$32,$7F,$64
	.dc.b	$8F,$CA,$D3,$E1,$5C,$9F,$98,$AF,$05,$6A,$73,$1F,$85,$32,$7E,$1E
	.dc.b	$B9,$57,$18,$57,$18,$C3,$F9,$79,$DC,$FE,$74,$E7,$A7,$F8,$DA,$32
	.dc.b	$FF,$8E,$23,$BE,$19,$C3,$C6,$E6,$3B,$9E,$38,$CE,$9F,$06,$B8,$63
	.dc.b	$FA,$7A,$EF,$3F,$35,$61,$E0,$8E,$06,$5C,$47,$C9,$3F,$68,$E1,$E5
	.dc.b	$8F,$B7,$0F,$8F,$87,$96,$31,$87,$86,$BC,$F8,$F9,$1F,$98,$63,$1F
	.dc.b	$06,$1E,$A8,$F0,$61,$B3,$A3,$E2,$8E,$F3,$E8,$8F,$C4,$63,$E5,$7D
	.dc.b	$E6,$27,$6F,$AE,$FC,$47,$42,$D7,$38,$F8,$B3,$C1,$F7,$9E,$1C,$3D
	.dc.b	$98,$FB,$A3,$EE,$1F,$53,$1F,$45,$FE,$4F,$0E,$B3,$E1,$8F,$7D,$FE
	.dc.b	$1A,$3A,$1C,$AF,$A5,$FC,$03,$9C,$7C,$4A,$E2,$EC,$E3,$1B,$8C,$BE
	.dc.b	$D7,$77,$67,$E2,$5F,$F3,$AE,$57,$0B,$C1,$FD,$0E,$9F,$DE,$DB,$5F
	.dc.b	$E0,$1F,$99,$B8,$7F,$80,$66,$E1,$C9,$FF,$03,$B6,$BF,$BA,$46,$30
	.dc.b	$ED,$1F,$DD,$FD,$0E,$31,$FE,$A4,$8F,$F0,$1F,$C6,$E2,$64,$7F,$81
	.dc.b	$BE,$66,$E3,$FE,$9D,$B4,$72,$F9,$69,$5F,$E0,$CE,$DB,$7B,$9F,$11
	.dc.b	$27,$FC,$10,$F9,$7E,$67,$DB,$B3,$E7,$1F,$07,$E6,$9C,$C6,$2F,$FC
	.dc.b	$CA,$DC,$3F,$5B,$1F,$94,$8C,$5F,$F8,$96,$B8,$DF,$C0,$F9,$AF,$97
	.dc.b	$83,$E8,$3E,$13,$ED,$8C,$C6,$2F,$F8,$AC,$EF,$6C,$9C,$F2,$FF,$45
	.dc.b	$56,$DF,$E0,$1F,$3C,$D3,$BB,$63,$FC,$FB,$FB,$8B,$E6,$33,$52,$7F
	.dc.b	$CC,$38,$AF,$83,$FE,$64,$6B,$FE,$92,$6B,$E0,$9F,$6F,$5B,$C8,$E9
	.dc.b	$3F,$18,$FE,$7A,$FE,$0B,$E6,$FE,$6F,$A5,$71,$F9,$AE,$1A,$7F,$92
	.dc.b	$23,$15,$CF,$67,$CF,$FF,$4B,$AF,$F4,$CB,$67,$F1,$1A,$7D,$E8,$C5
	.dc.b	$7E,$81,$B5,$FC,$31,$B5,$FE,$5F,$AE,$0F,$D8,$FA,$07,$C4,$53,$ED
	.dc.b	$FE,$C6,$E3,$C6,$3F,$EB,$DA,$DB,$F3,$1B,$7F,$E8,$E6,$BF,$EB,$C6
	.dc.b	$BF,$CE,$5D,$E5,$F8,$2F,$A1,$7C,$27,$6A,$F9,$82,$33,$7F,$E6,$EF
	.dc.b	$99,$FD,$79,$FB,$0E,$0B,$19,$25,$7E,$B6,$32,$FE,$BF,$D9,$FD,$FE
	.dc.b	$3F,$DD,$AB,$FC,$F3,$18,$E2,$91,$FC,$D0,$FE,$F9,$E0,$7F,$50,$8F
	.dc.b	$DF,$CE,$AF,$F0,$53,$FD,$53,$C2,$EB,$F3,$1D,$9B,$87,$3A,$56,$E3
	.dc.b	$0F,$F5,$CA,$E1,$18,$C5,$18,$AE,$F1,$89,$D5,$F6,$9C,$FC,$CF,$B8
	.dc.b	$F7,$FB,$96,$BA,$45,$9E,$73,$6F,$FC,$13,$8F,$F6,$EF,$99,$F8,$E7
	.dc.b	$C8,$F2,$AE,$3E,$58,$B8,$79,$26,$DF,$21,$F2,$FF,$40,$FA,$3E,$65
	.dc.b	$C3,$DD,$1F,$CC,$27,$71,$FE,$D8,$C7,$FB,$FC,$E4,$91,$F2,$F0,$FC
	.dc.b	$37,$AD,$CD,$FC,$4F,$1A,$3E,$FF,$C9,$CA,$F9,$9E,$1D,$BF,$D8,$76
	.dc.b	$3E,$17,$FD,$8F,$A7,$EC,$9B,$19,$BF,$F7,$5C,$E5,$F3,$2F,$93,$FC
	.dc.b	$6F,$C9,$7E,$39,$B3,$F6,$FB,$BF,$22,$7E,$4B,$E4,$AC,$7D,$FF,$73
	.dc.b	$E1,$11,$EA,$FC,$5F,$0F,$F3,$3F,$25,$F9,$A6,$CF,$E1,$FD,$EF,$CE
	.dc.b	$9F,$9E,$F9,$2B,$1F,$C2,$BD,$FF,$59,$1F,$5E,$C8,$FA,$5F,$25,$FA
	.dc.b	$63,$FB,$2F,$3F,$F0,$A3,$FC,$62,$E3,$E4,$A5,$FE,$6A,$5E,$DE,$39
	.dc.b	$3F,$4B,$0F,$9B,$FE,$DA,$DA,$FB,$CD,$9F,$F1,$F1,$FA,$67,$0F,$FB
	.dc.b	$F5,$FE,$06,$F4,$F1,$BC,$BF,$25,$E6,$7F,$B0,$FC,$E3,$E0,$E9,$EA
	.dc.b	$F7,$2E,$3F,$B3,$E9,$7D,$07,$F4,$98,$FB,$AD,$CB,$C7,$3A,$F3,$FF
	.dc.b	$0B,$39,$9F,$3F,$D0,$3E,$2F,$FA,$EC,$7F,$95,$E3,$18,$7F,$52,$6C
	.dc.b	$7B,$83,$87,$DD,$F2,$9F,$DE,$BE,$F5,$D5,$F4,$A3,$0C,$BF,$C9,$CE
	.dc.b	$9B,$1F,$9A,$8C,$57,$7F,$AB,$FF,$07,$D7,$EA,$5B,$1F,$D4,$67,$7A
	.dc.b	$7F,$76,$9C,$BF,$E2,$CD,$DF,$0F,$F3,$8F,$C2,$FC,$B5,$C7,$E6,$4E
	.dc.b	$9F,$D3,$B7,$1F,$96,$D7,$FC,$F0,$D7,$FC,$C6,$EE,$FF,$C0,$FF,$15
	.dc.b	$FA,$DA,$EB,$87,$F8,$42,$31,$7F,$8E,$F9,$6F,$CB,$9F,$98,$F8,$FB
	.dc.b	$7F,$95,$8F,$C8,$91,$FC,$8B,$F9,$4F,$1B,$F3,$E7,$E8,$3E,$3E,$CF
	.dc.b	$E7,$63,$F4,$7F,$2D,$8C,$3F,$3A,$FE,$9B,$E5,$7F,$4F,$F2,$D6,$FF
	.dc.b	$48,$E1,$A7,$E9,$E3,$15,$FD,$83,$E8,$3F,$31,$E0,$7F,$AF,$3D,$5F
	.dc.b	$F5,$04,$62,$7F,$D6,$F1,$FE,$A3,$F0,$3F,$8A,$46,$0F,$F6,$94,$EF
	.dc.b	$E3,$95,$CB,$B9,$87,$5B,$DC,$61,$E4,$7E,$DF,$5F,$F6,$5D,$F2,$8D
	.dc.b	$D7,$15,$87,$7C,$3B,$4E,$71,$F2,$DF,$1B,$DE,$3D,$E3,$18,$71,$9E
	.dc.b	$A7,$8A,$B8,$E1,$D2,$3C,$B3,$B7,$9E,$1D,$F1,$F3,$D6,$F0,$EB,$8F
	.dc.b	$B6,$FF,$86,$E1,$F2,$DB,$19,$7B,$1F,$C4,$BE,$3D,$E8,$7F,$A3,$A3
	.dc.b	$F7,$11,$8F,$8B,$FF,$26,$B6,$33,$3B,$7B,$FC,$8B,$DD,$FF,$29,$E2
	.dc.b	$8C,$4F,$F0,$38,$C4,$EA,$73,$3C,$A3,$E8,$4E,$A3,$E1,$CF,$2C,$BE
	.dc.b	$5F,$BF,$FE,$DE,$D3,$F1,$33,$96,$E1,$CA,$31,$7F,$88,$AE,$0F,$E2
	.dc.b	$A3,$2F,$DE,$6E,$1E,$1C,$3B,$CF,$3B,$8F,$91,$F3,$DF,$48,$E5,$1A
	.dc.b	$9F,$C0,$DF,$4A,$F0,$F8,$FF,$E2,$6F,$97,$FF,$7E,$71,$FF,$8C,$A1
	.dc.b	$18,$C7,$FA,$33,$7A,$DF,$35,$F2,$1F,$F6,$1B,$6F,$94,$F1,$BE,$D7
	.dc.b	$FD,$0E,$B4,$AF,$9E,$BB,$A7,$CC,$53,$C1,$7F,$9D,$AE,$F5,$82,$5F
	.dc.b	$5F,$80,$F4,$2F,$A0,$FE,$7A,$FF,$A5,$9A,$FF,$45,$6B,$FD,$91,$A8
	.dc.b	$8F,$0A,$FF,$56,$B7,$3D,$4F,$7F,$13,$CF,$EF,$F8,$A9,$E9,$FC,$9F
	.dc.b	$45,$EA,$D3,$51,$30,$F3,$3F,$EA,$C7,$D4,$DA,$E7,$1A,$3F,$C6,$CE
	.dc.b	$1F,$F3,$EB,$72,$FF,$7B,$FD,$03,$D3,$F8,$BE,$1B,$D0,$A4,$C3,$FC
	.dc.b	$81,$F4,$0F,$9B,$F9,$6D,$8F,$F5,$FB,$5F,$ED,$CD,$7F,$C1,$AD,$7E
	.dc.b	$2D,$7F,$A4,$9E,$B3,$D0,$F1,$FA,$7F,$17,$82,$4F,$F8,$27,$7F,$9C
	.dc.b	$FD,$3F,$9F,$FC,$17,$17,$D0,$FA,$07,$CB,$7C,$76,$D7,$FB,$51,$AF
	.dc.b	$F8,$0D,$AF,$F8,$8A,$9F,$0F,$F8,$86,$C7,$F9,$85,$AF,$FB,$27,$E8
	.dc.b	$3F,$C5,$CF,$F8,$CF,$4F,$CC,$11,$FF,$1B,$7A,$DF,$92,$FA,$A7,$A1
	.dc.b	$F6,$DB,$3F,$EE,$86,$BF,$E1,$76,$BF,$E3,$3F,$84,$FD,$21,$1F,$D5
	.dc.b	$7C,$1F,$E9,$F7,$91,$FE,$45,$7F,$61,$E2,$B8,$3C,$5E,$57,$84,$95
	.dc.b	$E1,$7F,$DF,$3E,$17,$F2,$9F,$9D,$E2,$49,$FD,$C7,$C0,$F4,$D7,$AD
	.dc.b	$C9,$FF,$56,$E3,$EA,$EC,$F0,$47,$EA,$7E,$AB,$FD,$0C,$FF,$8C,$C3
	.dc.b	$87,$FC,$6C,$D7,$FC,$AA,$D7,$FD,$8B,$3F,$CE,$DE,$FE,$47,$B8,$FE
	.dc.b	$1C,$FF,$92,$B4,$DF,$AD,$F0,$BE,$7A,$4F,$9A,$3F,$82,$E1,$FE,$73
	.dc.b	$8F,$23,$5F,$1E,$FF,$B1,$8C,$47,$EF,$B7,$FE,$9D,$F3,$9F,$38,$91
	.dc.b	$FE,$CE,$3F,$3C,$7F,$BB,$BE,$85,$F3,$3F,$53,$E7,$A4,$FF,$AB,$67
	.dc.b	$2D,$9F,$E0,$11,$88,$FF,$2D,$35,$FF,$3C,$B5,$FE,$22,$EF,$7F,$A3
	.dc.b	$08,$C6,$BA,$AF,$8F,$E1,$7E,$19,$95,$F6,$9F,$97,$EE,$E2,$3F,$DE
	.dc.b	$3C,$3C,$9F,$F7,$6B,$74,$FE,$4D,$3F,$90,$EF,$95,$F8,$47,$EA,$FB
	.dc.b	$98,$8E,$71,$74,$FF,$71,$C2,$7F,$33,$F3,$D2,$3F,$1A,$FD,$EF,$73
	.dc.b	$11,$E4,$8B,$5C,$A6,$EB,$F8,$89,$B4,$CB,$F9,$E5,$7E,$9B,$E6,$3F
	.dc.b	$4F,$F3,$D2,$BF,$37,$19,$69,$0F,$EB,$18,$7F,$45,$BF,$E8,$DE,$EF
	.dc.b	$EC,$3E,$76,$57,$E9,$1F,$D9,$FC,$C6,$23,$E5,$C5,$D3,$F4,$F7,$FE
	.dc.b	$19,$AF,$EB,$3D,$0F,$95,$FE,$B7,$CA,$F9,$7A,$6A,$26,$5F,$AD,$AE
	.dc.b	$0D,$8F,$E9,$51,$88,$FF,$52,$37,$1F,$C3,$B7,$4F,$EB,$6D,$9F,$EC
	.dc.b	$4B,$BF,$FC,$05,$F2,$FF,$EC,$19,$FF,$5F,$35,$F8,$A7,$F7,$D7,$17
	.dc.b	$FC,$10,$3F,$C1,$EF,$FB,$74,$63,$BF,$A9,$2B,$AF,$81,$E0,$EB,$FF
	.dc.b	$09,$7C,$2B,$DB,$A6,$6B,$C7,$35,$F1,$E1,$E0,$8F,$4F,$77,$F8,$1B
	.dc.b	$E8,$7F,$21,$5E,$2C,$FE,$0F,$C8,$7D,$23,$5D,$72,$7F,$8D,$57,$07
	.dc.b	$FD,$9C,$D7,$FD,$DA,$D7,$FC,$0A,$EF,$1F,$DA,$B6,$F3,$5F,$E1,$8F
	.dc.b	$A1,$7C,$47,$F8,$6B,$91,$F2,$F7,$DB,$99,$F9,$CF,$F0,$B4,$E9,$C1
	.dc.b	$F8,$79,$E4,$47,$31,$FE,$12,$7E,$CF,$57,$CC,$E4,$91,$82,$1F,$29
	.dc.b	$47,$E0,$8F,$C1,$F6,$FF,$50,$F9,$CF,$E9,$7E,$19,$A4,$BF,$F7,$A7
	.dc.b	$6F,$F9,$DE,$38,$47,$F4,$A7,$FC,$91,$F0,$DF,$C6,$B8,$BF,$4A,$FE
	.dc.b	$2B,$87,$FA,$1F,$C8,$7F,$5E,$37,$D5,$30,$F8,$98,$AF,$84,$7F,$01
	.dc.b	$6C,$67,$AB,$FB,$3F,$4F,$A0,$95,$E6,$D3,$F1,$CD,$8D,$BF,$D8,$1F
	.dc.b	$3F,$67,$F9,$43,$E1,$DF,$33,$FE,$5D,$AF,$F0,$6F,$93,$FE,$46,$1F
	.dc.b	$27,$FE,$48,$F5,$FF,$C6,$53,$FE,$4F,$9C,$BF,$E5,$18,$CA,$BE,$BC
	.dc.b	$39,$AF,$F2,$8C,$62,$5F,$43,$C2,$79,$C6,$1F,$16,$1F,$E5,$89,$EB
	.dc.b	$5C,$5B,$8F,$A5,$E5,$3B,$AE,$37,$FE,$65,$9D,$E1,$67,$C7,$7C,$5D
	.dc.b	$D7,$84,$F7,$5F,$37,$FC,$DD,$87,$86,$78,$DF,$B6,$76,$53,$FC,$B1
	.dc.b	$5F,$26,$77,$1F,$38,$E7,$7E,$17,$89,$E1,$38,$9E,$73,$75,$5F,$9D
	.dc.b	$1C,$5D,$9C,$67,$E5,$E1,$F0,$23,$A6,$1F,$47,$1F,$8F,$1E,$7A,$FA
	.dc.b	$31,$B3,$85,$6A,$7A,$63,$F0,$DF,$0E,$3E,$3C,$3E,$24,$7A,$AB,$81
	.dc.b	$CB,$0F,$83,$87,$32,$38,$3E,$2C,$70,$7E,$9B,$C6,$7E,$09,$E2,$3E
	.dc.b	$14,$7B,$8E,$24,$7C,$4F,$23,$0F,$A0,$FA,$47,$C8,$9F,$1B,$D9,$F0
	.dc.b	$3F,$32,$7C,$8F,$89,$F8,$D3,$F8,$89,$E3,$3D,$5C,$BF,$71,$CD,$7D
	.dc.b	$19,$FC,$1E,$9F,$E7,$88,$FD,$23,$72,$FD,$A3,$71,$FF,$5F,$3F,$E7
	.dc.b	$EC,$3F,$D8,$1F,$11,$8C,$7F,$D7,$B1,$81,$3B,$53,$B7,$6D,$7F,$D8
	.dc.b	$AF,$E6,$69,$F3,$7F,$E6,$6B,$FD,$ED,$75,$7F,$7F,$3A,$E4,$FD,$04
	.dc.b	$E7,$3F,$F6,$7C,$E6,$3E,$79,$93,$FB,$5D,$F5,$9F,$4C,$79,$8F,$D1
	.dc.b	$4F,$1D,$3C,$31,$97,$C3,$5F,$E7,$A7,$38,$FA,$63,$A4,$FB,$8F,$BF
	.dc.b	$5E,$3C,$77,$1D,$A7,$96,$1E,$2B,$FD,$75,$F8,$63,$15,$BC,$7F,$64
	.dc.b	$FC,$FC,$7C,$71,$88,$F3,$4F,$C2,$C3,$AC,$78,$EB,$6F,$EF,$CF,$88
	.dc.b	$2F,$CD,$5E,$0C,$76,$7A,$2B,$89,$63,$B6,$1E,$1C,$7D,$38,$F8,$31
	.dc.b	$FB,$A7,$A4,$F0,$1F,$3F,$0E,$B7,$DC,$E7,$1F,$60,$F5,$9F,$62,$FB
	.dc.b	$3C,$F0,$F1,$39,$3A,$E1,$E2,$9D,$DF,$69,$FA,$8D,$7C,$B5,$EC,$BF
	.dc.b	$9E,$D8,$F9,$11,$E4,$8E,$F1,$F1,$F0,$F7,$3F,$4E,$74,$72,$2B,$E0
	.dc.b	$7D,$B5,$F8,$23,$C9,$C3,$FA,$8E,$9F,$F2,$16,$3F,$EB,$C6,$DF,$F7
	.dc.b	$5F,$06,$D6,$EB,$FD,$9B,$5F,$E9,$CA,$E1,$1F,$DD,$63,$11,$FE,$94
	.dc.b	$6C,$FF,$80,$7C,$BE,$7C,$A6,$F9,$D8,$66,$1F,$ED,$AE,$CB,$8F,$FA
	.dc.b	$81,$F1,$FC,$E7,$F0,$09,$FF,$06,$B6,$7F,$CB,$F8,$FF,$83,$FC,$CD
	.dc.b	$F1,$CB,$FF,$0F,$DE,$EF,$FD,$94,$D7,$FD,$14,$D7,$F2,$DF,$42,$F9
	.dc.b	$BF,$8B,$CA,$8C,$5F,$F9,$06,$31,$F3,$9A,$9C,$DF,$F9,$81,$B3,$FE
	.dc.b	$71,$F5,$3F,$D1,$97,$C6,$72,$FF,$9C,$A3,$35,$FE,$7A,$1F,$43,$F2
	.dc.b	$67,$FC,$9A,$D8,$D3,$FE,$DE,$8C,$EF,$F7,$27,$8B,$7E,$99,$F0,$79
	.dc.b	$BE,$59,$7F,$B6,$27,$38,$7F,$96,$B6,$FF,$50,$37,$93,$A7,$D0,$B8
	.dc.b	$7B,$56,$B4,$F4,$35,$FE,$72,$D4,$7F,$40,$AF,$C4,$BF,$8F,$3F,$93
	.dc.b	$C7,$3C,$7F,$DD,$7C,$2E,$0E,$F7,$ED,$8C,$65,$F2,$FE,$13,$E9,$1F
	.dc.b	$4F,$D1,$B1,$93,$EC,$70,$FF,$39,$C7,$FD,$49,$C4,$BF,$BF,$39,$AF
	.dc.b	$F5,$2F,$CE,$BE,$57,$F0,$6D,$8F,$C1,$3E,$C6,$C6,$88,$EB,$6A,$7D
	.dc.b	$DF,$A0,$F9,$ED,$E4,$F6,$7D,$07,$D9,$9C,$E8,$BF,$93,$69,$E2,$FF
	.dc.b	$66,$6B,$A6,$BF,$41,$AF,$1B,$ED,$E8,$F7,$6D,$7E,$BB,$BB,$F6,$27
	.dc.b	$EC,$BE,$42,$C6,$49,$7F,$B2,$E1,$4E,$A3,$1F,$23,$FB,$A3,$F8,$6E
	.dc.b	$7F,$B9,$3F,$DE,$34,$8D,$5F,$EF,$BD,$8B,$7D,$5F,$E0,$7E,$94,$9F
	.dc.b	$F4,$7E,$1E,$07,$F8,$2C,$66,$BC,$8F,$83,$1C,$57,$82,$2D,$7E,$53
	.dc.b	$BB,$FA,$B5,$62,$D0,$CB,$F0,$91,$86,$C7,$F5,$DC,$BF,$8B,$F8,$93
	.dc.b	$0F,$E2,$EF,$C0,$8C,$5F,$C7,$F4,$BF,$8E,$9F,$13,$2B,$7F,$C7,$A2
	.dc.b	$CF,$F1,$EC,$D5,$87,$F0,$91,$F7,$7E,$13,$F1,$07,$CE,$AB,$7F,$ED
	.dc.b	$4E,$14,$FE,$C1,$B1,$FE,$16,$AF,$C6,$4E,$49,$1F,$CC,$3C,$9E,$73
	.dc.b	$F7,$DF,$E6,$BF,$2D,$27,$E9,$B6,$7F,$8D,$79,$4F,$67,$F7,$ED,$AF
	.dc.b	$03,$67,$FC,$2D,$3F,$D2,$B8,$3F,$7B,$1F,$8D,$F1,$7F,$A1,$F7,$FF
	.dc.b	$44,$7F,$A2,$FC,$2C,$C7,$F0,$7D,$36,$FC,$3E,$47,$80,$93,$E0,$7B
	.dc.b	$4F,$F4,$EF,$A0,$FE,$9E,$D9,$FE,$9F,$F2,$7F,$D4,$8F,$CC,$FC,$9D
	.dc.b	$8F,$F7,$FF,$D5,$3E,$23,$F8,$DB,$63,$E6,$37,$2F,$F2,$1B,$5F,$EA
	.dc.b	$7E,$2F,$F5,$E3,$FA,$FF,$89,$6B,$FA,$FE,$1F,$D8,$8F,$EE,$BD,$9F
	.dc.b	$C4,$6F,$F6,$11,$FC,$FF,$07,$A9,$E2,$25,$F8,$9F,$99,$EF,$FF,$6C
	.dc.b	$3F,$B6,$FC,$95,$9F,$EB,$9F,$40,$F6,$FF,$B0,$B6,$3F,$B0,$7C,$13
	.dc.b	$DC,$CD,$7F,$4F,$FA,$07,$C7,$7F,$3C,$6D,$FE,$35,$BB,$7F,$99,$1A
	.dc.b	$FF,$9B,$3E,$97,$B3,$FE,$47,$9F,$E8,$FE,$1A,$72,$FE,$80,$7D,$33
	.dc.b	$A3,$FB,$BC,$EA,$B8,$FD,$0F,$F3,$D7,$99,$FC,$AB,$E8,$7F,$B0,$BA
	.dc.b	$E6,$FE,$D1,$18,$CB,$FA,$BB,$FD,$BD,$FF,$49,$B5,$E5,$EC,$FF,$5A
	.dc.b	$8C,$47,$F9,$7F,$EB,$1F,$0F,$FE,$0A,$6E,$1E,$29,$EB,$79,$AF,$EF
	.dc.b	$1F,$09,$FE,$04,$9F,$EC,$ED,$D9,$C7,$F7,$B0,$E5,$4A,$E0,$DF,$07
	.dc.b	$FB,$6C,$7F,$85,$23,$FD,$56,$D7,$FA,$5B,$5E,$F1,$F2,$BD,$7F,$1B
	.dc.b	$6B,$BD,$3C,$2F,$39,$F9,$4F,$98,$B3,$A8,$F0,$91,$FE,$1D,$3C,$5F
	.dc.b	$F5,$2F,$77,$F2,$EF,$8E,$FC,$E9,$F9,$EF,$A0,$B3,$F1,$E7,$4D,$F3
	.dc.b	$7F,$B8,$69,$FE,$42,$95,$FF,$96,$A7,$3B,$7F,$26,$9F,$E7,$7F,$45
	.dc.b	$9F,$63,$F5,$6D,$C7,$A4,$5C,$7A,$4D,$BF,$E5,$F0,$FA,$1F,$D1,$57
	.dc.b	$FB,$63,$D3,$FE,$EB,$18,$9F,$F6,$AC,$7F,$78,$FA,$26,$BF,$CE,$3C
	.dc.b	$FF,$EC,$15,$FE,$24,$BE,$11,$88,$ED,$E0,$FF,$84,$5E,$66,$DC,$46
	.dc.b	$EF,$AE,$9E,$3A,$D6,$55,$E6,$DC,$3B,$E5,$E5,$CB,$C8,$FA,$B2,$EA
	.dc.b	$DA,$E1,$18,$BF,$F1,$DE,$3E,$07,$CA,$76,$2E,$1D,$9F,$19,$A8,$E1
	.dc.b	$97,$BE,$75,$3C,$4F,$F8,$02,$FC,$B6,$D7,$C3,$6D,$FF,$59,$F8,$87
	.dc.b	$A2,$F8,$5C,$36,$BF,$DE,$73,$A9,$C9,$FD,$87,$0F,$E7,$5F,$01,$7C
	.dc.b	$D3,$49,$7F,$EF,$C9,$DB,$EE,$9D,$B6,$7E,$77,$C5,$7F,$8E,$34,$F7
	.dc.b	$56,$9B,$3F,$4E,$31,$B7,$E2,$AB,$83,$F8,$C8,$CD,$FE,$1D,$AF,$E4
	.dc.b	$DB,$82,$B8,$CE,$EF,$B1,$1F,$AB,$1A,$79,$46,$70,$F2,$C6,$8F,$C6
	.dc.b	$71,$FF,$94,$7E,$2B,$FC,$A7,$C7,$FA,$DE,$57,$16,$39,$3F,$C5,$91
	.dc.b	$8C,$3F,$71,$79,$7F,$C5,$B0,$8C,$61,$FE,$5D,$BB,$3C,$6B,$FD,$7E
	.dc.b	$ED,$FE,$09,$2F,$A1,$DF,$BF,$27,$31,$D2,$75,$18,$21,$FD,$3C,$64
	.dc.b	$FA,$1F,$D9,$3E,$DF,$A4,$FF,$13,$57,$F9,$C5,$B8,$9A,$4B,$FE,$2C
	.dc.b	$FA,$8F,$5F,$C6,$7B,$3E,$93,$DB,$F3,$12,$BF,$8E,$46,$5F,$E4,$B6
	.dc.b	$8E,$E4,$8C,$BB,$7E,$2B,$F3,$71,$7C,$3B,$D8,$F9,$7E,$14,$BF,$E5
	.dc.b	$4F,$CE,$F8,$69,$1E,$07,$DA,$47,$FC,$8D,$0C,$F7,$1F,$52,$7E,$E2
	.dc.b	$9D,$3F,$EE,$7F,$D2,$7E,$5F,$27,$F7,$C6,$C7,$D0,$6E,$47,$FC,$BF
	.dc.b	$1F,$EB,$8F,$33,$FC,$AD,$3E,$C6,$E7,$FE,$B9,$9F,$F5,$29,$FE,$EB
	.dc.b	$8F,$F2,$E2,$FA,$17,$CD,$61,$B1,$FE,$61,$6B,$C8,$E1,$1F,$7B,$C3
	.dc.b	$FC,$99,$23,$FA,$97,$97,$F9,$6F,$1B,$F0,$64,$8F,$B2,$FE,$1B,$E6
	.dc.b	$BF,$36,$48,$FC,$CF,$C2,$7E,$7B,$C8,$91,$FD,$75,$FD,$0F,$D2,$FE
	.dc.b	$8B,$AF,$F1,$E4,$9F,$C2,$BF,$92,$FA,$0F,$D3,$92,$3F,$B4,$78,$2F
	.dc.b	$9F,$FD,$5F,$B3,$FA,$CC,$A5,$7F,$6B,$8F,$D7,$F1,$7E,$70,$91,$F9
	.dc.b	$27,$F4,$3F,$6B,$FD,$8B,$93,$FB,$63,$79,$BF,$D5,$4D,$AF,$F5,$4E
	.dc.b	$5F,$E6,$57,$FB,$D4,$78,$DB,$E7,$7F,$00,$8F,$F0,$D4,$6B,$3C,$79
	.dc.b	$DE,$02,$47,$F8,$22,$7F,$CA,$4D,$F2,$79,$11,$F8,$3F,$2D,$D7,$E0
	.dc.b	$3B,$7D,$04,$9F,$1C,$65,$E8,$FB,$88,$E0,$D9,$0F,$E1,$1E,$2F,$F4
	.dc.b	$6F,$89,$F5,$5C,$9F,$7B,$FC,$73,$57,$C3,$E6,$78,$E6,$AC,$70,$7F
	.dc.b	$85,$EF,$75,$F2,$30,$F8,$D8,$48,$C1,$0E,$71,$F2,$CF,$07,$67,$CD
	.dc.b	$3E,$77,$B1,$27,$E9,$C6,$5F,$98,$FD,$E2,$38,$36,$43,$E0,$DA,$7E
	.dc.b	$D7,$C9,$C9,$FB,$EF,$C1,$F9,$A9,$18,$21,$C9,$FC,$11,$F8,$3F,$9B
	.dc.b	$91,$F8,$88,$CB,$F8,$A8,$CB,$F2,$A3,$F8,$8F,$CF,$1F,$E2,$7F,$3A
	.dc.b	$EB,$A6,$A2,$57,$FB,$86,$B8,$36,$3F,$84,$C6,$23,$F8,$23,$76,$FF
	.dc.b	$21,$B7,$A1,$32,$30,$DF,$73,$F9,$A7,$D5,$FF,$90,$A3,$F4,$7F,$41
	.dc.b	$FA,$13,$3E,$2C,$8F,$F4,$36,$2A,$E1,$1F,$D5,$E3,$13,$FC,$FF,$E7
	.dc.b	$7F,$47,$3F,$8C,$F9,$FF,$D6,$11,$FF,$5C,$FB,$1F,$D1,$7E,$07,$FC
	.dc.b	$0D,$1F,$B4,$9D,$3D,$63,$2F,$F6,$51,$C0,$F1,$FF,$9F,$D3,$E3,$29
	.dc.b	$F4,$22,$57,$F2,$B3,$8B,$FC,$D2,$B9,$46,$5A,$FF,$31,$6E,$7F,$E7
	.dc.b	$06,$F0,$7C,$E9,$E6,$7F,$A8,$BB,$7C,$4F,$F0,$EF,$8A,$FE,$18,$47
	.dc.b	$FD,$4F,$C3,$FD,$31,$F6,$37,$0F,$E3,$73,$B9,$CF,$CA,$FF,$15,$DA
	.dc.b	$47,$F1,$A7,$E5,$6F,$23,$04,$3E,$53,$FC,$77,$0F,$E5,$24,$8F,$E3
	.dc.b	$4D,$7F,$81,$BB,$7F,$93,$C7,$F0,$9B,$FF,$12,$E5,$FE,$4E,$9F,$E2
	.dc.b	$2D,$C7,$F8,$44,$EE,$7F,$A0,$B5,$FE,$2F,$EF,$FE,$1A,$7F,$D4,$94
	.dc.b	$F9,$9F,$F0,$EB,$6B,$FD,$6A,$D7,$FB,$6F,$82,$8C,$47,$D3,$8C,$4F
	.dc.b	$98,$91,$FE,$B9,$97,$CC,$7F,$81,$BE,$09,$F2,$DF,$E3,$0A,$F1,$7A
	.dc.b	$FF,$74,$FF,$1F,$38,$3F,$5A,$E3,$0F,$F0,$6C,$2B,$9E,$5F,$B2,$9C
	.dc.b	$C6,$25,$F3,$FC,$2F,$AB,$7E,$07,$F1,$8F,$0C,$7C,$8D,$C7,$F7,$AE
	.dc.b	$EE,$DF,$9E,$F8,$BF,$95,$C3,$C1,$5C,$6B,$D4,$7F,$8A,$6B,$F3,$58
	.dc.b	$F3,$8C,$D7,$9E,$38,$19,$25,$F5,$C7,$E4,$47,$5C,$39,$12,$75,$8F
	.dc.b	$9B,$0F,$E2,$B8,$79,$27,$D1,$19,$3C,$85,$8D,$3C,$30,$EF,$3C,$23
	.dc.b	$D6,$FD,$33,$BB,$EA,$8F,$93,$1B,$BF,$24,$F3,$BE,$D5,$A8,$FB,$B8
	.dc.b	$FC,$A6,$9A,$98,$F8,$23,$E4,$DF,$8F,$1F,$8E,$F8,$B0,$EF,$5E,$C3
	.dc.b	$F0,$31,$92,$3D,$8C,$96,$3E,$64,$6A,$39,$47,$4C,$7E,$2C,$7E,$47
	.dc.b	$1F,$8A,$F3,$BF,$AE,$FB,$CF,$89,$8F,$CF,$C3,$F2,$27,$C5,$3B,$5F
	.dc.b	$E0,$5F,$61,$27,$95,$7B,$9E,$D1,$97,$C5,$5F,$A1,$8F,$A7,$E1,$7E
	.dc.b	$E6,$FF,$44,$DC,$3F,$60,$DA,$FE,$B2,$FE,$F2,$DF,$05,$8C,$7F,$CC
	.dc.b	$8D,$8F,$F3,$7B,$5F,$F6,$F1,$FE,$E8,$8D,$BF,$A0,$A7,$CC,$FF,$09
	.dc.b	$8F,$D1,$9F,$DA,$2B,$D0,$FF,$6A,$9D,$ED,$88,$EB,$7F,$C0,$27,$31
	.dc.b	$C8,$F3,$9D,$EF,$B4,$FF,$06,$8E,$C7,$DE,$3F,$D7,$F1,$E3,$D3,$51
	.dc.b	$C9,$ED,$19,$D3,$D7,$58,$C3,$C9,$7D,$23,$0F,$8F,$1F,$D6,$63,$E6
	.dc.b	$C7,$AC,$63,$1E,$95,$F2,$8F,$25,$6F,$1F,$93,$7D,$0E,$33,$AA,$E2
	.dc.b	$FD,$08,$D4,$F9,$8F,$F7,$AC,$FC,$27,$13,$F3,$63,$0F,$86,$78,$B6
	.dc.b	$78,$E3,$FC,$16,$3E,$4C,$E9,$F0,$4E,$9F,$98,$7C,$37,$18,$FA,$0F
	.dc.b	$8A,$7A,$A7,$94,$F4,$9E,$51,$96,$9F,$08,$B1,$F1,$E7,$9C,$7B,$23
	.dc.b	$ED,$D7,$D4,$7F,$07,$3A,$78,$4F,$90,$E8,$FC,$39,$F7,$3F,$81,$8E
	.dc.b	$EF,$B8,$D1,$EE,$34,$62,$BD,$D1,$D6,$3B,$9C,$2B,$FD,$C7,$BF,$F6
	.dc.b	$9F,$85,$FD,$16,$5F,$DC,$9B,$8F,$F9,$2A,$7F,$B7,$7C,$D6,$30,$47
	.dc.b	$19,$FE,$D9,$5F,$AB,$9F,$EE,$50,$8C,$47,$38,$C1,$FD,$F2,$33,$1F
	.dc.b	$EB,$6E,$4E,$D4,$D9,$18,$E3,$FF,$3F,$61,$FE,$CC,$8D,$CE,$A3,$FB
	.dc.b	$F7,$3E,$27,$9C,$22,$EF,$23,$FC,$09,$F3,$57,$2F,$F0,$DC,$F9,$32
	.dc.b	$FF,$07,$36,$BF,$7F,$3E,$E8,$F4,$EA,$F4,$9B,$78,$38,$FE,$44,$E6
	.dc.b	$31,$87,$F8,$4E,$6E,$1F,$E6,$BB,$E3,$39,$3F,$89,$1F,$E4,$FF,$9E
	.dc.b	$7C,$37,$B3,$E2,$7E,$A5,$A7,$35,$FE,$1F,$9D,$34,$E7,$3F,$E8,$69
	.dc.b	$DB,$FE,$3A,$9E,$7F,$3D,$FE,$99,$AD,$CA,$30,$FF,$9C,$FE,$71,$F3
	.dc.b	$9C,$3C,$A9,$7F,$E6,$FB,$DD,$FF,$99,$CE,$53,$F8,$06,$CE,$E3,$D7
	.dc.b	$8B,$DA,$E7,$5F,$E8,$18,$C5,$7F,$A5,$34,$FE,$63,$BF,$F9,$2B,$2F
	.dc.b	$9F,$18,$9F,$C8,$47,$E0,$88,$F1,$E0,$B8,$7F,$A6,$FE,$4D,$ED,$7D
	.dc.b	$86,$D7,$FA,$8B,$E7,$5F,$1F,$9A,$93,$FE,$4D,$F9,$1B,$87,$F9,$91
	.dc.b	$FA,$BF,$16,$F8,$1A,$63,$E2,$6D,$FC,$E9,$D3,$FE,$97,$6D,$FC,$90
	.dc.b	$E1,$F9,$8D,$7F,$D2,$8F,$0D,$FF,$69,$EA,$FE,$32,$7F,$D5,$53,$9A
	.dc.b	$FF,$4D,$4E,$6B,$FC,$E7,$F2,$3F,$A8,$F9,$DF,$D4,$B6,$BF,$CD,$4D
	.dc.b	$C3,$FD,$02,$DA,$FE,$D9,$D8,$E8,$E9,$7F,$C0,$7E,$5B,$F7,$B1,$8C
	.dc.b	$3F,$85,$CA,$7F,$6E,$47,$F2,$F5,$CF,$B7,$15,$FE,$90,$F8,$DF,$DE
	.dc.b	$9D,$3C,$0B,$7F,$E7,$F8,$FD,$C1,$1F,$F4,$5B,$67,$FD,$33,$F3,$BC
	.dc.b	$1B,$3F,$E8,$7F,$8C,$FE,$12,$7A,$3A,$2C,$64,$FF,$46,$7C,$EE,$2B
	.dc.b	$FD,$46,$FF,$10,$F5,$64,$FA,$B6,$78,$98,$BF,$F5,$4D,$5E,$0F,$93
	.dc.b	$5A,$BF,$F4,$C4,$FF,$1A,$F9,$C9,$1C,$FC,$5B,$87,$FA,$CA,$7F,$8B
	.dc.b	$12,$7F,$D5,$2D,$86,$1F,$5B,$EC,$51,$CA,$30,$FF,$32,$8F,$CF,$D7
	.dc.b	$D2,$6C,$72,$8F,$E8,$D8,$BC,$5E,$1D,$BC,$7D,$0F,$E5,$58,$1A,$4B
	.dc.b	$FC,$E7,$2A,$5F,$2D,$E1,$F2,$3C,$4D,$AF,$17,$CA,$F9,$4F,$C5,$F5
	.dc.b	$58,$F3,$79,$FE,$02,$3D,$B0,$47,$93,$E5,$3D,$DF,$2B,$EF,$6D,$7B
	.dc.b	$FE,$57,$E1,$1F,$98,$EB,$B1,$F0,$FC,$2F,$69,$1E,$CF,$C0,$F7,$FE
	.dc.b	$51,$F2,$FE,$4E,$CF,$CB,$F9,$3F,$9C,$7E,$93,$B2,$C7,$CF,$F7,$FE
	.dc.b	$49,$1F,$53,$F3,$3E,$4F,$EB,$1F,$5F,$C8,$B3,$FC,$E3,$E7,$1F,$27
	.dc.b	$FC,$BD,$B1,$FC,$2F,$B1,$18,$7F,$87,$35,$FC,$29,$F8,$8F,$39,$EF
	.dc.b	$7F,$60,$6D,$7F,$15,$EF,$46,$2B,$FA,$04,$62,$BF,$07,$F2,$9F,$8D
	.dc.b	$3F,$1D,$F1,$37,$0F,$EA,$B4,$F9,$97,$0F,$E4,$F4,$FE,$D7,$18,$AF
	.dc.b	$CA,$46,$5F,$BC,$FE,$40,$8F,$E0,$7D,$6F,$CE,$1B,$CF,$F3,$47,$F8
	.dc.b	$73,$AB,$FA,$6B,$8E,$FF,$CA,$7A,$BF,$A3,$3F,$A2,$7C,$15,$C3,$F4
	.dc.b	$91,$F9,$E2,$3F,$8F,$6D,$73,$F8,$04,$F1,$8C,$3F,$E2,$08,$7C,$AD
	.dc.b	$31,$EB,$59,$FE,$47,$3A,$7F,$52,$DA,$FE,$DD,$8F,$FA,$A2,$FF,$C3
	.dc.b	$53,$99,$FB,$73,$FD,$29,$FF,$17,$FA,$BF,$C0,$AF,$F9,$7C,$BA,$B3
	.dc.b	$52,$7C,$3F,$2B,$63,$D8,$FA,$7E,$57,$D5,$F0,$F6,$BC,$ED,$AF,$84
	.dc.b	$DE,$4F,$E8,$F9,$A3,$FC,$37,$F3,$2C,$7F,$B7,$DF,$FB,$0B,$E0,$FF
	.dc.b	$27,$39,$F3,$3F,$D7,$11,$F9,$6F,$29,$E1,$7F,$AF,$72,$FF,$29,$61
	.dc.b	$FC,$7B,$0E,$0F,$33,$18,$F6,$9D,$C2,$77,$18,$BE,$BA,$78,$2B,$8E
	.dc.b	$55,$E1,$87,$3C,$BC,$ED,$8F,$E5,$F7,$C9,$F1,$1C,$2B,$96,$3E,$0B
	.dc.b	$E3,$1D,$6B,$86,$3E,$6A,$D5,$F8,$8E,$6E,$1F,$E5,$79,$7C,$3C,$39
	.dc.b	$C6,$B1,$F9,$AF,$F8,$D6,$BE,$2B,$63,$2F,$03,$FC,$75,$F0,$DF,$CF
	.dc.b	$FE,$13,$F8,$53,$63,$FC,$2C,$D7,$7E,$7A,$7F,$88,$5B,$83,$FC,$FD
	.dc.b	$5F,$E1,$D8,$C7,$C2,$6A,$73,$3E,$56,$BF,$82,$78,$E9,$F0,$DB,$9E
	.dc.b	$6B,$E5,$FC,$47,$F1,$AA,$D4,$7C,$EA,$DD,$FE,$15,$AF,$E0,$6B,$85
	.dc.b	$F3,$8F,$9B,$3A,$C3,$CF,$1F,$42,$B6,$F3,$BE,$C4,$74,$5C,$3C,$6F
	.dc.b	$48,$F3,$BF,$8D,$E4,$4B,$E5,$BF,$99,$F2,$7E,$A1,$BC,$7F,$EB,$79
	.dc.b	$D5,$FF,$06,$F9,$5C,$BF,$EB,$88,$46,$30,$EB,$76,$78,$D7,$F1,$7A
	.dc.b	$C7,$CA,$48,$FE,$22,$EF,$E5,$71,$1F,$C7,$62,$E5,$FB,$C9,$E9,$F2
	.dc.b	$BF,$D3,$F2,$F5,$B7,$13,$49,$87,$F1,$E7,$8B,$FC,$83,$E5,$7C,$DF
	.dc.b	$2B,$E7,$F8,$A9,$3F,$C9,$23,$2F,$F2,$47,$B1,$1C,$1B,$21,$FC,$D7
	.dc.b	$E1,$BE,$17,$CA,$48,$FE,$5E,$FC,$5F,$92,$91,$FC,$C1,$F5,$11,$FE
	.dc.b	$67,$E6,$69,$FE,$81,$1F,$36,$3E,$77,$27,$CF,$F1,$65,$7F,$3C,$8F
	.dc.b	$A5,$F1,$DF,$14,$91,$FC,$F9,$F9,$5F,$86,$FE,$93,$8B,$FC,$DE,$DE
	.dc.b	$6F,$F7,$C3,$6B,$FD,$EE,$BE,$43,$F0,$70,$F9,$8F,$A2,$FF,$80,$DA
	.dc.b	$FF,$A3,$DA,$FF,$AE,$5A,$F6,$F9,$3F,$19,$1F,$F1,$DD,$23,$FC,$1D
	.dc.b	$83,$67,$17,$E5,$23,$0D,$8C,$DF,$F8,$09,$71,$7E,$5F,$0F,$E6,$B7
	.dc.b	$F0,$3B,$70,$67,$E2,$F5,$F1,$7F,$18,$98,$F9,$1F,$93,$F1,$3C,$C9
	.dc.b	$1F,$E8,$EC,$3F,$CE,$AD,$F8,$7F,$07,$87,$23,$06,$FE,$43,$E9,$91
	.dc.b	$F8,$2F,$D6,$F8,$FF,$41,$23,$B7,$C2,$7A,$FE,$3F,$D9,$F1,$F2,$7C
	.dc.b	$51,$EF,$F8,$2F,$BE,$48,$F9,$AF,$E0,$BE,$3F,$E2,$12,$3D,$5F,$1F
	.dc.b	$F7,$B7,$75,$7E,$03,$EF,$9C,$34,$D4,$4B,$FF,$22,$57,$06,$C7,$F9
	.dc.b	$1D,$46,$5F,$F6,$A3,$79,$3F,$C3,$B1,$FE,$8C,$8C,$75,$FF,$92,$9B
	.dc.b	$1F,$E9,$0C,$39,$BF,$47,$E1,$3E,$F1,$1F,$A9,$F1,$F5,$FF,$1A,$72
	.dc.b	$A7,$18,$FE,$8A,$BF,$0D,$F1,$F2,$FE,$E3,$FC,$7F,$D4,$FF,$29,$E7
	.dc.b	$FA,$68,$D4,$F4,$74,$EE,$78,$7C,$74,$8F,$F6,$9C,$7F,$68,$C3,$FD
	.dc.b	$BE,$DB,$FF,$55,$F9,$28,$FF,$50,$57,$0B,$E8,$F2,$8E,$8F,$F9,$3D
	.dc.b	$DB,$5F,$47,$AB,$E9,$F8,$E9,$3E,$28,$CB,$48,$73,$F8,$EF,$81,$F1
	.dc.b	$D2,$3D,$2F,$C3,$F8,$DC,$47,$86,$2E,$9E,$C9,$FE,$0F,$F7,$BF,$E2
	.dc.b	$CA,$FF,$63,$CF,$DE,$7F,$17,$EC,$FF,$0C,$E6,$FF,$1C,$46,$B7,$FF
	.dc.b	$44,$DF,$18,$FF,$37,$87,$0F,$FA,$38,$C7,$B5,$F8,$F8,$CC,$72,$8C
	.dc.b	$1E,$42,$3D,$CF,$F6,$DD,$3D,$FF,$C0,$7C,$7C,$9F,$B2,$FE,$0F,$E3
	.dc.b	$F1,$1F,$46,$2E,$7F,$72,$7F,$11,$F1,$FF,$89,$C6,$5F,$E0,$A3,$2D
	.dc.b	$21,$F6,$FE,$23,$F9,$7F,$D8,$BE,$3F,$F2,$5F,$07,$2B,$D0,$27,$4E
	.dc.b	$1A,$FC,$53,$53,$FB,$38,$CC,$7C,$96,$BF,$EB,$87,$8E,$5F,$5D,$BB
	.dc.b	$E6,$BF,$49,$F2,$1F,$A5,$7F,$4D,$F2,$13,$0F,$CF,$C6,$5A,$43,$F3
	.dc.b	$1F,$1D,$FA,$DF,$8E,$FD,$77,$C6,$C9,$FD,$3C,$65,$A4,$3F,$41,$F1
	.dc.b	$BF,$B5,$F8,$CF,$F0,$37,$69,$A4,$9F,$85,$DF,$FA,$D8,$E1,$1F,$B4
	.dc.b	$7F,$C3,$9F,$45,$F9,$5E,$0F,$E2,$74,$8F,$F5,$6D,$7E,$1E,$31,$F1
	.dc.b	$BC,$0E,$3D,$16,$75,$5C,$1F,$F6,$23,$6F,$AB,$F8,$F8,$DF,$5F,$F8
	.dc.b	$7E,$3F,$47,$1F,$EB,$46,$BF,$EB,$26,$BF,$C2,$1D,$CF,$49,$F5,$9E
	.dc.b	$CF,$89,$FE,$F3,$F1,$6F,$51,$4C,$3F,$BF,$78,$1F,$EF,$83,$F8,$5F
	.dc.b	$B8,$71,$E2,$FF,$BF,$23,$13,$C7,$0E,$97,$CE,$73,$29,$CA,$F9,$7C
	.dc.b	$E1,$FC,$61,$B8,$78,$E7,$53,$6B,$8D,$74,$6C,$79,$27,$D5,$3C,$9B
	.dc.b	$1C,$71,$F7,$C7,$43,$D0,$E2,$78,$46,$4F,$81,$7E,$38,$F0,$4E,$EB
	.dc.b	$A6,$3E,$F7,$C5,$7F,$28,$F9,$8F,$CA,$36,$E4,$F8,$93,$CD,$DD,$F7
	.dc.b	$9E,$77,$E0,$BF,$5D,$7A,$1A,$F9,$8B,$3F,$2C,$EF,$3C,$63,$C1,$1E
	.dc.b	$37,$B9,$E9,$3A,$1E,$5C,$3E,$6B,$A3,$E2,$47,$57,$73,$D4,$F9,$24
	.dc.b	$8F,$09,$F0,$31,$F9,$A4,$70,$6A,$7B,$4F,$17,$AC,$65,$F7,$1E,$2C
	.dc.b	$3E,$33,$A2,$1A,$9E,$4F,$CE,$C3,$E3,$E1,$F9,$2C,$3C,$53,$ED,$7E
	.dc.b	$31,$EB,$C3,$E8,$92,$3A,$CF,$18,$FB,$F3,$CA,$7B,$1F,$27,$1F,$C7
	.dc.b	$C6,$67,$E8,$4E,$A3,$EA,$7C,$57,$F9,$B6,$3E,$8B,$5E,$6D,$BF,$D4
	.dc.b	$B9,$8F,$DC,$7C,$37,$EE,$63,$11,$FE,$C0,$0E,$23,$55,$FD,$0D,$B3
	.dc.b	$FD,$A5,$FC,$13,$FC,$13,$C3,$FF,$19,$CF,$EE,$A3,$B9,$F9,$9C,$BF
	.dc.b	$C8,$7D,$9C,$23,$33,$FC,$02,$73,$1F,$9F,$33,$97,$FB,$4A,$FA,$BE
	.dc.b	$38,$FD,$84,$F0,$D3,$C1,$1E,$B7,$A5,$7F,$91,$9C,$E3,$E8,$AE,$F8
	.dc.b	$62,$FC,$D7,$D9,$F7,$1D,$9E,$58,$78,$70,$EB,$8F,$EB,$2B,$DF,$18
	.dc.b	$7F,$82,$18,$9F,$54,$EE,$3C,$D1,$87,$66,$0F,$F1,$0D,$78,$1F,$1D
	.dc.b	$F5,$3D,$05,$9F,$21,$CE,$FD,$87,$8C,$AE,$4D,$CF,$6B,$F2,$63,$F1
	.dc.b	$70,$F8,$53,$FC,$B6,$31,$3F,$66,$3C,$C7,$E0,$5E,$2F,$C7,$6D,$F8
	.dc.b	$6F,$BC,$EC,$F7,$D7,$47,$F0,$07,$E4,$F0,$F6,$3F,$4A,$B9,$36,$BD
	.dc.b	$4D,$3D,$47,$43,$C4,$58,$E2,$FA,$CC,$BF,$2A,$3E,$94,$D9,$E6,$FF
	.dc.b	$58,$E3,$FE,$F9,$F1,$BF,$E9,$0C,$FF,$B9,$B7,$2F,$ED,$9F,$17,$70
	.dc.b	$5F,$09,$FE,$FB,$5F,$E5,$B9,$FE,$E4,$D7,$7C,$68,$C5,$71,$9F,$EF
	.dc.b	$9F,$17,$D6,$3F,$C1,$CD,$A7,$C5,$D8,$98,$7F,$83,$1D,$46,$2F,$FC
	.dc.b	$1B,$F1,$7C,$4E,$5F,$17,$6F,$FC,$27,$1E,$2C,$0D,$24,$7F,$89,$A3
	.dc.b	$FD,$DE,$FF,$86,$BE,$0F,$C9,$F1,$7E,$56,$47,$F8,$7B,$E0,$EE,$3F
	.dc.b	$E1,$E7,$D5,$F1,$72,$7F,$C4,$7F,$0B,$6B,$FC,$47,$3E,$FF,$8B,$C7
	.dc.b	$17,$F8,$91,$FA,$12,$F9,$75,$61,$B8,$7F,$8B,$BE,$33,$13,$23,$FC
	.dc.b	$C1,$5C,$63,$EA,$36,$33,$5C,$A3,$FC,$CC,$DC,$3F,$C6,$BF,$17,$CE
	.dc.b	$7F,$C7,$2F,$D5,$F8,$C9,$7F,$E3,$EF,$7E,$D7,$F8,$FA,$7E,$E7,$E1
	.dc.b	$DF,$CC,$AB,$FC,$87,$18,$EE,$CF,$54,$9F,$F2,$65,$3E,$33,$18,$7F
	.dc.b	$8D,$A7,$FD,$8B,$2A,$CE,$7F,$E5,$68,$E1,$3C,$A7,$E4,$FC,$EB,$4A
	.dc.b	$7E,$6E,$DF,$08,$F5,$B3,$87,$FA,$32,$F8,$3F,$29,$B9,$7F,$A2,$DA
	.dc.b	$FF,$AC,$5A,$FF,$8E,$5D,$DF,$EB,$6D,$EA,$F6,$AF,$61,$FE,$52,$8C
	.dc.b	$DF,$C2,$F8,$DF,$92,$7C,$AF,$8D,$B1,$92,$57,$2E,$F6,$CA,$DC,$E6
	.dc.b	$1F,$0B,$F5,$0F,$AB,$E3,$5C,$3E,$8C,$7D,$8F,$97,$C5,$F1,$7E,$DF
	.dc.b	$93,$F7,$3E,$3A,$DF,$D8,$6D,$77,$6E,$FF,$55,$B3,$F6,$9B,$3F,$C3
	.dc.b	$32,$47,$E2,$3E,$3A,$57,$E1,$3E,$3A,$DF,$AE,$6E,$1F,$81,$9B,$5F
	.dc.b	$86,$F8,$EF,$C8,$9F,$92,$F8,$DB,$1F,$8E,$8B,$87,$C5,$9B,$7F,$EF
	.dc.b	$5F,$93,$7C,$0F,$EC,$9B,$87,$FB,$9E,$5E,$0F,$F0,$A8,$FF,$69,$CE
	.dc.b	$B0,$FF,$7B,$B5,$FF,$7B,$35,$FF,$3F,$35,$FF,$49,$35,$3C,$EB,$F9
	.dc.b	$B5,$3C,$8F,$26,$1F,$9B,$8C,$61,$FA,$AF,$8F,$FD,$89,$FB,$2F,$8F
	.dc.b	$B1,$FA,$EE,$8F,$E6,$7C,$3F,$A2,$8F,$F6,$2B,$74,$FF,$49,$1F,$56
	.dc.b	$77,$F1,$FF,$BC,$F8,$DF,$DE,$B6,$FF,$73,$F1,$BF,$C0,$0F,$E0,$3F
	.dc.b	$1B,$63,$FD,$91,$F8,$7F,$FA,$BD,$FF,$69,$B7,$1E,$FD,$88,$C4,$7F
	.dc.b	$A8,$E3,$0F,$F4,$A8,$7B,$DF,$D2,$57,$B3,$63,$0E,$7E,$36,$61,$E2
	.dc.b	$F8,$EB,$87,$EA,$70,$7C,$77,$52,$57,$99,$F0,$7B,$3D,$CF,$07,$C7
	.dc.b	$59,$F6,$FC,$6F,$8C,$F2,$73,$D8,$F8,$3F,$1B,$E9,$23,$F0,$9F,$E6
	.dc.b	$BF,$2E,$F8,$8A,$49,$FE,$77,$19,$7E,$4F,$97,$F0,$4F,$85,$E5,$58
	.dc.b	$F9,$FF,$1D,$F1,$C8,$FD,$07,$E5,$7C,$6F,$C9,$F8,$CF,$94,$D9,$FA
	.dc.b	$FF,$1B,$F3,$7E,$37,$E7,$36,$3E,$DC,$7D,$32,$3F,$71,$FA,$DF,$1B
	.dc.b	$F5,$4E,$1C,$9F,$3D,$FB,$1A,$CA,$FC,$2F,$C6,$DA,$FA,$93,$F7,$C9
	.dc.b	$3F,$88,$6C,$7D,$9F,$8D,$E1,$E3,$2B,$FD,$F9,$D9,$F8,$B9,$FF,$07
	.dc.b	$7A,$B9,$9F,$F8,$01,$5B,$BF,$F0,$EC,$E5,$FE,$58,$BE,$37,$8D,$43
	.dc.b	$F4,$B3,$C1,$FF,$0E,$CF,$18,$FC,$C3,$5F,$1A,$F8,$DF,$19,$E4,$F8
	.dc.b	$DB,$5E,$43,$71,$8A,$F1,$CD,$C3,$D1,$97,$F6,$3F,$8D,$F6,$FC,$4F
	.dc.b	$B9,$B8,$F3,$C3,$DB,$38,$7D,$B3,$6F,$E1,$43,$E4,$D8,$FE,$BA,$1F
	.dc.b	$1A,$F6,$91,$2B,$7E,$2F,$FA,$D2,$73,$2F,$13,$FB,$76,$99,$8E,$18
	.dc.b	$F5,$C7,$A4,$63,$0D,$3E,$39,$D6,$1D,$30,$E7,$3E,$48,$EF,$8F,$8A
	.dc.b	$30,$73,$29,$CE,$FA,$D7,$2C,$3C,$33,$93,$75,$CE,$32,$DB,$F1,$75
	.dc.b	$FE,$26,$73,$8F,$9A,$FC,$11,$8C,$7B,$D6,$EF,$CF,$7B,$BF,$01,$FC
	.dc.b	$C3,$0F,$53,$63,$E9,$36,$FF,$C7,$BF,$1D,$FD,$A3,$E5,$2F,$AE,$69
	.dc.b	$2B,$F2,$F8,$37,$FF,$72,$C6,$1A,$F8,$5A,$FC,$C6,$BF,$7D,$E3,$A7
	.dc.b	$C4,$6E,$7B,$F8,$0F,$B3,$7F,$E3,$1B,$B5,$A7,$94,$E0,$DF,$BF,$FE
	.dc.b	$B5,$BE,$53,$F4,$AB,$83,$F9,$56,$DF,$58,$D5,$F3,$69,$F6,$AF,$A3
	.dc.b	$E0,$33,$87,$A6,$7C,$31,$F9,$5E,$3F,$F2,$77,$C7,$7E,$6B,$8F,$F5
	.dc.b	$FC,$C8,$C7,$27,$EF,$1B,$7F,$C2,$BE,$3A,$E1,$FE,$57,$6D,$F1,$9F
	.dc.b	$DA,$4F,$66,$C2,$77,$F1,$DC,$3E,$3A,$56,$A3,$2D,$21,$FC,$63,$D8
	.dc.b	$F1,$BF,$B7,$DB,$FD,$06,$D3,$FD,$0B,$C3,$F2,$5F,$EF,$ED,$7F,$72
	.dc.b	$FC,$DD,$7F,$82,$B7,$5C,$D7,$C0,$E2,$FE,$F1,$D3,$E0,$A7,$C2,$E7
	.dc.b	$F3,$DB,$97,$F1,$3B,$46,$23,$FC,$C6,$7F,$0D,$9E,$47,$A2,$B5,$1E
	.dc.b	$68,$C1,$1F,$E3,$C7,$12,$C7,$5C,$BF,$8E,$F8,$DF,$DB,$67,$E7,$FC
	.dc.b	$6F,$D4,$23,$CB,$EB,$DF,$05,$F2,$9B,$1F,$EA,$C6,$BF,$D0,$9A,$FF
	.dc.b	$8A,$1A,$FF,$40,$F8,$FE,$53,$23,$EF,$C7,$E1,$3E,$3E,$47,$E0,$9F
	.dc.b	$C3,$FC,$6F,$87,$B7,$EC,$E1,$FE,$66,$AF,$14,$60,$FF,$5F,$F4,$7F
	.dc.b	$A9,$7E,$3F,$F1,$E4,$9F,$C6,$FC,$0F,$E4,$BE,$3E,$47,$E4,$1F,$CA
	.dc.b	$F2,$FE,$23,$E5,$BF,$C9,$D3,$FE,$CA,$6B,$FD,$81,$AF,$F9,$19,$AF
	.dc.b	$F7,$0F,$95,$FF,$5D,$3F,$9E,$6B,$FE,$97,$6F,$17,$EB,$39,$3F,$4B
	.dc.b	$F2,$12,$BF,$46,$FE,$AB,$F1,$1F,$E2,$B5,$FD,$5B,$C7,$7C,$0F,$E7
	.dc.b	$1B,$7F,$EE,$06,$BF,$DD,$9A,$FF,$92,$EB,$FD,$94,$FE,$C9,$AF,$FA
	.dc.b	$AD,$B5,$FD,$B9,$F3,$F6,$7F,$83,$1F,$E8,$F1,$9F,$8F,$F3,$1E,$7E
	.dc.b	$AF,$D8,$FE,$49,$FE,$8A,$C3,$FD,$E6,$D7,$FC,$12,$D7,$FC,$DE,$D7
	.dc.b	$B7,$C7,$FC,$12,$3E,$7F,$90,$F8,$9F,$21,$F1,$7B,$24,$FA,$E3,$F8
	.dc.b	$0F,$E0,$9F,$E1,$98,$FF,$0B,$87,$0F,$F8,$69,$AF,$FA,$01,$AF,$BF
	.dc.b	$A7,$E9,$11,$F9,$1F,$23,$FC,$3B,$17,$67,$F1,$0F,$99,$7C,$95,$35
	.dc.b	$12,$FC,$75,$C1,$B1,$FC,$5A,$31,$1F,$E2,$E6,$F0,$7F,$85,$9B,$C5
	.dc.b	$FE,$0E,$6C,$FF,$85,$30,$FD,$33,$F6,$F0,$FC,$51,$1F,$C0,$7C,$95
	.dc.b	$74,$7E,$0E,$3F,$21,$1F,$91,$F9,$29,$3F,$87,$7F,$29,$F2,$38,$8F
	.dc.b	$C4,$45,$CB,$F1,$13,$F9,$9F,$91,$95,$F8,$F7,$F3,$9F,$21,$88,$FC
	.dc.b	$84,$5A,$FC,$84,$DD,$3F,$21,$36,$9E,$13,$A1,$7F,$A7,$EA,$FD,$47
	.dc.b	$C8,$4B,$FC,$F4,$65,$A4,$3F,$41,$F2,$1F,$AF,$F9,$09,$1F,$A5,$7F
	.dc.b	$AD,$FC,$BF,$FB,$13,$5F,$D2,$46,$A7,$F5,$0E,$9F,$EA,$3F,$82,$23
	.dc.b	$FA,$9F,$5B,$CA,$A6,$A2,$57,$E5,$AB,$83,$63,$FA,$CC,$62,$3F,$CE
	.dc.b	$2D,$BF,$F2,$CB,$78,$3F,$CA,$2D,$9F,$F2,$BE,$1F,$C6,$AB,$FC,$D3
	.dc.b	$8F,$F8,$B1,$BE,$DF,$B8,$F7,$FC,$95,$BF,$EE,$F5,$BC,$7F,$D1,$AD
	.dc.b	$8F,$A6,$FF,$B4,$9F,$EA,$7F,$83,$3E,$3F,$E4,$FC,$04,$9C,$C7,$F7
	.dc.b	$69,$D4,$60,$83,$2F,$3F,$C9,$FF,$7C,$F9,$B7,$A7,$FE,$10,$97,$DE
	.dc.b	$3E,$3B,$0D,$C3,$FB,$2D,$F0,$6D,$FF,$81,$23,$11,$FE,$9C,$6B,$FE
	.dc.b	$7A,$0E,$1F,$F3,$EB,$C2,$FF,$D1,$2D,$D7,$35,$F8,$8F,$94,$FC,$4F
	.dc.b	$CA,$7E,$2B,$59,$8F,$CC,$8C,$B4,$87,$C1,$BF,$9F,$F2,$9F,$92,$F9
	.dc.b	$4F,$C9,$FC,$A4,$AF,$AB,$19,$69,$0E,$47,$D6,$F8,$9F,$CD,$EF,$FD
	.dc.b	$A3,$8F,$E3,$4F,$BF,$BD,$18,$8F,$F4,$DC,$62,$78,$1C,$A3,$F0,$AF
	.dc.b	$E9,$B6,$FD,$3F,$C9,$7E,$A3,$E4,$E5,$7E,$2A,$32,$D2,$1E,$6B,$F0
	.dc.b	$FC,$9F,$EC,$3E,$4F,$CC,$49,$FF,$68,$CF,$9A,$31,$1C,$72,$47,$18
	.dc.b	$FC,$CF,$C9,$FF,$91,$3E,$3B,$FC,$3B,$CD,$FE,$BA,$F8,$97,$27,$FA
	.dc.b	$E0,$FC,$D5,$A3,$2F,$FB,$02,$33,$18,$85,$74,$78,$4F,$FB,$0A,$31
	.dc.b	$8E,$E3,$AB,$7B,$3F,$D8,$EF,$39,$DC,$F1,$F4,$FF,$45,$7F,$52,$3C
	.dc.b	$D7,$C1,$B1,$FE,$D4,$BE,$55,$C1,$B8,$E6,$B7,$5C,$CF,$65,$F3,$7F
	.dc.b	$DB,$B8,$78,$9B,$87,$8B,$1F,$44,$7C,$9B,$E7,$1D,$A3,$E0,$9C,$30
	.dc.b	$F0,$3D,$8A,$7D,$0C,$3C,$0F,$CA,$BF,$45,$FC,$1A,$DB,$F2,$0F,$4E
	.dc.b	$3D,$CF,$06,$1E,$4B,$F0,$9D,$4F,$31,$E3,$33,$8F,$B5,$D1,$F0,$CB
	.dc.b	$3D,$6B,$C5,$19,$7C,$4F,$AC,$E1,$3C,$CF,$C1,$57,$8C,$93,$C9,$E9
	.dc.b	$3C,$6B,$D0,$61,$F8,$57,$F0,$31,$F8,$B7,$E0,$9C,$9E,$0C,$7E,$3B
	.dc.b	$CE,$3E,$29,$E8,$BF,$8A,$FD,$62,$47,$C3,$C3,$C1,$1D,$4E,$D8,$FA
	.dc.b	$EF,$F0,$11,$C2,$BE,$3D,$74,$79,$46,$4F,$D2,$9E,$F3,$E4,$FC,$5F
	.dc.b	$F9,$3A,$FF,$4C,$DB,$FC,$EB,$72,$5A,$3F,$CA,$AE,$3E,$33,$31,$88
	.dc.b	$46,$31,$FE,$71,$18,$39,$38,$8D,$47,$F3,$D7,$F0,$AF,$F4,$FF,$05
	.dc.b	$D3,$88,$F0,$1F,$8C,$C1,$3A,$8F,$C7,$9A,$BF,$E0,$93,$AA,$E5,$3E
	.dc.b	$39,$CD,$71,$7F,$85,$47,$6D,$33,$5F,$A6,$8C,$69,$FE,$F3,$BF,$05
	.dc.b	$70,$9E,$93,$9A,$EF,$8F,$B3,$1F,$15,$79,$8F,$89,$7D,$9F,$D0,$1D
	.dc.b	$8F,$A3,$1C,$23,$B4,$FE,$E2,$31,$7F,$2E,$BF,$51,$18,$7E,$61,$8C
	.dc.b	$3C,$F3,$BB,$F0,$1D,$27,$E6,$4F,$C3,$9E,$98,$F9,$4D,$57,$08,$E4
	.dc.b	$7B,$EB,$D2,$7B,$1C,$47,$06,$9D,$0E,$13,$E1,$3B,$3D,$27,$E0,$D7
	.dc.b	$A1,$F5,$D7,$85,$B3,$D1,$A7,$0C,$7E,$C3,$F5,$0C,$1F,$47,$1F,$A9
	.dc.b	$7E,$F8,$E2,$7E,$46,$78,$CF,$CF,$9F,$14,$FE,$90,$F7,$1D,$2F,$DC
	.dc.b	$DB,$F5,$3C,$CF,$8B,$5C,$EF,$AC,$7D,$07,$F0,$ED,$3F,$9C,$F0,$7F
	.dc.b	$80,$BE,$4B,$F3,$D9,$FF,$92,$1B,$5F,$E4,$2F,$93,$B7,$FE,$42,$9B
	.dc.b	$7C,$6B,$FB,$BE,$49,$FD,$45,$7F,$83,$23,$13,$A9,$CC,$FF,$93,$BE
	.dc.b	$4F,$9D,$1E,$BF,$89,$FE,$DA,$FF,$CA,$B1,$A3,$3F,$29,$63,$F3,$67
	.dc.b	$F0,$CC,$54,$9E,$7F,$29,$2B,$78,$FF,$A3,$A9,$EF,$E6,$3F,$C2,$D0
	.dc.b	$AF,$3C,$6D,$B8,$76,$6C,$EE,$FF,$C3,$2B,$E5,$3D,$67,$B3,$E3,$AC
	.dc.b	$F8,$AF,$CF,$3F,$03,$E4,$F3,$8F,$89,$F8,$7F,$27,$2F,$86,$2F,$92
	.dc.b	$C5,$FA,$2B,$E4,$FC,$9E,$38,$FF,$8D,$67,$F1,$23,$ED,$D3,$E6,$5F
	.dc.b	$1B,$86,$E7,$DB,$E5,$31,$32,$3F,$DD,$F3,$AA,$FE,$66,$D7,$FC,$D8
	.dc.b	$D7,$F3,$8E,$F1,$FC,$7F,$DE,$FD,$0A,$FA,$75,$E3,$BF,$C7,$D2,$B5
	.dc.b	$3F,$E5,$E1,$FB,$4E,$27,$F1,$ED,$BF,$C1,$6C,$F9,$4F,$7C,$E6,$FF
	.dc.b	$A8,$B7,$5F,$C3,$FA,$FA,$E5,$FC,$8B,$F9,$2F,$94,$99,$7D,$0F,$85
	.dc.b	$B8,$FE,$1A,$6E,$9F,$3B,$45,$62,$FF,$78,$3E,$1F,$E8,$B4,$FC,$9D
	.dc.b	$3F,$01,$FC,$C4,$7C,$43,$58,$FC,$16,$BF,$E9,$96,$BD,$AB,$F2,$EF
	.dc.b	$F5,$E6,$C7,$2A,$FD,$FD,$7E,$5E,$BA,$57,$D0,$8C,$4F,$AF,$E5,$3F
	.dc.b	$66,$7E,$D3,$86,$E1,$F9,$28,$B8,$7E,$A2,$6D,$7C,$3F,$94,$FD,$D9
	.dc.b	$FB,$CF,$94,$B8,$7E,$6A,$3F,$7F,$F2,$58,$C3,$F6,$2F,$F0,$2F,$92
	.dc.b	$95,$FA,$7E,$FF,$F0,$43,$FA,$28,$CD,$7E,$E1,$FE,$2B,$3A,$8C,$4E
	.dc.b	$C8,$FF,$A4,$E3,$58,$FD,$9A,$E3,$E3,$CB,$F0,$7C,$ED,$C3,$84,$DB
	.dc.b	$FD,$6E,$0A,$C5,$FF,$13,$87,$C9,$79,$0F,$2F,$C9,$D9,$F6,$45,$C7
	.dc.b	$C9,$37,$0F,$E5,$7F,$3A,$F5,$3F,$72,$DC,$3F,$13,$C0,$E1,$CD,$49
	.dc.b	$FF,$13,$E2,$BE,$0F,$F8,$A9,$AF,$FB,$8D,$AF,$F5,$37,$78,$7F,$46
	.dc.b	$C3,$F7,$55,$9B,$FD,$BC,$62,$DD,$3F,$CC,$7E,$43,$E9,$9F,$53,$CD
	.dc.b	$B8,$7F,$00,$E0,$78,$9F,$98,$6C,$7A,$1B,$87,$F9,$01,$AF,$FB,$63
	.dc.b	$81,$F2,$7F,$80,$F4,$3F,$02,$DB,$FC,$47,$4F,$E1,$0F,$C2,$FC,$95
	.dc.b	$8F,$17,$37,$84,$8F,$85,$FE,$7F,$83,$E4,$7F,$1B,$F2,$3F,$8E,$6D
	.dc.b	$7E,$57,$E4,$7F,$22,$7E,$4B,$E4,$6C,$7F,$1E,$FB,$D5,$FE,$D9,$B1
	.dc.b	$F5,$9B,$B7,$F9,$79,$AF,$BF,$E4,$BF,$3A,$7E,$7B,$E4,$6D,$7F,$4B
	.dc.b	$F9,$D7,$C2,$7F,$76,$6C,$7E,$A7,$E4,$71,$3F,$D0,$BE,$77,$FB,$AE
	.dc.b	$3F,$E6,$C6,$BF,$E1,$30,$E1,$FF,$0B,$35,$FB,$FB,$BE,$4F,$F6,$27
	.dc.b	$EC,$BE,$4E,$E3,$FB,$9F,$93,$FD,$A9,$FE,$0A,$FC,$57,$FD,$07,$19
	.dc.b	$3F,$A1,$FB,$29,$FE,$36,$D8,$FC,$DB,$71,$FF,$41,$B5,$FF,$15,$77
	.dc.b	$2D,$F2,$7D,$CF,$07,$C1,$DC,$3F,$B9,$4B,$E4,$FF,$B2,$FE,$37,$E4
	.dc.b	$C5,$E2,$FF,$6A,$AD,$5F,$F5,$8A,$D4,$FF,$88,$1E,$15,$F2,$5A,$FF
	.dc.b	$8F,$5A,$FF,$A8,$9B,$97,$9F,$E4,$FE,$21,$E2,$9B,$7D,$F8,$BC,$A7
	.dc.b	$F7,$AF,$A6,$FE,$33,$1D,$63,$2F,$F0,$0C,$3F,$03,$5F,$EA,$86,$BE
	.dc.b	$59,$FE,$8F,$1A,$C7,$FD,$41,$F3,$EF,$CD,$BF,$8E,$61,$E1,$3F,$C5
	.dc.b	$31,$9F,$3B,$E3,$FC,$84,$9F,$EF,$79,$FF,$20,$7E,$34,$67,$3F,$13
	.dc.b	$E0,$24,$7F,$AF,$35,$FB,$1F,$29,$F8,$53,$F0,$DF,$13,$67,$EB,$C7
	.dc.b	$E2,$7E,$53,$18,$7D,$69,$FF,$21,$70,$A7,$F1,$DF,$29,$F8,$FF,$94
	.dc.b	$B5,$F8,$07,$0D,$3F,$07,$18,$BF,$F0,$87,$D0,$FF,$91,$B9,$FF,$DA
	.dc.b	$DE,$AF,$F9,$DE,$31,$3F,$ED,$68,$FF,$3D,$73,$B7,$FF,$65,$7C,$6F
	.dc.b	$F4,$2A,$FF,$70,$DF,$08,$C4,$23,$17,$D7,$DF,$FC,$CC,$F5,$3F,$DD
	.dc.b	$38,$75,$CF,$C7,$1D,$27,$78,$74,$C3,$9C,$E7,$2D,$BF,$9F,$CB,$CB
	.dc.b	$97,$79,$CD,$70,$AF,$26,$1D,$2B,$DB,$7C,$27,$A1,$E5,$37,$7C,$AB
	.dc.b	$84,$7A,$27,$83,$A3,$85,$FC,$A7,$F8,$83,$5F,$21,$FF,$00,$1C,$21
	.dc.b	$F1,$FF,$E2,$DE,$5F,$EB,$ED,$8F,$E3,$46,$BC,$D4,$E9,$B1,$D7,$E2
	.dc.b	$5F,$1F,$FD,$A3,$0F,$F3,$D3,$63,$33,$B8,$C1,$F2,$23,$3A,$FD,$16
	.dc.b	$EB,$F8,$7F,$77,$FD,$7D,$39,$AE,$04,$BF,$87,$3A,$AF,$C7,$5F,$29
	.dc.b	$FC,$4B,$5F,$C4,$37,$05,$71,$9D,$DF,$5C,$3C,$11,$D8,$C6,$1D,$A6
	.dc.b	$C7,$E5,$AB,$97,$3F,$F2,$BE,$DF,$F6,$B7,$67,$F2,$3D,$EF,$17,$EC
	.dc.b	$1B,$87,$F8,$0B,$9F,$FA,$9B,$FC,$BA,$F3,$87,$F3,$5A,$B5,$FA,$7A
	.dc.b	$D1,$FC,$D3,$E5,$73,$F2,$B2,$30,$43,$F8,$01,$DD,$D3,$8B,$FE,$85
	.dc.b	$3D,$BE,$57,$FC,$47,$1F,$D1,$5B,$66,$92,$BF,$A2,$BC,$5F,$E8,$FE
	.dc.b	$DF,$A3,$E5,$7D,$3F,$2B,$27,$FA,$5C,$65,$FE,$39,$68,$EE,$48,$CB
	.dc.b	$B7,$D1,$1B,$7F,$AA,$FC,$AF,$C6,$F9,$49,$1F,$D6,$1F,$93,$F2,$B2
	.dc.b	$3B,$BE,$C2,$3E,$2F,$94,$D3,$DA,$3E,$8C,$7D,$4F,$C6,$3F,$DF,$3C
	.dc.b	$7F,$EC,$86,$F1,$FF,$BB,$5B,$7F,$E3,$E7,$FD,$DC,$FF,$B6,$DF,$F7
	.dc.b	$43,$FC,$ED,$AF,$F8,$2F,$74,$62,$3F,$BB,$B5,$ED,$DB,$E3,$EA,$91
	.dc.b	$D6,$3F,$1B,$83,$F4,$A7,$26,$1B,$5F,$EA,$C6,$BF,$86,$9E,$1C,$7F
	.dc.b	$92,$F7,$DF,$11,$F8,$46,$D7,$F5,$36,$BE,$96,$BF,$E2,$5F,$A1,$7C
	.dc.b	$9F,$F9,$BA,$11,$87,$FC,$E2,$D7,$F8,$9B,$5F,$83,$F2,$FF,$44,$8F
	.dc.b	$CB,$F8,$4F,$D3,$FD,$0F,$F4,$68,$FF,$0E,$B5,$FE,$45,$C2,$8C,$47
	.dc.b	$F9,$C9,$B1,$9F,$87,$E6,$49,$FB,$1F,$2F,$F8,$1F,$98,$91,$F6,$DF
	.dc.b	$C2,$7C,$C7,$E1,$74,$FD,$4F,$2F,$A6,$BF,$09,$18,$33,$F2,$DF,$8B
	.dc.b	$F9,$59,$3F,$85,$8F,$C7,$7C,$AC,$8F,$C4,$3F,$91,$F9,$4F,$C9,$7C
	.dc.b	$8F,$62,$4F,$E2,$DF,$0F,$C8,$7E,$5F,$96,$4F,$E4,$63,$F8,$8A,$DD
	.dc.b	$B6,$3E,$7D,$F0,$F4,$DC,$EA,$BF,$A6,$74,$BB,$BF,$A7,$47,$3C,$3E
	.dc.b	$F5,$6E,$3F,$A6,$B5,$FF,$34,$35,$E2,$EE,$BA,$47,$F5,$98,$C5,$6A
	.dc.b	$BC,$DF,$15,$FB,$3F,$2E,$5F,$EB,$5F,$DB,$7C,$A6,$23,$F8,$F7,$27
	.dc.b	$F4,$07,$47,$EC,$E3,$F7,$9F,$1D,$FC,$C3,$F1,$AF,$EB,$33,$FC,$95
	.dc.b	$BA,$7E,$FE,$77,$3F,$BB,$6B,$BE,$EE,$5E,$F7,$3F,$46,$4F,$C8,$DB
	.dc.b	$F2,$78,$FF,$B8,$E3,$07,$FB,$11,$C3,$FC,$0D,$AF,$F9,$E1,$AF,$2A
	.dc.b	$FE,$6F,$1C,$31,$F9,$3D,$88,$C5,$EE,$FB,$CE,$7E,$57,$E0,$70,$FC
	.dc.b	$1F,$94,$95,$AC,$3F,$0C,$D9,$DB,$4D,$7C,$9F,$C9,$F9,$4F,$95,$EC
	.dc.b	$49,$E5,$19,$7A,$46,$5F,$9D,$1F,$94,$F9,$F3,$F9,$6C,$5C,$34,$D4
	.dc.b	$4A,$FF,$7C,$D7,$06,$C7,$F0,$F8,$C4,$7F,$87,$9B,$B7,$E3,$A1,$18
	.dc.b	$DB,$F3,$6D,$8F,$F1,$56,$0E,$2F,$C3,$7C,$AF,$E1,$FE,$56,$57,$C7
	.dc.b	$8C,$B4,$87,$E2,$FE,$53,$F1,$DF,$2B,$F8,$FE,$99,$3F,$3E,$32,$D2
	.dc.b	$1F,$93,$EB,$FD,$CF,$D4,$7F,$73,$8F,$AD,$F2,$BF,$99,$33,$E6,$C8
	.dc.b	$FF,$1E,$89,$E0,$FF,$90,$1A,$FF,$B6,$DA,$FE,$F1,$DF,$07,$C7,$6F
	.dc.b	$06,$6B,$F5,$1F,$29,$FA,$9F,$92,$FD,$57,$C9,$4C,$7F,$15,$19,$69
	.dc.b	$0F,$D7,$7C,$97,$F8,$83,$E4,$7F,$DD,$99,$FE,$69,$FF,$3E,$CF,$F8
	.dc.b	$5D,$B1,$98,$E0,$43,$F3,$B1,$CB,$2F,$E8,$28,$ED,$F1,$1F,$E1,$E7
	.dc.b	$FD,$B9,$3F,$EA,$D6,$BF,$DC,$DA,$FF,$BA,$1C,$D7,$96,$79,$CF,$83
	.dc.b	$E4,$7D,$5F,$1B,$EB,$F8,$E9,$3E,$48,$CB,$E6,$8C,$BE,$D8,$F8,$5F
	.dc.b	$1D,$E5,$24,$79,$5A,$FC,$37,$6F,$C0,$F8,$DF,$E7,$BF,$19,$FA,$AC
	.dc.b	$BF,$D0,$5E,$3B,$1F,$E6,$E7,$C3,$BC,$B8,$D3,$F9,$DC,$EE,$7F,$D1
	.dc.b	$57,$CD,$46,$5F,$F4,$6F,$73,$0F,$F4,$87,$CE,$E7,$1E,$D4,$8C,$63
	.dc.b	$FD,$21,$E5,$8F,$56,$E1,$CE,$B8,$5F,$95,$EA,$D7,$27,$9A,$38,$4F
	.dc.b	$2C,$37,$3F,$EA,$3A,$F1,$CF,$8A,$32,$7A,$A6,$DF,$BF,$1D,$4F,$C0
	.dc.b	$BF,$04,$75,$8C,$1E,$1A,$E7,$3D,$8F,$35,$EC,$8E,$09,$3C,$4F,$59
	.dc.b	$E3,$3A,$9F,$67,$1F,$AB,$1E,$23,$FB,$8E,$1E,$76,$E3,$F0,$9E,$B5
	.dc.b	$C4,$F6,$C6,$5F,$05,$FB,$89,$3C,$B0,$D9,$F8,$69,$F9,$58,$78,$9A
	.dc.b	$64,$F4,$63,$EE,$72,$7B,$2B,$E2,$3E,$07,$B3,$99,$FC,$24,$7A,$31
	.dc.b	$FC,$23,$AC,$3C,$B5,$D1,$EC,$F8,$23,$E2,$3E,$0B,$FA,$C6,$A3,$C3
	.dc.b	$1F,$0C,$F9,$F5,$D9,$FA,$53,$F8,$02,$4F,$99,$E6,$F6,$8E,$51,$87
	.dc.b	$ED,$7C,$57,$E9,$9B,$1E,$E8,$C4,$7D,$D6,$BF,$EF,$7F,$8B,$FF,$7C
	.dc.b	$46,$21,$18,$9F,$F7,$BB,$63,$FC,$62,$68,$FF,$7A,$C6,$9F,$C4,$E4
	.dc.b	$F0,$3F,$C5,$B3,$FB,$99,$F0,$1A,$8C,$E1,$C6,$7F,$7D,$5C,$78,$BE
	.dc.b	$F5,$F0,$9C,$DF,$08,$F1,$E1,$FA,$CB,$E1,$1F,$9C,$3B,$E7,$C3,$0F
	.dc.b	$D8,$C7,$3C,$7F,$DD,$38,$73,$8F,$F2,$6E,$1F,$8D,$C3,$9E,$3E,$99
	.dc.b	$F8,$18,$F3,$7B,$18,$C3,$9E,$3D,$A3,$07,$F9,$6B,$1F,$F2,$7D,$73
	.dc.b	$C7,$D1,$18,$BF,$95,$3D,$4D,$57,$ED,$B1,$F6,$C6,$31,$F1,$16,$BC
	.dc.b	$35,$FB,$A3,$E0,$B8,$9E,$67,$57,$B9,$92,$C7,$9F,$1F,$56,$3F,$0F
	.dc.b	$1F,$8D,$3F,$24,$D1,$DC,$F7,$9F,$CD,$5F,$63,$F5,$8F,$53,$87,$E7
	.dc.b	$98,$9E,$ED,$3E,$6C,$F8,$63,$F0,$71,$F9,$1C,$3E,$0D,$FA,$23,$C1
	.dc.b	$3C,$B0,$F7,$47,$85,$AF,$81,$AF,$C6,$3D,$6F,$C6,$9F,$89,$1E,$9A
	.dc.b	$EE,$D8,$F9,$F1,$FD,$67,$C8,$FF,$00,$7C,$7F,$FA,$33,$3F,$F0,$63
	.dc.b	$71,$FF,$02,$FC,$65,$C5,$7C,$27,$FC,$17,$5F,$E4,$79,$FF,$05,$B5
	.dc.b	$DE,$E8,$C5,$71,$9F,$F0,$57,$C6,$75,$8F,$F0,$97,$C6,$5B,$4D,$8C
	.dc.b	$37,$0F,$F0,$DB,$6F,$FC,$45,$F0,$5C,$4E,$5D,$B6,$BF,$C4,$31,$E2
	.dc.b	$C3,$FE,$00,$C7,$F8,$76,$30,$67,$E3,$3C,$47,$8F,$E2,$EC,$FF,$8A
	.dc.b	$A3,$CD,$F1,$98,$C7,$FC,$64,$FA,$7E,$2E,$57,$F8,$C3,$E2,$ED,$7F
	.dc.b	$8D,$E7,$DD,$F1,$78,$E2,$FF,$1E,$3F,$3E,$5F,$4E,$F7,$70,$DC,$7F
	.dc.b	$C7,$7F,$19,$89,$91,$FE,$5C,$9D,$57,$F2,$23,$9C,$FD,$56,$CE,$E3
	.dc.b	$E9,$7E,$37,$FE,$D9,$8F,$F5,$8C,$E8,$FF,$0D,$27,$F0,$71,$B7,$26
	.dc.b	$3E,$5B,$99,$D3,$E5,$AD,$7F,$AB,$A9,$3E,$0F,$8E,$CD,$F2,$E7,$52
	.dc.b	$7C,$BF,$1F,$30,$E5,$F1,$B7,$0E,$16,$AF,$5F,$C6,$E7,$83,$FD,$83
	.dc.b	$28,$F9,$DF,$35,$EB,$C7,$DE,$DC,$3D,$76,$F9,$B7,$C4,$F0,$99,$7F
	.dc.b	$E8,$AB,$E1,$1F,$D4,$1A,$FC,$70,$E1,$F9,$4F,$7B,$FD,$95,$BE,$27
	.dc.b	$AD,$78,$A3,$17,$F2,$7E,$3B,$EB,$FC,$3F,$D8,$6C,$64,$F9,$D1,$96
	.dc.b	$9F,$47,$E3,$BE,$F9,$F8,$0F,$8E,$B1,$93,$EB,$7C,$6E,$31,$F9,$AF
	.dc.b	$E1,$BE,$3A,$4F,$DB,$DA,$D7,$D2,$9F,$DF,$F0,$1A,$4A,$FC,$A7,$27
	.dc.b	$DE,$8F,$C8,$ED,$CB,$7F,$AE,$FE,$53,$9E,$5F,$E2,$7E,$36,$E5,$F8
	.dc.b	$09,$FC,$CF,$C7,$63,$0F,$BE,$FE,$73,$E3,$66,$1F,$91,$F8,$3B,$87
	.dc.b	$E2,$26,$DF,$E1,$A6,$D7,$D9,$E5,$76,$FE,$20,$FE,$49,$4F,$C1,$DE
	.dc.b	$8E,$6F,$F1,$75,$FA,$FB,$E3,$1F,$87,$8C,$11,$FD,$65,$FF,$33,$BF
	.dc.b	$F4,$5C,$72,$32,$69,$CE,$5F,$A0,$F8,$6F,$E5,$5F,$8E,$78,$AB,$F5
	.dc.b	$D1,$8C,$B3,$E7,$FF,$32,$AE,$37,$F9,$5E,$34,$62,$3F,$7D,$18,$9E
	.dc.b	$1F,$1B,$D8,$EF,$F1,$D7,$2F,$2D,$FF,$46,$C1,$F1,$DE,$4F,$8D,$F2
	.dc.b	$B6,$B7,$97,$83,$D7,$F2,$12,$3F,$0B,$7F,$E2,$6C,$3C,$CB,$E3,$7D
	.dc.b	$E7,$C0,$F6,$AD,$74,$F9,$C7,$BD,$FD,$35,$B1,$E0,$B7,$C7,$66,$B5
	.dc.b	$ED,$7F,$56,$BF,$F6,$25,$FE,$59,$B9,$7F,$93,$34,$46,$23,$CB,$F1
	.dc.b	$9F,$48,$FA,$7F,$1B,$6F,$F9,$5F,$CE,$3D,$3F,$E6,$4D,$8F,$E6,$D6
	.dc.b	$8C,$3F,$CE,$5A,$FD,$7F,$3D,$F0,$DF,$DB,$1B,$5F,$AC,$6C,$F9,$DA
	.dc.b	$FC,$AF,$8E,$FC,$39,$F8,$8F,$8D,$B5,$F1,$7E,$3B,$F1,$7F,$1D,$F8
	.dc.b	$C6,$C7,$C9,$F8,$EF,$C5,$11,$FA,$EF,$E3,$7E,$37,$F2,$67,$E5,$33
	.dc.b	$B3,$F4,$7E,$37,$F2,$FF,$1B,$F9,$86,$C7,$D5,$8F,$CB,$11,$FC,$1B
	.dc.b	$FE,$0A,$FC,$77,$F9,$44,$7E,$59,$BA,$FE,$46,$7F,$C1,$91,$8E,$42
	.dc.b	$F9,$AC,$9A,$F8,$3B,$87,$E5,$27,$4F,$F1,$66,$E1,$F8,$16,$BF,$A9
	.dc.b	$6B,$E3,$AF,$0F,$C7,$F4,$8F,$EB,$92,$8F,$2F,$C7,$CB,$F0,$75,$5B
	.dc.b	$DC,$FB,$C9,$3E,$37,$E1,$75,$7B,$BE,$3F,$DE,$D9,$F5,$36,$BC,$2D
	.dc.b	$F2,$53,$F2,$7E,$3F,$E3,$FC,$5F,$C8,$6C,$7C,$36,$DF,$A9,$B5,$CA
	.dc.b	$3F,$C0,$9F,$33,$FD,$DF,$6F,$F5,$C7,$8F,$FD,$86,$31,$3F,$EB,$88
	.dc.b	$FE,$C9,$B7,$F8,$FB,$1D,$CF,$2F,$81,$FA,$CE,$CC,$61,$A5,$87,$4F
	.dc.b	$8F,$63,$D3,$E3,$3C,$91,$8A,$E2,$73,$38,$E1,$DA,$35,$18,$C3,$B6
	.dc.b	$5E,$58,$E1,$8F,$7C,$7D,$98,$71,$6D,$F0,$C7,$C3,$18,$BE,$35,$B7
	.dc.b	$C7,$8F,$83,$1E,$CF,$8B,$1E,$D3,$E8,$3F,$DB,$6F,$CA,$6D,$7C,$66
	.dc.b	$E3,$FC,$2B,$E4,$3F,$8E,$FC,$D5,$F1,$8D,$25,$7F,$91,$70,$FF,$14
	.dc.b	$79,$FF,$C4,$9B,$19,$F8,$4F,$EB,$AD,$9D,$46,$33,$F4,$37,$5F,$C0
	.dc.b	$36,$BF,$0B,$EB,$7F,$B6,$27,$37,$F6,$F0,$EF,$87,$28,$C5,$6E,$FF
	.dc.b	$1B,$87,$37,$F0,$2D,$BE,$38,$7E,$3E,$B8,$56,$AF,$B5,$EC,$E1,$3A
	.dc.b	$BE,$35,$DA,$B6,$78,$9F,$CA,$F5,$7F,$0A,$F9,$0F,$F1,$E7,$1F,$E8
	.dc.b	$63,$13,$FC,$26,$73,$C7,$FC,$57,$8E,$CF,$FA,$AB,$AB,$FD,$3A,$EC
	.dc.b	$87,$F0,$68,$DD,$FF,$18,$F9,$0B,$46,$98,$F8,$FB,$7F,$C6,$62,$BF
	.dc.b	$C4,$23,$8F,$C8,$7F,$4E,$BF,$E4,$6D,$B3,$49,$5F,$C8,$DE,$2F,$F2
	.dc.b	$5F,$8F,$F1,$FC,$87,$93,$E4,$24,$FF,$29,$8C,$BF,$CA,$5E,$C4,$70
	.dc.b	$6C,$87,$F2,$6F,$90,$F7,$F9,$D2,$3F,$9A,$3F,$0B,$E4,$24,$7F,$35
	.dc.b	$7D,$44,$7F,$9C,$7B,$BA,$7F,$98,$C7,$CA,$8F,$9B,$ED,$7F,$19,$9F
	.dc.b	$F4,$83,$FE,$94,$7F,$C2,$AD,$73,$C3,$C3,$E3,$FE,$54,$9A,$49,$FF
	.dc.b	$2E,$C7,$FA,$65,$E7,$DF,$E0,$3C,$3D,$3E,$2F,$8F,$F1,$FC,$7C,$AF
	.dc.b	$0C,$79,$BE,$3B,$B9,$23,$AB,$E2,$F8,$EF,$59,$23,$D3,$F1,$DF,$90
	.dc.b	$5F,$30,$F8,$FF,$F0,$7C,$61,$B1,$98,$FF,$09,$B5,$FF,$4D,$0E,$DF
	.dc.b	$F6,$A7,$C7,$FF,$86,$1B,$1F,$EC,$F6,$BC,$1A,$FF,$93,$F6,$E9,$1F
	.dc.b	$A9,$6F,$07,$F9,$41,$B2,$EA,$42,$FE,$CF,$C6,$FD,$AF,$8F,$97,$F6
	.dc.b	$23,$EE,$FC,$77,$D6,$24,$7D,$47,$EC,$FC,$6F,$E0,$89,$1F,$80,$F8
	.dc.b	$EF,$C2,$FC,$74,$8F,$C1,$BF,$88,$F1,$FF,$13,$F1,$DF,$82,$24,$FE
	.dc.b	$01,$FC,$2F,$C6,$7E,$38,$91,$FE,$60,$F9,$EF,$E5,$6F,$FB,$7D,$7C
	.dc.b	$6F,$F0,$07,$DF,$1B,$DF,$13,$23,$FD,$53,$3C,$1B,$1F,$BD,$8C,$47
	.dc.b	$F9,$5D,$BC,$5F,$C5,$DB,$C5,$FE,$7E,$6C,$FC,$F7,$FB,$1C,$EF,$E3
	.dc.b	$3C,$9F,$1B,$2B,$D5,$1E,$73,$0F,$AE,$30,$FB,$FE,$37,$D9,$F1,$B2
	.dc.b	$30,$43,$E1,$39,$7E,$24,$7C,$3F,$8D,$F8,$9E,$A4,$9F,$95,$19,$7E
	.dc.b	$1B,$F0,$48,$E0,$D9,$0F,$A1,$F1,$7F,$3F,$E2,$FE,$87,$C4,$C9,$FA
	.dc.b	$D1,$96,$90,$FB,$1F,$13,$F5,$FE,$22,$47,$DD,$7E,$D7,$C4,$E2,$3E
	.dc.b	$F4,$5D,$FE,$94,$FE,$D7,$E5,$5B,$FE,$DF,$E6,$3F,$03,$87,$EB,$F8
	.dc.b	$CC,$8C,$E1,$FC,$66,$F8,$CF,$FA,$C8,$38,$7F,$72,$62,$FF,$80,$C6
	.dc.b	$63,$F7,$AD,$7B,$9B,$3A,$3C,$E7,$C1,$3E,$7E,$4F,$CC,$7C,$5F,$E6
	.dc.b	$7E,$2E,$61,$F9,$F8,$CB,$FA,$18,$CB,$FA,$18,$FD,$07,$C5,$FE,$87
	.dc.b	$E2,$E4,$FE,$9E,$32,$D2,$1F,$A9,$F8,$BF,$E3,$DF,$2C,$F6,$BF,$B7
	.dc.b	$E9,$FE,$A6,$F5,$AE,$07,$73,$30,$FF,$58,$DF,$16,$C7,$F2,$68,$C4
	.dc.b	$7C,$A6,$DF,$F4,$96,$F0,$7C,$86,$CF,$F2,$7C,$1E,$A7,$F7,$CF,$8E
	.dc.b	$FE,$49,$7F,$EE,$E7,$FC,$75,$87,$F8,$2D,$B1,$9B,$E0,$43,$FD,$D3
	.dc.b	$1C,$9F,$E4,$CB,$E3,$7B,$7C,$3F,$7F,$8C,$95,$C2,$32,$D2,$1F,$8F
	.dc.b	$39,$FC,$6F,$9F,$E3,$3F,$C1,$18,$1A,$49,$FC,$6F,$9F,$FE,$D9,$8E
	.dc.b	$11,$E6,$7E,$07,$C6,$FC,$1F,$8C,$93,$EA,$7E,$27,$C6,$E2,$3C,$57
	.dc.b	$FE,$E3,$9C,$6D,$F1,$6B,$E5,$7C,$37,$90,$97,$E2,$C9,$18,$8E,$71
	.dc.b	$87,$E4,$C7,$D3,$F8,$6F,$EC,$DF,$18,$F1,$7F,$AF,$57,$F3,$4D,$7F
	.dc.b	$46,$7C,$D7,$F6,$7D,$9F,$0D,$FD,$AC,$65,$E0,$7E,$E2,$BA,$67,$C2
	.dc.b	$7F,$5B,$18,$97,$CB,$73,$7E,$77,$1F,$EC,$A3,$15,$D5,$D5,$71,$6E
	.dc.b	$3E,$87,$75,$C6,$BD,$2D,$BF,$25,$6E,$7C,$6D,$8F,$5C,$ED,$F1,$9F
	.dc.b	$C4,$AB,$BE,$16,$F8,$BE,$D9,$E2,$53,$89,$F1,$06,$1F,$07,$1E,$71
	.dc.b	$F0,$AB,$CE,$F5,$39,$99,$3D,$35,$97,$9E,$3E,$97,$BE,$1F,$09,$D6
	.dc.b	$1F,$1E,$FC,$F1,$E3,$7E,$C6,$1B,$8F,$34,$F9,$67,$75,$92,$38,$24
	.dc.b	$F9,$A3,$2F,$98,$F4,$1E,$6C,$3E,$24,$F8,$0F,$7C,$78,$30,$FC,$8E
	.dc.b	$1F,$12,$F9,$CF,$B2,$BB,$BC,$CE,$AF,$E1,$4E,$71,$C5,$D9,$98,$F3
	.dc.b	$3C,$B0,$F7,$47,$CE,$AF,$84,$7D,$43,$93,$C0,$87,$86,$B5,$5E,$A9
	.dc.b	$E2,$FC,$E7,$33,$D6,$34,$64,$FA,$7F,$13,$FE,$18,$BF,$D2,$37,$0F
	.dc.b	$D9,$B7,$1F,$EC,$6F,$F8,$7A,$FF,$B2,$FC,$66,$31,$FE,$BE,$D8,$FF
	.dc.b	$2C,$B5,$FE,$BA,$D7,$89,$F9,$5A,$79,$4E,$7F,$CD,$4E,$AF,$F6,$D5
	.dc.b	$A9,$CC,$AF,$F4,$13,$98,$FD,$01,$C7,$3F,$F6,$4E,$1D,$63,$27,$13
	.dc.b	$F0,$F3,$C7,$4F,$E0,$91,$E6,$7C,$33,$C5,$CC,$71,$C7,$38,$F9,$31
	.dc.b	$F1,$DF,$79,$DC,$7E,$A7,$0E,$D8,$FC,$8C,$3C,$37,$D6,$31,$8F,$49
	.dc.b	$FC,$D9,$E3,$8F,$5E,$3F,$B3,$8C,$1F,$E1,$8B,$EA,$FD,$23,$AC,$6D
	.dc.b	$F4,$4F,$27,$E1,$E1,$E8,$9F,$54,$70,$3C,$27,$C7,$BF,$21,$EC,$74
	.dc.b	$D3,$35,$ED,$C7,$D4,$6C,$AF,$B8,$F1,$47,$16,$99,$9F,$66,$3F,$15
	.dc.b	$F6,$4F,$67,$DD,$5D,$1B,$5E,$D8,$E9,$1C,$AF,$E1,$61,$F4,$F0,$E9
	.dc.b	$3F,$96,$3D,$77,$E1,$7F,$D6,$51,$A7,$E7,$98,$AE,$6F,$03,$C2,$77
	.dc.b	$3E,$69,$F9,$F8,$FC,$24,$FF,$84,$FE,$2B,$F2,$99,$7F,$84,$DB,$7F
	.dc.b	$E1,$7E,$CB,$87,$F8,$42,$6F,$2B,$8D,$1F,$EE,$D9,$E9,$7F,$DD,$63
	.dc.b	$11,$C6,$3F,$BC,$7C,$4F,$2A,$37,$E2,$70,$C9,$FF,$10,$B9,$69,$FE
	.dc.b	$24,$EF,$E4,$73,$F8,$9B,$19,$3F,$C0,$D1,$E2,$CB,$D0,$DC,$3F,$C8
	.dc.b	$D3,$FE,$0B,$F8,$9F,$21,$E5,$F8,$9B,$1F,$E3,$28,$F4,$78,$18,$D3
	.dc.b	$FC,$1E,$FA,$FE,$26,$57,$E8,$CF,$F3,$DE,$4F,$B8,$FE,$1D,$1F,$E3
	.dc.b	$C9,$DD,$FF,$88,$A5,$F2,$18,$E2,$FE,$11,$87,$18,$FA,$74,$F9,$77
	.dc.b	$BB,$86,$E5,$E1,$9F,$F7,$D7,$E1,$6F,$27,$57,$2B,$FD,$0F,$5C,$E3
	.dc.b	$EB,$B6,$33,$7D,$23,$FD,$1A,$DC,$BF,$13,$E5,$FF,$A6,$2B,$FD,$29
	.dc.b	$39,$7F,$82,$46,$6B,$DD,$F1,$BF,$8A,$D4,$DF,$8D,$C3,$2F,$D5,$B7
	.dc.b	$C9,$F8,$DC,$61,$F1,$DE,$5F,$19,$CF,$E3,$6D,$F0,$C9,$18,$BF,$9C
	.dc.b	$DD,$7E,$6C,$EA,$BF,$D5,$9F,$2C,$F3,$7E,$7B,$63,$E6,$57,$D0,$D7
	.dc.b	$35,$27,$E5,$D6,$F1,$FF,$59,$B5,$FF,$57,$87,$0F,$FA,$C9,$E1,$7F
	.dc.b	$B4,$E0,$69,$D6,$BF,$CF,$13,$F8,$08,$F9,$5F,$1D,$30,$F8,$1F,$1D
	.dc.b	$6F,$F1,$93,$6F,$DF,$36,$BF,$2B,$A3,$E3,$BE,$A1,$F5,$7E,$3A,$CF
	.dc.b	$CB,$8F,$B1,$F1,$B8,$CB,$F3,$0F,$DB,$F8,$C9,$5F,$47,$E3,$6D,$7E
	.dc.b	$76,$7F,$65,$89,$A4,$AF,$F4,$6C,$65,$FB,$5F,$1B,$CB,$4F,$AE,$FE
	.dc.b	$27,$E3,$7F,$15,$C1,$6B,$F0,$0D,$C7,$F5,$8D,$BF,$BE,$FE,$47,$E3
	.dc.b	$7F,$25,$F1,$B6,$FF,$10,$DB,$FD,$A3,$6F,$F0,$EF,$E6,$7E,$36,$57
	.dc.b	$E3,$FE,$36,$DF,$EE,$E7,$DE,$49,$F6,$B6,$3F,$D4,$3D,$6F,$47,$8E
	.dc.b	$3F,$C0,$23,$17,$FA,$25,$E8,$7F,$32,$6C,$67,$E3,$71,$39,$F9,$1F
	.dc.b	$E4,$14,$69,$2B,$FD,$77,$0F,$8D,$E8,$75,$F8,$3B,$3E,$1F,$4F,$C0
	.dc.b	$78,$7E,$36,$C7,$97,$D7,$EE,$47,$AB,$E2,$F8,$DF,$47,$C6,$FA,$5B
	.dc.b	$3E,$DF,$8C,$F6,$1E,$DF,$8D,$B1,$FC,$0B,$E5,$1F,$05,$FC,$ED,$B1
	.dc.b	$FE,$D4,$6F,$27,$EB,$BC,$A4,$62,$3F,$63,$EA,$3D,$4F,$E3,$ED,$BF
	.dc.b	$E4,$6D,$9F,$DB,$9F,$D7,$E7,$E5,$FC,$77,$D1,$3E,$97,$C7,$5B,$FA
	.dc.b	$DF,$1D,$F4,$08,$FC,$E7,$E9,$FA,$1F,$67,$E3,$3E,$D3,$67,$EF,$7C
	.dc.b	$57,$DD,$3E,$F7,$C6,$58,$FC,$17,$AB,$F7,$08,$FD,$A7,$FA,$F7,$D0
	.dc.b	$7F,$2F,$EF,$FE,$92,$BE,$57,$FB,$CB,$9A,$FE,$AA,$E5,$FE,$5E,$BE
	.dc.b	$1F,$7E,$A7,$06,$CE,$A3,$B9,$1F,$8F,$C8,$AF,$E3,$BC,$48,$C7,$C6
	.dc.b	$F8,$8E,$70,$CB,$CF,$1E,$6F,$8D,$98,$79,$3E,$1A,$DE,$E3,$D3,$3F
	.dc.b	$00,$8F,$8B,$14,$62,$78,$1E,$38,$7C,$6F,$4C,$7E,$3B,$F1,$BA,$66
	.dc.b	$1F,$13,$E3,$6E,$1C,$B4,$F7,$57,$D0,$25,$78,$5F,$A7,$F1,$BF,$40
	.dc.b	$FA,$3F,$1B,$67,$E7,$C5,$BE,$99,$AB,$19,$FD,$AF,$89,$FB,$3F,$15
	.dc.b	$F6,$9B,$8F,$D8,$8B,$87,$29,$B5,$FE,$91,$87,$CA,$3D,$FF,$C2,$47
	.dc.b	$FA,$6F,$CB,$32,$75,$CD,$FE,$B0,$9C,$CF,$F6,$5C,$FF,$D3,$9F,$19
	.dc.b	$FC,$36,$38,$61,$D2,$31,$1E,$08,$C5,$72,$C3,$AC,$F8,$4F,$0F,$37
	.dc.b	$C1,$D7,$C1,$1D,$32,$AF,$36,$D7,$2C,$7B,$DF,$1A,$E8,$6C,$B3,$BC
	.dc.b	$3C,$58,$F9,$27,$C3,$97,$B2,$73,$8F,$6A,$D5,$73,$C7,$BD,$EE,$3C
	.dc.b	$C7,$69,$F9,$AF,$F8,$03,$0F,$7B,$63,$2F,$63,$F8,$7F,$C7,$3C,$67
	.dc.b	$CA,$FF,$8D,$A3,$0D,$8C,$CF,$06,$BD,$FE,$07,$FA,$63,$FC,$6B,$0F
	.dc.b	$F0,$5F,$A5,$FE,$40,$6C,$66,$76,$F1,$9F,$9B,$1C,$B3,$FB,$75,$C1
	.dc.b	$BE,$AF,$FB,$BB,$6F,$B9,$5B,$6D,$F2,$3E,$BC,$EF,$0F,$C7,$56,$DB
	.dc.b	$3B,$7E,$BC,$ED,$FC,$CC,$ED,$B3,$BB,$F1,$C5,$BF,$0B,$F8,$0A,$E1
	.dc.b	$83,$0F,$17,$3F,$F1,$7F,$8D,$FC,$EF,$1F,$F2,$86,$F1,$FF,$26,$6E
	.dc.b	$3F,$C2,$7E,$3A,$DF,$F1,$F6,$F3,$FF,$A7,$27,$F9,$5D,$62,$73,$F1
	.dc.b	$BA,$F8,$E9,$5F,$CC,$23,$2D,$21,$FC,$13,$E3,$BC,$2F,$F7,$BF,$6F
	.dc.b	$FC,$F2,$D3,$FC,$F7,$C0,$8C,$B6,$33,$1C,$A3,$11,$FC,$90,$E4,$7F
	.dc.b	$26,$E4,$FE,$D3,$EB,$EF,$83,$E0,$C3,$EB,$1E,$F6,$1B,$7F,$E9,$13
	.dc.b	$FC,$21,$5C,$5F,$2B,$E4,$23,$83,$86,$1F,$55,$B1,$E0,$F4,$5F,$15
	.dc.b	$F3,$BD,$19,$5F,$D6,$1F,$8B,$F5,$8F,$62,$9F,$19,$C4,$FF,$A8,$1A
	.dc.b	$FF,$7E,$6B,$FC,$7B,$97,$FA,$D7,$C8,$72,$8F,$64,$F0,$E8,$FE,$DB
	.dc.b	$3F,$80,$F9,$09,$5F,$DA,$9F,$C2,$FB,$CF,$87,$F2,$FB,$9E,$32,$57
	.dc.b	$8D,$FF,$59,$E5,$FE,$3E,$F8,$6E,$24,$9F,$F5,$1F,$E3,$DF,$CE,$F3
	.dc.b	$FF,$62,$BF,$EC,$88,$CF,$C8,$FE,$52,$31,$1F,$D5,$DF,$F5,$7B,$F5
	.dc.b	$7F,$05,$75,$7E,$29,$B1,$FE,$C4,$6B,$FE,$22,$6B,$F1,$7E,$13,$E9
	.dc.b	$BF,$EB,$66,$BF,$D9,$5B,$87,$9B,$C8,$FD,$4A,$FA,$17,$9D,$86,$DF
	.dc.b	$FA,$B9,$AF,$FA,$81,$FD,$63,$F9,$7F,$99,$7C,$77,$E8,$9B,$1F,$ED
	.dc.b	$C6,$BF,$E3,$A6,$BF,$BE,$ED,$FA,$8F,$FA,$F1,$AF,$FA,$B9,$B8,$3C
	.dc.b	$6F,$EC,$AF,$E8,$3E,$37,$E7,$79,$3F,$3F,$B2,$61,$F5,$A3,$F6,$E7
	.dc.b	$F6,$27,$FD,$82,$D7,$FB,$5B,$5F,$EE,$2D,$E4,$71,$7F,$B6,$A7,$F0
	.dc.b	$3F,$25,$F7,$4F,$BD,$F2,$3F,$7F,$E4,$7F,$01,$F2,$32,$BF,$0F,$1F
	.dc.b	$C2,$3E,$B9,$EC,$FF,$08,$6C,$FF,$97,$83,$87,$FC,$C2,$D7,$F4,$98
	.dc.b	$FC,$CD,$BF,$23,$C5,$F8,$F3,$F2,$1D,$9F,$C5,$BE,$33,$F2,$DC,$5F
	.dc.b	$23,$EE,$3F,$CF,$38,$FF,$BA,$9A,$FF,$11,$9D,$BF,$E1,$27,$FC,$30
	.dc.b	$FF,$BA,$1A,$FF,$84,$9A,$EE,$7F,$91,$C7,$28,$34,$EB,$83,$F0,$9F
	.dc.b	$F4,$14,$FE,$B3,$BB,$F5,$04,$7F,$5F,$CB,$5F,$D5,$45,$AF,$D8,$CF
	.dc.b	$EB,$FE,$4E,$57,$ED,$9F,$D9,$73,$E2,$3F,$65,$16,$BF,$73,$3F,$B7
	.dc.b	$F9,$39,$5F,$BE,$7F,$75,$E7,$62,$3F,$75,$16,$BF,$80,$CD,$D3,$FC
	.dc.b	$73,$CF,$F8,$8A,$C5,$FF,$9C,$78,$7F,$9F,$E1,$FC,$23,$CA,$FE,$0D
	.dc.b	$F2,$52,$BF,$86,$C7,$F0,$92,$38,$3F,$88,$FC,$9F,$F6,$0F,$C7,$BF
	.dc.b	$C0,$F7,$FA,$59,$FF,$3F,$46,$33,$F1,$B6,$33,$E3,$F1,$F6,$64,$7E
	.dc.b	$B2,$B7,$1D,$67,$F1,$5C,$1F,$D8,$3E,$71,$F2,$54,$CE,$1F,$D3,$F9
	.dc.b	$95,$F8,$22,$47,$E7,$63,$2D,$8E,$2F,$E4,$5E,$DE,$E7,$F7,$CF,$B0
	.dc.b	$7C,$8D,$25,$79,$CF,$F4,$E4,$3E,$47,$31,$FE,$85,$8E,$2E,$F9,$51
	.dc.b	$88,$D4,$F7,$F9,$0F,$07,$C7,$C9,$E1,$49,$D4,$61,$AF,$7F,$8F,$FF
	.dc.b	$02,$FC,$EB,$E4,$3F,$C0,$FF,$2C,$EF,$A6,$A2,$5F,$F0,$1A,$E0,$D8
	.dc.b	$FF,$06,$C6,$23,$FD,$C6,$DE,$7F,$F4,$63,$79,$BF,$D0,$AD,$9F,$F4
	.dc.b	$5E,$1F,$C1,$E7,$FD,$13,$F6,$1F,$EB,$09,$F8,$9F,$3A,$F8,$BF,$F1
	.dc.b	$5B,$63,$FC,$15,$3B,$7F,$C2,$CD,$C7,$FD,$38,$DB,$EB,$F1,$FC,$BE
	.dc.b	$4B,$9F,$C9,$74,$89,$87,$82,$32,$D2,$1F,$EB,$DC,$7F,$8A,$E3,$F9
	.dc.b	$DE,$47,$6B,$D6,$FE,$EE,$DB,$FF,$70,$B5,$FF,$53,$9F,$A4,$8D,$9F
	.dc.b	$A9,$23,$F8,$58,$FF,$53,$D7,$75,$F3,$3F,$AF,$EB,$95,$B8,$FD,$91
	.dc.b	$1C,$1C,$F8,$3F,$6D,$EF,$F9,$89,$1E,$66,$BF,$ED,$BB,$FF,$0E,$6F
	.dc.b	$FB,$89,$FD,$F7,$B9,$FE,$50,$F8,$D3,$FE,$F7,$B7,$82,$E0,$FF,$62
	.dc.b	$8F,$94,$FF,$63,$C6,$67,$FD,$9B,$39,$7F,$D9,$F1,$95,$7D,$B6,$CD
	.dc.b	$7F,$B4,$23,$12,$E1,$E1,$3C,$E3,$0F,$8B,$0F,$F6,$BC,$F5,$AE,$2D
	.dc.b	$C7,$D4,$F2,$9D,$D7,$1B,$FF,$71,$CE,$F0,$B3,$E3,$BE,$2E,$EB,$C2
	.dc.b	$7B,$AF,$9B,$FE,$EC,$C3,$C3,$3C,$6F,$DB,$3B,$29,$F2,$EB,$E4,$CE
	.dc.b	$E3,$E7,$B7,$0F,$44,$79,$0E,$47,$88,$F9,$78,$79,$9F,$13,$5E,$C7
	.dc.b	$12,$9C,$70,$F3,$BC,$30,$F2,$DF,$CF,$C7,$C3,$3E,$43,$45,$3C,$33
	.dc.b	$D6,$7E,$B6,$1C,$0F,$99,$3D,$F0,$F2,$4E,$8F,$99,$3A,$37,$7E,$02
	.dc.b	$47,$C5,$74,$F0,$25,$79,$27,$2F,$C6,$7C,$51,$F2,$C9,$19,$78,$BE
	.dc.b	$A3,$CD,$7F,$0B,$1F,$AB,$1F,$42,$FC,$B5,$CA,$3E,$5D,$F6,$7C,$51
	.dc.b	$E7,$3F,$11,$3E,$A8,$F9,$2F,$27,$A4,$68,$E0,$FC,$EF,$0F,$F6,$2D
	.dc.b	$9F,$C3,$46,$23,$D0,$D4,$3E,$3A,$C6,$63,$01,$C4,$6A,$7F,$8D,$CE
	.dc.b	$A7,$2F,$DF,$C5,$BF,$F8,$0E,$BF,$79,$3E,$0B,$FC,$ED,$6B,$6F,$CF
	.dc.b	$CE,$4E,$57,$C7,$1E,$F7,$C2,$31,$3F,$CB,$23,$F3,$17,$CE,$3F,$40
	.dc.b	$73,$CB,$F8,$64,$65,$F1,$C7,$BB,$1C,$DF,$48,$F8,$A6,$70,$B7,$D6
	.dc.b	$B8,$E1,$E0,$BF,$F1,$06,$18,$C3,$F6,$17,$DF,$0F,$4C,$62,$B7,$18
	.dc.b	$AF,$0E,$3F,$B1,$AF,$44,$F8,$2B,$71,$EF,$39,$4E,$DF,$31,$FE,$94
	.dc.b	$8F,$53,$F2,$F1,$EA,$FE,$CC,$EA,$6E,$79,$63,$EB,$6C,$76,$3E,$59
	.dc.b	$F8,$4C,$39,$9E,$FA,$EA,$5C,$3D,$15,$F7,$2B,$84,$61,$E7,$7D,$DA
	.dc.b	$77,$BF,$94,$76,$9F,$93,$7D,$9A,$73,$76,$F5,$8B,$3E,$D3,$AB,$CD
	.dc.b	$F9,$4D,$7E,$FD,$78,$E3,$8C,$31,$FB,$4F,$D1,$BF,$13,$6F,$C4,$7F
	.dc.b	$97,$BC,$1F,$F3,$15,$7F,$79,$6E,$3F,$E6,$2F,$99,$B7,$FE,$66,$9B
	.dc.b	$CB,$FE,$0C,$7F,$D3,$AD,$AF,$F0,$5C,$63,$88,$DF,$13,$8C,$7F,$99
	.dc.b	$27,$FC,$15,$F3,$38,$99,$1F,$E0,$DF,$12,$E3,$FD,$FD,$E3,$E2,$49
	.dc.b	$FF,$09,$7C,$C5,$AF,$F0,$24,$F8,$49,$3F,$E1,$E7,$CF,$F5,$6F,$67
	.dc.b	$D6,$D9,$FF,$30,$2F,$63,$33,$27,$3E,$CD,$C3,$E8,$47,$FA,$6A,$D3
	.dc.b	$F2,$25,$F5,$8F,$81,$C3,$6F,$BF,$B1,$89,$91,$E2,$F2,$2D,$F6,$1C
	.dc.b	$6C,$BF,$CC,$D5,$FE,$7D,$9F,$F8,$01,$C7,$E1,$8F,$A2,$7F,$A0,$79
	.dc.b	$5F,$4C,$F8,$5D,$54,$AF,$F7,$61,$C2,$FF,$94,$B5,$FF,$22,$B5,$FE
	.dc.b	$22,$EE,$FF,$17,$6F,$8E,$E9,$5E,$99,$F0,$C7,$D0,$F5,$E5,$FD,$0F
	.dc.b	$52,$DF,$BF,$4F,$75,$7D,$6F,$57,$38,$FF,$97,$75,$47,$E5,$35,$FF
	.dc.b	$1E,$E5,$FE,$99,$6E,$3C,$DF,$F3,$14,$67,$B7,$81,$1E,$1E,$45,$AF
	.dc.b	$A0,$3E,$79,$F2,$5F,$75,$B5,$F3,$7D,$97,$B5,$9A,$93,$FD,$4A,$F9
	.dc.b	$47,$FB,$45,$FF,$37,$DF,$27,$FA,$BB,$5C,$C7,$EC,$B5,$61,$2F,$F0
	.dc.b	$17,$C6,$7F,$77,$39,$BF,$F3,$37,$CE,$7E,$C0,$FD,$8F,$AB,$70,$FF
	.dc.b	$3D,$45,$9F,$F3,$64,$DB,$FF,$3E,$7A,$FF,$B8,$3F,$73,$EB,$5C,$3F
	.dc.b	$D0,$51,$FB,$CF,$5B,$18,$7F,$9D,$5F,$E0,$FE,$AF,$F7,$AA,$FF,$66
	.dc.b	$36,$7E,$F2,$F9,$8D,$C4,$9D,$BC,$A3,$0F,$02,$E3,$FE,$9F,$AE,$55
	.dc.b	$C0,$EF,$E7,$78,$8F,$1C,$58,$C9,$2B,$C7,$19,$7C,$BE,$AF,$A8,$F5
	.dc.b	$FC,$E5,$AF,$5C,$59,$F5,$CD,$C3,$EF,$79,$DF,$C3,$67,$F1,$6D,$8F
	.dc.b	$E6,$3F,$6F,$FE,$93,$C3,$F7,$53,$97,$F9,$FF,$8B,$F8,$7B,$B1,$F8
	.dc.b	$B9,$E9,$1F,$E0,$D9,$D3,$FE,$19,$7F,$11,$F5,$8F,$88,$FE,$29,$5D
	.dc.b	$6B,$FA,$44,$E6,$FE,$AF,$A1,$F6,$0F,$B3,$EC,$D8,$FE,$1D,$D6,$F7
	.dc.b	$FB,$E1,$FE,$AC,$6C,$66,$7F,$40,$D9,$FF,$66,$F8,$3F,$EE,$9F,$9E
	.dc.b	$FC,$2F,$C3,$FE,$19,$B7,$F8,$6F,$9E,$FC,$49,$F8,$AF,$9D,$B1,$FC
	.dc.b	$7F,$D7,$72,$78,$9B,$1F,$AF,$6E,$5F,$E4,$26,$BF,$E2,$8C,$BF,$DF
	.dc.b	$5F,$3D,$F9,$6F,$8F,$FC,$BB,$6F,$FC,$A3,$CD,$FE,$EB,$F9,$66,$BF
	.dc.b	$D8,$9B,$3A,$3F,$CA,$F8,$21,$B6,$EB,$FA,$27,$CF,$BE,$5F,$E4,$B7
	.dc.b	$0F,$E0,$CD,$D3,$FC,$B2,$D7,$D1,$F4,$CF,$6F,$FB,$9B,$6B,$C5,$3F
	.dc.b	$D1,$A1,$ED,$6A,$BF,$BC,$FA,$0F,$1A,$9F,$B6,$F9,$EB,$8F,$F4,$BF
	.dc.b	$A1,$7A,$FF,$79,$B1,$FC,$6D,$B9,$FF,$9B,$9A,$FC,$DF,$0F,$F8,$A4
	.dc.b	$FC,$46,$C7,$F5,$A5,$F3,$DF,$C1,$BC,$DF,$F0,$6F,$D6,$26,$9F,$C2
	.dc.b	$BE,$82,$E1,$FD,$7F,$E8,$7F,$85,$63,$FE,$62,$0E,$1F,$F3,$2B,$5F
	.dc.b	$F4,$5B,$79,$FB,$7C,$DF,$F1,$63,$C3,$E4,$DB,$FE,$2F,$CD,$FD,$EE
	.dc.b	$FF,$86,$E1,$FC,$62,$31,$E5,$CA,$FE,$7D,$5A,$AF,$23,$5F,$1C,$FF
	.dc.b	$8A,$33,$7C,$EF,$F2,$63,$94,$FF,$24,$7C,$1E,$0C,$C3,$FB,$F7,$8E
	.dc.b	$F9,$FC,$5F,$F5,$6C,$97,$FB,$46,$DF,$92,$30,$47,$F0,$FF,$42,$FA
	.dc.b	$D7,$C7,$F4,$F0,$95,$FE,$B2,$AD,$4F,$C3,$F0,$E4,$F9,$3E,$2A,$E7
	.dc.b	$E3,$9F,$11,$27,$E2,$BF,$07,$E8,$3F,$A2,$9F,$E4,$CF,$A3,$EF,$3F
	.dc.b	$D1,$A3,$2F,$E7,$39,$93,$FE,$19,$8D,$19,$71,$7F,$E0,$97,$FC,$25
	.dc.b	$19,$F8,$4E,$DE,$07,$76,$46,$4E,$51,$8A,$FF,$00,$D7,$F8,$E7,$97
	.dc.b	$FD,$89,$C5,$FE,$B0,$8C,$4F,$FB,$56,$3F,$D6,$7C,$BF,$42,$30,$7F
	.dc.b	$B6,$A7,$7E,$87,$D9,$8C,$65,$FD,$85,$D9,$EE,$71,$1B,$8F,$D2,$71
	.dc.b	$FD,$47,$75,$D3,$0F,$25,$F0,$6E,$3E,$8A,$FF,$75,$E3,$E5,$C7,$A6
	.dc.b	$1C,$67,$C7,$5B,$7C,$4D,$AE,$17,$D2,$7C,$38,$77,$78,$9E,$A9,$DB
	.dc.b	$8C,$3B,$CF,$BE,$3F,$71,$87,$CA,$6C,$65,$EC,$7E,$F7,$E3,$FF,$D1
	.dc.b	$7C,$5F,$E2,$B6,$C7,$FB,$F9,$AE,$FE,$11,$3F,$E3,$1B,$7C,$F3,$E3
	.dc.b	$F0,$DB,$FE,$14,$7E,$39,$C3,$E6,$6B,$F8,$17,$86,$5F,$29,$BA,$66
	.dc.b	$BE,$55,$BE,$17,$FD,$73,$39,$C3,$EC,$DF,$07,$F0,$CD,$7F,$3F,$3C
	.dc.b	$1C,$DF,$49,$D5,$FB,$23,$BD,$6E,$BD,$51,$B2,$3F,$88,$8D,$4F,$18
	.dc.b	$F1,$CE,$B0,$FC,$2D,$78,$78,$BE,$25,$7E,$BF,$D6,$34,$FD,$96,$38
	.dc.b	$7F,$64,$DE,$8F,$CC,$57,$EE,$AB,$1C,$9F,$DA,$33,$FF,$0B,$61,$C8
	.dc.b	$FD,$E7,$CF,$69,$FF,$50,$C7,$FA,$E5,$FE,$01,$EC,$67,$0F,$E0,$B5
	.dc.b	$68,$D9,$95,$D0,$FE,$11,$F1,$7F,$E7,$5B,$FE,$1A,$DB,$FF,$61,$4F
	.dc.b	$E1,$27,$2D,$8C,$BF,$C3,$E7,$C0,$78,$7E,$2F,$C5,$F3,$BE,$3F,$2A
	.dc.b	$57,$F1,$78,$CB,$CD,$F0,$11,$C1,$B2,$1F,$C7,$7E,$2F,$DD,$F3,$72
	.dc.b	$3F,$91,$3F,$07,$E6,$64,$79,$DF,$61,$1F,$57,$C3,$E9,$FE,$51,$1F
	.dc.b	$26,$3E,$57,$C3,$49,$FE,$5C,$FC,$DF,$8C,$F9,$DC,$BF,$1C,$93,$F1
	.dc.b	$5F,$95,$F1,$7F,$4F,$E0,$24,$FF,$37,$8F,$B1,$F8,$F7,$EB,$B9,$FF
	.dc.b	$0A,$DE,$6F,$F1,$5B,$6B,$FD,$C2,$BE,$07,$8F,$C5,$FD,$98,$34,$93
	.dc.b	$F6,$7E,$73,$FD,$C0,$FF,$4B,$9F,$E5,$98,$78,$3B,$BF,$15,$5A,$AF
	.dc.b	$F6,$95,$3E,$63,$CE,$48,$E5,$E5,$FE,$57,$07,$D0,$BE,$0B,$FD,$81
	.dc.b	$18,$6C,$66,$BF,$87,$CE,$5F,$F6,$37,$D0,$BA,$3F,$D9,$10,$8C,$3A
	.dc.b	$8C,$C6,$23,$FC,$DB,$93,$CE,$F9,$A4,$8F,$81,$F3,$5F,$43,$E6,$BE
	.dc.b	$41,$23,$F3,$D4,$8F,$99,$E4,$7D,$6F,$9A,$95,$F2,$63,$EC,$FC,$CC
	.dc.b	$8F,$98,$FD,$CF,$81,$FB,$BF,$05,$F5,$49,$3D,$9F,$B1,$F3,$1F,$82
	.dc.b	$F9,$99,$3F,$56,$3F,$0B,$F0,$9F,$86,$F9,$49,$3F,$66,$32,$FA,$1F
	.dc.b	$C1,$11,$C1,$B2,$1F,$73,$E4,$FF,$23,$F2,$92,$3F,$08,$FE,$53,$E5
	.dc.b	$31,$1F,$82,$8B,$CB,$F9,$69,$FE,$21,$F8,$F7,$F9,$46,$3F,$8E,$BF
	.dc.b	$D7,$5B,$EC,$6A,$E1,$FE,$1B,$9D,$78,$1F,$E8,$86,$C3,$CE,$47,$59
	.dc.b	$D7,$CA,$F8,$3D,$99,$7C,$E3,$C6,$61,$63,$E4,$9C,$C6,$CF,$13,$B8
	.dc.b	$F4,$CF,$B3,$E0,$FD,$BF,$23,$2F,$D1,$19,$69,$0E,$F8,$7A,$FE,$53
	.dc.b	$E2,$FC,$A7,$C6,$F9,$09,$5F,$06,$32,$FC,$38,$CB,$C8,$F8,$5A,$27
	.dc.b	$FA,$14,$7F,$95,$BE,$02,$FC,$FB,$8A,$9A,$F1,$53,$4F,$D5,$B8,$AF
	.dc.b	$EF,$E7,$32,$B9,$7F,$C0,$6D,$DF,$FC,$61,$1B,$3C,$C5,$73,$6E,$4F
	.dc.b	$C2,$FC,$97,$E1,$BE,$56,$5F,$E0,$63,$2F,$E0,$E3,$2F,$C1,$E4,$FC
	.dc.b	$0C,$FE,$3B,$E0,$FF,$1F,$F0,$B2,$BF,$15,$19,$69,$0E,$5B,$FF,$0F
	.dc.b	$7C,$FF,$50,$AF,$E3,$1F,$70,$E3,$A7,$CE,$9D,$57,$FB,$30,$38,$7F
	.dc.b	$C4,$ED,$7F,$D6,$4D,$7F,$C0,$4D,$7F,$C0,$EF,$1D,$BE,$5B,$79,$37
	.dc.b	$87,$93,$E2,$FF,$5B,$F2,$7F,$AE,$F9,$19,$8F,$EA,$63,$2D,$21,$F0
	.dc.b	$3E,$4B,$F6,$BF,$23,$FB,$6F,$91,$93,$FB,$18,$CB,$48,$7C,$8F,$91
	.dc.b	$FD,$EE,$3F,$E0,$5E,$0F,$F5,$BC,$FF,$80,$DA,$FF,$B4,$78,$FF,$6F
	.dc.b	$3B,$38,$C7,$EF,$DF,$E0,$FF,$1D,$FE,$17,$F5,$BF,$17,$C3,$FE,$24
	.dc.b	$7F,$A8,$37,$E6,$B5,$C3,$30,$FF,$51,$56,$A3,$FC,$63,$8F,$FB,$0E
	.dc.b	$B5,$5F,$89,$AF,$CD,$3F,$E3,$39,$FF,$64,$75,$FF,$87,$E7,$33,$C9
	.dc.b	$AF,$F1,$17,$F4,$17,$99,$F4,$1E,$9F,$8F,$FD,$9F,$C7,$FF,$47,$CB
	.dc.b	$FC,$87,$ED,$FE,$2F,$1F,$F2,$2F,$3B,$C4,$FF,$22,$4E,$67,$F2,$11
	.dc.b	$89,$FF,$24,$5F,$39,$FE,$FD,$18,$F4,$7F,$8D,$C3,$E6,$39,$4F,$58
	.dc.b	$FF,$29,$C6,$2F,$FC,$A4,$F7,$BE,$8D,$BE,$6E,$EB,$8D,$FF,$96,$4E
	.dc.b	$8F,$77,$C5,$8E,$A3,$FC,$C3,$7D,$27,$86,$1E,$06,$CE,$C4,$ED,$E3
	.dc.b	$87,$B3,$1F,$14,$7C,$03,$E4,$DF,$C5,$77,$87,$CE,$C3,$C7,$1E,$53
	.dc.b	$67,$77,$D3,$8F,$BE,$7C,$58,$F8,$6B,$E8,$9B,$BF,$43,$D8,$E8,$7A
	.dc.b	$31,$F3,$61,$EC,$8F,$31,$C8,$B5,$F6,$71,$F0,$47,$29,$F1,$9C,$71
	.dc.b	$C9,$1C,$12,$3C,$51,$CA,$7E,$29,$27,$E0,$BC,$67,$E0,$9D,$F0,$F5
	.dc.b	$BE,$9C,$7E,$0E,$3F,$80,$BE,$D3,$E9,$C3,$93,$B3,$91,$F1,$4E,$35
	.dc.b	$1F,$83,$3F,$86,$C3,$C9,$1F,$25,$E7,$1F,$02,$3C,$B5,$DB,$1F,$8D
	.dc.b	$3F,$2E,$B8,$D7,$85,$F8,$11,$92,$47,$57,$F0,$3F,$1F,$FE,$63,$8F
	.dc.b	$B0,$D7,$D4,$DA,$FD,$63,$98,$FF,$5C,$FC,$7F,$FA,$EA,$31,$08,$C5
	.dc.b	$FF,$B9,$1B,$1F,$EB,$D3,$47,$FB,$AE,$34,$FE,$13,$1F,$F2,$96,$DF
	.dc.b	$97,$8E,$15,$5F,$F0,$AF,$AE,$9C,$47,$90,$E5,$19,$7F,$1E,$66,$FF
	.dc.b	$85,$5F,$68,$C9,$E4,$CF,$35,$C5,$FE,$21,$1D,$5C,$DF,$ED,$63,$A6
	.dc.b	$7F,$ED,$7A,$F0,$61,$C2,$7D,$EF,$07,$BE,$3F,$BD,$C7,$C5,$8E,$E3
	.dc.b	$17,$E1,$C3,$E2,$CF,$13,$57,$FA,$EB,$E3,$3A,$3C,$58,$F9,$67,$D9
	.dc.b	$18,$7F,$6A,$7A,$0F,$F7,$9C,$FC,$3C,$7A,$9E,$73,$BD,$F8,$6B,$B5
	.dc.b	$70,$C3,$CC,$FD,$63,$C8,$E1,$FF,$7E,$63,$EC,$C7,$C8,$FD,$BC,$7C
	.dc.b	$17,$F8,$2C,$7C,$0F,$6B,$FB,$F7,$C1,$EA,$DC,$7C,$91,$E1,$6B,$E0
	.dc.b	$3F,$07,$1F,$86,$7E,$9C,$FC,$3A,$F6,$BF,$16,$AC,$FB,$6B,$B4,$E4
	.dc.b	$F2,$5F,$D0,$9F,$04,$65,$A7,$13,$E7,$9E,$03,$17,$F3,$A3,$C4,$7F
	.dc.b	$6B,$F8,$FF,$CC,$57,$F7,$D6,$DF,$F6,$FF,$8F,$B8,$7F,$6F,$9B,$C6
	.dc.b	$E3,$47,$F9,$42,$7A,$5F,$EC,$5E,$0D,$AE,$13,$B9,$FE,$F5,$F1,$FD
	.dc.b	$2D,$F1,$F6,$30,$DC,$3F,$C1,$AD,$AF,$F0,$53,$C7,$E3,$A4,$FF,$84
	.dc.b	$BE,$3A,$D7,$F8,$3E,$7C,$19,$7A,$5B,$89,$A6,$3D,$FF,$21,$E5,$F8
	.dc.b	$EB,$8F,$F8,$7A,$3D,$1F,$1D,$8C,$7F,$C3,$8F,$AF,$E3,$A5,$7F,$89
	.dc.b	$3E,$3A,$D7,$F8,$86,$7E,$07,$C7,$63,$8B,$FC,$5C,$FD,$09,$7D,$43
	.dc.b	$E3,$B0,$DC,$7F,$56,$7E,$B3,$F0,$FF,$E1,$D8,$7F,$A0,$A7,$57,$FE
	.dc.b	$82,$6B,$FC,$95,$A8,$8E,$1B,$7F,$5D,$F1,$BF,$84,$95,$3E,$1B,$CB
	.dc.b	$39,$8C,$61,$FE,$44,$F9,$0F,$BC,$7D,$F9,$B1,$93,$FC,$93,$F2,$18
	.dc.b	$C7,$FC,$8C,$FE,$17,$E3,$FF,$0D,$F2,$16,$BF,$C9,$ED,$AF,$F2,$63
	.dc.b	$73,$FF,$2C,$3F,$2F,$B9,$D7,$4F,$57,$D6,$FC,$8C,$3F,$84,$C6,$5F
	.dc.b	$F5,$8C,$F0,$6C,$7C,$38,$C6,$1F,$D4,$63,$9E,$DC,$63,$FD,$8D,$8B
	.dc.b	$0F,$87,$1F,$EA,$9F,$A9,$78,$5F,$AC,$5E,$05,$8C,$37,$3F,$BF,$5F
	.dc.b	$EF,$1A,$DC,$63,$AD,$1D,$23,$0E,$A7,$C1,$F2,$5D,$CF,$07,$65,$AD
	.dc.b	$C7,$8F,$E4,$71,$5E,$77,$CD,$EE,$79,$FC,$5B,$7D,$9B,$5E,$D6,$ED
	.dc.b	$FD,$97,$3F,$2C,$62,$BF,$D0,$B9,$FD,$1B,$F6,$61,$D6,$7D,$3B,$FC
	.dc.b	$5F,$1F,$E3,$36,$32,$7B,$A3,$2D,$3E,$0F,$C6,$7C,$D3,$E7,$7C,$6D
	.dc.b	$8C,$93,$0F,$A9,$19,$7E,$B7,$C6,$FD,$53,$EB,$7C,$6D,$AF,$9B,$1F
	.dc.b	$C3,$33,$34,$91,$FC,$86,$32,$FD,$4F,$8D,$E5,$7F,$83,$7F,$05,$F1
	.dc.b	$9F,$83,$F8,$DB,$5F,$69,$FC,$37,$C6,$77,$24,$75,$7C,$5F,$1B,$F8
	.dc.b	$B3,$F1,$9F,$09,$67,$F0,$7F,$0D,$F8,$FF,$8D,$FC,$83,$63,$F1,$5F
	.dc.b	$53,$FC,$0D,$FE,$70,$DD,$3F,$D9,$4D,$7F,$D7,$0D,$7F,$D7,$EB,$E3
	.dc.b	$BF,$A3,$B6,$7A,$FA,$D8,$9F,$E3,$DF,$38,$F7,$BF,$87,$36,$BF,$A5
	.dc.b	$60,$8C,$3F,$D3,$5A,$FF,$B2,$7E,$71,$C9,$FC,$F9,$B5,$FD,$45,$B3
	.dc.b	$FE,$E3,$6B,$EA,$F9,$0F,$39,$E8,$F8,$FB,$5F,$AD,$87,$C3,$7B,$3E
	.dc.b	$43,$FB,$1C,$62,$7E,$4C,$7C,$0F,$6F,$FA,$7E,$E6,$92,$7E,$6F,$C7
	.dc.b	$FC,$63,$E3,$FC,$85,$8F,$A7,$F1,$DF,$2B,$E3,$FE,$5B,$63,$EB,$FC
	.dc.b	$7F,$62,$3F,$61,$F0,$FC,$6F,$D2,$3E,$9F,$CA,$59,$FB,$FF,$1D,$F5
	.dc.b	$B7,$FA,$ED,$8F,$C1,$C7,$A8,$8F,$E1,$1F,$6F,$C7,$7D,$E3,$87,$13
	.dc.b	$CA,$FE,$01,$C0,$9F,$C1,$FC,$6C,$C3,$F1,$FF,$1D,$70,$E1,$1F,$8A
	.dc.b	$9F,$92,$47,$F2,$6D,$8D,$8F,$31,$87,$E1,$B1,$F0,$D7,$E4,$BA,$BF
	.dc.b	$27,$F1,$B3,$1F,$CF,$FC,$6D,$CB,$9E,$5C,$AB,$EC,$92,$BF,$46,$FF
	.dc.b	$82,$BE,$4F,$FD,$09,$1F,$CC,$57,$C7,$7F,$89,$9C,$DF,$47,$33,$B3
	.dc.b	$53,$B6,$9C,$FE,$3B,$97,$C7,$73,$6C,$64,$F0,$46,$5A,$7F,$46,$B7
	.dc.b	$CB,$27,$F6,$91,$FE,$A9,$F5,$BF,$C0,$0E,$A3,$3E,$DF,$F8,$F6,$3F
	.dc.b	$C0,$9D,$EE,$9F,$F4,$F4,$E6,$78,$78,$9F,$EB,$6C,$39,$E3,$DE,$31
	.dc.b	$3F,$E0,$D8,$C3,$E1,$3A,$DF,$4A,$E3,$8F,$89,$B7,$CA,$BB,$9C,$46
	.dc.b	$3E,$5C,$B5,$8F,$6C,$BC,$CD,$C3,$9D,$F1,$BE,$0F,$39,$D3,$E7,$C7
	.dc.b	$C4,$DC,$7C,$55,$C3,$0E,$B5,$AA,$F2,$9F,$ED,$1B,$F7,$B6,$3E,$8B
	.dc.b	$6F,$F8,$EF,$C7,$BD,$27,$CB,$E1,$B5,$FE,$6B,$9D,$4E,$4F,$E4,$B8
	.dc.b	$7F,$9C,$BD,$BB,$F1,$FF,$E0,$CB,$FF,$39,$4E,$5B,$19,$7C,$CE,$F3
	.dc.b	$F8,$ED,$D7,$ED,$B6,$7E,$D7,$C0,$FF,$AB,$A7,$38,$7D,$0A,$D5,$7E
	.dc.b	$41,$B8,$2F,$84,$FE,$0E,$72,$4A,$FB,$33,$A9,$E5,$1D,$48,$FD,$78
	.dc.b	$D5,$7B,$70,$F4,$D6,$9F,$C0,$CF,$83,$95,$C9,$FE,$98,$C3,$F7,$18
	.dc.b	$FE,$8B,$A7,$FC,$2D,$39,$9F,$DC,$7B,$F9,$E3,$FF,$33,$61,$CA,$31
	.dc.b	$3F,$C4,$6B,$85,$F2,$C3,$37,$BA,$FE,$0D,$F1,$F8,$BF,$F4,$FC,$7F
	.dc.b	$09,$F8,$FC,$4C,$8D,$90,$FE,$15,$F1,$DC,$7E,$3F,$FB,$2E,$1E,$C6
	.dc.b	$E0,$69,$2B,$F8,$B3,$C5,$FE,$31,$F1,$DE,$3F,$8F,$F2,$7C,$7C,$9F
	.dc.b	$E3,$91,$97,$FD,$75,$68,$EE,$48,$CB,$B7,$F9,$03,$EF,$F8,$EF,$81
	.dc.b	$F1,$F2,$7F,$94,$3F,$0F,$E3,$E4,$76,$7D,$64,$7C,$3F,$1B,$A7,$F9
	.dc.b	$64,$7C,$B8,$F9,$9F,$1B,$F3,$79,$65,$7F,$35,$8F,$A1,$F1,$BF,$18
	.dc.b	$91,$EC,$7E,$4F,$C6,$FD,$52,$47,$F9,$B7,$C0,$FD,$47,$37,$F9,$6D
	.dc.b	$FF,$3B,$C6,$7E,$2F,$EF,$C1,$BF,$19,$8E,$EF,$AF,$3F,$CA,$6B,$F9
	.dc.b	$8F,$2F,$F3,$59,$C9,$AF,$8C,$E6,$49,$F1,$7C,$5F,$6F,$8C,$EF,$F1
	.dc.b	$92,$7C,$D1,$E2,$F8,$CE,$E4,$8F,$F7,$96,$48,$F1,$F8,$DE,$82,$4F
	.dc.b	$BF,$E3,$3F,$3A,$FE,$A3,$AF,$FC,$66,$D3,$FC,$6B,$D4,$8C,$B6,$3B
	.dc.b	$FC,$67,$C0,$23,$83,$C7,$F0,$3F,$20,$8F,$CF,$F8,$BF,$97,$F1,$7F
	.dc.b	$33,$E2,$E4,$FD,$38,$F9,$FF,$17,$F3,$09,$1E,$E7,$E7,$FC,$5F,$D4
	.dc.b	$24,$7D,$BF,$8B,$F6,$FE,$3D,$FD,$6F,$A9,$E5,$FF,$64,$FC,$AF,$FB
	.dc.b	$45,$FF,$6B,$6B,$FE,$10,$15,$C6,$7F,$B5,$BF,$76,$2F,$4F,$C8,$8C
	.dc.b	$47,$FB,$BD,$AF,$28,$F1,$91,$F0,$7A,$BF,$8C,$F8,$CF,$C6,$FC,$6C
	.dc.b	$9F,$CA,$46,$5F,$CB,$46,$5F,$F0,$DF,$42,$7F,$89,$FE,$41,$FE,$00
	.dc.b	$9F,$F0,$FB,$74,$FF,$5B,$37,$3F,$E3,$EF,$F5,$88,$CC,$F2,$21,$FE
	.dc.b	$D1,$8D,$4F,$0F,$8E,$EF,$EC,$4A,$D4,$78,$8C,$3E,$28,$C3,$A3,$A3
	.dc.b	$AB,$F2,$CF,$AB,$E3,$3D,$7F,$19,$2F,$C5,$19,$69,$0F,$7F,$C6,$7C
	.dc.b	$2F,$8C,$F8,$7F,$19,$27,$D5,$19,$7D,$91,$97,$E3,$C7,$E9,$3F,$0A
	.dc.b	$FF,$41,$C7,$F8,$BF,$F2,$BF,$82,$83,$85,$1C,$1F,$C5,$35,$FA,$8D
	.dc.b	$7F,$C6,$AD,$7A,$35,$E7,$F1,$3E,$32,$3E,$33,$EF,$FC,$6C,$7E,$B3
	.dc.b	$F8,$2F,$8E,$C4,$7E,$0A,$2E,$DF,$82,$9F,$C3,$FC,$1F,$E2,$3E,$3A
	.dc.b	$5F,$DF,$8C,$B4,$87,$E3,$3E,$3B,$F8,$4F,$CB,$FF,$92,$63,$F0,$DF
	.dc.b	$1F,$F9,$23,$35,$23,$FD,$7F,$3B,$7F,$2C,$1C,$3F,$99,$6B,$FE,$5A
	.dc.b	$78,$63,$FB,$A8,$FA,$51,$8E,$5D,$57,$E8,$B4,$FD,$1F,$C7,$FE,$93
	.dc.b	$E3,$66,$1F,$9B,$8C,$B4,$87,$EA,$3E,$3F,$F5,$7F,$1D,$FA,$CF,$8D
	.dc.b	$93,$FA,$38,$CB,$48,$7E,$C3,$E3,$BF,$C0,$BF,$91,$79,$E3,$F6,$9E
	.dc.b	$B2,$31,$1F,$D1,$9A,$FF,$84,$9A,$FF,$A8,$5E,$5B,$FC,$19,$E5,$B7
	.dc.b	$63,$BF,$C7,$F8,$3C,$3F,$0F,$85,$31,$FC,$C6,$1C,$71,$56,$E7,$38
	.dc.b	$78,$63,$11,$FE,$A8,$9F,$F1,$0C,$F9,$7E,$27,$FC,$2B,$06,$92,$7E
	.dc.b	$EE,$5E,$F8,$F8,$64,$8E,$E7,$B7,$E1,$FF,$A9,$FC,$77,$F5,$9C,$BF
	.dc.b	$5D,$ED,$FF,$BB,$6B,$F6,$1E,$03,$C3,$FF,$4E,$CE,$67,$FD,$E5,$18
	.dc.b	$9F,$F5,$0D,$F3,$95,$6E,$31,$F2,$FC,$6B,$A5,$7E,$F6,$31,$7F,$EA
	.dc.b	$A7,$87,$B1,$FE,$E1,$94,$63,$1E,$17,$FE,$B0,$39,$CE,$9C,$5F,$37
	.dc.b	$CB,$7C,$EF,$95,$6C,$F7,$D6,$DA,$76,$3F,$89,$61,$EA,$BE,$78,$7C
	.dc.b	$87,$BC,$D3,$89,$F0,$9D,$B4,$E7,$87,$8E,$38,$C7,$8C,$F9,$A7,$84
	.dc.b	$57,$85,$E3,$59,$9F,$83,$87,$CA,$9E,$15,$B7,$8D,$FC,$BC,$75,$1E
	.dc.b	$33,$B9,$67,$ED,$47,$C6,$AD,$92,$FD,$6E,$5F,$4B,$9A,$F2,$9E,$BA
	.dc.b	$E7,$3E,$EA,$F0,$9F,$48,$E5,$3F,$22,$B9,$3E,$8C,$3D,$78,$7C,$78
	.dc.b	$D5,$F8,$0F,$19,$1C,$90,$E4,$78,$E3,$E1,$47,$8C,$91,$E8,$BF,$1B
	.dc.b	$F1,$2F,$C7,$87,$CB,$3D,$71,$F7,$A7,$D7,$5F,$79,$F6,$3D,$23,$E6
	.dc.b	$C6,$1F,$8B,$F1,$BF,$E6,$07,$CA,$DA,$FD,$53,$70,$FD,$13,$B9,$FE
	.dc.b	$C9,$F1,$BF,$D9,$63,$10,$8C,$5F,$F4,$06,$C7,$F8,$5D,$AF,$F6,$D3
	.dc.b	$81,$F8,$3C,$9E,$DF,$F4,$1A,$FE,$01,$3E,$03,$51,$9C,$BB,$46,$38
	.dc.b	$7F,$44,$F5,$C2,$E1,$D2,$73,$3C,$F4,$FE,$19,$39,$8F,$D1,$19,$BF
	.dc.b	$F6,$66,$1F,$D4,$EB,$C7,$3E,$8C,$78,$CF,$8E,$3D,$C7,$8C,$FC,$F1
	.dc.b	$CC,$FE,$6A,$F3,$AF,$4D,$72,$C3,$F6,$15,$B1,$87,$96,$7C,$38,$FE
	.dc.b	$D6,$31,$7C,$8F,$3D,$F2,$BF,$9B,$7D,$5F,$69,$E9,$9D,$BE,$63,$07
	.dc.b	$9A,$7E,$3B,$CA,$37,$3E,$28,$FE,$11,$5E,$27,$E0,$C7,$73,$45,$9F
	.dc.b	$AF,$8F,$8C,$E4,$57,$D5,$3E,$13,$CF,$87,$BF,$1F,$91,$1D,$A7,$C6
	.dc.b	$70,$7E,$3C,$69,$FC,$11,$F4,$EB,$BC,$FC,$07,$CA,$F2,$8E,$87,$86
	.dc.b	$F2,$D8,$ED,$87,$BE,$3A,$CF,$B5,$E2,$78,$B1,$FA,$58,$FE,$4B,$1F
	.dc.b	$C3,$4F,$93,$1F,$80,$FF,$80,$7E,$33,$FC,$41,$8F,$F7,$96,$E1,$FE
	.dc.b	$07,$F8,$CB,$5B,$8F,$ED,$BC,$5F,$E0,$FA,$FF,$06,$B6,$3F,$BA,$F1
	.dc.b	$A7,$2D,$8C,$FC,$67,$4A,$FF,$0C,$36,$7F,$D8,$93,$D9,$46,$4D,$7C
	.dc.b	$5D,$9D,$3C,$7E,$32,$47,$1F,$8C,$B9,$7F,$83,$E7,$C6,$49,$F3,$BE
	.dc.b	$6F,$8C,$F1,$1E,$3F,$8C,$B3,$E3,$8F,$37,$C6,$62,$FF,$C3,$4F,$A7
	.dc.b	$E2,$E5,$7A,$7E,$32,$D7,$F8,$7E,$7D,$DF,$19,$8E,$2F,$F1,$43,$F3
	.dc.b	$A5,$F2,$EB,$C3,$70,$F8,$9F,$1B,$89,$91,$FE,$60,$AE,$31,$FE,$3E
	.dc.b	$AE,$11,$FE,$64,$6B,$FE,$BB,$73,$87,$F9,$07,$F2,$37,$5F,$4A,$FE
	.dc.b	$07,$87,$FA,$57,$91,$5A,$8F,$C1,$BF,$E2,$26,$F3,$FD,$D8,$7C,$75
	.dc.b	$89,$3C,$3D,$DB,$87,$61,$87,$68,$E5,$F1,$D2,$BB,$FC,$75,$BE,$17
	.dc.b	$E4,$AC,$69,$EB,$AB,$7F,$CD,$78,$FE,$77,$69,$FC,$ED,$3F,$74,$72
	.dc.b	$6A,$65,$FC,$DB,$E0,$D8,$FF,$35,$C6,$33,$FE,$A5,$1D,$2F,$E6,$D7
	.dc.b	$84,$FF,$2A,$E1,$DF,$8B,$E3,$77,$7C,$7F,$8F,$B5,$F3,$1B,$8F,$BB
	.dc.b	$89,$18,$C3,$EC,$3F,$3F,$C2,$97,$F5,$3E,$3E,$E1,$C6,$7E,$A7,$C7
	.dc.b	$E3,$0F,$C0,$3F,$5F,$E3,$7E,$C7,$C6,$DC,$7E,$EB,$70,$F1,$36,$FF
	.dc.b	$10,$D9,$FC,$36,$68,$C5,$7D,$C8,$C4,$FE,$23,$E3,$7F,$0C,$7E,$1F
	.dc.b	$E3,$2C,$64,$95,$F0,$63,$2F,$C3,$F8,$DF,$C7,$7C,$6F,$E3,$DB,$5F
	.dc.b	$94,$6C,$EE,$3F,$2F,$3F,$AF,$F9,$37,$C3,$53,$F0,$7F,$BB,$A7,$15
	.dc.b	$FE,$8C,$B5,$F0,$8F,$D8,$36,$3F,$DA,$2D,$7F,$C2,$07,$0B,$FF,$58
	.dc.b	$C6,$F5,$FE,$63,$83,$E3,$FB,$DF,$E8,$E5,$39,$C3,$F5,$F5,$FA,$D3
	.dc.b	$F5,$DF,$1F,$63,$F6,$7F,$19,$FC,$D3,$73,$49,$1F,$EA,$A8,$7C,$6F
	.dc.b	$F3,$3F,$8B,$67,$FB,$98,$C3,$9D,$25,$FF,$42,$AF,$F5,$43,$63,$97
	.dc.b	$C6,$7F,$44,$93,$49,$3F,$C5,$FE,$5D,$EE,$FF,$B5,$1B,$19,$8C,$1C
	.dc.b	$EB,$CC,$79,$F8,$2C,$F8,$FD,$5F,$01,$1F,$F6,$3C,$A3,$C9,$F1,$9E
	.dc.b	$FF,$8D,$F8,$0D,$AF,$67,$C5,$FC,$33,$E2,$7C,$5D,$8F,$81,$F0,$FE
	.dc.b	$D2,$3D,$9F,$81,$F1,$7F,$2F,$E2,$FE,$63,$67,$E4,$7C,$5F,$CF,$3E
	.dc.b	$87,$C5,$58,$F9,$9F,$13,$F2,$48,$FA,$9F,$99,$F1,$5F,$5F,$E2,$BE
	.dc.b	$C3,$67,$EA,$71,$7D,$BF,$8A,$FB,$8D,$8F,$B1,$1F,$54,$8F,$C6,$7E
	.dc.b	$C7,$C5,$7E,$0C,$E1,$D1,$FC,$BB,$91,$1F,$DE,$7F,$26,$5D,$31,$F1
	.dc.b	$75,$FF,$03,$C6,$9F,$F7,$CB,$6F,$F4,$BE,$A2,$31,$1F,$E7,$78,$C4
	.dc.b	$FF,$83,$A7,$87,$85,$E0,$AF,$E8,$F0,$F8,$AF,$04,$7F,$44,$8C,$57
	.dc.b	$8B,$E2,$FD,$07,$A7,$E5,$2C,$64,$F2,$C7,$72,$3E,$8F,$8C,$F8,$07
	.dc.b	$C1,$F8,$CB,$1E,$E8,$F8,$9F,$11,$8C,$7C,$4F,$C7,$F8,$9F,$91,$F1
	.dc.b	$16,$FE,$2B,$86,$9F,$22,$31,$5F,$D9,$47,$CA,$B2,$FE,$27,$1F,$EA
	.dc.b	$3F,$7B,$FD,$3D,$39,$F1,$BF,$D5,$31,$FE,$A1,$F1,$9D,$FF,$27,$DB
	.dc.b	$4F,$0C,$79,$C6,$31,$FE,$07,$8F,$37,$B1,$89,$DD,$75,$BE,$8F,$3C
	.dc.b	$7C,$35,$B6,$BE,$AA,$E2,$FF,$10,$C7,$CB,$9E,$B1,$F0,$65,$E3,$AE
	.dc.b	$18,$67,$0E,$31,$FC,$3B,$2F,$0B,$DC,$F3,$9C,$0B,$7D,$32,$F4,$63
	.dc.b	$CF,$09,$5F,$CE,$0E,$07,$F3,$AA,$F8,$4D,$C3,$E8,$BF,$CF,$BE,$31
	.dc.b	$EB,$BE,$5B,$FC,$69,$18,$6C,$66,$7B,$4E,$5F,$05,$B8,$FF,$C6,$71
	.dc.b	$86,$BE,$36,$BF,$80,$6B,$F6,$1E,$3A,$FC,$99,$E9,$F0,$7F,$36,$73
	.dc.b	$AF,$E2,$3D,$CF,$F6,$D3,$CE,$74,$4C,$3E,$4C,$E9,$ED,$19,$C3,$B5
	.dc.b	$F4,$9F,$C4,$D6,$DF,$CC,$C6,$5B,$5C,$6F,$C5,$14,$EA,$7D,$89,$E1
	.dc.b	$87,$F9,$96,$3F,$1B,$D3,$FC,$27,$E5,$BF,$DF,$DD,$5F,$C1,$63,$13
	.dc.b	$FC,$4A,$73,$C3,$FE,$5B,$E9,$7C,$66,$2B,$F8,$5B,$6F,$8C,$7F,$0B
	.dc.b	$BC,$57,$08,$FE,$3B,$F1,$78,$F8,$C9,$18,$21,$FC,$4E,$38,$9C,$BE
	.dc.b	$0B,$C1,$1F,$B2,$BF,$F3,$D3,$4F,$F3,$E6,$C8,$CB,$63,$31,$CA,$31
	.dc.b	$1F,$C1,$DB,$A7,$F2,$7E,$47,$E0,$3F,$DD,$2B,$FC,$39,$5E,$B7,$FC
	.dc.b	$E5,$2F,$D0,$5E,$47,$C5,$8C,$3D,$F7,$95,$E6,$7C,$A4,$70,$76,$76
	.dc.b	$F6,$C3,$FB,$A3,$FE,$10,$C7,$CD,$CE,$F8,$3F,$03,$67,$5C,$5F,$E1
	.dc.b	$6C,$BF,$5F,$5F,$2B,$E6,$FF,$4B,$87,$F4,$5B,$47,$E9,$1F,$F0,$EB
	.dc.b	$5C,$F3,$7D,$08,$D5,$BE,$A5,$C7,$D6,$A5,$F8,$8F,$DF,$4E,$9F,$EB
	.dc.b	$FF,$21,$F8,$62,$3B,$BF,$E0,$BF,$09,$E7,$F8,$BF,$01,$27,$FD,$75
	.dc.b	$92,$3F,$D7,$95,$E3,$F8,$9F,$A5,$8F,$FA,$5A,$FA,$35,$FE,$C1,$F2
	.dc.b	$1F,$96,$99,$1F,$DC,$9F,$CC,$F6,$FE,$43,$E6,$FF,$1F,$26,$91,$FF
	.dc.b	$66,$76,$7F,$75,$F9,$1F,$D0,$92,$3F,$BE,$FC,$8F,$E9,$FF,$28,$FA
	.dc.b	$3C,$BF,$EC,$76,$CF,$FA,$F1,$B6,$7D,$D1,$FE,$6D,$C3,$F9,$5B,$FE
	.dc.b	$68,$6B,$FE,$54,$3C,$DE,$2F,$E9,$A3,$31,$D7,$E4,$7C,$44,$7E,$B5
	.dc.b	$F7,$ED,$F3,$7C,$8F,$9F,$E4,$65,$6B,$24,$FE,$D3,$E7,$3F,$57,$3F
	.dc.b	$E7,$A6,$BE,$36,$BF,$E6,$86,$E1,$ED,$EB,$F8,$84,$9E,$DF,$23,$F1
	.dc.b	$FE,$47,$E4,$77,$49,$F1,$47,$F0,$4F,$2F,$FD,$09,$1F,$EF,$D6,$BF
	.dc.b	$D5,$5A,$FF,$A3,$9A,$FD,$1E,$BF,$A8,$47,$DD,$F2,$3F,$C4,$3F,$13
	.dc.b	$FE,$55,$1F,$E0,$DF,$9D,$7A,$7F,$C5,$1B,$1F,$EA,$19,$DF,$87,$FA
	.dc.b	$D7,$8C,$7F,$A7,$9A,$FD,$76,$BF,$E8,$C6,$A7,$9F,$3E,$FA,$BC,$87
	.dc.b	$87,$D5,$FC,$67,$C9,$CB,$FA,$CF,$E3,$FE,$4F,$10,$C7,$FA,$07,$37
	.dc.b	$FB,$46,$36,$7D,$D9,$FC,$B7,$C9,$C8,$FC,$03,$F9,$9F,$93,$C4,$73
	.dc.b	$CF,$F3,$53,$8C,$3A,$D7,$E7,$FE,$4E,$5F,$E2,$5F,$D1,$7C,$96,$23
	.dc.b	$91,$FA,$18,$C5,$79,$A6,$E1,$FA,$79,$B5,$FA,$79,$FD,$67,$C9,$7E
	.dc.b	$B7,$E4,$A4,$FE,$5E,$32,$D2,$1E,$0E,$AF,$D7,$7C,$8F,$F5,$BF,$CA
	.dc.b	$7F,$DA,$79,$FF,$65,$6E,$7F,$EC,$0A,$46,$2D,$18,$F0,$F9,$44,$3F
	.dc.b	$5B,$3B,$9D,$47,$FB,$9F,$E6,$FF,$A7,$3F,$E5,$9F,$9C,$7C,$7F,$F6
	.dc.b	$06,$C6,$FD,$FF,$EC,$75,$C2,$3F,$DD,$2D,$7F,$81,$35,$FF,$6B,$B5
	.dc.b	$3E,$0C,$FF,$A1,$37,$3E,$35,$E1,$F9,$1F,$89,$F3,$73,$0F,$5B,$F1
	.dc.b	$E7,$11,$F2,$F4,$4E,$31,$F9,$55,$FD,$EB,$F2,$46,$1F,$3F,$E2,$66
	.dc.b	$3F,$D1,$C4,$E9,$C3,$5F,$F3,$F3,$6F,$FA,$D3,$72,$FA,$D9,$3E,$57
	.dc.b	$E3,$FC,$6F,$DD,$F9,$59,$8F,$D1,$8C,$BF,$E5,$B3,$E2,$BF,$E0,$ED
	.dc.b	$8D,$24,$71,$21,$F0,$7E,$57,$F1,$5A,$FE,$2F,$86,$4F,$E0,$A3,$2D
	.dc.b	$21,$F2,$7E,$57,$F2,$5F,$11,$F9,$3F,$94,$93,$F8,$A8,$CB,$48,$7D
	.dc.b	$1F,$95,$FC,$D7,$CA,$FF,$54,$DC,$D2,$4F,$F9,$E7,$9F,$F3,$71,$C2
	.dc.b	$3F,$2E,$FE,$8F,$E5,$7F,$49,$F0,$7F,$A5,$DE,$57,$E7,$23,$2D,$21
	.dc.b	$F8,$3F,$95,$FD,$67,$CA,$F8,$C9,$1E,$1C,$91,$88,$FC,$53,$C1,$FD
	.dc.b	$4F,$CA,$FF,$90,$BE,$3D,$E8,$FF,$1B,$9F,$F0,$17,$37,$F6,$13,$F2
	.dc.b	$28,$C9,$5C,$B8,$13,$D3,$AB,$87,$C0,$FF,$8C,$27,$37,$DF,$E7,$31
	.dc.b	$87,$26,$F7,$FF,$71,$78,$63,$CE,$D1,$8B,$E5,$7E,$7B,$E0,$DA,$F4
	.dc.b	$56,$F1,$B3,$FD,$F7,$0E,$93,$E0,$AC,$9E,$DB,$E6,$FC,$19,$DD,$71
	.dc.b	$C3,$E4,$63,$C6,$BE,$14,$FC,$48,$F0,$47,$C5,$AE,$2F,$53,$B8,$9F
	.dc.b	$84,$75,$C7,$33,$C2,$75,$87,$91,$A6,$B0,$F8,$6F,$9D,$F2,$DF,$C3
	.dc.b	$8F,$2C,$FF,$87,$70,$F2,$9E,$6C,$7D,$44,$AF,$15,$7D,$06,$C6,$63
	.dc.b	$D4,$64,$F0,$63,$E3,$3F,$03,$3E,$1A,$F8,$C7,$81,$E6,$79,$6B,$A5
	.dc.b	$7E,$1A,$B9,$3D,$4F,$79,$67,$E3,$BE,$03,$58,$FC,$78,$CC,$70,$AF
	.dc.b	$74,$7B,$2F,$C6,$F9,$4C,$1E,$49,$F6,$CF,$92,$BE,$91,$F9,$3A,$E3
	.dc.b	$3F,$09,$F6,$3E,$37,$2F,$A1,$CB,$ED,$1F,$0B,$FE,$54,$70,$D9,$FD
	.dc.b	$03,$72,$FD,$B3,$B9,$FD,$A7,$C8,$7E,$D6,$31,$08,$C5,$FF,$96,$E3
	.dc.b	$67,$81,$C3,$B3,$06,$E3,$EC,$BF,$EA,$7F,$61,$DF,$88,$FF,$33,$4D
	.dc.b	$D3,$FD,$55,$39,$9E,$F7,$FC,$06,$73,$1F,$C1,$CC,$E9,$FE,$EA,$C3
	.dc.b	$AC,$64,$F2,$69,$C6,$32,$FF,$0B,$8F,$5B,$9C,$3F,$57,$1D,$0F,$43
	.dc.b	$E3,$7D,$D8,$6D,$ED,$8F,$EC,$B1,$F2,$DF,$59,$F0,$E3,$E5,$C3,$AC
	.dc.b	$60,$EA,$7E,$54,$F1,$C6,$DC,$1B,$3F,$C9,$18,$FA,$A3,$E8,$E1,$E1
	.dc.b	$8C,$4F,$D1,$3F,$96,$61,$F2,$31,$F2,$57,$CF,$AF,$8F,$18,$AF,$D3
	.dc.b	$57,$AA,$FA,$C6,$2B,$AB,$F6,$0C,$61,$EC,$C3,$C4,$D8,$F0,$CF,$6A
	.dc.b	$E7,$87,$C9,$3C,$18,$FC,$B2,$C7,$C1,$6D,$7C,$66,$BF,$0A,$FB,$57
	.dc.b	$3A,$EA,$7C,$97,$78,$FA,$EB,$E2,$5F,$7A,$F6,$B6,$33,$1E,$13,$25
	.dc.b	$72,$78,$DE,$EF,$DA,$AF,$94,$78,$DF,$F2,$9F,$03,$E5,$BF,$2B,$87
	.dc.b	$F9,$6D,$B8,$FF,$96,$FD,$BF,$D2,$BF,$E5,$68,$CD,$F2,$7F,$3A,$FF
	.dc.b	$B3,$7A,$3F,$CC,$31,$88,$FF,$17,$B5,$E1,$3F,$E0,$DA,$7C,$B7,$4B
	.dc.b	$FF,$64,$B6,$1F,$2D,$62,$57,$FB,$52,$5E,$D6,$30,$FF,$3C,$46,$27
	.dc.b	$5F,$27,$CC,$E9,$F2,$D6,$32,$78,$E3,$CD,$97,$FB,$69,$B4,$6F,$AD
	.dc.b	$E5,$3C,$DE,$D5,$BF,$24,$7A,$7E,$57,$19,$79,$1F,$67,$CB,$4A,$F5
	.dc.b	$7C,$95,$AF,$4C,$FC,$1F,$91,$C7,$17,$F0,$AD,$7F,$DD,$B1,$F4,$AB
	.dc.b	$F9,$27,$CA,$5E,$93,$F2,$4C,$13,$2F,$F2,$D4,$69,$B1,$F5,$67,$9C
	.dc.b	$60,$F9,$DF,$23,$C6,$3E,$6B,$E1,$F9,$17,$AF,$4F,$B1,$F2,$96,$FF
	.dc.b	$CF,$13,$A7,$EC,$39,$BF,$AE,$FD,$FF,$8F,$93,$F7,$7E,$3E,$D7,$DC
	.dc.b	$9B,$B7,$C3,$E5,$56,$2F,$FC,$79,$6F,$C0,$5F,$17,$F1,$DB,$3F,$E4
	.dc.b	$1C,$3F,$43,$E5,$7F,$92,$23,$FD,$1D,$3A,$BF,$F6,$89,$CE,$FE,$34
	.dc.b	$66,$3F,$A9,$35,$F0,$B5,$FF,$27,$3C,$EF,$BF,$C4,$F8,$EB,$F5,$71
	.dc.b	$FE,$8A,$8F,$D0,$7C,$2C,$C3,$F3,$DF,$21,$71,$FC,$EC,$DB,$F8,$53
	.dc.b	$6B,$F7,$1F,$21,$FA,$93,$F5,$9F,$95,$7F,$74,$8F,$D4,$46,$5F,$84
	.dc.b	$7F,$B6,$61,$E2,$67,$1F,$F7,$14,$62,$77,$1F,$ED,$C8,$7C,$87,$63
	.dc.b	$BF,$C8,$5A,$DF,$52,$B7,$18,$6D,$F6,$AE,$0F,$82,$EE,$7F,$A1,$A5
	.dc.b	$62,$FD,$1C,$5E,$E3,$DF,$BD,$8E,$D1,$71,$F5,$CD,$C3,$A7,$93,$F1
	.dc.b	$8F,$8F,$F1,$F7,$0F,$34,$5B,$F8,$73,$6B,$F9,$0F,$C1,$3D,$7F,$DA
	.dc.b	$37,$0F,$8F,$2F,$85,$4E,$AA,$57,$FB,$36,$38,$61,$FE,$1A,$6B,$FE
	.dc.b	$1F,$6B,$F9,$E7,$77,$FC,$EA,$FF,$DD,$B5,$FE,$E7,$9C,$BF,$9F,$8C
	.dc.b	$D7,$DC,$F3,$FF,$0A,$7E,$1B,$96,$D7,$FB,$BF,$EF,$9F,$19,$FC,$15
	.dc.b	$B1,$FE,$E3,$6D,$FF,$8C,$1A,$FF,$43,$E1,$7C,$87,$E4,$BE,$43,$F2
	.dc.b	$6D,$BF,$C1,$61,$F9,$63,$F2,$FF,$1B,$63,$FE,$00,$7C,$D7,$FB,$09
	.dc.b	$FF,$79,$37,$1F,$F1,$FB,$5F,$F2,$23,$5F,$CE,$7C,$77,$E8,$BE,$3B
	.dc.b	$F4,$6D,$AF,$C9,$7C,$77,$E9,$8F,$D3,$F0,$58,$FC,$FD,$FF,$52,$E4
	.dc.b	$FE,$4D,$E9,$9A,$49,$FF,$74,$6E,$F8,$EF,$D8,$9F,$B2,$F1,$EC,$FE
	.dc.b	$87,$E3,$BF,$6A,$7F,$61,$FC,$AF,$FD,$A9,$19,$3C,$38,$FF,$15,$CB
	.dc.b	$FC,$BA,$D7,$FC,$C4,$D7,$F9,$0B,$57,$7A,$BF,$23,$5F,$B4,$6C,$BA
	.dc.b	$7F,$A7,$47,$E1,$9B,$D3,$FD,$3F,$EF,$1E,$85,$3D,$1F,$1F,$71,$F8
	.dc.b	$FF,$5E,$F8,$CF,$E9,$8D,$8F,$F3,$AE,$68,$C3,$FE,$78,$6B,$FE,$0B
	.dc.b	$CD,$F2,$1F,$14,$FF,$08,$FC,$C3,$E3,$5E,$AE,$71,$FA,$FA,$BD,$2F
	.dc.b	$EC,$D9,$EE,$73,$18,$B9,$1E,$49,$D6,$7E,$39,$F4,$9F,$E1,$9C,$7C
	.dc.b	$1E,$97,$8A,$0F,$29,$F9,$56,$55,$C2,$3E,$2F,$C3,$F9,$0F,$EF,$5F
	.dc.b	$88,$B8,$31,$3A,$3C,$93,$F2,$88,$E0,$FE,$67,$4F,$CE,$DF,$25,$CB
	.dc.b	$1F,$19,$E2,$8C,$D7,$D6,$F8,$AF,$8A,$78,$BE,$16,$C6,$4F,$89,$1F
	.dc.b	$68,$8F,$FA,$BB,$A3,$F1,$98,$7E,$64,$FC,$D5,$59,$FC,$74,$7E,$77
	.dc.b	$97,$18,$F9,$2F,$F2,$91,$FA,$1D,$FF,$45,$F2,$37,$0F,$CB,$38,$69
	.dc.b	$F9,$98,$C5,$FF,$76,$F9,$FF,$F2,$1F,$4F,$FB,$2B,$D1,$FF,$01,$C6
	.dc.b	$27,$FD,$95,$1F,$E0,$6E,$1F,$E5,$FE,$EF,$F4,$C9,$FB,$15,$B8,$C4
	.dc.b	$23,$15,$D3,$E1,$3F,$9D,$CF,$43,$F0,$17,$D2,$75,$87,$8E,$7A,$D7
	.dc.b	$0C,$3A,$E1,$D2,$78,$FA,$1F,$EC,$89,$CD,$6F,$0E,$75,$DC,$F2,$46
	.dc.b	$5E,$39,$F7,$73,$8D,$C7,$B4,$60,$E5,$97,$9D,$F2,$CE,$EF,$DB,$1F
	.dc.b	$EE,$FA,$F9,$4D,$8C,$BE,$43,$FD,$E9,$F1,$4F,$43,$F4,$31,$FE,$F4
	.dc.b	$8C,$72,$FF,$02,$6C,$66,$76,$F7,$F8,$47,$C5,$7F,$08,$C7,$FC,$E4
	.dc.b	$D8,$CC,$FF,$C0,$09,$CB,$63,$3B,$7C,$E6,$ED,$F8,$46,$CF,$E0,$3D
	.dc.b	$BF,$F8,$03,$39,$C3,$E4,$56,$AB,$F0,$4D,$C1,$7C,$27,$F1,$33,$92
	.dc.b	$57,$D1,$9D,$4F,$28,$EA,$47,$EE,$57,$47,$18,$7A,$66,$E1,$F9,$CB
	.dc.b	$F0,$72,$7F,$3D,$F9,$2F,$F3,$0F,$89,$FC,$EB,$8E,$F1,$FF,$13,$AF
	.dc.b	$E8,$51,$8C,$FF,$A2,$F8,$6C,$31,$5C,$E7,$FA,$35,$62,$BF,$C3,$D1
	.dc.b	$FD,$26,$B1,$F2,$59,$F9,$29,$3F,$D3,$63,$2F,$F0,$7A,$47,$F4,$2C
	.dc.b	$3F,$AA,$36,$33,$3F,$E7,$A3,$F9,$57,$1B,$E7,$B3,$19,$8C,$57,$F4
	.dc.b	$C6,$C7,$13,$F9,$8C,$F0,$7F,$D0,$B1,$A7,$F9,$2B,$74,$69,$E0,$AE
	.dc.b	$F1,$FD,$93,$E4,$FE,$45,$78,$FE,$4F,$E2,$13,$0F,$E6,$76,$8D,$B6
	.dc.b	$3C,$93,$C5,$B1,$97,$D8,$D7,$CE,$E6,$FC,$B1,$EC,$AE,$A7,$D5,$F9
	.dc.b	$3F,$AD,$F0,$92,$7F,$BF,$3E,$8E,$7F,$F5,$1C,$7F,$48,$6B,$F2,$DA
	.dc.b	$FC,$66,$BF,$49,$AF,$E2,$3C,$3E,$9E,$F4,$8F,$F0,$64,$7E,$1F,$C3
	.dc.b	$91,$FE,$0F,$7F,$15,$F2,$7E,$3B,$FB,$07,$F0,$79,$F8,$AD,$7C,$8F
	.dc.b	$E1,$7E,$43,$F2,$BF,$96,$FC,$29,$FE,$94,$DD,$7F,$C3,$33,$AF,$91
	.dc.b	$95,$FE,$92,$A4,$72,$F4,$79,$F8,$3F,$93,$93,$49,$5F,$D7,$E1,$E0
	.dc.b	$78,$BE,$46,$4F,$58,$F2,$FC,$84,$8F,$03,$E8,$F9,$1F,$4F,$C8,$F7
	.dc.b	$24,$F5,$7C,$5F,$23,$EE,$F5,$A4,$FA,$63,$E0,$FA,$BF,$0B,$E4,$64
	.dc.b	$FB,$63,$F7,$8B,$F1,$BF,$F5,$AB,$FE,$30,$6B,$FA,$56,$BF,$AF,$6B
	.dc.b	$FE,$EA,$F9,$CF,$F1,$FB,$FE,$38,$6B,$FA,$D6,$BF,$B7,$6B,$FC,$3F
	.dc.b	$67,$C9,$FD,$8F,$62,$4F,$D3,$8F,$3E,$9F,$E6,$78,$FF,$22,$B5,$FD
	.dc.b	$D3,$5F,$E0,$8D,$7B,$7A,$FF,$84,$23,$F7,$FE,$4F,$E1,$FC,$DB,$5F
	.dc.b	$8B,$F3,$9F,$E4,$8A,$FF,$3E,$B5,$FF,$7F,$CE,$DF,$F4,$0B,$5F,$E1
	.dc.b	$8D,$7F,$8B,$35,$FF,$20,$07,$0F,$F9,$0D,$E5,$D3,$F9,$96,$F8,$BC
	.dc.b	$1B,$3F,$E4,$9A,$F9,$F5,$FC,$03,$C8,$6C,$F9,$5F,$D1,$92,$FF,$3D
	.dc.b	$E6,$D7,$F4,$B1,$6F,$95,$7F,$4D,$9F,$D4,$FC,$A7,$EB,$3E,$52,$47
	.dc.b	$E9,$E3,$F5,$E4,$70,$7F,$BF,$3A,$BF,$B8,$BF,$E3,$B5,$F2,$9F,$D2
	.dc.b	$9F,$C7,$E1,$E9,$6E,$DF,$D4,$DD,$D7,$3A,$FE,$89,$18,$8F,$E8,$5A
	.dc.b	$7E,$C2,$72,$48,$E0,$43,$C5,$E8,$F8,$7E,$4F,$C5,$E3,$C9,$F4,$46
	.dc.b	$5A,$43,$D5,$F2,$7E,$9F,$88,$F5,$7C,$9C,$9F,$81,$19,$69,$0F,$85
	.dc.b	$E3,$FE,$CB,$EA,$7F,$DD,$31,$FE,$CA,$F9,$C7,$99,$FB,$66,$C7,$FB
	.dc.b	$3A,$76,$FF,$9E,$03,$87,$FB,$23,$5F,$C7,$B5,$FF,$48,$35,$E3,$F2
	.dc.b	$5E,$32,$3E,$3F,$3F,$EB,$7C,$9F,$D7,$F9,$49,$3F,$76,$32,$D2,$1F
	.dc.b	$7E,$BE,$F7,$CA,$7D,$FF,$92,$93,$F8,$58,$CB,$48,$7E,$1F,$E4,$FF
	.dc.b	$8B,$FD,$47,$F8,$46,$3F,$19,$F2,$9F,$89,$33,$EF,$48,$FF,$64,$CE
	.dc.b	$DF,$EF,$CD,$7F,$C1,$0D,$7F,$66,$EF,$AF,$FA,$43,$7A,$F3,$5F,$97
	.dc.b	$F9,$4F,$CC,$7C,$9F,$E6,$7E,$4E,$63,$F9,$F8,$CB,$48,$7E,$87,$E4
	.dc.b	$FF,$C5,$3F,$80,$FF,$86,$73,$FD,$C3,$FE,$6A,$6F,$C4,$EA,$A1,$FE
	.dc.b	$C7,$8D,$B8,$3F,$D4,$D2,$F9,$CE,$BF,$17,$FE,$29,$8F,$F6,$44,$FF
	.dc.b	$AC,$DA,$FE,$A1,$AF,$F1,$4F,$09,$1A,$BF,$2D,$74,$AF,$16,$DF,$E4
	.dc.b	$7A,$FF,$78,$C6,$05,$F2,$F8,$FF,$7F,$C8,$49,$D4,$E5,$B5,$FE,$D4
	.dc.b	$6D,$7C,$63,$E3,$FC,$87,$C8,$F8,$FF,$61,$27,$D8,$D7,$E1,$BB,$7A
	.dc.b	$FC,$87,$F6,$0F,$90,$78,$6D,$BF,$DE,$5E,$43,$5F,$E6,$A7,$CB,$C3
	.dc.b	$2E,$38,$3F,$9C,$CE,$E7,$FD,$F5,$87,$4B,$DC,$63,$0E,$CB,$E7,$38
	.dc.b	$3D,$F0,$FE,$87,$18,$C7,$C1,$3B,$7F,$A3,$E1,$D6,$30,$75,$C0,$E0
	.dc.b	$E8,$62,$BA,$5F,$6B,$E7,$87,$92,$7C,$78,$75,$77,$7C,$6B,$E0,$34
	.dc.b	$F0,$1D,$E7,$C9,$8F,$08,$FE,$7B,$3C,$1A,$70,$8E,$F7,$E1,$9D,$5F
	.dc.b	$13,$38,$F7,$B6,$1E,$5A,$D9,$23,$C7,$5C,$1E,$E7,$C2,$2C,$7C,$CC
	.dc.b	$34,$6E,$7C,$11,$CE,$7B,$E1,$E5,$7A,$C6,$4F,$AB,$3C,$5F,$97,$87
	.dc.b	$E0,$27,$C1,$7F,$0C,$EF,$3A,$8F,$AC,$F2,$3C,$18,$7B,$63,$B9,$F7
	.dc.b	$23,$4D,$C7,$ED,$3A,$3C,$F7,$F1,$E7,$71,$F0,$70,$F3,$3F,$46,$3E
	.dc.b	$A6,$1E,$57,$C8,$EA,$38,$47,$93,$1F,$85,$1D,$8F,$97,$87,$E8,$A7
	.dc.b	$99,$C8,$91,$E7,$3E,$23,$F4,$4C,$BC,$FE,$3F,$FB,$D6,$5F,$9D,$6D
	.dc.b	$FE,$A1,$B9,$7E,$79,$FF,$00,$57,$E8,$3E,$43,$19,$7F,$AE,$DB,$1F
	.dc.b	$EC,$56,$BF,$E3,$06,$BC,$4F,$C3,$D3,$C8,$FF,$54,$D7,$ED,$A7,$4F
	.dc.b	$EE,$27,$5C,$7F,$7E,$73,$9F,$FB,$B2,$F8,$BE,$38,$EB,$18,$7C,$87
	.dc.b	$ED,$EB,$9B,$FB,$93,$C9,$AF,$79,$FF,$5D,$BE,$38,$63,$F8,$FA,$E3
	.dc.b	$8E,$63,$17,$DE,$BE,$01,$DD,$E7,$87,$97,$0E,$D8,$FE,$BE,$BC,$33
	.dc.b	$EB,$8C,$47,$80,$E4,$E2,$FF,$6D,$8F,$AA,$FC,$55,$A3,$D8,$71,$C7
	.dc.b	$BB,$FB,$63,$E4,$9F,$D7,$71,$F4,$3E,$23,$18,$77,$7A,$57,$63,$99
	.dc.b	$71,$F4,$47,$83,$1F,$87,$5F,$5E,$3C,$53,$AC,$38,$47,$63,$81,$DA
	.dc.b	$3E,$31,$D2,$F7,$1F,$D2,$70,$F9,$6F,$DC,$AF,$69,$F0,$8B,$5C,$A3
	.dc.b	$B5,$72,$C7,$E4,$4E,$67,$A6,$1E,$FA,$F6,$CE,$6B,$75,$F8,$A9,$CB
	.dc.b	$E1,$9F,$61,$F9,$1B,$F0,$C7,$F7,$4E,$27,$53,$1F,$EE,$2D,$9F,$F0
	.dc.b	$5F,$C8,$5C,$BF,$C2,$B3,$6B,$FC,$00,$FF,$BB,$6A,$C7,$F8,$36,$31
	.dc.b	$1D,$DB,$CB,$FE,$0C,$F8,$EE,$53,$E3,$6C,$7F,$84,$3E,$42,$E4,$9A
	.dc.b	$63,$E3,$EE,$3F,$E0,$87,$7F,$1B,$23,$FC,$17,$E4,$5C,$BF,$C4,$B3
	.dc.b	$E2,$C0,$D2,$57,$F8,$97,$E3,$7C,$27,$8B,$CC,$B1,$FE,$16,$8F,$2F
	.dc.b	$C6,$E3,$0F,$F1,$8B,$E8,$F8,$D9,$5F,$E1,$CF,$8D,$B5,$FE,$36,$9F
	.dc.b	$7F,$E1,$9F,$46,$BF,$C4,$31,$87,$E2,$CB,$EA,$5E,$C6,$1B,$3F,$E2
	.dc.b	$AB,$7C,$7E,$6A,$4F,$FB,$3A,$FF,$D9,$F7,$C1,$C1,$D6,$7E,$BB,$67
	.dc.b	$2D,$3F,$C4,$CF,$F8,$3B,$8D,$83,$E3,$FB,$D6,$8F,$F2,$0C,$66,$BC
	.dc.b	$DF,$1F,$F0,$0F,$83,$ED,$D8,$C9,$EB,$F8,$FC,$5F,$3C,$7B,$C7,$C8
	.dc.b	$F8,$F9,$5F,$07,$E3,$ED,$EF,$25,$7C,$EF,$88,$CF,$67,$FB,$96,$51
	.dc.b	$F3,$BE,$75,$E8,$D3,$F2,$B4,$E6,$79,$F9,$D3,$FC,$BD,$9B,$0E,$31
	.dc.b	$F2,$1B,$1F,$39,$AF,$D3,$6B,$FE,$38,$39,$61,$E0,$FD,$DB,$FC,$77
	.dc.b	$5D,$AB,$9E,$3F,$B3,$9C,$DF,$BF,$E4,$3F,$16,$7E,$33,$E4,$6E,$1F
	.dc.b	$85,$8F,$C7,$FC,$8E,$2B,$E3,$3F,$92,$F8,$B9,$5F,$8B,$F8,$EB,$5F
	.dc.b	$2E,$7F,$2F,$F1,$F8,$CF,$E5,$3F,$9A,$F8,$F9,$87,$E4,$FE,$3E,$E1
	.dc.b	$F4,$E7,$F8,$0F,$19,$A4,$AF,$C9,$F3,$FE,$32,$3F,$4B,$F1,$9C,$B2
	.dc.b	$FB,$0F,$EA,$3E,$36,$5F,$E8,$7B,$6E,$5F,$81,$9F,$D6,$FC,$66,$30
	.dc.b	$FC,$03,$FC,$3B,$F2,$F7,$A7,$4C,$79,$F7,$2F,$F4,$AC,$E9,$B3,$A6
	.dc.b	$9F,$82,$F7,$3F,$0C,$DE,$BF,$C4,$F5,$3D,$AF,$DD,$D6,$B5,$FF,$7C
	.dc.b	$35,$FD,$B3,$5F,$C2,$35,$FF,$3E,$B5,$3C,$F4,$FE,$6D,$4F,$6B,$BE
	.dc.b	$1E,$27,$E0,$7C,$FC,$CB,$DB,$F1,$D7,$2F,$D7,$6A,$F8,$DF,$E6,$ED
	.dc.b	$AF,$AA,$D9,$FE,$19,$CA,$8C,$47,$FB,$25,$7C,$77,$CF,$F6,$7E,$83
	.dc.b	$6F,$F5,$3F,$3C,$F3,$FF,$88,$B6,$3F,$88,$B7,$0F,$E2,$ED,$7F,$51
	.dc.b	$C0,$FC,$F9,$CF,$FD,$42,$30,$D8,$CD,$FF,$20,$9C,$BE,$CF,$90,$FC
	.dc.b	$27,$BD,$F8,$56,$C7,$E0,$BE,$43,$F1,$1F,$1B,$F8,$96,$C7,$E1,$BE
	.dc.b	$3B,$E3,$91,$F8,$EF,$CA,$F8,$EF,$C8,$1F,$91,$F8,$DB,$3F,$8E,$F8
	.dc.b	$EF,$CA,$7C,$77,$E5,$5B,$1F,$92,$F8,$EF,$A6,$47,$E9,$BF,$5B,$B3
	.dc.b	$F3,$87,$E7,$7E,$32,$CF,$E6,$BE,$37,$F4,$1E,$8F,$E8,$5B,$1F,$EF
	.dc.b	$7F,$97,$FC,$47,$A1,$FD,$E1,$B1,$FD,$19,$BC,$DF,$D2,$DA,$FF,$19
	.dc.b	$EC,$FF,$7C,$4F,$FB,$F3,$E3,$7F,$8A,$F3,$7F,$82,$E5,$1F,$98,$9D
	.dc.b	$57,$F8,$02,$1F,$19,$FE,$04,$C3,$F3,$5E,$76,$63,$1E,$1A,$29,$E7
	.dc.b	$71,$87,$9D,$B8,$79,$DA,$F9,$F0,$7C,$77,$84,$E7,$5E,$97,$C5,$F3
	.dc.b	$4F,$9E,$A7,$9C,$B3,$F2,$63,$4F,$C7,$6D,$FC,$76,$BF,$CF,$F3,$46
	.dc.b	$23,$FD,$F3,$18,$AF,$89,$87,$69,$FE,$C4,$FC,$AB,$F0,$79,$3E,$63
	.dc.b	$E5,$7C,$65,$8C,$9E,$38,$FB,$98,$FF,$13,$8C,$1F,$CC,$2B,$17,$FB
	.dc.b	$C6,$BF,$1E,$BF,$C7,$31,$88,$FF,$01,$3C,$37,$FB,$CF,$F1,$1A,$FC
	.dc.b	$15,$7D,$DD,$3E,$F7,$C8,$7D,$FF,$8C,$B5,$F8,$76,$E3,$EE,$6E,$3D
	.dc.b	$1F,$F1,$1F,$CD,$BE,$07,$F9,$0D,$7F,$A8,$BC,$BF,$ED,$53,$9F,$3B
	.dc.b	$FD,$89,$1F,$DB,$78,$9F,$1D,$FE,$5D,$F9,$04,$F0,$C7,$9C,$63,$10
	.dc.b	$C7,$A3,$DC,$CB,$88,$E1,$1F,$8C,$BE,$D7,$D7,$1F,$1C,$F0,$C3,$BC
	.dc.b	$E6,$3A,$1C,$74,$DB,$E1,$9D,$BC,$6B,$38,$F8,$6F,$DD,$87,$82,$B4
	.dc.b	$74,$9D,$38,$C7,$CF,$97,$C4,$CB,$9E,$5D,$AB,$C1,$97,$F9,$0B,$0F
	.dc.b	$83,$1F,$E4,$B6,$BE,$93,$FC,$9D,$F2,$2F,$11,$CD,$FE,$4A,$8C,$36
	.dc.b	$33,$3F,$E5,$29,$CB,$FE,$54,$A7,$C2,$7F,$76,$8C,$35,$F1,$35,$FC
	.dc.b	$0B,$5F,$AA,$F1,$D3,$EA,$CF,$4F,$4F,$E4,$CE,$77,$FC,$4F,$97,$FE
	.dc.b	$33,$BD,$CE,$70,$E7,$7C,$0F,$A3,$5B,$BF,$0C,$66,$7E,$85,$71,$7F
	.dc.b	$15,$5B,$79,$9A,$9E,$17,$E6,$9D,$4F,$E4,$6B,$96,$1E,$98,$E0,$7D
	.dc.b	$D9,$E9,$D9,$FC,$BB,$87,$FB,$2F,$07,$ED,$9B,$49,$D3,$FC,$DF,$E3
	.dc.b	$F3,$1F,$CD,$9B,$D1,$FA,$89,$FE,$79,$58,$9C,$FC,$86,$BE,$42,$57
	.dc.b	$F4,$18,$CB,$48,$7F,$1A,$F8,$EF,$1F,$1B,$F1,$0F,$F3,$6C,$FF,$4F
	.dc.b	$6B,$FD,$3D,$AF,$F2,$96,$BF,$CA,$1D,$F4,$FF,$B6,$67,$A7,$89,$DA
	.dc.b	$73,$B7,$F1,$8F,$91,$F0,$FD,$4F,$F5,$0C,$7F,$AC,$7B,$5F,$00,$CF
	.dc.b	$8B,$23,$FB,$14,$ED,$FE,$C4,$D7,$F9,$E3,$5F,$E7,$6E,$F3,$F2,$37
	.dc.b	$0C,$D7,$CE,$F9,$1F,$9F,$EB,$4C,$3F,$B5,$3F,$1B,$EA,$3F,$D2,$F1
	.dc.b	$FD,$DD,$AF,$F7,$36,$BF,$D3,$5A,$FF,$4C,$6B,$F8,$1E,$47,$C8,$F5
	.dc.b	$F8,$B9,$3F,$DE,$63,$F0,$3F,$25,$F8,$4F,$9A,$69,$4C,$D7,$8F,$E2
	.dc.b	$3F,$D8,$FC,$8C,$D1,$86,$CF,$F8,$2D,$AF,$F9,$2A,$FF,$DB,$2F,$F3
	.dc.b	$0E,$0F,$F0,$44,$62,$3F,$D8,$4D,$8C,$FC,$8F,$97,$C5,$F3,$77,$CB
	.dc.b	$E5,$9F,$FB,$5F,$0F,$E9,$11,$83,$F4,$4B,$E4,$7F,$DB,$FD,$C8,$C4
	.dc.b	$FF,$A5,$1A,$F9,$E7,$71,$FE,$CE,$8E,$93,$C7,$E4,$3E,$41,$F2,$7E
	.dc.b	$47,$E5,$71,$7C,$BF,$8F,$97,$C3,$0F,$B7,$B7,$C0,$F4,$FA,$92,$7C
	.dc.b	$4F,$83,$E3,$FE,$A1,$23,$97,$17,$D7,$F8,$F9,$1D,$5F,$B5,$EA,$FF
	.dc.b	$02,$87,$A1,$FE,$22,$7F,$DF,$ED,$7F,$DA,$87,$F6,$27,$0E,$1A,$FC
	.dc.b	$DF,$8D,$FC,$31,$1F,$67,$C7,$F7,$F9,$B6,$7D,$27,$FD,$C1,$18,$70
	.dc.b	$1C,$3B,$6B,$E4,$F9,$CF,$E7,$CF,$F7,$66,$A6,$9A,$E3,$FB,$9F,$21
	.dc.b	$F9,$B2,$4F,$D1,$F3,$3F,$3D,$F2,$1F,$9F,$F9,$09,$3F,$56,$32,$FE
	.dc.b	$0D,$FC,$09,$1C,$1B,$20,$E2,$FF,$31,$7C,$CF,$F5,$96,$FC,$8F,$F3
	.dc.b	$08,$7C,$8D,$31,$12,$BF,$1B,$3A,$6C,$E5,$ED,$D9,$FE,$18,$D1,$5F
	.dc.b	$E1,$4C,$3C,$3F,$23,$DC,$F0,$7A,$D3,$0E,$B1,$97,$BC,$79,$7E,$3F
	.dc.b	$CD,$F2,$12,$7C,$71,$97,$FD,$F3,$87,$FB,$01,$B1,$FA,$C8,$C4,$7F
	.dc.b	$B2,$1B,$87,$C4,$9E,$53,$EA,$3F,$AB,$4B,$E3,$7E,$37,$C7,$7C,$7F
	.dc.b	$72,$57,$C4,$8C,$B4,$87,$3D,$3E,$1F,$C7,$7C,$FF,$8E,$93,$F3,$1F
	.dc.b	$A5,$F1,$D8,$8E,$19,$27,$1A,$7D,$AA,$FD,$54,$FF,$87,$EB,$F8,$60
	.dc.b	$F9,$A7,$9B,$FA,$F6,$C6,$A7,$83,$EF,$3E,$BE,$9F,$A2,$69,$FA,$48
	.dc.b	$DB,$FB,$46,$BF,$E5,$63,$67,$37,$AE,$1F,$CE,$23,$FC,$B7,$18,$E2
	.dc.b	$E1,$7F,$D0,$A7,$F1,$BF,$1F,$F9,$3F,$8E,$91,$82,$1F,$11,$CB,$F1
	.dc.b	$A3,$F3,$5F,$1F,$F9,$BF,$8F,$93,$F9,$88,$CB,$48,$7C,$EF,$8F,$FE
	.dc.b	$29,$F6,$BF,$DD,$A3,$F5,$1F,$20,$F8,$1F,$F0,$CE,$C8,$C4,$FD,$0A
	.dc.b	$E0,$9F,$F0,$2B,$FE,$7A,$6B,$9F,$7F,$E7,$46,$0E,$24,$7B,$CB,$DA
	.dc.b	$FD,$B3,$FB,$7F,$90,$97,$FB,$48,$CB,$48,$7E,$2A,$BF,$C5,$BF,$4D
	.dc.b	$FC,$B3,$1F,$F7,$23,$FE,$9A,$9F,$CB,$36,$33,$1C,$08,$7F,$4E,$8E
	.dc.b	$5C,$3F,$BF,$7F,$C1,$91,$FE,$3D,$7F,$9A,$B7,$E2,$FF,$AE,$C6,$3A
	.dc.b	$4D,$3F,$A0,$5F,$92,$7C,$E7,$A3,$E4,$7D,$3E,$EF,$AB,$E4,$66,$3E
	.dc.b	$38,$CB,$48,$7B,$3E,$43,$E0,$FC,$87,$88,$91,$FD,$0F,$3F,$1C,$62
	.dc.b	$3E,$1B,$C1,$F8,$1E,$87,$F5,$9F,$8E,$78,$3F,$CB,$65,$E4,$33,$FD
	.dc.b	$21,$FE,$37,$71,$1F,$EB,$C8,$C1,$FA,$9A,$E7,$B7,$FB,$02,$B7,$49
	.dc.b	$CA,$F9,$BC,$DF,$F9,$0D,$BF,$09,$FB,$37,$8D,$70,$6E,$3E,$77,$77
	.dc.b	$FE,$CE,$6B,$D6,$FC,$93,$B6,$E3,$B7,$DB,$87,$76,$D6,$4F,$F6,$CE
	.dc.b	$1E,$18,$E3,$1F,$19,$DB,$4F,$89,$3F,$C2,$F1,$F3,$DF,$C3,$AE,$38
	.dc.b	$F9,$88,$E0,$97,$C8,$F1,$1F,$4E,$FC,$CF,$93,$1F,$63,$6F,$B3,$C4
	.dc.b	$F2,$1C,$4C,$9F,$4E,$3F,$01,$87,$9A,$3E,$41,$C7,$0F,$2B,$D4,$C9
	.dc.b	$C6,$BC,$B8,$7D,$93,$E1,$CF,$33,$A5,$7D,$42,$47,$CA,$74,$F0,$25
	.dc.b	$77,$9C,$BE,$38,$CB,$E1,$3C,$78,$FD,$57,$44,$35,$3E,$29,$FA,$F5
	.dc.b	$E2,$9F,$63,$D9,$F8,$C7,$BA,$3D,$8F,$8D,$F4,$1C,$88,$F2,$72,$FD
	.dc.b	$18,$EC,$F9,$DF,$9D,$F1,$FF,$AE,$8F,$98,$D7,$D8,$DB,$FD,$3B,$98
	.dc.b	$0F,$8F,$B1,$98,$C0,$71,$1A,$AF,$F2,$43,$67,$FC,$AA,$FE,$13,$17
	.dc.b	$93,$FD,$7E,$BF,$73,$5E,$0A,$D4,$E7,$1E,$73,$F8,$EB,$E3,$CB,$F8
	.dc.b	$E8,$E0,$66,$B8,$3E,$03,$F1,$73,$C1,$FC,$69,$C3,$3F,$E1,$31,$97
	.dc.b	$C7,$1E,$EC,$73,$5D,$23,$FB,$D6,$3E,$6C,$3C,$51,$F1,$2F,$71,$FA
	.dc.b	$C8,$ED,$8F,$F8,$07,$1E,$CF,$B2,$B7,$3E,$43,$94,$6F,$0F,$F0,$24
	.dc.b	$F9,$30,$FD,$BD,$F2,$BF,$74,$F2,$34,$FC,$3C,$3C,$B5,$E7,$8F,$DB
	.dc.b	$9E,$43,$FD,$75,$7D,$B0,$F2,$1E,$1A,$F5,$4F,$90,$DB,$8A,$F1,$D7
	.dc.b	$19,$FA,$18,$FB,$0F,$19,$6F,$AC,$72,$3E,$05,$72,$BE,$91,$F6,$2F
	.dc.b	$A3,$F4,$0C,$46,$AF,$A4,$78,$2F,$D0,$FB,$8F,$41,$FE,$7D,$AE,$6F
	.dc.b	$C1,$31,$5E,$0C,$3E,$55,$79,$A7,$27,$98,$F0,$96,$36,$F8,$4F,$65
	.dc.b	$FC,$09,$F1,$BF,$D7,$38,$3F,$BD,$FC,$8F,$F8,$FF,$3F,$F0,$6B,$71
	.dc.b	$FF,$03,$7C,$7D,$C7,$FC,$0D,$36,$F8,$D7,$F7,$3E,$74,$FF,$92,$6B
	.dc.b	$FC,$9D,$5A,$7F,$D8,$AD,$73,$2F,$8F,$E8,$DB,$FF,$0C,$7C,$7E,$26
	.dc.b	$47,$F8,$6F,$BE,$DF,$F8,$89,$E3,$F1,$F2,$7F,$C4,$1F,$1D,$6B,$FC
	.dc.b	$4F,$3E,$57,$FC,$44,$D7,$FC,$44,$DC,$BF,$C3,$ED,$C9,$F1,$DE,$63
	.dc.b	$CF,$E0,$DC,$7F,$C5,$D1,$EA,$F8,$DC,$63,$FE,$36,$7D,$BF,$1B,$2B
	.dc.b	$FC,$69,$E6,$DA,$FF,$1D,$CF,$CD,$F4,$95,$9E,$3F,$F3,$A6,$A9,$CC
	.dc.b	$FF,$A2,$BE,$65,$EC,$7D,$7C,$51,$89,$FE,$57,$87,$DF,$FA,$E7,$C5
	.dc.b	$6E,$B5,$7F,$E8,$D6,$BF,$EF,$C6,$BF,$E8,$96,$BF,$CD,$9A,$F8,$30
	.dc.b	$FC,$45,$AD,$F1,$5D,$E7,$31,$8C,$7C,$A7,$FB,$4F,$27,$CD,$FC,$33
	.dc.b	$E2,$7C,$75,$9D,$E4,$D7,$E2,$57,$C9,$F8,$ED,$4F,$FA,$EF,$04,$7C
	.dc.b	$DF,$8E,$F9,$DB,$DC,$79,$E3,$F3,$63,$18,$71,$6E,$DF,$4B,$1F,$93
	.dc.b	$7F,$EB,$AF,$9B,$79,$BF,$45,$B1,$E6,$A7,$D7,$3D,$CD,$5C,$AF,$F5
	.dc.b	$A6,$5F,$EB,$5C,$38,$B8,$0E,$1D,$B5,$E1,$3F,$42,$FF,$DA,$55,$E5
	.dc.b	$8C,$7B,$9F,$8D,$E3,$93,$F0,$33,$7C,$86,$31,$F9,$15,$74,$F8,$13
	.dc.b	$6F,$EB,$7C,$87,$E5,$CF,$CC,$7C,$85,$8E,$91,$F9,$BF,$89,$C6,$3F
	.dc.b	$4D,$FC,$F7,$C7,$CA,$F1,$7C,$7D,$AF,$B1,$3F,$B3,$D0,$D2,$57,$FA
	.dc.b	$0E,$32,$FA,$FE,$3F,$96,$9F,$5D,$FD,$67,$C7,$FE,$B7,$E3,$ED,$7C
	.dc.b	$16,$E3,$F8,$76,$DF,$DF,$7F,$69,$F1,$FF,$B5,$F8,$DB,$7F,$25,$B7
	.dc.b	$F8,$F6,$DF,$E1,$DF,$DE,$7C,$6C,$AF,$A1,$F1,$B6,$FF,$2D,$3E,$F2
	.dc.b	$4F,$B5,$B1,$F9,$BF,$03,$86,$1F,$93,$7F,$9A,$7E,$09,$FE,$23,$BF
	.dc.b	$CD,$B6,$95,$FC,$2A,$3F,$D4,$33,$AF,$57,$8F,$C7,$7F,$21,$93,$49
	.dc.b	$3F,$93,$B7,$CD,$3C,$8F,$E5,$51,$86,$C6,$67,$83,$67,$FA,$1D,$BE
	.dc.b	$3B,$D2,$7A,$BE,$3E,$DF,$AB,$93,$C8,$47,$FD,$85,$5F,$C8,$DB,$EC
	.dc.b	$7C,$23,$E1,$FC,$7D,$8F,$87,$EE,$7C,$63,$E3,$FC,$75,$8F,$8F,$E2
	.dc.b	$7C,$12,$3D,$9F,$89,$E0,$FC,$EF,$8F,$F9,$ED,$9F,$9F,$F1,$9F,$48
	.dc.b	$FA,$7F,$1B,$63,$E9,$FC,$67,$CD,$23,$EA,$7E,$87,$C6,$FD,$AF,$8D
	.dc.b	$FB,$6D,$9F,$B7,$F1,$7F,$78,$FB,$FF,$19,$63,$F9,$1F,$2B,$E9,$9E
	.dc.b	$1F,$D4,$6C,$FD,$96,$CF,$F0,$D6,$BF,$DE,$39,$3F,$88,$F6,$7F,$50
	.dc.b	$FC,$3F,$FC,$E7,$7F,$8C,$F3,$3F,$C0,$51,$8B,$FC,$6C,$FF,$9E,$3D
	.dc.b	$04,$BE,$6B,$FA,$A6,$4F,$13,$8C,$61,$97,$E5,$7D,$04,$7F,$E0,$05
	.dc.b	$3E,$27,$C2,$70,$C7,$C2,$7F,$43,$8D,$4F,$C4,$FA,$67,$C7,$7F,$65
	.dc.b	$6C,$7F,$40,$97,$C8,$66,$A4,$FE,$7E,$B7,$7F,$CC,$DA,$FF,$38,$6B
	.dc.b	$FE,$98,$0F,$0D,$BE,$57,$C8,$74,$9F,$3E,$DF,$1A,$3C,$9F,$1F,$30
	.dc.b	$F2,$63,$6F,$DA,$74,$9F,$D7,$E2,$69,$2B,$E3,$FC,$5F,$D6,$3E,$BF
	.dc.b	$C8,$58,$F8,$11,$F6,$A7,$17,$E4,$D7,$A5,$7C,$5F,$90,$FB,$FF,$17
	.dc.b	$F8,$06,$C7,$C9,$8B,$5C,$66,$ED,$FE,$43,$F9,$BF,$EE,$3C,$7F,$EA
	.dc.b	$DF,$23,$E3,$46,$2F,$FD,$7D,$1F,$27,$89,$F1,$BF,$E2,$29,$C9,$FE
	.dc.b	$C4,$AE,$1F,$21,$FC,$26,$31,$F0,$1F,$8F,$76,$27,$78,$75,$BE,$97
	.dc.b	$C6,$B5,$7C,$63,$17,$C2,$7A,$E3,$E9,$7F,$DA,$B8,$F9,$B5,$C4,$72
	.dc.b	$9B,$5C,$63,$07,$92,$B8,$3E,$68,$D1,$30,$F1,$65,$E7,$C3,$9C,$62
	.dc.b	$BC,$67,$F8,$16,$BE,$53,$FE,$0A,$6B,$F3,$A3,$E2,$3F,$E0,$EF,$91
	.dc.b	$78,$EF,$99,$FE,$79,$18,$6C,$66,$7F,$C2,$53,$97,$C1,$F0,$5F,$EF
	.dc.b	$16,$FA,$5F,$D1,$71,$4E,$67,$C8,$D7,$F0,$2F,$1D,$BE,$7B,$74,$CD
	.dc.b	$FC,$AF,$53,$FD,$BF,$18,$BF,$91,$5B,$7F,$08,$D7,$F0,$13,$C1,$CD
	.dc.b	$7D,$49,$D6,$1E,$68,$ED,$39,$63,$D9,$CB,$49,$87,$A2,$32,$FE,$4F
	.dc.b	$AF,$F9,$0F,$0F,$FB,$BF,$7F,$E4,$9C,$BF,$9C,$C3,$F8,$84,$E5,$FE
	.dc.b	$31,$F1,$F6,$FF,$94,$B6,$B8,$4F,$ED,$6B,$13,$9F,$90,$D7,$C8,$4A
	.dc.b	$FE,$39,$19,$69,$0F,$DF,$7C,$7F,$85,$FE,$E5,$F0,$3F,$DF,$5A,$7F
	.dc.b	$7F,$E1,$46,$5B,$19,$8E,$51,$88,$FE,$2E,$DE,$8F,$DD,$72,$BE,$6B
	.dc.b	$FC,$BF,$1F,$EC,$79,$7C,$0F,$C3,$A7,$D2,$B3,$C3,$71,$FE,$BF,$7C
	.dc.b	$1B,$7E,$52,$3E,$96,$BE,$47,$CE,$77,$C7,$FA,$0C,$7F,$A3,$30,$FD
	.dc.b	$2F,$B7,$E3,$E9,$FF,$11,$CE,$5A,$FB,$BD,$CF,$82,$7C,$52,$3F,$25
	.dc.b	$FF,$14,$53,$E6,$FE,$27,$2F,$F4,$7C,$FF,$D7,$37,$EC,$BF,$85,$5A
	.dc.b	$8F,$EB,$9F,$27,$F8,$32,$3E,$3F,$6B,$ED,$7D,$77,$EB,$63,$FA,$93
	.dc.b	$50,$70,$DE,$AF,$2D,$FF,$82,$23,$E1,$35,$FE,$DB,$9F,$A7,$CF,$FC
	.dc.b	$A7,$C9,$7E,$57,$E4,$E5,$7F,$6B,$7F,$31,$F0,$3E,$4F,$47,$F1,$99
	.dc.b	$FF,$A9,$AB,$FC,$94,$D7,$FD,$3C,$D7,$FB,$57,$77,$F8,$2B,$E4,$BF
	.dc.b	$44,$48,$FE,$F1,$F2,$7F,$A7,$FC,$5B,$FA,$34,$7F,$6C,$6F,$5F,$FB
	.dc.b	$A1,$BC,$BF,$E1,$57,$FD,$9B,$3F,$E2,$8F,$AF,$FD,$02,$69,$1F,$F3
	.dc.b	$06,$29,$7C,$9F,$84,$93,$FE,$CA,$8F,$F6,$7B,$7C,$6F,$37,$C9,$48
	.dc.b	$C1,$E1,$E0,$FD,$86,$26,$92,$3F,$D8,$58,$F8,$7E,$43,$E0,$12,$3C
	.dc.b	$1F,$21,$F0,$FE,$43,$E2,$78,$72,$7C,$91,$FB,$FF,$9D,$FD,$44,$7F
	.dc.b	$B3,$5A,$FF,$B4,$9A,$FF,$86,$DA,$FF,$BB,$F5,$E3,$C1,$F3,$7C,$2F
	.dc.b	$A4,$7D,$3F,$90,$FE,$19,$F3,$DF,$D0,$AF,$FD,$C5,$EE,$FF,$8E,$5B
	.dc.b	$EE,$6B,$8E,$47,$FA,$36,$B8,$3F,$EE,$16,$BF,$EE,$66,$BF,$E2,$F7
	.dc.b	$78,$FE,$09,$BC,$B9,$AF,$E5,$D2,$FC,$21,$EB,$7E,$27,$E4,$E6,$3F
	.dc.b	$65,$FC,$67,$7E,$23,$F1,$91,$6F,$F1,$93,$F9,$1F,$93,$95,$F8,$17
	.dc.b	$F2,$9F,$27,$88,$FC,$A4,$5A,$FC,$A4,$FE,$67,$E4,$E5,$7E,$25,$FC
	.dc.b	$E7,$C1,$62,$3F,$39,$16,$BF,$39,$37,$5F,$CE,$4D,$AF,$AD,$3F,$A5
	.dc.b	$F5,$FF,$4D,$F2,$72,$7F,$2D,$19,$69,$0F,$D5,$79,$7F,$D4,$FB,$FF
	.dc.b	$BD,$65,$FE,$33,$ED,$46,$33,$FF,$40,$36,$D3,$AF,$9A,$E5,$E5,$49
	.dc.b	$FE,$27,$87,$3B,$E1,$7B,$9F,$F4,$DF,$2B,$3F,$EA,$DF,$52,$EF,$A7
	.dc.b	$E3,$A2,$57,$E2,$5F,$EB,$55,$C6,$30,$43,$FC,$D6,$E5,$B1,$DB,$C0
	.dc.b	$FE,$E5,$F6,$0F,$8D,$A4,$AF,$4E,$FF,$E9,$F9,$7C,$6E,$67,$FC,$FB
	.dc.b	$1C,$5D,$9F,$11,$C3,$A8,$EF,$F1,$7E,$0F,$8B,$93,$C3,$65,$AB,$73
	.dc.b	$98,$C3,$63,$D3,$E6,$7F,$7D,$F9,$CF,$EB,$B3,$FE,$8A,$F9,$FF,$D0
	.dc.b	$BF,$EC,$86,$BF,$E8,$E8,$DF,$1F,$F2,$C7,$8C,$7F,$A7,$43,$87,$FD
	.dc.b	$40,$D7,$FA,$7B,$5D,$FC,$27,$8C,$8F,$87,$E1,$FF,$C2,$5F,$61,$FE
	.dc.b	$9B,$8F,$81,$F6,$0F,$8E,$FF,$12,$36,$3F,$DB,$F3,$B7,$FD,$4A,$DC
	.dc.b	$7F,$A1,$B7,$87,$FD,$68,$D9,$FC,$AD,$7F,$B1,$1E,$F8,$76,$F0,$E4
	.dc.b	$60,$87,$FA,$E2,$79,$70,$1A,$3E,$2B,$F5,$1E,$EF,$F4,$8E,$BF,$F0
	.dc.b	$3D,$FF,$5C,$6B,$FA,$88,$D1,$C1,$FC,$04,$7F,$83,$B4,$FD,$37,$B9
	.dc.b	$FB,$3F,$99,$FD,$A7,$CC,$CB,$DC,$65,$A4,$3F,$D9,$F6,$F7,$7F,$79
	.dc.b	$F3,$3E,$52,$4F,$F4,$E9,$46,$23,$8B,$C1,$F2,$FC,$CF,$F9,$27,$E3
	.dc.b	$0B,$3F,$F0,$7F,$88,$F8,$FF,$F7,$21,$FD,$06,$D1,$97,$FC,$2B,$19
	.dc.b	$8C,$42,$FA,$D7,$4A,$FF,$77,$46,$25,$F4,$1C,$27,$B3,$7D,$1F,$F7
	.dc.b	$A3,$D2,$B8,$37,$1F,$45,$F2,$BF,$3D,$F0,$6C,$7F,$8A,$AB,$86,$16
	.dc.b	$7C,$77,$CA,$7C,$15,$93,$DB,$7C,$DF,$F1,$9E,$1E,$2B,$E3,$87,$B6
	.dc.b	$76,$53,$E5,$57,$F8,$E9,$F9,$33,$C1,$A7,$4C,$3C,$51,$E8,$8F,$61
	.dc.b	$D0,$A7,$CD,$8D,$BC,$6B,$D0,$F8,$0B,$3F,$30,$F0,$61,$F0,$A3,$C2
	.dc.b	$6C,$B1,$F6,$CF,$36,$3B,$9F,$8B,$3E,$29,$E8,$F8,$4F,$2E,$3F,$15
	.dc.b	$F4,$4F,$03,$94,$F4,$3D,$D5,$D0,$F2,$1A,$9E,$31,$F0,$E7,$11,$F0
	.dc.b	$6F,$DB,$8F,$D5,$8F,$54,$EB,$0F,$14,$62,$F9,$1F,$08,$F1,$3F,$37
	.dc.b	$0F,$14,$62,$3B,$1E,$FB,$ED,$1F,$02,$33,$7E,$88,$F5,$92,$3E,$25
	.dc.b	$65,$ED,$19,$7A,$C7,$D3,$F1,$FF,$5B,$1F,$2D,$AF,$89,$B8,$7E,$A5
	.dc.b	$CC,$7F,$A1,$BE,$37,$FD,$0F,$18,$84,$62,$BF,$A1,$46,$04,$EE,$75
	.dc.b	$1F,$D1,$DF,$CD,$65,$FC,$47,$A7,$F3,$93,$DC,$FC,$4C,$65,$FC,$59
	.dc.b	$9B,$FE,$07,$19,$7F,$82,$C6,$71,$FD,$24,$E5,$F2,$DF,$E6,$A3,$A6
	.dc.b	$5C,$2B,$D0,$F4,$DD,$8F,$E8,$27,$A1,$FE,$16,$8C,$BE,$43,$E0,$63
	.dc.b	$C6,$7D,$71,$D3,$1F,$1D,$70,$BF,$06,$3E,$38,$C4,$F9,$AB,$F6,$78
	.dc.b	$F9,$A3,$17,$D6,$7F,$6E,$7C,$77,$07,$47,$8D,$74,$7E,$81,$F3,$1F
	.dc.b	$4E,$3B,$3A,$1C,$67,$9D,$78,$0F,$4E,$3F,$0E,$FD,$87,$C3,$71,$8F
	.dc.b	$AD,$B1,$DD,$F8,$71,$F8,$0C,$3C,$35,$D6,$B9,$1E,$C3,$A9,$FD,$26
	.dc.b	$3D,$4F,$F1,$F9,$EC,$ED,$B7,$EB,$AF,$3C,$FD,$59,$F0,$5F,$8A,$3E
	.dc.b	$2B,$E6,$AF,$7C,$7C,$66,$BD,$CD,$9D,$CA,$FC,$7A,$EE,$F8,$63,$D7
	.dc.b	$85,$7D,$55,$FC,$EF,$9B,$FC,$CF,$F0,$BF,$8A,$CF,$FC,$C0,$D9,$FF
	.dc.b	$34,$FC,$D5,$BF,$F3,$4F,$12,$B1,$87,$2B,$FE,$ED,$1F,$EE,$0A,$FF
	.dc.b	$38,$B5,$E2,$D9,$FF,$3B,$FC,$CF,$2B,$37,$E6,$B0,$CC,$3F,$C2,$3F
	.dc.b	$35,$71,$FF,$04,$BC,$7C,$89,$3F,$E1,$6F,$22,$D7,$F8,$3A,$7C,$18
	.dc.b	$FF,$9A,$6B,$FC,$19,$18,$3C,$9D,$0F,$7B,$DA,$D9,$FF,$30,$2F,$6F
	.dc.b	$33,$27,$3E,$DD,$C7,$FC,$21,$07,$FD,$23,$69,$F9,$90,$F5,$5E,$66
	.dc.b	$1B,$7E,$0F,$5B,$13,$23,$C7,$36,$BB,$9D,$9E,$BF,$3A,$F0,$29,$F6
	.dc.b	$F9,$9F,$0F,$FE,$26,$C3,$FD,$07,$82,$F8,$46,$1B,$1F,$CB,$9A,$FF
	.dc.b	$9C,$5A,$FF,$80,$DA,$FC,$17,$8D,$7E,$3E,$BF,$15,$5D,$6F,$C9,$18
	.dc.b	$BF,$5F,$2B,$57,$AB,$F5,$0F,$C8,$FD,$7F,$F3,$EC,$7E,$AC,$65,$FF
	.dc.b	$4B,$47,$F8,$F1,$C9,$8E,$AE,$35,$23,$79,$5C,$7C,$37,$E1,$9E,$75
	.dc.b	$E0,$AD,$67,$FC,$CA,$DF,$2C,$F9,$3F,$BA,$DC,$3F,$D5,$F4,$F9,$CD
	.dc.b	$C7,$D0,$9D,$5F,$DA,$39,$DF,$EA,$5B,$3F,$D7,$9A,$7F,$60,$9E,$2F
	.dc.b	$FA,$8D,$E7,$7F,$BD,$B7,$89,$E4,$C3,$F0,$51,$8C,$3F,$CF,$BF,$39
	.dc.b	$FB,$33,$F6,$91,$63,$24,$C7,$FC,$F3,$19,$7F,$CF,$7E,$A7,$EE,$CF
	.dc.b	$DE,$72,$5A,$FF,$44,$C7,$EF,$FD,$4C,$57,$FA,$09,$FE,$17,$F8,$EF
	.dc.b	$E9,$EB,$FD,$9C,$D9,$FC,$1B,$F2,$1A,$70,$24,$70,$D6,$C6,$A3,$9B
	.dc.b	$DA,$33,$AF,$F6,$AE,$35,$FC,$3A,$EF,$B7,$93,$D3,$F2,$B3,$1F,$2B
	.dc.b	$96,$9E,$8F,$4F,$D6,$7B,$3E,$72,$C6,$49,$7E,$C8,$CB,$F8,$1F,$A6
	.dc.b	$7C,$DF,$F0,$66,$D7,$E2,$BF,$02,$7C,$A6,$6A,$4F,$E1,$6B,$77,$FE
	.dc.b	$24,$6B,$FE,$E7,$6B,$FE,$79,$77,$B7,$F4,$3F,$C0,$3F,$8B,$57,$F0
	.dc.b	$49,$CB,$FC,$6E,$53,$AA,$FA,$BE,$67,$DA,$3E,$DF,$CF,$5B,$FE,$1B
	.dc.b	$F4,$2F,$95,$FC,$E3,$63,$F3,$8D,$BF,$F1,$DB,$5F,$F4,$77,$1F,$D6
	.dc.b	$F6,$FF,$0F,$ED,$FE,$21,$B7,$F8,$8F,$6B,$F1,$67,$E3,$3B,$6C,$7F
	.dc.b	$1E,$FC,$0B,$D1,$8F,$FA,$64,$38,$7F,$D3,$8D,$7F,$9D,$37,$1F,$C3
	.dc.b	$FC,$F7,$E6,$3E,$73,$F3,$2D,$BF,$CC,$FC,$F7,$E7,$0F,$CE,$FC,$F5
	.dc.b	$8F,$E5,$FF,$2F,$FC,$C7,$2F,$F2,$CB,$5F,$23,$5F,$F5,$73,$56,$3F
	.dc.b	$98,$E7,$FD,$49,$FA,$AF,$26,$DF,$EA,$BE,$5B,$F5,$A7,$EB,$B8,$EC
	.dc.b	$78,$BD,$7F,$E1,$D9,$7F,$31,$CB,$C3,$18,$3F,$BC,$78,$0F,$99,$FD
	.dc.b	$B9,$FB,$8F,$99,$B8,$7F,$4A,$F9,$57,$A3,$FD,$79,$B1,$F8,$93,$FB
	.dc.b	$26,$EF,$C8,$BF,$CB,$F1,$C6,$BF,$B2,$FD,$A3,$A7,$F9,$13,$67,$F9
	.dc.b	$13,$75,$FF,$40,$B5,$FC,$3D,$BE,$9B,$FC,$2E,$DA,$F1,$4B,$E7,$BB
	.dc.b	$7A,$BA,$BF,$47,$C7,$F8,$BB,$AE,$1F,$D9,$B2,$46,$63,$C6,$F9,$88
	.dc.b	$FF,$A5,$30,$F5,$7C,$9F,$F8,$AD,$B1,$CE,$71,$A2,$BE,$0F,$2C,$AF
	.dc.b	$EE,$BF,$48,$F0,$F1,$B7,$F4,$7C,$FF,$4C,$FF,$BC,$DA,$FF,$84,$5A
	.dc.b	$9A,$7E,$BF,$C5,$FF,$45,$E8,$FA,$CF,$99,$1F,$D7,$67,$58,$FF,$86
	.dc.b	$EB,$5A,$7D,$6E,$69,$3E,$6D,$AE,$5E,$59,$FB,$24,$9F,$B2,$FF,$90
	.dc.b	$F8,$9D,$3F,$58,$FA,$FE,$45,$AF,$AF,$CD,$F8,$79,$FB,$7F,$37,$9C
	.dc.b	$7F,$C0,$F8,$23,$EF,$FB,$FF,$80,$F9,$3B,$8F,$2A,$46,$63,$0E,$AF
	.dc.b	$9B,$FE,$36,$FA,$5F,$E6,$7E,$97,$FB,$4F,$E3,$BF,$C7,$94,$9C,$BF
	.dc.b	$ED,$48,$CF,$A7,$FD,$F3,$E2,$51,$C2,$B7,$18,$AF,$F2,$34,$62,$75
	.dc.b	$BF,$F8,$32,$7A,$1F,$E4,$BB,$E9,$3A,$C3,$C9,$3C,$AB,$86,$1D,$63
	.dc.b	$CD,$18,$AF,$2F,$97,$FE,$CA,$AD,$63,$E1,$BE,$58,$74,$78,$B7,$1F
	.dc.b	$0C,$68,$F2,$9A,$2C,$F3,$51,$C1,$D5,$78,$31,$F7,$47,$F3,$3A,$F9
	.dc.b	$6D,$8C,$BE,$F3,$FD,$FF,$F1,$9F,$E9,$3E,$5F,$F1,$23,$63,$3E,$27
	.dc.b	$EF,$5B,$3F,$E2,$9B,$FC,$37,$C6,$5E,$23,$49,$5F,$D1,$4F,$BD,$1A
	.dc.b	$7D,$AD,$CB,$E4,$B7,$6F,$BD,$DF,$FE,$C7,$7F,$01,$19,$AF,$BE,$D7
	.dc.b	$AD,$FC,$FC,$17,$C6,$77,$39,$AE,$15,$EA,$9F,$9F,$5B,$79,$9E,$6A
	.dc.b	$E0,$75,$8D,$4F,$1C,$3D,$33,$AA,$E6,$7E,$33,$97,$F9,$B7,$CB,$7E
	.dc.b	$83,$AF,$FC,$51,$CA,$F9,$FC,$73,$7F,$3C,$8C,$61,$FA,$FB,$CB,$FD
	.dc.b	$62,$11,$8C,$3F,$A1,$5D,$9E,$35,$FB,$CA,$C7,$C2,$E7,$E1,$A4,$FE
	.dc.b	$EE,$32,$FE,$F6,$32,$FF,$04,$8E,$9F,$0D,$FE,$69,$C7,$FA,$83,$70
	.dc.b	$FF,$5A,$57,$DF,$9C,$B6,$32,$FF,$0E,$9E,$6F,$F0,$BF,$9C,$F4,$7C
	.dc.b	$E7,$A7,$C9,$93,$FC,$3E,$32,$F3,$7C,$04,$70,$6C,$87,$F1,$CF,$27
	.dc.b	$E2,$7C,$E4,$8F,$E3,$4F,$C7,$F9,$C9,$1E,$77,$D8,$47,$D5,$CB,$A7
	.dc.b	$F9,$3C,$7C,$F8,$FA,$1E,$54,$9F,$E4,$CF,$D3,$F8,$9F,$AD,$F9,$5F
	.dc.b	$E8,$67,$FC,$51,$BA,$31,$1F,$E3,$66,$BF,$DF,$9A,$FF,$86,$17,$C5
	.dc.b	$74,$F9,$69,$3F,$CF,$A7,$2D,$EC,$F0,$91,$D7,$C5,$79,$3E,$2B,$EE
	.dc.b	$57,$FB,$7E,$7E,$D3,$5F,$3B,$5F,$37,$CA,$7B,$3E,$5A,$47,$86,$3D
	.dc.b	$FF,$2D,$F0,$3C,$39,$3E,$58,$F8,$7F,$2B,$E0,$24,$78,$1F,$1F,$CA
	.dc.b	$7C,$8F,$95,$93,$ED,$8F,$97,$F2,$BF,$33,$E5,$24,$FC,$18,$F9,$FF
	.dc.b	$2B,$EC,$24,$7B,$1F,$7F,$C9,$FD,$4F,$93,$93,$F2,$63,$EB,$FC,$97
	.dc.b	$D8,$F9,$39,$3F,$36,$3F,$67,$E4,$3F,$4D,$47,$FB,$36,$30,$D8,$CC
	.dc.b	$FF,$23,$9C,$BF,$ED,$3F,$A1,$78,$FF,$ED,$58,$46,$1D,$46,$63,$11
	.dc.b	$F5,$3E,$57,$F8,$17,$BD,$FE,$73,$BF,$F5,$A3,$7C,$CF,$C5,$4F,$E4
	.dc.b	$8F,$BE,$4A,$FB,$ED,$55,$FE,$BD,$E8,$4F,$19,$FF,$05,$DF,$F5,$0A
	.dc.b	$EF,$F2,$BE,$2F,$56,$56,$A3,$CA,$61,$EB,$8F,$F9,$6A,$FF,$AE,$C6
	.dc.b	$CE,$BF,$2B,$EC,$F9,$59,$18,$21,$C8,$FF,$3D,$CA,$37,$19,$F0,$BF
	.dc.b	$C8,$15,$F1,$AB,$E3,$FC,$94,$AF,$5C,$65,$F6,$C6,$5E,$51,$FE,$49
	.dc.b	$76,$68,$86,$9F,$A7,$F0,$1F,$53,$E0,$24,$FC,$88,$CB,$48,$76,$F9
	.dc.b	$1F,$B7,$EE,$FD,$CF,$8F,$93,$F4,$23,$2F,$D2,$8C,$BE,$68,$FE,$8E
	.dc.b	$BE,$7B,$FC,$6E,$FF,$9E,$07,$CF,$BC,$FD,$B6,$35,$3C,$1F,$9C,$73
	.dc.b	$F8,$1F,$CB,$46,$63,$FD,$84,$D7,$FC,$7A,$D4,$71,$3A,$C7,$9A,$3B
	.dc.b	$72,$38,$BF,$C8,$B7,$FE,$A2,$BF,$C7,$FC,$97,$E7,$BE,$4A,$46,$08
	.dc.b	$7D,$67,$2F,$D8,$8F,$D2,$FC,$97,$E9,$BE,$4A,$4F,$E6,$A3,$2D,$21
	.dc.b	$F8,$0F,$92,$F9,$7F,$3C,$7F,$80,$FE,$00,$EA,$A6,$A2,$57,$F8,$84
	.dc.b	$57,$17,$0D,$7F,$C4,$CD,$7F,$DD,$8E,$75,$F8,$91,$F2,$A3,$1A,$FF
	.dc.b	$9A,$9B,$1F,$B8,$C1,$97,$F0,$4F,$94,$FE,$0B,$F2,$12,$BF,$73,$19
	.dc.b	$69,$0F,$CF,$FC,$6F,$F0,$CF,$93,$FF,$07,$7E,$65,$E2,$9F,$D6,$37
	.dc.b	$4F,$E1,$F3,$B9,$FE,$0A,$D7,$7F,$27,$CB,$E4,$3F,$C1,$57,$FE,$98
	.dc.b	$9F,$F0,$43,$5F,$EC,$AD,$7F,$D0,$FE,$02,$35,$87,$C4,$8C,$57,$5A
	.dc.b	$FF,$4F,$D3,$E3,$BF,$C4,$DF,$20,$F6,$FD,$7F,$1B,$2F,$FD,$4B,$39
	.dc.b	$6C,$FF,$87,$E3,$11,$FD,$E1,$AF,$91,$AF,$FB,$D2,$D1,$EE,$24,$7B
	.dc.b	$9A,$FC,$A7,$E6,$9F,$3B,$AF,$F7,$3F,$1E,$F5,$FF,$99,$D7,$F8,$03
	.dc.b	$C4,$FF,$54,$9F,$91,$7F,$5B,$EF,$7F,$15,$7F,$DB,$DD,$5F,$EA,$E8
	.dc.b	$C5,$F3,$E0,$DD,$FF,$AC,$A3,$16,$F9,$DE,$53,$D6,$30,$F8,$30,$FF
	.dc.b	$5C,$CF,$4C,$7C,$AD,$C3,$F7,$EF,$29,$DD,$71,$BC,$BE,$1B,$E2,$DB
	.dc.b	$CC,$FF,$08,$8F,$15,$F3,$8C,$8C,$3D,$31,$E1,$8F,$8C,$71,$C3,$E1
	.dc.b	$57,$86,$32,$71,$C3,$E6,$E3,$C6,$FB,$4F,$C0,$39,$1E,$0C,$3B,$DF
	.dc.b	$9E,$34,$7D,$03,$CA,$FA,$0D,$4F,$96,$BA,$92,$39,$3E,$17,$CD,$1E
	.dc.b	$23,$13,$FC,$AA,$BC,$2F,$90,$F0,$CF,$77,$57,$F3,$1B,$8F,$E0,$1E
	.dc.b	$18,$78,$5C,$9F,$12,$FA,$BE,$B3,$AE,$3F,$69,$DE,$3E,$47,$4F,$9F
	.dc.b	$1F,$79,$23,$CD,$19,$78,$C7,$C0,$3E,$A5,$F8,$E3,$EA,$61,$E7,$69
	.dc.b	$E2,$3E,$6D,$F8,$DC,$DF,$CD,$BE,$64,$8F,$6D,$E5,$F8,$31,$97,$AC
	.dc.b	$7E,$27,$E3,$7F,$CD,$31,$E0,$6E,$3F,$A6,$6E,$5F,$B4,$77,$3F,$D7
	.dc.b	$BE,$3B,$FA,$FC,$62,$11,$8B,$FF,$72,$36,$3F,$B2,$9C,$4F,$F7,$5C
	.dc.b	$69,$FC,$BE,$6F,$77,$FA,$1C,$FE,$FE,$78,$3F,$C0,$67,$5D,$BF,$26
	.dc.b	$73,$9F,$FB,$46,$F8,$BD,$6F,$84,$62,$7C,$67,$A0,$FC,$ED,$F4,$7F
	.dc.b	$3E,$79,$F4,$E3,$87,$ED,$23,$B9,$FE,$00,$9E,$17,$E7,$8E,$8F,$07
	.dc.b	$F6,$67,$3C,$3C,$D5,$CE,$3C,$57,$B3,$96,$3F,$C0,$B0,$F2,$46,$27
	.dc.b	$CF,$8F,$F6,$88,$C1,$C7,$0F,$98,$D9,$F7,$DF,$E3,$EF,$CF,$5F,$14
	.dc.b	$F1,$CF,$98,$F1,$B8,$AE,$AD,$3A,$96,$3A,$9E,$4C,$3D,$97,$DF,$0F
	.dc.b	$A1,$3F,$82,$8C,$4F,$D9,$8E,$E7,$CB,$9F,$65,$76,$8C,$4F,$47,$B1
	.dc.b	$F5,$8E,$27,$F9,$F2,$78,$BE,$4C,$7E,$91,$F2,$4B,$5D,$63,$85,$75
	.dc.b	$C3,$E4,$47,$9C,$EE,$FC,$27,$C6,$78,$70,$FA,$D8,$78,$23,$84,$F2
	.dc.b	$7F,$0E,$76,$3E,$74,$FC,$27,$F9,$DF,$2F,$F8,$07,$E3,$7F,$C4,$79
	.dc.b	$FF,$80,$78,$91,$8C,$BF,$B9,$7C,$75,$C1,$87,$19,$FF,$02,$D7,$F8
	.dc.b	$6E,$7F,$BC,$35,$DD,$F4,$9F,$F0,$AC,$E6,$BF,$BE,$FB,$77,$90,$D2
	.dc.b	$57,$EA,$B0,$F0,$CE,$A3,$FC,$2B,$F1,$DE,$5B,$E7,$18,$3F,$C3,$3E
	.dc.b	$AE,$72,$FF,$62,$C7,$F8,$73,$E3,$71,$32,$3F,$C5,$2F,$96,$77,$8F
	.dc.b	$F8,$AA,$31,$F1,$BC,$CE,$9F,$19,$6F,$FC,$4F,$1D,$FE,$37,$17,$FE
	.dc.b	$3E,$7C,$5F,$1B,$2B,$FC,$5B,$E7,$5A,$FF,$21,$CF,$9F,$E3,$71,$C7
	.dc.b	$F5,$5F,$A5,$2F,$D8,$9D,$58,$6E,$3F,$C1,$CF,$E1,$1F,$6A,$F3,$B5
	.dc.b	$52,$7F,$CD,$95,$C2,$FE,$73,$5F,$E3,$2D,$7F,$96,$BB,$C7,$F1,$38
	.dc.b	$3C,$FE,$95,$FE,$CB,$94,$E6,$FF,$C9,$7F,$1F,$F4,$CF,$A9,$F1,$F6
	.dc.b	$32,$7F,$94,$3E,$3F,$18,$7F,$99,$9F,$BD,$E1,$78,$67,$FD,$47,$CF
	.dc.b	$FE,$19,$F0,$73,$B6,$9B,$3F,$0F,$0F,$CC,$DA,$73,$1F,$ED,$D8,$C4
	.dc.b	$FF,$AF,$7E,$71,$E5,$7D,$06,$C7,$3D,$9F,$0D,$F3,$63,$85,$6C,$FD
	.dc.b	$16,$0B,$E2,$FE,$6D,$AF,$F4,$66,$BF,$D6,$9D,$DF,$ED,$AF,$E5,$4F
	.dc.b	$FB,$2A,$73,$7F,$D1,$67,$35,$FE,$7B,$C3,$FC,$ED,$3F,$07,$3F,$F1
	.dc.b	$F7,$C8,$7C,$AF,$27,$E5,$B6,$75,$48,$C6,$5F,$15,$FA,$3F,$21,$2B
	.dc.b	$7F,$21,$6B,$E5,$4F,$D6,$F1,$F1,$9F,$8D,$FB,$3F,$21,$F6,$BC,$AB
	.dc.b	$8F,$81,$B8,$7D,$36,$DF,$AD,$B3,$F8,$0C,$FF,$CF,$8F,$A2,$FA,$3E
	.dc.b	$AF,$90,$FC,$3F,$C8,$7E,$21,$B1,$93,$DF,$19,$69,$F0,$BE,$3F,$F1
	.dc.b	$FF,$21,$F9,$06,$C6,$4F,$8F,$19,$69,$F2,$BE,$43,$F2,$E7,$E6,$3C
	.dc.b	$AB,$19,$3F,$8E,$71,$FF,$9C,$67,$F9,$26,$66,$92,$7E,$AF,$C8,$72
	.dc.b	$DF,$F3,$0F,$F0,$4F,$9B,$7B,$5F,$90,$6D,$7F,$BA,$5B,$5F,$CA,$DA
	.dc.b	$FF,$B1,$30,$7C,$8F,$88,$95,$E6,$8C,$B7,$A3,$F6,$7F,$21,$FB,$46
	.dc.b	$C7,$E2,$BE,$47,$F6,$FF,$1D,$FB,$86,$C7,$E4,$FF,$1B,$65,$FD,$0D
	.dc.b	$B1,$F1,$1B,$1F,$E2,$AD,$D1,$88,$FF,$6D,$AF,$96,$FE,$84,$BD,$9B
	.dc.b	$18,$73,$F2,$1A,$F8,$FB,$8F,$8D,$E3,$F2,$1E,$32,$47,$8D,$F3,$7B
	.dc.b	$9D,$BE,$43,$BB,$67,$D9,$F1,$FE,$2F,$90,$F1,$B6,$3E,$07,$C8,$7B
	.dc.b	$C8,$FB,$DF,$85,$F1,$FE,$A3,$D7,$F2,$16,$7E,$47,$A7,$EE,$F4,$BD
	.dc.b	$ED,$8F,$99,$F1,$DF,$2C,$8F,$CB,$7E,$77,$C5,$FC,$63,$E3,$FC,$75
	.dc.b	$9F,$A9,$F1,$9F,$2B,$BF,$E5,$B6,$3E,$C4,$7D,$72,$3F,$5D,$FB,$5F
	.dc.b	$19,$F4,$8E,$1D,$0F,$29,$5F,$CE,$3A,$55,$F6,$3E,$32,$4F,$F1,$7F
	.dc.b	$33,$6F,$E1,$E3,$38,$7E,$0E,$D1,$F8,$92,$4F,$E2,$5F,$F0,$25,$7F
	.dc.b	$26,$AF,$F0,$CD,$70,$AE,$DE,$96,$BE,$32,$47,$87,$E3,$2D,$FA,$63
	.dc.b	$FA,$7E,$0A,$FF,$0D,$36,$3F,$76,$D9,$FC,$C3,$5F,$F4,$AF,$7F,$FA
	.dc.b	$47,$E3,$3C,$BF,$0F,$E6,$6D,$FC,$16,$D7,$84,$FF,$18,$DA,$33,$B7
	.dc.b	$CC,$9C,$FC,$67,$C0,$F8,$9F,$82,$DC,$3E,$63,$67,$A6,$5F,$EA,$38
	.dc.b	$C5,$F3,$97,$CA,$B5,$FD,$74,$7F,$AC,$BD,$7F,$F5,$7C,$E7,$C5,$FF
	.dc.b	$5C,$47,$FA,$CB,$C4,$76,$F9,$7E,$33,$FA,$66,$1F,$C0,$30,$E0,$F3
	.dc.b	$31,$8F,$69,$DC,$78,$A3,$18,$75,$BE,$8F,$8A,$33,$3C,$B1,$F1,$B4
	.dc.b	$EC,$31,$F2,$EB,$A8,$ED,$37,$2F,$2E,$5E,$26,$E3,$E1,$BD,$63,$E0
	.dc.b	$C3,$89,$DA,$B6,$FF,$1E,$C7,$BB,$99,$E1,$3F,$0A,$3F,$D8,$ED,$7D
	.dc.b	$98,$FF,$98,$5E,$A7,$FB,$33,$CB,$7A,$8F,$96,$D4,$61,$B1,$99,$FF
	.dc.b	$35,$4E,$5F,$F3,$65,$FE,$2F,$DD,$EC,$D8,$CF,$1F,$F9,$39,$B3,$A8
	.dc.b	$C6,$DF,$15,$BB,$7E,$11,$B5,$F7,$FE,$03,$FD,$69,$39,$BF,$9D,$5A
	.dc.b	$AF,$C7,$B7,$05,$F0,$9F,$C4,$4E,$49,$5F,$5E,$75,$3C,$A3,$A9,$1F
	.dc.b	$A9,$1A,$AF,$76,$1E,$9A,$D3,$F7,$E7,$9F,$E0,$9F,$06,$FF,$12,$D9
	.dc.b	$FC,$AB,$71,$FE,$B7,$28,$FF,$27,$3F,$8D,$74,$FF,$60,$6B,$BB,$EF
	.dc.b	$EF,$FB,$F9,$5C,$EF,$8E,$EB,$CD,$FB,$F9,$CC,$63,$97,$F8,$AB,$71
	.dc.b	$FE,$0F,$F1,$D6,$B7,$1F,$CB,$67,$18,$75,$8F,$E6,$15,$C5,$7E,$08
	.dc.b	$F2,$F0,$D8,$FF,$36,$1A,$8F,$07,$2B,$83,$FD,$77,$3A,$AE,$0F,$F1
	.dc.b	$B3,$81,$9A,$E3,$3C,$23,$52,$F2,$BF,$45,$1E,$46,$C6,$63,$F4,$CD
	.dc.b	$7F,$D9,$0D,$D7,$F8,$E3,$75,$FA,$ED,$9F,$D5,$5F,$F9,$F1,$F4,$5F
	.dc.b	$C0,$23,$FE,$8B,$C7,$DD,$E9,$FC,$0C,$BF,$2F,$39,$6B,$FC,$E6,$7E
	.dc.b	$57,$1F,$CA,$F0,$E4,$FC,$56,$BB,$E9,$FE,$7C,$F7,$6B,$F1,$23,$7A
	.dc.b	$FF,$41,$BF,$F1,$17,$3F,$DC,$F8,$D9,$7F,$DA,$DF,$AB,$FA,$27,$E5
	.dc.b	$23,$E6,$F8,$BF,$70,$8F,$7F,$8D,$FB,$79,$9A,$48,$FE,$ED,$F2,$1F
	.dc.b	$8D,$FC,$25,$E6,$52,$4F,$FA,$46,$1E,$5F,$E3,$7B,$9F,$13,$49,$5F
	.dc.b	$8C,$F3,$FF,$CB,$6F,$FA,$67,$9F,$C4,$DC,$FF,$39,$18,$8F,$EB,$6B
	.dc.b	$E3,$FF,$3B,$32,$7F,$C1,$8F,$E8,$3E,$43,$C5,$F1,$72,$7F,$37,$92
	.dc.b	$3F,$37,$F5,$AE,$EF,$D3,$E3,$2F,$FC,$2C,$FE,$AB,$E4,$3F,$57,$F2
	.dc.b	$32,$7F,$C3,$4F,$EB,$BE,$47,$F5,$9F,$5B,$FB,$C9,$FD,$4B,$5F,$D2
	.dc.b	$37,$AF,$F7,$DE,$1F,$E2,$9B,$C5,$FB,$82,$57,$F8,$9F,$5F,$DF,$7D
	.dc.b	$AF,$F1,$7A,$FE,$0C,$DF,$13,$FD,$CE,$DA,$78,$C8,$FC,$25,$7F,$80
	.dc.b	$72,$7C,$67,$A0,$97,$FE,$CB,$F9,$EF,$F1,$CB,$FE,$CD,$B7,$C8,$F8
	.dc.b	$89,$19,$E4,$FE,$19,$F3,$8F,$07,$F7,$6D,$9F,$EB,$CD,$7F,$95,$9F
	.dc.b	$E8,$67,$0E,$1D,$67,$FE,$17,$6E,$BF,$CC,$9B,$3F,$E8,$EF,$B7,$F6
	.dc.b	$CE,$6F,$C6,$49,$FE,$37,$3A,$7E,$2E,$4F,$92,$F9,$A4,$8F,$81,$F2
	.dc.b	$75,$FF,$15,$CE,$DF,$89,$3F,$52,$7E,$AF,$1C,$9F,$90,$FD,$8D,$F1
	.dc.b	$1C,$A2,$E9,$F6,$A7,$F9,$A7,$E8,$8F,$66,$92,$FF,$99,$C7,$F3,$B7
	.dc.b	$F9,$73,$B7,$31,$B7,$3E,$7D,$BF,$F2,$05,$6E,$78,$D7,$82,$BF,$A2
	.dc.b	$7A,$DF,$E3,$4B,$FF,$23,$37,$7F,$EC,$2D,$BF,$F5,$EF,$E8,$9F,$86
	.dc.b	$7F,$A0,$46,$5F,$F2,$67,$17,$43,$FA,$84,$EE,$FF,$33,$F2,$BF,$9C
	.dc.b	$25,$7E,$3F,$D8,$AF,$D0,$8B,$3F,$A0,$9F,$3F,$A3,$FD,$23,$1F,$E7
	.dc.b	$97,$F8,$C3,$FC,$B9,$19,$23,$FD,$72,$37,$5E,$E9,$FE,$C9,$9E,$33
	.dc.b	$F7,$C7,$6B,$E0,$E8,$FD,$9F,$CA,$C8,$FD,$33,$FB,$6F,$95,$C4,$7E
	.dc.b	$22,$2E,$1F,$B8,$9F,$EE,$5F,$82,$7F,$A3,$EB,$FB,$7B,$5F,$ED,$4D
	.dc.b	$9F,$F3,$C7,$E4,$4F,$88,$E3,$F0,$52,$F8,$BF,$DF,$67,$51,$FD,$EB
	.dc.b	$F0,$5F,$F2,$5B,$FE,$4F,$6B,$F1,$DA,$FE,$D1,$AF,$F8,$81,$DD,$FF
	.dc.b	$A2,$5B,$A6,$77,$FF,$06,$70,$3E,$0B,$FC,$1D,$5A,$C0,$1C,$36,$B8
	.dc.b	$CF,$F8,$57,$EC,$7F,$6D,$26,$98,$FD,$0B,$EB,$47,$48,$CF,$BD,$C4
	.dc.b	$FF,$0F,$4E,$BC,$4F,$E4,$5F,$35,$FC,$A8,$97,$FC,$77,$E6,$6B,$FB
	.dc.b	$E8,$BB,$7F,$2E,$9F,$F1,$47,$D2,$7F,$01,$AF,$F0,$F3,$67,$FD,$58
	.dc.b	$DD,$FF,$99,$FC,$DF,$0F,$06,$5F,$F8,$B2,$B4,$FF,$12,$9F,$E7,$BF
	.dc.b	$37,$23,$F9,$83,$FD,$03,$E6,$B1,$1F,$C4,$A2,$E3,$FD,$0A,$7F,$C8
	.dc.b	$5F,$60,$F9,$2A,$4B,$FA,$35,$FE,$4D,$97,$C9,$66,$7F,$D8,$F1,$C5
	.dc.b	$DE,$9F,$63,$74,$E6,$77,$5F,$E4,$BF,$94,$FE,$77,$87,$FB,$35,$B8
	.dc.b	$7F,$9A,$1B,$5F,$CC,$3B,$B8,$30,$FF,$27,$C6,$B1,$FE,$B1,$3F,$D8
	.dc.b	$09,$3F,$D4,$DF,$EC,$7F,$37,$88,$FE,$6B,$17,$0F,$EC,$B3,$FE,$5F
	.dc.b	$F8,$D7,$52,$7F,$27,$F4,$2C,$BC,$DF,$35,$FE,$7E,$8C,$65,$F9,$89
	.dc.b	$4E,$5F,$F4,$1E,$1D,$2F,$9C,$63,$0E,$D6,$C3,$9B,$DF,$0E,$4F,$56
	.dc.b	$AC,$3A,$3C,$6F,$A6,$1F,$AA,$E8,$E7,$1E,$49,$C3,$EA,$69,$FB,$0A
	.dc.b	$E7,$3E,$27,$AE,$1C,$B0,$F0,$DF,$3B,$F0,$1C,$71,$F5,$D7,$16,$CF
	.dc.b	$79,$D9,$ED,$3E,$19,$C7,$1A,$FB,$67,$E2,$E3,$E1,$8F,$39,$C4,$3D
	.dc.b	$E7,$38,$FA,$2F,$C1,$87,$1B,$F0,$CE,$5F,$01,$E4,$9F,$64,$EC,$E9
	.dc.b	$1C,$30,$F5,$BF,$4C,$F3,$CF,$03,$E8,$BB,$BC,$E3,$F2,$0F,$B3,$87
	.dc.b	$86,$7D,$D3,$A6,$1C,$49,$87,$8E,$FD,$67,$67,$CA,$66,$3C,$87,$08
	.dc.b	$D1,$C1,$CC,$70,$8F,$3F,$C6,$FF,$A0,$EF,$F1,$6D,$C7,$F2,$AD,$C3
	.dc.b	$F9,$AB,$FE,$89,$C3,$F9,$BF,$BD,$8C,$FF,$B6,$36,$3F,$C7,$2D,$7F
	.dc.b	$DC,$AD,$78,$9F,$84,$AF,$F6,$FE,$FF,$82,$8C,$E7,$FA,$49,$CC,$7E
	.dc.b	$28,$CE,$7F,$EE,$4F,$31,$7C,$67,$FD,$A4,$75,$8C,$BC,$4C,$DF,$58
	.dc.b	$CB,$FB,$28,$EA,$7F,$24,$8C,$D7,$BA,$3A,$57,$7C,$3F,$33,$87,$9B
	.dc.b	$0E,$55,$DA,$FC,$D7,$B8,$E2,$78,$EF,$8E,$1D,$A3,$E4,$9A,$BE,$73
	.dc.b	$AA,$E5,$1E,$96,$DF,$6A,$E8,$71,$7D,$18,$EE,$3F,$53,$18,$38,$47
	.dc.b	$68,$E0,$78,$AB,$81,$92,$CF,$C3,$6C,$7A,$27,$B1,$CC,$B3,$B3,$CD
	.dc.b	$3C,$DA,$66,$7E,$85,$7B,$1A,$7F,$10,$E0,$FF,$4C,$7C,$DF,$DC,$CF
	.dc.b	$FD,$3E,$DA,$FF,$4F,$FC,$E5,$BF,$F3,$DC,$DB,$E3,$5F,$E9,$98,$C7
	.dc.b	$A0,$BB,$1C,$6F,$FA,$26,$F7,$1F,$E8,$F8,$71,$3F,$B1,$CF,$F6,$A9
	.dc.b	$D1,$23,$51,$FD,$35,$FF,$0F,$BD,$1E,$38,$FF,$A3,$FE,$27,$CD,$D8
	.dc.b	$6F,$CA,$61,$95,$C2,$74,$D9,$CD,$F7,$3F,$D7,$FF,$47,$FE,$2E,$9E
	.dc.b	$ED,$8D,$60,$87,$C4,$FC,$77,$59,$FE,$AA,$75,$59,$35,$56,$FE,$1C
	.dc.b	$7F,$B3,$BE,$77,$E5,$51,$E9,$CF,$86,$57,$09,$D3,$E6,$9C,$D7,$9B
	.dc.b	$17,$CA,$F9,$4F,$36,$16,$FD,$B1,$6B,$84,$DC,$BF,$81,$C7,$F7,$8F
	.dc.b	$9C,$FB,$D8,$9F,$81,$F3,$B8,$66,$3E,$C9,$D3,$F2,$A7,$35,$F3,$C7
	.dc.b	$E1,$1F,$8B,$9F,$86,$D8,$FA,$98,$3F,$27,$7C,$87,$E3,$EB,$53,$E6
	.dc.b	$6D,$FE,$21,$B7,$FE,$17,$BE,$33,$DB,$97,$91,$EE,$9E,$0D,$7F,$CD
	.dc.b	$F5,$E3,$7F,$DD,$ED,$7F,$2B,$EF,$B8,$BF,$44,$DB,$FC,$FE,$2F,$C9
	.dc.b	$71,$7F,$06,$AC,$FC,$E6,$D7,$4B,$FD,$A4,$F1,$7A,$9F,$8F,$E9,$93
	.dc.b	$F9,$6D,$EE,$1F,$81,$9B,$7F,$93,$E2,$FA,$73,$9B,$FC,$8F,$E1,$0F
	.dc.b	$2F,$F7,$4D,$8F,$4E,$8F,$92,$CD,$49,$E2,$F9,$A7,$51,$82,$5F,$FB
	.dc.b	$2E,$FA,$36,$BE,$72,$8C,$BF,$C0,$39,$BF,$5D,$F4,$3F,$AD,$3F,$5D
	.dc.b	$F3,$17,$0F,$DA,$C5,$AF,$D0,$4D,$D3,$F3,$3F,$44,$E6,$FE,$2C,$DC
	.dc.b	$3E,$AF,$E4,$BE,$4A,$DD,$68,$93,$AD,$2E,$3F,$94,$D5,$F1,$BF,$04
	.dc.b	$F8,$5F,$1B,$6F,$E4,$C5,$AE,$13,$71,$FE,$69,$7C,$AF,$67,$D0,$9D
	.dc.b	$3F,$E8,$9E,$17,$CD,$FF,$0F,$3F,$88,$7C,$FD,$9F,$E2,$F1,$71,$FE
	.dc.b	$05,$37,$1F,$C1,$7B,$9F,$C5,$CF,$E3,$1F,$2B,$70,$FE,$3F,$16,$FF
	.dc.b	$85,$4D,$AF,$97,$F4,$5F,$28,$FC,$07,$C8,$DC,$3F,$8D,$F1,$27,$18
	.dc.b	$7C,$7A,$B7,$CA,$FF,$9E,$CF,$F3,$EB,$E1,$18,$8F,$F6,$FC,$61,$C3
	.dc.b	$5F,$C3,$69,$FE,$B7,$9F,$E6,$FE,$9F,$F3,$3F,$4B,$F9,$A3,$FD,$57
	.dc.b	$E3,$BF,$9B,$CF,$F3,$06,$DF,$F3,$CF,$17,$B6,$5F,$C8,$27,$2F,$EA
	.dc.b	$7E,$67,$F9,$F1,$FC,$FF,$E8,$EE,$1F,$D1,$BE,$67,$FA,$11,$FD,$0F
	.dc.b	$7B,$1D,$FC,$7F,$11,$FD,$1F,$E7,$6C,$7F,$4E,$F2,$3F,$A5,$1F,$DC
	.dc.b	$7E,$A7,$FC,$6F,$19,$3F,$B0,$AF,$A3,$D1,$BF,$86,$B3,$ED,$F0,$3E
	.dc.b	$01,$FD,$D7,$EB,$DF,$43,$89,$E7,$F1,$BD,$CF,$07,$A5,$63,$FB,$7F
	.dc.b	$CE,$7C,$47,$FD,$B4,$DE,$67,$8E,$86,$FF,$E0,$97,$52,$E6,$FC,$CD
	.dc.b	$FD,$1F,$03,$FC,$11,$F6,$4F,$73,$EB,$47,$F8,$47,$F1,$AF,$A3,$5F
	.dc.b	$EE,$96,$E1,$E0,$7F,$D9,$71,$FE,$04,$73,$F4,$69,$78,$3F,$E1,$D8
	.dc.b	$C3,$F3,$7D,$0C,$4E,$AF,$FB,$9F,$E2,$0F,$92,$FF,$0E,$36,$3F,$C3
	.dc.b	$38,$72,$77,$3F,$5F,$E3,$24,$6B,$F4,$37,$CF,$E3,$1F,$F0,$76,$01
	.dc.b	$EE,$76,$FA,$1F,$34,$63,$0F,$1C,$65,$A7,$8A,$35,$5F,$DA,$F3,$57
	.dc.b	$BF,$DE,$7B,$34,$E5,$5E,$28,$CE,$1F,$03,$2F,$79,$F1,$FC,$67,$99
	.dc.b	$89,$D1,$2F,$D7,$3A,$8C,$3F,$E0,$8C,$BE,$0C,$7E,$EE,$51,$87,$F7
	.dc.b	$CD,$7F,$81,$B5,$FF,$20,$AF,$A0,$FA,$67,$97,$C8,$B5,$F6,$22,$D7
	.dc.b	$92,$6D,$FF,$8C,$FC,$37,$CC,$FE,$B4,$FD,$77,$1D,$C7,$F5,$B1,$70
	.dc.b	$FF,$1F,$D7,$E0,$5B,$1F,$E5,$FF,$29,$EB,$FF,$AA,$3E,$15,$F3,$3F
	.dc.b	$E5,$89,$CE,$3F,$EC,$28,$C7,$BB,$B7,$57,$71,$FC,$C5,$F0,$58,$74
	.dc.b	$73,$3B,$7F,$C7,$B7,$6B,$85,$6E,$5F,$50,$F1,$7F,$CA,$F1,$D3,$1F
	.dc.b	$F3,$6C,$E7,$E0,$FC,$D3,$E2,$F5,$BB,$4E,$63,$81,$FA,$D7,$18,$77
	.dc.b	$6D,$FC,$1A,$E3,$87,$82,$B4,$D9,$F0,$B6,$FC,$1A,$78,$9F,$4C,$65
	.dc.b	$B8,$E7,$1E,$F1,$C0,$F2,$9A,$2B,$A7,$18,$7C,$5C,$99,$79,$AF,$A6
	.dc.b	$1F,$1A,$FF,$AA,$E1,$F0,$5B,$19,$7D,$33,$F9,$DF,$89,$7C,$A7,$EF
	.dc.b	$67,$FA,$D4,$63,$AF,$53,$99,$FE,$C3,$39,$7F,$B0,$61,$FE,$9B,$F0
	.dc.b	$FF,$50,$D8,$CF,$93,$FC,$9D,$B3,$A8,$C6,$BF,$4D,$BB,$FD,$E6,$D7
	.dc.b	$D9,$F2,$BF,$D5,$B5,$BB,$F9,$95,$AA,$FC,$4B,$70,$FE,$93,$1A,$98
	.dc.b	$FC,$E9,$D5,$F4,$C3,$C1,$5A,$C3,$B5,$FD,$4A,$D3,$F8,$BB,$EC,$4A
	.dc.b	$FB,$D3,$A9,$7D,$47,$C1,$7E,$FB,$6F,$F4,$0D,$C3,$FD,$92,$BE,$2B
	.dc.b	$93,$9A,$FE,$8C,$D7,$AE,$1A,$9F,$2F,$98,$F8,$BF,$E8,$95,$FD,$CB
	.dc.b	$C2,$FF,$6A,$F5,$FF,$0D,$6F,$37,$F5,$1B,$E1,$D7,$8B,$FE,$CF,$87
	.dc.b	$27,$FD,$96,$DC,$3F,$D9,$0D,$BE,$4F,$F3,$C9,$FF,$63,$FD,$43,$D0
	.dc.b	$FE,$59,$08,$C3,$FC,$BD,$AF,$F8,$AD,$AF,$F8,$C1,$CD,$FF,$AF,$9B
	.dc.b	$7F,$EF,$06,$CF,$F5,$7D,$DF,$25,$EB,$8F,$07,$99,$ED,$E7,$7D,$07
	.dc.b	$F9,$5A,$7F,$B3,$B5,$FF,$60,$61,$72,$F0,$5F,$AA,$35,$3E,$58,$C1
	.dc.b	$1F,$82,$F8,$63,$96,$5F,$DF,$F0,$FE,$4F,$F2,$1E,$3D,$3F,$2F,$87
	.dc.b	$F7,$06,$E2,$BE,$11,$87,$E0,$7A,$5D,$4F,$21,$2B,$E6,$3F,$D6,$F0
	.dc.b	$FF,$80,$3C,$FB,$9F,$AB,$3F,$0F,$AD,$3F,$81,$E2,$FC,$57,$D0,$DE
	.dc.b	$23,$4F,$95,$F5,$1F,$97,$C7,$EF,$FB,$1F,$14,$8F,$DD,$7C,$58,$BE
	.dc.b	$9F,$C0,$49,$FF,$11,$6F,$F8,$CF,$8A,$F1,$7D,$43,$C9,$A4,$AD,$F0
	.dc.b	$FE,$66,$5E,$F7,$DF,$EF,$FF,$0D,$D7,$F7,$B6,$BF,$E7,$76,$BF,$E7
	.dc.b	$C6,$BF,$A3,$EA,$F3,$3D,$8F,$F2,$4B,$85,$97,$D5,$73,$9F,$C9,$97
	.dc.b	$D3,$FC,$98,$93,$CF,$8B,$F6,$D1,$F3,$7E,$A3,$D1,$E6,$FE,$9B,$2F
	.dc.b	$F0,$55,$FA,$63,$07,$F8,$56,$FF,$7D,$87,$D5,$24,$F5,$E4,$FF,$5F
	.dc.b	$B7,$E1,$FE,$E7,$9D,$F6,$A3,$FC,$EB,$CD,$F7,$67,$F4,$DF,$51,$FC
	.dc.b	$26,$7F,$C4,$4D,$7F,$D3,$CD,$7F,$D4,$6D,$7F,$80,$FA,$5F,$87,$23
	.dc.b	$E3,$CF,$F1,$9F,$2B,$FC,$8F,$87,$FC,$3D,$CD,$E4,$33,$E3,$78,$BE
	.dc.b	$9B,$F8,$3E,$46,$93,$1F,$F5,$87,$03,$1F,$F2,$4C,$7C,$D6,$BF,$5B
	.dc.b	$9F,$F8,$F4,$63,$2F,$B1,$5C,$78,$BF,$83,$7D,$47,$C5,$C3,$81,$FE
	.dc.b	$75,$F1,$3F,$C2,$D7,$F7,$E3,$F8,$4D,$70,$E1,$91,$FE,$49,$BE,$4F
	.dc.b	$FB,$21,$AF,$F9,$FC,$38,$7F,$D0,$4F,$86,$BB,$37,$2E,$0D,$9F,$27
	.dc.b	$91,$F3,$3D,$89,$1F,$05,$F9,$FF,$51,$88,$F5,$E9,$F1,$E7,$18,$FC
	.dc.b	$2D,$57,$F5,$B9,$A6,$1F,$2D,$FB,$3B,$E2,$39,$D7,$E1,$6F,$95,$6A
	.dc.b	$7F,$50,$3E,$A3,$FC,$B6,$E1,$AF,$CE,$35,$EB,$7F,$37,$8F,$E9,$91
	.dc.b	$FE,$A5,$77,$5F,$45,$BB,$7C,$B6,$CE,$FC,$BE,$57,$23,$C0,$F5,$8F
	.dc.b	$E8,$9E,$97,$DC,$BE,$47,$84,$CF,$14,$BF,$5C,$ED,$FF,$5C,$35,$7D
	.dc.b	$47,$FB,$31,$FC,$97,$B5,$FA,$02,$3F,$95,$F5,$2B,$F2,$A2,$E9,$F4
	.dc.b	$E7,$CB,$E6,$7F,$59,$CD,$F3,$54,$CE,$1E,$49,$FE,$DB,$91,$A4,$8F
	.dc.b	$F4,$E4,$65,$E0,$43,$E9,$56,$5F,$BB,$F5,$1F,$B8,$23,$FA,$DF,$A8
	.dc.b	$AE,$8F,$D8,$47,$EF,$A3,$FB,$87,$D4,$3E,$27,$FB,$97,$D4,$3E,$3E
	.dc.b	$9A,$89,$87,$E2,$2B,$83,$63,$8F,$C9,$FF,$7B,$FA,$8F,$F2,$CC,$FE
	.dc.b	$3B,$87,$F0,$E9,$B1,$A8,$FC,$73,$5F,$D9,$07,$0F,$ED,$5A,$FF,$BD
	.dc.b	$5A,$EF,$83,$C4,$47,$C5,$F0,$DF,$E0,$FF,$A8,$FE,$C9,$1F,$9D,$FA
	.dc.b	$8F,$B2,$FE,$91,$AF,$E8,$63,$7C,$9F,$C3,$5B,$CF,$FC,$1D,$B2,$0E
	.dc.b	$15,$7D,$9A,$FD,$E7,$CC,$FF,$40,$F3,$A5,$7F,$25,$8F,$E6,$64,$70
	.dc.b	$7F,$29,$FA,$8F,$F1,$47,$D4,$7F,$C0,$08,$FF,$72,$35,$F9,$21,$C3
	.dc.b	$F3,$1A,$FF,$1F,$78,$5F,$F2,$96,$E5,$9F,$0B,$FA,$2F,$BD,$FD,$1B
	.dc.b	$EA,$3F,$AB,$FA,$53,$1F,$E7,$91,$96,$90,$FE,$27,$F5,$1F,$E1,$9F
	.dc.b	$A8,$FC,$74,$7F,$46,$6B,$F8,$0F,$21,$18,$8F,$E6,$91,$8E,$9F,$ED
	.dc.b	$D3,$C3,$5F,$E8,$ED,$D3,$FC,$D0,$DB,$5F,$4F,$9A,$FF,$28,$7D,$17
	.dc.b	$F9,$82,$4D,$23,$FE,$64,$7F,$27,$87,$F4,$D6,$E1,$FD,$1D,$B3,$DB
	.dc.b	$C8,$FB,$D3,$FD,$73,$C2,$EF,$30,$FE,$C4,$FF,$70,$23,$83,$FB,$37
	.dc.b	$D4,$7F,$99,$7E,$27,$F5,$18,$FE,$37,$CD,$F3,$D7,$E3,$F2,$7C,$D7
	.dc.b	$F5,$99,$CC,$FB,$23,$13,$FD,$76,$F9,$CF,$FA,$6E,$31,$DD,$FE,$28
	.dc.b	$87,$D4,$72,$9E,$B1,$F9,$D8,$C5,$FF,$67,$7B,$DF,$46,$DF,$37,$75
	.dc.b	$C6,$FE,$45,$F1,$7C,$38,$E6,$7F,$55,$87,$9B,$1E,$31,$E7,$3E,$05
	.dc.b	$FB,$1D,$E3,$E9,$9F,$0D,$F3,$9F,$84,$5C,$3D,$6F,$B9,$F0,$B4,$E7
	.dc.b	$1B,$BF,$34,$78,$0E,$98,$7B,$EB,$8D,$F9,$88,$E0,$EB,$8F,$A8,$8E
	.dc.b	$09,$5D,$4F,$11,$E8,$2D,$77,$7C,$A6,$A3,$D7,$3D,$23,$2F,$C2,$7E
	.dc.b	$41,$CE,$BC,$AF,$27,$51,$C6,$3C,$0E,$27,$D5,$1E,$59,$E2,$AB,$EB
	.dc.b	$4E,$A3,$B3,$F7,$9F,$CC,$39,$8F,$C8,$FA,$9F,$CC,$9B,$5F,$4A,$31
	.dc.b	$97,$E4,$DF,$E6,$F7,$FC,$B7,$DB,$FE,$5D,$18,$84,$62,$BF,$90,$B6
	.dc.b	$3F,$A5,$B5,$FF,$1F,$3E,$47,$FD,$3B,$E7,$FF,$A1,$E7,$F5,$31,$DC
	.dc.b	$F9,$3A,$7F,$BE,$77,$E1,$1C,$8C,$D7,$1C,$FC,$31,$89,$DB,$CF,$4F
	.dc.b	$DA,$DF,$38,$FB,$E6,$6B,$FC,$A5,$19,$9F,$1C,$FB,$8E,$87,$DC,$3C
	.dc.b	$D8,$74,$C3,$C5,$1F,$00,$FC,$5D,$73,$BF,$2D,$EF,$0F,$D3,$57,$84
	.dc.b	$F8,$6E,$1F,$04,$6D,$F3,$98,$9F,$3D,$FF,$AB,$E7,$CF,$8F,$EC,$70
	.dc.b	$F0,$5F,$8F,$1E,$46,$9F,$5B,$6B,$B1,$70,$EB,$8F,$8F,$0E,$98,$F7
	.dc.b	$9F,$59,$AC,$3C,$B3,$F5,$B1,$F9,$D8,$F7,$3C,$35,$C2,$BD,$E7,$03
	.dc.b	$E3,$DF,$57,$E5,$63,$E4,$9E,$11,$FE,$CE,$9D,$3F,$83,$AE,$AF,$BE
	.dc.b	$7C,$91,$FC,$E3,$8B,$FD,$49,$E5,$FF,$B5,$71,$FF,$3F,$37,$3F,$F5
	.dc.b	$1F,$D3,$5C,$7F,$A6,$4D,$BE,$35,$FE,$7A,$BF,$F1,$63,$FE,$AD,$9E
	.dc.b	$15,$FE,$E2,$DA,$D6,$A3,$FC,$F8,$D3,$C3,$5C,$E7,$67,$FA,$AE,$7F
	.dc.b	$26,$D7,$FC,$3D,$3C,$DF,$EC,$B1,$9B,$75,$BE,$0F,$D0,$DC,$3F,$DD
	.dc.b	$CB,$E5,$73,$32,$7F,$D9,$33,$BC,$3F,$C0,$8D,$7F,$C0,$6D,$7F,$C6
	.dc.b	$16,$F9,$FF,$F1,$95,$7F,$BE,$DB,$3F,$EA,$E5,$F0,$7F,$2D,$CE,$7F
	.dc.b	$ED,$06,$D7,$4F,$7D,$F0,$1B,$C3,$E9,$46,$27,$FD,$57,$8B,$EA,$3E
	.dc.b	$8B,$63,$C5,$D1,$89,$91,$FE,$36,$BF,$F0,$AC,$7C,$D6,$C6,$6B,$94
	.dc.b	$7F,$89,$71,$7C,$8F,$A4,$F5,$5D,$BF,$81,$1E,$6F,$A3,$C5,$FF,$B3
	.dc.b	$B2,$47,$E1,$B2,$7C,$9D,$3E,$27,$D2,$5C,$7F,$D8,$F3,$A7,$F0,$27
	.dc.b	$C5,$F8,$5D,$63,$D5,$FC,$8F,$D3,$B9,$E9,$F9,$BE,$17,$D4,$3E,$2F
	.dc.b	$38,$FF,$B5,$2F,$83,$F1,$5B,$8F,$F9,$71,$AF,$FA,$19,$AF,$FB,$E5
	.dc.b	$DD,$FE,$B7,$E9,$95,$DA,$93,$98,$FD,$3F,$CD,$FE,$4D,$FF,$72,$B6
	.dc.b	$77,$1F,$A0,$AC,$57,$49,$FF,$40,$56,$2B,$FD,$05,$08,$C3,$E0,$A7
	.dc.b	$A7,$E0,$EA,$F8,$91,$8B,$FC,$C7,$77,$E5,$4F,$CB,$70,$D8,$C9,$30
	.dc.b	$FC,$E4,$65,$FF,$7C,$7D,$43,$D1,$FE,$06,$DA,$FB,$33,$E3,$F8,$9C
	.dc.b	$D4,$9F,$CA,$56,$EB,$FD,$16,$D7,$FD,$6A,$D7,$DE,$EF,$5F,$A1,$E9
	.dc.b	$F3,$9F,$92,$7C,$17,$F6,$1F,$35,$2F,$F6,$DF,$27,$6B,$F6,$D3,$6F
	.dc.b	$F6,$93,$6B,$E6,$FD,$3B,$A3,$F8,$FB,$63,$DB,$4F,$A8,$F2,$47,$0A
	.dc.b	$D9,$F9,$3F,$AE,$7C,$EE,$27,$8C,$61,$E7,$C5,$FA,$86,$F6,$7D,$8E
	.dc.b	$BF,$B2,$C8,$C9,$C2,$31,$B7,$C2,$CD,$7C,$E3,$0D,$8F,$CC,$4F,$07
	.dc.b	$F8,$A7,$CC,$FF,$17,$3F,$8C,$65,$70,$FE,$3D,$17,$0F,$E3,$D3,$6B
	.dc.b	$F8,$E7,$BF,$FC,$7C,$FE,$41,$F3,$F7,$0F,$E4,$D1,$6F,$F9,$34,$DA
	.dc.b	$FC,$F4,$37,$7D,$47,$D0,$3F,$33,$F4,$57,$1F,$C6,$D6,$E7,$F8,$E5
	.dc.b	$70,$9F,$E4,$0D,$9F,$F6,$BB,$5F,$F8,$02,$E6,$FF,$B3,$7C,$2B,$DD
	.dc.b	$ED,$5F,$2E,$31,$7F,$CF,$FD,$AF,$E7,$87,$F3,$DF,$66,$C7,$93,$E9
	.dc.b	$7F,$A0,$BF,$CC,$5B,$7F,$EE,$86,$BF,$BA,$0E,$1F,$E9,$FE,$53,$E6
	.dc.b	$7F,$A1,$ED,$FD,$2D,$B7,$FD,$47,$E6,$3F,$A7,$1F,$D3,$F7,$B1,$E3
	.dc.b	$F2,$7F,$A0,$65,$FE,$BA,$8F,$24,$60,$FE,$D9,$8F,$F2,$78,$FD,$CF
	.dc.b	$9A,$F2,$DF,$41,$FD,$6F,$CA,$FE,$AA,$FF,$56,$FA,$89,$7C,$29,$F1
	.dc.b	$38,$E0,$FD,$A9,$EE,$F0,$7C,$64,$AF,$8C,$FF,$7D,$A7,$C7,$7F,$62
	.dc.b	$3F,$B1,$FA,$B6,$BC,$BF,$4F,$F6,$5F,$EE,$AD,$C3,$FD,$E6,$BE,$7B
	.dc.b	$FC,$38,$E6,$3E,$7F,$7F,$C4,$25,$7D,$77,$FC,$1D,$F4,$DF,$63,$B3
	.dc.b	$F9,$97,$0A,$31,$5F,$E0,$47,$E6,$5F,$F1,$29,$D7,$AF,$FE,$2D,$FA
	.dc.b	$04,$BA,$B0,$DC,$3E,$47,$A1,$89,$FF,$16,$FD,$2B,$E4,$FF,$BC,$36
	.dc.b	$BF,$BE,$36,$FF,$14,$1C,$3F,$D1,$B2,$FB,$5E,$9F,$F8,$30,$F2,$7D
	.dc.b	$33,$E8,$31,$87,$F3,$8D,$07,$F8,$B9,$FD,$7E,$2F,$A4,$F6,$4E,$72
	.dc.b	$FB,$B3,$9A,$DD,$7A,$08,$FA,$1F,$F1,$F5,$3D,$FF,$EB,$53,$E1,$6D
	.dc.b	$F4,$AF,$95,$18,$E1,$FE,$0F,$87,$A6,$FC,$B8,$FC,$F9,$79,$1F,$2A
	.dc.b	$24,$FC,$87,$E5,$91,$C1,$F0,$7F,$06,$FE,$69,$97,$B2,$7E,$D4,$62
	.dc.b	$FC,$3F,$45,$F5,$8F,$07,$8B,$70,$F8,$F1,$F6,$08,$FD,$87,$F3,$DF
	.dc.b	$37,$FE,$3E,$3F,$C7,$FC,$56,$7F,$55,$1F,$E4,$2C,$71,$97,$F8,$B5
	.dc.b	$FF,$38,$FB,$EF,$57,$FC,$0D,$F0,$EF,$9B,$FF,$02,$CE,$71,$FF,$6E
	.dc.b	$46,$3D,$CD,$BA,$BB,$8F,$E5,$AF,$83,$FE,$E0,$9D,$5F,$28,$D5,$F0
	.dc.b	$7F,$C8,$F7,$7E,$9F,$F7,$3D,$7B,$94,$61,$E8,$77,$F4,$F8,$78,$5F
	.dc.b	$E6,$88,$C9,$C1,$C6,$9E,$D8,$FD,$5D,$73,$8C,$47,$F9,$62,$B9,$DF
	.dc.b	$86,$B5,$9F,$F8,$A1,$B5,$E1,$CF,$58,$F6,$AD,$36,$7B,$B5,$ED,$3C
	.dc.b	$2E,$47,$A2,$FA,$BA,$31,$5C,$63,$E1,$47,$F4,$16,$BE,$09,$FF,$3C
	.dc.b	$46,$23,$F8,$4F,$C5,$BE,$5F,$F6,$B1,$FC,$02,$31,$D5,$FE,$7F,$6C
	.dc.b	$66,$76,$F7,$F1,$1F,$19,$D9,$B1,$FC,$2B,$0F,$F5,$64,$E4,$91,$ED
	.dc.b	$9D,$ED,$F4,$2B,$73,$F0,$7C,$AF,$F2,$AC,$E6,$B7,$B7,$D1,$6D,$7D
	.dc.b	$A6,$BF,$86,$AE,$33,$BC,$3E,$D5,$6B,$0E,$F8,$78,$6F,$78,$73,$BD
	.dc.b	$91,$F2,$16,$7C,$91,$70,$FC,$25,$72,$A7,$D3,$FC,$47,$F2,$6D,$BF
	.dc.b	$C7,$B6,$BF,$CE,$6B,$E2,$F9,$39,$AF,$E8,$2D,$7A,$E1,$A9,$F4,$F8
	.dc.b	$4F,$8C,$FE,$81,$5F,$DA,$FC,$1F,$F3,$FF,$6F,$F5,$2F,$2B,$F9,$94
	.dc.b	$FF,$76,$AD,$BF,$D4,$5B,$3B,$AE,$3C,$BF,$EC,$F6,$D3,$0F,$07,$66
	.dc.b	$21,$F4,$EF,$33,$7C,$CF,$A4,$FF,$5D,$D7,$F4,$76,$BF,$DF,$AB,$83
	.dc.b	$FE,$36,$6B,$FD,$59,$AF,$F8,$85,$AF,$FA,$71,$AF,$F6,$85,$E5,$F8
	.dc.b	$A6,$CF,$86,$3D,$13,$88,$F1,$46,$C8,$F8,$DD,$91,$C1,$FD,$87,$8D
	.dc.b	$F2,$3F,$03,$1F,$F1,$0C,$FF,$87,$7C,$2F,$93,$A7,$F8,$95,$D7,$37
	.dc.b	$8B,$CE,$B5,$FD,$AE,$F8,$BF,$16,$34,$FC,$56,$BF,$E9,$AD,$5D,$8E
	.dc.b	$AE,$FC,$8F,$A4,$F0,$F9,$7F,$95,$E6,$FF,$0A,$B7,$9B,$FC,$1F,$BF
	.dc.b	$E2,$EB,$E2,$FA,$5F,$6B,$E9,$FF,$2D,$AF,$F7,$F6,$BF,$66,$3E,$FE
	.dc.b	$8F,$85,$E3,$3F,$6E,$73,$1F,$0B,$E9,$FF,$CA,$D3,$FE,$12,$C5,$18
	.dc.b	$8F,$F2,$F4,$63,$CA,$FF,$0B,$C7,$E4,$67,$4F,$F9,$8F,$C0,$FF,$0A
	.dc.b	$3F,$E1,$3F,$C6,$AB,$F1,$5B,$BF,$FA,$22,$6E,$DF,$03,$A3,$D1,$7E
	.dc.b	$36,$BE,$26,$BF,$E2,$56,$BF,$23,$87,$77,$FE,$40,$6D,$67,$4F,$A7
	.dc.b	$F3,$5F,$A7,$C5,$D1,$FA,$6F,$13,$FD,$E9,$97,$F8,$AD,$AF,$F9,$8A
	.dc.b	$76,$FE,$E7,$E4,$BE,$E9,$1F,$DE,$7C,$37,$E0,$3E,$17,$F1,$FE,$1F
	.dc.b	$E0,$67,$FC,$DD,$CF,$97,$FD,$47,$F5,$1F,$C2,$5F,$F3,$CF,$4F,$DC
	.dc.b	$E1,$FC,$2F,$8B,$FB,$DA,$D5,$7F,$07,$F3,$FF,$17,$F1,$F2,$3F,$86
	.dc.b	$47,$E7,$BE,$63,$F1,$91,$FC,$1F,$37,$C7,$F9,$0D,$79,$5D,$7E,$9B
	.dc.b	$F4,$BF,$05,$FE,$BA,$BF,$F4,$5B,$5F,$F4,$1E,$69,$ED,$9F,$E7,$38
	.dc.b	$E5,$7F,$A4,$B3,$F9,$3B,$7F,$B1,$BE,$A3,$E1,$4F,$FA,$33,$CB,$FF
	.dc.b	$00,$CF,$FB,$29,$BD,$5F,$E7,$09,$57,$0A,$E5,$C7,$FC,$7E,$F7,$5F
	.dc.b	$E7,$66,$BC,$23,$F9,$25,$78,$23,$FD,$70,$D7,$B3,$5F,$F3,$FB,$7C
	.dc.b	$06,$2F,$97,$F9,$84,$AF,$77,$D4,$57,$FD,$15,$3B,$7E,$14,$FD,$39
	.dc.b	$FA,$9D,$D2,$7E,$33,$F5,$F7,$C4,$72,$8B,$A7,$FA,$6B,$0F,$A1,$3F
	.dc.b	$EE,$AF,$A8,$7B,$3F,$D0,$5B,$19,$8C,$7A,$19,$64,$64,$E5,$18,$7F
	.dc.b	$D7,$F3,$C9,$B3,$C9,$FF,$7C,$FA,$BF,$3A,$7F,$E0,$0E,$0E,$1F,$B1
	.dc.b	$18,$7E,$D4,$FD,$43,$FC,$7F,$5F,$EE,$96,$BA,$E2,$FE,$1A,$FF,$B2
	.dc.b	$9F,$E1,$FB,$FD,$2F,$77,$F3,$84,$9F,$C7,$78,$15,$FA,$31,$72,$F9
	.dc.b	$13,$FD,$97,$0F,$E8,$B5,$FE,$BC,$9F,$43,$FE,$BE,$D1,$F3,$5F,$DA
	.dc.b	$DB,$1F,$03,$D1,$E4,$66,$64,$64,$87,$FA,$F2,$FB,$B6,$7D,$B1,$88
	.dc.b	$FF,$5F,$6E,$FA,$7F,$DC,$12,$3F,$55,$F4,$F5,$FC,$6C,$5C,$BF,$09
	.dc.b	$3F,$E0,$1F,$A7,$FF,$7C,$D7,$B7,$83,$F5,$5F,$1E,$F8,$FF,$F0,$33
	.dc.b	$67,$47,$FB,$7A,$B8,$BF,$F0,$01,$AF,$EC,$C3,$87,$F6,$CD,$79,$7C
	.dc.b	$4F,$DA,$FA,$7F,$F3,$04,$76,$3F,$D9,$1C,$1F,$E1,$7E,$2F,$F0,$FE
	.dc.b	$5F,$BF,$AD,$D3,$4F,$AA,$A3,$31,$8A,$E3,$3F,$EF,$66,$BB,$E2,$FF
	.dc.b	$12,$57,$48,$F5,$35,$F0,$47,$9C,$FC,$8F,$D3,$FF,$31,$C3,$F8,$97
	.dc.b	$D3,$FE,$0A,$BC,$91,$8E,$4F,$F1,$85,$6E,$41,$FE,$EA,$FC,$96,$BF
	.dc.b	$C7,$5A,$F0,$6F,$2F,$F1,$0F,$9B,$FE,$93,$E7,$CA,$FE,$59,$1F,$CF
	.dc.b	$4F,$D2,$F2,$9A,$48,$FE,$EF,$F4,$DF,$C4,$DF,$49,$FE,$49,$F2,$F1
	.dc.b	$C1,$FE,$12,$F0,$3F,$C9,$55,$8D,$7F,$99,$3C,$6F,$F0,$4D,$73,$3F
	.dc.b	$B8,$8C,$47,$F9,$57,$17,$D3,$7F,$62,$F4,$64,$FF,$45,$8F,$EA,$C6
	.dc.b	$1F,$2E,$AF,$2D,$5A,$9F,$CB,$CB,$E9,$7F,$35,$1C,$35,$78,$9F,$E2
	.dc.b	$E3,$FC,$7F,$3A,$DB,$F7,$F7,$C7,$AF,$F8,$6C,$F2,$C3,$FA,$84,$E5
	.dc.b	$FF,$12,$46,$5F,$EA,$9F,$3F,$FA,$3B,$FC,$AE,$CF,$A2,$F9,$27,$F9
	.dc.b	$CE,$75,$C3,$C2,$FF,$8E,$C7,$F1,$E4,$D3,$F9,$0E,$3F,$D4,$3C,$1E
	.dc.b	$53,$0F,$ED,$8F,$F7,$D3,$0F,$DC,$39,$99,$3F,$61,$F0,$CE,$BF,$D6
	.dc.b	$43,$E9,$36,$F8,$BE,$5B,$FB,$84,$63,$14,$61,$FE,$E9,$7C,$F5,$FC
	.dc.b	$FD,$72,$8C,$5F,$FA,$AE,$BE,$3D,$FF,$7B,$F4,$FA,$36,$39,$E3,$E4
	.dc.b	$9E,$4F,$9B,$0D,$DF,$6A,$E6,$D3,$FD,$DB,$3B,$8E,$78,$E6,$7D,$85
	.dc.b	$8F,$5B,$9C,$7C,$F7,$E0,$AF,$DC,$E1,$E0,$AF,$51,$FB,$CB,$E2,$E0
	.dc.b	$5F,$C0,$9E,$51,$F3,$67,$CE,$D3,$96,$3F,$2D,$F4,$DF,$8F,$1F,$54
	.dc.b	$68,$F4,$4E,$DE,$F1,$A6,$CE,$98,$77,$32,$5C,$3D,$F7,$E4,$C7,$DC
	.dc.b	$7A,$08,$E4,$93,$E6,$39,$5F,$86,$3E,$1C,$F2,$7E,$8C,$EA,$FC,$73
	.dc.b	$98,$E4,$F6,$7D,$C7,$9E,$79,$9C,$A3,$51,$F6,$8D,$61,$F0,$F0,$F5
	.dc.b	$C7,$97,$D1,$FF,$51,$C7,$E4,$9B,$8F,$E8,$1B,$8F,$F3,$67,$F9,$9F
	.dc.b	$B1,$73,$FE,$BD,$82,$30,$FF,$61,$6B,$FE,$55,$6B,$FD,$01,$FC,$85
	.dc.b	$3C,$BF,$F0,$6E,$7F,$AE,$AD,$1F,$56,$75,$BF,$E5,$E3,$33,$FB,$49
	.dc.b	$CC,$7D,$B3,$3A,$FF,$99,$63,$33,$D6,$7D,$07,$5B,$F0,$D7,$B3,$0F
	.dc.b	$2C,$62,$7D,$86,$F0,$F1,$3C,$AF,$C5,$3B,$8F,$80,$6C,$F2,$1B,$BF
	.dc.b	$24,$6E,$3C,$E7,$F8,$BD,$C3,$B3,$15,$E0,$BF,$2B,$63,$B5,$FB,$67
	.dc.b	$31,$E0,$6C,$66,$37,$7E,$F6,$9E,$EC,$7D,$77,$FB,$E9,$D4,$62,$7F
	.dc.b	$78,$7A,$0F,$F0,$96,$1E,$8C,$3B,$D7,$5C,$38,$1E,$12,$BD,$0F,$87
	.dc.b	$87,$7B,$F1,$46,$1E,$E7,$C3,$3B,$C7,$C1,$9E,$71,$FD,$AF,$BB,$FC
	.dc.b	$AF,$97,$F9,$C3,$4B,$7F,$E1,$0E,$0F,$F4,$BD,$62,$FF,$CB,$D5,$67
	.dc.b	$35,$FE,$33,$BF,$E8,$8F,$FA,$DE,$F8,$5F,$F8,$C6,$ED,$6A,$3F,$4A
	.dc.b	$D7,$FB,$14,$7F,$8D,$9F,$CA,$35,$F0,$D7,$7F,$03,$9C,$F4,$7C,$9D
	.dc.b	$8F,$7F,$86,$5F,$EA,$F6,$C6,$6F,$F6,$B3,$96,$C6,$70,$FF,$01,$6D
	.dc.b	$D2,$FE,$07,$4F,$EA,$B3,$FF,$00,$FC,$B7,$F8,$DB,$1F,$F0,$5F,$63
	.dc.b	$D9,$F8,$D3,$9C,$BF,$C2,$73,$9A,$FF,$08,$CE,$6F,$FC,$17,$39,$AF
	.dc.b	$F0,$CC,$E6,$BE,$A4,$E6,$BF,$CB,$7A,$3E,$93,$E3,$60,$8C,$5F,$FA
	.dc.b	$82,$73,$18,$9F,$83,$1E,$CF,$43,$EB,$CE,$6F,$FC,$3B,$9F,$FB,$D7
	.dc.b	$0F,$3F,$C7,$7C,$CC,$CD,$24,$7C,$C8,$CB,$F1,$3D,$BF,$87,$3F,$89
	.dc.b	$D1,$1F,$86,$EA,$7C,$77,$A9,$B7,$FC,$AB,$6F,$F0,$4C,$E4,$91,$F4
	.dc.b	$3D,$CE,$51,$D5,$FC,$8F,$D1,$FF,$8A,$E7,$93,$6B,$F9,$AB,$5F,$F4
	.dc.b	$A3,$5F,$F3,$88,$78,$6B,$FA,$8F,$9F,$FF,$6B,$57,$FA,$4E,$72,$FF
	.dc.b	$B0,$E3,$35,$FE,$2B,$9C,$BF,$E9,$38,$CD,$7F,$A5,$27,$2A,$33,$5E
	.dc.b	$2F,$69,$F2,$DF,$79,$B1,$F1,$5B,$87,$F5,$06,$BF,$EA,$F7,$37,$E9
	.dc.b	$E0,$FD,$AC,$62,$BD,$11,$F9,$8E,$C9,$1F,$96,$D2,$E1,$F9,$49,$B8
	.dc.b	$7E,$8A,$6D,$7C,$8F,$A3,$FF,$45,$3F,$65,$B7,$FD,$91,$AF,$FB,$0D
	.dc.b	$AF,$1F,$2F,$F8,$85,$FC,$3F,$33,$F8,$5D,$72,$8F,$A5,$18,$9D,$EA
	.dc.b	$8C,$61,$FE,$D4,$AE,$4F,$EC,$3E,$4B,$F6,$A7,$ED,$BE,$2E,$C6,$49
	.dc.b	$97,$EC,$A3,$2F,$F9,$97,$E7,$DF,$45,$FB,$06,$D7,$E5,$DB,$3F,$E0
	.dc.b	$26,$BF,$EE,$17,$38,$7F,$32,$FA,$27,$CA,$FF,$28,$9C,$DF,$FB,$0A
	.dc.b	$95,$A8,$F6,$69,$FA,$A6,$F3,$FC,$2E,$6F,$86,$C8,$C9,$C2,$31,$97
	.dc.b	$FA,$EB,$0F,$C8,$56,$1F,$CF,$4D,$AF,$E8,$9D,$4F,$91,$FE,$3A,$7F
	.dc.b	$1E,$D2,$CF,$F1,$C8,$B8,$FF,$1A,$9B,$87,$F2,$8F,$A2,$FE,$4A,$7F
	.dc.b	$26,$F9,$8B,$87,$F2,$48,$B7,$FC,$8A,$6D,$7F,$BE,$B6,$7C,$FF,$D1
	.dc.b	$26,$1F,$45,$B7,$ED,$F7,$5F,$45,$F3,$67,$F1,$B1,$8A,$F9,$C7,$F0
	.dc.b	$0B,$EA,$FE,$26,$35,$7F,$81,$AE,$37,$FC,$FB,$D4,$FE,$8B,$E5,$5C
	.dc.b	$3F,$D9,$BD,$9F,$D7,$F3,$47,$F0,$CF,$89,$7C,$8F,$F6,$BF,$11,$F2
	.dc.b	$9F,$DA,$27,$35,$FE,$D6,$9C,$BF,$D2,$3E,$47,$F9,$57,$CE,$FF,$4E
	.dc.b	$6C,$7F,$4C,$FA,$1F,$EA,$FE,$87,$F5,$26,$C7,$BB,$84,$31,$FF,$5A
	.dc.b	$C7,$A3,$E1,$BC,$5E,$4F,$F6,$07,$FA,$DF,$CA,$7F,$39,$F9,$7B,$5F
	.dc.b	$D8,$1F,$E7,$9F,$43,$23,$FB,$1F,$C3,$DC,$BF,$D9,$99,$3C,$4F,$F0
	.dc.b	$67,$CC,$3C,$67,$CA,$7C,$19,$CC,$63,$1F,$ED,$BF,$19,$FD,$24,$FE
	.dc.b	$95,$EB,$58,$FF,$7D,$FD,$0F,$D8,$CF,$F9,$C9,$FD,$8D,$C3,$FE,$F4
	.dc.b	$6B,$FE,$1F,$6B,$F1,$33,$FC,$F4,$FC,$7C,$DF,$89,$BD,$1F,$AD,$E3
	.dc.b	$7A,$A3,$15,$F3,$7D,$57,$D1,$7D,$AC,$BC,$71,$8E,$DF,$F0,$AC,$E6
	.dc.b	$7F,$A7,$CE,$6B,$FD,$13,$97,$D6,$EB,$FF,$15,$FD,$0B,$E3,$BE,$4E
	.dc.b	$09,$CC,$4B,$F4,$7D,$17,$C9,$AF,$F7,$E5,$F1,$9C,$F2,$FF,$73,$F0
	.dc.b	$BF,$C3,$BF,$23,$67,$F3,$2F,$83,$E7,$E4,$7C,$3D,$1F,$01,$8C,$BA
	.dc.b	$D7,$DC,$F1,$BE,$1C,$62,$FE,$54,$63,$C6,$FD,$1C,$B0,$4E,$6B,$FC
	.dc.b	$A5,$97,$F3,$F7,$F2,$F3,$AE,$5F,$F1,$9F,$AD,$F1,$35,$95,$F5,$F2
	.dc.b	$79,$B8,$CF,$E0,$D7,$E4,$3E,$8B,$F9,$54,$62,$FE,$F4,$62,$FC,$B2
	.dc.b	$E0,$F9,$51,$E8,$CD,$5B,$BF,$F0,$3D,$69,$F3,$C6,$6F,$F5,$F2,$F9
	.dc.b	$CF,$F1,$BB,$6B,$FC,$3C,$D9,$FF,$64,$B5,$FC,$DB,$9C,$3C,$56,$F2
	.dc.b	$FF,$0B,$87,$B5,$FF,$17,$7C,$BF,$F8,$BF,$3B,$97,$F9,$59,$B9,$7F
	.dc.b	$95,$1B,$8F,$F9,$49,$FF,$3C,$F9,$7F,$B5,$AF,$F6,$A7,$CB,$7F,$89
	.dc.b	$23,$17,$FE,$DD,$8C,$7B,$9F,$73,$E1,$D1,$BF,$13,$0D,$C3,$9D,$FF
	.dc.b	$12,$C3,$9D,$F0,$8C,$61,$C1,$D6,$15,$7D,$13,$C5,$FE,$57,$1E,$33
	.dc.b	$F8,$EC,$67,$C0,$EF,$1F,$E5,$8F,$5B,$94,$E6,$38,$1D,$2B,$9E,$3E
	.dc.b	$38,$C5,$F2,$CB,$F1,$33,$9B,$E4,$DB,$E3,$A7,$F9,$2A,$72,$FC,$9C
	.dc.b	$7C,$33,$AC,$7C,$ED,$7B,$9E,$4C,$3B,$D7,$39,$D5,$F4,$AE,$38,$F9
	.dc.b	$70,$EA,$FC,$09,$F7,$4F,$F2,$46,$BC,$E3,$FC,$DD,$18,$8F,$E8,$BF
	.dc.b	$16,$F4,$3F,$6D,$1F,$CB,$04,$67,$9B,$FC,$EC,$D8,$D3,$5F,$07,$90
	.dc.b	$F8,$BE,$ED,$8E,$98,$7F,$AE,$67,$24,$8F,$AB,$19,$6C,$F1,$DB,$E9
	.dc.b	$79,$BF,$E7,$87,$EA,$46,$76,$FA,$4D,$AF,$C2,$B5,$FC,$5D,$71,$9D
	.dc.b	$E1,$F2,$6B,$58,$78,$70,$EF,$86,$6A,$3E,$47,$E8,$61,$CE,$3B,$C6
	.dc.b	$A7,$CA,$EA,$3F,$47,$E2,$BB,$9F,$51,$C6,$31,$EA,$D9,$FF,$0D,$D7
	.dc.b	$07,$6D,$AD,$4C,$8C,$1B,$F9,$3F,$F4,$EF,$4F,$FA,$FB,$EA,$2F,$4F
	.dc.b	$F3,$06,$D7,$FB,$2F,$15,$E3,$2F,$33,$63,$33,$FE,$20,$8C,$7C,$7E
	.dc.b	$7E,$A2,$4F,$F8,$91,$E1,$F1,$D8,$86,$7F,$ED,$09,$C3,$FE,$D2,$9E
	.dc.b	$DF,$23,$2B,$FC,$5A,$F8,$FF,$39,$FE,$FF,$5F,$ED,$88,$D4,$F4,$D7
	.dc.b	$FA,$5C,$6E,$33,$3B,$79,$9F,$DE,$6B,$9D,$7F,$89,$B4,$47,$F0,$E9
	.dc.b	$E8,$7F,$BB,$DC,$57,$F5,$C9,$DF,$4F,$97,$D9,$97,$D5,$AF,$29,$EB
	.dc.b	$DB,$EB,$24,$78,$DF,$77,$C7,$7B,$FE,$1F,$E0,$7C,$54,$AF,$44,$7C
	.dc.b	$3F,$0A,$47,$AD,$F8,$DE,$17,$C7,$F6,$24,$FB,$DF,$95,$EB,$7C,$BF
	.dc.b	$0E,$4F,$C3,$7E,$77,$C1,$FC,$FF,$32,$4F,$C7,$7E,$97,$93,$F4,$FE
	.dc.b	$9F,$EA,$7C,$04,$AF,$99,$1F,$5F,$EB,$7E,$C7,$C0,$C9,$FA,$11,$F6
	.dc.b	$FE,$0E,$47,$D3,$7E,$F7,$C1,$FD,$FF,$82,$93,$F5,$DF,$CA,$7E,$75
	.dc.b	$FD,$E3,$AB,$E9,$2F,$83,$C9,$2B,$FC,$6D,$0F,$83,$E3,$E8,$C9,$CB
	.dc.b	$D3,$CD,$EB,$F1,$D2,$78,$BE,$0F,$1F,$C3,$EA,$49,$EA,$F9,$3D,$FF
	.dc.b	$2F,$B9,$27,$C2,$FA,$3E,$87,$D3,$F4,$92,$7C,$AF,$B3,$E3,$FF,$67
	.dc.b	$0F,$A3,$7C,$B6,$1B,$1F,$E7,$D3,$51,$9F,$95,$D3,$23,$53,$C1,$FC
	.dc.b	$D9,$D6,$74,$E2,$35,$C7,$FE,$24,$6F,$97,$FE,$8A,$9E,$07,$87,$E8
	.dc.b	$D9,$FF,$47,$8F,$F4,$7C,$61,$C3,$5F,$E9,$8D,$7F,$CF,$2E,$6F,$FC
	.dc.b	$66,$DA,$F0,$36,$7F,$D3,$76,$F9,$9F,$C1,$FC,$EF,$E1,$3E,$66,$57
	.dc.b	$E0,$23,$2F,$E0,$A3,$2F,$EC,$F9,$D3,$F8,$CF,$94,$FC,$6F,$CA,$4A
	.dc.b	$FC,$4C,$65,$FC,$5C,$65,$E5,$19,$7A,$46,$74,$FC,$AC,$FE,$63,$E3
	.dc.b	$3F,$33,$EE,$CB,$FC,$AC,$65,$A4,$3C,$DF,$07,$FA,$0F,$84,$91,$F9
	.dc.b	$C7,$F9,$B7,$E7,$7F,$DB,$63,$D9,$1A,$69,$FE,$97,$8D,$63,$FE,$60
	.dc.b	$FA,$D7,$A1,$F9,$06,$D7,$E3,$5A,$FE,$21,$7A,$3B,$99,$3B,$79,$46
	.dc.b	$23,$84,$5A,$F0,$4F,$A3,$87,$D3,$F1,$52,$FC,$D1,$97,$D1,$19,$7F
	.dc.b	$95,$63,$F8,$17,$F4,$5F,$89,$BC,$8F,$D4,$FE,$88,$F0,$E9,$FA,$C9
	.dc.b	$D5,$70,$AC,$11,$FD,$65,$7F,$07,$7F,$5D,$F1,$5F,$EC,$36,$E3,$FB
	.dc.b	$68,$C4,$7E,$B1,$FF,$25,$69,$FE,$8C,$72,$D9,$E0,$71,$F8,$7B,$5D
	.dc.b	$08,$73,$7F,$81,$FE,$29,$FE,$C3,$8F,$E0,$1F,$27,$FC,$87,$B7,$F6
	.dc.b	$8D,$8F,$F7,$FC,$ED,$FD,$FB,$70,$FF,$4F,$B6,$BF,$78,$D9,$1F,$EF
	.dc.b	$7F,$33,$F1,$B8,$7E,$3B,$E4,$24,$FE,$2E,$32,$D2,$1F,$5B,$E8,$BF
	.dc.b	$2B,$E1,$C8,$FC,$93,$F9,$8F,$0F,$11,$F7,$22,$E7,$FD,$B7,$8D,$5F
	.dc.b	$9E,$F9,$19,$7F,$9B,$7F,$43,$F1,$F8,$8F,$C2,$C5,$BE,$53,$FA,$6F
	.dc.b	$97,$95,$FA,$37,$F5,$3F,$27,$88,$FC,$6C,$5A,$F1,$4F,$EB,$BE,$4E
	.dc.b	$57,$EA,$DF,$D8,$FC,$6E,$23,$F2,$B1,$6B,$D5,$3F,$B6,$F8,$D9,$5F
	.dc.b	$B3,$7F,$73,$F1,$98,$8F,$CE,$C5,$AF,$85,$3F,$D1,$3E,$47,$FD,$D3
	.dc.b	$1F,$5B,$E8,$3F,$67,$7F,$63,$BD,$F1,$1F,$CD,$E7,$33,$FB,$88,$C4
	.dc.b	$FF,$3B,$BE,$72,$AD,$C6,$1E,$33,$D2,$7C,$0F,$F4,$33,$0F,$17,$A6
	.dc.b	$FF,$E3,$E8,$C1,$FE,$8E,$71,$3D,$2F,$C7,$3E,$96,$BF,$E9,$09,$DD
	.dc.b	$74,$C3,$C3,$3B,$29,$ED,$2E,$1E,$99,$E1,$1E,$23,$45,$7D,$38,$7C
	.dc.b	$0B,$EF,$3E,$7B,$EB,$5A,$3E,$06,$1E,$6C,$3B,$56,$CE,$91,$EA,$9E
	.dc.b	$11,$9C,$3B,$5F,$4A,$D1,$CF,$0F,$0D,$7C,$19,$D3,$0F,$56,$1E,$07
	.dc.b	$47,$80,$D5,$F6,$77,$5D,$30,$F1,$C7,$C2,$9D,$12,$3D,$B1,$C9,$F9
	.dc.b	$31,$98,$FA,$07,$73,$AD,$FD,$0A,$ED,$87,$B2,$71,$31,$FF,$62,$78
	.dc.b	$1F,$53,$E4,$3F,$9D,$47,$A9,$B9,$7E,$51,$CC,$7F,$14,$F9,$1F,$E2
	.dc.b	$B1,$88,$46,$2B,$FD,$B0,$D8,$FF,$0C,$1A,$3C,$26,$1F,$C4,$DF,$F8
	.dc.b	$BB,$AF,$F1,$73,$DC,$FC,$44,$65,$FC,$51,$9C,$FF,$5B,$19,$7F,$5F
	.dc.b	$19,$D3,$FD,$63,$CC,$BE,$B3,$E3,$AE,$33,$98,$EA,$66,$B8,$D7,$0C
	.dc.b	$3C,$F5,$BB,$EF,$7B,$3F,$9C,$E1,$E3,$C3,$E1,$C7,$AA,$B8,$46,$2B
	.dc.b	$C3,$87,$5A,$D3,$63,$31,$E4,$C3,$C3,$39,$AF,$2C,$E4,$E2,$5A,$E2
	.dc.b	$79,$AB,$8C,$66,$78,$CE,$63,$8B,$4E,$31,$E7,$36,$58,$F7,$DF,$B1
	.dc.b	$FB,$87,$28,$D3,$FE,$D0,$8D,$3F,$E5,$BA,$F1,$D7,$7A,$EB,$3C,$A3
	.dc.b	$C2,$D8,$CF,$C5,$39,$DF,$23,$F6,$73,$FE,$B0,$DC,$FF,$D9,$92,$F9
	.dc.b	$1C,$62,$C3,$95,$E1,$B7,$FE,$37,$8C,$5F,$18,$C1,$FE,$3B,$F9,$1C
	.dc.b	$9A,$F9,$1B,$19,$26,$1F,$E2,$88,$CB,$FE,$2A,$F9,$1E,$A7,$87,$E3
	.dc.b	$7F,$C0,$35,$FE,$47,$8C,$B4,$FC,$69,$FE,$E7,$8D,$CF,$38,$CB,$ED
	.dc.b	$87,$C8,$75,$3B,$72,$DA,$DE,$8A,$DC,$61,$B7,$FE,$E9,$7F,$AE,$C6
	.dc.b	$6F,$D1,$F2,$12,$38,$7C,$85,$BE,$13,$78,$3F,$DF,$18,$2B,$17,$EB
	.dc.b	$F3,$7E,$11,$F0,$F8,$AC,$78,$E3,$E3,$56,$31,$F2,$3F,$71,$FE,$E3
	.dc.b	$F0,$B7,$53,$F3,$7E,$43,$0C,$C3,$DD,$F2,$16,$FD,$CF,$E1,$B4,$7C
	.dc.b	$E7,$FB,$76,$BE,$6C,$63,$E2,$35,$3F,$ED,$D9,$CB,$BB,$57,$E5,$63
	.dc.b	$0D,$AE,$3F,$11,$C9,$97,$C9,$CB,$4F,$F6,$87,$9C,$F8,$1F,$80,$D8
	.dc.b	$FB,$1F,$23,$89,$91,$F6,$FE,$46,$E7,$F6,$DF,$C8,$7C,$8F,$E4,$7E
	.dc.b	$46,$D7,$E0,$5B,$5F,$81,$6E,$9F,$69,$F8,$DF,$5E,$F3,$29,$8F,$32
	.dc.b	$E1,$F2,$27,$4D,$9D,$34,$98,$7C,$87,$F9,$5C,$F0,$8C,$1F,$31,$57
	.dc.b	$68,$FC,$B7,$C9,$7E,$AC,$FD,$67,$C8,$DB,$FC,$E4,$59,$FC,$E4,$DC
	.dc.b	$BF,$2B,$F2,$5F,$B3,$3F,$69,$ED,$5C,$3F,$49,$1F,$B7,$F9,$2C,$61
	.dc.b	$FA,$57,$F7,$5F,$25,$FB,$BF,$62,$DF,$EA,$DB,$5F,$AB,$6E,$3F,$A4
	.dc.b	$6C,$FE,$F5,$B3,$FB,$B7,$F8,$37,$4F,$F1,$7F,$CF,$7F,$AE,$57,$ED
	.dc.b	$E3,$0E,$8F,$DE,$C6,$AB,$F7,$53,$A9,$FE,$0F,$0F,$9D,$63,$FA,$B6
	.dc.b	$CF,$E7,$3E,$13,$13,$23,$FB,$ED,$7F,$A6,$2B,$6E,$1A,$FF,$80,$9A
	.dc.b	$FF,$A7,$BC,$56,$3F,$A4,$AC,$36,$7F,$9B,$3F,$AD,$7F,$03,$3C,$3A
	.dc.b	$BE,$11,$2B,$DF,$D7,$6B,$77,$CE,$F8,$56,$8F,$8B,$7F,$2F,$E5,$24
	.dc.b	$7C,$8F,$82,$B8,$72,$F9,$4F,$A0,$7D,$1F,$94,$B5,$F3,$BE,$53,$EA
	.dc.b	$1F,$57,$E5,$2C,$7D,$2F,$94,$FB,$07,$D9,$F9,$3B,$1F,$5B,$E5,$3E
	.dc.b	$E1,$F7,$7C,$2B,$1F,$6B,$E4,$3F,$00,$7E,$07,$E4,$AC,$7D,$EF,$90
	.dc.b	$FC,$21,$F8,$5F,$91,$B1,$F8,$2E,$8F,$C4,$1F,$89,$F9,$1B,$1F,$86
	.dc.b	$F9,$1F,$C6,$7C,$8F,$E3,$5B,$1F,$D1,$BF,$3E,$7C,$0F,$F6,$75,$F0
	.dc.b	$36,$30,$E7,$CE,$95,$C7,$E2,$EF,$47,$DA,$C1,$F1,$9C,$CE,$9F,$19
	.dc.b	$6F,$C1,$F0,$9D,$CF,$07,$C6,$58,$F2,$7C,$3F,$8C,$F2,$7C,$65,$8F
	.dc.b	$47,$A3,$E7,$F4,$3D,$0D,$8F,$67,$C6,$7A,$CF,$67,$C6,$58,$F8,$1F
	.dc.b	$19,$EF,$31,$BF,$A5,$F8,$5F,$19,$27,$67,$C7,$7E,$37,$C4,$48,$F9
	.dc.b	$5F,$13,$6F,$E1,$4D,$F4,$7F,$CA,$5E,$5B,$93,$FC,$21,$5A,$C3,$E9
	.dc.b	$5E,$DF,$D5,$4E,$AF,$8B,$FE,$24,$8F,$B1,$F1,$1F,$5C,$FB,$1F,$11
	.dc.b	$63,$24,$C3,$EA,$46,$5F,$AD,$F1,$1F,$7C,$FF,$13,$7E,$7F,$FE,$10
	.dc.b	$C7,$BC,$66,$FF,$C1,$73,$FE,$08,$9C,$C6,$3E,$27,$8F,$C2,$7F,$9B
	.dc.b	$FE,$31,$D9,$FC,$0F,$07,$C5,$66,$31,$57,$3F,$F1,$C5,$6B,$1F,$F4
	.dc.b	$A5,$69,$B3,$AB,$E1,$C0,$BF,$F0,$CC,$FF,$A6,$2F,$C1,$5D,$61,$F3
	.dc.b	$0F,$4F,$FC,$53,$1B,$06,$9C,$E7,$EA,$8C,$D3,$E0,$FA,$5F,$3D,$7C
	.dc.b	$B5,$C5,$B1,$98,$EC,$7C,$29,$E3,$87,$6C,$7C,$33,$C4,$DD,$71,$C7
	.dc.b	$C5,$1C,$74,$EB,$7C,$63,$D3,$5B,$BE,$6D,$8F,$4E,$1E,$0C,$78,$92
	.dc.b	$3C,$B9,$7B,$71,$C6,$5E,$0C,$3D,$55,$FE,$C0,$6B,$DB,$2F,$96,$E5
	.dc.b	$AF,$B3,$E3,$9D,$8F,$99,$FE,$9B,$29,$CC,$61,$B1,$FD,$49,$AF,$87
	.dc.b	$D6,$FE,$2E,$FF,$56,$C1,$18,$8D,$78,$DD,$A7,$2D,$9E,$39,$7B,$5B
	.dc.b	$BF,$C8,$6D,$FC,$CF,$27,$FD,$39,$39,$BF,$BB,$5A,$AF,$C6,$B7,$05
	.dc.b	$F0,$9F,$AB,$39,$25,$7E,$1A,$75,$3C,$A3,$A9,$1F,$C1,$C6,$AB,$D7
	.dc.b	$1F,$86,$AE,$6E,$8F,$D3,$F1,$3D,$57,$77,$8E,$31,$DD,$64,$D9,$95
	.dc.b	$CA,$72,$D9,$CB,$FA,$9A,$ED,$DD,$F9,$AE,$8F,$E3,$F7,$FE,$BA,$AF
	.dc.b	$F4,$9B,$5F,$EF,$EA,$78,$71,$FF,$82,$63,$13,$C3,$B7,$2F,$FA,$F6
	.dc.b	$39,$5F,$6B,$CE,$1E,$FE,$CB,$46,$98,$EC,$B8,$79,$22,$BE,$48,$E3
	.dc.b	$D9,$23,$D2,$F6,$F9,$9F,$DB,$C7,$AA,$34,$D3,$FD,$63,$1A,$CB,$E1
	.dc.b	$CE,$9B,$3A,$3A,$75,$47,$8B,$DF,$B6,$47,$46,$BF,$E2,$8B,$FE,$5A
	.dc.b	$F8,$3A,$FC,$C4,$7C,$7F,$1D,$E9,$F8,$E9,$1E,$67,$D9,$F1,$DE,$DF
	.dc.b	$8E,$93,$EA,$7E,$07,$C7,$7C,$1F,$8E,$93,$EE,$7E,$27,$C6,$FC,$5F
	.dc.b	$0E,$4F,$C2,$7E,$47,$C6,$FC,$9F,$8D,$93,$F1,$9F,$99,$F1,$9F,$37
	.dc.b	$E3,$3E,$77,$C5,$CA,$F9,$71,$F4,$7E,$33,$E9,$7C,$64,$9F,$9F,$1F
	.dc.b	$57,$E2,$A4,$7D,$27,$EC,$7C,$67,$D9,$F8,$C9,$3F,$59,$FC,$4F,$CB
	.dc.b	$BC,$2B,$F1,$52,$BF,$DD,$72,$8D,$FC,$67,$0F,$42,$56,$9E,$7E,$47
	.dc.b	$4F,$8B,$FC,$AF,$A5,$FE,$E5,$AF,$ED,$7D,$99,$7A,$FC,$5E,$BB,$24
	.dc.b	$60,$F0,$FC,$57,$98,$8F,$93,$E2,$BD,$3F,$15,$23,$CE,$FB,$3E,$2B
	.dc.b	$F5,$30,$F9,$87,$C0,$61,$B1,$FE,$88,$6B,$FE,$B4,$38,$7C,$06,$E6
	.dc.b	$46,$DE,$51,$88,$E1,$3C,$BC,$9F,$99,$1F,$DB,$A3,$1C,$BF,$D8,$15
	.dc.b	$79,$BE,$65,$9F,$97,$D7,$98,$7F,$A7,$27,$2D,$9F,$E9,$6A,$32,$FF
	.dc.b	$A6,$9A,$FE,$E1,$AF,$FA,$5F,$9D,$F1,$9F,$82,$F8,$C9,$3F,$80,$8F
	.dc.b	$C2,$98,$7F,$D2,$F2,$9C,$C7,$E1,$BE,$33,$F1,$7F,$19,$23,$04,$38
	.dc.b	$B9,$79,$C7,$E4,$8C,$6D,$FE,$48,$B5,$66,$BF,$27,$F1,$9F,$98,$F6
	.dc.b	$E4,$60,$87,$9A,$3F,$3A,$7E,$7B,$E2,$FF,$3F,$F1,$92,$7F,$3A,$FF
	.dc.b	$2A,$F9,$2F,$ED,$59,$78,$63,$53,$F0,$1D,$7B,$FF,$8C,$F9,$97,$55
	.dc.b	$25,$7F,$B4,$A1,$E9,$62,$3F,$19,$1C,$1D,$4F,$39,$D6,$1F,$88,$95
	.dc.b	$6D,$E3,$5B,$7F,$C7,$33,$FE,$6E,$C7,$FD,$33,$18,$E7,$FF,$6A,$43
	.dc.b	$E3,$7D,$C7,$BF,$E3,$A5,$6C,$7B,$AA,$E1,$19,$8C,$57,$08,$FD,$37
	.dc.b	$CC,$BC,$BF,$D7,$FC,$CF,$F8,$52,$7F,$DE,$6D,$77,$1F,$B4,$9E,$11
	.dc.b	$FE,$F8,$6B,$A3,$8F,$91,$FB,$AF,$99,$F1,$C7,$ED,$5B,$AF,$C7,$6E
	.dc.b	$BF,$E4,$C7,$FC,$93,$19,$CB,$91,$2B,$E0,$09,$DD,$71,$9D,$4F,$77
	.dc.b	$C5,$DB,$FC,$33,$E6,$1F,$27,$FC,$3B,$E6,$7E,$C6,$5F,$65,$B8,$FF
	.dc.b	$97,$AB,$78,$F8,$1A,$FF,$0F,$6D,$7E,$05,$B3,$FE,$66,$F6,$9E,$87
	.dc.b	$E5,$F2,$FC,$C7,$CA,$4C,$7E,$B7,$1F,$A9,$FB,$6D,$E3,$FB,$95,$C6
	.dc.b	$77,$F2,$9F,$A1,$F1,$E4,$72,$7F,$49,$E3,$62,$1B,$7E,$3E,$70,$FE
	.dc.b	$46,$7F,$55,$F2,$92,$BC,$6F,$EB,$7E,$3F,$11,$CB,$7F,$CA,$CE,$2F
	.dc.b	$F3,$55,$FB,$3F,$93,$97,$EC,$7F,$6D,$F2,$98,$8E,$47,$BA,$31,$5F
	.dc.b	$A3,$9F,$DE,$7C,$A4,$AF,$8A,$FE,$FF,$E3,$B1,$1C,$B7,$FD,$04,$E2
	.dc.b	$FF,$59,$5F,$C1,$7E,$3A,$5F,$CE,$7F,$84,$7C,$6E,$23,$97,$17,$F9
	.dc.b	$21,$B3,$A7,$F1,$FF,$22,$DB,$E5,$C3,$E6,$7F,$83,$BF,$EB,$4F,$11
	.dc.b	$F1,$5F,$EA,$69,$CC,$FF,$0C,$8C,$4F,$FA,$3E,$B9,$D6,$AF,$B6,$9F
	.dc.b	$EB,$E8,$E3,$5D,$30,$C3,$FE,$B2,$30,$F8,$DE,$93,$A8,$FF,$63,$D7
	.dc.b	$47,$47,$9F,$0F,$1B,$4E,$07,$FA,$7F,$0F,$2D,$F4,$8E,$27,$AC,$DB
	.dc.b	$E2,$C3,$D5,$3C,$27,$C2,$D7,$81,$CA,$78,$1D,$67,$8E,$1D,$2F,$CB
	.dc.b	$5C,$27,$2F,$8B,$0F,$64,$7C,$69,$DD,$F4,$AF,$0D,$FB,$2F,$73,$F2
	.dc.b	$6B,$67,$68,$CD,$70,$3C,$D1,$A7,$96,$1E,$B9,$DC,$79,$6B,$A3,$AB
	.dc.b	$8F,$23,$D8,$63,$0F,$59,$E9,$72,$7B,$6F,$B1,$F2,$6B,$D9,$1D,$CF
	.dc.b	$34,$F2,$8F,$0C,$77,$39,$57,$8B,$C6,$FE,$9F,$1F,$76,$0A,$33,$87
	.dc.b	$E1,$DB,$87,$E8,$1E,$33,$C3,$E4,$B8,$C6,$21,$18,$F1,$7F,$D2,$31
	.dc.b	$83,$B3,$88,$D4,$7F,$A5,$9F,$C8,$E1,$FE,$40,$E5,$FC,$9C,$F7,$3F
	.dc.b	$23,$19,$7F,$26,$67,$3F,$D7,$C6,$5F,$D8,$C6,$73,$FF,$60,$73,$FF
	.dc.b	$8F,$A7,$33,$E5,$78,$C6,$67,$CD,$3E,$08,$C3,$E7,$31,$7E,$67,$94
	.dc.b	$F6,$8C,$61,$EC,$AE,$35,$F4,$0D,$9F,$E0,$47,$2F,$9C,$C4,$F2,$7B
	.dc.b	$9F,$00,$E2,$7F,$3E,$8E,$2E,$30,$F8,$57,$EA,$C3,$DB,$7C,$DB,$5C
	.dc.b	$F0,$F3,$4E,$6B,$D5,$39,$3C,$35,$F2,$2B,$71,$98,$E6,$65,$DC,$F8
	.dc.b	$4E,$71,$93,$6E,$63,$DE,$73,$76,$65,$F7,$57,$95,$DC,$FF,$32,$E3
	.dc.b	$FF,$11,$7C,$8F,$F9,$EB,$1F,$F1,$73,$6F,$FC,$5B,$F2,$B7,$0F,$ED
	.dc.b	$F3,$6F,$8D,$7F,$74,$8C,$71,$2E,$C6,$3E,$46,$57,$F7,$AF,$91,$B8
	.dc.b	$FF,$4C,$9B,$97,$F8,$FE,$6D,$7F,$8F,$BE,$47,$A9,$E1,$F8,$EF,$D9
	.dc.b	$C7,$F8,$12,$32,$D3,$FC,$79,$1A,$C3,$FA,$CC,$F1,$F9,$19,$5A,$E6
	.dc.b	$B5,$FD,$9E,$D5,$8C,$7F,$D3,$FB,$27,$FA,$CE,$1D,$AB,$C9,$F2,$12
	.dc.b	$3C,$3F,$01,$71,$E9,$37,$1E,$78,$2B,$17,$FC,$3B,$1E,$DE,$7F,$BC
	.dc.b	$F8,$1C,$76,$7D,$B1,$F0,$EF,$19,$7A,$5F,$B0,$FF,$6E,$F8,$6B,$A9
	.dc.b	$F9,$5E,$3E,$19,$87,$C8,$F9,$0B,$7F,$0D,$FC,$16,$AF,$99,$FF,$74
	.dc.b	$57,$F9,$B6,$31,$F1,$1A,$97,$A1,$62,$56,$E7,$8E,$1F,$E1,$EB,$E9
	.dc.b	$7A,$8C,$4E,$A7,$F2,$CD,$AF,$F4,$2B,$D9,$E9,$7F,$03,$CF,$7C,$0D
	.dc.b	$3F,$11,$F2,$36,$BF,$CC,$B3,$A7,$F0,$FF,$23,$9C,$BF,$04,$FE,$3F
	.dc.b	$E4,$7F,$21,$F2,$36,$BF,$1A,$DA,$FC,$43,$74,$FB,$AF,$C8,$F9,$97
	.dc.b	$99,$4C,$79,$97,$0F,$95,$3A,$6C,$E9,$A4,$C3,$E4,$4E,$DB,$5C,$DF
	.dc.b	$CD,$7C,$97,$E9,$CF,$D4,$7C,$8D,$BF,$D2,$C5,$9F,$D0,$4D,$CB,$F2
	.dc.b	$BF,$25,$FA,$F3,$F6,$1E,$DD,$C3,$F5,$B1,$FB,$3F,$92,$C6,$1F,$A9
	.dc.b	$7F,$6D,$F2,$5F,$B7,$F6,$2D,$FE,$D1,$B5,$FA,$F6,$E3,$FA,$46,$CF
	.dc.b	$E8,$FB,$3F,$D0,$AF,$EF,$2F,$A3,$FB,$EE,$AF,$E2,$9F,$30,$8F,$E0
	.dc.b	$D1,$86,$9F,$EB,$F8,$D3,$C4,$C1,$FA,$8F,$99,$57,$FB,$5F,$2F,$EA
	.dc.b	$51,$89,$FE,$EF,$08,$C3,$FD,$E9,$AF,$F5,$93,$53,$FA,$AF,$9F,$7A
	.dc.b	$7D,$E3,$0F,$AB,$AE,$61,$E7,$ED,$B8,$7F,$9E,$F2,$FD,$13,$FE,$DC
	.dc.b	$6F,$86,$7F,$49,$1B,$BE,$3B,$FF,$1A,$9D,$9A,$71,$1F,$2B,$E5,$66
	.dc.b	$3F,$1F,$E5,$6E,$1E,$1F,$95,$F9,$E7,$D0,$F9,$7B,$5F,$37,$E5,$7F
	.dc.b	$A3,$7C,$B3,$C9,$FA,$B1,$87,$EB,$FC,$B6,$B9,$AD,$FF,$BC,$EF,$F9
	.dc.b	$73,$F5,$FE,$5B,$EE,$9F,$7B,$C6,$B1,$F6,$FE,$4F,$F0,$27,$E0,$BE
	.dc.b	$4E,$C7,$DF,$F9,$3F,$C2,$9F,$86,$F9,$2B,$1F,$83,$F9,$1F,$C4,$9F
	.dc.b	$8A,$F9,$2B,$1F,$87,$F9,$2F,$C6,$FC,$97,$E3,$9B,$1F,$E4,$4E,$8F
	.dc.b	$E8,$BF,$35,$FE,$45,$8C,$1F,$CE,$D7,$C0,$FF,$69,$75,$CF,$FE,$B5
	.dc.b	$73,$3C,$21,$D8,$97,$9F,$C4,$E5,$F1,$76,$F8,$56,$FE,$33,$B1,$DF
	.dc.b	$E3,$2C,$EB,$0F,$F2,$26,$7E,$8F,$84,$F2,$7C,$67,$95,$F3,$7C,$3C
	.dc.b	$BE,$1F,$19,$76,$F4,$7A,$3E,$B3,$D9,$E9,$DA,$ED,$F1,$9E,$F3,$E0
	.dc.b	$7C,$65,$8F,$17,$95,$F0,$CC,$61,$F0,$DF,$8D,$F1,$92,$76,$7A,$5F
	.dc.b	$95,$F0,$D2,$3D,$9F,$0F,$6F,$E6,$4D,$E0,$FA,$B3,$6B,$EA,$4D,$AF
	.dc.b	$9F,$C3,$FE,$19,$9F,$8D,$3F,$5F,$E1,$E4,$7C,$9F,$87,$B7,$F6,$66
	.dc.b	$E3,$F8,$29,$B5,$F8,$1F,$87,$FD,$1F,$C3,$FE,$05,$B1,$F4,$DB,$87
	.dc.b	$E0,$DB,$7F,$E1,$0F,$1B,$FC,$1F,$3F,$E8,$8E,$AF,$F0,$34,$3E,$1F
	.dc.b	$F9,$04,$E4,$FC,$4C,$70,$F8,$A5,$70,$AD,$C6,$2B,$7F,$0D,$8A,$D6
	.dc.b	$5F,$E9,$97,$BC,$66,$7A,$DD,$5F,$32,$F4,$BF,$C5,$31,$E3,$32,$E3
	.dc.b	$1F,$04,$7A,$7E,$0F,$95,$71,$DB,$C1,$3F,$EB,$08,$C3,$CD,$B5,$CB
	.dc.b	$05,$F2,$C3,$B6,$5F,$E8,$9B,$DB,$63,$38,$7F,$AF,$2F,$8E,$7E,$08
	.dc.b	$F3,$E5,$70,$E3,$1E,$3B,$E2,$FA,$6B,$51,$89,$DC,$F2,$3E,$15,$F6
	.dc.b	$7D,$53,$B3,$E3,$BF,$D1,$E7,$DE,$D8,$CB,$E9,$BF,$EA,$3F,$1A,$EB
	.dc.b	$7C,$D6,$1B,$5F,$EC,$C9,$D4,$E4,$FE,$B3,$83,$D5,$FF,$6E,$57,$FB
	.dc.b	$4A,$31,$F1,$DA,$9C,$CF,$95,$AF,$CD,$78,$EF,$F7,$2B,$8C,$FD,$79
	.dc.b	$CE,$DF,$4F,$C9,$FF,$5E,$5E,$E7,$35,$F7,$DA,$FE,$31,$B6,$AE,$37
	.dc.b	$E1,$C3,$EB,$56,$8D,$CE,$AF,$A4,$E8,$8F,$E2,$E3,$4F,$78,$CE,$1E
	.dc.b	$57,$F0,$53,$CF,$C5,$71,$BE,$2B,$0B,$E2,$A9,$88,$98,$FE,$72,$B8
	.dc.b	$CE,$A3,$15,$CD,$BF,$13,$FE,$E1,$E7,$FE,$05,$CD,$79,$BF,$80,$B6
	.dc.b	$BF,$7F,$57,$1F,$2B,$70,$34,$C7,$C4,$E7,$E2,$75,$F1,$13,$0F,$EE
	.dc.b	$51,$97,$FB,$A4,$65,$67,$FC,$56,$7B,$7C,$47,$7F,$87,$95,$FD,$EE
	.dc.b	$32,$FF,$7D,$8C,$BC,$B5,$FE,$37,$3A,$8C,$1F,$C6,$A7,$77,$FC,$8E
	.dc.b	$7D,$9F,$11,$ED,$F8,$89,$7F,$E0,$98,$F8,$1F,$09,$A8,$C1,$FE,$0D
	.dc.b	$7E,$27,$C4,$6A,$30,$7F,$84,$5F,$91,$F1,$12,$3F,$C2,$8F,$CD,$F4
	.dc.b	$3F,$97,$C7,$9B,$89,$0F,$84,$D6,$D2,$BF,$BA,$CE,$5B,$F0,$DC,$C8
	.dc.b	$EB,$C2,$ED,$D9,$23,$8B,$E1,$F8,$5F,$17,$C2,$F8,$FE,$16,$57,$68
	.dc.b	$F3,$7C,$8C,$8F,$0B,$E9,$F8,$5F,$57,$CC,$49,$F2,$BE,$DF,$85,$F7
	.dc.b	$7C,$2F,$BF,$E1,$65,$7A,$A3,$E1,$7C,$2C,$8F,$6B,$F1,$7E,$17,$E3
	.dc.b	$7C,$2F,$C7,$F8,$59,$5F,$0A,$3E,$57,$8F,$F2,$FE,$12,$4F,$C6,$8F
	.dc.b	$9D,$F0,$B2,$3E,$4B,$F9,$DF,$D0,$3F,$89,$F8,$3F,$EF,$47,$FC,$5B
	.dc.b	$3F,$E2,$26,$BF,$55,$B1,$BF,$85,$E5,$E7,$73,$F8,$59,$7C,$63,$B7
	.dc.b	$C2,$F7,$F8,$59,$3D,$63,$F6,$0B,$E6,$BF,$D7,$AF,$F8,$E4,$C7,$C2
	.dc.b	$EA,$0D,$F8,$59,$3B,$9C,$B6,$72,$FE,$88,$ED,$3C,$E2,$47,$18,$F8
	.dc.b	$9E,$04,$8F,$84,$FC,$7F,$86,$C4,$7C,$38,$BA,$FC,$99,$F5,$FC,$D3
	.dc.b	$AB,$FD,$3B,$68,$C3,$FE,$A0,$6B,$FE,$9D,$35,$1C,$23,$35,$FA,$66
	.dc.b	$F8,$6A,$91,$87,$19,$FD,$69,$D7,$8A,$E2,$FE,$BD,$5F,$77,$4F,$C1
	.dc.b	$1F,$83,$F6,$A4,$64,$87,$E0,$5C,$BF,$83,$8E,$2F,$E1,$E3,$E0,$F4
	.dc.b	$48,$FB,$59,$23,$EF,$FD,$6F,$9E,$7E,$24,$6D,$FA,$D3,$B7,$F0,$53
	.dc.b	$ED,$C7,$D5,$C1,$F9,$9F,$88,$95,$F9,$78,$FC,$E1,$87,$F3,$11,$87
	.dc.b	$F1,$9F,$34,$F4,$7F,$22,$BD,$2A,$62,$24,$F0,$8C,$BC,$A3,$2F,$E3
	.dc.b	$67,$93,$B3,$47,$E6,$FC,$97,$C3,$7E,$11,$B1,$FE,$E2,$6B,$F8,$36
	.dc.b	$BF,$EE,$17,$3C,$7F,$CA,$1B,$C5,$F8,$39,$FF,$24,$57,$EB,$3C,$6C
	.dc.b	$C7,$E1,$FE,$67,$FA,$CB,$FA,$D6,$BF,$A5,$6E,$1F,$CD,$3E,$69,$E2
	.dc.b	$7E,$DA,$73,$8F,$76,$DF,$5F,$5F,$F7,$77,$C2,$3F,$75,$F3,$4F,$87
	.dc.b	$FD,$FA,$F8,$7A,$62,$24,$F0,$8C,$BC,$A3,$2F,$EE,$E7,$93,$B3,$47
	.dc.b	$F0,$9F,$96,$7B,$5F,$C3,$9B,$19,$8F,$E2,$3F,$7F,$FA,$78,$FD,$64
	.dc.b	$E9,$FE,$10,$DD,$BF,$AB,$B7,$5F,$A1,$5C,$69,$F8,$BB,$D1,$FE,$3A
	.dc.b	$D8,$FE,$38,$D7,$F8,$A7,$C6,$7D,$67,$8C,$F0,$23,$FC,$8E,$3E,$DF
	.dc.b	$57,$EE,$FE,$3F,$F9,$D7,$C7,$CA,$FE,$71,$1F,$CF,$4C,$3F,$CE,$63
	.dc.b	$0F,$F3,$DF,$8F,$FE,$85,$F1,$F2,$30,$43,$FA,$03,$97,$FA,$14,$7F
	.dc.b	$49,$E8,$FE,$95,$C5,$27,$FA,$44,$65,$FE,$95,$19,$7F,$A5,$47,$F5
	.dc.b	$1E,$3F,$EA,$5F,$1F,$27,$FA,$84,$65,$FE,$A5,$19,$7F,$A9,$47,$F8
	.dc.b	$D7,$F4,$1F,$D0,$47,$F8,$25,$BC,$7F,$EB,$AD,$51,$8E,$2F,$E4,$6F
	.dc.b	$FA,$1A,$74,$E6,$95,$C1,$EC,$77,$F8,$EF,$EA,$FF,$1C,$F2,$FF,$C7
	.dc.b	$32,$F9,$B7,$CA,$7E,$CC,$7C,$97,$ED,$63,$33,$FC,$E6,$72,$FF,$3B
	.dc.b	$8C,$AB,$ED,$9E,$6B,$F7,$D1,$89,$F0,$47,$7B,$EC,$D8,$E0,$7F,$47
	.dc.b	$9E,$17,$D9,$B3,$B9,$E6,$79,$6B,$9B,$C7,$0F,$7D,$F1,$6D,$71,$8F
	.dc.b	$4B,$EC,$9E,$67,$AE,$76,$F3,$C3,$CA,$FC,$8B,$EE,$D3,$38,$79,$9E
	.dc.b	$07,$4B,$ED,$87,$82,$B8,$D6,$EB,$C1,$7D,$AF,$67,$9C,$F9,$78,$79
	.dc.b	$27,$67,$5B,$F1,$CF,$08,$F0,$E1,$F0,$6F,$A5,$68,$F0,$DF,$3C,$3C
	.dc.b	$58,$71,$C3,$D8,$7C,$48,$F3,$1A,$AE,$51,$F0,$2B,$91,$E5,$AE,$47
	.dc.b	$3A,$E3,$19,$AE,$35,$D6,$38,$DF,$9A,$21,$97,$B1,$FE,$0B,$F8,$EF
	.dc.b	$F1,$FD,$FE,$35,$B5,$F9,$E6,$CE,$FE,$4F,$FC,$31,$F2,$17,$0F,$D9
	.dc.b	$B5,$FF,$6A,$8A,$FF,$25,$BF,$E1,$98,$D3,$4F,$C3,$D3,$D4,$70,$7E
	.dc.b	$1A,$75,$AF,$E9,$AB,$53,$99,$5F,$E3,$27,$31,$F9,$23,$8E,$BF,$EC
	.dc.b	$3C,$3A,$C6,$4E,$27,$89,$CE,$1F,$B4,$8E,$27,$F8,$C6,$BA,$C6,$63
	.dc.b	$8B,$E5,$9E,$35,$F1,$A7,$A6,$1E,$7C,$3C,$D5,$A9,$F9,$26,$8F,$24
	.dc.b	$6A,$78,$46,$6F,$C1,$39,$8D,$99,$BF,$04,$F8,$EB,$8B,$67,$6D,$39
	.dc.b	$BC,$4F,$1B,$99,$E6,$7C,$99,$E6,$D9,$E6,$D9,$E6,$71,$8E,$71,$93
	.dc.b	$9B,$F2,$EF,$E1,$3F,$88,$9E,$0D,$9E,$0F,$23,$85,$7F,$1E,$DF,$F9
	.dc.b	$FE,$DF,$73,$2F,$E7,$AD,$9F,$E8,$3F,$27,$6F,$F9,$FF,$0F,$F3,$69
	.dc.b	$FB,$87,$F4,$6A,$D6,$1D,$23,$35,$C2,$F3,$7C,$63,$FA,$97,$76,$3E
	.dc.b	$6B,$2D,$BE,$38,$7F,$54,$8C,$57,$09,$CD,$AF,$AE,$7D,$BE,$56,$D6
	.dc.b	$9B,$7F,$D8,$5B,$A7,$F5,$C7,$CB,$F2,$7E,$6F,$93,$B7,$D9,$B7,$FD
	.dc.b	$A5,$B7,$FD,$91,$F6,$F8,$DE,$EF,$93,$B7,$E6,$6D,$FF,$71,$6D,$FF
	.dc.b	$6C,$7E,$74,$3F,$1E,$FF,$7E,$57,$FA,$F5,$B1,$FD,$E7,$E4,$F7,$18
	.dc.b	$BF,$DA,$CF,$1F,$8D,$C9,$F4,$2A,$DF,$FB,$1F,$F2,$57,$C4,$78,$1B
	.dc.b	$1F,$33,$D2,$C4,$C8,$E0,$FD,$A8,$C6,$0C,$7F,$15,$F3,$8F,$4B,$F1
	.dc.b	$10,$87,$C7,$61,$D6,$3F,$86,$9D,$4E,$6B,$4D,$9D,$57,$26,$CE,$9A
	.dc.b	$6A,$7E,$5F,$C2,$3E,$1E,$9E,$4F,$9C,$7C,$1E,$2F,$FD,$D5,$5B,$6D
	.dc.b	$6E,$75,$9F,$FB,$97,$EA,$5E,$DF,$D8,$6D,$F9,$BE,$37,$13,$23,$FD
	.dc.b	$B1,$4A,$DB,$FE,$DB,$6B,$FE,$36,$6B,$FC,$A7,$5F,$AA,$FC,$F9,$C7
	.dc.b	$B3,$F2,$F8,$7E,$63,$63,$E3,$B6,$78,$1F,$5E,$BF,$D2,$F3,$F3,$2B
	.dc.b	$8B,$4E,$27,$EB,$BD,$CF,$DC,$FC,$BF,$EE,$9B,$19,$3F,$67,$19,$69
	.dc.b	$FC,$0A,$DF,$3A,$E6,$FC,$FB,$67,$F7,$7E,$CE,$26,$47,$19,$D5,$7F
	.dc.b	$96,$6B,$2F,$F9,$6F,$E5,$FF,$86,$FC,$C7,$F0,$E6,$DF,$F0,$66,$CE
	.dc.b	$E3,$F8,$4C,$ED,$A6,$CB,$1F,$20,$F4,$FC,$C7,$A0,$FC,$D7,$C2,$D9
	.dc.b	$F3,$47,$EC,$A7,$FD,$9D,$8A,$31,$03,$9C,$FE,$12,$74,$FF,$9D,$A3
	.dc.b	$F9,$7C,$FF,$28,$3F,$94,$FC,$C5,$9F,$E4,$31,$67,$FC,$F9,$37,$AB
	.dc.b	$F9,$E7,$43,$D7,$FD,$07,$B1,$FA,$16,$E3,$FB,$19,$D3,$67,$4D,$35
	.dc.b	$3C,$30,$FC,$5D,$AF,$8C,$61,$5F,$F4,$0F,$98,$FE,$8B,$FA,$17,$F8
	.dc.b	$5B,$1F,$E7,$31,$87,$3E,$0E,$BE,$66,$C7,$F4,$2A,$47,$2F,$02,$4E
	.dc.b	$7C,$0B,$9F,$E7,$BC,$D6,$0F,$99,$F0,$9E,$2F,$80,$B8,$75,$F9,$9F
	.dc.b	$2F,$C3,$F9,$9B,$1E,$1F,$99,$F4,$FC,$8F,$A9,$B1,$E5,$F1,$3D,$A7
	.dc.b	$BB,$E6,$6C,$7A,$7C,$1F,$83,$E1,$FC,$26,$C7,$B7,$E6,$7E,$29,$F1
	.dc.b	$BE,$66,$C7,$C1,$EE,$F9,$27,$CA,$F9,$9B,$1F,$17,$CA,$F9,$BF,$33
	.dc.b	$F3,$9B,$1F,$DA,$B7,$FF,$26,$37,$DE,$FE,$EF,$0F,$7E,$C6,$3C,$E9
	.dc.b	$5C,$7E,$62,$F4,$FC,$DC,$1F,$33,$CB,$E6,$79,$B6,$FC,$1F,$03,$D8
	.dc.b	$C5,$F1,$7C,$3E,$97,$8B,$E0,$ED,$79,$DC,$3E,$9F,$96,$B7,$E1,$7D
	.dc.b	$27,$AB,$E5,$EC,$FB,$DF,$6F,$CA,$FB,$BD,$6B,$3F,$0D,$B7,$EB,$6F
	.dc.b	$95,$FE,$E3,$DB,$F9,$AB,$6F,$95,$FC,$B6,$DF,$C3,$F8,$EF,$97,$F1
	.dc.b	$DF,$31,$B5,$F4,$5B,$3B,$8F,$A9,$3F,$4B,$CD,$91,$F5,$FC,$CB,$5F
	.dc.b	$3E,$7E,$BF,$C6,$E3,$2F,$33,$F6,$BE,$3B,$FC,$EB,$F1,$4E,$6F,$E1
	.dc.b	$DE,$53,$77,$C6,$EA,$73,$0B,$C6,$7F,$EF,$C8,$DE,$BF,$E9,$1A,$DD
	.dc.b	$6B,$0E,$98,$5B,$E5,$7B,$C3,$94,$F4,$5F,$37,$FE,$05,$F8,$ED,$46
	.dc.b	$3C,$F4,$63,$E2,$DF,$07,$C6,$F8,$5F,$FA,$63,$5F,$E5,$4F,$1B,$E7
	.dc.b	$3A,$DB,$E1,$4F,$F2,$FC,$39,$E7,$AC,$BB,$CF,$33,$F9,$A5,$F1,$C7
	.dc.b	$AB,$63,$96,$5E,$3A,$E3,$8F,$69,$F1,$1C,$CA,$E7,$1F,$0C,$66,$FA
	.dc.b	$3A,$C3,$BD,$F4,$7D,$B5,$C2,$FB,$5F,$03,$E1,$DF,$C2,$3C,$25,$72
	.dc.b	$F4,$CB,$FC,$23,$F1,$CE,$97,$CE,$61,$B5,$FD,$46,$75,$39,$3F,$C3
	.dc.b	$37,$FE,$3A,$F0,$6F,$A7,$FE,$D7,$BF,$EB,$13,$96,$C6,$5F,$23,$BC
	.dc.b	$BD,$6D,$D3,$E7,$B6,$7E,$9F,$8D,$F7,$AB,$78,$7D,$EA,$D5,$7E,$31
	.dc.b	$B8,$2F,$86,$1D,$2F,$81,$2F,$E9,$4E,$9F,$C0,$DF,$22,$4F,$E3,$27
	.dc.b	$53,$DE,$39,$61,$E4,$8E,$67,$E8,$78,$DF,$38,$F8,$9E,$D1,$8F,$89
	.dc.b	$B2,$6C,$CA,$E5,$39,$6C,$E5,$FC,$C5,$76,$F8,$9F,$D1,$F5,$7F,$BD
	.dc.b	$A3,$C3,$CB,$8E,$6F,$E0,$78,$FE,$B3,$0F,$F5,$6A,$C3,$50,$BE,$8E
	.dc.b	$AF,$31,$D2,$38,$CE,$29,$F0,$F7,$E1,$65,$76,$8D,$18,$7F,$BB,$4A
	.dc.b	$73,$1E,$4F,$85,$E9,$F0,$92,$30,$43,$8B,$97,$9C,$78,$CC,$69,$F0
	.dc.b	$ED,$59,$AF,$83,$F0,$9E,$9F,$84,$91,$83,$E2,$FC,$1F,$B6,$64,$60
	.dc.b	$87,$AA,$3E,$11,$F0,$FE,$0F,$E2,$7C,$17,$C8,$F0,$5F,$23,$4C,$DF
	.dc.b	$FB,$7B,$CD,$7C,$1F,$CE,$7F,$D7,$7B,$7D,$B6,$F2,$FF,$3C,$77,$5C
	.dc.b	$7E,$0F,$81,$1E,$9E,$27,$3E,$D9,$1D,$DE,$DF,$05,$DF,$E0,$BC,$1A
	.dc.b	$CA,$F2,$47,$8F,$E0,$A4,$79,$DF,$37,$C0,$F9,$FE,$0A,$4F,$AD,$F5
	.dc.b	$7C,$17,$AF,$E0,$BD,$9F,$03,$2B,$E0,$47,$BF,$E0,$A4,$7C,$37,$E1
	.dc.b	$7C,$17,$C3,$F8,$2F,$89,$F0,$52,$BE,$44,$7C,$7F,$27,$E4,$7C,$0C
	.dc.b	$9F,$99,$1F,$2F,$E0,$BF,$33,$FA,$1F,$F2,$7E,$9F,$D3,$37,$A7,$E6
	.dc.b	$4F,$E7,$BE,$73,$FC,$49,$87,$F9,$DE,$F7,$E9,$F2,$35,$F0,$9B,$F4
	.dc.b	$38,$7C,$14,$BE,$19,$7F,$72,$C7,$FA,$36,$3F,$B5,$5F,$07,$DF,$E0
	.dc.b	$E5,$F0,$3F,$65,$3A,$7F,$D5,$5F,$3A,$F8,$3F,$DB,$C2,$1F,$05,$86
	.dc.b	$C7,$09,$C1,$99,$91,$92,$1F,$E9,$F9,$E2,$D8,$E2,$F9,$7D,$15,$FF
	.dc.b	$BA,$A3,$0D,$7F,$CE,$4D,$CB,$E0,$7C,$27,$CD,$F1,$65,$70,$7E,$84
	.dc.b	$7B,$3E,$75,$DD,$EB,$E1,$97,$FE,$CB,$9C,$B6,$7F,$DD,$AA,$32,$FF
	.dc.b	$AA,$DA,$FF,$67,$6B,$F5,$7D,$CF,$C0,$6D,$23,$D7,$1F,$83,$30,$FB
	.dc.b	$23,$0E,$8F,$82,$EB,$2F,$BD,$3E,$BE,$AF,$AF,$F3,$AE,$EA,$67,$0F
	.dc.b	$64,$D7,$FD,$A1,$8A,$38,$E5,$F9,$5E,$0F,$CB,$7C,$34,$BF,$A3,$19
	.dc.b	$7E,$9C,$65,$FA,$71,$F8,$8F,$9D,$FE,$6F,$1F,$EF,$57,$F9,$93,$7D
	.dc.b	$5D,$7A,$D2,$30,$43,$FD,$C9,$3C,$9B,$1D,$66,$47,$58,$FC,$BF,$96
	.dc.b	$F8,$3F,$CC,$F9,$1F,$E5,$4A,$FD,$00,$70,$FF,$BC,$67,$83,$A6,$BF
	.dc.b	$EE,$76,$BF,$E2,$36,$BF,$E4,$66,$BF,$EF,$1F,$9D,$FE,$C0,$F8,$3E
	.dc.b	$75,$E2,$FE,$B1,$B1,$E4,$9D,$AE,$47,$77,$F9,$4A,$73,$18,$9E,$DE
	.dc.b	$D7,$ED,$A7,$35,$FE,$58,$B7,$91,$FB,$AF,$9D,$FF,$39,$61,$F0,$9F
	.dc.b	$83,$8F,$38,$46,$2F,$FC,$C2,$F0,$8F,$24,$63,$E1,$BA,$7C,$34,$9E
	.dc.b	$91,$DC,$C3,$C0,$1D,$A3,$4F,$D4,$F9,$8F,$ED,$2F,$D6,$F9,$DF,$E3
	.dc.b	$77,$FE,$72,$6C,$72,$9D,$BD,$43,$8C,$7D,$8D,$AF,$F3,$DB,$67,$C7
	.dc.b	$5F,$CC,$9E,$5F,$3A,$F4,$BF,$90,$FC,$EF,$C3,$9F,$4B,$54,$ED,$FC
	.dc.b	$6B,$5E,$73,$FC,$AA,$31,$8F,$CF,$AF,$C6,$BF,$CD,$FE,$3B,$FA,$27
	.dc.b	$C7,$4C,$3F,$8F,$C7,$F4,$73,$0F,$F2,$08,$C3,$FC,$93,$E3,$7F,$A6
	.dc.b	$7C,$74,$8C,$10,$FE,$4E,$E5,$FE,$53,$1F,$D4,$FA,$3F,$AA,$70,$49
	.dc.b	$FE,$5F,$19,$7F,$98,$C6,$5F,$E6,$31,$FD,$6F,$83,$FA,$E7,$BB,$27
	.dc.b	$F9,$BC,$65,$FE,$73,$19,$7F,$9C,$C7,$F9,$03,$E6,$FF,$B3,$47,$F9
	.dc.b	$15,$BE,$59,$A7,$FA,$B2,$3F,$19,$19,$79,$10,$FE,$95,$1B,$D7,$F3
	.dc.b	$DF,$1C,$F2,$BF,$5D,$4F,$9D,$7C,$DF,$FB,$00,$7C,$EF,$FB,$0A,$33
	.dc.b	$3F,$9A,$9C,$BF,$9C,$8C,$AB,$ED,$86,$6B,$FD,$97,$18,$9F,$04,$77
	.dc.b	$BE,$CD,$8E,$07,$EA,$27,$85,$F6,$6C,$EE,$79,$9E,$5A,$E6,$F1,$C3
	.dc.b	$DF,$7C,$5B,$5C,$63,$D2,$FB,$27,$99,$EB,$9D,$BC,$F0,$F2,$BF,$22
	.dc.b	$FB,$B4,$CE,$1E,$67,$81,$D2,$FB,$61,$E0,$AE,$35,$BA,$F0,$5F,$6B
	.dc.b	$D9,$E7,$3E,$5E,$1E,$49,$D9,$D6,$FC,$73,$C2,$3C,$38,$7C,$1B,$E9
	.dc.b	$5A,$3C,$37,$CF,$0F,$16,$1C,$70,$F6,$1F,$12,$3C,$C6,$AB,$94,$7C
	.dc.b	$0A,$E4,$79,$6B,$91,$CE,$B8,$C6,$6B,$8D,$75,$8E,$37,$E6,$88,$65
	.dc.b	$EC,$7F,$2C,$F8,$FF,$E3,$57,$F8,$D6,$D7,$E7,$9B,$3B,$F5,$7F,$DF
	.dc.b	$96,$F8,$FC,$3F,$EF,$F6,$BF,$EE,$51,$5F,$C9,$DF,$E7,$71,$A6,$9F
	.dc.b	$88,$A7,$A9,$FD,$5F,$4F,$D2,$CE,$9F,$D3,$CE,$BB,$3F,$17,$39,$D7
	.dc.b	$FD,$9F,$7C,$5F,$3E,$1F,$90,$BE,$15,$F9,$63,$81,$FE,$6B,$AE,$0F
	.dc.b	$AC,$F2,$38,$7D,$98,$78,$A3,$11,$C8,$FE,$E4,$E1,$F8,$C7,$2C,$3C
	.dc.b	$EE,$27,$94,$62,$FC,$D5,$C5,$B1,$9A,$E4,$7A,$4E,$58,$7C,$08,$D5
	.dc.b	$F4,$9D,$57,$26,$9C,$A3,$E6,$BC,$8F,$85,$3C,$AB,$A4,$70,$9E,$87
	.dc.b	$0A,$E8,$D8,$E9,$3F,$41,$A7,$CB,$8E,$31,$96,$C6,$63,$BC,$73,$33
	.dc.b	$1F,$42,$32,$7F,$81,$37,$FE,$7D,$B7,$F8,$77,$2F,$F0,$F4,$63,$1F
	.dc.b	$E7,$7F,$25,$97,$87,$07,$F3,$67,$F9,$FD,$E7,$1F,$E9,$4D,$BF,$E9
	.dc.b	$57,$6F,$85,$7F,$4A,$7F,$C1,$37,$FD,$23,$E7,$75,$DD,$B6,$C6,$4F
	.dc.b	$E9,$B1,$97,$FA,$7E,$7D,$7E,$77,$B3,$63,$27,$F5,$48,$CB,$4F,$EA
	.dc.b	$FF,$25,$E5,$F9,$2F,$33,$63,$27,$F5,$C8,$CB,$4F,$EB,$FE,$37,$B7
	.dc.b	$E4,$BD,$CD,$8C,$9F,$D9,$23,$2D,$3F,$CC,$BF,$3A,$F8,$BF,$95,$B2
	.dc.b	$31,$3F,$DB,$3E,$6B,$31,$89,$FF,$19,$37,$8F,$FD,$C3,$C7,$AC,$3E
	.dc.b	$97,$C1,$7D,$09,$FF,$28,$7C,$EB,$E1,$FB,$B7,$3F,$EF,$3E,$8E,$26
	.dc.b	$47,$FB,$16,$DC,$D8,$D3,$FC,$07,$18,$9F,$F6,$F6,$2F,$9C,$7A,$3F
	.dc.b	$85,$8C,$7C,$6D,$9C,$D7,$05,$19,$AD,$36,$74,$DC,$3F,$CB,$D7,$D2
	.dc.b	$79,$57,$07,$FD,$23,$F3,$2F,$89,$F9,$AD,$8F,$3F,$CD,$BE,$0F,$35
	.dc.b	$27,$D7,$F5,$6E,$7C,$67,$E0,$8C,$4F,$FA,$7B,$E7,$9E,$EF,$F9,$F6
	.dc.b	$BF,$CE,$22,$33,$3F,$D0,$A3,$13,$B7,$FC,$EF,$3A,$7F,$A3,$B7,$5F
	.dc.b	$DB,$6F,$FE,$CF,$BF,$05,$FC,$BF,$81,$F9,$9C,$7F,$37,$D7,$B5,$F3
	.dc.b	$5B,$97,$C8,$C3,$FD,$2F,$3F,$46,$B8,$34,$E0,$59,$E1,$5F,$BB,$F7
	.dc.b	$3F,$79,$F3,$56,$BF,$C8,$2D,$AF,$DF,$37,$3F,$E1,$BE,$22,$3F,$3F
	.dc.b	$F3,$CE,$EA,$7F,$07,$F9,$AB,$8F,$FB,$2A,$74,$FE,$4F,$C5,$CE,$5F
	.dc.b	$92,$7F,$87,$7C,$CF,$F0,$FF,$9A,$B5,$FE,$54,$6D,$7F,$11,$6E,$9E
	.dc.b	$46,$CF,$F1,$E6,$CF,$F9,$95,$F3,$FC,$D7,$D7,$86,$28,$C3,$6B,$EF
	.dc.b	$36,$7F,$C0,$2D,$BF,$E6,$1F,$40,$F6,$7A,$DF,$EA,$A3,$0F,$F9,$B3
	.dc.b	$0F,$E5,$47,$F2,$BF,$9B,$B1,$92,$63,$FC,$B6,$32,$FE,$9B,$B1,$EC
	.dc.b	$FE,$CD,$7B,$56,$30,$DB,$FD,$2B,$6B,$93,$73,$FE,$91,$7C,$AB,$F5
	.dc.b	$77,$C9,$FF,$41,$FC,$DF,$F4,$3F,$9F,$FF,$76,$C6,$0F,$F6,$AD,$FF
	.dc.b	$50,$8C,$E3,$FE,$8E,$7F,$8E,$4F,$F5,$75,$F3,$9C,$FC,$9E,$8F,$5F
	.dc.b	$13,$F8,$FF,$A3,$6F,$F8,$C3,$E1,$F9,$CF,$E4,$7E,$95,$9F,$E3,$8F
	.dc.b	$97,$E7,$3C,$DF,$13,$67,$CA,$78,$5F,$57,$CE,$7A,$FE,$36,$CE,$B6
	.dc.b	$7A,$ED,$D5,$F0,$7E,$72,$5E,$EB,$5E,$36,$30,$F8,$1E,$2F,$C7,$F1
	.dc.b	$7E,$43,$6F,$86,$4F,$9C,$F9,$87,$CD,$F8,$DB,$3A,$F9,$0F,$A0,$7D
	.dc.b	$1F,$9C,$B1,$CB,$CF,$FA,$9F,$39,$F5,$5B,$1E,$5E,$3F,$F2,$8B,$7E
	.dc.b	$07,$FB,$E4,$3E,$0A,$C6,$3D,$49,$5C,$7E,$5E,$F0,$FF,$B2,$30,$7C
	.dc.b	$DF,$2F,$9B,$E6,$DB,$F0,$7C,$2F,$F5,$8F,$83,$EE,$D8,$F2,$7C,$37
	.dc.b	$88,$F1,$FB,$56,$32,$7A,$7E,$63,$19,$F8,$5F,$49,$EA,$F9,$9B,$3E
	.dc.b	$F7,$DB,$F2,$FE,$EF,$6A,$CF,$C3,$6D,$FA,$DB,$E4,$7F,$8A,$37,$FE
	.dc.b	$6D,$5F,$E2,$D5,$7C,$E3,$E6,$36,$FE,$27,$C7,$FC,$CF,$47,$E6,$B6
	.dc.b	$BE,$93,$67,$71,$F5,$67,$E9,$F9,$9F,$53,$E3,$AC,$FD,$96,$D7,$D1
	.dc.b	$6E,$5E,$77,$ED,$FC,$7A,$BF,$D1,$BC,$BE,$9F,$8E,$C6,$1F,$E9,$78
	.dc.b	$F6,$56,$3E,$35,$1B,$BE,$35,$FA,$2D,$97,$BA,$D4,$F3,$35,$19,$78
	.dc.b	$61,$57,$CF,$7F,$82,$BE,$77,$AC,$63,$CF,$46,$3E,$1D,$F1,$9C,$6F
	.dc.b	$85,$F7,$38,$57,$28,$CE,$BE,$A9,$1F,$E1,$EA,$E5,$87,$68,$E5,$5C
	.dc.b	$33,$FE,$FA,$D8,$E3,$8F,$8B,$5E,$D7,$C7,$1E,$D5,$E2,$39,$C6,$70
	.dc.b	$FF,$03,$DF,$19,$CE,$5C,$88,$FA,$E3,$57,$D3,$1F,$14,$F5,$3F,$C2
	.dc.b	$25,$3C,$F8,$7A,$DC,$C7,$C6,$77,$F2,$0E,$77,$CF,$FF,$85,$E3,$0D
	.dc.b	$8C,$CF,$FB,$72,$72,$FF,$B7,$AB,$FC,$77,$E0,$7F,$87,$DB,$19,$F8
	.dc.b	$9F,$F6,$EB,$67,$51,$8C,$BD,$8D,$D7,$E7,$B6,$BE,$BD,$BC,$5F,$F1
	.dc.b	$7D,$F0,$C3,$EF,$5E,$EF,$F1,$6D,$BE,$11,$A9,$8F,$DE,$9D,$5F,$4C
	.dc.b	$3C,$35,$AC,$3B,$5F,$E1,$2B,$4F,$92,$FB,$12,$BF,$21,$3A,$F3,$5F
	.dc.b	$14,$F9,$0C,$37,$0F,$D5,$CE,$A7,$3F,$21,$A6,$4E,$AB,$83,$FA,$DA
	.dc.b	$EB,$5D,$3E,$33,$FC,$A7,$D5,$FE,$BE,$BF,$F2,$E5,$FF,$9F,$67,$3C
	.dc.b	$BF,$EC,$59,$FD,$5C,$E4,$87,$FA,$02,$76,$FF,$60,$8D,$AA,$FE,$12
	.dc.b	$BF,$7B,$FC,$63,$FD,$9C,$C7,$17,$26,$9F,$E8,$FF,$87,$C4,$F0,$8E
	.dc.b	$98,$FF,$9B,$23,$17,$D9,$B1,$E9,$AE,$9F,$BD,$7E,$7E,$BE,$07,$C7
	.dc.b	$FA,$35,$FA,$2E,$BA,$66,$5E,$5B,$E1,$8F,$F9,$E9,$B1,$EF,$F8,$CD
	.dc.b	$BE,$43,$F8,$4C,$69,$3C,$30,$FB,$F3,$E1,$D7,$FC,$C5,$5F,$EA,$DF
	.dc.b	$8B,$C1,$9F,$8B,$91,$FE,$B1,$6B,$FE,$B0,$F8,$AE,$46,$35,$FB,$93
	.dc.b	$89,$FF,$5C,$7C,$57,$7F,$88,$91,$83,$FD,$79,$F1,$1E,$32,$3F,$EB
	.dc.b	$FF,$86,$F8,$DF,$A3,$7D,$AC,$7F,$11,$1E,$93,$3C,$1A,$F8,$79,$7F
	.dc.b	$E1,$DB,$FF,$6E,$37,$E1,$F9,$F8,$72,$34,$F6,$EE,$EF,$F0,$D2,$79
	.dc.b	$3E,$2F,$98,$F1,$F9,$52,$7B,$3E,$6F,$86,$F3,$FC,$34,$9F,$13,$EA
	.dc.b	$F8,$6F,$5F,$C3,$49,$F3,$3E,$EF,$86,$F7,$FC,$2C,$9F,$53,$F0,$BE
	.dc.b	$1B,$E1,$FC,$2C,$9F,$73,$F1,$BE,$17,$E3,$FC,$2C,$9F,$84,$FC,$AF
	.dc.b	$2B,$E5,$FC,$24,$9F,$8C,$FC,$EF,$85,$FC,$A7,$DA,$FF,$24,$9F,$C6
	.dc.b	$78,$88,$C7,$67,$F7,$48,$C7,$9D,$C3,$E1,$B8,$FC,$34,$C3,$FC,$9D
	.dc.b	$82,$3A,$FA,$5F,$AF,$F9,$FF,$F7,$AD,$7E,$CF,$E8,$BF,$D0,$AD,$F7
	.dc.b	$75,$DD,$23,$F7,$15,$C1,$FD,$00,$71,$AF,$F0,$0B,$7B,$FC,$6B,$75
	.dc.b	$AF,$0F,$FD,$CA,$D8,$FE,$E3,$5C,$1F,$4F,$C3,$FC,$4F,$88,$91,$82
	.dc.b	$1F,$1B,$64,$6A,$31,$3F,$33,$E1,$FE,$6F,$C3,$49,$F8,$B1,$97,$E3
	.dc.b	$C6,$5E,$91,$F1,$FB,$BF,$DF,$11,$EF,$6E,$5F,$E7,$E8,$FF,$7E,$35
	.dc.b	$FE,$E2,$6C,$F9,$31,$CD,$FA,$1C,$55,$F8,$B1,$76,$F9,$53,$F8,$4F
	.dc.b	$87,$95,$F7,$1F,$C3,$FA,$F8,$8F,$85,$16,$BF,$0D,$AF,$F2,$49,$FF
	.dc.b	$4A,$AF,$87,$FC,$7F,$C3,$49,$FC,$3C,$7E,$48,$C3,$F3,$63,$0F,$F3
	.dc.b	$9E,$3F,$F0,$DD,$FF,$3E,$FA,$1F,$E0,$DC,$7F,$E2,$AA,$46,$1F,$F1
	.dc.b	$6D,$70,$9D,$36,$33,$CD,$FD,$0E,$31,$0F,$A1,$79,$F9,$A9,$3E,$37
	.dc.b	$FA,$54,$EA,$3F,$43,$F1,$3F,$AC,$E8,$91,$82,$1F,$89,$72,$FE,$2E
	.dc.b	$3F,$AA,$7D,$0B,$E1,$69,$27,$FA,$D7,$95,$F9,$07,$FA,$E4,$ED,$B3
	.dc.b	$B7,$F4,$F3,$B7,$F5,$3F,$6F,$FE,$DC,$7F,$B2,$7C,$93,$DB,$FE,$C8
	.dc.b	$D8,$FE,$D1,$3B,$78,$FA,$5F,$DA,$63,$1C,$5F,$DC,$1B,$C3,$F4,$ED
	.dc.b	$EF,$EE,$E6,$1D,$1E,$F1,$FD,$CF,$E5,$1F,$35,$49,$5F,$DE,$7E,$B1
	.dc.b	$38,$8F,$E0,$6F,$07,$FB,$F3,$5F,$E0,$F1,$C1,$FF,$01,$7E,$41,$FE
	.dc.b	$56,$D3,$FC,$0E,$DC,$FF,$14,$D9,$FF,$3A,$FD,$2F,$FB,$45,$E2,$7F
	.dc.b	$82,$63,$49,$B1,$B8,$FE,$0D,$F2,$3F,$CB,$88,$FF,$26,$F1,$EB,$A3
	.dc.b	$F9,$4C,$7F,$34,$8F,$E6,$BF,$2F,$27,$F9,$6B,$FC,$E3,$E4,$B1,$1F
	.dc.b	$C4,$E2,$E5,$FB,$29,$FE,$7B,$F2,$52,$BF,$9A,$BF,$D0,$3E,$4B,$11
	.dc.b	$FC,$6E,$2D,$7E,$EA,$7F,$A2,$FD,$04,$AF,$E7,$AF,$F8,$6F,$E3,$BE
	.dc.b	$37,$A1,$FD,$1E,$35,$3F,$C9,$5D,$3B,$F8,$AA,$FF,$A8,$30,$FE,$99
	.dc.b	$7C,$8F,$F0,$C1,$23,$67,$F5,$47,$FC,$3D,$F2,$2F,$1F,$FC,$81,$2F
	.dc.b	$A0,$7C,$BF,$F0,$33,$F0,$4E,$23,$FD,$5F,$18,$3F,$84,$57,$3C,$3F
	.dc.b	$D6,$75,$BA,$4E,$5E,$D5,$D7,$9F,$F8,$85,$6E,$3A,$4F,$FA,$F2,$B8
	.dc.b	$4F,$36,$CF,$19,$EE,$D7,$A9,$C7,$0F,$6D,$71,$6C,$EE,$3C,$17,$E2
	.dc.b	$8F,$5C,$EA,$B9,$61,$D9,$F8,$35,$B6,$99,$3D,$F3,$C5,$B3,$9A,$F6
	.dc.b	$C7,$CE,$AE,$4E,$8F,$49,$6F,$CA,$F1,$8D,$E1,$F0,$A3,$CA,$70,$AE
	.dc.b	$A7,$68,$D4,$73,$8F,$80,$47,$8D,$7C,$5B,$ED,$10,$F1,$D7,$23,$9D
	.dc.b	$76,$BF,$1E,$1E,$F9,$CC,$7B,$27,$C5,$39,$3E,$23,$C6,$BE,$45,$75
	.dc.b	$9C,$D7,$13,$AD,$EA,$3F,$2F,$EA,$7F,$05,$AF,$C6,$36,$33,$E6,$7D
	.dc.b	$AA,$60,$AF,$E1,$8E,$3E,$4B,$31,$88,$46,$2B,$FC,$06,$D8,$FF,$7A
	.dc.b	$B5,$FF,$6C,$9F,$62,$9E,$83,$A7,$F0,$93,$DC,$D4,$65,$FB,$E6,$75
	.dc.b	$FD,$45,$72,$AE,$33,$E4,$38,$B8,$BF,$C8,$4E,$E7,$99,$BA,$E7,$7D
	.dc.b	$DF,$DA,$47,$7D,$7F,$D9,$B3,$97,$DF,$1D,$8E,$77,$DA,$75,$1F,$08
	.dc.b	$D5,$74,$7C,$15,$D3,$0F,$34,$6A,$3A,$CF,$92,$72,$D9,$EE,$6C,$EC
	.dc.b	$E4,$ED,$7E,$39,$E3,$5D,$EB,$8D,$6E,$3B,$99,$78,$C7,$53,$EC,$BB
	.dc.b	$7B,$C6,$CE,$2F,$F2,$FB,$F8,$2F,$08,$D3,$C0,$D5,$79,$1A,$79,$23
	.dc.b	$7C,$2D,$FF,$BB,$D7,$F3,$B3,$F2,$F1,$9F,$0B,$F0,$D9,$FF,$3E,$E0
	.dc.b	$B5,$AB,$FF,$0F,$3C,$71,$61,$D6,$B7,$18,$AD,$BF,$D0,$E3,$33,$DE
	.dc.b	$B9,$3E,$5E,$CC,$DF,$F8,$22,$72,$F2,$C3,$FA,$8C,$E8,$93,$AC,$6E
	.dc.b	$3F,$D4,$7E,$63,$99,$D3,$3B,$5D,$22,$CF,$F5,$79,$B8,$7F,$5E,$F9
	.dc.b	$7F,$19,$E4,$CA,$E1,$E4,$8B,$7F,$D7,$E6,$D7,$F6,$6F,$94,$F5,$9E
	.dc.b	$CF,$94,$B8,$7B,$22,$DF,$F6,$79,$B5,$FB,$78,$7C,$57,$F6,$9B,$FA
	.dc.b	$11,$8E,$5F,$8F,$F2,$FF,$42,$31,$3A,$AE,$11,$FE,$B2,$6D,$FF,$7A
	.dc.b	$AE,$11,$8A,$FF,$3A,$7D,$97,$EF,$6B,$C4,$D8,$FC,$17,$89,$DE,$31
	.dc.b	$87,$F2,$06,$A7,$3C,$5F,$91,$9E,$1C,$9E,$7F,$71,$F1,$54,$FC,$4E
	.dc.b	$4F,$9D,$CE,$7F,$02,$AE,$7F,$EC,$C6,$D7,$C7,$6C,$FF,$85,$EA,$D7
	.dc.b	$F3,$49,$AF,$FB,$7E,$FB,$CB,$E2,$FC,$57,$F9,$DC,$3C,$51,$F3,$7E
	.dc.b	$7B,$FD,$93,$3F,$49,$B8,$7F,$87,$FE,$5F,$5F,$01,$2B,$4F,$E7,$23
	.dc.b	$13,$FE,$25,$F3,$FF,$3F,$F2,$BF,$A0,$6D,$FE,$81,$B3,$B8,$FD,$1C
	.dc.b	$FD,$FF,$9D,$75,$D3,$E4,$FC,$EB,$D5,$C5,$FE,$22,$B6,$DA,$DC,$EB
	.dc.b	$0F,$F7,$BD,$69,$6D,$FC,$13,$E7,$DD,$1B,$8C,$3F,$B8,$F9,$7F,$DC
	.dc.b	$FC,$C5,$CF,$F7,$2D,$C3,$FC,$88,$DB,$FF,$25,$BF,$C0,$3E,$62,$57
	.dc.b	$F0,$0F,$98,$B7,$FE,$4B,$9B,$7F,$E5,$29,$B5,$FE,$CB,$CD,$1F,$14
	.dc.b	$FB,$BF,$1B,$67,$E1,$45,$C3,$DB,$1F,$87,$9C,$5F,$EC,$6F,$9C,$FF
	.dc.b	$BD,$A9,$18,$7F,$DF,$2D,$7F,$DB,$86,$AF,$F7,$4F,$F2,$9F,$7F,$F8
	.dc.b	$EF,$CC,$7F,$1E,$6D,$FF,$1E,$6C,$FF,$9A,$5B,$87,$F7,$9E,$D7,$D5
	.dc.b	$FF,$7F,$BF,$E5,$2F,$F0,$66,$9F,$CB,$9C,$5F,$B8,$FE,$6E,$F1,$71
	.dc.b	$3D,$3C,$0E,$15,$C5,$FE,$19,$F8,$63,$E6,$69,$D7,$DC,$B5,$FC,$52
	.dc.b	$74,$D9,$D3,$49,$87,$83,$17,$C2,$72,$8C,$3E,$9A,$95,$EE,$F0,$2D
	.dc.b	$71,$EF,$F7,$1E,$FF,$99,$B5,$F1,$3E,$07,$E1,$1F,$0F,$E6,$6C,$7C
	.dc.b	$8F,$8C,$F8,$C7,$C7,$F9,$9B,$1F,$33,$E6,$7E,$51,$F2,$FE,$66,$C7
	.dc.b	$D0,$F9,$9F,$9C,$7C,$FF,$99,$B1,$F5,$3E,$67,$E9,$1F,$4F,$B2,$C7
	.dc.b	$D8,$F9,$9F,$AC,$7D,$7F,$99,$B1,$F7,$3E,$67,$ED,$1F,$DA,$3E,$9B
	.dc.b	$FC,$99,$19,$3F,$B4,$AF,$7B,$46,$FD,$0B,$3A,$F9,$1E,$47,$3F,$99
	.dc.b	$B1,$CB,$D0,$EC,$77,$F4,$EC,$76,$F9,$9F,$17,$AF,$E3,$6C,$78,$BE
	.dc.b	$63,$CC,$63,$B3,$F1,$BA,$A3,$D5,$F2,$FE,$BF,$97,$B7,$EA,$7D,$DE
	.dc.b	$8F,$BF,$E5,$6C,$FB,$9C,$34,$F8,$31,$8C,$7B,$FC,$7F,$C7,$31,$E5
	.dc.b	$7E,$A7,$34,$7F,$67,$E7,$7C,$65,$3E,$77,$C7,$DC,$BF,$B7,$CE,$9B
	.dc.b	$3A,$69,$31,$F7,$FC,$7F,$16,$C7,$F7,$8F,$9E,$73,$7F,$77,$6C,$7D
	.dc.b	$AF,$90,$C4,$C8,$E0,$FE,$06,$31,$5F,$33,$0F,$C2,$1F,$85,$F9,$0B
	.dc.b	$7F,$84,$8B,$3F,$4A,$6E,$DF,$0F,$E4,$3F,$D0,$9F,$1B,$E4,$AF,$F1
	.dc.b	$DF,$02,$73,$9F,$F9,$1B,$E4,$FF,$D3,$B9,$FF,$A6,$E3,$FC,$93,$5B
	.dc.b	$8C,$5F,$2B,$E2,$D8,$CF,$A6,$C3,$9C,$FF,$A8,$B0,$E7,$7C,$21,$F3
	.dc.b	$FF,$C1,$BE,$4B,$AC,$63,$D1,$46,$3E,$37,$FA,$AE,$7A,$BE,$17,$D4
	.dc.b	$FF,$5A,$B8,$D7,$E0,$47,$3B,$E5,$EF,$78,$6B,$59,$FF,$AF,$2B,$47
	.dc.b	$32,$DF,$2C,$7C,$31,$9A,$E2,$F8,$32,$F3,$B9,$6D,$71,$C3,$C1,$18
	.dc.b	$C3,$C1,$7C,$AF,$BC,$78,$F0,$EF,$87,$AA,$B4,$7C,$0A,$FE,$A6,$D7
	.dc.b	$9D,$F8,$9C,$B5,$F7,$7C,$97,$ED,$FE,$83,$FB,$03,$63,$FA,$E8,$70
	.dc.b	$FF,$60,$6B,$E8,$F5,$7F,$DC,$4F,$0B,$FE,$95,$F2,$7F,$DA,$DB,$19
	.dc.b	$9D,$BE,$87,$8E,$3E,$D6,$E7,$D6,$FE,$F7,$95,$F5,$6B,$51,$F8,$1A
	.dc.b	$DD,$FE,$01,$AF,$E3,$2B,$85,$F3,$8F,$B1,$3A,$C3,$CB,$87,$CD,$BD
	.dc.b	$D7,$3C,$3B,$11,$FC,$04,$69,$FC,$CD,$F3,$24,$FA,$23,$F4,$7F,$3F
	.dc.b	$FE,$B5,$6F,$C7,$F6,$E4,$4E,$67,$F3,$51,$8F,$8F,$D4,$E7,$2F,$DD
	.dc.b	$4F,$0B,$C9,$AA,$DF,$C7,$FE,$6F,$A3,$FC,$19,$CF,$F8,$36,$F2,$E7
	.dc.b	$9B,$FD,$8B,$9A,$F1,$A7,$FB,$02,$30,$FF,$AE,$9A,$FF,$83,$1B,$7F
	.dc.b	$E1,$97,$FC,$93,$1F,$C8,$FC,$5F,$F1,$62,$ED,$FE,$8C,$FF,$8B,$DA
	.dc.b	$FE,$41,$AF,$F8,$B9,$CD,$7F,$A9,$DB,$87,$ED,$DB,$3E,$7F,$8E,$F4
	.dc.b	$AF,$A0,$F8,$15,$EE,$F8,$DF,$6F,$07,$D0,$77,$3A,$A1,$C7,$5F,$4C
	.dc.b	$7E,$A6,$3F,$83,$35,$DB,$5E,$07,$FA,$E9,$E8,$E0,$F9,$1F,$1B,$E2
	.dc.b	$D5,$F1,$B4,$93,$F3,$FE,$2B,$47,$F9,$57,$2F,$C3,$B6,$8D,$31,$F1
	.dc.b	$5C,$2B,$91,$F6,$5F,$07,$C5,$78,$7E,$26,$4F,$DD,$8F,$27,$C4,$75
	.dc.b	$24,$7F,$04,$C7,$F2,$F1,$E3,$FA,$0F,$F6,$D7,$1F,$E0,$7E,$2B,$D8
	.dc.b	$7C,$E8,$CF,$17,$CC,$B4,$7F,$B8,$30,$F9,$F7,$FE,$61,$8C,$47,$F9
	.dc.b	$61,$B1,$9F,$89,$E9,$E1,$F8,$49,$5E,$5A,$FF,$27,$B7,$E1,$FC,$44
	.dc.b	$7B,$E3,$E3,$9F,$F6,$24,$60,$63,$E6,$F8,$6F,$51,$23,$51,$EA,$F8
	.dc.b	$6F,$77,$C3,$7C,$32,$4F,$8D,$F8,$DF,$0B,$F0,$C9,$1C,$B0,$7C,$2F
	.dc.b	$C7,$F8,$5F,$CA,$F8,$0F,$95,$A6,$6F,$87,$C2,$FD,$12,$3F,$05,$FA
	.dc.b	$9E,$AF,$D1,$24,$77,$F2,$7E,$A7,$C2,$76,$F3,$E4,$F8,$E3,$F4,$BF
	.dc.b	$3D,$FC,$DA,$3F,$B2,$70,$A3,$11,$F6,$DA,$FC,$86,$C6,$FE,$0F,$99
	.dc.b	$27,$E4,$CF,$F7,$4E,$87,$2F,$F7,$1F,$9F,$FB,$35,$FE,$9B,$1F,$40
	.dc.b	$F6,$FF,$7A,$D8,$FB,$02,$78,$DF,$48,$FD,$DC,$74,$7F,$D7,$2D,$7C
	.dc.b	$1D,$5F,$C5,$A3,$11,$C2,$3F,$85,$C7,$4C,$37,$3E,$0B,$7B,$DF,$23
	.dc.b	$E1,$7E,$4F,$C2,$CA,$F7,$C6,$5F,$83,$19,$7F,$D6,$BA,$27,$E8,$FC
	.dc.b	$2F,$D2,$F8,$49,$5F,$22,$32,$FC,$A8,$CB,$CA,$3C,$FF,$15,$F1,$BD
	.dc.b	$89,$3F,$1B,$54,$EA,$3F,$DC,$11,$86,$C7,$C3,$23,$C5,$E0,$FD,$5D
	.dc.b	$3F,$D2,$8F,$DC,$F8,$6F,$CC,$FC,$C4,$8C,$10,$F7,$C7,$E3,$8F,$C7
	.dc.b	$FA,$7F,$90,$F8,$7F,$C8,$FC,$3C,$AF,$C3,$C6,$5F,$C4,$C6,$5F,$93
	.dc.b	$1F,$D0,$38,$FF,$19,$C7,$FE,$3B,$FA,$0F,$E2,$35,$FC,$F1,$AF,$FB
	.dc.b	$53,$6E,$07,$F5,$6A,$FF,$23,$C7,$FB,$F9,$AF,$F9,$15,$AF,$F4,$67
	.dc.b	$BB,$63,$9B,$F9,$4F,$A0,$79,$7F,$AC,$F8,$9F,$D6,$F3,$CA,$FD,$1C
	.dc.b	$65,$FD,$2C,$65,$FC,$2C,$7F,$5C,$FA,$07,$77,$F5,$EF,$25,$CF,$4D
	.dc.b	$44,$BF,$D2,$D7,$06,$C7,$1E,$AF,$EC,$FF,$40,$F3,$7F,$B4,$FD,$03
	.dc.b	$D6,$A6,$A2,$61,$FE,$46,$15,$C5,$C3,$5F,$F2,$5B,$5F,$F3,$71,$C2
	.dc.b	$39,$BF,$BF,$8E,$1D,$FF,$B9,$85,$BD,$AF,$EF,$BD,$5F,$C9,$67,$F8
	.dc.b	$0F,$92,$F9,$DF,$F0,$0B,$63,$F8,$2C,$ED,$FE,$18,$DC,$3B,$37,$A3
	.dc.b	$FD,$04,$D9,$FF,$01,$D2,$3F,$C1,$BF,$41,$EA,$BF,$37,$D0,$7E,$0A
	.dc.b	$CD,$9D,$61,$FE,$8D,$6C,$FF,$83,$9A,$9B,$3F,$AC,$8F,$E0,$BF,$1F
	.dc.b	$FE,$01,$F1,$7F,$DB,$B1,$FD,$CD,$AF,$CC,$F2,$3F,$9B,$4E,$CF,$E7
	.dc.b	$64,$8F,$E6,$3E,$3F,$F3,$EF,$8F,$FE,$9F,$F3,$52,$7F,$9A,$C6,$5A
	.dc.b	$43,$F8,$BF,$C8,$7F,$46,$F9,$0F,$21,$23,$C8,$D7,$F8,$EB,$B7,$FA
	.dc.b	$0F,$C8,$7F,$4E,$F8,$FF,$61,$23,$D8,$D7,$F9,$2B,$B7,$FB,$8F,$C4
	.dc.b	$7C,$AD,$3F,$CD,$10,$E1,$FE,$81,$18,$6C,$EC,$FF,$1F,$41,$A6,$67
	.dc.b	$FC,$8A,$FF,$8E,$3E,$35,$E5,$75,$97,$C8,$FE,$35,$FC,$87,$AD,$FE
	.dc.b	$C9,$8C,$74,$7E,$3D,$D5,$F2,$CB,$FC,$9B,$18,$C3,$A7,$D0,$70,$C3
	.dc.b	$FD,$A4,$DC,$7F,$CA,$6E,$EE,$E3,$FE,$BA,$BE,$2F,$FB,$64,$FC,$F4
	.dc.b	$F1,$A3,$C5,$B1,$C6,$BC,$6F,$1B,$F5,$36,$72,$7F,$99,$31,$F5,$3F
	.dc.b	$0A,$39,$3D,$0F,$F3,$4B,$E2,$39,$39,$3F,$D9,$98,$7A,$1E,$8D,$7E
	.dc.b	$59,$EE,$39,$D7,$C4,$9E,$F8,$7A,$F1,$F7,$12,$3C,$6F,$BE,$3E,$11
	.dc.b	$23,$89,$2B,$A9,$DC,$F0,$9E,$7C,$3D,$F3,$CE,$7D,$6F,$39,$F4,$E3
	.dc.b	$C6,$BE,$2D,$F8,$E7,$A9,$F1,$CD,$94,$F8,$73,$E1,$BF,$91,$1C,$2F
	.dc.b	$C0,$F0,$AE,$8F,$99,$F8,$53,$F0,$0F,$9E,$EA,$FC,$71,$F5,$BE,$37
	.dc.b	$F8,$64,$7D,$A6,$E5,$F9,$D6,$E3,$FE,$A2,$7F,$89,$61,$FE,$A4,$F8
	.dc.b	$DC,$63,$FC,$EA,$11,$87,$F9,$EB,$5F,$F7,$C0,$9E,$6F,$03,$E1,$53
	.dc.b	$D3,$FF,$1C,$5F,$EB,$67,$C0,$6A,$33,$97,$87,$0F,$CA,$CF,$F0,$19
	.dc.b	$CB,$E2,$38,$75,$69,$E7,$A7,$EE,$71,$F0,$4F,$E0,$67,$38,$7F,$BA
	.dc.b	$6F,$FC,$95,$7E,$39,$F4,$63,$C6,$7B,$C7,$BA,$31,$7D,$E3,$DD,$58
	.dc.b	$7F,$44,$76,$3F,$D4,$35,$DA,$FD,$91,$FA,$7C,$3C,$AF,$03,$9C,$79
	.dc.b	$6F,$D1,$18,$39,$B8,$C7,$F3,$95,$CB,$0F,$54,$7C,$88,$EB,$8E,$2F
	.dc.b	$E3,$9F,$95,$3C,$C7,$CD,$BE,$8F,$97,$1F,$43,$F3,$F0,$F7,$1C,$CB
	.dc.b	$8F,$B2,$FE,$11,$DC,$AF,$C0,$8D,$CF,$38,$F7,$3F,$07,$1F,$5D,$F8
	.dc.b	$23,$F1,$B5,$C1,$FA,$55,$C2,$FB,$34,$FF,$69,$6D,$FE,$AD,$C7,$FA
	.dc.b	$5F,$87,$FE,$87,$CF,$FA,$3E,$F6,$77,$37,$1F,$F6,$2D,$FF,$8C,$A7
	.dc.b	$89,$FD,$52,$B5,$C1,$FD,$36,$ED,$7F,$82,$6B,$FA,$CF,$66,$27,$2C
	.dc.b	$8F,$F7,$74,$EF,$0F,$EF,$8D,$7F,$D6,$6D,$7D,$7C,$FD,$8E,$F9,$DA
	.dc.b	$FE,$CF,$1F,$E2,$B9,$DC,$76,$8C,$47,$B1,$B8,$FF,$65,$FA,$0F,$49
	.dc.b	$EA,$C2,$DF,$F7,$28,$F7,$F7,$7F,$8C,$63,$07,$FB,$99,$C6,$1F,$DD
	.dc.b	$63,$E8,$FC,$77,$C4,$3E,$2F,$C7,$58,$FE,$FB,$1E,$2C,$0D,$24,$7F
	.dc.b	$94,$7E,$6B,$FB,$AB,$FE,$C2,$6D,$F7,$3E,$BF,$C9,$FD,$95,$1A,$E1
	.dc.b	$FF,$27,$36,$92,$D1,$9B,$4E,$57,$BA,$D4,$E6,$FF,$DA,$7F,$41,$FB
	.dc.b	$8C,$7F,$CB,$4D,$8F,$C8,$F8,$9E,$38,$C5,$F8,$1B,$97,$83,$5D,$73
	.dc.b	$7C,$E7,$D5,$ED,$BE,$1A,$9F,$95,$CD,$F4,$19,$D3,$FD,$CB,$57,$3F
	.dc.b	$F7,$63,$5F,$F3,$9B,$5F,$82,$1C,$3F,$11,$AE,$6F,$C1,$4F,$9A,$F1
	.dc.b	$DF,$CA,$C3,$F5,$93,$98,$FF,$7B,$FD,$03,$CE,$FA,$8D,$8F,$F1,$EF
	.dc.b	$C9,$E2,$64,$7E,$2F,$96,$E7,$F7,$77,$FA,$B1,$FA,$EF,$3F,$F5,$FF
	.dc.b	$23,$6F,$FC,$8E,$DB,$E1,$C0,$8C,$6B,$FE,$46,$7F,$0D,$F3,$CE,$AA
	.dc.b	$7C,$DF,$A0,$7A,$D8,$CB,$DC,$7C,$8A,$E0,$FF,$A9,$5B,$97,$F6,$76
	.dc.b	$BF,$EA,$26,$BC,$6F,$ED,$7D,$03,$B3,$B4,$FE,$6F,$34,$E7,$0F,$F3
	.dc.b	$1F,$C9,$FF,$0E,$F9,$3F,$E1,$ED,$8C,$9F,$E6,$88,$CB,$4F,$F3,$5F
	.dc.b	$C9,$7F,$16,$3F,$8B,$FC,$9D,$8C,$9F,$BB,$F0,$71,$97,$D1,$EF,$47
	.dc.b	$C8,$A9,$5F,$12,$BF,$CE,$9D,$2F,$8B,$CC,$7E,$D4,$70,$2F,$F8,$9C
	.dc.b	$61,$B1,$9B,$F9,$91,$83,$F4,$1E,$BB,$D4,$EB,$C1,$FC,$EE,$31,$5F
	.dc.b	$E8,$8F,$7B,$F9,$9F,$C9,$FF,$34,$6C,$64,$FF,$46,$C6,$5A,$7A,$FA
	.dc.b	$DF,$3E,$8F,$E7,$EF,$F0,$C7,$4E,$5B,$E3,$69,$DB,$5C,$9F,$1F,$57
	.dc.b	$D0,$3E,$6B,$F8,$C3,$67,$AF,$BD,$89,$91,$FE,$F3,$9D,$57,$F8,$D5
	.dc.b	$AF,$FB,$C1,$AF,$EF,$9D,$EB,$FE,$A8,$C5,$F0,$BD,$2B,$F5,$D1,$8B
	.dc.b	$F6,$E1,$F0,$0F,$83,$E3,$58,$F8,$3E,$37,$F4,$AA,$FE,$01,$7F,$ED
	.dc.b	$56,$BF,$C3,$B5,$F9,$73,$BF,$82,$FE,$97,$DB,$F3,$1B,$87,$CC,$F8
	.dc.b	$BF,$9E,$7D,$0F,$94,$B1,$F4,$3E,$4F,$C0,$47,$FC,$AD,$8F,$D8,$8F
	.dc.b	$27,$D0,$7F,$58,$F3,$FE,$CB,$6B,$EC,$FD,$07,$DC,$3E,$EF,$65,$8F
	.dc.b	$BB,$F2,$7E,$D2,$3D,$B8,$11,$F0,$7E,$83,$F0,$A7,$F8,$13,$E8,$3F
	.dc.b	$CC,$55,$F8,$68,$C7,$B3,$23,$F1,$D3,$FC,$C5,$B1,$C7,$E3,$7F,$BA
	.dc.b	$C9,$A4,$9F,$7C,$7F,$2B,$E4,$FE,$4C,$DE,$8F,$F7,$2B,$FD,$38,$C4
	.dc.b	$FF,$60,$E0,$43,$E2,$FD,$1E,$FF,$A5,$B5,$E6,$F9,$1F,$19,$1F,$0B
	.dc.b	$E6,$F9,$0F,$81,$F2,$3F,$05,$B3,$EE,$F3,$FE,$27,$C7,$FC,$56,$C7
	.dc.b	$C2,$8F,$79,$1F,$6B,$F0,$BC,$6F,$98,$70,$DB,$DE,$FF,$7F,$E4,$7C
	.dc.b	$3D,$3E,$97,$C5,$5C,$3F,$61,$3A,$7F,$D3,$97,$FC,$5E,$BF,$89,$CE
	.dc.b	$63,$FD,$40,$D7,$FB,$39,$B9,$FB,$3F,$15,$D6,$3E,$8B,$FE,$11,$F9
	.dc.b	$EF,$B1,$3F,$BF,$6C,$FD,$6F,$8B,$D7,$2C,$AF,$1D,$6A,$BB,$B5,$FE
	.dc.b	$EE,$D7,$FC,$31,$C7,$E7,$E5,$FB,$F9,$FE,$47,$E2,$FF,$24,$CC,$3F
	.dc.b	$C6,$DE,$ED,$C3,$F9,$E3,$F9,$7F,$8B,$FE,$3D,$1F,$EB,$EC,$FF,$0D
	.dc.b	$92,$73,$F1,$3F,$93,$AF,$F1,$19,$8A,$DC,$D8,$F2,$B7,$A7,$F1,$ED
	.dc.b	$F5,$3F,$C6,$39,$7F,$B3,$EF,$8B,$6F,$FC,$6B,$F3,$FF,$C5,$BE,$7F
	.dc.b	$10,$8C,$3E,$73,$FC,$77,$F1,$3F,$D9,$DC,$1B,$F3,$55,$CA,$B8,$E9
	.dc.b	$EB,$AD,$4F,$6B,$E5,$39,$AD,$E3,$FE,$28,$6D,$F9,$B0,$EA,$F8,$2F
	.dc.b	$8D,$6E,$73,$7C,$DB,$19,$C3,$C1,$E5,$7C,$9C,$3C,$67,$AB,$1E,$EE
	.dc.b	$30,$F4,$63,$DA,$3C,$57,$ED,$C7,$C7,$1F,$98,$BF,$8E,$D8,$CB,$E1
	.dc.b	$3F,$DF,$DF,$1A,$EE,$7D,$07,$E9,$A3,$0D,$8C,$CF,$06,$BD,$FC,$F7
	.dc.b	$C7,$7E,$BB,$1F,$EB,$CD,$8C,$CF,$EB,$0F,$B8,$E1,$E5,$1F,$5E,$37
	.dc.b	$A7,$D3,$6E,$DF,$4F,$C6,$FF,$2D,$CE,$6B,$81,$2F,$E5,$4E,$AB,$ED
	.dc.b	$5F,$29,$FC,$7B,$5F,$1D,$7E,$1E,$B8,$46,$A3,$A5,$F7,$8E,$A6,$2F
	.dc.b	$AC,$D8,$FC,$DD,$72,$E3,$7C,$FB,$E0,$F0,$DA,$FD,$14,$EA,$73,$F0
	.dc.b	$BA,$64,$EA,$B8,$3F,$A3,$AE,$B5,$D3,$DD,$FF,$3F,$F6,$7E,$FA,$7F
	.dc.b	$D3,$FC,$78,$E6,$FE,$89,$CF,$F8,$8A,$D6,$5F,$E4,$0C,$BF,$C4,$8D
	.dc.b	$7F,$C8,$97,$C5,$FF,$52,$B7,$0F,$6E,$39,$97,$6F,$E3,$5F,$F6,$47
	.dc.b	$C2,$EF,$23,$B9,$D4,$FF,$92,$5B,$97,$A6,$B7,$3C,$8F,$E5,$DC,$9D
	.dc.b	$FE,$83,$D5,$AF,$F5,$FD,$5D,$BE,$56,$DF,$F8,$CE,$B6,$FF,$B1,$1B
	.dc.b	$AF,$D6,$AE,$07,$5D,$A5,$71,$6B,$FE,$97,$BF,$2B,$7E,$1B,$C0,$47
	.dc.b	$E5,$FC,$1F,$8C,$C3,$B3,$27,$D0,$F8,$3D,$38,$3B,$3E,$BF,$80,$91
	.dc.b	$F5,$9F,$77,$C1,$FB,$FE,$02,$CF,$DA,$8C,$3F,$73,$E0,$3E,$21,$1F
	.dc.b	$F3,$17,$CF,$7F,$44,$7F,$D6,$4F,$FB,$B1,$FF,$80,$19,$7D,$93,$F0
	.dc.b	$11,$98,$DF,$83,$F7,$73,$91,$D2,$37,$EF,$7E,$07,$49,$3E,$08,$E7
	.dc.b	$F4,$12,$3C,$6F,$6F,$7B,$BF,$D0,$7E,$27,$59,$5E,$88,$F1,$FB,$D2
	.dc.b	$3D,$6F,$9B,$DE,$F3,$FD,$07,$E4,$7E,$4E,$57,$C0,$8F,$5F,$D0,$7E
	.dc.b	$57,$92,$4F,$C4,$8F,$7F,$A7,$F9,$9E,$59,$3F,$22,$3E,$1F,$D0,$7E
	.dc.b	$77,$CD,$93,$F3,$23,$F4,$DF,$2F,$FC,$7F,$C1,$FC,$E7,$4F,$E8,$CC
	.dc.b	$F9,$1A,$F9,$1F,$D6,$7B,$DF,$4E,$BF,$74,$DF,$0B,$FC,$79,$87,$FB
	.dc.b	$EF,$2F,$F4,$77,$CF,$FF,$9F,$9F,$F4,$87,$CF,$BD,$AF,$DA,$36,$37
	.dc.b	$DB,$FB,$7A,$E1,$1F,$EB,$99,$D3,$6B,$55,$23,$AC,$74,$30,$FF,$BE
	.dc.b	$6D,$F4,$1F,$03,$E4,$E4,$FF,$AB,$E7,$2D,$9C,$BC,$67,$E3,$C7,$C8
	.dc.b	$F9,$5F,$93,$F0,$72,$B9,$46,$5E,$91,$97,$C5,$1E,$DF,$8A,$F8,$3E
	.dc.b	$9C,$9F,$8F,$19,$7E,$3C,$65,$FF,$6D,$6A,$9D,$47,$F5,$98,$C3,$FE
	.dc.b	$4C,$6B,$98,$FF,$27,$F9,$4C,$DC,$1F,$C8,$7E,$66,$5F,$C3,$8F,$C2
	.dc.b	$98,$7F,$DE,$51,$87,$E4,$7C,$27,$E2,$BE,$16,$46,$08,$7D,$17,$2F
	.dc.b	$D3,$8F,$E6,$9C,$3F,$E2,$1E,$4F,$DF,$7D,$07,$E2,$E8,$1C,$3F,$EC
	.dc.b	$8D,$B8,$9F,$CF,$63,$26,$AB,$A7,$2F,$2A,$E0,$7F,$B8,$67,$71,$CA
	.dc.b	$3F,$05,$F0,$DF,$A4,$F8,$69,$18,$21,$F8,$B7,$2F,$E3,$63,$FA,$77
	.dc.b	$CF,$BB,$7F,$A8,$79,$1F,$E9,$DA,$FF,$78,$35,$69,$FC,$76,$B8,$47
	.dc.b	$FB,$D5,$AE,$B1,$F5,$57,$1A,$FE,$48,$D8,$F7,$57,$09,$FE,$51,$F3
	.dc.b	$EF,$91,$FE,$C5,$F2,$7F,$D1,$27,$F9,$63,$6F,$E3,$56,$F1,$FA,$35
	.dc.b	$CA,$F8,$FB,$32,$73,$7C,$27,$FB,$75,$F1,$9F,$41,$F6,$B0,$EC,$DB
	.dc.b	$DD,$7F,$75,$E3,$79,$1F,$DD,$FD,$87,$CE,$53,$51,$30,$F0,$9F,$86
	.dc.b	$9D,$B8,$0E,$1F,$18,$AE,$EE,$FE,$7F,$FD,$5B,$AF,$F4,$FF,$9F,$FC
	.dc.b	$4C,$FE,$49,$AA,$77,$B7,$E3,$5B,$1B,$E6,$FF,$08,$57,$09,$CF,$5F
	.dc.b	$EC,$7E,$37,$F9,$5F,$CB,$4C,$3F,$89,$47,$F3,$43,$0F,$CA,$9F,$DB
	.dc.b	$CE,$63,$F8,$C7,$C7,$7F,$86,$7E,$6A,$46,$08,$71,$D9,$1A,$8F,$E4
	.dc.b	$93,$FD,$0B,$E3,$7F,$9E,$F1,$49,$FE,$49,$19,$7E,$7E,$3F,$C2,$27
	.dc.b	$51,$FC,$B6,$7F,$A5,$FC,$67,$F4,$6E,$D9,$3F,$CB,$63,$2F,$2F,$03
	.dc.b	$F9,$EF,$17,$F4,$AA,$CC,$FB,$FE,$17,$E3,$91,$FF,$1E,$3F,$E2,$43
	.dc.b	$0F,$91,$DF,$BF,$FE,$3F,$F8,$BF,$E3,$B7,$FC,$2B,$E7,$FF,$05,$D3
	.dc.b	$F8,$0C,$11,$8E,$9F,$E0,$CF,$E1,$EF,$8C,$EA,$BF,$85,$46,$27,$84
	.dc.b	$F5,$9F,$E1,$C7,$58,$E5,$9F,$EF,$A7,$91,$FC,$72,$B9,$57,$0B,$F0
	.dc.b	$B6,$7A,$1B,$AE,$57,$D3,$0C,$CF,$E9,$AF,$8B,$FC,$A2,$FC,$47,$A2
	.dc.b	$F8,$B6,$B8,$BA,$9E,$15,$C9,$61,$E6,$7C,$55,$C2,$75,$5C,$67,$77
	.dc.b	$D9,$D9,$F0,$0F,$31,$C4,$F4,$DF,$57,$8D,$E5,$E1,$5D,$67,$46,$DF
	.dc.b	$01,$9A,$F0,$4F,$23,$8D,$F9,$8F,$65,$72,$9D,$D7,$29,$CC,$72,$9E
	.dc.b	$D1,$C6,$BC,$F3,$93,$8C,$66,$3D,$D3,$D6,$FD,$58,$7E,$1A,$73,$7C
	.dc.b	$67,$FA,$EF,$9E,$F9,$1F,$A9,$5F,$ED,$86,$C6,$63,$E5,$B5,$FF,$67
	.dc.b	$9B,$3F,$AF,$0F,$8A,$EB,$3F,$ED,$D6,$C6,$9A,$F2,$C3,$F9,$35,$BC
	.dc.b	$B7,$67,$E3,$AB,$FB,$9D,$FE,$42,$B5,$39,$9F,$01,$F7,$E3,$2F,$E0
	.dc.b	$4E,$26,$6B,$AB,$93,$8D,$F8,$A3,$2F,$F3,$98,$E2,$E6,$BC,$17,$E3
	.dc.b	$AD,$EB,$EF,$BE,$D5,$A6,$9D,$EB,$84,$66,$FA,$4E,$4D,$BE,$17,$CB
	.dc.b	$5F,$ED,$AB,$EF,$1F,$2C,$DC,$78,$67,$99,$C6,$FB,$CF,$58,$CC,$EC
	.dc.b	$E9,$7E,$36,$CE,$DB,$3C,$E3,$B9,$E1,$79,$B6,$36,$75,$7F,$BD,$DF
	.dc.b	$A1,$FC,$04,$69,$E0,$6A,$BC,$4D,$3C,$45,$AF,$ED,$FC,$8E,$0F,$F1
	.dc.b	$05,$7F,$33,$F0,$FE,$EE,$7F,$EF,$F9,$FE,$FB,$39,$26,$1F,$CF,$EB
	.dc.b	$71,$FE,$40,$BE,$9C,$18,$A4,$77,$AC,$3E,$1E,$EB,$C0,$F8,$BC,$E5
	.dc.b	$FD,$FF,$2F,$EA,$11,$8A,$DB,$63,$27,$F4,$BE,$9E,$7E,$FF,$26,$C6
	.dc.b	$4F,$EB,$51,$96,$9F,$D4,$FE,$7F,$CB,$D9,$FE,$1A,$8C,$1F,$DA,$9C
	.dc.b	$6B,$FD,$7E,$77,$C3,$E8,$6D,$7F,$5D,$F8,$BF,$6F,$CC,$7B,$1B,$19
	.dc.b	$3F,$B6,$46,$5A,$7F,$88,$FB,$FF,$CF,$EF,$86,$BE,$66,$09,$CB,$FE
	.dc.b	$07,$8C,$C6,$38,$FF,$6F,$C3,$FE,$2C,$9C,$F0,$FB,$9E,$7F,$3F,$FC
	.dc.b	$02,$B7,$AF,$E1,$72,$FF,$1A,$4E,$5F,$BD,$19,$CB,$FC,$6B,$39,$7E
	.dc.b	$FC,$66,$BF,$83,$C2,$75,$8F,$F8,$FB,$8B,$D7,$EB,$B8,$3F,$11,$83
	.dc.b	$E7,$1B,$EA,$73,$18,$D3,$FC,$91,$7C,$5F,$F2,$6D,$F1,$BE,$11,$8A
	.dc.b	$E1,$9F,$F9,$32,$F8,$3F,$E6,$0B,$E1,$5D,$30,$EF,$87,$D0,$F9,$F7
	.dc.b	$AF,$F8,$6E,$2C,$36,$FF,$CB,$53,$A7,$F1,$13,$A6,$9F,$E2,$1F,$27
	.dc.b	$F3,$FF,$13,$FA,$06,$C6,$4F,$F1,$34,$65,$A7,$F9,$9F,$E7,$7F,$C8
	.dc.b	$AF,$C0,$F9,$E7,$CA,$FE,$1A,$73,$B7,$B6,$73,$5C,$7A,$7E,$7C,$62
	.dc.b	$7F,$CD,$F5,$C3,$7F,$E0,$73,$97,$ED,$C6,$6F,$FD,$1F,$39,$7F,$D7
	.dc.b	$11,$9A,$FF,$53,$CE,$5F,$E0,$7C,$4F,$8B,$FD,$FF,$C5,$FF,$00,$6E
	.dc.b	$1F,$E4,$56,$D7,$F9,$35,$B8,$FF,$15,$F8,$AF,$DB,$FC,$97,$F0,$86
	.dc.b	$DF,$F9,$35,$B7,$FE,$55,$6D,$7A,$BE,$1F,$F2,$B3,$E8,$6B,$FE,$DA
	.dc.b	$D1,$18,$84,$62,$FF,$BC,$C6,$E7,$F3,$3E,$9B,$D0,$FA,$51,$86,$D7
	.dc.b	$E5,$A7,$4F,$DB,$9D,$3F,$E6,$5F,$86,$FE,$47,$F0,$9F,$C9,$1B,$19
	.dc.b	$3F,$CD,$51,$96,$9F,$E8,$0F,$9A,$FF,$01,$B7,$C2,$FE,$63,$18,$CB
	.dc.b	$FD,$2C,$FF,$0B,$9E,$06,$A7,$6D,$3F,$D6,$DF,$3D,$FC,$DD,$ED,$E8
	.dc.b	$7F,$14,$8C,$5F,$48,$C7,$8D,$E0,$6D,$70,$39,$C6,$5A,$66,$7C,$B1
	.dc.b	$FD,$1B,$E3,$BD,$2D,$9F,$77,$77,$F4,$AF,$6F,$DA,$D8,$F8,$51,$FC
	.dc.b	$F3,$AB,$E0,$B6,$3E,$37,$C1,$7F,$52,$F8,$EF,$8A,$D8,$F9,$5F,$1B
	.dc.b	$FD,$5A,$3E,$4B,$63,$E7,$7C,$EF,$F4,$8F,$9E,$F9,$AD,$8F,$A5,$F3
	.dc.b	$DF,$40,$FA,$3F,$1D,$63,$EB,$7C,$C7,$F6,$2F,$92,$FA,$AD,$8F,$B5
	.dc.b	$F3,$DF,$60,$FE,$D3,$F3,$8F,$6F,$13,$FE,$BE,$8F,$F2,$A3,$FE,$B7
	.dc.b	$D3,$7E,$DF,$F6,$E7,$8F,$C2,$7F,$70,$F9,$4B,$5F,$EB,$D7,$AF,$9D
	.dc.b	$FD,$D3,$7B,$3F,$EC,$17,$C3,$EB,$48,$F0,$7C,$7D,$EA,$FF,$70,$68
	.dc.b	$F9,$9F,$3F,$C6,$FA,$1B,$7E,$7F,$8E,$F5,$98,$BE,$2F,$BB,$E3,$3D
	.dc.b	$FF,$1B,$6B,$DC,$E1,$F8,$1E,$45,$BF,$0B,$F1,$89,$1C,$0D,$88,$79
	.dc.b	$5F,$D5,$BA,$1F,$0F,$FD,$B9,$B5,$F3,$7E,$4B,$13,$23,$E8,$7C,$55
	.dc.b	$C3,$E0,$B7,$0F,$EE,$3B,$A3,$15,$FE,$27,$8C,$3E,$39,$FE,$CF,$F3
	.dc.b	$4E,$6F,$F1,$2F,$CB,$7F,$80,$23,$15,$FE,$F3,$8C,$4E,$A3,$91,$ED
	.dc.b	$C5,$5C,$1F,$7C,$7F,$8D,$F9,$17,$F8,$CC,$FF,$C4,$3F,$33,$6B,$FC
	.dc.b	$40,$DB,$E2,$7F,$7E,$71,$97,$F8,$21,$FC,$AF,$C6,$7F,$A1,$BE,$2B
	.dc.b	$D5,$5F,$E4,$2D,$53,$9D,$7F,$C1,$73,$4D,$45,$F8,$15,$70,$CB,$FD
	.dc.b	$37,$7F,$26,$B5,$F0,$7F,$6E,$73,$87,$26,$AF,$9B,$79,$AF,$8D,$CC
	.dc.b	$61,$DF,$C4,$FF,$63,$D7,$D3,$7C,$6F,$83,$FE,$B7,$CF,$E7,$61,$C6
	.dc.b	$30,$7E,$35,$C5,$72,$BE,$95,$C2,$F9,$CF,$6A,$DC,$E7,$0F,$1F,$99
	.dc.b	$F1,$71,$F1,$47,$A3,$0E,$35,$B9,$F0,$8C,$7B,$BE,$39,$DB,$6B,$91
	.dc.b	$DB,$2B,$1C,$B1,$F6,$3F,$ED,$66,$BD,$CC,$57,$C4,$7F,$DB,$7F,$1B
	.dc.b	$FC,$07,$E7,$3F,$B2,$C2,$30,$FF,$68,$6B,$C3,$B9,$3C,$2F,$FD,$09
	.dc.b	$E3,$5F,$43,$F9,$05,$FF,$70,$9C,$B6,$32,$F9,$DD,$EF,$F3,$5B,$B7
	.dc.b	$D8,$6C,$FD,$7F,$27,$F5,$15,$BC,$3F,$0B,$5A,$AF,$C5,$37,$05,$F0
	.dc.b	$C3,$A5,$F0,$25,$FC,$B9,$D3,$F8,$5B,$E4,$49,$FC,$9C,$EA,$7B,$C7
	.dc.b	$2C,$3C,$91,$CC,$FC,$FF,$23,$A5,$E5,$F6,$8C,$36,$33,$3C,$17,$76
	.dc.b	$4D,$4C,$9D,$47,$1C,$7F,$86,$74,$FF,$A3,$2D,$F3,$38,$E3,$FE,$18
	.dc.b	$DC,$3F,$D8,$33,$F1,$9F,$E2,$A6,$2F,$59,$7A,$5B,$3B,$9F,$E1,$F3
	.dc.b	$ED,$F2,$F7,$F4,$FF,$7B,$2F,$F4,$EB,$72,$FF,$69,$36,$FF,$DA,$2B
	.dc.b	$CC,$E3,$BC,$9F,$1E,$2A,$E0,$DA,$EF,$F1,$12,$B7,$48,$F1,$F1,$66
	.dc.b	$7C,$F3,$97,$73,$A6,$90,$DB,$97,$8F,$CC,$FB,$BC,$C9,$1E,$08,$F8
	.dc.b	$24,$70,$79,$3E,$47,$E2,$91,$C1,$E8,$F3,$3E,$4F,$04,$8F,$5B,$F3
	.dc.b	$3A,$64,$7B,$9A,$FB,$3C,$CF,$A3,$E6,$7D,$26,$47,$C4,$8F,$AB,$E6
	.dc.b	$48,$F8,$EF,$D8,$F3,$3E,$CF,$99,$27,$E5,$BF,$73,$CC,$FB,$BE,$67
	.dc.b	$DE,$F9,$F9,$5F,$42,$3F,$03,$E5,$FE,$0B,$E7,$E4,$FD,$48,$FC,$2F
	.dc.b	$95,$23,$EB,$BF,$8E,$F7,$FF,$B2,$47,$E3,$1F,$E3,$1D,$9F,$EE,$66
	.dc.b	$F6,$7F,$81,$7A,$D1,$8A,$46,$3E,$4B,$A7,$95,$27,$F1,$75,$FC,$C5
	.dc.b	$BF,$25,$E1,$F8,$99,$1A,$7C,$9F,$25,$E5,$F9,$29,$3C,$9F,$47,$C5
	.dc.b	$7A,$7E,$3A,$4F,$67,$D9,$F1,$3E,$DF,$4A,$4F,$89,$F8,$1E,$E7,$C1
	.dc.b	$F5,$A4,$F9,$9F,$89,$F1,$DF,$B7,$87,$CE,$7F,$88,$9F,$DB,$98,$E6
	.dc.b	$D4,$66,$3F,$C0,$67,$0E,$7C,$CC,$8E,$33,$BE,$67,$37,$F8,$56,$17
	.dc.b	$BA,$E7,$1B,$9D,$5F,$82,$BE,$D7,$C7,$FD,$BF,$8D,$98,$7D,$08,$CB
	.dc.b	$F4,$A3,$2F,$DF,$8F,$C1,$7C,$77,$E0,$FE,$3A,$4F,$D8,$8C,$BF,$6A
	.dc.b	$32,$FE,$1E,$3F,$15,$F1,$B2,$3E,$FB,$F8,$DF,$8D,$C4,$7E,$36,$2E
	.dc.b	$BF,$8D,$9B,$A7,$E3,$7A,$7F,$D9,$B7,$C8,$FC,$4C,$FD,$3F,$9C,$7C
	.dc.b	$37,$F6,$A9,$46,$1F,$ED,$AD,$7F,$DA,$6D,$7F,$3E,$E7,$0F,$A3,$1F
	.dc.b	$E6,$18,$C6,$1F,$E4,$76,$C7,$F9,$97,$83,$F9,$43,$F9,$CF,$8E,$FD
	.dc.b	$51,$1F,$E5,$3F,$51,$D1,$FD,$64,$65,$A6,$AF,$F1,$DF,$38,$F8,$DA
	.dc.b	$49,$FC,$7C,$3E,$33,$11,$F8,$E8,$E0,$EA,$79,$CE,$B3,$F1,$56,$AB
	.dc.b	$85,$69,$F3,$FC,$7F,$F5,$0C,$9F,$31,$4C,$44,$9F,$6C,$65,$E2,$7E
	.dc.b	$93,$E1,$1E,$4F,$D5,$6C,$7F,$88,$1A,$FF,$10,$3F,$CA,$31,$B3,$F6
	.dc.b	$11,$AC,$BE,$53,$74,$FF,$3D,$69,$6F,$FD,$31,$97,$B6,$FE,$83,$6B
	.dc.b	$FD,$00,$D9,$DD,$7E,$F6,$31,$06,$92,$35,$9C,$9F,$E0,$33,$97,$F8
	.dc.b	$1F,$DC,$FF,$AF,$5F,$F1,$EB,$5F,$73,$5F,$E5,$ED,$7F,$D2,$8E,$F0
	.dc.b	$FC,$43,$70,$CF,$A3,$FC,$4F,$D2,$7C,$87,$F1,$5F,$59,$F1,$9F,$A3
	.dc.b	$6E,$5F,$E4,$E6,$BF,$3D,$AF,$F4,$07,$39,$7E,$5D,$B8,$FF,$A9,$DB
	.dc.b	$3F,$EC,$5F,$A4,$F0,$BF,$9E,$DF,$F1,$04,$7F,$43,$F0,$B5,$FF,$0B
	.dc.b	$ED,$FE,$8E,$78,$1A,$3F,$4F,$F2,$FF,$91,$23,$FA,$AD,$2B,$CA,$2E
	.dc.b	$7F,$B8,$F4,$3E,$C5,$7E,$67,$86,$5F,$EC,$9F,$CE,$71,$62,$3C,$71
	.dc.b	$6F,$96,$3F,$AD,$9F,$DC,$FC,$97,$E8,$C8,$FE,$F3,$AE,$BE,$C8,$B5
	.dc.b	$CA,$7F,$55,$F2,$5F,$AB,$65,$7F,$04,$8C,$BC,$C8,$7C,$67,$FA,$87
	.dc.b	$C5,$FF,$96,$6B,$FD,$E9,$F3,$4F,$43,$FD,$B4,$3D,$4F,$F6,$E4,$67
	.dc.b	$0F,$F7,$D4,$E5,$E2,$FF,$B8,$6F,$A4,$27,$83,$FE,$93,$E7,$FF,$80
	.dc.b	$2E,$6B,$BD,$F6,$6C,$74,$07,$1C,$30,$D3,$91,$E4,$9E,$4D,$37,$87
	.dc.b	$93,$0F,$1B,$6F,$BC,$F2,$3E,$15,$72,$7B,$57,$0C,$3D,$2F,$B2,$BE
	.dc.b	$14,$F1,$6C,$6F,$0F,$0D,$70,$9E,$F3,$C6,$FB,$61,$E3,$C3,$DE,$D7
	.dc.b	$C6,$7C,$0A,$EA,$7C,$FB,$E9,$1C,$1A,$F0,$3C,$95,$DA,$33,$5D,$6B
	.dc.b	$9D,$7C,$73,$E1,$47,$84,$D5,$7C,$39,$DC,$FC,$47,$C5,$87,$96,$74
	.dc.b	$77,$7E,$09,$99,$87,$6B,$F1,$9E,$8A,$EA,$76,$BF,$14,$66,$FA,$D7
	.dc.b	$43,$95,$FE,$1B,$E1,$3F,$B4,$C7,$E1,$5B,$7F,$94,$6D,$FE,$49,$DC
	.dc.b	$FE,$3B,$E2,$B9,$7C,$2F,$F4,$96,$C7,$F9,$69,$AF,$FB,$60,$4F,$25
	.dc.b	$1F,$96,$C1,$F0,$3F,$E0,$2D,$3F,$47,$5D,$CF,$C8,$4E,$BA,$BF,$3B
	.dc.b	$19,$9F,$D4,$CE,$63,$F2,$C6,$75,$FF,$62,$DF,$57,$D1,$1D,$23,$0F
	.dc.b	$A4,$C5,$F6,$7C,$E6,$0E,$A6,$2F,$C3,$3C,$A3,$E0,$46,$27,$96,$1E
	.dc.b	$FA,$D5,$79,$4F,$DC,$E1,$E9,$BF,$83,$1D,$70,$F4,$E1,$CA,$3C,$53
	.dc.b	$AC,$3A,$4E,$63,$C1,$3E,$69,$CB,$4E,$35,$F0,$A7,$67,$38,$C9,$F0
	.dc.b	$DC,$C6,$CF,$1C,$73,$8C,$BB,$8C,$C7,$33,$31,$F1,$23,$27,$37,$E0
	.dc.b	$E1,$F1,$1F,$80,$62,$FD,$93,$C2,$3F,$D3,$FC,$2F,$93,$FF,$24,$E5
	.dc.b	$FD,$2B,$E4,$EE,$1F,$EA,$7A,$D3,$FE,$44,$8C,$DF,$2B,$BC,$FF,$E1
	.dc.b	$19,$B7,$C2,$BF,$B4,$C6,$3B,$66,$1C,$CF,$F2,$7B,$89,$FE,$DE,$F2
	.dc.b	$9C,$63,$FD,$7B,$E4,$BA,$9D,$BE,$36,$E1,$FE,$58,$8B,$87,$F7,$89
	.dc.b	$B5,$FD,$9B,$E3,$7C,$A7,$9B,$E3,$2E,$1F,$E6,$08,$B7,$FE,$00,$9B
	.dc.b	$5F,$DB,$BE,$37,$DA,$7B,$BE,$36,$E1,$FE,$68,$8B,$7F,$E0,$89,$B5
	.dc.b	$FE,$6B,$FB,$C7,$A7,$F2,$A9,$18,$C3,$F3,$67,$FB,$5F,$BF,$F8,$0B
	.dc.b	$FE,$B8,$8D,$4F,$FA,$D9,$AF,$FB,$89,$AF,$F5,$C3,$AD,$7F,$9E,$DC
	.dc.b	$46,$9F,$EC,$3F,$36,$F8,$3F,$23,$6F,$C7,$2C,$5F,$80,$3C,$FD,$E1
	.dc.b	$2F,$FC,$C9,$6C,$39,$3F,$E6,$76,$BF,$EF,$86,$BF,$BA,$F9,$CF,$F4
	.dc.b	$54,$F8,$31,$46,$27,$C5,$3B,$D3,$F2,$75,$AF,$12,$DE,$E3,$F3,$AD
	.dc.b	$AE,$07,$57,$17,$ED,$6C,$F3,$9D,$5F,$4B,$FB,$1F,$38,$F8,$BA,$78
	.dc.b	$3C,$4B,$7F,$EE,$E9,$D3,$F1,$1B,$97,$FA,$25,$AF,$D7,$0E,$1F,$C1
	.dc.b	$78,$3F,$AE,$C5,$F2,$16,$30,$DB,$FA,$FD,$9F,$EA,$F8,$C5,$FC,$81
	.dc.b	$19,$9F,$85,$D3,$F2,$0F,$93,$D3,$63,$24,$AE,$5B,$7D,$49,$D4,$63
	.dc.b	$75,$E2,$BE,$8D,$5B,$FC,$E7,$C9,$7A,$CF,$67,$9D,$63,$D3,$AA,$71
	.dc.b	$97,$A7,$3F,$F6,$55,$E3,$1F,$F6,$B7,$CD,$BD,$AF,$D6,$37,$2F,$BD
	.dc.b	$DB,$89,$91,$F9,$69,$D6,$3F,$EB,$B6,$BF,$4D,$AF,$EC,$FE,$4B,$E9
	.dc.b	$1F,$4F,$C4,$B5,$E8,$8B,$3E,$7D,$7F,$B6,$65,$FE,$7F,$53,$A8,$C3
	.dc.b	$CE,$B9,$1F,$0B,$E4,$BF,$01,$E4,$FE,$05,$B1,$93,$E3,$C6,$5A,$7C
	.dc.b	$AF,$4F,$F1,$1F,$25,$F8,$96,$C6,$4F,$9F,$19,$69,$FA,$5F,$0F,$F5
	.dc.b	$67,$EB,$3C,$3B,$19,$26,$9F,$81,$8C,$BF,$83,$9B,$9F,$EC,$A6,$D7
	.dc.b	$F1,$DF,$92,$FC,$E1,$F9,$DE,$CB,$1F,$82,$F9,$2F,$D0,$1F,$A1,$F9
	.dc.b	$2B,$1F,$86,$F9,$2F,$D2,$1F,$A5,$F9,$2B,$1F,$8A,$F6,$FF,$50,$7E
	.dc.b	$A7,$E1,$EC,$7E,$3B,$E1,$FF,$58,$7E,$B7,$C6,$B1,$FA,$AF,$A6,$FF
	.dc.b	$42,$B7,$87,$FA,$FC,$3E,$5E,$C6,$3D,$79,$5C,$7E,$52,$CF,$E6,$6D
	.dc.b	$F2,$9C,$8E,$7F,$29,$6F,$C1,$EB,$76,$F9,$4E,$ED,$8F,$27,$CA,$78
	.dc.b	$BE,$53,$C6,$D8,$F4,$7C,$A7,$98,$F3,$FC,$A5,$8F,$67,$CA,$7A,$8F
	.dc.b	$5F,$CA,$58,$F8,$1F,$29,$EE,$F6,$7D,$ED,$8F,$89,$F2,$9F,$08,$C7
	.dc.b	$0F,$B5,$F8,$BF,$27,$27,$67,$CB,$7E,$4F,$CA,$48,$F9,$DE,$65,$BF
	.dc.b	$8D,$F2,$7F,$38,$C7,$89,$FE,$E7,$E6,$46,$6F,$EB,$7B,$76,$FE,$73
	.dc.b	$FE,$14,$F9,$14,$D3,$1F,$23,$5F,$E4,$F1,$A6,$9A,$29,$2B,$F8,$94
	.dc.b	$6D,$B3,$CD,$CF,$C8,$F8,$4F,$17,$C7,$DB,$EF,$16,$79,$CD,$CF,$CF
	.dc.b	$1F,$CD,$B0,$7C,$7F,$AC,$F6,$7C,$7D,$C7,$D3,$1E,$FF,$8F,$C5,$F9
	.dc.b	$9F,$85,$F0,$3D,$3E,$2F,$F9,$64,$F5,$F5,$71,$ED,$FF,$B8,$BD,$B6
	.dc.b	$9F,$DB,$EF,$71,$8A,$DD,$F1,$CF,$FD,$39,$87,$1C,$15,$AB,$E7,$8D
	.dc.b	$9C,$DF,$0C,$3F,$DE,$3F,$2C,$DF,$F5,$91,$FE,$9E,$CB,$FC,$35,$39
	.dc.b	$F1,$F8,$4B,$C6,$77,$F9,$21,$5B,$79,$1D,$B0,$EB,$B7,$C2,$8D,$C6
	.dc.b	$31,$FF,$14,$61,$D3,$2E,$F8,$73,$C3,$83,$6B,$CB,$7C,$70,$1E,$5C
	.dc.b	$7C,$95,$B8,$F1,$65,$D2,$B8,$46,$F1,$C4,$C8,$F4,$9E,$67,$E0,$E5
	.dc.b	$E5,$8C,$DF,$57,$47,$F8,$2C,$A7,$F8,$35,$F8,$8D,$BF,$9E,$FF,$84
	.dc.b	$7E,$39,$EB,$3E,$5F,$FC,$77,$18,$6C,$66,$7B,$4E,$5F,$05,$FE,$63
	.dc.b	$E3,$FF,$C8,$6D,$8C,$F2,$7F,$B7,$1B,$3A,$8C,$6F,$F3,$9B,$BF,$E0
	.dc.b	$9B,$5F,$88,$F5,$FF,$CB,$D5,$BB,$FA,$F5,$AA,$FC,$1B,$70,$5F,$0C
	.dc.b	$3A,$5F,$02,$5F,$CC,$9D,$3F,$96,$BE,$44,$9F,$C2,$CE,$A7,$BC,$72
	.dc.b	$C3,$C9,$1C,$CF,$D0,$71,$BD,$97,$C0,$76,$8C,$36,$33,$3C,$17,$BF
	.dc.b	$99,$93,$97,$84,$71,$C7,$FD,$05,$D5,$FC,$33,$0F,$07,$B5,$8E,$5F
	.dc.b	$E0,$19,$23,$19,$7F,$91,$A3,$E3,$BF,$E4,$C6,$BF,$E3,$37,$83,$98
	.dc.b	$E3,$5E,$DC,$2F,$CB,$FD,$FA,$FF,$1D,$37,$1F,$23,$71,$FE,$24,$BE
	.dc.b	$17,$8E,$D2,$FF,$CB,$B3,$B7,$FC,$B6,$ED,$CF,$C2,$F8,$08,$F0,$E1
	.dc.b	$AF,$8A,$2E,$3E,$29,$B8,$F8,$A6,$D7,$69,$F5,$FC,$1F,$B3,$E0,$64
	.dc.b	$F9,$23,$2D,$21,$F0,$3E,$0B,$E1,$9A,$7E,$1C,$61,$F6,$F0,$7C,$AE
	.dc.b	$D7,$C9,$53,$E5,$C7,$CC,$E9,$95,$F1,$1F,$9F,$F0,$5F,$43,$E0,$AC
	.dc.b	$FC,$88,$C3,$F2,$BE,$0B,$FD,$6F,$CA,$E2,$FF,$4E,$4F,$F0,$98,$C4
	.dc.b	$EA,$73,$3F,$E9,$79,$E1,$F0,$3F,$78,$91,$F5,$BE,$07,$F0,$3F,$03
	.dc.b	$F8,$2F,$81,$93,$F6,$A3,$F0,$BF,$03,$F8,$6F,$81,$93,$F7,$A3,$F1
	.dc.b	$3F,$03,$23,$F0,$2F,$E3,$BE,$5F,$FD,$51,$D5,$F9,$D7,$F8,$DB,$7C
	.dc.b	$8D,$1B,$F8,$2E,$1E,$4C,$C3,$FD,$4F,$7F,$D2,$DB,$F0,$5D,$7E,$0A
	.dc.b	$46,$9F,$07,$C0,$F8,$7E,$0B,$EB,$7C,$9C,$AE,$71,$E5,$F8,$3F,$B5
	.dc.b	$F1,$52,$7B,$C7,$A7,$D1,$91,$E2,$7D,$9F,$05,$ED,$F8,$29,$3E,$67
	.dc.b	$E0,$7C,$0F,$ED,$A1,$F2,$CE,$5C,$36,$3F,$DF,$26,$A3,$3C,$DA,$64
	.dc.b	$6A,$78,$3F,$EC,$73,$AC,$E9,$C4,$6B,$97,$FC,$2F,$1D,$4E,$53,$F5
	.dc.b	$7E,$0F,$EB,$7C,$1C,$9F,$9F,$19,$7E,$8C,$65,$FF,$5F,$72,$27,$EF
	.dc.b	$7C,$1F,$DF,$F8,$39,$5F,$62,$32,$FD,$A8,$CB,$CA,$3F,$0D,$F0,$72
	.dc.b	$3E,$FB,$F8,$9F,$82,$C4,$78,$22,$E7,$F8,$B9,$BE,$7F,$F8,$97,$2F
	.dc.b	$F6,$75,$F2,$3F,$13,$3F,$4F,$E5,$DC,$FF,$4B,$CF,$95,$F4,$A7,$2D
	.dc.b	$9F,$EE,$CF,$F8,$22,$32,$FF,$79,$6B,$FE,$0C,$35,$1F,$B8,$B7,$C1
	.dc.b	$FE,$8C,$93,$FE,$12,$CB,$FA,$07,$C9,$FD,$67,$F4,$D3,$A9,$F9,$EF
	.dc.b	$F3,$89,$D6,$7D,$FC,$E7,$BF,$F8,$F6,$DF,$E3,$1A,$FE,$1D,$FF,$20
	.dc.b	$B9,$8E,$04,$78,$7C,$05,$78,$45,$C7,$F1,$75,$CA,$7F,$E0,$06,$0F
	.dc.b	$84,$F3,$7C,$24,$9F,$76,$5F,$E5,$9A,$D4,$60,$79,$68,$CC,$74,$AE
	.dc.b	$4F,$6B,$FD,$67,$C1,$3C,$AF,$D7,$6A,$E9,$A6,$A2,$5F,$DA,$AE,$0D
	.dc.b	$8F,$BA,$FF,$97,$5B,$8F,$C8,$D7,$F5,$2D,$D3,$EF,$36,$BF,$C9,$6F
	.dc.b	$C3,$9D,$ED,$FC,$06,$35,$3E,$D8,$C6,$59,$64,$7F,$02,$9D,$9C,$EF
	.dc.b	$F1,$9F,$2F,$FE,$40,$9F,$E0,$FE,$9E,$5F,$F2,$DB,$54,$6D,$FE,$0C
	.dc.b	$D7,$F2,$C6,$BD,$3F,$83,$19,$8E,$F1,$F1,$3D,$57,$CA,$7F,$1B,$F9
	.dc.b	$7E,$C7,$78,$CD,$7F,$1A,$6B,$FE,$5D,$6B,$E0,$6B,$FA,$76,$BF,$82
	.dc.b	$CB,$FD,$1C,$FE,$17,$8F,$F0,$9C,$1F,$A1,$3F,$45,$F0,$56,$75,$5C
	.dc.b	$1F,$C9,$CE,$9C,$FC,$3F,$EA,$48,$F0,$F8,$7A,$FE,$6A,$2F,$17,$ED
	.dc.b	$27,$F6,$1C,$32,$BB,$BF,$B3,$E2,$C4,$7E,$8A,$2D,$7E,$F2,$7F,$71
	.dc.b	$ED,$4A,$F3,$BF,$BB,$EB,$C4,$7E,$AA,$2D,$7F,$04,$9F,$E0,$1E,$5E
	.dc.b	$EB,$DF,$18,$7E,$0C,$65,$FD,$9C,$65,$FE,$8D,$F0,$DF,$E5,$BA,$FF
	.dc.b	$50,$FC,$BB,$CF,$FE,$30,$3D,$3F,$E3,$51,$9B,$FF,$53,$CE,$5E,$2F
	.dc.b	$F1,$FB,$E9,$09,$E0,$A3,$31,$89,$E5,$3B,$7C,$35,$E0,$7A,$74,$30
	.dc.b	$EF,$5F,$EB,$8C,$39,$D7,$08,$C9,$EB,$BE,$4E,$EF,$93,$DC,$C5,$F0
	.dc.b	$AF,$1E,$1E,$5B,$E3,$5E,$9B,$E3,$3E,$CC,$3C,$33,$F1,$2B,$77,$C8
	.dc.b	$F4,$E1,$E4,$AD,$D7,$94,$F0,$57,$68,$C9,$AA,$E5,$3D,$4E,$58,$78
	.dc.b	$A3,$AC,$7A,$CE,$85,$78,$BE,$9B,$E2,$78,$18,$71,$AF,$0D,$75,$8F
	.dc.b	$01,$CE,$FC,$D1,$98,$E8,$64,$E7,$87,$C5,$8D,$57,$AE,$BE,$29,$CA
	.dc.b	$FC,$77,$C4,$97,$C6,$32,$FD,$38,$EB,$87,$C0,$8D,$43,$E0,$3F,$D1
	.dc.b	$8F,$91,$B8,$7E,$51,$B8,$FE,$09,$DC,$87,$C3,$58,$CC,$60,$ED,$F0
	.dc.b	$5F,$EF,$79,$D3,$63,$27,$F2,$C8,$FC,$9E,$4F,$80,$76,$FE,$5A,$BC
	.dc.b	$15,$A9,$CC,$7D,$B3,$27,$DD,$AE,$55,$C6,$3C,$99,$FE,$BA,$73,$1C
	.dc.b	$4C,$E9,$FE,$CC,$BE,$B5,$FB,$49,$EA,$F1,$8C,$CF,$AE,$31,$7D,$F0
	.dc.b	$F3,$36,$FC,$97,$C1,$F8,$06,$23,$55,$CA,$3C,$95,$C9,$A6,$CB,$1B
	.dc.b	$3C,$CF,$A8,$FF,$13,$57,$77,$FC,$5D,$3C,$E7,$6D,$BF,$51,$F0,$6F
	.dc.b	$C6,$DA,$EE,$DA,$E3,$1E,$36,$9E,$38,$E7,$1E,$29,$E6,$D3,$99,$67
	.dc.b	$8C,$F5,$38,$C7,$F1,$BB,$F5,$3F,$4A,$34,$FF,$61,$F9,$7F,$C1,$E3
	.dc.b	$FE,$44,$F9,$7B,$8F,$F8,$A6,$6D,$7F,$67,$9B,$D5,$FE,$29,$9B,$D2
	.dc.b	$9B,$18,$73,$DB,$30,$FE,$D3,$17,$0F,$F1,$DC,$DC,$7C,$55,$FD,$BA
	.dc.b	$3A,$FC,$8C,$8F,$EE,$3F,$13,$6F,$FC,$87,$36,$FF,$CB,$B3,$6B,$FC
	.dc.b	$85,$F1,$3E,$63,$CF,$F1,$16,$3F,$BD,$C5,$C3,$FC,$9B,$36,$FF,$CC
	.dc.b	$FF,$13,$EE,$3D,$FF,$2F,$70,$FF,$01,$C5,$BF,$F2,$AC,$DA,$FF,$51
	.dc.b	$FC,$BF,$FA,$66,$FF,$CF,$AD,$AF,$F1,$28,$70,$FF,$8A,$1A,$FF,$9F
	.dc.b	$DE,$13,$FB,$AE,$07,$96,$E0,$FA,$F7,$CF,$D4,$FA,$F3,$9C,$7F,$D8
	.dc.b	$3F,$2C,$F8,$3F,$F5,$BD,$FF,$A2,$E3,$13,$FE,$92,$6C,$79,$9A,$FF
	.dc.b	$8F,$1C,$E5,$F8,$AF,$95,$FE,$23,$5D,$AF,$F1,$32,$9C,$DF,$F6,$FA
	.dc.b	$FF,$4A,$47,$FA,$31,$FF,$3B,$4E,$FC,$3E,$32,$F0,$EC,$4A,$DC,$F1
	.dc.b	$BE,$87,$F7,$B7,$84,$6A,$78,$4E,$AF,$B4,$F2,$BC,$C7,$FB,$6B,$E5
	.dc.b	$9F,$15,$F1,$AF,$F9,$D4,$62,$FF,$D6,$ED,$8F,$B6,$D7,$FC,$C2,$E6
	.dc.b	$BE,$36,$DF,$5A,$31,$5F,$0A,$3F,$5B,$83,$BA,$FC,$76,$1B,$5F,$E5
	.dc.b	$79,$D3,$67,$4D,$3E,$2F,$47,$CA,$3E,$5F,$45,$8C,$93,$3F,$89,$B2
	.dc.b	$75,$18,$AB,$A7,$DF,$9B,$7F,$D7,$78,$FE,$B7,$AD,$EA,$3D,$7E,$65
	.dc.b	$9F,$44,$5C,$7E,$BD,$7D,$DA,$C6,$1F,$D5,$A5,$F2,$CF,$67,$E9,$46
	.dc.b	$1B,$19,$C7,$FD,$D2,$D9,$FD,$6B,$6B,$F8,$FF,$AE,$F8,$BF,$D3,$CE
	.dc.b	$6F,$FD,$CD,$39,$7E,$77,$C7,$FD,$43,$EA,$F8,$96,$32,$4C,$7D,$7C
	.dc.b	$5F,$D9,$B2,$FD,$B2,$BE,$13,$A7,$9D,$71,$AD,$FC,$B7,$E0,$BC,$9F
	.dc.b	$C1,$B6,$7F,$00,$DB,$EE,$DA,$EF,$EA,$7E,$2B,$E3,$FF,$16,$DB,$FC
	.dc.b	$43,$6F,$CE,$DA,$FD,$D7,$87,$FB,$03,$F6,$3E,$1D,$BF,$D7,$C5,$BF
	.dc.b	$C3,$CD,$AE,$35,$A3,$F6,$B1,$97,$F0,$DA,$3E,$43,$F3,$E7,$E8,$3B
	.dc.b	$6C,$FE,$77,$E5,$BF,$46,$7E,$93,$8E,$C7,$E8,$7E,$43,$F8,$EF,$93
	.dc.b	$FA,$86,$C7,$E9,$7D,$AF,$E4,$3E,$57,$EB,$1B,$1F,$A9,$F6,$FF,$5E
	.dc.b	$7E,$C3,$C5,$B1,$FE,$C3,$F9,$4F,$F1,$C3,$78,$7F,$B0,$C3,$E4,$2C
	.dc.b	$63,$D4,$95,$C7,$E4,$2C,$FE,$7A,$DF,$21,$C8,$E7,$F2,$16,$FC,$1E
	.dc.b	$9F,$6F,$90,$EE,$D8,$F2,$7C,$B7,$8B,$E4,$7C,$6D,$8F,$47,$CB,$79
	.dc.b	$8F,$3F,$C9,$58,$F6,$7C,$97,$F1,$BF,$25,$AA,$30,$D9,$FE,$AB,$3A
	.dc.b	$9C,$FC,$8F,$C2,$F5,$BE,$1B,$6B,$E4,$7C,$7F,$C6,$31,$C3,$F0,$5F
	.dc.b	$93,$F1,$DF,$D5,$3D,$0B,$5F,$EF,$1F,$61,$1F,$3B,$E3,$FE,$7F,$C7
	.dc.b	$DA,$FA,$8E,$1A,$7D,$78,$C6,$3F,$3E,$7E,$BF,$C7,$63,$C7,$FE,$D1
	.dc.b	$C0,$8F,$EC,$B8,$43,$C5,$F1,$B7,$0F,$E2,$B9,$FF,$58,$C3,$FB,$83
	.dc.b	$6B,$F9,$02,$AE,$13,$A9,$A6,$A2,$D3,$87,$FB,$CF,$C7,$7F,$76,$3F
	.dc.b	$BB,$FC,$5D,$BF,$89,$16,$BC,$73,$70,$F1,$FC,$67,$F7,$EF,$8B,$FE
	.dc.b	$FE,$DC,$3E,$63,$6F,$D6,$DA,$F5,$F9,$7F,$E8,$4F,$86,$7B,$3E,$C8
	.dc.b	$C3,$EE,$F8,$0B,$87,$A7,$7F,$ED,$AF,$FA,$72,$78,$65,$FE,$90,$6C
	.dc.b	$64,$E1,$5C,$B6,$FF,$0C,$E1,$D2,$5F,$2A,$DB,$F5,$D1,$C4,$CB,$8C
	.dc.b	$FC,$F1,$FE,$AD,$F0,$DD,$9C,$EF,$97,$8B,$E3,$AF,$F5,$94,$61,$E8
	.dc.b	$DB,$E7,$0B,$E7,$AF,$18,$F0,$DD,$BE,$1A,$FF,$AF,$67,$27,$9E,$F8
	.dc.b	$BD,$B2,$F4,$4E,$EB,$8C,$F9,$27,$2F,$AF,$2F,$54,$62,$3A,$E5,$E6
	.dc.b	$AE,$31,$E7,$AD,$E3,$E5,$BF,$93,$3E,$53,$CA,$5A,$F8,$4D,$8D,$BE
	.dc.b	$BF,$89,$7A,$6F,$96,$FF,$2B,$CA,$73,$18,$6C,$7F,$97,$1A,$F7,$E3
	.dc.b	$FF,$75,$3F,$E6,$0C,$11,$88,$D7,$C5,$76,$9C,$B6,$78,$E9,$F0,$5B
	.dc.b	$AF,$CB,$6D,$FD,$EF,$57,$FD,$73,$39,$BF,$B1,$5A,$AF,$C7,$B7,$05
	.dc.b	$F0,$9F,$C3,$CE,$49,$5F,$81,$9D,$4F,$28,$EA,$47,$F1,$B1,$AA,$F5
	.dc.b	$C7,$E4,$2B,$9B,$A3,$F2,$DC,$5F,$E7,$CF,$7F,$C4,$FF,$9F,$BE,$22
	.dc.b	$F3,$1B,$32,$B8,$CE,$5B,$39,$7F,$65,$5F,$E8,$8C,$7F,$31,$CF,$FC
	.dc.b	$71,$BC,$BF,$C7,$1B,$5F,$BF,$F5,$ED,$F1,$8F,$E7,$13,$8C,$7E,$4C
	.dc.b	$7F,$3B,$AC,$7C,$47,$0F,$92,$FB,$D9,$7F,$A6,$5B,$97,$F2,$A9,$DD
	.dc.b	$8F,$F2,$2C,$E9,$FE,$50,$DC,$BF,$D4,$33,$D7,$E1,$7B,$7C,$2C,$BE
	.dc.b	$58,$F7,$9D,$46,$31,$FF,$2C,$62,$BE,$13,$E0,$AD,$BE,$28,$CD,$F3
	.dc.b	$8E,$47,$93,$8B,$DC,$47,$39,$30,$7C,$27,$C3,$35,$3E,$EA,$C3,$C2
	.dc.b	$76,$D9,$CB,$F1,$3E,$03,$E6,$7C,$0F,$CD,$F8,$09,$3C,$63,$E8,$7C
	.dc.b	$0E,$A3,$07,$57,$EA,$75,$EA,$30,$78,$5F,$B1,$F0,$32,$3C,$8F,$DB
	.dc.b	$F8,$1F,$B9,$F0,$3F,$77,$E0,$65,$7A,$63,$F0,$1F,$03,$23,$D8,$FE
	.dc.b	$0F,$E0,$3F,$09,$F2,$B2,$7E,$03,$F8,$7F,$80,$FC,$47,$CA,$7E,$27
	.dc.b	$E0,$65,$7C,$58,$FC,$67,$C0,$C8,$F9,$0F,$E5,$BD,$CF,$E4,$1D,$1F
	.dc.b	$EE,$77,$F9,$1B,$7E,$07,$46,$FC,$AE,$1E,$67,$1F,$81,$98,$FD,$AC
	.dc.b	$3F,$A2,$37,$E0,$7B,$11,$D7,$C0,$78,$7E,$06,$47,$17,$C9,$F0,$3E
	.dc.b	$5B,$F3,$7C,$0C,$AE,$D1,$E9,$F4,$BD,$5F,$03,$27,$C5,$1F,$B4,$53
	.dc.b	$EC,$F1,$BF,$A9,$2D,$F4,$D9,$FC,$B9,$C3,$7C,$9A,$99,$1A,$9E,$8F
	.dc.b	$EC,$DA,$F8,$23,$6F,$FB,$C0,$E8,$7C,$7F,$95,$FA,$44,$7E,$57,$C0
	.dc.b	$D7,$FD,$83,$49,$C7,$17,$D7,$AF,$B5,$F0,$52,$FE,$93,$F7,$7E,$0B
	.dc.b	$11,$CA,$2D,$7D,$F9,$FC,$17,$C1,$7E,$0F,$E0,$65,$FD,$B8,$CB,$F7
	.dc.b	$63,$2F,$92,$32,$F9,$A3,$35,$F8,$B9,$FC,$77,$C3,$CA,$FC,$33,$F9
	.dc.b	$1F,$82,$C4,$7B,$A2,$D7,$E4,$E7,$ED,$7C,$AB,$E4,$3F,$B6,$DA,$30
	.dc.b	$FF,$70,$6B,$FD,$EC,$D4,$70,$8C,$D7,$FB,$A5,$BC,$3F,$E1,$56,$D7
	.dc.b	$F2,$DF,$31,$E8,$7E,$2F,$D3,$FF,$38,$CF,$E1,$97,$AD,$93,$5F,$0B
	.dc.b	$27,$51,$97,$84,$65,$FC,$24,$F2,$76,$68,$FD,$67,$C2,$7E,$DC,$8F
	.dc.b	$F2,$8F,$BE,$67,$88,$FC,$3C,$DD,$33,$1F,$E6,$BA,$AF,$E4,$FE,$01
	.dc.b	$E6,$FE,$9D,$B1,$FA,$26,$BF,$99,$F8,$67,$7E,$EA,$4E,$DF,$D8,$4E
	.dc.b	$A3,$F5,$7F,$05,$FD,$25,$BE,$B7,$ED,$63,$FC,$97,$97,$91,$D4,$FE
	.dc.b	$9D,$B1,$99,$E0,$71,$F6,$25,$74,$21,$C5,$FE,$05,$F4,$3F,$36,$3F
	.dc.b	$64,$DA,$F8,$6D,$DB,$FC,$AD,$F3,$1F,$E3,$CA,$FE,$07,$39,$7F,$B8
	.dc.b	$35,$36,$38,$47,$EF,$FE,$EB,$E1,$37,$5F,$E2,$91,$8C,$3E,$9F,$D3
	.dc.b	$7F,$B3,$6B,$F8,$8C,$E7,$1F,$F3,$43,$63,$FC,$00,$D5,$3C,$55,$7E
	.dc.b	$03,$DC,$70,$FF,$AC,$67,$F1,$1F,$13,$FE,$16,$83,$7E,$2A,$4E,$88
	.dc.b	$74,$CB,$FC,$21,$7D,$27,$97,$1F,$3F,$83,$97,$F9,$58,$CB,$F9,$78
	.dc.b	$CB,$CC,$FC,$A4,$78,$FE,$63,$C9,$F1,$72,$7F,$3D,$19,$7F,$41,$19
	.dc.b	$79,$61,$EE,$9F,$67,$C5,$FB,$79,$65,$7E,$9E,$32,$FE,$A6,$32,$F2
	.dc.b	$D5,$3F,$17,$E6,$37,$39,$8C,$3F,$B1,$8C,$EF,$F3,$EB,$8C,$E4,$FE
	.dc.b	$97,$F1,$6F,$80,$FA,$B0,$F9,$77,$A7,$FE,$A5,$3F,$A9,$46,$5C,$79
	.dc.b	$DF,$EA,$99,$DC,$FF,$90,$6F,$A6,$3F,$E4,$39,$CC,$FF,$A4,$70,$EF
	.dc.b	$18,$C3,$AB,$70,$E9,$1F,$E9,$B9,$DB,$6F,$A3,$E2,$BA,$7F,$94,$5E
	.dc.b	$66,$DF,$41,$E5,$9E,$6E,$EB,$BB,$4E,$67,$9E,$AE,$1E,$C8,$E6,$7C
	.dc.b	$69,$E2,$D3,$75,$D9,$E4,$70,$BF,$2B,$DC,$E7,$5D,$70,$F8,$ED,$9F
	.dc.b	$74,$F1,$BE,$F3,$98,$F6,$61,$E8,$AE,$B1,$E3,$3A,$E1,$E3,$AE,$11
	.dc.b	$9C,$3E,$83,$E8,$B8,$F7,$3A,$61,$E0,$BE,$47,$B7,$0E,$E7,$83,$0E
	.dc.b	$A7,$67,$33,$C6,$FA,$E1,$E0,$C3,$B1,$E6,$8C,$D7,$C0,$9D,$1F,$13
	.dc.b	$0F,$0C,$FF,$A3,$FE,$2F,$FB,$1D,$FC,$20,$E1,$FA,$2D,$C7,$ED,$39
	.dc.b	$8F,$F4,$C7,$C6,$73,$F1,$3F,$BB,$36,$3F,$DE,$4D,$7F,$D3,$C1,$C3
	.dc.b	$C8,$FC,$2D,$BE,$13,$FA,$FE,$5F,$A2,$9D,$3F,$A4,$9D,$73,$FE,$42
	.dc.b	$73,$A7,$FA,$FA,$F8,$BE,$48,$EB,$18,$7C,$A6,$27,$93,$E5,$30,$7E
	.dc.b	$60,$EA,$7F,$AB,$A3,$93,$EC,$C3,$F4,$37,$E4,$8E,$07,$3C,$3D,$15
	.dc.b	$C2,$39,$9F,$CF,$9C,$3B,$C3,$D5,$87,$93,$0E,$B5,$A6,$C6,$6B,$CF
	.dc.b	$7F,$1D,$E4,$79,$E3,$51,$E2,$6D,$F6,$9F,$3D,$74,$69,$D2,$3E,$55
	.dc.b	$71,$8C,$B4,$F0,$4F,$16,$C7,$13,$E7,$BB,$9E,$67,$85,$D9,$98,$DC
	.dc.b	$64,$DB,$EB,$3F,$8A,$F1,$FF,$64,$F2,$3F,$C4,$B9,$7F,$39,$F9,$8B
	.dc.b	$97,$F5,$EE,$1F,$F2,$83,$FE,$2F,$8C,$E1,$FE,$2B,$BB,$87,$F6,$E9
	.dc.b	$B7,$C2,$BF,$A6,$46,$0C,$FC,$A5,$C3,$FC,$74,$D9,$FE,$A4,$DC,$7F
	.dc.b	$A6,$3D,$3E,$47,$AF,$C8,$DB,$FF,$21,$36,$FF,$AD,$36,$FF,$AA,$3E
	.dc.b	$4F,$27,$CB,$E5,$5B,$FF,$25,$36,$FF,$B1,$36,$FF,$AE,$3E,$CF,$8F
	.dc.b	$F6,$FC,$7D,$BF,$F2,$93,$6F,$FB,$53,$6F,$FB,$23,$F2,$21,$F9,$5F
	.dc.b	$F0,$8A,$FF,$57,$B6,$3F,$B9,$7C,$86,$E3,$17,$FE,$58,$6D,$7F,$04
	.dc.b	$6D,$7F,$97,$6B,$87,$3F,$7F,$B6,$FF,$54,$5F,$FA,$F1,$B5,$FE,$11
	.dc.b	$C1,$F5,$4F,$2F,$87,$B5,$30,$E0,$FE,$02,$B7,$39,$BE,$5F,$2E,$F0
	.dc.b	$34,$FE,$06,$BF,$1A,$E7,$05,$70,$E8,$DD,$F0,$6C,$EA,$F8,$C6,$2B
	.dc.b	$84,$E4,$D5,$F5,$8C,$9F,$EC,$FF,$98,$7C,$47,$D1,$6D,$7F,$A2,$7E
	.dc.b	$53,$13,$23,$FD,$C3,$3A,$AF,$F1,$25,$65,$FF,$17,$FE,$2F,$F4,$3B
	.dc.b	$0D,$3F,$D4,$F7,$E5,$94,$EB,$87,$F4,$F5,$B6,$F8,$38,$7E,$77,$0B
	.dc.b	$E4,$2F,$C7,$E1,$B9,$75,$E0,$FB,$31,$8C,$FE,$E5,$F1,$AD,$2B,$E5
	.dc.b	$18,$6C,$F0,$7F,$2D,$F3,$2F,$76,$9F,$73,$AA,$DF,$B6,$D5,$B8,$FF
	.dc.b	$68,$46,$2B,$FD,$5E,$FF,$5A,$9D,$3F,$ED,$56,$BF,$D7,$B6,$F9,$7C
	.dc.b	$BF,$18,$F8,$FE,$B5,$C3,$E6,$4E,$9B,$C1,$F9,$96,$CE,$EF,$EB,$FC
	.dc.b	$B7,$D0,$F9,$6F,$A2,$DA,$E2,$D9,$DC,$74,$9D,$B4,$D9,$63,$F0,$7E
	.dc.b	$47,$DC,$3E,$EF,$CB,$59,$F2,$C5,$9F,$C4,$CD,$E2,$FF,$4B,$65,$FE
	.dc.b	$65,$BF,$5F,$97,$F8,$6F,$96,$FC,$3B,$23,$27,$C0,$8C,$61,$F9,$4F
	.dc.b	$73,$F1,$BF,$2D,$F8,$E6,$D7,$C6,$6C,$EE,$3E,$4C,$ED,$A6,$CB,$1F
	.dc.b	$A0,$F9,$6F,$CC,$F3,$FE,$69,$B3,$F4,$B9,$BF,$3B,$F2,$DF,$9E,$6C
	.dc.b	$7D,$6F,$96,$FD,$09,$FA,$2F,$96,$B1,$F6,$BE,$5B,$F4,$A7,$E9,$BC
	.dc.b	$0B,$1F,$7B,$E5,$BF,$A7,$FE,$01,$FC,$2A,$30,$7E,$11,$C6,$BF,$EA
	.dc.b	$45,$F2,$1C,$3D,$7E,$2D,$AE,$1F,$2F,$D0,$EB,$F2,$F6,$3A,$7C,$BF
	.dc.b	$83,$D7,$F0,$B6,$3C,$1F,$2F,$E4,$3C,$BF,$0D,$63,$C9,$F2,$9E,$83
	.dc.b	$D3,$F2,$F6,$3D,$1F,$2D,$EC,$F9,$7F,$6B,$63,$D9,$E9,$7C,$03,$E0
	.dc.b	$FC,$BD,$8F,$81,$F2,$FF,$13,$E5,$FE,$2B,$63,$E2,$47,$C8,$F9,$6C
	.dc.b	$72,$FC,$57,$13,$F2,$7E,$5E,$CF,$C9,$7E,$7F,$9D,$27,$E7,$7C,$BD
	.dc.b	$AF,$9D,$3F,$53,$E5,$31,$E4,$7F,$BC,$7A,$91,$9C,$3E,$B3,$A7,$FD
	.dc.b	$D7,$1A,$7F,$7B,$E3,$1F,$F0,$6A,$F8,$EB,$18,$6B,$C1,$B8,$FF,$B9
	.dc.b	$27,$84,$F4,$38,$14,$E0,$78,$3E,$3F,$BF,$C7,$78,$1B,$19,$3C,$B1
	.dc.b	$96,$9E,$8F,$8F,$F3,$9E,$8F,$8E,$B1,$92,$61,$E5,$8C,$BE,$7F,$8F
	.dc.b	$F8,$0F,$F9,$FF,$D3,$FE,$CF,$EA,$33,$FF,$7A,$E1,$FE,$8A,$9C,$C2
	.dc.b	$F8,$FC,$62,$73,$F1,$D8,$77,$37,$1F,$EA,$B7,$D1,$F0,$18,$8D,$63
	.dc.b	$E1,$BE,$12,$F9,$86,$BF,$AE,$8E,$26,$5C,$65,$E3,$8F,$F0,$0F,$88
	.dc.b	$EC,$E7,$7C,$BC,$5F,$65,$7F,$81,$A3,$0F,$46,$DF,$38,$5F,$3D,$B8
	.dc.b	$C7,$86,$ED,$F0,$DB,$FD,$56,$D9,$F1,$E3,$E2,$CB,$CF,$5A,$6C,$F8
	.dc.b	$DA,$F8,$27,$85,$F6,$9C,$DF,$5A,$D6,$19,$A9,$1F,$EB,$C3,$A3,$FD
	.dc.b	$A2,$7D,$F1,$96,$BF,$EC,$2F,$8F,$FC,$17,$CC,$FF,$6E,$6C,$73,$0E
	.dc.b	$1E,$CD,$7E,$0F,$17,$FB,$61,$E1,$48,$C4,$78,$5A,$F0,$F7,$BC,$D1
	.dc.b	$C7,$6F,$A5,$5C,$67,$E7,$CE,$75,$F9,$FF,$11,$FE,$BF,$79,$61,$F6
	.dc.b	$E7,$4F,$DF,$6B,$F8,$D9,$E0,$E6,$BA,$91,$FC,$24,$69,$FC,$A4,$F2
	.dc.b	$BF,$24,$77,$AF,$C1,$D6,$E7,$38,$7A,$AF,$81,$C6,$7C,$3C,$2F,$97
	.dc.b	$7C,$56,$17,$C0,$53,$11,$33,$FC,$DD,$71,$9D,$46,$2B,$9B,$6F,$FD
	.dc.b	$FF,$D9,$FC,$0A,$FF,$80,$7C,$96,$38,$BF,$7E,$DB,$FF,$26,$E0,$C3
	.dc.b	$17,$E7,$6F,$42,$E9,$AA,$E1,$7F,$E2,$C5,$F0,$9C,$BE,$63,$F0,$79
	.dc.b	$F8,$DB,$9F,$F1,$4A,$DD,$F1,$BE,$1C,$32,$30,$7F,$8E,$2D,$F0,$BD
	.dc.b	$7E,$0E,$4E,$63,$C0,$61,$65,$E1,$9C,$C6,$CE,$8E,$E3,$CD,$3E,$9E
	.dc.b	$49,$5E,$17,$D9,$F0,$72,$30,$43,$BE,$9F,$CC,$31,$4E,$E7,$27,$F9
	.dc.b	$F2,$32,$FB,$63,$E4,$7C,$1F,$C9,$F8,$29,$3F,$05,$F9,$9F,$07,$F3
	.dc.b	$7E,$0A,$CF,$C5,$8C,$3F,$1F,$8A,$BE,$0E,$46,$4B,$FA,$D5,$F5,$FE
	.dc.b	$0A,$5F,$CF,$7E,$D7,$C1,$7D,$BF,$81,$FB,$9F,$05,$2B,$EA,$47,$DF
	.dc.b	$F8,$19,$1F,$5D,$FC,$17,$C1,$7E,$0F,$E0,$A4,$FD,$B7,$F0,$DF,$05
	.dc.b	$F8,$7F,$82,$FC,$47,$C1,$4A,$FC,$04,$7E,$2F,$E0,$A4,$7E,$0D,$FC
	.dc.b	$BF,$B9,$FC,$8F,$AB,$F4,$0F,$F4,$B9,$F8,$0B,$E0,$B6,$70,$F2,$78
	.dc.b	$F9,$9C,$BE,$0A,$5F,$FA,$1A,$FF,$A2,$B7,$E0,$7F,$47,$F1,$32,$30
	.dc.b	$6F,$E0,$BF,$4F,$F1,$32,$30,$73,$BF,$3F,$C1,$48,$C1,$DF,$D2,$F5
	.dc.b	$FC,$14,$8C,$1F,$EA,$7A,$F6,$6F,$FD,$FA,$11,$ED,$75,$BF,$EF,$5C
	.dc.b	$EF,$67,$8F,$87,$FE,$EA,$9E,$35,$93,$55,$23,$99,$0F,$CC,$9D,$0F
	.dc.b	$8D,$F0,$5F,$48,$8F,$C9,$F8,$2A,$FD,$28,$B8,$FD,$69,$FB,$3F,$05
	.dc.b	$2B,$E8,$3F,$73,$E0,$B1,$1F,$6A,$2D,$7D,$E9,$FC,$0F,$C1,$7E,$0B
	.dc.b	$E0,$A5,$FD,$98,$CB,$F6,$E3,$2F,$E1,$23,$2F,$E1,$A3,$35,$F8,$A9
	.dc.b	$FC,$6F,$C0,$FF,$22,$F9,$19,$7F,$85,$8C,$B4,$87,$E4,$3E,$0F,$E9
	.dc.b	$7C,$CB,$97,$EA,$79,$1F,$E1,$1C,$BE,$D3,$78,$BE,$A5,$6F,$0F,$F7
	.dc.b	$33,$5F,$EE,$CD,$7F,$C0,$4D,$78,$DF,$FB,$A9,$FE,$A1,$F1,$FF,$83
	.dc.b	$F5,$3F,$A3,$C7,$DA,$6E,$3F,$8B,$53,$B3,$87,$B1,$2B,$FD,$F1,$3A
	.dc.b	$6C,$E5,$E9,$3C,$5F,$D5,$7C,$2F,$ED,$88,$FF,$BC,$BE,$4F,$F3,$58
	.dc.b	$7E,$E2,$73,$3F,$B8,$7F,$A9,$CE,$B3,$FC,$C7,$C0,$3C,$CF,$CF,$7C
	.dc.b	$CB,$BA,$9F,$A1,$9D,$62,$3F,$32,$E1,$B5,$B3,$9F,$C4,$7E,$A7,$DB
	.dc.b	$7B,$37,$D8,$97,$CF,$6F,$D9,$56,$A3,$1A,$7E,$B6,$31,$1F,$A8,$7E
	.dc.b	$F3,$E3,$6B,$FD,$B6,$FC,$07,$67,$85,$E1,$EA,$7A,$FE,$67,$EC,$DF
	.dc.b	$EE,$FE,$65,$E9,$FF,$01,$6D,$70,$15,$C1,$E5,$C5,$FC,$1E,$31,$1A
	.dc.b	$8E,$5E,$A7,$F0,$BF,$99,$6D,$FC,$37,$E6,$5E,$15,$35,$13,$0F,$05
	.dc.b	$70,$6C,$71,$F1,$FE,$EF,$C5,$FD,$EF,$8B,$95,$F8,$58,$CB,$F8,$78
	.dc.b	$CB,$FD,$6F,$A3,$FD,$78,$FE,$2F,$9B,$F0,$FF,$09,$23,$04,$38,$39
	.dc.b	$79,$47,$E3,$BE,$2F,$F1,$FF,$17,$27,$F2,$D1,$97,$F3,$11,$97,$C5
	.dc.b	$1F,$96,$F8,$BF,$CB,$F4,$49,$FC,$F4,$65,$FD,$04,$65,$F5,$47,$E7
	.dc.b	$BD,$0D,$46,$0F,$D3,$39,$79,$90,$F8,$4F,$F3,$FF,$8A,$FE,$F9,$D7
	.dc.b	$F9,$FF,$99,$7C,$77,$F3,$F3,$FA,$35,$A3,$2F,$E9,$63,$31,$88,$FF
	.dc.b	$41,$FC,$07,$F8,$A9,$E1,$3F,$E8,$B8,$C4,$EE,$38,$9D,$FD,$1F,$E9
	.dc.b	$D1,$CF,$0F,$2D,$F2,$3F,$68,$E2,$39,$4F,$0A,$E5,$5E,$B7,$C1,$85
	.dc.b	$8E,$38,$78,$A7,$71,$D4,$F8,$35,$CD,$DD,$71,$7D,$AD,$9F,$4C,$7C
	.dc.b	$39,$E0,$E8,$E9,$87,$9E,$7B,$E1,$E8,$9E,$93,$97,$B9,$E3,$9E,$AF
	.dc.b	$90,$DE,$1E,$97,$75,$CD,$E1,$7E,$2B,$E9,$1E,$39,$D1,$2F,$E2,$C7
	.dc.b	$9A,$FA,$3D,$CE,$67,$83,$08,$E0,$95,$F5,$E7,$8D,$F8,$2F,$D2,$7A
	.dc.b	$9F,$19,$9C,$3E,$3C,$66,$78,$9E,$27,$8C,$7A,$E3,$38,$7F,$35,$F8
	.dc.b	$7F,$F3,$EE,$3F,$21,$AF,$D1,$6E,$1F,$8A,$73,$1F,$CF,$7E,$2F,$97
	.dc.b	$C4,$7F,$B6,$DB,$1F,$EE,$76,$BF,$E2,$16,$BC,$4F,$C4,$5B,$E1,$5E
	.dc.b	$0F,$E1,$E7,$5A,$FE,$96,$B5,$39,$95,$FE,$36,$73,$1F,$93,$38,$EB
	.dc.b	$FE,$B6,$C3,$AC,$64,$E2,$78,$9C,$E1,$FB,$38,$E2,$7F,$51,$8C,$CF
	.dc.b	$BE,$38,$61,$E3,$C3,$C5,$5A,$9F,$84,$68,$F0,$46,$A7,$84,$66,$FC
	.dc.b	$0F,$91,$F8,$15,$CE,$3E,$51,$BB,$F1,$4E,$AB,$9B,$67,$B9,$B3,$C6
	.dc.b	$E4,$EF,$7E,$38,$DC,$66,$BA,$9F,$26,$BA,$B6,$76,$D9,$E7,$1D,$CF
	.dc.b	$3B,$CD,$F0,$C6,$CE,$AF,$CD,$9E,$AD,$3A,$96,$7F,$94,$FC,$2B,$9B
	.dc.b	$FC,$7D,$5F,$E1,$DF,$43,$EE,$E7,$FE,$F3,$9C,$E1,$FD,$66,$73,$5F
	.dc.b	$E1,$5B,$E5,$E6,$62,$91,$D6,$B0,$EA,$BB,$12,$7F,$B2,$7C,$CE,$B0
	.dc.b	$5D,$8C,$36,$B8,$63,$E5,$9E,$1F,$33,$C7,$E6,$6D,$7F,$8D,$DB,$8F
	.dc.b	$F7,$16,$DF,$B5,$F0,$79,$DE,$1F,$16,$DF,$F8,$FD,$B7,$FD,$E5,$B7
	.dc.b	$F1,$5F,$47,$C7,$7A,$7E,$3A,$DF,$F9,$1D,$B7,$FE,$01,$6D,$FC,$D7
	.dc.b	$E4,$43,$E6,$5C,$D8,$6D,$FF,$94,$7E,$67,$13,$23,$FC,$5B,$C2,$F2
	.dc.b	$5C,$19,$C7,$FD,$D5,$18,$AF,$F1,$67,$CC,$BD,$DF,$0B,$63,$FC,$6F
	.dc.b	$AB,$E5,$FF,$C9,$71,$C2,$B6,$7D,$F9,$7D,$37,$FA,$F2,$74,$72,$71
	.dc.b	$8A,$7F,$1B,$F3,$4F,$07,$93,$F8,$E8,$CC,$63,$A7,$2D,$C3,$CB,$5C
	.dc.b	$5B,$3A,$69,$A9,$E1,$7F,$EC,$2F,$9A,$7C,$5F,$D0,$6D,$7F,$9F,$7E
	.dc.b	$57,$13,$23,$E6,$EE,$EA,$C6,$1F,$31,$FC,$EF,$CC,$FD,$68,$F7,$57
	.dc.b	$E9,$63,$1A,$FF,$28,$E3,$43,$C2,$E3,$CB,$2F,$ED,$CF,$07,$94,$E7
	.dc.b	$0E,$A7,$CB,$E3,$7C,$9F,$D3,$5F,$25,$63,$0D,$BF,$BB,$3A,$6C,$E9
	.dc.b	$A6,$A7,$85,$78,$33,$FC,$24,$7E,$63,$E6,$9F,$07,$FE,$FD,$AF,$F7
	.dc.b	$B4,$62,$FF,$DF,$11,$89,$D3,$FE,$9D,$9D,$3A,$0E,$2F,$F8,$9F,$CD
	.dc.b	$F9,$E7,$FD,$F7,$39,$BF,$F4,$E4,$E6,$BE,$76,$9F,$98,$BF,$8F,$B7
	.dc.b	$D0,$F7,$3E,$BF,$B9,$F6,$19,$8F,$DE,$F0,$ED,$FE,$A1,$FB,$BF,$31
	.dc.b	$F7,$BE,$62,$D7,$D6,$6D,$76,$C9,$18,$CF,$C6,$DA,$FC,$23,$6B,$F9
	.dc.b	$3D,$70,$9F,$C0,$F9,$7F,$8C,$3F,$1B,$F3,$16,$32,$4A,$F0,$46,$5F
	.dc.b	$17,$9D,$F9,$43,$F2,$BD,$16,$BF,$1B,$16,$7D,$13,$72,$F9,$3F,$01
	.dc.b	$F9,$CF,$80,$FC,$EB,$70,$FC,$AB,$6F,$E0,$36,$B8,$CF,$E6,$AB,$8B
	.dc.b	$4F,$CE,$FC,$C7,$EA,$39,$FF,$52,$D8,$FD,$0F,$3F,$EB,$3E,$63,$F5
	.dc.b	$AD,$8F,$D2,$FC,$C7,$F4,$CD,$3F,$E0,$0C,$60,$FD,$BF,$9B,$FB,$4F
	.dc.b	$97,$D7,$9D,$67,$F7,$6F,$EE,$3E,$5F,$FA,$AF,$CD,$FF,$BB,$E7,$F6
	.dc.b	$31,$8E,$1C,$FC,$96,$BD,$8B,$5F,$CC,$2D,$1C,$BE,$62,$4E,$7E,$62
	.dc.b	$E1,$F9,$2B,$7C,$C7,$83,$D9,$F0,$B6,$FA,$7C,$C7,$90,$F2,$FC,$35
	.dc.b	$8F,$07,$CB,$FA,$0F,$4F,$C6,$D8,$F2,$7C,$C7,$B3,$E2,$7D,$AD,$8F
	.dc.b	$47,$AD,$F0,$0F,$83,$F1,$76,$3D,$9F,$31,$F1,$3E,$63,$E2,$B6,$3E
	.dc.b	$04,$7C,$8F,$97,$C7,$0F,$B5,$F9,$7F,$31,$2B,$E3,$45,$9F,$85,$E8
	.dc.b	$7D,$03,$E8,$FC,$C5,$AF,$99,$1F,$53,$E5,$71,$5F,$25,$C4,$FD,$17
	.dc.b	$3E,$2F,$F2,$DE,$64,$7E,$16,$7F,$0D,$F2,$15,$FE,$76,$DB,$FE,$CA
	.dc.b	$DB,$FE,$6D,$E0,$23,$15,$FE,$1F,$8C,$3F,$D1,$0E,$A7,$F4,$6F,$90
	.dc.b	$FC,$A7,$C7,$FE,$55,$B1,$93,$FA,$5C,$65,$A7,$F4,$EF,$90,$FC,$E7
	.dc.b	$C3,$FE,$75,$B1,$93,$FA,$9C,$65,$A7,$F5,$6F,$90,$FF,$3A,$FC,$63
	.dc.b	$D6,$F4,$7C,$76,$1F,$54,$DD,$BF,$C6,$FA,$2B,$1E,$C7,$F8,$2E,$AC
	.dc.b	$E7,$1F,$F4,$C6,$1C,$9B,$87,$2C,$3A,$2F,$9A,$6A,$F1,$B7,$39,$8C
	.dc.b	$3F,$E9,$2C,$0C,$F1,$5D,$BD,$30,$E7,$87,$26,$D7,$13,$8B,$8D,$BE
	.dc.b	$14,$73,$AE,$58,$F8,$27,$33,$E3,$BE,$93,$9C,$7C,$72,$C7,$C7,$9E
	.dc.b	$AE,$4F,$84,$E4,$59,$DB,$9C,$BC,$97,$D1,$B1,$E8,$C3,$C0,$F9,$F0
	.dc.b	$EB,$8F,$9A,$BD,$F3,$FD,$B1,$AF,$3A,$F9,$6E,$5A,$FA,$BE,$3F,$FC
	.dc.b	$8D,$F3,$7F,$DB,$E1,$18,$7F,$B9,$35,$FE,$F2,$D7,$D1,$C5,$FE,$56
	.dc.b	$7F,$BC,$5A,$31,$1E,$56,$BC,$3E,$03,$C1,$3B,$7E,$D4,$F3,$C7,$DC
	.dc.b	$DC,$FE,$37,$C6,$FF,$B8,$5E,$58,$7C,$B9,$D3,$F8,$06,$BF,$91,$9E
	.dc.b	$0E,$6F,$B9,$1F,$A9,$1A,$7F,$31,$3C,$B0,$F3,$47,$7A,$FC,$85,$6E
	.dc.b	$73,$87,$A2,$7A,$3F,$5E,$7C,$3C,$7F,$E7,$56,$F1,$1A,$63,$E3,$2E
	.dc.b	$3F,$AC,$AE,$0D,$AE,$0D,$9E,$7F,$0B,$F9,$6E,$9F,$E3,$2D,$E1,$FE
	.dc.b	$30,$DA,$FF,$5A,$FE,$8F,$F5,$AB,$71,$FE,$5E,$9D,$63,$FE,$8B,$6D
	.dc.b	$6E,$FF,$47,$F8,$13,$C9,$E9,$82,$30,$FE,$EB,$15,$70,$AF,$17,$1B
	.dc.b	$67,$B1,$CB,$1E,$38,$FF,$97,$8E,$97,$93,$8D,$70,$8E,$D8,$79,$3C
	.dc.b	$AF,$CE,$47,$8D,$B6,$69,$23,$89,$FC,$42,$34,$E0,$E4,$6B,$1F,$47
	.dc.b	$AB,$E7,$A9,$3B,$AE,$2F,$F9,$DA,$B6,$F1,$97,$A9,$F0,$9B,$19,$8F
	.dc.b	$85,$E1,$C8,$E8,$D7,$A7,$5F,$F5,$57,$C8,$D7,$E4,$61,$FE,$74,$78
	.dc.b	$9A,$F0,$7E,$EF,$DF,$38,$A9,$27,$CD,$CD,$F4,$FC,$59,$1E,$D7,$EC
	.dc.b	$E3,$FB,$F8,$F0,$F5,$60,$8F,$FB,$53,$97,$FD,$3C,$FF,$B5,$BE,$81
	.dc.b	$E5,$FE,$39,$B1,$98,$C7,$97,$F8,$CE,$EF,$CF,$CF,$E0,$DB,$D3,$F8
	.dc.b	$C8,$FE,$8F,$18,$E6,$FE,$93,$CA,$FB,$D7,$CB,$FE,$73,$A5,$F1,$D4
	.dc.b	$98,$FF,$BA,$A1,$F2,$9F,$A0,$F8,$79,$3E,$B7,$F4,$79,$F9,$7C,$09
	.dc.b	$3F,$87,$84,$7E,$93,$D1,$FA,$53,$FE,$92,$E2,$7C,$37,$EB,$09,$1F
	.dc.b	$1F,$E1,$BF,$5F,$F0,$B2,$3E,$53,$FB,$2F,$85,$FD,$7F,$E0,$5F,$E8
	.dc.b	$E9,$FC,$B3,$5F,$D4,$37,$93,$FD,$A1,$8B,$E1,$BF,$77,$E4,$4B,$FA
	.dc.b	$CF,$EF,$BE,$1B,$F7,$FE,$44,$9F,$B4,$FF,$02,$F8,$5F,$0E,$92,$7C
	.dc.b	$2F,$F0,$83,$E9,$3F,$C3,$BE,$1B,$F8,$49,$1F,$C2,$79,$3F,$C2,$3F
	.dc.b	$17,$7B,$5F,$B8,$6C,$FF,$55,$6B,$FE,$C5,$6B,$FE,$CC,$73,$B7,$E0
	.dc.b	$E1,$18,$ED,$FE,$F2,$D8,$FF,$06,$F9,$6F,$81,$F1,$61,$27,$F8,$05
	.dc.b	$6D,$FE,$11,$3F,$C8,$BE,$1E,$47,$E7,$5F,$E4,$BF,$0F,$88,$FB,$D1
	.dc.b	$72,$FA,$33,$FC,$AB,$E8,$A5,$7E,$95,$FE,$5B,$F1,$18,$8F,$C3,$45
	.dc.b	$AF,$B3,$3F,$CB,$BE,$88,$D2,$57,$F1,$78,$7C,$8E,$23,$F9,$54,$70
	.dc.b	$75,$3C,$E7,$55,$F2,$6B,$55,$C2,$B4,$FF,$23,$FA,$2F,$F1,$DB,$7D
	.dc.b	$9F,$E8,$31,$88,$EF,$C5,$89,$91,$B3,$CD,$1F,$D3,$31,$91,$FC,$1D
	.dc.b	$FE,$9F,$F1,$58,$8F,$D4,$45,$CB,$F3,$73,$E0,$F1,$7F,$9F,$E3,$FB
	.dc.b	$CB,$FF,$0E,$35,$FF,$1C,$9F,$E4,$C7,$0E,$1D,$57,$E2,$5B,$C7,$FE
	.dc.b	$40,$6C,$FF,$95,$34,$FE,$36,$FF,$55,$8C,$A6,$98,$8D,$D7,$F1,$7F
	.dc.b	$5D,$F8,$13,$9F,$FB,$7B,$70,$FE,$A7,$3A,$9C,$F7,$E9,$93,$AA,$E0
	.dc.b	$AF,$B4,$2B,$FB,$AE,$FF,$E5,$E9,$FD,$33,$73,$FF,$2E,$B6,$FF,$CE
	.dc.b	$3F,$44,$F4,$78,$F8,$D2,$7F,$B2,$56,$D5,$6D,$CF,$27,$F7,$29,$34
	.dc.b	$91,$FE,$51,$8C,$BC,$08,$7F,$80,$7E,$7B,$2F,$F8,$0E,$2E,$5E,$AA
	.dc.b	$F6,$9F,$E1,$2A,$CE,$5F,$E7,$8F,$A1,$7C,$27,$F8,$59,$B8,$FF,$7D
	.dc.b	$6B,$FE,$0F,$F9,$27,$C5,$6E,$A4,$ED,$FF,$0F,$CE,$A3,$FC,$0D,$F8
	.dc.b	$1F,$DE,$C3,$FD,$0E,$D8,$FF,$4D,$B5,$7D,$97,$99,$CF,$CB,$FF,$8A
	.dc.b	$A7,$46,$E7,$5C,$AF,$D2,$3F,$37,$BF,$E2,$DB,$8F,$FA,$A2,$75,$8F
	.dc.b	$F8,$EA,$7F,$D0,$75,$B8,$C1,$0F,$C7,$C6,$AB,$92,$F9,$3E,$44,$9D
	.dc.b	$FC,$3D,$7F,$21,$5C,$27,$17,$E5,$AF,$F2,$5F,$AA,$F4,$E9,$30,$FC
	.dc.b	$FC,$7E,$47,$2F,$EE,$F3,$99,$EA,$23,$2F,$0E,$6F,$77,$89,$23,$D6
	.dc.b	$FC,$1F,$92,$C4,$72,$AF,$95,$38,$CB,$E5,$56,$EB,$E1,$B5,$E1,$AB
	.dc.b	$E4,$F1,$5C,$27,$67,$CA,$8F,$F3,$1F,$C4,$FF,$6F,$9F,$F6,$DF,$D1
	.dc.b	$71,$AF,$F6,$ED,$7C,$D7,$2F,$FB,$38,$C2,$8E,$0F,$FA,$DB,$D1,$FD
	.dc.b	$F3,$99,$FF,$5E,$46,$27,$71,$C9,$BC,$1F,$ED,$A7,$9E,$1E,$5B,$E4
	.dc.b	$7F,$BC,$1C,$72,$70,$8F,$F7,$9B,$53,$CB,$0F,$1B,$71,$F5,$CF,$BB
	.dc.b	$0F,$06,$3E,$D6,$C7,$7B,$F4,$4F,$AA,$B8,$CF,$CF,$AD,$B4,$CE,$1E
	.dc.b	$9C,$7C,$11,$E4,$BE,$CF,$A1,$AE,$4F,$05,$E5,$A7,$C2,$3B,$16,$3E
	.dc.b	$51,$E5,$75,$5E,$18,$EB,$5F,$14,$F1,$91,$C9,$27,$BC,$65,$F8,$C7
	.dc.b	$C0,$8E,$EE,$9F,$04,$75,$79,$47,$E0,$EB,$E1,$E1,$F2,$A3,$99,$E9
	.dc.b	$C3,$E4,$E1,$E7,$8E,$86,$B0,$E8,$E0,$E8,$71,$3C,$93,$CA,$BB,$46
	.dc.b	$BE,$27,$FB,$1D,$FE,$2D,$B3,$F9,$66,$E0,$A4,$7F,$6A,$71,$ED,$66
	.dc.b	$31,$1F,$EC,$53,$79,$FF,$7A,$71,$1A,$9E,$13,$97,$F1,$35,$FD,$B7
	.dc.b	$93,$F1,$77,$D8,$FB,$D1,$97,$F0,$06,$73,$FD,$7C,$65,$FD,$8C,$67
	.dc.b	$4F,$F7,$77,$85,$FE,$91,$9C,$CF,$9D,$E3,$19,$9F,$24,$F8,$23,$0F
	.dc.b	$F8,$53,$0F,$3C,$62,$3E,$31,$CE,$7B,$E1,$EA,$8C,$1F,$EE,$D9,$E0
	.dc.b	$F9,$4C,$46,$EB,$AB,$63,$C4,$DB,$F3,$B6,$FB,$CF,$C4,$3B,$9F,$E0
	.dc.b	$DB,$EE,$E3,$0F,$8D,$3E,$EC,$79,$CF,$EB,$CF,$5E,$3F,$AF,$79,$61
	.dc.b	$5F,$41,$83,$C1,$1E,$86,$9E,$82,$E3,$D8,$FA,$31,$D7,$0F,$B9,$78
	.dc.b	$AF,$14,$63,$0F,$A7,$7C,$5A,$7C,$07,$81,$99,$FA,$C7,$85,$FE,$75
	.dc.b	$C3,$FD,$CF,$E5,$3F,$DA,$57,$FD,$CD,$B9,$7F,$6A,$F8,$AB,$87,$F5
	.dc.b	$9E,$15,$63,$0E,$57,$FD,$CA,$31,$2B,$B1,$8F,$42,$56,$CF,$EF,$2E
	.dc.b	$30,$FE,$AE,$F0,$8C,$9F,$DB,$BD,$1F,$07,$41,$BE,$7E,$19,$3B,$3F
	.dc.b	$BA,$C6,$2F,$FB,$1F,$CA,$79,$8F,$3F,$C6,$DA,$FF,$05,$C5,$9F,$ED
	.dc.b	$33,$76,$FF,$0C,$78,$3E,$E3,$DF,$EE,$5C,$3F,$C2,$71,$6F,$FB,$8C
	.dc.b	$DA,$FF,$5F,$74,$FC,$AB,$FE,$19,$18,$DB,$E6,$7E,$06,$EB,$FA,$38
	.dc.b	$FF,$86,$E6,$E1,$C6,$3F,$C1,$A6,$70,$FF,$38,$DF,$17,$A9,$FB,$49
	.dc.b	$E1,$87,$FB,$3F,$B9,$C1,$D9,$B5,$C3,$0F,$F3,$A7,$CF,$61,$EF,$36
	.dc.b	$F9,$5F,$13,$F0,$EA,$B5,$7F,$C1,$2A,$DF,$F2,$BF,$9D,$7B,$1E,$E6
	.dc.b	$D7,$87,$0E,$5F,$03,$9A,$93,$E5,$B5,$F0,$7F,$C5,$F7,$C2,$75,$39
	.dc.b	$8F,$F7,$73,$B9,$FF,$56,$B5,$FE,$32,$D7,$CD,$5F,$EF,$57,$F8,$EF
	.dc.b	$CD,$FC,$D7,$E8,$B6,$BF,$57,$F3,$9D,$F0,$DF,$07,$DA,$AD,$46,$0F
	.dc.b	$74,$EE,$FF,$A7,$4E,$9F,$F2,$46,$EE,$4F,$C2,$9F,$86,$E4,$B8,$7E
	.dc.b	$02,$2D,$7F,$91,$E9,$58,$C9,$7E,$AF,$90,$F2,$1F,$63,$81,$F2,$B8
	.dc.b	$9F,$9D,$9A,$B1,$97,$CE,$BD,$D6,$A7,$33,$FE,$D8,$AE,$13,$FD,$86
	.dc.b	$11,$87,$FB,$23,$5F,$E9,$3D,$9F,$E8,$7F,$89,$FE,$16,$7F,$0C,$F8
	.dc.b	$9B,$7F,$E8,$08,$B5,$FE,$5E,$9B,$9F,$BA,$3F,$BC,$7D,$83,$E3,$3F
	.dc.b	$56,$DC,$3F,$DA,$FF,$63,$FE,$AE,$7F,$19,$19,$AF,$D6,$B6,$FA,$42
	.dc.b	$31,$3F,$E9,$9A,$E5,$3C,$2F,$E6,$7C,$7C,$AF,$CE,$6C,$F8,$BC,$5F
	.dc.b	$CB,$AB,$97,$F3,$4A,$D5,$F0,$9C,$BF,$C2,$21,$1F,$66,$F3,$7F,$CD
	.dc.b	$FE,$6F,$F9,$99,$FC,$D3,$E6,$EE,$1F,$EA,$88,$B3,$FE,$86,$9B,$8F
	.dc.b	$FA,$D7,$E1,$BF,$9D,$9F,$CF,$38,$AE,$1F,$EB,$08,$B7,$FE,$8E,$9B
	.dc.b	$5F,$1B,$67,$5F,$D0,$3F,$82,$F4,$7F,$BB,$31,$F9,$F1,$96,$9F,$A4
	.dc.b	$F6,$31,$8F,$D2,$9F,$D9,$3F,$71,$AF,$D3,$6D,$F8,$2B,$FC,$61,$C9
	.dc.b	$FE,$D1,$F6,$FF,$AA,$9F,$D5,$BE,$6A,$CF,$F0,$BF,$C0,$FF,$D9,$0D
	.dc.b	$F0,$BF,$99,$D7,$F3,$29,$CE,$5F,$EB,$3F,$D2,$7F,$57,$7F,$C9,$F1
	.dc.b	$46,$27,$5F,$2F,$FD,$9E,$31,$BF,$E5,$6B,$7E,$DE,$4D,$7B,$D7,$1F
	.dc.b	$27,$B3,$E1,$39,$75,$58,$F0,$63,$FC,$B3,$0F,$0F,$8F,$DF,$E7,$3C
	.dc.b	$0F,$87,$E7,$25,$F0,$F8,$0B,$BF,$F4,$AE,$0F,$E1,$D3,$F0,$FC,$6F
	.dc.b	$EA,$3F,$83,$FF,$98,$23,$07,$83,$E8,$FD,$87,$B7,$E2,$2C,$7C,$3F
	.dc.b	$63,$DE,$FF,$BF,$5B,$8F,$FB,$8D,$72,$FF,$82,$DC,$C3,$2F,$E3,$35
	.dc.b	$F5,$7C,$9F,$1F,$BF,$F4,$DF,$F0,$97,$D2,$BE,$02,$98,$F3,$AD,$FF
	.dc.b	$5B,$F9,$80,$F9,$7F,$F0,$EB,$87,$EB,$78,$38,$8D,$4F,$F4,$FF,$C1
	.dc.b	$5C,$3F,$DD,$1B,$1D,$67,$7B,$7F,$74,$9C,$BF,$EF,$9E,$7F,$E7,$DD
	.dc.b	$07,$FA,$05,$3E,$09,$EA,$FF,$8B,$67,$32,$F8,$9F,$07,$37,$95,$B9
	.dc.b	$F9,$62,$E4,$F0,$7F,$91,$5F,$C0,$BC,$C6,$32,$7C,$C7,$C2,$6E,$3C
	.dc.b	$CF,$90,$E3,$0F,$67,$CE,$7C,$13,$E4,$DB,$CD,$C5,$FC,$79,$B5,$E3
	.dc.b	$8F,$F5,$56,$5E,$5A,$CD,$FF,$2C,$DD,$87,$EF,$1B,$7F,$D2,$5B,$87
	.dc.b	$FA,$AD,$AF,$FA,$40,$7E,$57,$F9,$77,$FC,$4A,$DB,$F8,$67,$F5,$13
	.dc.b	$F9,$DB,$9A,$FF,$4C,$B6,$7F,$D8,$0D,$C7,$E6,$79,$9F,$E3,$48,$D7
	.dc.b	$0F,$F0,$0F,$A8,$FD,$F1,$FB,$FF,$A7,$B8,$FE,$DA,$2E,$3F,$C2,$66
	.dc.b	$E5,$FC,$0F,$C4,$7D,$3F,$F5,$9F,$87,$77,$7F,$B4,$73,$FE,$1B,$39
	.dc.b	$8C,$53,$E8,$F1,$92,$F7,$A7,$FB,$5A,$F7,$86,$A7,$34,$C7,$BB,$C7
	.dc.b	$1E,$98,$71,$8C,$5F,$65,$F4,$8F,$2D,$F4,$5B,$9C,$C6,$1E,$D9,$FF
	.dc.b	$B1,$BC,$CF,$DE,$C6,$0F,$F7,$75,$F3,$BE,$4D,$8D,$1D,$23,$B4,$F2
	.dc.b	$CF,$D3,$39,$37,$5C,$63,$2D,$BF,$17,$87,$EB,$8C,$57,$8B,$1F,$45
	.dc.b	$F0,$6D,$F0,$CF,$CB,$18,$AE,$19,$60,$F2,$E3,$D6,$75,$3E,$53,$F3
	.dc.b	$18,$FB,$5B,$1F,$1D,$B5,$FE,$29,$F8,$D7,$CF,$7F,$1A,$9F,$D0,$08
	.dc.b	$CF,$36,$E7,$53,$FA,$59,$CB,$C3,$0F,$F1,$27,$C7,$7F,$8D,$5B,$19
	.dc.b	$F3,$D3,$96,$C6,$63,$EA,$C6,$36,$F9,$AD,$D9,$5F,$73,$DA,$FF,$75
	.dc.b	$C6,$23,$F1,$33,$B7,$F0,$2D,$73,$7F,$82,$AE,$31,$A8,$F0,$91,$FB
	.dc.b	$F1,$A7,$89,$F8,$29,$E9,$87,$BB,$0F,$C1,$D6,$E3,$18,$79,$5B,$7D
	.dc.b	$CE,$A7,$E8,$F8,$FE,$9D,$79,$1B,$19,$CF,$FC,$6A,$DF,$6B,$51,$88
	.dc.b	$FD,$04,$F0,$6C,$67,$1F,$D9,$CF,$E0,$2B,$8F,$67,$C9,$E7,$FE,$11
	.dc.b	$F4,$D7,$93,$F8,$1B,$72,$FE,$05,$97,$F2,$C7,$FC,$2C,$D7,$F9,$A1
	.dc.b	$AB,$E1,$8F,$E3,$9B,$F4,$1F,$80,$9F,$F6,$7F,$D2,$FF,$BF,$5F,$F3
	.dc.b	$E3,$5F,$C9,$B5,$FE,$68,$D7,$F0,$0E,$F2,$FF,$35,$B7,$2C,$FC,$6F
	.dc.b	$A7,$93,$F8,$A7,$D0,$7A,$F0,$5B,$E3,$B5,$F3,$2F,$8B,$A6,$E7,$55
	.dc.b	$FE,$C7,$6B,$E6,$AF,$F4,$A3,$D1,$B1,$A8,$F4,$C7,$32,$3F,$E0,$BE
	.dc.b	$27,$C7,$F9,$E7,$F9,$94,$FF,$83,$DA,$FC,$EC,$BD,$1E,$07,$97,$C4
	.dc.b	$93,$F2,$9F,$09,$1F,$DE,$CF,$F2,$CF,$57,$71,$F4,$A7,$E3,$76,$FF
	.dc.b	$8A,$A3,$FC,$A5,$F4,$BF,$77,$8B,$FC,$9C,$DE,$2F,$E9,$4D,$9F,$F1
	.dc.b	$1E,$7F,$93,$AF,$A7,$EE,$FC,$22,$4F,$DE,$7F,$D7,$59,$FF,$3E,$ED
	.dc.b	$F0,$92,$7F,$CC,$5C,$AF,$5F,$F1,$7F,$4A,$F8,$9A,$4A,$FF,$34,$43
	.dc.b	$E2,$7E,$F7,$83,$FA,$59,$FE,$F8,$D7,$F8,$43,$5F,$E1,$AD,$EC,$FE
	.dc.b	$B5,$E8,$FE,$6B,$9C,$D3,$1E,$AF,$E7,$3D,$49,$87,$A1,$FC,$FE,$1E
	.dc.b	$2F,$0B,$F0,$AF,$FA,$DE,$7F,$DB,$4D,$7F,$98,$B5,$FF,$6C,$ED,$FE
	.dc.b	$4D,$7E,$1F,$AD,$FA,$A3,$F5,$7E,$9F,$EB,$3E,$57,$F5,$FE,$77,$FB
	.dc.b	$C2,$BF,$CF,$7F,$19,$97,$F9,$4F,$3F,$FB,$43,$E9,$3F,$C6,$F1,$FE
	.dc.b	$EB,$6B,$FD,$09,$AF,$F8,$8D,$AF,$F8,$27,$2F,$E9,$2F,$D6,$F1,$FF
	.dc.b	$7A,$7E,$FB,$E9,$3F,$7F,$E3,$79,$3C,$A9,$5F,$72,$3F,$65,$83,$4F
	.dc.b	$D4,$B6,$7C,$91,$83,$FE,$00,$72,$B2,$7D,$0F,$F0,$C2,$57,$E2,$3C
	.dc.b	$5F,$DF,$7D,$23,$E3,$3F,$80,$78,$2F,$99,$A6,$A2,$5F,$F6,$21,$5C
	.dc.b	$5C,$35,$FE,$CE,$D7,$FC,$90,$E7,$87,$FD,$FA,$DF,$13,$FA,$B3,$67
	.dc.b	$FA,$DE,$1F,$E8,$F7,$F3,$FF,$2F,$FC,$98,$8F,$E8,$BE,$92,$BF,$CB
	.dc.b	$B6,$F8,$53,$8C,$7F,$D8,$5B,$2F,$F9,$7F,$C7,$4C,$3F,$56,$FF,$32
	.dc.b	$FA,$4C,$47,$3A,$4E,$2F,$9D,$7F,$27,$F3,$CD,$25,$FE,$33,$5F,$EE
	.dc.b	$52,$FA,$4C,$CF,$F4,$41,$1C,$9E,$07,$F7,$C7,$0E,$DE,$4E,$8E,$35
	.dc.b	$D2,$D7,$E0,$8F,$35,$FF,$3D,$FA,$3F,$F2,$55,$F1,$C1,$A7,$AC,$D7
	.dc.b	$B1,$27,$B5,$70,$7D,$0D,$CB,$BC,$79,$F9,$3F,$AD,$E3,$23,$F8,$BB
	.dc.b	$FD,$7B,$E9,$31,$1F,$36,$2D,$7D,$69,$F3,$78,$9F,$37,$C1,$97,$E2
	.dc.b	$8C,$BF,$93,$C7,$F8,$8B,$63,$FA,$7C,$62,$3F,$C7,$2D,$BF,$3C,$23
	.dc.b	$1B,$7F,$89,$DB,$1F,$47,$C9,$FE,$EF,$F4,$9F,$B1,$8F,$C7,$36,$FE
	.dc.b	$0F,$22,$31,$87,$FA,$2D,$FF,$66,$57,$DC,$9D,$75,$74,$24,$66,$76
	.dc.b	$FE,$17,$7F,$F3,$7E,$1E,$8F,$AE,$F4,$B7,$1F,$EF,$13,$A7,$FC,$80
	.dc.b	$DA,$FF,$22,$B5,$FF,$45,$6B,$FD,$FF,$E9,$5E,$7F,$2E,$6F,$F0,$17
	.dc.b	$C7,$FE,$3A,$FF,$2C,$DC,$BE,$4E,$3F,$A4,$9F,$C9,$35,$FF,$33,$B5
	.dc.b	$FF,$64,$35,$F2,$D7,$F8,$12,$38,$4F,$48,$F1,$4F,$E7,$6B,$39,$7F
	.dc.b	$96,$FE,$95,$F2,$7F,$E1,$7F,$A5,$FE,$C1,$7F,$A3,$6B,$FA,$D9,$DB
	.dc.b	$FA,$76,$B9,$9F,$F1,$04,$63,$1F,$C2,$D3,$E8,$7F,$C4,$9F,$4B,$FD
	.dc.b	$BF,$0F,$D8,$7D,$2F,$C3,$CB,$FC,$DC,$D8,$FF,$39,$CE,$DF,$DF,$37
	.dc.b	$0F,$C8,$B6,$BF,$CE,$CD,$9F,$F3,$DD,$7C,$07,$FD,$9D,$F4,$9F,$E2
	.dc.b	$AE,$53,$49,$1F,$D0,$9F,$F7,$B6,$1F,$94,$6E,$1F,$ED,$D6,$CF,$4F
	.dc.b	$67,$FC,$A7,$0F,$99,$FE,$CF,$E8,$CA,$FC,$C4,$E5,$B3,$C0,$8E,$BE
	.dc.b	$1F,$FC,$BB,$F3,$2F,$53,$FA,$AB,$67,$FD,$57,$A2,$31,$1F,$E9,$73
	.dc.b	$FD,$D8,$E1,$FF,$4D,$46,$E7,$A9,$23,$F5,$93,$F9,$F9,$F8,$BF,$29
	.dc.b	$F0,$C9,$3E,$0C,$7F,$A6,$3D,$E7,$6F,$BA,$7E,$19,$F1,$3E,$52,$46
	.dc.b	$48,$7F,$4F,$A4,$6A,$30,$DD,$BE,$97,$99,$FE,$6D,$F8,$97,$C8,$A7
	.dc.b	$FD,$73,$E0,$FF,$71,$C3,$FD,$77,$7F,$3D,$FF,$1F,$3F,$EB,$B3,$11
	.dc.b	$B8,$FF,$6A,$57,$37,$67,$FB,$12,$FB,$DF,$FB,$1B,$E9,$78,$61,$D5
	.dc.b	$BC,$1F,$EC,$77,$A5,$70,$B4,$63,$0E,$CF,$08,$E9,$7E,$5B,$E4,$FF
	.dc.b	$B4,$6F,$95,$71,$C3,$DF,$3B,$BA,$7C,$0B,$FF,$6C,$63,$B8,$F5,$9E
	.dc.b	$EA,$EE,$EC,$A7,$F8,$66,$7C,$2F,$C1,$C7,$CF,$7E,$36,$C7,$13,$DB
	.dc.b	$1E,$36,$9C,$6B,$AE,$3F,$0F,$0F,$81,$3E,$4A,$FA,$67,$5A,$F5,$C7
	.dc.b	$9C,$E6,$79,$4F,$3C,$7D,$13,$83,$AA,$F8,$F1,$E3,$C7,$B9,$F0,$C8
	.dc.b	$E4,$93,$E7,$39,$61,$F8,$3C,$3C,$E4,$9E,$EF,$29,$F8,$55,$DA,$7C
	.dc.b	$6F,$D1,$9E,$13,$F1,$DC,$4F,$23,$E4,$3A,$8F,$80,$FD,$9F,$8B,$FE
	.dc.b	$45,$87,$E2,$DB,$87,$E5,$DB,$8A,$D1,$FC,$A5,$C7,$B9,$98,$C4,$23
	.dc.b	$18,$7F,$09,$8C,$1C,$9C,$46,$A3,$F8,$73,$F5,$AB,$FC,$DD,$CB,$F6
	.dc.b	$27,$B9,$F9,$08,$CB,$F9,$23,$35,$FB,$18,$CB,$FB,$38,$CE,$5E,$B7
	.dc.b	$F7,$DE,$27,$F8,$9A,$75,$1E,$6D,$39,$4F,$F8,$A9,$CE,$3F,$9C,$C3
	.dc.b	$C3,$18,$AF,$25,$73,$C3,$DB,$18,$C7,$D4,$F5,$8F,$15,$F2,$77,$7E
	.dc.b	$0A,$FD,$4D,$7C,$08,$C3,$F9,$83,$13,$CA,$FC,$CF,$23,$15,$CA,$7C
	.dc.b	$07,$FA,$B2,$76,$FD,$03,$CC,$7C,$B3,$EB,$E3,$E0,$3D,$D7,$E2,$2D
	.dc.b	$F6,$9F,$21,$D8,$F9,$B7,$E1,$C7,$E4,$61,$E0,$9F,$A5,$1F,$7E,$3E
	.dc.b	$44,$78,$F0,$F0,$5F,$AE,$3D,$17,$CD,$FD,$CE,$3F,$25,$F5,$63,$E5
	.dc.b	$AE,$11,$E2,$3C,$B8,$F7,$BF,$A6,$D8,$FE,$6B,$C3,$FD,$EB,$E1,$FF
	.dc.b	$CD,$38,$FF,$77,$6D,$FF,$77,$F8,$8B,$7F,$E0,$29,$B7,$C6,$BF,$BA
	.dc.b	$FA,$38,$E3,$55,$89,$3A,$3F,$BD,$C6,$2F,$FA,$D4,$FF,$99,$1F,$F0
	.dc.b	$33,$5F,$F0,$8B,$6F,$FC,$22,$DE,$57,$A9,$E4,$6E,$1F,$DD,$3C,$EC
	.dc.b	$4C,$8F,$F0,$0C,$EB,$0F,$EE,$4D,$7F,$BF,$B5,$FF,$0C,$7C,$CF,$B4
	.dc.b	$F7,$7C,$6D,$AF,$F0,$AC,$7C,$4F,$75,$F1,$79,$BF,$F6,$24,$62,$BF
	.dc.b	$C3,$7E,$17,$C7,$3E,$47,$D2,$D8,$C9,$FE,$07,$8F,$81,$C1,$FE,$43
	.dc.b	$8C,$1F,$EE,$3E,$BF,$9E,$FF,$AC,$DB,$87,$FB,$4F,$E8,$FF,$C0,$B3
	.dc.b	$9E,$2F,$B7,$3F,$E2,$F7,$ED,$FC,$4B,$E3,$B3,$87,$2E,$6F,$F4,$6C
	.dc.b	$E7,$0F,$F6,$07,$83,$FE,$29,$6F,$17,$E1,$A9,$F4,$39,$CF,$FD,$9D
	.dc.b	$76,$7F,$D4,$2D,$BF,$E6,$11,$C3,$E6,$3C,$B3,$9A,$FF,$0B,$5A,$7D
	.dc.b	$3F,$4A,$FA,$5A,$78,$65,$F1,$38,$CB,$FD,$AD,$5B,$7F,$9B,$60,$8C
	.dc.b	$3F,$CE,$5A,$FF,$90,$DA,$FF,$2B,$77,$97,$EA,$2D,$F2,$DD,$EB,$FC
	.dc.b	$59,$7E,$C9,$CB,$FE,$E7,$FA,$5F,$A7,$3F,$EF,$16,$DF,$EC,$BE,$8F
	.dc.b	$AD,$7F,$39,$9C,$B7,$C8,$F9,$DF,$52,$EC,$FB,$75,$AC,$7F,$C7,$93
	.dc.b	$9B,$FF,$60,$73,$3B,$3F,$72,$D8,$CE,$18,$99,$1C,$1D,$C6,$2F,$D1
	.dc.b	$F2,$7E,$03,$E8,$7C,$65,$BF,$74,$7E,$E6,$59,$23,$0D,$9F,$A6,$D9
	.dc.b	$FE,$D4,$DD,$7F,$8A,$FA,$5F,$B7,$7F,$DD,$31,$9A,$FC,$4C,$E5,$FF
	.dc.b	$47,$7C,$8F,$F1,$33,$F8,$A7,$C8,$58,$FF,$45,$C5,$C3,$FC,$D5,$37
	.dc.b	$0F,$F7,$AF,$D2,$3E,$4F,$F4,$ED,$C3,$FD,$D5,$F4,$DF,$EC,$27,$F3
	.dc.b	$27,$F6,$F8,$D6,$1F,$82,$3F,$C1,$8E,$2F,$FD,$EF,$4F,$73,$AE,$1F
	.dc.b	$CB,$23,$18,$74,$C3,$F4,$3F,$1F,$F9,$D3,$F3,$DE,$B5,$9D,$63,$FB
	.dc.b	$09,$C5,$78,$63,$F6,$75,$87,$F3,$13,$6B,$F2,$D3,$6B,$FD,$7F,$F3
	.dc.b	$DF,$CE,$CF,$E7,$9F,$4D,$6B,$FD,$67,$16,$7F,$D2,$53,$70,$FF,$60
	.dc.b	$7C,$7F,$F4,$33,$FA,$27,$15,$C3,$FD,$77,$16,$FF,$D3,$53,$6B,$E8
	.dc.b	$F6,$FD,$03,$F8,$67,$53,$E0,$31,$8F,$C5,$C1,$C3,$FC,$02,$3F,$7A
	.dc.b	$D8,$FF,$1A,$35,$FF,$5F,$35,$FB,$17,$FA,$CF,$23,$AC,$7C,$9E,$BF
	.dc.b	$E6,$0F,$FA,$BF,$E0,$7F,$AF,$1F,$D7,$FE,$9A,$E3,$E3,$FA,$4F,$F6
	.dc.b	$D3,$FC,$39,$B5,$FE,$45,$6B,$FE,$CF,$6B,$FE,$51,$F4,$9E,$72,$BF
	.dc.b	$B7,$FD,$1F,$F9,$51,$FC,$24,$E9,$CF,$B5,$FD,$47,$E0,$E4,$6A,$FF
	.dc.b	$17,$87,$F5,$29,$FE,$83,$8B,$DF,$FD,$B3,$6F,$D1,$18,$3F,$B8,$5F
	.dc.b	$F3,$3C,$59,$3D,$BF,$EC,$1E,$6F,$99,$FE,$C5,$F4,$D3,$1F,$17,$C2
	.dc.b	$DD,$7F,$48,$34,$FE,$24,$FC,$5F,$1F,$C6,$4A,$FF,$7B,$E9,$E0,$7D
	.dc.b	$99,$FF,$6F,$F8,$BF,$EE,$0F,$F8,$53,$3F,$91,$3F,$19,$B8,$7E,$CF
	.dc.b	$5F,$47,$CE,$6E,$BE,$1F,$93,$F0,$88,$FD,$97,$FC,$21,$F4,$AF,$84
	.dc.b	$FF,$10,$B6,$7E,$9B,$FC,$C6,$CE,$70,$FE,$05,$F3,$A1,$F3,$9F,$E2
	.dc.b	$C7,$0F,$F8,$27,$AF,$11,$AC,$3F,$3D,$F4,$9F,$06,$FF,$D1,$CD,$7F
	.dc.b	$CF,$C1,$C3,$FE,$81,$6A,$C7,$FB,$57,$A9,$FE,$3A,$F5,$7F,$A7,$CF
	.dc.b	$97,$E1,$9E,$BE,$6B,$C7,$F4,$BF,$E0,$8C,$5C,$98,$CF,$CA,$E4,$FE
	.dc.b	$81,$1A,$8F,$ED,$ED,$C7,$D3,$3F,$E3,$DD,$19,$3E,$7B,$E2,$37,$0E
	.dc.b	$A7,$F2,$39,$DC,$F8,$FE,$02,$4E,$7C,$77,$97,$8C,$BF,$1F,$9F,$F8
	.dc.b	$C5,$F9,$37,$E6,$25,$F9,$9F,$A9,$F5,$CF,$4F,$96,$3F,$B3,$BF,$B5
	.dc.b	$39,$BF,$E2,$5F,$4D,$FB,$59,$FC,$EB,$6B,$86,$87,$FA,$C4,$FF,$B3
	.dc.b	$E7,$55,$FE,$B7,$79,$47,$DC,$6B,$9F,$13,$FB,$64,$62,$3E,$E7,$6F
	.dc.b	$F8,$39,$FF,$27,$FB,$7F,$E5,$3F,$A6,$FF,$2A,$33,$0F,$E1,$3D,$FF
	.dc.b	$79,$FE,$0D,$1F,$C3,$B9,$BF,$CB,$5F,$4D,$EF,$F8,$97,$C9,$FF,$15
	.dc.b	$CD,$39,$8C,$7D,$25,$D3,$C7,$37,$B3,$F8,$DB,$6F,$8E,$7F,$EE,$CC
	.dc.b	$38,$B6,$F8,$B5,$7D,$37,$F6,$7F,$91,$C3,$63,$FD,$E5,$5F,$ED,$FF
	.dc.b	$2F,$FC,$41,$18,$3F,$DF,$93,$BD,$3D,$71,$8F,$6D,$87,$46,$C7,$13
	.dc.b	$F9,$37,$8B,$ED,$7C,$71,$AA,$E7,$18,$AE,$47,$A7,$0F,$1C,$EB,$0F
	.dc.b	$16,$3F,$EF,$56,$DF,$8E,$F8,$63,$DE,$73,$3D,$4F,$23,$8C,$3A,$DF
	.dc.b	$3C,$7C,$8D,$C3,$FC,$BF,$5F,$1E,$FF,$83,$B5,$D1,$FC,$28,$7C,$8B
	.dc.b	$E9,$3F,$8C,$46,$63,$1D,$1F,$E6,$A6,$C6,$67,$6F,$6F,$75,$D1,$E2
	.dc.b	$6C,$77,$C3,$FC,$5D,$39,$24,$7B,$A7,$7B,$7C,$96,$E7,$F2,$7D,$CF
	.dc.b	$F7,$6B,$F2,$E3,$35,$F8,$06,$BF,$8A,$6E,$0A,$E3,$3B,$C3,$F1,$B5
	.dc.b	$A3,$73,$A9,$E5,$3A,$23,$F5,$23,$55,$EC,$8F,$54,$EB,$0F,$C1,$57
	.dc.b	$83,$8F,$E7,$37,$2F,$25,$FE,$8E,$72,$D8,$7B,$9B,$A9,$39,$79,$47
	.dc.b	$3E,$E4,$FF,$80,$FE,$22,$BF,$ED,$46,$F6,$FF,$1D,$C7,$F0,$15,$8E
	.dc.b	$5F,$F0,$A6,$5F,$ED,$CC,$39,$10,$FE,$D8,$E5,$FE,$FB,$F4,$CF,$90
	.dc.b	$E5,$1F,$D8,$7E,$7B,$FB,$53,$FC,$ED,$AF,$08,$DA,$9E,$4F,$FA,$E1
	.dc.b	$AF,$E5,$DA,$FF,$62,$6B,$EE,$5F,$23,$E2,$E5,$5B,$E3,$70,$DB,$FF
	.dc.b	$26,$5F,$06,$DE,$AA,$46,$08,$7C,$08,$F4,$CF,$18,$FF,$00,$F0,$BE
	.dc.b	$43,$C5,$4E,$EF,$8B,$BF,$F8,$B3,$9E,$F8,$3F,$C3,$38,$BF,$A4,$DE
	.dc.b	$F2,$F2,$FC,$0F,$2C,$64,$60,$87,$D2,$8F,$19,$F0,$7A,$DE,$9F,$83
	.dc.b	$E3,$7E,$C6,$3F,$81,$CB,$FC,$2B,$87,$F4,$6F,$4B,$5C,$FE,$CF,$77
	.dc.b	$D3,$D7,$2B,$E3,$C7,$5E,$ED,$46,$0F,$EF,$1A,$7F,$AA,$DF,$F2,$CF
	.dc.b	$D3,$7F,$6D,$7F,$88,$71,$23,$11,$FE,$1C,$8C,$79,$7F,$EF,$86,$C7
	.dc.b	$FB,$EB,$8F,$F4,$EF,$F8,$BB,$EA,$7E,$33,$78,$FF,$17,$3A,$E0,$FC
	.dc.b	$F7,$37,$F8,$6A,$BC,$2D,$7C,$0D,$7F,$4F,$82,$31,$1F,$E7,$7F,$A2
	.dc.b	$FF,$13,$BF,$E7,$46,$BF,$E4,$F6,$D7,$EB,$5F,$B9,$DD,$FB,$2C,$3F
	.dc.b	$77,$5F,$E1,$DC,$AB,$EB,$F0,$3D,$27,$E1,$AF,$EA,$C7,$5E,$3C,$19
	.dc.b	$F4,$35,$F2,$FF,$8C,$F2,$5F,$1B,$4C,$DF,$FA,$A7,$E9,$7F,$C6,$0F
	.dc.b	$F8,$12,$FF,$22,$7E,$CA,$33,$1E,$BF,$13,$F2,$B8,$48,$F8,$11,$E5
	.dc.b	$FA,$6F,$07,$4C,$9F,$89,$1C,$EB,$F3,$FE,$CF,$FA,$26,$7F,$D3,$6D
	.dc.b	$7F,$D0,$B8,$27,$E0,$7D,$1F,$E9,$3D,$99,$5F,$3E,$3F,$83,$FD,$33
	.dc.b	$E3,$3F,$84,$F8,$8F,$9A,$A6,$A2,$5F,$FB,$26,$74,$D9,$CB,$FE,$96
	.dc.b	$AE,$8F,$F6,$50,$E1,$FE,$D0,$D7,$FC,$E6,$D7,$F0,$DF,$35,$F6,$48
	.dc.b	$FE,$1F,$E9,$AB,$F8,$78,$BD,$5F,$87,$9F,$C0,$7C,$74,$AF,$C7,$3F
	.dc.b	$83,$FA,$6C,$47,$E3,$E2,$D7,$E3,$E7,$F9,$5F,$9A,$69,$2B,$FD,$D5
	.dc.b	$0F,$A6,$C4,$7F,$AD,$A3,$83,$A9,$E7,$3A,$AF,$F3,$24,$FF,$2C,$9C
	.dc.b	$77,$7F,$4E,$BE,$93,$FE,$F7,$FA,$57,$AD,$E4,$D2,$4E,$63,$FA,$0C
	.dc.b	$EA,$3E,$83,$5F,$F0,$B3,$5F,$F5,$0B,$5F,$DA,$65,$FA,$72,$3F,$B5
	.dc.b	$FA,$6A,$FE,$D6,$2E,$7F,$B5,$9F,$0F,$C1,$FC,$8F,$0A,$5F,$A6,$32
	.dc.b	$FC,$88,$CB,$FE,$DC,$C3,$F8,$55,$FC,$47,$F8,$3C,$7F,$72,$6A,$69
	.dc.b	$AF,$11,$F4,$DF,$A1,$9F,$F7,$FB,$5F,$F6,$D3,$6B,$FB,$C2,$E7,$FE
	.dc.b	$96,$E7,$7F,$F0,$0D,$F4,$E8,$C9,$AA,$FE,$DF,$B7,$F1,$4C,$7F,$C0
	.dc.b	$FF,$4D,$F7,$32,$FB,$AD,$C7,$FC,$13,$5B,$AF,$B7,$A7,$E9,$5A,$7E
	.dc.b	$9E,$3F,$CC,$AF,$FB,$59,$AF,$2E,$3F,$EF,$FF,$1B,$F8,$68,$FF,$53
	.dc.b	$5F,$F8,$5A,$31,$8F,$E3,$AD,$7E,$83,$8F,$0C,$C3,$FC,$3F,$5B,$9D
	.dc.b	$47,$8D,$F3,$57,$28,$C1,$FB,$DF,$A5,$7C,$8F,$F8,$4B,$E6,$BF,$A9
	.dc.b	$E1,$F9,$E6,$E7,$FE,$2A,$AD,$DF,$82,$31,$7F,$0E,$73,$3F,$E6,$D6
	.dc.b	$BF,$EF,$86,$BF,$E3,$4F,$A5,$FE,$45,$87,$EB,$3E,$97,$E0,$E1,$FE
	.dc.b	$38,$6D,$7F,$8F,$2B,$73,$C7,$E4,$3E,$94,$7F,$2B,$9F,$F1,$F8,$70
	.dc.b	$EC,$FE,$E3,$F4,$7A,$C3,$FC,$05,$F1,$BF,$E1,$23,$FC,$2B,$19,$C3
	.dc.b	$F1,$6F,$DF,$E5,$34,$97,$FD,$DD,$7D,$1F,$F5,$3F,$3A,$4F,$C7,$15
	.dc.b	$CA,$7B,$9A,$8F,$C1,$9F,$ED,$E9,$CB,$C6,$5F,$2D,$FD,$BB,$0F,$CF
	.dc.b	$FE,$1D,$FE,$0F,$AE,$34,$8C,$67,$FE,$5F,$BE,$15,$E2,$7F,$BC,$BF
	.dc.b	$E0,$06,$BF,$AA,$6B,$BF,$8A,$F7,$7C,$A7,$F8,$E3,$CD,$93,$E5,$D7
	.dc.b	$FC,$BB,$87,$F9,$B5,$FF,$2F,$C7,$F8,$31,$AF,$CC,$AE,$6F,$D4,$FA
	.dc.b	$2F,$A3,$12,$3A,$47,$D4,$30,$FB,$A7,$BE,$9F,$5A,$7F,$CD,$7F,$12
	.dc.b	$E0,$63,$FE,$84,$FA,$6F,$EC,$38,$FF,$67,$E1,$FF,$42,$C6,$3C,$9F
	.dc.b	$B3,$3B,$7F,$D1,$38,$73,$BC,$4F,$FA,$3A,$30,$B9,$39,$61,$FE,$90
	.dc.b	$6C,$7F,$A5,$1F,$0C,$F0,$6E,$3E,$69,$63,$C1,$B8,$FF,$A6,$8F,$2B
	.dc.b	$83,$C5,$3C,$67,$9E,$3E,$69,$F7,$16,$B8,$CE,$A3,$91,$FE,$E2,$AF
	.dc.b	$15,$F8,$49,$5F,$0A,$32,$DB,$E9,$8F,$CA,$8F,$33,$6F,$CB,$3E,$3C
	.dc.b	$7E,$3E,$3E,$D2,$5F,$3A,$F1,$CE,$A3,$9C,$76,$39,$61,$EF,$8E,$15
	.dc.b	$CA,$7F,$D8,$31,$EE,$32,$47,$38,$79,$5F,$8A,$78,$2B,$A3,$E8,$3E
	.dc.b	$2C,$F3,$78,$9E,$2C,$23,$AA,$E9,$3B,$8F,$99,$3A,$BF,$34,$78,$1F
	.dc.b	$A4,$49,$F7,$C7,$33,$83,$F1,$7E,$27,$FC,$91,$8F,$E4,$1B,$8F,$E2
	.dc.b	$5B,$87,$FB,$89,$FE,$43,$EC,$5C,$FF,$92,$E3,$FE,$7D,$13,$A6,$C6
	.dc.b	$A7,$84,$7F,$2C,$7F,$19,$5F,$E5,$7D,$7F,$1D,$3F,$EF,$09,$CF,$84
	.dc.b	$9C,$CF,$7D,$3F,$63,$7C,$63,$F2,$27,$13,$F2,$6E,$63,$F6,$F1,$97
	.dc.b	$8D,$75,$8E,$27,$5D,$3F,$DF,$B1,$99,$E3,$39,$3F,$B3,$C6,$67,$DD
	.dc.b	$1E,$08,$C3,$EF,$31,$7E,$AC,$3D,$AD,$8C,$9E,$28,$CB,$4C,$DF,$8A
	.dc.b	$32,$D8,$CD,$F8,$A3,$DB,$5D,$5B,$3C,$DA,$71,$C7,$25,$AF,$83,$5F
	.dc.b	$C3,$DE,$86,$2B,$93,$6B,$C4,$74,$77,$1F,$1A,$7D,$D5,$DA,$7E,$3B
	.dc.b	$67,$E3,$57,$16,$CF,$53,$FB,$4F,$C3,$7D,$7C,$BF,$A2,$79,$B7,$1C
	.dc.b	$9F,$D2,$5C,$5F,$F6,$6E,$15,$E3,$9F,$FB,$8D,$59,$CD,$F1,$25,$FF
	.dc.b	$BA,$E1,$3F,$E7,$88,$FE,$F6,$D7,$B3,$6B,$AD,$7F,$6F,$F8,$17,$D4
	.dc.b	$77,$AF,$14,$62,$7F,$BA,$74,$3E,$1B,$93,$63,$FD,$FE,$DC,$7F,$DF
	.dc.b	$CD,$7F,$B8,$4F,$B7,$A9,$7E,$FA,$E1,$AF,$E7,$F3,$FE,$F9,$3F,$07
	.dc.b	$E6,$3E,$27,$B4,$F8,$4C,$69,$FE,$07,$8C,$3F,$DD,$67,$73,$AD,$3F
	.dc.b	$C0,$F5,$A7,$FB,$67,$CE,$FC,$3E,$03,$49,$87,$F8,$A9,$F8,$7E,$CF
	.dc.b	$D3,$6C,$7F,$87,$EB,$FC,$47,$C5,$F3,$DB,$1F,$EC,$DE,$BF,$B3,$1F
	.dc.b	$C5,$1A,$FF,$8E,$9C,$ED,$F0,$7E,$19,$F1,$9F,$86,$AE,$67,$F8,$82
	.dc.b	$33,$7E,$1F,$07,$FC,$4E,$FF,$B6,$DB,$3F,$C8,$9A,$FF,$92,$1A,$88
	.dc.b	$F0,$BF,$CD,$7D,$2B,$DC,$F1,$4E,$6F,$FC,$E7,$39,$7E,$17,$D3,$3E
	.dc.b	$67,$E1,$36,$3F,$DD,$CD,$C7,$F9,$AB,$5F,$F2,$DB,$9C,$7C,$3F,$29
	.dc.b	$F0,$63,$15,$E0,$8C,$5F,$EA,$A9,$82,$73,$5F,$E5,$39,$CC,$7C,$BF
	.dc.b	$A5,$78,$BF,$5D,$B5,$FD,$23,$24,$61,$FE,$96,$D7,$FC,$E2,$E6,$FF
	.dc.b	$80,$7D,$03,$E5,$7B,$61,$E5,$8C,$61,$FE,$5D,$FA,$67,$4F,$EF,$DB
	.dc.b	$19,$BC,$4C,$8E,$0E,$E3,$19,$7F,$9A,$8F,$3F,$C8,$FA,$C9,$5E,$BC
	.dc.b	$D1,$89,$FF,$38,$E6,$F4,$7F,$63,$7F,$83,$9E,$6F,$C9,$8C,$C7,$E1
	.dc.b	$EF,$9C,$E7,$C0,$F9,$98,$7D,$3B,$DF,$77,$FA,$2B,$E3,$3F,$8D,$1F
	.dc.b	$C6,$FE,$9A,$E1,$FE,$6B,$8B,$97,$F9,$7E,$6E,$7F,$D8,$61,$F4,$6F
	.dc.b	$91,$FD,$73,$70,$FB,$0D,$9F,$F0,$0B,$5F,$F5,$3B,$9A,$FE,$1D,$F4
	.dc.b	$CF,$5F,$F5,$D3,$9B,$FF,$76,$CE,$5F,$C5,$73,$3E,$2B,$F2,$0D,$8C
	.dc.b	$FA,$98,$99,$1C,$1D,$C6,$32,$FA,$35,$C5,$B3,$AA,$E0,$7F,$A3,$FE
	.dc.b	$7F,$FA,$11,$FD,$0F,$E9,$AC,$64,$95,$FE,$87,$8C,$BF,$E8,$9F,$8A
	.dc.b	$FE,$94,$7F,$4B,$E3,$B5,$FE,$9B,$8B,$3F,$E8,$F9,$BB,$7D,$1D,$5F
	.dc.b	$43,$F4,$09,$87,$D0,$6D,$FE,$33,$AB,$F9,$5C,$FF,$25,$AE,$2F,$EC
	.dc.b	$E3,$33,$F9,$99,$CB,$C7,$C5,$FA,$77,$FA,$97,$FB,$17,$CB,$7F,$54
	.dc.b	$EF,$B8,$FF,$59,$7F,$93,$7D,$1F,$FB,$52,$3F,$D9,$5E,$1F,$F5,$9C
	.dc.b	$7F,$4D,$39,$7F,$D9,$31,$9D,$3F,$A8,$79,$5F,$E3,$BB,$FE,$C9,$EB
	.dc.b	$7F,$50,$7F,$B1,$BF,$D4,$BD,$EF,$EC,$BF,$1D,$67,$FB,$53,$E2,$F9
	.dc.b	$DF,$ED,$3E,$05,$9F,$ED,$CF,$F5,$AF,$47,$FB,$6F,$AF,$67,$FB,$93
	.dc.b	$FD,$7B,$E9,$24,$7F,$5D,$F8,$4B,$BF,$AB,$C7,$FF,$04,$FC,$E3,$C0
	.dc.b	$79,$BF,$0A,$73,$18,$C3,$FB,$37,$AB,$FD,$A8,$FE,$D7,$F1,$76,$3F
	.dc.b	$DE,$1A,$7C,$1C,$BF,$81,$1F,$EF,$A7,$0F,$F0,$5F,$23,$FD,$85,$3F
	.dc.b	$E0,$B9,$CC,$FF,$09,$E7,$FE,$3E,$FF,$BF,$B4,$F3,$7C,$9F,$D6,$F3
	.dc.b	$3C,$AF,$F8,$7F,$E9,$1F,$05,$FE,$15,$6E,$5F,$E7,$AC,$D1,$8C,$7C
	.dc.b	$B1,$8A,$FF,$0F,$62,$1F,$2D,$FD,$A7,$E9,$3F,$C1,$F8,$FF,$5F,$EC
	.dc.b	$FF,$36,$56,$EB,$5C,$12,$B4,$7F,$65,$9F,$F7,$77,$1E,$73,$F6,$7C
	.dc.b	$1F,$CD,$7F,$17,$DF,$FE,$28,$F9,$1F,$F1,$D7,$81,$FD,$62,$5B,$FF
	.dc.b	$88,$27,$C9,$CB,$33,$EF,$5E,$9C,$DF,$1D,$98,$FE,$7D,$C0,$BF,$C0
	.dc.b	$E8,$E3,$D7,$0F,$B2,$31,$97,$4C,$91,$8C,$55,$FE,$2F,$E0,$7C,$A7
	.dc.b	$4C,$7F,$C8,$2F,$C9,$F7,$E6,$5C,$ED,$E6,$63,$1F,$2D,$7E,$53,$E9
	.dc.b	$1E,$A6,$6F,$E7,$C6,$30,$FF,$14,$6F,$FE,$33,$9C,$D6,$E7,$E8,$56
	.dc.b	$F0,$FE,$BD,$F4,$9F,$EB,$A9,$FE,$EC,$D8,$FF,$5F,$B5,$FC,$CB,$5F
	.dc.b	$EB,$8E,$EF,$C7,$E7,$FF,$8B,$63,$15,$E0,$8F,$F2,$7F,$B5,$FE,$62
	.dc.b	$E8,$B3,$FE,$67,$6E,$1F,$2A,$3F,$C5,$5C,$DF,$E5,$C9,$CE,$3F,$35
	.dc.b	$FF,$2C,$FD,$1F,$FA,$EF,$E2,$1F,$23,$C3,$CC,$4E,$63,$1F,$3B,$73
	.dc.b	$FC,$4D,$6B,$3F,$F6,$95,$71,$BD,$66,$AD,$5B,$1E,$EF,$2C,$77,$3A
	.dc.b	$9C,$CB,$E8,$FF,$B2,$7D,$07,$68,$FF,$23,$46,$1E,$99,$7F,$81,$FC
	.dc.b	$CF,$DF,$DF,$2A,$D3,$7D,$A6,$1C,$F0,$ED,$18,$79,$9F,$E5,$3B,$E7
	.dc.b	$87,$F8,$4F,$0E,$6D,$AE,$18,$75,$C3,$A5,$F0,$C3,$CF,$5A,$BE,$15
	.dc.b	$E1,$C3,$9D,$77,$F1,$7E,$73,$E4,$9D,$5F,$18,$E7,$1D,$A7,$75,$CA
	.dc.b	$78,$1F,$CA,$0A,$7F,$2A,$7E,$43,$6B,$E0,$3F,$CB,$BE,$39,$F1,$FF
	.dc.b	$C6,$63,$94,$63,$BF,$FC,$A8,$D8,$CC,$ED,$ED,$EE,$BD,$3F,$F2,$D6
	.dc.b	$3F,$E2,$76,$C6,$67,$CD,$39,$6C,$67,$6F,$B4,$DD,$7E,$93,$67,$F0
	.dc.b	$7E,$DF,$F9,$1E,$B7,$87,$CD,$AD,$57,$E2,$DB,$87,$F8,$A6,$35,$31
	.dc.b	$FC,$09,$B7,$EB,$5F,$5C,$3C,$31,$BC,$3B,$D7,$0B,$E8,$DC,$3B,$92
	.dc.b	$B9,$C7,$E6,$F8,$BF,$DA,$3F,$27,$D6,$3F,$9A,$F8,$48,$C4,$EA,$78
	.dc.b	$4B,$E6,$B2,$EE,$33,$58,$33,$3A,$F9,$5F,$F3,$67,$57,$F2,$FF,$99
	.dc.b	$BD,$5F,$C2,$31,$46,$31,$F1,$DD,$CB,$FD,$CE,$DA,$34,$D7,$E9,$5F
	.dc.b	$DE,$31,$FF,$75,$3F,$8E,$6F,$C5,$E8,$DF,$44,$C7,$FA,$8E,$2B,$56
	.dc.b	$E7,$31,$CF,$3F,$E3,$FF,$81,$BE,$A7,$CC,$FF,$40,$AE,$6D,$7F,$93
	.dc.b	$1F,$D8,$9E,$0E,$A3,$8F,$89,$23,$04,$38,$E2,$AE,$99,$B0,$F1,$DF
	.dc.b	$8B,$0F,$87,$D9,$F1,$3D,$19,$97,$AA,$3E,$3F,$C5,$6A,$30,$7B,$9F
	.dc.b	$97,$F2,$9A,$8C,$1F,$09,$F7,$FC,$2C,$8F,$7B,$F0,$BC,$6F,$87,$F0
	.dc.b	$B6,$7E,$1C,$61,$F8,$BF,$03,$F2,$08,$FC,$EF,$8F,$FB,$7F,$4D,$23
	.dc.b	$E8,$BF,$7B,$E5,$3E,$FF,$BD,$F8,$0F,$1E,$57,$D6,$8F,$C1,$FC,$2C
	.dc.b	$8F,$B2,$FE,$1B,$87,$F0,$FC,$52,$7E,$EB,$F8,$AF,$8B,$FC,$87,$B1
	.dc.b	$FF,$00,$67,$F0,$AB,$E0,$B2,$6B,$DB,$FC,$8F,$AD,$2B,$F0,$EF,$E5
	.dc.b	$38,$7F,$2B,$C7,$F9,$6E,$39,$5F,$8C,$8F,$CC,$F1,$FE,$6B,$EA,$24
	.dc.b	$FE,$42,$3F,$45,$FA,$5F,$FB,$0F,$AF,$F0,$13,$FE,$2C,$8C,$75,$FF
	.dc.b	$72,$6C,$67,$E9,$F8,$FC,$4C,$8F,$CF,$CA,$3A,$FA,$3F,$03,$CE,$F8
	.dc.b	$3F,$1D,$2F,$E0,$C7,$8B,$CC,$F1,$FC,$44,$9E,$71,$FB,$95,$D9,$1F
	.dc.b	$F1,$E9,$8E,$BD,$41,$AC,$8D,$B9,$6C,$65,$FF,$80,$11,$D5,$B1,$D0
	.dc.b	$D6,$DF,$1B,$C4,$6B,$F6,$EB,$F5,$6D,$9D,$37,$7F,$95,$CF,$F4,$39
	.dc.b	$E5,$FC,$87,$E9,$C7,$BF,$E8,$9C,$BF,$D8,$68,$23,$2E,$5A,$FF,$67
	.dc.b	$6B,$FE,$90,$73,$8F,$FC,$01,$6E,$1D,$27,$A5,$7C,$AF,$13,$F0,$BF
	.dc.b	$13,$F8,$6F,$2E,$57,$DE,$8C,$BF,$80,$8C,$BF,$0A,$32,$FC,$48,$CE
	.dc.b	$3F,$12,$7F,$23,$F0,$52,$BF,$12,$FE,$53,$E3,$31,$1F,$32,$2D,$7C
	.dc.b	$C9,$FC,$CF,$B9,$F9,$AF,$A9,$97,$F9,$28,$CB,$F9,$48,$CB,$F5,$A3
	.dc.b	$F1,$9F,$46,$F9,$BB,$F3,$32,$BF,$1B,$39,$6C,$E5,$FE,$62,$21,$3D
	.dc.b	$26,$4F,$28,$E1,$8F,$EB,$A3,$84,$FE,$77,$E8,$FF,$D2,$51,$F7,$DA
	.dc.b	$FF,$88,$5A,$FF,$B2,$9A,$FD,$97,$73,$D5,$BC,$19,$D3,$DF,$F1,$DD
	.dc.b	$FE,$96,$61,$FA,$AB,$47,$EB,$FE,$8F,$8C,$FE,$42,$72,$EE,$78,$BF
	.dc.b	$0A,$76,$BE,$AF,$FA,$9B,$F1,$9F,$E9,$B8,$FE,$3D,$B1,$F9,$C7,$79
	.dc.b	$7C,$56,$CE,$FC,$AF,$E0,$51,$C2,$BB,$79,$78,$33,$32,$3F,$82,$CF
	.dc.b	$27,$F5,$3F,$7B,$FC,$05,$FB,$2D,$C3,$EA,$37,$0F,$D2,$7D,$AF,$FA
	.dc.b	$2A,$BF,$87,$CE,$5E,$7D,$DF,$C5,$63,$10,$9E,$0F,$E1,$BA,$3F,$D7
	.dc.b	$AF,$F1,$86,$B9,$F2,$FF,$8F,$7D,$D3,$87,$F9,$05,$6B,$3F,$C4,$37
	.dc.b	$3F,$DC,$37,$0F,$E0,$B7,$C5,$FE,$67,$F0,$FF,$CF,$BE,$1E,$46,$08
	.dc.b	$7F,$18,$72,$FF,$1A,$8F,$E8,$BF,$41,$FD,$1B,$8E,$4F,$F3,$E8,$CB
	.dc.b	$FD,$02,$32,$FF,$22,$8F,$E9,$BE,$3F,$F4,$EF,$22,$4F,$F4,$68,$CB
	.dc.b	$FD,$22,$32,$FF,$2A,$8F,$EA,$BF,$1B,$FE,$37,$FD,$33,$FC,$DD,$3F
	.dc.b	$E1,$86,$ED,$FC,$BA,$77,$3F,$E3,$09,$CE,$DF,$EC,$A6,$C7,$F5,$D9
	.dc.b	$E9,$BC,$8E,$0F,$7F,$8B,$C4,$77,$8B,$87,$F4,$1B,$57,$F5,$5F,$89
	.dc.b	$7C,$8A,$BF,$9B,$FD,$23,$DC,$FE,$5C,$7F,$90,$B2,$CB,$8F,$81,$FE
	.dc.b	$65,$3B,$9F,$E7,$F8,$74,$BD,$C6,$30,$EC,$AB,$83,$DE,$F9,$E1,$D6
	.dc.b	$38,$8C,$3A,$3C,$63,$FA,$14,$62,$B9,$C7,$9E,$74,$7F,$52,$71,$1E
	.dc.b	$09,$E1,$1C,$AF,$CB,$3C,$AF,$C7,$7C,$5A,$F8,$E7,$75,$C7,$0F,$3B
	.dc.b	$D8,$F5,$4F,$88,$E1,$1E,$CC,$3C,$31,$D4,$E3,$7D,$A3,$31,$C0,$E5
	.dc.b	$5D,$5E,$07,$58,$D9,$C2,$35,$1F,$34,$F0,$5F,$A1,$FA,$97,$E0,$61
	.dc.b	$CE,$34,$F8,$70,$F8,$8E,$9F,$95,$1A,$BF,$06,$1F,$1B,$0F,$0C,$79
	.dc.b	$27,$9C,$79,$0C,$BC,$4C,$CF,$13,$E1,$39,$C3,$C1,$3F,$32,$FC,$31
	.dc.b	$EC,$3F,$C1,$FF,$0F,$FC,$2F,$1F,$90,$DC,$7F,$12,$DC,$3F,$26,$EE
	.dc.b	$7F,$C2,$FF,$13,$CB,$5F,$F7,$D3,$63,$FC,$20,$1C,$3F,$EE,$13,$FD
	.dc.b	$E1,$1C,$1F,$C0,$60,$FA,$5F,$F2,$BE,$5F,$A0,$9E,$E6,$A3,$39,$F1
	.dc.b	$9F,$D2,$D7,$1E,$0F,$D1,$DF,$09,$CD,$F0,$8F,$2E,$3F,$A8,$BE,$11
	.dc.b	$F9,$13,$86,$1F,$ED,$2A,$E1,$1E,$9B,$E7,$87,$68,$F7,$1E,$57,$18
	.dc.b	$75,$AE,$33,$D0,$FF,$1F,$E1,$E0,$BF,$74,$72,$7A,$BE,$33,$18,$78
	.dc.b	$F0,$EB,$1D,$CD,$61,$DE,$33,$5C,$8E,$B1,$99,$D9,$C6,$32,$DA,$F9
	.dc.b	$87,$77,$71,$CE,$33,$1E,$AC,$3D,$AF,$98,$C5,$F8,$A7,$95,$76,$39
	.dc.b	$4F,$D2,$3C,$05,$72,$FC,$F8,$E2,$78,$1F,$E6,$55,$F6,$FD,$07,$03
	.dc.b	$E2,$FF,$D3,$F9,$FF,$3B,$6E,$5F,$D1,$FE,$5A,$E2,$C3,$8D,$61,$B5
	.dc.b	$FD,$3E,$31,$82,$EC,$63,$E5,$33,$95,$BF,$EA,$CD,$C7,$FA,$E3,$72
	.dc.b	$FE,$9C,$F4,$F7,$BA,$F0,$DB,$FE,$BC,$DB,$FE,$C8,$DB,$FE,$AC,$F9
	.dc.b	$3E,$3B,$CB,$E4,$DB,$FE,$CC,$DB,$FE,$D8,$DB,$FE,$BC,$FB,$3C,$AF
	.dc.b	$6F,$CB,$5B,$FE,$DC,$DB,$FE,$E8,$DB,$FE,$CC,$FC,$A8,$7E,$04,$F3
	.dc.b	$F0,$DB,$FC,$F9,$FE,$9E,$E2,$7E,$04,$F2,$F7,$72,$BF,$D8,$37,$C7
	.dc.b	$0F,$EE,$11,$C6,$F8,$3A,$AF,$17,$63,$BE,$9F,$7B,$47,$D8,$B8,$B3
	.dc.b	$87,$FB,$3A,$F8,$36,$F8,$56,$E3,$07,$AB,$2F,$F0,$17,$D0,$F9,$71
	.dc.b	$FC,$7A,$F1,$3C,$0E,$6B,$F7,$EE,$5E,$52,$8D,$57,$4A,$E3,$9F,$26
	.dc.b	$DF,$0A,$E5,$5F,$10,$F1,$D6,$E7,$FD,$D3,$F4,$6F,$87,$FA,$CD,$8F
	.dc.b	$57,$0F,$D2,$F2,$B3,$52,$78,$BF,$A3,$E0,$4E,$71,$FF,$52,$E0,$F8
	.dc.b	$5E,$A7,$C0,$F2,$2E,$3E,$28,$FF,$52,$57,$FB,$4E,$B8,$4F,$FB,$85
	.dc.b	$B1,$FD,$20,$38,$7F,$DB,$5B,$7D,$AF,$9B,$FA,$BF,$37,$F5,$9B,$7F
	.dc.b	$33,$85,$18,$BF,$8D,$7C,$6B,$57,$C5,$A7,$1A,$E1,$3F,$E4,$3F,$A0
	.dc.b	$FE,$03,$F4,$1F,$C0,$9B,$5F,$E5,$36,$E1,$FE,$59,$6D,$7F,$92,$EF
	.dc.b	$F8,$49,$FC,$2B,$0B,$7F,$E5,$B8,$B7,$FE,$61,$9B,$5F,$E5,$3F,$97
	.dc.b	$FE,$23,$F2,$FF,$C4,$9B,$87,$F9,$8D,$B7,$FE,$69,$6D,$71,$9F,$F3
	.dc.b	$45,$71,$69,$F9,$BA,$7D,$0B,$DD,$FD,$E3,$67,$E6,$61,$89,$91,$F5
	.dc.b	$EB,$8C,$7D,$E4,$62,$35,$5C,$E3,$FC,$0B,$C9,$FC,$E3,$E5,$FF,$96
	.dc.b	$9F,$CB,$BE,$62,$DF,$F9,$FE,$2D,$7F,$A1,$66,$E6,$AF,$E1,$9E,$C7
	.dc.b	$81,$FC,$EA,$F9,$8F,$E7,$8E,$6F,$F5,$18,$A3,$55,$C2,$77,$5C,$1B
	.dc.b	$19,$3A,$C6,$5A,$78,$3D,$1E,$FF,$2D,$E0,$6C,$7E,$C3,$EE,$9F,$0F
	.dc.b	$FD,$3D,$7C,$25,$8C,$3E,$8F,$6E,$57,$B3,$E5,$EE,$7E,$AD,$3F,$D2
	.dc.b	$53,$F0,$3D,$2F,$7F,$CB,$FC,$07,$E0,$FC,$BF,$C2,$F8,$DB,$5F,$19
	.dc.b	$F8,$B7,$F1,$BE,$5E,$CF,$CA,$7E,$4F,$89,$23,$E6,$F9,$97,$3F,$97
	.dc.b	$C1,$FC,$A9,$BE,$E7,$B7,$CA,$F7,$3F,$4B,$E1,$26,$1F,$5B,$E5,$ED
	.dc.b	$FA,$FC,$8F,$AE,$7D,$8F,$97,$B5,$F7,$3E,$37,$ED,$F9,$3F,$71,$B1
	.dc.b	$F8,$0F,$96,$FB,$E7,$E0,$3E,$5E,$C7,$F2,$0E,$AF,$E7,$99,$7E,$09
	.dc.b	$BD,$FF,$8B,$8F,$F0,$27,$E9,$BF,$8D,$AE,$1E,$9F,$F8,$16,$FF,$D8
	.dc.b	$71,$CF,$D8,$C7,$B3,$97,$5E,$CC,$9D,$1E,$17,$FC,$1D,$83,$E6,$2F
	.dc.b	$CC,$61,$B5,$FC,$DA,$53,$C3,$E6,$33,$A7,$E6,$A5,$19,$AE,$1F,$2F
	.dc.b	$6B,$4F,$A7,$D1,$F5,$7A,$56,$BE,$23,$6B,$E2,$37,$83,$E2,$79,$5F
	.dc.b	$98,$8C,$57,$F7,$BE,$04,$62,$FF,$C5,$91,$89,$F5,$7C,$C7,$C0,$F9
	.dc.b	$7F,$82,$D8,$C9,$EF,$8C,$B4,$F8,$53,$F2,$0F,$93,$E4,$D8,$C9,$F1
	.dc.b	$FE,$5B,$18,$7C,$57,$E7,$FC,$A7,$DD,$F8,$8F,$F2,$34,$FD,$E8,$C7
	.dc.b	$C2,$FF,$85,$78,$BF,$8B,$56,$B6,$42,$B7,$7C,$71,$FF,$51,$E1,$C6
	.dc.b	$73,$7C,$3D,$4B,$87,$1B,$DE,$1F,$C9,$BE,$6F,$EC,$FA,$58,$84,$61
	.dc.b	$FF,$55,$65,$FC,$B1,$7A,$4F,$2B,$C7,$1C,$39,$FF,$90,$3C,$0E,$CE
	.dc.b	$36,$F8,$31,$BC,$7C,$17,$CF,$2F,$16,$5E,$1B,$DB,$72,$F1,$37,$0F
	.dc.b	$E7,$F7,$C5,$F0,$1D,$4B,$1B,$71,$5E,$29,$E0,$F6,$C7,$CD,$7D,$9B
	.dc.b	$5E,$AC,$7C,$F9,$7C,$0C,$7A,$1F,$1E,$7F,$DA,$2D,$79,$4F,$F5,$78
	.dc.b	$C4,$7F,$56,$F8,$C7,$63,$E7,$3F,$AE,$46,$1B,$19,$9F,$F0,$B4,$E5
	.dc.b	$FF,$0C,$53,$A9,$F1,$BD,$DB,$3A,$6B,$E6,$6B,$F3,$1D,$EB,$F4,$1B
	.dc.b	$AE,$6B,$EA,$FC,$77,$E0,$EB,$51,$F6,$2B,$77,$F8,$D6,$BF,$8C,$AE
	.dc.b	$17,$CE,$3E,$AC,$EB,$0F,$55,$F6,$9E,$8C,$76,$E5,$A4,$C3,$C6,$FE
	.dc.b	$03,$0E,$EF,$FA,$BF,$95,$ED,$63,$95,$18,$9F,$D2,$C6,$27,$52,$F6
	.dc.b	$EC,$4A,$DC,$F1,$7F,$41,$5D,$2B,$8F,$B5,$FE,$4A,$E5,$FF,$49,$5B
	.dc.b	$DA,$C6,$1E,$56,$E1,$E4,$BB,$46,$CC,$AF,$F3,$5C,$E5,$FF,$36,$FC
	.dc.b	$E7,$EC,$9F,$F0,$7A,$F6,$FA,$B6,$37,$CB,$83,$33,$23,$27,$F9,$E3
	.dc.b	$E7,$3F,$D5,$0D,$F4,$3C,$F0,$6F,$A9,$27,$44,$3F,$C4,$33,$C9,$B1
	.dc.b	$CA,$64,$77,$8F,$3F,$5F,$A3,$DC,$93,$EE,$8F,$5F,$B9,$A8,$C1,$F0
	.dc.b	$9C,$B4,$87,$C4,$F6,$FD,$3E,$87,$D0,$F0,$1F,$35,$4C,$D7,$C1,$F4
	.dc.b	$BD,$FE,$9C,$8C,$1F,$17,$D3,$F8,$64,$7E,$9F,$BB,$F4,$3D,$D9,$1F
	.dc.b	$59,$FA,$7E,$EF,$D4,$F7,$7E,$AF,$8F,$2B,$ED,$C7,$D8,$F7,$64,$7D
	.dc.b	$E7,$ED,$F1,$FD,$CE,$49,$3F,$82,$7E,$FF,$B5,$F8,$CF,$61,$E8,$5F
	.dc.b	$42,$5F,$FA,$D2,$72,$FE,$2B,$DA,$FC,$39,$1F,$C6,$F0,$FE,$2B,$97
	.dc.b	$F1,$7C,$72,$7F,$23,$1F,$8E,$E3,$FC,$7F,$B5,$27,$F2,$B1,$F9,$EF
	.dc.b	$9C,$FF,$10,$47,$F8,$2F,$E1,$BE,$91,$9F,$17,$5E,$CF,$D5,$F8,$89
	.dc.b	$5F,$EC,$AB,$FF,$0A,$BE,$9F,$2B,$DD,$EF,$C8,$C1,$BF,$3F,$C1,$EC
	.dc.b	$48,$C1,$FE,$93,$EB,$1F,$B4,$6F,$55,$8F,$F3,$C4,$6D,$C1,$98,$91
	.dc.b	$92,$1F,$EE,$73,$99,$DD,$E2,$FA,$F9,$BE,$21,$1F,$77,$35,$7F,$D1
	.dc.b	$39,$FD,$66,$CF,$CB,$9F,$47,$CE,$38,$FD,$1E,$0C,$BF,$F5,$84,$E5
	.dc.b	$B3,$FE,$AB,$51,$97,$FD,$C8,$D7,$FC,$04,$6A,$3F,$D6,$5E,$63,$1F
	.dc.b	$F4,$5C,$FD,$5F,$2B,$EF,$1F,$7F,$D9,$91,$92,$1E,$C7,$2F,$BA,$38
	.dc.b	$BF,$7E,$3F,$11,$EC,$FE,$DF,$E5,$64,$FE,$0E,$32,$D2,$1F,$23,$DA
	.dc.b	$FC,$87,$8D,$F9,$1E,$99,$3F,$8B,$8C,$BF,$8D,$8C,$BF,$42,$3F,$09
	.dc.b	$F3,$8F,$80,$BF,$1F,$2B,$FD,$B9,$39,$6C,$E5,$FC,$14,$27,$9C,$C9
	.dc.b	$E3,$1B,$F7,$FF,$4B,$5C,$E7,$F1,$5F,$39,$FE,$43,$7B,$FB,$7E,$43
	.dc.b	$F3,$11,$98,$DE,$9F,$76,$31,$1F,$69,$AF,$F8,$F5,$AF,$FB,$A9,$7C
	.dc.b	$5F,$E8,$FE,$73,$F9,$2C,$FF,$BE,$5A,$F9,$23,$C2,$47,$FD,$FD,$3C
	.dc.b	$9E,$15,$FA,$DF,$9C,$FE,$85,$97,$F0,$27,$FA,$54,$FF,$BE,$A3,$1E
	.dc.b	$56,$FC,$B9,$39,$F1,$3F,$CA,$97,$C2,$3F,$5C,$DB,$E9,$5C,$27,$37
	.dc.b	$E1,$31,$3F,$C4,$BE,$73,$F7,$B3,$FB,$7F,$9C,$7C,$DF,$F0,$36,$C7
	.dc.b	$84,$09,$E3,$5C,$FC,$0F,$E1,$51,$8C,$FF,$CD,$4D,$AE,$36,$F1,$BF
	.dc.b	$88,$7A,$CF,$37,$F8,$97,$CE,$7F,$9A,$33,$FE,$4E,$DC,$BF,$CE,$95
	.dc.b	$BC,$7C,$81,$C3,$FC,$45,$B5,$FC,$C1,$B3,$FE,$7D,$A7,$CE,$7F,$35
	.dc.b	$C7,$F9,$B7,$64,$C7,$F9,$84,$65,$FE,$65,$19,$7F,$8A,$C7,$F3,$DD
	.dc.b	$BF,$9F,$72,$49,$FE,$71,$19,$7F,$9D,$46,$5F,$E3,$B1,$FD,$17,$CA
	.dc.b	$FE,$8D,$E5,$49,$FE,$81,$19,$7F,$A1,$46,$5F,$E4,$B1,$FD,$37,$E3
	.dc.b	$3F,$C5,$7F,$37,$FE,$60,$DF,$FD,$96,$DD,$BE,$CF,$89,$FC,$66,$B8
	.dc.b	$61,$F7,$61,$5B,$84,$63,$E1,$37,$0C,$3A,$EF,$E0,$24,$73,$B7,$CB
	.dc.b	$62,$38,$65,$E2,$9F,$A3,$1B,$AF,$4D,$7F,$53,$F8,$77,$C7,$77,$A7
	.dc.b	$CE,$3C,$CF,$F5,$B1,$FE,$3D,$8C,$B8,$F8,$1F,$F5,$D4,$EE,$7F,$D2
	.dc.b	$55,$CE,$31,$0B,$EF,$A7,$E7,$A7,$38,$75,$6E,$1D,$23,$FD,$8D,$5C
	.dc.b	$2F,$C2,$76,$BE,$53,$E3,$8A,$7E,$A9,$E8,$72,$C3,$C8,$D9,$DC,$FB
	.dc.b	$EF,$C0,$F1,$6B,$E9,$AE,$2F,$84,$F1,$9C,$4E,$75,$E0,$75,$87,$C2
	.dc.b	$C3,$D4,$FB,$A3,$77,$DA,$73,$5E,$59,$EA,$F3,$7B,$1E,$08,$DB,$DA
	.dc.b	$3E,$13,$C2,$7A,$5F,$A2,$3D,$58,$78,$18,$73,$8D,$5F,$48,$EF,$87
	.dc.b	$9D,$D5,$7C,$89,$D6,$1E,$53,$CC,$F1,$33,$87,$CA,$8C,$E1,$D8,$F2
	.dc.b	$3C,$63,$CF,$5D,$67,$31,$C4,$CD,$F8,$8F,$F6,$AE,$5F,$1F,$5F,$CF
	.dc.b	$C7,$B8,$C9,$F5,$DC,$BF,$CB,$BE,$23,$F9,$7C,$62,$11,$89,$FF,$0A
	.dc.b	$C6,$0E,$67,$FB,$5C,$A7,$33,$F0,$78,$3E,$69,$BF,$E3,$A7,$5A,$FE
	.dc.b	$92,$B5,$39,$9E,$D3,$9A,$E3,$1F,$8E,$38,$E9,$FD,$19,$FE,$0D,$1A
	.dc.b	$9F,$D7,$C6,$8F,$EC,$11,$99,$F5,$E3,$FA,$0C,$3C,$35,$D4,$F0,$CF
	.dc.b	$58,$CD,$F4,$9C,$9D,$30,$F4,$5F,$3C,$3B,$4F,$F8,$F2,$31,$3A,$8F
	.dc.b	$8B,$3A,$7E,$55,$7B,$E7,$4D,$8F,$25,$72,$69,$CA,$77,$87,$9E,$FA
	.dc.b	$4F,$C1,$3E,$2C,$E9,$F4,$1C,$AF,$B1,$ED,$BE,$CD,$8E,$C6,$EF,$B5
	.dc.b	$73,$8E,$A6,$4E,$B3,$C4,$EB,$3F,$42,$39,$C6,$63,$9C,$71,$39,$C7
	.dc.b	$F2,$BE,$0F,$E8,$FE,$27,$E0,$B2,$FE,$9E,$DC,$3F,$9F,$7C,$E5,$C3
	.dc.b	$FA,$5F,$12,$B1,$87,$2B,$FE,$91,$18,$95,$D8,$C7,$CE,$67,$5B,$7F
	.dc.b	$D3,$DB,$8F,$F4,$E3,$BB,$8C,$BF,$B1,$BD,$7D,$CE,$DC,$56,$FF,$AC
	.dc.b	$36,$F9,$69,$FD,$92,$31,$87,$F6,$C7,$CD,$F1,$BE,$7F,$2E,$E1,$FD
	.dc.b	$85,$B8,$72,$DB,$FC,$D5,$3F,$D9,$6B,$83,$FD,$9F,$CB,$F8,$1F,$39
	.dc.b	$F0,$5B,$19,$3F,$B6,$46,$5A,$7F,$88,$BE,$71,$E7,$FC,$FC,$51,$89
	.dc.b	$FE,$FD,$F3,$6F,$2B,$53,$9A,$FF,$5C,$37,$6F,$DE,$EE,$FB,$A7,$C4
	.dc.b	$F0,$CF,$EC,$C6,$33,$FF,$19,$F7,$3C,$1F,$13,$63,$FC,$95,$8B,$E6
	.dc.b	$DC,$BA,$B9,$5A,$FA,$57,$9B,$8D,$FE,$E6,$29,$CD,$74,$F9,$EF,$43
	.dc.b	$FE,$6B,$7F,$C9,$CE,$75,$FC,$95,$27,$37,$D1,$E2,$FF,$87,$5C,$C7
	.dc.b	$E7,$5C,$BC,$7C,$8E,$71,$83,$77,$E1,$BC,$CF,$D3,$F9,$E7,$C4,$53
	.dc.b	$F4,$3C,$BF,$E6,$1B,$FF,$2D,$CE,$9B,$3F,$3B,$CD,$E1,$1F,$E3,$7E
	.dc.b	$2F,$F3,$03,$C3,$E1,$BF,$99,$EB,$FE,$D8,$C1,$18,$8F,$E6,$91,$89
	.dc.b	$FE,$93,$08,$C3,$FD,$31,$BA,$FE,$F7,$E9,$9F,$33,$EB,$C3,$E5,$C6
	.dc.b	$30,$FB,$BC,$8F,$9F,$F9,$EB,$E7,$EC,$61,$B3,$C1,$B9,$FD,$EA,$E1
	.dc.b	$3D,$2B,$83,$4E,$07,$F9,$3F,$E7,$3F,$84,$65,$FC,$25,$B1,$93,$FC
	.dc.b	$A9,$19,$69,$FE,$57,$9F,$E2,$07,$F1,$1C,$2C,$64,$FF,$33,$7C,$8F
	.dc.b	$FA,$99,$C6,$3F,$C1,$A3,$0F,$EF,$FE,$67,$F8,$D7,$CC,$FF,$1B,$6D
	.dc.b	$7F,$99,$5B,$3B,$8F,$F3,$44,$ED,$A6,$CB,$1F,$C7,$47,$CF,$BE,$0B
	.dc.b	$EF,$47,$FA,$DA,$31,$3F,$79,$B1,$FE,$08,$6B,$FD,$DC,$E7,$3F,$A6
	.dc.b	$E9,$FC,$94,$77,$9E,$3C,$3F,$CF,$E7,$2F,$FA,$3B,$E6,$BF,$9C,$1F
	.dc.b	$CE,$7E,$6A,$E3,$FE,$83,$8B,$87,$F2,$39,$B5,$FC,$3B,$B3,$FD,$9B
	.dc.b	$7F,$E9,$35,$F3,$3F,$D1,$9C,$DF,$F3,$47,$33,$B8,$D4,$ED,$B1,$93
	.dc.b	$A4,$65,$A7,$7F,$53,$B7,$85,$DD,$B1,$FE,$A5,$F9,$F7,$C6,$7F,$58
	.dc.b	$5F,$11,$63,$0F,$9F,$B7,$FA,$EF,$23,$DC,$C6,$1F,$8A,$7D,$9E,$AE
	.dc.b	$BD,$BB,$3F,$8E,$7E,$07,$CB,$FC,$1F,$8F,$B3,$F1,$5F,$89,$C3,$F1
	.dc.b	$7E,$5E,$CF,$C9,$7E,$47,$89,$23,$E6,$79,$D7,$93,$FA,$EF,$0F,$F3
	.dc.b	$56,$FB,$DE,$DF,$2F,$DC,$FD,$1F,$85,$98,$7D,$5F,$97,$B7,$F4,$FE
	.dc.b	$13,$EB,$1F,$5F,$E5,$ED,$7D,$BF,$90,$FB,$5E,$57,$DB,$6C,$7D,$FF
	.dc.b	$96,$FB,$C7,$DF,$F9,$7B,$1F,$ED,$5E,$BF,$E8,$ED,$F4,$FE,$6B,$FE
	.dc.b	$0A,$8C,$E7,$FD,$E1,$FF,$64,$C7,$1F,$67,$5E,$E6,$DE,$1E,$E7,$D6
	.dc.b	$E2,$B5,$FB,$E7,$EC,$FC,$BC,$8F,$E0,$3F,$3F,$72,$FE,$01,$3F,$7B
	.dc.b	$E7,$F1,$E3,$7F,$68,$F0,$51,$9C,$3F,$84,$FC,$CD,$C3,$F8,$43,$E9
	.dc.b	$F4,$BD,$5E,$95,$AF,$80,$DA,$F6,$37,$1F,$EE,$BC,$1F,$CC,$67,$E0
	.dc.b	$5F,$2B,$F9,$0D,$8F,$E2,$DF,$33,$F9,$4F,$99,$FC,$AB,$6B,$F8,$EB
	.dc.b	$67,$71,$FC,$82,$7F,$37,$7F,$9C,$F9,$7B,$3F,$C9,$1B,$5F,$C8,$DB
	.dc.b	$8F,$C3,$7F,$47,$F2,$7F,$E8,$8F,$8A,$FF,$21,$53,$E0,$5C,$0A,$D7
	.dc.b	$C4,$2B,$53,$FC,$6A,$F8,$DF,$0B,$DD,$69,$B1,$9C,$BF,$D5,$98,$73
	.dc.b	$8E,$FE,$AE,$B1,$F0,$56,$B0,$F0,$7C,$F7,$DA,$F4,$B1,$08,$C3,$FE
	.dc.b	$B4,$C7,$FD,$41,$E9,$3C,$9F,$1C,$EF,$9B,$F8,$DB,$B3,$AB,$8D,$BE
	.dc.b	$1C,$6F,$1E,$F7,$CF,$2F,$0E,$5E,$0B,$DB,$72,$FF,$58,$37,$0F,$91
	.dc.b	$7C,$5F,$01,$D4,$B1,$B7,$15,$E2,$9E,$0F,$6C,$7C,$D7,$D9,$B5,$EA
	.dc.b	$C7,$CF,$97,$C0,$C7,$A1,$F1,$A7,$FA,$73,$5E,$53,$FE,$12,$8C,$47
	.dc.b	$DF,$F8,$E7,$63,$E7,$FF,$C2,$F1,$86,$C6,$67,$FD,$C7,$39,$7F,$DC
	.dc.b	$B4,$E9,$7C,$7F,$76,$CE,$9A,$F9,$9A,$FC,$C7,$7A,$FD,$06,$EB,$9A
	.dc.b	$FA,$BE,$87,$FB,$86,$B5,$1F,$62,$B7,$7F,$8C,$6B,$F8,$BA,$E1,$7C
	.dc.b	$E3,$EA,$CE,$B0,$F5,$5F,$69,$E8,$C7,$6E,$5A,$4C,$3C,$6F,$E0,$30
	.dc.b	$EE,$FE,$B3,$CA,$BC,$8F,$D3,$BE,$1D,$FE,$B9,$B9,$FE,$A7,$91,$7C
	.dc.b	$1B,$87,$EA,$70,$E5,$97,$4C,$38,$CF,$73,$4B,$E4,$3F,$DF,$9C,$BF
	.dc.b	$B5,$87,$C4,$62,$BF,$83,$46,$1F,$E0,$9D,$5F,$E9,$68,$E5,$18,$AE
	.dc.b	$57,$C7,$1F,$63,$6B,$75,$FE,$BE,$7F,$63,$7C,$FE,$3F,$44,$9E,$87
	.dc.b	$8B,$E3,$F0,$F1,$8B,$3E,$39,$F0,$7D,$1F,$E3,$23,$FD,$56,$DC,$BF
	.dc.b	$89,$36,$FF,$91,$3E,$49,$DE,$9C,$89,$3E,$2C,$D5,$BB,$E7,$5A,$AF
	.dc.b	$F6,$CC,$F5,$AD,$E7,$F0,$7E,$43,$E1,$79,$32,$FA,$6A,$9D,$46,$1A
	.dc.b	$FC,$7F,$27,$E5,$11,$E2,$FC,$DE,$8F,$83,$B4,$9F,$46,$68,$F8,$BB
	.dc.b	$C9,$CB,$5F,$4E,$1F,$2B,$2F,$96,$C8,$E7,$1F,$6F,$C9,$91,$EA,$7E
	.dc.b	$F7,$C7,$FD,$FF,$8F,$93,$EE,$7F,$05,$E3,$FE,$0F,$C8,$FC,$27,$C3
	.dc.b	$4A,$F8,$71,$F8,$7F,$86,$FC,$47,$C3,$49,$F8,$F1,$F7,$A6,$47,$B9
	.dc.b	$FC,$0D,$7E,$3F,$E1,$A4,$FC,$E7,$F2,$5F,$09,$F9,$3F,$84,$93,$F4
	.dc.b	$9F,$CC,$7E,$9F,$FE,$75,$EB,$FF,$79,$BF,$D2,$A5,$18,$F1,$B6,$70
	.dc.b	$F8,$5E,$3E,$54,$AF,$C4,$5F,$FB,$E9,$BF,$09,$DB,$E8,$24,$69,$F0
	.dc.b	$FC,$27,$EA,$7E,$29,$14,$93,$F9,$38,$4F,$A3,$E8,$24,$F8,$1F,$5F
	.dc.b	$C3,$7E,$EE,$1F,$0C,$F8,$4C,$36,$3F,$5E,$6A,$33,$F0,$7A,$64,$6A
	.dc.b	$78,$3F,$EA,$55,$3A,$8C,$4D,$E0,$FF,$73,$D6,$A7,$E2,$BF,$43,$E1
	.dc.b	$A4,$60,$87,$D0,$F2,$3F,$C1,$5B,$A7,$73,$93,$FC,$39,$1E,$E7,$E8
	.dc.b	$FA,$7F,$73,$E1,$A4,$60,$87,$1F,$25,$49,$DC,$E6,$3F,$C1,$CF,$DC
	.dc.b	$CF,$F0,$E4,$7E,$FE,$95,$E1,$17,$4F,$C5,$CD,$CF,$D7,$36,$BD,$33
	.dc.b	$F5,$FE,$81,$F0,$5F,$EC,$79,$FF,$1A,$08,$CB,$93,$11,$5A,$66,$36
	.dc.b	$FF,$B6,$DB,$83,$0E,$B1,$87,$BE,$CE,$4F,$D2,$FC,$34,$C3,$F3,$AF
	.dc.b	$EA,$3E,$1B,$11,$F1,$E2,$E3,$FA,$A9,$FC,$97,$CF,$3B,$69,$2F,$F2
	.dc.b	$30,$ED,$C4,$7F,$BC,$63,$83,$A9,$E7,$3A,$BF,$54,$FF,$96,$6B,$6F
	.dc.b	$1A,$DB,$FA,$9F,$87,$FD,$76,$D2,$30,$43,$F0,$9D,$3F,$0D,$FF,$34
	.dc.b	$37,$90,$AB,$F5,$DF,$3B,$FE,$6B,$AF,$A8,$D7,$ED,$35,$FB,$61,$C3
	.dc.b	$F7,$9E,$19,$7B,$23,$5E,$F7,$63,$E5,$FA,$3F,$BD,$F9,$EF,$F7,$3C
	.dc.b	$FE,$E9,$AF,$E9,$43,$8A,$E8,$FF,$BB,$23,$87,$C0,$F3,$23,$CF,$E3
	.dc.b	$2B,$CE,$79,$72,$FE,$51,$B1,$FE,$89,$7A,$3E,$1F,$80,$79,$3F,$95
	.dc.b	$6C,$7E,$65,$AF,$E7,$DA,$FC,$67,$38,$FC,$36,$DF,$E9,$1B,$3F,$E9
	.dc.b	$9C,$2A,$FB,$EF,$F2,$EB,$FA,$76,$BF,$AD,$6B,$F5,$1A,$FE,$C1,$DD
	.dc.b	$FD,$36,$DE,$7D,$4F,$E5,$DE,$AB,$E1,$3F,$A5,$F0,$CC,$7F,$A1,$BF
	.dc.b	$D3,$B9,$31,$1F,$C3,$62,$E5,$FD,$42,$7F,$A9,$FC,$5C,$AF,$E9,$6F
	.dc.b	$F5,$6F,$8B,$C4,$7F,$15,$8B,$5F,$D6,$27,$FA,$DF,$C5,$CA,$FE,$A6
	.dc.b	$FF,$8D,$7D,$4F,$C6,$57,$F5,$E8,$D4,$FF,$1F,$75,$AE,$FE,$2A,$4F
	.dc.b	$F8,$8E,$D1,$CF,$E2,$B3,$3F,$CA,$E7,$2E,$E7,$4F,$1E,$2A,$EE,$BC
	.dc.b	$91,$23,$B4,$6B,$1F,$DE,$F5,$30,$FF,$7F,$D7,$F5,$8F,$86,$70,$FF
	.dc.b	$76,$87,$CC,$FF,$66,$7F,$DB,$9E,$6B,$D0,$FE,$57,$39,$9F,$ED,$71
	.dc.b	$89,$FE,$5F,$7D,$32,$FF,$72,$5E,$E3,$17,$DA,$BF,$9A,$9D,$5D,$B8
	.dc.b	$BE,$91,$FE,$AF,$C3,$CD,$7C,$DB,$5B,$9E,$27,$A2,$B8,$AB,$F0,$DF
	.dc.b	$1C,$37,$3E,$A2,$DF,$8D,$F3,$1E,$18,$DD,$72,$9F,$04,$E8,$E4,$58
	.dc.b	$D3,$C4,$D4,$75,$AF,$94,$D3,$E4,$E1,$EC,$9E,$13,$AC,$3D,$75,$B9
	.dc.b	$CB,$C3,$0F,$4C,$E8,$F8,$38,$78,$AB,$E1,$61,$D5,$F3,$E1,$D5,$F6
	.dc.b	$C6,$67,$86,$1D,$E7,$6F,$77,$DD,$3C,$E3,$46,$E7,$C0,$72,$BF,$25
	.dc.b	$76,$C2,$3F,$12,$32,$F4,$8C,$E1,$E5,$8D,$18,$BF,$17,$BF,$FE,$A6
	.dc.b	$C3,$F1,$AD,$C7,$F3,$8D,$C1,$7F,$EE,$EE,$CF,$F0,$37,$C2,$62,$BF
	.dc.b	$D5,$F3,$A8,$C4,$EA,$38,$1F,$D4,$A3,$2D,$3F,$05,$5F,$E9,$5E,$1F
	.dc.b	$C6,$C7,$33,$3E,$E7,$F7,$0D,$3F,$55,$19,$7F,$59,$5C,$63,$F1,$86
	.dc.b	$74,$FF,$0D,$C6,$67,$8C,$B0,$FC,$DC,$E6,$BD,$15,$D2,$FC,$11,$83
	.dc.b	$9B,$8B,$F5,$4E,$F0,$F4,$C7,$53,$D2,$E3,$0E,$F5,$C6,$7F,$C5,$F1
	.dc.b	$89,$D4,$70,$3D,$67,$3A,$EC,$E2,$78,$34,$E1,$1E,$E8,$CB,$63,$37
	.dc.b	$EB,$7F,$DD,$F7,$E3,$7E,$41,$8B,$F1,$D7,$80,$E4,$5A,$E5,$3F,$E9
	.dc.b	$8A,$E4,$F9,$E3,$E3,$D7,$26,$C7,$9E,$B9,$3F,$2A,$33,$1D,$9A,$76
	.dc.b	$C3,$EB,$4B,$7F,$E6,$DF,$0F,$FE,$34,$9F,$E7,$3C,$0F,$8B,$C6,$5F
	.dc.b	$9A,$BD,$BF,$D3,$63,$38,$73,$A6,$18,$BF,$E8,$55,$67,$37,$C4,$FE
	.dc.b	$AD,$D5,$93,$53,$63,$24,$C3,$FA,$C4,$65,$FE,$B5,$E0,$75,$3B,$77
	.dc.b	$5A,$FE,$C5,$16,$7F,$B0,$4D,$CF,$DF,$F1,$7E,$53,$CD,$E1,$5C,$3F
	.dc.b	$B5,$45,$BF,$ED,$13,$6B,$E3,$FC,$5F,$B4,$F7,$7D,$05,$C3,$FB,$94
	.dc.b	$5B,$FE,$E1,$36,$BF,$C9,$1F,$7C,$F6,$FE,$5D,$23,$18,$7F,$0A,$AF
	.dc.b	$AD,$F7,$EE,$FD,$D6,$AF,$FC,$94,$D7,$FC,$B2,$D7,$FD,$2A,$D7,$F8
	.dc.b	$AB,$57,$D6,$BE,$0B,$C6,$DC,$7C,$47,$8F,$E6,$DA,$EA,$A4,$FF,$A7
	.dc.b	$EB,$85,$FF,$A9,$5A,$FF,$9B,$1A,$F3,$F9,$FF,$C5,$E8,$C5,$18,$5F
	.dc.b	$3F,$62,$61,$E4,$BF,$F8,$03,$5C,$23,$17,$CE,$F7,$5F,$ED,$38,$ED
	.dc.b	$78,$7F,$45,$7D,$AF,$8B,$CE,$BF,$CE,$DF,$32,$F1,$FE,$BB,$6B,$7F
	.dc.b	$2D,$89,$91,$FE,$C1,$9D,$57,$FB,$1D,$AF,$FA,$39,$AF,$EB,$F9,$5F
	.dc.b	$13,$FA,$B7,$FC,$53,$5E,$FE,$B4,$6A,$30,$EA,$BF,$DD,$0E,$5B,$E1
	.dc.b	$F3,$BE,$8F,$CD,$F9,$29,$7F,$03,$E5,$6D,$FC,$7D,$55,$8D,$BE,$3D
	.dc.b	$5B,$FC,$24,$DB,$FA,$73,$73,$FF,$22,$7C,$EF,$F0,$33,$F8,$27,$CE
	.dc.b	$DC,$3F,$CB,$11,$6F,$FC,$AD,$37,$1F,$F5,$77,$CD,$7E,$42,$FF,$01
	.dc.b	$48,$C3,$F7,$7E,$6F,$FD,$37,$3A,$9C,$D7,$2E,$7F,$F5,$C5,$FF,$BE
	.dc.b	$27,$8B,$FE,$C0,$6B,$98,$DB,$FC,$7F,$93,$F8,$E1,$FC,$77,$96,$E1
	.dc.b	$FE,$71,$8B,$3F,$E6,$F9,$B8,$7F,$2E,$9B,$5F,$CB,$7C,$2F,$C8,$5F
	.dc.b	$F9,$D6,$7F,$94,$73,$7F,$29,$E0,$B7,$FE,$7D,$6D,$7F,$9E,$DB,$87
	.dc.b	$F3,$A7,$F9,$87,$07,$F3,$1F,$96,$B7,$FE,$85,$6D,$FF,$A0,$DB,$7F
	.dc.b	$D0,$9F,$D8,$12,$7F,$57,$EE,$3C,$BC,$5F,$E1,$6A,$DF,$F4,$BA,$D5
	.dc.b	$FE,$AA,$72,$FF,$09,$7F,$7F,$1A,$AE,$53,$F9,$DF,$D4,$31,$7F,$E8
	.dc.b	$C6,$C7,$F5,$35,$E6,$7F,$4F,$73,$5F,$D7,$7A,$5E,$EF,$EA,$4F,$D5
	.dc.b	$78,$57,$0E,$DF,$2D,$C4,$E5,$F3,$16,$3C,$5F,$29,$D4,$ED,$F2,$B6
	.dc.b	$3C,$DE,$87,$84,$F1,$78,$16,$3D,$5E,$07,$94,$F3,$7B,$56,$3D,$DD
	.dc.b	$3E,$93,$D5,$F2,$B6,$3F,$7D,$F4,$8F,$8D,$FE,$D8,$D8,$CF,$C6,$E2
	.dc.b	$7E,$17,$93,$F0,$FE,$5A,$D7,$CB,$7E,$37,$CA,$FC,$7F,$95,$B3,$F3
	.dc.b	$DF,$95,$F2,$D2,$3E,$97,$C2,$5E,$97,$67,$D3,$F8,$8F,$A0,$7D,$1E
	.dc.b	$AB,$7E,$8F,$3B,$CF,$F0,$7E,$86,$C7,$B3,$E0,$FD,$66,$31,$EB,$1F
	.dc.b	$E0,$FE,$24,$FC,$0F,$4E,$4F,$13,$E2,$BF,$81,$EE,$91,$FC,$97,$F1
	.dc.b	$7F,$F3,$8B,$F8,$B8,$FF,$0A,$43,$E5,$E9,$C3,$1F,$09,$FE,$25,$8D
	.dc.b	$57,$39,$E3,$F0,$5C,$23,$B3,$FE,$1F,$F8,$87,$C2,$53,$1F,$0B,$6B
	.dc.b	$F9,$C4,$E9,$B3,$A6,$93,$1F,$1F,$AA,$AE,$11,$85,$7D,$63,$79,$3E
	.dc.b	$5F,$DE,$7C,$0F,$8E,$B8,$FC,$58,$B5,$C2,$6F,$2F,$FB,$EA,$DF,$2F
	.dc.b	$F2,$0F,$93,$F0,$B7,$1F,$9D,$1F,$33,$E5,$F1,$8F,$89,$F9,$FF,$2E
	.dc.b	$AF,$F4,$F7,$95,$FE,$4B,$C5,$F2,$59,$C3,$FD,$19,$39,$BC,$7C,$52
	.dc.b	$38,$56,$1D,$61,$D2,$FE,$C4,$E5,$FB,$57,$78,$58,$F8,$2F,$85,$3E
	.dc.b	$6F,$E5,$F2,$E1,$B1,$E4,$3F,$D7,$7C,$2F,$17,$B4,$71,$E3,$FC,$63
	.dc.b	$B3,$17,$CE,$33,$B7,$AE,$79,$D7,$2C,$3B,$CE,$67,$89,$CE,$33,$B7
	.dc.b	$FB,$32,$3E,$F5,$71,$BE,$B3,$9A,$DD,$7F,$B4,$9A,$F0,$CB,$C9,$8C
	.dc.b	$8E,$D7,$C9,$F2,$CE,$71,$F2,$5F,$C4,$9F,$11,$E6,$2B,$97,$C5,$5F
	.dc.b	$11,$DF,$C6,$3C,$97,$CE,$72,$9F,$F7,$3C,$E6,$31,$08,$C3,$B6,$BE
	.dc.b	$0F,$8C,$FE,$A8,$F3,$C3,$FD,$27,$EB,$F2,$6C,$66,$76,$F1,$9F,$C0
	.dc.b	$47,$2C,$BD,$ED,$D7,$E6,$7B,$3F,$ED,$07,$E7,$46,$6F,$E7,$35,$FC
	.dc.b	$63,$70,$FF,$14,$C7,$6C,$23,$F7,$67,$57,$D6,$39,$CE,$B0,$F1,$BF
	.dc.b	$95,$9D,$4F,$68,$D6,$1E,$47,$51,$FA,$5E,$27,$CE,$3D,$9E,$D0,$F9
	.dc.b	$CF,$87,$97,$E5,$63,$18,$6A,$F8,$61,$CD,$AF,$E9,$67,$6F,$7B,$E7
	.dc.b	$8F,$FB,$26,$B7,$87,$59,$FE,$F3,$5F,$E5,$B5,$EB,$7E,$77,$B3,$F8
	.dc.b	$26,$3E,$58,$C7,$1F,$FB,$4E,$B8,$DF,$F2,$66,$DF,$07,$9C,$75,$AE
	.dc.b	$2F,$FB,$86,$37,$5C,$FD,$6C,$E1,$F0,$DB,$26,$FA,$D8,$65,$FA,$5C
	.dc.b	$BF,$E7,$5F,$9C,$FC,$6E,$5E,$A6,$E3,$FE,$EB,$6C,$FF,$BC,$DF,$1B
	.dc.b	$76,$F1,$C6,$B2,$34,$DB,$88,$D9,$27,$C5,$ED,$7A,$BD,$8F,$5F,$B1
	.dc.b	$27,$E8,$46,$5A,$43,$E6,$7B,$1F,$08,$D3,$F3,$E3,$0F,$BF,$5F,$93
	.dc.b	$DD,$FE,$17,$8F,$EE,$5E,$B7,$AA,$7D,$7B,$C9,$F8,$CF,$BA,$FD,$F8
	.dc.b	$D9,$F9,$51,$87,$F0,$DE,$C7,$F0,$4D,$7F,$30,$DE,$13,$4C,$7A,$FF
	.dc.b	$68,$95,$F8,$FF,$5F,$EE,$FA,$FF,$7B,$D7,$93,$F9,$38,$FC,$0F,$AF
	.dc.b	$F8,$2F,$5E,$4F,$E5,$E3,$EA,$CC,$8F,$C0,$3F,$62,$BF,$13,$EB,$C9
	.dc.b	$FC,$F3,$F8,$CF,$5F,$F1,$BE,$94,$9F,$D1,$3F,$92,$F9,$BF,$E5,$BD
	.dc.b	$7F,$A2,$7F,$95,$37,$C8,$D1,$BF,$4F,$87,$97,$F4,$BC,$B9,$8F,$FB
	.dc.b	$1F,$0F,$F0,$23,$7C,$FF,$AF,$F3,$92,$30,$7F,$B3,$7E,$5F,$FC,$02
	.dc.b	$D9,$FD,$2C,$62,$3F,$C2,$AB,$D2,$F4,$7A,$52,$7C,$0F,$AF,$D1,$FD
	.dc.b	$8C,$7F,$80,$3E,$72,$FA,$66,$98,$F9,$CB,$3F,$EB,$E9,$D3,$67,$4D
	.dc.b	$21,$FE,$3A,$3B,$1C,$67,$E6,$4F,$CD,$F0,$A4,$FC,$58,$CB,$F1,$E3
	.dc.b	$2F,$F8,$CB,$5F,$F0,$BB,$F2,$F6,$FA,$DB,$48,$C1,$0E,$0E,$5E,$51
	.dc.b	$F7,$71,$91,$F5,$5F,$C0,$63,$88,$F0,$45,$E2,$FC,$14,$DF,$21,$4E
	.dc.b	$F6,$5F,$1A,$38,$9F,$81,$AF,$A3,$F3,$8F,$77,$E8,$F0,$CA,$FF,$75
	.dc.b	$CE,$5B,$3F,$DC,$94,$65,$FE,$F0,$D7,$FC,$06,$D7,$CB,$87,$93,$E7
	.dc.b	$3F,$38,$48,$FC,$A7,$C1,$57,$E7,$45,$D3,$F4,$13,$F8,$8F,$9C,$76
	.dc.b	$52,$5F,$C2,$AF,$F2,$52,$9F,$F6,$0B,$9D,$BF,$C5,$11,$CA,$30,$7C
	.dc.b	$A7,$0F,$E2,$5B,$3D,$1E,$71,$9A,$F0,$B4,$E6,$78,$7B,$3F,$4F,$F3
	.dc.b	$5F,$A8,$D6,$4F,$EB,$63,$2F,$EB,$E3,$2F,$EB,$23,$F4,$FF,$38,$F9
	.dc.b	$5F,$85,$EA,$4A,$EF,$39,$6C,$FD,$78,$C4,$7F,$8B,$5A,$FF,$8E,$1A
	.dc.b	$FF,$92,$FD,$E7,$CB,$FE,$D9,$B1,$EC,$6B,$FA,$D5,$F0,$DB,$99,$3B
	.dc.b	$79,$46,$23,$84,$5D,$BE,$BD,$72,$9F,$93,$EF,$BC,$8F,$E0,$F8,$7D
	.dc.b	$0B,$FB,$AD,$E1,$FF,$31,$56,$EF,$8D,$FE,$5E,$31,$1F,$E4,$D6,$BF
	.dc.b	$E5,$86,$BF,$E6,$BF,$9C,$FF,$2A,$3F,$6B,$D3,$F2,$3F,$C2,$DB,$8F
	.dc.b	$E3,$A7,$73,$FC,$3D,$AF,$F9,$91,$AF,$5F,$80,$FD,$54,$62,$3F,$71
	.dc.b	$D6,$E1,$FE,$80,$48,$FE,$71,$C5,$5F,$E2,$51,$70,$FE,$89,$3F,$D1
	.dc.b	$FE,$1A,$57,$F3,$F7,$FA,$57,$C2,$E2,$3F,$8D,$45,$AF,$E9,$93,$FD
	.dc.b	$3F,$E1,$65,$7F,$47,$7F,$C4,$7E,$8F,$EB,$2B,$FA,$94,$6A,$7F,$92
	.dc.b	$3A,$E0,$DF,$CE,$49,$FD,$6F,$6F,$F7,$48,$E7,$F0,$D9,$9F,$F6,$EE
	.dc.b	$29,$C6,$3F,$E4,$08,$FF,$75,$CF,$27,$FD,$BE,$DA,$E6,$E7,$8B,$CE
	.dc.b	$61,$EB,$18,$7F,$6B,$F0,$8E,$36,$1F,$CA,$BE,$73,$FA,$D5,$7F,$2C
	.dc.b	$F0,$9E,$77,$FA,$BE,$73,$3F,$D7,$E3,$13,$FE,$B3,$C3,$A5,$F3,$8C
	.dc.b	$61,$DA,$D8,$72,$9E,$B1,$B3,$C1,$87,$47,$FD,$47,$5C,$6B,$C3,$87
	.dc.b	$69,$E7,$7F,$EC,$33,$9B,$83,$99,$D4,$FF,$55,$57,$38,$F2,$E1,$97
	.dc.b	$8D,$F9,$70,$F1,$5F,$0A,$E5,$87,$86,$B5,$3C,$6B,$73,$9B,$EB,$3E
	.dc.b	$33,$E4,$1B,$39,$E1,$E3,$7D,$F8,$65,$E1,$3E,$43,$9E,$1F,$0A,$B8
	.dc.b	$4F,$43,$78,$78,$A3,$38,$78,$E7,$E1,$9A,$3C,$38,$47,$04,$BE,$D1
	.dc.b	$97,$DF,$19,$C3,$C6,$47,$AC,$68,$E2,$FC,$1A,$F1,$3E,$78,$D4,$F1
	.dc.b	$8F,$9C,$75,$39,$B9,$9F,$C0,$7B,$BF,$EA,$06,$FB,$BF,$3A,$31,$8F
	.dc.b	$E3,$DF,$F5,$2D,$FF,$81,$BE,$03,$FC,$0F,$18,$84,$62,$BF,$DE,$4D
	.dc.b	$8F,$F7,$EB,$5F,$F0,$79,$C0,$FB,$99,$3D,$A7,$17,$E3,$E7,$C0,$6A
	.dc.b	$32,$FE,$0C,$CE,$9F,$AC,$AE,$55,$C6,$7B,$4E,$6B,$8B,$FB,$28,$E3
	.dc.b	$A7,$FB,$1E,$72,$FA,$A3,$C1,$5C,$F0,$FD,$16,$1D,$A7,$C4,$76,$AE
	.dc.b	$55,$B8,$E2,$6A,$B8,$DF,$68,$F8,$A6,$B0,$F2,$3C,$CF,$F0,$A4,$6D
	.dc.b	$F9,$33,$CA,$BB,$57,$B5,$B7,$E2,$AF,$01,$D5,$F6,$4E,$E3,$2D,$3C
	.dc.b	$35,$D6,$33,$3C,$E7,$31,$CC,$F4,$BC,$63,$A9,$B2,$C7,$AA,$BC,$AD
	.dc.b	$3A,$96,$7A,$9C,$DE,$B3,$CD,$FF,$2B,$70,$3E,$13,$F0,$79,$7F,$3F
	.dc.b	$F9,$BB,$8F,$F9,$22,$B4,$FF,$41,$8C,$DF,$2B,$BC,$9F,$EC,$09,$B7
	.dc.b	$C2,$BF,$A6,$C6,$3E,$72,$61,$CC,$FE,$98,$E2,$7F,$A9,$BC,$A7,$18
	.dc.b	$FF,$53,$F0,$7A,$9D,$BC,$0B,$87,$F5,$58,$B8,$7F,$5C,$9B,$5F,$D6
	.dc.b	$FE,$73,$CA,$79,$BC,$4B,$87,$F5,$D8,$B7,$FD,$92,$6D,$7F,$63,$F9
	.dc.b	$2F,$69,$EE,$F9,$CB,$87,$F6,$58,$B7,$FD,$B2,$6D,$7F,$99,$3E,$6D
	.dc.b	$ED,$FF,$99,$AB,$F7,$71,$8C,$3F,$7B,$18,$9D,$3F,$E8,$E9,$D3,$A0
	.dc.b	$E2,$BE,$DF,$CC,$FF,$02,$BF,$F3,$1C,$E5,$FF,$3B,$46,$6B,$FD,$6D
	.dc.b	$F3,$6F,$82,$EF,$08,$C5,$77,$8C,$4E,$A7,$31,$E3,$6E,$1E,$2C,$BF
	.dc.b	$0F,$3C,$2F,$FC,$FF,$2C,$15,$F8,$C8,$C7,$CA,$D9,$37,$E5,$B0,$CB
	.dc.b	$E7,$3A,$6C,$E7,$0F,$F0,$6C,$7C,$0A,$C5,$FC,$6A,$B3,$FC,$B3,$E6
	.dc.b	$DE,$37,$C2,$6C,$7C,$16,$E1,$F4,$5A,$FD,$87,$37,$FA,$CF,$79,$F0
	.dc.b	$7F,$0A,$73,$7F,$EA,$79,$CB,$FE,$36,$E4,$7C,$3F,$EB,$1F,$F1,$34
	.dc.b	$66,$31,$E1,$65,$B1,$93,$94,$7C,$9F,$8F,$91,$F1,$BE,$5A,$E9,$F1
	.dc.b	$0F,$F1,$8F,$02,$B3,$AF,$E2,$AA,$DF,$E2,$26,$DF,$09,$BD,$AE,$D7
	.dc.b	$CD,$FF,$02,$3F,$81,$FC,$DD,$C7,$FC,$95,$17,$0F,$F2,$8C,$DC,$BF
	.dc.b	$D9,$3F,$34,$EB,$FC,$9B,$70,$FC,$93,$6F,$F4,$4D,$7F,$52,$E6,$BF
	.dc.b	$8E,$FC,$F3,$E0,$BF,$27,$39,$BF,$F6,$C4,$E5,$FF,$2F,$74,$7F,$1A
	.dc.b	$3F,$8D,$F4,$58,$C9,$31,$FF,$34,$C6,$5F,$F3,$5C,$DC,$7F,$CD,$53
	.dc.b	$6B,$F9,$67,$CD,$FF,$26,$E1,$FE,$4E,$D8,$FF,$37,$37,$0F,$F3,$AB
	.dc.b	$6F,$FC,$E9,$C3,$FC,$BB,$E5,$7F,$97,$B6,$FF,$CE,$CD,$BF,$F3,$EB
	.dc.b	$6B,$F8,$2D,$7E,$B4,$FD,$87,$BE,$F3,$31,$87,$EB,$26,$E1,$F8,$7B
	.dc.b	$5E,$EB,$44,$9D,$72,$AF,$AE,$1C,$B0,$FF,$08,$F2,$BE,$6F,$96,$1F
	.dc.b	$93,$8C,$57,$F4,$C9,$79,$B6,$3F,$49,$EF,$4C,$FF,$51,$E2,$5C,$79
	.dc.b	$60,$F9,$5F,$EA,$BF,$28,$F8,$6E,$91,$87,$B7,$C9,$EB,$E1,$EE,$5F
	.dc.b	$EA,$8B,$47,$8F,$D3,$93,$FE,$AC,$C7,$F6,$ED,$8F,$47,$81,$23,$5E
	.dc.b	$FD,$D3,$CB,$E0,$FB,$4F,$77,$CD,$DA,$FF,$5B,$FC,$DB,$E2,$FF,$B6
	.dc.b	$B6,$33,$F1,$78,$9F,$8D,$E7,$7C,$7F,$94,$B5,$E5,$7E,$57,$99,$F2
	.dc.b	$FE,$4E,$CF,$A5,$F9,$DF,$29,$23,$D9,$F0,$F7,$6F,$97,$F1,$7F,$D4
	.dc.b	$BB,$7E,$C3,$79,$0D,$35,$BF,$F1,$EA,$52,$1F,$0B,$EC,$F4,$BD,$AF
	.dc.b	$BB,$E1,$66,$7F,$0A,$2E,$1F,$85,$B7,$C2,$FD,$CF,$8D,$F8,$AD,$BF
	.dc.b	$A9,$DF,$F8,$0F,$94,$FC,$53,$63,$EE,$45,$C3,$F0,$F3,$7C,$4F,$CA
	.dc.b	$F7,$2B,$F1,$3F,$09,$9C,$18,$7B,$2B,$FD,$AB,$1F,$EE,$EF,$8B,$7C
	.dc.b	$37,$F8,$91,$B1,$9F,$87,$C4,$C8,$E0,$EE,$31,$7F,$A3,$AE,$2D,$9D
	.dc.b	$57,$03,$F2,$DF,$29,$F9,$E3,$F3,$FF,$37,$63,$24,$AF,$D7,$62,$9D
	.dc.b	$46,$3E,$57,$F4,$FF,$2B,$FA,$86,$DF,$E8,$DB,$3B,$8F,$D3,$4F,$EB
	.dc.b	$BE,$57,$FD,$1B,$F1,$6F,$92,$F8,$FD,$2C,$DF,$2B,$A9,$CC,$2F,$19
	.dc.b	$7C,$DA,$B7,$C6,$6D,$72,$CB,$FD,$59,$87,$29,$CD,$F2,$C1,$F3,$4E
	.dc.b	$1F,$B9,$1F,$EB,$0B,$FF,$55,$CE,$78,$F9,$CF,$DE,$E0,$78,$BC,$EB
	.dc.b	$53,$C4,$FC,$1B,$8D,$BE,$04,$63,$1E,$F1,$88,$5F,$4D,$3B,$C7,$9A
	.dc.b	$B7,$97,$91,$B1,$CF,$2D,$DF,$18,$F1,$1C,$63,$9E,$3E,$3B,$E5,$1C
	.dc.b	$32,$F5,$61,$D2,$F8,$1C,$AF,$83,$F0,$67,$75,$D2,$BE,$4C,$FF,$95
	.dc.b	$AB,$E1,$B6,$32,$F8,$0F,$F2,$B7,$C7,$BC,$97,$CD,$FF,$92,$E3,$0D
	.dc.b	$8C,$CF,$06,$BD,$FE,$3B,$FA,$B3,$FE,$61,$C3,$FD,$35,$E9,$FF,$95
	.dc.b	$5B,$19,$9D,$BC,$67,$E2,$C7,$2D,$FE,$A5,$70,$6F,$B5,$FE,$C6,$DB
	.dc.b	$F0,$B5,$B6,$DF,$23,$E7,$CE,$F0,$FC,$1D,$6D,$B3,$B7,$E7,$4E,$DF
	.dc.b	$CA,$CE,$DB,$3B,$BF,$1C,$5B,$F0,$BF,$5E,$B8,$61,$FE,$39,$8F,$CC
	.dc.b	$F2,$7E,$BF,$D8,$F2,$36,$33,$CB,$FD,$E9,$BF,$25,$A8,$C4,$7E,$76
	.dc.b	$78,$36,$33,$97,$E7,$A7,$F7,$95,$C7,$C2,$FE,$4B,$7F,$C3,$9B,$DD
	.dc.b	$FC,$B1,$BC,$DF,$C0,$3C,$4B,$5F,$CA,$E6,$E1,$FC,$1D,$B9,$7F,$48
	.dc.b	$9F,$EE,$95,$C1,$7E,$06,$F4,$70,$D8,$FE,$6F,$93,$2F,$2C,$E6,$7F
	.dc.b	$91,$F8,$DC,$2E,$47,$0A,$DC,$78,$A7,$C1,$F7,$9F,$C5,$EB,$FD,$48
	.dc.b	$D7,$FD,$68,$D7,$FD,$76,$D7,$E2,$BB,$C3,$C9,$1F,$D3,$23,$17,$AE
	.dc.b	$DF,$8B,$CE,$E0,$7A,$3E,$29,$98,$79,$CE,$4F,$09,$DC,$66,$BF,$98
	.dc.b	$FC,$57,$4E,$FF,$F5,$8C,$E5,$AF,$FA,$14,$65,$F4,$5C,$37,$3F,$4C
	.dc.b	$6B,$87,$E2,$9E,$8F,$C7,$5B,$53,$33,$E7,$38,$1E,$58,$DB,$EB,$E3
	.dc.b	$7C,$7F,$F2,$9C,$3F,$C9,$2D,$E0,$FE,$9A,$DA,$FE,$1B,$9F,$FB,$21
	.dc.b	$ED,$9F,$FA,$F1,$BC,$7F,$68,$F9,$BD,$EF,$95,$A4,$9F,$91,$19,$7F
	.dc.b	$C8,$9E,$57,$E5,$BF,$2F,$F8,$3A,$7F,$97,$1B,$A7,$F6,$06,$D7,$F8
	.dc.b	$89,$E4,$FC,$DE,$D7,$C1,$7E,$6D,$B1,$F1,$5A,$FF,$AD,$F9,$1F,$80
	.dc.b	$7F,$83,$A3,$E3,$5D,$7F,$CD,$4D,$C7,$C9,$E5,$79,$FC,$5F,$D6,$F8
	.dc.b	$0F,$92,$A4,$C3,$E8,$47,$E0,$AF,$EE,$E7,$FC,$36,$7D,$1E,$27,$9C
	.dc.b	$FC,$AE,$6F,$94,$FD,$CF,$D8,$FF,$0A,$C3,$FC,$FE,$DE,$0F,$F0,$1B
	.dc.b	$70,$FF,$22,$BF,$8B,$8F,$91,$E2,$FC,$D2,$4F,$BA,$FF,$80,$7D,$E3
	.dc.b	$E0,$29,$27,$F3,$DF,$09,$FB,$DC,$BE,$A7,$CB,$FD,$72,$4F,$C8,$F9
	.dc.b	$7F,$B5,$F2,$72,$3E,$5B,$F7,$7B,$BB,$FD,$04,$9F,$CC,$66,$8F,$E2
	.dc.b	$FF,$7B,$DF,$7F,$C0,$72,$FE,$14,$97,$F5,$BE,$4B,$F8,$C7,$D0,$7F
	.dc.b	$83,$23,$F9,$39,$FC,$A1,$C3,$80,$E1,$DB,$5E,$1E,$9F,$42,$3F,$AD
	.dc.b	$F0,$FE,$73,$CA,$32,$FE,$F6,$7B,$F0,$7E,$EA,$BE,$A7,$88,$ED,$FC
	.dc.b	$16,$7F,$9B,$F2,$3F,$40,$7E,$87,$E5,$6E,$5F,$92,$8A,$FE,$4A,$3F
	.dc.b	$4D,$F2,$B2,$3F,$2E,$FE,$A7,$E4,$B1,$1F,$98,$8B,$9F,$EA,$E7,$F9
	.dc.b	$F6,$7F,$EB,$9A,$F6,$7C,$66,$0C,$FC,$65,$8C,$90,$FE,$35,$3F,$EB
	.dc.b	$A7,$17,$EF,$AE,$11,$AA,$FD,$E7,$C9,$48,$FD,$6B,$FB,$FF,$1B,$11
	.dc.b	$FA,$E8,$B7,$FC,$06,$7F,$82,$78,$32,$BF,$6A,$FF,$07,$F9,$2C,$47
	.dc.b	$ED,$A2,$D7,$F0,$99,$FC,$2F,$9F,$F8,$1A,$FF,$47,$F1,$FF,$31,$7F
	.dc.b	$03,$19,$7F,$0A,$D8,$FD,$54,$70,$8E,$6F,$E0,$FD,$EA,$FE,$06,$2D
	.dc.b	$FE,$06,$7F,$37,$F0,$D2,$BF,$99,$74,$A3,$F3,$FF,$0B,$99,$FF,$7C
	.dc.b	$4E,$5D,$CE,$9F,$E7,$DF,$3F,$FE,$71,$6F,$95,$FD,$E6,$31,$1F,$D3
	.dc.b	$FE,$81,$F1,$99,$A9,$3C,$0F,$E9,$13,$A7,$AF,$77,$C3,$BF,$E3,$91
	.dc.b	$F2,$5A,$FE,$AD,$AF,$EC,$1C,$E9,$FD,$69,$BD,$5F,$E1,$A6,$CF,$FB
	.dc.b	$02,$DE,$2F,$A7,$09,$3E,$23,$CB,$1F,$E1,$5C,$7F,$AA,$47,$FA,$47
	.dc.b	$CB,$C1,$FE,$19,$F8,$8F,$EB,$F1,$FD,$BE,$76,$49,$D7,$AD,$FE,$1F
	.dc.b	$FB,$CF,$E7,$9A,$7F,$8B,$9B,$7F,$ED,$46,$AF,$9A,$7C,$CF,$1F,$4A
	.dc.b	$4F,$F7,$1A,$DD,$6A,$3F,$A5,$FC,$B7,$F5,$D2,$3F,$DC,$F6,$FF,$35
	.dc.b	$F0,$FF,$5F,$9C,$CF,$F5,$07,$FC,$73,$3A,$CB,$FD,$82,$FC,$89,$E5
	.dc.b	$BE,$8D,$FC,$AD,$8C,$90,$61,$FE,$26,$F1,$BF,$AC,$4F,$6D,$E5,$79
	.dc.b	$5F,$0F,$9F,$88,$E3,$97,$8A,$71,$87,$3C,$BF,$A5,$57,$B7,$CA,$F4
	.dc.b	$91,$F8,$1F,$2B,$5E,$63,$3F,$ED,$53,$9B,$E9,$5F,$08,$95,$C8,$F9
	.dc.b	$5F,$35,$87,$9D,$FF,$73,$9C,$57,$96,$BF,$CB,$DF,$11,$FE,$5D,$85
	.dc.b	$BE,$7D,$F3,$FF,$E7,$01,$F1,$FF,$E7,$28,$CC,$F0,$9C,$BC,$A3,$2A
	.dc.b	$FB,$3D,$67,$FC,$F7,$18,$9D,$C7,$29,$D1,$D6,$BC,$11,$C3,$E1,$7E
	.dc.b	$CC,$E8,$EC,$E3,$8B,$9D,$7F,$A2,$1B,$3E,$E7,$CB,$87,$91,$B5,$D3
	.dc.b	$0D,$D7,$5C,$3C,$91,$AC,$7D,$D1,$E3,$6C,$F9,$CA,$7C,$AC,$7D,$2F
	.dc.b	$CE,$C7,$E0,$57,$18,$FE,$02,$FC,$F9,$F4,$36,$3D,$0E,$F1,$F1,$47
	.dc.b	$96,$7A,$BD,$CF,$0D,$76,$7C,$18,$FB,$6F,$A4,$71,$37,$19,$77,$87
	.dc.b	$AA,$7E,$09,$C3,$1F,$09,$1C,$12,$78,$9F,$56,$FC,$4F,$94,$F0,$19
	.dc.b	$BF,$03,$F6,$AF,$C1,$87,$B2,$FC,$33,$D3,$0F,$03,$F3,$CE,$6E,$1E
	.dc.b	$46,$23,$C8,$E2,$3E,$4F,$C5,$7F,$09,$8D,$B7,$1F,$CB,$B7,$1F,$C7
	.dc.b	$BB,$9F,$D3,$FC,$27,$EA,$23,$11,$FE,$AB,$0E,$23,$59,$FF,$8C,$5B
	.dc.b	$3F,$E3,$67,$F0,$F9,$B5,$FF,$5C,$E5,$FA,$AA,$EE,$7E,$7E,$75,$C5
	.dc.b	$F8,$E8,$CC,$FE,$C2,$73,$1F,$A5,$33,$A7,$FB,$A2,$FA,$BE,$38,$E9
	.dc.b	$18,$7C,$86,$2F,$B3,$ED,$BE,$D3,$B8,$FF,$7A,$E1,$EA,$8C,$47,$80
	.dc.b	$FF,$73,$B8,$79,$5F,$8E,$FA,$D7,$16,$CF,$84,$F5,$1C,$0E,$6E,$AF
	.dc.b	$C6,$DA,$E9,$3C,$2B,$A3,$4F,$D7,$9C,$8A,$7A,$8B,$87,$CA,$C7,$E9
	.dc.b	$47,$84,$FE,$93,$8F,$AE,$BA,$E3,$A6,$E1,$F3,$63,$CB,$5E,$1B,$A6
	.dc.b	$DE,$F7,$FE,$F8,$8C,$4F,$02,$D7,$B1,$B5,$E9,$6B,$E4,$C7,$F8,$9E
	.dc.b	$FF,$EB,$3D,$11,$FD,$BD,$B3,$FD,$1B,$85,$F1,$78,$D7,$FA,$D5,$5B
	.dc.b	$FE,$DE,$DB,$FE,$AD,$68,$C6,$1F,$EC,$1C,$2C,$E6,$FF,$BF,$7C,$C6
	.dc.b	$0C,$DD,$8C,$93,$0F,$24,$65,$F3,$7C,$47,$43,$AF,$CB,$DA,$FF,$05
	.dc.b	$45,$9F,$64,$DC,$FF,$BA,$78,$7E,$43,$CB,$F2,$F7,$0F,$F0,$94,$5B
	.dc.b	$F8,$93,$6B,$FB,$E7,$CB,$FB,$0F,$6F,$CB,$DC,$3F,$C3,$51,$6F,$E6
	.dc.b	$4D,$AF,$9F,$0E,$CF,$F6,$E5,$FF,$70,$9C,$C6,$27,$FC,$07,$0F,$9A
	.dc.b	$C3,$AF,$81,$FA,$53,$CD,$B5,$AB,$E9,$76,$BF,$BE,$CD,$5F,$3A,$F9
	.dc.b	$5E,$6D,$BF,$21,$E5,$FA,$56,$FA,$A9,$3A,$BF,$F0,$4D,$F1,$7D,$37
	.dc.b	$C2,$75,$1E,$0A,$EF,$F1,$B2,$79,$69,$EF,$F5,$71,$7F,$E2,$43,$CF
	.dc.b	$58,$9F,$CB,$68,$AC,$5F,$F8,$6A,$F6,$FF,$A7,$F2,$61,$F4,$A7,$51
	.dc.b	$89,$D4,$E6,$7F,$C4,$7F,$3C,$F8,$7F,$9E,$DA,$F8,$1D,$2F,$B6,$79
	.dc.b	$5A,$B9,$5F,$02,$F8,$61,$FE,$34,$BD,$CE,$6B,$F5,$3F,$72,$F9,$7A
	.dc.b	$63,$E5,$ED,$78,$27,$4D,$9D,$34,$F0,$7C,$37,$A0,$FD,$C7,$0B,$8F
	.dc.b	$FD,$61,$3A,$26,$7F,$3F,$45,$F0,$AD,$CE,$63,$A5,$7F,$B2,$EB,$6D
	.dc.b	$8F,$95,$29,$CB,$F5,$4F,$E1,$3E,$AF,$F0,$63,$F8,$3F,$AF,$6F,$FD
	.dc.b	$09,$16,$7F,$8B,$4D,$C7,$FC,$DD,$F3,$FF,$C3,$8F,$E1,$FF,$39,$70
	.dc.b	$FF,$46,$45,$BF,$E3,$D3,$6B,$F1,$3E,$87,$C4,$E7,$3F,$1B,$6F,$EF
	.dc.b	$38,$FE,$52,$72,$D9,$E0,$6F,$CA,$B1,$92,$63,$FB,$AB,$E6,$FE,$F2
	.dc.b	$F9,$BF,$E7,$6F,$CF,$BF,$CE,$35,$F8,$C6,$C6,$7E,$53,$F9,$6C,$63
	.dc.b	$2F,$F4,$C3,$6B,$FC,$F3,$CE,$8C,$5F,$FA,$2D,$5F,$28,$E7,$E7,$4B
	.dc.b	$F0,$7C,$DD,$9F,$24,$DC,$7B,$CD,$AF,$D7,$71,$3E,$07,$F4,$A7,$9B
	.dc.b	$DE,$B3,$FA,$18,$B8,$FB,$E6,$E1,$FC,$92,$F9,$5F,$E2,$EF,$F1,$31
	.dc.b	$8B,$FC,$41,$F8,$B8,$FD,$B6,$9E,$FE,$FF,$8A,$7C,$6F,$36,$E1,$FC
	.dc.b	$43,$EE,$1A,$7F,$28,$6C,$7E,$EB,$ED,$BF,$6B,$A7,$F3,$69,$CE,$7F
	.dc.b	$57,$E7,$BE,$89,$F4,$BE,$72,$E1,$F5,$FE,$27,$EA,$9F,$5B,$E8,$AC
	.dc.b	$78,$78,$7C,$07,$F3,$8F,$9A,$7C,$06,$27,$F0,$1F,$23,$F7,$8F,$BF
	.dc.b	$F0,$36,$3F,$87,$EB,$FC,$B7,$6F,$F6,$C3,$63,$F0,$2F,$FA,$E1,$DF
	.dc.b	$17,$BE,$E4,$7F,$32,$FC,$05,$EF,$E3,$87,$FD,$8F,$F1,$2F,$1F,$FB
	.dc.b	$DB,$6F,$F8,$FA,$F8,$9C,$CF,$E5,$3E,$36,$57,$E6,$7E,$36,$E1,$FC
	.dc.b	$E7,$56,$FF,$44,$FE,$B3,$A7,$F9,$5B,$7F,$A1,$1F,$9E,$FA,$29,$8F
	.dc.b	$E8,$FC,$2B,$8F,$FB,$9B,$E7,$03,$E1,$3F,$C2,$8E,$1F,$D3,$ED,$88
	.dc.b	$D5,$FF,$52,$FB,$97,$BF,$FD,$A9,$B1,$FD,$6E,$79,$3B,$9F,$F0,$EF
	.dc.b	$CB,$BD,$AA,$78,$3D,$F7,$C3,$E2,$F9,$60,$F8,$1E,$DE,$97,$8D,$B3
	.dc.b	$E3,$8B,$B7,$B6,$B8,$4E,$4F,$26,$1F,$E2,$B9,$F8,$4D,$E2,$FF,$06
	.dc.b	$EB,$FC,$B9,$FF,$1E,$DF,$F8,$D6,$7F,$C0,$FF,$73,$FC,$4A,$7F,$C2
	.dc.b	$6D,$C3,$FE,$00,$D2,$7F,$B2,$61,$FD,$C6,$B4,$83,$86,$F9,$DF,$4A
	.dc.b	$78,$9D,$9F,$09,$FD,$12,$7F,$91,$CF,$F4,$36,$BF,$DE,$5A,$F7,$FA
	.dc.b	$2F,$B2,$FF,$80,$DB,$5F,$C8,$BA,$BF,$C4,$71,$8C,$7F,$C0,$4D,$AE
	.dc.b	$5C,$7F,$CA,$AB,$6F,$F2,$CF,$6A,$5F,$F3,$1F,$02,$DF,$F3,$59,$B8
	.dc.b	$FF,$30,$9B,$5F,$CC,$3C,$F7,$85,$FE,$6C,$F8,$8F,$D6,$E3,$FC,$C6
	.dc.b	$31,$E0,$FF,$36,$FD,$47,$FC,$E7,$87,$FB,$2A,$39,$4E,$AB,$84,$E5
	.dc.b	$A2,$FA,$F9,$AC,$39,$CE,$70,$E7,$7C,$61,$F4,$0F,$8F,$FE,$87,$1B
	.dc.b	$3E,$E4,$67,$D3,$E1,$E7,$BC,$8F,$25,$72,$BE,$3A,$7B,$A3,$17,$DF
	.dc.b	$0E,$51,$D4,$FE,$63,$7C,$A3,$35,$C5,$B9,$7F,$8E,$5B,$5F,$8F,$C7
	.dc.b	$C5,$97,$5C,$BC,$37,$B6,$D7,$A9,$AF,$82,$78,$65,$23,$C9,$87,$67
	.dc.b	$46,$2B,$8C,$7A,$A3,$FC,$30,$D7,$C1,$3F,$D5,$1F,$82,$7F,$56,$F8
	.dc.b	$CF,$F5,$5F,$46,$5B,$19,$F8,$9F,$E3,$2D,$9E,$57,$FE,$4F,$E9,$BE
	.dc.b	$F1,$A4,$AF,$EC,$73,$B7,$D7,$3B,$6C,$FD,$3F,$53,$FD,$EB,$AF,$CD
	.dc.b	$AD,$6F,$F4,$5B,$19,$AE,$17,$F6,$E7,$2F,$E1,$9A,$FE,$31,$B8,$7F
	.dc.b	$80,$E3,$84,$C7,$C6,$7D,$0C,$3C,$6F,$38,$D6,$1E,$37,$C3,$1F,$95
	.dc.b	$E2,$7D,$27,$F9,$96,$3F,$49,$18,$F1,$B9,$B6,$33,$3F,$D6,$DB,$3A
	.dc.b	$8E,$51,$D5,$B7,$FA,$C7,$84,$67,$AF,$FD,$19,$D3,$FE,$E8,$C3,$F6
	.dc.b	$1F,$45,$8E,$4F,$DE,$36,$D3,$E8,$BC,$36,$3F,$D1,$38,$7F,$6B,$6C
	.dc.b	$EA,$3F,$D2,$F1,$FE,$A8,$ED,$FF,$77,$3F,$8F,$6B,$FE,$24,$6B,$FE
	.dc.b	$6D,$6B,$FC,$8B,$BF,$C8,$A3,$8E,$5F,$ED,$46,$E5,$FE,$4A,$BF,$F0
	.dc.b	$AD,$7F,$AE,$3E,$8D,$EA,$7B,$79,$1F,$28,$F4,$F0,$DB,$E1,$5E,$0B
	.dc.b	$E3,$1F,$A7,$6B,$FE,$3C,$6B,$FE,$85,$6B,$FE,$BB,$77,$9F,$FB,$A9
	.dc.b	$B9,$66,$BC,$9C,$0F,$8D,$F3,$7A,$CF,$9C,$A4,$CB,$D9,$C3,$E5,$3F
	.dc.b	$74,$B2,$FA,$8F,$96,$33,$18,$8F,$F6,$76,$CF,$A4,$F8,$F1,$FE,$D7
	.dc.b	$9E,$3A,$F9,$EF,$E4,$56,$A3,$F1,$5E,$6F,$F2,$96,$F5,$FE,$1A,$31
	.dc.b	$BF,$F7,$4D,$1F,$37,$F2,$FC,$7F,$19,$34,$F9,$8F,$F1,$3D,$7F,$CF
	.dc.b	$FD,$1F,$5F,$C0,$34,$FC,$77,$8B,$F9,$4F,$9E,$67,$49,$8F,$FB,$7A
	.dc.b	$38,$BF,$EF,$AD,$7F,$DE,$0F,$F2,$16,$BF,$E7,$66,$BF,$EB,$F6,$BF
	.dc.b	$E9,$68,$FF,$73,$BD,$7E,$27,$C4,$78,$FC,$4F,$D3,$F8,$3F,$E3,$69
	.dc.b	$FF,$77,$B5,$FF,$0B,$61,$FE,$A0,$8C,$70,$FE,$D5,$FC,$04,$F9,$FC
	.dc.b	$1F,$BB,$97,$F1,$58,$F3,$46,$0F,$E7,$5D,$EF,$A3,$FE,$0B,$87,$EF
	.dc.b	$DB,$8F,$FA,$B9,$B7,$FB,$E7,$EF,$C7,$B7,$C3,$FD,$FF,$91,$EE,$AF
	.dc.b	$39,$FC,$11,$C5,$FF,$0B,$D9,$5F,$B6,$FA,$07,$CB,$FE,$51,$B8,$7F
	.dc.b	$4F,$6B,$FE,$A4,$6B,$FE,$35,$3F,$88,$D3,$E8,$FE,$F1,$23,$EE,$7D
	.dc.b	$1F,$E0,$7E,$8F,$CF,$C5,$27,$F0,$11,$E3,$F5,$7C,$44,$8F,$13,$F2
	.dc.b	$30,$FA,$3B,$7D,$DF,$5F,$F1,$87,$E3,$7D,$5F,$E2,$3E,$FF,$F8,$2F
	.dc.b	$0F,$F2,$57,$07,$E4,$1F,$F2,$67,$77,$CA,$6F,$83,$C3,$D1,$91,$FD
	.dc.b	$BE,$F9,$3F,$EC,$46,$BF,$04,$38,$7E,$23,$C2,$FC,$11,$A9,$EA,$FD
	.dc.b	$61,$7F,$7B,$99,$A7,$E3,$27,$ED,$C6,$3C,$4F,$D5,$1F,$AB,$FA,$3B
	.dc.b	$3F,$A9,$8A,$F0,$A4,$FE,$C7,$E8,$E4,$FE,$BD,$FD,$A7,$87,$88,$E1
	.dc.b	$17,$3F,$C4,$4F,$F3,$CC,$9F,$03,$49,$7F,$3B,$87,$F0,$92,$F8,$1C
	.dc.b	$CF,$BE,$38,$BB,$3E,$9D,$75,$9D,$E1,$F2,$2B,$75,$C6,$B4,$FF,$07
	.dc.b	$FA,$3F,$D0,$F5,$C8,$C1,$0F,$85,$1F,$C4,$4F,$E2,$5F,$29,$FC,$4F
	.dc.b	$BE,$4F,$F1,$17,$F8,$B7,$D1,$E2,$3E,$54,$5C,$BF,$55,$3F,$8E,$E5
	.dc.b	$F8,$9F,$13,$2F,$D8,$7E,$12,$34,$FE,$2D,$FE,$83,$CD,$FC,$F6,$3F
	.dc.b	$0F,$18,$36,$FE,$16,$3F,$2C,$FE,$1B,$D1,$FE,$FB,$F4,$6F,$92,$A4
	.dc.b	$9F,$99,$D0,$97,$C9,$66,$7F,$17,$1C,$5D,$B5,$CC,$7F,$7A,$FA,$3F
	.dc.b	$F3,$D4,$7B,$3D,$57,$CD,$FF,$7E,$6C,$7C,$09,$DB,$F9,$D6,$EF,$FA
	.dc.b	$26,$F3,$FF,$09,$6C,$FF,$AF,$E9,$F3,$BE,$1F,$03,$B7,$95,$2B,$B4
	.dc.b	$65,$FF,$32,$5F,$F3,$77,$F6,$8D,$7B,$47,$F7,$08,$FE,$13,$1F,$EC
	.dc.b	$D6,$A6,$DB,$E7,$FF,$C3,$7E,$2F,$F6,$6A,$FF,$6A,$7D,$2F,$EF,$34
	.dc.b	$FD,$EB,$74,$FE,$49,$5B,$AF,$F5,$16,$7C,$88,$FE,$1B,$E9,$3F,$BB
	.dc.b	$57,$F2,$BF,$A4,$FE,$6E,$FF,$2D,$FA,$4F,$D5,$57,$E3,$1A,$FF,$09
	.dc.b	$9D,$BF,$EE,$76,$A6,$98,$87,$C7,$7F,$73,$F9,$8F,$E5,$84,$AF,$E3
	.dc.b	$87,$F3,$C7,$0F,$F0,$B7,$83,$F8,$EE,$0F,$F4,$13,$FC,$59,$F9,$0F
	.dc.b	$E4,$BD,$07,$D0,$FF,$56,$3F,$CA,$B3,$A9,$E1,$C1,$82,$3D,$7E,$87
	.dc.b	$7B,$7F,$68,$AF,$F6,$14,$FF,$51,$E8,$46,$20,$7F,$99,$1C,$36,$79
	.dc.b	$BE,$97,$C7,$E9,$79,$3C,$EE,$C4,$9F,$87,$7D,$A3,$11,$EF,$78,$3F
	.dc.b	$0F,$E9,$3D,$A6,$1D,$19,$21,$F1,$DD,$BF,$2F,$E3,$3F,$CD,$3F,$12
	.dc.b	$F8,$67,$47,$B3,$67,$87,$FD,$C4,$7C,$D7,$F7,$38,$CE,$3F,$0A,$72
	.dc.b	$F1,$7F,$BC,$61,$D6,$F1,$3F,$22,$30,$BE,$93,$9B,$E0,$C3,$FC,$01
	.dc.b	$18,$BE,$B3,$B7,$AE,$3E,$88,$C1,$D5,$C4,$F1,$AE,$75,$EB,$6D,$6F
	.dc.b	$0F,$B7,$5E,$2C,$3C,$B8,$F1,$AF,$49,$E5,$2C,$6B,$0F,$1D,$F2,$7B
	.dc.b	$4D,$3E,$36,$1E,$05,$5F,$0B,$0F,$06,$3E,$8B,$E3,$5E,$C3,$C0,$5A
	.dc.b	$E1,$1E,$08,$FE,$14,$F0,$C7,$D1,$18,$3D,$C5,$3C,$35,$DD,$F2,$9E
	.dc.b	$73,$BD,$F9,$31,$C6,$1E,$09,$D9,$E5,$C3,$EA,$D6,$E3,$CD,$7D,$5E
	.dc.b	$73,$EC,$9D,$1B,$C3,$CE,$48,$F6,$3A,$78,$12,$7D,$73,$C6,$3E,$01
	.dc.b	$23,$DF,$39,$7C,$EF,$57,$31,$F2,$7E,$2B,$F8,$36,$1F,$94,$6E,$1F
	.dc.b	$8D,$6E,$3F,$C0,$1F,$E1,$97,$FC,$0B,$E1,$B1,$8F,$F1,$C6,$C7,$F8
	.dc.b	$E5,$AF,$F9,$6D,$AF,$13,$EA,$53,$5F,$F2,$55,$7E,$92,$74,$FE,$9A
	.dc.b	$75,$D5,$F9,$49,$CE,$7F,$EF,$CB,$E2,$F8,$E3,$AC,$61,$F2,$1F,$8D
	.dc.b	$AE,$6F,$E3,$CF,$26,$9D,$E7,$FC,$B8,$F8,$E1,$8F,$E9,$2B,$8E,$39
	.dc.b	$8C,$5F,$7A,$F8,$07,$77,$9E,$1E,$5C,$3B,$63,$F9,$1A,$F0,$CF,$AE
	.dc.b	$31,$1E,$03,$93,$8B,$FC,$C6,$3E,$AB,$F1,$56,$8F,$61,$C7,$1E,$EF
	.dc.b	$E8,$CF,$92,$7F,$9E,$B1,$F4,$3E,$23,$18,$77,$7A,$57,$63,$99,$71
	.dc.b	$F4,$47,$83,$1F,$87,$5F,$5E,$3C,$53,$AC,$38,$47,$63,$81,$DA,$3E
	.dc.b	$31,$D2,$F7,$1F,$CE,$30,$F9,$6F,$DC,$AF,$69,$F0,$8B,$5F,$D6,$38
	.dc.b	$3F,$AD,$E9,$FE,$97,$CF,$FD,$31,$18,$CB,$FB,$87,$C5,$E5,$E1,$C9
	.dc.b	$FD,$1D,$FE,$E9,$79,$CB,$FB,$17,$C8,$DC,$39,$47,$F7,$78,$C5,$9A
	.dc.b	$62,$AD,$FF,$7F,$8F,$F4,$DD,$78,$1A,$F6,$8F,$EC,$F5,$CD,$BF,$11
	.dc.b	$D8,$EF,$F4,$76,$3F,$C1,$51,$CB,$1F,$EF,$6D,$A3,$4C,$78,$9E,$83
	.dc.b	$D3,$F4,$96,$FF,$C2,$91,$EF,$EC,$FF,$6A,$46,$0F,$F0,$A7,$BD,$F0
	.dc.b	$3C,$79,$1F,$E1,$BF,$94,$BA,$FF,$86,$E7,$C4,$49,$F6,$BF,$02,$1D
	.dc.b	$DF,$E1,$E9,$FF,$40,$B6,$3F,$DE,$7C,$4F,$9E,$C3,$FE,$C4,$9B,$5F
	.dc.b	$EE,$B6,$D7,$F8,$02,$BD,$2F,$FB,$1E,$33,$5C,$F1,$FB,$D2,$FA,$47
	.dc.b	$CD,$E1,$B8,$FF,$AB,$ED,$F4,$4E,$3D,$5C,$AD,$7D,$3B,$E0,$71,$97
	.dc.b	$AA,$31,$3F,$E9,$EC,$9F,$1B,$DC,$FC,$86,$7E,$FF,$63,$15,$B9,$F9
	.dc.b	$15,$C2,$3F,$C9,$F8,$7A,$67,$24,$8F,$47,$94,$E5,$E7,$5F,$32,$31
	.dc.b	$3F,$EB,$0F,$A1,$7B,$7F,$11,$B1,$FE,$A4,$FA,$57,$C5,$E6,$A4,$E7
	.dc.b	$D1,$B8,$FF,$BC,$2D,$1F,$A9,$FA,$57,$CB,$D3,$1F,$3B,$70,$FF,$58
	.dc.b	$CE,$9B,$3A,$69,$E7,$F8,$9F,$31,$F2,$39,$FF,$D7,$91,$92,$65,$FE
	.dc.b	$74,$D5,$5F,$91,$6C,$FC,$A6,$CE,$E7,$E9,$F2,$7F,$A0,$2B,$9F,$27
	.dc.b	$F0,$D8,$C5,$7F,$9E,$7D,$BF,$E0,$67,$F0,$4F,$6A,$DF,$F9,$FE,$2C
	.dc.b	$FF,$9F,$E6,$DF,$F1,$6F,$09,$F4,$BF,$C3,$0F,$E1,$BF,$4B,$71,$FF
	.dc.b	$44,$47,$ED,$7E,$37,$18,$FF,$97,$A7,$F9,$17,$86,$9F,$C5,$F8,$2F
	.dc.b	$5E,$9F,$8E,$E2,$7A,$18,$C3,$F3,$D5,$B6,$D6,$E7,$58,$FE,$D2,$B4
	.dc.b	$D9,$D5,$70,$3F,$DD,$7F,$39,$FE,$8C,$6C,$FF,$2B,$87,$CA,$58,$93
	.dc.b	$A3,$8C,$62,$3D,$C3,$CA,$57,$F9,$EA,$BB,$FA,$9D,$7E,$73,$B3,$23
	.dc.b	$27,$92,$31,$7D,$07,$3B,$E1,$BF,$44,$79,$FE,$0E,$DF,$E9,$E3,$F1
	.dc.b	$44,$7F,$14,$DA,$F3,$E4,$E1,$7C,$87,$29,$CC,$62,$FE,$3F,$91,$F1
	.dc.b	$4F,$8D,$E9,$D8,$FF,$5A,$F4,$1F,$F5,$BD,$BE,$89,$B7,$F2,$D6,$CF
	.dc.b	$F0,$66,$CF,$FB,$59,$AF,$FB,$0F,$81,$F4,$9F,$D0,$7E,$67,$EB,$36
	.dc.b	$FE,$DF,$C6,$FD,$93,$ED,$7D,$25,$8F,$29,$E4,$C5,$C5,$E3,$24,$78
	.dc.b	$DF,$EB,$18,$FF,$24,$CB,$E1,$F6,$FE,$17,$CE,$FE,$A4,$FF,$71,$C3
	.dc.b	$F9,$9C,$FC,$C6,$ED,$FD,$F9,$FF,$5E,$E3,$E2,$9F,$C3,$46,$A3,$C9
	.dc.b	$7F,$14,$8F,$C5,$7F,$B5,$7C,$53,$C6,$FF,$00,$36,$7A,$72,$A1,$F1
	.dc.b	$7A,$AF,$CE,$7C,$74,$BF,$D0,$FC,$7D,$CF,$F1,$D9,$FF,$08,$7E,$C9
	.dc.b	$F6,$7E,$1F,$EC,$12,$BF,$DD,$3B,$BE,$8F,$F5,$07,$EA,$7C,$2B,$3E
	.dc.b	$FF,$9B,$0F,$85,$FF,$0D,$38,$7F,$61,$AE,$23,$53,$F1,$F2,$7D,$1B
	.dc.b	$93,$FB,$B3,$67,$FA,$7B,$67,$D2,$70,$9F,$F0,$97,$4B,$DC,$FF,$14
	.dc.b	$36,$FC,$DF,$00,$F8,$7C,$D7,$8F,$E0,$7F,$B9,$E5,$FC,$CB,$A7,$F3
	.dc.b	$D9,$FB,$34,$FF,$12,$3F,$B7,$9D,$65,$FA,$27,$F9,$CC,$7C,$C6,$B9
	.dc.b	$F4,$3D,$67,$F8,$C2,$99,$FF,$90,$67,$F8,$6C,$62,$3C,$73,$F2,$9D
	.dc.b	$61,$E4,$A7,$D1,$FF,$2A,$C7,$E5,$36,$39,$CF,$F7,$77,$F8,$74,$66
	.dc.b	$39,$3B,$8F,$C1,$B5,$FD,$EB,$51,$E0,$F2,$FF,$C5,$51,$CE,$B3,$8B
	.dc.b	$EB,$7F,$31,$7E,$0B,$FC,$1C,$63,$0F,$57,$D1,$FE,$0A,$BF,$91,$36
	.dc.b	$3F,$CD,$3C,$DF,$9E,$C7,$F0,$D3,$97,$F8,$8B,$C6,$BF,$36,$D7,$3C
	.dc.b	$9F,$E3,$F8,$C4,$7F,$39,$AF,$89,$7D,$BD,$CF,$E8,$9F,$49,$FD,$15
	.dc.b	$98,$FE,$C3,$C8,$FB,$F8,$71,$9F,$F2,$B7,$07,$F4,$BF,$9F,$FC,$3F
	.dc.b	$C5,$BE,$3B,$FD,$09,$A2,$73,$1F,$D9,$B2,$B8,$F8,$E6,$F8,$7F,$D3
	.dc.b	$1B,$7C,$6B,$FD,$C9,$7C,$5B,$7C,$5A,$BE,$7F,$FC,$43,$E8,$E1,$B1
	.dc.b	$FE,$CE,$AF,$E7,$3E,$5F,$F2,$1C,$FD,$13,$F7,$AB,$71,$88,$E9,$EE
	.dc.b	$7F,$68,$79,$1B,$71,$1B,$AE,$97,$FD,$1A,$F9,$B6,$B6,$DC,$3D,$35
	.dc.b	$C3,$0E,$91,$E9,$AD,$C6,$0F,$09,$FD,$43,$C8,$F9,$6F,$9A,$78,$11
	.dc.b	$DD,$75,$74,$72,$C3,$E2,$BF,$C9,$1A,$F9,$4F,$E1,$EE,$1F,$D1,$FC
	.dc.b	$73,$E2,$FF,$90,$C7,$F1,$48,$C7,$4F,$F9,$99,$B1,$99,$DB,$DF,$E0
	.dc.b	$3F,$C4,$6F,$66,$B9,$EA,$FF,$37,$E1,$FE,$66,$6C,$F0,$DB,$EB,$4E
	.dc.b	$6B,$D9,$39,$DB,$E9,$7A,$7F,$DE,$2F,$73,$9A,$FC,$03,$5F,$C5,$36
	.dc.b	$FF,$C2,$91,$F8,$B8,$C1,$30,$FC,$75,$F5,$C3,$F0,$13,$9A,$F1,$D6
	.dc.b	$9E,$4C,$72,$7D,$3A,$E4,$FF,$8B,$FE,$93,$FC,$F2,$FE,$83,$91,$18
	.dc.b	$8F,$C9,$C6,$3C,$6E,$2E,$BC,$6B,$5B,$D3,$F3,$F3,$BB,$E9,$39,$CF
	.dc.b	$FD,$F8,$FF,$73,$8C,$F9,$9F,$E2,$0E,$CF,$F7,$EF,$5B,$E8,$71,$D1
	.dc.b	$FE,$62,$6D,$FA,$6E,$E5,$E8,$6D,$F1,$9F,$E1,$D5,$AC,$7F,$BE,$C7
	.dc.b	$F2,$0F,$0B,$7F,$3F,$23,$04,$3F,$B1,$F8,$1F,$D3,$1F,$F5,$AB,$5C
	.dc.b	$CF,$FA,$E1,$7D,$1B,$E9,$3C,$FD,$2F,$99,$FF,$14,$D7,$F7,$D6,$BF
	.dc.b	$E5,$7A,$E0,$FF,$79,$6B,$FD,$6A,$37,$C1,$FD,$AA,$BA,$7A,$7F,$12
	.dc.b	$BC,$38,$5C,$38,$DF,$F5,$D6,$F0,$FF,$93,$9B,$5F,$E0,$EE,$37,$C8
	.dc.b	$FA,$3D,$66,$FF,$33,$E8,$25,$FF,$85,$67,$2F,$9F,$93,$85,$48,$C1
	.dc.b	$0F,$74,$79,$4F,$8B,$F4,$8F,$23,$ED,$FA,$1F,$C9,$76,$FC,$1F,$D0
	.dc.b	$FD,$C3,$3D,$5E,$DF,$2B,$CD,$F4,$33,$3F,$91,$1F,$43,$C6,$FC,$14
	.dc.b	$7C,$D8,$E0,$FF,$8B,$63,$38,$7F,$85,$3B,$5F,$45,$FE,$03,$AF,$F3
	.dc.b	$9C,$63,$97,$FD,$EA,$D8,$FF,$13,$B5,$FF,$14,$E4,$F9,$1F,$C0,$FE
	.dc.b	$03,$F1,$A7,$FC,$26,$DC,$7F,$C2,$15,$B9,$FF,$29,$E3,$FE,$F4,$7A
	.dc.b	$B5,$FF,$26,$E1,$E6,$F8,$FF,$1C,$E6,$7C,$FE,$3F,$F8,$FD,$FF,$2E
	.dc.b	$B5,$FF,$4C,$B7,$0F,$80,$F8,$F3,$FF,$48,$E3,$FC,$52,$BC,$FB,$7A
	.dc.b	$8F,$C7,$46,$6F,$E4,$7D,$2F,$F0,$76,$F8,$BF,$B4,$8C,$61,$FE,$1E
	.dc.b	$D1,$1F,$B2,$C7,$FD,$57,$9B,$C8,$FD,$CE,$B2,$7E,$1C,$7E,$77,$E9
	.dc.b	$5F,$3B,$FB,$ED,$A5,$7F,$97,$E7,$2F,$CA,$F9,$7F,$E0,$7F,$4B,$23
	.dc.b	$07,$CE,$F8,$7F,$E0,$FE,$94,$8C,$1C,$FD,$CF,$C9,$91,$C1,$F5,$BD
	.dc.b	$DF,$E1,$FC,$F2,$30,$7F,$9C,$7E,$0D,$EA,$FE,$D7,$C2,$F8,$1C,$FF
	.dc.b	$E4,$F6,$F3,$7F,$B0,$2B,$77,$C6,$F8,$78,$32,$30,$43,$FC,$EF,$3C
	.dc.b	$1B,$3F,$C4,$E3,$11,$F9,$AE,$5F,$BC,$FF,$BB,$DB,$CF,$FA,$0F,$0B
	.dc.b	$F9,$5C,$3C,$3B,$12,$BF,$2A,$EA,$37,$E0,$49,$FC,$CB,$CB,$9B,$11
	.dc.b	$F7,$A2,$ED,$FA,$FC,$15,$F1,$7E,$0E,$5F,$FA,$67,$74,$7C,$9F,$83
	.dc.b	$CC,$FF,$BF,$67,$2E,$E7,$4D,$21,$B7,$2F,$E4,$39,$BF,$A3,$F5,$48
	.dc.b	$FD,$8C,$7C,$23,$F5,$A4,$7F,$66,$FE,$C3,$D3,$F8,$E4,$8F,$DC,$FD
	.dc.b	$2D,$7F,$3B,$17,$4F,$81,$3F,$CD,$F8,$7F,$03,$5F,$E8,$2C,$9F,$0F
	.dc.b	$F8,$73,$5C,$92,$7E,$24,$65,$FC,$74,$65,$FC,$1B,$F8,$12,$3F,$C5
	.dc.b	$BE,$95,$F2,$5F,$DA,$1B,$1F,$EE,$26,$AC,$BF,$92,$43,$E4,$B8,$54
	.dc.b	$AE,$0F,$38,$FE,$A5,$F4,$AE,$DF,$EA,$9E,$9B,$E7,$69,$A8,$98,$7F
	.dc.b	$B4,$05,$71,$70,$D7,$FD,$AA,$D7,$FD,$DE,$FF,$1B,$EC,$FF,$55,$C7
	.dc.b	$F6,$FF,$1F,$C2,$79,$3B,$E4,$7F,$B9,$71,$FE,$9D,$5C,$1B,$1F,$EB
	.dc.b	$4B,$E8,$DA,$FE,$FD,$1F,$CA,$E3,$C3,$18,$3F,$C0,$58,$72,$87,$D2
	.dc.b	$FF,$B3,$76,$FF,$67,$37,$1F,$3D,$6C,$FF,$42,$57,$A8,$F9,$59,$74
	.dc.b	$23,$F1,$FE,$93,$FA,$26,$1F,$ED,$9F,$9B,$75,$7F,$80,$DB,$5F,$EC
	.dc.b	$E9,$FF,$6F,$57,$0C,$7E,$3B,$FD,$26,$6F,$91,$FE,$1B,$8C,$47,$FB
	.dc.b	$59,$BE,$0F,$F6,$6F,$93,$FF,$17,$F5,$4A,$FE,$B3,$1F,$C2,$8F,$EA
	.dc.b	$A4,$7F,$AD,$BF,$DA,$B7,$7D,$0F,$F8,$E3,$E9,$25,$79,$F3,$FE,$91
	.dc.b	$5A,$8C,$0C,$51,$98,$FC,$C4,$FF,$4F,$F7,$D7,$D2,$72,$71,$AE,$1A
	.dc.b	$FF,$80,$5F,$C0,$E0,$B7,$6F,$F9,$30,$FF,$27,$CE,$8F,$AF,$7E,$6E
	.dc.b	$1F,$61,$1F,$F0,$2C,$3C,$8F,$F2,$AE,$B2,$7D,$78,$FA,$A7,$A9,$F1
	.dc.b	$49,$1C,$9F,$91,$F4,$7A,$8E,$37,$D5,$B1,$E6,$34,$FC,$0A,$EC,$E8
	.dc.b	$F9,$63,$E1,$1E,$F7,$F0,$18,$7D,$27,$F1,$44,$7F,$16,$F9,$7F,$D7
	.dc.b	$C6,$63,$06,$9C,$57,$4D,$F1,$E0,$FE,$DA,$31,$5F,$82,$8F,$DC,$8B
	.dc.b	$FD,$DF,$1F,$58,$C4,$F8,$A3,$AE,$3E,$76,$E1,$DA,$7C,$F5,$C1,$F5
	.dc.b	$BD,$AF,$D2,$F8,$EF,$E0,$63,$CA,$B6,$72,$AD,$B4,$EC,$7C,$99,$E3
	.dc.b	$5C,$EF,$CC,$F7,$9A,$71,$3E,$01,$FE,$E0,$9F,$54,$71,$8F,$09,$F3
	.dc.b	$0A,$FA,$EF,$CC,$47,$B4,$6A,$FC,$EF,$96,$FC,$F8,$FD,$0B,$F9,$95
	.dc.b	$AB,$E8,$78,$63,$A5,$F9,$DD,$BC,$B0,$F7,$BE,$9C,$3E,$1D,$72,$8E
	.dc.b	$0F,$32,$38,$21,$E5,$3D,$F8,$79,$63,$D7,$7E,$09,$CD,$F8,$1E,$11
	.dc.b	$92,$4F,$68,$CB,$C8,$8F,$9C,$E7,$7F,$9C,$F8,$5F,$E1,$58,$7E,$51
	.dc.b	$B8,$7E,$8D,$B3,$BC,$BF,$9B,$5B,$DD,$C3,$FC,$E5,$AF,$F5,$46,$BF
	.dc.b	$AA,$6B,$C4,$FB,$F2,$D5,$CB,$F8,$E9,$E0,$6A,$32,$FE,$18,$CE,$9F
	.dc.b	$B3,$AE,$55,$C6,$7B,$4E,$6B,$8B,$FB,$98,$E3,$A7,$FB,$CE,$72,$FB
	.dc.b	$B0,$FC,$FE,$1D,$2B,$C0,$74,$9E,$18,$77,$AD,$DF,$2A,$F8,$87,$A6
	.dc.b	$74,$D3,$9C,$F0,$8C,$E1,$E7,$6D,$F6,$AD,$9C,$67,$D4,$F8,$CF,$EC
	.dc.b	$17,$E1,$7E,$91,$B9,$F6,$34,$F2,$CF,$77,$AE,$34,$EB,$87,$EE,$A3
	.dc.b	$FC,$4D,$7F,$4A,$73,$1C,$CE,$F5,$CC,$C9,$6F,$C2,$D8,$F8,$B3,$DE
	.dc.b	$7C,$AD,$8F,$2C,$7C,$19,$E6,$D3,$FB,$3E,$7F,$84,$CB,$FA,$E7,$C7
	.dc.b	$DC,$F2,$7F,$48,$71,$87,$F6,$3A,$BB,$FF,$6C,$9B,$C1,$FD,$12,$F8
	.dc.b	$12,$FF,$DA,$70,$9F,$F7,$44,$7F,$6D,$6D,$73,$AF,$EC,$71,$DB,$F5
	.dc.b	$26,$FD,$A3,$0F,$83,$DF,$E3,$89,$A4,$BF,$ED,$AF,$3D,$B8,$47,$F6
	.dc.b	$CC,$3F,$C1,$AF,$F6,$6F,$A5,$F6,$F4,$3D,$5F,$6C,$61,$B3,$FD,$D2
	.dc.b	$74,$F1,$31,$1F,$13,$C4,$F2,$92,$3F,$BF,$B7,$6F,$EE,$18,$3E,$97
	.dc.b	$85,$79,$5F,$A3,$39,$EC,$F9,$2D,$9F,$F3,$24,$FF,$86,$9B,$3A,$8C
	.dc.b	$4F,$F8,$73,$89,$F3,$7E,$88,$FF,$0F,$9C,$CD,$39,$8F,$E1,$30,$FA
	.dc.b	$47,$CD,$F8,$9B,$8F,$F9,$D5,$B3,$FC,$81,$AF,$F9,$5D,$CD,$7E,$47
	.dc.b	$E8,$3F,$DC,$35,$FE,$0E,$9C,$BF,$E5,$38,$CD,$79,$A5,$F0,$5E,$12
	.dc.b	$57,$85,$B3,$EF,$F1,$5F,$49,$F2,$AF,$8D,$FF,$91,$B1,$56,$A3,$C5
	.dc.b	$87,$E8,$27,$37,$C6,$31,$5B,$C3,$E4,$FC,$F3,$D4,$FA,$AD,$C7,$FD
	.dc.b	$30,$D9,$FE,$86,$D7,$FC,$F4,$E7,$1F,$E6,$FD,$4F,$A4,$FF,$53,$5F
	.dc.b	$F8,$CA,$B4,$FF,$8B,$A7,$5E,$5F,$ED,$3E,$91,$F0,$74,$C7,$C8,$DB
	.dc.b	$FF,$25,$CE,$9B,$3A,$69,$DF,$E5,$BD,$64,$7D,$6D,$D3,$F1,$9C,$4F
	.dc.b	$9D,$FC,$BE,$8A,$F9,$53,$9B,$E8,$7D,$17,$17,$F2,$A7,$77,$FC,$33
	.dc.b	$D5,$97,$FE,$5A,$F5,$6D,$FF,$95,$FC,$1F,$F6,$BC,$FF,$97,$AB,$F8
	.dc.b	$9D,$6B,$87,$FC,$F1,$F4,$5F,$C1,$89,$5F,$E7,$06,$E1,$FE,$61,$EF
	.dc.b	$7D,$17,$F1,$C9,$FE,$06,$FE,$0F,$AD,$E7,$53,$F0,$DC,$3F,$C5,$B1
	.dc.b	$FC,$64,$E9,$B3,$B7,$F2,$13,$BC,$7F,$5F,$39,$6C,$EB,$C5,$C6,$7F
	.dc.b	$84,$F9,$8F,$F3,$6C,$FF,$A1,$1B,$5F,$E7,$95,$F1,$9F,$CD,$9C,$D7
	.dc.b	$19,$C1,$99,$B5,$CF,$CF,$E0,$71,$F9,$8B,$19,$25,$77,$8C,$BF,$A2
	.dc.b	$F5,$3F,$42,$49,$FD,$0B,$63,$E7,$7C,$47,$E1,$A3,$F9,$7B,$67,$F9
	.dc.b	$63,$67,$71,$EF,$F0,$FF,$A2,$7A,$9E,$D6,$C7,$FA,$5B,$E8,$18,$FF
	.dc.b	$56,$F8,$FF,$DB,$6B,$FC,$6A,$73,$5F,$EA,$C9,$CB,$FA,$BD,$9F,$45
	.dc.b	$FD,$0A,$7E,$53,$67,$E8,$7C,$0F,$F5,$5F,$8D,$F9,$CD,$8F,$17,$CA
	.dc.b	$FF,$59,$F6,$BF,$9C,$E4,$8C,$4F,$D9,$F8,$CF,$E9,$FD,$DF,$5D,$B1
	.dc.b	$FE,$C0,$AF,$E4,$7C,$3F,$ED,$B6,$F8,$C5,$EB,$FF,$74,$77,$39,$BF
	.dc.b	$EF,$3F,$2D,$FC,$2F,$3F,$1C,$63,$C9,$FF,$68,$65,$FE,$CE,$F8,$67
	.dc.b	$93,$FD,$F9,$B1,$9F,$86,$C4,$FE,$37,$E2,$3F,$BC,$F2,$9A,$4A,$FF
	.dc.b	$30,$E9,$E7,$F1,$FC,$1E,$DF,$C4,$6C,$FE,$6F,$E8,$79,$91,$FF,$34
	.dc.b	$F1,$31,$4F,$F8,$2B,$80,$3E,$77,$FB,$3F,$D0,$7F,$C0,$0C,$3F,$DA
	.dc.b	$7D,$7F,$E6,$2A,$DD,$6B,$9A,$56,$8F,$F6,$CC,$FF,$B7,$B9,$33,$B3
	.dc.b	$0E,$5B,$7C,$88,$FF,$74,$75,$BE,$87,$FC,$31,$F0,$3F,$91,$6E,$5F
	.dc.b	$EE,$AD,$7F,$92,$37,$CE,$FA,$BA,$7F,$89,$63,$04,$C7,$C0,$F9,$32
	.dc.b	$FF,$17,$47,$F7,$8B,$E6,$E8,$F8,$71,$AC,$FC,$EF,$F8,$AF,$6F,$C1
	.dc.b	$B9,$BF,$E9,$5E,$D7,$C9,$7F,$DF,$FF,$43,$FC,$56,$FF,$C0,$2D,$8F
	.dc.b	$8B,$1F,$DA,$B0,$FF,$04,$CE,$5F,$F8,$02,$DC,$9F,$43,$F8,$A9,$DF
	.dc.b	$93,$F9,$48,$C5,$F6,$8C,$74,$7F,$88,$30,$E5,$5A,$C5,$84,$F9,$DF
	.dc.b	$43,$FE,$AE,$6B,$A3,$5A,$FF,$8A,$23,$31,$97,$8E,$5F,$DC,$34,$F8
	.dc.b	$93,$C2,$FF,$C8,$DA,$FF,$2B,$F6,$7F,$C8,$BC,$16,$7F,$C8,$4D,$AF
	.dc.b	$E9,$4D,$C3,$FB,$DF,$A3,$F5,$9F,$E7,$5F,$43,$FE,$B4,$F8,$67,$95
	.dc.b	$FC,$46,$30,$FF,$7D,$AF,$F5,$EE,$3F,$EC,$38,$D3,$71,$76,$7E,$96
	.dc.b	$F8,$C6,$3B,$11,$89,$61,$D2,$BF,$DA,$98,$74,$69,$F3,$BE,$65,$ED
	.dc.b	$7F,$4C,$8C,$88,$CF,$A1,$E3,$8F,$E6,$3E,$2F,$89,$D5,$71,$D7,$DF
	.dc.b	$1D,$63,$18,$7D,$7C,$39,$CF,$6F,$33,$FC,$83,$7C,$23,$9E,$3D,$EB
	.dc.b	$73,$96,$DF,$E0,$6F,$8B,$D7,$0E,$F8,$78,$70,$E5,$18,$8F,$4E,$D8
	.dc.b	$99,$1F,$03,$0E,$AF,$E4,$B0,$EA,$F3,$35,$7F,$1D,$FC,$91,$E8,$29
	.dc.b	$4C,$CF,$C0,$77,$F1,$0F,$A1,$FE,$3B,$1C,$23,$1D,$9F,$CB,$1B,$19
	.dc.b	$9D,$BD,$3D,$DF,$F5,$4B,$E0,$BF,$F2,$B7,$6F,$F3,$46,$C6,$67,$6F
	.dc.b	$19,$F9,$D1,$CB,$6F,$9C,$DD,$BE,$C7,$A5,$FA,$F9,$CD,$70,$25,$FC
	.dc.b	$99,$D5,$7D,$FB,$E5,$3F,$8F,$6B,$F8,$D6,$E0,$AE,$33,$BB,$EB,$87
	.dc.b	$82,$3B,$18,$C3,$B4,$D8,$FC,$E5,$72,$E2,$74,$FC,$F9,$FD,$64,$63
	.dc.b	$3D,$4E,$67,$F4,$B3,$C3,$E2,$33,$52,$78,$D6,$F2,$FC,$FD,$EE,$73
	.dc.b	$F1,$1F,$C5,$7A,$3F,$88,$7D,$2D,$E5,$FE,$66,$DA,$FD,$EE,$3F,$ED
	.dc.b	$D7,$FD,$B1,$79,$C3,$D2,$D9,$D4,$FF,$B3,$3E,$23,$1F,$07,$23,$FD
	.dc.b	$C0,$EF,$E1,$F1,$1F,$ED,$28,$B9,$FF,$23,$9E,$DF,$5E,$F5,$A9,$2F
	.dc.b	$FB,$41,$FC,$4A,$1E,$5E,$63,$FA,$A6,$89,$E5,$1C,$23,$15,$DA,$B8
	.dc.b	$61,$FE,$A2,$BD,$DF,$26,$C6,$67,$E0,$79,$3F,$07,$E0,$65,$FF,$BF
	.dc.b	$63,$2D,$21,$FE,$F0,$F8,$1F,$90,$47,$C9,$B2,$3E,$67,$C0,$FC,$2F
	.dc.b	$22,$57,$A7,$3F,$F0,$83,$7C,$DF,$8E,$47,$06,$FC,$8F,$97,$CF,$F3
	.dc.b	$19,$1C,$E3,$ED,$7D,$2C,$8F,$1B,$F7,$FC,$4F,$E1,$7C,$1F,$EC,$67
	.dc.b	$F8,$73,$7C,$4D,$1B,$F7,$B8,$78,$BC,$7D,$A9,$8F,$D1,$C3,$FC,$CE
	.dc.b	$DF,$77,$B7,$B7,$23,$07,$BF,$B7,$F0,$64,$7E,$17,$67,$97,$DD,$91
	.dc.b	$D5,$F4,$7D,$17,$A7,$DE,$93,$E1,$7D,$9F,$4B,$ED,$CE,$4F,$95,$F8
	.dc.b	$1E,$D7,$C1,$F6,$E4,$FA,$5F,$89,$F2,$5F,$17,$DC,$93,$ED,$7E,$47
	.dc.b	$C1,$FE,$9F,$EB,$BE,$3F,$91,$FE,$E9,$6D,$7F,$6A,$6F,$8A,$5C,$DF
	.dc.b	$47,$DB,$95,$F2,$5F,$A9,$EF,$FE,$DE,$1C,$1F,$DC,$1F,$D2,$98,$F5
	.dc.b	$B5,$19,$8F,$F1,$C0,$FC,$19,$9E,$AA,$4F,$2A,$E1,$D9,$A8,$FF,$11
	.dc.b	$DD,$79,$CE,$AF,$FC,$44,$D8,$FF,$14,$B5,$F0,$47,$E3,$7D,$AF,$C7
	.dc.b	$7D,$1C,$9F,$C3,$46,$5F,$C4,$46,$5F,$DE,$F2,$27,$F2,$D9,$7F,$15
	.dc.b	$FD,$4F,$C9,$9F,$9F,$81,$18,$CB,$FC,$77,$0C,$3C,$CD,$BE,$53,$31
	.dc.b	$FC,$C4,$75,$FA,$89,$1F,$C2,$F0,$47,$87,$E5,$F3,$3F,$C4,$A7,$2E
	.dc.b	$E7,$4D,$21,$B7,$2F,$D6,$F8,$5F,$33,$FE,$9C,$6B,$FD,$F1,$AF,$F8
	.dc.b	$01,$AF,$F9,$4D,$DE,$7F,$5D,$B8,$E7,$3F,$C2,$F2,$7C,$9F,$72,$61
	.dc.b	$ED,$7E,$67,$B5,$88,$F6,$45,$C7,$E7,$66,$AF,$C7,$F4,$CB,$FC,$53
	.dc.b	$F9,$2E,$9C,$47,$E2,$22,$DF,$E2,$F8,$D5,$8D,$3F,$17,$56,$FF,$17
	.dc.b	$3F,$AD,$D3,$F5,$DD,$52,$7F,$9F,$74,$FB,$9F,$D7,$4E,$5F,$E8,$B0
	.dc.b	$9E,$51,$FA,$8D,$DC,$FF,$EC,$A8,$46,$1F,$A5,$19,$8C,$47,$D3,$8F
	.dc.b	$DA,$46,$34,$FA,$11,$FB,$2F,$0B,$F7,$1F,$35,$FC,$FE,$FF,$6E,$D7
	.dc.b	$F6,$AD,$AF,$D7,$2F,$33,$8F,$9B,$2B,$8B,$D2,$31,$1C,$A2,$D7,$29
	.dc.b	$FE,$11,$F3,$4F,$87,$F7,$36,$FF,$DD,$2D,$7F,$C9,$2D,$7F,$C9,$AE
	.dc.b	$6B,$F8,$13,$78,$7F,$D4,$4D,$9F,$F5,$4E,$2F,$96,$FE,$31,$CC,$F4
	.dc.b	$FE,$73,$6B,$FD,$F2,$D7,$FC,$B2,$D7,$FC,$BA,$E7,$1F,$E2,$4D,$C7
	.dc.b	$FD,$64,$D9,$FF,$5C,$FC,$5D,$ED,$FE,$07,$F2,$BF,$C1,$C9,$3F,$BB
	.dc.b	$F9,$6A,$FE,$E6,$2E,$3F,$9E,$9F,$E1,$DF,$21,$2B,$F8,$13,$FC,$47
	.dc.b	$0C,$47,$F0,$18,$B5,$FA,$69,$FE,$2D,$F1,$92,$BF,$85,$3F,$C6,$78
	.dc.b	$31,$1F,$C2,$62,$D7,$EB,$A7,$F8,$F7,$C4,$4A,$FE,$24,$FF,$21,$FA
	.dc.b	$1C,$47,$F1,$18,$B5,$FB,$69,$FE,$4D,$F4,$3B,$AF,$E3,$51,$87,$F8
	.dc.b	$E4,$65,$FE,$35,$19,$7F,$A4,$FC,$3B,$D7,$5F,$FB,$9B,$E6,$5E,$9F
	.dc.b	$FB,$8C,$FE,$B7,$79,$71,$E6,$FF,$B9,$E7,$73,$FE,$EE,$C3,$A5,$EE
	.dc.b	$31,$87,$65,$5C,$1E,$F7,$CF,$0E,$B1,$C4,$61,$D1,$E3,$1F,$EF,$78
	.dc.b	$C5,$73,$8F,$3C,$E8,$FF,$80,$2E,$23,$C1,$3C,$23,$95,$F9,$67,$95
	.dc.b	$F8,$EF,$8B,$5F,$1C,$EE,$B8,$E1,$E7,$7B,$1E,$A9,$F1,$1C,$23,$D9
	.dc.b	$87,$86,$3A,$9C,$6F,$B4,$66,$38,$1C,$AB,$AB,$C0,$EB,$1B,$38,$46
	.dc.b	$A3,$E6,$9E,$0B,$F4,$3F,$52,$FC,$0C,$39,$C6,$9F,$0E,$1F,$11,$D3
	.dc.b	$F2,$A3,$57,$E0,$C3,$E3,$61,$E1,$8F,$24,$F3,$8F,$21,$97,$89,$99
	.dc.b	$E2,$7C,$27,$38,$78,$27,$E6,$5F,$86,$3D,$83,$07,$C3,$7F,$17,$7D
	.dc.b	$C6,$39,$FE,$BB,$4C,$C6,$DF,$F6,$A1,$92,$B9,$7B,$7C,$27,$78,$C4
	.dc.b	$23,$13,$FE,$F8,$7F,$35,$4F,$6F,$FC,$65,$87,$E8,$6F,$99,$F8,$69
	.dc.b	$D7,$47,$E8,$63,$33,$FA,$79,$CC,$7E,$30,$CE,$5F,$EC,$5B,$EA,$F8
	.dc.b	$63,$A4,$61,$F8,$C7,$03,$E4,$4F,$4A,$F2,$47,$AC,$E3,$7E,$37,$11
	.dc.b	$E3,$AE,$91,$F2,$23,$13,$BC,$3E,$0D,$6A,$BC,$07,$F1,$07,$0F,$4C
	.dc.b	$3D,$55,$C6,$3C,$27,$BC,$D1,$E8,$9E,$35,$DA,$78,$5F,$8A,$B8,$4E
	.dc.b	$5A,$74,$8F,$7C,$65,$B1,$9B,$F9,$51,$CC,$C9,$5C,$9E,$88,$EB,$19
	.dc.b	$7E,$79,$E8,$3E,$05,$79,$A7,$84,$F4,$69,$D2,$7D,$D3,$FC,$AF,$9F
	.dc.b	$FC,$C7,$F4,$3F,$E7,$BC,$BF,$C7,$3F,$41,$6B,$FC,$CB,$E5,$2B,$18
	.dc.b	$30,$FF,$35,$DE,$1E,$77,$FE,$63,$8C,$5F,$18,$C1,$FE,$67,$FA,$0C
	.dc.b	$9A,$9B,$19,$25,$FF,$9A,$63,$2F,$F9,$AF,$E8,$BA,$9D,$BE,$2A,$D7
	.dc.b	$F9,$CE,$2C,$FF,$9C,$66,$E7,$FE,$56,$F9,$EF,$29,$E6,$F9,$7B,$87
	.dc.b	$F9,$EE,$2D,$FF,$9E,$66,$D7,$FA,$2F,$F5,$4F,$E0,$B7,$FA,$03,$FD
	.dc.b	$16,$E2,$BF,$B1,$C6,$38,$FE,$34,$E6,$B9,$72,$60,$CC,$DA,$FF,$5C
	.dc.b	$FD,$83,$DF,$FA,$38,$A3,$13,$E1,$8F,$EE,$3F,$34,$ED,$DD,$6A,$7F
	.dc.b	$D7,$ED,$7F,$16,$D7,$F7,$6D,$7F,$91,$35,$7C,$FF,$FA,$BE,$FF,$D9
	.dc.b	$AD,$8F,$79,$E2,$E0,$F3,$C6,$74,$FF,$63,$3B,$9F,$CD,$B5,$FE,$16
	.dc.b	$D7,$F9,$93,$61,$8B,$83,$F2,$71,$86,$C3,$82,$C4,$C7,$E3,$56,$DB
	.dc.b	$5B,$6C,$EF,$0F,$5D,$6D,$B5,$B6,$CF,$D4,$F9,$DF,$A1,$7F,$EF,$06
	.dc.b	$CF,$2F,$AB,$FF,$2F,$4E,$75,$E0,$7F,$BA,$67,$F9,$5D,$70,$7F,$7E
	.dc.b	$D7,$31,$B7,$F6,$5F,$24,$D7,$F6,$4F,$C5,$BF,$C2,$73,$A3,$51,$87
	.dc.b	$55,$F6,$27,$55,$CB,$2E,$6D,$BF,$8F,$C9,$F3,$8F,$A3,$EC,$7E,$6E
	.dc.b	$3F,$A0,$EA,$9D,$47,$03,$E8,$CE,$F1,$FC,$BC,$E9,$F9,$AF,$4E,$4C
	.dc.b	$6F,$FA,$FF,$A1,$FD,$49,$FA,$AF,$94,$B8,$FB,$B9,$D3,$8C,$7D,$90
	.dc.b	$AC,$5A,$FE,$7E,$BF,$64,$FE,$35,$F3,$0F,$4B,$13,$F6,$E7,$E8,$62
	.dc.b	$E3,$F9,$2F,$D6,$8C,$C6,$23,$F8,$1F,$C9,$7F,$00,$3F,$80,$FC,$95
	.dc.b	$BF,$44,$59,$F3,$CD,$CB,$F8,$74,$DA,$FE,$1B,$36,$BF,$99,$69,$FD
	.dc.b	$EA,$FF,$8A,$7C,$87,$F0,$EF,$13,$F8,$7B,$70,$F8,$CD,$BF,$8A,$FF
	.dc.b	$14,$F1,$7F,$8A,$FC,$7D,$AF,$98,$DA,$F9,$6D,$C3,$F9,$13,$FB,$2E
	.dc.b	$A7,$AD,$4C,$7A,$D7,$0F,$D8,$CE,$9B,$3A,$69,$30,$FD,$44,$65,$B1
	.dc.b	$F7,$FA,$FF,$95,$9F,$CB,$3E,$56,$C7,$E8,$3E,$C3,$E2,$3F,$E3,$BE
	.dc.b	$D4,$62,$FF,$AD,$2D,$BF,$AD,$39,$97,$07,$F2,$9D,$BF,$A0,$78,$5F
	.dc.b	$CE,$BE,$37,$F9,$DB,$FC,$F3,$E5,$A6,$7F,$93,$F9,$6B,$A7,$E4,$BE
	.dc.b	$5B,$FA,$09,$FD,$0B,$DA,$B5,$F9,$9E,$1F,$E8,$A7,$F4,$6E,$3B,$1F
	.dc.b	$9D,$F9,$4F,$E9,$27,$F6,$FF,$D5,$7F,$CA,$D1,$93,$E0,$61,$EA,$3D
	.dc.b	$78,$58,$FE,$07,$2F,$95,$E5,$E9,$F3,$6C,$EB,$E5,$7B,$78,$1D,$DB
	.dc.b	$1C,$BE,$57,$C4,$78,$FD,$3B,$1D,$BE,$57,$CC,$79,$FE,$56,$C7,$F1
	.dc.b	$4F,$AD,$F4,$F1,$7F,$11,$EA,$70,$FF,$82,$E7,$3D,$1F,$B7,$AF,$AF
	.dc.b	$39,$DA,$61,$F7,$E2,$CF,$F1,$BB,$6F,$CC,$CE,$1C,$FA,$93,$AC,$3E
	.dc.b	$2F,$C6,$7C,$D3,$E7,$7C,$B5,$8F,$93,$F1,$7F,$44,$C6,$5F,$CA,$79
	.dc.b	$11,$99,$F9,$FF,$15,$6B,$4F,$B7,$92,$4F,$C3,$E5,$B5,$E9,$9B,$AF
	.dc.b	$9F,$4F,$E2,$97,$F2,$67,$2F,$E7,$61,$1F,$26,$F3,$7F,$E3,$7C,$9F
	.dc.b	$1B,$F8,$93,$F1,$5F,$1B,$71,$FC,$2C,$5A,$F6,$4D,$CB,$E4,$5B,$E5
	.dc.b	$BF,$24,$7E,$4F,$D2,$B8,$FE,$3A,$3F,$2D,$F2,$D8,$C7,$E3,$3F,$99
	.dc.b	$F8,$DF,$B9,$F0,$3F,$CD,$67,$CD,$18,$F4,$BF,$71,$97,$F8,$36,$B5
	.dc.b	$CA,$85,$6E,$F8,$E5,$FE,$AC,$C3,$8C,$E6,$F8,$7C,$A5,$C3,$8D,$EF
	.dc.b	$0F,$F6,$07,$CC,$BE,$37,$F4,$D1,$FE,$B3,$CB,$FD,$3F,$39,$EB,$E1
	.dc.b	$3F,$43,$9F,$C4,$A7,$4F,$13,$AD,$F4,$D7,$E0,$46,$E3,$18,$7F,$AB
	.dc.b	$39,$7F,$17,$97,$6C,$39,$E1,$C1,$B5,$E4,$BE,$38,$7F,$B6,$70,$E2
	.dc.b	$F8,$4E,$D9,$73,$9D,$C6,$0D,$4F,$0C,$3B,$4E,$B0,$EF,$5E,$AB,$F8
	.dc.b	$37,$BC,$3C,$51,$D4,$FF,$4B,$CE,$CB,$5F,$11,$FF,$4D,$B5,$FF,$65
	.dc.b	$FC,$2B,$CF,$7C,$D7,$29,$4E,$63,$0D,$8E,$ED,$7A,$FC,$9F,$FB,$E9
	.dc.b	$F0,$B5,$CF,$B5,$D7,$04,$E6,$7E,$FB,$C7,$5F,$A0,$DC,$F3,$7F,$57
	.dc.b	$D8,$FF,$71,$D6,$A3,$EC,$56,$EF,$F1,$AD,$7F,$0D,$5C,$2F,$9C,$7D
	.dc.b	$7C,$3C,$77,$C8,$ED,$14,$D5,$F3,$9D,$C7,$1B,$E8,$F8,$2B,$86,$1E
	.dc.b	$58,$D1,$FA,$1E,$37,$CD,$7D,$18,$C3,$5F,$CD,$46,$B5,$EA,$F0,$F4
	.dc.b	$ED,$6E,$FF,$6E,$DD,$3F,$B9,$37,$0F,$F6,$53,$FE,$55,$7B,$FA,$67
	.dc.b	$FC,$B3,$1F,$EA,$16,$FA,$19,$7F,$CB,$10,$9C,$76,$7F,$3D,$8C,$4F
	.dc.b	$02,$1C,$27,$7C,$BE,$4F,$42,$CF,$91,$BA,$7C,$46,$DB,$D0,$B1,$2B
	.dc.b	$CE,$EA,$38,$FC,$D3,$D9,$A4,$AF,$F3,$AC,$3D,$9C,$47,$F9,$CE,$38
	.dc.b	$3A,$9E,$73,$AD,$3F,$AD,$56,$AB,$85,$69,$F9,$3E,$8F,$A3,$D1,$91
	.dc.b	$82,$1F,$2A,$3D,$A7,$BB,$D3,$91,$AC,$3E,$64,$E2,$3C,$DE,$B7,$C6
	.dc.b	$F0,$1F,$33,$4F,$83,$1F,$0B,$DA,$95,$F0,$1F,$8B,$ED,$7C,$6E,$9B
	.dc.b	$3F,$12,$30,$FE,$0B,$D1,$FE,$05,$C9,$F9,$66,$F2,$7F,$82,$27,$F8
	.dc.b	$7C,$E5,$BE,$27,$F3,$A7,$FD,$3C,$DD,$BF,$C0,$4D,$E6,$FF,$23,$3F
	.dc.b	$E0,$89,$FF,$54,$BF,$DB,$1B,$E6,$7E,$12,$3F,$D0,$D1,$AF,$13,$1E
	.dc.b	$B6,$7D,$69,$3C,$A3,$F0,$7D,$92,$3F,$02,$FE,$1B,$D8,$E9,$EB,$C9
	.dc.b	$F0,$BD,$FD,$7F,$07,$AB,$27,$CA,$F8,$FD,$5F,$26,$92,$7D,$2F,$9F
	.dc.b	$D5,$F4,$7A,$9F,$5F,$E6,$65,$7B,$A3,$D9,$E9,$FE,$83,$C5,$FF,$18
	.dc.b	$4F,$FB,$15,$BC,$9F,$22,$7F,$4B,$F2,$FF,$26,$9C,$C8,$C3,$7D,$0D
	.dc.b	$FB,$1F,$27,$D5,$97,$F3,$9F,$99,$EA,$FE,$CA,$1C,$2D,$B0,$D8,$FE
	.dc.b	$0D,$A7,$9A,$31,$08,$F4,$FC,$D7,$A3,$AF,$F8,$45,$6A,$78,$5F,$43
	.dc.b	$C1,$5B,$BF,$E2,$55,$BB,$E3,$5B,$9F,$0C,$7F,$13,$AE,$1A,$7F,$90
	.dc.b	$BD,$6F,$C3,$78,$72,$FF,$17,$1F,$89,$30,$FE,$52,$30,$FE,$43,$0F
	.dc.b	$E2,$3F,$17,$D0,$EB,$CB,$E3,$9F,$9E,$DC,$13,$BC,$BF,$2D,$7C,$BE
	.dc.b	$66,$61,$E8,$3E,$BD,$23,$BF,$CD,$6A,$7E,$55,$FF,$AA,$6B,$51,$C7
	.dc.b	$0E,$F7,$C2,$31,$3A,$78,$9D,$A3,$A6,$DF,$E7,$19,$61,$F7,$FE,$6B
	.dc.b	$FC,$D3,$7E,$53,$F0,$3C,$1F,$80,$35,$A4,$8F,$F0,$24,$F0,$7F,$C1
	.dc.b	$ED,$7F,$A3,$07,$0F,$AF,$DA,$F9,$A4,$7D,$3E,$C5,$7D,$3A,$7C,$C9
	.dc.b	$C6,$7E,$9A,$FC,$9F,$4C,$BF,$47,$1F,$F9,$B9,$BE,$E6,$5F,$86,$7E
	.dc.b	$72,$71,$7F,$99,$9B,$C1,$F9,$69,$B5,$F9,$29,$FD,$7E,$9F,$B0,$EB
	.dc.b	$93,$F4,$FA,$7E,$16,$3F,$E9,$D6,$E3,$FE,$E1,$55,$C2,$BD,$B1,$A7
	.dc.b	$FB,$86,$DF,$E6,$C7,$D7,$E1,$7E,$C0,$FA,$DC,$12,$3E,$8B,$FE,$35
	.dc.b	$7E,$73,$67,$94,$7F,$90,$BE,$65,$E7,$7E,$F5,$B1,$F7,$DA,$FC,$65
	.dc.b	$F0,$1B,$99,$3B,$79,$46,$23,$84,$5D,$3F,$8C,$D7,$29,$FC,$67,$CD
	.dc.b	$3E,$57,$C5,$EC,$49,$CC,$7F,$11,$9D,$47,$F9,$45,$AF,$F9,$71,$AF
	.dc.b	$E3,$9A,$FF,$85,$79,$5E,$9F,$F1,$BF,$97,$F1,$4F,$E0,$9B,$97,$F2
	.dc.b	$EA,$DC,$FE,$85,$AF,$F9,$A9,$AF,$E7,$1A,$F7,$AF,$EC,$BD,$4F,$9A
	.dc.b	$FE,$12,$49,$FD,$AF,$CD,$57,$F3,$B3,$B7,$F7,$33,$FC,$42,$7F,$88
	.dc.b	$FC,$D4,$9F,$DE,$BF,$C5,$32,$C4,$72,$8B,$97,$EF,$67,$F8,$CF,$CD
	.dc.b	$4A,$FE,$0A,$FF,$1C,$D7,$11,$E2,$8B,$5F,$C1,$67,$F9,$0F,$C8,$4A
	.dc.b	$FE,$1A,$FF,$24,$F9,$9C,$47,$AA,$2D,$7F,$0D,$9F,$E5,$3F,$35,$BA
	.dc.b	$FE,$2B,$18,$7F,$8B,$C6,$5F,$87,$19,$7F,$6F,$F0,$CF,$5F,$FA,$64
	.dc.b	$3E,$69,$D5,$FE,$53,$3F,$C1,$11,$97,$1E,$5F,$F9,$5E,$77,$3F,$E5
	.dc.b	$DB,$E9,$97,$F9,$7E,$73,$3F,$E9,$EC,$3B,$C6,$30,$EA,$DC,$3A,$47
	.dc.b	$FA,$A2,$76,$DB,$E8,$F8,$AE,$9F,$E6,$D7,$99,$B7,$D0,$79,$67,$9B
	.dc.b	$BA,$EE,$D3,$99,$E7,$AB,$87,$B2,$39,$9F,$1A,$78,$B4,$DD,$76,$79
	.dc.b	$1C,$2F,$CA,$F7,$39,$D7,$5C,$3E,$3B,$67,$DD,$3C,$6F,$BC,$E6,$3D
	.dc.b	$98,$7A,$2B,$AC,$78,$CE,$B8,$78,$EB,$84,$67,$0F,$A0,$FA,$2E,$3D
	.dc.b	$CE,$98,$78,$2F,$91,$ED,$C3,$B9,$E0,$C3,$A9,$D9,$CC,$F1,$BE,$B8
	.dc.b	$78,$30,$EC,$79,$A3,$35,$F0,$27,$47,$C4,$C3,$C3,$3F,$EA,$BF,$85
	.dc.b	$7C,$BF,$DF,$6E,$1F,$10,$38,$7E,$9B,$5F,$F5,$A7,$C2,$F3,$9F,$F7
	.dc.b	$EB,$63,$FB,$7B,$5F,$F7,$19,$D2,$36,$FE,$12,$9E,$EB,$C2,$FC,$5C
	.dc.b	$F6,$35,$19,$7F,$16,$67,$2F,$D3,$D7,$2A,$E3,$2C,$3F,$2B,$39,$8E
	.dc.b	$26,$4E,$B3,$C2,$77,$7D,$EF,$F6,$75,$DF,$3F,$F6,$6C,$E5,$F5,$CF
	.dc.b	$8A,$31,$3F,$08,$EA,$78,$2B,$AD,$F7,$8C,$D7,$67,$C8,$F3,$8D,$36
	.dc.b	$39,$1C,$2F,$C2,$79,$27,$6D,$8D,$CF,$13,$B5,$78,$A3,$33,$B9,$CC
	.dc.b	$75,$3D,$4F,$18,$EF,$1F,$E7,$0B,$F7,$BF,$7A,$74,$F0,$9E,$87,$07
	.dc.b	$FC,$E9,$3C,$1F,$98,$73,$BF,$4B,$4F,$29,$6B,$F9,$4F,$91,$FE,$5E
	.dc.b	$F9,$AF,$C2,$65,$FE,$4B,$F9,$9B,$87,$F9,$5E,$6E,$1F,$E5,$E9,$B7
	.dc.b	$C6,$BF,$CA,$F1,$8E,$A5,$D8,$C7,$CD,$4A,$D9,$FE,$5E,$71,$87,$F6
	.dc.b	$A7,$84,$64,$FF,$31,$FC,$BF,$53,$B7,$C5,$D8,$C9,$2B,$FB,$9C,$65
	.dc.b	$FE,$EB,$F3,$1E,$53,$CD,$F2,$F6,$BF,$CD,$F1,$67,$FB,$DC,$DD,$7F
	.dc.b	$51,$AB,$E4,$FF,$7B,$87,$F9,$E5,$B7,$FE,$5E,$7F,$82,$C6,$79,$3E
	.dc.b	$3C,$6A,$F9,$B6,$55,$FE,$C0,$8E,$2F,$3A,$FA,$2B,$E6,$9F,$01,$FE
	.dc.b	$CC,$8E,$26,$5C,$3D,$23,$77,$FE,$7A,$8C,$4E,$9E,$11,$FE,$9B,$C1
	.dc.b	$F3,$2E,$BF,$0E,$1C,$AF,$FD,$27,$5A,$9F,$2F,$CD,$3D,$9E,$8D,$C7
	.dc.b	$C1,$68,$C3,$E3,$6B,$FE,$8E,$73,$7F,$94,$FA,$47,$C8,$F6,$C3,$F2
	.dc.b	$72,$9C,$E1,$E9,$CD,$C5,$A6,$C6,$78,$B1,$32,$38,$3B,$8C,$67,$EE
	.dc.b	$AE,$2D,$9D,$57,$03,$FD,$5D,$F3,$2F,$0B,$E6,$36,$3E,$0B,$6B,$E9
	.dc.b	$35,$FF,$56,$B9,$CB,$F6,$7F,$47,$FE,$8E,$AE,$B7,$F0,$23,$17,$FE
	.dc.b	$AC,$F9,$47,$3F,$ED,$1F,$3E,$1F,$98,$73,$87,$FB,$3D,$CB,$F9,$0E
	.dc.b	$64,$6A,$FA,$67,$CF,$0E,$8F,$CB,$E7,$98,$7E,$1F,$E5,$FF,$35,$6C
	.dc.b	$BF,$13,$5A,$8F,$94,$FD,$89,$DC,$EA,$BE,$C4,$EE,$3F,$89,$CE,$B6
	.dc.b	$4F,$EA,$BE,$72,$5F,$EC,$7D,$0B,$8F,$EC,$66,$E3,$FA,$E9,$B5,$FA
	.dc.b	$6E,$0F,$B7,$7F,$43,$3F,$AE,$49,$FA,$EF,$F1,$EF,$9D,$79,$7F,$B2
	.dc.b	$BF,$93,$6E,$05,$6B,$0F,$E0,$9F,$29,$FC,$00,$FE,$03,$F2,$D6,$32
	.dc.b	$4B,$FE,$11,$19,$7F,$85,$4D,$D3,$F8,$3C,$DA,$FB,$1D,$DF,$DC,$E7
	.dc.b	$F8,$5F,$C9,$FF,$0E,$F1,$7F,$87,$B6,$BF,$8B,$37,$1F,$E2,$CF,$F1
	.dc.b	$4F,$27,$F8,$AF,$C9,$DA,$FE,$38,$DA,$FE,$38,$DC,$3F,$8C,$BF,$B2
	.dc.b	$F9,$CF,$F6,$9D,$7F,$09,$6D,$7F,$06,$7F,$DA,$31,$9F,$63,$81,$1E
	.dc.b	$1E,$CD,$AE,$13,$70,$FE,$AB,$7C,$A7,$37,$FC,$C7,$AB,$F9,$69,$FC
	.dc.b	$BB,$E2,$EC,$7F,$2C,$BF,$E7,$3F,$39,$F1,$E3,$F4,$0D,$BF,$B4,$B5
	.dc.b	$FE,$BA,$E7,$8F,$FA,$07,$8D,$FD,$83,$17,$C3,$7F,$3C,$8C,$3F,$CF
	.dc.b	$BC,$4D,$7C,$35,$CF,$C4,$7E,$BE,$D1,$FD,$13,$E5,$E4,$FF,$4B,$F6
	.dc.b	$EE,$BF,$D2,$F8,$7F,$A4,$9F,$D2,$B8,$AD,$7F,$51,$F9,$6F,$E9,$A7
	.dc.b	$F7,$4F,$97,$FE,$97,$19,$3C,$37,$EE,$C7,$DA,$7B,$B1,$B3,$AD,$D6
	.dc.b	$F8,$EE,$9E,$BF,$56,$D6,$E5,$F2,$BE,$1F,$96,$F1,$36,$75,$F2,$DE
	.dc.b	$53,$CD,$EC,$58,$E5,$F2,$BF,$D6,$33,$F0,$B7,$13,$4D,$7C,$E7,$BB
	.dc.b	$7F,$E2,$17,$F7,$38,$3F,$C2,$B3,$9E,$AF,$C0,$C6,$36,$95,$EF,$E9
	.dc.b	$4E,$30,$E6,$7E,$F7,$07,$07,$E0,$7E,$13,$AB,$70,$F8,$5F,$1D,$F4
	.dc.b	$7E,$33,$E9,$36,$3E,$34,$5C,$3A,$EF,$F3,$AB,$1A,$FF,$83,$B7,$5F
	.dc.b	$DB,$F8,$CC,$E3,$FC,$DE,$7F,$A0,$D7,$F8,$53,$4F,$A3,$3F,$65,$B9
	.dc.b	$76,$5C,$9F,$13,$92,$56,$DF,$8F,$18,$AE,$11,$6F,$B4,$DA,$F1,$7C
	.dc.b	$77,$E3,$CF,$C8,$7C,$75,$8F,$C2,$45,$C3,$DF,$92,$B1,$87,$BF,$E5
	.dc.b	$3F,$31,$F2,$9F,$99,$6E,$3F,$8D,$6E,$1C,$9B,$5F,$1F,$E3,$FF,$D3
	.dc.b	$9F,$08,$F4,$7E,$34,$61,$F9,$1E,$DD,$C3,$F9,$C6,$8A,$C7,$27,$E8
	.dc.b	$1B,$19,$CF,$FD,$6F,$87,$26,$DF,$2A,$E8,$BE,$6D,$CF,$FA,$98,$FF
	.dc.b	$55,$5F,$F8,$DA,$73,$D5,$CA,$7F,$53,$CD,$CA,$73,$1C,$0F,$F2,$2B
	.dc.b	$8D,$7D,$D1,$8C,$3B,$46,$23,$9E,$5D,$6B,$86,$3E,$1B,$DB,$71,$F1
	.dc.b	$72,$7E,$42,$F8,$56,$B0,$ED,$58,$37,$97,$9E,$75,$3F,$05,$B8,$F8
	.dc.b	$23,$FC,$BB,$87,$57,$D8,$6E,$FA,$3F,$E5,$59,$F8,$2F,$F9,$70,$D1
	.dc.b	$9A,$F9,$8E,$7E,$1D,$E5,$BE,$73,$FC,$9F,$18,$6C,$66,$7F,$CC,$53
	.dc.b	$97,$C1,$F2,$7F,$EF,$86,$FC,$1F,$F9,$63,$14,$E6,$7C,$91,$89,$D4
	.dc.b	$F1,$DF,$E9,$4E,$6B,$DE,$D8,$CF,$AF,$FE,$78,$77,$5C,$78,$3F,$0C
	.dc.b	$DB,$FC,$0B,$5F,$C7,$57,$19,$DE,$1F,$7E,$B5,$87,$79,$FC,$85,$6E
	.dc.b	$F9,$DE,$C8,$F9,$A3,$4F,$E6,$EF,$99,$27,$F1,$93,$CF,$8D,$E4,$FC
	.dc.b	$6B,$FD,$2C,$63,$E3,$F5,$26,$E1,$2B,$73,$96,$CE,$5F,$D1,$57,$6F
	.dc.b	$8F,$FF,$4F,$F4,$7F,$AD,$2D,$F1,$F8,$E6,$FE,$DC,$DC,$3F,$C6,$D2
	.dc.b	$BC,$61,$CE,$7F,$B0,$4E,$5B,$3A,$8F,$F2,$0C,$6B,$E3,$F0,$47,$8F
	.dc.b	$C7,$D7,$F9,$EC,$A7,$19,$79,$EB,$C3,$F3,$5F,$E9,$CB,$FE,$E0,$D7
	.dc.b	$FA,$2B,$67,$FD,$A2,$BE,$73,$8F,$A3,$2B,$8B,$D2,$31,$1C,$A2,$D7
	.dc.b	$29,$F6,$7C,$2F,$B7,$E2,$65,$FC,$28,$CB,$48,$7B,$3E,$47,$E2,$11
	.dc.b	$F9,$2F,$C7,$F8,$EF,$6F,$A3,$27,$DA,$FC,$0F,$4A,$47,$C0,$6B,$F0
	.dc.b	$3D,$4F,$8B,$EA,$7C,$66,$47,$C6,$8F,$A9,$F1,$D2,$3E,$CB,$F5,$FE
	.dc.b	$3B,$EC,$7C,$4C,$9F,$BA,$FD,$BE,$BF,$B9,$F1,$7F,$77,$E2,$25,$7E
	.dc.b	$0A,$3F,$0B,$FA,$B7,$EF,$F8,$FF,$E0,$0A,$F9,$2C,$9A,$F0,$3F,$17
	.dc.b	$E5,$7F,$14,$A7,$7F,$FA,$B6,$30,$DF,$27,$7E,$B7,$6F,$8A,$9A,$7F
	.dc.b	$8C,$73,$47,$8B,$C5,$F1,$FC,$3C,$AD,$3E,$6F,$87,$F3,$FC,$34,$9E
	.dc.b	$4F,$AB,$E0,$FD,$7F,$0D,$27,$B3,$EE,$F3,$BD,$FF,$0B,$27,$C4,$FC
	.dc.b	$2F,$1F,$DB,$C3,$27,$FC,$8C,$FF,$85,$9B,$EE,$7C,$8F,$85,$91,$EB
	.dc.b	$7E,$5F,$CE,$7E,$CE,$1F,$35,$FD,$99,$FF,$2A,$98,$F5,$75,$19,$8F
	.dc.b	$F1,$99,$C3,$D1,$CC,$C8,$E3,$3B,$E8,$FA,$30,$38,$1A,$9E,$0F,$49
	.dc.b	$E1,$D1,$B8,$D9,$FE,$36,$E6,$FD,$94,$7F,$B9,$DF,$D9,$47,$D7,$F2
	.dc.b	$7F,$0F,$E5,$48,$C1,$0E,$4F,$E3,$3C,$5F,$C6,$F8,$D2,$7F,$01,$19
	.dc.b	$7F,$05,$19,$7C,$51,$F9,$4F,$1A,$47,$E1,$DF,$CB,$F8,$D8,$8F,$44
	.dc.b	$5D,$7F,$33,$37,$3F,$54,$DA,$FF,$6D,$66,$AF,$AD,$F2,$FF,$E7,$09
	.dc.b	$FE,$E4,$7C,$B7,$0E,$1A,$FF,$77,$6B,$F2,$DD,$E1,$F5,$21,$18,$C3
	.dc.b	$5F,$09,$FB,$0F,$83,$FE,$6B,$FA,$BF,$F8,$07,$0F,$B7,$E9,$A3,$17
	.dc.b	$F7,$63,$18,$F0,$F0,$F8,$B0,$FF,$4E,$57,$07,$FD,$F0,$FE,$D1,$FC
	.dc.b	$37,$C2,$3E,$6B,$F2,$CD,$8F,$CA,$B5,$FC,$46,$1A,$99,$1A,$7F,$33
	.dc.b	$18,$8F,$CC,$CF,$1C,$7E,$13,$67,$F2,$1C,$9F,$16,$7D,$FD,$59,$B6
	.dc.b	$6F,$96,$D6,$1F,$05,$B3,$AA,$E1,$5F,$E8,$AC,$BF,$AF,$D6,$A3,$95
	.dc.b	$FF,$18,$9F,$BF,$5F,$7A,$FE,$F3,$5F,$AE,$DD,$FF,$80,$47,$E0,$63
	.dc.b	$8B,$FB,$A8,$CE,$3E,$EF,$92,$F9,$7F,$0B,$2B,$E5,$BF,$3B,$E1,$71
	.dc.b	$1F,$1A,$2D,$7C,$69,$B9,$7E,$D7,$65,$62,$FE,$CD,$7E,$63,$E0,$7C
	.dc.b	$C4,$8F,$66,$5F,$E7,$F6,$C6,$DF,$43,$F8,$0F,$86,$FE,$2D,$F3,$0F
	.dc.b	$6B,$E1,$36,$BE,$33,$5F,$D2,$B5,$FF,$2C,$39,$D3,$F4,$0D,$D3,$F8
	.dc.b	$73,$67,$F8,$B7,$CC,$FF,$4D,$7F,$2F,$E7,$7E,$88,$8F,$E6,$BE,$12
	.dc.b	$BF,$EC,$AE,$6F,$F6,$AB,$C0,$D1,$FA,$0C,$FF,$56,$47,$F4,$7F,$19
	.dc.b	$5E,$51,$73,$FD,$7D,$2B,$F6,$7F,$19,$2F,$F5,$6F,$ED,$BA,$F1,$1E
	.dc.b	$38,$B7,$CA,$7F,$77,$F3,$B2,$BF,$66,$FE,$FB,$93,$11,$EB,$8B,$5E
	.dc.b	$29,$FE,$07,$F3,$BB,$AF,$DD,$C6,$1F,$DE,$C6,$5F,$89,$19,$7F,$A4
	.dc.b	$FC,$3B,$CE,$5F,$F4,$5F,$9B,$67,$FD,$04,$F9,$76,$8C,$BF,$D2,$E3
	.dc.b	$31,$88,$5F,$5A,$E9,$5F,$D2,$E3,$13,$CE,$3B,$70,$7F,$4F,$BE,$51
	.dc.b	$D0,$E1,$87,$9A,$B8,$1F,$D6,$EF,$9D,$70,$BE,$57,$D2,$BC,$26,$EF
	.dc.b	$95,$79,$EF,$8D,$6E,$BA,$9F,$0A,$F9,$BC,$70,$E6,$FB,$B0,$E7,$5C
	.dc.b	$70,$F1,$4E,$B0,$EA,$F1,$3A,$CF,$B5,$F9,$58,$76,$7A,$61,$C7,$0E
	.dc.b	$B1,$A3,$C3,$5C,$5D,$9E,$8B,$EE,$FA,$4D,$DF,$58,$C9,$A9,$E4,$73
	.dc.b	$73,$5E,$0A,$E4,$7A,$2B,$94,$66,$79,$47,$A6,$B9,$4E,$4E,$4F,$68
	.dc.b	$E3,$7E,$D3,$CB,$70,$F5,$57,$18,$C6,$1F,$32,$73,$1C,$57,$AD,$FC
	.dc.b	$1D,$F8,$6D,$C7,$F1,$ED,$BF,$C5,$3B,$9F,$F1,$27,$C4,$7F,$89,$63
	.dc.b	$10,$8C,$7B,$1F,$E0,$06,$C7,$FB,$CD,$AF,$FB,$64,$FC,$4E,$0F,$53
	.dc.b	$FC,$63,$97,$E9,$67,$B9,$A8,$CE,$9D,$63,$1C,$FF,$8E,$77,$5F,$AD
	.dc.b	$BE,$93,$99,$F3,$1C,$0F,$CC,$9C,$2B,$FD,$8B,$7D,$A3,$D0,$6E,$FC
	.dc.b	$8F,$27,$B4,$6E,$3D,$91,$8A,$F0,$61,$FA,$2C,$3C,$75,$E0,$39,$61
	.dc.b	$E0,$AD,$C7,$78,$F7,$5F,$38,$E1,$7C,$F0,$F0,$47,$A4,$FF,$2C,$61
	.dc.b	$DA,$BC,$F5,$BC,$3A,$C7,$33,$35,$DC,$F7,$B9,$C3,$C2,$79,$63,$C2
	.dc.b	$6C,$F7,$B9,$8E,$67,$18,$E7,$19,$39,$BE,$6C,$3D,$2F,$C2,$31,$5E
	.dc.b	$69,$E1,$1C,$8C,$9F,$C7,$3C,$0F,$F1,$E7,$7F,$F9,$E3,$2F,$EA,$3E
	.dc.b	$85,$C3,$FC,$7D,$CC,$AC,$61,$FE,$3A,$AB,$87,$2B,$FF,$1F,$C6,$2F
	.dc.b	$84,$60,$FF,$21,$F7,$E4,$D4,$D8,$C9,$2F,$FB,$74,$65,$FE,$E1,$DF
	.dc.b	$D4,$ED,$F1,$96,$BF,$C9,$F1,$67,$FB,$B4,$DC,$FF,$B6,$F8,$3E,$53
	.dc.b	$CD,$F1,$B7,$0F,$F2,$BC,$5B,$FE,$FD,$36,$BF,$BA,$F8,$3E,$D3,$DD
	.dc.b	$F1,$B7,$0F,$F2,$FC,$5B,$FF,$03,$4D,$AF,$F3,$97,$CD,$23,$E5,$CA
	.dc.b	$31,$7F,$92,$3F,$27,$F3,$0F,$81,$DC,$EA,$BF,$D6,$AD,$7F,$D5,$47
	.dc.b	$3B,$FF,$40,$3F,$EA,$D8,$D3,$63,$2F,$FA,$DF,$E5,$DF,$0F,$E4,$6E
	.dc.b	$3E,$39,$FE,$F3,$F2,$AF,$4F,$57,$2B,$FC,$B5,$7C,$2F,$FD,$32,$D7
	.dc.b	$CA,$D7,$DB,$F0,$7F,$8C,$D5,$AA,$31,$C1,$FC,$C3,$0F,$15,$6A,$78
	.dc.b	$56,$E7,$2D,$BF,$F0,$7C,$7E,$F2,$F9,$46,$0F,$F0,$9C,$EE,$FF,$44
	.dc.b	$DA,$F2,$9E,$0A,$E3,$3A,$BF,$8D,$F0,$92,$BE,$37,$3B,$E7,$DE,$46
	.dc.b	$73,$FF,$75,$E8,$C3,$E8,$C2,$31,$87,$CA,$9C,$C6,$27,$EE,$E0,$F9
	.dc.b	$7F,$E4,$79,$FF,$AD,$97,$C0,$FE,$D1,$CE,$3F,$EB,$77,$2F,$99,$CD
	.dc.b	$FF,$A7,$9C,$BF,$16,$FB,$77,$73,$BF,$AD,$C4,$F5,$E9,$8F,$5A,$E5
	.dc.b	$F5,$A7,$4D,$9D,$34,$98,$FD,$3C,$D5,$C2,$30,$FF,$B9,$67,$73,$FE
	.dc.b	$66,$F9,$EF,$E1,$C7,$F0,$FF,$4E,$CF,$FB,$42,$2D,$FF,$9D,$E6,$D7
	.dc.b	$FB,$8F,$E5,$DE,$37,$EC,$9B,$87,$DA,$FD,$60,$E6,$63,$FE,$B7,$9D
	.dc.b	$E1,$FE,$F6,$6B,$FB,$26,$BF,$EE,$7A,$7C,$CF,$F2,$86,$D7,$7A,$C4
	.dc.b	$C8,$F1,$3B,$8C,$57,$D5,$AE,$2D,$9D,$57,$03,$D5,$EF,$F7,$F8,$0F
	.dc.b	$03,$63,$27,$BE,$32,$D3,$E1,$7C,$A7,$9F,$E4,$FD,$0D,$8C,$9F,$1E
	.dc.b	$32,$D3,$E5,$7C,$A7,$BC,$F8,$1F,$37,$63,$24,$CB,$D9,$19,$7D,$D3
	.dc.b	$76,$FA,$71,$FE,$F4,$D1,$59,$BF,$9D,$F2,$BF,$34,$F9,$DE,$E5,$8F
	.dc.b	$D0,$7C,$1F,$E9,$4F,$D3,$78,$76,$3F,$01,$E1,$FD,$53,$EB,$7C,$A5
	.dc.b	$8F,$C2,$7C,$A7,$D9,$3E,$D7,$CA,$58,$FC,$47,$CA,$7D,$D3,$EF,$7C
	.dc.b	$A5,$8F,$C6,$79,$7F,$81,$3F,$05,$E2,$58,$FC,$87,$CA,$7E,$14,$FC
	.dc.b	$37,$91,$63,$F8,$6F,$CC,$7F,$B1,$5B,$F0,$3F,$DC,$A3,$FD,$57,$18
	.dc.b	$3F,$89,$ED,$FC,$61,$72,$BC,$6F,$E7,$FD,$FF,$DE,$AB,$53,$97,$F3
	.dc.b	$BD,$1F,$42,$BF,$A2,$FC,$AE,$FE,$57,$83,$71,$EF,$F2,$BC,$CE,$9F
	.dc.b	$2B,$63,$C7,$E5,$77,$3C,$1E,$55,$8F,$6F,$C1,$FA,$FE,$0F,$D8,$D8
	.dc.b	$F8,$3E,$E7,$BC,$C6,$FF,$C5,$F8,$7F,$C1,$ED,$F6,$BE,$21,$BA,$F9
	.dc.b	$2F,$BF,$E0,$64,$7C,$6F,$95,$B5,$FC,$87,$8D,$F2,$BF,$14,$CF,$27
	.dc.b	$C3,$8C,$DF,$CE,$9F,$95,$F2,$BB,$8F,$A2,$FD,$5F,$7E,$47,$DA,$F7
	.dc.b	$EE,$3E,$99,$B8,$FC,$FD,$55,$8B,$FB,$3C,$E9,$5E,$70,$FB,$91,$C2
	.dc.b	$BF,$0F,$3F,$77,$E5,$64,$7E,$13,$E5,$6D,$FB,$E6,$E5,$ED,$9B,$5E
	.dc.b	$0D,$5F,$29,$F8,$73,$F1,$1F,$2B,$67,$F1,$F1,$F8,$BF,$23,$19,$7C
	.dc.b	$B7,$F1,$DE,$47,$EE,$BE,$1D,$DD,$D7,$D8,$C3,$DE,$6D,$FE,$4E,$6F
	.dc.b	$0F,$F8,$5A,$6E,$9F,$D1,$F0,$E7,$7F,$99,$9C,$DF,$28,$7C,$BB,$D4
	.dc.b	$7C,$A7,$FA,$D3,$3F,$D2,$D6,$A7,$2F,$FA,$D6,$33,$F1,$FF,$8E,$CB
	.dc.b	$BC,$EE,$F9,$3F,$AE,$C3,$B4,$62,$34,$73,$3F,$67,$5C,$EF,$C3,$5C
	.dc.b	$27,$47,$0B,$EF,$19,$7C,$35,$C6,$7C,$18,$7A,$27,$6D,$BE,$AD,$9F
	.dc.b	$0F,$67,$C5,$C3,$A5,$6B,$1F,$44,$62,$B7,$C1,$8C,$39,$47,$B0,$FF
	.dc.b	$28,$CF,$A5,$B1,$F2,$9B,$5F,$E5,$67,$F8,$9F,$C6,$3A,$DF,$31,$99
	.dc.b	$4E,$63,$0D,$8E,$4D,$7C,$3E,$C7,$FB,$D9,$ED,$8A,$31,$1A,$F8,$DE
	.dc.b	$D3,$96,$CF,$1D,$BE,$7B,$75,$FA,$AD,$BF,$BD,$E6,$7E,$22,$B7,$7F
	.dc.b	$22,$B5,$5F,$90,$6E,$1F,$E4,$18,$D4,$C7,$EF,$C7,$8D,$B7,$D7,$0F
	.dc.b	$0D,$6F,$0E,$F5,$A3,$13,$A7,$9E,$1E,$0B,$E5,$1F,$D0,$7B,$BE,$23
	.dc.b	$F9,$8E,$54,$3D,$9D,$C6,$1B,$1C,$27,$F3,$53,$C2,$F3,$52,$7A,$56
	.dc.b	$F1,$FD,$AD,$F6,$F6,$7F,$5F,$D9,$FD,$36,$9E,$CE,$39,$3F,$8F,$C6
	.dc.b	$31,$FF,$36,$AB,$D4,$7A,$DB,$87,$AA,$ED,$71,$AF,$F2,$E3,$C5,$E7
	.dc.b	$EB,$E8,$8F,$6F,$5E,$BC,$22,$E5,$E7,$9E,$FF,$31,$FC,$92,$BF,$D0
	.dc.b	$2D,$7F,$CD,$4D,$9F,$ED,$4B,$BF,$8F,$7C,$AE,$2F,$48,$C4,$72,$8B
	.dc.b	$5C,$A7,$DB,$DD,$EE,$F6,$25,$FC,$68,$CB,$48,$7C,$2F,$63,$E2,$9A
	.dc.b	$7E,$2C,$61,$F5,$F8,$1F,$32,$1F,$01,$4F,$83,$1F,$0B,$C2,$95,$F0
	.dc.b	$9F,$8B,$E1,$7C,$6A,$B3,$F1,$A3,$0F,$DC,$F6,$3F,$80,$71,$FE,$49
	.dc.b	$BC,$46,$98,$F5,$FE,$E9,$2B,$F0,$BD,$5F,$80,$F5,$FF,$03,$EB,$C9
	.dc.b	$FC,$4C,$7E,$1B,$E6,$3F,$81,$F1,$FF,$87,$5F,$F4,$1C,$FF,$AE,$9A
	.dc.b	$FF,$A8,$DD,$F5,$7F,$72,$F1,$BF,$8E,$57,$F8,$95,$B1,$9F,$17,$F2
	.dc.b	$70,$F9,$0F,$07,$AF,$31,$EC,$F8,$FC,$2F,$27,$AF,$27,$C4,$F9,$FD
	.dc.b	$7F,$47,$AF,$27,$CC,$FA,$FD,$7F,$67,$A9,$F5,$FE,$62,$57,$AE,$3E
	.dc.b	$07,$AB,$FA,$6F,$D6,$7F,$AF,$CF,$FA,$FD,$BC,$3F,$0E,$7D,$3B,$CA
	.dc.b	$FF,$6E,$A3,$DB,$EB,$FC,$BF,$52,$4F,$C9,$7E,$77,$A9,$FB,$68,$7C
	.dc.b	$CF,$F7,$17,$FC,$A7,$8F,$F9,$F1,$CF,$A9,$B9,$D4,$7F,$BD,$CE,$3E
	.dc.b	$7E,$66,$47,$2E,$7F,$EF,$35,$C2,$FA,$3F,$C5,$DB,$5C,$2B,$A4,$EE
	.dc.b	$B8,$DF,$E1,$7C,$8F,$C3,$79,$12,$FF,$09,$19,$7F,$0D,$19,$7F,$0D
	.dc.b	$1F,$8D,$F1,$3F,$1D,$E2,$C9,$FC,$64,$65,$FC,$74,$65,$FC,$74,$7E
	.dc.b	$57,$C4,$91,$F9,$37,$F3,$1E,$2E,$23,$F2,$91,$75,$FC,$D4,$DC,$FF
	.dc.b	$C8,$FD,$DF,$EF,$7B,$E4,$7E,$76,$7E,$A7,$CC,$3E,$5B,$F0,$9E,$04
	.dc.b	$AF,$98,$7F,$A2,$27,$4E,$03,$87,$FB,$E3,$5F,$C6,$B5,$FF,$16,$B5
	.dc.b	$FF,$49,$7C,$9F,$E4,$31,$FF,$2E,$B6,$32,$7F,$3C,$7F,$5A,$E0,$F2
	.dc.b	$F9,$08,$E0,$D8,$E1,$7F,$91,$F7,$5F,$0D,$49,$3E,$BE,$4F,$C9,$6D
	.dc.b	$F9,$3A,$DC,$FB,$DA,$FE,$4D,$B1,$F8,$EC,$B9,$CE,$6E,$CE,$63,$F5
	.dc.b	$31,$FD,$5B,$E6,$7B,$D7,$F9,$F1,$B3,$FE,$B5,$EA,$EB,$3F,$C3,$EB
	.dc.b	$83,$27,$99,$F4,$F5,$FD,$A3,$E7,$F9,$2F,$80,$7D,$EB,$91,$FE,$2D
	.dc.b	$9D,$BF,$62,$3F,$C6,$0F,$DC,$6E,$5F,$E8,$E8,$F8,$5E,$EF,$CB,$3E
	.dc.b	$67,$BF,$63,$24,$3E,$33,$97,$E4,$45,$D3,$E7,$4F,$E4,$B1,$FA,$DF
	.dc.b	$2B,$2F,$D9,$19,$7D,$0F,$9B,$23,$4F,$4F,$CC,$BD,$5F,$E2,$1F,$33
	.dc.b	$FE,$76,$CB,$F8,$73,$76,$FF,$3F,$D6,$F1,$FC,$E3,$5F,$F2,$B3,$5F
	.dc.b	$E4,$ED,$78,$DF,$F5,$8E,$4F,$E7,$D1,$F9,$46,$AC,$3F,$DD,$51,$86
	.dc.b	$FC,$0F,$E7,$7C,$D9,$5C,$5F,$D0,$47,$E8,$7E,$42,$4F,$57,$F4,$9F
	.dc.b	$0D,$88,$ED,$9F,$EA,$A7,$18,$78,$2B,$F5,$5B,$CB,$F3,$3F,$AD,$E9
	.dc.b	$C4,$72,$3F,$6B,$18,$AF,$4C,$FE,$CF,$E6,$65,$7B,$DF,$DB,$72,$62
	.dc.b	$39,$75,$7E,$D6,$71,$7F,$0A,$BF,$79,$F0,$DC,$2F,$E4,$46,$1F,$95
	.dc.b	$19,$79,$EE,$9D,$47,$F4,$5F,$83,$79,$7F,$3E,$1F,$32,$D7,$FD,$7A
	.dc.b	$7D,$47,$10,$35,$C5,$FE,$C2,$8C,$C7,$FA,$4A,$BA,$46,$2F,$BE,$7C
	.dc.b	$63,$78,$75,$3B,$E1,$CD,$FF,$66,$D7,$0C,$3C,$B1,$8C,$3A,$BC,$23
	.dc.b	$A3,$5D,$9F,$EA,$88,$DB,$70,$DB,$EB,$9F,$13,$6F,$C4,$EC,$E6,$7B
	.dc.b	$67,$8B,$F0,$8B,$3C,$9D,$1C,$2F,$D5,$87,$C0,$BE,$B3,$C4,$D6,$1E
	.dc.b	$67,$8C,$ED,$EB,$3C,$2B,$9D,$FC,$3C,$33,$50,$E6,$F9,$A7,$71,$ED
	.dc.b	$35,$7F,$57,$0F,$04,$EA,$7C,$93,$E1,$79,$91,$C1,$2F,$C5,$19,$7E
	.dc.b	$94,$66,$3E,$53,$DE,$FC,$33,$A3,$8C,$66,$7E,$7C,$7C,$B3,$E6,$57
	.dc.b	$94,$FC,$27,$C1,$FF,$94,$1B,$F1,$1E,$28,$C6,$3F,$9D,$7F,$CA,$B7
	.dc.b	$F9,$9F,$83,$FC,$D4,$62,$11,$8A,$FF,$78,$36,$3F,$C8,$2D,$7F,$DB
	.dc.b	$47,$E1,$32,$7A,$4E,$5F,$C2,$CF,$73,$51,$97,$F1,$66,$74,$FD,$4D
	.dc.b	$72,$AE,$33,$E4,$38,$B8,$BF,$C8,$CE,$E7,$99,$BA,$E7,$7D,$DF,$DA
	.dc.b	$C7,$7D,$3F,$D9,$73,$97,$DF,$1D,$8E,$77,$DA,$75,$1F,$08,$D5,$74
	.dc.b	$7C,$15,$D3,$0F,$34,$6A,$3A,$CF,$92,$72,$D9,$EE,$6C,$EC,$E4,$ED
	.dc.b	$7E,$39,$E3,$5D,$EB,$8D,$6E,$3B,$99,$78,$C7,$53,$EC,$BB,$7B,$C6
	.dc.b	$CE,$2F,$F9,$C2,$FE,$0B,$C2,$34,$F0,$35,$5E,$46,$9E,$48,$DF,$17
	.dc.b	$F8,$C7,$BB,$F0,$59,$7F,$5E,$F3,$AE,$1F,$E3,$C9,$B8,$7F,$90,$26
	.dc.b	$DF,$1A,$FE,$D1,$18,$E2,$FF,$47,$46,$0F,$ED,$9D,$F9,$35,$36,$32
	.dc.b	$4B,$FE,$8F,$19,$7F,$A4,$F7,$F5,$3B,$7C,$CD,$AF,$EF,$11,$67,$FA
	.dc.b	$7C,$DC,$FF,$BC,$FC,$AF,$94,$F3,$7C,$CD,$C3,$FC,$01,$16,$FF,$AB
	.dc.b	$CD,$AF,$F0,$0F,$83,$ED,$3D,$DF,$31,$70,$FF,$04,$45,$BF,$EB,$F3
	.dc.b	$6B,$FD,$4D,$F3,$28,$FF,$54,$4F,$F5,$91,$19,$BF,$F5,$EC,$7F,$5D
	.dc.b	$8C,$38,$6B,$FE,$C2,$73,$5F,$6B,$E7,$9E,$7F,$D8,$AE,$B8,$7D,$AA
	.dc.b	$56,$B0,$FF,$42,$7C,$C3,$E1,$3C,$B8,$7F,$6D,$8C,$4F,$92,$11,$87
	.dc.b	$CE,$D7,$FD,$AA,$E6,$FF,$1B,$F3,$CE,$1F,$24,$E6,$FF,$D9,$73,$97
	.dc.b	$FD,$33,$4F,$9B,$7C,$17,$A5,$B3,$AE,$17,$6E,$6A,$4F,$2A,$DE,$7F
	.dc.b	$E2,$5C,$18,$70,$AD,$5F,$0A,$D5,$F0,$BF,$D2,$DE,$E7,$35,$EA,$F8
	.dc.b	$3F,$8F,$1F,$11,$B9,$9A,$4C,$7F,$D4,$3F,$3C,$F9,$5E,$51,$87,$F5
	.dc.b	$BF,$37,$FC,$AF,$3F,$E6,$DD,$68,$7C,$0E,$BB,$F6,$DC,$3F,$D5,$B3
	.dc.b	$A6,$CE,$9F,$9B,$F3,$6F,$63,$F0,$8D,$8C,$FA,$F8,$99,$1C,$1D,$C6
	.dc.b	$37,$F9,$95,$C5,$B3,$AA,$E0,$7F,$9B,$7E,$6F,$F8,$31,$FC,$1F,$D5
	.dc.b	$B1,$92,$57,$F9,$22,$32,$FF,$97,$47,$CD,$BC,$8F,$A2,$DA,$FC,$B3
	.dc.b	$67,$F5,$6D,$7F,$60,$72,$9F,$E3,$7F,$06,$F3,$3F,$51,$5C,$78,$7E
	.dc.b	$8C,$E7,$0F,$F6,$CF,$CD,$3E,$6F,$F9,$1B,$63,$37,$89,$91,$C1,$DC
	.dc.b	$63,$2F,$F3,$7D,$71,$6C,$EA,$B8,$1E,$0F,$82,$EF,$F0,$5E,$06,$C6
	.dc.b	$4F,$2C,$65,$A7,$A3,$E5,$7C,$FF,$2B,$E8,$6C,$64,$F6,$C6,$5A,$7C
	.dc.b	$0F,$95,$F7,$9F,$03,$A6,$C6,$49,$97,$B2,$32,$FB,$A6,$ED,$F9,$CD
	.dc.b	$3F,$45,$7C,$23,$1F,$2B,$F3,$0F,$9B,$EE,$59,$FC,$EF,$C2,$7E,$98
	.dc.b	$FD,$3F,$C2,$D8,$FA,$BE,$3F,$DC,$F7,$BE,$AB,$63,$EC,$FC,$9F,$E0
	.dc.b	$3D,$FF,$B2,$D8,$FB,$BF,$25,$F7,$0F,$BB,$F3,$56,$3F,$03,$F2,$7F
	.dc.b	$80,$3F,$03,$E3,$58,$FC,$2F,$C9,$FE,$10,$FC,$2F,$93,$63,$FD,$31
	.dc.b	$C3,$FB,$CF,$9A,$7C,$27,$F6,$C6,$CE,$8F,$F4,$F7,$92,$8F,$F5,$24
	.dc.b	$67,$9D,$39,$AF,$EE,$BC,$1F,$C5,$F0,$FE,$12,$FE,$6E,$F9,$B9,$6F
	.dc.b	$83,$C3,$E5,$38,$FC,$A5,$9E,$D8,$FF,$45,$8E,$BF,$29,$27,$38,$3E
	.dc.b	$53,$1A,$7F,$B2,$A3,$FC,$88,$FF,$5F,$C3,$FA,$77,$97,$FD,$39,$B7
	.dc.b	$FD,$81,$B3,$FB,$67,$E0,$7C,$37,$C5,$F8,$6F,$8C,$DA,$F2,$F0,$7F
	.dc.b	$84,$2F,$FD,$B2,$DD,$FE,$75,$FC,$DF,$81,$F9,$DE,$BF,$F0,$0E,$2B
	.dc.b	$8F,$BB,$C9,$FA,$39,$7A,$7E,$17,$E5,$FC,$9F,$CC,$6D,$7C,$08,$B3
	.dc.b	$B8,$C1,$1F,$85,$F2,$BA,$D3,$E9,$BF,$7F,$E0,$E5,$7A,$BE,$0E,$DF
	.dc.b	$AA,$6D,$FE,$1E,$6D,$7E,$13,$75,$62,$FF,$07,$0A,$C7,$0F,$F7,$DF
	.dc.b	$94,$FC,$29,$F8,$6F,$94,$B8,$FD,$A8,$B8,$7D,$99,$B9,$7E,$2B,$E0
	.dc.b	$7F,$1B,$E3,$7E,$39,$B8,$7E,$09,$B7,$F8,$16,$D7,$E4,$BC,$6F,$F4
	.dc.b	$87,$C4,$BB,$3F,$27,$EB,$61,$FC,$AC,$DC,$3F,$3B,$D7,$FD,$36,$F8
	.dc.b	$46,$27,$5B,$7F,$86,$70,$B6,$9B,$1C,$E9,$F3,$0F,$49,$F2,$7F,$EB
	.dc.b	$3C,$FF,$C6,$93,$98,$C5,$07,$97,$F8,$FD,$3C,$71,$D3,$0E,$6F,$FA
	.dc.b	$EB,$0E,$71,$89,$DE,$1D,$6B,$47,$5B,$E9,$7C,$30,$F0,$CF,$56,$D7
	.dc.b	$1B,$F0,$DF,$0A,$F3,$4E,$6F,$94,$76,$39,$16,$78,$38,$AF,$2E,$1E
	.dc.b	$7A,$E3,$1D,$BB,$3E,$56,$1D,$6F,$BD,$FC,$07,$FB,$6C,$7B,$5B,$19
	.dc.b	$7C,$67,$F7,$2F,$8C,$7C,$D3,$E6,$BF,$C9,$D1,$86,$C6,$67,$FC,$CB
	.dc.b	$39,$7F,$CC,$F6,$F5,$BF,$C8,$71,$86,$BE,$46,$BF,$79,$AF,$C8,$78
	.dc.b	$E5,$F5,$9B,$9E,$FE,$13,$E8,$F9,$7F,$E7,$9A,$D4,$7D,$4A,$DD,$F3
	.dc.b	$8D,$B9,$63,$F4,$F0,$F4,$46,$A3,$13,$C9,$FC,$BB,$67,$71,$C6,$FC
	.dc.b	$58,$79,$63,$8D,$F6,$C3,$F1,$F5,$A9,$F2,$1A,$24,$69,$E1,$5C,$E3
	.dc.b	$DB,$1D,$CF,$D1,$F1,$BC,$07,$C6,$FF,$A6,$F9,$BF,$C4,$FF,$21,$7A
	.dc.b	$7F,$AC,$D2,$31,$87,$F5,$88,$E2,$FF,$31,$BC,$E1,$D5,$B3,$A9,$61
	.dc.b	$FE,$34,$F9,$0E,$70,$FD,$F1,$E9,$7F,$A5,$E3,$0D,$8C,$C7,$FA,$6D
	.dc.b	$AF,$FA,$C1,$B4,$CF,$FD,$6D,$18,$6D,$79,$3C,$8F,$2F,$93,$E6,$E8
	.dc.b	$98,$FF,$4B,$8C,$BF,$D7,$73,$47,$FA,$AF,$5F,$94,$D8,$CC,$EE,$38
	.dc.b	$BF,$03,$1F,$97,$1F,$13,$C9,$F8,$BE,$2C,$AF,$EB,$4F,$C8,$F1,$24
	.dc.b	$78,$5E,$C4,$7C,$9F,$1F,$A7,$B1,$F5,$63,$E8,$C7,$D2,$F8,$89,$3F
	.dc.b	$D9,$DF,$AB,$F1,$FF,$5B,$E3,$FD,$44,$9F,$80,$FB,$7E,$3B,$ED,$92
	.dc.b	$3F,$B8,$7C,$2F,$DE,$F8,$59,$1F,$DC,$DF,$C1,$FE,$B7,$F4,$E3,$FC
	.dc.b	$26,$FF,$9C,$5F,$F0,$43,$5F,$F0,$43,$5F,$F0,$83,$5F,$F1,$5B,$FE
	.dc.b	$8A,$79,$FC,$2F,$63,$BF,$89,$E0,$F0,$64,$F8,$1F,$1F,$C2,$79,$3E
	.dc.b	$13,$ED,$60,$69,$2B,$CB,$F0,$7E,$A2,$3E,$AF,$84,$F6,$FC,$1F,$BB
	.dc.b	$E0,$E4,$FB,$A3,$E0,$F9,$FD,$89,$1F,$01,$F0,$FC,$DF,$C6,$24,$7C
	.dc.b	$6F,$83,$F9,$3F,$3B,$F2,$BC,$19,$3F,$2A,$3F,$53,$F8,$FB,$D6,$FD
	.dc.b	$43,$67,$FE,$00,$35,$FF,$54,$87,$0F,$CE,$F8,$5F,$AE,$47,$EB,$FC
	.dc.b	$2F,$6F,$63,$F8,$0F,$A6,$FA,$DF,$F6,$DD,$7F,$96,$E3,$1F,$1D,$B8
	.dc.b	$FF,$2F,$1D,$EB,$CA,$47,$CA,$D7,$FD,$8D,$8A,$39,$47,$E7,$29,$F9
	.dc.b	$E3,$93,$F1,$BE,$94,$AF,$C6,$BF,$90,$F8,$8C,$47,$E9,$3C,$A7,$4F
	.dc.b	$FB,$F6,$36,$7E,$52,$7F,$2F,$F1,$12,$3F,$2E,$FF,$1C,$FD,$73,$FA
	.dc.b	$9F,$A1,$FE,$FC,$B4,$EE,$BB,$46,$1F,$F3,$8D,$70,$F3,$3F,$D0,$6F
	.dc.b	$FB,$86,$33,$F1,$3D,$4E,$DF,$13,$63,$24,$3D,$52,$8D,$46,$26,$E3
	.dc.b	$FA,$FC,$D7,$BA,$F1,$FA,$FE,$BF,$89,$91,$82,$1D,$A3,$E0,$1F,$07
	.dc.b	$E2,$7E,$17,$C4,$C9,$F6,$BF,$17,$9B,$11,$E6,$8B,$97,$8F,$1F,$C3
	.dc.b	$CF,$F1,$61,$F6,$DF,$E3,$47,$F0,$EF,$F8,$4E,$7F,$84,$46,$BE,$07
	.dc.b	$FA,$3C,$EE,$39,$35,$FF,$1F,$1F,$E3,$F7,$0E,$1A,$E7,$E0,$7F,$A7
	.dc.b	$47,$3A,$EE,$DE,$0E,$13,$FA,$0E,$6F,$E9,$57,$FE,$B8,$C5,$F1,$3F
	.dc.b	$85,$24,$7E,$03,$E0,$AB,$F6,$22,$E5,$F0,$E7,$F1,$9F,$13,$F8,$DF
	.dc.b	$89,$97,$F8,$88,$CB,$48,$7E,$0B,$E2,$3E,$1F,$C0,$7E,$D3,$C0,$7C
	.dc.b	$85,$33,$5F,$32,$7D,$E4,$7F,$81,$9F,$49,$C3,$F3,$1E,$2F,$EC,$7E
	.dc.b	$0F,$F3,$4F,$C1,$F1,$7F,$49,$F1,$32,$3F,$3F,$1F,$A7,$23,$83,$F4
	.dc.b	$7F,$41,$FA,$BF,$13,$F5,$9F,$15,$27,$F4,$F1,$96,$90,$FD,$0F,$C5
	.dc.b	$7F,$82,$3E,$6D,$E8,$FF,$81,$DB,$3E,$C6,$BF,$C8,$1A,$FE,$18,$3A
	.dc.b	$E3,$FC,$7B,$7A,$3F,$16,$D9,$FF,$45,$7C,$E7,$EE,$5F,$E1,$9F,$98
	.dc.b	$67,$C2,$FD,$7F,$17,$F0,$73,$FC,$47,$19,$26,$58,$33,$E9,$7F,$88
	.dc.b	$BF,$28,$FE,$29,$87,$F3,$66,$BE,$76,$BF,$9E,$6B,$FE,$9D,$78,$E9
	.dc.b	$FA,$86,$E9,$BF,$2B,$FC,$5F,$F8,$6F,$F5,$FB,$FE,$2B,$F4,$9F,$E1
	.dc.b	$BE,$E3,$49,$1F,$EF,$D8,$CB,$C0,$87,$F2,$DF,$86,$F9,$99,$7F,$11
	.dc.b	$6E,$3F,$C8,$2F,$F9,$51,$FE,$46,$8C,$D3,$74,$61,$B1,$B3,$87,$87
	.dc.b	$C7,$E5,$BF,$C6,$54,$69,$27,$FA,$FD,$74,$8E,$11,$FE,$C3,$C7,$CD
	.dc.b	$1E,$3F,$0F,$FC,$A7,$F2,$3F,$BC,$9F,$F6,$47,$D0,$7F,$96,$7B,$BB
	.dc.b	$61,$87,$F8,$79,$85,$1F,$E5,$FA,$E5,$18,$9F,$E2,$B1,$8F,$33,$A6
	.dc.b	$3F,$C6,$1B,$5F,$ED,$77,$97,$8B,$FE,$68,$84,$63,$1F,$14,$7F,$B7
	.dc.b	$70,$F1,$63,$CA,$3C,$07,$FB,$87,$1F,$1C,$71,$8C,$9F,$E8,$2C,$7F
	.dc.b	$94,$C7,$13,$46,$F0,$E8,$4B,$F5,$E3,$E5,$8D,$C7,$C4,$AE,$6D,$BF
	.dc.b	$38,$C3,$C7,$1C,$B1,$CB,$EA,$38,$4F,$9A,$76,$FA,$F0,$F8,$98,$F9
	.dc.b	$EB,$6F,$8A,$75,$7F,$2E,$79,$E3,$E9,$9C,$9F,$32,$3C,$4F,$C4,$7C
	.dc.b	$4F,$CD,$C7,$D6,$F0,$BF,$89,$3D,$09,$1D,$2B,$2F,$13,$C4,$53,$C0
	.dc.b	$6E,$3C,$3F,$23,$FC,$37,$2F,$C4,$B7,$2F,$C1,$37,$25,$A3,$F8,$A3
	.dc.b	$8F,$07,$31,$88,$46,$31,$FF,$21,$B6,$3F,$92,$35,$FF,$73,$1E,$63
	.dc.b	$FD,$73,$E4,$FF,$9E,$E7,$FD,$79,$1D,$CD,$5A,$75,$97,$69,$CC,$61
	.dc.b	$EE,$63,$8B,$E2,$BE,$0D,$3F,$D9,$17,$CE,$7F,$13,$3D,$E7,$35,$FE
	.dc.b	$8D,$7C,$71,$EE,$3A,$1E,$93,$CD,$87,$0A,$EF,$1C,$F0,$FF,$71,$5F
	.dc.b	$37,$D1,$7B,$9F,$D3,$9B,$3F,$85,$4E,$DF,$5C,$6C,$F1,$E3,$FA,$0C
	.dc.b	$7D,$53,$C8,$F3,$4E,$EF,$E3,$63,$DC,$E3,$3E,$A6,$CE,$E3,$BC,$FA
	.dc.b	$31,$F8,$B1,$D2,$30,$F7,$C3,$C5,$8F,$9E,$3E,$EE,$1D,$EB,$C2,$70
	.dc.b	$7B,$BC,$B0,$B3,$F0,$30,$FE,$23,$BF,$F3,$8F,$95,$FF,$5F,$67,$FD
	.dc.b	$05,$B3,$F9,$69,$FE,$7B,$F2,$B8,$BF,$E8,$1C,$4B,$C5,$73,$BF,$E7
	.dc.b	$33,$96,$CF,$F9,$12,$BA,$4E,$5F,$EF,$95,$FD,$2B,$2E,$72,$6F,$C9
	.dc.b	$E1,$98,$7F,$43,$39,$FC,$9E,$30,$FE,$B8,$F2,$24,$6E,$3D,$33,$87
	.dc.b	$71,$E8,$9F,$34,$7B,$9B,$87,$F5,$E7,$30,$E5,$7C,$8F,$C5,$8C,$3F
	.dc.b	$1F,$57,$89,$BC,$7F,$D5,$B7,$C1,$B5,$D2,$31,$7F,$EC,$06,$BF,$E2
	.dc.b	$16,$BF,$57,$F2,$BF,$8D,$5E,$DC,$91,$89,$FF,$55,$2D,$3E,$BB,$9C
	.dc.b	$BF,$D2,$0D,$BF,$AB,$F2,$7A,$D7,$E9,$BE,$DF,$A0,$7C,$65,$3D,$DF
	.dc.b	$3C,$F7,$F1,$87,$FB,$2A,$B6,$FF,$96,$1B,$A7,$FB,$71,$AF,$F8,$F5
	.dc.b	$AF,$E1,$BE,$53,$F0,$67,$E1,$38,$2D,$7B,$F8,$D3,$F8,$87,$35,$FE
	.dc.b	$1D,$7F,$17,$F2,$92,$B8,$7C,$A5,$AF,$F1,$14,$FE,$43,$E5,$31,$C3
	.dc.b	$FE,$23,$7E,$E7,$D7,$FF,$AE,$70,$FF,$72,$B6,$FF,$DB,$DC,$9F,$ED
	.dc.b	$D9,$CF,$BD,$C0,$FC,$3D,$6E,$BF,$CB,$4D,$7F,$D4,$2D,$7F,$A2,$BD
	.dc.b	$70,$57,$F4,$8F,$83,$7D,$7B,$E4,$BB,$4F,$FB,$EE,$30,$FC,$5F,$96
	.dc.b	$FD,$59,$FA,$CF,$96,$B1,$93,$E5,$7C,$B6,$33,$FF,$22,$3F,$B2,$E7
	.dc.b	$93,$F3,$FE,$5A,$D7,$F9,$26,$6E,$9F,$E4,$99,$B5,$EC,$F8,$3F,$92
	.dc.b	$7F,$09,$D7,$FC,$91,$F1,$F8,$9F,$97,$3F,$18,$8F,$FA,$7B,$74,$63
	.dc.b	$2F,$F6,$94,$3E,$13,$94,$61,$FE,$13,$C1,$2F,$F4,$9F,$AE,$FC,$A7
	.dc.b	$F1,$11,$9A,$FF,$31,$3F,$C6,$A7,$51,$89,$E3,$F3,$B9,$35,$F3,$16
	.dc.b	$32,$76,$F9,$8C,$53,$7F,$F7,$E6,$09,$ED,$F3,$12,$FC,$BF,$31,$70
	.dc.b	$E1,$1F,$CC,$6D,$59,$C3,$9D,$DA,$D4,$DB,$FF,$3A,$5A,$B1,$D0,$87
	.dc.b	$CC,$7B,$CF,$81,$F2,$37,$1F,$8F,$17,$0F,$19,$D2,$71,$97,$B3,$E6
	.dc.b	$3E,$41,$F2,$7E,$62,$E1,$F4,$22,$DF,$2C,$BC,$55,$8B,$FC,$EC,$FE
	.dc.b	$83,$E1,$BF,$5E,$7E,$1F,$D4,$FF,$13,$63,$FB,$08,$CB,$F8,$57,$F5
	.dc.b	$A4,$70,$7F,$3F,$E7,$7D,$1B,$A3,$8E,$1E,$C9,$CC,$7E,$07,$E0,$25
	.dc.b	$FE,$23,$E0,$2D,$FE,$43,$E1,$BF,$76,$49,$FC,$B3,$FD,$1B,$8B,$B4
	.dc.b	$7E,$43,$E6,$7F,$19,$8F,$E3,$5F,$C7,$7C,$C4,$9F,$CA,$FC,$CD,$C7
	.dc.b	$E3,$60,$F9,$9E,$84,$AF,$F5,$2C,$65,$BE,$77,$E6,$7E,$13,$F3,$4D
	.dc.b	$8F,$D0,$FC,$C7,$E7,$4F,$CF,$7C,$C5,$8F,$E9,$1F,$AF,$7C,$EC,$BF
	.dc.b	$B9,$B5,$FF,$2B,$9F,$EB,$F7,$0F,$F8,$15,$AE,$7D,$EF,$F0,$1C,$7F
	.dc.b	$A8,$27,$AF,$99,$89,$CF,$CB,$C9,$F0,$FC,$C5,$C3,$F0,$D8,$7F,$00
	.dc.b	$9F,$F6,$1B,$63,$CB,$E4,$7F,$74,$8C,$47,$9B,$B5,$1F,$E3,$8E,$CF
	.dc.b	$C6,$C6,$3E,$07,$CB,$E4,$79,$9F,$3F,$C1,$CC,$BE,$1F,$CB,$5C,$BC
	.dc.b	$7F,$2B,$DC,$93,$E4,$7C,$5F,$2D,$F0,$0F,$83,$F2,$D6,$7E,$6F,$85
	.dc.b	$F1,$3E,$57,$E2,$B6,$3E,$8C,$7A,$C8,$FC,$07,$DD,$F2,$BF,$30,$E1
	.dc.b	$8F,$C6,$7E,$7F,$81,$2B,$ED,$FC,$25,$AF,$97,$3F,$1C,$93,$F3,$1B
	.dc.b	$1F,$43,$E5,$78,$71,$7C,$8E,$34,$7F,$87,$7E,$71,$E4,$D3,$EF,$FC
	.dc.b	$A5,$CB,$FC,$5D,$8A,$B7,$1F,$8A,$8F,$C5,$C6,$1C,$1B,$AF,$F1,$6C
	.dc.b	$66,$3E,$E6,$0F,$90,$FC,$69,$FE,$50,$FD,$7F,$FD,$57,$87,$E6,$A3
	.dc.b	$2F,$F8,$8E,$3F,$C6,$15,$F8,$C9,$E1,$F1,$3F,$E3,$7F,$9E,$FE,$68
	.dc.b	$FF,$AC,$7B,$7F,$3F,$18,$AF,$F5,$8C,$7E,$8B,$E4,$BF,$C6,$BF,$20
	.dc.b	$8D,$FA,$DF,$E4,$3A,$D9,$FE,$BE,$BE,$55,$B8,$C4,$EE,$31,$0B,$EB
	.dc.b	$7D,$30,$F2,$4F,$06,$AC,$3A,$CF,$97,$E1,$BF,$95,$E1,$D2,$73,$5C
	.dc.b	$4F,$29,$C7,$1F,$15,$F1,$D3,$AD,$70,$AD,$63,$8B,$DC,$7F,$B6,$B2
	.dc.b	$F3,$BD,$71,$F0,$4E,$A3,$89,$E0,$9D,$B8,$9D,$D7,$53,$DB,$3A,$BF
	.dc.b	$11,$F4,$EB,$F1,$4D,$7E,$31,$F8,$A7,$0F,$FB,$BB,$E4,$5D,$8F,$47
	.dc.b	$F1,$F2,$9C,$C6,$1B,$1F,$94,$6B,$E0,$F6,$3F,$DF,$0F,$76,$B9,$F5
	.dc.b	$3A,$E1,$FD,$E1,$B3,$C3,$3F,$A8,$DD,$BF,$1E,$DB,$FC,$47,$97,$FC
	.dc.b	$C2,$B7,$7F,$52,$B5,$5F,$81,$6E,$0B,$E1,$87,$4B,$E0,$4B,$FC,$0D
	.dc.b	$6D,$C6,$1D,$9C,$E1,$D3,$0E,$71,$F9,$B8,$EE,$4B,$FC,$3C,$EA,$7C
	.dc.b	$91,$D3,$0F,$14,$78,$27,$58,$7C,$58,$EF,$5B,$7E,$09,$FA,$0F,$11
	.dc.b	$F3,$DF,$EC,$38,$FF,$4F,$3F,$E9,$B6,$FC,$77,$F5,$6E,$BF,$E7,$F3
	.dc.b	$B9,$FF,$08,$36,$33,$3B,$E3,$FF,$6F,$4F,$3A,$DF,$C6,$E5,$1F,$C6
	.dc.b	$FD,$1C,$62,$E8,$7C,$9E,$1B,$5F,$C8,$BA,$D7,$CE,$39,$1F,$92,$8C
	.dc.b	$B6,$3F,$2A,$E8,$FF,$72,$E3,$FE,$9E,$7E,$0F,$4F,$8A,$3F,$DA,$31
	.dc.b	$F1,$3E,$2F,$06,$66,$46,$48,$7C,$48,$DB,$F3,$3E,$2F,$A7,$C5,$7A
	.dc.b	$FA,$A4,$FD,$08,$F0,$7C,$5E,$A3,$07,$D4,$72,$D2,$1F,$4B,$E2,$FD
	.dc.b	$1F,$15,$F1,$FE,$46,$4F,$DC,$8F,$67,$C5,$7B,$7E,$2E,$4F,$E0,$23
	.dc.b	$E0,$7C,$5E,$A3,$07,$E1,$1F,$89,$F1,$3F,$BE,$E5,$FC,$4B,$78,$7F
	.dc.b	$D5,$15,$FD,$42,$72,$DF,$9D,$7C,$57,$DB,$F9,$39,$3F,$EE,$D9,$CB
	.dc.b	$7E,$2F,$B4,$1A,$48,$E9,$E1,$7E,$22,$7F,$CF,$71,$CE,$31,$F1,$79
	.dc.b	$F8,$BF,$C5,$76,$CA,$F0,$47,$1F,$8B,$91,$E3,$7A,$7C,$5F,$5F,$3F
	.dc.b	$B7,$C5,$4A,$F4,$47,$87,$E2,$BF,$2D,$E2,$49,$F6,$47,$97,$E2,$BF
	.dc.b	$35,$F2,$72,$7E,$04,$7A,$7E,$2B,$D5,$E1,$C9,$F8,$91,$F9,$DF,$9E
	.dc.b	$7B,$97,$DE,$95,$FE,$04,$9C,$BF,$2F,$E2,$FF,$51,$F2,$72,$30,$6F
	.dc.b	$DC,$FD,$97,$AA,$F9,$CF,$F6,$04,$FF,$9F,$5A,$FF,$A5,$86,$0A,$F2
	.dc.b	$1C,$A7,$7E,$87,$FA,$42,$FA,$7C,$2E,$B0,$F0,$1F,$D1,$B0,$E4,$DC
	.dc.b	$3F,$6B,$18,$8F,$F0,$E7,$7B,$B3,$F9,$64,$7F,$A7,$DB,$8F,$F9,$49
	.dc.b	$BE,$8F,$F0,$CA,$E1,$F1,$7F,$85,$F5,$66,$1F,$8F,$7F,$11,$1F,$89
	.dc.b	$F8,$B9,$3F,$93,$7F,$89,$FC,$BF,$F5,$2C,$BB,$46,$A7,$C4,$EB,$D6
	.dc.b	$DF,$C5,$C9,$FE,$3B,$5F,$E6,$6B,$4F,$4F,$8C,$D5,$7F,$AE,$67,$2F
	.dc.b	$0D,$3D,$51,$88,$D4,$F0,$BE,$97,$6F,$73,$E8,$F6,$BD,$3F,$19,$30
	.dc.b	$E8,$36,$65,$FE,$26,$CB,$CD,$7F,$BD,$8F,$F1,$4D,$7F,$A0,$5A,$F9
	.dc.b	$23,$5F,$17,$F1,$89,$1D,$AF,$AF,$1E,$23,$DD,$37,$0F,$89,$3F,$CE
	.dc.b	$7E,$7B,$FC,$6D,$5F,$E5,$E7,$FC,$80,$FE,$45,$B3,$FE,$DC,$84,$70
	.dc.b	$75,$EB,$73,$F8,$19,$1C,$E3,$B1,$FD,$0A,$78,$47,$F9,$A5,$AF,$FA
	.dc.b	$D1,$AF,$C7,$CD,$F1,$9F,$84,$24,$7B,$E3,$FC,$46,$D8,$CB,$F7,$5F
	.dc.b	$C5,$FC,$17,$E3,$3E,$33,$F1,$BD,$D2,$B9,$46,$5F,$E5,$56,$8F,$A8
	.dc.b	$48,$FF,$17,$E1,$FD,$C7,$8B,$FB,$5F,$5F,$EE,$FE,$1F,$F6,$87,$CD
	.dc.b	$F3,$25,$FE,$F6,$32,$FC,$F8,$CB,$FE,$88,$BF,$EF,$CF,$C5,$8F,$90
	.dc.b	$7E,$17,$C1,$FD,$2F,$C5,$C9,$F8,$51,$FA,$83,$4F,$A9,$AF,$EF,$33
	.dc.b	$FD,$F7,$0F,$E9,$7E,$2B,$F5,$E7,$EC,$3C,$09,$5F,$09,$FC,$A1,$1F
	.dc.b	$F2,$7F,$CE,$BC,$CF,$F0,$17,$CE,$FE,$6F,$83,$FD,$DA,$DA,$FF,$07
	.dc.b	$D6,$DF,$F5,$5B,$5F,$F5,$3B,$5F,$F5,$63,$5F,$F7,$BB,$5F,$D5,$7C
	.dc.b	$EF,$1C,$BF,$7D,$C1,$FC,$20,$FF,$0D,$C6,$70,$FE,$28,$DA,$FC,$64
	.dc.b	$ED,$FD,$44,$65,$7C,$EF,$F1,$C7,$FC,$4F,$F3,$AF,$4B,$FC,$40,$D8
	.dc.b	$FF,$62,$CE,$DE,$3E,$2F,$F8,$82,$3F,$94,$C7,$FB,$1D,$AF,$F8,$C8
	.dc.b	$D9,$F9,$0F,$8D,$FF,$18,$7C,$2C,$8F,$CF,$47,$F2,$43,$F8,$31,$1E
	.dc.b	$07,$F8,$07,$E1,$7F,$DA,$9A,$FF,$21,$9E,$13,$FC,$16,$73,$AF,$F3
	.dc.b	$76,$C7,$EB,$E7,$6A,$95,$B8,$FF,$23,$D7,$F9,$11,$D4,$7F,$41,$F8
	.dc.b	$8F,$F1,$CC,$60,$FE,$40,$E2,$B9,$C7,$F2,$3A,$E5,$18,$F0,$91,$D6
	.dc.b	$7E,$34,$63,$E2,$7F,$C4,$53,$99,$F5,$47,$F9,$3B,$E3,$3F,$83,$DF
	.dc.b	$F9,$7B,$E7,$7F,$CD,$1D,$7E,$76,$D7,$D1,$7F,$CD,$95,$C2,$D1,$8F
	.dc.b	$07,$E1,$46,$2F,$EC,$1C,$CB,$7C,$DF,$B9,$7C,$EF,$96,$1F,$E7,$0F
	.dc.b	$27,$C3,$85,$BE,$2D,$3F,$D7,$D8,$F8,$27,$8C,$78,$4F,$F3,$CC,$F8
	.dc.b	$5B,$8F,$85,$E3,$7F,$06,$7C,$77,$EA,$C3,$C3,$5C,$C4,$F3,$7F,$D9
	.dc.b	$53,$B8,$E3,$5C,$E7,$B9,$4E,$37,$E1,$78,$C6,$71,$F8,$55,$EE,$9F
	.dc.b	$03,$DE,$33,$1E,$83,$35,$CE,$75,$7E,$39,$D6,$1E,$A2,$4F,$9A,$3D
	.dc.b	$97,$D1,$F2,$9E,$78,$E7,$87,$96,$FA,$4E,$8F,$09,$98,$8E,$09,$7F
	.dc.b	$50,$FA,$BE,$57,$F8,$1F,$0F,$C5,$36,$FF,$12,$DB,$FF,$58,$BF,$E0
	.dc.b	$DB,$FF,$59,$FC,$6E,$33,$FE,$09,$18,$13,$B5,$3B,$76,$D7,$FD,$DC
	.dc.b	$7D,$DA,$79,$1F,$B0,$CB,$FA,$6C,$F2,$35,$19,$CF,$C3,$19,$9F,$F6
	.dc.b	$74,$E6,$3B,$9C,$38,$3F,$04,$F9,$E7,$71,$88,$E5,$5F,$ED,$BA,$E5
	.dc.b	$1E,$6A,$DB,$F9,$33,$94,$EE,$B9,$61,$F8,$FC,$3D,$71,$8A,$F1,$1C
	.dc.b	$4E,$E7,$F9,$0B,$0E,$EF,$43,$11,$C7,$0F,$09,$F1,$2F,$B5,$70,$AE
	.dc.b	$67,$F2,$37,$11,$AA,$E5,$87,$9E,$BC,$53,$96,$9F,$E5,$28,$D6,$5D
	.dc.b	$E3,$07,$90,$A7,$C5,$8E,$98,$6C,$F3,$BF,$5A,$BB,$46,$63,$67,$F3
	.dc.b	$DF,$87,$FF,$1C,$55,$C3,$FA,$34,$62,$3F,$99,$37,$83,$F9,$8B,$FC
	.dc.b	$DB,$3F,$E6,$9C,$3F,$CD,$2F,$FA,$54,$E5,$FE,$81,$C3,$70,$FE,$79
	.dc.b	$8D,$C7,$F9,$DF,$AF,$6B,$F8,$7D,$FF,$53,$8C,$56,$5B,$19,$25,$FF
	.dc.b	$A5,$29,$3E,$76,$CF,$F6,$18,$C1,$E1,$C5,$F1,$5D,$67,$FC,$CD,$18
	.dc.b	$7F,$CE,$B6,$F2,$9F,$17,$F1,$63,$0D,$8C,$CF,$E9,$A7,$2D,$8C,$C7
	.dc.b	$F6,$1D,$5E,$13,$9B,$E3,$4F,$03,$51,$9B,$FF,$3D,$7C,$EF,$C6,$9F
	.dc.b	$ED,$5B,$FF,$64,$F5,$BE,$64,$E7,$1F,$ED,$1B,$A7,$3C,$1F,$3D,$B8
	.dc.b	$FF,$80,$5F,$0F,$CE,$3E,$23,$DC,$D9,$F6,$37,$3F,$E0,$6D,$7F,$8A
	.dc.b	$B9,$C7,$F0,$DF,$32,$F2,$BA,$E1,$F8,$39,$4E,$70,$FF,$09,$77,$7D
	.dc.b	$92,$3F,$E1,$96,$E3,$F0,$34,$FC,$83,$63,$F1,$51,$F8,$BE,$4F,$C1
	.dc.b	$12,$BF,$C4,$4D,$AE,$5F,$1B,$B9,$FF,$11,$3F,$4B,$E6,$9E,$B7,$DF
	.dc.b	$6C,$FD,$D6,$D7,$F2,$46,$BF,$CD,$9C,$E7,$FE,$F1,$F9,$C7,$C6,$75
	.dc.b	$C3,$F4,$9E,$8A,$73,$8F,$E4,$30,$E2,$FF,$A2,$A3,$38,$FF,$C0,$19
	.dc.b	$CA,$AE,$58,$7E,$AF,$E3,$FF,$44,$4C,$3F,$C8,$4D,$AF,$95,$F1,$DB
	.dc.b	$9F,$F2,$13,$FB,$3E,$19,$1F,$E4,$5F,$9C,$B5,$F4,$66,$EB,$F8,$E9
	.dc.b	$B5,$FE,$59,$CD,$F3,$3F,$24,$91,$F2,$5B,$87,$AF,$DA,$FA,$11,$F4
	.dc.b	$5B,$3F,$0D,$B3,$B8,$FF,$2C,$69,$FC,$24,$FE,$19,$F3,$7F,$E5,$48
	.dc.b	$C9,$27,$F1,$13,$A8,$E0,$7F,$B9,$65,$F2,$1F,$C7,$DB,$1F,$D8,$3B
	.dc.b	$11,$8B,$FE,$E7,$18,$AE,$07,$8B,$E5,$BC,$6D,$BE,$DF,$1F,$E6,$3C
	.dc.b	$FF,$1F,$63,$24,$C3,$F7,$B8,$A7,$51,$8A,$BB,$FF,$99,$6B,$57,$C6
	.dc.b	$31,$F1,$FF,$10,$F8,$BE,$9D,$9F,$81,$17,$0E,$F3,$6B,$A7,$C7,$FC
	.dc.b	$C3,$E6,$FC,$7D,$C3,$E4,$45,$BF,$3C,$DA,$FF,$3D,$C3,$89,$EE,$FE
	.dc.b	$AC,$98,$7E,$AD,$B3,$FE,$85,$E8,$FE,$0B,$F1,$BF,$B3,$9F,$D3,$F8
	.dc.b	$08,$C5,$AF,$84,$E6,$3F,$55,$5C,$2B,$F0,$BE,$E7,$F4,$1F,$94,$B5
	.dc.b	$FD,$01,$FE,$01,$EF,$7F,$01,$F9,$8B,$3F,$B3,$7F,$1B,$7F,$8E,$F9
	.dc.b	$CB,$3F,$89,$7F,$23,$F1,$FF,$D3,$7C,$3B,$3F,$D3,$1F,$CA,$FC,$7F
	.dc.b	$F5,$1F,$0A,$CF,$F5,$07,$F3,$3F,$39,$F9,$AF,$7A,$CF,$E5,$5F,$CE
	.dc.b	$FC,$6C,$8F,$CC,$7C,$6D,$E3,$FE,$87,$F3,$5F,$E4,$49,$FE,$B7,$F3
	.dc.b	$1F,$DB,$E3,$13,$FD,$23,$D7,$4E,$7D,$C9,$7F,$2F,$AB,$FD,$61,$87
	.dc.b	$F0,$EF,$10,$AB,$A7,$C7,$7F,$81,$60,$D2,$5F,$EA,$63,$F8,$1B,$FD
	.dc.b	$BB,$CB,$F2,$79,$3F,$DC,$DF,$37,$BD,$CC,$93,$C5,$ED,$F0,$1E,$BF
	.dc.b	$8B,$F6,$36,$7D,$07,$E0,$F3,$F3,$CF,$C1,$F9,$CD,$70,$7C,$67,$E2
	.dc.b	$7C,$6F,$F8,$0F,$D0,$B8,$7F,$80,$9F,$91,$EB,$48,$EB,$6C,$7E,$1F
	.dc.b	$85,$98,$FA,$35,$F3,$FE,$73,$1C,$7F,$29,$C6,$1F,$E0,$EF,$4A,$DF
	.dc.b	$D3,$7E,$B7,$85,$27,$BE,$5F,$43,$DC,$C5,$FD,$EA,$FB,$9E,$1E,$32
	.dc.b	$FA,$ED,$C3,$E1,$72,$A3,$17,$E9,$8C,$4F,$F8,$7B,$E6,$DE,$57,$F8
	.dc.b	$75,$B1,$FE,$1B,$6D,$7F,$A0,$9A,$FF,$A5,$1C,$E1,$FE,$44,$F4,$FF
	.dc.b	$C3,$71,$8A,$FA,$9D,$9F,$D6,$AB,$F1,$BF,$0F,$FE,$33,$F8,$0F,$CB
	.dc.b	$36,$BE,$0E,$68,$C6,$5F,$9E,$6D,$FF,$0E,$F9,$C7,$AD,$FE,$37,$AF
	.dc.b	$F5,$1F,$77,$F0,$55,$3A,$CB,$FD,$5B,$39,$F9,$2F,$F1,$AF,$C3,$FE
	.dc.b	$6A,$FF,$86,$5E,$DE,$66,$30,$ED,$3B,$8E,$B1,$8B,$E9,$F0,$8D,$7B
	.dc.b	$D7,$1C,$AB,$C1,$B8,$F7,$C7,$A6,$5E,$4C,$7D,$53,$96,$E5,$DF,$0E
	.dc.b	$87,$98,$57,$17,$37,$E7,$BE,$51,$8B,$F3,$E3,$E2,$C3,$A4,$F9,$4E
	.dc.b	$95,$C1,$EB,$97,$B5,$CC,$F0,$9D,$1B,$C7,$C5,$3C,$F1,$F1,$57,$D7
	.dc.b	$8F,$F6,$ED,$7C,$F6,$C6,$5F,$7E,$5F,$E6,$DF,$89,$7C,$97,$F8,$86
	.dc.b	$7F,$DA,$B1,$8F,$4F,$53,$99,$FF,$74,$CE,$5F,$F6,$DE,$0F,$67,$FD
	.dc.b	$D7,$18,$6B,$E4,$6B,$F4,$9A,$FE,$19,$E3,$97,$CD,$6E,$DB,$F8,$AF
	.dc.b	$BB,$E5,$FE,$4E,$B5,$1F,$80,$AD,$DF,$38,$DB,$96,$3F,$4A,$36,$FE
	.dc.b	$72,$BA,$5F,$78,$D5,$70,$BF,$C0,$56,$A7,$B1,$A2,$46,$9E,$15,$CE
	.dc.b	$3C,$F1,$F8,$29,$E4,$FE,$B2,$FC,$78,$7C,$BC,$38,$CF,$EA,$AB,$C1
	.dc.b	$D6,$F0,$BF,$C0,$BC,$7F,$89,$8C,$7B,$9F,$B7,$8C,$4F,$F8,$D2,$FF
	.dc.b	$91,$BC,$39,$3F,$9F,$36,$3F,$9A,$36,$FB,$4F,$33,$7F,$35,$6B,$7C
	.dc.b	$76,$BC,$8D,$BF,$E1,$73,$E0,$D7,$13,$23,$C6,$EE,$38,$7B,$BC,$7D
	.dc.b	$C9,$5E,$88,$CB,$48,$7F,$AE,$AD,$EE,$6A,$31,$37,$4F,$87,$3E,$5F
	.dc.b	$72,$57,$C4,$7D,$1D,$F2,$3E,$3B,$5E,$DE,$D7,$B4,$8F,$CC,$7E,$07
	.dc.b	$CC,$FC,$1F,$6A,$4F,$D0,$7E,$27,$B5,$F1,$7D,$A9,$3F,$51,$F9,$1F
	.dc.b	$33,$F2,$7D,$A9,$3F,$61,$F9,$9E,$CF,$CD,$E8,$93,$F7,$1F,$A1,$EC
	.dc.b	$FD,$1F,$66,$4F,$E0,$1F,$AD,$FB,$07,$ED,$38,$BF,$10,$FF,$84,$E7
	.dc.b	$FB,$3B,$FE,$B8,$6F,$C1,$F0,$F6,$B8,$F8,$3C,$BE,$5E,$57,$08,$EB
	.dc.b	$ED,$C8,$E6,$F8,$3D,$AF,$0F,$C1,$F8,$BD,$B9,$5E,$08,$F2,$F9,$12
	.dc.b	$3C,$6F,$A3,$DB,$F4,$FB,$72,$7C,$EF,$B3,$BF,$DB,$ED,$49,$F5,$BF
	.dc.b	$03,$DA,$F8,$3E,$D4,$9F,$7B,$F1,$3D,$8F,$8B,$EC,$C9,$F8,$6F,$C8
	.dc.b	$F6,$7F,$41,$0F,$99,$FF,$39,$BF,$E1,$C3,$18,$EA,$33,$1F,$E7,$3F
	.dc.b	$9F,$FD,$67,$3B,$E5,$7E,$9F,$B1,$2B,$E8,$C7,$D6,$23,$83,$FC,$37
	.dc.b	$FB,$0F,$F7,$35,$D7,$F6,$A7,$71,$8F,$43,$FC,$75,$87,$78,$E2,$7E
	.dc.b	$DE,$FF,$A4,$5F,$F0,$95,$F2,$19,$35,$EC,$4A,$F0,$C6,$5F,$1C,$73
	.dc.b	$F6,$7A,$7C,$94,$9F,$3C,$65,$F4,$C6,$5F,$31,$FE,$43,$BF,$ED,$EF
	.dc.b	$BB,$E7,$BC,$DF,$23,$23,$04,$38,$FA,$AA,$4E,$E7,$30,$AF,$80,$67
	.dc.b	$3F,$81,$58,$AF,$9B,$F3,$DF,$14,$8F,$D0,$F9,$4A,$F4,$8B,$5F,$26
	.dc.b	$7E,$67,$C9,$FC,$DF,$92,$97,$F6,$23,$2F,$DA,$8C,$BE,$68,$FC,$47
	.dc.b	$17,$F2,$DF,$80,$7C,$F7,$E2,$B1,$FF,$03,$46,$3A,$37,$87,$E4,$7A
	.dc.b	$3E,$F3,$F3,$0F,$CB,$61,$A7,$1F,$3B,$F8,$9C,$3F,$D9,$B5,$C1,$B5
	.dc.b	$B7,$FD,$A3,$3B,$E7,$FF,$42,$70,$BE,$0F,$F1,$5F,$29,$32,$FC,$AC
	.dc.b	$7E,$34,$C3,$F4,$A3,$0F,$E6,$BE,$53,$F2,$44,$7F,$3B,$F2,$15,$FB
	.dc.b	$31,$72,$FC,$B4,$FE,$CB,$F4,$07,$27,$E6,$FE,$46,$61,$FB,$39,$CB
	.dc.b	$67,$81,$1F,$F4,$9D,$BE,$47,$F9,$04,$7C,$9A,$E2,$DA,$E0,$FE,$62
	.dc.b	$B8,$7A,$3F,$AB,$B9,$3F,$B6,$7F,$5D,$F2,$38,$8F,$C9,$C5,$AF,$D8
	.dc.b	$4F,$EC,$FE,$46,$57,$EF,$9F,$DB,$7C,$8E,$23,$F3,$71,$6B,$F7,$13
	.dc.b	$FD,$CD,$7D,$87,$FA,$5E,$BD,$AD,$7D,$0D,$8F,$E0,$1F,$25,$C6,$31
	.dc.b	$C7,$FE,$D9,$6D,$7F,$0E,$6F,$7F,$EB,$EB,$F9,$1F,$CE,$BE,$0B,$F8
	.dc.b	$A5,$FF,$A1,$E3,$0E,$03,$87,$6D,$7F,$8E,$39,$C7,$FC,$C9,$1C,$27
	.dc.b	$B9,$FC,$CB,$E7,$BF,$AC,$47,$E7,$FE,$71,$F1,$3F,$CD,$9B,$1F,$CD
	.dc.b	$5A,$FE,$97,$E7,$5F,$07,$BA,$93,$B7,$F9,$FC,$EA,$3F,$98,$FC,$B7
	.dc.b	$F2,$92,$3F,$EC,$EF,$D8,$BF,$C5,$1D,$A9,$CC,$FF,$12,$7F,$C4,$D3
	.dc.b	$FE,$2D,$3F,$C6,$91,$FC,$69,$C7,$0E,$BE,$42,$4F,$56,$BB,$F9,$0E
	.dc.b	$7F,$21,$23,$C4,$F6,$F9,$0C,$47,$58,$BC,$3F,$E4,$7C,$7F,$CF,$13
	.dc.b	$FD,$67,$E5,$1E,$FF,$F4,$C8,$7C,$F7,$F2,$C4,$7F,$2E,$F4,$FF,$D8
	.dc.b	$91,$98,$C1,$A7,$15,$D3,$0C,$79,$FF,$EC,$B8,$C5,$76,$8F,$F6,$71
	.dc.b	$FD,$49,$D9,$E0,$70,$EC,$F1,$C7,$2A,$EF,$87,$A5,$F2,$57,$06,$E1
	.dc.b	$E8,$8F,$14,$ED,$B8,$78,$30,$F5,$9D,$CB,$7D,$A3,$58,$78,$6B,$D3
	.dc.b	$87,$5B,$F4,$35,$4E,$DF,$4E,$1E,$17,$67,$86,$BD,$07,$91,$F1,$4F
	.dc.b	$17,$C7,$1E,$07,$C7,$3C,$DE,$ED,$8E,$71,$D8,$E7,$1B,$9E,$93,$A9
	.dc.b	$EA,$79,$AF,$A9,$E5,$BF,$64,$EE,$BC,$87,$28,$C9,$F0,$5C,$DF,$8E
	.dc.b	$33,$7D,$0F,$1D,$FE,$3B,$E4,$FF,$7B,$7F,$86,$6D,$FE,$55,$B3,$BF
	.dc.b	$86,$FE,$23,$F2,$B7,$0F,$F1,$33,$5F,$F6,$40,$AF,$E1,$2F,$E7,$23
	.dc.b	$4D,$3F,$01,$4F,$84,$FF,$7B,$69,$F9,$F9,$D3,$FA,$19,$D7,$17,$E2
	.dc.b	$67,$3A,$FF,$AC,$EF,$8B,$E6,$8E,$B1,$87,$CE,$62,$79,$2C,$3F,$29
	.dc.b	$7D,$A3,$F2,$87,$63,$F6,$AE,$70,$F3,$47,$13,$A1,$C0,$C3,$E9,$BF
	.dc.b	$14,$7C,$68,$C5,$F8,$AF,$C3,$5A,$AF,$01,$FE,$52,$70,$F4,$C3,$E0
	.dc.b	$57,$18,$F0,$DF,$92,$F9,$CE,$63,$BC,$F9,$67,$2D,$38,$D7,$C9,$9D
	.dc.b	$9C,$E3,$27,$CA,$73,$1B,$3C,$71,$CE,$32,$EE,$33,$1F,$C7,$FB,$5B
	.dc.b	$BE,$51,$8F,$F3,$F6,$E1,$FC,$E7,$A2,$E9,$FC,$B6,$6D,$7F,$28,$C3
	.dc.b	$9C,$62,$D1,$89,$FE,$95,$79,$8C,$4F,$F4,$5F,$94,$C6,$3F,$ED,$EB
	.dc.b	$FE,$A1,$F2,$98,$99,$1F,$D4,$FE,$52,$DF,$F4,$D7,$8F,$7C,$9D,$1F
	.dc.b	$D6,$5C,$4F,$F5,$37,$BC,$E3,$4F,$EE,$7F,$97,$3A,$BD,$56,$8C,$61
	.dc.b	$FE,$2C,$F8,$17,$9D,$A9,$CD,$7F,$95,$5B,$87,$4C,$3E,$22,$BE,$38
	.dc.b	$7F,$AB,$6B,$6F,$F5,$2B,$7C,$F3,$C7,$F1,$B6,$7F,$C7,$BF,$48,$F9
	.dc.b	$2C,$D4,$9F,$F0,$A5,$6E,$FF,$C3,$2D,$7F,$D5,$CD,$7E,$D7,$63,$E1
	.dc.b	$7E,$B3,$6B,$FC,$05,$F2,$F8,$99,$1F,$B4,$38,$BF,$6A,$33,$5F,$E0
	.dc.b	$1F,$97,$FB,$A7,$DE,$F9,$7B,$7F,$E0,$D8,$B3,$FE,$07,$9B,$BF,$F7
	.dc.b	$BF,$23,$F0,$A7,$E1,$BE,$5E,$E1,$FE,$15,$8B,$7F,$E0,$F9,$B5,$FE
	.dc.b	$2E,$FC,$C1,$F2,$7F,$71,$B8,$7F,$99,$BD,$37,$89,$9A,$93,$FE,$E1
	.dc.b	$AD,$E3,$FE,$65,$BD,$CE,$6B,$73,$FE,$1F,$F9,$8B,$8F,$F8,$4D,$FC
	.dc.b	$EF,$1C,$BF,$F1,$67,$CC,$5B,$FF,$13,$4D,$BF,$F0,$D4,$DA,$FD,$77
	.dc.b	$73,$E6,$3F,$4F,$F3,$1F,$A8,$6C,$FF,$8D,$DB,$8F,$F8,$C5,$B8,$7F
	.dc.b	$94,$3E,$79,$F2,$BF,$9E,$6D,$FF,$BF,$DF,$F1,$3B,$63,$44,$75,$E5
	.dc.b	$DC,$3F,$D6,$95,$C9,$FF,$70,$FC,$2B,$E7,$BF,$87,$5F,$33,$F1,$12
	.dc.b	$9D,$79,$1F,$C1,$7A,$3F,$AD,$56,$BC,$74,$7D,$0B,$FF,$72,$BA,$6F
	.dc.b	$37,$1F,$37,$93,$6F,$95,$71,$DB,$F8,$EE,$1D,$3D,$CF,$01,$E1,$E6
	.dc.b	$B7,$C2,$2E,$5E,$09,$B7,$CB,$07,$CD,$7A,$7E,$6B,$D4,$DC,$7C,$2D
	.dc.b	$C3,$D2,$DB,$E4,$D9,$FE,$59,$7D,$1F,$4D,$FC,$5F,$43,$E3,$7C,$D5
	.dc.b	$9F,$6B,$71,$F8,$AD,$C7,$CC,$FC,$DE,$3F,$9D,$F3,$56,$FE,$2B,$6F
	.dc.b	$E6,$B6,$FD,$CF,$D5,$E5,$95,$F3,$3E,$6A,$DF,$D4,$9B,$7F,$16,$6D
	.dc.b	$7B,$27,$EF,$7C,$D6,$2F,$FA,$3E,$A8,$FC,$A7,$89,$2F,$F0,$FC,$DF
	.dc.b	$E1,$DD,$B1,$87,$E0,$BE,$33,$F1,$07,$E2,$7E,$6A,$DF,$E0,$3E,$6B
	.dc.b	$F1,$87,$E3,$71,$B1,$F8,$4F,$37,$F2,$07,$E4,$7E,$6A,$C7,$E2,$3E
	.dc.b	$3B,$F2,$87,$E5,$7E,$66,$C7,$E5,$7F,$63,$FF,$44,$B7,$4F,$EB,$D0
	.dc.b	$F9,$9B,$18,$F7,$73,$F3,$36,$F9,$3C,$3E,$66,$47,$5F,$99,$B6,$E8
	.dc.b	$FC,$CE,$4F,$7F,$BF,$C7,$78,$1B,$87,$97,$E6,$7C,$7F,$31,$E4,$6C
	.dc.b	$7A,$7E,$67,$CE,$7A,$3E,$5A,$C7,$B7,$E5,$BD,$67,$B3,$DF,$B1,$F0
	.dc.b	$7E,$5B,$DE,$7C,$0F,$96,$B1,$F1,$7D,$6F,$86,$7C,$4F,$95,$B1,$F2
	.dc.b	$7E,$57,$E3,$9F,$23,$E0,$2C,$7C,$DF,$94,$F9,$67,$CC,$F8,$FB,$1F
	.dc.b	$C1,$7F,$64,$FF,$0C,$B4,$CF,$37,$C7,$7F,$C2,$D1,$A8,$C1,$FE,$16
	.dc.b	$7F,$B8,$B7,$E1,$BF,$C2,$D1,$BE,$7F,$E0,$EE,$BD,$22,$F9,$0D,$9C
	.dc.b	$3E,$3E,$E3,$E5,$8B,$8F,$82,$F8,$D6,$30,$EB,$5E,$0F,$90,$C6,$1E
	.dc.b	$87,$C7,$EC,$48,$F8,$1F,$1D,$72,$E9,$9F,$F2,$E6,$D7,$AA,$75,$8F
	.dc.b	$AB,$E3,$7D,$A7,$BB,$E3,$AE,$3F,$2E,$3E,$0F,$C6,$E3,$1E,$AF,$F8
	.dc.b	$93,$E7,$5B,$7E,$76,$7F,$CF,$5E,$D7,$F8,$66,$73,$E3,$FF,$A4,$63
	.dc.b	$FC,$39,$E3,$3B,$BF,$C9,$BF,$23,$FD,$D3,$0F,$F7,$8E,$1C,$1E,$66
	.dc.b	$31,$ED,$3B,$84,$EE,$31,$7D,$74,$F0,$57,$1C,$AB,$C3,$0E,$79,$79
	.dc.b	$DB,$01,$7C,$DF,$19,$C6,$B9,$E3,$E1,$BE,$51,$DA,$B8,$63,$E7,$AD
	.dc.b	$5F,$8C,$E6,$E1,$EF,$97,$C4,$C3,$9C,$6B,$1F,$2B,$E1,$3C,$78,$76
	.dc.b	$9D,$E1,$DE,$B7,$87,$5A,$D5,$F8,$6B,$58,$FF,$93,$E3,$B9,$6B,$E9
	.dc.b	$36,$FE,$F3,$BF,$8C,$7B,$1F,$DE,$63,$90,$8C,$FA,$98,$8C,$4E,$DE
	.dc.b	$4D,$7F,$11,$CB,$FE,$D9,$79,$E0,$8C,$47,$95,$AF,$0F,$7F,$B4,$ED
	.dc.b	$FA,$B3,$CF,$3F,$A4,$DD,$7F,$2D,$F0,$BF,$E6,$F9,$CD,$70,$26,$1F
	.dc.b	$5A,$75,$5D,$E3,$EC,$DF,$37,$F3,$2D,$7F,$2B,$3C,$1C,$D7,$4B,$EF
	.dc.b	$1D,$4C,$5F,$58,$C9,$F8,$9A,$E5,$3F,$A0,$BE,$D8,$7B,$E3,$C7,$87
	.dc.b	$E5,$2B,$73,$9B,$E7,$87,$53,$E0,$DF,$E9,$BA,$7F,$B8,$70,$FF,$A4
	.dc.b	$79,$7F,$C4,$BE,$BF,$FA,$8E,$BF,$C6,$4D,$7F,$8D,$3F,$EC,$66,$CF
	.dc.b	$13,$FD,$89,$1B,$9E,$7D,$79,$AF,$F6,$2C,$F1,$9E,$FC,$3F,$87,$C3
	.dc.b	$FC,$04,$D9,$FE,$1F,$5F,$8E,$9C,$B6,$32,$F9,$27,$B4,$EA,$3C,$FB
	.dc.b	$62,$BF,$DC,$11,$EC,$E8,$C1,$9E,$99,$1D,$5F,$11,$1C,$1F,$11,$DB
	.dc.b	$F1,$3A,$3B,$74,$C8,$F9,$2F,$87,$A6,$47,$A5,$F7,$11,$F6,$7C,$86
	.dc.b	$9F,$9F,$1E,$98,$F5,$7B,$32,$7E,$AB,$ED,$F8,$CF,$77,$C6,$7C,$92
	.dc.b	$4F,$C7,$7E,$67,$C5,$7C,$3F,$8C,$93,$F7,$A3,$E3,$74,$48,$FC,$0B
	.dc.b	$F2,$7E,$23,$E5,$7C,$4F,$D5,$24,$FD,$37,$ED,$7C,$E7,$FA,$B2,$3F
	.dc.b	$CF,$4B,$E1,$FE,$AB,$63,$3F,$11,$8E,$BC,$FC,$44,$9E,$31,$C3,$E2
	.dc.b	$3A,$92,$3F,$09,$68,$F0,$FC,$47,$6F,$88,$95,$E2,$8F,$0F,$87,$F8
	.dc.b	$6F,$25,$CB,$4C,$D7,$A3,$E2,$3D,$64,$7B,$BE,$EF,$86,$E5,$F1,$D2
	.dc.b	$7D,$F1,$EE,$F8,$6E,$DD,$32,$7E,$1C,$7C,$2F,$85,$F8,$E4,$8F,$5B
	.dc.b	$FC,$A5,$7C,$DA,$F0,$D8,$FE,$5C,$D7,$F9,$9B,$5F,$EB,$83,$C3,$70
	.dc.b	$BE,$17,$E9,$FC,$2C,$AF,$AF,$1F,$58,$FB,$64,$7E,$7B,$5F,$F0,$E4
	.dc.b	$F3,$78,$9F,$EB,$9F,$94,$FE,$9D,$E4,$7F,$AC,$1F,$E9,$6D,$F8,$6F
	.dc.b	$DB,$C1,$BE,$8C,$AE,$51,$A3,$0F,$F8,$AE,$7F,$BB,$BD,$FE,$1B,$99
	.dc.b	$D3,$D3,$91,$92,$1C,$5F,$0F,$C3,$78,$BE,$12,$4F,$A6,$32,$FA,$E3
	.dc.b	$2F,$86,$32,$F8,$E3,$39,$7B,$E7,$DB,$F0,$BF,$8F,$F1,$E5,$FC,$58
	.dc.b	$CB,$48,$7B,$7E,$13,$E2,$FC,$2F,$C6,$F8,$59,$3F,$36,$32,$FC,$F8
	.dc.b	$CB,$F1,$63,$F9,$83,$F8,$AE,$1F,$E6,$5E,$C3,$E7,$3F,$11,$A3,$D6
	.dc.b	$FF,$11,$4E,$7A,$38,$F4,$7D,$DC,$7F,$CD,$27,$33,$F2,$38,$FE,$50
	.dc.b	$8E,$0F,$F7,$F9,$B3,$8C,$73,$C3,$FA,$57,$CE,$7E,$1B,$1F,$F1,$97
	.dc.b	$4B,$97,$FC,$91,$87,$E2,$FE,$1F,$F0,$E7,$E2,$3D,$D9,$19,$21,$F8
	.dc.b	$57,$F1,$BF,$0F,$FB,$FF,$46,$4F,$E5,$63,$2D,$21,$F8,$DF,$87,$F9
	.dc.b	$3F,$38,$E6,$FC,$BF,$C4,$4A,$F6,$CE,$5B,$3F,$B9,$8C,$47,$F8,$99
	.dc.b	$AF,$F8,$C5,$AF,$ED,$9D,$EF,$FD,$B5,$BB,$66,$BF,$B5,$4B,$E2,$3F
	.dc.b	$95,$5F,$E8,$DF,$E3,$9E,$AC,$8F,$D9,$47,$EB,$88,$E0,$FD,$B7,$C4
	.dc.b	$7E,$CB,$E2,$3F,$92,$7A,$F2,$7F,$75,$19,$69,$0F,$D9,$7C,$47,$93
	.dc.b	$E7,$1F,$17,$FE,$41,$CD,$18,$AF,$F2,$3C,$62,$75,$39,$9F,$E0,$ED
	.dc.b	$9F,$CA,$47,$FB,$86,$31,$AF,$F9,$DE,$79,$4F,$F0,$FF,$8A,$FC,$FF
	.dc.b	$CE,$7F,$3C,$8F,$C8,$FC,$E7,$FA,$26,$BF,$C3,$0D,$BF,$E4,$B5,$BA
	.dc.b	$FC,$EB,$5F,$F3,$33,$5D,$FA,$3F,$CD,$63,$11,$FE,$89,$F9,$C7,$C1
	.dc.b	$FF,$39,$6C,$7F,$98,$DA,$FF,$9A,$1A,$FF,$9C,$DC,$E3,$FE,$2D,$6E
	.dc.b	$1F,$CE,$9B,$3F,$E9,$7F,$9C,$74,$FF,$2A,$F8,$EF,$F1,$57,$69,$A4
	.dc.b	$AF,$F1,$54,$65,$FE,$6D,$1C,$5F,$99,$87,$EC,$BE,$5B,$FC,$6C,$DF
	.dc.b	$8A,$FF,$16,$C6,$27,$FA,$63,$7C,$03,$FE,$41,$AE,$13,$D0,$DB,$C8
	.dc.b	$E7,$5D,$3E,$2B,$FC,$85,$5F,$E3,$CA,$FE,$B3,$7D,$27,$24,$8E,$91
	.dc.b	$96,$CF,$F9,$13,$5F,$C4,$CF,$93,$E2,$BF,$91,$FC,$73,$DE,$57,$FB
	.dc.b	$07,$E7,$1E,$8F,$F6,$83,$FA,$A6,$59,$71,$EA,$7F,$6C,$9D,$CF,$FB
	.dc.b	$27,$0E,$97,$B8,$C6,$1D,$97,$C1,$F0,$7B,$E1,$FD,$DA,$31,$8F,$82
	.dc.b	$76,$FF,$7A,$C3,$AC,$60,$EB,$87,$77,$99,$67,$96,$3D,$EB,$8B,$71
	.dc.b	$F2,$3E,$0C,$7D,$95,$EE,$BE,$97,$C2,$3C,$57,$EB,$C7,$85,$7F,$BA
	.dc.b	$5A,$F9,$4F,$F6,$24,$6E,$78,$E3,$E5,$C3,$BD,$7F,$BC,$67,$CE,$69
	.dc.b	$63,$E7,$AE,$11,$0C,$CE,$E7,$51,$83,$A5,$76,$7E,$29,$93,$CB,$5E
	.dc.b	$EB,$F3,$47,$03,$D9,$7F,$0D,$F6,$63,$F4,$B1,$EF,$8F,$C0,$25,$78
	.dc.b	$CF,$54,$7C,$42,$38,$25,$77,$3E,$43,$F8,$18,$F5,$9E,$CF,$8F,$FF
	.dc.b	$36,$4F,$A0,$38,$BF,$C2,$B7,$0F,$C8,$3C,$24,$3E,$3A,$C6,$63,$07
	.dc.b	$4B,$FF,$7E,$CE,$9B,$19,$3F,$C9,$A7,$07,$F3,$99,$3E,$53,$F9,$CC
	.dc.b	$7E,$7E,$7C,$35,$F3,$1F,$E7,$53,$AC,$BA,$D7,$28,$C4,$F6,$AE,$0F
	.dc.b	$EE,$23,$2F,$FB,$EC,$FB,$97,$E1,$E2,$FD,$24,$78,$0C,$CF,$F4,$CD
	.dc.b	$3C,$91,$E9,$C7,$FD,$C5,$3D,$23,$86,$3E,$3B,$F0,$C7,$3B,$C3,$F8
	.dc.b	$B3,$17,$CA,$7C,$31,$F3,$27,$67,$83,$1F,$DC,$63,$E7,$9E,$46,$EF
	.dc.b	$93,$89,$DE,$3E,$99,$F8,$E6,$B1,$E9,$5F,$B1,$BE,$91,$8C,$3C,$35
	.dc.b	$F6,$9B,$5F,$34,$F8,$B8,$FA,$31,$EE,$D8,$E9,$8F,$C5,$7E,$41,$F6
	.dc.b	$A3,$E2,$BE,$93,$07,$A4,$EA,$7E,$D7,$7F,$E8,$FF,$1D,$FD,$FB,$3F
	.dc.b	$E6,$FC,$28,$C5,$FF,$53,$E9,$B8,$7F,$50,$9B,$87,$2A,$FC,$9C,$E5
	.dc.b	$B1,$98,$FC,$BB,$5F,$E9,$B6,$F8,$EE,$6B,$E3,$6C,$61,$B7,$FD,$39
	.dc.b	$B5,$FD,$2A,$3F,$B4,$47,$2E,$D9,$1B,$3F,$AA,$B8,$9F,$1D,$FF,$B4
	.dc.b	$A7,$CE,$D8,$CC,$FF,$88,$9F,$D7,$FC,$E3,$A3,$E1,$C2,$31,$7F,$DB
	.dc.b	$F5,$7C,$03,$CE,$DD,$6A,$FF,$CF,$6D,$7A,$E1,$DD,$FF,$64,$C6,$5B
	.dc.b	$19,$7F,$85,$FC,$DB,$C6,$C4,$23,$19,$7D,$48,$C4,$EA,$73,$1F,$61
	.dc.b	$B9,$7E,$0F,$E6,$1F,$2F,$EE,$9C,$E1,$FE,$41,$9C,$BF,$E0,$2E,$B7
	.dc.b	$CB,$FE,$11,$B1,$FE,$04,$F8,$7C,$4C,$8F,$C2,$CE,$B2,$FC,$2B,$5F
	.dc.b	$F0,$8B,$5F,$F1,$27,$C3,$FE,$18,$FC,$3F,$C0,$5A,$FF,$08,$C7,$E5
	.dc.b	$3E,$25,$F3,$59,$BF,$C3,$1F,$EF,$38,$CD,$7F,$86,$3C,$7F,$C8,$9F
	.dc.b	$92,$F4,$EC,$64,$FF,$0F,$CB,$E4,$F3,$7D,$3A,$D3,$9A,$FF,$58,$FC
	.dc.b	$DB,$E4,$3F,$0E,$D8,$FC,$37,$AD,$FE,$14,$77,$3A,$3E,$DC,$EF,$0F
	.dc.b	$D6,$B5,$FE,$82,$D7,$FA,$68,$78,$72,$FE,$F3,$D5,$E9,$7F,$E3,$67
	.dc.b	$F5,$FB,$48,$FF,$1C,$E9,$70,$F6,$76,$FC,$67,$DD,$A3,$E5,$35,$97
	.dc.b	$F8,$FA,$31,$87,$F9,$0A,$7F,$77,$F2,$9F,$BC,$6C,$64,$FF,$23,$C6
	.dc.b	$5A,$7F,$B6,$3E,$6D,$F3,$7F,$AE,$6C,$67,$CC,$C4,$C8,$FC,$F4,$EB
	.dc.b	$2F,$E2,$6D,$7F,$B3,$B5,$F8,$EE,$F2,$FF,$4F,$FC,$2B,$AF,$F9,$15
	.dc.b	$F8,$F0,$FE,$43,$39,$F1,$BF,$90,$F4,$7F,$80,$AB,$FC,$C6,$FD,$BC
	.dc.b	$BF,$08,$E5,$BE,$67,$0F,$37,$8B,$6B,$8B,$D3,$E5,$A4,$74,$E7,$B7
	.dc.b	$D2,$7F,$3D,$C1,$8D,$BE,$93,$E4,$F9,$6F,$2F,$CB,$5C,$7C,$AD,$C3
	.dc.b	$CA,$DB,$FF,$3B,$E8,$8C,$57,$F8,$63,$AD,$39,$8F,$7E,$1D,$9F,$83
	.dc.b	$E6,$FC,$4F,$96,$F8,$AD,$8C,$9F,$1A,$32,$D3,$E4,$F2,$FC,$CF,$96
	.dc.b	$F9,$AD,$8C,$9F,$3A,$32,$D3,$E8,$F2,$7D,$43,$EA,$FC,$B5,$8C,$9F
	.dc.b	$D5,$32,$FC,$16,$CD,$3D,$B7,$FC,$B5,$B5,$B8,$FB,$FF,$31,$CE,$BE
	.dc.b	$EB,$F9,$5F,$17,$F9,$8F,$4F,$F1,$79,$FF,$1F,$B5,$39,$C3,$E5,$FC
	.dc.b	$67,$7C,$8D,$26,$5E,$7F,$97,$FE,$DF,$FB,$2F,$E2,$F9,$3E,$BC,$63
	.dc.b	$A7,$F2,$7E,$86,$BA,$AE,$1F,$EB,$6C,$3F,$29,$87,$DE,$F8,$EF,$01
	.dc.b	$27,$E2,$3F,$07,$E7,$3F,$CA,$F1,$C7,$AF,$FB,$FC,$63,$83,$F9,$06
	.dc.b	$09,$CF,$B5,$AF,$95,$B8,$7F,$B1,$B0,$47,$2F,$96,$EC,$49,$FF,$03
	.dc.b	$CA,$3C,$5E,$E7,$83,$E2,$BC,$2D,$AE,$9F,$2B,$E4,$F9,$4F,$2B,$63
	.dc.b	$C1,$F2,$BE,$B2,$3D,$9F,$77,$C9,$7B,$0F,$6F,$B7,$67,$D1,$F2,$3F
	.dc.b	$03,$E3,$FE,$0B,$63,$D9,$E9,$FC,$72,$3E,$A7,$E5,$7C,$7F,$C8,$3E
	.dc.b	$4F,$B9,$67,$E2,$7C,$6F,$F5,$77,$E6,$B5,$FE,$9E,$7E,$4F,$D1,$34
	.dc.b	$87,$F8,$B9,$FF,$06,$74,$FF,$33,$F7,$B9,$74,$FF,$38,$7F,$B2,$7C
	.dc.b	$E4,$BF,$EB,$7F,$11,$6B,$F9,$EC,$F7,$24,$F5,$6C,$7F,$42,$F8,$8E
	.dc.b	$18,$7F,$41,$7F,$B8,$7A,$52,$FF,$B3,$FC,$35,$BF,$E9,$33,$EB,$24
	.dc.b	$FA,$5F,$77,$C3,$7F,$79,$F8,$6F,$EF,$4D,$9F,$EE,$2D,$AF,$EA,$6D
	.dc.b	$C7,$FC,$6D,$F3,$8D,$1F,$33,$FE,$A0,$F5,$FF,$B2,$CE,$63,$18,$7F
	.dc.b	$AD,$3C,$07,$CA,$FC,$E8,$FF,$5B,$5F,$0C,$39,$C6,$0F,$F5,$DE,$1C
	.dc.b	$DC,$07,$1F,$15,$FE,$01,$8E,$D7,$D6,$31,$1B,$AE,$B1,$8A,$E3,$7E
	.dc.b	$1C,$7C,$F7,$FA,$AC,$39,$4E,$63,$C3,$5C,$1F,$2E,$1D,$27,$FC,$0B
	.dc.b	$8F,$BB,$4E,$77,$BB,$F1,$63,$8C,$38,$47,$A2,$B1,$32,$3B,$CE,$9F
	.dc.b	$24,$E7,$2F,$54,$E6,$FB,$61,$E1,$BE,$38,$73,$7F,$DB,$73,$F2,$DB
	.dc.b	$5F,$41,$B3,$FE,$5B,$F8,$57,$A8,$F6,$F0,$DA,$FF,$49,$CE,$A7,$27
	.dc.b	$F9,$8F,$07,$1F,$FB,$EA,$BF,$D3,$31,$8F,$33,$53,$99,$F2,$B5,$FC
	.dc.b	$5B,$C7,$4F,$B0,$DD,$B3,$5F,$3B,$DF,$FF,$1A,$D6,$A3,$E8,$D6,$EF
	.dc.b	$F1,$6D,$7F,$15,$5C,$2F,$9C,$7D,$E9,$D6,$1E,$78,$FC,$0D,$6D,$E7
	.dc.b	$7D,$88,$E8,$B8,$78,$DE,$91,$E7,$73,$87,$82,$74,$7E,$56,$BB,$4F
	.dc.b	$E7,$F0,$F1,$92,$F9,$47,$E9,$FD,$17,$97,$FA,$0E,$3F,$D7,$46,$3D
	.dc.b	$AF,$F2,$CC,$62,$51,$8E,$3F,$F0,$94,$71,$6E,$99,$C3,$FB,$D4,$E5
	.dc.b	$B1,$9C,$3C,$3E,$DE,$E5,$FB,$EF,$F8,$7D,$FF,$53,$35,$FF,$6D,$B5
	.dc.b	$FF,$36,$9C,$23,$FD,$58,$D0,$7F,$A3,$8C,$16,$3A,$65,$ED,$F7,$3C
	.dc.b	$98,$BD,$CB,$EE,$4C,$3F,$BC,$46,$8F,$F5,$D6,$5F,$29,$B5,$C6,$7F
	.dc.b	$DD,$ED,$7F,$DD,$ED,$C3,$E6,$4F,$39,$FF,$02,$72,$F8,$BD,$B9,$1F
	.dc.b	$E0,$77,$CB,$EE,$48,$EA,$F8,$08,$F8,$3D,$CD,$3F,$E0,$48,$F6,$C7
	.dc.b	$BB,$D9,$93,$FE,$13,$7E,$0F,$C4,$FC,$2F,$6B,$D6,$49,$F5,$BE,$EF
	.dc.b	$77,$E3,$92,$3F,$C3,$9E,$DF,$CA,$F9,$49,$1F,$E1,$F7,$E6,$FB,$7F
	.dc.b	$3B,$E7,$3E,$39,$27,$E3,$BF,$33,$D2,$FB,$BC,$3F,$94,$6F,$4F,$F9
	.dc.b	$85,$B5,$FE,$47,$7F,$8A,$4E,$FD,$BF,$C7,$41,$A6,$3C,$5C,$FB,$5D
	.dc.b	$C9,$7F,$8C,$3A,$3E,$3F,$9B,$E6,$48,$FC,$6E,$1E,$5F,$6B,$BF,$B7
	.dc.b	$E0,$F6,$A5,$6B,$4F,$44,$F9,$3D,$BF,$71,$27,$C1,$D8,$D9,$3F,$0F
	.dc.b	$DB,$F5,$92,$BA,$1F,$07,$CD,$F7,$FB,$7F,$02,$E4,$EB,$0E,$93,$F1
	.dc.b	$3D,$AF,$9C,$49,$F0,$D2,$3E,$97,$CC,$7C,$A2,$4F,$39,$7B,$7F,$BB
	.dc.b	$FA,$87,$C6,$FE,$F7,$E8,$9B,$D3,$51,$30,$F7,$D7,$06,$C7,$EB,$23
	.dc.b	$11,$FB,$ED,$11,$8E,$3F,$E2,$ED,$E1,$FF,$52,$36,$BF,$09,$EF,$FE
	.dc.b	$0B,$DF,$91,$E9,$7F,$0B,$EF,$E2,$18,$7E,$3E,$70,$FE,$46,$7F,$17
	.dc.b	$EF,$CA,$F8,$4F,$E3,$BE,$4F,$11,$CB,$2F,$CB,$4E,$2F,$F3,$55,$FC
	.dc.b	$87,$F4,$1F,$F0,$E5,$EC,$FF,$50,$38,$9F,$E6,$6D,$BF,$F6,$02,$F8
	.dc.b	$0E,$51,$2B,$EF,$4E,$DE,$73,$97,$B6,$A9,$D3,$6B,$C9,$5E,$5F,$80
	.dc.b	$F3,$7B,$93,$0E,$B1,$96,$90,$EB,$87,$F7,$A7,$C7,$EF,$FB,$C8,$F9
	.dc.b	$BE,$02,$BC,$22,$E7,$F0,$27,$FA,$27,$CE,$FF,$AC,$AB,$F3,$E1,$79
	.dc.b	$BA,$37,$E6,$49,$FD,$D1,$B7,$FA,$2C,$F3,$8F,$DF,$35,$FE,$96,$E6
	.dc.b	$FE,$AB,$75,$FE,$97,$3C,$CF,$D8,$FB,$AF,$A2,$F9,$7A,$3E,$07,$EF
	.dc.b	$12,$BE,$AF,$D0,$D7,$E3,$C5,$CB,$EF,$CF,$E1,$7E,$0F,$F0,$DF,$2D
	.dc.b	$2F,$EE,$C6,$5A,$43,$E8,$F5,$FC,$8F,$4B,$FB,$96,$8F,$A0,$FF,$62
	.dc.b	$57,$A1,$AF,$F6,$5D,$D5,$C6,$3E,$4C,$61,$B1,$F2,$88,$FD,$48,$D3
	.dc.b	$CE,$3C,$3B,$7F,$A0,$63,$E3,$AE,$4F,$D1,$7C,$54,$9C,$C7,$02,$38
	.dc.b	$3F,$3B,$F2,$BF,$A9,$F9,$5F,$D5,$7C,$AC,$9F,$D0,$C6,$5A,$43,$F2
	.dc.b	$BF,$29,$FE,$0F,$FC,$15,$F0,$1F,$E0,$86,$CF,$E2,$1A,$FF,$43,$6B
	.dc.b	$FE,$10,$73,$C9,$FB,$A6,$F3,$FE,$19,$B2,$EE,$EF,$F0,$FF,$E2,$1F
	.dc.b	$C1,$3F,$19,$39,$E6,$FF,$02,$61,$FE,$18,$B4,$62,$79,$57,$09,$34
	.dc.b	$9D,$7D,$78,$FF,$C5,$9F,$3C,$E9,$FF,$12,$B7,$0F,$D2,$B5,$FE,$C0
	.dc.b	$D7,$FC,$5A,$E7,$5F,$E2,$ED,$CF,$F4,$6D,$9F,$F1,$7F,$CF,$BE,$4B
	.dc.b	$F8,$F7,$CC,$7F,$89,$BB,$CD,$25,$7F,$60,$8C,$BF,$C6,$A3,$8B,$FC
	.dc.b	$73,$83,$F9,$41,$87,$F8,$7C,$61,$D1,$0F,$E2,$71,$FC,$BF,$E5,$FF
	.dc.b	$98,$7C,$B7,$8C,$93,$E1,$6B,$FC,$69,$DB,$FD,$A7,$E6,$7F,$D2,$1A
	.dc.b	$FF,$38,$F9,$27,$C5,$7F,$48,$A7,$CE,$3D,$CF,$E0,$E7,$F8,$46,$3F
	.dc.b	$CA,$71,$97,$FC,$AB,$19,$78,$BF,$C3,$AB,$AD,$F0,$8C,$5F,$7D,$B9
	.dc.b	$5F,$7F,$80,$FE,$2B,$18,$AE,$8E,$B0,$E7,$5F,$56,$79,$E1,$D6,$3C
	.dc.b	$77,$FC,$7B,$1E,$4D,$AD,$CF,$98,$FF,$36,$DF,$4A,$DC,$73,$3C,$85
	.dc.b	$8D,$4F,$2C,$37,$3E,$33,$C2,$7F,$9E,$1F,$46,$3E,$F9,$F0,$DF,$02
	.dc.b	$38,$25,$72,$3C,$C7,$88,$B7,$E2,$7D,$35,$F6,$6B,$9C,$71,$7C,$38
	.dc.b	$FA,$67,$C5,$3D,$70,$F4,$9A,$7D,$42,$FB,$3E,$1B,$EA,$E6,$BC,$18
	.dc.b	$7C,$33,$F0,$F8,$FD,$A6,$CF,$C3,$74,$FA,$31,$C3,$E3,$8D,$BE,$2C
	.dc.b	$3E,$FE,$3E,$5F,$90,$FD,$A4,$F1,$0E,$1F,$25,$FF,$AA,$23,$15,$F8
	.dc.b	$87,$53,$FE,$AC,$F9,$2D,$5F,$F9,$9D,$B1,$FE,$B4,$6B,$FE,$EE,$3B
	.dc.b	$C6,$DF,$8F,$7F,$CC,$B9,$3D,$71,$B9,$FC,$EC,$E6,$3E,$69,$E1,$38
	.dc.b	$F6,$2B,$8C,$66,$B8,$CE,$63,$88,$BF,$CB,$5F,$5D,$7F,$57,$3D,$5E
	.dc.b	$38,$7E,$6E,$3A,$67,$FE,$E4,$C3,$A6,$3C,$27,$A1,$E4,$AF,$1B,$8B
	.dc.b	$E6,$FE,$50,$ED,$7C,$E7,$B4,$6E,$30,$EC,$C1,$FE,$90,$BF,$65,$78
	.dc.b	$71,$F5,$C6,$2B,$E7,$4F,$80,$D5,$79,$4F,$87,$5F,$43,$1F,$9D,$8E
	.dc.b	$AB,$85,$F2,$3C,$6E,$5A,$66,$FC,$98,$FA,$B0,$E9,$8F,$C6,$AF,$C0
	.dc.b	$61,$F0,$8E,$57,$C0,$F7,$57,$AA,$FF,$8A,$75,$37,$FE,$99,$1F,$CC
	.dc.b	$EF,$EE,$67,$F7,$6B,$5F,$2F,$8A,$FE,$D5,$7C,$31,$FF,$30,$B5,$FF
	.dc.b	$1D,$D7,$5C,$3F,$CE,$71,$C1,$AF,$F7,$99,$E0,$F4,$F9,$1E,$CE,$5B
	.dc.b	$5E,$7F,$92,$B4,$9A,$63,$E4,$AE,$3F,$D3,$5C,$34,$FE,$A1,$18,$C3
	.dc.b	$D3,$3E,$4C,$3E,$23,$6C,$D3,$D3,$2F,$9B,$7A,$DE,$88,$C3,$EA,$D5
	.dc.b	$F4,$9F,$DD,$B3,$FF,$37,$DF,$06,$D7,$48,$C5,$FF,$B9,$DA,$FF,$99
	.dc.b	$8E,$B3,$FE,$DC,$FC,$0D,$EC,$7C,$66,$C7,$FB,$55,$7B,$19,$99,$3F
	.dc.b	$ED,$9F,$15,$F2,$18,$C3,$FD,$37,$18,$9F,$F4,$2F,$E0,$7F,$EF,$F6
	.dc.b	$F5,$FD,$4F,$C7,$3B,$DF,$FA,$BC,$56,$DC,$35,$F5,$35,$FF,$40,$B5
	.dc.b	$1C,$A7,$E9,$74,$FE,$00,$FC,$0F,$CA,$DC,$3F,$C1,$51,$F8,$4E,$EC
	.dc.b	$5F,$E5,$1F,$C5,$FA,$9F,$CD,$3A,$2F,$71,$FC,$5F,$CA,$E1,$98,$FF
	.dc.b	$85,$7E,$56,$E3,$F9,$E7,$EF,$7C,$EB,$E5,$A9,$D3,$AA,$DF,$FA,$B2
	.dc.b	$74,$FF,$AA,$1B,$7F,$6C,$38,$7E,$F3,$5F,$F5,$2B,$C3,$AB,$F9,$86
	.dc.b	$8F,$9E,$7A,$3E,$0B,$FF,$70,$4E,$63,$F4,$FE,$EC,$C7,$FC,$61,$CF
	.dc.b	$6F,$F7,$33,$FA,$CF,$43,$19,$7F,$01,$EF,$47,$EC,$3D,$09,$8F,$F8
	.dc.b	$E7,$7B,$8F,$F0,$29,$B8,$71,$9B,$5E,$AF,$97,$F9,$27,$D2,$E6,$B1
	.dc.b	$F1,$E3,$F5,$BC,$06,$92,$3E,$53,$F4,$BE,$5B,$8E,$5F,$91,$A4,$7F
	.dc.b	$06,$99,$5F,$E4,$FF,$97,$B7,$FC,$62,$6E,$5E,$F9,$B5,$FC,$7A,$FF
	.dc.b	$87,$9F,$C4,$3E,$5E,$C7,$F9,$66,$3F,$89,$FC,$BE,$31,$FE,$46,$FF
	.dc.b	$18,$FA,$5F,$CD,$D7,$E7,$31,$FE,$59,$18,$A4,$63,$E7,$78,$44,$AE
	.dc.b	$07,$FA,$66,$B8,$56,$E3,$33,$FE,$A3,$9C,$C6,$E3,$91,$D9,$E5,$58
	.dc.b	$AF,$1F,$77,$9C,$F4,$7C,$CD,$AE,$BE,$4B,$1F,$DE,$5F,$09,$CC,$2B
	.dc.b	$86,$3D,$CF,$27,$CC,$FC,$33,$E2,$7C,$CD,$BE,$31,$67,$86,$3D,$EB
	.dc.b	$18,$FF,$53,$FA,$AF,$CF,$E1,$E4,$6D,$F0,$CB,$FD,$0D,$39,$EA,$FE
	.dc.b	$93,$5A,$AF,$EA,$AD,$7F,$D8,$8D,$7F,$D9,$8D,$C4,$F6,$C3,$FB,$27
	.dc.b	$C8,$BE,$4B,$AE,$1F,$D5,$A3,$18,$FC,$4F,$92,$FC,$11,$F8,$3F,$9A
	.dc.b	$B1,$F2,$3E,$6B,$FA,$86,$A6,$92,$3F,$83,$69,$F4,$7E,$6B,$F1,$9F
	.dc.b	$35,$F8,$D6,$CF,$D2,$F9,$AF,$C8,$1F,$91,$F9,$8B,$1F,$5B,$E6,$BB
	.dc.b	$91,$EB,$92,$3C,$7F,$35,$F9,$9E,$0F,$CD,$36,$BE,$FF,$CC,$FE,$74
	.dc.b	$FC,$F7,$CC,$D8,$FC,$1F,$B1,$EC,$23,$D9,$F7,$FA,$BF,$A5,$3F,$4D
	.dc.b	$EB,$59,$FC,$7F,$8F,$FE,$51,$69,$FE,$02,$5F,$33,$63,$1F,$1F,$FD
	.dc.b	$E6,$4D,$24,$7C,$A8,$7C,$CF,$33,$A7,$CE,$D9,$F0,$FC,$D7,$73,$C1
	.dc.b	$F3,$56,$3C,$BF,$35,$DC,$8F,$57,$C5,$F3,$5E,$73,$D1,$EF,$59,$F6
	.dc.b	$FC,$87,$AF,$E6,$BD,$8D,$8F,$83,$1E,$B2,$3E,$97,$DD,$F2,$FF,$0C
	.dc.b	$E1,$E2,$FE,$A3,$64,$7C,$7F,$97,$97,$F3,$3E,$5A,$DE,$E7,$E4,$12
	.dc.b	$7E,$33,$63,$A7,$CB,$F0,$F1,$FF,$C8,$98,$A3,$FC,$41,$F3,$BF,$E4
	.dc.b	$39,$FF,$42,$52,$31,$87,$F7,$CF,$0F,$FA,$5F,$91,$C3,$E2,$A4,$FF
	.dc.b	$63,$BE,$11,$FE,$2E,$BF,$F2,$2C,$7F,$81,$1B,$7F,$D5,$3E,$4F,$F0
	.dc.b	$87,$E1,$7E,$4E,$E1,$F8,$A8,$FC,$47,$C9,$E3,$0F,$8C,$FF,$8D,$FE
	.dc.b	$7B,$F0,$7E,$9F,$FA,$C7,$E3,$BF,$C2,$34,$9C,$BF,$E9,$D8,$CF,$A6
	.dc.b	$F3,$3F,$D4,$5F,$21,$FE,$53,$AF,$F0,$CD,$F0,$8C,$47,$3E,$F6,$3D
	.dc.b	$E3,$71,$87,$B1,$FE,$22,$C3,$B6,$9B,$8E,$55,$70,$ED,$97,$A1,$B3
	.dc.b	$C7,$2F,$2C,$E7,$1E,$D9,$79,$2F,$77,$D0,$F0,$9F,$CE,$EF,$8B,$E6
	.dc.b	$C3,$C1,$7C,$B0,$E9,$1E,$5C,$BC,$F5,$CB,$0E,$A4,$BE,$11,$E5,$CB
	.dc.b	$9D,$C8,$E1,$3A,$7C,$53,$9C,$BD,$13,$F1,$E7,$FC,$C9,$7F,$59,$B1
	.dc.b	$97,$CA,$7F,$8B,$FE,$4D,$F1,$0F,$84,$FC,$84,$61,$B1,$99,$E0,$D7
	.dc.b	$BF,$85,$FE,$FE,$7F,$C7,$58,$23,$11,$AF,$89,$FC,$E4,$E5,$B3,$C7
	.dc.b	$2F,$A4,$DD,$BE,$D3,$6F,$F1,$7C,$5F,$94,$AD,$DF,$E0,$6B,$55,$F9
	.dc.b	$76,$E0,$BE,$18,$74,$BE,$04,$BF,$AD,$3A,$7F,$33,$7C,$89,$3F,$8F
	.dc.b	$9D,$4F,$78,$E5,$87,$92,$39,$98,$BF,$03,$F8,$3B,$F2,$DF,$2C,$3E
	.dc.b	$24,$E7,$0E,$0F,$C0,$3F,$4D,$C6,$F9,$EF,$D5,$F4,$7F,$B5,$5F,$F3
	.dc.b	$D4,$FF,$85,$9A,$FF,$82,$9D,$F0,$7F,$11,$6C,$2B,$A7,$C5,$7F,$00
	.dc.b	$BE,$8D,$BE,$1A,$5B,$FE,$25,$F1,$37,$0F,$F3,$2F,$CA,$DE,$D3,$BA
	.dc.b	$FE,$45,$57,$AB,$FD,$97,$1F,$C7,$DA,$FF,$A0,$DA,$FF,$AF,$DA,$FA
	.dc.b	$7E,$37,$C1,$AB,$E3,$69,$88,$93,$F0,$E3,$2F,$C5,$8C,$BF,$15,$BB
	.dc.b	$7C,$BE,$4E,$BF,$1B,$ED,$FD,$9F,$F3,$72,$E0,$FE,$75,$18,$7E,$64
	.dc.b	$F9,$3E,$33,$8E,$3F,$4E,$30,$FD,$58,$AF,$D5,$7D,$7F,$1B,$EC,$E0
	.dc.b	$F0,$78,$32,$BE,$E4,$7C,$0F,$8B,$F8,$3F,$19,$27,$F0,$11,$F1,$3E
	.dc.b	$2B,$51,$83,$F0,$8F,$C8,$F8,$8F,$E9,$DC,$DF,$8B,$6F,$1F,$FA,$22
	.dc.b	$BF,$C5,$ED,$7F,$DE,$13,$F9,$1F,$88,$F8,$9F,$29,$27,$F2,$11,$F8
	.dc.b	$3F,$7F,$FA,$D4,$7E,$35,$B4,$69,$3E,$27,$81,$27,$FC,$DF,$1F,$E6
	.dc.b	$B6,$FC,$4F,$D3,$EC,$91,$83,$7F,$13,$E1,$23,$CB,$E5,$7C,$9F,$11
	.dc.b	$E5,$F8,$89,$3D,$A3,$D1,$F1,$1F,$7F,$C2,$93,$E2,$8F,$67,$C4,$7E
	.dc.b	$0F,$C2,$93,$E6,$8F,$81,$F1,$1F,$07,$69,$3E,$A8,$F8,$9F,$0F,$F8
	.dc.b	$BF,$8E,$93,$EE,$8F,$91,$F0,$FF,$8F,$F1,$24,$FC,$28,$FD,$77,$CF
	.dc.b	$BD,$BF,$D8,$7C,$5B,$82,$9A,$89,$7F,$E9,$89,$D3,$67,$2F,$FA,$3A
	.dc.b	$BA,$3F,$E6,$E6,$BF,$80,$6B,$FE,$04,$E7,$46,$23,$FC,$39,$AB,$E2
	.dc.b	$3F,$04,$48,$FB,$7F,$11,$5F,$F4,$EC,$ED,$FB,$F3,$F8,$99,$FC,$57
	.dc.b	$C4,$49,$FC,$1B,$F8,$DF,$88,$C4,$72,$8B,$E5,$7E,$D6,$3F,$21,$3F
	.dc.b	$C5,$7E,$63,$FC,$67,$39,$CD,$18,$F2,$BF,$C6,$8D,$BF,$E4,$8F,$F9
	.dc.b	$B9,$FF,$60,$B9,$9E,$33,$D4,$ED,$F1,$36,$B7,$15,$FF,$5D,$D7,$36
	.dc.b	$FC,$4F,$97,$D9,$91,$DA,$3D,$07,$ED,$A7,$FC,$F7,$1F,$EA,$26,$BF
	.dc.b	$B4,$6B,$F5,$57,$C4,$FC,$12,$47,$AB,$E2,$6B,$E4,$8B,$9F,$C5,$A5
	.dc.b	$7F,$39,$F9,$E7,$99,$7C,$F9,$8F,$FB,$C0,$4E,$9C,$35,$FF,$71,$1A
	.dc.b	$8E,$2E,$23,$58,$7F,$9A,$1E,$11,$FE,$52,$3A,$D7,$F0,$C8,$CC,$7F
	.dc.b	$95,$9A,$FD,$8F,$8A,$FC,$01,$1F,$B3,$F1,$55,$FA,$71,$6F,$E0,$4F
	.dc.b	$E1,$FE,$2B,$F1,$1F,$15,$2F,$F0,$31,$97,$E9,$3F,$54,$8F,$F9,$E7
	.dc.b	$07,$A3,$FD,$9F,$DF,$7C,$FB,$CF,$C3,$6F,$E3,$D7,$CF,$33,$32,$32
	.dc.b	$43,$E2,$1F,$42,$3E,$21,$1C,$D3,$E7,$FF,$0F,$A7,$DD,$F8,$CF,$D2
	.dc.b	$1E,$58,$CF,$32,$FF,$C8,$91,$F8,$C7,$17,$CF,$97,$FD,$2C,$D8,$EF
	.dc.b	$3B,$7F,$53,$F1,$9F,$B1,$F8,$B9,$1F,$AA,$8F,$DA,$1F,$9D,$23,$F9
	.dc.b	$D7,$FA,$4F,$CF,$BC,$8F,$F0,$77,$3B,$D9,$A7,$F8,$36,$75,$7F,$93
	.dc.b	$6B,$F1,$9A,$FE,$2D,$AF,$FA,$79,$AF,$F8,$8F,$E7,$FF,$27,$AF,$FA
	.dc.b	$DF,$91,$E4,$FF,$87,$DB,$5F,$1F,$87,$F8,$65,$F0,$8E,$5D,$3B,$B9
	.dc.b	$1B,$C3,$FD,$85,$5B,$87,$CF,$BA,$3F,$C5,$1F,$3F,$FE,$36,$93,$67
	.dc.b	$51,$FB,$16,$BF,$84,$6B,$FA,$96,$BF,$EB,$C6,$BF,$C7,$BE,$3B,$FC
	.dc.b	$3D,$F0,$52,$3F,$8F,$C7,$F2,$93,$F8,$91,$1F,$E2,$4F,$F1,$5F,$87
	.dc.b	$D4,$70,$3F,$93,$BF,$CC,$4C,$BF,$C8,$A3,$0F,$F2,$CF,$85,$FF,$17
	.dc.b	$7C,$24,$8C,$10,$FE,$3D,$1F,$C8,$C8,$FF,$A6,$2D,$BF,$E6,$7E,$7D
	.dc.b	$EC,$FF,$B6,$87,$BB,$FE,$DC,$8C,$DF,$E7,$E7,$2F,$17,$FD,$C3,$F1
	.dc.b	$7F,$E4,$DB,$C4,$FE,$96,$30,$AB,$9B,$E0,$BE,$D7,$E1,$6C,$71,$C3
	.dc.b	$CE,$F1,$9E,$F1,$8E,$6E,$71,$E4,$9F,$D0,$3C,$DF,$DB,$47,$83,$0F
	.dc.b	$1B,$6B,$AE,$1C,$B0,$F0,$DF,$94,$F5,$8B,$F5,$CF,$1C,$73,$8F,$86
	.dc.b	$39,$57,$0A,$DC,$FA,$2B,$77,$DA,$38,$19,$AF,$71,$F0,$67,$6C,$32
	.dc.b	$78,$F0,$F4,$CF,$1C,$3C,$33,$A8,$EA,$F3,$38,$E1,$E4,$7E,$E6,$1E
	.dc.b	$33,$E0,$C6,$63,$B6,$1E,$1B,$EA,$7C,$18,$E7,$87,$89,$E6,$6F,$0F
	.dc.b	$21,$1C,$9E,$5C,$3B,$CF,$18,$EE,$3E,$23,$F8,$D3,$E9,$6B,$A6,$E3
	.dc.b	$F8,$E7,$31,$FC,$D7,$E2,$FF,$9B,$46,$21,$18,$CF,$FA,$2B,$63,$FB
	.dc.b	$53,$5F,$F7,$51,$F4,$33,$FE,$5D,$C9,$F8,$48,$EC,$67,$AB,$FC,$59
	.dc.b	$AF,$E7,$A3,$2F,$E8,$2B,$8C,$7D,$A3,$3A,$FF,$B2,$EB,$8B,$E4,$8E
	.dc.b	$97,$DD,$F4,$98,$39,$18,$BE,$D5,$CA,$3D,$66,$E3,$C8,$EE,$BC,$15
	.dc.b	$D7,$0F,$CF,$E1,$E6,$BD,$C6,$30,$E9,$87,$C4,$9E,$F7,$C6,$3B,$61
	.dc.b	$E9,$9C,$C7,$9A,$F9,$56,$E7,$27,$69,$E2,$7A,$1C,$D7,$C3,$C7,$99
	.dc.b	$B2,$BD,$CF,$5C,$F7,$39,$9E,$77,$33,$DC,$EB,$1F,$EF,$DF,$99,$FE
	.dc.b	$D1,$7F,$E0,$8F,$35,$18,$9F,$F5,$EB,$5F,$F3,$F3,$5F,$F0,$24,$70
	.dc.b	$C3,$FC,$DD,$9F,$F9,$AF,$E2,$5F,$13,$8C,$77,$8F,$5A,$D4,$E7,$1F
	.dc.b	$27,$C1,$5C,$BC,$7E,$F5,$C1,$36,$30,$DC,$7F,$B5,$5F,$26,$DF,$26
	.dc.b	$DA,$FF,$D0,$95,$EC,$6B,$E6,$6B,$F1,$DB,$5C,$E3,$FA,$6D,$7F,$C0
	.dc.b	$0F,$37,$C5,$18,$12,$CD,$3E,$F9,$7C,$F3,$D5,$EF,$18,$6C,$67,$0F
	.dc.b	$F1,$EC,$E5,$B1,$98,$F0,$6A,$F9,$BF,$F2,$7D,$7F,$9C,$EB,$53,$F5
	.dc.b	$A7,$35,$E7,$C9,$F3,$CF,$5F,$E9,$78,$8F,$8C,$FA,$93,$9C,$FE,$6C
	.dc.b	$E5,$F9,$F1,$9C,$3A,$4E,$5E,$D1,$9C,$BF,$D6,$13,$97,$E9,$FC,$EF
	.dc.b	$F7,$4B,$F7,$B6,$FF,$CB,$AD,$7F,$D0,$4D,$7F,$CA,$C1,$E1,$3F,$EB
	.dc.b	$DF,$23,$F2,$75,$FE,$95,$9C,$BF,$9A,$8C,$D7,$C3,$E2,$FE,$32,$FE
	.dc.b	$83,$A7,$F0,$31,$88,$FC,$DC,$60,$FC,$B7,$67,$E5,$DB,$7F,$94,$F3
	.dc.b	$3F,$9F,$7C,$2F,$E3,$1B,$1F,$E7,$4E,$D4,$E5,$B1,$98,$FD,$17,$4F
	.dc.b	$E9,$EB,$8E,$3E,$78,$FC,$67,$CE,$BE,$3B,$F1,$6D,$C3,$F1,$2D,$AF
	.dc.b	$F4,$3B,$5F,$F5,$63,$9D,$FF,$DD,$18,$7E,$2F,$E5,$BA,$E1,$E1,$E6
	.dc.b	$46,$31,$F7,$46,$1F,$F6,$16,$1C,$DF,$F4,$3C,$67,$1F,$9D,$D3,$F3
	.dc.b	$09,$3F,$C4,$F8,$3F,$4E,$D8,$FE,$05,$E6,$7F,$03,$6C,$7E,$F3,$D3
	.dc.b	$FE,$0C,$7F,$07,$CE,$C6,$49,$87,$DD,$8C,$BF,$ED,$2C,$9F,$1F,$F2
	.dc.b	$09,$5F,$21,$B3,$BF,$8F,$F9,$F1,$F4,$1B,$3F,$01,$B3,$B8,$FE,$23
	.dc.b	$3F,$C6,$CF,$E3,$9F,$3B,$63,$24,$9F,$C9,$46,$5F,$CA,$57,$E4,$89
	.dc.b	$3F,$1D,$B1,$F9,$9F,$8F,$FE,$51,$1F,$CA,$5F,$E5,$9F,$37,$FE,$BB
	.dc.b	$AF,$E4,$AD,$AF,$E0,$4B,$E5,$77,$12,$76,$F2,$8C,$3C,$0B,$C6,$9F
	.dc.b	$E8,$B5,$CD,$E2,$78,$3C,$AF,$19,$E4,$F9,$6B,$19,$25,$F9,$23,$2F
	.dc.b	$9B,$E4,$BD,$67,$B3,$E4,$AD,$7B,$22,$CF,$B2,$6E,$3E,$58,$FE,$6B
	.dc.b	$F3,$9F,$93,$7F,$84,$F8,$88,$C4,$7F,$5E,$6B,$C3,$A3,$FA,$1C,$7E
	.dc.b	$6E,$BA,$53,$E3,$1F,$39,$FD,$6E,$B5,$9F,$F6,$89,$CB,$FD,$5A,$33
	.dc.b	$8F,$F4,$89,$CB,$FC,$D6,$33,$7D,$6F,$ED,$FC,$6F,$F6,$7F,$2B,$F9
	.dc.b	$56,$3F,$42,$31,$E3,$FE,$03,$E4,$F5,$D9,$6B,$FB,$2D,$A3,$F0,$BF
	.dc.b	$27,$F8,$6F,$99,$B5,$F8,$67,$F1,$3F,$29,$DF,$BA,$CF,$27,$F1,$BF
	.dc.b	$29,$E3,$EF,$B3,$D9,$FC,$8F,$CA,$7E,$4B,$7B,$3F,$92,$7F,$2B,$F3
	.dc.b	$92,$3F,$2B,$F2,$57,$93,$F2,$BE,$97,$BF,$C2,$FC,$DB,$6B,$F3,$7E
	.dc.b	$8F,$E7,$8F,$CF,$FC,$95,$8F,$F0,$1F,$17,$E5,$5F,$F1,$03,$70,$FE
	.dc.b	$92,$7F,$78,$70,$FE,$11,$AF,$F8,$81,$7C,$47,$D1,$F1,$7F,$C3,$11
	.dc.b	$8A,$F9,$B1,$8F,$91,$FF,$04,$C9,$A4,$9F,$1F,$C8,$F2,$F9,$0E,$6D
	.dc.b	$8F,$4C,$76,$F5,$71,$97,$F7,$9C,$11,$E2,$F9,$0F,$C2,$79,$57,$0F
	.dc.b	$BE,$F9,$BE,$56,$47,$C4,$F8,$7B,$8F,$39,$F5,$FC,$76,$3C,$5F,$CE
	.dc.b	$E6,$8C,$E1,$F8,$AF,$9D,$B8,$7F,$82,$FC,$04,$7C,$4F,$9D,$95,$F4
	.dc.b	$BE,$3A,$DF,$A6,$7E,$4F,$C7,$E3,$3E,$ED,$BF,$A0,$DA,$FA,$CF,$F8
	.dc.b	$AF,$E6,$BF,$CA,$90,$0E,$1B,$F0,$7F,$E2,$E8,$C1,$FE,$53,$AF,$F2
	.dc.b	$DE,$6F,$4F,$A5,$7E,$B5,$FB,$BF,$15,$F7,$BE,$76,$DF,$E3,$1B,$8F
	.dc.b	$D3,$6E,$7F,$21,$FF,$20,$7C,$E7,$E0,$7D,$6F,$F5,$C7,$C3,$7F,$86
	.dc.b	$23,$17,$FE,$B8,$8F,$F0,$DF,$AA,$F2,$FF,$8C,$7C,$4F,$F9,$8A,$BD
	.dc.b	$D7,$C2,$31,$08,$C5,$F5,$EE,$FE,$79,$3D,$4F,$8D,$87,$5D,$3C,$71
	.dc.b	$D2,$77,$87,$4C,$39,$CE,$72,$DA,$C3,$93,$D4,$F1,$5F,$2C,$7C,$55
	.dc.b	$C2,$F9,$D7,$1C,$BD,$B5,$A8,$E5,$3A,$37,$87,$6C,$7C,$98,$F5,$AF
	.dc.b	$21,$C1,$C3,$D7,$2F,$04,$62,$FA,$3A,$BE,$98,$78,$70,$EB,$97,$F9
	.dc.b	$C2,$FE,$AE,$3F,$C1,$9A,$F9,$CF,$E1,$3F,$14,$F8,$C7,$BF,$FE,$78
	.dc.b	$8C,$36,$33,$3F,$C4,$04,$E9,$C5,$3E,$75,$F0,$7E,$36,$CE,$9A,$F9
	.dc.b	$5A,$FE,$35,$DE,$5F,$6D,$B9,$E6,$BE,$A7,$C7,$7F,$9A,$AB,$51,$F5
	.dc.b	$EB,$77,$F9,$06,$BF,$8F,$AE,$17,$CE,$3E,$CC,$EB,$0F,$55,$F6,$9E
	.dc.b	$8C,$76,$E5,$A4,$C3,$C6,$FE,$13,$0E,$EF,$8E,$79,$4E,$8F,$CC,$4E
	.dc.b	$A7,$F4,$97,$E3,$25,$77,$AE,$18,$7A,$5B,$5F,$AE,$BE,$58,$7A,$A3
	.dc.b	$C9,$5B,$3F,$75,$3A,$C3,$AC,$E8,$E4,$79,$A7,$83,$E9,$3D,$86,$4C
	.dc.b	$5F,$89,$F9,$C6,$4F,$5E,$17,$06,$1E,$18,$C6,$1F,$46,$31,$5E,$33
	.dc.b	$89,$E0,$31,$5E,$DB,$F0,$3D,$23,$E3,$61,$E8,$78,$CF,$CE,$72,$7A
	.dc.b	$23,$53,$C6,$3B,$6F,$FA,$9B,$FF,$5B,$B6,$3F,$CA,$47,$C6,$3F,$DE
	.dc.b	$73,$C6,$FB,$47,$9D,$FE,$29,$C3,$F8,$2F,$5D,$FB,$47,$F7,$A9,$FE
	.dc.b	$4D,$18,$E0,$D4,$FF,$22,$BF,$E4,$53,$C1,$C9,$E4,$3F,$A5,$39,$7C
	.dc.b	$AD,$7F,$C9,$47,$FB,$FA,$3F,$C1,$38,$FF,$92,$9F,$16,$7C,$9B,$3B
	.dc.b	$3C,$7E,$F7,$93,$8B,$CB,$EF,$49,$FE,$67,$8F,$B2,$75,$1E,$2B,$FE
	.dc.b	$60,$D8,$F4,$3E,$36,$CF,$18,$E7,$3C,$1F,$82,$DC,$FE,$3A,$F7,$BE
	.dc.b	$3F,$C7,$4A,$E2,$FC,$AF,$73,$E5,$FC,$EF,$6B,$34,$95,$E7,$F7,$3E
	.dc.b	$89,$1F,$1F,$B9,$F5,$3A,$E4,$79,$9F,$AF,$EE,$7D,$8F,$9C,$EC,$49
	.dc.b	$F8,$4F,$87,$DC,$FB,$A4,$8F,$7F,$B7,$FC,$07,$E6,$BF,$92,$F0,$7F
	.dc.b	$5B,$F9,$2F,$DD,$99,$ED,$97,$F9,$21,$5F,$CB,$3A,$3F,$C8,$AD,$8F
	.dc.b	$F0,$71,$A9,$FA,$2D,$3F,$B2,$C7,$83,$A3,$6D,$8F,$F0,$B2,$F1,$7F
	.dc.b	$2B,$EE,$C9,$FA,$EF,$E6,$3D,$DC,$47,$D8,$8B,$9F,$D8,$9F,$CE,$FB
	.dc.b	$BF,$9E,$F7,$25,$FE,$02,$32,$FE,$0A,$32,$FE,$0A,$3F,$4B,$EE,$7E
	.dc.b	$9B,$DC,$93,$F8,$88,$CB,$F8,$A8,$CB,$F8,$A8,$CB,$F8,$C8,$CE,$1F
	.dc.b	$8C,$9F,$DB,$7E,$D3,$FA,$88,$FD,$AA,$F9,$DC,$9A,$E0,$98,$7E,$8A
	.dc.b	$78,$6F,$FB,$46,$C3,$B9,$1F,$E7,$48,$E1,$E5,$78,$3C,$79,$18,$21
	.dc.b	$FC,$0A,$7F,$B7,$3D,$E3,$CF,$3F,$89,$F8,$DF,$C5,$61,$2B,$EE,$C6
	.dc.b	$5F,$EF,$47,$E2,$1A,$D3,$3C,$68,$CB,$63,$FC,$B1,$E6,$7F,$0C,$8C
	.dc.b	$42,$31,$3F,$90,$75,$F3,$17,$4F,$89,$EF,$5B,$F8,$9A,$23,$1A,$FC
	.dc.b	$66,$C3,$E4,$2C,$4A,$F1,$DF,$46,$9D,$27,$8F,$BF,$CB,$DE,$93,$AC
	.dc.b	$7E,$9C,$EA,$31,$8A,$31,$1E,$DC,$3E,$E7,$BF,$E9,$8F,$D8,$FA,$37
	.dc.b	$77,$C8,$53,$11,$27,$E6,$BF,$30,$8E,$0F,$49,$D0,$B3,$FE,$37,$6F
	.dc.b	$C7,$FC,$4F,$80,$F8,$BF,$01,$27,$D3,$19,$69,$0F,$51,$F4,$3D,$FF
	.dc.b	$9B,$EE,$FC,$EF,$7A,$4F,$C2,$8C,$B4,$87,$2E,$0F,$CD,$FB,$DF,$5B
	.dc.b	$E2,$BF,$1B,$1F,$9E,$6F,$11,$A4,$AF,$9E,$F1,$7E,$8F,$B9,$F8,$0F
	.dc.b	$6B,$F0,$3E,$CC,$9F,$AB,$19,$69,$0F,$25,$BC,$9F,$C5,$E7,$FE,$4E
	.dc.b	$9F,$EC,$ED,$7F,$68,$DC,$8D,$3F,$CA,$FD,$6E,$6E,$3E,$4F,$43,$AF
	.dc.b	$93,$F8,$3F,$96,$71,$FE,$B1,$B9,$7E,$C1,$AF,$F9,$AD,$AF,$FB,$F1
	.dc.b	$CE,$1F,$DE,$1B,$87,$F3,$A6,$CF,$FB,$27,$E5,$FF,$9A,$3F,$88,$6B
	.dc.b	$F7,$84,$70,$6D,$FE,$66,$77,$7F,$E0,$66,$DF,$F4,$86,$D7,$7C,$3F
	.dc.b	$71,$4F,$C5,$3F,$BB,$3F,$E0,$EE,$F7,$C9,$FF,$0C,$6D,$FF,$10,$6B
	.dc.b	$FE,$8B,$6B,$FD,$58,$FE,$05,$1F,$AD,$8D,$CF,$58,$C7,$C4,$FE,$C7
	.dc.b	$E0,$3F,$5F,$3F,$EE,$BF,$98,$7B,$7F,$D8,$07,$BF,$FD,$8A,$33,$97
	.dc.b	$FB,$C6,$72,$F1,$7F,$B3,$DF,$48,$4F,$07,$FD,$43,$E0,$FF,$1F,$7F
	.dc.b	$B2,$57,$7B,$FE,$DF,$AE,$E3,$AC,$E8,$FF,$80,$18,$75,$9E,$8F,$7C
	.dc.b	$38,$5F,$9E,$FA,$5F,$94,$E3,$87,$86,$B8,$B6,$77,$3E,$86,$CE,$B1
	.dc.b	$E5,$3E,$A4,$F9,$4F,$93,$7E,$AA,$E7,$8F,$19,$D6,$3E,$79,$EA,$48
	.dc.b	$F2,$DF,$69,$D4,$7C,$FB,$F6,$34,$D6,$3E,$97,$89,$E9,$34,$53,$CA
	.dc.b	$70,$A8,$FA,$6F,$E7,$E1,$F2,$A3,$DA,$FA,$4F,$56,$39,$8E,$93,$F5
	.dc.b	$A3,$2F,$E0,$AF,$EF,$56,$C9,$19,$BE,$AD,$BF,$16,$1F,$25,$F4,$D7
	.dc.b	$86,$75,$C5,$FB,$6C,$BF,$1C,$DC,$7F,$18,$DC,$7E,$AB,$FB,$CC,$3E
	.dc.b	$BF,$99,$8C,$7F,$DE,$8D,$8F,$F4,$61,$FE,$F2,$8C,$B4,$CB,$F9,$BA
	.dc.b	$FE,$FD,$C3,$EA,$9C,$E5,$FD,$E7,$C8,$FE,$41,$1F,$85,$3C,$27,$E1
	.dc.b	$EB,$8B,$91,$87,$37,$FC,$09,$3C,$B0,$FD,$54,$72,$D7,$C5,$3D,$9C
	.dc.b	$C7,$23,$F0,$17,$DA,$33,$7D,$63,$C3,$78,$AF,$25,$F7,$C3,$E0,$E1
	.dc.b	$DB,$1F,$D0,$D7,$86,$7D,$D1,$88,$F0,$4E,$DF,$C8,$1B,$3F,$C4,$35
	.dc.b	$E7,$AE,$A7,$28,$D9,$C9,$C4,$F9,$6F,$A1,$63,$83,$E9,$8F,$01,$63
	.dc.b	$89,$DA,$7C,$37,$E7,$C7,$F7,$78,$FC,$5A,$FB,$F8,$76,$69,$D8,$E5
	.dc.b	$7F,$52,$F9,$3F,$3B,$1F,$63,$FC,$13,$F5,$47,$C0,$7E,$0B,$1F,$F6
	.dc.b	$5C,$62,$FF,$C2,$FF,$AA,$BE,$27,$51,$FE,$90,$9D,$61,$FD,$4D,$AF
	.dc.b	$F3,$F6,$BB,$C3,$FD,$AE,$D9,$FE,$88,$FF,$93,$7E,$91,$E0,$76,$8C
	.dc.b	$37,$27,$EA,$EF,$43,$17,$C3,$0F,$F2,$2D,$23,$18,$7F,$67,$F9,$B7
	.dc.b	$AF,$C3,$2E,$D2,$9C,$E1,$FE,$36,$6C,$71,$AE,$13,$97,$8C,$EA,$7F
	.dc.b	$AF,$35,$FF,$5A,$B5,$FE,$66,$DF,$03,$FC,$27,$F4,$1E,$F6,$DF,$F9
	.dc.b	$67,$C1,$7A,$79,$55,$4F,$97,$87,$F8,$79,$B7,$CA,$3F,$B8,$35,$FF
	.dc.b	$63,$BF,$E2,$2E,$DE,$87,$5F,$9D,$B8,$F8,$63,$C1,$CB,$8B,$FF,$22
	.dc.b	$63,$FC,$13,$7F,$E1,$38,$FF,$8B,$3D,$4F,$31,$E7,$EC,$B9,$7A,$E3
	.dc.b	$CB,$66,$92,$3F,$07,$5F,$E3,$6F,$43,$E0,$1F,$07,$CF,$B3,$F1,$63
	.dc.b	$E2,$7C,$CE,$33,$F2,$BF,$1F,$E5,$65,$7C,$BE,$AB,$5E,$A9,$F9,$9F
	.dc.b	$2B,$8D,$BD,$7A,$A3,$F0,$BF,$32,$F8,$FA,$63,$E4,$AE,$7F,$E7,$C9
	.dc.b	$D3,$FE,$B2,$6E,$5F,$9E,$6B,$FE,$23,$6B,$FE,$FA,$77,$8F,$F3,$48
	.dc.b	$7C,$6F,$F7,$AA,$ED,$87,$E6,$23,$13,$F5,$3E,$5B,$F0,$C7,$E1,$FC
	.dc.b	$9B,$7F,$8B,$8F,$C5,$7C,$AE,$2B,$ED,$BF,$8D,$F9,$59,$5F,$91,$D2
	.dc.b	$D7,$E0,$26,$E7,$F1,$A6,$D7,$F8,$3A,$BC,$FE,$5F,$94,$F3,$7C,$8D
	.dc.b	$9F,$F5,$0E,$9F,$E3,$87,$FD,$AE,$DA,$FE,$06,$D7,$FC,$90,$D7,$B3
	.dc.b	$BF,$27,$F4,$B1,$FE,$1E,$87,$C8,$FF,$5C,$9C,$CF,$92,$BF,$B6,$F7
	.dc.b	$F2,$AD,$3F,$E8,$8E,$0B,$5F,$EF,$BF,$DA,$BE,$6E,$1D,$BA,$BF,$D1
	.dc.b	$35,$A5,$3A,$BE,$58,$F6,$C3,$91,$BE,$5F,$C1,$E3,$CE,$F8,$4F,$7D
	.dc.b	$3A,$41,$BA,$E1,$98,$75,$CB,$C3,$AE,$30,$FE,$EB,$9F,$F9,$02,$73
	.dc.b	$5B,$C7,$F9,$BE,$5E,$1D,$7B,$F8,$FE,$06,$D6,$DB,$8F,$43,$E2,$38
	.dc.b	$AE,$53,$D6,$B9,$3D,$FC,$AF,$5F,$C9,$7B,$1B,$19,$3C,$91,$96,$9E
	.dc.b	$7F,$91,$F8,$67,$C4,$F9,$1B,$19,$26,$5E,$5D,$3E,$9C,$EA,$3D,$07
	.dc.b	$EA,$AB,$EE,$B6,$BE,$B3,$FA,$EB,$CE,$CA,$A7,$F8,$9A,$BF,$26,$DA
	.dc.b	$FC,$9B,$5F,$C9,$35,$EC,$78,$A5,$7D,$2B,$C3,$18,$7E,$B7,$C8,$FE
	.dc.b	$08,$FC,$1F,$C8,$D8,$FB,$5F,$23,$DF,$3F,$C8,$B7,$80,$D3,$1F,$23
	.dc.b	$F8,$C3,$F1,$BF,$23,$6F,$F0,$9F,$23,$F9,$03,$F2,$3F,$23,$63,$F1
	.dc.b	$1F,$23,$D4,$8F,$F1,$0D,$FF,$4D,$1E,$1F,$91,$FC,$C9,$F9,$AF,$42
	.dc.b	$D7,$E4,$7E,$43,$F3,$A7,$E7,$BE,$46,$C6,$4E,$9B,$FE,$B2,$7D,$84
	.dc.b	$8C,$EE,$8F,$D9,$7E,$B5,$FE,$F9,$9E,$59,$FE,$0E,$30,$DF,$92,$D9
	.dc.b	$C3,$E6,$2D,$7E,$AE,$93,$F3,$5B,$1E,$46,$D7,$F3,$5C,$51,$89,$FE
	.dc.b	$DA,$EB,$AF,$F8,$43,$FE,$17,$8E,$D7,$E8,$F6,$3F,$1F,$CB,$27,$F0
	.dc.b	$7C,$D7,$2F,$C4,$4F,$E2,$F9,$CD,$25,$7D,$39,$FE,$8F,$1F,$9E,$F9
	.dc.b	$5F,$F1,$37,$81,$F8,$E8,$C6,$3F,$C3,$6B,$AD,$FE,$8A,$F7,$18,$3F
	.dc.b	$52,$E3,$4F,$E5,$6D,$F0,$3F,$A0,$36,$7F,$99,$61,$DB,$1E,$98,$F7
	.dc.b	$C3,$8B,$70,$FE,$77,$5B,$BE,$F3,$97,$FD,$B3,$87,$57,$DB,$8F,$4C
	.dc.b	$64,$70,$7D,$D1,$8C,$7B,$C5,$C7,$BC,$DA,$EB,$87,$75,$87,$7C,$BE
	.dc.b	$7D,$6E,$73,$5B,$AD,$65,$E9,$CB,$BD,$78,$CF,$94,$E0,$98,$FC,$8B
	.dc.b	$F0,$63,$DA,$B8,$56,$9F,$2E,$5E,$C7,$37,$C6,$3C,$95,$C6,$B8,$57
	.dc.b	$9A,$73,$87,$96,$F8,$4F,$23,$EB,$E5,$ED,$CB,$E7,$1E,$C7,$19,$7B
	.dc.b	$CF,$3D,$F3,$AF,$29,$CC,$FA,$A5,$AF,$C5,$BF,$E0,$53,$C8,$7F,$A0
	.dc.b	$1C,$BE,$43,$5F,$00,$F4,$1F,$2D,$FE,$86,$94,$E6,$30,$D8,$FF,$45
	.dc.b	$B5,$F0,$53,$E5,$5F,$05,$CE,$30,$D8,$CC,$F9,$5A,$F2,$73,$B7,$E8
	.dc.b	$9B,$AF,$E8,$9B,$3F,$98,$F8,$AF,$F1,$25,$6E,$FF,$5D,$5A,$AF,$C7
	.dc.b	$B7,$05,$F0,$C3,$A5,$F0,$25,$FE,$D2,$74,$FE,$9A,$F9,$12,$7F,$79
	.dc.b	$3A,$9E,$F1,$CB,$0F,$24,$73,$31,$7E,$07,$F8,$0D,$F9,$6F,$96,$1F
	.dc.b	$22,$73,$87,$07,$E0,$14,$EE,$68,$EC,$5C,$3C,$24,$C3,$F8,$5D,$F6
	.dc.b	$C3,$D5,$5B,$9C,$BD,$70,$F1,$5F,$18,$F0,$9B,$3E,$26,$1B,$C3,$A3
	.dc.b	$C7,$0F,$64,$D8,$D3,$4E,$47,$F0,$49,$E2,$E2,$FC,$38,$79,$23,$DB
	.dc.b	$87,$8D,$F8,$2F,$F2,$8A,$F9,$27,$27,$55,$F8,$09,$FA,$E7,$86,$38
	.dc.b	$38,$3A,$D7,$78,$FE,$67,$F1,$7F,$82,$E5,$7C,$B7,$EB,$E7,$FC,$B0
	.dc.b	$FF,$3F,$BF,$F1,$33,$63,$31,$C6,$38,$5F,$F9,$25,$D3,$78,$B5,$30
	.dc.b	$FF,$29,$4F,$05,$7E,$27,$A7,$93,$D7,$93,$B7,$91,$27,$8C,$65,$E7
	.dc.b	$19,$7C,$F1,$96,$9C,$0F,$0F,$93,$E9,$F2,$64,$79,$1F,$67,$93,$ED
	.dc.b	$F2,$7E,$37,$67,$F0,$3A,$FF,$00,$72,$65,$F5,$F9,$3F,$0C,$8F,$BB
	.dc.b	$B3,$E3,$75,$48,$F8,$2F,$C9,$F2,$7E,$57,$93,$DB,$E1,$A5,$7C,$68
	.dc.b	$F9,$DE,$54,$8F,$92,$FD,$1F,$2B,$F6,$DF,$25,$FC,$43,$A7,$FB,$7F
	.dc.b	$C6,$7E,$C0,$CF,$74,$C3,$FB,$C4,$65,$FF,$2D,$D7,$1E,$BF,$F5,$A3
	.dc.b	$FC,$6B,$C3,$FE,$64,$D9,$CD,$FF,$AE,$67,$2D,$8F,$C1,$79,$9F,$85
	.dc.b	$23,$F8,$0F,$32,$BF,$8C,$8B,$97,$E4,$27,$F1,$9E,$67,$E3,$7C,$C9
	.dc.b	$7F,$88,$8C,$BF,$8A,$8C,$BF,$96,$8F,$CA,$79,$5F,$95,$F3,$24,$FE
	.dc.b	$42,$32,$FE,$4A,$32,$FE,$7A,$32,$FE,$82,$33,$87,$E9,$27,$F5,$1F
	.dc.b	$1B,$FE,$EF,$8F,$F5,$F3,$FE,$42,$AF,$F5,$73,$5F,$F4,$73,$BD,$BF
	.dc.b	$D5,$CF,$FB,$DF,$5F,$F7,$17,$82,$8C,$56,$A7,$E6,$3F,$A6,$8C,$B7
	.dc.b	$CA,$F1,$F5,$48,$F0,$C7,$98,$C3,$E2,$8C,$3F,$80,$F7,$BF,$13,$8C
	.dc.b	$8C,$10,$FF,$51,$F1,$7F,$B9,$33,$FF,$72,$B6,$7F,$0F,$5F,$92,$3F
	.dc.b	$B8,$52,$51,$86,$C7,$F7,$56,$BF,$DF,$F3,$F8,$5C,$E8,$9E,$93,$5F
	.dc.b	$93,$39,$94,$E5,$F8,$D1,$98,$C5,$7C,$DF,$96,$C5,$F2,$BE,$AD,$79
	.dc.b	$B6,$B8,$EC,$AE,$35,$F4,$FE,$56,$E1,$F6,$7E,$5A,$D9,$B3,$2B,$FC
	.dc.b	$38,$71,$2D,$7E,$06,$3F,$C3,$0D,$8F,$8D,$E7,$F8,$1F,$99,$F0,$37
	.dc.b	$53,$66,$4F,$B0,$F9,$53,$A7,$07,$22,$BF,$E2,$96,$C7,$D7,$23,$E2
	.dc.b	$3C,$F8,$BE,$3F,$DF,$F2,$B2,$7C,$18,$FE,$6E,$7E,$19,$EC,$AE,$13
	.dc.b	$CA,$30,$67,$D6,$F9,$3F,$29,$27,$9D,$27,$E6,$9B,$8F,$1B,$5F,$C5
	.dc.b	$43,$E4,$FE,$AF,$8F,$F6,$67,$CA,$D7,$EF,$9F,$AD,$8F,$B2,$79,$C8
	.dc.b	$FC,$07,$D5,$E9,$FD,$EF,$4A,$4F,$9A,$3F,$02,$61,$EB,$9F,$EF,$67
	.dc.b	$31,$F1,$B1,$FF,$23,$F1,$FE,$5B,$17,$0F,$A0,$FC,$6C,$EB,$93,$F4
	.dc.b	$98,$FC,$3F,$07,$F0,$F3,$FE,$50,$C7,$F8,$5B,$5D,$B5,$FD,$67,$C9
	.dc.b	$FD,$27,$E0,$8F,$93,$75,$7E,$3D,$B1,$F8,$89,$E1,$A7,$EB,$E1,$1C
	.dc.b	$DF,$F3,$C3,$5F,$DB,$35,$F0,$F1,$7F,$06,$A7,$C8,$FE,$76,$BA,$E9
	.dc.b	$C2,$BC,$53,$2F,$C5,$1B,$30,$FA,$BA,$9F,$13,$FA,$7F,$B6,$71,$53
	.dc.b	$F5,$53,$AA,$FD,$1B,$5F,$DA,$B5,$FF,$49,$35,$FE,$20,$D7,$F8,$B7
	.dc.b	$B3,$FB,$2D,$7F,$8C,$7C,$9B,$D3,$FE,$2C,$3D,$7F,$E3,$11,$9B,$FE
	.dc.b	$3D,$39,$78,$BF,$C7,$6F,$A4,$27,$83,$FE,$9C,$EE,$FE,$42,$E6,$BB
	.dc.b	$DF,$66,$C7,$43,$CF,$3C,$2B,$9D,$F6,$DB,$AD,$F8,$EF,$EA,$BC,$9A
	.dc.b	$72,$C7,$F9,$B4,$F2,$C3,$A5,$6F,$0E,$78,$F8,$F0,$E0,$79,$CB,$1A
	.dc.b	$7C,$C7,$C9,$8E,$0F,$BF,$0C,$E3,$F1,$2B,$84,$F3,$38,$1A,$C3,$C6
	.dc.b	$F6,$C3,$C8,$47,$04,$38,$9C,$DF,$85,$3E,$DB,$E8,$79,$A3,$27,$39
	.dc.b	$F3,$1E,$3B,$F3,$46,$4D,$C7,$78,$CB,$F8,$2C,$3D,$F3,$A8,$E7,$7E
	.dc.b	$17,$C9,$3A,$3E,$2D,$72,$8F,$0F,$27,$EB,$1F,$9E,$D7,$39,$7E,$37
	.dc.b	$2F,$C4,$CE,$67,$FB,$87,$85,$FD,$C6,$31,$08,$C5,$7F,$5A,$6C,$7F
	.dc.b	$A2,$4F,$F7,$3C,$65,$FA,$99,$FF,$80,$B8,$7E,$EC,$76,$33,$E1,$7F
	.dc.b	$79,$D3,$F3,$B1,$97,$F3,$F5,$C6,$3E,$F1,$9D,$7F,$BF,$57,$17,$C1
	.dc.b	$5C,$23,$13,$E5,$8C,$5F,$8A,$3D,$87,$89,$C6,$3F,$92,$C3,$C1,$5E
	.dc.b	$4A,$DB,$C9,$F1,$98,$C3,$C7,$7C,$A3,$A1,$AB,$E9,$87,$92,$3F,$DF
	.dc.b	$E6,$AF,$B4,$EA,$B9,$47,$73,$33,$BB,$F7,$9E,$07,$B1,$93,$EA,$63
	.dc.b	$E7,$38,$96,$38,$C7,$F8,$BB,$0F,$23,$F3,$4C,$5F,$8A,$7B,$5F,$88
	.dc.b	$F2,$3F,$E3,$3F,$92,$7B,$1F,$D1,$6F,$0D,$7F,$C7,$91,$AB,$FF,$11
	.dc.b	$C2,$30,$FF,$89,$9B,$9F,$F4,$F5,$E9,$FF,$8F,$E3,$17,$6F,$C5,$36
	.dc.b	$FF,$C8,$BF,$22,$F9,$1C,$43,$E4,$5E,$76,$F2,$5E,$F1,$FF,$13,$D6
	.dc.b	$9F,$F1,$64,$EB,$0E,$73,$9C,$3F,$C1,$73,$9A,$E3,$86,$8F,$85,$3E
	.dc.b	$16,$D7,$18,$FF,$2E,$19,$9F,$8B,$1D,$6B,$E0,$7C,$F3,$E2,$FA,$57
	.dc.b	$92,$30,$FF,$83,$7E,$4B,$E8,$78,$3F,$E2,$96,$F8,$26,$92,$7F,$C4
	.dc.b	$BE,$5F,$C9,$AE,$11,$83,$FC,$31,$DD,$9B,$FD,$A6,$5F,$E6,$E7,$FB
	.dc.b	$8C,$FF,$87,$79,$75,$5B,$6C,$64,$FF,$11,$7C,$8F,$90,$8F,$F9,$92
	.dc.b	$91,$8D,$BF,$C5,$7E,$2B,$E6,$3C,$53,$E7,$8C,$5F,$F8,$B7,$CC,$F7
	.dc.b	$91,$E8,$DB,$FF,$1A,$E0,$F9,$1F,$64,$FB,$5F,$77,$C8,$7C,$82,$5F
	.dc.b	$95,$B7,$CB,$8B,$73,$FE,$3D,$7F,$1F,$F2,$0F,$8B,$FC,$0B,$67,$F4
	.dc.b	$2D,$AF,$F4,$2B,$5F,$C1,$B9,$E0,$FF,$18,$7C,$27,$FA,$96,$BF,$CF
	.dc.b	$73,$97,$FD,$49,$19,$AF,$F1,$EC,$E5,$FF,$56,$46,$6B,$FC,$75,$39
	.dc.b	$94,$E6,$BF,$C9,$3F,$1F,$F8,$82,$3F,$55,$B3,$F2,$BE,$3F,$F0,$91
	.dc.b	$F8,$57,$F0,$DF,$1F,$2B,$FC,$C1,$F1,$F6,$FE,$8C,$DC,$BF,$15,$36
	.dc.b	$BD,$7E,$47,$EA,$32,$FF,$15,$47,$F0,$86,$E1,$FE,$AC,$6B,$FA,$A7
	.dc.b	$3E,$2F,$F9,$6A,$3D,$BF,$11,$FE,$C7,$8C,$4E,$BD,$0F,$93,$39,$7C
	.dc.b	$37,$FD,$91,$CB,$6F,$E5,$E1,$B7,$E6,$D5,$EA,$FF,$4C,$70,$5F,$8F
	.dc.b	$FA,$99,$FF,$9B,$27,$38,$F3,$53,$AC,$BF,$CF,$37,$C7,$1E,$9C,$DF
	.dc.b	$87,$C7,$95,$EE,$58,$72,$F8,$BF,$F7,$26,$5F,$EE,$78,$C3,$F8,$59
	.dc.b	$DD,$2B,$71,$FE,$8E,$8C,$4E,$EB,$8C,$FF,$80,$F2,$D6,$BE,$DC,$3D
	.dc.b	$CF,$BF,$C5,$F8,$1A,$DC,$7D,$D9,$23,$18,$F3,$3C,$6E,$33,$F6,$B7
	.dc.b	$2F,$23,$67,$77,$D3,$C7,$F9,$DF,$15,$F3,$DB,$19,$3C,$31,$96,$9E
	.dc.b	$4F,$8A,$FA,$C7,$D7,$F8,$AB,$19,$25,$F9,$F7,$FA,$F3,$A8,$F5,$1F
	.dc.b	$AA,$AF,$96,$FE,$C1,$AE,$6B,$E9,$34,$FE,$D1,$B2,$7F,$19,$1F,$B4
	.dc.b	$AB,$39,$AF,$C9,$4E,$5B,$19,$8F,$1A,$9E,$35,$C9,$FC,$A7,$93,$2F
	.dc.b	$EA,$FC,$55,$C3,$DD,$BB,$E2,$BF,$39,$E0,$7F,$C0,$1B,$FB,$71,$8F
	.dc.b	$57,$F3,$DF,$15,$27,$EF,$FC,$55,$C7,$9F,$C5,$7E,$8C,$FD,$27,$C4
	.dc.b	$DA,$FC,$2F,$C4,$F6,$F8,$2F,$D4,$36,$3F,$13,$F1,$3F,$AB,$3F,$59
	.dc.b	$E8,$D8,$FC,$6F,$C4,$7E,$BC,$FD,$87,$C4,$D8,$C9,$FA,$FE,$03,$9D
	.dc.b	$3E,$E4,$62,$37,$18,$75,$D9,$FB,$99,$D4,$F0,$F8,$EF,$8A,$7C,$6F
	.dc.b	$8F,$B7,$DE,$3E,$4F,$C5,$E2,$BB,$B7,$0E,$8D,$EF,$FF,$7F,$E2,$8F
	.dc.b	$C0,$F3,$CC,$7E,$FF,$45,$C7,$EF,$CF,$E2,$3E,$3B,$87,$85,$F8,$48
	.dc.b	$C6,$1F,$BF,$AD,$E1,$F8,$AB,$DC,$60,$FC,$DB,$8C,$FF,$93,$DF,$2C
	.dc.b	$3A,$5F,$1C,$3A,$57,$F2,$FB,$E1,$9B,$19,$18,$7F,$A9,$E3,$CD,$EF
	.dc.b	$7C,$AB,$84,$2B,$83,$FD,$5A,$B8,$5F,$7B,$E1,$3D,$9A,$EF,$0F,$36
	.dc.b	$32,$3B,$BE,$A6,$92,$BA,$46,$5B,$19,$C7,$CB,$8F,$93,$B5,$97,$92
	.dc.b	$73,$97,$8A,$B5,$39,$BF,$36,$3E,$3C,$7E,$6E,$57,$1E,$33,$6B,$E3
	.dc.b	$CD,$7E,$26,$3D,$4F,$7E,$5E,$29,$D6,$5F,$20,$F0,$B8,$9E,$17,$ED
	.dc.b	$C3,$BE,$1E,$D3,$CA,$7E,$0F,$1F,$82,$48,$DB,$E8,$8E,$07,$92,$7C
	.dc.b	$D8,$F8,$E3,$38,$F8,$F0,$F3,$E5,$F0,$E7,$81,$FE,$62,$29,$FE,$64
	.dc.b	$9F,$CA,$B6,$7F,$34,$FF,$9A,$3D,$F7,$A0,$F8,$FF,$F4,$6C,$61,$B1
	.dc.b	$99,$ED,$39,$7C,$16,$F8,$D7,$C0,$7F,$A5,$A3,$0D,$8C,$CF,$06,$BE
	.dc.b	$77,$3A,$FE,$8D,$BB,$7E,$21,$B3,$FA,$4F,$88,$FC,$FD,$6E,$FF,$69
	.dc.b	$5A,$AF,$D1,$37,$05,$F0,$C3,$A5,$F0,$25,$FE,$EA,$74,$FE,$CA,$F9
	.dc.b	$12,$7F,$81,$4E,$A7,$BC,$72,$C3,$C9,$1C,$CC,$5F,$81,$FE,$0F,$7E
	.dc.b	$5B,$E5,$87,$C8,$9C,$E1,$C1,$F8,$05,$3B,$9A,$3B,$17,$0F,$09,$30
	.dc.b	$FD,$8D,$F6,$C3,$D5,$5B,$9C,$BD,$70,$F1,$5F,$18,$F0,$9B,$3E,$26
	.dc.b	$1B,$C3,$A3,$C7,$0F,$64,$D8,$D3,$4E,$47,$F1,$79,$E2,$E2,$FC,$38
	.dc.b	$79,$23,$17,$E8,$C3,$F9,$45,$FC,$83,$8B,$E7,$9E,$87,$D6,$3C,$0F
	.dc.b	$07,$07,$C8,$70,$FF,$3A,$F8,$4F,$C0,$F8,$AF,$8F,$FF,$67,$D7,$F9
	.dc.b	$4D,$FD,$5B,$78,$B7,$1A,$CB,$F9,$D5,$71,$C3,$35,$1F,$F2,$D5,$6D
	.dc.b	$FF,$29,$B6,$78,$46,$8E,$BE,$6F,$6E,$3E,$FE,$5C,$9E,$71,$97,$AC
	.dc.b	$65,$FF,$19,$5F,$FA,$61,$D9,$0C,$CF,$AB,$CB,$F5,$F9,$72,$7C,$EF
	.dc.b	$BB,$CB,$F7,$F9,$7F,$1B,$AF,$F8,$15,$7F,$81,$F8,$B2,$FB,$BC,$BF
	.dc.b	$8A,$47,$E0,$F6,$7C,$8E,$B9,$1F,$11,$F9,$7E,$57,$CC,$F2,$7B,$7C
	.dc.b	$2C,$AF,$93,$1F,$43,$CA,$91,$F3,$1F,$A7,$E5,$7E,$B3,$E3,$E4,$FF
	.dc.b	$79,$7F,$71,$F1,$DF,$C4,$BA,$BF,$D4,$0D,$EA,$FE,$AF,$3C,$2B,$FC
	.dc.b	$AF,$39,$7F,$BB,$35,$FE,$E1,$1B,$AE,$7D,$DF,$84,$F3,$3E,$37,$CC
	.dc.b	$4B,$FC,$0C,$65,$FF,$17,$46,$5F,$AF,$1F,$8C,$F2,$A4,$7E,$21,$FC
	.dc.b	$7F,$97,$88,$FB,$B1,$70,$FC,$B4,$FE,$53,$CA,$FA,$DE,$84,$BF,$C8
	.dc.b	$C6,$5A,$43,$F0,$FE,$66,$5F,$C4,$C5,$BF,$D1,$4F,$E9,$BE,$39,$E6
	.dc.b	$52,$5F,$FA,$DE,$3F,$C9,$2D,$8C,$BF,$E8,$58,$E2,$EF,$CC,$E9,$D5
	.dc.b	$F8,$DF,$8F,$93,$CE,$32,$FF,$95,$63,$2F,$38,$FB,$BC,$7F,$7B,$E0
	.dc.b	$A4,$FF,$A0,$B9,$3F,$DB,$DA,$7E,$02,$B7,$39,$7F,$0B,$F0,$18,$8F
	.dc.b	$C0,$C5,$9F,$C3,$CF,$E8,$7E,$63,$FC,$63,$EC,$7B,$BC,$97,$82,$9C
	.dc.b	$ED,$FE,$56,$6E,$5F,$ED,$2D,$DD,$B8,$C3,$8D,$5C,$39,$5F,$F9,$17
	.dc.b	$1E,$D1,$88,$47,$87,$0E,$71,$BC,$7F,$89,$C3,$E0,$FA,$7C,$84,$AD
	.dc.b	$1E,$38,$F1,$4E,$E3,$39,$7B,$2B,$84,$78,$2F,$AD,$FB,$21,$F3,$BF
	.dc.b	$C3,$DD,$1F,$EF,$0C,$9F,$11,$ED,$38,$4E,$F3,$FC,$3D,$72,$F2,$B2
	.dc.b	$47,$CE,$38,$9F,$1B,$98,$EE,$71,$37,$7D,$73,$FF,$33,$D7,$09,$F6
	.dc.b	$EC,$F6,$7E,$87,$A5,$27,$31,$F1,$0C,$3E,$CE,$0F,$BF,$39,$8E,$7D
	.dc.b	$1F,$67,$D2,$91,$82,$1C,$78,$11,$A8,$E9,$8A,$BF,$05,$E1,$C9,$FA
	.dc.b	$76,$8F,$C3,$7C,$96,$67,$A4,$5E,$1F,$C3,$CF,$DC,$C6,$57,$C0,$7E
	.dc.b	$FE,$58,$8F,$0C,$5A,$FC,$64,$FE,$57,$F6,$BF,$F1,$CC,$7E,$7D,$BB
	.dc.b	$FF,$60,$6D,$FD,$FF,$93,$E1,$E3,$48,$FC,$5C,$EB,$7F,$D3,$D7,$E9
	.dc.b	$30,$78,$2F,$82,$C3,$70,$FC,$BC,$EA,$7F,$21,$F2,$6E,$FD,$C4,$AD
	.dc.b	$DF,$17,$F0,$B5,$B7,$F1,$7F,$26,$F8,$0F,$F3,$64,$7F,$AC,$E3,$0E
	.dc.b	$1A,$FF,$9C,$5A,$FF,$AC,$83,$AC,$7F,$36,$DD,$BF,$D6,$ED,$9F,$F6
	.dc.b	$05,$BE,$4B,$F5,$5F,$26,$F2,$29,$2B,$F4,$31,$F8,$86,$C6,$5F,$D0
	.dc.b	$C7,$17,$7C,$7F,$B5,$F7,$7F,$4F,$9F,$FB,$3F,$E4,$DE,$C7,$FB,$28
	.dc.b	$7B,$9F,$EC,$C8,$CE,$1F,$ED,$59,$CB,$C5,$FF,$68,$DF,$48,$4F,$05
	.dc.b	$19,$8C,$4F,$29,$DB,$E1,$AF,$03,$D3,$C0,$FE,$3B,$3A,$3F,$DC,$D7
	.dc.b	$CE,$78,$47,$8C,$F3,$57,$27,$75,$CA,$FC,$97,$C6,$76,$71,$3E,$1D
	.dc.b	$71,$7D,$35,$C6,$FC,$98,$6D,$F5,$61,$EC,$BE,$18,$75,$C3,$8D,$EE
	.dc.b	$7A,$5F,$03,$9C,$FB,$4D,$95,$D3,$E6,$35,$5D,$4E,$D1,$A8,$F4,$9E
	.dc.b	$B8,$D6,$1E,$98,$E5,$87,$9D,$D4,$65,$F0,$C6,$5F,$15,$FB,$27,$55
	.dc.b	$E0,$8F,$59,$9C,$39,$9E,$FA,$87,$A2,$B8,$9F,$73,$9F,$F4,$11,$F5
	.dc.b	$A4,$FF,$B3,$35,$F8,$B3,$A8,$F8,$AD,$77,$E4,$70,$8C,$42,$31,$3F
	.dc.b	$EE,$76,$C7,$F9,$F0,$EC,$5C,$BF,$31,$6F,$23,$FC,$33,$1F,$9B,$9E
	.dc.b	$E7,$D7,$9D,$71,$7E,$26,$33,$3C,$AB,$8E,$BE,$19,$E1,$18,$7C,$11
	.dc.b	$FE,$0A,$8C,$47,$FB,$78,$DD,$75,$7D,$86,$0F,$C4,$1E,$09,$E5,$1B
	.dc.b	$C3,$F3,$98,$79,$AB,$67,$2C,$3C,$D5,$B8,$E9,$1E,$F7,$57,$D1,$E3
	.dc.b	$7D,$30,$F3,$C7,$84,$FF,$0B,$38,$78,$98,$BE,$D7,$E0,$8E,$67,$82
	.dc.b	$FE,$34,$64,$F1,$5F,$7A,$E3,$3E,$39,$CD,$71,$3E,$59,$DC,$EA,$F3
	.dc.b	$8E,$A7,$13,$F1,$F0,$F9,$37,$A1,$FE,$58,$CB,$FC,$27,$18,$C3,$FC
	.dc.b	$22,$BD,$7D,$47,$28,$D5,$FF,$84,$9B,$5C,$AF,$73,$F9,$E9,$E5,$5D
	.dc.b	$27,$32,$F9,$27,$A5,$86,$D7,$F6,$2F,$93,$FF,$15,$BF,$E2,$48,$CD
	.dc.b	$7F,$88,$9A,$FF,$55,$6B,$F0,$1A,$FE,$C1,$B8,$BF,$15,$EB,$3F,$E3
	.dc.b	$A8,$C7,$97,$EA,$5F,$28,$F8,$FC,$5F,$09,$F9,$4D,$C7,$FC,$B4,$DC
	.dc.b	$3F,$CB,$37,$C6,$7F,$BE,$CE,$7C,$27,$87,$FD,$96,$7F,$C1,$3C,$BF
	.dc.b	$2A,$7F,$CD,$FE,$75,$AD,$63,$FE,$35,$6E,$5F,$E1,$7B,$E1,$E4,$7F
	.dc.b	$98,$F1,$E6,$7F,$97,$BB,$BF,$4D,$5F,$E7,$B6,$C7,$FB,$F1,$7C,$A7
	.dc.b	$85,$CE,$2C,$FF,$BC,$74,$E4,$D5,$5B,$FF,$10,$FA,$7E,$73,$C7,$F2
	.dc.b	$96,$32,$4C,$3F,$C8,$D1,$97,$FC,$A1,$8B,$E4,$FE,$07,$91,$E3,$6D
	.dc.b	$FA,$DB,$5C,$1B,$A7,$B7,$8F,$E4,$67,$EB,$6D,$FC,$36,$DF,$81,$B5
	.dc.b	$FE,$76,$FA,$07,$CA,$FE,$61,$B7,$FE,$76,$F5,$3F,$9D,$BF,$E7,$88
	.dc.b	$CE,$1F,$E7,$66,$DF,$3B,$FD,$0D,$F0,$AD,$47,$53,$EF,$7C,$CD,$C3
	.dc.b	$FC,$96,$FD,$3F,$33,$F0,$5C,$36,$7F,$C9,$ED,$BF,$C2,$37,$1F,$F2
	.dc.b	$D3,$F8,$9C,$25,$7E,$0F,$E4,$ED,$FE,$2E,$6D,$FE,$16,$6D,$7F,$0A
	.dc.b	$C9,$F0,$FF,$EE,$0F,$13,$FC,$9D,$8A,$73,$5F,$D6,$38,$78,$78,$FB
	.dc.b	$79,$61,$6F,$FB,$1D,$F0,$7F,$AE,$4E,$AF,$94,$6A,$F8,$3D,$EE,$DF
	.dc.b	$1C,$3C,$6F,$F9,$02,$F8,$D6,$DB,$19,$8E,$E7,$E7,$7E,$47,$EA,$57
	.dc.b	$F9,$A7,$9F,$F6,$BE,$47,$9D,$4E,$B3,$FF,$42,$4E,$63,$17,$FA,$7E
	.dc.b	$BF,$F5,$CE,$5E,$AC,$38,$CF,$39,$CB,$E2,$8E,$B9,$73,$3A,$7C,$5D
	.dc.b	$BE,$58,$7C,$F6,$CE,$9A,$6B,$4F,$21,$E5,$F8,$BB,$19,$26,$3E,$1B
	.dc.b	$4E,$A3,$15,$E7,$F8,$DC,$6D,$F1,$74,$47,$C1,$F8,$B9,$97,$9B,$CA
	.dc.b	$B8,$F5,$9B,$87,$CD,$CF,$E9,$36,$3E,$57,$C5,$FC,$5F,$99,$B8,$FC
	.dc.b	$9C,$D1,$8C,$7E,$44,$62,$7E,$2D,$71,$7E,$FB,$F8,$6F,$8C,$FC,$3B
	.dc.b	$63,$27,$E1,$23,$2D,$25,$7E,$18,$FC,$F4,$69,$C6,$E8,$FC,$17,$9B
	.dc.b	$2B,$EB,$FC,$65,$BF,$9D,$F1,$9F,$8B,$F1,$7F,$D6,$95,$F7,$23,$1E
	.dc.b	$BF,$E2,$FE,$32,$4F,$E0,$3E,$32,$E1,$F5,$FE,$33,$F2,$07,$E4,$7D
	.dc.b	$2B,$5F,$86,$F9,$2E,$DF,$07,$F9,$56,$C7,$E2,$BE,$33,$F3,$07,$E6
	.dc.b	$7D,$5B,$1F,$8E,$F8,$CF,$CE,$1F,$9D,$F7,$EC,$64,$FC,$E7,$01,$CE
	.dc.b	$9F,$9B,$8C,$46,$E3,$0E,$BA,$7F,$5D,$3A,$9E,$1F,$19,$F1,$4F,$8D
	.dc.b	$F2,$56,$FB,$C7,$C9,$F9,$2C,$57,$76,$E1,$D1,$BE,$2F,$F9,$C3,$C4
	.dc.b	$64,$9F,$E1,$3C,$F3,$2F,$E0,$BD,$17,$2F,$BB,$3F,$C4,$3E,$47,$87
	.dc.b	$8B,$FC,$2A,$31,$82,$31,$87,$F1,$0A,$DD,$F2,$8C,$56,$E3,$F1,$93
	.dc.b	$AD,$3F,$95,$B7,$4F,$E6,$18,$F6,$C7,$89,$2F,$A1,$CB,$0E,$98,$72
	.dc.b	$6D,$FF,$90,$B0,$E9,$5F,$CC,$30,$E8,$F1,$C3,$A6,$1F,$CA,$A7,$2F
	.dc.b	$38,$CD,$F9,$31,$B7,$FE,$9C,$C3,$AA,$6F,$63,$1E,$77,$E3,$CB,$CB
	.dc.b	$7C,$5A,$FF,$4B,$9D,$B9,$C7,$BB,$9C,$6E,$3E,$03,$BD,$70,$9F,$15
	.dc.b	$F4,$9C,$D7,$08,$C1,$E0,$C7,$CD,$19,$C3,$BC,$62,$FE,$31,$32,$F8
	.dc.b	$8F,$D2,$69,$32,$F7,$57,$E1,$27,$84,$61,$A7,$D3,$BF,$03,$83,$DC
	.dc.b	$5C,$7D,$D8,$F9,$F2,$F7,$47,$E3,$A7,$FC,$FE,$7C,$0C,$2B,$F2,$4B
	.dc.b	$1B,$7F,$D0,$BE,$FF,$F3,$AF,$92,$DC,$23,$0F,$26,$BD,$9A,$FE,$37
	.dc.b	$E4,$1F,$03,$CA,$D1,$89,$F0,$B5,$F3,$C6,$A7,$84,$EF,$5F,$D1,$B7
	.dc.b	$5E,$D3,$FA,$DF,$87,$FF,$4E,$D6,$A3,$F6,$15,$BC,$3F,$28,$D7,$F2
	.dc.b	$95,$C2,$F9,$C7,$EC,$A7,$58,$78,$E3,$F6,$B5,$B7,$9D,$F6,$23,$A2
	.dc.b	$E1,$E3,$7A,$47,$9D,$CE,$1E,$09,$D1,$FC,$02,$BB,$4F,$EA,$70,$F1
	.dc.b	$92,$F9,$47,$42,$57,$2B,$FD,$F5,$F2,$C3,$D5,$39,$AF,$49,$E7,$38
	.dc.b	$D7,$8C,$8F,$EE,$EB,$96,$1E,$18,$D9,$93,$D6,$74,$AE,$CF,$CE,$32
	.dc.b	$7A,$F0,$B8,$2F,$AC,$62,$FE,$8C,$62,$BB,$9C,$4F,$45,$F6,$BF,$6E
	.dc.b	$1E,$89,$F0,$1C,$23,$2F,$96,$3E,$2B,$F2,$A7,$95,$F8,$23,$13,$CE
	.dc.b	$B8,$C7,$4A,$FA,$CF,$F9,$E4,$F8,$47,$F6,$13,$A9,$4C,$ED,$FE,$F2
	.dc.b	$E2,$7C,$97,$ED,$AB,$FC,$9C,$FF,$43,$AF,$F2,$13,$63,$31,$F9,$19
	.dc.b	$E2,$D9,$D6,$1F,$D7,$5B,$1F,$EB,$77,$B4,$66,$38,$10,$FF,$2A,$43
	.dc.b	$CD,$E9,$C5,$D7,$CC,$95,$A8,$CB,$C2,$32,$FA,$23,$2D,$38,$1D,$BC
	.dc.b	$CF,$47,$97,$23,$C2,$FA,$FC,$AF,$67,$91,$EF,$E9,$FE,$13,$5F,$E0
	.dc.b	$EE,$1C,$BE,$8F,$23,$E1,$11,$F5,$F5,$7C,$5E,$99,$1E,$E7,$E4,$79
	.dc.b	$1F,$27,$C7,$ED,$F2,$32,$BE,$1C,$7C,$DF,$22,$47,$C6,$7E,$87,$91
	.dc.b	$FA,$DF,$8B,$93,$FD,$FD,$FD,$CF,$C7,$7F,$1D,$E9,$FF,$52,$B7,$A3
	.dc.b	$FB,$94,$F0,$AF,$F2,$64,$E5,$07,$0F,$F6,$B8,$E1,$5D,$3B,$7F,$05
	.dc.b	$E5,$7C,$7F,$92,$97,$F7,$23,$2F,$F8,$DA,$32,$FE,$3A,$3F,$15,$E4
	.dc.b	$7E,$37,$CD,$FF,$78,$4F,$F9,$BF,$DB,$CB,$F8,$69,$AF,$E5,$9F,$C9
	.dc.b	$79,$7F,$93,$F2,$BE,$C7,$B9,$2B,$F1,$B1,$97,$F5,$1E,$17,$E8,$A3
	.dc.b	$DF,$1F,$A7,$8C,$3A,$F0,$93,$AF,$0F,$F5,$51,$89,$EA,$43,$F5,$3E
	.dc.b	$67,$E9,$FE,$3D,$E6,$52,$4F,$FA,$CE,$3F,$6D,$C5,$FE,$BF,$77,$19
	.dc.b	$9D,$BF,$EB,$58,$DF,$2F,$F9,$2D,$FC,$AC,$7F,$97,$FE,$3F,$F3,$07
	.dc.b	$87,$BE,$47,$F9,$9A,$32,$F8,$63,$F0,$F5,$F8,$8F,$80,$93,$FE,$D8
	.dc.b	$E8,$FF,$76,$E5,$F8,$AA,$F8,$53,$F2,$27,$FD,$E3,$AF,$F8,$01,$B1
	.dc.b	$99,$DB,$F8,$C8,$E7,$19,$8E,$CF,$E6,$3D,$AF,$D8,$7C,$97,$D4,$F6
	.dc.b	$BF,$8E,$71,$39,$DF,$2F,$9C,$FF,$CD,$91,$8C,$BF,$33,$7C,$3C,$56
	.dc.b	$3B,$96,$3F,$E7,$28,$C3,$63,$8E,$1F,$E7,$7A,$E3,$18,$38,$F0,$BC
	.dc.b	$1E,$32,$F9,$3A,$4A,$E1,$82,$DF,$0B,$E0,$6C,$EA,$F9,$36,$F5,$5E
	.dc.b	$98,$7C,$9B,$CD,$F0,$FC,$4C,$9F,$31,$A8,$E1,$3B,$8F,$F1,$6B,$5F
	.dc.b	$F1,$AB,$5F,$F1,$CB,$5F,$D2,$BB,$DB,$FA,$63,$75,$CD,$79,$78,$DE
	.dc.b	$F7,$C4,$F4,$E6,$3E,$D7,$E0,$7B,$52,$30,$41,$97,$E0,$37,$4F,$CE
	.dc.b	$F8,$E9,$5F,$1D,$F9,$5D,$72,$30,$43,$9E,$9F,$A7,$8F,$AF,$E3,$FF
	.dc.b	$93,$B8,$FF,$D6,$6D,$E3,$34,$97,$F5,$5E,$2F,$D7,$F9,$3F,$B5,$8F
	.dc.b	$E2,$B9,$DF,$33,$4C,$D7,$E0,$B5,$56,$23,$CA,$7E,$31,$FC,$26,$FC
	.dc.b	$FC,$9F,$36,$1F,$E5,$7A,$FE,$F4,$D7,$F4,$31,$A3,$83,$F9,$08,$FF
	.dc.b	$2F,$70,$7E,$8F,$E5,$3F,$44,$FF,$1C,$F8,$5F,$F3,$7B,$6B,$FC,$EE
	.dc.b	$D7,$FC,$F6,$D7,$F6,$9E,$2A,$79,$B6,$3F,$4D,$3C,$4F,$F0,$97,$C9
	.dc.b	$BC,$0F,$CD,$7C,$A3,$E0,$BF,$39,$39,$BF,$CC,$B5,$F8,$A2,$78,$3F
	.dc.b	$9E,$6B,$FE,$8E,$6B,$FE,$92,$6B,$FC,$F9,$AF,$F5,$EF,$94,$79,$5F
	.dc.b	$E9,$08,$46,$1F,$F4,$9B,$5F,$F4,$DB,$5F,$F4,$FB,$9C,$7F,$C6,$0D
	.dc.b	$E8,$FE,$9A,$D9,$F0,$70,$FE,$EF,$DE,$FD,$3D,$FF,$70,$F9,$47,$B7
	.dc.b	$FD,$50,$FF,$12,$5E,$5C,$7B,$DF,$D6,$27,$73,$FD,$E2,$F9,$A8,$CB
	.dc.b	$FE,$AB,$EE,$FD,$6A,$F8,$FC,$E1,$DA,$30,$F6,$7F,$C0,$71,$CA,$FA
	.dc.b	$B6,$B9,$DF,$2C,$3C,$98,$71,$C3,$C4,$FF,$AF,$AB,$6D,$C7,$D9,$3F
	.dc.b	$E1,$08,$F1,$5F,$38,$C9,$E9,$9B,$8F,$8A,$BB,$63,$BA,$F4,$9F,$0F
	.dc.b	$0F,$7C,$79,$EF,$9D,$76,$3C,$D8,$66,$FC,$E7,$19,$DE,$3E,$C8,$F3
	.dc.b	$35,$F2,$1C,$0B,$3E,$32,$57,$BE,$35,$8F,$A5,$E6,$7C,$E8,$F5,$9B
	.dc.b	$C3,$C9,$5C,$A7,$67,$2C,$7E,$5B,$E7,$3E,$09,$F8,$AC,$3F,$C3,$B5
	.dc.b	$F1,$4F,$71,$2F,$BC,$F4,$8C,$3C,$F9,$BF,$0C,$D8,$F5,$46,$23,$E0
	.dc.b	$B5,$FF,$24,$F9,$5F,$E4,$A8,$C4,$23,$13,$FE,$30,$8D,$9F,$EF,$87
	.dc.b	$0E,$CC,$1F,$E9,$73,$DB,$9B,$CA,$FF,$50,$4F,$E7,$A7,$9B,$FA,$09
	.dc.b	$D6,$FF,$36,$79,$4E,$6B,$8E,$3E,$0C,$35,$5F,$D4,$A7,$2F,$87,$4E
	.dc.b	$18,$7E,$52,$3D,$31,$87,$F0,$47,$8A,$FF,$C3,$55,$CA,$3F,$C9,$F7
	.dc.b	$DB,$1F,$2C,$7D,$83,$9D,$70,$C3,$B4,$7C,$3C,$3B,$1E,$F7,$13,$B7
	.dc.b	$C2,$60,$FF,$11,$E1,$FE,$C5,$9E,$18,$7E,$B3,$0F,$0C,$62,$BF,$35
	.dc.b	$18,$BF,$89,$1F,$42,$FD,$4F,$B4,$EB,$87,$C8,$7D,$26,$0F,$1E,$1D
	.dc.b	$E7,$C5,$5D,$30,$F5,$B4,$F5,$3E,$3C,$7E,$61,$63,$DC,$F9,$E3,$8D
	.dc.b	$FD,$59,$E2,$7F,$9F,$3E,$51,$EA,$7F,$B9,$73,$FF,$72,$46,$33,$FF
	.dc.b	$13,$AF,$6F,$51,$C6,$AE,$1F,$EA,$4A,$E0,$FE,$59,$AF,$E6,$DA,$F1
	.dc.b	$9F,$CD,$AF,$93,$7A,$BF,$EF,$58,$FF,$48,$C6,$27,$6D,$8F,$D1,$B5
	.dc.b	$FD,$43,$9C,$7D,$9F,$26,$F9,$0F,$3D,$77,$C3,$D3,$4A,$D6,$1F,$DD
	.dc.b	$BE,$4D,$E6,$E9,$B1,$FD,$EA,$77,$87,$F8,$5A,$D5,$A8,$C3,$5F,$F3
	.dc.b	$C3,$5F,$F3,$9B,$5F,$F1,$EB,$7C,$8F,$F3,$E7,$0F,$CE,$6E,$1F,$E6
	.dc.b	$3F,$3B,$1E,$1F,$F5,$78,$FF,$0E,$57,$C9,$C3,$FC,$F7,$39,$7F,$86
	.dc.b	$5F,$26,$C6,$5F,$F5,$FC,$7F,$80,$39,$3C,$1D,$9F,$B2,$CF,$FD,$1E
	.dc.b	$BE,$4B,$33,$32,$FF,$16,$70,$5C,$7F,$C7,$38,$FF,$A5,$63,$CF,$EA
	.dc.b	$4A,$FF,$3E,$7C,$9D,$BD,$E8,$AF,$56,$1F,$E9,$D6,$C6,$6F,$F9,$1C
	.dc.b	$E5,$B9,$BD,$BF,$8B,$E2,$7A,$9B,$3F,$E4,$18,$B8,$F9,$AB,$84,$E4
	.dc.b	$FF,$4A,$BF,$7B,$6F,$F5,$7E,$3F,$06,$73,$A7,$D2,$DA,$47,$F9,$35
	.dc.b	$F9,$31,$8C,$3F,$C6,$FF,$26,$F9,$2F,$C9,$B6,$FF,$D2,$BE,$9A,$B3
	.dc.b	$52,$7F,$4F,$5B,$9F,$E8,$6D,$7F,$A4,$35,$FF,$7C,$BB,$CB,$D3,$F2
	.dc.b	$7C,$E7,$E7,$BF,$7B,$CA,$FC,$5F,$1D,$C3,$FC,$CC,$E1,$A7,$C4,$8C
	.dc.b	$57,$D8,$C7,$F2,$67,$E2,$3E,$4E,$D7,$F9,$BE,$3F,$2F,$F2,$F8,$AF
	.dc.b	$BE,$FE,$D3,$47,$C5,$2F,$FB,$3F,$17,$C3,$BF,$F5,$E4,$E5,$FE,$D1
	.dc.b	$19,$A7,$8B,$9F,$13,$FC,$CD,$1D,$30,$E7,$7F,$36,$30,$FF,$0F,$6C
	.dc.b	$7F,$71,$6C,$75,$C3,$C1,$39,$7B,$61,$E1,$F0,$ED,$75,$9A,$FF,$B4
	.dc.b	$E3,$3C,$EF,$93,$FB,$D8,$79,$23,$FD,$F5,$E3,$EA,$73,$7E,$E3,$F7
	.dc.b	$71,$97,$9F,$3F,$FB,$32,$F9,$57,$0D,$3F,$D1,$18,$74,$9C,$E9,$D4
	.dc.b	$ED,$F1,$F6,$32,$4C,$7C,$15,$E7,$9D,$46,$30,$F9,$38,$76,$6E,$BF
	.dc.b	$99,$BF,$77,$C6,$FA,$F6,$99,$7B,$31,$57,$C3,$9F,$9B,$1C,$70,$FA
	.dc.b	$2E,$9B,$3B,$AF,$91,$A3,$E5,$3A,$DF,$A2,$31,$3F,$13,$E3,$BE,$61
	.dc.b	$F3,$78,$2C,$64,$E7,$F3,$38,$BE,$4F,$D3,$F8,$FF,$9F,$F2,$B6,$BC
	.dc.b	$2D,$AF,$03,$78,$3E,$8E,$48,$FC,$1F,$13,$0B,$F2,$18,$6E,$3F,$96
	.dc.b	$C5,$5B,$8C,$2A,$E1,$18,$78,$CF,$E1,$CE,$2E,$9C,$4F,$19,$FC,$3F
	.dc.b	$9B,$F8,$F3,$F2,$1F,$21,$63,$F1,$7F,$21,$F9,$8F,$7F,$FD,$A9,$18
	.dc.b	$3F,$30,$E3,$0F,$C2,$7C,$87,$E6,$4F,$CD,$7C,$85,$AF,$CA,$FC,$87
	.dc.b	$E7,$4F,$CF,$7A,$76,$3F,$33,$F2,$BD,$FE,$17,$F4,$4D,$8F,$CE,$FC
	.dc.b	$87,$E9,$4F,$D3,$7A,$96,$3F,$43,$F1,$FF,$A9,$3F,$55,$F0,$16,$32
	.dc.b	$7A,$0C,$6B,$F9,$ED,$BF,$88,$46,$A3,$0E,$B9,$FF,$57,$3A,$9E,$1F
	.dc.b	$1F,$F1,$4F,$8D,$F2,$B6,$FE,$2C,$7C,$9F,$95,$C5,$77,$6E,$1D,$1B
	.dc.b	$E1,$FE,$A7,$14,$7E,$23,$97,$F4,$B7,$FC,$65,$AF,$E1,$9B,$CE,$69
	.dc.b	$FC,$4B,$E5,$39,$F8,$BF,$8A,$8C,$5A,$31,$A7,$E4,$6B,$78,$7E,$22
	.dc.b	$F7,$59,$BE,$13,$F8,$FA,$DE,$7F,$CC,$2F,$73,$9A,$DC,$B3,$FE,$79
	.dc.b	$7C,$1F,$E7,$57,$C2,$75,$7F,$D8,$EB,$4D,$BE,$77,$C3,$1F,$26,$32
	.dc.b	$BC,$0F,$B1,$A4,$C3,$AA,$C3,$AB,$F2,$EB,$94,$EA,$72,$6B,$2F,$9D
	.dc.b	$3A,$C3,$D7,$8F,$93,$0F,$36,$3E,$1C,$79,$4E,$67,$C0,$DC,$7A,$65
	.dc.b	$F0,$70,$E7,$87,$5B,$FC,$06,$35,$E3,$17,$2F,$1C,$7B,$6B,$8B,$4E
	.dc.b	$25,$38,$96,$38,$E5,$E6,$AE,$D8,$F9,$30,$B9,$79,$CF,$55,$F0,$C7
	.dc.b	$BE,$1C,$71,$E9,$97,$AD,$F8,$66,$30,$E7,$97,$A6,$75,$3E,$43,$FD
	.dc.b	$C5,$97,$E4,$9B,$97,$E4,$1B,$5F,$E8,$AF,$7F,$FA,$17,$CA,$DF,$0C
	.dc.b	$D2,$57,$FB,$AB,$07,$C9,$BE,$07,$FD,$27,$18,$6C,$66,$78,$35,$F2
	.dc.b	$B9,$D7,$F0,$CD,$DB,$F3,$AD,$9F,$CD,$7C,$47,$FB,$7A,$73,$7F,$AF
	.dc.b	$AD,$57,$E4,$DB,$82,$F8,$4F,$E8,$E7,$24,$AF,$DC,$4E,$A7,$94,$75
	.dc.b	$23,$FB,$78,$D5,$7B,$E3,$F7,$55,$CD,$D1,$C3,$0F,$6D,$DC,$3F,$73
	.dc.b	$87,$2B,$E1,$5A,$BE,$27,$4B,$E0,$FA,$5F,$3B,$83,$CE,$6C,$E3,$1C
	.dc.b	$C9,$7F,$22,$32,$D8,$CC,$75,$32,$57,$2F,$67,$F7,$97,$E9,$AF,$16
	.dc.b	$1F,$0E,$73,$58,$C3,$DB,$87,$C0,$8C,$5F,$0C,$39,$47,$A6,$FB,$E1
	.dc.b	$F6,$63,$66,$6F,$93,$4F,$33,$EB,$33,$3C,$B0,$F0,$CF,$E2,$8E,$71
	.dc.b	$DA,$FC,$2E,$F0,$F9,$B3,$D4,$F8,$17,$D5,$E4,$67,$0F,$81,$19,$8F
	.dc.b	$E4,$97,$D2,$31,$3A,$9E,$37,$DD,$E5,$5B,$7D,$64,$39,$9E,$8A,$E7
	.dc.b	$1B,$8F,$29,$BA,$F9,$0F,$97,$0F,$82,$F2,$31,$87,$D1,$7A,$1C,$6B
	.dc.b	$A5,$79,$E3,$A9,$C7,$0F,$41,$F2,$5E,$26,$A3,$C3,$1F,$82,$8E,$E7
	.dc.b	$E3,$0E,$58,$7C,$4C,$3F,$6F,$7F,$AC,$3C,$2E,$27,$E9,$D7,$82,$BE
	.dc.b	$E9,$23,$E0,$4E,$DE,$D7,$E9,$25,$F5,$8F,$3C,$79,$AB,$94,$7A,$6B
	.dc.b	$95,$FA,$4F,$04,$7D,$89,$DD,$75,$AD,$C6,$1F,$9F,$1D,$67,$81,$E6
	.dc.b	$C3,$D7,$5B,$7E,$94,$7F,$00,$DF,$F9,$CE,$3F,$EB,$96,$C7,$FA,$04
	.dc.b	$FB,$4E,$1F,$7C,$7F,$14,$F8,$3F,$74,$60,$FF,$0A,$38,$CB,$D9,$87
	.dc.b	$AA,$30,$FB,$2B,$95,$70,$3C,$E7,$A8,$FF,$00,$4E,$9E,$C7,$F8,$6D
	.dc.b	$C5,$FF,$52,$7D,$51,$FE,$18,$6B,$FE,$15,$6B,$D0,$F0,$1C,$CF,$F6
	.dc.b	$8E,$6F,$D6,$1E,$BF,$F3,$C8,$C3,$E3,$CB,$FC,$21,$79,$AD,$DF,$26
	.dc.b	$BF,$EC,$5B,$E2,$E6,$FF,$A3,$F9,$4F,$1D,$E9,$E1,$7A,$77,$E6,$66
	.dc.b	$1F,$EA,$C8,$D1,$1C,$1F,$EB,$0F,$99,$EB,$E2,$FF,$BD,$62,$62,$8C
	.dc.b	$47,$FB,$E6,$30,$E0,$FF,$60,$C5,$69,$FE,$FF,$F6,$3F,$DF,$D3,$C2
	.dc.b	$02,$33,$3E,$16,$F4,$71,$E1,$FF,$69,$4B,$D4,$C5,$F8,$3B,$2E,$3E
	.dc.b	$76,$D6,$E7,$BE,$3D,$AF,$24,$8C,$E1,$DB,$D3,$E3,$E9,$FB,$72,$F8
	.dc.b	$4D,$C4,$D2,$5F,$23,$E1,$BC,$9C,$4F,$CB,$F4,$7D,$30,$FB,$47,$8B
	.dc.b	$F2,$5B,$1F,$35,$AF,$D2,$6B,$F5,$DC,$E8,$CF,$E6,$CF,$5A,$EF,$5A
	.dc.b	$9F,$A3,$CB,$FE,$87,$ED,$F9,$6D,$BF,$51,$C4,$B5,$C5,$FF,$19,$F0
	.dc.b	$7F,$8C,$1F,$81,$3C,$5F,$B9,$7D,$A7,$51,$83,$8D,$FD,$8F,$A5,$7C
	.dc.b	$0F,$C4,$F8,$A7,$C6,$7E,$2D,$B7,$F8,$F6,$BF,$94,$6B,$F9,$97,$3A
	.dc.b	$FC,$38,$FC,$7C,$63,$2F,$8A,$D8,$FC,$9F,$C6,$78,$9F,$83,$E2,$3D
	.dc.b	$AF,$CB,$43,$E2,$DC,$5F,$38,$FC,$74,$65,$C7,$3F,$D2,$9D,$CF,$E8
	.dc.b	$2F,$A6,$7F,$A2,$9C,$CA,$72,$FC,$4B,$E6,$FE,$A3,$5D,$C7,$81,$E3
	.dc.b	$3C,$1B,$8F,$A2,$F7,$8F,$8E,$F9,$E1,$E7,$3C,$95,$B8,$FC,$4E,$1E
	.dc.b	$2B,$E3,$5E,$79,$E3,$1D,$B1,$C4,$F2,$7C,$D3,$DB,$1E,$33,$EB,$34
	.dc.b	$53,$E0,$61,$E1,$7F,$82,$CF,$CC,$BE,$27,$78,$F4,$1C,$2F,$C2,$FD
	.dc.b	$4C,$7D,$CF,$26,$BB,$38,$E3,$ED,$BF,$59,$23,$CB,$7D,$F1,$F4,$92
	.dc.b	$7E,$2E,$3F,$84,$8F,$24,$F9,$A7,$95,$77,$8E,$51,$F0,$C9,$1F,$0A
	.dc.b	$72,$F9,$CF,$92,$53,$E3,$5F,$DD,$9F,$0E,$3F,$CA,$BC,$4F,$CC,$57
	.dc.b	$B5,$B5,$F8,$A6,$E3,$F9,$A7,$73,$FC,$CB,$C2,$E5,$8F,$F3,$E6,$C7
	.dc.b	$F4,$36,$BF,$C3,$1A,$F1,$3E,$D6,$0F,$7B,$F6,$15,$F9,$C9,$D3,$F9
	.dc.b	$E9,$D6,$FE,$E9,$E8,$72,$AE,$39,$F8,$E3,$13,$B7,$C1,$A7,$3A,$FD
	.dc.b	$F3,$DE,$18,$FE,$13,$0E,$F1,$F8,$83,$C7,$87,$F5,$5B,$EF,$1E,$F3
	.dc.b	$B9,$F9,$53,$BC,$6F,$0F,$C7,$C6,$0F,$E9,$B3,$E2,$9F,$46,$1E,$5A
	.dc.b	$DD,$79,$0F,$CC,$DF,$92,$7D,$18,$79,$F0,$E5,$5E,$39,$E4,$FA,$4E
	.dc.b	$58,$E0,$E2,$FA,$1B,$1C,$0B,$8F,$A6,$79,$5F,$7C,$7C,$AD,$8C,$9C
	.dc.b	$E7,$E9,$5F,$77,$DA,$73,$BE,$B3,$F0,$AF,$F6,$7F,$14,$F3,$FF,$99
	.dc.b	$67,$FC,$CE,$93,$9A,$FE,$D1,$2F,$9B,$7C,$5F,$09,$FE,$B7,$5B,$BF
	.dc.b	$E9,$47,$2C,$3A,$3A,$6B,$FE,$0E,$8E,$0F,$F8,$75,$AF,$F6,$2E,$87
	.dc.b	$C2,$77,$84,$48,$C7,$63,$F1,$77,$AD,$BA,$D6,$5F,$D8,$9E,$6F,$F6
	.dc.b	$AF,$C4,$1F,$15,$B9,$CE,$5E,$A9,$CD,$6E,$BC,$52,$9C,$BF,$DA,$6B
	.dc.b	$BB,$C6,$77,$3A,$8F,$F1,$D3,$5F,$F0,$1B,$5F,$F1,$8B,$71,$FF,$1F
	.dc.b	$7C,$17,$BB,$C3,$37,$E0,$30,$CC,$BF,$C8,$75,$C1,$B5,$C1,$B3,$F4
	.dc.b	$A7,$FC,$95,$F0,$1D,$4E,$DF,$01,$6B,$FC,$4B,$69,$F1,$7A,$D9,$CF
	.dc.b	$FC,$11,$87,$F9,$4E,$3C,$DF,$01,$30,$E3,$F0,$16,$F7,$8A,$BC,$4D
	.dc.b	$AF,$3C,$60,$CC,$7F,$81,$FA,$DE,$AF,$F9,$42,$30,$FF,$94,$BE,$02
	.dc.b	$E1,$FE,$60,$9B,$9F,$F9,$9D,$A3,$E0,$2B,$87,$8F,$1B,$E8,$18,$61
	.dc.b	$BA,$7F,$35,$CB,$FC,$C3,$F0,$39,$A9,$39,$CF,$FC,$D7,$E8,$38,$B3
	.dc.b	$1F,$C8,$B0,$FF,$21,$CE,$6B,$84,$A3,$0E,$CE,$75,$CB,$6F,$87,$18
	.dc.b	$AE,$35,$C1,$FF,$44,$36,$3B,$D7,$F8,$C6,$72,$D8,$CC,$72,$E1,$F9
	.dc.b	$71,$E3,$9B,$49,$B1,$86,$D7,$12,$57,$A4,$F6,$46,$9C,$57,$39,$DD
	.dc.b	$FB,$27,$71,$EC,$BE,$CE,$B0,$F9,$38,$73,$BE,$18,$73,$C3,$AC,$F0
	.dc.b	$7B,$E1,$6B,$73,$F0,$67,$EB,$B5,$F0,$9F,$C6,$AB,$A6,$1F,$32,$7E
	.dc.b	$2F,$B4,$F7,$1E,$77,$97,$0F,$F5,$5C,$E6,$31,$08,$C3,$E6,$BF,$C0
	.dc.b	$C6,$31,$F2,$C6,$2F,$AD,$6C,$F9,$95,$BC,$FE,$1D,$6F,$0E,$F3,$9A
	.dc.b	$F6,$CE,$5E,$58,$7C,$F9,$D4,$7C,$08,$C6,$1E,$07,$C7,$1E,$08,$C6
	.dc.b	$1E,$08,$C6,$1E,$09,$CC,$78,$0C,$9E,$09,$FF,$51,$DF,$83,$2F,$B1
	.dc.b	$87,$83,$0E,$44,$BE,$07,$CC,$C3,$AB,$E7,$C3,$9C,$62,$F9,$C6,$2B
	.dc.b	$95,$FA,$8E,$D7,$D0,$FA,$97,$C7,$2F,$BA,$F6,$C7,$E2,$9E,$78,$CD
	.dc.b	$FC,$BB,$E6,$75,$C3,$C1,$87,$C0,$8C,$C7,$C7,$6C,$7A,$E7,$94,$E4
	.dc.b	$E3,$3A,$9F,$2C,$FA,$A7,$24,$8F,$9D,$3B,$BF,$3E,$1E,$BC,$BA,$D7
	.dc.b	$73,$C9,$7D,$DC,$5F,$1A,$E1,$1A,$33,$3D,$DF,$A0,$60,$87,$98,$F9
	.dc.b	$97,$E2,$C3,$C3,$1C,$0D,$5F,$78,$F6,$47,$29,$F9,$D8,$70,$79,$47
	.dc.b	$4C,$3E,$A1,$2B,$8C,$74,$C3,$CD,$97,$E0,$A3,$A5,$6D,$E5,$5D,$63
	.dc.b	$DF,$8F,$8E,$7E,$01,$E7,$C3,$C3,$1F,$4E,$FA,$4F,$73,$AD,$F8,$AB
	.dc.b	$78,$7C,$D9,$D1,$23,$DA,$F6,$76,$FD,$A8,$F9,$D1,$88,$F1,$61,$E9
	.dc.b	$C3,$C5,$97,$E4,$A7,$94,$FA,$AF,$8B,$B7,$C5,$7C,$CF,$0D,$FB,$CE
	.dc.b	$79,$7C,$3C,$7E,$BE,$5E,$EA,$E1,$87,$5C,$7D,$E7,$C6,$9D,$47,$CD
	.dc.b	$81,$FA,$22,$9F,$91,$29,$F8,$72,$F9,$0F,$73,$FC,$7F,$1A,$3F,$DE
	.dc.b	$4E,$1D,$B5,$FF,$72,$B5,$FE,$63,$E7,$BC,$9E,$51,$FC,$C6,$30,$F5
	.dc.b	$C5,$18,$8F,$2C,$62,$78,$B6,$3F,$98,$CF,$2D,$3F,$95,$37,$5F,$E3
	.dc.b	$BE,$E7,$FA,$D5,$FE,$3F,$19,$AF,$2E,$1F,$D0,$23,$11,$FD,$16,$B8
	.dc.b	$3A,$3B,$CE,$CF,$E8,$F1,$A8,$F6,$46,$5D,$9F,$D1,$A3,$58,$7A,$70
	.dc.b	$F3,$D6,$DF,$E8,$53,$AB,$F1,$D6,$F0,$F4,$DF,$83,$0E,$11,$87,$BE
	.dc.b	$1C,$9C,$9E,$97,$BB,$E3,$32,$7A,$0E,$31,$F2,$1C,$D7,$85,$E5,$19
	.dc.b	$76,$60,$87,$C9,$38,$DF,$A8,$F5,$E1,$ED,$C3,$27,$AE,$79,$9E,$3C
	.dc.b	$3C,$E7,$AD,$CE,$1E,$37,$58,$78,$E7,$6F,$C3,$6C,$7F,$80,$AF,$C0
	.dc.b	$F2,$C3,$C6,$F8,$8E,$B1,$A9,$E3,$1D,$8E,$46,$DC,$DA,$78,$5F,$84
	.dc.b	$95,$C0,$F0,$61,$F3,$27,$71,$F2,$2B,$D8,$73,$7B,$57,$27,$E6,$1C
	.dc.b	$8F,$39,$A3,$94,$F4,$7D,$04,$9E,$8F,$1B,$F4,$CE,$E7,$C3,$87,$C4
	.dc.b	$AD,$5F,$86,$78,$DF,$C0,$AF,$B9,$87,$D9,$8C,$3E,$DB,$F2,$1E,$1C
	.dc.b	$3B,$57,$08,$FC,$91,$D8,$F4,$9C,$F0,$F0,$47,$8E,$FF,$04,$76,$9E
	.dc.b	$2E,$27,$77,$C6,$7E,$84,$EC,$D3,$89,$DB,$C9,$FC,$51,$88,$DC,$7D
	.dc.b	$07,$94,$6A,$39,$1F,$66,$BC,$D1,$98,$FA,$11,$87,$66,$2B,$CC,$F5
	.dc.b	$9F,$01,$C7,$0F,$9E,$E2,$38,$C6,$1F,$04,$F2,$8F,$BB,$B7,$F2,$8C
	.dc.b	$7F,$9F,$B6,$BF,$A1,$B5,$37,$1F,$7D,$75,$9F,$F3,$37,$B9,$FC,$62
	.dc.b	$E4,$77,$8A,$FC,$93,$CA,$FF,$1E,$33,$5D,$1C,$47,$9E,$6B,$FE,$6A
	.dc.b	$8D,$E3,$F0,$5C,$E1,$FD,$09,$F5,$CF,$F9,$61,$AF,$F9,$F9,$BC,$AF
	.dc.b	$71,$EE,$76,$CB,$FC,$DD,$5F,$CF,$2B,$53,$6B,$77,$F1,$8F,$ED,$33
	.dc.b	$A7,$01,$C3,$F2,$8D,$C7,$27,$03,$A9,$ED,$BD,$AF,$E9,$B1,$8F,$6E
	.dc.b	$C9,$B3,$2B,$C3,$5C,$A3,$E0,$B6,$33,$5D,$5D,$F6,$FF,$7E,$F3,$3C
	.dc.b	$D4,$87,$C0,$FF,$81,$27,$31,$8A,$FF,$04,$C6,$1D,$6F,$FE,$0C,$BE
	.dc.b	$51,$FD,$F1,$AF,$F8,$1D,$AF,$F8,$39,$AF,$F8,$49,$F1,$7A,$7D,$23
	.dc.b	$E3,$5F,$FB,$C5,$F0,$D7,$78,$C4,$2F,$CF,$5C,$6F,$9C,$F0,$39,$F4
	.dc.b	$E6,$F8,$DF,$09,$D5,$7C,$3F,$6B,$5E,$6C,$9C,$DA,$79,$1E,$F7,$11
	.dc.b	$EF,$6B,$EF,$6B,$FE,$2E,$F6,$5E,$D7,$BF,$89,$E3,$FF,$34,$CF,$F1
	.dc.b	$51,$8E,$8F,$89,$83,$0E,$2F,$B4,$C6,$1F,$E2,$C6,$9F,$E2,$E8,$DB
	.dc.b	$FE,$39,$6B,$FE,$42,$AF,$F7,$3F,$5F,$69,$DE,$1F,$98,$6F,$07,$E8
	.dc.b	$BB,$FF,$C6,$B8,$7C,$DC,$38,$BE,$3A,$DB,$FE,$50,$6C,$6A,$3E,$C5
	.dc.b	$7F,$94,$BD,$87,$B3,$EE,$F6,$5D,$B4,$F8,$33,$AB,$F8,$0D,$7F,$CA
	.dc.b	$8D,$7F,$CB,$6D,$7B,$57,$C5,$F0,$5E,$CF,$E2,$65,$EB,$3D,$7F,$F3
	.dc.b	$28,$E9,$FF,$33,$C6,$67,$FC,$D7,$39,$7F,$CD,$B1,$95,$7D,$B8,$33
	.dc.b	$5F,$E7,$08,$C4,$F8,$23,$BD,$F6,$6C,$70,$3F,$CF,$33,$C2,$FB,$3C
	.dc.b	$BD,$8F,$05,$78,$E9,$18,$BF,$F3,$FB,$C3,$1F,$80,$DC,$7B,$57,$C1
	.dc.b	$AF,$13,$EE,$6B,$EC,$AF,$91,$8F,$18,$F5,$1B,$39,$E1,$E3,$8D,$11
	.dc.b	$F6,$C7,$03,$C0,$78,$B1,$F7,$47,$C7,$AE,$EF,$0A,$F6,$56,$E7,$2F
	.dc.b	$C6,$6B,$D0,$ED,$3E,$33,$A6,$1C,$4F,$3E,$1E,$E8,$D4,$F3,$9D,$3C
	.dc.b	$A7,$47,$C4,$C3,$CF,$1B,$34,$74,$38,$E1,$F0,$5F,$8C,$68,$F8,$51
	.dc.b	$C9,$F9,$D5,$FE,$B8,$F1,$3E,$DC,$75,$6B,$A3,$FD,$59,$97,$B1,$A7
	.dc.b	$FA,$F3,$C3,$E5,$1F,$EB,$96,$C6,$4D,$9F,$EC,$47,$2D,$3E,$B5,$3D
	.dc.b	$9F,$EB,$D8,$FE,$7E,$7F,$D9,$B5,$F8,$AA,$D7,$0F,$E3,$27,$27,$2B
	.dc.b	$E3,$A7,$8E,$31,$3B,$57,$CE,$B7,$1E,$7C,$3F,$5D,$87,$68,$FC,$51
	.dc.b	$DA,$33,$3C,$23,$D5,$18,$BE,$D7,$D5,$B3,$E2,$39,$9E,$23,$13,$EF
	.dc.b	$70,$F3,$77,$3C,$DA,$70,$C3,$DD,$3B,$39,$4E,$DB,$3B,$3D,$D3,$FD
	.dc.b	$CE,$FD,$51,$F2,$B1,$F0,$1C,$0B,$3D,$8E,$18,$7B,$2F,$C1,$19,$AF
	.dc.b	$EF,$77,$D1,$F5,$C7,$07,$F6,$9E,$BB,$B3,$FC,$33,$02,$46,$70,$FF
	.dc.b	$7D,$B6,$34,$47,$0E,$03,$BC,$FF,$C5,$03,$E3,$BF,$C6,$F5,$FD,$76
	.dc.b	$74,$FF,$84,$A3,$35,$FD,$28,$FF,$19,$72,$BD,$8E,$75,$E2,$F5,$5E
	.dc.b	$B6,$2F,$75,$54,$EA,$BF,$AE,$5F,$37,$FB,$74,$66,$7C,$13,$FE,$41
	.dc.b	$AE,$53,$9C,$BF,$C7,$EE,$EB,$E0,$35,$F2,$61,$F3,$1F,$99,$19,$6C
	.dc.b	$79,$EF,$FA,$BF,$C6,$76,$8C,$72,$A3,$D9,$E1,$7F,$8E,$FD,$6E,$8D
	.dc.b	$D3,$FC,$29,$18,$F0,$8D,$38,$BF,$E1,$48,$FF,$0F,$C6,$3C,$0C,$F8
	.dc.b	$5A,$F3,$ED,$7F,$89,$1B,$AF,$D3,$C7,$FC,$99,$18,$DF,$FC,$96,$F6
	.dc.b	$F5,$7C,$44,$BF,$F1,$43,$6F,$96,$6F,$83,$E1,$5F,$E3,$27,$CF,$1F
	.dc.b	$E6,$F8,$FF,$25,$3F,$E5,$B7,$FC,$BE,$E6,$7F,$C9,$0D,$7F,$C9,$AF
	.dc.b	$F9,$DF,$AB,$FC,$1F,$1F,$E5,$0F,$5B,$1E,$0F,$0B,$FE,$3A,$FF,$94
	.dc.b	$63,$33,$FE,$81,$F5,$9E,$AF,$5E,$54,$63,$0F,$F4,$3C,$62,$75,$39
	.dc.b	$AF,$F4,$2C,$E6,$BB,$C7,$EB,$FC,$A7,$AB,$E0,$AD,$4F,$F2,$6C,$BF
	.dc.b	$5F,$5A,$BE,$32,$8C,$3F,$BA,$6C,$7E,$CD,$B1,$CA,$FE,$33,$C5,$F7
	.dc.b	$BC,$F7,$B3,$D3,$4F,$93,$39,$BF,$0C,$E7,$0F,$2D,$F0,$C3,$8B,$6B
	.dc.b	$6D,$BF,$61,$F0,$5D,$C6,$8E,$B3,$89,$D9,$E1,$2D,$AB,$A6,$1E,$68
	.dc.b	$C5,$F3,$AD,$D6,$B0,$F3,$CE,$4F,$8D,$1F,$12,$B8,$D7,$CB,$8C,$1F
	.dc.b	$1B,$D3,$7A,$AF,$2B,$B6,$5F,$1A,$73,$18,$CB,$F9,$3C,$63,$0F,$1E
	.dc.b	$5F,$6A,$31,$7E,$9C,$7B,$63,$D6,$F6,$7C,$2A,$DE,$1E,$A8,$C4,$EE
	.dc.b	$F9,$61,$D6,$31,$5B,$9F,$34,$62,$BE,$95,$6F,$1F,$44,$7F,$43,$C3
	.dc.b	$BB,$A3,$BC,$69,$46,$9F,$AD,$7D,$A7,$53,$9C,$BE,$6C,$E6,$BD,$AD
	.dc.b	$3D,$2F,$63,$38,$FC,$AC,$7D,$11,$DA,$31,$87,$69,$CD,$F6,$9C,$D7
	.dc.b	$23,$D9,$3A,$CB,$E7,$DF,$12,$4F,$87,$0F,$25,$FC,$93,$9B,$88,$E7
	.dc.b	$3C,$23,$11,$F1,$2F,$B6,$5F,$36,$3A,$5F,$23,$A3,$C8,$C9,$6B,$E1
	.dc.b	$9D,$32,$F9,$C7,$CB,$C7,$D3,$3C,$0F,$06,$3E,$3B,$ED,$1C,$DE,$EE
	.dc.b	$8E,$15,$CF,$0F,$36,$1D,$F1,$F4,$D6,$9F,$A3,$5B,$8C,$61,$C3,$0E
	.dc.b	$B3,$E6,$AD,$9C,$AF,$A5,$6E,$FF,$19,$18,$7D,$59,$7D,$97,$D0,$43
	.dc.b	$D1,$1F,$13,$0F,$6C,$61,$B3,$C6,$30,$71,$C3,$D1,$5D,$CF,$AF,$3B
	.dc.b	$8F,$45,$F9,$6B,$F1,$C7,$1C,$3D,$71,$F3,$63,$0D,$7B,$92,$BA,$4F
	.dc.b	$08,$F6,$E5,$F6,$30,$F1,$63,$F2,$2B,$51,$E4,$9D,$4F,$BB,$1F,$4B
	.dc.b	$6B,$6F,$1C,$7D,$13,$AC,$3C,$A7,$83,$2C,$DF,$08,$FA,$37,$CE,$FE
	.dc.b	$06,$5E,$87,$67,$B7,$0F,$8B,$5F,$AB,$3D,$EE,$32,$FA,$91,$E0,$3F
	.dc.b	$C8,$91,$FC,$85,$B8,$7F,$1D,$6D,$7F,$92,$5F,$F7,$1F,$AA,$D7,$11
	.dc.b	$FE,$F0,$8C,$3B,$F1,$91,$88,$FF,$29,$46,$27,$8E,$0F,$31,$E3,$E1
	.dc.b	$B7,$E1,$9D,$4E,$49,$1F,$CC,$67,$79,$FF,$3F,$6E,$BF,$CF,$BC,$DF
	.dc.b	$F5,$73,$FD,$02,$33,$5F,$CA,$DA,$FF,$28,$6E,$0A,$E3,$3B,$C3,$FA
	.dc.b	$35,$68,$DC,$EA,$79,$4E,$88,$FF,$47,$8D,$57,$B2,$3D,$53,$AC,$3F
	.dc.b	$AC,$56,$AF,$B4,$E7,$0F,$65,$D3,$C8,$72,$3B,$19,$3C,$66,$6F,$AB
	.dc.b	$9C,$3C,$51,$F2,$6B,$FB,$16,$1E,$17,$58,$7B,$AE,$C7,$19,$D4,$75
	.dc.b	$C3,$E7,$4F,$28,$D9,$93,$FA,$D4,$F0,$BF,$04,$64,$FE,$B3,$19,$AE
	.dc.b	$71,$9C,$3E,$A3,$D2,$B8,$9E,$C7,$58,$7C,$87,$E1,$98,$C3,$D0,$F1
	.dc.b	$BE,$B5,$C4,$E6,$E0,$91,$E8,$38,$1F,$42,$38,$46,$A3,$E7,$C7,$27
	.dc.b	$81,$E1,$3D,$58,$7B,$5E,$53,$C7,$0F,$6C,$6E,$B9,$3F,$10,$F0,$9D
	.dc.b	$A7,$F0,$6E,$DE,$55,$D6,$34,$76,$3C,$73,$D2,$BF,$03,$5C,$27,$AC
	.dc.b	$61,$DC,$79,$70,$F9,$95,$F0,$27,$A1,$D8,$95,$EA,$8E,$D3,$C2,$BC
	.dc.b	$F1,$8C,$3E,$9D,$FA,$4E,$77,$C6,$BC,$13,$AA,$E1,$5F,$28,$F3,$DF
	.dc.b	$E5,$5E,$47,$8B,$0F,$85,$87,$AE,$30,$70,$BF,$CB,$3C,$23,$47,$85
	.dc.b	$F1,$D7,$A5,$F8,$66,$27,$C9,$87,$9A,$3C,$87,$39,$F2,$61,$E8,$3E
	.dc.b	$86,$1E,$48,$E9,$3F,$74,$E1,$7E,$4A,$E1,$7F,$56,$72,$7E,$72,$32
	.dc.b	$FC,$8B,$E9,$5E,$48,$F3,$1E,$49,$DB,$F0,$8E,$6F,$DB,$7E,$CD,$F9
	.dc.b	$30,$F8,$51,$8A,$E8,$D3,$A6,$1F,$06,$FD,$B1,$F3,$4F,$34,$64,$F5
	.dc.b	$47,$8A,$BE,$7E,$1E,$8A,$E6,$7A,$27,$9B,$E6,$AF,$C4,$9C,$DF,$80
	.dc.b	$77,$8F,$55,$7D,$AC,$3D,$57,$D6,$79,$9E,$18,$FA,$B7,$CE,$FE,$24
	.dc.b	$FF,$AB,$BC,$B7,$97,$FD,$82,$51,$87,$FB,$1B,$5F,$ED,$0D,$44,$38
	.dc.b	$C5,$EA,$71,$66,$B9,$46,$2B,$8C,$62,$79,$79,$8F,$2F,$94,$71,$C2
	.dc.b	$DF,$06,$BE,$03,$C7,$87,$55,$7C,$DC,$D7,$9F,$8D,$F0,$1F,$05,$B3
	.dc.b	$FE,$02,$6B,$FE,$07,$6B,$EF,$3F,$D6,$B1,$E2,$8D,$DF,$85,$B5,$F1
	.dc.b	$7E,$36,$DF,$C8,$9A,$F2,$3C,$A6,$6F,$DB,$3A,$99,$1D,$CC,$9F,$0B
	.dc.b	$F1,$AF,$F0,$AD,$72,$1E,$2A,$FE,$3B,$63,$53,$C1,$63,$F2,$63,$0D
	.dc.b	$E6,$EC,$72,$F9,$5B,$5C,$A2,$BC,$B3,$FC,$26,$0A,$FA,$FF,$A0,$3E
	.dc.b	$62,$92,$FD,$1F,$9F,$BD,$FD,$FB,$B2,$7D,$96,$8F,$C0,$D3,$E1,$7F
	.dc.b	$07,$DB,$F0,$36,$FC,$AF,$67,$E0,$71,$4E,$5F,$F1,$AF,$37,$E6,$63
	.dc.b	$13,$B6,$C7,$F9,$1F,$A3,$9B,$6B,$89,$F8,$38,$CB,$4C,$8F,$21,$E1
	.dc.b	$7E,$87,$1F,$CB,$46,$38,$BF,$03,$0F,$A4,$7B,$5B,$9F,$D7,$56,$F0
	.dc.b	$F9,$C7,$EC,$97,$8D,$B8,$C3,$70,$7E,$C0,$F2,$F5,$5E,$5A,$BF,$8E
	.dc.b	$BE,$26,$9C,$74,$79,$B8,$D3,$F1,$D5,$B6,$D6,$E7,$58,$79,$85,$6D
	.dc.b	$F4,$E6,$F8,$8F,$E0,$F8,$72,$3E,$1B,$89,$E5,$F8,$93,$E8,$FD,$2D
	.dc.b	$AD,$FC,$46,$26,$47,$E8,$E7,$55,$E0,$B5,$EA,$3A,$DD,$C3,$EB,$56
	.dc.b	$B0,$E5,$3B,$AF,$E0,$58,$74,$6C,$7F,$11,$7F,$92,$46,$5F,$E4,$DE
	.dc.b	$D3,$DB,$FA,$2D,$BF,$7F,$C5,$62,$64,$7E,$EE,$75,$5E,$9A,$CB,$F3
	.dc.b	$63,$2F,$F1,$56,$BE,$0A,$F2,$3F,$CE,$23,$2F,$FA,$E3,$9B,$FD,$56
	.dc.b	$FF,$AB,$BC,$6F,$F5,$C3,$FE,$D4,$31,$E7,$FF,$AE,$3D,$1F,$F6,$3C
	.dc.b	$71,$F3,$3F,$D8,$8F,$FB,$D8,$FF,$5C,$79,$FF,$EB,$78,$FF,$73,$CF
	.dc.b	$09,$E6,$6A,$FB,$D7,$FB,$2E,$33,$87,$FB,$57,$4F,$F6,$2C,$7F,$B3
	.dc.b	$4E,$D1,$96,$E1,$DA,$77,$39,$3F,$DC,$33,$FE,$FD,$F4,$7F,$DA,$0F
	.dc.b	$03,$FD,$AA,$7A,$CD,$63,$E7,$C3,$D5,$1A,$3C,$B8,$77,$79,$1F,$EF
	.dc.b	$39,$E1,$8F,$CA,$BE,$F8,$78,$AB,$E6,$61,$D1,$E1,$8F,$A3,$1E,$95
	.dc.b	$BA,$F8,$86,$CB,$87,$C8,$19,$7F,$BA,$5F,$06,$5E,$28,$FB,$18,$75
	.dc.b	$C7,$DA,$D7,$C2,$7A,$AF,$AC,$E7,$2F,$4E,$5F,$EF,$9A,$FB,$13,$9C
	.dc.b	$3B,$47,$D5,$CB,$D1,$97,$92,$36,$7C,$6C,$39,$BF,$3C,$F8,$71,$EB
	.dc.b	$BF,$51,$E8,$8F,$24,$73,$AF,$CB,$E3,$F0,$E3,$18,$FE,$4A,$7E,$76
	.dc.b	$1D,$AF,$EB,$9F,$66,$F8,$61,$DA,$3D,$17,$ED,$BF,$C4,$34,$FA,$65
	.dc.b	$8F,$92,$79,$27,$94,$F0,$8F,$65,$7C,$A7,$E0,$63,$F1,$A7,$51,$E4
	.dc.b	$3E,$8B,$F2,$2B,$9D,$F8,$27,$9E,$3E,$E7,$D4,$FC,$77,$E6,$D7,$C0
	.dc.b	$BE,$D5,$C6,$7B,$61,$E3,$AD,$65,$F8,$18,$C5,$7B,$6B,$E9,$E3,$F7
	.dc.b	$6B,$58,$F8,$67,$27,$B2,$33,$3E,$88,$F8,$B8,$7A,$DF,$2D,$F5,$CB
	.dc.b	$EA,$62,$C7,$F0,$D5,$FE,$2D,$8C,$61,$FE,$2F,$8F,$F2,$7C,$61,$C3
	.dc.b	$5F,$F1,$A7,$73,$C8,$7A,$3C,$67,$FC,$03,$18,$7A,$4E,$E7,$4D,$8E
	.dc.b	$BE,$83,$C2,$E5,$82,$31,$3A,$9C,$C6,$23,$D4,$FF,$A3,$CF,$F3,$9B
	.dc.b	$FE,$81,$D3,$FD,$0E,$7F,$9D,$6B,$FD,$0F,$1F,$E7,$F8,$FF,$41,$BB
	.dc.b	$3F,$D0,$67,$FA,$1C,$FF,$5B,$9F,$E9,$7D,$7F,$D0,$31,$CE,$BB,$CE
	.dc.b	$63,$BB,$CC,$FF,$41,$BB,$BF,$F5,$A3,$63,$FD,$3A,$6E,$FC,$AF,$84
	.dc.b	$FF,$5E,$5F,$FA,$3E,$B8,$C6,$EF,$C3,$5F,$E9,$79,$FF,$56,$B6,$7F
	.dc.b	$D4,$67,$94,$F6,$D7,$37,$C2,$64,$F3,$9D,$67,$BB,$99,$F2,$CF,$13
	.dc.b	$FD,$4F,$5F,$46,$B7,$3E,$19,$F8,$F5,$CC,$EF,$19,$BF,$8F,$1D,$6F
	.dc.b	$D6,$73,$76,$DA,$EF,$7F,$EA,$47,$FD,$85,$E8,$FF,$AD,$6F,$9E,$1D
	.dc.b	$70,$F7,$C7,$CF,$AE,$31,$FE,$F7,$AE,$0E,$8F,$34,$68,$EA,$FE,$1C
	.dc.b	$FC,$1D,$75,$7E,$39,$E9,$BF,$83,$87,$9F,$0F,$C6,$CE,$4F,$5C,$FA
	.dc.b	$63,$F2,$98,$78,$A7,$9D,$F5,$3E,$7C,$FA,$4E,$6E,$6F,$E3,$DF,$3A
	.dc.b	$FC,$55,$F3,$AE,$31,$EC,$BE,$37,$E5,$8F,$5B,$5F,$C7,$C7,$77,$37
	.dc.b	$D5,$FC,$15,$F5,$AF,$0C,$7C,$B3,$E3,$1C,$8F,$01,$C8,$D1,$F9,$DB
	.dc.b	$4F,$3A,$F9,$27,$78,$F8,$51,$99,$F9,$27,$2C,$3F,$25,$87,$B2,$F8
	.dc.b	$61,$FE,$F8,$AF,$1E,$1D,$9B,$87,$3C,$3E,$F6,$1F,$4A,$7D,$DE,$8F
	.dc.b	$FB,$9D,$FF,$78,$FA,$09,$FF,$71,$B6,$3F,$D8,$53,$B7,$FD,$CC,$7F
	.dc.b	$BE,$BD,$1F,$F8,$01,$1F,$EF,$B8,$FF,$77,$1F,$F0,$02,$3F,$DD,$B1
	.dc.b	$C4,$E5,$3F,$EA,$BE,$AF,$F6,$F7,$A3,$FE,$D6,$F8,$59,$87,$FB,$47
	.dc.b	$8F,$FD,$AE,$FF,$BA,$E7,$FD,$DC,$47,$06,$DF,$F6,$E3,$FE,$D9,$F4
	.dc.b	$B8,$BF,$E8,$99,$B7,$FE,$FD,$9F,$F7,$4F,$A5,$FE,$D6,$E7,$FF,$67
	.dc.b	$C7,$FB,$78,$FF,$69,$CF,$FB,$66,$FF,$A1,$57,$FB,$7D,$FF,$7D,$35
	.dc.b	$11,$C3,$79,$79,$DE,$A1,$3E,$17,$8F,$A5,$FE,$DF,$7F,$DA,$FE,$8F
	.dc.b	$F5,$15,$E8,$7F,$B8,$5F,$EA,$8D,$7F,$AF,$35,$EC,$70,$F4,$39,$47
	.dc.b	$FB,$86,$38,$C6,$0E,$DD,$1F,$ED,$88,$F5,$B5,$FF,$70,$DF,$FB,$F9
	.dc.b	$FF,$6A,$C6,$5F,$E3,$07,$FB,$60,$BD,$1F,$EE,$96,$F9,$DF,$EE,$18
	.dc.b	$FF,$76,$79,$9F,$ED,$39,$F3,$B5,$FF,$6B,$F9,$DF,$E9,$07,$FD,$C4
	.dc.b	$D8,$FF,$6B,$EB,$FD,$D5,$B3,$FD,$FD,$AF,$4F,$2F,$97,$91,$FE,$C9
	.dc.b	$E1,$5F,$D3,$8C,$3F,$6A,$3F,$DD,$5E,$4F,$FB,$49,$DD,$FD,$EA,$FF
	.dc.b	$69,$3F,$39,$AF,$3F,$2B,$B9,$2F,$FD,$A6,$73,$3B,$79,$3F,$ED,$5B
	.dc.b	$EB,$4F,$23,$FD,$9F,$18,$6D,$70,$F2,$BF,$D9,$CF,$FB,$1B,$CA,$B6
	.dc.b	$BF,$F7,$25,$62,$3A,$4F,$CB,$BF,$F6,$64,$7E,$61,$AE,$AF,$FC,$41
	.dc.b	$1E,$08,$E9,$83,$E9,$7B,$E3,$FE,$F1,$7F,$D9,$6F,$AB,$C9,$FF,$67
	.dc.b	$BF,$EE,$2F,$2F,$38,$FF,$B2,$CF,$F1,$64,$FF,$B2,$B0,$34,$9E,$5F
	.dc.b	$D0,$B9,$EF,$3E,$23,$9D,$FB,$E3,$18,$7D,$93,$FC,$8D,$E6,$FF,$96
	.dc.b	$A7,$FC,$97,$E3,$7F,$95,$7C,$9F,$F2,$BC,$7F,$A2,$7C,$DF,$F2,$5F
	.dc.b	$87,$FE,$A0,$C3,$FC,$A7,$3F,$E5,$03,$58,$73,$7F,$CD,$F8,$73,$AF
	.dc.b	$F2,$E5,$F2,$F3,$7F,$D2,$1E,$57,$F9,$64,$FF,$47,$CF,$F9,$BD,$B1
	.dc.b	$FE,$65,$BF,$F4,$11,$DC,$FF,$3E,$57,$37,$FD,$1A,$7F,$9E,$A3,$FC
	.dc.b	$CC,$F8,$8C,$1C,$72,$FF,$62,$4F,$1A,$FF,$43,$C7,$FA,$62,$75,$7E
	.dc.b	$F7,$C3,$3E,$47,$FC,$E8,$63,$0F,$F5,$21,$F3,$32,$F8,$B3,$B3,$FD
	.dc.b	$07,$3B,$3F,$D1,$B8,$FB,$63,$E1,$57,$29,$F9,$75,$CA,$31,$1F,$03
	.dc.b	$2F,$9D,$8F,$D6,$8C,$56,$DE,$A7,$9A,$FA,$B8,$7C,$66,$1F,$F4,$7B
	.dc.b	$8B,$FB,$18,$7C,$AB,$EC,$7C,$8C,$3C,$38,$FD,$C6,$CE,$E3,$18,$F8
	.dc.b	$F1,$EB,$87,$C5,$7E,$11,$83,$DE,$63,$2F,$A9,$1F,$EA,$DC,$7E,$A6
	.dc.b	$5F,$2E,$39,$3E,$99,$DD,$7E,$68,$FB,$F1,$B7,$AB,$BC,$BC,$F1,$F7
	.dc.b	$B2,$FC,$9D,$FC,$08,$C1,$F8,$C3,$FD,$7B,$97,$B1,$F3,$57,$2B,$ED
	.dc.b	$1F,$48,$CE,$5F,$07,$1F,$CB,$47,$FA,$C6,$FF,$D6,$6F,$F5,$FC,$BF
	.dc.b	$D5,$CD,$8C,$BF,$D3,$BC,$4F,$F5,$63,$FE,$B2,$6B,$FD,$9F,$BF,$FD
	.dc.b	$58,$E7,$CD,$FF,$53,$F7,$C8,$FF,$55,$10,$FF,$54,$BF,$E2,$83,$FD
	.dc.b	$13,$E6,$FF,$90,$F1,$FF,$14,$4F,$F8,$DF,$CD,$FF,$14,$6F,$FE,$2F
	.dc.b	$8F,$F1,$93,$FE,$47,$7F,$C8,$07,$F9,$7F,$6F,$F1,$6B,$FE,$54,$3F
	.dc.b	$C6,$24,$3A,$15,$FF,$1D,$BF,$E4,$77,$FC,$E2,$E8,$8E,$9E,$F5,$FE
	.dc.b	$63,$C3,$FC,$93,$1E,$43,$FC,$A2,$6B,$0F,$66,$1E,$A8,$ED,$87,$F9
	.dc.b	$05,$EA,$FA,$EF,$E0,$E1,$DD,$B1,$9A,$E5,$5D,$70,$F9,$4F,$FA,$C2
	.dc.b	$BC,$8F,$F9,$62,$FD,$67,$3A,$E3,$3D,$27,$CE,$D7,$E5,$5F,$CF,$AE
	.dc.b	$2F,$86,$3E,$43,$5F,$49,$CA,$BA,$91,$D9,$92,$9D,$8F,$86,$7D,$73
	.dc.b	$99,$93,$DC,$7C,$D3,$15,$E7,$3D,$F5,$F8,$08,$FF,$43,$9D,$A3,$35
	.dc.b	$EB,$3E,$84,$FC,$39,$D6,$1F,$59,$B8,$7C,$C9,$FF,$65,$C7,$83,$0F
	.dc.b	$63,$E6,$C3,$AB,$EA,$3C,$26,$6F,$A3,$6F,$A6,$1F,$EC,$89,$CE,$1F
	.dc.b	$E7,$DA,$FF,$3D,$F1,$FF,$84,$3C,$BC,$36,$FF,$CF,$3E,$6C,$AF,$ED
	.dc.b	$87,$FA,$22,$3F,$D0,$E6,$5F,$F0,$87,$0F,$FA,$09,$FF,$43,$1F,$E8
	.dc.b	$32,$1F,$E5,$41,$C0,$F2,$FF,$D0,$31,$C8,$FF,$35,$F9,$6F,$33,$FC
	.dc.b	$FB,$1F,$E7,$B9,$DE,$5F,$E7,$B9,$CB,$FE,$8C,$9F,$F4,$21,$0E,$82
	.dc.b	$BF,$D0,$78,$7F,$9C,$F0,$B3,$E0,$C3,$FC,$EE,$FF,$9D,$2B,$FD,$08
	.dc.b	$FC,$23,$FC,$F3,$C7,$A5,$E4,$75,$74,$D7,$C3,$E6,$7F,$9F,$DE,$27
	.dc.b	$37,$06,$63,$FD,$05,$1E,$8F,$2F,$FC,$FD,$9F,$F9,$1B,$D3,$79,$5F
	.dc.b	$E8,$B8,$C3,$75,$FF,$47,$57,$06,$F9,$99,$C3,$FD,$2F,$83,$E0,$31
	.dc.b	$08,$C3,$EF,$BF,$F4,$43,$F2,$BA,$BF,$D0,$84,$CF,$FD,$09,$CE,$F3
	.dc.b	$7F,$9C,$46,$1F,$F4,$01,$2B,$FD,$2B,$CF,$FE,$A8,$AF,$F4,$34,$23
	.dc.b	$13,$FE,$82,$D3,$F2,$3B,$3D,$0F,$F2,$6C,$66,$FF,$11,$39,$F0,$BF
	.dc.b	$C9,$13,$FE,$AF,$F3,$BF,$D0,$B3,$FE,$4B,$F4,$9F,$15,$AC,$3F,$D7
	.dc.b	$35,$AB,$FF,$41,$56,$9F,$C0,$79,$DF,$E8,$6A,$FF,$41,$79,$DC,$71
	.dc.b	$FF,$43,$D6,$A3,$FC,$FD,$3F,$E6,$88,$B2,$F2,$FF,$D0,$11,$FE,$87
	.dc.b	$3F,$D0,$A7,$12,$CF,$E6,$3C,$CF,$F3,$93,$FE,$75,$3D,$F2,$F3,$BF
	.dc.b	$CF,$D3,$A3,$FC,$F1,$7F,$E8,$2B,$FC,$C4,$67,$CE,$F4,$5F,$F9,$EE
	.dc.b	$73,$18,$F3,$BF,$CF,$B3,$9C,$7C,$53,$9A,$FD,$17,$9B,$FE,$76,$7F
	.dc.b	$CF,$7C,$1F,$EB,$78,$C5,$7D,$4F,$33,$FC,$EA,$FE,$CA,$3E,$CB,$FE
	.dc.b	$79,$F3,$7F,$CF,$7A,$2B,$6F,$FA,$2E,$FE,$74,$63,$0F,$F1,$2F,$89
	.dc.b	$FE,$1B,$E6,$FF,$0A,$F9,$1F,$E3,$99,$FF,$0A,$79,$9F,$E1,$9A,$FF
	.dc.b	$26,$CE,$73,$FF,$0C,$92,$FF,$C5,$17,$C0,$91,$9A,$EC,$72,$F1,$7F
	.dc.b	$CD,$15,$C2,$74,$7F,$8D,$27,$5E,$6F,$F8,$C2,$30,$7F,$95,$63,$D2
	.dc.b	$E0,$FF,$23,$1F,$E2,$D9,$E8,$48,$F0,$C7,$13,$FC,$7E,$E1,$E3,$87
	.dc.b	$B6,$78,$47,$5F,$37,$FC,$92,$FF,$A1,$CF,$F1,$FD,$FF,$96,$C8,$FF
	.dc.b	$91,$27,$FD,$0F,$87,$FA,$5F,$0F,$5C,$FF,$91,$F0,$FF,$2B,$C6,$2B
	.dc.b	$77,$FE,$7F,$8F,$26,$1E,$0A,$FF,$36,$E1,$F2,$27,$39,$7D,$5C,$7D
	.dc.b	$87,$C1,$8C,$E5,$F4,$B2,$F5,$CF,$A0,$F3,$4F,$F9,$A6,$FC,$D1,$CB
	.dc.b	$0F,$85,$3C,$A3,$EA,$9C,$6B,$F0,$F5,$E3,$8F,$8D,$87,$8D,$F9,$71
	.dc.b	$97,$85,$7B,$CF,$06,$5E,$F8,$E3,$8F,$9E,$7C,$18,$7A,$31,$F8,$F5
	.dc.b	$E5,$C3,$A4,$62,$3B,$E1,$F8,$77,$B9,$B7,$FB,$42,$8F,$F3,$5A,$31
	.dc.b	$3F,$E6,$C6,$CF,$0E,$EF,$F3,$4C,$7F,$9A,$FC,$5F,$F3,$6B,$FE,$71
	.dc.b	$F3,$75,$18,$8E,$DE,$63,$C0,$FF,$37,$47,$F9,$B0,$FF,$37,$1F,$E3
	.dc.b	$5C,$FF,$C3,$CF,$F8,$44,$FF,$15,$67,$FE,$18,$7F,$C5,$B1,$FE,$3E
	.dc.b	$7F,$C6,$C6,$DC,$3F,$E2,$13,$FC,$4A,$72,$AF,$F0,$CE,$9D,$C9,$1C
	.dc.b	$A7,$C2,$48,$CC,$73,$3B,$3F,$E3,$43,$35,$E5,$3F,$C7,$D5,$DA,$7A
	.dc.b	$4E,$A3,$11,$CE,$BB,$3A,$21,$99,$E1,$7F,$E3,$28,$FA,$D6,$BF,$99
	.dc.b	$1F,$E6,$9A,$F2,$3C,$A7,$C9,$E6,$FF,$90,$B0,$EF,$3D,$4D,$5F,$92
	.dc.b	$B7,$1F,$22,$7E,$5C,$62,$3E,$09,$FE,$61,$8E,$71,$96,$CF,$32,$61
	.dc.b	$E9,$AF,$C1,$BF,$E9,$29,$EB,$3A,$9F,$AD,$3B,$AF,$73,$BC,$3D,$B3
	.dc.b	$E9,$24,$F8,$63,$E3,$C7,$22,$5F,$6A,$E9,$7F,$22,$3F,$CE,$57,$DA
	.dc.b	$BF,$D2,$93,$E3,$9E,$11,$FE,$7B,$FC,$31,$EA,$FE,$AA,$11,$87,$F5
	.dc.b	$AD,$7F,$CA,$CD,$7F,$5C,$7E,$BD,$D7,$2B,$87,$31,$86,$D7,$F8,$B3
	.dc.b	$DC,$7A,$FF,$E2,$EA,$D6,$1C,$EB,$06,$E7,$FC,$D9,$1E,$01,$EB,$3D
	.dc.b	$1F,$53,$63,$FC,$DA,$6D,$C3,$FE,$60,$6E,$5E,$78,$FE,$11,$18,$CB
	.dc.b	$DB,$E9,$7F,$8E,$67,$FC,$BD,$39,$77,$7F,$E7,$FB,$FF,$44,$79,$17
	.dc.b	$0F,$F4,$04,$75,$AF,$F4,$6C,$74,$9F,$F3,$AB,$F2,$BD,$0F,$F3,$E4
	.dc.b	$FF,$9A,$DA,$FF,$9F,$4F,$F3,$C5,$7F,$9E,$DC,$1F,$E7,$A8,$FF,$3B
	.dc.b	$BF,$E8,$AF,$47,$AB,$FE,$79,$3F,$CB,$18,$7F,$97,$3C,$CF,$54,$1F
	.dc.b	$65,$BF,$64,$7C,$06,$FD,$89,$7F,$E6,$17,$FC,$CB,$E8,$3D,$AF,$F3
	.dc.b	$3F,$6F,$F9,$87,$1F,$F2,$BD,$FF,$33,$8C,$3F,$CA,$77,$FF,$30,$3F
	.dc.b	$8A,$E6,$DB,$4D,$1F,$E6,$42,$8F,$3B,$FC,$A8,$FF,$95,$BD,$9F,$F2
	.dc.b	$87,$9E,$F6,$BF,$CB,$53,$B9,$B5,$FE,$2E,$8E,$5E,$3F,$F9,$5F,$CF
	.dc.b	$7B,$DA,$8F,$F2,$4F,$99,$FE,$44,$F4,$2F,$4F,$F9,$10,$FF,$38,$F9
	.dc.b	$DF,$E4,$F7,$18,$FF,$91,$1C,$F9,$FF,$E4,$83,$C7,$E7,$F8,$4F,$F1
	.dc.b	$FF,$1F,$F9,$87,$DA,$CE,$72,$FF,$C9,$6F,$F9,$52,$31,$87,$F9,$1E
	.dc.b	$3F,$C8,$B8,$A7,$37,$FE,$45,$BF,$F2,$89,$2F,$FD,$9D,$83,$0F,$CA
	.dc.b	$47,$EE,$BF,$0B,$F5,$3F,$EC,$46,$DF,$E9,$47,$E1,$3F,$ED,$5C,$3F
	.dc.b	$D8,$13,$A7,$96,$9F,$42,$31,$6B,$E3,$DD,$9A,$97,$FD,$EA,$FC,$37
	.dc.b	$FB,$D6,$BD,$23,$FB,$FF,$8F,$FE,$1A,$8F,$F0,$69,$FE,$1A,$7F,$C1
	.dc.b	$1E,$87,$F8,$B7,$CF,$79,$FF,$E3,$AE,$9F,$F0,$EB,$FE,$0B,$AD,$FA
	.dc.b	$1F,$E1,$DB,$E1,$3B,$F3,$7F,$C3,$98,$FF,$56,$C7,$9B,$FE,$2E,$31
	.dc.b	$3B,$7F,$C2,$87,$F9,$0E,$3F,$C6,$2F,$F8,$68,$C4,$7F,$89,$5F,$F0
	.dc.b	$D9,$9C,$7F,$C8,$EF,$FA,$FE,$BF,$CC,$6E,$AF,$CA,$F8,$4F,$F2,$C6
	.dc.b	$3E,$AC,$BF,$C5,$D1,$F1,$2F,$FC,$5D,$8F,$8A,$F8,$C7,$C2,$9F,$19
	.dc.b	$DC,$FF,$34,$9C,$70,$FF,$1B,$91,$C1,$0F,$26,$5F,$24,$98,$78,$A3
	.dc.b	$2D,$C7,$CD,$1F,$5F,$1F,$25,$70,$8C,$65,$F1,$6B,$78,$FA,$A3,$17
	.dc.b	$C3,$0E,$8D,$BE,$27,$AC,$FF,$34,$1E,$7B,$E7,$8F,$AE,$3C,$A7,$0C
	.dc.b	$11,$F4,$27,$73,$C6,$BB,$57,$8E,$3B,$4F,$E4,$DE,$71,$93,$9B,$F3
	.dc.b	$0C,$D7,$F6,$89,$57,$E3,$2B,$75,$FE,$56,$3F,$AF,$14,$FE,$C1,$3F
	.dc.b	$E5,$E7,$07,$F9,$67,$D0,$FF,$2B,$91,$E3,$DD,$FE,$5E,$7F,$C5,$FE
	.dc.b	$6F,$F8,$B9,$FF,$2C,$10,$F2,$39,$7F,$C2,$37,$FE,$08,$26,$3F,$E1
	.dc.b	$8C,$3F,$C2,$D3,$FE,$30,$F3,$BF,$C2,$67,$F8,$75,$FF,$16,$4F,$23
	.dc.b	$85,$FF,$84,$DB,$3C,$27,$FC,$80,$70,$3A,$1F,$E2,$08,$DB,$4D,$1F
	.dc.b	$E2,$89,$E8,$E2,$BC,$AE,$6B,$B5,$FF,$89,$AB,$FC,$51,$7F,$02,$7D
	.dc.b	$15,$FE,$3C,$3F,$C5,$2F,$08,$FB,$1E,$7F,$F9,$45,$F6,$4F,$7A,$F0
	.dc.b	$BC,$6F,$ED,$E1,$FE,$33,$7C,$18,$7F,$90,$F0,$E5,$5E,$E9,$F9,$AF
	.dc.b	$39,$F0,$3F,$E6,$69,$DB,$EB,$3E,$11,$DF,$0F,$53,$F2,$8C,$1F,$80
	.dc.b	$9F,$6D,$F8,$E7,$77,$E4,$AF,$79,$D2,$74,$F1,$3A,$38,$7A,$9F,$E9
	.dc.b	$A9,$FA,$B3,$EA,$8C,$CE,$CF,$F4,$5D,$7B,$23,$81,$EF,$3C,$18,$7D
	.dc.b	$47,$C1,$5E,$53,$E5,$C6,$63,$89,$D9,$C5,$76,$8E,$75,$F3,$DF,$F4
	.dc.b	$15,$F9,$88,$F8,$6F,$D1,$5E,$7B,$E7,$87,$D0,$8F,$C5,$92,$3E,$FB
	.dc.b	$BC,$3E,$6C,$78,$8E,$37,$FE,$D3,$BF,$DE,$F9,$DF,$E5,$E8,$79,$EF
	.dc.b	$63,$FC,$C8,$7F,$98,$23,$FD,$B5,$CD,$FE,$49,$AF,$F3,$84,$7F,$9A
	.dc.b	$8F,$F3,$AF,$A1,$FE,$79,$8C,$C7,$F9,$A4,$E3,$C3,$B7,$FB,$93,$6B
	.dc.b	$FC,$A9,$87,$F9,$CA,$7F,$CB,$BE,$67,$F9,$F2,$7F,$CB,$1E,$7F,$F1
	.dc.b	$46,$F0,$7F,$97,$6B,$FC,$C2,$7F,$9B,$8F,$F4,$1F,$9F,$FE,$6E,$7F
	.dc.b	$CE,$17,$FE,$74,$38,$3F,$E6,$23,$FD,$ED,$7F,$E6,$D7,$FC,$E9,$1A
	.dc.b	$9F,$F2,$AC,$E6,$FF,$CC,$07,$F9,$AF,$CF,$FF,$32,$B7,$DC,$FF,$2F
	.dc.b	$E1,$FE,$61,$3F,$94,$94,$FE,$57,$CC,$EC,$FF,$50,$36,$36,$FF,$9A
	.dc.b	$8C,$3F,$E6,$D8,$FF,$34,$65,$FE,$A7,$F4,$3D,$D3,$FE,$A2,$6A,$3F
	.dc.b	$CD,$91,$BF,$3A,$47,$92,$51,$88,$FE,$6F,$EC,$FF,$89,$DF,$F2,$FF
	.dc.b	$4F,$F9,$A3,$9F,$E3,$29,$D4,$7C,$99,$FC,$71,$FE,$66,$2B,$95,$CD
	.dc.b	$FE,$6E,$73,$E6,$FF,$AE,$BC,$A7,$97,$FE,$44,$9F,$F3,$67,$37,$CA
	.dc.b	$9C,$B6,$75,$3F,$E6,$A8,$D7,$89,$8A,$FF,$27,$46,$0F,$A7,$EB,$3C
	.dc.b	$AF,$F3,$AB,$63,$FC,$D7,$E2,$FF,$9B,$B0,$FE,$C3,$39,$6D,$78,$23
	.dc.b	$11,$FE,$C3,$75,$CB,$FE,$57,$F2,$5E,$57,$F9,$C7,$1F,$F3,$59,$2F
	.dc.b	$F4,$67,$F9,$56,$7F,$CD,$51,$9A,$FF,$35,$4F,$F9,$BE,$3F,$CF,$FE
	.dc.b	$57,$F9,$B9,$FF,$70,$F9,$5E,$E9,$FF,$36,$EB,$F1,$23,$11,$FE,$D7
	.dc.b	$6B,$C2,$3F,$CD,$93,$FB,$EF,$55,$EB,$75,$AF,$F3,$44,$62,$7F,$CD
	.dc.b	$FB,$7F,$9A,$27,$D3,$97,$F9,$B2,$72,$48,$FF,$04,$F9,$7F,$E6,$A7
	.dc.b	$F2,$8D,$4D,$3F,$CE,$F2,$FC,$33,$FD,$55,$5C,$E3,$FD,$05,$18,$AF
	.dc.b	$92,$7F,$8B,$8F,$F0,$E9,$FE,$27,$F2,$BF,$C6,$E4,$7F,$C4,$FD,$BF
	.dc.b	$E1,$FF,$0B,$FC,$44,$D8,$FF,$34,$F3,$FF,$98,$8F,$ED,$31,$98,$FF
	.dc.b	$2F,$BF,$E3,$29,$FF,$1D,$79,$1F,$E3,$69,$F0,$91,$C1,$27,$84,$79
	.dc.b	$4F,$1D,$FF,$A0,$E3,$31,$C5,$B3,$C6,$7D,$57,$E4,$78,$18,$BF,$F2
	.dc.b	$45,$70,$24,$66,$FF,$C8,$57,$CD,$F0,$9F,$E7,$2B,$E8,$DB,$E7,$5C
	.dc.b	$63,$E7,$57,$1A,$D9,$F1,$EB,$73,$D2,$30,$F8,$2F,$C6,$E6,$7E,$8D
	.dc.b	$FF,$A6,$63,$FD,$25,$18,$7F,$D1,$D7,$FE,$56,$9D,$79,$9F,$E8,$AA
	.dc.b	$FF,$52,$D7,$D4,$9F,$35,$7F,$98,$CF,$34,$F9,$30,$F3,$BD,$5F,$F4
	.dc.b	$91,$E6,$AF,$6E,$1E,$68,$DC,$7F,$AB,$8F,$1C,$7E,$0F,$0F,$14,$7E
	.dc.b	$22,$BA,$1E,$4B,$F4,$47,$0C,$3D,$3E,$67,$FA,$3F,$1F,$F3,$B5,$F9
	.dc.b	$49,$5F,$1E,$BD,$78,$FE,$47,$0E,$77,$E8,$BE,$F7,$FE,$C0,$3F,$CE
	.dc.b	$B5,$FE,$C6,$6D,$FF,$B6,$23,$11,$FE,$77,$3F,$CE,$BE,$BB,$CD,$FF
	.dc.b	$3C,$47,$F4,$93,$FA,$57,$7F,$F9,$E5,$FF,$2A,$1F,$E7,$73,$FC,$68
	.dc.b	$53,$FC,$89,$8F,$F8,$6B,$CC,$FF,$09,$77,$FF,$85,$0F,$F1,$14,$FF
	.dc.b	$A6,$23,$FC,$32,$69,$FF,$1E,$B6,$7F,$C6,$19,$FF,$91,$8F,$F3,$0C
	.dc.b	$FF,$8A,$8F,$F1,$43,$FE,$2A,$7F,$C6,$55,$DD,$FF,$25,$CF,$8E,$3C
	.dc.b	$2F,$F8,$E9,$FF,$21,$69,$FE,$37,$8D,$4F,$0A,$E9,$3F,$E2,$D8,$F2
	.dc.b	$D7,$29,$FF,$38,$C7,$80,$C9,$C4,$F2,$90,$D5,$7B,$1F,$F2,$6C,$7B
	.dc.b	$1F,$89,$3E,$4F,$37,$FC,$81,$87,$C6,$8F,$04,$FF,$9D,$EB,$CB,$3F
	.dc.b	$1D,$E5,$7E,$D9,$FF,$28,$9E,$33,$C3,$1A,$3D,$38,$7B,$67,$C9,$3F
	.dc.b	$E6,$8C,$3C,$38,$78,$E3,$E8,$1C,$8F,$F4,$21,$8A,$E5,$7D,$63,$CC
	.dc.b	$7C,$18,$CB,$C0,$F7,$9E,$6B,$F4,$1E,$28,$FF,$3C,$5F,$8A,$75,$5F
	.dc.b	$83,$79,$57,$08,$F8,$77,$F1,$A7,$F2,$71,$87,$E1,$5F,$89,$FF,$3C
	.dc.b	$B4,$FC,$14,$6E,$BA,$4F,$68,$F8,$F8,$7C,$F8,$C1,$EE,$3E,$21,$D1
	.dc.b	$D3,$6F,$C9,$7F,$EC,$4F,$2F,$FC,$E5,$E5,$FF,$9C,$6B,$FC,$CE,$7F
	.dc.b	$8E,$5C,$3F,$D7,$A3,$FC,$F0,$70,$F3,$7F,$CE,$F1,$C0,$46,$BC,$DC
	.dc.b	$C7,$F6,$6A,$E7,$CB,$FE,$7B,$9F,$F3,$1F,$99,$FE,$6D,$E7,$FE,$00
	.dc.b	$DB,$FF,$36,$9F,$E6,$43,$B3,$94,$DF,$35,$E5,$7F,$9E,$6B,$FC,$E6
	.dc.b	$FF,$9C,$6B,$FC,$EC,$74,$3F,$CE,$A7,$FA,$00,$87,$F7,$CA,$FF,$38
	.dc.b	$4F,$F1,$3E,$5F,$F3,$1E,$3F,$E6,$FA,$F8,$1E,$6F,$F9,$CA,$B5,$E6
	.dc.b	$7F,$9C,$A3,$13,$FE,$63,$6B,$FE,$73,$AF,$F3,$DD,$70,$8F,$95,$7D
	.dc.b	$A3,$F9,$15,$59,$FF,$36,$79,$9E,$E7,$EB,$DB,$F6,$17,$CE,$FC,$19
	.dc.b	$4E,$63,$FC,$F3,$1F,$5E,$BF,$CE,$2F,$D9,$F5,$9E,$F7,$18,$46,$1E
	.dc.b	$1E,$86,$7A,$FF,$CE,$9E,$7F,$F3,$76,$CF,$F3,$93,$FC,$DE,$7F,$9C
	.dc.b	$CB,$D1,$FE,$73,$EB,$FF,$39,$46,$2B,$FC,$E5,$E7,$FF,$91,$23,$D3
	.dc.b	$EB,$3D,$CC,$43,$D1,$FF,$39,$BF,$E4,$29,$D7,$1F,$E4,$E7,$FC,$A9
	.dc.b	$C2,$8F,$F6,$5F,$13,$C0,$E7,$5A,$9C,$C7,$F9,$C6,$BF,$CD,$5E,$45
	.dc.b	$9F,$F6,$9F,$9D,$FE,$6C,$7F,$41,$E8,$7F,$9C,$1C,$FA,$1F,$E6,$FD
	.dc.b	$3F,$C8,$B3,$97,$FC,$DD,$1F,$E5,$D5,$E7,$F8,$5F,$F3,$97,$A1,$98
	.dc.b	$FF,$33,$C7,$FA,$AD,$AF,$0F,$43,$FC,$DC,$4C,$3F,$07,$E7,$FF,$99
	.dc.b	$E7,$FC,$F0,$7F,$99,$9C,$69,$FE,$7C,$3E,$24,$7F,$9B,$BD,$0F,$F3
	.dc.b	$84,$27,$75,$FD,$CF,$D0,$FF,$32,$FA,$1F,$E6,$1A,$E3,$87,$F9,$AB
	.dc.b	$7F,$B1,$1F,$E6,$9F,$5F,$FC,$EC,$FF,$9A,$1B,$87,$FB,$F2,$F9,$3D
	.dc.b	$6B,$8C,$62,$FF,$CE,$F1,$8A,$FC,$0F,$95,$FE,$2D,$F0,$BF,$C4,$DE
	.dc.b	$67,$F8,$62,$FF,$C5,$3E,$7F,$F8,$60,$FF,$2A,$BF,$E3,$F3,$FC,$55
	.dc.b	$D9,$FE,$19,$6C,$FF,$89,$EB,$FC,$59,$E8,$7F,$89,$63,$FC,$DB,$1D
	.dc.b	$A7,$FC,$48,$73,$3F,$C8,$05,$C3,$FC,$4B,$3C,$F0,$FF,$17,$E9,$FE
	.dc.b	$E3,$C7,$FC,$5F,$7C,$9F,$F1,$EB,$C9,$B7,$CA,$3F,$C7,$87,$98,$A6
	.dc.b	$EF,$E0,$47,$C3,$BF,$81,$3F,$E6,$93,$B4,$EB,$0F,$F1,$D1,$EE,$C3
	.dc.b	$C9,$1E,$EA,$E3,$87,$F9,$4F,$0E,$98,$7B,$70,$E0,$DA,$F9,$B5,$E8
	.dc.b	$9C,$BE,$E9,$FA,$66,$CF,$F2,$91,$1F,$A4,$FA,$CF,$39,$0E,$2F,$59
	.dc.b	$F9,$71,$CE,$3C,$D3,$EC,$9C,$C7,$86,$BD,$2F,$1A,$F0,$5F,$D5,$69
	.dc.b	$F3,$27,$E8,$3F,$E8,$23,$E8,$14,$F3,$1E,$48,$F9,$95,$F0,$9B,$1F
	.dc.b	$44,$D1,$EB,$9F,$1C,$6A,$BD,$35,$C6,$B5,$7F,$E6,$67,$FD,$00,$7F
	.dc.b	$9A,$8F,$E5,$82,$FF,$CF,$AF,$F2,$FE,$FF,$F3,$31,$3D,$0F,$F3,$64
	.dc.b	$7F,$9B,$BD,$0E,$2F,$F9,$A8,$FF,$37,$53,$D0,$78,$5A,$8C,$36,$33
	.dc.b	$3E,$29,$FF,$0C,$46,$33,$FF,$14,$37,$4F,$F1,$95,$7F,$88,$8F,$F1
	.dc.b	$94,$7F,$85,$2B,$FC,$3F,$18,$BF,$F0,$AC,$FF,$93,$E3,$FC,$4E,$EF
	.dc.b	$B3,$FC,$65,$87,$F9,$98,$FF,$12,$CF,$F8,$74,$DD,$72,$3F,$C9,$0F
	.dc.b	$F8,$FA,$38,$61,$FE,$39,$8E,$11,$CB,$0F,$1C,$7F,$8A,$9A,$7F,$8D
	.dc.b	$6F,$C3,$5F,$E3,$53,$94,$6B,$0F,$35,$7F,$94,$DA,$F8,$67,$83,$FE
	.dc.b	$70,$3F,$C7,$27,$C5,$3C,$15,$D9,$F8,$87,$B0,$97,$E7,$8D,$47,$03
	.dc.b	$C1,$87,$C1,$7F,$CA,$E7,$73,$DF,$5E,$89,$E3,$87,$B6,$78,$9D,$A7
	.dc.b	$4E,$2F,$C0,$D8,$CB,$ED,$3C,$F1,$A7,$07,$94,$B3,$DD,$B8,$7C,$69
	.dc.b	$FA,$4D,$AF,$F2,$FE,$1E,$B9,$FB,$35,$EB,$BF,$A1,$39,$AF,$9B,$3F
	.dc.b	$E8,$93,$CC,$6F,$0F,$7E,$1D,$63,$FC,$D1,$7E,$3C,$3C,$38,$7F,$9E
	.dc.b	$7C,$FF,$F2,$F5,$7F,$97,$7C,$FF,$F2,$CF,$07,$FB,$56,$7F,$CB,$D3
	.dc.b	$9F,$43,$FC,$BF,$CD,$FE,$5D,$8F,$F3,$24,$F2,$7F,$CC,$53,$FE,$2D
	.dc.b	$87,$A1,$FE,$62,$DB,$FC,$61,$E7,$7F,$96,$FC,$FC,$61,$FE,$60,$9F
	.dc.b	$F1,$A7,$A1,$FE,$41,$8F,$F3,$21,$FE,$57,$8E,$26,$A6,$47,$F9,$80
	.dc.b	$4F,$17,$FC,$C2,$7F,$0C,$E5,$FF,$27,$2F,$3F,$FC,$AE,$7F,$96,$23
	.dc.b	$26,$9C,$FA,$9F,$E4,$78,$CF,$A3,$FC,$3A,$90,$8C,$36,$3F,$89,$B5
	.dc.b	$FE,$3B,$E8,$7B,$77,$FF,$0A,$9F,$E5,$98,$FF,$2F,$F9,$3F,$E5,$E7
	.dc.b	$FC,$B5,$9F,$F9,$5E,$BF,$CB,$71,$FE,$60,$E1,$33,$CF,$FF,$2D,$39
	.dc.b	$31,$E8,$71,$F4,$3F,$CB,$9E,$A7,$F9,$A2,$7F,$91,$FA,$3F,$E5,$D8
	.dc.b	$C7,$42,$31,$AF,$68,$C5,$78,$3B,$D7,$C2,$FF,$9C,$F5,$7F,$98,$09
	.dc.b	$5F,$E6,$57,$94,$7F,$98,$B4,$FF,$3E,$F9,$FF,$E5,$CA,$F4,$1F,$E6
	.dc.b	$1F,$47,$FC,$5F,$19,$CF,$FC,$67,$E8,$7F,$92,$DF,$F4,$27,$81,$FE
	.dc.b	$5E,$F2,$FF,$C6,$BC,$9F,$E6,$1C,$BF,$CB,$3E,$A7,$F8,$EB,$3F,$8E
	.dc.b	$3D,$04,$66,$3E,$5C,$62,$7F,$A5,$B6,$3F,$AA,$BF,$E5,$FC,$5E,$8F
	.dc.b	$F9,$9E,$7F,$CC,$77,$FE,$46,$8C,$4F,$F9,$2A,$B8,$46,$2F,$C7,$18
	.dc.b	$F4,$7F,$D0,$3E,$87,$F9,$88,$99,$FF,$98,$72,$FE,$CB,$1F,$07,$CF
	.dc.b	$FF,$2C,$C7,$F9,$A6,$7F,$CA,$B1,$88,$FF,$35,$61,$FE,$68,$F4,$3F
	.dc.b	$CC,$33,$9B,$E5,$18,$3F,$CB,$35,$FE,$AC,$F3,$FF,$CA,$CF,$F9,$86
	.dc.b	$FF,$D6,$71,$8F,$43,$FC,$CF,$E7,$F0,$DB,$FC,$C9,$5F,$EE,$79,$CB
	.dc.b	$7D,$0F,$F2,$F3,$FE,$EB,$F2,$FF,$C0,$F8,$FF,$81,$3C,$FF,$F0,$33
	.dc.b	$FE,$0C,$24,$69,$BE,$17,$F8,$8D,$FF,$0E,$9F,$E1,$98,$FF,$0E,$FA
	.dc.b	$8F,$47,$FC,$BD,$3F,$E1,$E9,$E4,$4A,$E5,$D9,$FE,$36,$8E,$35,$C2
	.dc.b	$BF,$C2,$55,$FE,$65,$7C,$38,$79,$AB,$93,$67,$9F,$93,$FE,$31,$C3
	.dc.b	$A4,$7F,$92,$23,$FC,$A5,$7F,$E3,$D7,$44,$7F,$C6,$D1,$B7,$D0,$F4
	.dc.b	$73,$5F,$E4,$29,$E7,$5F,$E4,$13,$BC,$66,$78,$BE,$52,$3F,$E3,$9B
	.dc.b	$F4,$B6,$FB,$3F,$2D,$F1,$3F,$0A,$7F,$CD,$37,$FE,$2C,$9C,$D7,$F9
	.dc.b	$30,$E7,$7F,$E4,$0B,$E9,$7E,$08,$C1,$2F,$94,$79,$E7,$B1,$F1,$4F
	.dc.b	$79,$F7,$B1,$FB,$99,$7F,$9F,$23,$DA,$76,$25,$F5,$7F,$CB,$B3,$C1
	.dc.b	$E8,$7F,$97,$B0,$F3,$E1,$E2,$8E,$D9,$7E,$06,$7D,$99,$7C,$4B,$E5
	.dc.b	$87,$F9,$C6,$31,$5B,$7F,$0D,$5D,$2F,$C3,$3F,$ED,$2A,$FF,$3B,$47
	.dc.b	$F9,$7C,$87,$84,$FF,$2F,$7A,$3F,$E5,$C6,$FA,$3F,$E5,$CF,$07,$FC
	.dc.b	$C0,$FF,$95,$FC,$29,$3F,$E5,$7F,$47,$FC,$B2,$FF,$1E,$3F,$CC,$47
	.dc.b	$F8,$09,$FF,$02,$3F,$E3,$3C,$BF,$C0,$9A,$FF,$84,$A3,$FC,$48,$7F
	.dc.b	$82,$CF,$F0,$69,$FE,$0F,$3F,$C3,$75,$FE,$13,$7F,$C4,$71,$FE,$26
	.dc.b	$CB,$FC,$1B,$1C,$8E,$0E,$0F,$F1,$41,$63,$81,$E1,$72,$D3,$31,$E9
	.dc.b	$9F,$03,$FE,$2F,$9E,$4F,$9C,$F2,$9E,$1A,$FF,$3E,$A8,$FF,$3C,$CF
	.dc.b	$E2,$63,$FC,$8E,$71,$8E,$6F,$A2,$BC,$4E,$5A,$E4,$F2,$1D,$EF,$FD
	.dc.b	$08,$FF,$98,$2F,$FD,$13,$7D,$DF,$F1,$DC,$6E,$31,$DD,$FE,$38,$7D
	.dc.b	$B3,$F5,$27,$AE,$1F,$11,$B8,$7C,$47,$ED,$61,$F7,$2B,$BC,$7B,$E7
	.dc.b	$FD,$2B,$18,$AF,$83,$39,$3C,$D5,$FE,$93,$8E,$D1,$D2,$FC,$73,$D4
	.dc.b	$D5,$F8,$6F,$A3,$DA,$33,$5F,$1C,$F3,$94,$E8,$7C,$13,$C4,$7D,$23
	.dc.b	$E6,$1A,$BF,$1D,$76,$8F,$3D,$FF,$A8,$27,$EB,$CF,$DE,$57,$FE,$B4
	.dc.b	$F3,$FF,$A1,$5F,$F9,$5B,$5F,$F2,$A1,$FE,$02,$E4,$70,$FF,$81,$3D
	.dc.b	$17,$C1,$EA,$F8,$10,$FF,$0A,$C6,$9C,$57,$27,$FC,$67,$87,$81,$FF
	.dc.b	$0D,$B6,$35,$09,$E4,$EB,$D1,$FF,$34,$11,$FF,$3A,$7B,$95,$D1,$FE
	.dc.b	$70,$8F,$56,$3E,$99,$D5,$FF,$9D,$6B,$67,$A2,$33,$3F,$E7,$DA,$B3
	.dc.b	$FE,$4A,$B5,$7F,$9B,$23,$FD,$E7,$E8,$FF,$9A,$DF,$F1,$4B,$5F,$F2
	.dc.b	$8B,$5F,$F3,$8F,$9D,$C7,$1F,$7C,$FF,$97,$F1,$FF,$18,$FA,$5F,$E4
	.dc.b	$AB,$FF,$3E,$D6,$AF,$85,$6B,$CA,$FF,$30,$7C,$34,$BF,$F2,$FD,$FC
	.dc.b	$AF,$41,$3F,$0A,$5E,$53,$F6,$0F,$9F,$7B,$CB,$FC,$D4,$4A,$E7,$97
	.dc.b	$4F,$79,$EF,$7F,$99,$AF,$FC,$AF,$5F,$E5,$FF,$43,$FC,$BC,$D8,$CA
	.dc.b	$ED,$FF,$31,$C7,$F9,$83,$AB,$DD,$3F,$E7,$6E,$9C,$70,$7F,$99,$A3
	.dc.b	$FC,$E2,$7F,$9E,$B6,$79,$3F,$E5,$BD,$7C,$B2,$C5,$EE,$BD,$DD,$4F
	.dc.b	$F9,$A3,$07,$C3,$7F,$A3,$30,$FF,$2F,$C6,$1E,$6F,$C5,$F9,$6D,$57
	.dc.b	$86,$BF,$CA,$39,$4B,$F0,$E5,$FE,$5E,$F7,$31,$7F,$E7,$6F,$37,$CF
	.dc.b	$5E,$68,$FF,$27,$9F,$E9,$1A,$E1,$19,$9D,$CF,$1B,$FF,$2E,$3F,$D5
	.dc.b	$E3,$FC,$BF,$3F,$D6,$63,$FD,$3B,$E5,$3C,$AF,$F2,$77,$BF,$FE,$AF
	.dc.b	$AF,$F2,$61,$2B,$FD,$13,$5B,$8F,$4C,$EE,$73,$5F,$E5,$E7,$FC,$B8
	.dc.b	$79,$E3,$58,$FE,$DA,$F9,$4F,$0B,$ED,$3F,$EB,$8A,$DC,$7F,$A3,$2B
	.dc.b	$73,$C7,$C6,$F9,$B6,$9F,$F4,$A3,$FE,$B5,$D9,$F8,$A7,$EC,$30,$FF
	.dc.b	$34,$5F,$0B,$DE,$1C,$E7,$EC,$D6,$F0,$F7,$DE,$C9,$3A,$EF,$B3,$FE
	.dc.b	$AC,$F1,$BF,$C2,$8F,$F8,$87,$CA,$FF,$0C,$D7,$F5,$57,$FC,$32,$62
	.dc.b	$3F,$C4,$F3,$FE,$64,$9D,$79,$9F,$E4,$29,$FF,$0E,$C6,$1F,$F1,$6F
	.dc.b	$99,$FE,$26,$F2,$7F,$CA,$A7,$F8,$B7,$A7,$FC,$69,$3E,$1B,$FF,$22
	.dc.b	$4E,$4F,$F3,$14,$7F,$94,$63,$FC,$89,$87,$F8,$C9,$DC,$78,$0F,$2E
	.dc.b	$3F,$E5,$23,$D3,$1A,$8E,$C7,$F9,$12,$BB,$3F,$E4,$CB,$FF,$33,$BF
	.dc.b	$E4,$0C,$7D,$51,$FE,$6B,$9F,$1E,$3E,$BA,$FF,$27,$CF,$AB,$1F,$89
	.dc.b	$1E,$2C,$3C,$F8,$F8,$A3,$C0,$79,$EB,$FC,$91,$3E,$5A,$E3,$18,$9D
	.dc.b	$9E,$87,$B5,$7B,$6B,$E8,$4E,$70,$F5,$BF,$1C,$F5,$1E,$32,$D7,$53
	.dc.b	$9C,$F9,$8E,$77,$E2,$AF,$AF,$5F,$EB,$33,$C7,$19,$7E,$0C,$78,$30
	.dc.b	$F8,$91,$FE,$78,$BF,$90,$FC,$D9,$F9,$3E,$4F,$F9,$97,$0E,$F8,$FA
	.dc.b	$EB,$F0,$F3,$96,$E5,$F4,$AF,$FC,$C9,$10,$1F,$E8,$63,$FC,$D2,$63
	.dc.b	$B3,$FC,$CA,$FF,$09,$6B,$FE,$63,$CF,$FC,$86,$DF,$2B,$FC,$CE,$F4
	.dc.b	$3F,$CC,$E4,$8F,$F0,$83,$FE,$15,$7F,$C1,$D5,$FE,$1F,$ED,$FF,$09
	.dc.b	$3F,$E4,$48,$CC,$FF,$85,$8F,$F2,$9E,$9F,$E3,$02,$4F,$18,$CB,$63
	.dc.b	$FC,$37,$BF,$F9,$0E,$7F,$C4,$D1,$C1,$E8,$E9,$FF,$3A,$D7,$81,$F3
	.dc.b	$61,$FE,$52,$6E,$1F,$E4,$E7,$71,$99,$FF,$29,$D7,$F8,$BA,$3F,$C7
	.dc.b	$33,$FE,$32,$9F,$82,$6B,$0F,$F2,$04,$EA,$3C,$06,$67,$FC,$7C,$F0
	.dc.b	$C3,$D5,$7D,$70,$F8,$F5,$FE,$44,$3B,$5F,$33,$C9,$5C,$30,$EF,$5C
	.dc.b	$2B,$FC,$89,$18,$7E,$49,$F1,$1F,$B5,$87,$F9,$A6,$36,$79,$2F,$B5
	.dc.b	$7C,$38,$E2,$78,$9E,$D3,$E9,$8F,$8F,$1B,$C3,$FC,$9F,$5E,$69,$C9
	.dc.b	$FE,$5A,$AF,$84,$F3,$3A,$B8,$3E,$3E,$1E,$D8,$F2,$5F,$82,$BC,$B1
	.dc.b	$CE,$74,$F3,$BF,$C0,$3D,$27,$C4,$DA,$F3,$61,$F9,$39,$CD,$7A,$27
	.dc.b	$35,$FE,$56,$CF,$FC,$50,$FF,$94,$FC,$8B,$E4,$1F,$F3,$14,$7F,$96
	.dc.b	$A7,$FB,$A3,$C4,$BE,$4F,$F9,$7D,$FF,$68,$35,$FD,$63,$FE,$61,$DA
	.dc.b	$CF,$4E,$4F,$F3,$1E,$1F,$E6,$27,$FC,$B0,$63,$DE,$F1,$9F,$E5,$6C
	.dc.b	$18,$79,$5F,$0E,$3E,$47,$2D,$8D,$47,$F8,$2D,$B5,$DB,$1F,$F3,$14
	.dc.b	$FF,$9B,$8F,$F3,$61,$0F,$F3,$4C,$79,$B9,$3F,$CA,$B1,$FC,$4B,$C9
	.dc.b	$FF,$47,$BF,$E2,$16,$BF,$E5,$CA,$FF,$33,$F8,$DE,$5F,$2B,$FC,$CC
	.dc.b	$47,$07,$49,$FF,$34,$9C,$FC,$4F,$F3,$03,$FE,$23,$F2,$BF,$CB,$EF
	.dc.b	$D2,$6B,$9F,$2B,$E1,$47,$F9,$CC,$95,$83,$E3,$73,$FF,$95,$AB,$FC
	.dc.b	$A2,$7F,$96,$7C,$7F,$F2,$EB,$75,$FF,$2E,$CE,$F8,$FF,$CB,$73,$FE
	.dc.b	$EB,$38,$9F,$E5,$D2,$99,$77,$C9,$FE,$66,$EE,$FF,$20,$CF,$F2,$F3
	.dc.b	$FC,$EF,$E1,$7F,$A9,$1F,$F2,$BB,$5F,$F3,$23,$5F,$F3,$EF,$89,$67
	.dc.b	$FC,$B9,$1F,$E4,$76,$BD,$5A,$FF,$3D,$F0,$FF,$CA,$39,$7F,$AC,$9B
	.dc.b	$C3,$FE,$8C,$F5,$1E,$2F,$F9,$7A,$B5,$97,$F9,$5B,$0E,$4D,$7F,$A9
	.dc.b	$E1,$C9,$FF,$3B,$7A,$4F,$4B,$96,$5F,$E9,$88,$C3,$FE,$4A,$3F,$C9
	.dc.b	$1E,$3F,$F9,$CB,$17,$8F,$FE,$4C,$C3,$FC,$C3,$3F,$E5,$7F,$1F,$FC
	.dc.b	$9B,$7F,$EA,$2A,$D5,$7F,$96,$AF,$FC,$B1,$7F,$1E,$FF,$CB,$33,$F3
	.dc.b	$DA,$F8,$2F,$F7,$71,$D2,$3F,$D1,$DE,$93,$D2,$FF,$29,$5F,$96,$31
	.dc.b	$7F,$E5,$62,$57,$F9,$6F,$C9,$FF,$2A,$63,$FE,$57,$8F,$F2,$AC,$F6
	.dc.b	$F1,$7C,$84,$70,$7F,$A5,$3F,$0F,$7B,$FF,$E5,$8B,$FF,$76,$72,$27
	.dc.b	$35,$FE,$8F,$F2,$3F,$CB,$67,$F8,$57,$1F,$F0,$97,$91,$FE,$24,$8C
	.dc.b	$1F,$E4,$98,$FF,$0D,$F9,$3F,$E1,$1E,$9F,$F0,$BC,$7F,$8A,$5F,$F2
	.dc.b	$19,$2B,$FC,$33,$7F,$E1,$39,$FF,$12,$4F,$F8,$A9,$FF,$16,$CF,$F8
	.dc.b	$A1,$F0,$9F,$E3,$F7,$17,$E6,$BF,$04,$79,$5B,$3D,$23,$FC,$59,$E4
	.dc.b	$7F,$88,$EF,$CB,$E5,$FF,$8B,$B0,$F5,$47,$84,$F3,$5F,$F8,$BC,$87
	.dc.b	$AE,$FD,$2F,$F9,$0E,$36,$E4,$DD,$75,$C3,$FC,$B1,$3D,$6B,$8B,$FE
	.dc.b	$49,$7C,$B8,$FC,$3B,$F7,$9F,$E5,$33,$E0,$1F,$E4,$62,$9A,$C3,$E1
	.dc.b	$79,$7F,$E5,$18,$FF,$1F,$9E,$5A,$EE,$7C,$8C,$3F,$D2,$58,$78,$67
	.dc.b	$9D,$FC,$2A,$F1,$C7,$13,$D7,$97,$C9,$8F,$9D,$87,$7B,$F9,$73,$9A
	.dc.b	$FA,$07,$4C,$3E,$0D,$F4,$AF,$04,$FB,$49,$87,$E3,$2B,$71,$C7,$0F
	.dc.b	$F5,$13,$FE,$54,$3F,$CA,$C7,$F9,$64,$77,$7F,$95,$1F,$F2,$83,$5F
	.dc.b	$F2,$ED,$7F,$95,$9F,$F0,$DF,$97,$FE,$56,$72,$D7,$FC,$AC,$FF,$82
	.dc.b	$30,$FF,$23,$47,$F9,$00,$FF,$05,$1F,$E1,$67,$33,$FE,$13,$9F,$F0
	.dc.b	$2E,$3F,$E0,$7E,$FF,$F0,$2D,$FF,$8A,$0F,$F0,$84,$78,$A3,$FC,$E3
	.dc.b	$7F,$E3,$B9,$FF,$0E,$F1,$7F,$85,$B0,$FF,$0E,$C7,$7B,$FF,$2E,$1C
	.dc.b	$5F,$F2,$DB,$87,$D7,$7D,$23,$FC,$79,$7F,$E1,$E3,$FC,$31,$3E,$A3
	.dc.b	$FC,$3C,$57,$27,$B2,$FC,$58,$7B,$27,$FC,$9D,$87,$91,$E3,$5B,$7C
	.dc.b	$C7,$F9,$24,$F1,$9E,$63,$02,$3F,$C9,$87,$F9,$15,$E7,$7E,$B7,$55
	.dc.b	$FE,$6D,$3C,$4F,$CB,$9F,$15,$7A,$2B,$84,$74,$AF,$B5,$18,$79,$61
	.dc.b	$F5,$27,$C1,$5C,$6F,$CB,$1C,$67,$FC,$94,$48,$FA,$A7,$B8,$FC,$24
	.dc.b	$FC,$17,$D2,$72,$AE,$EF,$CD,$BF,$C2,$34,$F9,$71,$EC,$BF,$1C,$EA
	.dc.b	$BA,$1E,$5A,$EF,$1E,$BB,$FB,$D1,$FE,$60,$3C,$A7,$88,$F9,$D8,$7F
	.dc.b	$AC,$EB,$FC,$57,$E4,$7F,$92,$67,$FD,$83,$E5,$FF,$8E,$B7,$FF,$5C
	.dc.b	$36,$3F,$CA,$87,$F9,$5C,$FF,$63,$F1,$FF,$93,$BC,$AF,$F2,$E9,$FE
	.dc.b	$51,$AF,$F2,$9C,$E9,$E7,$1A,$8C,$1D,$27,$FC,$90,$F8,$BC,$BF,$F2
	.dc.b	$AB,$FE,$47,$8C,$AB,$4E,$A3,$FC,$97,$7F,$E5,$05,$5F,$E5,$F7,$B9
	.dc.b	$C6,$7F,$CC,$18,$7F,$96,$A3,$FC,$DB,$87,$F9,$74,$8F,$F9,$D3,$C9
	.dc.b	$78,$DF,$E7,$CF,$2B,$FC,$A7,$8B,$2F,$F6,$F4,$61,$FF,$2A,$DF,$82
	.dc.b	$F8,$BF,$E5,$BD,$F9,$1F,$E5,$58,$FF,$2C,$47,$7C,$3F,$CB,$2F,$F9
	.dc.b	$4F,$C9,$FF,$75,$63,$FD,$E3,$D1,$7C,$CE,$32,$FF,$7C,$F5,$A7,$38
	.dc.b	$7C,$0F,$2F,$FC,$B7,$87,$9E,$3F,$DF,$F1,$E4,$7E,$3D,$FD,$CF,$65
	.dc.b	$F0,$5D,$67,$E5,$C6,$1E,$9A,$7F,$A7,$BA,$7F,$CB,$5E,$77,$F3,$0F
	.dc.b	$3B,$FC,$BF,$18,$A1,$FE,$61,$3F,$CB,$7C,$F8,$E9,$FF,$2D,$C6,$08
	.dc.b	$AF,$FC,$B3,$9F,$E2,$35,$66,$E9,$F6,$47,$B7,$0F,$F3,$0D,$6E,$B4
	.dc.b	$FE,$0B,$83,$FC,$BB,$5F,$2F,$B7,$8C,$F9,$AF,$FC,$C6,$FE,$5A,$BE
	.dc.b	$64,$EA,$7F,$CA,$BE,$5F,$F9,$5D,$79,$7F,$E7,$27,$FC,$B7,$7F,$EB
	.dc.b	$C8,$FF,$2D,$60,$F2,$BF,$CA,$F5,$F4,$0E,$75,$ED,$8C,$CF,$F9,$9A
	.dc.b	$31,$E5,$FF,$98,$2B,$FC,$C6,$FF,$92,$A3,$FC,$9F,$E4,$7F,$95,$48
	.dc.b	$E0,$F9,$1E,$47,$F9,$3A,$BC,$07,$F9,$76,$3F,$77,$19,$BF,$F2,$FC
	.dc.b	$70,$F2,$7B,$57,$9E,$55,$BA,$FF,$30,$4E,$63,$0F,$F9,$BA,$39,$1F
	.dc.b	$E5,$C9,$FF,$28,$F8,$FF,$E5,$09,$FF,$2D,$9F,$0F,$6F,$F2,$A7,$91
	.dc.b	$FE,$57,$BF,$CE,$D6,$EF,$FC,$C4,$72,$9D,$5A,$F8,$DF,$73,$8C,$FD
	.dc.b	$99,$FF,$14,$6F,$FE,$46,$6F,$8D,$FE,$1B,$EF,$FF,$0E,$C7,$F9,$48
	.dc.b	$FF,$18,$1F,$E1,$69,$E1,$E2,$FF,$8C,$23,$47,$23,$FC,$C7,$5C,$DF
	.dc.b	$F1,$19,$C5,$C3,$BE,$7F,$F2,$4F,$91,$FE,$47,$9F,$F2,$15,$7F,$88
	.dc.b	$DB,$7E,$03,$FC,$73,$87,$F8,$EE,$3F,$C9,$15,$C7,$0F,$F2,$0E,$1F
	.dc.b	$E2,$EC,$39,$46,$2F,$83,$E6,$3D,$31,$FE,$41,$7F,$C6,$06,$CF,$82
	.dc.b	$F2,$9F,$F1,$FE,$5E,$4C,$3C,$3E,$47,$F9,$06,$FD,$98,$7F,$94,$1E
	.dc.b	$39,$7A,$6B,$77,$C6,$7C,$58,$7C,$2C,$39,$63,$F3,$63,$15,$B7,$FC
	.dc.b	$D5,$5F,$E5,$48,$C6,$3F,$E4,$D3,$DB,$3A,$9F,$F3,$A1,$32,$F8,$D1
	.dc.b	$EA,$8F,$F2,$F1,$9C,$3C,$39,$7C,$E8,$F4,$E5,$F6,$32,$F2,$63,$ED
	.dc.b	$9F,$79,$F2,$30,$E8,$7D,$1B,$F8,$11,$FE,$F5,$3F,$CB,$27,$F9,$51
	.dc.b	$5E,$63,$27,$F9,$73,$D7,$7A,$F9,$EC,$FF,$2E,$4E,$63,$86,$5F,$E5
	.dc.b	$D8,$FF,$17,$9F,$C6,$CF,$F2,$DF,$66,$BC,$AF,$F1,$1C,$7F,$85,$63
	.dc.b	$FC,$12,$FF,$95,$0F,$F1,$BE,$7F,$E5,$33,$FC,$91,$A7,$F8,$4C,$FF
	.dc.b	$08,$47,$F8,$EC,$FF,$09,$9C,$CC,$5F,$F9,$8E,$BF,$C5,$D7,$E0,$9D
	.dc.b	$92,$74,$FF,$88,$CF,$F1,$E4,$7F,$88,$CE,$11,$AB,$FF,$11,$DF,$8E
	.dc.b	$BC,$11,$E7,$8F,$F3,$23,$63,$D7,$1F,$E6,$23,$25,$72,$7F,$8A,$63
	.dc.b	$DE,$FF,$8D,$2F,$D2,$7F,$90,$63,$31,$E6,$8C,$BF,$E3,$08,$FF,$1B
	.dc.b	$57,$9E,$BE,$3C,$78,$E7,$FC,$B7,$5F,$42,$BF,$CB,$35,$EB,$9F,$89
	.dc.b	$D9,$FE,$70,$AE,$B7,$E0,$BF,$F2,$2D,$79,$4F,$90,$E1,$F2,$18,$C3
	.dc.b	$E4,$DF,$38,$C5,$FA,$F0,$F8,$71,$F0,$5A,$FE,$0C,$F3,$9F,$72,$79
	.dc.b	$61,$F4,$A3,$C4,$D7,$C2,$A7,$FB,$66,$1F,$DD,$E7,$F1,$FF,$23,$FE
	.dc.b	$3B,$7F,$C8,$79,$3C,$77,$C6,$7F,$84,$63,$0F,$F8,$4E,$55,$C5,$FE
	.dc.b	$B6,$D7,$F4,$0D,$7F,$BE,$35,$FF,$09,$B5,$E7,$E6,$75,$8E,$9C,$FF
	.dc.b	$E0,$B6,$F2,$FF,$86,$B0,$68,$EC,$E1,$F0,$1E,$7D,$0D,$33,$9F,$F9
	.dc.b	$4B,$1C,$BF,$BF,$6B,$FC,$15,$AF,$AA,$3F,$82,$3F,$01,$E6,$FC,$07
	.dc.b	$E3,$5B,$E0,$B3,$9F,$A6,$6E,$7E,$9A,$D5,$F8,$F9,$FD,$1F,$09,$23
	.dc.b	$04,$3C,$91,$E2,$C8,$D2,$46,$F1,$F9,$11,$E4,$F6,$7E,$C7,$9B,$FE
	.dc.b	$5E,$AF,$E4,$4D,$7F,$8B,$75,$79,$1F,$E4,$CD,$73,$AF,$C2,$8C,$47
	.dc.b	$F9,$3F,$0F,$F2,$57,$8D,$61,$CF,$FE,$49,$7F,$CA,$3D,$3F,$E5,$37
	.dc.b	$FC,$A5,$1F,$E5,$02,$4F,$F3,$5D,$B9,$CF,$F2,$96,$AF,$79,$EF,$FF
	.dc.b	$93,$E3,$FC,$91,$18,$E1,$FE,$74,$D8,$FE,$80,$D7,$F9,$A7,$8D,$FE
	.dc.b	$47,$7F,$C8,$27,$F4,$57,$0F,$F9,$33,$C7,$EC,$FF,$9A,$BE,$37,$FC
	.dc.b	$B6,$FF,$8E,$9B,$AF,$F9,$F7,$A3,$FD,$43,$CD,$FE,$1F,$9C,$E5,$FE
	.dc.b	$5D,$78,$DF,$F4,$56,$B9,$ED,$EE,$A3,$2F,$86,$3F,$CC,$56,$FC,$25
	.dc.b	$F0,$FD,$E7,$31,$8C,$3F,$D6,$38,$3E,$2B,$C8,$72,$CD,$D5,$8A,$DC
	.dc.b	$DC,$FF,$CC,$D9,$FF,$AF,$2B,$17,$FE,$7F,$C3,$A6,$1F,$DA,$5B,$7F
	.dc.b	$DB,$DA,$FB,$F0,$F6,$C6,$AB,$E4,$63,$F9,$CB,$F0,$9F,$C5,$1C,$5F
	.dc.b	$CF,$FA,$9F,$98,$7A,$7E,$0E,$D7,$BA,$3A,$91,$F2,$36,$7D,$74,$F8
	.dc.b	$1E,$31,$87,$FC,$AD,$E1,$7F,$83,$A3,$FC,$3C,$7F,$85,$BC,$7F,$F0
	.dc.b	$CF,$89,$FE,$56,$E7,$FF,$12,$CE,$7B,$FF,$C3,$3E,$3F,$F8,$BA,$31
	.dc.b	$A7,$F8,$99,$FF,$0B,$5F,$13,$FC,$35,$8F,$F9,$06,$3F,$CA,$C7,$F9
	.dc.b	$13,$B7,$FC,$2A,$FF,$90,$CF,$F0,$F9,$AC,$7C,$27,$F8,$7F,$0F,$F1
	.dc.b	$34,$FF,$8A,$67,$75,$FE,$49,$BF,$04,$7B,$2F,$83,$DE,$FC,$CF,$F8
	.dc.b	$CF,$0B,$7C,$63,$FC,$D6,$D7,$C7,$97,$C4,$73,$8F,$F8,$C3,$0F,$F2
	.dc.b	$14,$E6,$77,$97,$C6,$7F,$CA,$39,$7C,$2B,$F7,$DF,$F8,$EA,$3F,$C9
	.dc.b	$67,$F9,$36,$7F,$C8,$67,$9C,$F8,$98,$7C,$93,$C8,$7D,$97,$D4,$7D
	.dc.b	$22,$4F,$C0,$3E,$4C,$F6,$8F,$AA,$7A,$B2,$F9,$67,$DB,$9E,$13,$E7
	.dc.b	$8F,$8F,$8F,$B6,$FC,$0D,$C3,$EC,$E1,$FE,$43,$94,$62,$3F,$BA,$B5
	.dc.b	$FF,$2A,$57,$F9,$3F,$C7,$FF,$28,$C7,$F9,$67,$BB,$FC,$A2,$23,$31
	.dc.b	$D7,$C7,$FE,$62,$FF,$97,$F0,$FF,$2A,$3F,$E5,$23,$11,$FE,$14,$7F
	.dc.b	$C4,$4D,$7F,$B6,$E5,$FE,$23,$9F,$F1,$54,$7F,$8F,$A3,$FC,$70,$FF
	.dc.b	$85,$33,$FF,$16,$4A,$BC,$31,$86,$D7,$09,$D7,$7F,$80,$FF,$16,$C7
	.dc.b	$F8,$82,$3F,$C6,$77,$FE,$1C,$8F,$F1,$C9,$C6,$B9,$C7,$F8,$9C,$FF
	.dc.b	$24,$9F,$E2,$79,$EA,$FF,$89,$CE,$C7,$F9,$16,$7D,$9D,$1F,$E2,$A7
	.dc.b	$AC,$73,$3D,$8F,$7B,$E3,$87,$F8,$B6,$FF,$C7,$C4,$C3,$E4,$47,$F9
	.dc.b	$43,$0E,$98,$7C,$73,$A5,$79,$27,$E2,$5F,$4C,$3C,$35,$E9,$C3,$DA
	.dc.b	$D9,$D3,$CC,$C1,$ED,$3E,$B9,$F1,$CD,$5F,$CA,$7F,$C8,$75,$DA,$30
	.dc.b	$7A,$67,$FC,$A5,$3F,$2A,$33,$5E,$CA,$F8,$F3,$E2,$3E,$29,$EC,$8F
	.dc.b	$3C,$F0,$AE,$73,$A3,$CF,$7F,$01,$F4,$3F,$1D,$E9,$3F,$16,$FE,$84
	.dc.b	$69,$F0,$BE,$E3,$A4,$6D,$F4,$60,$7F,$CB,$BD,$3F,$E4,$3F,$80,$91
	.dc.b	$FE,$41,$8F,$F2,$71,$FE,$2B,$23,$FE,$43,$6B,$FE,$5C,$76,$FF,$99
	.dc.b	$B0,$7D,$1F,$F9,$5F,$C8,$FF,$28,$79,$32,$BF,$C9,$D1,$97,$FC,$9C
	.dc.b	$FF,$8E,$48,$E0,$D9,$0F,$ED,$F5,$FE,$5D,$71,$D3,$FE,$63,$9E,$91
	.dc.b	$AF,$27,$FC,$9A,$FF,$B5,$BC,$9F,$F2,$DD,$7F,$84,$DB,$CF,$FE,$72
	.dc.b	$F0,$3F,$CB,$35,$FE,$5D,$3F,$CC,$4F,$F9,$34,$8F,$F9,$9D,$EB,$E0
	.dc.b	$7F,$9C,$3C,$6F,$F4,$17,$B4,$F6,$BF,$CC,$AD,$BF,$F7,$53,$5F,$F7
	.dc.b	$83,$5F,$F7,$27,$93,$FE,$51,$7E,$C0,$70,$FF,$92,$3D,$46,$9F,$E4
	.dc.b	$2E,$AF,$F2,$51,$23,$33,$FE,$50,$F2,$BF,$C9,$4D,$3F,$C9,$45,$3B
	.dc.b	$F0,$FF,$93,$1F,$C4,$3F,$E4,$EF,$23,$FC,$82,$4F,$23,$FC,$66,$FF
	.dc.b	$91,$3D,$D7,$A7,$FE,$36,$6F,$17,$FB,$FF,$0F,$C9,$46,$1C,$35,$FF
	.dc.b	$11,$47,$F8,$79,$FF,$16,$E9,$FE,$32,$8C,$3F,$E2,$DF,$27,$FC,$4E
	.dc.b	$48,$FF,$17,$36,$7F,$C3,$7E,$27,$F8,$D2,$7F,$D4,$DE,$37,$F8,$A2
	.dc.b	$FF,$C6,$DE,$3F,$F8,$79,$FF,$2B,$0F,$23,$15,$FE,$55,$9B,$83,$9B
	.dc.b	$E8,$D6,$23,$FC,$A9,$3F,$E2,$B7,$FC,$B4,$D7,$95,$7F,$8E,$7C,$8E
	.dc.b	$77,$FE,$57,$51,$FB,$7C,$2D,$7E,$E7,$EC,$3F,$D1,$33,$93,$F0,$AE
	.dc.b	$23,$E3,$F0,$70,$AE,$30,$9D,$46,$1E,$11,$F8,$E3,$5E,$27,$F9,$BA
	.dc.b	$33,$1D,$A3,$07,$F9,$AB,$E6,$BB,$61,$93,$FC,$A8,$FF,$91,$BC,$9F
	.dc.b	$F1,$77,$8B,$FE,$11,$7F,$C4,$1E,$37,$F8,$5F,$0F,$F1,$8C,$61,$FF
	.dc.b	$06,$9F,$E2,$2B,$FF,$16,$C6,$63,$FC,$79,$3F,$E0,$D7,$FC,$3B,$1F
	.dc.b	$E2,$48,$FF,$1C,$C7,$F8,$73,$9B,$FC,$26,$FF,$8E,$E3,$FC,$34,$F2
	.dc.b	$30,$7F,$8A,$23,$FC,$3E,$D8,$ED,$1F,$E2,$F8,$C1,$C0,$F4,$F9,$3F
	.dc.b	$E2,$27,$FC,$6E,$67,$0F,$81,$19,$9F,$F1,$FF,$1F,$F8,$F5,$B8,$FB
	.dc.b	$A7,$58,$7C,$5B,$E3,$87,$F8,$EE,$FC,$4F,$80,$C4,$FC,$0F,$2B,$FC
	.dc.b	$59,$7F,$E3,$38,$F7,$DF,$F9,$03,$2F,$A6,$7C,$7B,$FF,$1D,$1C,$8B
	.dc.b	$87,$C2,$7B,$36,$7C,$6D,$9E,$53,$F5,$CF,$F2,$4D,$78,$23,$EC,$57
	.dc.b	$86,$FC,$95,$F2,$8F,$AD,$5B,$8F,$E1,$EF,$F2,$0C,$3D,$73,$FE,$37
	.dc.b	$BF,$F2,$12,$BF,$F2,$3D,$FF,$0D,$8C,$4E,$BA,$FF,$C8,$7E,$57,$F9
	.dc.b	$0A,$7F,$95,$C7,$F9,$17,$8F,$86,$1F,$E4,$22,$79,$5F,$E4,$6A,$FF
	.dc.b	$24,$70,$6A,$31,$1F,$CD,$27,$73,$92,$47,$F9,$29,$AF,$23,$04,$30
	.dc.b	$43,$A4,$F6,$7F,$BC,$4F,$67,$81,$A9,$EA,$D3,$27,$09,$E4,$FF,$97
	.dc.b	$4E,$C6,$63,$8B,$4E,$26,$CE,$87,$03,$06,$CC,$4F,$95,$B3,$E0,$8C
	.dc.b	$47,$22,$1C,$4A,$74,$23,$C1,$D1,$0C,$C7,$52,$4F,$AA,$32,$D9,$E5
	.dc.b	$18,$3D,$AE,$09,$3C,$A3,$2D,$7D,$73,$CE,$31,$1E,$89,$E3,$39,$24
	.dc.b	$72,$6B,$C8,$C1,$0C,$10,$E9,$3D,$9F,$81,$3D,$9E,$06,$A7,$AB,$4C
	.dc.b	$9C,$27,$93,$F0,$4E,$C6,$63,$8B,$4E,$26,$CE,$87,$03,$06,$CC,$4F
	.dc.b	$95,$B3,$E0,$8C,$47,$22,$1C,$4A,$74,$23,$C1,$D1,$0C,$C7,$52,$4F
	.dc.b	$AA,$32,$D9,$E5,$18,$3D,$AE,$09,$3C,$A3,$2D,$7D,$73,$CE,$31,$1E
	.dc.b	$89,$E3,$39,$24,$72,$6B,$C8,$C1,$0C,$10,$E9,$3D,$9F,$81,$3D,$9E
	.dc.b	$06,$A7,$AB,$4C,$9C,$27,$93,$F0,$4E,$C6,$63,$8B,$4E,$26,$CE,$87
	.dc.b	$03,$06,$CC,$4F,$EB,$AF,$F9,$64,$3D,$C7,$B3,$FC,$C3,$2F,$F7,$BC
	.dc.b	$62,$7F,$D2,$95,$C2,$31,$08,$C5,$71,$CB,$FD,$35,$5A,$9C,$D7,$2B
	.dc.b	$EB,$7B,$8F,$13,$FC,$DF,$3F,$E9,$38,$FF,$48,$9F,$E9,$9A,$7F,$42
	.dc.b	$9C,$91,$C1,$23,$64,$3F,$C5,$13,$CD,$B3,$C5,$D1,$E1,$9E,$4E,$9A
	.dc.b	$E4,$D9,$FD,$02,$39,$3B,$32,$7F,$5E,$8E,$2E,$DA,$E4,$DC,$F9,$5E
	.dc.b	$C7,$32,$18,$21,$CC,$E3,$3E,$27,$27,$53,$89,$B2,$99,$37,$1E,$17
	.dc.b	$46,$23,$A3,$82,$1C,$4B,$1C,$5C,$11,$C1,$23,$9C,$65,$B1,$99,$F8
	.dc.b	$AD,$9F,$6C,$62,$76,$47,$04,$8D,$90,$F7,$CF,$36,$CF,$17,$47,$BE
	.dc.b	$79,$3A,$6B,$93,$67,$C0,$8E,$4E,$CC,$9E,$D8,$E2,$ED,$AE,$4D,$CF
	.dc.b	$95,$EC,$73,$21,$82,$1C,$CE,$33,$E2,$72,$75,$38,$9B,$29,$93,$71
	.dc.b	$E1,$74,$62,$3A,$38,$21,$C4,$B1,$C5,$C1,$1C,$12,$39,$C6,$5B,$19
	.dc.b	$9F,$8A,$D9,$F6,$C6,$27,$64,$70,$48,$D9,$0F,$7C,$F3,$6C,$F1,$74
	.dc.b	$7B,$E7,$93,$A6,$B9,$36,$7C,$08,$E4,$EC,$C9,$ED,$8E,$2E,$DA,$E4
	.dc.b	$DC,$F9,$5E,$C7,$32,$18,$21,$CC,$E3,$3E,$27,$27,$53,$89,$B2,$99
	.dc.b	$37,$1F,$B9,$C3,$FC,$E7,$97,$F9,$DD,$FE,$8F,$3C,$27,$CC,$F3,$38
	.dc.b	$D7,$85,$C1,$0E,$C5,$8E,$2E,$08,$E0,$91,$CE,$32,$D8,$CD,$FC,$A6
	.dc.b	$CF,$C2,$8C,$4E,$C8,$E0,$91,$B2,$1E,$D8,$DB,$83,$DB,$1C,$1D,$35
	.dc.b	$C9,$B3,$DD,$1C,$9D,$99,$3D,$71,$C5,$DB,$5C,$9B,$9F,$6B,$C8,$87
	.dc.b	$80,$A7,$23,$84,$F8,$5D,$10,$D1,$C0,$EE,$58,$EE,$EC,$C9,$5C,$BC
	.dc.b	$8C,$47,$77,$04,$38,$96,$38,$B8,$23,$82,$47,$38,$CB,$63,$33,$F1
	.dc.b	$DB,$3F,$02,$31,$3F,$89,$AF,$F0,$8C,$FF,$85,$F1,$FF,$08,$BF,$E1
	.dc.b	$3C,$3F,$C2,$F8,$FF,$84,$F2,$FF,$0C,$56,$AB,$85,$69,$E0,$7F,$86
	.dc.b	$0B,$3F,$E2,$4C,$BF,$C8,$51,$FE,$30,$D3,$FC,$7B,$18,$8F,$F1,$B9
	.dc.b	$0F,$F1,$B1,$4F,$F1,$C9,$1E,$0E,$88,$66,$3A,$93,$4F,$F2,$3C,$65
	.dc.b	$B3,$CA,$30,$7F,$92,$9C,$12,$79,$46,$5A,$FA,$E7,$9C,$62,$3D,$13
	.dc.b	$C6,$72,$48,$E4,$D7,$91,$82,$18,$21,$D2,$7B,$3F,$02,$7B,$3C,$0D
	.dc.b	$4F,$56,$99,$38,$4F,$27,$E0,$9D,$8C,$C7,$16,$9C,$4D,$9D,$0E,$06
	.dc.b	$0D,$98,$9F,$2B,$67,$C1,$18,$8E,$44,$38,$94,$E8,$47,$83,$A2,$19
	.dc.b	$8E,$A4,$9F,$54,$65,$B3,$CA,$3A,$E5,$FC,$F9,$10,$FC,$E9,$FC,$0F
	.dc.b	$36,$9A,$DD,$BF,$F9,$8D,$B5,$CB,$77,$0B,$87,$85,$EC,$95,$FE,$59
	.dc.b	$BE,$2A,$F8,$D7,$08,$D5,$70,$7F,$CD,$15,$C2,$B7,$08,$E7,$9F,$FA
	.dc.b	$25,$B9,$7A,$A7,$57,$CE,$F8,$D6,$B2,$F0,$57,$6A,$D6,$5E,$B8,$E2
	.dc.b	$79,$08,$6C,$F8,$4E,$EB,$A4,$70,$34,$4B,$F0,$D7,$B7,$1F,$43,$E0
	.dc.b	$CB,$C7,$18,$AE,$ED,$7E,$46,$1D,$EB,$96,$5E,$FC,$7C,$27,$C7,$9D
	.dc.b	$57,$29,$D5,$77,$8F,$35,$F5,$24,$EA,$3D,$B8,$79,$23,$11,$EC,$9F
	.dc.b	$26,$3E,$C8,$ED,$96,$CF,$9B,$19,$8E,$15,$CA,$B8,$E3,$E6,$9C,$DF
	.dc.b	$17,$E5,$5F,$1A,$F2,$CE,$72,$F7,$63,$72,$F8,$73,$E0,$C7,$A5,$F0
	.dc.b	$C7,$DF,$97,$AB,$2E,$8F,$8C,$97,$C7,$1E,$95,$EC,$3C,$78,$74,$AD
	.dc.b	$C7,$4A,$E0,$D2,$65,$F5,$32,$FD,$86,$3E,$0A,$EB,$5F,$85,$C3,$A4
	.dc.b	$63,$2F,$8B,$95,$CB,$BE,$5F,$03,$0E,$B9,$7C,$07,$CB,$5B,$C3,$AE
	.dc.b	$5D,$E7,$67,$E6,$9C,$5F,$96,$31,$97,$89,$D9,$E5,$BE,$8E,$2B,$8E
	.dc.b	$1E,$1B,$F2,$1C,$E7,$ED,$9E,$43,$9E,$5F,$4E,$37,$3C,$4F,$CA,$3C
	.dc.b	$63,$51,$F2,$30,$F2,$4E,$72,$FB,$51,$89,$E5,$7F,$98,$3B,$CF,$C8
	.dc.b	$8E,$58,$FA,$AF,$CD,$8F,$AA,$FE,$F4,$79,$1B,$5E,$43,$AB,$F0,$CE
	.dc.b	$B9,$7C,$EC,$7C,$51,$C0,$F1,$57,$9E,$75,$87,$86,$FA,$56,$CE,$95
	.dc.b	$B7,$E7,$5F,$47,$E5,$C7,$95,$B3,$C9,$A7,$DA,$AE,$CD,$3C,$E7,$47
	.dc.b	$2F,$82,$3D,$B9,$7E,$42,$37,$5F,$1A,$FC,$53,$F5,$27,$DB,$5C,$32
	.dc.b	$FA,$78,$74,$AE,$58,$7A,$CE,$51,$99,$DD,$F6,$BF,$55,$F5,$8F,$21
	.dc.b	$F1,$5E,$71,$DF,$1F,$D7,$E1,$D6,$38,$DF,$DF,$BF,$81,$39,$6C,$F1
	.dc.b	$7F,$29,$19,$7B,$9F,$CE,$38,$DA,$FF,$43,$E4,$FE,$89,$C2,$E9,$72
	.dc.b	$E7,$6F,$ED,$98,$70,$7F,$B7,$DF,$0A,$DC,$60,$DE,$1F,$D3,$F1,$FE
	.dc.b	$E8,$2F,$93,$83,$11,$B7,$8C,$E5,$CB,$D5,$FE,$FF,$3C,$AF,$D2,$39
	.dc.b	$BD,$13,$AB,$AB,$CD,$74,$EB,$4F,$81,$D3,$A9,$CE,$7D,$E7,$2F,$86
	.dc.b	$F9,$5F,$9D,$FF,$08,$D7,$07,$D8,$D7,$E0,$B5,$FF,$0B,$3B,$CB,$E3
	.dc.b	$36,$BF,$C3,$76,$97,$57,$97,$09,$5E,$3B,$E1,$83,$0E,$17,$A8,$E5
	.dc.b	$5E,$0E,$BF,$81,$B7,$F8,$43,$2F,$0C,$63,$97,$E1,$3E,$DB,$F9,$0D
	.dc.b	$C8,$D2,$61,$E3,$7E,$65,$BC,$EE,$73,$F2,$7B,$5D,$DD,$5B,$1F,$7D
	.dc.b	$B8,$FE,$15,$AF,$E2,$BB,$FE,$F4,$FE,$15,$B1,$F8,$B6,$BB,$E2,$F5
	.dc.b	$3F,$E4,$13,$F4,$3E,$07,$E4,$79,$BF,$C9,$3C,$88,$C4,$E8,$FC,$CC
	.dc.b	$EA,$3F,$C9,$DD,$8F,$07,$EB,$56,$C4,$67,$6E,$11,$FE,$56,$8C,$3B
	.dc.b	$6E,$1E,$18,$FF,$2D,$F8,$4F,$0B,$F4,$50,$8C,$3F,$A5,$6B,$FA,$B6
	.dc.b	$BF,$B0,$F0,$DE,$27,$26,$C7,$EB,$5A,$FE,$CD,$AF,$EE,$3D,$27,$8F
	.dc.b	$FE,$6D,$CF,$F7,$91,$89,$FD,$23,$5F,$C7,$9F,$93,$D9,$E3,$F1,$E2
	.dc.b	$FF,$3B,$56,$EB,$B6,$1E,$E8,$C4,$6A,$30,$71,$7F,$3D,$E4,$77,$C7
	.dc.b	$F8,$CB,$6F,$84,$F9,$3C,$AE,$0B,$04,$61,$B1,$FB,$FC,$9E,$56,$6A
	.dc.b	$4F,$33,$51,$87,$F8,$D5,$72,$6C,$FA,$C8,$FF,$0B,$C1,$E5,$FA,$FB
	.dc.b	$D1,$89,$EC,$7C,$A3,$D1,$E6,$F9,$C8,$F9,$FC,$DA,$F8,$A3,$E5,$11
	.dc.b	$FE,$55,$8A,$3C,$14,$F3,$A9,$CB,$6F,$E7,$1E,$53,$CC,$FE,$8F,$8F
	.dc.b	$F9,$B1,$BB,$FF,$A5,$9B,$E4,$FF,$AC,$69,$E7,$FF,$AC,$EB,$5D,$3F
	.dc.b	$E8,$2B,$DE,$1F,$EB,$8B,$ED,$3F,$E8,$6B,$E1,$87,$3B,$DD,$AF,$6F
	.dc.b	$8E,$F8,$3D,$58,$E4,$FF,$41,$DF,$47,$C1,$5D,$2B,$9E,$1E,$69,$D9
	.dc.b	$C9,$C1,$30,$F7,$5F,$39,$F2,$5F,$32,$61,$E4,$8F,$15,$F1,$7C,$31
	.dc.b	$98,$C4,$72,$8F,$05,$F6,$8A,$70,$39,$B9,$69,$99,$E7,$87,$D2,$8E
	.dc.b	$27,$53,$67,$CB,$C3,$D5,$1F,$0A,$BE,$85,$F0,$8E,$55,$C1,$D3,$DE
	.dc.b	$7C,$57,$E0,$C3,$E0,$CF,$CA,$38,$38,$8F,$F6,$EE,$1E,$08,$C5,$78
	.dc.b	$2B,$85,$73,$C3,$C5,$7B,$7C,$97,$F6,$63,$15,$DA,$7E,$65,$F3,$BE
	.dc.b	$53,$D2,$30,$47,$D4,$7E,$4E,$B9,$CF,$03,$CF,$1E,$0A,$F1,$B4,$F8
	.dc.b	$93,$C2,$BA,$47,$CF,$AE,$2F,$94,$D9,$C5,$FA,$F3,$C6,$3C,$67,$19
	.dc.b	$F9,$A7,$4C,$3E,$8C,$64,$F4,$1C,$E3,$53,$CE,$34,$49,$FB,$75,$ED
	.dc.b	$AD,$CF,$13,$C3,$3A,$8F,$06,$1F,$5A,$7B,$9E,$BA,$F4,$BC,$4E,$D7
	.dc.b	$E1,$BF,$39,$1C,$57,$39,$FC,$91,$C6,$74,$D8,$FB,$CD,$C3,$EE,$BE
	.dc.b	$68,$E5,$3E,$78,$C4,$FD,$67,$CD,$87,$B9,$A7,$BA,$FC,$73,$BC,$3E
	.dc.b	$24,$E4,$F0,$1F,$89,$3A,$9F,$68,$E6,$7C,$03,$6F,$E0,$0E,$6E,$4D
	.dc.b	$DF,$B9,$B8,$7B,$9B,$5E,$E7,$EC,$47,$BA,$FC,$F1,$E1,$BE,$B7,$CD
	.dc.b	$B5,$CD,$B7,$CE,$7E,$9C,$F3,$AF,$3E,$1F,$8E,$7D,$B7,$F7,$1F,$19
	.dc.b	$F0,$4F,$1D,$FC,$79,$FC,$2B,$63,$8C,$7C,$F9,$F0,$C7,$1B,$F0,$D6
	.dc.b	$E3,$99,$EB,$AE,$6D,$3F,$90,$1F,$E8,$A9,$FE,$82,$DB,$FE,$AA,$DA
	.dc.b	$FF,$44,$75,$3C,$FF,$F3,$24,$FF,$A3,$E3,$1C,$3A,$9C,$CF,$FA,$5A
	.dc.b	$72,$FF,$A5,$70,$79,$8F,$37,$FD,$29,$18,$6C,$66,$78,$35,$F4,$3F
	.dc.b	$EB,$95,$E7,$7F,$BD,$5F,$F6,$E9,$8F,$23,$FD,$A7,$3B,$9D,$5F,$FB
	.dc.b	$25,$B1,$99,$DD,$FF,$BC,$AF,$FD,$A0,$7F,$B1,$0F,$F6,$74,$FF,$B5
	.dc.b	$36,$FF,$62,$61,$FE,$D0,$9C,$D7,$FB,$43,$6F,$F6,$55,$7F,$BC,$EB
	.dc.b	$FD,$B0,$6E,$BC,$ED,$33,$7D,$6B,$88,$9F,$F7,$59,$D8,$FF,$6A,$BA
	.dc.b	$72,$74,$8E,$4E,$EB,$C5,$1F,$26,$BC,$0D,$37,$3C,$6F,$D4,$F1,$BF
	.dc.b	$13,$F3,$7A,$7F,$DE,$D3,$A8,$E0,$79,$0E,$07,$92,$FD,$B1,$B6,$C6
	.dc.b	$CE,$38,$79,$A7,$E0,$3F,$EF,$1C,$BE,$BC,$7D,$58,$FA,$38,$FA,$2F
	.dc.b	$9C,$71,$BF,$B2,$79,$CB,$1F,$02,$BA,$E3,$FE,$2F,$BF,$C1,$63,$EC
	.dc.b	$BF,$25,$F4,$AF,$83,$8F,$8A,$58,$A7,$96,$3E,$DC,$7E,$63,$4E,$F5
	.dc.b	$FE,$37,$8E,$EF,$91,$F5,$1C,$1F,$0B,$6B,$D8,$7C,$3C,$BF,$2F,$3E
	.dc.b	$2B,$F0,$BC,$2F,$BB,$FA,$4C,$3D,$17,$E7,$8F,$CE,$1F,$42,$BD,$B7
	.dc.b	$D2,$75,$3F,$E4,$7C,$3C,$B8,$74,$C7,$F5,$57,$BC,$3C,$38,$79,$9A
	.dc.b	$4C,$3F,$05,$87,$AE,$3B,$E5,$F3,$EB,$85,$7C,$5C,$7E,$F5,$7E,$76
	.dc.b	$3E,$D4,$FE,$02,$7E,$64,$7D,$EC,$BD,$B5,$DE,$36,$F1,$8C,$C7,$C0
	.dc.b	$CB,$EE,$3E,$4A,$F6,$63,$F5,$30,$ED,$3F,$2C,$ED,$97,$BD,$E5,$3E
	.dc.b	$B7,$91,$AC,$BD,$99,$7B,$A3,$EF,$61,$E8,$C3,$ED,$57,$82,$3E,$BF
	.dc.b	$90,$F2,$D3,$FC,$07,$8E,$E0,$E0,$79,$4F,$33,$35,$AA,$63,$CF,$0E
	.dc.b	$47,$F0,$BC,$38,$DF,$08,$FD,$EC,$EE,$BF,$7E,$7F,$11,$70,$E1,$AF
	.dc.b	$0C,$7C,$0D,$9E,$EF,$18,$EC,$FA,$70,$79,$98,$8F,$3C,$DB,$79,$9C
	.dc.b	$32,$F7,$F0,$66,$31,$A2,$31,$9F,$A3,$1F,$45,$79,$1B,$19,$7E,$2C
	.dc.b	$65,$B1,$E8,$69,$FE,$F2,$B6,$6F,$33,$BC,$FF,$B3,$72,$F3,$DF,$16
	.dc.b	$C7,$1A,$E1,$39,$C3,$DD,$D1,$EA,$EB,$B3,$F2,$32,$FB,$B1,$8C,$3D
	.dc.b	$98,$FF,$B1,$2B,$D4,$D8,$CD,$FA,$E3,$18,$FE,$0D,$B5,$F8,$18,$78
	.dc.b	$3D,$5F,$C4,$79,$4F,$2B,$E2,$36,$BE,$43,$71,$F9,$CD,$7F,$0F,$E8
	.dc.b	$3D,$1F,$98,$DA,$FA,$0D,$9F,$AC,$D7,$EA,$73,$7F,$5F,$ED,$79,$9F
	.dc.b	$6E,$30,$D9,$4E,$9F,$CB,$CE,$9F,$ED,$1D,$AF,$33,$EC,$C7,$0B,$FE
	.dc.b	$B7,$39,$D7,$84,$A3,$0E,$CE,$75,$CB,$0F,$04,$7F,$87,$3C,$8F,$C6
	.dc.b	$34,$FC,$77,$6A,$32,$D8,$FC,$97,$96,$F3,$38,$B6,$33,$18,$3A,$F8
	.dc.b	$DF,$9F,$8F,$CB,$35,$FC,$AE,$FF,$DD,$A3,$1A,$7F,$08,$E7,$FE,$AF
	.dc.b	$E9,$7F,$7A,$8F,$E0,$D3,$C6,$FF,$7E,$D7,$97,$5F,$80,$8F,$E5,$7C
	.dc.b	$7F,$D1,$37,$B3,$A6,$5C,$23,$CB,$D1,$87,$FC,$35,$87,$E4,$1C,$9F
	.dc.b	$9F,$E4,$F8,$8F,$A8,$DC,$F1,$53,$C5,$D1,$88,$ED,$A7,$F7,$4E,$0E
	.dc.b	$15,$BA,$F6,$D7,$CA,$23,$FC,$0E,$3A,$69,$F1,$F3,$B9,$7F,$27,$6D
	.dc.b	$79,$0F,$A3,$E5,$76,$23,$EB,$F2,$EB,$FC,$4F,$04,$FD,$A2,$47,$85
	.dc.b	$F4,$D3,$CB,$A7,$2C,$7F,$9E,$78,$DF,$E8,$0B,$FF,$3E,$1F,$E8,$0D
	.dc.b	$BF,$D1,$C7,$FA,$18,$FF,$54,$78,$DF,$E8,$27,$2D,$3F,$D1,$85,$8F
	.dc.b	$F5,$B1,$FE,$96,$8E,$2F,$FA,$48,$CB,$D6,$3F,$D3,$D1,$E1,$79,$BB
	.dc.b	$3F,$D6,$D1,$B7,$FD,$BE,$D7,$FD,$AE,$77,$8E,$73,$ED,$9F,$3C,$F3
	.dc.b	$8F,$F5,$A4,$F9,$E3,$AB,$63,$31,$BF,$27,$FD,$71,$39,$6C,$67,$9F
	.dc.b	$FD,$75,$1F,$EE,$03,$31,$EB,$3D,$F3,$C6,$76,$7F,$BB,$7C,$0F,$F6
	.dc.b	$4E,$1E,$6A,$E0,$79,$0E,$C7,$FA,$DE,$FB,$57,$28,$FF,$6C,$D7,$2B
	.dc.b	$F9,$31,$CB,$0F,$5B,$6F,$CB,$1F,$16,$76,$D9,$F3,$9E,$0C,$3E,$CB
	.dc.b	$C7,$0F,$04,$7F,$B7,$B0,$F0,$4F,$03,$47,$9E,$7D,$CF,$FB,$70,$FB
	.dc.b	$37,$F0,$DD,$1E,$A3,$47,$DC,$AF,$BD,$6B,$EF,$3F,$14,$EF,$3E,$69
	.dc.b	$F5,$46,$5A,$7C,$22,$E1,$F0,$8F,$97,$1E,$29,$F6,$3E,$29,$F0,$36
	.dc.b	$BC,$31,$CE,$B8,$D7,$85,$B7,$F9,$26,$DF,$C0,$3E,$64,$67,$0F,$8B
	.dc.b	$1F,$99,$34,$7B,$23,$4F,$C5,$34,$7C,$57,$E5,$57,$86,$79,$BF,$23
	.dc.b	$0F,$3D,$FB,$E7,$35,$C0,$EB,$5F,$02,$FF,$D0,$47,$FB,$F8,$FF,$5B
	.dc.b	$E1,$FE,$FE,$6D,$FF,$BF,$9B,$7F,$EF,$9E,$29,$1F,$EF,$DF,$27,$FD
	.dc.b	$F6,$FF,$66,$6B,$FE,$75,$3F,$DE,$BE,$47,$FB,$E0,$8E,$0F,$19,$FE
	.dc.b	$89,$7F,$CC,$19,$7F,$9D,$71,$FF,$59,$4F,$F9,$9C,$FF,$32,$9F,$E6
	.dc.b	$FC,$BF,$CC,$AE,$1F,$F4,$2B,$FE,$9B,$E6,$FF,$39,$3D,$4E,$15,$D7
	.dc.b	$AB,$FD,$2F,$1D,$6B,$FC,$F6,$72,$BF,$F4,$CF,$27,$F9,$DA,$7F,$CF
	.dc.b	$33,$E0,$37,$7F,$E7,$96,$C6,$A3,$B6,$3E,$C8,$F8,$93,$FE,$A8,$BF
	.dc.b	$6B,$CE,$BF,$D2,$58,$FA,$27,$31,$CA,$BD,$27,$27,$FD,$35,$E4,$FF
	.dc.b	$A6,$23,$DD,$7E,$9C,$BF,$D2,$D5,$D3,$2F,$55,$7F,$A8,$0F,$F7,$89
	.dc.b	$4F,$F5,$F1,$E9,$3B,$1F,$E9,$AC,$BF,$D2,$4F,$D8,$C7,$CA,$E8,$F0
	.dc.b	$D7,$C3,$3C,$0F,$D5,$70,$F8,$31,$F4,$4F,$28,$FB,$17,$E6,$9C,$E1
	.dc.b	$F5,$B1,$F8,$D5,$C7,$2F,$8D,$39,$C7,$E3,$57,$C1,$CB,$FD,$8B,$7B
	.dc.b	$C7,$A6,$1C,$27,$EA,$4F,$80,$F3,$1F,$04,$C5,$7D,$C7,$E3,$65,$FE
	.dc.b	$C4,$6E,$3F,$EB,$D6,$D7,$C6,$AF,$BD,$3E,$59,$F8,$D5,$EF,$3F,$1A
	.dc.b	$F8,$0E,$27,$E4,$F1,$F9,$35,$E1,$CB,$E1,$E1,$E8,$BE,$33,$9C,$7D
	.dc.b	$98,$7E,$18,$F4,$E1,$C8,$8F,$C3,$3C,$18,$74,$78,$5F,$49,$DC,$7D
	.dc.b	$C0,$3C,$EF,$F4,$3A,$F5,$1E,$AB,$D5,$FF,$5C,$56,$A7,$31,$8E,$3F
	.dc.b	$F5,$FC,$63,$7F,$F5,$65,$6E,$F9,$46,$2B,$71,$8C,$3F,$C7,$B5,$BC
	.dc.b	$3B,$FA,$5F,$EC,$AE,$57,$AC,$F5,$F3,$9F,$FB,$42,$73,$18,$C7,$FC
	.dc.b	$99,$18,$9F,$1D,$F3,$BE,$51,$83,$93,$8A,$F1,$CE,$EF,$A4,$7F,$B9
	.dc.b	$B1,$4F,$FB,$2D,$B7,$FE,$CF,$6C,$FF,$AA,$3C,$9F,$57,$B0,$EE,$FF
	.dc.b	$71,$79,$5F,$EA,$DD,$3F,$DF,$73,$9C,$7F,$DA,$93,$97,$EB,$CA,$75
	.dc.b	$87,$F9,$92,$B4,$FF,$B6,$23,$33,$FE,$BE,$C3,$FD,$7A,$4B,$FF,$5E
	.dc.b	$BF,$EE,$3E,$07,$C1,$FD,$AF,$75,$D3,$FE,$F7,$9C,$DF,$FB,$F9,$4E
	.dc.b	$A3,$FD,$D5,$1E,$2C,$1E,$DF,$E3,$33,$FC,$66,$5F,$6E,$31,$1F,$92
	.dc.b	$E9,$FF,$59,$B4,$FF,$59,$1F,$96,$0E,$9A,$6B,$CA,$FF,$5C,$F7,$7F
	.dc.b	$AD,$9B,$E5,$FF,$B0,$3C,$BF,$F5,$D3,$72,$FF,$60,$FB,$BF,$EB,$1C
	.dc.b	$3F,$0F,$93,$DD,$FF,$4D,$57,$A9,$FF,$4A,$56,$E1,$B7,$96,$1E,$F7
	.dc.b	$FA,$87,$7F,$8B,$EF,$5E,$13,$CA,$7F,$D3,$DE,$7F,$87,$4F,$70,$70
	.dc.b	$FC,$16,$BF,$B8,$6B,$BF,$3F,$FD,$51,$E4,$BC,$1A,$76,$89,$3F,$8A
	.dc.b	$D3,$E6,$56,$E3,$07,$F0,$47,$0F,$FA,$C2,$FE,$04,$FD,$3F,$6D,$EE
	.dc.b	$7F,$B8,$9B,$5E,$A3,$F8,$6B,$87,$0D,$7D,$EE,$77,$F9,$71,$FC,$53
	.dc.b	$CC,$7B,$FF,$EC,$36,$C7,$98,$57,$1C,$3C,$98,$6E,$E5,$6E,$B4,$F9
	.dc.b	$27,$2F,$FB,$32,$FA,$BF,$EC,$E9,$DB,$DB,$C8,$F1,$54,$8E,$91,$FE
	.dc.b	$D7,$9F,$F7,$8E,$1F,$97,$BE,$27,$28,$CD,$7C,$4A,$E4,$F8,$6F,$94
	.dc.b	$7F,$B8,$7C,$7F,$F3,$0F,$4F,$F9,$EB,$7F,$F5,$07,$8F,$FE,$46,$8C
	.dc.b	$61,$FE,$A7,$8C,$3F,$E6,$59,$DB,$FE,$63,$21,$FE,$54,$AF,$F3,$3B
	.dc.b	$67,$8F,$5F,$F9,$56,$3F,$CB,$4F,$F9,$55,$E2,$D7,$81,$8A,$FF,$2B
	.dc.b	$3F,$E6,$4A,$F4,$BE,$33,$CC,$6C,$FF,$36,$9B,$F3,$3F,$CE,$2F,$F9
	.dc.b	$9C,$C9,$4F,$F3,$71,$BC,$3F,$CC,$AE,$A7,$D7,$7E,$52,$5F,$92,$79
	.dc.b	$4F,$92,$7D,$C7,$F9,$AC,$DC,$73,$C3,$E1,$D7,$F9,$C0,$F9,$71,$F0
	.dc.b	$09,$3C,$1F,$29,$DB,$0F,$5D,$F9,$2B,$C4,$D9,$EC,$7C,$33,$8B,$96
	.dc.b	$DF,$89,$F8,$07,$D4,$9E,$71,$98,$F0,$61,$F1,$1E,$27,$5A,$F8,$15
	.dc.b	$EB,$8F,$A6,$7A,$9E,$6E,$EF,$D3,$7D,$6B,$C8,$72,$9F,$1E,$1F,$26
	.dc.b	$F8,$C7,$33,$DA,$53,$C6,$59,$E4,$FA,$0F,$61,$C2,$7E,$64,$7F,$A6
	.dc.b	$E3,$FD,$3C,$40,$E5,$76,$3C,$BF,$F4,$CC,$61,$B1,$9F,$33,$FD,$43
	.dc.b	$3F,$CC,$2B,$FA,$FF,$99,$FE,$9C,$8C,$78,$28,$C4,$7F,$95,$32,$FF
	.dc.b	$30,$BF,$E6,$E3,$FC,$7F,$97,$F9,$71,$C7,$99,$FE,$33,$8F,$F2,$D4
	.dc.b	$7F,$9B,$EF,$FC,$94,$FF,$8F,$0F,$F1,$D4,$7F,$90,$78,$FF,$CB,$A4
	.dc.b	$8F,$F1,$EF,$07,$F9,$12,$79,$9F,$E5,$4C,$7F,$C9,$57,$FE,$86,$AE
	.dc.b	$78,$FF,$90,$32,$FF,$2F,$65,$E3,$8F,$4C,$63,$1E,$91,$8B,$FF,$31
	.dc.b	$56,$EF,$FC,$D5,$7F,$E4,$B7,$A9,$0F,$7D,$F2,$77,$5E,$37,$39,$7A
	.dc.b	$F2,$F7,$47,$B2,$31,$8F,$9E,$BC,$37,$CE,$BE,$6C,$FD,$17,$AE,$5F
	.dc.b	$E8,$47,$A9,$CC,$98,$76,$24,$F5,$9F,$F3,$E3,$FE,$B0,$AE,$D9,$7B
	.dc.b	$EF,$B9,$FE,$7E,$C3,$EB,$E5,$F0,$63,$BE,$3D,$1F,$85,$1F,$2F,$2F
	.dc.b	$5C,$62,$BD,$71,$88,$E3,$8F,$8D,$FF,$43,$1B,$CB,$CA,$FF,$AD,$CC
	.dc.b	$9C,$0D,$18,$BF,$03,$F0,$27,$A6,$1F,$93,$9C,$BF,$87,$9F,$F7,$BF
	.dc.b	$8F,$FE,$6F,$C3,$FC,$E8,$7F,$8F,$7C,$CF,$F3,$A7,$3F,$F7,$4A,$FF
	.dc.b	$43,$70,$FF,$97,$C9,$5F,$E5,$F6,$C7,$F9,$E9,$FF,$41,$FC,$03,$DD
	.dc.b	$FF,$43,$57,$F9,$BE,$BF,$CE,$BE,$6F,$F9,$BF,$0E,$D9,$7F,$9A,$1F
	.dc.b	$F0,$19,$FE,$79,$9F,$F3,$9D,$7F,$9B,$7C,$CF,$F3,$81,$1C,$1D,$67
	.dc.b	$8F,$C0,$BD,$3F,$F4,$24,$76,$AB,$5F,$E6,$CA,$D3,$C4,$FF,$28,$79
	.dc.b	$BF,$E5,$BF,$37,$FC,$D7,$18,$F3,$3F,$CF,$7E,$77,$F9,$E9,$FF,$36
	.dc.b	$67,$FE,$7F,$9F,$F3,$69,$1F,$F3,$B4,$7F,$A6,$FE,$0D,$F0,$7C,$6B
	.dc.b	$FC,$D3,$E8,$7F,$9C,$C8,$FF,$8A,$EF,$FD,$69,$1F,$E7,$7B,$93,$9F
	.dc.b	$2D,$18,$AE,$07,$F9,$9B,$C1,$79,$FF,$E3,$5E,$8F,$F3,$51,$27,$FC
	.dc.b	$D5,$E8,$7F,$99,$9F,$F6,$2F,$83,$FE,$6A,$6C,$65,$FE,$5E,$DF,$3B
	.dc.b	$FC,$A6,$4F,$2F,$C2,$E3,$CC,$A4,$AF,$64,$7F,$9A,$1F,$77,$97,$FE
	.dc.b	$4E,$7F,$CB,$D5,$FE,$58,$5C,$FF,$93,$F3,$3D,$71,$FE,$43,$AF,$F3
	.dc.b	$19,$2B,$E0,$B6,$3F,$C9,$13,$FE,$5C,$AE,$5E,$0F,$57,$F3,$DE,$57
	.dc.b	$F8,$DF,$CA,$FF,$1E,$1F,$E5,$1E,$FB,$87,$F9,$2E,$2B,$FE,$49,$8A
	.dc.b	$FF,$97,$23,$FC,$C8,$47,$ED,$4F,$F9,$13,$CC,$FF,$28,$DF,$82,$BF
	.dc.b	$C8,$35,$FE,$4C,$21,$FE,$55,$F7,$F2,$FF,$9A,$BC,$9F,$F3,$81,$1E
	.dc.b	$AF,$F9,$32,$BF,$71,$E6,$FF,$94,$63,$FC,$99,$E6,$FF,$7C,$AF,$EF
	.dc.b	$DC,$DF,$E0,$7F,$23,$FB,$99,$FD,$FA,$75,$3F,$DD,$DB,$CD,$FD,$D1
	.dc.b	$BB,$FF,$80,$23,$17,$CE,$73,$5F,$DE,$EB,$44,$70,$43,$FC,$0B,$5C
	.dc.b	$9F,$F0,$19,$88,$E0,$D3,$85,$FA,$3C,$7F,$F0,$B1,$23,$DA,$72,$8D
	.dc.b	$38,$AF,$F0,$C3,$CA,$FC,$EF,$52,$4F,$07,$E1,$61,$ED,$23,$D6,$78
	.dc.b	$CE,$EB,$D4,$76,$9E,$73,$A2,$57,$29,$D5,$71,$C3,$DE,$DB,$F3,$56
	.dc.b	$E7,$FC,$A7,$5D,$9A,$70,$BE,$B5,$E5,$AE,$0F,$F9,$02,$B8,$36,$B8
	.dc.b	$4E,$A3,$FC,$5B,$48,$FF,$7B,$4F,$FB,$CE,$D3,$9A,$FF,$3F,$1C,$2B
	.dc.b	$68,$C1,$A9,$E3,$1A,$8C,$DF,$F9,$D2,$FD,$71,$FE,$3F,$AF,$F1,$F1
	.dc.b	$2F,$C1,$1F,$EE,$8E,$7F,$F1,$E3,$FE,$D1,$6B,$9F,$37,$FC,$82,$7F
	.dc.b	$91,$2B,$FC,$7F,$1F,$ED,$76,$BF,$E4,$0F,$2B,$FC,$7E,$FF,$80,$5F
	.dc.b	$F1,$69,$FE,$0F,$8F,$ED,$78,$7F,$6E,$3F,$C2,$93,$FE,$08,$3F,$C7
	.dc.b	$87,$F7,$72,$47,$07,$FB,$CE,$3F,$DD,$7C,$BF,$F0,$2C,$7F,$79,$33
	.dc.b	$5D,$3C,$CF,$F0,$A1,$36,$FF,$0B,$93,$BF,$FC,$19,$8F,$5B,$EF,$87
	.dc.b	$4C,$7F,$C2,$D1,$8C,$BF,$C1,$AF,$F9,$AA,$7F,$C3,$A7,$F8,$DE,$B9
	.dc.b	$11,$E5,$97,$BE,$7E,$20,$BF,$F2,$9C,$7F,$8A,$F0,$F1,$D7,$0A,$F5
	.dc.b	$9C,$2B,$6F,$F8,$75,$F8,$58,$79,$30,$FA,$B1,$A9,$F3,$4F,$D2,$7F
	.dc.b	$CD,$4D,$8F,$F0,$E1,$B9,$E3,$87,$F9,$6E,$7A,$5F,$B9,$FB,$4F,$96
	.dc.b	$FE,$21,$FE,$2B,$71,$5F,$06,$78,$9D,$EB,$DD,$5F,$5C,$E3,$7E,$A7
	.dc.b	$94,$F0,$7E,$34,$F8,$DC,$92,$78,$12,$3F,$01,$8F,$DD,$9E,$D7,$E3
	.dc.b	$78,$4E,$B1,$F9,$F8,$77,$C3,$E1,$63,$EF,$BF,$8E,$78,$8F,$DA,$64
	.dc.b	$C9,$87,$E1,$3C,$97,$07,$F8,$E2,$FF,$B8,$CF,$EE,$FC,$07,$97,$FE
	.dc.b	$3D,$8F,$F2,$4C,$FF,$90,$5B,$C9,$D2,$15,$C3,$A7,$A4,$FF,$90,$30
	.dc.b	$B7,$D7,$0E,$98,$73,$3F,$C8,$24,$AF,$F1,$EF,$95,$FE,$3A,$8F,$E1
	.dc.b	$D9,$7F,$92,$6F,$F8,$84,$23,$0F,$F8,$2B,$C9,$FF,$1F,$BC,$E7,$73
	.dc.b	$A8,$E1,$E5,$7F,$94,$A7,$FC,$8D,$7D,$E7,$31,$FE,$47,$7A,$46,$7C
	.dc.b	$AF,$F1,$9C,$FF,$90,$67,$C6,$23,$2F,$F9,$1C,$98,$FF,$9B,$E3,$2F
	.dc.b	$F9,$8B,$C8,$FF,$23,$D7,$F2,$B8,$D7,$8D,$FE,$49,$BF,$F2,$6C,$E4
	.dc.b	$D5,$70,$8C,$9F,$68,$FB,$5E,$3F,$B5,$B3,$C5,$FA,$FE,$7B,$E1,$35
	.dc.b	$1F,$E5,$29,$D3,$F7,$30,$CF,$17,$A2,$7C,$DE,$CB,$E0,$73,$5C,$AF
	.dc.b	$FA,$27,$47,$F9,$47,$BF,$FC,$77,$1F,$62,$31,$E5,$FD,$AE,$FF,$F2
	.dc.b	$94,$FF,$94,$38,$3F,$A6,$5F,$09,$C9,$AF,$27,$FC,$95,$46,$9F,$E4
	.dc.b	$DA,$7C,$33,$E1,$BF,$0F,$1F,$E9,$F9,$FF,$2A,$57,$C7,$39,$79,$9A
	.dc.b	$DF,$F1,$B3,$AA,$FC,$F7,$03,$8B,$FD,$51,$1F,$EA,$9F,$2F,$C6,$4F
	.dc.b	$33,$96,$5F,$D9,$7C,$CF,$83,$5B,$9F,$F2,$A6,$5F,$9A,$BF,$5D,$6A
	.dc.b	$31,$3F,$E5,$E2,$5F,$F9,$A9,$FF,$28,$3E,$9F,$33,$FC,$B8,$FD,$8F
	.dc.b	$27,$FD,$68,$FF,$96,$C8,$6F,$CB,$FF,$30,$38,$CB,$FC,$AC,$DC,$3F
	.dc.b	$5B,$5C,$A3,$81,$93,$FC,$01,$6B,$E7,$1E,$3F,$0F,$FC,$B7,$87,$F9
	.dc.b	$C1,$F8,$70,$B7,$97,$AC,$7F,$77,$8F,$F1,$7A,$FF,$32,$E1,$FE,$5E
	.dc.b	$AD,$47,$83,$0A,$F3,$8F,$F3,$0F,$27,$F8,$77,$C9,$FF,$08,$CF,$F2
	.dc.b	$33,$FC,$23,$E4,$FF,$87,$7C,$5F,$F2,$54,$7F,$90,$0F,$F1,$51,$3A
	.dc.b	$7F,$C6,$FE,$5F,$F8,$B9,$B8,$75,$8F,$F1,$24,$2F,$FC,$51,$18,$7B
	.dc.b	$57,$F9,$4D,$FF,$0F,$BF,$E4,$6A,$E4,$7F,$89,$70,$F3,$DF,$F8,$96
	.dc.b	$7F,$C7,$07,$F9,$10,$F2,$57,$29,$FF,$24,$47,$33,$27,$03,$13,$E8
	.dc.b	$72,$7C,$23,$FC,$71,$1E,$09,$F3,$C7,$C3,$9E,$04,$8F,$05,$FC,$36
	.dc.b	$DF,$AA,$BE,$75,$F3,$7A,$E1,$F3,$B0,$FF,$29,$C7,$84,$93,$FE,$93
	.dc.b	$AE,$53,$F2,$E7,$CB,$87,$CB,$79,$9F,$2C,$F7,$57,$C2,$24,$79,$63
	.dc.b	$2D,$7B,$BC,$D4,$78,$5F,$8A,$FA,$5F,$2D,$FA,$70,$FF,$34,$C6,$AB
	.dc.b	$95,$6C,$E5,$7F,$1E,$7C,$38,$7B,$2F,$F0,$AF,$DF,$C3,$A4,$FD,$DA
	.dc.b	$E0,$F9,$A7,$4F,$BC,$C1,$FE,$6D,$C3,$C2,$78,$AB,$9C,$7E,$02,$FE
	.dc.b	$AB,$E7,$AF,$C7,$56,$E3,$FD,$52,$7F,$58,$C3,$FC,$E1,$3F,$D6,$A3
	.dc.b	$E8,$B5,$74,$FF,$9C,$1F,$E7,$1E,$5F,$F9,$C3,$D3,$D4,$E6,$7F,$1B
	.dc.b	$E5,$BC,$CF,$F3,$7C,$7F,$8F,$8F,$F3,$A4,$65,$B1,$FE,$38,$3F,$C3
	.dc.b	$D9,$FF,$8A,$F5,$FF,$14,$4F,$F9,$36,$30,$7F,$89,$5F,$F1,$4D,$7F
	.dc.b	$8F,$09,$5C,$A3,$FC,$47,$BF,$F8,$BE,$FF,$CC,$FA,$7F,$8B,$DE,$B3
	.dc.b	$FE,$7C,$8C,$61,$D6,$77,$87,$4A,$FF,$1A,$9E,$33,$AF,$17,$F8,$D9
	.dc.b	$C9,$CE,$38,$4F,$6B,$FF,$2C,$CF,$F9,$96,$FB,$1F,$E7,$17,$86,$3E
	.dc.b	$97,$FD,$21,$3C,$1F,$95,$5F,$E4,$29,$FF,$2C,$47,$33,$C8,$70,$BF
	.dc.b	$F4,$01,$E2,$BE,$F1,$FE,$78,$3B,$9E,$8B,$EC,$77,$F3,$3F,$D2,$98
	.dc.b	$FB,$71,$F5,$61,$D3,$1F,$35,$6B,$0F,$A7,$18,$CB,$E8,$65,$ED,$CB
	.dc.b	$AE,$1F,$E5,$9C,$7F,$CF,$B3,$F7,$30,$F4,$E3,$CE,$7E,$8E,$3C,$F2
	.dc.b	$F1,$5F,$AA,$FF,$CE,$AD,$C7,$EC,$C7,$2C,$BC,$2F,$BE,$3C,$37,$F2
	.dc.b	$23,$E5,$D7,$85,$F8,$78,$7B,$A7,$F2,$91,$F1,$72,$F8,$0F,$F4,$DF
	.dc.b	$27,$FC,$A3,$C6,$F4,$9E,$BF,$F9,$66,$30,$D8,$CE,$BF,$E4,$D9,$CB
	.dc.b	$63,$37,$FE,$59,$13,$A7,$FC,$7E,$FF,$9C,$47,$A6,$F4,$FF,$C9,$11
	.dc.b	$98,$C5,$FF,$8F,$3C,$BF,$F1,$D5,$F4,$9C,$BD,$27,$FC,$A5,$18,$AF
	.dc.b	$F2,$9C,$7F,$92,$5E,$F7,$DA,$36,$FF,$95,$EB,$FC,$C7,$7F,$E6,$26
	.dc.b	$F8,$DF,$E8,$76,$D7,$F0,$DF,$75,$EF,$7F,$95,$F1,$79,$AF,$87,$FF
	.dc.b	$2A,$D6,$AF,$C5,$39,$7F,$CA,$F1,$9C,$3F,$CC,$57,$CA,$F8,$E1,$D7
	.dc.b	$0E,$15,$FE,$41,$BE,$0F,$F9,$E9,$AF,$FA,$39,$A8,$38,$7F,$CC,$D8
	.dc.b	$7C,$5F,$6B,$FD,$5A,$FF,$9A,$BC,$6F,$9F,$92,$73,$5F,$E5,$19,$7C
	.dc.b	$4F,$C1,$F8,$97,$C0,$FF,$9C,$23,$0F,$F9,$B6,$BF,$DE,$83,$BF,$FC
	.dc.b	$DA,$FF,$9C,$0C,$79,$DF,$CE,$FA,$7F,$CD,$F1,$E9,$EE,$FF,$35,$FB
	.dc.b	$3F,$8C,$8D,$99,$71,$7F,$E7,$16,$DF,$F9,$93,$E2,$BF,$C7,$7F,$16
	.dc.b	$F8,$3A,$7F,$8F,$BC,$F4,$DF,$33,$FC,$C7,$C5,$F0,$7E,$02,$F1,$9D
	.dc.b	$6B,$FE,$90,$F0,$1E,$C7,$F5,$18,$46,$1F,$EA,$8D,$7F,$AC,$35,$FF
	.dc.b	$59,$B9,$BF,$C8,$7C,$0B,$E3,$3F,$CA,$6D,$AF,$05,$FF,$95,$AF,$87
	.dc.b	$47,$FB,$3A,$B5,$5F,$E8,$76,$E1,$F9,$BA,$DC,$FE,$17,$D2,$69,$FE
	.dc.b	$78,$8F,$ED,$91,$87,$FC,$F4,$43,$FD,$25,$E9,$7F,$88,$5B,$E3,$FF
	.dc.b	$A1,$9E,$95,$FE,$6A,$9D,$61,$FA,$B6,$E5,$FE,$88,$AF,$F5,$CE,$1F
	.dc.b	$E8,$CC,$38,$1C,$E1,$E8,$7F,$9E,$5F,$F4,$47,$BF,$FE,$76,$F4,$8F
	.dc.b	$F9,$E2,$7F,$D1,$B3,$FE,$A7,$9C,$BC,$0F,$F4,$21,$D3,$D3,$E5,$52
	.dc.b	$33,$E6,$7F,$8D,$BA,$BF,$C4,$91,$FE,$0E,$3F,$C3,$8F,$F8,$58,$FF
	.dc.b	$10,$BF,$E5,$3C,$7F,$C3,$7D,$9F,$E2,$6A,$FF,$18,$7A,$3F,$E1,$BA
	.dc.b	$D1,$FE,$30,$3F,$C4,$07,$F9,$51,$D4,$FF,$94,$A7,$8D,$F6,$8C,$4F
	.dc.b	$F8,$B2,$31,$7F,$E4,$09,$F0,$4E,$6F,$CE,$75,$AD,$BF,$E3,$BA,$F3
	.dc.b	$91,$F2,$57,$29,$F7,$47,$28,$FC,$05,$FF,$92,$FD,$2F,$F1,$DB,$F1
	.dc.b	$E7,$D7,$3A,$C3,$D7,$87,$94,$ED,$3F,$E4,$69,$FF,$45,$36,$3F,$CF
	.dc.b	$B1,$EE,$C3,$C2,$EA,$FE,$1D,$78,$EB,$FC,$AE,$D3,$EB,$4E,$EF,$E1
	.dc.b	$36,$F9,$D7,$63,$C9,$5F,$01,$F2,$DF,$6A,$E1,$3F,$24,$EC,$7F,$9C
	.dc.b	$0F,$F3,$04,$6E,$7A,$47,$CD,$3A,$10,$F7,$1F,$E8,$63,$F0,$66,$0F
	.dc.b	$F3,$FE,$1F,$33,$0F,$F4,$25,$75,$9E,$17,$F1,$5F,$23,$E6,$3E,$86
	.dc.b	$1F,$7A,$3F,$CD,$C7,$86,$79,$3D,$8F,$4D,$7C,$6C,$3D,$31,$C0,$F2
	.dc.b	$4F,$FA,$22,$3E,$2C,$EE,$7C,$71,$FE,$90,$7F,$CE,$D1,$FC,$FD,$FF
	.dc.b	$3B,$4F,$F9,$E1,$FF,$3C,$77,$7F,$9C,$5F,$F3,$BF,$A5,$FE,$75,$72
	.dc.b	$BD,$2F,$F3,$9D,$7F,$2C,$F4,$2F,$37,$F9,$CE,$FF,$D0,$8F,$F8,$7B
	.dc.b	$4F,$F1,$14,$7F,$AF,$A3,$FC,$71,$9F,$F6,$38,$FF,$14,$9F,$E6,$23
	.dc.b	$FC,$82,$7F,$89,$EB,$FC,$53,$5C,$8F,$F1,$FC,$6D,$FF,$1E,$BB,$3B
	.dc.b	$47,$F8,$D6,$FF,$CA,$5E,$8F,$F8,$EA,$7F,$C7,$E7,$F8,$F7,$93,$FC
	.dc.b	$86,$D3,$C3,$7D,$F0,$F3,$5F,$2A,$DD,$7F,$8F,$A7,$FC,$91,$8F,$F9
	.dc.b	$96,$39,$5F,$90,$E4,$FF,$9B,$4E,$06,$A7,$FC,$B9,$3F,$1D,$E7,$8F
	.dc.b	$F9,$5E,$7C,$71,$87,$DD,$5F,$09,$FF,$3B,$4F,$C6,$9D,$4F,$82,$7E
	.dc.b	$69,$A3,$E2,$87,$FD,$04,$D8,$FF,$2D,$61,$FE,$5E,$8C,$1F,$3E,$3C
	.dc.b	$F5,$DE,$3D,$47,$F9,$96,$FD,$4E,$0F,$9E,$7C,$02,$57,$5B,$F0,$92
	.dc.b	$FC,$35,$CA,$B6,$DA,$E1,$3A,$BF,$15,$69,$B3,$A8,$F4,$1E,$58,$CD
	.dc.b	$73,$23,$83,$9F,$A3,$FE,$74,$F4,$3F,$CD,$7D,$1F,$E5,$18,$C7,$A3
	.dc.b	$FE,$73,$26,$9F,$E7,$16,$E3,$FE,$63,$3F,$CC,$DE,$87,$F9,$7D,$B3
	.dc.b	$FE,$6B,$F3,$FF,$CB,$1E,$8F,$F9,$9E,$3F,$49,$E8,$FF,$90,$23,$15
	.dc.b	$FE,$49,$8F,$F3,$23,$AA,$FF,$32,$36,$BF,$DE,$13,$A8,$E1,$3E,$1C
	.dc.b	$BF,$CC,$BF,$01,$E3,$9F,$F2,$9F,$A3,$FE,$5D,$F4,$BF,$CB,$F8,$FF
	.dc.b	$99,$9B,$3F,$E6,$4F,$4B,$FC,$C3,$87,$F9,$CF,$D2,$FF,$34,$3F,$E8
	.dc.b	$B8,$CD,$FF,$98,$70,$E3,$39,$F7,$6C,$FF,$01,$F4,$FF,$CB,$BE,$57
	.dc.b	$F9,$72,$BF,$CF,$D1,$8E,$FF,$F2,$D5,$7F,$A2,$7D,$3F,$F2,$EB,$FE
	.dc.b	$65,$F4,$FF,$CB,$AF,$F9,$B6,$FE,$1E,$BF,$E5,$EF,$07,$FC,$BA,$DF
	.dc.b	$47,$FC,$B8,$FF,$96,$0F,$E2,$5D,$BF,$E7,$59,$FF,$2C,$55,$AF,$E2
	.dc.b	$DE,$0F,$F9,$82,$31,$7F,$E5,$F6,$AF,$3F,$FC,$B2,$F8,$F1,$FE,$3D
	.dc.b	$E8,$FF,$97,$E3,$07,$FA,$0D,$C3,$FE,$B6,$8F,$F3,$33,$A3,$A7,$97
	.dc.b	$FE,$60,$F3,$E4,$66,$FF,$C8,$8F,$13,$FD,$1D,$E0,$7F,$93,$1F,$F3
	.dc.b	$0F,$9D,$FE,$55,$3F,$CB,$25,$7B,$79,$DF,$E5,$4C,$7F,$CB,$1E,$7E
	.dc.b	$67,$A1,$C8,$DB,$FE,$62,$35,$A7,$FA,$D7,$CE,$F1,$B7,$4F,$F2,$47
	.dc.b	$9E,$E0,$FE,$81,$39,$C3,$FC,$B4,$7F,$43,$17,$EC,$70,$D7,$FC,$AC
	.dc.b	$7F,$97,$8F,$8D,$5F,$E5,$A7,$FC,$BE,$7F,$96,$AF,$FC,$B2,$EB,$C7
	.dc.b	$FF,$2D,$79,$FF,$E5,$98,$FF,$2A,$79,$BF,$E4,$F7,$DF,$1F,$E5,$8C
	.dc.b	$7F,$DA,$1E,$7F,$F9,$62,$76,$7F,$96,$27,$94,$EA,$B9,$3F,$ED,$78
	.dc.b	$CE,$BF,$E1,$CA,$FE,$F3,$D3,$FE,$0B,$E6,$FF,$2D,$C7,$F8,$64,$FF
	.dc.b	$08,$F9,$5F,$E0,$43,$FB,$F3,$FE,$07,$F3,$BF,$BF,$3F,$E0,$73,$6E
	.dc.b	$23,$71,$D8,$FF,$05,$CF,$F8,$B9,$FF,$02,$9F,$E2,$33,$47,$27,$4F
	.dc.b	$5E,$9F,$F0,$E1,$FE,$1A,$7F,$C5,$F3,$FE,$4F,$7A,$9F,$E2,$0D,$7F
	.dc.b	$C2,$F1,$FE,$51,$9F,$F2,$C4,$E4,$F0,$9F,$E1,$B8,$F7,$61,$F0,$E3
	.dc.b	$FC,$AF,$5F,$E4,$AA,$FF,$15,$34,$EB,$87,$F8,$C8,$F8,$15,$C6,$7B
	.dc.b	$9E,$73,$8D,$7F,$94,$AB,$8E,$1F,$43,$0F,$54,$FC,$13,$CC,$7C,$19
	.dc.b	$FB,$11,$E6,$8F,$7D,$75,$9E,$71,$E8,$3B,$1E,$E9,$F8,$51,$FE,$36
	.dc.b	$AF,$9B,$1F,$58,$E9,$3E,$1A,$E8,$F1,$C3,$DD,$5C,$30,$F7,$61,$CE
	.dc.b	$BC,$F8,$73,$BF,$87,$1E,$69,$F9,$0F,$3B,$F8,$13,$FE,$82,$8F,$C1
	.dc.b	$57,$8B,$0F,$BF,$7C,$27,$E1,$1A,$AE,$AF,$9E,$35,$4C,$3F,$95,$C7
	.dc.b	$FA,$0A,$7F,$CA,$38,$7E,$F5,$FF,$5C,$18,$3F,$C9,$BE,$6E,$5F,$F2
	.dc.b	$8B,$7A,$BF,$CA,$87,$F9,$57,$0F,$F2,$84,$FF,$96,$B9,$FF,$CA,$0F
	.dc.b	$F8,$D9,$A8,$47,$07,$FC,$AB,$E6,$7F,$7F,$7F,$C4,$7A,$7F,$84,$5F
	.dc.b	$F0,$39,$FE,$18,$9F,$F0,$53,$FE,$29,$8F,$F1,$87,$A8,$F3,$BF,$C9
	.dc.b	$95,$A3,$FC,$8D,$5F,$E3,$83,$FC,$44,$E7,$0F,$F3,$0C,$7F,$90,$E3
	.dc.b	$AD,$73,$DF,$FC,$9B,$18,$BE,$B7,$FE,$4D,$BF,$F0,$F1,$E8,$D7,$FC
	.dc.b	$44,$D3,$D3,$5D,$F2,$FF,$31,$57,$82,$7F,$C8,$78,$F7,$7C,$C7,$4C
	.dc.b	$BD,$B8,$F8,$2F,$FC,$A5,$1F,$E5,$BC,$7F,$C9,$A4,$AF,$6E,$1E,$5C
	.dc.b	$7D,$19,$7A,$A3,$FC,$5B,$3E,$53,$E5,$57,$D7,$C3,$CA,$48,$FF,$27
	.dc.b	$9A,$3E,$34,$F2,$C3,$CB,$7D,$6B,$ED,$46,$2B,$D7,$87,$89,$B7,$C7
	.dc.b	$0F,$04,$76,$CB,$E1,$5F,$B6,$3D,$78,$7C,$0A,$FF,$2F,$C7,$A6,$B7
	.dc.b	$18,$AF,$01,$C6,$7E,$54,$76,$8F,$C6,$47,$AC,$FA,$39,$7C,$D7,$FC
	.dc.b	$F7,$3B,$3E,$DC,$F4,$7B,$19,$8F,$BB,$87,$CF,$3D,$8F,$FA,$37,$CD
	.dc.b	$FF,$2A,$E7,$FE,$53,$3F,$C1,$4B,$C3,$FF,$12,$CA,$31,$3F,$B1,$6C
	.dc.b	$7E,$DD,$AF,$F7,$57,$3C,$7F,$E6,$3A,$57,$F8,$72,$1E,$F3,$B7,$38
	.dc.b	$7F,$82,$B0,$E3,$7F,$E5,$78,$FF,$0C,$5B,$CC,$FF,$2F,$C3,$CC,$FF
	.dc.b	$31,$79,$BC,$AB,$9C,$E7,$CD,$F5,$D7,$F9,$7E,$31,$87,$F9,$8A,$72
	.dc.b	$7F,$98,$7C,$A7,$9B,$FE,$61,$B4,$E6,$3F,$CC,$9E,$67,$F9,$72,$FF
	.dc.b	$C9,$97,$FE,$5C,$C3,$F8,$64,$E5,$62,$9D,$47,$F9,$B2,$BF,$8E,$53
	.dc.b	$CC,$FA,$2D,$8C,$C6,$0F,$F3,$2C,$F4,$AF,$8E,$F8,$23,$E1,$4F,$F9
	.dc.b	$A7,$1F,$B7,$E5,$FC,$12,$38,$F2,$A4,$FF,$2D,$94,$62,$FF,$98,$1F
	.dc.b	$E6,$53,$FC,$65,$CB,$FE,$5C,$7F,$95,$35,$D0,$EA,$FF,$32,$3F,$E6
	.dc.b	$B6,$F6,$7F,$8F,$E3,$F9,$97,$91,$FE,$64,$7F,$A1,$35,$FF,$20,$35
	.dc.b	$FF,$30,$9A,$F0,$FC,$B1,$FE,$41,$8E,$71,$83,$D3,$D1,$FE,$5F,$27
	.dc.b	$37,$F9,$7A,$7F,$CD,$51,$F9,$37,$FA,$73,$5F,$EA,$8D,$7F,$CA,$4D
	.dc.b	$7F,$A5,$1F,$E6,$13,$FC,$C7,$E4,$7F,$99,$DF,$E9,$CD,$75,$C5,$FE
	.dc.b	$5D,$7F,$CD,$3E,$37,$F9,$93,$7F,$F2,$CC,$FE,$6B,$C6,$F8,$D5,$AC
	.dc.b	$BF,$CC,$B5,$A2,$47,$F5,$B9,$DD,$75,$8F,$F3,$2D,$7D,$AB,$F8,$D1
	.dc.b	$FE,$6A,$8C,$1F,$E6,$27,$FC,$D0,$77,$3F,$CB,$5E,$8F,$81,$FE,$D1
	.dc.b	$E2,$FF,$96,$1F,$ED,$5E,$3B,$E0,$35,$09,$FF,$35,$3F,$E7,$96,$BF
	.dc.b	$21,$AF,$4A,$FC,$7C,$75,$85,$7F,$9B,$71,$FF,$35,$47,$FA,$86,$75
	.dc.b	$E2,$7F,$8A,$A3,$FC,$C5,$C3,$FE,$23,$7F,$C3,$C7,$F8,$C0,$FF,$0B
	.dc.b	$9F,$E4,$DF,$1F,$FC,$2F,$C9,$FE,$1C,$8D,$57,$F8,$9E,$B9,$47,$F8
	.dc.b	$6C,$CF,$8F,$FE,$2E,$7B,$9F,$E2,$E7,$0F,$F9,$A2,$BA,$3F,$E7,$96
	.dc.b	$D7,$F9,$31,$B7,$FE,$2F,$9F,$25,$F8,$27,$2F,$EB,$BC,$3F,$F2,$5C
	.dc.b	$7F,$96,$9F,$F1,$6E,$1E,$0C,$3F,$C6,$D1,$FE,$54,$C3,$DA,$DB,$E8
	.dc.b	$DC,$3D,$07,$C6,$3C,$CE,$70,$FF,$27,$C7,$C8,$BF,$9C,$FB,$67,$47
	.dc.b	$89,$E5,$1E,$D8,$CC,$7C,$83,$37,$EF,$9E,$67,$A2,$3F,$C9,$57,$E6
	.dc.b	$69,$E6,$3E,$91,$E6,$33,$7E,$69,$FB,$11,$B8,$F9,$2F,$FA,$52,$BC
	.dc.b	$07,$F9,$A1,$D5,$F9,$9E,$87,$D2,$32,$7E,$12,$3E,$5B,$FE,$56,$3F
	.dc.b	$CB,$44,$30,$7F,$93,$79,$BF,$CA,$BE,$93,$C8,$FF,$2C,$36,$7F,$CB
	.dc.b	$A7,$F9,$5F,$C8,$FF,$2F,$65,$FE,$53,$8F,$F0,$99,$FE,$54,$34,$7F
	.dc.b	$83,$FC,$7F,$F0,$36,$6C,$FF,$C6,$13,$FE,$0E,$3F,$C4,$B7,$FE,$22
	.dc.b	$8F,$F0,$34,$7F,$83,$EF,$FC,$A7,$5F,$E1,$57,$FC,$48,$DE,$8F,$F0
	.dc.b	$4C,$FF,$85,$39,$FF,$C5,$8F,$85,$B3,$FE,$13,$8F,$F3,$65,$7F,$93
	.dc.b	$5F,$F1,$56,$3E,$6C,$3C,$B3,$D9,$FF,$0C,$CF,$98,$E8,$7F,$8D,$31
	.dc.b	$F4,$39,$BE,$CF,$F9,$C7,$1F,$2D,$FA,$AF,$9C,$F9,$CE,$47,$4A,$F6
	.dc.b	$5F,$48,$D4,$61,$D4,$F2,$AE,$91,$EC,$BF,$F1,$65,$F1,$9C,$CF,$8E
	.dc.b	$31,$3F,$E6,$B9,$C9,$F0,$EF,$FC,$7F,$E4,$7F,$91,$AF,$A5,$FF,$92
	.dc.b	$DA,$7F,$90,$0F,$81,$87,$D8,$9E,$11,$DC,$F3,$9C,$2B,$F0,$4F,$F9
	.dc.b	$58,$E4,$7B,$A7,$D9,$87,$94,$97,$FE,$6D,$AF,$2E,$3F,$5E,$F9,$47
	.dc.b	$47,$F1,$87,$33,$C1,$7F,$72,$D7,$FC,$5E,$7F,$CA,$1E,$47,$F9,$33
	.dc.b	$C7,$FF,$23,$C7,$F6,$86,$BF,$EB,$36,$BF,$E5,$18,$FF,$0D,$B9,$6F
	.dc.b	$27,$F9,$2B,$B3,$33,$BD,$BF,$CB,$75,$8B,$FF,$12,$DE,$E7,$27,$FB
	.dc.b	$05,$FF,$23,$F8,$BF,$E4,$52,$78,$3F,$E4,$49,$FF,$23,$1B,$3F,$CA
	.dc.b	$93,$FE,$4F,$F0,$FF,$C8,$F1,$9C,$7F,$D9,$97,$FE,$48,$8C,$47,$F9
	.dc.b	$40,$FF,$21,$BC,$8F,$F2,$4F,$8B,$FE,$4F,$C3,$FC,$9D,$1C,$2B,$95
	.dc.b	$70,$3F,$C7,$D1,$FE,$6B,$BF,$F2,$5F,$8B,$FE,$4F,$C7,$FC,$9D,$E8
	.dc.b	$7F,$91,$AF,$FC,$9C,$7F,$9A,$AB,$FC,$8E,$E2,$7F,$CC,$F5,$FE,$57
	.dc.b	$8C,$5F,$F9,$73,$4E,$5E,$2F,$8A,$FF,$CD,$AD,$9F,$95,$EE,$BD,$DF
	.dc.b	$A7,$39,$C3,$FD,$DC,$FF,$93,$2B,$E1,$B6,$3F,$CF,$BE,$8B,$D3,$E3
	.dc.b	$E4,$FD,$C9,$FF,$26,$8E,$FF,$F1,$1B,$86,$BF,$E5,$03,$1E,$7F,$F8
	.dc.b	$D3,$BF,$FC,$9E,$DF,$5F,$FC,$E7,$3F,$E5,$36,$D7,$F8,$BB,$CB,$FF
	.dc.b	$29,$6F,$FE,$8A,$6E,$BF,$E5,$08,$FF,$46,$FA,$6F,$86,$FF,$0F,$D7
	.dc.b	$F8,$E7,$1F,$F2,$89,$F8,$07,$14,$F7,$9E,$F7,$F8,$9B,$BB,$FC,$A7
	.dc.b	$7F,$E5,$2E,$DF,$F2,$A3,$F7,$CF,$F2,$3E,$7F,$E4,$E9,$DC,$7F,$93
	.dc.b	$A7,$91,$FE,$30,$F3,$FF,$C9,$C7,$F9,$3E,$7F,$CB,$58,$FF,$93,$C9
	.dc.b	$7F,$E4,$8B,$79,$FF,$E4,$88,$C7,$9F,$EB,$9F,$F2,$31,$F9,$A9,$D2
	.dc.b	$32,$7F,$91,$C9,$5F,$E5,$9E,$1F,$F2,$1D,$7F,$94,$09,$E7,$FF,$94
	.dc.b	$2B,$FC,$D2,$7F,$90,$FC,$EF,$F2,$4F,$9B,$FE,$41,$3F,$C8,$5E,$77
	.dc.b	$F9,$36,$7F,$CA,$67,$F9,$06,$7F,$C9,$F1,$98,$F9,$13,$FE,$4C,$7F
	.dc.b	$CA,$53,$FE,$4C,$23,$FD,$BB,$CD,$FF,$01,$79,$1F,$E0,$9F,$3B,$FC
	.dc.b	$55,$1F,$DF,$A7,$FC,$05,$39,$7F,$C3,$3C,$FF,$C7,$5F,$F1,$24,$F2
	.dc.b	$F2,$7F,$C4,$51,$FE,$16,$F3,$3F,$C4,$4F,$F8,$5A,$33,$7F,$E1,$43
	.dc.b	$F9,$64,$FF,$86,$1E,$D3,$C6,$BF,$C2,$F7,$DE,$BF,$C4,$D3,$E4,$BE
	.dc.b	$67,$F8,$66,$3B,$C6,$4F,$F1,$81,$CE,$79,$61,$EE,$7F,$C5,$D1,$B3
	.dc.b	$FC,$43,$3F,$E3,$4B,$F8,$4D,$9D,$B6,$7F,$C4,$A7,$17,$C2,$67,$0F
	.dc.b	$59,$EF,$C3,$FC,$B2,$7F,$AA,$23,$D6,$6D,$F5,$CF,$C7,$3D,$33,$C5
	.dc.b	$FF,$06,$E1,$E9,$7C,$35,$C5,$DD,$7C,$3C,$3E,$4C,$2F,$FC,$D0,$E8
	.dc.b	$EC,$7C,$08,$FB,$4E,$4F,$63,$CA,$38,$C6,$F0,$FA,$93,$F1,$8E,$31
	.dc.b	$93,$C1,$5E,$F8,$D4,$F0,$3E,$5B,$FC,$76,$7F,$46,$2D,$19,$8F,$D3
	.dc.b	$C6,$27,$F3,$47,$F9,$33,$CE,$93,$FE,$4C,$3F,$5B,$D8,$F3,$BF,$CA
	.dc.b	$11,$FE,$4B,$21,$FE,$4E,$21,$FE,$00,$C7,$FB,$E7,$99,$FE,$1C,$3F
	.dc.b	$C0,$CF,$FB,$83,$CE,$FF,$09,$E5,$FE,$1D,$8C,$79,$DF,$E0,$9B,$FF
	.dc.b	$03,$CF,$F8,$8B,$A3,$FC,$53,$1F,$E1,$83,$FC,$35,$B7,$F8,$F2,$79
	.dc.b	$47,$F0,$DC,$3A,$D7,$0A,$D3,$70,$FF,$12,$61,$DF,$0F,$F2,$66,$1F
	.dc.b	$E1,$43,$D4,$43,$C7,$5F,$E2,$07,$FC,$AB,$D1,$FE,$5A,$BE,$6D,$8F
	.dc.b	$F1,$89,$0F,$F1,$01,$E8,$8D,$BF,$E4,$EC,$3C,$B9,$7F,$8E,$E3,$47
	.dc.b	$1C,$3E,$14,$65,$F7,$C7,$F9,$CF,$0F,$25,$7F,$8C,$B0,$F1,$4F,$82
	.dc.b	$B7,$97,$C4,$C3,$96,$1F,$0E,$BB,$3C,$B1,$FA,$78,$72,$C3,$C3,$5F
	.dc.b	$E4,$46,$E3,$FC,$4A,$FF,$CD,$67,$19,$FA,$59,$7C,$D8,$CE,$3F,$82
	.dc.b	$9C,$E1,$D6,$78,$47,$AE,$74,$D3,$E7,$9F,$1F,$1F,$93,$87,$86,$7D
	.dc.b	$67,$F2,$DF,$2F,$FC,$15,$D0,$F4,$9E,$97,$F9,$02,$73,$18,$CF,$FC
	.dc.b	$51,$E7,$7F,$91,$A3,$19,$FF,$6B,$AD,$DF,$1A,$FF,$1F,$4F,$EC,$27
	.dc.b	$FC,$83,$7F,$B2,$8F,$F1,$18,$C3,$FC,$5B,$1F,$E3,$CF,$33,$FC,$61
	.dc.b	$E6,$FF,$8C,$BD,$2F,$F2,$AB,$67,$FC,$6D,$E7,$FF,$8C,$EB,$7E,$7D
	.dc.b	$BF,$F2,$CC,$62,$BA,$47,$F9,$73,$E3,$5E,$87,$49,$C3,$5F,$F2,$6C
	.dc.b	$69,$FF,$22,$1A,$29,$A8,$C6,$7F,$E4,$BC,$DF,$1C,$F9,$0C,$C7,$F8
	.dc.b	$FB,$CE,$FF,$37,$FA,$0F,$7B,$FC,$6B,$97,$F9,$7E,$93,$98,$FF,$1E
	.dc.b	$FB,$DF,$E7,$6C,$7F,$C8,$59,$3D,$34,$E7,$93,$FC,$7D,$3F,$E3,$92
	.dc.b	$74,$7F,$83,$7C,$1F,$F1,$EB,$7C,$9F,$F1,$D6,$96,$BF,$DD,$DE,$97
	.dc.b	$F8,$EA,$73,$1F,$E3,$D9,$D2,$F4,$1F,$15,$FE,$19,$8F,$F1,$77,$A6
	.dc.b	$F6,$B1,$7F,$E3,$E7,$FC,$7B,$5F,$E3,$EF,$7F,$3C,$3C,$6F,$D7,$87
	.dc.b	$F9,$F9,$AF,$F8,$E7,$1F,$F2,$4B,$FE,$27,$F9,$0F,$E6,$4F,$FA,$11
	.dc.b	$AF,$F8,$FD,$B9,$2F,$CD,$A3,$E4,$7C,$37,$CF,$C8,$AA,$5E,$9E,$1B
	.dc.b	$1D,$E7,$78,$F6,$C3,$AD,$EC,$CC,$FF,$93,$2B,$FC,$A8,$43,$FC,$A0
	.dc.b	$5C,$7C,$CF,$F9,$0F,$C6,$FF,$2A,$C6,$23,$C7,$EA,$73,$7F,$CB,$97
	.dc.b	$FE,$56,$F7,$B1,$5B,$BF,$F2,$BE,$3F,$56,$B5,$7D,$E3,$EC,$7A,$9F
	.dc.b	$E4,$A8,$FD,$B3,$5F,$ED,$2F,$F9,$67,$D5,$FF,$6C,$61,$E3,$6D,$7F
	.dc.b	$97,$BD,$5F,$93,$18,$BF,$F2,$E3,$5F,$F2,$ED,$7A,$FC,$DF,$F0,$74
	.dc.b	$7F,$7D,$E7,$FF,$01,$C7,$F8,$83,$D1,$FF,$00,$B7,$A7,$FB,$F5,$7F
	.dc.b	$85,$A7,$27,$F8,$2F,$CD,$FF,$02,$4F,$F8,$08,$DD,$FF,$8A,$23,$FC
	.dc.b	$35,$1F,$E3,$8A,$FF,$28,$9F,$E0,$C3,$FC,$46,$62,$7F,$C9,$7E,$97
	.dc.b	$F8,$95,$E5,$1F,$E1,$EC,$3C,$8F,$13,$FC,$1D,$87,$F8,$51,$C4,$7F
	.dc.b	$8D,$23,$CF,$7E,$19,$FF,$25,$1A,$BE,$D8,$7F,$8E,$23,$FC,$4F,$3B
	.dc.b	$C3,$FC,$71,$1F,$E6,$68,$F6,$CE,$F0,$F9,$67,$92,$BE,$5E,$1F,$16
	.dc.b	$FF,$D1,$D3,$E3,$BE,$31,$FE,$35,$8E,$33,$B8,$F5,$CF,$CB,$7D,$71
	.dc.b	$ED,$AE,$B3,$E5,$BF,$9B,$5E,$B3,$E5,$3F,$E5,$67,$C3,$5C,$63,$EB
	.dc.b	$C7,$DB,$9F,$1B,$F8,$38,$CD,$7D,$08,$F8,$67,$53,$E4,$BF,$59,$EE
	.dc.b	$7D,$A3,$31,$E1,$7D,$67,$73,$D7,$3E,$D9,$F5,$C7,$57,$E7,$DF,$FA
	.dc.b	$2E,$BE,$75,$FD,$AA,$D4,$79,$88,$FC,$B8,$FF,$27,$3E,$E3,$FC,$9E
	.dc.b	$E1,$F8,$77,$FE,$4E,$7F,$C9,$BC,$5F,$E4,$C7,$FD,$E0,$D7,$FD,$8D
	.dc.b	$E8,$FF,$94,$7D,$2F,$ED,$98,$B8,$9E,$86,$1B,$7F,$E6,$29,$D4,$E4
	.dc.b	$91,$FE,$49,$8F,$F1,$1F,$A3,$FE,$15,$9F,$F2,$B4,$FF,$83,$B5,$FF
	.dc.b	$06,$3F,$E1,$FB,$FF,$0A,$FA,$5F,$E2,$98,$C5,$7F,$85,$A3,$15,$C0
	.dc.b	$FF,$1D,$CE,$8E,$9B,$7F,$A6,$6F,$FC,$61,$1F,$E3,$56,$E3,$FE,$2C
	.dc.b	$9F,$F3,$4C,$F9,$E7,$27,$8B,$1F,$44,$67,$1F,$F1,$0B,$DF,$0F,$F2
	.dc.b	$13,$FE,$27,$3F,$C7,$B1,$E7,$9E,$33,$E6,$3D,$17,$FE,$2C,$3C,$D3
	.dc.b	$D2,$33,$5C,$8F,$59,$FE,$67,$3F,$CC,$B1,$D7,$8F,$FC,$78,$D8,$F1
	.dc.b	$1E,$F7,$A5,$7B,$AB,$C4,$D3,$C5,$3D,$71,$FF,$3A,$BF,$E6,$43,$91
	.dc.b	$EF,$8F,$F2,$BB,$DA,$3E,$53,$FE,$79,$78,$9F,$80,$BF,$2C,$7A,$8E
	.dc.b	$91,$C1,$F5,$99,$78,$3D,$71,$FA,$53,$FE,$ED,$7F,$23,$6B,$FF,$03
	.dc.b	$7A,$1F,$E4,$8A,$FF,$26,$9F,$E2,$1F,$4B,$FC,$9B,$C7,$FD,$7E,$7F
	.dc.b	$C9,$38,$FF,$92,$FC,$3F,$F2,$6B,$FE,$0F,$11,$FE,$24,$3F,$CB,$BE
	.dc.b	$8B,$DB,$FF,$2A,$63,$FE,$42,$3F,$C9,$25,$AF,$F1,$C4,$FF,$93,$5F
	.dc.b	$F6,$3F,$A0,$F4,$3F,$B8,$47,$F8,$76,$31,$5F,$E5,$B6,$C7,$09,$FF
	.dc.b	$28,$9F,$E5,$8F,$47,$FC,$A9,$5D,$F3,$FE,$FA,$30,$4F,$09,$FF,$26
	.dc.b	$1F,$E4,$EB,$FF,$33,$3F,$E2,$DF,$47,$FC,$A9,$1E,$88,$FF,$28,$BF
	.dc.b	$0E,$32,$DF,$47,$D2,$84,$69,$E9,$1F,$E5,$69,$D6,$1F,$E5,$AF,$47
	.dc.b	$15,$C7,$CD,$7A,$3D,$23,$FC,$A4,$D8,$CD,$FF,$9F,$DB,$3F,$EF,$27
	.dc.b	$FC,$AF,$19,$F3,$B8,$11,$FF,$43,$4F,$9F,$93,$FC,$9E,$D1,$DB,$FE
	.dc.b	$5A,$6F,$9D,$FE,$FE,$9F,$F2,$AF,$6F,$F8,$E7,$4B,$E4,$7F,$96,$EF
	.dc.b	$FC,$BB,$18,$3F,$CB,$7E,$57,$F9,$6A,$3F,$D2,$CD,$7F,$1C,$7F,$97
	.dc.b	$FC,$DF,$2C,$66,$3F,$24,$BC,$DC,$92,$39,$BE,$CF,$2F,$FC,$8D,$1F
	.dc.b	$E4,$F3,$FC,$94,$7D,$0D,$7F,$CA,$BE,$67,$F9,$22,$3F,$D5,$13,$FE
	.dc.b	$5C,$7F,$40,$D7,$FC,$E0,$D7,$A1,$F8,$83,$F0,$19,$7F,$96,$1C,$1F
	.dc.b	$E5,$DF,$2B,$B3,$8E,$FF,$F2,$7E,$1F,$E7,$B7,$FC,$93,$1F,$89,$9C
	.dc.b	$BF,$E5,$5F,$33,$FC,$9F,$5F,$E4,$D7,$FC,$A9,$1F,$B5,$A4,$7D,$C6
	.dc.b	$C7,$FB,$0D,$AF,$CE,$7B,$BF,$E7,$0F,$33,$FC,$E7,$67,$BC,$7E,$6F
	.dc.b	$CC,$79,$9F,$E7,$86,$C7,$CB,$35,$1F,$E5,$4E,$3D,$3F,$8E,$9F,$F2
	.dc.b	$F4,$FF,$0B,$7F,$CC,$1C,$3F,$E5,$DF,$0F,$FC,$B6,$70,$9E,$87,$F8
	.dc.b	$4B,$93,$FC,$20,$FF,$83,$8F,$F0,$83,$FE,$08,$F2,$FF,$C7,$71,$FE
	.dc.b	$03,$E6,$FF,$0E,$F8,$BF,$E3,$5B,$FF,$03,$CF,$F8,$7C,$FF,$01,$9F
	.dc.b	$E1,$5B,$FF,$09,$9D,$8E,$8F,$F8,$78,$FF,$21,$1F,$E1,$A3,$84,$74
	.dc.b	$9E,$D1,$D2,$7F,$C5,$51,$FE,$3E,$AF,$F1,$C4,$FF,$96,$FC,$BF,$F2
	.dc.b	$04,$FF,$91,$63,$A5,$70,$9F,$F1,$03,$E3,$BF,$F1,$FB,$67,$95,$F4
	.dc.b	$7E,$34,$FF,$8C,$6F,$A5,$7B,$23,$17,$FE,$43,$AF,$34,$7F,$8B,$AF
	.dc.b	$E3,$CF,$90,$D4,$FF,$8E,$AB,$C9,$3A,$3C,$17,$ED,$9F,$F1,$A9,$E4
	.dc.b	$29,$E4,$3E,$BD,$FA,$8F,$1C,$FF,$90,$DF,$95,$7E,$18,$C4,$F9,$8F
	.dc.b	$B8,$7C,$69,$F4,$34,$CD,$FB,$DE,$E7,$8D,$F9,$0F,$8C,$FB,$6E,$DB
	.dc.b	$1E,$29,$F6,$D7,$D2,$8F,$3C,$FE,$00,$F2,$1F,$8B,$3E,$33,$E4,$AF
	.dc.b	$EE,$51,$AB,$FF,$4C,$BF,$E5,$4B,$FF,$25,$DF,$F9,$56,$3F,$CC,$E7
	.dc.b	$F9,$53,$CA,$FF,$2B,$3F,$E5,$46,$BF,$E8,$CE,$0F,$F2,$7B,$FD,$D6
	.dc.b	$FF,$C9,$EF,$F9,$46,$32,$FF,$8A,$8C,$79,$1F,$E4,$EC,$7F,$C6,$A7
	.dc.b	$F8,$B2,$3F,$C1,$EE,$32,$E0,$E6,$FF,$C2,$87,$F8,$6A,$FF,$C8,$77
	.dc.b	$FE,$2D,$9D,$47,$F8,$4A,$BF,$C5,$37,$FE,$1C,$8F,$F0,$BB,$FE,$19
	.dc.b	$AE,$13,$D8,$FF,$18,$47,$F8,$F6,$BC,$4F,$F8,$C6,$BA,$39,$9F,$F1
	.dc.b	$7C,$F8,$2B,$D9,$18,$3A,$4F,$F8,$C6,$B8,$BF,$E3,$8A,$E0,$FF,$8B
	.dc.b	$63,$D6,$DA,$FF,$22,$BC,$5E,$AF,$F7,$DF,$2B,$FC,$DF,$7E,$88,$E8
	.dc.b	$DC,$3E,$3D,$F8,$1B,$7C,$A9,$7C,$9E,$C7,$F9,$AC,$FF,$26,$1D,$FC
	.dc.b	$AF,$F3,$3E,$1F,$32,$FF,$C8,$78,$FA,$8F,$F2,$CD,$72,$6B,$F7,$27
	.dc.b	$E2,$3F,$2F,$0F,$16,$1E,$18,$D1,$DF,$97,$FC,$AB,$7F,$4A,$3E,$D5
	.dc.b	$7D,$87,$91,$E7,$8F,$15,$F6,$8D,$9F,$2C,$F5,$9F,$03,$1F,$5B,$FE
	.dc.b	$79,$C3,$C5,$E4,$7F,$94,$A3,$FC,$A0,$D7,$FC,$B7,$E4,$FF,$93,$9F
	.dc.b	$F2,$5F,$93,$78,$BF,$CB,$53,$74,$FF,$16,$F8,$FF,$E5,$57,$FC,$C3
	.dc.b	$7F,$DD,$3C,$9F,$F2,$5B,$FE,$5D,$6B,$FE,$51,$F2,$7F,$C6,$11,$88
	.dc.b	$FF,$1E,$3F,$E5,$2C,$15,$B3,$FB,$13,$FE,$4F,$F2,$7F,$CA,$75,$DE
	.dc.b	$FF,$CA,$D5,$FE,$55,$38,$46,$5E,$38,$7F,$95,$9D,$47,$F9,$2F,$C9
	.dc.b	$FF,$29,$E1,$FE,$5C,$AF,$0D,$7F,$96,$AB,$FC,$AC,$D9,$F2,$C7,$1F
	.dc.b	$27,$FC,$C7,$3F,$E5,$0F,$3D,$E5,$7F,$94,$2B,$FC,$AF,$5F,$E5,$2B
	.dc.b	$FF,$2A,$5F,$33,$FC,$B9,$3F,$E5,$27,$FB,$AF,$93,$FE,$50,$9F,$F3
	.dc.b	$7F,$A0,$F6,$BB,$DF,$F9,$6F,$CC,$AA,$FF,$CA,$AF,$F8,$99,$B7,$FE
	.dc.b	$78,$CD,$EA,$F0,$9C,$E5,$FE,$51,$BF,$F2,$99,$FE,$95,$E8,$FF,$2C
	.dc.b	$A6,$9F,$E5,$17,$63,$CC,$FF,$2C,$38,$6F,$7B,$C3,$F7,$C6,$1F,$83
	.dc.b	$E5,$FF,$95,$F4,$F9,$B7,$FE,$57,$3F,$C2,$F1,$9C,$7E,$3F,$A2,$F7
	.dc.b	$7E,$4D,$F0,$AF,$F2,$D5,$F0,$F3,$73,$87,$66,$C7,$FA,$2F,$07,$B7
	.dc.b	$FE,$4A,$C3,$C3,$E6,$FF,$97,$A8,$DF,$35,$E7,$6F,$2F,$F2,$CF,$A5
	.dc.b	$6F,$FC,$AF,$35,$FF,$2B,$1F,$E3,$0A,$FF,$2D,$61,$DA,$BF,$C8,$72
	.dc.b	$F1,$FF,$CA,$AB,$CE,$B1,$86,$F9,$FC,$1F,$F2,$7C,$7F,$94,$A7,$FC
	.dc.b	$B1,$3F,$E5,$72,$3F,$E8,$21,$E6,$FF,$93,$9F,$F3,$5B,$70,$FA,$2D
	.dc.b	$CF,$F2,$BE,$9F,$D9,$AF,$F2,$4C,$E5,$FF,$2A,$98,$F3,$FF,$CA,$CD
	.dc.b	$3C,$11,$FE,$56,$7F,$3E,$F8,$5F,$EF,$07,$F8,$3C,$FF,$00,$79,$9F
	.dc.b	$E2,$EF,$2F,$FC,$11,$1F,$E2,$08,$FF,$12,$74,$FF,$84,$1B,$3F,$E0
	.dc.b	$A8,$FF,$15,$CF,$39,$FF,$24,$BF,$E0,$A3,$FC,$37,$3D,$27,$97,$9D
	.dc.b	$FE,$13,$9F,$F0,$E9,$FE,$1B,$A7,$9F,$FE,$62,$8F,$2C,$FF,$8E,$A7
	.dc.b	$CD,$1F,$E1,$38,$FF,$17,$3F,$E1,$B3,$C0,$70,$EB,$FF,$18,$B8,$9F
	.dc.b	$29,$C9,$C4,$F2,$BF,$59,$FE,$2B,$9F,$1C,$FF,$9A,$5A,$FF,$9E,$4F
	.dc.b	$F1,$79,$E3,$AF,$F1,$79,$FE,$4D,$7E,$55,$7B,$DD,$9E,$08,$DB,$F2
	.dc.b	$63,$C5,$87,$F8,$C4,$F9,$07,$CF,$31,$1D,$CE,$B3,$DE,$FE,$6B,$67
	.dc.b	$84,$FD,$23,$84,$E9,$F0,$4F,$8A,$3E,$06,$1F,$47,$0F,$8F,$87,$D1
	.dc.b	$25,$7C,$C8,$F9,$44,$8C,$DF,$D3,$8C,$4E,$E7,$37,$F3,$B0,$F6,$DF
	.dc.b	$E0,$AF,$F1,$AD,$AF,$91,$1F,$42,$BF,$D3,$73,$F4,$63,$15,$FC,$F1
	.dc.b	$57,$F9,$4D,$B7,$FE,$56,$25,$7F,$BD,$BA,$DC,$4F,$3F,$FC,$AB,$3F
	.dc.b	$CB,$63,$0E,$DA,$FF,$95,$F8,$BF,$CA,$AF,$F9,$6F,$CF,$FE,$01,$8F
	.dc.b	$F9,$6A,$35,$18,$9F,$F1,$51,$FE,$4A,$D3,$FC,$1F,$3F,$E4,$33,$FC
	.dc.b	$30,$7F,$93,$4F,$F0,$83,$A7,$FC,$21,$C3,$FE,$1C,$D7,$FC,$2D,$87
	.dc.b	$F8,$A9,$FF,$23,$CF,$F8,$88,$D9,$CE,$BF,$C4,$53,$B8,$FF,$10,$C7
	.dc.b	$F8,$BA,$3F,$CC,$98,$7F,$92,$27,$97,$9D,$FE,$2D,$9C,$C7,$F8,$8A
	.dc.b	$31,$8F,$B2,$BF,$C5,$93,$FE,$28,$BE,$F3,$FE,$54,$6C,$F5,$BE,$71
	.dc.b	$ED,$BF,$F1,$D1,$3C,$FF,$F0,$0C,$7F,$90,$A3,$19,$31,$FF,$1C,$65
	.dc.b	$F0,$B2,$FF,$32,$D6,$A7,$27,$CC,$BE,$58,$74,$BF,$F2,$95,$F0,$8C
	.dc.b	$57,$F8,$ED,$AF,$F9,$90,$E5,$3C,$32,$F1,$5F,$FA,$9A,$7E,$7B,$DB
	.dc.b	$1C,$D6,$E7,$ED,$BE,$7A,$CC,$74,$3C,$13,$ED,$3F,$C2,$EF,$DA,$6E
	.dc.b	$2C,$3F,$C4,$B1,$86,$CF,$F9,$66,$75,$3D,$A3,$FC,$EF,$E9,$7B,$27
	.dc.b	$31,$FE,$0F,$C3,$85,$E6,$B4,$CB,$E5,$7C,$31,$F8,$0F,$F8,$89,$B3
	.dc.b	$C6,$76,$DC,$3B,$1E,$F3,$C7,$1A,$3A,$DF,$78,$D9,$A2,$DF,$87,$0F
	.dc.b	$F1,$DB,$E1,$7E,$04,$FC,$29,$FF,$4F,$35,$E5,$5F,$2A,$35,$AB,$E4
	.dc.b	$9E,$2F,$9F,$3F,$F7,$0C,$E6,$31,$8F,$E4,$63,$18,$7D,$7C,$71,$87
	.dc.b	$28,$C4,$7F,$0D,$8C,$3F,$E3,$09,$E2,$FF,$8C,$BC,$37,$BB,$2F,$F0
	.dc.b	$B1,$8F,$83,$FC,$AB,$A9,$A2,$2B,$F0,$63,$53,$E2,$6D,$77,$9E,$CD
	.dc.b	$F2,$3F,$CB,$F3,$E2,$F2,$ED,$7F,$98,$EF,$F6,$1D,$9F,$E7,$4A,$D4
	.dc.b	$60,$FD,$A3,$EA,$BF,$F2,$6E,$18,$BE,$4F,$E7,$27,$55,$F9,$E8,$FF
	.dc.b	$36,$4E,$9F,$F3,$BC,$FC,$4F,$73,$C7,$C5,$E4,$6D,$F8,$6B,$8C,$61
	.dc.b	$63,$E7,$C1,$39,$8C,$7C,$5F,$BB,$E2,$3D,$ED,$C3,$A6,$5F,$52,$3D
	.dc.b	$78,$F9,$9B,$19,$9F,$53,$67,$D9,$F1,$AF,$88,$FF,$45,$F0,$3C,$47
	.dc.b	$C0,$FF,$A3,$2B,$58,$FA,$EB,$53,$9B,$F0,$E0,$BD,$BD,$A3,$31,$88
	.dc.b	$FF,$4F,$D7,$F5,$DF,$93,$E5,$87,$58,$46,$27,$E3,$CF,$E7,$67,$2F
	.dc.b	$CD,$8F,$05,$FA,$A7,$31,$C0,$FE,$69,$7E,$4D,$7F,$DB,$77,$FE,$CC
	.dc.b	$F7,$5F,$03,$FE,$DD,$BF,$AB,$18,$9F,$F6,$EC,$7F,$BA,$3D,$AF,$7F
	.dc.b	$B9,$FE,$E9,$8F,$F5,$30,$F9,$57,$C0,$7F,$A5,$3E,$57,$FD,$8F,$18
	.dc.b	$9D,$B5,$FC,$1C,$6A,$7F,$D6,$53,$9C,$D3,$9A,$E3,$5D,$71,$FE,$A1
	.dc.b	$7C,$E7,$FA,$09,$FE,$AF,$9D,$34,$F3,$FA,$BD,$6F,$FD,$B6,$F1,$8F
	.dc.b	$F5,$F1,$EB,$9E,$35,$9E,$FE,$F1,$83,$CE,$59,$F2,$9F,$B4,$BF,$43
	.dc.b	$FE,$DE,$9E,$2E,$70,$AE,$4F,$01,$92,$99,$BF,$44,$70,$8F,$7C,$61
	.dc.b	$D9,$8B,$F2,$3C,$5A,$E4,$ED,$1E,$D7,$BE,$3F,$42,$BB,$47,$C4,$9E
	.dc.b	$C6,$CB,$5F,$0A,$39,$B5,$FB,$B3,$CA,$FD,$F7,$D6,$7C,$E7,$5A,$F2
	.dc.b	$BF,$4B,$C7,$FF,$58,$3F,$15,$AF,$DE,$6B,$BC,$BF,$DD,$BC,$9F,$C3
	.dc.b	$A3,$FD,$E4,$D9,$E7,$E3,$FF,$BC,$67,$FC,$30,$D7,$FD,$F5,$3C,$73
	.dc.b	$FC,$B6,$0F,$96,$FF,$0F,$57,$D4,$8C,$3F,$6A,$FF,$E0,$0D,$7E,$75
	.dc.b	$B3,$C0,$DC,$F1,$AE,$75,$BB,$F0,$57,$AB,$2F,$C2,$BC,$27,$C7,$5C
	.dc.b	$23,$57,$F8,$6C,$3B,$BF,$E3,$38,$E0,$78,$67,$84,$7C,$0C,$3B,$B7
	.dc.b	$0E,$F3,$A8,$FD,$DB,$6F,$AC,$7C,$69,$DB,$0F,$15,$74,$74,$7B,$CD
	.dc.b	$61,$EF,$9F,$91,$3B,$C3,$CF,$5E,$9C,$3C,$13,$CF,$0F,$0D,$6A,$BC
	.dc.b	$A7,$2B,$8F,$EF,$CC,$96,$3E,$8E,$5E,$18,$D5,$74,$C3,$E0,$57,$58
	.dc.b	$F7,$99,$9F,$4C,$7D,$0C,$7F,$8D,$79,$7F,$EF,$D9,$FF,$31,$35,$2F
	.dc.b	$33,$F1,$39,$36,$FC,$5F,$9B,$FE,$57,$8C,$5F,$F9,$AE,$31,$3F,$90
	.dc.b	$BE,$53,$FD,$4F,$C8,$FF,$56,$BF,$D0,$FE,$45,$DD,$DA,$FF,$2F,$18
	.dc.b	$9E,$1E,$6E,$5F,$F5,$75,$F2,$7F,$A4,$09,$E0,$E1,$B9,$79,$BA,$95
	.dc.b	$C6,$3D,$35,$C7,$1F,$F0,$53,$FE,$85,$BE,$6F,$F9,$A9,$AF,$9A,$36
	.dc.b	$F1,$3F,$B8,$78,$1F,$EF,$EA,$F0,$37,$2E,$13,$FE,$04,$F0,$5F,$2F
	.dc.b	$E0,$F0,$7F,$C7,$2D,$8C,$CF,$F8,$11,$AF,$F8,$85,$B8,$7F,$85,$7A
	.dc.b	$FE,$35,$FF,$90,$9B,$3A,$C3,$FC,$55,$CB,$8F,$2E,$63,$E3,$8D,$98
	.dc.b	$7F,$BA,$61,$FE,$4C,$9C,$C7,$A4,$F3,$10,$F3,$14,$E3,$6F,$67,$C6
	.dc.b	$79,$23,$CF,$F3,$0F,$99,$FE,$2F,$97,$F8,$E2,$31,$3F,$E5,$16,$BF
	.dc.b	$E5,$F6,$C6,$E7,$FA,$C7,$99,$FE,$4A,$97,$C2,$E2,$D9,$A7,$31,$FE
	.dc.b	$66,$7A,$F7,$FF,$9E,$3E,$6B,$FC,$1F,$3F,$E7,$2E,$64,$62,$3F,$CE
	.dc.b	$4D,$CF,$FC,$4E,$D8,$FF,$3F,$3B,$AF,$F5,$F7,$A4,$F8,$7F,$25,$23
	.dc.b	$0F,$FA,$F3,$0E,$FE,$36,$A7,$33,$FE,$64,$6E,$3F,$EB,$E6,$C7,$F4
	.dc.b	$06,$BF,$ED,$F7,$8E,$1E,$2F,$57,$A1,$D7,$19,$5F,$EB,$FA,$4E,$A3
	.dc.b	$13,$A7,$FD,$5B,$3A,$78,$4E,$6B,$B4,$60,$87,$0A,$E8,$F2,$9B,$3F
	.dc.b	$1A,$6E,$DE,$B9,$B5,$EB,$9E,$38,$7C,$19,$FC,$4C,$EC,$FF,$54,$9F
	.dc.b	$35,$CB,$F0,$DF,$99,$D6,$F1,$3E,$03,$67,$C8,$7A,$31,$FB,$31,$96
	.dc.b	$CE,$71,$F0,$57,$C4,$AD,$47,$D3,$6C,$78,$4E,$07,$D5,$72,$FE,$0D
	.dc.b	$B9,$7C,$D8,$F1,$9D,$6B,$CE,$72,$7D,$C6,$0D,$1C,$2F,$E9,$B8,$7D
	.dc.b	$2E,$6F,$D3,$18,$23,$82,$47,$9E,$7B,$4F,$42,$4F,$BE,$32,$F0,$9F
	.dc.b	$29,$E1,$74,$F8,$4F,$88,$64,$EC,$64,$E3,$3E,$47,$2F,$81,$B1,$D2
	.dc.b	$35,$1C,$23,$04,$76,$78,$23,$74,$BF,$A2,$7D,$2C,$3F,$0F,$87,$B2
	.dc.b	$F8,$9D,$4B,$5C,$63,$51,$93,$EE,$38,$8D,$47,$B5,$B7,$E2,$BF,$04
	.dc.b	$ED,$F1,$D7,$DF,$AD,$E1,$EB,$AE,$A7,$3A,$DB,$F2,$CF,$14,$F0,$7C
	.dc.b	$F1,$D9,$D4,$65,$E7,$7F,$35,$B5,$DE,$74,$7E,$56,$7D,$17,$ED,$C3
	.dc.b	$E3,$DF,$4C,$3B,$56,$A3,$BC,$EA,$30,$77,$BE,$B5,$EA,$8C,$37,$0F
	.dc.b	$15,$71,$7C,$91,$D9,$B8,$78,$F0,$F9,$4D,$7C,$8D,$BF,$0C,$F4,$9D
	.dc.b	$9E,$62,$C7,$E1,$AF,$C1,$1F,$2A,$BE,$A5,$F9,$63,$F0,$73,$A7,$E4
	.dc.b	$98,$39,$DF,$77,$C6,$77,$2D,$F7,$76,$71,$33,$1C,$63,$D3,$19,$7F
	.dc.b	$7A,$FF,$45,$AF,$E3,$8D,$BF,$F0,$FE,$1F,$92,$8F,$F4,$38,$F0,$9E
	.dc.b	$AB,$6F,$F1,$2C,$61,$B1,$99,$E2,$71,$70,$FF,$55,$F5,$7F,$C6,$2E
	.dc.b	$AD,$18,$8D,$77,$F9,$23,$11,$FD,$6D,$E3,$9F,$F6,$A6,$E9,$FE,$9E
	.dc.b	$E8,$79,$FF,$E5,$4A,$D6,$1F,$D6,$EF,$78,$7F,$63,$6A,$BE,$3C,$1B
	.dc.b	$A9,$3F,$D7,$AB,$75,$A8,$C4,$EA,$FB,$C1,$63,$E5,$C7,$C8,$FA,$EB
	.dc.b	$83,$6B,$83,$BC,$73,$1E,$79,$AF,$9E,$B9,$B9,$3C,$64,$BF,$0A,$38
	.dc.b	$57,$2C,$BE,$8D,$F4,$C7,$D9,$7E,$A2,$63,$E5,$79,$E1,$C8,$ED,$19
	.dc.b	$23,$DE,$79,$C6,$23,$FC,$3D,$97,$D0,$C3,$C3,$1F,$00,$E1,$97,$B1
	.dc.b	$DE,$3E,$7A,$FA,$33,$9A,$E3,$8F,$9A,$FE,$13,$E6,$C7,$C3,$18,$C7
	.dc.b	$C5,$5C,$1F,$9D,$1E,$A3,$44,$3D,$D5,$D5,$F1,$9B,$33,$1D,$30,$F6
	.dc.b	$1E,$6C,$BD,$19,$7C,$DC,$BD,$13,$A8,$C4,$EA,$FA,$65,$E8,$7B,$E3
	.dc.b	$E1,$CB,$D1,$1F,$1A,$B8,$4F,$96,$B8,$56,$AB,$8E,$5F,$46,$73,$8F
	.dc.b	$86,$7B,$E5,$E4,$CB,$9D,$FC,$4C,$7C,$17,$B7,$E6,$56,$EF,$E1,$63
	.dc.b	$DF,$1E,$B8,$F2,$9F,$31,$AC,$39,$63,$D4,$F9,$13,$F0,$A7,$4D,$C7
	.dc.b	$C2,$D9,$DC,$FA,$32,$EC,$FB,$8E,$F5,$F0,$F2,$F3,$3E,$89,$E5,$3F
	.dc.b	$77,$1F,$3B,$F9,$13,$ED,$99,$AE,$37,$E7,$8C,$57,$27,$C0,$70,$2D
	.dc.b	$74,$BF,$04,$79,$C8,$F9,$67,$96,$5F,$27,$0E,$F5,$EA,$C3,$EC,$46
	.dc.b	$23,$B4,$EE,$B8,$E5,$F8,$F9,$EB,$7C,$27,$39,$7F,$01,$C7,$C7,$1F
	.dc.b	$CF,$B2,$FE,$34,$DC,$BF,$DF,$6D,$3F,$CD,$31,$F9,$32,$7A,$BF,$B5
	.dc.b	$71,$34,$C4,$59,$F0,$61,$F2,$71,$F8,$D6,$AD,$47,$D1,$8E,$75,$C2
	.dc.b	$3F,$85,$4E,$B0,$FD,$E3,$6F,$FA,$03,$67,$75,$C5,$EF,$19,$7F,$9F
	.dc.b	$3E,$E7,$8C,$7B,$A3,$11,$C8,$47,$C7,$8F,$6B,$FE,$B5,$F5,$1E,$AE
	.dc.b	$2B,$F8,$D6,$E9,$CD,$78,$F2,$F8,$0F,$1B,$E1,$97,$F3,$6B,$E1,$1F
	.dc.b	$CF,$1B,$3F,$0F,$A1,$D8,$F9,$AD,$65,$DB,$3F,$E5,$55,$AA,$FE,$8A
	.dc.b	$F0,$F5,$AC,$72,$C3,$A2,$8C,$B2,$33,$95,$BE,$B2,$F5,$7E,$2C,$7F
	.dc.b	$82,$30,$F2,$C6,$23,$FA,$81,$FE,$F0,$9E,$17,$FE,$06,$AD,$B6,$33
	.dc.b	$2F,$0B,$AD,$6A,$31,$5F,$D6,$E3,$E1,$FC,$4F,$8A,$B7,$7C,$9B,$E2
	.dc.b	$E6,$F8,$4F,$F7,$99,$CC,$7F,$65,$78,$F1,$EF,$1F,$45,$DC,$7E,$8B
	.dc.b	$70,$FB,$73,$E4,$37,$87,$33,$BF,$AD,$F0,$E2,$C7,$BE,$31,$1F,$15
	.dc.b	$B7,$F7,$9E,$C7,$7F,$72,$D7,$5A,$FE,$F9,$EA,$3D,$5E,$70,$8C,$4F
	.dc.b	$F8,$D2,$31,$3A,$9C,$C7,$F8,$E1,$BA,$7D,$96,$D7,$F8,$27,$8F,$FC
	.dc.b	$85,$EB,$7C,$5B,$E7,$3E,$39,$CB,$E5,$8C,$9F,$16,$32,$FC,$73,$37
	.dc.b	$F8,$A8,$CB,$55,$7D,$1F,$9B,$7C,$47,$D5,$9D,$E1,$F8,$EF,$5D,$E6
	.dc.b	$7C,$26,$C7,$C1,$6E,$5F,$E5,$66,$BF,$E6,$07,$39,$7E,$65,$B3,$F8
	.dc.b	$5F,$86,$7C,$57,$C1,$9C,$DF,$F9,$A3,$AF,$FC,$5D,$EA,$BD,$6F,$D0
	.dc.b	$E4,$F6,$7F,$4F,$39,$BE,$01,$C3,$FA,$18,$DD,$7F,$9B,$2F,$FC,$75
	.dc.b	$EC,$6B,$0E,$9A,$62,$B5,$82,$B5,$87,$ED,$9E,$B1,$FD,$0C,$F0,$4F
	.dc.b	$9A,$7B,$C7,$93,$1E,$37,$EC,$AE,$B1,$89,$E2,$6E,$39,$4E,$E3,$DB
	.dc.b	$7B,$3C,$2F,$F9,$94,$F3,$4F,$06,$9C,$CF,$DA,$5F,$91,$FE,$17,$8F
	.dc.b	$57,$E3,$5D,$7D,$98,$79,$4E,$F3,$BC,$7E,$3C,$FA,$0E,$65,$73,$3C
	.dc.b	$23,$17,$E1,$7E,$46,$3E,$BC,$7E,$7B,$63,$D0,$79,$0E,$C7,$92,$3D
	.dc.b	$D5,$D6,$78,$4F,$89,$AF,$69,$F8,$91,$B6,$9F,$7B,$C8,$FB,$D3,$F8
	.dc.b	$A6,$BE,$C6,$BF,$7D,$CF,$90,$F6,$7F,$97,$4A,$73,$18,$6E,$5F,$E8
	.dc.b	$96,$CE,$F0,$FE,$7B,$7C,$EB,$F2,$53,$FE,$8F,$C1,$E9,$FF,$42,$BF
	.dc.b	$CE,$C6,$27,$67,$FB,$3F,$1F,$1C,$FE,$8B,$0E,$93,$F8,$43,$F5,$97
	.dc.b	$CE,$FF,$A6,$9C,$E3,$53,$C2,$B9,$5F,$BB,$2F,$0E,$1F,$D5,$1B,$7F
	.dc.b	$8B,$BE,$F1,$C3,$1F,$2E,$3D,$71,$DD,$F6,$C7,$CF,$3F,$09,$EC,$F9
	.dc.b	$31,$EC,$D9,$ED,$3A,$7F,$D6,$98,$F8,$27,$51,$F9,$57,$D9,$8F,$8E
	.dc.b	$3F,$31,$3B,$7E,$75,$F6,$69,$DB,$18,$FC,$9B,$F5,$1D,$B1,$E8,$75
	.dc.b	$C3,$DA,$48,$F6,$57,$28,$C5,$76,$3C,$78,$FA,$DF,$B5,$3F,$23,$0F
	.dc.b	$AD,$1B,$C3,$BE,$1D,$23,$DD,$3E,$97,$EE,$9D,$E3,$FC,$13,$EB,$BD
	.dc.b	$9F,$E5,$7B,$E1,$B5,$FE,$DD,$9D,$3F,$D4,$7D,$9F,$D1,$E5,$FE,$13
	.dc.b	$9C,$CF,$F8,$50,$E1,$5F,$D9,$54,$69,$D3,$5F,$EC,$5E,$53,$CE,$F0
	.dc.b	$F3,$A3,$13,$FD,$8D,$AF,$F8,$17,$BD,$F3,$9E,$4D,$39,$E1,$EA,$A7
	.dc.b	$B4,$F6,$B0,$D8,$E9,$38,$99,$1F,$DB,$E7,$4F,$F6,$FC,$7C,$3A,$FF
	.dc.b	$8C,$E7,$33,$F0,$5A,$FF,$8D,$9A,$8F,$1D,$7F,$82,$72,$FF,$28,$FB
	.dc.b	$5F,$41,$B3,$FE,$13,$6B,$EF,$F1,$3F,$AB,$67,$FE,$30,$AE,$50,$69
	.dc.b	$27,$FC,$85,$C5,$D6,$7F,$79,$97,$F8,$F2,$31,$F0,$B9,$75,$E7,$4B
	.dc.b	$F1,$5F,$0C,$7F,$C7,$F1,$CC,$86,$B5,$AF,$F9,$66,$7F,$05,$9F,$F9
	.dc.b	$BA,$3C,$E7,$C1,$F6,$1E,$D7,$F5,$2F,$4B,$FC,$A5,$18,$9F,$F3,$B3
	.dc.b	$5F,$F3,$9A,$F8,$8E,$10,$B4,$EF,$BD,$7D,$3D,$BD,$D6,$9B,$5B,$9F
	.dc.b	$F1,$A7,$B6,$F6,$FA,$F4,$23,$19,$7F,$AA,$E3,$13,$CB,$E0,$37,$5A
	.dc.b	$AF,$F7,$17,$B2,$F7,$3C,$F0,$8C,$3C,$E7,$86,$BE,$88,$FC,$44,$E9
	.dc.b	$E7,$5C,$A3,$13,$FE,$9F,$AE,$39,$7F,$A1,$A7,$87,$17,$ED,$5B,$86
	.dc.b	$70,$FF,$45,$CE,$5B,$19,$3F,$D8,$12,$F7,$3C,$07,$86,$64,$EA,$32
	.dc.b	$F0,$8A,$FC,$28,$B8,$FB,$63,$E1,$4E,$30,$F4,$55,$9F,$83,$35,$F9
	.dc.b	$11,$6F,$E0,$D7,$09,$F1,$10,$F8,$E5,$C3,$E1,$57,$BA,$3E,$6B,$5E
	.dc.b	$A7,$ED,$E3,$6F,$C2,$EB,$FE,$F8,$FD,$C6,$E3,$EB,$C3,$E3,$4F,$86
	.dc.b	$31,$3A,$C3,$EA,$36,$7C,$26,$E3,$3E,$2F,$73,$11,$A3,$C0,$E5,$F9
	.dc.b	$86,$8F,$C6,$4F,$56,$C6,$6F,$CB,$1C,$89,$7E,$9C,$7E,$65,$F1,$AF
	.dc.b	$E6,$97,$C6,$7F,$21,$39,$34,$F0,$8F,$25,$73,$BF,$05,$6E,$79,$46
	.dc.b	$B0,$F7,$BE,$A2,$57,$C8,$9D,$47,$C1,$8E,$8F,$A2,$78,$C6,$0E,$11
	.dc.b	$93,$F5,$35,$E6,$79,$9F,$10,$D1,$F8,$02,$E1,$F0,$1B,$87,$C0,$C3
	.dc.b	$8C,$E6,$FC,$18,$78,$89,$7E,$0A,$F8,$27,$0C,$3C,$51,$D9,$F8,$8F
	.dc.b	$C0,$9E,$2F,$96,$BD,$67,$67,$A1,$C6,$BA,$B4,$E4,$6A,$38,$CF,$4B
	.dc.b	$FC,$0C,$62,$75,$39,$C3,$DD,$1C,$E7,$C0,$F8,$E7,$96,$1E,$A9,$E5
	.dc.b	$39,$39,$9F,$5E,$77,$87,$C8,$BE,$B5,$C3,$0F,$95,$5F,$B9,$BF,$03
	.dc.b	$E3,$7E,$45,$F1,$69,$F5,$0B,$5C,$EF,$AD,$78,$27,$E1,$9E,$52,$5F
	.dc.b	$81,$F3,$1E,$83,$38,$7B,$67,$95,$F5,$9E,$F1,$E5,$7C,$E7,$1A,$F9
	.dc.b	$B1,$C0,$D4,$FB,$EB,$EF,$BF,$32,$7C,$55,$E0,$C3,$ED,$CE,$DB,$7C
	.dc.b	$E7,$CE,$6C,$F8,$47,$88,$F6,$E1,$F6,$9F,$E8,$18,$7F,$18,$6D,$7F
	.dc.b	$8D,$DF,$C5,$9F,$E3,$9F,$15,$EE,$38,$35,$29,$CC,$61,$B1,$CD,$AF
	.dc.b	$F4,$FF,$73,$FC,$74,$F7,$D1,$18,$8D,$74,$76,$9C,$B6,$78,$EB,$FC
	.dc.b	$B5,$BA,$FF,$8A,$ED,$D2,$F5,$5C,$1F,$D6,$2F,$75,$FD,$A9,$AF,$F7
	.dc.b	$16,$D7,$19,$E1,$8C,$AF,$EF,$35,$AB,$EB,$18,$9D,$63,$E5,$7F,$C7
	.dc.b	$CF,$F5,$DC,$3A,$4E,$C9,$1B,$4E,$EB,$AB,$92,$63,$FD,$EF,$0F,$16
	.dc.b	$3E,$08,$C6,$37,$0F,$F2,$76,$3E,$0C,$38,$C7,$C1,$9D,$3F,$23,$1E
	.dc.b	$F7,$CA,$3C,$AD,$AF,$26,$1C,$32,$FA,$B5,$A3,$BD,$F2,$7B,$19,$3E
	.dc.b	$76,$1F,$13,$0F,$B1,$8F,$95,$EC,$49,$E3,$8F,$81,$E5,$87,$82,$3D
	.dc.b	$71,$E9,$C7,$C9,$97,$DC,$AD,$5F,$6C,$BC,$97,$E2,$39,$D7,$CA,$6E
	.dc.b	$5F,$13,$2F,$24,$78,$AF,$C8,$F1,$CB,$C8,$78,$A7,$71,$99,$F0,$65
	.dc.b	$CF,$1F,$0E,$5E,$4A,$F8,$F3,$93,$FC,$1C,$F0,$3F,$1A,$68,$A7,$BC
	.dc.b	$CF,$3B,$E7,$5F,$2D,$F4,$63,$1A,$D9,$CD,$FD,$1B,$FF,$33,$47,$EC
	.dc.b	$A3,$15,$CE,$F8,$5F,$16,$CE,$AB,$9D,$FE,$63,$2F,$F4,$E6,$1C,$6B
	.dc.b	$5F,$33,$C5,$EB,$F4,$1F,$EB,$17,$91,$E7,$71,$7C,$CF,$2C,$67,$C3
	.dc.b	$F5,$34,$E2,$74,$71,$1C,$6D,$4F,$84,$CF,$07,$F9,$AE,$73,$F0,$5E
	.dc.b	$13,$C5,$E9,$DC,$BC,$FF,$03,$F2,$7D,$5F,$33,$63,$D7,$EF,$FA,$4F
	.dc.b	$57,$C0,$58,$F7,$FB,$FE,$D3,$DD,$E3,$D8,$F8,$7F,$01,$F5,$7D,$0F
	.dc.b	$84,$D8,$F8,$FF,$05,$F1,$4F,$8D,$D5,$63,$ED,$71,$FF,$20,$F8,$1F
	.dc.b	$CE,$F6,$1B,$F0,$18,$6D,$70,$9D,$36,$74,$D3,$53,$C2,$BE,$A1,$DE
	.dc.b	$3E,$DC,$FE,$3F,$C9,$7C,$FF,$DB,$8C,$36,$7F,$D7,$2D,$AF,$F5,$CB
	.dc.b	$F8,$1F,$82,$E5,$F4,$16,$BF,$D7,$ED,$AF,$B9,$D4,$AF,$C4,$7B,$19
	.dc.b	$EA,$FB,$93,$FE,$E1,$C3,$FD,$8F,$E3,$7E,$37,$D4,$F3,$36,$3F,$5F
	.dc.b	$8F,$FB,$1E,$B8,$3F,$AC,$6C,$7F,$4D,$EC,$FC,$5D,$FF,$40,$8C,$4F
	.dc.b	$0C,$3C,$33,$F9,$BF,$4F,$B6,$1E,$8E,$1F,$CF,$1F,$9F,$F1,$AE,$3F
	.dc.b	$A4,$8B,$3E,$E9,$B8,$7B,$23,$E7,$7B,$5F,$A7,$6D,$FE,$B1,$A7,$C6
	.dc.b	$2C,$FF,$51,$C7,$F0,$BA,$FE,$19,$B5,$F8,$37,$D9,$5F,$CD,$1A,$FF
	.dc.b	$4A,$6B,$FD,$B9,$B8,$FC,$1F,$0F,$85,$7E,$37,$D4,$EC,$FF,$BE,$3E
	.dc.b	$8B,$FB,$3F,$AD,$FD,$A3,$83,$FD,$EB,$F4,$58,$AF,$F7,$FD,$5C,$FF
	.dc.b	$DF,$5C,$EA,$C5,$FF,$BF,$2A,$C9,$CF,$CC,$BD,$8C,$C6,$BD,$F9,$93
	.dc.b	$26,$7E,$3C,$1E,$B6,$A7,$35,$FC,$72,$72,$F1,$AF,$56,$4C,$3C,$38
	.dc.b	$71,$9F,$E7,$3F,$49,$E5,$BE,$07,$F2,$6B,$7D,$27,$BF,$25,$6D,$F0
	.dc.b	$9F,$0A,$35,$19,$8F,$F1,$01,$C3,$3F,$E8,$AF,$B6,$34,$FF,$89,$5A
	.dc.b	$E6,$39,$9E,$1E,$1F,$EA,$38,$FA,$23,$E6,$FC,$34,$CB,$D3,$6D,$DB
	.dc.b	$3D,$AD,$4F,$A2,$73,$18,$C7,$E7,$36,$FE,$17,$1F,$DA,$3E,$CF,$D1
	.dc.b	$D8,$EF,$7D,$A3,$18,$75,$C5,$18,$C7,$E9,$FC,$67,$C2,$B7,$02,$30
	.dc.b	$DC,$7F,$CC,$13,$A9,$C9,$F8,$C3,$F1,$FE,$93,$E2,$B1,$5E,$4C,$BF
	.dc.b	$00,$F9,$B0,$F8,$4E,$79,$3F,$2B,$B7,$E5,$19,$1C,$77,$BB,$7C,$7B
	.dc.b	$7C,$27,$98,$95,$E6,$8C,$B6,$3F,$42,$79,$BE,$36,$C7,$D6,$DD,$F4
	.dc.b	$AF,$89,$FC,$2C,$61,$B1,$99,$FF,$41,$4E,$5B,$19,$C7,$F2,$73,$96
	.dc.b	$FC,$77,$C0,$73,$3B,$9F,$F0,$CF,$C5,$3D,$5F,$DE,$42,$F8,$3A,$8C
	.dc.b	$C6,$30,$FF,$3E,$57,$17,$FD,$22,$DC,$3F,$D3,$0D,$7F,$46,$D7,$FD
	.dc.b	$09,$F4,$CF,$4F,$F4,$B0,$8C,$57,$FA,$71,$B1,$FE,$A3,$6B,$FA,$E7
	.dc.b	$38,$FF,$8D,$1B,$3D,$BC,$FF,$DD,$46,$2B,$F6,$BE,$27,$F0,$FF,$8B
	.dc.b	$78,$BF,$E9,$7C,$7F,$13,$49,$CB,$C0,$C4,$EF,$93,$FD,$1F,$3A,$9F
	.dc.b	$F5,$05,$FF,$1B,$E2,$C6,$1D,$1D,$4F,$FA,$48,$FC,$6E,$1C,$F3,$E0
	.dc.b	$FF,$B0,$6F,$CC,$E6,$7C,$27,$8F,$0E,$B8,$F8,$A7,$84,$E8,$FE,$57
	.dc.b	$19,$78,$47,$39,$F1,$C7,$FA,$78,$F1,$9F,$CA,$E7,$FD,$57,$86,$AF
	.dc.b	$CB,$39,$C2,$BF,$00,$F0,$63,$F2,$B0,$E1,$18,$C3,$C4,$EE,$3D,$35
	.dc.b	$EE,$8F,$29,$D4,$F5,$1F,$1E,$36,$E0,$F4,$38,$3B,$E1,$A2,$E3,$F2
	.dc.b	$5F,$1C,$6E,$3E,$3B,$67,$EF,$4F,$19,$F9,$6D,$8C,$C7,$33,$25,$7F
	.dc.b	$0D,$F0,$3F,$C2,$1B,$3F,$52,$31,$1F,$1D,$BF,$00,$F2,$3F,$B0,$CA
	.dc.b	$73,$18,$6E,$7F,$D9,$DB,$3B,$BF,$F7,$85,$F3,$AF,$AB,$83,$C7,$FF
	.dc.b	$7A,$57,$E5,$E3,$17,$F6,$AF,$83,$F8,$20,$E2,$78,$57,$2A,$FF,$7F
	.dc.b	$1C,$AB,$85,$6E,$3C,$78,$FB,$32,$E3,$5F,$7C,$FD,$8D,$FF,$80,$5F
	.dc.b	$CC,$5F,$8A,$7F,$58,$F3,$8F,$F0,$35,$F7,$C3,$9C,$78,$30,$F9,$31
	.dc.b	$C1,$F0,$4F,$9B,$0E,$D8,$70,$35,$8F,$8F,$1F,$05,$6F,$1F,$33,$ED
	.dc.b	$AD,$C7,$9B,$0E,$AF,$93,$0E,$B5,$C6,$35,$87,$5A,$F2,$CE,$71,$F4
	.dc.b	$9D,$0E,$B8,$79,$63,$93,$87,$E5,$D7,$93,$0E,$F8,$78,$4E,$D1,$1F
	.dc.b	$B5,$3F,$1C,$F0,$1C,$9F,$91,$1C,$D8,$7A,$0F,$E1,$5F,$4F,$FE,$0C
	.dc.b	$7F,$C1,$E1,$C3,$FE,$24,$31,$D7,$FC,$AF,$5F,$B9,$2F,$85,$C5,$7F
	.dc.b	$8F,$E3,$8B,$FE,$28,$6B,$C2,$BF,$C5,$6D,$28,$F8,$B7,$AF,$88,$C3
	.dc.b	$63,$31,$C1,$AF,$46,$E3,$FE,$FC,$F8,$C7,$8B,$E6,$A4,$61,$F8,$35
	.dc.b	$C3,$3F,$F2,$87,$BF,$FE,$56,$9F,$EF,$CD,$7F,$BA,$B7,$0E,$D8,$7C
	.dc.b	$27,$FD,$C5,$E5,$FB,$1B,$5F,$E0,$0A,$E4,$FF,$55,$6B,$FE,$57,$E4
	.dc.b	$F9,$7E,$37,$F8,$B1,$BE,$0F,$F0,$99,$FF,$1B,$F8,$AF,$A8,$CD,$7C
	.dc.b	$AB,$E1,$5F,$E1,$B8,$C5,$F1,$8F,$F2,$DC,$BE,$A7,$ED,$47,$F9,$09
	.dc.b	$AF,$F9,$35,$B9,$7F,$A2,$D7,$C4,$F9,$9C,$ED,$FE,$6F,$FA,$AF,$39
	.dc.b	$9E,$5F,$5E,$AF,$A9,$A4,$CB,$D7,$F2,$B5,$F2,$CE,$DF,$F2,$4B,$5C
	.dc.b	$CF,$BF,$17,$CC,$76,$C2,$57,$AE,$2B,$FE,$67,$BF,$3B,$FE,$6D,$6E
	.dc.b	$BF,$0E,$7D,$9F,$0B,$E8,$D6,$5F,$C5,$8A,$F3,$8B,$B7,$C5,$9F,$C7
	.dc.b	$7D,$63,$D2,$A4,$BF,$D7,$FD,$67,$F4,$57,$F6,$33,$B7,$F4,$6D,$7F
	.dc.b	$4E,$D7,$FD,$26,$D7,$F9,$73,$BB,$FF,$60,$67,$FE,$53,$9F,$C3,$F9
	.dc.b	$9E,$17,$EE,$79,$BF,$53,$EB,$64,$7E,$06,$2B,$F1,$E2,$DF,$E0,$67
	.dc.b	$ED,$FD,$7C,$AF,$C9,$BF,$83,$FA,$FC,$47,$F9,$BF,$E8,$BF,$32,$62
	.dc.b	$DC,$FF,$D1,$75,$FD,$95,$3C,$AB,$18,$6F,$77,$F8,$C8,$E1,$E5,$5A
	.dc.b	$FF,$67,$57,$2B,$FF,$5B,$B6,$3B,$57,$F5,$39,$CB,$63,$27,$FB,$5E
	.dc.b	$FC,$3E,$AF,$58,$5F,$FA,$75,$D5,$70,$7D,$0F,$FB,$7F,$9B,$F5,$9F
	.dc.b	$27,$E5,$5C,$5F,$96,$AE,$27,$E4,$79,$F5,$D2,$BA,$71,$AE,$15,$D8
	.dc.b	$E0,$72,$BE,$ED,$A7,$C8,$74,$24,$67,$5A,$F3,$AF,$E0,$53,$8D,$7F
	.dc.b	$92,$47,$F0,$5A,$F5,$7D,$83,$C2,$A4,$C3,$C5,$0F,$90,$C4,$7D,$E3
	.dc.b	$E2,$47,$14,$64,$FF,$6F,$39,$71,$68,$C5,$FF,$82,$E3,$17,$F1,$7C
	.dc.b	$89,$7E,$28,$F3,$FA,$B2,$3C,$AF,$AB,$E4,$31,$1F,$07,$C4,$FF,$6E
	.dc.b	$BF,$2D,$AF,$C5,$CB,$FD,$58,$FB,$BD,$4F,$A9,$F4,$B2,$30,$43,$E6
	.dc.b	$70,$A3,$51,$CE,$7F,$19,$CB,$FC,$CF,$83,$F4,$3F,$29,$FC,$75,$BE
	.dc.b	$2E,$BC,$89,$87,$E2,$05,$6F,$A3,$FC,$5C,$D8,$FD,$1C,$FE,$66,$31
	.dc.b	$1C,$2B,$07,$47,$DD,$F1,$FF,$78,$8F,$DC,$F9,$1F,$C0,$FC,$8F,$E4
	.dc.b	$FC,$09,$3F,$80,$8F,$C2,$FC,$84,$8F,$C1,$BF,$88,$F9,$1F,$C4,$FC
	.dc.b	$8C,$9F,$C3,$BF,$8C,$F7,$3F,$1B,$F2,$3F,$9F,$F0,$25,$7E,$32,$3F
	.dc.b	$23,$EB,$C8,$FC,$7B,$F9,$4F,$8F,$FC,$AF,$9B,$27,$F2,$6F,$E6,$3E
	.dc.b	$2F,$11,$F6,$FE,$37,$F3,$66,$39,$BF,$3F,$D7,$FC,$BD,$B1,$A9,$E6
	.dc.b	$7E,$78,$FD,$19,$0F,$D0,$FB,$75,$FC,$47,$C6,$FE,$A0,$C5,$76,$8C
	.dc.b	$1F,$A8,$8C,$F4,$FD,$AF,$4F,$F8,$A4,$FE,$9B,$A7,$FB,$7B,$FA,$8F
	.dc.b	$17,$FC,$D8,$FE,$0A,$76,$D3,$47,$33,$F1,$13,$FC,$37,$AE,$F1,$7F
	.dc.b	$82,$F2,$FC,$8D,$FE,$1C,$FC,$44,$67,$1F,$C6,$B7,$0F,$24,$EA,$F9
	.dc.b	$92,$3F,$8B,$1E,$32,$E1,$E1,$6D,$73,$7F,$36,$D8,$F7,$C6,$A7,$A6
	.dc.b	$1F,$A6,$AD,$5F,$86,$B4,$49,$F5,$BC,$30,$F4,$DD,$8F,$49,$4F,$0B
	.dc.b	$70,$F0,$D7,$78,$C6,$1E,$0A,$B1,$BC,$3C,$55,$B9,$F1,$61,$F4,$5A
	.dc.b	$7D,$BB,$F7,$1C,$DC,$9E,$22,$E1,$E2,$9F,$4C,$62,$78,$57,$72,$4E
	.dc.b	$9F,$83,$3B,$C3,$CC,$FA,$0E,$B8,$7A,$2B,$81,$23,$E2,$9C,$4D,$9F
	.dc.b	$74,$FC,$8E,$1F,$6D,$B8,$7F,$93,$5F,$39,$F9,$6F,$80,$7C,$7B,$E5
	.dc.b	$7F,$CF,$F1,$86,$C6,$67,$FD,$0B,$39,$7F,$D0,$F6,$F9,$56,$DF,$E8
	.dc.b	$88,$C3,$63,$33,$C1,$AF,$A1,$CE,$BF,$CC,$1B,$9F,$FB,$5F,$8D,$D5
	.dc.b	$FD,$66,$B5,$87,$EA,$AF,$78,$7E,$D5,$AF,$F8,$FE,$B8,$61,$FC,$D2
	.dc.b	$73,$8F,$9A,$BF,$47,$1C,$31,$EF,$C1,$D7,$19,$5A,$7F,$A0,$4E,$A3
	.dc.b	$04,$CB,$CA,$FE,$C2,$35,$18,$26,$3E,$3A,$E7,$97,$9A,$FD,$51,$E0
	.dc.b	$AD,$C7,$F7,$6B,$E7,$7C,$63,$C0,$D9,$FF,$1C,$CE,$A7,$99,$A2,$C6
	.dc.b	$AF,$C1,$1F,$D6,$B1,$F8,$A4,$C3,$CB,$8F,$8F,$1E,$C4,$AF,$25,$F2
	.dc.b	$CB,$E6,$63,$E8,$9E,$98,$72,$7C,$46,$0B,$97,$9D,$C4,$F1,$AE,$78
	.dc.b	$7E,$0F,$0E,$D9,$79,$6F,$8C,$63,$1F,$34,$7B,$AB,$86,$58,$8E,$B1
	.dc.b	$E9,$7F,$0D,$19,$CB,$C5,$8F,$82,$7D,$B3,$9C,$3B,$47,$DD,$6E,$5F
	.dc.b	$2E,$7D,$CF,$D2,$E8,$F9,$B8,$7D,$3F,$27,$FC,$A3,$E5,$FD,$F7,$FC
	.dc.b	$A7,$AD,$F1,$7F,$C8,$F8,$FF,$92,$DF,$D7,$36,$B5,$3B,$6C,$64,$99
	.dc.b	$FF,$88,$3E,$57,$51,$FE,$55,$F9,$2F,$F2,$23,$D4,$F1,$38,$9F,$F1
	.dc.b	$3C,$63,$CD,$F2,$34,$E0,$73,$71,$1F,$E5,$99,$FF,$2A,$4E,$4F,$07
	.dc.b	$95,$D8,$EF,$E5,$D8,$F2,$7C,$AF,$C1,$F6,$BC,$6D,$8F,$47,$8F,$E6
	.dc.b	$3C,$FE,$5D,$8F,$67,$8F,$EA,$3D,$7E,$4D,$8F,$81,$E4,$7C,$DF,$82
	.dc.b	$F7,$B6,$3E,$27,$37,$C2,$3E,$1F,$45,$8F,$DF,$F9,$3F,$68,$FB,$7E
	.dc.b	$4D,$8F,$E0,$6D,$E2,$FE,$00,$DE,$AF,$D9,$B7,$B3,$F8,$E3,$65,$D4
	.dc.b	$E5,$4F,$E1,$7C,$E7,$CA,$FD,$68,$C3,$72,$FF,$33,$37,$0F,$F3,$53
	.dc.b	$F7,$3C,$A9,$3F,$81,$F2,$ED,$7D,$79,$BA,$7F,$A3,$B2,$56,$34,$FF
	.dc.b	$55,$43,$CB,$F2,$78,$9F,$88,$6E,$5F,$8D,$70,$FE,$3F,$CA,$B9,$FE
	.dc.b	$16,$6E,$5D,$E6,$E3,$DB,$C2,$FC,$99,$F9,$4F,$12,$E3,$F9,$98,$B8
	.dc.b	$FE,$3E,$6E,$1E,$98,$F9,$1F,$0F,$F9,$C6,$DF,$E8,$5A,$7E,$5C,$B3
	.dc.b	$FC,$A3,$2F,$DB,$EB,$FB,$86,$D7,$F4,$07,$0F,$F4,$2F,$2E,$DF,$F3
	.dc.b	$F9,$B7,$FC,$CA,$6E,$7F,$ED,$3E,$27,$C8,$7F,$AC,$FD,$8F,$F5,$A7
	.dc.b	$2A,$73,$9F,$FA,$D2,$73,$EA,$3E,$4F,$84,$F3,$C2,$EB,$F8,$AC,$38
	.dc.b	$AC,$3A,$C6,$1E,$77,$C3,$1F,$05,$DC,$5F,$09,$9C,$1E,$D6,$63,$0D
	.dc.b	$CF,$CC,$7B,$5C,$67,$E1,$8C,$1A,$7F,$D9,$FF,$25,$E3,$8F,$89,$F2
	.dc.b	$7E,$A9,$4E,$5F,$64,$66,$31,$1A,$CB,$87,$2F,$F9,$BB,$6F,$F6,$D5
	.dc.b	$68,$93,$AA,$E4,$FB,$FE,$47,$DD,$F2,$37,$2F,$36,$FF,$4E,$31,$7E
	.dc.b	$B3,$E7,$38,$C7,$E6,$BF,$27,$E4,$FE,$47,$15,$C3,$9F,$02,$31,$87
	.dc.b	$3C,$7F,$01,$18,$C7,$EB,$BF,$47,$DF,$98,$79,$3D,$17,$C9,$E3,$1E
	.dc.b	$59,$2F,$18,$FE,$32,$F7,$87,$8A,$7E,$F7,$14,$8F,$7F,$23,$6C,$65
	.dc.b	$E5,$8F,$C0,$CE,$AB,$F0,$D5,$AC,$7F,$1B,$F1,$FF,$87,$3F,$19,$93
	.dc.b	$E1,$31,$87,$C2,$9B,$97,$C1,$AD,$4E,$4F,$C3,$5B,$E4,$DF,$0F,$F9
	.dc.b	$E8,$C3,$7E,$4D,$C9,$9A,$DD,$F2,$6E,$3F,$67,$0E,$33,$DA,$F8,$BB
	.dc.b	$C3,$F2,$5F,$26,$F4,$D3,$FE,$BE,$C0,$27,$51,$98,$C4,$F0,$78,$4E
	.dc.b	$9B,$1D,$8F,$67,$CD,$F9,$B1,$E1,$3C,$AF,$F5,$0B,$E5,$1E,$97,$E9
	.dc.b	$DB,$3A,$5E,$76,$66,$4E,$5E,$11,$8A,$FC,$E1,$EC,$1E,$17,$BB,$E5
	.dc.b	$3F,$A1,$C7,$F1,$3E,$5F,$64,$62,$38,$4E,$FC,$5E,$59,$FE,$22,$55
	.dc.b	$B7,$E4,$D6,$DE,$99,$57,$99,$C2,$F9,$E1,$E0,$7E,$95,$6F,$0E,$D8
	.dc.b	$75,$BF,$49,$C4,$EA,$53,$AB,$E9,$8E,$35,$D2,$75,$18,$7F,$DF,$B5
	.dc.b	$DE,$FC,$51,$83,$6E,$1E,$B5,$E8,$C3,$D9,$1F,$00,$DB,$87,$CF,$87
	.dc.b	$8E,$B9,$46,$E3,$13,$BB,$F1,$36,$79,$9E,$D3,$85,$F8,$5C,$1D,$9C
	.dc.b	$57,$C7,$AE,$73,$A6,$C6,$63,$AB,$4F,$1B,$B3,$33,$F2,$63,$27,$1F
	.dc.b	$81,$72,$FC,$D6,$D7,$C9,$6B,$F8,$06,$FB,$CF,$23,$FA,$D4,$A7,$31
	.dc.b	$86,$E9,$FB,$46,$CE,$EF,$F3,$97,$CE,$BF,$1D,$8B,$C7,$FD,$06,$5F
	.dc.b	$9B,$8C,$61,$F9,$3B,$E0,$FE,$25,$B3,$BA,$E3,$5F,$A9,$38,$D6,$F2
	.dc.b	$F4,$D7,$E9,$EF,$8D,$6E,$F8,$DF,$F0,$C8,$E3,$5B,$BE,$31,$EA,$6C
	.dc.b	$EE,$B8,$BF,$E3,$86,$DF,$68,$FE,$2E,$D7,$47,$C2,$8D,$CF,$48,$4E
	.dc.b	$EB,$C6,$79,$E3,$E2,$4E,$E7,$55,$E2,$AD,$D7,$82,$B4,$75,$8C,$C7
	.dc.b	$88,$F8,$D8,$74,$3C,$8C,$32,$4C,$3E,$83,$97,$E3,$47,$18,$FD,$F4
	.dc.b	$F6,$3B,$E1,$F0,$4F,$3C,$71,$8F,$99,$3F,$C4,$7E,$51,$D9,$FC,$6B
	.dc.b	$6F,$E4,$91,$8B,$FE,$66,$D7,$F9,$4A,$F8,$FE,$11,$FD,$1A,$35,$7F
	.dc.b	$CF,$1B,$5D,$2F,$74,$9E,$4F,$F4,$0F,$93,$7A,$DA,$6C,$66,$38,$FC
	.dc.b	$A7,$EA,$B1,$FE,$93,$5A,$AF,$EB,$0D,$7F,$9F,$07,$0F,$F4,$16,$BF
	.dc.b	$E4,$9F,$8B,$FF,$21,$BF,$DD,$1A,$FF,$6B,$6C,$FF,$58,$AF,$F0,$F4
	.dc.b	$EF,$CC,$F8,$72,$AD,$E7,$FE,$5D,$6D,$7F,$6C,$BE,$2F,$F8,$73,$8F
	.dc.b	$E5,$F8,$FF,$E6,$56,$F8,$BF,$C0,$E7,$FC,$49,$F2,$98,$99,$1F,$E3
	.dc.b	$07,$8F,$86,$9E,$7F,$1B,$F3,$AB,$FC,$64,$DD,$3F,$CE,$6D,$D3,$FB
	.dc.b	$FE,$48,$7C,$47,$2E,$39,$87,$F9,$22,$78,$3F,$EB,$C7,$6F,$F8,$E7
	.dc.b	$E2,$FC,$64,$7F,$C7,$9F,$27,$5F,$F1,$24,$5C,$BF,$C7,$53,$E8,$F8
	.dc.b	$FF,$5E,$12,$FC,$11,$97,$FC,$8D,$19,$7D,$B1,$F0,$3E,$0F,$E1,$E5
	.dc.b	$27,$D1,$19,$7F,$C9,$D1,$97,$E2,$C7,$E5,$BE,$4D,$E9,$D2,$4F,$F8
	.dc.b	$5B,$E3,$FF,$9C,$3F,$E1,$89,$DB,$67,$6F,$F8,$42,$76,$FF,$84,$A7
	.dc.b	$6F,$F9,$17,$C9,$FB,$27,$0D,$3E,$AF,$9B,$F7,$3E,$4E,$57,$D1,$8F
	.dc.b	$BA,$61,$FF,$27,$46,$1F,$F3,$57,$C9,$FE,$10,$8F,$D9,$F9,$1A,$FF
	.dc.b	$95,$62,$DF,$E1,$27,$F6,$F9,$7E,$5B,$C6,$FC,$BB,$7D,$4F,$F1,$53
	.dc.b	$70,$FF,$23,$BF,$EA,$A7,$FC,$43,$1A,$3F,$C5,$25,$72,$70,$72,$F2
	.dc.b	$F2,$2A,$9E,$CD,$F4,$F3,$1F,$E8,$38,$E6,$6E,$B9,$F3,$FE,$6F,$E4
	.dc.b	$2F,$5B,$96,$9A,$E9,$4D,$33,$87,$9A,$D7,$C6,$31,$5C,$1B,$F1,$BC
	.dc.b	$C8,$EB,$AE,$BF,$E5,$0C,$BF,$5F,$4A,$CE,$DF,$C3,$31,$FD,$96,$1E
	.dc.b	$AF,$91,$FE,$19,$8F,$8D,$AF,$31,$28,$CD,$73,$5F,$1B,$CF,$C2,$97
	.dc.b	$CD,$ED,$18,$8E,$97,$F0,$3C,$7F,$92,$67,$0E,$9E,$97,$94,$95,$E5
	.dc.b	$F8,$DA,$F2,$CB,$FD,$1C,$FA,$67,$D9,$3E,$DF,$47,$EA,$FC,$94,$AF
	.dc.b	$74,$65,$FA,$5C,$29,$D4,$75,$9F,$C1,$F8,$BF,$88,$F9,$17,$8B,$4C
	.dc.b	$D7,$E1,$B9,$FF,$96,$3F,$24,$4F,$07,$F1,$6D,$75,$1C,$35,$FE,$73
	.dc.b	$39,$6D,$77,$AE,$D7,$F6,$7E,$32,$47,$D9,$7E,$E7,$C7,$7D,$DF,$8E
	.dc.b	$FC,$AF,$6C,$AF,$BD,$1F,$81,$F8,$D9,$1F,$81,$7F,$09,$F1,$DF,$85
	.dc.b	$F8,$E9,$3F,$85,$7F,$11,$EC,$FE,$27,$E3,$BF,$43,$DF,$2B,$F1,$51
	.dc.b	$F8,$DF,$56,$47,$E3,$5F,$C8,$7C,$6F,$E4,$7C,$B9,$3F,$91,$7F,$29
	.dc.b	$F1,$38,$8F,$BD,$F1,$7F,$97,$31,$C9,$F9,$1A,$C4,$FE,$66,$73,$E2
	.dc.b	$7F,$50,$ED,$FF,$48,$B6,$7F,$40,$7E,$87,$DC,$AF,$E8,$7E,$2F,$F4
	.dc.b	$86,$1F,$C7,$18,$34,$7E,$9C,$FC,$27,$A3,$1F,$BA,$FE,$1F,$D2,$C4
	.dc.b	$7E,$0F,$AF,$F8,$1F,$57,$E0,$B8,$BF,$B9,$6D,$F8,$0C,$36,$FE,$0A
	.dc.b	$33,$9F,$FB,$36,$F8,$E1,$CE,$FF,$29,$5A,$8F,$F7,$6C,$EA,$F9,$61
	.dc.b	$E4,$9F,$03,$1F,$25,$71,$7F,$2F,$7D,$27,$55,$CA,$7C,$06,$30,$F2
	.dc.b	$5F,$1C,$3C,$11,$E7,$C3,$95,$76,$C3,$83,$E6,$8C,$E1,$D4,$95,$C6
	.dc.b	$76,$7E,$AB,$0E,$C7,$7A,$E3,$3E,$2B,$DE,$1F,$E0,$8B,$F0,$13,$0E
	.dc.b	$AF,$84,$DE,$1D,$2B,$E3,$46,$EF,$A3,$4E,$05,$3C,$91,$E0,$C3,$C7
	.dc.b	$1F,$4A,$BA,$3F,$20,$E1,$87,$C6,$75,$5E,$0C,$3C,$B1,$D0,$DD,$F4
	.dc.b	$71,$1A,$AF,$73,$6F,$BB,$F8,$B9,$FB,$C7,$F9,$62,$7F,$12,$D9,$FE
	.dc.b	$2E,$FB,$8F,$F2,$EF,$B6,$F8,$C7,$C9,$7F,$98,$27,$FD,$47,$18,$71
	.dc.b	$3B,$9D,$47,$FA,$8A,$74,$F8,$7E,$49,$B6,$B1,$46,$27,$B4,$62,$75
	.dc.b	$39,$9E,$51,$FC,$AA,$38,$65,$FE,$60,$E2,$75,$3C,$AD,$5C,$C3,$F6
	.dc.b	$58,$70,$E0,$44,$C3,$8D,$EF,$1F,$E6,$77,$BA,$FD,$6B,$5F,$F1,$DE
	.dc.b	$3D,$DE,$B8,$F1,$C7,$F7,$15,$BC,$7A,$D6,$89,$87,$6B,$FF,$07,$CE
	.dc.b	$A3,$13,$AB,$E7,$5A,$C3,$C1,$5A,$24,$EA,$BF,$A9,$5F,$38,$C4,$EA
	.dc.b	$7D,$77,$CF,$1F,$44,$60,$F0,$1F,$17,$0E,$F7,$C8,$91,$BB,$FA,$53
	.dc.b	$A8,$C4,$E9,$FE,$C7,$8F,$C3,$9D,$46,$30,$F0,$B4,$F4,$C6,$B2,$FC
	.dc.b	$24,$79,$0D,$61,$E0,$C7,$8C,$EA,$3D,$D3,$F5,$A7,$38,$7D,$9A,$D6
	.dc.b	$3F,$82,$AD,$1D,$63,$2F,$A8,$F8,$55,$E6,$CB,$E6,$46,$4E,$38,$78
	.dc.b	$23,$31,$E6,$BE,$D1,$8B,$EB,$3C,$27,$E3,$4E,$70,$F8,$D8,$7D,$5A
	.dc.b	$FB,$7D,$3F,$22,$BC,$3E,$47,$F9,$2B,$CB,$FA,$0F,$F9,$33,$0B,$E3
	.dc.b	$7F,$90,$F1,$FF,$0F,$C6,$2B,$2D,$8C,$9F,$E2,$48,$C6,$3F,$E3,$8A
	.dc.b	$7C,$96,$35,$FF,$26,$7C,$87,$F9,$0B,$0F,$F1,$59,$E2,$71,$7C,$63
	.dc.b	$1E,$67,$91,$A7,$03,$FC,$C9,$3C,$2B,$AC,$F3,$F3,$77,$7F,$E4,$AA
	.dc.b	$79,$7D,$CF,$07,$99,$71,$E7,$F2,$5F,$0B,$D9,$F2,$36,$3B,$F9,$3E
	.dc.b	$73,$D1,$E6,$D8,$F1,$F9,$3E,$B3,$D9,$E5,$D8,$F3,$F9,$5F,$3B,$E0
	.dc.b	$3E,$03,$63,$D7,$D1,$F0,$CF,$89,$D7,$63,$F7,$FE,$57,$DB,$F2,$FE
	.dc.b	$E3,$63,$27,$F0,$48,$CB,$4C,$F1,$7F,$11,$3F,$8D,$4F,$F1,$38,$FC
	.dc.b	$E4,$7F,$B4,$23,$0F,$E4,$9A,$FF,$9F,$B8,$91,$F6,$4F,$B5,$56,$BE
	.dc.b	$AC,$5A,$FA,$53,$79,$7F,$07,$E5,$FE,$07,$CB,$FC,$13,$70,$FB,$AD
	.dc.b	$BF,$B4,$DA,$FC,$5F,$95,$F8,$9F,$2F,$F1,$4D,$BF,$C2,$B6,$FF,$04
	.dc.b	$DA,$E3,$3F,$8A,$AE,$2D,$3F,$1B,$E1,$7E,$57,$C0,$FC,$B3,$63,$27
	.dc.b	$F4,$43,$F8,$27,$80,$DB,$FA,$CC,$67,$2F,$E9,$33,$9A,$F9,$6D,$9D
	.dc.b	$4F,$FA,$AB,$D0,$7C,$CF,$E9,$63,$F6,$F1,$89,$FD,$0C,$62,$75,$1F
	.dc.b	$EC,$79,$D4,$FE,$84,$FE,$05,$E6,$7F,$03,$CE,$D7,$F4,$A6,$CF,$F4
	.dc.b	$26,$E9,$FD,$9D,$CC,$7E,$1B,$45,$FE,$2E,$3F,$8B,$B5,$FE,$CC,$DC
	.dc.b	$98,$BE,$4B,$FB,$17,$AF,$FD,$8E,$73,$18,$CF,$E3,$D3,$E6,$71,$8F
	.dc.b	$C5,$AB,$7C,$2F,$FB,$AE,$1C,$F0,$ED,$18,$39,$E1,$C2,$33,$82,$B8
	.dc.b	$7C,$BE,$7E,$0F,$4D,$C3,$95,$7D,$58,$F6,$59,$A4,$9F,$F0,$EF,$D8
	.dc.b	$7D,$2C,$3F,$61,$AF,$E0,$DF,$D9,$46,$67,$F9,$FE,$1E,$E1,$19,$7F
	.dc.b	$C2,$06,$5C,$2A,$E9,$3F,$D5,$67,$F0,$33,$FB,$C6,$B9,$E0,$FA,$CF
	.dc.b	$67,$D7,$F2,$FF,$17,$E4,$3E,$33,$6F,$E0,$ED,$F6,$A3,$71,$FE,$3F
	.dc.b	$8C,$37,$44,$FC,$6F,$3B,$E8,$7C,$9F,$D1,$6D,$74,$AF,$F2,$53,$C6
	.dc.b	$38,$34,$E9,$F2,$9F,$80,$F3,$1F,$29,$F6,$E3,$0D,$9D,$93,$5F,$E3
	.dc.b	$B9,$FE,$4A,$BE,$CF,$CA,$7E,$0F,$E0,$7F,$08,$DB,$F2,$36,$FF,$0D
	.dc.b	$1F,$94,$8C,$71,$7E,$3F,$E0,$7C,$9F,$09,$F8,$D6,$D7,$B5,$FC,$39
	.dc.b	$1F,$8C,$FD,$2F,$92,$79,$DF,$9E,$B7,$C9,$B9,$3F,$09,$5A,$C3,$F2
	.dc.b	$75,$A9,$CD,$F3,$24,$FF,$9B,$AB,$86,$7F,$CF,$AD,$F2,$6F,$88,$E3
	.dc.b	$7D,$23,$0F,$E8,$29,$7C,$AF,$F7,$35,$A9,$F8,$A7,$F9,$9A,$75,$5F
	.dc.b	$B4,$9D,$57,$F5,$19,$CD,$7F,$9B,$3E,$4D,$E2,$7E,$85,$79,$56,$30
	.dc.b	$DC,$7F,$D1,$33,$C1,$B3,$F9,$C8,$C5,$7F,$A3,$87,$87,$FE,$7A,$F9
	.dc.b	$17,$8F,$FE,$3B,$CD,$C9,$FE,$75,$9C,$DF,$F6,$4B,$E1,$7B,$AD,$3F
	.dc.b	$E9,$F8,$CC,$FF,$12,$9C,$BC,$45,$74,$7F,$88,$46,$9F,$F2,$64,$FF
	.dc.b	$A3,$BE,$4F,$58,$79,$33,$C6,$3D,$ED,$5A,$C3,$FD,$18,$F5,$8F,$F4
	.dc.b	$BE,$1E,$47,$BD,$D7,$27,$92,$FD,$91,$98,$F8,$E6,$4F,$63,$99,$E9
	.dc.b	$87,$96,$74,$7A,$8A,$E7,$1F,$24,$74,$7D,$15,$CE,$37,$8F,$C3,$7C
	.dc.b	$B3,$A7,$A9,$8C,$3C,$EE,$0E,$8E,$23,$BE,$1E,$6A,$E9,$5A,$3D,$67
	.dc.b	$18,$D5,$F8,$EB,$C0,$72,$9F,$2C,$66,$FB,$1F,$4E,$76,$D3,$33,$F8
	.dc.b	$1F,$81,$FA,$ED,$CB,$ED,$46,$23,$E4,$37,$DF,$7C,$9F,$F5,$59,$4E
	.dc.b	$63,$0D,$D3,$FD,$AA,$D9,$DD,$FF,$A5,$EF,$9D,$7E,$0F,$0F,$EF,$35
	.dc.b	$7C,$8F,$F4,$E6,$5F,$95,$BD,$CF,$DC,$6C,$7E,$66,$B8,$CF,$E0,$1B
	.dc.b	$1C,$72,$F4,$37,$0E,$D5,$0E,$93,$AA,$F3,$36,$3F,$DD,$F1,$D0,$FF
	.dc.b	$58,$15,$FF,$79,$DF,$92,$38,$D7,$5B,$F3,$61,$EC,$9C,$DF,$82,$B5
	.dc.b	$1D,$27,$53,$D3,$0F,$12,$C3,$B3,$E4,$9D,$BF,$C2,$9B,$1A,$8E,$13
	.dc.b	$CA,$BD,$11,$B3,$F6,$93,$B3,$84,$68,$DB,$99,$FC,$16,$3E,$CA,$FE
	.dc.b	$06,$78,$CA,$78,$30,$B5,$D3,$0F,$3E,$1D,$30,$F6,$CE,$6F,$E1,$9F
	.dc.b	$06,$FF,$84,$7C,$2B,$AB,$F9,$53,$67,$FD,$0E,$03,$97,$2D,$7F,$D1
	.dc.b	$AE,$67,$FD,$25,$F2,$6F,$85,$FE,$65,$BF,$FB,$B6,$31,$7F,$D4,$6F
	.dc.b	$84,$FF,$BC,$7E,$4B,$F5,$18,$7F,$A6,$FE,$4D,$EB,$72,$C3,$C9,$18
	.dc.b	$7F,$AD,$57,$0B,$E5,$C1,$B9,$56,$DF,$EB,$66,$2F,$57,$C6,$3A,$C7
	.dc.b	$F4,$D9,$E1,$1F,$EA,$A6,$BF,$14,$E4,$7C,$82,$9D,$A5,$F0,$7D,$8F
	.dc.b	$24,$F8,$F1,$FF,$2E,$71,$F1,$F0,$BF,$CD,$6D,$F1,$BF,$82,$E3,$FE
	.dc.b	$31,$8C,$1F,$0E,$5E,$2B,$E4,$B0,$DA,$FF,$19,$56,$FE,$35,$F1,$37
	.dc.b	$8E,$57,$F9,$0A,$72,$D9,$CB,$FE,$02,$C5,$3F,$6A,$30,$F0,$6F,$C8
	.dc.b	$7A,$E0,$D2,$47,$F9,$36,$32,$FF,$94,$B1,$7C,$97,$A8,$F4,$57,$A7
	.dc.b	$4E,$84,$AE,$4D,$7C,$7F,$21,$EF,$38,$3E,$DE,$2F,$21,$1F,$13,$5F
	.dc.b	$5F,$C1,$FC,$73,$E4,$47,$E2,$FE,$3D,$E7,$7F,$84,$5B,$5F,$81,$6B
	.dc.b	$FA,$A6,$BF,$EF,$27,$3C,$5F,$EF,$77,$FA,$1C,$7F,$8C,$DA,$FE,$59
	.dc.b	$DF,$3F,$E3,$DB,$1F,$B5,$77,$3C,$FB,$3C,$D7,$F7,$F9,$3E,$91,$2F
	.dc.b	$E8,$35,$FB,$3E,$67,$E1,$4E,$0F,$E0,$FA,$BF,$10,$47,$F0,$DC,$15
	.dc.b	$FC,$0C,$5C,$7F,$D8,$1C,$5F,$6A,$7D,$9E,$17,$C7,$23,$F1,$DA,$FF
	.dc.b	$8E,$BA,$7D,$C7,$F8,$A0,$AF,$F8,$81,$F8,$51,$98,$F8,$4D,$5A,$FE
	.dc.b	$C1,$E0,$7C,$79,$D4,$7F,$91,$23,$3C,$BF,$9C,$F8,$AE,$8F,$E7,$BE
	.dc.b	$36,$E2,$6C,$CA,$E3,$39,$6C,$E5,$F6,$0A,$FD,$4F,$0F,$F0,$93,$F8
	.dc.b	$53,$88,$FE,$61,$18,$A3,$B8,$FE,$69,$3B,$72,$62,$6B,$E0,$8F,$DB
	.dc.b	$FC,$5F,$A8,$F5,$D7,$BB,$E3,$9F,$17,$7E,$2E,$61,$F1,$27,$2D,$9C
	.dc.b	$BD,$61,$3C,$E6,$4F,$18,$DF,$91,$F4,$F1,$5F,$4A,$E5,$F1,$7E,$3F
	.dc.b	$42,$47,$86,$3C,$C6,$1F,$04,$61,$F3,$7C,$4F,$AC,$8F,$A7,$E2,$EB
	.dc.b	$E7,$8B,$8F,$8F,$05,$7E,$13,$C3,$F9,$D7,$FA,$D6,$BF,$8C,$6F,$17
	.dc.b	$F3,$48,$F9,$E6,$4F,$C7,$8F,$8D,$7C,$7F,$E5,$5B,$1A,$9E,$58,$FE
	.dc.b	$5E,$39,$1F,$88,$39,$57,$13,$CC,$FD,$5F,$8B,$FB,$44,$7E,$C7,$C6
	.dc.b	$FD,$DF,$8D,$FC,$DF,$8D,$FD,$EA,$7F,$3E,$D7,$ED,$47,$E0,$FE,$32
	.dc.b	$47,$E0,$5F,$C3,$7C,$6F,$E1,$FE,$32,$4F,$E1,$5F,$C5,$7A,$FF,$8B
	.dc.b	$F8,$DE,$FD,$92,$BF,$15,$1F,$8F,$F5,$24,$7E,$35,$FC,$97,$C6,$FE
	.dc.b	$4F,$E1,$A4,$FE,$45,$FC,$B7,$C5,$E2,$3F,$21,$ED,$FE,$64,$C7,$1F
	.dc.b	$DB,$AC,$4F,$E6,$67,$3E,$1F,$EA,$3A,$7F,$D8,$0D,$9F,$D1,$1F,$A3
	.dc.b	$F6,$6B,$FA,$1F,$8C,$FD,$31,$87,$F4,$06,$0D,$1F,$A7,$3E,$F7,$A7
	.dc.b	$1F,$C1,$F5,$7F,$65,$CF,$EF,$3F,$DA,$5B,$3F,$85,$E0,$FE,$FF,$1A
	.dc.b	$6E,$3F,$E0,$FA,$E4,$DA,$E1,$D7,$DA,$B7,$87,$F8,$E6,$F6,$DA,$DB
	.dc.b	$F8,$D8,$FE,$39,$5E,$69,$CE,$1E,$18,$E8,$66,$A9,$C6,$FB,$BF,$9F
	.dc.b	$6D,$F8,$23,$13,$C2,$BA,$47,$86,$FB,$E1,$ED,$C3,$89,$CA,$33,$3E
	.dc.b	$B8,$CD,$F8,$E3,$C4,$F5,$AE,$4E,$6F,$B9,$23,$33,$E1,$38,$1D,$66
	.dc.b	$BF,$B9,$BF,$0B,$EA,$3B,$9C,$4B,$5D,$5E,$46,$5C,$3C,$5B,$87,$C5
	.dc.b	$3E,$8C,$72,$BF,$2C,$FE,$12,$3C,$73,$A8,$E3,$5E,$48,$C5,$70,$3F
	.dc.b	$09,$5C,$30,$F8,$EF,$8E,$34,$F9,$AF,$9B,$F7,$0F,$F2,$E1,$EF,$2D
	.dc.b	$79,$9B,$5F,$C8,$5F,$F6,$97,$9A,$F8,$FF,$F3,$17,$B7,$FE,$26,$6C
	.dc.b	$66,$7F,$D1,$B3,$97,$FD,$1F,$6F,$25,$AF,$F9,$AE,$30,$D8,$CC,$F0
	.dc.b	$6B,$E6,$73,$97,$EB,$9B,$9F,$F9,$DF,$85,$D3,$FD,$A2,$B5,$87,$ED
	.dc.b	$AF,$78,$7E,$A5,$AA,$F8,$F8,$FB,$A9,$3F,$AB,$AD,$D6,$A3,$13,$AB
	.dc.b	$EF,$1F,$CF,$67,$51,$82,$63,$FB,$FB,$E7,$8F,$9E,$31,$87,$36,$E1
	.dc.b	$CE,$74,$FA,$67,$4D,$35,$8E,$1F,$EA,$D7,$EE,$CB,$E7,$C6,$8F,$49
	.dc.b	$31,$F2,$BF,$EC,$5C,$BD,$39,$78,$A7,$33,$D3,$1F,$EC,$75,$A8,$F1
	.dc.b	$4E,$AF,$BB,$FD,$86,$75,$3C,$4F,$05,$F5,$3E,$D6,$5E,$E8,$D5,$73
	.dc.b	$C7,$E3,$E5,$ED,$C7,$AE,$1C,$6F,$D1,$87,$19,$CC,$79,$27,$4F,$DF
	.dc.b	$9D,$5F,$C5,$26,$5E,$C8,$CB,$70,$F9,$99,$7A,$B1,$EB,$87,$39,$CD
	.dc.b	$F0,$8F,$9D,$1F,$B0,$AF,$C4,$36,$7F,$C7,$AD,$3F,$C8,$B3,$F0,$9F
	.dc.b	$03,$F8,$BE,$8F,$89,$5F,$3F,$C5,$FF,$62,$F9,$1F,$84,$7F,$D8,$F9
	.dc.b	$DE,$E4,$D8,$C3,$E0,$C3,$FC,$9F,$3A,$C7,$C3,$1F,$E5,$B8,$C3,$FE
	.dc.b	$50,$BE,$F3,$A6,$C7,$F9,$4E,$77,$B7,$F9,$0D,$7C,$7F,$82,$FF,$CE
	.dc.b	$7F,$1F,$FE,$54,$7F,$C2,$E7,$A5,$C5,$F3,$8C,$79,$5E,$78,$FF,$37
	.dc.b	$4E,$A7,$FC,$DA,$D8,$FF,$34,$B5,$DC,$F2,$9F,$0F,$95,$67,$FC,$D5
	.dc.b	$4F,$2B,$C4,$78,$FC,$AB,$7E,$8F,$27,$C5,$AA,$FF,$C9,$B3,$F0,$DB
	.dc.b	$87,$C7,$6C,$79,$BC,$6F,$71,$EF,$F2,$AC,$7C,$4F,$1F,$E1,$1F,$0F
	.dc.b	$C8,$B1,$F2,$3C,$8F,$0E,$06,$92,$3C,$5C,$DF,$2C,$F9,$9C,$F6,$3F
	.dc.b	$6F,$E3,$FE,$2B,$C8,$FC,$5B,$63,$27,$E0,$B5,$FE,$49,$1F,$45,$B1
	.dc.b	$9E,$94,$E5,$B1,$CA,$7F,$91,$78,$FD,$A3,$F0,$35,$F7,$CF,$C0,$5D
	.dc.b	$8C,$92,$BE,$AC,$65,$FA,$FE,$47,$E1,$EF,$F1,$0D,$AF,$C6,$B6,$77
	.dc.b	$1F,$90,$9F,$C7,$78,$FF,$8F,$F2,$2C,$FE,$51,$B5,$F8,$46,$F5,$FF
	.dc.b	$A7,$77,$46,$3C,$0F,$C7,$B6,$BF,$2A,$FE,$6F,$C0,$FC,$E6,$D7,$1F
	.dc.b	$D0,$B6,$BF,$23,$3F,$A1,$F1,$F1,$97,$FA,$AF,$D6,$7C,$57,$E9,$6F
	.dc.b	$F7,$D1,$8C,$7F,$57,$18,$9D,$47,$FB,$82,$75,$3F,$9E,$3F,$84,$DF
	.dc.b	$F0,$AC,$AD,$7F,$3A,$6C,$FF,$81,$8F,$EC,$93,$FE,$A2,$7F,$A2,$35
	.dc.b	$CC,$ED,$A3,$CF,$FF,$6B,$E6,$AF,$F6,$DC,$63,$3B,$7F,$DE,$66,$DF
	.dc.b	$FB,$AF,$17,$C7,$E3,$AD,$18,$B5,$77,$7F,$DD,$30,$F8,$FE,$0E,$E7
	.dc.b	$47,$5B,$C9,$A9,$B5,$D3,$2F,$F7,$8E,$1E,$0F,$8E,$E7,$F1,$FD,$1B
	.dc.b	$5B,$3C,$2E,$5C,$63,$EB,$F9,$9C,$3E,$21,$F2,$1F,$5B,$1F,$8F,$1F
	.dc.b	$08,$CF,$0D,$E0,$FE,$8B,$2C,$3E,$0D,$69,$FA,$57,$F8,$08,$FF,$1F
	.dc.b	$B5,$DB,$5F,$92,$F8,$63,$37,$E1,$8F,$95,$1E,$A9,$EC,$7C,$AC,$FE
	.dc.b	$9F,$C8,$7C,$EF,$8B,$F9,$ED,$AF,$83,$3A,$6F,$02,$7E,$AF,$CC,$F0
	.dc.b	$9C,$F9,$FF,$63,$E4,$7E,$CB,$63,$24,$CB,$C5,$83,$E6,$77,$3D,$BE
	.dc.b	$47,$F3,$19,$FE,$21,$B4,$69,$27,$81,$E6,$F8,$0F,$C4,$C7,$E2,$BE
	.dc.b	$2B,$F1,$7E,$ED,$AF,$5B,$72,$F1,$CF,$E4,$3E,$67,$1C,$7F,$8A,$C9
	.dc.b	$F2,$5E,$22,$63,$FA,$FD,$3F,$3B,$18,$C7,$DA,$FE,$6B,$BF,$F3,$6F
	.dc.b	$E1,$7E,$3D,$E7,$FE,$CD,$B8,$7E,$CD,$B7,$FB,$B6,$BF,$E8,$27,$39
	.dc.b	$7E,$2E,$3F,$57,$18,$9F,$D1,$7C,$83,$E4,$7C,$18,$7E,$DB,$04,$E6
	.dc.b	$BF,$0D,$7C,$70,$FE,$05,$39,$7F,$41,$19,$AE,$38,$73,$AD,$47,$F0
	.dc.b	$6F,$B1,$FF,$44,$61,$FA,$A8,$DB,$F9,$3F,$92,$78,$FF,$C4,$DB,$1F
	.dc.b	$C4,$DB,$97,$F1,$B6,$BF,$EA,$27,$35,$F0,$61,$18,$9F,$D6,$FD,$03
	.dc.b	$A7,$F6,$93,$9A,$FF,$55,$F8,$FF,$C4,$BE,$4D,$E5,$FE,$D3,$27,$3F
	.dc.b	$CD,$9C,$DF,$00,$E1,$FE,$51,$1B,$AF,$E6,$57,$FC,$7B,$E4,$B5,$87
	.dc.b	$4D,$31,$5A,$C1,$5A,$C3,$EE,$3D,$63,$F9,$3E,$1E,$7A,$F3,$57,$79
	.dc.b	$F2,$63,$C7,$0D,$46,$2B,$71,$88,$FE,$23,$77,$1F,$24,$7B,$2B,$C9
	.dc.b	$8F,$38,$F2,$C6,$0E,$6E,$0E,$C7,$F3,$1C,$7C,$8F,$63,$A1,$9C,$3C
	.dc.b	$4F,$B0,$C9,$F2,$9C,$AC,$3C,$D3,$D2,$38,$63,$E5,$3E,$33,$B3,$9E
	.dc.b	$3E,$C8,$CD,$7D,$09,$F1,$1E,$93,$CD,$8F,$BC,$DB,$CA,$FC,$0F,$5A
	.dc.b	$F0,$63,$F0,$E3,$9D,$70,$39,$46,$63,$D8,$77,$3E,$04,$69,$FC,$67
	.dc.b	$C1,$FE,$09,$B1,$F8,$58,$C4,$7C,$B6,$FC,$03,$B3,$FB,$C4,$A7,$31
	.dc.b	$86,$E7,$FD,$C1,$B3,$BB,$FE,$A7,$7C,$EB,$E3,$4F,$E9,$3C,$AF,$EA
	.dc.b	$F5,$F9,$FC,$7F,$C0,$93,$AB,$E0,$FC,$B6,$CE,$EB,$8D,$7F,$60,$38
	.dc.b	$D6,$F1,$F4,$D7,$EA,$EF,$9E,$5C,$6B,$F3,$17,$C2,$35,$7F,$01,$B8
	.dc.b	$FA,$27,$AE,$1C,$6F,$B4,$F9,$EF,$85,$78,$F1,$FE,$E7,$86,$CE,$77
	.dc.b	$F9,$4C,$7C,$75,$B8,$F1,$D7,$08,$F9,$37,$D2,$7F,$C3,$B5,$C1,$FF
	.dc.b	$0F,$CF,$A4,$E4,$F5,$9E,$0A,$7E,$4C,$EB,$1F,$51,$E4,$3C,$B8,$79
	.dc.b	$E3,$9B,$87,$C7,$8F,$9A,$E1,$DD,$FE,$05,$7D,$A3,$CB,$8F,$D6,$69
	.dc.b	$F1,$E6,$3F,$B6,$AF,$53,$A3,$A1,$FB,$AF,$88,$76,$FF,$2B,$DF,$FC
	.dc.b	$3F,$18,$9F,$F0,$A3,$5F,$F1,$62,$F1,$F8,$47,$F9,$1A,$34,$FF,$8A
	.dc.b	$03,$87,$AD,$7F,$96,$DC,$B7,$E4,$BF,$53,$8F,$F4,$76,$E1,$FE,$32
	.dc.b	$6D,$7F,$98,$7E,$91,$E4,$F8,$A1,$3A,$5F,$26,$F8,$FD,$36,$3F,$C6
	.dc.b	$13,$B8,$C4,$7F,$92,$5B,$8F,$F5,$96,$E3,$E3,$E1,$E7,$87,$AE,$78
	.dc.b	$DF,$C1,$6B,$FE,$4A,$6B,$FE,$71,$E3,$F9,$3E,$0F,$F9,$C1,$BE,$17
	.dc.b	$F0,$59,$FF,$1D,$E1,$89,$FF,$41,$56,$E7,$FC,$39,$5B,$7E,$84,$65
	.dc.b	$BF,$18,$F8,$9E,$ED,$8F,$F2,$4B,$5F,$F2,$82,$F8,$BD,$CF,$0F,$93
	.dc.b	$95,$C2,$39,$FC,$9C,$8E,$6F,$6F,$0F,$11,$E2,$8F,$4E,$06,$98,$CB
	.dc.b	$CA,$49,$F2,$FC,$C5,$7C,$BF,$29,$EA,$31,$C5,$FE,$8A,$C3,$EA,$CB
	.dc.b	$54,$61,$BC,$3F,$0B,$E6,$3E,$1B,$F1,$3E,$56,$5F,$C4,$8F,$C6,$7C
	.dc.b	$93,$D0,$A4,$9F,$F3,$C4,$7F,$99,$5B,$1F,$9D,$8E,$11,$FE,$B9,$6B
	.dc.b	$FE,$C1,$6B,$FA,$77,$39,$7F,$89,$DF,$E9,$11,$DB,$C1,$F0,$3F,$5B
	.dc.b	$E5,$7E,$D9,$1F,$B7,$F3,$15,$FB,$7F,$2D,$F8,$03,$18,$7C,$7E,$FF
	.dc.b	$C1,$92,$7F,$07,$F2,$D5,$FC,$1C,$59,$F9,$D3,$F1,$3C,$69,$5F,$09
	.dc.b	$F8,$FE,$9E,$23,$DD,$1F,$0C,$8F,$C3,$7F,$C3,$BE,$97,$FA,$BB,$53
	.dc.b	$89,$FF,$68,$C6,$1F,$F1,$6C,$6E,$75,$C5,$F7,$BE,$4A,$F2,$3E,$56
	.dc.b	$98,$89,$87,$23,$F0,$D3,$B7,$15,$F8,$8A,$DC,$7F,$87,$70,$FF,$71
	.dc.b	$CF,$E3,$3C,$1E,$E4,$9E,$36,$F9,$1C,$47,$13,$BF,$C9,$F9,$8C,$F9
	.dc.b	$BE,$DC,$1F,$2C,$F9,$3F,$6A,$F9,$3A,$62,$26,$1E,$89,$D3,$67,$2F
	.dc.b	$86,$BA,$D7,$28,$D1,$C5,$EF,$F2,$7E,$0F,$66,$4F,$58,$CB,$DE,$32
	.dc.b	$F5,$C1,$3E,$8F,$91,$93,$E5,$7D,$7F,$27,$88,$E1,$17,$3F,$85,$6A
	.dc.b	$FC,$15,$39,$3F,$8F,$57,$E0,$E3,$1E,$4E,$A7,$3C,$5F,$8B,$F2,$B1
	.dc.b	$3F,$EF,$AE,$3F,$D7,$36,$FF,$20,$D7,$F1,$ED,$7A,$9F,$87,$1F,$2B
	.dc.b	$FC,$95,$C3,$5F,$BC,$70,$3E,$27,$C9,$7D,$A2,$3F,$5F,$E5,$3E,$EF
	.dc.b	$CA,$78,$78,$FF,$02,$DD,$CD,$31,$F2,$5F,$84,$25,$FE,$07,$E5,$3F
	.dc.b	$0F,$F2,$52,$3F,$08,$FE,$2B,$E5,$3F,$17,$F2,$9D,$49,$23,$FD,$6D
	.dc.b	$97,$E4,$A3,$C5,$ED,$7E,$48,$93,$F8,$FF,$94,$FC,$AF,$B1,$23,$F2
	.dc.b	$4F,$E6,$3E,$4F,$11,$D3,$6F,$EE,$EF,$7E,$3F,$03,$FC,$DE,$32,$EE
	.dc.b	$BE,$11,$FA,$2F,$83,$91,$D8,$FD,$01,$FA,$62,$1F,$A3,$F7,$AB,$E1
	.dc.b	$3E,$61,$0F,$11,$FB,$CA,$FF,$67,$F3,$DF,$93,$FD,$89,$CB,$AB,$F2
	.dc.b	$1C,$FF,$D9,$DB,$EC,$7E,$5A,$31,$5E,$C5,$EC,$E6,$7F,$19,$1F,$DB
	.dc.b	$F5,$FF,$13,$37,$2E,$13,$A3,$A7,$8D,$FE,$03,$C7,$C2,$DC,$77,$5F
	.dc.b	$92,$AD,$F5,$FE,$52,$B5,$3F,$77,$0F,$2B,$63,$2F,$E7,$1B,$1F,$CA
	.dc.b	$67,$A4,$F3,$63,$F9,$E8,$D3,$6B,$C2,$E6,$B9,$9D,$23,$38,$7A,$AF
	.dc.b	$9B,$F2,$AA,$E1,$F2,$6F,$8D,$78,$A7,$A3,$A7,$D4,$E6,$BA,$46,$88
	.dc.b	$F5,$6D,$75,$9E,$51,$FB,$CB,$F2,$CE,$63,$67,$C9,$8C,$BE,$03,$27
	.dc.b	$13,$9E,$1F,$12,$FC,$67,$84,$B5,$D5,$B3,$D4,$E3,$87,$C0,$7E,$61
	.dc.b	$8C,$3A,$E1,$E2,$8F,$B1,$18,$74,$79,$30,$FB,$EE,$8F,$F3,$A1,$4F
	.dc.b	$C8,$D7,$F9,$25,$F8,$07,$F9,$E7,$E0,$3F,$C4,$1F,$2D,$C2,$3F,$D4
	.dc.b	$1E,$9A,$31,$3A,$9E,$AD,$7F,$3B,$E5,$37,$E5,$8A,$31,$3E,$48,$C4
	.dc.b	$EA,$73,$3C,$A3,$F9,$7C,$70,$D3,$FA,$17,$1B,$B1,$C1,$AB,$98,$7E
	.dc.b	$5F,$0E,$1C,$1F,$C4,$30,$E1,$7A,$C7,$F7,$B5,$AB,$FD,$43,$5F,$F4
	.dc.b	$7E,$3D,$9E,$98,$F1,$C7,$F8,$05,$6F,$1E,$B5,$A2,$61,$DA,$FF,$A6
	.dc.b	$4E,$A3,$13,$AB,$E7,$5A,$C3,$C1,$5A,$24,$EA,$BF,$A8,$DF,$38,$C4
	.dc.b	$FF,$83,$7E,$11,$EE,$7A,$7D,$0F,$57,$62,$7F,$8B,$EB,$FD,$0E,$73
	.dc.b	$1C,$23,$07,$F8,$89,$DE,$49,$DD,$F5,$AD,$72,$DA,$FF,$3A,$5F,$37
	.dc.b	$84,$E8,$EA,$FF,$53,$FB,$37,$89,$C7,$27,$D9,$BE,$4B,$A5,$AF,$78
	.dc.b	$72,$BD,$D6,$A7,$35,$FA,$26,$BF,$ED,$76,$BF,$D5,$1A,$FA,$3D,$EF
	.dc.b	$16,$4F,$7A,$C6,$1B,$8F,$86,$74,$D9,$D3,$49,$5F,$08,$F8,$BE,$F7
	.dc.b	$27,$19,$7D,$5F,$7B,$E1,$68,$F9,$5F,$1C,$61,$F2,$FB,$F8,$C7,$C4
	.dc.b	$DC,$39,$E6,$8D,$CF,$9F,$DF,$B5,$AF,$73,$BF,$CB,$7C,$06,$D7,$B7
	.dc.b	$DF,$F8,$7E,$FF,$C4,$6C,$7C,$1F,$7F,$0D,$38,$1F,$17,$BB,$F5,$58
	.dc.b	$FF,$80,$B8,$3F,$25,$18,$8F,$C7,$35,$FC,$4B,$63,$F3,$1E,$DF,$C4
	.dc.b	$F9,$6F,$AA,$DB,$FA,$46,$F7,$FE,$0A,$FF,$22,$AF,$D0,$39,$3A,$7C
	.dc.b	$DF,$F9,$C5,$FC,$53,$71,$FA,$6F,$F9,$CE,$33,$ED,$FE,$56,$1E,$DF
	.dc.b	$D6,$F8,$9C,$36,$B9,$FB,$5B,$38,$7B,$96,$3B,$FB,$3C,$FD,$CE,$8D
	.dc.b	$8F,$1F,$B7,$F8,$2F,$97,$F0,$36,$3C,$FE,$DF,$E1,$BE,$5B,$C8,$D8
	.dc.b	$F5,$FB,$7E,$7F,$33,$D0,$D8,$F7,$FB,$9F,$8E,$E4,$F6,$36,$3E,$1F
	.dc.b	$B5,$F9,$2F,$96,$F8,$0D,$8F,$D8,$7D,$A7,$FB,$F5,$FD,$97,$DA,$B8
	.dc.b	$3F,$75,$39,$E7,$FC,$EC,$E6,$B8,$FC,$4F,$EF,$E3,$13,$FA,$2A,$E1
	.dc.b	$E7,$FF,$18,$9C,$BF,$C6,$A3,$A4,$FE,$2D,$AF,$FA,$E1,$AB,$A9,$18
	.dc.b	$5D,$2F,$77,$EE,$1F,$77,$DE,$B9,$FE,$12,$2E,$1F,$84,$9B,$8F,$F2
	.dc.b	$CB,$7B,$DF,$85,$3F,$0D,$EA,$DC,$7F,$1B,$17,$0F,$C6,$CD,$BE,$51
	.dc.b	$F8,$D3,$F1,$DE,$B5,$BF,$CA,$C5,$3F,$2B,$16,$78,$CE,$8F,$CD,$C6
	.dc.b	$5F,$F6,$DF,$D0,$3C,$8F,$E5,$CF,$F2,$D8,$FE,$54,$E5,$BE,$96,$A7
	.dc.b	$F6,$53,$9C,$3F,$CF,$67,$CD,$70,$FE,$F5,$AF,$83,$C1,$EE,$78,$3C
	.dc.b	$4B,$87,$82,$2C,$FF,$22,$A5,$63,$2F,$E1,$3F,$34,$BF,$E8,$6F,$8E
	.dc.b	$33,$18,$C7,$2D,$C7,$FA,$7E,$2B,$E5,$18,$79,$3F,$E9,$67,$31,$FE
	.dc.b	$2A,$6B,$F0,$1A,$EE,$7E,$2F,$6F,$CC,$3E,$6F,$7D,$BF,$9B,$16,$BD
	.dc.b	$D3,$71,$FE,$EB,$9B,$E6,$BE,$AF,$67,$D6,$6E,$3F,$59,$FA,$64,$7F
	.dc.b	$8F,$4F,$17,$FC,$75,$9B,$0F,$EE,$9E,$E3,$E7,$BF,$5B,$5E,$F9,$CC
	.dc.b	$63,$3F,$F6,$35,$F1,$AF,$EE,$F7,$C7,$AB,$E8,$E1,$C1,$B3,$E0,$8F
	.dc.b	$E7,$57,$F3,$DB,$7C,$E3,$F1,$36,$F8,$07,$C9,$7E,$53,$E6,$FC,$F1
	.dc.b	$8C,$7F,$2B,$18,$9F,$E8,$D5,$C5,$B5,$BB,$EA,$D7,$F2,$9F,$19,$FA
	.dc.b	$0F,$8F,$FD,$0B,$71,$FD,$0B,$F9,$E2,$3F,$31,$F5,$FD,$B7,$EF,$A3
	.dc.b	$FC,$21,$F2,$0C,$FF,$C2,$B3,$9C,$3F,$BD,$CE,$6B,$F2,$33,$99,$FD
	.dc.b	$4B,$5F,$F7,$2B,$5F,$F7,$B3,$7B,$DF,$6F,$FD,$57,$97,$FC,$77,$EA
	.dc.b	$3C,$6F,$F1,$44,$E7,$1F,$F7,$F5,$AB,$51,$FE,$27,$9D,$63,$FC,$96
	.dc.b	$72,$FF,$84,$A3,$35,$C6,$FF,$C6,$3F,$37,$FD,$83,$0F,$F1,$CE,$3F
	.dc.b	$8B,$E9,$CF,$A1,$FE,$41,$8C,$4F,$EC,$2B,$86,$5F,$E9,$46,$BF,$C2
	.dc.b	$9A,$88,$E1,$FE,$1D,$F3,$BF,$E2,$4E,$AF,$E3,$2D,$AF,$E3,$2F,$F1
	.dc.b	$62,$3F,$A4,$7F,$8C,$FC,$2F,$08,$FE,$3E,$DA,$FE,$20,$DC,$3F,$66
	.dc.b	$D9,$FE,$53,$F0,$5F,$E3,$4E,$2F,$E5,$2D,$8F,$E5,$2F,$F2,$62,$3F
	.dc.b	$B4,$7F,$47,$7E,$DF,$85,$F7,$7C,$C3,$E3,$7E,$04,$FF,$9B,$79,$BF
	.dc.b	$D2,$13,$98,$FE,$5F,$5C,$2B,$79,$AF,$71,$FE,$C0,$9D,$57,$F9,$F6
	.dc.b	$72,$F2,$17,$D5,$F0,$63,$C1,$F1,$D7,$D3,$C7,$58,$F7,$AC,$63,$DE
	.dc.b	$FF,$CF,$2F,$93,$1F,$F5,$45,$F5,$9F,$80,$78,$27,$55,$DE,$33,$8F
	.dc.b	$6B,$F3,$C7,$88,$E0,$7A,$E7,$82,$C7,$D1,$8F,$96,$36,$71,$BE,$95
	.dc.b	$C2,$B7,$7D,$AB,$C3,$87,$91,$E2,$60,$F0,$18,$3C,$07,$6C,$3C,$18
	.dc.b	$7B,$E3,$2F,$81,$E1,$7E,$0C,$3C,$51,$C7,$0F,$13,$C3,$0F,$15,$F3
	.dc.b	$BE,$13,$C8,$E1,$7C,$A7,$B1,$B8,$C9,$B7,$35,$F5,$FA,$3F,$D5,$F3
	.dc.b	$F8,$36,$BF,$3D,$AF,$D8,$73,$F1,$2E,$9F,$F6,$14,$FF,$39,$D3,$F3
	.dc.b	$33,$9D,$3F,$75,$7C,$2F,$6D,$8C,$D7,$09,$E7,$8F,$E4,$F1,$FF,$15
	.dc.b	$E3,$F9,$36,$BF,$73,$1E,$CD,$AE,$33,$FA,$A3,$75,$CF,$C8,$FD,$6B
	.dc.b	$F9,$FB,$E2,$DA,$E3,$7F,$C4,$6F,$89,$CE,$F8,$BE,$7C,$7F,$4D,$87
	.dc.b	$28,$D5,$79,$EB,$6F,$F1,$B9,$D3,$6F,$BB,$FC,$06,$76,$D3,$AC,$6A
	.dc.b	$78,$C6,$4F,$CE,$D7,$99,$E9,$7C,$0F,$13,$0F,$8C,$E5,$E3,$7E,$09
	.dc.b	$E5,$0A,$EE,$78,$23,$26,$9C,$DF,$E0,$B1,$F9,$D5,$FC,$10,$F9,$57
	.dc.b	$FC,$72,$3E,$76,$19,$9E,$D5,$F0,$B0,$EB,$7F,$0B,$0E,$B7,$E9,$C7
	.dc.b	$AE,$1E,$A3,$F8,$1F,$33,$EE,$DF,$2D,$FE,$54,$9F,$EC,$31,$87,$FD
	.dc.b	$84,$D7,$FC,$B3,$3C,$67,$84,$7F,$A1,$E7,$56,$DB,$FB,$5F,$C2,$3E
	.dc.b	$33,$F9,$8C,$61,$B7,$C7,$0B,$8F,$97,$E3,$AE,$4F,$8C,$A6,$22,$5F
	.dc.b	$F6,$1C,$FC,$0F,$8E,$76,$E7,$2F,$6D,$62,$3D,$77,$D5,$F7,$D7,$B3
	.dc.b	$0F,$8C,$D9,$D4,$B0,$EF,$4C,$FF,$7D,$3D,$CF,$F5,$53,$87,$86,$3F
	.dc.b	$3B,$0F,$83,$0F,$97,$7C,$2F,$CC,$8C,$36,$33,$1F,$41,$AF,$D9,$6E
	.dc.b	$3F,$0A,$1F,$30,$E9,$EE,$D8,$FF,$3E,$D7,$0C,$7F,$C4,$3F,$4C,$F3
	.dc.b	$3E,$9F,$86,$F9,$9A,$7D,$69,$D6,$1E,$FD,$D1,$88,$F7,$9F,$E2,$79
	.dc.b	$E2,$EE,$35,$A7,$9F,$EE,$7F,$BA,$4F,$E3,$1A,$FE,$55,$AF,$FA,$A9
	.dc.b	$AF,$F8,$09,$DC,$7D,$4F,$1F,$F9,$AB,$7A,$FE,$7C,$EB,$4F,$CF,$FC
	.dc.b	$6F,$DD,$26,$1F,$A0,$6B,$EA,$CF,$F5,$1F,$1F,$F9,$43,$F2,$B3,$F9
	.dc.b	$AF,$2D,$F7,$4F,$88,$FD,$2C,$61,$B1,$9C,$3F,$61,$39,$6C,$65,$F0
	.dc.b	$74,$3E,$3F,$F4,$67,$79,$D6,$DF,$7F,$E6,$1E,$5F,$E2,$5B,$87,$ED
	.dc.b	$9A,$FE,$FD,$AF,$FB,$51,$CC,$FE,$AB,$17,$DD,$BE,$8B,$DC,$DB,$FF
	.dc.b	$62,$DF,$0E,$0F,$F1,$9B,$61,$7C,$AF,$89,$FA,$09,$D7,$07,$F1,$3E
	.dc.b	$6F,$D7,$12,$FF,$88,$B5,$F8,$7D,$7F,$C1,$0F,$E0,$B1,$FC,$1B,$E3
	.dc.b	$E4,$FE,$4F,$81,$1F,$C2,$BE,$43,$33,$F3,$22,$ED,$FB,$AA,$57,$97
	.dc.b	$AB,$E5,$12,$FE,$53,$5F,$CE,$8C,$FF,$DA,$4F,$C4,$F3,$79,$3F,$1A
	.dc.b	$7E,$04,$48,$C1,$E7,$F9,$0F,$E3,$E4,$7D,$5D,$55,$FC,$1C,$5C,$BD
	.dc.b	$13,$FC,$9B,$9B,$F8,$B1,$2B,$F9,$CB,$5F,$C5,$FC,$87,$F2,$C3,$83
	.dc.b	$F1,$F8,$FF,$97,$91,$F9,$5F,$39,$5F,$C9,$C5,$C3,$E4,$4D,$8F,$E9
	.dc.b	$25,$9F,$E8,$91,$FC,$F3,$E6,$DF,$25,$49,$3F,$C2,$61,$F3,$98,$8F
	.dc.b	$D8,$47,$07,$53,$CE,$7A,$7C,$94,$9E,$4F,$7F,$5B,$11,$CA,$2E,$5F
	.dc.b	$CE,$30,$57,$F5,$0F,$BC,$7C,$AF,$F4,$16,$E1,$FD,$97,$A5,$18,$8F
	.dc.b	$EC,$6D,$7F,$D2,$71,$A9,$FE,$B7,$F1,$8F,$8E,$A7,$78,$95,$FC,$A6
	.dc.b	$74,$D9,$CB,$F0,$BE,$53,$FB,$2F,$9B,$FC,$CE,$3F,$98,$35,$D1,$F2
	.dc.b	$FE,$53,$7E,$CC,$8C,$1F,$3F,$E5,$3E,$A6,$D2,$30,$7D,$3F,$93,$EF
	.dc.b	$EE,$C8,$C1,$F5,$FE,$4F,$C7,$F0,$72,$30,$7D,$BD,$7F,$01,$F3,$B2
	.dc.b	$30,$7D,$FF,$91,$FC,$21,$1F,$C1,$7B,$BF,$87,$F9,$2F,$7F,$AF,$27
	.dc.b	$F0,$D1,$F8,$BF,$7A,$47,$E2,$5F,$C7,$7C,$8F,$E3,$E3,$FB,$D6,$FF
	.dc.b	$DA,$F9,$CD,$31,$F7,$AF,$91,$F9,$BE,$74,$BF,$EB,$93,$96,$FC,$87
	.dc.b	$F8,$16,$0D,$24,$7E,$1E,$1F,$21,$FE,$1C,$5F,$1B,$4C,$45,$9E,$FE
	.dc.b	$A6,$CC,$3F,$D4,$6B,$FB,$3F,$8D,$FE,$2C,$F6,$BF,$06,$75,$F3,$7B
	.dc.b	$79,$B2,$FD,$2F,$87,$E2,$B1,$1C,$FE,$27,$C8,$63,$C7,$FF,$0D,$E4
	.dc.b	$8F,$C7,$FC,$44,$AF,$87,$1E,$BF,$7E,$47,$C6,$7D,$DF,$0B,$88,$F4
	.dc.b	$7C,$47,$C1,$31,$8F,$6A,$C4,$FC,$E9,$CE,$DF,$EF,$CF,$9A,$78,$5F
	.dc.b	$E2,$36,$D7,$F9,$41,$AF,$F9,$65,$A8,$E5,$1F,$E4,$CD,$FF,$CA,$B5
	.dc.b	$9E,$8F,$F0,$F7,$C3,$FF,$8C,$B0,$E7,$9F,$C2,$ED,$FB,$1E,$14,$CF
	.dc.b	$F0,$71,$F6,$CC,$3F,$56,$30,$FF,$90,$47,$CE,$3E,$75,$C5,$E5,$F8
	.dc.b	$CF,$F1,$95,$6A,$72,$FF,$BD,$70,$F5,$39,$AE,$11,$8B,$7C,$56,$B0
	.dc.b	$E7,$1D,$EB,$77,$F0,$E3,$0F,$53,$18,$78,$2B,$8D,$F3,$C3,$B4,$E5
	.dc.b	$B8,$74,$BF,$21,$CC,$F7,$D7,$17,$AC,$6F,$0F,$0E,$1E,$6C,$3A,$57
	.dc.b	$96,$F8,$3D,$27,$57,$C6,$B7,$3D,$8C,$44,$7C,$57,$DB,$0F,$8D,$87
	.dc.b	$48,$C4,$F9,$0D,$DF,$48,$F9,$71,$CC,$FA,$27,$B8,$AF,$CA,$6C,$7D
	.dc.b	$87,$7F,$19,$FE,$9F,$F8,$1F,$F5,$2B,$63,$FD,$5C,$D7,$7C,$BF,$E2
	.dc.b	$67,$FD,$65,$6F,$61,$F2,$BF,$EA,$E8,$C3,$63,$33,$C1,$AF,$99,$CE
	.dc.b	$BF,$84,$6E,$9F,$8C,$F1,$1F,$33,$FE,$A8,$AD,$61,$F5,$EF,$78,$7E
	.dc.b	$25,$AA,$F8,$F2,$EE,$A4,$FD,$2A,$DD,$6A,$31,$3A,$BE,$F1,$F8,$99
	.dc.b	$D4,$60,$98,$FE,$7E,$F9,$E3,$E7,$8C,$61,$CD,$B8,$73,$9D,$3E,$99
	.dc.b	$D3,$4D,$63,$8E,$5F,$F5,$07,$C6,$FF,$B2,$B9,$3F,$D7,$BC,$EF,$9D
	.dc.b	$C5,$FF,$09,$8F,$F6,$A3,$FE,$A1,$74,$DA,$DE,$0C,$3A,$CE,$F9,$33
	.dc.b	$18,$9F,$F2,$5F,$3B,$C0,$FF,$5D,$47,$F2,$36,$BF,$C7,$63,$57,$FC
	.dc.b	$8E,$73,$29,$CB,$D2,$3B,$E2,$F9,$4F,$2C,$EE,$73,$F3,$BA,$F8,$CB
	.dc.b	$8F,$B5,$B5,$FA,$4C,$BF,$D8,$D5,$E4,$6D,$7F,$89,$F1,$4E,$5B,$1F
	.dc.b	$1E,$3C,$7F,$19,$D3,$7F,$9F,$F1,$9E,$63,$CF,$ED,$DB,$F9,$CF,$32
	.dc.b	$3D,$DE,$DF,$15,$C2,$3E,$A3,$73,$FF,$5D,$65,$FE,$7E,$9F,$B1,$F0
	.dc.b	$FF,$13,$E2,$BE,$2B,$F1,$BE,$2B,$D0,$49,$F6,$3E,$BF,$88,$F9,$67
	.dc.b	$CC,$F8,$8B,$3F,$83,$EB,$F9,$E7,$D0,$F0,$AC,$7E,$1F,$E2,$3E,$21
	.dc.b	$1F,$90,$FC,$77,$EB,$9F,$63,$E2,$2B,$FE,$EF,$F9,$9F,$F3,$5B,$7E
	.dc.b	$27,$F1,$71,$FE,$6A,$8C,$1B,$F8,$7F,$AD,$1F,$E6,$08,$FF,$2E,$37
	.dc.b	$83,$EE,$35,$FF,$2D,$AF,$9C,$EA,$76,$F8,$9B,$5E,$3F,$87,$F0,$FC
	.dc.b	$E7,$89,$B1,$E7,$F8,$7F,$09,$1F,$0B,$E4,$F8,$6F,$49,$EA,$F7,$AC
	.dc.b	$FB,$FE,$1B,$DB,$E4,$7B,$9B,$1F,$0F,$E1,$BD,$A4,$7D,$AF,$C0,$F8
	.dc.b	$6F,$8A,$7C,$6F,$86,$B3,$FE,$07,$F0,$0F,$F8,$17,$E7,$3F,$4C,$FF
	.dc.b	$9D,$76,$FD,$BC,$63,$89,$5C,$1F,$DF,$7C,$52,$B7,$E4,$7E,$96,$F8
	.dc.b	$57,$FA,$19,$AF,$78,$D1,$C3,$0F,$BD,$F0,$F6,$7B,$C7,$F8,$67,$95
	.dc.b	$F0,$FF,$80,$3F,$03,$F1,$16,$7F,$0D,$16,$B8,$9F,$C4,$E9,$59,$D3
	.dc.b	$D5,$E4,$A5,$F1,$1F,$8C,$3F,$1B,$EA,$5C,$BF,$25,$17,$1E,$B7,$C6
	.dc.b	$B1,$87,$4C,$3F,$C6,$98,$7E,$66,$FF,$2F,$E9,$7E,$61,$91,$93,$F3
	.dc.b	$D1,$88,$EB,$3C,$9A,$72,$9E,$0F,$8B,$E6,$FF,$CE,$B1,$F7,$4F,$F2
	.dc.b	$03,$FE,$7F,$71,$1F,$CD,$63,$6F,$1F,$89,$FE,$6B,$18,$C3,$E2,$B6
	.dc.b	$B9,$FC,$EE,$BE,$27,$7F,$11,$27,$AE,$AB,$7C,$EE,$73,$F3,$5D,$CB
	.dc.b	$AC,$DC,$3F,$C8,$3F,$39,$FE,$8A,$7F,$C9,$8F,$86,$FF,$CB,$CE,$5F
	.dc.b	$BA,$1D,$36,$FF,$A2,$C6,$E3,$FC,$3A,$D7,$BC,$68,$FE,$8B,$1D,$6F
	.dc.b	$F3,$DE,$FF,$F9,$6B,$E3,$73,$5F,$1B,$E2,$E6,$3F,$03,$3F,$F4,$FB
	.dc.b	$63,$39,$7D,$57,$E6,$FC,$66,$32,$F9,$BF,$19,$F4,$3D,$3F,$A2,$DC
	.dc.b	$3A,$38,$7B,$7C,$65,$C7,$E0,$ED,$E0,$7F,$3F,$B3,$E7,$7F,$71,$97
	.dc.b	$ED,$E7,$E3,$67,$FD,$DA,$B5,$D9,$F0,$23,$E3,$93,$2F,$F2,$1D,$7D
	.dc.b	$08,$C3,$E0,$79,$7C,$37,$FA,$62,$7E,$BC,$63,$E1,$FF,$17,$F1,$96
	.dc.b	$7E,$33,$6F,$CF,$6A,$FC,$8F,$C3,$E7,$3F,$C8,$F2,$7E,$53,$E2,$FF
	.dc.b	$2A,$DC,$BE,$83,$87,$E9,$7C,$65,$CB,$F0,$BF,$3B,$FB,$C9,$F5,$C2
	.dc.b	$30,$FB,$9A,$FF,$98,$9A,$FF,$A5,$1D,$E1,$F2,$DB,$3F,$EA,$DF,$BE
	.dc.b	$FE,$AB,$5E,$0F,$07,$FC,$3F,$F3,$D8,$C3,$F8,$7F,$BC,$E6,$CF,$AD
	.dc.b	$FE,$24,$8C,$65,$FA,$AB,$E0,$FF,$AE,$5B,$8F,$F8,$B5,$FE,$C0,$FF
	.dc.b	$12,$72,$BE,$7B,$FA,$E3,$FE,$AC,$6B,$FE,$C7,$6B,$BF,$5F,$FC,$57
	.dc.b	$1F,$D9,$72,$FC,$43,$6F,$FA,$47,$D2,$FF,$90,$AF,$F6,$33,$97,$F8
	.dc.b	$67,$C8,$FF,$89,$7B,$FF,$09,$1F,$E2,$96,$BF,$EB,$EF,$0B,$F8,$34
	.dc.b	$EC,$FE,$22,$7F,$12,$F8,$DB,$8F,$EB,$5B,$5F,$C7,$DB,$97,$F1,$67
	.dc.b	$84,$FE,$CF,$9E,$D7,$F2,$5F,$89,$F2,$12,$7F,$3B,$8F,$F1,$78,$C4
	.dc.b	$EE,$7F,$79,$F2,$7F,$C9,$27,$FC,$93,$3F,$ED,$B1,$F3,$EF,$3F,$FD
	.dc.b	$B3,$9B,$A7,$FB,$34,$E6,$BF,$D5,$BF,$21,$FE,$F7,$8C,$47,$F6,$C9
	.dc.b	$DD,$FE,$C2,$BF,$DF,$98,$6B,$0E,$99,$62,$B5,$87,$F9,$A5,$F0,$E1
	.dc.b	$FD,$D6,$FA,$CF,$9F,$1E,$53,$DA,$3C,$67,$AA,$B8,$DF,$95,$EB,$5C
	.dc.b	$23,$13,$C2,$30,$F1,$39,$DE,$0F,$0A,$C7,$D0,$FF,$84,$70,$F3,$61
	.dc.b	$D8,$F0,$C7,$33,$FD,$FC,$5C,$7E,$0B,$93,$D9,$1C,$30,$EF,$7E,$69
	.dc.b	$F9,$98,$D7,$67,$85,$C3,$E5,$3B,$61,$F1,$5F,$0E,$3E,$BC,$3E,$7B
	.dc.b	$63,$E1,$47,$84,$F5,$3D,$4F,$31,$D6,$38,$3D,$23,$DF,$1E,$88,$FB
	.dc.b	$47,$17,$07,$07,$F2,$1D,$BF,$EA,$77,$EC,$B5,$F8,$ED,$77,$06,$7C
	.dc.b	$93,$B7,$FC,$77,$0A,$E5,$39,$D3,$84,$62,$B6,$D8,$CD,$F5,$B5,$F1
	.dc.b	$7D,$67,$E9,$7A,$B5,$3F,$A0,$C7,$FC,$5F,$3F,$84,$38,$E3,$E4,$9F
	.dc.b	$D2,$DF,$37,$85,$F2,$BF,$01,$CA,$FB,$D6,$E3,$C8,$F8,$F5,$E3,$3F
	.dc.b	$85,$9E,$11,$DB,$0F,$86,$DC,$7C,$93,$D7,$1E,$37,$DA,$FC,$D3,$D9
	.dc.b	$E2,$F9,$31,$F8,$98,$70,$3A,$57,$E7,$71,$F2,$56,$E3,$C9,$5C,$23
	.dc.b	$A6,$3F,$9B,$AD,$D7,$DB,$9F,$49,$C5,$E9,$87,$89,$FB,$EF,$99,$E1
	.dc.b	$7F,$25,$A7,$D1,$AF,$90,$F0,$BF,$1B,$0E,$AF,$8E,$BB,$47,$DD,$35
	.dc.b	$87,$B8,$F7,$E3,$F2,$09,$8F,$B2,$7E,$55,$F0,$8F,$55,$71,$7F,$31
	.dc.b	$BB,$E6,$5F,$29,$FC,$BB,$7C,$C6,$31,$FF,$28,$B5,$FF,$32,$6D,$B5
	.dc.b	$7C,$EF,$FC,$06,$D8,$FE,$87,$5C,$5F,$EA,$CD,$7B,$FC,$8F,$39,$F2
	.dc.b	$36,$4D,$F9,$3C,$39,$F9,$39,$7E,$68,$AF,$FA,$2A,$93,$8D,$7F,$D2
	.dc.b	$B1,$D2,$7D,$B3,$E8,$8D,$C7,$A2,$30,$7F,$87,$97,$CF,$3E,$3B,$E1
	.dc.b	$46,$1B,$19,$8F,$F4,$28,$F9,$FF,$E1,$98,$7F,$92,$67,$53,$FE,$44
	.dc.b	$6B,$FC,$04,$E8,$7F,$8B,$9C,$BF,$6B,$E7,$DE,$7F,$C9,$6C,$7F,$A2
	.dc.b	$1A,$FF,$AB,$17,$A1,$B9,$FA,$1E,$1B,$C8,$A6,$6F,$D3,$F3,$FF,$DC
	.dc.b	$5B,$D7,$F0,$23,$11,$FE,$B8,$8F,$98,$FF,$A2,$A3,$2F,$AD,$AF,$F1
	.dc.b	$D6,$BF,$CA,$DB,$C5,$F9,$5F,$94,$FC,$11,$27,$F2,$7E,$75,$7E,$F7
	.dc.b	$CB,$7E,$1C,$C6,$FF,$E8,$EF,$9F,$7C,$77,$E5,$FD,$24,$62,$FF,$D7
	.dc.b	$31,$89,$FC,$87,$CB,$EE,$B5,$5F,$8B,$F9,$7F,$CA,$19,$AF,$C0,$FC
	.dc.b	$F7,$CC,$9F,$BE,$D7,$FD,$56,$DC,$7C,$9F,$27,$C3,$93,$F1,$93,$BA
	.dc.b	$FC,$03,$5F,$E8,$4D,$7F,$A8,$3C,$FB,$FF,$A9,$68,$F9,$F7,$91,$EF
	.dc.b	$6D,$F9,$30,$FD,$67,$34,$8F,$DE,$BF,$AF,$F9,$9C,$47,$EA,$7B,$3F
	.dc.b	$64,$63,$5F,$D4,$FC,$C7,$ED,$49,$3F,$C1,$BE,$66,$BF,$B1,$8B,$3F
	.dc.b	$B0,$9F,$8F,$D7,$2B,$E3,$BF,$3F,$6C,$47,$FA,$9B,$7F,$CC,$BF,$12
	.dc.b	$7E,$39,$1F,$9F,$EE,$FD,$23,$8E,$1F,$BE,$F9,$9F,$E1,$44,$9F,$E3
	.dc.b	$7F,$33,$5F,$E0,$D1,$67,$F8,$2C,$FF,$11,$E4,$95,$FC,$89,$FE,$29
	.dc.b	$F3,$38,$8F,$E1,$DC,$BF,$C5,$CC,$5F,$F0,$ED,$3F,$8D,$12,$7F,$95
	.dc.b	$FC,$CD,$7F,$8B,$45,$9F,$E2,$B3,$63,$F9,$99,$67,$F9,$7F,$8A,$9F
	.dc.b	$E5,$1F,$28,$F6,$A9,$2B,$FC,$2D,$83,$0F,$E8,$34,$F9,$FD,$52,$B4
	.dc.b	$FF,$3C,$9E,$8F,$2A,$E3,$7E,$0B,$F0,$FC,$CC,$AD,$1C,$5F,$2F,$B7
	.dc.b	$98,$ED,$7E,$29,$C6,$3E,$0F,$0D,$93,$0F,$E8,$9F,$3E,$E9,$A4,$CB
	.dc.b	$BD,$F0,$F9,$BC,$47,$F2,$98,$57,$18,$FF,$4A,$B5,$D3,$5F,$F5,$53
	.dc.b	$9C,$7C,$9F,$26,$F9,$5F,$EA,$CD,$AF,$0D,$6F,$5F,$9B,$F3,$7F,$50
	.dc.b	$93,$E6,$D5,$F2,$BF,$63,$E6,$FF,$B3,$C1,$A4,$9F,$F5,$5C,$3E,$6F
	.dc.b	$EF,$B2,$3A,$47,$E0,$BE,$6A,$47,$77,$F0,$BF,$37,$F8,$6F,$9B,$EE
	.dc.b	$49,$EA,$F8,$B6,$FC,$5F,$CD,$C9,$F4,$47,$E3,$BE,$66,$47,$AD,$FC
	.dc.b	$8F,$B9,$F9,$2F,$9A,$F5,$92,$7D,$2F,$BB,$E0,$BF,$2E,$48,$F8,$9F
	.dc.b	$35,$F9,$A6,$1F,$1C,$FC,$EF,$CD,$7F,$81,$7E,$79,$F3,$5F,$E0,$E6
	.dc.b	$CF,$FB,$41,$AF,$E0,$5A,$FF,$B9,$57,$CC,$F3,$24,$7E,$2A,$1F,$35
	.dc.b	$DF,$E6,$64,$E5,$F1,$7C,$D7,$8F,$E6,$3C,$44,$9F,$13,$E5,$F3,$7D
	.dc.b	$04,$8E,$DE,$9F,$AF,$E5,$3D,$9E,$7C,$9F,$14,$57,$D5,$1E,$F2,$3E
	.dc.b	$FF,$98,$F8,$84,$7D,$51,$F1,$FE,$06,$47,$B5,$F9,$5F,$31,$88,$F8
	.dc.b	$D1,$F2,$C8,$FC,$B7,$E8,$7C,$6C,$8E,$47,$F8,$73,$E7,$5F,$27,$FE
	.dc.b	$22,$6C,$66,$3E,$C7,$C2,$E5,$FF,$15,$4E,$A7,$FC,$3C,$1C,$3F,$E2
	.dc.b	$16,$BF,$E5,$8E,$27,$C9,$FE,$0C,$91,$F6,$6A,$BF,$C3,$8F,$F2,$AC
	.dc.b	$BE,$4F,$F1,$46,$BC,$AF,$E3,$F2,$F9,$E7,$AD,$FE,$B9,$8F,$F7,$76
	.dc.b	$3F,$EB,$39,$CF,$C8,$7F,$B0,$7E,$53,$C7,$5F,$EC,$39,$CB,$C0,$C7
	.dc.b	$C7,$FF,$AE,$E7,$71,$DF,$0E,$D1,$8A,$E2,$D8,$CE,$1E,$1B,$E4,$F8
	.dc.b	$E3,$37,$E1,$6B,$D7,$0F,$34,$F1,$8A,$6E,$2B,$B3,$D0,$5B,$F2,$5F
	.dc.b	$82,$38,$1A,$C3,$CF,$7E,$3B,$E7,$1F,$E4,$2B,$F5,$46,$30,$F2,$CF
	.dc.b	$0B,$F8,$A6,$66,$3D,$67,$78,$78,$63,$CC,$7F,$99,$6B,$E7,$B6,$7E
	.dc.b	$C3,$6B,$F1,$AF,$FB,$AB,$85,$F0,$98,$F9,$44,$62,$7F,$27,$18,$9D
	.dc.b	$4E,$67,$C1,$ED,$BC,$0E,$B9,$23,$13,$F9,$C8,$C4,$EA,$73,$3C,$A3
	.dc.b	$E3,$C7,$0D,$3F,$0B,$E0,$BC,$9F,$EE,$71,$F8,$59,$D6,$FF,$93,$6E
	.dc.b	$1F,$91,$6A,$BE,$5E,$87,$6C,$64,$FD,$9A,$DD,$6A,$31,$3A,$C3,$C0
	.dc.b	$FE,$1A,$75,$18,$9D,$91,$F4,$E3,$E8,$C7,$CD,$7C,$DB,$5F,$8E,$BE
	.dc.b	$71,$8B,$E8,$45,$32,$76,$FB,$EB,$F3,$7E,$9B,$EF,$DF,$05,$FE,$1A
	.dc.b	$E4,$FF,$06,$43,$E1,$31,$B7,$FB,$8A,$36,$FF,$82,$9D,$36,$B7,$8F
	.dc.b	$FB,$7A,$72,$FF,$0D,$C3,$C1,$3A,$7C,$57,$69,$F0,$3F,$C3,$AF,$C0
	.dc.b	$D7,$A3,$6B,$D4,$D9,$E3,$87,$96,$3A,$7C,$0F,$57,$B7,$C0,$F7,$E7
	.dc.b	$B5,$EB,$6E,$3F,$E2,$3C,$3F,$D4,$73,$FC,$C3,$1F,$91,$7D,$C9,$1F
	.dc.b	$0D,$F5,$71,$72,$D7,$E5,$FC,$07,$B4,$F7,$55,$3E,$51,$E9,$21,$DF
	.dc.b	$0F,$F2,$93,$E1,$FC,$03,$FC,$B0,$F0,$3F,$C7,$6E,$32,$F9,$D3,$C1
	.dc.b	$FA,$10,$F7,$FE,$83,$6B,$8F,$BF,$89,$CF,$B5,$F3,$6B,$FC,$63,$5E
	.dc.b	$18,$C1,$9F,$6B,$9F,$BD,$D1,$B3,$E4,$F6,$BB,$9E,$0F,$6E,$C7,$A3
	.dc.b	$DB,$EC,$47,$DC,$F8,$7D,$AF,$3F,$B7,$E8,$6C,$FC,$0F,$67,$D7,$ED
	.dc.b	$FB,$1B,$1F,$13,$DA,$F5,$11,$F9,$4F,$B7,$DA,$F8,$7D,$FF,$11,$B3
	.dc.b	$F3,$38,$FF,$2D,$E5,$7F,$B1,$A3,$07,$F8,$87,$E1,$FF,$35,$F1,$1F
	.dc.b	$E2,$28,$FC,$33,$6B,$FC,$D8,$D7,$F6,$A3,$B1,$19,$8F,$DC,$C6,$3D
	.dc.b	$8F,$0F,$CB,$76,$6D,$FF,$9E,$AD,$EC,$F9,$BC,$DF,$1B,$67,$5E,$CF
	.dc.b	$E9,$E8,$D2,$47,$F4,$0F,$9E,$7B,$7F,$AF,$D1,$18,$9D,$4E,$63,$13
	.dc.b	$EE,$F6,$7E,$1F,$B5,$F1,$1B,$7E,$77,$C4,$47,$E3,$B6,$7F,$83,$62
	.dc.b	$8F,$9B,$ED,$73,$AF,$81,$EC,$FD,$0F,$4B,$E8,$B6,$32,$7C,$58,$CB
	.dc.b	$4F,$91,$EC,$FD,$8F,$6B,$EC,$B6,$32,$7C,$D8,$CB,$4F,$A1,$EC,$7E
	.dc.b	$03,$D9,$FC,$0B,$63,$27,$D5,$8C,$B4,$CE,$9E,$E8,$CB,$63,$EE,$73
	.dc.b	$7F,$23,$F9,$6F,$C6,$B6,$32,$49,$F8,$FE,$A6,$A3,$F0,$9C,$BF,$94
	.dc.b	$E7,$FC,$AB,$63,$27,$E2,$63,$2D,$3F,$DA,$75,$F8,$0E,$0F,$E7,$7E
	.dc.b	$C3,$E7,$BF,$96,$CF,$F9,$1C,$D5,$7F,$B6,$A7,$5E,$97,$F4,$29,$C9
	.dc.b	$FE,$58,$E6,$FB,$95,$F8,$FC,$3F,$14,$47,$C3,$1A,$9E,$CB,$8F,$A3
	.dc.b	$9F,$3F,$FD,$C5,$F8,$0B,$E6,$3F,$A6,$2E,$1B,$18,$6E,$3F,$EE,$29
	.dc.b	$D3,$67,$4E,$7E,$63,$FA,$A4,$74,$6C,$64,$99,$7F,$5D,$A7,$BD,$B9
	.dc.b	$F9,$7F,$35,$FE,$72,$7C,$1F,$29,$F1,$63,$17,$C6,$31,$C9,$FB,$19
	.dc.b	$CC,$FF,$08,$0E,$1F,$E1,$6D,$7F,$C0,$2F,$4D,$3F,$CE,$ED,$BF,$F0
	.dc.b	$2C,$7F,$BB,$38,$FB,$C7,$C3,$F9,$5F,$EE,$59,$7C,$A6,$C6,$49,$87
	.dc.b	$BF,$E5,$75,$1F,$2F,$BB,$FB,$D6,$DF,$49,$B1,$92,$57,$C7,$F9,$5D
	.dc.b	$47,$90,$FE,$43,$F0,$7F,$C3,$27,$F9,$3F,$8C,$8C,$46,$BC,$3F,$F0
	.dc.b	$14,$62,$3F,$96,$3F,$C5,$36,$5E,$CF,$C0,$7C,$6F,$F8,$46,$75,$1F
	.dc.b	$CD,$3E,$6D,$EC,$FF,$86,$AF,$F4,$1F,$0F,$F9,$B9,$CD,$79,$F1,$56
	.dc.b	$A3,$F5,$33,$A9,$FF,$5B,$B5,$FB,$4D,$7C,$17,$FC,$BF,$F0,$27,$BB
	.dc.b	$49,$7F,$E7,$48,$CB,$63,$F5,$D1,$8A,$FF,$5D,$60,$F3,$1E,$B6,$AF
	.dc.b	$F4,$BF,$2B,$FE,$29,$EB,$34,$99,$7F,$87,$E3,$2F,$E9,$E3,$F5,$BF
	.dc.b	$07,$CA,$3F,$55,$F0,$98,$69,$FA,$D8,$FD,$9C,$67,$3F,$D3,$37,$3F
	.dc.b	$DC,$F8,$7E,$32,$5F,$85,$B5,$FA,$D7,$F7,$BF,$09,$FB,$E7,$F8,$74
	.dc.b	$7F,$B3,$47,$CE,$3D,$7F,$ED,$F8,$3C,$EF,$C1,$D6,$A7,$FD,$D7,$F0
	.dc.b	$FF,$E6,$8B,$E1,$3F,$A6,$AD,$C6,$23,$FD,$A8,$3C,$0C,$61,$D7,$3F
	.dc.b	$F3,$8C,$6E,$7C,$8E,$AF,$C2,$E6,$18,$78,$71,$F3,$D7,$81,$F2,$CF
	.dc.b	$9C,$F2,$5E,$DF,$E0,$95,$5F,$C9,$D7,$17,$27,$13,$33,$B7,$CB,$5E
	.dc.b	$08,$F4,$BE,$0B,$F4,$46,$27,$9E,$38,$7A,$1F,$45,$C1,$ED,$C3,$C9
	.dc.b	$8F,$C5,$9D,$9F,$00,$F2,$39,$9E,$11,$8C,$7D,$4F,$D0,$8E,$B8,$F9
	.dc.b	$4F,$44,$7D,$37,$0F,$C6,$30,$7A,$1D,$C7,$63,$9C,$7C,$6C,$7D,$87
	.dc.b	$88,$B3,$D2,$3E,$49,$D0,$FC,$1C,$69,$F9,$FE,$5F,$E1,$5B,$1F,$6A
	.dc.b	$31,$1F,$8A,$6F,$47,$F5,$DC,$7F,$D7,$31,$8F,$23,$F8,$6D,$23,$13
	.dc.b	$AB,$FF,$25,$DF,$1A,$D4,$62,$F8,$BF,$99,$C1,$EB,$7F,$94,$A3,$0D
	.dc.b	$8E,$18,$7D,$FB,$FE,$61,$7C,$A7,$FC,$B2,$72,$8D,$35,$F8,$98,$78
	.dc.b	$F0,$FD,$43,$E5,$CB,$BC,$FE,$5E,$B8,$46,$AF,$DE,$DC,$3A,$E3,$E5
	.dc.b	$BF,$F6,$ED,$F7,$C7,$51,$E0,$C3,$8C,$F9,$AB,$FC,$DF,$87,$82,$BF
	.dc.b	$45,$8F,$8E,$BD,$F1,$FA,$57,$C6,$71,$7F,$59,$1B,$7F,$DD,$AD,$7F
	.dc.b	$87,$57,$69,$E1,$1D,$AF,$C6,$F0,$7E,$53,$EA,$9F,$63,$C8,$FA,$15
	.dc.b	$F2,$6B,$CB,$7E,$92,$38,$24,$F3,$3E,$65,$78,$DF,$06,$3E,$AC,$3C
	.dc.b	$31,$DC,$FC,$D5,$7C,$B3,$E1,$E3,$F3,$30,$EF,$3D,$4F,$65,$FF,$11
	.dc.b	$ED,$72,$3E,$13,$F9,$7F,$07,$FA,$6A,$31,$7F,$E2,$F9,$CB,$FE,$A1
	.dc.b	$87,$C3,$62,$F9,$D7,$F8,$7E,$B9,$35,$E0,$D8,$DC,$7F,$57,$E7,$E5
	.dc.b	$7F,$CF,$A3,$FC,$69,$86,$26,$47,$F9,$06,$36,$61,$FE,$F6,$FF,$AC
	.dc.b	$BF,$03,$7C,$9F,$95,$B3,$FE,$01,$C1,$18,$8E,$E6,$5C,$3F,$E1,$A6
	.dc.b	$BD,$67,$D5,$18,$80,$FC,$15,$CF,$F2,$23,$0D,$8C,$C7,$F9,$E3,$F0
	.dc.b	$57,$91,$9A,$F9,$13,$C0,$F9,$55,$C1,$F9,$CD,$7F,$C9,$DF,$3D,$FE
	.dc.b	$04,$7E,$83,$5F,$A8,$D7,$7D,$BD,$A3,$5C,$3F,$47,$E7,$DE,$6F,$A5
	.dc.b	$B7,$FE,$8E,$AD,$EF,$FE,$54,$FC,$19,$F2,$7F,$86,$C5,$18,$BF,$F1
	.dc.b	$84,$62,$7E,$A7,$CC,$EE,$B5,$5F,$E6,$7A,$D3,$F8,$B6,$B9,$8D,$BF
	.dc.b	$57,$E3,$7E,$F1,$1F,$B5,$E3,$D7,$F0,$A7,$FB,$E7,$E3,$9E,$9F,$E6
	.dc.b	$5B,$1F,$81,$AF,$C4,$7C,$6C,$8F,$C2,$3F,$8B,$F8,$DC,$47,$4E,$85
	.dc.b	$BD,$3E,$86,$B4,$FC,$E7,$D2,$BE,$27,$F0,$8D,$C3,$FD,$72,$D7,$FC
	.dc.b	$87,$93,$DE,$DD,$7E,$12,$B5,$5F,$E8,$D6,$BF,$92,$0E,$1F,$CA,$BD
	.dc.b	$6F,$FC,$57,$97,$E8,$BE,$63,$F4,$C7,$6A,$FD,$17,$9D,$FA,$92,$3F
	.dc.b	$A5,$F8,$FA,$FC,$1D,$7F,$32,$FE,$B5,$AF,$C5,$EE,$FF,$2D,$BF,$AD
	.dc.b	$F9,$3F,$DA,$7C,$84,$8C,$10,$E0,$E5,$E5,$1F,$AB,$F9,$C7,$CB,$D2
	.dc.b	$4F,$FB,$8B,$ED,$3F,$15,$D7,$F0,$EB,$73,$FC,$4D,$AF,$F1,$76,$BF
	.dc.b	$EB,$76,$BC,$6B,$F9,$0F,$C2,$3E,$7D,$F1,$B8,$6D,$7F,$BC,$AF,$AF
	.dc.b	$17,$CE,$D5,$F2,$7F,$C4,$FE,$52,$5F,$F0,$F8,$FE,$2C,$61,$F9,$11
	.dc.b	$87,$F8,$A7,$37,$F1,$B2,$3F,$C5,$FE,$52,$BF,$4F,$E0,$7F,$38,$63
	.dc.b	$1F,$17,$CD,$FF,$23,$F4,$E5,$7F,$1F,$8F,$E4,$C6,$1F,$B1,$18,$74
	.dc.b	$7F,$25,$75,$3D,$23,$F9,$47,$CA,$7F,$2F,$F9,$59,$18,$21,$F8,$37
	.dc.b	$2F,$E1,$63,$F9,$B7,$CC,$7F,$37,$F6,$24,$FF,$33,$8C,$BF,$CD,$63
	.dc.b	$2F,$E3,$63,$FA,$97,$E0,$FF,$43,$9B,$F9,$9F,$72,$31,$48,$C7,$3F
	.dc.b	$F4,$58,$E2,$FF,$55,$9F,$C0,$C6,$DB,$3C,$23,$FA,$F3,$5F,$E0,$50
	.dc.b	$F9,$CF,$19,$D2,$3F,$09,$CF,$FC,$F3,$9B,$F0,$F1,$F9,$18,$C1,$E1
	.dc.b	$F8,$29,$1D,$AF,$FD,$88,$D8,$F3,$7C,$BF,$F1,$9F,$89,$93,$FD,$76
	.dc.b	$93,$F1,$2F,$F8,$F7,$C4,$4A,$D4,$7C,$8F,$96,$F8,$04,$8F,$03,$F0
	.dc.b	$FE,$5B,$F9,$37,$A9,$27,$B4,$7D,$0F,$96,$FA,$3F,$03,$27,$C5,$1F
	.dc.b	$53,$E5,$7E,$61,$23,$D8,$FC,$FF,$5F,$EC,$FC,$B4,$9F,$54,$7D,$CF
	.dc.b	$83,$FB,$BE,$74,$9F,$74,$7F,$84,$7E,$5D,$F2,$9F,$E0,$D6,$CF,$D6
	.dc.b	$6B,$FC,$94,$FC,$13,$87,$F9,$47,$3A,$1F,$25,$D4,$93,$FE,$09,$87
	.dc.b	$CA,$78,$7D,$7F,$17,$AB,$2B,$51,$E5,$F8,$0F,$21,$23,$C8,$F9,$FA
	.dc.b	$7D,$44,$8E,$DE,$37,$B7,$97,$DD,$F1,$D2,$7C,$51,$5F,$66,$BE,$99
	.dc.b	$F8,$44,$8F,$85,$F4,$1F,$1C,$8F,$AE,$3E,$57,$C7,$48,$F7,$3F,$37
	.dc.b	$E3,$F1,$1D,$B7,$4F,$CF,$24,$7C,$F7,$EA,$7C,$24,$8E,$67,$C9,$F8
	.dc.b	$9F,$B0,$47,$E6,$7C,$55,$7B,$47,$DB,$23,$F6,$FC,$6F,$C0,$7C,$4C
	.dc.b	$8F,$A9,$1F,$83,$30,$FF,$75,$C5,$39,$8F,$F1,$BF,$99,$FE,$4F,$F9
	.dc.b	$CF,$F2,$DF,$C7,$7F,$8E,$BE,$43,$FC,$3F,$49,$CB,$FF,$00,$3B,$7F
	.dc.b	$B7,$46,$27,$6F,$E2,$27,$94,$EB,$4F,$F6,$14,$71,$3F,$C8,$F3,$C6
	.dc.b	$75,$7C,$27,$3D,$5F,$DE,$F0,$EB,$87,$26,$D6,$EB,$CD,$39,$BF,$2D
	.dc.b	$F1,$7A,$CF,$84,$DD,$71,$C3,$C5,$1E,$03,$17,$CD,$F3,$B5,$E2,$73
	.dc.b	$8F,$80,$D8,$D4,$74,$8D,$E1,$F1,$B0,$F2,$47,$4B,$EE,$47,$A4,$69
	.dc.b	$EE,$7E,$E6,$BE,$9B,$6F,$E6,$B6,$BF,$D2,$1F,$1E,$F0,$DE,$CE,$1B
	.dc.b	$5F,$E9,$79,$D4,$E4,$FF,$4C,$61,$FE,$A5,$F3,$AF,$09,$A4,$BF,$F5
	.dc.b	$04,$ED,$F8,$93,$B7,$E8,$7A,$0F,$6D,$C7,$F4,$6B,$5B,$FE,$29,$BC
	.dc.b	$1F,$8E,$6E,$1F,$E8,$EC,$39,$5F,$1C,$A4,$F8,$1F,$C2,$4E,$A3,$04
	.dc.b	$C3,$B4,$EB,$1F,$1C,$E7,$1F,$04,$F4,$AF,$D1,$DF,$37,$8D,$6C,$FC
	.dc.b	$DE,$3E,$65,$3B,$BE,$71,$B3,$F5,$93,$B9,$D4,$72,$BF,$D1,$74,$FF
	.dc.b	$9B,$39,$FC,$BC,$7F,$E5,$8F,$75,$D3,$8D,$BF,$BF,$C2,$B1,$8F,$F9
	.dc.b	$FE,$AE,$1C,$EF,$FC,$AD,$58,$B7,$55,$8C,$75,$4A,$FF,$2E,$75,$5C
	.dc.b	$BF,$CC,$13,$73,$F2,$4D,$AF,$1F,$CE,$75,$E8,$EC,$D8,$FF,$33,$38
	.dc.b	$7F,$CD,$1C,$17,$1F,$F3,$64,$DC,$3E,$04,$DA,$F7,$F5,$7A,$4F,$53
	.dc.b	$4F,$F3,$91,$83,$FD,$C1,$F3,$8F,$89,$C6,$3F,$E9,$98,$C4,$EA,$7E
	.dc.b	$3C,$7F,$BA,$A3,$07,$2F,$87,$C9,$AF,$88,$B1,$DB,$E1,$F8,$9C,$BE
	.dc.b	$22,$C7,$8B,$E1,$7A,$9D,$BE,$1E,$C7,$9B,$E1,$BC,$27,$8B,$E1,$EC
	.dc.b	$7A,$BE,$0F,$CA,$79,$BB,$EC,$7B,$BE,$0B,$D3,$F0,$9E,$A6,$C7,$C2
	.dc.b	$F8,$2F,$6F,$C2,$7B,$9B,$1F,$1B,$E0,$FE,$09,$F0,$BE,$12,$C7,$CA
	.dc.b	$F8,$4F,$8B,$E2,$7C,$66,$C7,$CE,$E7,$F9,$27,$CA,$F1,$2C,$7F,$50
	.dc.b	$3F,$CA,$BE,$87,$FA,$69,$C4,$FF,$95,$63,$0F,$FB,$35,$FF,$26,$46
	.dc.b	$BC,$44,$6B,$B7,$FD,$76,$BE,$07,$F3,$ED,$AE,$0F,$FB,$D9,$B1,$AF
	.dc.b	$80,$91,$AF,$96,$B8,$F8,$7E,$03,$99,$D3,$E5,$AD,$7A,$25,$F3,$2B
	.dc.b	$FD,$8C,$61,$FD,$9F,$9E,$F4,$9F,$2B,$AB,$E1,$87,$9B,$E6,$BD,$3E
	.dc.b	$B7,$A9,$B1,$92,$65,$F1,$7E,$13,$51,$8E,$1F,$87,$5F,$0A,$73,$5F
	.dc.b	$2F,$E1,$3E,$2F,$AF,$F1,$9B,$87,$C6,$6C,$EE,$3E,$4C,$FC,$CC,$FE
	.dc.b	$6F,$C2,$59,$F9,$AD,$AF,$B0,$DC,$BE,$AB,$F5,$3E,$0F,$EA,$FC,$25
	.dc.b	$BF,$AA,$DB,$FC,$03,$6F,$EE,$B6,$7F,$12,$D9,$FC,$43,$F8,$0F,$84
	.dc.b	$93,$F8,$0F,$62,$D7,$E2,$E6,$DF,$E2,$26,$D7,$E5,$38,$7F,$13,$F0
	.dc.b	$9F,$8A,$6C,$7E,$29,$B8,$7E,$55,$B7,$FD,$2A,$3F,$07,$C1,$F8,$0F
	.dc.b	$9B,$FF,$23,$DF,$E1,$F8,$D3,$9F,$5B,$F0,$4C,$9D,$71,$5C,$18,$F5
	.dc.b	$C3,$8D,$73,$BF,$C5,$5E,$EB,$FA,$DF,$63,$E5,$BF,$CE,$F7,$FA,$97
	.dc.b	$FD,$0E,$DF,$85,$FE,$8B,$18,$CB,$BC,$FF,$B2,$A3,$8C,$60,$FF,$65
	.dc.b	$3B,$BE,$EE,$9B,$F0,$DD,$3E,$6A,$5F,$87,$E0,$AE,$1C,$F2,$F3,$56
	.dc.b	$30,$E7,$56,$D7,$3E,$6D,$C7,$E8,$F8,$7F,$D7,$C6,$2B,$DD,$18,$99
	.dc.b	$1F,$C3,$B0,$7A,$8E,$5C,$E3,$ED,$18,$F3,$70,$DA,$FE,$E7,$5C,$6F
	.dc.b	$F8,$14,$E5,$F9,$1F,$0F,$F3,$0F,$9B,$F1,$16,$3E,$94,$5C,$3E,$17
	.dc.b	$02,$B1,$8F,$CC,$F1,$BE,$A9,$F5,$BE,$22,$E3,$F6,$E2,$E1,$CA,$6D
	.dc.b	$7E,$0B,$E1,$FF,$86,$CF,$CF,$6C,$7D,$8F,$89,$FF,$02,$C6,$30,$F9
	.dc.b	$FC,$68,$C5,$A3,$1E,$2F,$E2,$2F,$8B,$FC,$83,$E6,$DE,$DF,$E8,$EF
	.dc.b	$EC,$88,$CE,$3F,$ED,$38,$FB,$B1,$87,$0D,$7F,$DB,$0E,$72,$FD,$73
	.dc.b	$67,$FD,$37,$F4,$8E,$7F,$16,$1F,$CC,$BE,$73,$FD,$43,$5F,$CD,$DB
	.dc.b	$8F,$E6,$F1,$FE,$91,$EC,$EF,$E0,$65,$ED,$FF,$15,$CE,$A3,$07,$EA
	.dc.b	$FE,$33,$F5,$27,$EA,$BE,$2A,$C6,$49,$87,$DE,$8C,$BF,$80,$F8,$8E
	.dc.b	$13,$FB,$56,$C6,$63,$07,$EE,$7E,$27,$F6,$F3,$FB,$8F,$12,$4F,$EF
	.dc.b	$7E,$26,$DF,$E3,$26,$E5,$FB,$69,$B5,$F4,$7E,$2F,$F5,$BF,$38,$F7
	.dc.b	$3E,$BD,$FF,$56,$AC,$C6,$F1,$56,$DF,$F6,$5F,$9F,$FE,$6A,$8C,$47
	.dc.b	$FA,$E4,$DC,$F4,$9D,$3C,$E7,$77,$DE,$B6,$7E,$20,$61,$E2,$7B,$E1
	.dc.b	$5E,$D5,$CB,$0E,$B1,$C6,$B5,$7D,$6B,$CB,$3F,$ED,$F3,$B9,$FC,$BA
	.dc.b	$B8,$BD,$30,$F4,$61,$DE,$38,$57,$78,$C4,$79,$8E,$77,$DE,$BD,$C7
	.dc.b	$28,$EE,$72,$71,$5E,$48,$E4,$F5,$76,$62,$36,$74,$7E,$19,$AC,$3D
	.dc.b	$F8,$7A,$A3,$B1,$AA,$ED,$3A,$C3,$D2,$D9,$E4,$7A,$AB,$93,$E1,$39
	.dc.b	$5F,$85,$F9,$15,$C9,$E0,$DF,$2F,$F9,$74,$7D,$33,$FC,$45,$E1,$7E
	.dc.b	$2E,$33,$1F,$87,$6F,$5B,$C8,$FE,$C3,$3C,$0F,$E8,$71,$9D,$3F,$A2
	.dc.b	$D7,$0A,$DB,$63,$37,$FC,$4A,$B8,$5F,$E0,$31,$7B,$3F,$C5,$B5,$FC
	.dc.b	$CC,$62,$7F,$1D,$5C,$27,$F3,$AD,$9F,$E3,$A6,$E7,$A5,$73,$8F,$06
	.dc.b	$5E,$AB,$E2,$D8,$E3,$3B,$AF,$EB,$F1,$CE,$B7,$8F,$8A,$F8,$56,$AF
	.dc.b	$83,$FB,$96,$E1,$E5,$8F,$8F,$3C,$1D,$1F,$A9,$C3,$B1,$E6,$BE,$C6
	.dc.b	$6F,$AC,$6A,$79,$CE,$9F,$79,$B3,$18,$7A,$1C,$91,$C1,$27,$E0,$C7
	.dc.b	$13,$F6,$78,$7A,$5F,$EA,$58,$7A,$67,$C5,$1E,$A3,$71,$ED,$9E,$13
	.dc.b	$A3,$6F,$33,$38,$7B,$0F,$8F,$3F,$33,$B7,$FC,$51,$F1,$3F,$C8,$B5
	.dc.b	$FF,$11,$37,$1F,$F1,$0B,$67,$FC,$5D,$F1,$77,$C3,$4D,$25,$7F,$8C
	.dc.b	$67,$5D,$12,$3F,$C6,$8F,$1F,$8B,$C4,$7F,$81,$31,$E8,$63,$7F,$F1
	.dc.b	$EF,$C9,$3E,$73,$C4,$DA,$FF,$53,$35,$E4,$62,$36,$47,$FC,$89,$5E
	.dc.b	$D5,$F3,$6E,$7C,$36,$3F,$D4,$0D,$7F,$C5,$1F,$38,$F2,$77,$52,$76
	.dc.b	$F5,$9D,$47,$FA,$A7,$E6,$DD,$FE,$16,$C6,$63,$C9,$F3,$9F,$C1,$B5
	.dc.b	$FF,$37,$56,$AB,$FC,$14,$D7,$FA,$43,$5F,$F0,$8B,$5F,$F3,$5F,$CD
	.dc.b	$BE,$73,$EE,$52,$31,$3F,$E4,$76,$BF,$E5,$7D,$DF,$27,$C2,$73,$4D
	.dc.b	$FF,$CA,$EF,$3C,$34,$7D,$6A,$FA,$FF,$25,$31,$FA,$EF,$DA,$F1,$F1
	.dc.b	$1D,$E2,$DF,$E0,$27,$EF,$FA,$32,$BE,$FB,$F8,$2F,$92,$C4,$79,$E2
	.dc.b	$D7,$E2,$27,$EF,$7C,$E3,$E3,$29,$2F,$FC,$CD,$6F,$77,$11,$FE,$65
	.dc.b	$AE,$0F,$FB,$85,$AF,$F8,$AE,$78,$39,$AF,$CA,$FC,$94,$8C,$1F,$96
	.dc.b	$F3,$3F,$32,$47,$F3,$3F,$27,$5F,$A1,$17,$3F,$D0,$CF,$E8,$3E,$47
	.dc.b	$F4,$3F,$27,$2F,$F4,$31,$97,$F4,$71,$97,$ED,$47,$E7,$FE,$61,$F3
	.dc.b	$54,$93,$FE,$FD,$F9,$CF,$F0,$96,$3F,$EB,$DA,$DC,$FF,$C0,$16,$BF
	.dc.b	$E8,$96,$C7,$FB,$0E,$78,$F2,$FF,$98,$A7,$2F,$FA,$73,$E2,$DF,$38
	.dc.b	$F8,$EF,$E1,$D1,$87,$C7,$8F,$E4,$7E,$5B,$F7,$FF,$2F,$23,$F7,$F1
	.dc.b	$FC,$08,$C3,$F9,$78,$C3,$FC,$13,$8B,$F8,$39,$1F,$E0,$FF,$2F,$5F
	.dc.b	$CF,$45,$D7,$F8,$7C,$FF,$0E,$F9,$6F,$E1,$FE,$84,$BF,$E1,$F1,$97
	.dc.b	$F8,$8C,$65,$FD,$44,$65,$FD,$54,$66,$BF,$8D,$CF,$F1,$AF,$96,$FE
	.dc.b	$37,$F2,$F2,$FF,$8D,$C6,$5F,$E3,$B1,$97,$F6,$91,$FC,$8B,$E5,$FF
	.dc.b	$91,$FA,$52,$7F,$91,$C6,$5F,$E4,$B1,$97,$F7,$91,$FC,$EF,$E6,$FE
	.dc.b	$AC,$7F,$A7,$DB,$E9,$1B,$E2,$FD,$C9,$D7,$3F,$12,$3F,$CC,$8A,$7F
	.dc.b	$B7,$E5,$F3,$9D,$CC,$EB,$F8,$DE,$3F,$CA,$12,$BF,$85,$F3,$7E,$1F
	.dc.b	$E8,$BF,$87,$46,$0F,$37,$CD,$7A,$48,$E7,$4F,$47,$8F,$ED,$F9,$A9
	.dc.b	$39,$D1,$1F,$07,$E1,$BE,$16,$72,$B4,$FC,$5F,$9A,$F8,$DF,$33,$27
	.dc.b	$93,$F2,$7E,$6B,$E5,$7C,$D4,$9E,$CF,$CD,$F9,$5F,$9D,$F3,$5F,$3F
	.dc.b	$DE,$95,$E3,$8F,$A5,$F3,$32,$3C,$CF,$D5,$F8,$4F,$AD,$F2,$7F,$5F
	.dc.b	$E6,$A5,$7A,$E3,$ED,$7C,$2F,$DB,$F4,$64,$FB,$E3,$FB,$C6,$AF,$97
	.dc.b	$A4,$9F,$89,$F0,$FA,$F9,$79,$1F,$2E,$91,$CB,$E6,$39,$FC,$C4,$AD
	.dc.b	$3D,$BE,$47,$BF,$C1,$49,$E4,$F8,$BE,$5F,$C7,$F2,$F2,$7B,$3E,$6F
	.dc.b	$97,$F3,$FC,$1C,$9F,$13,$EA,$F9,$7F,$5F,$CA,$49,$F3,$3B,$5E,$37
	.dc.b	$E2,$3C,$37,$03,$E5,$3E,$19,$2B,$89,$EF,$F9,$3F,$8E,$47,$E1,$7C
	.dc.b	$A5,$7A,$F9,$25,$87,$6A,$D6,$3E,$1C,$58,$63,$0E,$17,$E4,$F8,$AF
	.dc.b	$AA,$67,$1E,$77,$FE,$2C,$7E,$97,$C8,$7D,$A3,$ED,$FC,$85,$8C,$90
	.dc.b	$F2,$9C,$9F,$C0,$E1,$23,$EE,$3F,$84,$F8,$FC,$47,$0C,$7E,$3F,$C8
	.dc.b	$7E,$20,$CE,$9F,$B2,$97,$CE,$B1,$75,$FC,$3F,$95,$78,$1F,$16,$F7
	.dc.b	$5A,$7F,$DB,$B1,$9B,$E5,$7C,$63,$1D,$BC,$70,$E7,$7F,$3E,$F6,$F2
	.dc.b	$31,$87,$59,$DC,$78,$30,$E8,$F1,$BE,$98,$78,$6F,$83,$6B,$75,$C4
	.dc.b	$EC,$7C,$0A,$E2,$F8,$23,$78,$79,$30,$F0,$B6,$FD,$75,$AC,$3B,$BC
	.dc.b	$4E,$B3,$B7,$CC,$6E,$E3,$A9,$E2,$F8,$CC,$DF,$82,$33,$5C,$E7,$47
	.dc.b	$1A,$FA,$8F,$BC,$A7,$BC,$CC,$7C,$E7,$7F,$28,$F2,$1E,$FF,$09,$4E
	.dc.b	$63,$0D,$8E,$AD,$7A,$F0,$BD,$6E,$58,$7F,$A2,$9B,$19,$9F,$2C,$E5
	.dc.b	$B1,$9C,$FE,$7B,$74,$FC,$87,$B0,$F8,$3F,$E9,$35,$AC,$7E,$BD,$EF
	.dc.b	$0F,$C6,$35,$5F,$1E,$4D,$D4,$9F,$AF,$5B,$AD,$46,$27,$57,$DE,$3F
	.dc.b	$13,$3A,$8C,$13,$1F,$D1,$5F,$3C,$7C,$F1,$8C,$39,$B7,$0E,$73,$A7
	.dc.b	$D3,$3A,$69,$AC,$71,$C6,$FC,$23,$FB,$9C,$7F,$67,$FA,$2F,$A4,$FF
	.dc.b	$69,$8C,$D7,$F0,$39,$CB,$FC,$16,$33,$C9,$FC,$B6,$72,$FF,$2F,$8C
	.dc.b	$D7,$F1,$E9,$CB,$FC,$0F,$F0,$97,$6F,$F7,$CB,$E5,$F2,$D6,$B5,$37
	.dc.b	$1F,$E5,$B8,$7F,$B0,$67,$FC,$05,$0C,$39,$C7,$E4,$DB,$1F,$9C,$6B
	.dc.b	$FA,$06,$BF,$CC,$B1,$7A,$9D,$F3,$F8,$DE,$BD,$9C,$E1,$F1,$63,$15
	.dc.b	$96,$C6,$4F,$95,$18,$BF,$91,$EC,$5D,$7E,$37,$AF,$DF,$4F,$F5,$6E
	.dc.b	$1F,$46,$31,$F2,$3E,$1F,$5F,$C5,$16,$74,$7D,$78,$C6,$3F,$57,$D8
	.dc.b	$B8,$FD,$3F,$63,$B7,$C8,$7A,$DB,$87,$DF,$F6,$7D,$DE,$27,$BD,$B1
	.dc.b	$F8,$3F,$67,$0D,$38,$1F,$87,$F5,$FF,$A2,$67,$F2,$DA,$FF,$64,$E3
	.dc.b	$46,$39,$3F,$24,$D8,$7B,$1F,$0F,$E4,$7E,$83,$70,$FC,$A7,$66,$C8
	.dc.b	$FF,$18,$C3,$F3,$91,$CF,$D7,$FB,$07,$D9,$F6,$2D,$7E,$77,$C6,$FA
	.dc.b	$1F,$23,$F7,$5B,$1F,$E4,$9F,$AA,$7B,$3F,$83,$6C,$67,$DA,$C4,$E7
	.dc.b	$D9,$D7,$9B,$6B,$93,$C7,$DA,$FB,$DF,$23,$67,$F6,$0F,$5F,$6F,$F0
	.dc.b	$5F,$23,$67,$F6,$8F,$87,$DB,$F1,$7B,$76,$7C,$CF,$E9,$7A,$1F,$3D
	.dc.b	$F9,$67,$FC,$02,$63,$C7,$D4,$67,$C0,$FE,$F4,$A3,$DD,$EE,$7E,$4F
	.dc.b	$E4,$6E,$3F,$C1,$5E,$3E,$1F,$EC,$FD,$47,$CF,$B5,$CD,$FF,$9D,$67
	.dc.b	$31,$8D,$3F,$C2,$71,$89,$FD,$CE,$7F,$EC,$17,$85,$F3,$73,$87,$E2
	.dc.b	$AF,$DA,$DA,$E1,$1F,$D6,$1A,$FF,$83,$FB,$1C,$FF,$E3,$FB,$FB,$9E
	.dc.b	$EF,$DB,$F3,$BE,$E3,$23,$27,$E0,$63,$17,$FE,$23,$A7,$B7,$F8,$33
	.dc.b	$F0,$9E,$E5,$BF,$C4,$45,$AE,$13,$78,$BF,$CA,$57,$FE,$27,$BF,$C7
	.dc.b	$7B,$5F,$8D,$F7,$7F,$1C,$C8,$C9,$F9,$38,$C6,$1E,$39,$B7,$D2,$6D
	.dc.b	$7F,$93,$2D,$EE,$7E,$67,$D7,$FC,$D3,$67,$F3,$CF,$EC,$E7,$71,$FA
	.dc.b	$B6,$E3,$FE,$06,$DB,$E5,$D7,$EE,$7E,$4B,$F9,$94,$FF,$91,$DB,$97
	.dc.b	$BB,$0F,$E5,$51,$87,$FC,$BE,$D9,$FF,$33,$B6,$FE,$BC,$F0,$F5,$BF
	.dc.b	$2F,$E3,$FE,$63,$E0,$AC,$FE,$21,$B8,$7E,$7B,$8D,$18,$CB,$F3,$AE
	.dc.b	$A7,$FC,$3F,$3F,$C3,$AB,$FC,$30,$D8,$FC,$55,$FB,$7D,$0F,$49,$EA
	.dc.b	$F4,$ED,$FF,$B3,$74,$4E,$2F,$E1,$55,$CB,$FD,$21,$4F,$03,$E1,$F9
	.dc.b	$7F,$11,$B8,$F3,$7E,$61,$1F,$E3,$93,$C5,$F3,$F9,$3F,$B1,$FC,$27
	.dc.b	$E5,$65,$FC,$02,$31,$BF,$EF,$9B,$5F,$AE,$6C,$FC,$77,$F5,$44,$8E
	.dc.b	$A7,$83,$E5,$3E,$CF,$87,$F6,$9B,$1F,$05,$FC,$11,$1E,$07,$A3,$E4
	.dc.b	$3F,$03,$97,$E0,$9B,$1F,$25,$FC,$51,$1E,$07,$F9,$2F,$E7,$FE,$6B
	.dc.b	$FE,$FF,$DB,$F8,$7F,$AD,$FC,$7E,$73,$9F,$FA,$7A,$73,$5F,$E0,$49
	.dc.b	$CC,$FF,$97,$9A,$FF,$A4,$DA,$FF,$AE,$5B,$CA,$FC,$29,$F2,$1F,$9A
	.dc.b	$F9,$3F,$F0,$24,$63,$0F,$C1,$46,$27,$51,$FE,$C3,$C7,$FD,$A1,$5B
	.dc.b	$C7,$F4,$D7,$FC,$DF,$DF,$FF,$23,$D7,$F8,$87,$DD,$7C,$7E,$7E,$07
	.dc.b	$FC,$41,$29,$CE,$5F,$E2,$0B,$FF,$27,$E3,$FE,$7A,$6B,$FE,$AF,$6B
	.dc.b	$FE,$D5,$6E,$8E,$07,$C9,$7F,$84,$78,$BF,$6E,$DC,$3F,$3C,$FE,$FC
	.dc.b	$8F,$E5,$1F,$E0,$5B,$F0,$8F,$D2,$37,$1F,$E0,$8D,$CF,$F4,$6D,$9F
	.dc.b	$E0,$7F,$23,$FE,$21,$C7,$F8,$3B,$63,$F5,$CF,$F0,$F2,$3F,$A4,$7F
	.dc.b	$A2,$5F,$FB,$3C,$7C,$F7,$D8,$7E,$97,$B8,$AF,$F7,$B4,$E6,$7F,$D9
	.dc.b	$87,$0C,$3F,$DA,$D8,$FF,$85,$AB,$71,$FE,$A4,$F8,$5B,$5D,$26,$BC
	.dc.b	$AF,$A5,$F1,$8E,$87,$FA,$BA,$8B,$DB,$DE,$3E,$58,$C6,$3D,$DA,$77
	.dc.b	$29,$DD,$F3,$C7,$1C,$3D,$93,$F0,$DE,$B8,$F9,$A3,$15,$C2,$31,$0C
	.dc.b	$3D,$78,$FA,$6F,$9C,$E4,$F5,$88,$D5,$F9,$E7,$BC,$E8,$EE,$7A,$5E
	.dc.b	$75,$DB,$0F,$45,$71,$3C,$95,$C4,$F0,$57,$43,$8B,$E2,$BF,$0B,$F3
	.dc.b	$4C,$57,$47,$13,$C6,$BB,$47,$8D,$AF,$AE,$78,$3F,$34,$D1,$CA,$3C
	.dc.b	$B3,$E2,$3A,$47,$87,$DA,$FB,$8D,$CB,$ED,$46,$23,$F1,$ED,$F8,$87
	.dc.b	$BF,$F2,$E5,$39,$8C,$37,$2F,$C4,$B6,$77,$7F,$8C,$7F,$DE,$35,$CF
	.dc.b	$0F,$CF,$63,$FC,$32,$7F,$2E,$D7,$ED,$E3,$E2,$C7,$F3,$6D,$8F,$BF
	.dc.b	$3B,$AE,$7C,$BD,$DF,$D5,$56,$DB,$87,$86,$BC,$87,$12,$D7,$17,$D1
	.dc.b	$5C,$5B,$3C,$6F,$F4,$F1,$C5,$B5,$C5,$FD,$5D,$71,$6D,$71,$7F,$9C
	.dc.b	$CE,$DA,$7B,$A3,$53,$C6,$32,$7C,$09,$E1,$19,$35,$5E,$39,$DD,$FC
	.dc.b	$0C,$3A,$1F,$3A,$FC,$15,$BB,$E5,$87,$96,$39,$19,$7C,$67,$E0,$F1
	.dc.b	$F9,$F8,$FF,$08,$6B,$FC,$0A,$BE,$B4,$78,$63,$D0,$78,$4F,$19,$F7
	.dc.b	$A7,$C1,$7E,$43,$DA,$FC,$FE,$2F,$E8,$91,$F8,$B9,$FF,$35,$70,$7E
	.dc.b	$1D,$B1,$97,$FC,$76,$D7,$FC,$76,$FF,$2A,$DF,$FE,$00,$72,$D9,$FF
	.dc.b	$7C,$B6,$FB,$C7,$F8,$AC,$C7,$E1,$6F,$91,$A6,$F1,$F2,$6D,$FE,$2C
	.dc.b	$AC,$9F,$DB,$F1,$FE,$5B,$83,$0E,$53,$D4,$D0,$8D,$61,$E9,$6C,$7F
	.dc.b	$B2,$E7,$8B,$FE,$61,$6B,$FE,$8D,$73,$7F,$CE,$65,$F0,$CE,$DF,$73
	.dc.b	$6F,$C3,$3D,$9F,$F5,$AF,$CF,$3D,$DF,$95,$08,$C3,$F3,$5A,$FD,$16
	.dc.b	$BF,$5D,$CE,$1F,$D2,$E1,$F3,$CF,$47,$E9,$46,$1B,$5C,$35,$F9,$3F
	.dc.b	$6E,$F6,$FE,$9F,$CA,$CC,$3D,$33,$96,$CE,$5F,$E8,$95,$F4,$B7,$5F
	.dc.b	$DF,$BE,$15,$FE,$D4,$F9,$EA,$FE,$3D,$AF,$F1,$B6,$BF,$DA,$1A,$9E
	.dc.b	$7A,$7B,$FE,$7D,$F0,$FE,$E6,$D7,$E0,$0E,$19,$FE,$3B,$E3,$7F,$01
	.dc.b	$99,$A4,$9F,$C1,$3F,$94,$F1,$24,$7D,$A9,$E0,$DF,$73,$99,$1F,$C5
	.dc.b	$35,$F8,$A7,$E4,$61,$E4,$FE,$80,$FD,$0C,$FD,$6F,$9D,$FE,$E3,$5F
	.dc.b	$E0,$86,$A6,$93,$BF,$F5,$DE,$FB,$E7,$3E,$8D,$7F,$BC,$DA,$F8,$8E
	.dc.b	$53,$DC,$E3,$1A,$DF,$FC,$29,$5C,$F8,$B5,$7F,$07,$E3,$3E,$49,$1F
	.dc.b	$D4,$35,$F7,$F4,$FE,$CF,$E3,$BF,$7A,$7E,$FA,$7F,$7F,$E0,$CA,$F9
	.dc.b	$8F,$F0,$2F,$93,$C4,$72,$E0,$FE,$CE,$E8,$FA,$51,$F1,$FE,$4B,$F5
	.dc.b	$3B,$9A,$49,$FE,$2D,$19,$7F,$08,$BC,$DF,$A0,$73,$8C,$CE,$1D,$C6
	.dc.b	$5F,$27,$93,$FC,$60,$8F,$E1,$7D,$BA,$E8,$FC,$44,$7F,$1E,$8F,$E3
	.dc.b	$FD,$3F,$92,$24,$FF,$11,$6B,$F0,$7E,$5F,$F9,$29,$C1,$FC,$97,$91
	.dc.b	$FC,$A0,$8F,$E5,$7D,$4A,$FC,$98,$BC,$9F,$C5,$F6,$56,$3B,$FF,$71
	.dc.b	$56,$FF,$2D,$3F,$CD,$FE,$A1,$F1,$F4,$93,$FA,$3F,$3A,$BF,$80,$9D
	.dc.b	$BF,$C8,$1A,$FD,$B8,$E9,$E3,$48,$D3,$DF,$E3,$F1,$1C,$E2,$E9,$FC
	.dc.b	$BE,$D5,$FC,$73,$E7,$BF,$67,$7F,$D5,$5A,$FF,$5B,$6B,$FE,$27,$6B
	.dc.b	$FE,$74,$78,$D7,$E8,$3B,$1C,$3F,$A4,$3B,$78,$32,$FF,$59,$19,$7F
	.dc.b	$87,$5F,$2D,$FF,$A9,$E4,$F8,$FF,$E8,$7E,$B4,$BF,$89,$1E,$9F,$07
	.dc.b	$D5,$DB,$27,$9C,$7D,$1F,$8F,$FA,$5F,$33,$27,$E6,$47,$D5,$F8,$DF
	.dc.b	$E9,$FE,$C4,$9F,$A1,$1F,$67,$E6,$7F,$A9,$FC,$F4,$9F,$A9,$1F,$77
	.dc.b	$E3,$7E,$F7,$97,$27,$EC,$47,$E0,$7D,$09,$1F,$6D,$FC,$27,$C7,$7E
	.dc.b	$17,$E3,$BF,$AF,$F9,$92,$BF,$01,$1F,$89,$F8,$E9,$1F,$83,$7F,$19
	.dc.b	$F1,$DF,$8D,$F9,$AF,$EE,$5C,$C6,$92,$BF,$C2,$1E,$6A,$5F,$1B,$FD
	.dc.b	$BB,$E7,$A5,$7E,$3E,$3F,$C1,$DF,$86,$3E,$23,$FB,$F3,$67,$9B,$5F
	.dc.b	$F1,$06,$08,$C4,$7F,$9B,$63,$1F,$1B,$D3,$E3,$24,$FF,$7C,$A4,$E2
	.dc.b	$3F,$90,$C3,$E5,$FC,$66,$7C,$4F,$F0,$44,$3E,$37,$FC,$0F,$F1,$12
	.dc.b	$FA,$47,$AB,$E5,$24,$77,$7D,$BF,$1B,$88,$EF,$F1,$7F,$00,$C6,$3D
	.dc.b	$9F,$F0,$9F,$AB,$27,$D1,$1F,$1B,$E1,$24,$7A,$DF,$93,$F0,$D8,$8F
	.dc.b	$5F,$C3,$FC,$C3,$18,$7A,$AB,$13,$F0,$E7,$3E,$4F,$FA,$57,$E6,$DE
	.dc.b	$37,$F8,$79,$B5,$FE,$49,$6B,$FE,$4D,$6B,$FE,$53,$73,$87,$F8,$BF
	.dc.b	$E1,$FF,$D3,$4D,$F8,$9F,$F0,$EC,$EB,$0F,$F2,$1F,$17,$FB,$D1,$BE
	.dc.b	$2F,$E1,$0F,$C2,$F8,$13,$2F,$B5,$19,$7E,$E4,$7F,$8A,$3E,$71,$E0
	.dc.b	$7F,$AD,$A7,$FD,$D1,$ED,$7F,$AC,$67,$3E,$AF,$FB,$FB,$E3,$7F,$C2
	.dc.b	$34,$9C,$BC,$01,$9F,$5B,$FD,$75,$3C,$27,$6D,$C3,$C0,$DB,$ED,$87
	.dc.b	$86,$F8,$BE,$3A,$DB,$67,$78,$78,$30,$EB,$3C,$63,$27,$9E,$B8,$BB
	.dc.b	$6B,$9A,$DD,$7F,$95,$30,$F2,$4F,$B4,$D1,$63,$BB,$E1,$C3,$CE,$F5
	.dc.b	$AE,$77,$DA,$B7,$3F,$14,$CC,$C7,$A4,$EE,$B9,$C7,$80,$FF,$4D,$57
	.dc.b	$D0,$6C,$FD,$96,$D7,$FA,$67,$E3,$1F,$00,$F8,$4C,$36,$BF,$D4,$13
	.dc.b	$A9,$C9,$FE,$A1,$C1,$F2,$BF,$E0,$BA,$FF,$53,$46,$3E,$07,$53,$99
	.dc.b	$F2,$B5,$FC,$53,$C7,$4F,$C1,$37,$5F,$F7,$F7,$5B,$A1,$C5,$F4,$AB
	.dc.b	$58,$7D,$D6,$BF,$91,$6D,$71,$9E,$18,$CA,$FA,$D5,$AB,$EB,$18,$9D
	.dc.b	$63,$E4,$7F,$0F,$87,$39,$D1,$27,$4F,$E0,$E3,$51,$82,$63,$F0,$2B
	.dc.b	$9D,$F8,$27,$37,$EB,$9F,$0D,$7C,$2A,$D3,$C6,$7F,$59,$E1,$BE,$6B
	.dc.b	$FD,$95,$9F,$FA,$C3,$A7,$FC,$7C,$FF,$AC,$A3,$35,$FE,$56,$9C,$BF
	.dc.b	$E5,$88,$CF,$4F,$FA,$F5,$B5,$FE,$CE,$AF,$DC,$BC,$FE,$3B,$E3,$D7
	.dc.b	$FA,$C7,$A5,$E7,$FF,$AC,$F1,$FE,$44,$D7,$F8,$0C,$6A,$FF,$91,$4E
	.dc.b	$65,$39,$7F,$CF,$0D,$AF,$27,$C3,$F8,$2F,$A9,$ED,$F7,$31,$7F,$ED
	.dc.b	$FC,$DF,$31,$8D,$7E,$35,$59,$CD,$F1,$3F,$B7,$BF,$EA,$56,$C7,$A6
	.dc.b	$78,$C6,$0E,$17,$F2,$7E,$2B,$A5,$75,$F9,$C9,$5F,$47,$E2,$ED,$FC
	.dc.b	$F7,$A6,$3F,$EA,$49,$F0,$C6,$0C,$FB,$7E,$77,$D0,$D8,$FB,$9E,$E7
	.dc.b	$AC,$F6,$7C,$E5,$8F,$C0,$7C,$67,$62,$3E,$E7,$C3,$F1,$9F,$0C,$F8
	.dc.b	$9D,$F6,$7F,$11,$F1,$7F,$1C,$F9,$1F,$09,$63,$F1,$9F,$2F,$EA,$23
	.dc.b	$F2,$9F,$6F,$57,$CF,$3E,$87,$C6,$59,$FC,$A7,$C5,$FD,$33,$EA,$7C
	.dc.b	$5D,$8F,$CC,$78,$DF,$18,$8F,$D6,$7E,$A7,$CC,$FF,$99,$23,$FA,$6F
	.dc.b	$5F,$EE,$BA,$53,$9E,$6F,$F5,$AB,$FE,$7C,$AD,$9F,$D5,$27,$7F,$17
	.dc.b	$F8,$79,$7C,$E5,$F2,$F0,$DB,$E3,$83,$0E,$FF,$13,$E4,$24,$FF,$8E
	.dc.b	$B0,$FF,$32,$37,$E2,$7B,$1D,$FD,$2B,$1D,$2F,$D3,$F1,$1E,$3F,$9C
	.dc.b	$F2,$36,$7F,$DF,$DF,$37,$F9,$7C,$FF,$B6,$91,$C3,$80,$E1,$D7,$83
	.dc.b	$F2,$BE,$27,$E0,$1F,$07,$E2,$AD,$F6,$F0,$3F,$5B,$E8,$BE,$71,$9F
	.dc.b	$EC,$32,$EB,$5A,$9C,$DF,$13,$C5,$3F,$E0,$2C,$D5,$FC,$05,$B3,$A9
	.dc.b	$E6,$D8,$CC,$7F,$BE,$DB,$CB,$FE,$F8,$E6,$1E,$38,$CD,$7D,$DE,$27
	.dc.b	$C5,$7D,$D3,$EF,$79,$76,$7E,$84,$E9,$B3,$A6,$9A,$F8,$AF,$C3,$1F
	.dc.b	$87,$F8,$BB,$19,$26,$7F,$90,$8C,$BF,$92,$F5,$FF,$1C,$7E,$3F,$E2
	.dc.b	$ED,$78,$A2,$CF,$E6,$26,$EB,$F9,$DC,$BF,$C6,$B7,$B3,$D3,$3A,$7E
	.dc.b	$07,$C4,$7E,$6F,$D7,$FC,$E3,$63,$DE,$E1,$F8,$3F,$17,$71,$63,$FA
	.dc.b	$BF,$80,$FF,$46,$57,$E5,$5B,$3F,$AC,$8F,$E5,$D1,$89,$E1,$F1,$38
	.dc.b	$3F,$59,$34,$FC,$CF,$8D,$F5,$FC,$7F,$B0,$D8,$C9,$F8,$17,$F0,$C4
	.dc.b	$70,$6B,$0F,$3B,$FE,$4E,$E8,$FD,$A7,$47,$48,$F4,$7C,$47,$98,$F3
	.dc.b	$FC,$E5,$8C,$92,$F9,$9F,$BD,$D1,$3B,$8C,$F7,$A7,$FD,$39,$5F,$17
	.dc.b	$E2,$BE,$1F,$C4,$7C,$46,$C7,$C9,$8F,$8F,$F1,$58,$C3,$F3,$9E,$4F
	.dc.b	$EB,$BE,$73,$E5,$5F,$EE,$E3,$1C,$3F,$9D,$7F,$50,$48,$FD,$33,$6B
	.dc.b	$9C,$7F,$16,$DF,$F8,$B7,$4F,$18,$FB,$5F,$11,$F6,$3E,$27,$EC,$B6
	.dc.b	$32,$4B,$F7,$EC,$F8,$AD,$CF,$E0,$FE,$2B,$F0,$33,$F8,$26,$C6,$49
	.dc.b	$5C,$BE,$07,$51,$FC,$AA,$1F,$39,$FE,$CC,$8F,$E0,$4D,$8F,$E1,$2D
	.dc.b	$7F,$8E,$35,$FE,$42,$EF,$A7,$FC,$77,$1D,$39,$BF,$C0,$B0,$FA,$4F
	.dc.b	$E7,$B7,$E1,$BE,$0F,$F4,$3F,$9D,$7C,$57,$E7,$3E,$4B,$FC,$15,$18
	.dc.b	$9F,$D1,$46,$26,$47,$F3,$99,$7C,$66,$33,$FC,$58,$F8,$8F,$F2,$53
	.dc.b	$FC,$D5,$AF,$F3,$A6,$BF,$EA,$86,$BC,$75,$FE,$61,$1F,$E5,$3D,$3F
	.dc.b	$C6,$7F,$07,$FE,$3C,$7E,$0C,$67,$1F,$EB,$DC,$0F,$92,$FF,$0E,$36
	.dc.b	$3F,$9A,$F6,$FE,$6E,$72,$48,$FE,$00,$FE,$FB,$5F,$DF,$B6,$32,$7F
	.dc.b	$05,$8C,$B4,$99,$FE,$2B,$4E,$0F,$E3,$3E,$43,$C6,$49,$F1,$E2,$8C
	.dc.b	$4E,$E7,$F8,$97,$C8,$FF,$10,$9F,$EA,$53,$FE,$D3,$1F,$3B,$F6,$9F
	.dc.b	$F0,$2F,$BF,$FE,$7E,$8C,$47,$FB,$48,$DE,$5F,$DA,$B6,$56,$E3,$FA
	.dc.b	$CF,$C6,$FF,$B6,$BD,$EE,$35,$AC,$F1,$5A,$C1,$5A,$C3,$FD,$1A,$F8
	.dc.b	$23,$CD,$8E,$6F,$FD,$CA,$E6,$18,$E3,$0F,$26,$1E,$39,$FE,$F9,$39
	.dc.b	$7D,$B7,$5F,$65,$F5,$7D,$A7,$03,$CB,$7D,$27,$D3,$19,$C3,$81,$E3
	.dc.b	$BE,$98,$FB,$E7,$99,$DA,$F9,$38,$3D,$AE,$30,$E5,$1C,$27,$94,$60
	.dc.b	$F3,$16,$78,$1F,$2E,$FC,$6D,$3B,$BB,$AF,$82,$78,$4F,$92,$77,$C7
	.dc.b	$E4,$3F,$7E,$39,$BE,$F9,$E6,$75,$79,$C7,$A0,$DB,$F8,$7F,$4B,$EE
	.dc.b	$BF,$2D,$AE,$63,$F0,$51,$8F,$90,$7C,$1F,$F6,$CA,$FF,$0B,$D7,$F8
	.dc.b	$E2,$73,$92,$72,$58,$CB,$80,$E2,$FB,$5F,$5B,$FC,$EE,$0F,$92,$FF
	.dc.b	$22,$C6,$1B,$1E,$19,$E0,$DC,$7F,$07,$8F,$F2,$2A,$DB,$E5,$9E,$8E
	.dc.b	$B1,$F2,$E1,$FA,$C7,$CF,$9F,$79,$FC,$A5,$70,$7F,$13,$3A,$8F,$04
	.dc.b	$7E,$50,$ED,$7E,$07,$C4,$77,$C7,$CE,$FC,$57,$E6,$5F,$78,$FC,$AD
	.dc.b	$77,$8F,$24,$EA,$3B,$C6,$C8,$E0,$93,$E2,$3D,$35,$FE,$6E,$9E,$78
	.dc.b	$FE,$BF,$0F,$83,$7F,$E7,$23,$DD,$8F,$8C,$DE,$1E,$D3,$EC,$E3,$E3
	.dc.b	$7B,$1E,$8C,$7B,$57,$8A,$37,$87,$89,$87,$CA,$AE,$4E,$8E,$D3,$E2
	.dc.b	$3C,$95,$F2,$F0,$F3,$CF,$C9,$BE,$78,$FC,$28,$C1,$FC,$23,$8F,$FC
	.dc.b	$3D,$7F,$C9,$76,$FE,$96,$D9,$FF,$13,$36,$7F,$C4,$9F,$21,$78,$BF
	.dc.b	$01,$5F,$E9,$69,$E6,$FF,$3E,$AF,$F4,$04,$62,$3F,$D3,$0F,$F8,$B5
	.dc.b	$E4,$43,$FC,$A5,$CB,$5F,$F1,$9F,$CB,$BE,$43,$C4,$D8,$FF,$2C,$18
	.dc.b	$88,$FF,$AB,$E3,$4F,$F8,$78,$C9,$E2,$72,$FA,$5A,$FF,$B2,$5A,$FE
	.dc.b	$E7,$E7,$1F,$09,$F1,$61,$18,$7F,$C4,$0D,$7F,$C0,$BF,$3A,$F6,$38
	.dc.b	$4E,$67,$FD,$74,$D7,$FD,$7E,$D7,$C1,$5F,$2B,$E7,$5F,$01,$F4,$DB
	.dc.b	$3F,$EC,$D6,$BF,$E5,$A5,$F0,$1B,$9F,$93,$F2,$52,$B8,$47,$AF,$E7
	.dc.b	$78,$F2,$FF,$8B,$9B,$C9,$FE,$D2,$6D,$7F,$1B,$F9,$DF,$E4,$B5,$BB
	.dc.b	$FF,$5F,$4A,$BC,$6F,$FB,$5D,$AF,$38,$FC,$2E,$EF,$84,$FB,$A4,$9F
	.dc.b	$B5,$F2,$95,$FC,$1F,$89,$F8,$23,$18,$B3,$FB,$1F,$09,$F8,$62,$57
	.dc.b	$E0,$FE,$52,$BF,$8C,$F9,$3F,$C5,$98,$BF,$F2,$6F,$CE,$FF,$80,$B0
	.dc.b	$FF,$78,$B6,$BF,$9D,$37,$D4,$FE,$F7,$E1,$FE,$0E,$3F,$BF,$4E,$BC
	.dc.b	$1F,$87,$18,$83,$4E,$5F,$38,$E4,$FD,$86,$08,$C5,$7F,$9A,$EB,$C5
	.dc.b	$DA,$E2,$7C,$9F,$E9,$09,$5F,$A1,$F9,$4A,$FE,$AB,$E5,$7F,$52,$63
	.dc.b	$7F,$91,$0C,$3A,$78,$DF,$AE,$24,$7E,$AF,$E5,$6B,$FB,$48,$B5,$CA
	.dc.b	$BC,$95,$F0,$FC,$59,$7E,$97,$DB,$F3,$EF,$90,$CC,$FF,$65,$3E,$7C
	.dc.b	$FC,$82,$39,$32,$7C,$3F,$17,$E8,$1C,$B1,$F1,$6D,$FE,$3E,$9F,$E0
	.dc.b	$5E,$5F,$F0,$83,$F8,$4F,$C2,$D8,$C9,$0F,$E1,$EF,$F0,$EF,$96,$91
	.dc.b	$FC,$2D,$FE,$23,$F0,$D8,$8F,$E2,$BF,$2D,$FC,$50,$C6,$3E,$2F,$3F
	.dc.b	$F8,$B9,$27,$F8,$A7,$C0,$57,$F8,$EC,$59,$F4,$CD,$EC,$FE,$25,$36
	.dc.b	$BE,$B6,$8A,$FE,$69,$F3,$8F,$9E,$A4,$AF,$E4,$5D,$D5,$FF,$6E,$CE
	.dc.b	$DF,$F0,$A6,$5F,$66,$31,$1F,$CC,$27,$AF,$95,$23,$79,$FF,$86,$F9
	.dc.b	$97,$FC,$7F,$0E,$5F,$31,$BB,$EF,$85,$9C,$C7,$FB,$17,$E7,$BF,$80
	.dc.b	$37,$C7,$FE,$59,$18,$8F,$F7,$BF,$80,$E9,$7C,$26,$AF,$F8,$7D,$EE
	.dc.b	$FF,$0C,$D7,$F1,$8D,$7F,$C6,$4F,$5C,$3F,$43,$87,$EB,$F9,$BF,$63
	.dc.b	$E0,$9E,$D7,$F0,$0F,$D3,$6A,$F9,$8F,$A2,$48,$D6,$FF,$13,$E6,$7D
	.dc.b	$7E,$47,$98,$93,$F9,$B3,$C6,$FF,$66,$B7,$CC,$FD,$CE,$F9,$5D,$27
	.dc.b	$2D,$F8,$EF,$C0,$91,$D7,$CC,$FE,$13,$D1,$E8,$48,$FF,$61,$D7,$F4
	.dc.b	$46,$FC,$BF,$E2,$BD,$09,$1D,$E3,$F1,$BC,$B2,$3C,$4F,$E4,$3E,$5F
	.dc.b	$F2,$3E,$B7,$A4,$93,$D9,$F6,$7C,$BF,$E5,$89,$1E,$BF,$96,$FC,$CF
	.dc.b	$CA,$48,$F7,$3F,$9C,$F9,$6F,$CE,$FC,$B7,$C5,$24,$FA,$9F,$F0,$6F
	.dc.b	$CF,$3E,$4B,$FC,$2C,$D9,$FA,$6D,$7F,$43,$E1,$6B,$E3,$A4,$69,$E3
	.dc.b	$F2,$FC,$BE,$5F,$A9,$27,$FD,$D7,$68,$F0,$FC,$BF,$80,$93,$E0,$F9
	.dc.b	$7F,$1F,$CB,$79,$3E,$56,$4F,$92,$2B,$FE,$24,$C1,$3E,$C2,$47,$83
	.dc.b	$E5,$BD,$A4,$7D,$B1,$F0,$3E,$5A,$47,$C0,$7E,$1F,$C5,$62,$3B,$47
	.dc.b	$C8,$23,$EC,$7E,$4F,$C9,$48,$E4,$7E,$84,$7C,$E7,$F3,$97,$0D,$5E
	.dc.b	$4F,$F4,$C7,$7E,$0F,$D4,$F9,$5F,$F1,$45,$6E,$7F,$9E,$35,$FE,$84
	.dc.b	$D7,$FD,$50,$DF,$19,$AF,$F6,$22,$7C,$17,$E0,$3E,$53,$F0,$2F,$E0
	.dc.b	$BE,$52,$61,$F8,$28,$FF,$1C,$7C,$F3,$B9,$EA,$7F,$AD,$7E,$4D,$EE
	.dc.b	$7F,$53,$AD,$4E,$63,$15,$C3,$0F,$F7,$DC,$6F,$D5,$FE,$B5,$87,$29
	.dc.b	$FF,$60,$61,$C9,$D9,$8B,$E5,$3B,$84,$EE,$30,$FF,$92,$DB,$3C,$2F
	.dc.b	$C7,$87,$96,$FA,$35,$E7,$87,$92,$77,$1E,$23,$BD,$73,$78,$99,$C3
	.dc.b	$C1,$3C,$30,$F0,$E1,$D6,$7C,$D3,$F0,$6B,$17,$CD,$EB,$58,$38,$61
	.dc.b	$E3,$8E,$F8,$74,$7C,$15,$C7,$0F,$0C,$79,$0F,$F0,$13,$F2,$DB,$1F
	.dc.b	$4D,$B7,$FC,$B5,$FF,$75,$FC,$03,$D3,$C7,$C9,$A3,$13,$FC,$CE,$31
	.dc.b	$3A,$9C,$CF,$83,$C3,$7C,$07,$5C,$91,$89,$FE,$79,$18,$9D,$4E,$67
	.dc.b	$94,$7C,$98,$E1,$9F,$E0,$7E,$05,$D1,$FE,$34,$8F,$C0,$CE,$B7,$FC
	.dc.b	$A3,$70,$FC,$6B,$55,$F2,$F0,$FB,$63,$27,$EE,$56,$EB,$51,$89,$D6
	.dc.b	$1E,$07,$F1,$53,$A8,$C4,$EC,$8F,$A7,$1F,$46,$3E,$6B,$FC,$A7,$2F
	.dc.b	$FA,$4B,$E2,$B3,$87,$F9,$3F,$CF,$7C,$5E,$2B,$5A,$F0,$9C,$B6,$F5
	.dc.b	$18,$BF,$D5,$D7,$06,$C6,$73,$FD,$94,$FF,$A8,$BE,$2B,$FC,$BF,$DB
	.dc.b	$FE,$57,$AF,$F4,$86,$3F,$DE,$14,$EA,$3F,$C3,$D1,$AC,$3F,$8E,$C6
	.dc.b	$27,$51,$EB,$F6,$EE,$1D,$EB,$A4,$62,$DF,$0F,$4C,$7A,$F6,$BC,$51
	.dc.b	$70,$EE,$7C,$79,$C6,$5F,$E6,$FF,$C3,$5E,$2F,$81,$B8,$7F,$84,$9F
	.dc.b	$E9,$6D,$8D,$11,$D7,$8B,$70,$EF,$A2,$F9,$C6,$1B,$87,$86,$B8,$3F
	.dc.b	$2F,$E2,$7D,$DF,$13,$EF,$6E,$3F,$35,$C3,$F3,$FE,$26,$E3,$E0,$D9
	.dc.b	$18,$C3,$EB,$36,$BF,$CA,$19,$3C,$8F,$9F,$E2,$3E,$5B,$E1,$46,$1F
	.dc.b	$89,$D3,$2F,$E2,$78,$F7,$4F,$67,$93,$F2,$63,$E5,$36,$7F,$09,$F1
	.dc.b	$5F,$D6,$EB,$F0,$6D,$7F,$34,$DE,$2F,$F7,$7B,$77,$7C,$4F,$E0,$4F
	.dc.b	$C1,$7C,$55,$C3,$F2,$1F,$11,$F8,$5E,$FF,$C3,$36,$3F,$29,$E3,$7E
	.dc.b	$27,$E2,$BF,$14,$D8,$FB,$DC,$1F,$81,$3F,$05,$E5,$58,$FF,$29,$FE
	.dc.b	$1C,$F8,$9F,$CA,$B6,$33,$F1,$58,$9C,$FC,$2C,$9E,$3E,$75,$E3,$FC
	.dc.b	$2E,$1F,$EB,$99,$EB,$E9,$74,$F8,$6E,$AF,$6F,$84,$FA,$BF,$2F,$6B
	.dc.b	$FC,$66,$F8,$B9,$7E,$CF,$99,$67,$FC,$6E,$FE,$AB,$F0,$FF,$D8,$78
	.dc.b	$3F,$EC,$D6,$CF,$FA,$45,$79,$7B,$F2,$7D,$BF,$09,$2B,$E0,$FC,$2D
	.dc.b	$E0,$F4,$FC,$2F,$C2,$3E,$1F,$C2,$DA,$F8,$F1,$67,$DF,$37,$3F,$F6
	.dc.b	$F7,$17,$FA,$4A,$F6,$7C,$E9,$D3,$FE,$A5,$87,$C2,$7D,$0F,$85,$FA
	.dc.b	$2D,$9F,$AA,$DC,$3E,$6B,$70,$EF,$EA,$7D,$8F,$67,$EC,$B6,$FE,$EB
	.dc.b	$6F,$EA,$B6,$BC,$FF,$09,$F8,$03,$F0,$3F,$0B,$6F,$F0,$B1,$6F,$EE
	.dc.b	$CD,$AE,$35,$A3,$F1,$71,$97,$FD,$F0,$BC,$0F,$D9,$FC,$BF,$F2,$4B
	.dc.b	$4E,$6B,$FC,$E7,$39,$8C,$4F,$FA,$8D,$AF,$FB,$FD,$AF,$6A,$FD,$77
	.dc.b	$E2,$0F,$85,$FE,$5E,$DC,$3F,$D6,$2F,$FA,$0D,$B1,$A2,$3A,$F8,$7B
	.dc.b	$7F,$CB,$B2,$7E,$17,$FD,$35,$FE,$8E,$DA,$FF,$73,$AF,$92,$FE,$83
	.dc.b	$3C,$3D,$CB,$3B,$BF,$F3,$1B,$67,$AD,$6A,$A5,$FF,$B3,$67,$93,$6B
	.dc.b	$B3,$67,$F4,$9E,$67,$E9,$7E,$36,$CF,$FB,$43,$B3,$FD,$B0,$D9,$D3
	.dc.b	$4D,$4F,$D5,$FC,$2D,$E2,$7F,$58,$FC,$41,$D5,$98,$C3,$6F,$FC,$59
	.dc.b	$5C,$1B,$3F,$C8,$F6,$4E,$51,$0F,$F4,$2F,$CF,$BC,$9F,$92,$DB,$F8
	.dc.b	$5F,$17,$FE,$74,$7E,$24,$66,$BE,$B3,$78,$3F,$BD,$B6,$7B,$C2,$B9
	.dc.b	$57,$0B,$EE,$4A,$EF,$E6,$D9,$FE,$DE,$7A,$BC,$5F,$BB,$F8,$77,$F4
	.dc.b	$FA,$FF,$51,$66,$9C,$FA,$FF,$2D,$F8,$51,$99,$F8,$F1,$83,$E3,$7D
	.dc.b	$03,$E3,$FF,$13,$F4,$9F,$E1,$0A,$FF,$57,$4E,$5F,$C2,$46,$72,$F2
	.dc.b	$CE,$5F,$98,$7C,$E7,$0E,$1A,$FD,$16,$AF,$5D,$D9,$FE,$DB,$BF,$D4
	.dc.b	$FC,$87,$EB,$31,$FD,$6B,$23,$27,$EB,$A3,$17,$FE,$08,$AF,$A1,$3F
	.dc.b	$B2,$E0,$FD,$AF,$C8,$7E,$D9,$91,$93,$F6,$F1,$8A,$E5,$F0,$5F,$BC
	.dc.b	$3F,$7B,$F2,$16,$BF,$7B,$16,$7C,$13,$79,$15,$F9,$0D,$3D,$BF,$09
	.dc.b	$FC,$18,$FE,$0F,$C5,$67,$F8,$3C,$5C,$3D,$53,$6F,$9C,$7C,$7D,$3F
	.dc.b	$CC,$F1,$8C,$3E,$4D,$7F,$12,$F7,$FA,$E3,$FC,$49,$B8,$7C,$56,$DF
	.dc.b	$F8,$27,$E2,$3F,$4B,$F4,$0F,$6B,$F4,$D2,$EF,$F0,$CE,$70,$FB,$F7
	.dc.b	$C2,$F7,$5A,$7F,$07,$29,$D3,$FA,$49,$D3,$C8,$57,$57,$FD,$F3,$1A
	.dc.b	$7F,$DB,$66,$AB,$BE,$3D,$23,$35,$C7,$0F,$44,$E6,$7F,$DD,$87,$13
	.dc.b	$C4,$E5,$F3,$19,$3C,$2E,$70,$ED,$3D,$27,$47,$4A,$F0,$61,$E7,$8C
	.dc.b	$61,$E6,$76,$76,$9D,$E1,$D6,$3C,$17,$D6,$7E,$14,$72,$3C,$E7,$2C
	.dc.b	$3C,$78,$77,$8E,$98,$77,$AE,$31,$83,$8B,$8A,$E3,$87,$86,$3C,$73
	.dc.b	$D4,$F1,$DF,$5A,$DC,$66,$76,$73,$77,$5C,$EF,$C4,$7E,$0A,$B8,$D6
	.dc.b	$CE,$31,$9A,$E2,$75,$78,$BF,$85,$E9,$FE,$F8,$DF,$1B,$84,$62,$3F
	.dc.b	$0A,$DF,$8E,$FF,$3C,$E7,$FD,$02,$31,$F0,$A8,$C5,$FF,$43,$AD,$DA
	.dc.b	$F7,$5A,$8F,$E9,$33,$AC,$7F,$2D,$8B,$CA,$FE,$99,$18,$6C,$75,$CB
	.dc.b	$C1,$7F,$8D,$BE,$53,$FD,$4C,$E5,$1F,$9B,$6B,$B9,$E3,$7F,$A5,$AE
	.dc.b	$33,$FD,$72,$B7,$3E,$18,$FD,$24,$E9,$36,$76,$78,$6F,$CF,$3E,$EB
	.dc.b	$EF,$3E,$9C,$3C,$95,$AA,$FE,$0C,$DB,$EC,$75,$75,$5D,$A7,$A5,$79
	.dc.b	$2F,$AC,$7A,$6B,$67,$38,$CC,$7A,$1B,$5C,$E3,$C2,$73,$A8,$FE,$E6
	.dc.b	$34,$6E,$FC,$B5,$CA,$FB,$D7,$0C,$3E,$64,$E6,$3D,$E6,$6F,$E1,$3D
	.dc.b	$6B,$8B,$F5,$70,$D3,$EB,$38,$99,$25,$FD,$57,$91,$C3,$83,$FB,$0F
	.dc.b	$C1,$7F,$30,$DB,$FA,$47,$13,$E0,$F1,$7F,$D0,$5B,$7C,$23,$FC,$9F
	.dc.b	$78,$AE,$11,$FE,$52,$AC,$4E,$7E,$42,$4F,$F9,$59,$E1,$38,$8F,$ED
	.dc.b	$51,$72,$FE,$9D,$7D,$27,$FB,$76,$9D,$C8,$FF,$98,$3E,$42,$BA,$3F
	.dc.b	$CC,$71,$E5,$8F,$37,$B7,$27,$FC,$CE,$FA,$7E,$07,$11,$FD,$EE,$2E
	.dc.b	$1E,$29,$F7,$6F,$2B,$FC,$DE,$FC,$1F,$73,$11,$FE,$03,$8B,$5E,$A9
	.dc.b	$F9,$72,$F9,$B7,$B5,$FD,$62,$3F,$C8,$F1,$87,$0D,$7F,$AE,$B5,$FF
	.dc.b	$10,$B9,$AF,$EE,$F0,$F9,$C7,$C2,$F8,$1B,$1F,$E4,$6E,$3F,$F2,$A5
	.dc.b	$F1,$7F,$CA,$DF,$86,$3C,$4F,$25,$23,$0F,$9D,$AF,$F7,$06,$BF,$E3
	.dc.b	$47,$38,$F9,$38,$FF,$B3,$5F,$F9,$8D,$B5,$FE,$66,$AD,$CE,$7B,$FE
	.dc.b	$3D,$1A,$63,$C1,$B7,$EB,$3F,$CD,$F3,$B7,$07,$E6,$E7,$83,$88,$D3
	.dc.b	$FE,$60,$6E,$1E,$79,$DC,$FF,$99,$30,$77,$FC,$46,$C7,$C3,$8C,$11
	.dc.b	$ED,$97,$A8,$FC,$0C,$F0,$8F,$E8,$99,$7F,$8D,$E9,$EA,$7E,$BB,$F1
	.dc.b	$1F,$B7,$97,$F8,$BD,$B5,$B9,$FD,$9D,$70,$BF,$85,$D4,$BD,$C7,$F9
	.dc.b	$7D,$B1,$EA,$AF,$9A,$49,$F8,$8F,$E5,$BF,$12,$E3,$1F,$C0,$E3,$4F
	.dc.b	$C7,$75,$86,$31,$FB,$D1,$63,$FD,$41,$5E,$0E,$9F,$F3,$BD,$7E,$F2
	.dc.b	$FF,$84,$1F,$C2,$7E,$4A,$BF,$ED,$A2,$9F,$E6,$D7,$E4,$F4,$FD,$12
	.dc.b	$47,$E4,$DA,$FC,$C3,$FC,$FD,$F1,$89,$FE,$32,$D8,$FC,$2E,$CA,$FE
	.dc.b	$3B,$F8,$9B,$D7,$A4,$AF,$9B,$1F,$EE,$16,$C6,$5F,$06,$C9,$E5,$1C
	.dc.b	$3D,$9D,$67,$FE,$69,$9C,$CF,$06,$C6,$67,$C7,$F0,$9E,$4F,$85,$97
	.dc.b	$E2,$8C,$BE,$48,$CB,$CE,$3D,$7F,$27,$EC,$FA,$09,$3E,$A8,$CB,$EC
	.dc.b	$8C,$BE,$38,$FC,$17,$C0,$C8,$FD,$5F,$12,$3F,$0D,$F0,$59,$9F,$F6
	.dc.b	$4C,$E5,$DC,$E9,$A4,$36,$E5,$F8,$5F,$29,$F4,$88,$FD,$0F,$94,$FC
	.dc.b	$FF,$4C,$8F,$CF,$BF,$AB,$DB,$FD,$28,$BD,$0F,$B7,$F2,$7F,$80,$A8
	.dc.b	$7D,$A7,$EF,$F9,$9F,$D7,$3D,$87,$D2,$53,$35,$F8,$0F,$4F,$F0,$A4
	.dc.b	$7F,$07,$F0,$3F,$88,$F9,$39,$1F,$86,$7F,$17,$F2,$7F,$8C,$F9,$39
	.dc.b	$3F,$8A,$7F,$1F,$E5,$7E,$EB,$C0,$93,$EC,$7F,$7B,$E6,$FE,$53,$D5
	.dc.b	$FC,$AF,$AD,$2B,$F2,$71,$FD,$D3,$E5,$5F,$29,$7E,$0A,$57,$F0,$49
	.dc.b	$CB,$7E,$53,$81,$1D,$7C,$2F,$3F,$93,$FE,$A9,$E2,$AD,$18,$6D,$7F
	.dc.b	$AC,$27,$53,$C3,$F1,$4F,$D3,$4F,$E9,$1F,$CE,$F9,$9F,$ED,$56,$C7
	.dc.b	$FB,$CD,$DF,$67,$FB,$35,$78,$9F,$E1,$06,$CF,$F1,$E6,$BD,$7C,$8D
	.dc.b	$78,$DF,$6F,$C7,$93,$B8,$F8,$DF,$29,$F1,$FB,$64,$FC,$58,$AF,$EE
	.dc.b	$B6,$4E,$33,$FF,$70,$42,$B1,$A7,$FB,$BE,$1F,$39,$F4,$CC,$E1,$ED
	.dc.b	$F2,$3D,$44,$7D,$5E,$4D,$7E,$34,$5C,$7E,$14,$DC,$3E,$44,$DA,$F9
	.dc.b	$13,$F8,$1F,$95,$91,$F7,$DF,$C2,$7B,$18,$8F,$64,$5C,$3D,$73,$F8
	.dc.b	$9F,$93,$FC,$57,$CA,$CB,$FC,$44,$65,$FC,$54,$65,$F8,$D1,$F9,$1F
	.dc.b	$35,$13,$E0,$FF,$C1,$FC,$FF,$89,$F9,$4C,$CF,$FB,$D7,$C1,$F9,$11
	.dc.b	$8A,$DF,$0B,$0C,$57,$CD,$6D,$B0,$E8,$F7,$9E,$6D,$35,$5C,$DC,$0F
	.dc.b	$98,$ED,$DF,$86,$C7,$94,$F5,$B8,$C7,$FC,$53,$1C,$7B,$3C,$6F,$3C
	.dc.b	$BF,$D7,$4F,$13,$17,$D6,$33,$5C,$AD,$7C,$9E,$D8,$FC,$66,$E1,$E2
	.dc.b	$BE,$33,$9A,$E1,$18,$9D,$61,$DC,$F6,$4E,$9C,$47,$96,$FA,$BE,$B3
	.dc.b	$27,$D6,$3D,$25,$29,$97,$E3,$BB,$F8,$57,$CC,$BE,$6B,$FC,$B9,$18
	.dc.b	$6C,$66,$7A,$4E,$5E,$D5,$FD,$37,$C5,$FF,$32,$36,$33,$E9,$7F,$79
	.dc.b	$7C,$4D,$7E,$43,$FD,$F7,$5F,$9C,$DD,$7F,$23,$F0,$3F,$E3,$AE,$9F
	.dc.b	$C0,$4E,$6F,$EF,$B5,$FC,$33,$67,$84,$7C,$98,$D6,$1E,$5C,$3B,$E3
	.dc.b	$F8,$EB,$DC,$62,$F8,$4F,$80,$D7,$A2,$C7,$13,$A2,$57,$E4,$2F,$9E
	.dc.b	$5E,$EC,$3F,$3D,$C5,$FC,$5B,$BB,$96,$1F,$99,$8F,$E3,$3E,$06,$31
	.dc.b	$FD,$9F,$7A,$F1,$8F,$EE,$2F,$79,$F4,$AD,$37,$0D,$E3,$FB,$19,$CB
	.dc.b	$DD,$FF,$14,$4F,$1E,$DF,$1F,$5F,$93,$0F,$E5,$7E,$16,$2B,$C1,$6B
	.dc.b	$C6,$1D,$6A,$DF,$F3,$29,$B5,$BB,$FF,$2F,$AE,$EC,$CE,$BD,$39,$5A
	.dc.b	$F4,$EE,$5F,$E3,$6C,$15,$8C,$FC,$D0,$AC,$5F,$C5,$F9,$AF,$F4,$93
	.dc.b	$FE,$6E,$6E,$1F,$E4,$05,$F0,$1E,$87,$38,$7C,$16,$D7,$3F,$7F,$55
	.dc.b	$B6,$47,$57,$2D,$3D,$FE,$AF,$C2,$EF,$F8,$6D,$8C,$93,$1F,$77,$AB
	.dc.b	$A8,$F8,$EF,$CA,$3E,$5F,$65,$8F,$67,$AD,$E9,$3D,$5D,$56,$3E,$03
	.dc.b	$87,$E1,$7C,$05,$C3,$E1,$7C,$3F,$C1,$8F,$84,$D9,$FA,$FE,$BF,$DA
	.dc.b	$3E,$DF,$AF,$63,$ED,$FA,$FF,$78,$FB,$FE,$BD,$8F,$BF,$EB,$FE,$0B
	.dc.b	$B7,$F0,$6D,$8F,$C1,$F8,$DF,$86,$F5,$BF,$0E,$D8,$FB,$1F,$09,$F5
	.dc.b	$4F,$AD,$F0,$D6,$3F,$D5,$3F,$33,$FE,$2A,$6F,$B3,$F9,$28,$7B,$36
	.dc.b	$31,$EC,$CA,$E3,$E7,$5E,$4F,$F4,$1E,$1F,$E2,$89,$EB,$EC,$F3,$F9
	.dc.b	$AE,$8F,$5F,$62,$4F,$8B,$E7,$6E,$1C,$BC,$4F,$11,$E3,$F6,$6D,$7F
	.dc.b	$B0,$3E,$5F,$FC,$4E,$DF,$2F,$F5,$31,$FB,$6F,$1B,$F6,$4E,$B0,$F2
	.dc.b	$FB,$5E,$D3,$DD,$ED,$5C,$3E,$1F,$B5,$F0,$4F,$85,$F2,$D6,$32,$4B
	.dc.b	$F7,$C6,$5F,$83,$37,$1F,$E0,$5C,$2A,$C5,$FF,$11,$87,$B5,$F3,$FD
	.dc.b	$BF,$A0,$D9,$FA,$AD,$C7,$E6,$B7,$0E,$BE,$C7,$D7,$F6,$FE,$C3,$6F
	.dc.b	$EE,$B6,$FE,$AB,$6B,$CB,$F3,$7F,$7C,$FC,$07,$CD,$DB,$FC,$2C,$5B
	.dc.b	$FB,$B3,$6B,$8D,$68,$FC,$5C,$65,$FF,$10,$AE,$FF,$E6,$F5,$FD,$BE
	.dc.b	$31,$3F,$DB,$E1,$18,$7F,$B9,$35,$E1,$3F,$CA,$1D,$BF,$97,$9C,$4F
	.dc.b	$1A,$FE,$65,$F0,$EF,$9B,$FE,$5A,$DA,$FF,$1C,$CE,$BE,$27,$35,$27
	.dc.b	$27,$F4,$7A,$E1,$8F,$FB,$81,$56,$E3,$FD,$E5,$6F,$9B,$FE,$8B,$18
	.dc.b	$6C,$7F,$46,$5F,$05,$FD,$0A,$38,$7B,$76,$77,$8F,$99,$B6,$AE,$D7
	.dc.b	$BB,$97,$E8,$AD,$5F,$0B,$DC,$E5,$EE,$FE,$AF,$CB,$FD,$4F,$97,$FA
	.dc.b	$A6,$C6,$4F,$D8,$46,$5A,$77,$B7,$CC,$7F,$8F,$A3,$EE,$36,$3F,$C5
	.dc.b	$8D,$7F,$C6,$0D,$7F,$CC,$EE,$EF,$F0,$ED,$9F,$EF,$1F,$8A,$3E,$47
	.dc.b	$F0,$53,$9B,$F8,$9F,$37,$FC,$62,$FD,$CD,$8F,$AB,$F2,$EF,$33,$E8
	.dc.b	$4E,$72,$FE,$F2,$DB,$FC,$2B,$6F,$E8,$5F,$19,$EB,$E5,$72,$23,$F3
	.dc.b	$63,$53,$F8,$4A,$E4,$FC,$6F,$0F,$F4,$6F,$C0,$6B,$FE,$8C,$CB,$FD
	.dc.b	$67,$7F,$2A,$73,$3C,$63,$E8,$4E,$A3,$F1,$DF,$03,$F1,$2B,$F4,$DF
	.dc.b	$32,$F9,$9E,$2D,$C3,$F6,$0D,$BF,$F3,$03,$5F,$F3,$23,$9D,$3F,$72
	.dc.b	$D9,$F8,$BE,$67,$6B,$FD,$6F,$C9,$C9,$FC,$97,$C2,$3A,$78,$65,$FE
	.dc.b	$9B,$55,$B8,$FD,$63,$FB,$6B,$C6,$5F,$BA,$DF,$F7,$27,$EE,$BE,$66
	.dc.b	$E5,$DB,$14,$E3,$1F,$DB,$55,$BF,$E0,$7F,$01,$FC,$08,$FE,$07,$F3
	.dc.b	$37,$1E,$51,$6F,$F7,$D3,$6B,$F8,$5F,$C0,$FF,$0A,$3F,$85,$FB,$B7
	.dc.b	$0F,$14,$5B,$FE,$0D,$36,$BF,$89,$FB,$BC,$2F,$D0,$DA,$FE,$18,$D9
	.dc.b	$DC,$FB,$6B,$F9,$05,$7F,$B3,$C7,$CC,$BD,$7F,$F6,$7D,$BB,$7F,$CC
	.dc.b	$B3,$9A,$FF,$4F,$7C,$47,$FB,$62,$31,$1F,$C2,$E7,$77,$FE,$A1,$AD
	.dc.b	$3D,$23,$33,$C8,$F0,$C6,$71,$DC,$66,$FA,$E1,$E3,$9C,$CF,$FB,$84
	.dc.b	$E2,$7F,$32,$BE,$AE,$6B,$8C,$62,$3E,$04,$62,$79,$C6,$2F,$C4,$F4
	.dc.b	$39,$38,$9F,$7C,$6C,$F0,$4F,$25,$87,$9B,$0F,$44,$73,$38,$1F,$1E
	.dc.b	$78,$4E,$F0,$F3,$4F,$23,$C5,$5C,$AF,$C3,$19,$7D,$4F,$53,$18,$79
	.dc.b	$CF,$A4,$F2,$AE,$B7,$DF,$0F,$0C,$E5,$B7,$D0,$EF,$3B,$9E,$CD,$8C
	.dc.b	$C7,$13,$25,$72,$73,$BF,$1B,$63,$EF,$47,$9D,$FC,$27,$5F,$F7,$F6
	.dc.b	$F8,$BE,$A8,$C4,$7E,$05,$BF,$18,$F8,$2F,$EB,$93,$FC,$D6,$8C,$9D
	.dc.b	$69,$8B,$E3,$7C,$1B,$19,$BE,$36,$BE,$2F,$E1,$70,$7C,$7E,$B2,$FD
	.dc.b	$1C,$62,$7F,$2F,$5C,$27,$F4,$ED,$9E,$A6,$E7,$A5,$73,$8F,$0C,$EE
	.dc.b	$7D,$59,$7B,$AF,$BD,$EE,$7D,$F7,$CE,$7B,$D6,$E3,$9B,$67,$75,$C5
	.dc.b	$FE,$04,$DB,$F0,$47,$C1,$9E,$0E,$8F,$D9,$D7,$73,$AE,$3F,$02,$77
	.dc.b	$19,$9E,$E6,$66,$3F,$00,$E7,$7E,$1A,$EF,$7E,$2B,$E5,$5E,$13,$F6
	.dc.b	$D1,$E6,$35,$5E,$B9,$DC,$66,$BC,$73,$93,$B5,$F8,$B0,$F9,$78,$78
	.dc.b	$8C,$61,$EC,$7C,$C4,$8F,$9B,$3A,$8F,$69,$99,$E1,$3A,$8C,$1A,$70
	.dc.b	$7F,$2D,$E8,$FE,$BD,$F0,$3F,$CC,$B6,$FF,$24,$7C,$15,$BF,$EC,$0D
	.dc.b	$AD,$C7,$F5,$9A,$C7,$42,$A9,$88,$CF,$CC,$CA,$FE,$C0,$E9,$A4,$3F
	.dc.b	$CA,$CE,$5F,$F2,$C6,$5D,$88,$FF,$69,$F9,$4A,$E8,$FE,$D9,$1E,$48
	.dc.b	$F2,$F8,$92,$7F,$B8,$3E,$8F,$80,$C4,$7F,$99,$62,$ED,$FE,$72,$9F
	.dc.b	$6E,$F2,$BF,$BC,$3F,$03,$E3,$B1,$1F,$E6,$D8,$B5,$FE,$7A,$9F,$95
	.dc.b	$2F,$98,$7B,$18,$6C,$7F,$94,$DA,$FF,$99,$F8,$7F,$81,$FC,$CF,$F1
	.dc.b	$2C,$3F,$D2,$75,$AA,$FE,$0E,$75,$BF,$F4,$A3,$4F,$F4,$B4,$6D,$FE
	.dc.b	$1B,$F2,$FF,$E1,$A7,$FC,$C4,$D7,$C0,$7F,$1A,$71,$5F,$EA,$2E,$1F
	.dc.b	$F2,$D4,$F0,$F0,$FB,$D7,$18,$F1,$B5,$EF,$1E,$49,$77,$7C,$46,$CF
	.dc.b	$F9,$CD,$AF,$88,$FC,$67,$81,$B9,$91,$B7,$94,$62,$38,$4F,$2C,$3D
	.dc.b	$15,$A9,$F7,$F7,$FD,$4C,$4D,$3C,$9E,$15,$9F,$F5,$F4,$E9,$B3,$F0
	.dc.b	$0F,$F6,$17,$47,$2F,$43,$F4,$DF,$33,$FE,$C8,$C9,$C3,$FD,$12,$93
	.dc.b	$98,$C5,$FF,$AD,$30,$FF,$28,$B5,$FA,$75,$C2,$7A,$9F,$B3,$9D,$5F
	.dc.b	$D5,$C3,$E6,$12,$FF,$D3,$DF,$2F,$DD,$FE,$AB,$AF,$0C,$3F,$A7,$B6
	.dc.b	$33,$5F,$98,$AE,$11,$FB,$D8,$E8,$FC,$E7,$55,$FC,$0A,$21,$FE,$5D
	.dc.b	$3F,$82,$FC,$9E,$1F,$F6,$91,$70,$FF,$6F,$C7,$D3,$E8,$93,$F9,$47
	.dc.b	$F8,$87,$CC,$3E,$63,$33,$FA,$01,$BA,$BF,$26,$48,$D6,$CB,$7A,$9D
	.dc.b	$09,$3F,$E8,$08,$FF,$6E,$B6,$32,$FD,$28,$E2,$EF,$D5,$CE,$7F,$7A
	.dc.b	$17,$DE,$F3,$A7,$E0,$70,$EB,$5C,$FE,$0B,$DB,$E8,$48,$F1,$47,$C0
	.dc.b	$30,$F8,$63,$0F,$9F,$E5,$3E,$2F,$CA,$C8,$C1,$0F,$43,$97,$D5,$1F
	.dc.b	$77,$DC,$91,$F8,$67,$F0,$1E,$EE,$23,$F6,$5C,$89,$C6,$DE,$F9,$4A
	.dc.b	$B3,$A7,$F8,$6A,$38,$4F,$53,$F2,$73,$F7,$3E,$56,$47,$D1,$7E,$FF
	.dc.b	$CA,$FE,$77,$A2,$4F,$E9,$1F,$D0,$7A,$1F,$84,$F9,$39,$3F,$65,$FC
	.dc.b	$3F,$CA,$7E,$23,$CD,$93,$F7,$5F,$C5,$FC,$A7,$E3,$3D,$39,$3F,$81
	.dc.b	$7F,$1F,$EF,$7E,$43,$E5,$24,$FE,$15,$FC,$9F,$CA,$7E,$53,$E4,$E4
	.dc.b	$FE,$25,$FC,$BF,$CA,$7E,$D7,$C1,$93,$FB,$C7,$F7,$1E,$6F,$E7,$3D
	.dc.b	$7F,$CE,$FA,$F2,$BF,$25,$1F,$DB,$F7,$7C,$A5,$F8,$29,$5F,$ED,$69
	.dc.b	$CB,$7D,$EE,$04,$75,$F2,$9C,$FE,$4E,$47,$17,$B7,$CA,$7F,$7D,$F9
	.dc.b	$8F,$CA,$F7,$FE,$6F,$C7,$47,$E8,$2F,$F8,$9B,$67,$4F,$0F,$03,$F8
	.dc.b	$F2,$F1,$7A,$12,$BF,$DD,$B1,$C6,$77,$E2,$F8,$49,$3A,$AF,$85,$F2
	.dc.b	$B2,$3E,$03,$F1,$7E,$57,$E3,$7C,$34,$9F,$88,$ED,$FE,$53,$C0,$8F
	.dc.b	$99,$F2,$B2,$76,$79,$3C,$8F,$84,$47,$CF,$E4,$D7,$E0,$C5,$DB,$E0
	.dc.b	$CD,$D7,$E0,$CD,$AF,$E0,$7B,$2B,$EE,$7C,$A4,$9F,$B4,$FD,$FF,$67
	.dc.b	$11,$E8,$8B,$8F,$D3,$D3,$FC,$0A,$DF,$91,$FC,$31,$2B,$F0,$9F,$27
	.dc.b	$5F,$83,$F2,$9F,$8B,$31,$5C,$FC,$CF,$F4,$7F,$C3,$BC,$1F,$EA,$D9
	.dc.b	$3E,$4F,$31,$86,$E3,$FC,$7A,$7F,$80,$3F,$EA,$0A,$E4,$E6,$F7,$D0
	.dc.b	$AE,$93,$BC,$3A,$34,$E0,$51,$F2,$FF,$C3,$7C,$0C,$36,$3F,$DC,$33
	.dc.b	$FE,$A8,$8C,$63,$E2,$9E,$3D,$BE,$87,$9E,$4C,$39,$B6,$39,$9F,$EB
	.dc.b	$AB,$E7,$87,$5B,$E5,$18,$9D,$57,$1C,$3C,$13,$99,$F0,$9D,$B0,$8F
	.dc.b	$63,$C6,$FB,$2F,$AD,$F3,$C3,$D8,$DA,$EF,$5B,$79,$9F,$E4,$2B,$F8
	.dc.b	$8D,$9F,$63,$67,$FB,$43,$FE,$CD,$F2,$5F,$31,$8F,$8A,$46,$27,$FB
	.dc.b	$74,$62,$75,$39,$9F,$07,$A6,$F1,$FA,$E5,$FD,$1D,$B1,$99,$FE,$F1
	.dc.b	$39,$6C,$67,$5F,$92,$DD,$3F,$DC,$7F,$07,$FE,$36,$EB,$FC,$0C,$E6
	.dc.b	$FE,$B3,$5F,$C3,$B6,$78,$47,$CF,$C3,$C1,$87,$6C,$79,$9F,$8F,$AD
	.dc.b	$C6,$6B,$85,$F5,$BE,$38,$F9,$3D,$1F,$F0,$05,$EC,$91,$F9,$3A,$F1
	.dc.b	$4E,$A3,$18,$FE,$7F,$89,$F3,$0F,$82,$D6,$7F,$E0,$A8,$C6,$BB,$5F
	.dc.b	$1D,$AB,$E3,$76,$77,$7F,$9A,$9C,$DF,$1B,$D5,$EF,$E3,$7F,$C3,$1D
	.dc.b	$1F,$CF,$F1,$7C,$76,$30,$FF,$2C,$42,$B1,$97,$96,$AE,$5E,$A9,$B5
	.dc.b	$CA,$D7,$FE,$5A,$8E,$51,$8C,$3C,$18,$70,$F8,$49,$39,$3C,$3F,$0B
	.dc.b	$8C,$39,$4D,$BF,$44,$DA,$FF,$37,$7D,$83,$93,$CA,$D8,$E9,$87,$F8
	.dc.b	$DA,$1F,$15,$AA,$95,$AD,$55,$C6,$FF,$CE,$6F,$09,$D3,$9C,$7F,$B4
	.dc.b	$36,$3C,$77,$E2,$C3,$AD,$FC,$3E,$2F,$8F,$ED,$7C,$86,$DF,$87,$6F
	.dc.b	$CC,$E5,$F5,$67,$E6,$FC,$2E,$EB,$4D,$C3,$E7,$36,$BC,$7E,$A7,$D6
	.dc.b	$EF,$FF,$1D,$5F,$A2,$31,$CF,$F1,$B9,$A4,$FC,$1F,$C5,$7F,$CE,$0F
	.dc.b	$BA,$3E,$5E,$5F,$80,$9C,$BF,$10,$C7,$C2,$FE,$0B,$E1,$3F,$C8,$D1
	.dc.b	$F8,$A6,$BF,$9D,$6F,$1F,$E8,$9B,$BB,$E2,$FF,$16,$7E,$33,$E3,$2E
	.dc.b	$7F,$17,$93,$F1,$FC,$9F,$90,$6C,$7C,$9D,$7F,$27,$BF,$E5,$1B,$1E
	.dc.b	$DF,$4B,$D4,$7A,$FD,$3B,$1F,$81,$F9,$AF,$6F,$27,$F8,$8D,$B1,$FA
	.dc.b	$4E,$0F,$D0,$C6,$1F,$F4,$0B,$FE,$8E,$8E,$5F,$0D,$AF,$87,$DB,$C3
	.dc.b	$19,$3D,$F1,$B8,$7F,$B0,$30,$7A,$5F,$67,$E6,$3B,$B6,$FC,$DF,$0D
	.dc.b	$F7,$7E,$67,$C6,$D8,$FB,$5F,$09,$F5,$CF,$B1,$F0,$96,$3D,$DF,$0D
	.dc.b	$EA,$3D,$7D,$B6,$3F,$4F,$C2,$F9,$8F,$66,$C8,$FE,$13,$18,$6F,$17
	.dc.b	$F0,$67,$87,$15,$9D,$5F,$F0,$16,$DF,$35,$7B,$A9,$5E,$29,$DD,$F8
	.dc.b	$EB,$37,$F1,$FE,$1F,$E9,$F9,$BF,$51,$B8,$7D,$B6,$CE,$E3,$EF,$4F
	.dc.b	$DA,$D7,$ED,$EB,$67,$F0,$4D,$AF,$A8,$DC,$7F,$08,$FE,$0B,$E1,$E5
	.dc.b	$7E,$27,$E1,$ED,$FD,$B9,$B7,$F8,$B9,$B5,$FC,$9B,$BD,$CA,$BC,$E1
	.dc.b	$F8,$CB,$AB,$E5,$9E,$AF,$F3,$2F,$9E,$7B,$3F,$CD,$AB,$53,$98,$C5
	.dc.b	$70,$CF,$FC,$06,$DA,$FF,$76,$35,$FE,$72,$7F,$3A,$70,$F9,$F2,$7C
	.dc.b	$4F,$E8,$8F,$D1,$FA,$56,$FF,$55,$16,$BF,$39,$37,$6F,$EA,$3F,$30
	.dc.b	$F8,$8F,$C4,$37,$0F,$C8,$FC,$8F,$F8,$DD,$FC,$9C,$67,$0F,$C9,$47
	.dc.b	$F4,$B8,$C5,$F4,$BF,$C9,$5F,$1B,$D9,$F9,$9A,$DD,$F1,$AF,$09,$C3
	.dc.b	$DA,$B1,$FD,$0F,$AD,$F1,$3F,$D8,$1F,$91,$97,$FA,$9D,$CC,$EF,$83
	.dc.b	$55,$B6,$C6,$4F,$EA,$79,$7D,$BF,$85,$FD,$47,$CF,$3D,$2F,$AF,$18
	.dc.b	$6D,$7D,$88,$FE,$57,$5C,$63,$07,$D8,$7E,$25,$FC,$C3,$E6,$FC,$5D
	.dc.b	$9F,$97,$16,$7E,$2E,$6D,$BF,$DA,$35,$A2,$4E,$A3,$E0,$C7,$D1,$BE
	.dc.b	$54,$8F,$C4,$FC,$CF,$F8,$61,$FE,$17,$F3,$DF,$C7,$1F,$83,$19,$C3
	.dc.b	$C3,$39,$7E,$B8,$8D,$3E,$D6,$BD,$BD,$CF,$87,$18,$8F,$F0,$8F,$E2
	.dc.b	$4F,$17,$F1,$BF,$15,$FA,$58,$C6,$3F,$8C,$8C,$4F,$CC,$9C,$D7,$CB
	.dc.b	$0E,$1F,$9C,$D7,$F4,$9C,$7F,$D1,$0F,$E8,$BF,$1F,$70,$FE,$99,$16
	.dc.b	$7F,$9F,$4D,$D3,$FD,$E1,$F2,$BF,$42,$7F,$2F,$E0,$23,$13,$AE,$7F
	.dc.b	$F1,$8C,$63,$1F,$E1,$4D,$B5,$F1,$E0,$DD,$F0,$78,$F3,$4C,$3F,$C4
	.dc.b	$18,$7D,$CC,$38,$BF,$9C,$7C,$6D,$C7,$CF,$5E,$2E,$3E,$55,$AF,$8F
	.dc.b	$F2,$9E,$6F,$90,$B8,$F6,$8B,$3D,$26,$E1,$F0,$BE,$3F,$85,$F9,$1B
	.dc.b	$5E,$26,$CE,$E7,$D3,$5F,$C8,$6B,$FD,$A8,$3E,$6D,$EB,$7F,$1C,$B7
	.dc.b	$8F,$F9,$7A,$D4,$FD,$FF,$85,$FF,$2C,$DF,$09,$FF,$2C,$56,$E3,$11
	.dc.b	$FE,$EA,$38,$9A,$C7,$8D,$F3,$8C,$9F,$DA,$6B,$94,$66,$B9,$E1,$E4
	.dc.b	$BE,$37,$FB,$BC,$2B,$C6,$2B,$EE,$AE,$38,$74,$8E,$07,$82,$FA,$3E
	.dc.b	$23,$86,$1F,$0E,$78,$47,$AE,$B9,$DF,$68,$61,$E2,$C3,$9D,$7C,$29
	.dc.b	$E6,$7C,$13,$B1,$C0,$F1,$5F,$67,$67,$AC,$DD,$75,$8D,$47,$19,$F1
	.dc.b	$BA,$39,$DF,$96,$B8,$36,$32,$6E,$78,$9E,$77,$A6,$1F,$06,$78,$4E
	.dc.b	$A3,$EA,$19,$9E,$4D,$8E,$55,$E0,$9C,$9F,$06,$32,$F8,$9A,$75,$7E
	.dc.b	$AF,$1B,$8F,$F0,$CD,$8C,$C6,$0F,$EE,$7F,$10,$F2,$FF,$C6,$52,$AE
	.dc.b	$0A,$75,$A7,$FA,$CA,$73,$18,$9E,$17,$D2,$F9,$D6,$9E,$B8,$7E,$7B
	.dc.b	$27,$AD,$FE,$04,$8C,$36,$39,$65,$F9,$6C,$BB,$56,$B2,$FC,$CD,$72
	.dc.b	$75,$87,$68,$E3,$1F,$A6,$74,$DC,$3B,$3F,$EC,$CB,$EB,$3C,$2F,$B4
	.dc.b	$EC,$EB,$84,$70,$48,$FD,$CD,$72,$8E,$98,$78,$DF,$F6,$CD,$F8,$AB
	.dc.b	$A9,$C2,$75,$3F,$3D,$B3,$B3,$FC,$52,$58,$EB,$7E,$3A,$E7,$3A,$C3
	.dc.b	$DB,$5C,$EB,$53,$98,$F9,$33,$AC,$3D,$11,$B3,$D9,$3B,$7F,$87,$C6
	.dc.b	$8D,$C7,$8E,$3F,$91,$CF,$3C,$3D,$75,$A8,$F1,$D6,$E3,$94,$FB,$A7
	.dc.b	$27,$1B,$F3,$DF,$C8,$8C,$B8,$E6,$FE,$B5,$F2,$3F,$CC,$76,$FE,$AF
	.dc.b	$D5,$6B,$FA,$7B,$6B,$71,$FD,$52,$B1,$CC,$AA,$62,$33,$DF,$2B,$FA
	.dc.b	$DB,$A6,$90,$FE,$B6,$E5,$FE,$BB,$F1,$DD,$88,$FF,$64,$F8,$FA,$E8
	.dc.b	$FE,$CF,$1E,$48,$F2,$FC,$74,$9F,$ED,$6F,$A3,$CB,$C4,$7F,$6A,$8B
	.dc.b	$B7,$F8,$02,$7D,$BF,$1D,$2B,$FB,$9B,$F0,$3E,$3F,$11,$FD,$CA,$2D
	.dc.b	$7F,$82,$27,$E5,$4B,$E1,$1E,$DF,$F8,$66,$06,$5C,$34,$D4,$70,$6C
	.dc.b	$74,$C3,$FD,$6B,$1F,$E5,$86,$BF,$E3,$69,$E0,$FF,$8E,$3E,$5D,$9F
	.dc.b	$66,$C7,$8D,$AF,$F8,$B5,$AF,$F9,$65,$CE,$3F,$DD,$61,$F3,$0F,$3B
	.dc.b	$DE,$D8,$FF,$35,$D7,$0E,$4F,$F4,$CC,$BD,$9D,$92,$BF,$CE,$31,$EA
	.dc.b	$C3,$B3,$B8,$CC,$ED,$FF,$29,$C6,$EB,$AC,$E6,$B8,$BE,$4F,$63,$E9
	.dc.b	$FC,$BB,$E3,$FE,$7B,$71,$FA,$8D,$7F,$C9,$CD,$7F,$CF,$4E,$6F,$E1
	.dc.b	$7C,$4F,$F8,$1B,$1F,$F5,$AB,$6B,$EE,$D6,$E7,$FD,$99,$F8,$B3,$E2
	.dc.b	$FF,$58,$FF,$A3,$6B,$F0,$78,$7F,$88,$7A,$D3,$99,$FF,$28,$BC,$5D
	.dc.b	$DF,$4B,$E7,$E9,$49,$C9,$FE,$50,$AE,$0F,$F9,$EB,$27,$C7,$7E,$59
	.dc.b	$B1,$FE,$A2,$6B,$BF,$8E,$D4,$C8,$D3,$C6,$31,$1D,$62,$E7,$F1,$A7
	.dc.b	$F8,$3F,$24,$AF,$F2,$EB,$FC,$2B,$E4,$F1,$1F,$E5,$C8,$B5,$FE,$72
	.dc.b	$9F,$B3,$9F,$D8,$25,$7D,$86,$BE,$DD,$FD,$87,$07,$FC,$D7,$F0,$7F
	.dc.b	$C6,$08,$FF,$9B,$BE,$16,$BF,$E6,$C8,$B8,$7F,$A0,$A6,$EF,$FE,$7C
	.dc.b	$9B,$5F,$E7,$49,$FE,$4D,$83,$E6,$7F,$92,$DC,$AF,$A7,$C3,$F8,$19
	.dc.b	$D9,$98,$FF,$3D,$CE,$1D,$D3,$D6,$EB,$F3,$D2,$73,$1E,$03,$0F,$FA
	.dc.b	$5A,$53,$98,$FB,$97,$F7,$C8,$FF,$B8,$3E,$67,$F7,$9A,$7E,$72,$73
	.dc.b	$3F,$87,$7F,$8D,$CE,$B1,$FC,$FC,$6A,$B8,$57,$39,$F8,$7F,$27,$23
	.dc.b	$D6,$FC,$6D,$BF,$43,$F1,$72,$7E,$2F,$2A,$3F,$4B,$E4,$7C,$CF,$94
	.dc.b	$FE,$A5,$F1,$5E,$1B,$FF,$55,$B7,$87,$F5,$93,$F4,$B9,$FF,$AC,$7C
	.dc.b	$87,$D3,$BF,$A2,$D7,$E3,$47,$D8,$E2,$91,$F4,$1F,$B7,$D5,$F7,$3A
	.dc.b	$A4,$FD,$47,$EF,$FC,$E7,$E0,$3E,$4A,$4F,$D8,$7F,$07,$F3,$9F,$DA
	.dc.b	$7F,$17,$FF,$70,$4F,$E0,$9A,$FF,$B3,$1B,$CB,$FC,$05,$7C,$E7,$E3
	.dc.b	$3E,$57,$F1,$BE,$A4,$BF,$C2,$C7,$E4,$3E,$53,$F2,$3F,$2B,$27,$F1
	.dc.b	$31,$F9,$4E,$69,$1F,$8C,$7F,$2F,$CD,$FD,$F7,$E7,$1F,$2F,$FD,$57
	.dc.b	$9D,$18,$BF,$F7,$84,$62,$75,$39,$9F,$EB,$67,$F5,$C9,$6A,$F9,$4F
	.dc.b	$85,$EC,$4A,$F8,$51,$F1,$7D,$E9,$1F,$15,$D3,$F1,$7E,$53,$ED,$F8
	.dc.b	$72,$3E,$54,$7A,$7C,$CF,$57,$CB,$C9,$FD,$7E,$9F,$D2,$5F,$F7,$CA
	.dc.b	$8D,$46,$1E,$4F,$07,$E2,$7A,$72,$30,$7D,$9F,$95,$FB,$84,$7E,$E7
	.dc.b	$CA,$D7,$EE,$45,$E8,$FB,$93,$7C,$6F,$8D,$C0,$AC,$5B,$87,$E4,$DF
	.dc.b	$D5,$F9,$79,$3E,$C7,$EC,$7C,$E6,$23,$DB,$17,$2F,$E9,$5A,$AB,$EF
	.dc.b	$78,$BF,$7F,$E3,$66,$1F,$1A,$32,$FC,$88,$CB,$F2,$23,$F0,$DC,$9F
	.dc.b	$7B,$E1,$7F,$BF,$69,$F8,$68,$C7,$AF,$F8,$1C,$BF,$01,$87,$F8,$7A
	.dc.b	$F7,$0F,$92,$DC,$FB,$E3,$6F,$FB,$72,$B9,$5E,$B0,$ED,$3C,$E7,$B5
	.dc.b	$F2,$87,$CD,$BD,$3F,$F5,$94,$73,$3F,$C5,$F1,$9F,$86,$E5,$18,$9D
	.dc.b	$E7,$FE,$19,$8E,$7F,$17,$D6,$F8,$47,$33,$BD,$70,$C7,$F4,$33,$AB
	.dc.b	$E4,$AF,$95,$6E,$FC,$18,$78,$9E,$6D,$AE,$11,$67,$83,$A3,$C5,$1D
	.dc.b	$F0,$F0,$DF,$6B,$E0,$4A,$DC,$65,$FD,$B9,$E2,$29,$D4,$CD,$7C,$07
	.dc.b	$7F,$0D,$FB,$BF,$9C,$FF,$2D,$36,$38,$B5,$DE,$A9,$FF,$30,$57,$F9
	.dc.b	$8F,$C1,$BE,$C7,$F3,$FB,$F0,$CE,$5B,$19,$7C,$6E,$F5,$F9,$AD,$D3
	.dc.b	$F8,$8F,$87,$FE,$2D,$EA,$F9,$53,$9B,$FC,$0B,$5F,$C3,$B6,$78,$47
	.dc.b	$CC,$8D,$61,$E3,$C3,$BE,$3F,$5E,$F7,$18,$BE,$13,$E0,$35,$E6,$B1
	.dc.b	$C4,$E8,$95,$F8,$FB,$E7,$97,$E2,$B2,$F4,$71,$3C,$FF,$F2,$C6,$3F
	.dc.b	$98,$8C,$7A,$5D,$72,$FC,$F4,$6F,$D6,$C6,$7F,$B6,$C3,$8D,$69,$B1
	.dc.b	$9B,$DE,$5F,$CE,$67,$FC,$21,$8F,$77,$71,$9F,$4B,$C1,$D7,$FE,$57
	.dc.b	$97,$A9,$8B,$D1,$FD,$C1,$C6,$5F,$C7,$65,$78,$D3,$DC,$D8,$7C,$AD
	.dc.b	$8C,$7A,$93,$1F,$EF,$7E,$D5,$C5,$9F,$8E,$B0,$F9,$66,$D7,$F5,$5F
	.dc.b	$9C,$71,$78,$5B,$1F,$E3,$65,$E7,$66,$64,$E5,$E1,$18,$C7,$C3,$3D
	.dc.b	$67,$18,$7F,$9E,$EF,$9C,$F1,$3F,$C2,$7C,$9E,$FF,$5B,$E0,$36,$32
	.dc.b	$4C,$3A,$9F,$27,$D6,$DC,$7F,$87,$1F,$90,$7C,$9F,$5A,$C7,$AF,$D1
	.dc.b	$F7,$1E,$FE,$9B,$1E,$F7,$0F,$F9,$5B,$E7,$3F,$C0,$6E,$4F,$89,$87
	.dc.b	$D7,$8C,$BE,$4E,$46,$9F,$ED,$CA,$E5,$E9,$7D,$BF,$5E,$47,$F8,$CB
	.dc.b	$D7,$B8,$71,$E7,$7A,$BF,$81,$3F,$05,$EB,$5B,$FF,$1C,$F2,$7E,$17
	.dc.b	$97,$F0,$CD,$8F,$F1,$EE,$FF,$89,$E0,$FC,$53,$63,$D9,$E8,$FA,$4F
	.dc.b	$57,$A3,$63,$FC,$EB,$F3,$3E,$CE,$8F,$F5,$93,$63,$F4,$9B,$7E,$5E
	.dc.b	$30,$FF,$77,$7F,$DB,$91,$CB,$D3,$D7,$BF,$B7,$86,$32,$7B,$E3,$70
	.dc.b	$F9,$58,$3D,$1E,$C7,$7F,$46,$DF,$9B,$D1,$F1,$1E,$3F,$66,$C7,$D9
	.dc.b	$F6,$3E,$B1,$F5,$FD,$7B,$1E,$EF,$4B,$D4,$7A,$FD,$4B,$1F,$EC,$8E
	.dc.b	$17,$CE,$7B,$36,$FC,$8C,$63,$87,$F7,$6E,$B8,$AD,$6E,$3F,$82,$B6
	.dc.b	$FF,$DB,$AD,$BF,$DB,$AA,$E3,$72,$BC,$53,$BB,$ED,$18,$9F,$A5,$ED
	.dc.b	$7D,$3F,$46,$CF,$DA,$6E,$3F,$31,$B7,$F7,$1F,$B5,$AF,$DB,$D6,$DF
	.dc.b	$E0,$9B,$7F,$51,$B7,$F8,$47,$F0,$5E,$C4,$AF,$C4,$FB,$16,$FE,$DC
	.dc.b	$DB,$FC,$5C,$DA,$FF,$33,$72,$7F,$90,$5F,$C8,$D7,$E3,$67,$56,$1F
	.dc.b	$37,$FE,$0B,$AF,$EE,$F1,$8F,$5B,$FA,$13,$FD,$F2,$74,$FF,$7C,$6B
	.dc.b	$98,$FE,$66,$F1,$9F,$F6,$AF,$CF,$FF,$16,$8F,$E1,$E6,$4F,$CA,$FC
	.dc.b	$E7,$E8,$4F,$D1,$7A,$36,$BF,$53,$16,$7F,$37,$37,$4F,$E8,$FF,$36
	.dc.b	$F9,$8F,$C4,$B7,$0F,$C8,$7A,$4F,$6F,$35,$27,$F2,$95,$BC,$7F,$27
	.dc.b	$7B,$9C,$D6,$E7,$F2,$F5,$BC,$3F,$AB,$DE,$E7,$35,$FD,$47,$A5,$F2
	.dc.b	$F7,$6C,$37,$2F,$F7,$CC,$E9,$B3,$A6,$9F,$5F,$1F,$A7,$BB,$E3,$FE
	.dc.b	$AE,$5F,$EA,$68,$C5,$7D,$A6,$C7,$D7,$9D,$BF,$81,$8C,$D7,$49,$FD
	.dc.b	$7F,$81,$F6,$A7,$37,$E1,$D7,$F0,$37,$F4,$31,$F9,$FF,$31,$F4,$19
	.dc.b	$5F,$31,$CB,$49,$97,$CA,$8C,$B6,$3D,$C7,$FB,$03,$E6,$BF,$C0,$F1
	.dc.b	$FE,$4A,$6B,$FE,$4E,$6B,$BF,$6F,$DA,$A3,$8E,$1F,$9A,$6D,$7A,$F6
	.dc.b	$FC,$BC,$62,$BF,$49,$F3,$4F,$0B,$9B,$6F,$F4,$AD,$9F,$F2,$EB,$5F
	.dc.b	$F3,$0B,$9C,$3F,$52,$D9,$F9,$5B,$FE,$9A,$31,$5F,$CF,$F8,$FF,$A1
	.dc.b	$1F,$D0,$FE,$5E,$DF,$F4,$A8,$B3,$FC,$F2,$6E,$1F,$E3,$7F,$49,$F2
	.dc.b	$CE,$0F,$F1,$13,$71,$F9,$B8,$7F,$B0,$E3,$F3,$51,$FE,$ED,$75,$1F
	.dc.b	$DC,$5B,$7D,$5B,$3B,$E5,$EB,$7D,$B7,$93,$DB,$1F,$1F,$0E,$32,$E9
	.dc.b	$57,$2E,$93,$6F,$C5,$F3,$5E,$83,$D3,$F3,$36,$3A,$68,$9C,$63,$E7
	.dc.b	$AB,$87,$9F,$E6,$B8,$E1,$C1,$B5,$ED,$6C,$EE,$7A,$D7,$F1,$AA,$FF
	.dc.b	$67,$0F,$9A,$7A,$3F,$CA,$63,$FC,$8B,$E1,$FF,$99,$E7,$31,$FE,$A2
	.dc.b	$F8,$2F,$E5,$D1,$88,$DE,$18,$BF,$F5,$2D,$59,$FD,$F5,$F3,$9C,$C7
	.dc.b	$85,$F0,$62,$9D,$4F,$73,$47,$2B,$E8,$FF,$BB,$4C,$9E,$22,$9E,$EC
	.dc.b	$3C,$D5,$D6,$78,$CE,$6F,$C3,$7E,$08,$E9,$7C,$23,$11,$B3,$C4,$76
	.dc.b	$18,$79,$30,$ED,$19,$8C,$4F,$28,$D9,$E8,$9D,$D7,$CB,$8E,$C7,$AD
	.dc.b	$C3,$C1,$F2,$61,$EA,$8C,$61,$EA,$8C,$61,$E0,$6C,$F7,$C3,$C5,$39
	.dc.b	$8F,$31,$99,$DD,$F6,$8C,$B4,$F1,$CF,$16,$C7,$13,$77,$E3,$69,$F2
	.dc.b	$23,$67,$38,$FC,$1B,$63,$F0,$1C,$6E,$3F,$C3,$B6,$33,$18,$3F,$31
	.dc.b	$F0,$CF,$1F,$FD,$3B,$3F,$D9,$67,$31,$8A,$E5,$2A,$E1,$A7,$F8,$26
	.dc.b	$73,$18,$9E,$47,$F8,$35,$C6,$1F,$A1,$C5,$E8,$FE,$A6,$30,$D8,$F0
	.dc.b	$E1,$F9,$8A,$E3,$1F,$AF,$38,$DF,$E7,$1A,$FE,$72,$36,$F8,$71,$F6
	.dc.b	$E1,$DA,$38,$61,$DD,$D6,$1D,$EB,$78,$77,$AD,$57,$F0,$06,$E1,$DA
	.dc.b	$3E,$25,$70,$7F,$C5,$95,$E4,$C3,$C1,$1E,$59,$D3,$C3,$0F,$1B,$0D
	.dc.b	$3E,$33,$13,$C8,$FE,$15,$7E,$13,$C5,$7D,$CF,$45,$71,$39,$9F,$C0
	.dc.b	$1D,$3E,$6C,$3D,$78,$79,$2B,$9C,$7C,$0A,$DD,$79,$CE,$77,$E3,$8E
	.dc.b	$F1,$87,$EC,$BD,$E7,$91,$DB,$0F,$25,$71,$9F,$11,$8A,$E3,$19,$8F
	.dc.b	$A0,$64,$FE,$B1,$F3,$3F,$C8,$B5,$FE,$B1,$F3,$56,$7F,$AE,$37,$9B
	.dc.b	$FB,$1D,$5E,$43,$66,$57,$F4,$C8,$D7,$54,$8F,$E9,$EE,$9F,$EC,$8E
	.dc.b	$5A,$7F,$54,$9E,$DF,$35,$23,$FA,$BB,$E1,$F9,$6C,$47,$F6,$D8,$BA
	.dc.b	$FF,$6D,$9F,$37,$CD,$4A,$FE,$BE,$FA,$7E,$63,$11,$FD,$D6,$2D,$7F
	.dc.b	$75,$9F,$77,$CD,$4A,$FE,$CE,$FC,$1F,$9A,$C4,$7F,$7D,$8B,$5F,$DF
	.dc.b	$67,$E5,$4B,$E3,$1E,$3E,$1B,$1F,$E0,$A3,$51,$C3,$0C,$C7,$F9,$FA
	.dc.b	$6B,$FD,$9C,$D9,$FE,$8E,$72,$FF,$6B,$6B,$FE,$95,$F9,$A6,$7E,$46
	.dc.b	$C7,$85,$AF,$FA,$33,$E6,$BF,$8B,$E5,$FE,$98,$9C,$CF,$85,$AF,$F7
	.dc.b	$96,$BD,$EB,$FD,$42,$A9,$EA,$7C,$46,$C7,$C1,$6B,$E1,$F5,$B5,$32
	.dc.b	$34,$F1,$8C,$47,$9E,$2E,$1F,$EA,$EA,$FF,$6C,$4F,$CD,$F9,$A7,$C7
	.dc.b	$FD,$46,$C6,$63,$EB,$FC,$4B,$E0,$75,$1E,$AA,$DD,$7D,$10,$E1,$FA
	.dc.b	$8D,$7F,$D8,$9D,$0F,$96,$7C,$5F,$EA,$D7,$C6,$53,$11,$2B,$FC,$81
	.dc.b	$3A,$6C,$E5,$FF,$60,$E4,$F8,$EF,$49,$23,$F1,$90,$F8,$EC,$47,$09
	.dc.b	$BB,$7F,$B5,$2B,$53,$C2,$73,$DD,$FE,$35,$9E,$D3,$D2,$FF,$81,$72
	.dc.b	$48,$FF,$25,$BF,$C1,$7E,$4F,$11,$FE,$61,$8B,$8F,$F9,$86,$7E,$96
	.dc.b	$92,$BE,$B3,$F4,$BD,$7C,$47,$CF,$DF,$E4,$BF,$EE,$C6,$B9,$9F,$F7
	.dc.b	$7E,$FF,$E7,$8F,$9A,$FE,$2C,$49,$FF,$2F,$FC,$D5,$7F,$CE,$91,$72
	.dc.b	$FF,$3A,$4D,$DB,$FC,$F5,$36,$BF,$CF,$D3,$FC,$8F,$07,$CC,$FF,$23
	.dc.b	$B9,$5F,$E7,$79,$CB,$67,$2F,$FA,$0E,$79,$47,$3F,$5F,$A7,$CC,$CA
	.dc.b	$E5,$19,$7A,$46,$56,$5F,$E9,$7C,$15,$F9,$AB,$95,$F8,$27,$F8,$9F
	.dc.b	$CD,$31,$CC,$FE,$2A,$6E,$19,$80,$AC,$3B,$3A,$10,$F1,$D7,$F9,$EE
	.dc.b	$37,$3E,$5F,$96,$F9,$04,$AF,$8D,$CF,$FA,$3F,$8C,$91,$FA,$07,$F4
	.dc.b	$DE,$5F,$CF,$F9,$69,$3F,$35,$FA,$5C,$7F,$4F,$A2,$4F,$D1,$7E,$B7
	.dc.b	$3F,$D7,$E0,$93,$F5,$5F,$B5,$D3,$F6,$FA,$64,$FD,$97,$EF,$7C,$B7
	.dc.b	$DF,$F9,$39,$3F,$75,$FC,$17,$C9,$FF,$65,$F9,$AF,$F6,$F4,$FF,$AD
	.dc.b	$DA,$FC,$41,$D4,$8C,$F1,$FF,$B3,$D7,$C7,$7E,$37,$E4,$FF,$1D,$E9
	.dc.b	$4C,$3F,$19,$1F,$91,$F9,$3F,$C9,$78,$72,$7F,$21,$1F,$95,$E6,$91
	.dc.b	$F9,$37,$F3,$1C,$DF,$DE,$FE,$69,$F2,$D4,$95,$F9,$93,$D9,$F2,$1E
	.dc.b	$DF,$62,$47,$B6,$3E,$07,$B9,$23,$E0,$3A,$7E,$07,$0F,$BF,$E4,$3E
	.dc.b	$31,$23,$67,$FB,$77,$1F,$E8,$3C,$5F,$EF,$16,$F8,$1F,$CA,$9B,$3F
	.dc.b	$29,$FE,$7A,$F6,$5F,$2B,$ED,$24,$75,$F3,$2B,$BF,$47,$E1,$18,$C3
	.dc.b	$E9,$7C,$87,$DB,$24,$FD,$BF,$90,$AF,$B3,$A1,$38,$AF,$B5,$57,$3F
	.dc.b	$E8,$35,$AB,$FC,$25,$FE,$1E,$BE,$97,$CA,$C9,$F8,$4F,$D5,$F9,$5C
	.dc.b	$47,$B6,$2E,$3F,$16,$7E,$D7,$CA,$CA,$F9,$4F,$DD,$EB,$C4,$7C,$5F
	.dc.b	$8C,$FC,$01,$8B,$F9,$DC,$9F,$EF,$1F,$84,$7C,$8F,$CF,$8C,$36,$FF
	.dc.b	$81,$CE,$9F,$A5,$F2,$3A,$7E,$A5,$5E,$67,$2A,$30,$D9,$E7,$87,$DD
	.dc.b	$BE,$50,$F9,$97,$A8,$F8,$6D,$CF,$F1,$49,$CC,$62,$38,$4E,$DF,$F7
	.dc.b	$2D,$FF,$1A,$7A,$7C,$5F,$6B,$E3,$1A,$3A,$57,$1D,$3D,$D1,$D3,$0E
	.dc.b	$D4,$BE,$57,$BC,$3B,$4F,$4C,$38,$B6,$33,$5C,$2F,$9D,$F1,$C3,$C3
	.dc.b	$1D,$A7,$51,$E5,$3C,$45,$BE,$8F,$C3,$8D,$BF,$E4,$58,$F6,$B6,$BE
	.dc.b	$23,$6B,$FC,$34,$FF,$B4,$F3,$7C,$D6,$3E,$23,$EC,$36,$33,$3F,$E2
	.dc.b	$19,$CB,$FE,$23,$C3,$FC,$BD,$E0,$F4,$6C,$67,$D9,$4E,$5B,$19,$8F
	.dc.b	$9B,$18,$D7,$E8,$F8,$7F,$E2,$8E,$BC,$D4,$BF,$85,$7B,$DF,$EF,$5E
	.dc.b	$E7,$35,$BC,$3E,$9D,$EE,$7F,$AE,$D6,$E7,$35,$BB,$FB,$77,$BB,$5E
	.dc.b	$E7,$35,$B9,$FC,$A5,$6E,$B5,$18,$9F,$CC,$F1,$3E,$57,$FC,$13,$96
	.dc.b	$1B,$5F,$93,$97,$CE,$FF,$77,$AD,$7A,$DF,$DD,$74,$FD,$95,$73,$8C
	.dc.b	$69,$FE,$50,$C7,$9D,$6B,$0E,$58,$F5,$9C,$D6,$FD,$7F,$F0,$07,$3F
	.dc.b	$F8,$07,$2F,$0F,$C3,$E3,$4D,$1F,$E0,$49,$D6,$49,$D6,$5F,$D7,$B2
	.dc.b	$FF,$09,$DE,$CF,$F3,$1D,$6D,$C4,$EF,$1E,$38,$F8,$27,$FD,$49,$F8
	.dc.b	$C3,$C5,$CF,$53,$C7,$EF,$8A,$B5,$7F,$D2,$E7,$31,$89,$FF,$3C,$35
	.dc.b	$FF,$55,$35,$FE,$0C,$D7,$FC,$43,$F6,$1F,$C4,$6B,$CF,$D0,$FA,$F7
	.dc.b	$B5,$E2,$AD,$5F,$F7,$FA,$D4,$E7,$AF,$E2,$C6,$30,$FF,$60,$5F,$0E
	.dc.b	$2F,$F1,$84,$E7,$0F,$F0,$40,$9D,$57,$F1,$B6,$C7,$EE,$2B,$E7,$BC
	.dc.b	$23,$35,$F1,$F3,$F9,$3A,$F8,$9F,$95,$87,$F7,$18,$F9,$7F,$88,$7D
	.dc.b	$5C,$3D,$55,$A7,$C7,$19,$C7,$CF,$39,$7F,$C8,$B1,$9B,$FE,$60,$DA
	.dc.b	$FE,$1B,$83,$BF,$BF,$27,$E0,$3E,$2B,$FB,$CE,$1F,$E0,$3F,$33,$1E
	.dc.b	$67,$E0,$1F,$2F,$C2,$FA,$7C,$9B,$5F,$53,$64,$7C,$EC,$FF,$01,$B7
	.dc.b	$92,$F1,$BF,$0D,$BF,$C2,$BD,$2F,$73,$FC,$BB,$1F,$A9,$6B,$F9,$B6
	.dc.b	$F4,$7E,$E5,$CC,$3F,$03,$FF,$5D,$EB,$F8,$6F,$C6,$7E,$23,$FD,$46
	.dc.b	$33,$3F,$85,$9C,$BF,$A7,$F7,$7A,$E3,$FE,$BB,$BF,$F6,$76,$3F,$C1
	.dc.b	$E7,$27,$8B,$E1,$3F,$23,$3F,$AA,$9F,$F6,$9F,$4F,$EA,$A7,$FD,$BF
	.dc.b	$39,$77,$87,$E3,$32,$FE,$4D,$B7,$E4,$33,$F9,$37,$FB,$4A,$F6,$7B
	.dc.b	$BF,$98,$9C,$C7,$F8,$F2,$DE,$BF,$EE,$1B,$7E,$35,$EC,$66,$7F,$7B
	.dc.b	$B7,$82,$BF,$81,$F0,$E3,$2F,$E0,$3E,$0A,$3F,$7B,$8F,$FB,$CE,$BF
	.dc.b	$C8,$DC,$98,$F6,$3F,$81,$BD,$3D,$CF,$E0,$BE,$A5,$AF,$F2,$63,$E0
	.dc.b	$F6,$FA,$F6,$59,$F9,$19,$A3,$BF,$95,$F9,$3F,$56,$D7,$CD,$94,$7A
	.dc.b	$B8,$3F,$89,$F9,$16,$BF,$CB,$4F,$F1,$4F,$9B,$FF,$03,$47,$6F,$8A
	.dc.b	$C7,$A7,$EE,$6C,$64,$FA,$13,$C3,$9B,$FC,$D3,$A2,$B7,$1F,$E1,$B3
	.dc.b	$FD,$6B,$3B,$70,$D7,$FC,$40,$D7,$F9,$A7,$B9,$F4,$8F,$A7,$AD,$C3
	.dc.b	$F1,$5C,$C9,$C5,$7E,$9A,$AF,$4F,$F4,$5A,$FF,$62,$DF,$0A,$FB,$BE
	.dc.b	$7F,$DE,$64,$64,$EB,$18,$8F,$F5,$25,$3E,$6D,$F3,$3F,$D2,$A3,$0D
	.dc.b	$F8,$3F,$E3,$F1,$8C,$3F,$A4,$B6,$F9,$DF,$29,$CD,$70,$8C,$4E,$AF
	.dc.b	$B4,$62,$FF,$B0,$D6,$E7,$FA,$9F,$53,$E2,$FF,$2B,$F1,$F8,$6E,$5F
	.dc.b	$B1,$9D,$3F,$B8,$6E,$5F,$5F,$C2,$FC,$E5,$7F,$0A,$9C,$CF,$F9,$56
	.dc.b	$7F,$41,$EF,$F8,$0F,$0F,$C0,$5C,$3B,$45,$AF,$F5,$6D,$AB,$1A,$7F
	.dc.b	$BD,$B8,$BF,$98,$2F,$A3,$FE,$22,$E7,$CD,$F9,$D1,$8C,$BF,$0F,$18
	.dc.b	$F8,$5F,$CD,$36,$3F,$2A,$FF,$20,$BF,$E8,$D9,$7E,$42,$30,$7F,$26
	.dc.b	$FC,$0D,$C3,$FD,$F1,$B7,$FC,$5D,$FE,$27,$97,$F8,$3D,$CD,$FF,$B2
	.dc.b	$1B,$19,$AE,$5E,$47,$37,$F5,$DE,$F7,$F8,$23,$C8,$7B,$78,$CB,$F9
	.dc.b	$24,$61,$FD,$6C,$FA,$48,$FF,$98,$B4,$FF,$47,$BF,$E6,$4F,$9D,$FE
	.dc.b	$D7,$97,$F8,$93,$E9,$3F,$80,$CF,$F2,$59,$CB,$FE,$23,$8C,$EB,$FD
	.dc.b	$E2,$72,$A9,$3A,$55,$FE,$95,$AD,$C6,$BC,$2F,$F0,$EC,$7F,$7C,$BF
	.dc.b	$E0,$0D,$78,$7B,$7F,$D4,$A3,$10,$0F,$F3,$47,$F5,$FC,$5F,$CC,$5D
	.dc.b	$46,$6F,$FC,$83,$39,$9E,$3E,$1F,$52,$63,$F1,$5F,$EC,$3F,$3E,$F9
	.dc.b	$4F,$D5,$FC,$A7,$F8,$D2,$31,$5F,$B6,$EE,$4E,$6B,$89,$27,$FD,$27
	.dc.b	$5C,$30,$FD,$47,$4B,$E1,$FF,$C8,$33,$FB,$46,$D7,$EC,$5F,$D6,$11
	.dc.b	$FD,$93,$FA,$FF,$86,$FE,$9B,$8F,$EF,$1B,$3F,$B9,$6E,$1F,$AC,$6E
	.dc.b	$BF,$E6,$3C,$5F,$0F,$FE,$52,$F9,$9F,$E0,$AD,$C3,$F8,$13,$FB,$D2
	.dc.b	$3F,$C0,$DF,$E0,$1E,$E7,$F9,$6B,$DF,$FE,$1A,$D9,$FE,$14,$E9,$FE
	.dc.b	$14,$DC,$BE,$1F,$B7,$C2,$7F,$88,$36,$3F,$81,$B7,$2F,$43,$67,$FA
	.dc.b	$44,$BE,$0F,$F0,$DF,$33,$FD,$DA,$3F,$CF,$9E,$F7,$E5,$14,$65,$FF
	.dc.b	$75,$4F,$0F,$67,$FD,$D9,$8A,$F8,$5E,$EF,$F2,$57,$F9,$3D,$79,$5E
	.dc.b	$F1,$C6,$3E,$4A,$B7,$E2,$9A,$F8,$0F,$1C,$F8,$31,$EE,$F1,$8C,$9C
	.dc.b	$4A,$71,$C7,$C3,$3C,$23,$FB,$55,$FA,$DE,$13,$E9,$7E,$30,$C7,$8E
	.dc.b	$1E,$48,$F6,$99,$C3,$C9,$19,$8E,$35,$C1,$F1,$9C,$1C,$1D,$EF,$C5
	.dc.b	$1F,$C1,$F0,$F3,$3E,$9B,$EA,$F2,$3A,$B8,$AE,$51,$F0,$31,$F9,$CE
	.dc.b	$EF,$AC,$62,$76,$78,$AB,$E5,$36,$BE,$29,$CE,$FC,$6F,$A6,$7C,$07
	.dc.b	$02,$BC,$27,$93,$E9,$3E,$6C,$6D,$A6,$DE,$2E,$CF,$85,$87,$C4,$7C
	.dc.b	$07,$03,$F0,$9A,$7D,$57,$CE,$D7,$39,$7E,$16,$31,$F0,$FF,$EC,$9C
	.dc.b	$BF,$93,$46,$34,$FE,$6D,$68,$C4,$EA,$FF,$96,$4A,$F9,$3B,$8C,$C6
	.dc.b	$32,$FD,$46,$5F,$E2,$BA,$FB,$AD,$7F,$2B,$3B,$8E,$38,$F8,$AB,$8F
	.dc.b	$9D,$E0,$6D,$FE,$1A,$F8,$3F,$CF,$DA,$FF,$9C,$27,$85,$FA,$0F,$05
	.dc.b	$7E,$EE,$B6,$70,$8C,$9B,$9E,$35,$D6,$FC,$93,$E6,$C7,$D7,$5A,$AF
	.dc.b	$01,$E5,$AD,$BC,$67,$6C,$3E,$0C,$69,$F9,$8D,$AF,$1C,$7F,$5B,$6B
	.dc.b	$A9,$EC,$73,$C3,$E2,$C7,$48,$CB,$FC,$0A,$78,$9C,$B0,$F8,$37,$F8
	.dc.b	$0C,$78,$92,$7D,$D5,$FB,$D3,$D4,$73,$C7,$E0,$47,$3A,$F5,$5F,$CB
	.dc.b	$9C,$9F,$15,$EF,$87,$F1,$CA,$FC,$16,$1D,$F0,$EA,$FD,$5C,$3E,$16
	.dc.b	$1C,$A7,$D7,$8F,$F3,$7F,$23,$FD,$4D,$AF,$F3,$2D,$7F,$C0,$11,$FE
	.dc.b	$A9,$F5,$F1,$68,$C3,$FE,$57,$AC,$C7,$28,$FF,$2D,$4E,$38,$BF,$DD
	.dc.b	$71,$B3,$27,$F9,$76,$3F,$A5,$BD,$63,$6F,$F7,$88,$CC,$F8,$23,$FD
	.dc.b	$37,$E6,$78,$3D,$1F,$F3,$FC,$74,$3F,$CB,$7E,$9F,$0C,$0D,$24,$7F
	.dc.b	$9B,$5E,$5F,$3D,$23,$FC,$E5,$1E,$9D,$FB,$75,$C9,$FF,$3B,$46,$5F
	.dc.b	$F6,$0C,$65,$FF,$35,$C7,$C4,$FC,$6F,$CB,$1F,$E7,$96,$BF,$E7,$76
	.dc.b	$BF,$41,$AF,$CF,$78,$F1,$FF,$6B,$D1,$9B,$D5,$F8,$47,$7D,$7E,$BE
	.dc.b	$0F,$C6,$DF,$2B,$F3,$A3,$0D,$8C,$C7,$D2,$6B,$FD,$F9,$B8,$FF,$7B
	.dc.b	$8F,$F2,$8F,$C9,$76,$F1,$CF,$82,$B7,$C9,$FC,$7B,$CB,$FA,$37,$F7
	.dc.b	$9A,$FF,$82,$9A,$FF,$B4,$DA,$F1,$E1,$F5,$FD,$6F,$87,$2F,$13,$6F
	.dc.b	$55,$FE,$57,$9F,$37,$CD,$7F,$C0,$18,$FC,$6B,$5F,$F0,$CB,$5F,$F7
	.dc.b	$13,$5F,$E6,$AE,$E3,$E0,$7C,$D7,$A1,$B8,$7C,$E9,$D5,$F0,$E6,$FA
	.dc.b	$37,$B8,$FE,$7E,$DC,$3F,$CD,$95,$C1,$B9,$7D,$58,$FF,$48,$C7,$A2
	.dc.b	$30,$47,$DD,$5D,$3C,$4E,$E7,$8E,$7F,$67,$F8,$2B,$D6,$A4,$BE,$27
	.dc.b	$EE,$E3,$FD,$80,$E3,$95,$1C,$63,$0D,$7F,$D2,$58,$7F,$06,$8C,$6B
	.dc.b	$F9,$57,$9C,$E8,$FF,$45,$7C,$23,$E5,$FF,$24,$D8,$FE,$02,$D7,$FC
	.dc.b	$9A,$D7,$FB,$38,$75,$5F,$93,$F8,$2F,$F2,$1E,$1F,$DB,$1B,$5E,$0B
	.dc.b	$FD,$1D,$F0,$AD,$78,$5F,$A6,$BD,$DF,$5B,$FD,$C7,$44,$8F,$89,$D5
	.dc.b	$FA,$F6,$F6,$E5,$F7,$6A,$9D,$46,$1B,$A7,$EE,$55,$7F,$05,$FA,$1F
	.dc.b	$E3,$B0,$69,$2B,$FC,$DF,$7D,$63,$F9,$07,$A7,$27,$CB,$1E,$BF,$87
	.dc.b	$91,$F7,$86,$88,$F8,$1D,$5A,$9F,$AC,$6A,$31,$9F,$CD,$9B,$BF,$CD
	.dc.b	$9B,$5F,$EC,$5B,$FE,$02,$FC,$6F,$A1,$FE,$40,$47,$E4,$FD,$0D,$7E
	.dc.b	$24,$5C,$7E,$24,$FF,$27,$F9,$F9,$5F,$41,$FE,$55,$F4,$18,$8F,$99
	.dc.b	$16,$BE,$64,$FD,$EF,$8B,$FC,$69,$2B,$F1,$AD,$7D,$F8,$FE,$0D,$F5
	.dc.b	$71,$F1,$7E,$47,$2A,$BF,$0D,$32,$73,$1F,$CF,$7B,$3F,$A9,$7D,$14
	.dc.b	$9F,$C3,$C7,$F3,$2F,$C1,$7F,$AD,$E7,$FC,$E7,$E2,$6F,$0F,$F6,$E6
	.dc.b	$DF,$FB,$BE,$AF,$2F,$E2,$FE,$27,$FA,$F7,$6C,$AF,$C9,$C7,$F4,$DF
	.dc.b	$57,$FB,$17,$3C,$9F,$CB,$C7,$87,$E2,$FF,$B3,$72,$C9,$FC,$26,$49
	.dc.b	$FE,$AD,$D3,$FD,$AF,$C7,$95,$FA,$08,$FE,$FB,$F8,$E3,$E0,$69,$27
	.dc.b	$F8,$44,$7C,$B7,$FB,$7F,$7F,$F8,$23,$E4,$5F,$8E,$B4,$C3,$6B,$F2
	.dc.b	$DE,$47,$E9,$B9,$3F,$C2,$3F,$35,$FC,$97,$90,$D2,$77,$7F,$6D,$DF
	.dc.b	$F5,$24,$C3,$F4,$FE,$17,$F5,$A7,$F8,$4E,$2E,$AF,$0D,$C9,$F0,$C7
	.dc.b	$F7,$7C,$BF,$66,$48,$E6,$FF,$85,$B5,$7C,$63,$E8,$FF,$B7,$FD,$AB
	.dc.b	$E1,$29,$F0,$67,$58,$7E,$CE,$74,$FF,$32,$BF,$DA,$F0,$FC,$7F,$0B
	.dc.b	$E2,$4F,$ED,$2B,$6F,$F0,$EE,$DF,$E9,$5C,$5F,$E0,$2F,$13,$FC,$44
	.dc.b	$7F,$8C,$A3,$FA,$74,$AB,$EA,$13,$C2,$FA,$9F,$8F,$3E,$97,$D3,$1E
	.dc.b	$1E,$29,$7E,$78,$AF,$F1,$EF,$51,$E3,$FF,$8C,$4C,$F8,$9F,$E0,$5C
	.dc.b	$9F,$47,$F1,$09,$7E,$FF,$7D,$F0,$38,$8F,$E8,$DC,$6F,$4D,$F2,$BF
	.dc.b	$89,$6C,$FC,$9C,$38,$64,$E9,$4B,$0E,$58,$71,$37,$3F,$EB,$0F,$9D
	.dc.b	$7C,$07,$F8,$8D,$B7,$FE,$52,$6B,$FD,$40,$38,$7F,$A9,$BA,$C3,$FC
	.dc.b	$69,$BB,$87,$FC,$A8,$7F,$8C,$A7,$58,$7F,$9B,$3E,$B3,$F8,$87,$3F
	.dc.b	$F8,$EB,$B3,$FC,$B4,$7F,$9A,$A3,$26,$B0,$FC,$4E,$5F,$46,$75,$C9
	.dc.b	$BF,$17,$FD,$97,$C7,$FE,$2E,$C5,$8F,$ED,$F0,$F8,$B3,$98,$C4,$23
	.dc.b	$1F,$1D,$CB,$C0,$F3,$61,$6B,$FD,$01,$7C,$2F,$94,$E7,$0E,$F8,$6E
	.dc.b	$FF,$6F,$E3,$FF,$B8,$5E,$D8,$FF,$B8,$3C,$44,$E7,$1F,$F2,$1C,$63
	.dc.b	$E1,$3C,$F5,$BF,$1F,$A4,$E7,$0E,$67,$E0,$1C,$5F,$2C,$3A,$5F,$1A
	.dc.b	$E6,$6E,$32,$DC,$32,$FA,$1B,$7F,$DE,$AF,$A4,$62,$B8,$D6,$CF,$19
	.dc.b	$FC,$5B,$0F,$3C,$6A,$FC,$B8,$7C,$18,$C4,$EA,$FF,$D4,$35,$ED,$BF
	.dc.b	$F0,$FB,$5E,$11,$FE,$A6,$F8,$47,$D0,$FE,$7E,$7F,$2D,$18,$E8,$D4
	.dc.b	$E6,$7F,$C5,$33,$97,$F3,$96,$F4,$3F,$8D,$C6,$1A,$F8,$9A,$FD,$96
	.dc.b	$BF,$45,$E3,$A7,$DA,$6E,$9F,$B4,$97,$CB,$FF,$6B,$E5,$F9,$B5,$A9
	.dc.b	$E5,$E3,$60,$FA,$75,$B8,$F9,$B8,$F9,$E7,$51,$8A,$E4,$F6,$F4,$FF
	.dc.b	$82,$39,$9D,$E5,$E5,$7A,$CE,$A3,$13,$B2,$3F,$E9,$6C,$3C,$4E,$8F
	.dc.b	$D7,$7C,$BB,$E8,$79,$69,$FE,$E8,$8C,$69,$F9,$0E,$4F,$F5,$BB,$AC
	.dc.b	$7F,$DD,$53,$A7,$F3,$72,$8C,$61,$DB,$0E,$13,$CA,$FA,$56,$B0,$F0
	.dc.b	$79,$5F,$03,$B7,$F8,$9F,$CA,$5B,$C9,$FE,$7D,$EC,$46,$74,$F1,$DD
	.dc.b	$CB,$C4,$DC,$BF,$A3,$61,$FE,$85,$7F,$AF,$3F,$E4,$DE,$D7,$CE,$7F
	.dc.b	$B1,$70,$FE,$5C,$D7,$F9,$AC,$6A,$7F,$97,$4E,$67,$F9,$6B,$E3,$6C
	.dc.b	$77,$ED,$C5,$71,$9F,$F2,$BF,$CF,$BA,$7D,$5D,$2F,$A0,$76,$F3,$AD
	.dc.b	$65,$FD,$BE,$72,$FF,$99,$63,$38,$7F,$3F,$6B,$FE,$90,$6B,$F9,$B6
	.dc.b	$BF,$D4,$1E,$3C,$3F,$EE,$56,$CF,$F5,$4C,$13,$F2,$2B,$D3,$8F,$A3
	.dc.b	$C2,$FE,$A7,$7F,$E0,$36,$CF,$F9,$D3,$E8,$3E,$BC,$E6,$7E,$34,$62
	.dc.b	$7F,$AB,$35,$FF,$53,$B5,$FD,$93,$5F,$EC,$D8,$3E,$3B,$9F,$1F,$CA
	.dc.b	$8C,$61,$FE,$89,$BE,$D9,$7F,$96,$7E,$83,$F0,$DE,$71,$F5,$7C,$66
	.dc.b	$1F,$2F,$BB,$F7,$6C,$D2,$61,$FD,$B7,$1F,$EF,$1C,$9E,$13,$F2,$1C
	.dc.b	$56,$7F,$21,$F4,$1F,$93,$3F,$29,$E3,$58,$FF,$4F,$7D,$03,$E7,$3F
	.dc.b	$18,$D8,$FE,$F5,$C4,$8C,$3F,$DF,$5A,$FF,$17,$EB,$FD,$97,$C1,$7E
	.dc.b	$84,$FD,$17,$C1,$DB,$FF,$73,$7C,$1F,$E2,$AF,$FD,$82,$DE,$4F,$E2
	.dc.b	$38,$7F,$B2,$EB,$7A,$FE,$77,$2F,$E3,$98,$7F,$B7,$1A,$FF,$27,$6B
	.dc.b	$FE,$1C,$6D,$7A,$7E,$5B,$F6,$D0,$6F,$C3,$61,$FD,$C7,$8D,$FB,$BF
	.dc.b	$0B,$E8,$E5,$FB,$98,$C7,$27,$8F,$D1,$F8,$E4,$9F,$8F,$8F,$F8,$C9
	.dc.b	$FF,$68,$F8,$5F,$EC,$A7,$F8,$17,$B5,$FC,$17,$E9,$7F,$83,$3F,$C1
	.dc.b	$FE,$87,$C1,$DB,$6B,$F8,$83,$F0,$7C,$AF,$AA,$48,$EE,$FF,$07,$E0
	.dc.b	$FE,$45,$BF,$F1,$13,$F8,$97,$05,$AF,$F6,$E7,$CE,$3D,$1F,$77,$C4
	.dc.b	$7F,$09,$8C,$4F,$F1,$F8,$C4,$FF,$0F,$9C,$F2,$7F,$8F,$5A,$FF,$57
	.dc.b	$0E,$1F,$EB,$4F,$0E,$6F,$E9,$B1,$A9,$FE,$4D,$C5,$DB,$0F,$E4,$DF
	.dc.b	$2D,$FC,$A4,$FE,$55,$AD,$8C,$92,$FF,$82,$BF,$D2,$63,$51,$83,$FA
	.dc.b	$39,$FC,$CC,$FE,$69,$E7,$D9,$FE,$69,$14,$E1,$B7,$F4,$B9,$C6,$5F
	.dc.b	$D1,$7E,$65,$F2,$7F,$D0,$61,$18,$C3,$FC,$D7,$1F,$D2,$FB,$FF,$8D
	.dc.b	$E3,$FC,$6A,$B4,$F9,$67,$8D,$F3,$9C,$C7,$1F,$27,$13,$30,$EA,$76
	.dc.b	$AD,$BE,$0C,$3B,$DF,$48,$DE,$0C,$BF,$85,$FD,$07,$E7,$7E,$87,$FA
	.dc.b	$C7,$5A,$31,$7C,$2B,$6D,$9D,$37,$E9,$78,$D7,$2F,$91,$93,$BF,$A2
	.dc.b	$B9,$7C,$CC,$15,$8C,$FF,$7F,$0A,$C5,$FE,$23,$CC,$F9,$DF,$3D,$FC
	.dc.b	$42,$FF,$13,$CF,$FC,$B2,$BE,$7C,$E5,$E0,$7F,$18,$8F,$E7,$19,$7E
	.dc.b	$52,$FF,$34,$49,$FD,$1B,$F6,$FE,$AF,$CB,$18,$3E,$17,$D1,$FF,$BE
	.dc.b	$5B,$C3,$FD,$FE,$3F,$8C,$47,$FA,$AB,$81,$1B,$C3,$F1,$6D,$CF,$F8
	.dc.b	$EB,$6B,$83,$CA,$7E,$B3,$97,$FD,$6B,$97,$FA,$02,$7E,$A7,$D1,$FF
	.dc.b	$94,$5F,$DD,$FB,$FF,$E1,$18,$C5,$7E,$84,$FD,$9C,$7E,$F4,$D7,$C7
	.dc.b	$5A,$FC,$2C,$BE,$8B,$FB,$76,$5F,$E2,$3F,$9C,$7D,$1F,$F8,$36,$73
	.dc.b	$7F,$9E,$9C,$BC,$A3,$FB,$B5,$FF,$03,$6B,$FE,$5B,$6B,$FE,$D0,$6B
	.dc.b	$E1,$F7,$3B,$1F,$A3,$F9,$0B,$5D,$7C,$4F,$C1,$F9,$7E,$36,$C6,$49
	.dc.b	$7F,$AD,$E3,$7D,$2E,$E7,$DF,$F4,$8F,$A1,$FD,$87,$CC,$FF,$8A,$63
	.dc.b	$13,$FA,$8E,$0F,$DE,$CE,$6A,$4E,$7C,$0D,$E5,$F3,$F0,$B8,$7E,$73
	.dc.b	$E7,$7F,$9A,$76,$9A,$4C,$3F,$DC,$DD,$3F,$DF,$A3,$F6,$67,$ED,$3D
	.dc.b	$BB,$3F,$B1,$FA,$4F,$EA,$9A,$FE,$E1,$B1,$92,$67,$FA,$FE,$6D,$47
	.dc.b	$EE,$7E,$43,$C6,$47,$C3,$9A,$31,$5F,$B7,$7F,$82,$F0,$7F,$06,$7F
	.dc.b	$83,$FB,$FE,$C2,$5F,$26,$D7,$EF,$DF,$E1,$BF,$01,$C2,$7F,$85,$36
	.dc.b	$7F,$82,$37,$7F,$E0,$6F,$09,$FE,$21,$F1,$76,$BF,$38,$7F,$BD,$07
	.dc.b	$D1,$3D,$1F,$F5,$46,$0F,$A3,$FF,$4D,$CE,$6B,$FA,$B7,$C5,$1E,$37
	.dc.b	$C2,$B6,$FF,$5B,$8F,$F7,$F0,$FA,$4E,$4F,$F9,$52,$7F,$D4,$F1,$88
	.dc.b	$F2,$3A,$C3,$C5,$97,$78,$D4,$F4,$75,$5C,$AF,$C1,$86,$DF,$25,$DB
	.dc.b	$F4,$9E,$83,$FA,$7D,$7A,$1C,$E3,$D2,$7D,$CF,$82,$F9,$4F,$A6,$6B
	.dc.b	$E8,$BE,$EF,$BD,$F1,$B4,$EF,$1C,$1F,$54,$70,$C3,$B4,$79,$4E,$C5
	.dc.b	$3E,$C5,$F9,$5C,$1C,$8E,$A7,$73,$FD,$C5,$7E,$8C,$7E,$5C,$64,$F4
	.dc.b	$1C,$2B,$C2,$75,$2D,$F9,$4F,$5B,$97,$DB,$7E,$33,$D2,$57,$67,$19
	.dc.b	$FA,$33,$C5,$F9,$F1,$F2,$27,$8B,$4E,$A5,$8D,$9E,$E8,$F8,$FE,$77
	.dc.b	$E6,$A3,$EE,$B5,$F8,$27,$FA,$DE,$3E,$9C,$67,$E3,$9A,$FF,$A9,$AB
	.dc.b	$47,$E7,$63,$38,$FE,$7E,$B8,$56,$DB,$19,$BF,$D2,$5A,$F8,$BF,$6F
	.dc.b	$07,$A5,$AA,$FD,$1C,$62,$7F,$1E,$F2,$AE,$33,$FA,$86,$C7,$63,$84
	.dc.b	$F5,$AE,$91,$C7,$1F,$65,$F1,$7D,$D9,$78,$2B,$CE,$DB,$FC,$EE,$1E
	.dc.b	$88,$E1,$8F,$92,$7F,$80,$63,$E0,$C7,$84,$7E,$5E,$FC,$AF,$C9,$C7
	.dc.b	$BD,$79,$F1,$EE,$D9,$EF,$3A,$78,$61,$FA,$E9,$F6,$E3,$D8,$F1,$CF
	.dc.b	$E7,$70,$F3,$4F,$38,$FD,$03,$C0,$F8,$55,$D8,$F5,$47,$F1,$FC,$3E
	.dc.b	$3B,$EC,$6B,$F6,$70,$F9,$35,$CA,$3A,$CF,$A9,$E0,$75,$3C,$93,$EB
	.dc.b	$7B,$1E,$B8,$E0,$F5,$AE,$18,$FC,$A2,$5F,$77,$DF,$8F,$C1,$8F,$73
	.dc.b	$87,$C8,$E6,$BC,$67,$8A,$BA,$CF,$C5,$AE,$B3,$F3,$8F,$E4,$5D,$1F
	.dc.b	$EE,$6D,$BF,$A0,$ED,$FE,$33,$E3,$FE,$85,$5F,$E5,$99,$CB,$FE,$63
	.dc.b	$87,$C1,$E2,$F9,$4F,$F9,$7F,$E1,$B1,$46,$DE,$19,$3B,$3B,$14,$FF
	.dc.b	$52,$57,$FB,$EA,$3F,$DD,$6D,$8F,$EE,$31,$8A,$7C,$5F,$80,$93,$FE
	.dc.b	$71,$F9,$DA,$F4,$9D,$BF,$DE,$A7,$91,$CF,$3F,$49,$23,$FC,$F5,$DB
	.dc.b	$5F,$F4,$FF,$C0,$7F,$88,$DF,$F4,$03,$5F,$F3,$C4,$FC,$5F,$9B,$FF
	.dc.b	$50,$C7,$FA,$15,$AF,$FA,$A1,$BC,$1F,$E7,$CE,$5E,$1E,$B7,$C0,$9D
	.dc.b	$D7,$F0,$ED,$91,$88,$FF,$35,$35,$FE,$ED,$F4,$8F,$2F,$EC,$C2,$30
	.dc.b	$FF,$96,$9A,$FF,$98,$7C,$B7,$C3,$F0,$9C,$E7,$FE,$D2,$97,$C3,$74
	.dc.b	$6D,$F3,$C3,$BF,$C8,$7C,$E8,$FF,$36,$36,$FF,$DB,$6D,$E7,$FE,$53
	.dc.b	$F8,$FB,$E6,$B8,$F2,$7A,$3E,$97,$FD,$C3,$87,$FB,$89,$B7,$FC,$C3
	.dc.b	$0E,$58,$FF,$85,$7E,$95,$CF,$F0,$1B,$1F,$E7,$86,$BF,$E7,$FF,$9A
	.dc.b	$FE,$91,$8F,$FB,$BA,$73,$3F,$EF,$56,$BF,$CF,$9A,$F7,$AF,$F1,$03
	.dc.b	$4F,$F4,$2D,$79,$13,$67,$C3,$5D,$EF,$CC,$4A,$E3,$E4,$57,$CD,$E5
	.dc.b	$FD,$AC,$3F,$CC,$91,$EB,$AF,$D9,$7D,$2F,$FB,$1A,$3F,$DA,$4D,$7F
	.dc.b	$DE,$AD,$C1,$9F,$FA,$C6,$3F,$D9,$53,$C3,$DD,$E6,$47,$9F,$BB,$F8
	.dc.b	$AF,$8D,$7C,$AD,$24,$FF,$A9,$3E,$29,$F0,$D8,$8F,$F2,$25,$70,$7E
	.dc.b	$3B,$5C,$D7,$CA,$51,$9C,$FC,$F1,$FE,$51,$BF,$F2,$36,$7F,$DC,$1B
	.dc.b	$1F,$E5,$5A,$E1,$3F,$A4,$F8,$7F,$D9,$12,$7F,$65,$F4,$B5,$FD,$94
	.dc.b	$5C,$3B,$F7,$7E,$BA,$BF,$75,$F4,$B2,$FF,$74,$FE,$F7,$DB,$C4,$7F
	.dc.b	$04,$F7,$BF,$92,$78,$A7,$38,$FA,$A5,$AB,$E4,$7F,$6A,$D8,$F9,$51
	.dc.b	$8F,$03,$2C,$8C,$9F,$3E,$30,$FF,$B6,$67,$93,$67,$93,$FC,$47,$E8
	.dc.b	$7F,$8A,$11,$FE,$29,$F4,$55,$FE,$29,$17,$0F,$87,$E2,$AB,$F8,$E7
	.dc.b	$CF,$4B,$FE,$38,$FF,$1F,$FA,$1C,$47,$F1,$D8,$B7,$CA,$7F,$15,$F0
	.dc.b	$32,$BF,$DD,$7C,$A8,$F8,$7F,$0F,$99,$FC,$4C,$7E,$16,$75,$1F,$8A
	.dc.b	$8C,$1F,$2B,$9E,$47,$48,$FE,$65,$E0,$C8,$FE,$64,$FF,$35,$F8,$8F
	.dc.b	$E4,$BF,$4B,$FD,$5E,$7F,$76,$D7,$FD,$38,$D7,$FC,$7E,$1C,$3F,$CC
	.dc.b	$BE,$43,$F0,$3E,$04,$8F,$E8,$11,$FD,$0F,$E0,$24,$7F,$43,$7F,$A3
	.dc.b	$7C,$77,$8B,$D3,$FD,$16,$06,$92,$BF,$DC,$3C,$7F,$8D,$7F,$A8,$F9
	.dc.b	$9F,$90,$3F,$A8,$74,$FF,$75,$FA,$57,$15,$25,$7F,$AE,$F9,$3F,$3F
	.dc.b	$3F,$A8,$8F,$F5,$F7,$96,$F0,$7F,$68,$D8,$FE,$3E,$D7,$FD,$80,$D7
	.dc.b	$FA,$B7,$CF,$FF,$1E,$F0,$FF,$D9,$1E,$1A,$31,$28,$C7,$2F,$F8,$2E
	.dc.b	$5D,$5F,$DA,$76,$97,$FD,$A5,$FE,$D9,$CB,$FD,$73,$27,$83,$E6,$25
	.dc.b	$79,$9F,$ED,$59,$FF,$73,$24,$7F,$73,$FA,$1F,$F0,$EF,$CD,$BE,$4D
	.dc.b	$18,$6C,$FC,$7A,$FF,$71,$56,$EB,$FC,$45,$F4,$2F,$A1,$FE,$DF,$F4
	.dc.b	$6F,$8F,$FE,$38,$DC,$7F,$A3,$B5,$FF,$74,$35,$FE,$FD,$F3,$1F,$D7
	.dc.b	$38,$CD,$3F,$AF,$E5,$FC,$2B,$C4,$F4,$FD,$27,$F6,$9B,$F3,$1F,$EF
	.dc.b	$7C,$3E,$E2,$7E,$F5,$27,$84,$BA,$3F,$8B,$57,$F3,$58,$FC,$FB,$5F
	.dc.b	$F4,$FB,$5E,$DE,$1F,$F6,$2B,$FE,$3B,$CB,$F3,$3C,$8F,$B7,$5E,$49
	.dc.b	$CB,$E6,$3C,$95,$F2,$3E,$92,$46,$70,$47,$D6,$F8,$87,$C5,$EA,$BE
	.dc.b	$4C,$FC,$E2,$3F,$39,$FB,$BE,$A7,$F5,$48,$F1,$35,$F8,$1C,$BE,$C9
	.dc.b	$FF,$13,$FD,$27,$FB,$1A,$7C,$CD,$76,$03,$9D,$BF,$A4,$F2,$FC,$C7
	.dc.b	$FA,$04,$71,$9F,$E1,$2D,$7F,$D9,$2D,$7A,$37,$8D,$5F,$B5,$6E,$BF
	.dc.b	$90,$3F,$23,$C9,$6F,$FC,$A7,$CD,$FD,$48,$C3,$E8,$AE,$7E,$07,$FB
	.dc.b	$63,$7F,$F3,$1C,$62,$7D,$91,$8C,$7F,$84,$C6,$3A,$BF,$77,$0F,$A0
	.dc.b	$E5,$18,$C2,$D7,$EF,$6B,$8A,$BE,$93,$AC,$3D,$B8,$70,$AF,$E0,$DE
	.dc.b	$13,$E0,$7C,$11,$F3,$FC,$0F,$ED,$91,$88,$DB,$63,$F8,$A6,$3F,$E5
	.dc.b	$17,$73,$E1,$CF,$FD,$EF,$18,$9D,$5F,$7A,$D1,$DF,$C2,$F2,$56,$EF
	.dc.b	$8D,$6E,$F8,$E1,$E1,$BD,$CE,$6B,$73,$E5,$8C,$61,$D4,$C4,$47,$C2
	.dc.b	$7B,$0B,$87,$B2,$B9,$C7,$F5,$23,$47,$F5,$48,$F7,$B6,$BD,$8F,$F5
	.dc.b	$8F,$8B,$7C,$63,$A7,$FC,$B3,$18,$6C,$66,$7B,$4E,$5F,$05,$BC,$2F
	.dc.b	$E5,$11,$86,$BE,$36,$BF,$41,$AF,$DB,$78,$E9,$F5,$DB,$97,$EB,$65
	.dc.b	$F3,$9F,$E0,$5E,$8F,$BD,$5A,$9E,$5E,$16,$0F,$C2,$56,$E3,$E4,$CE
	.dc.b	$DF,$EA,$4D,$C7,$D5,$5C,$3C,$EF,$F2,$7E,$5E,$38,$E9,$7C,$9D,$63
	.dc.b	$89,$D4,$E4,$FB,$38,$F8,$63,$27,$EB,$BE,$7F,$FA,$35,$7E,$73,$8D
	.dc.b	$18,$BF,$C8,$C3,$D5,$FF,$3B,$D6,$EB,$57,$CF,$C5,$F3,$ED,$FD,$BA
	.dc.b	$72,$FF,$70,$BE,$38,$F4,$9C,$BF,$DC,$5F,$EA,$D8,$7F,$48,$C7,$C7
	.dc.b	$D3,$FE,$70,$EC,$FF,$31,$D7,$FA,$D6,$B4,$F0,$BF,$E1,$8D,$9D,$D7
	.dc.b	$1A,$F2,$7C,$05,$BF,$04,$D7,$F9,$4F,$C5,$5E,$E3,$BC,$BE,$2F,$76
	.dc.b	$62,$63,$FC,$D1,$D6,$7F,$E9,$48,$C1,$FD,$37,$E8,$FF,$85,$3F,$EB
	.dc.b	$66,$E3,$FA,$96,$BF,$E3,$C6,$BF,$91,$77,$A7,$F9,$E5,$B3,$FE,$02
	.dc.b	$8F,$80,$EF,$AB,$D9,$3A,$AF,$E8,$CD,$E9,$FE,$DD,$AE,$AB,$E2,$43
	.dc.b	$E3,$E6,$1F,$D9,$DB,$87,$F5,$01,$F4,$6F,$57,$8C,$EB,$3F,$F4,$74
	.dc.b	$E6,$FF,$B7,$36,$FF,$AC,$E1,$FC,$07,$E4,$7E,$94,$E6,$FF,$D5,$F3
	.dc.b	$9B,$E3,$8F,$9A,$17,$C2,$3E,$B7,$98,$F9,$4A,$79,$3E,$3A,$E1,$E2
	.dc.b	$F4,$7B,$9C,$F8,$AC,$7F,$81,$72,$FC,$0B,$63,$EE,$F8,$9B,$8F,$C8
	.dc.b	$FD,$1F,$E0,$0F,$C0,$F8,$D6,$3F,$C1,$9F,$46,$F8,$DF,$C7,$36,$3F
	.dc.b	$C2,$1F,$8A,$BE,$4B,$35,$F8,$AF,$81,$95,$F9,$EF,$82,$BC,$DF,$DA
	.dc.b	$7E,$03,$FB,$04,$FE,$9D,$B1,$9E,$DF,$CC,$46,$39,$FF,$D8,$1A,$7F
	.dc.b	$13,$6F,$8F,$FA,$1D,$5E,$1F,$9A,$3F,$41,$4F,$9B,$A6,$3E,$52,$E9
	.dc.b	$FB,$2A,$FD,$09,$FA,$2D,$EC,$7C,$2F,$87,$F8,$27,$EB,$FC,$7F,$DD
	.dc.b	$6F,$F8,$BC,$BF,$DE,$F5,$FB,$CF,$5F,$F5,$67,$EB,$3E,$92,$C7,$F0
	.dc.b	$0F,$A4,$FD,$79,$FB,$0F,$5A,$C7,$D0,$F3,$7E,$79,$FB,$9F,$C5,$FF
	.dc.b	$DC,$51,$93,$F8,$4F,$07,$EE,$0F,$DC,$F0,$58,$FF,$22,$7D,$0F,$FA
	.dc.b	$45,$FE,$BD,$AA,$30,$FF,$61,$6B,$FE,$B4,$6B,$FD,$25,$DF,$27,$FB
	.dc.b	$C5,$B3,$FC,$9F,$83,$F8,$94,$62,$BF,$C7,$8D,$7C,$1E,$B7,$F0,$E9
	.dc.b	$DB,$FC,$43,$E6,$A5,$FF,$24,$D2,$DF,$F2,$49,$B8,$FF,$2A,$9B,$5F
	.dc.b	$CD,$EB,$F8,$C9,$FC,$6B,$D2,$B1,$FC,$BA,$29,$FC,$BA,$2D,$7F,$35
	.dc.b	$F9,$D7,$CC,$7F,$2D,$84,$62,$FF,$D0,$71,$FE,$C2,$F3,$35,$39,$C3
	.dc.b	$FC,$B7,$3C,$2F,$27,$3F,$03,$F8,$D4,$67,$0E,$77,$C6,$3C,$55,$BC
	.dc.b	$38,$D7,$F4,$BF,$15,$F4,$9F,$CE,$DB,$19,$FA,$5C,$4C,$8E,$0E,$E3
	.dc.b	$19,$FF,$96,$69,$F2,$5C,$CE,$9F,$25,$70,$F0,$45,$AE,$13,$73,$FC
	.dc.b	$17,$9A,$BF,$C6,$F2,$7F,$25,$BF,$C5,$4E,$5F,$C7,$35,$CC,$FE,$5A
	.dc.b	$3E,$87,$D0,$F9,$23,$04,$9F,$E7,$15,$FC,$B6,$3F,$DE,$13,$B9,$CD
	.dc.b	$7F,$5F,$FA,$4F,$F7,$BD,$7F,$73,$6C,$FF,$9E,$5F,$E1,$91,$FE,$EC
	.dc.b	$73,$87,$9F,$D6,$46,$27,$5C,$5D,$7C,$BE,$CF,$E8,$3D,$F9,$3F,$E0
	.dc.b	$1F,$A4,$7C,$56,$33,$F4,$D5,$CF,$B5,$6A,$FF,$6F,$3F,$DE,$FE,$93
	.dc.b	$FC,$09,$7F,$E3,$90,$E1,$FF,$1E,$35,$FF,$55,$35,$DF,$B1,$CC,$FC
	.dc.b	$BF,$C5,$DB,$FF,$64,$57,$F7,$0A,$E3,$5A,$9C,$BE,$3F,$02,$47,$E8
	.dc.b	$B8,$5F,$45,$8B,$FE,$D9,$87,$F2,$FA,$EB,$7F,$72,$B5,$A7,$E0,$7E
	.dc.b	$85,$F4,$9F,$E1,$96,$DF,$E9,$23,$04,$7F,$DB,$D1,$AC,$7F,$5F,$DB
	.dc.b	$6F,$A5,$FE,$97,$E6,$E6,$1F,$A9,$F6,$6E,$1F,$A1,$9B,$7F,$7F,$AF
	.dc.b	$FB,$85,$FE,$BE,$7F,$5D,$F3,$92,$3F,$65,$F4,$97,$0F,$D5,$4D,$C3
	.dc.b	$94,$DA,$FF,$6E,$74,$B1,$FF,$77,$C7,$EE,$FE,$37,$F7,$5B,$FE,$ED
	.dc.b	$91,$93,$F8,$04,$63,$2F,$DC,$7B,$DF,$C0,$4F,$E0,$5F,$01,$6B,$F8
	.dc.b	$34,$59,$FD,$F4,$DC,$FD,$CF,$0A,$FE,$19,$F1,$56,$B7,$1F,$C3,$E7
	.dc.b	$F8,$74,$E6,$BF,$71,$5F,$ED,$C1,$F4,$8F,$33,$FD,$1D,$3F,$E0,$3C
	.dc.b	$3F,$DF,$93,$9A,$FC,$D7,$C4,$62,$F7,$EE,$FF,$05,$AD,$57,$FB,$AC
	.dc.b	$EB,$3A,$BE,$11,$C8,$C8,$18,$78,$5C,$9C,$FE,$93,$AC,$65,$EF,$1E
	.dc.b	$B7,$38,$7F,$17,$73,$1F,$EE,$F3,$9C,$F5,$C3,$38,$F9,$27,$C3,$7E
	.dc.b	$BC,$7F,$D9,$93,$9B,$E1,$18,$AE,$75,$B7,$D9,$87,$8B,$1F,$63,$DA
	.dc.b	$0E,$5B,$5C,$F1,$C1,$B2,$B9,$3E,$54,$7C,$43,$ED,$46,$9F,$98,$D7
	.dc.b	$E9,$35,$DC,$72,$8E,$CD,$3B,$47,$D3,$C3,$DE,$F4,$8D,$34,$D6,$1F
	.dc.b	$05,$EF,$1A,$69,$AC,$3E,$73,$E2,$30,$7A,$CF,$AD,$7E,$76,$CF,$37
	.dc.b	$F3,$1E,$4F,$CA,$7F,$14,$D7,$F2,$0D,$77,$8F,$FB,$A7,$E3,$3F,$D1
	.dc.b	$99,$7E,$3E,$B8,$46,$29,$18,$D7,$F2,$71,$8B,$E5,$18,$9D,$5F,$E2
	.dc.b	$70,$79,$DF,$EF,$7C,$BF,$47,$18,$C7,$F1,$D9,$77,$9F,$D3,$B6,$7F
	.dc.b	$DF,$C6,$E7,$F4,$2D,$9D,$E3,$E4,$6E,$1E,$1C,$3C,$15,$C9,$B7,$FA
	.dc.b	$DC,$3B,$37,$0E,$55,$B8,$E4,$79,$A3,$2E,$EF,$2F,$92,$FA,$AB,$EA
	.dc.b	$DA,$E2,$FF,$0D,$9D,$DF,$AC,$F9,$77,$DC,$EB,$87,$C8,$9E,$B8,$74
	.dc.b	$3A,$D7,$1B,$FE,$10,$D7,$C1,$7C,$DF,$D6,$9A,$AF,$E1,$CE,$0F,$21
	.dc.b	$FB,$69,$F1,$B9,$AE,$F3,$E8,$3C,$78,$7B,$23,$33,$E8,$AE,$38,$7D
	.dc.b	$48,$ED,$1F,$6F,$2F,$64,$77,$3E,$61,$DE,$FE,$3C,$FA,$CF,$65,$FC
	.dc.b	$18,$CE,$1E,$3A,$F7,$61,$FC,$D7,$8F,$FA,$B6,$BF,$CE,$B5,$FE,$B3
	.dc.b	$C7,$FE,$BD,$BF,$F2,$94,$E5,$57,$F5,$DF,$81,$C3,$CA,$36,$43,$FC
	.dc.b	$A1,$1A,$A7,$C1,$D2,$4E,$F7,$FE,$E9,$18,$8F,$F2,$C3,$C7,$0E,$5E
	.dc.b	$E4,$9F,$F3,$33,$D7,$E5,$B1,$1F,$E5,$F8,$B7,$FD,$FE,$7C,$5C,$32
	.dc.b	$BF,$CD,$CF,$97,$BB,$11,$FE,$67,$8B,$5F,$E0,$79,$F7,$FC,$B7,$FA
	.dc.b	$22,$BF,$90,$6D,$FE,$00,$8C,$47,$F5,$48,$C7,$A3,$F0,$79,$78,$B2
	.dc.b	$78,$9F,$32,$B8,$E1,$FE,$15,$AF,$A1,$2F,$A4,$FF,$26,$BC,$0C,$79
	.dc.b	$3A,$8C,$C7,$84,$FE,$CD,$F4,$9F,$EF,$B7,$84,$6A,$FF,$CE,$8D,$CB
	.dc.b	$F3,$ED,$9E,$3E,$FF,$8B,$E4,$3F,$DD,$F1,$E1,$6B,$DC,$38,$BF,$F4
	.dc.b	$07,$D2,$3E,$3F,$97,$14,$BE,$4F,$BA,$75,$1E,$DF,$A3,$FF,$15,$37
	.dc.b	$9B,$E1,$C6,$23,$C1,$F2,$BF,$CE,$F1,$FF,$47,$4E,$5B,$5C,$0F,$97
	.dc.b	$3B,$D7,$F7,$B5,$CE,$FF,$CE,$0D,$73,$5F,$EE,$A5,$19,$7F,$C2,$4F
	.dc.b	$C0,$7D,$19,$F9,$8E,$FE,$2D,$8C,$90,$F4,$BF,$AA,$FA,$3F,$F5,$AC
	.dc.b	$7E,$13,$AD,$18,$8F,$B9,$18,$F6,$38,$7B,$1C,$59,$3C,$6B,$86,$BF
	.dc.b	$E9,$D5,$5F,$8C,$F8,$87,$C4,$D2,$5F,$E3,$7E,$15,$F0,$38,$8F,$BD
	.dc.b	$5C,$1F,$C9,$35,$FC,$14,$F0,$73,$5F,$26,$75,$1F,$E9,$D6,$BF,$CB
	.dc.b	$9A,$FE,$8F,$E0,$7F,$5E,$47,$E2,$75,$BE,$8F,$11,$F1,$21,$38,$D3
	.dc.b	$FD,$79,$1B,$9E,$3F,$47,$FB,$A2,$3D,$3D,$8A,$F3,$8B,$5C,$A7,$F6
	.dc.b	$79,$BE,$36,$92,$FE,$47,$C6,$D7,$E1,$C5,$BF,$F7,$15,$2B,$1C,$FF
	.dc.b	$D3,$A3,$A5,$FD,$1B,$EF,$5F,$C4,$3E,$8E,$4F,$C2,$7F,$89,$FC,$EE
	.dc.b	$23,$E1,$45,$C7,$E0,$CF,$F1,$8F,$9C,$95,$F2,$9F,$E3,$7F,$47,$88
	.dc.b	$F9,$51,$6B,$E4,$CF,$E1,$FE,$42,$57,$E0,$1F,$87,$F0,$58,$8F,$C2
	.dc.b	$C5,$AF,$7C,$E6,$7F,$01,$3F,$83,$23,$FE,$DF,$3F,$91,$76,$FF,$2E
	.dc.b	$23,$F8,$0F,$85,$FE,$47,$F4,$1F,$D2,$A3,$F9,$13,$5F,$E2,$4D,$C3
	.dc.b	$F4,$1E,$97,$F3,$8F,$A3,$95,$F8,$A7,$F9,$DF,$CB,$7F,$3C,$FA,$39
	.dc.b	$3F,$8E,$7F,$9F,$FD,$1F,$83,$D1,$93,$E0,$7F,$B2,$7C,$C7,$F9,$DB
	.dc.b	$1F,$E8,$BF,$11,$2B,$F2,$EF,$F4,$8F,$83,$FE,$CF,$F4,$6E,$2A,$4A
	.dc.b	$FE,$01,$1F,$AB,$D9,$D5,$FE,$12,$F3,$3F,$3B,$5F,$D0,$5A,$FF,$31
	.dc.b	$6C,$F0,$86,$6F,$C8,$3F,$DE,$31,$FC,$37,$E8,$3F,$93,$61,$FE,$D8
	.dc.b	$6C,$7F,$7E,$6F,$92,$5D,$3F,$D8,$BD,$39,$5F,$B4,$7F,$B2,$F9,$9F
	.dc.b	$D6,$34,$78,$34,$95,$FD,$83,$4F,$25,$FF,$6C,$CE,$4F,$EF,$9F,$ED
	.dc.b	$FF,$45,$F3,$BE,$4C,$D2,$4F,$F2,$7F,$A2,$7C,$DF,$F6,$8F,$A3,$7C
	.dc.b	$85,$25,$78,$70,$E6,$FF,$2F,$EC,$FE,$B1,$97,$FB,$E9,$B1,$FE,$20
	.dc.b	$6E,$8F,$1D,$E2,$7E,$0B,$E9,$3F,$AC,$DF,$F6,$3C,$FE,$D6,$3F,$E4
	.dc.b	$A6,$C7,$D4,$9D,$F3,$FD,$78,$4E,$9B,$E7,$7A,$CE,$DE,$BF,$E4,$32
	.dc.b	$FA,$73,$F2,$DA,$FF,$80,$73,$F3,$73,$71,$8F,$F0,$46,$0F,$9A,$F8
	.dc.b	$C4,$8F,$A9,$F0,$EF,$8B,$C4,$7A,$E6,$E9,$F8,$DC,$57,$BA,$EB,$5F
	.dc.b	$D9,$F0,$ED,$1F,$BD,$3D,$B1,$93,$FB,$67,$D1,$BC,$2F,$F0,$FB,$67
	.dc.b	$FC,$3C,$D7,$FB,$0F,$37,$C8,$E1,$FF,$3A,$CE,$73,$F9,$4D,$8C,$DF
	.dc.b	$F9,$5A,$B8,$5F,$F9,$69,$FA,$7D,$A6,$93,$9B,$87,$14,$AF,$C7,$3C
	.dc.b	$E3,$A7,$81,$FE,$86,$DB,$FC,$4B,$87,$38,$C7,$63,$0F,$04,$61,$F1
	.dc.b	$7C,$AD,$AE,$11,$BB,$E7,$97,$F0,$78,$C6,$1C,$EB,$71,$F4,$A3,$12
	.dc.b	$BE,$D3,$FB,$9F,$0F,$FD,$AE,$EC,$F2,$78,$08,$CE,$3F,$E3,$C8,$C7
	.dc.b	$C0,$78,$2B,$7E,$9F,$49,$CC,$73,$3B,$4E,$DA,$66,$3F,$8B,$D7,$5B
	.dc.b	$EF,$97,$FA,$BA,$79,$CE,$6F,$94,$70,$9D,$57,$0B,$F4,$B5,$EF,$7E
	.dc.b	$9C,$3C,$D3,$9B,$EB,$7C,$B0,$EB,$5A,$7F,$2D,$3E,$CC,$3F,$D8,$27
	.dc.b	$A8,$AE,$5F,$F6,$2F,$C4,$BE,$2F,$F9,$7C,$7F,$5B,$8C,$73,$FE,$85
	.dc.b	$B1,$99,$DB,$D3,$C4,$78,$1E,$26,$C7,$F6,$5C,$3F,$CD,$B3,$92,$47
	.dc.b	$CD,$9D,$EB,$FE,$74,$F9,$A7,$A1,$FD,$0E,$3E,$E4,$EB,$6F,$A0,$DB
	.dc.b	$FC,$1B,$55,$F2,$E7,$FA,$73,$9C,$BC,$75,$E1,$CB,$F0,$17,$B8,$C1
	.dc.b	$31,$F0,$CE,$B2,$F2,$65,$E2,$8E,$A4,$AF,$C6,$5F,$3B,$E3,$1F,$99
	.dc.b	$E3,$FD,$C7,$57,$28,$FD,$E7,$75,$9C,$EB,$F9,$97,$F7,$D1,$9E,$5F
	.dc.b	$E4,$53,$97,$F9,$E3,$6F,$53,$BF,$93,$FE,$D9,$D9,$FD,$B7,$0F,$E1
	.dc.b	$1F,$39,$8C,$3F,$63,$57,$2F,$E1,$D3,$6B,$F8,$74,$DC,$93,$63,$0E
	.dc.b	$74,$98,$7F,$5D,$F6,$EE,$3F,$C8,$A6,$E5,$E3,$C7,$FC,$EF,$7F,$D8
	.dc.b	$E7,$C1,$F4,$AF,$26,$98,$E4,$B5,$FE,$8F,$9D,$36,$74,$D2,$63,$FE
	.dc.b	$92,$9D,$B6,$B9,$BF,$CD,$B8,$7D,$DF,$1D,$EF,$6D,$FF,$77,$70,$FF
	.dc.b	$A7,$FE,$C5,$F3,$98,$BF,$F4,$FC,$62,$78,$1B,$9D,$6B,$F4,$FE,$1B
	.dc.b	$FC,$03,$3F,$E0,$3F,$A2,$FA,$11,$8C,$3F,$A4,$C6,$3C,$99,$5F,$EA
	.dc.b	$FA,$FF,$51,$B6,$38,$F8,$7C,$BC,$4B,$3B,$7E,$E6,$7F,$EA,$C8,$FF
	.dc.b	$2D,$73,$23,$19,$7F,$0A,$8C,$7C,$17,$F7,$F8,$FE,$75,$3A,$7F,$A3
	.dc.b	$09,$E5,$F2,$FD,$8E,$FF,$0B,$70,$F9,$1F,$2D,$E2,$F9,$7F,$1B,$63
	.dc.b	$E6,$7C,$BF,$9B,$E2,$7C,$ED,$8F,$81,$F4,$1F,$18,$F8,$FE,$1D,$8F
	.dc.b	$A9,$F3,$1E,$E3,$DF,$E8,$D8,$FB,$1E,$7F,$C2,$3E,$1F,$91,$63,$EE
	.dc.b	$7C,$A7,$C6,$3E,$3F,$CB,$58,$FC,$07,$BD,$FA,$3F,$87,$F9,$6D,$8F
	.dc.b	$C2,$7C,$A7,$E9,$FE,$27,$E7,$B6,$3F,$DC,$DF,$08,$F9,$FF,$D7,$36
	.dc.b	$33,$E0,$E2,$7E,$BF,$A7,$27,$F2,$1F,$2D,$7A,$DE,$1F,$EF,$BE,$69
	.dc.b	$F2,$FF,$BB,$84,$62,$FF,$BE,$47,$F0,$4F,$69,$8E,$EB,$58,$74,$BD
	.dc.b	$C6,$2F,$B6,$1D,$70,$E3,$5D,$0D,$C7,$32,$39,$BF,$3C,$F0,$7B,$1E
	.dc.b	$AB,$ED,$3F,$E0,$7E,$6F,$F7,$43,$C7,$BD,$1A,$AF,$C8,$3A,$79,$C6
	.dc.b	$B0,$FE,$35,$39,$C7,$F9,$13,$C6,$79,$FC,$86,$AB,$7F,$09,$C3,$DA
	.dc.b	$B9,$78,$70,$FF,$7B,$B6,$74,$D3,$53,$E0,$F8,$19,$1C,$7E,$3E,$E9
	.dc.b	$E6,$C3,$FC,$E7,$3D,$6B,$CF,$5A,$D3,$CF,$16,$7E,$24,$DB,$FF,$1B
	.dc.b	$7C,$FB,$E3,$7F,$01,$F4,$EF,$97,$FE,$79,$4A,$D6,$1F,$4F,$25,$EE
	.dc.b	$3E,$B5,$6F,$0F,$E0,$0F,$66,$B9,$8F,$F3,$35,$F4,$8F,$E1,$19,$BE
	.dc.b	$47,$E8,$9F,$4B,$E4,$AD,$FC,$B8,$B5,$F0,$E6,$E5,$F9,$BD,$BF,$28
	.dc.b	$7E,$57,$86,$E1,$E9,$E3,$FA,$78,$7E,$2A,$FF,$3F,$7B,$AD,$3C,$27
	.dc.b	$55,$C2,$BF,$51,$79,$7F,$B3,$74,$BB,$BF,$6B,$E0,$FE,$D9,$B9,$7E
	.dc.b	$E9,$B8,$FE,$42,$B8,$4F,$EF,$AB,$E8,$FC,$EF,$FA,$F2,$3E,$87,$CD
	.dc.b	$3C,$9C,$FA,$1F,$B7,$8C,$4F,$D2,$AE,$18,$FF,$8F,$03,$87,$FC,$7E
	.dc.b	$D7,$F9,$9B,$5F,$D6,$73,$FE,$F4,$FD,$F7,$3D,$AF,$95,$16,$7E,$2C
	.dc.b	$DD,$7F,$86,$4D,$AF,$E0,$73,$6B,$FD,$0D,$F5,$EF,$2B,$F8,$57,$C5
	.dc.b	$7E,$42,$31,$7F,$5E,$31,$3F,$8A,$9C,$D7,$B5,$AF,$F9,$69,$AF,$F4
	.dc.b	$EF,$C7,$7F,$B6,$3F,$6A,$FF,$3B,$18,$CB,$F9,$14,$63,$C3,$FC,$BC
	.dc.b	$F0,$AF,$F3,$2B,$5E,$91,$F9,$EE,$FF,$E1,$BD,$7F,$C2,$0F,$E1,$39
	.dc.b	$DC,$3F,$81,$45,$BF,$DD,$4D,$C7,$F8,$6F,$C6,$FF,$10,$3F,$88,$FC
	.dc.b	$B5,$C3,$F8,$54,$5B,$FE,$05,$36,$BF,$8A,$FC,$A7,$F1,$83,$F8,$CF
	.dc.b	$35,$C3,$F8,$94,$5B,$FE,$15,$36,$BF,$8E,$FC,$AF,$F2,$03,$F9,$0F
	.dc.b	$CA,$DC,$3F,$8D,$45,$BF,$E2,$53,$6B,$F9,$2F,$C6,$F0,$BF,$E4,$0D
	.dc.b	$AF,$E3,$0D,$9D,$CF,$F2,$5A,$FE,$1F,$5F,$EC,$31,$F3,$CF,$9F,$FF
	.dc.b	$03,$DB,$D0,$61,$FE,$CF,$AD,$5F,$FA,$6B,$E2,$F1,$87,$0B,$DC,$F4
	.dc.b	$8F,$D2,$4E,$A3,$FA,$EC,$EA,$B9,$C7,$FB,$7C,$E0,$6A,$F8,$39,$8D
	.dc.b	$E1,$E5,$9C,$CF,$AE,$73,$5D,$70,$F0,$CF,$43,$FB,$9C,$65,$F5,$19
	.dc.b	$BE,$8E,$CE,$93,$ED,$BE,$F0,$C3,$CB,$5C,$30,$F5,$CF,$03,$D4,$73
	.dc.b	$38,$1D,$F0,$F1,$3E,$DC,$3B,$3B,$BE,$B3,$A8,$E8,$7A,$B0,$F1,$CE
	.dc.b	$9B,$1D,$5B,$7F,$11,$B7,$DA,$3C,$F3,$E4,$74,$F8,$4E,$45,$72,$70
	.dc.b	$AF,$04,$66,$7A,$1F,$42,$76,$D3,$31,$F4,$23,$2D,$8C,$BF,$83,$F2
	.dc.b	$DE,$67,$DA,$6C,$7D,$86,$BB,$8F,$F7,$4F,$7F,$F8,$C2,$3F,$C5,$31
	.dc.b	$8C,$D1,$82,$FC,$72,$B8,$5F,$48,$C5,$70,$8C,$4F,$5C,$7F,$33,$8B
	.dc.b	$E0,$BF,$CD,$F1,$86,$C7,$87,$1F,$C9,$63,$D1,$FC,$A4,$71,$7F,$3E
	.dc.b	$D7,$77,$DE,$39,$5F,$59,$DD,$F6,$7D,$75,$B6,$E3,$A6,$1D,$1F,$D6
	.dc.b	$E1,$E3,$BE,$93,$E4,$BE,$8A,$75,$87,$88,$FF,$46,$9E,$7A,$E2,$7E
	.dc.b	$F2,$37,$3E,$23,$A1,$E2,$8D,$61,$E7,$8E,$CD,$AE,$87,$6B,$F1,$DF
	.dc.b	$4A,$DE,$1C,$EB,$51,$EC,$BE,$95,$B9,$E7,$3F,$82,$6B,$C8,$F8,$87
	.dc.b	$98,$98,$79,$AF,$A1,$DE,$39,$61,$E4,$C3,$B9,$E9,$7E,$4D,$F6,$C3
	.dc.b	$D0,$78,$E3,$27,$1B,$FE,$4D,$CF,$FE,$6A,$F8,$3F,$E6,$7B,$7F,$9B
	.dc.b	$FE,$16,$E7,$FE,$6C,$6D,$6E,$3F,$CC,$55,$8F,$2B,$FD,$ED,$1F,$E6
	.dc.b	$5A,$C4,$E7,$E5,$64,$FF,$9A,$5E,$13,$88,$FF,$31,$45,$C7,$FC,$EB
	.dc.b	$3E,$0F,$AF,$7C,$B5,$25,$FF,$9C,$F3,$AF,$FA,$02,$76,$FF,$A1,$DA
	.dc.b	$FF,$9B,$E3,$A7,$8D,$23,$4F,$7F,$84,$C4,$73,$8B,$8F,$F5,$0D,$58
	.dc.b	$2F,$CD,$F2,$73,$0F,$03,$E9,$F8,$5C,$47,$96,$2E,$1C,$ED,$5F,$42
	.dc.b	$9F,$8D,$3D,$DF,$F2,$04,$7F,$0C,$8C,$38,$6B,$FE,$45,$6B,$FE,$F4
	.dc.b	$39,$C7,$5E,$57,$75,$3C,$15,$CB,$1F,$ED,$B6,$E2,$7C,$1F,$8A,$30
	.dc.b	$D8,$CC,$79,$9A,$FF,$95,$DB,$3D,$2F,$FD,$C1,$3F,$E8,$E6,$BF,$EC
	.dc.b	$23,$A9,$FE,$92,$F8,$A7,$83,$F1,$5B,$19,$8C,$4E,$59,$19,$39,$46
	.dc.b	$1E,$D3,$C9,$B3,$C9,$FF,$50,$FC,$F3,$C3,$F8,$0D,$8F,$A6,$D7,$FC
	.dc.b	$E8,$D7,$E8,$39,$C3,$E5,$FC,$FB,$E5,$3E,$F3,$6B,$BD,$FE,$BB,$D0
	.dc.b	$71,$DE,$29,$5F,$EE,$19,$CB,$67,$2F,$FB,$82,$3F,$16,$E4,$FF,$5D
	.dc.b	$DB,$E2,$BC,$24,$8C,$FB,$F5,$FA,$53,$B7,$8C,$FA,$27,$59,$FB,$22
	.dc.b	$F4,$FF,$60,$A5,$FD,$CA,$FC,$F7,$C6,$7E,$EC,$93,$FB,$5F,$80,$AF
	.dc.b	$EE,$62,$CF,$E9,$67,$F4,$9F,$3D,$EC,$AF,$F5,$13,$5F,$CF,$35,$F1
	.dc.b	$35,$FE,$E6,$F1,$9F,$CE,$FC,$A7,$F8,$FB,$1F,$43,$6B,$7B,$7F,$11
	.dc.b	$F8,$49,$1F,$C2,$DF,$E2,$9F,$0D,$88,$FE,$21,$17,$0F,$DE,$4D,$C3
	.dc.b	$FC,$07,$DD,$FC,$22,$B1,$68,$FE,$17,$5F,$C7,$FE,$23,$F9,$07,$C1
	.dc.b	$CA,$FE,$35,$19,$7F,$8E,$46,$5F,$E4,$11,$FC,$9F,$3F,$E5,$1F,$13
	.dc.b	$27,$F9,$14,$65,$FE,$49,$19,$7F,$94,$47,$F1,$11,$8D,$24,$7F,$87
	.dc.b	$33,$F9,$38,$E2,$3F,$71,$5C,$1D,$67,$F3,$2F,$A5,$6E,$FE,$8C,$59
	.dc.b	$FA,$33,$FC,$FF,$E4,$E5,$FF,$39,$7F,$A1,$7C,$AF,$E9,$7E,$56,$4F
	.dc.b	$E8,$DF,$D4,$79,$9F,$D1,$DF,$EA,$FF,$28,$DE,$99,$AF,$E8,$9F,$29
	.dc.b	$AE,$09,$18,$3F,$A4,$7C,$A7,$F5,$12,$3F,$D2,$FC,$EF,$EC,$7F,$5E
	.dc.b	$F7,$29,$27,$FA,$81,$AE,$B8,$FC,$43,$F0,$13,$FA,$D5,$C7,$FD,$36
	.dc.b	$3F,$51,$C5,$F8,$8F,$AD,$7C,$27,$F7,$26,$DF,$3E,$84,$62,$39,$D7
	.dc.b	$1F,$93,$F3,$FC,$A7,$A3,$CA,$95,$C7,$24,$FB,$32,$F6,$DC,$AD,$47
	.dc.b	$C0,$F8,$39,$1C,$5F,$87,$F2,$9F,$13,$D6,$FE,$AF,$AB,$24,$61,$B7
	.dc.b	$FD,$54,$F8,$71,$C3,$57,$C4,$7B,$FD,$D9,$5E,$F9,$CB,$7E,$4F,$E2
	.dc.b	$11,$D4,$FC,$7F,$A3,$FC,$1F,$85,$27,$94,$7D,$AF,$4F,$ED,$D4,$9F
	.dc.b	$89,$15,$F9,$67,$D8,$8C,$72,$7D,$79,$BC,$9F,$BE,$D7,$CB,$7C,$36
	.dc.b	$F8,$5A,$EA,$3F,$80,$6A,$F7,$7E,$E1,$23,$DD,$EF,$57,$94,$ED,$F8
	.dc.b	$53,$F8,$39,$D6,$7F,$E1,$5A,$D5,$70,$AD,$3F,$7F,$7F,$CC,$91,$FC
	.dc.b	$16,$D5,$F4,$66,$9C,$61,$F9,$5A,$FD,$0E,$3F,$A2,$F9,$49,$87,$E2
	.dc.b	$E3,$2F,$E3,$63,$2F,$48,$FF,$25,$FE,$42,$F9,$5F,$E9,$1F,$05,$FD
	.dc.b	$8F,$7F,$C4,$6B,$8F,$4F,$E6,$37,$1F,$E5,$18,$71,$7F,$D5,$F7,$C6
	.dc.b	$31,$1F,$EE,$9A,$E2,$E9,$AE,$4E,$78,$75,$5F,$34,$F8,$5F,$A1,$87
	.dc.b	$F7,$79,$CC,$62,$FF,$B0,$56,$F1,$ED,$5F,$EC,$28,$7C,$8F,$57,$51
	.dc.b	$E1,$39,$CF,$06,$99,$8E,$2E,$5E,$38,$78,$DB,$7E,$2C,$3C,$95,$E1
	.dc.b	$C3,$C1,$18,$8E,$B7,$C6,$31,$7D,$A7,$C1,$18,$26,$1E,$1B,$E6,$FF
	.dc.b	$95,$2B,$D7,$5F,$E4,$F6,$BE,$38,$FF,$2C,$FC,$27,$F9,$6F,$E6,$EF
	.dc.b	$BE,$69,$2B,$F8,$25,$FF,$94,$7C,$5E,$DE,$F1,$A4,$9D,$C6,$5F,$90
	.dc.b	$DD,$3E,$C7,$90,$F2,$1D,$7F,$2E,$B5,$C1,$F6,$9B,$87,$D9,$6A,$BE
	.dc.b	$55,$C7,$19,$3E,$07,$ED,$4E,$A3,$04,$C7,$EC,$5F,$49,$D4,$62,$B9
	.dc.b	$B6,$BF,$0D,$3A,$C3,$CD,$5F,$8E,$E3,$79,$4F,$37,$DB,$87,$E7,$67
	.dc.b	$3E,$8E,$32,$FD,$75,$EF,$4E,$4A,$75,$97,$F8,$62,$35,$87,$37,$55
	.dc.b	$C3,$1D,$CF,$F8,$6E,$72,$EF,$CD,$F1,$F5,$FF,$AA,$65,$E7,$E3,$0E
	.dc.b	$D5,$6F,$94,$DD,$3D,$B3,$6C,$DF,$3F,$0E,$77,$98,$FF,$65,$A7,$CD
	.dc.b	$E3,$2F,$E6,$91,$F6,$E7,$55,$D6,$B5,$9F,$FA,$FF,$DE,$7C,$2F,$8D
	.dc.b	$B7,$FD,$C5,$FF,$53,$36,$34,$47,$5C,$77,$2F,$E8,$95,$C9,$B5,$D6
	.dc.b	$75,$7F,$1F,$8F,$E0,$7A,$3F,$05,$B8,$FB,$5C,$3F,$E0,$3F,$9A,$FF
	.dc.b	$0A,$39,$31,$E9,$7C,$B8,$CE,$5F,$08,$C4,$7C,$FF,$9B,$FF,$77,$D7
	.dc.b	$F8,$35,$B3,$FD,$6D,$FF,$70,$46,$7E,$1F,$87,$C3,$48,$E1,$EC,$DB
	.dc.b	$E1,$F0,$FD,$63,$B3,$67,$D1,$EB,$F9,$4F,$37,$BF,$63,$D9,$EB,$FA
	.dc.b	$4F,$57,$B1,$63,$E0,$7A,$FE,$DF,$63,$DC,$D8,$F8,$9E,$C7,$C1,$F6
	.dc.b	$3E,$13,$63,$DD,$EB,$FC,$13,$E1,$78,$B6,$3E,$67,$B1,$F2,$4F,$95
	.dc.b	$F3,$76,$3E,$87,$B1,$F3,$4F,$9D,$E5,$D8,$FA,$9E,$CF,$D1,$3E,$97
	.dc.b	$B3,$63,$EC,$7A,$FF,$54,$FA,$DE,$BD,$8F,$B9,$EB,$FD,$93,$ED,$79
	.dc.b	$96,$3F,$C8,$5F,$0C,$E6,$FD,$43,$63,$3D,$D8,$9F,$C1,$7A,$72,$7F
	.dc.b	$11,$EC,$5E,$DF,$F0,$FF,$88,$F9,$6F,$EE,$D5,$FE,$59,$8C,$7B,$1F
	.dc.b	$B7,$5F,$01,$FE,$1B,$9D,$57,$0B,$E6,$DA,$FE,$06,$D9,$FF,$46,$57
	.dc.b	$1A,$DD,$F5,$AD,$47,$F9,$99,$B7,$E0,$C7,$55,$FC,$1F,$99,$F2,$D6
	.dc.b	$30,$DC,$BF,$CD,$13,$A6,$CE,$9A,$6B,$E6,$FA,$7B,$5D,$5B,$19,$39
	.dc.b	$C6,$5A,$76,$F9,$6F,$21,$E5,$F9,$6B,$19,$26,$BF,$E4,$7C,$53,$A8
	.dc.b	$C5,$5D,$3F,$9E,$FA,$BF,$E6,$CA,$C6,$1F,$EF,$1F,$9A,$7C,$AF,$F4
	.dc.b	$2C,$7F,$DE,$82,$33,$3F,$E7,$18,$C4,$ED,$E4,$D4,$6A,$3F,$D3,$4D
	.dc.b	$3F,$0F,$F4,$AF,$1B,$F9,$D4,$67,$2F,$3E,$8F,$9B,$FA,$C7,$D7,$F9
	.dc.b	$BB,$8F,$D4,$8B,$7F,$0E,$6D,$FE,$23,$7F,$C2,$1F,$85,$E2,$B8,$7E
	.dc.b	$42,$2D,$FE,$42,$6D,$71,$AD,$1F,$96,$8C,$BF,$EF,$6E,$9F,$F4,$8B
	.dc.b	$FA,$CF,$07,$F5,$7D,$FF,$AC,$6C,$64,$FD,$8C,$65,$F8,$BF,$36,$F3
	.dc.b	$3E,$56,$48,$C4,$FD,$16,$C7,$FA,$8D,$AF,$D8,$73,$97,$DB,$6C,$FF
	.dc.b	$BC,$FA,$5C,$FF,$42,$73,$7F,$A7,$E7,$FD,$51,$FA,$BE,$6B,$87,$ED
	.dc.b	$A2,$CF,$ED,$A6,$E3,$FB,$69,$B5,$DE,$6D,$7E,$13,$E6,$DE,$47,$B9
	.dc.b	$B5,$F8,$D6,$CF,$FA,$FD,$AF,$E5,$1C,$E1,$F9,$76,$CF,$B7,$9F,$F1
	.dc.b	$71,$8A,$FC,$9F,$CD,$FF,$5A,$7F,$3A,$D7,$F2,$E1,$C3,$C3,$C0,$FA
	.dc.b	$71,$FD,$8B,$0F,$D3,$37,$0F,$CF,$C3,$F1,$4F,$EC,$2E,$AB,$F8,$76
	.dc.b	$32,$BF,$85,$FC,$E5,$C7,$F7,$53,$6F,$F4,$33,$6B,$F5,$7D,$BF,$EE
	.dc.b	$4F,$9C,$FE,$30,$7F,$19,$F9,$4B,$3F,$C5,$A2,$E3,$FC,$16,$6E,$1F
	.dc.b	$AC,$F9,$6F,$E4,$07,$F2,$1E,$8B,$87,$F1,$E8,$B7,$FC,$36,$6D,$7E
	.dc.b	$D3,$E7,$3F,$94,$1F,$CA,$7E,$72,$E1,$FC,$9A,$2D,$FF,$15,$9B,$5F
	.dc.b	$BC,$F8,$CE,$17,$FC,$B1,$B5,$FC,$6D,$B3,$B9,$FE,$63,$5F,$C1,$AB
	.dc.b	$FD,$72,$3E,$71,$F0,$DF,$6A,$DE,$7F,$F9,$72,$73,$5F,$C7,$3E,$2B
	.dc.b	$F8,$AC,$62,$15,$C3,$1C,$E1,$FC,$8A,$57,$87,$EF,$E1,$D6,$75,$39
	.dc.b	$9E,$73,$97,$C0,$79,$E7,$CD,$19,$8F,$F6,$D9,$93,$AD,$73,$7D,$F8
	.dc.b	$79,$F0,$F3,$46,$CE,$31,$9A,$E9,$1C,$2F,$BD,$F3,$8F,$59,$C6,$F9
	.dc.b	$DF,$C3,$8E,$07,$88,$EC,$60,$F6,$CF,$3C,$3E,$05,$72,$BF,$87,$5D
	.dc.b	$70,$F1,$BE,$0C,$3C,$D8,$71,$8D,$36,$33,$1E,$03,$85,$7A,$9C,$61
	.dc.b	$E0,$BE,$AD,$BE,$55,$BB,$F6,$5F,$23,$D8,$E2,$B8,$47,$43,$C6,$58
	.dc.b	$F1,$9B,$72,$D3,$33,$F0,$1B,$1F,$80,$F2,$5E,$5F,$C4,$6C,$66,$30
	.dc.b	$7F,$12,$F8,$D7,$85,$F9,$49,$57,$05,$3A,$D3,$FB,$7C,$E6,$31,$3C
	.dc.b	$2F,$A5,$F3,$AD,$3D,$70,$FD,$16,$2F,$81,$FE,$47,$18,$6C,$72,$CB
	.dc.b	$F3,$D9,$76,$AD,$65,$FA,$1A,$E4,$EB,$0E,$D1,$C6,$3F,$30,$E9,$B8
	.dc.b	$76,$7F,$C1,$37,$D6,$78,$5F,$69,$D9,$D7,$08,$E0,$91,$FB,$DA,$E5
	.dc.b	$1D,$30,$F1,$BF,$E1,$6B,$F1,$57,$53,$84,$EA,$7E,$7B,$67,$67,$F4
	.dc.b	$82,$C7,$5B,$F1,$D7,$39,$D6,$1E,$DA,$E7,$5A,$9C,$C7,$C9,$9D,$61
	.dc.b	$E8,$8D,$9E,$C9,$DB,$FC,$06,$34,$6E,$3C,$71,$FC,$92,$79,$E1,$EB
	.dc.b	$AD,$47,$8E,$B7,$1C,$A7,$DD,$39,$38,$DF,$9E,$FE,$44,$65,$C7,$2B
	.dc.b	$C8,$7C,$17,$F3,$3D,$FF,$AC,$C7,$FB,$92,$BF,$CD,$7F,$07,$95,$7C
	.dc.b	$6F,$0D,$8D,$90,$FF,$30,$46,$A9,$F3,$94,$93,$BC,$BF,$C9,$71,$88
	.dc.b	$FF,$34,$3C,$70,$EF,$F3,$8F,$82,$A4,$AF,$F4,$0E,$75,$FE,$8E,$FF
	.dc.b	$74,$9E,$11,$FE,$02,$8C,$4F,$95,$EB,$E4,$49,$D1,$F0,$2B,$FC,$80
	.dc.b	$E3,$44,$3E,$0B,$84,$F8,$0E,$15,$8C,$FF,$CF,$B5,$B9,$E4,$BE,$6B
	.dc.b	$DA,$48,$CF,$C1,$D7,$A6,$3E,$09,$C5,$F4,$AF,$AD,$4F,$9C,$7B,$98
	.dc.b	$6C,$7F,$9F,$9A,$F9,$CF,$E3,$7D,$BB,$9F,$F4,$54,$73,$6C,$F0,$8F
	.dc.b	$F4,$8B,$5F,$F1,$AF,$17,$FA,$1D,$FF,$45,$B5,$EC,$DC,$3F,$D3,$5D
	.dc.b	$D7,$7E,$3E,$9F,$82,$B8,$5F,$FA,$79,$AE,$8F,$F4,$F7,$C4,$3C,$1F
	.dc.b	$82,$D8,$FF,$4D,$B5,$F7,$2C,$37,$32,$76,$F2,$8C,$47,$08,$B8,$FF
	.dc.b	$AC,$6B,$94,$FF,$AA,$3E,$71,$E1,$FD,$26,$C6,$63,$EA,$FC,$DF,$F4
	.dc.b	$5B,$FF,$5D,$56,$AB,$FC,$34,$D7,$FD,$7A,$D7,$FC,$C2,$D7,$F6,$7F
	.dc.b	$39,$F8,$97,$F2,$8B,$87,$F6,$AD,$8C,$F1,$60,$CC,$C8,$C9,$FE,$C2
	.dc.b	$C1,$F1,$1E,$82,$46,$7D,$DA,$FE,$26,$76,$F1,$9F,$7C,$EB,$2F,$F6
	.dc.b	$C5,$6A,$B8,$56,$9F,$D1,$FC,$E7,$EF,$88,$FE,$9B,$DE,$AF,$E8,$BA
	.dc.b	$13,$8C,$3F,$5B,$5F,$A2,$F9,$CF,$81,$7F,$9D,$6C,$FF,$BA,$5B,$CB
	.dc.b	$FE,$8C,$F9,$47,$C7,$F1,$DB,$F4,$35,$CB,$0F,$F7,$7B,$5F,$F4,$9B
	.dc.b	$5F,$E0,$DF,$39,$FC,$58,$8F,$EF,$BE,$72,$BE,$98,$B9,$7F,$04,$9B
	.dc.b	$97,$F1,$79,$B5,$FC,$46,$7F,$90,$FC,$4F,$F2,$2F,$9C,$93,$FC,$2A
	.dc.b	$32,$FF,$0C,$8C,$BF,$1E,$3F,$94,$E9,$FC,$AB,$E2,$E4,$FF,$12,$8C
	.dc.b	$BF,$C5,$23,$2F,$CF,$8F,$E1,$A3,$0A,$48,$FE,$4E,$21,$86,$63,$F8
	.dc.b	$74,$71,$76,$D7,$31,$BC,$99,$7D,$5B,$C3,$F5,$E7,$FA,$1F,$CB,$4B
	.dc.b	$FE,$56,$FF,$46,$F9,$7F,$D0,$FC,$BC,$9F,$D2,$BF,$A4,$F3,$7F,$A5
	.dc.b	$B2,$3F,$99,$BF,$D3,$A3,$FA,$7F,$CB,$49,$FE,$6E,$FF,$52,$E2,$FE
	.dc.b	$A7,$F2,$D2,$7F,$9D,$BF,$D5,$BC,$DF,$EC,$BF,$36,$F6,$E9,$2B,$FA
	.dc.b	$09,$AE,$A8,$FF,$4D,$9F,$D3,$70,$FC,$28,$FE,$E3,$0E,$CF,$EE,$3F
	.dc.b	$38,$F9,$4A,$4A,$FE,$AD,$0F,$95,$F1,$FC,$B7,$93,$C9,$95,$DE,$3C
	.dc.b	$F7,$E8,$A9,$3E,$38,$F5,$FC,$1C,$8F,$33,$EE,$F9,$6F,$7F,$AB,$27
	.dc.b	$D4,$FC,$2F,$96,$F3,$FC,$5F,$A3,$DE,$95,$E6,$8F,$5F,$CB,$48,$F4
	.dc.b	$BA,$7F,$B0,$EC,$F9,$6F,$82,$48,$D9,$F2,$7D,$BF,$A2,$47,$E6,$7A
	.dc.b	$9F,$53,$19,$1F,$3D,$D3,$FD,$BB,$8F,$FD,$AA,$FD,$3C,$BE,$D9,$C1
	.dc.b	$F1,$FB,$DF,$48,$8F,$D0,$F7,$EB,$A3,$E9,$C7,$D8,$8C,$F3,$7D,$DF
	.dc.b	$61,$AF,$FA,$2A,$F6,$73,$3F,$09,$B7,$E3,$08,$FE,$1F,$6A,$FA,$62
	.dc.b	$DF,$5C,$FE,$44,$FE,$37,$1F,$CA,$91,$FC,$87,$CA,$D7,$E1,$7C,$B7
	.dc.b	$E6,$8C,$5F,$9F,$E7,$1F,$2E,$A9,$F0,$8C,$31,$B3,$3F,$65,$7F,$BD
	.dc.b	$A7,$31,$8A,$E1,$3F,$E4,$EC,$3A,$BD,$2B,$9C,$62,$39,$1F,$A4,$DF
	.dc.b	$C4,$E3,$0F,$04,$62,$FA,$C2,$B9,$47,$82,$9F,$36,$F8,$7F,$A5,$39
	.dc.b	$8C,$61,$E3,$AF,$81,$18,$7C,$67,$2F,$94,$F0,$65,$E4,$9D,$5F,$3C
	.dc.b	$3A,$CE,$5B,$7D,$2F,$8B,$DA,$32,$F9,$9B,$7D,$2B,$B5,$70,$AD,$61
	.dc.b	$E7,$69,$E6,$AE,$0F,$73,$4E,$1D,$E1,$E7,$9C,$D7,$5A,$F8,$15,$F4
	.dc.b	$0E,$A5,$72,$FB,$67,$4F,$FB,$23,$95,$F3,$98,$8F,$F7,$3C,$61,$DF
	.dc.b	$C3,$FF,$96,$A7,$4D,$8E,$77,$FE,$9A,$F1,$79,$36,$33,$F0,$09,$CB
	.dc.b	$63,$31,$F0,$23,$19,$FE,$0F,$C8,$7A,$5F,$CC,$63,$F0,$53,$AD,$BE
	.dc.b	$E3,$70,$FC,$1B,$55,$F2,$F3,$FB,$63,$27,$E9,$56,$EB,$51,$89,$D6
	.dc.b	$1E,$07,$ED,$4E,$A3,$13,$B2,$3E,$9C,$7D,$18,$F9,$AF,$F3,$1C,$5F
	.dc.b	$A1,$F7,$73,$87,$EB,$DB,$1F,$A5,$F8,$CB,$C9,$FE,$5C,$D7,$8B,$6F
	.dc.b	$F5,$CD,$CF,$F5,$AA,$B8,$E1,$CB,$E0,$3F,$AD,$74,$7F,$09,$F8,$BB
	.dc.b	$97,$EF,$26,$E5,$FC,$06,$6E,$7F,$B7,$9B,$92,$6C,$61,$CF,$C5,$EB
	.dc.b	$E1,$6E,$3F,$EB,$7B,$46,$31,$FF,$55,$C6,$27,$55,$C5,$FF,$5F,$60
	.dc.b	$F1,$BC,$6F,$F6,$F8,$CC,$63,$C8,$CB,$63,$27,$28,$CB,$4C,$E1,$FE
	.dc.b	$BC,$AE,$8F,$93,$93,$DF,$D7,$F0,$1B,$5E,$BF,$89,$C3,$4F,$74,$7C
	.dc.b	$68,$91,$F0,$7E,$2A,$DF,$7F,$27,$E5,$FC,$AF,$B1,$B5,$F0,$3C,$9F
	.dc.b	$7F,$77,$C0,$6C,$7C,$4F,$23,$E1,$9F,$51,$A7,$FB,$97,$DB,$FF,$72
	.dc.b	$37,$E2,$BE,$DC,$7F,$BE,$63,$06,$FD,$FC,$FC,$56,$9B,$1C,$FE,$07
	.dc.b	$89,$CB,$E1,$6C,$77,$F7,$BA,$9D,$BE,$0E,$C7,$FB,$DF,$F2,$27,$07
	.dc.b	$E3,$1F,$EB,$31,$98,$C7,$99,$E6,$F8,$39,$3E,$BF,$82,$BC,$9F,$EF
	.dc.b	$2D,$C7,$B3,$C7,$F6,$B6,$BE,$17,$C0,$FC,$0F,$83,$F8,$2D,$8F,$8D
	.dc.b	$F0,$3F,$9C,$F4,$3E,$2B,$63,$E5,$7C,$0F,$E8,$3E,$57,$E4,$B6,$3D
	.dc.b	$9E,$07,$AF,$E2,$FD,$8D,$8F,$A5,$F0,$7F,$43,$D0,$FA,$2D,$8F,$AD
	.dc.b	$BF,$D4,$F3,$FE,$AB,$63,$27,$DB,$8C,$B4,$CC,$7C,$38,$CB,$63,$FC
	.dc.b	$09,$E6,$7F,$87,$7F,$26,$FF,$2C,$46,$0F,$F0,$2B,$89,$4E,$FE,$13
	.dc.b	$F8,$0C,$E6,$39,$7C,$1E,$0C,$CD,$BF,$F4,$8F,$CD,$FF,$11,$7B,$61
	.dc.b	$D3,$E7,$DD,$1F,$C3,$A1,$5B,$7C,$35,$E8,$AD,$BD,$3D,$3F,$F3,$94
	.dc.b	$FF,$5F,$AE,$0D,$8C,$BF,$E1,$28,$CB,$EE,$F8,$5F,$5F,$C3,$7B,$1B
	.dc.b	$7E,$6F,$51,$B2,$73,$7E,$41,$3A,$8C,$ED,$F3,$5B,$7D,$CE,$97,$FE
	.dc.b	$FA,$6F,$C2,$FF,$2A,$F9,$6F,$9A,$D8,$C9,$31,$F0,$63,$E4,$F8,$7D
	.dc.b	$CF,$4F,$4F,$EA,$FA,$7F,$59,$B1,$93,$C3,$19,$69,$F5,$34,$7C,$E7
	.dc.b	$FA,$56,$30,$AF,$FA,$1C,$EB,$2B,$39,$AD,$C6,$2B,$2D,$8C,$9D,$23
	.dc.b	$2D,$3F,$DA,$FF,$32,$F3,$3B,$FC,$7F,$F3,$2C,$7E,$FC,$E6,$BD,$13
	.dc.b	$97,$FC,$9F,$19,$D7,$FC,$3A,$D7,$FC,$48,$D7,$77,$F8,$67,$FC,$D1
	.dc.b	$1F,$8E,$7E,$7F,$E4,$8E,$EF,$CB,$4E,$71,$FF,$43,$4E,$70,$F4,$F3
	.dc.b	$E8,$FE,$CF,$F9,$2B,$CE,$C4,$FC,$FF,$9C,$7C,$F7,$CC,$5F,$3D,$63
	.dc.b	$0D,$9E,$0D,$D7,$F2,$59,$AF,$8D,$75,$BE,$2D,$38,$9F,$B2,$A7,$C4
	.dc.b	$7F,$32,$8F,$C1,$76,$FF,$2A,$75,$19,$9E,$F3,$97,$F4,$47,$2F,$3B
	.dc.b	$F7,$53,$BC,$7F,$C1,$BC,$AF,$57,$35,$FB,$FF,$C9,$BE,$0E,$5F,$AF
	.dc.b	$E1,$FF,$1A,$BE,$58,$CE,$5F,$4B,$C5,$E1,$AC,$AF,$DB,$70,$BC,$FC
	.dc.b	$5F,$E9,$BB,$9F,$1F,$FC,$63,$E4,$3F,$8C,$B7,$0F,$D9,$B6,$B8,$37
	.dc.b	$83,$F9,$37,$B5,$FC,$83,$DB,$FE,$42,$DB,$FD,$DB,$6F,$C0,$DA,$FE
	.dc.b	$5D,$C3,$FC,$A3,$DB,$FE,$52,$DB,$FE,$06,$DB,$F4,$36,$BF,$9B,$7C
	.dc.b	$77,$F3,$0F,$90,$FE,$62,$DB,$FE,$16,$DB,$F8,$0D,$AF,$E7,$DF,$11
	.dc.b	$FC,$E0,$FE,$73,$F1,$F6,$FF,$89,$C5,$BF,$91,$36,$BF,$49,$F1,$3F
	.dc.b	$EC,$BF,$9A,$7A,$BF,$EB,$C9,$66,$CB,$53,$FE,$8C,$BB,$8F,$F5,$0A
	.dc.b	$D2,$C3,$9B,$FE,$97,$9D,$BF,$D4,$27,$73,$A8,$FF,$64,$1C,$6B,$C5
	.dc.b	$3A,$7C,$11,$93,$C1,$7D,$1F,$F7,$19,$93,$B3,$95,$7E,$2C,$3D,$B3
	.dc.b	$98,$C5,$F7,$AF,$41,$CE,$B6,$F5,$3B,$E1,$E9,$8C,$BE,$B8,$CE,$1E
	.dc.b	$AC,$3B,$47,$3B,$EC,$FC,$0A,$E5,$18,$8E,$27,$B8,$F0,$1E,$A3,$94
	.dc.b	$F4,$9E,$2D,$AF,$04,$F0,$C3,$DB,$39,$8E,$95,$E7,$7A,$3C,$9E,$07
	.dc.b	$62,$B9,$38,$4F,$48,$CD,$78,$27,$31,$C5,$A7,$18,$FB,$6D,$8F,$9E
	.dc.b	$D7,$F0,$7F,$1C,$F9,$0F,$C0,$B6,$3E,$AB,$5D,$C1,$9F,$1C,$E0,$FF
	.dc.b	$0F,$C2,$B8,$4E,$74,$E1,$18,$AD,$B6,$33,$7D,$6D,$7C,$5F,$C8,$62
	.dc.b	$F5,$75,$97,$E7,$63,$13,$FA,$2A,$E1,$3F,$A1,$6C,$F5,$37,$3D,$2B
	.dc.b	$9C,$78,$67,$73,$EE,$CB,$DB,$7D,$EF,$73,$F0,$EF,$9C,$F7,$AD,$C7
	.dc.b	$36,$CE,$EB,$8B,$FC,$11,$B7,$E0,$8F,$83,$3C,$1D,$1F,$BE,$AE,$E7
	.dc.b	$5C,$7E,$04,$EE,$33,$3D,$CC,$CC,$7E,$01,$CE,$FC,$35,$DE,$FC,$57
	.dc.b	$CA,$BC,$27,$F0,$18,$F3,$1A,$AF,$5C,$EE,$33,$5E,$39,$C9,$DA,$FC
	.dc.b	$58,$7C,$BC,$3C,$46,$30,$F6,$3E,$62,$47,$CD,$9D,$47,$B4,$CC,$F0
	.dc.b	$9D,$46,$0D,$38,$3F,$94,$F9,$BF,$E5,$6C,$BF,$9A,$6B,$FE,$7A,$6D
	.dc.b	$7F,$95,$B9,$FF,$45,$8F,$FA,$07,$C0,$CE,$1C,$6B,$FC,$AB,$78,$BD
	.dc.b	$E1,$FE,$84,$BD,$FE,$54,$F7,$F5,$C1,$31,$FF,$34,$4E,$5B,$39,$7F
	.dc.b	$BA,$4F,$28,$E7,$ED,$F4,$F0,$A5,$72,$8C,$BD,$23,$2F,$F9,$8A,$D3
	.dc.b	$E4,$F4,$7C,$BE,$DC,$AF,$1C,$65,$F2,$C6,$5E,$51,$EC,$F2,$3D,$BF
	.dc.b	$41,$27,$D7,$19,$7D,$B1,$97,$C5,$1F,$39,$FF,$23,$79,$57,$CD,$FF
	.dc.b	$20,$47,$F8,$BD,$AF,$F8,$A3,$F2,$8F,$9A,$FF,$13,$8C,$F2,$7F,$A7
	.dc.b	$25,$DB,$D5,$B5,$CE,$FC,$1F,$95,$7F,$B8,$23,$FC,$78,$D7,$FC,$6A
	.dc.b	$DB,$FF,$53,$FE,$55,$FE,$4E,$9D,$FC,$17,$A6,$1A,$D3,$AC,$7B,$7E
	.dc.b	$93,$FC,$5D,$5F,$E4,$95,$F0,$F9,$35,$E4,$C9,$D4,$65,$E1,$19,$7F
	.dc.b	$C9,$33,$C9,$D9,$A3,$FC,$AD,$F3,$AE,$2F,$9C,$D8,$CC,$7D,$1F,$9E
	.dc.b	$7C,$AE,$A3,$1F,$43,$2B,$67,$49,$FE,$6C,$CB,$E2,$E0,$FE,$8C,$E1
	.dc.b	$C3,$5E,$11,$F8,$1F,$9F,$FA,$B1,$FE,$64,$6E,$FF,$ED,$6A,$E4,$FF
	.dc.b	$B6,$21,$E6,$71,$F3,$65,$FF,$9D,$67,$2D,$9C,$BD,$67,$C1,$12,$38
	.dc.b	$46,$AB,$B4,$FE,$E3,$1F,$DC,$FA,$12,$FF,$6F,$19,$7F,$73,$19,$7F
	.dc.b	$5D,$1F,$C0,$32,$91,$FB,$E7,$F8,$1F,$C1,$62,$3F,$DA,$BF,$2A,$EB
	.dc.b	$FE,$1A,$D8,$E0,$7F,$06,$F8,$8F,$E1,$7F,$2F,$23,$04,$3F,$7C,$E5
	.dc.b	$FE,$01,$1F,$C4,$8C,$69,$FE,$4C,$F1,$DE,$02,$F5,$7F,$E9,$FF,$9E
	.dc.b	$7C,$8F,$DD,$8F,$F0,$14,$61,$C3,$5F,$C1,$35,$FC,$3B,$9C,$3F,$CB
	.dc.b	$70,$F8,$DF,$F1,$F7,$91,$FE,$01,$6C,$76,$BF,$E5,$3F,$31,$23,$F9
	.dc.b	$3B,$FC,$B3,$E3,$71,$1F,$C7,$62,$E1,$ED,$CD,$5F,$C3,$F0,$F0,$5F
	.dc.b	$FA,$ED,$FD,$F6,$3E,$1E,$DD,$44,$9D,$47,$13,$0F,$FA,$EA,$39,$B6
	.dc.b	$34,$F5,$8F,$E8,$3F,$2F,$FD,$0B,$E6,$24,$FF,$40,$8F,$E9,$7E,$C7
	.dc.b	$F3,$48,$FF,$17,$67,$F5,$61,$E5,$68,$DF,$95,$FD,$33,$E5,$FF,$0F
	.dc.b	$E5,$CC,$3F,$A5,$C7,$F5,$0D,$7F,$17,$D7,$27,$FA,$7C,$7F,$5A,$FC
	.dc.b	$B1,$F2,$F7,$E8,$A5,$78,$AB,$FC,$81,$D0,$BD,$CE,$44,$67,$9B,$A7
	.dc.b	$C4,$49,$D4,$E5,$B5,$FC,$E2,$FA,$B6,$39,$B6,$7C,$1F,$31,$F1,$BA
	.dc.b	$3E,$3F,$3C,$BE,$BC,$49,$F6,$79,$32,$7B,$BE,$FF,$98,$F8,$1F,$13
	.dc.b	$27,$C6,$FC,$3F,$2F,$E2,$75,$FC,$5F,$97,$95,$E8,$8F,$91,$F2,$F2
	.dc.b	$3D,$6F,$CB,$ED,$FF,$02,$FE,$59,$F1,$A5,$D9,$FA,$F8,$C3,$79,$FE
	.dc.b	$97,$D0,$F0,$E6,$E3,$C7,$30,$F9,$38,$27,$AF,$3F,$F0,$6F,$22,$56
	.dc.b	$A3,$EE,$7D,$0C,$8F,$9E,$FD,$FF,$7B,$11,$E4,$EA,$FA,$3F,$13,$F8
	.dc.b	$33,$3B,$7F,$05,$EB,$47,$E1,$FD,$99,$7F,$6A,$3D,$FC,$BF,$03,$A2
	.dc.b	$4F,$A2,$32,$FA,$A3,$2F,$84,$FC,$4E,$FF,$D9,$9C,$76,$23,$19,$7C
	.dc.b	$9A,$E2,$FC,$BB,$FC,$D7,$CD,$7E,$6F,$E0,$66,$3F,$8E,$8C,$BF,$90
	.dc.b	$8C,$BD,$B4,$4F,$F8,$F3,$C9,$7C,$D5,$F9,$A9,$7F,$9A,$9C,$B6,$72
	.dc.b	$FA,$21,$3C,$E7,$A7,$CC,$7F,$A7,$7E,$1F,$FC,$27,$BF,$F2,$F8,$C7
	.dc.b	$BF,$FA,$EC,$11,$8C,$7F,$5B,$0F,$99,$C5,$F1,$9F,$F5,$A6,$1D,$2B
	.dc.b	$8C,$63,$0E,$98,$70,$9E,$53,$93,$FD,$79,$F3,$CF,$37,$FD,$75,$1F
	.dc.b	$EF,$CB,$FF,$5A,$4E,$7E,$0F,$84,$A3,$0E,$CE,$7E,$6F,$A5,$D5,$F4
	.dc.b	$CB,$FB,$FD,$70,$6C,$67,$0E,$F7,$C6,$B5,$1E,$29,$D5,$FA,$1A,$F8
	.dc.b	$EB,$84,$62,$B8,$CE,$F0,$F4,$C7,$F2,$F8,$F4,$1E,$68,$CC,$F1,$39
	.dc.b	$57,$48,$FE,$C5,$5F,$02,$BF,$C4,$6D,$7C,$D1,$FD,$9F,$E2,$5F,$1A
	.dc.b	$F9,$FC,$36,$BF,$C5,$13,$A9,$C9,$DB,$E7,$FF,$2B,$3F,$E2,$DC,$11
	.dc.b	$88,$F0,$B5,$FA,$ED,$7E,$EB,$CB,$4F,$A2,$DD,$3F,$DE,$1D,$4F,$21
	.dc.b	$DB,$F1,$AB,$58,$72,$E6,$C1,$F8,$2B,$E1,$58,$9F,$97,$97,$9A,$B1
	.dc.b	$97,$F8,$FA,$F7,$5A,$C3,$AB,$F8,$B6,$DF,$18,$FA,$B3,$C2,$BD,$37
	.dc.b	$D3,$2F,$7B,$E3,$CB,$D3,$7D,$63,$F3,$DC,$6F,$5F,$FC,$BF,$A6,$1B
	.dc.b	$5F,$99,$9D,$7B,$3C,$B2,$FD,$BD,$69,$BE,$D6,$73,$FE,$CB,$39,$6D
	.dc.b	$67,$0F,$EC,$B3,$96,$FA,$1E,$6E,$EF,$F1,$94,$FF,$8C,$73,$FE,$CE
	.dc.b	$FF,$86,$1B,$CF,$B7,$BD,$5B,$E7,$1C,$3D,$9C,$6B,$EF,$AB,$83,$D7
	.dc.b	$BE,$8E,$1B,$87,$81,$B8,$FF,$95,$1B,$9F,$F9,$1F,$07,$CE,$76,$5F
	.dc.b	$39,$63,$0D,$C3,$FC,$03,$3A,$6C,$E9,$A6,$A7,$86,$1F,$D0,$AB,$83
	.dc.b	$FE,$66,$E6,$F5,$F6,$7B,$1B,$7F,$23,$D7,$C3,$4F,$97,$1F,$0B,$E7
	.dc.b	$FE,$1C,$D9,$D5,$F7,$F7,$BD,$87,$B7,$B6,$CF,$AB,$E3,$F0,$D3,$DB
	.dc.b	$1F,$0F,$D8,$91,$F6,$8B,$7F,$D7,$3D,$6F,$F1,$7C,$7F,$89,$1B,$19
	.dc.b	$F6,$7E,$CC,$63,$87,$FC,$EC,$FF,$9C,$AB,$87,$A5,$AF,$47,$6F,$0F
	.dc.b	$89,$93,$D7,$D4,$B9,$78,$7C,$FE,$A7,$6F,$6A,$D7,$F8,$DB,$E7,$FF
	.dc.b	$C9,$ED,$F8,$CF,$C5,$C6,$2F,$FB,$AA,$8F,$3F,$B9,$2B,$D9,$E9,$DB
	.dc.b	$F8,$1B,$7B,$3C,$7F,$6B,$6B,$E1,$7B,$9F,$03,$DC,$F8,$2D,$8F,$8D
	.dc.b	$EC,$7C,$43,$E2,$FB,$76,$3E,$57,$B3,$F2,$0F,$93,$ED,$58,$F6,$78
	.dc.b	$1E,$AF,$1F,$D6,$D8,$FA,$5E,$CF,$D0,$F3,$FE,$8B,$63,$EB,$6F,$F5
	.dc.b	$3C,$FF,$AA,$D8,$C9,$F6,$E3,$2D,$33,$1F,$81,$8C,$B6,$3F,$CC,$BF
	.dc.b	$1D,$FE,$10,$6F,$B1,$FB,$A8,$FF,$05,$C6,$09,$3F,$E2,$68,$E0,$DF
	.dc.b	$67,$85,$7F,$05,$87,$CF,$BD,$0F,$F5,$8C,$09,$19,$C3,$FD,$7C,$D8
	.dc.b	$FF,$7E,$35,$FF,$80,$2E,$7C,$0F,$F5,$DB,$67,$FD,$33,$F3,$3F,$E1
	.dc.b	$A7,$FD,$0D,$5E,$2B,$F6,$7B,$5E,$DF,$6E,$D7,$C2,$6D,$7B,$5B,$8F
	.dc.b	$FB,$5F,$4F,$F1,$15,$7C,$8B,$E4,$D3,$91,$F3,$3E,$7F,$E5,$1F,$2F
	.dc.b	$E7,$EC,$64,$98,$FC,$B8,$CB,$F3,$7D,$0F,$A5,$E8,$7D,$36,$D7,$D8
	.dc.b	$6C,$EE,$3E,$DC,$FF,$35,$C9,$AB,$1C,$2C,$6C,$61,$B7,$FE,$AD,$AE
	.dc.b	$31,$8C,$3F,$CD,$11,$89,$D5,$F5,$6D,$6E,$7F,$A7,$4F,$79,$FA,$7F
	.dc.b	$3D,$FE,$8B,$7F,$DF,$8D,$BF,$C2,$11,$C3,$80,$E1,$D7,$8F,$F8,$F8
	.dc.b	$E5,$7F,$D7,$7E,$79,$F1,$B9,$AD,$78,$73,$1F,$F6,$2F,$CD,$BC,$AC
	.dc.b	$67,$FE,$85,$3B,$D6,$32,$FE,$FB,$EC,$2F,$86,$1E,$CA,$FA,$19,$BE
	.dc.b	$1A,$FC,$3E,$1B,$5F,$36,$74,$D9,$D3,$4D,$4F,$0C,$BE,$8D,$70,$6D
	.dc.b	$7E,$BF,$E5,$7F,$93,$67,$FE,$12,$94,$63,$0F,$D2,$FA,$3B,$F8,$99
	.dc.b	$1F,$A6,$9D,$57,$FB,$23,$89,$7A,$8F,$89,$F3,$9E,$1C,$7B,$37,$0F
	.dc.b	$F7,$BE,$CF,$33,$F8,$14,$E6,$FF,$80,$35,$39,$E0,$F9,$93,$C3,$7F
	.dc.b	$E2,$5F,$1D,$FC,$4F,$E7,$2E,$5F,$C6,$1B,$8F,$F1,$36,$E7,$FB,$77
	.dc.b	$F8,$D7,$AD,$FC,$6F,$D8,$B7,$FC,$81,$B7,$FC,$6D,$B7,$FB,$F7,$F9
	.dc.b	$16,$DF,$C8,$FD,$8B,$7F,$CA,$1B,$7F,$C8,$DB,$7F,$C1,$DF,$E5,$5F
	.dc.b	$39,$FC,$AF,$E6,$ED,$FF,$30,$6D,$FF,$2B,$6D,$FF,$0F,$7F,$99,$7C
	.dc.b	$1C,$AF,$E6,$FF,$39,$6F,$F9,$94,$DB,$FE,$2D,$36,$BF,$51,$4F,$86
	.dc.b	$FE,$61,$F3,$8F,$4F,$F5,$F8,$32,$FF,$2C,$CE,$6F,$F6,$17,$C2,$F7
	.dc.b	$5A,$7A,$C6,$67,$F9,$E5,$F3,$C3,$11,$FE,$D1,$18,$75,$53,$B7,$37
	.dc.b	$D1,$A7,$4C,$3C,$0F,$90,$C9,$C9,$CB,$DE,$BB,$57,$57,$C1,$1F,$ED
	.dc.b	$83,$8C,$66,$BC,$13,$C2,$B9,$D6,$E3,$17,$EC,$8C,$57,$C4,$8C,$47
	.dc.b	$CF,$3A,$D7,$27,$15,$C9,$F4,$9C,$8F,$74,$F2,$C3,$E4,$E1,$E5,$9F
	.dc.b	$81,$5C,$4F,$5B,$D3,$0F,$13,$D0,$C1,$D0,$C5,$74,$C3,$C8,$DC,$3C
	.dc.b	$13,$E8,$6D,$F7,$BF,$8A,$76,$7C,$35,$C6,$BA,$9C,$4B,$1C,$5F,$4C
	.dc.b	$F1,$39,$BE,$58,$E6,$BE,$2D,$F3,$9F,$83,$6C,$7E,$05,$AF,$E3,$5B
	.dc.b	$F1,$8D,$FF,$1B,$3C,$EB,$FA,$6C,$E7,$4F,$F0,$95,$F0,$BD,$B6,$33
	.dc.b	$5C,$27,$9E,$1F,$99,$C5,$E8,$7F,$58,$8C,$36,$38,$5F,$E8,$B1,$F0
	.dc.b	$65,$F9,$89,$E2,$FE,$99,$AF,$E7,$23,$77,$FA,$76,$CF,$F6,$63,$71
	.dc.b	$E0,$9E,$D8,$79,$A7,$C3,$1D,$F0,$F2,$3F,$E3,$18,$E2,$6A,$BB,$57
	.dc.b	$19,$D5,$70,$7F,$84,$36,$FC,$71,$FE,$3B,$6B,$A3,$D9,$1B,$9E,$90
	.dc.b	$9D,$DF,$8A,$3B,$9E,$AC,$23,$E9,$39,$98,$C3,$CF,$18,$8D,$1F,$26
	.dc.b	$FB,$DF,$6C,$3D,$58,$77,$AF,$15,$68,$F2,$C6,$4D,$57,$B0,$E5,$3D
	.dc.b	$F0,$F0,$BD,$F0,$C5,$7D,$5C,$3E,$04,$79,$E3,$13,$A3,$E4,$61,$F3
	.dc.b	$FA,$1E,$57,$E7,$2F,$FC,$A1,$9F,$F3,$6D,$BF,$CD,$15,$A8,$C6,$08
	.dc.b	$E3,$DB,$87,$51,$FE,$7E,$9C,$61,$D6,$F9,$BF,$E0,$69,$FE,$73,$1D
	.dc.b	$BE,$4B,$FD,$5D,$1F,$E8,$6D,$F8,$9D,$63,$31,$FC,$EA,$78,$7B,$9A
	.dc.b	$8E,$07,$43,$CD,$3C,$33,$FF,$15,$AF,$8E,$E9,$E0,$CA,$E1,$A2,$B5
	.dc.b	$18,$6B,$E1,$3A,$CF,$17,$13,$C6,$BD,$1F,$33,$E9,$F2,$25,$73,$8C
	.dc.b	$BD,$63,$2F,$2C,$13,$F0,$7C,$4F,$85,$F2,$B2,$BC,$91,$97,$CD,$19
	.dc.b	$79,$47,$D8,$5E,$D7,$F9,$45,$FF,$4B,$B5,$FF,$29,$37,$4F,$F1,$E7
	.dc.b	$CE,$7D,$99,$FF,$55,$4E,$77,$FF,$34,$B6,$33,$E5,$77,$AE,$10,$F9
	.dc.b	$C7,$C0,$78,$DB,$1F,$EA,$86,$BF,$E6,$3F,$9C,$67,$BA,$93,$B7,$D9
	.dc.b	$3A,$8F,$F5,$8F,$CE,$7F,$90,$DB,$F0,$DF,$06,$0D,$F2,$A4,$E8,$87
	.dc.b	$F9,$CE,$79,$36,$39,$4C,$8E,$F1,$F3,$7E,$6F,$FD,$03,$1F,$EC,$26
	.dc.b	$BF,$E7,$D6,$EF,$FE,$67,$F9,$C7,$CA,$71,$E5,$97,$C5,$FB,$13,$A8
	.dc.b	$E9,$3D,$C8,$FF,$B4,$3E,$73,$E7,$E1,$B8,$FF,$19,$B5,$FF,$4E,$35
	.dc.b	$FF,$50,$B5,$FF,$73,$43,$CA,$FC,$42,$F3,$69,$88,$93,$E1,$8C,$BE
	.dc.b	$38,$CB,$FE,$9D,$9E,$4E,$CD,$1F,$AD,$F9,$AF,$DF,$79,$F2,$30,$43
	.dc.b	$F4,$8E,$5F,$D3,$47,$F0,$5C,$64,$7E,$DD,$FE,$11,$EE,$E2,$3F,$59
	.dc.b	$17,$5F,$E0,$73,$FC,$37,$E1,$3F,$87,$7C,$B4,$BF,$E0,$11,$97,$F8
	.dc.b	$14,$65,$FD,$B4,$65,$FD,$C4,$66,$BF,$87,$CF,$EC,$7E,$71,$F1,$F4
	.dc.b	$97,$FC,$5C,$7C,$5F,$ED,$9C,$4F,$07,$FC,$DA,$1C,$3F,$E7,$06,$BF
	.dc.b	$ED,$96,$BF,$CB,$7E,$5B,$F9,$51,$1F,$E3,$9F,$13,$5F,$E2,$11,$71
	.dc.b	$FE,$4F,$3F,$C2,$6F,$CB,$5E,$77,$F6,$F3,$FE,$C1,$6F,$7F,$FB,$B1
	.dc.b	$DC,$FF,$B7,$5B,$3B,$EA,$E8,$75,$89,$3D,$63,$2F,$78,$FE,$85,$F2
	.dc.b	$BF,$D0,$FE,$5A,$4F,$F3,$58,$FE,$A5,$EB,$FF,$36,$C3,$FD,$D2,$BC
	.dc.b	$8C,$9A,$F1,$FF,$A6,$7C,$A4,$C3,$F9,$F3,$FD,$3F,$E5,$BF,$A8,$6B
	.dc.b	$27,$FA,$13,$FD,$4F,$E5,$BF,$B0,$7C,$D3,$E7,$2F,$CC,$4B,$FE,$97
	.dc.b	$39,$6F,$2F,$0F,$85,$91,$83,$7E,$57,$42,$3C,$BE,$63,$DB,$CF,$EE
	.dc.b	$E7,$93,$FD,$63,$89,$3E,$4F,$1A,$4F,$89,$F3,$FC,$C7,$A3,$E1,$E4
	.dc.b	$F9,$9F,$5F,$91,$EC,$EB,$F6,$FC,$BC,$AF,$5C,$7C,$0F,$97,$91,$EE
	.dc.b	$7E,$1F,$6F,$F8,$0F,$E7,$3E,$04,$FF,$64,$6F,$63,$A3,$FC,$1B,$18
	.dc.b	$7F,$C1,$FC,$AF,$98,$E5,$CB,$CF,$92,$63,$FD,$DB,$04,$F7,$E6,$93
	.dc.b	$97,$6F,$F7,$4C,$3F,$D6,$B8,$F6,$F9,$8F,$B2,$7D,$AF,$7A,$D7,$D8
	.dc.b	$F9,$8F,$BA,$61,$E4,$FE,$03,$D5,$91,$B3,$C9,$C7,$F0,$79,$A4,$60
	.dc.b	$87,$C3,$72,$FC,$58,$F9,$06,$38,$3E,$34,$7F,$8B,$F6,$FE,$B2,$F2
	.dc.b	$8C,$FB,$2D,$5F,$31,$F9,$4F,$81,$97,$F9,$28,$FC,$B9,$87,$E2,$46
	.dc.b	$1F,$F1,$0F,$92,$F9,$7F,$F1,$E3,$63,$31,$FA,$07,$2E,$BE,$62,$46
	.dc.b	$8E,$7F,$2C,$67,$C3,$BE,$2B,$FC,$9B,$82,$B5,$4A,$D3,$FE,$4F,$9D
	.dc.b	$3C,$AB,$8F,$CB,$F0,$AD,$EA,$BD,$DF,$3B,$EF,$86,$DB,$1C,$CE,$D8
	.dc.b	$65,$F4,$7C,$DB,$CD,$FF,$63,$47,$33,$2E,$32,$F3,$C2,$78,$7C,$27
	.dc.b	$58,$C7,$9B,$D5,$D5,$F1,$C7,$E0,$DF,$09,$EC,$6A,$F9,$47,$F9,$4B
	.dc.b	$0E,$EE,$DB,$7C,$EB,$CB,$5C,$30,$F0,$5E,$F0,$EF,$7B,$BE,$C7,$82
	.dc.b	$7B,$4E,$AF,$C3,$5B,$9E,$55,$F1,$AF,$AC,$FC,$68,$FB,$67,$42,$CF
	.dc.b	$B5,$B1,$B7,$FD,$CF,$F1,$CF,$9C,$FE,$BF,$F1,$88,$C4,$F4,$3F,$D7
	.dc.b	$8E,$1E,$F1,$B9,$FE,$CF,$6F,$3F,$F4,$B5,$C6,$31,$C9,$A9,$CC,$F9
	.dc.b	$9A,$FD,$47,$8E,$BF,$41,$BA,$FE,$63,$AD,$E3,$7E,$E2,$B5,$87,$D8
	.dc.b	$BD,$E1,$F8,$C6,$AB,$E3,$DF,$BA,$93,$F4,$EB,$75,$A8,$C4,$EA,$FB
	.dc.b	$C7,$E2,$27,$51,$82,$63,$EB,$AF,$CA,$4E,$A3,$15,$F9,$DE,$47,$DA
	.dc.b	$7F,$9C,$B2,$FD,$34,$63,$E0,$BD,$D7,$FE,$C6,$A7,$C2,$E3,$0F,$F6
	.dc.b	$34,$71,$7A,$37,$3F,$DB,$B6,$B9,$57,$1C,$3B,$63,$E0,$A7,$B5,$FC
	.dc.b	$93,$B7,$F9,$0F,$4B,$E2,$F1,$7E,$3A,$BA,$7F,$09,$9B,$8F,$B2,$6D
	.dc.b	$F1,$BF,$F6,$0E,$3D,$9F,$E4,$77,$CB,$07,$DC,$3A,$7A,$F2,$BE,$8D
	.dc.b	$E4,$EE,$B5,$8F,$FB,$4E,$B5,$39,$BF,$E7,$B8,$7E,$9A,$F6,$FE,$75
	.dc.b	$AE,$63,$6F,$F2,$6F,$B8,$7C,$67,$8B,$2F,$F4,$EE,$5F,$E4,$B9,$CB
	.dc.b	$E1,$8C,$D7,$FB,$64,$4E,$9F,$F6,$E6,$5E,$BF,$1B,$E0,$4E,$78,$7F
	.dc.b	$A9,$CE,$70,$ED,$8E,$8F,$87,$2F,$57,$13,$F0,$7E,$3D,$F2,$7F,$43
	.dc.b	$E6,$BE,$4C,$62,$7F,$A8,$C7,$83,$E0,$B3,$56,$79,$37,$5F,$56,$9F
	.dc.b	$EA,$A9,$F9,$DA,$3D,$8D,$69,$FD,$9B,$D8,$7C,$AF,$E0,$3E,$6F,$F0
	.dc.b	$71,$8C,$3F,$B0,$C7,$C0,$E8,$F2,$F3,$D9,$F5,$3F,$80,$C9,$F2,$D9
	.dc.b	$D7,$E4,$C6,$2B,$E9,$5F,$E4,$3E,$C3,$FD,$E3,$1F,$C6,$7B,$11,$8B
	.dc.b	$FE,$4F,$18,$F4,$BE,$FC,$23,$15,$FE,$82,$FB,$7F,$E0,$5B,$FD,$4F
	.dc.b	$B6,$F9,$57,$F7,$67,$35,$FD,$C7,$35,$6A,$FF,$C1,$3F,$2F,$FE,$33
	.dc.b	$AF,$23,$70,$F1,$B5,$FE,$EB,$8A,$31,$1F,$C0,$74,$FD,$6F,$7F,$F7
	.dc.b	$1B,$FD,$5F,$36,$27,$F5,$5A,$3D,$DF,$D4,$E9,$FB,$1C,$7D,$11,$8E
	.dc.b	$1E,$7F,$01,$32,$FD,$AF,$C4,$5B,$6F,$FE,$22,$D5,$F0,$FF,$BC,$DF
	.dc.b	$C6,$DC,$3F,$D5,$DA,$FF,$2F,$7F,$C3,$FF,$3F,$E4,$F2,$BF,$81,$BE
	.dc.b	$AF,$87,$FE,$0B,$F4,$16,$BF,$C4,$AF,$B9,$FE,$13,$E2,$D7,$FC,$52
	.dc.b	$FB,$BD,$9F,$C6,$FC,$75,$9F,$F1,$16,$DF,$93,$8F,$E2,$7F,$6F,$FE
	.dc.b	$89,$9F,$F1,$8F,$7E,$3E,$27,$F8,$AB,$FC,$47,$E8,$9F,$1D,$FC,$73
	.dc.b	$17,$D3,$3E,$2B,$53,$98,$C6,$3F,$EC,$DB,$E2,$F8,$8F,$1C,$6B,$9B
	.dc.b	$F8,$24,$E9,$F9,$B3,$AB,$EA,$FF,$B4,$A7,$53,$FD,$F9,$AF,$F8,$E9
	.dc.b	$AF,$E6,$73,$FF,$70,$CF,$86,$31,$F0,$DF,$0B,$E1,$FE,$1B,$27,$5E
	.dc.b	$02,$F8,$DF,$90,$E9,$7B,$71,$3B,$F3,$3E,$69,$F3,$BD,$1B,$87,$4C
	.dc.b	$3F,$8B,$36,$74,$D3,$CD,$B3,$E8,$9E,$FF,$D1,$6C,$E9,$FF,$67,$36
	.dc.b	$34,$47,$5F,$1F,$73,$F5,$5F,$5A,$C3,$FD,$76,$F9,$CE,$9E,$EF,$D6
	.dc.b	$F8,$0F,$EB,$7E,$27,$F8,$77,$1F,$BD,$18,$F7,$3F,$10,$FF,$0A,$A7
	.dc.b	$0F,$F9,$6E,$BE,$D4,$E6,$31,$3F,$C4,$B5,$FE,$8F,$97,$FA,$39,$B7
	.dc.b	$FE,$7B,$DF,$EA,$7C,$6F,$E6,$BD,$59,$3F,$35,$FC,$EC,$63,$0F,$EA
	.dc.b	$1F,$47,$FE,$B8,$AF,$9B,$F4,$EF,$03,$FB,$74,$E6,$BE,$2C,$E6,$BF
	.dc.b	$11,$09,$D6,$7F,$E0,$3A,$D4,$F6,$E5,$FD,$34,$62,$7F,$B7,$D7,$5A
	.dc.b	$FF,$2A,$B5,$FF,$48,$B5,$FF,$4C,$37,$37,$E3,$9F,$C6,$71,$FC,$AB
	.dc.b	$6B,$F2,$2F,$F3,$07,$F9,$C3,$9B,$FF,$5B,$B6,$7E,$53,$67,$7F,$03
	.dc.b	$FE,$16,$8C,$47,$BF,$E3,$7F,$79,$3F,$19,$B5,$F1,$31,$47,$DF,$6C
	.dc.b	$7F,$99,$EB,$93,$63,$31,$FE,$6B,$FA,$EF,$EC,$4F,$E3,$FD,$EF,$F1
	.dc.b	$04,$63,$2F,$CA,$46,$3B,$FF,$44,$DC,$3F,$44,$DB,$E9,$3F,$91,$BE
	.dc.b	$B1,$FE,$84,$6B,$FE,$C0,$F3,$7F,$9C,$DF,$EA,$FD,$6F,$F1,$84,$62
	.dc.b	$7F,$45,$18,$F2,$BF,$39,$39,$9F,$F4,$7B,$5E,$91,$FA,$1F,$15,$F1
	.dc.b	$FF,$E4,$2F,$C3,$5F,$41,$FE,$3D,$9C,$E5,$FC,$42,$72,$FE,$42,$33
	.dc.b	$87,$FA,$45,$AF,$FA,$79,$AF,$FB,$69,$AF,$F2,$1F,$13,$FC,$9F,$F3
	.dc.b	$3F,$D0,$1B,$5F,$C4,$DF,$EF,$FC,$BF,$C7,$DB,$D7,$FC,$D6,$FF,$B8
	.dc.b	$B5,$FD,$93,$F8,$99,$FF,$3B,$FC,$07,$F4,$CF,$53,$FB,$9B,$70,$FF
	.dc.b	$2E,$B6,$FF,$93,$FA,$5F,$E3,$D9,$FD,$E6,$FF,$A7,$7F,$7D,$F9,$6F
	.dc.b	$F7,$BA,$FE,$5B,$39,$7F,$C2,$91,$9E,$65,$72,$ED,$F5,$DF,$ED,$2F
	.dc.b	$FC,$12,$D7,$53,$FE,$23,$D1,$F2,$9F,$E7,$4F,$3B,$FB,$2B,$67,$F9
	.dc.b	$C3,$F1,$08,$F1,$7E,$3F,$27,$08,$FE,$7C,$DB,$FE,$8B,$7F,$6B,$2F
	.dc.b	$E3,$97,$FD,$CA,$B5,$1F,$E3,$0F,$87,$5B,$E9,$1F,$2B,$88,$FF,$08
	.dc.b	$FD,$2F,$C0,$75,$1F,$ED,$5A,$E3,$5C,$27,$FC,$03,$5B,$8C,$47,$39
	.dc.b	$EF,$EC,$63,$05,$F5,$8D,$D7,$92,$B7,$87,$8E,$F7,$A6,$2B,$FD,$E3
	.dc.b	$5E,$2A,$D6,$1E,$17,$CD,$3E,$83,$CB,$5B,$78,$99,$3D,$EE,$5F,$41
	.dc.b	$9B,$F0,$BB,$3C,$C7,$F9,$1E,$7C,$33,$E5,$8E,$38,$7B,$EF,$DD,$5E
	.dc.b	$4C,$71,$7D,$EF,$9C,$61,$EC,$31,$F2,$63,$E2,$9F,$8F,$39,$3C,$D8
	.dc.b	$7C,$17,$DB,$1C,$CD,$D7,$29,$D4,$FB,$A7,$58,$79,$4F,$5C,$78,$AB
	.dc.b	$86,$1E,$78,$F0,$57,$07,$BD,$70,$8E,$F1,$E4,$9E,$15,$CA,$3A,$19
	.dc.b	$2C,$F3,$7C,$95,$CC,$F2,$3F,$68,$DB,$97,$ED,$FA,$9F,$E7,$96,$E5
	.dc.b	$F8,$28,$C4,$7D,$C6,$FC,$73,$E1,$FF,$CD,$F3,$FE,$E1,$AF,$D1,$4E
	.dc.b	$74,$FD,$3D,$F0,$BD,$B6,$33,$5C,$27,$9E,$3F,$A2,$C5,$D5,$FA,$D8
	.dc.b	$C3,$63,$85,$FE,$47,$1F,$06,$5F,$95,$9E,$2F,$E5,$5A,$FE,$66,$37
	.dc.b	$7F,$B0,$6C,$FE,$EC,$DC,$78,$27,$B6,$1E,$69,$F0,$C7,$7C,$3C,$8F
	.dc.b	$F0,$68,$E2,$6A,$BB,$57,$19,$D5,$70,$7F,$85,$B6,$FC,$71,$FC,$41
	.dc.b	$AE,$8F,$64,$6E,$7A,$42,$77,$7E,$28,$EE,$7A,$B0,$8F,$A4,$E6,$63
	.dc.b	$0F,$3C,$62,$34,$7C,$9B,$EF,$7F,$43,$1F,$8D,$3F,$BD,$3B,$94,$F2
	.dc.b	$4F,$E2,$23,$AC,$74,$BE,$26,$09,$87,$D3,$C7,$E3,$3E,$B9,$E7,$8F
	.dc.b	$AA,$FC,$24,$9F,$AA,$F2,$9F,$1D,$F6,$9F,$A8,$7F,$68,$F0,$FF,$94
	.dc.b	$6B,$FD,$AB,$E0,$2D,$FF,$A5,$DB,$D5,$FE,$91,$AB,$C7,$FE,$2E,$AF
	.dc.b	$F4,$55,$60,$CC,$D8,$FE,$B9,$1F,$E3,$1A,$FF,$4D,$36,$8D,$3B,$FB
	.dc.b	$CF,$8F,$A7,$3B,$FE,$E3,$E0,$F3,$E8,$34,$91,$FD,$54,$FE,$F6,$FA
	.dc.b	$BA,$9E,$75,$3D,$B5,$CA,$BF,$D4,$9C,$5F,$03,$B1,$E9,$D3,$E0,$C4
	.dc.b	$9F,$F5,$AC,$E9,$B3,$97,$FB,$79,$EF,$21,$FD,$FC,$A7,$FA,$E3,$3F
	.dc.b	$E1,$CF,$FB,$01,$AF,$23,$FD,$03,$F9,$77,$F1,$67,$DC,$D7,$FB,$A3
	.dc.b	$5F,$EF,$2D,$4F,$F1,$EE,$5F,$EF,$10,$FC,$BD,$F2,$BF,$82,$8C,$36
	.dc.b	$B8,$72,$7F,$80,$7E,$F1,$EF,$FF,$9B,$6D,$18,$7F,$CE,$0D,$7E,$A1
	.dc.b	$A8,$E1,$19,$C7,$C7,$E8,$3D,$3F,$13,$6F,$FD,$43,$5B,$C3,$FC,$25
	.dc.b	$EA,$BE,$57,$C4,$DA,$FF,$40,$35,$F0,$B5,$FC,$33,$9C,$7E,$1F,$C1
	.dc.b	$FC,$29,$FF,$57,$B5,$FF,$78,$09,$E1,$7E,$2B,$F9,$15,$BB,$E7,$AF
	.dc.b	$CB,$BD,$E1,$FE,$2C,$5F,$23,$E8,$24,$7A,$1A,$F8,$3E,$87,$EB,$DF
	.dc.b	$D8,$6E,$5F,$0A,$BF,$6F,$F9,$87,$F8,$A2,$FF,$C7,$37,$FE,$35,$ED
	.dc.b	$4E,$65,$39,$8C,$37,$E3,$79,$F8,$F2,$B9,$E3,$FE,$E1,$BD,$B6,$BF
	.dc.b	$11,$F4,$4F,$8A,$FC,$23,$6F,$FD,$64,$D7,$F0,$6D,$7F,$70,$E7,$1F
	.dc.b	$CA,$FD,$17,$C6,$9F,$F7,$0B,$5F,$A7,$3B,$C3,$FC,$A8,$D8,$FF,$2C
	.dc.b	$CE,$EB,$9E,$9E,$6C,$3F,$8E,$FD,$EF,$D7,$C7,$FD,$F8,$DC,$1A,$A3
	.dc.b	$0D,$F9,$BE,$3C,$32,$BE,$29,$C2,$B6,$27,$E2,$F0,$FC,$42,$4F,$C4
	.dc.b	$6B,$F8,$ED,$1F,$23,$FB,$36,$C7,$61,$B7,$E4,$DF,$8B,$F3,$7B,$23
	.dc.b	$F2,$32,$AF,$58,$BB,$FC,$E9,$BA,$FF,$9F,$70,$56,$2F,$E5,$D7,$C8
	.dc.b	$E5,$91,$F0,$36,$F8,$D1,$F3,$2B,$33,$F8,$EC,$3E,$6C,$E3,$1F,$CA
	.dc.b	$6D,$DA,$B9,$3E,$E7,$81,$F2,$F4,$FA,$E7,$D8,$F9,$5B,$1C,$22,$BC
	.dc.b	$32,$FB,$13,$FC,$6F,$E8,$9F,$15,$FC,$5D,$B5,$FC,$50,$38,$7F,$8B
	.dc.b	$B5,$FE,$98,$E7,$1F,$E5,$3E,$9F,$FA,$5F,$4F,$F5,$23,$6B,$FD,$53
	.dc.b	$5B,$9F,$F5,$43,$63,$FD,$5B,$3B,$7F,$D8,$3E,$67,$9F,$0F,$CA,$FC
	.dc.b	$77,$F6,$4E,$79,$5F,$1A,$3F,$9B,$7D,$E7,$EC,$A3,$F2,$FE,$F7,$F3
	.dc.b	$66,$FA,$3A,$EC,$FC,$FF,$B9,$FD,$77,$E2,$A5,$7D,$08,$FD,$1F,$9F
	.dc.b	$FD,$CA,$A4,$FD,$48,$EF,$F2,$FF,$DD,$89,$1F,$93,$C9,$1F,$AC,$AF
	.dc.b	$ED,$5F,$4B,$2B,$EE,$C7,$EC,$3E,$5E,$47,$E0,$1F,$D9,$D7,$F5,$1F
	.dc.b	$07,$FB,$64,$FE,$0B,$E5,$9F,$80,$E3,$9F,$FD,$A1,$18,$F3,$B8,$1F
	.dc.b	$BC,$F5,$25,$FE,$31,$FD,$FE,$1F,$E1,$CF,$73,$F8,$84,$FF,$B7,$BB
	.dc.b	$3F,$D5,$5E,$37,$F8,$5E,$73,$3F,$C0,$9A,$FF,$38,$E3,$FF,$13,$E5
	.dc.b	$E4,$F9,$CB,$A9,$F5,$4F,$F8,$AB,$F3,$1F,$EB,$57,$FE,$15,$C9,$18
	.dc.b	$8F,$F7,$64,$63,$CE,$FF,$14,$4E,$67,$FD,$E7,$8B,$E7,$FF,$C3,$71
	.dc.b	$FC,$0F,$EF,$1F,$31,$FD,$E5,$B1,$FC,$52,$B8,$7B,$39,$AF,$4F,$82
	.dc.b	$D2,$99,$BF,$DE,$7C,$43,$E1,$BF,$C6,$BD,$12,$73,$1D,$FE,$42,$47
	.dc.b	$8D,$F8,$BA,$BE,$9F,$33,$F6,$F2,$F8,$CF,$95,$AF,$F6,$ED,$97,$E6
	.dc.b	$E2,$FF,$22,$F3,$4A,$E1,$1E,$FB,$91,$F0,$DF,$AB,$CA,$F8,$8C,$CF
	.dc.b	$E2,$FC,$F7,$C3,$FD,$A6,$C7,$12,$1B,$8D,$1C,$BB,$3F,$C5,$3A,$3E
	.dc.b	$9B,$FC,$44,$DA,$FE,$18,$D7,$FC,$42,$D7,$FC,$AC,$E6,$7F,$37,$7F
	.dc.b	$E6,$49,$FB,$0D,$7E,$25,$7F,$9D,$3C,$2F,$A9,$1F,$E5,$2F,$8C,$FF
	.dc.b	$2C,$19,$E2,$91,$92,$1F,$C9,$67,$8B,$79,$4C,$D7,$FB,$87,$5F,$F9
	.dc.b	$CE,$73,$29,$CD,$7E,$26,$38,$70,$FE,$AE,$3F,$73,$3A,$AE,$2D,$AD
	.dc.b	$9F,$DB,$A7,$93,$D6,$B9,$77,$F0,$7C,$3F,$47,$FE,$7A,$F8,$AE,$34
	.dc.b	$8C,$3C,$4F,$5B,$8C,$3F,$C1,$0F,$66,$F5,$AB,$9C,$EE,$FB,$CE,$6F
	.dc.b	$A1,$DF,$2F,$23,$DA,$35,$87,$4C,$39,$C6,$2F,$D3,$7C,$2F,$D1,$39
	.dc.b	$79,$57,$03,$C7,$3B,$C3,$C9,$3A,$3C,$6E,$1F,$21,$C4,$FF,$37,$BF
	.dc.b	$11,$FF,$52,$35,$F6,$CF,$F6,$1F,$8A,$FF,$58,$76,$DD,$0D,$25,$74
	.dc.b	$F5,$FF,$89,$37,$AF,$FB,$56,$3F,$E6,$56,$CE,$A7,$C2,$F0,$D7,$E9
	.dc.b	$B7,$5F,$C6,$FC,$E3,$E4,$DE,$97,$D4,$AD,$61,$F7,$1A,$FE,$05,$B5
	.dc.b	$C6,$78,$63,$2B,$E3,$D6,$AF,$AC,$62,$75,$8F,$91,$FC,$6E,$1C,$E7
	.dc.b	$44,$9D,$3F,$84,$8D,$46,$0F,$D4,$FC,$23,$E8,$FB,$C6,$1B,$9F,$F5
	.dc.b	$19,$FE,$91,$9F,$E9,$39,$11,$C6,$30,$DC,$7F,$3D,$3A,$BE,$71,$FE
	.dc.b	$58,$9E,$6E,$9E,$18,$79,$3E,$4F,$E0,$F5,$7F,$A1,$7E,$1B,$FD,$D7
	.dc.b	$97,$FB,$72,$72,$FF,$65,$8C,$E1,$FE,$2D,$C1,$5A,$8F,$03,$72,$F4
	.dc.b	$FD,$45,$AD,$E1,$FE,$59,$FA,$4F,$D8,$3F,$E3,$16,$E1,$FE,$C6,$6B
	.dc.b	$FE,$FB,$6B,$FC,$49,$DE,$5F,$ED,$A6,$CF,$F6,$68,$FE,$92,$F7,$E6
	.dc.b	$F3,$53,$E8,$FF,$AE,$4F,$F6,$5F,$A4,$FF,$68,$3F,$E6,$D8,$CC,$FF
	.dc.b	$84,$67,$2F,$FB,$0A,$32,$FF,$24,$3A,$76,$7F,$BB,$1F,$E4,$F1,$A6
	.dc.b	$C6,$5F,$E5,$F1,$DB,$0F,$F7,$9B,$6F,$D5,$F3,$DE,$88,$C5,$7E,$D3
	.dc.b	$E4,$FF,$C5,$F5,$FD,$F5,$B1,$E5,$F3,$FE,$A7,$52,$73,$8F,$69,$F1
	.dc.b	$57,$C3,$6E,$1F,$EB,$9C,$3F,$CC,$8D,$9F,$76,$1F,$C1,$3D,$9F,$25
	.dc.b	$F6,$84,$62,$FF,$D0,$9E,$C3,$E6,$3F,$08,$D8,$F4,$7A,$78,$9F,$77
	.dc.b	$37,$88,$93,$E2,$7E,$46,$1F,$C5,$3C,$EF,$01,$F8,$DF,$A1,$B5,$FE
	.dc.b	$93,$FA,$47,$C9,$FE,$4D,$B1,$9F,$95,$C4,$FE,$13,$E8,$FF,$7D,$C1
	.dc.b	$F6,$DA,$FF,$60,$6B,$FD,$99,$BD,$6F,$7B,$F8,$BE,$1F,$EA,$36,$D7
	.dc.b	$FA,$4D,$FF,$11,$46,$4F,$D2,$6B,$1F,$F5,$57,$5F,$CF,$E5,$FF,$7C
	.dc.b	$29,$F3,$7C,$CF,$D8,$C8,$D2,$5F,$E4,$F1,$FE,$05,$F4,$9F,$B1,$3F
	.dc.b	$65,$F1,$F6,$7B,$FC,$B7,$ED,$F1,$F5,$46,$0F,$F5,$E3,$8B,$FF,$59
	.dc.b	$8C,$5E,$2F,$8C,$95,$FE,$AB,$8C,$BE,$3D,$BF,$8D,$FC,$9F,$F0,$33
	.dc.b	$F8,$27,$93,$67,$E0,$9E,$3F,$07,$F8,$43,$4F,$F7,$05,$BE,$5B,$E1
	.dc.b	$12,$3C,$0F,$EB,$70,$7C,$C7,$F9,$92,$7F,$DA,$CD,$CF,$FD,$8C,$BE
	.dc.b	$A3,$F8,$C3,$9E,$3F,$E8,$B2,$FA,$17,$4F,$F1,$F9,$7D,$2B,$E7,$BF
	.dc.b	$86,$D6,$B2,$FF,$15,$D7,$F2,$EA,$D4,$7F,$BE,$67,$58,$7D,$76,$BF
	.dc.b	$C2,$DA,$FF,$8C,$1A,$FF,$8D,$9E,$35,$E6,$6C,$F8,$3E,$C5,$F2,$5E
	.dc.b	$4C,$BD,$BF,$1B,$2B,$E1,$FC,$0D,$CB,$E1,$CD,$CB,$FB,$2E,$FF,$2A
	.dc.b	$B1,$7F,$EA,$7D,$9E,$4F,$C8,$3E,$4F,$D3,$59,$F9,$F1,$72,$F9,$F3
	.dc.b	$71,$FE,$A3,$F4,$DF,$E9,$1B,$F8,$E7,$C8,$71,$49,$DF,$B7,$F4,$67
	.dc.b	$31,$CB,$DA,$C1,$99,$B7,$D2,$F9,$5F,$08,$ED,$5B,$75,$69,$D5,$7C
	.dc.b	$FF,$6F,$FA,$E7,$D3,$3A,$BE,$FC,$61,$B8,$7D,$D8,$DF,$5E,$2F,$F1
	.dc.b	$FF,$13,$F8,$E3,$F1,$FF,$4D,$6F,$F2,$B1,$67,$F2,$B3,$73,$FC,$9F
	.dc.b	$D2,$FF,$5C,$9F,$89,$7F,$8E,$F9,$FF,$F4,$F3,$EF,$84,$EE,$BF,$BF
	.dc.b	$D6,$A7,$F7,$73,$CD,$FE,$B4,$D7,$FC,$FE,$D7,$FD,$0A,$D7,$AF,$1F
	.dc.b	$F6,$C8,$ED,$7F,$67,$E4,$31,$3F,$E0,$3F,$A6,$FE,$2F,$3F,$4F,$99
	.dc.b	$18,$C3,$D6,$23,$3E,$FF,$1F,$22,$5F,$1B,$E1,$A7,$E5,$9F,$F1,$B3
	.dc.b	$67,$C9,$E6,$7E,$DF,$BF,$37,$EC,$8C,$61,$A8,$F9,$1A,$A7,$F7,$DD
	.dc.b	$DF,$D5,$F4,$F8,$F1,$FE,$0A,$FA,$6C,$67,$F9,$AF,$73,$1E,$97,$E5
	.dc.b	$5B,$97,$F3,$3C,$17,$CD,$FF,$04,$35,$FF,$56,$B5,$FF,$59,$BD,$A3
	.dc.b	$35,$FA,$1F,$03,$AD,$7F,$34,$F4,$DE,$5F,$E9,$7D,$4F,$F0,$FC,$63
	.dc.b	$0F,$CE,$F6,$A7,$35,$27,$2D,$BF,$E7,$FE,$47,$FB,$9A,$7E,$C7,$C6
	.dc.b	$BE,$17,$FC,$57,$39,$AF,$B3,$CE,$AD,$5F,$E4,$EB,$5C,$1F,$CB,$1B
	.dc.b	$1F,$ED,$0A,$FC,$1F,$0F,$4F,$89,$FE,$67,$E5,$7F,$3B,$F1,$FF,$94
	.dc.b	$E7,$FE,$2F,$BF,$E7,$B3,$92,$61,$FC,$F9,$FE,$77,$F5,$3C,$A3,$FA
	.dc.b	$0F,$57,$F9,$4F,$DD,$FE,$7E,$D8,$C9,$30,$FE,$8B,$D5,$A8,$FC,$C7
	.dc.b	$0F,$EA,$67,$FD,$C7,$97,$D0,$F8,$BF,$EA,$13,$99,$F6,$35,$FC,$E3
	.dc.b	$FB,$1E,$0F,$E5,$AD,$BD,$FC,$D7,$F9,$A6,$31,$5F,$A3,$D9,$F1,$BF
	.dc.b	$0C,$97,$FD,$B6,$32,$D8,$FE,$B6,$7F,$5C,$E0,$B1,$FD,$89,$B5,$FD
	.dc.b	$89,$B8,$FF,$40,$BF,$AB,$3E,$19,$FE,$CB,$F1,$7C,$EF,$F8,$74,$87
	.dc.b	$C2,$DF,$69,$F3,$19,$9F,$E7,$5F,$2B,$FE,$B0,$9C,$C7,$F9,$A6,$B8
	.dc.b	$56,$F3,$5E,$E3,$FC,$39,$3D,$FE,$9B,$8D,$F4,$C7,$15,$A8,$4E,$B0
	.dc.b	$E7,$5A,$9F,$01,$AA,$F0,$61,$FE,$C4,$BE,$B3,$ED,$C3,$83,$DC,$E0
	.dc.b	$72,$C7,$B5,$79,$E3,$17,$D3,$0F,$6C,$78,$30,$DC,$74,$8E,$C7,$43
	.dc.b	$B4,$78,$9F,$9A,$67,$1F,$5B,$B3,$8C,$74,$70,$70,$C3,$07,$3C,$7C
	.dc.b	$71,$EB,$8C,$3D,$71,$F8,$58,$FC,$BB,$E8,$68,$AE,$4F,$74,$F8,$2F
	.dc.b	$E3,$E1,$E6,$AE,$15,$A7,$BE,$1E,$48,$FB,$0D,$BF,$8D,$1B,$3B,$3F
	.dc.b	$E2,$CC,$3C,$AF,$CE,$31,$8F,$B1,$F9,$75,$CD,$FC,$96,$3F,$03,$1F
	.dc.b	$93,$3E,$2C,$3D,$2F,$C2,$F6,$3F,$CF,$97,$F8,$66,$BC,$83,$87,$EC
	.dc.b	$BA,$F8,$EF,$F5,$85,$FE,$96,$31,$F0,$DF,$E1,$BA,$E3,$7F,$A5,$BD
	.dc.b	$C6,$0F,$D6,$38,$AE,$37,$F6,$30,$FF,$31,$55,$EB,$FD,$76,$1F,$A4
	.dc.b	$C3,$9B,$71,$F0,$3F,$A8,$BE,$37,$FE,$99,$7F,$6C,$62,$38,$61,$F8
	.dc.b	$6B,$F2,$EB,$C7,$1F,$DE,$37,$0F,$D2,$CE,$B0,$F6,$63,$E0,$7A,$DF
	.dc.b	$99,$B5,$D6,$74,$FC,$07,$E3,$5F,$58,$FD,$24,$FA,$23,$CB,$1F,$A7
	.dc.b	$BF,$03,$FC,$4E,$74,$76,$75,$5D,$23,$CB,$87,$EF,$1D,$CF,$1C,$3D
	.dc.b	$B1,$A8,$E3,$3F,$3D,$87,$C9,$3D,$0F,$AC,$E9,$3C,$A3,$A1,$F1,$AB
	.dc.b	$C4,$F0,$3D,$B1,$B7,$F9,$4B,$6F,$DB,$07,$EE,$DF,$B7,$1F,$8C,$7C
	.dc.b	$43,$E2,$E3,$F2,$49,$1E,$B7,$2F,$C5,$73,$5E,$37,$94,$75,$24,$79
	.dc.b	$E7,$F9,$CF,$2F,$FA,$BB,$AF,$F9,$C6,$DF,$E8,$56,$D7,$F5,$96,$CF
	.dc.b	$FA,$D2,$EF,$17,$F8,$BE,$BF,$D6,$BF,$01,$83,$37,$63,$24,$3F,$D7
	.dc.b	$2E,$5F,$F0,$24,$FF,$C0,$07,$FB,$24,$FF,$8C,$6F,$FD,$86,$DE,$57
	.dc.b	$83,$E4,$6D,$7F,$7E,$8C,$7C,$5E,$5E,$04,$8F,$F6,$E4,$F0,$7C,$3B
	.dc.b	$FB,$08,$FF,$B4,$3D,$DA,$FF,$89,$BE,$05,$EB,$FC,$46,$C7,$FB,$5A
	.dc.b	$BE,$2B,$0F,$F6,$C9,$F2,$3E,$33,$0F,$C0,$3D,$DA,$FF,$93,$DF,$F0
	.dc.b	$41,$7E,$93,$F8,$CF,$37,$FB,$71,$BC,$5F,$DD,$5B,$3F,$DE,$FE,$97
	.dc.b	$FD,$92,$FF,$A2,$E3,$2A,$1F,$41,$F7,$A3,$0D,$9D,$FD,$2B,$D8,$D3
	.dc.b	$6B,$FD,$2C,$D7,$FD,$02,$3D,$6F,$F7,$63,$A9,$ED,$0F,$62,$9D,$23
	.dc.b	$C5,$ED,$7A,$2B,$9B,$74,$FF,$7D,$B7,$4F,$F0,$A7,$C2,$DA,$FF,$0D
	.dc.b	$4F,$17,$FC,$30,$7F,$AF,$FE,$07,$E6,$46,$63,$FC,$38,$D7,$8C,$7C
	.dc.b	$39,$F8,$B1,$FE,$B3,$6B,$E4,$8F,$3F,$3C,$8F,$F5,$C5,$23,$E6,$7D
	.dc.b	$53,$E6,$35,$5F,$1B,$E9,$99,$7D,$16,$C7,$2B,$FD,$7F,$D3,$7F,$8A
	.dc.b	$A3,$FD,$60,$D7,$FC,$EE,$DD,$3F,$DF,$EB,$E3,$B8,$F8,$F2,$B8,$BD
	.dc.b	$23,$F0,$1F,$4A,$F9,$0A,$4A,$FF,$69,$7C,$C7,$E2,$27,$51,$FE,$A3
	.dc.b	$6B,$F3,$83,$87,$E8,$B5,$FF,$25,$3C,$31,$F7,$C2,$7E,$43,$E0,$CB
	.dc.b	$C4,$FF,$A5,$3E,$93,$E8,$D7,$C6,$F0,$D1,$8D,$BF,$D2,$4D,$AF,$F6
	.dc.b	$92,$FA,$6F,$E3,$2E,$6B,$8F,$56,$0C,$CF,$C4,$E3,$97,$FE,$F3,$D1
	.dc.b	$1F,$A7,$D1,$F4,$7A,$AF,$D0,$C3,$D1,$D1,$9D,$7D,$F0,$FA,$6E,$44
	.dc.b	$AF,$9F,$9D,$79,$C5,$AE,$33,$76,$FF,$3A,$5F,$3B,$FA,$F7,$E5,$AF
	.dc.b	$7F,$3C,$9F,$F7,$C6,$FF,$A7,$DF,$E4,$DE,$67,$E7,$55,$CB,$E7,$5F
	.dc.b	$59,$FA,$59,$FC,$A2,$3D,$3E,$9A,$BA,$3B,$C7,$D0,$8F,$E1,$DF,$4C
	.dc.b	$F8,$FA,$4A,$FF,$49,$61,$F8,$4F,$57,$FA,$14,$7F,$0C,$AE,$0F,$FB
	.dc.b	$CD,$AF,$E4,$5A,$FF,$00,$6B,$FE,$99,$77,$3D,$7C,$DF,$E4,$15,$FE
	.dc.b	$8D,$E0,$D6,$1F,$86,$E1,$98,$FC,$87,$F1,$3E,$B7,$F2,$7F,$A5,$FD
	.dc.b	$94,$FF,$0E,$6B,$FE,$A7,$6B,$FE,$AD,$6F,$0F,$E6,$3E,$9B,$FA,$47
	.dc.b	$C7,$CA,$FA,$B1,$F9,$6F,$A0,$91,$F6,$1F,$CC,$FD,$47,$87,$E9,$FF
	.dc.b	$A3,$62,$69,$2B,$F6,$98,$FE,$9B,$BF,$FA,$B7,$C7,$C9,$FC,$1C,$7E
	.dc.b	$93,$E6,$E4,$7E,$19,$FD,$3D,$FF,$45,$EF,$FE,$BB,$3F,$BF,$6B,$FC
	.dc.b	$B9,$AF,$FB,$1D,$AF,$EC,$BD,$4F,$D8,$91,$FC,$9C,$7F,$83,$3D,$CF
	.dc.b	$E0,$9C,$7F,$ED,$3C,$FF,$DA,$7E,$16,$4D,$7C,$47,$8F,$8B,$FB,$36
	.dc.b	$06,$93,$1F,$83,$F4,$DF,$D3,$DF,$C6,$E3,$D6,$31,$0E,$9F,$E1,$91
	.dc.b	$86,$FC,$97,$F8,$8B,$E7,$5E,$A3,$E5,$30,$DC,$3A,$89,$DC,$FF,$10
	.dc.b	$F7,$7F,$B7,$7D,$27,$F4,$3A,$FE,$50,$1C,$3F,$CA,$DA,$FF,$5A,$6B
	.dc.b	$F1,$7E,$29,$F1,$1F,$E3,$46,$C7,$8B,$89,$18,$8F,$76,$DF,$EF,$FC
	.dc.b	$9E,$07,$C9,$9F,$4C,$7B,$BE,$9B,$CF,$F2,$D2,$73,$B2,$71,$1E,$79
	.dc.b	$FF,$0C,$68,$69,$EE,$BE,$FF,$4D,$F2,$F1,$34,$E7,$C3,$FD,$12,$FC
	.dc.b	$B8,$7C,$C3,$E3,$57,$CA,$3E,$A7,$3B,$C1,$C4,$FC,$CA,$E2,$FC,$BA
	.dc.b	$F2,$91,$F7,$F9,$8F,$89,$FB,$CD,$8F,$05,$72,$C7,$C9,$F1,$3F,$68
	.dc.b	$93,$F2,$32,$78,$98,$8F,$B5,$3F,$10,$8F,$ED,$3C,$4F,$A9,$7F,$D6
	.dc.b	$DB,$C5,$FA,$16,$CF,$F5,$AF,$89,$FF,$38,$C6,$39,$5F,$02,$F8,$EF
	.dc.b	$F2,$91,$AE,$8D,$F6,$FF,$BA,$3B,$BA,$37,$26,$0C,$7F,$05,$0D,$FF
	.dc.b	$B9,$CE,$7E,$0F,$83,$BB,$B7,$C2,$B6,$DC,$3D,$57,$DF,$07,$D3,$7E
	.dc.b	$A7,$54,$77,$AA,$EB,$BB,$FD,$EA,$E4,$DD,$71,$70,$74,$CB,$B4,$67
	.dc.b	$2F,$F0,$5C,$F3,$C3,$BC,$62,$75,$87,$81,$F4,$57,$6C,$3C,$35,$B6
	.dc.b	$D6,$DB,$3E,$4C,$3C,$31,$88,$F0,$1E,$6B,$EE,$EC,$F1,$D7,$3C,$2C
	.dc.b	$78,$5A,$F2,$C3,$D1,$87,$69,$FE,$13,$7F,$0E,$BF,$CB,$ED,$7A,$1F
	.dc.b	$E3,$5F,$8A,$FB,$DD,$5F,$C4,$1B,$19,$D3,$F8,$C3,$67,$AF,$AB,$FC
	.dc.b	$82,$3F,$C7,$D7,$FE,$22,$EE,$C9,$23,$33,$BC,$FE,$73,$74,$FB,$BF
	.dc.b	$0C,$F8,$CF,$F2,$FD,$6B,$0F,$7E,$3D,$71,$F9,$18,$F1,$7F,$0F,$87
	.dc.b	$1E,$5D,$CE,$67,$F1,$AD,$AE,$58,$7E,$0A,$F8,$57,$4C,$38,$91,$FC
	.dc.b	$5C,$6A,$7C,$91,$B2,$3C,$F1,$F4,$E5,$F8,$9C,$7C,$DC,$5F,$B4,$FA
	.dc.b	$6C,$E1,$FA,$EE,$64,$62,$7F,$73,$DF,$67,$3B,$7E,$0A,$79,$36,$33
	.dc.b	$7F,$A3,$9E,$33,$9B,$E5,$F2,$BF,$E8,$AE,$9F,$F3,$36,$08,$C5,$7F
	.dc.b	$6D,$87,$D3,$63,$1F,$35,$5C,$7F,$86,$B6,$B9,$5F,$F4,$D7,$8B,$63
	.dc.b	$95,$3E,$99,$CD,$86,$D7,$F4,$EC,$BF,$A3,$3F,$CF,$63,$35,$FE,$CA
	.dc.b	$6C,$FF,$A1,$9B,$3F,$D6,$63,$8C,$EE,$BC,$3F,$4B,$FD,$92,$BF,$B6
	.dc.b	$B6,$7F,$D2,$CD,$7F,$D8,$AD,$7F,$88,$BB,$AF,$F4,$AB,$67,$C1,$F3
	.dc.b	$FD,$BB,$24,$7F,$55,$9D,$5F,$F8,$06,$D7,$A8,$FE,$E2,$7F,$81,$7E
	.dc.b	$4B,$FC,$71,$7F,$DC,$1B,$1F,$EE,$AB,$FF,$5A,$4E,$7C,$FF,$6F,$9D
	.dc.b	$23,$FB,$BF,$57,$FA,$FD,$F1,$47,$DB,$8F,$F0,$97,$C0,$EF,$0F,$0C
	.dc.b	$62,$B6,$7F,$81,$FE,$59,$F3,$1F,$81,$6C,$79,$7D,$2C,$4F,$CE,$FA
	.dc.b	$39,$3E,$2F,$52,$E3,$DB,$14,$78,$DB,$3F,$89,$F5,$31,$3F,$8F,$FA
	.dc.b	$57,$CA,$53,$1F,$2B,$6F,$FC,$05,$F4,$9F,$BA,$E5,$FF,$64,$B6,$3F
	.dc.b	$A4,$FE,$01,$FE,$F1,$77,$E1,$7E,$6F,$D7,$FF,$77,$C7,$F8,$95,$BB
	.dc.b	$7D,$4E,$E4,$7F,$87,$67,$47,$E8,$78,$23,$FA,$1E,$5B,$7F,$70,$64
	.dc.b	$F9,$9F,$D1,$B6,$BF,$C3,$59,$FE,$6A,$BF,$B0,$CE,$5F,$17,$7B,$E9
	.dc.b	$BF,$60,$7E,$C7,$E4,$2E,$1F,$B1,$F9,$9F,$DA,$1F,$B5,$E9,$B1,$F1
	.dc.b	$BC,$9F,$09,$FB,$9F,$3E,$C7,$EE,$7E,$4B,$F7,$87,$EF,$7C,$9B,$1F
	.dc.b	$BD,$3F,$80,$1F,$C0,$7D,$BA,$FD,$2F,$99,$F6,$9F,$C1,$FE,$89,$F2
	.dc.b	$78,$9F,$F2,$97,$CD,$61,$B1,$FC,$36,$1F,$4D,$63,$F8,$4F,$D3,$3C
	.dc.b	$15,$FF,$99,$A3,$0D,$8C,$F2,$7F,$82,$0F,$F0,$5C,$E9,$C0,$71,$DB
	.dc.b	$C2,$31,$3F,$C8,$61,$F4,$4F,$91,$F1,$CC,$C3,$FC,$A5,$87,$27,$FD
	.dc.b	$AF,$5B,$73,$3E,$DF,$8C,$91,$F0,$BE,$12,$E5,$F8,$FD,$7F,$DC,$53
	.dc.b	$F1,$AB,$E2,$56,$B3,$F6,$F8,$0B,$FE,$17,$7F,$33,$CD,$F9,$5F,$4D
	.dc.b	$F2,$D9,$19,$3E,$8C,$63,$2F,$E2,$38,$3E,$8B,$FD,$51,$5F,$E9,$88
	.dc.b	$C7,$BB,$F0,$E5,$EF,$58,$95,$B9,$E3,$7D,$1D,$C6,$67,$73,$97,$A5
	.dc.b	$62,$FE,$97,$07,$FA,$AE,$FE,$6F,$4F,$F5,$1F,$A6,$75,$7E,$12,$24
	.dc.b	$E8,$FF,$4F,$4E,$9C,$65,$F3,$7E,$73,$F1,$C7,$E3,$FE,$2A,$CF,$E5
	.dc.b	$22,$CF,$D3,$9B,$97,$E1,$7E,$97,$FA,$6C,$FE,$35,$AF,$F9,$59,$AF
	.dc.b	$F9,$69,$A9,$FE,$AB,$1C,$F8,$FE,$67,$C7,$7F,$66,$9C,$E3,$CE,$70
	.dc.b	$69,$FF,$50,$46,$AF,$FC,$EF,$D7,$FC,$1F,$0F,$C2,$B6,$BF,$DD,$0F
	.dc.b	$F2,$39,$DF,$8F,$FB,$52,$4F,$FB,$16,$75,$CC,$79,$57,$AF,$CC,$F3
	.dc.b	$1F,$A5,$EA,$FE,$87,$1F,$C7,$F2,$4E,$B0,$F3,$BB,$F4,$71,$5F,$C9
	.dc.b	$7E,$97,$8B,$FB,$76,$BF,$E8,$06,$BF,$E8,$46,$A7,$C3,$A7,$EE,$5B
	.dc.b	$7F,$E8,$A8,$FC,$E7,$67,$F8,$4A,$73,$5C,$EF,$07,$E7,$66,$DF,$FB
	.dc.b	$37,$D3,$78,$DF,$E1,$A6,$C7,$E7,$E3,$04,$78,$3F,$A7,$31,$7F,$D0
	.dc.b	$FC,$2F,$F6,$64,$FF,$B6,$5B,$1F,$D3,$F6,$F9,$33,$9F,$8A,$FF,$14
	.dc.b	$D6,$B4,$FE,$2C,$D9,$FB,$BA,$EA,$BF,$95,$7D,$2C,$9F,$E5,$FF,$4B
	.dc.b	$72,$FE,$1D,$37,$1F,$E1,$7E,$0A,$B1,$7F,$EB,$2E,$3F,$E1,$DC,$9F
	.dc.b	$CD,$4F,$E6,$DE,$95,$9F,$E7,$71,$72,$FE,$35,$37,$1F,$E2,$FB,$FF
	.dc.b	$BC,$EF,$EA,$1D,$5C,$57,$F6,$F5,$F3,$1F,$E4,$F7,$35,$CB,$E3,$F0
	.dc.b	$66,$6D,$FF,$4C,$F8,$FF,$E9,$47,$F4,$BD,$EC,$64,$98,$FF,$2F,$8C
	.dc.b	$B6,$3F,$AA,$C7,$F5,$3F,$8C,$E5,$18,$24,$FF,$37,$8C,$BF,$DF,$CF
	.dc.b	$F6,$58,$FA,$57,$CB,$FF,$B6,$B3,$7C,$A7,$F5,$69,$CD,$7F,$57,$F8
	.dc.b	$AF,$F7,$0C,$62,$15,$C3,$1C,$E1,$FD,$6A,$ED,$7E,$7F,$0E,$93,$98
	.dc.b	$F1,$3E,$18,$CC,$77,$3F,$C3,$B1,$9A,$FF,$78,$FB,$5E,$19,$F5,$BE
	.dc.b	$3C,$3C,$51,$C0,$E5,$8E,$70,$EB,$1F,$EF,$8C,$7C,$AF,$AA,$32,$FC
	.dc.b	$69,$E1,$86,$A3,$18,$F8,$DF,$49,$E7,$C7,$6F,$84,$EC,$5C,$7C,$B1
	.dc.b	$E9,$86,$3E,$29,$F4,$E1,$98,$C6,$3E,$4A,$D9,$E0,$3D,$18,$F9,$5C
	.dc.b	$D7,$88,$E4,$EE,$FA,$E3,$F4,$67,$4D,$3D,$8F,$03,$33,$F2,$A3,$26
	.dc.b	$DF,$B9,$5E,$1C,$3C,$ED,$8F,$06,$1E,$99,$CC,$7A,$4F,$96,$72,$30
	.dc.b	$7A,$AB,$BB,$E1,$35,$7E,$CF,$84,$FF,$30,$C7,$E1,$DA,$FD,$06,$BF
	.dc.b	$A0,$73,$F1,$CF,$84,$FC,$CC,$FF,$B0,$34,$32,$75,$A6,$2F,$8D,$F0
	.dc.b	$6C,$66,$F8,$DA,$F8,$BF,$5F,$17,$56,$B2,$FD,$54,$62,$7F,$27,$5C
	.dc.b	$27,$F5,$CD,$9E,$A6,$E7,$A5,$73,$8F,$0C,$EE,$7D,$59,$7B,$AF,$BD
	.dc.b	$EE,$7D,$F7,$CE,$7B,$D6,$E3,$9B,$67,$75,$C5,$FE,$12,$DB,$F0,$47
	.dc.b	$C1,$9E,$0E,$8F,$D7,$D7,$73,$AE,$3F,$02,$77,$19,$9E,$E6,$66,$3F
	.dc.b	$00,$E7,$7E,$1A,$EF,$7E,$2B,$E5,$5E,$13,$F5,$11,$E6,$3E,$AE,$3F
	.dc.b	$12,$3F,$86,$CF,$CE,$BF,$B7,$3F,$13,$0C,$D7,$CC,$7C,$2C,$3A,$E1
	.dc.b	$E2,$3D,$75,$F1,$EF,$C7,$87,$86,$7F,$23,$87,$86,$FF,$84,$57,$DD
	.dc.b	$8F,$04,$7F,$0B,$AF,$65,$F3,$3E,$06,$1F,$13,$1E,$44,$C3,$E6,$DF
	.dc.b	$F1,$6E,$6F,$F3,$A7,$57,$F3,$FD,$FF,$B5,$74,$FF,$89,$2B,$FD,$43
	.dc.b	$39,$55,$FE,$7B,$BC,$3C,$A3,$64,$3F,$D0,$D1,$AA,$FF,$72,$C7,$FA
	.dc.b	$B3,$DE,$C1,$9B,$B1,$92,$1F,$E8,$F7,$BF,$3B,$B6,$92,$7F,$BB,$9E
	.dc.b	$68,$7C,$5E,$63,$FB,$C4,$DD,$7F,$BF,$CF,$97,$69,$5F,$EC,$07,$D1
	.dc.b	$EB,$E2,$3F,$D4,$11,$6B,$FC,$27,$3E,$D9,$87,$FB,$20,$F8,$1F,$19
	.dc.b	$87,$FD,$50,$59,$FF,$0D,$C7,$CD,$87,$D2,$7F,$18,$9F,$F5,$F3,$78
	.dc.b	$3F,$85,$B6,$7F,$8A,$FD,$2F,$FA,$FD,$F8,$27,$2E,$4E,$93,$B8,$C1
	.dc.b	$A9,$E0,$FF,$C0,$1F,$A5,$FF,$4B,$BC,$9A,$E7,$CF,$F3,$7D,$2B,$C4
	.dc.b	$F4,$56,$B3,$FF,$6A,$37,$3F,$E4,$2D,$C3,$F9,$5D,$F1,$7C,$FE,$E7
	.dc.b	$8A,$FF,$DB,$8D,$BF,$E5,$2D,$9F,$E6,$7E,$F5,$51,$49,$ED,$7C,$B9
	.dc.b	$B3,$FE,$EA,$37,$15,$A6,$67,$E1,$BF,$EF,$37,$F1,$2F,$9B,$8B,$EB
	.dc.b	$46,$23,$FC,$3F,$7F,$CF,$BE,$95,$EE,$6A,$E5,$71,$3E,$AC,$E9,$FC
	.dc.b	$BF,$5B,$E9,$3F,$C5,$F1,$86,$BD,$8F,$D8,$B8,$D3,$FD,$DA,$BE,$2B
	.dc.b	$F6,$6E,$72,$FF,$32,$36,$B8,$F8,$7A,$AF,$C8,$7D,$1B,$E1,$E9,$2B
	.dc.b	$F1,$7F,$33,$F8,$99,$D4,$7E,$39,$AF,$94,$38,$C9,$97,$F6,$2A,$E5
	.dc.b	$7F,$46,$B8,$DF,$F6,$63,$8C,$64,$FB,$19,$BE,$85,$F4,$7F,$C4,$1B
	.dc.b	$5F,$29,$AF,$C7,$9E,$1F,$49,$BA,$93,$B7,$94,$7C,$2E,$29,$3E,$D7
	.dc.b	$F6,$99,$3E,$87,$33,$EF,$CF,$F2,$2E,$CF,$8F,$1C,$3E,$92,$47,$CA
	.dc.b	$79,$E3,$88,$F9,$16,$9C,$6D,$F2,$27,$FC,$23,$3E,$C8,$C4,$23,$83
	.dc.b	$D4,$87,$4A,$F8,$DC,$B2,$3E,$33,$F2,$6F,$11,$F1,$F5,$FD,$76,$BF
	.dc.b	$3E,$B5,$3F,$37,$CF,$FA,$04,$7E,$87,$D2,$57,$94,$5C,$FD,$57,$F6
	.dc.b	$6B,$F8,$F7,$D2,$3E,$22,$93,$0F,$E3,$3E,$97,$F4,$67,$F1,$93,$B7
	.dc.b	$F8,$EB,$5C,$CF,$F2,$99,$E1,$87,$C6,$AD,$57,$E3,$5A,$FF,$8D,$1A
	.dc.b	$FF,$8F,$9B,$87,$E3,$F7,$FC,$71,$27,$F1,$DE,$9F,$F3,$7F,$A3,$FD
	.dc.b	$64,$1A,$7E,$BE,$BF,$D5,$3F,$0D,$F9,$7F,$A0,$95,$F9,$77,$F3,$5E
	.dc.b	$E7,$E6,$FE,$7E,$4F,$E6,$DF,$CF,$7D,$27,$7F,$98,$93,$FC,$FB,$24
	.dc.b	$7E,$8F,$D6,$FD,$27,$7C,$AF,$D2,$3F,$A7,$EF,$FD,$47,$CB,$C9,$FD
	.dc.b	$43,$FA,$BB,$FE,$A5,$E1,$BE,$2E,$92,$BF,$A6,$FB,$EF,$85,$FD,$97
	.dc.b	$A7,$27,$F6,$4F,$ED,$6F,$FC,$0B,$EE,$7F,$03,$F1,$BF,$DE,$CD,$AF
	.dc.b	$F0,$5B,$79,$7F,$D4,$3F,$05,$E2,$E6,$95,$FD,$73,$04,$7F,$84,$FE
	.dc.b	$61,$F7,$EF,$63,$FC,$27,$F4,$AF,$8B,$A4,$C7,$F9,$7C,$3E,$2F,$97
	.dc.b	$B2,$F8,$DA,$4A,$FE,$19,$ED,$7F,$75,$FA,$4F,$E7,$11,$FC,$7D,$BE
	.dc.b	$27,$ED,$1B,$C7,$FE,$88,$D9,$8F,$5B,$FE,$7F,$F4,$AF,$9E,$FF,$14
	.dc.b	$B6,$33,$1E,$DF,$1F,$DF,$1F,$E1,$CB,$E7,$F2,$DE,$3E,$99,$39,$C5
	.dc.b	$38,$8F,$34,$DD,$3F,$86,$5F,$29,$E9,$83,$E8,$FC,$04,$8F,$F0,$5E
	.dc.b	$BF,$7D,$B3,$F0,$C8,$F9,$7A,$5E,$3E,$23,$E3,$4D,$C7,$C3,$5A,$9E
	.dc.b	$67,$F7,$5B,$57,$F7,$56,$C6,$63,$1F,$13,$F4,$C9,$1E,$1F,$22,$BF
	.dc.b	$56,$2E,$1E,$09,$FF,$2B,$F8,$FF,$3E,$BF,$A3,$36,$7F,$8B,$B7,$8B
	.dc.b	$FD,$83,$DB,$FD,$47,$E2,$78,$F1,$CC,$3F,$BC,$DE,$DF,$F3,$2F,$A0
	.dc.b	$AE,$FD,$BF,$EC,$2D,$FF,$C1,$78,$B0,$FB,$B8,$7F,$C0,$0E,$EE,$F1
	.dc.b	$8A,$DB,$FB,$39,$FF,$69,$C6,$2F,$A5,$70,$6C,$FF,$9D,$AB,$94,$F4
	.dc.b	$F0,$B3,$2F,$A4,$FD,$06,$5D,$2B,$FD,$AF,$18,$9F,$21,$FE,$E2,$71
	.dc.b	$8F,$F8,$85,$EB,$19,$EB,$F8,$0F,$03,$B3,$89,$E0,$DA,$EF,$97,$92
	.dc.b	$FA,$C7,$8A,$75,$7D,$70,$F3,$3E,$3B,$E1,$3A,$8E,$F3,$A6,$9E,$3B
	.dc.b	$F3,$B8,$C3,$C7,$3B,$9F,$0C,$F5,$31,$5E,$C7,$FC,$2D,$1F,$15,$B1
	.dc.b	$97,$D1,$3F,$E4,$FF,$8A,$FB,$DC,$F7,$43,$49,$5F,$E2,$13,$EA,$7A
	.dc.b	$FF,$C2,$1C,$63,$FE,$5E,$EB,$FF,$2D,$36,$33,$3B,$7C,$AF,$1C,$FE
	.dc.b	$B3,$75,$4C,$F8,$47,$C6,$7F,$83,$EF,$78,$FC,$3C,$38,$5F,$E0,$5A
	.dc.b	$AF,$8F,$26,$EA,$4F,$CF,$AD,$D6,$A3,$13,$AB,$EF,$1F,$87,$9D,$46
	.dc.b	$09,$8F,$E5,$2F,$9E,$3E,$78,$C6,$1F,$84,$E2,$FD,$67,$CC,$E7,$2F
	.dc.b	$D1,$7C,$D5,$9C,$E9,$BD,$D3,$95,$76,$30,$D9,$E1,$C5,$FE,$35,$E7
	.dc.b	$FE,$0F,$F2,$96,$7F,$73,$37,$2F,$D9,$4D,$CB,$F8,$04,$DC,$93,$63
	.dc.b	$0E,$7E,$1B,$5B,$DC,$7F,$87,$B7,$1F,$E3,$2D,$CB,$F8,$CB,$D6,$7B
	.dc.b	$78,$35,$FE,$2E,$DA,$FE,$42,$DA,$FE,$42,$D9,$FE,$26,$D9,$FE,$18
	.dc.b	$FA,$FE,$A5,$F3,$74,$C7,$87,$6F,$FD,$3B,$3F,$E0,$76,$CE,$9A,$6A
	.dc.b	$71,$7F,$E7,$9A,$FF,$55,$CE,$4E,$7E,$C7,$84,$F1,$7C,$D5,$8E,$FF
	.dc.b	$37,$86,$9E,$28,$F4,$79,$3E,$99,$B3,$AC,$3C,$1E,$87,$B4,$F7,$79
	.dc.b	$96,$7D,$7D,$3F,$07,$E2,$BE,$13,$63,$DF,$E3,$FC,$53,$E3,$7C,$A5
	.dc.b	$8F,$87,$F2,$9F,$24,$F9,$5E,$4D,$8F,$8F,$F2,$1F,$37,$E4,$FE,$73
	.dc.b	$63,$E5,$F8,$5F,$44,$FA,$5F,$25,$63,$E7,$F8,$3F,$57,$E4,$BE,$B3
	.dc.b	$63,$FD,$9F,$F9,$93,$E2,$BE,$D7,$1F,$E5,$63,$13,$F6,$23,$1F,$0F
	.dc.b	$F9,$7F,$A7,$B3,$FD,$81,$E1,$F0,$DF,$9B,$F5,$6C,$FF,$64,$7A,$7A
	.dc.b	$BD,$7E,$1E,$CF,$89,$F0,$7C,$2F,$87,$E1,$AC,$F9,$9F,$27,$C3,$79
	.dc.b	$7D,$8B,$3E,$A6,$EB,$FE,$BF,$D9,$18,$F1,$BF,$D2,$31,$8A,$F8,$37
	.dc.b	$CD,$A7,$33,$E2,$FC,$2F,$C3,$3E,$27,$99,$63,$24,$C7,$C1,$19,$7C
	.dc.b	$5F,$0B,$F2,$CF,$99,$F0,$B6,$BE,$8C,$59,$F4,$4D,$CB,$CF,$F0,$BF
	.dc.b	$4F,$E1,$7E,$A3,$70,$FB,$2D,$BF,$80,$DA,$E3,$3F,$7A,$B8,$B4,$FC
	.dc.b	$0F,$C2,$FF,$11,$FA,$7F,$C0,$36,$32,$4C,$3E,$67,$AD,$A8,$FE,$93
	.dc.b	$87,$FB,$4B,$E6,$BF,$79,$48,$FE,$49,$F2,$7F,$E4,$2A,$D9,$FD,$42
	.dc.b	$55,$C2,$B7,$CD,$FC,$9E,$73,$E7,$FF,$B7,$9E,$15,$D9,$E4,$6A,$FF
	.dc.b	$B4,$34,$CD,$BE,$6F,$FD,$F1,$3E,$29,$F3,$47,$95,$F4,$47,$B2,$37
	.dc.b	$18,$37,$87,$F3,$BA,$E9,$CD,$6B,$94,$7F,$43,$78,$B6,$75,$3B,$6D
	.dc.b	$F8,$E7,$CD,$1F,$B1,$F7,$A4,$FE,$DB,$E1,$6E,$3F,$9D,$9B,$7F,$9C
	.dc.b	$9B,$5F,$37,$E6,$DF,$1D,$FD,$37,$E9,$FE,$64,$62,$7F,$CA,$F1,$89
	.dc.b	$91,$F1,$7D,$C7,$E0,$BF,$E9,$58,$F9,$19,$B0,$E4,$FC,$5C,$9F,$40
	.dc.b	$FA,$4F,$8C,$BE,$56,$C6,$1B,$87,$C1,$9D,$36,$74,$D3,$53,$C2,$FC
	.dc.b	$15,$C1,$FA,$C3,$E9,$5E,$B7,$D7,$FA,$AF,$F4,$5E,$1E,$A9,$CB,$F4
	.dc.b	$63,$35,$CA,$72,$DA,$FB,$DD,$3F,$95,$9C,$E6,$C3,$F3,$9F,$99,$FF
	.dc.b	$9F,$DF,$F4,$1F,$D5,$7E,$6E,$31,$8F,$FA,$3A,$31,$C9,$2B,$E8,$70
	.dc.b	$7F,$BB,$DF,$9F,$7C,$30,$ED,$83,$87,$EA,$FC,$1F,$F8,$2F,$0F,$F7
	.dc.b	$1E,$1C,$A3,$1F,$17,$96,$C6,$4E,$F1,$C7,$CA,$E5,$E6,$59,$F1,$B7
	.dc.b	$1F,$D1,$63,$F8,$48,$C6,$5E,$48,$C5,$7A,$B9,$FC,$9E,$07,$95,$B1
	.dc.b	$93,$DF,$19,$69,$F0,$BE,$53,$D9,$C1,$ED,$6C,$64,$F8,$F1,$96,$9F
	.dc.b	$2B,$9B,$19,$FA,$37,$FC,$36,$7E,$75,$87,$C9,$AD,$CE,$5B,$7F,$E7
	.dc.b	$3F,$89,$FF,$33,$FD,$13,$D7,$FE,$35,$6F,$23,$EA,$CE,$73,$FE,$2B
	.dc.b	$7C,$2F,$75,$A7,$A4,$AB,$83,$FC,$A6,$74,$F2,$35,$3C,$B0,$E9,$19
	.dc.b	$AE,$38,$78,$E7,$33,$FE,$CF,$38,$9F,$CE,$2F,$AB,$9A,$E3,$18,$8F
	.dc.b	$F6,$69,$C0,$F2,$4F,$47,$33,$C2,$31,$0C,$3C,$37,$E3,$8E,$47,$0C
	.dc.b	$3D,$4E,$0E,$07,$78,$E1,$7F,$12,$38,$1E,$9B,$EF,$7D,$B0,$F4,$4F
	.dc.b	$2A,$F2,$3D,$67,$55,$C2,$3B,$57,$09,$CC,$6C,$CC,$EE,$B9,$9E,$38
	.dc.b	$ED,$87,$AA,$31,$5C,$9B,$3C,$8E,$8F,$29,$FC,$0B,$63,$ED,$47,$33
	.dc.b	$25,$72,$7C,$13,$CD,$CA,$EE,$FC,$5B,$63,$F1,$0D,$7F,$04,$DF,$8D
	.dc.b	$79,$7F,$DD,$E5,$39,$8C,$37,$4F,$F6,$3B,$67,$77,$FD,$EE,$F9,$D7
	.dc.b	$E4,$30,$7A,$BF,$DF,$F2,$FC,$EC,$63,$0F,$CA,$DF,$07,$F1,$CD,$9D
	.dc.b	$D7,$1A,$FF,$04,$1C,$6B,$79,$7A,$6B,$F5,$37,$C6,$B7,$7C,$6F,$FD
	.dc.b	$BF,$1C,$6B,$77,$C6,$3D,$4D,$9D,$D7,$17,$F8,$03,$6F,$B4,$7F,$BA
	.dc.b	$9A,$E8,$F8,$51,$B9,$E9,$09,$DD,$78,$CF,$3C,$7C,$49,$DC,$EA,$BC
	.dc.b	$55,$BA,$F0,$56,$8E,$B1,$98,$F1,$1F,$1B,$0E,$87,$91,$86,$49,$87
	.dc.b	$D0,$72,$FC,$68,$E3,$1F,$C0,$E7,$B1,$DF,$0F,$82,$79,$E3,$8C,$7C
	.dc.b	$C9,$E3,$5D,$67,$96,$1F,$16,$73,$5F,$0A,$72,$60,$FE,$C1,$C1,$FE
	.dc.b	$0D,$7F,$B3,$FC,$5F,$F3,$2D,$91,$A8,$C3,$C3,$C0,$BC,$DF,$E1,$C6
	.dc.b	$DF,$FC,$00,$9E,$67,$F7,$28,$CD,$6B,$E7,$A4,$E8,$FE,$99,$18,$7F
	.dc.b	$A7,$47,$07,$FB,$DF,$CA,$F6,$F8,$D9,$18,$21,$FD,$55,$CB,$FD,$5E
	.dc.b	$3C,$BF,$17,$E6,$F8,$D9,$3F,$E0,$68,$CB,$FE,$08,$8C,$BF,$D7,$E3
	.dc.b	$DB,$F1,$9E,$EF,$8D,$93,$FE,$0E,$8C,$BF,$E1,$08,$CB,$FD,$9E,$3E
	.dc.b	$4A,$FC,$D1,$F0,$58,$6C,$7F,$91,$1A,$FF,$8C,$BE,$E9,$F1,$1B,$AE
	.dc.b	$70,$F8,$9A,$66,$BF,$C9,$BF,$42,$F1,$3C,$2D,$8C,$C7,$93,$EB,$9E
	.dc.b	$DE,$A3,$1E,$F4,$AD,$9F,$E5,$8E,$37,$9B,$EC,$5E,$CD,$31,$12,$78
	.dc.b	$46,$5E,$51,$97,$FC,$9F,$3C,$9D,$9A,$3F,$CC,$FC,$4F,$77,$E5,$B6
	.dc.b	$33,$1F,$3B,$E8,$7E,$4F,$77,$F8,$FA,$55,$B7,$FC,$BB,$2F,$8A,$E0
	.dc.b	$74,$BF,$CE,$FC,$5B,$E4,$A9,$2F,$FC,$9B,$E0,$57,$FC,$77,$3B,$7F
	.dc.b	$D0,$0D,$7F,$C5,$11,$F8,$2F,$AE,$7A,$17,$86,$57,$FA,$12,$72,$D9
	.dc.b	$CB,$FE,$74,$8F,$B4,$D8,$E8,$6B,$5F,$81,$3A,$6C,$7E,$4B,$E8,$7F
	.dc.b	$D1,$55,$FE,$BE,$6B,$FE,$C2,$0E,$1F,$F6,$2B,$5F,$F6,$83,$C2,$7E
	.dc.b	$77,$D0,$FD,$99,$FF,$44,$B6,$3E,$94,$EF,$1F,$E1,$BE,$23,$37,$C0
	.dc.b	$7D,$02,$63,$F2,$BE,$8E,$BF,$E6,$8E,$7F,$F2,$7B,$FE,$AD,$6B,$F4
	.dc.b	$A7,$EC,$FC,$A7,$DA,$F9,$59,$3F,$52,$32,$FD,$68,$CB,$F6,$A3,$2F
	.dc.b	$DC,$8C,$E9,$FE,$75,$C5,$5F,$86,$F9,$59,$7F,$80,$7F,$13,$F2,$B8
	.dc.b	$8F,$C2,$C5,$BE,$53,$F8,$EF,$76,$57,$E2,$1F,$C8,$F7,$E2,$3F,$1B
	.dc.b	$16,$BC,$53,$FC,$33,$17,$C8,$5F,$3A,$61,$FE,$CE,$9C,$B6,$72,$FF
	.dc.b	$AF,$A3,$FC,$05,$7F,$A7,$77,$19,$9D,$CE,$6B,$C1,$1F,$A6,$F9,$5F
	.dc.b	$D3,$F8,$B2,$FF,$43,$1F,$AA,$C3,$F5,$7D,$72,$7F,$4B,$1F,$AE,$F9
	.dc.b	$5F,$E9,$7C,$5F,$85,$9F,$F5,$A3,$79,$3F,$0D,$3F,$B4,$F9,$5D,$F9
	.dc.b	$F2,$FF,$5F,$1F,$B8,$F9,$5F,$DC,$FC,$AC,$9F,$D9,$C7,$EF,$36,$FD
	.dc.b	$EE,$F2,$7F,$6F,$1F,$D7,$3A,$DF,$2B,$49,$3F,$BC,$35,$F2,$BB,$F8
	.dc.b	$99,$1F,$EF,$FA,$4F,$3F,$95,$93,$97,$B7,$D1,$77,$F8,$C9,$3C,$5F
	.dc.b	$17,$CB,$78,$FE,$5B,$C9,$F1,$F2,$BB,$47,$9F,$C9,$91,$E1,$7D,$5E
	.dc.b	$7F,$AF,$9F,$D9,$EF,$CA,$F3,$45,$7F,$DA,$7B,$7F,$AC,$5F,$57,$C1
	.dc.b	$E7,$BB,$FD,$A9,$2F,$95,$F8,$C4,$AD,$9F,$07,$E5,$7E,$51,$1F,$89
	.dc.b	$F2,$B5,$EF,$EA,$7C,$F3,$18,$76,$71,$1F,$2D,$CF,$97,$FC,$23,$1F
	.dc.b	$F3,$95,$79,$3D,$4F,$B0,$7D,$9F,$84,$98,$7D,$38,$CB,$F5,$63,$2F
	.dc.b	$C0,$8C,$BF,$0A,$3F,$C3,$BE,$83,$E7,$3F,$09,$F0,$D2,$BF,$C4,$B8
	.dc.b	$7F,$BD,$EB,$51,$8C,$BF,$8E,$C6,$23,$E6,$D7,$17,$FE,$00,$C6,$6B
	.dc.b	$97,$C6,$FE,$8B,$D8,$95,$CC,$78,$CB,$57,$98,$DC,$FF,$25,$AC,$3C
	.dc.b	$6B,$87,$AB,$FE,$3C,$8F,$F6,$66,$FE,$6F,$8E,$C7,$C4,$FF,$71,$BE
	.dc.b	$D1,$89,$F5,$DD,$BF,$1E,$5F,$D4,$B0,$E0,$AF,$58,$72,$AF,$F2,$5D
	.dc.b	$6A,$F9,$43,$E6,$5F,$07,$FE,$A8,$8F,$29,$FE,$52,$8C,$ED,$CB,$3F
	.dc.b	$54,$62,$76,$73,$D7,$AC,$67,$1F,$EA,$8D,$AE,$38,$77,$9C,$D7,$08
	.dc.b	$C4,$EB,$0F,$06,$1E,$39,$CB,$E5,$AE,$13,$A8,$F6,$4E,$9B,$19,$C3
	.dc.b	$EC,$C5,$9D,$47,$96,$3B,$4E,$EB,$AC,$7F,$75,$9F,$80,$FD,$B3,$A9
	.dc.b	$9A,$F9,$4E,$7E,$2B,$FC,$7B,$F3,$5F,$DF,$63,$FD,$51,$D4,$8C,$4E
	.dc.b	$A7,$FC,$04,$D7,$B5,$FF,$8A,$BC,$7E,$4D,$8F,$27,$BD,$FD,$21,$FA
	.dc.b	$AD,$7E,$D3,$FD,$4B,$5F,$A6,$DD,$7F,$29,$E5,$3D,$17,$37,$C7,$AD
	.dc.b	$61,$CB,$2C,$1F,$3E,$F8,$56,$27,$F0,$79,$79,$72,$EF,$7B,$C7,$AD
	.dc.b	$6B,$1E,$AF,$E4,$DB,$7C,$72,$F2,$D7,$0C,$39,$DF,$1C,$3C,$11,$FE
	.dc.b	$21,$CB,$C7,$39,$3F,$37,$C5,$FE,$6E,$F8,$7E,$58,$FF,$88,$B7,$7C
	.dc.b	$46,$2F,$FB,$0D,$5B,$FF,$3A,$CD,$A4,$D8,$C3,$76,$FC,$BD,$7F,$9F
	.dc.b	$38,$7D,$FD,$1F,$02,$31,$3F,$E2,$9F,$88,$CB,$E3,$AB,$97,$F7,$26
	.dc.b	$D7,$08,$FF,$19,$56,$32,$FD,$CD,$FF,$8C,$FE,$23,$1C,$59,$6C,$64
	.dc.b	$FF,$1C,$46,$5A,$7F,$8E,$CE,$9F,$11,$D5,$A6,$4F,$F2,$03,$96,$99
	.dc.b	$C3,$FC,$8D,$19,$6C,$7F,$23,$F9,$AF,$F3,$E3,$7E,$0F,$DB,$1F,$E5
	.dc.b	$28,$C1,$BF,$80,$CC,$E9,$B1,$92,$53,$6F,$9D,$6F,$81,$E1,$5E,$1F
	.dc.b	$80,$EF,$F0,$3E,$06,$C7,$97,$E0,$30,$D3,$81,$E9,$F8,$1F,$3C,$7A
	.dc.b	$1B,$1E,$DF,$81,$F5,$9E,$CF,$81,$B1,$F0,$7B,$3D,$FF,$03,$F0,$1B
	.dc.b	$1F,$17,$E0,$3E,$19,$F1,$3E,$06,$C7,$C9,$F8,$1F,$8E,$7C,$8F,$88
	.dc.b	$B1,$F3,$7E,$07,$E5,$FC,$5F,$CC,$6C,$7D,$1F,$07,$E7,$9F,$43,$E0
	.dc.b	$AC,$7D,$5F,$13,$E9,$FC,$17,$D4,$6C,$7F,$5B,$F9,$AF,$F1,$AB,$F6
	.dc.b	$FA,$7F,$2B,$18,$7E,$94,$7F,$A3,$BA,$11,$AF,$81,$93,$CF,$E0,$AC
	.dc.b	$F0,$3F,$B7,$DB,$E0,$BA,$9D,$BE,$0A,$DF,$8F,$E0,$BC,$3F,$05,$E2
	.dc.b	$6C,$79,$FE,$07,$CB,$F0,$5E,$66,$C7,$AF,$E0,$BD,$3F,$05,$EA,$6C
	.dc.b	$64,$F8,$11,$96,$99,$D3,$D1,$E5,$B5,$56,$E7,$31,$8A,$F9,$5F,$05
	.dc.b	$F2,$0F,$93,$E5,$D8,$C9,$2B,$8C,$65,$E7,$F0,$5F,$40,$FA,$3F,$05
	.dc.b	$6B,$EB,$45,$9F,$0C,$DF,$23,$F8,$96,$2F,$82,$FB,$3F,$05,$F6,$9B
	.dc.b	$8F,$DF,$6E,$1E,$A6,$DF,$26,$CF,$FB,$32,$FA,$3F,$87,$BF,$C2,$FC
	.dc.b	$0C,$8F,$C5,$7A,$B7,$1F,$8B,$37,$1F,$2C,$DA,$FF,$14,$7C,$CF,$F0
	.dc.b	$5B,$FF,$73,$35,$FE,$E4,$D7,$7E,$77,$F4,$E7,$2F,$F2,$C5,$1C,$EF
	.dc.b	$F9,$AF,$CB,$FF,$8A,$6F,$FC,$81,$5A,$BF,$E6,$74,$F9,$97,$36,$17
	.dc.b	$3D,$8C,$37,$0F,$F2,$55,$71,$8C,$61,$F9,$F8,$C4,$EA,$FA,$B6,$B7
	.dc.b	$3F,$D2,$A7,$BC,$FE,$CF,$DE,$FD,$89,$FB,$2F,$84,$B1,$92,$61,$F9
	.dc.b	$28,$CB,$FA,$F1,$F3,$2E,$EF,$8C,$DA,$FF,$2D,$7C,$87,$F9,$CD,$F8
	.dc.b	$B1,$9A,$E4,$DC,$BE,$83,$67,$F8,$0F,$52,$F9,$5F,$0C,$3E,$6E,$EF
	.dc.b	$89,$BF,$25,$86,$E3,$F2,$A7,$4D,$9D,$34,$D4,$F0,$BF,$9D,$5C,$1B
	.dc.b	$5F,$83,$F9,$87,$A5,$4F,$B9,$CD,$71,$FC,$1C,$E9,$B3,$A6,$9F,$88
	.dc.b	$F9,$87,$0F,$E3,$DB,$1F,$2B,$E6,$F1,$32,$38,$3F,$DF,$B6,$4E,$73
	.dc.b	$FE,$1B,$1F,$95,$6F,$C3,$6F,$E2,$38,$37,$0F,$AE,$D9,$DC,$7D,$AF
	.dc.b	$8F,$ED,$F1,$3D,$DB,$19,$3E,$FC,$65,$A7,$E0,$BE,$67,$CD,$F3,$3E
	.dc.b	$76,$C6,$4F,$C3,$C6,$5A,$7E,$2B,$E6,$3D,$DF,$33,$EF,$6C,$64,$FC
	.dc.b	$7C,$65,$A7,$E4,$B8,$F1,$97,$C4,$6F,$4F,$F8,$B7,$64,$62,$BF,$23
	.dc.b	$18,$7B,$57,$E8,$4F,$F5,$28,$F9,$97,$AD,$FE,$B7,$C1,$E3,$FF,$96
	.dc.b	$27,$35,$FE,$81,$F8,$BF,$EB,$B1,$88,$FF,$09,$4E,$EF,$FD,$0D,$0A
	.dc.b	$DC,$75,$9D,$4F,$33,$C5,$19,$C3,$71,$9B,$ED,$87,$92,$73,$3F,$EC
	.dc.b	$A3,$89,$FE,$E6,$BE,$AE,$6B,$8C,$62,$3E,$0C,$62,$79,$C6,$2F,$C4
	.dc.b	$F4,$39,$38,$9F,$81,$1B,$3C,$13,$C9,$F8,$07,$1B,$F4,$3C,$CE,$07
	.dc.b	$C8,$8E,$18,$7B,$27,$75,$DA,$B9,$BE,$77,$91,$8B,$F2,$1F,$42,$B9
	.dc.b	$61,$EF,$9D,$1E,$63,$8C,$6B,$0F,$0C,$79,$E7,$84,$E6,$36,$66,$77
	.dc.b	$7E,$33,$CD,$3B,$6C,$EC,$E2,$EE,$FC,$D1,$98,$D9,$97,$F0,$9C,$8E
	.dc.b	$ED,$B6,$33,$18,$3F,$C5,$EB,$E3,$5E,$5F,$F9,$96,$3F,$8A,$CE,$A7
	.dc.b	$3A,$7F,$8D,$AF,$A5,$71,$BE,$11,$8A,$DC,$F2,$3F,$C7,$AE,$30,$FD
	.dc.b	$16,$2F,$57,$C7,$18,$6C,$78,$70,$F2,$4F,$EA,$6B,$94,$FA,$8E,$57
	.dc.b	$F9,$E6,$BF,$93,$8D,$BE,$2A,$E7,$3C,$6F,$BC,$FF,$94,$A3,$8D,$F7
	.dc.b	$AD,$CF,$1C,$7B,$5F,$08,$FE,$06,$DC,$3B,$C7,$C5,$9D,$B0,$F2,$57
	.dc.b	$47,$47,$EA,$70,$F0,$9E,$57,$F7,$35,$D0,$EB,$1A,$8F,$45,$F8,$6B
	.dc.b	$A5,$6A,$3C,$D5,$BA,$EF,$87,$B7,$0F,$23,$F4,$63,$BE,$1C,$8F,$35
	.dc.b	$71,$39,$61,$E3,$8C,$E1,$D8,$F1,$61,$C8,$EB,$5E,$BB,$F8,$31,$97
	.dc.b	$E0,$4F,$17,$18,$79,$5C,$9C,$8F,$A4,$7F,$1A,$F0,$3F,$C1,$1D,$FF
	.dc.b	$CC,$F6,$FE,$77,$F1,$96,$95,$FE,$0D,$8C,$3C,$23,$FC,$33,$58,$E1
	.dc.b	$FF,$7F,$C7,$F8,$72,$B1,$39,$75,$F2,$12,$74,$43,$FB,$9B,$97,$FB
	.dc.b	$AC,$75,$F8,$BE,$DF,$1B,$27,$FC,$53,$19,$7F,$C5,$71,$97,$FB,$EC
	.dc.b	$79,$7E,$2F,$CD,$F1,$B2,$7F,$C6,$31,$97,$FC,$67,$19,$7F,$C0,$B1
	.dc.b	$ED,$F8,$CF,$77,$C6,$C9,$FF,$1C,$C6,$5F,$F1,$DC,$65,$FF,$06,$C7
	.dc.b	$CB,$5F,$32,$F8,$2C,$36,$3F,$CC,$AD,$7F,$C9,$5F,$32,$F8,$8D,$D4
	.dc.b	$9D,$BD,$67,$51,$FE,$6E,$F9,$97,$87,$E1,$6C,$66,$3C,$9F,$32,$F6
	.dc.b	$B5,$12,$74,$7F,$9E,$27,$4F,$F9,$EB,$89,$E6,$FB,$57,$AD,$4C,$44
	.dc.b	$9E,$11,$97,$94,$65,$FF,$34,$4F,$27,$66,$8F,$F4,$4F,$03,$DC,$F9
	.dc.b	$8D,$8C,$C7,$CF,$F9,$9F,$95,$DB,$FE,$97,$AD,$57,$F9,$D5,$B1,$9A
	.dc.b	$E1,$BF,$E7,$BE,$29,$F2,$54,$98,$7F,$9E,$3B,$6B,$FE,$D6,$9D,$BF
	.dc.b	$EF,$4C,$13,$FE,$76,$9C,$CF,$F9,$42,$3F,$09,$F2,$AF,$36,$F0,$CA
	.dc.b	$FF,$53,$CE,$5B,$39,$7F,$D1,$D1,$F7,$1B,$1D,$0D,$61,$FE,$B7,$9D
	.dc.b	$36,$3F,$29,$F3,$3F,$E3,$CA,$FF,$5A,$B5,$FF,$4D,$36,$7F,$D8,$1F
	.dc.b	$31,$FD,$69,$FF,$61,$46,$E7,$FB,$28,$70,$FF,$68,$6A,$35,$E5,$FF
	.dc.b	$8E,$9F,$E1,$F9,$BD,$CF,$A4,$48,$F9,$9A,$7C,$BF,$95,$C4,$7D,$19
	.dc.b	$B8,$FD,$08,$E7,$3F,$6F,$E5,$3E,$E7,$CA,$C9,$E6,$78,$23,$4E,$3C
	.dc.b	$66,$4A,$E1,$3F,$7E,$B6,$FE,$06,$33,$87,$86,$F8,$D7,$E2,$FE,$52
	.dc.b	$5F,$5C,$11,$F8,$FF,$95,$CC,$FE,$2E,$2E,$1C,$E9,$5F,$95,$F6,$A5
	.dc.b	$F3,$7F,$31,$F2,$B8,$8F,$CA,$45,$BE,$53,$FC,$4E,$DF,$1F,$7C,$D9
	.dc.b	$87,$FB,$DA,$72,$D9,$CB,$FE,$DD,$8F,$F0,$3D,$FC,$F8,$F8,$F3,$D4
	.dc.b	$DC,$66,$77,$C1,$FE,$2A,$7A,$D7,$EA,$7E,$57,$F6,$1E,$24,$8F,$8D
	.dc.b	$1F,$B3,$9F,$DA,$7C,$AC,$9F,$95,$1F,$B7,$F9,$59,$1F,$35,$FD,$D7
	.dc.b	$67,$EE,$FE,$56,$4F,$D1,$7F,$7D,$E9,$7E,$FF,$E5,$7F,$80,$7C,$AC
	.dc.b	$AF,$AD,$1F,$C0,$F7,$FE,$09,$C1,$27,$ED,$47,$F6,$4E,$C7,$CC,$D2
	.dc.b	$4F,$DF,$35,$F2,$BB,$F8,$79,$1F,$8F,$A4,$F3,$F9,$59,$39,$7B,$7C
	.dc.b	$AF,$7F,$8C,$93,$C5,$F1,$7C,$AF,$8F,$E5,$7C,$9F,$1F,$2B,$B4,$79
	.dc.b	$FC,$A9,$1E,$17,$D5,$F2,$BE,$BF,$4B,$D9,$EF,$CA,$F3,$45,$5D,$DF
	.dc.b	$9F,$DF,$FD,$77,$1E,$BF,$84,$D4,$E5,$F8,$DF,$2B,$27,$47,$C1,$F9
	.dc.b	$5F,$94,$47,$E2,$7C,$A5,$7B,$F8,$5F,$C2,$7C,$86,$5F,$D1,$E7,$27
	.dc.b	$F8,$61,$C6,$5E,$5C,$3E,$B7,$AD,$2B,$E8,$3D,$67,$D5,$F0,$DF,$6F
	.dc.b	$D7,$95,$F5,$63,$EF,$18,$7B,$E3,$E2,$9C,$C6,$CF,$B8,$EE,$BA,$C7
	.dc.b	$F0,$DF,$49,$F3,$3F,$89,$F8,$79,$39,$8F,$F1,$7C,$EA,$3F,$0D,$3A
	.dc.b	$7C,$31,$96,$FC,$6F,$02,$3D,$FE,$3E,$BA,$3C,$51,$DA,$3B,$FA,$FF
	.dc.b	$E9,$EF,$8A,$FE,$E5,$A3,$07,$C7,$E7,$7F,$A5,$18,$CB,$F9,$E5,$F0
	.dc.b	$8F,$EA,$8D,$C3,$84,$FF,$93,$27,$83,$FE,$F9,$AE,$77,$AC,$3C,$17
	.dc.b	$C2,$9F,$30,$F8,$56,$BE,$0C,$7F,$D6,$B3,$98,$C5,$7F,$4F,$8C,$3D
	.dc.b	$CC,$6B,$D6,$33,$A7,$E2,$DB,$7D,$30,$EF,$7C,$4F,$1D,$F1,$8C,$DF
	.dc.b	$16,$D7,$1C,$3C,$77,$C0,$B8,$76,$7C,$D5,$C1,$B3,$AC,$3B,$3C,$27
	.dc.b	$53,$F1,$1B,$4A,$DC,$FF,$74,$35,$5F,$DB,$E7,$E1,$C7,$F7,$86,$BF
	.dc.b	$ED,$9E,$87,$CC,$FF,$75,$F8,$DF,$F2,$33,$63,$33,$D6,$72,$F7,$A7
	.dc.b	$BF,$FC,$FA,$3E,$AB,$5F,$A8,$D7,$EE,$B6,$38,$F9,$1F,$1A,$3F,$A9
	.dc.b	$EB,$FB,$FF,$29,$E9,$7F,$7F,$8F,$B1,$3A,$DF,$BD,$F1,$3E,$8D,$F0
	.dc.b	$EB,$C6,$5F,$6F,$2E,$ED,$C3,$93,$FE,$17,$AE,$0E,$6F,$8E,$3E,$07
	.dc.b	$AE,$5E,$3C,$BC,$35,$B2,$57,$A6,$75,$1F,$0A,$BF,$2F,$C5,$FA,$FF
	.dc.b	$81,$CB,$74,$CF,$2B,$E2,$B2,$AA,$92,$77,$37,$0F,$D0,$5F,$EE,$F9
	.dc.b	$3F,$C4,$DD,$3F,$E1,$7C,$5F,$15,$8C,$7F,$7B,$57,$0F,$DC,$4D,$C3
	.dc.b	$95,$7F,$1B,$8C,$5F,$ED,$63,$07,$F1,$FF,$89,$C9,$AE,$1B,$19,$26
	.dc.b	$1F,$C7,$23,$2F,$F1,$E8,$EA,$76,$F8,$9A,$FF,$2B,$2C,$7F,$24,$8B
	.dc.b	$9F,$B6,$69,$FC,$62,$2C,$FF,$B7,$BE,$15,$F0,$1E,$D7,$FD,$03,$5F
	.dc.b	$E4,$37,$2D,$F8,$5D,$4C,$8D,$7C,$0D,$BF,$E8,$B7,$FE,$49,$9E,$55
	.dc.b	$D6,$B5,$8F,$F9,$1E,$9F,$03,$F2,$BE,$17,$C4,$DC,$3C,$5F,$03,$E5
	.dc.b	$3C,$DF,$05,$63,$CD,$F0,$38,$69,$E9,$8F,$67,$C0,$FF,$8D,$F5,$FE
	.dc.b	$D5,$97,$FB,$36,$BF,$83,$72,$23,$13,$FE,$7C,$8C,$57,$18,$F9,$1F
	.dc.b	$03,$F1,$CF,$91,$D5,$70,$F9,$1E,$FF,$CB,$F8,$1F,$98,$D8,$F9,$9E
	.dc.b	$FF,$CF,$F8,$1F,$A0,$D8,$FA,$1F,$01,$F4,$CF,$A9,$F0,$36,$3E,$A7
	.dc.b	$C0,$FD,$73,$EC,$78,$96,$3E,$C6,$BF,$6C,$FB,$9E,$15,$8F,$F3,$F7
	.dc.b	$84,$F8,$0F,$CE,$36,$33,$F0,$38,$9C,$FB,$FF,$9E,$F9,$0B,$5F,$DE
	.dc.b	$DE,$3E,$FF,$E8,$BE,$46,$CF,$F7,$F7,$AF,$C0,$48,$F0,$FC,$0D,$D7
	.dc.b	$FC,$17,$A3,$DF,$F1,$9E,$4F,$81,$B7,$FE,$98,$F0,$9F,$29,$FC,$FE
	.dc.b	$30,$DC,$3F,$C4,$73,$AF,$63,$F7,$4F,$0F,$5A,$D6,$E3,$F8,$1B,$6F
	.dc.b	$FD,$B4,$AB,$75,$27,$BB,$CA,$31,$7E,$AF,$53,$E5,$FC,$17,$CC,$6D
	.dc.b	$FD,$06,$CE,$E3,$E9,$CF,$D2,$F8,$0F,$A7,$F0,$56,$7E,$BB,$6B,$E3
	.dc.b	$B7,$3F,$F5,$86,$C8,$FB,$7F,$03,$2F,$EF,$FC,$15,$C3,$E7,$CD,$C3
	.dc.b	$8C,$DA,$FF,$27,$60,$AC,$5B,$C5,$FE,$61,$78,$87,$CB,$FF,$B6,$70
	.dc.b	$FA,$DF,$36,$E4,$FE,$55,$5A,$F8,$8F,$A7,$39,$AF,$F6,$9C,$E5,$FF
	.dc.b	$72,$5F,$38,$FE,$F4,$D7,$FD,$DE,$D7,$CF,$ED,$7E,$68,$FC,$DF,$C2
	.dc.b	$5A,$FC,$FC,$59,$FC,$7C,$DD,$BF,$03,$E8,$7E,$10,$FC,$2F,$C0,$5C
	.dc.b	$3F,$1B,$16,$FF,$1B,$CE,$AC,$5F,$2B,$D9,$F9,$48,$CB,$F3,$3E,$13
	.dc.b	$F5,$FF,$09,$FB,$06,$C7,$ED,$1B,$87,$E9,$1B,$7F,$EC,$FF,$97,$75
	.dc.b	$FE,$6B,$E5,$3E,$6C,$63,$0F,$D1,$C6,$27,$F8,$C6,$0A,$D3,$FE,$25
	.dc.b	$6A,$21,$B8,$FE,$41,$7F,$B0,$F7,$FF,$52,$7E,$AB,$9E,$E3,$FB,$28
	.dc.b	$B7,$F1,$A6,$D7,$EB,$A6,$D7,$EB,$A6,$D7,$87,$EF,$FF,$D9,$73,$FB
	.dc.b	$C3,$F1,$2E,$23,$E8,$C6,$3C,$6F,$6C,$E6,$7D,$21,$C3,$EC,$6B,$EF
	.dc.b	$FA,$07,$AB,$FD,$EB,$5F,$CB,$C6,$30,$FF,$4C,$46,$27,$E5,$CE,$6B
	.dc.b	$FC,$9C,$D7,$C2,$D7,$FD,$71,$F6,$7F,$C2,$EF,$E7,$68,$73,$E0,$FF
	.dc.b	$83,$67,$FD,$4B,$1A,$32,$5A,$FD,$94,$73,$7F,$D6,$6D,$AE,$91,$AF
	.dc.b	$89,$E9,$7D,$7E,$26,$61,$CF,$B9,$8F,$F2,$1C,$3A,$D6,$9E,$AF,$97
	.dc.b	$C4,$C6,$5D,$AF,$8F,$47,$A4,$F5,$73,$DC,$BB,$64,$9C,$65,$E8,$AB
	.dc.b	$87,$3A,$79,$3F,$0C,$F8,$9E,$45,$C7,$9C,$5C,$3E,$0C,$DB,$E5,$C3
	.dc.b	$C2,$FC,$2D,$AF,$8E,$D9,$DC,$F9,$AB,$F8,$BD,$7F,$AF,$07,$CB,$BD
	.dc.b	$2F,$D2,$DB,$C3,$F9,$F3,$9A,$FC,$8F,$C3,$7E,$AE,$31,$0A,$E1,$8E
	.dc.b	$70,$FF,$4B,$5D,$AF,$C2,$E1,$D2,$73,$1E,$27,$C3,$19,$8E,$E6,$4E
	.dc.b	$35,$CD,$FF,$6E,$19,$3C,$51,$C5,$CE,$1E,$6C,$3C,$B1,$B3,$E1,$46
	.dc.b	$67,$84,$62,$FC,$77,$DE,$30,$78,$8E,$43,$0F,$26,$1D,$B0,$F8,$B3
	.dc.b	$C0,$F2,$D7,$37,$81,$E9,$9E,$17,$F1,$2B,$C9,$5B,$BE,$B8,$79,$B0
	.dc.b	$E3,$5A,$7C,$73,$AA,$E1,$1D,$6B,$84,$E6,$36,$66,$77,$5C,$CF,$4C
	.dc.b	$76,$BF,$86,$F0,$8D,$47,$9C,$CB,$C2,$BB,$36,$7B,$1C,$67,$67,$17
	.dc.b	$E6,$46,$FE,$29,$DB,$F8,$66,$CF,$E1,$1A,$FE,$09,$BF,$13,$FE,$07
	.dc.b	$CF,$FA,$04,$63,$C8,$FC,$B5,$23,$13,$AB,$FE,$8B,$7C,$6B,$51,$8B
	.dc.b	$E3,$87,$E8,$31,$7A,$7F,$D3,$23,$0D,$8E,$18,$FE,$5F,$1E,$95,$AC
	.dc.b	$7F,$35,$5C,$9D,$61,$DA,$38,$C7,$E7,$5D,$37,$0E,$CF,$F5,$DB,$EB
	.dc.b	$3C,$2F,$B4,$EC,$EB,$84,$70,$48,$FD,$DD,$72,$8E,$98,$78,$DF,$ED
	.dc.b	$57,$E2,$AE,$A7,$09,$D4,$AB,$93,$FC,$37,$0F,$3B,$4F,$15,$71,$C3
	.dc.b	$E1,$CE,$B0,$EC,$7A,$C9,$5E,$28,$D1,$D2,$FB,$9E,$09,$E3,$3B,$3B
	.dc.b	$BC,$9F,$11,$F0,$5C,$47,$19,$EB,$3C,$0F,$87,$5E,$66,$1F,$C4,$67
	.dc.b	$C3,$87,$C2,$BF,$0C,$F6,$C3,$DC,$FF,$1B,$E4,$FE,$AB,$F3,$1F,$CC
	.dc.b	$36,$FE,$D7,$F2,$16,$BF,$B1,$B6,$B7,$1F,$D4,$2B,$1C,$8A,$A6,$23
	.dc.b	$3F,$2F,$2B,$FA,$BB,$A6,$90,$FE,$DE,$E5,$FE,$E3,$E0,$F6,$23,$FD
	.dc.b	$83,$C1,$AE,$8F,$EC,$71,$E4,$8F,$2F,$CB,$C9,$FE,$CE,$FA,$3E,$2F
	.dc.b	$11,$FD,$FA,$2F,$0F,$F6,$79,$F6,$FC,$4C,$AF,$ED,$EF,$C0,$F8,$CC
	.dc.b	$47,$F8,$1A,$2D,$7F,$6F,$9F,$95,$2F,$2B,$FB,$8B,$FE,$2A,$3F,$C8
	.dc.b	$8E,$1C,$35,$FF,$17,$B5,$FF,$24,$BB,$9F,$F0,$9C,$3E,$E9,$97,$81
	.dc.b	$B1,$FE,$82,$AE,$1C,$3F,$E5,$2F,$BE,$7A,$BE,$3A,$FF,$2B,$46,$1C
	.dc.b	$35,$F4,$35,$FF,$21,$B9,$C7,$C7,$CD,$E0,$9F,$EE,$97,$FE,$60,$8C
	.dc.b	$4F,$A2,$B8,$4F,$F9,$93,$8D,$C1,$F2,$1B,$87,$FA,$69,$AF,$43,$FC
	.dc.b	$D8,$F4,$9E,$53,$C4,$93,$C7,$4A,$F1,$3F,$C6,$B1,$8C,$3E,$5D,$F3
	.dc.b	$9F,$F3,$C6,$CF,$67,$E3,$42,$30,$FC,$96,$BF,$75,$AF,$F9,$A1,$CD
	.dc.b	$7D,$8C,$1E,$D7,$E1,$9B,$5F,$EB,$8A,$DE,$5F,$E5,$EF,$8C,$FD,$7F
	.dc.b	$C6,$CA,$FF,$20,$C7,$EC,$8C,$3F,$E5,$88,$C3,$FE,$C2,$F8,$EF,$93
	.dc.b	$97,$E2,$1B,$7F,$E9,$46,$CF,$FA,$5D,$F8,$49,$CB,$75,$FC,$EC,$60
	.dc.b	$D9,$5C,$9A,$CB,$F8,$57,$AA,$F8,$EA,$4B,$FF,$2D,$E1,$5F,$BD,$D6
	.dc.b	$AE,$11,$FE,$6D,$6B,$FE,$76,$9F,$D7,$FC,$AF,$F1,$38,$FC,$E3,$5F
	.dc.b	$D8,$35,$FF,$4D,$B5,$FF,$5A,$BB,$CF,$F5,$BF,$24,$F5,$FF,$76,$DA
	.dc.b	$FF,$76,$D6,$EF,$FD,$7D,$8F,$9F,$D9,$F7,$92,$BD,$1F,$21,$5F,$6E
	.dc.b	$69,$C6,$7F,$E8,$18,$56,$30,$EB,$1B,$BF,$85,$7F,$32,$BE,$6F,$C7
	.dc.b	$FC,$EF,$90,$95,$F1,$E3,$2F,$C9,$8C,$BE,$28,$FA,$BE,$AF,$D6,$F9
	.dc.b	$09,$3F,$3E,$32,$FD,$18,$CB,$EA,$8F,$8B,$F2,$92,$3F,$D5,$83,$74
	.dc.b	$7C,$AE,$8D,$4F,$EC,$E7,$2F,$06,$B9,$8D,$ED,$F5,$2A,$ED,$FE,$EC
	.dc.b	$BF,$D4,$DF,$CC,$F9,$5F,$C7,$91,$FC,$47,$C9,$7E,$4B,$E3,$E4,$7E
	.dc.b	$2D,$FC,$AF,$C8,$FE,$5B,$E4,$3F,$B0,$75,$3C,$9A,$66,$FF,$25,$F1
	.dc.b	$FA,$F2,$E4,$60,$FC,$B7,$C7,$7E,$80,$8F,$E6,$7C,$EF,$D1,$FC,$74
	.dc.b	$8F,$CE,$3F,$A6,$F8,$FF,$D3,$FC,$7C,$9F,$D0,$3F,$AA,$F4,$FF,$BA
	.dc.b	$7C,$B3,$E1,$29,$2B,$F4,$C6,$B9,$B7,$D1,$23,$FD,$AD,$5F,$3B,$A3
	.dc.b	$FD,$A0,$BE,$5B,$AF,$95,$2B,$71,$E0,$F8,$B9,$1C,$9F,$1F,$C8,$F9
	.dc.b	$3E,$23,$FA,$67,$8A,$C5,$18,$6D,$FF,$63,$9D,$4E,$7E,$43,$DB,$F2
	.dc.b	$12,$BC,$F1,$F0,$3D,$39,$1E,$A7,$4F,$F0,$8C,$1E,$9F,$DA,$F6,$64
	.dc.b	$FC,$08,$F9,$3E,$FF,$CA,$F5,$A4,$FC,$48,$AF,$38,$BC,$3F,$1A,$6F
	.dc.b	$8B,$FC,$03,$17,$C7,$FD,$43,$38,$37,$F8,$5F,$1B,$F6,$09,$1F,$4B
	.dc.b	$D0,$AF,$A6,$2E,$5F,$52,$6E,$3D,$E6,$D7,$5C,$7F,$8A,$3F,$07,$E0
	.dc.b	$3F,$0A,$47,$EF,$FC,$05,$74,$7E,$0A,$3F,$17,$1F,$E3,$FD,$9F,$0F
	.dc.b	$7E,$26,$58,$FE,$6B,$3A,$71,$A3,$3F,$E4,$73,$99,$DD,$72,$78,$DF
	.dc.b	$7F,$87,$E9,$1F,$ED,$BD,$78,$C6,$3E,$13,$F9,$86,$FF,$72,$31,$3D
	.dc.b	$BE,$26,$DF,$CF,$C9,$58,$BE,$13,$E9,$C3,$9D,$6A,$31,$87,$36,$9E
	.dc.b	$FF,$95,$7C,$17,$FA,$E6,$3C,$C6,$5C,$5F,$D2,$9D,$F0,$72,$A4,$7C
	.dc.b	$39,$D5,$70,$BF,$83,$3C,$2B,$A9,$C2,$74,$F8,$F4,$F7,$B9,$C3,$B5
	.dc.b	$71,$C3,$AC,$62,$B7,$39,$C3,$BE,$1D,$30,$E4,$78,$27,$CD,$3A,$BE
	.dc.b	$B5,$B9,$E5,$5E,$33,$33,$FE,$4B,$8F,$54,$7D,$C3,$A9,$5C,$BF,$EE
	.dc.b	$0E,$77,$CB,$7F,$94,$FE,$1F,$FC,$AA,$D8,$CC,$FF,$9A,$67,$2F,$F9
	.dc.b	$AE,$BF,$C6,$9E,$2F,$87,$E0,$4D,$24,$EE,$32,$FB,$DB,$AF,$EA,$BC
	.dc.b	$87,$A3,$F9,$BA,$D5,$FC,$EB,$DE,$DF,$61,$B3,$FA,$5B,$E3,$CB,$BB
	.dc.b	$93,$C8,$FB,$B3,$A7,$18,$FE,$26,$F9,$C7,$E2,$1B,$3B,$9D,$57,$46
	.dc.b	$D7,$4C,$7D,$11,$83,$F5,$3E,$7F,$FA,$1B,$1F,$CE,$F1,$FF,$8F,$7D
	.dc.b	$ED,$E9,$FE,$1A,$AF,$F3,$9D,$69,$E1,$8F,$E8,$2F,$FC,$85,$39,$BE
	.dc.b	$33,$FC,$7E,$B8,$BB,$78,$E1,$E2,$E2,$F9,$9D,$1F,$37,$1F,$F2,$77
	.dc.b	$C2,$E3,$2F,$15,$5C,$7A,$4D,$AF,$F2,$A4,$DC,$7F,$6B,$7F,$C4,$8F
	.dc.b	$57,$C3,$66,$75,$F0,$52,$75,$8F,$9F,$87,$0F,$F9,$43,$05,$63,$3F
	.dc.b	$34,$7F,$A0,$6F,$8E,$5F,$E6,$5A,$F1,$44,$33,$2F,$83,$C6,$1D,$66
	.dc.b	$E1,$D2,$6C,$7D,$12,$CF,$CB,$8B,$83,$E5,$BF,$C8,$B5,$E2,$7F,$CD
	.dc.b	$8D,$F7,$FE,$34,$62,$FF,$CD,$6F,$06,$C6,$BD,$FD,$B6,$BA,$47,$2F
	.dc.b	$80,$C5,$72,$F8,$0E,$A7,$6F,$80,$B7,$E3,$70,$F9,$7E,$02,$D7,$8B
	.dc.b	$DE,$F2,$C7,$99,$B3,$EB,$F7,$BD,$27,$AB,$DF,$B1,$EF,$F9,$6F,$69
	.dc.b	$EE,$EB,$B1,$F0,$FE,$5B,$E0,$FC,$0F,$C2,$6C,$7C,$7F,$77,$E2,$FC
	.dc.b	$07,$C6,$6C,$7C,$BF,$7B,$E4,$9F,$2B,$DF,$B1,$F3,$FD,$FF,$9A,$7C
	.dc.b	$EF,$12,$C7,$D3,$E1,$FA,$27,$D2,$F1,$2C,$7E,$23,$C3,$FF,$60,$B7
	.dc.b	$DD,$FC,$E4,$7F,$8E,$E3,$06,$FD,$DC,$9A,$F7,$AC,$73,$F7,$38,$9C
	.dc.b	$BD,$EB,$1D,$FE,$5B,$A9,$DB,$DF,$B1,$E3,$F7,$7C,$27,$8B,$DF,$B1
	.dc.b	$FA,$6F,$15,$F2,$9F,$E0,$48,$C3,$78,$3F,$DE,$93,$AF,$63,$F7,$0F
	.dc.b	$0F,$56,$D6,$E3,$F8,$0B,$6F,$FC,$12,$AB,$75,$27,$BB,$CA,$31,$7F
	.dc.b	$EC,$6E,$07,$A7,$F2,$BE,$03,$E5,$B7,$0F,$A0,$DA,$E0,$DC,$BF,$89
	.dc.b	$5B,$E5,$BE,$9F,$B1,$F5,$1B,$87,$D9,$6E,$1E,$16,$DF,$2F,$7B,$ED
	.dc.b	$9F,$73,$DF,$B7,$F8,$18,$B7,$E9,$9B,$5C,$6B,$47,$E1,$E3,$2F,$F8
	.dc.b	$9B,$E5,$BF,$D9,$0F,$F6,$DA,$FF,$1A,$46,$1C,$35,$FE,$E4,$D7,$97
	.dc.b	$95,$FC,$B5,$47,$1B,$FA,$BE,$EE,$63,$17,$FE,$DE,$D3,$FD,$C8,$FE
	.dc.b	$6F,$D8,$FC,$CF,$C0,$FE,$69,$91,$93,$F4,$11,$8C,$7F,$0F,$C6,$F9
	.dc.b	$2F,$C2,$9F,$86,$F7,$ED,$FE,$3A,$2D,$70,$9B,$A7,$E2,$A6,$D7,$E2
	.dc.b	$A6,$D7,$F9,$C7,$27,$C0,$7E,$BB,$E0,$BF,$5E,$DB,$FD,$A3,$6B,$F1
	.dc.b	$8D,$C7,$FC,$05,$0F,$96,$FF,$4D,$D7,$D1,$8C,$74,$E9,$FF,$2D,$CE
	.dc.b	$9F,$AC,$D7,$FC,$48,$E6,$BA,$B6,$BF,$47,$F0,$FF,$56,$31,$4A,$FD
	.dc.b	$9F,$C6,$7E,$B0,$FD,$6E,$77,$0F,$DA,$C5,$AF,$9F,$36,$BF,$63,$36
	.dc.b	$BF,$63,$36,$BF,$CF,$DF,$2D,$FC,$4D,$FC,$73,$5F,$F1,$E3,$5F,$F3
	.dc.b	$CB,$5E,$3F,$11,$F8,$D8,$E5,$87,$F8,$03,$EF,$FF,$AA,$3F,$32,$33
	.dc.b	$87,$FA,$0B,$E5,$DE,$B7,$CB,$C7,$FD,$21,$18,$9F,$A0,$D8,$FD,$13
	.dc.b	$5F,$F2,$83,$9B,$FB,$2D,$9F,$CD,$7C,$47,$E8,$63,$15,$FE,$B7,$F9
	.dc.b	$0F,$E2,$55,$FD,$F9,$B1,$F4,$63,$FD,$45,$EF,$25,$F2,$FF,$E1,$57
	.dc.b	$78,$FF,$87,$5B,$5C,$1B,$7D,$EB,$87,$87,$C7,$B2,$5F,$87,$C4,$B7
	.dc.b	$E1,$3C,$D3,$8C,$7B,$4D,$AE,$DD,$3E,$33,$C9,$D3,$63,$D9,$17,$0E
	.dc.b	$9A,$7A,$EB,$18,$7A,$7E,$5F,$D8,$7B,$7E,$5A,$E3,$F1,$62,$E1,$CB
	.dc.b	$65,$62,$FE,$17,$CB,$F1,$C3,$E6,$36,$F8,$36,$77,$3F,$46,$BF,$8B
	.dc.b	$57,$FA,$E4,$7C,$BB,$D6,$FF,$64,$DB,$C4,$FF,$3C,$4E,$6B,$FD,$25
	.dc.b	$F1,$7F,$ED,$98,$C4,$7E,$AE,$77,$7F,$CA,$EB,$4F,$48,$CC,$F2,$3C
	.dc.b	$31,$9C,$77,$19,$BE,$B8,$78,$E7,$33,$FE,$DB,$38,$9F,$EE,$8B,$EA
	.dc.b	$E6,$B8,$C6,$23,$E0,$46,$27,$9C,$62,$FC,$4F,$43,$93,$89,$F7,$C6
	.dc.b	$CF,$04,$F2,$58,$79,$B0,$F4,$47,$33,$81,$F1,$E7,$84,$EF,$0F,$34
	.dc.b	$F2,$3C,$55,$CA,$FC,$31,$97,$D4,$F5,$31,$87,$9C,$FA,$4F,$2A,$EB
	.dc.b	$7D,$F0,$F0,$CE,$5B,$7D,$0E,$F3,$B9,$EC,$D8,$CC,$71,$32,$57,$27
	.dc.b	$3B,$F1,$B6,$3E,$F4,$79,$DF,$C2,$72,$FF,$4E,$6F,$7F,$AA,$31,$1F
	.dc.b	$81,$6F,$C6,$BC,$9F,$E7,$92,$9C,$C6,$1B,$A2,$75,$75,$5F,$39,$D5
	.dc.b	$70,$9F,$C2,$60,$F5,$B8,$65,$FA,$28,$C4,$FE,$5B,$0E,$4F,$E2,$DB
	.dc.b	$3B,$AE,$35,$E1,$38,$D6,$F2,$F5,$D7,$EB,$AF,$8D,$6E,$F8,$DF,$AE
	.dc.b	$38,$D6,$EF,$8C,$7B,$6F,$8D,$6E,$F8,$BF,$C0,$5B,$1A,$8F,$87,$5C
	.dc.b	$9E,$07,$EC,$AB,$B9,$D6,$35,$1C,$31,$D5,$F6,$AD,$CF,$D1,$BE,$C7
	.dc.b	$78,$CC,$78,$CF,$95,$7D,$27,$55,$CC,$F3,$C6,$4F,$0D,$72,$39,$CF
	.dc.b	$68,$CC,$F1,$AE,$B7,$1F,$BB,$19,$7E,$94,$67,$0F,$B1,$1E,$C3,$18
	.dc.b	$7C,$19,$C4,$72,$32,$7B,$EF,$C2,$EA,$BE,$B4,$FF,$20,$E3,$FE,$C5
	.dc.b	$E0,$FF,$32,$DB,$FA,$F7,$C8,$DC,$3F,$B6,$36,$B7,$1F,$DB,$6B,$1C
	.dc.b	$6A,$A6,$23,$3F,$23,$2B,$FB,$A3,$A6,$90,$FE,$A4,$E5,$FE,$A9,$E1
	.dc.b	$76,$23,$FD,$F7,$C2,$AE,$8F,$F0,$04,$79,$23,$CB,$F1,$B2,$7F,$C0
	.dc.b	$8F,$A3,$E3,$B1,$1F,$D8,$62,$EB,$FE,$01,$9F,$6F,$C6,$CA,$FF,$06
	.dc.b	$3F,$03,$E3,$B1,$1F,$DA,$62,$D7,$F8,$26,$7E,$54,$BC,$B7,$66,$1B
	.dc.b	$1F,$C1,$6B,$FC,$AF,$18,$8F,$F0,$40,$F9,$76,$3C,$6F,$FC,$B5,$3D
	.dc.b	$5B,$5C,$63,$F8,$71,$93,$FC,$C1,$F2,$FF,$EB,$07,$C1,$C7,$FC,$4E
	.dc.b	$31,$1E,$06,$E3,$FE,$6A,$E5,$F0,$CF,$03,$F8,$E4,$F0,$7B,$B5,$E9
	.dc.b	$E9,$F9,$A3,$11,$FC,$8F,$91,$C3,$F1,$9B,$1D,$4F,$F1,$B4,$67,$87
	.dc.b	$4C,$8D,$4F,$07,$E3,$A9,$D4,$62,$6E,$3F,$E8,$4A,$D4,$F7,$E0,$7B
	.dc.b	$7F,$61,$B1,$98,$FB,$78,$62,$3E,$AC,$D7,$E7,$B5,$FE,$7C,$D7,$FA
	.dc.b	$13,$79,$BF,$CA,$5F,$1D,$FA,$AF,$90,$95,$FE,$59,$8F,$D6,$98,$7F
	.dc.b	$C6,$D1,$87,$E3,$7C,$87,$C1,$CF,$FD,$3E,$DE,$0F,$E9,$CD,$9F,$EA
	.dc.b	$4F,$C0,$7F,$1A,$E5,$34,$99,$7E,$69,$C6,$B4,$C4,$7F,$05,$F9,$57
	.dc.b	$CB,$D2,$5F,$F9,$CF,$0A,$FF,$96,$27,$6F,$E2,$F4,$46,$23,$FC,$A9
	.dc.b	$3F,$AA,$F9,$7F,$E1,$D1,$F8,$56,$BF,$A6,$6E,$7F,$EC,$0F,$97,$FE
	.dc.b	$F2,$FF,$B0,$A3,$73,$FA,$46,$BF,$DC,$DA,$FF,$77,$6E,$1E,$4F,$67
	.dc.b	$DC,$49,$F4,$62,$F9,$3C,$47,$B6,$6E,$0C,$BF,$D2,$37,$BA,$E3,$5A
	.dc.b	$9C,$BF,$2E,$FE,$67,$C9,$7C,$DF,$93,$97,$E0,$8C,$BE,$28,$CB,$F3
	.dc.b	$63,$EA,$7A,$7F,$57,$E4,$E4,$FA,$23,$2F,$AA,$32,$FD,$58,$FE,$33
	.dc.b	$B3,$E1,$A9,$27,$F6,$D0,$F8,$7C,$47,$F1,$88,$E0,$EA,$79,$CE,$B6
	.dc.b	$FA,$D1,$75,$FF,$12,$5F,$3B,$F9,$BF,$23,$F8,$D2,$3F,$43,$E4,$FF
	.dc.b	$21,$F2,$32,$3E,$9B,$F9,$3F,$93,$FC,$A7,$C9,$49,$FA,$EF,$E5,$FE
	.dc.b	$4F,$F3,$1F,$25,$27,$ED,$BF,$9B,$F9,$3F,$CE,$7C,$94,$9F,$BE,$FE
	.dc.b	$7F,$D0,$FD,$07,$C9,$49,$FC,$1B,$FA,$3F,$93,$FD,$27,$C9,$C9,$FC
	.dc.b	$3B,$FA,$7F,$53,$FB,$67,$CB,$BE,$1A,$92,$BF,$1A,$6B,$97,$7C,$F2
	.dc.b	$3F,$51,$5F,$39,$FF,$61,$AF,$93,$EB,$E4,$C9,$DC,$78,$3E,$4A,$47
	.dc.b	$27,$C7,$F2,$9E,$4F,$8B,$93,$D9,$F3,$FC,$A7,$A3,$E5,$3D,$3F,$29
	.dc.b	$2B,$C7,$1E,$CF,$4A,$47,$99,$D3,$E6,$F4,$FE,$09,$1D,$1E,$CF,$82
	.dc.b	$F8,$A4,$7D,$FF,$29,$F2,$3D,$69,$1F,$09,$D3,$F0,$9F,$9B,$F2,$92
	.dc.b	$34,$7C,$8F,$91,$FA,$24,$7E,$5F,$C9,$D7,$E5,$C5,$F1,$FF,$75,$0F
	.dc.b	$3D,$CA,$BD,$79,$1F,$E7,$F9,$CE,$18,$BE,$17,$B7,$EC,$FC,$27,$E0
	.dc.b	$88,$FD,$CF,$85,$AF,$DC,$8B,$97,$8B,$0F,$0D,$7F,$8D,$77,$7C,$5D
	.dc.b	$26,$1F,$CB,$61,$F1,$D8,$86,$1C,$E5,$F1,$DC,$CD,$46,$7E,$3B,$FD
	.dc.b	$31,$F1,$5E,$9C,$FF,$D4,$5F,$1B,$8E,$1F,$81,$8A,$31,$97,$C7,$86
	.dc.b	$18,$BE,$33,$FE,$A5,$C3,$A5,$71,$8C,$61,$D3,$0E,$13,$CA,$72,$7F
	.dc.b	$AB,$7E,$59,$F0,$DF,$EB,$38,$FF,$72,$DF,$FA,$BE,$73,$C1,$C2,$51
	.dc.b	$87,$67,$3D,$FD,$2E,$AF,$A6,$5F,$66,$31,$7C,$EF,$94,$E5,$FA,$73
	.dc.b	$9A,$E3,$87,$8D,$B1,$E9,$6B,$E4,$C3,$C3,$5D,$2B,$86,$1C,$CF,$04
	.dc.b	$EA,$F8,$D6,$EF,$A5,$69,$E0,$7F,$91,$0F,$49,$63,$E1,$3F,$E4,$86
	.dc.b	$BF,$EB,$BE,$87,$CB,$F1,$8F,$F6,$CC,$61,$E9,$F1,$A8,$C4,$7F,$94
	.dc.b	$A3,$13,$C6,$9F,$06,$F2,$30,$DB,$F2,$35,$F2,$35,$FB,$8E,$F5,$FA
	.dc.b	$0D,$D7,$FD,$DD,$E5,$3D,$2F,$DC,$D6,$B0,$FB,$17,$BC,$3F,$14,$D5
	.dc.b	$7C,$7A,$B7,$52,$7F,$0E,$7E,$2A,$B8,$56,$E3,$39,$7B,$AB,$86,$1E
	.dc.b	$19,$E3,$5B,$8C,$13,$2F,$6D,$7E,$5E,$75,$18,$AF,$CB,$F1,$BE,$E3
	.dc.b	$FA,$56,$7F,$A4,$8C,$79,$5F,$02,$FF,$8A,$7E,$6A,$C3,$07,$F1,$BC
	.dc.b	$38,$BD,$1B,$9F,$EF,$DB,$5C,$A7,$8E,$1D,$B1,$F0,$57,$F8,$1D,$FF
	.dc.b	$1E,$60,$F5,$BF,$94,$F4,$FF,$19,$AD,$61,$FC,$4A,$FC,$B5,$A7,$F9
	.dc.b	$54,$EA,$BF,$90,$5F,$39,$FF,$24,$4E,$5E,$27,$FA,$27,$F0,$3F,$E6
	.dc.b	$8F,$F2,$1F,$93,$E9,$8F,$C1,$9C,$BE,$13,$F8,$B4,$69,$46,$B0,$FE
	.dc.b	$4D,$39,$8C,$57,$F2,$08,$E1,$D7,$E3,$BF,$99,$E5,$E3,$CE,$F0,$EB
	.dc.b	$F5,$E3,$18,$7A,$9B,$5E,$39,$D4,$FF,$A8,$EF,$A4,$FF,$2A,$BE,$79
	.dc.b	$BC,$BE,$3A,$7C,$DF,$33,$26,$BC,$BB,$97,$C8,$6E,$3D,$F0,$46,$2F
	.dc.b	$EB,$79,$9C,$6F,$E7,$B6,$F8,$79,$7E,$23,$C7,$E6,$5A,$FA,$BD,$D8
	.dc.b	$69,$F6,$23,$D3,$E1,$48,$FB,$7E,$5D,$C3,$C9,$E5,$7C,$8A,$FF,$74
	.dc.b	$D7,$E0,$23,$1F,$15,$F0,$BA,$75,$F2,$56,$BF,$B0,$3F,$1B,$CA,$E5
	.dc.b	$F2,$76,$7F,$B2,$3F,$53,$EC,$3F,$B2,$C7,$FA,$4D,$B9,$FC,$5F,$1F
	.dc.b	$5E,$66,$FC,$AF,$27,$C9,$5C,$3F,$B7,$3C,$FC,$C9,$1C,$FB,$2E,$3F
	.dc.b	$77,$2F,$5B,$F7,$FC,$17,$B9,$E3,$E1,$FC,$3C,$63,$0F,$26,$DF,$0F
	.dc.b	$E4,$FD,$0D,$8F,$44,$7C,$7F,$93,$F6,$36,$3D,$9F,$15,$D3,$CD,$F8
	.dc.b	$0D,$8F,$81,$F2,$3F,$3E,$3E,$23,$63,$E2,$7A,$3F,$4F,$E3,$3E,$43
	.dc.b	$63,$FD,$F1,$F6,$3F,$DC,$9F,$95,$E8,$FE,$AE,$31,$C1,$F0,$63,$15
	.dc.b	$F4,$5B,$5C,$0E,$71,$96,$99,$9F,$94,$FD,$9F,$90,$FB,$4D,$3E,$D1
	.dc.b	$63,$6F,$DD,$8F,$D9,$FC,$07,$FA,$E6,$3F,$DA,$4F,$FB,$69,$BF,$15
	.dc.b	$AF,$8E,$DB,$63,$27,$08,$FE,$11,$F3,$6F,$8D,$FE,$1F,$E4,$7F,$B7
	.dc.b	$3C,$BC,$C6,$3E,$39,$18,$C3,$F9,$A5,$7F,$26,$9C,$C6,$0F,$E7,$1A
	.dc.b	$30,$ED,$19,$8F,$F1,$DC,$F0,$7F,$8E,$1B,$C3,$FC,$B2,$F2,$31,$3F
	.dc.b	$DE,$EB,$9C,$7F,$81,$5A,$F8,$3C,$F6,$8F,$8E,$FC,$AF,$C4,$7C,$A6
	.dc.b	$E1,$F2,$9B,$7F,$C6,$EB,$84,$FF,$44,$AF,$A3,$F1,$BF,$4B,$E3,$AC
	.dc.b	$FD,$26,$DF,$16,$EB,$FE,$61,$C1,$1F,$43,$E2,$7E,$7F,$D3,$7F,$71
	.dc.b	$AF,$7E,$A9,$CF,$56,$F2,$F9,$14,$BE,$11,$F3,$AF,$83,$A8,$CE,$5F
	.dc.b	$DA,$67,$2F,$F4,$9A,$E5,$7D,$63,$15,$C3,$6F,$93,$F1,$BF,$90,$F8
	.dc.b	$FF,$C8,$B7,$2F,$C8,$B6,$FE,$43,$71,$FB,$3F,$36,$F3,$BF,$82,$FC
	.dc.b	$7F,$F5,$78,$C6,$1F,$DC,$23,$13,$A8,$FC,$36,$0A,$DB,$FC,$46,$BF
	.dc.b	$B0,$7C,$3B,$E5,$FF,$33,$F3,$8F,$07,$31,$86,$E3,$F2,$A7,$4F,$6A
	.dc.b	$E4,$E6,$77,$87,$C7,$AD,$BF,$96,$AD,$B9,$9F,$EE,$5F,$36,$F8,$CF
	.dc.b	$E4,$7C,$D8,$6E,$3E,$C9,$D3,$FE,$10,$9D,$34,$F8,$3D,$EF,$57,$F9
	.dc.b	$6F,$C9,$7F,$7F,$8C,$4F,$F8,$62,$31,$3A,$8E,$2D,$CB,$F3,$9B,$7F
	.dc.b	$55,$CF,$F8,$5B,$6B,$F8,$5B,$67,$71,$FE,$B3,$9F,$E2,$37,$FC,$FB
	.dc.b	$E4,$AC,$FF,$8A,$1B,$5F,$D1,$5B,$C5,$FE,$22,$7F,$8C,$F4,$FF,$65
	.dc.b	$E1,$B7,$FE,$BF,$6D,$FF,$82,$F3,$F9,$B8,$7E,$A6,$F8,$5F,$F2,$0F
	.dc.b	$92,$FE,$DD,$E2,$7F,$24,$6C,$64,$FF,$65,$46,$5A,$7F,$28,$F8,$8F
	.dc.b	$E5,$67,$F2,$CF,$92,$B1,$92,$65,$FC,$0E,$32,$FF,$16,$3F,$D8,$43
	.dc.b	$E6,$9A,$FF,$20,$CF,$EE,$CE,$63,$FD,$35,$F1,$3F,$E5,$BA,$DC,$FF
	.dc.b	$2A,$E5,$FF,$7B,$C6,$23,$FD,$B2,$71,$35,$8F,$1B,$E7,$19,$3F,$DC
	.dc.b	$B5,$CA,$33,$5C,$F0,$F1,$BD,$6B,$D2,$E6,$B8,$47,$F2,$FC,$3B,$3E
	.dc.b	$03,$B1,$C2,$FD,$D3,$C2,$3C,$F5,$CF,$0E,$31,$8A,$DC,$62,$39,$1D
	.dc.b	$67,$C8,$AF,$8E,$1E,$8B,$F5,$9C,$A7,$C6,$6C,$F8,$51,$B7,$D4,$63
	.dc.b	$0F,$33,$E0,$3E,$3E,$1E,$37,$96,$1E,$97,$AE,$1E,$2A,$E3,$1A,$C3
	.dc.b	$C5,$5E,$8A,$D5,$F3,$6D,$72,$3C,$51,$96,$99,$9F,$01,$CD,$CB,$4C
	.dc.b	$D7,$C2,$8F,$90,$D7,$D6,$77,$F6,$5E,$D7,$E1,$9B,$1F,$11,$AF,$E3
	.dc.b	$9B,$F1,$DF,$C3,$F3,$FF,$58,$46,$38,$51,$8B,$FE,$FB,$5B,$B5,$EE
	.dc.b	$B5,$1F,$EB,$99,$D6,$3F,$9D,$C5,$AF,$F8,$1A,$30,$D8,$EB,$97,$82
	.dc.b	$FF,$3D,$7C,$A7,$FC,$1A,$72,$8F,$C7,$B5,$DC,$F1,$BF,$D6,$57,$19
	.dc.b	$FF,$66,$56,$E7,$C3,$1F,$98,$9D,$26,$CE,$CF,$0D,$F9,$E7,$DD,$7D
	.dc.b	$E7,$D3,$87,$92,$B5,$5F,$C1,$9B,$7D,$8E,$AE,$AB,$B4,$F4,$AF,$25
	.dc.b	$F5,$8F,$4D,$6C,$E7,$19,$8F,$43,$6B,$9C,$78,$4E,$75,$1F,$E1,$51
	.dc.b	$A3,$77,$E5,$AE,$57,$DE,$B8,$61,$F3,$27,$31,$EF,$33,$7F,$09,$EB
	.dc.b	$5C,$5F,$AB,$86,$9F,$59,$C4,$C9,$2F,$EA,$BC,$8E,$1D,$FF,$E3,$E7
	.dc.b	$F9,$86,$85,$19,$7F,$CC,$E6,$8C,$F4,$3E,$33,$13,$B9,$FF,$21,$18
	.dc.b	$9F,$F2,$96,$1F,$E3,$BA,$DF,$D7,$3E,$37,$8E,$AF,$3E,$9A,$89,$8F
	.dc.b	$F9,$87,$2F,$EF,$93,$A8,$C5,$F3,$6F,$C4,$F4,$F9,$89,$18,$21,$FD
	.dc.b	$9D,$CB,$FE,$F3,$94,$F9,$BE,$2F,$DB,$EC,$FF,$93,$6B,$C4,$D7,$84
	.dc.b	$6D,$CF,$5F,$03,$0F,$5F,$47,$DF,$E0,$C8,$F4,$C7,$C4,$30,$FF,$7B
	.dc.b	$8C,$3F,$E3,$CF,$CD,$BE,$5F,$1F,$FA,$1B,$F3,$77,$C2,$FD,$5A,$46
	.dc.b	$1E,$06,$6B,$8B,$F7,$E3,$4F,$FA,$29,$BC,$9F,$E8,$A9,$DC,$BC,$0F
	.dc.b	$E0,$D3,$FE,$8E,$D7,$FB,$36,$3F,$E9,$86,$D7,$E2,$EB,$73,$FE,$9C
	.dc.b	$6C,$7E,$3E,$76,$9B,$1A,$87,$D0,$B8,$7D,$DE,$AB,$CC,$A7,$C1,$9D
	.dc.b	$57,$F9,$89,$AF,$FA,$9A,$78,$3F,$E6,$66,$BF,$EA,$A8,$DE,$5F,$EA
	.dc.b	$A6,$E5,$FE,$B4,$AD,$CF,$FA,$B9,$B1,$FE,$B8,$9D,$D7,$EA,$FE,$F1
	.dc.b	$E1,$FD,$5F,$6D,$F1,$14,$FB,$13,$AA,$FF,$3C,$B5,$FF,$5E,$CF,$07
	.dc.b	$FC,$FA,$D7,$FD,$83,$1B,$C9,$8F,$EE,$FA,$5D,$0F,$8C,$FD,$8F,$C6
	.dc.b	$CC,$7F,$53,$1F,$B4,$30,$FE,$8E,$30,$FF,$A3,$3A,$FF,$0B,$87,$FB
	.dc.b	$5B,$E7,$3F,$D9,$F9,$7F,$B4,$1B,$1F,$ED,$A9,$DD,$7F,$0D,$6A,$9D
	.dc.b	$BF,$C4,$5A,$FF,$03,$E7,$FC,$E4,$F0,$9C,$9A,$7A,$E7,$EE,$F8,$CF
	.dc.b	$E1,$BF,$1D,$30,$FE,$0B,$1F,$C4,$0C,$3F,$BF,$8C,$3F,$C2,$BC,$0F
	.dc.b	$DF,$F3,$C8,$C1,$0F,$EF,$79,$FC,$DE,$8F,$D2,$47,$F0,$49,$FE,$3B
	.dc.b	$5F,$B3,$F9,$C7,$C1,$FF,$BC,$EB,$FD,$FB,$18,$70,$1C,$3B,$6B,$FE
	.dc.b	$FB,$73,$BF,$FB,$FB,$E7,$3E,$59,$F3,$23,$39,$76,$BF,$E0,$93,$AE
	.dc.b	$0F,$24,$7F,$32,$F3,$7F,$99,$F3,$CC,$BF,$95,$C6,$5F,$E5,$B1,$97
	.dc.b	$F9,$34,$7F,$3A,$F9,$AF,$E7,$7E,$C4,$9F,$E6,$71,$97,$F9,$AC,$65
	.dc.b	$FE,$5D,$1F,$17,$8F,$CB,$F2,$72,$7F,$DB,$3A,$FF,$43,$6C,$E5,$FA
	.dc.b	$27,$F5,$7D,$15,$CE,$36,$7F,$B8,$A3,$2D,$8D,$3F,$CF,$FD,$7F,$EA
	.dc.b	$3B,$48,$FE,$97,$1F,$D5,$BC,$FF,$E8,$3F,$07,$27,$FA,$7C,$7F,$60
	.dc.b	$FC,$E1,$F3,$5F,$D1,$BC,$09,$5F,$D5,$E7,$2D,$F8,$EF,$E9,$7F,$0F
	.dc.b	$23,$06,$FE,$5B,$FA,$7F,$C4,$48,$C1,$CF,$E5,$3F,$A9,$FA,$B2,$30
	.dc.b	$77,$F9,$2F,$21,$1F,$17,$C7,$79,$FE,$3B,$FA,$DF,$C6,$49,$F3,$47
	.dc.b	$AF,$E3,$A4,$7A,$5F,$77,$97,$EF,$F8,$0F,$EC,$7F,$17,$2B,$DD,$1F
	.dc.b	$0F,$AB,$FB,$3F,$C6,$49,$F8,$51,$F1,$FD,$5F,$ED,$7F,$13,$27,$E3
	.dc.b	$47,$F8,$2B,$E9,$1D,$54,$93,$F2,$FB,$35,$CD,$BE,$09,$3F,$E0,$5B
	.dc.b	$4F,$3E,$59,$39,$76,$FE,$2B,$CC,$53,$7F,$16,$D2,$BA,$45,$78,$4B
	.dc.b	$5F,$41,$9F,$1F,$F9,$8C,$FF,$8A,$A7,$4F,$96,$76,$F9,$EF,$FB,$C7
	.dc.b	$E7,$3F,$E2,$B8,$FF,$19,$5F,$25,$19,$7F,$42,$63,$DE,$F9,$1E,$B7
	.dc.b	$27,$E5,$7C,$D4,$9E,$7C,$1F,$E3,$D7,$81,$83,$FC,$81,$7D,$70,$F4
	.dc.b	$EB,$F1,$2F,$95,$78,$2B,$57,$CF,$E6,$9F,$33,$FE,$26,$CE,$5F,$51
	.dc.b	$3A,$70,$D7,$E9,$7A,$1F,$DA,$9F,$F2,$8B,$5E,$35,$F8,$5F,$92,$FF
	.dc.b	$19,$EB,$27,$F0,$11,$97,$FC,$6B,$19,$7E,$BC,$7E,$37,$E0,$FE,$17
	.dc.b	$C4,$7F,$89,$76,$FC,$0C,$63,$D1,$FE,$D1,$C7,$FE,$61,$AD,$2F,$8C
	.dc.b	$D6,$3F,$76,$EE,$1C,$A3,$FC,$D5,$7C,$27,$37,$C6,$73,$87,$D3,$F9
	.dc.b	$B7,$C5,$FF,$B0,$A7,$FD,$F5,$39,$8C,$61,$FE,$36,$AD,$F9,$7C,$AF
	.dc.b	$FC,$F1,$0F,$99,$F2,$C2,$36,$F4,$37,$7D,$30,$ED,$7C,$E7,$85,$F2
	.dc.b	$BE,$38,$78,$70,$E1,$5C,$AF,$85,$78,$62,$CE,$9F,$13,$63,$9E,$1F
	.dc.b	$07,$0E,$D3,$92,$5F,$48,$CB,$FD,$9A,$BD,$77,$F9,$B3,$AC,$66,$BE
	.dc.b	$54,$67,$E2,$FF,$2B,$F3,$9F,$DB,$9B,$1F,$DC,$DA,$EF,$85,$3D,$FC
	.dc.b	$1F,$D8,$4F,$D7,$6B,$F4,$1A,$FC,$F6,$E3,$FE,$9B,$E2,$F8,$B1,$FB
	.dc.b	$DD,$7F,$CD,$FE,$6B,$E5,$3F,$83,$C7,$DD,$9D,$6F,$DE,$F8,$9F,$56
	.dc.b	$F8,$74,$63,$2F,$B8,$78,$67,$18,$76,$C7,$AE,$1C,$EB,$79,$79,$5B
	.dc.b	$97,$81,$FF,$61,$65,$D9,$CE,$3D,$2F,$96,$3D,$72,$C5,$7F,$B2,$3E
	.dc.b	$73,$FD,$79,$1F,$95,$E4,$46,$27,$F2,$51,$FE,$7F,$7C,$A7,$1F,$8C
	.dc.b	$B1,$C1,$B3,$C6,$FF,$71,$D1,$FE,$D7,$C3,$B4,$E7,$1F,$F3,$03,$FE
	.dc.b	$BA,$8F,$F5,$BF,$67,$F3,$4E,$DF,$E3,$B3,$9C,$FF,$88,$CE,$6B,$8F
	.dc.b	$C6,$59,$F6,$37,$4F,$5F,$CF,$DC,$39,$63,$FC,$DF,$B7,$FD,$67,$8F
	.dc.b	$F8,$F1,$B8,$FF,$90,$5A,$FE,$65,$AF,$F2,$BC,$FC,$0F,$F2,$87,$B7
	.dc.b	$CF,$E2,$79,$DF,$F4,$5F,$4B,$C3,$87,$C0,$F8,$EE,$31,$8C,$7E,$A4
	.dc.b	$63,$5F,$E9,$91,$FD,$0E,$7E,$13,$FA,$B6,$B9,$8D,$A1,$1D,$5B,$1F
	.dc.b	$0F,$D7,$ED,$7F,$EF,$7D,$1F,$1B,$F1,$CF,$91,$F0,$B6,$74,$F6,$E0
	.dc.b	$FC,$3B,$68,$D3,$1F,$0D,$CA,$BB,$37,$2F,$EF,$94,$F8,$5F,$AD,$F0
	.dc.b	$BF,$5D,$B7,$E5,$EE,$EC,$47,$FC,$ED,$97,$E6,$23,$C5,$F0,$5F,$7F
	.dc.b	$E1,$3F,$00,$DA,$F7,$7C,$17,$E0,$CF,$C2,$7C,$1D,$8F,$85,$CF,$EB
	.dc.b	$23,$DB,$4F,$C0,$EB,$FA,$9F,$91,$FC,$61,$F8,$DF,$81,$B5,$FD,$DF
	.dc.b	$57,$CE,$3E,$7B,$F2,$8D,$9F,$A1,$E4,$E2,$73,$E5,$7C,$DA,$FE,$8D
	.dc.b	$5E,$E6,$E5,$FC,$AF,$0E,$B5,$BF,$9E,$EA,$76,$EA,$B8,$78,$F0,$7C
	.dc.b	$EF,$E9,$7C,$27,$CF,$79,$23,$0F,$9B,$E0,$3C,$44,$BF,$13,$E5,$F8
	.dc.b	$AF,$59,$EC,$F8,$2B,$3D,$FE,$03,$B7,$9D,$F0,$1B,$1E,$3F,$98,$F7
	.dc.b	$11,$F7,$3F,$07,$D2,$F8,$E7,$C8,$F8,$1B,$3F,$E1,$7F,$9E,$FE,$E6
	.dc.b	$FB,$7D,$1F,$D9,$C6,$39,$7F,$C9,$1C,$29,$CE,$1F,$A6,$95,$6D,$FE
	.dc.b	$AC,$D7,$FC,$64,$D7,$F5,$8F,$5F,$27,$FC,$80,$DC,$3F,$DF,$58,$3E
	.dc.b	$4F,$BC,$7D,$13,$F0,$3F,$07,$F8,$26,$99,$3E,$B3,$96,$9F,$8B,$F6
	.dc.b	$3F,$CC,$ED,$F8,$3F,$E3,$D1,$8C,$BF,$CC,$CF,$F9,$D2,$78,$7C,$16
	.dc.b	$AB,$6D,$8F,$F2,$07,$CF,$3E,$0F,$F9,$56,$AF,$94,$FE,$07,$5A,$9C
	.dc.b	$C6,$32,$FF,$2E,$DF,$1A,$FF,$72,$B6,$3F,$DD,$4D,$7F,$C0,$4D,$7F
	.dc.b	$CB,$CE,$EF,$FC,$D4,$D9,$FF,$36,$FC,$E7,$F1,$A7,$C4,$7C,$3F,$86
	.dc.b	$FC,$D7,$C0,$FC,$26,$C6,$49,$87,$C4,$CB,$51,$F2,$FE,$13,$E4,$FC
	.dc.b	$37,$CA,$6C,$64,$FA,$11,$96,$9F,$23,$E6,$3E,$3F,$CF,$7F,$71,$D7
	.dc.b	$D3,$18,$EB,$FE,$8B,$B2,$3F,$20,$DB,$F8,$F1,$82,$3D,$4F,$F4,$05
	.dc.b	$3E,$A9,$EA,$74,$AF,$C0,$FC,$3F,$E0,$BE,$26,$E1,$F8,$76,$EF,$F8
	.dc.b	$46,$E3,$FE,$88,$E6,$FF,$52,$57,$FA,$07,$E7,$DE,$97,$F0,$5F,$89
	.dc.b	$FE,$BB,$18,$9F,$E3,$9E,$47,$F9,$B5,$E1,$5A,$3F,$27,$F0,$36,$3F
	.dc.b	$D1,$9E,$83,$C3,$FC,$6F,$53,$C4,$FE,$D3,$39,$AF,$7C,$E5,$E3,$5F
	.dc.b	$B7,$9D,$6F,$FE,$4B,$0E,$1F,$F2,$73,$5F,$F5,$53,$C3,$6F,$F5,$B3
	.dc.b	$6B,$F1,$B5,$E2,$3F,$BD,$FC,$FB,$D5,$FE,$4D,$C1,$86,$D7,$FA,$9E
	.dc.b	$74,$D8,$FD,$9F,$97,$B8,$FF,$56,$78,$4F,$5F,$F9,$87,$CF,$FF,$82
	.dc.b	$23,$13,$F3,$23,$13,$23,$F8,$07,$5D,$DB,$F7,$BC,$BF,$DC,$74,$FA
	.dc.b	$51,$FD,$0D,$AF,$FB,$27,$D2,$46,$23,$FC,$09,$18,$9F,$E1,$27,$F0
	.dc.b	$AF,$0A,$E1,$FC,$3F,$2F,$E8,$DF,$3F,$FC,$39,$B1,$92,$61,$FC,$43
	.dc.b	$1D,$47,$F1,$7E,$BF,$31,$1F,$EE,$F1,$FA,$19,$F4,$4F,$F1,$A3,$F8
	.dc.b	$DD,$DA,$FE,$41,$F3,$1E,$F2,$3C,$9B,$5F,$C8,$1F,$E4,$7F,$13,$FC
	.dc.b	$91,$B3,$FC,$A6,$3F,$93,$FC,$8E,$2B,$F9,$3B,$79,$BF,$48,$DA,$FC
	.dc.b	$77,$C4,$7F,$8A,$FD,$A6,$5F,$EF,$5B,$FC,$0E,$09,$CB,$FE,$9F,$AE
	.dc.b	$15,$BE,$56,$18,$AD,$0F,$9F,$E2,$B1,$EB,$18,$8F,$F6,$B9,$BB,$E9
	.dc.b	$5F,$E3,$F9,$D4,$B0,$C5,$6A,$38,$1E,$09,$EF,$7D,$A7,$DF,$3E,$B7
	.dc.b	$86,$3D,$2F,$D1,$3E,$5C,$7D,$6E,$CE,$95,$DF,$0E,$11,$E0,$38,$63
	.dc.b	$E4,$76,$7F,$95,$F1,$F1,$38,$36,$71,$3E,$0E,$3E,$8C,$3C,$2F,$63
	.dc.b	$15,$E6,$9F,$8F,$78,$79,$1F,$14,$F2,$5F,$81,$FF,$34,$19,$8E,$07
	.dc.b	$A8,$B8,$7C,$66,$C7,$9E,$79,$3E,$69,$F9,$27,$CC,$BF,$11,$CD,$F8
	.dc.b	$C7,$32,$C7,$E1,$23,$71,$C4,$E7,$19,$6B,$E8,$C7,$EA,$D7,$E5,$2B
	.dc.b	$E8,$FB,$7F,$95,$AF,$C7,$B5,$E0,$1C,$3F,$8E,$75,$F1,$DF,$C6,$6B
	.dc.b	$FD,$19,$18,$E6,$FC,$B5,$71,$BF,$93,$7B,$8C,$1F,$E9,$37,$15,$C6
	.dc.b	$FF,$17,$8B,$5F,$A3,$87,$E8,$A3,$18,$7E,$43,$0E,$93,$FA,$66,$CF
	.dc.b	$D8,$37,$3F,$93,$3F,$6F,$5D,$23,$CB,$8F,$A2,$F9,$3E,$AC,$FC,$15
	.dc.b	$F8,$26,$DF,$E5,$C6,$1E,$D8,$E3,$8F,$AA,$BB,$57,$FA,$C7,$1F,$1E
	.dc.b	$3A,$8F,$26,$1C,$67,$D1,$5D,$DB,$3D,$E7,$4F,$C9,$C7,$CD,$3A,$8F
	.dc.b	$D6,$3F,$03,$1E,$C4,$74,$FA,$A3,$F6,$55,$D1,$F5,$D7,$C3,$AE,$98
	.dc.b	$6C,$F4,$1C,$E7,$D9,$87,$9D,$A7,$9E,$BC,$AF,$9F,$1F,$91,$7E,$53
	.dc.b	$C6,$7D,$27,$81,$F2,$62,$1F,$1A,$7C,$8F,$33,$DF,$5E,$17,$EE,$9C
	.dc.b	$8F,$6C,$78,$9C,$9C,$0D,$63,$EC,$BF,$79,$23,$B6,$1F,$CE,$B9,$7F
	.dc.b	$B4,$FC,$67,$F3,$FD,$BF,$B4,$F1,$FF,$06,$BF,$F3,$79,$91,$5C,$1F
	.dc.b	$F3,$C1,$97,$FB,$9B,$63,$96,$9F,$E3,$3B,$D7,$CD,$BE,$36,$92,$F9
	.dc.b	$9F,$EC,$D8,$D3,$FE,$E2,$6B,$FE,$C5,$6B,$FE,$03,$6B,$FE,$02,$5F
	.dc.b	$13,$E0,$24,$75,$F8,$DA,$F3,$8F,$F1,$83,$EE,$9E,$2F,$86,$33,$92
	.dc.b	$5F,$17,$EC,$25,$7A,$7E,$36,$BF,$D7,$F5,$FF,$0A,$F8,$FF,$E6,$78
	.dc.b	$FF,$27,$F5,$6B,$6F,$E9,$F9,$75,$F8,$BF,$8E,$4A,$F8,$9F,$19,$5F
	.dc.b	$85,$1E,$4D,$0D,$3D,$70,$F9,$EF,$A7,$9F,$F4,$26,$C0,$D1,$F7,$C7
	.dc.b	$CF,$FF,$2B,$CB,$FB,$1C,$68,$FE,$B9,$5F,$D3,$25,$F1,$3C,$5B,$3A
	.dc.b	$3B,$5F,$81,$77,$FF,$1C,$C3,$AB,$70,$FE,$AE,$D9,$F3,$6D,$C1,$E8
	.dc.b	$72,$EA,$F4,$47,$F8,$02,$BF,$AF,$B6,$7D,$D5,$CD,$7C,$FF,$FB,$6D
	.dc.b	$E6,$D7,$EC,$1F,$15,$C6,$1F,$0B,$D9,$78,$FF,$1E,$78,$57,$EA,$5A
	.dc.b	$FF,$70,$6B,$CF,$DC,$FA,$B1,$D3,$2F,$EE,$B6,$F4,$BE,$93,$6F,$CB
	.dc.b	$7F,$57,$E7,$DE,$2D,$24,$FC,$4F,$82,$FF,$4F,$BF,$1A,$76,$E7,$E3
	.dc.b	$3E,$F4,$9E,$31,$F0,$39,$FF,$3D,$EB,$7E,$E3,$E4,$24,$FE,$D6,$3F
	.dc.b	$76,$61,$E1,$D0,$9C,$C7,$D6,$F8,$4F,$C1,$BF,$30,$38,$E2,$FC,$6B
	.dc.b	$78,$3F,$C2,$FF,$4C,$F8,$1F,$CA,$4F,$2B,$FE,$2C,$D7,$FC,$3A,$D7
	.dc.b	$F3,$8D,$7B,$F5,$7E,$77,$E8,$3F,$37,$8F,$E7,$1B,$8F,$29,$FE,$2F
	.dc.b	$F2,$3F,$C6,$3E,$4A,$61,$FC,$56,$32,$FF,$17,$8C,$BF,$0A,$3F,$8F
	.dc.b	$F9,$72,$3F,$8E,$3F,$C8,$BE,$4B,$11,$F3,$FB,$3F,$38,$DE,$CC,$E9
	.dc.b	$FB,$0F,$9F,$7C,$67,$ED,$5B,$5F,$A2,$6B,$F3,$7E,$7F,$F7,$1B,$7F
	.dc.b	$90,$27,$33,$FC,$F5,$AE,$6B,$F8,$3C,$62,$3F,$82,$3D,$2F,$F8,$0D
	.dc.b	$FD,$59,$FE,$10,$D5,$3B,$BF,$E6,$5E,$9F,$F4,$0E,$89,$5F,$CF,$63
	.dc.b	$FA,$19,$87,$F1,$31,$87,$FA,$17,$D0,$7F,$48,$F8,$09,$18,$21,$F9
	.dc.b	$07,$2F,$E4,$A3,$E3,$72,$48,$F8,$CF,$F1,$DC,$BF,$2D,$18,$3F,$A0
	.dc.b	$E5,$F9,$46,$CF,$C4,$23,$FC,$42,$5E,$3F,$33,$B4,$7E,$8B,$DD,$FE
	.dc.b	$BE,$47,$FA,$E7,$2F,$F6,$2F,$57,$F9,$C1,$23,$FA,$F3,$FD,$3B,$E7
	.dc.b	$DF,$41,$FD,$B7,$93,$F7,$35,$F8,$7F,$0B,$FB,$85,$6A,$3F,$97,$78
	.dc.b	$08,$7C,$9E,$89,$3A,$87,$CA,$F2,$FA,$0F,$01,$27,$FC,$04,$7F,$80
	.dc.b	$25,$1E,$5F,$A0,$F2,$79,$92,$BA,$47,$93,$E5,$24,$77,$7C,$FF,$2B
	.dc.b	$E8,$FA,$0F,$71,$27,$BE,$9E,$98,$F8,$5F,$41,$EE,$24,$FA,$7C,$AF
	.dc.b	$83,$F0,$F2,$3D,$8F,$C4,$F8,$9F,$8B,$D3,$F2,$C9,$3D,$B7,$FB,$CF
	.dc.b	$F8,$7F,$2F,$91,$F0,$FF,$30,$91,$F1,$FC,$FF,$9F,$E0,$FF,$87,$FE
	.dc.b	$7D,$F2,$B4,$CD,$7D,$28,$FE,$30,$FB,$1A,$F7,$6B,$FE,$B4,$C1,$18
	.dc.b	$8F,$EE,$91,$8F,$95,$F0,$7C,$AC,$9E,$D1,$E3,$F9,$4F,$27,$CA,$C9
	.dc.b	$F1,$45,$7C,$11,$FE,$1A,$C0,$D3,$1F,$25,$ED,$24,$FA,$E3,$FC,$09
	.dc.b	$F4,$0F,$85,$A4,$9F,$81,$ED,$57,$F1,$33,$B7,$FC,$5A,$D7,$33,$FE
	.dc.b	$16,$8C,$79,$9F,$E2,$2C,$5F,$25,$F3,$CE,$91,$FE,$12,$F9,$E7,$A3
	.dc.b	$FE,$17,$6E,$3F,$E0,$46,$BF,$D6,$EF,$FB,$BB,$99,$FF,$12,$CE,$A3
	.dc.b	$FC,$C0,$D7,$FD,$CE,$D7,$FC,$C3,$2F,$8F,$FC,$11,$23,$F0,$1F,$23
	.dc.b	$5F,$E7,$99,$FE,$E4,$C6,$3E,$DF,$8A,$FF,$6E,$79,$5F,$7B,$1F,$F3
	.dc.b	$4D,$BE,$3F,$FC,$53,$CB,$F8,$4B,$DC,$F8,$BE,$46,$CE,$63,$FC,$75
	.dc.b	$5C,$2F,$9B,$6F,$70,$BE,$74,$F9,$F7,$CD,$FF,$B9,$E3,$C5,$97,$F9
	.dc.b	$15,$D4,$67,$1F,$EF,$75,$C3,$CE,$F0,$DE,$E3,$F8,$6D,$E0,$FF,$27
	.dc.b	$3B,$7A,$98,$BE,$93,$C5,$B1,$C3,$CA,$F3,$D7,$1C,$3C,$35,$AA,$F2
	.dc.b	$E1,$DD,$F6,$5F,$09,$D5,$F1,$C3,$B5,$EE,$3C,$F7,$E4,$39,$95,$E0
	.dc.b	$F9,$E3,$85,$F5,$7F,$D1,$33,$ED,$9F,$F5,$53,$5E,$E7,$F2,$CF,$92
	.dc.b	$FC,$CF,$D0,$7F,$A4,$1B,$19,$E4,$4E,$5B,$1F,$CD,$2B,$FC,$53,$C3
	.dc.b	$E0,$F0,$BF,$59,$3F,$45,$AF,$D0,$6B,$F7,$DF,$DA,$6B,$F6,$9B,$A7
	.dc.b	$DF,$F3,$9C,$8F,$03,$E3,$56,$B0,$E9,$8E,$0F,$A1,$7C,$2B,$13,$F2
	.dc.b	$6B,$83,$E1,$AD,$DF,$58,$D4,$70,$C7,$C8,$EA,$F1,$8F,$B6,$2D,$7F
	.dc.b	$A6,$71,$F0,$3C,$8F,$C6,$F2,$3E,$83,$FD,$A1,$87,$E4,$E3,$1F,$23
	.dc.b	$E1,$CB,$F5,$FE,$8F,$FA,$8E,$B5,$7C,$30,$E6,$DA,$FD,$AB,$73,$FF
	.dc.b	$5E,$CF,$37,$F9,$CC,$66,$3F,$DD,$2F,$FB,$52,$BF,$BB,$B6,$3F,$CE
	.dc.b	$0B,$E4,$7E,$27,$4F,$FB,$12,$34,$DF,$91,$CD,$F6,$6D,$F6,$BB,$A7
	.dc.b	$C1,$9C,$CB,$B3,$FD,$6F,$7F,$C9,$F2,$DE,$4F,$90,$DD,$6A,$FF,$CF
	.dc.b	$0D,$CF,$F9,$95,$F0,$AD,$BD,$BD,$4F,$3E,$3F,$0B,$E4,$71,$9F,$66
	.dc.b	$C7,$FA,$EA,$77,$97,$C6,$F6,$32,$FF,$A1,$E3,$2F,$F8,$72,$FC,$F3
	.dc.b	$AC,$BE,$37,$C8,$7A,$8F,$5F,$C4,$5C,$7E,$7B,$87,$E8,$FC,$55,$C7
	.dc.b	$E8,$B6,$FE,$83,$67,$ED,$FC,$2F,$C6,$F8,$6F,$8E,$D8,$FB,$7D,$9F
	.dc.b	$28,$F9,$7F,$0D,$63,$EF,$FC,$1E,$1A,$7E,$0A,$3E,$8F,$C1,$C8,$FC
	.dc.b	$2F,$C2,$5D,$3F,$0B,$C7,$F5,$8F,$AF,$F0,$96,$BF,$17,$F2,$5F,$68
	.dc.b	$FB,$7F,$07,$63,$FC,$8B,$F3,$8F,$A0,$FC,$2B,$63,$F2,$7E,$46,$27
	.dc.b	$3E,$4C,$9E,$3F,$0B,$74,$FF,$63,$60,$FA,$0E,$67,$4E,$BB,$7E,$0F
	.dc.b	$A0,$EE,$78,$3E,$1E,$C7,$93,$E1,$3C,$67,$93,$E1,$AC,$7A,$3E,$2B
	.dc.b	$CE,$7A,$3E,$1E,$C7,$B3,$E1,$7D,$67,$B3,$E1,$EC,$7C,$0F,$85,$F7
	.dc.b	$9F,$03,$E1,$EC,$7C,$4F,$5B,$E1,$9F,$13,$E1,$EC,$38,$BF,$DB,$1F
	.dc.b	$41,$FD,$9A,$3E,$5F,$AB,$FA,$C8,$C4,$FC,$48,$C5,$CA,$FA,$3F,$0D
	.dc.b	$6B,$9C,$DF,$1B,$F9,$17,$9E,$E7,$5E,$70,$FF,$21,$C7,$F3,$37,$ED
	.dc.b	$FC,$3F,$DA,$7E,$DB,$5C,$9F,$80,$32,$7F,$A1,$BD,$DF,$F7,$63,$7E
	.dc.b	$1F,$F8,$3C,$63,$36,$5E,$48,$FF,$2F,$D7,$1F,$8E,$DD,$F0,$6C,$7C
	.dc.b	$7A,$FE,$79,$6F,$9F,$7C,$4E,$23,$FC,$19,$18,$AF,$F4,$7C,$6E,$74
	.dc.b	$FD,$19,$D3,$FE,$94,$6D,$76,$6D,$78,$7E,$77,$FD,$42,$F8,$CE,$F7
	.dc.b	$FC,$8A,$77,$7D,$63,$F9,$4D,$AF,$8C,$FC,$2B,$E2,$7C,$9F,$8A,$F8
	.dc.b	$E7,$C8,$BB,$19,$24,$F1,$1B,$B4,$F0,$B6,$FE,$8F,$C4,$7D,$2F,$8A
	.dc.b	$B8,$7D,$76,$E1,$CE,$F9,$46,$33,$FA,$CF,$B7,$E6,$7D,$7F,$41,$FD
	.dc.b	$92,$7E,$34,$63,$E0,$B5,$8F,$C5,$C7,$E7,$4F,$C2,$FA,$07,$AB,$A9
	.dc.b	$F5,$1F,$6E,$71,$87,$F5,$6A,$DD,$F3,$AD,$3F,$8F,$F8,$CF,$C6,$FC
	.dc.b	$6F,$E3,$9B,$19,$3F,$29,$19,$69,$CF,$E8,$1E,$9F,$F5,$A6,$C7,$E6
	.dc.b	$FE,$37,$13,$23,$83,$F9,$C8,$C6,$1E,$AF,$41,$E0,$7E,$03,$9D,$E1
	.dc.b	$7F,$66,$9C,$E3,$F1,$76,$FC,$24,$E6,$E4,$E6,$F7,$87,$59,$CB,$67
	.dc.b	$51,$8B,$FD,$2F,$D0,$3D,$5F,$EE,$AD,$AF,$DB,$7C,$DE,$26,$47,$CD
	.dc.b	$9D,$61,$F8,$3E,$A5,$EA,$3E,$BF,$7B,$D7,$FF,$00,$B7,$0F,$E0,$5C
	.dc.b	$F8,$99,$1F,$66,$75,$5D,$6B,$2F,$7F,$8D,$FE,$0E,$7F,$08,$F8,$FB
	.dc.b	$7F,$C3,$62,$CF,$96,$6E,$BF,$C2,$FE,$83,$F8,$79,$FC,$43,$4B,$87
	.dc.b	$F1,$58,$B7,$ED,$9B,$5F,$C4,$FA,$7F,$8B,$9F,$C6,$3E,$7E,$E1,$FC
	.dc.b	$76,$2D,$FC,$59,$B5,$FC,$6F,$E5,$FF,$8F,$9F,$C8,$3E,$36,$E1,$FC
	.dc.b	$96,$2D,$FC,$D9,$B5,$FC,$8F,$E5,$78,$5F,$F2,$B6,$D7,$D3,$6C,$EE
	.dc.b	$7F,$98,$57,$F8,$CA,$BF,$D6,$E3,$E8,$18,$FF,$AB,$EF,$EE,$C6,$27
	.dc.b	$FD,$64,$7F,$B1,$4E,$13,$BE,$4F,$F3,$54,$EE,$7F,$D7,$1F,$1B,$FD
	.dc.b	$66,$31,$7D,$2F,$9D,$73,$9C,$DF,$29,$CD,$F1,$7B,$63,$6F,$C7,$35
	.dc.b	$FE,$8F,$5C,$5C,$9E,$63,$FA,$23,$97,$CA,$64,$E3,$3C,$DF,$8B,$7E
	.dc.b	$48,$C4,$79,$EF,$C6,$F1,$36,$E2,$7E,$14,$60,$F0,$CF,$2B,$F4,$61
	.dc.b	$F0,$E7,$CF,$5D,$DE,$87,$73,$89,$89,$E3,$87,$AE,$3B,$1E,$A3,$47
	.dc.b	$B6,$BB,$3F,$02,$BB,$1E,$67,$B1,$C5,$C5,$76,$BF,$44,$F3,$BF,$03
	.dc.b	$C2,$FA,$BE,$26,$B9,$8D,$D7,$38,$F2,$DF,$8B,$0F,$BF,$1D,$70,$EF
	.dc.b	$EE,$BD,$EC,$B6,$7E,$EB,$5F,$B6,$DF,$93,$FE,$21,$9F,$F8,$A6,$31
	.dc.b	$CD,$FE,$0C,$A4,$62,$75,$7F,$E2,$F9,$5F,$27,$71,$98,$C6,$5F,$9F
	.dc.b	$C9,$B7,$18,$C3,$63,$86,$5E,$19,$FD,$35,$72,$9F,$01,$CB,$0F,$CE
	.dc.b	$35,$FC,$B4,$6D,$FF,$22,$35,$FD,$CD,$F7,$7F,$C9,$13,$C6,$BB,$DF
	.dc.b	$86,$3C,$78,$F6,$BE,$17,$C8,$FD,$66,$1D,$4E,$D5,$CC,$F3,$CF,$08
	.dc.b	$D5,$F5,$AF,$0C,$F4,$AE,$11,$E8,$9D,$3F,$39,$B1,$A8,$F1,$1F,$02
	.dc.b	$37,$5E,$2A,$E8,$78,$63,$27,$27,$33,$EA,$8C,$DF,$BE,$21,$D5,$D1
	.dc.b	$8C,$3E,$CC,$60,$D1,$FC,$1F,$0F,$31,$E7,$C3,$E0,$D7,$18,$F8,$B3
	.dc.b	$C6,$BE,$8C,$E6,$3E,$29,$99,$FF,$3F,$78,$DF,$E3,$5F,$91,$FE,$65
	.dc.b	$AE,$38,$FF,$98,$E1,$FE,$63,$C3,$27,$F8,$F2,$15,$8B,$E5,$3F,$E6
	.dc.b	$F3,$13,$CA,$FA,$56,$FE,$6D,$F2,$54,$97,$C4,$FF,$1B,$60,$9D,$C6
	.dc.b	$1A,$FF,$91,$5B,$8F,$9E,$35,$F1,$9D,$C8,$F1,$F9,$1A,$F1,$8B,$3F
	.dc.b	$E9,$4A,$57,$9F,$E3,$A5,$F8,$5F,$57,$C7,$E2,$3C,$51,$6F,$94,$FB
	.dc.b	$FE,$3A,$57,$A5,$F8,$5F,$1F,$88,$F5,$45,$AF,$14,$FD,$09,$7C,$FF
	.dc.b	$CB,$7F,$D6,$CD,$72,$6F,$E8,$3F,$8B,$57,$FA,$52,$74,$49,$D7,$3F
	.dc.b	$5E,$CF,$E0,$15,$FE,$9A,$6F,$17,$F1,$56,$D7,$FB,$17,$E5,$11,$C7
	.dc.b	$A6,$4F,$F9,$8E,$36,$A7,$6F,$F9,$9F,$E8,$3F,$D7,$ED,$F6,$FD,$D1
	.dc.b	$88,$FF,$38,$FD,$07,$F8,$F9,$D1,$0F,$F3,$AC,$6F,$2F,$F5,$8C,$EE
	.dc.b	$73,$5F,$26,$1E,$8D,$3B,$C7,$CD,$FA,$07,$85,$49,$7F,$E7,$DF,$83
	.dc.b	$FF,$39,$BF,$E7,$F9,$DB,$67,$6F,$FA,$22,$76,$FE,$57,$D5,$FD,$07
	.dc.b	$CA,$C8,$FC,$D4,$7E,$8C,$C3,$F9,$B8,$C3,$FE,$8D,$F8,$17,$C8,$FE
	.dc.b	$0D,$B1,$98,$FC,$37,$D0,$3E,$03,$51,$27,$47,$FA,$5E,$74,$E0,$FF
	.dc.b	$4F,$FD,$07,$FB,$6B,$83,$FD,$B6,$DE,$5F,$EA,$D8,$F8,$2B,$F6,$1F
	.dc.b	$2D,$FB,$BF,$97,$91,$FB,$58,$FD,$F1,$87,$F6,$D1,$87,$F7,$5E,$4F
	.dc.b	$F0,$32,$3F,$BD,$F9,$7A,$FE,$F6,$2E,$5F,$B5,$9F,$D2,$7D,$03,$E3
	.dc.b	$69,$2F,$FD,$69,$F4,$1F,$A2,$D3,$FB,$75,$6E,$7F,$D7,$0D,$7F,$D7
	.dc.b	$ED,$C3,$FD,$F7,$5C,$64,$27,$83,$FE,$56,$6B,$DA,$FF,$DF,$AE,$DF
	.dc.b	$F7,$FF,$8F,$FC,$6F,$D0,$FE,$45,$F3,$32,$BF,$8E,$47,$F2,$53,$0F
	.dc.b	$F1,$D8,$C3,$FC,$87,$E6,$7F,$95,$7C,$04,$8C,$10,$FE,$48,$E5,$FE
	.dc.b	$4D,$1F,$23,$86,$47,$C2,$7F,$87,$66,$E6,$CC,$FC,$49,$BA,$FC,$49
	.dc.b	$BB,$7F,$86,$6D,$E2,$F2,$33,$83,$5F,$E7,$5E,$FF,$F4,$12,$47,$F3
	.dc.b	$BE,$7F,$E8,$9E,$9C,$8F,$E7,$CF,$F4,$7F,$17,$FA,$77,$D0,$3E,$62
	.dc.b	$92,$BF,$A3,$C3,$E6,$B7,$F3,$12,$72,$F2,$F9,$AE,$7F,$31,$27,$8B
	.dc.b	$DB,$E6,$BB,$FC,$C4,$9E,$AF,$8B,$E6,$BC,$7F,$33,$27,$C2,$F9,$BE
	.dc.b	$6B,$CF,$F3,$52,$7C,$AF,$AB,$BB,$D7,$F3,$52,$7D,$2F,$BB,$C9,$F7
	.dc.b	$FC,$54,$9F,$6B,$F0,$BE,$33,$E1,$F7,$49,$F8,$2F,$C6,$F9,$9F,$8F
	.dc.b	$F1,$52,$7E,$2B,$F2,$BD,$0F,$97,$DF,$FD,$E7,$E7,$DF,$33,$4C,$DF
	.dc.b	$F7,$28,$FE,$0E,$BC,$35,$7F,$82,$A7,$31,$89,$FC,$04,$62,$75,$5D
	.dc.b	$FE,$5F,$C1,$F3,$12,$FB,$47,$8F,$E5,$FC,$9F,$31,$27,$C5,$15,$F1
	.dc.b	$61,$E1,$9C,$65,$E1,$F1,$19,$B0,$F7,$7C,$B4,$C7,$D8,$F0,$C7,$FC
	.dc.b	$E5,$F4,$1F,$E1,$C9,$FD,$FE,$3E,$B5,$19,$7C,$27,$F0,$47,$11,$E1
	.dc.b	$C3,$FA,$7C,$F2,$AE,$BF,$11,$FD,$EE,$F7,$18,$9F,$9D,$ED,$66,$BE
	.dc.b	$A7,$CB,$49,$E4,$77,$8F,$F0,$27,$CE,$BD,$3A,$49,$FB,$1F,$2B,$5F
	.dc.b	$38,$9E,$0F,$43,$13,$F8,$39,$FC,$27,$C9,$C9,$FC,$13,$F8,$7F,$95
	.dc.b	$C4,$7C,$2E,$1F,$9F,$38,$DF,$E1,$57,$E3,$7E,$3B,$FD,$EB,$E7,$7E
	.dc.b	$1D,$B8,$7E,$19,$B8,$7D,$8F,$95,$BC,$49,$A4,$AF,$E6,$7F,$27,$FD
	.dc.b	$8A,$F8,$46,$30,$FF,$56,$7C,$FB,$B7,$FD,$63,$2B,$FE,$B5,$39,$F4
	.dc.b	$78,$4F,$F3,$E8,$C3,$BB,$E7,$E6,$7A,$21,$3A,$BE,$78,$76,$6E,$1D
	.dc.b	$2B,$77,$C6,$F8,$61,$E6,$C3,$83,$6F,$84,$AB,$85,$6A,$FC,$CD,$7C
	.dc.b	$6D,$9E,$11,$CA,$30,$4A,$EB,$5D,$1F,$F6,$94,$F8,$4A,$74,$6B,$4C
	.dc.b	$FC,$A7,$E4,$7E,$83,$2D,$8C,$F1,$A7,$2D,$8E,$B8,$BC,$2F,$D6,$47
	.dc.b	$D8,$6B,$F4,$DA,$FD,$06,$C7,$1E,$6F,$8F,$1F,$B9,$D7,$FC,$2D,$F0
	.dc.b	$8E,$3F,$E2,$91,$F8,$09,$D6,$FD,$EF,$89,$F5,$EF,$87,$56,$32,$FC
	.dc.b	$11,$E3,$9C,$61,$DB,$1E,$B8,$73,$AD,$9E,$7A,$DC,$7F,$BF,$F0,$EB
	.dc.b	$3A,$CB,$D4,$DC,$BC,$AF,$73,$9C,$F0,$C7,$D5,$87,$E6,$38,$5E,$3B
	.dc.b	$E4,$BC,$78,$7F,$71,$8F,$F4,$DE,$BB,$74,$DF,$91,$DD,$F2,$BB,$5C
	.dc.b	$2F,$E0,$CE,$6F,$81,$CF,$D8,$FE,$81,$D3,$FC,$C6,$DF,$21,$8A,$FF
	.dc.b	$67,$C7,$F8,$19,$FF,$5E,$BA,$6D,$6F,$0F,$45,$5C,$3D,$B3,$6F,$85
	.dc.b	$7F,$11,$7F,$C3,$17,$D3,$E4,$35,$EE,$6D,$B1,$93,$C3,$19,$7C,$7F
	.dc.b	$07,$D7,$E2,$3B,$36,$32,$7A,$23,$2D,$33,$87,$8E,$32,$D8,$F7,$F3
	.dc.b	$7A,$7E,$17,$D4,$D8,$F8,$7F,$0B,$86,$9C,$0B,$5E,$F8,$F8,$5F,$07
	.dc.b	$F0,$DF,$89,$F0,$9F,$1F,$C1,$7C,$BE,$30,$FE,$F4,$FC,$9F,$84,$F9
	.dc.b	$5F,$0B,$67,$E9,$BF,$37,$E1,$24,$7D,$6E,$EB,$A7,$CF,$F0,$BE,$91
	.dc.b	$F4,$FE,$12,$D7,$FB,$E3,$E9,$3F,$C6,$CD,$F8,$3F,$B5,$18,$BF,$E9
	.dc.b	$0F,$F8,$DE,$78,$7C,$1E,$8D,$F8,$F6,$3A,$7C,$1F,$23,$9F,$C1,$D8
	.dc.b	$F0,$7C,$1F,$6F,$83,$EE,$D8,$F2,$7C,$1F,$88,$F1,$FC,$1D,$8F,$47
	.dc.b	$C1,$F9,$BE,$0F,$CE,$D8,$F6,$78,$3E,$AF,$83,$F5,$B6,$3E,$07,$C1
	.dc.b	$FD,$1F,$96,$F7,$B6,$3E,$27,$9D,$F5,$7E,$63,$E1,$B6,$3E,$47,$C2
	.dc.b	$7C,$6F,$84,$F8,$ED,$8F,$F0,$0F,$94,$FC,$0B,$FC,$6F,$1F,$3C,$FD
	.dc.b	$33,$8E,$8F,$9B,$1F,$AB,$F3,$5F,$3B,$FB,$19,$FF,$0C,$C6,$3E,$5F
	.dc.b	$53,$98,$FD,$DA,$8C,$F9,$DA,$6D,$FF,$85,$2B,$93,$6B,$93,$FE,$2A
	.dc.b	$79,$39,$AE,$4F,$B3,$07,$C3,$7A,$8F,$5F,$C7,$D9,$D4,$5C,$BF,$C2
	.dc.b	$15,$FE,$DF,$9C,$9F,$E1,$BF,$9F,$E6,$F9,$7E,$49,$E9,$7F,$0D,$B5
	.dc.b	$69,$F8,$73,$AA,$E5,$87,$F9,$06,$7F,$D8,$27,$07,$0F,$FB,$12,$B9
	.dc.b	$3F,$DA,$9A,$FF,$B2,$36,$FB,$1F,$0F,$F5,$BE,$27,$EB,$B7,$0F,$84
	.dc.b	$DA,$F6,$B7,$7F,$F2,$E6,$9F,$E5,$BA,$F9,$17,$C9,$A7,$C1,$F8,$3F
	.dc.b	$E0,$5F,$2B,$EF,$CD,$18,$9D,$13,$2F,$97,$F1,$5B,$9F,$C5,$FC,$47
	.dc.b	$E3,$3E,$2A,$DF,$D7,$6D,$FD,$26,$E9,$E6,$7F,$21,$F3,$AF,$0E,$F8
	.dc.b	$98,$6E,$1F,$89,$9D,$36,$74,$D3,$53,$C3,$0F,$CC,$D7,$06,$D7,$DD
	.dc.b	$F9,$E7,$B3,$E2,$F1,$70,$DC,$7F,$CC,$13,$A7,$FC,$3A,$DB,$FF,$10
	.dc.b	$B5,$FF,$12,$B5,$FC,$1F,$A1,$FC,$6F,$4F,$EC,$3F,$3E,$F8,$AF,$C7
	.dc.b	$4E,$70,$FF,$45,$CE,$5F,$F3,$84,$66,$C1,$C5,$FE,$E3,$D9,$E7,$F2
	.dc.b	$3F,$33,$C3,$F9,$BF,$13,$71,$F9,$4D,$C7,$E2,$B7,$0F,$EE,$BC,$1F
	.dc.b	$E6,$3A,$FA,$17,$C9,$A7,$23,$FD,$09,$F8,$17,$CD,$AF,$F4,$E7,$C3
	.dc.b	$7E,$A6,$31,$3F,$EC,$18,$C7,$1C,$8F,$DE,$7C,$75,$CF,$FD,$23,$F9
	.dc.b	$D3,$C7,$FE,$AD,$F3,$7F,$B3,$8C,$5F,$FB,$3E,$31,$32,$3F,$83,$75
	.dc.b	$5B,$FE,$03,$C9,$FC,$83,$E2,$FF,$90,$B6,$BF,$86,$36,$77,$1F,$C3
	.dc.b	$E7,$F9,$3F,$C7,$7F,$28,$F8,$FB,$3F,$C4,$DB,$5F,$C3,$9B,$C1,$FB
	.dc.b	$27,$F9,$7F,$9F,$FC,$C3,$E0,$2D,$FF,$1B,$6D,$FF,$16,$6D,$FE,$E9
	.dc.b	$FE,$6F,$E1,$7F,$38,$F3,$ED,$FF,$23,$6D,$FF,$1E,$6D,$FF,$02,$7F
	.dc.b	$9F,$EF,$2B,$F9,$57,$1D,$BF,$E4,$B3,$6F,$F8,$4C,$DA,$FF,$3E,$FC
	.dc.b	$67,$E4,$3E,$81,$A7,$F1,$EB,$7C,$27,$FB,$EE,$73,$97,$FA,$1A,$F8
	.dc.b	$5E,$EB,$4F,$49,$57,$05,$7D,$30,$E3,$1C,$CD,$D7,$38,$CC,$7F,$30
	.dc.b	$C2,$D7,$18,$D9,$C6,$FB,$46,$5F,$04,$7F,$B1,$8D,$9E,$2C,$3C,$EF
	.dc.b	$1C,$3C,$0E,$8E,$13,$D2,$77,$1E,$33,$9D,$77,$AF,$81,$5B,$C3,$D3
	.dc.b	$5C,$A7,$AC,$F9,$23,$FD,$A4,$72,$BE,$AE,$CF,$E0,$77,$D5,$F3,$18
	.dc.b	$8D,$BC,$A7,$C0,$FA,$4C,$57,$80,$F2,$95,$E6,$7C,$3B,$F3,$BC,$0E
	.dc.b	$CE,$30,$FA,$37,$E0,$6C,$66,$7D,$4D,$BF,$19,$EF,$9D,$CF,$95,$B1
	.dc.b	$98,$E2,$64,$AE,$7E,$01,$F0,$1F,$8D,$6C,$FE,$25,$AF,$D5,$6F,$C8
	.dc.b	$3E,$1B,$F2,$32,$9C,$C6,$1B,$A7,$F3,$26,$CE,$EF,$FC,$03,$7C,$EB
	.dc.b	$F2,$BA,$36,$FF,$02,$E5,$F9,$F8,$C6,$1F,$9A,$BE,$0F,$E4,$9B,$3B
	.dc.b	$AE,$35,$FE,$0E,$38,$D6,$F2,$F4,$D7,$EA,$EF,$8D,$6E,$F8,$DF,$F4
	.dc.b	$A8,$E3,$5B,$BE,$31,$EA,$6C,$EE,$B8,$BF,$C0,$9B,$7D,$A3,$FA,$AB
	.dc.b	$5D,$1F,$0A,$37,$3D,$21,$3B,$AF,$19,$E7,$8F,$89,$3B,$9D,$57,$8A
	.dc.b	$B7,$5E,$0A,$D1,$D6,$33,$1E,$23,$E3,$61,$D0,$F2,$30,$C9,$30,$FA
	.dc.b	$0E,$5F,$8D,$1C,$63,$F8,$3C,$F6,$3B,$E1,$F0,$4F,$3C,$71,$8F,$99
	.dc.b	$3C,$6B,$AC,$F2,$C3,$E2,$CE,$6B,$E1,$4E,$4C,$0F,$3B,$FC,$FC,$FF
	.dc.b	$33,$D7,$FC,$95,$E6,$FF,$94,$A3,$FC,$C2,$64,$E0,$BE,$3B,$15,$CA
	.dc.b	$7F,$CD,$C6,$27,$96,$1F,$E4,$CA,$DF,$CF,$3E,$3E,$FD,$04,$C3,$FD
	.dc.b	$23,$39,$6C,$E5,$FE,$7B,$97,$F9,$7E,$55,$D2,$BA,$FC,$7F,$6F,$90
	.dc.b	$98,$74,$8C,$BD,$A3,$2F,$43,$E2,$5A,$7C,$FE,$CF,$A3,$DB,$95,$E6
	.dc.b	$8C,$BE,$88,$CB,$CF,$65,$AB,$E0,$FC,$87,$C2,$F6,$A5,$FC,$08,$CB
	.dc.b	$F0,$A3,$2F,$3A,$4F,$D3,$87,$E7,$8F,$97,$C3,$63,$F9,$EE,$FF,$E8
	.dc.b	$C8,$C4,$7B,$C7,$E7,$9E,$F7,$67,$F9,$1A,$37,$C4,$AF,$F4,$AD,$7D
	.dc.b	$5A,$E4,$D9,$D4,$71,$BF,$37,$7B,$D8,$A4,$9E,$E3,$E9,$BF,$96,$DF
	.dc.b	$28,$57,$17,$72,$F8,$EF,$84,$D8,$E7,$3B,$7D,$DF,$4E,$E4,$F9,$0F
	.dc.b	$9F,$4F,$F5,$7B,$67,$FC,$CA,$ED,$F2,$B5,$E7,$CB,$C8,$E7,$52,$37
	.dc.b	$19,$78,$C7,$D7,$FA,$0F,$F3,$C4,$7C,$06,$BF,$4D,$B9,$FC,$1F,$A8
	.dc.b	$7B,$7C,$7C,$BF,$C0,$43,$E3,$29,$9C,$3E,$5F,$A7,$F6,$DF,$DD,$07
	.dc.b	$0F,$EF,$5A,$FF,$B4,$1A,$FF,$A1,$9E,$58,$7D,$3F,$89,$F8,$D2,$6C
	.dc.b	$EB,$6F,$F6,$CF,$D0,$BD,$3F,$CB,$2F,$80,$A6,$22,$61,$F8,$29,$D3
	.dc.b	$67,$2F,$E5,$E3,$A3,$92,$19,$9F,$E1,$BF,$31,$FC,$3B,$DF,$93,$FC
	.dc.b	$32,$32,$FF,$0E,$8C,$BF,$AD,$8F,$D6,$7D,$03,$E4,$A9,$27,$F2,$FF
	.dc.b	$42,$F9,$0C,$47,$EA,$E1,$F0,$9F,$B5,$6C,$7E,$7A,$FA,$3F,$C7,$BE
	.dc.b	$67,$F9,$1F,$CD,$48,$C1,$0F,$E0,$0E,$5F,$E0,$51,$FC,$A8,$C7,$17
	.dc.b	$EC,$3C,$6F,$E1,$F5,$9A,$F0,$65,$F8,$27,$FD,$83,$87,$A5,$C3,$5F
	.dc.b	$C3,$B5,$FE,$84,$E6,$FF,$69,$F4,$3F,$AF,$9F,$E1,$4D,$8F,$DC,$4E
	.dc.b	$F8,$BF,$9B,$7C,$27,$F4,$02,$5F,$F3,$EF,$4A,$BF,$C7,$62,$DF,$AF
	.dc.b	$C7,$FF,$16,$4F,$F0,$6C,$59,$FF,$21,$5C,$34,$C4,$49,$E1,$19,$79
	.dc.b	$46,$5F,$4C,$65,$F5,$C7,$F5,$6F,$99,$FE,$AF,$EB,$C9,$FE,$AB,$1F
	.dc.b	$D6,$BC,$4F,$EB,$7F,$35,$27,$FA,$CC,$7F,$5E,$F5,$BF,$B2,$FC,$DF
	.dc.b	$F6,$79,$FE,$5B,$C6,$8C,$47,$F3,$D9,$FE,$D7,$BB,$E6,$F8,$CC,$38
	.dc.b	$C6,$5B,$F3,$5C,$3E,$26,$46,$0D,$F8,$3D,$3A,$E4,$60,$E7,$E2,$78
	.dc.b	$08,$F6,$F9,$AF,$1F,$6F,$93,$E1,$E4,$F8,$A3,$CF,$D9,$23,$CA,$FA
	.dc.b	$BE,$6B,$D7,$E2,$C9,$F4,$BE,$EF,$9A,$F7,$FA,$1F,$03,$E6,$A5,$7B
	.dc.b	$A3,$E1,$FC,$6C,$8F,$82,$FC,$6F,$9A,$F8,$FF,$35,$27,$E2,$BF,$2B
	.dc.b	$CB,$F9,$7F,$19,$F3,$3C,$59,$5F,$2A,$3F,$C3,$3F,$81,$BE,$87,$FC
	.dc.b	$37,$A7,$F8,$32,$BE,$BB,$5F,$EF,$93,$B7,$ED,$AF,$99,$E6,$48,$CF
	.dc.b	$CD,$D7,$8F,$CD,$78,$0C,$79,$3F,$B9,$B4,$79,$3E,$6A,$57,$68,$F3
	.dc.b	$FC,$CF,$A3,$DC,$93,$E2,$8C,$BE,$48,$CB,$E8,$8C,$BE,$A8,$CE,$1E
	.dc.b	$19,$F8,$7E,$97,$C4,$F5,$25,$FB,$A3,$2F,$C0,$8C,$BF,$12,$3E,$5F
	.dc.b	$AD,$F3,$3E,$4A,$4F,$C6,$8C,$BF,$22,$32,$FC,$C8,$FA,$7F,$25,$FE
	.dc.b	$F3,$F2,$93,$FE,$57,$D9,$F2,$59,$8C,$37,$CB,$4A,$F0,$DA,$E7,$F1
	.dc.b	$FF,$E0,$7A,$DD,$E8,$DD,$F5,$A7,$CF,$3B,$3F,$D7,$90,$C3,$F9,$14
	.dc.b	$E7,$E4,$3A,$E7,$E8,$AE,$37,$C2,$F7,$E5,$75,$C3,$85,$73,$BE,$11
	.dc.b	$8A,$EB,$7C,$67,$37,$C6,$31,$7F,$1A,$F8,$57,$29,$CC,$FA,$AB,$86
	.dc.b	$1E,$68,$E1,$3A,$BE,$31,$CE,$B6,$F9,$4D,$D7,$17,$DA,$D9,$4F,$09
	.dc.b	$FF,$45,$47,$B9,$B5,$F2,$9B,$5F,$D5,$7E,$41,$EC,$BE,$7F,$0D,$AF
	.dc.b	$F0,$DC,$EA,$72,$7F,$5D,$C3,$FC,$3D,$F1,$B7,$C0,$FE,$5B,$7F,$31
	.dc.b	$AF,$C7,$6B,$F7,$9F,$E6,$DA,$FD,$C6,$EB,$F6,$FE,$49,$C9,$FE,$5E
	.dc.b	$AD,$61,$F0,$6F,$78,$75,$8F,$8D,$7C,$3C,$1D,$44,$9F,$98,$78,$5D
	.dc.b	$E1,$D7,$1F,$44,$6F,$1E,$F5,$C3,$0E,$8F,$2C,$BE,$0B,$72,$F1,$BD
	.dc.b	$4E,$18,$F5,$7C,$C7,$DD,$E3,$77,$FF,$99,$F4,$FC,$34,$63,$E0,$F8
	.dc.b	$3A,$F8,$3B,$5F,$EB,$BA,$DB,$C6,$FA,$39,$6C,$F0,$BE,$3F,$17,$FE
	.dc.b	$88,$ED,$FF,$0D,$5F,$FA,$C7,$0F,$E6,$33,$97,$F8,$1E,$3E,$6F,$8D
	.dc.b	$C6,$5F,$ED,$BB,$FF,$57,$57,$03,$F9,$CF,$C6,$EA,$FD,$B8,$5A,$FF
	.dc.b	$55,$D7,$F3,$EF,$8D,$C7,$C6,$E5,$B1,$93,$FA,$1C,$65,$FE,$8B,$EF
	.dc.b	$F3,$F8,$5E,$8D,$8C,$9F,$D2,$A3,$2D,$33,$9F,$FB,$1F,$04,$EA,$30
	.dc.b	$F9,$FA,$7D,$1F,$01,$6F,$FA,$AB,$EB,$F8,$3C,$5F,$36,$F0,$7C,$CF
	.dc.b	$6F,$57,$FD,$71,$B5,$E1,$F8,$1F,$88,$7C,$5F,$81,$B5,$FD,$8B,$E0
	.dc.b	$70,$D3,$FB,$2C,$7C,$BF,$99,$91,$FD,$A3,$BA,$E1,$ED,$F0,$7E,$81
	.dc.b	$F4,$7E,$06,$D7,$F8,$93,$E7,$DF,$03,$F5,$DF,$F1,$14,$66,$31,$F0
	.dc.b	$59,$F8,$19,$3C,$7C,$8B,$97,$FB,$C7,$07,$C0,$73,$3A,$7C,$15,$BF
	.dc.b	$07,$C0,$F7,$F8,$2F,$03,$63,$C9,$F0,$1E,$33,$C9,$F0,$56,$3D,$1F
	.dc.b	$01,$E7,$F8,$2F,$43,$63,$D9,$DD,$EB,$F8,$2F,$63,$63,$E0,$7C,$0F
	.dc.b	$BC,$F8,$1F,$05,$63,$E2,$79,$9F,$0C,$F8,$9E,$9D,$8F,$91,$F0,$3F
	.dc.b	$1F,$E0,$BE,$43,$63,$FC,$9F,$F2,$3F,$E1,$D7,$E7,$1F,$A3,$71,$CB
	.dc.b	$F1,$E3,$F5,$7E,$13,$E7,$FF,$57,$28,$FD,$B4,$3C,$DD,$BA,$8C,$C6
	.dc.b	$3C,$CC,$B6,$FF,$BD,$CE,$9A,$62,$B9,$36,$78,$5F,$86,$30,$7B,$3E
	.dc.b	$0F,$D4,$7A,$FE,$0E,$C6,$49,$3C,$47,$9E,$CD,$5C,$27,$43,$E7,$BF
	.dc.b	$C7,$97,$FE,$A9,$6C,$7F,$BB,$1A,$FF,$7F,$38,$60,$9E,$0F,$F1,$68
	.dc.b	$CC,$7C,$07,$E6,$7C,$E7,$FB,$3E,$31,$1A,$3D,$18,$F9,$FE,$13,$EC
	.dc.b	$7C,$2F,$D9,$6D,$7D,$E6,$D7,$0C,$FF,$C9,$73,$F8,$2A,$E0,$D3,$81
	.dc.b	$67,$E0,$E4,$F9,$AF,$43,$67,$F1,$7F,$0D,$89,$91,$E8,$9D,$57,$86
	.dc.b	$B2,$F8,$FE,$17,$F2,$1F,$0D,$F9,$16,$DF,$E5,$9B,$3B,$8F,$CC,$CF
	.dc.b	$E5,$35,$78,$17,$C1,$C3,$6B,$E6,$72,$2B,$71,$85,$5C,$23,$0D,$C3
	.dc.b	$F0,$D5,$C6,$30,$FF,$55,$AE,$35,$F5,$BE,$7B,$F9,$BB,$FE,$ED,$6D
	.dc.b	$FE,$0C,$38,$7F,$0C,$D7,$FC,$8C,$F0,$AF,$C4,$47,$F9,$82,$31,$3F
	.dc.b	$A8,$A7,$85,$F8,$B9,$CD,$7F,$99,$BC,$DF,$E5,$37,$F8,$A6,$C7,$B3
	.dc.b	$9D,$E8,$FE,$4E,$73,$97,$82,$91,$8B,$FF,$2D,$36,$3F,$81,$5F,$29
	.dc.b	$ED,$E5,$FD,$5E,$EF,$AD,$DD,$67,$E8,$36,$FE,$6E,$BF,$E6,$B9,$FA
	.dc.b	$95,$C1,$A7,$02,$CF,$0A,$FD,$57,$CE,$FD,$C8,$FD,$BB,$73,$F9,$3D
	.dc.b	$3F,$BD,$F8,$8D,$F1,$4C,$7F,$5B,$7B,$7F,$C1,$7D,$2A,$D6,$3F,$C0
	.dc.b	$BE,$75,$E0,$7E,$E5,$B7,$FC,$B3,$AB,$13,$23,$F7,$33,$AA,$FD,$95
	.dc.b	$65,$FD,$A7,$C4,$FF,$31,$F8,$0F,$E6,$4D,$BF,$E6,$ED,$9D,$C7,$F3
	.dc.b	$A9,$FE,$71,$F0,$FF,$CE,$7E,$22,$CF,$F3,$E6,$D7,$F0,$16,$F0,$7C
	.dc.b	$4D,$7F,$D5,$35,$FD,$13,$E6,$7F,$A1,$7C,$EF,$F4,$36,$C6,$4F,$E9
	.dc.b	$31,$96,$9F,$D3,$3E,$5B,$FA,$57,$CE,$FF,$4B,$6C,$64,$FE,$A3,$19
	.dc.b	$69,$FD,$53,$E7,$3F,$A9,$1F,$D4,$FE,$26,$C6,$49,$A7,$F1,$B8,$CB
	.dc.b	$FC,$D8,$FF,$5D,$8F,$9D,$65,$FE,$B1,$D1,$EF,$FE,$B2,$73,$5F,$CD
	.dc.b	$FE,$2F,$FD,$9B,$18,$8F,$EB,$D3,$BB,$FD,$D5,$69,$E9,$19,$9E,$47
	.dc.b	$86,$33,$8E,$E3,$37,$D7,$0F,$1C,$E6,$7F,$D8,$D8,$59,$CC,$73,$38
	.dc.b	$DF,$47,$07,$02,$C7,$0B,$F5,$47,$C4,$7C,$18,$79,$E3,$84,$63,$0F
	.dc.b	$14,$6E,$FB,$4F,$28,$FF,$6F,$1D,$AB,$93,$B3,$FB,$DB,$87,$71,$D5
	.dc.b	$F3,$9B,$3E,$6D,$78,$1D,$D7,$80,$F9,$EF,$78,$D6,$1F,$16,$FB,$36
	.dc.b	$33,$3B,$3B,$9D,$0F,$2B,$AB,$F2,$1E,$D9,$E5,$3C,$1B,$19,$8E,$8D
	.dc.b	$38,$DF,$99,$B1,$F5,$8E,$E7,$31,$EE,$BD,$DF,$6B,$63,$F1,$2D,$7E
	.dc.b	$E3,$7E,$39,$F0,$5F,$E3,$19,$E7,$46,$4E,$B4,$C5,$F1,$BE,$0D,$8C
	.dc.b	$DF,$1B,$5F,$17,$F1,$B8,$B6,$D6,$5F,$A5,$8C,$4F,$E6,$6B,$84,$FE
	.dc.b	$A5,$B3,$D4,$DC,$F4,$AE,$71,$E1,$9D,$CF,$AB,$2F,$75,$F7,$BD,$CF
	.dc.b	$BE,$F9,$CF,$7A,$DC,$73,$6C,$EE,$B8,$BF,$C1,$1B,$7E,$08,$F8,$33
	.dc.b	$C1,$D1,$FB,$5A,$EE,$75,$C7,$E0,$4E,$E3,$33,$DC,$CC,$C7,$E0,$1C
	.dc.b	$EF,$C3,$5D,$EF,$C5,$7C,$AB,$C2,$7E,$FA,$3C,$C6,$AB,$D7,$3B,$8C
	.dc.b	$D7,$8E,$72,$76,$BF,$16,$1F,$2F,$0F,$11,$8C,$3D,$8F,$98,$91,$F3
	.dc.b	$67,$51,$ED,$33,$3C,$27,$51,$83,$4E,$0F,$E5,$FD,$BF,$E6,$F8,$FE
	.dc.b	$69,$AF,$F7,$06,$9F,$E7,$4E,$3F,$F3,$56,$1F,$D5,$FE,$37,$37,$C6
	.dc.b	$BF,$D3,$66,$2B,$77,$FD,$EA,$B7,$F2,$6F,$8C,$BF,$3B,$30,$FF,$38
	.dc.b	$62,$AD,$46,$1A,$FF,$9C,$14,$F1,$8C,$57,$6F,$8D,$EF,$F1,$D2,$79
	.dc.b	$46,$5E,$91,$97,$A4,$79,$BD,$7F,$3F,$AF,$27,$C5,$19,$7C,$91,$97
	.dc.b	$C9,$1E,$EF,$8D,$F7,$FC,$74,$9F,$54,$65,$F6,$46,$5F,$64,$7D,$05
	.dc.b	$F3,$AF,$73,$0D,$8F,$F4,$9B,$5F,$F5,$2F,$CE,$B2,$DD,$49,$DB,$D6
	.dc.b	$75,$1F,$E9,$CE,$F7,$AF,$E1,$6C,$66,$3C,$9F,$3B,$FC,$8F,$9B,$FC
	.dc.b	$A9,$5A,$AF,$F5,$AB,$63,$35,$C3,$B3,$D9,$F3,$AE,$1B,$C5,$31,$FF
	.dc.b	$56,$CE,$5B,$39,$7F,$D8,$31,$F8,$2C,$3F,$CC,$8E,$E3,$33,$B9,$CD
	.dc.b	$78,$23,$E6,$7C,$EB,$C8,$A4,$BF,$F5,$EF,$CE,$FF,$A4,$9F,$F5,$FC
	.dc.b	$ED,$B3,$B7,$FD,$A7,$3B,$7F,$DA,$BE,$8F,$CC,$6F,$0F,$DB,$8C,$65
	.dc.b	$FE,$7B,$F8,$4B,$D9,$FC,$FE,$78,$5F,$FB,$5D,$AE,$6F,$FD,$30,$FF
	.dc.b	$4D,$8C,$BF,$D3,$FE,$77,$FC,$CC,$FF,$B5,$97,$BB,$F8,$B6,$C6,$7D
	.dc.b	$1C,$19,$99,$19,$21,$FE,$F0,$9E,$2D,$8E,$2F,$ED,$7E,$57,$F7,$BF
	.dc.b	$2B,$23,$04,$3F,$70,$E5,$FD,$D4,$7E,$7F,$E7,$1F,$21,$49,$3F,$EF
	.dc.b	$0F,$9D,$7C,$66,$21,$C1,$FE,$A5,$BE,$31,$FE,$F6,$6B,$AD,$FF,$B6
	.dc.b	$5F,$28,$FE,$19,$F2,$FF,$C5,$3E,$5E,$46,$08,$7F,$0F,$72,$FF,$11
	.dc.b	$8F,$E3,$66,$30,$FF,$4E,$F8,$AA,$CD,$78,$F2,$F7,$11,$FF,$03,$57
	.dc.b	$FA,$F7,$E7,$7F,$59,$97,$F8,$07,$3F,$35,$7F,$B5,$DA,$FF,$81,$43
	.dc.b	$87,$85,$FE,$1B,$C7,$FE,$55,$F0,$5F,$CC,$09,$3F,$CA,$FC,$EA,$FF
	.dc.b	$2B,$8B,$97,$AA,$7F,$89,$65,$EB,$AF,$3E,$5F,$E1,$58,$70,$68,$DE
	.dc.b	$92,$B6,$7F,$88,$E7,$6E,$03,$87,$EB,$1B,$8E,$4E,$0D,$CF,$77,$94
	.dc.b	$F9,$5F,$E8,$9F,$2F,$FD,$3B,$D3,$91,$FD,$26,$3F,$A8,$FC,$BF,$F5
	.dc.b	$2F,$98,$93,$FD,$36,$3F,$AA,$F8,$52,$3F,$A8,$3F,$D6,$3E,$17,$FA
	.dc.b	$FE,$EF,$98,$A4,$AF,$EA,$DE,$B6,$BE,$63,$7F,$0D,$27,$FA,$ED,$A7
	.dc.b	$9F,$CB,$74,$EB,$95,$A8,$EF,$F1,$72,$38,$BE,$2F,$97,$F1,$FC,$BF
	.dc.b	$93,$E2,$25,$76,$8F,$3F,$5C,$8F,$0B,$EA,$F8,$DF,$5F,$8F,$27,$CA
	.dc.b	$FB,$BE,$5F,$DF,$E9,$FC,$0F,$97,$95,$EA,$8F,$87,$F1,$92,$3D,$AF
	.dc.b	$C6,$F9,$7F,$8F,$F2,$F2,$7E,$0B,$F2,$BC,$CF,$97,$F1,$BF,$33,$C2
	.dc.b	$95,$F1,$A3,$FC,$1B,$F3,$6F,$9C,$A4,$9F,$97,$5A,$F9,$9D,$FC,$CC
	.dc.b	$9F,$F0,$5D,$A7,$11,$F8,$EF,$25,$2F,$99,$EE,$6B,$CA,$63,$FE,$7B
	.dc.b	$AD,$CF,$93,$E6,$66,$1D,$A3,$CF,$F3,$1E,$8F,$72,$4F,$8A,$32,$F9
	.dc.b	$23,$2F,$8F,$24,$EA,$31,$3A,$CB,$C7,$0C,$7D,$35,$F1,$BD,$5F,$8F
	.dc.b	$EA,$4A,$F8,$31,$97,$E1,$C6,$5F,$04,$7C,$EF,$62,$47,$C8,$7E,$8F
	.dc.b	$C9,$E2,$3C,$DF,$27,$F5,$0C,$65,$E4,$CB,$E5,$FC,$9F,$FB,$A3,$CC
	.dc.b	$F9,$1B,$3E,$4F,$18,$FA,$63,$0F,$FA,$9F,$E4,$15,$C2,$F7,$87,$4A
	.dc.b	$D4,$F4,$C3,$38,$72,$96,$1C,$9E,$D8,$78,$17,$CD,$BB,$3F,$D8,$31
	.dc.b	$C3,$1F,$F5,$84,$E7,$CC,$EF,$3F,$EB,$B8,$C3,$B3,$1E,$57,$58,$CE
	.dc.b	$3E,$19,$EF,$87,$68,$C4,$EB,$0E,$F7,$C6,$B5,$1E,$26,$DF,$85,$AF
	.dc.b	$AA,$78,$61,$E6,$8E,$ED,$BE,$D1,$E6,$31,$5C,$A3,$31,$C0,$F6,$1D
	.dc.b	$4D,$CF,$F8,$42,$3E,$0C,$7F,$51,$6B,$E7,$8F,$F0,$AF,$C8,$BD,$A7
	.dc.b	$CE,$61,$B5,$FD,$64,$FF,$6E,$B8,$7F,$AE,$4F,$19,$F0,$77,$BE,$3A
	.dc.b	$92,$B9,$62,$9D,$46,$1B,$AF,$D0,$6E,$BF,$EE,$DF,$93,$72,$7E,$B2
	.dc.b	$B5,$8F,$C9,$BD,$E1,$F8,$16,$AB,$E3,$DB,$BA,$93,$F2,$B2,$F2,$47
	.dc.b	$D5,$AD,$CE,$6F,$9D,$6A,$31,$5B,$C7,$C9,$97,$E5,$2F,$64,$9D,$63
	.dc.b	$E0,$CB,$F4,$1C,$9F,$C7,$3E,$4F,$8E,$5F,$AF,$D2,$CE,$6F,$F3,$AF
	.dc.b	$F2,$28,$CF,$22,$EC,$61,$B5,$C3,$E2,$BF,$D6,$2F,$F2,$EC,$BF,$8A
	.dc.b	$F3,$FF,$03,$9C,$E5,$FC,$76,$73,$5F,$CA,$E7,$35,$FC,$27,$1F,$BD
	.dc.b	$3F,$EC,$29,$CB,$FC,$2A,$B8,$DF,$F2,$19,$FF,$45,$4E,$FE,$1B,$39
	.dc.b	$A7,$FD,$9D,$39,$25,$7F,$41,$8E,$2D,$F8,$7E,$31,$8C,$BD,$5F,$1F
	.dc.b	$83,$3F,$0D,$70,$F3,$45,$C3,$A4,$DA,$E3,$5A,$3D,$91,$97,$E1,$7C
	.dc.b	$7F,$80,$F0,$FC,$7D,$8F,$84,$E1,$F8,$9F,$0F,$70,$F5,$3E,$83,$D3
	.dc.b	$F1,$F6,$7E,$57,$C1,$FB,$0F,$6F,$C2,$58,$F9,$DF,$0B,$F0,$0F,$83
	.dc.b	$DB,$63,$E9,$7C,$27,$C4,$3E,$2F,$BD,$63,$EB,$7C,$1F,$C8,$3E,$4F
	.dc.b	$C2,$58,$FB,$5F,$07,$F3,$0F,$9B,$DD,$63,$EF,$78,$7F,$43,$E1,$3E
	.dc.b	$8B,$63,$FC,$93,$D2,$FC,$F5,$F0,$9F,$79,$FF,$19,$4E,$9C,$B7,$E1
	.dc.b	$75,$F0,$B2,$35,$E5,$5E,$7F,$F0,$15,$BE,$0F,$A1,$D7,$E1,$6D,$F5
	.dc.b	$F8,$4F,$01,$E1,$F8,$5B,$1E,$1F,$83,$F2,$1E,$5F,$85,$B1,$E5,$F8
	.dc.b	$3F,$CA,$F9,$0F,$96,$F5,$46,$1F,$67,$C2,$6B,$E5,$6D,$FF,$86,$9F
	.dc.b	$81,$F0,$92,$3E,$07,$C2,$DE,$1F,$5F,$A3,$F1,$4F,$8D,$EB,$5A,$F8
	.dc.b	$DF,$09,$F2,$4F,$95,$F0,$B6,$32,$4B,$F8,$B1,$97,$E3,$CD,$F3,$3F
	.dc.b	$C0,$1C,$CA,$C5,$AE,$7C,$E3,$E1,$7F,$73,$83,$E8,$1E,$6F,$EF,$2D
	.dc.b	$7B,$C3,$FD,$13,$7B,$AD,$4E,$6B,$FA,$53,$5F,$F5,$23,$5F,$11,$D3
	.dc.b	$0F,$F1,$4B,$8E,$FF,$A5,$F0,$FF,$8B,$C7,$F8,$4F,$27,$F9,$3D,$FC
	.dc.b	$61,$88,$D3,$F9,$48,$D5,$7E,$2A,$38,$DF,$2F,$87,$B1,$B6,$CF,$F9
	.dc.b	$46,$BF,$CA,$93,$CE,$B8,$34,$E0,$59,$E1,$5E,$6F,$85,$F3,$FC,$3D
	.dc.b	$AF,$23,$6B,$B4,$FB,$3E,$1F,$1B,$7F,$8D,$F2,$7C,$37,$C0,$3E,$0F
	.dc.b	$C3,$DC,$BD,$D1,$71,$F4,$4D,$BF,$E7,$FF,$39,$FE,$DF,$BF,$AE,$DA
	.dc.b	$FF,$5D,$CA,$3F,$D8,$F3,$AF,$4B,$F0,$71,$B9,$E7,$18,$F4,$F2,$CA
	.dc.b	$E0,$76,$8C,$5F,$CB,$F8,$6F,$B3,$F1,$1F,$69,$B5,$F5,$DB,$3B,$8F
	.dc.b	$B5,$3F,$91,$F3,$DE,$1F,$E0,$BE,$0B,$0D,$AF,$F6,$C4,$E9,$B1,$F8
	.dc.b	$8F,$88,$DC,$7E,$63,$E7,$1D,$DF,$A0,$6C,$67,$B7,$13,$23,$83,$B8
	.dc.b	$C6,$7F,$2A,$B8,$B6,$75,$5C,$0F,$F7,$67,$C1,$7E,$53,$83,$EF,$7C
	.dc.b	$4F,$EB,$63,$13,$FE,$9B,$E5,$FF,$3A,$3C,$3A,$F4,$7E,$8B,$E2,$6C
	.dc.b	$7E,$7F,$E6,$BF,$4B,$5F,$A6,$6C,$64,$99,$FE,$77,$C7,$D4,$7E,$9F
	.dc.b	$E1,$BF,$5A,$7E,$BB,$E5,$6C,$64,$95,$FA,$58,$CB,$FA,$7F,$87,$FD
	.dc.b	$A9,$FB,$6D,$2D,$7E,$CE,$2C,$FE,$B6,$6E,$DF,$E9,$5F,$CF,$BF,$C1
	.dc.b	$73,$FE,$A9,$6F,$87,$FC,$CE,$3F,$C1,$31,$82,$7C,$CF,$F0,$48,$CB
	.dc.b	$AF,$89,$99,$7F,$56,$AF,$F0,$7D,$7F,$B3,$5F,$F0,$54,$7F,$B2,$5B
	.dc.b	$1E,$19,$EF,$F1,$3C,$A3,$87,$C4,$78,$8F,$1F,$C4,$5C,$3B,$45,$9E
	.dc.b	$D3,$73,$FE,$3B,$1F,$EA,$B1,$F3,$6E,$1F,$E4,$72,$F8,$4F,$F6,$5C
	.dc.b	$E6,$BF,$93,$9F,$EA,$A3,$8D,$70,$9D,$BF,$E7,$9F,$8B,$FE,$7D,$5C
	.dc.b	$69,$7D,$72,$CD,$EE,$79,$9F,$D1,$67,$4E,$0F,$E6,$F7,$C1,$D6,$1E
	.dc.b	$07,$CB,$19,$7B,$C6,$5E,$27,$39,$EB,$87,$96,$3E,$23,$E7,$8C,$DF
	.dc.b	$96,$31,$5C,$EB,$73,$F0,$70,$F0,$46,$0F,$2B,$87,$E4,$57,$28,$C6
	.dc.b	$1E,$7C,$3B,$D7,$A8,$F3,$61,$DD,$E3,$87,$77,$E2,$1D,$CF,$7D,$F4
	.dc.b	$78,$B7,$0F,$04,$70,$39,$9E,$63,$18,$78,$EF,$C0,$DA,$E5,$7E,$1B
	.dc.b	$F9,$17,$C8,$F0,$9C,$8F,$09,$5F,$09,$C2,$BA,$46,$67,$A1,$F3,$CA
	.dc.b	$FC,$3F,$80,$7C,$0F,$E1,$DB,$19,$8C,$1F,$DF,$3E,$39,$F1,$1F,$90
	.dc.b	$95,$70,$53,$AD,$3F,$C2,$93,$98,$C4,$F0,$BE,$97,$CE,$B4,$F5,$CB
	.dc.b	$F3,$D8,$FF,$A9,$6F,$F2,$4D,$7F,$35,$87,$49,$FD,$23,$67,$FC,$26
	.dc.b	$6E,$39,$F2,$7F,$85,$9D,$DF,$79,$F1,$E5,$DA,$B5,$1C,$9B,$87,$F8
	.dc.b	$BF,$0E,$CD,$BE,$4F,$F8,$CE,$BC,$07,$17,$F5,$53,$C1,$F8,$F3,$C1
	.dc.b	$FE,$11,$5D,$9A,$70,$8D,$4F,$1A,$EB,$3D,$EB,$84,$F5,$34,$F3,$9D
	.dc.b	$CF,$82,$FA,$1E,$F6,$1F,$32,$FB,$9E,$59,$E2,$72,$BE,$D1,$9B,$EC
	.dc.b	$78,$9E,$37,$D3,$0F,$A3,$87,$80,$ED,$19,$C3,$E3,$46,$09,$3C,$1F
	.dc.b	$E3,$11,$E1,$AE,$B1,$8B,$F0,$BD,$6B,$F9,$77,$23,$8F,$FC,$C3,$3F
	.dc.b	$D2,$7D,$1F,$E6,$BB,$FF,$47,$6C,$7F,$57,$F8,$BB,$7C,$66,$DF,$2A
	.dc.b	$FE,$A1,$18,$AC,$FC,$5C,$9F,$EA,$CF,$0F,$8B,$C4,$7F,$57,$8B,$B7
	.dc.b	$F6,$79,$EB,$E5,$CA,$FE,$BC,$F8,$3C,$BC,$47,$F5,$F8,$B5,$FD,$BE
	.dc.b	$7C,$BB,$CA,$FE,$CC,$FA,$3E,$2F,$11,$FD,$9E,$2D,$7F,$77,$9F,$6F
	.dc.b	$34,$AF,$ED,$CF,$C0,$F8,$BC,$47,$F6,$F8,$B5,$FD,$FE,$7E,$2F,$C4
	.dc.b	$CA,$FE,$EC,$FC,$8F,$8A,$C4,$7F,$77,$F8,$BF,$96,$62,$FF,$D0,$FF
	.dc.b	$9F,$BE,$43,$E9,$D2,$30,$FF,$5B,$E5,$FF,$47,$C6,$23,$FB,$1C,$7F
	.dc.b	$A4,$BC,$77,$B3,$CA,$B5,$87,$FA,$53,$EA,$FF,$D2,$AD,$BD,$63,$E0
	.dc.b	$F9,$A7,$A5,$49,$9F,$84,$FF,$58,$43,$6C,$C7,$88,$CC,$E3,$0F,$F5
	.dc.b	$1D,$6E,$79,$07,$BB,$86,$BC,$CD,$4F,$96,$BE,$2F,$CD,$3D,$AA,$4A
	.dc.b	$EB,$F3,$5F,$EC,$77,$BC,$ED,$CF,$C1,$7C,$E9,$3C,$63,$FC,$59,$E2
	.dc.b	$7E,$83,$19,$1F,$E3,$18,$FD,$19,$87,$FC,$65,$18,$7D,$BF,$34,$F3
	.dc.b	$FE,$DB,$63,$31,$8F,$13,$2C,$8C,$9C,$A3,$0F,$C3,$9E,$4D,$9E,$4F
	.dc.b	$D5,$F0,$9F,$A0,$3E,$4B,$F7,$6D,$9D,$35,$FF,$25,$79,$1A,$99,$1A
	.dc.b	$7F,$80,$C7,$2F,$92,$93,$C9,$EB,$F2,$78,$8F,$C1,$61,$E2,$F9,$3F
	.dc.b	$09,$9E,$57,$3F,$FB,$E6,$7C,$FF,$27,$E6,$F9,$39,$87,$9A,$3D,$26
	.dc.b	$1E,$BA,$A7,$31,$B3,$DA,$EF,$0E,$F0,$BE,$EF,$C2,$F9,$3F,$8B,$F2
	.dc.b	$72,$30,$43,$BB,$97,$C3,$1F,$B6,$F9,$97,$CC,$7F,$97,$24,$48,$CB
	.dc.b	$90,$E1,$E0,$D7,$FC,$CA,$E7,$0F,$4C,$3E,$67,$F5,$4F,$E9,$5B,$1A
	.dc.b	$9E,$1A,$F8,$AB,$87,$8D,$FC,$32,$FF,$CE,$91,$FB,$26,$BC,$FC,$6D
	.dc.b	$4C,$8D,$3C,$63,$11,$E1,$8B,$87,$9A,$95,$F9,$0F,$96,$97,$F9,$07
	.dc.b	$F2,$7F,$09,$F9,$4F,$95,$93,$F9,$47,$F2,$FA,$7E,$63,$E5,$7F,$A9
	.dc.b	$FC,$B7,$E8,$33,$FB,$7C,$28,$C6,$DF,$8B,$E3,$5F,$E8,$3E,$5B,$8F
	.dc.b	$C0,$4D,$3F,$43,$1F,$A4,$F9,$59,$1F,$A4,$7F,$4F,$E1,$7E,$A3,$E1
	.dc.b	$A4,$FE,$A1,$FD,$5F,$C4,$7F,$66,$FD,$05,$F0,$F4,$95,$FA,$ED,$75
	.dc.b	$F2,$72,$3F,$6D,$48,$E5,$E8,$73,$F2,$FA,$7C,$9C,$BD,$C7,$7F,$91
	.dc.b	$F0,$7C,$9C,$9E,$71,$E3,$F9,$09,$1D,$9F,$37,$C9,$F9,$FE,$46,$4F
	.dc.b	$89,$F5,$7C,$97,$AF,$E4,$24,$F9,$9F,$77,$C9,$7B,$FE,$46,$4F,$A9
	.dc.b	$F8,$5E,$DF,$C3,$F4,$A4,$FB,$9F,$8D,$E8,$FC,$7F,$91,$93,$F0,$9F
	.dc.b	$95,$D9,$F2,$FE,$26,$4F,$C6,$76,$FF,$0C,$E6,$7C,$8F,$D1,$24,$6C
	.dc.b	$FE,$45,$F2,$EF,$92,$FF,$0D,$B6,$33,$1F,$65,$CB,$1C,$9F,$58,$C1
	.dc.b	$FC,$72,$39,$36,$39,$3E,$0F,$92,$EE,$47,$C7,$F2,$55,$E5,$17,$C7
	.dc.b	$FC,$DD,$AB,$CF,$F1,$7E,$8F,$92,$98,$7B,$23,$2F,$BA,$32,$F9,$63
	.dc.b	$DF,$F1,$7F,$ED,$EF,$81,$FE,$5D,$87,$F8,$E7,$E4,$7F,$B9,$79,$88
	.dc.b	$C4,$EA,$B8,$61,$FC,$9E,$51,$8A,$E1,$F1,$D6,$78,$D6,$EB,$8A,$F9
	.dc.b	$9F,$ED,$BE,$F7,$F2,$4C,$BF,$8C,$C6,$25,$18,$F2,$BF,$D5,$B1,$8C
	.dc.b	$FF,$8D,$53,$DC,$EA,$F1,$BE,$18,$75,$AD,$E1,$CA,$7B,$DF,$08,$C4
	.dc.b	$EA,$B8,$61,$FC,$E3,$0E,$93,$E0,$BE,$15,$AA,$F6,$5F,$27,$83,$5F
	.dc.b	$81,$3C,$1F,$51,$A9,$E5,$7D,$A3,$C1,$3B,$7F,$9C,$1D,$A7,$FA,$44
	.dc.b	$FC,$78,$CB,$5F,$F6,$B7,$C5,$BE,$6B,$FA,$77,$C9,$A3,$13,$FD,$56
	.dc.b	$31,$3A,$9C,$CF,$4F,$3F,$F5,$AF,$F5,$BB,$46,$23,$5C,$FE,$48,$C4
	.dc.b	$7D,$67,$8E,$BF,$79,$BA,$FF,$53,$EB,$78,$7F,$DD,$2B,$58,$7C,$4B
	.dc.b	$DE,$1F,$4D,$AA,$F8,$F9,$9B,$A9,$3F,$38,$FC,$85,$70,$AD,$C6,$72
	.dc.b	$F5,$D7,$0C,$3C,$33,$C6,$B7,$18,$26,$5E,$DA,$FC,$14,$EA,$31,$5F
	.dc.b	$7B,$93,$FA,$A7,$67,$B6,$7F,$C3,$DE,$9F,$A7,$4F,$C7,$D6,$9B,$19
	.dc.b	$BE,$4A,$F9,$47,$F8,$66,$35,$97,$EE,$9D,$57,$2C,$77,$83,$E6,$BF
	.dc.b	$91,$4F,$F9,$16,$BF,$94,$BF,$E4,$78,$CE,$1F,$E3,$89,$CB,$FE,$3A
	.dc.b	$8C,$F5,$FF,$70,$6D,$7B,$27,$FC,$5B,$39,$ED,$B3,$FE,$3D,$A7,$6E
	.dc.b	$2F,$FC,$65,$18,$9F,$17,$6E,$72,$F4,$5D,$AD,$5F,$8E,$31,$59,$75
	.dc.b	$DB,$30,$F6,$7C,$AD,$BF,$54,$DC,$FD,$53,$6B,$CD,$36,$BC,$B3,$72
	.dc.b	$F9,$5D,$DE,$33,$C9,$DD,$70,$F9,$4E,$1F,$99,$F2,$77,$0F,$94,$FA
	.dc.b	$CF,$67,$75,$9F,$A5,$F2,$7E,$FF,$93,$F8,$0D,$8F,$AD,$F2,$7F,$0C
	.dc.b	$F8,$9D,$96,$3E,$D7,$C0,$7C,$73,$E4,$7B,$F6,$3E,$F7,$C9,$7C,$B3
	.dc.b	$E6,$7C,$9D,$8F,$C1,$7C,$67,$CF,$3E,$87,$6D,$8F,$C3,$78,$9F,$4F
	.dc.b	$E3,$7E,$A3,$63,$FD,$21,$F3,$4F,$8B,$FB,$4F,$FA,$46,$33,$18,$F8
	.dc.b	$DC,$FC,$6C,$9E,$3E,$55,$E6,$FF,$44,$E0,$F8,$DE,$67,$4F,$8D,$B7
	.dc.b	$E0,$F8,$CE,$E7,$83,$E3,$6C,$79,$3E,$2F,$C6,$79,$3E,$2E,$C7,$A3
	.dc.b	$E2,$FC,$E7,$A3,$E2,$EC,$7B,$3E,$2B,$D6,$7B,$3E,$4A,$C7,$C0,$F8
	.dc.b	$BF,$79,$F0,$3D,$7B,$1F,$13,$D3,$F8,$67,$C4,$F5,$EC,$7C,$8F,$8B
	.dc.b	$F8,$E7,$C8,$F8,$BB,$19,$27,$27,$C4,$8C,$BF,$1A,$6F,$99,$FE,$64
	.dc.b	$E5,$56,$2F,$FD,$05,$F3,$3F,$E4,$F7,$FD,$24,$DC,$3F,$D3,$0D,$7F
	.dc.b	$A8,$B5,$E3,$F1,$7F,$B9,$51,$C7,$0F,$E0,$3F,$30,$EA,$D4,$E7,$0F
	.dc.b	$F5,$F7,$C1,$7F,$99,$9F,$C3,$9F,$BF,$71,$87,$E0,$23,$1F,$1B,$F8
	.dc.b	$A6,$9C,$0F,$F6,$F4,$F5,$9F,$F5,$E2,$F8,$DE,$99,$F5,$6D,$70,$6D
	.dc.b	$7F,$B5,$AB,$84,$F4,$AE,$0D,$38,$1E,$0F,$8B,$F4,$7C,$6F,$A5,$B1
	.dc.b	$92,$5F,$5F,$8D,$D4,$7A,$3E,$2F,$E0,$1F,$07,$E3,$6C,$64,$95,$E5
	.dc.b	$8C,$AE,$0F,$E5,$FF,$32,$F4,$3F,$96,$42,$31,$7F,$ED,$28,$DF,$A3
	.dc.b	$A9,$CD,$F3,$AD,$46,$2B,$AD,$EE,$F8,$32,$BA,$1C,$1D,$38,$BF,$35
	.dc.b	$F3,$8F,$C0,$FC,$6F,$DF,$F8,$EF,$C0,$36,$3E,$DB,$6B,$EA,$37,$1F
	.dc.b	$F7,$0F,$9A,$F0,$BF,$11,$F0,$BF,$93,$8C,$61,$F8,$48,$C4,$C8,$FC
	.dc.b	$3F,$C6,$DC,$7E,$01,$F2,$FA,$5D,$9F,$29,$B5,$EB,$C9,$1F,$AC,$8C
	.dc.b	$3A,$AE,$73,$AA,$E4,$73,$EB,$B7,$C2,$2D,$70,$BE,$8D,$8F,$D7,$7C
	.dc.b	$27,$E9,$30,$FF,$6F,$F4,$3E,$3B,$3D,$7F,$A9,$6E,$5F,$47,$1F,$F0
	.dc.b	$44,$FE,$B7,$E3,$B3,$1F,$9F,$F8,$DF,$D8,$61,$FB,$16,$DF,$EB,$1C
	.dc.b	$3F,$AE,$F8,$EB,$5F,$A7,$F8,$7F,$DC,$1F,$B9,$F8,$EB,$5F,$B4,$8B
	.dc.b	$3F,$AD,$9B,$A7,$AB,$E2,$BF,$80,$1F,$C0,$76,$B8,$7E,$F2,$2D,$FE
	.dc.b	$D6,$6D,$7F,$C0,$0F,$98,$FF,$08,$CB,$E0,$D1,$85,$AA,$7F,$C1,$93
	.dc.b	$92,$7C,$4F,$F1,$18,$CB,$AF,$87,$99,$FE,$6F,$0E,$BF,$11,$8B,$FC
	.dc.b	$4C,$78,$2B,$19,$79,$6A,$DF,$86,$BC,$BF,$11,$E2,$3C,$7F,$11,$67
	.dc.b	$B6,$69,$C6,$5D,$A1,$58,$C5,$7F,$E8,$6A,$FF,$54,$8F,$99,$71,$7C
	.dc.b	$18,$FE,$11,$F0,$7C,$1D,$47,$1F,$8C,$F8,$D5,$C2,$7F,$84,$56,$E3
	.dc.b	$11,$CE,$77,$5D,$70,$5E,$E3,$A5,$76,$39,$E1,$C1,$FF,$4E,$DD,$7C
	.dc.b	$B1,$5F,$F6,$31,$C4,$C9,$4C,$E1,$E4,$C3,$C3,$3C,$67,$E1,$CE,$70
	.dc.b	$F2,$47,$80,$E9,$5C,$1E,$87,$0C,$3D,$93,$C2,$3E,$01,$C1,$C3,$E2
	.dc.b	$8D,$DF,$87,$0F,$65,$F5,$C3,$E4,$DE,$CF,$86,$78,$27,$78,$74,$8F
	.dc.b	$11,$A9,$E2,$E0,$F8,$6E,$27,$8D,$F8,$67,$CC,$D9,$F8,$4D,$C3,$CF
	.dc.b	$3E,$13,$65,$9D,$B6,$37,$1D,$E3,$6D,$36,$58,$D9,$F8,$27,$2F,$E1
	.dc.b	$3D,$F7,$C0,$52,$57,$D5,$8C,$BF,$B9,$F8,$F7,$C3,$FF,$61,$84,$E9
	.dc.b	$4E,$B4,$FE,$01,$39,$8C,$4F,$0B,$FD,$24,$62,$75,$7D,$71,$FC,$FE
	.dc.b	$3F,$EB,$1C,$3F,$2C,$D7,$F2,$97,$D2,$7F,$4A,$D9,$FD,$81,$B8,$E7
	.dc.b	$CB,$FB,$37,$77,$DE,$7C,$6D,$BF,$E3,$11,$C4,$E7,$5C,$5D,$E5,$FA
	.dc.b	$9C,$3C,$15,$C2,$B4,$D3,$57,$D5,$F8,$ED,$BE,$D1,$B3,$DB,$1B,$AE
	.dc.b	$B1,$FC,$32,$BA,$B4,$EB,$71,$FD,$C5,$F4,$74,$74,$8D,$61,$F1,$23
	.dc.b	$E6,$D7,$4B,$EE,$78,$2F,$91,$CD,$CE,$1E,$99,$F1,$CE,$63,$D0,$66
	.dc.b	$BA,$9E,$08,$F9,$2F,$9A,$78,$9E,$19,$E3,$87,$09,$D1,$C4,$FC,$A4
	.dc.b	$75,$72,$7B,$4F,$E4,$1C,$3F,$D5,$3E,$2B,$F9,$36,$BF,$D0,$F8,$51
	.dc.b	$8A,$FE,$B5,$18,$7F,$A1,$FC,$5E,$6F,$94,$7F,$46,$AC,$56,$E3,$FA
	.dc.b	$35,$63,$E2,$E4,$7F,$4C,$77,$F1,$78,$8F,$E9,$91,$73,$FE,$D9,$3D
	.dc.b	$3C,$B9,$5F,$D5,$1E,$FE,$66,$23,$FA,$A4,$5A,$FE,$E9,$3E,$4D,$65
	.dc.b	$7F,$5C,$7C,$FF,$17,$88,$FE,$B9,$16,$BF,$BE,$4F,$B3,$9A,$57,$F6
	.dc.b	$47,$DF,$F1,$78,$8F,$EC,$91,$6B,$FC,$09,$3F,$13,$E2,$65,$7F,$6C
	.dc.b	$7E,$3F,$C5,$62,$3F,$B6,$7C,$5F,$CA,$31,$7F,$D9,$61,$F3,$2F,$81
	.dc.b	$FA,$D1,$86,$C6,$63,$FC,$61,$E3,$FF,$A4,$5F,$F1,$94,$65,$FF,$27
	.dc.b	$8F,$99,$ED,$7D,$DB,$CF,$FD,$DE,$B8,$4F,$8F,$E6,$5E,$97,$89,$B8
	.dc.b	$7F,$8F,$DA,$F6,$5A,$EE,$64,$ED,$E5,$18,$8E,$11,$6B,$C1,$3F,$03
	.dc.b	$E6,$5E,$CD,$25,$FF,$93,$FE,$67,$FD,$9E,$FF,$94,$67,$6E,$7E,$0B
	.dc.b	$E4,$C9,$E3,$1F,$E1,$DF,$0F,$F3,$79,$48,$FF,$11,$C7,$E7,$8C,$3F
	.dc.b	$E2,$38,$C3,$FE,$5F,$F9,$97,$9F,$F5,$9B,$19,$8C,$78,$59,$64,$64
	.dc.b	$E5,$18,$7E,$BC,$F2,$6C,$F2,$7F,$C8,$9F,$2E,$F9,$1F,$DB,$36,$33
	.dc.b	$1F,$B6,$F1,$F2,$C8,$C9,$FE,$40,$8E,$3F,$09,$23,$AB,$D3,$E4,$B1
	.dc.b	$1F,$E5,$0B,$7C,$97,$80,$CF,$1F,$F8,$F7,$C7,$4B,$E4,$BC,$BF,$25
	.dc.b	$30,$F4,$C7,$A0,$C3,$D6,$30,$E8,$F7,$BA,$BE,$D5,$D6,$7E,$1F,$C9
	.dc.b	$7C,$2F,$92,$91,$82,$1E,$97,$2F,$AE,$3F,$5B,$F3,$2F,$91,$A4,$9F
	.dc.b	$F4,$E7,$CC,$FF,$7F,$C7,$F6,$D5,$B9,$FF,$31,$07,$0F,$F9,$95,$AF
	.dc.b	$F9,$8C,$D5,$70,$55,$FE,$5A,$7B,$F8,$BF,$C0,$23,$11,$FB,$A5,$E2
	.dc.b	$E6,$64,$F0,$3C,$F1,$87,$84,$5C,$3F,$87,$57,$29,$FC,$5F,$CA,$7E
	.dc.b	$28,$8F,$E3,$BE,$0F,$F1,$BF,$29,$23,$F2,$2F,$E4,$32,$FC,$8F,$CA
	.dc.b	$49,$FC,$AB,$F9,$4F,$1F,$F2,$BF,$29,$27,$F3,$2F,$E6,$3E,$53,$F3
	.dc.b	$3F,$29,$27,$F3,$AF,$E7,$3C,$1F,$CE,$FC,$2C,$9F,$D0,$BF,$A0,$F8
	.dc.b	$7F,$EA,$DF,$32,$F8,$6A,$4A,$FD,$3F,$99,$AF,$92,$91,$FE,$AD,$A4
	.dc.b	$72,$F9,$3E,$7E,$5F,$4F,$93,$97,$B8,$EF,$F2,$3E,$0F,$93,$93,$CE
	.dc.b	$3C,$7F,$23,$23,$B3,$E6,$F9,$3F,$3F,$C8,$C9,$F1,$3E,$AF,$92,$F5
	.dc.b	$FC,$84,$9F,$33,$EE,$F9,$2F,$7F,$C7,$49,$F5,$3F,$0B,$DA,$F8,$7E
	.dc.b	$87,$C4,$F4,$A5,$7B,$E3,$E3,$FC,$8C,$8F,$84,$FC,$AF,$91,$F9,$7F
	.dc.b	$11,$27,$E3,$3B,$7F,$DA,$3C,$CF,$91,$FA,$24,$8D,$9F,$ED,$9E,$E7
	.dc.b	$C8,$FF,$82,$DB,$19,$8F,$B2,$E5,$8E,$4F,$AC,$60,$FF,$70,$C7,$26
	.dc.b	$C7,$27,$C1,$F2,$3D,$C8,$F8,$FE,$46,$BC,$A2,$F9,$5F,$C6,$6D,$5E
	.dc.b	$7F,$8A,$97,$EB,$7D,$5F,$21,$88,$F1,$FC,$8F,$B4,$C6,$1C,$FE,$2F
	.dc.b	$FD,$B1,$F0,$5F,$EF,$3C,$BF,$C5,$1F,$23,$FD,$9B,$07,$B2,$9C,$D6
	.dc.b	$EF,$8D,$6A,$51,$8A,$E1,$F1,$F6,$78,$D6,$EB,$8A,$F9,$8F,$EC,$5E
	.dc.b	$FA,$19,$78,$DD,$79,$7F,$E9,$E7,$39,$FF,$BE,$EF,$8C,$60,$E4,$FF
	.dc.b	$44,$7A,$57,$38,$DB,$6B,$A3,$6F,$77,$D7,$DD,$F3,$DF,$08,$F0,$5F
	.dc.b	$06,$D7,$0C,$3C,$37,$C2,$B5,$1E,$46,$BE,$C9,$E0,$FC,$39,$E1,$18
	.dc.b	$BF,$04,$F1,$8C,$4E,$A3,$F9,$D1,$CC,$B3,$F1,$1F,$E7,$CD,$7F,$D7
	.dc.b	$FF,$16,$F9,$9F,$E8,$FF,$24,$8C,$4F,$48,$C4,$EA,$73,$3D,$3D,$0F
	.dc.b	$D8,$3E,$06,$B9,$E8,$FE,$A7,$82,$73,$3F,$65,$E3,$AF,$D5,$6E,$BF
	.dc.b	$83,$97,$63,$C4,$79,$BF,$16,$F7,$5F,$80,$6B,$F8,$D6,$D7,$19,$E1
	.dc.b	$94,$AF,$C4,$1B,$7E,$9C,$EE,$33,$97,$A6,$78,$65,$E7,$77,$8F,$92
	.dc.b	$B7,$8F,$8A,$F8,$46,$2B,$78,$F2,$CB,$DD,$93,$C1,$E5,$9F,$E8,$63
	.dc.b	$15,$E4,$F0,$B3,$7C,$2E,$E3,$FA,$9B,$DB,$EC,$6E,$1C,$EF,$73,$F3
	.dc.b	$2B,$8F,$33,$EE,$9E,$A7,$AF,$A7,$FC,$55,$0F,$82,$C6,$BF,$B7,$BE
	.dc.b	$17,$8C,$7F,$84,$DE,$EB,$F7,$98,$72,$AD,$5F,$7C,$2D,$F8,$2F,$67
	.dc.b	$57,$96,$1F,$E8,$8F,$C1,$7F,$C9,$D3,$A8,$FC,$D7,$C5,$BC,$BF,$14
	.dc.b	$2B,$6B,$3F,$AD,$CC,$BE,$18,$73,$AE,$1F,$0D,$FE,$A1,$7F,$25,$87
	.dc.b	$27,$FC,$50,$D7,$31,$B7,$F8,$D4,$7A,$30,$EF,$7F,$8B,$7F,$D5,$91
	.dc.b	$9B,$FE,$41,$87,$8F,$9B,$D3,$39,$C3,$DB,$18,$D3,$E7,$56,$EF,$D1
	.dc.b	$7B,$BF,$81,$5A,$C1,$E6,$F1,$BF,$91,$83,$E4,$3F,$D5,$71,$FD,$97
	.dc.b	$E4,$FA,$C6,$31,$FE,$BB,$1E,$4F,$73,$3F,$15,$6F,$9D,$EF,$2F,$6D
	.dc.b	$67,$0F,$A1,$5E,$EC,$5F,$21,$9B,$FC,$6F,$87,$F8,$41,$F2,$5F,$84
	.dc.b	$70,$FC,$4F,$7B,$E0,$FD,$67,$BF,$87,$FD,$A9,$97,$B2,$3E,$46,$5F
	.dc.b	$D2,$FE,$33,$84,$FC,$0D,$11,$87,$CA,$7F,$60,$FB,$7F,$F0,$54,$FF
	.dc.b	$2C,$E9,$46,$27,$FC,$D1,$18,$F0,$3F,$00,$DA,$FF,$38,$FE,$08,$F5
	.dc.b	$BE,$BF,$CA,$7C,$59,$FE,$39,$39,$7E,$9C,$67,$0F,$EE,$3F,$7B,$FE
	.dc.b	$42,$9F,$13,$63,$C2,$D7,$F9,$36,$08,$C4,$7E,$A7,$C6,$FD,$69,$FA
	.dc.b	$EF,$2E,$DF,$E7,$F1,$7B,$3F,$8C,$F0,$FF,$15,$18,$9F,$C3,$C7,$E4
	.dc.b	$F2,$FE,$0D,$F0,$7F,$B8,$3F,$73,$F0,$76,$BF,$69,$2F,$C0,$FF,$CC
	.dc.b	$51,$8F,$17,$FD,$6E,$EA,$33,$AF,$D2,$AF,$E2,$5E,$77,$EE,$B3,$7B
	.dc.b	$3F,$85,$F9,$8C,$37,$1E,$5F,$05,$F8,$9F,$43,$F8,$6E,$68,$C4,$F7
	.dc.b	$F8,$1F,$E1,$FD,$3A,$6C,$7C,$3E,$8F,$E2,$BE,$7B,$E0,$FC,$51,$87
	.dc.b	$AF,$9B,$F9,$7F,$1A,$DF,$F9,$35,$FE,$23,$F3,$BF,$5F,$C5,$FE,$55
	.dc.b	$F4,$2F,$63,$3F,$0D,$FC,$9E,$53,$9C,$3F,$89,$5F,$17,$F9,$4C,$67
	.dc.b	$9F,$F2,$E2,$74,$FF,$1D,$8D,$57,$F5,$38,$D5,$FF,$15,$6D,$FF,$B1
	.dc.b	$AF,$F0,$F1,$FE,$31,$6F,$3F,$D9,$F8,$4F,$96,$7C,$CF,$3A,$DF,$D9
	.dc.b	$8B,$7F,$67,$93,$F8,$7C,$E4,$FB,$DE,$1F,$D4,$3E,$AF,$A1,$63,$24
	.dc.b	$AE,$51,$97,$F6,$BA,$3E,$77,$FD,$F5,$5F,$D0,$63,$1F,$2B,$F2,$E5
	.dc.b	$F2,$B6,$25,$6E,$78,$DF,$47,$71,$99,$DC,$E5,$E9,$58,$BF,$99,$B3
	.dc.b	$E1,$BE,$D7,$D6,$7F,$3D,$C7,$FC,$31,$39,$7C,$D1,$9C,$7E,$75,$7D
	.dc.b	$3A,$D4,$7F,$8E,$9A,$FF,$BD,$5A,$FF,$6F,$D7,$FB,$EE,$1F,$E0,$FE
	.dc.b	$0F,$D7,$78,$9F,$A0,$F1,$7F,$42,$CC,$FF,$5C,$E5,$A7,$F5,$FF,$AA
	.dc.b	$7A,$BF,$27,$E3,$FF,$B4,$C6,$27,$E4,$46,$27,$51,$C5,$B9,$FE,$52
	.dc.b	$B8,$3F,$E5,$46,$BF,$E3,$B0,$E1,$FF,$1F,$35,$FF,$48,$E6,$F8,$AF
	.dc.b	$DD,$9F,$BC,$F7,$2E,$1F,$C2,$62,$D7,$E4,$66,$EB,$FA,$1F,$61,$F1
	.dc.b	$DF,$85,$F5,$FF,$C0,$B1,$8C,$7F,$09,$18,$9F,$D3,$4E,$70,$FF,$36
	.dc.b	$B5,$F5,$35,$FF,$2B,$7D,$9F,$FB,$32,$FF,$1C,$7F,$84,$9C,$47,$E3
	.dc.b	$E3,$1F,$15,$F9,$76,$9C,$0F,$F0,$94,$F0,$9F,$90,$D7,$BC,$FE,$EE
	.dc.b	$31,$1F,$EA,$9F,$25,$9F,$F8,$B3,$9F,$FC,$47,$18,$C3,$F9,$94,$62
	.dc.b	$7F,$82,$4E,$6B,$FD,$16,$D7,$EC,$35,$FC,$97,$8B,$FE,$38,$E9,$FE
	.dc.b	$5E,$DC,$3F,$9E,$3F,$D9,$38,$3F,$CF,$31,$FD,$AD,$B5,$FD,$A9,$AF
	.dc.b	$F6,$CF,$8C,$FF,$01,$7C,$A7,$F8,$EB,$A1,$18,$BF,$F1,$B4,$61,$FF
	.dc.b	$50,$D7,$FB,$62,$7F,$8F,$4E,$63,$FB,$D4,$78,$DF,$F2,$7F,$CB,$FF
	.dc.b	$46,$6D,$7F,$53,$69,$FC,$6C,$BC,$7F,$AF,$F7,$5E,$A7,$F8,$5F,$E5
	.dc.b	$7F,$CB,$71,$8B,$FE,$C3,$CC,$9C,$D7,$F8,$1A,$B5,$87,$F8,$29,$AF
	.dc.b	$F7,$A6,$BF,$EA,$1C,$3F,$C6,$D8,$7F,$95,$5F,$F5,$B3,$7C,$8F,$F2
	.dc.b	$FC,$62,$7F,$A2,$BC,$1F,$F6,$2B,$5F,$D9,$35,$ED,$1F,$E1,$AF,$C0
	.dc.b	$5F,$1D,$D2,$3F,$CE,$F1,$8B,$F0,$1F,$DF,$9C,$57,$82,$77,$5C,$6B
	.dc.b	$FC,$D9,$3F,$EF,$21,$F4,$3F,$BD,$7F,$D0,$B9,$FF,$AA,$63,$11,$FE
	.dc.b	$C1,$F8,$73,$C3,$B3,$E2,$56,$DF,$F7,$0C,$EB,$E7,$38,$E9,$F4,$71
	.dc.b	$C5,$6B,$05,$6B,$0F,$F5,$C3,$E0,$8F,$36,$39,$BF,$0B,$98,$63,$8C
	.dc.b	$3C,$75,$67,$F8,$F4,$D7,$FD,$21,$87,$9D,$C9,$EB,$3E,$0C,$ED,$F8
	.dc.b	$D8,$78,$5F,$3D,$FA,$63,$07,$8E,$FB,$63,$EF,$AE,$87,$C9,$2B,$9C
	.dc.b	$78,$46,$31,$F2,$3B,$8F,$01,$A2,$E1,$F2,$5B,$1E,$A6,$BC,$63,$E1
	.dc.b	$BE,$23,$E5,$B8,$8E,$8D,$3A,$16,$3A,$35,$E8,$7C,$63,$A1,$F4,$8F
	.dc.b	$56,$3F,$25,$C4,$F9,$1F,$09,$C5,$F0,$B6,$7C,$2D,$3E,$AF,$C6,$7E
	.dc.b	$19,$B9,$7E,$26,$31,$1F,$35,$BF,$19,$FE,$AB,$CB,$F2,$F1,$8D,$7F
	.dc.b	$DA,$14,$8C,$4E,$AF,$F3,$B2,$BE,$4E,$E3,$31,$8C,$7F,$4B,$8F,$F9
	.dc.b	$92,$BF,$1C,$D7,$F0,$33,$B8,$E3,$8F,$8A,$B8,$F4,$F8,$1B,$7F,$97
	.dc.b	$BE,$0F,$EC,$1A,$FF,$73,$9E,$17,$E8,$3C,$15,$FB,$8A,$D9,$C2,$32
	.dc.b	$6E,$78,$D7,$5B,$F2,$4F,$9B,$1F,$5D,$6A,$BC,$07,$96,$B6,$F1,$9D
	.dc.b	$B0,$F8,$31,$A7,$E6,$36,$BC,$71,$FC,$31,$AE,$A7,$B1,$CF,$0F,$8B
	.dc.b	$1D,$23,$2F,$EB,$A7,$89,$CB,$0F,$83,$7F,$80,$C7,$89,$27,$DD,$5F
	.dc.b	$BC,$3D,$47,$3C,$7E,$04,$73,$AF,$55,$FC,$B9,$C9,$F1,$5E,$F8,$7F
	.dc.b	$1E,$BF,$64,$FE,$33,$0F,$04,$76,$7E,$B5,$FC,$39,$E7,$3E,$CC,$7F
	.dc.b	$9A,$72,$B8,$1F,$1F,$FD,$4F,$6F,$C0,$5D,$3E,$51,$6B,$84,$DA,$E3
	.dc.b	$3F,$CE,$B3,$FF,$4A,$6B,$FD,$4B,$0E,$73,$FD,$2E,$FA,$3F,$CB,$CC
	.dc.b	$9F,$E4,$93,$07,$80,$FF,$00,$CF,$37,$FA,$3F,$E8,$4E,$FE,$6D,$8F
	.dc.b	$15,$F4,$C3,$FC,$D5,$F7,$4F,$4F,$FC,$17,$A2,$31,$7E,$18,$C4,$EA
	.dc.b	$73,$3F,$E5,$D9,$CB,$FE,$02,$FA,$2F,$89,$D2,$7E,$35,$70,$C3,$FC
	.dc.b	$F5,$B3,$E4,$FE,$57,$BE,$F9,$DA,$7C,$B8,$99,$7F,$BA,$E3,$2F,$F7
	.dc.b	$38,$F7,$F5,$FF,$89,$21,$BA,$30,$D8,$FF,$14,$B5,$F9,$CD,$C7,$FD
	.dc.b	$1D,$A5,$DF,$FB,$9E,$C8,$C5,$FD,$0C,$3F,$01,$E4,$FC,$AD,$7F,$D1
	.dc.b	$B3,$F2,$5A,$FC,$76,$BF,$C8,$AF,$F2,$13,$F0,$BC,$A9,$3F,$E0,$F8
	.dc.b	$FC,$2F,$E0,$7F,$CC,$A3,$FD,$18,$D7,$F9,$5B,$5F,$F0,$4B,$53,$FC
	.dc.b	$EB,$3F,$F7,$37,$DB,$FF,$88,$27,$F0,$A7,$F8,$41,$C3,$FE,$F0,$9E
	.dc.b	$1A,$7F,$38,$FB,$97,$C0,$7C,$F6,$D7,$F9,$51,$AF,$E6,$5A,$FF,$A8
	.dc.b	$DC,$E1,$FE,$F7,$EB,$7C,$6F,$E8,$9B,$5E,$41,$5C,$30,$F1,$61,$BB
	.dc.b	$95,$BA,$D3,$FE,$2A,$FB,$77,$C3,$7E,$15,$B1,$FE,$6A,$6B,$FA,$F6
	.dc.b	$BF,$EB,$97,$39,$31,$FE,$B3,$EE,$FE,$6D,$BD,$5F,$BA,$AD,$C7,$E2
	.dc.b	$38,$9F,$09,$FC,$14,$98,$FF,$92,$FE,$3E,$AC,$BF,$4B,$38,$7F,$4F
	.dc.b	$3F,$C1,$BE,$27,$F8,$6D,$7F,$05,$6B,$FE,$CF,$6B,$FD,$B9,$AF,$F9
	.dc.b	$31,$E3,$3E,$B8,$FF,$27,$ED,$F8,$08,$C4,$7D,$4F,$23,$C1,$E0,$C8
	.dc.b	$FA,$95,$BC,$FF,$BD,$B6,$3B,$4F,$F2,$3E,$29,$1F,$E6,$97,$F9,$37
	.dc.b	$16,$23,$E0,$5F,$C7,$9C,$63,$FC,$1A,$BE,$07,$CA,$FE,$F7,$33,$49
	.dc.b	$87,$FA,$92,$32,$FE,$27,$EB,$7D,$77,$EC,$6E,$3F,$16,$FF,$98,$C3
	.dc.b	$E5,$A9,$2F,$96,$95,$F3,$9F,$E7,$69,$E0,$F7,$9E,$73,$BC,$3C,$11
	.dc.b	$70,$FF,$10,$DF,$E7,$AF,$9F,$57,$F5,$1F,$2B,$FA,$51,$23,$FD,$21
	.dc.b	$E7,$D7,$7C,$7F,$EA,$F7,$FD,$29,$3F,$D4,$67,$FA,$97,$6C,$9F,$F4
	.dc.b	$C3,$FD,$57,$C4,$C4,$72,$8B,$A7,$F3,$49,$FE,$A7,$F4,$4F,$A0,$FE
	.dc.b	$58,$DB,$FF,$68,$B5,$FE,$AC,$D7,$FC,$7A,$1D,$57,$E3,$3E,$8B,$EA
	.dc.b	$CF,$CA,$0E,$23,$F2,$33,$C3,$6F,$F2,$3B,$6B,$FD,$4B,$5B,$9F,$F2
	.dc.b	$5B,$63,$FD,$53,$3B,$D7,$FB,$E7,$B3,$FD,$53,$8A,$57,$FA,$EE,$3F
	.dc.b	$B8,$FE,$04,$DB,$FB,$E7,$C7,$3E,$22,$9A,$8F,$EF,$DE,$97,$F5,$8F
	.dc.b	$A3,$98,$FF,$B1,$63,$FC,$09,$C6,$F9,$9F,$EC,$79,$CA,$FE,$5F,$AF
	.dc.b	$FB,$9E,$7C,$35,$FD,$A3,$D3,$95,$FC,$87,$5F,$BC,$FE,$1A,$5E,$EF
	.dc.b	$F6,$8E,$49,$3D,$27,$2D,$F6,$3C,$64,$75,$EE,$7F,$84,$38,$9F,$37
	.dc.b	$FE,$1D,$F9,$7F,$E7,$BB,$DF,$00,$EE,$7D,$BE,$74,$C7,$C0,$FC,$0F
	.dc.b	$4B,$FC,$49,$ED,$23,$FC,$1A,$D9,$FF,$06,$35,$FF,$05,$35,$FC,$4E
	.dc.b	$5F,$25,$FE,$AB,$93,$F0,$0C,$47,$EC,$A3,$1C,$7E,$B3,$FC,$69,$F7
	.dc.b	$8F,$92,$FF,$0E,$B6,$BF,$C3,$8D,$7F,$C3,$4D,$7F,$D2,$B4,$F9,$2F
	.dc.b	$37,$81,$FC,$93,$F0,$27,$D3,$7F,$8E,$1B,$7E,$6A,$DF,$85,$FE,$9E
	.dc.b	$9C,$CF,$9F,$EF,$FF,$B5,$AC,$51,$88,$4F,$0C,$7F,$9B,$76,$3A,$3F
	.dc.b	$39,$EC,$7F,$AB,$6F,$F1,$6D,$7F,$0D,$D5,$E5,$F9,$59,$1F,$DA,$F1
	.dc.b	$47,$DF,$D5,$F0,$7A,$AF,$99,$F1,$8F,$98,$FC,$23,$63,$95,$E2,$7F
	.dc.b	$C7,$3D,$52,$7E,$BD,$6D,$FD,$F6,$2F,$93,$F9,$04,$8F,$BB,$D3,$FD
	.dc.b	$53,$8F,$E4,$4F,$E4,$FC,$9E,$15,$F8,$26,$BD,$23,$93,$F8,$38,$CF
	.dc.b	$17,$F5,$DE,$4F,$E1,$B5,$A8,$FB,$59,$7F,$91,$F2,$34,$98,$7F,$9C
	.dc.b	$4F,$A6,$F8,$FE,$2A,$47,$CB,$8F,$DF,$FA,$1F,$E5,$AF,$9F,$93,$FB
	.dc.b	$38,$CB,$FC,$03,$E0,$93,$A8,$FD,$E4,$FF,$A0,$BF,$1F,$FB,$5B,$FF
	.dc.b	$26,$6F,$7F,$FB,$A1,$B4,$0F,$FB,$DA,$78,$79,$3F,$EE,$E6,$CE,$DA
	.dc.b	$9E,$DE,$4E,$EF,$FD,$EB,$F4,$5F,$84,$BE,$11,$8E,$7F,$C8,$5F,$F3
	.dc.b	$AA,$D6,$1C,$EB,$8F,$37,$38,$F3,$4E,$AB,$85,$F2,$BE,$98,$76,$BD
	.dc.b	$E1,$C5,$61,$C6,$B7,$1D,$CE,$B7,$CD,$E2,$73,$38,$B9,$6D,$E6,$BA
	.dc.b	$47,$03,$A1,$A2,$BA,$71,$87,$C1,$C3,$CD,$3F,$0A,$7F,$B0,$B5,$F3
	.dc.b	$98,$BF,$DB,$3F,$64,$7C,$57,$F9,$B3,$93,$F2,$0D,$8F,$F6,$13,$5F
	.dc.b	$ED,$AD,$7E,$BF,$AD,$FC,$5D,$FF,$64,$5A,$31,$1E,$46,$BC,$3A,$BF
	.dc.b	$3D,$3B,$7E,$24,$F3,$CB,$FB,$BF,$C6,$BE,$5B,$FC,$91,$1F,$81,$9D
	.dc.b	$6D,$CF,$1F,$8F,$18,$7F,$2D,$7C,$7A,$BA,$E3,$27,$6F,$D7,$9D,$46
	.dc.b	$09,$87,$79,$D6,$3E,$2C,$7C,$13,$9A,$DC,$7C,$DB,$E7,$39,$3D,$7F
	.dc.b	$25,$C6,$BC,$AD,$52,$E7,$C2,$9A,$4C,$BD,$13,$6B,$F2,$9E,$0B,$9F
	.dc.b	$FC,$99,$5F,$C0,$7C,$6F,$F3,$87,$5F,$F3,$C9,$CD,$FF,$90,$9F,$F0
	.dc.b	$45,$7F,$07,$F6,$76,$FF,$B4,$A3,$75,$C7,$A7,$36,$C7,$C1,$86,$21
	.dc.b	$F4,$3F,$DF,$B1,$FE,$38,$D8,$FF,$1B,$9A,$F0,$3F,$D0,$2F,$E5,$63
	.dc.b	$4F,$F8,$F1,$E1,$8F,$FA,$19,$B5,$FC,$4F,$9B,$FD,$35,$E4,$78,$67
	.dc.b	$37,$FC,$5D,$F6,$BF,$DE,$7E,$8B,$EF,$E1,$CF,$6F,$5E,$09,$CE,$3E
	.dc.b	$39,$F2,$DF,$36,$BE,$E6,$E1,$DA,$3D,$11,$E3,$3A,$E0,$F9,$AC,$5F
	.dc.b	$F8,$3B,$DD,$FF,$0F,$B7,$CF,$F5,$78,$3E,$49,$CE,$3F,$E5,$9C,$BF
	.dc.b	$9D,$DF,$CC,$AF,$09,$1F,$9B,$A7,$F7,$FF,$6F,$9C,$7F,$87,$21,$18
	.dc.b	$BF,$C1,$79,$3F,$73,$8B,$F1,$4F,$C4,$FA,$2F,$93,$B7,$E2,$63,$E7
	.dc.b	$72,$7C,$02,$65,$F7,$9F,$8F,$97,$F4,$7F,$9D,$F9,$B9,$1E,$DC,$CF
	.dc.b	$96,$CC,$78,$7E,$8B,$FC,$45,$8F,$F9,$61,$B7,$FE,$98,$5F,$29,$BF
	.dc.b	$03,$F2,$1E,$F7,$F7,$FB,$FE,$D4,$DB,$FC,$7B,$5F,$F4,$3B,$5F,$EB
	.dc.b	$9C,$0F,$A1,$FF,$39,$57,$F9,$A1,$B3,$FE,$A8,$7F,$C3,$71,$9F,$7F
	.dc.b	$D7,$E2,$48,$FE,$E1,$E3,$DB,$FF,$48,$63,$FE,$00,$9F,$2F,$99,$F8
	.dc.b	$A8,$C4,$F9,$A3,$F1,$F9,$7E,$EF,$E8,$7E,$39,$F2,$3E,$32,$E1,$FE
	.dc.b	$7C,$F9,$8F,$F2,$2B,$FD,$FF,$CC,$FB,$51,$8C,$7F,$7D,$1F,$1B,$83
	.dc.b	$C0,$4A,$F8,$EF,$82,$9F,$43,$C8,$E7,$BD,$AF,$07,$87,$F6,$B9,$3B
	.dc.b	$B6,$3F,$C0,$DD,$5F,$EF,$CE,$0F,$B3,$A7,$D0,$E4,$F8,$04,$AF,$E0
	.dc.b	$7C,$0D,$1F,$1B,$EA,$3D,$7F,$4F,$6B,$A9,$FD,$6F,$E7,$7E,$AE,$7F
	.dc.b	$C7,$BE,$41,$F4,$5F,$C4,$E7,$35,$E4,$9C,$BC,$B4,$FF,$0E,$4F,$DE
	.dc.b	$7F,$D9,$6D,$7B,$70,$7F,$00,$8C,$47,$08,$EC,$7F,$24,$C1,$F4,$3F
	.dc.b	$9F,$BE,$D4,$BF,$1B,$FC,$A7,$47,$D1,$7D,$83,$EC,$F9,$16,$FE,$F4
	.dc.b	$5A,$FF,$7C,$D7,$F0,$B9,$C9,$F8,$3F,$67,$F0,$27,$E0,$BD,$3B,$19
	.dc.b	$25,$72,$8C,$AC,$FF,$A3,$E6,$F8,$5F,$63,$6F,$E4,$4F,$BB,$E1,$F3
	.dc.b	$52,$73,$5B,$BE,$91,$F8,$48,$FF,$65,$BA,$9E,$13,$A3,$C5,$3C,$6F
	.dc.b	$C7,$7A,$57,$F8,$1F,$A1,$FA,$1F,$45,$FC,$E3,$0F,$B7,$18,$CB,$F0
	.dc.b	$4C,$AF,$4F,$D7,$BE,$6B,$17,$F3,$3D,$6F,$D4,$9F,$AA,$F0,$AD,$FE
	.dc.b	$BE,$2C,$FD,$29,$B9,$F9,$23,$FC,$CD,$F4,$7D,$6F,$E1,$FC,$4F,$F6
	.dc.b	$58,$C4,$FB,$EB,$F0,$F3,$9C,$7F,$A6,$D6,$AB,$FC,$70,$1C,$3F,$E3
	.dc.b	$B6,$BE,$08,$E5,$7F,$4B,$E3,$AC,$FE,$23,$E8,$3F,$82,$9F,$C1,$BD
	.dc.b	$AB,$5F,$C2,$E2,$CF,$E3,$A6,$E5,$ED,$8F,$F7,$AF,$99,$F8,$4F,$43
	.dc.b	$FC,$01,$EB,$7F,$3B,$C7,$F0,$31,$89,$91,$F8,$46,$D7,$DF,$1F,$46
	.dc.b	$FA,$1F,$C7,$D7,$F8,$2E,$30,$FE,$30,$C4,$7F,$72,$8C,$CF,$F8,$31
	.dc.b	$AF,$C8,$6B,$FE,$61,$77,$8F,$AD,$B5,$FC,$63,$F0,$0F,$E5,$13,$9B
	.dc.b	$FF,$0E,$79,$5F,$E3,$5A,$FF,$56,$74,$63,$5F,$EB,$6D,$8F,$DF,$CE
	.dc.b	$F2,$FD,$BF,$36,$5F,$DC,$FB,$DF,$D8,$BA,$FF,$9B,$B6,$FF,$9E,$38
	.dc.b	$7F,$9F,$7D,$25,$AF,$91,$E3,$A3,$FC,$26,$D9,$FF,$65,$36,$7F,$96
	.dc.b	$CE,$8F,$F1,$B4,$67,$5F,$F7,$FF,$7F,$F6,$CB,$FE,$2F,$39,$7F,$B2
	.dc.b	$C7,$FB,$05,$FF,$66,$7C,$97,$F4,$EC,$3F,$BD,$B6,$3F,$AA,$47,$F5
	.dc.b	$2F,$27,$0F,$EF,$6F,$FC,$23,$E8,$FF,$5E,$F6,$4D,$25,$FF,$44,$8C
	.dc.b	$BF,$CD,$E7,$8B,$63,$F2,$15,$D5,$DB,$F8,$DC,$9E,$E7,$F8,$E1,$B5
	.dc.b	$FE,$EF,$6C,$FF,$95,$1A,$FE,$C1,$C9,$FA,$12,$C7,$F8,$23,$E5,$DF
	.dc.b	$35,$FE,$61,$9C,$D7,$F9,$96,$73,$5F,$D0,$A7,$38,$A7,$35,$E7,$C3
	.dc.b	$FC,$65,$18,$AF,$E1,$95,$BC,$33,$CB,$A7,$FC,$E1,$87,$FB,$BC,$7C
	.dc.b	$43,$B3,$FD,$81,$83,$E3,$BF,$8B,$4E,$6B,$F8,$6F,$C2,$FF,$BF,$63
	.dc.b	$11,$FD,$0E,$77,$7F,$E7,$8A,$4B,$E9,$37,$87,$5D,$B3,$1C,$A7,$C8
	.dc.b	$EB,$0F,$E4,$CE,$61,$7C,$30,$F3,$D6,$9F,$04,$F7,$73,$7B,$8F,$49
	.dc.b	$F0,$23,$2D,$33,$8F,$86,$7D,$D3,$D3,$0C,$57,$53,$9B,$83,$CD,$87
	.dc.b	$78,$F8,$95,$5F,$6D,$78,$2B,$71,$E1,$36,$53,$EB,$5F,$80,$DC,$77
	.dc.b	$9F,$21,$CC,$B3,$FE,$2C,$C3,$E2,$3E,$68,$D3,$4D,$5F,$95,$F9,$A7
	.dc.b	$B4,$F1,$D7,$9A,$FC,$55,$F3,$9B,$3C,$CD,$C7,$89,$A7,$CC,$7E,$7C
	.dc.b	$73,$69,$CC,$B1,$B3,$CA,$7C,$33,$CB,$F1,$5F,$98,$7E,$DB,$5F,$C7
	.dc.b	$9A,$9F,$C1,$C6,$7E,$2B,$F7,$B9,$FE,$52,$31,$C1,$FE,$1B,$B4,$62
	.dc.b	$75,$7D,$2F,$8D,$6A,$31,$7C,$5F,$D1,$E1,$FE,$90,$9F,$C4,$9F,$A2
	.dc.b	$70,$FD,$FA,$E7,$3F,$AA,$6C,$FA,$8D,$C7,$4E,$AF,$6B,$BB,$F0,$4F
	.dc.b	$8B,$0F,$C8,$3D,$B2,$E1,$8F,$E4,$F1,$F4,$C7,$1A,$F2,$CF,$F0,$1C
	.dc.b	$7A,$DF,$85,$B1,$FB,$AB,$F2,$B7,$0F,$91,$87,$68,$FC,$45,$F8,$4F
	.dc.b	$23,$F8,$BA,$F0,$C7,$58,$F3,$DF,$D8,$AF,$24,$6E,$FA,$D6,$E3,$AD
	.dc.b	$6E,$3A,$D7,$0C,$7C,$44,$8F,$0B,$D2,$3C,$C4,$8E,$24,$AF,$AD,$5D
	.dc.b	$5A,$7C,$9B,$F1,$BF,$07,$1F,$93,$5F,$58,$F3,$61,$E3,$7A,$4F,$CD
	.dc.b	$3D,$E7,$39,$8F,$B0,$C9,$F0,$5C,$CF,$AE,$3C,$67,$E2,$6B,$F0,$2F
	.dc.b	$F4,$9F,$9B,$FE,$73,$A7,$F4,$F6,$C6,$4F,$EC,$9C,$4F,$A8,$CF,$1F
	.dc.b	$F4,$DA,$DD,$7F,$33,$CB,$FC,$75,$9F,$F5,$DC,$7A,$4F,$FA,$5A,$30
	.dc.b	$FE,$74,$E2,$E2,$FA,$46,$CF,$F2,$A9,$F9,$D2,$9C,$4B,$18,$50,$FA
	.dc.b	$5F,$F5,$FC,$7F,$7D,$6B,$FE,$07,$6D,$FF,$62,$5E,$B7,$B9,$DC,$FF
	.dc.b	$06,$AE,$9E,$BF,$F8,$8A,$73,$3F,$EF,$E6,$BF,$DD,$7E,$97,$E3,$EA
	.dc.b	$69,$8F,$A8,$B3,$FE,$11,$F8,$47,$CC,$FC,$E6,$C7,$D2,$AF,$9F,$F4
	.dc.b	$CE,$CA,$49,$FA,$FE,$E5,$7E,$4C,$ED,$F9,$2D,$73,$3F,$E3,$CC,$BF
	.dc.b	$D3,$11,$F8,$0F,$3B,$D2,$7A,$BE,$A2,$C7,$9F,$D1,$F3,$18,$7F,$CE
	.dc.b	$1F,$4D,$FC,$B7,$8F,$FC,$F8,$DE,$5F,$E5,$ED,$9F,$EF,$FF,$4E,$E7
	.dc.b	$FC,$5C,$E6,$53,$FE,$22,$7F,$DB,$ED,$7B,$F5,$7E,$4B,$E9,$DF,$3F
	.dc.b	$49,$3F,$EE,$6E,$C7,$A5,$88,$FC,$6D,$70,$7F,$1A,$D7,$35,$F8,$D8
	.dc.b	$C6,$9F,$E9,$7B,$7C,$6F,$E8,$9B,$7F,$EF,$4A,$DE,$BF,$D3,$BE,$9D
	.dc.b	$F1,$3F,$A8,$6D,$7F,$BE,$9A,$FF,$A6,$7E,$13,$FD,$F2,$E6,$7F,$5F
	.dc.b	$1F,$98,$AF,$EB,$2D,$9F,$F1,$4D,$72,$CB,$FC,$7B,$C8,$FA,$7F,$C9
	.dc.b	$92,$BF,$7B,$F4,$F5,$FF,$35,$4E,$DF,$E0,$33,$F9,$A9,$FD,$FF,$C6
	.dc.b	$FD,$79,$FF,$1B,$DF,$FA,$E6,$31,$1F,$EB,$16,$C7,$FA,$F7,$6B,$9F
	.dc.b	$F7,$09,$DD,$7F,$B4,$9A,$E7,$93,$F7,$91,$88,$FF,$22,$BC,$9F,$F2
	.dc.b	$3D,$7E,$0E,$FB,$79,$7E,$6B,$FD,$8F,$04,$BF,$E3,$6F,$ED,$38,$71
	.dc.b	$1F,$02,$2D,$FC,$89,$F7,$FC,$C4,$AF,$7E,$7E,$98,$FE,$43,$F5,$8F
	.dc.b	$9E,$D5,$7F,$B4,$7E,$A3,$FC,$F9,$7F,$E0,$E6,$C7,$AA,$FF,$94,$C3
	.dc.b	$E6,$E9,$27,$B9,$FE,$64,$8F,$C6,$BD,$63,$4D,$36,$23,$79,$7F,$BB
	.dc.b	$5E,$D1,$A6,$BC,$4F,$46,$DF,$C6,$3C,$F9,$1F,$D0,$9F,$E3,$7F,$4D
	.dc.b	$88,$FC,$3C,$5B,$FC,$7C,$FF,$20,$F0,$25,$7F,$4A,$7F,$91,$FD,$46
	.dc.b	$23,$F1,$F1,$6B,$F2,$F3,$FD,$33,$EA,$1F,$47,$49,$7E,$EF,$A3,$FE
	.dc.b	$0D,$3A,$8F,$E8,$AD,$7F,$A4,$B5,$FF,$1F,$87,$0F,$F9,$09,$E1,$6D
	.dc.b	$FF,$C8,$B4,$DF,$FA,$FB,$6F,$E7,$DF,$0A,$FF,$5C,$7C,$27,$F4,$02
	.dc.b	$4F,$F6,$AB,$FE,$CF,$F5,$0D,$7F,$AC,$36,$7F,$AE,$B5,$FE,$C2,$D7
	.dc.b	$FA,$97,$5F,$F9,$41,$FE,$91,$F0,$DF,$D1,$CF,$E9,$FE,$BF,$F7,$6E
	.dc.b	$17,$CF,$D2,$57,$FB,$7E,$1F,$43,$E4,$7C,$64,$8F,$1B,$FF,$00,$07
	.dc.b	$C2,$7F,$55,$E2,$93,$CE,$3C,$1E,$FC,$8E,$CF,$8F,$E0,$FF,$C0,$FB
	.dc.b	$F8,$27,$FB,$DB,$5F,$EF,$ED,$7E,$BB,$5F,$F6,$2F,$AD,$FE,$CF,$F3
	.dc.b	$3D,$C4,$8F,$67,$A7,$FE,$16,$F7,$53,$49,$3F,$EF,$F8,$63,$E4,$DB
	.dc.b	$FC,$15,$8F,$C8,$AF,$2C,$60,$FF,$08,$E2,$FA,$77,$CB,$FF,$89,$1B
	.dc.b	$5F,$B8,$9D,$4E,$7E,$67,$FC,$5B,$F3,$CF,$84,$F2,$DF,$EF,$A3,$1A
	.dc.b	$EA,$73,$2E,$DF,$F7,$BF,$1B,$9B,$FC,$63,$F4,$AF,$A7,$FF,$11,$36
	.dc.b	$FF,$C4,$ED,$7F,$C5,$6D,$7F,$C3,$95,$F9,$97,$FC,$52,$DD,$7F,$DF
	.dc.b	$9C,$DE,$8F,$A7,$FE,$F1,$AF,$A7,$E1,$FC,$3E,$3C,$C7,$D5,$B2,$71
	.dc.b	$1E,$09,$F2,$91,$FF,$1E,$D3,$E4,$5F,$3D,$F1,$9B,$3F,$15,$AF,$C3
	.dc.b	$F0,$BF,$D7,$3B,$7F,$31,$8F,$E4,$2B,$E6,$3E,$29,$23,$E7,$75,$7F
	.dc.b	$49,$53,$B7,$E3,$CF,$CA,$23,$83,$EB,$78,$BB,$7C,$43,$24,$67,$93
	.dc.b	$F8,$54,$8E,$F5,$F6,$32,$91,$E8,$7F,$0B,$8B,$DE,$CC,$FD,$89,$FB
	.dc.b	$A4,$7F,$CC,$BC,$4E,$3F,$E7,$FA,$FE,$99,$B3,$F7,$3C,$9F,$E0,$9F
	.dc.b	$47,$FC,$15,$FF,$27,$32,$3F,$82,$47,$F9,$F3,$E9,$FF,$87,$43,$3F
	.dc.b	$E4,$0A,$33,$18,$6F,$77,$FB,$B1,$FF,$77,$46,$7C,$4F,$F7,$7B,$6B
	.dc.b	$E2,$35,$ED,$5C,$7D,$BD,$BF,$EF,$6F,$A7,$FC,$65,$F0,$8C,$73,$FF
	.dc.b	$23,$C3,$CB,$5A,$C3,$99,$F4,$5F,$C4,$C6,$A7,$85,$72,$BE,$97,$B9
	.dc.b	$CB,$C7,$9B,$C6,$E6,$76,$B0,$F0,$61,$D6,$73,$1E,$0C,$3B,$57,$90
	.dc.b	$E1,$87,$66,$D7,$0C,$3B,$C6,$0E,$17,$DA,$F9,$E1,$F0,$2B,$73,$F0
	.dc.b	$6B,$FA,$FB,$5F,$09,$8C,$3E,$EB,$F9,$51,$F1,$9F,$E6,$AE,$6F,$F5
	.dc.b	$7B,$63,$FD,$66,$D7,$FB,$5B,$5F,$AD,$EB,$7F,$1D,$7F,$D7,$36,$8C
	.dc.b	$47,$91,$AF,$0E,$AF,$F5,$EC,$ED,$F9,$33,$CF,$2F,$C8,$FC,$83,$E6
	.dc.b	$7F,$5B,$1F,$80,$9D,$6B,$CF,$1F,$8F,$18,$7F,$65,$7C,$79,$3A,$E3
	.dc.b	$27,$6F,$DC,$9D,$46,$09,$87,$79,$D6,$3E,$2C,$7C,$13,$9A,$DC,$7C
	.dc.b	$DB,$E7,$39,$3D,$7F,$31,$C6,$BC,$AD,$52,$F9,$4C,$29,$A4,$CB,$D1
	.dc.b	$36,$BF,$43,$CE,$F8,$8F,$F2,$75,$7F,$B3,$FD,$6F,$E2,$BD,$BF,$C5
	.dc.b	$AB,$5D,$18,$AF,$ED,$F7,$C3,$0F,$E3,$F5,$CF,$0D,$47,$FB,$C2,$75
	.dc.b	$87,$37,$FD,$4B,$7D,$CF,$F2,$AF,$D3,$7F,$84,$1B,$E7,$F2,$E5,$FE
	.dc.b	$BD,$E4,$F4,$C5,$7B,$7F,$97,$98,$AD,$46,$71,$FE,$5F,$7C,$B0,$EA
	.dc.b	$78,$3C,$BF,$E6,$78,$7F,$90,$AF,$FC,$4B,$3F,$E4,$48,$FE,$6D,$18
	.dc.b	$70,$D7,$84,$F8,$0F,$45,$BE,$7B,$17,$DC,$95,$FD,$1B,$0F,$EF,$53
	.dc.b	$A8,$C7,$B6,$F5,$7E,$13,$6F,$E3,$F8,$18,$9F,$7B,$23,$FC,$11,$83
	.dc.b	$D8,$C6,$1D,$E3,$13,$C8,$DC,$EB,$2F,$AD,$F0,$9F,$6B,$8B,$F0,$CF
	.dc.b	$D3,$FA,$6F,$85,$3A,$BF,$B3,$C9,$F4,$0F,$B1,$93,$E8,$F1,$3F,$E1
	.dc.b	$B8,$C3,$63,$B7,$23,$E6,$F8,$C7,$87,$E9,$BF,$C4,$95,$FE,$29,$6D
	.dc.b	$FE,$41,$7C,$AE,$FB,$FF,$0D,$F4,$9F,$DF,$EF,$FD,$42,$DB,$FC,$3F
	.dc.b	$E0,$EF,$1F,$7E,$6F,$BF,$E9,$DF,$4D,$4C,$7A,$57,$2F,$F1,$D7,$07
	.dc.b	$F1,$96,$F8,$FE,$AF,$2B,$D6,$F8,$BC,$F9,$3F,$EB,$2F,$D0,$FF,$C3
	.dc.b	$6F,$92,$3E,$0F,$D4,$48,$EB,$F1,$D7,$83,$F5,$BF,$3D,$FE,$3B,$9F
	.dc.b	$F2,$53,$63,$F6,$39,$FF,$A5,$67,$3E,$CF,$C2,$E5,$91,$C3,$6B,$7D
	.dc.b	$3D,$CF,$BD,$89,$BC,$18,$73,$E4,$4C,$3C,$3F,$5B,$70,$FD,$8F,$3F
	.dc.b	$D4,$ED,$FD,$5B,$6B,$FC,$EB,$E6,$FC,$C7,$E9,$F5,$78,$3E,$6A,$47
	.dc.b	$B3,$E6,$EE,$3F,$BC,$FA,$9E,$0F,$EC,$DA,$FF,$AE,$1A,$FF,$57,$6B
	.dc.b	$FE,$2D,$78,$E5,$FB,$E6,$D7,$F8,$C6,$1F,$51,$F9,$4A,$F0,$6F,$FC
	.dc.b	$3E,$F7,$1C,$E3,$F8,$ED,$70,$F0,$35,$1F,$2B,$EA,$66,$3F,$5B,$CC
	.dc.b	$B8,$7D,$A9,$B8,$7F,$2D,$E3,$FE,$0D,$7F,$72,$7E,$9F,$7C,$8F,$BF
	.dc.b	$EA,$DC,$3F,$07,$37,$0E,$53,$6B,$D9,$AB,$E1,$BC,$ED,$8F,$80,$6B
	.dc.b	$E1,$F3,$32,$33,$8F,$FA,$36,$B8,$3F,$E5,$CC,$3E,$3D,$EE,$31,$3A
	.dc.b	$9F,$E9,$2D,$AE,$C7,$48,$F9,$F7,$F3,$7E,$92,$4E,$7E,$A1,$F3,$98
	.dc.b	$C7,$F0,$55,$70,$FE,$AB,$5A,$BF,$3D,$38,$97,$F9,$CF,$62,$57,$EA
	.dc.b	$3C,$2B,$7F,$AC,$9B,$8F,$D3,$9B,$5F,$E2,$1F,$A8,$ED,$7F,$D0,$DA
	.dc.b	$FF,$89,$43,$87,$86,$7F,$D3,$1B,$3E,$69,$E7,$7F,$1B,$E1,$FF,$AE
	.dc.b	$46,$27,$75,$FD,$72,$30,$FD,$39,$DD,$FE,$E3,$E8,$A5,$FF,$04,$F6
	.dc.b	$ED,$FF,$08,$9B,$8F,$E3,$E6,$D7,$F8,$F7,$D6,$FE,$60,$FC,$A6,$DF
	.dc.b	$DC,$F3,$FF,$B8,$46,$2F,$FB,$3B,$6B,$94,$FE,$38,$D9,$FD,$DF,$EA
	.dc.b	$1F,$4D,$F0,$F0,$FF,$50,$46,$2B,$FB,$74,$23,$0F,$F7,$16,$BF,$3D
	.dc.b	$CE,$1F,$DE,$1B,$3F,$BB,$FA,$8F,$E4,$93,$9B,$F1,$12,$7F,$25,$3A
	.dc.b	$7F,$CB,$D5,$CA,$FF,$D6,$9E,$3F,$F8,$B6,$7F,$1A,$D8,$FE,$67,$C5
	.dc.b	$FE,$13,$8C,$63,$FE,$0A,$6D,$7F,$9A,$5B,$4F,$59,$D9,$F9,$6A,$EB
	.dc.b	$B7,$F3,$2F,$0E,$47,$F3,$EF,$A8,$B8,$7F,$42,$9B,$97,$F0,$99,$B5
	.dc.b	$FE,$77,$C5,$1F,$E2,$76,$C7,$F4,$8C,$B1,$32,$3F,$53,$3A,$C3,$FA
	.dc.b	$7D,$65,$FE,$A2,$7F,$46,$3F,$A3,$F8,$96,$BF,$A9,$C5,$3F,$AB,$C5
	.dc.b	$C3,$FC,$95,$EC,$7F,$C0,$1A,$FE,$1B,$87,$F8,$57,$D1,$FF,$07,$4F
	.dc.b	$E2,$F0,$F8,$CE,$DF,$F2,$6B,$6F,$9C,$F0,$E2,$E2,$FF,$91,$61,$ED
	.dc.b	$D3,$FC,$35,$F2,$2F,$9A,$C6,$7F,$C2,$2B,$6D,$AD,$CE,$B2,$FF,$2E
	.dc.b	$DF,$1A,$F4,$B6,$33,$CB,$C2,$73,$56,$73,$18,$3F,$8B,$BF,$ED,$E1
	.dc.b	$F5,$0E,$8F,$F5,$86,$5F,$EF,$9A,$56,$A7,$FC,$A9,$F0,$78,$BE,$13
	.dc.b	$FD,$6F,$E2,$97,$9B,$DC,$F1,$35,$19,$7F,$8D,$5D,$58,$78,$5E,$07
	.dc.b	$FB,$D8,$7C,$4F,$81,$DE,$1E,$07,$38,$78,$27,$55,$CD,$E1,$3F,$07
	.dc.b	$1C,$DF,$17,$30,$C3,$C9,$7E,$39,$CC,$62,$B8,$61,$EE,$9E,$78,$FA
	.dc.b	$F0,$F9,$D8,$E1,$F2,$9F,$00,$A7,$AA,$3E,$7C,$76,$8F,$83,$8F,$86
	.dc.b	$FB,$1F,$5B,$0F,$0C,$F3,$C3,$B1,$C4,$B3,$C2,$3E,$89,$F3,$0F,$84
	.dc.b	$F3,$C3,$D1,$1F,$04,$F0,$D7,$67,$E9,$57,$47,$C1,$5C,$5E,$87,$18
	.dc.b	$D3,$C2,$35,$1E,$56,$9E,$58,$F8,$47,$2C,$7E,$75,$F9,$1B,$19,$3C
	.dc.b	$B1,$97,$FC,$D9,$8F,$E4,$3D,$AF,$C3,$C7,$CF,$6B,$F8,$B6,$FC,$47
	.dc.b	$EE,$33,$FC,$FC,$63,$7F,$C9,$D7,$43,$F4,$75,$B7,$01,$C4,$F1,$AF
	.dc.b	$CE,$65,$FE,$86,$AB,$CB,$FA,$C7,$F3,$B7,$CF,$4F,$CE,$36,$B4,$71
	.dc.b	$9F,$D1,$B6,$77,$8F,$99,$B8,$76,$A8,$76,$9D,$57,$86,$FB,$4F,$EF
	.dc.b	$A3,$A1,$E0,$C3,$CD,$87,$49,$E5,$87,$48,$FD,$95,$70,$9F,$8B,$87
	.dc.b	$66,$D6,$E7,$A4,$7E,$FF,$0F,$0B,$F2,$6B,$83,$FC,$4F,$0F,$0D,$70
	.dc.b	$8F,$09,$30,$F0,$CE,$A3,$D4,$D7,$A6,$1D,$EF,$ED,$E5,$C2,$7F,$5F
	.dc.b	$87,$58,$F7,$1E,$FC,$35,$3F,$43,$0E,$98,$FA,$0F,$4D,$FC,$19,$F4
	.dc.b	$1D,$F1,$8F,$EF,$A3,$ED,$61,$D0,$EE,$FD,$3C,$7E,$0B,$CC,$E9,$7D
	.dc.b	$E3,$38,$7B,$67,$31,$F3,$F9,$1C,$2F,$9B,$FE,$77,$B7,$F3,$A9,$CC
	.dc.b	$63,$3F,$F0,$25,$6D,$FE,$9B,$F5,$19,$AD,$C7,$F5,$0A,$C5,$6E,$3F
	.dc.b	$AA,$D7,$37,$F9,$6E,$BF,$CC,$DB,$87,$E6,$9E,$37,$BC,$BF,$CB,$55
	.dc.b	$B6,$D6,$EB,$C0,$7F,$78,$FA,$87,$A5,$E5,$6C,$66,$31,$E9,$65,$91
	.dc.b	$93,$94,$7A,$BE,$9E,$47,$F6,$C7,$DB,$F5,$18,$8F,$EE,$31,$72,$FE
	.dc.b	$9D,$B7,$CE,$AF,$81,$F5,$0E,$CA,$4C,$3F,$BB,$FC,$8D,$7F,$AE,$EB
	.dc.b	$FC,$06,$B8,$47,$FA,$29,$AF,$F8,$02,$7E,$7F,$8F,$23,$FC,$02,$FD
	.dc.b	$2F,$A8,$C4,$7F,$81,$A2,$E7,$E6,$E3,$57,$D5,$FA,$8F,$E4,$33,$F5
	.dc.b	$5B,$97,$F1,$36,$E1,$FE,$00,$FA,$8F,$BD,$5B,$E9,$E4,$49,$FF,$50
	.dc.b	$CE,$AF,$F8,$E7,$D4,$3E,$93,$F0,$8D,$AF,$F5,$3B,$5F,$86,$7F,$83
	.dc.b	$BA,$5F,$19,$C2,$A4,$F0,$7F,$1D,$3A,$8C,$1F,$E7,$AA,$E8,$FF,$A9
	.dc.b	$5A,$F9,$27,$FC,$FE,$BE,$9D,$F0,$BF,$99,$6C,$7F,$AF,$1A,$F9,$87
	.dc.b	$D4,$3D,$CE,$15,$27,$83,$FA,$39,$D4,$7F,$8E,$3E,$A3,$F5,$24,$7F
	.dc.b	$C7,$7F,$51,$5F,$F1,$FC,$5E,$0F,$B5,$3F,$AD,$F8,$BF,$9D,$5F,$E3
	.dc.b	$46,$A6,$FC,$7E,$19,$E0,$F1,$3F,$6D,$18,$9F,$F5,$0C,$FE,$ED,$FF
	.dc.b	$66,$35,$FE,$AC,$D7,$F8,$3E,$FF,$C0,$88,$FF,$95,$36,$AF,$F9,$62
	.dc.b	$2E,$3F,$93,$9F,$57,$CB,$CA,$FF,$74,$DA,$3F,$85,$7D,$43,$E2,$B5
	.dc.b	$5F,$46,$B4,$F0,$AD,$CF,$F1,$1F,$A8,$7C,$CD,$25,$7F,$BD,$A1,$A6
	.dc.b	$23,$EE,$47,$07,$53,$CE,$75,$9F,$F1,$BA,$D5,$70,$AD,$3F,$E7,$4F
	.dc.b	$37,$F9,$29,$1F,$F3,$BF,$D3,$57,$FC,$F7,$16,$FF,$6D,$3F,$CB,$3C
	.dc.b	$19,$5F,$E7,$E7,$F9,$7F,$D4,$62,$3F,$D0,$71,$6B,$F7,$D3,$FC,$C7
	.dc.b	$EA,$1F,$47,$49,$7F,$E7,$AA,$FE,$69,$F4,$3F,$BE,$85,$70,$75,$DB
	.dc.b	$FE,$15,$8C,$47,$E8,$5B,$5C,$BB,$FA,$CF,$F8,$6C,$FF,$43,$CF,$07
	.dc.b	$0D,$7A,$61,$FE,$F0,$F1,$FF,$AB,$FB,$FF,$D3,$09,$3F,$E9,$EE,$3F
	.dc.b	$E9,$9F,$50,$DE,$92,$7F,$D7,$7F,$1F,$F8,$77,$FD,$4B,$EA,$7F,$56
	.dc.b	$23,$FE,$A8,$D3,$FA,$E7,$13,$E7,$69,$27,$FD,$8B,$0F,$9F,$F0,$B2
	.dc.b	$3E,$29,$FE,$A4,$C9,$1D,$3E,$87,$AF,$C6,$CA,$E6,$F8,$3E,$4B,$C3
	.dc.b	$F4,$F2,$7B,$BE,$4F,$8E,$FE,$D5,$C1,$E1,$9E,$E7,$E9,$73,$C9,$FD
	.dc.b	$E7,$E7,$DF,$38,$FA,$0F,$6F,$D3,$4B,$F5,$BF,$03,$EA,$3F,$BE,$7C
	.dc.b	$0A,$29,$27,$FD,$CB,$0D,$3C,$7C,$52,$7C,$7A,$7C,$97,$CB,$F2,$EF
	.dc.b	$98,$FF,$04,$7D,$43,$E6,$E9,$2B,$FD,$E3,$0F,$9C,$F5,$7C,$5F,$8E
	.dc.b	$7E,$6F,$0A,$31,$1F,$D1,$A3,$1B,$7D,$AF,$93,$FF,$08,$7D,$3B,$E7
	.dc.b	$A9,$2F,$9E,$1C,$BC,$4F,$EF,$3D,$7F,$9A,$F1,$7F,$C4,$8D,$A7,$87
	.dc.b	$F7,$3C,$4F,$9D,$F5,$1F,$D9,$36,$F2,$7C,$77,$83,$CE,$98,$76,$8A
	.dc.b	$FF,$47,$DB,$F2,$8D,$9F,$7C,$FD,$9F,$98,$78,$54,$95,$F8,$0B,$E3
	.dc.b	$B7,$F8,$AD,$7C,$F7,$C0,$24,$FC,$8E,$CF,$E8,$73,$A8,$E5,$3A,$7E
	.dc.b	$54,$EA,$3F,$B1,$CE,$E7,$F7,$41,$C3,$FB,$D6,$BF,$E9,$95,$87,$D2
	.dc.b	$24,$78,$FE,$0E,$BF,$0E,$2E,$1F,$EF,$7D,$3F,$DD,$D5,$FB,$4F,$46
	.dc.b	$5F,$EC,$9F,$DB,$F8,$18,$8F,$F0,$3F,$0F,$EC,$27,$18,$7D,$CA,$FF
	.dc.b	$29,$7D,$43,$C9,$FF,$57,$6D,$FD,$23,$14,$E5,$1F,$EF,$FE,$FF,$F5
	.dc.b	$A4,$66,$30,$69,$FF,$64,$CF,$37,$4D,$AE,$F5,$C9,$B5,$C7,$C9,$AB
	.dc.b	$EA,$1C,$7E,$3C,$3C,$D1,$8B,$F2,$C6,$27,$58,$7F,$B2,$6B,$F1,$37
	.dc.b	$A8,$C4,$EA,$3D,$B3,$CD,$F5,$9E,$18,$D3,$E3,$E3,$F2,$CE,$EF,$C1
	.dc.b	$87,$5A,$E7,$7C,$2B,$C3,$1D,$1E,$18,$78,$AB,$84,$F7,$6C,$F0,$BE
	.dc.b	$B3,$C6,$FA,$1D,$70,$F1,$BF,$4E,$3A,$1F,$96,$AF,$63,$FD,$45,$AF
	.dc.b	$FB,$77,$D6,$70,$FE,$72,$3F,$DE,$D1,$87,$F3,$FF,$1F,$FE,$57,$9D
	.dc.b	$36,$3B,$7A,$DF,$C4,$DB,$D9,$CF,$0F,$F1,$83,$67,$53,$E3,$78,$67
	.dc.b	$F7,$1B,$AF,$FC,$01,$F9,$57,$D0,$3B,$BE,$2D,$6B,$0F,$B2,$D7,$F1
	.dc.b	$0D,$AE,$33,$C3,$19,$5F,$4E,$B5,$7D,$63,$13,$AC,$7C,$8F,$D6,$C3
	.dc.b	$9C,$E8,$93,$A7,$F1,$31,$A8,$C1,$FE,$34,$E6,$FE,$C7,$F0,$7D,$F9
	.dc.b	$BF,$97,$FB,$17,$4F,$E5,$F3,$71,$FE,$5F,$E0,$7F,$8E,$27,$FB,$8C
	.dc.b	$E7,$0F,$ED,$B7,$6B,$FD,$0B,$5F,$DD,$A3,$1C,$79,$F9,$5B,$3F,$DE
	.dc.b	$9B,$8F,$F0,$8B,$46,$32,$FE,$86,$F5,$F8,$1F,$F6,$C5,$7F,$44,$6B
	.dc.b	$E6,$6E,$5F,$E0,$06,$DF,$FA,$2B,$0F,$F0,$4C,$79,$F8,$3A,$47,$F8
	.dc.b	$22,$9F,$9F,$7F,$57,$8F,$F0,$7F,$D2,$7B,$E3,$18,$F9,$E3,$1E,$2F
	.dc.b	$82,$73,$3F,$C6,$0F,$EA,$EE,$1F,$F7,$F3,$5C,$C7,$2A,$FF,$0E,$96
	.dc.b	$3F,$95,$CF,$F9,$BE,$3C,$CD,$8C,$F9,$7E,$5F,$13,$CC,$F8,$C9,$5F
	.dc.b	$D2,$31,$F6,$C7,$9F,$C9,$DC,$F4,$6D,$F8,$3C,$BF,$80,$7C,$1F,$A3
	.dc.b	$B5,$E2,$F9,$0F,$80,$47,$B6,$68,$F8,$9E,$77,$C9,$F9,$1F,$94,$DA
	.dc.b	$F5,$F5,$FC,$D3,$E7,$7C,$3D,$8F,$7F,$97,$F3,$48,$F6,$7E,$87,$D0
	.dc.b	$7D,$5F,$2F,$EB,$36,$7E,$3F,$CA,$7D,$9F,$9B,$FB,$4D,$8F,$97,$F3
	.dc.b	$BF,$64,$8F,$A9,$FB,$9E,$0F,$E0,$7E,$73,$F0,$4D,$9F,$A7,$E1,$FE
	.dc.b	$17,$E6,$7F,$0C,$D8,$FA,$FC,$BF,$85,$23,$F1,$9F,$CB,$F8,$EF,$9B
	.dc.b	$FD,$62,$F9,$CB,$18,$73,$F1,$9A,$F9,$BB,$87,$E6,$6D,$1C,$BE,$67
	.dc.b	$A9,$27,$F0,$52,$8F,$D4,$E2,$FA,$1F,$F2,$35,$7F,$33,$6C,$7F,$84
	.dc.b	$1A,$FF,$A2,$DA,$FF,$76,$EF,$5B,$E5,$FD,$7F,$37,$EC,$6E,$1E,$77
	.dc.b	$E0,$91,$F2,$B6,$FF,$C1,$75,$CE,$7D,$DE,$C7,$37,$E0,$7C,$97,$C8
	.dc.b	$F9,$AF,$92,$D8,$C9,$F1,$63,$2D,$3E,$47,$A9,$F4,$3E,$53,$E8,$B6
	.dc.b	$32,$7C,$D8,$CB,$4F,$A1,$F1,$9F,$63,$E5,$3E,$CB,$63,$27,$D5,$8C
	.dc.b	$B4,$CE,$7E,$E8,$CB,$63,$EE,$7C,$97,$E1,$3C,$FF,$C2,$B6,$32,$49
	.dc.b	$F8,$FF,$25,$A8,$FC,$27,$C8,$7E,$33,$D8,$FC,$6B,$63,$27,$E2,$63
	.dc.b	$2D,$3F,$D3,$BF,$97,$3D,$1F,$E8,$9E,$AB,$A9,$E5,$7F,$47,$CD,$7B
	.dc.b	$C3,$FD,$67,$7B,$AD,$4E,$6B,$FC,$B6,$D7,$FD,$AC,$D7,$F9,$53,$5F
	.dc.b	$F2,$1B,$C6,$7F,$B4,$47,$83,$0F,$2E,$0F,$2B,$BC,$7E,$B3,$DD,$FD
	.dc.b	$6F,$BD,$73,$FD,$43,$70,$FC,$7B,$71,$FE,$E9,$D7,$F9,$C8,$FE,$C1
	.dc.b	$F8,$67,$FA,$12,$FF,$A4,$3F,$ED,$36,$FB,$FA,$F8,$DD,$B2,$37,$F1
	.dc.b	$57,$26,$5E,$AF,$CB,$FF,$CC,$95,$F2,$3F,$44,$F9,$AF,$FB,$3A,$33
	.dc.b	$5C,$23,$FC,$CB,$4A,$DC,$E4,$FF,$34,$D7,$07,$FA,$A0,$70,$FF,$57
	.dc.b	$6B,$C3,$5F,$E0,$B1,$E0,$AF,$34,$7E,$0A,$BE,$75,$F9,$63,$E2,$FC
	.dc.b	$87,$C6,$F9,$1B,$3F,$19,$B8,$FF,$A1,$B5,$57,$CC,$F9,$2C,$E5,$F4
	.dc.b	$27,$FD,$C5,$87,$CE,$F9,$0F,$A3,$F1,$FF,$49,$B1,$F4,$A3,$EA,$FC
	.dc.b	$8E,$32,$ED,$F1,$FF,$E0,$4C,$0D,$25,$FF,$65,$8C,$BF,$72,$3E,$FF
	.dc.b	$C8,$7E,$01,$FF,$10,$7C,$2B,$C7,$FC,$1F,$C6,$E1,$B7,$FE,$0B,$9D
	.dc.b	$3F,$DC,$5B,$8F,$D2,$E8,$FF,$0F,$D7,$F8,$62,$73,$3F,$E0,$99,$65
	.dc.b	$FE,$31,$8E,$0D,$74,$D7,$F9,$93,$5F,$F7,$3B,$C7,$6F,$99,$F1,$1F
	.dc.b	$9A,$F8,$EF,$CD,$BF,$B9,$AF,$F4,$D8,$EA,$7B,$1F,$E1,$BB,$78,$BF
	.dc.b	$E5,$79,$CD,$7E,$33,$E2,$FF,$D6,$F1,$88,$FC,$EC,$EE,$FF,$DB,$35
	.dc.b	$FE,$BC,$FA,$1D,$61,$D3,$7C,$56,$B0,$56,$B0,$FF,$33,$3D,$63,$CE
	.dc.b	$E6,$B8,$C7,$8C,$E3,$5E,$9A,$E2,$F5,$AE,$13,$A3,$83,$83,$B2,$C3
	.dc.b	$DB,$3C,$EE,$CF,$7C,$79,$9C,$2B,$C7,$1F,$0A,$6B,$E8,$AE,$4F,$84
	.dc.b	$FF,$71,$97,$1F,$88,$E4,$E0,$7F,$00,$C7,$D6,$F8,$67,$9E,$1E,$18
	.dc.b	$FB,$51,$83,$E0,$1E,$0C,$74,$7B,$DC,$C7,$5C,$3C,$EF,$C1,$3C,$97
	.dc.b	$E0,$7C,$F7,$F0,$2B,$EC,$36,$3E,$5C,$77,$3A,$BF,$0C,$F6,$1D,$6F
	.dc.b	$C4,$F0,$3C,$46,$B1,$F9,$78,$FC,$49,$F4,$D7,$5A,$F8,$47,$0B,$F8
	.dc.b	$47,$8D,$F9,$B3,$D9,$B3,$C5,$F5,$BF,$1D,$E6,$79,$E7,$9C,$66,$36
	.dc.b	$77,$7F,$31,$D3,$FA,$98,$FC,$93,$5F,$BA,$D7,$F5,$EE,$7C,$E7,$93
	.dc.b	$FD,$12,$53,$98,$C3,$71,$4E,$B0,$FF,$48,$5F,$29,$CD,$6E,$BF,$06
	.dc.b	$70,$C1,$EB,$FF,$A5,$F0,$FD,$C4,$62,$7F,$3A,$7F,$8F,$31,$F2,$4F
	.dc.b	$EF,$71,$ED,$3D,$0F,$E0,$F7,$CE,$FF,$D4,$47,$38,$D4,$F0,$AE,$57
	.dc.b	$EF,$CB,$C3,$87,$FA,$A5,$B7,$FA,$AB,$EF,$1C,$32,$F2,$E3,$D7,$1D
	.dc.b	$DF,$6C,$7C,$F3,$F1,$1E,$CF,$93,$1E,$CD,$9E,$D3,$A7,$E6,$E3,$E0
	.dc.b	$9D,$47,$ED,$DF,$66,$3E,$38,$FD,$D4,$ED,$F9,$D7,$D9,$A7,$6C,$63
	.dc.b	$F2,$6F,$D4,$76,$C7,$A1,$D7,$0F,$69,$23,$D9,$5C,$A3,$15,$D8,$F1
	.dc.b	$E3,$EB,$7E,$D4,$FC,$8C,$3E,$B4,$6F,$0E,$F8,$74,$8F,$74,$FA,$5F
	.dc.b	$BA,$77,$8D,$61,$E0,$8C,$9E,$0B,$F6,$DF,$D4,$25,$75,$7D,$8F,$CC
	.dc.b	$AE,$58,$FC,$9B,$F3,$9E,$23,$F9,$5C,$FD,$03,$C2,$47,$E0,$4F,$2C
	.dc.b	$7E,$05,$7E,$3E,$BE,$61,$2B,$EE,$CF,$E1,$9F,$EC,$FC,$5F,$D7,$F6
	.dc.b	$FE,$D6,$D8,$C9,$FE,$40,$F8,$CC,$71,$FF,$76,$9F,$F1,$6C,$F0,$3F
	.dc.b	$C1,$8E,$1E,$27,$F8,$05,$79,$39,$94,$62,$7F,$C1,$13,$98,$C4,$71
	.dc.b	$8F,$0F,$23,$E3,$29,$E4,$AD,$EF,$FE,$2B,$F8,$CD,$93,$0F,$F2,$CF
	.dc.b	$C6,$57,$FC,$1B,$1E,$3C,$0D,$3C,$DB,$3E,$33,$E0,$B6,$BF,$CA,$7B
	.dc.b	$7F,$B0,$E3,$11,$FE,$3F,$F2,$37,$5E,$8A,$93,$C0,$6D,$FE,$10,$9E
	.dc.b	$BF,$15,$27,$53,$FE,$14,$8F,$0F,$C6,$66,$7F,$CC,$13,$FE,$18,$F8
	.dc.b	$CF,$29,$9D,$BF,$D0,$FF,$44,$F8,$CF,$B9,$68,$C3,$E3,$31,$1F,$81
	.dc.b	$97,$D7,$3E,$3B,$8D,$F0,$BF,$C0,$B5,$FF,$37,$B5,$FE,$6E,$74,$3F
	.dc.b	$08,$F1,$A6,$BF,$E7,$CB,$FE,$51,$18,$AF,$F4,$3C,$62,$79,$D7,$17
	.dc.b	$E8,$7C,$6F,$D4,$25,$7C,$8E,$0D,$3F,$39,$47,$DA,$C6,$4E,$E3,$EE
	.dc.b	$FA,$3F,$7B,$D4,$93,$CE,$32,$F5,$8C,$BF,$4A,$FF,$D6,$4F,$CB,$FC
	.dc.b	$A1,$D5,$F4,$1B,$19,$8F,$A7,$F5,$4F,$33,$51,$FA,$99,$57,$07,$F5
	.dc.b	$ED,$7F,$D3,$AD,$7E,$7B,$BD,$FF,$D8,$F6,$F3,$7B,$9D,$AB,$E3,$7C
	.dc.b	$7F,$E7,$3E,$3E,$46,$08,$7B,$5C,$BE,$F8,$FD,$1F,$C6,$C8,$FA,$0F
	.dc.b	$E9,$BE,$3F,$11,$F1,$7E,$3F,$F5,$26,$32,$FF,$67,$ED,$FE,$64,$E3
	.dc.b	$FD,$57,$C7,$FE,$B8,$FD,$7C,$4B,$FB,$91,$97,$EF,$46,$5F,$A5,$19
	.dc.b	$7E,$A4,$7E,$E3,$E3,$BF,$73,$F1,$F2,$7F,$0D,$19,$7F,$11,$19,$7E
	.dc.b	$E4,$7F,$00,$F2,$BF,$80,$FC,$7C,$9F,$C7,$46,$5F,$C8,$46,$5F,$C2
	.dc.b	$47,$F2,$8F,$A4,$7C,$75,$24,$FE,$5F,$E4,$2B,$FE,$A7,$9D,$BF,$CB
	.dc.b	$BC,$04,$62,$35,$5F,$CA,$E3,$1B,$7C,$C9,$FE,$2E,$74,$7F,$58,$BE
	.dc.b	$6F,$FD,$8A,$FF,$37,$6B,$FC,$F1,$AF,$FB,$A9,$AF,$FB,$CB,$F1,$77
	.dc.b	$C8,$FF,$3F,$A4,$62,$7F,$A2,$C6,$27,$53,$99,$F7,$47,$C5,$F9,$29
	.dc.b	$1F,$C8,$30,$D5,$FF,$03,$CB,$E0,$FC,$97,$CB,$F8,$69,$7B,$8F,$9D
	.dc.b	$F2,$3F,$0C,$91,$F2,$70,$47,$C7,$F9,$1F,$A9,$EE,$4A,$F0,$47,$D7
	.dc.b	$F9,$1F,$B1,$F2,$32,$7C,$91,$F6,$FE,$43,$E8,$12,$3B,$3F,$4F,$D7
	.dc.b	$FC,$07,$C8,$C9,$F6,$47,$E0,$FD,$EF,$C2,$7C,$8C,$9F,$81,$1F,$87
	.dc.b	$F9,$0F,$B8,$48,$F5,$3F,$7F,$E4,$3F,$18,$48,$F9,$1F,$23,$F8,$FF
	.dc.b	$80,$FC,$87,$BD,$27,$E6,$47,$E4,$FE,$07,$F1,$04,$8F,$8C,$FF,$75
	.dc.b	$FD,$01,$F1,$FF,$E0,$06,$CF,$F5,$46,$BF,$D4,$97,$C8,$EF,$E4,$78
	.dc.b	$7C,$7C,$AE,$11,$5F,$E8,$A7,$F6,$2A,$4F,$84,$91,$FA,$18,$FE,$34
	.dc.b	$F8,$7C,$FF,$21,$CF,$B7,$F0,$D2,$F8,$DF,$41,$2B,$D1,$F1,$F5,$F0
	.dc.b	$F7,$AB,$84,$77,$AF,$84,$47,$E0,$7A,$BF,$0C,$E5,$87,$83,$E3,$BE
	.dc.b	$31,$27,$E3,$7C,$7D,$7B,$E0,$9F,$9C,$48,$F7,$7C,$77,$CF,$F8,$F9
	.dc.b	$1F,$3E,$3E,$91,$87,$E8,$46,$1F,$A9,$F1,$DF,$5F,$E3,$E4,$60,$87
	.dc.b	$D7,$72,$FD,$98,$FF,$02,$FC,$AF,$F8,$B7,$97,$FD,$5F,$F2,$0E,$AF
	.dc.b	$EF,$73,$9B,$4E,$5F,$F5,$AC,$66,$31,$08,$72,$FF,$81,$27,$31,$BA
	.dc.b	$EB,$F1,$DF,$EC,$2B,$EB,$18,$C3,$C6,$F8,$B0,$E4,$E6,$76,$78,$AB
	.dc.b	$96,$1D,$9B,$3C,$67,$99,$BB,$AF,$9E,$76,$E4,$EA,$7A,$EF,$C2,$E6
	.dc.b	$FB,$57,$2C,$3D,$93,$F1,$2B,$6F,$59,$FF,$33,$4F,$CB,$8F,$D1,$B5
	.dc.b	$F6,$1F,$A6,$F8,$F7,$82,$F9,$6F,$F3,$74,$61,$B1,$99,$FF,$3A,$CE
	.dc.b	$5F,$F3,$BD,$3E,$3D,$E9,$78,$9B,$3A,$6B,$E4,$6B,$F8,$27,$79,$FD
	.dc.b	$A6,$EB,$FE,$40,$F5,$1E,$EF,$F8,$EE,$B5,$87,$C5,$BD,$E1,$F8,$86
	.dc.b	$AB,$E3,$ED,$6E,$A4,$FD,$2C,$BD,$51,$F6,$2B,$73,$9B,$E7,$5A,$8C
	.dc.b	$56,$F1,$F2,$65,$EB,$8E,$B9,$79,$AB,$EF,$E3,$D2,$30,$7E,$7F,$AF
	.dc.b	$F8,$47,$C0,$FF,$95,$FA,$BF,$88,$B7,$56,$1F,$C7,$7E,$43,$0F,$F1
	.dc.b	$F9,$B7,$FC,$46,$6F,$57,$F9,$22,$FF,$97,$46,$2B,$3F,$1F,$AE,$6B
	.dc.b	$5F,$CC,$DB,$8F,$F3,$39,$E7,$C3,$8C,$FF,$C0,$3F,$3A,$F8,$DF,$F0
	.dc.b	$1D,$FF,$9D,$A3,$18,$FF,$AE,$5B,$1F,$EB,$E6,$BF,$CE,$1C,$D7,$4A
	.dc.b	$F6,$CB,$F4,$57,$C0,$E2,$F7,$87,$7F,$8E,$F8,$7D,$2F,$C0,$1F,$0B
	.dc.b	$B9,$CC,$62,$BF,$D2,$D6,$9F,$4E,$5F,$3A,$31,$3E,$99,$CC,$A7,$2F
	.dc.b	$AE,$3B,$4F,$FA,$87,$27,$C1,$E2,$9A,$FF,$A4,$F3,$F2,$F9,$7E,$8D
	.dc.b	$3F,$B2,$E3,$E4,$D7,$13,$EF,$7C,$3A,$7C,$29,$F9,$91,$83,$CB,$F5
	.dc.b	$FF,$C9,$23,$91,$FE,$B3,$71,$97,$8A,$3E,$B4,$FE,$1B,$E4,$E4,$7F
	.dc.b	$B2,$F2,$F9,$3E,$56,$2F,$81,$F1,$73,$F1,$92,$BE,$F3,$E2,$F2,$7E
	.dc.b	$CE,$5F,$E9,$39,$FE,$FC,$62,$3F,$2F,$F9,$5F,$DA,$BF,$F6,$CB,$7A
	.dc.b	$8F,$FB,$9E,$BF,$DB,$2E,$E7,$FD,$DB,$E1,$7D,$4E,$7F,$E3,$75,$FE
	.dc.b	$02,$9F,$D3,$C6,$1C,$35,$FF,$65,$DF,$6F,$93,$F1,$9E,$4E,$5B,$7E
	.dc.b	$0F,$90,$F3,$9E,$8F,$93,$B1,$FE,$DF,$C5,$EC,$78,$C9,$1E,$3C,$93
	.dc.b	$A8,$FF,$39,$68,$F8,$FF,$86,$7C,$4F,$91,$B3,$F0,$3E,$3B,$E3,$FC
	.dc.b	$8F,$C8,$6C,$7F,$BB,$FF,$0C,$FF,$01,$3D,$DB,$97,$C3,$6B,$FE,$26
	.dc.b	$6B,$FC,$33,$83,$E7,$FC,$87,$D4,$3E,$AF,$95,$6F,$E8,$F3,$7D,$8E
	.dc.b	$AF,$B2,$D8,$FF,$80,$1F,$37,$FA,$4C,$7F,$C6,$6D,$7F,$DF,$2D,$7F
	.dc.b	$8D,$B7,$1F,$BF,$DF,$F8,$43,$F0,$BF,$23,$6F,$F9,$0F,$81,$FE,$90
	.dc.b	$F9,$BF,$EA,$11,$83,$3F,$65,$F0,$27,$5E,$37,$F2,$65,$19,$9F,$E5
	.dc.b	$79,$AF,$9B,$F8,$96,$BF,$E4,$C6,$BC,$1A,$FF,$34,$78,$D7,$82,$7C
	.dc.b	$93,$FC,$03,$F4,$67,$57,$F1,$CB,$DD,$7D,$1E,$77,$C9,$FE,$90,$FD
	.dc.b	$2F,$CA,$59,$FD,$0C,$5C,$7F,$67,$BF,$E9,$AB,$19,$FF,$A7,$70,$7C
	.dc.b	$A7,$EB,$8F,$D7,$FA,$17,$2F,$D5,$C5,$C7,$9D,$7F,$67,$9C,$9F,$B0
	.dc.b	$8F,$DB,$9F,$B8,$F4,$6C,$64,$91,$CA,$32,$F4,$2E,$1E,$58,$B3,$F9
	.dc.b	$1F,$3F,$FD,$82,$FC,$8C,$D1,$FE,$DE,$8C,$3A,$BF,$82,$E5,$BF,$23
	.dc.b	$FD,$8A,$35,$87,$FA,$FD,$7C,$8E,$67,$5F,$21,$30,$EB,$F2,$37,$1E
	.dc.b	$B3,$71,$E5,$36,$BF,$13,$9F,$F8,$1A,$7F,$BC,$4F,$D0,$9F,$C0,$DF
	.dc.b	$F8,$93,$AF,$F2,$53,$98,$FE,$56,$66,$7F,$83,$5F,$F8,$42,$3F,$D3
	.dc.b	$6D,$7B,$D7,$F5,$28,$F0,$4E,$0F,$C7,$DA,$FF,$C6,$71,$E2,$9D,$DB
	.dc.b	$47,$E8,$EF,$88,$FF,$04,$B6,$FE,$92,$F8,$9C,$CC,$9C,$BC,$23,$17
	.dc.b	$F9,$48,$FF,$5C,$CF,$2F,$88,$FF,$0A,$C6,$27,$F3,$36,$9C,$FC,$44
	.dc.b	$BF,$0F,$BD,$6F,$E3,$FE,$3B,$E3,$9F,$7F,$97,$FC,$F7,$74,$7E,$F2
	.dc.b	$75,$5F,$D6,$25,$5B,$AD,$69,$FD,$F2,$F7,$7F,$EC,$A6,$BA,$9F,$E7
	.dc.b	$51,$D2,$FB,$CF,$F8,$D2,$31,$3F,$C5,$AF,$F9,$F4,$FA,$2F,$FC,$37
	.dc.b	$E8,$FE,$4A,$BF,$3D,$AA,$33,$C9,$F1,$9B,$3F,$8A,$78,$5F,$CD,$F4
	.dc.b	$32,$FC,$FF,$84,$F9,$BF,$0B,$F3,$9B,$7D,$9C,$3E,$0F,$85,$B5,$FA
	.dc.b	$AF,$91,$FF,$10,$F2,$7E,$C1,$BD,$46,$92,$FF,$66,$FE,$BF,$E4,$BF
	.dc.b	$60,$FE,$C7,$E4,$3F,$65,$F2,$56,$BF,$72,$DC,$7F,$73,$3F,$B7,$F9
	.dc.b	$2C,$72,$FE,$DF,$D8,$F0,$93,$0F,$EF,$9A,$7F,$04,$8C,$61,$FC,$09
	.dc.b	$FE,$01,$F2,$5F,$C0,$5F,$F2,$4C,$7F,$B4,$0F,$C7,$79,$9F,$D0,$30
	.dc.b	$7C,$8F,$DA,$EC,$FF,$1B,$D7,$F4,$43,$FD,$A6,$70,$3F,$A8,$C6,$EB
	.dc.b	$FC,$1D,$F0,$DF,$E7,$4F,$9D,$E3,$5F,$E3,$EA,$FC,$2C,$62,$3C,$4E
	.dc.b	$B0,$F0,$E9,$C7,$0E,$74,$AD,$5F,$29,$CC,$62,$B8,$57,$9F,$0A,$F8
	.dc.b	$62,$BE,$0A,$F8,$17,$CE,$38,$1E,$D3,$81,$4F,$91,$8F,$77,$07,$C4
	.dc.b	$C3,$07,$B5,$C5,$76,$BF,$0C,$60,$F9,$2E,$30,$F3,$E3,$F0,$A7,$E0
	.dc.b	$57,$C2,$C3,$C0,$75,$AF,$F5,$0B,$88,$F8,$67,$79,$E2,$7B,$30,$F2
	.dc.b	$C6,$A3,$C2,$68,$FF,$00,$D7,$8B,$1F,$89,$18,$3E,$26,$1E,$5C,$3C
	.dc.b	$51,$89,$FC,$13,$DE,$FC,$58,$68,$F0,$C6,$4F,$0B,$98,$D9,$E5,$8F
	.dc.b	$A3,$19,$77,$1F,$0C,$F7,$57,$88,$F8,$51,$D4,$E3,$1C,$EF,$E6,$56
	.dc.b	$E3,$89,$CE,$7E,$1D,$65,$B1,$97,$F0,$71,$F0,$3B,$7F,$9B,$47,$E2
	.dc.b	$1A,$FC,$76,$BF,$AF,$73,$F1,$5F,$E6,$4C,$3F,$DB,$31,$8F,$6F,$FD
	.dc.b	$47,$48,$C4,$EA,$FF,$C6,$B7,$C6,$B5,$18,$BE,$2F,$EC,$30,$FF,$0F
	.dc.b	$CF,$E1,$9F,$F1,$E9,$87,$F3,$D5,$CE,$7F,$6E,$D9,$DF,$91,$FE,$44
	.dc.b	$9E,$8D,$8F,$06,$1D,$6F,$FD,$E7,$7E,$5C,$B7,$1F,$B2,$AF,$04,$76
	.dc.b	$C7,$DE,$DB,$E5,$5F,$EC,$6C,$3B,$3F,$EF,$EB,$F4,$63,$B8,$F4,$E1
	.dc.b	$CA,$7C,$F1,$DE,$BF,$CB,$78,$6C,$93,$A9,$FD,$03,$E4,$3D,$6F,$E8
	.dc.b	$F1,$F5,$46,$E7,$D7,$39,$BF,$03,$F0,$1F,$A3,$3D,$70,$D9,$E7,$3C
	.dc.b	$33,$EB,$C3,$CB,$3D,$4F,$0D,$EA,$7E,$11,$EB,$3F,$8D,$E1,$E4,$7C
	.dc.b	$18,$FC,$8A,$F8,$A7,$84,$F5,$E1,$CD,$F7,$4F,$84,$E7,$8F,$A9,$5F
	.dc.b	$57,$C2,$7C,$6C,$7C,$EF,$80,$F8,$58,$78,$63,$EB,$C7,$0C,$3E,$06
	.dc.b	$1E,$C3,$DD,$1E,$13,$89,$31,$F9,$A4,$8E,$11,$DE,$30,$F1,$7E,$36
	.dc.b	$1F,$97,$8E,$95,$F4,$2B,$E0,$9F,$E6,$BF,$93,$FE,$BD,$B5,$AF,$F3
	.dc.b	$4C,$61,$FF,$21,$73,$FF,$B6,$1C,$9F,$DB,$94,$7F,$B6,$78,$BF,$CE
	.dc.b	$6D,$8C,$CE,$FB,$BA,$57,$F9,$E1,$B3,$FE,$4D,$9F,$F3,$CF,$06,$27
	.dc.b	$3D,$F2,$7F,$CF,$B1,$C3,$E4,$E4,$7F,$A0,$1E,$7F,$17,$88,$FE,$ED
	.dc.b	$8F,$FA,$1A,$7C,$84,$8F,$F0,$3F,$D8,$3D,$7F,$2B,$63,$FD,$1C,$D7
	.dc.b	$FD,$56,$BD,$9D,$CF,$0F,$66,$57,$08,$F9,$4B,$F4,$8E,$5C,$DF,$E4
	.dc.b	$48,$C4,$EA,$73,$C3,$FE,$EF,$C5,$E9,$66,$BC,$BE,$9C,$C3,$FD,$13
	.dc.b	$87,$04,$3F,$49,$7A,$9E,$26,$C7,$FA,$41,$AF,$FA,$3B,$E7,$79,$3C
	.dc.b	$67,$FC,$3C,$FF,$B8,$A7,$6F,$F8,$85,$AE,$63,$6F,$8A,$3A,$E9,$EF
	.dc.b	$A6,$BF,$11,$B5,$E3,$BF,$C2,$FD,$8B,$D5,$A4,$9F,$F6,$04,$3E,$73
	.dc.b	$11,$FE,$DE,$8F,$F5,$CB,$AF,$3B,$39,$7F,$9C,$B9,$15,$FE,$57,$AE
	.dc.b	$7E,$47,$80,$F0,$F9,$36,$32,$43,$B3,$F7,$3E,$79,$F2,$94,$93,$FE
	.dc.b	$BC,$F9,$E7,$C9,$E2,$3E,$5D,$70,$7E,$3B,$5C,$D7,$CA,$51,$9C,$FF
	.dc.b	$11,$6F,$96,$FC,$6B,$6B,$FD,$97,$5B,$E9,$FF,$74,$79,$1F,$AE,$24
	.dc.b	$FF,$B9,$BC,$4A,$FF,$B9,$A2,$E3,$FE,$E3,$9F,$DB,$FE,$5F,$CE,$BF
	.dc.b	$D8,$8D,$E0,$34,$7A,$FC,$35,$E3,$F2,$32,$FF,$DF,$15,$BF,$9B,$EA
	.dc.b	$48,$CF,$CE,$55,$9F,$D9,$A5,$66,$31,$57,$5F,$D0,$E7,$E4,$AC,$61
	.dc.b	$E8,$AF,$5F,$AF,$23,$C6,$FB,$BD,$8C,$47,$83,$0E,$79,$2B,$39,$78
	.dc.b	$EF,$E2,$FC,$DC,$C3,$DA,$FC,$8F,$9C,$C4,$73,$3E,$4C,$62,$BD,$93
	.dc.b	$F8,$BF,$99,$95,$F6,$9F,$E6,$52,$FD,$2B,$F4,$F3,$CC,$FC,$2C,$FE
	.dc.b	$1F,$90,$D2,$47,$E3,$9E,$7F,$29,$FE,$1F,$BF,$F6,$A8,$71,$AF,$F0
	.dc.b	$AA,$EA,$7D,$8F,$6F,$F0,$64,$7E,$DF,$BB,$FC,$7B,$E8,$13,$F9,$96
	.dc.b	$CF,$F0,$46,$BF,$C3,$1A,$F7,$E3,$FC,$5F,$CF,$7E,$43,$DF,$93,$F8
	.dc.b	$88,$FC,$9F,$6C,$8F,$C5,$BF,$96,$ED,$FE,$5F,$F8,$BB,$E5,$FF,$54
	.dc.b	$DA,$FE,$2A,$D7,$F8,$EB,$5F,$17,$3F,$CA,$7F,$2F,$EA,$FE,$88,$FD
	.dc.b	$1F,$17,$E9,$3E,$5A,$4F,$E7,$1F,$D3,$FC,$7F,$8A,$3C,$24,$3C,$27
	.dc.b	$F6,$7B,$FE,$85,$C9,$F9,$DE,$3F,$D7,$9F,$B0,$F9,$9F,$EC,$FF,$A6
	.dc.b	$3E,$7A,$93,$0F,$E7,$F7,$FC,$76,$5F,$49,$FD,$43,$EB,$BF,$93,$57
	.dc.b	$F3,$46,$BF,$CF,$1A,$FF,$B6,$DA,$FE,$AA,$7E,$35,$75,$E0,$F1,$9E
	.dc.b	$4F,$9C,$FF,$03,$7C,$EB,$EB,$9D,$F8,$6E,$3E,$6D,$BF,$AA,$D6,$EB
	.dc.b	$DB,$F1,$9E,$7F,$2B,$F9,$06,$BF,$E7,$FC,$3D,$11,$83,$FC,$1D,$AB
	.dc.b	$E5,$7E,$29,$27,$B0,$C5,$F1,$3F,$EF,$A8,$CC,$79,$FB,$3F,$BA,$FB
	.dc.b	$FF,$C6,$27,$88,$70,$FE,$E6,$78,$3A,$8C,$BF,$A7,$38,$79,$9F,$CE
	.dc.b	$63,$31,$FA,$B6,$BE,$5B,$7B,$FF,$29,$B1,$E6,$6B,$FB,$F9,$F1,$F1
	.dc.b	$FF,$82,$5F,$D5,$B7,$C2,$F5,$9F,$53,$F4,$CF,$EE,$3C,$5E,$18,$F2
	.dc.b	$12,$FF,$BB,$DF,$EC,$9D,$1F,$1B,$F1,$87,$CF,$7D,$E6,$C7,$C1,$AF
	.dc.b	$93,$B4,$8F,$88,$F8,$5C,$1F,$E0,$AB,$47,$C5,$24,$7F,$83,$BC,$47
	.dc.b	$0F,$68,$FE,$51,$18,$C7,$F4,$67,$E9,$3E,$7E,$CF,$E0,$27,$8F,$D0
	.dc.b	$7E,$A0,$CC,$72,$B7,$D0,$7E,$B3,$E6,$24,$F1,$9C,$B6,$32,$F1,$7F
	.dc.b	$67,$3F,$E3,$CF,$A5,$FE,$5F,$CF,$FE,$30,$F9,$1F,$C4,$5F,$F8,$CA
	.dc.b	$72,$A3,$2A,$33,$18,$7F,$C0,$4D,$F1,$7F,$44,$DA,$ED,$BF,$F3,$5A
	.dc.b	$E2,$DF,$5B,$F9,$C5,$6F,$0F,$04,$EF,$0E,$6D,$C3,$9C,$F9,$EF,$8C
	.dc.b	$E6,$38,$9E,$BA,$E2,$DA,$E2,$FA,$EB,$8B,$93,$89,$EF,$9E,$2D,$9E
	.dc.b	$38,$7C,$0B,$F0,$4F,$EA,$E3,$E4,$B6,$BF,$82,$36,$BF,$C9,$FF,$1D
	.dc.b	$FB,$EE,$AB,$D1,$FC,$06,$BF,$CB,$33,$96,$C6,$5E,$DE,$33,$E6,$3F
	.dc.b	$CB,$39,$23,$13,$FE,$62,$8C,$4E,$A7,$33,$CA,$3D,$71,$C3,$5F,$BF
	.dc.b	$F0,$2E,$07,$4E,$AE,$61,$F1,$B1,$EA,$FC,$8B,$E1,$5A,$C3,$F0,$D8
	.dc.b	$F5,$DF,$F0,$4D,$BF,$F4,$0D,$F0,$7A,$46,$5F,$C8,$37,$0E,$F7,$E6
	.dc.b	$CB,$CB,$5C,$32,$F0,$63,$D9,$FB,$F8,$F5,$BF,$1C,$7E,$2B,$93,$F7
	.dc.b	$3F,$15,$FE,$0C,$D3,$FA,$53,$71,$4F,$EF,$BE,$13,$0F,$0A,$FD,$94
	.dc.b	$63,$97,$FD,$5D,$18,$3F,$6D,$79,$F8,$FD,$7C,$4D,$9F,$DD,$37,$1F
	.dc.b	$E0,$13,$CF,$E7,$F1,$E0,$7F,$AC,$7E,$8F,$FD,$81,$87,$FA,$61,$B7
	.dc.b	$FE,$A0,$6B,$FC,$F5,$AF,$F8,$95,$DC,$F3,$C3,$D1,$F1,$9E,$FA,$56
	.dc.b	$AF,$84,$FB,$7E,$33,$53,$FE,$BE,$9F,$E2,$51,$89,$FF,$63,$57,$09
	.dc.b	$F3,$46,$23,$FD,$7F,$5C,$5F,$44,$71,$9F,$E3,$9C,$36,$7F,$91,$7C
	.dc.b	$9F,$CD,$3C,$DE,$D3,$E0,$31,$7F,$D8,$A3,$0F,$FA,$F1,$B5,$FE,$C4
	.dc.b	$6B,$FD,$85,$AF,$F8,$27,$8F,$AD,$7D,$B6,$D7,$F3,$7F,$D3,$7F,$D7
	.dc.b	$4F,$F3,$1A,$FC,$1C,$61,$FE,$70,$79,$BC,$DF,$80,$FF,$95,$DB,$87
	.dc.b	$F8,$42,$BC,$D1,$F1,$79,$57,$FE,$CB,$9F,$07,$C5,$F6,$F8,$AE,$EF
	.dc.b	$83,$E7,$7C,$3F,$3B,$6B,$CC,$F9,$3E,$2B,$A6,$06,$92,$7F,$AD,$43
	.dc.b	$E6,$FD,$67,$B3,$E3,$2C,$FC,$1F,$9B,$F7,$FC,$DF,$C0,$6C,$7C,$5F
	.dc.b	$9B,$EE,$47,$AB,$E2,$F9,$BF,$8E,$7C,$8F,$1A,$CF,$CD,$F9,$9F,$96
	.dc.b	$7C,$CF,$9A,$B1,$F4,$7E,$6B,$D6,$47,$D2,$FB,$BC,$6F,$A6,$7D,$4F
	.dc.b	$9A,$B3,$FE,$26,$FA,$27,$BF,$F5,$1B,$1F,$E2,$3E,$A0,$8C,$B9,$6B
	.dc.b	$FE,$81,$E2,$FF,$11,$CF,$CA,$F6,$BE,$79,$1F,$9B,$E8,$7E,$18,$FC
	.dc.b	$3F,$CA,$58,$FF,$01,$69,$FE,$F4,$6F,$B9,$FC,$3A,$31,$8F,$F8,$31
	.dc.b	$FF,$6B,$CF,$0F,$53,$F1,$11,$A9,$FF,$4A,$35,$FF,$66,$B5,$FE,$36
	.dc.b	$F4,$F2,$3F,$94,$F7,$23,$17,$C3,$E5,$3C,$58,$78,$FE,$53,$C9,$E2
	.dc.b	$5A,$F4,$B7,$2F,$33,$72,$EB,$8F,$EC,$9B,$F2,$9E,$DF,$94,$F7,$37
	.dc.b	$0F,$86,$DC,$3E,$03,$6B,$9F,$C8,$FC,$5F,$17,$E3,$36,$FE,$5B,$F1
	.dc.b	$08,$FF,$8D,$F1,$46,$30,$F9,$8F,$CF,$F9,$4E,$75,$F5,$3E,$47,$E9
	.dc.b	$7C,$0F,$D3,$6C,$64,$97,$F5,$7E,$07,$51,$F7,$3E,$47,$ED,$1F,$6F
	.dc.b	$E4,$6C,$64,$95,$F7,$63,$2F,$FB,$9F,$E7,$7F,$86,$6D,$FE,$E0,$FD
	.dc.b	$39,$E2,$FF,$37,$9C,$E1,$FE,$AC,$9C,$BF,$E8,$48,$FE,$29,$7F,$E7
	.dc.b	$F6,$BF,$EE,$76,$BF,$A0,$6A,$7A,$B9,$F9,$9F,$EC,$98,$2D,$E3,$78
	.dc.b	$E7,$F3,$1F,$05,$F9,$9F,$93,$B9,$7E,$79,$B8,$FE,$6D,$B8,$FE,$19
	.dc.b	$FD,$07,$C8,$FE,$87,$E0,$AD,$FE,$99,$B7,$FA,$39,$FD,$37,$C9,$E3
	.dc.b	$0F,$F0,$1C,$BF,$40,$FD,$3A,$8F,$F0,$38,$FD,$3D,$C5,$FD,$86,$B7
	.dc.b	$3A,$9C,$E5,$FE,$AE,$9E,$77,$FE,$A9,$6B,$FC,$20,$F0,$78,$3F,$D0
	.dc.b	$23,$31,$FC,$31,$EB,$3B,$6C,$79,$69,$A7,$8E,$FF,$C1,$FF,$3B,$FE
	.dc.b	$EB,$C3,$F8,$39,$8F,$95,$FE,$0C,$D3,$FD,$19,$1B,$C3,$FD,$26,$FF
	.dc.b	$BB,$A3,$33,$D3,$E4,$BA,$F9,$56,$78,$B6,$F8,$CF,$8B,$E5,$31,$AF
	.dc.b	$F0,$8C,$5F,$1D,$FD,$65,$B8,$FF,$A9,$5B,$87,$FB,$3F,$C7,$46,$23
	.dc.b	$E4,$46,$2B,$DE,$7C,$0F,$92,$B7,$EC,$F8,$EF,$87,$F1,$FF,$11,$B1
	.dc.b	$92,$61,$F0,$3E,$43,$51,$F1,$3E,$27,$CC,$47,$CC,$D9,$F8,$EF,$D0
	.dc.b	$F8,$DF,$A2,$FF,$89,$63,$FD,$50,$3C,$E7,$C5,$7E,$07,$77,$AF,$FE
	.dc.b	$54,$9C,$D7,$FB,$BF,$E3,$7F,$A6,$C6,$23,$F2,$93,$BB,$FE,$A5,$5F
	.dc.b	$D4,$BE,$83,$58,$74,$CF,$15,$AC,$15,$AC,$3F,$CC,$6F,$58,$F3,$B9
	.dc.b	$AE,$31,$E3,$38,$D7,$A6,$B8,$BD,$6B,$84,$E8,$E0,$E0,$EC,$FF,$B1
	.dc.b	$CF,$89,$1C,$EB,$DE,$F9,$F1,$E6,$70,$C7,$C7,$5F,$0A,$AB,$E8,$AE
	.dc.b	$4F,$84,$FF,$63,$97,$1F,$88,$E4,$E0,$7F,$7A,$73,$87,$78,$F2,$99
	.dc.b	$C7,$E8,$3F,$0F,$1F,$2D,$FC,$13,$C3,$5C,$4F,$80,$E6,$79,$61,$E8
	.dc.b	$7E,$14,$E9,$E9,$3E,$AA,$F1,$1D,$A3,$E5,$C7,$63,$A3,$F0,$CF,$61
	.dc.b	$D1,$D3,$C0,$F0,$9A,$C7,$E5,$E3,$F0,$27,$9B,$67,$C6,$70,$AF,$1C
	.dc.b	$F9,$4D,$E3,$F8,$29,$EC,$78,$F0,$DB,$C6,$7E,$E1,$C5,$EA,$7A,$6B
	.dc.b	$6D,$8F,$CC,$7A,$7F,$A9,$8F,$C9,$35,$FB,$AD,$7F,$5E,$E7,$E4,$1E
	.dc.b	$E7,$F3,$B9,$4E,$63,$0D,$CB,$FC,$76,$D9,$DE,$1F,$CD,$EF,$9D,$7E
	.dc.b	$5A,$7F,$C8,$38,$3E,$1F,$F9,$E5,$FE,$DE,$31,$3B,$17,$C9,$FC,$69
	.dc.b	$8C,$7C,$D5,$C6,$F9,$DF,$F4,$73,$9D,$70,$BE,$55,$FC,$15,$E9,$97
	.dc.b	$19,$FD,$3D,$70,$8D,$61,$F0,$DB,$8F,$96,$7A,$E1,$C6,$FB,$5F,$92
	.dc.b	$7E,$13,$C5,$F2,$63,$FD,$6F,$0E,$07,$4A,$FD,$96,$3E,$4A,$DC,$79
	.dc.b	$2B,$84,$74,$C7,$F7,$B5,$BA,$F9,$B5,$A6,$9A,$C3,$C4,$FF,$9C,$A7
	.dc.b	$E1,$DF,$68,$F9,$AF,$B0,$EC,$FB,$2F,$C4,$C3,$CF,$5D,$1F,$A9,$19
	.dc.b	$C3,$DB,$5C,$9F,$76,$3F,$1C,$91,$D9,$F8,$51,$C2,$BD,$27,$6A,$F5
	.dc.b	$1D,$6B,$D1,$3C,$1E,$47,$AE,$BA,$D7,$BE,$FE,$A3,$6F,$93,$DA,$F9
	.dc.b	$39,$3F,$92,$5F,$E0,$CE,$E7,$AC,$8F,$17,$2F,$18,$F7,$C7,$8C,$91
	.dc.b	$D6,$FF,$AC,$F2,$FF,$78,$F8,$CF,$EB,$DB,$FF,$A9,$A3,$15,$FE,$10
	.dc.b	$9C,$A8,$FE,$FF,$F1,$F8,$77,$1F,$E0,$3E,$85,$7F,$AC,$63,$84,$61
	.dc.b	$FF,$05,$46,$67,$FC,$13,$F1,$7C,$E8,$D3,$1F,$1D,$6F,$FC,$47,$9E
	.dc.b	$CC,$3F,$E2,$4F,$8C,$E4,$47,$FC,$53,$F1,$D5,$F3,$47,$87,$03,$4C
	.dc.b	$47,$8C,$87,$F8,$CB,$8A,$BF,$E3,$2F,$8D,$F4,$18,$E1,$FF,$58,$FE
	.dc.b	$A0,$F8,$CF,$8C,$DA,$FE,$B5,$C3,$F3,$E3,$11,$FE,$5B,$8C,$79,$BC
	.dc.b	$2B,$8F,$C6,$CA,$E2,$7D,$3B,$E3,$87,$F9,$9B,$EF,$58,$7D,$8A,$46
	.dc.b	$1E,$26,$23,$EE,$CF,$9F,$C3,$E1,$5B,$AF,$F7,$53,$5F,$F3,$21,$C4
	.dc.b	$FB,$CE,$5F,$F3,$43,$C6,$7F,$CE,$B8,$BC,$1E,$ED,$AF,$1D,$6F,$3F
	.dc.b	$89,$F1,$9F,$20,$93,$F0,$39,$5A,$3E,$37,$33,$F2,$4F,$89,$F1,$DF
	.dc.b	$40,$CE,$9F,$53,$E3,$BE,$9F,$C7,$4A,$E9,$82,$7E,$B9,$98,$E9,$8F
	.dc.b	$82,$B3,$1F,$2A,$9F,$1B,$F3,$DB,$19,$8F,$A5,$F4,$0F,$37,$51,$FA
	.dc.b	$59,$F5,$B6,$B8,$C7,$CE,$32,$7E,$AD,$CE,$5F,$EA,$4B,$7C,$CF,$63
	.dc.b	$AD,$7B,$3D,$BF,$C8,$91,$F7,$FC,$85,$7D,$B8,$7F,$B4,$9E,$06,$8F
	.dc.b	$8B,$E2,$7E,$68,$8F,$C8,$F6,$AB,$CB,$3F,$CF,$98,$CF,$F4,$7F,$07
	.dc.b	$FA,$22,$4F,$D0,$F9,$0A,$FF,$A2,$F1,$4F,$F0,$56,$C7,$C3,$E0,$FC
	.dc.b	$0C,$E4,$FD,$6F,$95,$27,$ED,$C7,$EC,$3E,$43,$F6,$3F,$21,$27,$EF
	.dc.b	$C6,$5F,$C0,$C6,$5F,$87,$1F,$B8,$F2,$7F,$73,$F1,$D2,$7F,$0F,$19
	.dc.b	$7F,$13,$19,$7E,$5C,$7F,$25,$F9,$D7,$C7,$D2,$4F,$E4,$3C,$FA,$FF
	.dc.b	$AA,$E7,$6F,$BD,$AE,$67,$F9,$44,$63,$77,$6F,$F3,$98,$7B,$3E,$13
	.dc.b	$AC,$EA,$FF,$9D,$4B,$E3,$7C,$C4,$8F,$D7,$43,$E3,$FF,$85,$7D,$BF
	.dc.b	$F8,$D2,$7F,$9E,$35,$FC,$B3,$5F,$E0,$AD,$7F,$DD,$5F,$3C,$F8,$BF
	.dc.b	$D0,$36,$3F,$A3,$35,$FD,$03,$5F,$F6,$5D,$BE,$43,$E6,$12,$3C,$FF
	.dc.b	$23,$FC,$7F,$D5,$FE,$77,$1F,$D4,$5A,$FE,$AD,$AF,$F1,$A6,$BF,$27
	.dc.b	$E4,$3E,$C1,$1F,$87,$F2,$3F,$6F,$E4,$64,$7C,$67,$EF,$7C,$8F,$DF
	.dc.b	$D3,$EA,$92,$7E,$D3,$F6,$3E,$47,$F0,$9F,$23,$27,$E7,$C7,$E1,$FE
	.dc.b	$2A,$47,$D2,$7F,$15,$F2,$3F,$8B,$F8,$9F,$C0,$92,$7F,$0C,$FE,$13
	.dc.b	$E3,$BF,$21,$F1,$F2,$7E,$DC,$7E,$4F,$E3,$E4,$7D,$E7,$F2,$DF,$1F
	.dc.b	$FD,$FB,$E6,$1F,$38,$DF,$FB,$D4,$61,$B1,$9B,$FE,$2D,$39,$7F,$6D
	.dc.b	$F0,$7F,$C5,$1F,$C5,$F7,$7E,$8C,$FF,$04,$46,$63,$F9,$6B,$FC,$95
	.dc.b	$FE,$BB,$3E,$2D,$FC,$96,$78,$FA,$38,$7C,$64,$AF,$1C,$ED,$F1,$FC
	.dc.b	$7F,$80,$8F,$68,$D3,$FD,$26,$DF,$37,$FE,$7A,$F0,$3F,$51,$18,$F8
	.dc.b	$0F,$8D,$39,$D3,$FC,$51,$A3,$DD,$CD,$7C,$8A,$FF,$41,$E2,$26,$1C
	.dc.b	$DF,$93,$1A,$72,$F5,$7F,$C6,$16,$F9,$0F,$92,$76,$8F,$8F,$F4,$DF
	.dc.b	$DA,$55,$A3,$0D,$F8,$0F,$A3,$F2,$1F,$7E,$7F,$BC,$57,$06,$D6,$E3
	.dc.b	$EE,$C6,$BD,$0F,$F1,$AF,$C7,$FD,$A2,$4F,$F7,$5C,$9F,$23,$88,$FB
	.dc.b	$BF,$07,$F8,$03,$1C,$3F,$C4,$F0,$FF,$1F,$FC,$8F,$E1,$3E,$46,$5F
	.dc.b	$38,$AF,$E1,$A2,$D7,$1C,$55,$FE,$1F,$F9,$B7,$6F,$F1,$7F,$91,$7A
	.dc.b	$9F,$47,$25,$6B,$D2,$FA,$95,$A9,$FA,$D3,$95,$19,$78,$F9,$FC,$E3
	.dc.b	$13,$B6,$E1,$E1,$C3,$C1,$39,$6D,$71,$73,$3B,$BF,$0B,$5F,$3B,$6B
	.dc.b	$F9,$A4,$7A,$27,$8E,$16,$38,$B5,$E2,$D9,$E3,$1E,$C9,$E2,$D3,$89
	.dc.b	$4E,$25,$8E,$38,$7B,$AB,$FC,$FD,$1F,$1E,$7F,$C8,$AD,$7A,$9F,$E4
	.dc.b	$8F,$91,$FF,$13,$F1,$FF,$A1,$5B,$1F,$E1,$F6,$BB,$F6,$FF,$36,$FF
	.dc.b	$88,$E9,$E1,$BE,$2B,$C2,$D9,$D3,$5F,$1B,$5F,$A6,$EF,$2F,$B2,$DC
	.dc.b	$FF,$0D,$F1,$6E,$2F,$ED,$15,$AC,$3E,$55,$EF,$0F,$C3,$B5,$FF,$45
	.dc.b	$D7,$0C,$3B,$D7,$02,$4F,$CD,$C3,$B7,$76,$31,$ED,$3F,$3F,$1E,$95
	.dc.b	$A8,$C5,$EF,$0E,$B1,$CF,$1F,$03,$E4,$9F,$C2,$72,$7F,$AB,$3C,$EF
	.dc.b	$F0,$17,$2F,$EA,$7E,$4A,$EB,$FC,$6A,$6E,$1F,$C4,$26,$DF,$1A,$FD
	.dc.b	$F4,$63,$99,$76,$31,$F2,$19,$F9,$1B,$7F,$C1,$5B,$8F,$FA,$2B,$1F
	.dc.b	$F5,$FD,$7F,$85,$63,$F6,$D1,$8A,$FE,$97,$39,$9F,$E5,$3F,$56,$8C
	.dc.b	$47,$F2,$08,$C4,$FF,$54,$6C,$7F,$A6,$39,$51,$88,$FE,$09,$18,$AF
	.dc.b	$82,$DA,$FE,$39,$F2,$58,$BE,$7A,$FF,$44,$5F,$25,$F0,$F6,$D3,$6F
	.dc.b	$FA,$DC,$E9,$F2,$47,$A3,$0F,$1B,$67,$59,$79,$DB,$87,$79,$FA,$07
	.dc.b	$6B,$FE,$61,$C5,$F4,$CF,$4E,$AF,$4F,$13,$FD,$B2,$30,$FF,$78,$6E
	.dc.b	$3F,$15,$AF,$C4,$6B,$FD,$7B,$C5,$EB,$5F,$81,$6D,$7F,$41,$9F,$C2
	.dc.b	$1F,$85,$E2,$B1,$E1,$F4,$BE,$06,$1F,$E4,$38,$F1,$46,$0F,$B7,$C2
	.dc.b	$FD,$45,$F2,$5F,$92,$7F,$DB,$B5,$FE,$C3,$72,$DF,$1B,$5F,$1F,$23
	.dc.b	$5F,$25,$74,$FF,$5D,$DB,$E4,$3B,$B6,$BF,$D1,$CD,$9F,$CB,$35,$FC
	.dc.b	$AA,$F8,$FF,$27,$C8,$F9,$5B,$5E,$5F,$8D,$F4,$1E,$9F,$91,$B1,$E9
	.dc.b	$F8,$FF,$09,$1F,$0B,$E4,$F1,$3E,$07,$C7,$7C,$16,$CF,$C1,$F8,$EF
	.dc.b	$89,$F1,$FF,$15,$B1,$FE,$0F,$EB,$FD,$FF,$E0,$FF,$E7,$B8,$FF,$66
	.dc.b	$42,$31,$BF,$FB,$F2,$31,$E6,$7E,$B1,$B5,$FE,$D3,$97,$CD,$BD,$8F
	.dc.b	$D8,$C6,$1B,$19,$9E,$0D,$9F,$97,$F1,$FF,$6F,$E4,$3E,$E3,$6B,$EE
	.dc.b	$7C,$87,$D4,$23,$F5,$1F,$AF,$F2,$1F,$83,$3F,$09,$E6,$59,$F8,$07
	.dc.b	$F7,$FF,$03,$FC,$50,$E3,$16,$8E,$B7,$81,$FC,$82,$7F,$A6,$7B,$3F
	.dc.b	$C8,$AB,$53,$9C,$39,$D7,$FB,$D5,$AF,$F8,$34,$E3,$87,$C7,$8C,$C7
	.dc.b	$8E,$F9,$1D,$70,$FE,$17,$59,$3F,$39,$E9,$7E,$7C,$FD,$07,$C8,$58
	.dc.b	$C9,$2F,$F3,$71,$97,$DB,$1F,$E4,$FF,$09,$FA,$97,$DE,$D3,$FC,$60
	.dc.b	$FF,$B1,$A7,$E2,$4E,$B9,$3F,$9F,$CE,$6F,$98,$9D,$57,$09,$E8,$F7
	.dc.b	$7F,$DB,$EE,$6B,$9F,$C5,$F4,$F8,$FB,$3C,$5B,$87,$F4,$3B,$FF,$5F
	.dc.b	$4F,$FB,$FD,$B1,$9A,$FE,$0B,$18,$C7,$D7,$1E,$8F,$8C,$E7,$1E,$EF
	.dc.b	$84,$F5,$F9,$5E,$C6,$E3,$E9,$70,$FA,$FE,$32,$E5,$F0,$35,$F9,$3F
	.dc.b	$17,$F1,$3E,$33,$E2,$B6,$FE,$13,$87,$E2,$79,$96,$FE,$48,$F9,$BF
	.dc.b	$C8,$4F,$E4,$F4,$7D,$A7,$FA,$16,$3F,$2D,$5B,$E5,$FC,$94,$E6,$39
	.dc.b	$DF,$27,$F2,$06,$EB,$F6,$D1,$98,$FC,$A3,$5F,$F3,$13,$C6,$FF,$B8
	.dc.b	$36,$7F,$84,$73,$7F,$2E,$AE,$F8,$7E,$13,$E3,$BF,$0B,$F1,$36,$FF
	.dc.b	$04,$DA,$F9,$1C,$8A,$FC,$5F,$C7,$67,$3F,$C8,$FB,$BF,$8E,$F8,$EF
	.dc.b	$C7,$B7,$2F,$C6,$38,$7F,$1D,$F0,$57,$2F,$97,$F3,$7F,$8E,$7B,$FC
	.dc.b	$77,$EE,$63,$18,$FE,$7A,$31,$F2,$9F,$C5,$A7,$85,$7E,$DD,$AF,$FA
	.dc.b	$15,$AF,$FA,$DD,$E7,$86,$67,$F4,$BF,$21,$D2,$3F,$53,$FA,$9B,$E4
	.dc.b	$7F,$53,$1F,$E2,$88,$C3,$FA,$73,$11,$1E,$3F,$2F,$71,$FC,$85,$7F
	.dc.b	$0E,$CF,$FD,$4C,$FF,$29,$6F,$27,$5F,$33,$B3,$30,$F1,$FC,$8D,$C3
	.dc.b	$9F,$C6,$FF,$6C,$C7,$FD,$6B,$5F,$88,$ED,$46,$2B,$FC,$68,$D7,$FC
	.dc.b	$66,$E7,$2F,$1D,$FB,$7E,$43,$DC,$FB,$FE,$37,$E0,$7C,$75,$CB,$E2
	.dc.b	$B7,$1F,$6C,$FC,$5F,$8F,$C6,$3E,$EF,$8A,$F3,$13,$0F,$36,$68,$C6
	.dc.b	$1F,$19,$F9,$DF,$19,$F3,$DF,$E7,$51,$FE,$BC,$1E,$5B,$E2,$BF,$D8
	.dc.b	$58,$BD,$46,$5F,$EF,$5A,$D5,$FF,$3B,$F8,$DC,$61,$C2,$F7,$3D,$23
	.dc.b	$F6,$33,$FE,$FD,$F9,$BE,$33,$FD,$3A,$B7,$18,$8F,$23,$AC,$7C,$58
	.dc.b	$F1,$C3,$9D,$2B,$57,$CA,$73,$18,$AE,$15,$E7,$38,$4F,$9E,$33,$1F
	.dc.b	$ED,$73,$27,$F8,$16,$BA,$3C,$71,$E8,$FB,$4C,$57,$37,$07,$3B,$D9
	.dc.b	$ED,$7F,$DC,$E7,$63,$E1,$5F,$89,$F8,$33,$CE,$78,$63,$E6,$C7,$E1
	.dc.b	$D7,$13,$ED,$61,$E4,$36,$F8,$A3,$71,$8A,$F1,$DF,$C6,$3B,$16,$BE
	.dc.b	$06,$1E,$37,$CF,$3C,$1F,$79,$A3,$DE,$78,$A7,$E1,$57,$77,$AC,$7D
	.dc.b	$19,$E4,$FC,$58,$CB,$4E,$91,$C5,$B1,$E5,$7E,$39,$F0,$4F,$97,$3F
	.dc.b	$06,$3F,$0A,$D8,$F0,$47,$67,$E6,$9D,$0F,$59,$D2,$3E,$49,$C5,$FC
	.dc.b	$23,$5F,$71,$F0,$5F,$9F,$E8,$7E,$AA,$3F,$1C,$D7,$F1,$8D,$7F,$60
	.dc.b	$E7,$E4,$1E,$DF,$F8,$C2,$53,$98,$C3,$72,$4E,$AE,$AB,$E7,$3A,$AE
	.dc.b	$13,$F9,$4C,$1F,$11,$C3,$0F,$D9,$C6,$27,$F7,$67,$6C,$39,$BF,$9C
	.dc.b	$6C,$F0,$AE,$55,$E2,$39,$56,$F1,$F6,$57,$F0,$0B,$E7,$97,$1A,$FD
	.dc.b	$8D,$F0,$8D,$5F,$C3,$6E,$5E,$4A,$FE,$2D,$8F,$81,$F8,$77,$E0,$C7
	.dc.b	$71,$E1,$AE,$B5,$F1,$AF,$C1,$3F,$B1,$C7,$C7,$87,$CA,$6C,$74,$AE
	.dc.b	$13,$ED,$6C,$7C,$5C,$7C,$D3,$C5,$B1,$C4,$F0,$9F,$C3,$AB,$AB,$F5
	.dc.b	$E3,$E2,$CF,$58,$F7,$3E,$B3,$A6,$1F,$06,$B8,$C7,$C7,$38,$4F,$AE
	.dc.b	$BC,$2F,$5A,$F0,$BF,$C3,$E7,$83,$E9,$8F,$24,$F8,$A3,$D0,$7E,$1E
	.dc.b	$FE,$73,$CC,$F9,$87,$37,$E1,$BF,$50,$E7,$31,$D1,$E2,$C3,$EA,$9F
	.dc.b	$5B,$1F,$A6,$4C,$3D,$F3,$DA,$7E,$B9,$23,$2F,$27,$E0,$B9,$8F,$61
	.dc.b	$F0,$EB,$C1,$3A,$38,$9D,$AB,$C1,$3F,$01,$F2,$47,$FA,$17,$83,$FD
	.dc.b	$43,$3F,$E1,$AF,$57,$FB,$2E,$B9,$6C,$7F,$A0,$65,$F1,$F8,$E8,$FF
	.dc.b	$04,$D7,$FB,$15,$B5,$FD,$E2,$31,$5D,$A3,$FC,$1B,$F1,$DC,$60,$D3
	.dc.b	$1F,$1F,$67,$FC,$5B,$96,$CC,$3F,$E2,$6F,$8C,$E4,$47,$FC,$67,$F1
	.dc.b	$B5,$FF,$0D,$C7,$88,$8F,$F8,$30,$F1,$10,$FF,$1D,$F0,$57,$FC,$67
	.dc.b	$ED,$79,$CC,$70,$7F,$80,$B5,$78,$EF,$8B,$F8,$8D,$BF,$F5,$6B,$5F
	.dc.b	$F4,$62,$F8,$CD,$CF,$0F,$8D,$95,$C2,$3E,$8A,$FD,$03,$FC,$44,$FF
	.dc.b	$A1,$9A,$FF,$A3,$9B,$8F,$97,$0F,$F6,$F4,$EF,$B3,$FD,$F4,$F3,$6D
	.dc.b	$72,$8F,$F6,$D3,$5F,$E6,$2F,$4E,$5F,$F6,$65,$3E,$3B,$DE,$79,$A7
	.dc.b	$55,$FE,$C7,$97,$C5,$7C,$42,$47,$C4,$F8,$DA,$FC,$4F,$8D,$F9,$26
	.dc.b	$33,$F5,$7C,$5F,$CC,$24,$FC,$CF,$8D,$AF,$CC,$F8,$DF,$A2,$62,$BF
	.dc.b	$C6,$F0,$F9,$95,$7C,$C6,$C7,$FA,$C1,$AF,$FA,$E3,$E6,$9E,$5E,$EB
	.dc.b	$E4,$D6,$AB,$FE,$00,$35,$FE,$AE,$D7,$FA,$DB,$D2,$FC,$76,$F3,$3E
	.dc.b	$EB,$6B,$C5,$7F,$8A,$F8,$E9,$1F,$8A,$7F,$1B,$F1,$F8,$8F,$C5,$F8
	.dc.b	$BF,$90,$31,$8F,$E0,$7E,$33,$F2,$64,$9F,$C9,$F9,$55,$FC,$9F,$C6
	.dc.b	$FE,$60,$C5,$7E,$27,$C7,$FC,$D9,$27,$F3,$7F,$19,$5F,$EE,$A7,$DF
	.dc.b	$8F,$E0,$5C,$DF,$E0,$38,$FC,$3C,$60,$FD,$2F,$C6,$C8,$FD,$2C,$7F
	.dc.b	$0B,$FC,$5D,$EC,$D2,$4F,$EA,$BC,$EA,$FF,$BD,$65,$F3,$7C,$4C,$E7
	.dc.b	$FB,$58,$7C,$77,$52,$57,$F8,$3A,$75,$F1,$F8,$8E,$93,$67,$31,$FE
	.dc.b	$F4,$F9,$97,$C6,$FE,$D1,$B1,$98,$FD,$BF,$D4,$BE,$2F,$51,$FB,$0A
	.dc.b	$DD,$7F,$BE,$1A,$FC,$A6,$BF,$E2,$07,$77,$F9,$AB,$7C,$6F,$63,$AD
	.dc.b	$7A,$FC,$5F,$F1,$4F,$94,$D5,$F2,$3F,$34,$95,$B1,$3B,$F9,$1F,$E1
	.dc.b	$FF,$17,$FC,$5E,$7E,$EB,$5F,$F1,$9B,$5F,$F1,$BB,$67,$C1,$F2,$1F
	.dc.b	$6F,$E4,$65,$77,$9C,$BE,$11,$F3,$7F,$1D,$FC,$59,$FE,$42,$70,$E1
	.dc.b	$AF,$E4,$1A,$FF,$91,$BE,$71,$D3,$FC,$79,$B1,$D0,$38,$7B,$B5,$FF
	.dc.b	$27,$D7,$C9,$DB,$F2,$9C,$52,$7D,$5B,$7D,$26,$FC,$9F,$E6,$79,$24
	.dc.b	$60,$DE,$DF,$D9,$68,$D2,$47,$FB,$CE,$1F,$1F,$FA,$32,$47,$83,$E3
	.dc.b	$FF,$4D,$DB,$FA,$7F,$8F,$93,$E4,$8F,$D5,$7C,$7F,$72,$47,$57,$C5
	.dc.b	$F1,$BF,$AF,$24,$7B,$3E,$3F,$FB,$E7,$E4,$0F,$8E,$FD,$A7,$54,$AF
	.dc.b	$B5,$39,$7F,$15,$F3,$2F,$2A,$47,$DA,$8C,$47,$D1,$6C,$66,$36,$FF
	.dc.b	$0A,$E5,$53,$E3,$FC,$64,$9F,$0F,$C8,$79,$BD,$0F,$F0,$AE,$3E,$1F
	.dc.b	$2A,$E4,$75,$7E,$68,$C3,$E1,$8F,$24,$62,$3F,$8D,$CF,$37,$F2,$79
	.dc.b	$79,$3E,$43,$E2,$92,$3A,$5F,$8B,$E9,$9F,$23,$49,$3F,$23,$D5,$AF
	.dc.b	$69,$DB,$F1,$DA,$E6,$7E,$44,$63,$C6,$FF,$08,$6F,$78,$3F,$7E,$DB
	.dc.b	$68,$BC,$37,$E4,$3C,$C4,$9F,$B7,$F2,$35,$F2,$47,$80,$8F,$C2,$B4
	.dc.b	$7F,$8F,$DB,$19,$8C,$11,$FC,$3F,$C8,$FE,$30,$FC,$6C,$7E,$3B,$E0
	.dc.b	$3F,$1F,$F2,$32,$BF,$1B,$19,$7F,$1F,$19,$5A,$7C,$29,$FF,$12,$FC
	.dc.b	$DF,$F8,$CB,$D1,$FF,$64,$FC,$93,$D5,$FF,$64,$CE,$70,$4E,$5F,$F6
	.dc.b	$7C,$66,$31,$08,$7A,$5F,$ED,$29,$CC,$6E,$BA,$F7,$FF,$B7,$2F,$AC
	.dc.b	$63,$0F,$1B,$E2,$C3,$93,$99,$D9,$E2,$AE,$58,$76,$6C,$F1,$9E,$66
	.dc.b	$EE,$BE,$79,$DB,$93,$A9,$EB,$BF,$0B,$9B,$ED,$5C,$B0,$F6,$4F,$C4
	.dc.b	$AD,$BD,$67,$F9,$54,$FC,$B8,$FF,$00,$B5,$F6,$1F,$E0,$4F,$92,$FE
	.dc.b	$09,$C3,$FC,$C5,$B1,$FC,$F9,$AE,$FA,$7F,$32,$FF,$42,$A7,$80,$F8
	.dc.b	$8F,$0B,$67,$4D,$7C,$6D,$7F,$02,$EF,$3F,$B2,$DD,$53,$3E,$29,$C2
	.dc.b	$E7,$F8,$97,$BA,$FC,$3B,$5F,$C3,$36,$B8,$CF,$0C,$65,$7D,$2C,$BD
	.dc.b	$71,$F6,$2B,$73,$9B,$E9,$5A,$8C,$56,$F0,$F2,$46,$09,$97,$B2,$75
	.dc.b	$8F,$82,$3F,$3D,$D9,$FC,$B3,$E4,$78,$F2,$7F,$0A,$6D,$7F,$0A,$F9
	.dc.b	$1B,$8F,$F8,$13,$A3,$FD,$7F,$3F,$D0,$27,$38,$73,$8C,$1F,$D1,$2F
	.dc.b	$35,$FE,$BC,$8F,$E8,$FF,$1F,$8F,$90,$CB,$63,$FA,$5C,$70,$F9,$0C
	.dc.b	$65,$FB,$E9,$FA,$AB,$EC,$FF,$A2,$47,$F7,$B6,$C7,$F8,$21,$AF,$F9
	.dc.b	$51,$AF,$27,$7A,$7F,$A1,$1B,$3D,$BE,$99,$7F,$0A,$FC,$1E,$35,$BF
	.dc.b	$F3,$C5,$FF,$60,$F9,$DF,$75,$79,$CF,$C5,$CE,$9C,$74,$A7,$79,$78
	.dc.b	$7C,$6F,$F4,$AD,$79,$5A,$F2,$6B,$F0,$5B,$7D,$67,$96,$1F,$E2,$BC
	.dc.b	$DC,$FE,$7B,$FB,$79,$38,$BF,$C8,$77,$FD,$E1,$B9,$7A,$4F,$9C,$F0
	.dc.b	$33,$8F,$C7,$7C,$24,$9F,$A4,$F9,$BE,$99,$EC,$7C,$78,$F9,$51,$89
	.dc.b	$FF,$20,$69,$E3,$7F,$C1,$7F,$29,$F8,$83,$C1,$EB,$58,$DF,$1F,$C0
	.dc.b	$23,$F8,$17,$C1,$EA,$76,$3F,$23,$BD,$9F,$F0,$BF,$4F,$E5,$0F,$CA
	.dc.b	$FC,$A5,$8F,$F2,$97,$9E,$F8,$6F,$9E,$D8,$FF,$24,$37,$87,$FC,$AA
	.dc.b	$D7,$FC,$8F,$F5,$FF,$EF,$B9,$FF,$2F,$36,$3F,$09,$D9,$FE,$1C,$7F
	.dc.b	$53,$19,$E3,$FD,$DC,$A1,$F2,$7A,$37,$F2,$97,$0F,$3F,$C5,$79,$88
	.dc.b	$F9,$9F,$C4,$DF,$3F,$94,$EE,$78,$3E,$6A,$D7,$9B,$E4,$BC,$7F,$27
	.dc.b	$E4,$6C,$7F,$9D,$7E,$77,$F5,$3B,$7F,$26,$6A,$69,$AC,$7D,$5F,$21
	.dc.b	$EE,$3D,$FF,$2B,$6F,$E2,$FC,$AF,$C2,$F9,$5F,$86,$D8,$FF,$42,$7E
	.dc.b	$0F,$FE,$F6,$7F,$CF,$0D,$CB,$FD,$0A,$D7,$FD,$A2,$D7,$FD,$43,$C5
	.dc.b	$F5,$F8,$FE,$6B,$6B,$3E,$7F,$D2,$E6,$FD,$F7,$47,$F3,$2B,$FE,$4B
	.dc.b	$C4,$E8,$CF,$99,$FC,$92,$53,$9C,$7F,$D3,$57,$FC,$EB,$3F,$F7,$1B
	.dc.b	$5F,$E3,$2D,$7F,$A7,$B5,$FE,$B9,$7D,$A7,$FD,$69,$7D,$AB,$15,$F8
	.dc.b	$AC,$DF,$27,$F8,$A3,$F1,$7C,$F7,$0F,$C8,$C5,$AE,$13,$70,$F8,$7F
	.dc.b	$29,$F9,$23,$F2,$7E,$ED,$C3,$F3,$31,$6F,$C1,$36,$BE,$5F,$C5,$7E
	.dc.b	$68,$FC,$DF,$A7,$70,$FD,$0C,$5B,$F4,$4D,$AE,$35,$A3,$F4,$F1,$97
	.dc.b	$FD,$B3,$87,$FB,$F1,$FD,$5F,$CA,$7F,$50,$8C,$5F,$C3,$8C,$7C,$9C
	.dc.b	$AF,$D8,$FA,$F6,$BF,$63,$7F,$E3,$A9,$FF,$69,$36,$33,$F2,$9D,$7D
	.dc.b	$99,$5A,$7C,$11,$89,$FE,$4D,$F3,$BF,$E8,$DA,$FA,$FF,$07,$FE,$A1
	.dc.b	$7F,$DB,$91,$99,$FE,$89,$AA,$B5,$1C,$B7,$FE,$7B,$5B,$8F,$EB,$E1
	.dc.b	$C3,$FD,$89,$AF,$F8,$61,$ED,$5F,$E0,$48,$E1,$3F,$97,$AF,$D0,$46
	.dc.b	$E3,$E5,$7C,$8C,$C3,$E1,$FC,$AD,$AF,$8F,$37,$0F,$E2,$BA,$2B,$17
	.dc.b	$FD,$D3,$E4,$DE,$37,$D5,$7F,$B4,$18,$8F,$9E,$E2,$24,$7D,$2F,$46
	.dc.b	$E5,$FD,$47,$DB,$7C,$87,$DF,$F9,$3F,$EE,$51,$8B,$FB,$31,$89,$FE
	.dc.b	$CB,$39,$AF,$92,$7F,$9F,$A3,$24,$AF,$F1,$FE,$9F,$1E,$B8,$3E,$E9
	.dc.b	$E9,$3F,$86,$F9,$1F,$C8,$7A,$3F,$91,$6C,$64,$98,$7E,$3B,$D0,$D4
	.dc.b	$7E,$3B,$E3,$FF,$31,$F2,$3F,$99,$6C,$64,$95,$F9,$6F,$90,$D4,$7E
	.dc.b	$5B,$E1,$7F,$9D,$F1,$9A,$48,$FF,$6A,$C6,$5E,$07,$E7,$BE,$43,$F5
	.dc.b	$13,$FA,$9F,$90,$FD,$57,$C3,$5A,$FD,$1B,$71,$FD,$3C,$FE,$BF,$D6
	.dc.b	$C7,$0F,$F5,$6F,$B0,$7C,$8F,$F4,$E6,$E3,$FE,$E6,$6E,$1F,$EF,$B6
	.dc.b	$BF,$E3,$AF,$4D,$1A,$BE,$97,$FB,$DB,$FF,$26,$CE,$5F,$B7,$CB,$FE
	.dc.b	$76,$F0,$FF,$C6,$B8,$FD,$AF,$1F,$FC,$C7,$18,$8F,$F5,$B0,$71,$8F
	.dc.b	$F2,$8B,$E5,$F0,$9D,$29,$F3,$9A,$C3,$FC,$DA,$F3,$8F,$F6,$59,$A9
	.dc.b	$FF,$29,$DF,$38,$C1,$FB,$8C,$7F,$D1,$75,$CF,$0B,$7D,$E3,$15,$C0
	.dc.b	$F3,$B8,$3C,$0F,$FB,$58,$D5,$F7,$8C,$DF,$AA,$7E,$15,$D7,$E0,$E3
	.dc.b	$D8,$DB,$CB,$1F,$6E,$3E,$33,$D6,$E2,$3D,$46,$EB,$D2,$E3,$0F,$35
	.dc.b	$7C,$9B,$EC,$D3,$9D,$78,$EB,$E4,$D7,$07,$94,$E9,$F8,$46,$9C,$47
	.dc.b	$A0,$E1,$1E,$C8,$E1,$5E,$C3,$C3,$1D,$0F,$86,$74,$3E,$54,$F2,$7B
	.dc.b	$1A,$7B,$4F,$80,$F8,$91,$96,$AC,$7E,$94,$7B,$E3,$BD,$78,$E3,$CA
	.dc.b	$70,$9D,$CF,$94,$E5,$87,$CF,$7D,$87,$29,$E8,$EB,$0F,$94,$FE,$0E
	.dc.b	$FD,$EF,$03,$47,$E7,$E7,$F3,$9E,$2F,$E8,$63,$ED,$B5,$FD,$2B,$5F
	.dc.b	$CE,$B9,$F8,$AF,$F3,$16,$3F,$EE,$48,$C7,$A1,$F0,$69,$18,$9D,$61
	.dc.b	$FE,$B6,$BE,$35,$A8,$C5,$F1,$7F,$6D,$87,$F8,$B6,$7F,$30,$D7,$D4
	.dc.b	$7F,$BD,$2B,$9C,$FE,$F5,$B3,$FE,$C4,$38,$47,$4F,$43,$FD,$8E,$EE
	.dc.b	$FC,$13,$E2,$C3,$F3,$78,$7F,$AF,$AF,$FD,$9B,$A7,$18,$CE,$3E,$A8
	.dc.b	$E5,$5E,$6A,$EB,$7E,$16,$DF,$9F,$0E,$ED,$AE,$F5,$ED,$BE,$D3,$FC
	.dc.b	$07,$0E,$D3,$E5,$8F,$E0,$8E,$CE,$CF,$92,$FE,$95,$79,$23,$78,$7A
	.dc.b	$AB,$71,$D6,$B7,$1D,$71,$F0,$91,$C1,$E1,$BF,$29,$1C,$12,$87,$C9
	.dc.b	$3D,$B7,$E1,$7E,$01,$E1,$C7,$E4,$30,$FB,$A3,$1F,$B6,$F0,$BF,$03
	.dc.b	$DA,$FE,$13,$97,$C4,$F8,$4F,$82,$7C,$5A,$8F,$68,$FC,$29,$F1,$30
	.dc.b	$F9,$07,$C9,$AE,$F3,$CD,$C4,$F6,$3C,$31,$97,$C3,$7F,$62,$3C,$D8
	.dc.b	$7C,$17,$47,$9C,$F8,$77,$F3,$DD,$9F,$CD,$E7,$E0,$F0,$7F,$64,$F8
	.dc.b	$6F,$E9,$BA,$FC,$E8,$C5,$7F,$86,$23,$1C,$0A,$DC,$AB,$FB,$6F,$C9
	.dc.b	$61,$E3,$1F,$DC,$E7,$FC,$4A,$EE,$B9,$3F,$DE,$63,$33,$FD,$DB,$E4
	.dc.b	$B9,$51,$A6,$3E,$1A,$DF,$94,$E9,$E4,$F0,$32,$FF,$B8,$A9,$F2,$1D
	.dc.b	$09,$1C,$70,$FF,$02,$FC,$96,$23,$FC,$15,$3E,$4A,$E3,$1F,$E0,$F8
	.dc.b	$C7,$AB,$E6,$24,$73,$97,$C8,$E2,$3C,$5F,$23,$EC,$31,$CB,$FE,$1D
	.dc.b	$F8,$9F,$79,$27,$97,$C8,$D7,$D5,$F2,$1F,$10,$C5,$7F,$9C,$7E,$9B
	.dc.b	$FC,$FD,$3F,$0C,$38,$7F,$A5,$72,$7E,$A9,$CF,$07,$CE,$F8,$9F,$B7
	.dc.b	$18,$3E,$BD,$71,$7F,$D0,$AD,$79,$35,$FD,$AB,$DE,$FF,$C7,$1D,$3F
	.dc.b	$68,$EB,$19,$C3,$C5,$F6,$0D,$3C,$2D,$C3,$D4,$D7,$ED,$7C,$8E,$A7
	.dc.b	$F0,$BF,$19,$27,$71,$F8,$8E,$C9,$1F,$4D,$FC,$5F,$C6,$62,$3F,$09
	.dc.b	$17,$4F,$F2,$74,$F9,$89,$3F,$2F,$34,$7C,$6F,$9A,$F2,$57,$D1,$8F
	.dc.b	$CE,$D7,$D0,$6D,$70,$8F,$97,$18,$39,$70,$48,$ED,$1F,$A0,$61,$F8
	.dc.b	$B3,$F4,$7F,$03,$87,$F3,$85,$BF,$F3,$34,$7E,$C7,$EC,$1F,$1D,$49
	.dc.b	$5F,$9F,$87,$CF,$62,$3F,$0A,$7E,$EE,$3F,$D5,$AE,$FD,$8D,$60,$E2
	.dc.b	$FE,$0D,$4F,$8F,$F0,$12,$FB,$7C,$35,$7A,$F7,$AB,$84,$78,$EB,$D1
	.dc.b	$5B,$C3,$B4,$F8,$E7,$18,$78,$2B,$DD,$F1,$D2,$BD,$8F,$C1,$F8,$FC
	.dc.b	$47,$83,$14,$E3,$1E,$B9,$78,$EF,$E4,$7C,$74,$C3,$E3,$3F,$2F,$E3
	.dc.b	$B1,$1C,$E2,$DF,$2C,$55,$FA,$AF,$9C,$7C,$35,$26,$1F,$B5,$F9,$C7
	.dc.b	$B5,$88,$FF,$64,$D7,$07,$F9,$3B,$5C,$D7,$E3,$A3,$18,$FF,$BC,$6D
	.dc.b	$ED,$7E,$E9,$B7,$FC,$0E,$B7,$C1,$F6,$3E,$3F,$F0,$C4,$9F,$C2,$7C
	.dc.b	$87,$F0,$2F,$65,$C7,$F9,$C6,$CF,$F3,$A6,$BF,$EE,$06,$BF,$EA,$FD
	.dc.b	$DE,$9F,$E4,$FE,$3E,$4F,$E4,$63,$F2,$DE,$B4,$8F,$CA,$3F,$99,$F8
	.dc.b	$EF,$E2,$7F,$38,$F8,$1F,$D6,$B6,$BF,$A5,$B5,$FF,$78,$B5,$F1,$7C
	.dc.b	$07,$E8,$FE,$2A,$47,$E8,$63,$F4,$DF,$1F,$23,$F4,$8F,$EA,$7E,$3F
	.dc.b	$F9,$1F,$D7,$BC,$3F,$DE,$B6,$BF,$AD,$B5,$FF,$80,$2D,$7E,$17,$BD
	.dc.b	$FB,$32,$3F,$B1,$F7,$7F,$B3,$7E,$26,$F9,$0A,$49,$FE,$A9,$0F,$91
	.dc.b	$F1,$C7,$88,$87,$88,$FE,$D3,$0F,$8F,$EA,$48,$E7,$DF,$FD,$E3,$E5
	.dc.b	$DF,$53,$B9,$FE,$D5,$18,$AD,$4E,$67,$F7,$36,$F9,$2F,$E8,$FF,$5F
	.dc.b	$FE,$7B,$9F,$E9,$8D,$7F,$C6,$ED,$4E,$7C,$85,$3E,$4F,$E1,$92,$BC
	.dc.b	$7D,$6F,$03,$C7,$F2,$5E,$0E,$29,$5E,$8D,$13,$88,$FE,$BC,$78,$A7
	.dc.b	$F9,$AE,$7F,$E8,$A8,$FF,$09,$B5,$DE,$CC,$5D,$9F,$1D,$B3,$E4,$8E
	.dc.b	$DE,$37,$F1,$68,$73,$BE,$47,$FB,$5B,$63,$FB,$A3,$5F,$E5,$F8,$3E
	.dc.b	$A9,$ED,$F0,$BF,$E9,$77,$BB,$F5,$B5,$FE,$F4,$D7,$FC,$CA,$F5,$C3
	.dc.b	$F9,$4E,$0F,$9F,$FF,$15,$57,$DE,$6B,$FD,$00,$F4,$DF,$A3,$C0,$91
	.dc.b	$E8,$7F,$11,$97,$F5,$B8,$E1,$8F,$EB,$DF,$1C,$FA,$08,$FD,$23,$FC
	.dc.b	$8D,$C6,$F9,$4F,$D0,$92,$3F,$3F,$F2,$B5,$FC,$7F,$17,$E9,$8C,$6B
	.dc.b	$FC,$E2,$FD,$DE,$4F,$F9,$5D,$FE,$05,$19,$C7,$F5,$BA,$7E,$B9,$98
	.dc.b	$7E,$B1,$CB,$FE,$27,$F9,$F7,$C9,$FE,$17,$E4,$3E,$DC,$FF,$A4,$27
	.dc.b	$3C,$DF,$E9,$29,$CC,$A7,$2A,$33,$1F,$CE,$5B,$5D,$77,$FF,$19,$4F
	.dc.b	$09,$CD,$F5,$9C,$F6,$FF,$8D,$E7,$38,$75,$9E,$38,$73,$8F,$33,$5F
	.dc.b	$25,$71,$9D,$DF,$59,$DB,$5F,$55,$71,$C3,$D1,$5C,$27,$C0,$70,$29
	.dc.b	$C2,$B9,$4F,$08,$D1,$FB,$27,$E2,$B6,$BF,$60,$F5,$3F,$6F,$F1,$FF
	.dc.b	$B1,$E4,$FF,$43,$B6,$3F,$14,$D7,$7C,$E9,$FF,$46,$DB,$CB,$FE,$65
	.dc.b	$5F,$65,$AF,$D7,$6C,$6F,$C4,$E3,$1F,$CF,$75,$FB,$2D,$D3,$FC,$D1
	.dc.b	$ED,$3E,$27,$FC,$2B,$5A,$BF,$83,$7B,$C7,$C1,$1F,$36,$F8,$72,$6A
	.dc.b	$24,$FC,$AC,$BC,$4F,$CC,$AD,$B9,$BE,$8F,$A6,$B8,$E1,$F9,$2A,$DB
	.dc.b	$E3,$9E,$13,$AC,$BD,$39,$7A,$1F,$C8,$E1,$E7,$F0,$3F,$99,$FC,$B7
	.dc.b	$F8,$C3,$97,$F5,$CD,$D7,$F5,$DF,$2D,$73,$FD,$DC,$DB,$FE,$77,$37
	.dc.b	$9D,$36,$30,$E7,$E5,$B5,$EA,$DC,$7F,$AF,$B7,$1F,$E9,$F3,$CF,$D6
	.dc.b	$C6,$7F,$C0,$74,$FE,$61,$5F,$C0,$5A,$FF,$0E,$6E,$9F,$D3,$9B,$A7
	.dc.b	$E1,$72,$FE,$DB,$1E,$7E,$6F,$43,$E9,$8F,$57,$CA,$D7,$FB,$9B,$70
	.dc.b	$FE,$9B,$87,$F5,$A9,$EA,$43,$F9,$03,$97,$FB,$D9,$F1,$7E,$5B,$96
	.dc.b	$5F,$E5,$2F,$86,$7C,$B7,$CB,$6D,$7F,$B5,$57,$CA,$E6,$74,$FB,$A3
	.dc.b	$33,$E0,$6E,$3E,$66,$BF,$EC,$1C,$51,$FE,$FB,$9E,$D5,$E1,$F9,$5E
	.dc.b	$B1,$E1,$F8,$DF,$27,$CA,$79,$5B,$1E,$5F,$95,$F1,$11,$F1,$3E,$5F
	.dc.b	$94,$F6,$1E,$DF,$95,$B3,$ED,$F9,$3F,$81,$F2,$9F,$05,$B1,$F0,$7B
	.dc.b	$3D,$C4,$7D,$CF,$C1,$F8,$CF,$90,$7C,$9F,$94,$B3,$F2,$7E,$53,$E6
	.dc.b	$7C,$97,$CD,$6C,$7C,$DF,$47,$E5,$11,$F9,$4F,$CD,$E2,$FA,$87,$D5
	.dc.b	$F9,$3B,$3F,$EE,$FF,$9A,$F9,$5C,$DF,$D7,$DA,$FF,$84,$9A,$FF,$BC
	.dc.b	$5A,$B9,$BE,$9F,$8D,$F6,$C9,$5F,$6D,$FB,$DE,$2F,$E1,$4F,$C3,$7C
	.dc.b	$A5,$95,$FF,$AB,$BF,$05,$FF,$5E,$C7,$E2,$63,$F6,$D1,$87,$FD,$67
	.dc.b	$9F,$09,$E3,$D9,$FE,$A4,$A7,$C8,$FE,$A9,$B3,$F9,$C6,$DF,$E8,$9A
	.dc.b	$FE,$35,$FF,$44,$C7,$43,$CB,$9A,$33,$3E,$7F,$93,$F2,$FC,$57,$F0
	.dc.b	$D8,$FF,$38,$6F,$FC,$3F,$27,$D7,$70,$7F,$C6,$1B,$27,$53,$9C,$FE
	.dc.b	$1D,$F2,$7F,$C2,$4D,$3F,$C2,$95,$C6,$32,$D8,$CD,$78,$E3,$17,$FC
	.dc.b	$96,$FF,$DC,$17,$ED,$BF,$A3,$5A,$8F,$7E,$4F,$91,$FA,$3F,$27,$F4
	.dc.b	$9B,$3C,$75,$F8,$91,$8C,$3E,$D3,$72,$FA,$9F,$23,$F6,$8F,$B7,$F2
	.dc.b	$77,$0E,$59,$7F,$B4,$9B,$3A,$69,$AF,$90,$FC,$1F,$C9,$FE,$11,$B1
	.dc.b	$93,$9C,$65,$A6,$72,$FC,$74,$65,$B1,$E2,$F9,$0F,$C7,$FC,$27,$E4
	.dc.b	$1B,$19,$24,$FE,$5F,$D7,$D4,$7A,$BE,$0B,$F2,$FF,$21,$F9,$86,$C6
	.dc.b	$4F,$7C,$65,$A7,$D3,$AF,$F1,$97,$C0,$7E,$27,$EB,$BF,$DE,$73,$9F
	.dc.b	$65,$39,$F7,$BF,$21,$18,$9F,$C3,$B6,$FF,$CB,$4D,$7F,$D1,$AD,$7F
	.dc.b	$C6,$D5,$C8,$F0,$7C,$4D,$AF,$AB,$F1,$FF,$B2,$3F,$67,$F2,$16,$32
	.dc.b	$4B,$FD,$D4,$65,$FD,$E7,$C7,$7E,$EB,$DE,$FD,$DB,$6B,$F0,$2E,$1F
	.dc.b	$C1,$FC,$85,$AF,$F3,$F7,$D4,$7C,$27,$F8,$37,$C2,$7E,$BE,$31,$D5
	.dc.b	$FC,$06,$31,$BF,$F6,$CA,$57,$17,$FD,$5A,$D7,$F5,$ED,$7F,$89,$BD
	.dc.b	$30,$CC,$FF,$81,$FF,$05,$FF,$6E,$47,$D7,$9D,$57,$F2,$39,$FF,$6D
	.dc.b	$4F,$E9,$5F,$F7,$33,$76,$E5,$F2,$3C,$D9,$1B,$C2,$DB,$24,$FF,$12
	.dc.b	$C3,$EE,$46,$27,$C1,$F2,$1E,$5E,$1F,$33,$6F,$C5,$1E,$9F,$90,$C5
	.dc.b	$75,$9E,$9F,$19,$FE,$10,$C7,$FD,$3D,$1F,$E0,$26,$CF,$F8,$21,$AF
	.dc.b	$F7,$37,$38,$F8,$FC,$14,$62,$BD,$1F,$19,$F2,$70,$F9,$5F,$0D,$F2
	.dc.b	$FE,$32,$D7,$C7,$6E,$5E,$53,$E6,$4F,$D1,$F8,$CC,$E7,$DF,$27,$C4
	.dc.b	$7A,$49,$8F,$A5,$B8,$73,$D3,$CB,$1F,$6F,$E2,$7E,$E3,$FC,$E6,$7F
	.dc.b	$D4,$C3,$9F,$FD,$BA,$FF,$4E,$EE,$FA,$11,$88,$FF,$4F,$1B,$F4,$FF
	.dc.b	$9C,$E7,$FE,$F2,$CB,$FA,$ED,$6E,$BA,$FC,$E7,$1A,$D4,$23,$4F,$FA
	.dc.b	$CC,$C5,$F4,$C3,$F0,$13,$AE,$EF,$ED,$2F,$38,$CC,$FF,$00,$AE,$27
	.dc.b	$A7,$1D,$C6,$6B,$84,$62,$7F,$82,$9C,$0F,$C9,$57,$47,$38,$F0,$7E
	.dc.b	$01,$8B,$F3,$57,$C0,$7B,$18,$8E,$87,$C2,$71,$7E,$B8,$F0,$63,$F1
	.dc.b	$23,$07,$73,$57,$E1,$2D,$75,$76,$F4,$3C,$35,$F0,$4F,$09,$F3,$67
	.dc.b	$A4,$6A,$3D,$F7,$F0,$DF,$BB,$5D,$E7,$E6,$36,$3E,$3D,$72,$6C,$F8
	.dc.b	$5A,$78,$63,$CA,$D8,$F0,$5F,$BD,$B8,$7C,$F9,$F0,$C7,$B4,$F0,$95
	.dc.b	$EA,$7A,$0F,$6E,$3E,$9A,$E9,$1F,$54,$F8,$58,$7C,$86,$D7,$46,$9B
	.dc.b	$3A,$47,$D5,$EA,$FD,$24,$7E,$19,$AF,$E8,$9A,$FE,$A9,$CF,$C6,$BC
	.dc.b	$3F,$E3,$92,$9C,$C6,$1B,$87,$F9,$11,$B3,$BB,$FF,$12,$DF,$3A,$FC
	.dc.b	$74,$FF,$92,$B0,$7A,$7F,$E2,$AB,$FD,$B4,$62,$76,$2F,$93,$F9,$A3
	.dc.b	$18,$F9,$AB,$8D,$F3,$BF,$F1,$A9,$CE,$B8,$5F,$2A,$FE,$08,$F4,$D3
	.dc.b	$8C,$FE,$2E,$B8,$46,$B0,$F8,$6D,$C7,$CB,$3D,$70,$E3,$7D,$AF,$C9
	.dc.b	$3F,$09,$E2,$F9,$31,$FF,$24,$E1,$C0,$E9,$5F,$BB,$C7,$C9,$5B,$8F
	.dc.b	$25,$70,$8E,$98,$FE,$86,$B7,$5F,$36,$B4,$D3,$58,$78,$9F,$F3,$DC
	.dc.b	$FC,$3B,$ED,$1F,$35,$F6,$1D,$9F,$65,$F8,$98,$79,$EB,$A3,$F5,$23
	.dc.b	$38,$7B,$6B,$93,$EE,$C7,$E3,$92,$3B,$3F,$0A,$38,$57,$A4,$ED,$5E
	.dc.b	$A3,$AD,$7A,$27,$83,$C8,$F5,$D7,$5A,$F7,$DF,$D4,$6D,$F2,$7B,$5F
	.dc.b	$27,$27,$F2,$DB,$FC,$19,$DC,$F5,$91,$E2,$E5,$E3,$1E,$F8,$F1,$92
	.dc.b	$3A,$DF,$F5,$8E,$0F,$EE,$7F,$17,$FD,$53,$7F,$F0,$53,$6B,$FC,$30
	.dc.b	$D9,$FE,$F3,$F1,$77,$8D,$3F,$E9,$29,$E7,$5F,$E0,$58,$C4,$EA,$3F
	.dc.b	$C1,$1F,$11,$C4,$8F,$F8,$8F,$E2,$AB,$FD,$CE,$5F,$15,$DC,$CE,$DF
	.dc.b	$DF,$73,$F1,$12,$BF,$C5,$9F,$15,$5F,$F0,$A4,$78,$88,$FF,$AA,$8F
	.dc.b	$F5,$CF,$E2,$6F,$89,$F7,$36,$3F,$D3,$4D,$7F,$C0,$CB,$E2,$F7,$3C
	.dc.b	$3E,$2E,$57,$08,$E7,$F1,$32,$39,$BD,$BE,$2F,$11,$D3,$E2,$FC,$26
	.dc.b	$34,$FF,$1D,$43,$EA,$5F,$11,$F5,$63,$0D,$8C,$C7,$A3,$E2,$F1,$3F
	.dc.b	$EC,$68,$D9,$FE,$C8,$9D,$BF,$0C,$C9,$5C,$D7,$F8,$E7,$57,$C6,$7C
	.dc.b	$63,$AD,$7C,$6B,$F9,$24,$7E,$4F,$C6,$57,$E4,$FC,$67,$CE,$31,$87
	.dc.b	$5F,$8C,$FA,$24,$9F,$A3,$F1,$95,$FA,$31,$67,$C9,$3F,$0F,$E6,$DA
	.dc.b	$D2,$5F,$FA,$A7,$E7,$3F,$01,$7F,$32,$B7,$3F,$EE,$A6,$BF,$0D,$AF
	.dc.b	$E1,$DA,$FF,$BC,$9D,$C7,$D6,$97,$C2,$7D,$56,$D7,$FA,$E6,$B7,$CD
	.dc.b	$F9,$0F,$8D,$FC,$7F,$C7,$4A,$FC,$7C,$7E,$48,$C3,$F9,$08,$C3,$F9
	.dc.b	$4F,$8C,$FC,$B9,$1F,$CB,$F8,$75,$FC,$BF,$C4,$FE,$70,$C6,$9F,$6F
	.dc.b	$C8,$FC,$FF,$C6,$CA,$FC,$FC,$7E,$88,$C3,$FA,$08,$C3,$A3,$F4,$AE
	.dc.b	$A7,$A4,$7E,$A0,$FD,$59,$0F,$D5,$ED,$5F,$D5,$96,$3F,$1D,$1F,$C3
	.dc.b	$FB,$DF,$1B,$7E,$72,$58,$FF,$3C,$4E,$9C,$5F,$FA,$3E,$31,$1A,$9E
	.dc.b	$2F,$0A,$FD,$57,$D4,$BD,$FA,$49,$FF,$78,$7C,$EB,$C7,$C4,$7E,$9E
	.dc.b	$B8,$3F,$A9,$0E,$1D,$57,$EC,$63,$19,$7E,$F2,$FF,$8E,$D7,$6C,$BA
	.dc.b	$CF,$A7,$E3,$FE,$11,$27,$DB,$C9,$F0,$7C,$AF,$E4,$BF,$21,$F1,$FE
	.dc.b	$3F,$F9,$96,$26,$92,$FF,$7D,$5F,$CB,$3E,$3B,$E7,$7C,$7C,$9E,$A2
	.dc.b	$74,$E1,$FE,$67,$F1,$9F,$54,$91,$9C,$1F,$1D,$F6,$7E,$3B,$C6,$49
	.dc.b	$F0,$CA,$3F,$8A,$7C,$DB,$DD,$A4,$AF,$99,$19,$78,$FC,$77,$E1,$48
	.dc.b	$F8,$7E,$3F,$F1,$1E,$CF,$C2,$24,$78,$1F,$8B,$EE,$FE,$37,$E3,$E4
	.dc.b	$FA,$63,$F2,$1F,$1D,$23,$D8,$FE,$4F,$E3,$BF,$29,$D5,$F3,$89,$3E
	.dc.b	$C7,$E8,$FB,$9F,$99,$F8,$E9,$3F,$16,$3F,$B5,$6A,$F9,$AA,$49,$F9
	.dc.b	$5F,$03,$AF,$8E,$FE,$D9,$46,$92,$7E,$DC,$3E,$3B,$AF,$C7,$49,$FB
	.dc.b	$13,$96,$FC,$77,$8B,$E3,$64,$60,$DB,$E0,$23,$D9,$F1,$FC,$77,$A8
	.dc.b	$91,$C0,$EF,$F1,$9E,$E2,$3E,$2F,$8E,$AF,$F7,$1E,$D6,$2A,$F7,$92
	.dc.b	$7D,$AF,$C8,$F8,$E9,$1D,$8F,$6F,$C7,$7C,$C2,$3F,$03,$AE,$BD,$ED
	.dc.b	$3F,$30,$91,$F2,$BE,$27,$EA,$91,$F9,$3F,$1D,$57,$90,$F1,$FF,$65
	.dc.b	$83,$E3,$BE,$E9,$B9,$D5,$7E,$AE,$7F,$C4,$4F,$2A,$E3,$F1,$5F,$84
	.dc.b	$3F,$0B,$F1,$76,$7E,$CC,$57,$81,$CE,$3F,$C0,$5F,$2A,$9F,$F5,$44
	.dc.b	$BE,$3F,$F8,$94,$E7,$A7,$F9,$44,$62,$B6,$EB,$9F,$F9,$65,$F1,$ED
	.dc.b	$B8,$71,$8C,$1F,$CC,$5C,$57,$3C,$3A,$D7,$0B,$E7,$39,$BE,$22,$F8
	.dc.b	$D7,$6C,$3C,$6F,$19,$E6,$77,$8C,$B7,$0C,$BC,$E3,$2F,$5C,$3C,$35
	.dc.b	$D6,$B7,$7E,$0A,$E4,$79,$15,$70,$8C,$9F,$E4,$78,$F8,$4D,$BF,$9E
	.dc.b	$DA,$FE,$85,$F1,$EF,$15,$F2,$D8,$6D,$7F,$8B,$E7,$53,$93,$FA,$56
	.dc.b	$1F,$E6,$0F,$56,$F9,$3F,$E0,$AB,$F1,$35,$FA,$AD,$7E,$A3,$FE,$10
	.dc.b	$D7,$F0,$2D,$D7,$EE,$FA,$DF,$E4,$AE,$0F,$87,$39,$BE,$38,$60,$F9
	.dc.b	$B5,$B8,$F8,$D3,$B7,$ED,$37,$1F,$44,$7D,$E9,$E1,$8F,$8A,$FA,$4F
	.dc.b	$27,$AF,$C2,$7E,$9A,$75,$39,$23,$FD,$AE,$B9,$E5,$F0,$2F,$F0,$BD
	.dc.b	$7F,$C2,$BD,$5F,$D7,$F2,$FE,$C2,$3F,$86,$E9,$8D,$7F,$6F,$57,$3F
	.dc.b	$DA,$4D,$AE,$55,$FC,$A2,$31,$CB,$FE,$01,$8C,$1F,$CB,$0C,$FC,$76
	.dc.b	$9A,$64,$FE,$62,$E5,$A7,$F6,$BF,$9B,$7C,$77,$F9,$FB,$1F,$EF,$91
	.dc.b	$89,$FE,$EE,$D8,$FE,$50,$D7,$F6,$F8,$A3,$59,$FC,$16,$D7,$4F,$8F
	.dc.b	$F7,$46,$2F,$FB,$F2,$F8,$CC,$FC,$7E,$9B,$7E,$77,$CD,$5C,$1B,$8F
	.dc.b	$85,$B7,$CE,$BD,$31,$CD,$F8,$1F,$1D,$6F,$F8,$DD,$FA,$7E,$3F,$CA
	.dc.b	$79,$BE,$22,$DF,$C7,$7A,$91,$F8,$AF,$83,$E3,$78,$47,$CD,$6D,$F6
	.dc.b	$C9,$F1,$9F,$08,$F8,$7F,$1B,$6F,$EA,$74,$FA,$88,$FD,$27,$DB,$F1
	.dc.b	$BF,$2B,$E3,$7E,$5B,$67,$EE,$7C,$5F,$CE,$3E,$7F,$C5,$D8,$FC,$07
	.dc.b	$C5,$7C,$62,$3F,$79,$F9,$7D,$BF,$E9,$A8,$FC,$37,$C5,$7E,$56,$31
	.dc.b	$B7,$AE,$3E,$DC,$49,$FC,$67,$C5,$DA,$F9,$BF,$84,$3E,$0B,$EB,$B6
	.dc.b	$BF,$C8,$6D,$9F,$EE,$8D,$7F,$DC,$DF,$2F,$FE,$E5,$9F,$F2,$F3,$63
	.dc.b	$FD,$90,$D7,$7D,$3F,$82,$F0,$7F,$1E,$7E,$43,$E3,$6D,$FE,$86,$BF
	.dc.b	$0E,$47,$F3,$EF,$E3,$3E,$6F,$FD,$03,$1F,$E5,$15,$F1,$DF,$B7,$73
	.dc.b	$3F,$8A,$73,$3B,$F8,$BD,$1B,$F8,$BB,$1C,$FE,$2F,$B1,$1F,$F3,$2D
	.dc.b	$7F,$9A,$5B,$F1,$7D,$FE,$2F,$C0,$D8,$F2,$7C,$57,$8F,$E2,$FC,$8D
	.dc.b	$8C,$9D,$1F,$61,$1C,$1A,$CF,$D5,$C5,$CA,$3E,$07,$A3,$EF,$3E,$07
	.dc.b	$C5,$D8,$C9,$2B,$E1,$46,$5F,$89,$F1,$5F,$1F,$E2,$FE,$43,$6B,$E6
	.dc.b	$36,$77,$1F,$3E,$7E,$77,$C4,$FC,$FF,$8B,$B3,$F4,$DB,$5F,$45,$BE
	.dc.b	$67,$F3,$1E,$04,$62,$BF,$D3,$11,$87,$EE,$5F,$57,$EF,$7C,$57,$DD
	.dc.b	$F8,$BF,$BC,$D8,$C9,$31,$FC,$07,$A9,$A8,$FC,$37,$B3,$F8,$5F,$8B
	.dc.b	$FC,$33,$63,$27,$E2,$E3,$2D,$3F,$D4,$35,$FE,$29,$F7,$7F,$A3,$6A
	.dc.b	$FD,$51,$C1,$FD,$26,$B5,$7F,$EA,$8A,$D4,$E7,$D9,$61,$FC,$A2,$F6
	.dc.b	$7F,$8B,$8F,$F1,$9B,$98,$DD,$FF,$B2,$DB,$3F,$E4,$03,$8D,$FF,$5E
	.dc.b	$C3,$E5,$46,$67,$73,$FD,$87,$3F,$F4,$45,$FF,$AE,$5F,$F4,$83,$7E
	.dc.b	$2F,$FB,$04,$62,$FF,$99,$BC,$1B,$F1,$9A,$9D,$BC,$3E,$2F,$8F,$C0
	.dc.b	$DC,$3B,$36,$F9,$CF,$6F,$8C,$C6,$5F,$E6,$49,$7C,$B7,$FA,$67,$0F
	.dc.b	$95,$FA,$AF,$CE,$7E,$9C,$66,$7F,$CC,$34,$AD,$47,$C5,$9D,$3F,$35
	.dc.b	$AF,$F9,$71,$A8,$F0,$DF,$D5,$8E,$D5,$E5,$6D,$EE,$3C,$1C,$FF,$81
	.dc.b	$9C,$DF,$FB,$9F,$57,$C6,$7C,$CF,$8E,$F9,$AD,$C7,$E9,$38,$7E,$A7
	.dc.b	$C7,$5B,$FA,$3F,$19,$F5,$3E,$37,$EA,$B6,$BE,$D3,$87,$EE,$7C,$75
	.dc.b	$AF,$B3,$F1,$9F,$DD,$33,$34,$95,$F9,$19,$F4,$47,$E0,$8F,$C1,$FC
	.dc.b	$75,$9F,$C4,$7C,$77,$E1,$BE,$3B,$F0,$ED,$8C,$9F,$8D,$8C,$B4,$FC
	.dc.b	$87,$C7,$78,$C8,$F8,$7E,$0D,$97,$FA,$66,$30,$DC,$BF,$27,$5F,$96
	.dc.b	$EC,$FC,$BB,$FB,$18,$FF,$4D,$8E,$E7,$B1,$FE,$9A,$E6,$FD,$2C,$62
	.dc.b	$7F,$61,$DB,$F8,$18,$C4,$7F,$AF,$39,$FF,$D6,$F1,$88,$EB,$F2,$FA
	.dc.b	$BF,$D7,$B9,$8F,$F5,$70,$75,$5F,$EC,$07,$57,$F8,$F7,$3A,$5C,$7E
	.dc.b	$94,$DB,$FC,$75,$F1,$9C,$C7,$57,$D9,$19,$8F,$59,$93,$8D,$73,$7F
	.dc.b	$D6,$27,$C5,$9E,$31,$F0,$5E,$B8,$F9,$27,$47,$4C,$30,$7A,$9E,$C7
	.dc.b	$C7,$AE,$CF,$C6,$AE,$AE,$8C,$1E,$43,$07,$C8,$AF,$01,$D4,$B1,$F0
	.dc.b	$2B,$C2,$D3,$AB,$E3,$33,$3E,$78,$C9,$EC,$3C,$73,$D5,$A7,$38,$F2
	.dc.b	$4F,$38,$E3,$1C,$CC,$BB,$7E,$01,$EA,$C3,$E7,$63,$F8,$49,$D3,$67
	.dc.b	$EA,$DF,$B1,$B3,$E2,$7B,$1D,$2F,$C4,$F8,$0D,$9F,$26,$7D,$4D,$9F
	.dc.b	$73,$4D,$9E,$68,$FB,$1E,$0F,$E9,$63,$F0,$ED,$7F,$0A,$D7,$F5,$6E
	.dc.b	$7E,$3D,$E3,$B1,$FE,$51,$39,$8C,$52,$31,$3A,$BF,$E5,$D2,$BE,$73
	.dc.b	$98,$FE,$65,$18,$7F,$22,$7E,$EB,$D3,$E3,$5F,$B6,$C5,$18,$7A,$D7
	.dc.b	$19,$FD,$DB,$63,$C2,$70,$AF,$D0,$1F,$D0,$5C,$3E,$08,$E0,$F9,$31
	.dc.b	$F6,$E5,$E1,$C3,$D6,$D9,$FD,$75,$7A,$63,$C3,$5F,$13,$0D,$57,$B3
	.dc.b	$0E,$F1,$AC,$7C,$73,$F1,$6B,$BD,$7A,$31,$EE,$D9,$EF,$3A,$7C,$D5
	.dc.b	$FB,$F9,$F5,$E1,$E0,$3B,$4F,$EE,$31,$F3,$4F,$38,$FD,$DB,$C0,$F9
	.dc.b	$77,$D6,$3C,$97,$E6,$9F,$49,$C7,$1E,$A7,$6C,$3C,$11,$D5,$C3,$D2
	.dc.b	$7C,$D8,$FB,$2F,$E9,$35,$EE,$79,$F0,$F9,$55,$DC,$91,$F0,$0F,$B1
	.dc.b	$1E,$78,$F5,$47,$43,$E3,$D4,$7D,$F5,$DE,$77,$1E,$0B,$F5,$C7,$68
	.dc.b	$FC,$91,$F5,$B1,$FC,$2B,$B8,$F7,$11,$F2,$4F,$C0,$7C,$47,$80,$F1
	.dc.b	$39,$7A,$99,$3E,$05,$F9,$F1,$FC,$49,$2B,$E5,$4F,$F3,$6E,$1F,$ED
	.dc.b	$3F,$1B,$FD,$7F,$6F,$F0,$E4,$62,$7F,$C8,$13,$97,$FC,$2D,$F1,$B6
	.dc.b	$75,$1F,$E5,$36,$BF,$B8,$6B,$D3,$8D,$5F,$E0,$98,$FF,$10,$5F,$39
	.dc.b	$7C,$65,$31,$12,$7F,$CA,$51,$97,$FC,$A9,$1E,$3F,$C1,$1F,$1B,$49
	.dc.b	$3F,$E3,$19,$F6,$CB,$E3,$73,$3E,$98,$E2,$FF,$83,$5A,$E6,$BD,$19
	.dc.b	$25,$F1,$BE,$12,$57,$71,$3B,$F7,$33,$1F,$C8,$F1,$F1,$FC,$6F,$A4
	.dc.b	$CF,$0A,$3C,$FF,$17,$ED,$25,$75,$C3,$C5,$F1,$B8,$8E,$87,$83,$37
	.dc.b	$C6,$FC,$53,$58,$FF,$9A,$7E,$5D,$F1,$BF,$76,$D1,$87,$FD,$36,$D7
	.dc.b	$FC,$D3,$F8,$07,$C4,$AE,$55,$F6,$E7,$8B,$6B,$84,$7F,$A5,$DA,$FF
	.dc.b	$3C,$7A,$4F,$BE,$3F,$A3,$57,$F8,$9F,$4F,$F3,$43,$63,$FC,$E1,$5C
	.dc.b	$27,$ED,$FC,$6F,$E0,$49,$3F,$17,$27,$C7,$62,$3E,$2E,$BF,$67,$3F
	.dc.b	$C4,$19,$CF,$C5,$1F,$77,$E3,$7F,$19,$F1,$D2,$FA,$47,$E3,$CC,$3D
	.dc.b	$31,$4E,$63,$FD,$87,$ED,$7E,$81,$BD,$3F,$3E,$31,$1F,$EB,$5F,$B5
	.dc.b	$7C,$2E,$6B,$E6,$56,$AB,$FD,$80,$D7,$FB,$2B,$5F,$F4,$A3,$D3,$3F
	.dc.b	$F5,$35,$BE,$5B,$EF,$36,$BC,$67,$C7,$F9,$0F,$D5,$7C,$84,$8C,$10
	.dc.b	$F7,$39,$7E,$04,$7E,$C7,$E3,$A4,$7D,$17,$F6,$9E,$0E,$23,$E3,$7C
	.dc.b	$87,$ED,$CC,$65,$F4,$37,$FE,$D6,$DF,$90,$FD,$E1,$FB,$DF,$1E,$5F
	.dc.b	$DD,$8C,$BF,$7E,$32,$FD,$38,$CB,$F5,$63,$F8,$33,$FC,$1F,$E3,$A3
	.dc.b	$F8,$73,$27,$E2,$5C,$BF,$74,$FE,$41,$F8,$27,$C9,$7F,$16,$DB,$FC
	.dc.b	$4B,$7A,$7F,$95,$AF,$8F,$E3,$C3,$2F,$F7,$53,$B7,$F8,$3B,$B7,$3F
	.dc.b	$1D,$E0,$23,$C3,$96,$BF,$CC,$65,$3F,$C5,$9B,$1F,$EA,$76,$BB,$8F
	.dc.b	$5F,$C7,$C8,$EE,$79,$3E,$3B,$DF,$F1,$F2,$30,$7A,$3E,$3F,$B1,$1E
	.dc.b	$CF,$87,$E3,$FE,$39,$23,$DF,$F1,$FF,$2B,$E3,$FE,$5F,$C7,$C9,$F8
	.dc.b	$71,$F3,$BD,$4F,$51,$23,$D4,$FB,$7D,$7F,$A6,$48,$F9,$7F,$19,$FC
	.dc.b	$6F,$E4,$DF,$03,$FC,$55,$B3,$FD,$45,$AF,$FB,$6D,$AF,$FB,$07,$F2
	.dc.b	$FF,$E1,$2F,$FB,$7F,$9D,$18,$8F,$F2,$A9,$FE,$5D,$70,$F1,$F2,$3E
	.dc.b	$97,$C8,$7E,$20,$91,$F8,$0F,$13,$FB,$57,$E2,$3F,$D8,$23,$F0,$7E
	.dc.b	$BF,$E3,$8C,$FA,$FF,$90,$F9,$1F,$B8,$49,$FB,$8F,$DF,$F7,$7F,$2A
	.dc.b	$48,$FC,$6F,$B7,$FD,$DF,$E5,$9F,$23,$7E,$46,$57,$F4,$B9,$CB,$7E
	.dc.b	$47,$FA,$FC,$1A,$48,$FF,$79,$43,$E4,$7B,$12,$3A,$7C,$87,$87,$E3
	.dc.b	$FC,$5F,$1D,$27,$C1,$15,$FF,$05,$5A,$7C,$24,$8E,$FF,$1F,$EB,$23
	.dc.b	$E9,$8F,$77,$44,$8F,$63,$F0,$7E,$3F,$11,$DA,$3D,$A4,$7D,$6F,$C7
	.dc.b	$F8,$F9,$1C,$8F,$8D,$F1,$FF,$2C,$8F,$C9,$F8,$FA,$FA,$A3,$E4,$91
	.dc.b	$F8,$FF,$1F,$F4,$FE,$3E,$47,$D1,$8F,$AC,$61,$FE,$AF,$A2,$73,$1F
	.dc.b	$5F,$E3,$BE,$E7,$C7,$C8,$C1,$0E,$2E,$5E,$71,$FD,$FE,$8F,$FA,$C6
	.dc.b	$1F,$1E,$EA,$FF,$58,$D6,$BA,$15,$A9,$CD,$F1,$BE,$13,$AB,$FC,$43
	.dc.b	$C2,$32,$A3,$A3,$CC,$E1,$3C,$EB,$AD,$F1,$7C,$7C,$BF,$EC,$4C,$39
	.dc.b	$E1,$C9,$B7,$C7,$0E,$78,$71,$7C,$97,$C7,$0E,$B1,$E1,$3C,$95,$C5
	.dc.b	$B5,$C5,$F4,$57,$17,$A9,$C7,$0F,$15,$F4,$C3,$CB,$5B,$AE,$67,$F0
	.dc.b	$82,$9F,$C2,$AB,$E1,$B7,$0F,$A8,$FF,$0E,$F8,$F7,$8C,$F9,$7F,$F3
	.dc.b	$54,$61,$B1,$99,$ED,$39,$7C,$16,$F8,$F7,$AD,$FE,$73,$8C,$36,$33
	.dc.b	$3C,$1A,$F9,$DC,$E7,$F4,$9B,$A7,$FB,$E3,$D8,$7C,$2F,$F8,$F6,$B5
	.dc.b	$87,$CB,$C7,$BE,$3F,$27,$1E,$2F,$E3,$B0,$E3,$EC,$6E,$73,$3F,$89
	.dc.b	$6D,$72,$C3,$F1,$B7,$C2,$BA,$61,$C4,$8F,$DF,$8D,$4F,$92,$36,$47
	.dc.b	$9E,$3E,$9C,$BF,$47,$8F,$9B,$E1,$3C,$59,$FF,$29,$8F,$F2,$D7,$93
	.dc.b	$FE,$4A,$D3,$FD,$3F,$39,$3F,$97,$7C,$16,$9F,$F5,$7C,$7D,$37,$26
	.dc.b	$F3,$FF,$2B,$E9,$FE,$97,$8F,$E6,$DE,$2F,$19,$FF,$46,$CE,$5F,$E0
	.dc.b	$F1,$9C,$93,$9C,$7F,$CB,$AD,$BF,$09,$B7,$17,$FB,$EF,$AC,$FF,$42
	.dc.b	$CF,$FA,$DD,$B1,$F9,$96,$BF,$D8,$9A,$FF,$06,$0F,$0A,$FF,$3F,$46
	.dc.b	$3E,$1B,$D4,$BE,$FD,$E8,$62,$F8,$61,$CD,$FB,$98,$7F,$A1,$F9,$5F
	.dc.b	$89,$BC,$CE,$1C,$7F,$16,$73,$18,$C3,$D5,$3E,$2C,$FA,$35,$F8,$0D
	.dc.b	$7D,$2F,$17,$FD,$5A,$79,$F8,$F1,$8F,$CC,$F9,$1F,$94,$7A,$79,$AE
	.dc.b	$1F,$EC,$3B,$F8,$71,$D7,$4F,$D8,$CF,$CD,$8C,$1E,$3E,$1F,$B3,$F2
	.dc.b	$1C,$E7,$FD,$81,$C4,$8C,$61,$F8,$6F,$90,$FB,$47,$E0,$7F,$09,$7C
	.dc.b	$3E,$30,$F1,$67,$FC,$E9,$FB,$97,$E3,$23,$FC,$57,$25,$FE,$2A,$3F
	.dc.b	$DA,$0D,$9F,$CA,$FC,$97,$7C,$FF,$15,$F2,$12,$FC,$FF,$23,$6F,$E4
	.dc.b	$FE,$04,$F2,$7E,$73,$6B,$F9,$24,$A3,$0F,$F2,$76,$BF,$8D,$CD,$DB
	.dc.b	$FE,$60,$AF,$F7,$23,$67,$F5,$3F,$27,$FA,$38,$C7,$27,$FA,$FB,$E6
	.dc.b	$DE,$FF,$E9,$DB,$87,$FB,$B5,$7C,$0E,$67,$F2,$FF,$37,$F4,$6B,$CE
	.dc.b	$DC,$BF,$9E,$35,$FF,$22,$B5,$FE,$D3,$9B,$EA,$5F,$23,$FB,$76,$DF
	.dc.b	$FB,$F5,$7C,$A6,$67,$F6,$DE,$1F,$ED,$FD,$FB,$7F,$84,$7F,$57,$F3
	.dc.b	$0F,$91,$FC,$D3,$67,$E6,$B7,$3F,$EA,$4D,$7E,$67,$3F,$FB,$6A,$7F
	.dc.b	$1F,$F2,$BF,$C1,$7C,$2F,$E0,$CF,$F0,$7C,$FF,$84,$69,$6B,$F2,$AF
	.dc.b	$EF,$3D,$F7,$BD,$FA,$B6,$CF,$DE,$6E,$5F,$D8,$DA,$F8,$BE,$4F,$F8
	.dc.b	$A1,$FC,$57,$E5,$2D,$7E,$AF,$CF,$FF,$12,$3F,$E1,$4E,$8F,$E6,$3E
	.dc.b	$63,$D5,$FE,$1B,$5A,$C5,$5A,$9C,$FA,$DF,$BF,$AD,$57,$F9,$F1,$AF
	.dc.b	$F8,$3E,$34,$70,$C3,$C0,$78,$7F,$02,$67,$9A,$FE,$73,$E3,$7F,$2D
	.dc.b	$3F,$97,$72,$DB,$FD,$DC,$59,$FE,$7F,$37,$3F,$E6,$FF,$29,$FC,$D4
	.dc.b	$FE,$6D,$CB,$70,$FE,$07,$16,$FF,$A3,$CD,$AF,$E7,$F1,$FC,$F4,$FE
	.dc.b	$7D,$E5,$5B,$FE,$17,$14,$FE,$9F,$16,$78,$CE,$8F,$E2,$31,$97,$F3
	.dc.b	$7F,$94,$3E,$4F,$FA,$83,$FE,$DF,$8F,$8B,$6E,$D4,$EE,$73,$18,$F9
	.dc.b	$5E,$37,$27,$34,$F9,$5C,$61,$F6,$A7,$C3,$F2,$9E,$03,$C3,$F2,$96
	.dc.b	$38,$C5,$9E,$07,$8E,$D5,$9D,$3F,$AA,$FE,$78,$F3,$3F,$B7,$47,$FA
	.dc.b	$77,$E2,$7F,$AB,$CE,$72,$F6,$CE,$5F,$E7,$D1,$9C,$7F,$C7,$CD,$7F
	.dc.b	$D8,$ED,$7F,$B3,$35,$FF,$36,$BC,$67,$C7,$7E,$2B,$7C,$47,$8A,$3F
	.dc.b	$BF,$7B,$4E,$DA,$63,$E2,$EE,$7F,$3A,$74,$D9,$D3,$4F,$A7,$F8,$FB
	.dc.b	$E2,$FF,$C1,$4F,$FB,$22,$33,$18,$F8,$CC,$B2,$38,$FC,$25,$DF,$FB
	.dc.b	$3C,$7F,$96,$0F,$8F,$FA,$B7,$F8,$92,$33,$F0,$9F,$3A,$31,$C1,$FE
	.dc.b	$06,$AE,$15,$FE,$67,$6B,$FE,$F0,$6B,$FE,$0F,$6B,$FD,$76,$BF,$C1
	.dc.b	$B1,$CA,$7F,$C6,$B4,$D3,$35,$FE,$12,$FD,$5F,$FC,$65,$87,$E1,$BA
	.dc.b	$5A,$BE,$33,$5F,$0B,$F1,$1B,$A7,$F7,$DC,$38,$E8,$8F,$E7,$F7,$FD
	.dc.b	$6A,$3F,$C4,$33,$A9,$CF,$87,$F3,$BE,$2F,$E7,$B6,$3E,$9B,$87,$EA
	.dc.b	$FC,$5D,$D3,$CD,$D4,$F9,$8F,$F0,$43,$6B,$FB,$D3,$67,$FC,$68,$7E
	.dc.b	$86,$76,$7E,$E0,$FD,$CF,$CC,$5A,$FD,$8F,$CC,$7E,$F3,$D8,$FD,$EB
	.dc.b	$63,$24,$BF,$1F,$B1,$A8,$FD,$CF,$1F,$90,$8F,$91,$B3,$EA,$7F,$84
	.dc.b	$FC,$C7,$F0,$A7,$FC,$8B,$1F,$EC,$93,$F3,$DE,$57,$F9,$37,$87,$FC
	.dc.b	$8F,$7F,$ED,$4E,$BF,$E6,$35,$FC,$E4,$D9,$FD,$17,$E2,$BE,$A3,$83
	.dc.b	$FD,$94,$3D,$6E,$2F,$F4,$68,$DC,$62,$3C,$4E,$AF,$C3,$1C,$70,$FF
	.dc.b	$5E,$79,$DF,$EE,$6A,$D5,$F2,$9C,$D6,$CF,$26,$3B,$8C,$D7,$08,$C4
	.dc.b	$FA,$E3,$0F,$13,$D3,$8E,$0F,$33,$FE,$FA,$3D,$D8,$F9,$A3,$38,$F4
	.dc.b	$9F,$75,$D7,$DB,$87,$CD,$C7,$D1,$1E,$18,$C4,$EE,$3B,$63,$E4,$3A
	.dc.b	$BE,$0A,$F9,$F8,$60,$F3,$5F,$B8,$E7,$3E,$B3,$C2,$7C,$43,$07,$67
	.dc.b	$75,$D8,$F0,$3F,$32,$7A,$CF,$E0,$5A,$FD,$E9,$DB,$EB,$3B,$3C,$E7
	.dc.b	$A9,$E5,$8C,$B5,$F8,$E7,$93,$0F,$95,$8F,$E0,$AB,$DD,$5E,$CB,$B7
	.dc.b	$EB,$7B,$3D,$0F,$11,$D2,$76,$F6,$8B,$5F,$66,$BD,$C7,$52,$CF,$57
	.dc.b	$F1,$BF,$1D,$FD,$0E,$3F,$44,$D7,$F3,$ED,$7F,$58,$E7,$E3,$5F,$31
	.dc.b	$FE,$26,$94,$E6,$30,$DB,$4E,$B0,$FF,$60,$5F,$29,$CD,$6E,$BF,$61
	.dc.b	$83,$CE,$FF,$63,$61,$FB,$08,$C4,$FE,$ED,$E7,$8F,$8E,$7F,$6B,$8F
	.dc.b	$5A,$E7,$7C,$AF,$FD,$A0,$72,$8F,$DF,$4E,$EB,$8D,$FB,$72,$F0,$61
	.dc.b	$FE,$D7,$6D,$FE,$96,$FB,$47,$0C,$BC,$93,$F0,$31,$ED,$8F,$08,$EF
	.dc.b	$8F,$9E,$7D,$35,$FA,$F7,$C9,$7B,$3A,$57,$EE,$F0,$EF,$3E,$AC,$7C
	.dc.b	$8D,$9F,$26,$1E,$98,$EB,$1F,$C0,$DE,$A6,$DE,$91,$B7,$EB,$D7,$C3
	.dc.b	$9D,$63,$E9,$3C,$67,$93,$0F,$34,$73,$70,$F8,$1C,$D7,$C3,$C3,$CE
	.dc.b	$F6,$C3,$C0,$B0,$F0,$3B,$3E,$59,$FE,$68,$9F,$B4,$F1,$C7,$33,$F1
	.dc.b	$23,$81,$EF,$9E,$EF,$5A,$EF,$3B,$8E,$87,$D1,$9D,$C6,$4F,$1C,$68
	.dc.b	$FC,$59,$E0,$9F,$B4,$F8,$CF,$01,$CC,$FA,$75,$1C,$92,$FE,$3B,$C5
	.dc.b	$F8,$EF,$77,$2F,$A6,$3E,$1E,$1F,$8C,$3E,$A9,$F7,$B1,$FE,$BB,$DB
	.dc.b	$FE,$6A,$F9,$8F,$EC,$5B,$FF,$9C,$9B,$87,$F9,$75,$B3,$FE,$70,$CA
	.dc.b	$F3,$A7,$FD,$DF,$3C,$EB,$FC,$0F,$18,$9D,$47,$F8,$2B,$E5,$B8,$91
	.dc.b	$FF,$3A,$F0,$D7,$FC,$27,$F1,$BD,$8C,$6B,$FD,$C3,$1F,$F4,$0F,$8D
	.dc.b	$E2,$25,$7F,$9F,$BE,$36,$BF,$E1,$69,$DB,$FE,$1A,$9F,$21,$E8,$E0
	.dc.b	$FE,$E5,$B7,$FB,$B1,$B3,$FE,$8F,$6D,$7F,$7D,$F6,$FE,$1C,$60,$7B
	.dc.b	$38,$33,$C5,$F2,$AD,$FA,$C6,$9F,$F4,$D9,$89,$D4,$66,$3F,$C9,$14
	.dc.b	$F5,$F3,$5E,$0F,$5A,$57,$2B,$E1,$B7,$F8,$52,$1F,$82,$3D,$6F,$13
	.dc.b	$63,$C2,$D7,$FD,$36,$38,$9F,$2B,$C2,$BF,$C7,$F3,$FC,$D8,$EA,$E1
	.dc.b	$FE,$73,$5C,$DF,$F2,$2B,$5E,$0F,$7B,$F3,$5F,$F9,$8E,$7E,$1B,$5F
	.dc.b	$1C,$EF,$83,$FA,$2F,$03,$F5,$9F,$F5,$7D,$7F,$B0,$1A,$FF,$9B,$DB
	.dc.b	$87,$E1,$DF,$F6,$04,$6F,$D5,$FC,$DC,$7F,$B3,$E3,$97,$B1,$89,$CF
	.dc.b	$89,$27,$B3,$C3,$C6,$C4,$77,$8B,$87,$79,$FC,$07,$43,$E5,$E9,$2F
	.dc.b	$F0,$3F,$3C,$F1,$31,$1F,$3A,$B8,$3F,$35,$AE,$6B,$E8,$28,$CE,$1F
	.dc.b	$81,$B7,$8F,$F9,$16,$D7,$E4,$2B,$7B,$FF,$BC,$3C,$4F,$DD,$92,$7F
	.dc.b	$DC,$FE,$15,$7F,$D0,$B1,$71,$FD,$97,$81,$FE,$03,$7F,$69,$F3,$4F
	.dc.b	$4F,$F8,$3B,$6F,$FD,$EA,$D7,$FD,$48,$BE,$07,$73,$C3,$D6,$95,$C2
	.dc.b	$39,$FB,$32,$39,$BD,$BD,$7C,$47,$48,$B8,$F9,$26,$EB,$F8,$AC,$95
	.dc.b	$8B,$74,$27,$FD,$B1,$1E,$8E,$2F,$59,$1F,$5F,$B3,$5F,$5C,$5C,$BE
	.dc.b	$1C,$FC,$2F,$56,$57,$C2,$7E,$2F,$CD,$62,$3E,$1C,$5A,$F9,$73,$F8
	.dc.b	$AC,$E5,$7D,$97,$F9,$AC,$BF,$08,$7C,$AE,$AB,$F1,$55,$F8,$6E,$5F
	.dc.b	$E0,$91,$FE,$B8,$6B,$FB,$B7,$F8,$45,$FF,$A7,$07,$DD,$BC,$9E,$ED
	.dc.b	$AF,$11,$F7,$3C,$3F,$BE,$47,$EF,$FC,$8F,$F2,$1F,$C0,$5F,$11,$F9
	.dc.b	$86,$CF,$F0,$96,$BF,$C3,$9A,$FF,$0F,$E3,$FA,$9E,$C7,$E3,$3D,$A9
	.dc.b	$3F,$8C,$8F,$C7,$FC,$EC,$8F,$C7,$BF,$92,$F6,$FF,$99,$7C,$EB,$D3
	.dc.b	$FD,$4B,$6B,$F8,$D3,$5F,$E3,$ED,$7C,$5D,$28,$FC,$EF,$C4,$C9,$FC
	.dc.b	$EC,$7D,$4E,$96,$F4,$93,$FE,$5E,$E3,$F9,$D2,$E8,$FE,$8D,$F3,$FF
	.dc.b	$CC,$2B,$F9,$63,$5F,$E6,$2D,$7F,$DE,$2D,$7A,$7C,$FB,$E3,$FF,$0A
	.dc.b	$D8,$F0,$74,$EA,$73,$F2,$BF,$DF,$3D,$86,$D4,$97,$FB,$8F,$84,$FE
	.dc.b	$B5,$F5,$2F,$9C,$FE,$22,$D9,$FE,$84,$D7,$FA,$3B,$5F,$93,$F0,$FF
	.dc.b	$C1,$08,$FF,$04,$F9,$7F,$F0,$6F,$99,$FD,$07,$8B,$FB,$F3,$76,$FE
	.dc.b	$82,$DA,$FD,$BF,$2F,$E6,$27,$FA,$A7,$D4,$FF,$A4,$27,$FA,$A3,$5F
	.dc.b	$EB,$2D,$7F,$AD,$07,$0F,$F4,$EF,$9D,$7D,$37,$F8,$81,$B1,$F2,$8F
	.dc.b	$E0,$EE,$16,$3E,$8E,$7E,$1F,$3F,$FE,$01,$C3,$FC,$43,$5F,$DB,$9B
	.dc.b	$5F,$E1,$BB,$56,$FE,$0F,$F9,$ED,$70,$BF,$3B,$5F,$7C,$7F,$8C,$2B
	.dc.b	$F8,$0C,$79,$7C,$04,$62,$3C,$B9,$FB,$7C,$7E,$D1,$FE,$12,$ED,$FE
	.dc.b	$69,$9F,$F8,$25,$B5,$FD,$FD,$B2,$FA,$07,$CD,$7E,$22,$39,$79,$FF
	.dc.b	$D5,$2B,$F9,$F4,$7C,$E6,$BC,$E3,$47,$F9,$3A,$9F,$8D,$3C,$0C,$36
	.dc.b	$BA,$9E,$7F,$43,$FC,$59,$87,$F3,$68,$FF,$0C,$B7,$36,$7F,$EC,$CC
	.dc.b	$7C,$FC,$DC,$30,$F3,$13,$0F,$F0,$D6,$1C,$DF,$E2,$1B,$3E,$7F,$F0
	.dc.b	$44,$8F,$9D,$1F,$9F,$CB,$ED,$4F,$E1,$FE,$82,$56,$A3,$F1,$5F,$27
	.dc.b	$F8,$BF,$02,$4F,$28,$AF,$D6,$3F,$1B,$18,$BF,$C3,$4F,$F9,$2F,$E8
	.dc.b	$FF,$89,$7B,$1F,$EE,$8F,$8E,$FE,$31,$18,$C1,$18,$F2,$3F,$DE,$31
	.dc.b	$8D,$7F,$C7,$55,$FD,$CE,$B5,$7C,$E7,$37,$C2,$73,$7C,$4F,$F7,$BD
	.dc.b	$F1,$AD,$DF,$19,$CD,$F1,$9C,$C7,$11,$5C,$5D,$D7,$17,$35,$C5,$C9
	.dc.b	$C6,$78,$3A,$9E,$11,$F0,$6F,$CE,$7C,$4E,$4F,$F2,$C3,$EA,$AF,$E0
	.dc.b	$33,$F1,$DB,$7F,$19,$B7,$FE,$5D,$F8,$DF,$F1,$B7,$15,$F7,$7F,$45
	.dc.b	$5F,$E6,$49,$CB,$63,$2F,$6F,$11,$F3,$1F,$E6,$7C,$91,$89,$FF,$36
	.dc.b	$C6,$27,$53,$99,$E5,$1E,$E8,$E1,$9F,$F5,$7F,$88,$7C,$EF,$E8,$2F
	.dc.b	$E3,$56,$B7,$FB,$ED,$8F,$B2,$D5,$5C,$BA,$3B,$63,$27,$9B,$F5,$67
	.dc.b	$51,$82,$63,$F9,$2B,$E9,$3A,$8C,$57,$36,$D7,$E0,$E7,$58,$79,$AB
	.dc.b	$F1,$5D,$BF,$C6,$BC,$AF,$F0,$A6,$5F,$CD,$5B,$87,$F3,$2F,$9E,$B5
	.dc.b	$BA,$FD,$AC,$63,$A7,$FB,$24,$60,$FD,$CD,$E7,$E7,$A4,$7E,$F3,$6B
	.dc.b	$8F,$EC,$F1,$FE,$57,$58,$E7,$FE,$71,$56,$FF,$A4,$7B,$1F,$E2,$CB
	.dc.b	$FE,$62,$DC,$BF,$C5,$53,$FE,$AA,$9C,$B6,$33,$8F,$76,$E1,$FD,$17
	.dc.b	$D6,$F5,$61,$EB,$F9,$CF,$67,$C6,$DC,$7F,$89,$B7,$1F,$F7,$06,$2A
	.dc.b	$F1,$12,$BC,$0E,$F1,$FE,$A5,$F2,$16,$7F,$B1,$7C,$7F,$C9,$3E,$57
	.dc.b	$1D,$8F,$E4,$2F,$B4,$8F,$AD,$F8,$1F,$27,$C2,$3F,$93,$B7,$2F,$71
	.dc.b	$F5,$78,$FE,$B3,$67,$FA,$D7,$87,$FE,$D9,$7E,$43,$63,$FD,$F8,$D7
	.dc.b	$FC,$8C,$D7,$FD,$8E,$BE,$2F,$F0,$AD,$9E,$BF,$39,$89,$CF,$CE,$6B
	.dc.b	$E3,$6D,$78,$9E,$3F,$31,$F7,$70,$34,$93,$E6,$F8,$EE,$E7,$83,$E3
	.dc.b	$AC,$7F,$C0,$0F,$9F,$FF,$3B,$BF,$EE,$96,$E9,$FE,$EE,$6B,$FE,$16
	.dc.b	$6B,$FE,$E2,$F3,$BF,$0E,$FF,$99,$5A,$E7,$9F,$F2,$51,$88,$FF,$4E
	.dc.b	$E5,$FE,$C4,$BF,$9B,$F1,$FF,$13,$E6,$7E,$2B,$F1,$BC,$4F,$8F,$F3
	.dc.b	$36,$BE,$9B,$F2,$BE,$67,$E1,$12,$3E,$AB,$F1,$7E,$57,$E7,$9F,$43
	.dc.b	$D5,$B3,$F7,$3E,$67,$E9,$9F,$53,$E3,$2C,$7E,$03,$C7,$F9,$C4,$7F
	.dc.b	$02,$FD,$1F,$77,$ED,$9F,$73,$E0,$AC,$FF,$88,$7E,$85,$F2,$5F,$80
	.dc.b	$F4,$7F,$82,$C6,$27,$F1,$B1,$89,$FD,$C4,$E7,$8F,$FC,$8B,$E4,$23
	.dc.b	$11,$FE,$93,$6A,$3F,$D9,$2F,$1F,$13,$F9,$94,$F1,$AF,$0F,$C9,$F7
	.dc.b	$C7,$CD,$EC,$F9,$3D,$5F,$2B,$63,$27,$AE,$32,$D3,$DD,$ED,$FB,$3E
	.dc.b	$4F,$DA,$D8,$C9,$F0,$E3,$2D,$3E,$37,$C9,$7C,$4F,$93,$F8,$AD,$8C
	.dc.b	$9F,$C5,$F5,$47,$C6,$24,$65,$E2,$FD,$1F,$8B,$BB,$FF,$29,$A7,$C7
	.dc.b	$FD,$2F,$84,$FA,$6D,$BF,$B0,$E1,$FB,$5F,$23,$6F,$97,$C7,$7D,$A3
	.dc.b	$ED,$FC,$85,$AF,$C0,$45,$9F,$04,$DE,$4F,$EC,$90,$F9,$EF,$EB,$F7
	.dc.b	$FE,$7B,$96,$FF,$EB,$D8,$FF,$3F,$56,$FD,$0F,$E9,$F5,$AF,$80,$FE
	.dc.b	$A3,$39,$AF,$F4,$5B,$F8,$3C,$3A,$B8,$6B,$FE,$90,$6B,$F8,$93,$91
	.dc.b	$FE,$41,$2D,$7F,$BC,$F0,$FF,$03,$D7,$92,$BA,$CF,$E3,$B6,$FF,$5C
	.dc.b	$CF,$E8,$7E,$47,$F3,$FF,$1F,$FA,$06,$C6,$4F,$D3,$46,$5F,$D4,$78
	.dc.b	$FF,$A6,$CF,$F4,$ED,$8C,$93,$3F,$B9,$EF,$EA,$3E,$3F,$D4,$7C,$A7
	.dc.b	$E6,$7A,$3F,$A5,$73,$B7,$CA,$8C,$5F,$BE,$B8,$3F,$29,$AF,$FA,$D9
	.dc.b	$AF,$C7,$6B,$FB,$17,$8E,$BF,$79,$B7,$FE,$EF,$FD,$6B,$FA,$44,$E7
	.dc.b	$0F,$F0,$3F,$9D,$FE,$BE,$AF,$E1,$1F,$1F,$8F,$8E,$FE,$0E,$D8,$FF
	.dc.b	$56,$4E,$F2,$FC,$E4,$65,$FF,$03,$47,$FB,$96,$37,$F1,$DD,$BE,$47
	.dc.b	$BB,$63,$9C,$78,$BE,$3B,$17,$CB,$E3,$3F,$C1,$58,$7F,$A4,$6F,$FC
	.dc.b	$10,$D9,$FE,$2A,$D7,$F1,$5E,$2A,$35,$AF,$F1,$26,$D7,$F1,$79,$D7
	.dc.b	$C6,$FC,$2C,$3E,$1F,$C6,$CA,$F7,$FC,$6D,$CB,$DB,$37,$2E,$93,$6B
	.dc.b	$A6,$4F,$89,$F4,$92,$3D,$2D,$C3,$E3,$BF,$4B,$E2,$7E,$9B,$FA,$C8
	.dc.b	$FF,$4E,$0F,$35,$F0,$7F,$A7,$C9,$E9,$FF,$97,$E7,$35,$FE,$FC,$F8
	.dc.b	$AF,$E6,$91,$88,$FF,$7A,$CE,$EF,$F6,$F5,$FC,$E7,$E7,$75,$87,$4C
	.dc.b	$F1,$5A,$C1,$5A,$C3,$FC,$DC,$F5,$8F,$3B,$9A,$E3,$1E,$33,$8D,$7A
	.dc.b	$6B,$8B,$D6,$B8,$4E,$8E,$0E,$0E,$CF,$FA,$F0,$F8,$91,$CE,$BD,$EF
	.dc.b	$9F,$1E,$67,$0C,$7C,$75,$F0,$AA,$BE,$8A,$E4,$F8,$4F,$F5,$39,$71
	.dc.b	$F8,$8E,$4E,$07,$F6,$17,$38,$77,$8F,$29,$9C,$7E,$83,$F0,$F1,$F2
	.dc.b	$DF,$C1,$3C,$35,$C4,$F8,$0E,$67,$96,$1E,$87,$E1,$4E,$9E,$93,$EA
	.dc.b	$AF,$11,$DA,$3E,$5C,$76,$3A,$3F,$0C,$F6,$1D,$1D,$3C,$0F,$09,$AC
	.dc.b	$7E,$5E,$3F,$02,$79,$B6,$7C,$67,$0A,$F1,$CF,$94,$DE,$3F,$82,$9E
	.dc.b	$C7,$8F,$0D,$BC,$67,$EE,$1C,$5E,$A7,$A6,$B6,$D8,$FC,$C7,$A1,$FA
	.dc.b	$98,$FC,$93,$5F,$BA,$D7,$F5,$EE,$7E,$35,$EB,$B2,$FF,$0F,$4E,$63
	.dc.b	$14,$8C,$4E,$B0,$FE,$B9,$7C,$A3,$15,$FE,$27,$8C,$3F,$86,$3F,$7B
	.dc.b	$F0,$BF,$D8,$EB,$F7,$18,$A3,$0F,$4C,$7C,$33,$FB,$DB,$E6,$F0,$AD
	.dc.b	$DF,$F8,$D6,$F9,$3F,$A6,$3C,$57,$CA,$3A,$61,$FA,$67,$73,$FA,$36
	.dc.b	$E5,$DE,$BE,$0B,$6B,$A5,$7A,$8B,$5C,$B0,$D4,$7B,$F1,$F8,$37,$E7
	.dc.b	$8F,$06,$3F,$E0,$38,$D9,$2F,$FC,$0C,$78,$8B,$1B,$7D,$33,$FB,$9A
	.dc.b	$EB,$1A,$8E,$A7,$F0,$28,$CB,$F0,$B0,$F3,$D7,$C8,$8F,$51,$E7,$C7
	.dc.b	$E3,$3B,$3E,$24,$78,$5F,$9C,$D5,$5D,$9B,$3F,$22,$FC,$91,$C8,$F2
	.dc.b	$46,$5E,$D8,$FC,$57,$F0,$D3,$F2,$CF,$7C,$7C,$58,$F0,$9E,$E3,$E1
	.dc.b	$4F,$37,$C3,$86,$71,$E8,$71,$8C,$9D,$B1,$F8,$58,$77,$8F,$1B,$F3
	.dc.b	$A7,$8B,$EA,$79,$63,$EF,$8F,$AD,$3E,$43,$C7,$7E,$E3,$C5,$1C,$4E
	.dc.b	$87,$B0,$F8,$B8,$FC,$13,$F1,$27,$F5,$EE,$5F,$EF,$DF,$19,$FD,$8F
	.dc.b	$7F,$EE,$2D,$C1,$3A,$7F,$C3,$3F,$19,$98,$C4,$7F,$82,$A7,$FC,$B2
	.dc.b	$F0,$E3,$46,$27,$73,$FE,$0E,$97,$C0,$73,$6C,$7F,$89,$A3,$1F,$19
	.dc.b	$96,$46,$4F,$F1,$5C,$71,$CE,$47,$F8,$B9,$E9,$F1,$98,$8F,$F0,$FC
	.dc.b	$78,$EF,$FB,$1C,$79,$E3,$1E,$5F,$8C,$91,$FE,$38,$F8,$BA,$FF,$8D
	.dc.b	$BE,$2F,$D0,$63,$83,$FC,$4D,$1E,$B2,$1F,$E4,$0E,$3A,$FF,$8F,$BE
	.dc.b	$2F,$E0,$18,$9F,$F4,$6F,$D4,$3A,$FE,$95,$23,$0F,$FA,$31,$AF,$F2
	.dc.b	$DE,$3F,$ED,$5F,$16,$F0,$F8,$D6,$AB,$FD,$0C,$73,$BF,$AE,$D3,$EC
	.dc.b	$C6,$DF,$F4,$63,$DB,$06,$1D,$BC,$0F,$C3,$43,$14,$78,$30,$FC,$37
	.dc.b	$DF,$67,$1F,$EF,$D8,$7F,$92,$63,$11,$E9,$3F,$1E,$E3,$0F,$F7,$0A
	.dc.b	$F9,$EE,$78,$F4,$F8,$E9,$7D,$0E,$31,$E0,$F3,$3C,$3F,$1D,$27,$5D
	.dc.b	$2C,$55,$B9,$CC,$2B,$84,$79,$6B,$85,$7D,$4F,$9E,$7C,$2D,$25,$7F
	.dc.b	$B0,$3E,$77,$FB,$36,$BF,$EC,$8A,$DC,$FE,$99,$AF,$FB,$01,$AF,$D6
	.dc.b	$6B,$FE,$BD,$77,$1F,$ED,$29,$7C,$2F,$E1,$9B,$5F,$ED,$0A,$DE,$FF
	.dc.b	$3F,$E3,$BE,$99,$27,$DF,$F2,$15,$F7,$6C,$9C,$65,$F4,$AB,$ED,$FC
	.dc.b	$6C,$BF,$8E,$FD,$EF,$90,$C4,$72,$F8,$FF,$C0,$98,$BF,$B7,$E4,$7E
	.dc.b	$10,$93,$F4,$3E,$3A,$BF,$96,$E2,$FA,$D3,$FC,$27,$4F,$F0,$5C,$FF
	.dc.b	$B9,$A7,$2D,$8C,$BE,$AF,$8F,$FC,$89,$D3,$0F,$C5,$F9,$9F,$94,$F8
	.dc.b	$F9,$5F,$80,$8F,$CB,$98,$7E,$1C,$61,$FC,$2F,$C7,$FE,$73,$96,$46
	.dc.b	$08,$7C,$A7,$2F,$CC,$8F,$D1,$F9,$72,$3F,$1E,$FE,$9B,$E3,$F1,$1F
	.dc.b	$07,$3F,$F1,$63,$FA,$79,$F8,$44,$7F,$2F,$CF,$FA,$E3,$8E,$7F,$B6
	.dc.b	$8F,$E7,$30,$F8,$EF,$E7,$0D,$8F,$D1,$B5,$FC,$0F,$A1,$AF,$8F,$FE
	.dc.b	$71,$26,$92,$7E,$94,$3E,$3F,$A9,$23,$AF,$C7,$F8,$3E,$3E,$47,$81
	.dc.b	$F1,$FB,$DE,$4F,$8F,$EE,$49,$EA,$F8,$BE,$3F,$D3,$F1,$F2,$7D,$31
	.dc.b	$EC,$F8,$D9,$1E,$C7,$DF,$F1,$FF,$C5,$B1,$3F,$D5,$B8,$11,$8A,$FE
	.dc.b	$4A,$7F,$95,$DC,$3F,$CA,$23,$73,$FC,$A3,$E6,$57,$49,$5F,$D8,$A3
	.dc.b	$2F,$C6,$F8,$EF,$A2,$47,$E8,$FC,$87,$D4,$F9,$0F,$9E,$48,$F9,$AF
	.dc.b	$D2,$F8,$DF,$B3,$F1,$D2,$7E,$CC,$7D,$CF,$8C,$91,$F7,$1F,$BF,$F1
	.dc.b	$DF,$80,$F8,$EF,$B6,$49,$FB,$2F,$DE,$F2,$BF,$0B,$F1,$72,$7F,$0B
	.dc.b	$1F,$DF,$FF,$1A,$7C,$6D,$F6,$25,$7F,$2C,$9C,$B6,$3F,$C0,$11,$FD
	.dc.b	$4E,$3F,$70,$D7,$F7,$CD,$44,$F1,$D0,$F8,$CF,$11,$27,$B1,$E1,$F6
	.dc.b	$3C,$C4,$7C,$9F,$1B,$5E,$F1,$E4,$23,$E4,$7D,$DF,$1B,$23,$A1,$ED
	.dc.b	$F8,$DF,$84,$47,$E0,$7C,$6D,$7D,$71,$F0,$08,$FC,$0F,$8D,$F9,$44
	.dc.b	$7E,$47,$BB,$5F,$E9,$D9,$A7,$1D,$9F,$BB,$D7,$E1,$DF,$D3,$F8,$AF
	.dc.b	$A9,$F1,$53,$1F,$A5,$15,$E7,$16,$F9,$62,$AF,$F0,$6F,$CA,$39,$BF
	.dc.b	$D6,$58,$3E,$39,$E7,$7D,$48,$56,$FC,$BF,$AF,$5A,$9F,$B3,$39,$79
	.dc.b	$57,$18,$79,$1C,$EB,$85,$74,$C3,$C3,$18,$AE,$4D,$8C,$DF,$3B,$E4
	.dc.b	$F5,$8C,$DF,$89,$AF,$5C,$3C,$F3,$C6,$29,$B8,$AE,$CF,$49,$92,$9F
	.dc.b	$0A,$BC,$2E,$6F,$CD,$3C,$B0,$F4,$DF,$7A,$E3,$1F,$15,$FF,$19,$B5
	.dc.b	$F5,$4F,$F9,$AA,$31,$1F,$ED,$FF,$4D,$F2,$BF,$E6,$EF,$8E,$46,$27
	.dc.b	$F8,$CC,$62,$75,$39,$9E,$FD,$CF,$85,$EC,$D8,$FE,$43,$82,$31,$1C
	.dc.b	$27,$67,$B6,$38,$67,$F6,$BE,$27,$F1,$9D,$D9,$A9,$7F,$02,$F7,$AF
	.dc.b	$E2,$EF,$73,$9A,$DE,$1F,$81,$BD,$CF,$F0,$4A,$DC,$E6,$B7,$7F,$8D
	.dc.b	$BD,$DA,$F7,$39,$AD,$CF,$E5,$AB,$75,$A8,$C4,$FE,$13,$97,$F8,$8F
	.dc.b	$C7,$7F,$6B,$E4,$FD,$4F,$C7,$5D,$3F,$8B,$72,$FF,$A6,$9F,$E4,$51
	.dc.b	$9C,$7F,$D5,$33,$97,$F7,$B1,$9C,$3A,$D7,$03,$F9,$57,$C6,$66,$DF
	.dc.b	$1D,$63,$0D,$9F,$E6,$0D,$C3,$F8,$4B,$C7,$E3,$3F,$C0,$51,$FB,$9C
	.dc.b	$3F,$95,$C6,$23,$D2,$D7,$CE,$D8,$DD,$7F,$3D,$8F,$27,$C6,$F5,$8F
	.dc.b	$E8,$1A,$F9,$F6,$F4,$36,$32,$7F,$40,$7A,$67,$F9,$18,$C1,$31,$FE
	.dc.b	$43,$3F,$07,$E3,$79,$CF,$F4,$D9,$F8,$87,$C9,$FA,$6F,$F4,$24,$64
	.dc.b	$FF,$3E,$B8,$8F,$F2,$53,$5F,$84,$D7,$FA,$73,$5F,$F6,$FA,$F8,$CE
	.dc.b	$DF,$1B,$DD,$B5,$DB,$E3,$3C,$47,$8F,$E3,$6C,$78,$BE,$37,$E5,$61
	.dc.b	$FE,$8E,$8F,$F2,$C3,$71,$F2,$CE,$BE,$2F,$D9,$F1,$7E,$D6,$DF,$B3
	.dc.b	$E2,$7E,$01,$F0,$7E,$2A,$C7,$FA,$73,$D0,$FF,$59,$3F,$E6,$56,$E3
	.dc.b	$FE,$6B,$6B,$F8,$36,$BF,$CF,$7E,$3F,$FB,$74,$FF,$BE,$9B,$1F,$E2
	.dc.b	$46,$BF,$EC,$A6,$BF,$2B,$E2,$FE,$A1,$F5,$7E,$32,$D7,$FA,$C3,$E5
	.dc.b	$5D,$7F,$8D,$6C,$7F,$9E,$DB,$7F,$E8,$16,$BF,$99,$5D,$1F,$A9,$6C
	.dc.b	$FE,$03,$C3,$C4,$FE,$17,$E2,$E4,$FE,$13,$E3,$6E,$1F,$EB,$BE,$37
	.dc.b	$C6,$FD,$62,$57,$D6,$7E,$CF,$C6,$7E,$3F,$E3,$7F,$20,$D9,$FC,$7F
	.dc.b	$C4,$FE,$4C,$FC,$A7,$8F,$63,$F2,$7F,$19,$F8,$22,$3F,$82,$7F,$5D
	.dc.b	$F9,$2F,$F9,$D6,$3F,$33,$E7,$7F,$05,$8C,$69,$E7,$8F,$E1,$8B,$E5
	.dc.b	$BF,$C6,$75,$FE,$E1,$F2,$5F,$5C,$C7,$75,$AB,$FF,$77,$56,$A7,$3E
	.dc.b	$F7,$F0,$19,$E8,$DA,$E1,$1F,$E0,$13,$38,$7F,$99,$9A,$7F,$BE,$4B
	.dc.b	$7F,$C9,$F0,$63,$DA,$BF,$DE,$39,$7F,$80,$9B,$E9,$7B,$FE,$3B,$E0
	.dc.b	$32,$32,$70,$8C,$63,$FE,$F3,$AF,$F3,$CC,$F4,$F8,$EF,$91,$F1,$5F
	.dc.b	$25,$B1,$E0,$8F,$99,$F1,$D8,$BE,$5F,$19,$F3,$FE,$3B,$E8,$36,$FC
	.dc.b	$EF,$D6,$23,$FE,$D1,$C5,$18,$AE,$73,$EE,$F8,$EE,$6F,$C0,$F8,$AF
	.dc.b	$B9,$F1,$DF,$75,$B1,$92,$57,$B7,$E3,$B5,$1F,$23,$E3,$3F,$09,$F1
	.dc.b	$BF,$85,$6C,$64,$F9,$B1,$96,$9F,$E6,$CE,$4F,$E8,$7C,$BF,$E6,$E8
	.dc.b	$FC,$47,$D3,$37,$FC,$3E,$5F,$D6,$6D,$39,$AE,$47,$F9,$0E,$F9,$57
	.dc.b	$FA,$A1,$AF,$E1,$DA,$EC,$FC,$61,$DA,$3C,$3A,$D8,$FB,$9F,$01,$FA
	.dc.b	$1F,$8E,$FD,$13,$6B,$F1,$2E,$1F,$C5,$FC,$75,$AF,$C2,$79,$7F,$A9
	.dc.b	$F7,$BF,$54,$DA,$FC,$8B,$87,$F2,$7F,$1D,$6B,$F3,$5F,$2F,$F8,$C7
	.dc.b	$F1,$D3,$FB,$08,$C6,$DF,$E7,$3E,$64,$E7,$A3,$F5,$55,$C6,$FF,$50
	.dc.b	$D7,$FC,$D4,$D7,$FD,$3E,$F4,$9C,$CF,$F8,$17,$D8,$FD,$C6,$1F,$DE
	.dc.b	$AB,$56,$F7,$54,$F8,$DE,$1F,$1F,$C5,$B9,$7D,$8B,$E1,$5B,$8C,$5F
	.dc.b	$2A,$F0,$7C,$7F,$19,$CD,$3E,$37,$C9,$F1,$DE,$56,$CE,$9C,$3C,$3E
	.dc.b	$3E,$E4,$D7,$FD,$35,$83,$E3,$7F,$C1,$CD,$BF,$F7,$8E,$9E,$C9,$C9
	.dc.b	$23,$C6,$FC,$5F,$8F,$F8,$CF,$C7,$F8,$DF,$91,$F1,$F6,$BD,$2D,$CB
	.dc.b	$C5,$1F,$EC,$DC,$7F,$D5,$CF,$BA,$B8,$BF,$03,$E2,$7C,$A4,$7C,$B2
	.dc.b	$8C,$57,$43,$E6,$BF,$67,$E3,$FE,$D3,$FC,$E2,$3F,$D4,$E7,$F3,$5F
	.dc.b	$67,$FD,$59,$AF,$F5,$2C,$3F,$DC,$3E,$77,$F6,$6A,$FE,$61,$F1,$98
	.dc.b	$BD,$FA,$9F,$90,$9E,$46,$87,$CB,$71,$7E,$84,$6E,$31,$1E,$27,$58
	.dc.b	$F8,$63,$8D,$7F,$3F,$7F,$D8,$46,$0F,$E1,$55,$CE,$35,$7F,$EF,$8B
	.dc.b	$E7,$39,$AF,$24,$63,$1D,$5F,$09,$E8,$79,$A3,$2F,$AF,$1E,$0F,$B8
	.dc.b	$FF,$64,$17,$1F,$44,$7B,$23,$CF,$7C,$0C,$63,$D2,$3B,$1E,$03,$E1
	.dc.b	$61,$F3,$30,$D4,$60,$F7,$B8,$7E,$1D,$F9,$0E,$AF,$8F,$1F,$A9,$18
	.dc.b	$C3,$C5,$7E,$F6,$D7,$C4,$8F,$03,$E3,$3D,$2E,$23,$AC,$F7,$8F,$90
	.dc.b	$D9,$E4,$79,$8E,$58,$7B,$F1,$FB,$97,$E3,$BE,$87,$8C,$B5,$C8,$F8
	.dc.b	$B1,$B7,$A3,$E2,$8F,$3D,$D9,$F0,$3B,$78,$C7,$57,$8B,$E8,$38,$95
	.dc.b	$E2,$D7,$EA,$C7,$56,$C7,$E1,$FC,$FF,$F0,$E4,$7E,$7D,$AF,$E1,$1A
	.dc.b	$FE,$B9,$CF,$C8,$3D,$8F,$E2,$53,$FE,$14,$B3,$27,$58,$E2,$F8,$DF
	.dc.b	$06,$C6,$6F,$8D,$AF,$8B,$F9,$7C,$3F,$9A,$BF,$B9,$3B,$B8,$7F,$39
	.dc.b	$1C,$27,$F7,$0D,$9E,$A6,$E3,$F7,$8D,$3C,$31,$B9,$F5,$61,$EE,$C7
	.dc.b	$7B,$F9,$9B,$3F,$A3,$AF,$3C,$78,$2B,$E2,$E1,$AA,$F5,$5F,$BA,$35
	.dc.b	$8F,$AE,$7E,$05,$77,$C7,$CF,$7D,$DB,$3D,$E7,$4F,$9A,$BF,$69,$3F
	.dc.b	$0F,$0F,$01,$DA,$7F,$82,$E3,$E6,$9E,$71,$FC,$21,$E0,$7C,$2B,$EB
	.dc.b	$1E,$4B,$F3,$4F,$C7,$8F,$A3,$87,$AE,$38,$4F,$AE,$FB,$56,$8F,$2D
	.dc.b	$7A,$DF,$41,$E4,$BE,$F7,$E2,$8E,$06,$6F,$E0,$1E,$2A,$F0,$46,$B1
	.dc.b	$F7,$1E,$FC,$7E,$0D,$FB,$9C,$3E,$69,$F8,$15,$E3,$7F,$D5,$6D,$C7
	.dc.b	$D3,$1F,$04,$FB,$27,$84,$F7,$63,$F7,$AB,$A5,$F8,$27,$D2,$F8,$1F
	.dc.b	$7B,$97,$B9,$EC,$3D,$C7,$90,$F8,$E6,$5F,$6C,$7D,$8C,$3F,$18,$62
	.dc.b	$7F,$C2,$7F,$1B,$FD,$8B,$6F,$EE,$8D,$8C,$9F,$E1,$2F,$8E,$C7,$37
	.dc.b	$F8,$82,$7F,$D2,$4D,$9F,$EE,$B1,$8D,$0D,$F7,$38,$C7,$F6,$0C,$3F
	.dc.b	$C0,$59,$62,$65,$FF,$88,$23,$66,$1F,$EF,$D8,$7F,$8C,$3E,$37,$99
	.dc.b	$23,$FC,$4F,$F1,$D5,$FF,$54,$4E,$DF,$F0,$4C,$F9,$4F,$33,$E4,$21
	.dc.b	$FE,$30,$C6,$BE,$05,$F1,$BE,$A3,$3C,$3F,$E0,$3B,$7E,$56,$F8,$BF
	.dc.b	$8E,$DC,$3F,$C7,$ED,$7D,$6B,$E3,$B7,$3C,$3E,$3A,$57,$08,$FA,$8B
	.dc.b	$E5,$DD,$F8,$6C,$7F,$93,$CD,$47,$8B,$E3,$F4,$FF,$B2,$63,$27,$FB
	.dc.b	$2E,$78,$BF,$E1,$16,$BF,$ED,$17,$39,$FB,$F2,$7C,$7F,$B8,$EB,$5E
	.dc.b	$EF,$8C,$F8,$24,$7E,$0F,$C7,$D7,$E0,$FC,$7F,$C6,$31,$87,$5F,$8D
	.dc.b	$F9,$24,$9F,$93,$F1,$F5,$F9,$3F,$1D,$F3,$8C,$57,$F8,$D6,$1F,$2D
	.dc.b	$FD,$45,$F8,$8D,$7F,$DD,$0D,$8E,$1E,$1B,$E1,$3E,$7C,$EE,$BF,$DE
	.dc.b	$4D,$7F,$A8,$B5,$FE,$B6,$D7,$C3,$E0,$FD,$8A,$7C,$AD,$3A,$C7,$E1
	.dc.b	$8F,$C3,$FC,$7C,$AF,$C3,$C6,$5F,$C4,$C6,$5F,$C4,$C7,$E3,$BE,$36
	.dc.b	$47,$E3,$9F,$C8,$FC,$7E,$23,$F1,$F8,$7E,$50,$C6,$7F,$6F,$E3,$7F
	.dc.b	$2E,$49,$FC,$BF,$C7,$57,$F5,$7C,$A9,$FE,$02,$D8,$FF,$44,$B5,$DC
	.dc.b	$7E,$8B,$E3,$A4,$7E,$8A,$3F,$85,$FE,$B7,$F1,$23,$FD,$46,$D7,$F7
	.dc.b	$ED,$C3,$F2,$1E,$47,$0D,$FF,$5B,$E5,$CB,$E3,$1D,$3E,$36,$47,$47
	.dc.b	$BF,$1E,$23,$AC,$5C,$3F,$75,$8F,$E0,$67,$F5,$5C,$0F,$86,$FD,$8B
	.dc.b	$63,$31,$FB,$4F,$98,$74,$EA,$3F,$5D,$5B,$AF,$F3,$E3,$5F,$F0,$C8
	.dc.b	$70,$FF,$87,$1E,$18,$7E,$6E,$DD,$5D,$CE,$D5,$F2,$3E,$37,$E5,$FC
	.dc.b	$74,$8C,$1F,$33,$DE,$FE,$6F,$A1,$A4,$8F,$F6,$0C,$3E,$1B,$EA,$92
	.dc.b	$3E,$AF,$C4,$FD,$8F,$8D,$FB,$3F,$1D,$27,$EC,$C7,$DC,$F8,$DE,$E4
	.dc.b	$8E,$AF,$8B,$83,$F0,$24,$8F,$C0,$FC,$77,$F4,$EF,$C0,$7E,$24,$7F
	.dc.b	$B9,$9A,$FF,$42,$6F,$0F,$D8,$F8,$FE,$1D,$7F,$D5,$24,$D2,$57,$BB
	.dc.b	$E2,$FF,$22,$47,$F2,$3F,$1B,$FC,$BB,$E6,$1A,$7F,$27,$6C,$FF,$AD
	.dc.b	$43,$87,$FD,$70,$D7,$FC,$AB,$E8,$BB,$9C,$BF,$DA,$1B,$3F,$EF,$06
	.dc.b	$BF,$D6,$1A,$FF,$B5,$BE,$87,$FD,$F0,$FF,$9A,$36,$FF,$69,$08,$CC
	.dc.b	$7F,$B8,$E3,$1F,$07,$FD,$ED,$D7,$9F,$FB,$0F,$8E,$FD,$8F,$C7,$CB
	.dc.b	$FD,$8C,$7E,$D3,$E3,$BF,$0E,$48,$FC,$2B,$F8,$AF,$77,$F7,$24,$8F
	.dc.b	$DC,$FC,$7F,$F8,$1B,$EB,$1F,$29,$7E,$3E,$57,$A6,$FF,$D0,$3C,$2B
	.dc.b	$DC,$E4,$46,$6B,$FC,$1F,$26,$92,$7F,$DD,$13,$CF,$E3,$BC,$44,$8E
	.dc.b	$91,$CD,$B4,$F8,$CF,$41,$23,$3F,$1F,$5F,$F0,$5C,$ED,$EE,$D7,$33
	.dc.b	$E6,$3E,$17,$C7,$C8,$EF,$1F,$17,$82,$47,$89,$F9,$1F,$1F,$88,$EF
	.dc.b	$1F,$00,$8F,$83,$E3,$FE,$77,$24,$8F,$5C,$7D,$13,$0F,$F8,$A3,$1F
	.dc.b	$C2,$F2,$7F,$85,$D4,$FD,$7F,$8E,$FB,$1F,$1F,$2F,$E2,$C6,$5F,$8F
	.dc.b	$19,$7A,$47,$F8,$53,$E6,$5C,$EF,$4B,$FD,$7B,$F1,$FF,$B2,$BF,$F5
	.dc.b	$FD,$6A,$73,$E0,$FF,$AE,$A3,$12,$AE,$0A,$1E,$8F,$FB,$02,$73,$1C
	.dc.b	$1B,$87,$86,$B9,$5F,$58,$C4,$F0,$BE,$53,$9C,$3C,$77,$E6,$BD,$CE
	.dc.b	$63,$99,$EA,$AE,$2E,$EB,$8B,$E7,$AE,$2E,$4E,$27,$AE,$78,$BB,$9E
	.dc.b	$38,$7A,$EF,$C1,$87,$9E,$B3,$87,$B5,$FE,$93,$5F,$21,$B1,$97,$D0
	.dc.b	$7F,$4E,$F8,$F7,$B6,$F9,$AF,$E3,$51,$86,$C6,$67,$FA,$A9,$FE,$ED
	.dc.b	$70,$F8,$3B,$BF,$CC,$6E,$B0,$46,$23,$5F,$11,$FC,$9E,$72,$D9,$E3
	.dc.b	$97,$DA,$6E,$7F,$EF,$CF,$8A,$70,$3B,$7E,$2D,$6B,$2F,$C5,$B5,$FC
	.dc.b	$33,$6B,$8C,$F0,$C6,$57,$D2,$CB,$C6,$FD,$4A,$DB,$9B,$E9,$3A,$CB
	.dc.b	$CD,$7F,$7A,$F7,$18,$BE,$13,$DA,$3F,$33,$3A,$CB,$F3,$D8,$7A,$7A
	.dc.b	$FF,$90,$7C,$8F,$F3,$AC,$FF,$91,$F2,$A3,$19,$FF,$22,$87,$C8,$62
	.dc.b	$F8,$DF,$F1,$A8,$C4,$F1,$8C,$1F,$C7,$AF,$3F,$1F,$AF,$90,$B3,$FC
	.dc.b	$89,$B8,$FF,$07,$9E,$7F,$21,$8E,$AF,$F3,$67,$CF,$7C,$6A,$FF,$0C
	.dc.b	$B5,$11,$C3,$78,$FF,$CE,$8E,$84,$7D,$09,$EB,$3F,$02,$5F,$59,$FD
	.dc.b	$EB,$2F,$F4,$E5,$ED,$E4,$BA,$FE,$3F,$3B,$EB,$7F,$CA,$D5,$C2,$73
	.dc.b	$18,$C1,$97,$B2,$7C,$B1,$E9,$6B,$E6,$6E,$1D,$EF,$FC,$04,$D9,$E7
	.dc.b	$1E,$49,$F5,$63,$FC,$8E,$3C,$47,$F5,$0F,$93,$FA,$E7,$BE,$FE,$4C
	.dc.b	$64,$F7,$DF,$80,$8F,$81,$F3,$F1,$7F,$69,$E5,$E5,$1F,$82,$6C,$FF
	.dc.b	$5E,$E7,$FC,$29,$F8,$6F,$93,$B1,$FE,$C7,$F2,$BF,$DA,$4F,$F8,$75
	.dc.b	$B7,$FC,$09,$AF,$F9,$61,$AF,$F7,$AF,$2B,$7B,$7F,$B8,$DB,$5F,$9C
	.dc.b	$9F,$CC,$46,$1F,$EF,$AF,$FB,$9E,$3F,$BA,$B6,$3F,$DF,$EB,$E4,$B8
	.dc.b	$FC,$97,$E1,$7E,$71,$E5,$FD,$66,$E1,$FE,$35,$6E,$1F,$C6,$9A,$FF
	.dc.b	$8D,$3E,$73,$EC,$65,$FE,$F9,$6D,$7A,$FE,$53,$F5,$D1,$8A,$F3,$FC
	.dc.b	$17,$A8,$F5,$FC,$9D,$BF,$F7,$87,$CF,$3E,$4B,$F1,$2D,$8F,$F2,$43
	.dc.b	$6F,$F9,$6B,$5F,$F2,$3E,$EF,$94,$F8,$E7,$C8,$F8,$9B,$7F,$1B,$E4
	.dc.b	$FE,$5E,$3F,$31,$B1,$FF,$00,$7D,$8F,$F6,$7B,$FE,$57,$6E,$3F,$CF
	.dc.b	$1A,$FF,$A4,$9A,$FD,$BF,$2B,$EB,$9F,$63,$E5,$2D,$7D,$6F,$92,$FB
	.dc.b	$7F,$27,$F7,$1B,$1E,$1F,$7F,$F5,$38,$9A,$48,$FF,$1B,$ED,$FE,$A1
	.dc.b	$7F,$03,$F1,$BF,$86,$F9,$3F,$C3,$BF,$C0,$FE,$69,$EC,$FF,$24,$E8
	.dc.b	$7C,$74,$C3,$FC,$91,$C0,$BD,$C7,$F9,$2E,$B7,$5C,$EF,$F7,$B5,$AB
	.dc.b	$FE,$BE,$D7,$FD,$6A,$D7,$F8,$C3,$BA,$FE,$A9,$1F,$C8,$23,$17,$FE
	.dc.b	$75,$BF,$F5,$73,$E4,$8C,$CF,$E8,$3D,$1F,$D1,$9F,$A4,$F9,$4B,$5F
	.dc.b	$A2,$8B,$3F,$A2,$9B,$87,$EB,$76,$FD,$59,$FA,$CF,$93,$B8,$7E,$AA
	.dc.b	$2D,$FE,$AA,$6D,$7A,$3C,$57,$D6,$7E,$12,$30,$DC,$3F,$79,$EC,$FF
	.dc.b	$4F,$8C,$61,$FE,$74,$6D,$F2,$F9,$4F,$DE,$D6,$D9,$7F,$E7,$0A,$DF
	.dc.b	$17,$F2,$3C,$3F,$93,$9C,$EB,$B7,$C2,$F7,$F9,$4B,$87,$16,$D7,$0C
	.dc.b	$BD,$35,$E5,$F7,$73,$A7,$AB,$E4,$FD,$07,$A7,$8E,$E5,$E3,$8B,$8F
	.dc.b	$3D,$7F,$14,$D8,$FE,$A5,$3F,$E2,$83,$F1,$5F,$38,$F2,$B3,$DB,$F9
	.dc.b	$38,$C4,$FE,$48,$E1,$7F,$E2,$A6,$BF,$F0,$01,$AF,$F6,$36,$BF,$D0
	.dc.b	$A7,$AC,$78,$38,$2C,$F9,$FE,$4F,$EA,$1F,$57,$E4,$ED,$7C,$F8,$B3
	.dc.b	$ED,$9B,$AF,$E6,$21,$F5,$9F,$E3,$8B,$FA,$DF,$29,$FD,$C6,$31,$3F
	.dc.b	$16,$31,$EF,$CA,$FB,$9F,$29,$6B,$ED,$8F,$D5,$FE,$5E,$3F,$ED,$6F
	.dc.b	$D7,$39,$BE,$18,$CD,$FF,$B6,$E7,$2F,$F8,$16,$3E,$9D,$7F,$81,$CE
	.dc.b	$9C,$7F,$DD,$23,$31,$B8,$C1,$FE,$09,$8E,$B1,$F9,$7F,$91,$C4,$FE
	.dc.b	$6F,$E5,$3F,$39,$F2,$B7,$0F,$CB,$36,$FF,$09,$3F,$A1,$F9,$6C,$67
	.dc.b	$FA,$6F,$8A,$FD,$27,$8F,$FA,$56,$E3,$F9,$F7,$0F,$E8,$7E,$56,$E3
	.dc.b	$F8,$FE,$8F,$EF,$3D,$5F,$88,$8F,$77,$6F,$ED,$27,$24,$BF,$D5,$3F
	.dc.b	$B3,$E3,$FD,$A3,$FE,$43,$F9,$DF,$F7,$7C,$FE,$BF,$A7,$1F,$29,$FB
	.dc.b	$96,$C7,$FB,$8A,$77,$8F,$E8,$63,$2F,$EB,$63,$FB,$3B,$FE,$F6,$6B
	.dc.b	$FE,$F4,$6E,$26,$9F,$DC,$36,$FE,$1C,$DA,$E5,$F2,$9E,$7F,$67,$FC
	.dc.b	$7D,$87,$FB,$18,$78,$CF,$86,$FF,$23,$62,$F3,$7F,$97,$CE,$6B,$F0
	.dc.b	$9F,$19,$FE,$50,$8C,$47,$F8,$F2,$77,$7F,$E9,$AA,$FF,$2A,$E7,$AC
	.dc.b	$3A,$67,$8A,$D6,$1E,$B7,$C3,$87,$F3,$EB,$EB,$3E,$77,$94,$66,$78
	.dc.b	$C6,$2B,$A5,$79,$8E,$87,$F9,$9A,$FC,$2E,$0F,$F6,$E1,$5C,$CF,$82
	.dc.b	$3F,$DC,$87,$C6,$AF,$2D,$7C,$17,$D3,$8F,$86,$74,$78,$6B,$07,$3C
	.dc.b	$7C,$B1,$F1,$EE,$BE,$DB,$EA,$F9,$4C,$1E,$9C,$7E,$03,$ED,$C3,$C0
	.dc.b	$75,$7C,$27,$52,$9C,$8C,$61,$B3,$CB,$5F,$28,$F0,$B9,$69,$EE,$3C
	.dc.b	$B3,$DC,$F0,$BC,$67,$CD,$7E,$C7,$E4,$D7,$67,$E2,$4F,$46,$9C,$0F
	.dc.b	$80,$70,$74,$FD,$68,$E4,$FD,$8A,$F0,$BF,$0E,$7C,$A7,$8D,$F7,$B7
	.dc.b	$1F,$BA,$FA,$E3,$C1,$3E,$B8,$E3,$1C,$CC,$BB,$7E,$64,$7C,$FF,$23
	.dc.b	$F5,$31,$F8,$F6,$BF,$8C,$6B,$FA,$F7,$3F,$1C,$F4,$7F,$D4,$72,$9C
	.dc.b	$C6,$1B,$87,$FA,$F9,$B3,$BB,$FF,$5E,$5F,$3A,$FC,$A4,$FF,$B1,$30
	.dc.b	$7B,$FF,$EC,$1B,$FD,$BC,$62,$76,$2F,$93,$F8,$A6,$CF,$0A,$E5,$5F
	.dc.b	$EC,$B3,$95,$6F,$1F,$5D,$7F,$04,$BE,$79,$71,$AF,$D0,$DF,$08,$D5
	.dc.b	$FC,$26,$E3,$E4,$9E,$B8,$71,$BE,$D3,$E8,$C3,$AD,$78,$F1,$FF,$6F
	.dc.b	$61,$B3,$9D,$FE,$FE,$B6,$FC,$59,$DD,$71,$8F,$DA,$CE,$9F,$89,$8F
	.dc.b	$96,$75,$1F,$C2,$5F,$51,$F4,$A3,$51,$FE,$F4,$9F,$93,$3C,$31,$F5
	.dc.b	$1E,$43,$CB,$87,$9E,$3A,$B8,$7A,$63,$F2,$31,$F3,$E1,$E9,$3C,$F8
	.dc.b	$F8,$1F,$3C,$FC,$99,$FA,$F1,$B3,$E4,$92,$FC,$0E,$5F,$03,$E9,$AE
	.dc.b	$97,$EE,$8D,$3E,$47,$E6,$47,$E0,$23,$DC,$48,$E5,$7F,$5E,$79,$DF
	.dc.b	$AA,$7F,$A8,$37,$1F,$A0,$7D,$1B,$F0,$4E,$8F,$7D,$C7,$E9,$57,$39
	.dc.b	$DD,$72,$BF,$25,$7C,$1C,$7E,$8D,$7A,$5F,$EB,$9C,$1F,$DA,$FE,$4F
	.dc.b	$FB,$0E,$BF,$DC,$5B,$7F,$DC,$5B,$3F,$DC,$7E,$52,$EA,$9F,$F1,$45
	.dc.b	$71,$2D,$75,$6C,$ED,$BF,$27,$C8,$8F,$F8,$1F,$E5,$2B,$FE,$05,$F5
	.dc.b	$3B,$98,$E0,$FE,$FF,$F2,$7E,$22,$4F,$F8,$3F,$E5,$2B,$FE,$1E,$C5
	.dc.b	$3E,$68,$FF,$06,$37,$2F,$F0,$63,$65,$F2,$3E,$E2,$57,$F8,$6F,$E3
	.dc.b	$EB,$FE,$19,$F8,$DF,$86,$63,$2F,$F0,$9F,$17,$C6,$24,$FF,$88,$FE
	.dc.b	$4E,$BF,$E2,$1F,$92,$F9,$66,$2B,$FC,$D1,$F8,$7B,$6F,$A7,$48,$C3
	.dc.b	$FE,$D2,$6B,$FE,$2B,$F9,$3D,$C6,$23,$FC,$B8,$D4,$47,$0D,$9F,$8C
	.dc.b	$F7,$7F,$07,$4F,$D7,$7E,$5C,$FF,$B7,$DA,$FF,$9D,$2B,$C7,$3F,$E8
	.dc.b	$1F,$B0,$63,$EA,$6D,$7F,$BA,$9A,$FF,$90,$FE,$4B,$53,$F8,$EF,$91
	.dc.b	$93,$B8,$FC,$8F,$C8,$C8,$FF,$27,$3F,$94,$EE,$C4,$7F,$93,$62,$E3
	.dc.b	$FE,$46,$9F,$97,$F3,$CF,$A0,$A4,$BF,$F7,$EF,$84,$F4,$B1,$1F,$9E
	.dc.b	$AE,$0F,$E2,$DA,$E6,$BE,$4C,$63,$0F,$F5,$15,$BD,$6F,$B6,$DB,$FF
	.dc.b	$1B,$78,$1F,$51,$5F,$18,$FA,$FD,$6F,$B0,$C5,$23,$19,$7E,$DA,$73
	.dc.b	$1F,$E6,$79,$C4,$C7,$B9,$C6,$77,$D7,$FE,$A1,$CF,$FD,$F1,$0F,$91
	.dc.b	$E8,$4A,$E3,$87,$8F,$E5,$31,$1F,$EE,$7A,$7C,$A7,$8C,$CE,$1D,$3E
	.dc.b	$4B,$CC,$4A,$E7,$4F,$92,$C4,$73,$9B,$5E,$49,$BA,$FC,$4B,$56,$2F
	.dc.b	$D9,$5F,$0F,$E4,$64,$76,$7E,$37,$C8,$E2,$3D,$11,$71,$F8,$33,$F2
	.dc.b	$FE,$42,$57,$99,$F9,$DF,$23,$88,$F8,$11,$6B,$E4,$CF,$EE,$3E,$3D
	.dc.b	$F0,$74,$97,$F4,$87,$CF,$BE,$1B,$31,$FC,$B2,$B8,$BF,$C6,$5A,$E6
	.dc.b	$BF,$6F,$18,$C3,$FD,$E1,$6C,$7F,$84,$B6,$FC,$15,$BD,$FE,$56,$3F
	.dc.b	$89,$24,$FD,$4F,$92,$FE,$23,$F8,$63,$D8,$FE,$0A,$D9,$FE,$58,$D7
	.dc.b	$FA,$1B,$5F,$CE,$72,$FE,$67,$E4,$FF,$2D,$EA,$49,$FC,$0C,$7E,$67
	.dc.b	$39,$1F,$84,$7F,$39,$D9,$FC,$8B,$EB,$BF,$9A,$CF,$F3,$B6,$BF,$D4
	.dc.b	$9A,$FF,$BE,$DA,$FE,$1F,$E4,$FF,$4D,$F2,$72,$3F,$23,$1F,$A9,$F5
	.dc.b	$E4,$7E,$51,$FD,$67,$C9,$FF,$32,$F8,$0F,$D2,$4F,$F4,$B6,$BF,$D8
	.dc.b	$9A,$FE,$EC,$38,$7F,$77,$2F,$93,$FE,$DA,$D8,$FD,$6B,$5F,$D6,$7B
	.dc.b	$9F,$DC,$BE,$81,$85,$24,$F0,$F9,$4F,$E8,$FF,$41,$FC,$66,$3F,$B0
	.dc.b	$35,$FE,$EE,$D7,$CC,$D7,$CB,$F2,$BE,$42,$3E,$4F,$95,$F3,$FC,$AC
	.dc.b	$8F,$3B,$EA,$F8,$FF,$EA,$FF,$59,$F9,$B9,$FE,$E0,$D7,$FC,$0E,$D7
	.dc.b	$E3,$35,$F8,$BF,$2B,$F1,$08,$FC,$4F,$07,$E3,$F9,$DD,$BA,$64,$FC
	.dc.b	$88,$AF,$68,$FE,$B7,$7F,$E9,$18,$F1,$C7,$F8,$07,$85,$F2,$9F,$48
	.dc.b	$93,$FB,$5A,$E0,$BC,$37,$83,$FC,$DF,$9D,$C5,$FD,$D9,$B5,$FB,$D6
	.dc.b	$BC,$67,$F9,$F7,$B2,$F8,$8E,$17,$FD,$C2,$F7,$7F,$89,$6B,$DE,$34
	.dc.b	$7F,$8D,$BE,$79,$F5,$9D,$B2,$F2,$D2,$73,$5D,$A3,$C9,$C9,$E7,$23
	.dc.b	$FD,$12,$5F,$36,$F9,$0C,$CF,$F8,$FB,$25,$F9,$C9,$1F,$E4,$0E,$07
	.dc.b	$C9,$FE,$70,$91,$F9,$CF,$5E,$BF,$0C,$FE,$B5,$2F,$95,$FD,$19,$AD
	.dc.b	$7D,$7E,$32,$5F,$2B,$FA,$8F,$95,$98,$7E,$A2,$2B,$D6,$B8,$4E,$2F
	.dc.b	$9D,$F2,$AF,$E6,$5F,$42,$F9,$6F,$F6,$A4,$3E,$49,$BF,$F7,$5A,$D7
	.dc.b	$89,$FE,$2A,$9C,$D7,$F8,$B2,$72,$F1,$AE,$1C,$1F,$DF,$6B,$72,$AD
	.dc.b	$C6,$1B,$5D,$F0,$F1,$CE,$6F,$94,$E7,$0F,$13,$70,$F0,$1E,$3B,$E3
	.dc.b	$1E,$28,$D9,$99,$AF,$8E,$76,$E4,$E6,$79,$EB,$C2,$E6,$FD,$13,$C2
	.dc.b	$39,$57,$4C,$3F,$0F,$16,$7E,$6B,$FC,$96,$3E,$3B,$63,$2F,$AC,$FE
	.dc.b	$55,$F2,$5F,$D9,$79,$3F,$CF,$4D,$8F,$F4,$8B,$5D,$FA,$1F,$A1,$7F
	.dc.b	$D2,$B7,$FC,$CB,$E0,$3F,$D0,$4D,$8C,$F9,$29,$CB,$63,$31,$EB,$8C
	.dc.b	$67,$F6,$BE,$09,$F1,$7F,$B0,$8F,$B5,$3A,$DF,$F1,$0D,$C3,$F0,$CD
	.dc.b	$57,$CB,$BB,$B6,$32,$7E,$8E,$5E,$7C,$7C,$19,$77,$CB,$AD,$EE,$7C
	.dc.b	$2F,$DC,$C7,$A6,$1C,$B2,$F0,$63,$CF,$0E,$33,$F7,$AB,$78,$72,$8F
	.dc.b	$C5,$F2,$7E,$C7,$A7,$F8,$6F,$27,$EA,$9B,$AF,$F5,$CE,$1B,$7F,$BC
	.dc.b	$9B,$7F,$B9,$9B,$C9,$FC,$36,$FF,$AF,$C6,$2B,$3F,$2B,$27,$FB,$1E
	.dc.b	$D7,$1F,$DE,$60,$AC,$65,$FB,$CA,$B7,$FD,$B8,$FE,$47,$7F,$E1,$2B
	.dc.b	$FE,$73,$1F,$C5,$CC,$C6,$DC,$1F,$DC,$8F,$3F,$A7,$E8,$7D,$7D,$9F
	.dc.b	$CE,$5F,$F3,$1A,$F9,$5C,$F8,$1A,$6C,$FF,$B9,$27,$85,$FF,$81,$76
	.dc.b	$FE,$5F,$7E,$28,$FF,$3E,$4E,$9F,$F1,$4B,$C3,$3F,$E4,$6D,$BF,$F4
	.dc.b	$2A,$F9,$5E,$F5,$FC,$86,$1F,$2B,$E8,$3D,$3F,$2B,$6F,$D2,$F8,$72
	.dc.b	$FE,$82,$DA,$FF,$7F,$4F,$A6,$31,$F2,$9C,$A3,$E0,$B7,$0F,$07,$CA
	.dc.b	$FC,$6F,$95,$F8,$ED,$AF,$8F,$E3,$F5,$23,$F1,$5F,$07,$5F,$CE,$3E
	.dc.b	$7F,$CA,$D9,$FF,$34,$7C,$EF,$C5,$C7,$FD,$50,$D7,$FD,$BA,$D4,$E7
	.dc.b	$5F,$EB,$BF,$2B,$ED,$25,$7D,$57,$E0,$7C,$A7,$DE,$3E,$FF,$CB,$59
	.dc.b	$FB,$FF,$2D,$F8,$23,$F0,$7F,$23,$63,$F0,$7F,$2D,$F2,$48,$FE,$05
	.dc.b	$F9,$9E,$07,$E2,$8F,$C5,$FC,$9D,$9F,$C5,$F8,$9F,$8E,$F9,$3F,$C7
	.dc.b	$B6,$3F,$1F,$BF,$D5,$23,$F8,$D7,$ED,$7E,$02,$F9,$0F,$D8,$B6,$7F
	.dc.b	$D7,$2F,$FB,$85,$B1,$AF,$91,$DF,$C8,$D9,$DB,$CB,$E3,$FA,$92,$3F
	.dc.b	$CB,$D8,$7F,$B4,$F0,$FF,$5D,$46,$38,$D0,$F8,$EF,$19,$E4,$F9,$0B
	.dc.b	$5F,$EA,$3F,$9D,$72,$7F,$0A,$F0,$9F,$01,$FE,$F3,$AD,$4E,$63,$19
	.dc.b	$FF,$82,$C5,$F2,$AF,$2C,$62,$BF,$5C,$D8,$CC,$EE,$3D,$6D,$BF,$E1
	.dc.b	$2D,$94,$FF,$90,$EF,$FC,$31,$1E,$43,$E6,$FC,$87,$D0,$3E,$8F,$C8
	.dc.b	$D8,$C9,$2B,$FC,$2D,$A2,$75,$18,$F8,$FF,$B2,$7D,$AF,$91,$B7,$F6
	.dc.b	$A2,$CF,$29,$B9,$FF,$8A,$30,$7C,$0F,$E0,$BD,$4F,$C1,$B7,$1F,$C1
	.dc.b	$B7,$0F,$1B,$6F,$93,$67,$E0,$F0,$FD,$C8,$C5,$7E,$07,$E3,$BF,$1F
	.dc.b	$F1,$FF,$90,$6D,$FE,$41,$C3,$F9,$2F,$8F,$B5,$F0,$BD,$6F,$CB,$9F
	.dc.b	$98,$F8,$EB,$5F,$98,$8B,$3F,$22,$6E,$7F,$CF,$BE,$6B,$FC,$AB,$7F
	.dc.b	$5A,$3F,$AB,$F4,$3C,$6F,$EA,$B8,$2F,$6F,$F9,$4E,$95,$BB,$FF,$7B
	.dc.b	$56,$AB,$FD,$78,$D7,$F0,$67,$3F,$77,$F1,$B1,$98,$FC,$3B,$DA,$FC
	.dc.b	$D3,$F9,$DB,$FE,$EB,$18,$BF,$DA,$FA,$9F,$B8,$3F,$73,$F1,$B6,$32
	.dc.b	$4C,$7F,$13,$19,$7F,$CC,$1F,$88,$7F,$A5,$27,$F7,$FE,$FF,$F7,$28
	.dc.b	$C4,$FE,$42,$31,$F0,$B2,$BE,$41,$F7,$AB,$FD,$AC,$E2,$9C,$89,$E7
	.dc.b	$97,$F6,$EF,$A8,$7C,$5F,$7D,$BE,$7C,$63,$1E,$58,$71,$9C,$D7,$ED
	.dc.b	$2B,$58,$FF,$0C,$6B,$FE,$EB,$6B,$F4,$1D,$E1,$E1,$9F,$77,$C7,$74
	.dc.b	$C7,$BE,$5A,$7D,$9D,$DF,$0E,$BE,$23,$63,$71,$F1,$FE,$3B,$19,$7C
	.dc.b	$CF,$8C,$F9,$5F,$1D,$F2,$DB,$7D,$9C,$3E,$0F,$8E,$B5,$F5,$3E,$2B
	.dc.b	$FB,$76,$7F,$C2,$23,$FC,$04,$D7,$F1,$7C,$1F,$36,$76,$7D,$93,$ED
	.dc.b	$7C,$5D,$C3,$DB,$F1,$7F,$74,$FB,$DF,$19,$63,$24,$C3,$F0,$B1,$97
	.dc.b	$F0,$FF,$0D,$E4,$24,$FF,$39,$CB,$FC,$6D,$3E,$69,$FC,$51,$F8,$BF
	.dc.b	$8C,$B5,$F8,$59,$FF,$4F,$1F,$EB,$7F,$5B,$F4,$BC,$8F,$03,$FD,$AB
	.dc.b	$D6,$9F,$E9,$3F,$13,$FC,$B2,$31,$7F,$B9,$AD,$D6,$A1,$F2,$DA,$7F
	.dc.b	$DE,$93,$FE,$E0,$8C,$47,$FA,$D4,$DC,$F8,$A7,$4F,$39,$E5,$87,$97
	.dc.b	$1F,$E7,$B1,$CA,$FF,$D8,$47,$23,$C9,$1A,$AE,$47,$57,$11,$FC,$10
	.dc.b	$E3,$3D,$AF,$C3,$19,$C7,$E0,$5F,$18,$F9,$06,$4A,$7B,$AF,$6E,$6F
	.dc.b	$A4,$78,$23,$E8,$C6,$1F,$41,$89,$ED,$1E,$DA,$E1,$87,$C1,$AD,$9F
	.dc.b	$27,$1E,$46,$DF,$1C,$6E,$31,$5D,$CB,$7D,$8F,$8B,$5F,$20,$EA,$7C
	.dc.b	$47,$4F,$BE,$7B,$C7,$DF,$6C,$7B,$E3,$6F,$AC,$E9,$3E,$06,$C7,$81
	.dc.b	$F9,$07,$82,$FD,$F8,$FD,$28,$E4,$6A,$BD,$D7,$E0,$8F,$CD,$B5,$FC
	.dc.b	$64,$7B,$EF,$B6,$1F,$26,$38,$C7,$DB,$AB,$5F,$05,$AF,$91,$B5,$F2
	.dc.b	$E7,$F3,$DD,$89,$A4,$AF,$C2,$46,$5F,$F0,$9F,$C6,$7F,$7E,$BF,$F0
	.dc.b	$AC,$E6,$53,$95,$19,$8C,$37,$C6,$57,$2B,$FE,$BB,$5C,$5F,$D8,$60
	.dc.b	$F5,$7F,$B0,$C6,$1B,$1C,$2F,$9E,$3F,$9A,$BE,$53,$FD,$A0,$E5,$1F
	.dc.b	$9C,$3F,$82,$57,$27,$DB,$5C,$AF,$F8,$05,$FA,$F2,$F0,$57,$E8,$70
	.dc.b	$E6,$FE,$82,$75,$1E,$3C,$7B,$3A,$BE,$35,$E5,$C3,$B4,$EE,$3F,$52
	.dc.b	$F4,$C3,$67,$8A,$7D,$AD,$BF,$16,$3F,$E0,$48,$DD,$73,$75,$7D,$27
	.dc.b	$53,$F0,$DB,$1F,$02,$BF,$79,$8F,$C8,$8F,$F2,$8D,$7A,$4E,$0F,$C9
	.dc.b	$8D,$3F,$80,$C3,$CB,$3A,$8F,$01,$D2,$32,$F7,$32,$4C,$3C,$E7,$A3
	.dc.b	$1F,$8E,$4A,$F8,$CE,$9E,$D1,$F0,$27,$A3,$97,$D3,$8F,$B9,$FB,$E7
	.dc.b	$8F,$18,$FF,$20,$AF,$73,$C0,$F7,$61,$F5,$9E,$B1,$E0,$23,$82,$1D
	.dc.b	$8F,$63,$F4,$C8,$F5,$32,$70,$C7,$E5,$C7,$8A,$39,$63,$F0,$E3,$FC
	.dc.b	$4D,$F1,$3F,$D7,$B6,$B5,$FE,$10,$8C,$3F,$E2,$8E,$17,$C5,$E3,$9F
	.dc.b	$FC,$55,$5F,$E9,$66,$D7,$F7,$18,$C5,$1B,$F1,$5C,$60,$D3,$1F,$17
	.dc.b	$6B,$FC,$43,$15,$FF,$22,$C7,$87,$D9,$7C,$5D,$24,$FF,$8A,$33,$AF
	.dc.b	$F5,$BC,$95,$C2,$3F,$D7,$CD,$74,$D7,$FC,$36,$BE,$27,$C0,$48,$FF
	.dc.b	$21,$C3,$E2,$B1,$1F,$C4,$AF,$FD,$1F,$2F,$8B,$F4,$1A,$E3,$4E,$BE
	.dc.b	$23,$D8,$4B,$EB,$F1,$55,$E8,$7F,$96,$B0,$7C,$57,$C3,$35,$87,$49
	.dc.b	$FF,$5F,$FD,$43,$0F,$B7,$08,$C3,$FE,$21,$6B,$DB,$5F,$8F,$F1,$7C
	.dc.b	$27,$33,$FE,$05,$6B,$FE,$8A,$6B,$E0,$AF,$F6,$CB,$DD,$FB,$1F,$17
	.dc.b	$23,$C1,$1F,$6F,$E2,$64,$7C,$C7,$EF,$7A,$78,$8F,$26,$8F,$8B,$FC
	.dc.b	$11,$9D,$3E,$06,$9F,$7B,$E2,$BF,$0C,$4B,$FA,$FF,$03,$5E,$71,$6B
	.dc.b	$8E,$1F,$EF,$A9,$FF,$22,$7C,$DF,$76,$E3,$F2,$A3,$11,$FE,$F9,$F0
	.dc.b	$DD,$19,$AF,$83,$5A,$AF,$F1,$63,$5F,$F5,$8B,$5F,$F7,$AB,$D3,$0F
	.dc.b	$B5,$1F,$E5,$1E,$BF,$B3,$87,$EA,$E3,$C6,$7E,$4B,$B7,$F4,$84,$7F
	.dc.b	$2B,$F1,$75,$F9,$31,$6F,$DF,$3F,$AB,$F8,$A9,$5F,$9C,$7F,$5D,$F1
	.dc.b	$78,$8F,$A3,$F1,$7F,$B1,$31,$7F,$23,$6F,$DA,$78,$B2,$BF,$4B,$1F
	.dc.b	$B7,$30,$FD,$A8,$C3,$A3,$F5,$6E,$A7,$A4,$7F,$AC,$3E,$4D,$F3,$3F
	.dc.b	$C4,$5B,$19,$8F,$E0,$8E,$58,$E4,$FD,$B9,$C7,$E2,$E3,$D5,$E9,$F1
	.dc.b	$B8,$8F,$D1,$52,$71,$7F,$8E,$85,$7E,$B3,$E0,$9F,$2B,$49,$87,$FA
	.dc.b	$37,$E6,$13,$88,$FF,$37,$18,$9F,$D3,$C7,$28,$C9,$92,$B9,$7F,$81
	.dc.b	$89,$FF,$4B,$3D,$F8,$BB,$63,$E7,$F8,$1F,$90,$49,$F9,$9F,$1F,$F2
	.dc.b	$FE,$3B,$F9,$D6,$86,$92,$7C,$FD,$89,$7C,$67,$D3,$F8,$F9,$5F,$5E
	.dc.b	$3E,$B7,$C6,$48,$FB,$4F,$D9,$F8,$EF,$B5,$F1,$DE,$02,$4F,$6C,$11
	.dc.b	$E4,$F5,$3F,$01,$F1,$B2,$BF,$09,$1F,$83,$E0,$91,$F8,$77,$F0,$DF
	.dc.b	$1B,$F8,$7E,$3F,$69,$27,$B3,$F0,$3E,$37,$F1,$84,$8F,$C8,$7C,$6F
	.dc.b	$F6,$2E,$47,$C2,$BE,$13,$0D,$A1,$FC,$DA,$77,$3A,$F8,$5F,$CC,$7C
	.dc.b	$67,$CE,$25,$7B,$DF,$E4,$FF,$2C,$F8,$5A,$49,$FC,$1C,$65,$F8,$DF
	.dc.b	$01,$FA,$32,$3F,$A7,$F7,$BF,$BB,$7C,$0F,$E1,$E3,$FD,$EA,$D7,$CF
	.dc.b	$D4,$8C,$42,$31,$DF,$FD,$F6,$5F,$19,$D3,$E3,$65,$FC,$0D,$FF,$98
	.dc.b	$D6,$A3,$1E,$D7,$68,$7C,$67,$93,$E3,$25,$68,$F0,$4E,$9C,$67,$E5
	.dc.b	$9F,$11,$23,$F2,$72,$8F,$37,$1F,$C0,$24,$F2,$9D,$37,$E2,$FE,$29
	.dc.b	$1D,$7C,$65,$7F,$C1,$12,$9F,$84,$48,$F0,$BF,$3B,$92,$47,$53,$C1
	.dc.b	$C7,$F4,$88,$F8,$F9,$6B,$DA,$3E,$71,$1F,$6F,$C5,$7D,$A2,$3E,$BF
	.dc.b	$8C,$AF,$F7,$7C,$53,$8E,$0F,$A7,$5F,$82,$F8,$CF,$C1,$FC,$5C,$C3
	.dc.b	$E2,$45,$79,$45,$AF,$B7,$3F,$E0,$8F,$93,$FF,$20,$F4,$7E,$7B,$27
	.dc.b	$67,$E8,$31,$56,$9F,$C6,$46,$63,$13,$C2,$7F,$4F,$1F,$87,$76,$63
	.dc.b	$E2,$BF,$D7,$EE,$0E,$AE,$27,$94,$62,$FC,$97,$DF,$A3,$C9,$5A,$C3
	.dc.b	$C3,$3E,$3C,$39,$36,$F8,$E1,$E4,$C3,$8C,$E6,$3C,$57,$C9,$F3,$57
	.dc.b	$07,$E0,$D7,$07,$27,$0B,$EE,$E3,$0F,$24,$EE,$3D,$47,$F4,$4B,$F9
	.dc.b	$4F,$CD,$32,$57,$2F,$FB,$7B,$C2,$7C,$A7,$F4,$8F,$8C,$46,$27,$FA
	.dc.b	$7C,$62,$75,$39,$9E,$9E,$CF,$F9,$99,$F1,$35,$CF,$A9,$FD,$53,$0F
	.dc.b	$F3,$E3,$67,$86,$5F,$3D,$BA,$FE,$17,$D7,$FF,$1C,$76,$FC,$7C,$38
	.dc.b	$E3,$F3,$2F,$6F,$E1,$1A,$AB,$8B,$BC,$79,$4E,$AB,$9E,$3F,$84,$BD
	.dc.b	$BE,$3A,$D9,$2F,$F1,$93,$AC,$7F,$2D,$87,$32,$57,$DD,$9D,$57,$96
	.dc.b	$3F,$0F,$D9,$FC,$A7,$7F,$F3,$87,$17,$F2,$2F,$8E,$BA,$FF,$21,$9B
	.dc.b	$87,$F1,$E9,$B7,$C6,$BF,$8B,$46,$34,$FF,$57,$C6,$0F,$E4,$7F,$AA
	.dc.b	$7F,$4E,$7F,$8E,$69,$CA,$31,$87,$F0,$38,$C3,$27,$F9,$5E,$0F,$84
	.dc.b	$C4,$FF,$09,$F8,$DF,$F2,$7B,$FC,$DD,$AF,$F3,$46,$CF,$F4,$76,$E3
	.dc.b	$FE,$BA,$BF,$E5,$D1,$E9,$F8,$0F,$53,$EE,$F9,$3F,$F5,$BC,$FF,$33
	.dc.b	$31,$F1,$DE,$F6,$9F,$E0,$E8,$DE,$3F,$EB,$96,$BF,$C3,$F1,$46,$23
	.dc.b	$FA,$DB,$C3,$7F,$F4,$1B,$6B,$FA,$B7,$D2,$3C,$6E,$F7,$FE,$9F,$AF
	.dc.b	$F5,$D5,$7F,$50,$6C,$67,$E3,$7D,$5F,$1D,$EB,$7E,$1C,$FF,$AF,$6B
	.dc.b	$D1,$18,$3E,$4E,$2F,$B8,$7C,$4F,$1A,$FE,$C1,$39,$8C,$63,$E3,$F8
	.dc.b	$FF,$90,$7C,$9F,$90,$B7,$C3,$5F,$ED,$6F,$F9,$2B,$04,$62,$FF,$84
	.dc.b	$46,$3C,$CF,$1C,$E6,$7F,$86,$7C,$B7,$A5,$FF,$29,$B5,$CE,$9F,$82
	.dc.b	$51,$97,$E7,$FC,$77,$DD,$3E,$F7,$C8,$DC,$BE,$57,$C8,$FC,$62,$3F
	.dc.b	$75,$F9,$3D,$1F,$85,$3F,$47,$FA,$AF,$F9,$BE,$7E,$94,$7E,$27,$E3
	.dc.b	$A4,$7D,$5E,$1B,$97,$D9,$F9,$1F,$A6,$49,$FC,$53,$F5,$BE,$47,$F2
	.dc.b	$47,$6F,$84,$B3,$F7,$FE,$3B,$F2,$DF,$23,$F9,$76,$C7,$E0,$FE,$43
	.dc.b	$EF,$91,$FC,$B3,$F8,$2F,$90,$FC,$F1,$EA,$F8,$5B,$3F,$8B,$F2,$3F
	.dc.b	$45,$F2,$1F,$A3,$6C,$7E,$3F,$E4,$3F,$16,$47,$F4,$4F,$E3,$BC,$AF
	.dc.b	$D5,$1F,$1B,$CD,$B3,$FE,$26,$3F,$D4,$A7,$E6,$78,$3F,$86,$B8,$7F
	.dc.b	$3C,$63,$E4,$3F,$77,$19,$8F,$D6,$35,$FF,$22,$B5,$FF,$22,$3C,$FB
	.dc.b	$7F,$DD,$ED,$3F,$C9,$7C,$CF,$86,$ED,$1E,$5F,$8F,$F1,$9F,$7B,$E1
	.dc.b	$AC,$64,$98,$30,$FF,$7D,$E0,$AD,$CE,$6B,$CB,$F1,$BE,$EF,$90,$F7
	.dc.b	$B6,$3E,$23,$87,$FC,$97,$6F,$89,$C5,$F3,$3D,$BC,$FF,$23,$E3,$FE
	.dc.b	$4B,$6F,$E7,$3F,$20,$8F,$FA,$90,$FF,$54,$15,$E2,$F8,$3E,$2A,$D7
	.dc.b	$7F,$31,$2F,$8F,$FA,$E7,$E8,$BE,$2E,$C7,$A1,$B5,$C7,$34,$63,$7F
	.dc.b	$A7,$D1,$F8,$0F,$8F,$FC,$0B,$70,$FC,$33,$70,$F1,$42,$31,$5E,$49
	.dc.b	$FF,$4D,$FA,$AF,$17,$FA,$CC,$3E,$1B,$C3,$87,$F9,$E6,$15,$B8,$C4
	.dc.b	$EA,$7F,$DA,$F3,$97,$FD,$C2,$D7,$FC,$3C,$D7,$FC,$98,$D7,$C7,$5E
	.dc.b	$28,$E5,$7E,$79,$F0,$F8,$59,$AF,$ED,$F3,$FC,$93,$D4,$FD,$1F,$C7
	.dc.b	$E3,$CC,$46,$33,$FE,$58,$E4,$E3,$F6,$EE,$EC,$4F,$EB,$FE,$3F,$F5
	.dc.b	$A7,$6F,$5B,$F1,$91,$93,$F9,$EB,$8D,$7E,$7C,$EF,$A7,$8D,$7C,$9F
	.dc.b	$96,$7C,$4F,$CE,$FD,$7B,$FA,$8E,$FE,$D9,$CD,$7C,$F9,$CB,$FE,$71
	.dc.b	$8C,$E3,$FE,$47,$0E,$1F,$F2,$53,$5E,$17,$FC,$F6,$38,$DF,$78,$FB
	.dc.b	$FC,$FF,$EF,$2B,$F3,$47,$F0,$CF,$0E,$47,$8F,$C6,$B8,$FE,$AC,$64
	.dc.b	$8C,$E3,$F5,$E3,$15,$FC,$66,$FF,$8B,$F8,$72,$3E,$4F,$55,$C7,$FD
	.dc.b	$D9,$BF,$E2,$F1,$FE,$23,$7F,$46,$B5,$1F,$C5,$36,$FE,$3A,$4C,$3F
	.dc.b	$6C,$DC,$3F,$42,$FF,$26,$F0,$BF,$93,$BF,$E3,$4F,$84,$7C,$77,$F2
	.dc.b	$AF,$03,$0D,$C3,$FD,$AB,$3A,$6C,$7F,$8F,$3E,$0D,$F1,$7C,$27,$F9
	.dc.b	$B7,$C1,$FF,$30,$23,$FC,$25,$B9,$FE,$D1,$FE,$73,$AF,$F3,$A7,$F6
	.dc.b	$31,$FE,$B1,$1F,$1E,$F3,$BF,$D6,$5B,$FE,$3E,$31,$3F,$E8,$9F,$81
	.dc.b	$3C,$2B,$7B,$FE,$FE,$B7,$3F,$EA,$81,$F2,$F8,$C3,$F3,$4E,$A3,$C4
	.dc.b	$E6,$BF,$84,$39,$86,$1E,$08,$D6,$77,$1F,$CB,$4D,$AF,$2E,$1D,$A7
	.dc.b	$31,$E7,$33,$87,$9E,$33,$8F,$49,$D1,$E9,$70,$78,$EF,$E0,$B8,$9E
	.dc.b	$B8,$F1,$35,$5F,$02,$3E,$2E,$3E,$58,$F8,$93,$B9,$E2,$6A,$38,$1F
	.dc.b	$3C,$F1,$E3,$F4,$E3,$CE,$6E,$FB,$D7,$3A,$E0,$D7,$BC,$F5,$8D,$B4
	.dc.b	$E7,$1F,$12,$FD,$B1,$88,$FE,$D5,$8F,$C7,$7A,$C7,$16,$9C,$4F,$49
	.dc.b	$E0,$31,$5E,$06,$C7,$C0,$69,$C8,$EF,$5F,$1D,$FC,$B5,$FC,$F7,$E2
	.dc.b	$35,$EC,$70,$3C,$27,$E3,$27,$CC,$FD,$B7,$E4,$1E,$DC,$3E,$64,$74
	.dc.b	$8C,$B4,$F6,$4F,$DC,$DF,$F4,$F1,$F5,$5A,$FE,$7D,$AF,$EB,$5C,$FC
	.dc.b	$3F,$F8,$6E,$FF,$53,$18,$E3,$FF,$0B,$52,$31,$3A,$BF,$D7,$CA,$F9
	.dc.b	$3B,$8C,$C6,$1F,$D8,$E0,$F3,$F8,$C6,$1B,$1C,$31,$F0,$CF,$EE,$2B
	.dc.b	$94,$F8,$0E,$57,$F8,$43,$F8,$0B,$87,$F4,$F1,$C1,$FE,$0A,$F9,$B3
	.dc.b	$F0,$47,$E8,$67,$84,$77,$BF,$82,$DC,$7B,$57,$F1,$3C,$7C,$EF,$81
	.dc.b	$AF,$C3,$AF,$4D,$FA,$EB,$B5,$7C,$8B,$F0,$DF,$E5,$30,$F2,$5E,$E3
	.dc.b	$F2,$EF,$90,$F8,$15,$C5,$D5,$79,$67,$73,$97,$C8,$7E,$F6,$BA,$BF
	.dc.b	$0B,$0F,$2B,$A3,$F8,$6D,$FC,$9C,$7A,$1D,$70,$F3,$92,$3C,$24,$9E
	.dc.b	$51,$97,$B6,$3E,$1B,$EF,$1F,$24,$F9,$75,$DC,$F8,$17,$DD,$F7,$C7
	.dc.b	$C8,$39,$54,$7D,$E6,$4F,$95,$5F,$62,$79,$9C,$A3,$DB,$8F,$D6,$C7
	.dc.b	$F0,$91,$F5,$EF,$C7,$3F,$6A,$BC,$93,$EA,$3F,$17,$7D,$5F,$7C,$75
	.dc.b	$C7,$D1,$1B,$7E,$13,$EF,$3C,$B7,$E9,$C7,$F0,$24,$AF,$3C,$FF,$A4
	.dc.b	$A3,$FA,$FE,$7F,$E9,$66,$C6,$4F,$F5,$27,$C9,$63,$CA,$FF,$37,$CF
	.dc.b	$F8,$96,$74,$7E,$D8,$A7,$FA,$89,$D4,$7E,$F7,$DD,$E5,$B1,$A7,$5E
	.dc.b	$47,$C2,$E3,$0F,$F5,$A5,$78,$72,$FE,$0D,$1F,$E0,$8B,$F2,$7C,$C3
	.dc.b	$E1,$69,$27,$FB,$1D,$7F,$78,$C9,$5F,$E1,$49,$C8,$8C,$D7,$A2,$7A
	.dc.b	$4F,$F9,$0E,$72,$D8,$CB,$F0,$76,$7E,$AF,$D6,$3F,$D2,$D2,$F8,$BF
	.dc.b	$39,$F3,$AB,$7D,$58,$AF,$1E,$1E,$6A,$C4,$7F,$AB,$6B,$6F,$FA,$1F
	.dc.b	$17,$5F,$C2,$D2,$4E,$62,$BE,$28,$B5,$FE,$7F,$AE,$33,$FE,$91,$FC
	.dc.b	$71,$F0,$1F,$82,$84,$61,$D4,$67,$E2,$9E,$C6,$EB,$55,$FE,$69,$6B
	.dc.b	$F8,$56,$BF,$CF,$1A,$FF,$8B,$1E,$38,$7F,$A4,$FE,$53,$AC,$FF,$A9
	.dc.b	$5A,$FF,$99,$67,$7C,$3F,$E2,$0E,$1F,$9D,$3F,$2B,$85,$7F,$86,$3E
	.dc.b	$84,$EA,$7E,$AC,$65,$E0,$EB,$7F,$AD,$F7,$0F,$82,$A4,$AF,$C4,$E7
	.dc.b	$FE,$5E,$E5,$C4,$7E,$03,$E5,$7F,$2A,$62,$7F,$D5,$DF,$8B,$7F,$A5
	.dc.b	$A7,$FD,$80,$D7,$FD,$00,$D8,$DF,$17,$D3,$8D,$4F,$EB,$DA,$FF,$66
	.dc.b	$6B,$FE,$56,$6A,$7A,$6D,$F4,$7E,$27,$EC,$1D,$7C,$BF,$D5,$9F,$AC
	.dc.b	$F5,$E5,$FC,$B8,$CB,$F3,$63,$2F,$E7,$A3,$F7,$A7,$ED,$3E,$5A,$47
	.dc.b	$F9,$FA,$2B,$FA,$48,$B9,$7F,$B9,$FA,$7F,$C0,$55,$FB,$CF,$96,$97
	.dc.b	$F6,$DF,$E1,$3E,$4E,$23,$F8,$85,$7E,$F2,$7F,$70,$48,$FD,$B3,$FC
	.dc.b	$1A,$64,$73,$3F,$0D,$F2,$DF,$C2,$88,$FF,$A5,$26,$BF,$BA,$F9,$6F
	.dc.b	$E1,$E6,$31,$F4,$73,$BE,$57,$F8,$99,$2B,$FD,$3B,$55,$FE,$07,$16
	.dc.b	$BE,$07,$80,$AF,$E3,$7D,$92,$FF,$34,$FF,$28,$CB,$11,$F0,$31,$4F
	.dc.b	$F1,$72,$47,$F1,$57,$F9,$2F,$CB,$48,$E6,$7E,$97,$E5,$7F,$94,$91
	.dc.b	$FF,$59,$E7,$FC,$B3,$9F,$F9,$11,$23,$F9,$03,$FC,$97,$E5,$7F,$98
	.dc.b	$FB,$52,$7F,$61,$1F,$CD,$39,$A4,$7E,$CD,$FE,$83,$87,$F4,$BF,$7D
	.dc.b	$E9,$FF,$46,$6D,$7F,$2F,$6B,$F7,$9A,$FF,$91,$39,$10,$F9,$1E,$9F
	.dc.b	$2B,$2B,$F9,$5C,$E5,$BD,$5E,$12,$3F,$ED,$2C,$FC,$9F,$27,$E3,$24
	.dc.b	$78,$DF,$37,$C9,$7A,$49,$1D,$BE,$0B,$D9,$E4,$48,$F0,$BF,$DA,$74
	.dc.b	$67,$F0,$7E,$4B,$E0,$12,$BE,$03,$F0,$FC,$3F,$8C,$48,$F5,$FC,$97
	.dc.b	$F3,$5F,$97,$77,$7F,$73,$94,$61,$FE,$EC,$D7,$FB,$B1,$A8,$FE,$F5
	.dc.b	$DD,$F4,$BE,$32,$F2,$1A,$4A,$E3,$2F,$CF,$9E,$DF,$F5,$B6,$CF,$F2
	.dc.b	$46,$BF,$D6,$5F,$E9,$EA,$5D,$1F,$E7,$EF,$94,$FF,$73,$3F,$DD,$F9
	.dc.b	$DF,$13,$F8,$73,$FA,$A4,$EA,$FF,$BC,$4E,$9F,$C4,$4F,$E2,$C8,$FE
	.dc.b	$2E,$7F,$20,$47,$F0,$B1,$F9,$3F,$93,$91,$F8,$87,$FB,$B6,$8F,$8E
	.dc.b	$CC,$FE,$4E,$7F,$30,$47,$F3,$0F,$E7,$7E,$52,$47,$33,$F2,$7F,$29
	.dc.b	$FA,$12,$3F,$96,$F6,$EB,$F9,$F8,$FD,$21,$1F,$F0,$97,$3B,$AF,$F1
	.dc.b	$84,$8F,$EA,$7F,$88,$7D,$0E,$0F,$E5,$95,$B9,$FD,$1F,$C3,$FE,$C4
	.dc.b	$E0,$FD,$3E,$1F,$A1,$F0,$1F,$B5,$F9,$59,$3F,$AA,$8F,$DC,$18,$78
	.dc.b	$1F,$4A,$32,$FF,$8B,$07,$CC,$BC,$4F,$F1,$CF,$C2,$FE,$12,$31,$E7
	.dc.b	$29,$39,$AD,$F4,$FF,$B0,$AF,$FB,$C4,$E6,$F8,$36,$F8,$D6,$EF,$8C
	.dc.b	$75,$BE,$33,$9B,$E2,$D9,$E2,$6E,$B8,$BE,$AF,$5F,$FD,$A7,$87,$72
	.dc.b	$57,$17,$D9,$1A,$8C,$12,$7A,$5F,$81,$F2,$4E,$DC,$9F,$B6,$C3,$E2
	.dc.b	$B7,$0F,$F3,$0B,$FE,$6F,$8C,$BF,$E7,$1F,$84,$7C,$43,$E0,$7F,$CE
	.dc.b	$D1,$86,$C6,$67,$F8,$1C,$E5,$F0,$79,$8F,$8D,$FF,$3F,$61,$FE,$74
	.dc.b	$6C,$66,$77,$18,$3D,$8F,$E0,$23,$5A,$FF,$02,$DB,$FC,$55,$C1,$AC
	.dc.b	$3A,$5F,$D0,$38,$61,$9C,$38,$EF,$F8,$59,$61,$C9,$E1,$7C,$B1,$FC
	.dc.b	$65,$F0,$9F,$C0,$35,$E3,$E2,$7F,$10,$78,$C6,$6B,$96,$1F,$8F,$8E
	.dc.b	$18,$77,$C3,$AE,$3E,$52,$61,$E9,$9D,$5F,$19,$FB,$DC,$7F,$D5,$9F
	.dc.b	$07,$27,$F8,$7F,$BF,$E7,$6B,$FE,$80,$A7,$93,$A8,$FE,$45,$77,$2F
	.dc.b	$F4,$B5,$6A,$7F,$AE,$CE,$63,$FD,$35,$56,$FA,$DF,$ED,$A7,$FD,$3D
	.dc.b	$E3,$66,$38,$42,$31,$87,$8D,$C9,$AD,$EC,$7E,$F7,$CD,$7C,$8F,$CB
	.dc.b	$C3,$FD,$85,$18,$9F,$F6,$33,$63,$FA,$8B,$5D,$CF,$B5,$46,$F4,$FE
	.dc.b	$33,$E1,$E6,$31,$7F,$83,$5F,$23,$F1,$69,$F2,$3F,$EC,$3A,$F3,$4E
	.dc.b	$63,$18,$7F,$18,$3B,$C7,$A5,$E8,$D7,$F8,$A5,$F6,$AF,$5C,$E6,$78
	.dc.b	$C7,$63,$FD,$07,$F0,$1F,$E3,$47,$11,$FE,$B9,$F9,$87,$99,$E7,$6C
	.dc.b	$FF,$27,$FB,$97,$B5,$9A,$F8,$91,$89,$FF,$78,$36,$FF,$BD,$B5,$FF
	.dc.b	$79,$35,$FE,$59,$A3,$F2,$4F,$EF,$95,$DF,$0F,$91,$18,$AF,$89,$7D
	.dc.b	$6F,$F2,$FA,$FF,$BD,$F0,$FF,$4F,$B6,$BF,$DA,$9F,$91,$3E,$5F,$E7
	.dc.b	$7C,$47,$DB,$F9,$BF,$F1,$A6,$1F,$E1,$0B,$46,$1F,$F0,$9B,$5F,$DF
	.dc.b	$1A,$F6,$7E,$96,$4F,$C9,$7C,$0F,$EA,$9B,$7F,$EC,$1E,$FF,$C1,$46
	.dc.b	$2F,$FD,$F9,$F2,$BF,$ED,$6A,$FF,$76,$B6,$3F,$48,$BE,$73,$F5,$4E
	.dc.b	$70,$FC,$E7,$D1,$7D,$BA,$FF,$14,$B5,$F3,$B5,$FE,$2C,$D7,$C3,$F3
	.dc.b	$BD,$0F,$D7,$F8,$17,$1F,$1F,$CE,$F8,$0F,$0F,$B1,$63,$FB,$56,$1F
	.dc.b	$DA,$5F,$94,$DC,$3C,$7E,$57,$E4,$E3,$11,$FE,$65,$CF,$FC,$31,$5F
	.dc.b	$03,$D5,$FE,$0D,$A3,$4F,$7C,$7C,$0F,$9D,$F8,$3C,$96,$FE,$3B,$FB
	.dc.b	$5F,$23,$FD,$2D,$1F,$6D,$B9,$7F,$93,$1A,$FD,$A6,$BF,$DF,$7C,$46
	.dc.b	$7F,$BA,$8F,$AB,$E1,$79,$31,$FA,$2F,$D2,$F9,$CF,$A7,$E0,$DC,$BE
	.dc.b	$CB,$E3,$F8,$6F,$D9,$66,$69,$27,$FC,$BF,$9F,$83,$CD,$FB,$A7,$C2
	.dc.b	$2C,$7F,$85,$FC,$8F,$F3,$AB,$F8,$6F,$97,$FE,$29,$18,$D7,$F0,$F1
	.dc.b	$8F,$7F,$F8,$14,$E6,$7F,$32,$D7,$FA,$C3,$5F,$F0,$F3,$CF,$8F,$F9
	.dc.b	$43,$6F,$E3,$43,$E6,$BB,$47,$E6,$1F,$CA,$9F,$6B,$E5,$2C,$64,$97
	.dc.b	$FC,$FE,$93,$A8,$FC,$FF,$CC,$FE,$78,$FC,$1F,$9F,$6F,$F4,$B1,$67
	.dc.b	$94,$DB,$FF,$7B,$72,$74,$BF,$D5,$E1,$F8,$CF,$98,$FD,$4B,$23,$27
	.dc.b	$F1,$23,$DB,$19,$C3,$F6,$33,$6F,$FB,$A6,$FF,$E8,$5A,$C5,$FE,$27
	.dc.b	$0F,$F4,$7B,$FB,$AF,$9A,$FD,$BE,$5F,$B8,$6C,$7E,$FA,$3F,$3F,$EB
	.dc.b	$E3,$3F,$DF,$FC,$CF,$EF,$8F,$DF,$E3,$6F,$F8,$2C,$59,$FE,$0B,$37
	.dc.b	$3F,$AF,$ED,$7E,$13,$E0,$BF,$0A,$DC,$3F,$14,$FE,$10,$8F,$F2,$1E
	.dc.b	$14,$63,$0F,$EA,$51,$89,$FF,$34,$6A,$EA,$EB,$1F,$26,$BF,$D5,$1F
	.dc.b	$3B,$FD,$F3,$2F,$F0,$5B,$67,$F9,$FB,$FE,$6B,$78,$42,$35,$3C,$BC
	.dc.b	$3F,$ED,$11,$89,$D4,$E2,$7F,$B6,$7E,$BF,$FD,$16,$7F,$C3,$3F,$08
	.dc.b	$E5,$CF,$89,$FD,$C6,$31,$97,$82,$F8,$3F,$D4,$9B,$8F,$FB,$05,$78
	.dc.b	$BC,$A7,$FD,$11,$C7,$FD,$DB,$2F,$F0,$66,$CE,$FF,$F0,$34,$E6,$BF
	.dc.b	$C8,$93,$97,$8F,$CA,$FF,$75,$9F,$EF,$F9,$7F,$4C,$6B,$FE,$4B,$6B
	.dc.b	$FE,$91,$7A,$36,$3F,$C2,$9F,$3C,$F9,$4F,$0C,$7F,$85,$FF,$5A,$FE
	.dc.b	$CD,$87,$F9,$57,$E7,$FE,$D6,$1F,$E0,$D9,$CB,$FE,$9D,$8C,$E9,$FE
	.dc.b	$27,$9C,$C9,$A7,$F2,$C8,$D7,$AB,$F4,$BC,$5F,$9D,$DB,$71,$FB,$BC
	.dc.b	$48,$C6,$3F,$EB,$4C,$1C,$5F,$E1,$98,$C3,$F1,$DB,$87,$F6,$0D,$FF
	.dc.b	$C4,$1D,$DF,$CD,$9B,$C2,$69,$33,$FE,$D6,$FF,$94,$7E,$07,$E3,$46
	.dc.b	$0F,$CC,$57,$FA,$4F,$5F,$F2,$B5,$3E,$7B,$84,$61,$B5,$C2,$FE,$4D
	.dc.b	$FF,$A6,$FE,$5B,$39,$FF,$96,$F1,$FB,$33,$E5,$E2,$F4,$C6,$08,$F8
	.dc.b	$0F,$37,$C9,$DC,$7F,$DA,$D7,$FE,$CA,$1E,$0B,$E2,$BF,$D2,$1C,$3F
	.dc.b	$9C,$9C,$C7,$F5,$F6,$9F,$E2,$EE,$BF,$F4,$65,$71,$8F,$EC,$43,$E7
	.dc.b	$F0,$AF,$A4,$62,$3C,$4E,$A3,$37,$DE,$31,$7F,$E9,$97,$85,$7F,$BA
	.dc.b	$0E,$D3,$E8,$9D,$3C,$4C,$61,$FE,$EF,$AE,$3E,$5D,$9C,$DF,$23,$D3
	.dc.b	$8F,$82,$FF,$DE,$CF,$08,$F8,$27,$AF,$0F,$83,$5D,$E3,$E2,$57,$57
	.dc.b	$E1,$E1,$CD,$EC,$7F,$BB,$0B,$8F,$C0,$77,$5E,$49,$F0,$61,$C4,$B1
	.dc.b	$E8,$7C,$91,$D4,$B1,$C4,$F2,$4F,$9E,$FC,$8F,$86,$78,$4F,$DD,$AE
	.dc.b	$CD,$3B,$1C,$AF,$CD,$5C,$63,$B3,$E8,$3E,$09,$EE,$3E,$EB,$A7,$81
	.dc.b	$AA,$F3,$B6,$3C,$0F,$2A,$F7,$B6,$3C,$58,$7D,$E8,$F0,$1E,$D3,$C0
	.dc.b	$7E,$16,$7D,$6F,$DC,$9F,$5B,$4E,$85,$39,$47,$DF,$D3,$F4,$B1,$F9
	.dc.b	$C6,$BF,$65,$AF,$EA,$DC,$FC,$33,$CA,$FD,$04,$FF,$0B,$AF,$F7,$E4
	.dc.b	$E7,$2F,$E6,$51,$8A,$DD,$F2,$6C,$66,$7F,$27,$83,$08,$BE,$23,$11
	.dc.b	$F9,$68,$C3,$C0,$E1,$8F,$85,$FC,$DB,$6F,$F6,$F3,$C2,$79,$57,$49
	.dc.b	$FD,$DD,$6E,$BC,$95,$F9,$C7,$86,$7C,$AF,$C6,$F0,$C7,$E1,$B7,$1F
	.dc.b	$25,$FF,$11,$AE,$D7,$EA,$3C,$58,$EE,$7C,$F8,$72,$9D,$57,$66,$CF
	.dc.b	$E7,$B1,$F7,$63,$C4,$F2,$47,$E8,$A7,$CE,$7A,$A3,$4F,$28,$F6,$57
	.dc.b	$57,$E4,$DF,$85,$B8,$FA,$AB,$CE,$71,$C7,$D0,$74,$9F,$AF,$7C,$DF
	.dc.b	$80,$79,$EF,$A2,$C7,$D3,$7E,$3C,$7D,$33,$FC,$2A,$72,$F8,$8F,$AE
	.dc.b	$78,$EF,$E5,$B9,$C3,$C4,$F8,$0F,$69,$E1,$C6,$38,$25,$F0,$BF,$C0
	.dc.b	$47,$A8,$EC,$7D,$A7,$2F,$A7,$1F,$5E,$3F,$3A,$38,$1E,$98,$F7,$CF
	.dc.b	$AA,$3C,$95,$EE,$3C,$13,$CB,$1F,$91,$7C,$EB,$B6,$1E,$8C,$7E,$29
	.dc.b	$27,$F2,$AF,$F5,$5E,$0F,$EF,$9F,$11,$FD,$7B,$6F,$F5,$B6,$A8,$C6
	.dc.b	$2A,$DB,$FD,$D7,$E5,$B3,$18,$8F,$F7,$CB,$5F,$D8,$B5,$EB,$3F,$EF
	.dc.b	$E7,$FD,$89,$C8,$F6,$39,$B6,$3F,$C0,$51,$8F,$83,$CB,$E0,$C5,$EE
	.dc.b	$53,$35,$FE,$10,$87,$CE,$EC,$91,$FE,$0D,$D2,$BE,$89,$DB,$FE,$C8
	.dc.b	$6B,$99,$F4,$3F,$DE,$7A,$5F,$35,$F1,$B8,$7F,$C3,$8D,$8C,$CF,$F9
	.dc.b	$C9,$7C,$D6,$67,$87,$C3,$7F,$88,$EB,$FC,$EE,$D7,$FC,$75,$5C,$9F
	.dc.b	$F0,$46,$5F,$C7,$3A,$FF,$D8,$D1,$FE,$85,$0E,$1D,$9F,$C9,$5C,$62
	.dc.b	$9E,$5E,$CF,$DE,$AE,$3E,$A5,$BF,$F6,$4C,$75,$6B,$D0,$D9,$E1,$FC
	.dc.b	$75,$E9,$73,$6C,$66,$3B,$7C,$FB,$E1,$F5,$1F,$32,$55,$C1,$FC,$33
	.dc.b	$5F,$E7,$66,$A3,$84,$6F,$3F,$64,$FF,$43,$E8,$7D,$07,$B5,$B1,$E3
	.dc.b	$3E,$A7,$B7,$F5,$C8,$FD,$7F,$0E,$BF,$E6,$BE,$1F,$D2,$3F,$5B,$AB
	.dc.b	$38,$F8,$0F,$F2,$E7,$47,$E0,$49,$3F,$81,$E9,$AF,$DC,$8B,$3C,$79
	.dc.b	$15,$E7,$F2,$65,$FD,$DD,$3F,$CF,$CD,$F4,$32,$FE,$AE,$32,$D9,$F9
	.dc.b	$91,$88,$F9,$2E,$B0,$F8,$F0,$F9,$EF,$C4,$BF,$8B,$6B,$F7,$E7,$87
	.dc.b	$3F,$F0,$6E,$8F,$D9,$12,$7F,$61,$F3,$F5,$FF,$3B,$45,$C3,$FD,$FB
	.dc.b	$D8,$AF,$DD,$7A,$12,$FF,$70,$FF,$0C,$5F,$35,$97,$FD,$01,$DB,$FC
	.dc.b	$04,$C6,$1D,$3C,$BF,$E0,$84,$9F,$F4,$88,$F8,$FC,$3F,$E8,$98,$B3
	.dc.b	$F8,$8F,$0F,$FC,$15,$58,$D7,$F9,$1D,$5B,$FE,$44,$73,$3F,$87,$7C
	.dc.b	$CF,$F1,$32,$3E,$0E,$CA,$FF,$A5,$E2,$E3,$DE,$3F,$31,$7F,$CC,$3C
	.dc.b	$DF,$E3,$84,$8F,$4F,$CB,$D7,$FD,$41,$16,$B9,$4F,$EE,$78,$D3,$49
	.dc.b	$5F,$0B,$C8,$AF,$F3,$19,$DB,$FC,$71,$AE,$67,$F6,$91,$8B,$FB,$3D
	.dc.b	$BF,$CC,$4E,$6F,$DC,$ED,$FE,$68,$4B,$FA,$3E,$E7,$F0,$DF,$C2,$5E
	.dc.b	$A7,$EF,$DB,$3F,$CB,$1A,$FF,$45,$6B,$FD,$2B,$9D,$D3,$FD,$07,$E7
	.dc.b	$E4,$FF,$3F,$8F,$E8,$9F,$11,$23,$FA,$13,$F8,$B7,$C5,$F0,$7E,$12
	.dc.b	$3E,$17,$F9,$EF,$3F,$FB,$7F,$BF,$FA,$77,$83,$2B,$FA,$64,$7F,$51
	.dc.b	$F9,$E9,$1F,$D3,$DF,$E8,$5F,$30,$F0,$BF,$96,$7E,$27,$FD,$6E,$BF
	.dc.b	$B0,$35,$FE,$36,$D7,$FC,$AE,$D7,$FB,$86,$3F,$13,$8F,$F2,$CD,$E9
	.dc.b	$FF,$30,$3F,$DD,$78,$BF,$B2,$47,$F7,$48,$6B,$49,$7F,$E6,$7B,$FE
	.dc.b	$31,$93,$E2,$3C,$FA,$FF,$33,$CF,$F8,$9D,$FF,$60,$6B,$EA,$8F,$8F
	.dc.b	$8B,$E8,$3C,$04,$9F,$F8,$03,$0F,$8F,$FE,$9D,$F8,$D7,$FA,$23,$3F
	.dc.b	$E2,$2D,$D7,$FC,$EE,$DB,$FF,$73,$7E,$4E,$F9,$BF,$EA,$DF,$42,$F0
	.dc.b	$7F,$A4,$36,$BF,$BA,$B5,$FF,$07,$35,$FF,$7D,$F1,$3E,$1B,$FB,$C3
	.dc.b	$63,$E2,$65,$F1,$63,$11,$FD,$7F,$27,$C6,$7F,$67,$F9,$A7,$77,$F6
	.dc.b	$9F,$C1,$7E,$15,$FF,$73,$6B,$FD,$9E,$76,$E7,$8B,$FB,$04,$EA,$3F
	.dc.b	$AE,$B5,$FF,$4C,$B5,$FF,$4F,$4B,$C4,$F1,$12,$3A,$1F,$E0,$3F,$AE
	.dc.b	$7C,$9F,$F7,$D6,$C6,$63,$FC,$31,$FB,$03,$E5,$B5,$1F,$DC,$EB,$75
	.dc.b	$FE,$37,$6B,$FD,$C9,$AF,$FA,$B9,$DF,$2F,$F6,$3D,$DE,$E7,$CE,$3A
	.dc.b	$D7,$97,$8B,$C8,$47,$E8,$8A,$76,$7F,$84,$1F,$EA,$F8,$7D,$49,$F1
	.dc.b	$57,$9C,$8F,$F9,$36,$78,$F0,$FE,$D0,$91,$DA,$3A,$FC,$96,$23,$FC
	.dc.b	$4F,$D5,$FE,$5F,$FA,$1F,$D5,$19,$AF,$D3,$F8,$DF,$AD,$FA,$59,$7C
	.dc.b	$F4,$4E,$23,$9E,$0A,$C5,$F2,$9F,$F1,$C7,$D1,$3C,$6E,$30,$F8,$AF
	.dc.b	$C7,$BF,$ED,$D9,$D6,$3F,$ED,$F9,$CC,$A7,$2F,$1D,$7F,$DC,$31,$FE
	.dc.b	$18,$8C,$1C,$EF,$BD,$72,$9E,$2D,$8C,$D7,$2F,$6B,$FD,$D7,$39,$C3
	.dc.b	$C5,$3D,$AB,$78,$75,$9C,$BE,$9B,$E3,$5B,$9C,$CF,$A6,$B8,$B6,$33
	.dc.b	$87,$A7,$0F,$04,$FA,$8D,$16,$3D,$37,$E7,$8E,$37,$F3,$1F,$F2,$4C
	.dc.b	$FC,$66,$C6,$5F,$21,$FE,$95,$F8,$97,$CB,$BE,$1B,$FD,$03,$18,$6C
	.dc.b	$66,$78,$35,$EF,$E5,$BE,$4F,$FD,$13,$8F,$FA,$7D,$B1,$99,$FF,$52
	.dc.b	$4E,$5B,$19,$CF,$EB,$37,$34,$CF,$11,$F4,$3F,$E6,$2B,$DE,$3F,$2B
	.dc.b	$0E,$17,$F8,$B6,$BF,$CB,$6B,$86,$1F,$88,$9C,$E3,$E3,$C3,$9D,$FD
	.dc.b	$1C,$7B,$F8,$7D,$72,$93,$F4,$F0,$EB,$5A,$8C,$4E,$AF,$99,$F9,$A9
	.dc.b	$D3,$8C,$7F,$AE,$F1,$BE,$8B,$FC,$DD,$96,$1B,$8F,$F6,$6F,$6E,$D7
	.dc.b	$FB,$5E,$6B,$FE,$4A,$AF,$EA,$71,$89,$CB,$AD,$77,$ED,$5A,$FE,$B2
	.dc.b	$DC,$3F,$77,$3D,$3E,$13,$1C,$DF,$DD,$3C,$3F,$F3,$5D,$7A,$1A,$FF
	.dc.b	$7A,$6E,$7F,$EE,$B6,$A7,$CE,$7A,$3D,$AB,$97,$FB,$13,$D3,$F7,$DF
	.dc.b	$FA,$1B,$E4,$3E,$0E,$2F,$83,$FF,$12,$DF,$FB,$0F,$D0,$56,$A7,$37
	.dc.b	$FE,$AA,$BE,$18,$79,$9B,$87,$F2,$9A,$DD,$7E,$38,$FE,$68,$F2,$9F
	.dc.b	$F5,$03,$67,$EB,$FB,$9D,$EF,$CF,$F4,$0F,$7A,$9E,$8F,$A1,$FE,$E5
	.dc.b	$7F,$ED,$99,$D3,$FC,$D6,$33,$97,$9D,$AF,$8F,$24,$62,$3F,$C3,$8D
	.dc.b	$7F,$C5,$3F,$49,$DF,$C6,$FC,$1B,$6F,$FD,$D1,$F4,$1F,$E1,$57,$FD
	.dc.b	$FA,$DA,$FF,$67,$87,$0F,$FB,$49,$AF,$F8,$AF,$E8,$1F,$1F,$F2,$DB
	.dc.b	$5F,$D4,$DB,$3F,$45,$AF,$AF,$77,$77,$D7,$3F,$33,$E8,$5B,$FF,$15
	.dc.b	$FA,$7F,$9C,$3F,$3B,$F4,$96,$3F,$C5,$FE,$07,$E3,$08,$FA,$BA,$47
	.dc.b	$F8,$AD,$C3,$FC,$55,$FF,$1B,$F8,$3F,$A8,$F9,$2F,$D4,$BF,$AC,$E3
	.dc.b	$FF,$03,$4F,$F7,$6A,$FE,$ED,$E9,$7E,$BB,$6D,$3F,$94,$F9,$CF,$F3
	.dc.b	$FD,$7F,$73,$6C,$FE,$45,$AF,$DB,$6B,$FE,$5C,$BF,$F0,$CC,$F7,$6C
	.dc.b	$7F,$76,$7F,$DF,$B1,$9F,$4F,$CD,$EF,$F9,$F9,$AC,$F1,$7D,$5E,$9F
	.dc.b	$EC,$AF,$FC,$2B,$3F,$15,$B8,$7E,$7C,$F8,$EE,$1F,$D1,$67,$FE,$15
	.dc.b	$F9,$DF,$88,$7C,$5F,$A0,$B7,$E7,$EF,$F9,$1F,$01,$F2,$5B,$1F,$E1
	.dc.b	$1D,$BF,$92,$63,$F0,$7C,$5F,$D9,$46,$23,$FC,$EC,$D7,$C7,$E6,$7D
	.dc.b	$33,$EA,$7C,$F5,$BF,$8D,$F2,$9F,$C1,$BE,$7F,$FD,$03,$18,$3F,$C5
	.dc.b	$B6,$8C,$BF,$2E,$7E,$21,$27,$E2,$36,$35,$1F,$4F,$E6,$3F,$09,$3F
	.dc.b	$85,$FA,$09,$3F,$5F,$E6,$ED,$FD,$79,$BC,$5F,$EE,$F1,$38,$AF,$C0
	.dc.b	$7A,$1F,$8F,$3F,$21,$F2,$76,$3F,$05,$17,$0F,$C1,$4D,$BE,$9F,$35
	.dc.b	$F9,$7F,$77,$F3,$0D,$C3,$F1,$4D,$BF,$C5,$36,$B8,$CF,$E3,$EB,$8B
	.dc.b	$4F,$C9,$7B,$3F,$A3,$3F,$A7,$FA,$EF,$95,$C4,$FE,$96,$30,$F0,$31
	.dc.b	$1A,$C7,$FC,$9F,$2F,$6F,$99,$D3,$E6,$6D,$F3,$8B,$5C,$26,$E4,$AF
	.dc.b	$E6,$DF,$38,$EB,$FC,$17,$8D,$FD,$22,$31,$3F,$82,$E5,$FF,$4C,$BF
	.dc.b	$84,$8C,$E3,$FC,$81,$AF,$E8,$DA,$F6,$BF,$EB,$AA,$38,$E1,$E3,$8F
	.dc.b	$F0,$AC,$62,$BF,$A5,$7E,$C0,$F9,$7F,$26,$1F,$C1,$EE,$57,$C9,$F3
	.dc.b	$6C,$7F,$A0,$E3,$F4,$6E,$31,$ED,$E0,$7F,$8D,$6B,$15,$FE,$DD,$F9
	.dc.b	$EF,$F7,$93,$FE,$7F,$6D,$FD,$0F,$9D,$FE,$D3,$18,$C3,$9F,$0A,$73
	.dc.b	$E2,$FD,$E6,$E1,$FA,$AF,$A1,$7C,$D7,$F7,$4E,$67,$27,$F7,$69,$CD
	.dc.b	$7E,$C2,$72,$F1,$AF,$93,$3A,$D3,$FA,$7B,$5F,$E2,$4D,$7F,$B9,$3B
	.dc.b	$CF,$E7,$42,$31,$5F,$EF,$DF,$D8,$7F,$A7,$D6,$A7,$35,$FE,$48,$FA
	.dc.b	$17,$CD,$FF,$83,$1B,$87,$E6,$78,$F1,$3F,$9E,$6C,$64,$FE,$17,$3C
	.dc.b	$2B,$EB,$1F,$EB,$9E,$07,$85,$F3,$8F,$D3,$FC,$ED,$BF,$D3,$36,$BF
	.dc.b	$46,$DC,$FE,$9E,$BF,$E2,$8E,$2F,$85,$5F,$D6,$9B,$C9,$FE,$75,$9D
	.dc.b	$1F,$B4,$3F,$6B,$F3,$B7,$1F,$DA,$6D,$FB,$8F,$6F,$F7,$2D,$8C,$93
	.dc.b	$1F,$DA,$FC,$9E,$A3,$F7,$9F,$37,$E3,$23,$FD,$B7,$34,$62,$BF,$7A
	.dc.b	$FF,$06,$F9,$CF,$E0,$EF,$F8,$EA,$3F,$D7,$43,$C2,$FF,$7E,$3F,$E7
	.dc.b	$CD,$BF,$92,$E1,$FE,$7E,$9C,$CF,$FA,$C8,$E1,$EB,$FF,$88,$FC,$6F
	.dc.b	$D8,$4E,$A7,$AF,$D0,$F0,$C3,$9C,$23,$4F,$FB,$3C,$C5,$F3,$C3,$FD
	.dc.b	$1B,$3A,$F8,$9B,$39,$AD,$9F,$D9,$E3,$2F,$A7,$1F,$E0,$55,$C6,$36
	.dc.b	$7F,$BF,$63,$35,$FE,$2E,$AE,$33,$9B,$AF,$C6,$C7,$D5,$7E,$EB,$E1
	.dc.b	$3F,$08,$F0,$57,$C7,$8C,$E1,$C2,$39,$CF,$BE,$30,$F9,$0F,$8F,$3E
	.dc.b	$D8,$F4,$1A,$3D,$57,$C0,$B1,$F0,$DD,$BC,$4F,$09,$67,$E3,$1F,$0B
	.dc.b	$1F,$A7,$5C,$23,$D2,$78,$CF,$A9,$5D,$EB,$DE,$D8,$E7,$3E,$59,$EF
	.dc.b	$1C,$A3,$A3,$F2,$CE,$98,$FC,$EB,$F0,$D7,$09,$D5,$70,$9C,$C7,$A2
	.dc.b	$FC,$0F,$E4,$5A,$FC,$83,$C1,$5C,$E7,$8C,$6E,$3E,$31,$EA,$2B,$D5
	.dc.b	$AE,$CF,$87,$3F,$89,$EA,$FD,$34,$7C,$A6,$BF,$A3,$6B,$FA,$C7,$3F
	.dc.b	$16,$F3,$18,$7F,$BA,$27,$31,$8A,$46,$27,$57,$FE,$A1,$BE,$51,$8A
	.dc.b	$FF,$78,$C6,$1F,$C9,$0C,$1E,$C7,$FA,$A6,$FF,$73,$18,$9F,$CB,$1C
	.dc.b	$B1,$F1,$4F,$EF,$AF,$A3,$C2,$F9,$5F,$FA,$CC,$E5,$7E,$0A,$DC,$79
	.dc.b	$EF,$D9,$97,$1A,$FD,$6C,$F0,$8E,$D3,$A8,$E1,$8F,$92,$BD,$58,$78
	.dc.b	$11,$83,$58,$78,$1F,$54,$F6,$9F,$F6,$46,$1E,$19,$FE,$03,$5F,$1E
	.dc.b	$3C,$35,$B9,$86,$B0,$F2,$3D,$0F,$E1,$38,$7A,$4F,$85,$87,$67,$D3
	.dc.b	$1C,$6B,$D5,$1F,$1B,$1F,$11,$E0,$9F,$91,$7D,$9F,$61,$DA,$FA,$3E
	.dc.b	$1C,$38,$1E,$B2,$5F,$81,$E9,$7D,$A3,$CB,$3F,$1A,$FE,$EC,$6C,$CC
	.dc.b	$77,$3F,$90,$E3,$F1,$63,$51,$F4,$1F,$80,$DC,$3C,$AA,$78,$3F,$1E
	.dc.b	$FC,$4E,$CF,$16,$31,$F0,$13,$1F,$6E,$1E,$08,$F8,$71,$EE,$7E,$14
	.dc.b	$64,$F1,$BA,$78,$5F,$E0,$E3,$E5,$DF,$89,$F3,$63,$F0,$9F,$EF,$5F
	.dc.b	$05,$FE,$FB,$D7,$F1,$BC,$EE,$24,$E6,$B6,$7F,$9B,$7E,$73,$15,$B9
	.dc.b	$FF,$7B,$CE,$8F,$F5,$D4,$65,$FE,$ED,$1A,$8E,$DF,$0B,$CA,$CD,$31
	.dc.b	$F1,$B6,$FF,$CF,$9E,$FE,$CC,$3F,$E7,$AC,$B9,$11,$FF,$41,$7C,$E5
	.dc.b	$7C,$78,$7F,$9A,$31,$F7,$57,$8E,$31,$2F,$93,$F8,$1C,$5F,$E8,$DE
	.dc.b	$CB,$E7,$9F,$83,$5B,$3F,$C1,$18,$7F,$82,$67,$51,$83,$F9,$0B,$87
	.dc.b	$FC,$6F,$5D,$5F,$F7,$E1,$CC,$FF,$1D,$E3,$FC,$03,$E7,$3E,$94,$7F
	.dc.b	$97,$BD,$CB,$E4,$1F,$1C,$E1,$93,$FE,$D0,$9D,$3D,$A7,$9F,$D0,$BE
	.dc.b	$36,$92,$7D,$18,$BE,$03,$C8,$FF,$98,$2B,$84,$FE,$01,$AF,$FB,$E4
	.dc.b	$38,$7F,$DF,$8D,$7F,$8D,$9A,$AF,$6F,$D0,$BE,$36,$9E,$39,$F8,$63
	.dc.b	$E7,$7E,$9F,$85,$23,$E4,$C5,$7E,$6E,$FF,$E4,$47,$D6,$D7,$9C,$7D
	.dc.b	$CE,$79,$1A,$7E,$EF,$CD,$62,$39,$C5,$D3,$FD,$43,$09,$FA,$BD,$2F
	.dc.b	$46,$92,$7F,$E0,$0E,$CF,$95,$C4,$7C,$7A,$E0,$FE,$A9,$AE,$6B,$F5
	.dc.b	$AF,$ED,$1E,$FE,$DC,$8F,$C7,$19,$3F,$17,$5F,$96,$23,$F8,$EF,$97
	.dc.b	$AF,$CC,$8B,$97,$B2,$7F,$3B,$F2,$9F,$9E,$F9,$59,$7F,$96,$8A,$FD
	.dc.b	$48,$B5,$F1,$27,$F4,$BC,$7F,$A6,$F9,$59,$7F,$9E,$8C,$BF,$A0,$8C
	.dc.b	$BF,$7A,$32,$FE,$02,$33,$5F,$42,$7F,$63,$C7,$2B,$F5,$2F,$ED,$3E
	.dc.b	$17,$11,$F8,$88,$B5,$F6,$27,$F7,$3F,$29,$2B,$F6,$2F,$EF,$3E,$57
	.dc.b	$11,$F9,$08,$B5,$F8,$09,$FE,$03,$E4,$4A,$FD,$CB,$FC,$13,$E4,$F1
	.dc.b	$1F,$C8,$8F,$F3,$F8,$EE,$FE,$98,$FE,$9D,$AE,$E7,$F7,$FF,$41,$FE
	.dc.b	$C6,$8F,$BA,$D7,$F4,$71,$FC,$4B,$E0,$A4,$7F,$0A,$7F,$8A,$FC,$8F
	.dc.b	$F1,$AF,$C2,$5C,$5F,$CD,$B4,$46,$2F,$F7,$11,$89,$D4,$E6,$7D,$5E
	.dc.b	$7F,$F2,$0F,$92,$91,$FC,$62,$3F,$91,$B0,$FE,$36,$7F,$26,$F9,$2F
	.dc.b	$E5,$7F,$32,$F9,$3F,$13,$67,$C2,$D7,$F8,$29,$FE,$48,$70,$FF,$08
	.dc.b	$C9,$1F,$C5,$63,$F8,$BA,$FA,$2E,$E7,$83,$C5,$F0,$F6,$CA,$DB,$E4
	.dc.b	$F9,$4F,$8D,$C1,$F0,$09,$3F,$13,$64,$7A,$3E,$53,$D6,$49,$F0,$F9
	.dc.b	$5E,$EF,$94,$91,$E4,$7E,$0F,$C4,$FC,$2F,$94,$F8,$24,$9F,$82,$FC
	.dc.b	$4F,$93,$F8,$E4,$8F,$6F,$C2,$7F,$54,$FA,$07,$35,$24,$FF,$38,$87
	.dc.b	$47,$F4,$8F,$9F,$FF,$3A,$CF,$F3,$06,$BF,$E4,$C6,$A6,$FD,$03,$E0
	.dc.b	$FF,$88,$B6,$7F,$BF,$B5,$FE,$80,$D7,$EB,$73,$7F,$A0,$5F,$E2,$BE
	.dc.b	$07,$CA,$6F,$17,$E1,$3D,$CF,$EB,$B3,$99,$FE,$94,$D7,$F0,$6D,$7F
	.dc.b	$A4,$B5,$FF,$32,$FD,$07,$F3,$37,$F9,$1B,$5F,$F4,$8F,$AD,$FD,$42
	.dc.b	$31,$D6,$8E,$78,$7F,$80,$3E,$83,$F2,$53,$9B,$FF,$4C,$C6,$2B,$FC
	.dc.b	$E7,$28,$C4,$FE,$5F,$E6,$39,$77,$7E,$BE,$7F,$33,$3C,$1A,$F4,$8D
	.dc.b	$1F,$DF,$BE,$83,$F3,$91,$E8,$3F,$D0,$2E,$2F,$C8,$78,$67,$F5,$BF
	.dc.b	$33,$2B,$F4,$6F,$EC,$3E,$1F,$11,$FA,$CA,$FD,$4C,$FE,$C8,$91,$FB
	.dc.b	$2F,$98,$FD,$B9,$1F,$D6,$FA,$55,$FF,$0C,$F0,$BE,$67,$F7,$A6,$70
	.dc.b	$F2,$C7,$FB,$19,$BF,$11,$FC,$08,$FE,$07,$F0,$D6,$FF,$73,$15,$EB
	.dc.b	$1F,$E2,$0F,$A0,$7C,$37,$FB,$1A,$5F,$15,$E1,$7F,$D4,$B3,$AB,$FF
	.dc.b	$53,$CE,$65,$39,$78,$E5,$FE,$12,$87,$CD,$7F,$85,$2B,$84,$6D,$B8
	.dc.b	$78,$5B,$87,$6B,$E5,$7C,$5E,$95,$B6,$CE,$F0,$F3,$CE,$5F,$4D,$71
	.dc.b	$73,$3B,$7D,$33,$B6,$C6,$70,$F4,$C5,$9D,$47,$A6,$FC,$33,$F0,$0E
	.dc.b	$E6,$EB,$FC,$F3,$3F,$16,$3F,$98,$B5,$E2,$7F,$34,$F8,$A7,$83,$FE
	.dc.b	$87,$8F,$F2,$04,$63,$E1,$FF,$D0,$6D,$8C,$CE,$DE,$BF,$12,$F3,$BC
	.dc.b	$2D,$8F,$F2,$56,$08,$C4,$70,$9D,$9F,$06,$38,$6B,$F7,$FC,$37,$AD
	.dc.b	$FB,$98,$FB,$73,$AD,$7F,$00,$DC,$3F,$12,$D7,$F9,$6D,$71,$9D,$E1
	.dc.b	$DC,$93,$F2,$CD,$F5,$E3,$0E,$F3,$6B,$FC,$BF,$8F,$92,$B1,$7E,$0A
	.dc.b	$AF,$23,$A6,$5F,$06,$3F,$11,$C4,$F9,$CF,$E6,$97,$FE,$76,$FA,$67
	.dc.b	$C2,$7F,$04,$8F,$E2,$B3,$AC,$7F,$D1,$F5,$A9,$FF,$3E,$CE,$67,$FB
	.dc.b	$63,$5E,$3E,$27,$F2,$17,$2D,$7F,$90,$FC,$53,$CB,$F0,$43,$2F,$F5
	.dc.b	$54,$61,$F4,$7C,$35,$AE,$18,$F8,$A3,$15,$BC,$51,$89,$F7,$7C,$4F
	.dc.b	$87,$1E,$D9,$A3,$19,$23,$15,$FE,$1C,$6D,$7F,$21,$6C,$FF,$2A,$6C
	.dc.b	$FF,$A5,$3E,$19,$E7,$7A,$67,$37,$ED,$D1,$E7,$E7,$B3,$D8,$DC,$7F
	.dc.b	$D5,$33,$A7,$FC,$5A,$DA,$FE,$64,$D7,$FC,$38,$D7,$E6,$CF,$49,$FF
	.dc.b	$3E,$E2,$F3,$F1,$5F,$73,$E6,$7C,$C7,$CF,$FA,$67,$A7,$8C,$3F,$D7
	.dc.b	$71,$87,$FC,$7E,$DA,$FE,$88,$D7,$FC,$5C,$D7,$FD,$6F,$F4,$1F,$EB
	.dc.b	$19,$FF,$7F,$B6,$3F,$D8,$0D,$7F,$C7,$2D,$7F,$B3,$EA,$F9,$BF,$98
	.dc.b	$7E,$0F,$7B,$7F,$ED,$0F,$A0,$7C,$B7,$DC,$6C,$7F,$95,$9B,$87,$F5
	.dc.b	$B6,$BF,$E4,$55,$F2,$DF,$94,$87,$DE,$FD,$1E,$4F,$ED,$35,$AB,$FF
	.dc.b	$54,$CE,$5F,$3F,$CD,$FE,$04,$FC,$C7,$0D,$8F,$D8,$7C,$CF,$E4,$36
	.dc.b	$34,$91,$FE,$6C,$FA,$17,$CA,$7E,$2F,$E5,$7F,$49,$18,$9F,$DD,$47
	.dc.b	$E3,$F7,$91,$F8,$AF,$1E,$EF,$FB,$AE,$0F,$01,$27,$F3,$CF,$E8,$B0
	.dc.b	$7C,$77,$E6,$7C,$0E,$ED,$AF,$CA,$FB,$5F,$9D,$F9,$AF,$DA,$36,$3F
	.dc.b	$86,$7C,$D7,$B4,$8F,$EA,$DE,$DF,$1D,$FA,$5F,$0B,$D6,$D9,$FD,$0F
	.dc.b	$B5,$FA,$9F,$99,$FE,$00,$D8,$FE,$31,$F2,$9F,$24,$8F,$ED,$DF,$57
	.dc.b	$C1,$7E,$C7,$D4,$F8,$ED,$9F,$F6,$FF,$D0,$B1,$FB,$CF,$FB,$B6,$33
	.dc.b	$18,$F9,$5C,$BF,$C2,$78,$D5,$C6,$BF,$D1,$F0,$33,$3B,$9D,$46,$9B
	.dc.b	$DD,$FC,$95,$B5,$C6,$7A,$FC,$07,$82,$3F,$C8,$E7,$F9,$23,$93,$FC
	.dc.b	$F2,$FF,$39,$F5,$3F,$07,$E8,$FF,$0F,$6C,$64,$E3,$7F,$CB,$6B,$8B
	.dc.b	$FC,$F2,$7F,$17,$E8,$7F,$17,$F2,$6D,$FF,$45,$6D,$7F,$42,$6E,$7F
	.dc.b	$C8,$5F,$E6,$DF,$39,$FE,$B5,$AF,$E9,$9F,$37,$8F,$80,$FE,$44,$D8
	.dc.b	$C9,$E4,$B5,$71,$C3,$FA,$7D,$6D,$B5,$B9,$D5,$7F,$31,$AF,$E8,$9F
	.dc.b	$01,$2B,$5F,$15,$70,$FF,$31,$5B,$D0,$5E,$63,$FC,$C7,$8A,$F1,$87
	.dc.b	$F5,$6F,$81,$FE,$74,$FF,$51,$6B,$FC,$D9,$B8,$F6,$C5,$18,$C3,$FB
	.dc.b	$17,$D0,$FF,$31,$BF,$A7,$6F,$9F,$78,$BF,$C3,$EB,$53,$FE,$79,$35
	.dc.b	$19,$F2,$7F,$25,$3A,$9F,$F6,$FB,$5F,$C5,$35,$FC,$A3,$D3,$0F,$1C
	.dc.b	$FF,$1D,$F7,$FF,$AA,$D2,$B5,$8F,$FA,$5B,$5F,$ED,$F3,$F2,$BE,$AF
	.dc.b	$E5,$4F,$D3,$F6,$7F,$BB,$46,$27,$E7,$D7,$0C,$BF,$D1,$F3,$97,$E7
	.dc.b	$D7,$0D,$BF,$97,$7D,$3F,$FB,$9B,$1F,$8A,$DC,$7E,$17,$D0,$FF,$2F
	.dc.b	$7E,$24,$67,$0F,$F0,$73,$6B,$F5,$4D,$9F,$DA,$57,$1A,$DF,$57,$F3
	.dc.b	$DE,$0C,$61,$FE,$B4,$CF,$F9,$66,$1F,$46,$72,$FF,$85,$63,$37,$FE
	.dc.b	$1D,$9F,$24,$FE,$E4,$FB,$9E,$4F,$F0,$78,$CC,$7E,$F5,$AF,$F0,$77
	.dc.b	$AB,$C2,$FF,$0B,$BB,$C1,$F3,$5F,$F8,$87,$E1,$65,$7E,$8F,$D3,$B8
	.dc.b	$7E,$46,$6E,$5F,$9C,$9B,$5F,$E3,$0F,$7F,$FC,$54,$7E,$E3,$CD,$B1
	.dc.b	$FE,$10,$8B,$87,$DA,$9B,$7F,$6B,$E7,$FF,$C4,$FB,$7F,$75,$AF,$89
	.dc.b	$18,$26,$1F,$BD,$9F,$EF,$D5,$FA,$EF,$1F,$F0,$F5,$FE,$3B,$F8,$79
	.dc.b	$7F,$BD,$EA,$B8,$FE,$29,$B8,$7E,$D9,$B3,$F5,$FE,$7F,$C6,$47,$F9
	.dc.b	$9C,$B2,$4E,$6F,$F3,$4F,$F1,$8D,$3F,$32,$FF,$7F,$8F,$F6,$08,$F8
	.dc.b	$77,$CB,$FF,$35,$C3,$F8,$4E,$3F,$CF,$E7,$35,$F8,$1F,$85,$C5,$EF
	.dc.b	$C0,$5E,$2B,$43,$E8,$78,$BF,$E6,$19,$DB,$9A,$FE,$9B,$7F,$D3,$6B
	.dc.b	$78,$F8,$AB,$47,$48,$CD,$F2,$7A,$C7,$B1,$CE,$1C,$F0,$EB,$3C,$8F
	.dc.b	$14,$65,$E9,$1E,$B8,$C4,$7F,$B7,$CF,$8D,$19,$69,$EE,$C7,$BB,$83
	.dc.b	$85,$60,$F2,$BF,$EF,$43,$C5,$87,$B6,$33,$8F,$82,$31,$7E,$89,$E4
	.dc.b	$FA,$AF,$B9,$C2,$FC,$15,$F1,$E3,$15,$C2,$7D,$A7,$F7,$99,$EA,$E0
	.dc.b	$F8,$67,$C5,$3E,$55,$7B,$E7,$B5,$78,$23,$C4,$7C,$37,$BB,$B3,$D0
	.dc.b	$59,$ED,$5E,$07,$E2,$C6,$5A,$79,$71,$FB,$91,$CA,$31,$8F,$D4,$7F
	.dc.b	$33,$7E,$37,$E9,$4F,$8C,$E0,$FB,$30,$F2,$BC,$8F,$71,$F4,$E7,$93
	.dc.b	$F7,$DA,$F6,$9E,$8F,$63,$94,$7D,$1E,$6F,$D5,$47,$E8,$1A,$FE,$25
	.dc.b	$AF,$EC,$1C,$FC,$43,$D2,$FF,$62,$4A,$73,$18,$6E,$7F,$E3,$16,$CE
	.dc.b	$EF,$FC,$59,$7C,$EB,$F6,$53,$FB,$CF,$97,$FF,$18,$57,$EC,$B1,$46
	.dc.b	$1E,$38,$F8,$A7,$F6,$F8,$73,$BF,$CE,$57,$07,$AB,$5F,$D0,$46,$E7
	.dc.b	$8D,$7B,$32,$F0,$61,$FE,$41,$6D,$FE,$BA,$FB,$47,$0C,$7A,$D7,$96
	.dc.b	$BF,$CA,$18,$F7,$C7,$51,$E0,$C7,$B4,$FF,$93,$4F,$26,$1D,$A7,$F8
	.dc.b	$06,$3D,$A7,$DF,$8F,$EF,$9D,$9C,$71,$F3,$3A,$AE,$38,$76,$7C,$73
	.dc.b	$A6,$9E,$B8,$D2,$BF,$89,$5D,$63,$DA,$7D,$68,$CB,$FE,$6D,$7C,$CF
	.dc.b	$43,$F8,$C4,$79,$9D,$4F,$69,$DC,$7D,$F3,$D9,$5E,$57,$C5,$3B,$3E
	.dc.b	$F5,$78,$E3,$EE,$CF,$CF,$77,$7E,$5A,$F5,$D7,$86,$7B,$1F,$12,$7E
	.dc.b	$01,$E0,$61,$92,$63,$F4,$C9,$3F,$35,$FB,$78,$FD,$72,$47,$D1,$72
	.dc.b	$F9,$DC,$CF,$C2,$AF,$0B,$F1,$B0,$F8,$15,$E7,$C7,$E2,$9F,$3A,$7D
	.dc.b	$38,$78,$9E,$47,$F3,$CB,$FE,$D9,$F3,$FF,$D8,$F4,$FF,$14,$36,$7F
	.dc.b	$C4,$CD,$9F,$F3,$EE,$97,$85,$3F,$EA,$8A,$D9,$FD,$E2,$32,$D8,$D5
	.dc.b	$3E,$21,$F0,$FD,$7C,$1F,$E1,$73,$FD,$A0,$FF,$3F,$67,$E1,$AD,$E1
	.dc.b	$FE,$8F,$AD,$E7,$FC,$FD,$B0,$BE,$2F,$FA,$56,$B8,$BF,$E6,$1F,$47
	.dc.b	$D4,$47,$5F,$0B,$5F,$F2,$15,$7F,$84,$67,$D3,$A1,$A7,$B3,$0F,$E3
	.dc.b	$95,$E5,$D7,$D8,$7A,$7E,$0A,$C6,$48,$7F,$A7,$32,$47,$CF,$F8,$8F
	.dc.b	$8D,$3F,$EA,$D6,$BF,$EA,$26,$E5,$F1,$E9,$EF,$F1,$65,$71,$7E,$54
	.dc.b	$7D,$D8,$7D,$87,$AD,$FF,$4D,$B5,$FF,$4C,$36,$BF,$C3,$70,$FC,$89
	.dc.b	$E8,$F3,$AE,$13,$FE,$69,$6B,$FE,$E3,$6B,$FE,$ED,$6B,$E2,$D7,$F1
	.dc.b	$9E,$F4,$8F,$1C,$79,$F8,$FF,$D6,$71,$FE,$B0,$6B,$FE,$AD,$6E,$3E
	.dc.b	$65,$E0,$F1,$EF,$95,$C5,$E9,$1F,$8D,$F8,$0F,$CE,$7C,$DC,$AF,$C4
	.dc.b	$46,$5F,$F3,$14,$65,$FB,$11,$F2,$FC,$87,$CD,$D2,$4F,$FB,$1B,$E8
	.dc.b	$DE,$E6,$23,$FD,$1C,$7F,$A1,$67,$F5,$4D,$3F,$59,$5D,$1B,$1F,$67
	.dc.b	$E8,$D8,$D3,$A6,$BF,$A9,$E5,$FC,$FB,$FE,$B8,$FA,$3F,$DC,$1F,$8D
	.dc.b	$DE,$63,$FE,$7D,$8C,$BF,$8F,$8F,$C9,$7C,$7C,$8F,$C9,$3F,$B3,$F8
	.dc.b	$4C,$47,$E4,$E2,$DF,$E4,$E7,$F7,$1A,$7F,$07,$FA,$39,$7F,$B3,$8C
	.dc.b	$BF,$E8,$D8,$CB,$FA,$38,$CB,$FE,$D7,$8C,$D7,$E7,$E7,$F8,$27,$BD
	.dc.b	$FC,$4F,$E4,$A5,$FE,$FE,$32,$FF,$A6,$63,$2F,$EB,$E3,$F8,$67,$BB
	.dc.b	$FC,$6F,$86,$4F,$F0,$78,$CB,$FE,$A1,$8C,$BF,$B7,$8F,$D5,$7D,$07
	.dc.b	$CE,$F0,$D3,$23,$0D,$DB,$F0,$DE,$9F,$07,$EF,$CF,$F1,$49,$FF,$68
	.dc.b	$35,$E6,$1C,$3C,$CF,$F2,$F5,$6A,$1E,$EF,$90,$ED,$3A,$3F,$92,$7C
	.dc.b	$F3,$C4,$FE,$75,$F3,$52,$75,$C0,$AD,$47,$FB,$07,$E8,$FF,$90,$91
	.dc.b	$FE,$7C,$FF,$3C,$E6,$79,$1F,$D1,$7E,$8E,$57,$F1,$98,$FE,$95,$7F
	.dc.b	$C2,$F5,$FF,$7B,$79,$BF,$D2,$CC,$F8,$3F,$D3,$7E,$23,$F9,$79,$30
	.dc.b	$FE,$98,$F7,$F2,$7F,$A9,$7B,$92,$7F,$93,$47,$6E,$8F,$EA,$DF,$0F
	.dc.b	$27,$DB,$9A,$7F,$AD,$7A,$3F,$CF,$09,$3F,$D6,$5F,$E7,$FF,$0D,$FD
	.dc.b	$7C,$91,$FC,$CF,$E4,$3D,$7A,$FF,$63,$88,$76,$7F,$AE,$7C,$3F,$F4
	.dc.b	$52,$47,$FB,$73,$99,$1F,$D2,$7C,$5F,$EC,$44,$9F,$E8,$1E,$77,$F3
	.dc.b	$7F,$09,$F0,$FF,$CE,$7C,$1B,$5F,$74,$38,$7F,$B1,$72,$FE,$2A,$B1
	.dc.b	$D5,$FC,$E9,$FE,$19,$3F,$C4,$9A,$FE,$40,$38,$79,$63,$FD,$7F,$E8
	.dc.b	$BF,$BC,$12,$3F,$AB,$76,$FF,$4D,$F9,$E7,$C8,$7F,$1C,$6C,$FF,$81
	.dc.b	$DA,$FF,$22,$6B,$E1,$F9,$F7,$BF,$FC,$91,$B1,$FE,$0F,$6B,$FC,$A9
	.dc.b	$AF,$FB,$F7,$87,$DF,$AF,$F8,$17,$7F,$F0,$71,$FD,$86,$B8,$ED,$FC
	.dc.b	$9F,$CE,$FE,$E0,$4C,$3E,$67,$CE,$FC,$69,$D4,$7F,$38,$5E,$0A,$BF
	.dc.b	$44,$49,$D4,$65,$FF,$0C,$7D,$17,$F8,$7C,$EC,$FD,$63,$FC,$31,$9B
	.dc.b	$E8,$BF,$C4,$84,$8F,$1F,$0D,$7F,$C2,$B3,$B7,$F1,$93,$FE,$13,$3F
	.dc.b	$C5,$FF,$40,$AF,$F0,$67,$C4,$4A,$FD,$FC,$E5,$FF,$07,$46,$5F,$F0
	.dc.b	$73,$FE,$41,$8D,$47,$9F,$34,$31,$EE,$49,$F8,$D1,$96,$C7,$F8,$FF
	.dc.b	$E7,$24,$78,$A3,$FC,$29,$F3,$CF,$94,$FF,$15,$FC,$37,$DE,$8C,$7B
	.dc.b	$FF,$CB,$A3,$13,$AE,$5F,$E4,$F4,$F5,$3F,$94,$D6,$AF,$85,$6A,$F8
	.dc.b	$63,$FC,$E3,$0E,$AA,$FC,$18,$72,$BE,$18,$72,$AE,$18,$72,$58,$72
	.dc.b	$C3,$8C,$6C,$CC,$EE,$75,$7D,$27,$C0,$FA,$B0,$F0,$47,$6C,$39,$57
	.dc.b	$08,$C1,$C0,$F5,$4F,$07,$BC,$FF,$9D,$67,$DB,$1F,$BD,$6B,$E1,$3F
	.dc.b	$80,$7C,$1B,$E2,$9F,$3F,$FE,$7D,$8C,$36,$33,$3F,$E1,$A9,$CB,$FE
	.dc.b	$1C,$AF,$E2,$7E,$27,$87,$C0,$34,$93,$B8,$CB,$EB,$3E,$5B,$8D,$7F
	.dc.b	$7F,$E4,$7F,$39,$EC,$F8,$73,$9D,$BF,$02,$DB,$FB,$8D,$55,$C5,$ED
	.dc.b	$8F,$99,$B3,$B8,$C6,$1E,$0F,$1D,$8F,$83,$29,$87,$07,$EF,$63,$D6
	.dc.b	$FF,$31,$8F,$32,$4E,$B0,$E7,$7C,$4F,$B5,$C9,$FC,$33,$BF,$FD,$45
	.dc.b	$97,$F8,$32,$30,$FF,$A0,$FC,$2C,$D7,$FA,$7F,$AF,$F9,$7C,$E4,$FF
	.dc.b	$44,$5E,$AF,$F8,$BD,$DA,$DD,$FF,$A2,$FE,$5F,$5D,$8F,$91,$D4,$61
	.dc.b	$B9,$7F,$A4,$E3,$97,$CB,$E2,$FF,$8D,$1F,$E8,$4F,$8F,$FF,$43,$BF
	.dc.b	$E1,$AF,$84,$46,$23,$FC,$4F,$18,$BF,$F6,$53,$63,$FC,$CA,$DD,$3F
	.dc.b	$D2,$2D,$9F,$89,$9B,$DB,$F8,$77,$E1,$9F,$EC,$94,$F9,$5F,$6F,$CF
	.dc.b	$EA,$31,$3F,$EC,$B8,$C4,$F9,$6B,$FD,$3A,$DA,$DC,$79,$23,$07,$0A
	.dc.b	$FF,$5B,$76,$D9,$FE,$4D,$F3,$EF,$03,$E9,$36,$BF,$BC,$7C,$FB,$E4
	.dc.b	$F3,$5E,$A8,$C4,$FF,$BB,$5B,$7F,$E8,$86,$BF,$DD,$DA,$FE,$1F,$E7
	.dc.b	$FF,$C6,$73,$DA,$FF,$05,$18,$9F,$AF,$5D,$37,$FF,$78,$72,$7C,$D9
	.dc.b	$FE,$A7,$C3,$C3,$C8,$FC,$2B,$F3,$BE,$7D,$C3,$F1,$DB,$A7,$FA,$72
	.dc.b	$D1,$87,$FD,$3E,$D7,$FC,$BB,$8B,$37,$8F,$E6,$3F,$2D,$C5,$70,$FF
	.dc.b	$80,$1C,$5F,$96,$3F,$34,$D3,$FC,$45,$F5,$5F,$D2,$5F,$13,$71,$FF
	.dc.b	$57,$B5,$FF,$0F,$35,$FD,$77,$CE,$3E,$7F,$F5,$0D,$AF,$F3,$2B,$F1
	.dc.b	$DD,$C7,$EA,$3C,$59,$1A,$F9,$3B,$7F,$A4,$C3,$FC,$6B,$3E,$4F,$07
	.dc.b	$F1,$D1,$89,$F2,$C7,$E9,$76,$64,$F8,$EF,$19,$E4,$F8,$AB,$8F,$93
	.dc.b	$E2,$BF,$73,$F2,$BE,$86,$C7,$F9,$17,$C6,$FF,$1D,$BF,$55,$B9,$FF
	.dc.b	$B4,$9A,$FF,$90,$5A,$FC,$4F,$8A,$F8,$67,$C4,$F8,$6B,$5F,$13,$E1
	.dc.b	$BF,$84,$FC,$77,$C8,$6C,$78,$7C,$5F,$D6,$62,$69,$23,$FC,$59,$B3
	.dc.b	$E1,$3E,$89,$F4,$BE,$12,$CF,$F9,$7F,$E7,$91,$FC,$01,$FF,$76,$C6
	.dc.b	$63,$1A,$65,$FD,$E4,$F0,$9F,$CB,$37,$5F,$F7,$9B,$5F,$F3,$13,$9E
	.dc.b	$2F,$E6,$F0,$8C,$4F,$FA,$DB,$E7,$BF,$94,$3E,$13,$F0,$FE,$7F,$E2
	.dc.b	$8F,$C5,$FC,$25,$8C,$92,$FF,$C5,$9A,$27,$51,$8F,$43,$F2,$67,$E5
	.dc.b	$1B,$5F,$94,$8B,$1C,$8B,$97,$F4,$E1,$F3,$DF,$E9,$0A,$FF,$5F,$B6
	.dc.b	$7F,$C7,$8F,$F9,$1E,$33,$F0,$7F,$D2,$63,$55,$CD,$B3,$FD,$55,$E1
	.dc.b	$5C,$AB,$81,$27,$BF,$B7,$67,$BF,$BF,$FD,$43,$E6,$7B,$36,$BC,$AE
	.dc.b	$1F,$3F,$B9,$6B,$FC,$A5,$E5,$3E,$4F,$FA,$C9,$E7,$F3,$AD,$FF,$B7
	.dc.b	$27,$4D,$9D,$34,$FF,$55,$7B,$8F,$8E,$FE,$97,$C9,$FD,$03,$83,$F0
	.dc.b	$33,$9A,$FF,$58,$4E,$5F,$F7,$BC,$67,$2F,$F5,$50,$70,$FF,$AB,$5A
	.dc.b	$F0,$BF,$EA,$11,$C6,$FB,$C7,$F5,$CF,$9C,$FF,$47,$5F,$9A,$3E,$A7
	.dc.b	$AF,$23,$ED,$F9,$D7,$1F,$B7,$37,$0F,$B1,$36,$95,$3E,$77,$F9,$8C
	.dc.b	$7F,$B2,$FE,$7B,$F3,$B3,$9F,$3B,$FB,$6C,$62,$15,$C1,$FF,$00,$B7
	.dc.b	$0F,$D3,$AF,$9F,$FE,$EA,$E7,$0F,$8F,$EC,$7B,$71,$FE,$F3,$B7,$F7
	.dc.b	$B9,$FF,$6E,$6D,$BC,$BC,$F3,$97,$E6,$B5,$F9,$97,$F2,$89,$5F,$71
	.dc.b	$FF,$11,$7D,$1F,$F3,$B9,$FF,$0B,$D7,$3C,$FF,$39,$EA,$7F,$88,$BE
	.dc.b	$53,$F4,$0D,$C7,$F4,$EE,$1F,$D4,$FA,$57,$0F,$D4,$FA,$3F,$E2,$AF
	.dc.b	$81,$FD,$43,$6B,$F5,$EE,$1F,$D8,$F9,$B6,$BF,$63,$E5,$FF,$41,$F0
	.dc.b	$0D,$25,$7F,$B4,$63,$2F,$EE,$A3,$F6,$DD,$DF,$B7,$7F,$71,$D5,$FE
	.dc.b	$3C,$FA,$0B,$5F,$A3,$6E,$1F,$C0,$27,$F7,$DB,$63,$7F,$DF,$6B,$E3
	.dc.b	$26,$1E,$16,$DF,$F0,$87,$F8,$36,$1F,$C1,$DF,$F1,$9C,$7F,$AD,$C7
	.dc.b	$CF,$3E,$3F,$FC,$03,$A7,$F9,$8A,$31,$3F,$D4,$4D,$CE,$B0,$FC,$BC
	.dc.b	$EA,$7F,$D7,$A3,$E7,$F1,$93,$0F,$F3,$6C,$6E,$7C,$4E,$A3,$38,$7F
	.dc.b	$61,$9E,$4F,$76,$9E,$39,$E1,$87,$EA,$1E,$91,$FE,$C7,$36,$2F,$C5
	.dc.b	$3B,$AE,$D5,$CA,$B6,$7C,$5C,$30,$74,$F3,$BD,$F5,$E3,$8C,$4F,$6C
	.dc.b	$3A,$BC,$8F,$15,$F5,$36,$5A,$F0,$9E,$D7,$9D,$78,$23,$E1,$36,$7E
	.dc.b	$04,$6C,$E2,$FF,$A4,$F1,$F9,$2E,$8C,$63,$EF,$70,$7C,$38,$DB,$F6
	.dc.b	$2B,$B3,$6B,$DC,$76,$78,$57,$37,$EF,$DF,$AD,$EB,$87,$C2,$8F,$7D
	.dc.b	$79,$E7,$38,$78,$1F,$C0,$C6,$5F,$01,$E8,$3F,$1E,$7E,$47,$0F,$2B
	.dc.b	$C0,$D5,$76,$6C,$FC,$83,$D2,$F1,$6B,$D8,$F5,$9D,$8C,$5F,$DA,$AF
	.dc.b	$BB,$C9,$FA,$69,$FD,$13,$5F,$CF,$B5,$FD,$63,$9F,$86,$71,$7E,$46
	.dc.b	$7F,$A8,$57,$F3,$E5,$3A,$BF,$E5,$73,$98,$C4,$F0,$3F,$A2,$B8,$BE
	.dc.b	$D7,$D6,$BF,$65,$87,$F8,$5E,$FF,$2E,$7F,$4C,$70,$FB,$EF,$AC,$FE
	.dc.b	$E5,$B3,$E5,$37,$1D,$3C,$8F,$43,$BB,$F0,$4F,$9D,$F7,$65,$BC,$3F
	.dc.b	$0D,$3E,$08,$E1,$97,$E2,$71,$EF,$5E,$5C,$7A,$DF,$85,$B1,$F0,$30
	.dc.b	$EE,$DB,$EB,$3F,$2F,$0F,$04,$FF,$00,$C3,$E4,$C7,$82,$3F,$81,$E1
	.dc.b	$E3,$3B,$3E,$3C,$3D,$31,$EB,$C3,$AC,$C3,$A4,$E9,$F1,$1F,$89,$9F
	.dc.b	$8E,$74,$7E,$74,$7A,$1F,$BE,$F1,$7F,$98,$61,$E9,$9E,$17,$E7,$23
	.dc.b	$82,$4F,$82,$3D,$98,$FC,$02,$4F,$A6,$7C,$38,$7A,$6B,$E2,$9D,$23
	.dc.b	$94,$F4,$7C,$67,$C3,$98,$FB,$6F,$E9,$BA,$3A,$92,$BA,$D7,$3C,$7E
	.dc.b	$8B,$E3,$8C,$CF,$DF,$3A,$BE,$37,$A3,$A7,$E3,$DF,$8B,$0F,$6C,$71
	.dc.b	$3C,$2F,$F5,$68,$FF,$35,$E5,$FE,$DA,$D3,$FC,$C4,$DA,$FE,$85,$C7
	.dc.b	$FE,$41,$7F,$CD,$F7,$9B,$E3,$1F,$DE,$EB,$FD,$D1,$18,$8F,$DC,$1C
	.dc.b	$A3,$3F,$3D,$FE,$55,$8F,$F1,$CC,$FF,$9F,$63,$11,$FD,$47,$0C,$C6
	.dc.b	$27,$4F,$84,$C4,$72,$F0,$FF,$87,$4F,$FA,$1D,$B1,$CB,$C2,$E1,$F2
	.dc.b	$D3,$0F,$0C,$F6,$8E,$9F,$15,$99,$F4,$E0,$C5,$5E,$A2,$4F,$F9,$9A
	.dc.b	$75,$F2,$BE,$52,$47,$63,$FB,$65,$3E,$53,$31,$FE,$82,$D3,$E6,$69
	.dc.b	$F3,$E3,$E1,$1C,$2F,$E0,$7C,$A4,$8E,$98,$F9,$63,$E2,$7C,$9E,$67
	.dc.b	$8C,$BE,$23,$D5,$1C,$1A,$EB,$35,$FE,$06,$1F,$3C,$F8,$2C,$36,$3D
	.dc.b	$86,$A3,$F0,$9F,$2E,$EC,$DC,$7F,$A2,$A7,$FC,$EF,$5C,$5F,$F4,$6B
	.dc.b	$5F,$C3,$B5,$FE,$80,$EF,$2F,$F6,$CF,$CF,$30,$F3,$36,$BC,$75,$BE
	.dc.b	$3F,$E9,$1C,$9F,$E5,$29,$F3,$B5,$F9,$A7,$FA,$9F,$C1,$F8,$10,$F9
	.dc.b	$3B,$18,$F0,$A5,$72,$8F,$C9,$7C,$44,$8F,$A7,$C0,$8F,$CB,$79,$59
	.dc.b	$9F,$9D,$17,$6F,$D0,$4F,$CF,$F9,$E7,$B9,$49,$7F,$3B,$E7,$9F,$3D
	.dc.b	$88,$FF,$5A,$D7,$07,$E7,$35,$CD,$7D,$15,$19,$C1,$BF,$F6,$DF,$35
	.dc.b	$F3,$5F,$81,$6D,$FD,$EB,$E1,$5F,$B8,$E8,$FD,$8C,$FF,$91,$DB,$87
	.dc.b	$E3,$F8,$BF,$6D,$E8,$FF,$07,$9C,$92,$33,$5F,$23,$E4,$BF,$78,$47
	.dc.b	$E5,$FC,$15,$7F,$37,$4F,$0B,$A1,$9C,$7F,$85,$69,$FC,$14,$95,$F5
	.dc.b	$3E,$46,$BF,$9F,$8B,$3F,$C4,$66,$E9,$FC,$42,$6D,$7E,$E2,$7F,$88
	.dc.b	$7C,$4C,$8F,$BE,$FF,$13,$EC,$C4,$7E,$AE,$2E,$1F,$C7,$67,$F8,$C7
	.dc.b	$CF,$4A,$FC,$3B,$FC,$6F,$1C,$47,$EC,$E2,$D7,$F2,$59,$FA,$9E,$84
	.dc.b	$AF,$D8,$E4,$8F,$B1,$EE,$66,$7F,$DF,$33,$97,$F6,$4D,$73,$3F,$85
	.dc.b	$8C,$5F,$E1,$27,$FD,$DB,$F3,$DF,$79,$36,$77,$87,$F3,$AF,$9D,$78
	.dc.b	$DF,$D1,$5B,$7F,$C5,$9A,$FF,$14,$5F,$1F,$FC,$3F,$BF,$C1,$3F,$D2
	.dc.b	$03,$87,$FA,$5B,$5F,$F2,$23,$5F,$F4,$C4,$BE,$83,$F2,$1E,$5C,$9F
	.dc.b	$14,$79,$7C,$59,$1E,$57,$D1,$F1,$DF,$C8,$7E,$AB,$CF,$3F,$D6,$9A
	.dc.b	$FF,$94,$DA,$83,$8D,$D8,$BE,$3B,$F4,$5E,$64,$BF,$89,$1F,$1F,$E6
	.dc.b	$64,$7C,$77,$E5,$78,$3F,$CC,$FE,$A9,$D5,$FB,$B6,$D7,$F2,$D6,$BF
	.dc.b	$25,$AF,$F9,$A7,$2F,$E8,$5E,$87,$EC,$C9,$1F,$57,$E8,$7E,$C7,$AB
	.dc.b	$23,$EC,$3F,$6F,$E3,$FC,$5D,$DE,$12,$4F,$85,$FF,$00,$F8,$1F,$D2
	.dc.b	$A3,$FA,$77,$A3,$FC,$04,$FC,$27,$B1,$FD,$37,$C4,$7C,$45,$25,$7E
	.dc.b	$1F,$E2,$FF,$A9,$63,$A8,$FE,$94,$D7,$F0,$CD,$7F,$C1,$ED,$7F,$8B
	.dc.b	$7A,$BF,$C4,$48,$FF,$58,$BF,$E3,$AF,$F4,$2F,$A0,$FF,$03,$BF,$6F
	.dc.b	$E8,$5F,$3B,$FD,$8D,$B1,$FD,$82,$76,$A3,$2F,$E6,$5A,$FF,$88,$9A
	.dc.b	$FF,$A6,$DA,$FE,$26,$3F,$B4,$3D,$7D,$8F,$EE,$51,$88,$FE,$AF,$3D
	.dc.b	$B8,$FF,$4F,$F4,$3E,$A6,$FA,$1F,$DE,$63,$1E,$2F,$ED,$6F,$13,$FD
	.dc.b	$0E,$7E,$0C,$7F,$8D,$9A,$FF,$AB,$DA,$FE,$75,$E6,$FF,$3F,$F5,$3F
	.dc.b	$9E,$1D,$1F,$F0,$1C,$3E,$3B,$C4,$48,$FF,$0B,$FD,$0B,$E4,$31,$1F
	.dc.b	$BF,$AE,$0F,$F8,$1D,$AE,$6B,$FC,$9B,$58,$CB,$F8,$B0,$71,$97,$F9
	.dc.b	$31,$B4,$9D,$65,$FE,$59,$9C,$C7,$F8,$97,$D7,$FE,$AC,$67,$0D,$7B
	.dc.b	$5B,$F3,$E6,$3E,$A8,$AF,$F8,$BB,$04,$E3,$0F,$2C,$7F,$82,$E7,$FC
	.dc.b	$4F,$F4,$1F,$BA,$F7,$BF,$84,$7C,$2F,$DD,$8C,$4A,$31,$E1,$7F,$9C
	.dc.b	$5D,$7A,$BF,$EE,$37,$37,$CB,$97,$FD,$CD,$87,$29,$CE,$1C,$E7,$37
	.dc.b	$C5,$DE,$1D,$5B,$87,$3A,$FE,$39,$87,$39,$E3,$19,$11,$97,$75,$CD
	.dc.b	$B5,$B9,$FE,$F1,$16,$74,$F2,$6C,$73,$8E,$B7,$E7,$79,$9F,$D1,$63
	.dc.b	$E3,$BF,$E8,$E6,$BE,$A3,$FA,$57,$C1,$BE,$1D,$F4,$3F,$E5,$38,$C3
	.dc.b	$63,$33,$C1,$AF,$5E,$C7,$6F,$F9,$6F,$04,$62,$7C,$8D,$78,$4E,$E3
	.dc.b	$5A,$FD,$F6,$EB,$FD,$26,$9F,$28,$F2,$7F,$B9,$5E,$F0,$F9,$58,$70
	.dc.b	$BF,$AA,$D7,$8D,$70,$C3,$F0,$73,$9C,$7C,$F7,$C8,$FB,$58,$F7,$E7
	.dc.b	$C6,$3F,$87,$3A,$CE,$32,$FE,$25,$5B,$9D,$1F,$82,$C3,$B1,$2B,$AE
	.dc.b	$5E,$39,$E6,$75,$E2,$FF,$4A,$77,$7E,$EB,$2F,$F2,$67,$75,$AF,$F2
	.dc.b	$64,$DA,$FF,$4D,$4D,$F0,$D3,$4C,$7B,$97,$0F,$F4,$F7,$B7,$B3,$87
	.dc.b	$D0,$D8,$FF,$51,$37,$2F,$F2,$6B,$72,$FF,$09,$FB,$8F,$8C,$F2,$57
	.dc.b	$FA,$8D,$F1,$7B,$58,$9D,$61,$FE,$AD,$8C,$4E,$A3,$FD,$4F,$3A,$AF
	.dc.b	$F4,$8A,$F8,$CE,$47,$B7,$AA,$D7,$08,$B5,$FE,$62,$9B,$8F,$F2,$2D
	.dc.b	$BF,$C0,$B7,$B3,$FD,$7D,$3A,$7F,$1D,$F3,$EF,$A0,$FF,$14,$36,$3F
	.dc.b	$D6,$CD,$C3,$FC,$64,$D7,$FD,$F6,$E7,$0F,$F1,$BB,$67,$FD,$F7,$F3
	.dc.b	$FF,$EE,$67,$FD,$8B,$19,$AF,$E6,$53,$97,$FD,$AA,$DA,$F5,$72,$FD
	.dc.b	$F8,$C5,$F4,$F1,$F0,$7D,$2F,$A0,$7C,$AE,$27,$FD,$A7,$18,$6C,$73
	.dc.b	$F1,$B8,$3F,$89,$E0,$C4,$FF,$B8,$DB,$9F,$FA,$3A,$3F,$1C,$7E,$3F
	.dc.b	$E0,$EB,$FE,$E0,$F9,$FF,$9F,$3F,$EC,$A6,$EB,$FD,$53,$E6,$BF,$CC
	.dc.b	$75,$BE,$1F,$CD,$47,$C3,$93,$FE,$62,$9F,$F7,$DB,$93,$17,$FE,$FA
	.dc.b	$5F,$41,$C8,$E7,$F0,$16,$BC,$9C,$3F,$96,$CB,$F8,$AC,$74,$F8,$0C
	.dc.b	$4F,$8B,$E8,$3C,$7F,$03,$6B,$C2,$F9,$BE,$2B,$CF,$EF,$D9,$F2,$BF
	.dc.b	$A2,$F1,$7F,$63,$1F,$E6,$1D,$FF,$CF,$31,$8D,$FD,$1E,$1E,$FC,$2F
	.dc.b	$83,$F4,$1F,$0B,$E0,$2E,$3F,$01,$F8,$BE,$F7,$C6,$FA,$0B,$3F,$11
	.dc.b	$F0,$78,$B2,$3E,$3F,$BF,$73,$F8,$DF,$41,$F3,$BE,$03,$E7,$B6,$BC
	.dc.b	$C7,$F2,$FF,$A0,$47,$EF,$1B,$1A,$AF,$F3,$E7,$CF,$BE,$7F,$57,$2B
	.dc.b	$54,$BE,$2F,$F8,$3E,$3F,$9E,$5F,$17,$73,$AE,$2F,$F4,$FC,$E6,$B8
	.dc.b	$1F,$E8,$B9,$D5,$72,$C3,$C1,$3F,$83,$F8,$2F,$C5,$FA,$1F,$8C,$6D
	.dc.b	$7E,$29,$B3,$B8,$FC,$6C,$FE,$4A,$3F,$27,$E8,$59,$FC,$8B,$63,$F2
	.dc.b	$0D,$CB,$F3,$8F,$F3,$3F,$97,$FF,$1F,$57,$C7,$94,$7F,$8F,$63,$0E
	.dc.b	$AB,$DB,$3A,$F7,$F9,$7B,$BF,$EB,$B8,$E2,$DA,$E1,$5F,$E9,$96,$BF
	.dc.b	$8F,$7A,$4F,$F9,$09,$B7,$9F,$7B,$AB,$E2,$F7,$7D,$1F,$07,$E9,$6C
	.dc.b	$64,$F3,$C6,$5A,$7A,$BE,$83,$E0,$1F,$07,$E0,$2C,$64,$99,$7B,$B1
	.dc.b	$4E,$A3,$EE,$8F,$7D,$F2,$3F,$D3,$1B,$7F,$8C,$F9,$E7,$A3,$9A,$93
	.dc.b	$F9,$0A,$DD,$F4,$BD,$CE,$6B,$FA,$BC,$62,$7C,$35,$C3,$6F,$F7,$7C
	.dc.b	$E5,$FF,$78,$C6,$6F,$E7,$79,$FF,$7F,$97,$F0,$0D,$8F,$BA,$DC,$3E
	.dc.b	$03,$6F,$FD,$F7,$F3,$EF,$6F,$FB,$4A,$F4,$AC,$61,$B8,$7E,$B2,$74
	.dc.b	$D9,$D3,$F0,$7D,$AF,$90,$7F,$81,$7E,$7F,$F9,$3E,$1E,$A8,$CC,$EC
	.dc.b	$FC,$B4,$E9,$FF,$18,$68,$9E,$5F,$23,$C3,$4F,$CB,$DE,$DF,$F7,$F5
	.dc.b	$6E,$FF,$3B,$ED,$FE,$8F,$DC,$FD,$23,$63,$27,$E8,$A3,$2D,$3F,$4B
	.dc.b	$F3,$BF,$AB,$F8,$2F,$D6,$36,$32,$7E,$AA,$32,$D3,$F5,$BF,$33,$FB
	.dc.b	$33,$F6,$9F,$3D,$63,$F6,$2D,$CB,$F2,$8D,$DB,$F5,$1F,$29,$FB,$B3
	.dc.b	$F7,$9F,$2F,$6F,$F7,$2D,$BF,$CE,$36,$BF,$61,$F0,$DF,$C0,$CF,$E0
	.dc.b	$9F,$3F,$6F,$F8,$0B,$6F,$F4,$8D,$A4,$FF,$98,$67,$FD,$46,$3E,$7D
	.dc.b	$E5,$FF,$99,$69,$9F,$0A,$D4,$F1,$7F,$D5,$06,$0F,$EA,$91,$BA,$E7
	.dc.b	$1D,$FD,$33,$B3,$FD,$8D,$3B,$AE,$93,$AA,$E3,$5A,$C3,$15,$A8,$E5
	.dc.b	$87,$9A,$73,$1D,$CC,$E1,$D5,$C5,$F4,$AE,$11,$E3,$3E,$1C,$66,$78
	.dc.b	$47,$FB,$04,$E7,$5D,$9C,$4E,$E3,$11,$CA,$FE,$55,$72,$8C,$1F,$1A
	.dc.b	$76,$FC,$AA,$E4,$F8,$CC,$57,$29,$DE,$1E,$1C,$3D,$37,$F1,$AF,$A9
	.dc.b	$F1,$CE,$53,$BC,$3C,$51,$D0,$F1,$1A,$3D,$A5,$9F,$19,$D5,$E0,$75
	.dc.b	$71,$3D,$6F,$CB,$3C,$A7,$2D,$C3,$EC,$47,$46,$9E,$D8,$F0,$36,$38
	.dc.b	$9B,$8F,$B3,$19,$38,$BE,$A8,$DC,$66,$36,$65,$DB,$E9,$6E,$1F,$29
	.dc.b	$B8,$7C,$6C,$3D,$95,$F1,$78,$9F,$3F,$F9,$16,$C7,$E3,$DA,$FE,$1D
	.dc.b	$BF,$16,$E2,$65,$FC,$72,$FF,$DA,$53,$98,$C5,$FE,$9A,$30,$FF,$59
	.dc.b	$9D,$D7,$28,$C1,$B7,$19,$FE,$A3,$37,$95,$FE,$DF,$8C,$36,$39,$E3
	.dc.b	$F9,$BC,$7B,$D6,$B0,$FC,$F5,$72,$75,$87,$68,$E3,$1F,$BB,$74,$DC
	.dc.b	$3B,$3E,$5C,$38,$9D,$23,$31,$FC,$41,$B8,$78,$AB,$AE,$1E,$3A,$E8
	.dc.b	$72,$3F,$85,$4E,$DF,$E1,$B3,$E2,$9E,$07,$F0,$A9,$D4,$9B,$3B,$3F
	.dc.b	$8A,$C7,$8C,$ED,$3C,$70,$F8,$11,$DA,$FD,$35,$C6,$3D,$C7,$92,$76
	.dc.b	$F3,$3D,$26,$70,$EB,$3E,$03,$94,$6E,$FC,$13,$A8,$F0,$1F,$24,$C1
	.dc.b	$EB,$AE,$F5,$C8,$E7,$87,$C0,$3D,$F8,$7A,$0F,$9F,$3E,$68,$CD,$75
	.dc.b	$3C,$C4,$C3,$E5,$DF,$9E,$F8,$E1,$E2,$8F,$04,$62,$B8,$46,$DE,$95
	.dc.b	$EC,$72,$7B,$0F,$8A,$E4,$FE,$93,$5F,$E5,$EC,$7F,$AA,$6B,$FE,$53
	.dc.b	$CE,$DF,$EB,$3C,$2F,$F1,$F3,$FE,$62,$AC,$DF,$F9,$9A,$AD,$6A,$BF
	.dc.b	$C9,$11,$8A,$DF,$CE,$8B,$E0,$CA,$FF,$38,$46,$5B,$19,$7F,$CD,$F1
	.dc.b	$FE,$4E,$76,$F4,$F8,$DE,$BF,$33,$27,$9C,$57,$8C,$5C,$BF,$B8,$1E
	.dc.b	$39,$F2,$FC,$AF,$9B,$E5,$E5,$F9,$23,$2F,$9A,$32,$F9,$23,$DB,$F2
	.dc.b	$BE,$EF,$96,$93,$EC,$8C,$BE,$E8,$CB,$EC,$8F,$97,$F0,$4E,$9F,$A3
	.dc.b	$2F,$9E,$7C,$56,$1B,$1F,$E9,$7A,$E0,$FF,$AE,$A3,$2F,$F9,$D9,$BA
	.dc.b	$7F,$83,$E3,$A5,$78,$2F,$A3,$23,$93,$97,$FD,$8B,$BF,$FA,$09,$BD
	.dc.b	$DE,$68,$37,$CC,$93,$A2,$1F,$E8,$89,$E4,$FD,$CF,$23,$F0,$BE,$64
	.dc.b	$8F,$C1,$C7,$E2,$0C,$3F,$82,$8C,$3F,$ED,$8F,$9E,$7C,$17,$C8,$F9
	.dc.b	$E7,$BF,$4F,$97,$3A,$AF,$F6,$FC,$E9,$B3,$B7,$A5,$57,$FD,$3F,$3B
	.dc.b	$6C,$F4,$9F,$CE,$F9,$92,$7F,$36,$FE,$83,$E4,$F1,$1F,$9A,$8B,$C1
	.dc.b	$F9,$2E,$95,$7E,$9B,$DE,$FD,$3F,$C1,$4C,$3F,$4B,$19,$7F,$4F,$19
	.dc.b	$7F,$4B,$1F,$AE,$F8,$C9,$1F,$AC,$7F,$63,$F2,$78,$8F,$F2,$B6,$3F
	.dc.b	$B3,$9F,$F2,$84,$7F,$6A,$C5,$18,$9D,$4F,$F9,$55,$AF,$ED,$3D,$1F
	.dc.b	$DE,$9D,$63,$D9,$CB,$FC,$03,$E2,$65,$7E,$FA,$3F,$81,$98,$7F,$7B
	.dc.b	$18,$7F,$81,$F8,$7F,$C2,$3E,$3A,$46,$08,$7F,$04,$72,$FF,$06,$8F
	.dc.b	$D4,$E0,$F6,$AF,$9D,$2B,$FC,$CF,$97,$E8,$EB,$51,$8B,$FE,$AF,$0A
	.dc.b	$E1,$3F,$EE,$08,$C3,$B3,$FC,$E2,$F2,$75,$1C,$2B,$B3,$E4,$BF,$E5
	.dc.b	$3F,$36,$F5,$6F,$CE,$CC,$3A,$60,$AD,$47,$71,$E9,$7E,$EE,$7F,$78
	.dc.b	$DC,$FF,$B7,$B6,$7F,$C4,$B4,$F9,$4F,$0F,$C8,$78,$BE,$42,$57,$82
	.dc.b	$93,$E6,$F8,$EF,$3F,$8B,$2B,$51,$FC,$5B,$E7,$5A,$D2,$4F,$3F,$90
	.dc.b	$F7,$F6,$FC,$0F,$90,$93,$DE,$3E,$1F,$7F,$C4,$F9,$09,$3E,$38,$FE
	.dc.b	$47,$F3,$BF,$94,$8F,$84,$D7,$FC,$83,$C5,$FE,$0E,$8C,$71,$FF,$3E
	.dc.b	$E0,$7C,$8F,$D0,$F9,$19,$87,$BD,$FA,$7F,$23,$F5,$3E,$47,$EA,$FB
	.dc.b	$32,$BE,$24,$78,$BB,$A4,$78,$9F,$B7,$F2,$3F,$73,$DB,$93,$F2,$DF
	.dc.b	$BF,$F1,$FF,$80,$F8,$FF,$E7,$BE,$43,$E3,$29,$9B,$FB,$D8,$E5,$B1
	.dc.b	$FD,$12,$31,$1F,$00,$FF,$60,$CE,$38,$3F,$9F,$2F,$8D,$EE,$49,$E6
	.dc.b	$7E,$2F,$E7,$7F,$CD,$B5,$FD,$15,$BC,$9F,$E3,$86,$CF,$F9,$BB,$E2
	.dc.b	$5A,$7F,$56,$F9,$DF,$E3,$33,$F9,$56,$BE,$50,$E3,$0E,$DE,$FF,$1F
	.dc.b	$3F,$F8,$8D,$69,$B5,$CA,$3F,$DD,$8D,$7F,$73,$CF,$FC,$6B,$3F,$F2
	.dc.b	$6F,$CE,$FB,$E7,$F4,$CD,$7E,$36,$3F,$AB,$7F,$B6,$C6,$63,$FB,$23
	.dc.b	$67,$9E,$BF,$B6,$AD,$B1,$EA,$E5,$FD,$BF,$CE,$BE,$4F,$AF,$C9,$C9
	.dc.b	$CC,$57,$E2,$60,$9C,$63,$FE,$14,$BE,$55,$D3,$0F,$F1,$7C,$6E,$3B
	.dc.b	$D6,$27,$2C,$8C,$9C,$4F,$2F,$B5,$C8,$8F,$A3,$92,$BF,$1A,$FF,$8D
	.dc.b	$61,$E3,$F3,$32,$FF,$82,$7E,$73,$FA,$2F,$C2,$AD,$F1,$5F,$E1,$25
	.dc.b	$3A,$73,$EF,$FF,$A0,$A3,$13,$B5,$F0,$BF,$E8,$68,$C7,$AD,$FE,$50
	.dc.b	$6E,$1D,$DB,$87,$43,$C5,$7C,$9D,$61,$DE,$F9,$36,$F3,$3C,$27,$2F
	.dc.b	$2C,$3C,$55,$B8,$EC,$7B,$27,$6D,$AE,$6F,$C0,$AE,$2F,$58,$CD,$FC
	.dc.b	$0B,$E9,$87,$C1,$BD,$DF,$33,$FD,$4E,$71,$29,$93,$35,$F3,$5D,$FC
	.dc.b	$4B,$E3,$5F,$3B,$FD,$F2,$30,$D8,$CC,$F4,$9C,$BD,$A9,$DC,$EF,$FF
	.dc.b	$02,$C6,$1B,$19,$9E,$0D,$7C,$AE,$72,$FB,$CD,$CF,$FD,$87,$D8,$F2
	.dc.b	$5D,$3F,$1A,$B5,$87,$DC,$6B,$F8,$C6,$D7,$19,$E1,$8C,$AF,$AF,$97
	.dc.b	$9D,$FB,$35,$B7,$37,$D2,$75,$97,$9A,$FE,$DD,$EE,$31,$7C,$27,$B4
	.dc.b	$7E,$7E,$75,$97,$E7,$70,$F4,$F0,$BF,$61,$7B,$DF,$E9,$0D,$3F,$C2
	.dc.b	$31,$87,$FB,$FA,$F0,$B5,$A2,$F9,$5E,$1B,$7C,$23,$17,$C6,$30,$74
	.dc.b	$F7,$78,$73,$B9,$77,$18,$6C,$FF,$1A,$6E,$3F,$C0,$DF,$1A,$FA,$B7
	.dc.b	$C1,$F8,$FE,$B3,$FC,$E3,$7F,$E1,$59,$CB,$FE,$1A,$8C,$D7,$F4,$39
	.dc.b	$CB,$FC,$D2,$33,$A7,$FA,$56,$72,$FF,$AD,$A3,$35,$FC,$3E,$9F,$3F
	.dc.b	$FE,$1B,$8F,$F1,$3F,$D2,$3E,$6B,$CF,$39,$AF,$E8,$33,$9A,$FF,$15
	.dc.b	$4E,$70,$FF,$61,$CE,$6B,$FA,$C4,$E6,$BF,$C8,$F3,$9A,$FF,$55,$4F
	.dc.b	$5A,$FD,$BB,$5F,$33,$5F,$E4,$CD,$8F,$C2,$7C,$07,$D4,$C3,$EC,$B7
	.dc.b	$0F,$F1,$F7,$90,$8C,$C6,$1B,$87,$F6,$D9,$D4,$E4,$FC,$5F,$77,$D8
	.dc.b	$8C,$CF,$F6,$2F,$BD,$B8,$FF,$4F,$F9,$AF,$C6,$46,$30,$FF,$32,$F1
	.dc.b	$27,$96,$3F,$91,$FB,$C7,$9F,$8B,$FF,$2B,$C6,$1B,$D5,$9C,$FF,$B8
	.dc.b	$CF,$F8,$22,$7D,$B7,$F8,$CF,$B1,$FF,$24,$C7,$F9,$83,$F0,$76,$D9
	.dc.b	$F7,$3F,$41,$18,$9F,$F3,$15,$3D,$EF,$D2,$7C,$98,$E8,$E1,$ED,$EC
	.dc.b	$EB,$D2,$B7,$FE,$6E,$9F,$F5,$8B,$FE,$6D,$F9,$E7,$CA,$FE,$81,$B1
	.dc.b	$FE,$74,$FC,$3D,$F3,$59,$AF,$D2,$CF,$F3,$13,$FC,$F3,$F8,$0F,$FB
	.dc.b	$E1,$FF,$3C,$19,$E2,$FF,$09,$C6,$5F,$F1,$36,$1E,$99,$F1,$63,$FB
	.dc.b	$FF,$C2,$FF,$C1,$91,$83,$F8,$3F,$C6,$FF,$00,$F8,$FF,$E0,$2D,$8F
	.dc.b	$F4,$37,$DA,$FB,$31,$FE,$F9,$F4,$5F,$BA,$8C,$76,$FF,$83,$E3,$F8
	.dc.b	$67,$B5,$FE,$93,$BF,$F0,$DF,$D0,$63,$E4,$3F,$84,$BA,$F8,$EF,$E1
	.dc.b	$FF,$2F,$6B,$F8,$CB,$D7,$CD,$FC,$5F,$C9,$D9,$FF,$4D,$F8,$9E,$FD
	.dc.b	$3F,$9F,$ED,$FC,$77,$4F,$B5,$5F,$E2,$AF,$A1,$F1,$F5,$FF,$18,$7F
	.dc.b	$88,$FD,$8F,$EC,$EF,$F9,$27,$A1,$FE,$08,$BF,$F5,$24,$E7,$1F,$E6
	.dc.b	$1E,$DF,$0F,$93,$FE,$4A,$DA,$FE,$5E,$DC,$3F,$C8,$B8,$73,$7F,$CB
	.dc.b	$18,$73,$8C,$61,$E8,$3F,$C7,$DF,$1D,$F2,$0F,$93,$F1,$F7,$1F,$F5
	.dc.b	$CE,$BF,$C6,$31,$F7,$56,$A7,$2F,$D0,$F8,$F9,$1C,$3D,$CB,$5F,$E5
	.dc.b	$8B,$56,$31,$F1,$47,$F5,$B6,$C7,$DA,$F8,$D9,$87,$8B,$C4,$B8,$73
	.dc.b	$9B,$7C,$A6,$D7,$FB,$93,$05,$62,$FE,$B5,$5B,$FF,$2F,$7C,$6F,$F5
	.dc.b	$0F,$0B,$F1,$2D,$C7,$E0,$B6,$FD,$4D,$AF,$4F,$9B,$F9,$03,$F2,$3F
	.dc.b	$1B,$6F,$E4,$C5,$BF,$85,$36,$BF,$13,$0E,$1F,$C4,$BF,$E7,$EF,$D8
	.dc.b	$1C,$3F,$88,$E6,$FE,$D1,$18,$CB,$F0,$F1,$89,$D4,$7F,$BC,$27,$55
	.dc.b	$F8,$2A,$D4,$F0,$35,$5F,$DC,$FB,$5F,$2D,$4C,$7C,$B5,$AF,$F7,$DC
	.dc.b	$E9,$B3,$A6,$9F,$F0,$03,$F6,$27,$89,$D7,$CF,$FF,$00,$46,$27,$9C
	.dc.b	$62,$75,$1C,$5B,$9F,$E6,$B9,$BF,$C1,$1F,$1F,$FA,$98,$F9,$11,$8E
	.dc.b	$CF,$D6,$B6,$FF,$56,$D9,$E5,$1F,$E7,$FE,$1F,$D1,$5F,$3E,$1F,$CA
	.dc.b	$CE,$4D,$47,$7F,$53,$D1,$7F,$EB,$2F,$96,$7D,$1F,$F8,$32,$73,$8F
	.dc.b	$EB,$67,$38,$73,$BF,$F4,$75,$6A,$3C,$53,$A9,$FF,$67,$4E,$5F,$F1
	.dc.b	$44,$66,$BF,$71,$B3,$E7,$3E,$5F,$D0,$7C,$C6,$E3,$F4,$5B,$87,$FB
	.dc.b	$22,$B8,$4F,$D5,$AF,$E4,$FF,$2B,$FE,$2C,$F8,$BB,$3F,$EC,$26,$DF
	.dc.b	$FA,$AD,$B8,$FF,$AF,$5F,$E5,$F1,$FE,$38,$F3,$ED,$7F,$90,$9B,$1F
	.dc.b	$C5,$1B,$5F,$C4,$7C,$3F,$E8,$D3,$FE,$DC,$FD,$8B,$FD,$FD,$7F,$ED
	.dc.b	$2D,$FF,$C8,$D1,$89,$FF,$5E,$46,$3E,$4B,$FC,$86,$DA,$E0,$7C,$F9
	.dc.b	$FF,$59,$0F,$A4,$78,$1F,$E3,$1C,$BF,$B0,$D7,$FB,$06,$73,$5F,$E9
	.dc.b	$8F,$8B,$C5,$EF,$D6,$FE,$7F,$5A,$AE,$AD,$C3,$9C,$77,$AF,$F4,$FD
	.dc.b	$F2,$8F,$C8,$5F,$29,$CC,$6C,$CE,$1B,$7A,$C7,$86,$31,$1F,$EC,$93
	.dc.b	$81,$FE,$6D,$9E,$8F,$A4,$CD,$F9,$30,$F4,$47,$17,$D1,$3C,$23,$B9
	.dc.b	$E2,$3A,$9B,$BF,$13,$83,$FD,$CE,$76,$8E,$4E,$CE,$53,$E0,$76,$F8
	.dc.b	$2F,$D7,$18,$AE,$B3,$BC,$3E,$16,$1E,$99,$F8,$F3,$C6,$BC,$55,$AC
	.dc.b	$3C,$31,$CC,$F0,$D7,$03,$A9,$5C,$9F,$1E,$BA,$1E,$17,$FC,$73,$1F
	.dc.b	$03,$AB,$07,$84,$D4,$F0,$8E,$93,$F4,$F1,$F6,$E1,$E5,$6C,$7C,$76
	.dc.b	$D7,$18,$F1,$57,$99,$A7,$98,$FA,$A7,$53,$E0,$9C,$CB,$1C,$DD,$E3
	.dc.b	$F2,$5E,$67,$18,$F7,$7C,$3F,$F9,$5A,$3F,$10,$D7,$F5,$2D,$7F,$20
	.dc.b	$E7,$E3,$BF,$90,$D2,$7F,$D6,$99,$7F,$A7,$67,$31,$83,$4E,$2B,$96
	.dc.b	$9F,$EC,$59,$D0,$9E,$38,$7E,$BF,$17,$17,$FA,$A6,$30,$D8,$F0,$5F
	.dc.b	$39,$FD,$B5,$72,$9F,$F5,$81,$CA,$FF,$5C,$D7,$F5,$71,$B7,$CF,$5C
	.dc.b	$E7,$8D,$F7,$9F,$81,$1C,$6F,$BD,$6E,$78,$E5,$DA,$F8,$47,$F1,$26
	.dc.b	$E1,$DE,$3E,$05,$71,$74,$71,$3F,$82,$46,$CE,$18,$78,$4F,$2C,$6A
	.dc.b	$79,$4F,$D2,$CB,$E3,$CF,$D7,$6B,$F9,$E3,$CE,$53,$C7,$87,$85,$FE
	.dc.b	$1F,$7D,$DD,$E3,$F0,$27,$4F,$43,$11,$F3,$30,$F6,$57,$18,$F6,$99
	.dc.b	$AF,$41,$CA,$3D,$71,$92,$57,$58,$CB,$D8,$8E,$09,$1E,$67,$F0,$B3
	.dc.b	$D6,$3E,$CE,$3F,$4A,$FE,$39,$E6,$7E,$BD,$79,$F1,$FA,$13,$9A,$E4
	.dc.b	$7C,$63,$CB,$3E,$07,$DB,$87,$91,$F1,$1E,$03,$8D,$7E,$3F,$06,$EF
	.dc.b	$91,$FE,$AD,$AF,$FB,$6B,$E8,$2E,$0B,$7C,$1A,$73,$1C,$AE,$E1,$CE
	.dc.b	$FF,$C2,$53,$C2,$B7,$7C,$67,$35,$B7,$FB,$A7,$C9,$F4,$E4,$7D,$1D
	.dc.b	$3A,$C4,$9E,$B3,$A6,$CE,$5F,$EF,$DD,$9E,$0E,$59,$1E,$08,$F3,$98
	.dc.b	$7D,$DA,$FF,$BE,$A7,$31,$EA,$FD,$91,$E7,$FC,$05,$E8,$53,$11,$27
	.dc.b	$84,$65,$E5,$1F,$31,$7E,$C9,$FE,$88,$7E,$3F,$A4,$F5,$BE,$6B,$63
	.dc.b	$FC,$1D,$3B,$7C,$E7,$F2,$37,$1C,$5F,$EA,$F9,$E1,$3A,$84,$61,$FF
	.dc.b	$41,$CF,$0C,$FF,$96,$53,$E9,$1E,$C7,$9F,$E9,$3F,$C2,$15,$A6,$D7
	.dc.b	$F3,$3A,$DB,$FE,$23,$6B,$E6,$11,$C1,$BD,$3F,$0E,$77,$8F,$FA,$59
	.dc.b	$B7,$DE,$7C,$46,$F0,$FA,$5F,$13,$F4,$FD,$39,$8F,$D7,$8C,$BF,$66
	.dc.b	$32,$FC,$3E,$04,$FE,$6B,$47,$AF,$49,$5F,$81,$29,$F2,$E3,$6F,$E7
	.dc.b	$9A,$F5,$3D,$9E,$2F,$C9,$F2,$E4,$7C,$58,$CB,$FE,$49,$DF,$FD,$AE
	.dc.b	$D9,$CB,$F2,$76,$57,$28,$F9,$95,$C1,$D4,$7E,$CB,$DC,$91,$FA,$E7
	.dc.b	$F6,$BE,$EE,$23,$FC,$C5,$17,$87,$FD,$ED,$D6,$AF,$DD,$F9,$1F,$B3
	.dc.b	$C2,$61,$FB,$88,$CB,$FB,$28,$CB,$FE,$6D,$8F,$E0,$7D,$D2,$3F,$7E
	.dc.b	$FF,$06,$F0,$B1,$1F,$E7,$38,$B8,$F8,$A6,$F2,$FF,$AA,$72,$FE,$11
	.dc.b	$7B,$3F,$84,$CF,$F1,$0F,$4A,$47,$F0,$C7,$F8,$9F,$5E,$23,$FD,$01
	.dc.b	$17,$0F,$7C,$FF,$18,$F9,$49,$5F,$C5,$1F,$E3,$7E,$96,$23,$FD,$11
	.dc.b	$16,$BE,$3C,$FE,$67,$E1,$7F,$8C,$7D,$2C,$BF,$D1,$46,$5F,$E2,$F1
	.dc.b	$97,$EE,$46,$5F,$DB,$78,$49,$D5,$FE,$8E,$BF,$96,$F9,$5F,$CA,$B3
	.dc.b	$98,$7F,$2A,$8F,$E1,$9F,$7A,$F5,$BF,$9A,$FC,$67,$FB,$F2,$BF,$99
	.dc.b	$B5,$FD,$A4,$ED,$FE,$6F,$F0,$1F,$C1,$9F,$E5,$BE,$9F,$F3,$33,$F9
	.dc.b	$FC,$66,$3F,$86,$2F,$4B,$F9,$EE,$F2,$79,$C7,$93,$D7,$FE,$83,$C3
	.dc.b	$27,$FD,$51,$8A,$7A,$F0,$7F,$41,$F9,$39,$5E,$48,$FE,$51,$F6,$5E
	.dc.b	$9E,$27,$3F,$F1,$88,$C3,$7D,$BF,$EA,$BF,$39,$ED,$C3,$F8,$97,$E0
	.dc.b	$9F,$B6,$C3,$8C,$7F,$92,$FA,$3F,$DB,$CF,$FA,$ED,$AF,$0F,$13,$C3
	.dc.b	$E6,$C8,$FF,$60,$67,$FD,$09,$FE,$BF,$F2,$BD,$F0,$EB,$0E,$97,$BD
	.dc.b	$FD,$07,$EF,$99,$D3,$FA,$24,$EA,$BF,$D9,$8F,$F0,$07,$FD,$9D,$2F
	.dc.b	$13,$91,$C3,$C3,$E3,$F3,$3D,$7D,$49,$7C,$A3,$F8,$F7,$B5,$FD,$47
	.dc.b	$F0,$37,$CD,$D3,$35,$FE,$DB,$8C,$BF,$C7,$7C,$2F,$E4,$C6,$DF,$C5
	.dc.b	$F9,$9F,$DF,$3E,$B2,$46,$0E,$5F,$17,$DC,$8F,$E3,$BB,$2B,$F8,$AF
	.dc.b	$1B,$F1,$A6,$39,$3E,$A8,$F0,$D1,$FE,$0A,$D6,$57,$E6,$23,$FB,$67
	.dc.b	$D1,$EA,$3F,$AD,$9F,$D7,$1C,$38,$0E,$1D,$B5,$E1,$1F,$DD,$FE,$91
	.dc.b	$E1,$D3,$BC,$49,$FF,$4A,$63,$F8,$0A,$DC,$63,$C0,$FE,$DB,$18,$8F
	.dc.b	$EC,$98,$79,$2B,$F0,$FF,$74,$F8,$FF,$F0,$37,$C6,$FF,$67,$3F,$B4
	.dc.b	$46,$6F,$A1,$FE,$A3,$70,$FF,$6F,$8E,$1E,$CF,$F8,$3E,$73,$7F,$33
	.dc.b	$67,$37,$F8,$BF,$59,$8F,$F0,$28,$FD,$F9,$87,$F8,$AC,$61,$FE,$0F
	.dc.b	$F5,$7F,$C1,$08,$FF,$0A,$FA,$0A,$FF,$06,$8B,$87,$86,$7F,$C4,$5F
	.dc.b	$44,$F9,$C7,$47,$FA,$F7,$E2,$3E,$34,$E6,$31,$C9,$FE,$16,$8C,$64
	.dc.b	$AD,$FB,$FF,$E1,$A7,$8E,$1D,$27,$57,$C6,$31,$3A,$C3,$AD,$7F,$B3
	.dc.b	$70,$FE,$3D,$39,$8E,$26,$6B,$8D,$F3,$BE,$32,$BE,$33,$98,$E2,$66
	.dc.b	$B8,$BC,$EB,$8D,$70,$8C,$1C,$1F,$7D,$F8,$CF,$7C,$70,$AF,$73,$FA
	.dc.b	$09,$F8,$AD,$8C,$BE,$33,$F4,$BF,$0C,$F6,$1F,$47,$FA,$59,$FF,$72
	.dc.b	$46,$1F,$D4,$4E,$E7,$4D,$8F,$D6,$5F,$F0,$CF,$A3,$E4,$D8,$CF,$80
	.dc.b	$9C,$B6,$33,$1F,$0A,$31,$97,$E9,$FD,$FF,$F3,$8F,$5E,$6A,$5F,$C5
	.dc.b	$BD,$ED,$FE,$5A,$BD,$CE,$6B,$78,$7E,$32,$F7,$23,$FC,$87,$F0,$9F
	.dc.b	$D1,$27,$55,$C3,$0F,$C8,$5F,$0B,$DE,$39,$7F,$11,$19,$CB,$E0,$61
	.dc.b	$E9,$C7,$A6,$1C,$B1,$FC,$7F,$13,$D9,$FF,$40,$67,$FE,$9C,$F8,$4F
	.dc.b	$E4,$2F,$F0,$C8,$FD,$BB,$93,$FD,$43,$CF,$AD,$7F,$D6,$D3,$C9,$FF
	.dc.b	$05,$C6,$6F,$B5,$DA,$72,$E7,$E1,$FF,$63,$8F,$FA,$B2,$31,$CB,$B6
	.dc.b	$47,$FA,$C3,$9E,$E7,$FE,$7D,$FA,$37,$85,$FD,$82,$3F,$D1,$71,$8A
	.dc.b	$FC,$63,$63,$F9,$93,$5F,$E6,$EE,$74,$FE,$D2,$D9,$FE,$5B,$F7,$BF
	.dc.b	$ED,$77,$F9,$0D,$27,$4F,$DB,$FA,$3F,$F0,$64,$FF,$B4,$BE,$91,$F4
	.dc.b	$9F,$1E,$95,$A7,$FD,$4D,$3A,$AE,$B3,$99,$FE,$96,$D7,$FD,$4E,$D7
	.dc.b	$C1,$ED,$7C,$16,$CF,$28,$FA,$5F,$6A,$F9,$9F,$15,$F8,$74,$FE,$BF
	.dc.b	$C4,$FA,$5F,$AC,$79,$FC,$AC,$65,$FE,$E0,$8C,$1D,$F3,$4E,$9F,$EB
	.dc.b	$CD,$7F,$B1,$B5,$FC,$67,$83,$CA,$BF,$DD,$6D,$8F,$F1,$ED,$7E,$27
	.dc.b	$C4,$E9,$8F,$F7,$4F,$82,$7C,$66,$6B,$CD,$87,$8F,$2F,$E7,$99,$7F
	.dc.b	$0B,$6E,$5E,$49,$F8,$D8,$B1,$FF,$59,$DF,$F9,$5B,$C1,$FC,$B7,$83
	.dc.b	$F9,$99,$CE,$1F,$EC,$BF,$A3,$7D,$2F,$DB,$6E,$1F,$D9,$31,$7D,$2F
	.dc.b	$E7,$2F,$85,$EF,$E6,$FE,$46,$06,$93,$0F,$E4,$50,$F3,$7B,$1F,$AB
	.dc.b	$F8,$5F,$D7,$65,$FE,$3E,$9D,$57,$E5,$3E,$8F,$EF,$37,$9B,$F6,$5F
	.dc.b	$4B,$FE,$40,$8C,$1F,$9A,$3F,$99,$1F,$CC,$4B,$A7,$F8,$5D,$AF,$F8
	.dc.b	$71,$AF,$FB,$E2,$BF,$CE,$51,$E1,$C7,$F8,$07,$D2,$7F,$B7,$23,$07
	.dc.b	$C5,$E7,$7D,$27,$F0,$3F,$80,$FE,$08,$D9,$FD,$6F,$D2,$7F,$8C,$1F
	.dc.b	$E8,$8D,$CB,$FC,$54,$D7,$FC,$5E,$D7,$FC,$47,$F0,$BF,$E9,$39,$FA
	.dc.b	$8D,$86,$FF,$E3,$E8,$FB,$31,$89,$E3,$F1,$BF,$C6,$67,$FC,$49,$1E
	.dc.b	$9F,$96,$C7,$B1,$FC,$6D,$FD,$E6,$3F,$C8,$70,$F1,$F9,$7F,$A5,$8C
	.dc.b	$47,$F3,$F0,$E1,$F1,$E1,$F8,$3C,$3C,$FC,$BE,$1F,$9E,$F1,$3F,$C2
	.dc.b	$7E,$83,$F7,$13,$FC,$D7,$D7,$C7,$8B,$FC,$B5,$B1,$FD,$AE,$7F,$83
	.dc.b	$62,$69,$2B,$71,$F0,$4F,$1F,$D5,$E3,$83,$E3,$7C,$57,$C3,$3E,$27
	.dc.b	$C6,$DC,$7E,$6C,$5B,$E5,$4D,$57,$98,$FA,$1F,$05,$F3,$4F,$9D,$EB
	.dc.b	$5C,$7E,$BC,$5C,$39,$E1,$FD,$B2,$F8,$D7,$DC,$FA,$4F,$AC,$7D,$7E
	.dc.b	$FB,$19,$25,$F2,$8C,$BD,$26,$D7,$E0,$E6,$D7,$E2,$BD,$FF,$EB,$1D
	.dc.b	$FF,$83,$6C,$7E,$31,$C3,$F8,$EF,$AA,$B8,$FA,$3C,$AF,$C5,$1F,$8B
	.dc.b	$FA,$5B,$5F,$94,$8B,$3E,$E9,$B9,$7F,$AE,$B8,$FF,$09,$3F,$33,$E9
	.dc.b	$7E,$5E,$1F,$D4,$A7,$35,$FE,$98,$9C,$D7,$DB,$3F,$01,$CD,$FC,$DE
	.dc.b	$B9,$C7,$FA,$59,$AF,$E5,$5A,$FF,$85,$5E,$B7,$99,$FE,$F9,$4E,$3F
	.dc.b	$CA,$47,$F7,$FE,$C7,$CA,$53,$1F,$1B,$70,$FE,$4D,$3A,$6C,$E9,$A7
	.dc.b	$F2,$BF,$A4,$FF,$6C,$BC,$3C,$EF,$F0,$64,$63,$4F,$C7,$DF,$F5,$2A
	.dc.b	$E1,$E1,$68,$F2,$7B,$56,$3E,$47,$47,$F8,$57,$E9,$7F,$57,$87,$BE
	.dc.b	$31,$DB,$FA,$F7,$E4,$12,$BF,$58,$D9,$F9,$DC,$28,$E9,$E1,$F1,$AF
	.dc.b	$F0,$3F,$D0,$79,$F9,$3F,$D9,$BE,$0B,$E6,$7F,$C0,$93,$9A,$FF,$5D
	.dc.b	$CE,$5E,$3B,$68,$FF,$16,$43,$E6,$F1,$3F,$1F,$E6,$BE,$2F,$C8,$FC
	.dc.b	$66,$C6,$49,$B7,$ED,$B8,$DF,$33,$B9,$FE,$4F,$F4,$9F,$D9,$BC,$33
	.dc.b	$49,$1F,$ED,$BD,$FF,$7D,$1F,$C9,$CF,$E5,$1E,$B5,$9F,$E6,$38,$FF
	.dc.b	$90,$7C,$FF,$E5,$8D,$8C,$91,$FD,$FF,$C2,$EA,$3F,$B6,$7C,$E3,$E9
	.dc.b	$7F,$BC,$B6,$3C,$4D,$AF,$F7,$8E,$48,$C4,$7F,$31,$8C,$57,$F9,$56
	.dc.b	$78,$47,$FC,$00,$7F,$AC,$4F,$F1,$A7,$35,$FE,$F9,$F8,$9F,$F3,$C7
	.dc.b	$6B,$C6,$E9,$F3,$0A,$CC,$6F,$0F,$F1,$2B,$A8,$CC,$7F,$43,$3F,$D9
	.dc.b	$43,$E9,$7A,$3F,$EF,$CA,$E8,$E6,$BA,$4E,$A3,$FA,$DC,$EA,$31,$1F
	.dc.b	$ED,$B3,$81,$CE,$BF,$D2,$11,$A8,$F0,$BC,$F2,$FF,$03,$47,$03,$FD
	.dc.b	$B8,$57,$31,$C7,$1E,$78,$7B,$6B,$81,$E1,$C7,$A6,$3E,$A7,$84,$7C
	.dc.b	$43,$C8,$53,$BD,$F0,$C7,$AC,$7C,$7C,$2B,$F0,$22,$BE,$F9,$E0,$F8
	.dc.b	$8F,$6E,$1E,$58,$F0,$DE,$8B,$1E,$97,$CB,$1C,$8B,$1C,$4F,$A2,$59
	.dc.b	$F2,$1F,$17,$0F,$A5,$1F,$81,$AE,$CD,$3B,$1C,$A7,$E0,$D7,$66,$C6
	.dc.b	$5F,$71,$EB,$C3,$E0,$E3,$F4,$A7,$4D,$9F,$1D,$7A,$67,$C1,$1F,$3A
	.dc.b	$FD,$ED,$8E,$38,$7A,$23,$8C,$7B,$5F,$C0,$CF,$A9,$F7,$B5,$CC,$7B
	.dc.b	$1C,$B5,$FC,$4F,$7F,$FA,$19,$F8,$ED,$73,$3F,$A7,$8C,$7C,$6F,$F2
	.dc.b	$5C,$7F,$DC,$D1,$8E,$BF,$F1,$CD,$A3,$13,$AB,$FF,$17,$5F,$1A,$D4
	.dc.b	$62,$F8,$BF,$A9,$C3,$F8,$EC,$FE,$D0,$FD,$E7,$17,$F8,$DE,$7F,$5D
	.dc.b	$7C,$DB,$7C,$AF,$FC,$78,$72,$BF,$CD,$36,$77,$87,$5B,$F2,$67,$E1
	.dc.b	$C7,$FC,$88,$DC,$3F,$3B,$3A,$C3,$D1,$7E,$57,$A1,$E0,$BF,$33,$AA
	.dc.b	$FD,$55,$71,$47,$B0,$FD,$4C,$79,$EF,$CB,$1F,$BF,$9E,$0E,$CF,$E0
	.dc.b	$53,$B3,$B4,$F0,$3D,$11,$AA,$F5,$CE,$8F,$26,$1A,$23,$83,$53,$E6
	.dc.b	$7B,$36,$79,$BE,$AB,$F2,$9E,$69,$DC,$79,$EF,$C9,$1D,$63,$91,$F2
	.dc.b	$6B,$E3,$BC,$F1,$FB,$4D,$C7,$E2,$CF,$82,$3A,$57,$99,$61,$E1,$9D
	.dc.b	$4F,$8A,$3C,$B7,$D2,$BE,$56,$3F,$01,$EC,$7D,$68,$FC,$DE,$3F,$28
	.dc.b	$97,$DE,$FB,$57,$90,$FC,$2D,$70,$9F,$21,$F0,$8F,$13,$98,$EF,$7E
	.dc.b	$17,$C7,$3F,$17,$1F,$EC,$DF,$4B,$FD,$97,$6F,$C3,$C6,$2B,$FB,$8C
	.dc.b	$E5,$FF,$54,$EE,$F9,$EC,$56,$E7,$FB,$8C,$AB,$FA,$F4,$74,$8C,$3F
	.dc.b	$DE,$23,$2F,$4F,$99,$E7,$26,$98,$FA,$8B,$3F,$EC,$BF,$A6,$F0,$FB
	.dc.b	$47,$31,$FD,$E7,$C1,$E6,$47,$FC,$17,$F1,$35,$F3,$C7,$8B,$3F,$F0
	.dc.b	$43,$72,$39,$FA,$67,$AB,$F0,$1B,$7F,$F0,$05,$AF,$FA,$89,$FF,$50
	.dc.b	$B9,$8E,$1E,$74,$8C,$1F,$43,$83,$FD,$8D,$F4,$CF,$5B,$E8,$4A,$30
	.dc.b	$EC,$FD,$EB,$87,$FD,$63,$EC,$BE,$5F,$8C,$E7,$5F,$F0,$58,$97,$B9
	.dc.b	$F6,$9B,$7D,$4F,$06,$1E,$2F,$A6,$7B,$D4,$93,$BF,$A6,$7B,$38,$8F
	.dc.b	$C0,$D7,$07,$EF,$B5,$CD,$7E,$09,$46,$72,$FF,$10,$C2,$FF,$C2,$DA
	.dc.b	$FF,$BA,$DB,$1F,$C7,$AB,$84,$FE,$6B,$81,$E3,$7D,$72,$57,$D7,$F6
	.dc.b	$6B,$F5,$F3,$FB,$86,$32,$F8,$FB,$FF,$B6,$E7,$E8,$47,$D3,$F2,$3F
	.dc.b	$43,$39,$9F,$C0,$18,$88,$F0,$8D,$4F,$D2,$C7,$FC,$AF,$97,$F3,$6A
	.dc.b	$F8,$3E,$67,$CF,$23,$F3,$77,$AF,$CC,$8F,$91,$A9,$A4,$8F,$A9,$E8
	.dc.b	$FE,$2B,$7F,$F4,$B4,$7D,$DA,$E0,$FF,$9B,$BE,$83,$F6,$24,$7F,$CE
	.dc.b	$3F,$4D,$5F,$F3,$24,$5D,$3F,$CD,$13,$FB,$8F,$9B,$95,$FE,$78,$7F
	.dc.b	$77,$F4,$F8,$85,$7E,$4B,$BB,$C3,$18,$DB,$F8,$8C,$62,$7F,$CE,$97
	.dc.b	$FC,$13,$D2,$97,$FE,$84,$7F,$83,$F8,$D8,$8F,$F3,$C4,$5A,$FF,$3E
	.dc.b	$4D,$C3,$F8,$94,$DA,$FE,$37,$3F,$C4,$3D,$B9,$1F,$E9,$07,$F8,$9F
	.dc.b	$99,$88,$FF,$42,$45,$C3,$FD,$11,$3F,$C6,$3E,$6A,$57,$FA,$61,$FE
	.dc.b	$37,$EE,$62,$3F,$D1,$91,$6B,$FD,$21,$3F,$AD,$EE,$95,$F5,$5F,$E4
	.dc.b	$7F,$43,$88,$FD,$5C,$7E,$00,$8F,$EB,$1F,$E4,$7E,$A4,$8E,$47,$FA
	.dc.b	$AB,$D0,$FE,$5A,$47,$FD,$5B,$E6,$FF,$0E,$FA,$77,$BD,$FB,$36,$CF
	.dc.b	$F3,$66,$BF,$CF,$1A,$FF,$AB,$7E,$4B,$F8,$63,$FA,$6E,$47,$B3,$FC
	.dc.b	$D8,$FE,$8F,$3A,$9F,$C0,$1F,$84,$5E,$C6,$89,$1F,$B0,$87,$C1,$F9
	.dc.b	$FD,$FF,$31,$27,$CC,$FF,$3B,$9F,$F6,$EC,$78,$FE,$63,$FA,$39,$E2
	.dc.b	$DB,$F9,$6F,$D3,$7F,$36,$9F,$18,$71,$C7,$FC,$49,$BD,$3F,$D9,$BD
	.dc.b	$5F,$87,$3F,$C5,$BE,$9F,$F8,$15,$FF,$AF,$5A,$FE,$F5,$AF,$F8,$D1
	.dc.b	$AF,$7F,$3B,$C6,$47,$F8,$7F,$D3,$F6,$C3,$F8,$37,$5B,$E0,$7F,$A2
	.dc.b	$B6,$BF,$9C,$35,$F5,$FD,$3A,$37,$E3,$FF,$2F,$7F,$63,$3F,$ED,$36
	.dc.b	$BD,$A3,$5F,$41,$D8,$91,$FD,$16,$9E,$B7,$F2,$2E,$A9,$3F,$C7,$DF
	.dc.b	$BF,$CF,$FD,$6B,$E9,$DF,$2F,$FC,$FD,$B5,$FE,$02,$6B,$FE,$08,$6B
	.dc.b	$FE,$F9,$F9,$7F,$E3,$2F,$E2,$38,$91,$87,$FA,$EB,$77,$FF,$0D,$F9
	.dc.b	$FF,$E8,$7A,$FC,$57,$3F,$8F,$E9,$E4,$60,$FE,$C1,$8A,$3C,$C4,$8F
	.dc.b	$33,$F8,$DF,$07,$F3,$64,$8E,$31,$FD,$E3,$E9,$BF,$B9,$47,$F6,$B6
	.dc.b	$BF,$8E,$6E,$7F,$D7,$BE,$9F,$FD,$48,$FF,$60,$8D,$CF,$F4,$50,$E1
	.dc.b	$FE,$90,$D7,$FC,$FE,$D7,$77,$FE,$0A,$A7,$8D,$4F,$34,$7F,$83,$BE
	.dc.b	$9D,$F1,$74,$97,$FE,$03,$F9,$1F,$EF,$AF,$F8,$16,$76,$FF,$8E,$9A
	.dc.b	$FF,$90,$1A,$FA,$5A,$FF,$93,$BC,$07,$3F,$F0,$42,$47,$EF,$FD,$CA
	.dc.b	$FD,$FF,$17,$E3,$F1,$FF,$46,$CF,$3B,$FF,$87,$31,$7D,$27,$F0,$EF
	.dc.b	$A4,$98,$FF,$09,$8A,$FE,$26,$53,$8A,$E5,$5F,$E3,$0F,$A6,$7C,$FF
	.dc.b	$E2,$7E,$33,$E4,$E3,$FE,$AC,$9C,$F9,$BF,$EA,$E9,$CC,$A7,$2A,$33
	.dc.b	$1F,$C8,$5B,$5D,$78,$FF,$C0,$53,$C2,$73,$7D,$67,$3F,$03,$FE,$07
	.dc.b	$9C,$E1,$D6,$78,$E1,$CE,$3C,$CD,$7C,$95,$C6,$77,$7D,$67,$6D,$7D
	.dc.b	$55,$C7,$0F,$45,$70,$9F,$01,$C0,$A7,$0A,$E5,$3C,$23,$47,$E6,$DF
	.dc.b	$8A,$DA,$FD,$B3,$67,$FD,$0B,$F1,$4F,$55,$F4,$F8,$6D,$7F,$A3,$A7
	.dc.b	$53,$93,$FD,$1B,$87,$F1,$0F,$8F,$BE,$01,$A4,$BF,$F4,$BC,$ED,$F5
	.dc.b	$B7,$5F,$BB,$F0,$7F,$E8,$2E,$4F,$6E,$1C,$EF,$6D,$AF,$BE,$DD,$BF
	.dc.b	$C7,$D7,$C6,$B7,$8F,$22,$3F,$2A,$35,$7C,$A7,$2D,$9D,$7C,$2B,$1C
	.dc.b	$E1,$F8,$AC,$7C,$15,$F9,$76,$BF,$E6,$4A,$E0,$F8,$8F,$B7,$C4,$E6
	.dc.b	$FD,$AE,$1F,$DE,$F9,$FF,$D1,$2F,$F8,$DB,$A2,$DF,$F8,$F2,$BF,$D1
	.dc.b	$73,$FE,$01,$9C,$DF,$3B,$B5,$D2,$FF,$79,$18,$E5,$98,$7F,$82,$79
	.dc.b	$AD,$7F,$01,$9B,$87,$F9,$26,$6D,$7F,$9A,$7E,$95,$E0,$F9,$30,$46
	.dc.b	$27,$FC,$F3,$F4,$DF,$ED,$57,$FD,$3D,$19,$9F,$F0,$0B,$70,$FF,$48
	.dc.b	$B6,$7F,$DC,$B5,$C6,$B7,$7D,$FE,$91,$E9,$FF,$A6,$27,$FC,$F9,$18
	.dc.b	$BF,$F6,$FB,$63,$FD,$CA,$D7,$F9,$F3,$9C,$7F,$D4,$2D,$9F,$F7,$AF
	.dc.b	$D2,$FF,$88,$9F,$F5,$B4,$66,$81,$C3,$6F,$E9,$D7,$18,$FF,$0F,$CF
	.dc.b	$1B,$FC,$07,$D2,$FD,$63,$E5,$79,$47,$18,$FF,$B1,$63,$0D,$8F,$BB
	.dc.b	$DB,$C1,$B1,$93,$FC,$89,$19,$7F,$C7,$91,$F8,$6E,$FE,$78,$FF,$7D
	.dc.b	$F2,$7F,$D9,$75,$A3,$C1,$7E,$43,$F1,$B1,$FE,$BE,$8C,$9F,$EB,$67
	.dc.b	$13,$CF,$AB,$F2,$3C,$FC,$1F,$C3,$7D,$1B,$E9,$29,$F9,$8F,$A5,$78
	.dc.b	$58,$CB,$FD,$C9,$4F,$9C,$E2,$7E,$7B,$DB,$7C,$8E,$2B,$7E,$4F,$63
	.dc.b	$5D,$96,$3F,$DD,$DF,$49,$F4,$F9,$7F,$DB,$ED,$8F,$B7,$0F,$A5,$7C
	.dc.b	$1F,$E7,$6B,$5C,$5F,$80,$BF,$E4,$C7,$FB,$E4,$AF,$F8,$9F,$EA,$1F
	.dc.b	$4B,$FA,$59,$CE,$3E,$0D,$7A,$1F,$B8,$FA,$67,$C8,$E2,$BF,$CF,$97
	.dc.b	$FB,$A3,$F7,$7E,$A5,$8F,$39,$FE,$DB,$FA,$67,$A3,$FB,$36,$C7,$FA
	.dc.b	$14,$C4,$7F,$06,$BF,$F4,$8C,$7F,$BB,$9B,$A7,$F8,$FD,$FA,$D1,$9F
	.dc.b	$86,$E1,$EA,$48,$E1,$E1,$DB,$FF,$3D,$E3,$FC,$51,$BC,$7D,$FE,$B7
	.dc.b	$C8,$F8,$79,$E6,$1E,$1F,$A1,$B7,$FE,$84,$FA,$2F,$DB,$3F,$EF,$76
	.dc.b	$BF,$B0,$6A,$69,$FB,$23,$F5,$DA,$BE,$9F,$F8,$B4,$67,$4A,$7C,$38
	.dc.b	$B8,$7C,$3F,$84,$F8,$C7,$C7,$FA,$7B,$19,$26,$3F,$1E,$32,$FC,$9F
	.dc.b	$95,$F9,$C7,$CF,$F4,$AD,$7C,$F8,$B3,$F3,$E6,$F4,$7F,$46,$3F,$07
	.dc.b	$F2,$DF,$5C,$FB,$1D,$D7,$0F,$B1,$16,$FE,$C4,$DA,$E5,$B2,$B1,$7F
	.dc.b	$9E,$85,$62,$D7,$F9,$2F,$A0,$FC,$31,$F8,$7F,$A7,$B8,$7E,$1E,$2D
	.dc.b	$7E,$1E,$6D,$78,$3E,$6B,$F1,$C7,$E3,$FE,$6E,$E1,$F8,$F8,$B7,$F8
	.dc.b	$F9,$B5,$FD,$9A,$3E,$E7,$27,$F3,$D7,$F0,$0D,$73,$7F,$E7,$77,$2D
	.dc.b	$72,$7E,$2E,$3F,$B2,$FD,$3B,$87,$F0,$71,$2B,$F5,$75,$C8,$FF,$50
	.dc.b	$56,$E3,$31,$D2,$FE,$B7,$53,$E5,$3F,$B8,$B6,$B9,$CF,$FB,$FB,$E2
	.dc.b	$F3,$52,$73,$5B,$BF,$F5,$73,$FE,$C9,$8C,$CF,$F7,$CF,$A6,$FF,$69
	.dc.b	$47,$46,$D7,$0F,$27,$5F,$1B,$2B,$4F,$92,$31,$3F,$AD,$E6,$F9,$5D
	.dc.b	$46,$92,$FE,$54,$65,$F8,$2D,$9F,$F6,$27,$D3,$FF,$A6,$A7,$35,$FA
	.dc.b	$E2,$3F,$0A,$35,$3E,$18,$F9,$7F,$41,$EA,$DB,$F0,$0D,$8F,$E6,$7D
	.dc.b	$AF,$99,$FF,$07,$4E,$6F,$FB,$8B,$53,$9B,$57,$1B,$EF,$A7,$CB,$F8
	.dc.b	$D9,$1F,$2F,$E5,$6E,$3F,$ED,$FD,$D5,$8C,$7F,$69,$0A,$C5,$FE,$AF
	.dc.b	$E9,$BF,$96,$1F,$CB,$7D,$3B,$1F,$CB,$62,$E3,$FC,$B6,$6E,$1F,$B6
	.dc.b	$8F,$E6,$87,$F3,$5F,$85,$B7,$FC,$D6,$29,$FC,$D6,$2C,$FF,$07,$FA
	.dc.b	$27,$D3,$FF,$90,$9B,$7F,$92,$7F,$A6,$B6,$34,$47,$5E,$05,$C3,$F8
	.dc.b	$1C,$7E,$B7,$E1,$FF,$C3,$FF,$4E,$F0,$7F,$C4,$35,$FC,$23,$D3,$FF
	.dc.b	$2E,$CE,$63,$FD,$D5,$5C,$2B,$79,$FF,$6C,$C3,$15,$A1,$7C,$5E,$91
	.dc.b	$A7,$FD,$64,$6A,$B9,$E1,$E0,$7C,$D1,$93,$FB,$DC,$65,$ED,$5D,$6F
	.dc.b	$C5,$1B,$3A,$C6,$67,$84,$62,$BB,$DF,$8E,$3F,$DA,$47,$1B,$E7,$7E
	.dc.b	$98,$CB,$CE,$78,$47,$FB,$54,$E1,$7D,$9D,$DF,$67,$C8,$76,$F7,$B9
	.dc.b	$27,$B3,$CD,$C9,$8C,$7E,$46,$3E,$4A,$F8,$B8,$F2,$35,$7E,$2C,$7C
	.dc.b	$B7,$C3,$1F,$24,$F5,$BF,$26,$3C,$0E,$EF,$43,$D1,$87,$43,$B3,$99
	.dc.b	$FA,$78,$76,$7C,$07,$B8,$F8,$93,$C1,$FA,$F3,$D0,$E4,$FE,$10,$E1
	.dc.b	$87,$9E,$BA,$57,$19,$FB,$07,$80,$D9,$E0,$7C,$D5,$D6,$3B,$9B,$8C
	.dc.b	$B6,$3C,$38,$FC,$B8,$D9,$E3,$9F,$97,$D5,$F5,$63,$F3,$8D,$7F,$32
	.dc.b	$D7,$F1,$2E,$7E,$2D,$F3,$BF,$DC,$A7,$FC,$C7,$9F,$EB,$A7,$38,$7F
	.dc.b	$A4,$2F,$85,$ED,$B1,$9A,$E1,$3C,$F0,$FD,$66,$2E,$2F,$DC,$C6,$1B
	.dc.b	$1C,$2F,$F6,$18,$F8,$32,$FD,$34,$F1,$7F,$68,$D7,$F5,$51,$BB,$FD
	.dc.b	$D3,$67,$F8,$39,$B8,$F0,$4F,$6C,$3C,$D3,$E1,$8E,$F8,$79,$1F,$F5
	.dc.b	$7C,$71,$35,$5D,$AB,$8C,$EA,$B8,$3F,$C5,$DB,$7E,$38,$FF,$5C,$B5
	.dc.b	$D1,$EC,$8D,$CF,$48,$4E,$EF,$C5,$1D,$CF,$56,$11,$F4,$9C,$CC,$61
	.dc.b	$E7,$8F,$F6,$4E,$3F,$0A,$7F,$79,$5F,$1B,$1F,$E5,$B3,$B9,$C9,$E8
	.dc.b	$C7,$C9,$5E,$28,$F4,$9F,$4A,$3D,$18,$F9,$EF,$E1,$D6,$E3,$E0,$1F
	.dc.b	$76,$FC,$44,$8F,$A4,$F8,$71,$F3,$12,$3E,$89,$30,$F8,$47,$C0,$3C
	.dc.b	$58,$FC,$08,$F6,$47,$C3,$32,$49,$E8,$F2,$7C,$11,$D6,$3D,$B1,$A7
	.dc.b	$E9,$E3,$EE,$3E,$1E,$3E,$E3,$EC,$DF,$F4,$E1,$C0,$FA,$7F,$EA,$7A
	.dc.b	$FF,$B0,$23,$18,$7F,$67,$9C,$BF,$E6,$48,$7C,$D6,$2F,$9D,$59,$E3
	.dc.b	$5F,$DA,$0E,$11,$9A,$D7,$D3,$C9,$FE,$E8,$F1,$FA,$7C,$47,$1C,$3C
	.dc.b	$F3,$87,$FB,$F4,$F7,$EC,$95,$FD,$F5,$F1,$7C,$C6,$23,$96,$5F,$E2
	.dc.b	$47,$47,$F8,$0E,$3D,$5F,$4C,$F3,$A9,$27,$FC,$B3,$0F,$3F,$11,$E5
	.dc.b	$8E,$0E,$A7,$9C,$FC,$88,$7C,$CF,$FA,$45,$FF,$34,$B5,$F8,$C6,$BD
	.dc.b	$0F,$99,$F2,$BF,$ED,$A8,$DC,$73,$9E,$1A,$FF,$A1,$5B,$C3,$FE,$F8
	.dc.b	$BE,$17,$D3,$E9,$5F,$4F,$E3,$6C,$7F,$9E,$9A,$F8,$BE,$9D,$E9,$6E
	.dc.b	$A4,$ED,$F6,$4E,$A3,$07,$FA,$3E,$B9,$BF,$E2,$40,$E1,$FF,$13,$B5
	.dc.b	$FF,$57,$EC,$F0,$BE,$99,$23,$E8,$FA,$55,$FA,$11,$75,$FF,$63,$52
	.dc.b	$BF,$29,$AB,$C1,$A4,$C3,$ED,$94,$FC,$74,$6D,$FF,$4B,$B5,$FB,$C7
	.dc.b	$C9,$F9,$E8,$FC,$97,$E1,$F0,$62,$3E,$84,$5C,$7F,$D7,$5A,$7D,$69
	.dc.b	$1F,$12,$7F,$D6,$CD,$3E,$53,$8D,$0F,$FA,$EE,$BF,$73,$F4,$FF,$AF
	.dc.b	$25,$7F,$97,$7E,$9E,$BF,$A0,$8B,$3F,$E6,$59,$FD,$B7,$CD,$CA,$FF
	.dc.b	$32,$3F,$B9,$FA,$7C,$47,$EA,$22,$D7,$F9,$B6,$7F,$7D,$E7,$4A,$FF
	.dc.b	$36,$3F,$C0,$7C,$0C,$47,$EC,$22,$D7,$F9,$D6,$6F,$37,$F0,$F9,$B5
	.dc.b	$FC,$36,$7F,$85,$7A,$92,$3F,$CF,$0F,$F0,$DF,$23,$11,$FB,$C8,$B8
	.dc.b	$7F,$A0,$67,$F8,$97,$CD,$4A,$FF,$40,$3F,$C5,$7D,$4C,$47,$F0,$48
	.dc.b	$B5,$FE,$89,$9F,$D3,$7C,$0C,$AF,$D3,$3F,$C7,$7E,$9F,$11,$FA,$D8
	.dc.b	$B5,$FE,$91,$9C,$CF,$E8,$78,$BF,$66,$DA,$FB,$87,$DB,$F0,$1E,$5F
	.dc.b	$F2,$92,$4F,$FA,$5F,$CA,$FE,$15,$F4,$EF,$62,$92,$7F,$DD,$BF,$24
	.dc.b	$F5,$7F,$9A,$2F,$0E,$92,$77,$E1,$6B,$E2,$24,$72,$78,$FD,$3F,$83
	.dc.b	$E0,$A4,$FF,$C0,$0C,$D1,$DB,$E3,$7B,$FA,$F2,$BC,$6F,$8B,$C0,$FE
	.dc.b	$45,$F4,$FF,$CA,$67,$F4,$3C,$CE,$DF,$F5,$4C,$E6,$7F,$91,$B6,$75
	.dc.b	$07,$FA,$8F,$0F,$F2,$DF,$A7,$FD,$B4,$6A,$51,$8B,$FF,$5A,$B6,$3F
	.dc.b	$63,$D7,$FC,$9A,$B8,$60,$F8,$2F,$39,$2B,$CF,$87,$A3,$E9,$FE,$94
	.dc.b	$60,$FD,$DF,$77,$D1,$70,$1C,$3F,$CF,$9B,$87,$9F,$E9,$F5,$3C,$27
	.dc.b	$F7,$ED,$D7,$FA,$54,$62,$3F,$A3,$CB,$C3,$CF,$D0,$C9,$E4,$FF,$51
	.dc.b	$97,$A3,$FC,$93,$AE,$57,$F1,$E7,$F0,$FE,$FF,$F5,$7F,$A7,$7C,$BD
	.dc.b	$25,$6C,$F8,$1A,$BE,$61,$F0,$BF,$E0,$36,$CE,$9A,$FF,$2E,$F0,$7F
	.dc.b	$2B,$F3,$BD,$FE,$46,$4F,$E6,$A2,$BF,$D9,$32,$4E,$31,$FC,$18,$F0
	.dc.b	$97,$FA,$0F,$16,$61,$FA,$47,$87,$37,$FB,$DB,$E9,$FF,$00,$D8,$FE
	.dc.b	$ED,$18,$9F,$E6,$BF,$4F,$FE,$97,$C7,$F1,$93,$99,$34,$C4,$EE,$FA
	.dc.b	$4F,$F6,$13,$F9,$FC,$FF,$02,$72,$65,$C3,$FD,$13,$E9,$DF,$29,$FE
	.dc.b	$09,$6C,$7E,$31,$AF,$F8,$23,$E4,$3F,$B9,$3F,$E0,$A8,$CE,$3F,$A3
	.dc.b	$6E,$1F,$D4,$25,$18,$7F,$A9,$D7,$2A,$F9,$BA,$39,$FF,$85,$92,$BF
	.dc.b	$87,$FC,$95,$7F,$C6,$D3,$B7,$F8,$9C,$FF,$12,$9F,$E2,$7F,$49,$FC
	.dc.b	$53,$E9,$25,$7F,$18,$8C,$BF,$C6,$A3,$2F,$48,$FF,$15,$7D,$33,$E8
	.dc.b	$3F,$93,$CB,$E3,$1E,$A7,$B6,$B5,$CF,$FD,$BA,$73,$5F,$DC,$27,$2F
	.dc.b	$1A,$E1,$E7,$FC,$5A,$DC,$AB,$71,$86,$D7,$7C,$3C,$73,$9B,$E5,$39
	.dc.b	$C3,$C4,$DC,$3C,$07,$8E,$F8,$C7,$8A,$36,$66,$6B,$E3,$9D,$B9,$39
	.dc.b	$9E,$7A,$F0,$B9,$BF,$44,$F0,$8E,$55,$D3,$0F,$E8,$D1,$67,$E6,$BF
	.dc.b	$E3,$98,$F8,$ED,$8C,$BE,$B3,$FC,$7D,$F1,$9F,$C4,$FE,$9F,$F4,$11
	.dc.b	$FE,$C6,$F0,$91,$89,$D4,$FE,$95,$AF,$7E,$E7,$C7,$F2,$C7,$F8,$9B
	.dc.b	$63,$33,$FA,$D9,$CB,$63,$39,$FD,$76,$EB,$FA,$1F,$83,$FF,$3F,$73
	.dc.b	$FC,$19,$CD,$FE,$15,$AF,$E0,$5B,$3C,$23,$E7,$E1,$E0,$C3,$B6,$3C
	.dc.b	$CF,$AF,$5B,$8C,$D7,$0B,$EB,$7C,$71,$F2,$7C,$37,$F5,$EB,$D9,$23
	.dc.b	$F3,$95,$E2,$9D,$46,$31,$FC,$3F,$37,$F3,$3E,$7F,$E4,$BA,$7F,$6E
	.dc.b	$F0,$11,$8D,$3F,$B5,$43,$E0,$31,$7C,$6F,$F9,$C5,$E2,$B8,$C7,$F3
	.dc.b	$BA,$C4,$E7,$E0,$F5,$F2,$D6,$BF,$A0,$37,$1F,$F5,$D6,$08,$C6,$3F
	.dc.b	$CE,$A3,$15,$FD,$A3,$E8,$FF,$C1,$AF,$F4,$EF,$A8,$FE,$B2,$E7,$D5
	.dc.b	$F3,$C6,$31,$FF,$54,$D7,$07,$F9,$83,$5F,$F4,$93,$5F,$E4,$4D,$7F
	.dc.b	$CE,$7F,$6A,$F8,$3F,$87,$F4,$CF,$2F,$E4,$52,$B5,$97,$F6,$2A,$D3
	.dc.b	$FE,$00,$8C,$CF,$FB,$66,$72,$FF,$AF,$A3,$3A,$7F,$5A,$9C,$BF,$E1
	.dc.b	$98,$CD,$7F,$37,$6D,$7F,$88,$6B,$FA,$FC,$F7,$C3,$EB,$7C,$57,$DA
	.dc.b	$F0,$DF,$4D,$8A,$FF,$22,$BE,$AF,$3B,$F9,$54,$7F,$75,$AC,$77,$EF
	.dc.b	$2F,$F6,$9D,$69,$B1,$F8,$3E,$CD,$DF,$F7,$DE,$7F,$04,$FE,$1D,$B1
	.dc.b	$FE,$22,$BF,$F0,$5B,$FE,$03,$F9,$2F,$17,$5F,$E3,$5F,$C8,$FB,$3F
	.dc.b	$EB,$79,$E1,$38,$F8,$BD,$C7,$F8,$B3,$E9,$3F,$D5,$2F,$F0,$E6,$E9
	.dc.b	$FD,$BD,$AF,$FB,$45,$AF,$F6,$18,$FF,$14,$1F,$E1,$AF,$A2,$FF,$06
	.dc.b	$BF,$E3,$83,$31,$FE,$4F,$7F,$CA,$F1,$A8,$FF,$7B,$FD,$07,$F9,$B6
	.dc.b	$7F,$8F,$B6,$3F,$C0,$0D,$7F,$DC,$4D,$7F,$CC,$38,$BE,$53,$E0,$9F
	.dc.b	$0B,$E6,$2D,$FF,$92,$AD,$F0,$9F,$77,$4F,$C9,$C6,$27,$FD,$03,$1F
	.dc.b	$2B,$E5,$24,$7F,$8D,$B1,$7C,$B6,$33,$F6,$1F,$E6,$B8,$FD,$57,$57
	.dc.b	$F8,$B1,$7C,$B7,$D4,$F6,$FE,$AB,$E0,$F4,$7F,$1D,$EC,$DC,$BF,$D2
	.dc.b	$2F,$DA,$EC,$FC,$97,$D0,$D9,$FF,$4A,$BF,$7B,$E4,$7E,$FF,$C9,$59
	.dc.b	$F3,$3E,$CF,$5B,$F3,$51,$4F,$F4,$E9,$F8,$6F,$91,$FE,$29,$F4,$D6
	.dc.b	$3F,$C9,$CF,$ED,$BE,$79,$F4,$1F,$C7,$3E,$81,$F3,$59,$8C,$36,$FF
	.dc.b	$CB,$73,$A7,$FC,$F3,$5C,$9C,$CE,$F9,$3F,$C7,$93,$96,$DF,$FA,$45
	.dc.b	$FF,$17,$46,$5F,$F1,$87,$27,$FA,$E7,$E2,$FF,$3E,$7E,$83,$E5,$EE
	.dc.b	$3F,$5E,$2D,$7D,$6A,$FF,$60,$56,$1F,$AD,$F2,$FF,$A8,$3F,$53,$F1
	.dc.b	$37,$0F,$C0,$45,$BE,$5D,$1F,$E9,$39,$C9,$F8,$5F,$97,$FD,$89,$FB
	.dc.b	$2F,$5E,$C6,$49,$5C,$A3,$2F,$49,$B8,$FE,$26,$6D,$7F,$AF,$BE,$7B
	.dc.b	$FD,$A8,$FE,$4F,$C7,$FE,$9F,$18,$C3,$CB,$18,$F9,$6F,$95,$1F,$EB
	.dc.b	$F9,$D3,$F2,$9A,$F5,$8D,$E0,$9F,$D9,$3F,$E8,$66,$DF,$8F,$E5,$BC
	.dc.b	$3F,$05,$23,$FD,$37,$3A,$C3,$C1,$59,$7F,$4E,$38,$7F,$D3,$F5,$FE
	.dc.b	$A2,$FA,$0F,$EA,$18,$FF,$B4,$67,$35,$F6,$67,$35,$F8,$4F,$8D,$FE
	.dc.b	$DF,$18,$91,$D6,$33,$EC,$7E,$6D,$F8,$71,$9A,$FF,$16,$DF,$68,$FF
	.dc.b	$3E,$19,$2E,$1F,$0F,$C7,$F9,$E7,$D0,$F9,$7B,$7F,$02,$2D,$7E,$3F
	.dc.b	$45,$62,$FF,$DD,$9F,$3B,$F5,$DF,$F5,$8F,$D0,$3B,$7F,$B2,$CE,$72
	.dc.b	$FF,$7E,$CE,$5F,$F7,$1C,$7D,$8C,$7D,$ED,$4D,$31,$F4,$9F,$E9,$19
	.dc.b	$F9,$DF,$47,$FD,$CA,$31,$5F,$36,$31,$E0,$7C,$C6,$D7,$03,$FD,$E7
	.dc.b	$5C,$2B,$E3,$B5,$FF,$52,$35,$FA,$6D,$CB,$F1,$9E,$DF,$E6,$3E,$53
	.dc.b	$F3,$2D,$BF,$C4,$BF,$92,$DB,$FB,$FE,$1F,$97,$8C,$19,$F2,$3F,$43
	.dc.b	$F2,$5F,$A2,$6D,$FE,$49,$D3,$FE,$C9,$C7,$F3,$91,$8B,$FB,$B8,$3E
	.dc.b	$37,$FC,$6D,$DB,$FA,$C6,$E1,$F9,$C7,$C8,$47,$A4,$F1,$6F,$B1,$FB
	.dc.b	$4F,$8D,$FD,$AB,$63,$F4,$AD,$C3,$F0,$CD,$CB,$F5,$DE,$67,$F9,$1F
	.dc.b	$A7,$F7,$AD,$BF,$D6,$BF,$14,$8F,$89,$FC,$ED,$FF,$AD,$C7,$D0,$BE
	.dc.b	$8B,$F9,$4F,$8C,$BF,$D3,$CE,$6B,$F4,$3F,$15,$8B,$DD,$59,$C8,$F0
	.dc.b	$B8,$2C,$D8,$F5,$8D,$5F,$FB,$E7,$0F,$DA,$DF,$0B,$DE,$18,$AD,$5F
	.dc.b	$A1,$F0,$B4,$F0,$E3,$C7,$1E,$CF,$8A,$3E,$03,$9A,$F4,$47,$FB,$4C
	.dc.b	$E4,$79,$E7,$8D,$78,$2B,$84,$E8,$E7,$1E,$28,$ED,$1F,$EE,$73,$DA
	.dc.b	$7F,$10,$BF,$5B,$83,$C1,$87,$43,$8E,$3E,$49,$DD,$62,$BC,$B1,$8B
	.dc.b	$F2,$1E,$12,$C7,$7C,$3D,$2F,$D4,$C3,$B4,$FC,$A6,$C7,$C8,$8E,$4D
	.dc.b	$39,$63,$F1,$9F,$79,$83,$E1,$1E,$FC,$3E,$29,$E7,$2C,$F9,$23,$E6
	.dc.b	$4F,$37,$84,$78,$1F,$A5,$3C,$1F,$11,$C0,$EF,$8F,$C5,$C3,$DB,$1C
	.dc.b	$2B,$93,$E6,$38,$57,$8E,$33,$1E,$03,$D5,$19,$6C,$79,$F1,$F9,$D1
	.dc.b	$E2,$6B,$DC,$D9,$FA,$0F,$8F,$FD,$03,$72,$FC,$BC,$62,$3F,$1C,$DF
	.dc.b	$8D,$FE,$79,$5F,$D5,$23,$1D,$9F,$A7,$A4,$62,$75,$7F,$D6,$A5,$7C
	.dc.b	$9D,$C6,$63,$18,$FE,$CB,$17,$BB,$C6,$30,$D8,$E1,$97,$86,$7F,$73
	.dc.b	$5C,$A7,$C0,$72,$C3,$F5,$2D,$7F,$53,$1B,$7F,$B7,$B5,$FE,$1D,$7D
	.dc.b	$DF,$EE,$73,$C6,$BB,$DF,$86,$3C,$78,$F6,$BE,$17,$C8,$FD,$CE,$1D
	.dc.b	$4E,$D5,$CC,$F3,$CF,$08,$D5,$F5,$AF,$0C,$F4,$AE,$11,$E8,$9D,$3F
	.dc.b	$39,$B1,$A8,$F1,$1F,$02,$37,$5E,$2A,$E8,$78,$63,$27,$27,$33,$EA
	.dc.b	$8C,$DF,$F1,$F8,$F2,$94,$EF,$8F,$C6,$8C,$5F,$79,$F8,$C7,$38,$8F
	.dc.b	$F1,$CC,$3D,$07,$A6,$7E,$1C,$F5,$8F,$B5,$5F,$84,$AE,$B7,$F9,$38
	.dc.b	$F0,$47,$3C,$3D,$2F,$C1,$BE,$F8,$F1,$8D,$18,$C3,$ED,$B8,$24,$74
	.dc.b	$3A,$DF,$12,$4F,$07,$4F,$DC,$9F,$79,$E1,$3F,$A1,$57,$F9,$57,$5F
	.dc.b	$F1,$BC,$7F,$99,$5A,$FF,$97,$DB,$1B,$8F,$EE,$F1,$FE,$8A,$3F,$9C
	.dc.b	$7C,$3B,$E4,$71,$3F,$D6,$E3,$0F,$F9,$72,$73,$4F,$A2,$71,$61,$73
	.dc.b	$D3,$11,$2F,$FA,$35,$FF,$47,$AD,$C6,$0F,$ED,$4E,$1F,$F4,$C3,$D9
	.dc.b	$D5,$F8,$7B,$7D,$B2,$69,$27,$FC,$41,$19,$78,$E3,$5F,$F5,$23,$6B
	.dc.b	$FB,$C1,$D6,$3B,$CB,$E8,$FB,$92,$3A,$B5,$F3,$F9,$FF,$0C,$F8,$91
	.dc.b	$F1,$7E,$3F,$CE,$49,$F2,$B5,$F7,$FC,$D7,$CB,$38,$3F,$14,$7C,$DF
	.dc.b	$C3,$7F,$02,$D7,$FC,$F2,$D7,$F9,$53,$5F,$F4,$2B,$C7,$1F,$F3,$8F
	.dc.b	$A1,$F7,$3C,$63,$F8,$2A,$EF,$BD,$BE,$1B,$FF,$9D,$2B,$81,$99,$E1
	.dc.b	$F4,$6F,$92,$FC,$9E,$08,$C3,$F9,$66,$BF,$9A,$6B,$FE,$92,$73,$97
	.dc.b	$D4,$F3,$DE,$1F,$E4,$9B,$5F,$16,$B7,$97,$E3,$70,$FC,$A1,$27,$ED
	.dc.b	$FC,$95,$7F,$1B,$17,$1F,$C0,$F1,$2B,$E1,$7D,$13,$D1,$F5,$B7,$0F
	.dc.b	$D7,$B5,$FD,$9B,$5F,$F5,$5B,$9B,$FF,$22,$43,$E8,$FF,$C8,$59,$7F
	.dc.b	$5F,$6C,$7F,$AE,$AB,$84,$FF,$63,$6C,$7F,$AF,$A7,$6F,$3E,$AF,$3E
	.dc.b	$1F,$B3,$64,$7E,$69,$FD,$B7,$C9,$E2,$3F,$5B,$17,$1F,$83,$3F,$BB
	.dc.b	$F8,$CF,$D6,$92,$BF,$56,$D7,$F6,$BF,$1B,$FC,$04,$E0,$FE,$A3,$E3
	.dc.b	$7F,$82,$11,$FD,$5F,$83,$5F,$DE,$C5,$C3,$E8,$CD,$D7,$F8,$34,$DA
	.dc.b	$FE,$0B,$3F,$C4,$B8,$5F,$1B,$49,$3F,$B8,$29,$FE,$75,$BF,$F2,$B4
	.dc.b	$F0,$8F,$E1,$6D,$7F,$87,$9D,$7E,$2E,$3A,$7C,$1F,$17,$88,$E8,$7F
	.dc.b	$A0,$65,$39,$CB,$F8,$5C,$2B,$F1,$7E,$8F,$E2,$89,$7F,$8A,$6B,$D0
	.dc.b	$FC,$37,$B1,$FB,$3E,$23,$CA,$3F,$82,$71,$AB,$53,$27,$31,$F1,$FE
	.dc.b	$27,$F9,$CF,$84,$EE,$A6,$6B,$F8,$77,$D1,$3E,$0F,$7E,$24,$9C,$C7
	.dc.b	$D1,$F8,$9F,$A5,$F1,$32,$7E,$5C,$7D,$5F,$89,$EF,$E3,$C9,$F9,$F1
	.dc.b	$FC,$37,$1F,$E7,$DC,$7F,$CC,$34,$7C,$2F,$98,$D7,$95,$F7,$FE,$23
	.dc.b	$F0,$1F,$11,$32,$FB,$31,$F8,$3F,$42,$47,$DC,$7F,$0D,$F3,$1D,$F0
	.dc.b	$F8,$3E,$64,$AF,$7C,$7E,$2F,$D4,$91,$F8,$47,$F1,$DF,$11,$E7,$F3
	.dc.b	$FF,$8D,$EE,$69,$2B,$FB,$8F,$D0,$7E,$2F,$83,$F7,$F8,$3D,$EF,$E5
	.dc.b	$8D,$AF,$E7,$95,$BF,$07,$FC,$C0,$B6,$FE,$52,$D9,$EB,$18,$3F,$BE
	.dc.b	$F8,$1F,$CE,$23,$FA,$37,$CF,$3D,$BF,$E7,$4D,$8F,$CF,$FA,$58,$9F
	.dc.b	$D5,$7C,$44,$9E,$31,$FA,$DF,$89,$FC,$3E,$F2,$7F,$4F,$1D,$3E,$3A
	.dc.b	$47,$81,$FD,$A7,$B3,$88,$FD,$2F,$0F,$F4,$1F,$73,$F7,$06,$5F,$D2
	.dc.b	$BF,$94,$F1,$64,$FE,$D2,$3F,$A9,$5B,$E3,$A9,$27,$FB,$27,$C2,$E5
	.dc.b	$FE,$CD,$35,$FD,$54,$7E,$C5,$DF,$95,$C5,$FD,$74,$EF,$1F,$E6,$B5
	.dc.b	$B9,$E3,$5B,$9F,$E7,$3F,$38,$F9,$EF,$EB,$AD,$AF,$F1,$8B,$5F,$F1
	.dc.b	$AB,$5F,$F4,$E0,$75,$3F,$D1,$3E,$79,$C3,$FD,$C5,$B5,$FE,$01,$AD
	.dc.b	$F3,$FF,$BB,$FD,$67,$B5,$FE,$34,$7F,$A2,$E3,$FD,$5D,$C4,$4B,$FF
	.dc.b	$14,$C6,$89,$1C,$7E,$2E,$4E,$9E,$9F,$17,$88,$FB,$74,$9C,$5F,$F8
	.dc.b	$A2,$15,$FE,$09,$F9,$CF,$C9,$71,$30,$7C,$2F,$B1,$4E,$9C,$ED,$FE
	.dc.b	$C1,$8C,$4E,$DF,$7C,$F3,$8C,$74,$7F,$B2,$A3,$89,$DE,$78,$D7,$39
	.dc.b	$D5,$73,$D3,$FD,$9C,$F2,$C3,$A5,$6D,$B8,$73,$BE,$33,$E5,$BE,$2E
	.dc.b	$E7,$C2,$7B,$27,$6F,$1A,$E6,$FC,$0A,$E2,$E4,$E3,$87,$A1,$C6,$1E
	.dc.b	$77,$8C,$7C,$77,$F1,$6D,$7C,$46,$2F,$E2,$3F,$90,$F8,$57,$8A,$F9
	.dc.b	$DF,$F4,$3C,$61,$B1,$99,$FF,$46,$CE,$5F,$F4,$7D,$B9,$9E,$CF,$FA
	.dc.b	$42,$30,$D8,$CC,$F0,$6B,$E8,$73,$97,$DD,$6E,$BF,$F0,$07,$DB,$FE
	.dc.b	$5D,$C7,$F0,$27,$37,$F7,$5A,$FE,$29,$B3,$C2,$3E,$5C,$6B,$1F,$2C
	.dc.b	$63,$1F,$A7,$5B,$C7,$C5,$1D,$C9,$3F,$72,$75,$8F,$E5,$70,$E8,$4A
	.dc.b	$E7,$DA,$CB,$8E,$3F,$BA,$E3,$79,$9F,$E4,$D9,$FF,$49,$FB,$DF,$E9
	.dc.b	$07,$FD,$67,$F0,$17,$5F,$F5,$94,$DB,$E3,$36,$F9,$57,$FA,$6A,$31
	.dc.b	$59,$F9,$DD,$7B,$F7,$1F,$F5,$0B,$6F,$FD,$43,$3C,$FE,$03,$19,$7F
	.dc.b	$95,$BE,$73,$FD,$7B,$87,$FB,$31,$B7,$FE,$D0,$6B,$FC,$71,$AF,$EF
	.dc.b	$C3,$C2,$7A,$7C,$27,$AF,$E6,$FF,$83,$DF,$FA,$76,$B4,$F1,$F9,$C7
	.dc.b	$AD,$BF,$3F,$E2,$64,$9C,$D7,$FB,$13,$0F,$ED,$CF,$FA,$9E,$33,$3F
	.dc.b	$EE,$66,$BF,$CC,$9A,$F7,$BF,$F3,$AB,$67,$94,$79,$FE,$3F,$C3,$5D
	.dc.b	$33,$FF,$0C,$7C,$2F,$D3,$3C,$BE,$57,$F7,$1C,$3F,$D9,$F1,$8E,$EE
	.dc.b	$B9,$7F,$23,$AD,$3F,$EF,$A6,$BF,$D1,$9A,$FF,$BD,$BA,$B9,$5F,$FB
	.dc.b	$69,$B1,$FE,$DA,$E8,$FC,$21,$F8,$5F,$85,$B5,$FE,$C9,$BF,$DC,$3F
	.dc.b	$D4,$AB,$C5,$D5,$E4,$53,$A7,$C9,$1E,$8C,$FF,$DF,$BF,$0B,$F9,$3F
	.dc.b	$6B,$FC,$E3,$87,$FB,$B2,$F1,$3F,$5B,$E6,$BF,$A0,$4F,$FB,$65,$B3
	.dc.b	$FE,$04,$F9,$A7,$D0,$EE,$3E,$ED,$7F,$16,$3C,$CF,$EA,$FC,$3F,$34
	.dc.b	$66,$3D,$4D,$7F,$B6,$FC,$DF,$F6,$48,$FF,$74,$36,$3F,$C2,$7F,$0B
	.dc.b	$FC,$17,$4F,$F5,$9C,$E5,$FF,$80,$3A,$7F,$9E,$AF,$FA,$C4,$63,$E2
	.dc.b	$7E,$07,$A1,$F0,$5F,$C9,$DF,$F6,$3A,$F3,$B7,$1F,$9C,$D7,$CB,$E6
	.dc.b	$FC,$BC,$59,$7C,$B3,$E6,$7A,$57,$0F,$09,$E2,$97,$C1,$7D,$1F,$83
	.dc.b	$FA,$4D,$9F,$7F,$C5,$FD,$BC,$FF,$DC,$F1,$E4,$8C,$1E,$1E,$DF,$B4
	.dc.b	$7D,$BF,$95,$B3,$E0,$D5,$C0,$CD,$F0,$7F,$81,$F8,$4F,$C1,$36,$FE
	.dc.b	$7F,$B1,$E1,$23,$F1,$1F,$81,$EF,$7E,$24,$FC,$57,$2D,$9F,$06,$4F
	.dc.b	$96,$7C,$47,$F2,$1F,$9A,$FF,$42,$D6,$A7,$2F,$9F,$54,$EA,$FF,$D5
	.dc.b	$55,$A7,$F4,$0D,$7F,$C6,$6D,$7F,$8C,$35,$FF,$0C,$3C,$67,$FD,$F2
	.dc.b	$DA,$FC,$2D,$3E,$1F,$F8,$AC,$E6,$FE,$4E,$CF,$84,$FD,$41,$FA,$9F
	.dc.b	$86,$B8,$FA,$67,$4D,$9D,$34,$D7,$C1,$7E,$C4,$FD,$97,$CD,$58,$C9
	.dc.b	$30,$FF,$38,$DA,$75,$18,$F8,$3F,$DD,$1F,$BB,$F6,$2D,$F8,$E2,$CF
	.dc.b	$29,$BB,$7E,$9A,$6D,$7E,$56,$6D,$7E,$23,$E5,$9F,$35,$FC,$1F,$C7
	.dc.b	$FE,$B5,$18,$BF,$81,$18,$99,$1F,$0F,$DF,$B7,$FE,$8A,$BF,$DB,$4F
	.dc.b	$ED,$1B,$1D,$FE,$07,$AF,$C0,$4A,$FC,$25,$6B,$42,$AD,$5F,$E1,$BE
	.dc.b	$17,$FB,$77,$CD,$7F,$52,$8F,$C0,$C6,$3D,$7F,$BA,$FF,$27,$E5,$FF
	.dc.b	$79,$E5,$FC,$59,$B5,$F8,$E9,$D1,$C2,$7F,$2D,$6F,$95,$C5,$7B,$7C
	.dc.b	$7F,$8A,$7C,$6F,$84,$B7,$EA,$8B,$3F,$0E,$6E,$BF,$E0,$F8,$FF,$37
	.dc.b	$78,$3F,$35,$FF,$58,$FC,$7F,$FB,$2F,$0F,$D6,$CE,$6B,$80,$9D,$57
	.dc.b	$F0,$0A,$FE,$F5,$5B,$6F,$17,$F8,$4E,$9D,$5F,$DD,$2B,$57,$FE,$7F
	.dc.b	$F9,$A7,$83,$F3,$3E,$5B,$FB,$CC,$63,$2F,$87,$18,$9F,$EA,$53,$9A
	.dc.b	$FE,$AC,$D7,$FD,$52,$D7,$FC,$1C,$EF,$3F,$85,$08,$C5,$FE,$1F,$E8
	.dc.b	$BB,$3F,$84,$F7,$3F,$27,$F2,$7F,$94,$6C,$64,$97,$F9,$1F,$47,$51
	.dc.b	$F8,$CF,$2B,$F3,$7E,$E7,$E7,$1B,$19,$25,$7E,$67,$E3,$35,$1F,$94
	.dc.b	$F9,$BF,$ED,$9D,$86,$92,$3F,$BB,$6D,$F9,$88,$FD,$49,$FA,$AD,$EC
	.dc.b	$FE,$7B,$DC,$FD,$6F,$CD,$FE,$B9,$B1,$92,$65,$FA,$BF,$9D,$D4,$7E
	.dc.b	$9B,$E7,$3C,$64,$7C,$39,$A3,$15,$FB,$37,$F7,$7E,$E7,$EF,$1F,$F1
	.dc.b	$AC,$7F,$AD,$4F,$F0,$5F,$C2,$7F,$90,$F5,$FE,$3B,$18,$3F,$DC,$7F
	.dc.b	$15,$FE,$23,$9D,$79,$9F,$9B,$F1,$D1,$FE,$17,$9F,$F6,$57,$CE,$62
	.dc.b	$FF,$BA,$DF,$E8,$63,$11,$FE,$BC,$0E,$30,$F0,$BA,$8F,$33,$99,$E5
	.dc.b	$7D,$E7,$8B,$E4,$D3,$C2,$E6,$B8,$46,$27,$D6,$70,$3C,$AE,$5F,$26
	.dc.b	$3C,$1F,$84,$62,$BB,$38,$3C,$AE,$2B,$D1,$1F,$EE,$A3,$B1,$FD,$DA
	.dc.b	$FD,$2F,$8A,$3B,$1C,$31,$F1,$63,$F0,$AF,$DD,$87,$8C,$D5,$F9,$27
	.dc.b	$EB,$C6,$2F,$9D,$7C,$13,$C8,$5A,$F7,$9E,$43,$D2,$62,$3A,$CF,$79
	.dc.b	$F0,$9F,$01,$EC,$F4,$9F,$53,$5F,$B9,$3E,$27,$27,$58,$F1,$3F,$0F
	.dc.b	$1F,$A0,$F8,$8C,$1E,$53,$E5,$57,$8D,$FB,$2D,$7B,$CF,$85,$E5,$7F
	.dc.b	$25,$DC,$7A,$4F,$5B,$F0,$A7,$DF,$19,$8D,$99,$7F,$1D,$E4,$7F,$9C
	.dc.b	$23,$F1,$2D,$7F,$34,$D7,$F5,$8E,$7E,$39,$E9,$FF,$8D,$A5,$39,$8C
	.dc.b	$37,$04,$EB,$0F,$F5,$B5,$F2,$9C,$D6,$EB,$F2,$B8,$3E,$07,$FD,$79
	.dc.b	$87,$EB,$A3,$13,$FB,$93,$AE,$1C,$DF,$CE,$B6,$78,$57,$2A,$FF,$63
	.dc.b	$1C,$AB,$78,$FA,$EB,$F7,$77,$CF,$3E,$35,$FA,$BB,$E1,$1A,$BF,$84
	.dc.b	$DC,$BC,$95,$FC,$2B,$1F,$03,$F0,$AF,$C1,$8E,$E3,$C3,$5D,$6B,$FD
	.dc.b	$B5,$7E,$09,$FD,$7E,$3E,$3C,$3E,$4E,$3E,$07,$D7,$7C,$1B,$5F,$C0
	.dc.b	$6B,$83,$F1,$AB,$83,$E3,$AE,$0D,$38,$46,$94,$FC,$BA,$EB,$1F,$5A
	.dc.b	$BD,$AE,$AB,$E5,$4F,$9B,$0F,$64,$C3,$D7,$3D,$6F,$CB,$1F,$1C,$D6
	.dc.b	$1F,$14,$F8,$D8,$FC,$32,$61,$D6,$7E,$49,$FC,$51,$FA,$A7,$9A,$79
	.dc.b	$3B,$7E,$61,$1C,$1C,$E7,$AC,$FC,$33,$AB,$C9,$C4,$F1,$7E,$66,$1F
	.dc.b	$20,$F4,$D7,$84,$EB,$1A,$3C,$D8,$FD,$07,$E0,$1F,$12,$79,$3F,$34
	.dc.b	$FC,$5C,$FE,$20,$F3,$1F,$E5,$3C,$FF,$AF,$16,$3F,$CA,$2E,$1F,$F3
	.dc.b	$27,$CE,$DA,$FF,$29,$CF,$FB,$79,$B3,$FD,$C2,$31,$C3,$FE,$F5,$97
	.dc.b	$CE,$3C,$BE,$4D,$BF,$F5,$13,$5F,$F1,$D1,$FE,$BA,$5F,$35,$C2,$78
	.dc.b	$FA,$52,$B8,$C7,$4F,$0E,$47,$47,$BF,$95,$88,$FE,$16,$7F,$B9,$B0
	.dc.b	$4F,$AD,$B1,$FE,$1A,$6B,$C0,$F2,$7C,$D7,$A8,$8F,$AB,$C8,$AF,$AB
	.dc.b	$3F,$83,$EB,$FC,$03,$2F,$F9,$6A,$DF,$23,$F1,$09,$5F,$13,$E6,$AB
	.dc.b	$CE,$7F,$D7,$0F,$C8,$F6,$33,$7F,$E4,$AF,$9D,$F9,$4D,$DB,$ED,$52
	.dc.b	$73,$C5,$F0,$7D,$2C,$C7,$D1,$9F,$73,$5F,$8E,$D7,$F9,$9B,$5E,$95
	.dc.b	$F8,$5D,$5E,$CD,$3A,$C7,$8B,$E6,$5F,$27,$49,$7F,$EE,$3F,$61,$E4
	.dc.b	$E2,$3E,$54,$3C,$AF,$C9,$36,$3F,$DE,$37,$D1,$FC,$5D,$FE,$38,$8F
	.dc.b	$E3,$BE,$4E,$BF,$2E,$2E,$FF,$4E,$7E,$3F,$CE,$7F,$AB,$EB,$FD,$FA
	.dc.b	$D7,$F2,$A1,$C4,$FF,$9E,$BE,$75,$F2,$1C,$BD,$CF,$C4,$5F,$E4,$2F
	.dc.b	$93,$F8,$B6,$B9,$8D,$BF,$D8,$23,$AF,$1F,$E6,$2F,$FC,$88,$DF,$03
	.dc.b	$B5,$7E,$A9,$FD,$69,$1F,$D6,$FC,$AD,$7F,$15,$17,$2F,$C9,$4F,$ED
	.dc.b	$3E,$46,$57,$ED,$1F,$DB,$FC,$AE,$23,$F2,$5A,$7E,$E8,$C5,$FE,$6F
	.dc.b	$E4,$BF,$7A,$49,$FD,$EF,$8B,$5F,$CD,$45,$9F,$D1,$4D,$CF,$F6,$93
	.dc.b	$6B,$F8,$14,$FF,$11,$E1,$7C,$95,$24,$FF,$0A,$29,$FB,$18,$DB,$FE
	.dc.b	$0A,$C3,$FD,$93,$18,$8F,$D6,$1D,$7E,$3E,$3B,$C7,$84,$A7,$C5,$F2
	.dc.b	$1A,$AE,$D1,$6F,$AC,$FE,$2B,$DA,$95,$F8,$A7,$F6,$7C,$3F,$97,$8F
	.dc.b	$C5,$F1,$27,$F0,$31,$F8,$D6,$C6,$63,$F0,$AD,$7F,$1A,$DB,$FC,$97
	.dc.b	$6F,$63,$C3,$3F,$0F,$E4,$3E,$51,$1F,$70,$96,$6F,$90,$FE,$13,$F3
	.dc.b	$BF,$CC,$AB,$F2,$AD,$7F,$3A,$D7,$F9,$33,$5F,$93,$F2,$1F,$67,$E4
	.dc.b	$64,$78,$30,$EF,$3F,$77,$E4,$24,$E7,$14,$7E,$07,$E0,$FC,$7E,$6F
	.dc.b	$C6,$C8,$D2,$5F,$FB,$4A,$DF,$1F,$F8,$9F,$90,$93,$D6,$3F,$19,$EB
	.dc.b	$48,$F0,$3F,$8F,$F9,$0F,$83,$E6,$77,$24,$FC,$27,$FB,$16,$0F,$6B
	.dc.b	$F2,$C4,$9F,$57,$C3,$FC,$AF,$3E,$47,$D0,$1A,$FF,$42,$71,$5F,$ED
	.dc.b	$97,$5F,$3B,$F9,$77,$FA,$93,$5F,$EB,$6D,$7F,$CC,$ED,$4F,$F9,$C1
	.dc.b	$E7,$9F,$F4,$48,$C4,$7F,$1C,$9F,$E1,$EF,$E0,$F1,$FE,$76,$FD,$3F
	.dc.b	$33,$F6,$47,$EC,$FE,$43,$FA,$6F,$CE,$BE,$0B,$F3,$2D,$AF,$E3,$6D
	.dc.b	$7F,$91,$B5,$FF,$77,$5B,$E4,$7F,$7C,$48,$FC,$17,$C8,$FF,$01,$F7
	.dc.b	$BB,$FC,$1C,$9F,$C3,$45,$7A,$C7,$88,$8F,$E0,$FE,$07,$F8,$51,$1F
	.dc.b	$A5,$8A,$7F,$B3,$9F,$A1,$8F,$AC,$D7,$FA,$0B,$7A,$DE,$2F,$F8,$33
	.dc.b	$E1,$B0,$FF,$83,$A7,$85,$7F,$82,$5A,$E7,$BF,$F7,$F1,$88,$FF,$4F
	.dc.b	$BC,$9F,$F5,$0D,$3C,$5F,$EF,$6D,$8F,$2D,$7F,$7E,$F9,$CF,$ED,$31
	.dc.b	$F8,$D6,$BF,$D6,$1B,$8F,$A2,$3F,$D5,$BF,$3B,$FE,$B9,$8C,$42,$78
	.dc.b	$BF,$E2,$16,$BF,$E2,$A6,$BF,$EB,$66,$C3,$D1,$7A,$9F,$E4,$96,$D7
	.dc.b	$F0,$16,$BF,$E2,$75,$F1,$FB,$9E,$11,$27,$84,$73,$F9,$0E,$9F,$23
	.dc.b	$27,$A4,$57,$A4,$5C,$3E,$01,$E5,$9F,$F1,$2F,$CE,$7E,$67,$8B,$FD
	.dc.b	$9F,$F1,$2E,$5F,$F2,$3C,$E7,$14,$E5,$FF,$6A,$C6,$63,$10,$8F,$6C
	.dc.b	$61,$D4,$F4,$E8,$FF,$70,$D7,$48,$C5,$F8,$9F,$0E,$7F,$E5,$8C,$39
	.dc.b	$5F,$06,$D6,$EF,$C7,$39,$BE,$31,$93,$C9,$19,$77,$39,$C3,$CB,$1E
	.dc.b	$CC,$3C,$17,$E4,$8F,$31,$C0,$A7,$4A,$E5,$3C,$23,$47,$F9,$A6,$BE
	.dc.b	$53,$6F,$E0,$B6,$BF,$C4,$3F,$12,$F0,$DF,$3B,$86,$D7,$F8,$B2,$75
	.dc.b	$39,$3F,$C5,$38,$7E,$EF,$D4,$BC,$E6,$92,$FF,$C6,$B3,$B7,$D6,$DD
	.dc.b	$7E,$DF,$AF,$FE,$76,$E9,$F8,$D8,$73,$BD,$B6,$BE,$F3,$77,$FF,$47
	.dc.b	$DF,$1A,$DE,$3C,$88,$FC,$A8,$D5,$F2,$9C,$B6,$75,$E2,$7E,$E6,$30
	.dc.b	$4B,$FC,$BE,$1E,$1A,$0C,$3A,$CE,$B0,$FC,$47,$13,$F6,$5F,$F5,$1E
	.dc.b	$7F,$DD,$5F,$F5,$C3,$7E,$1B,$F5,$2F,$F8,$AE,$33,$8F,$FA,$7D,$E2
	.dc.b	$DF,$80,$DB,$FE,$0D,$8C,$DF,$EC,$9E,$2F,$F1,$0F,$9D,$B7,$FE,$37
	.dc.b	$FA,$07,$A3,$CB,$05,$69,$FF,$0C,$4E,$AB,$FB,$64,$F9,$27,$FD,$70
	.dc.b	$D7,$FD,$82,$D7,$E2,$B6,$FE,$17,$C3,$FF,$22,$9F,$73,$5F,$53,$63
	.dc.b	$75,$FE,$7E,$6C,$FC,$39,$F8,$31,$F3,$7E,$12,$E4,$9B,$F0,$D8,$F8
	.dc.b	$8B,$8F,$D1,$6E,$3F,$ED,$1C,$7E,$B5,$78,$70,$34,$97,$F6,$5E,$FE
	.dc.b	$87,$2A,$FA,$1C,$BE,$23,$C7,$EC,$5B,$FC,$03,$D4,$8F,$B7,$04,$78
	.dc.b	$7E,$13,$8C,$FE,$1D,$B9,$7E,$0F,$E0,$FD,$E7,$C0,$F8,$3B,$5F,$8D
	.dc.b	$F8,$2F,$51,$1E,$CF,$B7,$E0,$BE,$3F,$C1,$FC,$86,$CF,$E5,$70,$F9
	.dc.b	$67,$CC,$E8,$B1,$FE,$4D,$FC,$1B,$FB,$43,$F5,$5B,$3F,$2D,$AF,$E9
	.dc.b	$8F,$F3,$13,$87,$FD,$10,$FF,$98,$E3,$A7,$C1,$FD,$C8,$C4,$FF,$BF
	.dc.b	$DF,$F4,$34,$F8,$7D,$FD,$1B,$F8,$0B,$1E,$5F,$7F,$F1,$54,$69,$23
	.dc.b	$D1,$EF,$77,$F8,$0F,$03,$63,$DD,$E3,$FE,$3B,$CD,$7C,$B7,$96,$30
	.dc.b	$F9,$FD,$FE,$E4,$AF,$88,$F8,$BD,$A9,$1F,$1D,$BB,$7C,$7F,$17,$B7
	.dc.b	$C1,$FC,$16,$D7,$CF,$F6,$7D,$64,$7E,$83,$FA,$7F,$D9,$BF,$BE,$C7
	.dc.b	$FA,$05,$B7,$F2,$7C,$FF,$D3,$C6,$3B,$FF,$C2,$7F,$34,$E9,$CF,$07
	.dc.b	$F0,$38,$C6,$3C,$E3,$13,$FB,$7C,$55,$A7,$FD,$E6,$D7,$F0,$06,$A3
	.dc.b	$84,$6E,$FC,$31,$E4,$F4,$BA,$E1,$E9,$F5,$3C,$E7,$A3,$D4,$B1,$92
	.dc.b	$63,$FE,$AD,$7F,$DA,$59,$A7,$73,$93,$FD,$65,$1F,$EC,$39,$FF,$13
	.dc.b	$7C,$BF,$C4,$E0,$FB,$CD,$8F,$95,$1F,$23,$8F,$17,$CF,$46,$0F,$97
	.dc.b	$F9,$B8,$FC,$E6,$E3,$F4,$DF,$AC,$47,$F8,$E4,$F1,$6C,$71,$7E,$D7
	.dc.b	$CC,$5C,$3C,$14,$F8,$CF,$CA,$4F,$DC,$6D,$7E,$05,$C3,$F8,$3F,$8D
	.dc.b	$B7,$CB,$E5,$FF,$07,$EF,$FE,$11,$B5,$F8,$97,$0F,$E2,$FE,$62,$D7
	.dc.b	$CB,$3F,$03,$E2,$7F,$32,$F4,$3F,$D6,$D5,$FE,$B1,$E6,$4E,$7E,$1B
	.dc.b	$F1,$8F,$E0,$89,$3F,$86,$BF,$ED,$75,$C2,$31,$1D,$9E,$93,$F9,$58
	.dc.b	$E7,$1F,$9F,$F7,$7F,$6D,$C1,$F9,$E6,$C6,$49,$3F,$2B,$E6,$35,$1F
	.dc.b	$A7,$F9,$8F,$DF,$6F,$FA,$66,$C6,$49,$5F,$4B,$DB,$D4,$7F,$B3,$FE
	.dc.b	$65,$BF,$EB,$BE,$63,$FB,$64,$62,$7F,$69,$18,$9F,$D5,$4E,$6F,$E1
	.dc.b	$35,$FD,$4B,$5F,$E5,$4E,$F9,$FF,$D1,$6D,$BF,$F0,$9C,$3E,$69,$C5
	.dc.b	$DE,$72,$6B,$E1,$DF,$31,$FC,$77,$9F,$FB,$5C,$62,$7F,$D1,$51,$89
	.dc.b	$91,$FC,$34,$B9,$7E,$8E,$BF,$C1,$18,$FF,$A5,$79,$F9,$FB,$DF,$CA
	.dc.b	$5B,$7E,$2B,$57,$FA,$C6,$BC,$3F,$23,$A9,$E9,$F0,$1F,$E0,$3C,$7F
	.dc.b	$C4,$31,$FD,$2E,$FF,$D7,$13,$92,$61,$D6,$32,$DA,$7C,$C7,$BA,$BD
	.dc.b	$FE,$F7,$F3,$DE,$1B,$7F,$EA,$96,$E3,$ED,$9F,$8B,$BE,$32,$FF,$4F
	.dc.b	$E4,$F7,$3C,$C4,$C7,$FC,$31,$19,$6C,$7C,$E3,$E7,$FB,$D6,$3E,$26
	.dc.b	$3F,$E9,$31,$E5,$3D,$9F,$81,$DF,$FE,$CD,$8C,$4F,$F2,$EF,$83,$FF
	.dc.b	$29,$C6,$23,$E6,$CE,$EB,$F5,$D5,$FE,$58,$F9,$AD,$69,$FB,$8B,$47
	.dc.b	$17,$FD,$6E,$62,$BC,$38,$F1,$79,$C6,$4F,$C8,$E1,$FE,$E1,$AE,$51
	.dc.b	$FE,$BC,$34,$79,$2B,$93,$E9,$33,$7E,$39,$E3,$86,$A7,$81,$C8,$F2
	.dc.b	$E3,$EA,$C3,$DD,$1C,$E3,$18,$7C,$4C,$7C,$F8,$62,$B9,$1E,$57,$07
	.dc.b	$81,$F1,$DF,$AE,$78,$9F,$0A,$38,$E1,$D5,$F8,$37,$5F,$B5,$87,$73
	.dc.b	$A6,$1D,$AF,$9C,$F0,$39,$9E,$0A,$F8,$0F,$D2,$9E,$0F,$B6,$7C,$55
	.dc.b	$E7,$3B,$3D,$23,$E3,$CF,$82,$3D,$0D,$8F,$03,$5D,$C7,$1C,$7E,$93
	.dc.b	$71,$F8,$53,$E1,$3D,$17,$E1,$9F,$19,$BC,$3D,$33,$F3,$63,$C2,$7E
	.dc.b	$22,$7C,$F1,$CC,$FB,$AF,$37,$AC,$6C,$EA,$FE,$1F,$CD,$FD,$2B,$67
	.dc.b	$F2,$71,$88,$FD,$53,$7E,$1D,$F2,$3F,$C4,$25,$39,$8C,$37,$2F,$F5
	.dc.b	$EB,$67,$77,$FE,$5B,$BE,$75,$F7,$F0,$FE,$BF,$57,$C3,$FF,$30,$5F
	.dc.b	$EC,$EF,$73,$FB,$CA,$E5,$8F,$8E,$BF,$71,$1C,$A7,$47,$FB,$39,$C4
	.dc.b	$73,$9E,$11,$CB,$0F,$D3,$3C,$74,$E7,$8F,$F9,$C5,$E1,$3F,$AB,$C7
	.dc.b	$C7,$86,$B1,$F4,$1F,$99,$8C,$E1,$D6,$3F,$72,$F5,$3F,$CF,$85,$7F
	.dc.b	$DC,$71,$EE,$BE,$F1,$C3,$0F,$15,$70,$6C,$70,$C7,$EA,$57,$03,$C9
	.dc.b	$1C,$4F,$44,$6A,$BC,$98,$68,$8E,$0F,$45,$72,$23,$82,$57,$CD,$BF
	.dc.b	$1B,$FC,$BD,$B8,$7A,$E7,$A4,$EC,$F8,$E7,$D6,$11,$F2,$1F,$66,$3F
	.dc.b	$52,$BA,$D7,$C7,$9D,$E3,$EB,$76,$7B,$C9,$87,$C3,$9F,$0C,$79,$4F
	.dc.b	$0E,$3F,$38,$D3,$E5,$38,$C7,$B9,$EF,$87,$8D,$87,$F3,$4C,$7D,$CF
	.dc.b	$8A,$FE,$3B,$CD,$EA,$F2,$3E,$05,$7E,$50,$FB,$8F,$E2,$70,$FC,$AB
	.dc.b	$FD,$FB,$81,$F0,$3F,$DE,$75,$FF,$09,$36,$74,$DC,$95,$FE,$F2,$9F
	.dc.b	$EF,$53,$FE,$44,$9C,$B6,$7A,$CF,$F7,$08,$C7,$CB,$F3,$93,$4C,$7C
	.dc.b	$0D,$AF,$F0,$47,$A4,$F2,$7C,$4D,$8F,$F0,$55,$73,$F8,$09,$1F,$E0
	.dc.b	$D7,$B7,$C0,$E2,$3F,$D4,$39,$FA,$A7,$CC,$48,$FF,$02,$FA,$9E,$52
	.dc.b	$3F,$E1,$7D,$EB,$FE,$8F,$C5,$E4,$F9,$0C,$E1,$FE,$12,$F8,$7F,$71
	.dc.b	$2B,$FC,$41,$F0,$55,$E5,$E6,$7B,$0C,$53,$8B,$FD,$ED,$F3,$0F,$0F
	.dc.b	$EA,$5A,$30,$FF,$85,$9A,$FF,$A8,$FB,$DE,$CF,$09,$CC,$FF,$B0,$9A
	.dc.b	$F6,$AF,$F3,$EB,$D5,$B1,$D3,$0F,$F1,$6F,$5F,$D9,$3A,$4F,$DB,$A8
	.dc.b	$7F,$8F,$4F,$BD,$F0,$78,$7E,$47,$9B,$F6,$8C,$5F,$BE,$3F,$CA,$3F
	.dc.b	$91,$7E,$39,$FF,$24,$35,$FF,$5F,$36,$38,$05,$F3,$3F,$97,$75,$AF
	.dc.b	$F9,$CB,$89,$E8,$7E,$28,$D7,$5F,$CC,$FC,$AD,$E1,$D2,$67,$FE,$42
	.dc.b	$8F,$EA,$ED,$8F,$8F,$1C,$23,$FD,$4A,$D7,$FD,$54,$D7,$FD,$E2,$E7
	.dc.b	$3F,$4F,$CE,$BC,$6F,$D7,$36,$BB,$1F,$0B,$E1,$7D,$84,$7E,$2F,$C3
	.dc.b	$D7,$E2,$FC,$97,$C1,$31,$87,$C6,$DD,$F1,$BF,$14,$95,$F3,$BE,$26
	.dc.b	$BF,$9D,$F5,$FF,$5E,$62,$F9,$7C,$37,$CD,$24,$FD,$6F,$88,$AF,$EB
	.dc.b	$B1,$4F,$F0,$56,$C7,$FB,$36,$BE,$C4,$E4,$FB,$1F,$11,$27,$F0,$31
	.dc.b	$F6,$FE,$22,$47,$E1,$1F,$BD,$F3,$38,$8F,$DA,$73,$FF,$06,$31,$87
	.dc.b	$C1,$F8,$9F,$C2,$12,$7F,$1B,$E6,$D7,$F1,$B1,$67,$E4,$4F,$E1,$7E
	.dc.b	$26,$57,$E0,$DF,$C7,$73,$E2,$3E,$FC,$7E,$10,$8F,$E1,$1F,$CA,$78
	.dc.b	$32,$39,$1F,$9E,$8F,$CC,$11,$FD,$0F,$CB,$FE,$6F,$D2,$F9,$44,$8F
	.dc.b	$95,$C1,$FE,$2D,$7F,$D6,$5F,$B3,$BE,$4F,$F9,$C7,$C7,$7F,$B3,$27
	.dc.b	$FD,$64,$D7,$FD,$7F,$BF,$FB,$5A,$B8,$47,$FA,$DD,$7C,$6F,$F3,$D6
	.dc.b	$C7,$D9,$6B,$FC,$B1,$AF,$8B,$E3,$78,$11,$E0,$F3,$F8,$CF,$2F,$8F
	.dc.b	$27,$C1,$1F,$D3,$E9,$F1,$37,$E2,$E5,$79,$63,$5F,$11,$E5,$24,$79
	.dc.b	$5F,$47,$C2,$73,$24,$7B,$7E,$1B,$B7,$B3,$DF,$E0,$A4,$FC,$18,$F1
	.dc.b	$7A,$5F,$04,$91,$F0,$5F,$89,$E5,$F9,$C9,$1F,$27,$CE,$F5,$71,$7F
	.dc.b	$69,$FD,$A3,$E8,$76,$FF,$18,$6C,$7D,$13,$5C,$5F,$DA,$E3,$F9,$AD
	.dc.b	$FF,$BD,$1B,$46,$9F,$EE,$A8,$FF,$77,$35,$FB,$A7,$1F,$85,$F1,$7C
	.dc.b	$37,$8F,$D5,$F2,$7C,$2C,$AE,$31,$5F,$EC,$F6,$9F,$21,$23,$F7,$51
	.dc.b	$FC,$5D,$EF,$87,$97,$CF,$F8,$07,$37,$E0,$7C,$07,$C3,$24,$FA,$3E
	.dc.b	$16,$BE,$08,$F8,$44,$7B,$BF,$2F,$E4,$24,$7C,$08,$FE,$CB,$F3,$AF
	.dc.b	$7E,$92,$7E,$2F,$C1,$D4,$42,$CF,$17,$06,$5C,$38,$0E,$1D,$FC,$B3
	.dc.b	$AB,$FB,$5B,$63,$80,$70,$F3,$6B,$CA,$3F,$BF,$FC,$BB,$C7,$A7,$F8
	.dc.b	$0A,$B8,$EB,$F8,$BF,$83,$FC,$77,$C2,$4C,$7F,$07,$1F,$91,$30,$FD
	.dc.b	$6E,$14,$E6,$3E,$1D,$7F,$73,$5F,$2C,$EB,$FF,$5A,$C3,$E2,$BC,$8F
	.dc.b	$CA,$9D,$7C,$57,$CC,$9C,$CF,$2A,$E3,$18,$AF,$E1,$2D,$8C,$E9,$F4
	.dc.b	$E7,$9D,$77,$8C,$4E,$A3,$C1,$E3,$FD,$9B,$E5,$87,$86,$F6,$DA,$F3
	.dc.b	$5F,$17,$DB,$7C,$6B,$71,$DC,$DD,$71,$7D,$4D,$C3,$D3,$3C,$30,$F0
	.dc.b	$D7,$13,$65,$9D,$BD,$0F,$C5,$3E,$E7,$F1,$ED,$7C,$67,$E4,$BE,$29
	.dc.b	$E9,$BE,$5F,$F1,$71,$86,$C6,$67,$F2,$D3,$97,$BF,$80,$F8,$3F,$C9
	.dc.b	$E0,$8C,$4E,$A7,$31,$88,$F2,$BA,$C7,$ED,$37,$2F,$D5,$7C,$3F,$FA
	.dc.b	$33,$C2,$F8,$D3,$9B,$FC,$33,$5F,$C6,$B6,$78,$47,$C7,$8D,$63,$E5
	.dc.b	$C7,$99,$F4,$EB,$78,$F8,$A3,$37,$C5,$D7,$57,$F9,$36,$75,$39,$23
	.dc.b	$F9,$6A,$E7,$97,$E4,$71,$F4,$70,$BC,$D7,$C6,$7F,$94,$35,$FD,$3C
	.dc.b	$3C,$2C,$66,$BE,$17,$87,$94,$DA,$E7,$7F,$DD,$2B,$1E,$1E,$5E,$78
	.dc.b	$3E,$5B,$B5,$7F,$0A,$6C,$7F,$13,$6B,$FC,$61,$AF,$F1,$D7,$78,$FE
	.dc.b	$F5,$B3,$FC,$03,$E9,$1F,$2D,$E0,$9C,$DF,$AF,$2F,$E8,$57,$FE,$0A
	.dc.b	$F1,$7D,$5E,$37,$AD,$91,$93,$FC,$1F,$18,$AF,$E8,$B4,$F1,$3E,$17
	.dc.b	$8D,$F0,$DB,$7F,$E1,$67,$0F,$F8,$63,$C6,$B7,$CA,$6E,$BF,$E8,$2C
	.dc.b	$15,$8B,$FA,$BC,$BF,$3C,$FA,$1E,$2D,$8F,$F1,$1B,$87,$FC,$4B,$DF
	.dc.b	$72,$F3,$BF,$5C,$FB,$1F,$1D,$67,$FC,$57,$F1,$DF,$6C,$FB,$9E,$1D
	.dc.b	$8F,$F1,$7F,$C6,$FD,$F3,$F0,$1E,$4D,$8F,$F1,$9F,$C6,$7E,$0C,$FC
	.dc.b	$27,$65,$8F,$F1,$BE,$7F,$87,$3F,$11,$DD,$63,$F9,$B9,$FC,$E3,$E5
	.dc.b	$FF,$DE,$AE,$27,$F2,$51,$87,$FB,$AB,$FE,$A6,$8D,$71,$23,$4F,$FA
	.dc.b	$B1,$7C,$37,$E6,$57,$C3,$D8,$C3,$9F,$86,$97,$D2,$9F,$11,$8C,$7C
	.dc.b	$3F,$0D,$CC,$E9,$F1,$16,$F8,$7C,$37,$7F,$87,$F0,$36,$3A,$79,$5E
	.dc.b	$33,$C9,$F0,$F6,$3C,$1F,$0F,$E7,$3D,$1F,$0F,$63,$C9,$F0,$DE,$B3
	.dc.b	$D9,$F0,$F6,$3D,$1E,$67,$BC,$F8,$1F,$0F,$63,$D9,$F0,$FF,$0C,$F8
	.dc.b	$9F,$35,$63,$E5,$FC,$BF,$F9,$C5,$BF,$11,$F2,$7D,$6F,$E0,$53,$9E
	.dc.b	$4F,$F1,$A8,$F9,$7F,$F0,$C5,$F2,$F8,$FF,$E0,$31,$89,$FF,$16,$D2
	.dc.b	$73,$E0,$4B,$F0,$FB,$36,$FA,$E1,$C7,$BB,$FB,$E3,$6B,$FB,$FE,$3F
	.dc.b	$E7,$E7,$F1,$15,$E3,$C3,$CF,$F1,$52,$7D,$BE,$8D,$AF,$1C,$DC,$7F
	.dc.b	$DE,$78,$7F,$0A,$BF,$87,$3F,$03,$E2,$A4,$7C,$6F,$98,$B8,$7B,$26
	.dc.b	$E1,$CA,$6D,$79,$BE,$63,$E4,$FC,$5F,$CA,$6C,$7D,$07,$0F,$D2,$F8
	.dc.b	$BB,$8F,$C8,$9B,$87,$9A,$6D,$7C,$2F,$8A,$FA,$A7,$D6,$F8,$5B,$1F
	.dc.b	$72,$2E,$1F,$4A,$6D,$FC,$0F,$89,$FB,$BF,$17,$F7,$9B,$87,$E1,$1C
	.dc.b	$3F,$86,$F8,$BB,$87,$D6,$F3,$7F,$A3,$7B,$2E,$EF,$C0,$C6,$1B,$7F
	.dc.b	$5B,$99,$3F,$86,$BC,$9C,$E7,$51,$96,$9F,$96,$F8,$5F,$C9,$9F,$94
	.dc.b	$F9,$4B,$19,$25,$7E,$42,$32,$FE,$4B,$E3,$3F,$36,$7E,$73,$E0,$ED
	.dc.b	$7E,$8A,$2C,$FE,$62,$6E,$FF,$E4,$DF,$98,$6D,$FA,$4F,$90,$FE,$BF
	.dc.b	$18,$C7,$F5,$71,$89,$91,$FC,$63,$E6,$1C,$18,$CB,$F5,$78,$71,$6D
	.dc.b	$FF,$71,$AD,$DF,$E7,$AB,$58,$A3,$FC,$C3,$F2,$8F,$98,$FE,$F0,$D8
	.dc.b	$FD,$E9,$88,$8F,$F2,$A8,$D6,$3F,$B6,$8C,$BF,$EA,$48,$FE,$D7,$46
	.dc.b	$8F,$91,$EF,$F1,$32,$FA,$3E,$28,$C5,$F3,$F8,$8F,$29,$E6,$F8,$9B
	.dc.b	$7E,$38,$B3,$E1,$9B,$8F,$BE,$3F,$C2,$D8,$FF,$7C,$7C,$DF,$0F,$F0
	.dc.b	$37,$F8,$2C,$C7,$DC,$E5,$A7,$C1,$F8,$4F,$8E,$7C,$8F,$87,$B1,$92
	.dc.b	$5F,$C3,$8C,$BF,$C8,$8F,$F4,$40,$F9,$87,$C1,$7F,$AD,$73,$FD,$25
	.dc.b	$FF,$AE,$27,$35,$FE,$B6,$F8,$BC,$5E,$FB,$3D,$D5,$AA,$FF,$43,$1D
	.dc.b	$67,$57,$C2,$39,$19,$3F,$D4,$75,$CA,$7F,$D8,$17,$B9,$FF,$1F,$4E
	.dc.b	$5E,$37,$DF,$0F,$3C,$73,$3C,$31,$99,$E1,$18,$BF,$1B,$B3,$E1,$CE
	.dc.b	$F0,$F0,$46,$27,$73,$ED,$8C,$1E,$89,$DB,$FE,$B7,$3C,$57,$E4,$9E
	.dc.b	$11,$97,$B3,$E1,$31,$5D,$8F,$82,$F2,$BF,$4B,$E0,$3E,$1D,$F8,$30
	.dc.b	$F8,$D8,$77,$BE,$38,$76,$3C,$33,$AC,$39,$E1,$E0,$3D,$47,$8E,$35
	.dc.b	$87,$58,$F1,$36,$BA,$5F,$63,$D1,$3F,$51,$AF,$CF,$3A,$C7,$89,$E6
	.dc.b	$77,$2B,$93,$AB,$CC,$E3,$1D,$DA,$64,$F8,$67,$38,$E2,$F8,$4C,$BD
	.dc.b	$63,$CB,$1F,$2F,$0F,$90,$FE,$53,$CC,$79,$FF,$49,$B1,$98,$C0,$67
	.dc.b	$C6,$BD,$4F,$F4,$A4,$2B,$8A,$9D,$5F,$19,$CC,$62,$78,$5F,$57,$85
	.dc.b	$F4,$AD,$BD,$B1,$FD,$86,$3F,$D5,$AB,$F5,$0D,$7F,$49,$87,$59,$FD
	.dc.b	$B3,$67,$B1,$B8,$E7,$DF,$E1,$77,$7D,$E7,$C9,$5C,$6F,$E1,$47,$13
	.dc.b	$9D,$F1,$77,$5C,$67,$E3,$B5,$FE,$23,$3B,$79,$B5,$F8,$39,$7F,$03
	.dc.b	$BF,$04,$70,$3D,$F3,$B8,$C9,$AB,$F2,$4F,$43,$95,$F8,$67,$55,$DE
	.dc.b	$B7,$5E,$03,$E4,$D7,$9E,$32,$6A,$BB,$C6,$6B,$91,$E3,$7B,$19,$AF
	.dc.b	$39,$E2,$B8,$FF,$19,$BE,$B5,$C6,$73,$87,$D4,$AE,$27,$92,$33,$5F
	.dc.b	$42,$33,$3F,$62,$73,$1F,$04,$ED,$5D,$5E,$46,$0E,$35,$F6,$27,$8C
	.dc.b	$7C,$F3,$47,$0A,$EC,$E1,$F3,$39,$9E,$4E,$5F,$33,$93,$FB,$2F,$C4
	.dc.b	$7F,$51,$D3,$FC,$0A,$D8,$C9,$FE,$01,$F8,$AC,$61,$FD,$F6,$6F,$21
	.dc.b	$A6,$27,$3F,$0F,$2F,$FC,$14,$F0,$F8,$9C,$47,$F8,$26,$2E,$1F,$DE
	.dc.b	$4F,$F0,$54,$F6,$F8,$89,$5F,$E1,$37,$C3,$F1,$38,$8F,$F0,$94,$5A
	.dc.b	$E5,$92,$BC,$FE,$9C,$BF,$F0,$E3,$EA,$D7,$11,$FE,$1B,$8B,$7C,$A7
	.dc.b	$DF,$F1,$12,$BF,$C4,$8F,$C2,$F8,$9C,$47,$F8,$8E,$2D,$78,$A7,$E6
	.dc.b	$4B,$E6,$3F,$C4,$8F,$FA,$51,$AF,$FA,$A1,$B3,$F0,$FB,$1E,$EF,$29
	.dc.b	$CF,$17,$F1,$76,$DF,$F2,$56,$DF,$F9,$2A,$B8,$78,$1F,$63,$96,$5F
	.dc.b	$86,$3E,$DB,$0F,$F2,$01,$F7,$BE,$2B,$0F,$F8,$FC,$B7,$F3,$A3,$F0
	.dc.b	$BF,$2B,$FE,$41,$9F,$F7,$8B,$54,$FD,$D8,$C3,$7E,$63,$8F,$14,$AE
	.dc.b	$2F,$48,$F9,$3F,$31,$FE,$89,$9F,$F5,$DB,$5F,$F6,$7B,$6B,$FC,$85
	.dc.b	$08,$FE,$8C,$F0,$EF,$E6,$47,$9F,$5D,$7C,$FB,$FF,$4F,$B7,$CC,$3C
	.dc.b	$1F,$D1,$B6,$BA,$1F,$03,$E2,$BE,$19,$1F,$87,$F1,$B5,$EA,$7C,$4D
	.dc.b	$9F,$1D,$F2,$8D,$63,$F1,$FE,$2F,$E6,$92,$BE,$6F,$C7,$57,$9C,$AD
	.dc.b	$59,$8F,$A1,$5F,$5B,$E2,$E5,$FD,$67,$EC,$FC,$6E,$23,$9E,$09,$C5
	.dc.b	$FD,$9A,$BB,$7C,$5A,$7C,$6F,$E0,$8C,$E1,$ED,$F8,$DF,$C2,$91,$FC
	.dc.b	$2F,$C6,$D7,$BC,$5C,$7F,$0D,$3F,$8C,$F8,$D9,$5F,$8C,$7F,$1F,$C7
	.dc.b	$88,$F3,$C5,$AF,$C7,$4F,$E2,$3E,$36,$57,$E2,$1F,$CB,$F2,$E2,$3F
	.dc.b	$13,$16,$BF,$2D,$37,$4F,$77,$67,$E7,$8C,$5F,$D9,$BF,$D0,$91,$FD
	.dc.b	$0F,$C6,$FE,$93,$CD,$91,$FA,$47,$F4,$FF,$1B,$FC,$B3,$E6,$1F,$1D
	.dc.b	$49,$5F,$AA,$F4,$35,$F1,$B2,$3F,$29,$A7,$E5,$6D,$3C,$FE,$3B,$A7
	.dc.b	$C6,$4B,$D9,$FB,$2A,$47,$87,$E3,$BC,$5F,$1B,$2B,$7C,$4B,$4F,$9F
	.dc.b	$E3,$7D,$1F,$17,$E9,$F8,$D9,$87,$0B,$4F,$B7,$E2,$E4,$E5,$F8,$1F
	.dc.b	$1B,$F0,$7E,$32,$4F,$17,$E2,$7C,$6F,$C5,$F8,$DF,$8D,$F0,$F2,$BB
	.dc.b	$47,$C9,$F8,$C9,$1E,$17,$E6,$7C,$67,$CD,$F8,$C9,$3E,$57,$E8,$7C
	.dc.b	$67,$D1,$E7,$93,$E9,$7E,$A7,$C6,$7F,$60,$F9,$5F,$E5,$64,$F5,$71
	.dc.b	$F1,$79,$F8,$99,$87,$C8,$A4,$71,$F8,$BE,$5E,$DC,$AD,$3D,$7E,$2F
	.dc.b	$B7,$C5,$C9,$E4,$ED,$FD,$EE,$1F,$C3,$DB,$EB,$F9,$4E,$0E,$BE,$1F
	.dc.b	$D0,$47,$C9,$F1,$75,$D1,$E7,$8F,$74,$67,$BF,$FB,$95,$BE,$23,$E1
	.dc.b	$19,$E1,$FF,$33,$7C,$C3,$E1,$7F,$AE,$B7,$2F,$C7,$87,$0F,$AB,$E2
	.dc.b	$37,$32,$37,$3A,$C7,$C3,$5F,$DA,$3E,$5D,$DF,$49,$7D,$BE,$63,$FC
	.dc.b	$FF,$7F,$D0,$2B,$73,$FE,$7B,$3F,$CF,$8E,$1C,$07,$0E,$DA,$FF,$41
	.dc.b	$D1,$F1,$5F,$83,$24,$7D,$FF,$8B,$AF,$D6,$F8,$BF,$C4,$18,$C3,$FA
	.dc.b	$9F,$CB,$BB,$FF,$D2,$D4,$F8,$F7,$A9,$FE,$91,$AD,$7A,$1F,$E9,$F9
	.dc.b	$CD,$7F,$A8,$67,$2F,$1A,$E1,$E8,$7F,$A6,$6B,$72,$AD,$C6,$1B,$5D
	.dc.b	$F0,$F1,$CE,$6F,$94,$E7,$0F,$13,$70,$F0,$1E,$3B,$E3,$1E,$28,$D9
	.dc.b	$99,$AF,$8E,$76,$E4,$E6,$79,$EB,$C2,$E6,$FD,$13,$C2,$39,$57,$4C
	.dc.b	$3F,$90,$C5,$9F,$9A,$FE,$06,$3E,$3B,$63,$2F,$AC,$FC,$37,$C7,$BD
	.dc.b	$B7,$CC,$7E,$1A,$7F,$D9,$51,$87,$F1,$33,$B9,$D3,$63,$F1,$B6,$F1
	.dc.b	$1F,$15,$CE,$30,$D8,$CC,$F0,$6B,$E9,$73,$9F,$D9,$6E,$BF,$ED,$7F
	.dc.b	$8E,$FF,$39,$75,$7C,$39,$CD,$FE,$1D,$AF,$E0,$DB,$3C,$23,$E8,$E3
	.dc.b	$E4,$6C,$F0,$8C,$61,$F8,$CA,$E0,$F8,$F1,$F0,$12,$FE,$DC,$EB,$1F
	.dc.b	$CE,$61,$CC,$95,$F8,$49,$D5,$79,$E3,$F0,$FF,$2E,$F0,$FF,$D0,$99
	.dc.b	$FF,$8C,$FE,$67,$FC,$3A,$FF,$0E,$9D,$3F,$D6,$23,$37,$FE,$2E,$13
	.dc.b	$A7,$F8,$0B,$5F,$E0,$CD,$44,$70,$FF,$96,$B9,$DF,$1D,$E4,$8C,$37
	.dc.b	$95,$3F,$80,$C3,$FA,$24,$E6,$31,$5F,$CF,$23,$19,$65,$D4,$7F,$96
	.dc.b	$8F,$E2,$45,$3C,$45,$BE,$ED,$AF,$4D,$FF,$53,$8F,$30,$F8,$47,$A9
	.dc.b	$DA,$7F,$BF,$E8,$F8,$9F,$52,$F8,$DB,$18,$6C,$FF,$2A,$CB,$EB,$4F
	.dc.b	$8E,$B9,$4E,$5B,$8F,$F3,$3A,$DC,$F8,$2B,$BF,$C7,$74,$AF,$F2,$DD
	.dc.b	$FF,$99,$2B,$CB,$F1,$7E,$4F,$8C,$F2,$B6,$3B,$47,$84,$8F,$87,$E3
	.dc.b	$3C,$84,$70,$79,$7E,$27,$DE,$7C,$0D,$AC,$7C,$4F,$8B,$F6,$11,$F6
	.dc.b	$3E,$FE,$CF,$8F,$F1,$5F,$21,$B3,$F3,$3E,$2B,$E5,$9F,$33,$E2,$AC
	.dc.b	$7D,$0F,$89,$F9,$04,$7E,$43,$F2,$FB,$3E,$9F,$C4,$FD,$46,$CF,$D8
	.dc.b	$F8,$9F,$AE,$7D,$8F,$89,$B1,$FE,$3A,$FD,$A7,$FC,$9A,$FD,$76,$F2
	.dc.b	$7D,$36,$BF,$A0,$3F,$D6,$EE,$1F,$F5,$13,$FE,$4E,$8E,$9F,$13,$F9
	.dc.b	$F8,$C4,$FF,$A4,$94,$67,$E2,$7F,$41,$F2,$56,$FF,$C4,$CF,$1F,$89
	.dc.b	$FC,$D5,$FF,$93,$A7,$CD,$18,$33,$F0,$FD,$FB,$BC,$0D,$9F,$77,$A5
	.dc.b	$FA,$DF,$57,$C8,$D8,$F8,$5F,$0D,$D8,$8F,$C3,$7C,$3F,$09,$EB,$F8
	.dc.b	$5F,$63,$67,$FC,$37,$E1,$7F,$45,$6F,$4F,$EB,$23,$1C,$3F,$C3,$D4
	.dc.b	$67,$A7,$F7,$13,$FE,$E2,$6C,$F0,$8F,$D2,$46,$0E,$15,$E0,$E8,$B3
	.dc.b	$F3,$7C,$FF,$A1,$F0,$BF,$45,$B5,$F7,$DB,$3B,$8F,$C1,$4F,$D7,$F8
	.dc.b	$5F,$B1,$E6,$D9,$FC,$33,$6B,$ED,$37,$24,$FE,$0A,$3F,$00,$FE,$07
	.dc.b	$E1,$6B,$F8,$F6,$D7,$E0,$DB,$5C,$A3,$F1,$B1,$89,$FC,$6B,$63,$31
	.dc.b	$FE,$92,$FD,$A9,$F0,$DF,$CB,$5B,$19,$F8,$6C,$4C,$8E,$0E,$E3,$15
	.dc.b	$FE,$E6,$AF,$F3,$AD,$75,$F2,$3A,$7C,$37,$56,$C6,$4F,$14,$65,$FC
	.dc.b	$1F,$C6,$FE,$07,$C4,$FC,$13,$63,$27,$F8,$27,$44,$7E,$00,$91,$97
	.dc.b	$8B,$FE,$68,$C7,$FC,$80,$3E,$51,$DD,$FB,$5B,$E7,$5B,$AF,$F2,$51
	.dc.b	$B9,$D5,$FC,$FB,$E3,$3F,$E6,$3F,$57,$FA,$EC,$62,$18,$74,$79,$1B
	.dc.b	$29,$AA,$E9,$7F,$E6,$AB,$E6,$F8,$30,$B5,$C7,$0E,$D8,$75,$7C,$11
	.dc.b	$EE,$73,$87,$48,$E1,$7C,$DF,$09,$B7,$13,$CA,$3E,$0B,$CA,$FC,$31
	.dc.b	$FE,$63,$3D,$37,$E1,$AF,$44,$3C,$3E,$0D,$8C,$C6,$0E,$D8,$F4,$AE
	.dc.b	$55,$E6,$3F,$C0,$EE,$1F,$06,$1D,$0E,$2F,$8B,$1F,$73,$AC,$3D,$67
	.dc.b	$92,$78,$34,$CC,$77,$3E,$96,$3E,$E7,$E2,$1E,$2A,$EA,$E0,$D9,$F2
	.dc.b	$8F,$C3,$E3,$ED,$AF,$79,$F3,$DF,$3D,$F8,$A7,$C1,$87,$B5,$E8,$D7
	.dc.b	$E8,$D7,$99,$F6,$1E,$3C,$3C,$CE,$8C,$1E,$3C,$7F,$C5,$F8,$78,$E3
	.dc.b	$F2,$0D,$79,$9E,$AC,$3C,$B7,$DE,$3F,$4F,$3B,$6C,$7A,$CF,$94,$F8
	.dc.b	$27,$8B,$63,$2F,$56,$E1,$ED,$BF,$C0,$63,$F3,$2B,$D4,$6C,$F3,$D7
	.dc.b	$76,$9E,$EC,$3B,$BB,$3D,$77,$D5,$FC,$26,$3F,$1B,$0E,$ED,$8F,$CA
	.dc.b	$1C,$EF,$EB,$B4,$FA,$B8,$79,$6B,$8C,$7C,$D3,$A9,$71,$FB,$8F,$E9
	.dc.b	$71,$F9,$0F,$E1,$2F,$D4,$F9,$AF,$A3,$CA,$BD,$78,$FA,$5E,$71,$F8
	.dc.b	$49,$DC,$62,$3A,$1E,$73,$C1,$5D,$9E,$18,$FB,$E7,$86,$3F,$84,$AD
	.dc.b	$57,$23,$F3,$78,$FE,$B0,$F4,$CE,$E3,$B1,$FC,$AB,$D9,$FD,$F3,$FC
	.dc.b	$6D,$AF,$F0,$56,$BB,$9F,$F1,$17,$B0,$E7,$FE,$4B,$2A,$E3,$39,$D3
	.dc.b	$FD,$0F,$18,$AD,$B6,$33,$7D,$27,$9B,$FC,$C7,$1F,$F4,$5E,$1E,$93
	.dc.b	$F9,$FF,$A7,$FE,$33,$9F,$E6,$D8,$F7,$9F,$C1,$87,$11,$CE,$78,$57
	.dc.b	$F8,$EC,$EB,$5C,$2B,$71,$E3,$C3,$F9,$3B,$AC,$F9,$57,$F2,$AA,$F0
	.dc.b	$47,$8B,$1F,$4D,$75,$C3,$BD,$FF,$AA,$23,$BE,$3B,$C7,$D9,$8F,$29
	.dc.b	$E1,$5E,$0B,$FF,$28,$63,$B8,$FE,$9F,$5B,$7A,$4E,$AB,$8C,$6B,$1F
	.dc.b	$AB,$7C,$4F,$4C,$7F,$58,$9F,$61,$EF,$78,$61,$F3,$E3,$CE,$74,$C7
	.dc.b	$C4,$73,$7B,$D7,$43,$15,$FE,$C6,$C7,$CC,$FD,$83,$D8,$7C,$9C,$3C
	.dc.b	$B5,$E1,$BF,$29,$1C,$1E,$5B,$EC,$F8,$0C,$12,$BE,$19,$D8,$F5,$92
	.dc.b	$7A,$3A,$3F,$01,$5F,$61,$F4,$3E,$47,$55,$E0,$73,$87,$A4,$ED,$19
	.dc.b	$8F,$C8,$4F,$16,$CE,$AF,$D1,$1C,$8F,$23,$F3,$48,$E6,$FE,$5B,$F2
	.dc.b	$4F,$B1,$8F,$DF,$79,$61,$F5,$2B,$8C,$6E,$B0,$EA,$3C,$2F,$39,$EB
	.dc.b	$8F,$D6,$9F,$C1,$CF,$C0,$7E,$63,$EA,$3C,$27,$AC,$E8,$79,$DF,$09
	.dc.b	$F8,$87,$06,$8F,$03,$E2,$8F,$C2,$9E,$1B,$F6,$4E,$8E,$11,$97,$C0
	.dc.b	$78,$EB,$E2,$D7,$A2,$38,$19,$7D,$26,$4D,$4F,$83,$1F,$C0,$4F,$C7
	.dc.b	$9F,$93,$3F,$C0,$A3,$D1,$1C,$6F,$CB,$3C,$71,$F1,$CF,$4B,$FC,$04
	.dc.b	$F9,$EF,$E3,$9F,$92,$C3,$CF,$87,$BA,$F8,$61,$F8,$BA,$DD,$7E,$2E
	.dc.b	$F9,$4F,$FA,$16,$7F,$04,$DA,$FE,$32,$E6,$7C,$13,$FE,$D7,$3F,$10
	.dc.b	$59,$FE,$3E,$DB,$D3,$FC,$22,$BA,$7A,$9F,$E6,$BD,$7F,$93,$36,$7F
	.dc.b	$51,$3F,$E9,$71,$C4,$F8,$8D,$46,$3E,$22,$C9,$B3,$2B,$FD,$35,$1A
	.dc.b	$30,$FF,$A4,$FE,$1F,$89,$1F,$F5,$07,$C4,$57,$47,$FA,$8E,$3B,$C7
	.dc.b	$83,$E1,$BC,$3F,$0F,$2B,$FD,$51,$19,$7F,$D5,$51,$97,$FD,$45,$19
	.dc.b	$7F,$D4,$91,$9C,$BE,$06,$4A,$F6,$FC,$37,$BB,$E2,$26,$1F,$EB,$58
	.dc.b	$CB,$FE,$B7,$8C,$BF,$EA,$F8,$F8,$BF,$07,$F1,$BE,$12,$4F,$FA,$F6
	.dc.b	$32,$FF,$AF,$E3,$2F,$FA,$DE,$3E,$A2,$F9,$27,$BD,$86,$C7,$F9,$CB
	.dc.b	$88,$C8,$CB,$FE,$9E,$8F,$EC,$8F,$19,$FC,$2B,$B6,$BD,$DB,$3C,$A3
	.dc.b	$F2,$0F,$4D,$3F,$30,$BD,$DF,$21,$D6,$7F,$DB,$5C,$BF,$86,$F8,$69
	.dc.b	$18,$3F,$DB,$DF,$29,$FE,$C7,$E0,$34,$91,$F9,$AC,$7F,$27,$CF,$F9
	.dc.b	$2F,$81,$DC,$FC,$13,$FD,$76,$E1,$FF,$70,$47,$E5,$9F,$CB,$F4,$47
	.dc.b	$FD,$DC,$7E,$73,$F6,$BF,$53,$DD,$AF,$EB,$ED,$18,$8F,$F6,$A3,$5F
	.dc.b	$F5,$1A,$F8,$1F,$37,$C2,$C9,$F3,$57,$29,$F0,$7C,$9F,$87,$E0,$E5
	.dc.b	$6B,$04,$F9,$7E,$07,$FD,$D9,$06,$92,$7D,$1F,$03,$EB,$23,$CF,$E0
	.dc.b	$BD,$DE,$97,$BF,$D3,$93,$DE,$3E,$17,$95,$D8,$91,$F0,$DF,$0F,$C1
	.dc.b	$7C,$72,$47,$9F,$CB,$F9,$5F,$05,$F2,$FE,$0A,$4F,$AE,$3E,$77,$C1
	.dc.b	$7A,$89,$1F,$3D,$F6,$F9,$BF,$4C,$91,$F0,$FC,$DF,$E2,$3F,$12,$F8
	.dc.b	$3F,$CC,$F3,$23,$15,$F8,$76,$BF,$65,$AF,$E9,$8F,$F8,$02,$F1,$78
	.dc.b	$3D,$0C,$3C,$5C,$7C,$8F,$88,$91,$F7,$23,$73,$FC,$A5,$7C,$17,$A0
	.dc.b	$93,$BF,$84,$AF,$96,$3C,$04,$7D,$4F,$C0,$F9,$19,$1D,$E3,$E1,$FC
	.dc.b	$24,$8F,$13,$FC,$23,$9D,$79,$9F,$85,$3E,$D2,$3F,$1F,$E4,$3E,$6F
	.dc.b	$C2,$C8,$F6,$45,$7F,$85,$1F,$22,$31,$9F,$BE,$7F,$8B,$FC,$6B,$E1
	.dc.b	$7F,$85,$B6,$FF,$87,$B5,$FE,$AA,$D7,$FA,$E3,$9A,$FE,$C9,$F1,$CE
	.dc.b	$5F,$E4,$6D,$AF,$E0,$02,$B8,$79,$9F,$77,$E3,$5F,$09,$FE,$99,$A7
	.dc.b	$B9,$C5,$4E,$B9,$FF,$94,$CE,$63,$13,$C3,$2E,$9C,$9F,$89,$AD,$61
	.dc.b	$D5,$B8,$72,$AF,$14,$E5,$E3,$79,$3F,$9B,$E1,$DA,$F8,$C7,$73,$35
	.dc.b	$C5,$DE,$1E,$1B,$EB,$87,$39,$F5,$D6,$DB,$19,$C3,$F3,$78,$76,$25
	.dc.b	$F2,$AF,$01,$F1,$30,$EC,$F8,$AF,$FA,$3D,$FC,$48,$FF,$09,$35,$E4
	.dc.b	$7F,$85,$7D,$A7,$73,$E3,$BF,$A7,$C6,$1B,$19,$9F,$F0,$BC,$E5,$FF
	.dc.b	$0C,$D3,$E0,$DE,$4F,$89,$B3,$A6,$BE,$46,$BF,$83,$77,$A7,$DD,$6E
	.dc.b	$9F,$C0,$FC,$97,$AC,$E0,$F9,$15,$AC,$3E,$BB,$5F,$C4,$B6,$B8,$CF
	.dc.b	$0C,$65,$7C,$CC,$BD,$51,$F5,$EB,$73,$9B,$E9,$5A,$8C,$56,$F0,$F2
	.dc.b	$46,$09,$97,$AE,$75,$8F,$82,$3F,$3D,$F1,$AF,$81,$FF,$27,$5F,$F2
	.dc.b	$18,$C5,$7F,$80,$9B,$1F,$E4,$86,$BF,$E4,$F7,$3C,$BF,$E5,$76,$CF
	.dc.b	$F8,$42,$1F,$38,$EF,$FF,$25,$4F,$7C,$2E,$1F,$DD,$B2,$B8,$3C,$F7
	.dc.b	$9F,$87,$EF,$E1,$FD,$A1,$CB,$6B,$53,$B6,$C6,$EF,$91,$E0,$3F,$B4
	.dc.b	$76,$E2,$34,$48,$FE,$3A,$69,$C6,$1F,$E6,$3C,$5F,$1D,$E8,$7F,$C6
	.dc.b	$73,$FD,$F1,$CB,$7D,$7D,$4C,$8D,$7A,$D7,$0F,$E5,$78,$7D,$5A,$C3
	.dc.b	$DE,$6D,$7C,$DC,$3F,$C1,$8F,$7F,$4B,$C5,$E9,$78,$DB,$1E,$38,$F3
	.dc.b	$7C,$76,$32,$EF,$68,$CD,$7A,$7D,$3B,$5A,$F4,$7D,$C7,$BF,$C7,$B5
	.dc.b	$EF,$F4,$BE,$11,$F0,$FD,$2B,$1F,$0F,$BF,$0D,$3E,$34,$7C,$9F,$4A
	.dc.b	$47,$C9,$F8,$3B,$97,$9B,$D3,$F9,$C7,$CF,$F4,$ED,$7C,$FF,$0B,$E9
	.dc.b	$7A,$7F,$4D,$B1,$F4,$FD,$2F,$AC,$7D,$7F,$4E,$C7,$F8,$CB,$E3,$9F
	.dc.b	$27,$F9,$26,$C7,$DB,$F2,$B1,$39,$F2,$F5,$EA,$DA,$E4,$F1,$F4,$E4
	.dc.b	$75,$F1,$2F,$07,$DE,$CB,$FC,$07,$3E,$1F,$53,$C1,$EA,$F8,$5F,$17
	.dc.b	$AB,$E3,$F1,$2D,$79,$DF,$37,$A9,$23,$D5,$EA,$DC,$BB,$7C,$77,$AC
	.dc.b	$F6,$7C,$95,$AF,$81,$F1,$DE,$FF,$8F,$F8,$0D,$8F,$ED,$7E,$47,$F3
	.dc.b	$86,$F7,$7E,$A6,$31,$8F,$E9,$94,$67,$BE,$57,$2E,$EB,$7B,$9B,$7F
	.dc.b	$ED,$F1,$38,$AF,$E1,$5C,$5F,$DE,$1F,$AD,$F1,$FF,$57,$D8,$FA,$CD
	.dc.b	$8C,$9F,$6E,$32,$FD,$DF,$8F,$FB,$9E,$B7,$DD,$6C,$64,$FC,$14,$65
	.dc.b	$A7,$E1,$4F,$C2,$7C,$7F,$E1,$5A,$64,$FC,$53,$96,$99,$D3,$F0,$11
	.dc.b	$96,$C7,$FA,$0B,$DF,$7C,$7F,$F2,$26,$C6,$7E,$3F,$13,$23,$83,$B8
	.dc.b	$C5,$7F,$0A,$A7,$C7,$F3,$F3,$BA,$37,$0F,$03,$6B,$83,$78,$3E,$A7
	.dc.b	$8F,$F7,$FC,$9F,$C0,$36,$FF,$0A,$E1,$FF,$18,$DF,$F8,$A0,$7C,$7B
	.dc.b	$E1,$7F,$CD,$D8,$2F,$85,$7E,$25,$FE,$68,$F1,$AE,$0F,$F4,$B8,$CB
	.dc.b	$FE,$38,$3A,$D7,$48,$FE,$C7,$3C,$27,$FC,$68,$68,$E5,$7F,$EE,$48
	.dc.b	$D4,$77,$AE,$53,$97,$FD,$09,$19,$38,$9C,$CE,$B3,$DE,$FD,$95,$C6
	.dc.b	$78,$FB,$75,$4E,$9E,$47,$53,$38,$7A,$B0,$F0,$C7,$BF,$0E,$D3,$98
	.dc.b	$EA,$78,$EB,$A4,$FB,$23,$85,$7B,$1E,$E6,$0F,$5C,$6E,$FC,$D5,$CA
	.dc.b	$BC,$31,$8B,$F0,$5F,$C0,$9E,$97,$C6,$3C,$53,$A7,$D9,$7C,$DF,$3C
	.dc.b	$FA,$B0,$F8,$91,$AA,$E0,$F9,$4F,$87,$1A,$71,$5C,$DF,$11,$EA,$BF
	.dc.b	$3B,$E2,$9E,$18,$7A,$5B,$19,$3E,$1C,$65,$A6,$70,$F6,$1E,$77,$2D
	.dc.b	$33,$5F,$08,$F0,$B9,$69,$98,$F8,$2D,$7D,$27,$B8,$EA,$E5,$F8,$95
	.dc.b	$E8,$9E,$91,$C8,$F8,$05,$39,$3E,$47,$27,$6B,$F9,$0F,$43,$B9,$A3
	.dc.b	$11,$C5,$E7,$5E,$A7,$47,$D2,$C3,$EB,$E1,$F2,$6B,$85,$F6,$AD,$61
	.dc.b	$E9,$C3,$D9,$18,$3C,$27,$63,$D5,$3C,$6F,$A4,$F8,$67,$47,$AF,$0F
	.dc.b	$0B,$F5,$E3,$55,$C6,$30,$74,$9D,$E1,$F2,$A3,$96,$1E,$57,$DB,$87
	.dc.b	$81,$E1,$19,$7C,$27,$C6,$C3,$BD,$F4,$7D,$B5,$C6,$7F,$88,$F8,$EF
	.dc.b	$27,$F2,$AD,$8F,$DC,$B5,$DC,$19,$EF,$FF,$64,$CF,$15,$B8,$C4,$A3
	.dc.b	$1D,$1F,$C3,$A3,$17,$CD,$57,$0A,$DD,$FF,$15,$C7,$FC,$DF,$57,$D9
	.dc.b	$E1,$AF,$F2,$BC,$3B,$B7,$2E,$CF,$F2,$FB,$E3,$7F,$C9,$5B,$1B,$C3
	.dc.b	$F9,$15,$F2,$9F,$21,$A2,$E1,$E0,$7D,$11,$BB,$F8,$78,$78,$27,$71
	.dc.b	$E7,$3A,$C6,$5D,$DE,$5F,$25,$F5,$72,$7F,$3B,$BE,$CF,$F5,$78,$D3
	.dc.b	$4E,$B3,$1E,$67,$5C,$3D,$CF,$3B,$F1,$3C,$4D,$99,$9E,$53,$F2,$6B
	.dc.b	$BE,$1F,$0E,$BC,$B3,$D0,$D3,$DE,$3E,$2D,$F6,$C3,$D3,$1C,$67,$C1
	.dc.b	$5C,$63,$78,$71,$9F,$C7,$46,$EB,$E0,$9C,$B0,$F1,$9F,$09,$CE,$1E
	.dc.b	$C9,$EC,$71,$7C,$46,$67,$8D,$7B,$A7,$C2,$FC,$2C,$3D,$6F,$C3,$23
	.dc.b	$82,$47,$07,$D5,$1E,$89,$EA,$F0,$9D,$1F,$84,$7B,$38,$7E,$5B,$99
	.dc.b	$EB,$18,$21,$F2,$23,$47,$17,$F5,$35,$ED,$7E,$51,$8C,$3D,$CF,$94
	.dc.b	$CE,$1F,$22,$37,$5F,$1D,$DC,$F7,$38,$1E,$B3,$C7,$1F,$3B,$0F,$A5
	.dc.b	$7F,$3A,$B9,$C6,$1E,$73,$BA,$EB,$7F,$6D,$F1,$DF,$17,$E2,$9E,$0C
	.dc.b	$3E,$6B,$EB,$32,$79,$4C,$3A,$AE,$D1,$E7,$32,$FC,$87,$A3,$F2,$6B
	.dc.b	$C5,$1D,$1F,$88,$F6,$8F,$88,$FE,$9A,$DA,$FD,$B6,$C7,$C6,$6B,$E0
	.dc.b	$31,$1B,$23,$F2,$B1,$FF,$28,$C7,$F1,$0F,$66,$CF,$E8,$9A,$FE,$DE
	.dc.b	$BB,$8E,$27,$C7,$F0,$8C,$7C,$7D,$93,$66,$57,$F9,$F6,$34,$61,$FF
	.dc.b	$3B,$7C,$7F,$12,$3F,$E8,$3F,$8F,$AE,$8F,$F4,$34,$77,$8F,$07,$C7
	.dc.b	$78,$7E,$3E,$57,$FA,$2E,$32,$FF,$A3,$63,$2F,$FA,$16,$32,$FF,$A1
	.dc.b	$E3,$39,$7C,$0C,$95,$ED,$F8,$FF,$77,$C2,$4C,$3F,$D3,$11,$97,$FD
	.dc.b	$35,$19,$7F,$D2,$51,$F1,$7E,$3B,$E3,$7C,$7C,$9F,$F5,$04,$65,$FF
	.dc.b	$51,$46,$5F,$F4,$D4,$7D,$4F,$8F,$7C,$25,$24,$FF,$67,$CF,$F0,$7C
	.dc.b	$9F,$E7,$87,$F0,$B5,$C2,$75,$C7,$FD,$8E,$31,$1F,$E7,$48,$C5,$FE
	.dc.b	$0A,$F9,$46,$27,$89,$FD,$D6,$9F,$07,$E5,$3C,$71,$FE,$B1,$AF,$C2
	.dc.b	$7C,$7C,$8F,$F5,$F4,$7E,$1F,$E3,$E4,$7F,$B0,$5F,$C5,$7C,$87,$E2
	.dc.b	$FA,$24,$FF,$B1,$5F,$C7,$7C,$67,$E3,$E3,$F2,$1D,$31,$FF,$65,$1F
	.dc.b	$95,$F9,$0F,$2B,$E8,$DD,$18,$8F,$80,$7F,$A5,$1C,$47,$E8,$1F,$8A
	.dc.b	$DF,$88,$E8,$49,$E9,$2F,$89,$F0,$7C,$3F,$87,$E2,$65,$6A,$3C,$9F
	.dc.b	$0F,$23,$8B,$E7,$F8,$9F,$47,$C4,$49,$EA,$FA,$FE,$27,$D9,$EA,$FB
	.dc.b	$7D,$69,$5E,$28,$F8,$1E,$64,$8F,$2B,$F0,$FC,$8F,$89,$F1,$32,$7D
	.dc.b	$2F,$C7,$F3,$7E,$47,$C4,$FC,$9F,$89,$95,$EE,$8F,$99,$F1,$32,$3E
	.dc.b	$0B,$A7,$E5,$F2,$FC,$EF,$3B,$E9,$12,$3E,$3F,$9F,$FC,$27,$E2,$1F
	.dc.b	$11,$49,$3F,$5E,$1F,$13,$B6,$46,$5D,$BE,$05,$F1,$1D,$49,$1B,$3A
	.dc.b	$7C,$3F,$84,$8F,$7F,$89,$AF,$38,$B9,$FF,$22,$B5,$63,$C7,$FD,$84
	.dc.b	$3E,$27,$D8,$67,$0F,$F6,$FC,$3E,$23,$E0,$12,$3F,$56,$3A,$FA,$BF
	.dc.b	$06,$34,$FB,$5F,$E6,$B1,$A8,$F8,$1F,$0F,$F2,$FE,$26,$47,$C9,$8F
	.dc.b	$9C,$61,$F8,$D1,$87,$C3,$F1,$EF,$8A,$FE,$82,$D8,$CC,$7F,$44,$F9
	.dc.b	$0F,$F3,$B6,$DF,$D4,$2B,$55,$FE,$F6,$6B,$FE,$6F,$0E,$1F,$F3,$93
	.dc.b	$5F,$F3,$F7,$C7,$BE,$1B,$FC,$F5,$0F,$7D,$D5,$FE,$79,$AD,$74,$7F
	.dc.b	$A1,$67,$35,$CA,$F8,$C6,$21,$4E,$0F,$F4,$14,$E6,$39,$CF,$7C,$3B
	.dc.b	$5F,$26,$E1,$DD,$F1,$61,$C9,$CD,$71,$3C,$53,$B6,$99,$3C,$73,$C5
	.dc.b	$CE,$1E,$3B,$F0,$46,$2F,$BB,$CC,$F5,$C6,$9B,$79,$C3,$D7,$7E,$DC
	.dc.b	$3B,$56,$8F,$A8,$53,$EB,$4F,$C2,$6D,$FD,$27,$ED,$7B,$CF,$01,$F2
	.dc.b	$1C,$A7,$FD,$57,$18,$7A,$CE,$E7,$4D,$8F,$0D,$BE,$19,$E6,$73,$8C
	.dc.b	$36,$33,$3C,$1A,$FE,$37,$EC,$9C,$9F,$E1,$6E,$8F,$EC,$5F,$5B,$FD
	.dc.b	$E1,$FD,$6C,$E9,$FE,$85,$19,$AF,$E8,$22,$74,$FC,$66,$BD,$EB,$FC
	.dc.b	$5E,$FF,$B2,$E3,$2F,$FB,$37,$E4,$9F,$0D,$E3,$87,$2B,$E4,$DE,$66
	.dc.b	$65,$7C,$33,$FE,$AD,$5A,$7F,$82,$C6,$67,$F9,$6C,$E5,$F4,$C6,$6F
	.dc.b	$FC,$6F,$39,$7F,$C8,$B1,$C6,$7F,$68,$D7,$D1,$87,$B1,$FF,$78,$46
	.dc.b	$5F,$F8,$02,$3E,$19,$EB,$FF,$C0,$0C,$7F,$DE,$D1,$89,$DB,$5F,$D7
	.dc.b	$9C,$67,$FD,$E5,$83,$C8,$FF,$2D,$D6,$EB,$F7,$86,$CF,$F0,$E4,$6A
	.dc.b	$BC,$11,$93,$AC,$F7,$AE,$97,$E4,$AD,$CF,$5A,$FE,$18,$71,$AD,$BC
	.dc.b	$CF,$0C,$73,$BF,$04,$7B,$2B,$99,$1C,$10,$F1,$BE,$E8,$E3,$1F,$0B
	.dc.b	$C6,$E9,$82,$11,$E5,$E8,$C3,$A8,$CD,$F7,$7A,$9F,$12,$BD,$CF,$84
	.dc.b	$F4,$C7,$86,$33,$1C,$0F,$C9,$DF,$91,$D9,$CC,$F8,$47,$C6,$C3,$C4
	.dc.b	$79,$5F,$91,$1C,$5D,$57,$77,$8E,$1E,$0C,$7E,$6D,$F1,$39,$D7,$AE
	.dc.b	$BC,$27,$93,$0F,$3B,$8A,$E3,$3B,$8E,$E7,$9E,$BA,$47,$85,$B1,$98
	.dc.b	$E1,$7E,$3C,$3D,$0D,$C3,$BD,$74,$3C,$47,$A8,$F7,$BA,$8E,$11,$9B
	.dc.b	$F5,$BF,$44,$E5,$87,$A1,$C4,$F0,$9F,$2C,$76,$6C,$EE,$3E,$B3,$5F
	.dc.b	$85,$7E,$87,$C6,$78,$70,$F8,$07,$B9,$CD,$74,$9E,$0F,$9A,$3A,$C6
	.dc.b	$27,$8C,$79,$8E,$2F,$39,$E1,$1E,$2C,$3E,$13,$A3,$84,$69,$F8,$35
	.dc.b	$D5,$FA,$46,$AB,$BB,$E2,$AE,$75,$C7,$0F,$8D,$5E,$13,$75,$E1,$3E
	.dc.b	$01,$D0,$E4,$60,$F9,$8F,$17,$47,$19,$F8,$95,$E0,$9E,$B3,$E8,$8C
	.dc.b	$3C,$0C,$1D,$70,$FA,$8E,$2B,$9B,$9C,$3E,$03,$E2,$3E,$01,$CF,$0F
	.dc.b	$6B,$83,$47,$AA,$79,$93,$0F,$54,$68,$F5,$38,$BF,$1D,$74,$9E,$53
	.dc.b	$9B,$F1,$D6,$A7,$EC,$1A,$3C,$4E,$6F,$E9,$46,$8E,$E6,$A7,$99,$30
	.dc.b	$F8,$F1,$E5,$3E,$25,$F8,$63,$37,$D6,$3E,$7E,$1E,$7A,$EA,$78,$C8
	.dc.b	$60,$87,$43,$C5,$5D,$2F,$C5,$1F,$46,$30,$47,$04,$C3,$F8,$D6,$3F
	.dc.b	$CB,$4B,$87,$F1,$97,$E8,$4F,$63,$84,$7A,$DA,$66,$76,$FA,$0F,$31
	.dc.b	$FD,$83,$A7,$F9,$EB,$FE,$63,$C7,$FA,$04,$62,$35,$7F,$E3,$49,$CB
	.dc.b	$FD,$16,$1C,$AF,$2B,$FC,$17,$0B,$E5,$39,$D3,$8C,$FF,$4D,$8C,$3B
	.dc.b	$BE,$91,$8A,$E3,$7D,$31,$FE,$F1,$93,$D5,$F1,$46,$1B,$1C,$A7,$83
	.dc.b	$72,$FE,$9E,$F1,$CA,$D7,$07,$C9,$5C,$1B,$5A,$61,$FD,$F7,$1F,$4E
	.dc.b	$1F,$E0,$B6,$E1,$F1,$4E,$D3,$FE,$03,$6C,$EE,$FB,$4F,$CA,$6C,$E8
	.dc.b	$E5,$87,$96,$75,$87,$F6,$8C,$3B,$C7,$F7,$7A,$E0,$DA,$E1,$3A,$8E
	.dc.b	$27,$F7,$83,$47,$F7,$C9,$EE,$D3,$E0,$9B,$38,$E1,$F3,$DF,$86,$FA
	.dc.b	$EF,$C9,$87,$AB,$0F,$45,$F9,$2F,$8B,$F4,$DF,$89,$1A,$3C,$F3,$F1
	.dc.b	$30,$F4,$57,$13,$93,$EB,$33,$3C,$6B,$C3,$71,$EA,$79,$AB,$C1,$87
	.dc.b	$9E,$32,$47,$04,$8F,$87,$19,$78,$91,$F9,$26,$4F,$8F,$1C,$5C,$BF
	.dc.b	$1D,$FC,$46,$1E,$D3,$E3,$3E,$53,$8C,$F5,$7F,$CB,$77,$ED,$78,$DF
	.dc.b	$B6,$FA,$12,$B8,$61,$F1,$2B,$A1,$F4,$0E,$A7,$92,$3D,$B3,$E7,$8D
	.dc.b	$1E,$03,$8C,$FC,$FC,$3C,$F1,$EF,$37,$1D,$DC,$3B,$BF,$0D,$FD,$8A
	.dc.b	$F8,$11,$8C,$3E,$FB,$EB,$9F,$19,$2F,$CE,$79,$1F,$C1,$4F,$08,$F8
	.dc.b	$67,$8E,$3B,$4F,$38,$F7,$10,$F3,$D7,$AB,$0F,$91,$1D,$B0,$F5,$46
	.dc.b	$23,$75,$E7,$7A,$1E,$3C,$3E,$A4,$FC,$9A,$E3,$87,$A6,$B8,$9E,$5A
	.dc.b	$F4,$57,$CF,$9F,$5C,$F4,$AE,$2F,$9D,$F9,$4F,$13,$58,$7A,$67,$B4
	.dc.b	$EA,$BD,$11,$F0,$AB,$E2,$C6,$2F,$C3,$3F,$8D,$3C,$33,$A7,$11,$EC
	.dc.b	$7F,$17,$7D,$E7,$C8,$FD,$A7,$C8,$71,$3E,$BC,$74,$AF,$89,$3C,$67
	.dc.b	$F1,$86,$CF,$94,$7E,$39,$E9,$1E,$78,$F7,$E1,$F4,$5E,$18,$7C,$67
	.dc.b	$9C,$F6,$9D,$36,$3C,$AD,$3A,$3C,$AB,$C9,$7F,$1E,$7E,$A6,$1F,$06
	.dc.b	$BF,$09,$5A,$8F,$93,$1E,$E7,$4F,$DE,$9E,$A7,$7A,$FC,$11,$D4,$F9
	.dc.b	$B1,$EF,$BF,$C5,$C7,$C7,$3E,$7B,$EA,$6C,$7A,$9B,$3D,$9A,$76,$AF
	.dc.b	$2C,$FC,$C3,$F6,$77,$F3,$27,$51,$F3,$0F,$E7,$8F,$F5,$6C,$BF,$8B
	.dc.b	$1F,$C9,$4F,$87,$1E,$73,$EC,$47,$9D,$F8,$B8,$79,$DA,$79,$CB,$87
	.dc.b	$9D,$B3,$E2,$7E,$94,$F1,$7E,$54,$66,$7F,$8A,$76,$BB,$FF,$C2,$15
	.dc.b	$23,$0E,$0F,$F6,$C3,$87,$0D,$78,$47,$FB,$FB,$6F,$F0,$D7,$7B,$BF
	.dc.b	$15,$FE,$E1,$8C,$3F,$EE,$3C,$95,$FE,$21,$C3,$B4,$62,$7C,$38,$72
	.dc.b	$57,$C9,$4E,$DC,$E1,$BB,$E7,$87,$F8,$B7,$EB,$1F,$11,$ED,$BF,$F7
	.dc.b	$A4,$61,$C3,$5F,$84,$D7,$E3,$B9,$C3,$C1,$EF,$B5,$F3,$36,$BD,$D8
	.dc.b	$78,$70,$F5,$77,$3C,$1F,$F1,$E5,$7C,$F8,$C7,$27,$95,$B1,$99,$DB
	.dc.b	$F6,$39,$1E,$0B,$C1,$F6,$46,$1B,$19,$9F,$B9,$39,$7C,$1E,$03,$C1
	.dc.b	$F0,$36,$33,$18,$3E,$47,$7B,$C0,$FA,$13,$FE,$59,$8C,$6D,$A9,$CC
	.dc.b	$84,$E9,$FC,$73,$78,$7E,$D3,$6B,$F0,$DF,$64,$F6,$3F,$09,$5A,$C2
	.dc.b	$D7,$F9,$A7,$BD,$E0,$63,$2F,$C2,$F8,$2F,$0B,$10,$AD,$DF,$D6,$8D
	.dc.b	$CE,$A7,$31,$FA,$36,$BF,$A8,$6B,$E1,$BF,$7B,$D5,$E4,$78,$F1,$7C
	.dc.b	$AB,$CC,$CE,$3E,$1B,$FF,$46,$78,$7F,$EA,$77,$FD,$1A,$7F,$A9,$1C
	.dc.b	$11,$FF,$5C,$1A,$23,$9F,$07,$FD,$41,$D7,$FE,$8F,$9F,$F2,$86,$2F
	.dc.b	$13,$F5,$D5,$FE,$9C,$9E,$C7,$4B,$ED,$E1,$7F,$B1,$89,$87,$FA,$F2
	.dc.b	$74,$48,$CB,$FE,$E1,$31,$1B,$8C,$C7,$FB,$26,$7A,$CE,$4E,$95,$E8
	.dc.b	$9F,$F5,$51,$E8,$BF,$3C,$FF,$AD,$6F,$FD,$77,$3F,$C2,$E7,$2F,$0A
	.dc.b	$E4,$FC,$C3,$C6,$79,$67,$C7,$87,$FB,$5E,$31,$7E,$36,$90,$F4,$1C
	.dc.b	$8B,$3C,$AF,$FD,$AA,$FC,$B8,$FA,$66,$EB,$B3,$C2,$BB,$3D,$2F,$CE
	.dc.b	$FE,$9F,$0F,$F4,$43,$FC,$7B,$0F,$6B,$EB,$C3,$C1,$8F,$FB,$85,$F0
	.dc.b	$57,$CC,$9F,$46,$3E,$02,$5F,$9B,$1F,$94,$49,$F9,$F7,$FC,$65,$F2
	.dc.b	$47,$FB,$9C,$FC,$56,$3E,$68,$F6,$CF,$89,$FF,$73,$9E,$59,$F6,$DF
	.dc.b	$83,$1F,$E0,$55,$F6,$2B,$D9,$1E,$E7,$E0,$1F,$7D,$C3,$DA,$B9,$E1
	.dc.b	$E9,$8F,$F7,$A4,$79,$0F,$33,$F8,$16,$BE,$33,$E1,$C7,$17,$FD,$E1
	.dc.b	$8F,$CE,$C3,$F4,$97,$E0,$8C,$57,$D0,$AF,$D2,$E1,$F1,$6F,$70,$8F
	.dc.b	$64,$63,$1B,$5C,$6B,$D4,$7E,$96,$3E,$55,$F6,$C3,$C1,$3F,$87,$AF
	.dc.b	$EC,$58,$72,$9C,$DF,$0A,$D4,$7E,$07,$1F,$B5,$1F,$9F,$9D,$1B,$AF
	.dc.b	$AC,$73,$8F,$F6,$91,$F4,$23,$53,$F3,$AB,$A4,$FC,$E3,$F0,$B1,$FB
	.dc.b	$63,$E6,$BC,$5A,$FB,$8F,$CF,$1F,$42,$3F,$03,$1F,$5E,$3F,$71,$1E
	.dc.b	$C8,$E3,$87,$E6,$67,$89,$31,$FC,$53,$F8,$E3,$E9,$CF,$9D,$C1,$E1
	.dc.b	$8D,$E1,$FA,$48,$F8,$E6,$AB,$F9,$57,$5B,$1F,$E4,$CD,$9F,$E2,$2D
	.dc.b	$7F,$0E,$1C,$73,$BC,$3F,$F0,$9C,$E6,$BF,$DE,$95,$A9,$CB,$71,$FE
	.dc.b	$C8,$DA,$FF,$1F,$1F,$E3,$83,$FD,$5D,$97,$F9,$7D,$FF,$33,$1F,$E4
	.dc.b	$9C,$BF,$C9,$6F,$F9,$52,$BF,$CC,$3E,$27,$F9,$AB,$2F,$F4,$D5,$FF
	.dc.b	$90,$4F,$F2,$ED,$74,$9F,$F2,$EC,$FF,$A0,$67,$B4,$FF,$97,$EF,$FD
	.dc.b	$71,$3F,$EA,$8C,$BF,$CB,$75,$FE,$FA,$7F,$CB,$8F,$F9,$CA,$BC,$47
	.dc.b	$F9,$72,$FC,$59,$7F,$A7,$6B,$8C,$FF,$BF,$9E,$2F,$B0,$E3,$87,$B0
	.dc.b	$F0,$CF,$F9,$BE,$3D,$07,$F9,$BE,$79,$BF,$EA,$27,$9B,$E9,$BF,$F5
	.dc.b	$DC,$F8,$5F,$41,$23,$FC,$29,$87,$D9,$C7,$F9,$7B,$F0,$F1,$F8,$B7
	.dc.b	$FE,$97,$AF,$F3,$F6,$3F,$E7,$17,$FD,$A7,$8F,$8B,$1F,$F4,$01,$FE
	.dc.b	$1F,$86,$1D,$5F,$C7,$47,$C7,$AF,$6E,$3F,$A4,$9F,$1E,$1F,$8A,$C3
	.dc.b	$A6,$3C,$DF,$6E,$3E,$5A,$F8,$58,$FA,$DF,$84,$F8,$63,$C1,$3F,$2E
	.dc.b	$3E,$1E,$1D,$EB,$CB,$87,$79,$E4,$7A,$A3,$C7,$3F,$12,$7E,$14,$7F
	.dc.b	$A3,$8F,$CB,$DF,$E1,$6B,$8E,$1E,$77,$C5,$5D,$6B,$EF,$63,$F0,$1F
	.dc.b	$C3,$1F,$9B,$9F,$88,$79,$9F,$E5,$8F,$2C,$3F,$B1,$BE,$58,$F7,$9E
	.dc.b	$63,$E6,$9D,$9E,$53,$F0,$63,$FD,$DF,$3F,$A6,$7C,$06,$CC,$9E,$03
	.dc.b	$A6,$3F,$49,$F8,$51,$1F,$83,$7F,$95,$7F,$DD,$37,$E0,$3D,$6F,$4B
	.dc.b	$F1,$3A,$3F,$D7,$27,$E7,$0F,$98,$FE,$37,$1F,$DC,$C6,$63,$FD,$C4
	.dc.b	$D9,$FF,$6D,$E1,$E5,$8F,$82,$7D,$CC,$3C,$D5,$CF,$0F,$35,$FF,$B3
	.dc.b	$EB,$F8,$71,$E3,$26,$1E,$07,$11,$FC,$12,$3F,$21,$1E,$68,$FE,$18
	.dc.b	$FA,$8D,$E3,$FA,$8C,$3F,$DA,$55,$CE,$BC,$0D,$9F,$E0,$B3,$F4,$EB
	.dc.b	$F8,$0D,$7D,$F7,$FD,$D7,$3F,$8B,$7F,$84,$46,$71,$F9,$6F,$97,$0F
	.dc.b	$C7,$4F,$E1,$09,$87,$E8,$A7,$E9,$C7,$BA,$7E,$1D,$F8,$A3,$F2,$67
	.dc.b	$88,$FC,$21,$BC,$3C,$D8,$FE,$64,$F0,$CF,$BE,$7F,$07,$3A,$BF,$C1
	.dc.b	$47,$27,$CD,$1F,$9A,$9F,$64,$F9,$AF,$D3,$87,$CE,$C3,$EF,$CF,$CC
	.dc.b	$3F,$88,$E1,$F4,$1F,$B2,$FC,$CA,$F5,$C7,$E7,$07,$83,$FE,$E2,$7F
	.dc.b	$DD,$67,$FA,$3D,$C3,$FE,$EC,$CB,$FD,$D1,$E2,$7F,$B9,$63,$8B,$67
	.dc.b	$FE,$00,$47,$FB,$DE,$30,$7F,$BD,$7C,$4F,$F7,$A6,$9F,$EF,$36,$EF
	.dc.b	$BA,$FF,$4F,$78,$7F,$EE,$98,$FF,$72,$46,$3C,$2F,$F7,$4F,$B9,$FE
	.dc.b	$FF,$AF,$F7,$84,$E8,$FF,$79,$CF,$96,$33,$1F,$EC,$DF,$0F,$FD,$CE
	.dc.b	$FF,$31,$0E,$1F,$E6,$8D,$78,$78,$7F,$EE,$C7,$F9,$47,$85,$FE,$E7
	.dc.b	$7F,$DD,$CD,$73,$E1,$FF,$BB,$0F,$F7,$23,$FE,$E6,$21,$E1,$A7,$87
	.dc.b	$FE,$EB,$8F,$27,$85,$FE,$DD,$7F,$DC,$8D,$D9,$E1,$7F,$B7,$F6,$FF
	.dc.b	$62,$4E,$EB,$FD,$D3,$E0,$FF,$B6,$A3,$85,$6E,$57,$FE,$E5,$BF,$F6
	.dc.b	$67,$7F,$FB,$5A,$7F,$DE,$4D,$7F,$DA,$D1,$F1,$DC,$F7,$FF,$B6,$A7
	.dc.b	$FD,$C3,$4A,$FF,$6F,$1F,$EE,$77,$FD,$B7,$DD,$FE,$DA,$BF,$0C,$E9
	.dc.b	$A7,$47,$87,$7F,$FB,$85,$F2,$E1,$FE,$F4,$9F,$F1,$D9,$FE,$47,$3F
	.dc.b	$C6,$C7,$F8,$E3,$AB,$FC,$74,$7F,$9C,$7B,$FF,$CC,$BD,$7F,$E4,$B9
	.dc.b	$FF,$45,$9F,$E5,$C1,$DF,$FE,$6D,$70,$FF,$93,$73,$FF,$22,$CF,$FB
	.dc.b	$96,$FA,$63,$FD,$96,$FF,$CB,$87,$F9,$58,$E2,$7F,$A5,$4B,$1F,$E6
	.dc.b	$C9,$FF,$38,$C7,$03,$C7,$D7,$FE,$6F,$3D,$31,$FE,$57,$9F,$F2,$A4
	.dc.b	$7F,$9D,$A3,$FC,$C3,$3E,$F7,$9B,$EA,$3C,$11,$CE,$33,$87,$AF,$1F
	.dc.b	$87,$5F,$EA,$13,$D1,$5B,$7F,$CF,$47,$FA,$8C,$FF,$5E,$9A,$AF,$95
	.dc.b	$8F,$FA,$65,$EB,$1E,$33,$A1,$EC,$77,$5F,$E9,$DB,$F1,$63,$FE,$88
	.dc.b	$AF,$F4,$95,$7B,$30,$FF,$46,$C7,$DD,$C7,$E1,$B6,$7B,$C7,$CB,$8F
	.dc.b	$39,$F8,$43,$FD,$05,$87,$FA,$E5,$C6,$3F,$69,$D9,$FE,$81,$8D,$37
	.dc.b	$0F,$23,$F5,$67,$DC,$FB,$1A,$FC,$57,$A3,$67,$9E,$3E,$E8,$FC,$AD
	.dc.b	$F9,$DF,$79,$F2,$CE,$93,$F1,$5F,$94,$73,$BF,$1B,$E5,$39,$E1,$F8
	.dc.b	$53,$E6,$C7,$FA,$96,$3A,$1C,$4D,$9F,$24,$FF,$5A,$9F,$EC,$D8,$FA
	.dc.b	$07,$E3,$2B,$CD,$1F,$06,$BC,$98,$FD,$48,$F9,$B7,$E4,$C3,$DD,$7F
	.dc.b	$96,$C3,$9D,$78,$8F,$A3,$87,$C4,$7F,$C5,$07,$E3,$0F,$C3,$1F,$0A
	.dc.b	$FF,$DB,$B5,$E2,$8F,$80,$FE,$55,$FF,$56,$E1,$E3,$C7,$E8,$C7,$6A
	.dc.b	$F2,$3D,$A3,$FD,$E3,$5C,$63,$D9,$3D,$B0,$F9,$51,$E1,$3F,$73,$7F
	.dc.b	$AF,$7D,$47,$EB,$5F,$F7,$67,$7F,$FA,$FE,$07,$EC,$5C,$36,$3F,$D8
	.dc.b	$E7,$FA,$F8,$8F,$F9,$4E,$3F,$D7,$87,$FA,$F6,$3F,$C6,$C7,$F9,$EB
	.dc.b	$B7,$FC,$2F,$5F,$E5,$B3,$FC,$B9,$DD,$FE,$17,$BF,$F0,$47,$7F,$F8
	.dc.b	$36,$F8,$D7,$F8,$84,$FF,$0D,$9F,$E0,$C9,$FF,$08,$9F,$E1,$AC,$BF
	.dc.b	$C2,$D1,$B8,$FF,$14,$90,$FF,$23,$1F,$E4,$8B,$FF,$08,$47,$FB,$1A
	.dc.b	$3F,$C2,$07,$F9,$FE,$BC,$2F,$F9,$92,$3C,$37,$FE,$21,$BF,$F2,$76
	.dc.b	$7F,$E8,$DC,$3D,$57,$FE,$33,$BF,$6E,$3F,$E1,$E3,$FD,$17,$7F,$E1
	.dc.b	$E8,$FF,$0D,$5F,$F8,$A6,$BF,$CD,$11,$E9,$79,$61,$FE,$28,$C3,$A9
	.dc.b	$E8,$8F,$F4,$15,$FF,$B0,$AB,$E6,$56,$F0,$F0,$61,$FE,$54,$7F,$C4
	.dc.b	$F7,$FE,$3F,$3F,$D9,$07,$80,$F7,$1F,$1E,$BD,$33,$FE,$6C,$7F,$C6
	.dc.b	$58,$7F,$96,$E7,$FC,$60,$F8,$DF,$F1,$AD,$FF,$93,$70,$FF,$23,$C7
	.dc.b	$FA,$E0,$E9,$1E,$5B,$E9,$1F,$E7,$D3,$FC,$D9,$7F,$11,$FB,$8F,$F9
	.dc.b	$31,$F3,$9F,$E5,$53,$27,$A4,$CB,$F8,$5C,$7E,$CC,$7E,$01,$F1,$BF
	.dc.b	$E7,$B3,$C0,$7A,$0F,$5E,$3F,$5A,$7F,$08,$78,$F0,$E2,$7D,$9C,$3D
	.dc.b	$D1,$F8,$4A,$FF,$2B,$9D,$31,$F8,$CF,$CF,$7E,$31,$8A,$FC,$BC,$7E
	.dc.b	$80,$F8,$C7,$E0,$0F,$F5,$86,$1E,$0A,$F1,$DF,$82,$7E,$B9,$B3,$F3
	.dc.b	$98,$7E,$01,$FD,$2D,$F1,$7F,$0F,$5F,$8D,$3F,$0E,$F6,$7D,$4F,$C5
	.dc.b	$3F,$47,$08,$FF,$58,$9F,$EE,$08,$E0,$7A,$4E,$0F,$C7,$3B,$57,$FA
	.dc.b	$01,$FC,$5C,$79,$B1,$FC,$A1,$FA,$17,$EE,$57,$E6,$EB,$81,$F8,$78
	.dc.b	$E0,$7B,$2B,$CF,$3E,$58,$F6,$9F,$12,$3D,$AF,$0C,$3E,$13,$F7,$CF
	.dc.b	$A9,$87,$F1,$4B,$EF,$1E,$C3,$F1,$6F,$EF,$EB,$F4,$33,$FE,$50,$9F
	.dc.b	$E8,$38,$FF,$00,$9E,$F7,$FC,$7F,$1F,$6D,$7E,$58,$F6,$CF,$A0,$FB
	.dc.b	$6F,$FA,$67,$BF,$FD,$2A,$4E,$FF,$F4,$75,$FF,$A5,$A7,$FD,$35,$15
	.dc.b	$FF,$27,$9F,$E9,$B3,$27,$F9,$74,$FF,$4D,$96,$7F,$D3,$6F,$FA,$90
	.dc.b	$FF,$46,$9F,$E9,$32,$1D,$3C,$47,$7F,$FA,$57,$BE,$4E,$B2,$FF,$33
	.dc.b	$3F,$D2,$2B,$6F,$F4,$A8,$CB,$FE,$94,$EE,$FF,$48,$3F,$E9,$2E,$0F
	.dc.b	$37,$67,$FA,$56,$35,$1F,$E8,$F7,$FC,$ED,$B3,$BB,$FD,$29,$1F,$D5
	.dc.b	$A3,$1D,$DB,$8F,$F4,$97,$77,$FA,$45,$FF,$50,$B5,$FF,$54,$35,$FF
	.dc.b	$53,$B7,$AF,$FD,$03,$CD,$E9,$F0,$9E,$17,$F9,$FF,$6F,$78,$8C,$CF
	.dc.b	$F9,$F0,$FA,$15,$FE,$8F,$27,$5F,$C2,$C7,$FB,$68,$FA,$B7,$AD,$D2
	.dc.b	$FE,$47,$73,$CA,$C5,$FF,$74,$BD,$D7,$FA,$32,$73,$3C,$EF,$79,$B0
	.dc.b	$E5,$7F,$4C,$F0,$E1,$CA,$FF,$D2,$F5,$BC,$39,$CE,$67,$FD,$98,$D7
	.dc.b	$8E,$1E,$B7,$C8,$D7,$FC,$6D,$DB,$FE,$5E,$7F,$C5,$A7,$F8,$E0,$AE
	.dc.b	$6F,$FC,$3F,$1F,$E2,$4E,$BF,$F3,$19,$F1,$A3,$FC,$3E,$FF,$91,$23
	.dc.b	$FC,$36,$FF,$8F,$4F,$F0,$FB,$E0,$78,$BF,$E4,$53,$07,$F9,$9B,$BF
	.dc.b	$FC,$7B,$1A,$8F,$1B,$67,$C7,$3A,$9E,$07,$F9,$68,$FF,$1A,$1E,$B8
	.dc.b	$DB,$93,$D4,$70,$AF,$F2,$04,$F3,$8F,$19,$0E,$71,$F0,$FB,$FF,$C9
	.dc.b	$8F,$08,$E9,$3C,$5D,$DF,$9E,$B6,$FC,$28,$C9,$E1,$70,$F9,$0F,$01
	.dc.b	$4D,$98,$9E,$CE,$48,$E0,$8F,$90,$95,$E7,$24,$F8,$27,$71,$89,$FA
	.dc.b	$87,$80,$87,$63,$9C,$F1,$69,$C0,$D4,$7C,$5E,$DF,$F2,$E4,$7F,$9C
	.dc.b	$0F,$51,$C0,$FB,$A7,$C3,$39,$57,$AB,$9B,$FC,$BA,$FF,$98,$1B,$06
	.dc.b	$9F,$E5,$C9,$FF,$2F,$C7,$F9,$2A,$30,$FF,$61,$AF,$F0,$23,$FE,$0A
	.dc.b	$C7,$FA,$0C,$7F,$50,$E5,$FF,$00,$63,$FE,$1B,$7F,$C0,$7D,$DF,$D2
	.dc.b	$E7,$FB,$05,$FF,$9C,$B1,$FE,$B8,$FF,$71,$EF,$FE,$C5,$1C,$EB,$FC
	.dc.b	$77,$5F,$E0,$AA,$FE,$AE,$7F,$84,$8F,$EF,$38,$7F,$4E,$7F,$AF,$E3
	.dc.b	$E4,$8F,$EA,$51,$FE,$30,$C3,$FC,$18,$F2,$3F,$BD,$B8,$C7,$CF,$5F
	.dc.b	$DD,$EB,$FC,$E5,$7F,$E6,$2B,$FE,$BD,$18,$3F,$B3,$CE,$EF,$FB,$54
	.dc.b	$FF,$89,$70,$F4,$D7,$C3,$7F,$C6,$11,$E3,$3E,$51,$FD,$FC,$F9,$D1
	.dc.b	$EB,$BF,$81,$87,$92,$7C,$37,$D9,$F8,$27,$33,$E0,$9C,$CF,$74,$FD
	.dc.b	$87,$2F,$AC,$F9,$D8,$7A,$63,$E6,$5F,$F8,$75,$F1,$9D,$67,$C7,$87
	.dc.b	$AA,$3F,$C1,$D1,$8A,$FA,$11,$C0,$CD,$7C,$18,$CC,$79,$23,$DC,$65
	.dc.b	$F4,$61,$EF,$AF,$2C,$7C,$03,$CB,$87,$D7,$C3,$C5,$5F,$90,$9F,$F3
	.dc.b	$33,$C6,$FF,$03,$1E,$23,$33,$C0,$FA,$AE,$30,$F2,$D7,$18,$F5,$3D
	.dc.b	$5F,$F0,$E9,$89,$E2,$E0,$F6,$1F,$E3,$63,$FC,$E9,$5F,$E4,$B7,$FC
	.dc.b	$6C,$F9,$27,$A3,$F8,$B8,$FC,$45,$7D,$37,$F2,$A7,$E1,$B0,$FB,$13
	.dc.b	$C6,$FC,$2B,$0F,$EE,$33,$FA,$CC,$15,$FB,$1A,$FE,$EB,$CC,$F1,$BF
	.dc.b	$C5,$58,$A3,$0F,$F8,$B5,$AF,$F7,$C3,$FD,$D2,$E1,$C3,$AC,$BF,$C4
	.dc.b	$7F,$2A,$F8,$0F,$EB,$DC,$9F,$E0,$48,$C5,$FF,$77,$39,$E1,$A6,$99
	.dc.b	$11,$C1,$FF,$21,$B6,$FF,$C6,$38,$78,$6F,$FC,$23,$E0,$FF,$90,$AF
	.dc.b	$FC,$8A,$7A,$4A,$7A,$DD,$0F,$0B,$C5,$0F,$21,$E5,$61,$B1,$FE,$52
	.dc.b	$3F,$CA,$BE,$07,$F8,$F4,$9E,$0F,$F9,$0A,$3F,$C7,$E7,$F8,$FB,$C1
	.dc.b	$FF,$1E,$AE,$2F,$F1,$EB,$9F,$07,$FC,$7C,$FF,$8F,$3C,$AF,$F2,$4E
	.dc.b	$DF,$E3,$BC,$7E,$B4,$60,$FF,$19,$F7,$FF,$85,$7C,$0F,$F1,$9F,$06
	.dc.b	$09,$87,$F9,$AA,$3F,$C8,$2B,$87,$FC,$55,$E3,$7F,$90,$6B,$FC,$42
	.dc.b	$7F,$8C,$4F,$A5,$DF,$FE,$2B,$9F,$F2,$2F,$83,$FE,$30,$BF,$F3,$AC
	.dc.b	$FF,$8A,$5F,$F1,$3D,$7F,$9D,$63,$11,$FE,$39,$8C,$78,$1F,$E3,$23
	.dc.b	$FC,$77,$4F,$27,$13,$F8,$1A,$FE,$D3,$E0,$7F,$6E,$EB,$FE,$D1,$DB
	.dc.b	$FD,$BE,$7F,$B0,$9F,$E5,$09,$FE,$99,$3F,$E2,$DF,$07,$FC,$F3,$E0
	.dc.b	$7F,$4E,$A0,$AD,$BF,$E7,$B9,$DD,$7F,$84,$70,$DE,$1F,$E1,$96,$C6
	.dc.b	$6B,$FC,$4D,$5F,$E1,$67,$FB,$89,$FD,$D7,$5F,$F1,$4D,$FF,$5E,$7F
	.dc.b	$C7,$F1,$FA,$BC,$3C,$2F,$F7,$D3,$CA,$E2,$15,$E4,$9F,$F1,$BE,$3F
	.dc.b	$E0,$49,$CE,$3C,$A7,$E3,$9E,$1A,$F2,$3F,$18,$CE,$1F,$26,$BC,$B8
	.dc.b	$FC,$C2,$4F,$AA,$7F,$C2,$51,$F2,$EF,$B4,$FC,$91,$87,$B7,$0F,$34
	.dc.b	$7F,$85,$AF,$B4,$78,$B0,$EF,$3D,$70,$EE,$F9,$4F,$F1,$35,$FB,$6B
	.dc.b	$9C,$7A,$27,$75,$CA,$38,$63,$F1,$EB,$71,$E5,$BF,$26,$3E,$8A,$FC
	.dc.b	$54,$F8,$4F,$AF,$3D,$E7,$51,$F2,$5F,$BE,$FD,$71,$8F,$46,$E3,$FE
	.dc.b	$79,$AF,$C7,$47,$86,$3F,$CF,$4F,$67,$E5,$3F,$E7,$E3,$F2,$15,$E0
	.dc.b	$AF,$9F,$18,$34,$7E,$14,$F8,$B3,$CA,$BB,$DF,$C1,$BE,$53,$F1,$B0
	.dc.b	$FA,$07,$FB,$6B,$0D,$E1,$E8,$3B,$38,$C3,$E1,$47,$DC,$AF,$E5,$18
	.dc.b	$7F,$99,$27,$27,$E8,$E7,$F4,$38,$7C,$67,$FC,$AB,$1A,$AF,$D3,$5F
	.dc.b	$80,$F9,$07,$77,$D4,$7C,$93,$64,$9F,$F2,$E1,$F7,$8F,$A8,$78,$09
	.dc.b	$1D,$63,$E6,$BE,$EF,$07,$FC,$BB,$5E,$AA,$FF,$30,$36,$3F,$CB,$B1
	.dc.b	$FE,$5F,$8F,$F3,$0F,$37,$F6,$8C,$FF,$CD,$11,$FD,$BD,$FE,$E9,$7F
	.dc.b	$E2,$B3,$FC,$41,$5F,$E1,$8F,$03,$FC,$19,$3A,$C7,$FC,$3B,$3F,$E0
	.dc.b	$D3,$51,$FE,$0E,$9F,$E6,$51,$F3,$CF,$E2,$B3,$FD,$ED,$FE,$0A,$D7
	.dc.b	$FC,$12,$7F,$84,$67,$B3,$FE,$0F,$C3,$D0,$F3,$3F,$CD,$87,$2F,$07
	.dc.b	$FC,$DB,$5E,$B8,$FF,$16,$1F,$E0,$E3,$FC,$93,$8F,$F9,$D2,$3C,$0D
	.dc.b	$9E,$33,$E1,$3F,$14,$7F,$87,$8F,$4B,$FC,$B9,$F1,$4F,$23,$15,$E6
	.dc.b	$AF,$F1,$EE,$32,$7C,$8F,$FA,$1A,$B8,$DF,$81,$FF,$16,$9F,$DA,$71
	.dc.b	$FF,$0E,$3F,$CF,$30,$E9,$7F,$E2,$FC,$3F,$C4,$55,$F3,$9E,$D5,$B8
	.dc.b	$C4,$EC,$F2,$B8,$8F,$76,$1E,$3A,$F8,$51,$FC,$E2,$FB,$BA,$BF,$49
	.dc.b	$CB,$1F,$A1,$10,$FB,$27,$AA,$7F,$C8,$38,$FF,$9E,$EB,$FC,$E1,$3F
	.dc.b	$2C,$FB,$27,$C2,$3C,$18,$F9,$5E,$4F,$F9,$0A,$79,$1F,$E6,$5B,$FC
	.dc.b	$71,$E1,$C3,$D3,$5E,$2C,$3F,$CB,$CF,$29,$FC,$5B,$FE,$67,$3F,$CB
	.dc.b	$91,$A2,$5F,$E7,$A3,$F1,$38,$FC,$EA,$FE,$05,$5F,$12,$7D,$37,$E0
	.dc.b	$34,$FE,$56,$39,$CF,$9A,$3F,$4A,$4A,$FF,$2F,$BF,$34,$87,$9C,$F0
	.dc.b	$E3,$FA,$E7,$CF,$87,$96,$7E,$C4,$FE,$4C,$FC,$94,$FE,$06,$BE,$59
	.dc.b	$E1,$C3,$CD,$3B,$C7,$F2,$18,$F9,$AF,$C9,$3F,$EC,$07,$C2,$7F,$A9
	.dc.b	$5F,$F4,$D6,$1E,$88,$F3,$3F,$90,$6C,$7E,$15,$FB,$84,$BF,$8A,$6E
	.dc.b	$BC,$15,$CD,$B5,$E1,$8F,$F7,$4F,$3F,$F9,$86,$FF,$CD,$27,$F8,$F7
	.dc.b	$4F,$F3,$FF,$81,$FE,$5C,$F0,$35,$3F,$E7,$F8,$FF,$40,$D7,$F9,$EE
	.dc.b	$36,$7F,$A0,$63,$27,$1F,$06,$47,$F9,$D6,$79,$F8,$3F,$E6,$98,$FF
	.dc.b	$32,$6C,$C7,$FC,$E7,$E0,$FF,$9A,$76,$FF,$36,$DF,$F9,$C2,$3F,$D2
	.dc.b	$2D,$BF,$F3,$B1,$FE,$68,$3F,$D2,$1D,$9F,$E6,$CE,$2C,$AE,$FF,$F3
	.dc.b	$77,$0F,$F9,$AA,$BF,$D2,$CD,$7F,$87,$F7,$FF,$9C,$23,$11,$FC,$1E
	.dc.b	$1D,$FF,$E4,$4F,$90,$F5,$4F,$F9,$D6,$72,$A3,$35,$FE,$78,$6C,$E9
	.dc.b	$E3,$3F,$E3,$A3,$FD,$5B,$CF,$FE,$AE,$7F,$C5,$07,$B9,$C1,$1D,$90
	.dc.b	$D1,$4F,$EB,$46,$08,$60,$87,$F8,$5C,$C9,$4C,$9F,$E6,$73,$A1,$92
	.dc.b	$19,$38,$94,$E2,$7B,$4C,$94,$C9,$EE,$2B,$ED,$76,$64,$EE,$71,$29
	.dc.b	$93,$C0,$61,$E4,$F4,$3D,$26,$88,$73,$29,$CC,$C1,$0C,$11,$EA,$F7
	.dc.b	$3A,$9B,$72,$F4,$6B,$93,$67,$32,$BA,$70,$7C,$27,$04,$76,$43,$45
	.dc.b	$3E,$01,$82,$18,$21,$E9,$32,$53,$27,$A4,$E8,$64,$86,$4E,$25,$38
	.dc.b	$9E,$F3,$25,$32,$7B,$8A,$FB,$5D,$99,$3B,$9C,$4A,$64,$F0,$18,$79
	.dc.b	$3D,$0F,$49,$A2,$1C,$CA,$73,$30,$43,$04,$7A,$BD,$CE,$A6,$DC,$BD
	.dc.b	$1A,$E4,$D9,$CC,$AE,$9C,$1F,$09,$C1,$1D,$90,$D1,$4F,$80,$60,$86
	.dc.b	$08,$7A,$4C,$94,$C9,$E9,$3A,$19,$21,$93,$89,$4E,$27,$BC,$C9,$4C
	.dc.b	$9E,$E2,$BE,$D7,$66,$4E,$E7,$12,$99,$3C,$06,$1E,$4F,$F1,$AA,$FF
	.dc.b	$2F,$BF,$D0,$39,$FF,$CD,$6D,$E0,$FE,$97,$3A,$8F,$EE,$06,$4F,$F2
	.dc.b	$B1,$FE,$03,$29,$FE,$4D,$8E,$33,$D2,$38,$90,$E0,$73,$21,$93,$91
	.dc.b	$D2,$7C,$4F,$03,$4E,$1D,$9C,$8D,$10,$C9,$C2,$3C,$8E,$63,$91,$23
	.dc.b	$31,$C0,$D3,$88,$D4,$FB,$89,$3E,$79,$D1,$23,$31,$E5,$3A,$12,$79
	.dc.b	$BC,$63,$6F,$9D,$AF,$89,$AF,$19,$EC,$F1,$21,$DE,$7A,$B6,$36,$D9
	.dc.b	$E7,$3A,$8F,$01,$C8,$A6,$CC,$1E,$32,$9E,$38,$E7,$3D,$A3,$89,$0E
	.dc.b	$07,$32,$19,$39,$1D,$27,$C8,$F0,$34,$E1,$D9,$C8,$D1,$0C,$9C,$23
	.dc.b	$C8,$E6,$39,$12,$33,$1C,$0D,$38,$8D,$4F,$B8,$93,$E7,$9D,$12,$33
	.dc.b	$1E,$53,$A1,$27,$9B,$C6,$36,$F9,$DA,$F8,$9A,$F1,$9E,$CF,$12,$1D
	.dc.b	$E7,$AB,$63,$6D,$9E,$73,$A8,$F0,$1C,$8A,$6C,$C1,$E3,$29,$E3,$8E
	.dc.b	$73,$DA,$38,$90,$E0,$73,$21,$93,$91,$D2,$7C,$8F,$03,$4E,$1D,$9C
	.dc.b	$8D,$10,$C9,$C2,$3C,$8E,$63,$91,$23,$31,$C0,$D3,$88,$D4,$FB,$89
	.dc.b	$3E,$79,$D1,$23,$31,$E5,$3A,$12,$7C,$38,$7F,$98,$F6,$FF,$38,$E5
	.dc.b	$FE,$6D,$6D,$7F,$9A,$DB,$97,$F9,$AA,$7F,$9B,$D7,$F9,$DC,$8E,$09
	.dc.b	$5F,$CA,$E3,$2D,$8C,$BF,$E8,$28,$CB,$63,$2F,$88,$C1,$C4,$86,$08
	.dc.b	$6C,$87,$FA,$25,$E6,$D3,$89,$A3,$FD,$14,$E9,$A6,$8E,$07,$62,$9C
	.dc.b	$6F,$FD,$2E,$F0,$34,$53,$53,$EE,$6C,$F7,$9E,$24,$9D,$4F,$62,$1C
	.dc.b	$CB,$1B,$9E,$53,$C2,$74,$49,$D3,$EB,$31,$1B,$8C,$C7,$A1,$AF,$98
	.dc.b	$E0,$E1,$F0,$C7,$12,$4F,$88,$91,$C2,$32,$D8,$CB,$DE,$32,$D8,$CB
	.dc.b	$E1,$30,$71,$21,$82,$1B,$21,$E6,$79,$B4,$E2,$68,$F8,$6E,$9A,$68
	.dc.b	$E0,$76,$29,$C6,$7E,$23,$C0,$D1,$4D,$4F,$B9,$B3,$DE,$78,$92,$75
	.dc.b	$3D,$88,$73,$2C,$6E,$79,$4F,$09,$D1,$27,$4F,$AC,$C4,$6E,$33,$1E
	.dc.b	$86,$BE,$63,$83,$87,$C3,$1C,$49,$3E,$22,$47,$08,$CB,$63,$2F,$78
	.dc.b	$CB,$63,$2F,$84,$C1,$C4,$86,$08,$6C,$87,$99,$E6,$D3,$89,$A3,$E1
	.dc.b	$BA,$69,$A3,$81,$D8,$A7,$19,$F8,$8F,$03,$45,$35,$3E,$E6,$CF,$79
	.dc.b	$E2,$49,$D4,$F6,$21,$CC,$B1,$B9,$E5,$3C,$27,$44,$9D,$3E,$B3,$11
	.dc.b	$B8,$CC,$7A,$1A,$F9,$8E,$0E,$1F,$0C,$71,$24,$F8,$89,$1C,$23,$2D
	.dc.b	$8C,$BD,$E3,$2D,$8C,$BE,$13,$07,$12,$18,$21,$B2,$1E,$67,$9B,$4E
	.dc.b	$26,$8F,$86,$E9,$A6,$8E,$07,$62,$9C,$67,$E2,$3C,$0D,$14,$D4,$FB
	.dc.b	$9B,$3D,$E7,$89,$27,$53,$D8,$87,$32,$C6,$E7,$94,$F0,$9D,$12,$74
	.dc.b	$FA,$CC,$46,$E3,$31,$E8,$6B,$E6,$38,$38,$7C,$31,$C4,$FE,$B9,$F8
	.dc.b	$28,$80

;--------------------------------------------------------------------------------
;	FDIV.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fdiv_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#63,-(sp)
	peamsg	'FDIV.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FDIV.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fdiv_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdiv.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDIV.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdiv.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDIV.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fdiv_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3D,$61,$F3,$7F,$DA,$BD,$E7,$99,$C6,$30,$DA,$FE
	.dc.b	$50,$D9,$DC,$7E,$72,$3F,$B8,$BF,$98,$29,$F9,$E6,$B4,$CF,$9D,$C9
	.dc.b	$E2,$D7,$FB,$DB,$5E,$86,$A3,$9C,$66,$3C,$47,$48,$C9,$1C,$10,$E4
	.dc.b	$53,$BB,$B7,$24,$32,$6C,$F1,$98,$34,$60,$87,$90,$C9,$4E,$04,$36
	.dc.b	$7C,$53,$65,$34,$71,$37,$1F,$05,$D1,$88,$F5,$3C,$0C,$9C,$48,$60
	.dc.b	$91,$CC,$91,$CA,$3A,$1F,$04,$E3,$1A,$70,$72,$8E,$0E,$9A,$E4,$DC
	.dc.b	$71,$7D,$ED,$7D,$0D,$7C,$26,$A3,$9C,$66,$3B,$1D,$23,$24,$70,$43
	.dc.b	$91,$4E,$EE,$DC,$90,$C9,$B3,$C6,$60,$D1,$82,$1E,$43,$25,$38,$10
	.dc.b	$D9,$F1,$4D,$94,$D1,$C4,$DC,$7C,$17,$46,$23,$D4,$F0,$32,$71,$21
	.dc.b	$82,$47,$32,$47,$28,$E8,$7C,$13,$8C,$69,$C1,$CA,$38,$3A,$6B,$93
	.dc.b	$71,$C5,$F7,$B5,$F4,$35,$F0,$9A,$8E,$71,$98,$EC,$74,$8C,$91,$C1
	.dc.b	$0E,$45,$3B,$BB,$72,$43,$26,$CF,$19,$83,$46,$08,$79,$0C,$94,$E0
	.dc.b	$43,$67,$C5,$36,$53,$47,$13,$71,$F0,$5D,$18,$8F,$53,$F8,$1C,$BF
	.dc.b	$D1,$17,$FC,$F2,$3F,$A1,$35,$FE,$97,$87,$F4,$47,$F9,$F9,$A2,$9A
	.dc.b	$8E,$13,$D2,$32,$DC,$3F,$9E,$36,$FC,$93,$F2,$4F,$51,$0F,$49,$70
	.dc.b	$F6,$12,$3D,$71,$C8,$F8,$B1,$97,$A1,$C8,$8E,$5D,$C7,$27,$B3,$5D
	.dc.b	$B6,$38,$47,$D0,$32,$76,$72,$F8,$CD,$38,$7A,$9B,$8E,$84,$70,$43
	.dc.b	$CA,$53,$9B,$B7,$24,$32,$6C,$F1,$10,$F0,$9B,$30,$7A,$8C,$1A,$39
	.dc.b	$98,$34,$60,$87,$28,$F7,$3A,$31,$1E,$77,$67,$53,$27,$9C,$C1,$0E
	.dc.b	$45,$8E,$A4,$8E,$91,$C8,$F3,$46,$5E,$87,$22,$39,$77,$1C,$9E,$CD
	.dc.b	$76,$D9,$FC,$C4,$7E,$86,$BF,$37,$87,$E9,$5B,$87,$E8,$A3,$F5,$12
	.dc.b	$3F,$C5,$05,$3F,$54,$53,$F5,$47,$F8,$A2,$BB,$B8,$AE,$AF,$EB,$27
	.dc.b	$C1,$7F,$E4,$2C,$9F,$37,$FE,$2E,$C3,$FC,$CA,$D8,$CD,$F5,$6B,$E1
	.dc.b	$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7C,$51,$B7,$2D,$72,$71,$29,$A3
	.dc.b	$07,$F9,$61,$E0,$D3,$27,$F9,$7A,$3C,$0E,$4E,$44,$3A,$90,$D1,$4E
	.dc.b	$51,$D9,$A7,$69,$F8,$6F,$33,$89,$0D,$99,$29,$D5,$E2,$D9,$F3,$BA
	.dc.b	$39,$46,$9B,$1A,$8C,$46,$C8,$E0,$91,$B2,$4F,$AE,$32,$D8,$CC,$FB
	.dc.b	$5A,$FA,$67,$6E,$67,$81,$1C,$10,$E1,$3C,$9F,$74,$6D,$CB,$5C,$9C
	.dc.b	$4A,$68,$C1,$F0,$9E,$0D,$32,$7C,$28,$F0,$39,$39,$10,$EA,$43,$45
	.dc.b	$39,$47,$66,$9D,$A7,$E1,$BC,$CE,$24,$36,$64,$A7,$57,$8B,$67,$CE
	.dc.b	$E8,$E5,$1A,$6C,$6A,$31,$1B,$23,$82,$46,$C9,$3E,$B8,$CB,$63,$33
	.dc.b	$ED,$6B,$E9,$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7D,$D1,$B7,$2D,$72
	.dc.b	$71,$29,$A3,$07,$C2,$78,$34,$C9,$F0,$A3,$C0,$E4,$E4,$43,$A9,$0D
	.dc.b	$14,$E5,$1D,$9A,$76,$9F,$86,$F3,$38,$90,$D9,$92,$9D,$5E,$2D,$9F
	.dc.b	$3B,$A3,$94,$7E,$F9,$FE,$8D,$F2,$DF,$EE,$44,$D2,$3F,$EE,$86,$BF
	.dc.b	$EE,$C6,$BF,$EF,$06,$BF,$D3,$5A,$FF,$51,$6B,$E0,$3A,$10,$E8,$53
	.dc.b	$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10
	.dc.b	$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$A6,$7B,$4A,$79
	.dc.b	$8A,$78,$4A,$76,$29,$D4,$A7,$57,$E9,$B8,$23,$82,$3E,$16,$BE,$16
	.dc.b	$BE,$16,$BE,$16,$BE,$16,$BE,$13,$A1,$0E,$85,$3A,$14,$E8,$53,$A1
	.dc.b	$4E,$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$10,$FA,$67,$B4,$A7,$98,$A7,$84,$A7,$62
	.dc.b	$9D,$4A,$75,$7E,$9B,$82,$38,$23,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$6B
	.dc.b	$E1,$6B,$E1,$3A,$10,$E8,$53,$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8
	.dc.b	$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$FE,$09,$1F,$E7,$22,$61,$FE,$70,$8F,$F4,$16,$1F,$E7,$C6,$1F
	.dc.b	$E7,$C2,$8F,$83,$7C,$0B,$E3,$FF,$D0,$13,$FE,$85,$9C,$C6,$23,$84
	.dc.b	$ED,$DC,$E9,$FF,$4F,$FA,$5F,$E8,$BD,$1F,$09,$FE,$8D,$E0,$65,$98
	.dc.b	$C3,$6F,$FD,$29,$2B,$C4,$3D,$27,$81,$E3,$8C,$36,$33,$1E,$76,$BE
	.dc.b	$C5,$E7,$EC,$91,$B7,$97,$9F,$E0,$9D,$CF,$FA,$9F,$3F,$5B,$FE,$A9
	.dc.b	$8C,$BF,$E9,$EC,$FF,$D5,$50,$AF,$67,$CD,$BF,$3F,$F9,$F7,$CE,$FF
	.dc.b	$D6,$53,$9D,$F7,$5A,$AE,$82,$79,$78,$DA,$37,$52,$7A,$1B,$A4,$F4
	.dc.b	$F4,$24,$EA,$FC,$91,$E0,$F8,$59,$39,$B4,$F0,$8F,$1D,$7A,$EB,$CF
	.dc.b	$F0,$BF,$85,$8F,$73,$6C,$D3,$EC,$7A,$9D,$2F,$F1,$8D,$7F,$22,$D7
	.dc.b	$FD,$A4,$D7,$E3,$CB,$D8,$F8,$E4,$8F,$47,$27,$CA,$E3,$F1,$12,$3F
	.dc.b	$3B,$8F,$D3,$8F,$37,$8D,$F4,$3C,$C9,$5F,$06,$3E,$9F,$8F,$23,$E2
	.dc.b	$3F,$5B,$C7,$FA,$FF,$0B,$EF,$24,$F6,$D1,$1F,$0F,$E1,$7E,$EF,$93
	.dc.b	$2B,$E7,$47,$E0,$3C,$99,$1F,$45,$FC,$1F,$B5,$F8,$AF,$D0,$3F,$61
	.dc.b	$3F,$EF,$5D,$D1,$88,$FF,$7D,$46,$3C,$0F,$A9,$39,$9F,$C7,$AF,$A0
	.dc.b	$E9,$EE,$49,$FC,$24,$E5,$BE,$E7,$22,$3F,$2A,$11,$DB,$D3,$93,$96
	.dc.b	$BE,$79,$79,$9E,$42,$47,$37,$CF,$F0,$DE,$FF,$3E,$4F,$92,$3E,$17
	.dc.b	$97,$EC,$24,$78,$DF,$7F,$93,$F1,$FE,$4A,$4F,$B2,$3E,$57,$CB,$7C
	.dc.b	$BF,$76,$4F,$C0,$8F,$9D,$DF,$F2,$09,$1E,$F7,$E5,$FE,$85,$F4,$FF
	.dc.b	$4C,$93,$F2,$23,$EB,$79,$9F,$5F,$CC,$93,$F3,$23,$ED,$79,$5F,$50
	.dc.b	$91,$F2,$DF,$AF,$EC,$FD,$FF,$72,$4F,$D4,$8F,$E4,$79,$7F,$25,$8F
	.dc.b	$DF,$87,$0F,$F0,$83,$1E,$27,$F2,$AF,$47,$EB,$C7,$05,$CD,$FC,$AD
	.dc.b	$B1,$FC,$FD,$AF,$F4,$46,$A2,$07,$EE,$BC,$79,$7A,$11,$FD,$8F,$E8
	.dc.b	$7C,$1F,$E1,$6B,$E7,$3F,$93,$B6,$3F,$8B,$3B,$F4,$7F,$98,$F7,$21
	.dc.b	$F2,$FC,$2A,$1C,$1E,$7F,$2F,$D3,$E5,$E4,$F4,$8C,$B4,$87,$93,$E5
	.dc.b	$FC,$7E,$E4,$8F,$1B,$E6,$FC,$12,$47,$B1,$FE,$45,$89,$A6,$3E,$43
	.dc.b	$73,$F0,$23,$DF,$1F,$03,$E3,$FE,$0F,$C1,$4A,$F8,$31,$F1,$3E,$63
	.dc.b	$A9,$23,$E5,$BE,$0F,$98,$F9,$3F,$23,$27,$E4,$C7,$CC,$F9,$19,$1F
	.dc.b	$31,$F9,$FF,$17,$F8,$CF,$52,$4F,$E3,$3A,$91,$F8,$9F,$1F,$F1,$5E
	.dc.b	$47,$E1,$C9,$5F,$87,$7F,$15,$E4,$3A,$3F,$0D,$E6,$39,$3F,$0F,$D8
	.dc.b	$BF,$C4,$CE,$70,$FC,$64,$E5,$E3,$5C,$30,$E9,$1F,$C6,$63,$18,$7E
	.dc.b	$4F,$E9,$B9,$4F,$E4,$F1,$FC,$AC,$63,$0E,$92,$8C,$61,$D1,$B7,$CF
	.dc.b	$0F,$3E,$1C,$9B,$79,$3C,$F7,$C6,$3C,$71,$D5,$DB,$4F,$6C,$F3,$78
	.dc.b	$D7,$37,$F5,$55,$EE,$C7,$D1,$5F,$1A,$F9,$C7,$9A,$BE,$11,$C8,$EC
	.dc.b	$7C,$13,$A6,$3F,$03,$1E,$07,$19,$EB,$3F,$38,$F2,$94,$F1,$E3,$69
	.dc.b	$8F,$97,$0F,$19,$2B,$C5,$7D,$A3,$E0,$E1,$DA,$7C,$B7,$CB,$0F,$2E
	.dc.b	$1C,$E7,$C2,$7B,$0F,$51,$67,$83,$63,$E8,$3C,$2B,$B1,$E1,$7C,$77
	.dc.b	$E5,$79,$E1,$F0,$63,$EB,$BE,$52,$38,$24,$79,$23,$2F,$43,$E8,$4F
	.dc.b	$82,$7D,$E7,$E0,$9F,$14,$7A,$9D,$3C,$23,$AB,$E4,$7C,$D7,$EE,$AF
	.dc.b	$34,$F8,$67,$37,$F1,$E3,$EA,$9F,$17,$1F,$C3,$3B,$3C,$C6,$AB,$D2
	.dc.b	$F7,$33,$7F,$33,$0F,$2C,$79,$F0,$F5,$11,$F9,$CF,$B9,$F5,$1C,$AF
	.dc.b	$C2,$F5,$3C,$11,$EE,$AF,$D8,$DF,$89,$F7,$47,$59,$F0,$1D,$8E,$A7
	.dc.b	$23,$D8,$75,$8E,$2F,$84,$FC,$B5,$78,$A3,$31,$F1,$AF,$DB,$1F,$A3
	.dc.b	$AF,$03,$A3,$CA,$FC,$EC,$7F,$41,$8F,$E3,$AF,$CA,$4A,$F9,$15,$CB
	.dc.b	$0F,$DE,$4E,$6B,$85,$6A,$3D,$86,$6F,$F4,$27,$A1,$87,$E7,$5F,$09
	.dc.b	$FB,$AA,$EB,$3E,$99,$F8,$AF,$A6,$3E,$01,$F0,$6B,$F0,$D1,$99,$EB
	.dc.b	$8F,$E3,$1F,$C4,$F9,$DF,$E3,$F6,$C7,$F1,$A8,$C4,$FE,$C1,$D4,$FF
	.dc.b	$80,$7E,$0F,$FC,$05,$18,$84,$62,$BF,$D9,$6D,$8F,$F0,$03,$5F,$F0
	.dc.b	$4B,$FC,$3D,$FF,$01,$7C,$05,$79,$9F,$C0,$CC,$94,$CC,$FF,$40,$72
	.dc.b	$D7,$39,$FF,$83,$FB,$FF,$D7,$33,$9B,$FE,$61,$5D,$35,$C3,$FC,$A8
	.dc.b	$C5,$FA,$9F,$15,$F2,$8C,$5F,$86,$31,$3E,$E3,$93,$FE,$C8,$AE,$A6
	.dc.b	$F0,$F0,$CF,$BA,$30,$F8,$4C,$5F,$A6,$77,$87,$A2,$FA,$4F,$96,$F8
	.dc.b	$9E,$AC,$39,$96,$39,$61,$F2,$6F,$96,$1D,$23,$CE,$DA,$E3,$5E,$E3
	.dc.b	$8B,$CE,$7F,$C8,$38,$7C,$07,$ED,$9F,$1A,$BE,$D5,$F6,$8F,$63,$70
	.dc.b	$E5,$1F,$E4,$BA,$E4,$FE,$34,$E4,$79,$CB,$7C,$1F,$BF,$87,$C1,$3D
	.dc.b	$B5,$F3,$9A,$66,$3C,$F8,$7B,$E7,$E3,$CF,$B6,$72,$77,$AF,$82,$FD
	.dc.b	$B6,$BE,$E6,$BE,$76,$BF,$90,$39,$BE,$7B,$F6,$C6,$4F,$81,$7F,$19
	.dc.b	$FA,$27,$C6,$2D,$FA,$DF,$2B,$5E,$26,$EB,$C3,$87,$EA,$EF,$BB,$6F
	.dc.b	$B4,$EA,$BE,$5E,$1F,$05,$E7,$1D,$5F,$D0,$9F,$9B,$C3,$E3,$BF,$0E
	.dc.b	$7F,$06,$71,$7E,$99,$B7,$E8,$B4,$F4,$BB,$C3,$CB,$1E,$23,$A1,$F2
	.dc.b	$23,$57,$F1,$6B,$C3,$3F,$8E,$6B,$F4,$67,$C6,$F2,$3D,$6F,$E0,$5F
	.dc.b	$8F,$1E,$DA,$FD,$4C,$79,$4C,$CE,$DF,$82,$77,$7D,$6D,$3D,$67,$73
	.dc.b	$E0,$9F,$85,$7C,$6F,$D6,$3F,$D4,$07,$F1,$A2,$C2,$B3,$1B,$3F,$D5
	.dc.b	$BC,$0F,$81,$D4,$7F,$AF,$63,$13,$FE,$C2,$8F,$F5,$1B,$FE,$79,$9D
	.dc.b	$EF,$FE,$A8,$EB,$7C,$CE,$2B,$FD,$75,$E0,$FF,$AA,$E3,$18,$23,$FD
	.dc.b	$5D,$F2,$78,$75,$5F,$EB,$4F,$93,$C6,$0F,$93,$B1,$87,$3F,$33,$E8
	.dc.b	$C7,$FD,$8F,$87,$FA,$E6,$31,$1E,$2B,$46,$23,$C5,$18,$C3,$FD,$93
	.dc.b	$F3,$3E,$3F,$3F,$E4,$47,$FB,$1A,$30,$E7,$E6,$75,$F1,$76,$3F,$DB
	.dc.b	$1B,$FF,$B2,$AD,$3A,$9F,$AB,$9A,$73,$18,$6E,$5F,$ED,$99,$D5,$79
	.dc.b	$DE,$1C,$1E,$53,$A4,$FF,$A8,$EF,$E7,$DF,$D8,$8C,$1F,$3D,$7C,$4F
	.dc.b	$B4,$F7,$7C,$4D,$AF,$49,$CA,$3E,$17,$C4,$E6,$F9,$E3,$E9,$8F,$8D
	.dc.b	$F1,$3F,$1F,$E2,$6E,$1C,$AF,$C1,$19,$8C,$5F,$C1,$F8,$9C,$57,$4C
	.dc.b	$D1,$F8,$68,$0F,$37,$EC,$38,$6B,$9B,$FC,$EB,$4F,$CF,$F9,$DA,$24
	.dc.b	$69,$E7,$D3,$F1,$23,$FD,$A9,$AF,$86,$9E,$77,$8E,$B5,$1E,$7F,$89
	.dc.b	$FC,$37,$A7,$F8,$76,$E1,$EE,$D1,$F1,$1F,$65,$B1,$F7,$E3,$06,$7E
	.dc.b	$27,$F2,$1E,$4F,$E4,$5B,$3C,$FE,$27,$F2,$9F,$11,$F9,$56,$C7,$7F
	.dc.b	$87,$EC,$47,$F1,$8F,$87,$E2,$3F,$39,$F1,$1F,$9D,$6C,$F9,$FE,$23
	.dc.b	$F6,$5F,$A2,$3C,$CF,$D1,$46,$1C,$F9,$7F,$93,$93,$49,$5F,$9E,$87
	.dc.b	$C4,$F4,$F8,$9E,$AD,$9F,$C1,$CB,$DE,$E8,$48,$E4,$F7,$F7,$A4,$75
	.dc.b	$6F,$37,$E9,$E9,$EA,$79,$49,$5E,$37,$D1,$EA,$7B,$BD,$4F,$7B,$67
	.dc.b	$C7,$F0,$FE,$D2,$3E,$B7,$E0,$7C,$47,$C6,$3E,$3F,$C3,$D9,$F5,$FC
	.dc.b	$47,$CA,$EE,$F9,$6D,$8F,$7F,$67,$C9,$23,$F1,$DF,$99,$F1,$1F,$48
	.dc.b	$FA,$7F,$0F,$67,$E3,$FC,$47,$D6,$F8,$8F,$AE,$D8,$F9,$7F,$0F,$F5
	.dc.b	$48,$FD,$37,$EC,$7C,$47,$DE,$3F,$98,$7E,$8B,$FA,$E9,$FA,$71,$F9
	.dc.b	$FF,$47,$F4,$A4,$8F,$E9,$3A,$FF,$16,$E5,$FE,$5F,$7F,$D3,$DF,$E7
	.dc.b	$ED,$7F,$9F,$37,$CC,$34,$FE,$97,$1F,$C3,$1F,$DC,$7A,$5F,$D4,$A7
	.dc.b	$7F,$13,$E6,$F3,$66,$3F,$C2,$AF,$6F,$82,$73,$87,$3F,$37,$DB,$E6
	.dc.b	$7B,$9B,$5C,$62,$CE,$E3,$07,$F1,$95,$F3,$EE,$6F,$E2,$8B,$9A,$C6
	.dc.b	$19,$5F,$DB,$A7,$4D,$8F,$E3,$31,$8A,$FE,$EE,$B6,$C9,$2B,$93,$FC
	.dc.b	$83,$E7,$DE,$DF,$F2,$57,$FB,$24,$F9,$DC,$B7,$DB,$D1,$C4,$8F,$F8
	.dc.b	$0A,$38,$E1,$FE,$08,$F8,$FF,$F0,$17,$BF,$FE,$06,$6C,$67,$D9,$4E
	.dc.b	$5B,$1F,$E1,$0F,$83,$77,$BE,$43,$FC,$1F,$18,$6C,$66,$78,$35,$FE
	.dc.b	$7E,$7F,$40,$C3,$FA,$BF,$95,$E3,$F5,$BF,$8A,$4F,$9D,$AF,$01,$3C
	.dc.b	$1F,$F1,$1F,$C8,$7F,$4C,$7E,$E7,$E3,$8E,$FE,$FD,$9F,$E0,$98,$C4
	.dc.b	$FE,$06,$B8,$3F,$DA,$82,$F5,$FF,$A7,$B6,$3F,$C5,$AD,$7C,$31,$FD
	.dc.b	$53,$F6,$5C,$47,$9D,$AF,$C0,$6B,$FD,$81,$AF,$F8,$D7,$C3,$F3,$F1
	.dc.b	$EE,$AC,$ED,$AE,$CF,$CD,$70,$FB,$1F,$F2,$2B,$5F,$90,$D7,$E8,$B5
	.dc.b	$FF,$1B,$7C,$97,$F8,$21,$FD,$87,$C3,$78,$3C,$1F,$DD,$B6,$75,$3F
	.dc.b	$E0,$66,$EB,$FB,$B6,$E3,$FE,$44,$6C,$F6,$F8,$0E,$5F,$01,$D7,$C2
	.dc.b	$7C,$8D,$33,$5F,$E0,$AB,$56,$23,$C0,$F8,$FE,$07,$97,$C8,$FF,$98
	.dc.b	$35,$FE,$22,$DC,$4D,$35,$83,$E0,$B8,$E3,$FC,$3E,$7D,$D1,$E5,$F9
	.dc.b	$09,$3F,$C4,$DF,$85,$F0,$7F,$0F,$8B,$99,$27,$FC,$83,$87,$DE,$C8
	.dc.b	$BE,$03,$C2,$49,$DB,$E7,$F9,$1F,$88,$48,$F8,$5F,$05,$F1,$FE,$47
	.dc.b	$D8,$48,$FF,$27,$EC,$8F,$6F,$BF,$F5,$C9,$3F,$2F,$E0,$7E,$87,$C8
	.dc.b	$48,$FE,$5C,$FD,$DF,$82,$FA,$9F,$23,$F2,$49,$3D,$9F,$9D,$F2,$5F
	.dc.b	$E0,$08,$FF,$1F,$B5,$FF,$24,$B7,$AB,$F9,$F3,$FE,$0A,$8F,$E7,$9F
	.dc.b	$25,$F8,$12,$3C,$BE,$0F,$F0,$9F,$25,$F6,$49,$1E,$E7,$EE,$7C,$87
	.dc.b	$E2,$49,$1E,$2F,$83,$FC,$C6,$0F,$80,$FC,$77,$C8,$4A,$F3,$C7,$E4
	.dc.b	$7D,$2F,$C3,$12,$3E,$5B,$EE,$F8,$1E,$64,$8F,$7F,$9D,$DB,$D6,$FC
	.dc.b	$D7,$A9,$27,$E1,$C7,$E7,$7E,$43,$F2,$44,$8F,$AE,$FC,$AF,$81,$F3
	.dc.b	$92,$3E,$5F,$C1,$7A,$BD,$0F,$D3,$7C,$7C,$9F,$9F,$1F,$B5,$FD,$1F
	.dc.b	$FC,$CF,$E1,$1A,$4F,$33,$EB,$FC,$17,$C4,$25,$FD,$8F,$83,$FE,$17
	.dc.b	$EE,$7F,$22,$EB,$FF,$57,$36,$FF,$D9,$3E,$46,$BE,$3F,$F8,$8F,$B6
	.dc.b	$ED,$FF,$39,$B7,$2F,$E8,$71,$83,$3F,$13,$C1,$91,$FD,$C6,$7F,$C8
	.dc.b	$8F,$FA,$23,$E4,$7F,$A7,$BF,$E9,$26,$E7,$FE,$D1,$F7,$BF,$8A,$47
	.dc.b	$F4,$38,$FF,$6A,$4E,$08,$73,$F8,$4D,$FC,$27,$C7,$F9,$19,$3D,$E3
	.dc.b	$9F,$C3,$6A,$30,$78,$DF,$9D,$F0,$F2,$3F,$D3,$BC,$9F,$EC,$07,$D1
	.dc.b	$ED,$79,$08,$E0,$F6,$7B,$3E,$82,$38,$3E,$07,$AF,$EC,$F4,$A4,$7C
	.dc.b	$37,$DF,$EB,$48,$F8,$CD,$7E,$37,$A3,$F1,$3C,$D9,$1F,$2D,$F8,$FE
	.dc.b	$7C,$8F,$9C,$D7,$E7,$79,$DF,$8B,$F6,$E4,$7B,$9F,$C2,$FA,$9F,$80
	.dc.b	$F8,$AF,$BE,$49,$FF,$00,$BF,$82,$F3,$35,$1F,$76,$3F,$0F,$1F,$88
	.dc.b	$F8,$A7,$3F,$FB,$03,$E4,$9C,$5F,$E0,$4E,$BF,$DC,$63,$98,$C6,$1F
	.dc.b	$E0,$8A,$DC,$FF,$B2,$29,$87,$48,$4E,$94,$EB,$0E,$91,$C6,$FA,$61
	.dc.b	$DA,$32,$70,$C3,$A6,$1C,$F0,$E5,$3D,$AB,$83,$4E,$87,$28,$DC,$F1
	.dc.b	$AE,$75,$C2,$3A,$5F,$8A,$FD,$95,$A6,$E1,$CE,$75,$1E,$C6,$E1,$DB
	.dc.b	$0F,$95,$7F,$0F,$E3,$BA,$DF,$FB,$B9,$AF,$FA,$F2,$77,$86,$5B,$1E
	.dc.b	$B3,$B1,$E1,$3D,$95,$E5,$7E,$34,$E5,$E1,$7F,$69,$B3,$B8,$F3,$47
	.dc.b	$C6,$37,$87,$AA,$74,$6D,$F6,$99,$AF,$09,$DF,$1F,$59,$F3,$5C,$D7
	.dc.b	$B4,$ED,$8F,$C0,$34,$E4,$F4,$9E,$23,$A6,$3F,$0A,$B6,$76,$7E,$C1
	.dc.b	$92,$47,$AE,$34,$FD,$83,$E0,$D7,$CD,$7C,$73,$F8,$8A,$F1,$9D,$9E
	.dc.b	$B1,$CD,$EA,$F9,$A3,$D5,$1C,$4F,$AE,$F1,$8E,$04,$3D,$B5,$97,$ED
	.dc.b	$9E,$F3,$CA,$47,$D2,$E9,$E1,$8F,$E8,$27,$C3,$3E,$A3,$DC,$EB,$0F
	.dc.b	$AF,$1F,$4A,$B0,$FC,$C3,$E2,$3F,$02,$FD,$51,$CC,$F7,$3F,$02,$FF
	.dc.b	$15,$3E,$47,$F1,$33,$EC,$9F,$8E,$F9,$DE,$18,$FE,$85,$EB,$1F,$03
	.dc.b	$0F,$03,$E8,$8C,$D7,$E5,$E7,$33,$D4,$F9,$51,$E2,$9E,$B8,$FA,$E7
	.dc.b	$C0,$7B,$23,$E9,$9D,$31,$FF,$4F,$DF,$4B,$F4,$9C,$EB,$AE,$1F,$A0
	.dc.b	$C3,$F0,$D1,$E7,$C3,$CB,$3E,$B8,$F4,$1F,$1E,$BD,$13,$E0,$8F,$C8
	.dc.b	$7C,$37,$F8,$BD,$FC,$EB,$67,$F8,$CB,$6B,$F9,$43,$B9,$0F,$93,$B1
	.dc.b	$98,$C0,$71,$1A,$C7,$FD,$6D,$3A,$9C,$BF,$AF,$3F,$D9,$BE,$1F,$B9
	.dc.b	$EF,$57,$BD,$39,$9E,$F7,$FC,$ED,$E2,$73,$33,$97,$FB,$16,$B8,$C6
	.dc.b	$6F,$F9,$DB,$C3,$5E,$4F,$A9,$E9,$3C,$1C,$CF,$8F,$0F,$13,$6B,$A1
	.dc.b	$FC,$94,$E9,$7F,$02,$3B,$DF,$38,$C4,$EE,$30,$FF,$B6,$E7,$75,$E6
	.dc.b	$8F,$59,$FE,$E6,$9F,$5E,$1E,$48,$F9,$51,$8B,$EB,$3E,$AA,$F3,$D6
	.dc.b	$9B,$86,$63,$D0,$7C,$1A,$F9,$13,$E1,$C3,$C7,$3A,$6D,$78,$6F,$91
	.dc.b	$6F,$8D,$7C,$49,$EC,$79,$CB,$1F,$09,$B1,$F0,$4F,$89,$7E,$17,$D0
	.dc.b	$7C,$EA,$EF,$5C,$4E,$F5,$D5,$A6,$67,$E4,$57,$39,$EF,$1F,$9C,$3E
	.dc.b	$01,$A3,$DC,$78,$8B,$5E,$07,$CC,$D7,$95,$FC,$4B,$F0,$47,$EA,$30
	.dc.b	$FB,$AE,$8F,$5D,$7C,$77,$C0,$FA,$5F,$15,$D7,$96,$1E,$9B,$F6,$C7
	.dc.b	$C7,$9F,$5B,$FA,$83,$D0,$F1,$7F,$26,$7A,$23,$EA,$D7,$86,$3A,$9F
	.dc.b	$5E,$3A,$BE,$79,$EA,$FC,$63,$6F,$5B,$F6,$C7,$86,$37,$87,$C3,$8F
	.dc.b	$35,$74,$74,$D7,$A1,$E6,$3C,$0E,$63,$F5,$D7,$E7,$9F,$DF,$BE,$1C
	.dc.b	$3E,$24,$7C,$8A,$FA,$2F,$33,$B9,$93,$C6,$53,$B9,$4D,$9E,$E7,$9E
	.dc.b	$1E,$E8,$F3,$B6,$FD,$D1,$F2,$B4,$FE,$1E,$FF,$28,$EE,$FE,$23,$18
	.dc.b	$8F,$E2,$ED,$8C,$E1,$F9,$39,$46,$DF,$CB,$35,$E9,$1F,$CB,$1A,$71
	.dc.b	$F1,$FF,$D8,$1F,$0F,$D2,$11,$89,$FE,$77,$7F,$EB,$CE,$3D,$4E,$6B
	.dc.b	$FD,$5D,$E9,$63,$C2,$37,$D6,$C3,$9F,$46,$65,$FE,$B7,$8B,$7F,$EB
	.dc.b	$7F,$37,$91,$CF,$D0,$B5,$E2,$3F,$A6,$E4,$F3,$FF,$9F,$C6,$6F,$D3
	.dc.b	$18,$7A,$46,$67,$83,$EA,$97,$89,$4C,$78,$F6,$78,$4B,$6E,$06,$70
	.dc.b	$FF,$49,$CF,$92,$74,$FF,$AF,$BE,$0B,$E3,$F8,$C7,$8C,$64,$EB,$C5
	.dc.b	$E3,$F8,$2F,$23,$63,$C3,$1C,$8C,$61,$E0,$3E,$BD,$A3,$D9,$F0,$32
	.dc.b	$BD,$3F,$05,$6F,$86,$DF,$27,$CF,$F8,$46,$79,$BF,$B0,$C6,$6F,$E6
	.dc.b	$57,$F7,$CF,$33,$F0,$CF,$F7,$EF,$8C,$AA,$2C,$49,$FC,$84,$A3,$F2
	.dc.b	$8D,$8C,$C7,$E5,$FD,$1F,$8D,$1E,$06,$CF,$BC,$2D,$BC,$6E,$B1,$F4
	.dc.b	$F0,$3D,$1F,$BC,$7D,$FF,$83,$B8,$FF,$BC,$B6,$FD,$46,$AE,$14,$61
	.dc.b	$B3,$FE,$14,$9D,$4F,$0C,$1E,$3F,$DE,$E1,$FC,$63,$6F,$F1,$3F,$25
	.dc.b	$F8,$F3,$F2,$1F,$25,$63,$F2,$99,$38,$FC,$04,$8F,$04,$65,$F1,$62
	.dc.b	$F8,$0F,$C6,$EB,$F9,$B6,$CF,$E7,$FE,$43,$FC,$5C,$DF,$1F,$F4,$1F
	.dc.b	$25,$FB,$A9,$CF,$2F,$E1,$32,$F9,$17,$E0,$E2,$F8,$44,$7E,$15,$3D
	.dc.b	$6F,$CE,$FC,$9F,$36,$CF,$B3,$E1,$FF,$42,$77,$F6,$6C,$7C,$0E,$3F
	.dc.b	$9C,$47,$DE,$FD,$1E,$2F,$D4,$9E,$7F,$42,$CF,$B7,$CE,$F5,$1E,$BF
	.dc.b	$42,$C7,$CC,$E6,$FB,$44,$7E,$5B,$F7,$7E,$17,$F6,$A7,$C3,$F8,$1B
	.dc.b	$3F,$27,$B7,$E3,$1F,$1F,$E0,$2C,$7D,$8F,$86,$FC,$31,$1F,$AE,$FE
	.dc.b	$27,$A3,$F8,$09,$F3,$FE,$06,$CF,$D5,$F9,$2F,$C8,$F2,$9A,$48,$FC
	.dc.b	$87,$C4,$7F,$9B,$1B,$D3,$F9,$88,$C3,$FC,$ED,$AF,$F8,$5F,$C1,$FC
	.dc.b	$E7,$5F,$F1,$23,$F0,$1F,$05,$6F,$F9,$25,$FF,$4E,$7F,$96,$F3,$7E
	.dc.b	$13,$C3,$FE,$07,$39,$C7,$FA,$2C,$E6,$B7,$DD,$FA,$A8,$FC,$5C,$F8
	.dc.b	$5A,$EA,$7D,$7F,$03,$FD,$2A,$CD,$F4,$F0,$C9,$FE,$71,$6F,$5F,$18
	.dc.b	$7F,$1A,$8C,$4F,$AB,$C8,$E8,$75,$F2,$6C,$64,$E7,$F2,$78,$BE,$4F
	.dc.b	$F1,$CF,$92,$73,$FF,$69,$6D,$7F,$69,$6D,$7F,$A5,$9A,$FF,$73,$7F
	.dc.b	$8A,$C7,$F2,$8F,$D6,$9F,$11,$D7,$AB,$F9,$4C,$62,$75,$3F,$DD,$E3
	.dc.b	$17,$FE,$A2,$6B,$FD,$FD,$AF,$F9,$4F,$E4,$DE,$BF,$F7,$F6,$DF,$F7
	.dc.b	$F6,$CF,$FA,$B1,$AF,$F8,$25,$7A,$BF,$CF,$3F,$1D,$FE,$55,$5D,$63
	.dc.b	$1E,$9F,$F4,$0F,$93,$FE,$C5,$EE,$FF,$82,$DB,$8F,$FA,$E1,$AF,$F8
	.dc.b	$55,$AF,$F9,$A7,$E0,$1F,$16,$F9,$4F,$F0,$A4,$61,$B1,$99,$FF,$61
	.dc.b	$CE,$5E,$DE,$BF,$FB,$25,$BD,$9E,$38,$C4,$78,$5A,$FF,$9E,$7E,$83
	.dc.b	$E3,$BF,$D7,$1F,$F0,$BB,$7D,$EF,$F3,$F7,$7F,$F8,$D2,$77,$3F,$D4
	.dc.b	$5B,$19,$E2,$DF,$07,$56,$D7,$FB,$8B,$F5,$F6,$F8,$87,$CA,$BD,$7F
	.dc.b	$EC,$AD,$8F,$F1,$B3,$5F,$F7,$1B,$5F,$F2,$03,$FD,$75,$FE,$EA,$EF
	.dc.b	$1F,$F2,$5B,$71,$FF,$4C,$3B,$9F,$ED,$4D,$9D,$FA,$DE,$58,$C5,$7F
	.dc.b	$A7,$FC,$D7,$8B,$EC,$F1,$FF,$AF,$FA,$B7,$C3,$7A,$94,$C4,$4D,$3F
	.dc.b	$72,$D7,$F7,$1F,$23,$C7,$D4,$E5,$E9,$C9,$FD,$FC,$65,$A4,$3F,$80
	.dc.b	$7A,$9F,$2F,$C3,$F8,$91,$FE,$B5,$6A,$3F,$C0,$FC,$5F,$E1,$28,$CE
	.dc.b	$9F,$DE,$3A,$F9,$4C,$AF,$E1,$F1,$EF,$F3,$1F,$1B,$49,$3F,$E4,$BF
	.dc.b	$93,$7B,$BE,$3F,$7F,$B1,$27,$FC,$E1,$9F,$F8,$D5,$FF,$05,$71,$3D
	.dc.b	$7F,$94,$48,$FE,$3F,$F2,$FF,$37,$CC,$91,$FC,$89,$FA,$17,$F4,$7D
	.dc.b	$EF,$F3,$E7,$07,$F8,$49,$B8,$FF,$99,$6F,$BC,$63,$D7,$FB,$04,$8F
	.dc.b	$E5,$BF,$2B,$F6,$FE,$4E,$47,$F3,$07,$EF,$7B,$1F,$7F,$E4,$FA,$92
	.dc.b	$7D,$6F,$83,$B3,$F0,$84,$8F,$F3,$87,$E9,$5F,$E9,$7F,$0B,$FC,$9A
	.dc.b	$FE,$92,$33,$EC,$FE,$36,$3F,$D3,$71,$C3,$C4,$C7,$B3,$EE,$9F,$E8
	.dc.b	$F3,$FE,$33,$6B,$FE,$9B,$6B,$FD,$85,$FF,$2A,$3E,$2F,$63,$B1,$DF
	.dc.b	$D8,$F0,$7C,$67,$87,$E5,$25,$7A,$23,$C9,$D9,$E1,$24,$78,$5F,$27
	.dc.b	$C9,$FA,$49,$1F,$03,$E5,$3D,$9F,$29,$ED,$F9,$39,$3F,$12,$3E,$07
	.dc.b	$CA,$7B,$49,$1E,$D7,$E0,$7C,$A7,$C5,$24,$7C,$CF,$96,$F9,$1F,$25
	.dc.b	$F2,$7E,$02,$4F,$D0,$8F,$99,$EB,$7C,$92,$47,$C9,$7E,$67,$AB,$F4
	.dc.b	$49,$1F,$63,$DD,$FA,$9F,$25,$23,$ED,$BF,$C0,$FF,$4B,$FF,$6C,$F5
	.dc.b	$7F,$A2,$17,$C9,$7E,$51,$B5,$FE,$69,$6B,$FE,$D1,$6B,$FE,$17,$7F
	.dc.b	$D6,$6F,$FA,$DD,$DC,$F5,$9F,$07,$C9,$F8,$7E,$4F,$C5,$EE,$CC,$38
	.dc.b	$46,$5A,$43,$FD,$43,$87,$FA,$CD,$ED,$E6,$7A,$C8,$F8,$7E,$4F,$DC
	.dc.b	$61,$E3,$18,$7C,$DE,$77,$C3,$F9,$39,$18,$3D,$5E,$9F,$C7,$99,$18
	.dc.b	$3D,$DE,$EF,$CB,$99,$18,$3E,$17,$C9,$7C,$FF,$92,$91,$83,$E3,$79
	.dc.b	$DF,$4E,$64,$60,$F9,$5F,$25,$F5,$FE,$4A,$46,$0F,$9D,$EF,$FD,$B2
	.dc.b	$3F,$47,$E4,$BF,$AC,$7C,$37,$F9,$FE,$3F,$D8,$6D,$7F,$D6,$8D,$E6
	.dc.b	$7C,$17,$F7,$FF,$82,$E3,$F0,$3F,$EC,$2A,$34,$97,$FC,$DC,$83,$F6
	.dc.b	$9F,$E1,$17,$FE,$09,$9E,$73,$FE,$D4,$9C,$AF,$61,$F2,$BF,$C1,$32
	.dc.b	$FF,$6D,$B7,$0F,$E8,$93,$AF,$43,$FA,$E4,$FF,$B5,$6F,$FA,$45,$F0
	.dc.b	$F3,$F3,$52,$73,$87,$2B,$E1,$1F,$EE,$C5,$E4,$74,$F2,$E4,$F0,$3C
	.dc.b	$D3,$A7,$F3,$DF,$29,$E2,$A9,$18,$3E,$7F,$C1,$FD,$22,$3F,$4B,$E1
	.dc.b	$3E,$AF,$C2,$35,$FE,$C1,$F2,$4F,$4B,$FC,$61,$5F,$36,$31,$7F,$D9
	.dc.b	$78,$D3,$9A,$FF,$1B,$61,$CE,$BF,$CE,$11,$8C,$3F,$B6,$7C,$BF,$29
	.dc.b	$61,$C9,$CD,$78,$1A,$77,$BE,$95,$C6,$91,$8A,$E1,$4A,$D6,$1F,$85
	.dc.b	$78,$56,$E7,$A6,$1D,$9F,$1D,$6E,$F9,$DF,$0A,$D6,$1E,$06,$D6,$EF
	.dc.b	$E2,$35,$F6,$E1,$DD,$B3,$C1,$87,$9F,$0F,$8C,$D3,$CA,$7C,$A3,$A9
	.dc.b	$F5,$9F,$7E,$3E,$58,$D1,$D8,$E4,$7B,$0E,$F8,$EC,$93,$EE,$24,$78
	.dc.b	$23,$2F,$80,$E7,$8F,$D7,$6E,$1E,$D8,$F8,$91,$E2,$75,$7F,$4A,$3D
	.dc.b	$07,$B1,$EA,$F9,$4E,$B8,$FC,$99,$E7,$7E,$DA,$E5,$5D,$6B,$F2,$0F
	.dc.b	$3C,$3E,$4D,$68,$8E,$09,$1E,$B8,$CB,$EF,$8C,$BE,$38,$CB,$F2,$DF
	.dc.b	$14,$7D,$02,$47,$84,$E4,$75,$7E,$5B,$C5,$EB,$7F,$26,$7A,$BE,$E8
	.dc.b	$FC,$0D,$FB,$88,$E0,$E3,$7F,$14,$F8,$D8,$7D,$13,$E9,$39,$C3,$F2
	.dc.b	$F1,$9C,$7E,$8C,$7C,$FB,$EA,$FA,$4F,$31,$E9,$21,$82,$61,$E9,$C3
	.dc.b	$DA,$FA,$63,$E0,$9F,$26,$33,$7E,$18,$FB,$07,$8A,$7F,$0B,$3E,$D8
	.dc.b	$E3,$1D,$5F,$0B,$D5,$FC,$E5,$7A,$1F,$54,$F1,$72,$FD,$47,$27,$B6
	.dc.b	$7F,$2B,$F1,$6A,$FD,$7B,$67,$F5,$ED,$7F,$54,$D7,$FD,$09,$F2,$FF
	.dc.b	$E8,$58,$C4,$23,$13,$FD,$4A,$30,$27,$73,$A8,$FE,$AE,$FD,$47,$FD
	.dc.b	$7F,$C3,$5E,$35,$FC,$9D,$C9,$FB,$E3,$C6,$5F,$67,$FB,$4B,$93,$8E
	.dc.b	$3F,$D9,$23,$33,$CE,$7C,$BA,$F5,$9E,$2E,$70,$FE,$49,$3E,$08,$C3
	.dc.b	$DF,$0F,$34,$FA,$6B,$AE,$1F,$06,$38,$C7,$48,$C5,$7F,$14,$3A,$38
	.dc.b	$BF,$15,$F8,$27,$CB,$3E,$98,$C5,$F4,$6C,$F3,$BE,$CF,$CC,$3C,$C7
	.dc.b	$C3,$9F,$34,$FA,$23,$83,$CA,$FC,$31,$E5,$3C,$27,$F8,$36,$BA,$BF
	.dc.b	$E0,$FA,$F1,$3B,$7A,$BD,$CF,$0E,$1F,$22,$BA,$97,$0F,$64,$F1,$3C
	.dc.b	$65,$8D,$C7,$C8,$9F,$29,$F4,$8A,$FC,$93,$9C,$75,$9F,$46,$1E,$D9
	.dc.b	$C9,$B8,$C9,$DD,$CB,$F2,$67,$C2,$FC,$BA,$EE,$D9,$F4,$C7,$78,$FF
	.dc.b	$74,$61,$F1,$9F,$CA,$D7,$27,$E4,$1F,$12,$BD,$87,$23,$C8,$F2,$BF
	.dc.b	$43,$E6,$9E,$CF,$E2,$0F,$34,$FA,$1E,$B1,$DD,$F0,$C6,$5F,$D3,$DF
	.dc.b	$A9,$F1,$B5,$D9,$DE,$3C,$6F,$73,$AB,$CD,$FC,$29,$E7,$3E,$51,$C4
	.dc.b	$F2,$94,$E2,$53,$B9,$C6,$7D,$B7,$F2,$63,$A9,$E1,$3D,$33,$F1,$DF
	.dc.b	$A4,$7D,$AA,$FE,$0D,$ED,$7F,$02,$D7,$F8,$1C,$84,$67,$1F,$C8,$CE
	.dc.b	$A7,$31,$FE,$A9,$F9,$5D,$47,$F9,$43,$CE,$BD,$0F,$25,$58,$E6,$55
	.dc.b	$31,$E7,$5C,$7F,$CB,$58,$7F,$9B,$A3,$9F,$C1,$3E,$42,$F9,$D8,$6D
	.dc.b	$F0,$94,$E3,$4F,$F2,$C4,$FF,$9D,$27,$93,$B9,$F4,$CB,$E1,$1E,$76
	.dc.b	$21,$DD,$A9,$CD,$FF,$9B,$E7,$E4,$4E,$CF,$8C,$FF,$A3,$DA,$EA,$3F
	.dc.b	$A0,$34,$BD,$FD,$CF,$07,$BB,$67,$E1,$FE,$9B,$FD,$15,$BD,$5F,$3A
	.dc.b	$3F,$AB,$46,$0D,$FC,$B6,$7E,$37,$4D,$8E,$7E,$37,$1F,$1B,$93,$63
	.dc.b	$BC,$75,$F0,$B1,$A7,$CA,$3F,$D4,$58,$A3,$C5,$E1,$78,$FC,$2B,$87
	.dc.b	$A1,$C3,$EA,$F8,$DB,$87,$3E,$76,$29,$F6,$1E,$DF,$8B,$B7,$F0,$9F
	.dc.b	$81,$F1,$7F,$07,$E2,$EC,$FC,$67,$0D,$3E,$4C,$63,$1F,$06,$1F,$7E
	.dc.b	$7E,$6F,$2F,$CB,$E8,$F9,$8C,$8F,$A3,$F1,$76,$F8,$BF,$47,$E2,$FE
	.dc.b	$97,$C5,$DA,$FA,$EE,$1F,$EF,$DB,$7B,$A3,$F0,$AD,$8C,$E4,$9C,$AF
	.dc.b	$8B,$FB,$C7,$DF,$F8,$BB,$5F,$84,$F8,$BF,$C1,$1F,$B0,$F8,$4F,$DD
	.dc.b	$46,$4F,$D5,$BF,$E0,$4D,$3F,$DC,$59,$23,$11,$FE,$D8,$6B,$FE,$DF
	.dc.b	$6C,$6F,$C7,$ED,$DD,$DD,$B7,$DB,$B7,$C4,$78,$FB,$AC,$78,$BC,$9F
	.dc.b	$83,$8F,$FB,$26,$3E,$2B,$72,$FF,$7A,$E5,$E5,$AD,$FC,$57,$B7,$E2
	.dc.b	$BD,$CD,$C3,$DB,$E9,$7C,$13,$E1,$7C,$55,$8F,$83,$E9,$78,$08,$F8
	.dc.b	$1F,$1F,$C5,$7C,$9F,$8A,$F9,$4D,$9F,$93,$F1,$5F,$37,$6F,$9C,$D8
	.dc.b	$F9,$BF,$13,$EC,$23,$EC,$7D,$FF,$15,$F5,$7E,$2B,$EB,$36,$7E,$AF
	.dc.b	$C5,$7D,$93,$ED,$7A,$16,$3E,$CF,$A1,$F2,$08,$FC,$87,$F1,$7F,$A7
	.dc.b	$23,$F9,$8B,$5F,$E5,$6B,$E2,$F3,$1A,$F8,$B8,$EB,$E2,$EF,$47,$ED
	.dc.b	$6F,$F9,$5C,$FE,$69,$B1,$F5,$9A,$FE,$31,$AF,$F4,$BE,$84,$7F,$30
	.dc.b	$8E,$9E,$A7,$9B,$E2,$FC,$EF,$A3,$D4,$98,$7A,$3E,$2A,$DF,$F1,$13
	.dc.b	$F9,$B6,$7F,$CB,$3C,$83,$49,$5F,$B0,$F8,$7F,$E5,$15,$FC,$3D,$BE
	.dc.b	$DF,$F9,$8B,$57,$C7,$7F,$9A,$23,$13,$AB,$F5,$6C,$F8,$CF,$E8,$5A
	.dc.b	$7C,$E6,$CF,$F4,$3D,$13,$E5,$24,$7F,$40,$8C,$BF,$2F,$E2,$FE,$B9
	.dc.b	$F6,$3E,$32,$C7,$D8,$8F,$B7,$F1,$78,$DB,$EA,$3F,$7B,$E3,$25,$7D
	.dc.b	$ED,$2D,$7D,$99,$FC,$1F,$C6,$63,$6F,$81,$B2,$3F,$B0,$FC,$AB,$B2
	.dc.b	$9F,$D9,$3F,$5C,$6D,$8D,$3F,$B0,$56,$DB,$D9,$9D,$3E,$B4,$A3,$A9
	.dc.b	$FD,$89,$C4,$FF,$62,$6C,$FF,$A8,$DA,$FF,$84,$CF,$F0,$A7,$D5,$B8
	.dc.b	$3F,$BB,$B6,$BF,$B9,$62,$F8,$BC,$D4,$9F,$EE,$D5,$EA,$6C,$72,$CB
	.dc.b	$FC,$03,$F2,$CE,$C7,$C4,$FF,$76,$8C,$36,$33,$9F,$FA,$E6,$72,$FF
	.dc.b	$AE,$FE,$33,$FA,$D7,$CB,$7F,$7E,$6C,$67,$DE,$4E,$5B,$1C,$FE,$1F
	.dc.b	$FD,$84,$DF,$87,$F1,$46,$23,$E2,$0C,$3F,$C3,$F3,$FE,$0C,$F9,$6F
	.dc.b	$B5,$37,$93,$FC,$1E,$A3,$2F,$FB,$45,$AF,$47,$EC,$3F,$E1,$D7,$7C
	.dc.b	$FF,$E4,$66,$F0,$7D,$57,$73,$FB,$5F,$95,$FD,$33,$FE,$1E,$6B,$FE
	.dc.b	$DF,$6B,$EA,$6B,$F7,$DF,$BC,$F1,$E1,$F4,$C6,$23,$94,$63,$87,$F0
	.dc.b	$D8,$7F,$96,$AB,$4D,$9C,$E1,$D6,$AD,$BC,$07,$5B,$E1,$F3,$18,$6C
	.dc.b	$66,$7F,$CA,$AB,$E2,$72,$6A,$64,$EA,$38,$FC,$34,$8E,$2F,$4F,$87
	.dc.b	$91,$82,$1F,$C5,$DF,$1F,$73,$E4,$FF,$C9,$39,$7F,$2C,$6C,$66,$BF
	.dc.b	$9B,$47,$E0,$3C,$2F,$91,$1F,$88,$6B,$FA,$D6,$BF,$E8,$16,$D7,$F3
	.dc.b	$8F,$93,$F0,$92,$7F,$CE,$BE,$33,$D8,$F8,$FF,$2B,$FC,$82,$7F,$C3
	.dc.b	$ED,$7F,$BD,$B5,$FF,$08,$B5,$FE,$8D,$BB,$E2,$7E,$AC,$7F,$9C,$FC
	.dc.b	$DB,$99,$FA,$73,$F1,$3E,$52,$57,$F5,$E7,$EB,$FC,$57,$C8,$F1,$3E
	.dc.b	$29,$27,$FA,$D3,$F2,$B1,$7C,$47,$CE,$F8,$F9,$5F,$7A,$3F,$07,$FA
	.dc.b	$87,$EC,$7A,$FF,$A4,$37,$AB,$FA,$F3,$6B,$F1,$98,$7F,$8E,$DF,$F4
	.dc.b	$DE,$FF,$55,$F8,$7C,$9F,$6A,$31,$1F,$12,$3C,$9F,$0B,$E0,$F9,$29
	.dc.b	$3F,$90,$8F,$CA,$FC,$B7,$FA,$DA,$3F,$C9,$AD,$7F,$C5,$2D,$7F,$C7
	.dc.b	$ED,$7C,$FF,$23,$F8,$AC,$8D,$24,$74,$F8,$5F,$C7,$D1,$A4,$8F,$0F
	.dc.b	$C2,$7C,$6E,$E9,$1F,$EB,$AA,$4F,$EB,$3E,$77,$F5,$53,$FE,$3E,$6B
	.dc.b	$FE,$54,$6B,$FE,$09,$6B,$FB,$79,$7C,$2F,$D4,$F0,$64,$F7,$8F,$8B
	.dc.b	$F0,$5F,$63,$E4,$A4,$F8,$E3,$ED,$F8,$5F,$04,$91,$FE,$76,$DB,$FD
	.dc.b	$38,$FC,$CF,$82,$FC,$0F,$89,$23,$D9,$1F,$4B,$E0,$7F,$0B,$F2,$52
	.dc.b	$7E,$04,$7B,$FC,$1F,$DB,$EF,$FE,$CD,$9F,$F4,$1B,$5F,$85,$1F,$3B
	.dc.b	$D7,$FC,$7F,$A1,$27,$E5,$47,$E4,$BC,$29,$1F,$EE,$6D,$3F,$50,$FC
	.dc.b	$FF,$81,$FE,$33,$F2,$9F,$ED,$0F,$27,$FD,$E5,$DF,$FC,$22,$30,$4F
	.dc.b	$82,$C1,$2F,$FD,$2F,$1F,$E4,$76,$FA,$7F,$A2,$F9,$39,$1F,$EF,$69
	.dc.b	$FF,$77,$B7,$D7,$F0,$D4,$8C,$1B,$F2,$BF,$57,$E4,$C8,$C1,$0F,$F6
	.dc.b	$25,$FE,$FD,$C7,$8D,$FE,$37,$8F,$F7,$AC,$78,$7E,$03,$DE,$47,$C9
	.dc.b	$F0,$1F,$08,$F8,$04,$7D,$AF,$C3,$F4,$B5,$1C,$0F,$63,$F2,$BD,$FF
	.dc.b	$E0,$3C,$D2,$7E,$04,$7C,$EF,$80,$F9,$84,$8F,$92,$FC,$FF,$27,$F8
	.dc.b	$4F,$C9,$C9,$F9,$11,$F5,$BE,$06,$47,$CB,$7E,$CF,$A9,$FC,$83,$D6
	.dc.b	$93,$FE,$4D,$7F,$92,$FB,$BF,$D4,$60,$7F,$99,$5F,$D3,$A6,$93,$C7
	.dc.b	$FF,$01,$AF,$1A,$92,$7F,$C3,$0F,$F4,$EF,$94,$7C,$7F,$F6,$45,$F1
	.dc.b	$D4,$C4,$4A,$F1,$F2,$FF,$9B,$6B,$8C,$63,$B3,$FD,$09,$18,$84,$EF
	.dc.b	$CE,$ED,$E5,$F7,$F8,$D9,$5C,$4F,$EF,$17,$FC,$09,$C7,$80,$87,$C6
	.dc.b	$FA,$49,$3F,$EF,$28,$CB,$FE,$73,$F5,$7F,$A6,$91,$EC,$7C,$9F,$37
	.dc.b	$E7,$F9,$BF,$20,$91,$FE,$3D,$7E,$5F,$99,$EA,$87,$A8,$F8,$EF,$F2
	.dc.b	$D7,$22,$FF,$CB,$93,$9A,$FB,$33,$97,$ED,$C6,$5F,$B5,$7D,$63,$12
	.dc.b	$8F,$07,$CA,$E2,$FF,$CD,$0D,$BF,$C2,$BC,$27,$78,$79,$1E,$8F,$9A
	.dc.b	$BC,$AD,$AF,$1B,$BE,$1B,$1E,$8A,$E6,$FA,$AB,$C3,$5C,$61,$7E,$1B
	.dc.b	$F1,$CF,$39,$F8,$EE,$F1,$F4,$46,$0E,$D3,$E0,$7E,$1B,$71,$F9,$11
	.dc.b	$B3,$F4,$18,$7A,$9E,$26,$63,$E4,$4F,$19,$D4,$F0,$3A,$63,$F1,$1E
	.dc.b	$38,$F9,$09,$8F,$A5,$F6,$47,$BA,$3E,$04,$6A,$38,$C7,$BF,$1F,$69
	.dc.b	$1C,$12,$3E,$51,$E5,$3C,$78,$47,$D5,$87,$A3,$1F,$83,$7E,$22,$5F
	.dc.b	$CB,$25,$7C,$28,$F3,$E1,$E5,$8F,$80,$7A,$4F,$39,$0C,$BB,$AF,$93
	.dc.b	$1B,$3E,$99,$E4,$3D,$07,$CF,$C7,$E5,$3A,$AF,$80,$FB,$8D,$63,$ED
	.dc.b	$BE,$D8,$FE,$2A,$7C,$58,$78,$2F,$8C,$7A,$1F,$7B,$ED,$C7,$DE,$E5
	.dc.b	$F9,$8F,$9A,$FA,$BF,$92,$3A,$E3,$E6,$BF,$97,$5C,$9F,$03,$F7,$CF
	.dc.b	$BA,$49,$E3,$8F,$37,$E3,$C6,$5F,$03,$C5,$F8,$17,$F2,$EF,$F5,$31
	.dc.b	$F1,$67,$C1,$8F,$DD,$9F,$21,$AC,$3F,$3F,$7C,$70,$F0,$47,$9B,$0F
	.dc.b	$99,$8F,$C6,$8C,$47,$53,$38,$FB,$70,$EB,$7F,$D2,$BE,$0B,$F8,$CD
	.dc.b	$7E,$A1,$AF,$DE,$6D,$FE,$09,$CC,$7F,$52,$F9,$6E,$58,$FF,$B5,$A1
	.dc.b	$18,$7F,$DB,$2D,$7F,$DF,$CD,$79,$1F,$BB,$3F,$DE,$7C,$9F,$C0,$5C
	.dc.b	$CF,$F2,$57,$26,$CE,$E7,$1F,$86,$4F,$18,$CC,$F1,$9C,$C7,$1A,$FB
	.dc.b	$4F,$4D,$3B,$3D,$1E,$17,$A8,$E7,$A7,$FB,$D2,$F9,$E1,$C3,$0F,$44
	.dc.b	$62,$7E,$24,$76,$8F,$0E,$1B,$BF,$D6,$1E,$1A,$DB,$89,$DC,$62,$7C
	.dc.b	$73,$C0,$F3,$47,$8F,$0F,$35,$F7,$6C,$FC,$57,$E2,$47,$99,$B1,$DD
	.dc.b	$DD,$76,$3C,$D8,$78,$0B,$7D,$EB,$E5,$1D,$4B,$1B,$8F,$41,$B9,$F8
	.dc.b	$07,$57,$DF,$1B,$9F,$C3,$9C,$DE,$EF,$F8,$80,$FC,$0B,$87,$FC,$47
	.dc.b	$87,$CF,$78,$D7,$56,$D7,$84,$F1,$BC,$E7,$E6,$BA,$31,$7F,$22,$FC
	.dc.b	$91,$F9,$83,$C9,$5E,$83,$E1,$47,$8E,$3C,$B1,$97,$99,$F1,$1D,$99
	.dc.b	$7D,$33,$F4,$A7,$BC,$7E,$92,$FD,$4F,$85,$AE,$6B,$D8,$FA,$4E,$67
	.dc.b	$E1,$70,$FB,$CF,$84,$F1,$D7,$9F,$0F,$91,$1E,$D8,$F0,$BF,$9E,$37
	.dc.b	$5E,$77,$DA,$77,$3F,$30,$59,$F9,$6F,$38,$F7,$BF,$72,$7A,$D7,$E5
	.dc.b	$63,$CC,$7E,$44,$B1,$A7,$E7,$9F,$28,$FE,$1D,$5F,$BA,$6E,$1F,$C4
	.dc.b	$5F,$82,$7E,$1C,$78,$2F,$27,$F8,$0E,$BC,$63,$15,$FE,$42,$A7,$C8
	.dc.b	$EA,$3F,$21,$3A,$9F,$F2,$CC,$7F,$97,$3E,$3B,$C5,$16,$3F,$CA,$11
	.dc.b	$88,$FF,$2C,$B7,$9B,$FC,$B1,$E7,$75,$94,$3C,$3F,$E3,$37,$FE,$5E
	.dc.b	$E7,$5E,$EB,$53,$69,$E6,$53,$1F,$1D,$67,$FC,$CF,$3A,$6F,$9B,$9C
	.dc.b	$BD,$67,$8A,$51,$E5,$8F,$F4,$7F,$7F,$AD,$B5,$F3,$DF,$F4,$3B,$63
	.dc.b	$51,$FE,$88,$8D,$78,$18,$23,$F5,$F6,$F2,$4F,$88,$FA,$F9,$78,$57
	.dc.b	$81,$E3,$9D,$F0,$FF,$3A,$D5,$F2,$8F,$73,$E3,$37,$2F,$F4,$3A,$EA
	.dc.b	$CC,$EB,$DC,$F9,$9E,$9B,$DF,$C6,$1F,$EE,$F7,$9F,$8D,$F6,$71,$FF
	.dc.b	$55,$4F,$F5,$FC,$BC,$33,$93,$5E,$17,$4F,$96,$F1,$B6,$BC,$7E,$17
	.dc.b	$9B,$D8,$F3,$B6,$3C,$F1,$DC,$8F,$7C,$51,$E3,$F5,$FD,$E7,$1E,$3F
	.dc.b	$B7,$9F,$C4,$8F,$87,$EA,$CC,$3E,$1F,$AD,$6F,$78,$AB,$E1,$FC,$B6
	.dc.b	$79,$3E,$77,$0A,$3E,$6F,$2F,$CE,$E2,$B9,$FC,$E7,$0F,$FA,$DA,$9F
	.dc.b	$2D,$8C,$FC,$B1,$89,$FA,$B8,$7C,$D2,$3E,$F7,$E8,$7C,$B4,$8F,$84
	.dc.b	$DA,$FD,$F6,$EB,$4F,$DC,$3F,$07,$85,$BF,$C1,$FC,$17,$E0,$0F,$C3
	.dc.b	$95,$FD,$3F,$A5,$FB,$7E,$F3,$49,$1F,$81,$F9,$5F,$D0,$E7,$F4,$3F
	.dc.b	$51,$FF,$7A,$CF,$FB,$5A,$73,$C1,$F9,$79,$CD,$7D,$69,$FD,$33,$C7
	.dc.b	$C8,$FD,$F4,$62,$B8,$47,$F0,$1A,$FE,$2B,$7F,$91,$6D,$FC,$26,$BF
	.dc.b	$D9,$F6,$46,$20,$7F,$10,$C1,$F1,$3D,$7C,$2F,$0B,$70,$F5,$7A,$7E
	.dc.b	$1F,$0F,$CA,$D8,$FF,$79,$FC,$EF,$E9,$31,$FA,$6D,$7C,$4D,$7C,$39
	.dc.b	$A3,$11,$E3,$F9,$8F,$77,$89,$F0,$9B,$7F,$2F,$4F,$E5,$40,$A4,$7E
	.dc.b	$27,$9F,$FC,$BF,$6F,$F7,$A4,$60,$98,$7E,$22,$9F,$21,$D0,$93,$F7
	.dc.b	$5E,$FF,$05,$F4,$7E,$23,$EA,$B6,$7E,$0F,$2F,$F3,$DD,$7F,$9A,$E3
	.dc.b	$F0,$A3,$07,$8F,$87,$ED,$1F,$CE,$FE,$3B,$F9,$FC,$FE,$26,$31,$F2
	.dc.b	$DF,$7F,$8A,$E3,$F6,$39,$BF,$15,$1F,$83,$F6,$BD,$C4,$9F,$C2,$3F
	.dc.b	$07,$DA,$FC,$41,$DF,$CF,$B3,$CC,$FA,$79,$BE,$5F,$F9,$76,$5E,$56
	.dc.b	$CE,$F4,$FE,$AF,$F2,$1F,$E7,$27,$F1,$CD,$7E,$EB,$5D,$F5,$FD,$3F
	.dc.b	$8B,$FC,$D1,$FD,$6F,$DD,$FE,$69,$7D,$7A,$FF,$2F,$3F,$CF,$FE,$17
	.dc.b	$FA,$87,$8C,$69,$2B,$F4,$5D,$9F,$7C,$8E,$0F,$E6,$DB,$3E,$17,$E8
	.dc.b	$7C,$27,$D1,$6C,$F8,$65,$38,$D7,$F9,$5C,$71,$9D,$4F,$D9,$F8,$5F
	.dc.b	$B5,$F0,$96,$79,$38,$69,$D6,$31,$7F,$D7,$FE,$5F,$F2,$33,$F3,$61
	.dc.b	$0E,$CF,$EE,$4E,$A3,$84,$60,$FE,$E9,$1B,$9F,$F1,$13,$67,$FD,$68
	.dc.b	$D7,$FC,$48,$BD,$CE,$24,$9E,$2F,$F7,$FF,$96,$F8,$15,$FD,$D5,$7C
	.dc.b	$37,$F8,$15,$CF,$67,$F8,$CD,$CD,$FA,$DB,$5F,$EC,$56,$BF,$E3,$56
	.dc.b	$BF,$E7,$D6,$BF,$C5,$BE,$57,$FC,$71,$CD,$FE,$3D,$6C,$7F,$B3,$DA
	.dc.b	$FF,$8F,$DA,$FF,$A1,$FE,$17,$FB,$3F,$CB,$7F,$91,$5B,$1F,$ED,$76
	.dc.b	$BF,$E4,$76,$BF,$E8,$F9,$FF,$16,$FB,$BE,$1F,$7B,$FC,$15,$3E,$16
	.dc.b	$BE,$16,$BF,$E9,$7F,$96,$7C,$7F,$FA,$67,$5F,$F2,$C7,$EC,$9F,$E1
	.dc.b	$CA,$FC,$03,$5F,$F1,$15,$70,$7F,$08,$FF,$16,$7F,$CB,$CD,$7F,$DE
	.dc.b	$0D,$7F,$CC,$0D,$7F,$D5,$1F,$43,$FD,$E5,$FF,$33,$35,$FF,$7C,$35
	.dc.b	$FF,$34,$35,$FF,$58,$7C,$C3,$E3,$FD,$5C,$3D,$3D,$8B,$5F,$D3,$2F
	.dc.b	$A6,$5F,$E7,$5C,$38,$4F,$E7,$5F,$D9,$3F,$E7,$76,$BF,$E7,$70,$E1
	.dc.b	$FF,$3D,$35,$FE,$93,$E0,$BE,$3B,$C3,$E1,$7F,$5E,$F9,$3B,$E8,$9C
	.dc.b	$D6,$EF,$F8,$73,$5F,$F1,$3B,$71,$FF,$26,$37,$4E,$FF,$0B,$CB,$E3
	.dc.b	$FA,$FA,$73,$0E,$71,$96,$90,$EB,$F0,$9E,$4F,$9A,$FF,$4B,$6B,$FD
	.dc.b	$D1,$BA,$9A,$6B,$C4,$7C,$8F,$1C,$3F,$93,$56,$DF,$E3,$11,$F0,$39
	.dc.b	$E4,$69,$F8,$B5,$FE,$1A,$8F,$E5,$DE,$5F,$F8,$01,$FF,$2D,$35,$FF
	.dc.b	$15,$B5,$FE,$7B,$E9,$7F,$9F,$FC,$BF,$F4,$2C,$7F,$9E,$7E,$2B,$04
	.dc.b	$7F,$C2,$9E,$BF,$D2,$23,$FC,$B7,$E6,$7E,$AF,$91,$F0,$09,$1F,$E1
	.dc.b	$B7,$E0,$7C,$C7,$A0,$91,$F6,$3E,$0F,$EE,$F9,$32,$3E,$1B,$F8,$2F
	.dc.b	$83,$FD,$9F,$57,$FA,$17,$67,$AD,$F4,$1B,$7F,$02,$30,$7D,$4A,$7C
	.dc.b	$8F,$32,$4F,$E2,$3E,$0B,$F1,$BF,$1D,$23,$E9,$BF,$92,$F9,$7F,$F6
	.dc.b	$4C,$7F,$5B,$D3,$FD,$51,$D3,$FE,$8D,$EE,$FC,$0C,$60,$9F,$2D,$F9
	.dc.b	$A4,$D3,$A7,$C1,$7B,$7C,$0D,$7B,$73,$1F,$F3,$74,$7E,$8B,$83,$FC
	.dc.b	$49,$1F,$73,$C1,$F8,$04,$7F,$D3,$1C,$1F,$9D,$7F,$DA,$2D,$D3,$FC
	.dc.b	$D5,$E2,$7A,$FD,$EF,$99,$F1,$FE,$5F,$93,$98,$7F,$9F,$63,$D1,$E7
	.dc.b	$7B,$09,$1F,$EA,$5E,$0F,$F6,$AB,$E0,$F1,$BE,$B1,$1F,$F0,$07,$1F
	.dc.b	$ED,$DF,$F4,$47,$C7,$FC,$3F,$80,$91,$83,$F3,$3D,$7F,$6C,$8F,$E6
	.dc.b	$9F,$B5,$F1,$FF,$83,$24,$76,$EE,$F9,$DE,$0C,$8F,$AC,$FE,$27,$E3
	.dc.b	$FE,$D7,$5F,$ED,$B3,$FF,$78,$D7,$FA,$FD,$AF,$FA,$41,$AF,$F9,$BF
	.dc.b	$E5,$9D,$9C,$A3,$DB,$5F,$C5,$A7,$31,$FE,$A5,$F9,$0F,$CC,$11,$FF
	.dc.b	$11,$47,$C8,$7F,$D7,$6D,$F0,$7F,$DC,$0D,$BF,$F3,$5B,$65,$CE,$87
	.dc.b	$2F,$E9,$7E,$3F,$F1,$DF,$0B,$2F,$FD,$6D,$1E,$5D,$B5,$18,$3C,$6F
	.dc.b	$E5,$7E,$12,$47,$E7,$5F,$67,$AB,$23,$D2,$D7,$D3,$1F,$04,$8F,$B9
	.dc.b	$F8,$9F,$07,$FB,$9F,$2A,$4F,$FB,$25,$F9,$1F,$03,$23,$E2,$B5,$F8
	.dc.b	$BF,$07,$FC,$07,$E1,$E4,$7F,$B4,$5F,$A1,$E8,$C8,$F9,$AD,$7E,$6F
	.dc.b	$C2,$7F,$5A,$58,$7F,$58,$5C,$7F,$61,$CE,$9B,$9D,$4E,$5F,$F1,$2C
	.dc.b	$FF,$8E,$FD,$F7,$8F,$FD,$76,$3F,$D6,$31,$87,$FC,$29,$19,$51,$97
	.dc.b	$FC,$2D,$E2,$79,$63,$17,$E3,$6F,$37,$F8,$7E,$7F,$B1,$FC,$9B,$E3
	.dc.b	$3F,$B2,$FD,$07,$F0,$CC,$7F,$CD,$CD,$7D,$B7,$FE,$B4,$AE,$11,$FE
	.dc.b	$73,$5F,$1D,$D2,$E4,$E6,$3B,$FC,$5C,$8E,$0F,$8B,$C9,$91,$CD,$AF
	.dc.b	$80,$FC,$BD,$FF,$34,$71,$D8,$CB,$F3,$31,$E8,$EB,$F5,$9F,$47,$DC
	.dc.b	$F8,$9E,$E7,$C9,$DC,$D2,$57,$F9,$07,$C8,$DB,$F5,$63,$E6,$47,$CD
	.dc.b	$F2,$3E,$FC,$BE,$5D,$F1,$FF,$E4,$BA,$FF,$74,$CE,$63,$0F,$FB,$B2
	.dc.b	$76,$FD,$EA,$E7,$18,$8F,$BB,$18,$9E,$31,$CC,$FC,$2D,$76,$AE,$18
	.dc.b	$2F,$85,$75,$E4,$F0,$61,$CA,$F7,$87,$18,$FF,$31,$5F,$0B,$F0,$DF
	.dc.b	$07,$B9,$4F,$C8,$D7,$17,$B1,$E5,$38,$9F,$D9,$AB,$D3,$EA,$F8,$20
	.dc.b	$64,$E2,$FA,$AF,$C5,$8A,$FB,$57,$80,$F1,$47,$BF,$0B,$5C,$2B,$73
	.dc.b	$C6,$B3,$51,$F3,$46,$5B,$3C,$2B,$C2,$7D,$EB,$F4,$4F,$9C,$F1,$16
	.dc.b	$BC,$4D,$9F,$13,$E6,$AE,$97,$E1,$BE,$64,$C3,$E8,$C7,$C4,$8F,$84
	.dc.b	$F1,$A8,$7A,$2F,$CB,$18,$C7,$CA,$70,$9D,$46,$6F,$F0,$07,$68,$C9
	.dc.b	$E9,$C3,$F2,$F8,$78,$E3,$E9,$CE,$EF,$AB,$CA,$3E,$FC,$77,$32,$FE
	.dc.b	$18,$EA,$F8,$E7,$A4,$7C,$87,$13,$D4,$E9,$1F,$32,$7A,$46,$63,$A4
	.dc.b	$76,$7F,$1A,$F2,$7C,$13,$DB,$1F,$A6,$E6,$3C,$15,$F9,$23,$E1,$47
	.dc.b	$68,$EB,$1E,$29,$FB,$75,$99,$F2,$4E,$71,$FA,$31,$F2,$4F,$B1,$1D
	.dc.b	$5F,$92,$7C,$67,$A6,$1F,$82,$8F,$A3,$87,$9E,$3E,$8C,$F9,$09,$3E
	.dc.b	$C7,$A3,$FA,$C9,$FB,$DF,$0B,$FE,$52,$8F,$D8,$B5,$FD,$8B,$71,$FC
	.dc.b	$AB,$98,$FD,$8F,$CB,$FE,$CA,$31,$08,$C5,$7F,$AB,$DB,$1F,$C7,$0D
	.dc.b	$09,$E3,$1F,$0A,$3F,$90,$FC,$4F,$F8,$6A,$BF,$8E,$47,$80,$FD,$50
	.dc.b	$8D,$61,$DE,$33,$18,$7C,$15,$D5,$FE,$13,$19,$7F,$97,$C7,$F1,$57
	.dc.b	$A7,$3D,$7C,$FA,$78,$AB,$FC,$5A,$F9,$23,$D7,$86,$6B,$B4,$79,$E3
	.dc.b	$0F,$A0,$E7,$87,$C3,$BE,$6F,$A6,$F6,$75,$C3,$C9,$3A,$AF,$59,$FC
	.dc.b	$9C,$F3,$94,$F1,$D7,$5C,$3C,$71,$83,$E6,$B8,$AF,$3D,$7C,$39,$F4
	.dc.b	$57,$17,$D1,$3E,$2A,$F8,$8D,$8F,$03,$6F,$D1,$1E,$5B,$F4,$3F,$38
	.dc.b	$FA,$B5,$E1,$69,$D8,$E0,$F6,$69,$D8,$B3,$D8,$F8,$F7,$F8,$1C,$3E
	.dc.b	$79,$E2,$2B,$F3,$4F,$64,$71,$69,$99,$F8,$A7,$78,$EA,$FD,$06,$BE
	.dc.b	$B6,$BF,$20,$E3,$3E,$19,$F9,$4F,$78,$EB,$35,$FC,$9D,$7B,$E7,$CF
	.dc.b	$1F,$79,$AF,$7A,$F3,$BE,$36,$B9,$3A,$BE,$B7,$A9,$EB,$76,$FE,$AE
	.dc.b	$BD,$2F,$E1,$4D,$CF,$A5,$F5,$CF,$37,$F1,$F3,$E9,$6C,$66,$FE,$FC
	.dc.b	$75,$8E,$EF,$E8,$0F,$29,$E9,$29,$E9,$3C,$F3,$D6,$79,$9F,$E9,$CD
	.dc.b	$7F,$CC,$D2,$F9,$7C,$63,$F9,$79,$CC,$62,$7F,$C5,$6F,$06,$CF,$F9
	.dc.b	$8E,$31,$1F,$E0,$C7,$5C,$7F,$E4,$5F,$47,$93,$6B,$FC,$23,$7F,$E7
	.dc.b	$3E,$3C,$F2,$2F,$7C,$AF,$2A,$98,$F3,$EE,$3C,$8F,$F0,$DE,$4F,$0B
	.dc.b	$FD,$7B,$19,$D7,$FD,$5F,$18,$7F,$CB,$38,$75,$9D,$57,$17,$CF,$5E
	.dc.b	$37,$2B,$C7,$CE,$5A,$64,$FF,$9E,$1B,$87,$F9,$9B,$3E,$13,$E3,$8F
	.dc.b	$F1,$8F,$6F,$CD,$7F,$C6,$7E,$BD,$7F,$D0,$93,$B5,$19,$3E,$7B,$88
	.dc.b	$FF,$4A,$46,$0F,$A7,$E0,$BD,$1E,$CD,$9F,$13,$ED,$AF,$F3,$F1,$FE
	.dc.b	$AC,$9F,$AF,$F2,$BF,$EA,$4A,$F8,$AD,$9F,$F4,$0A,$F9,$7D,$F8,$9F
	.dc.b	$7B,$D6,$7A,$B4,$CE,$1F,$3A,$32,$F4,$F0,$7B,$9E,$0F,$5A,$C6,$FE
	.dc.b	$5F,$A1,$8C,$FF,$D2,$36,$8F,$3F,$AF,$2B,$CF,$E2,$5B,$F3,$FA,$DE
	.dc.b	$C3,$1D,$9F,$EF,$EE,$87,$02,$7F,$2F,$B3,$CB,$A7,$1C,$BF,$0F,$19
	.dc.b	$7E,$2F,$A5,$F0,$7E,$27,$E4,$B6,$3E,$04,$74,$F4,$B1,$A7,$FA,$AA
	.dc.b	$D1,$F4,$19,$5F,$43,$9A,$D7,$C0,$F9,$6F,$9C,$63,$6F,$57,$12,$3E
	.dc.b	$CF,$CB,$FD,$A6,$E1,$F6,$9C,$3F,$E0,$BB,$70,$7C,$29,$FD,$CF,$07
	.dc.b	$3A,$D4,$E5,$B9,$BE,$47,$E3,$9F,$81,$C6,$CF,$DF,$F8,$0F,$C2,$1F
	.dc.b	$8C,$2B,$FE,$DE,$E1,$FE,$0F,$DB,$F2,$F4,$FF,$0F,$65,$FE,$CD,$9C
	.dc.b	$9F,$C3,$3E,$23,$FD,$33,$7F,$4B,$D6,$FE,$1B,$18,$BF,$99,$18,$F2
	.dc.b	$25,$7E,$FB,$DE,$FD,$6D,$7F,$B6,$AF,$FC,$03,$C5,$F8,$3D,$DE,$C7
	.dc.b	$F1,$A6,$E1,$F0,$23,$07,$8F,$D5,$FC,$3F,$A7,$E2,$6C,$FF,$0C,$C3
	.dc.b	$BE,$1F,$8F,$6C,$7F,$18,$C9,$EA,$7A,$BC,$5F,$E4,$78,$FC,$42,$57
	.dc.b	$D2,$E1,$47,$C3,$F1,$7F,$2F,$EB,$FC,$26,$D7,$E5,$BD,$8F,$CD,$9F
	.dc.b	$CC,$B2,$18,$8F,$CD,$FC,$BF,$BC,$8F,$DE,$7E,$17,$C2,$FE,$90,$E5
	.dc.b	$F0,$D6,$7D,$DF,$2F,$F4,$8E,$DF,$05,$63,$F9,$6E,$AB,$FB,$E4,$8F
	.dc.b	$63,$F3,$FE,$6B,$F6,$27,$DC,$F8,$9B,$3F,$2F,$87,$EF,$9F,$D3,$3E
	.dc.b	$29,$F0,$F8,$9F,$DA,$FC,$57,$D9,$23,$F9,$87,$FA,$AF,$EA,$7F,$E6
	.dc.b	$98,$FE,$82,$DE,$BF,$C2,$7C,$5F,$F5,$88,$C7,$4A,$BF,$DD,$34,$F1
	.dc.b	$FC,$31,$FA,$B8,$C6,$3F,$CC,$89,$1F,$AA,$E8,$ED,$3C,$BC,$1F,$B7
	.dc.b	$F0,$1F,$DA,$27,$33,$FD,$71,$AF,$ED,$B2,$7B,$FE,$53,$F4,$3F,$31
	.dc.b	$71,$FC,$87,$CC,$7A,$7D,$6F,$D2,$B6,$3D,$F1,$F0,$88,$FE,$FA,$FF
	.dc.b	$9E,$3F,$C0,$38,$5E,$2F,$CF,$3E,$87,$9B,$67,$FB,$4C,$E9,$F9,$7E
	.dc.b	$B6,$74,$71,$7F,$2F,$C7,$FA,$53,$FC,$C6,$DE,$0F,$DB,$3E,$E7,$8B
	.dc.b	$71,$EB,$1F,$4F,$E1,$B1,$8F,$49,$FF,$00,$E2,$9F,$ED,$9F,$30,$EE
	.dc.b	$FF,$0A,$B6,$BE,$B3,$6F,$F9,$13,$5F,$F5,$72,$F7,$BF,$BB,$7C,$FF
	.dc.b	$F6,$99,$FE,$CF,$39,$AF,$EF,$93,$9D,$7F,$0D,$41,$1C,$5F,$F5,$D3
	.dc.b	$5F,$C5,$B5,$E0,$D5,$F3,$3F,$CB,$EB,$F1,$CD,$8F,$E7,$4D,$7F,$D8
	.dc.b	$CD,$7F,$2E,$FD,$57,$FB,$FF,$A9,$FE,$14,$7F,$C0,$31,$9B,$F4,$C2
	.dc.b	$75,$18,$F9,$87,$C1,$FF,$B2,$6B,$FA,$54,$63,$CF,$FD,$13,$63,$33
	.dc.b	$B7,$FD,$B5,$F1,$0F,$6D,$F3,$3F,$A9,$8C,$36,$33,$3F,$D5,$E7,$2F
	.dc.b	$83,$E0,$BF,$DC,$4D,$EF,$F0,$C6,$23,$FD,$D1,$8F,$F9,$36,$7F,$3D
	.dc.b	$EF,$FE,$EB,$4F,$CA,$7C,$05,$E0,$FC,$5B,$70,$FD,$85,$5F,$09,$CE
	.dc.b	$F8,$AC,$29,$A6,$23,$3F,$01,$FC,$16,$B8,$B6,$8D,$31,$BF,$0C,$3F
	.dc.b	$8A,$C7,$68,$EF,$F0,$16,$7F,$8C,$C6,$1F,$F2,$77,$CB,$7F,$83,$1F
	.dc.b	$E2,$8D,$7F,$C0,$0D,$7F,$8D,$87,$0F,$F1,$E7,$FB,$F3,$D3,$C0,$F8
	.dc.b	$8E,$DB,$1A,$8F,$DA,$1F,$C5,$FE,$43,$E2,$91,$FE,$5F,$F0,$5F,$20
	.dc.b	$C3,$C3,$3F,$9B,$39,$8F,$E6,$BE,$57,$C7,$23,$FC,$5D,$F9,$5F,$0D
	.dc.b	$F4,$99,$1F,$CF,$23,$EA,$FB,$92,$3F,$9F,$BF,$77,$8F,$FA,$A7,$97
	.dc.b	$F6,$BE,$13,$EA,$12,$BF,$92,$3F,$5F,$B7,$EF,$FC,$1C,$9F,$E9,$31
	.dc.b	$F8,$2F,$81,$91,$FD,$31,$FC,$2F,$C0,$FE,$27,$F5,$4F,$E8,$D3,$FE
	.dc.b	$30,$6E,$E6,$93,$C1,$F3,$7A,$EF,$5E,$9C,$E3,$A7,$C5,$4C,$3F,$D3
	.dc.b	$77,$FE,$BD,$6F,$C1,$78,$7E,$07,$F2,$52,$69,$27,$FC,$7F,$1F,$E9
	.dc.b	$07,$9F,$C6,$7A,$4F,$57,$C1,$7A,$FD,$E9,$3E,$07,$DD,$EE,$FB,$FD
	.dc.b	$DE,$E4,$9E,$AF,$8B,$E3,$BE,$27,$BB,$27,$D3,$1F,$1F,$DB,$91,$EC
	.dc.b	$7E,$57,$B7,$F2,$FE,$5F,$D6,$49,$F4,$BE,$EF,$83,$FA,$04,$8F,$8B
	.dc.b	$F2,$FF,$4F,$E5,$FF,$05,$FB,$5F,$F1,$F9,$FF,$64,$77,$A3,$11,$F9
	.dc.b	$5E,$25,$7D,$BF,$85,$F9,$64,$9F,$92,$FC,$EF,$6B,$F0,$1E,$D4,$9F
	.dc.b	$AB,$1F,$83,$F9,$0F,$C9,$7E,$D9,$F5,$E7,$F5,$39,$A3,$0F,$FA,$5F
	.dc.b	$2F,$F7,$24,$FF,$B6,$3F,$57,$FF,$78,$F7,$9A,$4F,$53,$F8,$82,$F9
	.dc.b	$AF,$E3,$4F,$F8,$A2,$B9,$7C,$3E,$27,$3E,$E4,$9E,$CF,$F0,$0F,$A8
	.dc.b	$78,$BC,$BD,$BE,$A4,$AF,$F4,$3E,$3F,$EF,$B6,$FC,$96,$DF,$44,$F8
	.dc.b	$A3,$C7,$F2,$7E,$4F,$98,$95,$ED,$8F,$3F,$CC,$6A,$30,$7C,$17,$D7
	.dc.b	$F2,$5E,$CF,$5E,$4F,$C5,$8F,$7F,$C9,$EA,$30,$7C,$97,$E1,$FC,$64
	.dc.b	$8F,$98,$FC,$6F,$8A,$91,$EC,$7E,$59,$1F,$7F,$A9,$A7,$F8,$D6,$DF
	.dc.b	$E8,$A7,$EB,$47,$D0,$F8,$A9,$1F,$65,$FA,$7F,$25,$23,$EE,$35,$E7
	.dc.b	$F2,$5F,$63,$D3,$91,$F8,$17,$ED,$FA,$7F,$73,$E4,$6C,$FE,$16,$30
	.dc.b	$FF,$C0,$09,$FE,$73,$ED,$FF,$9C,$E3,$07,$F8,$5F,$BB,$FC,$03,$E8
	.dc.b	$FE,$43,$B7,$FC,$BE,$BD,$3F,$E3,$2D,$AF,$F3,$03,$5F,$F6,$63,$5F
	.dc.b	$E9,$BF,$B5,$FE,$21,$ED,$5D,$2B,$FA,$5D,$71,$BF,$F0,$8D,$EE,$FF
	.dc.b	$A2,$6D,$FD,$EE,$7F,$DF,$0F,$F9,$8A,$39,$9F,$ED,$A7,$0F,$0F,$87
	.dc.b	$FE,$F1,$0F,$91,$B1,$27,$8F,$85,$FE,$52,$9D,$FC,$8C,$9C,$DA,$39
	.dc.b	$FC,$84,$9C,$B5,$F3,$FC,$4F,$83,$E2,$A4,$72,$7E,$8E,$FF,$4B,$C5
	.dc.b	$FE,$87,$C3,$FB,$5A,$FF,$45,$35,$FF,$7B,$35,$FF,$44,$65,$FE,$11
	.dc.b	$7E,$09,$FE,$93,$9E,$8F,$A3,$E2,$BE,$0F,$C4,$C8,$F5,$BF,$13,$E2
	.dc.b	$24,$7B,$9A,$FD,$0F,$89,$F9,$3F,$0F,$23,$E1,$BF,$33,$E2,$24,$7C
	.dc.b	$66,$BF,$63,$3F,$A3,$F1,$1F,$E2,$1D,$0D,$24,$FF,$84,$8F,$B5,$F1
	.dc.b	$1C,$1F,$A3,$1F,$6A,$3E,$DF,$C4,$49,$FA,$AF,$DE,$F8,$79,$18,$21
	.dc.b	$F8,$A7,$F0,$7F,$11,$E3,$24,$78,$79,$FF,$D2,$71,$E7,$F8,$8D,$CF
	.dc.b	$E0,$A3,$F1,$71,$F8,$CF,$3F,$D8,$49,$E5,$9F,$E6,$5F,$F5,$F7,$EB
	.dc.b	$3F,$E2,$EE,$CF,$CC,$D7,$19,$FC,$8C,$E7,$9B,$FD,$11,$92,$3F,$CD
	.dc.b	$17,$FE,$8C,$8C,$7C,$4F,$F9,$4E,$73,$5C,$1F,$DC,$5A,$B9,$7C,$4E
	.dc.b	$EE,$4F,$82,$39,$78,$3F,$E5,$5B,$FD,$C4,$F3,$8C,$19,$F8,$9D,$C7
	.dc.b	$A2,$3C,$51,$FE,$44,$F8,$DF,$F2,$DE,$7F,$EA,$1F,$99,$78,$9F,$EE
	.dc.b	$89,$FC,$C6,$09,$CB,$FE,$A7,$8C,$C6,$21,$1D,$1E,$13,$CF,$AF,$FD
	.dc.b	$9B,$1F,$EE,$48,$ED,$D9,$FE,$F6,$F9,$4E,$57,$E0,$6D,$F0,$6E,$3E
	.dc.b	$7B,$E3,$85,$3C,$98,$7F,$C0,$0C,$77,$1E,$23,$D1,$87,$8A,$3B,$CE
	.dc.b	$EB,$9D,$6C,$F5,$D6,$E3,$D5,$7F,$EC,$7C,$38,$D5,$9E,$31,$E0,$3C
	.dc.b	$38,$6A,$7A,$63,$F2,$2B,$9C,$63,$0E,$CF,$A8,$F6,$1F,$33,$1F,$33
	.dc.b	$EA,$C7,$37,$F3,$F0,$EC,$4C,$3A,$BE,$37,$CA,$4A,$E4,$7D,$32,$C7
	.dc.b	$17,$C4,$79,$6B,$BE,$1F,$4E,$B9,$4F,$94,$F8,$47,$4C,$7C,$0F,$27
	.dc.b	$E1,$18,$7E,$6D,$7F,$BC,$7C,$4F,$E8,$78,$FE,$25,$B8,$7E,$21,$B3
	.dc.b	$BB,$FE,$25,$F3,$57,$0F,$F5,$2B,$5F,$F7,$E0,$70,$FF,$BF,$DA,$E4
	.dc.b	$F8,$40,$70,$38,$BD,$B1,$D2,$6B,$A8,$AF,$3A,$FD,$23,$93,$A9,$E1
	.dc.b	$2D,$75,$72,$78,$F4,$E3,$19,$79,$C6,$74,$FF,$61,$CE,$5F,$87,$1D
	.dc.b	$2F,$C7,$87,$9A,$BB,$61,$CA,$3F,$1B,$5E,$5C,$30,$FE,$84,$C6,$1D
	.dc.b	$23,$C7,$87,$09,$F1,$1F,$3B,$0F,$13,$F4,$8E,$4E,$1F,$90,$62,$39
	.dc.b	$5F,$95,$F3,$C7,$A4,$F8,$71,$E6,$9E,$51,$F2,$6B,$C6,$F9,$CF,$19
	.dc.b	$F6,$63,$6E,$0F,$59,$E5,$C3,$89,$F0,$0B,$87,$BC,$F9,$B8,$79,$A7
	.dc.b	$D0,$7F,$0B,$E2,$FE,$81,$F1,$BF,$67,$3F,$E8,$1C,$5F,$CA,$AF,$FA
	.dc.b	$3C,$E5,$FE,$81,$85,$C3,$FA,$05,$6D,$FF,$7E,$E1,$FD,$2E,$7F,$09
	.dc.b	$1F,$D5,$1A,$F2,$AF,$EC,$54,$F8,$DE,$97,$FD,$4F,$E3,$6C,$E6,$BF
	.dc.b	$AC,$C6,$08,$FF,$5C,$AB,$7F,$EF,$DA,$7C,$6F,$2F,$8C,$E6,$DB,$FE
	.dc.b	$C7,$1E,$4A,$F5,$B7,$1F,$EC,$8D,$B3,$7E,$37,$CB,$F1,$BE,$66,$E1
	.dc.b	$FD,$BA,$38,$11,$F7,$BC,$FE,$37,$DB,$F1,$BE,$E6,$CF,$F7,$68,$B9
	.dc.b	$7A,$E6,$F0,$7D,$6F,$8D,$F8,$BF,$1B,$F1,$9B,$87,$F7,$E8,$F4,$11
	.dc.b	$F9,$EF,$AF,$E3,$3E,$69,$F3,$BE,$32,$CF,$F8,$1B,$E2,$FE,$89,$8C
	.dc.b	$BE,$63,$F5,$3E,$32,$4F,$F8,$33,$E3,$2D,$33,$FF,$05,$7E,$B5,$FC
	.dc.b	$46,$BF,$CC,$2D,$9F,$AD,$7F,$80,$8C,$57,$0F,$47,$F1,$D4,$FD,$49
	.dc.b	$F1,$3B,$9C,$C6,$31,$FF,$2A,$B6,$BF,$C7,$CD,$7F,$95,$9C,$EB,$F0
	.dc.b	$0D,$9F,$F1,$A6,$0F,$8C,$F2,$1D,$BE,$2A,$CF,$AB,$E3,$3D,$07,$A7
	.dc.b	$0B,$1E,$E9,$F6,$18,$C3,$E2,$47,$F8,$C3,$04,$FC,$1F,$8A,$F8,$5F
	.dc.b	$17,$6B,$E3,$B8,$7E,$4F,$C5,$DC,$39,$9F,$E3,$1C,$11,$F2,$3D,$57
	.dc.b	$C4,$7C,$AE,$07,$B9,$98,$C3,$31,$FF,$32,$63,$F6,$63,$E2,$D7,$09
	.dc.b	$CB,$67,$FD,$0F,$E8,$F7,$3B,$7B,$96,$BE,$FF,$C5,$7D,$DF,$8B,$FB
	.dc.b	$CD,$8F,$C1,$FC,$5F,$E0,$4C,$61,$EC,$F1,$1B,$A7,$F0,$DA,$CB,$FC
	.dc.b	$5F,$C5,$DC,$38,$5B,$E2,$BF,$18,$67,$6F,$C9,$BF,$8F,$DB,$F2,$1F
	.dc.b	$15,$71,$FC,$AB,$87,$F0,$6F,$F0,$3B,$34,$C7,$C5,$7E,$64,$E1,$8F
	.dc.b	$89,$FC,$E7,$C4,$7E,$77,$E2,$6E,$1F,$A2,$70,$FE,$93,$E2,$AD,$FA
	.dc.b	$9F,$E4,$7F,$AD,$BD,$6B,$E9,$E1,$B7,$FE,$CC,$94,$F0,$F8,$AC,$E1
	.dc.b	$F2,$E7,$FD,$A9,$5B,$F8,$AE,$BF,$15,$D9,$B1,$CE,$3F,$91,$49,$A4
	.dc.b	$8F,$F4,$D4,$7E,$0F,$8B,$FD,$10,$D8,$FF,$0C,$C3,$E2,$BD,$7D,$7E
	.dc.b	$CE,$BF,$6F,$BB,$6F,$D4,$FC,$0F,$73,$C2,$48,$EE,$F9,$2F,$FA,$96
	.dc.b	$0F,$DB,$5D,$DF,$22,$73,$18,$BF,$8D,$F1,$7F,$33,$E2,$FE,$6B,$63
	.dc.b	$E5,$7C,$5F,$C0,$23,$EE,$7E,$1F,$85,$D3,$AB,$EA,$B6,$7E,$97,$C5
	.dc.b	$FD,$83,$EC,$FC,$5D,$8F,$AD,$F1,$7F,$30,$8F,$CA,$7E,$7F,$C5,$FE
	.dc.b	$00,$FC,$0F,$C5,$D9,$FB,$DF,$17,$F8,$4F,$8B,$FC,$2B,$63,$FC,$81
	.dc.b	$C5,$FE,$B9,$F6,$3F,$C8,$51,$FE,$CF,$6F,$5F,$F7,$66,$BF,$E4,$EF
	.dc.b	$25,$18,$8F,$F6,$2B,$FE,$41,$AE,$9F,$0B,$FD,$DA,$1F,$17,$63,$1F
	.dc.b	$17,$9F,$8B,$B9,$79,$0E,$6F,$1F,$88,$FE,$DD,$81,$A4,$9F,$47,$C4
	.dc.b	$77,$3C,$1F,$0F,$63,$A6,$9F,$ED,$1B,$7C,$47,$E1,$3D,$EF,$F8,$03
	.dc.b	$5A,$DF,$DB,$39,$E6,$F5,$7C,$47,$88,$95,$F1,$DF,$2F,$C3,$FB,$CF
	.dc.b	$81,$F0,$F6,$7A,$64,$F8,$7F,$89,$F0,$DF,$15,$B3,$AF,$87,$F7,$91
	.dc.b	$FA,$8F,$C2,$F8,$7F,$98,$7A,$7A,$AC,$F6,$C3,$E8,$7C,$3F,$D1,$6C
	.dc.b	$78,$A3,$E5,$91,$FC,$03,$F3,$BE,$1F,$EC,$70,$FD,$96,$CF,$AA,$2F
	.dc.b	$17,$F3,$1B,$56,$3A,$BF,$C7,$7F,$31,$8C,$3E,$8D,$A1,$87,$F7,$87
	.dc.b	$51,$FD,$58,$D4,$64,$FE,$E3,$1C,$27,$FA,$8B,$67,$FB,$B9,$C2,$3F
	.dc.b	$35,$F0,$FF,$91,$F8,$8F,$C9,$36,$3E,$9C,$7F,$92,$BF,$5D,$FF,$05
	.dc.b	$C7,$E6,$A3,$0C,$3A,$F3,$A7,$FC,$29,$7F,$A3,$6B,$FE,$AB,$8D,$4F
	.dc.b	$F9,$16,$5F,$13,$DF,$E2,$3C,$0D,$9F,$C8,$4F,$F8,$2D,$BF,$11,$9C
	.dc.b	$3F,$AF,$4A,$3F,$C3,$FF,$2D,$7C,$1F,$EE,$6D,$AF,$86,$D7,$FC,$36
	.dc.b	$D7,$FD,$8F,$F1,$DF,$CB,$FE,$5F,$FB,$DB,$63,$E5,$B5,$FF,$11,$B5
	.dc.b	$FF,$67,$CF,$F8,$BB,$DB,$F0,$FC,$5F,$F5,$19,$F0,$B5,$F0,$B5,$F8
	.dc.b	$DF,$2E,$F5,$FF,$A0,$EB,$F2,$75,$FF,$1C,$D7,$C4,$BC,$AE,$06,$1C
	.dc.b	$63,$F7,$0F,$F8,$61,$B8,$70,$9D,$B6,$75,$3F,$33,$E3,$3F,$C3,$75
	.dc.b	$F7,$5B,$87,$F8,$C2,$7F,$02,$D7,$FD,$33,$4F,$5F,$B4,$E7,$9F,$FD
	.dc.b	$BD,$87,$3B,$F2,$D7,$87,$D6,$F1,$32,$3D,$31,$EE,$5F,$2E,$F8,$AC
	.dc.b	$7D,$1B,$D1,$A6,$E7,$55,$FA,$E7,$FA,$EB,$FE,$36,$6B,$FA,$66,$BF
	.dc.b	$E5,$E6,$BF,$53,$D8,$F5,$FC,$74,$8F,$9B,$1F,$24,$8E,$0F,$A3,$EC
	.dc.b	$7D,$6D,$5F,$0A,$F8,$AF,$F2,$1B,$6B,$F6,$F3,$A9,$C9,$F4,$B8,$5E
	.dc.b	$CF,$3B,$E1,$7F,$E5,$AB,$E0,$FE,$02,$7E,$8F,$C7,$48,$FD,$5B,$F7
	.dc.b	$FD,$AE,$5F,$17,$E2,$24,$F8,$B0,$FF,$21,$3F,$BE,$C5,$E4,$FE,$1F
	.dc.b	$DB,$93,$F8,$E8,$F2,$7C,$4C,$8F,$E9,$BA,$7C,$08,$FC,$77,$B7,$E9
	.dc.b	$C7,$E0,$92,$BC,$0F,$C4,$F8,$99,$1E,$36,$BF,$E9,$2B,$7C,$C7,$EA
	.dc.b	$DF,$F7,$93,$7B,$FF,$BD,$B6,$7F,$4D,$F2,$FF,$DC,$1F,$2C,$3E,$27
	.dc.b	$FC,$EA,$D8,$FA,$6D,$7C,$0D,$7F,$C0,$12,$F6,$FF,$58,$D8,$FE,$D0
	.dc.b	$D7,$FC,$E8,$FF,$77,$7F,$8E,$FC,$87,$E9,$88,$F0,$F7,$3F,$11,$6F
	.dc.b	$86,$F2,$92,$7C,$AF,$ED,$AD,$EC,$F8,$89,$3E,$2F,$6B,$F6,$7E,$04
	.dc.b	$8F,$E5,$AF,$A3,$D9,$EF,$D5,$F0,$89,$3F,$09,$ED,$EC,$7E,$F3,$E1
	.dc.b	$A4,$FB,$A3,$CF,$E1,$7F,$00,$F9,$09,$3F,$3A,$3D,$7C,$FF,$38,$91
	.dc.b	$F3,$9F,$57,$AB,$F2,$89,$1F,$2B,$D7,$FD,$57,$CC,$7D,$98,$FE,$83
	.dc.b	$F2,$3F,$C3,$8C,$FC,$2E,$BE,$13,$FD,$5F,$46,$92,$BF,$AC,$FC,$BF
	.dc.b	$FB,$5D,$FF,$0F,$35,$FF,$55,$37,$B3,$F8,$CB,$FD,$62,$3F,$AA,$F8
	.dc.b	$DF,$C7,$88,$F2,$F6,$3F,$8C,$7C,$C3,$E2,$3F,$D8,$B2,$8C,$3F,$EC
	.dc.b	$86,$BF,$DE,$1A,$FF,$93,$6F,$FA,$0B,$FE,$03,$8D,$E0,$F0,$13,$27
	.dc.b	$31,$FD,$C6,$B8,$D7,$AB,$C9,$FE,$85,$06,$98,$F3,$26,$1F,$11,$AF
	.dc.b	$C4,$F0,$38,$F3,$C8,$F9,$2F,$4E,$A9,$1F,$31,$AF,$CC,$E7,$FE,$8B
	.dc.b	$E4,$C8,$FE,$FE,$F9,$39,$24,$7D,$46,$BF,$53,$87,$D2,$47,$EC,$BE
	.dc.b	$C9,$F6,$FC,$9F,$F5,$22,$4F,$DE,$7E,$0F,$C5,$6A,$30,$7E,$09,$FE
	.dc.b	$B2,$47,$FC,$26,$FC,$8F,$1F,$E4,$F8,$F6,$7F,$13,$18,$7F,$C3,$3E
	.dc.b	$47,$F6,$22,$3F,$E4,$0E,$FF,$E8,$6F,$F9,$0B,$C7,$FC,$E3,$F6,$8F
	.dc.b	$B6,$E1,$C3,$5C,$F5,$FE,$32,$1F,$15,$FD,$9D,$FD,$37,$37,$E1,$6B
	.dc.b	$67,$79,$FF,$01,$69,$FE,$E0,$8F,$F3,$F3,$54,$FF,$78,$8C,$37,$D9
	.dc.b	$E2,$49,$E2,$F4,$F2,$BA,$FC,$87,$F8,$0B,$59,$5D,$A3,$C3,$F2,$1A
	.dc.b	$8C,$1E,$27,$FC,$13,$BC,$8F,$F2,$1B,$E8,$F6,$BF,$C0,$9B,$FB,$67
	.dc.b	$FC,$BE,$D7,$E5,$1F,$DA,$3B,$3F,$63,$19,$8F,$9C,$FE,$ED,$FF,$37
	.dc.b	$3C,$F4,$E8,$D9,$DF,$C0,$7C,$8F,$63,$FC,$3B,$F2,$53,$0F,$93,$1F
	.dc.b	$33,$D9,$D4,$60,$F9,$AF,$D0,$F5,$FF,$C6,$5F,$AF,$7F,$55,$F6,$CD
	.dc.b	$E4,$4F,$D6,$F6,$78,$61,$F5,$A3,$0F,$FA,$D6,$BE,$C3,$FE,$D3,$6B
	.dc.b	$FE,$71,$6B,$FE,$47,$6D,$AD,$F1,$9E,$4F,$93,$97,$F8,$48,$FC,$3F
	.dc.b	$B3,$E8,$E7,$93,$F8,$88,$CB,$48,$7E,$2B,$83,$FD,$05,$87,$F8,$DE
	.dc.b	$3D,$1C,$DF,$E4,$EC,$CD,$24,$7F,$A3,$A7,$FC,$22,$F0,$21,$D4,$FF
	.dc.b	$31,$3A,$71,$9F,$CC,$E6,$EC,$48,$FF,$4C,$DA,$3F,$CB,$F5,$FC,$A2
	.dc.b	$7E,$34,$7D,$63,$F4,$FE,$4F,$F9,$A7,$E5,$9F,$0B,$F6,$BE,$4E,$5F
	.dc.b	$FB,$9E,$72,$D9,$E0,$47,$5D,$FF,$82,$EA,$F8,$44,$8F,$2B,$FE,$72
	.dc.b	$AF,$DE,$C7,$DF,$8E,$0F,$7E,$EF,$F3,$37,$C2,$FF,$9C,$7B,$3F,$A1
	.dc.b	$FC,$BF,$F8,$46,$3F,$C1,$55,$8F,$07,$FD,$3F,$ED,$FF,$0F,$AD,$76
	.dc.b	$7F,$00,$A4,$63,$0F,$E9,$BD,$3C,$EF,$F8,$B3,$6B,$38,$7F,$AA,$B0
	.dc.b	$E5,$3C,$67,$9C,$F8,$1F,$11,$FD,$63,$1F,$2B,$A3,$15,$E3,$C7,$91
	.dc.b	$DE,$34,$27,$8C,$6A,$BC,$58,$F9,$2B,$B6,$3E,$89,$F1,$43,$1F,$31
	.dc.b	$C5,$D9,$E6,$3B,$4F,$C7,$9E,$B1,$E2,$C3,$E8,$36,$74,$F2,$35,$8F
	.dc.b	$AE,$B8,$E3,$EB,$23,$83,$AE,$1E,$F2,$38,$35,$5D,$67,$8B,$F2,$31
	.dc.b	$F4,$63,$E4,$BE,$33,$E3,$7D,$F7,$CE,$30,$FC,$48,$CC,$73,$1E,$17
	.dc.b	$F4,$5A,$F7,$B7,$1F,$C4,$37,$0F,$F0,$9B,$FC,$4B,$E6,$2E,$1F,$CD
	.dc.b	$30,$46,$1F,$E6,$ED,$7F,$B1,$35,$FF,$0E,$BF,$6D,$FF,$0F,$F0,$7F
	.dc.b	$5D,$9F,$F3,$74,$68,$FB,$99,$7F,$42,$9D,$3E,$1A,$E5,$3F,$BA,$E8
	.dc.b	$AF,$90,$E2,$7F,$73,$AE,$2F,$98,$FC,$89,$C4,$A7,$1D,$3C,$71,$FD
	.dc.b	$C5,$F2,$C7,$5C,$33,$5E,$0B,$F1,$DF,$6A,$F8,$27,$67,$96,$1E,$98
	.dc.b	$C5,$7E,$55,$F8,$18,$74,$7C,$77,$9B,$E3,$1E,$99,$D1,$E4,$9D,$57
	.dc.b	$39,$F3,$5F,$C9,$8F,$34,$7C,$43,$E8,$B8,$AE,$73,$E5,$6C,$F7,$8E
	.dc.b	$4F,$90,$F4,$97,$0F,$41,$E3,$7B,$1F,$35,$C4,$F4,$8F,$8A,$F6,$E7
	.dc.b	$70,$FF,$AC,$F0,$FE,$C9,$BF,$F8,$87,$5F,$F1,$15,$6B,$2C,$57,$F6
	.dc.b	$5A,$B8,$FF,$65,$7F,$C0,$D5,$FD,$D1,$BC,$FF,$EB,$3A,$FE,$93,$1A
	.dc.b	$C5,$F2,$5C,$AF,$C6,$D8,$7B,$F6,$25,$FF,$79,$E1,$B9,$7F,$58,$C9
	.dc.b	$1C,$7D,$CE,$90,$F0,$71,$7F,$E0,$17,$27,$98,$8F,$F5,$E7,$CD,$EE
	.dc.b	$72,$F9,$4F,$23,$67,$FC,$15,$1E,$E2,$3F,$D9,$9F,$77,$C8,$FA,$CF
	.dc.b	$67,$C9,$59,$FE,$C1,$1E,$6F,$13,$1A,$7A,$DF,$85,$F2,$3E,$AF,$1A
	.dc.b	$DF,$F8,$65,$C3,$FD,$D1,$F9,$E4,$7F,$84,$71,$7D,$EF,$92,$F9,$9E
	.dc.b	$EF,$CD,$6C,$FF,$71,$8F,$8F,$E5,$E3,$2F,$98,$FD,$3F,$7B,$EB,$7C
	.dc.b	$BF,$FA,$A6,$FF,$BC,$46,$18,$7F,$8B,$7C,$BB,$7F,$55,$FC,$3A,$F9
	.dc.b	$87,$C7,$61,$7C,$75,$8C,$32,$FF,$91,$57,$18,$FF,$2A,$36,$33,$5C
	.dc.b	$9F,$F0,$BF,$C5,$FE,$08,$F4,$FC,$8D,$8F,$F6,$0F,$81,$F9,$ED,$1F
	.dc.b	$05,$EE,$CB,$DE,$DA,$D9,$FE,$C8,$E1,$57,$7C,$FF,$25,$E8,$6D,$B5
	.dc.b	$C1,$57,$0F,$80,$F9,$67,$CC,$F8,$0B,$5B,$9E,$BF,$01,$F4,$0C,$DF
	.dc.b	$F8,$DB,$14,$7C,$CD,$1E,$7D,$F3,$B0,$DC,$7C,$58,$7F,$A5,$A7,$E4
	.dc.b	$D1,$A4,$9F,$F1,$FF,$C3,$7E,$A8,$FB,$3C,$DF,$E6,$18,$FF,$35,$47
	.dc.b	$E1,$3E,$07,$F0,$BF,$01,$67,$D5,$A2,$3F,$67,$F1,$72,$73,$17,$3F
	.dc.b	$67,$C0,$FE,$3B,$E0,$3F,$1E,$DA,$E7,$1F,$BC,$F8,$7C,$57,$F0,$17
	.dc.b	$F2,$BF,$03,$2B,$C3,$ED,$5A,$F9,$1F,$03,$F9,$A3,$1C,$3F,$81,$7F
	.dc.b	$83,$FB,$52,$FD,$31,$6B,$E8,$7C,$07,$E8,$8F,$D1,$FC,$0D,$AF,$7F
	.dc.b	$C0,$7E,$98,$C5,$7D,$77,$F8,$F7,$CC,$7F,$8A,$67,$EF,$36,$7F,$CE
	.dc.b	$2F,$F8,$22,$33,$EC,$70,$F8,$19,$18,$38,$BD,$3E,$03,$F0,$5E,$8D
	.dc.b	$9F,$BF,$B7,$F2,$16,$C7,$E0,$08,$FE,$B2,$35,$8F,$FB,$0B,$B5,$FA
	.dc.b	$FA,$3F,$9B,$36,$BD,$3D,$B8,$9F,$6F,$B3,$27,$DB,$DD,$67,$FD,$21
	.dc.b	$8B,$E6,$7F,$33,$5F,$E1,$D8,$C7,$81,$FE,$CF,$6C,$66,$77,$5F,$ED
	.dc.b	$69,$E5,$DD,$E4,$F7,$BF,$A9,$C6,$2F,$F5,$5A,$7F,$83,$E3,$FA,$DF
	.dc.b	$17,$FB,$1A,$7E,$8F,$A3,$F5,$3D,$4F,$AA,$F2,$CD,$DB,$E3,$6D,$78
	.dc.b	$E3,$06,$7B,$3E,$2F,$C1,$71,$6C,$FD,$EF,$3B,$E4,$FC,$17,$E0,$9B
	.dc.b	$1F,$C1,$74,$F8,$9D,$5D,$C9,$1F,$05,$FE,$DB,$B3,$E2,$7E,$99,$F8
	.dc.b	$CF,$1A,$D7,$E3,$3C,$0F,$C7,$9F,$90,$F0,$AC,$77,$F4,$BF,$83,$BF
	.dc.b	$45,$BC,$DF,$35,$AF,$E0,$46,$68,$CC,$7F,$23,$C9,$F3,$3F,$86,$3F
	.dc.b	$BE,$7A,$DF,$47,$0D,$46,$34,$91,$FC,$7E,$5B,$E3,$0F,$C3,$E3,$F5
	.dc.b	$DF,$27,$27,$E1,$63,$11,$E5,$CF,$C9,$39,$F5,$BF,$07,$47,$C7,$F0
	.dc.b	$18,$73,$C5,$CC,$99,$FC,$37,$B7,$9F,$23,$E3,$37,$2F,$8D,$C5,$D8
	.dc.b	$F7,$FB,$96,$BD,$9E,$D7,$C2,$F6,$BE,$1B,$63,$E9,$79,$DE,$B2,$3F
	.dc.b	$4D,$FB,$FA,$7E,$C0,$F9,$7F,$19,$67,$E4,$7B,$7F,$A0,$FD,$BB,$F6
	.dc.b	$9A,$7F,$BE,$B9,$13,$98,$95,$FA,$FA,$DC,$7C,$96,$BF,$D9,$1A,$FF
	.dc.b	$58,$AF,$C0,$4F,$D7,$E1,$FB,$BF,$33,$F7,$99,$1F,$B0,$F8,$2B,$7F
	.dc.b	$88,$8F,$C3,$47,$F8,$0B,$E6,$79,$CF,$FA,$75,$AF,$E3,$C3,$87,$F2
	.dc.b	$4D,$7F,$8A,$2F,$7F,$FC,$1B,$F3,$BF,$DC,$FA,$18,$70,$9C,$FB,$9F
	.dc.b	$94,$F9,$AF,$F2,$87,$CC,$7F,$86,$F2,$F9,$11,$8B,$87,$F0,$AF,$86
	.dc.b	$B3,$EB,$3F,$BD,$F8,$1F,$E5,$A9,$D1,$FE,$30,$71,$1F,$CA,$DA,$FE
	.dc.b	$AF,$D0,$FE,$5B,$5F,$E6,$8C,$1F,$01,$E4,$F8,$0F,$2B,$71,$FF,$02
	.dc.b	$4F,$F8,$59,$FE,$3B,$F3,$79,$C7,$C9,$8F,$F3,$88,$FF,$17,$FC,$D7
	.dc.b	$F8,$73,$CF,$FF,$14,$B6,$BF,$D9,$8D,$7E,$83,$5F,$E0,$1F,$1B,$FE
	.dc.b	$69,$F9,$BF,$F1,$8B,$63,$FD,$A8,$D7,$EC,$35,$FE,$11,$3F,$E3,$EF
	.dc.b	$8E,$F0,$FC,$47,$FB,$92,$7C,$2D,$7C,$2D,$7F,$86,$78,$5F,$BD,$D3
	.dc.b	$F8,$77,$CB,$59,$FA,$2D,$D7,$FA,$BD,$5E,$27,$C5,$58,$97,$FD,$12
	.dc.b	$35,$E1,$FF,$93,$2F,$FA,$FB,$6C,$D3,$1E,$27,$0B,$FE,$9B,$1D,$E3
	.dc.b	$C1,$E1,$59,$FE,$A3,$18,$7F,$9F,$78,$3F,$E9,$C7,$FA,$AF,$83,$E7
	.dc.b	$3D,$9D,$89,$D4,$FD,$CC,$7F,$76,$F2,$EF,$DC,$EA,$BF,$D6,$F5,$AA
	.dc.b	$FC,$B0,$F9,$A7,$AF,$FE,$6A,$8C,$36,$33,$1F,$E6,$E6,$BF,$B3,$39
	.dc.b	$7C,$2F,$D1,$78,$9F,$E4,$C8,$DD,$FE,$C2,$3F,$C1,$11,$FE,$41,$7F
	.dc.b	$B9,$F8,$9F,$34,$8F,$E9,$DF,$A1,$E2,$7D,$96,$47,$F7,$A8,$FB,$9E
	.dc.b	$14,$8F,$EF,$AF,$C2,$E4,$FC,$07,$83,$F6,$49,$3F,$B3,$7E,$F7,$93
	.dc.b	$E5,$8F,$D3,$AF,$92,$FC,$53,$63,$FC,$E4,$EF,$3F,$F7,$1B,$FB,$68
	.dc.b	$CF,$83,$C0,$E3,$E1,$72,$EB,$EE,$49,$FF,$59,$D7,$FA,$81,$BE,$27
	.dc.b	$DA,$E5,$91,$E0,$8F,$1F,$C4,$C8,$F1,$BE,$6F,$8A,$F3,$FB,$9E,$C2
	.dc.b	$4F,$67,$DF,$F2,$9F,$8C,$FD,$BF,$D5,$E0,$FF,$A8,$1B,$5F,$E9,$4C
	.dc.b	$D5,$F0,$BA,$3E,$1F,$44,$AF,$87,$1F,$1B,$E2,$BE,$59,$23,$DA,$FC
	.dc.b	$EF,$95,$F9,$7E,$EC,$9F,$97,$1F,$3B,$E2,$FE,$7F,$BB,$27,$E7,$C7
	.dc.b	$D2,$F8,$BF,$AE,$48,$F9,$2F,$DA,$F8,$AF,$AF,$F1,$72,$7E,$BC,$7D
	.dc.b	$AF,$80,$F8,$19,$49,$FB,$71,$FC,$1F,$F6,$2F,$E1,$D1,$FE,$BB,$6B
	.dc.b	$FE,$B6,$6F,$92,$68,$F8,$CF,$99,$84,$AF,$F5,$C4,$E5,$BF,$19,$F4
	.dc.b	$70,$91,$83,$7F,$2D,$FC,$46,$8D,$24,$7F,$BE,$A1,$F1,$BF,$C1,$BF
	.dc.b	$70,$DC,$FF,$92,$5A,$FF,$42,$9D,$BE,$4F,$8D,$F8,$04,$7F,$D8,$7F
	.dc.b	$33,$C3,$C3,$FD,$C6,$DE,$FF,$8D,$F2,$12,$BC,$4F,$9F,$E3,$BE,$53
	.dc.b	$23,$E0,$C7,$CD,$F9,$5F,$9D,$F2,$12,$7E,$2C,$65,$A4,$3D,$1F,$1F
	.dc.b	$ED,$99,$1E,$D7,$EC,$7C,$44,$8C,$10,$F8,$0F,$C5,$3E,$F7,$C4,$48
	.dc.b	$F8,$D1,$F8,$1F,$92,$D4,$60,$FB,$2F,$E1,$7E,$27,$51,$83,$EE,$BF
	.dc.b	$44,$8F,$D1,$7F,$19,$F2,$72,$3F,$06,$D7,$EA,$78,$7F,$64,$FC,$97
	.dc.b	$C5,$48,$FB,$51,$F9,$5F,$8A,$91,$F8,$C7,$F3,$1F,$1D,$FC,$C7,$F5
	.dc.b	$FF,$F3,$94,$FF,$BF,$1A,$FF,$40,$6F,$1F,$F1,$FE,$4F,$F7,$F4,$FE
	.dc.b	$03,$8B,$FA,$14,$1A,$48,$FE,$80,$3E,$81,$87,$F4,$1A,$E1,$3C,$CF
	.dc.b	$F6,$35,$FF,$2D,$AD,$C7,$E5,$6B,$F3,$04,$7F,$4F,$F2,$3F,$B3,$30
	.dc.b	$FF,$B1,$BA,$78,$BF,$9B,$8F,$EC,$B3,$FE,$C1,$F2,$FB,$B6,$7F,$D2
	.dc.b	$AD,$E8,$7E,$C9,$49,$3F,$E0,$77,$1F,$11,$7E,$4E,$4F,$AB,$0F,$F7
	.dc.b	$05,$F2,$9F,$EC,$1F,$AF,$FF,$93,$EB,$47,$FB,$95,$7B,$3F,$DB,$5B
	.dc.b	$1F,$29,$AF,$F6,$B6,$BF,$E4,$3C,$7F,$D4,$2F,$41,$1C,$36,$FE,$B0
	.dc.b	$DC,$7C,$B5,$C2,$BF,$DD,$B9,$DB,$FE,$E4,$7D,$8F,$92,$F9,$27,$C1
	.dc.b	$F9,$39,$5E,$AD,$3E,$3C,$62,$3F,$C9,$78,$7F,$AF,$1F,$F6,$B6,$15
	.dc.b	$EF,$F2,$9F,$50,$F9,$9F,$29,$23,$A5,$23,$11,$D3,$E1,$7E,$97,$C8
	.dc.b	$FF,$8C,$68,$D2,$4F,$F8,$47,$2F,$B7,$F2,$1C,$23,$AC,$FD,$B8,$FC
	.dc.b	$29,$23,$CE,$FD,$FF,$8C,$91,$82,$1E,$E7,$F0,$9F,$05,$E3,$24,$78
	.dc.b	$70,$47,$9F,$E3,$37,$3E,$E8,$FC,$64,$7E,$37,$E3,$3D,$84,$9E,$4F
	.dc.b	$BF,$E2,$F5,$1F,$1A,$3F,$29,$1F,$95,$F8,$CF,$CB,$7C,$14,$AF,$97
	.dc.b	$1F,$99,$F9,$4D,$46,$0F,$9E,$FE,$77,$E2,$BE,$61,$23,$D4,$FC,$FF
	.dc.b	$8A,$D4,$7D,$68,$FD,$24,$7F,$8F,$7E,$41,$E7,$7F,$C0,$0A,$7C,$DF
	.dc.b	$DB,$42,$7F,$81,$BA,$E5,$FF,$44,$F8,$9A,$8C,$1C,$27,$AF,$5F,$2B
	.dc.b	$E9,$3C,$F0,$F0,$7B,$18,$BE,$AD,$C3,$A3,$99,$DE,$3F,$E9,$87,$D1
	.dc.b	$86,$F0,$ED,$3E,$46,$9E,$39,$DC,$71,$C7,$33,$ED,$C3,$D5,$7C,$E3
	.dc.b	$D6,$79,$E7,$78,$FB,$23,$FC,$0D,$7D,$DA,$7C,$1B,$F4,$C7,$78,$C9
	.dc.b	$CC,$EB,$3F,$32,$FC,$07,$BE,$FB,$E1,$E8,$AE,$87,$8B,$0F,$3E,$1D
	.dc.b	$6F,$84,$F4,$9C,$BC,$0E,$F5,$CD,$F1,$63,$EF,$25,$F9,$DE,$77,$E4
	.dc.b	$8F,$73,$87,$47,$C1,$8E,$33,$F3,$F0,$FC,$4C,$79,$0C,$C7,$CF,$33
	.dc.b	$3C,$9E,$03,$E4,$3F,$C5,$CF,$C7,$6E,$5F,$45,$B8,$FE,$65,$DC,$FE
	.dc.b	$63,$E7,$3F,$33,$18,$84,$63,$0F,$F5,$03,$63,$F3,$46,$8F,$14,$69
	.dc.b	$FB,$23,$C3,$FF,$51,$47,$E7,$23,$C0,$6A,$33,$8F,$19,$D4,$62,$7B
	.dc.b	$D7,$C2,$7F,$55,$19,$7F,$59,$1F,$96,$7C,$1E,$05,$7C,$DA,$74,$8C
	.dc.b	$4F,$39,$FD,$CD,$73,$7C,$B7,$B7,$D6,$72,$C3,$E0,$CF,$FA,$F7,$0F
	.dc.b	$64,$E4,$F2,$E1,$D9,$FD,$29,$E4,$29,$B3,$C2,$EE,$3C,$07,$4C,$3D
	.dc.b	$95,$DE,$FE,$31,$E2,$AF,$55,$FA,$23,$0F,$A2,$B8,$CF,$59,$CD,$F5
	.dc.b	$AF,$2D,$F5,$7E,$79,$8A,$F5,$61,$F0,$CE,$05,$BF,$C0,$5F,$A1,$A7
	.dc.b	$63,$FB,$AF,$C7,$7F,$72,$AF,$EE,$FE,$1D,$C1,$DB,$FE,$16,$AC,$3F
	.dc.b	$E1,$89,$BD,$5F,$E0,$CB,$FE,$A5,$1F,$ED,$9B,$E7,$18,$3F,$B3,$7D
	.dc.b	$91,$E6,$FF,$B7,$5F,$EF,$CE,$5B,$F1,$FA,$37,$F1,$F6,$36,$F2,$F8
	.dc.b	$FE,$7F,$1F,$67,$9B,$87,$C2,$F9,$70,$F6,$E0,$8C,$66,$8C,$57,$A3
	.dc.b	$E3,$FC,$FF,$1F,$E8,$6C,$7A,$23,$8C,$9A,$48,$F8,$3F,$1F,$F3,$7A
	.dc.b	$3F,$CA,$FD,$BF,$0A,$31,$E4,$22,$92,$BE,$2C,$DD,$17,$FE,$A4,$C5
	.dc.b	$F1,$FF,$33,$E3,$BE,$6B,$70,$F9,$B1,$E6,$23,$F5,$1F,$4F,$C6,$FD
	.dc.b	$43,$EA,$FC,$75,$9F,$F1,$6F,$81,$E6,$31,$97,$F8,$F1,$FB,$7F,$1D
	.dc.b	$27,$ED,$FC,$75,$AF,$F2,$0F,$3F,$B8,$C6,$BF,$E4,$2E,$0F,$C5,$4F
	.dc.b	$B2,$FF,$CB,$BB,$3E,$2B,$F2,$92,$F9,$BF,$9D,$5F,$88,$9C,$C6,$27
	.dc.b	$91,$FE,$02,$70,$FF,$26,$6B,$FE,$01,$6B,$BE,$3F,$B7,$F1,$DF,$96
	.dc.b	$3A,$F1,$DC,$3F,$2F,$F1,$FF,$9A,$3F,$37,$F1,$F6,$3F,$CE,$3A,$BF
	.dc.b	$71,$76,$7E,$05,$B3,$C8,$DB,$8C,$FD,$3D,$5F,$93,$F8,$FE,$1D,$9F
	.dc.b	$A8,$6D,$F0,$8B,$5C,$38,$55,$8C,$BF,$61,$3F,$E2,$4F,$84,$69,$F3
	.dc.b	$32,$FE,$45,$5A,$A5,$6B,$2F,$F4,$05,$69,$FF,$14,$3F,$CA,$67,$FA
	.dc.b	$8B,$5C,$F8,$BF,$AF,$7F,$85,$B6,$9F,$21,$F7,$1B,$7D,$BD,$BC,$4F
	.dc.b	$F0,$62,$47,$F0,$6F,$8F,$B8,$7D,$EC,$BE,$77,$27,$F1,$6E,$2F,$EE
	.dc.b	$57,$F3,$23,$1E,$56,$7C,$99,$3F,$E8,$DA,$7C,$86,$31,$E9,$C8,$F0
	.dc.b	$7E,$C1,$9C,$FF,$D3,$31,$9C,$39,$7C,$87,$87,$E4,$3C,$4D,$8E,$D1
	.dc.b	$F8,$1F,$0B,$18,$F6,$71,$3E,$3F,$8F,$B3,$F8,$2D,$11,$ED,$F8,$FF
	.dc.b	$C5,$77,$DB,$FC,$4B,$86,$9E,$D8,$C5,$F3,$F4,$BE,$31,$F1,$FD,$1B
	.dc.b	$5F,$0E,$3E,$57,$C7,$62,$BC,$4F,$CD,$F8,$FF,$9D,$F1,$D6,$FE,$53
	.dc.b	$87,$FD,$49,$A2,$3F,$96,$B6,$33,$18,$F8,$FF,$B0,$7E,$4F,$E0,$AC
	.dc.b	$FD,$5E,$CF,$D4,$75,$7D,$D6,$C7,$D9,$F8,$FE,$A4,$7C,$2F,$83,$E3
	.dc.b	$FF,$98,$7E,$E4,$E6,$FC,$34,$61,$FC,$47,$C7,$EB,$96,$E1,$FB,$AB
	.dc.b	$FA,$2F,$E1,$FE,$43,$D6,$47,$DE,$FB,$BE,$43,$BF,$17,$E4,$DB,$3F
	.dc.b	$8F,$F3,$7F,$2D,$F1,$9F,$97,$6C,$7F,$B7,$FE,$DF,$FD,$32,$FF,$B3
	.dc.b	$5B,$CD,$FA,$E6,$BF,$B2,$6B,$FE,$A0,$F1,$D1,$FB,$78,$FE,$F6,$BC
	.dc.b	$6F,$EF,$6E,$67,$71,$8F,$8C,$CF,$C6,$5C,$3C,$8F,$0E,$6F,$EC,$F8
	.dc.b	$1A,$49,$F4,$7B,$DD,$BE,$33,$BB,$63,$DD,$F1,$9F,$D8,$FF,$73,$F2
	.dc.b	$46,$0F,$E5,$2E,$35,$FE,$47,$8B,$E3,$7C,$24,$AF,$8C,$F9,$3E,$37
	.dc.b	$DA,$7B,$BE,$36,$CF,$CD,$F8,$DF,$06,$5F,$09,$B1,$F4,$7E,$37,$DA
	.dc.b	$47,$E9,$3F,$03,$E3,$7E,$49,$F2,$BE,$0A,$CF,$D9,$F8,$DF,$9B,$F1
	.dc.b	$BF,$39,$B1,$F7,$7E,$37,$E4,$91,$FB,$CF,$CC,$F8,$3F,$AA,$7C,$5C
	.dc.b	$AC,$FE,$17,$E2,$BE,$CF,$C1,$7D,$A6,$C7,$E2,$63,$EA,$91,$FC,$53
	.dc.b	$FC,$CF,$E7,$18,$7F,$6C,$58,$58,$C3,$2B,$FD,$1B,$C7,$FE,$EC,$AD
	.dc.b	$C6,$21,$18,$76,$D7,$FA,$3F,$ED,$1F,$E3,$D7,$F9,$C7,$2B,$5F,$F2
	.dc.b	$6C,$E7,$D5,$FE,$AE,$27,$58,$7E,$12,$B7,$F1,$7F,$89,$6C,$7F,$C0
	.dc.b	$0A,$FE,$BF,$DC,$C3,$FB,$64,$61,$7C,$5F,$83,$E0,$FC,$2C,$BE,$47
	.dc.b	$EA,$E7,$FC,$87,$9F,$F7,$47,$8C,$6A,$3F,$CC,$F1,$AB,$ED,$F0,$7F
	.dc.b	$E4,$C8,$E3,$3F,$0A,$30,$67,$E2,$FE,$0F,$BF,$F0,$9B,$87,$5C,$FD
	.dc.b	$53,$F1,$BE,$17,$38,$78,$DF,$F1,$2F,$CD,$7B,$3E,$0F,$EA,$36,$7E
	.dc.b	$CB,$5F,$F6,$3B,$51,$3E,$41,$E5,$BE,$6F,$EF,$46,$1B,$19,$9F,$C2
	.dc.b	$CE,$5E,$FA,$FF,$B8,$DB,$F0,$9E,$48,$C4,$78,$9A,$FF,$19,$F9,$A7
	.dc.b	$85,$E4,$8F,$A1,$F1,$F7,$8F,$ED,$37,$87,$85,$5A,$FF,$4C,$DB,$C4
	.dc.b	$C7,$12,$3A,$CF,$CD,$6B,$FE,$30,$78,$3F,$47,$93,$FC,$04,$FE,$8A
	.dc.b	$7F,$19,$F0,$9F,$A1,$6C,$66,$76,$FF,$89,$29,$E2,$78,$A7,$3A,$7E
	.dc.b	$4F,$0F,$2D,$6A,$FC,$9E,$26,$30,$91,$83,$D1,$E1,$F3,$F8,$F9,$18
	.dc.b	$3F,$CA,$3F,$36,$F2,$FA,$D4,$9C,$C7,$C7,$5F,$37,$E4,$E1,$FA,$1F
	.dc.b	$50,$F5,$74,$DA,$FD,$01,$C3,$6E,$56,$CF,$F8,$24,$F4,$6C,$66,$3F
	.dc.b	$84,$35,$FE,$1B,$F9,$47,$F4,$B7,$F8,$63,$5C,$FA,$1D,$23,$11,$FE
	.dc.b	$52,$E0,$FE,$36,$F9,$23,$87,$0F,$F9,$6D,$BA,$7E,$37,$C8,$FA,$5F
	.dc.b	$0F,$2F,$EF,$3F,$4B,$CB,$FC,$56,$0E,$2F,$C2,$D7,$E4,$5A,$FF,$0F
	.dc.b	$6B,$FE,$B4,$6F,$33,$DA,$79,$1F,$7B,$CC,$97,$FB,$D9,$CB,$F6,$3A
	.dc.b	$3C,$38,$1A,$7E,$6A,$7F,$D1,$73,$F7,$BC,$EF,$E6,$59,$79,$DB,$7F
	.dc.b	$EC,$B9,$FE,$73,$39,$6C,$3A,$3F,$29,$F2,$72,$7F,$13,$1E,$2D,$E4
	.dc.b	$7E,$31,$F1,$78,$FE,$6D,$FB,$92,$7A,$BE,$4F,$9D,$FF,$46,$65,$FC
	.dc.b	$C7,$E5,$3F,$44,$7C,$E8,$CF,$1F,$F8,$67,$67,$1F,$E9,$FE,$2A,$65
	.dc.b	$F1,$E3,$87,$87,$D4,$91,$EE,$C3,$FD,$F8,$FE,$DE,$FF,$D9,$4F,$EE
	.dc.b	$FE,$4F,$F6,$47,$CB,$C3,$E6,$7C,$3F,$ED,$49,$3F,$55,$F3,$76,$7A
	.dc.b	$C9,$1D,$DF,$E0,$57,$FE,$9C,$8F,$E1,$1F,$23,$FB,$F3,$EC,$7C,$8F
	.dc.b	$BF,$0F,$E0,$5F,$25,$2B,$FD,$8F,$9F,$99,$FF,$4E,$2C,$BE,$51,$23
	.dc.b	$DC,$FC,$D9,$91,$F0,$5A,$FF,$29,$1C,$0A,$7C,$47,$F1,$3C,$65,$78
	.dc.b	$A3,$F4,$3F,$32,$F9,$3F,$C6,$B6,$7F,$16,$D7,$B6,$08,$C4,$7F,$37
	.dc.b	$7F,$CE,$0F,$FB,$91,$AF,$FA,$E5,$B1,$FD,$E3,$E7,$5F,$25,$FC,$9F
	.dc.b	$09,$7F,$CC,$DF,$E7,$73,$FE,$A4,$8F,$E2,$7F,$39,$D1,$F8,$0D,$7F
	.dc.b	$A5,$35,$FE,$BE,$DF,$05,$2F,$93,$FE,$49,$3E,$4E,$7E,$3F,$17,$E0
	.dc.b	$9F,$0F,$A7,$FD,$33,$F6,$9F,$EB,$7A,$FF,$84,$1B,$7F,$E7,$3C,$7F
	.dc.b	$B5,$E7,$FD,$ED,$7C,$A7,$28,$C6,$1E,$C8,$FA,$2F,$5E,$09,$3F,$26
	.dc.b	$32,$D2,$1F,$7A,$FC,$1F,$25,$F1,$BE,$4A,$4F,$85,$FB,$B1,$23,$ED
	.dc.b	$BF,$78,$8F,$F5,$FC,$5F,$25,$B8,$FB,$13,$F4,$A3,$F0,$FF,$25,$F0
	.dc.b	$FE,$5A,$57,$DF,$8F,$AF,$F2,$7A,$8C,$1F,$0D,$FB,$7E,$9C,$8F,$8C
	.dc.b	$FE,$4B,$BA,$47,$E2,$5A,$FB,$C7,$2F,$DA,$F9,$1F,$A9,$CB,$27,$FC
	.dc.b	$55,$C7,$F6,$DF,$F0,$FF,$CF,$7F,$98,$1F,$C4,$F3,$7F,$58,$E7,$FC
	.dc.b	$14,$7F,$97,$5A,$FF,$98,$9A,$FE,$3B,$E5,$BF,$0F,$32,$3F,$27,$1F
	.dc.b	$AA,$F9,$2F,$C3,$75,$C9,$FD,$14,$65,$A4,$3D,$9A,$7E,$53,$E4,$FF
	.dc.b	$A9,$CF,$EB,$B9,$FF,$D7,$DA,$7E,$86,$F6,$DE,$BF,$F3,$4D,$6E,$7E
	.dc.b	$2F,$CD,$59,$4D,$35,$8A,$6D,$FC,$0A,$FC,$E6,$9F,$9F,$3F,$C5,$73
	.dc.b	$FE,$CB,$C3,$FC,$07,$F3,$AF,$2F,$FC,$02,$D8,$FF,$43,$35,$FF,$45
	.dc.b	$B5,$FE,$2C,$BD,$3F,$F0,$A3,$BC,$1A,$7F,$90,$E3,$0D,$9F,$8B,$F2
	.dc.b	$9F,$B3,$E0,$98,$7F,$9D,$70,$F7,$CF,$A3,$E5,$37,$39,$DD,$39,$8C
	.dc.b	$7B,$3F,$BD,$F0,$A4,$60,$FF,$3A,$6C,$F9,$4F,$AB,$52,$73,$1F,$63
	.dc.b	$DE,$FE,$EB,$8F,$F8,$CE,$7F,$B9,$B5,$FE,$52,$D7,$FC,$C1,$C7,$F4
	.dc.b	$FE,$4B,$94,$79,$EB,$E8,$C7,$E1,$3E,$4F,$F8,$97,$8B,$2B,$D9,$19
	.dc.b	$69,$0F,$C6,$E7,$FE,$5E,$7F,$55,$BF,$E1,$3E,$13,$CE,$48,$F3,$BF
	.dc.b	$E5,$F9,$FF,$33,$CF,$E3,$A3,$F2,$A7,$E1,$0F,$E4,$9F,$31,$FE,$65
	.dc.b	$E8,$34,$97,$FE,$6B,$BF,$F3,$03,$F9,$A8,$E2,$FC,$EF,$83,$FC,$7F
	.dc.b	$B3,$FC,$BF,$C7,$93,$F4,$A3,$F2,$7E,$E6,$A3,$07,$D6,$7F,$9B,$79
	.dc.b	$5E,$32,$47,$F9,$EF,$24,$7B,$7D,$AD,$CF,$F0,$88,$FC,$FC,$7F,$99
	.dc.b	$FE,$23,$F9,$CE,$DF,$EA,$8F,$9C,$FC,$15,$FF,$AA,$7A,$FF,$D3,$11
	.dc.b	$88,$52,$73,$CF,$FE,$9D,$8E,$72,$AE,$11,$8F,$3A,$DF,$FA,$8D,$B5
	.dc.b	$FE,$B8,$77,$D5,$FD,$FA,$3F,$BC,$61,$D6,$31,$87,$09,$ED,$3E,$1B
	.dc.b	$F1,$C7,$FB,$06,$B7,$3C,$71,$F7,$3B,$9A,$7B,$CA,$7B,$A7,$8B,$CE
	.dc.b	$7C,$D3,$C6,$3C,$A7,$12,$9C,$63,$6F,$13,$CE,$6E,$FD,$47,$C0,$8F
	.dc.b	$0E,$3F,$2A,$33,$5D,$DD,$57,$23,$B1,$67,$EA,$3C,$B0,$F8,$6E,$A7
	.dc.b	$8E,$3F,$24,$97,$F1,$67,$D3,$7E,$18,$F0,$12,$3D,$55,$CA,$3E,$15
	.dc.b	$72,$9F,$73,$97,$E2,$39,$7C,$EF,$AB,$E2,$7F,$C6,$17,$F8,$86,$E5
	.dc.b	$F8,$16,$E5,$FE,$FC,$7F,$C6,$76,$F9,$DC,$61,$FE,$45,$8C,$09,$DC
	.dc.b	$EA,$3F,$C9,$4D,$7F,$DE,$8F,$D6,$43,$E3,$7F,$9B,$BF,$70,$CC,$6D
	.dc.b	$D7,$83,$5F,$0D,$FE,$9D,$C9,$C4,$CE,$7F,$E5,$BB,$EB,$3B,$9F,$4E
	.dc.b	$9C,$67,$C2,$E7,$0F,$CA,$CF,$48,$C3,$E5,$9E,$77,$F0,$EB,$CD,$18
	.dc.b	$9F,$82,$7E,$24,$E5,$3B,$71,$1E,$08,$C5,$78,$F0,$F0,$D7,$9A,$7D
	.dc.b	$51,$8B,$F2,$DF,$1A,$EF,$7C,$5F,$01,$F0,$8F,$9E,$7C,$6C,$3D,$66
	.dc.b	$CF,$6C,$71,$2E,$1E,$09,$E7,$7E,$0C,$3C,$51,$EA,$6D,$F3,$AF,$73
	.dc.b	$67,$CB,$1C,$EF,$AC,$79,$30,$F5,$76,$7F,$82,$FC,$1F,$C1,$E9,$FE
	.dc.b	$65,$F4,$2E,$7F,$88,$AD,$3F,$E3,$68,$CE,$1F,$E3,$CE,$3F,$F2,$8C
	.dc.b	$FF,$37,$9F,$C6,$CF,$F6,$69,$CB,$FE,$9F,$A4,$62,$3F,$B5,$FC,$E7
	.dc.b	$FB,$CA,$FF,$2F,$87,$E5,$AD,$2F,$92,$F0,$C6,$3E,$16,$CE,$6F,$89
	.dc.b	$E9,$71,$87,$F8,$CB,$D2,$DE,$1C,$B2,$47,$2F,$7F,$1E,$9D,$C7,$FC
	.dc.b	$AA,$E1,$FE,$A8,$FB,$B0,$FF,$1F,$B5,$F8,$8D,$CF,$D3,$F0,$5F,$23
	.dc.b	$5F,$F1,$EF,$77,$A2,$31,$E3,$22,$9C,$E7,$78,$FF,$85,$23,$13,$B8
	.dc.b	$FF,$0B,$F6,$F2,$3E,$1F,$C8,$58,$FF,$15,$47,$BF,$D4,$C6,$9F,$E5
	.dc.b	$47,$E4,$FC,$57,$8B,$E4,$ED,$FF,$8C,$5C,$3F,$E2,$47,$C2,$47,$F8
	.dc.b	$B6,$BF,$5B,$E0,$BD,$7F,$2D,$F5,$1B,$3F,$E3,$B8,$FA,$1E,$CE,$32
	.dc.b	$FF,$32,$BF,$6B,$E1,$FE,$1F,$55,$BF,$F2,$13,$87,$FD,$0F,$ED,$5B
	.dc.b	$FF,$36,$BF,$94,$5F,$3A,$E4,$C2,$E1,$B1,$86,$5F,$F2,$FA,$E3,$1F
	.dc.b	$6D,$B1,$9A,$E4,$FF,$85,$78,$BF,$14,$7E,$4B,$E1,$AC,$7E,$4A,$BF
	.dc.b	$D4,$3D,$6F,$84,$FC,$17,$CE,$FE,$53,$A3,$F3,$D3,$9B,$FF,$32,$4E
	.dc.b	$5F,$D1,$B6,$FF,$61,$ED,$72,$EE,$FC,$D3,$D3,$B2,$5F,$87,$9B,$F5
	.dc.b	$3F,$0B,$8C,$3F,$C9,$78,$AF,$F5,$7D,$1F,$D1,$B0,$F0,$C6,$2B,$E6
	.dc.b	$61,$FD,$03,$4F,$DE,$AF,$9B,$CF,$87,$A6,$61,$F2,$E7,$4D,$8F,$99
	.dc.b	$18,$BF,$E0,$98,$F5,$AF,$E2,$0D,$8F,$EB,$3E,$1B,$E1,$7C,$F8,$FC
	.dc.b	$AE,$69,$87,$BF,$87,$4F,$CA,$8F,$E0,$FC,$5F,$DA,$63,$E2,$37,$1F
	.dc.b	$87,$3A,$F2,$77,$E4,$7D,$6E,$FB,$87,$D4,$70,$FC,$DD,$7D,$5C,$1F
	.dc.b	$CA,$B8,$95,$F3,$38,$7C,$3E,$F7,$E0,$3E,$26,$E3,$F7,$9C,$3F,$DF
	.dc.b	$F0,$60,$9F,$E0,$4D,$9D,$37,$5F,$E5,$9B,$3E,$77,$DE,$74,$8F,$F6
	.dc.b	$BC,$66,$BF,$1B,$F1,$3F,$92,$3E,$2F,$C0,$D8,$C9,$ED,$F7,$B1,$7F
	.dc.b	$EE,$0D,$FF,$D6,$73,$F2,$BD,$2F,$9B,$ED,$FC,$E6,$C7,$CE,$8F,$C6
	.dc.b	$FB,$F8,$BE,$4F,$D4,$F6,$3F,$1B,$EF,$DB,$FD,$45,$3B,$13,$A9,$FC
	.dc.b	$63,$F8,$A2,$3F,$C4,$B2,$DB,$F5,$63,$F0,$3E,$C7,$E7,$9F,$D8,$FC
	.dc.b	$3C,$9F,$C1,$F7,$5C,$7F,$D5,$FF,$3A,$EA,$FE,$64,$DA,$FA,$CD,$9F
	.dc.b	$EF,$E3,$9D,$19,$8F,$F0,$1C,$63,$E1,$BF,$5F,$EB,$FF,$43,$A4,$62
	.dc.b	$FF,$DC,$91,$F5,$EF,$AD,$F4,$E7,$F1,$92,$BC,$7C,$7E,$62,$38,$3C
	.dc.b	$FC,$7F,$BE,$F8,$2F,$03,$63,$F4,$7E,$2F,$F0,$2F,$82,$FD,$3B,$63
	.dc.b	$C7,$BF,$C2,$23,$F0,$9F,$EB,$DF,$3B,$FA,$28,$FD,$CF,$C1,$FF,$76
	.dc.b	$8C,$6B,$F6,$F9,$53,$9E,$8F,$E2,$3F,$2B,$70,$FD,$63,$F1,$7C,$CF
	.dc.b	$A2,$48,$FA,$2F,$83,$6F,$E3,$3F,$09,$DD,$B3,$FC,$C7,$89,$F3,$7E
	.dc.b	$2E,$B6,$8F,$9D,$FE,$E9,$39,$8C,$5F,$18,$69,$FD,$CD,$B1,$FB,$76
	.dc.b	$ED,$F5,$9A,$FE,$01,$74,$76,$3E,$26,$D6,$BB,$71,$FF,$32,$F6,$BC
	.dc.b	$6D,$8F,$14,$F8,$48,$F8,$5F,$26,$FE,$A3,$E8,$7C,$95,$9F,$E8,$5C
	.dc.b	$7F,$01,$BF,$09,$F5,$76,$FA,$CF,$C2,$DF,$DC,$49,$F7,$3F,$E1,$B9
	.dc.b	$5F,$F4,$83,$EF,$7B,$36,$BB,$FA,$FF,$33,$DC,$FC,$13,$63,$E6,$47
	.dc.b	$CB,$23,$F2,$DF,$EE,$1F,$3A,$AA,$49,$FE,$ED,$19,$6F,$C5,$7F,$80
	.dc.b	$7E,$81,$F3,$BF,$E0,$3A,$4E,$63,$EF,$E7,$FE,$4A,$F9,$AF,$F2,$87
	.dc.b	$37,$EE,$71,$FF,$28,$F3,$7C,$78,$C1,$FC,$4E,$77,$D3,$FE,$55,$6D
	.dc.b	$7F,$58,$D7,$A7,$C2,$FF,$6C,$7B,$7C,$F7,$7F,$66,$D7,$F7,$5B,$FE
	.dc.b	$E4,$D8,$FE,$E7,$F1,$7B,$8F,$E9,$CE,$1F,$E3,$CF,$D3,$23,$FE,$32
	.dc.b	$9F,$F1,$03,$D7,$DF,$F7,$7B,$1E,$F6,$47,$F5,$9D,$AE,$BF,$CF,$5F
	.dc.b	$F1,$57,$CF,$33,$FF,$64,$4F,$FB,$4E,$31,$ED,$FF,$89,$DB,$19,$9D
	.dc.b	$BF,$C9,$3E,$31,$E4,$3E,$7B,$FC,$5D,$18,$6C,$66,$7F,$DC,$13,$97
	.dc.b	$C1,$BF,$FB,$85,$BE,$E7,$86,$31,$1F,$CC,$B1,$FF,$0E,$4F,$F4,$3E
	.dc.b	$DF,$E6,$3A,$FF,$83,$3A,$AD,$FF,$13,$6F,$8B,$FC,$12,$AE,$E6,$92
	.dc.b	$BF,$C0,$1D,$3A,$EC,$FF,$0F,$51,$A4,$9F,$F3,$2F,$ED,$5F,$EA,$4E
	.dc.b	$9F,$F4,$E5,$6E,$7F,$C1,$55,$E2,$AD,$67,$FE,$B1,$6C,$FF,$63,$6E
	.dc.b	$46,$8E,$9E,$5F,$03,$CF,$92,$5F,$F9,$0E,$BF,$D5,$4D,$9D,$B4,$D7
	.dc.b	$0F,$C6,$D3,$FC,$61,$F0,$3E,$5E,$D9,$5F,$CB,$7E,$05,$3B,$6C,$66
	.dc.b	$30,$7B,$7A,$BE,$8C,$3E,$75,$F1,$1F,$E7,$18,$C3,$63,$31,$FE,$75
	.dc.b	$6B,$FD,$C0,$E5,$EF,$F4,$8C,$E5,$FE,$D2,$6F,$0F,$FB,$D1,$FF,$16
	.dc.b	$57,$F8,$3B,$EF,$B2,$83,$86,$E1,$E8,$E0,$7C,$55,$F8,$AF,$44,$1A
	.dc.b	$4A,$FF,$46,$43,$E2,$79,$12,$3F,$C3,$FF,$43,$97,$FA,$C7,$EE,$9F
	.dc.b	$9C,$D7,$FD,$EA,$DD,$7F,$DC,$AD,$9F,$C2,$78,$7F,$33,$F2,$9F,$DC
	.dc.b	$47,$FA,$6D,$AF,$C7,$6B,$FE,$18,$6B,$F8,$FF,$A3,$FC,$91,$1F,$C4
	.dc.b	$7C,$57,$E0,$ED,$F3,$54,$93,$F8,$DE,$0F,$17,$47,$DB,$C7,$FD,$1D
	.dc.b	$3E,$38,$C1,$FE,$AF,$D1,$83,$E2,$BB,$12,$BF,$31,$CB,$F8,$8F,$C1
	.dc.b	$BF,$85,$C7,$F9,$19,$7C,$B6,$4D,$79,$1B,$F7,$BF,$19,$26,$92,$BF
	.dc.b	$D8,$1F,$83,$7F,$80,$DF,$F2,$8A,$F5,$BF,$2E,$D8,$CF,$AB,$FB,$7F
	.dc.b	$92,$FD,$C7,$D0,$C9,$FD,$74,$70,$F6,$FF,$39,$7F,$EF,$D9,$EF,$18
	.dc.b	$3F,$D4,$DE,$2B,$E3,$7F,$81,$12,$7F,$71,$F2,$5F,$47,$E6,$7F,$83
	.dc.b	$7C,$84,$9F,$C3,$C7,$9B,$CE,$EE,$48,$F6,$BF,$3F,$95,$F1,$BF,$6C
	.dc.b	$93,$FC,$17,$E8,$7E,$F7,$CB,$FF,$13,$F2,$A4,$FE,$52,$3E,$1F,$A1
	.dc.b	$EC,$24,$7C,$A7,$B7,$91,$F8,$72,$47,$F1,$1F,$33,$F8,$0F,$E0,$8E
	.dc.b	$4A,$49,$FF,$40,$C3,$97,$7C,$5F,$C1,$63,$E3,$4F,$FB,$DD,$AF,$F0
	.dc.b	$36,$BF,$E7,$45,$C5,$CA,$3A,$CF,$8A,$3F,$98,$7C,$7F,$F4,$2F,$DB
	.dc.b	$3F,$59,$E9,$9B,$CC,$9D,$E1,$FC,$AD,$AF,$F8,$1A,$B8,$7C,$87,$3F
	.dc.b	$1A,$4F,$FB,$CF,$0F,$D4,$B7,$E4,$3C,$06,$9E,$51,$A6,$C7,$97,$C3
	.dc.b	$F3,$7C,$7C,$9E,$91,$E9,$F9,$4D,$46,$0F,$03,$ED,$F9,$0D,$46,$0F
	.dc.b	$23,$FC,$FF,$F7,$57,$CA,$FC,$3F,$1E,$57,$F9,$AE,$72,$FA,$FC,$7E
	.dc.b	$15,$23,$07,$BF,$C0,$F9,$9F,$2B,$23,$07,$C3,$EF,$FA,$04,$7E,$37
	.dc.b	$C9,$FE,$1B,$C0,$91,$F8,$47,$F1,$3E,$07,$E2,$BB,$7F,$08,$49,$FC
	.dc.b	$23,$F8,$7E,$7D,$47,$E3,$23,$F2,$31,$F7,$FC,$09,$3F,$55,$FC,$17
	.dc.b	$CD,$48,$C1,$0F,$AA,$FF,$5A,$F0,$9F,$1F,$7E,$4A,$57,$FB,$6A,$72
	.dc.b	$DF,$94,$E1,$52,$30,$7E,$17,$C4,$FC,$91,$1F,$F7,$1F,$CF,$FF,$84
	.dc.b	$DB,$F0,$9F,$D9,$E3,$FC,$CD,$1F,$D8,$C3,$87,$FB,$33,$5F,$EB,$C7
	.dc.b	$FB,$B7,$B7,$FA,$C3,$D0,$87,$E4,$BD,$CF,$C2,$79,$DF,$A5,$F6,$E4
	.dc.b	$FE,$0E,$3F,$51,$F2,$BA,$8C,$1F,$9F,$7F,$19,$F2,$3F,$AD,$F4,$A4
	.dc.b	$FE,$2E,$3F,$C6,$3F,$82,$7F,$B7,$BD,$DF,$EE,$5E,$BB,$C0,$FF,$47
	.dc.b	$DE,$CC,$CF,$F7,$7F,$25,$7C,$67,$FD,$E5,$5F,$EB,$E7,$FC,$0D,$7C
	.dc.b	$FE,$1B,$FA,$84,$E6,$7F,$DD,$AD,$7E,$2D,$F4,$F8,$CF,$D2,$54,$9F
	.dc.b	$08,$9D,$3F,$A5,$E2,$F5,$FC,$D4,$8C,$10,$FF,$45,$E3,$ED,$9F,$D8
	.dc.b	$1F,$0B,$C3,$FF,$18,$7E,$EC,$D7,$FB,$33,$6F,$D6,$D7,$ED,$8B,$46
	.dc.b	$63,$D7,$AF,$FB,$69,$FF,$72,$47,$F9,$32,$77,$5F,$4B,$E6,$3E,$9F
	.dc.b	$CD,$7E,$2F,$4F,$9A,$DC,$4D,$31,$F3,$3C,$32,$F8,$B5,$C1,$B5,$FE
	.dc.b	$E2,$6F,$6F,$E0,$3A,$E4,$60,$87,$F2,$B9,$4F,$E1,$8F,$C3,$F6,$7E
	.dc.b	$23,$E6,$7C,$84,$AF,$F2,$26,$1F,$84,$8F,$47,$46,$E7,$C7,$1F,$90
	.dc.b	$8F,$F2,$9F,$C8,$FF,$82,$B7,$FF,$21,$7C,$FF,$9E,$D2,$F3,$7F,$53
	.dc.b	$39,$8C,$4F,$0D,$BF,$5B,$1C,$E1,$3B,$F3,$7E,$B4,$3D,$3E,$58,$7E
	.dc.b	$D1,$B1,$FE,$53,$78,$63,$71,$EB,$7C,$B0,$DD,$F6,$9F,$0E,$1E,$67
	.dc.b	$77,$E5,$9F,$F2,$F6,$3E,$5B,$E2,$4A,$DC,$F3,$37,$87,$A3,$1E,$E4
	.dc.b	$9D,$BE,$69,$E0,$7C,$8C,$7C,$4F,$81,$AF,$CC,$AF,$55,$78,$2F,$C5
	.dc.b	$3B,$3D,$06,$F1,$B8,$FA,$6B,$DF,$38,$C3,$AC,$E6,$3D,$35,$F2,$5E
	.dc.b	$87,$D1,$AE,$2D,$36,$53,$C0,$7B,$27,$EF,$B6,$BB,$E1,$EA,$8E,$38
	.dc.b	$FC,$57,$BB,$F1,$49,$7E,$07,$51,$D9,$F9,$5F,$23,$F8,$58,$F1,$B5
	.dc.b	$F1,$B7,$0F,$C7,$39,$8F,$F4,$D7,$D0,$7F,$A6,$E3,$10,$8C,$57,$FA
	.dc.b	$A5,$B1,$FC,$29,$AF,$F4,$67,$E4,$3F,$EA,$5E,$BF,$E2,$53,$FA,$58
	.dc.b	$EE,$7C,$71,$1A,$C3,$B4,$66,$30,$F7,$31,$CD,$D9,$E7,$AF,$59,$FE
	.dc.b	$40,$F7,$8F,$51,$F8,$E3,$BC,$FE,$19,$F1,$DF,$18,$E9,$1C,$23,$0F
	.dc.b	$9C,$E7,$38,$3B,$61,$CE,$FC,$38,$7A,$A7,$4D,$3F,$98,$C6,$AB,$D9
	.dc.b	$3A,$6C,$EB,$0F,$34,$78,$E3,$0F,$8E,$FE,$15,$FC,$68,$FA,$47,$BD
	.dc.b	$EC,$E2,$FB,$36,$B9,$C7,$AA,$38,$5F,$82,$FA,$BF,$11,$B1,$E1,$36
	.dc.b	$7C,$D3,$E2,$D7,$43,$FB,$87,$1B,$E0,$7F,$AA,$6F,$FD,$57,$8B,$FB
	.dc.b	$9F,$C0,$63,$2D,$D5,$C7,$FB,$46,$0A,$C5,$F0,$BF,$E7,$33,$96,$CF
	.dc.b	$F5,$BA,$E2,$DA,$FF,$00,$F7,$F2,$B3,$7E,$5F,$0C,$C3,$FA,$9F,$CB
	.dc.b	$DC,$7F,$A7,$3C,$7C,$0E,$5F,$29,$6B,$FA,$CB,$87,$CA,$F8,$71,$34
	.dc.b	$C7,$81,$EF,$DB,$FB,$9F,$C9,$F9,$63,$1E,$02,$29,$ED,$9D,$E1,$FD
	.dc.b	$FA,$31,$3B,$8F,$F0,$0F,$C4,$72,$3E,$07,$C2,$58,$FE,$FB,$1F,$0F
	.dc.b	$E1,$31,$B7,$F6,$B7,$E7,$F9,$5F,$E0,$DA,$FF,$01,$FC,$6E,$1F,$2B
	.dc.b	$1F,$F0,$7B,$E2,$23,$FD,$BA,$1F,$29,$EC,$F8,$DE,$AD,$9F,$F0,$7C
	.dc.b	$78,$34,$C6,$3F,$EA,$7B,$F6,$37,$E4,$3E,$2F,$C4,$79,$9B,$87,$F8
	.dc.b	$62,$3D,$3A,$62,$B9,$3E,$CF,$94,$FC,$74,$3E,$7D,$F2,$39,$C3,$FC
	.dc.b	$73,$39,$8D,$CE,$A7,$FC,$D4,$D9,$FF,$16,$9B,$8C,$CF,$E4,$72,$7C
	.dc.b	$AF,$CC,$39,$EB,$6B,$BF,$ED,$4F,$13,$CB,$EF,$FF,$B2,$E7,$31,$8F
	.dc.b	$72,$47,$0F,$1A,$E9,$FE,$3D,$8F,$F3,$CE,$1F,$E0,$97,$FC,$5F,$9B
	.dc.b	$93,$EE,$EB,$F7,$9B,$3A,$8B,$7F,$46,$6F,$47,$F9,$2F,$C0,$79,$9F
	.dc.b	$29,$79,$96,$30,$CC,$7F,$C8,$CF,$ED,$70,$FF,$32,$36,$B8,$CF,$F9
	.dc.b	$2F,$C4,$E8,$75,$F3,$2C,$7A,$B8,$BF,$2A,$7E,$5B,$E4,$EC,$7B,$BE
	.dc.b	$4B,$F3,$3D,$3F,$9A,$6C,$7C,$2F,$8F,$FD,$6F,$27,$E7,$9B,$1F,$E6
	.dc.b	$08,$FD,$0F,$C7,$63,$6F,$C9,$3F,$A4,$F8,$EF,$D2,$FC,$85,$BF,$9A
	.dc.b	$E1,$F9,$FF,$0B,$6F,$F3,$4F,$EB,$0F,$D6,$FC,$2D,$9F,$AA,$FE,$C3
	.dc.b	$E1,$7F,$63,$95,$9F,$B2,$E1,$A7,$DC,$8C,$5F,$E0,$4F,$F4,$87,$ED
	.dc.b	$FF,$8B,$9F,$F3,$7B,$63,$5C,$68,$FC,$34,$61,$D7,$37,$F3,$76,$D7
	.dc.b	$FA,$AD,$7C,$7F,$F3,$47,$35,$E0,$8F,$C2,$79,$DF,$85,$F3,$ED,$FE
	.dc.b	$B3,$14,$66,$3F,$D1,$4F,$E0,$C8,$FF,$A3,$7C,$EF,$C7,$1F,$8F,$F3
	.dc.b	$EC,$7F,$AB,$B3,$79,$6F,$5B,$F9,$63,$67,$E0,$FC,$7E,$27,$F8,$0F
	.dc.b	$E2,$BF,$B9,$DB,$FD,$4A,$D7,$FD,$C2,$D4,$1C,$61,$E0,$F8,$CF,$85
	.dc.b	$F1,$FE,$06,$E3,$E5,$C9,$F1,$9E,$43,$E4,$7C,$85,$9F,$E1,$F8,$BB
	.dc.b	$FF,$4C,$D8,$FF,$7C,$37,$3F,$E1,$ED,$7F,$DE,$F9,$FC,$2F,$8D,$FA
	.dc.b	$9F,$21,$F0,$9B,$7F,$82,$F7,$FE,$C1,$F6,$77,$B1,$E2,$EC,$F0,$91
	.dc.b	$F0,$BF,$DA,$FA,$1F,$21,$F8,$1E,$1F,$C1,$36,$BD,$7D,$BF,$3B,$B7
	.dc.b	$E7,$B6,$3F,$96,$9C,$8F,$E5,$3B,$7F,$B6,$EB,$F8,$34,$66,$3F,$96
	.dc.b	$B5,$FF,$6D,$AF,$88,$FE,$77,$F8,$AF,$D6,$8E,$B1,$89,$FC,$1B,$23
	.dc.b	$F4,$7F,$23,$67,$F5,$07,$FA,$BF,$D6,$FE,$2D,$E8,$1A,$4A,$FE,$7F
	.dc.b	$C6,$FD,$DF,$FC,$9B,$7F,$EF,$86,$C7,$EA,$7A,$3F,$BF,$46,$2B,$F5
	.dc.b	$5E,$9F,$EA,$4F,$D5,$72,$D8,$FE,$A3,$93,$D7,$F1,$92,$3E,$3C,$65
	.dc.b	$F1,$DB,$E2,$3B,$9F,$B5,$F9,$2B,$3F,$D7,$3F,$0B,$6F,$FD,$89,$B1
	.dc.b	$FD,$83,$E9,$7F,$C0,$1B,$FC,$49,$CB,$76,$FE,$75,$1F,$9C,$AF,$B2
	.dc.b	$D9,$FE,$C6,$D7,$F5,$27,$EA,$BE,$8D,$E2,$7F,$72,$6D,$79,$78,$31
	.dc.b	$3F,$DD,$BE,$8B,$FC,$53,$3F,$DD,$17,$C1,$67,$E0,$F4,$C9,$FC,$5C
	.dc.b	$E9,$B1,$FD,$46,$31,$8F,$90,$72,$7D,$BF,$8D,$FA,$E7,$F1,$DF,$95
	.dc.b	$B8,$7F,$0F,$E3,$4F,$F2,$1F,$73,$38,$7F,$29,$7F,$92,$7C,$9C,$BF
	.dc.b	$E6,$3F,$27,$6B,$F9,$74,$FF,$29,$EB,$C6,$7F,$CB,$DF,$F2,$FF,$EE
	.dc.b	$3F,$9B,$AF,$E6,$ED,$9F,$F3,$53,$F8,$B8,$79,$E8,$D6,$3F,$C6,$23
	.dc.b	$13,$D3,$E4,$FA,$B6,$7F,$8E,$DF,$49,$FF,$34,$59,$A4,$9F,$F0,$FD
	.dc.b	$F7,$F9,$3F,$39,$E8,$EB,$B3,$D0,$F1,$5A,$7D,$BB,$6B,$2F,$6B,$FE
	.dc.b	$34,$FA,$0F,$D8,$7C,$8F,$F0,$46,$CF,$F8,$B9,$AF,$D9,$6B,$F8,$EF
	.dc.b	$92,$FE,$2B,$F4,$3F,$C3,$1B,$1F,$E3,$66,$BF,$81,$6B,$F9,$69,$FD
	.dc.b	$87,$C3,$F8,$7C,$2F,$F1,$A4,$F8,$5A,$F8,$5A,$FC,$FE,$AF,$CA,$BF
	.dc.b	$19,$AF,$E3,$F7,$FD,$0F,$4F,$F5,$D3,$FA,$F7,$57,$F3,$AE,$0E,$51
	.dc.b	$6B,$EE,$DF,$2E,$3F,$F5,$53,$FE,$8C,$8F,$9C,$7E,$0B,$87,$FD,$C4
	.dc.b	$DF,$91,$FE,$5B,$97,$E1,$5B,$3A,$8F,$D9,$FC,$77,$F6,$F2,$57,$A5
	.dc.b	$E3,$F3,$4C,$FC,$95,$C2,$BC,$2D,$7F,$C1,$B3,$B7,$2F,$F2,$B6,$F8
	.dc.b	$FC,$A3,$11,$C3,$E3,$B0,$66,$64,$72,$8E,$1F,$1D,$23,$AB,$F1,$F5
	.dc.b	$7B,$CF,$3B,$0D,$AD,$90,$FF,$14,$4F,$09,$FA,$10,$FA,$1F,$EE,$8F
	.dc.b	$F5,$66,$BF,$E8,$06,$BF,$C2,$DA,$FF,$20,$5F,$0D,$FD,$7A,$3A,$36
	.dc.b	$33,$1D,$63,$EA,$FC,$B7,$CE,$E0,$34,$F2,$FD,$0F,$0A,$FE,$B3,$AB
	.dc.b	$C5,$EC,$D8,$FE,$DA,$D7,$FD,$2E,$D7,$F9,$12,$F0,$BE,$0A,$F1,$29
	.dc.b	$88,$F2,$7C,$C4,$AF,$C0,$E4,$8F,$A5,$F3,$9F,$A6,$DF,$F8,$BF,$C7
	.dc.b	$7E,$40,$97,$FC,$2A,$FF,$D1,$6F,$F0,$CE,$4F,$F5,$9B,$D3,$E3,$BE
	.dc.b	$11,$F8,$7F,$90,$F8,$9F,$1D,$27,$C2,$F8,$3E,$23,$E4,$7C,$7F,$E7
	.dc.b	$89,$3D,$DF,$F0,$7E,$6D,$1F,$0B,$F3,$FE,$5E,$5F,$F9,$36,$72,$FF
	.dc.b	$B1,$3E,$7F,$FC,$9C,$FF,$86,$1A,$FF,$B0,$9A,$FF,$4D,$6B,$FD,$89
	.dc.b	$79,$FF,$83,$6C,$73,$8C,$1F,$87,$F9,$F6,$94,$93,$FE,$28,$8C,$BF
	.dc.b	$27,$E4,$3F,$11,$F1,$B2,$3F,$63,$1F,$8B,$F8,$DF,$E0,$24,$8F,$A4
	.dc.b	$FE,$FB,$E4,$3F,$21,$F1,$D2,$7F,$73,$1F,$B7,$F8,$6F,$CA,$79,$32
	.dc.b	$7F,$7B,$1F,$97,$F2,$FF,$86,$92,$3E,$F3,$F4,$37,$FF,$31,$C7,$E2
	.dc.b	$3C,$BF,$CF,$1F,$88,$F7,$3F,$82,$78,$BF,$A1,$F8,$D9,$5F,$C2,$A3
	.dc.b	$F4,$9E,$67,$F1,$A2,$47,$E2,$DE,$DE,$1F,$F1,$6F,$DC,$BF,$89,$4F
	.dc.b	$FA,$81,$BD,$3F,$C5,$4F,$C1,$4F,$F0,$0F,$A0,$7C,$9D,$25,$FF,$2E
	.dc.b	$10,$F9,$4E,$1C,$BF,$BD,$83,$49,$3F,$E5,$E8,$71,$EC,$91,$97,$F9
	.dc.b	$BF,$CF,$FE,$A7,$D1,$FE,$F5,$A3,$E6,$BD,$87,$F3,$29,$D5,$FF,$80
	.dc.b	$BC,$47,$CA,$E7,$D6,$93,$FC,$CA,$38,$78,$9F,$17,$D3,$93,$FC,$EB
	.dc.b	$0F,$F5,$EB,$67,$6D,$35,$F2,$BE,$1F,$95,$91,$C5,$F2,$78,$D2,$3F
	.dc.b	$C0,$F7,$FA,$A7,$FB,$13,$5F,$F2,$2F,$AC,$8C,$47,$F8,$65,$7C,$9F
	.dc.b	$38,$FF,$6C,$C7,$3E,$7F,$E9,$D5,$C6,$BE,$27,$C9,$7E,$07,$D2,$98
	.dc.b	$7B,$63,$FA,$37,$CF,$FF,$84,$E3,$FC,$F6,$D7,$FC,$89,$82,$31,$1F
	.dc.b	$EA,$E6,$BE,$5F,$8B,$FC,$77,$B9,$23,$E5,$47,$D5,$F9,$29,$1F,$35
	.dc.b	$FB,$1E,$0F,$F4,$EF,$8A,$FF,$3C,$4F,$E3,$5A,$FF,$A5,$1B,$8F,$E6
	.dc.b	$BA,$DF,$25,$F9,$7F,$0B,$F2,$04,$BF,$C7,$3F,$93,$F7,$64,$7E,$45
	.dc.b	$AF,$D2,$F9,$0F,$D3,$7B,$F2,$3F,$05,$1F,$8D,$23,$83,$F0,$DF,$1F
	.dc.b	$FD,$9B,$C9,$FE,$AB,$1F,$E0,$2F,$01,$F3,$5F,$B1,$35,$EF,$EF,$DC
	.dc.b	$FE,$C5,$7F,$D1,$EF,$FD,$82,$D7,$FC,$E8,$7F,$35,$8F,$F0,$CD,$BE
	.dc.b	$67,$F7,$84,$9F,$CC,$79,$5F,$DE,$3E,$83,$FC,$49,$1C,$9A,$FF,$9F
	.dc.b	$DA,$FF,$A5,$1A,$FF,$BF,$9F,$F1,$BB,$BF,$3B,$FB,$C4,$62,$38,$47
	.dc.b	$ED,$BD,$FF,$E1,$DE,$E4,$9F,$D5,$47,$E3,$BD,$8F,$E9,$FD,$9F,$CD
	.dc.b	$A7,$F1,$AD,$7F,$19,$1F,$C0,$3E,$37,$F8,$C7,$CC,$C9,$FD,$A4,$7E
	.dc.b	$63,$D9,$E6,$48,$FC,$CB,$F1,$BE,$12,$47,$F9,$DB,$7F,$F0,$33,$FA
	.dc.b	$E6,$BF,$E8,$0C,$7F,$D6,$6F,$E8,$BD,$8F,$39,$1F,$F3,$FF,$CF,$7F
	.dc.b	$A0,$DF,$F1,$FB,$5F,$F5,$4F,$A3,$FE,$8F,$8C,$77,$7F,$2D,$BF,$F2
	.dc.b	$C4,$F7,$E0,$FF,$24,$D6,$A3,$87,$C7,$FE,$D2,$7F,$9C,$7A,$DF,$80
	.dc.b	$26,$3E,$B7,$FB,$CF,$CF,$7F,$3C,$8F,$F6,$6E,$48,$C4,$7F,$A7,$1A
	.dc.b	$FF,$AF,$9B,$E5,$35,$FE,$75,$3F,$1F,$D7,$FC,$79,$1F,$83,$F1,$DF
	.dc.b	$E6,$AD,$7F,$D8,$31,$82,$1F,$E7,$D7,$F3,$58,$FF,$14,$6E,$BF,$17
	.dc.b	$AB,$E0,$F6,$4B,$F9,$71,$F1,$0C,$3F,$D5,$FD,$8F,$D0,$91,$FB,$8F
	.dc.b	$8F,$E1,$24,$7D,$66,$BF,$E9,$FF,$33,$FD,$BA,$BE,$81,$F1,$7F,$EA
	.dc.b	$7F,$9B,$FF,$7A,$C6,$24,$3E,$39,$3A,$AF,$F5,$75,$F4,$BE,$17,$B8
	.dc.b	$EF,$F3,$18,$C3,$8C,$F4,$8C,$61,$DA,$11,$BA,$E3,$87,$6D,$F8,$3E
	.dc.b	$7A,$A7,$9B,$0E,$AF,$28,$E0,$D3,$55,$E7,$8E,$51,$F0,$DB,$8F,$86
	.dc.b	$FE,$0B,$67,$91,$F0,$F1,$F4,$63,$EA,$8F,$19,$F0,$6A,$E2,$C7,$C1
	.dc.b	$3E,$7B,$F9,$4D,$3B,$E3,$E8,$C7,$C0,$75,$8F,$31,$B2,$9D,$CB,$7D
	.dc.b	$F0,$E7,$1B,$9E,$47,$4A,$F8,$B8,$F5,$23,$83,$C4,$7C,$42,$4F,$CA
	.dc.b	$AE,$AF,$86,$78,$BE,$57,$E4,$12,$7E,$7B,$A8,$E1,$1F,$5B,$DB,$FC
	.dc.b	$0C,$75,$0E,$1F,$73,$70,$F7,$46,$AB,$FA,$3F,$D0,$7E,$62,$B8,$4F
	.dc.b	$E6,$A3,$13,$A8,$E0,$7E,$44,$46,$9C,$3F,$71,$7C,$97,$FB,$6E,$7F
	.dc.b	$3F,$1D,$4F,$C2,$9A,$E1,$E0,$F4,$33,$3C,$71,$F1,$57,$28,$C4,$F8
	.dc.b	$0F,$BA,$73,$29,$CF,$3E,$17,$F4,$A3,$C5,$58,$7C,$67,$85,$C1,$E8
	.dc.b	$79,$5F,$C1,$BF,$F7,$B5,$E7,$0E,$91,$89,$F9,$13,$D4,$D5,$79,$A7
	.dc.b	$4F,$A6,$35,$3C,$0E,$D8,$79,$E7,$C9,$18,$7C,$F7,$F3,$EF,$E1,$C7
	.dc.b	$F0,$58,$F1,$D7,$16,$C7,$C4,$78,$47,$D0,$6C,$74,$3E,$8C,$65,$A6
	.dc.b	$6F,$CE,$FA,$6F,$C9,$1F,$C8,$F9,$9C,$6F,$8A,$FE,$9B,$18,$6C,$67
	.dc.b	$0F,$EC,$F1,$AE,$7F,$F3,$FC,$7F,$73,$F8,$1C,$11,$FE,$DB,$E0,$7F
	.dc.b	$A4,$DF,$ED,$5F,$0D,$74,$65,$E2,$AC,$3E,$49,$B9,$1A,$62,$6D,$FF
	.dc.b	$78,$F8,$5F,$39,$1F,$EF,$AF,$9F,$E7,$A4,$7F,$82,$5B,$8F,$F9,$93
	.dc.b	$B5,$F0,$3F,$DB,$B9,$BF,$B8,$36,$33,$7F,$E3,$A9,$CB,$FE,$13,$F8
	.dc.b	$7F,$88,$47,$06,$E5,$ED,$FB,$3E,$1B,$E1,$5F,$F7,$D8,$C5,$7F,$84
	.dc.b	$A6,$E3,$FE,$12,$3F,$BF,$4E,$2B,$FD,$7D,$7F,$E4,$9B,$ED,$E7,$F9
	.dc.b	$C8,$FF,$89,$FD,$7F,$B0,$47,$07,$97,$CB,$F7,$11,$C1,$FE,$58,$E2
	.dc.b	$7C,$7F,$C2,$CF,$FD,$A1,$EC,$7D,$88,$C6,$7F,$E9,$8E,$64,$FB,$B4
	.dc.b	$FF,$66,$63,$F0,$FC,$6F,$98,$47,$B6,$1F,$19,$FF,$38,$7B,$1D,$77
	.dc.b	$FC,$2E,$28,$FC,$56,$6F,$A0,$7A,$5F,$E2,$46,$E1,$FE,$70,$9D,$4E
	.dc.b	$4D,$6F,$27,$54,$8C,$57,$F2,$BD,$FF,$C5,$9F,$0D,$E2,$3E,$07,$AF
	.dc.b	$70,$FF,$75,$7C,$CF,$FB,$59,$F0,$36,$FF,$D0,$2D,$7F,$9B,$35,$FF
	.dc.b	$04,$7D,$07,$F8,$22,$7F,$C2,$FC,$7F,$EC,$2E,$4F,$E9,$31,$FE,$F7
	.dc.b	$9D,$57,$E1,$72,$FF,$54,$F8,$BF,$4F,$69,$7C,$8F,$F1,$0C,$FD,$AE
	.dc.b	$6C,$DF,$1E,$3F,$8E,$F3,$CB,$AE,$7F,$ED,$29,$CB,$E3,$F1,$7E,$8E
	.dc.b	$0F,$27,$FC,$A2,$4C,$BE,$84,$65,$BE,$AF,$F9,$58,$8E,$0F,$AD,$F5
	.dc.b	$7F,$E3,$67,$D1,$9F,$F8,$E2,$30,$FE,$8C,$FE,$13,$F2,$3F,$82,$8C
	.dc.b	$E5,$E7,$F4,$7F,$28,$7E,$57,$96,$E1,$D3,$2F,$CD,$F5,$7F,$10,$DD
	.dc.b	$F1,$5F,$9B,$8C,$3F,$99,$E7,$FE,$04,$4B,$FD,$E3,$FC,$A3,$87,$F9
	.dc.b	$1E,$DF,$27,$E2,$BF,$49,$E3,$F5,$7F,$4D,$E0,$FE,$9F,$3B,$8F,$8F
	.dc.b	$0F,$F7,$0B,$63,$A1,$1F,$D2,$BD,$FC,$0F,$D8,$1C,$E3,$93,$FB,$3C
	.dc.b	$A4,$75,$ED,$B9,$FD,$F9,$F8,$1F,$13,$8E,$2F,$C7,$3F,$CC,$BE,$83
	.dc.b	$F1,$F5,$F8,$D7,$E2,$E1,$F2,$5C,$D2,$3F,$1B,$1A,$9E,$1B,$FC,$DF
	.dc.b	$8B,$E8,$DA,$FB,$B3,$FD,$0B,$E3,$FF,$D3,$24,$8F,$C4,$BF,$E9,$CF
	.dc.b	$2A,$47,$E3,$1B,$3F,$C5,$7C,$9F,$C9,$9F,$94,$F6,$ED,$7E,$2B,$37
	.dc.b	$C7,$FF,$96,$63,$F0,$5C,$2F,$8C,$FE,$67,$39,$C7,$F4,$51,$F8,$5F
	.dc.b	$57,$FB,$2D,$7F,$A2,$2F,$F9,$23,$6B,$F8,$33,$5F,$F4,$17,$4B,$81
	.dc.b	$C3,$F8,$FF,$5B,$E7,$37,$0F,$36,$6F,$89,$FC,$A7,$D0,$79,$1B,$3E
	.dc.b	$8F,$8B,$F1,$91,$F1,$BF,$DE,$B1,$F6,$47,$BB,$E3,$3E,$E6,$FF,$9F
	.dc.b	$7E,$F6,$FF,$A1,$C2,$D7,$C4,$7F,$49,$E5,$7C,$02,$47,$C0,$7B,$63
	.dc.b	$FA,$9F,$89,$FC,$43,$67,$EE,$FD,$0F,$F4,$27,$F9,$93,$78,$7F,$CA
	.dc.b	$C1,$C3,$FE,$59,$6B,$FE,$AB,$84,$FF,$B8,$1A,$7F,$3D,$2C,$74,$6B
	.dc.b	$FE,$63,$FA,$1F,$E1,$2D,$3F,$DD,$2E,$1F,$F5,$C3,$57,$27,$E2,$3E
	.dc.b	$1B,$F8,$37,$C5,$FE,$19,$B7,$F8,$AF,$8A,$FC,$29,$1F,$C2,$BF,$88
	.dc.b	$F0,$7F,$87,$1F,$E0,$0F,$9E,$FF,$31,$61,$FC,$F6,$3F,$61,$B7,$F1
	.dc.b	$3F,$8A,$B8,$7E,$59,$FE,$2D,$F2,$BF,$92,$24,$7E,$49,$FC,$AE,$5F
	.dc.b	$C6,$CE,$DF,$3D,$67,$F5,$FF,$40,$E5,$FE,$D2,$FE,$3E,$33,$18,$E4
	.dc.b	$CE,$5F,$BD,$E0,$FA,$F6,$8C,$47,$FA,$3E,$31,$7F,$DB,$DB,$1F,$DA
	.dc.b	$7F,$09,$FE,$6B,$3F,$D4,$9B,$1F,$EF,$4F,$93,$78,$1F,$E4,$29,$CC
	.dc.b	$BE,$85,$F1,$BE,$08,$FF,$77,$46,$30,$FE,$FC,$D8,$FC,$6B,$5F,$F5
	.dc.b	$23,$FC,$0A,$37,$DD,$CE,$47,$77,$0D,$8F,$F2,$07,$23,$E3,$7F,$93
	.dc.b	$72,$7D,$B6,$CF,$F2,$58,$B5,$FE,$39,$E5,$5F,$1A,$D1,$FC,$AF,$97
	.dc.b	$F9,$87,$27,$F3,$16,$C7,$F2,$F8,$B7,$FC,$CE,$6D,$FD,$5F,$09,$F3
	.dc.b	$7F,$31,$BF,$1F,$FE,$6C,$8C,$07,$07,$FA,$01,$C6,$3F,$E3,$F6,$E1
	.dc.b	$FE,$21,$6B,$FC,$73,$83,$FD,$74,$DF,$8B,$FF,$44,$C6,$23,$EF,$47
	.dc.b	$5F,$A1,$F1,$6F,$E3,$6E,$1F,$DC,$F0,$F8,$97,$DE,$BC,$84,$9F,$F1
	.dc.b	$D7,$D0,$3E,$63,$FD,$B3,$5F,$C2,$A3,$18,$EA,$73,$3F,$EE,$39,$CB
	.dc.b	$FC,$43,$E3,$DE,$3B,$E8,$7F,$C8,$D1,$86,$C6,$67,$83,$5E,$9F,$0D
	.dc.b	$FE,$FC,$7C,$4D,$73,$F0,$5E,$88,$C4,$7E,$87,$83,$F1,$EF,$F8,$E7
	.dc.b	$81,$C1,$8E,$2F,$B9,$18,$9F,$F0,$C3,$F6,$5B,$5B,$DF,$A4,$60,$DF
	.dc.b	$81,$97,$FD,$5F,$D6,$F3,$BF,$C6,$AD,$8F,$F3,$F3,$5F,$F3,$03,$5F
	.dc.b	$EB,$6F,$F6,$A7,$FC,$54,$EF,$1F,$C9,$36,$FF,$C9,$EE,$E7,$FC,$AC
	.dc.b	$D9,$DF,$57,$97,$9B,$F8,$0B,$73,$FD,$75,$FC,$0F,$8F,$C7,$C9,$C8
	.dc.b	$C1,$0F,$95,$1C,$4E,$5F,$27,$CF,$E7,$FE,$16,$4F,$DE,$3F,$D0,$F7
	.dc.b	$FE,$05,$6B,$F4,$E7,$83,$F4,$1A,$FD,$7F,$84,$F9,$8B,$F0,$AF,$E2
	.dc.b	$AF,$FA,$3D,$AF,$F5,$E6,$BF,$80,$38,$78,$7F,$E9,$57,$FD,$E7,$D9
	.dc.b	$D2,$78,$E3,$FC,$B9,$BC,$3F,$88,$F9,$2F,$CE,$CF,$F3,$2F,$A3,$7C
	.dc.b	$CF,$36,$E3,$FE,$A1,$6B,$FD,$C1,$AF,$8D,$7C,$2F,$12,$47,$17,$CD
	.dc.b	$EC,$BC,$5F,$82,$D9,$FF,$1D,$B5,$FF,$3B,$72,$38,$3E,$E7,$C4,$49
	.dc.b	$FC,$DB,$F7,$FE,$23,$F0,$1F,$2B,$FE,$88,$B3,$49,$5F,$ED,$38,$7C
	.dc.b	$AF,$E1,$C9,$1F,$A6,$F7,$FF,$15,$EF,$C8,$FD,$4B,$F8,$DE,$DF,$C7
	.dc.b	$77,$77,$24,$F5,$7C,$5C,$FF,$93,$24,$7E,$CB,$B7,$F2,$DF,$27,$F9
	.dc.b	$9F,$DD,$3F,$B1,$F3,$7F,$33,$6F,$47,$F5,$F9,$FC,$3F,$C1,$BE,$5B
	.dc.b	$F0,$0D,$C7,$FD,$AC,$D7,$FC,$4A,$D7,$FC,$97,$6F,$A1,$FC,$8F,$D1
	.dc.b	$FF,$0F,$CB,$FB,$9B,$72,$FF,$5F,$D6,$EF,$FD,$40,$BE,$63,$D1,$E7
	.dc.b	$C9,$CC,$7E,$73,$E8,$FF,$29,$06,$91,$FF,$5E,$D3,$CB,$F8,$44,$8E
	.dc.b	$FF,$41,$F1,$7E,$5F,$E3,$7C,$AC,$9F,$1C,$7C,$9F,$96,$E6,$48,$FF
	.dc.b	$22,$F9,$8C,$7F,$6B,$1F,$0B,$E5,$3E,$81,$1F,$6F,$8B,$F4,$FC,$4F
	.dc.b	$A9,$F2,$92,$7E,$0C,$7D,$7F,$9D,$F4,$92,$3B,$E0,$8F,$6F,$CA,$FD
	.dc.b	$D2,$4F,$CA,$E8,$FE,$03,$F4,$3F,$AA,$8F,$E7,$6D,$7F,$C1,$6D,$4E
	.dc.b	$74,$F6,$7C,$AF,$CA,$24,$F8,$9F,$9B,$E4,$7E,$35,$91,$F6,$A3,$F2
	.dc.b	$1F,$2B,$F9,$1F,$95,$93,$F7,$A3,$2D,$21,$FE,$5C,$E9,$FC,$2F,$CA
	.dc.b	$FF,$31,$FD,$D7,$F1,$12,$C9,$18,$6F,$B5,$BF,$99,$95,$FE,$EB,$A4
	.dc.b	$62,$3B,$1C,$E3,$FD,$DA,$EF,$E6,$3C,$04,$79,$7C,$B7,$8C,$FF,$42
	.dc.b	$BC,$5B,$87,$91,$B5,$FE,$AC,$9F,$C3,$AF,$0F,$94,$74,$3C,$CF,$BF
	.dc.b	$E5,$64,$7A,$5F,$85,$F2,$B2,$3D,$8D,$7D,$78,$7C,$4F,$97,$F9,$1F
	.dc.b	$03,$27,$E1,$3F,$2F,$E5,$3E,$67,$C8,$59,$F8,$D1,$87,$FD,$C3,$CC
	.dc.b	$E8,$FE,$84,$D8,$F8,$2D,$7F,$D4,$ED,$7F,$A5,$3F,$C1,$5F,$C8,$F4
	.dc.b	$72,$23,$CB,$A3,$F2,$F1,$23,$C3,$1F,$DC,$7E,$8D,$C5,$49,$3F,$77
	.dc.b	$E4,$7F,$0C,$47,$07,$E0,$7E,$47,$96,$B2,$3F,$CD,$7E,$5F,$F9,$57
	.dc.b	$0F,$F1,$7C,$7E,$1F,$E5,$FF,$20,$47,$07,$E2,$FE,$57,$F5,$BD,$BF
	.dc.b	$10,$91,$F1,$1F,$8F,$D7,$23,$E3,$B5,$FF,$66,$FD,$97,$F8,$29,$FC
	.dc.b	$99,$FB,$28,$FD,$C9,$FD,$EA,$35,$8F,$FA,$73,$E9,$3F,$AC,$57,$AB
	.dc.b	$BB,$FC,$6F,$18,$8E,$6D,$9F,$F3,$97,$5B,$E5,$FF,$58,$4A,$FD,$33
	.dc.b	$FA,$FF,$93,$FA,$1F,$2F,$27,$B1,$FB,$1D,$D1,$FB,$5F,$93,$93,$FA
	.dc.b	$F6,$BF,$94,$C3,$6F,$EC,$FE,$0B,$F4,$A4,$7F,$4A,$FE,$AB,$D9,$FF
	.dc.b	$09,$67,$E8,$8F,$E0,$47,$F8,$CE,$33,$AF,$FA,$93,$BB,$FC,$91,$F0
	.dc.b	$DF,$E2,$AF,$DE,$5C,$94,$E1,$13,$2F,$5D,$70,$6E,$1F,$E1,$8C,$3A
	.dc.b	$3F,$E5,$B8,$FD,$6B,$FE,$5F,$7F,$C4,$D3,$FE,$57,$F8,$FF,$41,$23
	.dc.b	$FA,$2C,$BE,$3B,$D1,$32,$7B,$C7,$AF,$E4,$FE,$01,$23,$8B,$F0,$FE
	.dc.b	$3B,$51,$E6,$8F,$85,$1F,$0F,$E3,$FE,$27,$C7,$CA,$F5,$C6,$5A,$43
	.dc.b	$D7,$F1,$5F,$2F,$E3,$BE,$81,$23,$D2,$FD,$3F,$8B,$D4,$7C,$68,$FA
	.dc.b	$51,$FE,$4E,$F9,$27,$94,$E5,$FF,$6B,$F9,$ED,$3F,$D1,$74,$F3,$3F
	.dc.b	$D1,$B3,$9C,$BF,$DB,$B3,$97,$8D,$70,$C3,$A4,$7F,$3A,$8C,$61,$A8
	.dc.b	$E3,$87,$49,$EE,$D8,$EB,$5C,$B0,$F0,$43,$0F,$4E,$1F,$EE,$EF,$A2
	.dc.b	$CC,$78,$E5,$18,$C3,$FD,$4A,$EB,$0F,$2B,$6F,$C9,$5D,$6B,$C4,$D3
	.dc.b	$37,$E0,$7D,$87,$9A,$73,$3C,$27,$C5,$87,$5A,$D9,$E9,$C3,$C3,$8F
	.dc.b	$C0,$8F,$53,$5F,$4B,$F2,$1A,$77,$29,$E5,$2D,$7C,$1C,$7C,$04,$9E
	.dc.b	$8F,$C3,$9E,$E7,$37,$CD,$79,$8F,$3D,$7A,$AB,$C5,$3C,$E7,$EC,$9C
	.dc.b	$8F,$15,$65,$F7,$63,$F1,$18,$74,$C3,$EE,$5F,$43,$E0,$5F,$84,$F1
	.dc.b	$BF,$07,$D1,$FF,$01,$D7,$C1,$6B,$D8,$38,$C7,$EB,$3A,$80,$FA,$3B
	.dc.b	$19,$8C,$1C,$AF,$FD,$58,$E2,$35,$3C,$27,$2F,$E2,$25,$CE,$E7,$C4
	.dc.b	$74,$AF,$CC,$BA,$8C,$D7,$1A,$E1,$3F,$86,$3C,$F1,$99,$DC,$7E,$00
	.dc.b	$E4,$53,$D3,$A7,$6A,$E6,$F8,$23,$C5,$97,$FB,$1E,$7B,$E1,$9C,$3B
	.dc.b	$57,$84,$E7,$5B,$C3,$EF,$47,$90,$F1,$5F,$69,$F8,$71,$89,$DB,$C8
	.dc.b	$FA,$27,$A3,$0F,$03,$B3,$D5,$7D,$AF,$A4,$7F,$B7,$63,$CB,$87,$03
	.dc.b	$DD,$7F,$11,$B3,$E9,$9D,$E1,$E4,$C3,$27,$80,$AF,$A2,$B8,$9D,$5F
	.dc.b	$81,$5C,$CF,$A9,$87,$C8,$8E,$06,$A7,$F9,$27,$3B,$9D,$F1,$DF,$73
	.dc.b	$7F,$BB,$39,$F8,$EC,$6B,$FD,$62,$AE,$7F,$E6,$7C,$3F,$DA,$F3,$FD
	.dc.b	$7A,$BF,$0B,$3F,$2A,$F9,$4F,$F5,$89,$CC,$FF,$62,$6B,$CF,$E3,$7A
	.dc.b	$D7,$95,$B6,$F8,$DB,$13,$0F,$ED,$BF,$1B,$70,$FE,$D6,$F0,$F8,$DE
	.dc.b	$38,$5A,$FE,$E6,$E1,$FE,$D6,$F8,$B1,$F8,$8D,$3E,$34,$7F,$70,$F8
	.dc.b	$DF,$1F,$C6,$F9,$1B,$1F,$DF,$A3,$81,$1F,$EE,$EF,$3F,$8D,$F5,$9E
	.dc.b	$EF,$A7,$FD,$94,$FF,$81,$A3,$2F,$F8,$93,$81,$E2,$7F,$04,$8D,$19
	.dc.b	$71,$AF,$F9,$DA,$FF,$DC,$D3,$FE,$12,$7F,$DC,$0D,$F8,$EF,$07,$C6
	.dc.b	$48,$FF,$67,$CE,$AB,$FC,$22,$D7,$ED,$35,$FF,$12,$2F,$8D,$F5,$1E
	.dc.b	$BF,$8C,$B5,$FE,$22,$F3,$74,$D9,$CD,$FF,$A3,$27,$2F,$A7,$E3,$3E
	.dc.b	$21,$F1,$7E,$2E,$DF,$F9,$1A,$5F,$43,$FC,$AA,$30,$D8,$E5,$18,$EF
	.dc.b	$FC,$BC,$70,$DA,$F2,$3C,$5C,$E1,$F3,$B0,$ED,$39,$6C,$FE,$5A,$FF
	.dc.b	$D8,$CD,$9D,$47,$FB,$E2,$31,$E2,$CA,$FF,$5C,$B6,$7E,$6F,$C6,$FE
	.dc.b	$04,$FC,$17,$C6,$DA,$F9,$1F,$11,$FA,$2D,$FF,$DE,$91,$F6,$EB,$13
	.dc.b	$F8,$9F,$8D,$93,$F7,$FE,$36,$E3,$F3,$7D,$AE,$46,$3A,$3F,$7B,$92
	.dc.b	$3E,$1F,$83,$F1,$3C,$0B,$8F,$C2,$70,$FF,$B4,$9F,$80,$47,$FD,$AB
	.dc.b	$E0,$FC,$32,$38,$3F,$1B,$F1,$9F,$9F,$F8,$DF,$D0,$36,$3F,$2D,$B7
	.dc.b	$E8,$CC,$69,$F9,$B7,$F4,$DF,$0B,$27,$F3,$9F,$09,$6B,$F4,$1F,$09
	.dc.b	$FA,$B3,$1B,$FF,$1A,$EE,$47,$EB,$FE,$33,$F6,$1F,$19,$71,$FD,$33
	.dc.b	$87,$F4,$6F,$F0,$3A,$34,$C7,$0F,$EE,$4E,$18,$FE,$C5,$FD,$E7,$C0
	.dc.b	$CB,$FF,$41,$FE,$EF,$F9,$5C,$FC,$93,$FC,$7A,$1E,$65,$38,$61,$FE
	.dc.b	$F0,$8C,$DF,$2F,$A1,$E0,$71,$F8,$8B,$19,$3B,$FC,$76,$2F,$93,$DF
	.dc.b	$E3,$BC,$1F,$1D,$6B,$CA,$E1,$FF,$05,$F8,$2C,$93,$FC,$95,$B3,$A6
	.dc.b	$E1,$C3,$CD,$FD,$21,$27,$FC,$2B,$0B,$4F,$3F,$63,$E1,$1F,$0F,$D8
	.dc.b	$B7,$F2,$3E,$27,$C0,$47,$C0,$F8,$FE,$37,$E5,$7C,$77,$CB,$6C,$FD
	.dc.b	$0F,$8D,$F9,$C7,$CF,$F8,$DB,$1F,$53,$E3,$7D,$84,$7D,$8F,$BF,$E1
	.dc.b	$7E,$B7,$C5,$FD,$76,$CF,$DC,$F1,$BE,$D1,$F6,$FE,$26,$C7,$F1,$73
	.dc.b	$FD,$32,$7C,$22,$F0,$FF,$8E,$1A,$FF,$84,$0F,$F1,$A6,$DF,$3F,$E2
	.dc.b	$BF,$0D,$E8,$7E,$1D,$B5,$FE,$CE,$FD,$E3,$FA,$0F,$6F,$F7,$0F,$A0
	.dc.b	$7C,$3F,$F6,$A9,$CD,$7F,$87,$27,$2F,$F2,$B5,$E5,$FF,$1D,$6C,$7D
	.dc.b	$86,$E7,$FE,$4E,$6B,$FD,$F6,$BF,$C7,$B8,$7F,$82,$17,$CC,$F8,$7E
	.dc.b	$0F,$C4,$F8,$FE,$2E,$63,$E9,$F8,$BB,$8F,$FA,$F6,$3F,$DE,$18,$3E
	.dc.b	$2F,$FB,$A3,$63,$F2,$AD,$9F,$F7,$3B,$5F,$27,$C3,$7C,$13,$E1,$7C
	.dc.b	$5D,$AF,$91,$BF,$C5,$3E,$37,$B3,$63,$E6,$7B,$3E,$02,$3E,$07,$C7
	.dc.b	$F1,$7F,$34,$F9,$DF,$11,$67,$FA,$B7,$D5,$3D,$5F,$EA,$2D,$8C,$FA
	.dc.b	$D8,$9F,$EA,$BF,$3F,$FE,$0E,$DB,$FC,$F6,$D7,$F0,$8D,$7F,$CA,$AD
	.dc.b	$7F,$D0,$EF,$EB,$2B,$F5,$ED,$8E,$7E,$37,$F5,$98,$C4,$7E,$3F,$0F
	.dc.b	$E8,$55,$FD,$B7,$F0,$67,$B7,$E8,$F4,$BF,$B4,$C6,$27,$53,$FD,$8E
	.dc.b	$31,$7F,$E9,$26,$BF,$9B,$6B,$F7,$3D,$BF,$CA,$1F,$95,$EC,$B7,$F9
	.dc.b	$B8,$FF,$26,$F5,$BF,$7A,$7C,$16,$B0,$FC,$30,$9D,$47,$E7,$FD,$D9
	.dc.b	$87,$E9,$7E,$32,$DF,$0F,$3D,$D7,$FE,$4B,$AF,$D5,$7C,$66,$B2,$FC
	.dc.b	$D3,$FE,$52,$FA,$07,$7D,$3F,$59,$F1,$B6,$FF,$D6,$F7,$FE,$68,$AD
	.dc.b	$C6,$0F,$F5,$DE,$28,$78,$3C,$73,$F2,$46,$2B,$84,$7F,$27,$8C,$3F
	.dc.b	$C1,$6B,$95,$7F,$95,$E7,$33,$FE,$6F,$6B,$FA,$AE,$7F,$E0,$47,$F0
	.dc.b	$3F,$8D,$B8,$7F,$09,$8F,$E0,$DF,$1B,$8A,$F8,$DB,$23,$FC,$53,$F4
	.dc.b	$1F,$E3,$FF,$1B,$F9,$DB,$6B,$FD,$A2,$D7,$F8,$D3,$5F,$F5,$5F,$C8
	.dc.b	$BD,$57,$D0,$7F,$44,$8C,$36,$33,$3F,$ED,$C9,$CB,$DB,$E3,$BF,$DD
	.dc.b	$2D,$F8,$CF,$1C,$62,$3C,$2D,$7E,$8F,$CE,$7D,$07,$E6,$F1,$3D,$1F
	.dc.b	$F4,$BF,$27,$FA,$81,$BC,$3C,$39,$3F,$D8,$73,$FE,$A4,$C3,$96,$1F
	.dc.b	$8B,$6C,$66,$77,$5E,$0F,$36,$DF,$F6,$1E,$8F,$EC,$2F,$F6,$E6,$BF
	.dc.b	$75,$AF,$FA,$A5,$AF,$E2,$9F,$F5,$C3,$EA,$78,$DF,$FA,$E1,$B5,$BE
	.dc.b	$7F,$0E,$69,$E3,$7F,$A3,$6D,$F1,$AD,$C3,$9B,$8E,$6F,$2F,$9E,$AF
	.dc.b	$91,$A6,$A2,$65,$F3,$AB,$83,$63,$9F,$9B,$F1,$33,$7D,$03,$E3,$BA
	.dc.b	$FC,$94,$C3,$FD,$45,$39,$6C,$F1,$8C,$47,$F0,$4A,$E7,$3E,$8F,$3F
	.dc.b	$E8,$CB,$E8,$3E,$0B,$FD,$BD,$AF,$FC,$00,$6A,$38,$70,$7F,$86,$9E
	.dc.b	$5D,$DD,$27,$E4,$F1,$F1,$AC,$4F,$83,$EF,$1E,$C5,$26,$1F,$E9,$BF
	.dc.b	$A0,$7C,$8F,$99,$FE,$63,$1F,$E2,$86,$BF,$E0,$50,$E1,$FF,$04,$35
	.dc.b	$F4,$1A,$F4,$3A,$11,$FF,$55,$7C,$6F,$FB,$31,$F9,$BE,$77,$CB,$23
	.dc.b	$FA,$6E,$7F,$F5,$F3,$FE,$B1,$FA,$E7,$27,$C3,$6C,$7F,$8F,$9A,$F9
	.dc.b	$DA,$F9,$3A,$FF,$61,$DD,$F8,$A2,$47,$E1,$FC,$EF,$C2,$7D,$1B,$D7
	.dc.b	$A4,$9F,$27,$CB,$7D,$CC,$7C,$B1,$DF,$E5,$FC,$84,$7E,$47,$53,$E7
	.dc.b	$7F,$CD,$F1,$FE,$ED,$6B,$F8,$96,$C7,$EF,$F9,$BF,$13,$F1,$7F,$5A
	.dc.b	$BF,$DA,$4F,$E4,$25,$D3,$F4,$23,$0D,$E6,$E3,$F2,$BF,$89,$93,$49
	.dc.b	$7F,$E6,$6F,$A2,$FE,$35,$87,$48,$FC,$95,$3E,$4B,$C6,$7E,$72,$B7
	.dc.b	$8F,$F4,$EE,$5F,$EC,$F7,$F6,$7E,$63,$F6,$E7,$EE,$3D,$6E,$1F,$09
	.dc.b	$F9,$BC,$3F,$05,$5F,$8D,$6B,$FE,$53,$6B,$D7,$C2,$FE,$02,$47,$F9
	.dc.b	$5F,$D0,$BA,$3F,$69,$F3,$3F,$C1,$B8,$65,$7E,$EA,$3F,$67,$F4,$5F
	.dc.b	$C7,$23,$F5,$CD,$4D,$35,$C3,$FC,$1B,$E6,$7F,$80,$12,$7C,$37,$FC
	.dc.b	$0F,$E6,$3F,$8A,$F7,$C9,$FE,$13,$1E,$0F,$37,$F7,$58,$FF,$1D,$9F
	.dc.b	$F4,$AB,$5F,$D7,$35,$FE,$83,$F3,$DF,$39,$FF,$08,$37,$23,$9F,$01
	.dc.b	$B7,$DE,$AF,$D5,$78,$7F,$CA,$08,$FE,$B7,$E1,$3F,$8A,$FD,$17,$EE
	.dc.b	$63,$F8,$83,$53,$4D,$5F,$6D,$F8,$ED,$2B,$E9,$C7,$F4,$0C,$7F,$DF
	.dc.b	$D1,$FB,$BE,$0F,$E7,$26,$7E,$0E,$47,$DA,$6B,$FE,$27,$9F,$F7,$0B
	.dc.b	$FC,$0F,$B3,$FA,$09,$1F,$ED,$7F,$41,$F0,$4F,$85,$D5,$F0,$63,$FC
	.dc.b	$60,$DE,$0F,$CE,$68,$AD,$FC,$EF,$F4,$CD,$26,$1F,$C4,$63,$FA,$CC
	.dc.b	$BE,$6B,$FA,$8C,$49,$F0,$C6,$BE,$6B,$6C,$8F,$2C,$7F,$4F,$E3,$FF
	.dc.b	$09,$47,$F1,$DF,$2B,$FA,$D9,$9F,$6F,$BF,$CD,$59,$FF,$19,$EB,$F7
	.dc.b	$A7,$31,$FC,$97,$A7,$FB,$1F,$14,$8C,$1F,$12,$7D,$24,$7E,$3F,$CB
	.dc.b	$FE,$03,$9E,$47,$F9,$00,$FB,$EF,$E1,$39,$7F,$0B,$C9,$FD,$47,$9F
	.dc.b	$E6,$57,$FB,$1D,$AF,$F4,$06,$BF,$8A,$E1,$E2,$F5,$8F,$A5,$E1,$6E
	.dc.b	$7F,$C0,$BF,$42,$F9,$CA,$4B,$FF,$13,$C7,$D1,$6F,$26,$5F,$F3,$64
	.dc.b	$71,$72,$7C,$8F,$AC,$FF,$12,$E1,$FE,$62,$97,$CA,$F8,$0F,$F0,$14
	.dc.b	$EA,$9D,$2A,$FA,$F1,$FD,$F6,$75,$C6,$78,$56,$F8,$7F,$57,$CB,$2F
	.dc.b	$E0,$3F,$AE,$E4,$FD,$7F,$27,$CB,$24,$FF,$6D,$3F,$5E,$FC,$FF,$1F
	.dc.b	$71,$F2,$63,$F6,$F1,$FE,$0C,$F2,$BF,$DD,$F3,$F6,$5F,$F0,$FB,$FE
	.dc.b	$96,$3F,$A8,$7B,$5B,$F9,$2E,$1E,$CD,$9E,$11,$87,$E4,$7C,$97,$D4
	.dc.b	$F0,$24,$60,$FA,$7F,$2B,$F2,$26,$46,$0F,$CB,$F7,$3E,$67,$D9,$C7
	.dc.b	$FC,$59,$B1,$99,$FE,$CA,$D7,$FD,$1D,$E0,$FF,$44,$AF,$EA,$47,$18
	.dc.b	$EB,$5F,$B4,$F3,$3F,$59,$F2,$FF,$E4,$4E,$09,$5E,$28,$CB,$48,$7F
	.dc.b	$AC,$FD,$C7,$C4,$FE,$D7,$DC,$F2,$92,$7C,$AF,$D5,$E8,$57,$E0,$B5
	.dc.b	$FF,$16,$D7,$C0,$AF,$F2,$FC,$FE,$13,$EB,$FF,$D0,$53,$FE,$B2,$6B
	.dc.b	$F8,$E6,$D9,$B2,$F9,$7F,$F3,$4C,$7F,$3D,$3F,$9F,$46,$6B,$FD,$16
	.dc.b	$FE,$09,$FF,$67,$47,$F8,$2B,$BB,$C1,$F3,$FE,$4E,$59,$5C,$1A,$FC
	.dc.b	$3E,$7E,$FE,$0F,$BF,$0F,$D6,$4F,$FB,$05,$AF,$F8,$AD,$AF,$FB,$3B
	.dc.b	$1F,$F1,$43,$FE,$9E,$8E,$3C,$7F,$CC,$1B,$96,$FD,$BF,$F3,$2F,$C5
	.dc.b	$7E,$D2,$BF,$CC,$1F,$44,$F8,$5F,$F5,$CF,$CF,$FF,$93,$23,$19,$7F
	.dc.b	$99,$6D,$39,$AF,$F5,$FF,$8B,$FE,$F5,$A5,$F0,$57,$DD,$70,$66,$39
	.dc.b	$C7,$F9,$C6,$31,$87,$F9,$71,$D5,$F5,$6E,$3E,$8C,$7C,$F7,$4F,$35
	.dc.b	$F9,$5F,$45,$F9,$71,$E3,$1E,$6A,$E2,$D3,$89,$CC,$F4,$CF,$18,$F7
	.dc.b	$DF,$9D,$E3,$87,$A2,$B0,$78,$26,$BC,$9E,$98,$FC,$08,$F0,$1F,$1B
	.dc.b	$1F,$43,$B3,$31,$F2,$71,$ED,$18,$9D,$CE,$70,$F2,$CE,$4E,$F3,$AC
	.dc.b	$7D,$51,$E5,$3E,$9C,$FA,$88,$E7,$1F,$8F,$87,$80,$93,$D9,$F9,$71
	.dc.b	$E4,$24,$78,$9C,$BD,$27,$DC,$7C,$4C,$3F,$09,$3F,$53,$E3,$3F,$CE
	.dc.b	$B1,$F8,$76,$E1,$F8,$D6,$E5,$FE,$B7,$7F,$CF,$77,$FE,$B9,$FA,$3C
	.dc.b	$61,$FE,$A9,$6C,$7F,$A9,$83,$87,$FD,$52,$D7,$91,$F1,$0F,$F5,$6F
	.dc.b	$3F,$F9,$D6,$7F,$39,$1A,$6C,$67,$2F,$04,$F1,$3F,$D9,$91,$97,$C5
	.dc.b	$5F,$84,$7F,$91,$ED,$5F,$29,$F8,$43,$25,$33,$A7,$82,$3F,$2D,$1E
	.dc.b	$3B,$F7,$5F,$78,$F6,$9E,$33,$9E,$1E,$A7,$9C,$F0,$AE,$71,$83,$E0
	.dc.b	$B8,$C3,$BD,$FF,$A5,$F0,$E5,$18,$9D,$57,$A6,$B5,$7F,$12,$FD,$13
	.dc.b	$EB,$BF,$85,$5E,$78,$F5,$1E,$03,$D0,$78,$CF,$97,$7D,$AB,$99,$E2
	.dc.b	$8F,$04,$66,$FD,$91,$DC,$F8,$55,$DD,$F9,$F8,$79,$0E,$57,$D2,$BD
	.dc.b	$07,$27,$F9,$1F,$2F,$F5,$4F,$97,$FE,$CF,$8F,$F4,$FF,$26,$E3,$FE
	.dc.b	$1C,$E3,$56,$2F,$FA,$85,$5E,$3F,$F2,$F6,$1F,$EF,$F9,$CB,$63,$38
	.dc.b	$7F,$43,$AF,$F6,$33,$FE,$ED,$C1,$F2,$BD,$5B,$7F,$D5,$BE,$17,$13
	.dc.b	$23,$FA,$37,$A7,$6F,$FA,$CB,$C6,$FC,$9F,$01,$6B,$FA,$63,$87,$C8
	.dc.b	$F8,$AB,$FB,$73,$5F,$ED,$CD,$CF,$FB,$CF,$B5,$F1,$39,$3F,$D5,$3E
	.dc.b	$2F,$9E,$31,$DE,$8A,$73,$9D,$E3,$FE,$18,$8C,$4E,$E3,$FD,$FF,$E1
	.dc.b	$32,$EC,$D8,$FE,$EF,$F2,$58,$99,$1F,$D9,$FD,$6B,$1F,$DE,$9F,$AD
	.dc.b	$F4,$4F,$88,$E7,$F2,$18,$6D,$FF,$A1,$B1,$4F,$9D,$B1,$E4,$8C,$1F
	.dc.b	$0F,$C5,$7C,$37,$C3,$FA,$D7,$B5,$F8,$59,$CD,$FF,$A4,$A7,$2D,$9D
	.dc.b	$7C,$17,$16,$C7,$FA,$2F,$E9,$1E,$17,$E4,$B7,$FF,$22,$79,$96,$73
	.dc.b	$5F,$E9,$1F,$35,$3B,$6C,$73,$F0,$B1,$3C,$5E,$9D,$0F,$46,$F9,$38
	.dc.b	$6D,$FF,$A7,$65,$3F,$85,$CB,$FC,$FE,$D8,$CC,$FF,$8F,$E3,$1F,$3D
	.dc.b	$D8,$EF,$E8,$D9,$FB,$DD,$7F,$EC,$CE,$27,$BB,$F2,$F5,$F9,$4D,$AF
	.dc.b	$95,$1F,$37,$DD,$C6,$8C,$39,$47,$D1,$F5,$BE,$95,$5B,$EE,$7E,$37
	.dc.b	$56,$5F,$1B,$1E,$53,$98,$C6,$3E,$5B,$E5,$5A,$3F,$D3,$F3,$AA,$FC
	.dc.b	$C4,$7F,$A8,$BC,$4F,$9B,$F3,$5F,$39,$B3,$F2,$E3,$E4,$11,$FF,$33
	.dc.b	$3F,$2F,$E5,$3E,$A9,$F5,$BC,$6B,$3E,$AB,$FF,$57,$F7,$7C,$88,$C6
	.dc.b	$88,$C4,$F9,$E3,$1D,$1F,$A4,$E0,$FE,$31,$E8,$7F,$19,$F3,$EE,$3F
	.dc.b	$E5,$17,$F5,$3F,$03,$FB,$2F,$3A,$CF,$E9,$DC,$3F,$E5,$7F,$72,$E7
	.dc.b	$FE,$6A,$7F,$63,$ED,$FE,$CB,$96,$D7,$9B,$64,$66,$3F,$CC,$1F,$03
	.dc.b	$70,$FF,$39,$BF,$CB,$4F,$E5,$DF,$47,$67,$FC,$CC,$FF,$33,$FA,$4F
	.dc.b	$F2,$9C,$7F,$9A,$7E,$93,$1D,$FF,$CD,$59,$19,$3E,$3C,$F0,$C3,$FD
	.dc.b	$03,$3C,$FD,$09,$5A,$F6,$2D,$7C,$BB,$57,$F4,$1F,$0B,$3D,$BF,$E7
	.dc.b	$77,$C9,$E5,$FE,$BB,$E3,$AE,$5F,$AB,$70,$FF,$A9,$5F,$D5,$11,$FE
	.dc.b	$91,$F4,$9F,$D3,$4F,$E9,$DE,$F5,$8F,$E3,$1D,$DF,$C9,$8F,$85,$D3
	.dc.b	$63,$F0,$DA,$BE,$A5,$F1,$9F,$B9,$6C,$FF,$BA,$5B,$3F,$E2,$D3,$84
	.dc.b	$FF,$23,$F5,$BF,$AF,$7D,$2F,$F3,$46,$DF,$F2,$9F,$5B,$F9,$B9,$F4
	.dc.b	$F9,$2C,$78,$B1,$7C,$5F,$84,$91,$E1,$7F,$A3,$7D,$0B,$E6,$FF,$B6
	.dc.b	$79,$1F,$D3,$E3,$15,$FE,$A5,$8F,$EE,$1F,$0D,$23,$ED,$C5,$DB,$FD
	.dc.b	$F5,$0F,$03,$F8,$C3,$4F,$74,$1C,$3F,$05,$AF,$FA,$DF,$DA,$FE,$FD
	.dc.b	$EB,$FF,$7F,$6C,$FC,$EF,$A5,$7C,$2F,$F3,$77,$F9,$64,$66,$31,$E9
	.dc.b	$E7,$9B,$F9,$55,$1A,$4A,$F2,$C3,$8B,$A7,$B7,$D5,$B3,$FC,$0F,$B5
	.dc.b	$F2,$EF,$9D,$FE,$EC,$D9,$FE,$0B,$F4,$17,$7F,$E3,$57,$C2,$FF,$A7
	.dc.b	$78,$8F,$6B,$F9,$F3,$70,$F8,$ED,$BF,$C3,$35,$FE,$43,$6F,$9D,$7C
	.dc.b	$D7,$F8,$1D,$B7,$FC,$5B,$F1,$4F,$EF,$97,$E2,$9C,$BF,$C5,$E3,$FC
	.dc.b	$CF,$7F,$56,$1E,$9A,$73,$7F,$CE,$F4,$FA,$2F,$5F,$88,$F5,$E5,$EA
	.dc.b	$24,$7A,$9F,$C7,$9F,$8D,$3F,$C7,$95,$FE,$00,$C3,$E2,$F5,$79,$32
	.dc.b	$FF,$1B,$FD,$33,$C0,$FE,$E8,$DC,$7F,$5F,$F4,$CF,$97,$CD,$6B,$C8
	.dc.b	$FE,$D1,$66,$92,$FF,$27,$F6,$7D,$DF,$E9,$6D,$CF,$FC,$49,$5F,$83
	.dc.b	$9F,$DE,$CE,$65,$F4,$FF,$AE,$9F,$E0,$2F,$8D,$BE,$1F,$F8,$2E,$31
	.dc.b	$3F,$D6,$CE,$39,$7F,$7D,$8C,$C7,$FA,$BD,$AF,$F5,$E6,$BF,$DC,$75
	.dc.b	$7C,$EF,$ED,$CF,$F3,$37,$17,$F9,$A3,$1F,$CB,$4E,$9B,$3F,$43,$9F
	.dc.b	$84,$7E,$15,$E3,$EE,$48,$FC,$CF,$6D,$C7,$FA,$E6,$2A,$FB,$3F,$4F
	.dc.b	$9E,$3F,$C4,$73,$FE,$1E,$3F,$CF,$1F,$4F,$FA,$19,$CB,$72,$FF,$02
	.dc.b	$29,$DE,$72,$7F,$B6,$CB,$E5,$31,$9F,$F6,$D8,$C4,$EA,$7F,$CD,$11
	.dc.b	$89,$FF,$6C,$62,$8C,$47,$F8,$22,$31,$3F,$E6,$FC,$DD,$FF,$28,$F9
	.dc.b	$7C,$B7,$1F,$DD,$C7,$E6,$2B,$18,$7E,$A5,$FF,$24,$7D,$3F,$F9,$5B
	.dc.b	$AF,$F8,$FB,$67,$F9,$F3,$5F,$F7,$63,$5F,$F0,$DF,$CA,$7F,$29,$FA
	.dc.b	$7F,$E4,$ED,$8F,$E8,$CD,$7F,$DE,$8D,$7F,$C4,$73,$FC,$2F,$E2,$BC
	.dc.b	$3E,$0F,$F9,$72,$7C,$2D,$7C,$2D,$7F,$C6,$DD,$CF,$7B,$F8,$6E,$B8
	.dc.b	$F9,$6B,$5F,$C1,$29,$18,$E1,$FE,$0D,$77,$CD,$FF,$2B,$13,$0F,$E8
	.dc.b	$32,$8F,$F2,$D5,$FF,$3A,$C3,$FD,$74,$D7,$EC,$B5,$FB,$0D,$7C,$FE
	.dc.b	$47,$86,$38,$73,$FE,$49,$B8,$6A,$BF,$C6,$FD,$1F,$C0,$DF,$F6,$43
	.dc.b	$5F,$C3,$B5,$FC,$33,$5F,$F2,$D6,$1F,$D9,$DF,$E5,$9F,$15,$E1,$3E
	.dc.b	$07,$9F,$23,$FC,$8B,$3B,$7F,$C8,$8D,$C7,$F4,$FF,$4F,$F2,$A7,$7F
	.dc.b	$21,$EB,$F8,$B9,$18,$21,$CA,$3E,$71,$F5,$2B,$FA,$17,$D0,$FD,$69
	.dc.b	$5F,$EA,$5F,$EE,$2D,$7F,$C2,$0D,$7F,$CD,$67,$0E,$8F,$EE,$EB,$E9
	.dc.b	$FF,$80,$3F,$E1,$E9,$E5,$7F,$EE,$C6,$ED,$FE,$74,$AD,$D7,$F1,$FF
	.dc.b	$9C,$FF,$6B,$3F,$E2,$66,$B9,$F4,$BD,$1F,$37,$FE,$3E,$8F,$F1,$5F
	.dc.b	$D3,$FF,$A2,$1F,$F0,$4B,$5F,$F1,$43,$5F,$4B,$5F,$F4,$2F,$D3,$FF
	.dc.b	$88,$9F,$F0,$8B,$5F,$F1,$83,$5F,$82,$D7,$F4,$7F,$25,$F8,$D2,$3F
	.dc.b	$DA,$FE,$6F,$F2,$1F,$23,$F9,$CE,$73,$49,$3F,$9C,$F2,$3F,$2C,$47
	.dc.b	$F0,$BD,$BF,$99,$F9,$19,$1F,$DD,$DF,$CE,$75,$7E,$77,$D1,$EC,$49
	.dc.b	$FD,$3B,$E1,$F1,$BF,$44,$48,$FF,$0B,$7C,$F3,$D6,$FC,$27,$CD,$3E
	.dc.b	$9A,$92,$BF,$CB,$30,$EB,$FD,$67,$B5,$ED,$24,$FE,$D9,$FC,$2C,$FF
	.dc.b	$A3,$23,$8F,$C7,$76,$3B,$FC,$AF,$ED,$7C,$DF,$CA,$7D,$93,$E8,$E9
	.dc.b	$9B,$FF,$55,$FE,$F5,$FE,$DD,$27,$7F,$F0,$87,$FD,$C5,$0F,$C1,$B9
	.dc.b	$3F,$E7,$06,$BF,$DB,$5B,$E1,$7F,$54,$F1,$BD,$93,$DB,$C7,$FE,$31
	.dc.b	$6F,$1A,$98,$8C,$FC,$EF,$84,$95,$FE,$77,$F2,$D6,$9F,$37,$5F,$42
	.dc.b	$57,$F9,$1F,$E0,$1F,$2F,$E0,$F8,$1F,$0F,$B5,$2B,$E0,$C7,$93,$A7
	.dc.b	$E0,$12,$3C,$19,$7D,$96,$F4,$FA,$88,$FF,$95,$7E,$7F,$F9,$F7,$0F
	.dc.b	$F7,$4A,$FF,$49,$CF,$CE,$F8,$CF,$82,$7C,$2F,$80,$FE,$71,$F0,$1F
	.dc.b	$CD,$B9,$3F,$C5,$95,$EF,$6B,$FD,$B1,$AF,$F6,$B5,$F3,$7C,$A3,$F9
	.dc.b	$0C,$FF,$43,$8F,$9E,$C8,$FB,$EF,$F4,$7F,$9B,$91,$FC,$A1,$AF,$FB
	.dc.b	$AF,$E9,$FF,$1C,$FF,$BB,$72,$EE,$FE,$22,$7E,$D9,$F7,$2F,$FA,$97
	.dc.b	$CE,$D9,$FC,$6C,$61,$FC,$7C,$7E,$08,$8F,$F3,$8E,$2F,$EB,$44,$70
	.dc.b	$7F,$3C,$EB,$FE,$BC,$47,$00,$6D,$FE,$24,$F3,$DF,$43,$F9,$0A,$95
	.dc.b	$FA,$07,$FB,$47,$A5,$23,$FA,$4B,$5F,$F4,$4F,$17,$FA,$B5,$FD,$3E
	.dc.b	$3F,$9A,$23,$FD,$43,$C2,$FE,$E4,$47,$07,$F5,$4F,$9D,$FE,$8D,$FB
	.dc.b	$C3,$E8,$A9,$27,$FD,$B7,$F5,$1F,$80,$DB,$FD,$09,$08,$FA,$2F,$FB
	.dc.b	$65,$AF,$F4,$A6,$BF,$0C,$E1,$C1,$FD,$32,$1C,$3F,$D5,$DE,$75,$D9
	.dc.b	$BE,$1F,$17,$77,$F3,$67,$FC,$9F,$F6,$7F,$B9,$8F,$E1,$3B,$FE,$F0
	.dc.b	$FF,$0A,$46,$63,$E9,$DF,$D5,$F0,$5F,$52,$F9,$FD,$F8,$B2,$FB,$5E
	.dc.b	$EB,$51,$83,$FD,$4B,$87,$F0,$17,$18,$23,$11,$E0,$37,$E2,$7F,$89
	.dc.b	$E9,$F4,$74,$93,$CA,$32,$FF,$B3,$FE,$A7,$FD,$9C,$F8,$47,$D7,$FF
	.dc.b	$70,$7F,$20,$D7,$3E,$FF,$F8,$0E,$3F,$CD,$D1,$F7,$36,$FA,$44,$7F
	.dc.b	$DE,$BC,$08,$FA,$DC,$BF,$55,$93,$F8,$48,$F9,$BE,$34,$8F,$83,$AF
	.dc.b	$FA,$51,$F3,$7A,$3F,$E1,$D7,$FD,$40,$DE,$3F,$C5,$B7,$87,$F0,$CD
	.dc.b	$9F,$D5,$3F,$E3,$A7,$E7,$F8,$BF,$70,$8F,$C7,$F0,$FE,$61,$1C,$1F
	.dc.b	$27,$E0,$3F,$9F,$E7,$FD,$2C,$91,$FC,$B1,$FE,$9D,$D7,$A8,$FE,$A7
	.dc.b	$1F,$D1,$A3,$FC,$B9,$E7,$FF,$48,$24,$FF,$8F,$9C,$B4,$87,$EE,$3E
	.dc.b	$77,$FC,$CD,$5F,$E6,$F8,$FC,$93,$5F,$F7,$7B,$79,$CD,$3F,$BD,$47
	.dc.b	$E3,$9F,$E7,$BA,$72,$F8,$6E,$BE,$A7,$6E,$BF,$F3,$7D,$7E,$62,$FF
	.dc.b	$D4,$6D,$7F,$D4,$4D,$7C,$7E,$1F,$17,$B4,$FA,$23,$FC,$FD,$F5,$1E
	.dc.b	$29,$FC,$E6,$5E,$6F,$A2,$F6,$9A,$C6,$4F,$E9,$2B,$87,$1F,$FB,$19
	.dc.b	$BA,$7E,$CF,$D8,$FB,$D8,$FC,$0F,$5B,$D2,$4C,$3F,$B8,$6A,$8F,$6F
	.dc.b	$D0,$FC,$C6,$4F,$C5,$8F,$F3,$17,$C5,$FE,$BB,$3F,$E6,$9F,$53,$FE
	.dc.b	$9D,$AF,$E6,$D2,$FA,$1F,$E2,$D3,$99,$4E,$5E,$3F,$17,$FE,$B6,$F0
	.dc.b	$FF,$9E,$C2,$B8,$57,$78,$FF,$18,$57,$FA,$F2,$FF,$91,$70,$69,$B1
	.dc.b	$CF,$1F,$E9,$13,$CB,$1E,$B8,$EF,$0F,$09,$E8,$C3,$CF,$87,$18,$F1
	.dc.b	$36,$78,$46,$0E,$07,$23,$DB,$5D,$A7,$95,$F0,$23,$82,$4F,$23,$DA
	.dc.b	$58,$F4,$61,$EA,$9F,$11,$F0,$67,$C1,$8F,$CB,$7C,$87,$8A,$73,$7E
	.dc.b	$38,$ED,$87,$A8,$EA,$F0,$AF,$9D,$87,$9A,$3C,$58,$79,$9E,$86,$F1
	.dc.b	$F8,$EF,$48,$F2,$BF,$82,$C3,$ED,$D7,$38,$FC,$0D,$7D,$67,$58,$FB
	.dc.b	$5A,$7C,$23,$66,$4A,$67,$0F,$3E,$1F,$5F,$93,$F5,$51,$F8,$76,$E3
	.dc.b	$F8,$C6,$DF,$F8,$2D,$FD,$85,$FF,$83,$7E,$AB,$19,$7F,$AE,$21,$18
	.dc.b	$7F,$D7,$6D,$7F,$B9,$35,$E4,$7A,$47,$C8,$38,$BD,$51,$AA,$FD,$23
	.dc.b	$A8,$CB,$E0,$39,$4F,$17,$A1,$99,$E3,$19,$79,$C7,$1D,$73,$5F,$95
	.dc.b	$8F,$2C,$61,$FC,$39,$E3,$D3,$FD,$AF,$7C,$DF,$85,$87,$A2,$31,$3F
	.dc.b	$08,$E2,$F9,$6B,$C3,$3D,$E3,$15,$CA,$B7,$7E,$37,$CD,$1E,$08,$C5
	.dc.b	$77,$7E,$1B,$67,$E0,$B7,$0E,$F1,$F1,$A7,$6F,$63,$DA,$76,$3F,$C5
	.dc.b	$17,$E4,$7D,$58,$7C,$33,$99,$70,$F4,$5F,$C7,$3B,$95,$F6,$46,$E7
	.dc.b	$9C,$7A,$DF,$E0,$DC,$3F,$D4,$36,$FB,$5A,$7D,$B8,$FE,$A7,$AE,$33
	.dc.b	$FE,$B3,$56,$FF,$AC,$CD,$AD,$D7,$F3,$88,$FF,$24,$CF,$F6,$06,$BC
	.dc.b	$23,$FD,$59,$C6,$E4,$E4,$D8,$FF,$16,$F2,$3E,$07,$35,$27,$FB,$27
	.dc.b	$CF,$5C,$3F,$AA,$F3,$23,$C8,$BE,$9E,$FD,$3E,$1B,$7F,$DB,$63,$CB
	.dc.b	$AF,$F7,$26,$D1,$A6,$3E,$8F,$AF,$D3,$76,$6D,$FF,$77,$8E,$24,$7F
	.dc.b	$AD,$BD,$B6,$74,$FC,$03,$E6,$71,$3E,$8B,$FD,$19,$7F,$DA,$E7,$4F
	.dc.b	$F8,$0F,$E9,$AE,$FF,$01,$F7,$F7,$FC,$0F,$A6,$B5,$FE,$09,$70,$FF
	.dc.b	$6C,$7D,$A4,$7F,$DB,$DF,$4A,$F8,$3F,$B8,$D8,$FF,$31,$66,$FA,$AF
	.dc.b	$F6,$0C,$7F,$98,$A7,$59,$7F,$81,$BC,$A7,$8D,$F8,$56,$D6,$F1,$C4
	.dc.b	$FE,$2E,$9F,$54,$EF,$C3,$6F,$FC,$C5,$C4,$F1,$F3,$5F,$91,$F3,$5E
	.dc.b	$25,$33,$7F,$E7,$68,$CB,$FE,$6B,$E9,$79,$3E,$15,$F2,$36,$30,$C9
	.dc.b	$FE,$61,$A2,$B7,$18,$6C,$7F,$35,$6B,$FE,$A3,$1F,$64,$F8,$CF,$80
	.dc.b	$DA,$FF,$3F,$3F,$80,$9F,$F4,$51,$C2,$7E,$9F,$C0,$4A,$FA,$7C,$9F
	.dc.b	$0D,$FA,$51,$F0,$FC,$C9,$18,$3E,$13,$F7,$37,$F9,$1E,$15,$9F,$8C
	.dc.b	$E1,$FF,$28,$73,$5C,$BF,$5A,$FC,$7F,$0F,$E4,$7C,$75,$AF,$98,$E1
	.dc.b	$F8,$8F,$F0,$0A,$F9,$0D,$D8,$E7,$F7,$87,$C4,$FE,$0D,$B7,$F5,$FC
	.dc.b	$7C,$4F,$F0,$6F,$22,$4F,$F9,$3F,$C9,$B9,$7F,$02,$F4,$7F,$80,$98
	.dc.b	$EF,$FE,$39,$E4,$23,$F8,$7F,$95,$30,$FC,$EF,$8C,$F2,$F1,$8F,$F0
	.dc.b	$CF,$96,$FE,$16,$63,$2E,$6F,$F1,$8F,$32,$61,$FE,$C4,$FA,$CF,$CF
	.dc.b	$DF,$F9,$A3,$CB,$FE,$25,$3F,$E6,$5F,$37,$F8,$F9,$C2,$74,$FF,$18
	.dc.b	$CE,$4E,$CF,$F4,$13,$FC,$9B,$CD,$FE,$61,$F8,$57,$E1,$F4,$FF,$37
	.dc.b	$46,$30,$97,$FE,$D6,$AF,$EF,$51,$FE,$7A,$78,$47,$FB,$95,$CD,$7F
	.dc.b	$9D,$FC,$6F,$E6,$67,$F3,$DF,$AD,$FE,$69,$7E,$B8,$F0,$FC,$14,$C3
	.dc.b	$FD,$01,$16,$7E,$07,$9D,$FC,$F3,$CF,$FE,$7A,$DA,$FF,$43,$47,$F2
	.dc.b	$32,$3F,$D1,$1F,$E6,$79,$3E,$8F,$DE,$7C,$0F,$A4,$B5,$ED,$F7,$BE
	.dc.b	$19,$FC,$AB,$77,$CB,$62,$7E,$17,$37,$42,$3F,$B7,$C7,$F0,$0D,$FA
	.dc.b	$0F,$9B,$F5,$9F,$39,$B1,$F2,$FE,$87,$E8,$9D,$FC,$BB,$1F,$3F,$E8
	.dc.b	$7D,$24,$7B,$3E,$CF,$3F,$EC,$F9,$FF,$69,$B3,$FE,$FF,$FC,$3D,$E6
	.dc.b	$FF,$25,$6C,$7F,$AB,$7C,$1C,$4F,$E0,$BE,$7B,$E3,$92,$7D,$8F,$CA
	.dc.b	$F9,$DF,$E5,$DF,$88,$7F,$2B,$E9,$FC,$14,$63,$D6,$FE,$65,$F5,$CD
	.dc.b	$A9,$8E,$0B,$97,$E2,$F9,$7E,$A9,$1F,$94,$FE,$FF,$E0,$3F,$C9,$90
	.dc.b	$CB,$F8,$DC,$61,$FE,$3B,$83,$E7,$BF,$A1,$4E,$7C,$1F,$F0,$97,$3F
	.dc.b	$F8,$07,$0F,$CD,$7A,$3F,$D9,$BE,$BB,$F0,$91,$83,$F3,$DE,$8F,$E2
	.dc.b	$08,$FD,$D7,$F1,$7E,$87,$EA,$0F,$D4,$FD,$05,$9F,$D3,$75,$7E,$B0
	.dc.b	$FF,$0E,$7D,$0F,$E9,$23,$F4,$D1,$FB,$0D,$7F,$2A,$48,$FC,$53,$FE
	.dc.b	$0A,$FA,$0F,$54,$7C,$1F,$55,$18,$E1,$FE,$94,$BE,$83,$FC,$54,$E6
	.dc.b	$1F,$5A,$CF,$F9,$53,$70,$FE,$A6,$D9,$FF,$42,$35,$FE,$A1,$F5,$BF
	.dc.b	$CA,$9F,$F4,$43,$5C,$FA,$FF,$CD,$63,$11,$F5,$C7,$D1,$F9,$6F,$ED
	.dc.b	$72,$7F,$9F,$E7,$39,$FF,$79,$F6,$9F,$29,$FD,$EF,$0E,$15,$FD,$6B
	.dc.b	$EB,$7F,$C0,$EF,$F5,$CA,$7C,$8F,$F8,$06,$73,$8F,$59,$CE,$5E,$03
	.dc.b	$C8,$E2,$FF,$D4,$4D,$7F,$B7,$35,$DD,$FF,$8F,$38,$FF,$0F,$0B,$FF
	.dc.b	$2F,$46,$1F,$F3,$0F,$D6,$F5,$73,$E7,$7E,$08,$FA,$BF,$2F,$72,$F2
	.dc.b	$61,$F6,$A3,$F0,$FE,$7C,$9C,$F2,$BE,$B3,$FA,$FC,$63,$35,$6F,$0F
	.dc.b	$24,$FE,$1F,$9B,$E5,$BB,$9F,$C2,$F0,$FF,$9B,$FE,$B5,$E8,$FF,$96
	.dc.b	$1B,$5F,$E1,$16,$E1,$FE,$11,$CD,$18,$8F,$BD,$1F,$94,$F8,$BF,$C8
	.dc.b	$F3,$FE,$49,$B7,$EA,$CF,$F0,$B7,$F9,$8A,$FC,$F1,$27,$FC,$41,$A7
	.dc.b	$59,$FD,$03,$5F,$F1,$43,$70,$36,$FE,$85,$7C,$1F,$A6,$FD,$1F,$1F
	.dc.b	$E9,$1F,$F2,$37,$D6,$7E,$3B,$E4,$BF,$C6,$6D,$CB,$FD,$C8,$D7,$FC
	.dc.b	$60,$D7,$F3,$9F,$27,$FE,$85,$FA,$DF,$F1,$DB,$63,$FD,$D8,$D7,$FC
	.dc.b	$70,$D7,$F4,$93,$FE,$65,$F8,$EF,$0E,$FF,$E5,$09,$F0,$B5,$F0,$B5
	.dc.b	$FB,$9F,$2F,$FA,$DE,$8F,$F5,$76,$CF,$A8,$C6,$FF,$81,$6D,$FF,$06
	.dc.b	$BB,$C3,$FC,$29,$B6,$69,$8F,$A8,$CF,$D4,$7F,$95,$64,$D2,$5F,$F9
	.dc.b	$5E,$1F,$43,$C2,$3F,$01,$3D,$A2,$47,$E0,$DA,$FE,$60,$FE,$37,$F4
	.dc.b	$3E,$8D,$1F,$BD,$BE,$07,$CF,$F2,$52,$FF,$D1,$73,$96,$CF,$18,$C4
	.dc.b	$75,$AE,$73,$F0,$3E,$9B,$E0,$FD,$2F,$C2,$F9,$79,$7F,$97,$8C,$B4
	.dc.b	$87,$9F,$83,$F9,$9F,$C8,$7C,$AF,$99,$F8,$C4,$9F,$88,$FC,$9F,$96
	.dc.b	$D4,$7E,$8E,$3E,$8C,$7D,$4F,$C1,$FF,$01,$A7,$FB,$A5,$FF,$4D,$57
	.dc.b	$F0,$26,$BE,$17,$7C,$DF,$D1,$57,$D0,$7F,$A2,$DB,$3F,$E8,$B6,$BF
	.dc.b	$BB,$6B,$FB,$A7,$FC,$0E,$F7,$F8,$1F,$21,$E5,$F9,$8F,$C5,$C3,$EB
	.dc.b	$1D,$1F,$E9,$C8,$C3,$63,$31,$FE,$A0,$6B,$FE,$9B,$C9,$F8,$4F,$F9
	.dc.b	$9E,$3F,$D3,$C7,$F2,$D7,$0E,$DA,$FE,$85,$AF,$F4,$FF,$D5,$BE,$2C
	.dc.b	$FF,$4D,$6E,$99,$F8,$2F,$9D,$F4,$7E,$AE,$D7,$CC,$52,$63,$F3,$FE
	.dc.b	$8B,$EA,$7C,$DF,$CB,$24,$7C,$B7,$E7,$7C,$E7,$D9,$24,$7D,$7F,$9C
	.dc.b	$FB,$9F,$39,$F9,$BF,$DF,$1F,$31,$4C,$D7,$DD,$F9,$BF,$B0,$47,$EC
	.dc.b	$3F,$6F,$E6,$FF,$0D,$F3,$52,$7F,$09,$1F,$89,$F9,$AF,$C5,$78,$72
	.dc.b	$7F,$11,$1F,$8D,$FA,$3F,$C2,$12,$3F,$08,$FE,$1F,$DE,$FC,$91,$23
	.dc.b	$F2,$1F,$29,$E9,$F9,$7F,$CB,$71,$C9,$FF,$7D,$ED,$FE,$E0,$7F,$2D
	.dc.b	$F3,$7F,$91,$23,$F9,$17,$FD,$BD,$F0,$FF,$81,$E9,$FF,$0E,$36,$7F
	.dc.b	$DC,$8D,$F1,$BF,$C3,$1E,$B7,$C9,$F8,$BF,$E3,$33,$F2,$AB,$FB,$2B
	.dc.b	$5F,$DF,$63,$FE,$C4,$73,$F0,$5E,$02,$3F,$EE,$D9,$FF,$42,$37,$E0
	.dc.b	$3A,$91,$EB,$EF,$FE,$83,$DE,$75,$78,$BE,$6E,$57,$F7,$A3,$FD,$A3
	.dc.b	$1B,$EB,$FD,$25,$1A,$49,$FC,$F9,$FE,$D8,$EA,$E0,$F3,$9E,$0D,$8F
	.dc.b	$83,$F3,$5F,$0B,$E6,$E4,$FC,$28,$CB,$48,$7F,$17,$D1,$F2,$BF,$2B
	.dc.b	$E5,$BE,$5F,$CD,$CA,$F9,$71,$F3,$BA,$35,$18,$3E,$7B,$F4,$BE,$6A
	.dc.b	$47,$D2,$7F,$9C,$7A,$2F,$7A,$9E,$EF,$2D,$19,$AF,$DF,$C6,$27,$53
	.dc.b	$99,$FF,$08,$66,$F9,$9E,$71,$FE,$28,$BE,$0F,$E0,$A3,$F0,$BF,$33
	.dc.b	$F8,$6F,$96,$93,$F8,$68,$FC,$4F,$56,$A3,$07,$E2,$9F,$C6,$FC,$AC
	.dc.b	$8F,$C6,$BF,$90,$C7,$F2,$3F,$2B,$67,$F2,$31,$87,$F0,$5F,$8E,$7F
	.dc.b	$A4,$DF,$CA,$FC,$B7,$E6,$0F,$E5,$31,$98,$FE,$C0,$3E,$9F,$FC,$C8
	.dc.b	$FF,$9A,$4C,$78,$5F,$CB,$63,$31,$FE,$6D,$6B,$FE,$A5,$7F,$D0,$4F
	.dc.b	$E4,$DF,$F3,$76,$DF,$E6,$F6,$CF,$1F,$95,$FF,$01,$FE,$11,$8B,$78
	.dc.b	$88,$E7,$1F,$D7,$DB,$82,$FF,$D2,$91,$C6,$B8,$7C,$9F,$49,$97,$FE
	.dc.b	$5A,$A4,$F8,$3E,$4B,$C3,$CB,$2B,$51,$E4,$EE,$D4,$60,$E4,$FA,$3E
	.dc.b	$43,$D3,$F2,$12,$7B,$47,$F5,$6F,$BF,$FE,$05,$1F,$D3,$1A,$FC,$D6
	.dc.b	$BF,$E9,$76,$BF,$0B,$2F,$F1,$23,$FB,$D8,$E7,$AF,$FA,$65,$BA,$7C
	.dc.b	$2F,$5F,$E5,$FC,$6F,$CC,$F7,$A6,$1F,$06,$3E,$7F,$C6,$6A,$30,$7C
	.dc.b	$57,$E9,$FC,$67,$D4,$F8,$C9,$3F,$26,$3E,$BF,$C6,$6A,$30,$7C,$D7
	.dc.b	$ED,$FC,$5F,$F0,$CE,$5F,$B4,$DD,$CD,$31,$F1,$5C,$2F,$EB,$47,$E1
	.dc.b	$23,$F0,$BF,$17,$F8,$6F,$8B,$95,$F6,$E3,$2D,$21,$F8,$7F,$8B,$FC
	.dc.b	$6F,$8B,$E1,$24,$7F,$7E,$CB,$F1,$D1,$E7,$FC,$23,$FC,$37,$3F,$46
	.dc.b	$38,$BF,$EA,$28,$CC,$9F,$F1,$1C,$7F,$A4,$9A,$FF,$B9,$DA,$FE,$71
	.dc.b	$B1,$C7,$E3,$3F,$CC,$71,$FE,$06,$9D,$7C,$36,$27,$3F,$17,$AF,$8B
	.dc.b	$95,$E3,$8C,$B4,$87,$F9,$16,$DF,$11,$DB,$E2,$3D,$C4,$9F,$85,$28
	.dc.b	$F8,$5F,$11,$B9,$F8,$11,$E6,$8F,$F1,$8F,$C8,$BE,$27,$FD,$A5,$4F
	.dc.b	$A9,$FF,$13,$CA,$BF,$DE,$51,$86,$FB,$7F,$E9,$AB,$E3,$3C,$1E,$7B
	.dc.b	$6F,$00,$70,$DA,$E5,$E0,$6A,$79,$3C,$2F,$64,$9D,$D7,$9D,$B1,$98
	.dc.b	$EF,$8F,$A5,$E5,$3A,$38,$DF,$88,$8E,$0D,$47,$2C,$7D,$ED,$39,$9D
	.dc.b	$B1,$F8,$33,$C2,$17,$E1,$7C,$C7,$12,$9A,$9E,$2F,$3C,$7A,$BC,$AF
	.dc.b	$C2,$D3,$35,$F6,$66,$D7,$D0,$BF,$F6,$6D,$FD,$0B,$F9,$C7,$87,$0F
	.dc.b	$9A,$E0,$F1,$5F,$49,$EF,$87,$A7,$1D,$9C,$2F,$E3,$DF,$3C,$3E,$43
	.dc.b	$E9,$3A,$57,$C1,$9D,$C7,$63,$38,$FE,$43,$E4,$7F,$D2,$75,$EB,$94
	.dc.b	$62,$75,$5F,$80,$8F,$DA,$FC,$F7,$E3,$1B,$5C,$27,$FD,$03,$39,$6C
	.dc.b	$64,$FE,$C9,$1F,$8F,$C3,$FB,$37,$C1,$7F,$53,$9F,$88,$7E,$62,$74
	.dc.b	$FC,$B8,$D5,$7F,$A2,$27,$33,$F9,$E9,$CD,$78,$2B,$FD,$1F,$18,$9F
	.dc.b	$09,$F8,$20,$F8,$72,$E9,$19,$8C,$57,$33,$CF,$E1,$78,$27,$CC,$7A
	.dc.b	$27,$B3,$F8,$1C,$3F,$16,$F6,$9A,$A3,$33,$DA,$3C,$B7,$E8,$AD,$D6
	.dc.b	$9B,$3E,$56,$CE,$E3,$D8,$75,$3C,$91,$E7,$71,$86,$CF,$7C,$F1,$2E
	.dc.b	$1E,$E9,$E7,$7E,$16,$9E,$12,$D7,$39,$F8,$D1,$F5,$EF,$C1,$5C,$6F
	.dc.b	$AC,$E4,$FD,$AF,$27,$F4,$3C,$FF,$05,$B7,$E0,$E3,$FA,$37,$BD,$8C
	.dc.b	$7F,$A2,$55,$BF,$E8,$93,$6B,$75,$FC,$D2,$3F,$C5,$F1,$89,$FF,$1A
	.dc.b	$35,$FE,$A1,$EF,$71,$E2,$FF,$3F,$5F,$F5,$7F,$7B,$13,$23,$FA,$D7
	.dc.b	$C5,$5B,$FE,$8C,$F1,$F8,$BE,$5F,$15,$6B,$FB,$03,$87,$FA,$3B,$E4
	.dc.b	$C4,$D3,$1F,$17,$E3,$F8,$BF,$23,$6B,$FB,$54,$71,$23,$FD,$49,$E9
	.dc.b	$F1,$7E,$B3,$D9,$F1,$76,$7F,$B9,$47,$BF,$E2,$F1,$A7,$F5,$F7,$E1
	.dc.b	$7C,$5F,$CB,$FB,$CF,$E9,$37,$FD,$E6,$31,$52,$3F,$C0,$53,$A8,$F9
	.dc.b	$AD,$7F,$91,$70,$A3,$11,$FD,$52,$31,$D3,$DF,$E3,$7C,$0D,$BF,$EB
	.dc.b	$3C,$DF,$83,$C3,$FD,$E3,$18,$39,$2F,$8C,$E1,$8A,$B8,$47,$AF,$E3
	.dc.b	$3D,$9F,$19,$70,$E5,$7F,$E6,$8C,$13,$A9,$CE,$7E,$9E,$CC,$5F,$59
	.dc.b	$F0,$C7,$C7,$F8,$BF,$CA,$4B,$F0,$07,$A9,$9B,$FF,$67,$46,$1B,$1F
	.dc.b	$E8,$28,$C4,$FF,$A2,$4E,$51,$93,$F3,$79,$3E,$37,$EB,$7C,$6F,$D7
	.dc.b	$6E,$1E,$89,$7C,$67,$DB,$F8,$DF,$B8,$D9,$D4,$7D,$FF,$8D,$C6,$5E
	.dc.b	$D3,$E2,$3F,$83,$F8,$DF,$C2,$7C,$6D,$BE,$CE,$1F,$07,$C6,$5B,$E5
	.dc.b	$C3,$F8,$18,$F8,$BE,$4B,$DF,$BF,$01,$86,$E1,$FE,$F3,$94,$FF,$A9
	.dc.b	$9B,$1F,$EF,$06,$E1,$F1,$DA,$FF,$AC,$57,$BF,$F9,$BF,$8D,$F0,$B6
	.dc.b	$BE,$37,$C6,$FE,$7F,$E3,$7F,$40,$D8,$F9,$5F,$19,$FA,$33,$18,$FB
	.dc.b	$38,$D1,$FA,$7F,$8D,$FD,$47,$C6,$DB,$FA,$4E,$1F,$C5,$F8,$7F,$EB
	.dc.b	$17,$26,$0F,$E3,$4B,$E2,$FF,$8F,$78,$2F,$DF,$5E,$3E,$A7,$31,$8B
	.dc.b	$FF,$76,$37,$4F,$F6,$95,$F1,$BF,$F6,$8B,$6B,$FD,$F3,$5B,$AF,$DB
	.dc.b	$E1,$FE,$BE,$9F,$3F,$C6,$76,$C3,$CA,$F9,$BE,$2F,$CF,$F1,$96,$BD
	.dc.b	$AE,$1F,$7F,$C6,$5C,$7F,$07,$8A,$3D,$FC,$F2,$BE,$2F,$3D,$BD,$CF
	.dc.b	$C4,$E5,$C6,$DF,$D1,$71,$FB,$B8,$7E,$EA,$DF,$94,$7C,$1A,$FF,$06
	.dc.b	$46,$2B,$F6,$B3,$9A,$FF,$1B,$28,$C9,$FE,$F4,$71,$1F,$B3,$9E,$4D
	.dc.b	$8C,$C7,$ED,$D7,$97,$F6,$BE,$33,$BB,$6B,$F0,$3E,$27,$DE,$3E,$FF
	.dc.b	$C6,$58,$FC,$2F,$C5,$FF,$58,$C8,$D2,$47,$F0,$5C,$DF,$BF,$3C,$9F
	.dc.b	$C4,$FC,$6F,$E2,$23,$15,$F9,$08,$FC,$6F,$C6,$C8,$FC,$9F,$C6,$5E
	.dc.b	$0F,$AB,$F1,$9E,$12,$4F,$FA,$FF,$44,$7E,$2B,$17,$91,$F9,$8F,$8B
	.dc.b	$F0,$36,$FF,$3F,$F1,$9F,$9C,$3F,$3B,$F1,$96,$3F,$47,$F1,$9E,$F2
	.dc.b	$3D,$DF,$EF,$3F,$83,$FF,$95,$23,$F9,$9B,$FE,$9B,$6F,$93,$FE,$01
	.dc.b	$8C,$6B,$FC,$41,$7C,$6E,$7E,$33,$4D,$BE,$7F,$1B,$D8,$8F,$FB,$76
	.dc.b	$91,$E2,$F8,$DF,$D9,$FE,$FF,$FB,$9C,$7B,$46,$2F,$C5,$F1,$D1,$F3
	.dc.b	$FC,$75,$AF,$27,$C5,$7B,$09,$3D,$DF,$DA,$FE,$21,$F8,$08,$FE,$AB
	.dc.b	$D6,$8C,$57,$B7,$E2,$FF,$82,$C6,$27,$FC,$52,$B9,$72,$7C,$5F,$8A
	.dc.b	$B8,$7C,$AF,$8B,$F9,$A4,$7E,$03,$FA,$8C,$7F,$D3,$5F,$49,$FC,$7A
	.dc.b	$7F,$44,$D8,$FE,$AC,$D7,$FD,$40,$1C,$3F,$DF,$F8,$5F,$19,$F5,$CF
	.dc.b	$B1,$F1,$96,$FE,$EF,$C5,$FE,$00,$8F,$D3,$7F,$07,$F1,$7F,$7F,$E2
	.dc.b	$FF,$00,$D9,$FC,$2F,$C5,$FE,$0F,$D3,$FC,$23,$63,$F1,$31,$F8,$C2
	.dc.b	$3F,$7D,$FC,$7F,$85,$F8,$B3,$F1,$9E,$55,$9F,$C8,$C7,$F9,$3F,$F0
	.dc.b	$9F,$F1,$07,$27,$F9,$5F,$6F,$E0,$6F,$F9,$66,$33,$5F,$E1,$AE,$AF
	.dc.b	$E2,$FC,$1F,$DB,$69,$18,$78,$7C,$5F,$E3,$BC,$F9,$7F,$E1,$2C,$3F
	.dc.b	$C4,$38,$71,$C3,$84,$FF,$8A,$E7,$38,$78,$7E,$2F,$CC,$79,$FE,$2E
	.dc.b	$D7,$38,$F5,$7C,$5E,$31,$FE,$EF,$87,$F8,$8A,$BC,$13,$FE,$6A,$8C
	.dc.b	$4F,$F9,$A9,$AF,$F8,$F9,$AF,$8F,$E2,$FE,$31,$F1,$FC,$EB,$7E,$B8
	.dc.b	$F9,$5F,$13,$8A,$F0,$BF,$E2,$AF,$9D,$7B,$5F,$EB,$69,$FF,$12,$46
	.dc.b	$38,$3F,$9A,$36,$33,$3B,$7F,$27,$F2,$4F,$05,$F3,$DF,$CE,$E3,$0D
	.dc.b	$8C,$CF,$F8,$C6,$72,$F8,$3D,$CF,$F7,$2B,$7C,$BF,$0C,$62,$3F,$43
	.dc.b	$8A,$31,$1F,$57,$E3,$4F,$DB,$F8,$DA,$FD,$06,$F2,$FE,$26,$6F,$07
	.dc.b	$1A,$FC,$14,$61,$7C,$53,$8D,$39,$8C,$7C,$6D,$9F,$E6,$F5,$FE,$4D
	.dc.b	$9E,$1E,$9F,$F5,$D6,$C6,$67,$6F,$DF,$F8,$DE,$55,$23,$F2,$91,$D4
	.dc.b	$8E,$0F,$CC,$7C,$6F,$87,$E3,$7C,$5F,$1B,$27,$F3,$91,$96,$90,$FC
	.dc.b	$9F,$C6,$7B,$F6,$7C,$EF,$F8,$8A,$7F,$49,$F1,$DE,$D3,$51,$98,$FD
	.dc.b	$73,$67,$FD,$9A,$DE,$0F,$F4,$6B,$67,$E5,$EB,$FE,$3C,$7F,$C1,$9F
	.dc.b	$9B,$F3,$7F,$8E,$B5,$FB,$2D,$7E,$EB,$77,$FA,$9F,$1B,$F5,$E5,$3D
	.dc.b	$DB,$E3,$FD,$59,$FA,$DF,$1B,$27,$F8,$03,$F6,$FF,$0A,$FF,$0D,$E9
	.dc.b	$FE,$1C,$C3,$FC,$31,$5F,$E1,$56,$BF,$E8,$96,$BE,$8F,$23,$EF,$57
	.dc.b	$F9,$36,$3F,$D3,$AD,$7F,$CE,$ED,$AF,$F7,$92,$F8,$CF,$57,$CE,$3D
	.dc.b	$DA,$4B,$FE,$AF,$FB,$43,$DB,$F8,$1F,$3F,$FE,$A2,$9F,$F1,$53,$5F
	.dc.b	$F4,$F3,$5F,$E9,$2D,$7F,$D8,$3F,$3E,$F2,$3F,$04,$D8,$FE,$CA,$D7
	.dc.b	$F9,$63,$FE,$23,$7F,$B4,$F9,$AF,$88,$FC,$53,$63,$31,$F3,$7C,$8F
	.dc.b	$AC,$3D,$2F,$F2,$04,$FF,$12,$6C,$66,$3F,$C8,$8D,$7F,$D7,$1E,$85
	.dc.b	$E1,$FF,$28,$B7,$AB,$F7,$ED,$9F,$DC,$F1,$7F,$07,$7E,$9F,$C8,$79
	.dc.b	$BE,$26,$46,$0F,$F2,$F7,$E4,$DF,$C9,$1F,$F2,$A3,$5F,$F6,$43,$5F
	.dc.b	$EE,$6D,$7F,$DE,$15,$FE,$B1,$7F,$C1,$BF,$41,$F2,$72,$65,$FE,$F0
	.dc.b	$8C,$3C,$3E,$47,$F2,$5F,$25,$F9,$3D,$26,$1F,$8A,$8F,$CF,$65,$EB
	.dc.b	$8F,$1F,$97,$F2,$A3,$11,$E4,$8F,$C9,$FC,$8F,$E7,$49,$1F,$94,$F9
	.dc.b	$1F,$E0,$19,$BC,$9F,$D1,$7C,$8C,$AF,$4C,$7E,$B7,$F2,$57,$83,$F6
	.dc.b	$1B,$3F,$E7,$96,$BF,$EE,$D6,$BF,$CC,$2D,$E5,$78,$09,$1F,$A6,$F9
	.dc.b	$2F,$1F,$93,$FB,$3E,$99,$3F,$2A,$3F,$7D,$F1,$5F,$E2,$98,$FE,$7A
	.dc.b	$D7,$FA,$0B,$53,$9E,$56,$3E,$67,$E9,$C9,$CE,$7F,$10,$FE,$09,$CF
	.dc.b	$FC,$17,$E4,$A5,$7E,$F1,$FE,$21,$D3,$FB,$78,$FF,$53,$3F,$E5,$C7
	.dc.b	$FD,$38,$D7,$FC,$72,$7F,$AD,$7E,$33,$F4,$31,$98,$FF,$1E,$BE,$87
	.dc.b	$9C,$76,$3B,$E5,$FC,$5F,$E4,$BF,$8C,$7C,$54,$AF,$E1,$F1,$96,$90
	.dc.b	$FE,$23,$F2,$3F,$C7,$FE,$46,$47,$F1,$67,$F9,$17,$C8,$C8,$C1,$0F
	.dc.b	$E3,$2F,$F6,$18,$FE,$E9,$B7,$CA,$F3,$D0,$F9,$DD,$1B,$F6,$F8,$7C
	.dc.b	$95,$C7,$FB,$BE,$1C,$6F,$9F,$85,$FD,$EB,$1E,$77,$DF,$CC,$E3,$87
	.dc.b	$0A,$D5,$FF,$A4,$A3,$07,$F7,$87,$EA,$3F,$E4,$3C,$DE,$17,$A7,$86
	.dc.b	$4F,$FA,$0A,$72,$F2,$CB,$D7,$C3,$FD,$96,$73,$3C,$EF,$E1,$4F,$C3
	.dc.b	$F2,$FE,$27,$C9,$4A,$E5,$97,$F3,$76,$FC,$97,$C9,$23,$AF,$80,$F9
	.dc.b	$9F,$23,$23,$8B,$FD,$77,$E8,$7F,$C2,$31,$FE,$F3,$5F,$01,$FC,$A9
	.dc.b	$B1,$FE,$EB,$6B,$F9,$06,$BF,$E8,$A5,$EF,$F5,$99,$3D,$63,$EF,$7C
	.dc.b	$94,$8F,$5B,$F8,$1F,$91,$91,$82,$1F,$AB,$E3,$FA,$B1,$FD,$D7,$46
	.dc.b	$BF,$89,$F9,$19,$7F,$60,$FC,$4C,$7E,$37,$E4,$78,$32,$3E,$5C,$7E
	.dc.b	$47,$E4,$24,$7C,$E7,$F2,$9F,$23,$FE,$03,$FA,$1F,$ED,$93,$FB,$00
	.dc.b	$E1,$FD,$9B,$5F,$F5,$6B,$5F,$C6,$FA,$BC,$FD,$C9,$1F,$72,$3F,$09
	.dc.b	$E2,$FE,$8F,$E0,$24,$FE,$2E,$3F,$4D,$EF,$6A,$30,$7E,$11,$FC,$67
	.dc.b	$89,$FA,$BF,$92,$93,$F9,$38,$FF,$27,$7D,$C7,$FB,$03,$C3,$FE,$2E
	.dc.b	$3B,$D3,$CA,$BF,$31,$5B,$E5,$FD,$6D,$7F,$B5,$E7,$F8,$8D,$F1,$F1
	.dc.b	$FE,$23,$63,$33,$B7,$FD,$9F,$E0,$B0,$FF,$40,$57,$8A,$B4,$DB,$F3
	.dc.b	$FC,$6F,$E9,$F9,$FD,$3F,$3B,$32,$FD,$74,$65,$A4,$3F,$10,$7F,$9C
	.dc.b	$31,$7C,$97,$F9,$57,$E6,$7F,$CF,$F3,$FC,$21,$AF,$F3,$96,$E4,$73
	.dc.b	$A7,$A7,$E5,$39,$79,$52,$FE,$3C,$7C,$EF,$94,$FF,$3B,$64,$69,$27
	.dc.b	$FC,$F9,$1F,$DE,$1F,$34,$71,$7E,$8F,$C1,$FD,$8F,$94,$FB,$3F,$29
	.dc.b	$27,$EA,$C6,$5A,$43,$E0,$E5,$F3,$FE,$47,$F0,$3E,$2F,$90,$93,$E2
	.dc.b	$7C,$FE,$2E,$A3,$F0,$31,$F8,$88,$FF,$32,$7C,$9B,$CA,$FE,$BD,$4F
	.dc.b	$9F,$FF,$1E,$4F,$ED,$23,$FD,$53,$18,$F1,$7F,$D5,$2E,$AB,$8B,$78
	.dc.b	$FF,$D5,$FE,$96,$EE,$EF,$FE,$B0,$C7,$A4,$7F,$43,$8C,$61,$E2,$AE
	.dc.b	$53,$E3,$9E,$18,$F8,$1B,$1E,$47,$8E,$16,$3F,$D7,$F8,$7B,$63,$10
	.dc.b	$8F,$0E,$3E,$58,$EC,$70,$C7,$D7,$7D,$23,$CC,$74,$C3,$FD,$93,$8F
	.dc.b	$13,$CE,$7A,$2F,$A1,$2B,$93,$97,$D3,$5E,$03,$C9,$5E,$D7,$C5,$8F
	.dc.b	$A3,$1F,$54,$6D,$F4,$D7,$2B,$E9,$3C,$0F,$39,$D0,$FA,$27,$A4,$A7
	.dc.b	$C3,$C7,$E8,$4F,$98,$E8,$48,$E9,$3E,$E8,$F4,$E3,$F0,$5F,$8C,$E6
	.dc.b	$7B,$1E,$2B,$FB,$F1,$98,$F0,$61,$FE,$F4,$E3,$FF,$02,$47,$A5,$AE
	.dc.b	$83,$8A,$F8,$6E,$A3,$FD,$FB,$F4,$1C,$EF,$FC,$56,$D8,$FF,$7F,$87
	.dc.b	$08,$E2,$76,$8E,$2F,$CA,$1D,$3F,$E4,$48,$FF,$50,$47,$06,$C6,$71
	.dc.b	$E3,$18,$EC,$F9,$F5,$B9,$F1,$D6,$E3,$CB,$7F,$96,$7B,$D5,$7D,$07
	.dc.b	$E1,$CC,$94,$CD,$F7,$7D,$06,$2F,$BC,$F8,$A3,$9E,$7B,$3D,$6E,$27
	.dc.b	$6F,$5B,$F2,$61,$FE,$64,$AD,$C6,$2F,$CC,$DA,$E0,$79,$EB,$85,$FC
	.dc.b	$43,$AE,$1C,$A7,$E7,$4F,$C3,$7E,$A1,$E3,$2B,$C9,$F8,$E7,$5B,$F0
	.dc.b	$B4,$F0,$E1,$ED,$BF,$04,$72,$9F,$85,$87,$D4,$BF,$04,$FE,$0A,$F8
	.dc.b	$3D,$9B,$C3,$FD,$7F,$E5,$3F,$B7,$E1,$FE,$53,$E6,$FF,$53,$3F,$DB
	.dc.b	$A3,$37,$FD,$5F,$DF,$B7,$FD,$9D,$FF,$4B,$36,$BF,$A4,$C6,$2F,$9D
	.dc.b	$7F,$6E,$5F,$29,$C9,$B5,$FD,$F3,$E5,$31,$32,$3F,$BF,$FC,$05,$BF
	.dc.b	$E9,$37,$FE,$AE,$9F,$F0,$1F,$CA,$73,$F7,$FA,$36,$3F,$C0,$F1,$E3
	.dc.b	$23,$FD,$79,$F3,$7C,$EF,$BB,$97,$FB,$67,$B5,$E5,$8C,$78,$08,$A7
	.dc.b	$FA,$D6,$77,$7F,$E1,$09,$D3,$67,$51,$FE,$B6,$8F,$F3,$CC,$FF,$79
	.dc.b	$F6,$3E,$17,$C9,$F6,$64,$FF,$87,$BE,$02,$EB,$F0,$1F,$A3,$F7,$5F
	.dc.b	$EE,$E9,$E7,$8F,$F9,$41,$FF,$3E,$46,$6B,$FC,$1F,$8F,$F8,$1F,$B7
	.dc.b	$D6,$79,$09,$5E,$97,$ED,$D3,$E7,$BE,$0F,$B7,$E0,$6D,$78,$23,$C7
	.dc.b	$E9,$63,$1F,$EF,$B6,$8F,$3F,$C7,$FC,$AF,$6A,$E1,$E8,$70,$FA,$3E
	.dc.b	$2E,$E1,$C9,$F7,$FC,$1F,$E4,$A3,$F9,$4B,$89,$FF,$36,$78,$48,$D3
	.dc.b	$63,$FC,$C8,$D9,$FF,$6A,$B5,$FE,$56,$D7,$FD,$69,$93,$E3,$BE,$79
	.dc.b	$F4,$3E,$3A,$DF,$CE,$F8,$2F,$27,$EF,$CF,$8C,$FA,$B1,$86,$4F,$D4
	.dc.b	$F8,$EB,$97,$C6,$7E,$DF,$C7,$7D,$CA,$B5,$F6,$9C,$34,$FB,$B1,$8A
	.dc.b	$FF,$25,$9F,$D3,$3E,$27,$E0,$FA,$BF,$09,$B5,$F1,$A0,$FC,$46,$CF
	.dc.b	$C4,$6B,$F1,$1F,$C1,$FC,$6F,$E4,$0F,$9B,$EC,$DA,$FC,$7F,$C6,$FE
	.dc.b	$50,$FC,$AF,$BD,$63,$F2,$7F,$19,$F9,$8F,$77,$F3,$2D,$8F,$CB,$FB
	.dc.b	$BF,$AE,$F8,$1F,$CE,$B6,$3F,$3B,$1F,$A0,$D7,$1A,$FE,$39,$FD,$1E
	.dc.b	$BF,$A4,$D2,$DF,$E8,$9C,$3F,$ED,$FF,$B7,$FE,$E2,$E4,$C7,$3F,$F1
	.dc.b	$C8,$CE,$1F,$E5,$67,$FD,$FD,$3A,$39,$7A,$9B,$F8,$DE,$1E,$3D,$9E
	.dc.b	$CE,$1A,$78,$63,$17,$D7,$E3,$7C,$07,$87,$C7,$B5,$E7,$8F,$27,$C6
	.dc.b	$E2,$BC,$8F,$E8,$FF,$2D,$FF,$44,$D7,$FB,$35,$7C,$0E,$7E,$0B,$4C
	.dc.b	$9F,$F6,$34,$E9,$B1,$F8,$0B,$FE,$AF,$39,$7F,$D4,$BA,$FF,$0D,$7E
	.dc.b	$27,$91,$E1,$F8,$2F,$90,$FF,$29,$FA,$0F,$F4,$66,$3F,$C7,$9B,$5F
	.dc.b	$2F,$DE,$D5,$FE,$EF,$DC,$78,$7E,$1A,$FF,$10,$B5,$FF,$0F,$36,$37
	.dc.b	$87,$FA,$FB,$07,$C1,$FD,$9F,$33,$C4,$D9,$F1,$7C,$0F,$94,$FB,$DF
	.dc.b	$1F,$63,$FD,$E9,$83,$C6,$EC,$48,$FF,$76,$46,$5F,$07,$B7,$F8,$7F
	.dc.b	$37,$E0,$36,$3F,$1D,$F1,$DF,$8B,$3F,$19,$F1,$D6,$3F,$C6,$DE,$9F
	.dc.b	$EB,$BF,$2C,$7A,$5E,$F6,$CF,$FB,$ED,$BC,$5F,$E4,$00,$E1,$FF,$06
	.dc.b	$FB,$EE,$77,$C7,$FE,$6F,$E3,$FF,$38,$DC,$3F,$45,$F1,$FF,$5F,$DB
	.dc.b	$FD,$03,$63,$CA,$7F,$86,$32,$7A,$FE,$52,$47,$95,$FF,$00,$E2,$F2
	.dc.b	$BF,$B5,$7B,$6E,$6F,$D6,$46,$1F,$ED,$DF,$43,$FC,$9F,$0F,$EC,$6D
	.dc.b	$DF,$F6,$1F,$1F,$AF,$57,$FA,$07,$E1,$CE,$6F,$C2,$37,$0F,$89,$68
	.dc.b	$C3,$F1,$DA,$FF,$8C,$B9,$5A,$3A,$3C,$E7,$F0,$1F,$90,$B8,$7C,$2B
	.dc.b	$73,$3E,$87,$84,$E6,$31,$7F,$C2,$7E,$47,$FC,$4B,$5A,$BF,$E6,$BF
	.dc.b	$44,$F8,$3F,$CE,$37,$1F,$4B,$67,$EF,$35,$F4,$7E,$B3,$F7,$27,$FC
	.dc.b	$12,$D8,$FF,$75,$F4,$FF,$A0,$67,$3D,$5F,$E3,$68,$FF,$36,$EF,$FC
	.dc.b	$FB,$54,$62,$BF,$D1,$11,$8F,$93,$E1,$F1,$DF,$DB,$FE,$13,$ED,$57
	.dc.b	$CE,$6E,$5F,$93,$6B,$FE,$94,$6B,$D3,$0F,$E2,$AF,$9B,$CE,$F2,$79
	.dc.b	$DE,$57,$CD,$E5,$F7,$E3,$B8,$FD,$1B,$FF,$05,$46,$63,$C4,$FF,$05
	.dc.b	$C7,$E7,$4F,$86,$3F,$C6,$9F,$35,$FC,$1A,$3F,$C4,$2D,$DB,$E1,$66
	.dc.b	$F9,$0F,$F2,$04,$E7,$1F,$DA,$B6,$FF,$03,$4F,$94,$DF,$9D,$F3,$5F
	.dc.b	$9D,$E6,$4C,$3F,$09,$4F,$94,$C6,$1F,$5A,$BE,$A7,$9B,$8D,$3C,$6F
	.dc.b	$D7,$F9,$49,$8F,$2F,$92,$B8,$7D,$E9,$FB,$9E,$6E,$30,$F5,$B8,$BF
	.dc.b	$E0,$DC,$FF,$7A,$3F,$CE,$BA,$7F,$9C,$B0,$FE,$0E,$D7,$EB,$7C,$8F
	.dc.b	$E1,$CF,$C4,$7C,$95,$C3,$D7,$1F,$8B,$F9,$2C,$57,$E4,$1F,$F1,$A7
	.dc.b	$D1,$7F,$1A,$EA,$FD,$9B,$67,$F8,$C3,$5F,$F6,$CB,$5F,$F5,$2F,$CA
	.dc.b	$BD,$57,$D1,$FE,$F2,$30,$D8,$CC,$FF,$23,$9C,$BD,$BC,$FF,$F7,$6B
	.dc.b	$7E,$07,$C7,$18,$8F,$0B,$5F,$CB,$7C,$A7,$F0,$5D,$7F,$31,$F2,$96
	.dc.b	$7F,$02,$DE,$2F,$D4,$D5,$E1,$FD,$E5,$7E,$02,$30,$67,$C6,$D4,$FF
	.dc.b	$16,$AF,$D8,$B4,$FD,$9D,$7D,$86,$BF,$EF,$E8,$FD,$BF,$95,$CA,$3F
	.dc.b	$1B,$3E,$18,$F2,$7D,$9F,$F8,$E3,$9B,$F3,$B1,$9A,$FE,$2D,$39,$CF
	.dc.b	$F7,$EF,$F9,$F6,$33,$E0,$F2,$39,$FC,$A7,$C1,$C1,$F4,$8F,$5F,$F9
	.dc.b	$7C,$61,$B1,$9B,$FE,$AD,$39,$7F,$27,$87,$F5,$E7,$C7,$C1,$E8,$3D
	.dc.b	$3E,$CC,$8F,$3B,$55,$7F,$19,$AF,$F3,$F4,$7F,$AA,$B4,$7B,$3E,$46
	.dc.b	$C7,$89,$AF,$FB,$C8,$E6,$E1,$FE,$D3,$3F,$E7,$F7,$A4,$FC,$19,$E3
	.dc.b	$5E,$1F,$93,$F9,$FE,$24,$BF,$9A,$FD,$2F,$77,$E4,$78,$CF,$90,$F8
	.dc.b	$ED,$AF,$8A,$D7,$C9,$82,$31,$1E,$4F,$92,$FB,$BF,$25,$23,$ED,$C7
	.dc.b	$97,$DC,$91,$F0,$DF,$C1,$FB,$3F,$99,$87,$D1,$7F,$B6,$5F,$EC,$ED
	.dc.b	$7F,$C1,$8D,$7E,$33,$51,$3E,$85,$87,$F7,$08,$46,$1F,$EE,$6D,$7F
	.dc.b	$C3,$0D,$7F,$B9,$6E,$F7,$3E,$71,$23,$F2,$FA,$FA,$76,$7B,$5D,$DB
	.dc.b	$3F,$E0,$06,$BF,$E2,$56,$BF,$E2,$4F,$AB,$7C,$37,$EE,$5B,$1F,$DE
	.dc.b	$5A,$FE,$3B,$B3,$F4,$BF,$2B,$FA,$CF,$95,$93,$FA,$A7,$F5,$FF,$29
	.dc.b	$F8,$5F,$7F,$F2,$04,$9F,$C8,$3F,$BF,$EA,$7C,$AF,$ED,$89,$3F,$B4
	.dc.b	$F9,$5F,$DC,$FC,$AC,$8F,$DB,$BF,$BC,$F9,$5F,$D9,$7F,$00,$3E,$03
	.dc.b	$F6,$2D,$AF,$D7,$B5,$FC,$5B,$5F,$CE,$FC,$B7,$F1,$9F,$C1,$FF,$91
	.dc.b	$77,$FF,$80,$DB,$C3,$FC,$12,$7F,$33,$A6,$FE,$5E,$61,$FE,$23,$C3
	.dc.b	$F2,$D3,$E4,$F8,$1F,$19,$27,$C6,$FF,$1F,$C5,$F2,$DE,$12,$4F,$F8
	.dc.b	$A6,$3F,$25,$7F,$ED,$23,$F4,$5B,$F9,$FE,$56,$57,$87,$C3,$4F,$C4
	.dc.b	$F7,$FE,$19,$27,$E1,$BE,$1F,$96,$F8,$04,$8F,$07,$CB,$7C,$3E,$B9
	.dc.b	$1E,$37,$F5,$1F,$95,$39,$AF,$37,$E7,$E0,$D2,$57,$F9,$4A,$1C,$7C
	.dc.b	$23,$FC,$35,$1C,$1B,$1F,$4F,$E5,$A4,$7C,$37,$EB,$7B,$32,$3E,$33
	.dc.b	$5F,$CB,$1F,$E5,$4E,$1F,$F3,$73,$F2,$FE,$1F,$EF,$11,$F9,$DF,$1F
	.dc.b	$F8,$12,$38,$3E,$97,$CA,$7E,$17,$E3,$24,$7D,$57,$F1,$1E,$84,$8F
	.dc.b	$B0,$D7,$C7,$EC,$31,$FE,$85,$18,$6F,$3F,$E3,$FD,$5F,$C9,$1F,$93
	.dc.b	$F3,$FF,$29,$E5,$7E,$57,$C9,$98,$7E,$16,$3F,$31,$E2,$6A,$30,$7E
	.dc.b	$25,$FE,$E9,$F8,$2F,$F2,$AF,$0B,$FC,$B6,$D9,$FF,$3C,$B7,$42,$F8
	.dc.b	$1E,$3E,$2F,$26,$57,$F9,$F6,$7F,$59,$AF,$F9,$A5,$7C,$67,$80,$93
	.dc.b	$FE,$57,$F9,$CF,$E3,$8D,$D7,$F9,$34,$7F,$14,$8F,$E5,$61,$C3,$FC
	.dc.b	$B9,$AE,$74,$FE,$75,$1F,$C7,$B0,$69,$CA,$30,$D9,$F8,$7F,$1D,$2F
	.dc.b	$D4,$FC,$6F,$16,$47,$B5,$AF,$DA,$CF,$FC,$C8,$FC,$1F,$8D,$F9,$BE
	.dc.b	$14,$8C,$1F,$17,$E3,$7E,$8C,$C8,$C1,$F2,$7E,$37,$EA,$91,$F9,$9F
	.dc.b	$19,$FD,$57,$EB,$DF,$15,$FD,$2D,$B3,$FE,$BC,$6B,$FE,$E1,$6B,$FE
	.dc.b	$E0,$C9,$F1,$DC,$EA,$4F,$D9,$8F,$C2,$FC,$77,$E1,$BE,$36,$4F,$DD
	.dc.b	$8F,$C4,$FC,$76,$A3,$07,$E0,$5F,$C6,$DF,$E3,$BE,$26,$4F,$E1,$63
	.dc.b	$F2,$3F,$15,$A8,$C1,$F8,$97,$F9,$CC,$7F,$38,$24,$7F,$38,$7F,$C8
	.dc.b	$99,$7F,$92,$A3,$F7,$BC,$CF,$53,$E0,$BF,$C9,$5B,$CD,$F9,$EF,$8A
	.dc.b	$FD,$0F,$C5,$7E,$8B,$DE,$98,$7E,$66,$32,$D2,$1F,$B8,$E9,$FF,$22
	.dc.b	$3F,$E2,$EC,$1A,$FF,$4E,$24,$7F,$4E,$7F,$CB,$53,$FE,$49,$8F,$3C
	.dc.b	$F0,$7F,$A3,$35,$ED,$F1,$3F,$E6,$99,$7C,$57,$22,$4F,$FA,$62,$32
	.dc.b	$DF,$88,$E1,$32,$3F,$DE,$13,$FE,$4B,$7F,$88,$FD,$73,$DB,$ED,$F1
	.dc.b	$52,$7F,$DE,$D3,$96,$CF,$F9,$EE,$31,$1C,$A3,$1E,$DF,$9F,$C1,$F7
	.dc.b	$92,$7C,$CF,$C2,$F8,$AD,$47,$92,$3D,$D1,$FE,$31,$F8,$D7,$A7,$FE
	.dc.b	$5A,$A7,$D0,$FE,$B9,$4F,$F8,$4A,$30,$DF,$91,$FF,$52,$D7,$19,$CB
	.dc.b	$CF,$2D,$FB,$5F,$C3,$1B,$87,$F8,$E1,$DF,$45,$C3,$9B,$FE,$3C,$C3
	.dc.b	$75,$E4,$9F,$F5,$9D,$FF,$16,$9E,$33,$93,$FC,$87,$8F,$B2,$F9,$5F
	.dc.b	$09,$A6,$AF,$94,$78,$23,$91,$F0,$31,$F5,$47,$18,$FB,$4D,$C7,$C2
	.dc.b	$78,$B1,$F3,$92,$3A,$39,$79,$39,$9F,$3D,$7C,$99,$D9,$A2,$46,$B0
	.dc.b	$F4,$BE,$57,$E4,$3E,$B2,$9F,$63,$0F,$A1,$3E,$DC,$3C,$53,$B3,$DF
	.dc.b	$8F,$DB,$7B,$18,$C3,$E0,$BB,$9F,$A6,$7C,$0A,$E8,$7C,$32,$47,$28
	.dc.b	$FA,$BF,$1F,$F9,$AB,$FA,$92,$8C,$5F,$E4,$5B,$8F,$F9,$F1,$FD,$0C
	.dc.b	$FF,$9F,$BE,$6B,$18,$7F,$A9,$E1,$18,$7F,$D5,$4D,$7F,$AD,$B5,$E4
	.dc.b	$7D,$A3,$FA,$FF,$99,$FD,$61,$C1,$C0,$A7,$F7,$3E,$BF,$9A,$F2,$AF
	.dc.b	$D3,$39,$36,$79,$0E,$33,$D6,$37,$19,$D7,$8C,$F8,$5C,$DF,$DE,$9E
	.dc.b	$91,$87,$9D,$F8,$9E,$D7,$CA,$31,$5C,$6F,$A6,$9F,$1E,$31,$3D,$6F
	.dc.b	$C5,$3F,$ED,$1A,$DC,$7C,$A3,$F4,$67,$8A,$39,$38,$8D,$BF,$1F,$0F
	.dc.b	$13,$C0,$F2,$CF,$5A,$F2,$C7,$A8,$F8,$F7,$E4,$78,$36,$BE,$01,$D0
	.dc.b	$B5,$CE,$76,$73,$C3,$E3,$1E,$23,$6F,$DE,$BF,$81,$7D,$6B,$FD,$93
	.dc.b	$E7,$39,$1F,$1F,$FE,$7F,$BC,$C3,$E2,$71,$8A,$F8,$5E,$1E,$53,$6B
	.dc.b	$85,$FF,$49,$8F,$C7,$5F,$F8,$BA,$B7,$39,$3F,$38,$B9,$5E,$B7,$56
	.dc.b	$CF,$F4,$8F,$5F,$13,$23,$FA,$8F,$C7,$DC,$3F,$B5,$3E,$3F,$8D,$F2
	.dc.b	$7C,$7D,$AF,$EA,$EE,$1F,$EB,$CF,$9A,$BE,$0B,$5F,$81,$A2,$31,$9F
	.dc.b	$C9,$F8,$DF,$8B,$CB,$FE,$05,$F1,$FE,$14,$63,$C1,$45,$3A,$4E,$F2
	.dc.b	$FF,$63,$C6,$27,$71,$FD,$C7,$E3,$3E,$61,$F3,$7E,$32,$C7,$F7,$08
	.dc.b	$FA,$1F,$19,$8D,$7F,$C1,$4F,$D3,$F8,$BF,$17,$93,$6F,$FC,$06,$E1
	.dc.b	$FF,$6B,$BE,$12,$3F,$E0,$1F,$89,$FB,$9F,$0F,$F7,$5B,$1F,$E0,$08
	.dc.b	$FC,$07,$C3,$E3,$1F,$F0,$D3,$F8,$3F,$86,$FC,$37,$DA,$7F,$21,$BF
	.dc.b	$F0,$3C,$63,$C1,$90,$BF,$F4,$CD,$7F,$82,$63,$1C,$3F,$C5,$F8,$D3
	.dc.b	$CE,$B7,$DF,$DB,$C7,$FC,$BC,$3E,$BF,$EC,$63,$C8,$FD,$0D,$FF,$B8
	.dc.b	$9C,$4F,$86,$76,$FF,$AF,$0E,$5C,$FF,$83,$6C,$FF,$33,$69,$FC,$D6
	.dc.b	$78,$BF,$9F,$A7,$CF,$FC,$23,$E1,$FC,$75,$BF,$0E,$7E,$CF,$8D,$F8
	.dc.b	$FF,$1B,$F2,$1B,$3A,$C1,$3F,$33,$CA,$CE,$5F,$21,$F9,$FE,$4C,$BE
	.dc.b	$1F,$19,$6B,$E6,$FC,$DF,$D4,$31,$C5,$F3,$72,$47,$CA,$FA,$E5,$D8
	.dc.b	$C3,$70,$F9,$38,$2B,$8C,$7C,$56,$C7,$F5,$96,$BF,$EB,$03,$85,$7E
	.dc.b	$B2,$DF,$1D,$F8,$83,$C4,$5B,$F8,$5F,$1D,$F8,$CF,$1F,$F1,$AD,$8F
	.dc.b	$8D,$E3,$FE,$40,$C6,$3F,$8C,$7F,$27,$F1,$DF,$94,$F8,$DB,$5F,$35
	.dc.b	$C3,$F3,$FE,$36,$DF,$E5,$1F,$CD,$FC,$6C,$9F,$A9,$F1,$76,$BF,$33
	.dc.b	$E5,$7E,$80,$C7,$17,$CD,$7F,$94,$7C,$AF,$E7,$EF,$F0,$2D,$AE,$31
	.dc.b	$87,$7E,$A4,$8D,$FC,$75,$9F,$F7,$FD,$7F,$B8,$A7,$9F,$C7,$67,$1F
	.dc.b	$D4,$4E,$6B,$7F,$1D,$E2,$3C,$7F,$1D,$70,$F1,$FA,$DE,$63,$15,$D9
	.dc.b	$FD,$6F,$D0,$3E,$3A,$FC,$6E,$1B,$7F,$84,$94,$FE,$A5,$B1,$FA,$96
	.dc.b	$DF,$F8,$75,$AF,$F4,$7C,$9F,$1D,$F2,$0F,$0F,$C7,$5B,$F9,$3F,$1F
	.dc.b	$F3,$0F,$9B,$95,$8F,$DE,$EC,$FE,$01,$FF,$7F,$47,$F8,$AE,$31,$F0
	.dc.b	$5F,$CD,$27,$39,$FE,$56,$73,$5B,$F8,$97,$89,$F6,$B7,$FE,$83,$18
	.dc.b	$9F,$B7,$1F,$7B,$49,$1F,$7B,$E3,$EE,$3F,$EC,$CC,$BF,$B3,$BD,$DA
	.dc.b	$FF,$23,$19,$7A,$7D,$FF,$31,$FD,$2B,$2F,$F4,$DE,$1F,$8A,$F8,$EF
	.dc.b	$1F,$8F,$F8,$E7,$F1,$FF,$1F,$F9,$0F,$8F,$B5,$F9,$07,$FA,$CE,$7F
	.dc.b	$C6,$33,$FE,$5E,$D7,$C2,$DC,$FE,$57,$A1,$FD,$8B,$95,$F1,$FF,$9C
	.dc.b	$3F,$3B,$F1,$B7,$2F,$CE,$FC,$07,$E8,$0F,$D0,$FB,$76,$3E,$AE,$4F
	.dc.b	$7F,$E9,$12,$3E,$B4,$65,$FF,$03,$E2,$F8,$EF,$D5,$9F,$AC,$E2,$B3
	.dc.b	$FC,$B3,$E7,$3E,$2E,$FF,$2A,$3F,$C4,$31,$8D,$FF,$A6,$7D,$42,$7F
	.dc.b	$C3,$93,$97,$FA,$8F,$CF,$BE,$17,$FB,$CB,$63,$F1,$4D,$C3,$F1,$0D
	.dc.b	$7E,$D6,$1F,$91,$F8,$2F,$9E,$7F,$8A,$71,$7C,$8E,$30,$FE,$57,$F3
	.dc.b	$CE,$1F,$EA,$0D,$8F,$EA,$0D,$C7,$FD,$08,$D7,$BF,$CF,$7F,$8A,$1F
	.dc.b	$CD,$35,$CF,$A3,$FE,$12,$8C,$47,$E0,$3E,$17,$F0,$67,$97,$E4,$AE
	.dc.b	$1F,$D5,$B3,$7C,$9F,$A7,$E4,$FD,$4D,$9F,$D7,$F1,$BD,$6F,$CB,$12
	.dc.b	$3F,$54,$FF,$8A,$AF,$F5,$FC,$3F,$4B,$C6,$F8,$9F,$27,$F1,$5F,$8D
	.dc.b	$F2,$1F,$96,$F8,$9B,$8F,$E2,$1C,$3F,$86,$7F,$74,$47,$FC,$9B,$87
	.dc.b	$EB,$B0,$FB,$FE,$2F,$D1,$F8,$FF,$A4,$CA,$F7,$7C,$7D,$E2,$FC,$41
	.dc.b	$FE,$61,$C1,$1F,$E4,$0F,$98,$B5,$FC,$2D,$B5,$FC,$25,$AF,$E0,$9A
	.dc.b	$FF,$9C,$AB,$FC,$1F,$8F,$E4,$BA,$FF,$09,$E2,$7E,$15,$FC,$37,$C7
	.dc.b	$FE,$1F,$C2,$B9,$7D,$37,$0F,$D5,$F8,$FB,$97,$A3,$E1,$D9,$A7,$F2
	.dc.b	$07,$D1,$25,$7C,$67,$EA,$77,$FE,$57,$AB,$F2,$CD,$9F,$C1,$47,$E6
	.dc.b	$7E,$3F,$18,$F7,$C1,$1F,$E3,$2F,$9E,$FC,$3F,$B7,$FE,$27,$6D,$7F
	.dc.b	$B6,$DA,$FF,$80,$DA,$FF,$21,$F9,$5F,$F0,$8F,$CF,$7F,$8B,$DB,$1F
	.dc.b	$EE,$36,$BF,$E0,$B6,$BF,$CA,$65,$DD,$FE,$F1,$8F,$03,$5F,$03,$5E
	.dc.b	$0D,$7F,$01,$D1,$FB,$AD,$3E,$C7,$C9,$59,$FC,$23,$78,$3E,$7D,$5D
	.dc.b	$4D,$25,$7F,$90,$2B,$C5,$CB,$FE,$5D,$7F,$E0,$0C,$7F,$AD,$9A,$FF
	.dc.b	$AC,$5B,$66,$8F,$90,$F1,$2F,$91,$A6,$22,$57,$85,$AF,$E1,$E7,$FC
	.dc.b	$AE,$F9,$7E,$3F,$9F,$C7,$C8,$C1,$0E,$11,$E0,$3C,$3F,$1F,$F1,$72
	.dc.b	$7E,$57,$E5,$AF,$81,$ED,$F9,$8D,$46,$63,$F5,$F3,$C3,$2C,$92,$3E
	.dc.b	$47,$B1,$C3,$2F,$F3,$95,$76,$6D,$9A,$7F,$9D,$F5,$F9,$1F,$21,$C7
	.dc.b	$0F,$F0,$2D,$6D,$FA,$91,$F2,$BE,$3F,$E5,$FC,$74,$9F,$B1,$1E,$DF
	.dc.b	$53,$FD,$D5,$1F,$E0,$B6,$BF,$E0,$96,$BF,$D0,$9A,$FD,$DF,$27,$EB
	.dc.b	$91,$FC,$27,$C7,$7D,$AF,$3F,$ED,$FC,$74,$9F,$C4,$47,$E4,$5F,$F1
	.dc.b	$97,$C3,$FF,$A4,$DA,$7F,$A5,$7E,$09,$19,$6C,$7F,$BC,$39,$9F,$1D
	.dc.b	$F8,$82,$47,$E5,$7E,$3B,$B7,$CF,$7F,$20,$8F,$F5,$03,$5F,$F3,$D0
	.dc.b	$70,$FF,$9F,$5A,$FF,$86,$FE,$7F,$F3,$69,$A4,$EB,$FB,$7F,$5A,$F8
	.dc.b	$DF,$C5,$36,$BF,$C7,$6D,$7F,$0E,$FB,$1F,$CD,$7C,$8F,$E3,$39,$64
	.dc.b	$7E,$BE,$3F,$1F,$CB,$F8,$62,$47,$F9,$0F,$9B,$FD,$22,$FC,$5F,$91
	.dc.b	$FC,$AF,$2C,$8F,$DC,$47,$EB,$BE,$47,$F8,$07,$D3,$FF,$91,$FC,$4F
	.dc.b	$F6,$D7,$5A,$31,$7F,$B9,$AF,$D2,$78,$9E,$78,$FF,$02,$B5,$FF,$6F
	.dc.b	$B5,$FF,$20,$B6,$BA,$79,$3E,$02,$4F,$FB,$2A,$DF,$25,$FA,$CD,$BC
	.dc.b	$BC,$32,$BF,$3F,$1F,$B0,$CB,$F4,$E4,$8F,$84,$FE,$9D,$F4,$E3,$FE
	.dc.b	$67,$CF,$DB,$18,$C7,$E8,$78,$1F,$EF,$58,$F0,$9E,$3F,$92,$F8,$A7
	.dc.b	$F2,$57,$27,$EF,$B5,$7C,$57,$EE,$9B,$1F,$E7,$56,$BF,$E7,$36,$BF
	.dc.b	$E6,$5C,$4F,$8B,$1F,$2F,$67,$73,$EA,$FC,$97,$D6,$F9,$1F,$AF,$F2
	.dc.b	$32,$FE,$44,$79,$67,$E0,$12,$3F,$3B,$87,$68,$FE,$15,$AF,$FA,$86
	.dc.b	$7F,$DF,$22,$39,$3F,$4F,$BF,$F0,$BF,$23,$F8,$6F,$91,$93,$F5,$E3
	.dc.b	$2D,$21,$FD,$E6,$DF,$23,$F8,$EF,$91,$FC,$7F,$04,$AF,$C0,$46,$5A
	.dc.b	$43,$97,$C8,$FE,$5B,$E4,$64,$7E,$1D,$FC,$CF,$C8,$C8,$FC,$53,$5F
	.dc.b	$17,$C8,$FE,$78,$8F,$E3,$DF,$D0,$FC,$67,$E8,$BE,$43,$F4,$7F,$13
	.dc.b	$2B,$F2,$91,$FD,$97,$E9,$7F,$CD,$D1,$82,$1F,$E1,$77,$2D,$F8,$FE
	.dc.b	$1F,$1B,$23,$EF,$F6,$72,$E6,$5F,$F9,$AA,$73,$D5,$FB,$28,$FF,$2B
	.dc.b	$C6,$04,$F0,$F8,$DF,$21,$E5,$F6,$FF,$96,$7C,$63,$D2,$F4,$32,$F9
	.dc.b	$07,$4E,$1F,$E5,$6F,$F2,$43,$5E,$87,$F3,$58,$CC,$7F,$28,$6B,$FE
	.dc.b	$BF,$3A,$7A,$1E,$29,$91,$D6,$3E,$57,$C6,$FC,$BF,$8D,$93,$ED,$CF
	.dc.b	$E1,$CE,$A3,$1A,$23,$11,$F3,$FE,$33,$EA,$7C,$6F,$D5,$F8,$D9,$3C
	.dc.b	$23,$EC,$7C,$66,$A3,$07,$47,$FA,$97,$D9,$FF,$43,$E3,$FA,$99,$7F
	.dc.b	$A9,$EB,$C3,$F1,$BA,$C7,$FA,$C7,$E3,$5F,$9C,$9F,$8C,$D7,$FA,$5B
	.dc.b	$5F,$88,$D7,$FD,$C9,$83,$E2,$B9,$F8,$52,$7E,$AC,$E5,$F8,$3F,$1D
	.dc.b	$F9,$1F,$8E,$91,$83,$E2,$FC,$77,$E5,$66,$46,$0F,$93,$F1,$DF,$99
	.dc.b	$C6,$46,$0F,$9B,$F1,$DF,$9D,$99,$18,$3E,$A7,$C7,$7E,$71,$FE,$30
	.dc.b	$D7,$FA,$3B,$5F,$CA,$B7,$97,$F3,$BF,$19,$FE,$4D,$8F,$D2,$F1,$7E
	.dc.b	$85,$BD,$DF,$9F,$9D,$77,$FE,$BF,$E0,$A6,$3E,$28,$CB,$48,$7F,$A1
	.dc.b	$FE,$3F,$FC,$A7,$7E,$4E,$EF,$F1,$46,$3F,$9F,$AF,$E4,$6D,$7F,$31
	.dc.b	$9A,$31,$1F,$CD,$6F,$FC,$A4,$FC,$2A,$F5,$1F,$E6,$78,$CC,$78,$7E
	.dc.b	$2B,$FC,$6D,$1F,$E4,$38,$E5,$18,$33,$F3,$1C,$59,$1F,$A1,$9F,$F2
	.dc.b	$CB,$EB,$F8,$EE,$DF,$17,$23,$51,$E1,$23,$83,$97,$C7,$79,$7E,$3B
	.dc.b	$C6,$48,$FE,$87,$68,$F3,$FC,$76,$E7,$C5,$1E,$D8,$FF,$1B,$7C,$81
	.dc.b	$71,$FF,$58,$F9,$EF,$E9,$17,$FD,$6A,$BA,$39,$F6,$7F,$D4,$58,$7F
	.dc.b	$5E,$85,$70,$55,$D6,$F8,$47,$3C,$3B,$BF,$E8,$3C,$39,$BF,$DA,$3E
	.dc.b	$2B,$95,$F8,$23,$47,$1B,$D5,$7F,$AC,$EF,$9E,$1C,$31,$F2,$36,$FC
	.dc.b	$27,$92,$77,$5C,$6F,$38,$77,$BE,$17,$EC,$AD,$34,$F1,$1E,$43,$DF
	.dc.b	$87,$C3,$C7,$89,$EA,$C7,$DE,$DC,$7B,$CF,$9A,$FE,$0D,$F5,$3C,$D3
	.dc.b	$E4,$BF,$3C,$7D,$3A,$E0,$D3,$53,$1E,$58,$FA,$EF,$AE,$3E,$19,$CE
	.dc.b	$1E,$49,$F0,$E3,$E8,$7D,$EE,$EB,$E1,$92,$3E,$1B,$CE,$38,$57,$8C
	.dc.b	$F8,$97,$F8,$16,$DF,$77,$53,$C1,$F7,$9A,$C3,$FC,$53,$F3,$7F,$B2
	.dc.b	$BE,$4D,$BF,$C4,$B6,$77,$7F,$E8,$0F,$9E,$B8,$7F,$59,$6B,$F9,$76
	.dc.b	$BF,$E3,$66,$A7,$F1,$AF,$E7,$7E,$1B,$FC,$0B,$3F,$9D,$8E,$07,$E0
	.dc.b	$44,$6B,$3F,$14,$66,$30,$F8,$CC,$73,$76,$7C,$87,$E0,$4E,$6F,$E6
	.dc.b	$1E,$FA,$78,$67,$FC,$2E,$F8,$E3,$D0,$74,$3C,$86,$2F,$85,$77,$8E
	.dc.b	$77,$FD,$F6,$F9,$BE,$8B,$DC,$F8,$CE,$AE,$0F,$06,$1E,$5A,$E3,$3E
	.dc.b	$D9,$D5,$73,$3B,$5F,$AE,$7C,$91,$87,$CF,$7E,$5B,$F8,$F1,$F1,$2F
	.dc.b	$9B,$D6,$B8,$47,$8A,$3E,$44,$F8,$9E,$13,$A3,$E9,$E1,$EC,$38,$15
	.dc.b	$F3,$47,$DF,$C3,$CD,$87,$B5,$FE,$33,$CC,$DD,$F1,$7F,$E4,$8D,$3F
	.dc.b	$C8,$10,$F9,$2C,$5F,$FA,$C2,$36,$FF,$43,$8C,$E3,$FE,$40,$9E,$13
	.dc.b	$FC,$EA,$7F,$B9,$61,$DD,$B3,$AA,$FE,$BA,$BE,$33,$9D,$79,$5B,$0F
	.dc.b	$8B,$B1,$30,$FE,$A9,$F2,$37,$0F,$ED,$CF,$0F,$91,$E3,$E9,$DA,$FE
	.dc.b	$B4,$E1,$FE,$9E,$F9,$08,$FF,$52,$F4,$FC,$7D,$CF,$4B,$C7,$18,$64
	.dc.b	$FF,$78,$7D,$F3,$B8,$C1,$FD,$F3,$D0,$E6,$7B,$3D,$3B,$1F,$E1,$A8
	.dc.b	$F7,$FC,$56,$34,$FF,$04,$3F,$0B,$E3,$7C,$7E,$8D,$BF,$F1,$0B,$87
	.dc.b	$FC,$12,$F8,$88,$FF,$6C,$F8,$BF,$9B,$ED,$FF,$AF,$23,$FB,$E4,$63
	.dc.b	$09,$1F,$E5,$F9,$FD,$CC,$7F,$6C,$8F,$F7,$8C,$71,$8C,$1F,$E8,$07
	.dc.b	$7C,$DF,$E8,$97,$FD,$CF,$3C,$FE,$33,$C3,$F1,$BE,$26,$47,$7F,$8B
	.dc.b	$B8,$74,$E8,$60,$9F,$C7,$CB,$F8,$05,$C1,$8F,$9D,$6D,$A9,$CE,$5F
	.dc.b	$E8,$C9,$CB,$67,$FC,$89,$18,$BF,$F4,$30,$70,$FF,$A2,$5A,$FF,$81
	.dc.b	$E9,$C1,$F3,$3E,$3F,$C0,$DC,$3E,$57,$7F,$D0,$F8,$DF,$A2,$D8,$F9
	.dc.b	$D1,$F5,$3E,$3B,$18,$FC,$1D,$91,$F6,$3E,$3A,$5F,$D5,$F8,$DB,$7B
	.dc.b	$9F,$BB,$F1,$D8,$E5,$FA,$4F,$C8,$F9,$DF,$D5,$61,$FE,$AE,$5D,$F9
	.dc.b	$F0,$74,$C9,$FF,$57,$4E,$9B,$1F,$E6,$B8,$C6,$3F,$21,$78,$1F,$8E
	.dc.b	$F7,$7C,$0D,$BF,$C6,$7C,$6F,$E4,$8F,$C9,$FC,$75,$8F,$C8,$7C,$77
	.dc.b	$E5,$BE,$3B,$F2,$ED,$8F,$CA,$7C,$77,$E6,$BE,$3B,$F3,$6D,$8F,$CC
	.dc.b	$47,$E7,$BE,$3B,$1A,$FC,$B7,$F4,$3F,$1D,$2B,$F3,$DC,$D6,$BF,$21
	.dc.b	$F3,$C9,$FE,$26,$D7,$F4,$95,$87,$3F,$1F,$1F,$F6,$34,$EE,$B9,$CF
	.dc.b	$2F,$7A,$56,$BE,$3E,$D7,$82,$7B,$FC,$7E,$3A,$3F,$DE,$58,$A3,$C7
	.dc.b	$F0,$13,$1E,$FF,$1F,$71,$F4,$CF,$A3,$CE,$C6,$3C,$9F,$5F,$C7,$FF
	.dc.b	$2E,$8F,$F3,$76,$1E,$88,$C7,$BB,$23,$F4,$7E,$57,$F9,$B2,$3F,$9D
	.dc.b	$D7,$0C,$7F,$D2,$AD,$7F,$CD,$52,$D1,$E7,$FC,$AF,$8D,$F9,$6D,$C7
	.dc.b	$FD,$D9,$87,$F4,$CC,$3F,$80,$34,$F8,$3F,$1F,$E7,$F6,$3F,$A7,$C3
	.dc.b	$E3,$DE,$1F,$FA,$FE,$30,$D8,$CE,$1F,$BC,$9C,$BE,$C7,$FA,$83,$F6
	.dc.b	$3C,$DF,$89,$EB,$FE,$01,$F8,$DE,$24,$9F,$BB,$F1,$77,$4F,$3F,$B1
	.dc.b	$F1,$09,$3D,$5F,$8F,$EC,$7E,$2B,$E3,$7E,$83,$67,$F0,$FF,$1B,$F8
	.dc.b	$E3,$F1,$FF,$17,$61,$87,$F2,$ED,$7E,$07,$81,$ED,$24,$FB,$77,$FE
	.dc.b	$4E,$FD,$8E,$5F,$F7,$D3,$F9,$4F,$8C,$FC,$DF,$C6,$FE,$71,$FC,$EF
	.dc.b	$C6,$FF,$70,$E9,$FE,$73,$5E,$07,$ED,$46,$7E,$0F,$5E,$DF,$F7,$C9
	.dc.b	$34,$95,$E1,$F8,$DE,$9E,$E7,$EA,$9B,$1E,$CD,$BF,$C2,$1F,$02,$F1
	.dc.b	$BE,$EB,$67,$47,$8F,$E7,$1F,$0F,$AA,$FC,$1F,$CD,$FB,$B9,$3F,$D7
	.dc.b	$47,$FB,$09,$C3,$FE,$BD,$C5,$18,$8F,$13,$5F,$9F,$F3,$6F,$37,$F8
	.dc.b	$14,$7F,$44,$8C,$65,$FD,$1A,$11,$87,$53,$FE,$20,$C9,$F1,$1F,$E1
	.dc.b	$C6,$DF,$B2,$9C,$09,$7C,$7E,$EF,$F3,$FF,$6F,$FD,$6A,$FD,$BA,$23
	.dc.b	$11,$ED,$6B,$F3,$5A,$F6,$B7,$C8,$F8,$FE,$43,$C8,$DC,$FC,$99,$F8
	.dc.b	$FE,$47,$D0,$7A,$7E,$46,$CF,$A7,$3F,$F2,$57,$9F,$FB,$1E,$43,$49
	.dc.b	$3F,$C2,$B3,$FE,$11,$5F,$33,$1F,$81,$F2,$1F,$17,$CE,$F8,$CF,$C7
	.dc.b	$F2,$FF,$29,$F0,$17,$1F,$2D,$A3,$31,$FC,$5A,$30,$D8,$FC,$B1,$F9
	.dc.b	$7F,$33,$F3,$C7,$D3,$F8,$EB,$3D,$23,$EB,$79,$B8,$E3,$FF,$09,$6B
	.dc.b	$FC,$19,$BF,$03,$F6,$CF,$B9,$E6,$DC,$3E,$4D,$7F,$6E,$AE,$3E,$FE
	.dc.b	$5B,$3A,$F3,$7F,$08,$7E,$17,$CC,$B1,$93,$9F,$C5,$62,$BC,$6F,$E2
	.dc.b	$CF,$C7,$12,$3E,$73,$F9,$1F,$87,$FC,$81,$F9,$1F,$8D,$B3,$E6,$8F
	.dc.b	$CA,$7C,$6E,$2F,$DE,$FF,$8B,$BE,$6F,$F6,$9E,$4F,$E7,$DB,$3F,$E2
	.dc.b	$66,$BF,$B0,$6B,$FC,$60,$7C,$97,$FA,$3B,$E4,$3F,$50,$D8,$FF,$17
	.dc.b	$B5,$FD,$CB,$5E,$53,$FD,$87,$C8,$F1,$7B,$BF,$C0,$A7,$C4,$D7,$C4
	.dc.b	$D7,$F8,$D7,$CB,$FE,$A3,$4B,$97,$F0,$36,$CF,$E8,$27,$4F,$ED,$1B
	.dc.b	$D2,$69,$3E,$47,$94,$FE,$EB,$ED,$1F,$19,$FC,$16,$D1,$89,$FE,$D6
	.dc.b	$D7,$F8,$BC,$6A,$78,$4F,$E7,$67,$F6,$6E,$FC,$0F,$D4,$37,$1F,$E6
	.dc.b	$CE,$E7,$FC,$D0,$D9,$DF,$6F,$9B,$0F,$1B,$67,$FB,$4D,$7C,$1F,$43
	.dc.b	$59,$FF,$9B,$E3,$E2,$7A,$18,$99,$1B,$21,$FE,$77,$97,$C8,$7C,$4C
	.dc.b	$BE,$47,$91,$D3,$E4,$25,$FF,$09,$BF,$E2,$6D,$9E,$11,$88,$FE,$C0
	.dc.b	$DA,$FE,$10,$BE,$43,$E8,$CB,$ED,$3F,$D0,$6F,$F3,$56,$BF,$E7,$C6
	.dc.b	$BF,$D1,$9A,$FE,$1B,$E4,$79,$C7,$0B,$FE,$83,$3D,$7B,$3E,$2D,$7C
	.dc.b	$6D,$E4,$FE,$25,$F9,$3C,$5C,$BA,$FF,$A3,$4F,$95,$B8,$FF,$A2,$23
	.dc.b	$CD,$1F,$80,$F8,$9F,$A8,$48,$FC,$A7,$8B,$EE,$C1,$BD,$24,$FE,$67
	.dc.b	$E4,$3E,$1F,$E1,$CF,$8C,$F0,$B6,$7F,$AC,$35,$FF,$4F,$B5,$FF,$1C
	.dc.b	$FB,$3F,$E7,$A7,$FA,$1B,$5F,$F5,$A3,$78,$3E,$9F,$C8,$FD,$FF,$80
	.dc.b	$F9,$1D,$7F,$DF,$5B,$6A,$FF,$5D,$F1,$FE,$4F,$EA,$1F,$0F,$F6,$BE
	.dc.b	$79,$F1,$5F,$21,$B8,$7F,$6F,$6B,$FE,$B6,$6B,$FE,$00,$F9,$EF,$ED
	.dc.b	$EF,$F6,$D0,$E1,$FF,$6B,$98,$F6,$FF,$49,$F5,$2F,$6B,$F4,$3F,$2B
	.dc.b	$FA,$7F,$13,$E3,$6D,$FD,$5D,$BC,$1F,$C3,$27,$F2,$3F,$66,$F9,$2F
	.dc.b	$BE,$DC,$FF,$C0,$CD,$7F,$D9,$AD,$7F,$CC,$FE,$0B,$E5,$7C,$1D,$B2
	.dc.b	$7F,$8A,$47,$ED,$3E,$5B,$F6,$BF,$2D,$27,$F8,$C4,$78,$BE,$1F,$C2
	.dc.b	$48,$F0,$BF,$C4,$F0,$F4,$C7,$F1,$DE,$CF,$DF,$9F,$C0,$3E,$17,$F8
	.dc.b	$37,$9F,$E7,$E1,$FD,$7E,$9F,$C1,$0C,$ED,$FB,$0F,$B3,$47,$E7,$9B
	.dc.b	$8F,$F8,$95,$AF,$FB,$B1,$AF,$FA,$1A,$FF,$D3,$8F,$F8,$41,$AF,$FC
	.dc.b	$01,$6D,$FF,$00,$EC,$F2,$FC,$0F,$F3,$09,$FF,$32,$F1,$DB,$FD,$BD
	.dc.b	$FF,$83,$67,$FC,$27,$1F,$83,$C1,$87,$9C,$91,$E7,$7F,$9C,$4B,$E5
	.dc.b	$7E,$19,$27,$FC,$71,$1F,$E7,$4F,$97,$F0,$69,$27,$27,$C9,$CB,$FC
	.dc.b	$F3,$5F,$E7,$AE,$0F,$E2,$FF,$57,$FE,$A4,$AF,$F2,$53,$5F,$F2,$C8
	.dc.b	$70,$FF,$97,$1A,$F4,$3E,$2D,$7F,$6F,$8E,$5A,$7F,$18,$6C,$6F,$D4
	.dc.b	$F8,$DB,$7C,$7F,$98,$98,$7A,$23,$2D,$21,$FB,$9A,$F0,$6B,$F3,$FE
	.dc.b	$62,$4F,$C0,$7E,$97,$CC,$48,$C1,$0E,$47,$F7,$97,$EC,$78,$3F,$67
	.dc.b	$C2,$93,$F2,$A3,$EE,$7C,$C6,$A3,$07,$CE,$7F,$01,$F3,$1A,$8C,$1F
	.dc.b	$49,$FC,$DC,$7E,$17,$E6,$24,$FE,$3E,$3F,$11,$F3,$1A,$8C,$1F,$69
	.dc.b	$FC,$67,$CC,$7E,$37,$E5,$A4,$FD,$E8,$FC,$87,$CB,$48,$FC,$0B,$F9
	.dc.b	$3F,$96,$FE,$E1,$F6,$8C,$29,$2B,$F8,$B7,$17,$F5,$DE,$2F,$F1,$42
	.dc.b	$F9,$3F,$E9,$2D,$9F,$F4,$83,$5F,$C3,$B5,$FF,$50,$BF,$C6,$DE,$B1
	.dc.b	$FD,$9A,$38,$73,$FF,$5C,$75,$3F,$53,$C7,$62,$CB,$D2,$49,$FF,$42
	.dc.b	$C3,$D0,$F6,$99,$8E,$51,$C5,$DC,$7F,$80,$C7,$2F,$F8,$4E,$BC,$79
	.dc.b	$FF,$B3,$DB,$5F,$D1,$EB,$6F,$FA,$52,$3C,$F9,$7F,$AD,$0F,$15,$F5
	.dc.b	$3A,$79,$3D,$7E,$3A,$5F,$17,$FC,$0F,$F6,$9F,$E2,$5C,$BF,$CA,$CF
	.dc.b	$F8,$CE,$7F,$D9,$2D,$73,$ED,$7F,$7A,$8C,$47,$FB,$D1,$FF,$14,$BF
	.dc.b	$D3,$1F,$F1,$A3,$98,$EF,$EC,$7F,$85,$B3,$7C,$97,$C1,$F8,$E9,$7F
	.dc.b	$D5,$E7,$2F,$BF,$E3,$78,$54,$8C,$1F,$EB,$1D,$DD,$3F,$2F,$DC,$93
	.dc.b	$98,$F9,$DF,$17,$F4,$59,$1F,$2E,$3F,$9F,$F4,$FF,$20,$24,$7F,$12
	.dc.b	$E5,$FF,$77,$C7,$FB,$7F,$C5,$4F,$E0,$4F,$E7,$11,$C9,$FE,$43,$8F
	.dc.b	$F9,$3F,$E2,$BF,$DF,$91,$FC,$A1,$FF,$3B,$73,$7F,$BB,$2D,$18,$E3
	.dc.b	$EB,$8E,$3C,$6E,$44,$C3,$FD,$AB,$5D,$1F,$E1,$77,$CB,$D5,$FE,$B3
	.dc.b	$DB,$FC,$E8,$93,$EB,$94,$7A,$EB,$FD,$DF,$3F,$D4,$63,$8B,$E7,$F9
	.dc.b	$6F,$E9,$D3,$FE,$53,$8F,$F4,$E5,$7A,$3E,$0F,$91,$23,$F9,$CC,$7B
	.dc.b	$3B,$B8,$CC,$9F,$85,$1F,$E7,$3F,$94,$F0,$C7,$AA,$FF,$3B,$AF,$F4
	.dc.b	$86,$C3,$55,$70,$DB,$FD,$D4,$DB,$F9,$2D,$AE,$0D,$AE,$38,$FB,$3E
	.dc.b	$57,$E1,$12,$BD,$38,$A3,$E3,$72,$EE,$7E,$D4,$7C,$48,$FF,$38,$FC
	.dc.b	$6B,$E5,$7F,$C7,$D4,$F9,$FF,$F2,$BB,$F8,$89,$FC,$EB,$9F,$84,$FF
	.dc.b	$63,$46,$27,$73,$FE,$C8,$AE,$71,$89,$ED,$B7,$E4,$B8,$F1,$8F,$FB
	.dc.b	$3D,$B8,$67,$1E,$B8,$F9,$27,$93,$D2,$3C,$35,$FE,$D7,$C7,$8D,$F9
	.dc.b	$4F,$45,$6D,$B8,$7E,$8A,$76,$4B,$EF,$19,$6C,$B0,$EA,$48,$CB,$E4
	.dc.b	$9E,$13,$F1,$F0,$F2,$57,$9A,$32,$7C,$59,$B8,$FC,$48,$F4,$1F,$EF
	.dc.b	$0A,$E2,$F3,$37,$8F,$AD,$E4,$70,$9F,$A5,$7D,$E3,$13,$B9,$ED,$87
	.dc.b	$A6,$7A,$1E,$DA,$F5,$9F,$03,$1F,$89,$1A,$7D,$CF,$4C,$7E,$B3,$CA
	.dc.b	$BA,$3E,$47,$66,$2F,$F1,$2E,$2F,$E6,$F0,$FF,$9F,$67,$E1,$37,$1F
	.dc.b	$C9,$B7,$0F,$F1,$9D,$FF,$8B,$BE,$7E,$E3,$FC,$70,$FE,$3A,$E1,$C3
	.dc.b	$5C,$8A,$FF,$46,$38,$9F,$CC,$AF,$8F,$71,$7B,$A3,$99,$F1,$CF,$04
	.dc.b	$69,$C1,$A9,$E7,$19,$7A,$C2,$FA,$3D,$EA,$B9,$3F,$06,$64,$A6,$74
	.dc.b	$EB,$1F,$8C,$8F,$56,$9F,$0D,$E1,$7E,$0B,$EF,$87,$8E,$BE,$55,$76
	.dc.b	$8F,$56,$1B,$8C,$61,$D5,$F0,$57,$5A,$DB,$CA,$BC,$73,$E6,$BF,$25
	.dc.b	$6A,$3C,$91,$8B,$E8,$F9,$1F,$82,$D8,$F9,$67,$D8,$AE,$A7,$94,$B7
	.dc.b	$E1,$3C,$95,$D2,$7D,$27,$8C,$E7,$7F,$05,$F1,$98,$8F,$0B,$C2,$BB
	.dc.b	$57,$48,$FE,$19,$8F,$F6,$0C,$FF,$C3,$57,$FC,$FF,$E5,$ED,$7E,$37
	.dc.b	$83,$F9,$64,$FF,$69,$9C,$DF,$F6,$87,$F1,$15,$FD,$52,$91,$88,$E9
	.dc.b	$5F,$CF,$23,$11,$FD,$13,$E5,$3B,$F1,$FF,$4C,$6F,$23,$BA,$C4,$C3
	.dc.b	$FB,$D7,$89,$70,$FE,$B0,$F9,$32,$7C,$BD,$F9,$7C,$36,$FF,$C0,$51
	.dc.b	$E2,$D7,$FC,$02,$DA,$34,$C7,$87,$E6,$9F,$F2,$A5,$7F,$82,$E3,$1E
	.dc.b	$1C,$8F,$EF,$CF,$BB,$6F,$DB,$36,$B8,$E7,$F9,$D9,$FF,$08,$FA,$1F
	.dc.b	$0B,$23,$EB,$DB,$0C,$C7,$FC,$35,$B5,$C3,$FB,$C3,$F0,$7E,$5B,$E5
	.dc.b	$FC,$EF,$F0,$5A,$FF,$10,$46,$33,$91,$FE,$2D,$9D,$5F,$8D,$AF,$F1
	.dc.b	$3C,$D1,$88,$FF,$12,$FD,$93,$C2,$FC,$3F,$5B,$E8,$FE,$8E,$3F,$E3
	.dc.b	$68,$FE,$19,$8A,$F8,$4E,$4D,$57,$0C,$3F,$D6,$B7,$CF,$8F,$F0,$B6
	.dc.b	$F6,$3B,$DE,$EB,$AC,$E7,$0E,$98,$7F,$85,$E3,$FC,$FF,$18,$87,$D9
	.dc.b	$7C,$E8,$DE,$5E,$5F,$3F,$F4,$35,$AE,$1B,$7B,$8F,$F4,$B6,$0F,$5F
	.dc.b	$35,$E6,$78,$F0,$7F,$8F,$B2,$F0,$3F,$80,$CF,$F4,$EE,$70,$FF,$69
	.dc.b	$9F,$86,$9D,$F2,$FE,$9A,$30,$7F,$8F,$1C,$4F,$E2,$A3,$EB,$62,$F5
	.dc.b	$7E,$39,$E2,$F2,$6E,$1E,$E9,$7A,$BF,$2B,$E3,$3E,$6B,$67,$E4,$E7
	.dc.b	$F6,$E7,$E8,$F9,$79,$CB,$E0,$BF,$B8,$FB,$17,$99,$7C,$BC,$37,$1F
	.dc.b	$9B,$29,$C5,$FF,$AF,$E3,$8C,$EA,$7E,$8F,$D5,$BE,$5E,$FC,$C6,$1B
	.dc.b	$5F,$6E,$53,$F1,$DB,$1F,$D2,$70,$FF,$43,$C6,$1C,$35,$FC,$F6,$1F
	.dc.b	$EA,$49,$F3,$F9,$BE,$2F,$98,$F7,$3F,$C3,$75,$93,$FE,$8B,$F5,$2E
	.dc.b	$1F,$E6,$9F,$4F,$F8,$97,$CE,$7F,$13,$6D,$7E,$53,$9B,$FC,$AA,$FE
	.dc.b	$43,$C1,$F8,$8E,$78,$FF,$D3,$AE,$78,$7F,$76,$7F,$30,$8F,$ED,$47
	.dc.b	$EB,$27,$8E,$3F,$E9,$28,$CD,$7F,$A9,$AD,$0F,$9B,$E5,$C5,$CD,$B5
	.dc.b	$E1,$8E,$DC,$58,$C3,$A3,$89,$F3,$70,$59,$F0,$3E,$5F,$37,$CD,$F3
	.dc.b	$56,$BD,$8E,$1A,$7B,$E3,$15,$E9,$F9,$AF,$71,$EF,$F9,$9B,$5F,$16
	.dc.b	$3E,$17,$B7,$8A,$F8,$0F,$C5,$F6,$FF,$13,$F1,$56,$FF,$0C,$E1,$FC
	.dc.b	$25,$3A,$53,$F8,$92,$4F,$FA,$AB,$81,$F3,$3F,$44,$FA,$5F,$31,$67
	.dc.b	$ED,$7A,$9F,$CC,$0F,$AD,$F2,$96,$3F,$88,$7B,$3F,$C4,$B3,$FF,$20
	.dc.b	$79,$3F,$BD,$8C,$47,$F8,$32,$3F,$A4,$74,$3D,$5F,$C0,$7C,$9F,$F3
	.dc.b	$C6,$E1,$F8,$9F,$5B,$F9,$F9,$F8,$5F,$92,$B1,$E0,$F2,$FB,$11,$F8
	.dc.b	$39,$FF,$19,$7F,$A1,$7A,$FF,$8D,$F9,$3F,$E9,$2D,$8F,$CA,$7C,$97
	.dc.b	$E4,$4F,$C9,$7C,$95,$8F,$E5,$9F,$65,$FC,$C7,$83,$F6,$0D,$7F,$C7
	.dc.b	$4D,$7F,$CB,$CD,$7F,$CB,$5C,$BF,$4A,$7F,$45,$F2,$5F,$9F,$E7,$FD
	.dc.b	$03,$FE,$10,$F5,$3F,$CB,$73,$FA,$9F,$23,$0F,$F6,$07,$F9,$CE,$2F
	.dc.b	$8A,$FE,$54,$D9,$FC,$0B,$72,$F3,$CF,$F8,$0E,$DF,$25,$FD,$AB,$C9
	.dc.b	$F0,$37,$0F,$E4,$BF,$43,$FC,$A9,$BE,$8F,$F8,$6B,$EB,$3E,$C4,$60
	.dc.b	$FE,$A1,$F4,$4E,$1F,$E7,$8D,$8F,$C9,$B7,$4F,$E3,$0D,$7F,$D0,$9C
	.dc.b	$2F,$A1,$F0,$47,$F6,$66,$C7,$F6,$EF,$A9,$7C,$BF,$F7,$49,$CE,$0B
	.dc.b	$FF,$7D,$47,$F0,$8B,$FE,$EA,$D9,$FE,$11,$BF,$27,$FB,$2C,$66,$B9
	.dc.b	$7D,$D7,$F6,$37,$FC,$EC,$D7,$FD,$34,$D7,$7E,$E7,$F9,$03,$07,$E0
	.dc.b	$FF,$AC,$C3,$FA,$7B,$6B,$F8,$BC,$78,$BE,$0B,$FC,$1B,$39,$AF,$F0
	.dc.b	$5F,$43,$CF,$FF,$24,$B6,$FF,$97,$69,$F9,$3A,$DD,$7F,$7C,$D7,$FC
	.dc.b	$1D,$97,$FA,$35,$AF,$FA,$C5,$AF,$C7,$6F,$A6,$BF,$C4,$5F,$FA,$9E
	.dc.b	$31,$CF,$F9,$5F,$98,$FC,$B3,$F9,$7F,$98,$F0,$78,$77,$4F,$F0,$96
	.dc.b	$5F,$E4,$E6,$C7,$F7,$F8,$C6,$7F,$D6,$5A,$F9,$E3,$FC,$99,$9B,$CA
	.dc.b	$FF,$34,$B6,$FE,$0F,$91,$89,$91,$F1,$3C,$4B,$87,$CE,$3F,$D9,$57
	.dc.b	$FC,$0D,$C6,$BE,$FF,$98,$FD,$81,$FB,$1E,$4B,$87,$CE,$8F,$DA,$7C
	.dc.b	$06,$2F,$8E,$1F,$E1,$3A,$FF,$16,$F1,$BC,$BF,$AE,$7D,$8F,$2A,$CE
	.dc.b	$A3,$ED,$F9,$38,$C3,$CD,$87,$F3,$68,$CD,$7F,$7D,$8C,$37,$04,$FF
	.dc.b	$9C,$5B,$1F,$E0,$3D,$3E,$D7,$CC,$7F,$0B,$3F,$86,$7C,$A5,$9F,$C4
	.dc.b	$C7,$F0,$FF,$98,$C6,$5E,$C7,$FC,$8B,$F4,$9F,$E6,$1F,$91,$FE,$96
	.dc.b	$D9,$FF,$08,$35,$FF,$6D,$35,$FF,$7F,$7C,$9F,$FA,$9B,$E3,$FF,$A9
	.dc.b	$B6,$3F,$C3,$0D,$7F,$DC,$4D,$7F,$74,$27,$FC,$C9,$9F,$8B,$E2,$CD
	.dc.b	$24,$FF,$BA,$DF,$CD,$7C,$B7,$F0,$CD,$3F,$7F,$E4,$5A,$FC,$1B,$6B
	.dc.b	$71,$FA,$CF,$87,$C7,$13,$0A,$62,$33,$F1,$5F,$DB,$AF,$F5,$0D,$A3
	.dc.b	$4C,$7C,$57,$0C,$3F,$73,$1D,$A3,$C5,$5E,$C9,$FD,$EC,$65,$FF,$16
	.dc.b	$46,$5F,$CC,$AF,$94,$E3,$F1,$52,$77,$1E,$FC,$3F,$C0,$5F,$0B,$DA
	.dc.b	$24,$FF,$A1,$3D,$DF,$F1,$43,$FE,$86,$A7,$B1,$E4,$F8,$69,$39,$8E
	.dc.b	$93,$A8,$C1,$0F,$2F,$CB,$78,$6F,$F8,$D4,$7E,$6D,$AF,$FB,$15,$AF
	.dc.b	$FA,$CD,$AF,$F8,$F7,$37,$CC,$73,$F2,$E4,$FC,$48,$F9,$3F,$2F,$23
	.dc.b	$E3,$BF,$33,$E3,$3C,$5F,$11,$FE,$24,$CF,$FC,$48,$DC,$8D,$3E,$E6
	.dc.b	$4F,$96,$FA,$BF,$07,$32,$FA,$71,$ED,$F8,$89,$1F,$2D,$FB,$7F,$07
	.dc.b	$F0,$7E,$1F,$B1,$27,$E1,$3F,$E3,$65,$2F,$89,$FC,$1F,$C9,$49,$FF
	.dc.b	$5E,$4E,$5F,$F4,$CF,$DB,$FF,$A3,$DB,$CB,$F9,$DF,$59,$39,$9F,$F6
	.dc.b	$7A,$E5,$D7,$C2,$3E,$3F,$FC,$8F,$68,$C5,$FF,$A3,$63,$13,$A9,$CC
	.dc.b	$FF,$A4,$7E,$85,$E8,$7A,$DB,$1F,$E5,$26,$BB,$8F,$9B,$EA,$FE,$80
	.dc.b	$91,$F9,$EF,$88,$FC,$47,$AB,$F5,$49,$1F,$89,$7E,$C7,$A3,$F8,$D2
	.dc.b	$47,$E9,$BE,$5F,$F5,$7F,$2F,$23,$F5,$2F,$EB,$BE,$5F,$F2,$9E,$0F
	.dc.b	$E0,$49,$3F,$95,$7F,$03,$B7,$ED,$3E,$2A,$4F,$EC,$A3,$F3,$9F,$0B
	.dc.b	$FB,$8F,$89,$93,$FA,$B8,$FD,$07,$0F,$E3,$49,$1F,$A1,$7F,$25,$F8
	.dc.b	$8B,$CE,$FD,$B3,$67,$FD,$EE,$D7,$F7,$FF,$29,$AF,$8B,$FE,$31,$F1
	.dc.b	$FF,$C9,$A7,$FA,$DB,$79,$F9,$78,$AC,$3F,$C7,$F5,$D3,$E1,$BC,$24
	.dc.b	$7F,$A7,$42,$3C,$BE,$9F,$90,$93,$FE,$BD,$C1,$EB,$F8,$BC,$E9,$39
	.dc.b	$7C,$BF,$29,$EC,$F8,$5F,$79,$27,$BB,$FC,$A6,$FF,$DB,$B1,$F0,$23
	.dc.b	$83,$DF,$E5,$3D,$FE,$84,$8F,$13,$F0,$BC,$69,$18,$21,$F1,$1F,$9B
	.dc.b	$F0,$9F,$23,$CD,$93,$F3,$A3,$2D,$21,$F3,$3E,$53,$E7,$F9,$12,$3E
	.dc.b	$13,$F6,$3C,$49,$1F,$29,$FB,$A4,$7F,$BA,$68,$F2,$B7,$1F,$E0,$0E
	.dc.b	$3F,$D2,$F2,$FF,$8B,$EB,$53,$F3,$FE,$53,$EF,$FC,$A4,$8C,$1F,$4F
	.dc.b	$C5,$FC,$1C,$C8,$C1,$F5,$FC,$5F,$C3,$FC,$24,$8C,$1F,$6F,$E1,$3F
	.dc.b	$16,$47,$EF,$78,$9F,$8E,$F9,$39,$1F,$81,$7F,$B5,$2F,$9D,$7C,$3E
	.dc.b	$17,$C2,$FF,$1D,$6C,$FF,$B5,$1A,$FE,$64,$E5,$1B,$F8,$6E,$4C,$8F
	.dc.b	$F3,$E7,$07,$E8,$E9,$5F,$C7,$FE,$A1,$B7,$E9,$3C,$79,$87,$0F,$1B
	.dc.b	$8D,$6A,$30,$31,$46,$63,$FC,$91,$18,$BE,$C7,$EC,$3E,$23,$F6,$3E
	.dc.b	$47,$F8,$27,$DA,$FD,$B5,$7F,$99,$DB,$8F,$FA,$4F,$DA,$73,$FF,$2D
	.dc.b	$CB,$C3,$1F,$EF,$5F,$27,$9C,$7E,$B7,$1E,$91,$D7,$E2,$E4,$F0,$C7
	.dc.b	$FD,$42,$FA,$87,$E1,$3F,$C0,$72,$FF,$14,$FD,$77,$F9,$4E,$7E,$70
	.dc.b	$70,$FD,$16,$BF,$C2,$DA,$FF,$8C,$BE,$FB,$FC,$36,$F7,$8E,$3A,$FF
	.dc.b	$15,$6D,$6F,$E1,$7E,$27,$9B,$F1,$7C,$D9,$87,$A6,$3E,$47,$95,$A8
	.dc.b	$C1,$ED,$7E,$67,$C8,$7C,$DF,$90,$93,$F0,$63,$E8,$79,$1A,$8C,$1F
	.dc.b	$15,$FE,$57,$8F,$F2,$E2,$47,$F1,$E7,$FC,$71,$BB,$E4,$37,$3F,$CC
	.dc.b	$27,$EE,$47,$F8,$A7,$91,$F1,$7F,$7F,$C8,$97,$F8,$AF,$4B,$FC,$A7
	.dc.b	$3B,$24,$66,$B8,$BF,$67,$E4,$3F,$C6,$9C,$3F,$E3,$98,$FF,$3F,$35
	.dc.b	$FE,$79,$7F,$D1,$7E,$23,$FC,$31,$39,$9F,$F3,$2D,$B4,$FD,$8C,$78
	.dc.b	$E7,$97,$27,$F4,$26,$9F,$93,$9D,$CF,$FA,$89,$FD,$9B,$D1,$AE,$FC
	.dc.b	$1F,$8E,$7C,$8F,$03,$F4,$0C,$8F,$CA,$47,$E8,$FE,$43,$F4,$9F,$1B
	.dc.b	$27,$F3,$11,$FE,$6F,$F9,$0F,$F0,$A4,$7A,$9A,$FE,$9D,$AF,$F9,$F9
	.dc.b	$B9,$FB,$F7,$FB,$24,$9F,$F5,$06,$5F,$39,$FF,$41,$2F,$8E,$FD,$B4
	.dc.b	$C9,$FD,$54,$7F,$1A,$F9,$0F,$F1,$F6,$C9,$7C,$FF,$F8,$BD,$4B,$E3
	.dc.b	$3F,$D2,$B3,$98,$C4,$F1,$8C,$7C,$5F,$F0,$D8,$E7,$EB,$FF,$A9,$23
	.dc.b	$1C,$7C,$70,$FF,$4F,$B6,$BC,$4F,$2C,$6D,$F1,$7B,$3F,$EA,$EA,$F1
	.dc.b	$B6,$B2,$7A,$F1,$F6,$BE,$23,$18,$79,$71,$F0,$46,$4F,$7D,$F4,$77
	.dc.b	$87,$47,$E1,$D7,$AB,$0F,$04,$7F,$A4,$DB,$8F,$A6,$30,$72,$3B,$63
	.dc.b	$F2,$1B,$87,$77,$AE,$1D,$E7,$D3,$8F,$98,$98,$7C,$3A,$E7,$8F,$AE
	.dc.b	$3C,$73,$A9,$F0,$3E,$3C,$7D,$E4,$9E,$4F,$9E,$38,$4F,$C2,$3C,$55
	.dc.b	$D5,$F5,$4F,$17,$E7,$63,$F6,$AF,$99,$23,$27,$DA,$29,$F1,$0F,$79
	.dc.b	$70,$FA,$FA,$FF,$03,$6D,$FB,$E3,$18,$7E,$35,$FE,$09,$4F,$9A,$FE
	.dc.b	$A1,$5F,$DB,$5B,$1F,$E4,$76,$BF,$EE,$E6,$BF,$A1,$7C,$CF,$E8,$FD
	.dc.b	$DF,$F2,$A4,$FE,$8E,$34,$7A,$4F,$F2,$1F,$2F,$07,$31,$B9,$E3,$97
	.dc.b	$8E,$78,$46,$1F,$21,$F9,$13,$99,$4F,$61,$97,$FC,$CD,$A7,$8E,$33
	.dc.b	$5D,$EF,$C5,$18,$BF,$F3,$35,$F7,$8F,$80,$73,$9F,$CE,$3E,$CB,$E4
	.dc.b	$F8,$70,$CD,$78,$2F,$D5,$5A,$3C,$53,$AB,$E7,$7E,$89,$D5,$7C,$1C
	.dc.b	$3A,$9C,$27,$E1,$D7,$A2,$F9,$47,$84,$F3,$16,$F9,$BF,$21,$AF,$B2
	.dc.b	$BA,$DF,$73,$E0,$DF,$59,$E7,$5E,$13,$9E,$3F,$D2,$FA,$BF,$D2,$5B
	.dc.b	$7F,$86,$FE,$46,$E1,$FD,$26,$6E,$0B,$FE,$A5,$58,$6D,$7F,$34,$8F
	.dc.b	$F4,$D4,$FF,$51,$6B,$C7,$89,$18,$A7,$1B,$E1,$F0,$DC,$3F,$C9,$13
	.dc.b	$A9,$C9,$C7,$D7,$F2,$4F,$F8,$B9,$6D,$E5,$73,$97,$F9,$06,$7F,$C6
	.dc.b	$D1,$A7,$87,$C8,$EF,$E3,$38,$36,$3A,$47,$72,$3E,$49,$47,$8F,$E2
	.dc.b	$FE,$3F,$6F,$FA,$FE,$3C,$51,$8A,$D9,$FD,$E6,$3E,$2C,$EE,$3F,$C0
	.dc.b	$B1,$87,$78,$7F,$81,$B3,$F9,$31,$FE,$0F,$D7,$DD,$F2,$3E,$09,$CD
	.dc.b	$7B,$C6,$1F,$66,$BE,$27,$C8,$ED,$B8,$7C,$9E,$9F,$34,$FC,$96,$C7
	.dc.b	$F8,$3E,$3C,$84,$7F,$C3,$6F,$D3,$E8,$FA,$1F,$23,$F4,$5B,$3F,$56
	.dc.b	$3E,$A7,$C8,$E3,$2F,$36,$7F,$ED,$98,$FB,$1F,$23,$F6,$78,$EE,$1F
	.dc.b	$75,$C3,$F7,$FE,$3E,$E1,$CB,$89,$1F,$98,$87,$E5,$6F,$8C,$C7,$D7
	.dc.b	$FF,$A6,$30,$FF,$36,$4E,$5F,$F3,$74,$66,$BE,$50,$70,$FF,$BC,$DA
	.dc.b	$FF,$82,$CD,$78,$DF,$A0,$6C,$FE,$1F,$E4,$7F,$28,$79,$6D,$F1,$F8
	.dc.b	$9F,$CD,$7C,$8F,$E6,$7E,$47,$F3,$4D,$8F,$CF,$47,$E7,$73,$C6,$3F
	.dc.b	$05,$FD,$BF,$80,$F9,$77,$B1,$86,$E1,$FA,$7C,$B1,$32,$3F,$55,$F2
	.dc.b	$57,$1F,$98,$FC,$BF,$09,$F0,$D7,$E1,$B0,$DB,$FF,$4B,$4A,$7E,$13
	.dc.b	$63,$FD,$2A,$DB,$FF,$62,$35,$FF,$2F,$F2,$3E,$4B,$F7,$47,$87,$E1
	.dc.b	$ED,$FF,$AB,$0F,$F1,$D1,$6B,$F8,$A3,$FD,$33,$63,$FE,$B5,$37,$1F
	.dc.b	$90,$5F,$25,$C3,$E4,$B8,$B6,$3B,$7C,$97,$4F,$92,$EA,$D8,$F1,$47
	.dc.b	$83,$CD,$C6,$1D,$5F,$1F,$C9,$7F,$01,$E7,$B7,$FE,$79,$70,$FF,$A1
	.dc.b	$1F,$E0,$04,$7F,$90,$8F,$92,$F6,$9C,$2F,$D4,$FF,$0D,$E7,$93,$F1
	.dc.b	$7E,$4A,$D7,$F0,$3F,$23,$E1,$D7,$C6,$F9,$2C,$F0,$7C,$6A,$FE,$57
	.dc.b	$1F,$2B,$E2,$BE,$5F,$C9,$5C,$BE,$83,$86,$9F,$4E,$31,$5F,$37,$E4
	.dc.b	$BF,$90,$72,$FD,$46,$D7,$FA,$66,$3F,$8F,$11,$FF,$52,$BF,$D1,$57
	.dc.b	$C9,$7D,$C3,$EE,$F5,$58,$FF,$5A,$7E,$05,$F7,$DF,$F3,$E6,$C8,$C6
	.dc.b	$3F,$D1,$BA,$FF,$9A,$C6,$27,$5C,$FF,$CC,$A3,$F0,$77,$FB,$BB,$E7
	.dc.b	$18,$7F,$D6,$ED,$7F,$80,$57,$7E,$8F,$C8,$F4,$F6,$6D,$F9,$BA,$3C
	.dc.b	$27,$E5,$BE,$4E,$C7,$AB,$A7,$C4,$47,$C5,$8F,$F3,$AC,$93,$E7,$E9
	.dc.b	$FC,$FF,$C8,$FB,$1B,$7F,$A5,$F9,$0F,$D1,$FC,$7F,$E9,$1B,$1F,$23
	.dc.b	$9F,$E0,$11,$EF,$BF,$C9,$8F,$EE,$D9,$7E,$AA,$7F,$D5,$1D,$1F,$87
	.dc.b	$F3,$7E,$74,$7F,$80,$E5,$F1,$F6,$3E,$7F,$44,$78,$9F,$EB,$1A,$FA
	.dc.b	$DE,$1E,$78,$3F,$2D,$87,$E0,$2B,$C6,$D8,$FA,$3D,$1F,$5C,$F2,$FC
	.dc.b	$5F,$84,$F1,$7C,$65,$BE,$F5,$FE,$BF,$F9,$A7,$1F,$F3,$56,$CF,$CE
	.dc.b	$94,$61,$FA,$2D,$7F,$9F,$6D,$FC,$B7,$E0,$FF,$C3,$AD,$AF,$AE,$D9
	.dc.b	$FC,$53,$5F,$C3,$E6,$E9,$FC,$99,$F2,$7E,$36,$DF,$A2,$DF,$30,$F0
	.dc.b	$7E,$77,$C6,$FF,$3E,$8C,$57,$BB,$8B,$F4,$13,$F9,$ED,$BF,$38,$49
	.dc.b	$F9,$6F,$E7,$F9,$FF,$46,$7D,$9F,$8E,$B3,$CF,$57,$C7,$7D,$DF,$8E
	.dc.b	$FB,$CD,$9F,$D8,$F5,$FE,$A4,$8F,$D8,$7F,$13,$F1,$DF,$85,$3F,$0D
	.dc.b	$F1,$D6,$7B,$7C,$AF,$F8,$A3,$E2,$7F,$13,$18,$3E,$CF,$7F,$FA,$7F
	.dc.b	$BB,$47,$F8,$A2,$77,$CF,$FB,$79,$D5,$7E,$6A,$73,$3F,$B6,$F4,$3F
	.dc.b	$A0,$F3,$7F,$1F,$B4,$62,$FF,$29,$BF,$F8,$06,$B5,$39,$BF,$F1,$5B
	.dc.b	$79,$3F,$87,$7C,$7F,$F9,$02,$31,$68,$F6,$76,$7E,$97,$E3,$AE,$5F
	.dc.b	$35,$C3,$F3,$FE,$3E,$E5,$FE,$24,$C1,$1F,$AD,$ED,$FD,$77,$C7,$DB
	.dc.b	$FA,$CE,$1F,$B1,$F1,$97,$0E,$4F,$ED,$4F,$DD,$92,$3F,$CD,$B8,$23
	.dc.b	$F7,$FF,$0D,$FE,$7D,$E1,$FF,$34,$F0,$7E,$FE,$BF,$C9,$AF,$F0,$18
	.dc.b	$CD,$FE,$ED,$B8,$7F,$A1,$DF,$F3,$73,$FE,$3E,$73,$C1,$FE,$D9,$F8
	.dc.b	$BF,$F1,$B7,$CC,$3C,$47,$C7,$FE,$CE,$30,$D8,$CE,$1F,$C9,$27,$2F
	.dc.b	$F2,$6F,$90,$7C,$3F,$FA,$32,$30,$D7,$F9,$E4,$6B,$E3,$FF,$80,$36
	.dc.b	$3A,$72,$FF,$A7,$DB,$9F,$8E,$31,$1F,$D1,$70,$46,$23,$E1,$7C,$7F
	.dc.b	$F0,$6D,$BE,$27,$C7,$59,$F8,$8D,$C3,$FD,$F5,$57,$73,$49,$5F,$BD
	.dc.b	$87,$C7,$EF,$E3,$FF,$8C,$5A,$51,$86,$D7,$FA,$A9,$AF,$F5,$D9,$E1
	.dc.b	$F1,$DC,$E3,$B5,$78,$E2,$47,$85,$AF,$DB,$F8,$FF,$47,$C7,$FA,$78
	.dc.b	$64,$F9,$E3,$2D,$21,$F8,$3F,$8F,$F8,$1F,$1F,$F0,$7E,$3E,$4F,$BE
	.dc.b	$32,$D2,$1F,$8B,$F8,$FF,$99,$E1,$FC,$1F,$1F,$F9,$D3,$67,$F8,$14
	.dc.b	$63,$73,$F1,$7E,$37,$8C,$62,$7F,$87,$B5,$FC,$C4,$7D,$5F,$86,$FA
	.dc.b	$DF,$19,$27,$EA,$47,$2F,$27,$B1,$23,$F8,$93,$F4,$FE,$17,$EF,$12
	.dc.b	$3E,$E7,$C5,$FE,$07,$E2,$FF,$05,$F0,$D2,$7F,$01,$1E,$6F,$27,$D4
	.dc.b	$48,$FE,$3C,$FE,$0B,$F3,$1F,$E9,$FD,$BF,$DB,$5B,$3F,$EF,$96,$F5
	.dc.b	$1F,$D5,$C7,$F9,$17,$77,$93,$F9,$59,$FF,$32,$CF,$F4,$D6,$BB,$8F
	.dc.b	$93,$E4,$7C,$D2,$47,$F3,$17,$E9,$CF,$EA,$3C,$7F,$E6,$0D,$C7,$FC
	.dc.b	$6E,$DA,$F3,$7C,$27,$E9,$D7,$CC,$7D,$37,$FC,$68,$D7,$FD,$60,$D4
	.dc.b	$70,$E5,$F0,$F8,$32,$3F,$C3,$B7,$FA,$E6,$C7,$F8,$07,$17,$9A,$E4
	.dc.b	$FD,$C4,$61,$B1,$9A,$F8,$9F,$19,$F1,$FE,$1B,$E4,$7C,$2C,$9F,$91
	.dc.b	$1F,$96,$CB,$F6,$31,$FE,$48,$6B,$F0,$BC,$4F,$8E,$7D,$BF,$85,$FA
	.dc.b	$44,$8F,$A5,$F1,$9F,$57,$E3,$3E,$B7,$C3,$C9,$FA,$D1,$DF,$C4,$FA
	.dc.b	$24,$8F,$82,$FE,$1F,$E3,$3E,$F1,$23,$EF,$7C,$17,$E0,$7E,$23,$F0
	.dc.b	$5F,$0F,$27,$F0,$51,$EB,$F0,$BE,$E9,$23,$E6,$BF,$93,$F8,$CF,$C5
	.dc.b	$12,$3F,$15,$F1,$9F,$77,$7F,$EA,$11,$F8,$DF,$88,$FC,$81,$9F,$6B
	.dc.b	$5D,$DF,$E3,$59,$FE,$3A,$DA,$34,$D7,$E2,$BF,$8B,$9D,$9F,$E4,$22
	.dc.b	$9F,$EF,$C7,$F2,$1E,$17,$E8,$FE,$33,$F4,$1F,$0F,$2B,$F4,$11,$FA
	.dc.b	$32,$38,$3F,$49,$F0,$FF,$A7,$F8,$4F,$E7,$9E,$EF,$FA,$AE,$5E,$07
	.dc.b	$F9,$9E,$30,$FF,$9A,$7A,$D5,$C6,$BF,$5C,$D7,$8F,$8B,$FE,$67,$BF
	.dc.b	$F3,$6C,$7F,$8F,$DF,$F7,$EB,$7E,$33,$C2,$78,$BD,$E9,$1B,$6B,$D0
	.dc.b	$6B,$FD,$C3,$DB,$F4,$92,$3B,$3E,$CF,$8B,$FE,$5F,$83,$C9,$F7,$FC
	.dc.b	$54,$BF,$F1,$BC,$E5,$F3,$7C,$57,$0A,$91,$83,$D5,$F1,$3F,$20,$8F
	.dc.b	$B7,$E2,$BE,$5F,$99,$F3,$19,$1F,$06,$3F,$83,$7E,$2A,$F1,$7E,$8F
	.dc.b	$C5,$CA,$FC,$04,$E5,$BE,$27,$E2,$E0,$D2,$47,$FA,$0E,$3F,$CE,$0F
	.dc.b	$02,$1F,$24,$3A,$7E,$67,$07,$F8,$6B,$CD,$FC,$1F,$C9,$49,$FB,$11
	.dc.b	$F8,$62,$38,$3E,$E7,$C5,$7C,$EC,$BF,$17,$E7,$C9,$F9,$B1,$F8,$EF
	.dc.b	$53,$51,$83,$F0,$8F,$E4,$BE,$2B,$F2,$7E,$A4,$9F,$C4,$47,$F2,$DF
	.dc.b	$BB,$79,$1F,$AD,$6C,$FF,$A8,$DA,$FF,$B4,$9A,$FF,$2E,$3F,$13,$CD
	.dc.b	$FB,$A7,$E2,$47,$0E,$FF,$D1,$37,$B7,$F3,$5E,$37,$E2,$B0,$FD,$3F
	.dc.b	$8B,$30,$FC,$4C,$7F,$89,$7E,$87,$FA,$27,$BD,$FE,$44,$8E,$13,$FE
	.dc.b	$9F,$9C,$E6,$F1,$78,$7C,$5C,$AF,$C9,$C7,$3F,$5F,$F3,$2C,$8E,$51
	.dc.b	$F9,$CF,$17,$FC,$13,$7F,$B4,$9F,$F7,$1B,$5F,$F4,$73,$5F,$85,$D0
	.dc.b	$C5,$F1,$1C,$E7,$F4,$77,$EB,$8F,$67,$C4,$7B,$7C,$C9,$5E,$B8,$CB
	.dc.b	$48,$7E,$32,$DF,$11,$FE,$20,$7F,$CD,$11,$FE,$F8,$6B,$FE,$A3,$6B
	.dc.b	$F5,$F7,$F3,$79,$1F,$E4,$89,$7C,$47,$38,$95,$F3,$63,$E9,$7C,$3E
	.dc.b	$C9,$1F,$81,$3F,$DF,$54,$8E,$9F,$11,$C2,$7E,$B4,$7D,$B8,$FB,$9E
	.dc.b	$BF,$DD,$F8,$79,$5F,$6E,$32,$D2,$1F,$07,$E2,$3F,$09,$F1,$1E,$92
	.dc.b	$47,$8C,$FE,$37,$8A,$3D,$DF,$11,$C2,$7F,$11,$1F,$8D,$8F,$F0,$77
	.dc.b	$C5,$7A,$7B,$BF,$C1,$3F,$2C,$F8,$2F,$CF,$4F,$C6,$C3,$31,$8D,$FF
	.dc.b	$45,$5B,$9F,$F0,$8D,$73,$51,$95,$19,$EB,$C3,$D6,$11,$8A,$FD,$73
	.dc.b	$AE,$0F,$E9,$73,$AC,$3C,$B1,$8C,$38,$4F,$98,$F1,$DF,$9A,$B7,$8E
	.dc.b	$67,$C7,$7C,$67,$D9,$19,$C7,$C9,$87,$86,$F1,$7E,$2C,$7B,$DF,$8C
	.dc.b	$EE,$7F,$8C,$9E,$0E,$4E,$F3,$D7,$1F,$24,$70,$3E,$3E,$3E,$66,$9F
	.dc.b	$32,$FC,$CE,$F0,$F0,$9D,$89,$3E,$A8,$E2,$7D,$0A,$EE,$F8,$8C,$E3
	.dc.b	$F4,$E7,$86,$3E,$C3,$D2,$74,$9F,$04,$75,$BC,$47,$39,$E1,$3C,$8F
	.dc.b	$53,$C9,$C4,$7B,$E3,$0F,$B7,$1F,$F2,$CF,$47,$E0,$A3,$07,$F7,$69
	.dc.b	$F8,$91,$98,$F5,$35,$FF,$30,$FC,$2F,$29,$FD,$23,$63,$F7,$0D,$7F
	.dc.b	$9D,$35,$E2,$7D,$73,$FD,$E7,$E0,$7D,$BC,$39,$C6,$7A,$D3,$98,$EA
	.dc.b	$65,$C3,$D8,$C1,$D4,$FF,$6C,$1F,$58,$EE,$53,$39,$7E,$A5,$CB,$5E
	.dc.b	$38,$66,$BC,$71,$D7,$0F,$0B,$D2,$78,$47,$E8,$63,$9E,$1E,$CB,$E5
	.dc.b	$86,$67,$C3,$3F,$C7,$F0,$EB,$5C,$A7,$D1,$18,$9E,$46,$EB,$95,$F7
	.dc.b	$9D,$57,$C6,$C3,$BB,$E7,$3E,$2B,$D7,$0C,$97,$0E,$93,$F1,$6F,$A6
	.dc.b	$1D,$A3,$C6,$DB,$EA,$79,$CB,$5E,$69,$FE,$F5,$87,$91,$FA,$D8,$78
	.dc.b	$27,$F9,$EF,$B9,$F3,$33,$F9,$B9,$FF,$5F,$F8,$AC,$67,$A3,$FB,$A3
	.dc.b	$8C,$7F,$99,$BF,$E6,$37,$FA,$3B,$5C,$C7,$F3,$99,$E8,$DE,$27,$C5
	.dc.b	$73,$9F,$ED,$CD,$87,$C5,$58,$95,$FE,$05,$F8,$9B,$87,$F6,$97,$87
	.dc.b	$C5,$73,$FC,$CB,$FD,$8F,$5F,$E0,$B8,$C5,$48,$F2,$4E,$A3,$FB,$06
	.dc.b	$48,$C4,$7F,$80,$9A,$FE,$E9,$FC,$DE,$9F,$94,$ED,$FF,$03,$7B,$5E
	.dc.b	$1E,$2F,$8D,$18,$C7,$FD,$B3,$5B,$D9,$7F,$1E,$B5,$8F,$F7,$D8,$C4
	.dc.b	$EA,$FC,$FF,$27,$ED,$3D,$DE,$CD,$8E,$A7,$96,$3E,$17,$C5,$E7,$0F
	.dc.b	$F2,$CE,$88,$F8,$DF,$17,$E6,$C5,$ED,$63,$1E,$78,$34,$FF,$11,$4E
	.dc.b	$CF,$29,$27,$FC,$2C,$FE,$07,$7F,$95,$CD,$D7,$D9,$FA,$8D,$AE,$B7
	.dc.b	$FE,$2D,$9F,$B1,$F1,$79,$C7,$CE,$FD,$BF,$8B,$F3,$7A,$F7,$0E,$3A
	.dc.b	$23,$31,$E8,$CF,$ED,$7C,$5E,$31,$F8,$2F,$E1,$7D,$7F,$CD,$43,$E5
	.dc.b	$F9,$5F,$1E,$2F,$F4,$83,$FE,$CD,$8C,$CF,$FA,$15,$AF,$F8,$9D,$AF
	.dc.b	$F9,$A5,$A8,$ED,$8F,$C5,$AF,$F4,$AD,$77,$F8,$CF,$CC,$7B,$DF,$99
	.dc.b	$7F,$35,$F1,$7F,$57,$D7,$B5,$F0,$1C,$3E,$7F,$8C,$B9,$7D,$E7,$F4
	.dc.b	$5E,$BC,$9F,$67,$C6,$5A,$FC,$1C,$FE,$9F,$E3,$31,$DD,$FC,$0B,$54
	.dc.b	$7C,$6F,$95,$7B,$B7,$DE,$C3,$73,$FF,$79,$4A,$7F,$81,$64,$8C,$4F
	.dc.b	$F8,$EE,$31,$3A,$9C,$C7,$F8,$FD,$6B,$FB,$BF,$73,$C4,$DB,$FA,$DF
	.dc.b	$19,$FB,$FF,$8C,$FE,$00,$D8,$FB,$5E,$EF,$F0,$33,$18,$FE,$89,$FE
	.dc.b	$21,$F9,$9F,$F5,$E9,$FC,$9E,$CF,$77,$31,$FB,$A8,$C3,$A8,$E3,$F2
	.dc.b	$99,$6C,$7E,$46,$9F,$1B,$D8,$EF,$EE,$D9,$D7,$C6,$F8,$8C,$67,$FE
	.dc.b	$CE,$B4,$79,$BC,$B9,$5C,$0E,$CF,$AB,$E3,$64,$78,$7E,$36,$E1,$CE
	.dc.b	$7D,$FF,$1B,$8E,$8F,$E7,$F9,$23,$E1,$F9,$53,$1F,$57,$C6,$DC,$7C
	.dc.b	$93,$F2,$3C,$9C,$63,$C9,$FD,$5F,$C9,$BC,$9B,$E3,$E1,$B9,$7F,$0C
	.dc.b	$C5,$4A,$FC,$3B,$67,$F0,$CD,$9F,$F4,$13,$5F,$F4,$B3,$FB,$67,$E4
	.dc.b	$7C,$67,$8B,$C8,$FB,$AF,$DE,$F8,$D9,$7F,$57,$C8,$B6,$D3,$F9,$D7
	.dc.b	$4B,$D8,$FE,$68,$DB,$FF,$49,$79,$DF,$D1,$67,$27,$F1,$FE,$37,$C6
	.dc.b	$FE,$33,$E3,$7F,$1A,$DB,$FC,$27,$8B,$F9,$03,$F2,$3E,$2D,$8F,$E4
	.dc.b	$77,$FD,$3B,$4F,$E6,$0D,$C8,$D3,$5E,$76,$3C,$4F,$CD,$FC,$67,$E7
	.dc.b	$1B,$87,$F2,$3F,$79,$5F,$DB,$DB,$19,$71,$1F,$D6,$2D,$EA,$7F,$55
	.dc.b	$6D,$7C,$26,$E5,$FE,$C7,$CB,$CB,$5F,$D0,$70,$7C,$67,$7F,$8C,$F0
	.dc.b	$37,$1F,$27,$C6,$78,$CF,$27,$8D,$63,$E1,$F4,$F8,$48,$F8,$5F,$E1
	.dc.b	$BF,$2C,$F2,$A9,$27,$F8,$74,$65,$F4,$7A,$9E,$D3,$E1,$7C,$6D,$8F
	.dc.b	$A3,$EA,$7B,$C8,$FB,$DF,$85,$EA,$7C,$9F,$86,$F8,$CD,$9F,$9D,$F0
	.dc.b	$FF,$34,$F9,$DF,$1B,$63,$EE,$FA,$BF,$2C,$8F,$CB,$7E,$47,$C6,$7D
	.dc.b	$5F,$8D,$FA,$CD,$9F,$B5,$F1,$BF,$64,$FB,$5F,$1B,$63,$F1,$3C,$1F
	.dc.b	$5C,$8F,$D7,$7E,$A7,$C6,$FE,$07,$E2,$BF,$04,$D9,$FC,$37,$89,$FE
	.dc.b	$18,$E9,$FE,$49,$D9,$FA,$59,$FF,$48,$4E,$58,$7E,$92,$77,$3F,$D7
	.dc.b	$CF,$F0,$43,$87,$FB,$17,$67,$FB,$8A,$7F,$0D,$F1,$3F,$93,$3F,$29
	.dc.b	$F1,$56,$3F,$31,$1F,$97,$F8,$AC,$5F,$8C,$FF,$54,$3F,$9B,$F8,$A9
	.dc.b	$5F,$E0,$1F,$CD,$3F,$C7,$9E,$5F,$F9,$73,$35,$6B,$0E,$5E,$22,$FF
	.dc.b	$C7,$78,$7F,$AB,$6B,$FC,$A9,$18,$AF,$D4,$AF,$89,$C9,$AF,$8A,$B7
	.dc.b	$E1,$8E,$3F,$15,$8B,$EF,$7F,$E3,$1A,$F3,$4F,$F8,$9E,$31,$3F,$E5
	.dc.b	$26,$BF,$2F,$1F,$34,$EF,$C6,$F2,$9E,$6F,$8A,$B8,$7C,$18,$F4,$FC
	.dc.b	$56,$30,$F0,$BF,$E2,$0F,$94,$7C,$03,$E3,$3F,$BA,$46,$1B,$19,$AF
	.dc.b	$B8,$7F,$86,$E7,$4E,$07,$C7,$BC,$07,$CA,$FF,$7F,$8C,$36,$33,$3C
	.dc.b	$67,$2F,$3A,$FE,$C3,$B7,$93,$E0,$FF,$8C,$CE,$5A,$F9,$27,$6F,$C9
	.dc.b	$F7,$3F,$25,$CD,$FE,$9C,$E2,$FE,$9D,$3F,$85,$F7,$33,$C3,$FE,$F4
	.dc.b	$9F,$EB,$13,$F8,$9B,$CB,$F8,$06,$E5,$F8,$76,$CB,$CC,$BE,$5F,$F8
	.dc.b	$76,$BF,$18,$D9,$34,$C7,$C6,$70,$C3,$F3,$11,$D2,$24,$7E,$6D,$AF
	.dc.b	$E2,$BC,$BF,$0F,$95,$E2,$F8,$B9,$3F,$A1,$8C,$B4,$87,$E4,$BE,$33
	.dc.b	$DB,$E1,$3B,$FF,$C0,$53,$FA,$9F,$2F,$DA,$6A,$33,$1F,$B1,$9E,$1F
	.dc.b	$07,$92,$47,$EC,$7E,$2F,$E4,$AF,$57,$FC,$88,$FF,$90,$5A,$FE,$25
	.dc.b	$AF,$E3,$9A,$FF,$9A,$7A,$BF,$B0,$F5,$F4,$9E,$3C,$EE,$6F,$E2,$11
	.dc.b	$87,$87,$AD,$F9,$EC,$3F,$D3,$78,$3F,$61,$FF,$2D,$C6,$1A,$FF,$0D
	.dc.b	$8D,$7B,$7F,$82,$78,$7B,$BE,$1F,$6B,$FC,$B5,$5E,$D6,$E5,$FE,$58
	.dc.b	$8F,$74,$6B,$97,$FC,$D7,$1F,$E2,$16,$BF,$C6,$C3,$8D,$3F,$31,$E7
	.dc.b	$FD,$1F,$83,$95,$FE,$A8,$E3,$47,$E4,$BE,$33,$E7,$FD,$A3,$D7,$F2
	.dc.b	$B6,$FF,$CE,$0D,$7F,$70,$D7,$F8,$77,$6B,$E3,$7F,$3A,$48,$FF,$56
	.dc.b	$7C,$8F,$CF,$7C,$5F,$0B,$F8,$52,$3A,$F8,$EF,$1F,$93,$E2,$24,$78
	.dc.b	$9F,$F4,$17,$67,$E4,$7E,$13,$F1,$C4,$9E,$FF,$1D,$FB,$0C,$7E,$D4
	.dc.b	$7E,$47,$83,$EB,$F1,$7F,$BC,$5B,$D7,$FE,$AF,$F1,$57,$C2,$BD,$16
	.dc.b	$F9,$77,$C6,$7D,$A6,$C7,$FA,$55,$AF,$F1,$D6,$BF,$CC,$EB,$FC,$EA
	.dc.b	$FF,$B6,$BE,$5B,$FC,$3E,$DF,$86,$F4,$79,$DE,$9F,$90,$F5,$7C,$7C
	.dc.b	$AF,$97,$1F,$A6,$F1,$1E,$DF,$E3,$1B,$3F,$EA,$76,$BF,$CB,$DA,$F8
	.dc.b	$FE,$43,$E3,$11,$FA,$FE,$47,$EF,$BA,$BF,$65,$1F,$6B,$E4,$3E,$61
	.dc.b	$9F,$0F,$F6,$FF,$A4,$7F,$26,$9F,$E7,$8D,$7F,$A0,$B5,$FE,$FC,$D5
	.dc.b	$D3,$FD,$AF,$E2,$BC,$04,$8F,$C4,$F2,$78,$FE,$23,$EE,$78,$72,$7F
	.dc.b	$DC,$DC,$FF,$8D,$C3,$FB,$A5,$7F,$0A,$FC,$6D,$DF,$FB,$06,$DF,$FB
	.dc.b	$2D,$AF,$F5,$36,$BF,$E1,$3A,$FE,$FD,$DF,$F0,$C9,$1F,$9C,$F9,$1F
	.dc.b	$E4,$7F,$84,$3C,$6A,$49,$FF,$4C,$6C,$F0,$FF,$8E,$AF,$92,$FE,$06
	.dc.b	$D9,$FF,$6F,$35,$FE,$CC,$D7,$FB,$5B,$FE,$88,$7A,$C7,$43,$AF,$89
	.dc.b	$FA,$1A,$FD,$14,$49,$FD,$9C,$65,$A4,$3F,$C3,$BD,$FD,$5B,$E5,$FE
	.dc.b	$AC,$8F,$EF,$3C,$DF,$D7,$18,$76,$64,$FE,$03,$F2,$3F,$B3,$F3,$64
	.dc.b	$60,$FE,$0B,$F2,$3F,$B7,$99,$18,$3E,$EF,$8F,$FB,$5F,$CD,$FF,$BC
	.dc.b	$F3,$FF,$0B,$F2,$3F,$6F,$87,$EF,$CF,$ED,$F3,$AA,$FF,$24,$F7,$7E
	.dc.b	$8A,$F8,$4F,$F5,$9C,$3A,$9F,$E9,$E7,$0F,$F5,$79,$E1,$5E,$1F,$1B
	.dc.b	$F4,$9E,$1C,$AE,$39,$30,$FD,$DC,$7F,$91,$9F,$D4,$B7,$E4,$BC,$A7
	.dc.b	$9B,$C0,$F3,$F2,$C9,$E7,$48,$F5,$FC,$8F,$F7,$0F,$C6,$3F,$89,$ED
	.dc.b	$F9,$F6,$EB,$FE,$66,$6E,$01,$F3,$8F,$8B,$FE,$DA,$FF,$97,$23,$FC
	.dc.b	$64,$6B,$E2,$FF,$AA,$C6,$63,$FC,$6C,$D7,$FC,$77,$F3,$9F,$E3,$CC
	.dc.b	$7F,$92,$CE,$DB,$5F,$DE,$A7,$58,$FD,$9F,$0F,$EA,$93,$0F,$7F,$A9
	.dc.b	$F6,$0C,$3F,$26,$30,$FF,$C0,$1F,$BA,$7C,$47,$DE,$F0,$E4,$E6,$3F
	.dc.b	$03,$F2,$9F,$E0,$AA,$DC,$E4,$F9,$B1,$F8,$6F,$95,$91,$F4,$1F,$F0
	.dc.b	$EF,$DD,$BE,$0F,$D5,$EE,$7F,$78,$DB,$FB,$7D,$FC,$86,$BF,$E5,$B6
	.dc.b	$BF,$E7,$1E,$67,$CB,$72,$8F,$C7,$89,$E3,$C5,$F9,$49,$FF,$16,$F8
	.dc.b	$6F,$8E,$FC,$DF,$CB,$4C,$3F,$2F,$E6,$7F,$56,$9F,$CF,$FC,$AF,$06
	.dc.b	$4F,$E2,$A3,$9F,$C6,$7E,$93,$E5,$A4,$F2,$C7,$F9,$C6,$7F,$A4,$F9
	.dc.b	$6D,$CE,$6F,$F5,$71,$E4,$F8,$BF,$EE,$36,$69,$2B,$FC,$F9,$E1,$FE
	.dc.b	$8F,$C4,$E1,$1D,$E7,$F6,$B1,$EE,$F8,$BF,$DB,$FC,$B4,$AF,$24,$65
	.dc.b	$A4,$3F,$59,$E0,$FC,$6F,$8B,$F1,$92,$3F,$D1,$9C,$3F,$DE,$A3,$FC
	.dc.b	$D9,$CB,$F9,$5F,$03,$FD,$97,$AF,$F8,$9D,$B3,$FC,$23,$C1,$FF,$3B
	.dc.b	$FA,$DF,$E3,$6D,$FF,$A1,$E5,$FE,$AB,$8C,$13,$E4,$B7,$10,$FE,$07
	.dc.b	$1F,$EA,$17,$FD,$8E,$FF,$B0,$5C,$FC,$B7,$63,$BF,$CB,$48,$E2,$D7
	.dc.b	$8F,$CB,$79,$3E,$5B,$FD,$0D,$66,$92,$7E,$8C,$7E,$25,$F1,$C7,$17
	.dc.b	$CB,$D5,$FE,$6E,$F9,$6F,$F2,$07,$4F,$F0,$7E,$0E,$15,$FE,$B4,$F1
	.dc.b	$FF,$DE,$91,$8A,$46,$2F,$FD,$F1,$1C,$AB,$F2,$87,$F1,$17,$1F,$13
	.dc.b	$C6,$BF,$DF,$CD,$9F,$E2,$CF,$5F,$25,$8E,$30,$EB,$18,$AE,$A7,$94
	.dc.b	$F4,$47,$8A,$FC,$91,$E1,$9E,$18,$F5,$BD,$CF,$96,$B6,$F3,$3A,$E3
	.dc.b	$EB,$AE,$11,$FE,$9F,$C3,$B3,$E5,$39,$14,$FF,$6D,$E1,$E7,$C7,$8D
	.dc.b	$7B,$E7,$AE,$1E,$0C,$3B,$61,$EE,$C7,$CB,$5E,$3B,$E0,$72,$9D,$47
	.dc.b	$23,$75,$F5,$1B,$8F,$C0,$9D,$E3,$E1,$23,$83,$9D,$F6,$7D,$B5,$F0
	.dc.b	$AB,$93,$F2,$F0,$F7,$61,$DE,$3E,$21,$F0,$8A,$7C,$28,$DC,$FC,$3A
	.dc.b	$EB,$87,$8A,$BA,$1C,$2F,$F1,$5C,$1F,$DC,$9B,$3F,$02,$31,$7F,$83
	.dc.b	$7F,$BB,$D7,$F5,$FF,$9D,$FF,$3D,$57,$F6,$18,$91,$87,$01,$C3,$A1
	.dc.b	$3D,$1D,$1E,$C1,$C3,$FE,$9B,$AF,$D3,$C7,$84,$D4,$67,$1E,$F1,$CA
	.dc.b	$77,$39,$8F,$1D,$FE,$25,$EB,$C9,$5F,$0E,$9C,$23,$F0,$B1,$DE,$30
	.dc.b	$FE,$90,$F1,$5E,$0F,$2B,$E1,$C3,$E0,$E1,$E1,$8F,$81,$5C,$8D,$E1
	.dc.b	$DA,$7F,$D7,$11,$89,$ED,$18,$AF,$6D,$6F,$0F,$2D,$7A,$27,$BC,$62
	.dc.b	$75,$3E,$73,$C8,$7D,$7A,$F2,$3E,$AA,$E2,$FD,$13,$A1,$BA,$E8,$F9
	.dc.b	$1B,$7E,$B9,$F2,$CE,$4F,$05,$FC,$FB,$EF,$1E,$23,$FC,$50,$E2,$3C
	.dc.b	$F5,$F0,$4E,$93,$F8,$6D,$FF,$C2,$78,$7F,$40,$F9,$6F,$F6,$F6,$DF
	.dc.b	$E5,$59,$CD,$7F,$42,$9C,$D7,$FB,$3E,$51,$FE,$CE,$9C,$35,$EC,$DD
	.dc.b	$3F,$D3,$6D,$97,$77,$4A,$F2,$36,$BF,$D4,$3F,$2D,$6B,$78,$7F,$56
	.dc.b	$AC,$7C,$B6,$5B,$1F,$D6,$BB,$F8,$78,$DC,$5B,$1F,$D7,$A3,$C4,$47
	.dc.b	$FA,$8B,$E5,$EE,$F2,$76,$FF,$BB,$3E,$47,$C5,$18,$E0,$45,$3E,$04
	.dc.b	$EE,$FF,$BD,$CE,$9B,$3A,$87,$C8,$73,$3D,$DE,$4D,$AF,$F0,$F4,$7C
	.dc.b	$1F,$96,$C7,$0F,$F7,$47,$E2,$78,$1E,$6F,$05,$F2,$38,$C3,$FC,$5F
	.dc.b	$18,$6C,$79,$A3,$17,$FC,$33,$4F,$F4,$14,$E4,$FA,$DF,$2D,$D7,$E4
	.dc.b	$3E,$93,$70,$EB,$1F,$57,$6C,$61,$FE,$08,$7E,$C7,$C2,$79,$7E,$36
	.dc.b	$DF,$91,$C3,$FE,$10,$DA,$DF,$F8,$41,$FC,$17,$E7,$09,$FC,$34,$BE
	.dc.b	$71,$F0,$9A,$9C,$C6,$1F,$F2,$BB,$67,$F8,$DB,$5F,$E5,$6D,$7F,$98
	.dc.b	$35,$FE,$57,$5F,$EF,$29,$FF,$53,$B6,$3F,$BF,$C7,$FA,$5F,$67,$C2
	.dc.b	$3D,$1F,$4F,$5F,$AB,$87,$D3,$F9,$83,$E4,$71,$8F,$F9,$B2,$30,$D8
	.dc.b	$F8,$27,$02,$BE,$2E,$2F,$8B,$F2,$5F,$19,$B1,$E6,$8B,$A7,$98,$F5
	.dc.b	$CE,$3C,$2F,$99,$1F,$EA,$FA,$78,$FF,$39,$7B,$96,$30,$C8,$FF,$58
	.dc.b	$9A,$2B,$FC,$06,$7E,$C3,$FA,$4F,$C3,$DD,$FF,$AC,$6D,$7F,$A3,$07
	.dc.b	$B7,$99,$FC,$0F,$CA,$49,$FA,$3D,$17,$1F,$67,$8C,$E6,$7C,$A7,$E2
	.dc.b	$3E,$53,$F1,$2D,$C3,$EC,$FC,$9F,$E9,$FA,$3F,$1A,$D8,$FD,$3C,$7E
	.dc.b	$43,$E4,$F1,$97,$6C,$51,$F9,$4F,$92,$FC,$AF,$C7,$DC,$3F,$0A,$E1
	.dc.b	$FC,$3F,$C8,$5C,$39,$3F,$9C,$3F,$3B,$F2,$36,$7F,$1A,$FE,$83,$C0
	.dc.b	$FD,$0E,$76,$7F,$22,$E1,$A7,$E5,$23,$17,$E7,$F9,$0F,$D4,$9F,$AA
	.dc.b	$F9,$1B,$5F,$9A,$8F,$D6,$FC,$8E,$2B,$DC,$FE,$BB,$E7,$9E,$75,$F3
	.dc.b	$70,$DC,$3F,$D8,$73,$FE,$CF,$7F,$DE,$6D,$BF,$F0,$4B,$5F,$F0,$CB
	.dc.b	$5F,$D4,$75,$6F,$B7,$F2,$7E,$2A,$31,$7F,$E2,$3B,$79,$BF,$89,$CB
	.dc.b	$F9,$AF,$8A,$F8,$EA,$7F,$49,$FC,$E7,$FC,$47,$87,$FB,$42,$74,$FF
	.dc.b	$B6,$BD,$97,$7F,$E5,$DB,$1F,$E2,$0D,$91,$87,$FC,$46,$D7,$FC,$6D
	.dc.b	$AA,$FF,$D1,$B4,$9F,$CA,$E1,$CA,$3C,$5E,$DF,$97,$0F,$4F,$87,$27
	.dc.b	$87,$07,$ED,$3C,$06,$0F,$92,$D6,$5F,$BB,$3C,$BC,$9E,$92,$61,$E9
	.dc.b	$C9,$1F,$D5,$34,$7C,$5F,$C5,$F9,$1F,$92,$DB,$F0,$9E,$9C,$FC,$3F
	.dc.b	$25,$F2,$CF,$99,$C5,$67,$79,$FB,$FF,$0A,$7C,$67,$F0,$16,$CF,$CB
	.dc.b	$6C,$FF,$93,$1A,$FF,$97,$F0,$F7,$FC,$9F,$D9,$E3,$FB,$4D,$BE,$F3
	.dc.b	$F1,$3C,$BF,$EE,$BF,$3A,$F7,$BF,$01,$18,$7F,$A3,$FC,$FF,$F7,$1C
	.dc.b	$BF,$CB,$6D,$7F,$CD,$6D,$7F,$CE,$0D,$7F,$CD,$7C,$BF,$C7,$5F,$E5
	.dc.b	$5D,$1F,$7F,$E2,$FF,$1D,$D3,$F9,$27,$F2,$7F,$15,$35,$F3,$E1,$FC
	.dc.b	$71,$FC,$94,$7F,$5A,$F5,$4F,$F2,$76,$C7,$E4,$5B,$8F,$F9,$C9,$AF
	.dc.b	$E4,$3C,$BF,$F3,$7B,$E4,$6F,$67,$E2,$70,$FB,$F7,$F5,$B3,$FE,$4D
	.dc.b	$F0,$9F,$A5,$E5,$93,$E6,$F5,$EE,$5F,$8D,$FC,$FB,$FA,$23,$FE,$82
	.dc.b	$6B,$FE,$92,$6B,$FE,$97,$6A,$C3,$F9,$A7,$C4,$FF,$8A,$B4,$3F,$B5
	.dc.b	$F9,$0C,$3F,$DF,$3E,$B5,$E5,$53,$F7,$5F,$23,$74,$FC,$7E,$4C,$FF
	.dc.b	$01,$1F,$B7,$6C,$79,$F9,$BF,$A9,$46,$23,$FD,$59,$F3,$EF,$D6,$BF
	.dc.b	$82,$C7,$F6,$58,$C7,$5F,$F9,$4D,$FF,$2E,$F8,$96,$B7,$CB,$FD,$A2
	.dc.b	$DF,$53,$FB,$D7,$E4,$7C,$57,$F8,$5A,$31,$3B,$C9,$3F,$E1,$A6,$E9
	.dc.b	$E9,$8F,$39,$1F,$F5,$8F,$1F,$F5,$DF,$8F,$F9,$47,$F1,$DF,$94,$B3
	.dc.b	$F7,$79,$53,$FC,$87,$E5,$33,$A7,$F1,$27,$F9,$27,$6C,$BF,$CA,$FC
	.dc.b	$A5,$AF,$E2,$F3,$FC,$A7,$BB,$1B,$7E,$2D,$FF,$3A,$7E,$79,$F8,$48
	.dc.b	$FE,$2C,$DC,$3F,$84,$3F,$92,$6D,$F0,$38,$FC,$A5,$8F,$DD,$CA,$33
	.dc.b	$1F,$ED,$0E,$54,$E6,$30,$DA,$FF,$71,$CE,$AB,$FD,$1D,$2F,$95,$F3
	.dc.b	$9E,$8E,$BB,$87,$82,$3D,$7B,$63,$1F,$F0,$5D,$FC,$AA,$FC,$7F,$D2
	.dc.b	$7F,$8D,$FC,$0F,$E8,$8D,$AF,$F7,$2B,$5F,$F7,$AB,$5F,$F7,$D7,$CB
	.dc.b	$7F,$8A,$BA,$BF,$A6,$36,$3F,$DD,$AD,$7F,$DF,$AD,$7F,$A5,$09,$7C
	.dc.b	$37,$FB,$FA,$3C,$2D,$7C,$2D,$78,$C7,$0D,$BE,$4F,$CB,$74,$E0,$FB
	.dc.b	$3F,$35,$74,$FB,$2D,$C1,$7F,$4A,$F0,$FD,$4F,$96,$BB,$26,$98,$8C
	.dc.b	$FC,$B7,$F6,$BA,$FB,$6D,$AF,$F6,$0D,$FD,$C8,$C7,$81,$C2,$3F,$0B
	.dc.b	$1D,$A3,$BF,$65,$9F,$C4,$C6,$1F,$C5,$FC,$B7,$93,$E2,$A4,$60,$FD
	.dc.b	$34,$FA,$26,$46,$0F,$C9,$FB,$DE,$CF,$6E,$46,$08,$7E,$1A,$3E,$09
	.dc.b	$F0,$BD,$CF,$91,$E0,$FF,$94,$E3,$FC,$14,$D9,$FD,$DF,$81,$FE,$2C
	.dc.b	$6C,$FC,$3F,$95,$E3,$39,$9F,$E7,$93,$F3,$A7,$E7,$FC,$A4,$9F,$D4
	.dc.b	$BF,$4B,$E5,$38,$E3,$E0,$C3,$FC,$4F,$5E,$08,$C1,$F4,$7E,$53,$EC
	.dc.b	$FC,$A4,$9F,$D9,$C7,$8A,$E4,$7F,$0D,$7F,$03,$E0,$FF,$21,$E9,$FF
	.dc.b	$1B,$E2,$BF,$01,$2F,$DE,$F8,$F0,$E3,$EF,$49,$FE,$31,$1F,$23,$D2
	.dc.b	$E9,$D7,$FE,$AB,$6C,$FF,$4B,$6F,$47,$F0,$7F,$4F,$81,$F3,$A3,$FD
	.dc.b	$C3,$3F,$CA,$1A,$FF,$B7,$5A,$FF,$8E,$1F,$F5,$93,$FE,$0C,$D2,$BF
	.dc.b	$E8,$C6,$E9,$FE,$54,$9B,$87,$F8,$96,$7F,$67,$5F,$E2,$2A,$7E,$D8
	.dc.b	$E5,$FD,$54,$61,$F9,$3D,$FF,$86,$F8,$0F,$F6,$3D,$FF,$3E,$6B,$FE
	.dc.b	$F3,$6B,$FE,$4E,$6B,$FC,$13,$E8,$DF,$1D,$FE,$57,$84,$61,$FF,$2D
	.dc.b	$35,$FE,$16,$7F,$95,$5C,$3F,$E3,$1C,$9F,$27,$F7,$BE,$16,$4F,$C5
	.dc.b	$8F,$CE,$FD,$1B,$E1,$7F,$3A,$D9,$FE,$AE,$D7,$BF,$62,$31,$1D,$FE
	.dc.b	$4F,$F1,$64,$7E,$97,$C9,$7E,$3B,$E4,$FF,$1F,$AC,$9F,$AD,$1F,$87
	.dc.b	$8F,$C0,$12,$3F,$00,$FF,$0F,$D3,$FD,$6F,$1F,$77,$E4,$FF,$32,$7E
	.dc.b	$6B,$E4,$FF,$37,$F2,$7F,$9C,$F9,$39,$5F,$83,$8F,$CA,$7C,$9F,$E3
	.dc.b	$49,$1F,$8D,$7B,$1F,$A4,$23,$F8,$BF,$92,$FD,$3F,$C9,$C8,$FC,$73
	.dc.b	$FC,$B5,$FF,$4C,$6D,$7E,$37,$FD,$B3,$1F,$E9,$16,$BF,$60,$FF,$4F
	.dc.b	$6F,$F7,$E3,$31,$F7,$16,$FC,$E6,$4F,$F8,$A2,$72,$FE,$7F,$E4,$FF
	.dc.b	$76,$47,$F4,$5F,$07,$FB,$E3,$0F,$E8,$E3,$0F,$F9,$2F,$BB,$FD,$1A
	.dc.b	$FE,$A7,$E0,$BF,$82,$9F,$BB,$23,$B6,$BF,$EF,$8F,$CF,$BF,$AC,$77
	.dc.b	$7F,$A7,$DF,$F5,$7C,$7F,$9F,$9A,$EF,$B3,$F9,$6F,$A3,$FE,$A3,$8F
	.dc.b	$F2,$E3,$5F,$C7,$B5,$FF,$5A,$B5,$FF,$5A,$74,$FF,$83,$3E,$4B,$FA
	.dc.b	$A5,$F1,$C3,$FD,$A3,$77,$C2,$DB,$70,$FE,$C7,$57,$0F,$F0,$AE,$DF
	.dc.b	$D9,$23,$F4,$18,$7A,$FB,$3C,$F3,$9F,$91,$B8,$7A,$23,$0F,$BD,$E4
	.dc.b	$47,$E1,$7C,$7F,$5F,$90,$EC,$C8,$F8,$D1,$E1,$F8,$F9,$1F,$25,$FC
	.dc.b	$77,$CF,$A6,$C7,$F0,$78,$34,$93,$FE,$E1,$87,$C7,$F0,$8F,$F3,$8C
	.dc.b	$70,$6C,$7D,$4F,$96,$F8,$4C,$8F,$AB,$19,$69,$0F,$C1,$3F,$73,$E3
	.dc.b	$FE,$57,$CA,$49,$FB,$B1,$F3,$7E,$47,$51,$83,$F0,$CF,$D1,$89,$1F
	.dc.b	$89,$7E,$A7,$C8,$FE,$BF,$EA,$BF,$B6,$4F,$F2,$26,$BF,$E2,$E6,$BF
	.dc.b	$EF,$96,$FA,$0E,$37,$C8,$F3,$F6,$A5,$FE,$5E,$3F,$2B,$F1,$7F,$83
	.dc.b	$F8,$E9,$3F,$96,$8F,$C3,$7C,$76,$A3,$07,$E7,$DF,$F0,$4F,$99,$FC
	.dc.b	$DB,$A7,$F4,$95,$F8,$D3,$35,$F8,$FF,$8E,$AF,$FA,$F6,$B8,$76,$FF
	.dc.b	$BE,$74,$57,$4F,$8C,$FF,$09,$D7,$F9,$96,$FF,$D1,$AD,$7F,$9B,$B5
	.dc.b	$FF,$7A,$57,$FB,$59,$ED,$3C,$DF,$DB,$7C,$77,$F9,$13,$F3,$FF,$0C
	.dc.b	$7F,$92,$17,$C6,$E4,$D4,$43,$FD,$F7,$1C,$38,$4D,$3F,$D0,$91,$ED
	.dc.b	$F8,$AF,$49,$1F,$4B,$FE,$4D,$8F,$56,$CD,$7F,$7F,$A2,$BE,$69,$E7
	.dc.b	$CF,$D1,$F1,$FC,$49,$8F,$FA,$47,$1F,$F6,$BB,$7E,$3F,$6F,$A2,$7E
	.dc.b	$04,$7C,$1F,$8C,$F8,$58,$4A,$F6,$C7,$C5,$F8,$CD,$46,$0F,$82,$FC
	.dc.b	$9F,$8C,$F5,$12,$3C,$4F,$B7,$E2,$B5,$1F,$22,$3E,$84,$7F,$4B,$F8
	.dc.b	$EF,$DD,$EE,$A7,$CE,$FF,$9D,$5C,$4F,$F0,$37,$51,$95,$19,$78,$E3
	.dc.b	$FE,$9A,$F0,$D7,$C2,$30,$F7,$F3,$39,$5F,$16,$F8,$BF,$E9,$A7,$37
	.dc.b	$CB,$04,$62,$FA,$47,$9A,$BF,$D4,$38,$F2,$8C,$9E,$6B,$E5,$1C,$63
	.dc.b	$CE,$71,$37,$87,$AE,$34,$6F,$0E,$93,$E2,$C7,$FD,$6F,$87,$9D,$F6
	.dc.b	$C6,$4F,$93,$87,$82,$37,$1D,$A7,$CF,$1C,$2F,$BB,$A9,$F2,$9F,$3A
	.dc.b	$76,$F4,$C7,$CF,$7D,$63,$D4,$70,$C7,$D5,$7D,$4F,$2E,$3E,$A2,$38
	.dc.b	$35,$3C,$B1,$F7,$9E,$47,$C0,$7B,$4F,$21,$F6,$E3,$D1,$3F,$1D,$E9
	.dc.b	$1E,$48,$EE,$F2,$7E,$97,$27,$F7,$8B,$FC,$9C,$FF,$B2,$A7,$E5,$CE
	.dc.b	$70,$FC,$13,$FD,$FA,$7F,$31,$F3,$DF,$99,$8C,$42,$31,$5F,$E2,$36
	.dc.b	$C7,$F9,$71,$AF,$FB,$A1,$F5,$3F,$EE,$AE,$6F,$E0,$8F,$BC,$EE,$57
	.dc.b	$33,$FA,$38,$CB,$63,$38,$F8,$67,$87,$37,$67,$C0,$7D,$E3,$F9,$0B
	.dc.b	$F9,$87,$BE,$7C,$E3,$11,$ED,$39,$9E,$53,$17,$B7,$18,$73,$C3,$D7
	.dc.b	$7E,$79,$E4,$EC,$C4,$EE,$FB,$B6,$7C,$C7,$0C,$3F,$D0,$B7,$D1,$F7
	.dc.b	$5F,$37,$E0,$61,$E3,$3A,$D7,$A1,$B3,$B8,$FA,$87,$C3,$AE,$8E,$23
	.dc.b	$4F,$C8,$6B,$F4,$2B,$C7,$87,$CD,$8C,$1F,$6D,$C3,$E9,$30,$7B,$30
	.dc.b	$F7,$57,$F2,$1C,$7F,$A6,$FC,$7F,$E0,$B5,$FC,$1C,$7F,$50,$F8,$3C
	.dc.b	$63,$F9,$1E,$65,$E3,$2F,$EB,$95,$6F,$85,$FF,$47,$8F,$C6,$CF,$F3
	.dc.b	$13,$71,$93,$FA,$67,$C7,$F1,$C1,$E0,$D8,$C3,$27,$FB,$2F,$C7,$DB
	.dc.b	$FE,$63,$1E,$58,$E5,$F1,$BC,$FE,$3E,$CF,$F6,$C7,$0F,$F5,$B7,$CB
	.dc.b	$5E,$D6,$BF,$DD,$DB,$9F,$F7,$2F,$8F,$F8,$1E,$93,$BD,$D5,$E9,$9C
	.dc.b	$C6,$31,$FF,$09,$4F,$4A,$E0,$D9,$D6,$3F,$E0,$E8,$EF,$B3,$E3,$FE
	.dc.b	$21,$F1,$7E,$3A,$E3,$FE,$06,$8F,$91,$F1,$B8,$C3,$E1,$9F,$E0,$F7
	.dc.b	$EA,$7E,$81,$FE,$01,$AF,$F1,$27,$16,$1F,$3B,$1F,$F1,$0B,$E5,$23
	.dc.b	$FE,$AF,$D9,$2F,$8F,$F7,$F1,$76,$6D,$7F,$8C,$63,$C3,$F1,$F8,$CB
	.dc.b	$CB,$87,$28,$F2,$FC,$E7,$C8,$E2,$B8,$7F,$8E,$5C,$3F,$DE,$7A,$3F
	.dc.b	$2B,$9B,$E3,$F3,$97,$48,$C5,$6C,$FE,$50,$FF,$BE,$9C,$47,$F8,$0F
	.dc.b	$EA,$DF,$0F,$F1,$BE,$3B,$F3,$92,$9C,$D7,$08,$C5,$7E,$19,$B1,$FE
	.dc.b	$D3,$69,$FE,$D5,$9E,$2F,$E8,$46,$BF,$2B,$E3,$FE,$A9,$F5,$BE,$1E
	.dc.b	$DF,$9A,$7E,$7F,$C3,$7E,$43,$CF,$FA,$13,$9D,$13,$9E,$39,$39,$F8
	.dc.b	$7B,$9F,$C5,$79,$6F,$27,$97,$C7,$5A,$FC,$4F,$6F,$F9,$F6,$BF,$13
	.dc.b	$F1,$D9,$E3,$FA,$2F,$C6,$C1,$E9,$7C,$8F,$9C,$FF,$17,$47,$C7,$32
	.dc.b	$7C,$97,$38,$FE,$58,$38,$7E,$33,$5F,$E1,$0D,$7F,$C7,$4D,$E0,$FC
	.dc.b	$9F,$C7,$FE,$78,$F1,$79,$56,$FE,$37,$C7,$7E,$8B,$E3,$FF,$46,$D8
	.dc.b	$F9,$5D,$1F,$A6,$31,$97,$E2,$DF,$D4,$FC,$7F,$D1,$E2,$B5,$F8,$AE
	.dc.b	$1F,$F5,$14,$66,$3E,$A7,$C7,$DC,$3F,$28,$FE,$CB,$E3,$BE,$D6,$B6
	.dc.b	$B9,$74,$23,$31,$F8,$7B,$63,$FE,$AF,$9F,$E1,$55,$BA,$D4,$E5,$FD
	.dc.b	$B7,$43,$E3,$FF,$80,$9D,$30,$FD,$33,$F8,$BD,$65,$7E,$2F,$E3,$AD
	.dc.b	$7E,$0B,$1F,$CC,$F3,$7F,$44,$F8,$EF,$CC,$DF,$A3,$0F,$E8,$11,$8B
	.dc.b	$FD,$EB,$FE,$5A,$9F,$CD,$23,$7E,$9F,$06,$46,$9C,$B4,$FC,$47,$91
	.dc.b	$F8,$DF,$1F,$F1,$CD,$8F,$C6,$47,$EC,$FA,$0D,$24,$7F,$19,$A7,$C6
	.dc.b	$FA,$0F,$DF,$FA,$7F,$C9,$72,$F4,$C7,$B3,$E3,$65,$FA,$BE,$36,$D7
	.dc.b	$E9,$BC,$DF,$73,$FB,$DE,$4F,$F5,$BE,$DF,$CF,$E7,$33,$FD,$86,$73
	.dc.b	$7E,$3F,$8D,$F9,$3E,$D7,$CA,$6C,$7C,$8F,$67,$E6,$9F,$3B,$E3,$2C
	.dc.b	$7E,$47,$15,$77,$24,$7E,$12,$32,$FC,$1F,$8C,$FA,$FF,$19,$F6,$1B
	.dc.b	$1F,$5B,$E3,$3E,$D9,$F7,$3E,$32,$C7,$E7,$FE,$6D,$EF,$FF,$5B,$6C
	.dc.b	$7E,$49,$BB,$FC,$36,$BF,$E0,$1F,$1F,$F0,$5F,$1B,$F8,$8F,$8C,$FC
	.dc.b	$4B,$6F,$F9,$77,$CD,$FF,$61,$7F,$96,$36,$FF,$48,$D7,$F9,$63,$5F
	.dc.b	$F5,$F7,$1F,$F5,$5F,$8B,$FE,$E1,$E8,$1A,$4B,$FC,$27,$C6,$FE,$6B
	.dc.b	$E3,$7F,$36,$D8,$F9,$D9,$7E,$7A,$BF,$C0,$3C,$68,$C4,$FD,$3A,$FC
	.dc.b	$01,$1F,$A4,$FE,$0E,$F8,$9F,$A8,$F8,$DB,$3F,$6D,$FD,$5F,$A1,$D9
	.dc.b	$B1,$F7,$E3,$F1,$84,$7E,$F3,$FE,$0D,$F2,$3F,$96,$F8,$BF,$E7,$CD
	.dc.b	$7F,$BB,$ED,$FE,$35,$F9,$CC,$E1,$F1,$AE,$CF,$F5,$49,$FF,$3A,$BF
	.dc.b	$DD,$9F,$F0,$F4,$FF,$53,$73,$7F,$E4,$67,$33,$FD,$22,$B8,$BF,$9F
	.dc.b	$6B,$DF,$D1,$F3,$1F,$E2,$C9,$7C,$77,$94,$F3,$7C,$75,$C7,$C1,$F1
	.dc.b	$DF,$E3,$0C,$53,$EB,$8C,$32,$7F,$91,$CE,$9B,$1F,$14,$F0,$74,$7C
	.dc.b	$33,$E2,$7C,$75,$9F,$74,$7C,$7F,$8E,$C6,$9F,$0D,$F1,$1F,$2F,$E1
	.dc.b	$6D,$7C,$57,$0F,$C7,$F8,$EB,$5F,$2D,$FA,$5F,$19,$F4,$FE,$32,$D7
	.dc.b	$CD,$70,$FC,$FF,$87,$B7,$F4,$DF,$B4,$7C,$92,$47,$F8,$D7,$81,$1F
	.dc.b	$37,$E3,$3F,$01,$F1,$1F,$81,$6D,$7D,$B8,$FC,$27,$C6,$63,$0F,$C0
	.dc.b	$3F,$E2,$2F,$97,$7C,$4B,$9B,$FC,$3D,$18,$6C,$66,$BF,$AC,$4E,$5F
	.dc.b	$EB,$5F,$1E,$F2,$DF,$31,$FE,$27,$8C,$36,$33,$3C,$1A,$FF,$8A,$6B
	.dc.b	$FC,$81,$E3,$78,$BE,$13,$FC,$57,$3E,$06,$BB,$6B,$FE,$E9,$E0,$7C
	.dc.b	$7F,$FA,$4F,$B7,$F3,$5F,$19,$78,$74,$DB,$FD,$25,$5C,$58,$72,$9F
	.dc.b	$D2,$4E,$63,$11,$FA,$9F,$8E,$C7,$B7,$FE,$45,$BE,$2D,$A3,$4C,$7B
	.dc.b	$5C,$2F,$F6,$11,$D6,$24,$7E,$CD,$AF,$F9,$B3,$CB,$FF,$57,$E9,$FE
	.dc.b	$BA,$7F,$CE,$D1,$FE,$4B,$5F,$1D,$EB,$6C,$67,$CC,$C1,$1C,$1C,$79
	.dc.b	$38,$70,$71,$F8,$E9,$3D,$63,$2D,$21,$E5,$F8,$DF,$99,$E3,$FF,$A2
	.dc.b	$BA,$FF,$CD,$0D,$9F,$DC,$C6,$3C,$5F,$F6,$1C,$7F,$A4,$73,$7C,$6F
	.dc.b	$29,$CC,$FF,$11,$AF,$64,$FB,$7E,$37,$DD,$F1,$12,$BE,$14,$71,$EE
	.dc.b	$EC,$48,$FE,$12,$FD,$7C,$5F,$1B,$F1,$C9,$3F,$2F,$E3,$7E,$57,$C4
	.dc.b	$7C,$BF,$8C,$93,$F3,$E3,$CD,$CF,$EB,$24,$7F,$16,$7B,$7C,$6F,$D3
	.dc.b	$24,$7D,$7F,$87,$F8,$FF,$0D,$F7,$3C,$04,$FF,$A2,$B4,$7C,$6F,$DB
	.dc.b	$37,$3C,$3A,$BE,$6D,$FD,$EB,$FF,$0B,$35,$FF,$0A,$35,$FE,$D2,$FF
	.dc.b	$00,$7F,$CB,$CD,$7F,$D0,$AD,$EB,$FC,$77,$C6,$F9,$BA,$BF,$19,$F1
	.dc.b	$B2,$F8,$C7,$AB,$9B,$C6,$48,$FF,$2C,$E1,$F7,$1F,$BD,$F1,$BF,$95
	.dc.b	$F8,$D9,$1F,$99,$8F,$DA,$2F,$96,$FA,$AF,$FA,$C5,$AF,$F9,$A5,$AF
	.dc.b	$FA,$C5,$AF,$FB,$2F,$81,$CB,$FE,$B9,$D7,$FD,$86,$D8,$CC,$FF,$9C
	.dc.b	$E7,$2F,$FB,$8F,$ED,$BF,$87,$C3,$83,$EA,$F8,$89,$CC,$61,$B5,$FE
	.dc.b	$C2,$5F,$13,$FC,$35,$FF,$33,$47,$FB,$05,$FF,$2B,$B7,$E3,$35,$B7
	.dc.b	$E1,$39,$BE,$E9,$27,$E5,$BF,$C3,$E9,$F1,$9D,$49,$3D,$7E,$33,$C1
	.dc.b	$F1,$5E,$1F,$8C,$93,$E1,$8F,$C8,$F1,$7E,$28,$91,$F6,$1E,$DF,$19
	.dc.b	$E9,$24,$7A,$7D,$0F,$67,$C5,$FB,$7E,$32,$4F,$B6,$3F,$81,$F9,$DF
	.dc.b	$0A,$3F,$D3,$6D,$7F,$CD,$8D,$7F,$CD,$4D,$7E,$67,$C6,$72,$E2,$91
	.dc.b	$F2,$63,$E6,$7C,$67,$CD,$F8,$89,$3F,$36,$32,$D2,$0E,$7F,$B8,$FD
	.dc.b	$7F,$8B,$FA,$DF,$13,$23,$EB,$3F,$67,$E2,$64,$7D,$96,$BC,$FA,$7F
	.dc.b	$05,$F1,$7F,$7F,$E2,$A4,$FD,$F7,$E1,$F1,$48,$F8,$6F,$F2,$6D,$CD
	.dc.b	$3E,$F7,$C5,$6E,$7F,$DF,$33,$A7,$F1,$51,$F3,$B8,$BF,$AD,$78,$DF
	.dc.b	$C2,$A7,$F8,$5E,$DE,$EE,$EF,$E0,$8F,$F9,$27,$05,$F0,$F7,$B9,$5F
	.dc.b	$FB,$86,$72,$F2,$D7,$F4,$38,$FF,$80,$E3,$E0,$0C,$7F,$A8,$FC,$8F
	.dc.b	$8B,$E2,$E4,$F1,$35,$49,$F3,$FC,$54,$9D,$5F,$F5,$A8,$F5,$FC,$5F
	.dc.b	$F0,$4F,$94,$FC,$FD,$6B,$0F,$F4,$BB,$7E,$2F,$5E,$0E,$FC,$1F,$F7
	.dc.b	$74,$FF,$2C,$B4,$62,$11,$87,$6D,$7F,$DE,$CF,$FA,$C9,$F5,$1C,$23
	.dc.b	$B3,$88,$D3,$F1,$32,$FE,$E7,$F1,$9F,$50,$91,$E9,$F8,$CF,$AE,$47
	.dc.b	$07,$B7,$E3,$3F,$53,$D1,$F7,$3E,$32,$4F,$D5,$8F,$BF,$F1,$9A,$8C
	.dc.b	$1F,$15,$FE,$57,$F8,$53,$DB,$FC,$2F,$C6,$CA,$F5,$CE,$5F,$6F,$CA
	.dc.b	$FF,$93,$9F,$92,$D7,$F1,$87,$F5,$67,$0E,$1A,$FE,$46,$1E,$7F,$F3
	.dc.b	$C9,$E5,$D9,$FB,$0A,$46,$3A,$78,$57,$F8,$2B,$D8,$EF,$7F,$99,$E6
	.dc.b	$D6,$1F,$8A,$F8,$ED,$3F,$A0,$F8,$0A,$FE,$3F,$E0,$98,$72,$CB,$F3
	.dc.b	$35,$D7,$B7,$F4,$FF,$1D,$30,$E1,$7F,$E9,$87,$F1,$DF,$03,$B7,$35
	.dc.b	$FA,$E8,$F3,$76,$7F,$87,$AB,$F2,$75,$F1,$B2,$F0,$CE,$4F,$F2,$36
	.dc.b	$9F,$E4,$AD,$7F,$D7,$57,$C1,$B3,$FB,$AF,$8A,$FD,$DF,$C5,$FE,$F3
	.dc.b	$E2,$E6,$5F,$A1,$8C,$B4,$87,$E4,$71,$FF,$28,$BE,$CE,$8F,$31,$1F
	.dc.b	$B6,$33,$F8,$F1,$E6,$F8,$BF,$A4,$C9,$FD,$84,$7F,$99,$FE,$6F,$ED
	.dc.b	$47,$FC,$01,$6B,$FD,$99,$AF,$F0,$56,$F7,$1F,$F2,$CC,$7E,$9D,$FF
	.dc.b	$11,$3B,$E3,$ED,$F1,$9D,$F2,$F0,$7C,$64,$AF,$E0,$F7,$FE,$21,$E0
	.dc.b	$FE,$3C,$DC,$3C,$57,$C6,$B7,$F1,$9E,$9F,$82,$FF,$3C,$59,$A4,$9F
	.dc.b	$8B,$9F,$9F,$E3,$38,$47,$82,$7E,$14,$7F,$86,$3E,$3F,$F9,$CE,$7F
	.dc.b	$E3,$AF,$96,$FC,$1C,$3D,$DF,$E6,$91,$86,$F3,$7F,$37,$9E,$38,$AB
	.dc.b	$8F,$A3,$FD,$5E,$3F,$C1,$75,$BB,$FE,$7D,$EC,$F3,$9F,$F2,$4B,$C3
	.dc.b	$1B,$5D,$6F,$96,$1B,$BE,$D3,$CD,$F4,$3B,$AB,$1F,$E5,$4C,$7C,$B5
	.dc.b	$C4,$95,$B9,$E6,$6E,$FD,$18,$77,$24,$ED,$F3,$4F,$01,$7F,$1D,$F0
	.dc.b	$98,$C7,$E3,$BD,$71,$CC,$F9,$2F,$D5,$3C,$E3,$93,$70,$F8,$33,$E7
	.dc.b	$C7,$D3,$3E,$DC,$3A,$BE,$2A,$DB,$DC,$FA,$26,$CF,$9A,$53,$27,$1C
	.dc.b	$32,$D3,$C9,$3D,$B0,$F2,$DF,$C2,$3E,$1E,$3F,$1A,$BA,$38,$C7,$E4
	.dc.b	$1E,$E7,$51,$D5,$F9,$7A,$7F,$8B,$6F,$F1,$6D,$CB,$F1,$4D,$CB,$FC
	.dc.b	$B7,$5F,$C0,$BE,$4E,$E1,$FC,$F9,$AF,$F8,$2D,$AF,$FA,$59,$A9,$F7
	.dc.b	$AF,$47,$FC,$5B,$3F,$C9,$23,$81,$F1,$8D,$71,$70,$79,$19,$9E,$39
	.dc.b	$F9,$63,$13,$BB,$FC,$2C,$72,$3F,$14,$76,$D3,$C1,$7F,$E3,$E7,$C5
	.dc.b	$1D,$B0,$CD,$F3,$8E,$A7,$94,$F1,$CE,$9E,$A6,$30,$FE,$C7,$7D,$5F
	.dc.b	$34,$EC,$EF,$87,$9A,$75,$5F,$34,$F3,$B8,$7E,$71,$8B,$F6,$BE,$7B
	.dc.b	$EF,$1E,$7C,$3C,$87,$69,$E0,$79,$AF,$C1,$3C,$9B,$19,$37,$87,$1A
	.dc.b	$E7,$7F,$3D,$B3,$E7,$9F,$64,$FB,$8F,$1C,$7F,$00,$CB,$F9,$E7,$C3
	.dc.b	$7F,$9B,$33,$FE,$67,$E3,$5C,$FF,$A2,$4D,$BF,$E8,$93,$6F,$FA,$23
	.dc.b	$FD,$29,$B5,$FC,$CB,$8B,$FC,$34,$D9,$D4,$3E,$3B,$93,$67,$FA,$2F
	.dc.b	$C7,$62,$64,$7F,$48,$F8,$6B,$8F,$F5,$17,$8F,$C6,$F2,$F8,$EB,$5F
	.dc.b	$D3,$9C,$3E,$37,$C2,$47,$CD,$F0,$DE,$DE,$FF,$F0,$8F,$99,$E4,$8C
	.dc.b	$78,$88,$A7,$F6,$89,$DD,$7F,$74,$6E,$5F,$E0,$BA,$E1,$3F,$0F,$CA
	.dc.b	$F7,$9F,$03,$E3,$6C,$7F,$64,$8F,$87,$F1,$B8,$BF,$ED,$CF,$C6,$F8
	.dc.b	$DF,$1F,$93,$6F,$FC,$42,$E1,$F1,$62,$8F,$33,$63,$31,$F4,$BE,$1F
	.dc.b	$DB,$E4,$7D,$26,$CF,$F8,$B2,$3E,$AF,$C6,$E3,$2F,$F0,$03,$F8,$1F
	.dc.b	$D0,$7F,$D1,$55,$FE,$32,$F2,$71,$D3,$F2,$19,$1F,$E8,$69,$FF,$76
	.dc.b	$BB,$31,$DD,$F8,$DC,$1F,$8D,$3E,$3B,$1F,$30,$E3,$D4,$E7,$0F,$F3
	.dc.b	$EC,$E5,$B3,$FE,$1A,$8C,$5F,$F8,$A1,$AF,$F9,$91,$AF,$FA,$C6,$BF
	.dc.b	$DE,$B5,$E7,$F8,$FE,$FE,$F7,$BD,$F8,$18,$FE,$1F,$E3,$DE,$6E,$2F
	.dc.b	$FC,$B5,$18,$6C,$7C,$7B,$E0,$D8,$F8,$9E,$FF,$CB,$3E,$67,$C7,$D8
	.dc.b	$F9,$1E,$67,$CF,$31,$A7,$FA,$5F,$83,$FC,$2D,$3F,$E9,$EF,$97,$FF
	.dc.b	$5B,$37,$D6,$F9,$91,$8A,$FF,$60,$85,$1A,$3F,$C8,$2E,$27,$FD,$66
	.dc.b	$D9,$FF,$28,$B5,$FF,$46,$AF,$83,$FC,$27,$9D,$E0,$6D,$7E,$07,$E4
	.dc.b	$3F,$10,$7E,$27,$E4,$2C,$7E,$17,$E2,$3F,$19,$F2,$1F,$8D,$6C,$7E
	.dc.b	$27,$E3,$FE,$87,$95,$F9,$16,$C7,$D0,$8F,$CA,$7C,$86,$35,$F8,$CF
	.dc.b	$E5,$FE,$43,$EC,$79,$76,$FE,$BB,$87,$F7,$5C,$E8,$FD,$F3,$63,$31
	.dc.b	$FC,$7F,$89,$F2,$1F,$A3,$39,$63,$F5,$5F,$D3,$7C,$84,$BF,$D1,$7C
	.dc.b	$3D,$AF,$B5,$F2,$1F,$AB,$31,$EA,$FF,$C0,$1C,$51,$FA,$FF,$90,$98
	.dc.b	$7F,$BC,$3F,$41,$61,$83,$F0,$93,$FC,$D6,$3F,$B7,$61,$CE,$31,$ED
	.dc.b	$E7,$E4,$3F,$35,$E5,$DC,$7F,$32,$E1,$F8,$6F,$C2,$23,$FE,$6E,$C3
	.dc.b	$F0,$98,$23,$C7,$E4,$A7,$F0,$91,$E7,$ED,$F2,$75,$F9,$7E,$42,$4F
	.dc.b	$AF,$DB,$BB,$32,$E8,$FF,$A7,$FC,$3F,$E0,$0F,$FA,$DD,$AF,$C2,$35
	.dc.b	$E2,$FF,$35,$8C,$C7,$93,$37,$C7,$FC,$5F,$8F,$F8,$CD,$CB,$E6,$7C
	.dc.b	$7F,$C9,$3E,$57,$C3,$D8,$FF,$6B,$62,$F1,$3B,$12,$3F,$A1,$46,$5F
	.dc.b	$8D,$F1,$FF,$4F,$D9,$FA,$8D,$8F,$B5,$F1,$BF,$5C,$FB,$1F,$1B,$63
	.dc.b	$FD,$C9,$F3,$0D,$7F,$A8,$36,$3F,$A8,$37,$7F,$F6,$D3,$5F,$F8,$03
	.dc.b	$EC,$3E,$33,$FD,$7A,$DA,$FF,$65,$CF,$FB,$C2,$72,$7F,$7B,$E1,$7C
	.dc.b	$77,$E2,$FD,$AF,$C6,$36,$FE,$1F,$8D,$F3,$C8,$FC,$27,$E9,$78,$9F
	.dc.b	$93,$F8,$EF,$ED,$DC,$5F,$A8,$AF,$CD,$46,$3C,$49,$1F,$9D,$F6,$EE
	.dc.b	$BF,$83,$8F,$CF,$FB,$FF,$CC,$3C,$4F,$BC,$48,$FA,$2F,$F7,$5F,$84
	.dc.b	$FD,$2C,$7E,$9F,$83,$FB,$F4,$62,$FA,$0F,$0B,$F3,$13,$FA,$9F,$94
	.dc.b	$97,$FA,$FF,$76,$DF,$0C,$BF,$DC,$9E,$27,$72,$57,$F8,$0F,$2F,$E4
	.dc.b	$D5,$FA,$49,$FD,$34,$FE,$E3,$C1,$FC,$EE,$7F,$CC,$BE,$3F,$FC,$59
	.dc.b	$5A,$BF,$2C,$E6,$F8,$1C,$F8,$76,$70,$F8,$FB,$1A,$A6,$7F,$EF,$9F
	.dc.b	$17,$E2,$12,$7C,$CF,$F8,$E7,$17,$C7,$F8,$BE,$3F,$C6,$DA,$E9,$9F
	.dc.b	$F4,$8F,$8F,$F9,$7D,$DE,$86,$CF,$87,$27,$07,$E4,$F4,$FB,$2D,$7C
	.dc.b	$AD,$8D,$DF,$8F,$E3,$3E,$19,$F1,$3E,$32,$E1,$E2,$D5,$3F,$23,$D5
	.dc.b	$CE,$5F,$22,$7F,$C9,$8D,$F8,$CF,$9A,$7C,$EF,$8C,$B8,$73,$8F,$A3
	.dc.b	$EA,$E2,$B9,$3F,$53,$E3,$25,$78,$7E,$32,$D7,$79,$FB,$3F,$19,$8E
	.dc.b	$9F,$BB,$9A,$33,$87,$F1,$7B,$FC,$24,$7D,$12,$46,$63,$F2,$93,$FE
	.dc.b	$1A,$AE,$7F,$11,$F8,$6F,$8C,$FC,$3B,$2F,$E2,$7C,$65,$C3,$E1,$3F
	.dc.b	$E2,$DF,$96,$75,$FF,$AF,$E7,$B4,$63,$D9,$FE,$06,$D8,$CC,$ED,$F4
	.dc.b	$7C,$83,$E1,$1F,$2F,$FC,$2A,$30,$D8,$CC,$FA,$E7,$2F,$83,$B7,$F6
	.dc.b	$0D,$ED,$F0,$C6,$23,$E3,$63,$F9,$49,$F8,$87,$F8,$63,$DB,$FE,$81
	.dc.b	$C1,$F3,$BE,$3E,$EF,$FA,$F6,$E1,$F2,$AA,$DF,$18,$FC,$54,$63,$88
	.dc.b	$EB,$1F,$16,$3F,$B6,$FA,$A7,$B3,$FC,$A3,$14,$63,$1F,$2B,$5F,$BD
	.dc.b	$1A,$9E,$13,$FC,$EE,$7F,$56,$EE,$BB,$4F,$9A,$35,$3F,$C0,$9B,$3E
	.dc.b	$3E,$ED,$1B,$F6,$7D,$1F,$21,$E9,$F8,$E9,$87,$EE,$33,$4E,$A3,$0D
	.dc.b	$7F,$87,$C7,$F9,$C5,$FD,$D7,$87,$FE,$8B,$79,$7C,$77,$C5,$3E,$6F
	.dc.b	$89,$FB,$EF,$52,$D6,$EB,$F9,$33,$74,$FE,$F7,$5B,$9F,$27,$C7,$7D
	.dc.b	$67,$FD,$01,$DB,$FD,$51,$A7,$F5,$6F,$7D,$19,$6C,$7E,$9F,$A9,$EC
	.dc.b	$FF,$82,$23,$FC,$BF,$3A,$6D,$7E,$12,$B7,$D1,$F7,$FE,$3A,$57,$C7
	.dc.b	$75,$F5,$EF,$0F,$CD,$8B,$C1,$F1,$12,$FC,$CF,$F6,$CC,$7C,$7E,$07
	.dc.b	$7D,$3F,$63,$5F,$CD,$67,$2F,$D5,$F8,$FF,$93,$E9,$C8,$FA,$2F,$E4
	.dc.b	$FE,$07,$E3,$7D,$7A,$BE,$33,$6B,$FB,$DB,$5F,$E1,$4D,$7F,$C4,$DD
	.dc.b	$CE,$8F,$DD,$47,$E1,$7C,$4F,$D0,$1F,$A1,$F8,$FF,$B7,$EA,$C9,$FC
	.dc.b	$1B,$FA,$5F,$8E,$F1,$E7,$E2,$24,$F8,$9F,$F0,$6F,$47,$E4,$3C,$CF
	.dc.b	$C3,$F7,$7F,$9B,$6B,$FA,$33,$5F,$C8,$FC,$7F,$EC,$7E,$3F,$F1,$BF
	.dc.b	$18,$F6,$69,$9A,$FE,$BC,$3F,$0B,$7B,$BF,$89,$6C,$7F,$87,$9A,$FF
	.dc.b	$2D,$6B,$F1,$38,$5F,$21,$F9,$CF,$19,$ED,$53,$F8,$24,$7F,$05,$F8
	.dc.b	$2F,$E0,$DF,$05,$30,$FD,$74,$7E,$6F,$E6,$9E,$C7,$E6,$DB,$3F,$E3
	.dc.b	$06,$BF,$CF,$DA,$FD,$6F,$A0,$FF,$56,$3F,$DC,$1A,$FF,$81,$1B,$E4
	.dc.b	$FF,$82,$B4,$7C,$8F,$C8,$F5,$7F,$8D,$7C,$24,$C3,$FB,$A7,$1F,$F7
	.dc.b	$E7,$E4,$CF,$EF,$DF,$EA,$0D,$7F,$AB,$35,$FF,$37,$35,$FF,$06,$C3
	.dc.b	$E0,$3E,$B9,$23,$F8,$87,$C8,$FD,$AF,$81,$FE,$59,$F0,$B2,$7C,$71
	.dc.b	$FC,$07,$EC,$3F,$94,$47,$F6,$56,$BF,$E7,$D6,$BF,$EA,$06,$F6,$B1
	.dc.b	$F9,$7B,$7F,$0E,$6F,$9F,$FE,$48,$6D,$FF,$8F,$FF,$41,$FE,$E5,$5F
	.dc.b	$E3,$27,$FA,$F3,$7D,$EF,$E8,$F1,$8B,$FF,$4E,$AF,$93,$FE,$31,$EA
	.dc.b	$7F,$0F,$C3,$FB,$E3,$5F,$F4,$A3,$5F,$F5,$BB,$7C,$85,$7F,$BC,$A7
	.dc.b	$F3,$73,$FC,$67,$1F,$E3,$4D,$C3,$C3,$EE,$F9,$3E,$4F,$CB,$F0,$F3
	.dc.b	$0F,$F1,$AE,$3F,$84,$BF,$3D,$6E,$72,$FB,$3E,$4F,$DB,$F2,$72,$78
	.dc.b	$BF,$03,$DE,$91,$82,$1F,$E5,$4C,$3F,$A9,$E1,$F1,$3E,$03,$E3,$FB
	.dc.b	$F2,$BC,$71,$F2,$BE,$4B,$E5,$B2,$3C,$F1,$F3,$BE,$4F,$51,$83,$D6
	.dc.b	$FF,$02,$F6,$7E,$9F,$C9,$C9,$FC,$DC,$7D,$6F,$93,$D4,$60,$F8,$6F
	.dc.b	$DA,$F9,$1F,$B7,$F0,$72,$7E,$3C,$7D,$EF,$91,$91,$F2,$9F,$C0,$F9
	.dc.b	$DF,$D7,$BE,$C1,$F2,$54,$95,$F6,$3E,$0B,$F0,$7F,$1D,$FD,$67,$A8
	.dc.b	$D2,$4F,$FA,$C2,$3F,$C8,$8F,$E0,$23,$8B,$F8,$2F,$8E,$FF,$0B,$FC
	.dc.b	$CB,$E3,$E9,$27,$FC,$E7,$1F,$D7,$7B,$BF,$D1,$6D,$E7,$FF,$79,$BC
	.dc.b	$67,$75,$D3,$E2,$7F,$BC,$FC,$67,$F0,$89,$FF,$3E,$FD,$67,$F8,$19
	.dc.b	$BE,$E6,$BC,$FF,$F1,$4C,$BD,$8C,$57,$F9,$FE,$F9,$3F,$E9,$89,$D3
	.dc.b	$FE,$83,$FB,$0F,$CD,$3F,$E9,$46,$F9,$47,$30,$F9,$F7,$9F,$FE,$05
	.dc.b	$5E,$7F,$F8,$15,$B3,$FE,$DA,$6B,$FE,$69,$6B,$FE,$6F,$FA,$EF,$DF
	.dc.b	$F0,$A2,$4F,$0E,$2F,$F0,$43,$67,$52,$AE,$2F,$1F,$B0,$78,$FF,$13
	.dc.b	$E1,$A4,$FF,$54,$9C,$BF,$33,$E1,$BF,$C3,$D1,$88,$FF,$68,$87,$19
	.dc.b	$7F,$B1,$37,$57,$F8,$BB,$99,$E5,$7D,$2F,$81,$98,$FF,$AB,$E7,$2F
	.dc.b	$DA,$F7,$B8,$54,$8C,$1F,$EB,$5F,$99,$FA,$AF,$FA,$4F,$91,$18,$8F
	.dc.b	$83,$18,$E6,$FE,$F7,$39,$9F,$E4,$7F,$15,$FE,$42,$8E,$7A,$FF,$8F
	.dc.b	$9B,$C5,$AA,$F0,$7C,$97,$F2,$CF,$06,$46,$08,$7F,$BB,$69,$3D,$4F
	.dc.b	$E6,$7F,$13,$DF,$E2,$FC,$64,$AF,$F5,$46,$9E,$A8,$FF,$3B,$7E,$83
	.dc.b	$F8,$EC,$3F,$C4,$99,$F9,$67,$FD,$07,$3F,$94,$F5,$8F,$F8,$3F,$0F
	.dc.b	$F4,$7B,$5F,$F3,$DB,$5F,$F1,$5B,$5F,$C6,$47,$2F,$9A,$EB,$F1,$12
	.dc.b	$7F,$DD,$34,$9F,$0F,$C8,$78,$BE,$02,$56,$A3,$2D,$20,$C9,$C1,$E0
	.dc.b	$C3,$FC,$2F,$3D,$BE,$03,$E0,$91,$F5,$BF,$13,$E6,$BE,$0B,$23,$CB
	.dc.b	$1F,$E7,$3F,$98,$7C,$77,$79,$74,$FE,$39,$4F,$F9,$3E,$30,$DE,$8F
	.dc.b	$D4,$BC,$B3,$FD,$84,$EA,$FA,$78,$FC,$38,$BF,$5E,$DC,$7F,$6C,$EE
	.dc.b	$EE,$3D,$2F,$C9,$86,$EF,$CB,$3E,$0C,$3F,$73,$3C,$67,$27,$F0,$1C
	.dc.b	$7C,$8E,$0F,$5C,$F0,$AE,$51,$E8,$7C,$18,$58,$E5,$87,$AA,$77,$1F
	.dc.b	$EA,$E6,$E3,$E1,$3C,$58,$F9,$C9,$1D,$27,$C5,$8F,$92,$73,$1E,$7C
	.dc.b	$7C,$8F,$83,$1F,$13,$F4,$B1,$F8,$78,$78,$AF,$8C,$E6,$B7,$87,$C0
	.dc.b	$C7,$E4,$12,$FC,$4F,$99,$E4,$4A,$F0,$1E,$6C,$3E,$03,$E5,$3E,$A9
	.dc.b	$4F,$84,$70,$9F,$85,$3F,$10,$91,$97,$E1,$71,$7F,$51,$8F,$40,$71
	.dc.b	$7F,$61,$B7,$F9,$57,$84,$FF,$3D,$FA,$0F,$D9,$E1,$FC,$EA,$30,$73
	.dc.b	$70,$A3,$6E,$DA,$FF,$42,$7D,$AF,$F4,$9F,$81,$FD,$8B,$EE,$34,$50
	.dc.b	$EA,$7F,$51,$1E,$26,$C6,$71,$F1,$D7,$6A,$E3,$C1,$5F,$31,$F9,$73
	.dc.b	$25,$33,$A7,$48,$FC,$2C,$77,$AF,$85,$7D,$E3,$DC,$79,$DC,$3E,$AA
	.dc.b	$E4,$F5,$BE,$73,$F0,$27,$D7,$3E,$0A,$EB,$7E,$08,$C1,$FD,$AA,$B8
	.dc.b	$4F,$BA,$3D,$51,$87,$C7,$5E,$03,$AD,$FC,$39,$F2,$CF,$AC,$E9,$7F
	.dc.b	$16,$7C,$AF,$85,$B1,$E9,$3B,$C6,$5A,$67,$0F,$79,$E0,$7E,$24,$75
	.dc.b	$77,$79,$3C,$D8,$7F,$31,$F8,$4F,$C1,$EB,$F8,$4C,$FF,$A5,$FC,$DE
	.dc.b	$31,$D1,$FD,$09,$C6,$3F,$9D,$E2,$47,$FA,$26,$B8,$D7,$F5,$28,$C4
	.dc.b	$EA,$3F,$AA,$FC,$57,$2A,$FE,$B2,$D9,$7C,$DD,$89,$5F,$D4,$BE,$46
	.dc.b	$E3,$FD,$2E,$3F,$A5,$47,$1F,$53,$97,$A7,$67,$FA,$CB,$87,$FB,$2B
	.dc.b	$E4,$23,$FD,$A3,$D3,$F6,$F7,$7F,$83,$3C,$9F,$24,$63,$BD,$14,$FF
	.dc.b	$54,$4E,$EB,$FB,$DC,$E9,$B3,$A8,$7C,$97,$33,$E0,$7A,$96,$BF,$C3
	.dc.b	$D1,$F0,$FD,$4C,$6B,$EF,$3E,$83,$F1,$FD,$CF,$27,$93,$6F,$FC,$4C
	.dc.b	$E1,$F1,$BE,$12,$3F,$C2,$B8,$DE,$27,$FA,$1A,$3F,$C9,$DC,$DF,$C1
	.dc.b	$B6,$79,$8F,$07,$EC,$5E,$F4,$F2,$56,$A7,$E4,$D6,$F0,$E6,$7F,$83
	.dc.b	$AD,$E3,$6A,$B9,$BF,$83,$FA,$AF,$F0,$75,$70,$C7,$FC,$C1,$F5,$EF
	.dc.b	$1B,$5E,$47,$E4,$A5,$39,$AE,$11,$8A,$E8,$6E,$75,$1F,$2F,$66,$18
	.dc.b	$DB,$F9,$2C,$FF,$85,$BC,$4E,$D8,$FD,$B8,$FF,$30,$BF,$E0,$E3,$B9
	.dc.b	$99,$E2,$EE,$3F,$9D,$35,$FF,$3A,$B5,$FF,$3E,$37,$77,$D7,$BE,$03
	.dc.b	$F2,$6D,$9E,$FE,$86,$27,$E4,$77,$7D,$0E,$6B,$5F,$27,$0F,$B3,$19
	.dc.b	$8F,$99,$AF,$C0,$EF,$C6,$7F,$0F,$6F,$D8,$D6,$9F,$F5,$D3,$6F,$EB
	.dc.b	$74,$FD,$76,$46,$4E,$11,$8A,$FF,$43,$FD,$0B,$E0,$7E,$C2,$F8,$0B
	.dc.b	$18,$65,$FF,$5E,$A5,$6E,$30,$D8,$FE,$C6,$D7,$FD,$03,$C1,$9E,$1F
	.dc.b	$CD,$36,$BF,$D4,$B4,$F8,$1F,$9B,$87,$F1,$1F,$9E,$95,$FE,$60,$F9
	.dc.b	$FB,$97,$E1,$FE,$73,$FB,$E4,$FC,$A5,$E6,$7F,$1B,$73,$3B,$8C,$7A
	.dc.b	$DF,$C2,$7E,$46,$DF,$FA,$6D,$C3,$CE,$ED,$F7,$7A,$7A,$BD,$7E,$0A
	.dc.b	$D7,$85,$C3,$E3,$F9,$EB,$7E,$77,$C8,$79,$7D,$5B,$3E,$97,$D1,$DB
	.dc.b	$E9,$F2,$6C,$FB,$5C,$34,$F8,$11,$8B,$F8,$9D,$DF,$04,$F8,$5E,$45
	.dc.b	$AF,$8D,$1F,$17,$E6,$B1,$5F,$2D,$F9,$1E,$DF,$EE,$FD,$5F,$E1,$13
	.dc.b	$F3,$E3,$1E,$3C,$C3,$F0,$95,$AA,$FF,$10,$4F,$EE,$1F,$F1,$03,$FD
	.dc.b	$4F,$81,$E5,$7D,$53,$EB,$7C,$BD,$C3,$F4,$DF,$23,$FE,$D6,$6F,$AD
	.dc.b	$F6,$FE,$63,$FA,$1C,$E7,$2F,$F7,$0F,$43,$A7,$C1,$1F,$E8,$06,$BF
	.dc.b	$84,$6C,$6F,$0F,$F5,$FD,$3E,$43,$F1,$1F,$11,$D9,$B3,$E4,$F8,$FF
	.dc.b	$09,$F8,$DE,$BB,$1F,$C7,$BD,$B7,$C9,$78,$1B,$1F,$A9,$6E,$9F,$E4
	.dc.b	$86,$BF,$EE,$6C,$9F,$1D,$F9,$9F,$89,$F7,$B6,$FF,$83,$EA,$EE,$FE
	.dc.b	$F0,$D8,$FD,$17,$C4,$E2,$7F,$9A,$7D,$03,$E1,$BE,$1B,$6B,$F7,$8D
	.dc.b	$CF,$FC,$B6,$D4,$7F,$B5,$3E,$73,$F9,$2B,$FC,$DD,$FF,$24,$F9,$5F
	.dc.b	$E1,$77,$2D,$F8,$3F,$F0,$64,$3E,$57,$FB,$DF,$B0,$E8,$CC,$61,$FE
	.dc.b	$95,$F4,$3F,$C8,$31,$FF,$36,$B5,$FF,$39,$35,$F2,$87,$0F,$F9,$FE
	.dc.b	$DC,$5F,$88,$F2,$FC,$6D,$D7,$F9,$A7,$A5,$A6,$C7,$A3,$E5,$BF,$C2
	.dc.b	$33,$9C,$3F,$C0,$03,$E8,$7E,$1B,$FE,$85,$6B,$E4,$6B,$F2,$9B,$EB
	.dc.b	$F9,$3E,$4F,$F3,$67,$C6,$EA,$B9,$7F,$44,$D3,$F3,$DC,$5F,$37,$C0
	.dc.b	$F9,$7F,$1D,$F3,$1F,$EE,$19,$7F,$9E,$B3,$F1,$72,$7E,$82,$31,$1E
	.dc.b	$38,$FF,$12,$EF,$FE,$07,$C3,$FC,$18,$DC,$BE,$F7,$7F,$EB,$BE,$37
	.dc.b	$FB,$E7,$D6,$FF,$2C,$9F,$EF,$6D,$C7,$F0,$5F,$2F,$AE,$DF,$1F,$CB
	.dc.b	$F8,$49,$7F,$D2,$75,$47,$F9,$57,$81,$E1,$FE,$18,$FF,$2D,$FD,$03
	.dc.b	$EB,$9E,$56,$71,$F0,$FA,$BF,$03,$C4,$F2,$B6,$3F,$C1,$F9,$FE,$A9
	.dc.b	$F0,$4F,$C3,$23,$FC,$03,$E1,$3F,$85,$9F,$96,$D2,$C7,$F1,$28,$FC
	.dc.b	$CC,$E3,$7F,$D0,$3F,$9C,$F8,$29,$5F,$5F,$DD,$B5,$FA,$59,$FD,$0F
	.dc.b	$7E,$39,$FF,$4C,$FF,$9B,$FE,$B3,$F1,$91,$FE,$7F,$6E,$1F,$B0,$F8
	.dc.b	$24,$3C,$1E,$1D,$52,$BF,$C1,$33,$FC,$32,$F9,$4F,$5E,$E9,$5F,$E4
	.dc.b	$DA,$FE,$1F,$9F,$56,$BF,$EE,$56,$E2,$E6,$FE,$E9,$5F,$93,$F7,$7F
	.dc.b	$77,$F3,$5F,$BC,$64,$7E,$5F,$C1,$B8,$FF,$05,$7F,$C8,$9F,$3D,$FE
	.dc.b	$47,$F0,$3F,$C1,$4D,$9F,$F7,$4B,$5F,$F7,$73,$5F,$DE,$FC,$9F,$F9
	.dc.b	$FB,$AF,$FC,$24,$D8,$FF,$78,$B5,$FF,$7B,$35,$FE,$0B,$2F,$95,$6B
	.dc.b	$E2,$6C,$F8,$9A,$F8,$9A,$FF,$21,$F3,$3F,$87,$F1,$7F,$1F,$F0,$AE
	.dc.b	$BF,$C9,$5B,$5F,$C4,$6A,$E1,$F2,$FA,$6F,$0F,$F3,$3A,$FD,$F5,$62
	.dc.b	$77,$F4,$5F,$D6,$2C,$D2,$63,$FE,$33,$8C,$FC,$5F,$08,$98,$7F,$93
	.dc.b	$2B,$FD,$16,$DF,$37,$AD,$48,$C1,$BF,$8B,$F0,$ED,$23,$07,$F9,$67
	.dc.b	$7F,$94,$E0,$FF,$3C,$61,$FE,$83,$75,$3E,$AF,$A2,$F5,$FC,$BC,$9F
	.dc.b	$24,$65,$A4,$39,$FB,$7F,$2B,$CD,$62,$F9,$7F,$F2,$94,$61,$B1,$9A
	.dc.b	$FE,$86,$7F,$58,$9D,$3F,$EA,$DC,$B4,$FF,$A9,$A6,$F2,$7F,$AB,$5B
	.dc.b	$BF,$EC,$6A,$D7,$E2,$1F,$F7,$6B,$F3,$7E,$5F,$EB,$9F,$63,$E3,$BD
	.dc.b	$16,$7C,$A4,$9F,$28,$C7,$DB,$1F,$6E,$FF,$DD,$F3,$F6,$7E,$5B,$F0
	.dc.b	$67,$E1,$3E,$3F,$AF,$CB,$49,$FA,$8F,$E2,$0F,$0F,$CB,$7C,$42,$47
	.dc.b	$83,$1F,$97,$1F,$23,$E5,$BF,$1E,$C9,$FB,$D1,$EA,$F9,$19,$1F,$81
	.dc.b	$7F,$2B,$1E,$EF,$95,$FA,$24,$9E,$DA,$A3,$EA,$FC,$87,$E7,$2E,$57
	.dc.b	$E2,$E3,$F3,$FF,$27,$FA,$0A,$93,$F9,$98,$FC,$17,$CF,$3E,$4B,$F0
	.dc.b	$2D,$9F,$63,$5F,$C0,$E8,$8C,$47,$F8,$1B,$B7,$F5,$78,$FE,$BB,$E1
	.dc.b	$E4,$FE,$9E,$3F,$6F,$F4,$5F,$C0,$B9,$DC,$FF,$CC,$E3,$FD,$1B,$5F
	.dc.b	$EB,$86,$BF,$ED,$97,$84,$7F,$BE,$3E,$BB,$FD,$91,$3F,$80,$A4,$62
	.dc.b	$3F,$2F,$18,$F4,$3F,$D5,$CD,$8F,$F5,$85,$FE,$2D,$BD,$1F,$EB,$35
	.dc.b	$19,$7F,$1A,$D7,$E0,$3F,$16,$7D,$71,$FC,$8A,$3F,$B6,$BE,$BF,$5B
	.dc.b	$E0,$EF,$F4,$3E,$47,$E7,$92,$7E,$7B,$FC,$46,$9E,$BF,$C8,$24,$FC
	.dc.b	$5F,$85,$F9,$7F,$29,$F3,$3E,$52,$4F,$C9,$8F,$BB,$F2,$1F,$70,$91
	.dc.b	$F7,$1E,$D8,$FD,$42,$47,$D1,$F9,$4F,$AF,$F2,$72,$3E,$A3,$F6,$BE
	.dc.b	$53,$F1,$3F,$13,$F8,$82,$4F,$E2,$1F,$57,$CA,$6A,$3F,$DE,$DD,$3F
	.dc.b	$E5,$57,$EF,$47,$E1,$32,$91,$F8,$17,$F0,$FF,$27,$23,$04,$3F,$08
	.dc.b	$FE,$33,$E4,$FF,$1B,$84,$9F,$C5,$47,$E4,$3E,$4F,$51,$83,$F1,$CF
	.dc.b	$E5,$3E,$4E,$47,$E4,$5F,$CB,$FC,$94,$8F,$CA,$35,$FC,$A7,$C9,$7E
	.dc.b	$70,$8F,$E6,$5F,$EA,$BE,$8F,$E4,$BB,$7F,$94,$7C,$D7,$F8,$3E,$BF
	.dc.b	$D0,$3D,$7F,$88,$F0,$7F,$BB,$36,$DE,$DF,$F3,$7B,$E1,$6A,$D7,$C9
	.dc.b	$67,$E0,$E4,$F4,$CB,$FC,$3D,$8F,$FA,$16,$30,$7F,$76,$7F,$CF,$2E
	.dc.b	$32,$47,$6C,$3C,$57,$EA,$F8,$3F,$5F,$C1,$CC,$38,$5F,$FA,$D1,$BF
	.dc.b	$1F,$F0,$3E,$42,$46,$9F,$EF,$2F,$FA,$B7,$DA,$BE,$03,$DB,$FD,$83
	.dc.b	$67,$F9,$D9,$FE,$B3,$8C,$9A,$C3,$8B,$23,$DA,$7F,$AE,$63,$E8,$FC
	.dc.b	$87,$D2,$F8,$F9,$3E,$88,$FA,$BF,$1D,$A8,$C1,$EC,$7E,$87,$C6,$7D
	.dc.b	$A2,$47,$CA,$F3,$DC,$BF,$E7,$08,$F8,$7F,$19,$C1,$D4,$65,$FB,$1C
	.dc.b	$BF,$52,$7F,$0B,$E6,$C9,$F9,$6F,$C6,$F4,$DD,$DF,$23,$EA,$3F,$DB
	.dc.b	$35,$F1,$1A,$FF,$60,$6B,$FD,$D9,$AF,$FA,$EB,$EE,$5E,$BF,$3C,$64
	.dc.b	$FC,$89,$CB,$E8,$F8,$EF,$CC,$91,$C1,$F8,$0F,$8E,$FC,$A4,$48,$C1
	.dc.b	$F0,$3E,$37,$F4,$3F,$1B,$23,$07,$E2,$23,$F3,$84,$70,$7C,$8F,$8A
	.dc.b	$FF,$09,$F0,$1A,$48,$FE,$FF,$E0,$7E,$63,$E3,$7F,$4D,$32,$7F,$2B
	.dc.b	$1F,$B2,$F8,$AF,$D9,$CC,$9F,$CC,$C6,$5A,$43,$EB,$9F,$A7,$F8,$AF
	.dc.b	$F1,$47,$B1,$F9,$58,$FE,$C6,$D7,$FD,$CF,$CD,$FE,$AC,$EC,$FF,$15
	.dc.b	$CE,$67,$FD,$21,$FA,$17,$F4,$67,$AD,$6D,$FC,$64,$F5,$F1,$7F,$82
	.dc.b	$56,$DB,$E0,$7F,$95,$63,$FC,$A3,$5F,$E3,$4D,$51,$88,$FE,$E8,$D7
	.dc.b	$C3,$8F,$F9,$A2,$38,$DF,$49,$FE,$05,$1E,$1E,$AF,$3F,$7E,$FE,$2E
	.dc.b	$56,$A3,$97,$C5,$73,$64,$72,$8E,$DF,$17,$E7,$24,$7D,$8C,$11,$EB
	.dc.b	$F8,$BD,$CF,$8A,$3C,$D1,$FC,$23,$E3,$5E,$CF,$F2,$1A,$7C,$BB,$C9
	.dc.b	$FF,$4D,$FB,$89,$93,$97,$16,$9D,$47,$2A,$FF,$51,$DF,$5B,$E1,$2B
	.dc.b	$F0,$EF,$D7,$E4,$F3,$1C,$E7,$18,$78,$9A,$78,$30,$F4,$57,$07,$B4
	.dc.b	$EE,$7C,$47,$9B,$1F,$2E,$3B,$C7,$D7,$5E,$66,$D6,$CA,$79,$EB,$8C
	.dc.b	$EE,$3A,$9E,$BB,$F1,$C7,$87,$0F,$5D,$7C,$47,$C1,$46,$E1,$A8,$EF
	.dc.b	$8F,$C1,$7A,$BF,$19,$B8,$7B,$9F,$31,$AB,$F2,$BF,$17,$1E,$73,$B8
	.dc.b	$F5,$1D,$A7,$6F,$A6,$76,$7C,$E7,$67,$9F,$1F,$61,$F0,$71,$F4,$BF
	.dc.b	$3F,$1F,$54,$F9,$AB,$9B,$C4,$F3,$D6,$8F,$86,$4A,$EB,$1A,$7B,$1F
	.dc.b	$3F,$8B,$FC,$B3,$97,$C3,$6E,$5F,$99,$6E,$1F,$97,$7F,$CB,$F7,$F9
	.dc.b	$9F,$98,$C6,$3F,$E9,$C6,$C7,$FA,$C9,$AF,$E6,$9A,$F1,$3D,$47,$E8
	.dc.b	$FD,$1F,$F2,$AC,$FE,$72,$34,$D8,$CE,$5D,$E7,$89,$FE,$12,$E3,$F9
	.dc.b	$CF,$08,$FF,$61,$CF,$08,$F2,$1F,$8E,$38,$14,$5F,$E5,$9C,$E9,$E3
	.dc.b	$72,$78,$27,$A3,$ED,$31,$7E,$27,$73,$E9,$8E,$91,$8B,$F2,$C6,$2F
	.dc.b	$C1,$7F,$E7,$6B,$F0,$46,$27,$57,$E9,$AD,$5F,$C3,$BE,$D3,$EB,$BF
	.dc.b	$87,$5E,$68,$F8,$87,$F0,$C7,$0F,$8C,$C5,$F6,$AE,$67,$8A,$3C,$11
	.dc.b	$9B,$F4,$C7,$73,$E0,$9F,$50,$F8,$F8,$79,$0E,$55,$D2,$7D,$07,$27
	.dc.b	$F9,$16,$3F,$D1,$BE,$3F,$FD,$2F,$9F,$F4,$EF,$8F,$B9,$7F,$56,$9B
	.dc.b	$92,$FF,$A6,$D6,$1B,$5F,$CE,$E3,$FA,$E4,$FF,$5E,$6B,$C7,$89,$18
	.dc.b	$AF,$EA,$FF,$09,$CE,$5F,$1B,$63,$0C,$C3,$FA,$AF,$C7,$5C,$7F,$BA
	.dc.b	$E1,$FD,$82,$39,$7C,$77,$3F,$8E,$B7,$FD,$71,$C3,$FD,$45,$F1,$64
	.dc.b	$69,$8F,$2B,$E0,$F7,$BB,$DD,$BE,$79,$CC,$63,$0F,$EF,$33,$F0,$2B
	.dc.b	$83,$67,$5A,$7F,$82,$23,$C3,$E6,$FC,$13,$E1,$7C,$65,$C3,$FB,$9C
	.dc.b	$7C,$5F,$8D,$C5,$FC,$0D,$11,$F2,$7E,$33,$C7,$C9,$70,$FF,$05,$B8
	.dc.b	$7F,$AF,$DF,$F6,$5D,$3D,$93,$E5,$24,$FF,$75,$F8,$DF,$6F,$27,$D5
	.dc.b	$6C,$7F,$85,$A3,$F0,$5F,$A1,$FE,$16,$30,$7F,$A2,$DC,$61,$F5,$9F
	.dc.b	$E9,$51,$FE,$1E,$E2,$E3,$D5,$F2,$99,$1C,$7E,$36,$DF,$17,$F2,$6B
	.dc.b	$F1,$FE,$6F,$F8,$D8,$7D,$63,$8B,$73,$AE,$BF,$F1,$84,$E7,$0F,$F4
	.dc.b	$3B,$6F,$E2,$9C,$AF,$F3,$2D,$9F,$F4,$9C,$EE,$3F,$C1,$3D,$2D,$3F
	.dc.b	$0D,$5F,$07,$9B,$C3,$9F,$C6,$7E,$3F,$C7,$7E,$23,$9A,$D7,$F9,$01
	.dc.b	$C3,$F2,$7D,$CB,$97,$C9,$7E,$7F,$C7,$7D,$0F,$16,$D7,$CF,$70,$FD
	.dc.b	$1F,$12,$DF,$D1,$7E,$4F,$89,$F9,$2E,$1F,$F2,$92,$F2,$33,$DB,$A6
	.dc.b	$61,$FE,$B2,$9D,$36,$3E,$2C,$63,$0F,$F2,$97,$03,$E3,$BF,$0A,$78
	.dc.b	$3C,$9B,$87,$E1,$7E,$3B,$F1,$27,$E2,$BE,$3A,$C7,$E2,$7E,$37,$F1
	.dc.b	$BF,$1B,$F8,$E6,$C7,$E3,$7E,$37,$F6,$BF,$26,$E8,$FC,$9C,$61,$93
	.dc.b	$F9,$3F,$7E,$ED,$F9,$37,$F3,$3F,$1D,$27,$F3,$1F,$1D,$6B,$F3,$1F
	.dc.b	$1D,$F9,$E3,$1C,$BF,$EF,$7E,$24,$7E,$8B,$E3,$A6,$1F,$A1,$8B,$7F
	.dc.b	$A1,$F8,$EF,$E5,$1F,$A2,$7F,$95,$6B,$F4,$D1,$89,$91,$FE,$F3,$9F
	.dc.b	$ED,$38,$A3,$FC,$0D,$1C,$67,$F9,$0B,$A8,$CF,$C7,$F6,$F8,$EE,$ED
	.dc.b	$8E,$91,$6B,$8C,$DB,$ED,$1F,$CF,$ED,$F1,$FF,$3B,$97,$D2,$DC,$7F
	.dc.b	$DC,$5C,$69,$FE,$40,$D8,$FD,$5F,$98,$8C,$61,$FC,$4A,$31,$5F,$D0
	.dc.b	$F0,$7C,$7F,$C6,$3E,$3F,$AF,$71,$F8,$9F,$1F,$F2,$8F,$97,$F1,$D6
	.dc.b	$3C,$7C,$BE,$22,$3E,$2D,$11,$DF,$E3,$FE,$9F,$C7,$FD,$46,$D7,$D1
	.dc.b	$F8,$FF,$AE,$7D,$8F,$8D,$B1,$FF,$00,$7E,$53,$E9,$70,$FF,$B0,$1A
	.dc.b	$FF,$2D,$6A,$E1,$46,$3C,$D7,$0B,$E4,$3F,$0B,$F1,$BF,$86,$6E,$3F
	.dc.b	$E9,$BE,$6F,$F6,$47,$CE,$3E,$27,$F8,$AB,$67,$F8,$CE,$88,$C3,$FC
	.dc.b	$71,$AF,$FB,$5B,$EC,$5E,$27,$F5,$76,$D7,$D8,$6C,$FD,$76,$BF,$C8
	.dc.b	$33,$7C,$87,$E7,$3E,$43,$F3,$AD,$BF,$CD,$7C,$87,$E8,$3E,$3F,$F4
	.dc.b	$2D,$8F,$95,$D7,$FD,$9F,$8B,$FD,$75,$1F,$88,$6E,$5F,$87,$6B,$E1
	.dc.b	$DB,$F5,$7F,$1F,$FA,$C6,$D7,$D5,$E7,$FD,$7F,$C7,$FF,$7A,$C1,$18
	.dc.b	$9F,$B5,$D1,$E1,$23,$E1,$7C,$9D,$1C,$4F,$F1,$07,$81,$FB,$39,$FC
	.dc.b	$14,$75,$E7,$CF,$C7,$D9,$FB,$07,$F9,$F6,$91,$E3,$E1,$F8,$04,$9F
	.dc.b	$80,$FC,$3E,$2F,$41,$E0,$F8,$FB,$3D,$38,$3F,$C6,$98,$3E,$3F,$CB
	.dc.b	$EA,$79,$9B,$5F,$97,$E4,$F9,$C4,$7E,$73,$FE,$33,$AF,$E7,$5D,$5F
	.dc.b	$C0,$BB,$3F,$80,$FA,$9F,$03,$D2,$F8,$2F,$C2,$F8,$FF,$99,$D1,$72
	.dc.b	$FD,$23,$87,$ED,$3F,$74,$8F,$F6,$EB,$FE,$1B,$F0,$9F,$30,$F9,$BF
	.dc.b	$19,$67,$DD,$3A,$6E,$6F,$84,$D7,$07,$F9,$0A,$31,$3F,$17,$C3,$7C
	.dc.b	$A3,$8F,$EC,$FC,$6F,$F8,$F2,$31,$5D,$23,$2F,$6F,$8D,$B8,$F5,$7F
	.dc.b	$03,$F0,$DF,$82,$F8,$DB,$5E,$47,$0F,$9B,$E3,$6D,$F9,$5F,$C4,$9F
	.dc.b	$3C,$91,$F0,$F4,$47,$D3,$F8,$3F,$C7,$FC,$3F,$E4,$1B,$5F,$0A,$3F
	.dc.b	$27,$F1,$98,$C3,$E1,$3F,$E2,$6F,$96,$FC,$27,$C8,$7F,$04,$6C,$FF
	.dc.b	$87,$9A,$FF,$91,$1A,$89,$F2,$3F,$DE,$FE,$5F,$F8,$6B,$63,$FC,$4E
	.dc.b	$D7,$FC,$9A,$D7,$90,$9F,$F3,$4F,$2F,$8F,$BF,$FD,$19,$3E,$36,$BE
	.dc.b	$36,$BF,$A9,$F9,$0E,$7D,$DF,$95,$E0,$7C,$86,$37,$FD,$73,$6F,$F4
	.dc.b	$B7,$71,$FE,$B6,$DB,$34,$D7,$E8,$BE,$2C,$7F,$44,$FF,$BD,$2F,$FD
	.dc.b	$EC,$D7,$FD,$48,$DB,$34,$FF,$0A,$47,$E6,$97,$C8,$F6,$3B,$FC,$24
	.dc.b	$9D,$35,$FF,$4E,$CF,$FA,$45,$E5,$F1,$DE,$6F,$8E,$91,$82,$1C,$23
	.dc.b	$D6,$7B,$3E,$3B,$E2,$77,$7E,$93,$C5,$F7,$FC,$74,$BF,$D4,$77,$A7
	.dc.b	$6D,$8C,$C6,$0F,$8F,$F1,$DA,$F0,$FF,$A8,$E1,$FB,$06,$DF,$FB,$02
	.dc.b	$FF,$C0,$6D,$7F,$D2,$7B,$FC,$8F,$8D,$E5,$1E,$2A,$FA,$71,$F5,$3E
	.dc.b	$37,$EA,$FC,$1C,$AF,$99,$1E,$6F,$13,$B9,$23,$BB,$FD,$93,$0F,$AB
	.dc.b	$ED,$FD,$E2,$4F,$D5,$F8,$DF,$C0,$FC,$6F,$E0,$BE,$36,$4F,$D9,$8F
	.dc.b	$87,$E2,$FB,$09,$1E,$C7,$B7,$37,$F8,$3A,$3F,$01,$F0,$5F,$8C,$3F
	.dc.b	$1B,$F1,$BF,$33,$C6,$FC,$7F,$C5,$CA,$FA,$31,$F4,$3B,$BE,$49,$23
	.dc.b	$E4,$BF,$33,$C0,$FA,$A4,$8F,$C7,$7C,$67,$E6,$BE,$33,$F4,$1F,$A3
	.dc.b	$7F,$B0,$3C,$2F,$E7,$8D,$F1,$3F,$37,$3F,$84,$F5,$BF,$1D,$87,$F8
	.dc.b	$D5,$AF,$EE,$BB,$3F,$A7,$7C,$87,$DE,$9C,$CF,$F0,$7E,$76,$2F,$2F
	.dc.b	$C3,$F1,$B2,$BF,$9D,$CE,$5B,$F1,$DE,$62,$3A,$F2,$7C,$9E,$17,$72
	.dc.b	$47,$F9,$33,$24,$7D,$DB,$FF,$24,$CF,$6F,$8E,$F8,$27,$C2,$F8,$5F
	.dc.b	$E0,$91,$FD,$13,$D6,$FF,$13,$B4,$FF,$14,$F8,$BF,$D3,$5B,$1E,$FF
	.dc.b	$0E,$47,$85,$AF,$C7,$F0,$7E,$51,$1F,$7F,$C7,$7C,$DF,$03,$E9,$FC
	.dc.b	$74,$9F,$EA,$DB,$27,$E9,$77,$FC,$D2,$4F,$BD,$FA,$1D,$FF,$5C,$91
	.dc.b	$F3,$3E,$13,$DB,$FA,$37,$F1,$08,$76,$7E,$03,$E3,$BF,$06,$6B,$D0
	.dc.b	$DF,$5F,$C2,$BF,$B7,$7F,$D0,$5A,$9A,$3C,$FE,$31,$FE,$13,$35,$3E
	.dc.b	$28,$FC,$8F,$C3,$49,$FC,$0B,$F9,$4F,$86,$91,$82,$1E,$8F,$07,$F1
	.dc.b	$9D,$29,$FC,$DF,$C7,$7E,$73,$E1,$E5,$FE,$32,$32,$D2,$1C,$C7,$83
	.dc.b	$F9,$6F,$86,$FD,$2F,$C3,$49,$FC,$AB,$FA,$8F,$86,$91,$F9,$86,$BC
	.dc.b	$EB,$F3,$FF,$17,$FA,$E2,$47,$E7,$9F,$D8,$FC,$5F,$F3,$0F,$94,$FF
	.dc.b	$55,$CF,$E8,$BE,$1F,$F6,$A6,$7C,$5F,$DB,$FC,$65,$9F,$F0,$D5,$70
	.dc.b	$7C,$1C,$09,$FE,$B9,$FA,$3F,$FA,$E2,$7C,$8D,$E9,$FF,$1D,$37,$8F
	.dc.b	$F6,$CF,$F9,$4A,$7F,$C5,$AF,$F8,$95,$BF,$19,$D0,$EB,$F1,$3F,$A8
	.dc.b	$FC,$41,$CB,$E0,$F0,$65,$FE,$06,$72,$DE,$5F,$D6,$C7,$FA,$1A,$3F
	.dc.b	$D2,$AD,$7F,$C0,$A7,$F8,$3D,$C3,$85,$CB,$E1,$99,$3D,$E3,$E0,$FC
	.dc.b	$17,$C2,$F8,$D9,$3E,$C8,$CB,$48,$7C,$6F,$7F,$EA,$75,$FC,$AF,$8D
	.dc.b	$93,$F4,$23,$E6,$FC,$6E,$A3,$07,$C8,$7E,$8F,$C0,$FD,$2F,$32,$4F
	.dc.b	$CC,$8F,$E1,$FF,$34,$BF,$E1,$ED,$9F,$F5,$BB,$5F,$86,$1C,$3F,$E2
	.dc.b	$AE,$D7,$23,$E0,$3A,$DC,$AF,$B7,$1F,$8F,$DF,$F0,$BF,$1D,$27,$F1
	.dc.b	$71,$F8,$8F,$8E,$D4,$60,$FC,$1B,$F9,$7E,$FF,$C6,$FC,$74,$9F,$C9
	.dc.b	$C7,$E4,$3E,$07,$51,$83,$F1,$6F,$F8,$3B,$E5,$3F,$29,$1F,$D2,$5A
	.dc.b	$FF,$93,$7D,$B4,$62,$3F,$DC,$CD,$DB,$DF,$F1,$FC,$FB,$A5,$7E,$62
	.dc.b	$3B,$76,$FE,$8B,$E0,$64,$FF,$A0,$70,$FC,$E3,$FC,$07,$1E,$F5,$C2
	.dc.b	$34,$FE,$AB,$E3,$FD,$1D,$7F,$E1,$AB,$34,$95,$FE,$99,$E4,$FF,$24
	.dc.b	$BD,$E3,$8B,$FA,$CF,$82,$FD,$AF,$C1,$7F,$84,$6B,$FC,$AF,$3C,$63
	.dc.b	$07,$F8,$57,$CD,$FF,$28,$6D,$FB,$68,$E3,$3F,$EA,$29,$CB,$FC,$69
	.dc.b	$FF,$25,$B9,$F4,$3F,$CB,$D1,$FD,$07,$C5,$FF,$5D,$3A,$9F,$E3,$0D
	.dc.b	$8C,$FC,$67,$03,$8F,$A1,$CA,$7C,$84,$AF,$F5,$C3,$FD,$05,$BF,$19
	.dc.b	$B7,$CD,$1E,$28,$FE,$DF,$F1,$DF,$DD,$B0,$FF,$46,$7C,$C3,$DF,$FF
	.dc.b	$42,$CB,$E3,$7F,$D0,$F3,$9C,$7F,$D2,$31,$8A,$DF,$C1,$7F,$5C,$95
	.dc.b	$6D,$5F,$77,$FC,$DF,$5D,$AF,$FD,$27,$A6,$EB,$C1,$5F,$E9,$F8,$C6
	.dc.b	$1D,$5B,$8F,$A3,$0E,$55,$4F,$29,$4F,$0D,$79,$71,$E3,$1E,$69,$E2
	.dc.b	$D3,$89,$4E,$35,$CF,$0F,$56,$1E,$7C,$7D,$37,$CA,$3A,$9D,$B0,$D6
	.dc.b	$1C,$6B,$E4,$3F,$EB,$AC,$7C,$97,$F2,$B1,$D4,$C3,$C5,$8F,$C9,$C3
	.dc.b	$A5,$7A,$1E,$B3,$AB,$E6,$F5,$BF,$95,$5E,$52,$46,$4F,$A6,$5A,$E6
	.dc.b	$48,$F2,$47,$59,$E6,$FC,$6C,$7C,$27,$C0,$9D,$B8,$7B,$C7,$CA,$3B
	.dc.b	$97,$A7,$F4,$58,$FE,$15,$B8,$FE,$2D,$B8,$7F,$96,$F0,$FE,$67,$F2
	.dc.b	$D7,$0F,$EA,$ED,$7F,$DB,$AD,$7F,$DD,$0D,$4F,$A9,$FF,$76,$7C,$33
	.dc.b	$8B,$11,$C0,$F3,$9F,$E6,$97,$2F,$23,$31,$C5,$D9,$9A,$FD,$13,$C4
	.dc.b	$CD,$7E,$55,$C6,$9A,$7A,$3C,$2B,$A3,$99,$F2,$C7,$7A,$C5,$78,$AF
	.dc.b	$CB,$5D,$6B,$8E,$9F,$22,$BF,$38,$78,$1C,$61,$DE,$FC,$93,$E2,$9F
	.dc.b	$F0,$1D,$72,$8F,$9E,$72,$70,$FD,$03,$13,$CA,$FC,$EF,$92,$3C,$C7
	.dc.b	$C2,$36,$79,$0F,$00,$3E,$AB,$CD,$DD,$FC,$03,$D0,$F1,$9E,$EF,$5C
	.dc.b	$32,$6F,$0F,$5D,$F3,$BE,$27,$F1,$DC,$BF,$CD,$5F,$1F,$F6,$F3,$FE
	.dc.b	$6F,$B5,$BF,$F5,$67,$02,$B1,$9F,$F4,$1A,$B8,$7F,$9C,$DF,$ED,$AD
	.dc.b	$BF,$E6,$F1,$8B,$FE,$5D,$1F,$CE,$B6,$E3,$1F,$EA,$A8,$FE,$9D,$F1
	.dc.b	$F8,$99,$1F,$D4,$7E,$3E,$E3,$FD,$29,$E3,$F0,$BC,$BE,$16,$D7,$F5
	.dc.b	$77,$0F,$F4,$87,$C4,$47,$FA,$5F,$C7,$FB,$3B,$FF,$C0,$7E,$77,$92
	.dc.b	$31,$DC,$8A,$7F,$82,$A7,$75,$FE,$28,$6E,$5F,$E4,$C6,$BF,$E1,$1F
	.dc.b	$33,$99,$F0,$3E,$12,$D7,$F8,$1E,$3E,$1F,$C7,$E2,$BF,$B3,$BF,$1B
	.dc.b	$E1,$7C,$7E,$65,$BF,$F0,$73,$87,$C5,$9A,$3C,$CD,$8C,$C7,$D3,$F8
	.dc.b	$5F,$6F,$97,$F4,$9B,$3F,$E1,$98,$FA,$BF,$1D,$8C,$BF,$BC,$BF,$63
	.dc.b	$E3,$BE,$2F,$95,$6F,$FC,$40,$E1,$FF,$04,$7C,$75,$BF,$F0,$0B,$F8
	.dc.b	$8F,$D2,$3F,$41,$3F,$EB,$9F,$98,$79,$99,$BF,$C7,$88,$CC,$EE,$7F
	.dc.b	$C7,$11,$89,$FF,$2C,$B5,$FF,$3E,$B5,$FD,$77,$6A,$FF,$8B,$77,$FF
	.dc.b	$1B,$8C,$7C,$87,$97,$C5,$F3,$3E,$7F,$90,$FC,$37,$CA,$BC,$6C,$6B
	.dc.b	$FB,$9C,$7F,$92,$75,$2B,$53,$93,$FC,$A0,$E3,$4F,$F5,$D6,$8F,$85
	.dc.b	$EB,$8F,$C5,$7E,$37,$C8,$CC,$39,$5F,$FB,$2A,$FF,$C3,$51,$8D,$BD
	.dc.b	$95,$BA,$EB,$F2,$17,$4F,$76,$08,$F9,$BE,$53,$D7,$BD,$18,$6E,$7F
	.dc.b	$EE,$D9,$4F,$F9,$9D,$B1,$FE,$88,$C7,$FA,$7C,$E4,$D7,$AD,$F7,$FE
	.dc.b	$43,$BB,$6B,$D1,$93,$E4,$3F,$09,$F2,$1F,$85,$6C,$EB,$DD,$FC,$41
	.dc.b	$8C,$BE,$4B,$F8,$BF,$8F,$FA,$BE,$5D,$AF,$8F,$CD,$F4,$E3,$31,$E0
	.dc.b	$F8,$FB,$87,$D2,$7F,$29,$F1,$12,$7C,$BF,$0D,$6B,$EB,$FC,$07,$E6
	.dc.b	$4C,$72,$FC,$C7,$F3,$9F,$1F,$2F,$DB,$16,$BE,$FF,$C7,$FE,$84,$FD
	.dc.b	$17,$C1,$5A,$F8,$71,$FA,$5F,$8F,$C5,$7E,$19,$FD,$47,$C6,$FF,$30
	.dc.b	$FD,$27,$FB,$D5,$FC,$A8,$C3,$0F,$C2,$46,$AB,$F1,$F3,$95,$9F,$FB
	.dc.b	$E7,$C3,$FC,$95,$FE,$3A,$50,$ED,$FD,$9B,$A8,$E3,$3F,$D6,$5D,$CE
	.dc.b	$8F,$F5,$64,$F1,$9F,$C3,$36,$7F,$D9,$ED,$7F,$B8,$FB,$7E,$DF,$8E
	.dc.b	$EE,$DA,$FB,$7E,$1B,$CC,$FD,$99,$F0,$FE,$16,$CE,$BC,$CF,$E6,$D4
	.dc.b	$69,$23,$F9,$C4,$3C,$BF,$99,$F1,$DF,$BE,$6C,$FE,$2B,$C7,$FE,$0E
	.dc.b	$FF,$B2,$5B,$A3,$C7,$71,$3E,$37,$FA,$65,$6B,$B7,$F7,$DE,$57,$F4
	.dc.b	$98,$34,$98,$FE,$AF,$E6,$1D,$3F,$ED,$56,$C6,$63,$07,$F7,$0C,$7F
	.dc.b	$D8,$F1,$F8,$0F,$88,$FC,$37,$C7,$7E,$1D,$F1,$F2,$78,$49,$3F,$D6
	.dc.b	$72,$FE,$17,$B7,$DB,$3D,$98,$7F,$61,$9F,$C5,$FB,$3F,$91,$F8,$EF
	.dc.b	$C9,$3F,$93,$F6,$7F,$29,$F1,$D6,$BF,$22,$FC,$2F,$1F,$E0,$12,$3C
	.dc.b	$0F,$A7,$E3,$7F,$39,$EC,$7E,$75,$B3,$F4,$7C,$6F,$96,$7E,$87,$D1
	.dc.b	$B1,$F5,$7C,$5F,$98,$47,$D8,$FC,$FF,$1B,$F5,$1E,$87,$D4,$6C,$FE
	.dc.b	$9B,$CF,$FD,$61,$FA,$DF,$8D,$B1,$F8,$1F,$23,$EC,$11,$F9,$0F,$D1
	.dc.b	$F6,$3F,$69,$EC,$7E,$D5,$B3,$FB,$2F,$8D,$FD,$C1,$FB,$9F,$8C,$B1
	.dc.b	$FC,$5B,$E5,$3F,$6D,$D1,$FA,$BF,$01,$18,$8F,$F2,$34,$62,$7F,$C2
	.dc.b	$53,$99,$FC,$97,$E9,$5F,$93,$9F,$E8,$8F,$F4,$FF,$37,$FA,$F3,$99
	.dc.b	$FD,$C9,$C3,$DA,$FF,$1D,$C3,$E3,$BF,$C4,$98,$3C,$8C,$C6,$19,$8F
	.dc.b	$F8,$86,$74,$D8,$F2,$9E,$6F,$07,$99,$DB,$E3,$AC,$FC,$18,$F0,$FC
	.dc.b	$4E,$33,$F2,$F9,$2C,$D3,$E6,$F8,$E9,$87,$BB,$E3,$2E,$1C,$2D,$5E
	.dc.b	$CF,$8C,$CF,$07,$C2,$7D,$FE,$CC,$C7,$E3,$FC,$65,$C7,$94,$FC,$4F
	.dc.b	$8C,$C6,$1F,$29,$C5,$FE,$9E,$BE,$8C,$7F,$98,$B2,$34,$C7,$C6,$7C
	.dc.b	$E3,$E7,$FC,$65,$BF,$AF,$1F,$4B,$E3,$31,$87,$AD,$FF,$11,$FC,$B7
	.dc.b	$D3,$F6,$BF,$B8,$36,$7C,$4D,$7F,$74,$D7,$F9,$B7,$C7,$B6,$7C,$BF
	.dc.b	$F7,$98,$C3,$63,$33,$ED,$9C,$BD,$BD,$6F,$E4,$0D,$E5,$F1,$C6,$23
	.dc.b	$C2,$D7,$FC,$63,$BB,$E3,$FF,$A4,$70,$FE,$53,$E3,$EF,$06,$9B,$7F
	.dc.b	$9C,$AB,$8F,$E0,$DB,$87,$39,$FC,$FC,$63,$E3,$F3,$EF,$7F,$85,$6F
	.dc.b	$F1,$4D,$BF,$E9,$57,$F6,$67,$2D,$87,$C7,$72,$8F,$D5,$D7,$82,$24
	.dc.b	$7E,$B9,$AF,$E4,$7D,$EF,$2F,$BB,$E8,$DB,$F9,$66,$DF,$E4,$F6,$ED
	.dc.b	$F1,$27,$77,$FB,$86,$BF,$A2,$9F,$EE,$6F,$F0,$7F,$47,$A7,$C6,$C8
	.dc.b	$C1,$0F,$DE,$F8,$BF,$C2,$A7,$E5,$B6,$3F,$C9,$15,$FD,$49,$FF,$00
	.dc.b	$78,$2F,$53,$FC,$79,$08,$C3,$FE,$3F,$6B,$FA,$16,$BF,$E0,$3E,$67
	.dc.b	$E0,$1E,$88,$FE,$1E,$73,$AF,$BA,$D3,$FC,$BB,$3C,$3A,$7E,$47,$B5
	.dc.b	$27,$CE,$FE,$12,$5F,$B1,$3D,$5E,$79,$3D,$6F,$4B,$70,$F4,$46,$0F
	.dc.b	$F2,$C5,$F9,$B0,$E9,$9F,$F9,$CE,$BC,$6D,$7E,$97,$C2,$78,$7B,$7F
	.dc.b	$DC,$91,$E8,$6B,$FC,$B2,$FF,$5D,$E3,$7E,$07,$DD,$F7,$FC,$E7,$E8
	.dc.b	$EB,$FC,$50,$D4,$1C,$37,$B3,$F3,$5C,$5F,$E7,$98,$FB,$3F,$21,$F8
	.dc.b	$D3,$F1,$DF,$1F,$F7,$BC,$E7,$AF,$49,$5F,$80,$F8,$DF,$1F,$A5,$F7
	.dc.b	$B3,$FF,$4E,$4F,$F9,$ED,$AF,$F1,$D6,$BF,$77,$ED,$9E,$B7,$D3,$6C
	.dc.b	$7D,$A6,$BF,$99,$F8,$FF,$D1,$7B,$FE,$5A,$93,$F9,$28,$FC,$1F,$CE
	.dc.b	$7E,$42,$0D,$23,$F0,$BE,$43,$DD,$ED,$48,$FC,$DC,$7E,$BF,$E4,$3F
	.dc.b	$61,$DB,$27,$F3,$F1,$DB,$BF,$F3,$B8,$7E,$56,$7F,$C4,$0D,$7F,$CB
	.dc.b	$8D,$7F,$C7,$DC,$8F,$90,$FD,$E1,$23,$F5,$7F,$05,$FA,$EF,$99,$FE
	.dc.b	$AF,$1F,$AE,$F8,$2F,$E0,$46,$7D,$BF,$D0,$FC,$E7,$EA,$27,$FC,$D4
	.dc.b	$D7,$F0,$6D,$4E,$7C,$75,$FE,$43,$CC,$FF,$68,$B7,$CC,$FF,$03,$B6
	.dc.b	$FD,$3F,$25,$E5,$F6,$FF,$8A,$7C,$64,$9F,$F5,$3D,$27,$F6,$BF,$01
	.dc.b	$FC,$9A,$7F,$AF,$35,$FF,$23,$35,$FF,$41,$35,$65,$E6,$F7,$7E,$21
	.dc.b	$23,$F8,$8F,$C2,$7F,$25,$F9,$29,$1F,$C5,$1F,$E0,$BF,$8B,$7E,$2B
	.dc.b	$66,$DF,$E8,$F8,$7C,$37,$F1,$66,$DF,$FB,$4D,$AF,$F7,$56,$BF,$E5
	.dc.b	$D5,$D7,$DA,$64,$FF,$A8,$67,$2F,$FB,$B7,$ED,$BC,$2D,$F9,$4F,$EA
	.dc.b	$70,$6E,$12,$78,$35,$FF,$0D,$CB,$E5,$38,$EF,$27,$B3,$D3,$A6,$46
	.dc.b	$08,$72,$7C,$3F,$0F,$E2,$F9,$19,$3E,$78,$F2,$FC,$7E,$A3,$07,$AD
	.dc.b	$F4,$FC,$86,$A3,$07,$BD,$FC,$D7,$BD,$EE,$F9,$09,$3F,$86,$8F,$83
	.dc.b	$ED,$EA,$30,$7C,$77,$E2,$FB,$3F,$1B,$DC,$93,$F2,$E3,$E4,$FC,$84
	.dc.b	$8F,$9C,$FC,$CF,$90,$F9,$BE,$8C,$9F,$A4,$FF,$48,$F9,$C7,$C0,$5F
	.dc.b	$94,$FE,$AF,$1F,$D6,$EB,$FC,$68,$D7,$FD,$38,$D7,$FD,$B0,$FF,$3D
	.dc.b	$EB,$FB,$B0,$8E,$51,$86,$F3,$F0,$F7,$FF,$03,$F2,$52,$FF,$0C,$FE
	.dc.b	$13,$D1,$91,$82,$1F,$82,$7F,$B0,$FE,$4A,$F6,$7F,$17,$DF,$2B,$FD
	.dc.b	$E5,$39,$7F,$25,$E0,$70,$A9,$18,$3F,$2D,$F2,$7F,$94,$23,$FD,$33
	.dc.b	$5F,$CC,$FB,$8F,$C0,$FF,$C6,$D1,$ED,$F3,$BE,$49,$2B,$FC,$B8,$7C
	.dc.b	$37,$FC,$55,$3F,$D1,$27,$84,$F5,$DF,$E5,$D7,$F8,$DA,$B5,$1F,$E9
	.dc.b	$8F,$9D,$F5,$57,$F9,$F2,$7F,$D1,$CF,$13,$FD,$D4,$E1,$FF,$0C,$C7
	.dc.b	$0B,$F4,$CF,$F8,$4E,$78,$79,$2A,$D5,$F7,$CF,$C5,$F0,$BF,$E2,$09
	.dc.b	$7B,$1E,$5F,$46,$63,$CA,$72,$F8,$F5,$F8,$DF,$29,$C6,$E4,$E7,$4F
	.dc.b	$F4,$6B,$C7,$D8,$F8,$FA,$F8,$DB,$8F,$F8,$56,$7F,$D0,$D3,$97,$FC
	.dc.b	$A1,$5F,$CD,$63,$11,$F0,$2B,$E5,$7C,$17,$A7,$D5,$F9,$9F,$29,$2B
	.dc.b	$D1,$19,$69,$0F,$9D,$E9,$7C,$1F,$53,$C6,$48,$F1,$E6,$8F,$BB,$F0
	.dc.b	$BB,$9F,$89,$1F,$6E,$3E,$E7,$A9,$F8,$22,$4F,$A9,$FC,$2F,$C9,$EA
	.dc.b	$3E,$34,$7E,$0E,$3F,$09,$F2,$9F,$85,$F9,$49,$5F,$2E,$32,$D2,$1F
	.dc.b	$87,$F9,$1F,$C6,$7C,$97,$E4,$89,$1F,$29,$FC,$AF,$C9,$6A,$3E,$B4
	.dc.b	$7E,$4E,$3F,$C9,$FF,$27,$FD,$77,$B3,$FC,$5F,$F3,$AF,$8C,$FF,$52
	.dc.b	$F1,$FF,$8C,$B1,$4E,$5F,$F1,$AD,$A7,$53,$99,$4F,$57,$8D,$7F,$AC
	.dc.b	$38,$98,$7F,$8F,$BE,$27,$85,$F3,$BC,$5F,$89,$A7,$82,$F8,$AA,$E3
	.dc.b	$18,$C3,$9E,$3E,$4B,$E5,$7E,$23,$96,$3F,$02,$3C,$8D,$8F,$6C,$F1
	.dc.b	$69,$93,$DB,$7D,$EB,$8C,$78,$DE,$F8,$78,$A3,$D2,$70,$BB,$8F,$C5
	.dc.b	$C7,$53,$E5,$3E,$4E,$3E,$17,$81,$99,$E1,$87,$B2,$B8,$9F,$0A,$FC
	.dc.b	$D3,$AA,$F4,$9C,$6F,$A4,$66,$3A,$E1,$E2,$26,$1E,$C7,$A4,$71,$3A
	.dc.b	$D7,$2C,$7E,$14,$7C,$C3,$E1,$1E,$09,$E6,$E5,$E6,$61,$F7,$9F,$12
	.dc.b	$3E,$EC,$78,$3A,$FF,$CE,$38,$7E,$2D,$B3,$F8,$26,$E3,$FE,$88,$7F
	.dc.b	$CE,$B6,$F9,$BC,$61,$F9,$B8,$FF,$46,$C6,$1C,$35,$FD,$03,$5D,$9E
	.dc.b	$C3,$FD,$2B,$F0,$BF,$E2,$29,$FD,$04,$69,$B1,$9C,$FC,$13,$E3,$39
	.dc.b	$08,$D3,$E2,$32,$E1,$F1,$9F,$85,$38,$78,$3F,$93,$7C,$BA,$F8,$23
	.dc.b	$11,$DE,$F0,$F8,$0C,$5F,$C3,$BE,$D1,$CA,$31,$5C,$B0,$ED,$5F,$E4
	.dc.b	$0C,$3D,$F7,$EB,$7A,$98,$9D,$D7,$19,$F4,$57,$1B,$F8,$75,$A9,$F6
	.dc.b	$1E,$43,$F8,$84,$F9,$1C,$1F,$42,$BF,$56,$F9,$CD,$BD,$EF,$C7,$7D
	.dc.b	$A7,$C3,$39,$8F,$6C,$78,$2F,$9D,$F7,$6D,$76,$9F,$75,$70,$BF,$74
	.dc.b	$74,$3F,$86,$63,$FE,$0B,$F9,$4F,$F3,$2E,$7F,$DD,$BD,$4B,$8F,$E5
	.dc.b	$7B,$95,$8C,$3F,$BF,$D5,$C3,$FC,$28,$FF,$8C,$A3,$83,$5D,$DF,$F6
	.dc.b	$88,$C4,$FF,$86,$FE,$53,$96,$06,$FC,$A6,$19,$7F,$DF,$7F,$4C,$FF
	.dc.b	$77,$BF,$DE,$E3,$C5,$18,$C7,$F9,$BC,$FE,$B6,$3F,$23,$18,$18,$23
	.dc.b	$FC,$A7,$3D,$77,$E7,$E8,$C8,$FD,$1D,$6E,$3D,$35,$C2,$FF,$6D,$39
	.dc.b	$AD,$CB,$D1,$F9,$1E,$1F,$F8,$7F,$E0,$FC,$F1,$8F,$05,$14,$F9,$13
	.dc.b	$BB,$FE,$FF,$18,$9D,$C7,$F8,$53,$C7,$FF,$4D,$3F,$E2,$C6,$E5,$EF
	.dc.b	$CB,$DD,$E8,$F8,$A7,$3C,$7F,$EA,$4C,$91,$8C,$3E,$47,$C9,$EF,$E0
	.dc.b	$FE,$6B,$F3,$B9,$F9,$7C,$05,$AF,$03,$87,$C8,$FA,$AC,$D3,$E4,$FC
	.dc.b	$97,$8F,$E0,$7E,$CB,$6B,$D5,$1F,$73,$CF,$C6,$FE,$BE,$3F,$F3,$85
	.dc.b	$7E,$0F,$E4,$7D,$9E,$FF,$E0,$9B,$1F,$0E,$3F,$0B,$F1,$F8,$BE,$4F
	.dc.b	$E2,$3D,$0F,$CE,$47,$F3,$1F,$7F,$17,$FE,$5D,$8C,$37,$C1,$4F,$E7
	.dc.b	$67,$27,$F0,$67,$11,$A9,$FC,$1F,$C7,$FE,$58,$FC,$BF,$C7,$D8,$FC
	.dc.b	$E7,$A7,$E2,$F7,$BF,$45,$18,$3C,$4E,$31,$FC,$D0,$71,$D1,$FE,$86
	.dc.b	$DD,$F1,$DC,$FC,$1F,$D2,$BF,$A6,$F5,$3F,$4F,$EC,$5C,$BF,$58,$E1
	.dc.b	$A7,$EB,$E3,$19,$78,$3E,$EB,$F0,$F3,$F1,$57,$B1,$F2,$9C,$CE,$E3
	.dc.b	$07,$C7,$8D,$CF,$FA,$59,$B3,$FD,$41,$AF,$F8,$EF,$C0,$FD,$B7,$C7
	.dc.b	$7E,$F8,$ED,$EC,$DB,$FE,$09,$EC,$FF,$02,$3F,$81,$FC,$6D,$8F,$E1
	.dc.b	$1E,$37,$F1,$7F,$09,$1F,$C2,$23,$0E,$7C,$3F,$D3,$7B,$96,$3F,$58
	.dc.b	$E1,$FF,$47,$53,$E2,$F1,$AF,$EB,$E3,$13,$AF,$8B,$F0,$78,$9E,$16
	.dc.b	$C7,$28,$F2,$7C,$5E,$2F,$93,$89,$EF,$E2,$59,$EC,$FA,$FE,$2F,$D9
	.dc.b	$F1,$76,$BC,$AE,$1A,$7A,$23,$15,$E7,$F1,$3E,$21,$F1,$7E,$2E,$D7
	.dc.b	$BA,$3E,$47,$C5,$E2,$BD,$CF,$EA,$3D,$17,$8F,$7C,$7C,$37,$0F,$D5
	.dc.b	$4F,$FA,$E9,$FF,$6A,$B6,$FF,$C1,$ED,$7F,$BF,$57,$F4,$6E,$0F,$D5
	.dc.b	$47,$CD,$F1,$FC,$1E,$37,$DB,$7F,$96,$FC,$EB,$D9,$A7,$DF,$F8,$CB
	.dc.b	$9F,$EE,$FE,$F9,$F0,$3F,$B9,$6C,$7F,$BA,$1B,$97,$F8,$89,$AF,$FA
	.dc.b	$37,$D2,$62,$F8,$4F,$C6,$7C,$6F,$89,$B8,$7F,$B7,$A7,$9F,$97,$FC
	.dc.b	$BD,$B1,$FE,$41,$DF,$F7,$4E,$F3,$FE,$27,$F3,$BF,$EB,$D8,$FF,$7D
	.dc.b	$B7,$3F,$F1,$AB,$5F,$F2,$33,$50,$CF,$CF,$F0,$9F,$A1,$F8,$DF,$8E
	.dc.b	$DB,$FC,$DF,$C6,$FE,$94,$FD,$37,$C6,$D8,$FE,$45,$E0,$FF,$AA,$5F
	.dc.b	$F2,$2E,$3F,$A1,$F8,$5F,$E4,$53,$98,$FF,$24,$B5,$FD,$D6,$AF,$8C
	.dc.b	$FD,$97,$89,$FB,$36,$E1,$FE,$F9,$F9,$0F,$EE,$AF,$C9,$F8,$1F,$F0
	.dc.b	$74,$63,$0F,$EF,$5B,$FD,$29,$FE,$73,$97,$E0,$6F,$CA,$DA,$F1,$F4
	.dc.b	$79,$8F,$F0,$8E,$1F,$EB,$3A,$F8,$FC,$9F,$D8,$FE,$77,$FC,$0E,$FF
	.dc.b	$06,$BF,$F0,$3D,$6D,$F0,$CE,$AB,$FD,$A1,$F8,$13,$C2,$FE,$96,$D8
	.dc.b	$F6,$7C,$F6,$27,$FA,$7F,$D6,$3D,$FF,$E9,$4D,$AF,$9A,$DC,$7F,$CE
	.dc.b	$6D,$7F,$25,$8B,$E0,$BF,$C1,$AD,$AF,$95,$F2,$38,$9F,$93,$F0,$52
	.dc.b	$7E,$77,$AF,$70,$FE,$97,$8F,$F6,$C7,$BF,$C1,$79,$09,$5F,$E8,$4F
	.dc.b	$9B,$47,$E6,$27,$FC,$59,$18,$F7,$FE,$BF,$AF,$FE,$47,$8C,$5F,$F6
	.dc.b	$BF,$C6,$3E,$45,$FE,$59,$BE,$D3,$93,$F1,$5C,$BF,$46,$95,$C1,$68
	.dc.b	$F6,$BE,$49,$31,$F9,$2F,$DE,$E7,$FC,$49,$E3,$FD,$77,$FC,$69,$3F
	.dc.b	$8F,$8F,$F0,$0E,$1F,$27,$0F,$87,$ED,$E3,$8F,$EB,$B2,$FE,$AB,$F6
	.dc.b	$08,$FF,$AA,$3B,$1C,$CE,$FF,$AA,$7E,$6B,$E5,$2D,$7E,$40,$FD,$BD
	.dc.b	$70,$9F,$CF,$F9,$3A,$CB,$FC,$75,$C3,$FC,$3E,$FF,$4D,$F2,$9F,$A4
	.dc.b	$3F,$4B,$E4,$D8,$C9,$FA,$BF,$94,$C5,$F2,$7F,$57,$E4,$49,$FD,$87
	.dc.b	$CA,$5A,$EF,$3F,$B0,$F9,$4C,$73,$FE,$8F,$A3,$EB,$46,$70,$FF,$32
	.dc.b	$E8,$8F,$D6,$92,$33,$F2,$3F,$BA,$3F,$77,$F2,$36,$3F,$80,$47,$EF
	.dc.b	$BE,$1F,$19,$FC,$07,$FC,$6B,$F4,$2F,$81,$FF,$69,$4F,$FB,$5A,$31
	.dc.b	$EC,$FF,$14,$6C,$66,$76,$FF,$BA,$FE,$55,$CE,$F0,$3F,$8D,$C6,$1B
	.dc.b	$19,$9F,$F7,$1C,$E5,$F0,$7C,$17,$FA,$29,$BC,$1E,$18,$C4,$7F,$BE
	.dc.b	$71,$FF,$30,$CF,$C6,$F3,$7F,$9A,$CF,$F9,$D7,$E9,$FF,$DF,$DC,$7F
	.dc.b	$35,$BC,$1F,$CE,$AF,$F4,$97,$C2,$B5,$B5,$AD,$71,$AB,$18,$FF,$18
	.dc.b	$6E,$0C,$7E,$1E,$3D,$FE,$4F,$F9,$DC,$E5,$B2,$69,$8F,$93,$E7,$19
	.dc.b	$C3,$F3,$47,$FA,$B1,$C3,$ED,$8D,$F8,$1C,$19,$1C,$2F,$C9,$3D,$3B
	.dc.b	$7A,$FC,$94,$AD,$61,$FA,$B7,$BF,$7F,$EB,$BE,$03,$84,$6A,$FE,$0C
	.dc.b	$E6,$38,$F9,$DE,$8F,$2E,$46,$0E,$7E,$1B,$DF,$FE,$BB,$7F,$BA,$6C
	.dc.b	$66,$7F,$7C,$D7,$ED,$9F,$8B,$FA,$1C,$3F,$7B,$A9,$18,$8F,$EE,$B1
	.dc.b	$8A,$FE,$D6,$D8,$FF,$67,$E8,$F9,$3F,$B7,$39,$9F,$E5,$F7,$E1,$E3
	.dc.b	$FC,$45,$7D,$4D,$A5,$7C,$87,$EB,$FC,$17,$2B,$FF,$42,$CF,$F3,$B6
	.dc.b	$BE,$AC,$3C,$FF,$1F,$E4,$9C,$CF,$F9,$CF,$2F,$07,$47,$E5,$57,$C9
	.dc.b	$7E,$10,$FC,$2E,$BF,$4B,$67,$C8,$52,$61,$F7,$3E,$4B,$DB,$E0,$7F
	.dc.b	$A2,$A3,$FA,$73,$5F,$23,$5F,$98,$D7,$F9,$4D,$7F,$43,$7F,$0F,$A7
	.dc.b	$E5,$8F,$17,$A1,$E3,$F3,$64,$FE,$01,$FC,$DF,$C8,$F8,$B0,$FA,$67
	.dc.b	$F0,$69,$F1,$B8,$3F,$07,$F1,$FF,$A2,$CE,$4F,$E2,$A3,$F4,$B8,$7E
	.dc.b	$9B,$09,$3F,$9A,$8F,$CA,$FD,$17,$F1,$B8,$FE,$90,$D7,$F1,$7C,$E8
	.dc.b	$C4,$7F,$94,$5A,$FD,$1F,$91,$FD,$9E,$12,$3F,$4F,$1F,$C0,$3E,$6B
	.dc.b	$FB,$57,$4F,$F3,$95,$F2,$59,$35,$CB,$E4,$F9,$0F,$D0,$E4,$69,$31
	.dc.b	$FC,$F8,$F0,$58,$7F,$A8,$E3,$F7,$8F,$AF,$1F,$07,$95,$E3,$F8,$1F
	.dc.b	$83,$F1,$DE,$5F,$2E,$61,$FB,$CC,$D3,$FC,$6A,$1F,$8D,$BE,$0B,$FC
	.dc.b	$A3,$18,$6C,$66,$3F,$CA,$AD,$7C,$FF,$3D,$FE,$35,$7F,$CB,$2D,$73
	.dc.b	$F0,$7D,$23,$11,$F3,$FE,$3F,$E5,$7C,$A4,$8F,$34,$79,$78,$5E,$AF
	.dc.b	$91,$B3,$FE,$66,$6B,$F2,$1A,$FD,$BF,$82,$FC,$19,$1F,$93,$F2,$9F
	.dc.b	$67,$E5,$24,$7C,$C7,$E2,$7A,$7F,$1A,$3F,$98,$9F,$E2,$9F,$93,$F8
	.dc.b	$B0,$69,$23,$FC,$F3,$0C,$B9,$CC,$9E,$73,$97,$ED,$E3,$F8,$BF,$94
	.dc.b	$91,$83,$EF,$ED,$F8,$F9,$91,$83,$EE,$7C,$5F,$E4,$F8,$A4,$60,$87
	.dc.b	$F9,$F3,$8F,$FD,$0C,$E3,$9F,$F3,$19,$FE,$6F,$49,$3F,$8E,$7F,$9F
	.dc.b	$F2,$BE,$6E,$9F,$CE,$9C,$4F,$F8,$D5,$AF,$EB,$9A,$FF,$74,$EA,$FC
	.dc.b	$DF,$CA,$72,$8F,$E9,$F3,$C3,$1F,$D5,$4F,$83,$E2,$FF,$5D,$C9,$2F
	.dc.b	$71,$FB,$1F,$92,$D4,$60,$FD,$2B,$FB,$5F,$92,$91,$FA,$87,$D7,$F2
	.dc.b	$32,$3C,$2D,$58,$FE,$EF,$E4,$BF,$7D,$C5,$27,$F6,$2F,$EA,$7D,$07
	.dc.b	$C4,$5F,$88,$FD,$4C,$7F,$B8,$AB,$FD,$76,$D7,$F7,$AD,$7F,$AE,$AF
	.dc.b	$88,$E5,$1F,$90,$8E,$18,$FF,$9D,$DD,$47,$C8,$F8,$9F,$E2,$5F,$25
	.dc.b	$23,$04,$3F,$D7,$7C,$5F,$EE,$B7,$17,$FE,$4F,$F9,$0F,$F1,$5B,$FE
	.dc.b	$B3,$F3,$15,$DA,$BE,$1F,$93,$FE,$27,$BF,$AC,$DE,$7F,$C2,$E9,$FE
	.dc.b	$22,$9F,$E0,$98,$F3,$C3,$93,$67,$5E,$36,$FE,$46,$65,$FE,$17,$C1
	.dc.b	$1C,$F9,$BF,$8B,$7D,$3B,$DC,$FE,$2A,$DB,$FF,$73,$B5,$FE,$52,$D7
	.dc.b	$F2,$17,$FD,$7F,$2F,$F5,$04,$ED,$B5,$E9,$9F,$CF,$F8,$DE,$BF,$92
	.dc.b	$98,$7E,$5A,$3D,$DD,$1A,$8C,$1E,$97,$FC,$4B,$F3,$DF,$E4,$E8,$FE
	.dc.b	$75,$6F,$91,$F8,$C6,$A3,$E4,$7A,$D5,$FF,$18,$D7,$1F,$01,$7F,$D9
	.dc.b	$E7,$FD,$EB,$2F,$86,$FF,$15,$B6,$3F,$D4,$CD,$7F,$AF,$B5,$F8,$7D
	.dc.b	$1E,$69,$91,$F4,$63,$D3,$F0,$9F,$6F,$B2,$4F,$92,$32,$D2,$1F,$86
	.dc.b	$3D,$FD,$9F,$0B,$E0,$FC,$A4,$8F,$2B,$FE,$64,$CB,$F0,$BD,$5F,$EB
	.dc.b	$6C,$FF,$DC,$35,$F0,$AB,$F1,$9D,$1F,$8D,$F8,$FE,$04,$9F,$EC,$FB
	.dc.b	$7C,$68,$E9,$F2,$1B,$9F,$C6,$47,$E5,$63,$F2,$DF,$21,$F9,$7F,$90
	.dc.b	$95,$F9,$18,$FC,$D7,$2E,$A3,$07,$E5,$5F,$CF,$7C,$87,$A4,$91,$E2
	.dc.b	$D9,$1E,$DF,$8D,$DC,$FE,$76,$3F,$4D,$1F,$DC,$7E,$43,$F0,$9D,$6A
	.dc.b	$7C,$EB,$E3,$7F,$D3,$74,$EB,$FF,$4E,$CE,$6B,$84,$E5,$E5,$19,$51
	.dc.b	$D9,$EB,$5D,$3A,$FF,$8F,$C7,$F3,$D8,$EF,$7F,$EA,$AF,$0F,$95,$F5
	.dc.b	$AD,$1E,$0B,$E3,$3A,$79,$E1,$C2,$FC,$F8,$74,$BF,$2B,$E1,$C6,$C7
	.dc.b	$36,$BE,$89,$E2,$FB,$9A,$FB,$2F,$BD,$73,$8F,$29,$C4,$F9,$98,$78
	.dc.b	$DB,$1B,$8F,$45,$F8,$31,$F7,$63,$BB,$F8,$E7,$94,$A7,$86,$78,$63
	.dc.b	$E4,$AE,$38,$78,$EF,$AD,$F0,$9C,$BE,$B3,$C1,$8F,$A9,$ED,$8F,$84
	.dc.b	$97,$E7,$3A,$1F,$2E,$3A,$61,$E1,$AE,$51,$E5,$3E,$1D,$FD,$9B,$E7
	.dc.b	$1D,$AF,$A3,$E3,$37,$7F,$81,$8C,$C7,$3E,$2F,$EF,$B7,$F8,$B6,$E1
	.dc.b	$F8,$A6,$E5,$FE,$ED,$C7,$FC,$B7,$F3,$77,$0F,$F5,$DB,$5F,$F7,$BB
	.dc.b	$5F,$D1,$35,$3F,$61,$FF,$7E,$FC,$8B,$A3,$11,$C8,$FC,$08,$8D,$38
	.dc.b	$39,$9A,$73,$5F,$A7,$8F,$09,$99,$EA,$E3,$5D,$57,$47,$84,$74,$73
	.dc.b	$3F,$9A,$8E,$75,$87,$C5,$7D,$2F,$C3,$A7,$C6,$9F,$C6,$9E,$9B,$E4
	.dc.b	$F5,$8F,$F7,$1C,$F8,$5E,$18,$75,$3C,$77,$E1,$9D,$57,$D0,$3A,$B8
	.dc.b	$7E,$44,$F1,$AF,$16,$1E,$73,$9C,$FC,$33,$D2,$72,$C3,$CB,$87,$76
	.dc.b	$C6,$4F,$59,$E9,$3E,$CB,$A9,$ED,$5E,$03,$DB,$1F,$42,$78,$1F,$C9
	.dc.b	$70,$F4,$3F,$4E,$76,$FE,$DB,$27,$DD,$FF,$2B,$BF,$E5,$99,$FE,$0A
	.dc.b	$73,$87,$E1,$23,$15,$BB,$FE,$B5,$5A,$BE,$27,$F3,$57,$11,$FC,$DC
	.dc.b	$C4,$FD,$77,$07,$FA,$E5,$C3,$FE,$61,$6B,$C3,$E4,$BC,$55,$CB,$BE
	.dc.b	$CF,$19,$B7,$EC,$9B,$7F,$E0,$1B,$F2,$DF,$7F,$92,$CD,$69,$DF,$77
	.dc.b	$0D,$2D,$F0,$C7,$E3,$C6,$63,$FB,$5E,$3F,$D4,$5B,$1F,$AA,$C7,$D4
	.dc.b	$D7,$E0,$FC,$97,$A3,$CA,$77,$FF,$8D,$EF,$B6,$0A,$D4,$E6,$FF,$C3
	.dc.b	$7E,$2F,$35,$87,$28,$C3,$70,$FF,$0E,$C6,$29,$F1,$BD,$0F,$8B,$E0
	.dc.b	$DA,$FF,$15,$C7,$C8,$F9,$2C,$63,$F7,$B3,$47,$CC,$C7,$CB,$F1,$97
	.dc.b	$0F,$F1,$9B,$87,$FC,$5A,$F8,$48,$FC,$8F,$92,$F6,$FC,$67,$D6,$6C
	.dc.b	$7F,$8F,$23,$EC,$E5,$8C,$BC,$4F,$DC,$F9,$0F,$8B,$F1,$56,$FF,$C8
	.dc.b	$6E,$1F,$97,$F2,$16,$FD,$4F,$E1,$32,$FC,$4C,$3E,$9D,$EB,$66,$FF
	.dc.b	$DB,$71,$86,$C7,$D1,$9E,$0D,$9D,$47,$E7,$0C,$CF,$E1,$BE,$43,$F2
	.dc.b	$67,$E5,$34,$B1,$F8,$0F,$8F,$FC,$27,$CF,$BE,$37,$F3,$31,$86,$4F
	.dc.b	$E1,$BE,$46,$E3,$F5,$5E,$3F,$15,$27,$F1,$7B,$DA,$FF,$28,$74,$32
	.dc.b	$5F,$E9,$BE,$47,$5E,$0F,$EF,$9F,$8B,$F3,$AF,$3A,$F9,$F8,$6E,$9F
	.dc.b	$F0,$02,$53,$F1,$5B,$1F,$0F,$34,$61,$F8,$CD,$7F,$DC,$1F,$77,$FD
	.dc.b	$5E,$7F,$60,$D8,$FF,$34,$1C,$BD,$0F,$DF,$C6,$72,$FF,$5A,$68,$F8
	.dc.b	$BF,$96,$6B,$E4,$EE,$3D,$3E,$4B,$89,$8C,$BF,$8D,$E2,$8E,$BC,$7F
	.dc.b	$53,$E2,$6D,$FF,$0A,$E2,$F9,$51,$98,$F3,$FC,$0D,$C7,$9B,$E6,$F9
	.dc.b	$0F,$BB,$F1,$16,$B9,$6C,$8C,$C7,$E2,$71,$FA,$D1,$FB,$C9,$D4,$E5
	.dc.b	$B8,$7F,$2A,$F2,$5A,$FD,$CA,$F9,$3C,$DD,$2B,$DA,$FE,$3B,$E1,$A4
	.dc.b	$FC,$EF,$90,$B5,$F8,$6C,$55,$F3,$BE,$3B,$3C,$9F,$D1,$72,$47,$D2
	.dc.b	$E4,$FA,$7F,$1F,$73,$FB,$2E,$1F,$B7,$F1,$F7,$3F,$9E,$FE,$CB,$D3
	.dc.b	$78,$37,$C1,$C3,$6F,$F9,$CD,$FF,$85,$E7,$F6,$54,$69,$27,$FD,$3F
	.dc.b	$9B,$9B,$F0,$E7,$7F,$06,$CF,$DB,$F9,$F7,$C4,$FF,$35,$6C,$7E,$4B
	.dc.b	$E3,$F1,$39,$F8,$9F,$E5,$F4,$69,$2B,$ED,$47,$FB,$51,$FC,$DF,$C7
	.dc.b	$F5,$F8,$1F,$CC,$BF,$C4,$BE,$75,$E4,$7F,$0D,$6D,$7E,$A5,$BC,$1F
	.dc.b	$C3,$5A,$FF,$6D,$E6,$7E,$BF,$F6,$2B,$F8,$94,$63,$A3,$FA,$43,$63
	.dc.b	$33,$BA,$FE,$E7,$DC,$D3,$FD,$55,$5F,$B2,$EE,$F8,$9F,$03,$FB,$07
	.dc.b	$F6,$3F,$23,$FB,$2F,$91,$B5,$FB,$87,$E5,$7C,$57,$C6,$24,$7C,$47
	.dc.b	$FA,$D7,$85,$F9,$8D,$BF,$77,$E7,$FE,$F1,$B5,$FC,$0B,$E4,$3F,$BE
	.dc.b	$FE,$20,$BF,$E0,$31,$87,$EC,$79,$DF,$50,$95,$F4,$9E,$F4,$F8,$2E
	.dc.b	$7D,$3D,$1B,$5F,$83,$F8,$8E,$FF,$23,$F8,$36,$C7,$F8,$2B,$E6,$9E
	.dc.b	$F3,$C6,$FE,$E9,$18,$6C,$66,$7F,$A2,$4E,$5F,$DF,$7C,$17,$F8,$AF
	.dc.b	$AB,$FC,$E3,$D0,$8C,$4F,$F8,$AF,$5F,$C7,$3F,$C4,$63,$3D,$3F,$E0
	.dc.b	$25,$F0,$BF,$1B,$BF,$FC,$1D,$18,$AF,$F0,$D5,$FE,$4E,$D8,$FF,$A0
	.dc.b	$32,$F5,$F7,$FA,$09,$7E,$8D,$11,$FE,$0E,$C7,$FC,$59,$87,$F4,$C6
	.dc.b	$E1,$FE,$49,$C5,$F2,$3E,$2F,$86,$FB,$3D,$BF,$6B,$E2,$6E,$9E,$47
	.dc.b	$FC,$77,$AB,$E7,$5F,$11,$F3,$1B,$5F,$29,$BA,$FF,$92,$AB,$85,$7D
	.dc.b	$EF,$92,$FC,$39,$F8,$8F,$92,$B8,$7C,$2E,$FE,$F8,$FF,$8B,$E3,$07
	.dc.b	$DD,$71,$8F,$C6,$82,$C7,$CD,$F1,$3E,$52,$4F,$95,$F8,$C4,$7E,$FB
	.dc.b	$67,$F9,$C6,$EF,$92,$FC,$EF,$C9,$7E,$79,$B7,$F7,$FE,$27,$F4,$3F
	.dc.b	$25,$F7,$1B,$1F,$72,$3F,$4B,$F2,$58,$C7,$FC,$EB,$8A,$3F,$53,$E0
	.dc.b	$FE,$AB,$E0,$AE,$1F,$84,$70,$FE,$1B,$E0,$6E,$1C,$9F,$D8,$9F,$AA
	.dc.b	$24,$7F,$87,$FC,$46,$29,$FD,$7F,$83,$FB,$8F,$92,$FD,$CB,$6F,$F2
	.dc.b	$71,$FE,$78,$F7,$BF,$37,$87,$28,$C6,$3F,$E3,$6F,$9C,$FF,$12,$7B
	.dc.b	$1F,$E0,$B6,$C7,$F8,$BD,$AF,$FA,$31,$AF,$FA,$47,$E4,$DD,$3F,$66
	.dc.b	$3F,$C6,$D1,$8E,$6F,$F0,$B3,$63,$33,$B7,$A7,$C0,$FF,$11,$7C,$0D
	.dc.b	$73,$97,$96,$31,$1F,$EA,$3C,$FF,$DD,$CF,$FA,$2D,$FD,$BB,$7C,$5F
	.dc.b	$E7,$5A,$EF,$4B,$3F,$C7,$EB,$8B,$FE,$B0,$6F,$83,$FE,$B7,$6D,$77
	.dc.b	$F0,$38,$E4,$9F,$D5,$DF,$D7,$6B,$FE,$09,$6B,$FE,$B7,$8E,$1D,$DF
	.dc.b	$E3,$97,$97,$85,$D2,$B8,$E3,$E1,$AC,$4C,$BF,$73,$5F,$47,$C8,$F1
	.dc.b	$ED,$E5,$E2,$C9,$F8,$91,$96,$90,$F8,$1D,$9F,$13,$C7,$FE,$4D,$F1
	.dc.b	$9E,$3F,$8F,$95,$FC,$96,$35,$91,$A4,$8F,$A5,$F1,$FF,$3B,$9D,$F8
	.dc.b	$53,$AB,$FC,$B1,$18,$6C,$66,$BF,$9D,$B5,$F9,$5F,$21,$CE,$31,$C3
	.dc.b	$FC,$B5,$BC,$BF,$83,$9F,$95,$97,$CB,$F9,$09,$87,$E1,$63,$C9,$F5
	.dc.b	$5F,$86,$8F,$E3,$CD,$7F,$DD,$99,$A3,$11,$FC,$5A,$31,$D3,$F6,$79
	.dc.b	$9F,$23,$F6,$49,$5F,$93,$F8,$4F,$B9,$F0,$9F,$77,$E0,$A4,$FE,$5E
	.dc.b	$3E,$37,$D4,$7F,$90,$23,$FA,$DB,$5F,$E6,$4D,$7C,$AD,$7F,$45,$F0
	.dc.b	$3F,$89,$23,$FA,$3F,$91,$FA,$FF,$23,$F8,$D2,$47,$F5,$F7,$C3,$F0
	.dc.b	$DF,$52,$FF,$C7,$93,$E2,$8C,$1F,$E8,$FE,$A7,$CD,$3E,$47,$F1,$0D
	.dc.b	$AF,$F3,$D3,$5F,$E7,$C7,$F8,$37,$F1,$AF,$F7,$D7,$4B,$A3,$FA,$0F
	.dc.b	$86,$9E,$13,$9A,$E1,$F2,$5F,$B0,$9F,$F4,$54,$7F,$48,$9D,$57,$1C
	.dc.b	$7C,$7C,$9E,$7C,$BF,$6D,$7F,$E8,$16,$BF,$E3,$46,$BF,$B5,$C3,$FD
	.dc.b	$8C,$F5,$CB,$CB,$0F,$1F,$C4,$78,$FE,$4B,$C9,$84,$BE,$07,$BD,$F4
	.dc.b	$7C,$97,$C6,$F7,$BC,$84,$9F,$23,$FC,$0F,$07,$C9,$7B,$C9,$3D,$35
	.dc.b	$F8,$DF,$23,$F4,$3E,$3A,$4F,$D0,$7E,$37,$C8,$FF,$1E,$87,$90,$F3
	.dc.b	$7F,$D9,$11,$86,$C6,$63,$FD,$98,$D7,$E3,$7C,$5F,$DE,$23,$E3,$CF
	.dc.b	$FC,$6C,$FF,$A2,$BE,$72,$BF,$EF,$76,$BF,$DF,$8F,$EF,$EE,$1C,$35
	.dc.b	$FF,$01,$FC,$EF,$F3,$E7,$83,$5C,$E9,$EA,$8C,$74,$7E,$53,$E3,$3F
	.dc.b	$26,$4A,$F3,$7C,$9F,$E2,$7A,$FF,$15,$BC,$9F,$54,$7E,$6B,$E3,$7F
	.dc.b	$10,$48,$FC,$43,$F5,$7E,$1F,$51,$F9,$E8,$FC,$A4,$7E,$57,$6F,$CB
	.dc.b	$7C,$74,$AF,$8D,$19,$69,$0F,$D2,$EB,$FA,$1C,$FF,$3D,$F1,$D2,$7E
	.dc.b	$73,$FA,$1C,$64,$7D,$16,$BC,$B8,$1F,$1D,$FD,$5B,$F2,$8F,$ED,$5E
	.dc.b	$07,$E9,$9B,$DD,$FB,$06,$DB,$89,$F3,$AF,$8D,$FE,$88,$D9,$FF,$75
	.dc.b	$35,$FF,$4E,$B5,$FB,$D3,$CA,$7F,$AF,$57,$FA,$16,$7F,$DF,$D5,$C7
	.dc.b	$0F,$F3,$AB,$70,$F2,$7C,$7F,$83,$E3,$3D,$5F,$21,$30,$E3,$1E,$DF
	.dc.b	$90,$D4,$60,$FD,$36,$48,$F8,$5F,$29,$27,$2F,$C5,$D3,$E3,$7C,$8D
	.dc.b	$9E,$31,$87,$FD,$CB,$A3,$1F,$99,$F2,$32,$73,$1F,$AF,$9F,$F1,$1C
	.dc.b	$7E,$7C,$FD,$03,$87,$07,$CE,$F8,$3F,$D9,$C6,$63,$82,$F9,$1E,$93
	.dc.b	$27,$C7,$1F,$76,$A4,$7C,$07,$F0,$13,$23,$E1,$B5,$FC,$7E,$9D,$5F
	.dc.b	$A7,$F1,$9F,$87,$F9,$19,$18,$3E,$5F,$C7,$FE,$0A,$64,$60,$FE,$3E
	.dc.b	$3F,$07,$79,$5F,$91,$F9,$19,$3A,$9C,$BC,$3F,$50,$FE,$FC,$FC,$76
	.dc.b	$BF,$C1,$DA,$FF,$A1,$9A,$F7,$F0,$FF,$8F,$47,$F8,$0A,$35,$F0,$5F
	.dc.b	$C8,$69,$5F,$9A,$F8,$EF,$D2,$7C,$94,$BF,$C9,$C7,$E9,$F3,$D4,$60
	.dc.b	$FC,$5B,$FE,$25,$FA,$1F,$B9,$D9,$FA,$BF,$93,$FD,$71,$9A,$FD,$87
	.dc.b	$C9,$D7,$FA,$1F,$7A,$F8,$F7,$FE,$C6,$BA,$79,$DF,$DB,$67,$F9,$85
	.dc.b	$FE,$A9,$AF,$FA,$9A,$F9,$4F,$EB,$33,$E3,$3D,$AB,$F9,$C6,$5F,$BF
	.dc.b	$9F,$47,$C4,$FF,$03,$F9,$29,$7B,$8C,$B4,$87,$E1,$B8,$7E,$07,$C6
	.dc.b	$F9,$49,$1F,$9A,$CF,$D9,$1F,$E5,$BF,$CF,$FD,$B3,$FC,$2F,$3F,$F7
	.dc.b	$DB,$E5,$9F,$F3,$4C,$EB,$CA,$3F,$E7,$29,$FF,$33,$35,$FF,$02,$B5
	.dc.b	$47,$0F,$91,$F0,$7C,$37,$46,$4F,$F6,$7A,$FF,$1F,$6D,$FE,$06,$7E
	.dc.b	$2F,$0F,$4F,$91,$F1,$9E,$4E,$19,$7C,$5A,$FF,$36,$A6,$FE,$AF,$91
	.dc.b	$F8,$04,$9F,$F5,$C4,$A3,$E2,$7C,$8E,$E7,$C9,$1F,$0A,$3F,$BB,$7C
	.dc.b	$97,$F0,$DC,$FF,$DC,$7F,$41,$C6,$FF,$DC,$B2,$F1,$BF,$B5,$4E,$65
	.dc.b	$39,$78,$E1,$FE,$B5,$F1,$18,$62,$79,$C6,$17,$99,$CE,$F9,$46,$2F
	.dc.b	$FD,$7D,$8F,$F7,$9A,$D5,$F1,$6E,$1D,$6F,$C7,$5E,$6B,$E3,$19,$C3
	.dc.b	$DC,$B1,$E2,$D9,$DB,$4F,$83,$5C,$DF,$4D,$F3,$C7,$C5,$5C,$31,$F5
	.dc.b	$5F,$9E,$3E,$15,$2F,$D8,$72,$3B,$1E,$83,$CD,$3D,$DF,$9B,$7F,$1F
	.dc.b	$0F,$39,$DA,$36,$78,$EA,$3D,$31,$F9,$73,$B8,$C5,$73,$C3,$CB,$1E
	.dc.b	$43,$C4,$5A,$FA,$78,$FC,$63,$C3,$3F,$1A,$FA,$18,$AF,$64,$F7,$24
	.dc.b	$F6,$79,$3F,$1C,$91,$C1,$F1,$F1,$7F,$77,$8F,$83,$8F,$F9,$DA,$B2
	.dc.b	$48,$FE,$BD,$F4,$19,$69,$FB,$68,$FE,$BF,$19,$8F,$F2,$13,$5F,$A8
	.dc.b	$FD,$58,$FC,$E4,$38,$BF,$D0,$51,$88,$4E,$F0,$FE,$DD,$5A,$7B,$CF
	.dc.b	$CB,$39,$71,$6D,$F2,$DF,$6A,$E1,$1E,$8C,$B1,$5F,$9B,$8F,$29,$F8
	.dc.b	$63,$99,$4F,$19,$DE,$79,$BE,$E3,$11,$E1,$71,$3D,$9E,$58,$77,$AF
	.dc.b	$F5,$04,$FB,$E3,$15,$E3,$AE,$F3,$C4,$D3,$E5,$AE,$38,$7C,$29,$D3
	.dc.b	$F1,$30,$F0,$1C,$0A,$EC,$F9,$77,$DF,$0F,$0E,$1E,$63,$B4,$65,$A6
	.dc.b	$6B,$9B,$F3,$AB,$9B,$F4,$CF,$4B,$8B,$F4,$9E,$68,$FE,$7B,$BB,$93
	.dc.b	$FD,$85,$9F,$F3,$9F,$6A,$E5,$FD,$12,$6D,$F0,$9B,$AF,$F4,$C7,$FD
	.dc.b	$95,$1F,$D3,$1A,$FF,$4C,$6B,$C6,$BF,$9C,$FC,$A7,$2C,$4D,$F9,$4C
	.dc.b	$32,$7F,$A7,$7C,$9D,$BF,$E8,$CF,$1F,$92,$E5,$F2,$76,$BF,$AA,$38
	.dc.b	$7F,$A3,$BE,$2C,$FF,$B1,$B7,$23,$7E,$47,$DF,$DB,$FE,$0B,$CF,$CD
	.dc.b	$18,$ED,$F3,$E3,$FE,$03,$9F,$6D,$6E,$3F,$BC,$4E,$B4,$FF,$2B,$4E
	.dc.b	$5B,$0D,$3A,$1F,$0F,$E4,$6D,$FF,$83,$E3,$E3,$7C,$86,$2F,$FB,$33
	.dc.b	$F2,$7E,$47,$C9,$9D,$BF,$F0,$B3,$87,$C6,$F9,$B1,$34,$F8,$5F,$23
	.dc.b	$ED,$BF,$A8,$DA,$FF,$10,$47,$D7,$F9,$1C,$63,$FD,$DD,$FB,$5F,$23
	.dc.b	$F1,$5B,$5F,$E2,$77,$0F,$FB,$B7,$A0,$E2,$BF,$D6,$B1,$88,$FF,$2F
	.dc.b	$FE,$88,$E4,$FC,$8D,$3E,$7F,$FD,$63,$5A,$9C,$C6,$27,$FB,$9C,$FF
	.dc.b	$B8,$E3,$11,$FE,$87,$6B,$9F,$77,$F2,$AF,$56,$DF,$E1,$2B,$D1,$F2
	.dc.b	$3E,$28,$C5,$7A,$23,$F4,$78,$BE,$4B,$F1,$3E,$DA,$73,$18,$66,$1F
	.dc.b	$0F,$E4,$2E,$1F,$09,$E7,$F2,$32,$7E,$47,$C8,$DA,$F8,$FE,$C7,$80
	.dc.b	$C7,$83,$F3,$23,$FC,$73,$A7,$FA,$1A,$FF,$D1,$3F,$3E,$F4,$3E,$5A
	.dc.b	$F4,$2C,$61,$93,$FE,$AA,$B5,$6E,$30,$D8,$FF,$57,$B5,$FE,$8E,$7C
	.dc.b	$4F,$91,$FB,$3E,$4F,$F8,$16,$BF,$05,$1F,$23,$E4,$7E,$4F,$C8,$D9
	.dc.b	$FC,$33,$F3,$3E,$4A,$46,$0F,$C5,$3F,$43,$E4,$3E,$7D,$59,$FF,$41
	.dc.b	$F2,$A3,$31,$F9,$1F,$90,$B9,$FE,$41,$FB,$3E,$E7,$D8,$C2,$D7,$27
	.dc.b	$0F,$F9,$CF,$04,$7F,$00,$6C,$66,$3F,$87,$70,$3E,$43,$F0,$A7,$2C
	.dc.b	$7F,$3E,$FE,$23,$E3,$E5,$FE,$97,$E4,$2D,$7E,$93,$E4,$3F,$1A,$63
	.dc.b	$97,$EF,$79,$BF,$E6,$18,$FC,$8F,$C8,$4C,$3F,$5D,$F2,$17,$0F,$D6
	.dc.b	$CF,$E5,$BC,$0C,$63,$CB,$5F,$D4,$3D,$3F,$50,$FF,$40,$E3,$F2,$AF
	.dc.b	$F9,$FC,$63,$0F,$D6,$8F,$35,$C2,$AF,$D0,$9F,$14,$97,$F2,$5F,$E6
	.dc.b	$57,$FE,$64,$8F,$F2,$46,$1E,$1F,$90,$EF,$EB,$78,$1F,$E6,$7F,$43
	.dc.b	$E1,$9F,$CF,$64,$8C,$5F,$F9,$47,$81,$8B,$E4,$78,$C7,$2C,$7F,$9B
	.dc.b	$D9,$A4,$BF,$D0,$7C,$FB,$CD,$FE,$0A,$D8,$CC,$60,$E7,$87,$98,$F8
	.dc.b	$FF,$25,$67,$FD,$AB,$87,$C6,$7F,$D7,$8D,$E0,$FE,$8A,$D7,$FC,$D4
	.dc.b	$7C,$D7,$0F,$FB,$55,$7C,$97,$C4,$3E,$AF,$0D,$AF,$96,$7C,$3D,$76
	.dc.b	$E2,$FF,$D7,$B1,$87,$ED,$EB,$F2,$3D,$FF,$BF,$87,$D0,$F9,$1F,$8E
	.dc.b	$4C,$3D,$2F,$DA,$F8,$0F,$C2,$9F,$86,$F9,$1B,$3D,$F0,$FA,$75,$F8
	.dc.b	$AF,$90,$FC,$5B,$67,$E2,$FC,$5F,$D4,$23,$F1,$9F,$C3,$FC,$87,$E4
	.dc.b	$BE,$43,$F2,$6D,$9F,$9B,$F1,$7F,$86,$3F,$2F,$F1,$D6,$3E,$06,$EF
	.dc.b	$A9,$78,$BF,$D0,$1B,$3F,$C7,$DB,$3F,$48,$FF,$4E,$38,$7E,$AF,$57
	.dc.b	$E4,$7C,$DF,$D2,$E9,$F9,$66,$DF,$C1,$E4,$7C,$7F,$EA,$8F,$D5,$FC
	.dc.b	$7D,$9F,$2E,$4F,$13,$C8,$48,$F2,$3F,$A8,$DB,$F6,$7E,$87,$ED,$1B
	.dc.b	$3D,$F2,$FD,$B9,$FE,$36,$F8,$AF,$E6,$1B,$7D,$ED,$3F,$9C,$78,$3F
	.dc.b	$E3,$8F,$A8,$FF,$77,$E1,$FC,$FD,$B5,$F0,$43,$87,$E2,$35,$FD,$85
	.dc.b	$FF,$9E,$2B,$F8,$AA,$F9,$0F,$17,$A5,$E3,$7C,$9F,$1F,$FD,$F3,$0F
	.dc.b	$F2,$96,$3F,$87,$B4,$E7,$E2,$E4,$F9,$5F,$EE,$F5,$C3,$2F,$33,$5F
	.dc.b	$AB,$F4,$4F,$2B,$FC,$76,$DB,$ED,$F2,$38,$99,$1E,$1F,$91,$B7,$FA
	.dc.b	$4C,$FE,$1E,$DE,$CB,$F9,$87,$CD,$F9,$1B,$87,$A6,$3E,$86,$98,$C3
	.dc.b	$93,$F4,$FE,$3E,$57,$BF,$E3,$2D,$77,$9F,$B1,$F1,$F8,$CB,$E5,$B8
	.dc.b	$BF,$A9,$E1,$B4,$4F,$F9,$75,$B3,$A6,$E1,$F9,$BA,$E1,$F1,$9F,$85
	.dc.b	$3F,$0D,$F0,$D6,$FE,$94,$7E,$27,$E3,$B1,$7F,$0D,$FF,$19,$FC,$F3
	.dc.b	$E1,$3F,$D9,$53,$F8,$78,$C7,$47,$F8,$25,$B1,$99,$DB,$FA,$DC,$DE
	.dc.b	$4B,$E7,$3F,$C2,$11,$86,$C6,$67,$F2,$93,$97,$C1,$F0,$AD,$BC,$0D
	.dc.b	$8C,$C6,$0F,$89,$F1,$BF,$CB,$B8,$7F,$2F,$F2,$17,$83,$F3,$6D,$BE
	.dc.b	$98,$FF,$9A,$1F,$CB,$5E,$70,$FA,$4D,$9D,$4F,$E2,$7E,$3B,$5F,$4C
	.dc.b	$F2,$3F,$C5,$98,$23,$15,$F7,$1A,$FD,$76,$BF,$D1,$23,$87,$97,$D2
	.dc.b	$A4,$FF,$2F,$A4,$F8,$09,$19,$7C,$7E,$4F,$9F,$E4,$7D,$1E,$C4,$AF
	.dc.b	$D9,$46,$5A,$43,$F4,$9F,$1F,$F1,$7C,$D7,$87,$FE,$77,$9F,$DD,$7C
	.dc.b	$87,$C3,$35,$19,$8F,$F0,$4C,$F0,$EC,$C9,$23,$FA,$47,$E2,$9F,$EC
	.dc.b	$CE,$5F,$F6,$6F,$B6,$F8,$9F,$F2,$A5,$7F,$7D,$6C,$66,$3F,$CB,$0D
	.dc.b	$7F,$9D,$75,$3E,$83,$FC,$C5,$1F,$D1,$27,$57,$DA,$FF,$01,$7B,$F3
	.dc.b	$FE,$F6,$3F,$E7,$BC,$3C,$BF,$1F,$83,$7F,$53,$FB,$28,$E9,$F1,$9E
	.dc.b	$42,$3E,$47,$FC,$D9,$83,$E8,$7F,$47,$9F,$F9,$E6,$31,$E1,$F8,$A7
	.dc.b	$3D,$FF,$E1,$DC,$FF,$D1,$D8,$FF,$7B,$BF,$D4,$FC,$1F,$A3,$C2,$FF
	.dc.b	$0E,$4E,$5A,$FF,$7E,$AF,$DC,$3F,$07,$E1,$BE,$49,$1F,$92,$FE,$5A
	.dc.b	$1E,$1F,$C2,$24,$FD,$6F,$91,$FB,$BE,$24,$8F,$F0,$77,$3F,$E1,$23
	.dc.b	$E4,$FC,$8F,$EA,$E5,$F6,$3F,$B5,$7F,$A7,$35,$FE,$E0,$D7,$FC,$46
	.dc.b	$D7,$F8,$37,$D1,$7F,$7A,$7E,$13,$5F,$F2,$AE,$C8,$C7,$17,$F9,$A3
	.dc.b	$47,$C9,$7D,$CF,$92,$FC,$AF,$81,$31,$FC,$8C,$79,$F2,$FF,$53,$47
	.dc.b	$FA,$8D,$AF,$A3,$DA,$FC,$9C,$62,$3F,$6D,$87,$E3,$39,$11,$FE,$67
	.dc.b	$EF,$78,$3F,$A4,$3F,$1B,$DB,$F8,$EE,$AF,$C7,$FC,$8C,$BF,$D1,$C7
	.dc.b	$C9,$FB,$1F,$D0,$C1,$A4,$7E,$C7,$C9,$FE,$60,$8F,$EB,$3E,$0B,$F7
	.dc.b	$1F,$49,$F7,$23,$F5,$FF,$29,$F9,$E3,$3D,$7D,$F8,$FF,$5F,$89,$A4
	.dc.b	$AF,$F0,$37,$53,$07,$4F,$62,$4F,$EF,$3E,$53,$C3,$CD,$FA,$CF,$94
	.dc.b	$93,$F8,$8B,$FF,$5C,$3F,$E3,$8F,$A3,$70,$7F,$00,$6C,$7F,$B5,$5A
	.dc.b	$FF,$85,$DA,$FF,$8D,$A7,$3D,$5F,$E8,$86,$F3,$7D,$4D,$3D,$18,$7E
	.dc.b	$FB,$E1,$7F,$7F,$F2,$93,$1F,$E2,$6F,$F0,$2F,$95,$FC,$D7,$D2,$7C
	.dc.b	$99,$FF,$13,$35,$FF,$24,$35,$FF,$3D,$B5,$FE,$AD,$D6,$F0,$F9,$F2
	.dc.b	$C9,$FE,$43,$1F,$D3,$7E,$43,$F4,$DB,$FF,$B0,$17,$CB,$64,$D7,$AD
	.dc.b	$2F,$FC,$E1,$3C,$38,$3F,$D4,$8D,$8F,$EC,$CF,$F0,$4B,$E9,$E7,$77
	.dc.b	$24,$F2,$F9,$5F,$11,$87,$FD,$0F,$5F,$EA,$D7,$47,$83,$E5,$3D,$11
	.dc.b	$23,$07,$93,$E5,$3D,$93,$23,$07,$A3,$E5,$3E,$04,$C8,$C1,$F8,$5E
	.dc.b	$2F,$89,$E6,$48,$C1,$F0,$3E,$4F,$F3,$B3,$23,$07,$C4,$F9,$4F,$99
	.dc.b	$F2,$72,$30,$7C,$8F,$94,$FA,$04,$7E,$5F,$B9,$FD,$FF,$95,$ED,$52
	.dc.b	$4F,$FA,$3B,$A3,$FC,$F4,$DC,$BF,$AB,$C7,$FB,$96,$3F,$CD,$81,$C3
	.dc.b	$F8,$A8,$DF,$4F,$E3,$99,$1D,$63,$FC,$21,$F5,$3F,$CE,$A3,$FD,$72
	.dc.b	$DE,$57,$37,$FB,$8E,$30,$DC,$B8,$9C,$BE,$0E,$5F,$FA,$DE,$BF,$39
	.dc.b	$7F,$EE,$EF,$B1,$7C,$A7,$83,$CD,$97,$FC,$FA,$72,$F1,$F0,$3F,$C0
	.dc.b	$91,$88,$FE,$FE,$DB,$F2,$C7,$F8,$DF,$0F,$D9,$B7,$C6,$F6,$91,$FF
	.dc.b	$5C,$FC,$EF,$FA,$2D,$BF,$27,$FE,$02,$8F,$F4,$3C,$7C,$C6,$BF,$E9
	.dc.b	$96,$BF,$EB,$B5,$F2,$5F,$D4,$DD,$E1,$E4,$94,$63,$7F,$F7,$9B,$C2
	.dc.b	$3F,$D9,$7F,$62,$F9,$0F,$A7,$C7,$27,$31,$F5,$B8,$FF,$B0,$56,$E7
	.dc.b	$27,$CF,$8F,$F1,$57,$D8,$E6,$3F,$DE,$7F,$59,$FB,$5C,$BF,$A4,$F8
	.dc.b	$88,$C4,$BC,$95,$F1,$D7,$E6,$DE,$FC,$5E,$BA,$4C,$3F,$DE,$98,$FE
	.dc.b	$E5,$FF,$19,$FB,$BF,$E0,$67,$FD,$BA,$D7,$F9,$33,$5F,$03,$6D,$E2
	.dc.b	$FF,$96,$31,$FB,$B9,$AB,$CE,$DF,$7B,$F8,$E7,$C4,$FF,$92,$65,$E4
	.dc.b	$7F,$1F,$F8,$89,$9F,$F8,$2A,$72,$D9,$E0,$47,$F8,$FF,$CD,$73,$F2
	.dc.b	$3F,$5B,$97,$85,$B6,$69,$FE,$72,$D3,$F2,$9C,$5E,$E8,$E5,$87,$F2
	.dc.b	$AE,$EF,$E6,$5F,$0D,$B2,$47,$FA,$CA,$FF,$C9,$6F,$FA,$B5,$FF,$02
	.dc.b	$EF,$FC,$DE,$78,$CF,$F3,$4F,$7B,$FD,$11,$F4,$BF,$E0,$48,$FF,$02
	.dc.b	$D7,$D8,$8F,$F1,$FC,$7F,$A6,$DB,$C4,$78,$42,$78,$3D,$08,$74,$CF
	.dc.b	$F1,$33,$FA,$98,$E1,$18,$EC,$FF,$49,$CF,$E9,$A7,$07,$FB,$39,$C3
	.dc.b	$E3,$EC,$F4,$CC,$8E,$C6,$4A,$7F,$B5,$7C,$CF,$32,$FA,$3F,$F7,$A3
	.dc.b	$FE,$24,$F8,$EF,$A8,$FF,$89,$A3,$2F,$FB,$77,$83,$E3,$46,$25,$5D
	.dc.b	$23,$0F,$6F,$43,$55,$F3,$5B,$7F,$E3,$27,$7A,$5F,$96,$FA,$58,$F4
	.dc.b	$8F,$14,$62,$FC,$2D,$BC,$9E,$9C,$3C,$51,$C9,$B1,$ED,$9E,$2D,$38
	.dc.b	$9F,$16,$AD,$F8,$63,$FB,$ED,$F7,$69,$C4,$A7,$12,$C7,$1C,$7D,$F1
	.dc.b	$F3,$E7,$94,$7C,$08,$C1,$E9,$AF,$0D,$F9,$A3,$E8,$1E,$4C,$3D,$6F
	.dc.b	$7A,$EA,$4B,$F8,$17,$B3,$E6,$4F,$9F,$1F,$19,$E1,$3C,$55,$C6,$3D
	.dc.b	$8E,$2B,$9C,$7B,$30,$FB,$51,$F0,$EB,$93,$A7,$AB,$93,$07,$2D,$BF
	.dc.b	$C4,$F5,$EF,$0E,$1F,$94,$DC,$BE,$93,$98,$FF,$3C,$FD,$27,$F3,$2B
	.dc.b	$FD,$8C,$23,$0F,$ED,$1A,$FE,$F5,$AF,$23,$EC,$8F,$89,$FE,$C7,$3F
	.dc.b	$A0,$8D,$36,$33,$87,$82,$BE,$6D,$7F,$A3,$27,$2F,$84,$EA,$E1,$F1
	.dc.b	$1F,$80,$38,$70,$7E,$49,$F2,$67,$DE,$31,$1E,$D3,$9B,$E0,$7C,$B7
	.dc.b	$FC,$4A,$BB,$47,$28,$C4,$F2,$BE,$D5,$FE,$00,$BF,$75,$F5,$7C,$0F
	.dc.b	$94,$E4,$7F,$25,$8E,$17,$F0,$63,$51,$87,$E0,$4F,$23,$AD,$79,$E7
	.dc.b	$CB,$3E,$BC,$3B,$9A,$79,$E1,$E7,$69,$E3,$3C,$D8,$7B,$1C,$5F,$B2
	.dc.b	$78,$57,$4A,$F3,$36,$BC,$B1,$C7,$0F,$01,$E7,$7F,$9A,$FC,$CF,$F8
	.dc.b	$93,$5F,$E8,$51,$70,$C9,$FC,$BB,$3F,$E8,$51,$9C,$3F,$A3,$3F,$E1
	.dc.b	$97,$F9,$FB,$5C,$CF,$F3,$B8,$C4,$FF,$4B,$F9,$9E,$4D,$9E,$DC,$5F
	.dc.b	$E1,$D6,$CE,$A0,$DF,$98,$C3,$67,$8F,$CC,$EF,$E6,$78,$36,$3A,$C7
	.dc.b	$8A,$CD,$24,$7F,$51,$F9,$9F,$77,$67,$F8,$2F,$D8,$F1,$46,$3B,$51
	.dc.b	$4F,$81,$3B,$AF,$F6,$AC,$E9,$B3,$A8,$7A,$FC,$CF,$77,$C9,$DA,$FF
	.dc.b	$09,$47,$C1,$F9,$9C,$70,$FF,$6A,$3F,$BB,$3F,$17,$E6,$3C,$9E,$C5
	.dc.b	$BF,$F0,$D3,$87,$C6,$F8,$48,$FF,$0C,$E2,$7D,$5B,$8F,$E0,$F8,$AF
	.dc.b	$57,$EB,$CE,$6F,$FB,$8F,$07,$F8,$AA,$76,$6B,$92,$CF,$E2,$69,$1C
	.dc.b	$B4,$FC,$56,$AF,$49,$E2,$67,$0F,$EF,$71,$86,$C7,$8B,$7E,$33,$9A
	.dc.b	$F8,$9B,$7A,$A7,$AF,$D0,$FF,$8F,$B0,$FF,$02,$AF,$7B,$3B,$E9,$93
	.dc.b	$FE,$A6,$9D,$36,$3F,$39,$82,$72,$FF,$89,$30,$F9,$3F,$0F,$F1,$8F
	.dc.b	$0F,$91,$70,$FD,$57,$73,$E5,$BE,$6F,$C0,$FE,$8A,$31,$3F,$33,$4F
	.dc.b	$F7,$B3,$67,$5F,$45,$C2,$39,$BC,$7E,$4B,$97,$C4,$D9,$EE,$E1,$A7
	.dc.b	$8A,$31,$AF,$E2,$CF,$E8,$9C,$9F,$53,$C7,$FD,$2D,$7F,$8B,$1F,$B2
	.dc.b	$DF,$03,$47,$C4,$8D,$D7,$F5,$66,$BF,$CD,$E3,$F8,$0E,$8F,$8E,$F8
	.dc.b	$07,$7F,$06,$E1,$FC,$0B,$E5,$7F,$80,$1F,$C0,$7E,$6E,$C7,$F0,$6E
	.dc.b	$3F,$E0,$9F,$35,$FC,$15,$B1,$FC,$2B,$E4,$FF,$6D,$E9,$FF,$09,$6C
	.dc.b	$7F,$9F,$63,$F8,$67,$C3,$63,$6F,$C4,$F9,$1F,$EE,$EA,$FE,$41,$FA
	.dc.b	$8F,$FB,$71,$BD,$1F,$C6,$23,$15,$F9,$07,$EC,$B6,$35,$CD,$B6,$D7
	.dc.b	$48,$E5,$F0,$B8,$AE,$4F,$5F,$86,$ED,$E5,$5B,$F1,$B8,$69,$E6,$8C
	.dc.b	$57,$8F,$E4,$3C,$C7,$9F,$A6,$D7,$B2,$3D,$5D,$58,$AF,$53,$ED,$F1
	.dc.b	$BF,$6B,$F3,$6C,$F1,$87,$F9,$BA,$30,$D8,$FD,$2C,$62,$FF,$C2,$ED
	.dc.b	$7F,$D1,$2D,$7F,$D9,$BB,$3E,$23,$E6,$1F,$37,$E5,$6E,$1F,$C3,$BE
	.dc.b	$A9,$E9,$7F,$34,$6C,$7D,$6E,$CC,$4F,$F0,$CF,$A4,$75,$7E,$E1,$B5
	.dc.b	$FE,$2A,$CD,$18,$7F,$C5,$AD,$7F,$D4,$DE,$6B,$1F,$EA,$91,$F7,$FE
	.dc.b	$5F,$C5,$F0,$9F,$83,$7C,$9E,$24,$C7,$F1,$3F,$2F,$6F,$F5,$DF,$80
	.dc.b	$B7,$FE,$16,$DA,$F2,$36,$7F,$C7,$ED,$7C,$79,$3C,$3F,$C9,$79,$DF
	.dc.b	$0D,$B7,$F9,$8F,$97,$FC,$B1,$F9,$7F,$95,$B1,$FE,$A6,$EE,$FF,$4D
	.dc.b	$BF,$0D,$B8,$FF,$93,$9A,$FF,$B0,$1A,$FD,$8F,$8D,$FD,$17,$C6,$7E
	.dc.b	$8D,$B5,$FA,$8F,$33,$EB,$78,$FF,$A7,$6C,$78,$7D,$AF,$E5,$78,$FF
	.dc.b	$65,$C7,$C1,$18,$3F,$C0,$FF,$A9,$7F,$9A,$2B,$F0,$5E,$4F,$F8,$22
	.dc.b	$31,$5F,$D8,$F9,$FE,$B4,$FF,$4C,$FA,$57,$2D,$38,$7C,$05,$C7,$FD
	.dc.b	$8F,$F6,$5F,$73,$1F,$F6,$EB,$5F,$F7,$33,$50,$71,$8B,$EB,$3F,$80
	.dc.b	$FC,$0A,$F7,$39,$F2,$FD,$1F,$1B,$FE,$0C,$BD,$D7,$4F,$5D,$BB,$E3
	.dc.b	$FD,$C7,$BF,$D7,$B5,$E8,$F7,$FF,$A9,$EB,$FD,$CF,$6F,$F4,$2B,$5F
	.dc.b	$5C,$7F,$8C,$FE,$81,$E1,$BA,$BF,$C7,$51,$87,$E6,$7B,$98,$9D,$63
	.dc.b	$FE,$EC,$FA,$57,$C6,$7F,$78,$6C,$7F,$91,$76,$7B,$B9,$AF,$EF,$BA
	.dc.b	$FE,$AF,$5F,$97,$E4,$FA,$09,$87,$F8,$47,$07,$C5,$7D,$D3,$CB,$F5
	.dc.b	$FF,$BE,$BF,$EF,$B3,$AA,$F2,$7C,$47,$6E,$DF,$23,$63,$EE,$C7,$C6
	.dc.b	$23,$FC,$07,$24,$7F,$8E,$39,$9D,$FF,$54,$FC,$77,$C6,$5B,$FF,$0F
	.dc.b	$CE,$9F,$81,$E0,$67,$3F,$2B,$F9,$5F,$9B,$93,$F1,$7E,$26,$D7,$AA
	.dc.b	$7F,$35,$F3,$78,$F0,$3F,$11,$3F,$EC,$6D,$D3,$FE,$66,$F9,$A7,$C9
	.dc.b	$53,$1F,$03,$73,$FF,$3B,$4E,$9B,$F1,$39,$C7,$FC,$4B,$3F,$E2,$7C
	.dc.b	$BB,$F8,$28,$FF,$15,$4F,$E7,$8F,$F6,$BB,$87,$F2,$CD,$73,$5F,$B6
	.dc.b	$F9,$BF,$DA,$1F,$B5,$F8,$DB,$8F,$E1,$63,$F7,$1F,$2B,$8B,$FB,$6F
	.dc.b	$F9,$0B,$E9,$BF,$A7,$7C,$5F,$F3,$56,$CF,$FB,$81,$AF,$94,$38,$7D
	.dc.b	$1F,$27,$FE,$74,$F1,$3F,$9F,$36,$3F,$DD,$2D,$79,$35,$F8,$12,$F8
	.dc.b	$5F,$F7,$BC,$78,$5A,$F8,$63,$5B,$79,$E7,$33,$FB,$BE,$37,$CB,$7C
	.dc.b	$BE,$1F,$9F,$1F,$6A,$7F,$51,$C7,$9D,$F8,$5D,$9F,$C1,$71,$DC,$BF
	.dc.b	$82,$CD,$C3,$AD,$7E,$C6,$B1,$C5,$FD,$76,$FF,$85,$B6,$BF,$CD,$B5
	.dc.b	$FA,$F8,$C7,$CB,$70,$8F,$DD,$47,$58,$F0,$7C,$37,$CF,$D3,$D3,$19
	.dc.b	$AF,$F6,$24,$E7,$3F,$E3,$4B,$87,$8F,$D3,$CA,$DC,$7B,$7C,$FF,$E5
	.dc.b	$1F,$11,$DA,$24,$FE,$BA,$3C,$3F,$2F,$E2,$F1,$E4,$F8,$23,$97,$7F
	.dc.b	$FA,$56,$30,$43,$FA,$23,$96,$FB,$DF,$E0,$1C,$7F,$9C,$36,$7F,$BA
	.dc.b	$4F,$E3,$44,$E9,$FF,$53,$66,$F8,$FF,$0D,$C9,$F8,$71,$F2,$3E,$0E
	.dc.b	$47,$C6,$7E,$5F,$8F,$E8,$F8,$2F,$01,$27,$C1,$8F,$F4,$B8,$FB,$58
	.dc.b	$FF,$38,$9F,$9F,$E8,$FD,$43,$EA,$F9,$FF,$0B,$DD,$93,$FC,$59,$FB
	.dc.b	$3F,$05,$F1,$BD,$EF,$71,$27,$BE,$E8,$F8,$5E,$F7,$E0,$3E,$42,$57
	.dc.b	$F2,$08,$F9,$FE,$B4,$8F,$E4,$6F,$E1,$BE,$03,$EE,$7D,$67,$E8,$A7
	.dc.b	$FB,$0B,$5F,$96,$DE,$03,$47,$BB,$F9,$0F,$96,$95,$FE,$CE,$9C,$BF
	.dc.b	$90,$F9,$5F,$CA,$FC,$AC,$8C,$1B,$F7,$3E,$F5,$7F,$AF,$E3,$E6,$B5
	.dc.b	$FE,$15,$8F,$8A,$7E,$7F,$CA,$7E,$83,$E5,$25,$7E,$7A,$3F,$47,$F0
	.dc.b	$DF,$A4,$F7,$64,$FE,$8A,$3E,$0F,$AB,$E0,$24,$78,$1F,$B7,$ED,$7E
	.dc.b	$B3,$E5,$24,$FE,$AA,$3E,$4F,$A9,$FB,$FF,$D4,$FF,$A1,$CF,$FB,$71
	.dc.b	$BE,$57,$F4,$6D,$BE,$4D,$7D,$0F,$57,$DE,$4B,$F7,$BF,$0B,$D0,$E9
	.dc.b	$F0,$B2,$7E,$64,$7D,$8F,$47,$C1,$EE,$C9,$FA,$11,$FC,$4F,$27,$ED
	.dc.b	$88,$FF,$35,$46,$1B,$19,$9F,$D5,$4E,$5F,$F3,$87,$D1,$FF,$9F,$DB
	.dc.b	$F2,$1F,$E7,$35,$19,$7F,$64,$D7,$FA,$A7,$AA,$DF,$FD,$29,$1F,$E3
	.dc.b	$4B,$FE,$B7,$18,$8F,$91,$F2,$7F,$2F,$D8,$FC,$AF,$AB,$F8,$42,$4F
	.dc.b	$E1,$1F,$E5,$14,$F9,$6D,$CF,$E4,$67,$EA,$47,$D5,$F2,$E4,$F8,$7D
	.dc.b	$6F,$F1,$D5,$A7,$ED,$7A,$92,$B4,$34,$F5,$46,$63,$EF,$7C,$B7,$E0
	.dc.b	$7E,$5B,$F0,$5C,$12,$78,$DF,$C1,$9F,$C3,$7A,$9B,$9C,$E5,$FC,$7F
	.dc.b	$83,$E1,$CF,$E2,$FE,$52,$56,$9F,$C7,$7A,$52,$38,$B5,$FC,$87,$C7
	.dc.b	$7E,$4C,$8F,$67,$F2,$DF,$23,$FC,$EF,$D6,$FE,$2D,$3F,$97,$F4,$FF
	.dc.b	$34,$67,$D4,$FC,$E7,$A5,$67,$E1,$F8,$3F,$EB,$C9,$CC,$7A,$BD,$1F
	.dc.b	$D1,$11,$F6,$FA,$3F,$A5,$F9,$1F,$D3,$32,$3E,$0C,$7F,$67,$FD,$57
	.dc.b	$FB,$BF,$0F,$F8,$79,$BD,$3F,$E6,$F6,$D7,$F9,$F3,$E1,$7F,$00,$FD
	.dc.b	$63,$FC,$49,$D5,$F9,$58,$FB,$91,$FE,$C9,$6B,$FD,$1D,$AF,$F8,$55
	.dc.b	$75,$7F,$7B,$8E,$67,$58,$E3,$87,$EF,$9D,$C7,$DE,$BF,$E3,$EA,$7D
	.dc.b	$5F,$23,$ED,$3D,$DE,$A7,$BF,$E4,$6D,$71,$8C,$3F,$3B,$8D,$E9,$FC
	.dc.b	$5F,$91,$93,$98,$F9,$1E,$9F,$DB,$64,$78,$E3,$E6,$7C,$84,$8F,$33
	.dc.b	$ED,$E7,$7C,$67,$BF,$E5,$9E,$B7,$EF,$5B,$7F,$EF,$16,$BF,$DB,$DA
	.dc.b	$FF,$8F,$7E,$FF,$FB,$9B,$F0,$79,$7F,$CA,$93,$C3,$87,$F9,$2B,$63
	.dc.b	$FC,$B1,$5C,$27,$F9,$36,$CF,$57,$F0,$9F,$25,$31,$F8,$91,$F8,$7F
	.dc.b	$90,$FC,$D3,$23,$E9,$47,$E7,$7D,$6F,$C6,$6B,$27,$E6,$47,$E3,$FD
	.dc.b	$7D,$46,$0F,$B4,$FF,$87,$FE,$B7,$FD,$1B,$1F,$DF,$9A,$FF,$52,$6B
	.dc.b	$FE,$71,$F0,$11,$88,$FC,$C7,$C8,$F3,$F5,$24,$7E,$1E,$3F,$41,$F2
	.dc.b	$3F,$A1,$F6,$64,$FE,$2E,$3F,$C6,$1E,$83,$B7,$FC,$44,$D9,$FF,$08
	.dc.b	$B5,$FE,$CA,$D7,$C9,$8F,$E9,$9F,$44,$7F,$8A,$E3,$11,$A8,$FF,$29
	.dc.b	$77,$3E,$47,$E1,$44,$9F,$CF,$47,$ED,$FD,$AD,$92,$3F,$C6,$5E,$82
	.dc.b	$D3,$D3,$D9,$E1,$5F,$A7,$8F,$E0,$11,$FC,$07,$E3,$BF,$81,$7B,$52
	.dc.b	$BF,$59,$1F,$C1,$7E,$37,$51,$83,$F6,$0F,$F0,$9F,$5F,$D2,$48,$F1
	.dc.b	$E2,$8F,$F3,$77,$EA,$DF,$E6,$19,$FE,$79,$39,$7F,$75,$3F,$E7,$99
	.dc.b	$D4,$3E,$45,$D3,$FB,$D8,$7C,$F3,$B7,$FD,$37,$DB,$FE,$B3,$8C,$48
	.dc.b	$E5,$19,$9F,$F5,$0D,$73,$AD,$D7,$78,$ED,$E4,$63,$0F,$F7,$7C,$F3
	.dc.b	$8C,$61,$FE,$C1,$75,$5C,$30,$EB,$DF,$E1,$C7,$78,$59,$F3,$B4,$F3
	.dc.b	$63,$EE,$AC,$47,$8D,$B1,$E2,$8F,$83,$3C,$9A,$72,$BF,$64,$78,$A3
	.dc.b	$FC,$8E,$EF,$0F,$1C,$60,$E1,$7E,$D9,$E1,$86,$88,$E0,$93,$E8,$C7
	.dc.b	$E6,$34,$F3,$E1,$E2,$7A,$9E,$63,$E9,$9E,$C8,$E2,$FB,$CD,$63,$F0
	.dc.b	$DF,$31,$C0,$AF,$C3,$C7,$E8,$E3,$E9,$9E,$F8,$FA,$1F,$21,$D3,$0F
	.dc.b	$23,$DE,$BE,$19,$31,$F9,$0F,$48,$EF,$8F,$DF,$9F,$94,$68,$F9,$DC
	.dc.b	$5F,$D4,$B2,$F4,$37,$0F,$CC,$37,$2F,$CA,$3F,$D6,$30,$FC,$B7,$CF
	.dc.b	$E3,$2F,$F5,$63,$63,$FD,$F4,$D7,$F3,$81,$C3,$C8,$F8,$E7,$FB,$37
	.dc.b	$C2,$F8,$8F,$08,$AE,$4D,$9C,$CE,$3F,$20,$9D,$C6,$6B,$F4,$EF,$16
	.dc.b	$BC,$75,$CD,$71,$8F,$C0,$1E,$58,$CD,$7E,$7A,$3D,$15,$87,$C7,$7E
	.dc.b	$6B,$E5,$3D,$A3,$84,$75,$BC,$57,$68,$C6,$9F,$2A,$31,$87,$77,$CD
	.dc.b	$1F,$29,$F8,$55,$C8,$F1,$16,$35,$5D,$27,$E9,$9F,$1A,$BE,$2D,$F9
	.dc.b	$DB,$1E,$03,$E0,$9E,$03,$EC,$CF,$37,$EA,$1B,$BF,$03,$A9,$F4,$57
	.dc.b	$58,$F8,$A7,$84,$B3,$E8,$BF,$80,$D8,$FF,$35,$6D,$F7,$F3,$FA,$99
	.dc.b	$A1,$F2,$3F,$5E,$73,$1B,$CB,$F2,$EE,$9F,$F3,$6B,$6F,$E5,$5F,$3B
	.dc.b	$FF,$59,$C7,$06,$BE,$26,$C7,$87,$E4,$3C,$17,$DB,$E4,$6D,$FA,$26
	.dc.b	$D7,$9E,$6E,$06,$FC,$8E,$1B,$87,$9B,$4D,$FC,$8F,$06,$C7,$AA,$3C
	.dc.b	$76,$69,$23,$FB,$6C,$3D,$7F,$2F,$89,$FE,$10,$F4,$FC,$71,$8F,$03
	.dc.b	$FC,$B7,$7F,$E7,$86,$ED,$FE,$19,$AF,$87,$7C,$6B,$FC,$3D,$18,$F0
	.dc.b	$3F,$C0,$51,$FE,$0E,$F0,$30,$47,$C7,$F2,$3F,$0B,$E4,$7E,$1B,$63
	.dc.b	$EA,$FC,$87,$8F,$E0,$FE,$3B,$63,$C9,$1E,$2D,$0D,$24,$7C,$0F,$90
	.dc.b	$F6,$7C,$0F,$D0,$6C,$7B,$63,$E9,$FC,$86,$37,$FF,$29,$1F,$90,$7E
	.dc.b	$BF,$C8,$7C,$5F,$81,$B7,$F1,$5C,$3F,$8C,$F6,$6D,$F2,$E9,$FC,$B4
	.dc.b	$7E,$12,$1F,$1E,$DF,$1F,$C0,$3F,$EC,$7C,$3F,$D2,$93,$97,$FD,$2D
	.dc.b	$19,$AF,$28,$70,$FF,$03,$6B,$FC,$B9,$AF,$F8,$2D,$B8,$7E,$A6,$BF
	.dc.b	$26,$79,$47,$0E,$2F,$F4,$FE,$C7,$E1,$7E,$0B,$F3,$2D,$8F,$C3,$47
	.dc.b	$E7,$3D,$7C,$65,$F0,$7A,$BF,$CE,$95,$FB,$0F,$90,$FD,$0F,$C8,$7E
	.dc.b	$89,$B1,$FB,$48,$FD,$2F,$B1,$8B,$E4,$FC,$9F,$9D,$FE,$79,$5F,$EB
	.dc.b	$95,$F2,$39,$8D,$31,$FE,$4D,$5F,$E2,$B9,$DC,$62,$11,$87,$6D,$7F
	.dc.b	$AB,$EF,$FA,$D9,$FD,$DF,$81,$E2,$9F,$A7,$4F,$67,$35,$AC,$37,$16
	.dc.b	$3F,$DF,$B7,$B6,$FC,$7F,$43,$6E,$0D,$FC,$7F,$EE,$3C,$F7,$C1,$F8
	.dc.b	$A3,$0C,$9E,$BE,$15,$D3,$F2,$58,$A3,$D3,$F0,$D2,$BC,$7E,$25,$BD
	.dc.b	$FC,$7F,$B8,$C7,$9B,$FC,$C3,$14,$7C,$2F,$8F,$98,$7B,$22,$DF,$83
	.dc.b	$C4,$F8,$FF,$1D,$F2,$1B,$5F,$0A,$3E,$5F,$89,$8A,$F3,$BF,$3B,$E3
	.dc.b	$A5,$7C,$8F,$8E,$B5,$EC,$F1,$3E,$99,$8D,$7C,$CF,$D6,$F8,$7F,$91
	.dc.b	$F0,$57,$0F,$F3,$65,$7F,$7D,$F3,$7F,$CE,$D3,$A9,$FC,$3C,$61,$BC
	.dc.b	$C9,$FE,$42,$D8,$FF,$4C,$F0,$BE,$3B,$F0,$A7,$E1,$BC,$2B,$3E,$03
	.dc.b	$FC,$43,$F3,$AF,$07,$FA,$6E,$C8,$C4,$FE,$55,$B1,$F9,$A6,$BF,$EC
	.dc.b	$9C,$5F,$03,$E4,$25,$7A,$1F,$3F,$C0,$7E,$63,$E3,$FF,$14,$D9,$F2
	.dc.b	$5B,$6F,$F4,$C7,$C0,$FE,$7B,$E3,$FF,$20,$DA,$DC,$BE,$9F,$FD,$2D
	.dc.b	$1F,$21,$B7,$FE,$15,$9D,$3F,$D2,$23,$2F,$FB,$85,$FE,$EC,$BC,$57
	.dc.b	$EB,$1F,$C7,$F9,$3E,$B5,$6A,$72,$7F,$6C,$97,$C0,$5F,$87,$C3,$72
	.dc.b	$F3,$DB,$DC,$F1,$12,$3D,$0F,$F7,$CB,$7C,$87,$5F,$17,$B3,$6B,$E3
	.dc.b	$E5,$E6,$F9,$0F,$17,$C8,$78,$DB,$3E,$0F,$4B,$E0,$11,$F8,$CF,$7F
	.dc.b	$86,$F5,$7C,$37,$AD,$B3,$E8,$F6,$BD,$DF,$1D,$FD,$E3,$74,$62,$7D
	.dc.b	$BE,$D7,$CD,$23,$F4,$DF,$A1,$EC,$71,$3E,$47,$C7,$59,$F8,$BE,$C7
	.dc.b	$CB,$F8,$EE,$CD,$8F,$93,$E0,$FD,$92,$3F,$7D,$FB,$9E,$C7,$94,$FA
	.dc.b	$9F,$1D,$67,$E3,$F3,$7E,$9F,$AF,$FB,$23,$FE,$06,$6F,$8F,$F6,$FE
	.dc.b	$0B,$DD,$ED,$7E,$1C,$97,$F8,$D7,$EB,$7C,$77,$E0,$4F,$C1,$78,$B6
	.dc.b	$7B,$F8,$9F,$E1,$EE,$AF,$E5,$D9,$78,$23,$1E,$CC,$AF,$DF,$D6,$AB
	.dc.b	$FD,$2B,$E1,$A3,$11,$FE,$F3,$8C,$57,$F8,$8A,$5E,$C7,$E4,$FE,$3B
	.dc.b	$C0,$DC,$3E,$0C,$7E,$5F,$C6,$C6,$3F,$09,$F3,$7B,$1F,$9B,$F8,$DB
	.dc.b	$7F,$21,$C3,$F9,$0F,$D6,$7F,$5B,$E4,$FF,$98,$B9,$15,$FE,$4E,$9C
	.dc.b	$D7,$FA,$D9,$FD,$41,$FE,$5F,$AF,$F3,$4B,$8F,$8B,$E0,$71,$F8,$BB
	.dc.b	$5D,$DC,$3F,$DA,$BC,$65,$27,$FC,$C6,$D9,$D3,$74,$FF,$36,$BF,$E5
	.dc.b	$F8,$F5,$79,$3E,$7F,$8B,$F4,$32,$3D,$DF,$17,$6F,$FC,$55,$68,$FF
	.dc.b	$15,$7C,$CB,$E0,$5F,$1F,$FD,$EE,$30,$D8,$CD,$FF,$80,$A7,$2F,$F8
	.dc.b	$13,$D3,$79,$2F,$9A,$FF,$02,$46,$1B,$19,$9E,$0D,$7F,$D6,$55,$F8
	.dc.b	$2D,$FC,$5C,$FF,$C6,$E7,$C0,$D7,$6D,$7F,$C6,$DD,$7F,$EE,$4E,$3F
	.dc.b	$F6,$84,$2D,$EE,$67,$83,$FD,$D3,$18,$9F,$F2,$2D,$E5,$FF,$1F,$35
	.dc.b	$FF,$23,$37,$E2,$0E,$7A,$9E,$EE,$ED,$EE,$7F,$89,$70,$F1,$36,$33
	.dc.b	$1F,$3E,$35,$F9,$D3,$CA,$E5,$12,$7F,$A6,$89,$D3,$86,$BF,$E5,$C6
	.dc.b	$EB,$F0,$9B,$A7,$E6,$FD,$DF,$77,$07,$E4,$BE,$3E,$F4,$7F,$9A,$F1
	.dc.b	$FD,$37,$C5,$E2,$64,$7E,$A5,$AF,$E3,$3D,$5E,$3F,$1F,$F2,$7C,$6F
	.dc.b	$E0,$1F,$0B,$7C,$6F,$F0,$FD,$FE,$C6,$B6,$D2,$1F,$96,$F8,$EF,$A5
	.dc.b	$D6,$F9,$17,$9F,$FE,$4F,$8C,$36,$33,$5F,$E5,$69,$CB,$FA,$AF,$5B
	.dc.b	$9C,$EA,$3F,$D2,$B3,$A7,$F4,$B3,$F7,$FF,$04,$FF,$3C,$C7,$F9,$75
	.dc.b	$AF,$F9,$89,$A8,$81,$CB,$99,$F5,$79,$DE,$72,$47,$9E,$D1,$FE,$6B
	.dc.b	$E2,$FE,$0B,$F2,$1F,$50,$95,$FC,$67,$D8,$FA,$FE,$C4,$8F,$E3,$8F
	.dc.b	$DA,$F9,$0F,$93,$E3,$FF,$9A,$E7,$E2,$35,$F3,$9F,$B7,$70,$FA,$B3
	.dc.b	$FE,$07,$E8,$F8,$FA,$1E,$C7,$E1,$89,$5F,$CA,$FD,$8F,$C4,$FB,$12
	.dc.b	$3F,$97,$3F,$8C,$F5,$BC,$7F,$15,$F6,$73,$FF,$67,$D7,$92,$30,$7C
	.dc.b	$1F,$D6,$BF,$73,$E1,$3C,$1F,$9D,$C6,$9C,$C6,$1B,$7F,$E2,$17,$FD
	.dc.b	$36,$FF,$83,$9F,$81,$5C,$E3,$FC,$4D,$F8,$43,$3F,$D1,$B6,$38,$6F
	.dc.b	$FC,$07,$A1,$C5,$FB,$B6,$CF,$FA,$85,$AF,$FA,$99,$AF,$FA,$83,$9D
	.dc.b	$E4,$7E,$C1,$FE,$13,$1F,$E2,$26,$BF,$ED,$F7,$6F,$A7,$CF,$FD,$BE
	.dc.b	$9F,$D9,$A3,$D9,$E6,$E2,$7F,$4F,$F6,$4F,$63,$F4,$CD,$AF,$D2,$35
	.dc.b	$FC,$0F,$42,$31,$1F,$EB,$6C,$3F,$D3,$CF,$C9,$F9,$0F,$09,$E2,$BF
	.dc.b	$1F,$C8,$79,$3C,$49,$5F,$3E,$3C,$FF,$21,$F6,$C9,$1F,$D8,$70,$47
	.dc.b	$DF,$E7,$F6,$92,$7E,$C7,$C8,$7C,$0F,$90,$F8,$3E,$A4,$9F,$B9,$1F
	.dc.b	$13,$D2,$FC,$41,$23,$B3,$F8,$BF,$90,$F9,$24,$8F,$C2,$7A,$3F,$CF
	.dc.b	$7E,$6F,$F2,$71,$FE,$DE,$6B,$FA,$66,$BF,$BE,$6B,$FC,$05,$63,$FC
	.dc.b	$FD,$B1,$FE,$E5,$6B,$D6,$3F,$A2,$7C,$8F,$F8,$42,$3F,$9E,$CE,$BB
	.dc.b	$7F,$AC,$B7,$BB,$F8,$6D,$6E,$95,$F9,$BE,$FF,$D2,$4F,$E1,$BE,$46
	.dc.b	$57,$E8,$63,$F1,$24,$70,$7E,$97,$E4,$3F,$1B,$F2,$12,$3F,$50,$FF
	.dc.b	$46,$F8,$AF,$EE,$18,$F7,$6E,$5F,$B4,$6D,$7F,$A4,$5B,$91,$FF,$46
	.dc.b	$C7,$BA,$FE,$81,$1F,$EB,$BF,$A1,$FE,$11,$BC,$9F,$DA,$63,$11,$FC
	.dc.b	$60,$38,$7F,$8D,$B5,$FE,$66,$FD,$CC,$7F,$C0,$6F,$26,$CF,$F7,$D8
	.dc.b	$D6,$3F,$CE,$BB,$1F,$1F,$FE,$01,$6E,$1F,$DC,$9A,$FD,$29,$FE,$83
	.dc.b	$FA,$1F,$F9,$7E,$3F,$BB,$B5,$D1,$FD,$1D,$C6,$3F,$E1,$66,$CF,$2F
	.dc.b	$47,$E8,$7C,$94,$BF,$F5,$3E,$A8,$FA,$9F,$23,$F5,$73,$B5,$A8,$C3
	.dc.b	$E7,$F7,$FF,$C2,$5A,$7F,$0E,$8F,$E8,$2D,$7E,$04,$6B,$C5,$91,$F3
	.dc.b	$CF,$F2,$33,$87,$E9,$7B,$BF,$C5,$9F,$69,$A3,$E3,$47,$F8,$A6,$35
	.dc.b	$B7,$E2,$7E,$47,$F1,$84,$AF,$6F,$B3,$F8,$F2,$38,$3E,$0E,$BF,$E3
	.dc.b	$3F,$C1,$7F,$CB,$11,$FE,$24,$6B,$E2,$6D,$FF,$90,$BE,$F3,$F5,$53
	.dc.b	$FD,$DD,$A9,$A6,$B9,$BF,$31,$F2,$7F,$E4,$1F,$C1,$7F,$CB,$B5,$F0
	.dc.b	$DA,$FE,$4D,$AF,$F6,$C6,$BF,$E0,$45,$F2,$7F,$E4,$36,$C7,$38,$C1
	.dc.b	$FE,$01,$CB,$AE,$3F,$E5,$AA,$DD,$7C,$E6,$C7,$EC,$FE,$1F,$FC,$AD
	.dc.b	$F6,$7F,$85,$AF,$F2,$3B,$5F,$AA,$D9,$FD,$D6,$1F,$BD,$24,$FE,$39
	.dc.b	$FE,$01,$F0,$DF,$C0,$79,$E4,$FE,$49,$FE,$09,$F2,$92,$3F,$2A,$D7
	.dc.b	$F3,$FA,$FF,$01,$F9,$4F,$E1,$5F,$01,$ED,$24,$FF,$A8,$F6,$FF,$10
	.dc.b	$B7,$E5,$36,$FE,$82,$7F,$89,$47,$F1,$3E,$7F,$F2,$77,$D6,$70,$AF
	.dc.b	$D6,$FC,$32,$31,$BF,$F1,$7A,$E1,$5F,$AA,$6B,$E6,$C7,$F2,$BD,$DF
	.dc.b	$E0,$EF,$AF,$FD,$C4,$FF,$8F,$35,$46,$23,$5D,$1F,$E3,$78,$C4,$7F
	.dc.b	$21,$F9,$6F,$F3,$6C,$7E,$F2,$15,$CA,$BF,$96,$57,$F2,$DF,$95,$FC
	.dc.b	$01,$2B,$E2,$3F,$E4,$3F,$D6,$FF,$C6,$FA,$7F,$31,$8E,$13,$FE,$51
	.dc.b	$8F,$E8,$53,$AC,$BF,$C4,$EF,$FA,$E2,$7E,$2E,$BF,$E0,$EA,$FF,$11
	.dc.b	$6D,$F1,$F6,$FF,$41,$DE,$E7,$9F,$C0,$60,$98,$77,$C7,$AF,$CB,$7F
	.dc.b	$9F,$FF,$3A,$FF,$66,$4F,$F3,$5C,$91,$88,$FF,$27,$C6,$3E,$53,$F8
	.dc.b	$63,$63,$FD,$C9,$97,$FB,$8D,$EC,$23,$4F,$DE,$AE,$99,$7C,$0F,$97
	.dc.b	$FF,$5B,$7C,$AB,$C2,$FF,$58,$36,$FF,$DA,$71,$83,$FD,$69,$F8,$C7
	.dc.b	$FA,$D6,$0D,$23,$F4,$7E,$63,$FD,$85,$1F,$36,$3F,$D0,$0D,$E5,$68
	.dc.b	$AF,$1F,$85,$FD,$DA,$DF,$37,$F0,$F2,$FE,$C9,$1F,$E8,$98,$C7,$4F
	.dc.b	$FA,$29,$D5,$71,$6E,$BB,$E0,$FF,$47,$E1,$CB,$0E,$B1,$8C,$39,$3A
	.dc.b	$C3,$93,$6F,$FD,$31,$F4,$3E,$0B,$B3,$C5,$A7,$FA,$77,$05,$FA,$EB
	.dc.b	$C7,$27,$93,$6B,$93,$4E,$53,$C6,$3C,$51,$91,$5E,$BA,$E6,$76,$3B
	.dc.b	$E3,$D8,$95,$C9,$D3,$E0,$8E,$2F,$67,$8B,$E4,$8E,$35,$E4,$C3,$E6
	.dc.b	$34,$F9,$67,$C1,$C7,$E5,$B4,$D9,$E8,$C7,$E6,$BC,$74,$FF,$0F,$B9
	.dc.b	$AF,$F4,$CD,$FC,$09,$CE,$5F,$32,$73,$5F,$EB,$4F,$84,$AF,$FB,$31
	.dc.b	$AF,$FB,$3D,$AF,$FB,$94,$FE,$1B,$7F,$E2,$58,$F2,$C3,$97,$FD,$41
	.dc.b	$3F,$B3,$8E,$47,$B8,$FF,$6D,$46,$B3,$E5,$19,$39,$72,$F4,$7C,$15
	.dc.b	$CE,$B8,$47,$A4,$FB,$87,$02,$9E,$7D,$78,$57,$E3,$A3,$DC,$77,$9E
	.dc.b	$6E,$2B,$6F,$8C,$E5,$7F,$ED,$A9,$FF,$5E,$5F,$B6,$72,$7F,$BD,$23
	.dc.b	$33,$EC,$9E,$66,$F0,$F1,$57,$9E,$30,$FA,$AB,$8B,$ED,$8E,$35,$F1
	.dc.b	$AF,$CA,$70,$2B,$93,$CD,$87,$C9,$C3,$EA,$E1,$F2,$33,$FE,$13,$F3
	.dc.b	$5F,$ED,$5C,$FF,$DA,$D9,$BC,$07,$BF,$9C,$BF,$9E,$DD,$A1,$B9,$C3
	.dc.b	$67,$FC,$75,$5C,$9B,$3D,$A7,$89,$B3,$FC,$81,$CB,$97,$2D,$7B,$B5
	.dc.b	$3B,$C7,$F9,$1E,$9E,$5E,$7F,$E8,$7F,$0F,$E7,$9C,$E3,$FC,$FA,$72
	.dc.b	$DE,$8F,$F7,$7B,$FD,$2D,$B8,$FF,$4A,$6B,$E8,$9E,$2F,$A8,$FE,$DF
	.dc.b	$CC,$F9,$8F,$03,$6F,$FA,$6E,$1F,$CE,$3E,$5F,$35,$27,$C9,$DA,$EF
	.dc.b	$C6,$3F,$DA,$9E,$7F,$31,$D3,$E6,$2D,$F4,$CD,$19,$8F,$EB,$AF,$C5
	.dc.b	$8C,$C6,$0F,$EC,$3F,$31,$F1,$7E,$C9,$C3,$E8,$8C,$32,$7F,$D4,$5B
	.dc.b	$2B,$71,$86,$E7,$F0,$E7,$55,$E1,$F9,$9F,$86,$7C,$4F,$8E,$B8,$7A
	.dc.b	$63,$EC,$7D,$2F,$E3,$23,$07,$D7,$71,$87,$F8,$FB,$37,$CC,$FC,$EF
	.dc.b	$99,$F9,$ED,$BF,$89,$1F,$04,$8F,$F8,$01,$F8,$9F,$33,$F5,$8F,$AF
	.dc.b	$CF,$67,$E6,$7C,$C7,$DA,$31,$8F,$85,$FB,$BF,$31,$27,$E9,$7C,$BD
	.dc.b	$AF,$F4,$AF,$95,$F1,$3C,$0F,$D6,$B6,$BF,$C4,$F7,$FA,$88,$C5,$7E
	.dc.b	$2B,$E6,$3F,$57,$F8,$57,$E0,$EF,$EF,$46,$3A,$64,$7F,$A9,$E7,$55
	.dc.b	$FE,$D7,$6B,$F9,$56,$BF,$B0,$5E,$77,$EB,$5B,$3C,$E3,$07,$EC,$7F
	.dc.b	$4C,$6F,$FB,$E6,$CF,$E7,$3F,$5C,$39,$9F,$E0,$7F,$06,$F9,$9A,$66
	.dc.b	$7F,$C5,$F9,$7F,$B3,$63,$F6,$D6,$FA,$3A,$63,$E6,$ED,$F7,$F8,$4D
	.dc.b	$98,$C3,$FD,$03,$68,$FE,$27,$FA,$4F,$92,$BB,$FE,$16,$F7,$B3,$1F
	.dc.b	$C6,$23,$0E,$A3,$F8,$84,$63,$0F,$DF,$07,$0F,$F0,$16,$BF,$C7,$17
	.dc.b	$BD,$C8,$95,$FE,$DD,$8C,$BF,$AC,$C7,$E3,$3F,$2F,$E6,$FE,$1E,$BF
	.dc.b	$11,$F8,$BF,$37,$27,$47,$D1,$7D,$F1,$23,$E6,$E6,$9C,$65,$FE,$E8
	.dc.b	$FA,$5F,$F2,$45,$7D,$26,$C7,$FB,$B2,$97,$FC,$F2,$73,$D3,$FD,$1A
	.dc.b	$7F,$9E,$56,$FE,$73,$E7,$FA,$1F,$41,$98,$FF,$BE,$E9,$1F,$CE,$DB
	.dc.b	$19,$8F,$E7,$D8,$3C,$AF,$01,$CB,$0F,$43,$F8,$6C,$25,$FE,$46,$ED
	.dc.b	$7E,$0B,$E3,$FF,$16,$63,$2F,$CA,$C7,$FA,$93,$64,$FE,$43,$E6,$E6
	.dc.b	$1F,$9B,$F9,$6B,$7F,$8C,$9F,$CA,$FC,$D6,$31,$E7,$E5,$B2,$4F,$E4
	.dc.b	$30,$F9,$1E,$1D,$CB,$E7,$38,$7F,$17,$93,$74,$FC,$22,$4E,$A1,$E1
	.dc.b	$7C,$62,$46,$5F,$EC,$58,$BE,$6F,$F5,$67,$EB,$3E,$6E,$D7,$ED,$FE
	.dc.b	$6F,$64,$7F,$24,$F2,$F9,$BF,$D9,$FC,$CF,$ED,$1B,$3F,$BF,$F9,$AF
	.dc.b	$DB,$9F,$B8,$F9,$AB,$1F,$C0,$FE,$6B,$C6,$47,$F3,$CF,$F7,$7F,$D7
	.dc.b	$7F,$DA,$B1,$FD,$91,$FF,$4B,$B7,$8F,$FB,$C4,$63,$87,$FD,$FA,$BC
	.dc.b	$4C,$9A,$F9,$8B,$7F,$C3,$3E,$D5,$B7,$F8,$25,$B1,$F4,$1B,$87,$EF
	.dc.b	$1A,$FF,$BC,$70,$7C,$D7,$8B,$E3,$3C,$6D,$BF,$EF,$7F,$0D,$FE,$27
	.dc.b	$6F,$8B,$FE,$0F,$8C,$60,$EB,$FC,$3C,$7F,$00,$8C,$57,$F8,$8B,$27
	.dc.b	$07,$9C,$91,$E7,$7E,$47,$C7,$7F,$86,$FC,$6F,$F2,$F6,$3F,$92,$D5
	.dc.b	$39,$F8,$6F,$93,$F1,$73,$1F,$9D,$F1,$77,$4F,$08,$FA,$0F,$F1,$DF
	.dc.b	$17,$F8,$C5,$B5,$FC,$81,$B3,$FE,$1F,$6B,$F1,$FE,$5B,$EB,$9F,$63
	.dc.b	$E5,$6D,$7D,$DF,$37,$ED,$9F,$73,$8E,$C7,$FB,$C3,$0F,$17,$3F,$85
	.dc.b	$FB,$D1,$E1,$6E,$3E,$16,$BF,$5F,$E2,$BF,$0E,$7E,$23,$E5,$6D,$7E
	.dc.b	$37,$E5,$7F,$16,$7E,$33,$E5,$6C,$78,$79,$1F,$0D,$F0,$5B,$1F,$05
	.dc.b	$B8,$FC,$16,$BF,$E4,$1F,$C7,$7E,$CC,$FE,$F1,$B1,$FD,$21,$AE,$FA
	.dc.b	$3F,$45,$F2,$BF,$E6,$BD,$7F,$0D,$F0,$5F,$A0,$8C,$63,$FE,$1E,$C7
	.dc.b	$CD,$3E,$52,$3E,$57,$F5,$9F,$39,$FA,$83,$F5,$3E,$F5,$9F,$D7,$47
	.dc.b	$F9,$97,$EC,$FF,$9C,$DF,$E2,$63,$1E,$9F,$EC,$3D,$E9,$7F,$E5,$88
	.dc.b	$FF,$18,$D7,$E3,$E7,$3A,$FE,$2B,$C5,$FF,$43,$F6,$E2,$BF,$92,$DF
	.dc.b	$E5,$B0,$FF,$3F,$FE,$81,$FE,$08,$C3,$F2,$21,$C5,$70,$F8,$2F,$F4
	.dc.b	$5C,$62,$B8,$FB,$FF,$E8,$EF,$92,$7C,$56,$A3,$0C,$9F,$F4,$5B,$FE
	.dc.b	$89,$9D,$C6,$0F,$E6,$3F,$50,$EA,$FF,$4C,$B6,$3F,$C2,$B8,$3E,$17
	.dc.b	$35,$27,$FC,$33,$7E,$5D,$3F,$D4,$78,$71,$7D,$B1,$95,$F5,$1F,$E1
	.dc.b	$5D,$FF,$A0,$B6,$3F,$C1,$AD,$7F,$D1,$8D,$7F,$D4,$FF,$30,$F0,$FF
	.dc.b	$71,$1F,$E1,$48,$C7,$4F,$F4,$C6,$C6,$67,$6F,$4F,$82,$72,$78,$9B
	.dc.b	$1E,$26,$BB,$8F,$3F,$AB,$FE,$B8,$E4,$FF,$63,$E5,$FD,$82,$BF,$C7
	.dc.b	$6D,$7E,$EB,$FD,$4D,$CC,$F1,$DB,$FD,$83,$39,$9D,$CE,$5F,$97,$D7
	.dc.b	$5F,$93,$E3,$7F,$68,$69,$F2,$9C,$57,$F4,$08,$F9,$BE,$9F,$18,$C4
	.dc.b	$79,$27,$13,$23,$D2,$D7,$F2,$BF,$41,$FE,$47,$7F,$D7,$3F,$4F,$FE
	.dc.b	$15,$7F,$49,$C5,$E7,$D7,$FD,$CE,$D9,$D4,$FE,$95,$BA,$FF,$82,$1B
	.dc.b	$AF,$FB,$AA,$D1,$8A,$FA,$AB,$E5,$7E,$27,$CB,$7F,$3E,$F8,$2F,$07
	.dc.b	$6B,$E6,$E9,$AF,$43,$FC,$5F,$8F,$FB,$7D,$AF,$F4,$0B,$E4,$E6,$FD
	.dc.b	$1E,$3F,$CF,$FA,$8F,$4C,$FF,$B7,$5A,$FF,$BE,$1A,$83,$8E,$97,$CD
	.dc.b	$FF,$98,$A7,$FC,$FF,$3E,$D7,$7F,$0D,$F7,$EB,$7E,$7F,$B3,$E6,$E5
	.dc.b	$7F,$11,$75,$F8,$57,$C3,$8E,$7F,$33,$F8,$3C,$7F,$DC,$31,$FE,$2C
	.dc.b	$6B,$FE,$78,$6B,$E7,$3F,$B0,$64,$F4,$3C,$4D,$8F,$EA,$8D,$7E,$D7
	.dc.b	$9B,$E7,$D1,$EF,$D2,$4F,$F2,$7F,$98,$F9,$5F,$27,$E5,$24,$79,$5F
	.dc.b	$F2,$0F,$43,$99,$F3,$7F,$70,$95,$F9,$0F,$88,$FB,$9F,$17,$23,$EE
	.dc.b	$3F,$82,$EA,$FB,$1E,$5F,$C3,$24,$FC,$37,$FC,$9F,$CC,$F2,$7F,$6D
	.dc.b	$DF,$FE,$64,$8C,$7C,$5F,$D3,$9C,$F8,$1F,$95,$FA,$7F,$F3,$4E,$3F
	.dc.b	$CF,$FE,$A7,$F9,$23,$F8,$1C,$7F,$CD,$BD,$EF,$C5,$BF,$D8,$F3,$9C
	.dc.b	$3F,$0D,$39,$F4,$BF,$51,$87,$2A,$FE,$87,$F0,$FF,$E2,$27,$FC,$CA
	.dc.b	$D7,$FD,$62,$D7,$ED,$07,$0F,$EA,$FA,$9E,$5F,$EA,$1F,$5C,$30,$FC
	.dc.b	$AE,$DF,$B9,$9D,$CF,$F5,$FE,$AF,$D6,$F7,$48,$C1,$FE,$D8,$C5,$F1
	.dc.b	$1F,$C4,$DB,$1E,$88,$C1,$F7,$BA,$3F,$C1,$B1,$FD,$AF,$C3,$FD,$C9
	.dc.b	$FC,$27,$8B,$FD,$5F,$1F,$B7,$F2,$BF,$7B,$F3,$92,$30,$7F,$BA,$3F
	.dc.b	$04,$FF,$1F,$BF,$E8,$A6,$BF,$ED,$66,$BF,$19,$AF,$D2,$F2,$3E,$67
	.dc.b	$37,$D8,$6F,$2E,$7D,$CF,$27,$67,$F1,$0F,$9F,$FE,$87,$7F,$DF,$3D
	.dc.b	$8F,$E2,$26,$7C,$1F,$E6,$D9,$BE,$17,$F9,$B3,$74,$F3,$46,$0F,$F5
	.dc.b	$0F,$43,$1F,$F3,$E4,$FE,$1C,$FA,$CE,$1F,$E3,$0D,$C7,$FD,$63,$C1
	.dc.b	$FC,$F1,$FC,$AF,$E9,$3F,$6A,$BF,$C1,$AD,$8C,$F5,$7F,$4B,$FA,$BF
	.dc.b	$F6,$0D,$77,$0B,$6F,$E9,$CD,$8F,$F5,$BB,$5F,$C8,$9C,$23,$7F,$27
	.dc.b	$D5,$91,$D6,$53,$F7,$FC,$8F,$E7,$3E,$4C,$AF,$81,$C0,$D3,$F8,$6C
	.dc.b	$61,$AF,$F8,$BD,$B9,$FC,$2B,$FC,$57,$C7,$FF,$44,$F9,$E9,$7F,$E3
	.dc.b	$97,$F1,$DD,$72,$30,$43,$A3,$F1,$79,$7F,$B8,$7D,$5F,$F4,$8C,$7F
	.dc.b	$80,$B7,$9F,$F9,$56,$FF,$88,$C3,$FA,$13,$FA,$27,$FC,$7A,$E5,$F9
	.dc.b	$B7,$F4,$70,$EB,$F2,$D6,$7E,$34,$61,$FA,$9F,$35,$FB,$D6,$FC,$57
	.dc.b	$F8,$0E,$3F,$05,$8F,$F3,$B7,$53,$FE,$4C,$7F,$D7,$2F,$2F,$03,$81
	.dc.b	$C7,$E4,$BF,$B4,$F8,$FF,$DA,$BE,$7E,$57,$F9,$7E,$3F,$BE,$76,$76
	.dc.b	$8F,$F1,$3B,$79,$7F,$3D,$9E,$B1,$FF,$0B,$7E,$89,$F8,$5E,$2F,$E4
	.dc.b	$ED,$D7,$FA,$83,$71,$FE,$D1,$AB,$E3,$3D,$9D,$7F,$E1,$D8,$FC,$D4
	.dc.b	$FF,$1E,$6B,$EF,$8D,$7C,$4F,$08,$F8,$F1,$F1,$63,$FC,$0B,$E8,$C9
	.dc.b	$FF,$3D,$3E,$CF,$95,$91,$82,$1F,$E7,$BD,$D1,$FE,$31,$FA,$B7,$CF
	.dc.b	$5F,$9F,$97,$FA,$99,$CB,$73,$E1,$52,$30,$7F,$A2,$FE,$6F,$FC,$32
	.dc.b	$47,$F6,$5F,$54,$F9,$6F,$F2,$03,$FE,$5F,$8F,$D7,$87,$0F,$EC,$9A
	.dc.b	$FF,$44,$6B,$FD,$FD,$7C,$B7,$4F,$8B,$93,$E0,$8F,$F2,$77,$D5,$3E
	.dc.b	$6E,$F4,$CA,$FE,$0D,$39,$6F,$4F,$0A,$91,$83,$F8,$D7,$97,$FC,$77
	.dc.b	$E1,$3F,$C6,$FE,$FC,$9C,$F9,$BF,$E2,$57,$E8,$4F,$F2,$F2,$3F,$C1
	.dc.b	$39,$75,$D9,$FD,$CB,$EE,$3F,$3B,$3F,$C4,$9A,$FF,$64,$6B,$FE,$1F
	.dc.b	$6B,$F9,$AF,$82,$FF,$36,$47,$29,$E9,$E1,$7E,$82,$7F,$A0,$1F,$E4
	.dc.b	$EF,$7E,$5F,$A2,$32,$D2,$1F,$89,$ED,$FF,$2B,$FE,$00,$F3,$BF,$9F
	.dc.b	$B6,$7F,$94,$EE,$8C,$47,$F7,$C8,$C4,$FF,$89,$BE,$B3,$FA,$EE,$7F
	.dc.b	$E0,$EB,$DB,$67,$FC,$E5,$3A,$F0,$FF,$C7,$DF,$09,$FE,$87,$C4,$D2
	.dc.b	$3F,$E3,$2D,$DF,$2D,$FE,$8F,$9F,$F0,$6E,$3F,$E2,$06,$DF,$F9,$DE
	.dc.b	$B7,$7F,$6B,$4F,$F4,$17,$EB,$DF,$E2,$6C,$17,$FE,$7D,$8C,$3F,$E2
	.dc.b	$2E,$3F,$EA,$8D,$AF,$ED,$C4,$BF,$E8,$7F,$51,$FE,$37,$7F,$17,$F5
	.dc.b	$6F,$8B,$FE,$D2,$D8,$F9,$6D,$7E,$53,$5F,$F3,$07,$0A,$3E,$1C,$74
	.dc.b	$9F,$F4,$6C,$71,$CF,$FC,$76,$EA,$7F,$DE,$FF,$2B,$FC,$BA,$7F,$C8
	.dc.b	$6D,$CB,$FD,$AB,$E5,$FF,$AD,$E3,$F1,$91,$FE,$D8,$F1,$FF,$D7,$11
	.dc.b	$FC,$32,$FF,$DA,$D1,$83,$FD,$79,$F5,$BF,$3A,$3F,$D1,$93,$B7,$C3
	.dc.b	$3F,$EC,$39,$D4,$27,$F0,$0F,$0F,$77,$C4,$78,$F8,$7C,$9E,$23,$3F
	.dc.b	$F4,$C7,$77,$B7,$0F,$F0,$B6,$1F,$C0,$79,$3F,$DD,$F8,$7F,$A1,$EB
	.dc.b	$58,$71,$AD,$1E,$67,$1F,$5B,$CE,$BF,$D1,$CD,$9F,$F5,$33,$CB,$1B
	.dc.b	$87,$FB,$F2,$F8,$BE,$0A,$C9,$F0,$AF,$9C,$6E,$B3,$3F,$06,$B8,$BB
	.dc.b	$33,$8F,$B3,$0F,$04,$1B,$8F,$96,$3D,$67,$CB,$C7,$E0,$63,$DE,$33
	.dc.b	$7F,$CD,$71,$EB,$5E,$49,$F0,$1D,$6B,$86,$1D,$9F,$64,$72,$9E,$AF
	.dc.b	$C9,$AF,$A7,$3D,$F0,$F2,$11,$C1,$E3,$F0,$BF,$99,$57,$C1,$AF,$F7
	.dc.b	$55,$64,$91,$FE,$86,$F8,$5C,$B4,$57,$27,$FD,$B5,$1A,$7F,$D8,$CD
	.dc.b	$72,$7C,$63,$FD,$55,$BF,$A1,$E9,$5F,$A7,$72,$6C,$FF,$71,$9C,$7E
	.dc.b	$27,$F8,$2B,$9A,$E4,$FF,$BA,$E3,$33,$CE,$38,$E7,$D6,$78,$B9,$AF
	.dc.b	$C1,$CF,$82,$30,$F7,$BE,$AF,$AA,$F9,$46,$23,$AE,$5F,$EA,$CC,$3A
	.dc.b	$BF,$E5,$EC,$3A,$C6,$27,$E5,$61,$C9,$FC,$01,$8A,$E5,$7E,$48,$EA
	.dc.b	$7F,$BC,$6B,$93,$F3,$8E,$27,$C1,$8E,$17,$E8,$9F,$87,$5E,$3A,$F6
	.dc.b	$1F,$4F,$E6,$FE,$5E,$7F,$ED,$68,$46,$30,$F1,$C7,$F0,$17,$FD,$93
	.dc.b	$19,$80,$F9,$FF,$F7,$A4,$FF,$A1,$61,$F2,$9B,$75,$F0,$F5,$CC,$EE
	.dc.b	$7C,$51,$D0,$C9,$FE,$1A,$72,$F8,$1A,$FF,$3D,$8F,$07,$3F,$5F,$93
	.dc.b	$EC,$F9,$FA,$7F,$9D,$7A,$F7,$A9,$E9,$D8,$C3,$30,$FE,$8A,$F9,$2B
	.dc.b	$FA,$13,$6B,$8C,$FF,$74,$EE,$7C,$F7,$76,$C7,$F5,$4F,$91,$C4,$C8
	.dc.b	$FF,$4C,$F8,$2F,$57,$1B,$7C,$F3,$F5,$F1,$99,$FE,$BB,$F1,$7F,$2F
	.dc.b	$93,$FC,$81,$18,$3F,$D6,$8E,$2B,$91,$FC,$E7,$B7,$E4,$C6,$63,$FB
	.dc.b	$BC,$60,$FA,$3D,$6F,$AD,$FF,$00,$DF,$F8,$D3,$CB,$7C,$DF,$E0,$6B
	.dc.b	$5D,$5F,$36,$73,$5B,$AF,$91,$18,$C7,$FC,$2C,$D7,$FB,$6B,$5F,$C5
	.dc.b	$F1,$BE,$B7,$F1,$B7,$FE,$44,$6C,$FC,$5C,$5E,$33,$E3,$3E,$F5,$6B
	.dc.b	$6F,$E0,$F8,$F3,$C3,$D1,$C1,$FE,$74,$CB,$A4,$63,$D8,$F6,$64,$F8
	.dc.b	$4F,$CC,$60,$EB,$CC,$61,$98,$7F,$BD,$1F,$CA,$CE,$E3,$07,$FB,$83
	.dc.b	$5F,$3F,$4F,$9B,$B3,$BB,$67,$DD,$1E,$6F,$6B,$1A,$7B,$F5,$47,$EA
	.dc.b	$FC,$0F,$AB,$3F,$79,$75,$67,$97,$4C,$C7,$FD,$F1,$2F,$94,$C5,$F1
	.dc.b	$7F,$6D,$F5,$BF,$98,$8F,$C4,$BF,$E3,$EA,$FF,$80,$0E,$78,$FF,$99
	.dc.b	$CA,$34,$F0,$F3,$BF,$6F,$18,$8F,$DE,$1B,$8C,$9F,$90,$D3,$F0,$D5
	.dc.b	$FC,$FB,$74,$63,$0F,$E0,$9F,$92,$7D,$D7,$FC,$8D,$D3,$FC,$37,$D8
	.dc.b	$D6,$5F,$C5,$17,$D6,$66,$74,$CA,$FE,$33,$97,$AA,$9F,$5B,$9C,$3F
	.dc.b	$53,$18,$AD,$9F,$C8,$33,$7D,$5F,$D4,$8F,$DD,$E3,$FA,$58,$C5,$FD
	.dc.b	$98,$FC,$89,$27,$9C,$7F,$6C,$9D,$E3,$F0,$74,$F4,$D6,$A9,$F5,$2F
	.dc.b	$A8,$FE,$58,$BD,$CB,$18,$65,$7F,$B0,$A7,$4D,$8F,$E5,$71,$8A,$FA
	.dc.b	$A3,$F2,$C7,$93,$FC,$E5,$B7,$E3,$AF,$E8,$0F,$96,$33,$5F,$B5,$CF
	.dc.b	$E4,$37,$7F,$89,$AF,$C5,$7F,$66,$7E,$D3,$CB,$B8,$7F,$24,$7F,$A4
	.dc.b	$7D,$6F,$EA,$EF,$F8,$DB,$FC,$FD,$BD,$DA,$F3,$25,$FC,$7A,$FD,$2F
	.dc.b	$0F,$E5,$9E,$0D,$EB,$FE,$A9,$F5,$CF,$A1,$EF,$F3,$78,$6E,$1C,$E3
	.dc.b	$C7,$E5,$E3,$4F,$1B,$89,$F8,$5F,$01,$67,$F9,$8F,$12,$3F,$09,$F1
	.dc.b	$DF,$C4,$B1,$B7,$FD,$09,$C3,$4F,$C2,$F4,$FE,$3A,$73,$87,$CD,$F9
	.dc.b	$CF,$C7,$1F,$8F,$F9,$7B,$7C,$39,$93,$FC,$82,$F3,$7F,$C2,$DF,$9B
	.dc.b	$E3,$7C,$EF,$8A,$B8,$7C,$97,$0F,$E1,$29,$C6,$9F,$EE,$6D,$9D,$36
	.dc.b	$7F,$17,$FA,$27,$E6,$9F,$EA,$6D,$E4,$FA,$9F,$0F,$FE,$05,$8C,$5F
	.dc.b	$F4,$2F,$3B,$FD,$C4,$FF,$41,$F8,$4F,$D6,$7D,$3F,$F3,$E7,$FC,$19
	.dc.b	$F7,$FF,$E4,$8C,$BF,$86,$B5,$FF,$05,$35,$FF,$0B,$B7,$CE,$FE,$9D
	.dc.b	$F2,$5F,$D1,$38,$FF,$A4,$B7,$1F,$EA,$1C,$1F,$D3,$0F,$DE,$7C,$C5
	.dc.b	$8F,$DA,$61,$FE,$28,$7E,$33,$71,$FD,$5B,$5F,$2F,$9B,$FE,$0B,$D7
	.dc.b	$F0,$95,$FE,$02,$F9,$AF,$EB,$3D,$3F,$D5,$9F,$EA,$FC,$D2,$7F,$07
	.dc.b	$F5,$CF,$2B,$19,$7F,$7E,$CD,$DD,$FE,$26,$6D,$F9,$23,$07,$F2,$FE
	.dc.b	$3F,$EC,$7F,$61,$FD,$91,$B3,$F9,$0F,$83,$FE,$09,$D1,$FD,$AB,$ED
	.dc.b	$3E,$E3,$FC,$D6,$33,$3F,$DE,$27,$2F,$D0,$F0,$7C,$44,$7E,$2B,$FE
	.dc.b	$0E,$F2,$7E,$35,$FF,$76,$F3,$FF,$D0,$2F,$F4,$18,$CD,$7F,$81,$27
	.dc.b	$2F,$F9,$43,$E9,$7F,$CA,$3F,$3B,$FE,$53,$6C,$7C,$6F,$87,$C7,$02
	.dc.b	$BF,$CA,$5F,$61,$FB,$FA,$FC,$E3,$67,$F9,$0B,$5F,$EA,$ED,$7F,$C2
	.dc.b	$BF,$61,$FC,$73,$3F,$F0,$6B,$6B,$FA,$27,$A6,$FA,$CF,$F2,$84,$E6
	.dc.b	$FF,$C7,$7F,$4B,$FE,$12,$3F,$C5,$5C,$FF,$CE,$F0,$FF,$21,$C7,$9B
	.dc.b	$C7,$FF,$2F,$E6,$69,$27,$F9,$BE,$CF,$83,$FC,$24,$7F,$4C,$8C,$74
	.dc.b	$FF,$88,$37,$F8,$D5,$AC,$FF,$C9,$D5,$FE,$FD,$FA,$6F,$95,$C0,$7C
	.dc.b	$BF,$2B,$87,$E4,$FD,$07,$90,$99,$FC,$A7,$FB,$D7,$85,$FE,$2E,$3E
	.dc.b	$A7,$E2,$5F,$92,$CB,$FC,$35,$1D,$BC,$8F,$A7,$F4,$37,$0F,$EB,$DC
	.dc.b	$28,$CC,$7D,$67,$E1,$91,$FE,$AF,$E4,$B9,$7F,$D2,$38,$FF,$9A,$63
	.dc.b	$1E,$B7,$CE,$F9,$FF,$A6,$CB,$FF,$38,$EB,$6F,$FC,$5C,$FF,$97,$67
	.dc.b	$F1,$F1,$FE,$E0,$6F,$A2,$E8,$FF,$0F,$F1,$7F,$85,$E7,$26,$BA,$BD
	.dc.b	$58,$FF,$3A,$8C,$6B,$FE,$A3,$F8,$37,$D3,$D3,$FD,$4D,$F6,$2F,$3B
	.dc.b	$18,$7F,$80,$6B,$6D,$E3,$CE,$DF,$E7,$E9,$FE,$4E,$D8,$FF,$53,$C6
	.dc.b	$23,$FC,$62,$D7,$FD,$5C,$D7,$FD,$59,$5E,$53,$FD,$71,$F6,$0F,$B1
	.dc.b	$C6,$1E,$39,$F5,$7B,$F8,$BF,$13,$FE,$BF,$FB,$1F,$F2,$BF,$CA,$7F
	.dc.b	$49,$6C,$FF,$83,$9A,$FF,$91,$1A,$FF,$B0,$7E,$5D,$F4,$4E,$FF,$EA
	.dc.b	$51,$86,$C6,$67,$FC,$33,$39,$7B,$7C,$33,$9F,$C4,$D8,$CC,$60,$F3
	.dc.b	$7A,$3F,$D7,$E1,$8F,$F6,$A7,$24,$3D,$5E,$77,$FB,$5F,$A3,$FC,$C7
	.dc.b	$3B,$9F,$F5,$8B,$63,$33,$BD,$FF,$DB,$D2,$F3,$FC,$0D,$BF,$91,$18
	.dc.b	$33,$E4,$C8,$F9,$8D,$7E,$67,$93,$D5,$FF,$7A,$D7,$F1,$37,$FC,$94
	.dc.b	$D7,$CB,$1B,$CD,$D5,$FE,$F8,$9F,$CC,$73,$66,$31,$1F,$EB,$1F,$B0
	.dc.b	$FE,$9E,$FF,$85,$9A,$FF,$AD,$63,$7A,$FE,$96,$BB,$56,$A3,$F5,$1F
	.dc.b	$07,$D4,$97,$FE,$64,$F0,$5F,$35,$FE,$35,$8C,$35,$ED,$3C,$1F,$E0
	.dc.b	$11,$9A,$6F,$FD,$1F,$EC,$1F,$0B,$FE,$1B,$E7,$46,$30,$FE,$56,$D7
	.dc.b	$FD,$2F,$1A,$9F,$C2,$7D,$87,$BA,$7F,$D3,$6D,$44,$70,$DF,$5B,$FC
	.dc.b	$55,$C0,$FD,$7F,$F2,$53,$FE,$DB,$6E,$7F,$EA,$16,$E7,$FE,$8F,$E1
	.dc.b	$F2,$4F,$B2,$BF,$49,$A7,$FA,$ED,$F4,$FC,$77,$B6,$31,$1E,$A6,$BE
	.dc.b	$63,$F5,$9F,$38,$F3,$FF,$11,$0D,$5E,$BE,$1B,$1F,$D3,$6B,$83,$FE
	.dc.b	$56,$D1,$EC,$FE,$2D,$B1,$98,$FC,$A7,$0F,$F8,$5B,$CF,$BC,$2E,$8F
	.dc.b	$A8,$DB,$FE,$BA,$D7,$E9,$B5,$F5,$71,$7E,$63,$E4,$3C,$1E,$AC,$9F
	.dc.b	$E6,$11,$E3,$FA,$09,$1E,$77,$F4,$DD,$3F,$8E,$F8,$2F,$1F,$9B,$E2
	.dc.b	$25,$78,$9F,$C4,$78,$DF,$BB,$E9,$FF,$61,$70,$BE,$5F,$E0,$1E,$E9
	.dc.b	$D5,$B1,$7D,$8F,$F3,$D8,$FE,$B2,$D7,$FD,$CB,$1A,$F2,$BF,$CE,$0D
	.dc.b	$8F,$EF,$FF,$93,$7C,$17,$FC,$44,$D7,$F9,$33,$79,$0E,$65,$CD,$FC
	.dc.b	$56,$BF,$05,$18,$F1,$FB,$CE,$67,$F3,$73,$97,$FD,$8B,$C7,$FC,$1B
	.dc.b	$A7,$EF,$7D,$14,$9F,$EB,$D1,$F8,$1F,$07,$F3,$FC,$1F,$BA,$9E,$2D
	.dc.b	$7F,$24,$7F,$35,$70,$FD,$CF,$03,$FC,$24,$FE,$37,$DD,$FC,$61,$FC
	.dc.b	$4F,$DA,$FE,$29,$F3,$5F,$8F,$F7,$25,$7F,$6F,$8F,$C9,$79,$FE,$22
	.dc.b	$47,$89,$F9,$B9,$3D,$8F,$CC,$7D,$44,$AF,$E1,$F1,$FC,$87,$D7,$FC
	.dc.b	$E7,$CA,$C9,$FE,$F9,$1F,$9F,$F9,$5F,$79,$23,$DE,$FD,$AF,$9F,$FF
	.dc.b	$3C,$47,$E9,$FE,$87,$F4,$C7,$E4,$A3,$30,$F3,$3F,$7C,$BE,$DF,$F1
	.dc.b	$CF,$F0,$E8,$CC,$62,$38,$78,$1F,$BD,$8C,$4F,$2F,$B1,$7D,$37,$EF
	.dc.b	$BE,$BF,$FC,$99,$87,$FA,$F5,$AF,$D4,$0E,$1F,$AE,$D7,$F8,$52,$F1
	.dc.b	$7F,$B4,$36,$3A,$46,$0F,$E4,$3C,$EF,$7B,$F8,$37,$BB,$2B,$FC,$4D
	.dc.b	$1F,$D4,$FE,$C5,$E7,$7F,$51,$89,$3F,$EE,$49,$CB,$FE,$2D,$FA,$0F
	.dc.b	$E2,$1E,$94,$8C,$10,$FE,$A9,$1F,$D6,$CE,$DE,$5F,$F2,$0F,$B1,$FE
	.dc.b	$91,$C5,$FC,$95,$B5,$FC,$F1,$AF,$C4,$E8,$F9,$47,$C0,$23,$F2,$DF
	.dc.b	$87,$5A,$8F,$B1,$5F,$22,$7E,$94,$FF,$30,$FA,$1F,$C0,$F1,$FD,$CF
	.dc.b	$0F,$F9,$9F,$29,$D7,$A1,$BF,$3F,$FC,$2D,$66,$93,$3F,$DB,$F1,$BE
	.dc.b	$07,$F9,$A7,$D8,$49,$FF,$2A,$47,$F7,$DF,$3A,$47,$F9,$5D,$FF,$02
	.dc.b	$78,$DE,$4E,$0F,$E0,$7F,$95,$BD,$3F,$E5,$2D,$BF,$EF,$0D,$7F,$26
	.dc.b	$7F,$64,$FB,$2F,$C2,$38,$1A,$B3,$F5,$92,$3F,$24,$69,$F7,$F6,$C8
	.dc.b	$CC,$FE,$71,$B1,$F8,$EF,$97,$FE,$A4,$C8,$FF,$39,$47,$F8,$6C,$8E
	.dc.b	$0F,$AA,$7F,$91,$F3,$FA,$0E,$07,$D9,$3E,$A7,$FA,$F7,$C6,$49,$D4
	.dc.b	$E5,$BF,$51,$FE,$4B,$83,$49,$1F,$C9,$3E,$AB,$FB,$41,$1F,$F4,$47
	.dc.b	$CF,$FF,$74,$FB,$2F,$F5,$F4,$7C,$46,$BF,$B7,$6B,$FC,$B1,$AF,$F7
	.dc.b	$F7,$FC,$D6,$F1,$F3,$BF,$CB,$D1,$88,$E5,$1F,$E6,$1F,$B2,$FF,$08
	.dc.b	$C7,$F3,$5F,$A9,$FE,$F8,$7F,$81,$A3,$30,$7F,$BF,$BF,$C2,$5A,$FF
	.dc.b	$3E,$6B,$FE,$11,$6B,$FE,$AD,$FA,$1F,$F2,$97,$EC,$1F,$D1,$3B,$FF
	.dc.b	$C3,$DE,$C2,$31,$D3,$FE,$23,$AF,$E5,$1E,$7F,$F2,$D2,$61,$FE,$C4
	.dc.b	$F3,$91,$FE,$30,$B7,$BF,$49,$5F,$D7,$23,$2F,$E4,$3E,$CF,$F9,$63
	.dc.b	$FD,$59,$AF,$F8,$A5,$AF,$F9,$01,$B9,$7F,$9A,$BE,$83,$FC,$D0,$49
	.dc.b	$FF,$2E,$63,$FE,$28,$7F,$DA,$CD,$D3,$FD,$95,$3B,$9C,$F9,$5F,$13
	.dc.b	$E8,$BD,$44,$8F,$43,$FE,$06,$FB,$3F,$E5,$91,$FE,$E4,$6B,$FE,$0C
	.dc.b	$6B,$FE,$00,$6B,$FE,$58,$C7,$E7,$7C,$77,$C7,$24,$7C,$57,$FD,$49
	.dc.b	$F5,$5F,$EA,$58,$34,$8F,$F9,$7B,$E6,$3F,$AC,$E7,$F6,$9B,$C5,$DD
	.dc.b	$B3,$FE,$25,$D7,$F7,$4F,$C4,$EF,$FF,$4A,$4E,$B2,$3C,$3E,$05,$7C
	.dc.b	$31,$FC,$3F,$0B,$E3,$FF,$00,$4C,$BE,$F3,$FE,$2C,$FB,$0F,$F3,$F4
	.dc.b	$7F,$88,$9A,$FF,$88,$1A,$FF,$8E,$5A,$BD,$A7,$D2,$7F,$9A,$1B,$1F
	.dc.b	$C1,$DA,$FF,$1F,$F9,$8F,$E6,$CF,$F0,$B7,$F1,$29,$A4,$F8,$BF,$F6
	.dc.b	$93,$FC,$4A,$3F,$DB,$BE,$8E,$08,$FF,$26,$AF,$F6,$43,$FC,$B5,$B9
	.dc.b	$FF,$82,$9B,$C3,$FE,$ED,$F8,$1F,$F3,$FE,$0F,$1F,$B9,$30,$FF,$77
	.dc.b	$BE,$2F,$83,$FF,$0A,$43,$EC,$7F,$02,$FF,$A8,$2B,$93,$FE,$A2,$8C
	.dc.b	$BE,$DA,$FF,$48,$46,$25,$5D,$23,$0D,$E6,$FF,$4B,$4F,$4B,$F0,$46
	.dc.b	$2F,$9B,$AA,$E1,$3E,$36,$BE,$13,$C9,$1C,$5A,$6F,$3A,$FD,$39,$F7
	.dc.b	$FC,$7F,$9E,$FF,$D5,$0D,$9F,$F5,$FB,$E8,$AB,$7E,$37,$9E,$1E,$9C
	.dc.b	$78,$5F,$A8,$F0,$14,$F0,$4F,$27,$E6,$61,$E0,$BF,$44,$7F,$AB,$9E
	.dc.b	$CD,$39,$E1,$D8,$91,$A7,$CD,$0E,$6F,$E6,$F8,$FD,$2C,$3F,$81,$57
	.dc.b	$49,$CE,$5F,$80,$7F,$9C,$FC,$4E,$B0,$FF,$3C,$B5,$FF,$66,$35,$FF
	.dc.b	$71,$B5,$FF,$48,$9E,$73,$FD,$CF,$DF,$FE,$DB,$BF,$C3,$C7,$03,$47
	.dc.b	$F3,$8D,$BE,$23,$D8,$CC,$F1,$CF,$FC,$B1,$7A,$9E,$07,$23,$C3,$AF
	.dc.b	$6A,$FF,$75,$3E,$08,$CD,$7F,$A5,$27,$2F,$48,$EF,$7D,$5F,$5E,$1C
	.dc.b	$E3,$18,$7E,$2A,$7C,$38,$75,$8C,$61,$D2,$7E,$49,$F8,$13,$78,$77
	.dc.b	$AE,$91,$93,$CA,$7F,$BE,$67,$93,$F0,$E3,$67,$07,$FC,$F5,$1F,$E8
	.dc.b	$A7,$D9,$7F,$E2,$C7,$E2,$9F,$EB,$2D,$FF,$C6,$71,$FC,$67,$CD,$3F
	.dc.b	$E2,$19,$76,$E1,$FE,$1D,$37,$6E,$DD,$9D,$E4,$47,$0D,$BF,$14,$62
	.dc.b	$3F,$9C,$47,$97,$9F,$04,$7F,$8F,$7C,$CF,$77,$AC,$FF,$33,$8C,$38
	.dc.b	$F9,$4A,$E6,$7F,$C7,$BD,$8E,$A7,$9F,$E7,$8C,$36,$33,$3F,$EA,$39
	.dc.b	$CB,$FE,$07,$BF,$77,$27,$F8,$55,$BD,$86,$9D,$7B,$5E,$5F,$9A,$BF
	.dc.b	$AD,$F9,$FE,$8C,$7F,$AD,$CE,$7C,$0F,$F2,$76,$5F,$E5,$4B,$DB,$63
	.dc.b	$3A,$FF,$82,$AB,$86,$1F,$03,$87,$F9,$6D,$FF,$9A,$9B,$5F,$EB,$D6
	.dc.b	$BF,$E2,$FE,$AF,$4F,$44,$AF,$F1,$23,$67,$FD,$EF,$9B,$BD,$C5,$F2
	.dc.b	$5B,$7F,$1D,$B5,$C1,$AF,$D8,$D7,$FC,$79,$E9,$FC,$82,$57,$F8,$CD
	.dc.b	$FA,$9F,$4D,$F8,$A8,$F8,$1E,$7F,$C7,$8C,$5F,$FA,$3A,$FF,$C0,$B3
	.dc.b	$F2,$49,$5F,$80,$AD,$EB,$FE,$13,$9F,$C2,$E7,$F9,$0C,$3E,$F7,$37
	.dc.b	$E5,$A3,$11,$F0,$23,$EE,$FC,$AC,$C7,$FC,$90,$D9,$F8,$B8,$BA,$3D
	.dc.b	$24,$AF,$C5,$46,$5F,$F7,$76,$7F,$E1,$CE,$9F,$C2,$73,$BE,$4B,$F0
	.dc.b	$51,$87,$F3,$7D,$7F,$EB,$1B,$FF,$52,$B7,$2F,$F7,$C3,$5F,$D2,$9F
	.dc.b	$B8,$E0,$FD,$8C,$34,$EF,$F1,$DD,$1B,$7F,$E9,$FF,$AF,$7A,$1F,$B8
	.dc.b	$6C,$7F,$AD,$1B,$8F,$60,$E1,$F0,$FD,$27,$D6,$69,$FC,$09,$C3,$FB
	.dc.b	$A6,$BF,$D5,$75,$7C,$7F,$C4,$3F,$49,$CF,$70,$FF,$5A,$7B,$EF,$AF
	.dc.b	$FE,$1A,$D8,$FF,$64,$B7,$0F,$33,$5F,$8A,$BE,$BF,$E8,$FC,$8F,$F0
	.dc.b	$88,$C5,$7E,$C7,$24,$FE,$D7,$09,$3F,$B5,$FA,$CF,$E5,$38,$FE,$4F
	.dc.b	$C1,$F0,$76,$7F,$98,$6F,$EE,$FC,$87,$0A,$FC,$C6,$48,$FE,$2F,$F5
	.dc.b	$1F,$E9,$8C,$BF,$DC,$2D,$AF,$9C,$D7,$EE,$B5,$FE,$46,$FF,$95,$63
	.dc.b	$8F,$91,$FC,$C2,$31,$3C,$A3,$F9,$07,$E9,$BF,$A4,$AF,$E6,$FD,$E8
	.dc.b	$C5,$7F,$3A,$FA,$F7,$C2,$FF,$17,$9C,$F7,$7F,$9D,$38,$FF,$8C,$1F
	.dc.b	$C7,$78,$7E,$4E,$1F,$0E,$3F,$8E,$7C,$37,$E7,$BE,$3A,$E9,$F3,$5F
	.dc.b	$E6,$FF,$5D,$FC,$A6,$3F,$E0,$0B,$6F,$F1,$CD,$7F,$3A,$D7,$FA,$1B
	.dc.b	$FE,$94,$8F,$F7,$C6,$8F,$7B,$AF,$91,$FD,$12,$31,$3F,$EF,$DA,$FD
	.dc.b	$83,$7B,$FC,$7E,$17,$91,$91,$E4,$7F,$AD,$C9,$A7,$F5,$EA,$7B,$9D
	.dc.b	$CE,$39,$3A,$3F,$B5,$29,$DE,$1F,$CF,$7E,$BF,$FA,$11,$F2,$FE,$B2
	.dc.b	$C6,$4F,$CA,$7C,$86,$2F,$F1,$6F,$D1,$F9,$AF,$E9,$55,$6B,$FA,$43
	.dc.b	$87,$F3,$9F,$19,$6F,$F2,$6F,$D2,$F8,$7F,$69,$27,$E9,$BF,$03,$86
	.dc.b	$47,$D6,$6D,$7F,$03,$F8,$0F,$D1,$EB,$EE,$6D,$7F,$78,$FA,$FF,$D2
	.dc.b	$BF,$DE,$A3,$34,$EB,$79,$AE,$6F,$E4,$BD,$9F,$D0,$5B,$87,$ED,$7D
	.dc.b	$8F,$EE,$24,$7F,$DF,$8F,$F7,$6F,$0D,$F2,$DF,$CB,$B7,$FE,$DC,$DA
	.dc.b	$FE,$8D,$5F,$CC,$B7,$FE,$9C,$D8,$FE,$05,$CD,$FD,$F8,$8F,$F3,$AF
	.dc.b	$0F,$EC,$ED,$FC,$C3,$BF,$FD,$AF,$F0,$5F,$DF,$FC,$DF,$F1,$5D,$FF
	.dc.b	$79,$8C,$3F,$DF,$F9,$F5,$EB,$58,$FF,$00,$57,$F7,$EE,$6F,$F1,$86
	.dc.b	$1F,$CE,$7D,$4F,$F0,$8C,$9A,$4A,$FE,$9B,$C2,$FA,$5F,$17,$A7,$FE
	.dc.b	$0D,$6C,$FF,$89,$3E,$B9,$C9,$FE,$23,$6C,$7F,$88,$5B,$BF,$F8,$09
	.dc.b	$AF,$F8,$CB,$E7,$FF,$A6,$7C,$A7,$F9,$05,$B5,$FD,$15,$B3,$FD,$51
	.dc.b	$AF,$F8,$EB,$A9,$E9,$7F,$95,$23,$F8,$9B,$5F,$C2,$36,$37,$EF,$F7
	.dc.b	$97,$D7,$7C,$FF,$4F,$FC,$87,$87,$F9,$4A,$31,$8F,$F9,$53,$8F,$FC
	.dc.b	$B5,$2F,$D8,$78,$CF,$9D,$B8,$79,$DA,$FB,$DA,$FF,$97,$B9,$3F,$B6
	.dc.b	$7A,$5F,$7C,$FF,$1E,$7D,$6F,$F2,$2C,$7E,$CF,$37,$E8,$2F,$FD,$65
	.dc.b	$CA,$FA,$4F,$C3,$79,$FF,$E4,$16,$E9,$E7,$3F,$CD,$7A,$FF,$BA,$34
	.dc.b	$F9,$DE,$FF,$FA,$EE,$7E,$73,$71,$F9,$CD,$7F,$C6,$19,$3D,$0F,$CA
	.dc.b	$1F,$1A,$5E,$BE,$2F,$C1,$6C,$7F,$8C,$7B,$1F,$E6,$CF,$AE,$7C,$D7
	.dc.b	$CE,$E0,$FE,$EF,$87,$0C,$15,$A8,$E7,$83,$EB,$FF,$97,$D7,$E0,$DA
	.dc.b	$FE,$2D,$AF,$C0,$6B,$FE,$D9,$D5,$E8,$7E,$B4,$FF,$34,$7C,$F5,$C3
	.dc.b	$C5,$8B,$B3,$FD,$22,$D8,$FF,$0C,$61,$E6,$9C,$9F,$99,$C7,$FB,$FC
	.dc.b	$FE,$65,$B3,$E7,$F7,$34,$7F,$A2,$79,$BE,$65,$7F,$9D,$76,$FC,$0F
	.dc.b	$17,$F9,$ED,$91,$93,$FD,$27,$3C,$2F,$FC,$11,$3F,$E8,$5A,$FF,$6C
	.dc.b	$EF,$FC,$03,$9B,$D1,$5F,$E5,$F9,$CE,$1F,$91,$9C,$D6,$F3,$FF,$45
	.dc.b	$1C,$FA,$ED,$7D,$5F,$81,$FF,$08,$37,$CC,$FF,$53,$C6,$32,$FE,$AD
	.dc.b	$3F,$EA,$E8,$FE,$64,$EC,$F9,$2E,$23,$FC,$84,$D7,$FC,$A2,$D7,$97
	.dc.b	$A7,$FE,$C2,$83,$49,$3D,$3D,$2F,$1F,$9F,$E4,$6C,$7A,$63,$91,$1F
	.dc.b	$1B,$FE,$CE,$FA,$A7,$8E,$CF,$EF,$C6,$1B,$19,$CB,$FC,$B5,$39,$7F
	.dc.b	$CB,$9F,$18,$F9,$5F,$D1,$46,$1A,$FF,$9B,$A3,$5D,$5F,$8C,$6C,$74
	.dc.b	$F1,$1D,$FE,$26,$C6,$63,$07,$FA,$13,$B1,$EB,$7F,$B7,$BE,$31,$F3
	.dc.b	$9F,$C3,$E7,$FA,$DD,$FF,$BB,$A7,$2F,$E9,$7D,$5E,$15,$BB,$B3,$FA
	.dc.b	$96,$E1,$FE,$83,$6E,$1E,$09,$E9,$5D,$FD,$1C,$FA,$3F,$A8,$C3,$FD
	.dc.b	$14,$DB,$34,$C7,$D6,$7F,$BD,$BA,$BF,$90,$C5,$E5,$F5,$37,$2F,$DE
	.dc.b	$55,$C3,$FD,$2D,$E8,$5C,$53,$4C,$46,$7E,$B2,$5F,$F6,$87,$87,$93
	.dc.b	$23,$FB,$5B,$5F,$F6,$45,$FF,$48,$7F,$B7,$F8,$BD,$C8,$FF,$22,$F3
	.dc.b	$BF,$D0,$DF,$2B,$E7,$6C,$67,$CC,$4E,$5B,$1F,$DE,$BC,$4F,$57,$85
	.dc.b	$F1,$C9,$1F,$7D,$F9,$9F,$90,$FE,$12,$3C,$B1,$C1,$FE,$67,$BA,$75
	.dc.b	$C5,$FE,$04,$5F,$25,$F2,$5B,$5C,$1A,$FF,$B9,$DF,$E7,$EE,$79,$7C
	.dc.b	$24,$7F,$DA,$B1,$F8,$C7,$FC,$38,$FF,$C0,$17,$A7,$CA,$F7,$3C,$1F
	.dc.b	$1D,$E1,$F9,$5F,$17,$CA,$CA,$F1,$47,$97,$D2,$F7,$92,$3B,$BF,$0B
	.dc.b	$C1,$F5,$12,$3D,$5F,$4F,$ED,$FA,$79,$1E,$D7,$E0,$7C,$D7,$C1,$F8
	.dc.b	$CF,$96,$49,$F5,$BF,$3B,$E9,$7E,$37,$34,$9F,$8D,$1F,$27,$E8,$24
	.dc.b	$7C,$97,$E6,$7C,$37,$CD,$FA,$0F,$AE,$49,$F8,$EF,$DA,$F9,$7F,$A5
	.dc.b	$F2,$92,$7E,$94,$7E,$77,$F0,$8F,$F7,$AF,$47,$F8,$F1,$BE,$2F,$F8
	.dc.b	$2D,$B5,$FC,$21,$FE,$2C,$FF,$34,$31,$F2,$DF,$B3,$83,$49,$1F,$EF
	.dc.b	$5E,$B4,$BD,$4F,$11,$27,$F1,$B0,$F9,$8F,$37,$CD,$79,$FE,$4E,$56
	.dc.b	$A3,$D5,$F3,$3E,$12,$47,$7B,$47,$F6,$EF,$11,$F8,$1F,$FB,$DE,$BF
	.dc.b	$58,$D7,$D9,$1A,$F8,$4F,$8D,$E4,$7E,$BE,$BF,$C3,$15,$C3,$DD,$C1
	.dc.b	$F2,$FE,$57,$E6,$7C,$A7,$C3,$24,$F8,$DF,$8D,$F0,$BF,$44,$91,$F0
	.dc.b	$3E,$0B,$EA,$7C,$CC,$8F,$86,$FF,$15,$5F,$51,$E2,$7E,$91,$8F,$87
	.dc.b	$FF,$05,$B4,$FF,$06,$C7,$F8,$59,$B1,$F5,$FE,$0F,$A4,$E6,$79,$B7
	.dc.b	$A7,$F0,$93,$AF,$67,$AC,$FF,$40,$A1,$0A,$E5,$C1,$E8,$8E,$3E,$C7
	.dc.b	$38,$FF,$1B,$4F,$3B,$FF,$69,$A9,$FC,$B7,$C1,$FE,$5F,$E0,$E5,$FE
	.dc.b	$1A,$32,$D2,$1F,$95,$EB,$FC,$7F,$C1,$FE,$7F,$E4,$A4,$FE,$39,$FD
	.dc.b	$17,$C7,$48,$FC,$43,$F9,$B2,$3F,$8B,$F9,$3D,$3D,$7A,$FF,$D2,$EF
	.dc.b	$E6,$23,$F0,$DF,$1B,$FA,$DF,$93,$93,$F8,$E8,$FE,$99,$F1,$DF,$EF
	.dc.b	$0E,$07,$89,$F2,$BB,$BF,$2B,$5C,$AF,$F9,$7D,$EE,$BF,$CF,$CB,$E4
	.dc.b	$B9,$FC,$94,$AD,$C7,$6F,$8A,$91,$C9,$F0,$FC,$3F,$E2,$F8,$1D,$94
	.dc.b	$95,$FE,$69,$A7,$A3,$F9,$15,$E8,$FE,$39,$B3,$FE,$5A,$6B,$F8,$03
	.dc.b	$FC,$43,$1F,$94,$B7,$6F,$38,$93,$FE,$77,$9C,$BF,$03,$C4,$F8,$FE
	.dc.b	$FC,$8C,$10,$FB,$DB,$7E,$E7,$8B,$FD,$AA,$DE,$0F,$8F,$E1,$FE,$A7
	.dc.b	$F1,$E7,$4D,$EC,$98,$FF,$A1,$E7,$2D,$D3,$85,$48,$C1,$F5,$3C,$1F
	.dc.b	$B6,$47,$FD,$21,$F4,$9F,$E7,$F7,$FD,$25,$F4,$9F,$DA,$DF,$D1,$B5
	.dc.b	$FD,$43,$5F,$D5,$B5,$7E,$91,$F7,$DE,$7D,$FF,$C0,$E7,$FB,$8E,$BF
	.dc.b	$B1,$AD,$57,$F0,$4F,$C7,$FF,$75,$3F,$B5,$7F,$84,$70,$7E,$42,$7F
	.dc.b	$73,$18,$9D,$65,$FC,$3E,$3F,$C6,$53,$CE,$30,$67,$E2,$FF,$C2,$5F
	.dc.b	$82,$FF,$75,$C3,$E8,$C7,$E7,$8C,$FC,$1F,$E8,$3C,$7B,$87,$F8,$6E
	.dc.b	$F8,$65,$F9,$C9,$FD,$2F,$8D,$FE,$3D,$E5,$FE,$05,$7F,$D2,$5A,$FF
	.dc.b	$0C,$DF,$FD,$D1,$3B,$F8,$CE,$33,$FA,$88,$FD,$7C,$7E,$C3,$E3,$24
	.dc.b	$FE,$B1,$FD,$9F,$C6,$48,$FD,$7B,$5F,$C6,$FC,$67,$EE,$3E,$33,$C8
	.dc.b	$48,$F2,$3E,$7F,$17,$51,$FB,$88,$FD,$FC,$7F,$00,$F8,$CF,$61,$27
	.dc.b	$D8,$FB,$FE,$33,$51,$FC,$02,$3F,$83,$C7,$F0,$8F,$8C,$FF,$2A,$7C
	.dc.b	$37,$E0,$2B,$FC,$F6,$DD,$9B,$FE,$46,$7F,$91,$4F,$F8,$C3,$97,$F4
	.dc.b	$B7,$CE,$7F,$C7,$73,$9E,$2F,$E7,$78,$23,$FC,$6B,$3F,$3D,$AF,$F3
	.dc.b	$F8,$D7,$CB,$7F,$9C,$E7,$33,$C9,$FB,$AF,$9A,$3D,$27,$AB,$97,$D7
	.dc.b	$F1,$52,$7C,$6F,$BB,$E2,$FD,$FF,$15,$D4,$93,$FE,$0B,$C1,$1E,$1F
	.dc.b	$8B,$DC,$FB,$23,$E3,$C7,$C8,$99,$3F,$01,$F9,$7F,$17,$F3,$3E,$2F
	.dc.b	$D4,$49,$EC,$FB,$7E,$2F,$E8,$F9,$B2,$7E,$4C,$7D,$4F,$2F,$EA,$FC
	.dc.b	$54,$9F,$9B,$1F,$DE,$7E,$3F,$E6,$72,$7F,$68,$FA,$47,$C2,$7F,$69
	.dc.b	$96,$DF,$DA,$E7,$38,$7F,$6E,$9C,$BC,$6B,$86,$CC,$31,$3B,$8C,$3F
	.dc.b	$E7,$49,$ED,$87,$F7,$7F,$8F,$EB,$2C,$3A,$BB,$39,$61,$FD,$FA,$77
	.dc.b	$3E,$A9,$63,$8B,$F4,$61,$DA,$7C,$67,$5C,$7C,$75,$C9,$F2,$E1,$C6
	.dc.b	$7B,$57,$17,$C4,$71,$C7,$2F,$90,$CE,$3E,$8A,$E5,$18,$BE,$B7,$F1
	.dc.b	$AF,$F3,$B5,$6B,$FD,$C6,$DE,$5F,$F4,$AE,$1E,$D7,$FD,$2F,$3F,$ED
	.dc.b	$D3,$67,$F8,$73,$E4,$39,$B4,$C9,$FE,$E7,$29,$C6,$BE,$7A,$F9,$3F
	.dc.b	$E6,$51,$88,$4F,$B9,$FF,$16,$CE,$B1,$FE,$D9,$39,$9E,$5D,$15,$E2
	.dc.b	$7C,$83,$25,$33,$A7,$8A,$FF,$A1,$3E,$1A,$FA,$F3,$C2,$30,$F3,$9F
	.dc.b	$13,$EF,$C3,$95,$F8,$6F,$8D,$74,$C3,$CF,$7F,$86,$7F,$C8,$71,$98
	.dc.b	$FE,$33,$F2,$1F,$D4,$F4,$FE,$A9,$9F,$F1,$1F,$90,$C6,$5F,$C2,$EA
	.dc.b	$DF,$F2,$09,$B5,$FC,$5F,$E4,$2E,$EF,$90,$B1,$87,$3F,$1F,$FD,$87
	.dc.b	$0F,$E3,$54,$8C,$47,$F1,$96,$BF,$CB,$9B,$7F,$C9,$7E,$43,$B9,$E2
	.dc.b	$F9,$6F,$F3,$1E,$9F,$CB,$23,$1B,$67,$E4,$6C,$FF,$BA,$AB,$FD,$4E
	.dc.b	$D8,$EB,$18,$C7,$FA,$F3,$5F,$E6,$0D,$7F,$9B,$BF,$EA,$B9,$E9,$F2
	.dc.b	$3E,$3F,$91,$F2,$3E,$5F,$91,$F3,$71,$DA,$F1,$38,$7F,$D6,$D7,$FC
	.dc.b	$51,$B1,$E4,$3C,$BF,$23,$EF,$F9,$1F,$80,$D9,$F6,$47,$C3,$F9,$0C
	.dc.b	$67,$FE,$73,$C1,$1F,$1F,$E4,$3E,$47,$C2,$5C,$3E,$23,$87,$C6,$FC
	.dc.b	$42,$3E,$6F,$90,$FA,$1F,$09,$F4,$5B,$1F,$36,$3E,$59,$1F,$6B,$F3
	.dc.b	$BE,$43,$EC,$1F,$67,$E3,$EC,$FD,$5F,$83,$FB,$86,$34,$F5,$3F,$7F
	.dc.b	$E0,$7F,$01,$F0,$56,$BE,$E3,$87,$E5,$BF,$70,$8F,$CE,$F8,$FF,$CD
	.dc.b	$FC,$67,$FB,$5B,$93,$F1,$51,$8E,$34,$53,$F1,$DF,$1D,$6F,$F1,$4F
	.dc.b	$E9,$33,$FF,$78,$B7,$A1,$F1,$D6,$30,$CA,$FF,$5C,$F8,$7F,$A3,$FC
	.dc.b	$B7,$FD,$27,$18,$3F,$D7,$97,$C9,$C6,$1F,$EC,$09,$DB,$FE,$D2,$FC
	.dc.b	$09,$C9,$E3,$F9,$A7,$57,$96,$73,$87,$2B,$56,$A3,$0D,$8F,$F4,$53
	.dc.b	$5F,$F6,$5B,$5F,$F6,$83,$57,$80,$EB,$F4,$47,$E4,$75,$F2,$75,$FE
	.dc.b	$DD,$F9,$1E,$B7,$FB,$3F,$1F,$22,$F8,$0C,$FC,$0E,$99,$3A,$F8,$1B
	.dc.b	$8F,$FB,$CA,$D1,$FC,$53,$F4,$4F,$8B,$C7,$FC,$71,$B7,$FE,$2E,$7F
	.dc.b	$D7,$4F,$CC,$D5,$1A,$C3,$D6,$D8,$CF,$C9,$75,$AE,$CF,$E0,$7E,$EC
	.dc.b	$FE,$0F,$E9,$5E,$6E,$63,$0C,$AF,$F5,$B6,$2A,$DC,$61,$B1,$FE,$BB
	.dc.b	$6B,$FE,$FC,$FA,$27,$77,$E4,$5B,$5F,$76,$9E,$76,$6B,$E2,$7C,$9C
	.dc.b	$AF,$81,$AB,$E5,$31,$97,$FB,$E6,$1F,$29,$F2,$FE,$53,$E6,$36,$F8
	.dc.b	$FC,$A7,$E9,$FC,$6F,$A0,$D8,$FC,$BF,$8A,$F1,$15,$F5,$7E,$3F,$59
	.dc.b	$F7,$7E,$C7,$C7,$FD,$9F,$8E,$B8,$F9,$9C,$3F,$ED,$DE,$7F,$AF,$1F
	.dc.b	$D1,$36,$34,$C7,$67,$E0,$CE,$38,$FB,$5F,$C3,$76,$4C,$3E,$1F,$5D
	.dc.b	$AF,$85,$3F,$8B,$F8,$EC,$76,$7E,$2D,$FE,$01,$F4,$7F,$42,$BF,$C8
	.dc.b	$58,$19,$0F,$2B,$7D,$1C,$19,$5F,$82,$9D,$36,$3F,$B0,$D2,$72,$FF
	.dc.b	$AC,$75,$7C,$7F,$8B,$C1,$FD,$03,$70,$FD,$F5,$FE,$CA,$FC,$87,$92
	.dc.b	$BA,$C6,$63,$2D,$7F,$43,$F1,$FF,$AB,$3F,$8F,$7E,$21,$FA,$3C,$51
	.dc.b	$FA,$FF,$90,$98,$FE,$33,$E4,$2E,$1F,$8B,$F9,$0F,$AC,$49,$FB,$58
	.dc.b	$23,$FB,$FF,$53,$F0,$67,$67,$F3,$66,$DF,$E6,$3E,$43,$13,$9E,$A9
	.dc.b	$3F,$BB,$FD,$8B,$F8,$9B,$F9,$2B,$4F,$F8,$62,$5F,$09,$FE,$0C,$6D
	.dc.b	$7E,$35,$BA,$7F,$B7,$DA,$FF,$AC,$57,$95,$CF,$E4,$7A,$36,$BC,$5F
	.dc.b	$23,$DC,$F0,$7C,$8D,$8F,$37,$C8,$F8,$48,$F8,$5F,$27,$C5,$F9,$CF
	.dc.b	$47,$C3,$59,$F7,$7C,$8F,$AF,$E4,$7D,$8D,$8F,$85,$F2,$1E,$D2,$3E
	.dc.b	$D7,$E0,$7C,$8F,$C3,$3E,$27,$C8,$59,$F9,$5F,$23,$F1,$FE,$3B,$E4
	.dc.b	$36,$3E,$77,$AB,$F2,$48,$FC,$97,$E6,$7A,$7F,$3C,$FA,$1F,$1D,$67
	.dc.b	$EB,$7C,$77,$D3,$3E,$A7,$C7,$58,$FE,$69,$E8,$7F,$8F,$3A,$3F,$32
	.dc.b	$7F,$4E,$70,$FF,$B8,$DA,$FF,$BA,$9A,$EF,$E3,$BF,$CA,$F1,$FE,$3D
	.dc.b	$F3,$3F,$CB,$EF,$F9,$D9,$BF,$07,$AF,$8D,$DF,$C7,$5C,$7C,$6F,$2F
	.dc.b	$87,$FF,$29,$E1,$FE,$06,$9F,$E2,$8D,$CB,$CF,$EA,$A3,$11,$E7,$F8
	.dc.b	$EF,$19,$E4,$F6,$2D,$FC,$0F,$85,$F3,$FC,$17,$A1,$B1,$F1,$23,$C2
	.dc.b	$47,$C2,$FE,$AB,$EE,$9C,$3F,$E1,$E6,$CF,$CC,$F8,$FC,$4F,$C4,$F8
	.dc.b	$F9,$3F,$CE,$2B,$79,$FF,$1C,$3E,$84,$FC,$AF,$89,$95,$F5,$BE,$3E
	.dc.b	$D6,$F7,$F8,$1F,$1F,$F4,$0C,$F2,$FD,$07,$E9,$FC,$7F,$D4,$F8,$1B
	.dc.b	$8F,$E0,$1C,$3F,$82,$F8,$FB,$8F,$5E,$44,$7D,$C3,$E9,$92,$7E,$9B
	.dc.b	$F5,$BD,$D4,$7F,$35,$F9,$07,$9B,$F8,$79,$CE,$1F,$EC,$28,$FC,$57
	.dc.b	$BD,$8A,$DC,$FE,$46,$B1,$5C,$6F,$A3,$F9,$1F,$90,$CD,$6E,$72,$FE
	.dc.b	$66,$32,$BE,$81,$F1,$3F,$ED,$58,$FC,$F9,$97,$0F,$E8,$A3,$7E,$07
	.dc.b	$F8,$29,$B1,$C3,$E4,$DE,$5B,$E8,$3F,$C1,$F1,$86,$C6,$67,$83,$5E
	.dc.b	$DF,$17,$FE,$3C,$7C,$77,$FE,$07,$E6,$F0,$4E,$67,$71,$FD,$D7,$BB
	.dc.b	$FB,$9C,$7C,$08,$FF,$29,$FC,$4F,$F5,$2E,$1F,$97,$18,$9D,$57,$F3
	.dc.b	$66,$FB,$1B,$E1,$FE,$D9,$5B,$9F,$AC,$7F,$94,$2B,$BB,$88,$FF,$30
	.dc.b	$D7,$18,$FE,$11,$F0,$9F,$E3,$09,$FD,$A3,$79,$BF,$CE,$F3,$FB,$A8
	.dc.b	$D5,$FF,$94,$E3,$F8,$9C,$7F,$6A,$EA,$FE,$12,$DE,$4F,$2D,$6E,$BF
	.dc.b	$44,$FF,$95,$1F,$F2,$73,$9A,$FF,$77,$B6,$78,$FC,$47,$3A,$FF,$35
	.dc.b	$FE,$C7,$FE,$E7,$BF,$F7,$AB,$6B,$E1,$B6,$7E,$17,$C3,$F1,$CF,$FD
	.dc.b	$02,$DC,$7F,$C0,$F3,$E2,$8C,$7C,$2E,$ED,$F3,$B4,$95,$E8,$F8,$5E
	.dc.b	$44,$70,$7F,$8F,$BE,$35,$EA,$73,$F9,$0F,$9D,$87,$F9,$92,$BF,$C7
	.dc.b	$0D,$7F,$92,$B5,$FF,$0A,$E9,$FE,$98,$F8,$4E,$51,$FE,$75,$9E,$1B
	.dc.b	$FC,$89,$F4,$76,$CA,$D3,$F0,$7E,$87,$FB,$EC,$7F,$9F,$5F,$EE,$EF
	.dc.b	$F9,$25,$AF,$F3,$66,$BE,$26,$BF,$90,$5E,$FF,$5E,$79,$3D,$63,$C1
	.dc.b	$EF,$7C,$BE,$B9,$3E,$18,$F9,$DF,$1F,$F1,$09,$1F,$11,$F8,$FC,$1F
	.dc.b	$4C,$91,$E9,$EE,$FA,$DF,$0D,$23,$E2,$3E,$FF,$7B,$ED,$76,$7D,$02
	.dc.b	$4F,$D0,$7F,$41,$AF,$D6,$8F,$99,$DB,$F8,$13,$E4,$72,$FE,$47,$E1
	.dc.b	$DF,$87,$3C,$FC,$36,$FF,$CF,$AD,$7F,$A0,$61,$FE,$A0,$7E,$C7,$85
	.dc.b	$F8,$22,$3F,$82,$7C,$7E,$8F,$E4,$7E,$42,$4F,$D7,$8F,$CA,$7C,$84
	.dc.b	$8F,$C1,$3F,$9D,$FA,$0F,$F0,$1F,$57,$E9,$3C,$4F,$C7,$92,$BF,$1E
	.dc.b	$FE,$4B,$E4,$7F,$3C,$48,$FF,$60,$53,$DC,$E9,$ED,$7E,$8F,$C5,$95
	.dc.b	$A8,$FD,$37,$8B,$F9,$C2,$47,$E7,$1F,$CF,$FC,$7F,$EA,$C9,$1F,$EA
	.dc.b	$2F,$A2,$FF,$55,$BF,$9B,$F1,$BF,$60,$47,$5D,$5F,$B3,$F8,$FF,$D4
	.dc.b	$92,$3F,$52,$FE,$B3,$C4,$FD,$C1,$23,$D7,$CF,$FC,$07,$F6,$4F,$E6
	.dc.b	$9C,$5F,$D7,$3C,$74,$63,$CB,$FE,$56,$DB,$FF,$7C,$F1,$3D,$9F,$E2
	.dc.b	$50,$FA,$3F,$F3,$9B,$FE,$D8,$6B,$FD,$59,$AF,$F8,$D9,$AF,$FA,$41
	.dc.b	$FF,$11,$3F,$EA,$5C,$7F,$B7,$BD,$75,$EF,$3D,$27,$2F,$09,$FF,$09
	.dc.b	$1E,$3E,$FF,$2F,$DC,$7F,$B7,$A3,$A6,$7F,$E9,$E8,$C4,$70,$AF,$E9
	.dc.b	$8D,$EC,$FF,$78,$BE,$D6,$C7,$F4,$A8,$F5,$47,$F4,$7C,$FF,$DB,$31
	.dc.b	$D8,$6B,$F3,$3E,$47,$E8,$1F,$95,$F6,$BF,$2D,$ED,$DA,$FC,$7C,$61
	.dc.b	$FC,$8F,$B3,$F5,$FE,$42,$46,$08,$7F,$8A,$F1,$FF,$7F,$B8,$DF,$EE
	.dc.b	$47,$FB,$36,$3F,$37,$ED,$FE,$94,$8F,$E7,$BD,$BF,$D4,$7A,$D2,$3F
	.dc.b	$42,$FE,$FF,$A9,$D7,$49,$3F,$99,$7F,$A6,$7C,$F7,$F4,$FE,$03,$FC
	.dc.b	$D3,$93,$FB,$AB,$6B,$E2,$B6,$77,$EB,$72,$24,$FC,$B8,$CB,$67,$FC
	.dc.b	$46,$E7,$A7,$FC,$62,$FF,$93,$5B,$B7,$E7,$5B,$87,$FB,$7B,$83,$F8
	.dc.b	$A4,$FE,$3B,$D4,$7C,$6F,$E0,$1B,$87,$F8,$B8,$FF,$18,$C6,$4D,$7C
	.dc.b	$6F,$E8,$A7,$73,$F9,$26,$D7,$59,$F8,$BE,$8F,$EC,$7E,$46,$5F,$AE
	.dc.b	$3E,$4F,$A7,$A8,$C1,$EF,$7F,$5F,$F4,$9F,$8E,$8F,$CF,$B7,$0F,$D3
	.dc.b	$B7,$4F,$F7,$DB,$FE,$62,$9F,$CA,$07,$0F,$2F,$81,$EA,$76,$D7,$F8
	.dc.b	$37,$B1,$27,$E7,$3F,$7B,$D4,$FE,$03,$F4,$8F,$1B,$F4,$ED,$AF,$90
	.dc.b	$D7,$FC,$AE,$D7,$FD,$6F,$F4,$9F,$C3,$54,$6F,$4F,$E0,$2D,$D1,$32
	.dc.b	$B8,$7D,$1F,$F0,$39,$CC,$A3,$0F,$06,$B9,$C7,$F8,$CC,$7F,$09,$C3
	.dc.b	$F8,$7B,$6B,$9B,$6B,$97,$8B,$F9,$E2,$3F,$93,$7F,$B2,$7D,$1F,$E3
	.dc.b	$A1,$EB,$A3,$0D,$F7,$76,$7E,$A3,$C6,$FE,$69,$F2,$DF,$E5,$1E,$A3
	.dc.b	$49,$87,$FA,$03,$B3,$F5,$BE,$BF,$08,$FD,$14,$FE,$CA,$3F,$9F,$7B
	.dc.b	$B2,$7F,$4C,$FE,$DB,$D6,$91,$FA,$B6,$BF,$EB,$5E,$F7,$AD,$FD,$1F
	.dc.b	$E5,$BC,$84,$9F,$16,$68,$FF,$34,$53,$D4,$E1,$5F,$B5,$9F,$E0,$B1
	.dc.b	$FD,$47,$DE,$F7,$12,$79,$BC,$BD,$5D,$47,$EF,$63,$F8,$6C,$7F,$96
	.dc.b	$6B,$FA,$A6,$DF,$D6,$3E,$5B,$FA,$E1,$97,$F8,$9F,$5D,$C3,$F8,$55
	.dc.b	$70,$CB,$F1,$91,$FC,$DB,$37,$BF,$F4,$49,$1E,$E7,$FD,$09,$5F,$E4
	.dc.b	$4E,$9F,$E1,$91,$FD,$AC,$F2,$79,$DF,$DB,$7C,$79,$7F,$C5,$5F,$47
	.dc.b	$8B,$E9,$F0,$39,$92,$7E,$E6,$3F,$E4,$A6,$F3,$6D,$F4,$CF,$C1,$8F
	.dc.b	$EF,$5E,$34,$9F,$E4,$4F,$C5,$AF,$8D,$C5,$E8,$24,$F6,$7D,$78,$FC
	.dc.b	$BE,$19,$3F,$1A,$3E,$77,$CB,$7C,$FD,$64,$FC,$A8,$FF,$00,$FC,$47
	.dc.b	$E0,$77,$FD,$AF,$D1,$BD,$BF,$D9,$CB,$AB,$F6,$B3,$9C,$3F,$77,$39
	.dc.b	$78,$D7,$0D,$58,$62,$77,$18,$5E,$AF,$35,$7C,$DC,$DF,$8B,$1F,$E0
	.dc.b	$B5,$BA,$F4,$56,$8E,$B8,$77,$79,$D7,$A2,$F9,$46,$6B,$8E,$3C,$F1
	.dc.b	$F2,$63,$D7,$0E,$F8,$70,$C3,$B1,$E6,$AD,$C7,$BF,$1F,$E3,$B7,$C6
	.dc.b	$AC,$F1,$8E,$D3,$89,$F3,$9F,$E0,$E8,$E4,$D3,$3C,$FF,$E9,$E7,$C1
	.dc.b	$3F,$EE,$69,$FE,$4F,$1F,$CB,$3E,$1F,$0E,$1A,$FF,$BC,$CE,$85,$AE
	.dc.b	$8F,$13,$DA,$3E,$29,$E2,$F9,$E3,$95,$FD,$77,$51,$97,$94,$66,$78
	.dc.b	$C7,$41,$7B,$8C,$CF,$39,$C8,$D3,$95,$FD,$B8,$F3,$46,$1D,$65,$F0
	.dc.b	$9E,$35,$93,$A6,$1D,$EF,$CD,$3D,$AF,$91,$A7,$99,$FD,$4B,$0F,$4C
	.dc.b	$7E,$7F,$1F,$E4,$DF,$2F,$EC,$CF,$FC,$E1,$A3,$E5,$F1,$97,$F5,$EB
	.dc.b	$DB,$FC,$62,$33,$97,$F9,$32,$9F,$2F,$8B,$FE,$65,$F2,$F6,$73,$87
	.dc.b	$F2,$08,$D7,$D0,$7F,$6A,$8F,$E9,$4D,$C3,$F0,$76,$8C,$47,$FB,$C1
	.dc.b	$AF,$FA,$13,$8F,$C1,$DA,$7C,$1F,$0D,$87,$BF,$CC,$78,$3E,$62,$E5
	.dc.b	$FE,$3B,$C5,$19,$8F,$1B,$FE,$80,$C7,$E1,$4E,$A7,$27,$F5,$0F,$8B
	.dc.b	$F0,$9E,$CF,$88,$B1,$D7,$9F,$CB,$C7,$F0,$1B,$1F,$E4,$78,$ED,$87
	.dc.b	$F4,$16,$ED,$FE,$A2,$6E,$C7,$39,$3E,$3F,$E5,$FC,$B7,$CC,$6E,$3E
	.dc.b	$D8,$F8,$BC,$98,$CB,$FC,$A8,$FD,$2F,$88,$F9,$3C,$B6,$FF,$CB,$4E
	.dc.b	$1F,$ED,$AF,$88,$8F,$D8,$F8,$8F,$A3,$CB,$F7,$1B,$1F,$E6,$38,$F4
	.dc.b	$91,$FE,$EE,$FA,$BE,$13,$F0,$67,$E1,$3E,$4E,$CF,$F9,$AF,$9F,$EE
	.dc.b	$98,$CF,$FC,$D8,$FE,$2B,$D9,$FC,$87,$DE,$FE,$27,$AF,$FD,$95,$C2
	.dc.b	$9C,$DC,$C3,$FC,$31,$7F,$8A,$8F,$F5,$2B,$79,$0E,$7D,$47,$CC,$3E
	.dc.b	$67,$F4,$31,$87,$F3,$7F,$0F,$89,$D6,$3F,$8D,$F8,$7F,$CF,$99,$AB
	.dc.b	$1F,$E6,$2E,$E7,$C3,$FE,$B3,$44,$62,$75,$79,$8C,$4F,$FB,$3B,$E9
	.dc.b	$3F,$8E,$D7,$F9,$B1,$B1,$FE,$D2,$9D,$F0,$7F,$B6,$A7,$2F,$FB,$73
	.dc.b	$E8,$FF,$15,$7F,$E7,$45,$D5,$E3,$73,$87,$FB,$71,$CC,$EE,$7F,$33
	.dc.b	$29,$CB,$FC,$6D,$AF,$F3,$E6,$BF,$C3,$77,$FE,$62,$FF,$6E,$AC,$78
	.dc.b	$BF,$C1,$B4,$FB,$AF,$C5,$F4,$9F,$23,$4C,$7B,$56,$FF,$D1,$B2,$F6
	.dc.b	$B8,$19,$C7,$FD,$EB,$19,$9F,$F5,$E7,$CF,$7C,$47,$F9,$02,$F9,$7F
	.dc.b	$E3,$8E,$6B,$FD,$D6,$E6,$78,$CF,$E4,$B8,$76,$D8,$FF,$4F,$FD,$0F
	.dc.b	$EE,$5B,$E1,$7D,$F8,$C6,$1F,$04,$FF,$46,$BF,$CC,$E7,$67,$FB,$05
	.dc.b	$C4,$7F,$47,$6B,$FD,$A9,$AF,$CD,$F1,$FF,$4A,$7E,$37,$E8,$BF,$49
	.dc.b	$5F,$ED,$BE,$14,$FB,$BE,$26,$4F,$C4,$F8,$BB,$87,$97,$B5,$B3,$E2
	.dc.b	$FE,$29,$AE,$8F,$EC,$F3,$A8,$FD,$27,$87,$31,$F9,$FF,$27,$71,$E9
	.dc.b	$8B,$E2,$3E,$79,$9C,$B9,$C6,$6F,$EB,$FC,$27,$D4,$3E,$AF,$C3,$D8
	.dc.b	$FB,$7D,$1F,$60,$C6,$3D,$9C,$47,$C1,$F4,$6C,$FB,$78,$BF,$9F,$3E
	.dc.b	$4E,$AF,$C0,$FC,$2F,$E0,$9B,$7F,$88,$8F,$E2,$3E,$16,$2B,$93,$FB
	.dc.b	$BF,$A2,$78,$BF,$A6,$6D,$FF,$BE,$DB,$5F,$E0,$16,$BF,$E2,$45,$E2
	.dc.b	$7F,$05,$FA,$C7,$47,$58,$C3,$DF,$C2,$FE,$15,$F4,$7F,$97,$C1,$87
	.dc.b	$F0,$2C,$7F,$DA,$B3,$93,$F8,$3C,$FF,$8C,$9A,$EA,$3F,$7B,$18,$8F
	.dc.b	$E4,$FE,$A7,$F3,$03,$F4,$BF,$11,$6B,$F1,$9E,$4F,$F3,$43,$F5,$3F
	.dc.b	$23,$63,$F2,$1E,$A7,$F5,$EE,$3F,$E1,$B8,$FF,$BA,$3B,$BF,$98,$CE
	.dc.b	$4D,$79,$5F,$CF,$8F,$E6,$3F,$46,$EE,$C6,$3F,$9B,$FA,$2F,$E8,$66
	.dc.b	$BA,$EC,$7E,$7F,$D6,$F0,$11,$FE,$86,$FF,$70,$FA,$37,$C6,$FF,$4C
	.dc.b	$FA,$2F,$F0,$E4,$62,$BF,$51,$1F,$D4,$3D,$89,$0D,$FF,$4D,$4F,$8B
	.dc.b	$CC,$7E,$37,$D8,$F8,$04,$AF,$EA,$6F,$C3,$F9,$8F,$EB,$67,$83,$E1
	.dc.b	$2C,$F5,$F1,$FF,$5B,$8B,$E2,$3F,$B0,$F9,$5E,$66,$D7,$EF,$7E,$53
	.dc.b	$E7,$11,$FE,$C2,$FF,$84,$3D,$97,$D1,$F9,$E7,$B1,$EB,$8C,$ED,$F0
	.dc.b	$BE,$2B,$FB,$0D,$6F,$7F,$7F,$C1,$FF,$72,$F7,$7E,$3B,$71,$FE,$1D
	.dc.b	$E4,$7D,$F2,$3F,$DC,$9F,$C1,$78,$FF,$DE,$8F,$9F,$EB,$59,$FE,$2D
	.dc.b	$E3,$FF,$7E,$3E,$9F,$C0,$58,$F3,$E4,$F0,$7F,$AA,$C7,$F9,$AD,$B1
	.dc.b	$FE,$4C,$F0,$3F,$35,$18,$8F,$C8,$7A,$5F,$DC,$2B,$9C,$FF,$62,$F9
	.dc.b	$7F,$F2,$E4,$E6,$BF,$91,$5A,$73,$E8,$6B,$CE,$B8,$F2,$C3,$FC,$07
	.dc.b	$87,$F2,$97,$FC,$17,$1E,$56,$C7,$F9,$81,$AF,$93,$C8,$FF,$21,$E3
	.dc.b	$FD,$3E,$3C,$3E,$CF,$93,$CB,$FF,$38,$4E,$7B,$3F,$98,$E1,$FD,$C9
	.dc.b	$FE,$67,$0F,$5B,$D6,$F0,$66,$5F,$1E,$34,$D8,$F0,$9F,$E5,$4A,$78
	.dc.b	$9F,$10,$FC,$1F,$C0,$DA,$F9,$36,$FA,$0F,$29,$9D,$3E,$4B,$F3,$3C
	.dc.b	$39,$5F,$1F,$C5,$B5,$F3,$BD,$1F,$69,$8D,$7D,$6F,$D4,$F0,$BE,$AF
	.dc.b	$83,$70,$FA,$0E,$1F,$1F,$B1,$70,$FB,$2F,$C8,$3E,$69,$23,$E1,$3E
	.dc.b	$FE,$EF,$F6,$0F,$C4,$FF,$04,$AF,$C7,$70,$FF,$34,$BF,$F6,$C7,$93
	.dc.b	$FE,$A7,$8F,$ED,$B3,$FE,$C6,$8C,$77,$72,$6D,$70,$9E,$73,$97,$AE
	.dc.b	$17,$AF,$FB,$24,$E6,$5F,$45,$F6,$FE,$07,$FD,$91,$18,$E7,$FF,$70
	.dc.b	$3F,$EC,$B9,$D3,$63,$2F,$FB,$C3,$E3,$1F,$1C,$FA,$3E,$71,$86,$C6
	.dc.b	$67,$FD,$AD,$39,$7B,$73,$FE,$C5,$B9,$F9,$31,$FC,$13,$67,$53,$FC
	.dc.b	$FF,$A3,$F9,$D3,$F0,$E7,$FC,$A3,$E8,$FF,$2B,$8F,$94,$7E,$F6,$76
	.dc.b	$E2,$7F,$CA,$FC,$3B,$E2,$FA,$D1,$89,$E9,$5C,$27,$2D,$E2,$FF,$6F
	.dc.b	$3F,$E1,$19,$DF,$A1,$F3,$2F,$FC,$C6,$D9,$34,$C7,$BB,$E7,$E9,$FD
	.dc.b	$7C,$5E,$5E,$AD,$C3,$D5,$57,$0F,$F3,$7F,$BB,$71,$4D,$31,$19,$F7
	.dc.b	$75,$F0,$93,$0F,$D7,$C7,$1F,$7B,$51,$83,$F6,$6F,$87,$D0,$7D,$CB
	.dc.b	$D7,$F0,$7D,$0C,$BF,$E5,$75,$A9,$F2,$7B,$BC,$6F,$85,$69,$AF,$F2
	.dc.b	$48,$F5,$7B,$BF,$1C,$91,$FC,$35,$F9,$5F,$2F,$A8,$FE,$0F,$1F,$06
	.dc.b	$3E,$17,$D0,$C9,$FE,$16,$FC,$5F,$98,$F9,$1F,$07,$F8,$A9,$FF,$04
	.dc.b	$35,$FE,$E5,$C6,$8C,$47,$F4,$56,$C6,$7E,$4F,$9F,$C5,$4A,$FE,$83
	.dc.b	$49,$EF,$D3,$27,$2F,$8B,$DB,$F1,$FC,$BF,$E1,$EC,$D2,$57,$F5,$38
	.dc.b	$7C,$BF,$AB,$E0,$24,$F8,$23,$DB,$F2,$9F,$3F,$DA,$93,$E4,$8F,$83
	.dc.b	$E3,$77,$24,$75,$7C,$5F,$01,$F1,$89,$1E,$CF,$96,$F9,$3F,$03,$23
	.dc.b	$DE,$FC,$CF,$97,$F9,$BF,$2D,$EB,$24,$FA,$5F,$77,$C1,$7D,$22,$47
	.dc.b	$C8,$F9,$5F,$AB,$F0,$5F,$5B,$E5,$64,$FC,$C8,$FB,$3F,$07,$F1,$C9
	.dc.b	$1F,$15,$F9,$5E,$E7,$DE,$24,$7F,$74,$FC,$73,$F8,$23,$FE,$93,$7F
	.dc.b	$8D,$F9,$9F,$E9,$07,$FD,$D3,$1B,$F9,$3F,$D7,$46,$BC,$DF,$C2,$BF
	.dc.b	$EA,$09,$FF,$03,$35,$EB,$DA,$8C,$47,$F9,$55,$AF,$2F,$84,$FD,$67
	.dc.b	$43,$BE,$9E,$08,$F0,$FC,$27,$EC,$7F,$03,$FC,$65,$FE,$B9,$AF,$AE
	.dc.b	$3C,$FF,$07,$FC,$47,$13,$49,$3E,$8F,$85,$F6,$91,$F8,$DF,$1B,$FC
	.dc.b	$3D,$7D,$CB,$BB,$0D,$8F,$F1,$49,$A8,$E1,$F4,$7F,$E8,$E8,$CB,$FE
	.dc.b	$BF,$7F,$A7,$B5,$FA,$61,$C3,$F5,$9A,$FF,$3B,$FB,$25,$F6,$F8,$29
	.dc.b	$3E,$29,$CB,$F7,$FD,$6F,$B3,$E7,$C8,$C1,$F8,$3F,$88,$FB,$B3,$23
	.dc.b	$07,$A3,$D7,$FE,$CA,$F9,$5B,$97,$F2,$E8,$F5,$C6,$3D,$7F,$82,$C8
	.dc.b	$F8,$31,$F8,$AF,$87,$FC,$5F,$C7,$C9,$FC,$A4,$65,$A4,$3F,$2D,$F1
	.dc.b	$3F,$92,$23,$F9,$B7,$F2,$BF,$1F,$F4,$3D,$49,$3F,$41,$FA,$7E,$A4
	.dc.b	$8F,$A6,$D7,$F4,$5F,$13,$F9,$EF,$89,$91,$FA,$77,$F4,$3F,$13,$23
	.dc.b	$F5,$4D,$7F,$55,$F1,$BF,$87,$FB,$A7,$97,$49,$3F,$86,$F8,$97,$85
	.dc.b	$F8,$D5,$E9,$7C,$E6,$CF,$F8,$51,$AF,$E1,$9A,$FE,$EA,$DE,$6F,$3A
	.dc.b	$93,$F9,$49,$CB,$E1,$F2,$FB,$11,$EC,$F8,$7C,$CF,$23,$23,$D1,$1F
	.dc.b	$A3,$FA,$3F,$EA,$FE,$5B,$CA,$FF,$07,$46,$1B,$DB,$BE,$D9,$87,$E9
	.dc.b	$AB,$FC,$C2,$D8,$FD,$77,$27,$C8,$8F,$D5,$FE,$84,$F4,$BF,$64,$D8
	.dc.b	$CC,$7E,$D7,$E9,$1D,$B7,$E0,$FF,$2F,$06,$92,$BF,$C7,$FC,$DF,$C5
	.dc.b	$F6,$F5,$47,$FB,$46,$B8,$3F,$BC,$6A,$8D,$A9,$E5,$B7,$F8,$BF,$CC
	.dc.b	$75,$7F,$09,$6D,$7F,$06,$6B,$CE,$78,$75,$6C,$91,$B7,$F9,$8F,$C7
	.dc.b	$7F,$13,$FC,$7B,$F8,$46,$5E,$26,$DF,$F8,$FD,$B7,$FE,$EB,$7F,$C7
	.dc.b	$F3,$BF,$8A,$E4,$49,$FE,$89,$E4,$FF,$8A,$FE,$FF,$DB,$1F,$E4,$9F
	.dc.b	$49,$18,$8F,$F0,$2B,$5F,$F6,$43,$63,$7F,$27,$CF,$29,$5F,$EC,$4A
	.dc.b	$FF,$60,$37,$E4,$FC,$04,$75,$DF,$E3,$23,$83,$97,$C7,$79,$FE,$4F
	.dc.b	$FC,$9D,$7F,$CB,$E7,$FA,$8B,$5F,$53,$5F,$4F,$9D,$C5,$F2,$4F,$C1
	.dc.b	$8F,$85,$AF,$42,$4F,$FB,$8E,$BF,$CD,$2F,$4F,$8F,$DB,$ED,$9F,$95
	.dc.b	$1F,$2F,$E3,$BE,$67,$9B,$2B,$E1,$46,$5A,$43,$F0,$87,$CC,$EA,$FA
	.dc.b	$9D,$5E,$E2,$47,$8D,$F8,$3F,$1D,$A8,$F9,$D1,$F6,$E3,$EE,$7C,$74
	.dc.b	$9F,$A4,$FD,$FF,$8E,$FC,$07,$C6,$7C,$A2,$4F,$BD,$F9,$BE,$6E,$A3
	.dc.b	$ED,$47,$E1,$E3,$F1,$1E,$14,$9F,$BC,$FE,$2F,$E3,$3F,$19,$EE,$FD
	.dc.b	$62,$4F,$CB,$7F,$D0,$DF,$0B,$FE,$9C,$EC,$FD,$9F,$5F,$FA,$B3,$E2
	.dc.b	$FF,$CF,$D3,$FD,$FA,$72,$EF,$AF,$FC,$AE,$BE,$33,$64,$C4,$7C,$CC
	.dc.b	$3F,$D6,$FF,$19,$FD,$F7,$E4,$7F,$98,$E3,$AF,$A2,$FF,$03,$DF,$09
	.dc.b	$FE,$6E,$E7,$E0,$3F,$D5,$91,$88,$4F,$05,$1D,$1C,$CF,$0F,$4B,$F8
	.dc.b	$54,$7F,$99,$63,$9F,$A9,$FE,$B7,$DB,$95,$F5,$6D,$F0,$6E,$3E,$3C
	.dc.b	$3C,$D7,$FE,$BF,$C7,$8B,$71,$F0,$56,$EB,$CB,$87,$39,$F0,$C7,$C6
	.dc.b	$C3,$8B,$63,$8B,$E7,$3E,$24,$F1,$7C,$26,$CA,$6C,$EF,$59,$68,$CF
	.dc.b	$FC,$73,$97,$C7,$6D,$7C,$66,$D7,$FA,$53,$0F,$F1,$3F,$C8,$DC,$3F
	.dc.b	$CB,$6D,$7F,$DB,$47,$58,$DB,$4D,$9E,$51,$F2,$9F,$D0,$A3,$ED,$47
	.dc.b	$E2,$27,$D0,$EB,$82,$A8,$CD,$71,$CF,$BD,$F2,$8C,$57,$83,$5C,$61
	.dc.b	$F4,$E3,$B1,$F8,$63,$FD,$F1,$3F,$D5,$27,$C7,$1D,$CF,$44,$74,$7D
	.dc.b	$F5,$D1,$F6,$1B,$9C,$1C,$B0,$F0,$57,$5C,$3F,$53,$F2,$1F,$D8,$B2
	.dc.b	$FD,$6B,$72,$FD,$6F,$C8,$5C,$BE,$C7,$12,$B1,$AA,$3F,$D6,$B7,$FB
	.dc.b	$0C,$3F,$6F,$C1,$A9,$CE,$1F,$B3,$7F,$C4,$75,$B3,$F7,$7B,$FF,$4F
	.dc.b	$C3,$A3,$67,$FC,$4F,$3F,$E6,$89,$CB,$63,$C4,$F3,$AF,$3C,$78,$7E
	.dc.b	$3B,$C4,$F8,$FD,$1F,$27,$C7,$5B,$E1,$87,$AE,$33,$1D,$5F,$11,$1F
	.dc.b	$07,$C7,$7B,$4F,$77,$B9,$63,$9E,$1F,$E3,$66,$FC,$77,$C3,$F7,$7E
	.dc.b	$23,$23,$D4,$FB,$C8,$FB,$7D,$CF,$94,$7C,$BF,$72,$C7,$68,$F9,$DF
	.dc.b	$1D,$8D,$FE,$73,$F4,$7E,$3B,$EB,$FD,$4F,$F2,$4B,$F2,$C6,$2A,$47
	.dc.b	$CD,$9D,$47,$FA,$F1,$AF,$F8,$15,$AF,$F2,$ED,$D1,$FA,$78,$FE,$60
	.dc.b	$BE,$3F,$C3,$D3,$E2,$65,$F7,$7F,$01,$3B,$8C,$1E,$3F,$9D,$F4,$9E
	.dc.b	$AF,$6A,$C7,$D8,$97,$C7,$FB,$8C,$E3,$E2,$7E,$0F,$67,$C2,$F8,$EB
	.dc.b	$7C,$5C,$3F,$05,$F1,$91,$FF,$5B,$D3,$F2,$DF,$F3,$14,$7F,$B0,$9B
	.dc.b	$1F,$03,$E3,$BF,$50,$EE,$6C,$BD,$AA,$62,$29,$FE,$CC,$FA,$2F,$F2
	.dc.b	$D3,$7D,$0F,$B3,$F1,$DD,$27,$38,$7F,$82,$02,$F4,$74,$6E,$AD,$FF
	.dc.b	$B6,$3E,$BD,$F1,$9E,$6F,$A3,$7C,$1F,$A2,$73,$5F,$ED,$E9,$CB,$67
	.dc.b	$FC,$85,$18,$CB,$FD,$B6,$D7,$FB,$89,$FD,$C9,$C3,$8F,$A2,$EA,$FA
	.dc.b	$5B,$E3,$FF,$9B,$A5,$F1,$BF,$2A,$73,$87,$4F,$41,$E9,$FC,$E6,$C7
	.dc.b	$B7,$E1,$F1,$39,$F5,$24,$FF,$BE,$2B,$79,$FC,$95,$3F,$63,$E8,$FF
	.dc.b	$D2,$55,$FE,$FE,$5E,$4E,$7C,$AD,$32,$7F,$0B,$3A,$6C,$7F,$99,$E3
	.dc.b	$18,$7F,$C0,$0F,$AF,$7C,$1F,$E1,$BE,$8F,$F5,$FC,$3F,$E8,$E9,$CE
	.dc.b	$3F,$C6,$ED,$5A,$8F,$9A,$FE,$CA,$3F,$C3,$ED,$76,$7F,$9D,$DC,$6F
	.dc.b	$FE,$99,$9E,$8D,$9F,$C5,$4F,$EE,$6F,$FD,$3F,$39,$7F,$C6,$BD,$EF
	.dc.b	$8F,$FE,$16,$7F,$0C,$F8,$EB,$8F,$EC,$FE,$4B,$F8,$79,$8C,$BF,$81
	.dc.b	$BE,$F9,$93,$FB,$9F,$93,$B5,$E2,$AF,$88,$63,$1F,$F5,$8E,$68,$FE
	.dc.b	$37,$D3,$30,$FE,$05,$F2,$77,$0F,$E2,$3F,$1F,$FC,$80,$C6,$3E,$A7
	.dc.b	$15,$FC,$27,$AA,$D7,$F1,$87,$EA,$74,$FD,$5E,$9B,$5F,$4F,$4F,$88
	.dc.b	$FD,$28,$F9,$C4,$7F,$83,$C7,$CD,$BF,$E5,$15,$FC,$CA,$BF,$99,$B6
	.dc.b	$BF,$8D,$47,$F3,$6F,$8F,$C5,$7F,$2A,$7F,$85,$FE,$24,$F8,$0F,$E1
	.dc.b	$BF,$49,$FC,$2F,$0F,$C3,$4E,$5F,$C4,$46,$6F,$F8,$33,$5F,$DA,$87
	.dc.b	$0F,$EE,$1A,$FF,$9A,$9A,$B9,$5E,$47,$ED,$4F,$1F,$C7,$DB,$FE,$69
	.dc.b	$F2,$1F,$D4,$0F,$F0,$3F,$EC,$BF,$C9,$B8,$BF,$17,$18,$F2,$3F,$C0
	.dc.b	$13,$E0,$F4,$4D,$26,$1F,$E1,$3F,$9F,$7E,$0E,$F9,$3E,$B3,$98,$C6
	.dc.b	$7F,$DF,$2E,$ED,$FE,$AE,$C3,$8F,$C9,$F9,$31,$E0,$F9,$BC,$1F,$19
	.dc.b	$30,$FF,$08,$61,$FB,$4C,$3E,$4E,$7E,$6B,$F6,$FC,$9F,$B9,$B7,$E1
	.dc.b	$F9,$2F,$82,$7C,$2F,$92,$B1,$E5,$F9,$1F,$82,$47,$BB,$F1,$3E,$4B
	.dc.b	$E4,$FC,$97,$CA,$6C,$FB,$7D,$0F,$9A,$7C,$EF,$46,$C7,$C1,$F9,$2F
	.dc.b	$9A,$47,$D6,$FD,$0F,$92,$FA,$A7,$D6,$F9,$2B,$3F,$27,$E4,$BE,$C9
	.dc.b	$F6,$BD,$3B,$1F,$37,$D2,$FB,$24,$7E,$3B,$F7,$3C,$1F,$B1,$C1,$FE
	.dc.b	$42,$ED,$FF,$5F,$F3,$7E,$5F,$9B,$EA,$4E,$70,$FF,$06,$7E,$CD,$FE
	.dc.b	$87,$7F,$C6,$AB,$E4,$7F,$CA,$2E,$6B,$FD,$7C,$E6,$77,$EA,$FF,$92
	.dc.b	$E3,$FD,$8D,$1F,$E4,$96,$CF,$DA,$6B,$FD,$63,$A5,$0F,$93,$EE,$78
	.dc.b	$3E,$3E,$DF,$93,$E4,$3C,$7F,$3F,$E4,$6C,$7A,$3E,$47,$C4,$47,$C4
	.dc.b	$F9,$7E,$47,$D6,$7B,$3E,$3A,$CF,$C0,$F9,$1F,$7F,$BB,$F0,$1B,$1F
	.dc.b	$12,$3D,$C4,$7D,$CF,$C1,$F7,$7E,$8F,$C7,$FC,$86,$CF,$F9,$9F,$C7
	.dc.b	$72,$7F,$95,$F8,$BE,$5F,$C8,$E9,$B5,$F4,$BE,$47,$E8,$E3,$F4,$9B
	.dc.b	$1F,$5A,$3E,$AF,$C8,$E2,$BE,$AB,$F6,$3E,$43,$EC,$FC,$8D,$BF,$BA
	.dc.b	$E1,$FB,$FF,$23,$6F,$EE,$BF,$80,$3E,$E1,$23,$EE,$3F,$7F,$E4,$51
	.dc.b	$FE,$BB,$E7,$FF,$2B,$79,$3F,$E5,$CF,$92,$FB,$98,$7F,$B4,$FA,$98
	.dc.b	$70,$8F,$F5,$F5,$F0,$71,$5C,$3A,$D8,$73,$AD,$63,$DE,$FA,$61,$C2
	.dc.b	$73,$1D,$00,$FA,$2F,$E3,$5F,$27,$FE,$4E,$6C,$7F,$36,$35,$E8,$FF
	.dc.b	$B6,$1C,$46,$9F,$E7,$7F,$2B,$FE,$B3,$FA,$3F,$F2,$BB,$63,$93,$5D
	.dc.b	$F5,$A7,$C1,$E9,$FF,$BD,$67,$AB,$5F,$E8,$F7,$FE,$18,$DB,$67,$FA
	.dc.b	$CF,$E3,$BF,$13,$3F,$06,$3E,$2E,$7F,$E4,$2E,$0F,$91,$C1,$E9,$AF
	.dc.b	$85,$4B,$E3,$18,$6D,$7C,$F6,$ED,$FE,$8F,$6C,$AB,$97,$A5,$DE,$BE
	.dc.b	$FF,$CA,$FF,$9E,$27,$EB,$B7,$1F,$F4,$A5,$FF,$B2,$DA,$FF,$3C,$C3
	.dc.b	$EE,$BF,$EC,$48,$CF,$91,$F8,$5B,$E7,$3F,$EC,$8A,$DA,$9D,$43,$D2
	.dc.b	$FF,$31,$C6,$2E,$F2,$7A,$DB,$7E,$BB,$E0,$F5,$F4,$73,$8F,$E4,$3E
	.dc.b	$56,$D9,$CD,$FF,$A0,$6F,$C9,$E9,$E3,$09,$18,$3D,$1E,$97,$2F,$95
	.dc.b	$91,$83,$F8,$27,$D1,$FF,$8F,$DC,$1E,$DD,$3E,$95,$7A,$3C,$45,$6E
	.dc.b	$B4,$7D,$5D,$3F,$D6,$AF,$CF,$F9,$6F,$A4,$47,$F1,$5E,$A7,$B2,$24
	.dc.b	$7C,$D8,$F7,$FB,$D2,$3E,$A3,$F0,$3D,$4F,$83,$EA,$7D,$FE,$0F,$F1
	.dc.b	$35,$7E,$56,$30,$7E,$0F,$E8,$DE,$9F,$CC,$7F,$3D,$CB,$F7,$5B,$C1
	.dc.b	$FE,$DD,$77,$7F,$83,$F9,$7F,$A2,$47,$F7,$75,$C7,$A7,$EA,$FC,$BF
	.dc.b	$E3,$C9,$3F,$8F,$7C,$5E,$97,$DA,$D2,$4F,$2C,$BC,$13,$F3,$78,$9F
	.dc.b	$03,$F8,$0F,$90,$97,$F5,$A3,$F0,$7F,$21,$F9,$D2,$47,$E7,$5F,$D0
	.dc.b	$7C,$7F,$E2,$09,$1D,$CF,$D5,$6E,$F5,$7E,$E6,$5F,$E9,$69,$FE,$8C
	.dc.b	$D7,$FC,$54,$DD,$FE,$47,$A9,$F9,$4F,$90,$FD,$71,$2B,$F5,$CF,$EC
	.dc.b	$7E,$43,$F3,$24,$8F,$17,$82,$C9,$E9,$FD,$5F,$47,$F3,$FF,$21,$2F
	.dc.b	$74,$9F,$D1,$FC,$87,$EF,$49,$3F,$BD,$7F,$80,$7C,$87,$EA,$09,$1C
	.dc.b	$BD,$2F,$D6,$FD,$17,$EF,$B9,$FF,$A3,$FC,$87,$EB,$8C,$F9,$7F,$B0
	.dc.b	$F9,$0F,$E1,$44,$C3,$F8,$53,$FC,$DD,$78,$CE,$5F,$C9,$C7,$E9,$38
	.dc.b	$7F,$A4,$FC,$87,$EE,$8F,$9F,$E1,$7E,$AF,$E8,$FF,$C3,$D5,$FC,$47
	.dc.b	$B3,$F7,$E6,$7E,$13,$F8,$1F,$EC,$FF,$D5,$E7,$FC,$A6,$D7,$FC,$34
	.dc.b	$D7,$FA,$7E,$28,$C4,$7F,$29,$5F,$1F,$D0,$91,$F5,$38,$5E,$2F,$F1
	.dc.b	$68,$7E,$09,$FC,$E3,$BB,$C4,$DE,$4F,$F8,$02,$D9,$FE,$BF,$F4,$BF
	.dc.b	$9A,$7F,$DB,$9F,$48,$F8,$FF,$91,$08,$C3,$F2,$DA,$FF,$63,$6B,$FE
	.dc.b	$9A,$5F,$21,$E5,$77,$8F,$C5,$6E,$3F,$DC,$2B,$B5,$7E,$03,$C6,$F9
	.dc.b	$7F,$25,$27,$E0,$47,$E1,$3C,$9D,$46,$0F,$88,$F7,$F4,$7E,$27,$CB
	.dc.b	$BE,$12,$99,$AF,$7F,$B3,$EE,$D3,$F9,$EC,$7A,$E3,$07,$F5,$7E,$27
	.dc.b	$91,$F7,$BE,$0A,$57,$D3,$8F,$CB,$F9,$9F,$98,$64,$7D,$78,$CB,$48
	.dc.b	$7E,$46,$9E,$57,$C7,$F9,$09,$3F,$51,$F9,$5E,$AC,$8F,$AE,$D7,$F9
	.dc.b	$E9,$F5,$FE,$43,$F2,$7F,$03,$23,$F7,$7D,$9F,$C1,$79,$BF,$23,$93
	.dc.b	$C9,$FD,$7F,$91,$FB,$06,$5F,$E3,$E3,$F3,$BD,$52,$39,$BF,$91,$FA
	.dc.b	$47,$C4,$FE,$4F,$8B,$F5,$D5,$FB,$16,$BF,$E2,$96,$BF,$EE,$06,$BF
	.dc.b	$83,$5F,$1B,$CF,$C8,$93,$FC,$94,$FF,$67,$47,$7F,$8A,$FD,$17,$E0
	.dc.b	$9F,$6B,$C0,$FD,$93,$77,$E9,$3E,$5E,$AF,$CE,$E5,$F8,$5C,$7C,$D1
	.dc.b	$83,$F4,$3F,$4B,$FE,$23,$8F,$03,$5F,$9E,$DC,$7F,$80,$3F,$DB,$DB
	.dc.b	$DF,$FB,$58,$C6,$3F,$9F,$6D,$7F,$94,$17,$AF,$FC,$2B,$C2,$93,$F6
	.dc.b	$1F,$DF,$7D,$2B,$E4,$3F,$80,$7D,$2B,$E5,$FF,$6E,$DB,$FE,$40,$D7
	.dc.b	$FC,$B0,$D7,$F6,$01,$7C,$37,$49,$93,$F3,$E7,$2F,$D1,$F8,$57,$B5
	.dc.b	$FC,$45,$B1,$98,$C7,$B1,$23,$83,$5F,$CB,$CF,$E4,$5F,$B9,$F4,$BF
	.dc.b	$C1,$1F,$F3,$9B,$53,$4F,$E1,$AF,$0D,$3F,$91,$46,$23,$F4,$6D,$E1
	.dc.b	$FE,$68,$FF,$21,$F4,$3F,$98,$7C,$9F,$F9,$4F,$C0,$FF,$2F,$DF,$F3
	.dc.b	$68,$C1,$FE,$57,$FA,$5F,$E7,$98,$7F,$93,$EB,$91,$CE,$33,$C9,$FC
	.dc.b	$F2,$1F,$17,$FD,$10,$98,$7F,$6C,$A7,$9F,$FE,$4A,$6C,$66,$30,$67
	.dc.b	$A7,$FA,$6F,$9D,$E7,$24,$FF,$A4,$F4,$FE,$95,$1E,$7F,$07,$73,$FC
	.dc.b	$02,$3B,$47,$F5,$6F,$33,$E0,$12,$79,$67,$FC,$89,$FF,$78,$65,$EE
	.dc.b	$9E,$11,$E9,$DB,$D1,$1F,$D8,$39,$64,$FB,$63,$2D,$21,$FD,$BE,$BF
	.dc.b	$CE,$BB,$FC,$BE,$6F,$A4,$49,$F1,$BF,$44,$FE,$DA,$47,$E5,$3F,$DC
	.dc.b	$3E,$2A,$47,$F1,$E7,$E6,$7C,$CF,$CD,$F8,$BF,$B4,$49,$F9,$18,$23
	.dc.b	$EF,$79,$9B,$9F,$AD,$1F,$56,$3F,$BE,$D4,$9F,$E5,$6F,$D9,$F2,$FE
	.dc.b	$D7,$C4,$FE,$1C,$93,$D9,$FF,$50,$7C,$8F,$F8,$2A,$3E,$17,$B9,$F8
	.dc.b	$13,$FD,$17,$19,$D7,$FD,$38,$FF,$98,$27,$FD,$58,$DE,$43,$4F,$F3
	.dc.b	$1C,$7F,$42,$F9,$2F,$E5,$35,$FE,$DC,$FA,$57,$C6,$FF,$6F,$9F,$E6
	.dc.b	$71,$8A,$FF,$70,$C6,$27,$5E,$57,$FA,$A6,$55,$B5,$5D,$97,$BB,$98
	.dc.b	$FE,$7E,$DF,$33,$FB,$F3,$9B,$E8,$DC,$7C,$D8,$F9,$6B,$FC,$09,$8F
	.dc.b	$1C,$72,$79,$B0,$F2,$47,$18,$F0,$D7,$17,$FD,$F7,$87,$17,$27,$1C
	.dc.b	$33,$3E,$E6,$C7,$C2,$C3,$CF,$3C,$E3,$89,$F0,$30,$F4,$C7,$18,$FF
	.dc.b	$70,$5D,$CF,$EF,$65,$F1,$5B,$7F,$2D,$FC,$15,$07,$C8,$7F,$91,$2F
	.dc.b	$F9,$AC,$60,$EA,$7F,$38,$8D,$34,$C9,$FE,$FD,$3C,$E3,$E4,$5E,$07
	.dc.b	$BA,$35,$5F,$6D,$D4,$65,$F2,$55,$AE,$31,$D0,$F2,$57,$18,$CB,$CE
	.dc.b	$3D,$1A,$E6,$FE,$CC,$79,$63,$0F,$E3,$33,$F7,$E1,$DE,$BC,$53,$DA
	.dc.b	$7F,$AB,$E1,$D9,$DC,$71,$3F,$C9,$51,$E6,$BF,$15,$F8,$27,$DA,$6F
	.dc.b	$1F,$D4,$FC,$C7,$B3,$6F,$D7,$B7,$2F,$F3,$AE,$EF,$99,$C6,$1F,$B2
	.dc.b	$AB,$87,$EC,$B8,$2D,$FF,$86,$30,$FF,$54,$BF,$E8,$F6,$EF,$AF,$8D
	.dc.b	$FF,$41,$5F,$EE,$9B,$5F,$EB,$FA,$FE,$78,$DC,$7F,$D1,$7C,$FE,$0E
	.dc.b	$EF,$E3,$15,$DA,$B1,$3E,$2F,$8D,$F1,$FA,$76,$BC,$6E,$1E,$AF,$80
	.dc.b	$8F,$F4,$2E,$AE,$C7,$B3,$E6,$2C,$7B,$3D,$3F,$17,$B5,$F0,$1B,$1F
	.dc.b	$E4,$E8,$F6,$E6,$69,$23,$DB,$F3,$1F,$20,$F9,$3E,$AD,$8F,$93,$1E
	.dc.b	$FE,$EC,$6F,$FE,$B7,$C1,$1F,$4B,$E0,$3F,$55,$7F,$E5,$EE,$FC,$5F
	.dc.b	$C7,$64,$7F,$5E,$7C,$A4,$7F,$B5,$4B,$E2,$FE,$77,$81,$D5,$B3,$FE
	.dc.b	$6A,$8F,$69,$1F,$ED,$2F,$DC,$97,$CA,$79,$4F,$37,$CA,$DA,$FF,$76
	.dc.b	$FE,$D0,$E8,$FC,$5B,$63,$FD,$E3,$7C,$9D,$CF,$BF,$E2,$37,$E1,$D9
	.dc.b	$E6,$E1,$FF,$02,$5F,$F3,$46,$C7,$E4,$23,$1A,$FF,$0E,$FA,$6F,$F6
	.dc.b	$7D,$7F,$9B,$5B,$1F,$E8,$C7,$F8,$4C,$67,$E2,$BF,$47,$29,$FF,$0D
	.dc.b	$CF,$1F,$47,$31,$CB,$C2,$FF,$72,$3A,$AF,$F4,$BF,$D3,$3C,$BE,$6D
	.dc.b	$C3,$FC,$FC,$FF,$95,$1B,$1A,$F2,$A4,$60,$DB,$E4,$FA,$5F,$F4,$E4
	.dc.b	$7F,$A1,$D7,$C1,$67,$E0,$F4,$C9,$FE,$5F,$8A,$B7,$18,$6C,$7F,$34
	.dc.b	$6B,$FE,$D9,$E1,$6E,$F9,$9F,$B6,$7C,$6F,$A0,$FF,$5E,$E1,$F8,$08
	.dc.b	$F9,$23,$A6,$98,$EA,$B3,$FE,$9F,$97,$4F,$03,$39,$7F,$B0,$23,$33
	.dc.b	$FE,$A5,$FA,$57,$89,$FC,$E2,$BF,$CF,$F1,$89,$F7,$B6,$3F,$A7,$B5
	.dc.b	$DF,$85,$E9,$E0,$FB,$F2,$8C,$57,$FA,$CF,$E9,$3F,$63,$7F,$ED,$27
	.dc.b	$F9,$7F,$37,$85,$CB,$7C,$FF,$C4,$C7,$F3,$49,$FF,$49,$B5,$FE,$C8
	.dc.b	$1C,$3F,$D9,$DB,$C3,$F4,$3E,$73,$F5,$27,$CF,$F9,$6B,$8F,$EA,$BE
	.dc.b	$1F,$F8,$37,$C6,$FF,$07,$6C,$7E,$CB,$D1,$FD,$89,$8C,$9A,$7F,$60
	.dc.b	$CB,$EE,$CF,$A3,$09,$7F,$07,$E1,$EE,$1F,$07,$E3,$BD,$A6,$31,$F8
	.dc.b	$4F,$F1,$7F,$84,$97,$FB,$EF,$8E,$B7,$E0,$8F,$F0,$26,$6F,$15,$7F
	.dc.b	$C7,$F9,$37,$97,$96,$7F,$3F,$87,$13,$F8,$63,$88,$F5,$75,$BE,$83
	.dc.b	$96,$5F,$C4,$9F,$A7,$C3,$F8,$7C,$3E,$8F,$1A,$31,$1F,$45,$AF,$D5
	.dc.b	$6C,$3E,$5B,$F8,$2F,$CD,$3E,$0B,$EF,$C6,$1F,$E7,$15,$FC,$E6,$2E
	.dc.b	$3F,$C8,$1C,$3F,$C9,$BD,$9B,$8F,$B7,$83,$F5,$F1,$FC,$4B,$E8,$FF
	.dc.b	$87,$57,$E2,$1F,$E1,$DB,$7E,$85,$CD,$FF,$8B,$DC,$CE,$FC,$8F,$5C
	.dc.b	$7F,$15,$9F,$F7,$2B,$5D,$4F,$F0,$6F,$45,$F2,$5F,$C8,$9B,$1F,$E3
	.dc.b	$BC,$3F,$61,$E2,$7F,$9C,$23,$07,$E7,$27,$77,$EB,$E2,$FF,$03,$37
	.dc.b	$DB,$FE,$B9,$F2,$FF,$D1,$1F,$E8,$BE,$E7,$F3,$92,$61,$FC,$F9,$FA
	.dc.b	$BF,$32,$FA,$37,$C4,$FF,$4C,$F7,$BF,$C3,$33,$98,$C5,$F9,$32,$4F
	.dc.b	$1E,$6F,$EA,$7E,$FF,$0C,$7E,$D5,$FF,$58,$8F,$EA,$FF,$21,$FD,$2C
	.dc.b	$91,$FD,$41,$FF,$0E,$FD,$1B,$E3,$BF,$AE,$F9,$7F,$E2,$88,$C5,$7F
	.dc.b	$62,$8F,$EC,$3E,$54,$8F,$27,$52,$9F,$17,$9D,$BE,$47,$C1,$7F,$5B
	.dc.b	$25,$FF,$60,$7C,$3F,$17,$FD,$B3,$E7,$BC,$0D,$9F,$EE,$3F,$07,$FD
	.dc.b	$C0,$F2,$7C,$5D,$8F,$EE,$BD,$5F,$D9,$C8,$FF,$6C,$7E,$7E,$5F,$DD
	.dc.b	$F2,$FD,$87,$27,$9B,$5F,$77,$C3,$7D,$0A,$D7,$6F,$EC,$BE,$7B,$FB
	.dc.b	$F9,$F0,$FE,$1A,$E7,$FE,$04,$F8,$6F,$EE,$C4,$7F,$BD,$BE,$0F,$13
	.dc.b	$FC,$0F,$F3,$DD,$5B,$3F,$23,$E8,$FF,$BE,$E3,$FC,$67,$F6,$9F,$E6
	.dc.b	$33,$FC,$32,$72,$FF,$0E,$8C,$EF,$FE,$13,$7F,$96,$93,$D7,$FF,$2B
	.dc.b	$B8,$F8,$8F,$F3,$0B,$FE,$27,$AF,$07,$B1,$89,$FF,$32,$E0,$CA,$F8
	.dc.b	$98,$6D,$F9,$FC,$3F,$F3,$1D,$7F,$B7,$23,$FD,$43,$C6,$8C,$69,$FD
	.dc.b	$10,$FF,$14,$38,$7F,$B9,$66,$F8,$7F,$29,$E2,$F1,$6E,$1F,$17,$C1
	.dc.b	$F4,$FC,$A7,$99,$B1,$F9,$0B,$4F,$C0,$24,$79,$23,$2F,$FA,$DB,$A3
	.dc.b	$F2,$BD,$BF,$03,$C0,$FC,$03,$67,$EA,$47,$C8,$F8,$9C,$69,$FC,$6F
	.dc.b	$0F,$F1,$F3,$FC,$77,$56,$1F,$2B,$BB,$E5,$B7,$1F,$BD,$1F,$4B,$E3
	.dc.b	$B1,$7C,$DF,$A3,$D7,$F4,$BA,$AD,$FE,$15,$C3,$EF,$F8,$AB,$7E,$37
	.dc.b	$EE,$9F,$80,$24,$7C,$17,$FD,$33,$F2,$5F,$E0,$0F,$8E,$75,$7F,$58
	.dc.b	$8C,$3F,$8F,$ED,$FF,$64,$CE,$B0,$FF,$30,$56,$9A,$6A,$BF,$29,$F4
	.dc.b	$56,$33,$8F,$FB,$52,$72,$D8,$FF,$75,$7D,$1F,$F2,$0E,$EF,$F1,$DB
	.dc.b	$63,$FD,$D6,$D7,$F1,$4D,$7F,$C1,$FF,$24,$EC,$7D,$27,$F9,$12,$30
	.dc.b	$D8,$CC,$FF,$BD,$A7,$2F,$6F,$4F,$FC,$F8,$DE,$EF,$1C,$62,$3C,$2D
	.dc.b	$7F,$65,$F4,$1F,$E2,$0D,$FE,$35,$3A,$1F,$09,$9D,$FF,$70,$DC,$3F
	.dc.b	$71,$85,$BE,$0D,$B3,$4C,$7C,$17,$1A,$7E,$25,$F5,$5F,$D7,$18,$C7
	.dc.b	$FC,$CA,$D3,$FC,$CF,$1B,$7F,$68,$FF,$83,$5F,$E9,$86,$CF,$F0,$A1
	.dc.b	$7C,$6F,$F2,$A4,$6E,$32,$BC,$5F,$F2,$0C,$EB,$DE,$BC,$9F,$E4,$4A
	.dc.b	$E1,$3C,$AB,$37,$FC,$67,$E1,$2E,$BF,$25,$BA,$9B,$F0,$92,$BF,$80
	.dc.b	$35,$FD,$A7,$C2,$74,$F1,$1F,$77,$F4,$A7,$FB,$44,$63,$8B,$53,$99
	.dc.b	$FD,$E7,$CD,$F8,$08,$FF,$0D,$7C,$7F,$0D,$E4,$F8,$6F,$F4,$8E,$BF
	.dc.b	$7D,$BA,$9A,$63,$E3,$B8,$61,$FC,$5E,$3D,$B1,$F0,$A1,$F5,$4E,$EF
	.dc.b	$83,$F0,$B2,$7E,$AC,$E5,$FE,$41,$F0,$BE,$02,$3F,$C7,$DF,$1F,$C0
	.dc.b	$FC,$CD,$FF,$D1,$9C,$BF,$E6,$56,$F3,$FF,$4B,$9D,$7C,$A4,$C3,$FB
	.dc.b	$DD,$23,$97,$C3,$73,$F8,$3F,$83,$3F,$E1,$8A,$FE,$82,$D7,$FA,$07
	.dc.b	$0F,$E8,$A7,$7E,$07,$8C,$93,$DB,$E0,$FC,$DF,$03,$23,$C2,$FA,$7D
	.dc.b	$AF,$57,$CD,$F8,$09,$3E,$07,$C7,$E5,$7B,$C9,$1E,$AF,$9B,$F8,$5F
	.dc.b	$01,$F0,$FE,$02,$4F,$BA,$3E,$37,$CA,$7B,$09,$1E,$C7,$DF,$F2,$9F
	.dc.b	$2F,$E0,$64,$FC,$68,$F9,$DF,$33,$F3,$FE,$0A,$4F,$CA,$8F,$A5,$F2
	.dc.b	$7F,$20,$91,$F2,$1F,$97,$F2,$9F,$5C,$91,$F4,$BE,$53,$F4,$BF,$B5
	.dc.b	$7F,$13,$EC,$76,$2B,$FC,$87,$F8,$4F,$FB,$D5,$FC,$FB,$5F,$F3,$95
	.dc.b	$71,$9F,$F4,$BA,$F9,$AF,$CD,$B6,$7F,$BD,$B5,$FB,$6D,$7F,$DD,$4F
	.dc.b	$FB,$A5,$FF,$20,$B5,$EF,$F2,$FE,$8F,$47,$D3,$F2,$DF,$B5,$C7,$FD
	.dc.b	$D5,$5E,$6D,$32,$F2,$F4,$BC,$64,$7C,$6F,$9B,$E2,$7E,$19,$23,$C3
	.dc.b	$E9,$7E,$AF,$E9,$7F,$25,$BF,$FB,$AD,$78,$79,$35,$E0,$EF,$C1,$FE
	.dc.b	$4F,$3F,$E8,$6C,$3F,$95,$87,$0F,$F2,$E6,$BF,$0B,$E2,$79,$3F,$CA
	.dc.b	$A3,$78,$7F,$B4,$F8,$55,$F6,$BD,$AF,$B7,$F2,$13,$0F,$93,$1F,$7B
	.dc.b	$E2,$B5,$18,$3E,$6B,$F8,$2F,$90,$FC,$1F,$C5,$49,$FA,$31,$F8,$6F
	.dc.b	$8F,$D4,$60,$FA,$AF,$C5,$F0,$FD,$E4,$8F,$7E,$A8,$F8,$7E,$16,$E7
	.dc.b	$D9,$1F,$36,$24,$7B,$DA,$FF,$48,$EF,$7C,$1F,$F3,$3F,$C0,$7E,$0C
	.dc.b	$FF,$32,$6B,$FE,$72,$6F,$0F,$C7,$F8,$AF,$CF,$92,$7F,$1A,$FE,$8B
	.dc.b	$E2,$BF,$47,$F2,$12,$7F,$22,$FE,$9B,$E2,$BF,$4F,$F1,$56,$7F,$2B
	.dc.b	$18,$7F,$AF,$F4,$7F,$69,$C7,$F5,$71,$FA,$B8,$FC,$D4,$62,$3F,$D6
	.dc.b	$BE,$0F,$F3,$79,$E1,$E1,$FE,$C6,$7F,$D7,$3E,$0F,$F7,$09,$34,$91
	.dc.b	$F8,$CF,$A5,$FF,$37,$3E,$28,$E8,$7F,$43,$8C,$C7,$EE,$37,$FE,$CB
	.dc.b	$FB,$5F,$F9,$AF,$2F,$3F,$A4,$8C,$65,$FB,$CA,$DF,$B3,$FC,$21,$98
	.dc.b	$7F,$AD,$AB,$FA,$53,$FD,$FB,$E9,$7F,$CE,$4F,$F8,$09,$AE,$7E,$4B
	.dc.b	$FA,$B4,$7F,$9B,$A3,$5F,$25,$82,$3D,$7E,$4F,$FA,$E7,$D2,$7E,$C2
	.dc.b	$3F,$C1,$AF,$F4,$B6,$FB,$5A,$F1,$B7,$C5,$FE,$18,$AF,$E9,$55,$FE
	.dc.b	$89,$6B,$F3,$DA,$FF,$21,$FC,$59,$D9,$CA,$3B,$D7,$F6,$D9,$CC,$7F
	.dc.b	$87,$B2,$65,$FD,$C5,$B1,$98,$C6,$92,$38,$35,$71,$7D,$2D,$BE,$3B
	.dc.b	$F8,$7F,$CB,$3E,$9C,$FF,$35,$6A,$69,$F4,$A3,$87,$0F,$F8,$0E,$31
	.dc.b	$C5,$F8,$D6,$D3,$8F,$F0,$19,$FE,$13,$E5,$BF,$C2,$79,$1A,$4A,$FF
	.dc.b	$5A,$C3,$E5,$78,$47,$E0,$E7,$F1,$B1,$FB,$7E,$C9,$3F,$B7,$7F,$23
	.dc.b	$F2,$B2,$3F,$14,$D7,$E1,$E7,$FD,$4F,$2F,$CA,$7C,$AF,$E6,$3E,$47
	.dc.b	$CA,$4A,$F1,$BE,$8F,$81,$D4,$7E,$56,$3F,$3F,$1F,$A0,$F9,$1F,$69
	.dc.b	$27,$D6,$FC,$0F,$90,$D4,$7E,$76,$3F,$4F,$1F,$A8,$F9,$09,$3F,$A1
	.dc.b	$7F,$57,$F2,$12,$3F,$48,$D7,$D6,$7C,$FE,$07,$C7,$7E,$CB,$E3,$BE
	.dc.b	$79,$27,$E6,$BF,$4B,$E3,$75,$1F,$AF,$8F,$DC,$C7,$F9,$57,$F3,$47
	.dc.b	$C2,$5F,$84,$97,$FE,$56,$9C,$B7,$E0,$FF,$A9,$FC,$8A,$51,$86,$CF
	.dc.b	$FB,$75,$AF,$F8,$CB,$37,$C7,$F8,$6A,$4F,$F0,$38,$FF,$34,$7E,$CE
	.dc.b	$E1,$BC,$72,$BF,$CD,$33,$97,$CD,$F0,$1E,$12,$3E,$17,$F2,$7F,$21
	.dc.b	$FC,$63,$E3,$E4,$FF,$11,$8F,$E3,$7F,$1F,$23,$F8,$A3,$FE,$82,$FD
	.dc.b	$B3,$FC,$AB,$1F,$E1,$DF,$96,$FE,$D3,$E2,$7F,$87,$FE,$9F,$F3,$53
	.dc.b	$9F,$84,$FE,$65,$18,$85,$7F,$A5,$5E,$1E,$17,$FB,$23,$E0,$BF,$73
	.dc.b	$0A,$E1,$5D,$A3,$FC,$49,$5F,$EC,$EB,$FE,$7D,$F1,$3A,$6C,$73,$C7
	.dc.b	$FC,$65,$3C,$B1,$EB,$8E,$F0,$F0,$9E,$72,$9E,$68,$DD,$F3,$8F,$09
	.dc.b	$B2,$9E,$CC,$3C,$2F,$C5,$9F,$5D,$F8,$70,$F1,$DF,$A6,$A9,$CE,$74
	.dc.b	$F4,$C7,$E0,$78,$3F,$E3,$F6,$CF,$8E,$31,$87,$C8,$7F,$C8,$54,$F9
	.dc.b	$7F,$E0,$15,$FE,$D2,$6C,$7F,$65,$34,$7F,$19,$8D,$34,$D4,$79,$9F
	.dc.b	$F2,$E7,$93,$FD,$D2,$7E,$E4,$7F,$97,$E7,$4F,$FA,$9F,$C4,$AF,$69
	.dc.b	$DD,$71,$CB,$BC,$62,$B7,$7F,$4A,$F9,$1F,$83,$AF,$2E,$BB,$8C,$1E
	.dc.b	$63,$E9,$9C,$2F,$CA,$F8,$67,$E0,$61,$E1,$8E,$15,$CC,$DD,$F8,$67
	.dc.b	$FC,$23,$1F,$9C,$C5,$B7,$FA,$C7,$2F,$F3,$F6,$BF,$E0,$08,$C6,$3F
	.dc.b	$C5,$AB,$7E,$87,$F8,$6A,$B5,$E8,$DC,$3F,$9A,$60,$F3,$F1,$8F,$F4
	.dc.b	$B8,$7C,$36,$32,$F0,$D5,$9C,$DF,$F1,$5F,$8B,$FF,$12,$61,$FE,$91
	.dc.b	$EF,$FE,$29,$18,$84,$62,$7F,$8A,$4E,$61,$D2,$E8,$F1,$C6,$39,$EC
	.dc.b	$E6,$FF,$88,$C7,$8F,$E4,$F7,$F9,$DF,$F8,$BA,$7E,$29,$EE,$8C,$FB
	.dc.b	$32,$3F,$25,$9F,$92,$B8,$3F,$95,$C5,$18,$8F,$F1,$E4,$62,$BD,$12
	.dc.b	$F8,$BF,$0F,$C5,$F8,$9B,$87,$8B,$0F,$99,$F2,$9F,$27,$E5,$3E,$53
	.dc.b	$67,$5C,$5F,$7A,$7E,$43,$63,$CE,$DA,$FF,$75,$F0,$AB,$6B,$E5,$3E
	.dc.b	$B1,$F5,$FE,$52,$E1,$DA,$7F,$CF,$8D,$F8,$7C,$E1,$F8,$99,$CB,$7D
	.dc.b	$CF,$C0,$79,$9F,$81,$6D,$F3,$8F,$09,$1F,$3D,$7F,$5C,$6F,$99,$F8
	.dc.b	$9F,$2B,$F1,$4D,$8F,$24,$7A,$88,$F2,$7D,$BF,$29,$F9,$13,$F2,$5F
	.dc.b	$29,$67,$E7,$72,$BD,$67,$AF,$F8,$A6,$CF,$C1,$BC,$4F,$E6,$FE,$27
	.dc.b	$F3,$9F,$29,$6B,$E3,$38,$7D,$6F,$CE,$23,$FE,$C4,$F0,$5E,$5F,$EA
	.dc.b	$B8,$51,$89,$FA,$1E,$6E,$63,$13,$FE,$0C,$37,$E5,$E4,$D5,$DB,$FC
	.dc.b	$54,$BF,$0B,$FC,$84,$61,$B1,$F6,$7C,$3F,$14,$63,$C6,$FF,$4C,$2E
	.dc.b	$EC,$9A,$EE,$B8,$76,$FB,$57,$AD,$EB,$F8,$7F,$DF,$E7,$E7,$9C,$D7
	.dc.b	$92,$73,$97,$FA,$95,$DD,$7F,$3F,$6B,$FE,$A5,$6A,$0E,$1B,$E2,$78
	.dc.b	$7C,$2F,$13,$6F,$E1,$FD,$37,$EE,$30,$FF,$5A,$36,$3F,$CA,$1F,$29
	.dc.b	$FE,$65,$77,$E3,$7D,$59,$F8,$91,$FE,$B7,$6E,$3E,$37,$93,$75,$FE
	.dc.b	$45,$6F,$46,$E9,$86,$E1,$FB,$A9,$F5,$B7,$E5,$F3,$87,$F3,$59,$47
	.dc.b	$E1,$FE,$D5,$AF,$E2,$7E,$99,$F0,$99,$8C,$33,$1F,$E3,$38,$AB,$71
	.dc.b	$86,$C7,$F1,$E6,$BF,$DC,$5F,$F4,$24,$7E,$31,$FF,$41,$37,$E1,$3C
	.dc.b	$5E,$B7,$C0,$E0,$F8,$3B,$DA,$F7,$3F,$13,$96,$46,$0F,$84,$FC,$8E
	.dc.b	$89,$1F,$17,$A6,$E7,$FE,$6C,$F0,$7F,$71,$ED,$FE,$DC,$CF,$37,$FB
	.dc.b	$CA,$33,$7E,$79,$F8,$C7,$C9,$F2,$2C,$64,$F6,$7B,$38,$C3,$F5,$DC
	.dc.b	$08,$F9,$C7,$F5,$58,$F9,$4D,$8D,$4F,$CB,$8F,$E6,$7C,$AF,$97,$FC
	.dc.b	$1B,$F8,$46,$CF,$E0,$63,$F0,$FF,$2F,$8C,$3F,$00,$FE,$2B,$E5,$BF
	.dc.b	$17,$F2,$F6,$FF,$10,$E1,$FC,$57,$CB,$DB,$FC,$43,$F9,$2F,$95,$FE
	.dc.b	$1D,$F4,$BF,$DE,$6B,$F2,$11,$8F,$6A,$47,$E1,$A7,$57,$FD,$81,$AF
	.dc.b	$F8,$CD,$AF,$EC,$C3,$FB,$C8,$C3,$F9,$67,$F8,$9B,$98,$FD,$D3,$97
	.dc.b	$F3,$DF,$17,$FA,$83,$F5,$3F,$2D,$6B,$F4,$DF,$2B,$F3,$08,$FF,$5E
	.dc.b	$D9,$1F,$43,$E5,$BF,$C0,$7F,$81,$FF,$5D,$AF,$D7,$46,$3D,$9F,$DA
	.dc.b	$FC,$B4,$9F,$ED,$7F,$B6,$FF,$0D,$7F,$69,$1F,$E1,$B8,$FF,$1E,$E9
	.dc.b	$EF,$F7,$51,$88,$FF,$6C,$35,$FF,$77,$36,$37,$F2,$9E,$4F,$63,$A3
	.dc.b	$70,$F0,$FC,$A7,$73,$C1,$E3,$58,$F2,$F8,$DF,$DF,$B1,$FD,$9E,$5F
	.dc.b	$E6,$5C,$BC,$73,$93,$5E,$37,$A7,$D6,$F5,$37,$0F,$7F,$B1,$ED,$F8
	.dc.b	$AF,$73,$63,$E1,$FC,$9F,$72,$3F,$E0,$CC,$D1,$E3,$F8,$CF,$8D,$F2
	.dc.b	$7F,$1D,$B5,$F3,$3E,$4F,$E5,$1F,$2F,$E4,$EC,$7D,$0F,$93,$F6,$11
	.dc.b	$EC,$FB,$FD,$8F,$A4,$7D,$3F,$93,$B3,$F6,$3C,$AF,$AC,$7D,$7F,$22
	.dc.b	$C7,$DC,$F8,$BF,$90,$47,$D4,$FC,$BF,$8A,$FB,$C7,$F9,$67,$F2,$14
	.dc.b	$E0,$FF,$17,$BF,$E6,$8F,$3F,$3D,$0E,$84,$BE,$3B,$97,$C0,$73,$7A
	.dc.b	$79,$1F,$E3,$69,$FD,$4D,$F1,$6D,$F2,$9F,$E1,$DF,$07,$F9,$CC,$BC
	.dc.b	$3F,$1D,$FE,$07,$8C,$61,$FE,$79,$C3,$F5,$98,$78,$BD,$9F,$57,$BF
	.dc.b	$EB,$6E,$3E,$A8,$F1,$11,$FF,$23,$D2,$3C,$DF,$19,$F0,$CF,$89,$E3
	.dc.b	$5A,$FF,$14,$FE,$0B,$E2,$E0,$FE,$60,$D8,$FE,$ED,$97,$C4,$BF,$F3
	.dc.b	$AC,$E7,$3F,$81,$F1,$BF,$3C,$FA,$1F,$1B,$70,$F9,$FE,$C7,$D3,$31
	.dc.b	$5E,$F7,$FC,$51,$F8,$B3,$A6,$F4,$E1,$B7,$FE,$22,$94,$F0,$E9,$CE
	.dc.b	$1F,$E2,$B9,$FE,$72,$D8,$FE,$7D,$18,$8F,$F6,$93,$5F,$E4,$0D,$7E
	.dc.b	$D7,$C7,$7E,$13,$E4,$FF,$BB,$C6,$30,$FF,$68,$7C,$1F,$C4,$E4,$7C
	.dc.b	$76,$E3,$18,$5B,$E0,$7F,$B9,$63,$87,$9D,$9E,$3F,$C7,$D5,$C1,$3C
	.dc.b	$6D,$F4,$BF,$EB,$7F,$94,$FF,$00,$B6,$3F,$D3,$6D,$7F,$DD,$CD,$7F
	.dc.b	$9E,$7C,$AF,$F5,$9F,$A5,$FF,$04,$B6,$3F,$D4,$6D,$7F,$DE,$CD,$7F
	.dc.b	$A2,$4F,$FB,$13,$2F,$0F,$6F,$E9,$E7,$C2,$D7,$C2,$D7,$F1,$3F,$47
	.dc.b	$FC,$39,$F9,$27,$EB,$DC,$3F,$14,$C7,$95,$FD,$4F,$87,$E8,$DA,$9E
	.dc.b	$6E,$78,$3A,$B7,$0E,$27,$6A,$E1,$86,$5E,$53,$B9,$7D,$2B,$E2,$7F
	.dc.b	$C5,$B1,$86,$C6,$63,$FC,$62,$D7,$FD,$96,$FE,$19,$FF,$3E,$39,$EE
	.dc.b	$FE,$1B,$1C,$CF,$F3,$DC,$6E,$32,$FF,$B2,$E3,$38,$FF,$73,$9C,$B7
	.dc.b	$97,$FB,$B5,$6E,$BF,$84,$7C,$9E,$5F,$E2,$4D,$CB,$F8,$DB,$74,$79
	.dc.b	$D4,$95,$FE,$D2,$8C,$B7,$CD,$E5,$CC,$F9,$4A,$49,$FF,$58,$FD,$2F
	.dc.b	$DA,$DB,$FC,$70,$DD,$7F,$CF,$DF,$10,$8C,$52,$31,$C5,$E5,$F3,$FD
	.dc.b	$52,$FA,$0F,$F1,$83,$FE,$60,$6B,$FE,$FA,$6B,$FE,$19,$38,$7C,$47
	.dc.b	$F9,$9B,$87,$F1,$1E,$17,$4A,$E3,$A7,$D4,$6E,$9F,$96,$9F,$8D,$E0
	.dc.b	$FC,$0F,$2E,$65,$F1,$63,$CB,$F4,$AE,$FF,$C3,$B6,$7F,$CE,$AD,$7C
	.dc.b	$81,$C3,$FE,$11,$DB,$F0,$B9,$FF,$52,$D3,$F3,$DE,$8F,$D7,$3E,$D4
	.dc.b	$EB,$4F,$E2,$1F,$2F,$F5,$09,$87,$FA,$AF,$0E,$55,$FE,$D2,$8F,$AB
	.dc.b	$E9,$FF,$A2,$38,$BE,$93,$78,$FF,$D7,$F3,$E6,$9C,$BF,$D2,$BE,$5F
	.dc.b	$F0,$24,$7B,$65,$FD,$13,$07,$81,$F8,$6F,$97,$95,$FD,$B9,$F4,$FA
	.dc.b	$1F,$95,$C7,$F0,$5B,$78,$49,$5E,$1C,$D1,$F6,$7F,$35,$FE,$19,$B7
	.dc.b	$C5,$6E,$9F,$D5,$1B,$7F,$DE,$74,$79,$DF,$23,$C9,$F0,$7D,$37,$C3
	.dc.b	$AF,$F4,$83,$71,$F8,$F3,$AF,$53,$E3,$49,$A4,$AF,$B7,$C9,$FE,$28
	.dc.b	$F0,$BA,$FA,$92,$7E,$29,$E3,$8F,$B3,$D5,$FB,$3D,$7F,$D9,$53,$AB
	.dc.b	$46,$21,$9F,$DA,$AF,$4E,$9E,$42,$4F,$D7,$CF,$DE,$F9,$2F,$FD,$D6
	.dc.b	$F5,$6B,$D3,$6F,$91,$F2,$DF,$8C,$E6,$F8,$BE,$77,$97,$C7,$97,$EA
	.dc.b	$E5,$65,$FD,$7A,$30,$B7,$F8,$E4,$8F,$23,$F2,$B5,$FA,$24,$8F,$81
	.dc.b	$F4,$DE,$4D,$FF,$D7,$D6,$CF,$E3,$79,$32,$FB,$C7,$E2,$7E,$97,$FA
	.dc.b	$64,$7F,$0C,$6B,$FC,$2D,$B9,$9A,$77,$DF,$86,$FC,$CF,$E4,$D9,$BE
	.dc.b	$57,$F8,$8F,$D3,$7F,$90,$71,$F9,$5F,$2F,$F6,$4C,$FB,$7A,$E8,$FE
	.dc.b	$8F,$5F,$87,$AF,$F7,$AB,$5F,$E0,$2D,$7F,$8B,$3F,$A8,$7F,$0E,$1E
	.dc.b	$1C,$1F,$D2,$1B,$9F,$FA,$D5,$E1,$3F,$6C,$FF,$2C,$64,$F8,$CF,$C7
	.dc.b	$67,$27,$E7,$C7,$BB,$C8,$F7,$B2,$3E,$9C,$7C,$2F,$92,$FC,$B7,$CB
	.dc.b	$C9,$FA,$F1,$F1,$BE,$53,$51,$83,$ED,$BF,$23,$AF,$F9,$CF,$D6,$BE
	.dc.b	$46,$99,$AF,$7E,$AF,$7B,$E4,$92,$3E,$4B,$F3,$3D,$EF,$A6,$C8,$E3
	.dc.b	$1F,$83,$F9,$8F,$D6,$12,$3F,$31,$CE,$8F,$BB,$F2,$BB,$9C,$C6,$1F
	.dc.b	$C9,$E3,$FB,$5D,$E4,$60,$FC,$BF,$C9,$FE,$1A,$64,$60,$FB,$3C,$DF
	.dc.b	$8D,$47,$F5,$9E,$8F,$EA,$07,$F5,$C8,$D4,$7F,$0D,$F0,$3F,$C1,$F1
	.dc.b	$FE,$09,$8F,$EB,$4D,$7F,$DE,$8D,$7A,$BF,$E0,$DE,$7C,$CF,$53,$B7
	.dc.b	$6F,$F8,$07,$E8,$BF,$B2,$75,$7F,$29,$E8,$FE,$24,$67,$1F,$D0,$FC
	.dc.b	$95,$7F,$8A,$57,$0C,$7F,$D0,$3D,$6A,$FE,$EB,$F4,$EF,$92,$BF,$27
	.dc.b	$31,$FC,$DC,$E5,$BF,$27,$C2,$A4,$60,$FD,$0F,$D3,$FF,$33,$CB,$FD
	.dc.b	$1F,$E2,$B9,$3F,$C3,$4D,$AD,$FC,$67,$F8,$1A,$FF,$6D,$1F,$DB,$9A
	.dc.b	$FF,$80,$DA,$FD,$40,$E1,$FA,$EF,$EB,$9B,$E1,$F5,$3B,$72,$7F,$85
	.dc.b	$33,$6B,$7D,$79,$7F,$BA,$9C,$B7,$5E,$15,$23,$07,$F0,$0F,$A6,$F9
	.dc.b	$CF,$E2,$3F,$33,$F9,$CF,$F0,$46,$BF,$BF,$11,$C1,$B8,$7F,$8A,$23
	.dc.b	$04,$7F,$94,$7C,$DF,$F8,$07,$B8,$D2,$57,$F5,$CE,$57,$CB,$F0,$8F
	.dc.b	$D5,$CF,$E2,$E3,$FA,$D7,$CE,$49,$FE,$9C,$FF,$35,$F9,$B9,$1F,$B2
	.dc.b	$6B,$FC,$5F,$D2,$60,$F9,$9F,$E7,$9C,$DE,$52,$57,$F8,$4F,$3F,$C0
	.dc.b	$3F,$E3,$DB,$E3,$F1,$3C,$23,$F7,$F5,$F9,$F8,$FE,$8D,$D5,$EF,$24
	.dc.b	$F3,$7F,$C8,$98,$BE,$5B,$73,$FC,$22,$7F,$51,$1F,$DD,$FC,$B9,$3F
	.dc.b	$DA,$DF,$D6,$7C,$3C,$8F,$D3,$B5,$F7,$5B,$E5,$BF,$C0,$1E,$77,$F8
	.dc.b	$6C,$93,$FD,$DD,$F1,$FC,$06,$A3,$FB,$C4,$7E,$E6,$3F,$C5,$1F,$4F
	.dc.b	$FD,$C6,$7F,$A0,$7A,$8F,$CD,$17,$FE,$2D,$6C,$FF,$79,$AE,$1E,$33
	.dc.b	$C3,$FF,$39,$C3,$DC,$FF,$22,$36,$BF,$C8,$4D,$7F,$CD,$98,$A3,$11
	.dc.b	$FA,$B8,$C7,$C9,$79,$AA,$4F,$86,$3F,$C8,$1F,$4F,$FE,$B8,$8F,$F0
	.dc.b	$63,$FE,$82,$A3,$4D,$47,$B9,$FA,$CF,$F9,$55,$AF,$91,$AF,$8D,$AF
	.dc.b	$F3,$2E,$5F,$C7,$FC,$07,$CA,$F6,$A4,$FF,$1E,$8F,$E4,$DF,$0F,$23
	.dc.b	$F9,$0B,$FC,$A7,$E6,$7F,$D2,$7F,$27,$FD,$2F,$7F,$E7,$BF,$50,$F9
	.dc.b	$DF,$F7,$6F,$BB,$FC,$36,$31,$8F,$F4,$2B,$4E,$6B,$FD,$E7,$EB,$FF
	.dc.b	$8F,$A9,$7C,$15,$F7,$5F,$47,$98,$E7,$1F,$D3,$A3,$18,$7F,$1D,$75
	.dc.b	$7D,$5B,$8F,$A3,$1F,$3D,$D3,$CD,$7E,$57,$D1,$7E,$5C,$78,$C7,$8A
	.dc.b	$B8,$B4,$E2,$78,$DF,$6D,$D8,$E6,$D7,$DB,$7E,$79,$E7,$87,$63,$BE
	.dc.b	$DF,$E6,$3C,$BE,$13,$72,$F9,$0D,$C7,$EF,$57,$FA,$BB,$E5,$6E,$5F
	.dc.b	$EC,$36,$BF,$D0,$5A,$FF,$76,$6A,$7C,$CF,$FA,$F7,$96,$B9,$3E,$01
	.dc.b	$FD,$FB,$C7,$FD,$63,$83,$A1,$5C,$D7,$DD,$8C,$B6,$33,$9F,$A6,$7F
	.dc.b	$06,$FF,$B6,$E3,$AE,$BC,$67,$C4,$F1,$BF,$B5,$1E,$D3,$C0,$FA,$30
	.dc.b	$F5,$D7,$69,$F8,$87,$63,$FB,$23,$DA,$FC,$57,$D6,$75,$93,$E9,$FF
	.dc.b	$D4,$98,$FF,$A6,$5B,$3F,$CF,$57,$A3,$F1,$34,$FF,$6E,$46,$9B,$F1
	.dc.b	$99,$BF,$E2,$7F,$19,$70,$63,$DE,$B1,$CF,$FC,$46,$AF,$3F,$FB,$1B
	.dc.b	$0F,$E0,$9C,$DF,$4B,$1F,$E0,$4D,$7F,$C6,$77,$FC,$3A,$31,$08,$FE
	.dc.b	$23,$CB,$DB,$CB,$FE,$4D,$7F,$C5,$A3,$0F,$8B,$EA,$1E,$CD,$F6,$F0
	.dc.b	$D9,$FF,$54,$4A,$7B,$E0,$8C,$4F,$FB,$BD,$B8,$FF,$2E,$6B,$FC,$B5
	.dc.b	$FF,$4A,$47,$FA,$CF,$BD,$E6,$78,$FE,$13,$E2,$C6,$27,$EA,$7C,$EF
	.dc.b	$CB,$E2,$FF,$66,$ED,$79,$13,$9F,$8A,$FF,$01,$DF,$FB,$1E,$BE,$2D
	.dc.b	$F0,$8F,$F6,$64,$62,$77,$A7,$F7,$3C,$3F,$CF,$55,$F8,$3F,$83,$ED
	.dc.b	$F3,$DD,$DE,$DF,$07,$27,$47,$FB,$55,$F2,$78,$BF,$7B,$E8,$1E,$D6
	.dc.b	$2B,$FD,$A5,$18,$6C,$7C,$C8,$C6,$5F,$D5,$9A,$FF,$55,$6B,$FD,$0F
	.dc.b	$27,$8F,$DF,$D6,$F8,$8D,$C3,$BC,$75,$23,$FE,$49,$AF,$C5,$DF,$F8
	.dc.b	$9A,$31,$1F,$8B,$8F,$F0,$E4,$7A,$3D,$5F,$A1,$F1,$7E,$D7,$E8,$72
	.dc.b	$C9,$D1,$FE,$F7,$7E,$B7,$C8,$FC,$3F,$56,$CF,$C2,$70,$F8,$3A,$58
	.dc.b	$A7,$E2,$12,$7C,$B9,$7E,$D3,$CA,$FD,$73,$FB,$6F,$26,$F8,$9F,$D5
	.dc.b	$AB,$FD,$A6,$FF,$AA,$5B,$1A,$F8,$6D,$B6,$3F,$DB,$3F,$51,$F9,$AF
	.dc.b	$1B,$F6,$4F,$E6,$30,$FF,$2C,$39,$9F,$F3,$93,$99,$DF,$77,$92,$3C
	.dc.b	$55,$C2,$31,$97,$A3,$E3,$1E,$3D,$F8,$FC,$37,$2F,$F7,$5C,$A7,$F7
	.dc.b	$F8,$23,$13,$F1,$1B,$1F,$69,$AF,$DF,$5F,$25,$FA,$BF,$71,$F3,$5E
	.dc.b	$38,$C3,$F3,$3E,$9F,$F8,$2E,$1F,$EF,$C7,$E0,$61,$FE,$C4,$72,$FF
	.dc.b	$9E,$DC,$A6,$9F,$16,$3F,$D3,$0E,$34,$FD,$E7,$C5,$7A,$E7,$FD,$9F
	.dc.b	$BF,$F5,$2F,$11,$2F,$8B,$DE,$BC,$19,$7C,$0F,$F1,$37,$89,$FE,$F7
	.dc.b	$71,$9F,$F3,$2B,$E5,$18,$C1,$5B,$AF,$F4,$A7,$D3,$7E,$CB,$2F,$E6
	.dc.b	$4F,$E9,$63,$3E,$0F,$E4,$23,$13,$AE,$FF,$AB,$1B,$AF,$D7,$35,$FF
	.dc.b	$5C,$35,$FF,$76,$BF,$E3,$1C,$7C,$F0,$FA,$EF,$E8,$8F,$93,$D2,$FC
	.dc.b	$FC,$63,$0F,$8D,$A7,$D4,$D3,$E3,$EF,$F3,$3C,$CF,$9A,$CC,$BE,$06
	.dc.b	$DF,$ED,$37,$E1,$5F,$D5,$CF,$FA,$BC,$E5,$FF,$4F,$47,$E6,$B1,$FF
	.dc.b	$02,$F7,$7E,$DB,$E8,$BE,$CF,$4F,$F1,$F6,$63,$DB,$E9,$6E,$3F,$EA
	.dc.b	$7C,$91,$F2,$6B,$E5,$7B,$36,$FE,$0F,$0A,$33,$1F,$C0,$23,$0D,$8F
	.dc.b	$9E,$7D,$0F,$4B,$F8,$1F,$D7,$FE,$8B,$5F,$B3,$18,$EC,$FC,$7E,$FF
	.dc.b	$90,$F8,$6B,$8F,$C2,$70,$FC,$6F,$85,$BB,$FA,$5F,$CB,$FB,$FF,$98
	.dc.b	$F4,$6D,$7C,$A7,$0F,$E3,$7D,$05,$E0,$FF,$64,$38,$8F,$BF,$F5,$1F
	.dc.b	$E0,$B7,$FD,$F8,$FF,$04,$73,$F1,$9F,$C7,$A3,$15,$FE,$E9,$6B,$FC
	.dc.b	$ED,$AF,$2F,$13,$F8,$3C,$62,$3C,$6F,$C5,$AF,$F4,$65,$FF,$91,$67
	.dc.b	$A6,$BF,$DB,$E3,$1B,$FE,$03,$EA,$3C,$2D,$C7,$F7,$3E,$6F,$F8,$72
	.dc.b	$73,$8F,$F8,$59,$FE,$1E,$FF,$3D,$6E,$3F,$EB,$26,$BF,$EB,$16,$BC
	.dc.b	$7D,$4F,$17,$C6,$75,$6D,$7E,$F3,$AD,$7F,$DE,$2F,$FD,$96,$DC,$BF
	.dc.b	$4B,$C3,$E5,$D3,$CD,$EA,$7F,$86,$AB,$FD,$91,$8F,$96,$30,$67,$D3
	.dc.b	$F8,$5F,$1B,$ED,$6C,$FD,$DF,$17,$E3,$1F,$E1,$5F,$A8,$56,$27,$CD
	.dc.b	$83,$D3,$F0,$12,$3E,$23,$FE,$16,$AF,$ED,$FA,$FE,$1B,$C7,$E9,$EA
	.dc.b	$7D,$37,$B7,$A7,$FE,$4A,$FA,$8F,$F5,$5E,$3F,$C6,$F0,$FE,$D0,$FC
	.dc.b	$58,$CE,$1F,$EE,$E7,$F9,$0B,$FE,$E2,$6B,$FE,$E1,$6B,$FE,$4E,$6B
	.dc.b	$FE,$5A,$9F,$E8,$F3,$D7,$C7,$F8,$1E,$77,$8D,$FC,$47,$97,$27,$C9
	.dc.b	$F1,$57,$0F,$17,$A9,$FE,$4A,$C8,$D2,$57,$F8,$4B,$CC,$72,$FC,$4E
	.dc.b	$7F,$E8,$8D,$9F,$F7,$26,$AF,$88,$FC,$C7,$0F,$E6,$7D,$79,$5F,$17
	.dc.b	$CA,$BA,$FC,$3F,$13,$C4,$49,$FF,$09,$6C,$8F,$F2,$4E,$7F,$32,$BF
	.dc.b	$73,$C7,$F7,$FD,$FF,$D3,$3F,$E5,$5F,$A3,$7C,$8D,$F1,$B0,$DC,$FF
	.dc.b	$9E,$FC,$DF,$F2,$DC,$7F,$C6,$4D,$7F,$83,$87,$0F,$F0,$A6,$F1,$B0
	.dc.b	$FD,$7C,$FF,$A3,$B3,$7C,$EF,$8F,$C6,$FD,$D3,$E5,$F6,$3F,$C4,$FF
	.dc.b	$43,$F8,$2E,$7F,$ED,$D8,$FE,$D1,$FF,$48,$46,$74,$F1,$BF,$82,$C8
	.dc.b	$D2,$7A,$FF,$6F,$E4,$BF,$28,$7C,$3F,$82,$B7,$FC,$8F,$E4,$BF,$86
	.dc.b	$18,$C3,$F9,$1B,$F2,$7E,$6A,$4F,$F7,$5C,$DF,$3B,$8B,$FE,$53,$E1
	.dc.b	$7F,$15,$31,$AF,$E1,$1F,$F2,$1F,$D4,$3A,$EF,$AF,$86,$E5,$FD,$BA
	.dc.b	$7F,$A5,$E0,$C3,$FD,$61,$B7,$F2,$58,$FE,$73,$3A,$9C,$FA,$3F,$E5
	.dc.b	$0A,$FF,$16,$D7,$D8,$6B,$FD,$01,$A9,$CF,$37,$D9,$F3,$3F,$DD,$FF
	.dc.b	$1D,$FE,$1D,$AF,$F6,$3E,$FF,$E7,$CC,$FF,$DF,$9C,$7F,$D2,$B0,$E1
	.dc.b	$18,$AD,$F9,$BC,$AB,$57,$C6,$58,$75,$C3,$F1,$17,$B9,$EF,$8D,$9C
	.dc.b	$E3,$D6,$5F,$4E,$ED,$7C,$7F,$F9,$CA,$30,$D8,$CC,$FF,$A2,$A7,$2F
	.dc.b	$FA,$33,$E4,$7F,$7F,$F5,$1F,$E7,$A6,$C6,$7C,$94,$E5,$B1,$CF,$BB
	.dc.b	$FD,$D6,$DC,$FC,$51,$88,$FF,$4E,$61,$FE,$6C,$9F,$E1,$5F,$4F,$FA
	.dc.b	$49,$F9,$11,$FC,$33,$CB,$FD,$7F,$07,$C5,$F0,$51,$8E,$3F,$A1,$87
	.dc.b	$1C,$3E,$5C,$63,$C1,$D4,$7F,$BF,$A7,$58,$7C,$F6,$D7,$2F,$13,$74
	.dc.b	$FA,$8F,$F3,$D3,$FE,$9C,$6B,$FE,$B2,$6B,$FE,$B7,$6B,$FC,$19,$FF
	.dc.b	$7C,$BF,$E4,$87,$8F,$4B,$93,$5F,$0B,$FC,$32,$31,$5F,$E1,$4A,$E2
	.dc.b	$FE,$86,$BA,$61,$B7,$B7,$C8,$59,$E0,$D7,$D5,$39,$DF,$F9,$BF,$CB
	.dc.b	$DE,$03,$4D,$75,$3E,$36,$93,$0F,$E0,$0F,$3D,$FF,$D2,$B1,$FE,$2B
	.dc.b	$FA,$7E,$84,$7F,$83,$7D,$47,$42,$38,$3F,$C0,$DF,$4D,$F6,$78,$3F
	.dc.b	$A4,$FA,$5F,$EC,$AD,$7F,$A6,$B7,$5F,$E9,$57,$FD,$45,$AF,$FB,$93
	.dc.b	$CA,$FE,$AE,$FF,$8F,$67,$DC,$7C,$38,$CC,$7C,$B3,$F7,$7E,$67,$C3
	.dc.b	$FA,$89,$1F,$C7,$63,$DD,$E3,$F8,$F6,$FF,$30,$4F,$E6,$23,$06,$7E
	.dc.b	$17,$E8,$7D,$47,$EB,$A7,$B0,$71,$AF,$FA,$A6,$D5,$F4,$FC,$99,$7F
	.dc.b	$E5,$57,$E9,$FB,$5F,$5F,$1F,$84,$49,$F8,$4F,$C5,$AF,$B8,$48,$FE
	.dc.b	$6D,$F3,$7F,$7F,$D3,$91,$FE,$61,$7E,$FF,$CC,$FE,$17,$F4,$3F,$F6
	.dc.b	$CF,$19,$A4,$EC,$FA,$8B,$E5,$78,$13,$0F,$F4,$CD,$7E,$4F,$E0,$DF
	.dc.b	$47,$B9,$FF,$5E,$5F,$F7,$26,$FB,$1B,$9D,$47,$D5,$BF,$17,$99,$E0
	.dc.b	$24,$7F,$84,$6F,$F9,$03,$FD,$D0,$FF,$3D,$E1,$E0,$F4,$7D,$9E,$54
	.dc.b	$9F,$F2,$A6,$3F,$EF,$77,$FD,$5F,$F3,$BF,$06,$A4,$60,$F3,$FB,$3F
	.dc.b	$00,$8F,$85,$F8,$7E,$CF,$C9,$24,$7F,$AE,$BC,$8F,$D8,$7C,$87,$F4
	.dc.b	$98,$F6,$68,$F8,$7F,$A0,$6B,$C8,$FA,$5E,$CF,$CE,$25,$7B,$DF,$A3
	.dc.b	$E1,$FD,$72,$47,$B3,$C0,$EF,$E5,$7D,$BF,$6A,$4F,$FB,$32,$3E,$F7
	.dc.b	$AD,$F6,$89,$1F,$2D,$FB,$BF,$53,$F8,$32,$47,$F9,$EF,$E5,$BF,$8C
	.dc.b	$F5,$FF,$BE,$BE,$13,$F1,$1E,$D4,$AF,$F6,$CC,$7F,$11,$FA,$8F,$BD
	.dc.b	$1F,$54,$38,$7E,$C3,$5F,$F1,$6B,$5F,$F5,$AB,$F7,$9F,$D5,$BD,$3C
	.dc.b	$7F,$E4,$F1,$D3,$BB,$C1,$DB,$F4,$3E,$57,$F3,$9E,$84,$AF,$F7,$6C
	.dc.b	$7F,$23,$FA,$77,$C9,$6E,$72,$7F,$B1,$31,$FE,$67,$E1,$2B,$73,$F9
	.dc.b	$AF,$47,$F5,$04,$7F,$DF,$3F,$2D,$D7,$E1,$3F,$02,$C8,$E9,$C3,$FE
	.dc.b	$07,$BF,$F2,$D4,$60,$FF,$69,$FA,$3F,$D0,$32,$FF,$1F,$F3,$7F,$19
	.dc.b	$3F,$9E,$7D,$4B,$EA,$7E,$77,$66,$19,$87,$33,$E7,$C7,$D3,$EC,$FA
	.dc.b	$51,$F3,$1A,$FD,$46,$F2,$FF,$9F,$23,$FC,$FD,$E0,$3E,$73,$F8,$37
	.dc.b	$74,$9F,$47,$1A,$B5,$1F,$87,$AF,$D0,$91,$C1,$C2,$3F,$99,$EB,$F3
	.dc.b	$BC,$1F,$F5,$94,$FF,$06,$F4,$3F,$53,$F0,$1F,$5F,$6B,$3C,$63,$0F
	.dc.b	$BF,$85,$D9,$FC,$6F,$CF,$93,$98,$FE,$C9,$D9,$FD,$DA,$3F,$E0,$0B
	.dc.b	$5F,$F1,$57,$8B,$F1,$23,$07,$EE,$74,$FC,$7B,$FA,$29,$E3,$3D,$3C
	.dc.b	$4F,$E5,$5E,$74,$9F,$E3,$EF,$F0,$2E,$DF,$EE,$5F,$49,$FC,$5F,$1F
	.dc.b	$C6,$F0,$FE,$32,$FF,$90,$36,$3F,$23,$3B,$EB,$FF,$48,$FD,$4B,$E5
	.dc.b	$3F,$BE,$7E,$DD,$FE,$2D,$BF,$CB,$B6,$B7,$3F,$E0,$3A,$FF,$70,$E3
	.dc.b	$F8,$57,$93,$63,$5F,$27,$FE,$07,$F4,$3F,$CC,$93,$F7,$B0,$7A,$1E
	.dc.b	$03,$5F,$13,$FE,$0F,$FA,$AF,$E7,$35,$FC,$CD,$AF,$FA,$A1,$A8,$38
	.dc.b	$6F,$93,$CF,$D5,$91,$F9,$B9,$FE,$01,$AF,$FA,$CB,$E5,$7F,$B5,$BF
	.dc.b	$E7,$1C,$BE,$A7,$79,$FA,$D6,$AD,$BE,$17,$E0,$BF,$DD,$9C,$BB,$F8
	.dc.b	$BF,$F1,$54,$6B,$87,$F8,$7F,$D5,$3E,$5F,$E8,$F4,$4C,$3F,$61,$39
	.dc.b	$7F,$65,$F5,$5F,$65,$F8,$11,$97,$FC,$0A,$DB,$FE,$37,$3B,$CF,$D5
	.dc.b	$E4,$FF,$88,$F4,$FF,$21,$E1,$FE,$D2,$9C,$BF,$EF,$38,$CB,$FD,$8B
	.dc.b	$D1,$7C,$DF,$E1,$AE,$4F,$F3,$B8,$FC,$4F,$7C,$8F,$BC,$FF,$58,$F9
	.dc.b	$C9,$1F,$B7,$6B,$FE,$29,$F7,$18,$3E,$6F,$F2,$7E,$9F,$9C,$95,$E5
	.dc.b	$7F,$88,$D7,$0E,$8D,$CF,$E3,$67,$FC,$11,$7F,$90,$8F,$F0,$16,$8F
	.dc.b	$47,$E1,$12,$3E,$03,$D2,$7E,$57,$37,$E0,$E3,$F4,$A7,$4E,$4F,$D3
	.dc.b	$F9,$72,$FF,$38,$FF,$76,$FA,$A9,$1F,$C5,$5A,$FF,$84,$FC,$1F,$F4
	.dc.b	$8B,$FA,$3F,$27,$EA,$11,$FA,$4F,$F9,$67,$C3,$FE,$7B,$1E,$A8,$FD
	.dc.b	$B9,$FE,$04,$F4,$FF,$C4,$DF,$4D,$AE,$9F,$DC,$F8,$DF,$BC,$33,$3F
	.dc.b	$E2,$BF,$AA,$7B,$3F,$29,$B8,$FF,$6D,$E5,$46,$23,$FC,$49,$18,$9E
	.dc.b	$1E,$D7,$36,$4F,$F5,$19,$E3,$3F,$E3,$9F,$A9,$FF,$3A,$71,$FE,$B7
	.dc.b	$A7,$F8,$79,$9F,$91,$D7,$C8,$7D,$CA,$34,$98,$F9,$73,$7A,$1D,$7E
	.dc.b	$62,$4F,$86,$72,$FF,$4C,$E4,$FF,$32,$B7,$E1,$BF,$94,$9F,$CA,$BE
	.dc.b	$1F,$FD,$3D,$F2,$DF,$D5,$79,$3F,$86,$7D,$53,$E7,$BF,$87,$60,$AF
	.dc.b	$C5,$3A,$8C,$C6,$21,$3C,$5F,$E2,$D1,$D5,$CC,$F0,$EB,$5F,$F1,$BF
	.dc.b	$3F,$85,$77,$EE,$F0,$C6,$30,$EC,$DB,$E3,$5F,$CB,$30,$E9,$87,$96
	.dc.b	$3C,$38,$7A,$B1,$E4,$DA,$DC,$F9,$0F,$3D,$F4,$C3,$71,$E5,$9F,$7D
	.dc.b	$79,$1E,$33,$E3,$C3,$55,$E0,$29,$EB,$2E,$3F,$0E,$78,$43,$C7,$FE
	.dc.b	$8A,$F8,$AB,$F9,$F4,$62,$3E,$96,$3F,$02,$31,$7F,$35,$D4,$FF,$45
	.dc.b	$F9,$5E,$38,$7F,$92,$5B,$1F,$E7,$F6,$BF,$79,$AF,$13,$E6,$1F,$82
	.dc.b	$F3,$7F,$DD,$B3,$F5,$A3,$4D,$8C,$E5,$DE,$78,$9F,$D5,$E3,$2F,$84
	.dc.b	$C8,$72,$F8,$CF,$A8,$71,$E7,$FB,$8E,$75,$F0,$46,$23,$C7,$5F,$E9
	.dc.b	$BA,$E7,$18,$3D,$53,$F5,$1F,$75,$72,$7B,$98,$AE,$53,$F8,$BC,$5F
	.dc.b	$50,$F8,$B4,$FF,$8E,$A3,$0D,$F6,$73,$87,$FA,$CE,$72,$FF,$41,$E4
	.dc.b	$7B,$1A,$E1,$DF,$6B,$AB,$FD,$BB,$97,$EF,$6B,$4F,$FA,$DA,$33,$87
	.dc.b	$F4,$B5,$F1,$9F,$E7,$C8,$D3,$55,$F5,$8C,$36,$BF,$91,$79,$BE,$C7
	.dc.b	$FC,$31,$E3,$78,$2F,$8C,$E5,$FE,$9B,$19,$A7,$D5,$7F,$92,$AF,$F9
	.dc.b	$22,$F2,$F4,$E6,$FF,$B1,$39,$9D,$CF,$92,$31,$3F,$CC,$9A,$FF,$14
	.dc.b	$AE,$F3,$EA,$F1,$BF,$B5,$5F,$F8,$BF,$C5,$F2,$46,$2F,$FB,$1C,$7A
	.dc.b	$F8,$FF,$CE,$19,$DF,$2F,$FB,$FE,$1F,$E6,$D8,$C1,$1F,$F0,$33,$F2
	.dc.b	$E7,$FC,$4C,$D9,$E3,$7F,$EE,$B9,$FF,$3E,$7D,$57,$7F,$28,$F8,$79
	.dc.b	$1F,$07,$98,$F1,$13,$2F,$F3,$F3,$70,$FF,$04,$7C,$FF,$F8,$CE,$7F
	.dc.b	$C9,$9E,$87,$D1,$8C,$4F,$F8,$02,$3D,$87,$94,$93,$FE,$1F,$7F,$03
	.dc.b	$2F,$A2,$F8,$5F,$03,$CD,$B5,$FE,$59,$8F,$71,$1F,$F1,$43,$F0,$F3
	.dc.b	$F9,$DD,$1F,$E7,$4C,$3C,$9F,$43,$F0,$5F,$9D,$E9,$CC,$3F,$D3,$51
	.dc.b	$73,$F6,$F9,$7F,$4F,$E9,$FD,$8D,$AF,$F3,$54,$7C,$F2,$3F,$E3,$97
	.dc.b	$F3,$D2,$F7,$3F,$D6,$13,$FE,$BB,$5E,$9F,$E6,$E7,$5E,$45,$9C,$D6
	.dc.b	$EB,$F0,$9E,$A6,$5B,$1F,$EC,$5F,$AA,$FC,$AE,$DF,$EE,$A7,$FC,$5D
	.dc.b	$19,$EB,$ED,$18,$CB,$FA,$FC,$FF,$6C,$8C,$57,$FB,$9B,$F6,$FF,$F3
	.dc.b	$E4,$F5,$F3,$3F,$2D,$18,$BE,$27,$FA,$07,$E1,$1F,$29,$EB,$5F,$29
	.dc.b	$63,$0C,$C3,$FD,$B3,$3A,$6C,$7C,$88,$C5,$7E,$67,$DA,$7D,$1F,$8B
	.dc.b	$CE,$FD,$44,$63,$0F,$F6,$CE,$5F,$CE,$5F,$F7,$47,$D5,$FE,$F7,$5F
	.dc.b	$E4,$38,$A3,$E3,$CF,$F8,$31,$D3,$77,$FA,$91,$A9,$FE,$C9,$7F,$EB
	.dc.b	$7A,$EB,$5F,$13,$E1,$FF,$75,$EE,$BE,$0B,$F5,$DD,$B8,$6D,$F9,$BE
	.dc.b	$1F,$E5,$98,$CB,$E5,$F8,$8C,$D3,$F8,$5F,$AA,$FD,$25,$FF,$AB,$F6
	.dc.b	$F5,$C7,$F8,$96,$30,$7F,$89,$9D,$E1,$F5,$83,$B7,$0D,$7F,$C5,$A6
	.dc.b	$BD,$3F,$CC,$C6,$63,$A4,$7E,$73,$EA,$FF,$9F,$DF,$E1,$FE,$1B,$F3
	.dc.b	$51,$8C,$3F,$C0,$F9,$7F,$02,$AF,$EA,$39,$B6,$F9,$8B,$C3,$78,$5F
	.dc.b	$2B,$8B,$F8,$B3,$31,$FB,$BE,$25,$DB,$E3,$E7,$EE,$CF,$FA,$BF,$8A
	.dc.b	$F9,$9F,$E3,$DE,$1F,$F1,$36,$E3,$F1,$B0,$FF,$0F,$68,$AF,$CE,$F0
	.dc.b	$EA,$7A,$BF,$47,$DB,$F8,$64,$BF,$8F,$3F,$EE,$1E,$34,$FC,$8F,$73
	.dc.b	$F7,$9F,$57,$FA,$4B,$F0,$5F,$D2,$9C,$F8,$1F,$C0,$7A,$5E,$C5,$33
	.dc.b	$7C,$23,$2F,$A7,$97,$E5,$7C,$17,$F3,$BF,$1B,$F2,$2D,$9F,$D1,$F9
	.dc.b	$7F,$B8,$31,$9F,$C8,$7F,$2F,$E8,$7F,$51,$F3,$BF,$79,$CD,$F5,$BA
	.dc.b	$FF,$A4,$BF,$62,$33,$87,$F9,$09,$B8,$7E,$93,$9F,$93,$FC,$19,$FE
	.dc.b	$39,$F5,$7F,$E0,$EB,$FD,$8B,$F0,$E7,$F2,$8E,$6F,$FC,$FA,$E6,$77
	.dc.b	$C1,$FC,$AA,$38,$57,$E5,$9A,$F0,$8F,$EC,$58,$7F,$01,$8F,$C2,$F9
	.dc.b	$BE,$DF,$17,$F8,$CB,$FC,$6B,$E0,$E6,$5F,$EA,$7F,$AB,$F1,$5F,$EF
	.dc.b	$A7,$F8,$FF,$C1,$FF,$6D,$EE,$34,$97,$F8,$2E,$4F,$ED,$3E,$37,$F2
	.dc.b	$6F,$27,$E6,$36,$7F,$8A,$FD,$37,$F2,$A3,$FC,$2D,$9F,$F7,$08,$C9
	.dc.b	$FC,$6B,$E9,$7C,$04,$7B,$64,$8F,$7F,$91,$FC,$D3,$E9,$3E,$D3,$6B
	.dc.b	$F9,$1F,$D1,$FF,$38,$3E,$F7,$95,$63,$F9,$3F,$D1,$FB,$48,$F6,$7E
	.dc.b	$5F,$91,$FD,$03,$C8,$FC,$33,$67,$FD,$9D,$F5,$4F,$86,$FE,$8B,$E3
	.dc.b	$7F,$8C,$63,$13,$FC,$D2,$3F,$A4,$F8,$BF,$2C,$91,$EC,$7F,$C8,$33
	.dc.b	$FC,$CB,$9F,$F9,$7F,$C0,$F3,$F8,$DF,$EA,$2F,$F8,$9F,$A3,$FC,$81
	.dc.b	$DF,$FC,$31,$B8,$FE,$57,$CA,$D3,$BF,$63,$FC,$4D,$3F,$DC,$E3,$EC
	.dc.b	$37,$1F,$F1,$33,$5F,$F6,$7F,$5F,$E0,$3E,$4B,$C0,$7E,$9F,$C6,$B7
	.dc.b	$FB,$3B,$FD,$AD,$FF,$3A,$6E,$5F,$A9,$F1,$3D,$1E,$57,$A7,$E1,$FC
	.dc.b	$64,$C3,$C6,$F9,$BC,$5F,$71,$FB,$5F,$0E,$CF,$F0,$1F,$A6,$EA,$FF
	.dc.b	$B9,$3E,$67,$F9,$4B,$9C,$3F,$C1,$91,$8D,$3F,$A6,$D7,$FA,$01,$F6
	.dc.b	$B5,$FF,$5E,$B5,$FF,$61,$37,$7F,$F2,$2F,$8B,$F4,$7D,$9F,$E1,$0D
	.dc.b	$C7,$F8,$EF,$CB,$FF,$61,$7E,$27,$D1,$FF,$87,$63,$18,$FC,$58,$FB
	.dc.b	$47,$D4,$24,$FD,$47,$F7,$7C,$3F,$C5,$4F,$C6,$7A,$B6,$7E,$87,$D1
	.dc.b	$FE,$0C,$C5,$FD,$17,$F8,$E7,$95,$27,$E4,$7D,$1D,$AF,$AD,$F3,$DF
	.dc.b	$8B,$31,$CD,$F8,$B7,$FC,$69,$F5,$5F,$CF,$2F,$F9,$CE,$5F,$C5,$55
	.dc.b	$C9,$D3,$FD,$BE,$74,$E9,$AF,$F5,$ED,$FF,$C8,$11,$83,$FC,$3F,$87
	.dc.b	$F8,$59,$FF,$10,$CF,$F7,$46,$BE,$03,$95,$FF,$52,$69,$FE,$7F,$E1
	.dc.b	$7D,$27,$FB,$03,$E4,$BF,$C2,$93,$F7,$23,$1E,$F7,$E0,$3E,$5A,$CE
	.dc.b	$70,$FF,$7C,$4F,$E1,$2B,$3C,$BF,$C8,$BD,$3B,$97,$6B,$DE,$0A,$D4
	.dc.b	$62,$1F,$53,$FD,$7B,$D4,$F4,$36,$3F,$B5,$B5,$FF,$09,$B5,$F8,$1F
	.dc.b	$2B,$FE,$68,$FA,$AF,$80,$D8,$FE,$E6,$D7,$FC,$36,$D7,$E4,$4F,$FB
	.dc.b	$9F,$0F,$0F,$7F,$9E,$7C,$2D,$7C,$2D,$7E,$FC,$7F,$42,$E5,$FA,$5C
	.dc.b	$1F,$32,$3F,$A2,$F0,$63,$83,$7D,$08,$C6,$1F,$16,$18,$62,$FA,$4F
	.dc.b	$CA,$6A,$C3,$B5,$70,$8D,$2F,$99,$E4,$FE,$02,$3F,$8E,$F1,$7E,$16
	.dc.b	$FF,$AC,$B6,$75,$3F,$C8,$3E,$9B,$F1,$9C,$BF,$CA,$EA,$F1,$7A,$9B
	.dc.b	$5E,$9A,$B8,$7F,$64,$6E,$06,$98,$FA,$6C,$F7,$EB,$E8,$A6,$1F,$CA
	.dc.b	$63,$2D,$21,$FC,$AF,$E7,$3C,$3F,$2A,$F8,$F5,$F8,$3E,$66,$5F,$F8
	.dc.b	$F6,$B5,$3C,$2B,$73,$96,$BF,$E9,$A9,$DB,$FE,$2D,$FD,$C3,$F0,$5C
	.dc.b	$1F,$E5,$87,$FA,$F4,$7F,$7F,$6B,$FE,$ED,$6B,$FD,$21,$AF,$E2,$17
	.dc.b	$C7,$7C,$87,$8C,$F5,$F5,$31,$39,$E7,$D7,$CF,$4C,$3C,$91,$C7,$D2
	.dc.b	$FB,$98,$9A,$49,$F3,$7C,$67,$72,$3E,$DF,$90,$F1,$6B,$E3,$F9,$29
	.dc.b	$3F,$06,$3C,$DF,$0D,$D8,$91,$F0,$5F,$E2,$1B,$BF,$07,$63,$F5,$9B
	.dc.b	$5F,$4D,$AF,$FA,$7B,$37,$C4,$EF,$19,$3B,$7E,$37,$CD,$FC,$7F,$99
	.dc.b	$F7,$92,$7E,$93,$F0,$BE,$2F,$E6,$7C,$BC,$9F,$B5,$1F,$3F,$E5,$FC
	.dc.b	$97,$27,$EF,$47,$D3,$F8,$FF,$96,$48,$FB,$CF,$CE,$F3,$7E,$C1,$23
	.dc.b	$F0,$DF,$25,$F6,$FE,$42,$47,$E2,$5F,$D2,$FE,$E3,$FE,$AC,$EB,$FF
	.dc.b	$50,$AF,$3B,$F6,$8D,$AF,$F2,$53,$5F,$F5,$9F,$88,$8C,$47,$F8,$51
	.dc.b	$7C,$87,$62,$47,$FA,$16,$3F,$D9,$2D,$F8,$9F,$1F,$C3,$C8,$C1,$BF
	.dc.b	$83,$F1,$91,$F1,$BF,$7B,$F0,$87,$C1,$7E,$C9,$B3,$FA,$F6,$BF,$ED
	.dc.b	$AC,$2E,$8F,$33,$FB,$3F,$62,$DF,$01,$CC,$98,$7A,$7E,$0B,$E4,$C7
	.dc.b	$C6,$24,$7C,$67,$E4,$FC,$07,$CE,$24,$7C,$1F,$3B,$F8,$BA,$FA,$8F
	.dc.b	$F8,$03,$9F,$8E,$3F,$BC,$FE,$02,$F5,$B6,$D8,$FF,$5B,$D7,$0D,$7F
	.dc.b	$B8,$2F,$86,$FE,$E1,$3D,$5B,$19,$8F,$EE,$8D,$7F,$D8,$FF,$80,$BE
	.dc.b	$37,$CB,$39,$CB,$C9,$5C,$CF,$F6,$3D,$6D,$FB,$9F,$05,$F8,$CF,$94
	.dc.b	$91,$83,$F0,$1F,$25,$F9,$09,$91,$83,$F0,$9F,$25,$F9,$4F,$91,$91
	.dc.b	$83,$F1,$19,$FE,$62,$64,$60,$FC,$67,$C8,$FE,$70,$8F,$E3,$FE,$47
	.dc.b	$F3,$FF,$23,$FA,$06,$47,$E4,$E3,$F4,$7F,$21,$A8,$C1,$F9,$77,$E1
	.dc.b	$FC,$3F,$CA,$24,$7C,$A7,$E6,$FC,$16,$A3,$EA,$C7,$CA,$8F,$EA,$7F
	.dc.b	$B9,$7D,$79,$FF,$09,$3F,$37,$93,$FB,$CB,$63,$31,$C3,$E4,$78,$B2
	.dc.b	$3F,$DD,$D2,$9F,$AF,$EF,$7E,$1B,$E3,$65,$7E,$1A,$3E,$DF,$BF,$F8
	.dc.b	$22,$47,$E0,$9F,$C3,$F4,$BE,$37,$73,$F8,$E9,$FC,$8C,$7A,$BE,$43
	.dc.b	$D7,$D9,$2B,$C7,$1E,$EF,$91,$D4,$60,$F3,$BF,$A0,$87,$9F,$F9,$66
	.dc.b	$CF,$F9,$1D,$AF,$F9,$2C,$FF,$2D,$B8,$70,$BC,$FE,$93,$27,$FC,$C3
	.dc.b	$DB,$F9,$8A,$D4,$63,$83,$FC,$AD,$F5,$5F,$ED,$78,$F9,$7F,$1F,$F5
	.dc.b	$8F,$DA,$C6,$61,$7F,$A2,$87,$4F,$EB,$9B,$3F,$E5,$F6,$BE,$90,$E1
	.dc.b	$F6,$7E,$7D,$FC,$C7,$0F,$1D,$78,$0F,$E0,$B1,$99,$FF,$3D,$FE,$1E
	.dc.b	$F7,$FF,$7A,$D8,$FF,$37,$35,$F0,$35,$F9,$4B,$DF,$FE,$22,$EF,$2F
	.dc.b	$DE,$B7,$1F,$F3,$D3,$BA,$F9,$07,$FA,$33,$E0,$FF,$C6,$AF,$D5,$6B
	.dc.b	$FE,$81,$6B,$BF,$8A,$FE,$27,$1F,$E3,$18,$E6,$1F,$F1,$FD,$FF,$0E
	.dc.b	$6D,$F3,$F8,$AF,$D4,$FC,$9F,$F9,$1F,$AC,$D2,$61,$F3,$FD,$64,$BA
	.dc.b	$F8,$CF,$EA,$2B,$F6,$71,$FB,$4E,$89,$3F,$AC,$7F,$6F,$DD,$23,$F5
	.dc.b	$ED,$7F,$07,$97,$EE,$3B,$BF,$7B,$F1,$DE,$52,$4F,$8F,$2F,$F3,$73
	.dc.b	$7E,$37,$6F,$EE,$A7,$F8,$2C,$7F,$06,$E0,$F7,$12,$79,$3F,$07,$6D
	.dc.b	$47,$F0,$28,$FE,$1B,$1F,$C3,$BE,$36,$4F,$F0,$67,$F8,$8F,$C6,$C8
	.dc.b	$C1,$0F,$59,$FC,$51,$FE,$2F,$F1,$BF,$3C,$91,$EB,$7E,$97,$85,$A8
	.dc.b	$FE,$27,$1F,$C7,$A3,$FC,$C7,$F4,$DF,$46,$7F,$CC,$0F,$F0,$D6,$F0
	.dc.b	$6B,$87,$7C,$3F,$E3,$8A,$34,$95,$FA,$BF,$59,$3F,$E8,$F8,$FF,$33
	.dc.b	$3C,$1E,$F3,$E2,$E1,$FF,$3A,$FD,$37,$F8,$03,$8F,$FB,$FA,$F6,$B2
	.dc.b	$6B,$DA,$DF,$B1,$FE,$4A,$93,$49,$8F,$DD,$F1,$FF,$9E,$91,$FD,$3F
	.dc.b	$4F,$F9,$28,$FF,$49,$FC,$3F,$F9,$B3,$2F,$F5,$BB,$FE,$7A,$6C,$6B
	.dc.b	$E3,$FF,$C0,$BF,$25,$FE,$8B,$EE,$FF,$7D,$FD,$3F,$9F,$1F,$F7,$EC
	.dc.b	$BE,$3B,$FD,$E1,$39,$94,$E5,$E3,$BF,$FB,$5B,$E2,$FF,$8D,$C2,$B8
	.dc.b	$57,$78,$63,$8C,$3F,$DF,$8D,$81,$AC,$3A,$63,$89,$E7,$8F,$6C,$78
	.dc.b	$5F,$7B,$FF,$73,$E1,$E8,$78,$C6,$4F,$6D,$79,$AF,$89,$27,$6F,$6C
	.dc.b	$3C,$11,$D2,$A9,$F1,$AF,$D3,$1C,$0E,$F9,$7F,$A9,$B2,$F8,$0D,$C7
	.dc.b	$E2,$B7,$0F,$F7,$6E,$3F,$CE,$FE,$4A,$E5,$FD,$55,$AF,$FB,$C0,$57
	.dc.b	$FA,$BA,$33,$1F,$21,$F2,$BF,$2B,$E1,$7F,$C8,$15,$F6,$63,$47,$CD
	.dc.b	$3E,$79,$FE,$75,$D7,$8B,$A8,$E1,$3C,$B3,$F3,$4F,$13,$B5,$70,$7F
	.dc.b	$DF,$07,$CF,$32,$53,$C3,$AF,$48,$FC,$4C,$7B,$CE,$87,$8C,$A7,$4B
	.dc.b	$E0,$E3,$0E,$2F,$6A,$F7,$3E,$27,$67,$E7,$71,$FF,$58,$FC,$97,$F9
	.dc.b	$8F,$3F,$F3,$24,$7F,$AD,$3E,$4B,$19,$7E,$AA,$AE,$1F,$43,$81,$58
	.dc.b	$BF,$F5,$DF,$8B,$67,$FA,$86,$1F,$B6,$76,$D8,$E7,$D2,$F8,$CF,$F3
	.dc.b	$64,$FF,$11,$6E,$3F,$E1,$59,$D7,$C0,$7F,$99,$A3,$FD,$17,$3A,$7E
	.dc.b	$CC,$F3,$F5,$7A,$CB,$D0,$B1,$87,$C3,$F4,$FF,$BD,$C7,$FB,$8D,$A1
	.dc.b	$CB,$AE,$7D,$B2,$BF,$93,$4E,$9B,$1E,$5C,$97,$CE,$30,$F5,$F3,$7C
	.dc.b	$3E,$6F,$89,$B7,$F1,$FE,$4F,$E5,$7C,$9F,$CB,$6C,$7C,$B8,$ED,$C5
	.dc.b	$FE,$EA,$6F,$17,$FB,$CA,$7F,$BB,$37,$3F,$F6,$5B,$EF,$7F,$A9,$D7
	.dc.b	$FB,$CB,$3F,$B1,$AF,$D9,$65,$FD,$97,$2D,$3E,$E6,$5F,$7F,$3F,$C0
	.dc.b	$36,$3F,$01,$1E,$42,$3E,$47,$CF,$8F,$E1,$FE,$3F,$F1,$0D,$9F,$C4
	.dc.b	$47,$B0,$8F,$B1,$F7,$FC,$8F,$E3,$CF,$C8,$71,$59,$FC,$87,$C8,$FE
	.dc.b	$4C,$C6,$BF,$03,$04,$7E,$5F,$E4,$7F,$31,$EC,$5B,$FC,$C3,$87,$E5
	.dc.b	$BF,$38,$8F,$F9,$46,$DF,$8F,$3E,$17,$F4,$F1,$8F,$62,$CE,$6B,$FC
	.dc.b	$6A,$FF,$97,$9B,$1F,$89,$9E,$11,$D6,$75,$1F,$E5,$F6,$BF,$E7,$9F
	.dc.b	$A5,$7C,$9F,$81,$7C,$9D,$8C,$32,$FD,$13,$A6,$C7,$86,$31,$5F,$E6
	.dc.b	$DF,$B7,$7C,$37,$B3,$C1,$7A,$FE,$E9,$CE,$3F,$E5,$09,$CB,$67,$D9
	.dc.b	$18,$BF,$F3,$C3,$5F,$F4,$93,$5F,$E6,$6B,$D7,$FE,$19,$F0,$1E,$16
	.dc.b	$DF,$F9,$7B,$E9,$5C,$9F,$41,$73,$D8,$C3,$27,$EC,$4E,$9B,$1F,$46
	.dc.b	$31,$8F,$EA,$37,$FE,$39,$BF,$F1,$D6,$DF,$F1,$D8,$FE,$45,$FB,$9F
	.dc.b	$F7,$88,$C1,$FE,$16,$71,$5F,$E4,$21,$F7,$1F,$E6,$99,$FF,$3B,$FD
	.dc.b	$2B,$E4,$FF,$13,$39,$AE,$13,$9A,$DD,$7E,$1E,$31,$3F,$EA,$36,$BF
	.dc.b	$EB,$D6,$BF,$D7,$97,$C9,$FE,$5B,$E9,$DF,$17,$F9,$89,$CE,$1F,$E8
	.dc.b	$A9,$CB,$EB,$F4,$BD,$FF,$33,$F0,$1B,$1E,$F8,$F8,$7F,$45,$8C,$BF
	.dc.b	$87,$EA,$8E,$DD,$1D,$FE,$3E,$E1,$F0,$9C,$3F,$AD,$CD,$1E,$F2,$47
	.dc.b	$FA,$4E,$DF,$35,$F4,$BE,$6B,$E9,$B6,$7E,$94,$7D,$6F,$5F,$19,$F9
	.dc.b	$1F,$B3,$EE,$4A,$FB,$11,$67,$D3,$ED,$FD,$EE,$3F,$BE,$DA,$FB,$D1
	.dc.b	$F8,$2E,$3C,$57,$C0,$7F,$0B,$F3,$1F,$86,$F9,$5B,$7F,$85,$70,$FE
	.dc.b	$1F,$E5,$AD,$FC,$87,$EB,$67,$FC,$33,$E9,$5F,$23,$8B,$FF,$5F,$C6
	.dc.b	$1B,$1F,$C2,$E3,$17,$FE,$E7,$6B,$FF,$00,$5A,$FF,$88,$9F,$F6,$0C
	.dc.b	$FF,$B2,$BC,$87,$C9,$78,$7E,$3F,$F8,$EC,$62,$7F,$3D,$E5,$7E,$88
	.dc.b	$FD,$1F,$95,$63,$F4,$5F,$2F,$FD,$63,$98,$D2,$47,$FA,$47,$E2,$3F
	.dc.b	$9A,$BF,$DD,$1B,$A7,$E2,$BE,$5F,$FC,$13,$18,$F8,$05,$7F,$9A,$3D
	.dc.b	$AF,$F6,$A6,$9A,$3F,$6B,$DB,$67,$F6,$9D,$9E,$32,$3F,$8B,$7C,$DF
	.dc.b	$2F,$FB,$CF,$0F,$F7,$AD,$9F,$F4,$8F,$EE,$9E,$06,$9F,$E2,$37,$FC
	.dc.b	$33,$97,$F8,$B1,$D6,$BF,$DC,$17,$BF,$FD,$E9,$B5,$FE,$ED,$6C,$FF
	.dc.b	$95,$0F,$F2,$FB,$87,$FC,$AF,$ED,$A3,$FC,$3D,$3E,$4F,$7F,$C5,$F2
	.dc.b	$FE,$37,$C9,$CF,$27,$D3,$D3,$71,$F2,$74,$78,$C9,$3E,$37,$CD,$C7
	.dc.b	$FE,$38,$FC,$55,$DB,$EF,$8C,$3F,$07,$E3,$BE,$17,$C6,$DC,$3E,$3B
	.dc.b	$F1,$7E,$37,$E0,$12,$3E,$03,$FE,$26,$C1,$D3,$DC,$F9,$9F,$29,$6B
	.dc.b	$FC,$6F,$3F,$D4,$34,$FF,$76,$F5,$23,$1C,$BF,$47,$E5,$3D,$1D,$5F
	.dc.b	$57,$D8,$FA,$04,$C7,$E8,$3F,$4F,$D6,$FB,$5F,$29,$F6,$DB,$3F,$80
	.dc.b	$F9,$4F,$BC,$7D,$FF,$94,$B1,$F8,$4F,$94,$FB,$84,$7E,$E3,$F7,$FE
	.dc.b	$4B,$FC,$D1,$FB,$AF,$FA,$6A,$7F,$17,$18,$EE,$CF,$C9,$6B,$DF,$B5
	.dc.b	$FE,$E0,$B4,$66,$3F,$0D,$C7,$FE,$A4,$9C,$C6,$2D,$18,$9F,$C6,$46
	.dc.b	$2B,$FC,$AB,$F8,$8B,$C0,$FF,$3A,$B7,$0F,$07,$C8,$E2,$64,$78,$FE
	.dc.b	$46,$E3,$FE,$4C,$CB,$FD,$02,$DF,$8F,$F7,$9F,$03,$E3,$ED,$FA,$E3
	.dc.b	$E1,$FC,$7E,$2B,$93,$F1,$BE,$3E,$57,$C1,$F8,$FB,$5D,$E7,$E5,$FC
	.dc.b	$7E,$3A,$FF,$D3,$79,$A3,$38,$7C,$38,$C3,$63,$FC,$D7,$18,$BF,$8A
	.dc.b	$FF,$83,$27,$FD,$99,$B7,$FB,$D3,$87,$E5,$FC,$3F,$F1,$0B,$7C,$9F
	.dc.b	$FA,$91,$FA,$5F,$11,$98,$C5,$70,$C3,$9D,$7D,$7B,$D6,$1D,$71,$7C
	.dc.b	$86,$2B,$57,$C3,$0F,$B7,$39,$97,$D1,$7F,$A7,$FE,$17,$FC,$A8,$D8
	.dc.b	$FC,$2B,$5F,$E1,$8D,$7F,$DC,$3F,$2A,$E1,$7D,$1F,$F9,$76,$30,$D8
	.dc.b	$CC,$FE,$42,$72,$F6,$E7,$FF,$7A,$B7,$2F,$1C,$62,$3C,$2D,$7E,$67
	.dc.b	$BC,$F1,$FF,$DE,$7C,$3F,$22,$31,$3F,$05,$AF,$D0,$E0,$E3,$E4,$F0
	.dc.b	$6C,$6E,$F0,$B1,$5F,$2C,$3B,$61,$A9,$FD,$3F,$C9,$FF,$9F,$5B,$7F
	.dc.b	$89,$8C,$71,$7F,$B6,$63,$F5,$D8,$7F,$99,$1F,$DB,$49,$A5,$E6,$FF
	.dc.b	$77,$CE,$E3,$FD,$66,$D8,$CF,$8B,$C6,$72,$D8,$E3,$E2,$5E,$5F,$93
	.dc.b	$E1,$DB,$FE,$05,$F2,$77,$4F,$E0,$2D,$C9,$54,$95,$E5,$BE,$6D,$3E
	.dc.b	$44,$BC,$5E,$5E,$27,$5D,$9F,$27,$4C,$DF,$1C,$FE,$7D,$62,$3D,$EF
	.dc.b	$9F,$CB,$FF,$77,$E7,$CF,$DA,$F9,$A4,$AF,$DC,$3F,$4F,$E8,$9E,$46
	.dc.b	$E7,$F9,$24,$FB,$E7,$31,$E1,$5E,$37,$9F,$E4,$64,$E6,$38,$74,$F5
	.dc.b	$24,$7F,$29,$AF,$DE,$3F,$CA,$17,$8D,$DC,$91,$DF,$C7,$F1,$77,$7C
	.dc.b	$5F,$90,$93,$E3,$8F,$91,$F2,$1F,$8A,$24,$7F,$15,$7F,$21,$D9,$FE
	.dc.b	$01,$EA,$FE,$1D,$C6,$8C,$74,$7F,$13,$6D,$B8,$DF,$46,$D3,$F0,$54
	.dc.b	$F1,$29,$2F,$EF,$5F,$0F,$27,$7F,$1F,$E2,$24,$FF,$4B,$B4,$7D,$3E
	.dc.b	$DE,$BF,$1F,$2B,$E8,$47,$DD,$CF,$FD,$67,$1F,$D5,$B1,$F2,$F9,$5F
	.dc.b	$EE,$96,$C6,$67,$6F,$F5,$6C,$7F,$CF,$AF,$87,$E3,$BE,$29,$1F,$27
	.dc.b	$C7,$7B,$88,$FE,$07,$A7,$E0,$FC,$74,$8F,$CA,$5A,$3F,$11,$C9,$F1
	.dc.b	$BE,$2B,$E8,$92,$BE,$03,$F5,$3E,$3B,$E5,$92,$3F,$3D,$F4,$7F,$11
	.dc.b	$B9,$7D,$0F,$8F,$91,$83,$C3,$F1,$FF,$78,$8F,$CD,$7F,$03,$E3,$FD
	.dc.b	$82,$47,$D6,$FA,$36,$5F,$73,$E0,$24,$FB,$1F,$D6,$4F,$FB,$2E,$3D
	.dc.b	$FF,$21,$F8,$C2,$3F,$6D,$FC,$7F,$C7,$FE,$18,$91,$EC,$C7,$F8,$62
	.dc.b	$FA,$1F,$F0,$33,$FC,$14,$7E,$CB,$FC,$F7,$C0,$FE,$C9,$18,$F0,$38
	.dc.b	$D7,$08,$FF,$71,$7D,$26,$9F,$F2,$F3,$5F,$E4,$0D,$7F,$95,$35,$2F
	.dc.b	$9C,$F2,$BE,$0B,$F0,$B7,$C2,$E7,$3F,$C8,$F8,$DA,$FE,$93,$E4,$E6
	.dc.b	$3F,$6A,$34,$47,$07,$DE,$E0,$FD,$5F,$C9,$48,$FC,$5B,$D4,$8F,$E1
	.dc.b	$1A,$B3,$FD,$DE,$0C,$3F,$6B,$F0,$D2,$BF,$2A,$F9,$FE,$12,$47,$E3
	.dc.b	$9A,$F3,$D3,$FC,$29,$7F,$EE,$68,$FC,$9F,$C8,$7B,$FE,$4A,$46,$0F
	.dc.b	$B1,$BB,$E2,$BE,$EF,$DB,$BD,$1A,$7C,$C8,$F9,$BF,$17,$FE,$4E,$BF
	.dc.b	$A6,$D7,$E9,$B6,$CF,$F6,$9F,$A1,$FA,$D3,$FB,$0D,$7F,$89,$9F,$D5
	.dc.b	$63,$31,$EB,$3F,$BB,$3F,$65,$BC,$9B,$38,$6D,$F8,$CF,$8D,$FE,$3B
	.dc.b	$AC,$AF,$CA,$73,$27,$F0,$BF,$15,$E5,$24,$FA,$F4,$47,$A7,$D8,$DC
	.dc.b	$FC,$68,$FC,$6C,$7F,$2A,$E1,$93,$FC,$29,$F5,$7C,$0C,$8F,$EC,$1A
	.dc.b	$FE,$7F,$9B,$F9,$CB,$75,$FF,$25,$2F,$91,$FD,$69,$FD,$AA,$BF,$59
	.dc.b	$F8,$57,$CA,$6F,$16,$BD,$CD,$F2,$FD,$1B,$FB,$15,$F5,$5A,$FF,$A1
	.dc.b	$1A,$FF,$B2,$30,$7C,$8F,$F4,$4F,$95,$93,$E8,$8F,$DF,$FD,$2B,$C3
	.dc.b	$FB,$CD,$9F,$C1,$35,$FF,$48,$B5,$FF,$6C,$3F,$9D,$7F,$83,$3B,$DB
	.dc.b	$F8,$2B,$76,$FF,$0F,$3B,$9F,$C1,$0F,$A4,$7A,$DF,$E1,$BF,$94,$FC
	.dc.b	$F4,$FE,$1F,$C4,$46,$31,$FE,$23,$7C,$2B,$F8,$B3,$FE,$5D,$7F,$34
	.dc.b	$D7,$FD,$52,$D7,$FD,$E0,$D7,$FC,$65,$EF,$BE,$4F,$FC,$59,$FB,$B7
	.dc.b	$E5,$71,$3F,$97,$AF,$F0,$7C,$FF,$28,$AF,$F1,$F4,$7F,$BE,$DA,$F4
	.dc.b	$F8,$7F,$D5,$46,$23,$FC,$AD,$5F,$A4,$7F,$DF,$ED,$7F,$D2,$4D,$77
	.dc.b	$DF,$FA,$FC,$FC,$AC,$8F,$C9,$C7,$A3,$E6,$24,$7F,$8D,$69,$1F,$B3
	.dc.b	$F0,$A4,$FE,$69,$AF,$CF,$E4,$FC,$A3,$F4,$6B,$F3,$4F,$FB,$3D,$AF
	.dc.b	$FB,$39,$AF,$AB,$97,$F8,$27,$C5,$FC,$A6,$47,$EA,$23,$E6,$F5,$FF
	.dc.b	$9C,$6B,$FA,$44,$FC,$28,$C1,$FC,$37,$5D,$C7,$AE,$3F,$86,$47,$D6
	.dc.b	$F9,$79,$3E,$F7,$F8,$87,$CB,$48,$C1,$0F,$73,$F7,$BA,$3C,$44,$8F
	.dc.b	$73,$FC,$77,$E2,$35,$1F,$2E,$3F,$8E,$47,$F9,$77,$E9,$7F,$97,$4F
	.dc.b	$F8,$BB,$DF,$FE,$28,$DF,$17,$5C,$1B,$7F,$C4,$95,$A9,$FF,$41,$1F
	.dc.b	$E3,$D7,$0F,$FA,$17,$DE,$46,$23,$FD,$21,$18,$F0,$7A,$47,$F8,$FA
	.dc.b	$7A,$75,$7F,$B6,$6B,$53,$FE,$42,$FA,$5F,$F1,$23,$FE,$44,$FD,$C7
	.dc.b	$F0,$AD,$E6,$FF,$3C,$4E,$A3,$F9,$C5,$23,$FC,$2B,$1F,$E3,$F6,$BE
	.dc.b	$A8,$D7,$AD,$FE,$41,$9C,$CF,$AD,$7B,$5D,$89,$1F,$CB,$3C,$8F,$F4
	.dc.b	$E7,$CA,$FF,$83,$FC,$3F,$E5,$71,$FD,$2C,$CF,$81,$FE,$18,$F9,$2F
	.dc.b	$E1,$3B,$7F,$AC,$FE,$9B,$D1,$0F,$98,$FF,$57,$46,$1B,$F0,$1F,$EE
	.dc.b	$4F,$1B,$FB,$C6,$8A,$E3,$18,$F8,$FE,$13,$FE,$BF,$6B,$FE,$96,$AE
	.dc.b	$58,$E5,$B1,$CD,$FF,$78,$E1,$E0,$8F,$F6,$46,$3E,$57,$D1,$3C,$31
	.dc.b	$F0,$DF,$39,$DD,$FF,$B1,$F0,$F0,$9A,$8F,$11,$D3,$1F,$46,$1E,$08
	.dc.b	$E4,$31,$F0,$3C,$0C,$4F,$07,$96,$1F,$ED,$DC,$BF,$D5,$F1,$A8,$C6
	.dc.b	$1E,$E9,$CD,$FD,$2A,$FF,$79,$4F,$E7,$17,$CA,$70,$7F,$3E,$D7,$FD
	.dc.b	$C4,$D7,$FD,$D8,$D7,$89,$E7,$3F,$D3,$DA,$FB,$1D,$99,$F8,$A4,$66
	.dc.b	$38,$D7,$E6,$DE,$27,$33,$39,$FE,$CA,$B8,$C6,$6F,$EA,$BF,$EF,$78
	.dc.b	$CE,$BC,$E3,$CA,$F5,$8C,$D7,$C1,$AF,$34,$74,$BE,$11,$8C,$3C,$CF
	.dc.b	$6C,$07,$9B,$0D,$BF,$8B,$3C,$53,$E4,$C9,$BB,$E4,$3D,$3A,$FF,$B3
	.dc.b	$23,$FD,$CB,$5F,$EF,$77,$4D,$F2,$36,$FF,$0C,$F9,$4F,$F3,$83,$FE
	.dc.b	$C8,$6E,$3F,$DC,$EB,$9B,$FC,$12,$33,$7F,$C3,$DF,$E9,$F5,$FE,$C3
	.dc.b	$6C,$FF,$05,$6B,$FC,$75,$AF,$87,$C6,$F1,$BF,$ED,$FE,$97,$85,$8E
	.dc.b	$95,$FF,$8B,$70,$FE,$35,$39,$FA,$6F,$F0,$5C,$7F,$B8,$17,$C7,$E7
	.dc.b	$9B,$4C,$9F,$F6,$C4,$E9,$B1,$FE,$97,$C1,$39,$CB,$F8,$F2,$F9,$9F
	.dc.b	$27,$85,$FE,$6F,$C3,$C1,$18,$F0,$BD,$7C,$7F,$5B,$E4,$2F,$93,$FE
	.dc.b	$1C,$BF,$A9,$18,$F3,$24,$7F,$6F,$7F,$D5,$15,$FE,$12,$6D,$71,$CF
	.dc.b	$FA,$4C,$FE,$0B,$6E,$BE,$2F,$83,$97,$E1,$D7,$F7,$B8,$CE,$3F,$E0
	.dc.b	$57,$55,$89,$FB,$DA,$79,$B4,$E4,$DB,$FF,$06,$47,$88,$8F,$F8,$15
	.dc.b	$F3,$E0,$F8,$EF,$7F,$8D,$E3,$6D,$7F,$85,$63,$D6,$47,$FC,$1C,$F2
	.dc.b	$DF,$D4,$7A,$FE,$66,$CF,$F8,$77,$E3,$3E,$59,$8C,$FF,$C3,$0F,$C1
	.dc.b	$F5,$BE,$7F,$9F,$6B,$FC,$48,$E1,$FF,$0E,$3F,$30,$8F,$FA,$8F,$E9
	.dc.b	$BF,$D7,$4D,$F8,$3F,$F6,$C4,$63,$0F,$E3,$CD,$7F,$BD,$AF,$84,$FB
	.dc.b	$FE,$AF,$E7,$25,$39,$BF,$F1,$B4,$E7,$E9,$9F,$1D,$FE,$E5,$6C,$FF
	.dc.b	$AD,$5B,$7F,$CB,$5A,$FF,$82,$BF,$06,$FE,$13,$3F,$C2,$BC,$2F,$E3
	.dc.b	$7D,$0F,$BF,$7A,$FE,$3B,$DE,$1F,$E6,$F9,$CE,$1F,$F0,$07,$E7,$DE
	.dc.b	$C7,$A5,$7C,$E5,$8C,$32,$7F,$D9,$53,$A6,$C7,$FB,$FE,$31,$7F,$98
	.dc.b	$E9,$7C,$9F,$E9,$30,$FC,$C3,$70,$F5,$1F,$C9,$29,$F4,$FF,$8D,$8F
	.dc.b	$F6,$C3,$63,$FA,$83,$5F,$F1,$6B,$5E,$4B,$A3,$EA,$7D,$33,$E3,$7E
	.dc.b	$B4,$E6,$FC,$37,$FC,$D2,$B5,$18,$E0,$7D,$27,$CA,$E2,$F9,$ED,$9F
	.dc.b	$ED,$5A,$FD,$89,$CE,$1F,$C0,$B7,$95,$CE,$F8,$4F,$FA,$1A,$7F,$11
	.dc.b	$F4,$CF,$96,$BF,$29,$86,$E1,$E6,$B5,$2B,$FA,$DC,$7F,$93,$A3,$13
	.dc.b	$FE,$FF,$6C,$7F,$77,$6B,$FE,$54,$D1,$F0,$FF,$C6,$3E,$0B,$E3,$B6
	.dc.b	$FD,$1F,$37,$FC,$7B,$47,$CD,$D8,$C3,$27,$FD,$29,$F4,$17,$2F,$F4
	.dc.b	$73,$EE,$F9,$1F,$3C,$7D,$FD,$11,$88,$FE,$CB,$18,$AF,$54,$E6,$7E
	.dc.b	$0F,$C7,$7E,$BF,$EC,$7F,$A4,$D7,$C5,$B4,$E7,$D4,$F4,$7C,$DF,$A7
	.dc.b	$DF,$B7,$F4,$DF,$67,$C2,$C8,$C1,$F5,$DF,$E7,$7F,$11,$23,$07,$FA
	.dc.b	$C5,$F8,$9C,$92,$3E,$F7,$2D,$D7,$FD,$5D,$F2,$3F,$D1,$0C,$74,$7E
	.dc.b	$E3,$D4,$47,$CD,$F9,$89,$87,$E1,$FE,$62,$E1,$FE,$E7,$E7,$57,$F4
	.dc.b	$EF,$87,$CE,$5C,$DF,$E2,$5F,$4E,$F8,$DB,$ED,$61,$B9,$FE,$2E,$7F
	.dc.b	$A7,$65,$FE,$96,$6B,$FB,$36,$BF,$E1,$86,$AF,$4B,$E8,$BF,$F0,$03
	.dc.b	$F1,$0F,$EF,$0B,$2F,$CA,$CE,$9F,$CB,$CE,$9F,$E3,$3F,$BB,$BC,$4F
	.dc.b	$ED,$6D,$C7,$F5,$1F,$29,$89,$CF,$C4,$FF,$68,$A3,$49,$5F,$E5,$59
	.dc.b	$4B,$CC,$EA,$7F,$83,$FD,$E7,$B3,$8B,$FE,$01,$5F,$E2,$AD,$7D,$51
	.dc.b	$FE,$67,$8C,$67,$FE,$B1,$BF,$85,$3C,$3E,$1F,$D3,$F0,$FE,$52,$47
	.dc.b	$8E,$FF,$C5,$B8,$7F,$1F,$CF,$F8,$6F,$8F,$F0,$3E,$8B,$F5,$CD,$BF
	.dc.b	$1F,$C1,$7C,$43,$FC,$1F,$F5,$3F,$E6,$28,$C9,$E8,$F8,$EF,$88,$47
	.dc.b	$C0,$FC,$7F,$84,$F9,$A7,$2F,$8A,$B3,$F0,$3E,$23,$E8,$9D,$BC,$7B
	.dc.b	$1F,$13,$E1,$BE,$81,$1F,$63,$FE,$2B,$FA,$97,$83,$49,$3F,$E8,$7F
	.dc.b	$89,$4E,$A3,$EB,$68,$F8,$6F,$C0,$FA,$DE,$E6,$CF,$D4,$F8,$7F,$C0
	.dc.b	$11,$F9,$8F,$E0,$FD,$AF,$C4,$9F,$E4,$BC,$3F,$9D,$F2,$FD,$A3,$F1
	.dc.b	$DC,$DF,$D0,$5C,$73,$7E,$1B,$E2,$BF,$22,$71,$F9,$CB,$97,$E1,$3B
	.dc.b	$7F,$20,$47,$EE,$3F,$E3,$AD,$9E,$C7,$E6,$BE,$67,$EB,$B6,$BF,$A6
	.dc.b	$4F,$F0,$06,$F9,$7F,$73,$D7,$FB,$AF,$DE,$F2,$BA,$12,$7E,$3B,$FE
	.dc.b	$48,$FA,$6F,$D3,$6D,$EA,$DF,$F2,$6F,$B2,$33,$E1,$FE,$BB,$C8,$FF
	.dc.b	$2B,$C7,$E8,$F0,$FC,$FF,$BF,$8F,$1B,$F6,$2C,$8F,$CC,$3F,$AF,$23
	.dc.b	$FA,$D9,$FF,$43,$DF,$F8,$49,$B1,$FE,$9E,$E9,$7D,$4F,$F8,$43,$2F
	.dc.b	$27,$8B,$E2,$3F,$7F,$F4,$D6,$FF,$62,$F9,$BC,$39,$1F,$DC,$F1,$78
	.dc.b	$D8,$C3,$DD,$5F,$C2,$3E,$1F,$1A,$7F,$0C,$7D,$FE,$0C,$C7,$97,$8B
	.dc.b	$70,$F8,$D3,$FE,$7A,$FD,$E3,$FB,$AF,$6B,$4F,$F4,$16,$1F,$E0,$39
	.dc.b	$C9,$FE,$2C,$BF,$DD,$35,$FF,$76,$35,$FE,$38,$DA,$E9,$E1,$7F,$B4
	.dc.b	$BE,$4F,$FD,$79,$5F,$C8,$70,$FC,$A4,$62,$B7,$87,$FB,$A3,$CA,$FF
	.dc.b	$7C,$56,$9B,$E1,$73,$6C,$EF,$0E,$98,$E5,$ED,$55,$7D,$53,$C9,$7C
	.dc.b	$AF,$FB,$F6,$73,$18,$7F,$E0,$04,$EE,$74,$D8,$FF,$1D,$1F,$ED,$EF
	.dc.b	$96,$FF,$5A,$FD,$57,$36,$C6,$83,$87,$8B,$5F,$F2,$24,$FF,$97,$35
	.dc.b	$EF,$F1,$9F,$D1,$E7,$C6,$D7,$C3,$3B,$7F,$D5,$3C,$1E,$95,$C1,$F3
	.dc.b	$63,$0D,$CB,$FC,$A4,$E7,$E1,$6F,$07,$DC,$6D,$FD,$CA,$B9,$30,$E7
	.dc.b	$7B,$AF,$A5,$39,$F8,$5F,$98,$D8,$FE,$AD,$5F,$E2,$C6,$CE,$A1,$F5
	.dc.b	$1F,$E6,$5E,$7F,$F3,$FC,$6E,$BF,$8C,$B6,$33,$3B,$E3,$FC,$9B,$6B
	.dc.b	$F2,$6D,$C3,$D0,$DB,$7C,$15,$F8,$29,$7F,$A1,$75,$F0,$52,$30,$43
	.dc.b	$C6,$F7,$F0,$3F,$80,$FC,$15,$F0,$4D,$3C,$1F,$33,$2F,$FC,$E7,$5C
	.dc.b	$1B,$5C,$1F,$D9,$4F,$9F,$E0,$BE,$31,$23,$E5,$BF,$27,$E0,$F5,$1F
	.dc.b	$B9,$8F,$74,$7B,$FE,$A3,$E0,$7A,$D2,$BF,$7D,$1F,$0F,$E0,$BE,$91
	.dc.b	$23,$EB,$BF,$57,$DE,$F9,$04,$8F,$E0,$DF,$29,$F2,$FE,$52,$47,$F0
	.dc.b	$97,$E7,$7C,$77,$CF,$F5,$FE,$F1,$27,$F0,$6F,$E0,$7E,$2F,$EA,$7C
	.dc.b	$4C,$9F,$E2,$51,$F5,$FE,$07,$ED,$6B,$FE,$82,$EB,$FF,$61,$B7,$AF
	.dc.b	$F8,$9C,$EB,$E7,$FF,$23,$26,$93,$0F,$F1,$44,$3E,$77,$A9,$23,$FD
	.dc.b	$49,$1F,$C1,$DB,$E2,$78,$88,$EB,$E0,$7C,$BF,$11,$E0,$24,$76,$7C
	.dc.b	$7F,$13,$EA,$F8,$69,$3D,$A3,$DB,$F3,$3E,$EF,$85,$93,$E2,$8F,$83
	.dc.b	$F3,$3E,$C2,$47,$A9,$F7,$FC,$CF,$C6,$24,$7A,$BE,$63,$E4,$FC,$AF
	.dc.b	$CA,$F9,$49,$3E,$E8,$F9,$BF,$0B,$F2,$09,$1F,$19,$F9,$7F,$09,$F9
	.dc.b	$FF,$07,$F5,$33,$F9,$A6,$BF,$E0,$EE,$24,$FD,$8F,$17,$87,$8D,$2B
	.dc.b	$E6,$C7,$DC,$F8,$4F,$AE,$48,$FA,$AF,$DA,$F8,$4F,$C0,$92,$3F,$DB
	.dc.b	$1F,$BC,$FF,$23,$EC,$FF,$62,$FE,$7B,$FE,$A4,$9C,$FC,$7F,$F0,$1A
	.dc.b	$FF,$34,$C7,$FB,$36,$74,$FF,$89,$C3,$FE,$C6,$7F,$A2,$35,$FE,$74
	.dc.b	$D7,$FA,$0B,$51,$FE,$7D,$8F,$F3,$93,$DF,$B7,$F8,$A3,$D9,$F2,$7C
	.dc.b	$2F,$97,$D3,$F3,$7C,$24,$AF,$6C,$65,$A4,$3F,$C7,$77,$FE,$06,$7E
	.dc.b	$1F,$C7,$FB,$C8,$FC,$6F,$83,$F8,$44,$70,$7C,$AF,$82,$F8,$DF,$1F
	.dc.b	$23,$E6,$BF,$27,$E0,$E4,$7D,$06,$BE,$2F,$82,$F9,$C4,$7E,$AB,$F4
	.dc.b	$7E,$3F,$E9,$7C,$1F,$E2,$7C,$A9,$5F,$6A,$3E,$B7,$C7,$6A,$30,$7D
	.dc.b	$E7,$E4,$FD,$2F,$F9,$A6,$3F,$BC,$35,$FE,$DC,$D7,$FB,$AB,$5F,$AA
	.dc.b	$17,$85,$D3,$C4,$93,$F6,$B9,$DE,$67,$0B,$DC,$E6,$31,$E0,$FD,$CF
	.dc.b	$C1,$3F,$D0,$73,$93,$F0,$6E,$1F,$C7,$52,$7D,$7E,$1F,$F8,$46,$7C
	.dc.b	$8D,$93,$4D,$57,$FB,$92,$7B,$1C,$0B,$3C,$1C,$C6,$DF,$C3,$FC,$87
	.dc.b	$E8,$C8,$FE,$AB,$E4,$3F,$4C,$47,$07,$EB,$BE,$3B,$F3,$3F,$12,$E3
	.dc.b	$BC,$92,$BC,$73,$97,$F6,$BF,$1D,$C2,$A4,$60,$F4,$FD,$2F,$F0,$8E
	.dc.b	$0F,$9F,$1F,$AA,$FD,$29,$E6,$FE,$A1,$B1,$C3,$E0,$BF,$57,$6F,$82
	.dc.b	$F9,$4D,$9F,$F1,$E3,$5F,$F1,$93,$5F,$F1,$CD,$7E,$27,$0F,$0C,$27
	.dc.b	$84,$61,$BC,$9F,$8A,$AE,$0F,$E3,$3F,$01,$FE,$3F,$7F,$51,$FF,$52
	.dc.b	$B7,$CE,$FE,$03,$1A,$8F,$E2,$1D,$6F,$6F,$3E,$6C,$9F,$BB,$39,$7E
	.dc.b	$FF,$E4,$EF,$43,$F8,$5A,$F8,$2A,$62,$3F,$90,$7C,$95,$AE,$31,$87
	.dc.b	$F9,$4F,$C8,$FF,$8D,$BC,$83,$49,$1F,$E5,$9F,$89,$7E,$F5,$FE,$60
	.dc.b	$E3,$FA,$F0,$BE,$35,$D0,$FD,$14,$70,$7F,$C6,$55,$FE,$55,$AE,$2D
	.dc.b	$E1,$FE,$53,$5D,$BE,$33,$7F,$1B,$30,$E7,$68,$C4,$7F,$B7,$23,$FC
	.dc.b	$89,$7C,$9F,$F2,$5C,$7A,$5A,$FF,$9E,$2F,$D7,$F1,$9F,$E5,$69,$CC
	.dc.b	$FF,$06,$9F,$7F,$47,$A2,$A4,$F8,$A3,$D7,$D1,$C4,$91,$DD,$E9,$EB
	.dc.b	$EA,$3D,$31,$F0,$A3,$E1,$F0,$FC,$4E,$09,$5E,$E8,$CB,$48,$7B,$FE
	.dc.b	$27,$E5,$FC,$4F,$A4,$91,$EF,$7D,$9E,$D6,$A3,$E4,$C7,$D2,$8F,$A7
	.dc.b	$F1,$3F,$53,$E2,$65,$7C,$E8,$FA,$FF,$13,$F1,$49,$1F,$2D,$F9,$1D
	.dc.b	$7A,$8F,$AB,$1F,$7A,$3E,$FF,$C4,$FE,$03,$E2,$65,$7D,$A8,$FC,$1F
	.dc.b	$67,$D1,$24,$7D,$77,$EA,$7C,$4F,$E2,$09,$1F,$82,$F6,$FF,$16,$47
	.dc.b	$07,$E4,$3F,$7B,$FF,$22,$3F,$EB,$5F,$5D,$D9,$FD,$DB,$E9,$BF,$82
	.dc.b	$7C,$93,$DA,$FE,$8B,$18,$6F,$47,$F4,$89,$5F,$2C,$17,$CA,$31,$F0
	.dc.b	$1C,$67,$FC,$06,$D4,$EF,$1C,$E1,$D6,$BA,$5E,$C9,$19,$9F,$43,$6F
	.dc.b	$9B,$E5,$C7,$FA,$EE,$1D,$CE,$D3,$A3,$E1,$57,$81,$FF,$6C,$4F,$0A
	.dc.b	$E5,$19,$3F,$DA,$F8,$7B,$23,$C1,$D9,$FE,$2A,$9C,$DF,$EE,$30,$C9
	.dc.b	$23,$FD,$CB,$F2,$39,$68,$AE,$4F,$F8,$A2,$34,$FF,$B5,$0D,$4F,$FB
	.dc.b	$A6,$3C,$91,$FE,$F8,$EF,$AF,$FB,$78,$F7,$1F,$E3,$2F,$07,$FA,$5D
	.dc.b	$7F,$3F,$8D,$3F,$D0,$63,$38,$F5,$9E,$27,$52,$9D,$74,$E1,$7F,$4A
	.dc.b	$3D,$71,$87,$F0,$73,$CE,$F0,$78,$70,$F4,$E1,$E0,$AE,$35,$FE,$14
	.dc.b	$7C,$17,$B7,$99,$FE,$18,$9F,$EC,$B1,$FD,$77,$16,$DF,$E9,$DC,$BF
	.dc.b	$41,$F1,$BF,$DA,$B4,$FE,$6B,$3F,$E9,$D9,$C9,$FA,$5F,$3F,$4F,$F3
	.dc.b	$08,$46,$32,$FA,$D1,$D5,$FD,$4C,$EE,$75,$87,$8B,$E3,$3F,$B4,$D7
	.dc.b	$36,$BC,$9B,$1B,$AF,$E9,$96,$F8,$CE,$6B,$E3,$2C,$61,$CF,$C5,$EB
	.dc.b	$E3,$2D,$FE,$FD,$C3,$D1,$EF,$66,$98,$F4,$BB,$FB,$BE,$06,$D7,$F0
	.dc.b	$9F,$7B,$C7,$F1,$9E,$46,$C7,$F0,$D8,$EA,$47,$D6,$F8,$3E,$33,$D6
	.dc.b	$7B,$3E,$02,$CF,$F1,$58,$F7,$FC,$06,$38,$7E,$69,$F4,$1F,$87,$EB
	.dc.b	$7C,$4F,$56,$DF,$F1,$D7,$0F,$CA,$7D,$E4,$7E,$6F,$C6,$7C,$CF,$8C
	.dc.b	$F9,$AD,$8F,$E4,$D1,$F1,$88,$FD,$47,$E4,$FC,$67,$D4,$3E,$FF,$C3
	.dc.b	$7E,$E6,$7F,$97,$46,$3B,$E4,$60,$FF,$00,$CF,$F2,$06,$FC,$5F,$2F
	.dc.b	$8A,$E6,$D8,$DC,$7E,$1A,$4D,$24,$7F,$84,$E1,$F1,$7F,$97,$D9,$FA
	.dc.b	$73,$2F,$3C,$E6,$31,$7F,$E6,$86,$D7,$FB,$07,$F3,$4F,$63,$FE,$C3
	.dc.b	$8C,$F7,$FF,$91,$27,$2F,$F9,$1B,$ED,$9E,$87,$87,$E9,$5E,$9F,$8E
	.dc.b	$73,$97,$F9,$DA,$72,$D9,$E1,$5C,$5B,$1F,$E7,$CA,$79,$BE,$DF,$A4
	.dc.b	$70,$F5,$8C,$32,$BF,$85,$6B,$FE,$D2,$AD,$C6,$21,$18,$76,$D7,$F8
	.dc.b	$93,$F4,$23,$FD,$1B,$F4,$9F,$E5,$57,$FD,$1F,$19,$9F,$F7,$44,$E5
	.dc.b	$FC,$CF,$88,$E2,$7C,$DF,$A7,$E2,$3F,$12,$DC,$79,$C7,$B7,$E0,$F1
	.dc.b	$87,$FB,$CF,$24,$7E,$43,$E4,$A5,$FE,$33,$D1,$B7,$E0,$CB,$FB,$AC
	.dc.b	$FE,$43,$D1,$CF,$27,$F9,$5A,$73,$5F,$EF,$FF,$B6,$78,$DF,$89,$FA
	.dc.b	$47,$B7,$F8,$B9,$CE,$5F,$EA,$E9,$CB,$67,$F8,$ED,$A7,$2F,$D2,$6B
	.dc.b	$FC,$F5,$AF,$FB,$49,$7B,$7F,$99,$FA,$1F,$F6,$7D,$F7,$8C,$3E,$5F
	.dc.b	$6F,$CD,$ED,$58,$F2,$38,$7E,$F5,$A3,$CA,$D8,$CC,$7E,$B7,$E5,$7F
	.dc.b	$77,$F1,$DF,$BC,$6C,$FE,$E2,$3F,$A1,$7C,$CB,$F1,$8F,$C4,$70,$FC
	.dc.b	$B9,$CC,$62,$BF,$CF,$3C,$3F,$EC,$57,$F8,$0F,$C8,$7F,$08,$8F,$E1
	.dc.b	$2F,$F0,$AF,$97,$93,$A3,$F8,$3B,$FC,$3B,$E5,$E4,$60,$FE,$16,$FF
	.dc.b	$12,$F9,$59,$1F,$C3,$BE,$56,$F1,$FE,$0F,$E5,$7F,$8B,$98,$D7,$F0
	.dc.b	$4F,$F1,$AE,$D9,$7F,$C5,$BB,$2D,$FE,$2E,$7F,$8F,$FC,$8E,$30,$FC
	.dc.b	$53,$FD,$83,$F7,$C5,$7F,$11,$FA,$4F,$E9,$D8,$FF,$A8,$27,$2F,$FB
	.dc.b	$DA,$33,$3F,$CF,$83,$87,$FC,$04,$D7,$A9,$AF,$6B,$F1,$F1,$98,$F0
	.dc.b	$BF,$EE,$69,$F0,$7C,$1F,$8F,$DB,$F3,$3E,$7F,$92,$93,$E5,$F9,$EB
	.dc.b	$87,$EE,$32,$7C,$97,$F7,$66,$D7,$FB,$05,$B3,$FE,$A9,$6B,$FE,$A7
	.dc.b	$5F,$25,$F8,$4E,$2F,$8C,$DA,$F8,$9F,$11,$F8,$8E,$1F,$94,$D8,$F9
	.dc.b	$1F,$25,$E1,$23,$E1,$7C,$9F,$23,$F4,$7E,$4B,$E9,$36,$7E,$87,$C9
	.dc.b	$7D,$53,$EB,$7C,$95,$8F,$A9,$EA,$7B,$48,$FB,$5F,$81,$EA,$FD,$DF
	.dc.b	$27,$EF,$36,$7E,$E7,$91,$F8,$1F,$92,$FC,$13,$63,$F0,$1F,$25,$F2
	.dc.b	$48,$FC,$97,$E6,$79,$5F,$89,$3F,$15,$E4,$D9,$FC,$47,$AD,$F8,$D3
	.dc.b	$F1,$DF,$25,$63,$FD,$D7,$FB,$EF,$F0,$6E,$8F,$F6,$D3,$5F,$F6,$C3
	.dc.b	$5F,$F7,$5B,$5F,$F3,$83,$FE,$3B,$F0,$7F,$DF,$2F,$F9,$16,$3F,$9C
	.dc.b	$B9,$9E,$3F,$3B,$FE,$3E,$8F,$F2,$2C,$78,$BE,$3F,$1F,$21,$97,$5F
	.dc.b	$21,$FE,$3E,$AF,$F1,$E5,$7F,$54,$F6,$91,$8C,$7D,$35,$BF,$90,$EC
	.dc.b	$77,$F9,$0B,$87,$C0,$F1,$FC,$47,$8F,$C6,$B1,$F1,$3D,$7F,$01,$1F
	.dc.b	$06,$5F,$CE,$B2,$F8,$FF,$1F,$EB,$F8,$FF,$63,$6B,$E6,$FC,$77,$BF
	.dc.b	$E3,$FE,$03,$63,$E8,$C7,$B4,$8F,$67,$E0,$7C,$7F,$F2,$8F,$65,$C5
	.dc.b	$F2,$63,$0C,$AF,$E1,$3C,$A7,$F8,$5F,$B2,$BE,$31,$F8,$0B,$E1,$A7
	.dc.b	$D6,$F8,$FF,$A9,$F1,$FF,$55,$B5,$F8,$68,$B3,$B8,$C1,$F8,$AF,$8F
	.dc.b	$FB,$87,$DD,$F8,$FB,$19,$3F,$1F,$F1,$B8,$AF,$C2,$BF,$83,$3E,$F1
	.dc.b	$23,$E7,$BF,$E7,$EF,$DF,$BF,$C5,$7F,$29,$FE,$3A,$AF,$F6,$04,$7F
	.dc.b	$90,$D4,$67,$AD,$3C,$3E,$43,$F9,$6C,$63,$E4,$7E,$35,$F0,$9C,$D7
	.dc.b	$3B,$E4,$D8,$CD,$F0,$9C,$DD,$9F,$E5,$BF,$43,$FE,$A0,$F6,$3F,$C4
	.dc.b	$CD,$C7,$FC,$56,$D7,$FD,$6C,$D7,$F9,$AF,$CB,$7D,$7F,$A2,$FF,$17
	.dc.b	$36,$3F,$C6,$6D,$7F,$D7,$CD,$7C,$3C,$DF,$EE,$37,$BB,$5E,$ED,$77
	.dc.b	$9F,$EA,$FE,$89,$8F,$F4,$6D,$3E,$0F,$C8,$5D,$3E,$2B,$70,$FF,$78
	.dc.b	$5F,$07,$FD,$E1,$3F,$3A,$31,$1F,$34,$DD,$F1,$C1,$3A,$C3,$EC,$2F
	.dc.b	$73,$5B,$FD,$FC,$3F,$DB,$97,$F7,$31,$FF,$29,$36,$3F,$82,$4E,$DF
	.dc.b	$EB,$F1,$DB,$DC,$FE,$C3,$D5,$F8,$08,$E1,$3F,$E4,$EA,$ED,$39,$C5
	.dc.b	$7C,$38,$FD,$8D,$BF,$82,$DC,$3B,$47,$EC,$A7,$E1,$7B,$99,$C7,$FD
	.dc.b	$1B,$39,$3F,$33,$F2,$B9,$39,$62,$F6,$69,$99,$E1,$8F,$E7,$2B,$31
	.dc.b	$8B,$FF,$32,$B7,$E4,$3C,$07,$9F,$B9,$F4,$4E,$BC,$36,$B8,$10,$FF
	.dc.b	$64,$CF,$09,$F5,$7C,$87,$D0,$C0,$D2,$57,$E2,$FD,$BD,$BE,$58,$F6
	.dc.b	$C7,$BB,$D8,$F8,$9F,$21,$2B,$DB,$BF,$FA,$81,$BE,$D7,$DE,$83,$49
	.dc.b	$1F,$99,$F6,$7E,$49,$1E,$7E,$D7,$D1,$F9,$09,$1F,$1D,$F9,$FE,$DF
	.dc.b	$D5,$F2,$3D,$84,$9F,$C5,$BE,$1F,$5F,$ED,$7C,$6C,$9F,$3C,$7E,$07
	.dc.b	$5F,$E1,$71,$F8,$BF,$8B,$FC,$01,$9E,$AF,$C1,$7C,$67,$C9,$24,$FE
	.dc.b	$61,$F9,$9F,$19,$F8,$72,$47,$E0,$2D,$ED,$7D,$BF,$C8,$59,$D2,$56
	.dc.b	$FC,$6F,$C8,$7C,$6F,$D7,$24,$7E,$99,$FD,$4E,$0F,$63,$F2,$DF,$1B
	.dc.b	$2B,$C1,$1E,$DF,$63,$F3,$5F,$11,$27,$C9,$1F,$9D,$F8,$DF,$C2,$12
	.dc.b	$3F,$66,$FE,$1F,$E2,$3F,$44,$48,$F6,$7B,$5F,$27,$D7,$FD,$37,$C6
	.dc.b	$C9,$F8,$11,$FA,$9F,$8D,$FC,$81,$23,$F7,$EF,$FB,$4F,$C4,$78,$3F
	.dc.b	$E4,$89,$FD,$3E,$1F,$E6,$07,$FD,$96,$DF,$33,$FA,$0D,$70,$F8,$BF
	.dc.b	$D3,$5B,$E2,$FF,$79,$7F,$C6,$AF,$F8,$43,$7A,$BF,$B6,$D6,$F1,$EF
	.dc.b	$E8,$7F,$14,$AF,$E3,$51,$FE,$17,$6B,$FC,$69,$AF,$F1,$67,$F8,$7B
	.dc.b	$E8,$F3,$FC,$27,$E3,$7D,$8F,$E1,$DF,$38,$F6,$7C,$BE,$94,$BF,$E0
	.dc.b	$B3,$97,$F8,$70,$FA,$3F,$F8,$02,$FF,$BE,$5A,$FF,$BF,$9A,$FC,$76
	.dc.b	$BF,$E8,$85,$F0,$FF,$C8,$63,$A7,$89,$67,$9E,$FF,$A2,$77,$3E,$4F
	.dc.b	$4B,$E5,$F1,$48,$C1,$0F,$F7,$FE,$3F,$C9,$9B,$3F,$03,$D5,$FA,$7E
	.dc.b	$AC,$9F,$BA,$FC,$4F,$42,$47,$C2,$6B,$D3,$D0,$FB,$7C,$DF,$73,$E3
	.dc.b	$64,$FE,$1A,$3E,$6F,$97,$FD,$33,$3F,$F1,$44,$FF,$93,$DA,$FF,$52
	.dc.b	$6B,$FD,$39,$FD,$9B,$F4,$BC,$6F,$AE,$7D,$8F,$2F,$F1,$7D,$32,$7F
	.dc.b	$2A,$FE,$43,$E2,$64,$7F,$45,$EC,$FC,$DC,$62,$3E,$2F,$97,$F4,$7E
	.dc.b	$D9,$E1,$D2,$4F,$E0,$3E,$91,$F0,$FF,$5D,$78,$DF,$A3,$9F,$F6,$1C
	.dc.b	$61,$C0,$70,$ED,$AF,$EA,$FE,$93,$F9,$B3,$FA,$CF,$90,$FD,$39,$F8
	.dc.b	$58,$CF,$5F,$F2,$4A,$7C,$37,$6B,$EC,$D8,$CC,$78,$9A,$F8,$7C,$7F
	.dc.b	$54,$C8,$F7,$47,$ED,$BD,$79,$1F,$C0,$1F,$D3,$F9,$B2,$3F,$4A,$D7
	.dc.b	$F1,$FE,$67,$E6,$FE,$37,$EE,$E5,$FC,$E9,$BE,$57,$FB,$A1,$B5,$FE
	.dc.b	$F6,$7E,$FB,$FE,$F1,$6C,$F1,$E7,$E8,$75,$E7,$FD,$47,$D1,$FD,$4B
	.dc.b	$FE,$2B,$E8,$7F,$0E,$33,$C7,$FA,$DF,$A4,$FE,$35,$3F,$5D,$AF,$F1
	.dc.b	$B6,$BF,$EF,$C6,$A0,$FF,$82,$DF,$E0,$6F,$4C,$7F,$AB,$B7,$3E,$5E
	.dc.b	$0F,$EF,$7E,$91,$C3,$49,$87,$F0,$C8,$7C,$56,$C9,$19,$79,$7C,$57
	.dc.b	$F0,$7F,$A4,$FE,$4D,$2C,$FF,$B3,$C6,$1B,$E1,$6F,$C3,$95,$FC,$6A
	.dc.b	$91,$88,$E5,$A7,$BD,$4F,$F4,$27,$FA,$0F,$BD,$FC,$CC,$8F,$F3,$37
	.dc.b	$F7,$DE,$9F,$F3,$E6,$47,$F2,$A8,$FE,$83,$EF,$48,$FE,$96,$FF,$8F
	.dc.b	$FE,$93,$FA,$14,$3E,$15,$18,$6F,$C3,$EC,$E1,$EB,$FF,$4C,$F9,$5F
	.dc.b	$E8,$C4,$BF,$E8,$CF,$F9,$72,$DE,$A6,$E7,$FA,$DC,$F8,$63,$FA,$AF
	.dc.b	$C0,$FF,$4F,$24,$FF,$4F,$79,$73,$6A,$3F,$B1,$C7,$A6,$3D,$5E,$5F
	.dc.b	$F5,$EF,$95,$95,$FE,$4E,$C7,$E5,$3F,$38,$F7,$38,$E4,$FF,$35,$37
	.dc.b	$8B,$FD,$5A,$D9,$DF,$1F,$F6,$AF,$95,$FE,$C8,$48,$FE,$C8,$FB,$FC
	.dc.b	$3D,$47,$F7,$88,$F9,$D1,$F3,$F8,$BF,$B9,$F9,$92,$BC,$D1,$F4,$F9
	.dc.b	$7E,$41,$23,$F8,$F7,$27,$D4,$8F,$99,$C5,$B9,$F7,$47,$DB,$8F,$B9
	.dc.b	$8F,$F8,$03,$D6,$95,$F0,$E3,$F0,$1B,$FD,$52,$47,$6D,$51,$F6,$73
	.dc.b	$FC,$31,$27,$E6,$5F,$E2,$48,$E0,$FA,$1E,$BF,$F6,$DF,$8C,$7C,$2F
	.dc.b	$E9,$E5,$F4,$9F,$BB,$52,$F5,$7F,$45,$39,$8C,$4F,$F0,$38,$C5,$73
	.dc.b	$6F,$27,$F0,$7F,$6F,$86,$1F,$AB,$6D,$E7,$9F,$FD,$AB,$87,$8E,$78
	.dc.b	$CF,$92,$7F,$DB,$35,$E2,$6C,$78,$5E,$37,$63,$F8,$B6,$3E,$E8,$ED
	.dc.b	$18,$AF,$15,$72,$8F,$F1,$03,$C9,$A7,$2C,$3C,$91,$BC,$3D,$33,$A3
	.dc.b	$E3,$CE,$9E,$17,$FC,$9F,$8B,$FC,$51,$1A,$8C,$57,$79,$CE,$3E,$7A
	.dc.b	$FE,$01,$3F,$DA,$FE,$33,$71,$FE,$4F,$6B,$FE,$EB,$36,$64,$A6,$63
	.dc.b	$EE,$3F,$EF,$2E,$7F,$86,$E7,$2F,$F2,$D7,$62,$73,$3F,$CF,$EE,$CE
	.dc.b	$63,$E0,$19,$2B,$9B,$EF,$09,$FA,$8F,$F4,$99,$E5,$AE,$A7,$33,$E5
	.dc.b	$3F,$00,$7A,$EB,$D3,$5D,$70,$F8,$93,$E0,$C3,$CE,$F7,$38,$9D,$DE
	.dc.b	$AE,$70,$E7,$1F,$87,$C9,$F4,$6F,$2F,$DD,$AF,$FB,$EE,$31,$5F,$EE
	.dc.b	$65,$F4,$9A,$8F,$BD,$1A,$C7,$F8,$E6,$1D,$3E,$1B,$51,$FC,$EE,$51
	.dc.b	$89,$F0,$E3,$67,$37,$DF,$E6,$3F,$D8,$38,$76,$6B,$CD,$B7,$EA,$6D
	.dc.b	$3E,$63,$B7,$C2,$7F,$9B,$B0,$F5,$46,$39,$7C,$1F,$0F,$E1,$F3,$6D
	.dc.b	$7B,$DC,$3F,$C8,$30,$FF,$20,$B6,$3B,$9E,$0F,$63,$D9,$FB,$FF,$E2
	.dc.b	$E7,$E4,$46,$1F,$76,$5A,$E0,$B1,$FE,$0D,$E1,$FC,$76,$CC,$3E,$7C
	.dc.b	$E4,$FD,$63,$8D,$FF,$D6,$53,$BB,$F8,$73,$99,$FE,$F5,$39,$AF,$57
	.dc.b	$9D,$F3,$CF,$A1,$F2,$D6,$3E,$F4,$7A,$BE,$33,$17,$F7,$30,$FF,$73
	.dc.b	$D7,$E0,$FC,$EF,$7F,$27,$D9,$6C,$79,$EB,$FB,$D5,$7E,$15,$F1,$B5
	.dc.b	$F5,$D7,$A0,$FB,$3F,$2D,$F2,$B9,$BF,$08,$DB,$E9,$A7,$C5,$9F,$79
	.dc.b	$23,$FA,$AE,$1F,$8A,$C3,$E2,$FC,$77,$FA,$F6,$78,$B6,$3F,$2F,$C6
	.dc.b	$F8,$9F,$CA,$4E,$78,$7E,$9E,$0E,$8F,$BA,$6B,$B2,$E5,$D3,$E1,$BF
	.dc.b	$03,$D5,$C9,$B1,$E8,$BF,$CA,$3E,$69,$FA,$E4,$7F,$D0,$FF,$49,$FE
	.dc.b	$21,$7F,$D9,$ED,$D7,$FC,$10,$3E,$97,$D1,$3A,$E0,$FD,$7C,$AB,$D5
	.dc.b	$3F,$E3,$B6,$C7,$F9,$6D,$E9,$A7,$F1,$D8,$E1,$9F,$2F,$A5,$79,$74
	.dc.b	$C7,$B3,$71,$FF,$71,$CB,$CD,$E0,$67,$0F,$3C,$66,$7F,$DD,$3F,$46
	.dc.b	$F9,$8F,$3A,$E1,$B1,$86,$4F,$FA,$AA,$74,$D8,$FE,$07,$69,$CB,$FE
	.dc.b	$B3,$FA,$47,$A9,$FB,$4F,$9E,$F8,$B1,$8C,$3E,$A4,$78,$BE,$4A,$47
	.dc.b	$9B,$E4,$EE,$1F,$43,$17,$B5,$E8,$33,$C5,$F2,$63,$37,$F8,$EF,$3B
	.dc.b	$F1,$67,$E3,$3E,$3E,$C7,$E0,$FC,$9F,$E0,$E6,$31,$FD,$73,$F6,$BE
	.dc.b	$8B,$E4,$47,$E0,$17,$AF,$9E,$4D,$32,$FF,$D9,$F3,$A6,$C7,$C6,$8C
	.dc.b	$57,$FB,$5F,$E8,$FF,$8B,$3F,$BD,$F3,$3F,$13,$18,$D7,$F4,$31,$E3
	.dc.b	$F5,$79,$13,$1E,$EF,$D2,$FA,$CF,$D3,$47,$F0,$1C,$11,$87,$D5,$EB
	.dc.b	$7E,$5A,$31,$87,$F6,$DF,$90,$7D,$27,$E1,$70,$FF,$7C,$37,$0F,$E9
	.dc.b	$BE,$2B,$E2,$3F,$A6,$DF,$0E,$CF,$E6,$3E,$94,$C7,$8C,$5C,$38,$FC
	.dc.b	$8F,$EF,$8F,$DF,$FC,$4D,$AF,$E0,$BF,$31,$FD,$33,$33,$49,$1F,$C3
	.dc.b	$BE,$87,$F9,$FF,$AD,$FD,$01,$B1,$FC,$3B,$E4,$BF,$85,$9F,$C3,$3E
	.dc.b	$3A,$C7,$F0,$3F,$5F,$FA,$39,$8D,$3F,$97,$3F,$C4,$BB,$E4,$FF,$18
	.dc.b	$F0,$2D,$7C,$29,$FE,$9F,$F1,$F8,$E5,$FE,$2E,$FF,$02,$FA,$47,$7D
	.dc.b	$EB,$C3,$72,$FA,$D3,$FD,$73,$1F,$F6,$01,$FE,$DF,$70,$E1,$AF,$FB
	.dc.b	$15,$AF,$FA,$DF,$E9,$3F,$C3,$58,$FF,$28,$F6,$BF,$BF,$46,$2F,$FA
	.dc.b	$CC,$7F,$66,$F8,$D9,$1F,$CF,$25,$F1,$78,$C3,$FA,$87,$B7,$FD,$70
	.dc.b	$95,$FC,$F9,$FF,$01,$F5,$FE,$23,$E4,$7F,$B8,$FC,$FF,$E1,$3B,$D1
	.dc.b	$8B,$FC,$5F,$1F,$F7,$6F,$73,$4D,$8F,$E6,$DE,$D7,$F6,$A2,$3F,$D3
	.dc.b	$1F,$F0,$7E,$AF,$8C,$FF,$0D,$7F,$00,$7F,$2A,$AF,$F3,$04,$E5,$FE
	.dc.b	$AF,$19,$DB,$FD,$B8,$FF,$32,$F6,$0D,$A7,$CB,$74,$F6,$3A,$F8,$9F
	.dc.b	$E1,$EA,$FD,$16,$1F,$DF,$1A,$FF,$9B,$DA,$FD,$3F,$37,$CB,$EA,$FD
	.dc.b	$26,$E7,$F3,$FD,$5F,$4F,$CD,$7D,$66,$C7,$A3,$E8,$BC,$04,$7F,$D8
	.dc.b	$38,$A3,$E8,$FE,$39,$FD,$1A,$7F,$C0,$1C,$E8,$C6,$7F,$E4,$4D,$DF
	.dc.b	$1F,$FB,$89,$CF,$6F,$F8,$83,$27,$A9,$F3,$0F,$C4,$FA,$17,$3F,$97
	.dc.b	$F1,$BF,$0C,$8F,$91,$FF,$21,$FD,$37,$F8,$5B,$4F,$F2,$9E,$1F,$EC
	.dc.b	$0A,$FF,$04,$CE,$71,$F0,$7A,$FF,$E5,$38,$FF,$04,$47,$8F,$E7,$71
	.dc.b	$F0,$DF,$79,$FB,$FE,$A7,$D3,$26,$1F,$11,$FA,$DE,$97,$E1,$0E,$9F
	.dc.b	$07,$67,$C3,$EA,$F7,$3C,$1E,$55,$8F,$2F,$A3,$EA,$23,$FE,$F0,$D1
	.dc.b	$1E,$EF,$4B,$F2,$3E,$A7,$A5,$B5,$EE,$F3,$FF,$2B,$EA,$FB,$5B,$1F
	.dc.b	$94,$8F,$C6,$11,$FB,$AF,$EB,$36,$F8,$71,$F8,$DF,$57,$F9,$7C,$63
	.dc.b	$93,$F3,$71,$FA,$22,$4F,$E0,$A7,$5E,$27,$E7,$A7,$E7,$79,$9F,$A8
	.dc.b	$F5,$2D,$FE,$99,$C3,$4F,$AF,$18,$C3,$FD,$1B,$B7,$F8,$81,$BE,$47
	.dc.b	$D9,$F3,$7E,$D3,$2F,$F6,$3E,$65,$AF,$C1,$BF,$B8,$3F,$58,$48,$FC
	.dc.b	$EB,$FE,$8C,$FA,$3F,$F2,$37,$C7,$FF,$85,$69,$AA,$73,$E4,$7F,$B7
	.dc.b	$DF,$F1,$B5,$6E,$75,$E5,$7F,$B2,$63,$13,$C2,$FF,$81,$FA,$F9,$C3
	.dc.b	$95,$D9,$CE,$5F,$ED,$39,$44,$FA,$6F,$F5,$C7,$C8,$7F,$38,$8C,$7C
	.dc.b	$5A,$31,$09,$DB,$B6,$BF,$D0,$3E,$45,$CA,$FA,$6E,$31,$86,$C6,$67
	.dc.b	$C1,$39,$7A,$78,$3F,$19,$BC,$1E,$2C,$3F,$D8,$2D,$9D,$4F,$F5,$09
	.dc.b	$7B,$5E,$E8,$FF,$75,$69,$67,$FA,$DC,$7C,$33,$E1,$77,$6F,$8B,$EB
	.dc.b	$47,$BE,$F1,$7C,$5B,$1F,$25,$AE,$CE,$83,$F1,$3F,$E2,$6E,$30,$7B
	.dc.b	$BF,$32,$3E,$63,$63,$31,$EF,$8D,$7E,$68,$E4,$F6,$78,$DF,$E6,$FB
	.dc.b	$DD,$FF,$9F,$1B,$1F,$E7,$3A,$E1,$3F,$E2,$B6,$BF,$E8,$26,$C6,$EB
	.dc.b	$AF,$27,$C9,$F5,$DE,$06,$30,$FE,$A3,$17,$87,$FA,$8F,$C2,$DF,$28
	.dc.b	$D9,$95,$FD,$69,$AF,$F2,$CF,$7F,$97,$F0,$0F,$F8,$23,$4F,$F1,$D7
	.dc.b	$3F,$FA,$AE,$BF,$81,$77,$69,$87,$FA,$EA,$B8,$6B,$F9,$CD,$11,$89
	.dc.b	$FD,$7B,$F9,$B7,$FD,$99,$D2,$E6,$FC,$BC,$7E,$86,$31,$8F,$E8,$1B
	.dc.b	$19,$9D,$BF,$EE,$8A,$FF,$07,$BD,$23,$39,$31,$FD,$7D,$F2,$A4,$FF
	.dc.b	$BD,$AF,$8C,$7F,$AD,$3F,$35,$FF,$02,$3F,$B0,$6B,$FB,$56,$BF,$E2
	.dc.b	$E6,$BD,$B0,$78,$FE,$96,$C7,$A6,$30,$7E,$F3,$37,$E4,$9F,$EB,$69
	.dc.b	$FF,$72,$B5,$F3,$E7,$F3,$63,$19,$7A,$7E,$A3,$EC,$FC,$EC,$BF,$A8
	.dc.b	$FD,$CF,$A7,$FC,$46,$CF,$D0,$DF,$33,$F3,$5B,$7F,$45,$AF,$F9,$2A
	.dc.b	$78,$4F,$9F,$81,$F3,$7F,$88,$25,$7E,$0F,$E8,$7F,$17,$F4,$12,$3F
	.dc.b	$0C,$FE,$3B,$E1,$3C,$7E,$33,$E6,$FE,$F3,$6B,$F0,$4D,$7C,$4D,$7F
	.dc.b	$86,$71,$FE,$3B,$E1,$7F,$32,$48,$FC,$A7,$D0,$7E,$73,$E5,$BF,$3B
	.dc.b	$F2,$F2,$7F,$31,$1F,$13,$F3,$07,$9B,$F8,$F6,$CF,$E4,$DA,$FC,$36
	.dc.b	$BE,$2E,$0F,$F7,$43,$FC,$F3,$9F,$F2,$0F,$F4,$EE,$BF,$F4,$8B,$9E
	.dc.b	$AF,$D3,$46,$23,$EE,$9F,$A6,$F8,$8F,$D9,$FC,$A7,$ED,$3E,$22,$4F
	.dc.b	$EB,$E3,$F8,$13,$FE,$FC,$FA,$0F,$E3,$5B,$1A,$48,$FE,$99,$C3,$FB
	.dc.b	$48,$FD,$F9,$23,$F7,$5F,$0D,$FC,$0B,$E4,$FF,$81,$FC,$2C,$9F,$DF
	.dc.b	$47,$F0,$6E,$BF,$C1,$12,$3A,$F0,$7E,$21,$FA,$B9,$7C,$27,$DB,$97
	.dc.b	$F9,$29,$B1,$9E,$0F,$8F,$28,$7C,$FF,$F1,$D7,$F0,$53,$A7,$E9,$B7
	.dc.b	$8F,$5D,$1F,$C8,$FF,$26,$FE,$A3,$8F,$C0,$6B,$FE,$A4,$C5,$18,$8F
	.dc.b	$C5,$B5,$FE,$35,$83,$D5,$F1,$12,$3E,$D1,$FE,$0A,$87,$CF,$FD,$4F
	.dc.b	$AB,$FF,$32,$CF,$F7,$86,$BF,$DF,$5A,$FF,$AD,$9A,$FF,$67,$5B,$FF
	.dc.b	$55,$6C,$73,$8C,$1F,$D6,$7F,$72,$FE,$59,$1F,$8A,$13,$C1,$C3,$75
	.dc.b	$FA,$93,$C1,$F9,$9F,$0D,$F3,$FE,$72,$5F,$BB,$5F,$A9,$3F,$4C,$91
	.dc.b	$9C,$D1,$F5,$FD,$AF,$B1,$EC,$CA,$D3,$F6,$FE,$1E,$47,$16,$BF,$84
	.dc.b	$F8,$7F,$CE,$FD,$4F,$E0,$A3,$E7,$87,$0F,$BD,$BA,$7E,$3F,$F0,$9F
	.dc.b	$D1,$4F,$06,$BF,$E2,$D6,$BF,$EE,$66,$BF,$E1,$76,$BF,$E3,$6F,$AB
	.dc.b	$79,$3F,$B0,$9E,$75,$DF,$B7,$F6,$75,$AB,$F3,$E1,$F8,$96,$F2,$FB
	.dc.b	$0F,$EF,$1F,$8D,$3E,$3A,$92,$7E,$84,$65,$F9,$DF,$19,$FA,$3F,$80
	.dc.b	$91,$F4,$5F,$D3,$7B,$BF,$A7,$F9,$BB,$3F,$56,$30,$F9,$38,$7F,$C4
	.dc.b	$19,$9A,$48,$FF,$27,$C7,$FA,$59,$BC,$3C,$08,$FF,$95,$3E,$29,$EE
	.dc.b	$7D,$89,$93,$F6,$27,$2F,$E2,$FB,$BF,$C1,$8F,$F0,$46,$F1,$7F,$02
	.dc.b	$6F,$95,$FE,$0B,$F9,$0F,$F2,$6C,$FF,$06,$6C,$67,$AF,$FC,$45,$18
	.dc.b	$BF,$F3,$EA,$F8,$7F,$E4,$FE,$1B,$E5,$A9,$2B,$F1,$BF,$86,$FF,$91
	.dc.b	$9F,$D5,$37,$2F,$E3,$4D,$CB,$FC,$B8,$D9,$FF,$3D,$BF,$DF,$9F,$D9
	.dc.b	$5F,$F8,$F9,$FF,$31,$1F,$D3,$BE,$13,$FC,$A1,$19,$8F,$F3,$33,$5F
	.dc.b	$F4,$2A,$F8,$2D,$92,$3F,$D0,$71,$97,$FD,$0B,$F0,$F5,$F0,$CF,$C0
	.dc.b	$3F,$A3,$C6,$71,$FE,$69,$F5,$7F,$CA,$B2,$FF,$3B,$36,$FF,$D2,$AD
	.dc.b	$5F,$8E,$FE,$0D,$FE,$1C,$DA,$FE,$7C,$DA,$FF,$40,$36,$7E,$9F,$5F
	.dc.b	$F5,$8F,$CD,$DF,$37,$49,$3F,$C3,$F8,$5F,$2D,$F7,$09,$19,$7E,$FF
	.dc.b	$CB,$7F,$45,$F9,$FF,$F1,$57,$92,$69,$2B,$FB,$0C,$3A,$38,$47,$F1
	.dc.b	$99,$FE,$9D,$1F,$D3,$FE,$4F,$FC,$DD,$F3,$6F,$C8,$3F,$C8,$BC,$3F
	.dc.b	$DF,$23,$1F,$41,$BC,$7F,$91,$5D,$7F,$AE,$BF,$D6,$7E,$7F,$FC,$ED
	.dc.b	$F3,$8F,$9A,$FF,$20,$36,$BF,$BA,$B5,$FE,$AC,$D7,$FB,$5F,$C8,$F2
	.dc.b	$99,$1F,$CC,$63,$FD,$05,$FA,$6F,$FA,$CA,$3F,$C4,$AD,$7F,$CE,$8D
	.dc.b	$7F,$DC,$2D,$F0,$0F,$E9,$E3,$FA,$03,$FA,$B8,$C9,$DB,$C0,$EF,$CB
	.dc.b	$F1,$B7,$7C,$65,$33,$87,$29,$AF,$F8,$0B,$0F,$C0,$BD,$BE,$5F,$FD
	.dc.b	$35,$F1,$9F,$B5,$8F,$F0,$BB,$5F,$EF,$0D,$7F,$C8,$2D,$7D,$BE,$1F
	.dc.b	$22,$3E,$A7,$ED,$FE,$0B,$FD,$22,$3F,$DA,$9C,$68,$C4,$7F,$BD,$23
	.dc.b	$1E,$EF,$FB,$C1,$B1,$FE,$F5,$78,$7B,$B9,$AF,$F2,$A4,$F1,$6A,$9D
	.dc.b	$BF,$E3,$2C,$9F,$43,$E4,$B9,$3F,$42,$3F,$C7,$BF,$35,$A8,$FF,$7C
	.dc.b	$C7,$FB,$5E,$31,$1F,$EF,$88,$FF,$6D,$7E,$77,$FE,$DB,$83,$48,$FF
	.dc.b	$A3,$FF,$2A,$7C,$9F,$FB,$25,$B1,$FD,$FD,$AF,$D7,$E7,$FC,$69,$1F
	.dc.b	$C2,$BF,$90,$F0,$7C,$EC,$8F,$EE,$F8,$FF,$25,$4A,$7D,$BE,$1F,$FA
	.dc.b	$CA,$7F,$9B,$7C,$91,$95,$B7,$1C,$FF,$E9,$07,$37,$B7,$89,$C2,$FA
	.dc.b	$4E,$E3,$C3,$1D,$FE,$C3,$17,$D6,$51,$8B,$E8,$EA,$B8,$61,$DE,$FF
	.dc.b	$BC,$77,$F9,$23,$17,$E3,$C7,$96,$39,$3C,$95,$5F,$1B,$6F,$C5,$3C
	.dc.b	$27,$D6,$68,$B1,$E0,$5D,$5F,$B7,$74,$D8,$D3,$71,$F9,$CE,$E7,$E7
	.dc.b	$7C,$67,$FB,$DE,$FF,$AA,$C7,$FB,$CC,$EA,$E2,$35,$1F,$D6,$9A,$FD
	.dc.b	$17,$D8,$BE,$47,$FD,$43,$3F,$46,$3B,$9E,$C3,$5E,$4F,$07,$C1,$1B
	.dc.b	$9E,$3A,$78,$67,$84,$61,$F1,$1F,$62,$3F,$B6,$C7,$FA,$C3,$5F,$04
	.dc.b	$FA,$DE,$97,$F8,$D7,$38,$7A,$2B,$A5,$7A,$CE,$F5,$EC,$F0,$BF,$D7
	.dc.b	$7A,$FF,$A6,$E5,$18,$9F,$09,$FE,$56,$8C,$BF,$CA,$B4,$D4,$7F,$30
	.dc.b	$6E,$9F,$CC,$3E,$66,$F3,$7F,$90,$EF,$F6,$5D,$DA,$E3,$FE,$5B,$1C
	.dc.b	$CC,$9F,$E6,$4F,$A2,$7C,$27,$97,$3F,$D5,$46,$27,$EE,$C2,$30,$FE
	.dc.b	$01,$AF,$E0,$BE,$5F,$93,$FB,$36,$A6,$9D,$F3,$FE,$93,$E5,$FB,$BB
	.dc.b	$1F,$82,$BE,$87,$DB,$39,$8C,$65,$FE,$AA,$AD,$D7,$A5,$B1,$FC,$BE
	.dc.b	$77,$8F,$F8,$AA,$72,$D8,$7A,$9C,$5B,$5C,$63,$07,$9F,$77,$9D,$F4
	.dc.b	$BC,$4B,$5F,$D8,$23,$04,$7F,$A9,$BE,$F8,$DE,$1F,$D5,$23,$A7,$CD
	.dc.b	$FD,$A3,$ED,$EF,$6F,$FB,$5C,$7C,$5F,$23,$FD,$17,$18,$3E,$13,$8B
	.dc.b	$FE,$CD,$F4,$7F,$84,$F2,$FF,$0A,$DA,$FE,$EB,$1E,$B2,$3F,$D3,$5F
	.dc.b	$76,$FF,$8E,$FC,$DD,$F2,$7F,$8E,$8C,$32,$BF,$80,$F1,$AB,$71,$86
	.dc.b	$E3,$FE,$E9,$6B,$FE,$0E,$E0,$FC,$C9,$F9,$AF,$5E,$D7,$F8,$2A,$3E
	.dc.b	$07,$CC,$E2,$9D,$BF,$E2,$BD,$93,$FA,$2F,$9F,$FD,$1F,$BD,$71,$FF
	.dc.b	$09,$B8,$7F,$B5,$3F,$6C,$8F,$AB,$AF,$FD,$D2,$FF,$5B,$F9,$0F,$D5
	.dc.b	$7C,$A7,$EA,$DF,$D9,$E7,$FB,$AF,$16,$F8,$47,$57,$FE,$D0,$8C,$3F
	.dc.b	$ED,$2F,$BD,$7C,$BF,$26,$C7,$FB,$9F,$F3,$73,$99,$EF,$EA,$BE,$2E
	.dc.b	$99,$9F,$F6,$EC,$65,$FF,$6F,$FD,$F7,$27,$FD,$C3,$F5,$FF,$90,$E4
	.dc.b	$EB,$AF,$F6,$7A,$D3,$FF,$00,$67,$55,$FD,$95,$FC,$BC,$7F,$A4,$9A
	.dc.b	$E7,$BF,$F7,$F1,$88,$FE,$E4,$BB,$B2,$4A,$FE,$17,$19,$7F,$86,$DB
	.dc.b	$E5,$FE,$6B,$63,$D7,$F2,$B8,$9F,$A3,$F5,$FF,$E1,$29,$FF,$7E,$B7
	.dc.b	$1F,$C2,$C7,$F7,$DB,$E5,$18,$3F,$C0,$D3,$BC,$FE,$1F,$6B,$83,$F8
	.dc.b	$EC,$61,$B8,$7F,$83,$67,$5F,$2D,$FC,$CA,$73,$87,$F6,$C5,$F2,$D9
	.dc.b	$35,$E5,$DC,$3E,$AF,$C5,$FA,$BE,$6F,$DE,$D8,$F5,$C7,$F2,$AD,$0D
	.dc.b	$24,$7F,$34,$8F,$CB,$E3,$FD,$C9,$AF,$FB,$21,$AF,$F8,$7D,$AF,$F8
	.dc.b	$0B,$E9,$BF,$C5,$13,$FE,$29,$D7,$E1,$C7,$F6,$E9,$F9,$1A,$A3,$71
	.dc.b	$87,$53,$C3,$A3,$FA,$AC,$7F,$5A,$9D,$7A,$58,$9F,$EB,$3E,$77,$FA
	.dc.b	$6F,$1F,$01,$F3,$5C,$57,$5F,$3F,$7E,$4F,$EC,$3E,$B5,$F1,$94,$CE
	.dc.b	$1C,$AF,$FC,$85,$3A,$8C,$43,$E5,$DE,$17,$EE,$BE,$C3,$F9,$46,$FF
	.dc.b	$A4,$AD,$4E,$7E,$6A,$61,$FD,$8A,$FF,$97,$57,$FB,$D1,$AF,$F9,$2D
	.dc.b	$AF,$8B,$AF,$D0,$72,$7F,$24,$32,$47,$C8,$F7,$A5,$FA,$F8,$FE,$87
	.dc.b	$AD,$8C,$3E,$87,$8B,$F3,$8C,$6B,$F6,$1F,$A3,$A4,$C3,$97,$2B,$E9
	.dc.b	$31,$87,$D9,$AF,$AF,$F4,$98,$C7,$F0,$2F,$EC,$BE,$27,$E8,$7C,$AD
	.dc.b	$CB,$E5,$F0,$23,$31,$FE,$06,$8C,$36,$3F,$C0,$71,$8C,$7F,$C8,$EB
	.dc.b	$B3,$24,$AE,$4F,$F8,$17,$4F,$E7,$B1,$EA,$F3,$BF,$1B,$F4,$9F,$8E
	.dc.b	$7F,$65,$F1,$7F,$E0,$BF,$08,$D2,$57,$F5,$6B,$7D,$27,$E5,$BE,$43
	.dc.b	$F2,$ED,$9F,$8F,$F4,$7F,$9A,$3F,$37,$EC,$58,$F9,$7E,$7F,$94,$8F
	.dc.b	$57,$D1,$DB,$FA,$2F,$17,$F4,$6D,$9F,$A7,$E8,$FE,$98,$FD,$3F,$7D
	.dc.b	$8F,$CB,$FE,$7C,$E6,$FF,$15,$B6,$3F,$C6,$6D,$E5,$FF,$38,$B5,$FF
	.dc.b	$4E,$76,$FE,$DF,$E9,$3F,$C8,$FF,$A7,$FF,$3E,$EB,$FD,$EC,$63,$0C
	.dc.b	$FC,$7F,$F9,$33,$E2,$3F,$C4,$F5,$FC,$92,$D3,$9F,$7B,$FB,$0F,$EA
	.dc.b	$2B,$FF,$1F,$B7,$0F,$F2,$3B,$73,$FF,$42,$B5,$F2,$E0,$F8,$FF,$2F
	.dc.b	$91,$E6,$6D,$FE,$86,$7F,$AE,$57,$F1,$26,$E7,$E5,$F4,$FD,$BE,$C7
	.dc.b	$F9,$9B,$E0,$C3,$D9,$FF,$2E,$38,$6B,$9C,$7F,$D3,$2E,$5F,$F0,$DE
	.dc.b	$1F,$B4,$7C,$5E,$5F,$CA,$F4,$FE,$5B,$F3,$3D,$59,$3E,$7F,$56,$EB
	.dc.b	$FD,$3F,$1F,$1F,$C7,$F8,$09,$5E,$08,$FF,$7D,$BE,$0F,$27,$EB,$9F
	.dc.b	$63,$C8,$B1,$FC,$0F,$F2,$9F,$F2,$AE,$1F,$CD,$1B,$C3,$F5,$FE,$87
	.dc.b	$FD,$0B,$18,$AF,$F7,$26,$3F,$E2,$8F,$0F,$E1,$D9,$A4,$9F,$F3,$FF
	.dc.b	$E7,$3F,$AA,$6F,$91,$FE,$7C,$8C,$65,$FE,$C0,$6B,$F7,$3E,$8B,$E9
	.dc.b	$61,$FC,$BE,$FE,$D4,$63,$C3,$91,$F4,$27,$55,$FE,$F3,$3E,$CB,$87
	.dc.b	$FD,$ED,$97,$F9,$02,$BF,$D1,$1C,$8F,$A1,$FC,$E9,$F9,$EF,$A2,$B8
	.dc.b	$FE,$26,$3F,$D1,$9F,$95,$38,$33,$97,$F8,$0F,$E3,$93,$9B,$FC,$87
	.dc.b	$CE,$7E,$A0,$FD,$4F,$CE,$58,$C9,$F9,$5F,$9C,$C6,$1F,$95,$7F,$C8
	.dc.b	$BE,$83,$E3,$A9,$FE,$48,$FC,$6D,$E4,$63,$0F,$EA,$B5,$B7,$FA,$BF
	.dc.b	$BF,$FE,$44,$C7,$FA,$3C,$E5,$FE,$93,$19,$E4,$FE,$B0,$EE,$3E,$53
	.dc.b	$51,$FE,$B3,$F4,$3F,$D6,$B1,$98,$FF,$5B,$79,$DF,$E4,$4F,$95,$7C
	.dc.b	$EF,$09,$FF,$5C,$4E,$63,$1F,$21,$FC,$E7,$3E,$B7,$C2,$F9,$C6,$2B
	.dc.b	$96,$1F,$DC,$23,$13,$FE,$C5,$D3,$A3,$C0,$5F,$37,$FD,$91,$1A,$5F
	.dc.b	$61,$FE,$02,$F8,$6F,$F6,$AB,$63,$F1,$46,$BC,$C4,$E9,$B1,$AF,$99
	.dc.b	$7C,$BF,$F0,$88,$C3,$5F,$CA,$46,$BB,$BF,$DC,$6D,$8E,$F8,$7F,$9D
	.dc.b	$7D,$3F,$1F,$9A,$69,$27,$7E,$57,$CD,$D7,$DD,$68,$C7,$07,$D1,$8C
	.dc.b	$57,$C8,$3E,$0F,$95,$A7,$EE,$C7,$F9,$41,$D3,$C1,$F9,$73,$CE,$B8
	.dc.b	$47,$23,$7D,$DF,$EF,$E7,$E6,$B6,$BF,$51,$1F,$3A,$31,$E1,$FE,$17
	.dc.b	$9F,$FC,$51,$5A,$AF,$F0,$EB,$63,$33,$BE,$3F,$63,$6B,$D2,$D7,$EE
	.dc.b	$B7,$A3,$FD,$EB,$1F,$E2,$3F,$B0,$7C,$97,$1D,$DE,$27,$F6,$7A,$FE
	.dc.b	$91,$EB,$E6,$F7,$1C,$6F,$04,$7A,$75,$78,$FC,$3F,$C1,$F9,$1D,$71
	.dc.b	$FF,$05,$D7,$F5,$38,$D7,$46,$23,$F8,$53,$77,$FF,$56,$DF,$29,$FF
	.dc.b	$2A,$35,$FF,$35,$FD,$67,$F1,$F9,$FF,$28,$7D,$87,$F8,$D6,$7F,$32
	.dc.b	$1C,$3F,$9C,$6B,$FC,$9D,$AF,$F3,$EE,$7F,$CD,$BF,$E6,$08,$E3,$3D
	.dc.b	$BE,$17,$E8,$75,$FE,$BB,$2F,$95,$F3,$1F,$97,$C3,$FC,$A0,$D7,$FD
	.dc.b	$02,$DB,$FF,$64,$D3,$F0,$37,$C4,$FD,$9C,$9F,$0F,$F8,$B6,$E3,$EC
	.dc.b	$6B,$EB,$8F,$7D,$FF,$00,$F9,$2F,$2E,$BF,$EC,$9A,$FF,$3E,$4E,$5F
	.dc.b	$F3,$27,$D8,$BE,$17,$D7,$83,$D0,$A4,$AF,$5F,$BD,$E5,$F8,$2F,$21
	.dc.b	$23,$C8,$FD,$0E,$B7,$0F,$F8,$26,$7E,$1F,$B9,$F1,$8F,$C8,$72,$FF
	.dc.b	$BF,$E3,$EE,$FE,$73,$F8,$1F,$03,$F9,$7F,$42,$31,$28,$C7,$4F,$D0
	.dc.b	$F1,$3E,$47,$99,$F1,$C9,$5F,$1D,$F8,$BE,$57,$FC,$01,$8F,$86,$D7
	.dc.b	$FC,$1C,$DE,$AF,$F6,$DE,$6F,$88,$FD,$37,$DD,$FF,$A8,$6F,$F2,$7F
	.dc.b	$87,$FF,$80,$5B,$ED,$6B,$D9,$FD,$96,$8F,$57,$F8,$23,$6B,$F2,$AD
	.dc.b	$7F,$97,$B5,$FF,$28,$F6,$7F,$61,$7F,$42,$FF,$A1,$DB,$EE,$FE,$E6
	.dc.b	$35,$1F,$8E,$FA,$6F,$C9,$FB,$F2,$3F,$27,$1E,$4F,$2B,$C6,$48,$F1
	.dc.b	$BF,$67,$A1,$F4,$9F,$9C,$24,$FF,$AA,$F8,$7E,$13,$FE,$01,$FC,$13
	.dc.b	$6F,$E8,$3E,$97,$F4,$67,$F0,$38,$CC,$7F,$AE,$3F,$40,$67,$FC,$9D
	.dc.b	$B1,$FB,$86,$BF,$D4,$9A,$FF,$6B,$E9,$57,$ED,$A3,$C5,$E4,$F7,$3F
	.dc.b	$8E,$78,$7F,$ED,$3E,$8F,$E1,$74,$F9,$CF,$DB,$FD,$2C,$C3,$F8,$6C
	.dc.b	$E5,$FF,$6F,$E5,$E8,$7F,$C5,$8D,$7F,$BE,$35,$FF,$02,$B7,$2F,$2E
	.dc.b	$FF,$04,$93,$FE,$D3,$CB,$F1,$DD,$FF,$E0,$5F,$B1,$FE,$03,$3F,$EE
	.dc.b	$E6,$BF,$ED,$E6,$E9,$FE,$ED,$FB,$4F,$A2,$DF,$90,$FE,$A1,$06,$92
	.dc.b	$BF,$91,$C3,$C5,$FE,$8C,$EE,$3F,$DF,$AD,$7F,$DD,$AE,$DF,$47,$C5
	.dc.b	$7F,$1E,$E1,$91,$82,$0C,$19,$7F,$8A,$3C,$6F,$E1,$D7,$F5,$0F,$AB
	.dc.b	$F1,$5F,$CA,$3E,$6E,$61,$FC,$A1,$FB,$3E,$E4,$8C,$10,$ED,$87,$5D
	.dc.b	$FF,$BC,$46,$0F,$F1,$BF,$D8,$FE,$06,$7F,$C6,$5F,$8B,$7D,$09,$FF
	.dc.b	$1D,$B5,$4E,$DE,$3F,$25,$FD,$AE,$31,$1A,$AC,$4F,$F6,$FF,$B4,$FA
	.dc.b	$71,$FE,$3F,$AF,$F5,$FD,$7F,$92,$36,$F8,$6F,$43,$13,$FE,$45,$AE
	.dc.b	$31,$E8,$F8,$3F,$07,$C8,$C8,$D9,$92,$9E,$8F,$83,$F3,$4C,$8C,$E3
	.dc.b	$EF,$E1,$FE,$67,$D9,$FD,$5B,$D6,$95,$FD,$5A,$3F,$4F,$F0,$12,$3F
	.dc.b	$2C,$FE,$AB,$E3,$7F,$C1,$9D,$DF,$E0,$A9,$F9,$C7,$EC,$BE,$9F,$27
	.dc.b	$C7,$E4,$F9,$1E,$47,$F8,$8F,$4F,$E3,$15,$FE,$4C,$6B,$FE,$97,$6B
	.dc.b	$F2,$7E,$47,$8B,$EA,$9F,$A5,$1F,$E1,$7E,$97,$CB,$52,$57,$C6,$DD
	.dc.b	$F6,$7F,$D2,$1B,$EA,$EA,$7F,$0D,$0B,$E7,$1C,$6B,$FA,$7A,$E2,$FF
	.dc.b	$35,$73,$FD,$A8,$C7,$AF,$FE,$2F,$9F,$EB,$D1,$F1,$1D,$F4,$7F,$A0
	.dc.b	$57,$99,$BF,$3A,$57,$DA,$9C,$BF,$73,$F1,$6F,$8F,$8F,$76,$E1,$FE
	.dc.b	$8B,$6C,$FF,$A5,$29,$F2,$FE,$3E,$4F,$F2,$94,$7D,$79,$FF,$62,$35
	.dc.b	$F2,$46,$BE,$13,$84,$78,$E3,$D9,$1F,$E4,$BF,$B2,$76,$DE,$F9,$7F
	.dc.b	$92,$9C,$B7,$0E,$15,$23,$07,$E6,$3E,$C5,$EE,$FF,$95,$DF,$F1,$26
	.dc.b	$BF,$EA,$76,$CF,$FA,$C1,$DB,$C7,$87,$85,$48,$C1,$FE,$46,$FA,$BF
	.dc.b	$F3,$97,$9A,$69,$23,$FC,$2D,$E2,$BE,$57,$FC,$B1,$3F,$E4,$BC,$BF
	.dc.b	$C2,$2D,$AF,$F4,$95,$6D,$ED,$F4,$BF,$E4,$0F,$3E,$46,$08,$7F,$B9
	.dc.b	$69,$3D,$8F,$EA,$BE,$C7,$F9,$9F,$E6,$BF,$13,$5F,$D7,$5A,$FF,$88
	.dc.b	$9A,$FF,$88,$5B,$D9,$FD,$DB,$E9,$79,$6F,$2B,$F4,$5E,$C2,$7D,$BF
	.dc.b	$49,$FE,$64,$E0,$F8,$8D,$D0,$D3,$FC,$75,$F5,$BF,$4A,$77,$87,$CD
	.dc.b	$9F,$F4,$44,$E6,$BF,$AA,$F0,$7F,$3F,$DF,$FC,$27,$D9,$FC,$83,$5F
	.dc.b	$F2,$1F,$CC,$FC,$EF,$8B,$99,$FD,$76,$BF,$B4,$CB,$F5,$77,$C7,$E2
	.dc.b	$BF,$DB,$53,$FE,$EB,$C1,$18,$8F,$E9,$CD,$7F,$D9,$0D,$8D,$F8,$3F
	.dc.b	$E9,$F8,$FD,$AC,$EB,$C1,$C4,$FF,$A6,$7F,$7E,$7C,$BD,$25,$7F,$A4
	.dc.b	$BE,$CB,$FA,$37,$11,$FE,$7F,$1F,$EF,$F3,$FC,$AC,$E1,$FF,$0C,$35
	.dc.b	$FF,$2E,$B5,$DF,$6F,$FA,$A6,$39,$E9,$FE,$A2,$6F,$0F,$09,$FF,$5F
	.dc.b	$7C,$C7,$AA,$CD,$26,$7A,$F7,$FF,$DA,$91,$FE,$2A,$91,$1C,$36,$3F
	.dc.b	$DB,$5F,$65,$FA,$18,$34,$FD,$27,$03,$EB,$BF,$13,$3F,$CC,$5A,$FF
	.dc.b	$2E,$6E,$9F,$83,$6E,$7F,$E0,$5E,$2F,$F2,$2C,$FF,$B0,$3F,$14,$FE
	.dc.b	$61,$7F,$EC,$06,$BF,$DF,$77,$46,$27,$F9,$A7,$33,$D3,$F8,$BF,$05
	.dc.b	$FE,$13,$C3,$F8,$7F,$D5,$7C,$07,$E3,$FC,$25,$F1,$7F,$D2,$51,$83
	.dc.b	$E2,$B8,$E2,$5F,$38,$C3,$D7,$EC,$F5,$7D,$5B,$7F,$39,$DE,$77,$1F
	.dc.b	$16,$3D,$23,$BC,$62,$FC,$8D,$BC,$9F,$5F,$0F,$14,$72,$6C,$7D,$B9
	.dc.b	$E2,$D3,$89,$F8,$68,$DB,$6B,$FC,$6D,$8F,$36,$C7,$76,$BB,$1C,$3F
	.dc.b	$CD,$1F,$99,$F1,$1A,$7F,$1A,$D4,$D2,$3F,$50,$FA,$9F,$1D,$FE,$26
	.dc.b	$BF,$A7,$1C,$8D,$1F,$98,$8D,$65,$FA,$F8,$EA,$68,$46,$9F,$DA,$57
	.dc.b	$C2,$73,$C5,$5E,$3A,$F8,$2B,$FC,$6E,$F8,$A3,$37,$FB,$D9,$CB,$E1
	.dc.b	$BE,$91,$DE,$7E,$19,$E3,$7C,$5C,$3F,$98,$F7,$7F,$C8,$7A,$7E,$7D
	.dc.b	$B9,$7E,$57,$87,$FC,$20,$FE,$86,$33,$87,$D0,$E4,$77,$63,$7F,$D9
	.dc.b	$37,$97,$F2,$F8,$7E,$CE,$38,$DF,$FB,$D3,$D3,$B9,$7F,$88,$6B,$47
	.dc.b	$F9,$31,$C4,$FF,$8F,$87,$CF,$7F,$88,$5F,$E3,$CD,$7F,$58,$D7,$F8
	.dc.b	$6B,$5F,$3C,$FF,$96,$1A,$75,$F0,$FC,$AE,$35,$FF,$33,$EC,$FB,$3F
	.dc.b	$F0,$ED,$FF,$98,$A3,$1E,$97,$C4,$75,$E9,$DA,$4D,$3C,$D3,$C2,$BF
	.dc.b	$10,$DC,$3F,$9C,$35,$FF,$67,$53,$D3,$E8,$75,$F9,$AB,$7F,$E4,$FF
	.dc.b	$1B,$FD,$98,$DF,$80,$FA,$91,$8C,$3F,$C9,$EA,$32,$7C,$9E,$3F,$EB
	.dc.b	$6D,$8D,$57,$E7,$67,$2F,$A3,$C9,$7D,$67,$C1,$6C,$7F,$98,$79,$78
	.dc.b	$E3,$FE,$11,$AD,$1F,$87,$E3,$FF,$02,$57,$F8,$9E,$31,$E5,$FE,$24
	.dc.b	$E9,$F2,$D6,$7F,$C5,$8E,$1F,$13,$F3,$31,$34,$F6,$7D,$9F,$F9,$DA
	.dc.b	$7E,$C7,$A0,$F3,$BF,$31,$39,$CF,$F9,$E4,$E6,$B7,$5F,$62,$31,$3F
	.dc.b	$59,$AF,$F8,$9E,$3C,$5F,$29,$F0,$CF,$D1,$72,$DB,$FF,$22,$47,$C1
	.dc.b	$E6,$C5,$FF,$A0,$B6,$FC,$6B,$FE,$11,$FB,$3F,$99,$F5,$7F,$AB,$6E
	.dc.b	$1F,$E4,$D8,$F8,$04,$7E,$A3,$FA,$1F,$55,$E1,$FE,$C1,$FA,$B7,$FE
	.dc.b	$A8,$73,$5F,$EB,$A7,$3D,$7F,$E8,$D6,$C7,$F5,$75,$F3,$BC,$F9,$3A
	.dc.b	$3F,$BC,$F5,$5E,$1D,$E2,$C3,$70,$FF,$63,$CA,$78,$7C,$5E,$73,$FF
	.dc.b	$5D,$CA,$3F,$83,$FD,$97,$E6,$74,$FF,$50,$3F,$8C,$6F,$C4,$EB,$B6
	.dc.b	$65,$FC,$CF,$15,$FE,$49,$FF,$12,$35,$FE,$E4,$D7,$FC,$27,$F6,$0F
	.dc.b	$96,$FE,$2E,$DA,$FF,$6F,$D3,$93,$35,$FC,$73,$EC,$7F,$CB,$35,$FE
	.dc.b	$E4,$7F,$85,$3F,$DF,$DC,$E1,$FE,$D8,$72,$FF,$88,$9A,$9F,$AA,$EF
	.dc.b	$C8,$ED,$C9,$FC,$BF,$EB,$DF,$17,$FC,$C7,$F1,$37,$D4,$66,$30,$CC
	.dc.b	$BF,$DA,$93,$A7,$FD,$AD,$5C,$9C,$CF,$AF,$E5,$BD,$9F,$2D,$67,$D8
	.dc.b	$E1,$F9,$DA,$23,$F9,$83,$63,$31,$FC,$CF,$F0,$1F,$C0,$47,$F9,$03
	.dc.b	$93,$FD,$B5,$18,$DB,$E2,$7C,$FF,$F4,$68,$C4,$FF,$9E,$7E,$9F,$FD
	.dc.b	$33,$7F,$EE,$D6,$DF,$FB,$A9,$FE,$D6,$DE,$DF,$E1,$DE,$17,$F0,$FF
	.dc.b	$32,$DF,$FB,$D7,$97,$F9,$63,$E3,$3F,$C6,$9F,$67,$FD,$83,$06,$5F
	.dc.b	$EE,$F8,$FC,$EC,$EE,$7F,$CB,$31,$B9,$D7,$77,$F5,$E8,$DC,$F2,$AE
	.dc.b	$BD,$D9,$AD,$3F,$8E,$F0,$38,$77,$5A,$ED,$6C,$3F,$49,$18,$6C,$75
	.dc.b	$3C,$7E,$97,$29,$D3,$E3,$3F,$B8,$EB,$F3,$23,$DF,$18,$C7,$FB,$D6
	.dc.b	$FF,$90,$D3,$FD,$51,$AB,$CE,$CF,$57,$F3,$96,$63,$F2,$DC,$B4,$FA
	.dc.b	$DE,$4F,$43,$AE,$F6,$32,$7D,$3F,$17,$1B,$7F,$23,$7E,$57,$C9,$7C
	.dc.b	$BF,$92,$B5,$F0,$5C,$3F,$BF,$E1,$47,$F8,$05,$B1,$98,$FF,$02,$70
	.dc.b	$FC,$8D,$93,$FE,$73,$8C,$3F,$82,$FA,$AF,$8F,$C7,$FD,$77,$A2,$4F
	.dc.b	$E4,$7C,$AB,$AF,$D1,$1F,$9F,$FE,$51,$FF,$59,$35,$FF,$40,$35,$FF
	.dc.b	$50,$B7,$D4,$7D,$7F,$92,$FF,$1D,$E9,$BF,$71,$FF,$53,$C7,$F8,$46
	.dc.b	$75,$87,$E0,$6B,$4F,$F8,$4E,$1F,$0D,$4F,$BB,$F4,$76,$FF,$1B,$97
	.dc.b	$F8,$B1,$FE,$50,$DB,$FF,$2E,$35,$F5,$FB,$5F,$E1,$AC,$3F,$15,$1F
	.dc.b	$E7,$2C,$BF,$63,$C3,$FE,$28,$9C,$E7,$E9,$F7,$3F,$C8,$D5,$FB,$8A
	.dc.b	$FF,$7A,$37,$2F,$F6,$1D,$3B,$3F,$85,$63,$ED,$AD,$3F,$23,$7F,$F1
	.dc.b	$F6,$86,$93,$0F,$F0,$C5,$7A,$F1,$EF,$ED,$F8,$B5,$F1,$BF,$E5,$9F
	.dc.b	$7F,$FB,$8C,$FF,$90,$1F,$9A,$FE,$1B,$2F,$F4,$7C,$6B,$0F,$F6,$CF
	.dc.b	$0A,$37,$1F,$7B,$ED,$7C,$A4,$7E,$6B,$FE,$2F,$F3,$FF,$A0,$D3,$F0
	.dc.b	$5F,$E8,$B8,$F7,$E0,$FE,$B3,$E7,$FF,$93,$67,$38,$FF,$BA,$3D,$D0
	.dc.b	$F8,$9F,$F3,$AB,$87,$FC,$E9,$F5,$BF,$C4,$B1,$F5,$4E,$6B,$F8,$6F
	.dc.b	$D9,$FF,$41,$DF,$FD,$34,$D7,$FD,$AC,$1C,$3F,$ED,$96,$AD,$3F,$BC
	.dc.b	$7A,$DF,$E6,$3E,$CF,$EE,$D7,$F7,$39,$3D,$F1,$8B,$FC,$47,$8D,$F8
	.dc.b	$43,$CF,$E8,$DB,$F2,$F8,$7F,$E7,$BC,$CD,$24,$7F,$AC,$74,$FF,$72
	.dc.b	$FC,$5F,$C2,$F8,$63,$F9,$0F,$1F,$0F,$C6,$E0,$78,$14,$FF,$4D,$7D
	.dc.b	$A7,$F9,$E3,$1F,$E7,$D8,$FF,$42,$9F,$F4,$CE,$6F,$9B,$F2,$12,$BF
	.dc.b	$C2,$11,$97,$F2,$3E,$EF,$E8,$0F,$AB,$F8,$EB,$DF,$C4,$F6,$CF,$FC
	.dc.b	$1D,$C5,$FC,$2F,$F5,$5F,$F6,$EC,$E7,$DD,$FF,$54,$C6,$27,$5D,$BF
	.dc.b	$E9,$F8,$C4,$7D,$86,$A2,$38,$7E,$C7,$8E,$F2,$BF,$84,$1F,$88,$F8
	.dc.b	$DB,$8F,$F0,$F8,$FD,$CF,$A1,$8B,$FD,$A3,$F8,$EF,$95,$95,$FC,$63
	.dc.b	$BD,$F2,$F8,$BF,$E2,$D0,$AF,$E0,$7E,$4E,$77,$FE,$06,$FF,$B1,$BC
	.dc.b	$97,$DA,$53,$FD,$93,$E6,$7F,$9C,$AB,$F6,$98,$FF,$B1,$27,$59,$FF
	.dc.b	$68,$BF,$97,$5A,$79,$C6,$78,$BB,$1F,$EC,$78,$FA,$37,$FE,$CC,$8C
	.dc.b	$19,$F3,$FF,$81,$7C,$6B,$E5,$BF,$D9,$FF,$2B,$87,$FD,$A3,$36,$FF
	.dc.b	$DA,$33,$5F,$F6,$CB,$70,$36,$70,$EF,$ED,$1F,$36,$F4,$FF,$CB,$91
	.dc.b	$86,$C6,$70,$FD,$B8,$9D,$38,$F9,$47,$D3,$FF,$08,$8C,$35,$FD,$FC
	.dc.b	$6B,$C2,$FF,$35,$36,$3A,$D7,$F9,$FB,$B3,$C7,$EF,$9A,$49,$DF,$93
	.dc.b	$F9,$17,$FD,$F1,$5F,$E5,$67,$FA,$D9,$97,$0E,$A3,$87,$17,$F1,$29
	.dc.b	$CF,$0F,$D1,$AF,$CE,$46,$38,$79,$5F,$7A,$D5,$7C,$3F,$88,$C5,$77
	.dc.b	$EC,$B3,$CB,$DA,$AF,$C9,$6E,$3F,$A5,$8F,$D5,$61,$D2,$31,$1F,$12
	.dc.b	$BF,$D2,$0F,$D1,$6C,$6A,$3D,$5F,$49,$F0,$F9,$FF,$33,$3C,$9B,$59
	.dc.b	$D3,$E2,$55,$BF,$9D,$D8,$FA,$CC,$6E,$8C,$4F,$EC,$67,$3F,$69,$FE
	.dc.b	$39,$8F,$F0,$5B,$5F,$D9,$C7,$29,$E3,$1F,$E0,$B9,$D6,$5F,$D4,$E7
	.dc.b	$F9,$04,$FE,$F7,$C8,$E0,$79,$7D,$3F,$C6,$B7,$9B,$73,$FC,$02,$36
	.dc.b	$FF,$9A,$67,$6F,$F2,$EF,$B4,$7C,$37,$FA,$E7,$A1,$18,$9F,$E7,$CD
	.dc.b	$7F,$AA,$C6,$A7,$E8,$FD,$A7,$FB,$5A,$7F,$D8,$4D,$7F,$4A,$D7,$FA
	.dc.b	$EB,$5F,$EA,$DE,$DF,$CC,$F5,$8D,$3E,$7F,$C2,$7E,$7E,$7D,$1E,$DF
	.dc.b	$9F,$23,$49,$1F,$D5,$79,$DE,$CF,$DB,$86,$4F,$2B,$0D,$8F,$EA,$35
	.dc.b	$C1,$FF,$21,$E7,$FD,$25,$BE,$77,$9C,$FC,$36,$DC,$FE,$CE,$E2,$E8
	.dc.b	$F9,$8D,$BF,$ED,$0D,$7E,$5B,$5F,$DC,$74,$7F,$B7,$FE,$47,$E2,$7D
	.dc.b	$0C,$9F,$F5,$54,$7C,$7F,$56,$47,$99,$FC,$27,$D9,$FE,$0B,$AB,$FC
	.dc.b	$23,$8F,$F4,$4A,$F0,$FA,$1E,$42,$3F,$DE,$FC,$87,$8D,$F4,$FE,$A2
	.dc.b	$4F,$F7,$49,$CB,$F8,$BF,$B2,$FE,$F6,$FF,$22,$EC,$46,$23,$FC,$37
	.dc.b	$18,$D3,$F9,$3C,$7F,$86,$67,$4F,$F1,$7E,$27,$E0,$9F,$0E,$3F,$C6
	.dc.b	$0D,$7F,$C6,$2D,$8D,$F1,$74,$C1,$CD,$FE,$18,$BF,$23,$5F,$1B,$5F
	.dc.b	$E4,$D1,$C3,$DE,$F2,$F2,$3B,$3F,$C3,$75,$F8,$2F,$47,$F2,$C7,$E5
	.dc.b	$3D,$CF,$C0,$E9,$F3,$67,$F4,$AD,$7D,$8D,$7D,$7C,$5D,$AD,$F4,$BF
	.dc.b	$A1,$F4,$E5,$7F,$B9,$E3,$F4,$3F,$43,$FA,$5F,$94,$93,$FE,$EB,$8F
	.dc.b	$D4,$7C,$9F,$A3,$BC,$D2,$4F,$D5,$FA,$CF,$EA,$6F,$E7,$3D,$6F,$D8
	.dc.b	$1F,$C2,$E3,$31,$FE,$9D,$CD,$F0,$7F,$B5,$F5,$E4,$FF,$BE,$E3,$F9
	.dc.b	$1F,$E5,$BC,$23,$F3,$2D,$7F,$D1,$AD,$7F,$AC,$37,$95,$87,$CF,$9F
	.dc.b	$F5,$8F,$EE,$3F,$E0,$EF,$03,$FC,$E8,$DF,$47,$FD,$03,$5B,$AF,$F5
	.dc.b	$94,$BE,$AF,$F9,$AB,$63,$9C,$7F,$01,$FB,$3F,$E0,$F1,$FE,$FE,$FB
	.dc.b	$37,$2F,$F4,$A6,$C7,$E5,$9A,$FF,$7E,$6B,$FE,$76,$5C,$9C,$49,$1C
	.dc.b	$5F,$BD,$E8,$F9,$3E,$46,$4F,$FB,$3B,$8B,$E7,$ED,$FE,$91,$F2,$5F
	.dc.b	$2F,$FC,$83,$09,$7F,$ED,$49,$CB,$C7,$EC,$FD,$FE,$FC,$8C,$10,$FB
	.dc.b	$51,$FC,$A8,$FE,$B3,$2F,$77,$F9,$27,$D9,$FE,$F2,$BF,$DD,$8D,$CB
	.dc.b	$FC,$7E,$DE,$4F,$F0,$67,$D9,$7D,$3A,$FF,$2F,$3F,$E4,$47,$FD,$32
	.dc.b	$D7,$71,$FD,$DA,$31,$1C,$38,$FF,$9A,$47,$38,$34,$FB,$4F,$FB,$FD
	.dc.b	$FF,$76,$BF,$EF,$6F,$C9,$FF,$6A,$FF,$BC,$DB,$87,$F9,$5A,$77,$3C
	.dc.b	$74,$F1,$11,$FE,$27,$1F,$E4,$07,$FC,$77,$E2,$7F,$01,$F5,$3F,$17
	.dc.b	$E7,$49,$FB,$11,$FD,$53,$E8,$A4,$7D,$B7,$E6,$F9,$5F,$D3,$FF,$01
	.dc.b	$FF,$63,$4F,$FA,$0F,$3F,$F3,$4C,$62,$11,$8F,$93,$FE,$AD,$DA,$F9
	.dc.b	$9F,$8B,$C7,$F2,$49,$7E,$87,$E6,$79,$92,$3D,$6D,$7F,$D9,$7B,$7E
	.dc.b	$33,$D2,$FD,$BF,$D4,$49,$F2,$72,$FF,$A5,$EB,$51,$83,$FC,$9D,$D4
	.dc.b	$BF,$F1,$0B,$63,$53,$C1,$F5,$D7,$3F,$B4,$FE,$A4,$FF,$94,$F9,$5E
	.dc.b	$BF,$C5,$F3,$E5,$F7,$3F,$CE,$31,$FD,$DE,$7F,$C9,$71,$FE,$E4,$6B
	.dc.b	$CF,$E1,$BF,$C8,$31,$88,$FF,$72,$7D,$AB,$E5,$7F,$C4,$CD,$8F,$F1
	.dc.b	$13,$5F,$39,$FE,$B7,$FB,$5F,$E8,$DC,$BF,$E9,$EB,$7D,$47,$62,$57
	.dc.b	$F5,$A8,$CB,$79,$3C,$64,$70,$7F,$60,$FB,$5F,$CD,$DB,$6F,$E4,$BF
	.dc.b	$49,$FC,$A8,$FF,$0B,$4E,$A3,$FC,$F4,$FE,$6A,$BF,$CF,$81,$C3,$AF
	.dc.b	$1B,$FC,$C7,$18,$8F,$F6,$97,$DA,$FF,$8A,$5F,$7F,$1B,$CE,$7D,$1F
	.dc.b	$F3,$A3,$FC,$4D,$5B,$CF,$FD,$7E,$BE,$83,$FC,$DE,$DC,$3F,$BD,$B5
	.dc.b	$FF,$51,$B5,$FF,$0D,$F9,$EF,$A5,$FE,$DA,$D8,$FD,$7B,$5F,$CB,$B5
	.dc.b	$FE,$EB,$F5,$BF,$E7,$EF,$48,$D2,$47,$F6,$BF,$3B,$FC,$8E,$47,$FC
	.dc.b	$C1,$1F,$E5,$9C,$BE,$84,$2B,$9F,$C5,$6E,$3F,$B0,$5F,$F9,$B6,$F8
	.dc.b	$70,$FE,$97,$F3,$DF,$F2,$25,$7E,$B1,$AF,$F8,$E1,$AE,$FC,$8F,$F0
	.dc.b	$A6,$9F,$A0,$8F,$F1,$33,$5F,$F1,$2B,$5F,$F0,$9B,$5F,$DA,$68,$FA
	.dc.b	$1F,$92,$48,$F9,$2F,$F8,$87,$17,$D2,$52,$4F,$F8,$D6,$32,$FF,$1E
	.dc.b	$CD,$F5,$BF,$E6,$E2,$47,$C8,$F8,$CF,$84,$61,$FF,$02,$46,$1F,$E0
	.dc.b	$3F,$AD,$3E,$2F,$FC,$52,$D8,$FE,$56,$D7,$F8,$9B,$5F,$F2,$FF,$C3
	.dc.b	$7E,$E7,$A0,$D2,$78,$5F,$E3,$DF,$CB,$5E,$BF,$FA,$85,$B7,$FE,$9D
	.dc.b	$6B,$FB,$CE,$67,$DC,$7F,$C0,$08,$FF,$51,$7D,$C7,$F5,$29,$34,$9C
	.dc.b	$FF,$E4,$B7,$FA,$8C,$FF,$91,$9A,$8F,$C5,$38,$6F,$AF,$FE,$D4,$E1
	.dc.b	$7C,$CF,$FA,$45,$B3,$FC,$7D,$AF,$F1,$21,$F3,$7F,$EE,$AE,$CF,$CB
	.dc.b	$C6,$3D,$3F,$F6,$FC,$FF,$49,$8F,$E0,$6D,$7F,$E0,$03,$5F,$37,$DC
	.dc.b	$7F,$08,$AE,$73,$FD,$9B,$3D,$F9,$BF,$EA,$5A,$D4,$7F,$43,$FB,$6F
	.dc.b	$C3,$E1,$E5,$9C,$BD,$0F,$F7,$1C,$6B,$0F,$F1,$BF,$A1,$FE,$32,$24
	.dc.b	$FF,$B4,$7E,$E3,$FB,$73,$EF,$C9,$F1,$3E,$C2,$47,$FA,$1A,$32,$DF
	.dc.b	$85,$FE,$CE,$BE,$E1,$F5,$7F,$02,$FE,$A4,$E6,$30,$FD,$79,$DB,$F6
	.dc.b	$6F,$FC,$BB,$96,$6B,$E4,$E1,$DB,$C8,$DD,$FD,$4C,$7A,$C6,$30,$F2
	.dc.b	$CF,$F2,$5B,$E3,$87,$E1,$6B,$58,$73,$9F,$1D,$6C,$F1,$61,$D2,$38
	.dc.b	$C6,$4F,$C1,$CF,$C0,$EB,$FE,$5D,$8F,$A2,$51,$89,$D6,$1C,$A3,$F9
	.dc.b	$94,$FF,$A9,$0D,$C7,$F4,$A7,$11,$AF,$90,$FF,$01,$4E,$5B,$3F,$D4
	.dc.b	$1F,$66,$1F,$D4,$BE,$4F,$F9,$5D,$7C,$6A,$FD,$25,$69,$FA,$73,$AA
	.dc.b	$E9,$1E,$D1,$19,$7B,$19,$70,$FD,$C9,$FA,$CF,$4E,$1A,$E4,$FC,$01
	.dc.b	$92,$9F,$B7,$9F,$F2,$8B,$E0,$8F,$F6,$9B,$4F,$F6,$AC,$7F,$96,$70
	.dc.b	$78,$AF,$A0,$FE,$93,$B7,$FA,$51,$B8,$7F,$03,$EA,$70,$FF,$BA,$A3
	.dc.b	$FD,$37,$3A,$CB,$EB,$F1,$BC,$3C,$47,$FB,$3F,$93,$5A,$2B,$58,$7F
	.dc.b	$07,$9D,$F4,$7F,$A1,$A3,$FC,$51,$6F,$A1,$CD,$74,$8C,$4E,$BC,$6F
	.dc.b	$82,$47,$E0,$BF,$E7,$4B,$F4,$4F,$F1,$36,$ED,$FE,$7D,$CF,$C3,$F0
	.dc.b	$5E,$BD,$DF,$71,$FE,$57,$CB,$F1,$91,$8F,$53,$FD,$04,$D8,$CC,$EF
	.dc.b	$2F,$F0,$DC,$F2,$FB,$1D,$C9,$A4,$AE,$BE,$1E,$5F,$E8,$11,$E3,$97
	.dc.b	$AB,$E5,$AD,$61,$FE,$20,$F1,$DE,$77,$DD,$6E,$3F,$E8,$06,$CF,$F9
	.dc.b	$A1,$AF,$F0,$CE,$47,$EB,$5F,$B0,$AF,$F4,$3B,$67,$FC,$DE,$D7,$83
	.dc.b	$5F,$C3,$79,$BF,$E5,$59,$FF,$61,$7B,$FF,$2E,$31,$3F,$EF,$18,$FC
	.dc.b	$1F,$19,$A6,$FD,$CB,$7F,$E1,$EF,$99,$F9,$F8,$1A,$48,$FF,$17,$FC
	.dc.b	$17,$E7,$7B,$7F,$D8,$51,$82,$65,$FE,$14,$F4,$7E,$CF,$A7,$FE,$06
	.dc.b	$9F,$F0,$CC,$60,$FD,$37,$E7,$DD,$E3,$FD,$52,$DA,$FF,$48,$B5,$F9
	.dc.b	$4D,$7E,$1F,$D3,$FB,$BD,$EF,$84,$DA,$FF,$73,$45,$9F,$EF,$9A,$2B
	.dc.b	$1D,$9F,$E3,$CE,$9F,$F2,$2F,$CE,$FC,$22,$63,$FE,$21,$7E,$07,$C2
	.dc.b	$C8,$FF,$28,$B7,$1F,$E4,$9E,$D7,$D2,$7F,$CD,$ED,$7F,$A6,$B5,$FF
	.dc.b	$01,$35,$2F,$AD,$FD,$DF,$E5,$BE,$7B,$FF,$4F,$4E,$5F,$F5,$4C,$66
	.dc.b	$BF,$C0,$5F,$0B,$FE,$67,$9F,$E6,$D1,$FE,$A8,$E1,$FF,$74,$CE,$63
	.dc.b	$F9,$D3,$5F,$F4,$2B,$FE,$47,$9E,$B9,$7F,$0E,$8C,$4F,$FA,$0A,$D3
	.dc.b	$FC,$4B,$EC,$BF,$3F,$3F,$DA,$9A,$FF,$87,$DA,$9A,$3A,$3F,$DF,$AD
	.dc.b	$C3,$D2,$D9,$FE,$A0,$D7,$FB,$BE,$AF,$95,$FF,$16,$57,$FB,$46,$31
	.dc.b	$3F,$C9,$BE,$E5,$F2,$DF,$C4,$6F,$73,$FE,$29,$FB,$7F,$E4,$EF,$F5
	.dc.b	$D6,$BF,$EA,$86,$BF,$E0,$66,$BF,$E3,$B7,$FC,$17,$87,$FB,$4E,$DF
	.dc.b	$33,$FC,$DA,$73,$5F,$E3,$79,$CD,$7F,$B8,$BE,$E7,$FA,$33,$7C,$EF
	.dc.b	$E7,$7F,$73,$FA,$EB,$FF,$58,$D6,$9B,$3F,$86,$F8,$FF,$E8,$33,$9E
	.dc.b	$7F,$F4,$BF,$86,$E3,$FF,$79,$65,$FE,$E7,$D1,$F3,$9F,$E4,$6C,$FD
	.dc.b	$8D,$7D,$8D,$51,$C2,$7F,$1A,$FA,$A3,$E4,$6D,$B8,$C4,$7C,$98,$E5
	.dc.b	$E5,$BE,$37,$FA,$CB,$6F,$FD,$F5,$F7,$3F,$95,$D3,$FC,$B3,$39,$C3
	.dc.b	$FD,$A3,$CD,$FD,$06,$7D,$AD,$BF,$F3,$66,$AF,$AF,$FD,$5C,$7F,$9B
	.dc.b	$EB,$78,$7F,$24,$F3,$3F,$B3,$F8,$9F,$E6,$EA,$D1,$CB,$1F,$E8,$6E
	.dc.b	$32,$FF,$13,$CE,$DF,$F3,$87,$DC,$FF,$60,$7F,$DD,$AD,$7F,$C9,$6D
	.dc.b	$7F,$D0,$0D,$4F,$BE,$7D,$3E,$07,$F8,$2F,$2F,$F3,$CC,$E6,$BF,$C6
	.dc.b	$F3,$9A,$EB,$F1,$5F,$E0,$2A,$FE,$F9,$5E,$48,$C7,$A9,$2B,$F8,$1D
	.dc.b	$6A,$BC,$F3,$F7,$63,$11,$FE,$62,$D3,$C0,$9B,$1D,$B1,$FA,$BE,$07
	.dc.b	$F4,$EF,$46,$63,$F1,$F8,$3E,$97,$B5,$8C,$BE,$CD,$7D,$0F,$AF,$C6
	.dc.b	$3F,$96,$DD,$1F,$E1,$AE,$2F,$E8,$53,$F7,$BE,$3B,$1E,$6F,$F5,$C6
	.dc.b	$69,$DF,$07,$CB,$E3,$36,$1D,$63,$E3,$7C,$37,$BC,$97,$F1,$1F,$85
	.dc.b	$9C,$8F,$8E,$DB,$FF,$42,$9F,$82,$FB,$2F,$C4,$9F,$9A,$E4,$B5,$FC
	.dc.b	$47,$CC,$FD,$26,$5E,$DF,$8E,$FE,$ED,$D9,$F9,$07,$F9,$EF,$D0,$FF
	.dc.b	$91,$FC,$03,$49,$87,$F9,$BF,$95,$F4,$5F,$C6,$31,$F9,$1F,$49,$FE
	.dc.b	$40,$8C,$57,$FC,$00,$8F,$F1,$77,$D9,$BE,$5F,$F3,$FF,$80,$E1,$B5
	.dc.b	$F2,$7E,$D3,$FB,$E1,$1F,$07,$32,$3F,$CB,$98,$FE,$47,$8F,$F4,$7F
	.dc.b	$67,$E1,$F8,$7F,$D5,$BE,$3F,$53,$F5,$BF,$63,$72,$FA,$CF,$F9,$6F
	.dc.b	$DF,$FE,$BC,$FF,$0F,$6C,$7F,$A3,$5B,$87,$D0,$6B,$FE,$2B,$C5,$F0
	.dc.b	$BF,$E6,$0F,$DC,$7F,$C5,$5C,$5F,$C8,$27,$38,$7E,$4E,$72,$FF,$9B
	.dc.b	$FE,$2C,$FF,$93,$9A,$FF,$94,$9B,$87,$F9,$39,$AF,$FB,$1B,$17,$E5
	.dc.b	$CF,$98,$F5,$36,$FD,$4D,$AF,$53,$5F,$EB,$3F,$77,$E6,$7F,$CB,$6D
	.dc.b	$73,$A7,$F9,$8A,$31,$01,$84,$F9,$5F,$85,$97,$CD,$84,$63,$1F,$E3
	.dc.b	$50,$F9,$0F,$F2,$65,$F2,$6C,$67,$A7,$EA,$46,$96,$AF,$7B,$E8,$1F
	.dc.b	$E7,$BF,$03,$FC,$C3,$C3,$FD,$42,$75,$97,$F0,$3E,$1F,$EC,$33,$9E
	.dc.b	$1F,$CC,$79,$3F,$65,$FC,$26,$7F,$17,$5F,$76,$BE,$79,$F1,$93,$1F
	.dc.b	$D2,$65,$EE,$7E,$F6,$DC,$BE,$57,$F0,$8F,$E3,$F3,$7D,$87,$FA,$A3
	.dc.b	$EE,$9C,$D9,$8C,$3F,$95,$FB,$AF,$F0,$EE,$3F,$93,$6E,$BF,$93,$6B
	.dc.b	$F8,$36,$BF,$D4,$B3,$67,$F9,$53,$F6,$DF,$07,$6F,$FA,$65,$3E,$EB
	.dc.b	$E9,$B7,$97,$FB,$84,$60,$FA,$B8,$FF,$7A,$9F,$17,$B9,$FA,$98,$C4
	.dc.b	$78,$E3,$FD,$53,$E1,$7F,$2C,$C3,$E9,$79,$DF,$C1,$3C,$BF,$E1,$0C
	.dc.b	$8C,$9F,$E2,$99,$E1,$7F,$CA,$A3,$FB,$6F,$3F,$F3,$47,$FB,$7F,$DC
	.dc.b	$7F,$05,$7E,$F7,$C3,$7F,$B4,$A3,$15,$C4,$6B,$FC,$6A,$7F,$8C,$12
	.dc.b	$BF,$A4,$57,$0E,$4F,$DD,$4F,$FB,$7B,$E4,$DF,$67,$7E,$17,$0D,$C3
	.dc.b	$E5,$CF,$F8,$F9,$FF,$30,$B7,$2F,$E7,$27,$FB,$7E,$BF,$C3,$91,$98
	.dc.b	$FE,$7A,$EF,$E5,$FF,$79,$1F,$AF,$87,$D8,$7F,$08,$84,$E9,$E9,$7F
	.dc.b	$B5,$8C,$57,$08,$C7,$5F,$EE,$6B,$FD,$E1,$E6,$F3,$BD,$B6,$75,$7F
	.dc.b	$E2,$D9,$CD,$F2,$BF,$25,$F8,$E3,$D2,$D8,$E2,$D7,$71,$FE,$60,$7F
	.dc.b	$85,$7D,$D7,$F8,$CF,$C9,$FF,$43,$B6,$3E,$3B,$5F,$EC,$2D,$7D,$5F
	.dc.b	$34,$FB,$1F,$E2,$71,$F3,$A3,$1E,$0F,$FA,$49,$B1,$99,$DB,$D3,$E6
	.dc.b	$3F,$0A,$F8,$1A,$E7,$93,$CB,$18,$8F,$D7,$7D,$9F,$CB,$D7,$EB,$D7
	.dc.b	$F1,$6F,$9D,$C7,$07,$EC,$DB,$7F,$92,$F0,$18,$62,$FF,$8E,$B6,$15
	.dc.b	$4C,$46,$FA,$9F,$03,$F6,$F8,$7F,$DF,$D1,$8C,$7E,$76,$3F,$EE,$BA
	.dc.b	$D4,$7F,$95,$A7,$4F,$E2,$CF,$D2,$B8,$CF,$D6,$DC,$F8,$57,$87,$E2
	.dc.b	$A5,$FE,$0F,$0E,$CD,$BE,$0F,$49,$F6,$78,$FF,$91,$EA,$FE,$6D,$C1
	.dc.b	$74,$FE,$30,$DB,$F8,$D5,$71,$FE,$83,$F4,$77,$83,$FA,$55,$7F,$6D
	.dc.b	$AC,$4F,$1F,$8F,$7C,$AB,$E0,$B8,$7D,$74,$C3,$FD,$6D,$39,$7F,$BB
	.dc.b	$7C,$D7,$18,$C4,$7F,$09,$BE,$77,$E1,$AF,$17,$C4,$FF,$B7,$B4,$FE
	.dc.b	$AA,$DD,$3F,$D9,$98,$7F,$46,$8C,$7C,$BF,$08,$FF,$02,$47,$AE,$3D
	.dc.b	$9F,$33,$27,$FC,$10,$FC,$1C,$DF,$3B,$F0,$F9,$1E,$F7,$85,$B7,$FE
	.dc.b	$37,$6B,$FE,$48,$CD,$18,$8F,$E0,$51,$FD,$B3,$E6,$3A,$7B,$B2,$7F
	.dc.b	$C3,$11,$F3,$FE,$32,$47,$F8,$6D,$FA,$5F,$1D,$F4,$F8,$7E,$01,$27
	.dc.b	$FB,$2B,$F0,$FE,$2B,$ED,$7C,$57,$F9,$6B,$AB,$FD,$8A,$DE,$AF,$EB
	.dc.b	$13,$AF,$17,$ED,$7D,$67,$F2,$0C,$7F,$DD,$8D,$7F,$9E,$6D,$FE,$70
	.dc.b	$9F,$C3,$7C,$87,$F6,$09,$FF,$2C,$35,$F8,$4D,$7E,$0B,$5F,$F5,$CE
	.dc.b	$2F,$82,$F3,$92,$3F,$D1,$90,$F2,$BD,$7F,$03,$27,$2F,$BB,$97,$DF
	.dc.b	$EE,$FE,$16,$CD,$25,$7F,$9B,$61,$F3,$7F,$18,$91,$E0,$F6,$7E,$4F
	.dc.b	$3F,$CA,$F4,$A4,$F9,$23,$E0,$F6,$7E,$4F,$63,$49,$3D,$FE,$E7,$E9
	.dc.b	$91,$F6,$FC,$DF,$D6,$FB,$3F,$AF,$EF,$C9,$F8,$31,$F6,$BD,$BF,$61
	.dc.b	$23,$D4,$FB,$FD,$7F,$D9,$7E,$6F,$FC,$37,$AF,$FA,$C7,$8E,$8C,$76
	.dc.b	$7D,$0A,$FC,$2F,$B1,$C3,$E2,$A6,$5F,$42,$3F,$81,$7E,$25,$F4,$63
	.dc.b	$FD,$28,$D7,$F3,$CD,$7F,$3A,$D7,$EE,$FB,$5F,$92,$23,$F7,$3E,$67
	.dc.b	$F2,$BF,$21,$23,$EF,$BF,$C4,$D7,$DB,$37,$C3,$FE,$04,$8F,$EF,$A6
	.dc.b	$B3,$AD,$33,$1B,$7F,$69,$D5,$FE,$B0,$7F,$C6,$71,$9E,$0F,$E6,$8D
	.dc.b	$DB,$F2,$BE,$EF,$F2,$3F,$DC,$FE,$4D,$FF,$87,$17,$D4,$E4,$D5,$6F
	.dc.b	$86,$D7,$FB,$96,$B8,$F8,$87,$FD,$6D,$1E,$2E,$0F,$4E,$5F,$AF,$8F
	.dc.b	$F6,$8B,$5F,$F0,$53,$5F,$F4,$22,$E0,$F8,$B3,$27,$DB,$1E,$9F,$9B
	.dc.b	$F5,$7B,$B2,$7C,$91,$96,$90,$FD,$FE,$0F,$99,$F8,$44,$8F,$63,$F1
	.dc.b	$7E,$33,$F2,$3F,$37,$27,$FD,$31,$C3,$F9,$2E,$84,$FC,$BF,$72,$57
	.dc.b	$C5,$6B,$E0,$F7,$3E,$87,$A9,$23,$E5,$3F,$4E,$A4,$7C,$D6,$BE,$8F
	.dc.b	$99,$FC,$57,$D9,$7D,$F8,$FF,$2A,$07,$19,$7F,$73,$6F,$2F,$F8,$DF
	.dc.b	$ED,$7E,$05,$7F,$9E,$16,$BF,$8E,$6C,$7E,$49,$AF,$F9,$41,$AF,$F9
	.dc.b	$04,$7D,$13,$C9,$FC,$F4,$F0,$C7,$FC,$08,$DC,$3F,$C9,$15,$BB,$F1
	.dc.b	$67,$E1,$23,$E1,$7C,$99,$F4,$64,$7A,$63,$BF,$B1,$23,$D8,$F8,$BD
	.dc.b	$89,$1E,$F6,$BF,$D0,$3C,$1F,$D2,$3F,$ED,$3F,$B5,$FE,$3E,$FF,$A7
	.dc.b	$1A,$FC,$E6,$E9,$FE,$1F,$E6,$FF,$16,$CF,$FA,$89,$FF,$05,$BF,$E7
	.dc.b	$66,$BF,$1D,$AF,$F9,$A5,$AF,$87,$DA,$FE,$0D,$1C,$A7,$B7,$C0,$7F
	.dc.b	$09,$B5,$7F,$0B,$FC,$69,$E7,$D2,$63,$FE,$B4,$87,$99,$B2,$46,$5F
	.dc.b	$E2,$9A,$7B,$38,$7F,$CF,$9E,$9F,$C4,$A3,$49,$E4,$FF,$1B,$4D,$24
	.dc.b	$7E,$9A,$7F,$DF,$DE,$13,$81,$E4,$FF,$11,$F8,$DF,$F1,$07,$80,$69
	.dc.b	$30,$FF,$4F,$DF,$F0,$DF,$23,$84,$7F,$11,$9F,$E3,$51,$FC,$6F,$E7
	.dc.b	$A4,$FF,$15,$7F,$8F,$7C,$EC,$8F,$E3,$0D,$7F,$81,$FC,$E7,$F2,$3E
	.dc.b	$8F,$19,$23,$C3,$97,$F8,$7D,$F5,$AE,$0E,$11,$FC,$8E,$7F,$96,$47
	.dc.b	$F2,$DF,$9B,$F6,$92,$79,$BF,$03,$E5,$75,$1F,$CA,$E3,$F9,$A4,$7F
	.dc.b	$95,$FF,$65,$FF,$44,$F5,$7F,$AD,$F7,$43,$E6,$B4,$6E,$26,$1F,$44
	.dc.b	$FF,$6B,$CF,$2E,$3F,$A6,$D8,$CF,$CD,$77,$F8,$CF,$F2,$94,$FF,$B0
	.dc.b	$E7,$FC,$7A,$D7,$8C,$6B,$E6,$B8,$47,$7C,$3C,$B3,$E9,$9F,$57,$CC
	.dc.b	$CA,$D5,$FF,$8A,$F3,$F6,$3F,$E9,$A8,$CF,$C8,$F8,$88,$FE,$0A,$3F
	.dc.b	$5E,$FF,$94,$DF,$D3,$67,$EC,$9F,$90,$7C,$9F,$03,$FD,$01,$F7,$4E
	.dc.b	$6A,$4C,$3D,$10,$E6,$DF,$37,$F9,$CE,$0D,$24,$F8,$FE,$43,$AF,$34
	.dc.b	$8F,$6C,$7D,$9F,$90,$91,$F0,$1F,$B9,$F2,$1F,$E9,$3F,$CF,$3C,$93
	.dc.b	$F1,$3E,$47,$F0,$06,$74,$F4,$F4,$78,$C9,$3F,$A4,$D1,$1F,$E8,$9C
	.dc.b	$9C,$FF,$E9,$F6,$D7,$D5,$6B,$FE,$60,$7F,$E0,$07,$1F,$FA,$E7,$E5
	.dc.b	$1C,$28,$57,$E3,$FE,$1B,$FD,$CF,$C7,$FF,$00,$7A,$BF,$CA,$73,$AC
	.dc.b	$3E,$AC,$62,$11,$D2,$F9,$CE,$63,$13,$D4,$E5,$5D,$6F,$BC,$70,$7F
	.dc.b	$DE,$F8,$79,$63,$81,$D9,$C7,$57,$E9,$63,$C9,$18,$C3,$8E,$3F,$C5
	.dc.b	$1B,$2A,$E0,$FF,$BC,$CC,$47,$F5,$F5,$F2,$7C,$5F,$EC,$4D,$78,$4F
	.dc.b	$27,$E4,$C7,$A5,$F9,$63,$E2,$DD,$19,$8F,$05,$FC,$97,$51,$97,$99
	.dc.b	$F0,$0C,$94,$F2,$99,$9E,$AF,$CD,$B6,$AE,$BC,$E5,$FC,$36,$31,$87
	.dc.b	$E1,$FA,$F2,$FF,$3A,$8C,$E1,$F8,$9D,$3F,$E0,$03,$96,$BF,$2E,$3F
	.dc.b	$17,$3F,$CE,$9B,$1F,$EC,$BE,$5F,$D8,$4F,$F0,$4F,$92,$EF,$3F,$4E
	.dc.b	$72,$FF,$AD,$BB,$13,$AD,$3F,$91,$FC,$35,$9F,$F4,$BE,$1F,$C3,$BC
	.dc.b	$BC,$4E,$7E,$4B,$86,$6E,$BC,$5F,$E9,$5C,$F5,$FA,$B2,$F8,$ED,$A3
	.dc.b	$4C,$67,$D0,$EB,$8D,$C3,$F8,$FF,$C9,$79,$7E,$4B,$CC,$D8,$FE,$47
	.dc.b	$1D,$F5,$34,$91,$FC,$87,$E4,$BD,$C7,$BF,$E4,$AC,$7F,$31,$FD,$AF
	.dc.b	$F6,$ED,$E7,$F8,$D1,$8E,$2F,$F6,$FB,$F8,$8E,$6F,$E9,$8F,$E2,$A3
	.dc.b	$FD,$F6,$D7,$FB,$CB,$5F,$D3,$37,$80,$FF,$B7,$EF,$C5,$F2,$5E,$0F
	.dc.b	$92,$F0,$B2,$3C,$DE,$85,$C3,$C6,$F9,$BE,$4B,$CF,$DD,$6B,$D8,$E1
	.dc.b	$FF,$01,$E3,$FE,$E2,$6C,$7D,$78,$C5,$FF,$95,$72,$F2,$FC,$97,$C2
	.dc.b	$F7,$3E,$1B,$70,$F9,$11,$C4,$8F,$67,$A7,$C9,$7C,$A3,$E5,$FA,$F6
	.dc.b	$7E,$87,$AD,$F3,$8C,$63,$F3,$9F,$A3,$F2,$3F,$4B,$E4,$2D,$7D,$77
	.dc.b	$0F,$B1,$F5,$11,$FF,$61,$FE,$79,$FE,$25,$7F,$BE,$5A,$30,$FF,$37
	.dc.b	$8F,$F1,$7F,$DD,$B2,$FD,$5C,$EF,$D7,$B6,$F5,$E9,$88,$B8,$76,$F3
	.dc.b	$3F,$C3,$CD,$F5,$B9,$FE,$1F,$FE,$B5,$8C,$1B,$F5,$FB,$D1,$B5,$87
	.dc.b	$C7,$F8,$7B,$E2,$FC,$BF,$53,$F9,$5E,$DF,$E2,$FB,$2B,$59,$7C,$0A
	.dc.b	$D3,$FC,$A9,$FC,$E4,$FF,$A4,$DA,$E4,$FE,$26,$1C,$BF,$D7,$3E,$B3
	.dc.b	$D4,$FF,$AA,$9B,$7E,$29,$FD,$35,$7F,$38,$9C,$BE,$1E,$4F,$49,$FC
	.dc.b	$4F,$89,$C7,$89,$F4,$72,$7B,$8C,$65,$FE,$B8,$C5,$1F,$0B,$E3,$FE
	.dc.b	$1F,$C7,$DB,$F8,$0E,$1F,$EA,$57,$EB,$8F,$6D,$FA,$A3,$E1,$47,$C0
	.dc.b	$FC,$67,$FC,$0B,$5F,$88,$FA,$BF,$F7,$EB,$F8,$A8,$CD,$7F,$B2,$A7
	.dc.b	$2D,$F9,$4F,$93,$18,$8F,$F6,$03,$5F,$EB,$6D,$7E,$67,$7F,$E4,$37
	.dc.b	$FD,$9F,$07,$DC,$A7,$CA,$66,$BF,$8A,$FC,$CC,$CF,$F8,$D7,$CC,$5C
	.dc.b	$BF,$8B,$FC,$C7,$F1,$A3,$19,$7F,$18,$7F,$8E,$FC,$44,$BF,$E4,$5C
	.dc.b	$36,$FF,$48,$3E,$65,$F9,$DF,$FA,$AA,$BF,$CC,$D1,$8F,$5F,$86,$5F
	.dc.b	$DA,$23,$13,$FC,$B7,$8F,$F9,$59,$FC,$B3,$E2,$6C,$7F,$31,$F9,$AF
	.dc.b	$E5,$E6,$32,$FE,$5E,$E2,$3F,$CC,$DB,$FF,$80,$1E,$06,$0F,$DB,$F1
	.dc.b	$3E,$42,$FC,$86,$1B,$3F,$6E,$3E,$A9,$1F,$C2,$CA,$3F,$AD,$57,$DC
	.dc.b	$D3,$FD,$14,$FD,$2A,$F1,$F3,$7F,$6B,$8C,$65,$F2,$3C,$A4,$62,$F8
	.dc.b	$FC,$86,$F9,$F8,$3C,$7C,$0F,$4F,$8D,$73,$FC,$83,$87,$CA,$FE,$30
	.dc.b	$8F,$FA,$FB,$07,$CC,$F8,$CF,$87,$F1,$16,$7E,$4F,$CA,$79,$CF,$47
	.dc.b	$CC,$58,$FF,$49,$7D,$5F,$F0,$8D,$FF,$A1,$35,$FF,$18,$EE,$8C,$47
	.dc.b	$FB,$59,$AF,$F9,$0F,$27,$CC,$FF,$85,$3A,$3F,$DD,$FE,$A7,$C7,$8C
	.dc.b	$7A,$1F,$EE,$6C,$BE,$EF,$CC,$E3,$C2,$F9,$8D,$8F,$C0,$F4,$FD,$42
	.dc.b	$3F,$05,$FA,$FD,$1F,$4F,$CF,$FA,$8D,$9F,$C4,$F2,$7D,$73,$EC,$71
	.dc.b	$D8,$F8,$2F,$F8,$67,$83,$FD,$41,$1F,$EE,$A6,$F0,$7C,$49,$FC,$2F
	.dc.b	$C8,$FE,$07,$E4,$7F,$04,$DB,$FC,$BF,$C8,$FE,$14,$FC,$37,$A1,$63
	.dc.b	$F3,$7E,$87,$E3,$88,$FD,$B7,$F2,$3F,$1F,$F8,$D3,$FC,$71,$FC,$03
	.dc.b	$FE,$6F,$9F,$D1,$C6,$3E,$36,$47,$FB,$2E,$7F,$D3,$6D,$8F,$F2,$07
	.dc.b	$D2,$E3,$5F,$EF,$CD,$7E,$EB,$5F,$F4,$CF,$52,$31,$1F,$E7,$FB,$7C
	.dc.b	$A7,$DD,$EC,$7C,$A7,$9A,$30,$FA,$3E,$43,$D3,$F2,$17,$4F,$33,$EC
	.dc.b	$FA,$1F,$F2,$46,$26,$92,$7F,$D1,$5A,$25,$F2,$7F,$10,$F8,$BF,$25
	.dc.b	$6B,$E1,$7C,$7F,$C0,$CF,$E4,$B6,$3F,$DE,$1B,$FF,$96,$DE,$AD,$DF
	.dc.b	$E2,$FC,$87,$F9,$6E,$31,$1F,$A2,$CB,$DF,$5F,$3F,$0F,$F3,$67,$A3
	.dc.b	$F9,$BA,$FE,$05,$5A,$6E,$1F,$6B,$E4,$3E,$DB,$70,$FB,$11,$F0,$88
	.dc.b	$FC,$0A,$FF,$34,$B7,$E4,$3F,$59,$F8,$AB,$93,$F0,$B1,$86,$4F,$E0
	.dc.b	$FE,$46,$E5,$E4,$7F,$17,$EA,$FE,$33,$E4,$AD,$7E,$25,$C3,$F8,$BF
	.dc.b	$92,$B7,$EC,$7F,$17,$F1,$FF,$EA,$6F,$94,$FC,$1C,$FF,$B0,$FC,$8F
	.dc.b	$F3,$44,$63,$9F,$F1,$55,$BE,$BF,$C6,$D6,$9B,$79,$39,$16,$B8,$E3
	.dc.b	$FE,$C4,$9D,$B6,$76,$FF,$87,$25,$5F,$E7,$98,$FE,$32,$D9,$E2,$7F
	.dc.b	$8A,$27,$F1,$9F,$25,$DF,$85,$F2,$56,$30,$DC,$3C,$15,$F9,$38,$C5
	.dc.b	$FF,$B7,$B1,$FD,$4E,$1C,$AF,$CD,$2F,$A4,$7A,$BF,$ED,$B8,$FF,$73
	.dc.b	$46,$3E,$4F,$FC,$2E,$D8,$CC,$ED,$FF,$71,$8F,$96,$FF,$6A,$FD,$27
	.dc.b	$F8,$79,$B1,$FE,$F1,$6B,$C1,$AF,$FB,$B2,$BF,$C9,$9A,$78,$3D,$8F
	.dc.b	$F7,$E4,$F9,$1A,$F8,$27,$6F,$FA,$1F,$E9,$3F,$66,$FC,$46,$BF,$0C
	.dc.b	$52,$33,$C5,$F8,$57,$F8,$04,$F1,$F0,$BF,$DF,$F3,$9F,$0A,$F0,$7F
	.dc.b	$14,$BE,$2D,$AC,$DF,$E4,$76,$B8,$FE,$09,$BC,$7F,$27,$2F,$CA,$7C
	.dc.b	$87,$EA,$6F,$ED,$B5,$F4,$5F,$17,$EF,$D7,$1C,$7F,$92,$D6,$EF,$9F
	.dc.b	$57,$82,$31,$3A,$8C,$7A,$BE,$4D,$1E,$1B,$2F,$E7,$1D,$15,$FE,$12
	.dc.b	$DF,$71,$CD,$F0,$6B,$0F,$C3,$F1,$6D,$FC,$EC,$3A,$BD,$6F,$8B,$9B
	.dc.b	$DE,$FF,$4E,$ED,$FF,$BF,$6B,$86,$1F,$D0,$BF,$61,$FC,$7E,$3C,$B0
	.dc.b	$46,$27,$EA,$78,$7E,$2F,$67,$F6,$3D,$9D,$22,$63,$FE,$48,$E3,$FD
	.dc.b	$6B,$D8,$FC,$2E,$1F,$9D,$7F,$C1,$ED,$7C,$5C,$9F,$E7,$AC,$BC,$2D
	.dc.b	$D1,$E1,$7B,$2F,$F4,$DA,$F9,$DB,$AF,$E2,$67,$5F,$2D,$32,$F1,$65
	.dc.b	$EC,$8F,$87,$E1,$7C,$6D,$1F,$2D,$E1,$6D,$FF,$54,$6B,$FE,$E9,$6B
	.dc.b	$FE,$28,$E7,$79,$FF,$41,$7A,$14,$C4,$7A,$7E,$5E,$57,$93,$64,$7D
	.dc.b	$6F,$17,$DD,$EE,$7C,$39,$34,$95,$FD,$A3,$D7,$FA,$CF,$CE,$D3,$E3
	.dc.b	$1F,$82,$8C,$F3,$BE,$95,$F0,$FF,$2F,$DD,$98,$7F,$A0,$27,$2F,$C8
	.dc.b	$E0,$7C,$B7,$C5,$24,$7C,$57,$CD,$F2,$FF,$54,$91,$F4,$32,$72,$7E
	.dc.b	$27,$E6,$B9,$CF,$FA,$39,$AF,$FA,$53,$44,$63,$A7,$F2,$97,$F0,$AB
	.dc.b	$E4,$63,$F1,$48,$FF,$82,$B8,$39,$BF,$8E,$E5,$FC,$41,$1F,$26,$7F
	.dc.b	$17,$E0,$7F,$18,$48,$F4,$3F,$85,$F9,$2F,$C2,$12,$3F,$08,$FD,$5F
	.dc.b	$94,$FC,$A1,$23,$E0,$3F,$47,$E1,$3F,$31,$F2,$72,$7E,$24,$7E,$6E
	.dc.b	$BE,$F1,$23,$D4,$FF,$B7,$37,$7D,$1F,$EF,$A7,$FD,$25,$A7,$FA,$49
	.dc.b	$FC,$55,$FE,$D6,$75,$BF,$FA,$DD,$FF,$66,$CF,$E3,$FE,$47,$F5,$46
	.dc.b	$FE,$2B,$F4,$5F,$1F,$FA,$12,$4F,$E8,$5F,$C9,$7C,$8F,$EC,$49,$1F
	.dc.b	$77,$A7,$F6,$9F,$23,$23,$F3,$8F,$E4,$7E,$47,$F8,$CC,$3E,$91,$97
	.dc.b	$FB,$E2,$30,$D8,$CC,$7F,$BF,$1A,$FF,$64,$39,$61,$D2,$33,$A7,$C5
	.dc.b	$6F,$8D,$FC,$BD,$4F,$F2,$4F,$9E,$7C,$8F,$F0,$BD,$26,$3F,$C1,$A7
	.dc.b	$2D,$C3,$85,$48,$C1,$FC,$74,$F6,$69,$EE,$FA,$2F,$E0,$F9,$FD,$76
	.dc.b	$DF,$8E,$7D,$18,$FB,$C9,$7E,$F7,$F9,$16,$2F,$63,$F9,$06,$F2,$BC
	.dc.b	$D1,$EB,$F6,$FD,$8C,$8F,$54,$65,$A4,$3F,$DB,$55,$FE,$CC,$7F,$89
	.dc.b	$70,$7F,$2D,$23,$FC,$E7,$97,$F2,$3F,$21,$F8,$D2,$46,$79,$3F,$DE
	.dc.b	$8F,$C9,$F5,$7F,$9B,$7C,$A4,$8F,$E3,$CF,$D1,$F0,$FE,$95,$D9,$FA
	.dc.b	$11,$87,$E6,$7E,$09,$FD,$DF,$8F,$F4,$0D,$E2,$FE,$D4,$D9,$FE,$DD
	.dc.b	$F4,$AC,$7F,$1F,$F4,$BE,$D9,$FC,$43,$5F,$F1,$6B,$5F,$EE,$8D,$7F
	.dc.b	$BC,$2F,$95,$E7,$F2,$D2,$7F,$01,$39,$6E,$7F,$9E,$F9,$B7,$9D,$4F
	.dc.b	$1C,$79,$36,$FD,$05,$FC,$8B,$F2,$C6,$0C,$FC,$9F,$A8,$91,$E0,$F4
	.dc.b	$BD,$A4,$70,$79,$3D,$2F,$F0,$0F,$99,$F0,$78,$BF,$BE,$37,$8B,$FD
	.dc.b	$82,$DA,$FF,$0B,$7D,$2B,$E2,$7F,$6F,$D5,$FE,$3E,$9F,$F2,$9B,$5F
	.dc.b	$F0,$D3,$5F,$F0,$F3,$5F,$F3,$0F,$B9,$CB,$E0,$E4,$7E,$9E,$53,$FC
	.dc.b	$13,$E9,$5E,$7D,$25,$7E,$B6,$1E,$76,$C9,$19,$79,$79,$DF,$C3,$BD
	.dc.b	$6F,$E0,$33,$FE,$31,$6B,$FD,$AB,$CA,$D7,$93,$23,$F7,$53,$FC,$19
	.dc.b	$B1,$CF,$C7,$FF,$15,$FE,$5B,$FC,$C6,$7F,$C7,$5E,$D2,$31,$1F,$EE
	.dc.b	$C6,$BF,$E6,$46,$C3,$C6,$7C,$D7,$4E,$59,$7F,$C4,$63,$BF,$99,$23
	.dc.b	$FC,$33,$87,$F1,$47,$F7,$3F,$31,$FC,$6C,$8F,$F1,$97,$F8,$F7,$CC
	.dc.b	$7A,$3D,$0F,$F1,$65,$7F,$42,$AF,$F4,$6E,$7E,$99,$C9,$FC,$43,$B3
	.dc.b	$84,$F8,$23,$F9,$44,$7C,$2E,$6E,$64,$9F,$F7,$CD,$FF,$A8,$1F,$F7
	.dc.b	$DE,$AF,$98,$E1,$1E,$AA,$FE,$65,$1F,$E5,$6F,$A4,$7C,$C7,$CE,$F3
	.dc.b	$E5,$FF,$8E,$27,$2D,$9E,$04,$75,$A7,$D6,$F1,$FD,$E4,$8F,$2B,$FE
	.dc.b	$4D,$AF,$F6,$DC,$3C,$0F,$F3,$06,$9F,$A3,$BF,$B9,$7E,$5F,$89,$FC
	.dc.b	$0F,$93,$30,$FA,$6F,$A7,$E5,$FD,$5F,$19,$D4,$93,$FD,$5F,$06,$7F
	.dc.b	$EA,$28,$FA,$2F,$FA,$AB,$3F,$F3,$83,$96,$D7,$C5,$F0,$7F,$77,$EE
	.dc.b	$CB,$FF,$2D,$F8,$A8,$FF,$3F,$FD,$2B,$A2,$F4,$7F,$9A,$E0,$D2,$5F
	.dc.b	$96,$3F,$CA,$EF,$F9,$99,$79,$1D,$0E,$BD,$1F,$5F,$E2,$A5,$7D,$57
	.dc.b	$ED,$7C,$BF,$DB,$F9,$6F,$D3,$FA,$92,$BE,$D4,$79,$BA,$BC,$84,$8F
	.dc.b	$13,$FE,$81,$BF,$F6,$9F,$0F,$FA,$4D,$BE,$07,$FA,$AF,$93,$FD,$5A
	.dc.b	$7F,$A7,$A1,$F4,$5F,$E0,$EF,$92,$FE,$1B,$7F,$EB,$1E,$C7,$C3,$7F
	.dc.b	$AC,$67,$FD,$B3,$18,$CF,$FD,$6D,$18,$9D,$79,$9F,$85,$9F,$CA,$C6
	.dc.b	$2F,$A3,$CF,$9F,$FD,$79,$87,$3A,$FF,$7C,$46,$27,$84,$72,$3B,$DF
	.dc.b	$5C,$FF,$3B,$1F,$EF,$F8,$C5,$F0,$9C,$E3,$EC,$AF,$F7,$7C,$FF,$88
	.dc.b	$FE,$47,$E2,$36,$32,$7F,$B3,$C4,$69,$C3,$5F,$F7,$99,$E5,$3F,$DB
	.dc.b	$1F,$1B,$FE,$CC,$9F,$64,$7F,$B6,$E7,$E0,$BF,$ED,$B9,$D6,$3E,$1A
	.dc.b	$DC,$62,$7B,$9D,$43,$9C,$3F,$DC,$FD,$5F,$ED,$AD,$3F,$00,$DD,$FF
	.dc.b	$01,$BB,$E6,$B1,$8B,$0E,$35,$FE,$BE,$A7,$27,$F3,$49,$FF,$79,$B6
	.dc.b	$33,$3F,$B5,$6D,$7E,$D3,$E3,$7A,$DF,$F3,$DF,$92,$BD,$DF,$ED,$2C
	.dc.b	$3F,$99,$C6,$3E,$5E,$47,$F1,$7F,$4E,$E3,$FC,$36,$3D,$3F,$23,$D3
	.dc.b	$37,$4D,$8C,$39,$BF,$31,$2B,$CC,$FA,$FC,$1F,$DA,$C7,$F0,$2E,$DE
	.dc.b	$9B,$7A,$27,$39,$78,$4F,$8D,$AB,$E5,$FD,$7F,$29,$EC,$6E,$3F,$CB
	.dc.b	$E3,$E4,$49,$A4,$8F,$E7,$FE,$DF,$C6,$F2,$4F,$C6,$F9,$4C,$3F,$2B
	.dc.b	$E9,$7E,$FD,$7F,$5A,$F7,$31,$D5,$C5,$91,$FD,$29,$FB,$18,$1A,$7D
	.dc.b	$BA,$FF,$71,$CF,$FB,$F1,$BA,$7B,$F9,$91,$FE,$D0,$9D,$7A,$BE,$33
	.dc.b	$D1,$F3,$D6,$3F,$B6,$BE,$7E,$EF,$67,$A9,$67,$FB,$8B,$87,$FA,$FB
	.dc.b	$E6,$23,$FC,$FE,$FF,$DC,$EF,$F7,$8F,$2F,$E1,$7C,$87,$C7,$64,$7F
	.dc.b	$68,$7D,$E4,$79,$7B,$3F,$24,$F9,$5E,$E5,$8F,$F0,$1F,$9D,$F4,$4C
	.dc.b	$6D,$F0,$78,$3F,$A1,$4F,$C5,$F8,$EF,$AD,$E7,$FD,$36,$C7,$F8,$32
	.dc.b	$3E,$79,$1F,$EF,$6F,$F8,$6B,$E9,$7F,$A7,$47,$FA,$7D,$B7,$F8,$D0
	.dc.b	$FF,$87,$23,$59,$FE,$DA,$73,$8D,$AF,$2C,$71,$79,$F8,$DF,$ED,$D7
	.dc.b	$2F,$FB,$7F,$CC,$FF,$05,$61,$FE,$AF,$6D,$78,$57,$C0,$77,$73,$5C
	.dc.b	$7E,$7B,$06,$7C,$FB,$87,$FA,$E7,$E8,$D8,$F9,$DF,$CB,$BE,$D7,$31
	.dc.b	$BE,$AD,$1F,$E3,$E8,$DC,$FF,$2F,$6B,$D0,$38,$7F,$8C,$7C,$4B,$D7
	.dc.b	$FE,$1E,$DB,$FF,$66,$7D,$13,$E3,$73,$5E,$5F,$93,$95,$E9,$E7,$B8
	.dc.b	$F9,$F9,$7D,$46,$37,$57,$8F,$44,$FB,$F2,$F5,$12,$7F,$DD,$D7,$E8
	.dc.b	$8F,$77,$C7,$7D,$BF,$A7,$7B,$BF,$1E,$30,$FD,$FF,$A3,$FF,$76,$E1
	.dc.b	$FE,$E5,$5F,$05,$9F,$83,$D3,$27,$BC,$E9,$B1,$FE,$6A,$8C,$65,$FF
	.dc.b	$00,$7E,$2F,$FD,$C9,$5F,$EF,$16,$C7,$E4,$F8,$BF,$BF,$CE,$7E,$0B
	.dc.b	$F8,$97,$CA,$7F,$13,$F8,$7B,$3F,$C2,$5F,$E3,$5E,$8C,$8C,$1F,$E7
	.dc.b	$07,$FA,$7F,$C8,$BE,$85,$F3,$98,$6D,$7F,$13,$F8,$AC,$4F,$F2,$5F
	.dc.b	$85,$93,$FE,$EB,$AD,$EB,$FB,$79,$FE,$4D,$F2,$32,$BF,$8E,$73,$5A
	.dc.b	$FE,$39,$E5,$7F,$B7,$67,$F2,$FD,$5F,$72,$7F,$B1,$6F,$F7,$A7,$F9
	.dc.b	$26,$1F,$D6,$A3,$13,$FC,$95,$B5,$FE,$60,$F0,$F5,$7F,$CE,$DF,$B5
	.dc.b	$E6,$FD,$12,$47,$D6,$7E,$A7,$99,$FB,$DE,$5F,$F7,$34,$FD,$F8,$C7
	.dc.b	$C4,$F8,$7A,$3C,$5F,$15,$6B,$D2,$7F,$8E,$F3,$46,$A3,$FD,$3F,$E2
	.dc.b	$DD,$7F,$8E,$3F,$8D,$F2,$BF,$1D,$E5,$5A,$FC,$73,$87,$F1,$6F,$E2
	.dc.b	$48,$FF,$A5,$BC,$8F,$EB,$07,$C5,$EF,$B1,$E4,$C3,$FC,$19,$ED,$7F
	.dc.b	$83,$63,$1E,$97,$FB,$94,$F9,$9B,$7F,$81,$A7,$55,$FD,$47,$C9,$7B
	.dc.b	$9F,$E1,$16,$E1,$FE,$83,$6C,$FF,$AC,$9A,$FF,$07,$ED,$FF,$6F,$FA
	.dc.b	$7D,$7E,$0F,$F2,$3F,$1B,$FD,$7B,$A7,$F0,$1A,$FF,$BD,$E7,$3E,$AF
	.dc.b	$8F,$D0,$BA,$7A,$36,$FF,$32,$B7,$CF,$FF,$12,$47,$F0,$FD,$7F,$D8
	.dc.b	$6D,$7F,$DB,$AD,$7E,$B7,$4F,$BF,$E0,$FF,$18,$DB,$ED,$F1,$7F,$0C
	.dc.b	$FC,$87,$3D,$8F,$93,$C4,$EC,$F8,$84,$8F,$88,$FF,$8D,$FA,$FF,$C3
	.dc.b	$3C,$3E,$ED,$1F,$0B,$F9,$BF,$1F,$E8,$3F,$47,$E2,$66,$3F,$5B,$07
	.dc.b	$8B,$8E,$0F,$F0,$26,$2F,$89,$F8,$24,$BF,$F5,$B4,$65,$FF,$34,$4B
	.dc.b	$A5,$D0,$F3,$FE,$99,$FA,$DE,$EB,$7F,$3F,$E1,$3F,$08,$7F,$92,$BE
	.dc.b	$83,$FD,$BB,$19,$3F,$17,$E5,$7F,$43,$ED,$FE,$19,$EB,$FF,$95,$E3
	.dc.b	$07,$F4,$7C,$DF,$17,$FE,$56,$E5,$F8,$9F,$13,$FC,$52,$31,$8F,$FA
	.dc.b	$97,$82,$F0,$7F,$99,$74,$75,$7E,$5F,$C2,$F2,$36,$FF,$0D,$F0,$FF
	.dc.b	$84,$23,$F1,$1F,$C3,$FC,$3F,$E7,$CF,$67,$45,$9F,$D2,$FD,$17,$83
	.dc.b	$88,$FE,$7B,$7F,$CB,$B9,$31,$3F,$EB,$BE,$1F,$F1,$CE,$BF,$D9,$EF
	.dc.b	$F9,$77,$8B,$F0,$C9,$5F,$5B,$7F,$D8,$1F,$37,$86,$C7,$ED,$B8,$3F
	.dc.b	$CD,$AF,$E8,$7D,$1F,$A5,$BF,$ED,$D9,$1F,$80,$7F,$46,$47,$F0,$3C
	.dc.b	$6F,$A2,$FA,$31,$D7,$1F,$E8,$71,$8D,$3F,$CA,$18,$A7,$F0,$07,$F0
	.dc.b	$5F,$8A,$B7,$FB,$37,$F0,$87,$E1,$7C,$0B,$1E,$97,$0F,$EE,$5B,$7F
	.dc.b	$9A,$7F,$D2,$BF,$0B,$FE,$14,$8F,$F5,$EF,$8D,$FE,$E5,$8C,$7C,$67
	.dc.b	$F8,$36,$7F,$D4,$DC,$7F,$E7,$BC,$3F,$D7,$57,$B9,$FE,$95,$5B,$E2
	.dc.b	$FF,$3D,$C6,$27,$EB,$7C,$2D,$8F,$F5,$86,$3D,$A7,$38,$74,$9F,$F5
	.dc.b	$A7,$27,$F9,$89,$F2,$98,$BE,$58,$76,$AE,$0F,$F6,$D8,$CE,$DF,$E9
	.dc.b	$79,$CA,$F8,$7C,$FC,$56,$99,$3A,$72,$D3,$C9,$83,$E8,$7F,$A7,$F7
	.dc.b	$7F,$87,$9B,$1F,$C0,$1A,$FF,$BA,$5A,$FF,$8E,$FE,$39,$E0,$3E,$8B
	.dc.b	$FC,$4F,$18,$6C,$66,$7F,$85,$CE,$5E,$DE,$F7,$FB,$65,$BB,$78,$E3
	.dc.b	$11,$E1,$6B,$FB,$DD,$3F,$C9,$9C,$5F,$5F,$9B,$E2,$4F,$CA,$6B,$F4
	.dc.b	$2F,$F2,$95,$C3,$8B,$94,$EE,$30,$FF,$60,$6C,$FE,$6F,$E3,$2F,$C3
	.dc.b	$3D,$AB,$E8,$4C,$3E,$CC,$E5,$B3,$97,$C5,$09,$E7,$32,$76,$43,$7E
	.dc.b	$87,$8B,$F8,$07,$FC,$D5,$1F,$EB,$D5,$C7,$FE,$6D,$F0,$FF,$B1,$C6
	.dc.b	$27,$FB,$BB,$C2,$33,$1B,$E5,$FF,$3F,$E8,$AD,$46,$1B,$29,$D7,$CF
	.dc.b	$F1,$8F,$E8,$9F,$9D,$BD,$7C,$57,$3F,$A3,$B8,$1E,$1E,$1A,$F8,$61
	.dc.b	$FE,$0F,$8C,$61,$A9,$CC,$FF,$BE,$67,$2F,$B3,$C7,$EB,$F4,$7F,$E7
	.dc.b	$6C,$BD,$4D,$C8,$D3,$1E,$CF,$0B,$F8,$B1,$E6,$8F,$4F,$E5,$3F,$DD
	.dc.b	$67,$E4,$FD,$1F,$A8,$CF,$8D,$EE,$FC,$D7,$CB,$3F,$CD,$DA,$FC,$C6
	.dc.b	$BF,$25,$B9,$17,$B9,$CC,$93,$F6,$3C,$5F,$93,$F2,$7F,$2B,$D6,$93
	.dc.b	$F7,$23,$E6,$F5,$FB,$C9,$1F,$DE,$31,$FB,$91,$FE,$9A,$E1,$FA,$1E
	.dc.b	$BF,$CD,$25,$7E,$1B,$E4,$3E,$AF,$D7,$7F,$66,$8F,$A8,$D7,$E9,$37
	.dc.b	$AB,$F1,$FF,$21,$F0,$FD,$3F,$A5,$C0,$69,$2F,$C9,$CC,$F7,$3F,$0B
	.dc.b	$F3,$72,$7F,$2F,$1F,$88,$F8,$DF,$C1,$5C,$9F,$CD,$C7,$E3,$3E,$37
	.dc.b	$EF,$12,$3A,$BF,$81,$FA,$2F,$C8,$92,$3F,$47,$F1,$DF,$94,$F8,$09
	.dc.b	$1F,$A6,$7F,$2F,$F1,$BF,$77,$AB,$D6,$49,$F9,$9B,$7E,$15,$FD,$9F
	.dc.b	$33,$F8,$07,$FD,$25,$0E,$2F,$F2,$AC,$61,$BE,$2F,$E9,$A1,$F2,$57
	.dc.b	$E3,$B1,$F1,$D3,$0F,$F3,$7D,$7F,$B2,$1B,$E9,$7E,$7E,$0D,$24,$7F
	.dc.b	$78,$FC,$EF,$F6,$0F,$EB,$5A,$FE,$AD,$B9,$79,$BE,$37,$F6,$B1,$FA
	.dc.b	$89,$E1,$E3,$60,$F4,$7C,$6F,$EF,$BF,$35,$79,$1F,$B1,$6D,$7E,$45
	.dc.b	$AF,$C6,$6B,$FE,$1D,$CF,$FC,$76,$FB,$BE,$23,$E2,$1F,$17,$E7,$7F
	.dc.b	$87,$43,$E9,$FF,$DE,$EF,$FA,$54,$7D,$13,$E2,$76,$D8,$D4,$F0,$7E
	.dc.b	$A7,$21,$91,$F9,$37,$FD,$FE,$D7,$FC,$C2,$D7,$F9,$8B,$5E,$43,$F7
	.dc.b	$6F,$C2,$70,$FE,$0A,$79,$37,$C0,$F5,$CE,$AF,$F9,$8E,$9F,$E0,$EC
	.dc.b	$3C,$A7,$F2,$F5,$F3,$DF,$87,$3F,$11,$F0,$BF,$89,$F6,$2D,$7E,$1A
	.dc.b	$30,$FE,$23,$D9,$FC,$77,$B3,$23,$07,$E3,$3E,$53,$F2,$53,$23,$07
	.dc.b	$E4,$3E,$77,$F2,$C4,$7F,$27,$F2,$7F,$99,$F9,$2F,$CD,$32,$3F,$2F
	.dc.b	$1F,$9D,$F9,$7D,$46,$0F,$CD,$BF,$73,$89,$E3,$FF,$A3,$A7,$E9,$B5
	.dc.b	$FC,$03,$63,$71,$F5,$BC,$7E,$53,$23,$EF,$C7,$EB,$7E,$5F,$F5,$DF
	.dc.b	$29,$27,$F5,$71,$FD,$5B,$F0,$2F,$F0,$6F,$3F,$ED,$A3,$84,$FF,$B7
	.dc.b	$A7,$3C,$BF,$7B,$E9,$7F,$74,$FE,$03,$EA,$3F,$63,$5F,$93,$6B,$F6
	.dc.b	$84,$F0,$7F,$20,$FF,$00,$7E,$03,$5F,$F5,$7B,$5F,$EF,$ED,$7F,$1F
	.dc.b	$F5,$0F,$3B,$F4,$B3,$C7,$2F,$E2,$2D,$C3,$73,$FB,$7F,$47,$F7,$1E
	.dc.b	$84,$C7,$FD,$F7,$E1,$27,$F7,$95,$B9,$CC,$61,$FD,$86,$1F,$94,$7D
	.dc.b	$3D,$7F,$5E,$31,$1E,$A8,$C7,$5F,$99,$91,$CA,$3D,$3B,$FE,$FB,$67
	.dc.b	$83,$4C,$D7,$83,$87,$B1,$1F,$92,$FE,$1E,$7F,$55,$BF,$C2,$8F,$8A
	.dc.b	$75,$EC,$FE,$17,$E4,$FF,$39,$BF,$F1,$5B,$5F,$DD,$B5,$FE,$0C,$D4
	.dc.b	$BE,$1B,$9F,$9F,$27,$F8,$A4,$E5,$FE,$2D,$F9,$B7,$F9,$99,$FC,$F3
	.dc.b	$5F,$F7,$6B,$5F,$F1,$EB,$5F,$E1,$AF,$F6,$B7,$9F,$C6,$7F,$24,$8E
	.dc.b	$59,$7F,$50,$75,$3F,$C2,$BD,$07,$47,$F8,$40,$9F,$09,$FE,$27,$71
	.dc.b	$F0,$1F,$A3,$89,$5F,$A1,$8F,$D3,$7C,$4C,$8F,$D2,$3F,$A9,$F9,$C9
	.dc.b	$1F,$A7,$6B,$E1,$BF,$F0,$37,$5F,$F9,$39,$7C,$DF,$84,$93,$FE,$17
	.dc.b	$8C,$B7,$E6,$7F,$6D,$32,$3F,$69,$1F,$B9,$F7,$FD,$44,$8E,$4F,$B7
	.dc.b	$7D,$47,$EE,$E3,$F8,$04,$7F,$01,$F6,$E4,$FE,$FD,$FE,$09,$F3,$72
	.dc.b	$30,$43,$D8,$FF,$09,$F9,$AF,$94,$48,$F5,$3F,$37,$D1,$D4,$7F,$0B
	.dc.b	$8F,$E2,$11,$FC,$47,$E3,$FF,$89,$7C,$D4,$AF,$E2,$11,$FC,$57,$E6
	.dc.b	$7E,$B1,$23,$E3,$3F,$67,$A3,$51,$FC,$5E,$3F,$8E,$47,$F9,$BB,$F0
	.dc.b	$6F,$D5,$CF,$F9,$B5,$FD,$2F,$0F,$F6,$1F,$86,$46,$25,$18,$EF,$E5
	.dc.b	$E1,$7F,$92,$6B,$F7,$B5,$FD,$82,$B8,$CE,$4D,$78,$5E,$12,$4F,$F2
	.dc.b	$BE,$EF,$E6,$1F,$2F,$FC,$C5,$91,$FC,$BE,$3F,$D0,$BC,$FF,$E1,$F8
	.dc.b	$FF,$40,$37,$B9,$A7,$A6,$30,$FA,$FE,$93,$FC,$1F,$3A,$AF,$5C,$3E
	.dc.b	$33,$FC,$C4,$D8,$FF,$55,$9A,$8C,$9F,$E8,$BE,$5F,$F4,$8D,$73,$EF
	.dc.b	$47,$F4,$4A,$FF,$52,$36,$33,$DF,$FE,$18,$F9,$67,$5F,$F9,$0E,$FF
	.dc.b	$D4,$DE,$E7,$FB,$A7,$B7,$FC,$13,$7F,$E2,$6B,$46,$27,$FD,$DD,$7C
	.dc.b	$6B,$4D,$8E,$57,$FE,$1B,$9E,$55,$C6,$3F,$C3,$D7,$E2,$CB,$57,$D1
	.dc.b	$FF,$60,$DD,$AF,$F5,$F3,$71,$FF,$58,$DF,$19,$C9,$FE,$D2,$F9,$6E
	.dc.b	$2D,$32,$7F,$BD,$0A,$7F,$8B,$DD,$B9,$AF,$F2,$1F,$CA,$FD,$77,$FC
	.dc.b	$8B,$09,$F3,$BF,$E2,$E9,$D6,$3F,$62,$73,$3F,$E3,$B9,$CB,$E3,$AE
	.dc.b	$BB,$F2,$C7,$EC,$F8,$7F,$85,$CF,$EE,$6C,$F0,$F1,$7F,$EE,$48,$FF
	.dc.b	$5A,$BF,$E4,$79,$DC,$EB,$1F,$D3,$F6,$7F,$9E,$AB,$F1,$7D,$08,$C4
	.dc.b	$FF,$BE,$23,$1F,$0B,$FE,$86,$8F,$D9,$4E,$9F,$D6,$72,$BE,$0F,$B5
	.dc.b	$FF,$3A,$6D,$7F,$52,$C3,$FD,$D1,$5B,$7F,$C3,$1F,$2D,$E5,$D9,$ED
	.dc.b	$FF,$97,$E7,$27,$AC,$FF,$80,$2E,$72,$FF,$77,$7E,$07,$F9,$29,$FF
	.dc.b	$7A,$36,$3F,$A6,$4E,$F6,$FD,$76,$8A,$D4,$61,$FD,$0D,$FC,$46,$DF
	.dc.b	$C3,$6B,$FC,$2D,$AF,$F1,$55,$E3,$7C,$27,$FD,$99,$3E,$48,$C7,$8D
	.dc.b	$EF,$C5,$C9,$7E,$5F,$0D,$BF,$81,$8B,$4F,$8B,$5F,$4E,$31,$68,$C4
	.dc.b	$EA,$BF,$DD,$93,$96,$FC,$BF,$87,$E5,$BC,$4D,$8E,$F5,$F4,$BE,$1F
	.dc.b	$CD,$D5,$E7,$6C,$EB,$0F,$F5,$5B,$DA,$7C,$31,$8C,$BF,$D9,$ED,$7F
	.dc.b	$D2,$37,$F7,$BE,$57,$E1,$77,$FC,$36,$E1,$D6,$2D,$7D,$A9,$BC,$3F
	.dc.b	$C7,$70,$7C,$A7,$CB,$F9,$5F,$98,$DC,$7C,$D1,$E4,$23,$F1,$1F,$3F
	.dc.b	$CA,$FD,$3F,$95,$FA,$8D,$9F,$74,$7B,$08,$FC,$C7,$DF,$F2,$9F,$6C
	.dc.b	$FB,$9F,$15,$67,$E3,$7B,$7F,$7C,$C6,$BF,$94,$7F,$05,$EF,$7E,$0F
	.dc.b	$DD,$B5,$F3,$5C,$3F,$69,$F9,$A4,$7F,$D2,$5F,$9A,$BC,$BF,$DF,$60
	.dc.b	$87,$76,$1D,$57,$FB,$4D,$CB,$63,$F1,$93,$C5,$B3,$A8,$FE,$A2,$D7
	.dc.b	$FD,$96,$3F,$01,$FE,$9F,$5F,$ED,$F5,$EB,$79,$1C,$CF,$FA,$3D,$CC
	.dc.b	$EE,$7C,$31,$89,$FE,$BE,$D7,$BB,$5F,$F6,$DF,$DA,$7F,$9C,$AB,$FD
	.dc.b	$DD,$F4,$FF,$84,$7F,$DE,$11,$99,$FF,$4F,$4E,$5B,$C7,$F0,$23,$11
	.dc.b	$EC,$6B,$FE,$E8,$6B,$FE,$F3,$5C,$7F,$43,$E9,$5F,$2F,$F4,$A7,$38
	.dc.b	$FF,$C0,$19,$CB,$E6,$F8,$7D,$7C,$3E,$DB,$1E,$A8,$F2,$11,$FF,$68
	.dc.b	$52,$3F,$7F,$F9,$A7,$F8,$1E,$7F,$72,$FB,$7A,$BF,$C0,$4E,$73,$FF
	.dc.b	$37,$39,$94,$7F,$0D,$8C,$37,$3F,$E6,$91,$8E,$D3,$EC,$9F,$F6,$46
	.dc.b	$6E,$CF,$C8,$FD,$3B,$E5,$7E,$24,$61,$99,$7F,$BA,$67,$4D,$8F,$C8
	.dc.b	$C6,$2F,$FC,$3E,$BE,$57,$F3,$BE,$EF,$F4,$99,$FF,$6C,$4E,$6F,$F9
	.dc.b	$3C,$27,$5A,$FF,$07,$E0,$FE,$61,$BF,$F3,$16,$DF,$F1,$B8,$FE,$69
	.dc.b	$F3,$58,$C3,$FA,$4B,$EC,$F7,$3D,$BE,$E5,$BE,$F9,$A3,$31,$ED,$7D
	.dc.b	$64,$7F,$99,$F2,$7F,$5F,$FC,$27,$FC,$8B,$1F,$CA,$E3,$0C,$3F,$99
	.dc.b	$C7,$F7,$4D,$3F,$C7,$8F,$0C,$BF,$D5,$0E,$67,$ED,$F1,$A1,$F3,$5D
	.dc.b	$CF,$07,$A9,$6B,$AF,$C3,$F8,$FE,$6B,$C8,$D8,$F0,$C7,$9F,$D6,$C6
	.dc.b	$5E,$07,$D5,$CB,$EB,$E5,$B7,$E8,$70,$FA,$BE,$0A,$DF,$A1,$FD,$47
	.dc.b	$67,$E2,$3E,$52,$D7,$CC,$70,$FE,$1D,$FC,$29,$1F,$C5,$7C,$A7,$F2
	.dc.b	$0F,$99,$FF,$10,$E1,$F8,$48,$C7,$BB,$F3,$FE,$32,$5F,$CC,$E8,$B7
	.dc.b	$FA,$4E,$AF,$D5,$92,$7F,$CD,$FC,$BF,$A0,$8F,$EE,$1E,$D3,$F4,$1E
	.dc.b	$FD,$1F,$E0,$C6,$DF,$D2,$E8,$FC,$23,$FD,$EA,$B8,$69,$FE,$26,$A7
	.dc.b	$C6,$53,$F0,$9F,$2F,$6F,$F0,$3F,$31,$FD,$F3,$13,$49,$1E,$68,$FF
	.dc.b	$14,$BF,$88,$E2,$FC,$77,$A9,$F8,$F7,$F2,$1C,$52,$7F,$1B,$F3,$17
	.dc.b	$87,$FD,$E5,$F9,$A7,$EA,$DF,$E7,$BC,$48,$C4,$7F,$7A,$6B,$F9,$06
	.dc.b	$BF,$E1,$7C,$9F,$33,$F9,$F3,$F4,$1F,$2B,$71,$FC,$EF,$CA,$7E,$8C
	.dc.b	$FF,$21,$7C,$7F,$F8,$7E,$32,$7F,$85,$17,$C4,$FF,$87,$9B,$1F,$59
	.dc.b	$B8,$FA,$1A,$FF,$84,$1F,$AA,$FF,$B2,$38,$DF,$1F,$E0,$E5,$FA,$8F
	.dc.b	$8B,$E2,$7C,$7C,$D7,$1F,$13,$FE,$57,$DF,$E7,$E3,$F1,$9A,$FC,$5E
	.dc.b	$4F,$F1,$9C,$62,$3F,$DB,$5F,$9A,$FF,$42,$5D,$3F,$E5,$48,$C3,$FE
	.dc.b	$B9,$D5,$F5,$3F,$E6,$C9,$CE,$5F,$EE,$3D,$BF,$BA,$5F,$E4,$7D,$7F
	.dc.b	$93,$F2,$BF,$8D,$7E,$5F,$C8,$7C,$72,$5F,$C7,$7E,$57,$C8,$FD,$0F
	.dc.b	$8B,$FA,$2D,$9F,$F2,$1F,$D1,$7C,$CC,$7F,$D9,$8D,$8F,$E1,$FA,$FF
	.dc.b	$97,$B6,$F5,$C7,$F1,$29,$D4,$7F,$96,$5F,$DA,$BF,$E4,$26,$BB,$E4
	.dc.b	$FF,$2C,$C6,$23,$FC,$C1,$F9,$AF,$D6,$AF,$C1,$7C,$8F,$FA,$1E,$31
	.dc.b	$3E,$03,$FC,$CD,$87,$F8,$3B,$1F,$95,$EE,$FE,$33,$E5,$FF,$1A,$DC
	.dc.b	$7F,$19,$1F,$8B,$23,$F8,$B7,$F7,$FF,$9D,$BE,$2F,$FD,$3A,$D9,$FC
	.dc.b	$A7,$C9,$E2,$64,$7E,$5F,$E6,$2E,$7E,$5F,$09,$A2,$7F,$3F,$F3,$12
	.dc.b	$FF,$3D,$EE,$5C,$38,$5A,$BF,$4B,$DF,$9D,$FF,$0C,$FF,$AA,$78,$BF
	.dc.b	$C5,$BF,$2D,$FE,$C5,$C3,$E8,$71,$B3,$FF,$6E,$78,$1F,$57,$8F,$15
	.dc.b	$FE,$D9,$8F,$B1,$58,$BE,$5B,$B1,$EB,$5F,$77,$8F,$4F,$0C,$BA,$E3
	.dc.b	$F7,$E7,$2F,$8A,$33,$7C,$E7,$2F,$5C,$6E,$09,$CE,$DF,$E5,$8C,$3C
	.dc.b	$C7,$DE,$F8,$6C,$F5,$69,$99,$76,$F9,$5F,$F6,$36,$BE,$AC,$39,$3E
	.dc.b	$C8,$CE,$BF,$E2,$2C,$5F,$4E,$F6,$7F,$D6,$B1,$FE,$5E,$32,$E1,$FF
	.dc.b	$30,$C6,$FE,$57,$51,$A6,$FC,$C7,$FB,$13,$E9,$EF,$79,$A4,$AE,$D6
	.dc.b	$F7,$3F,$D8,$11,$86,$BE,$06,$BE,$98,$E1,$C9,$FD,$63,$B3,$F8,$EC
	.dc.b	$FD,$07,$FD,$0C,$DE,$1F,$81,$BF,$DE,$9D,$CF,$CE,$A4,$62,$75,$5C
	.dc.b	$36,$FF,$07,$CF,$4B,$E0,$EE,$3F,$C2,$7F,$24,$F4,$FE,$77,$1F,$FB
	.dc.b	$0A,$31,$7F,$EF,$B3,$FD,$BC,$E1,$FC,$A9,$89,$E2,$F4,$7C,$4E,$4F
	.dc.b	$C6,$C7,$1C,$7F,$32,$FF,$8D,$63,$3F,$03,$C3,$34,$EF,$1E,$95,$A3
	.dc.b	$F2,$FF,$4B,$F7,$FB,$3F,$4F,$6E,$5C,$69,$FE,$30,$6D,$FF,$BD,$EE
	.dc.b	$D1,$A4,$E4,$FC,$86,$1D,$23,$FD,$56,$DF,$96,$CE,$3D,$7E,$62,$E2
	.dc.b	$C7,$C5,$18,$7B,$55,$C0,$D2,$57,$F8,$4F,$E9,$FF,$B5,$BF,$F0,$03
	.dc.b	$3D,$BF,$E2,$06,$BF,$E0,$26,$BF,$EF,$A6,$BF,$E5,$0A,$FE,$4C,$F3
	.dc.b	$AF,$24,$F2,$CB,$DD,$E1,$F9,$7E,$27,$3F,$33,$30,$FF,$49,$D7,$C3
	.dc.b	$9F,$67,$7F,$F9,$EE,$7F,$C6,$6D,$7F,$C2,$AD,$7F,$A6,$78,$9F,$B2
	.dc.b	$D5,$F4,$DE,$E9,$FE,$64,$DC,$3E,$DB,$78,$3F,$CD,$2B,$C9,$F9,$5F
	.dc.b	$1D,$EA,$D6,$61,$E6,$BF,$81,$3E,$EF,$9A,$F0,$6E,$69,$2B,$E6,$6D
	.dc.b	$F1,$08,$FD,$3F,$33,$E3,$FC,$D4,$8F,$A8,$FE,$03,$DF,$7D,$3F,$F8
	.dc.b	$12,$7F,$BA,$F8,$FE,$FE,$3F,$F7,$FC,$7C,$68,$C1,$F0,$F9,$9F,$31
	.dc.b	$F5,$3E,$1A,$57,$AC,$FC,$C4,$77,$E9,$7C,$E7,$F1,$CC,$D1,$8A,$FE
	.dc.b	$49,$18,$9D,$4E,$67,$F1,$1C,$3F,$10,$8F,$E3,$5F,$8F,$DD,$F8,$62
	.dc.b	$47,$8F,$A3,$F5,$1F,$23,$F8,$AF,$9C,$93,$F9,$77,$F3,$DF,$25,$FA
	.dc.b	$1F,$09,$DD,$F8,$16,$D7,$F9,$E9,$AF,$F9,$71,$AF,$E6,$FE,$73,$F2
	.dc.b	$E4,$7C,$5F,$4B,$FC,$4B,$BF,$F8,$0B,$FD,$16,$3F,$51,$F3,$BF,$9E
	.dc.b	$3F,$65,$19,$8F,$F0,$07,$67,$D0,$7F,$59,$C7,$F8,$62,$3F,$AF,$F9
	.dc.b	$DF,$D3,$91,$FF,$44,$7D,$3F,$EA,$7B,$FF,$87,$7C,$F7,$EB,$09,$3A
	.dc.b	$F0,$7E,$17,$2F,$C1,$24,$7C,$17,$F2,$BC,$7F,$B4,$24,$7B,$BE,$4B
	.dc.b	$F8,$6A,$F9,$1F,$EE,$4F,$FB,$B3,$F7,$7F,$85,$97,$F6,$76,$C0,$E1
	.dc.b	$EB,$F2,$FA,$77,$27,$F6,$49,$46,$1F,$EC,$ED,$7F,$B7,$35,$FE,$E8
	.dc.b	$70,$E3,$E9,$32,$3F,$DF,$F2,$9F,$A3,$F2,$5F,$C4,$3E,$4A,$57,$DA
	.dc.b	$8C,$B4,$87,$F8,$D7,$95,$F2,$7F,$C6,$3A,$A4,$FF,$25,$7E,$EF,$C9
	.dc.b	$C8,$FC,$2B,$5E,$5E,$9F,$F2,$0F,$99,$91,$FC,$B5,$FC,$2F,$CA,$48
	.dc.b	$FC,$6B,$5F,$17,$CA,$7E,$2B,$E5,$3F,$16,$C8,$FC,$9C,$7C,$DE,$9F
	.dc.b	$AF,$F6,$FF,$CC,$27,$FC,$C4,$1C,$77,$FF,$6A,$E2,$57,$D5,$E9,$FA
	.dc.b	$90,$69,$2F,$FD,$C3,$C0,$F9,$2F,$E9,$DD,$FF,$80,$F1,$3F,$85,$F5
	.dc.b	$7E,$82,$7F,$43,$F2,$77,$0F,$D4,$C6,$1F,$E1,$C3,$E9,$7F,$C3,$CE
	.dc.b	$3F,$3A,$FF,$0E,$ED,$FE,$00,$6E,$1F,$E3,$5A,$DD,$7B,$BE,$9F,$FC
	.dc.b	$0A,$FE,$19,$AF,$F8,$ED,$AF,$E5,$5A,$97,$CE,$7E,$81,$FC,$DC,$66
	.dc.b	$B1,$3F,$C3,$7C,$59,$3F,$C5,$7C,$74,$7F,$71,$FA,$8B,$3F,$C6,$5A
	.dc.b	$FF,$A0,$5B,$9F,$F8,$29,$B8,$7E,$A7,$64,$7F,$07,$AF,$77,$67,$ED
	.dc.b	$3E,$E3,$FB,$06,$1F,$E2,$46,$E1,$EF,$9F,$57,$77,$EB,$F5,$34,$98
	.dc.b	$7F,$95,$B6,$7C,$E7,$F2,$9F,$02,$4F,$D6,$9C,$BF,$62,$DC,$7D,$49
	.dc.b	$1F,$E6,$08,$CB,$78,$FE,$64,$C8,$E5,$1F,$C3,$BE,$27,$FD,$39,$1F
	.dc.b	$84,$FC,$5B,$CE,$DC,$F5,$F1,$FB,$F7,$FF,$85,$5F,$F7,$FA,$FF,$3C
	.dc.b	$35,$FE,$24,$D5,$F5,$4F,$73,$FC,$4D,$E6,$37,$A7,$F2,$09,$D5,$7E
	.dc.b	$59,$7C,$5F,$F1,$96,$C7,$F9,$F9,$AF,$FA,$15,$AF,$FA,$2F,$9F,$FB
	.dc.b	$7F,$57,$F8,$4F,$B3,$F1,$F3,$FE,$8B,$9C,$B6,$32,$FF,$20,$87,$D4
	.dc.b	$FF,$85,$63,$FC,$81,$3A,$E9,$FF,$29,$B7,$A3,$F1,$55,$BB,$FE,$FF
	.dc.b	$F4,$5F,$B8,$26,$1F,$DB,$67,$FD,$8C,$FF,$B5,$5B,$87,$F9,$79,$B7
	.dc.b	$D3,$E1,$FF,$C9,$D0,$F9,$DF,$E0,$7C,$3E,$D2,$61,$EB,$7F,$C6,$52
	.dc.b	$F9,$8D,$CF,$F7,$DD,$3F,$B3,$65,$DE,$7F,$86,$FC,$77,$C7,$25,$7C
	.dc.b	$57,$9F,$1E,$A3,$94,$79,$23,$F8,$AF,$D1,$49,$EC,$FA,$3D,$D9,$18
	.dc.b	$21,$F0,$5F,$E3,$BF,$45,$F5,$C9,$1F,$55,$F5,$FC,$AE,$A3,$D5,$1F
	.dc.b	$12,$3E,$2F,$CB,$FF,$26,$E8,$95,$F3,$E3,$E4,$FC,$87,$C4,$24,$7E
	.dc.b	$0F,$7F,$F5,$7B,$7A,$F6,$FD,$79,$FA,$31,$FE,$43,$F9,$5F,$EA,$F3
	.dc.b	$FE,$39,$F8,$36,$1F,$B8,$87,$0F,$FB,$FA,$73,$E2,$71,$F1,$79,$3F
	.dc.b	$E3,$DC,$38,$5F,$FB,$8E,$91,$88,$FE,$FD,$18,$9E,$1E,$07,$8C,$93
	.dc.b	$E3,$AF,$F5,$83,$F8,$DF,$97,$F4,$79,$52,$3F,$25,$1F,$D3,$2E,$47
	.dc.b	$E1,$5F,$F3,$06,$CE,$CF,$F3,$17,$D5,$3E,$33,$FC,$A8,$DB,$FF,$29
	.dc.b	$B5,$FF,$04,$E8,$8C,$47,$ED,$7E,$63,$FA,$6B,$FE,$AB,$FC,$EB,$FA
	.dc.b	$F3,$FE,$51,$6B,$F2,$BC,$0F,$E6,$D3,$C2,$1F,$33,$FC,$37,$8B,$FC
	.dc.b	$45,$F4,$5F,$E8,$19,$FF,$12,$4B,$E7,$7F,$C4,$D0,$9D,$3C,$A3,$31
	.dc.b	$89,$E9,$3B,$F9,$2E,$31,$89,$D5,$70,$9C,$B6,$BC,$3D,$DF,$82,$79
	.dc.b	$18,$C3,$FA,$85,$F4,$9D,$78,$BF,$EA,$7C,$7C,$2D,$8C,$CF,$F8,$D2
	.dc.b	$DF,$31,$C9,$FF,$1B,$B5,$F7,$B5,$FF,$23,$35,$E2,$79,$0F,$F2,$57
	.dc.b	$CC,$35,$F9,$7A,$77,$AB,$5F,$2A,$B4,$F1,$3A,$95,$CD,$FC,$E8,$CB
	.dc.b	$63,$35,$F7,$3C,$7E,$7A,$7F,$BF,$B6,$FE,$3D,$97,$FA,$4D,$4E,$A3
	.dc.b	$F0,$7E,$3D,$C9,$C4,$EE,$FC,$C5,$7F,$A9,$9B,$3F,$94,$6B,$FA,$F6
	.dc.b	$BF,$E0,$3F,$1B,$FC,$A7,$3F,$C9,$FB,$7B,$5F,$F4,$4F,$13,$C1,$C5
	.dc.b	$FE,$BC,$9B,$F4,$5F,$E7,$3B,$FE,$09,$F1,$7F,$91,$7F,$D5,$91,$9C
	.dc.b	$BF,$38,$7A,$27,$4E,$31,$47,$F6,$9A,$E7,$C9,$E2,$8D,$F0,$7F,$0D
	.dc.b	$75,$7C,$5F,$ED,$35,$FE,$C6,$6C,$7E,$D1,$AF,$F0,$F6,$BF,$C3,$9F
	.dc.b	$F6,$8C,$7F,$1D,$F0,$7E,$1C,$F5,$F1,$BE,$1C,$62,$FF,$07,$A3,$37
	.dc.b	$CC,$FC,$4F,$99,$F8,$AD,$CB,$E1,$C7,$F9,$22,$B8,$36,$8D,$31,$F1
	.dc.b	$DF,$43,$87,$FC,$5B,$7F,$32,$31,$E2,$FD,$2E,$7D,$78,$96,$BF,$B0
	.dc.b	$38,$7F,$C0,$0F,$86,$CD,$3C,$5F,$2F,$DF,$E1,$7E,$EB,$6B,$FB,$54
	.dc.b	$5E,$1F,$B7,$AF,$C1,$AC,$6F,$F1,$36,$7C,$6F,$A7,$E1,$FF,$0E,$DC
	.dc.b	$BF,$BA,$47,$8C,$8F,$F8,$55,$F8,$9F,$21,$F0,$6B,$F1,$ED,$9F,$EF
	.dc.b	$91,$EB,$23,$FE,$1D,$7E,$66,$BF,$96,$3F,$2F,$F1,$76,$7F,$C0,$9F
	.dc.b	$11,$F3,$4C,$6B,$EC,$A4,$7E,$7B,$E1,$7E,$97,$C4,$5B,$FF,$06,$38
	.dc.b	$7F,$C5,$8F,$CD,$23,$FC,$66,$DF,$56,$F9,$6F,$F4,$3B,$67,$4D,$9F
	.dc.b	$E4,$0D,$7F,$DA,$BF,$83,$7E,$AF,$83,$F2,$3F,$85,$BD,$7F,$E0,$94
	.dc.b	$AD,$61,$FE,$7F,$AD,$3F,$EA,$4F,$AD,$7A,$FF,$E9,$86,$C7,$9D,$B7
	.dc.b	$FC,$C5,$AF,$FB,$99,$7C,$8F,$83,$EA,$BF,$D7,$F5,$D6,$31,$BF,$93
	.dc.b	$EB,$1F,$35,$7A,$30,$DB,$FF,$5B,$CF,$E2,$1F,$90,$DC,$BF,$A1,$35
	.dc.b	$FF,$7B,$B5,$FE,$BF,$8B,$E4,$BF,$8B,$FE,$74,$E4,$F3,$46,$1F,$27
	.dc.b	$C2,$7F,$AE,$30,$FF,$3B,$63,$E0,$8C,$19,$F9,$4F,$9B,$F6,$EF,$37
	.dc.b	$D9,$18,$7F,$92,$FD,$77,$9F,$9F,$F8,$2B,$70,$FF,$58,$3F,$09,$6D
	.dc.b	$FA,$A8,$D2,$F0,$5B,$7F,$A8,$AB,$77,$FC,$42,$31,$5C,$78,$7D,$FF
	.dc.b	$17,$F8,$2F,$AE,$7C,$25,$F8,$4C,$37,$2F,$F7,$2C,$A7,$FD,$B2,$D8
	.dc.b	$FC,$C3,$70,$FE,$E4,$D7,$FC,$1A,$3E,$2B,$E2,$4F,$FB,$C3,$F0,$27
	.dc.b	$C4,$7E,$52,$73,$5F,$D0,$84,$EA,$BF,$A3,$F1,$FF,$4C,$E2,$FE,$9A
	.dc.b	$D8,$FE,$97,$1F,$C8,$FE,$63,$17,$FB,$77,$E0,$79,$7E,$A2,$57,$A9
	.dc.b	$F6,$F9,$7E,$3D,$BE,$33,$67,$E6,$9F,$EE,$5F,$AF,$FF,$02,$38,$C3
	.dc.b	$EA,$47,$FB,$EA,$31,$C8,$8D,$CF,$F2,$47,$4D,$F8,$4F,$ED,$31,$FD
	.dc.b	$D3,$97,$FA,$4B,$A6,$C7,$F3,$FF,$AF,$98,$76,$AE,$0F,$F7,$78,$E7
	.dc.b	$D1,$23,$5E,$2D,$C7,$F8,$D7,$37,$F4,$93,$19,$7F,$AE,$9F,$1F,$C3
	.dc.b	$CB,$ED,$CF,$6F,$D5,$3F,$D4,$7E,$43,$18,$7F,$B0,$9F,$C7,$79,$5F
	.dc.b	$8F,$EF,$B8,$FE,$31,$C3,$F8,$A7,$F1,$A4,$7F,$92,$7C,$EB,$E9,$7F
	.dc.b	$85,$76,$FF,$1A,$8C,$4F,$D1,$CD,$E1,$FE,$9B,$91,$FC,$01,$FA,$7A
	.dc.b	$FA,$6D,$AF,$A5,$EF,$7F,$86,$23,$1B,$7F,$78,$FA,$CF,$F1,$7D,$7F
	.dc.b	$B8,$5A,$FF,$BA,$DA,$FE,$DB,$04,$62,$3F,$6D,$5F,$E6,$DC,$3C,$3F
	.dc.b	$0B,$FE,$28,$FA,$DF,$4D,$7F,$66,$AD,$37,$6F,$21,$BF,$83,$B8,$7A
	.dc.b	$BE,$13,$FC,$0D,$7F,$E4,$28,$FF,$7D,$36,$FF,$DF,$0D,$7F,$CF,$F9
	.dc.b	$FE,$4F,$8B,$E0,$61,$F9,$06,$DF,$C6,$FA,$FF,$88,$7E,$53,$8E,$C7
	.dc.b	$E7,$36,$FF,$11,$E3,$F1,$7E,$07,$FC,$31,$18,$8F,$8D,$1E,$6F,$92
	.dc.b	$F9,$C7,$E7,$BB,$2D,$FD,$4F,$8B,$FA,$47,$E8,$BA,$2C,$7D,$8F,$85
	.dc.b	$F7,$91,$F7,$BF,$E7,$0C,$7F,$D8,$DC,$6F,$90,$C4,$7F,$B1,$A3,$15
	.dc.b	$FE,$87,$F8,$74,$62,$7F,$D6,$AD,$7E,$3F,$C8,$FE,$0B,$A3,$FC,$77
	.dc.b	$CB,$FE,$48,$7F,$C7,$B1,$9C,$3F,$05,$39,$7F,$09,$C1,$FE,$E2,$7F
	.dc.b	$D2,$DC,$BF,$86,$F8,$7F,$8E,$48,$FB,$3C,$5F,$93,$3F,$CD,$BF,$5F
	.dc.b	$FD,$5E,$3F,$33,$1F,$97,$F8,$B9,$1F,$EE,$FD,$7F,$C0,$9A,$79,$A7
	.dc.b	$F3,$9E,$D7,$E1,$89,$5F,$86,$7F,$C9,$DA,$F9,$F6,$FC,$BF,$B3,$FE
	.dc.b	$7A,$8C,$6B,$FA,$A8,$FF,$3E,$FD,$4F,$E6,$32,$FF,$2E,$B7,$1F,$D1
	.dc.b	$3F,$9A,$8C,$F6,$FF,$A1,$38,$3E,$66,$5F,$13,$8F,$F0,$E4,$C3,$E2
	.dc.b	$E6,$D7,$F6,$9F,$19,$E3,$6C,$FF,$2C,$9F,$F0,$6B,$FB,$CE,$2F,$46
	.dc.b	$BF,$BB,$64,$7E,$91,$FD,$39,$1E,$EF,$BC,$FF,$4D,$FD,$7B,$E9,$B1
	.dc.b	$3D,$E7,$F8,$37,$CB,$63,$4F,$DC,$3F,$1F,$49,$5E,$58,$B3,$FE,$62
	.dc.b	$CB,$F1,$DB,$FE,$2B,$D5,$CB,$6F,$FA,$3F,$0F,$C1,$7F,$DA,$BE,$17
	.dc.b	$CC,$DB,$FD,$9B,$F2,$3F,$EC,$0C,$BF,$C0,$57,$B7,$FD,$C3,$5B,$BE
	.dc.b	$31,$AB,$E3,$39,$8D,$9C,$B8,$32,$EE,$B8,$F2,$30,$F9,$F3,$FA,$F8
	.dc.b	$C6,$1F,$EE,$0A,$FE,$1B,$5A,$6C,$66,$BF,$93,$68,$FA,$DE,$F8,$7F
	.dc.b	$AF,$26,$CE,$6F,$C3,$F2,$3F,$EC,$5D,$3E,$05,$AB,$57,$EA,$8E,$57
	.dc.b	$E8,$9C,$D7,$0B,$E5,$A7,$F8,$B3,$07,$D7,$7F,$84,$BE,$4F,$FC,$E4
	.dc.b	$D8,$FF,$0C,$9A,$F6,$D3,$A6,$C6,$BE,$51,$F3,$6F,$AF,$FF,$3E,$46
	.dc.b	$1B,$19,$9E,$0D,$7F,$C4,$F8,$7F,$9D,$F6,$F1,$FA,$9F,$E3,$F9,$F1
	.dc.b	$CE,$5B,$19,$7F,$9E,$F8,$5F,$C8,$5F,$B2,$D7,$E3,$37,$6F,$96,$FD
	.dc.b	$38,$DF,$0F,$C5,$9C,$F4,$5D,$FF,$8E,$B6,$BF,$AE,$D5,$7E,$5F,$A5
	.dc.b	$72,$34,$C7,$9F,$27,$F2,$A7,$02,$CF,$06,$C6,$DF,$CB,$9F,$61,$C6
	.dc.b	$3F,$A3,$87,$E5,$8F,$2F,$CD,$B7,$F8,$0B,$E3,$3F,$01,$D7,$FE,$F4
	.dc.b	$FC,$4B,$FC,$AA,$DC,$33,$D3,$FC,$EA,$AC,$7F,$7A,$6D,$F3,$8F,$F0
	.dc.b	$5C,$62,$FF,$07,$1F,$EE,$07,$FD,$FE,$FF,$91,$7C,$97,$84,$FA,$7E
	.dc.b	$B1,$87,$97,$CC,$CA,$FF,$24,$DF,$27,$A4,$77,$F9,$9F,$07,$CC,$FF
	.dc.b	$B2,$73,$5A,$30,$DB,$FE,$A2,$D7,$FC,$EB,$3C,$3C,$6E,$71,$E0,$AF
	.dc.b	$64,$7C,$28,$7E,$3C,$F9,$8F,$F6,$C4,$61,$B1,$98,$FF,$6E,$35,$FF
	.dc.b	$41,$F6,$B9,$3F,$4F,$1F,$AE,$8C,$76,$FF,$B8,$DB,$19,$9D,$BE,$8F
	.dc.b	$C9,$BE,$27,$23,$93,$FC,$65,$18,$6F,$07,$D5,$F9,$AF,$AD,$F3,$D3
	.dc.b	$0F,$9A,$FD,$9D,$7E,$FE,$6F,$CB,$FE,$B5,$79,$5A,$FA,$9A,$FF,$0B
	.dc.b	$8E,$1C,$9F,$43,$E4,$BF,$08,$49,$FB,$7F,$01,$F8,$7F,$B0,$91,$F7
	.dc.b	$9F,$C5,$7C,$E7,$8B,$1F,$F6,$B4,$F8,$7E,$53,$E2,$46,$23,$C5,$1F
	.dc.b	$7B,$E6,$FF,$24,$48,$FC,$4F,$D7,$7E,$57,$E7,$3F,$2D,$F5,$52,$7F
	.dc.b	$1B,$1F,$03,$F2,$FF,$F1,$3C,$7F,$B4,$37,$46,$23,$F8,$E3,$63,$3C
	.dc.b	$9F,$8E,$F9,$AF,$D1,$70,$CB,$FC,$CC,$7E,$97,$E5,$3F,$4D,$F2,$92
	.dc.b	$7F,$3B,$1E,$1F,$43,$F0,$19,$7F,$B1,$27,$F0,$8D,$7F,$12,$D7,$FC
	.dc.b	$B7,$E0,$FC,$3E,$BF,$DE,$70,$FE,$CC,$FD,$BF,$F0,$0F,$F9,$16,$3F
	.dc.b	$0D,$5F,$EE,$98,$7D,$36,$BE,$4A,$4F,$FA,$E2,$72,$DC,$BF,$5D,$1E
	.dc.b	$18,$FF,$40,$35,$E9,$1A,$F4,$BC,$24,$8E,$9F,$23,$E4,$F9,$4F,$2F
	.dc.b	$C8,$C9,$F0,$47,$F0,$EF,$0B,$EF,$47,$F9,$E1,$AF,$E7,$B5,$FE,$01
	.dc.b	$18,$8F,$67,$A3,$F0,$08,$FA,$FD,$0F,$DD,$7D,$7B,$BF,$F7,$2D,$9F
	.dc.b	$EB,$E1,$C3,$FD,$89,$AF,$FB,$F1,$77,$F2,$24,$72,$8C,$BF,$E6,$0D
	.dc.b	$DF,$49,$E2,$A9,$3F,$3E,$3E,$BF,$CB,$FD,$8F,$97,$93,$F4,$E3,$2D
	.dc.b	$20,$E5,$FB,$BE,$73,$17,$D2,$7E,$0B,$E9,$25,$FD,$D7,$F0,$BF,$27
	.dc.b	$23,$F0,$0D,$7A,$78,$1F,$7B,$E4,$BE,$4F,$8B,$27,$FD,$A1,$C0,$8F
	.dc.b	$E4,$DF,$74,$F8,$6A,$7F,$26,$71,$5F,$39,$AF,$F8,$39,$AF,$F8,$54
	.dc.b	$7C,$C7,$F9,$8D,$FF,$6F,$47,$0D,$FF,$D4,$0D,$CB,$FC,$39,$5B,$AF
	.dc.b	$49,$F9,$FF,$9B,$FD,$19,$1F,$CF,$7C,$AF,$E9,$88,$E0,$FD,$17,$CD
	.dc.b	$7E,$AB,$E5,$64,$7E,$95,$FE,$C7,$F1,$BF,$E1,$1C,$FE,$8A,$F9,$9C
	.dc.b	$F2,$CB,$1E,$45,$3B,$8F,$97,$E5,$FE,$50,$8F,$E2,$DF,$CB,$F9,$32
	.dc.b	$3F,$1C,$D7,$FC,$6F,$F9,$87,$F7,$2C,$7F,$D8,$9C,$6A,$FD,$09,$FD
	.dc.b	$6E,$75,$C1,$FE,$F9,$F4,$1D,$FF,$DF,$9B,$8F,$A3,$1F,$C6,$C6,$23
	.dc.b	$D4,$FE,$71,$D5,$73,$ED,$E4,$48,$C8,$D5,$4E,$CF,$ED,$DF,$58,$F1
	.dc.b	$E9,$2F,$8D,$BE,$FD,$F3,$3F,$E1,$46,$CE,$9A,$FC,$F6,$E9,$EA,$F0
	.dc.b	$7B,$E9,$FD,$EB,$2F,$55,$FF,$98,$1A,$FE,$4D,$AF,$F9,$A5,$F4,$3F
	.dc.b	$C2,$1A,$F1,$CF,$A1,$FE,$06,$E8,$7C,$1F,$BA,$24,$FC,$41,$3A,$71
	.dc.b	$E5,$BE,$73,$FC,$24,$D8,$CC,$7F,$85,$7B,$3F,$C8,$11,$CD,$AF,$E8
	.dc.b	$8F,$D5,$FC,$2E,$CE,$1F,$2F,$FE,$1D,$FC,$83,$FA,$76,$FF,$D1,$A9
	.dc.b	$18,$E5,$FF,$44,$36,$3F,$37,$F9,$87,$C0,$AE,$0D,$AF,$F4,$73,$67
	.dc.b	$7A,$FF,$92,$BE,$77,$E9,$D7,$E9,$39,$7F,$47,$F3,$5F,$E3,$BE,$E3
	.dc.b	$49,$87,$FA,$52,$7F,$53,$F1,$FF,$B2,$99,$3F,$86,$AF,$F3,$26,$7F
	.dc.b	$B4,$E0,$FF,$27,$F2,$FE,$BE,$BF,$82,$FC,$57,$F3,$4F,$07,$F7,$75
	.dc.b	$27,$93,$5F,$DC,$60,$F9,$EF,$E0,$3C,$7E,$72,$4F,$96,$7E,$8B,$FE
	.dc.b	$AF,$D7,$F7,$FB,$F0,$8F,$35,$7F,$0A,$8F,$E1,$7D,$7F,$04,$93,$CD
	.dc.b	$F8,$9F,$0F,$A8,$F7,$47,$F1,$28,$FE,$27,$F3,$BE,$EB,$95,$F0,$E3
	.dc.b	$2D,$21,$EC,$F9,$9F,$E3,$7C,$FF,$44,$91,$EB,$7E,$A7,$CC,$6A,$3E
	.dc.b	$74,$7F,$22,$8F,$E4,$7F,$37,$F6,$49,$3F,$0D,$FB,$9A,$C8,$F8,$CD
	.dc.b	$7F,$CE,$6B,$E6,$FF,$96,$B2,$3E,$E4,$7F,$A2,$3E,$6B,$FC,$2B,$1F
	.dc.b	$D1,$F8,$FF,$DC,$9E,$F7,$F7,$A6,$C7,$FB,$69,$AE,$E5,$EF,$6D,$B1
	.dc.b	$B6,$BB,$8F,$F4,$77,$E5,$5F,$81,$8F,$E2,$6D,$7F,$9E,$37,$A3,$FA
	.dc.b	$1F,$A3,$FD,$18,$93,$F9,$57,$FA,$4F,$CB,$FF,$A3,$7E,$67,$F2,$FD
	.dc.b	$DF,$EF,$56,$D7,$F4,$86,$A7,$F8,$13,$FB,$DF,$C7,$BF,$D2,$69,$A4
	.dc.b	$7D,$3C,$BF,$EB,$AC,$BF,$82,$E5,$FE,$08,$6E,$1F,$B3,$5D,$1E,$5E
	.dc.b	$8F,$F0,$76,$1F,$D7,$7E,$3F,$CD,$3F,$D7,$23,$07,$9D,$C7,$6A,$7D
	.dc.b	$95,$CA,$31,$5E,$D8,$C4,$F4,$8F,$EC,$F3,$A7,$E0,$57,$49,$E7,$8E
	.dc.b	$9B,$3F,$D3,$6A,$F5,$FE,$FF,$5F,$E8,$F7,$FC,$3E,$72,$7F,$83,$EB
	.dc.b	$FD,$68,$71,$3F,$C2,$5F,$23,$B6,$99,$3E,$03,$96,$9F,$08,$FE,$FF
	.dc.b	$F3,$7F,$E7,$09,$F5,$47,$F8,$0A,$7D,$0F,$C9,$9D,$63,$E3,$9C,$CF
	.dc.b	$2D,$6B,$C4,$EC,$2B,$8B,$FD,$B7,$5F,$EC,$B3,$F9,$6D,$7F,$D0,$39
	.dc.b	$7D,$FE,$DB,$3F,$7E,$6D,$FF,$85,$BA,$9F,$09,$FE,$A1,$6D,$7F,$99
	.dc.b	$9B,$3F,$25,$AF,$CB,$F9,$4E,$6F,$F8,$81,$AF,$FB,$D6,$7F,$B0,$28
	.dc.b	$F0,$F0,$7F,$89,$3E,$6A,$F6,$1B,$E6,$61,$DE,$CF,$06,$99,$CB,$F6
	.dc.b	$AF,$B3,$B1,$F3,$9F,$C3,$67,$F8,$23,$74,$FE,$14,$D7,$F3,$93,$C3
	.dc.b	$BF,$FD,$8A,$F2,$EC,$B5,$FE,$5F,$BE,$71,$83,$FD,$97,$AB,$F8,$07
	.dc.b	$FC,$83,$3F,$E1,$A7,$EC,$37,$1F,$83,$18,$7F,$C8,$CD,$7F,$A5,$35
	.dc.b	$FE,$D0,$D7,$FC,$58,$D9,$F0,$7C,$B7,$8A,$0D,$25,$7F,$6B,$87,$CC
	.dc.b	$F8,$BE,$67,$C6,$D9,$F4,$47,$32,$3C,$5E,$DC,$5E,$A3,$D7,$D5,$67
	.dc.b	$E0,$47,$D3,$F2,$7F,$2B,$9F,$BA,$31,$CF,$F0,$BE,$63,$E1,$FC,$C5
	.dc.b	$C3,$E4,$38,$7D,$0F,$B4,$8F,$AF,$E6,$3E,$5F,$CC,$7C,$C6,$C7,$D1
	.dc.b	$8F,$88,$47,$E1,$3F,$1F,$E2,$FE,$99,$F5,$3E,$2E,$CF,$D9,$8F,$57
	.dc.b	$66,$34,$FF,$05,$EA,$8F,$B7,$EE,$7D,$CF,$86,$B8,$7E,$05,$C3,$F4
	.dc.b	$5F,$AC,$47,$FD,$A7,$F8,$8B,$E6,$3F,$56,$D8,$FF,$31,$FE,$36,$73
	.dc.b	$3F,$B0,$F5,$7F,$CD,$51,$FE,$4D,$6E,$7F,$E4,$3B,$E4,$FF,$91,$63
	.dc.b	$3E,$2F,$F9,$07,$F0,$27,$CA,$7E,$E9,$B8,$7F,$95,$D7,$CA,$66,$7F
	.dc.b	$7F,$F5,$9F,$BD,$C7,$FC,$A0,$DC,$3A,$4E,$BA,$BA,$7B,$7C,$3E,$26
	.dc.b	$E7,$FE,$63,$C1,$19,$8F,$F7,$AC,$61,$B1,$FE,$57,$C7,$FA,$94,$E5
	.dc.b	$FE,$78,$FF,$A8,$67,$8F,$C4,$6E,$7F,$D6,$73,$97,$71,$FB,$8F,$1B
	.dc.b	$FD,$6B,$5F,$E9,$06,$CF,$FA,$E7,$F1,$BF,$D3,$CE,$7C,$BF,$E3,$FF
	.dc.b	$54,$F8,$3A,$7F,$1C,$C3,$8E,$FF,$E6,$D9,$D5,$FF,$A7,$30,$F8,$CF
	.dc.b	$F9,$8D,$B8,$FF,$A9,$DF,$E4,$D1,$BD,$BF,$93,$C7,$F6,$D8,$E5,$E1
	.dc.b	$62,$72,$FD,$9F,$8D,$F2,$F5,$DA,$FF,$48,$62,$8C,$C7,$F6,$3C,$53
	.dc.b	$98,$C3,$63,$FB,$43,$5F,$F7,$C3,$FE,$BB,$8D,$F0,$FC,$F9,$E6,$D9
	.dc.b	$D4,$7F,$BF,$FC,$7F,$E4,$DF,$3B,$FC,$8B,$A7,$F9,$1B,$70,$FE,$51
	.dc.b	$F3,$9F,$C9,$8C,$5F,$F2,$67,$F9,$4F,$2C,$9F,$E5,$DC,$96,$BF,$95
	.dc.b	$FC,$E7,$F2,$E3,$1C,$5F,$A9,$8F,$E5,$31,$FD,$63,$EF,$1E,$1F,$F5
	.dc.b	$AF,$A9,$78,$B9,$8C,$33,$2F,$91,$3A,$6C,$7C,$18,$C6,$1F,$23,$E5
	.dc.b	$BF,$0E,$71,$7F,$D6,$DC,$28,$FE,$87,$F3,$93,$0F,$F7,$0F,$F0,$0F
	.dc.b	$EF,$DF,$E9,$51,$FD,$A2,$1F,$07,$4E,$18,$36,$F7,$72,$2B,$78,$74
	.dc.b	$F8,$2E,$47,$3E,$DB,$19,$3C,$3F,$23,$8C,$7A,$47,$FA,$E6,$FF,$84
	.dc.b	$B7,$C1,$FD,$1F,$C7,$79,$5B,$1F,$90,$8F,$C1,$11,$FC,$6B,$F8,$5F
	.dc.b	$96,$91,$F9,$06,$DF,$E1,$EF,$F8,$8C,$FC,$4F,$13,$E1,$7B,$9F,$0D
	.dc.b	$FD,$2F,$67,$F7,$2F,$0B,$F8,$FE,$1F,$EB,$36,$BF,$AF,$8F,$EE,$F8
	.dc.b	$3E,$73,$E6,$7C,$A7,$CD,$6E,$5F,$4B,$C1,$FA,$07,$D1,$F0,$6C,$7D
	.dc.b	$6F,$93,$E6,$47,$FD,$B1,$82,$3B,$FC,$DF,$D9,$F9,$BF,$B4,$DA,$FB
	.dc.b	$FD,$5F,$74,$FB,$DF,$37,$61,$B7,$FB,$F7,$F3,$17,$C3,$FE,$BD,$B3
	.dc.b	$FB,$B6,$CF,$F9,$7D,$AF,$FB,$57,$37,$2F,$E2,$F9,$7F,$18,$DB,$FC
	.dc.b	$8F,$91,$F8,$FF,$9B,$FC,$83,$63,$C7,$DB,$FA,$BC,$59,$3E,$4F,$C8
	.dc.b	$4A,$F2,$3F,$E2,$6F,$E0,$17,$7F,$F9,$09,$F3,$E1,$F9,$87,$35,$D8
	.dc.b	$F8,$32,$8D,$C6,$17,$A5,$CC,$E9,$F2,$B6,$79,$64,$FA,$29,$F2,$DF
	.dc.b	$D1,$5B,$3F,$E8,$56,$CF,$FA,$7C,$E0,$A7,$D5,$9F,$A4,$F5,$36,$BB
	.dc.b	$E5,$E7,$F9,$0F,$71,$EF,$F9,$0B,$3E,$4C,$3C,$9F,$03,$E3,$7F,$C4
	.dc.b	$F9,$7F,$AB,$DC,$CF,$95,$CC,$7C,$19,$F9,$47,$CB,$F8,$DB,$5E,$2F
	.dc.b	$C6,$1D,$5F,$E7,$06,$C7,$9B,$83,$FC,$D8,$EE,$7F,$D0,$B8,$FF,$0B
	.dc.b	$F4,$FE,$37,$11,$A4,$AF,$E9,$5F,$8D,$7D,$8D,$3F,$D5,$CD,$7E,$53
	.dc.b	$5F,$F6,$53,$5F,$C1,$E2,$F9,$8F,$C2,$7C,$C7,$E1,$5B,$7E,$78,$FC
	.dc.b	$51,$1F,$F5,$E7,$0F,$E1,$63,$F1,$DF,$05,$F8,$D3,$F1,$DF,$31,$6B
	.dc.b	$E0,$47,$FA,$1F,$F4,$7F,$F0,$26,$BF,$91,$8C,$7B,$1F,$95,$F7,$A5
	.dc.b	$FC,$7F,$90,$BB,$7F,$24,$3D,$33,$F9,$BF,$8E,$CE,$3F,$9B,$7F,$D4
	.dc.b	$BF,$18,$F7,$3F,$A9,$7C,$AB,$BB,$FD,$D9,$39,$CF,$FD,$91,$1F,$EE
	.dc.b	$FF,$98,$FE,$D3,$5B,$CF,$F8,$1D,$ED,$46,$5F,$F3,$EE,$1D,$AF,$AF
	.dc.b	$B3,$67,$8C,$D7,$FD,$07,$16,$CD,$31,$37,$1F,$2D,$7D,$8C,$7A,$71
	.dc.b	$FE,$D7,$E5,$2F,$32,$6C,$4B,$E3,$5B,$6D,$6D,$B3,$E5,$FA,$AF,$F1
	.dc.b	$77,$CB,$FF,$85,$DB,$8F,$FB,$ED,$AF,$ED,$DA,$FF,$28,$1F,$30,$EF
	.dc.b	$FE,$05,$1F,$BD,$32,$E1,$FE,$01,$1B,$ED,$FF,$13,$36,$3C,$1E,$D7
	.dc.b	$F1,$A6,$E9,$E5,$8C,$47,$8D,$AF,$F1,$AF,$AB,$F8,$8F,$C2,$6B,$F6
	.dc.b	$B8,$B5,$BF,$A7,$8F,$F0,$18,$2F,$FD,$CD,$7C,$6B,$57,$C2,$30,$75
	.dc.b	$DB,$2F,$F4,$AD,$7F,$B0,$3F,$ED,$63,$E3,$38,$7F,$CD,$CD,$DF,$E6
	.dc.b	$7A,$BF,$2D,$B3,$F8,$21,$3C,$73,$FC,$93,$70,$EB,$EB,$78,$AB,$6D
	.dc.b	$9D,$47,$97,$0C,$4F,$A3,$47,$D3,$7F,$33,$E8,$FE,$8D,$F3,$5F,$75
	.dc.b	$B8,$7F,$30,$9D,$69,$FD,$21,$B1,$FD,$41,$B7,$D2,$39,$4F,$D5,$BB
	.dc.b	$0F,$99,$7D,$57,$FA,$5A,$3F,$9E,$FA,$DC,$8E,$B3,$AF,$88,$E2,$D9
	.dc.b	$FE,$A2,$D7,$F8,$43,$73,$FF,$1B,$E7,$FE,$F8,$9F,$EA,$E3,$E8,$1E
	.dc.b	$5F,$F9,$DE,$30,$D8,$CC,$7F,$9E,$DA,$FF,$9B,$E5,$F1,$9E,$C8,$E2
	.dc.b	$76,$9D,$DF,$8A,$7E,$0E,$DF,$E5,$5E,$BB,$A1,$A7,$C6,$EA,$7E,$57
	.dc.b	$E9,$CB,$F9,$20,$FA,$B7,$CC,$7A,$5B,$1F,$E9,$16,$BF,$B3,$6B,$FD
	.dc.b	$87,$A3,$F7,$0F,$F6,$76,$BF,$CC,$6F,$F9,$0F,$77,$A3,$EA,$7C,$5B
	.dc.b	$7F,$9F,$E7,$FC,$F7,$EF,$F9,$FE,$33,$FC,$69,$87,$F7,$16,$BF,$CE
	.dc.b	$F7,$FF,$3A,$7C,$EF,$C7,$D3,$F0,$13,$E5,$E6,$F5,$11,$FE,$A3,$C2
	.dc.b	$F1,$BE,$39,$23,$F2,$98,$FE,$13,$AF,$F9,$B6,$BF,$E9,$3E,$84,$63
	.dc.b	$AF,$FC,$7F,$18,$EE,$FA,$3C,$FE,$3F,$2B,$E5,$13,$2F,$94,$FF,$AE
	.dc.b	$F8,$BE,$B7,$C5,$FD,$5E,$9F,$F4,$3D,$7F,$83,$DA,$FB,$FA,$DF,$31
	.dc.b	$F7,$BE,$73,$F3,$3F,$94,$FF,$8F,$78,$3C,$CD,$DF,$FC,$67,$C6,$AF
	.dc.b	$C6,$6F,$FE,$4C,$CB,$FD,$98,$D7,$F9,$73,$5F,$F1,$63,$5F,$DB,$72
	.dc.b	$78,$3E,$3A,$47,$EE,$E3,$F2,$3F,$29,$F9,$2F,$9D,$93,$FB,$28,$F0
	.dc.b	$FC,$27,$E9,$70,$34,$93,$FE,$06,$F9,$77,$C5,$7E,$79,$B1,$FE,$38
	.dc.b	$6B,$F1,$7D,$1F,$87,$E7,$7E,$8B,$C4,$93,$FB,$CE,$14,$FE,$B7,$F3
	.dc.b	$0F,$F3,$14,$FF,$BB,$5A,$FF,$4F,$6B,$FE,$4D,$6B,$F8,$09,$7C,$7F
	.dc.b	$D2,$24,$7F,$93,$3E,$AF,$BB,$E1,$F3,$7F,$6B,$E0,$48,$C1,$FC,$2B
	.dc.b	$0F,$F2,$C3,$E3,$F8,$3F,$E1,$51,$88,$F2,$47,$F6,$37,$F9,$06,$2F
	.dc.b	$8E,$FC,$19,$23,$C5,$87,$F0,$4F,$96,$91,$FC,$91,$FE,$1D,$F5,$5F
	.dc.b	$EE,$C8,$FE,$36,$BE,$7F,$F8,$D3,$63,$C0,$D7,$FB,$A3,$5F,$F3,$92
	.dc.b	$FA,$0E,$B3,$27,$AC,$7F,$18,$D7,$F7,$7D,$BF,$EE,$2E,$CE,$8D,$BD
	.dc.b	$79,$73,$0F,$98,$D7,$F8,$15,$FE,$0A,$B9,$61,$FC,$88,$93,$F4,$FE
	.dc.b	$4B,$C0,$61,$FE,$15,$8F,$9D,$D1,$F0,$BC,$0F,$E6,$5E,$8B,$DF,$A7
	.dc.b	$C8,$8F,$93,$F4,$16,$B8,$1F,$29,$FE,$75,$F5,$6F,$7F,$73,$93,$FA
	.dc.b	$4C,$65,$FC,$B7,$CA,$7F,$3A,$B9,$18,$3F,$19,$F2,$5F,$C0,$26,$46
	.dc.b	$0F,$EB,$F8,$7F,$42,$F9,$D9,$18,$3F,$29,$F2,$7F,$44,$8F,$FA,$8B
	.dc.b	$E1,$3F,$C4,$7C,$3F,$EC,$7F,$9D,$FE,$11,$3F,$D7,$9A,$FF,$48,$6C
	.dc.b	$7F,$B3,$79,$BF,$D2,$13,$F8,$AF,$23,$F1,$24,$7F,$12,$FE,$33,$D0
	.dc.b	$FC,$8B,$23,$F2,$31,$FD,$73,$85,$F3,$D4,$93,$FE,$B8,$E0,$FE,$66
	.dc.b	$DF,$53,$2F,$F4,$5A,$E2,$E4,$FE,$EF,$F5,$2F,$9D,$FE,$DC,$BC,$EA
	.dc.b	$62,$33,$E3,$5A,$E3,$18,$7F,$C0,$3F,$81,$7D,$97,$FB,$5B,$77,$FF
	.dc.b	$78,$37,$D1,$FF,$2F,$63,$E8,$AF,$EE,$2F,$EE,$5F,$DE,$35,$FF,$30
	.dc.b	$35,$FF,$6F,$35,$E7,$E6,$7A,$BC,$49,$1E,$E8,$FF,$05,$63,$FD,$5E
	.dc.b	$07,$C1,$F9,$6C,$9A,$F3,$F7,$F2,$D6,$7F,$C3,$B5,$C5,$F9,$11,$FE
	.dc.b	$15,$FA,$AF,$91,$1F,$E2,$27,$FA,$BE,$AD,$3F,$CF,$91,$86,$EF,$C7
	.dc.b	$0E,$4C,$BE,$51,$FD,$73,$A7,$F8,$F7,$B4,$F2,$91,$86,$D7,$FA,$16
	.dc.b	$75,$3F,$C9,$7D,$E7,$63,$7E,$97,$FD,$0E,$BF,$C5,$33,$98,$FF,$20
	.dc.b	$F0,$BE,$6F,$FC,$57,$F3,$92,$73,$1F,$E4,$89,$D4,$60,$87,$F8,$56
	.dc.b	$BF,$A5,$F0,$7F,$82,$67,$FD,$81,$3F,$E9,$FC,$FF,$9C,$71,$7B,$A7
	.dc.b	$33,$FE,$0C,$6A,$D3,$D7,$F3,$3E,$1A,$93,$FB,$98,$F2,$79,$FC,$C9
	.dc.b	$1F,$16,$FF,$A5,$3F,$EA,$FB,$F8,$BF,$0D,$C2,$3E,$0D,$7C,$48,$F7
	.dc.b	$5F,$F9,$33,$E5,$A5,$7C,$68,$CB,$48,$7C,$1E,$EF,$F2,$9F,$81,$F4
	.dc.b	$FC,$AF,$F1,$AC,$F8,$5A,$FF,$85,$A3,$EC,$6E,$F9,$4E,$33,$FE,$7A
	.dc.b	$9F,$AF,$1F,$E5,$EF,$0A,$4F,$F9,$F9,$FF,$30,$FC,$8F,$DC,$F9,$4E
	.dc.b	$64,$9F,$F6,$56,$FF,$84,$8E,$F8,$EE,$7F,$91,$47,$E1,$23,$FD,$0B
	.dc.b	$F6,$9F,$EE,$E9,$FF,$42,$B5,$FF,$42,$B7,$93,$FC,$8D,$F5,$5F,$E3
	.dc.b	$19,$FF,$4A,$3F,$E3,$A7,$F0,$0D,$7F,$C9,$2D,$77,$F0,$7F,$A0,$8F
	.dc.b	$D6,$C7,$FA,$49,$B3,$DB,$B7,$FD,$03,$F5,$3F,$E0,$58,$FD,$85,$7E
	.dc.b	$1E,$31,$DF,$FC,$03,$44,$62,$75,$CD,$FA,$2F,$43,$FD,$21,$E3,$FF
	.dc.b	$15,$C7,$9D,$7F,$61,$F2,$32,$6B,$C5,$FF,$4B,$7C,$A7,$E3,$27,$FD
	.dc.b	$F4,$D7,$F8,$7B,$5F,$F7,$EE,$3F,$B5,$FC,$97,$FD,$3F,$1F,$C5,$30
	.dc.b	$F4,$E8,$9C,$CF,$F9,$FD,$B3,$AD,$1D,$5E,$2E,$C4,$FF,$1E,$F0,$1F
	.dc.b	$05,$E4,$F9,$DF,$E4,$39,$7E,$F6,$30,$ED,$F4,$DF,$28,$D1,$EA,$71
	.dc.b	$5D,$23,$DB,$5D,$23,$10,$D5,$39,$BE,$B8,$77,$AE,$8F,$D5,$BC,$C7
	.dc.b	$EF,$DB,$A7,$F0,$E8,$E9,$8F,$96,$31,$1F,$CF,$3E,$3F,$93,$EE,$0E
	.dc.b	$1F,$82,$D7,$81,$DC,$E0,$F9,$5F,$E8,$FF,$11,$F8,$CB,$F6,$C7,$06
	.dc.b	$C6,$71,$F0,$4F,$80,$F9,$27,$2E,$4E,$8F,$18,$F3,$4F,$18,$FE,$3D
	.dc.b	$BF,$FA,$5E,$BF,$08,$7E,$65,$C4,$7F,$B0,$77,$FF,$77,$CE,$A3,$8F
	.dc.b	$C3,$FE,$8B,$5F,$CC,$C7,$FA,$FE,$7F,$DE,$9F,$17,$88,$FC,$1B,$5F
	.dc.b	$EC,$AD,$E5,$7D,$1F,$3A,$FC,$4C,$61,$FF,$80,$2F,$FA,$31,$B1,$BA
	.dc.b	$FD,$97,$87,$78,$3F,$62,$DE,$D7,$CA,$E7,$5F,$F0,$FE,$3E,$BF,$21
	.dc.b	$F5,$4F,$9A,$E9,$18,$78,$63,$89,$D5,$71,$7F,$CA,$0F,$F6,$56,$E7
	.dc.b	$F8,$5B,$EB,$DB,$F0,$DB,$5B,$F9,$8E,$87,$5C,$EC,$7C,$FD,$5F,$54
	.dc.b	$E5,$F7,$2F,$93,$B1,$86,$57,$F3,$29,$D3,$63,$EA,$57,$46,$F9,$4F
	.dc.b	$8C,$F9,$4D,$AF,$17,$EC,$9F,$E0,$F7,$C9,$19,$DB,$F8,$1C,$7F,$21
	.dc.b	$C5,$F2,$7F,$4E,$DE,$3F,$A2,$30,$CA,$FE,$8F,$3A,$6C,$79,$A3,$17
	.dc.b	$F2,$FC,$FF,$C9,$BE,$01,$C1,$FE,$B5,$F0,$3E,$2F,$67,$DA,$AD,$F6
	.dc.b	$7F,$81,$67,$C9,$E3,$E5,$B9,$FC,$FD,$5F,$1D,$C7,$A3,$E7,$B6,$7F
	.dc.b	$C3,$31,$F6,$F1,$FC,$5E,$9E,$58,$C1,$F3,$AD,$F1,$5E,$3F,$9E,$FB
	.dc.b	$2D,$C3,$FC,$47,$1C,$C8,$FF,$82,$DE,$DD,$FE,$B3,$F0,$3F,$17,$67
	.dc.b	$FC,$57,$1F,$3F,$D9,$C6,$5F,$17,$A1,$C7,$FE,$73,$BF,$77,$C1,$FC
	.dc.b	$5F,$07,$F1,$6D,$8F,$F1,$AC,$7B,$48,$FF,$87,$5F,$DD,$7B,$3F,$B3
	.dc.b	$C3,$E9,$1F,$E4,$1F,$0F,$F6,$8F,$F9,$0F,$DF,$B3,$FB,$B8,$DB,$79
	.dc.b	$BF,$B0,$BF,$EF,$EC,$3F,$87,$4F,$36,$C6,$63,$BF,$C8,$7E,$3A,$76
	.dc.b	$FF,$18,$AF,$ED,$6E,$67,$7D,$1E,$48,$D5,$7F,$56,$7F,$AC,$BC,$A3
	.dc.b	$5E,$46,$FD,$7E,$1E,$B5,$AE,$EE,$1F,$E6,$78,$30,$4F,$F1,$F6,$CE
	.dc.b	$9B,$3F,$CE,$BE,$EB,$E1,$4F,$F6,$86,$C7,$FA,$2A,$5E,$BF,$CC,$9C
	.dc.b	$CB,$EA,$3E,$E3,$F4,$9F,$A9,$A7,$F8,$31,$CB,$76,$FA,$B1,$AC,$7F
	.dc.b	$BB,$BD,$0C,$57,$FA,$69,$F1,$F1,$BE,$9F,$E3,$CF,$FB,$07,$C7,$FE
	.dc.b	$61,$18,$C7,$FC,$1B,$CF,$FC,$C3,$1F,$E3,$4D,$BE,$11,$FC,$76,$B8
	.dc.b	$3F,$E2,$99,$D5,$7A,$7B,$7C,$FE,$AF,$A1,$B5,$F8,$6C,$D3,$FD,$35
	.dc.b	$B1,$F5,$1B,$8F,$F6,$86,$BF,$DE,$7C,$AF,$C6,$61,$FE,$32,$6D,$7C
	.dc.b	$4F,$8F,$FC,$DC,$62,$BF,$D0,$DD,$EF,$73,$F9,$99,$FC,$D3,$DE,$B8
	.dc.b	$7F,$28,$E1,$FE,$45,$D5,$FC,$8D,$B1,$FC,$3F,$93,$F8,$B7,$17,$F3
	.dc.b	$C6,$C7,$FA,$9A,$3F,$95,$7C,$9E,$33,$FD,$E7,$77,$8D,$7D,$4B,$E5
	.dc.b	$3E,$EF,$99,$FB,$28,$C6,$3F,$1E,$32,$FC,$9F,$8C,$B7,$F8,$97,$F0
	.dc.b	$A7,$EE,$BE,$3F,$FD,$79,$3B,$3F,$6A,$E2,$7F,$87,$7D,$47,$F3,$6E
	.dc.b	$5F,$F2,$6E,$6F,$A9,$FE,$E9,$39,$C3,$FC,$CD,$39,$7F,$A6,$CF,$F4
	.dc.b	$73,$7E,$7D,$8F,$EC,$55,$F9,$4E,$D3,$47,$9D,$DB,$8E,$57,$F5,$8F
	.dc.b	$95,$BA,$7C,$87,$F4,$3F,$23,$FA,$2F,$97,$B5,$F8,$B7,$0F,$E2,$5F
	.dc.b	$C6,$11,$FE,$AF,$F0,$7F,$D7,$4F,$D5,$FD,$4D,$8F,$F0,$67,$D2,$FF
	.dc.b	$0E,$C7,$FA,$AB,$75,$F4,$F3,$FF,$83,$A3,$52,$FC,$C3,$FC,$54,$FF
	.dc.b	$8F,$9A,$FF,$93,$DA,$FF,$9B,$9A,$F6,$F9,$BF,$EE,$1E,$27,$56,$E3
	.dc.b	$E4,$EA,$F0,$1F,$52,$6C,$7F,$30,$BF,$E5,$17,$DF,$D8,$FF,$1C,$46
	.dc.b	$23,$C1,$1F,$E1,$BC,$5E,$97,$F7,$FF,$1B,$D8,$DC,$3F,$C3,$5F,$01
	.dc.b	$FC,$49,$FF,$53,$37,$0F,$F7,$87,$33,$B7,$FC,$75,$39,$9F,$89,$7E
	.dc.b	$22,$5F,$F3,$0C,$FC,$2F,$F8,$2B,$C6,$FF,$08,$7D,$47,$F8,$45,$B1
	.dc.b	$F8,$AF,$8E,$FA,$1F,$29,$F9,$46,$C7,$C1,$F9,$AF,$79,$1E,$CF,$F8
	.dc.b	$4F,$E6,$BE,$C7,$CC,$FF,$87,$5B,$3F,$E6,$5F,$9E,$1F,$E9,$47,$0D
	.dc.b	$D7,$FB,$A7,$D3,$FE,$A7,$C4,$FF,$65,$CF,$0F,$8C,$FE,$D3,$F9,$13
	.dc.b	$A3,$FC,$D0,$DC,$7F,$A7,$B7,$1F,$F4,$1B,$5F,$F5,$8F,$C1,$BB,$FF
	.dc.b	$CA,$F1,$FE,$27,$8C,$7C,$4F,$F8,$D7,$9F,$FC,$BF,$5A,$D3,$F7,$3E
	.dc.b	$57,$F8,$E4,$FF,$36,$6D,$FE,$C2,$CF,$FC,$87,$1E,$6E,$AF,$29,$23
	.dc.b	$CA,$FF,$A0,$77,$7C,$27,$F8,$CA,$5A,$FC,$68,$C2,$F0,$7F,$C2,$FE
	.dc.b	$27,$F9,$F6,$F8,$D5,$CF,$F9,$86,$1F,$3D,$B8,$7F,$87,$FE,$A9,$D5
	.dc.b	$BF,$4B,$FC,$27,$F5,$5D,$70,$FE,$EE,$DB,$FF,$55,$B5,$FF,$5E,$35
	.dc.b	$FE,$2F,$E3,$FF,$96,$CF,$27,$E1,$CE,$AC,$5F,$F9,$DF,$89,$F2,$FF
	.dc.b	$CA,$7A,$BE,$4B,$67,$F9,$4C,$7F,$A8,$FD,$0F,$C2,$65,$FE,$CB,$6B
	.dc.b	$FE,$EA,$6B,$FC,$2B,$DD,$FE,$10,$7D,$4F,$6A,$DF,$F1,$08,$FF,$38
	.dc.b	$79,$98,$AF,$E8,$EF,$DA,$F6,$65,$79,$FE,$0A,$D7,$E5,$27,$FC,$F1
	.dc.b	$E1,$E3,$97,$F9,$DB,$FE,$3C,$F2,$3F,$C2,$78,$7F,$B8,$FE,$63,$FD
	.dc.b	$7B,$F1,$48,$C6,$3F,$EB,$EF,$67,$0F,$FB,$9A,$B7,$3A,$85,$FF,$A2
	.dc.b	$FE,$49,$5A,$9E,$07,$3A,$E1,$5B,$9D,$47,$FB,$CE,$3F,$D2,$8D,$C3
	.dc.b	$C2,$D3,$F2,$27,$E4,$AB,$F2,$1C,$7F,$E1,$5C,$3C,$31,$8F,$6B,$D3
	.dc.b	$18,$EF,$53,$E3,$B3,$9F,$E7,$27,$37,$8C,$39,$63,$F1,$27,$2D,$C3
	.dc.b	$93,$6B,$85,$3E,$A9,$F1,$3F,$EC,$18,$FF,$80,$31,$8F,$8F,$FF,$28
	.dc.b	$B6,$33,$3B,$7D,$C3,$E5,$1F,$4F,$FC,$5A,$32,$65,$C3,$C2,$37,$E2
	.dc.b	$7F,$97,$9B,$1E,$2F,$6F,$F7,$0D,$D7,$C7,$18,$8F,$0E,$28,$C4,$7C
	.dc.b	$7F,$AB,$F9,$4F,$D4,$31,$C1,$F5,$76,$FA,$D8,$2B,$75,$F8,$C3,$F8
	.dc.b	$43,$89,$DB,$AE,$1F,$B0,$DE,$1E,$B5,$F4,$E1,$C1,$F8,$3A,$FC,$0C
	.dc.b	$FD,$1F,$73,$17,$F8,$7D,$3F,$C3,$93,$F9,$96,$BF,$96,$38,$19,$9F
	.dc.b	$53,$C5,$AF,$F2,$28,$46,$23,$89,$ED,$2D,$79,$27,$71,$D4,$F1,$79
	.dc.b	$7E,$67,$F3,$7F,$57,$F9,$DE,$9F,$D8,$EB,$F9,$88,$C6,$BF,$97,$E5
	.dc.b	$46,$2F,$F9,$97,$D3,$FE,$23,$0E,$2D,$BE,$15,$BC,$3A,$5F,$F3,$CB
	.dc.b	$DB,$9F,$3B,$83,$FC,$F5,$B3,$FD,$FE,$3C,$1E,$83,$EA,$DF,$19,$86
	.dc.b	$CF,$5C,$BF,$5D,$87,$23,$AD,$F1,$BF,$27,$EE,$DF,$8E,$BF,$DA,$35
	.dc.b	$FE,$9C,$DC,$4E,$61,$E2,$FF,$58,$7F,$D7,$AD,$7F,$4E,$D7,$F9,$C7
	.dc.b	$9F,$FE,$C2,$AE,$7F,$35,$E2,$24,$FA,$4F,$85,$DD,$F3,$FB,$1F,$58
	.dc.b	$F9,$8F,$F6,$5C,$61,$B1,$9A,$FD,$BC,$E5,$BC,$FF,$BC,$7F,$DD,$3F
	.dc.b	$56,$F9,$9F,$B9,$08,$C3,$FE,$17,$9E,$0B,$93,$F0,$FB,$BE,$57,$E5
	.dc.b	$92,$7F,$C3,$DF,$0A,$F8,$0F,$53,$FE,$17,$8F,$2B,$5F,$E2,$39,$FF
	.dc.b	$A5,$7D,$DF,$6C,$E6,$7F,$7B,$CD,$FD,$09,$F8,$DC,$5F,$53,$BF,$ED
	.dc.b	$9F,$8B,$F2,$3E,$EF,$D2,$4A,$DE,$A8,$FC,$5F,$7F,$EE,$B7,$FF,$16
	.dc.b	$E7,$FE,$04,$BF,$2F,$B1,$E2,$23,$F3,$39,$9F,$15,$F8,$AF,$9F,$93
	.dc.b	$FE,$3D,$9C,$BF,$6F,$B5,$F4,$7F,$B7,$8F,$F0,$EC,$63,$BF,$F9,$6C
	.dc.b	$23,$0F,$F3,$06,$BF,$1B,$F1,$BF,$88,$FF,$89,$B9,$51,$88,$FF,$26
	.dc.b	$B6,$33,$CB,$DA,$BC,$7F,$1B,$71,$FF,$52,$E1,$DD,$AF,$F4,$49,$DB
	.dc.b	$FD,$0B,$97,$FC,$E5,$F1,$9F,$BD,$FA,$8F,$F7,$6C,$FD,$46,$BF,$87
	.dc.b	$D3,$F6,$73,$F9,$BD,$3E,$74,$FA,$F8,$3F,$37,$18,$8F,$EE,$1A,$FB
	.dc.b	$27,$F5,$1F,$31,$D3,$8F,$F8,$BB,$F8,$96,$EF,$FC,$3E,$35,$97,$F0
	.dc.b	$2E,$EF,$DF,$FD,$14,$CB,$F1,$3D,$7F,$9E,$7C,$C7,$F0,$0F,$6F,$F6
	.dc.b	$DD,$46,$92,$3E,$37,$C8,$7F,$0A,$F1,$64,$7D,$FE,$54,$FF,$1D,$EA
	.dc.b	$FE,$1F,$E9,$CA,$F0,$C7,$F2,$5F,$C8,$DE,$E7,$F0,$56,$CF,$CF,$6B
	.dc.b	$FD,$DB,$A1,$18,$8F,$F3,$DE,$3F,$0F,$83,$F8,$DF,$D0,$7F,$1E,$3F
	.dc.b	$94,$C6,$67,$F8,$EF,$CF,$FF,$22,$23,$F1,$3E,$83,$F8,$1F,$59,$FC
	.dc.b	$AC,$23,$0F,$E6,$1A,$FF,$81,$9A,$FF,$83,$1F,$F6,$0B,$BF,$91,$E7
	.dc.b	$18,$8E,$11,$FC,$97,$8F,$1D,$1F,$C3,$BE,$0B,$F9,$B9,$FD,$1E,$7F
	.dc.b	$A6,$7D,$67,$F0,$DB,$FB,$71,$97,$F8,$2C,$65,$FB,$BB,$FF,$B2,$7E
	.dc.b	$43,$FA,$0F,$C2,$49,$FC,$1C,$7F,$53,$23,$83,$9F,$C8,$7F,$26,$F4
	.dc.b	$95,$49,$3F,$CE,$4F,$8D,$F2,$3F,$28,$F7,$91,$F9,$0F,$C2,$FA,$1D
	.dc.b	$47,$1D,$D3,$F4,$A7,$FB,$33,$FD,$63,$E9,$E5,$7E,$6A,$3E,$EF,$C8
	.dc.b	$FE,$EC,$91,$FB,$B7,$E6,$7C,$57,$F6,$0F,$89,$93,$FA,$28,$FE,$E5
	.dc.b	$84,$8F,$D2,$BF,$3B,$E2,$3F,$BD,$7E,$4B,$FC,$72,$7E,$5B,$79,$3F
	.dc.b	$C1,$2D,$EA,$FF,$34,$7D,$63,$E9,$3F,$17,$F2,$3F,$80,$26,$1F,$89
	.dc.b	$7F,$07,$EF,$48,$FC,$63,$5F,$D0,$71,$3B,$BC,$7F,$15,$27,$FA,$1C
	.dc.b	$E5,$B3,$FE,$11,$8C,$47,$E0,$9B,$A7,$F8,$83,$8B,$FD,$C5,$3F,$67
	.dc.b	$EA,$FF,$90,$CF,$F8,$93,$C0,$7A,$9F,$E1,$B6,$D6,$FD,$3F,$ED,$37
	.dc.b	$FE,$72,$8F,$E7,$6D,$7F,$CD,$2D,$7F,$CD,$ED,$7F,$D7,$5E,$BF,$E3
	.dc.b	$48,$FE,$2F,$EB,$1F,$1F,$DB,$D9,$FD,$9D,$FA,$2B,$AC,$60,$CF,$A5
	.dc.b	$23,$C6,$D7,$FC,$F7,$C1,$FD,$E5,$FC,$C7,$77,$B1,$FE,$46,$7F,$87
	.dc.b	$B5,$FF,$02,$CE,$E3,$2F,$F9,$89,$FF,$44,$B5,$FF,$54,$35,$FF,$7D
	.dc.b	$B5,$7D,$67,$E1,$DF,$F7,$E7,$E5,$BF,$E7,$8E,$0F,$F2,$9B,$7A,$3F
	.dc.b	$D6,$55,$BA,$E3,$C1,$FC,$52,$3F,$AB,$C7,$FA,$69,$AF,$08,$FE,$77
	.dc.b	$F3,$32,$3F,$AD,$7A,$48,$C4,$7F,$5F,$E3,$76,$7F,$65,$24,$7F,$65
	.dc.b	$7F,$8B,$7D,$1F,$F9,$64,$91,$FD,$FF,$81,$F0,$DF,$E5,$86,$C7,$F5
	.dc.b	$09,$DB,$FD,$43,$DC,$4E,$A3,$F8,$BE,$5F,$80,$7C,$FC,$BF,$D5,$08
	.dc.b	$FF,$81,$67,$E7,$57,$2F,$80,$FE,$FD,$18,$3E,$97,$93,$FE,$5A,$F8
	.dc.b	$FF,$EB,$A4,$BF,$86,$F0,$E5,$91,$D5,$AF,$F5,$1F,$39,$E0,$7F,$9E
	.dc.b	$49,$1F,$07,$D8,$F8,$06,$1F,$E6,$33,$C1,$D1,$FE,$34,$F8,$EF,$ED
	.dc.b	$E4,$7E,$AB,$FD,$0F,$C2,$91,$FE,$11,$9F,$F3,$FB,$E3,$8F,$EE,$FF
	.dc.b	$15,$FE,$7C,$24,$7F,$7E,$F5,$7F,$D1,$DF,$27,$FD,$34,$91,$CD,$FF
	.dc.b	$3A,$79,$6F,$36,$92,$7F,$C4,$11,$97,$FC,$69,$F1,$9F,$C2,$BB,$7F
	.dc.b	$C7,$6D,$8F,$F1,$D3,$79,$1F,$5A,$F9,$4F,$F4,$DF,$D6,$FF,$90,$EF
	.dc.b	$FC,$80,$DA,$34,$9D,$FC,$3E,$1E,$57,$F9,$36,$7F,$8B,$E4,$87,$3B
	.dc.b	$EB,$3F,$79,$E2,$2B,$FC,$A3,$7C,$2F,$C3,$8F,$FA,$DE,$F7,$5C,$F0
	.dc.b	$7E,$DF,$FE,$B4,$25,$FF,$A2,$3C,$DF,$E2,$59,$7F,$BB,$FE,$5F,$FC
	.dc.b	$DD,$3F,$E5,$C6,$BE,$52,$38,$70,$1C,$3E,$1F,$AE,$7D,$27,$FA,$5B
	.dc.b	$ED,$BF,$D8,$F2,$7F,$D8,$F3,$A8,$FE,$E5,$4F,$85,$FF,$4C,$47,$ED
	.dc.b	$3E,$BB,$C3,$F0,$FE,$3F,$21,$18,$9F,$95,$18,$7E,$65,$F3,$CF,$35
	.dc.b	$EA,$C3,$AD,$EE,$FD,$58,$75,$AE,$91,$E1,$39,$D7,$08,$E5,$7F,$3E
	.dc.b	$ED,$7D,$A6,$ED,$FC,$17,$1F,$14,$E4,$F9,$3F,$25,$C5,$A6,$4F,$A0
	.dc.b	$E5,$A6,$4F,$C4,$3E,$C7,$EB,$7C,$7F,$E1,$DF,$41,$C2,$FD,$CA,$35
	.dc.b	$18,$84,$F3,$7F,$29,$3A,$C7,$EF,$C6,$8E,$A7,$3E,$1F,$16,$2D,$7F
	.dc.b	$C1,$51,$FE,$2D,$E0,$7D,$3F,$97,$5F,$35,$EF,$D9,$C5,$FE,$3E,$F6
	.dc.b	$FF,$8D,$71,$FC,$5B,$FE,$5C,$E2,$FD,$94,$60,$FF,$59,$BF,$EB,$7B
	.dc.b	$FD,$0E,$9D,$27,$F5,$BC,$7E,$77,$FB,$3E,$0F,$AA,$D5,$F0,$C7,$FD
	.dc.b	$AF,$CB,$FE,$19,$7F,$D8,$9F,$5F,$FD,$45,$CF,$B7,$D7,$AD,$39,$CB
	.dc.b	$FA,$33,$77,$FA,$93,$AF,$37,$84,$E7,$0F,$F6,$6B,$FE,$08,$7B,$F8
	.dc.b	$FF,$ED,$1A,$EA,$D9,$34,$D7,$81,$8A,$FF,$6D,$37,$1F,$1F,$97,$E7
	.dc.b	$8C,$65,$F4,$BE,$B1,$F3,$3F,$25,$B8,$7F,$B2,$5B,$87,$FA,$6D,$AF
	.dc.b	$F0,$5D,$DF,$57,$FC,$5E,$BF,$DA,$0D,$9F,$F5,$13,$5E,$0D,$7F,$8C
	.dc.b	$71,$F6,$C7,$F0,$AD,$AF,$E3,$9E,$5F,$F3,$37,$F9,$44,$EB,$2F,$B3
	.dc.b	$87,$FB,$FA,$7C,$3E,$1F,$E0,$E3,$18,$F8,$A3,$F0,$D8,$BE,$5A,$92
	.dc.b	$FF,$DD,$11,$97,$FB,$C7,$81,$DB,$C3,$FC,$07,$A1,$FC,$5E,$7C,$67
	.dc.b	$F3,$1C,$93,$B8,$CE,$5F,$CF,$A7,$31,$FE,$11,$BD,$4F,$F8,$36,$FF
	.dc.b	$20,$FF,$BE,$DA,$FF,$33,$E4,$FF,$08,$F3,$7E,$16,$73,$3F,$2B,$AF
	.dc.b	$DA,$4C,$3D,$6E,$FE,$32,$47,$F8,$4D,$B3,$FE,$D7,$D3,$F7,$2D,$ED
	.dc.b	$F8,$E4,$AF,$F0,$D7,$67,$D0,$F6,$BE,$1B,$63,$DF,$CB,$F8,$49,$C5
	.dc.b	$FB,$6A,$F2,$7E,$93,$47,$95,$EB,$26,$3F,$E2,$77,$EB,$F8,$52,$3E
	.dc.b	$AB,$70,$FE,$2B,$F3,$1F,$BB,$9F,$E2,$F0,$8C,$3A,$F0,$3F,$BD,$46
	.dc.b	$23,$FA,$DA,$F9,$1F,$E0,$7F,$94,$3E,$2F,$F8,$55,$73,$9C,$C6,$3E
	.dc.b	$53,$E3,$CF,$F6,$46,$BF,$DA,$5A,$FF,$7F,$6A,$7F,$AD,$E3,$FE,$D3
	.dc.b	$FA,$D7,$A5,$D7,$4F,$0C,$62,$7F,$DB,$14,$F4,$FF,$C2,$8D,$8F,$82
	.dc.b	$DC,$3F,$9C,$B5,$F3,$61,$C0,$93,$C1,$FF,$0F,$78,$CE,$3F,$82,$D9
	.dc.b	$FF,$74,$7D,$87,$CC,$E3,$FE,$A1,$87,$FA,$82,$FE,$3F,$D7,$BE,$93
	.dc.b	$FC,$50,$DC,$3E,$AB,$6B,$FA,$6B,$5F,$F0,$6B,$D6,$3F,$C3,$9C,$BF
	.dc.b	$CE,$70,$EF,$39,$AF,$F1,$4C,$E6,$BF,$C4,$5C,$DF,$C7,$9B,$E4,$7F
	.dc.b	$3E,$FB,$0F,$9F,$7F,$EC,$D8,$FE,$83,$5B,$8F,$F1,$9F,$C6,$3E,$7B
	.dc.b	$35,$FD,$1A,$31,$87,$EF,$29,$93,$EB,$3D,$E4,$9F,$7B,$F0,$B0,$91
	.dc.b	$F0,$9B,$5F,$E2,$CE,$C6,$0F,$93,$F0,$1F,$4B,$A6,$E1,$FA,$EF,$17
	.dc.b	$E5,$70,$7F,$6C,$6C,$7F,$90,$FC,$BF,$F4,$0B,$BE,$2F,$E4,$FD,$DF
	.dc.b	$D8,$3C,$17,$CF,$E2,$BF,$73,$18,$7F,$D4,$DC,$5F,$C3,$C8,$FF,$A6
	.dc.b	$DF,$E5,$D9,$F3,$F9,$CF,$E6,$6D,$9F,$8D,$C6,$F3,$7F,$0B,$F2,$9F
	.dc.b	$86,$6C,$EA,$3F,$0E,$47,$F0,$EF,$F9,$C7,$E3,$3F,$79,$1F,$EE,$2F
	.dc.b	$13,$8F,$2F,$FB,$8A,$72,$B8,$BE,$27,$AB,$FD,$18,$FE,$FF,$F5,$FF
	.dc.b	$D1,$23,$38,$7C,$B9,$F8,$9F,$39,$B8,$F8,$6F,$0E,$3F,$E9,$FF,$4D
	.dc.b	$67,$E9,$38,$7E,$4F,$CE,$5C,$BC,$51,$FD,$7A,$3F,$4B,$F5,$5F,$A4
	.dc.b	$24,$7E,$91,$FD,$3F,$D4,$C8,$FD,$3B,$6B,$F9,$B7,$CE,$FD,$93,$FB
	.dc.b	$0F,$CA,$5A,$FE,$9F,$E9,$BE,$C3,$F9,$DE,$FF,$DB,$3C,$EF,$F0,$C4
	.dc.b	$62,$BF,$B8,$C7,$E7,$73,$FF,$15,$78,$06,$92,$7F,$BE,$F8,$DF,$89
	.dc.b	$F9,$9F,$EE,$2D,$8F,$0E,$0E,$2F,$C7,$7C,$DF,$81,$B3,$F8,$EF,$9C
	.dc.b	$FE,$E6,$47,$FB,$8B,$FD,$E7,$C7,$FC,$B7,$C9,$FF,$7E,$6C,$FF,$4E
	.dc.b	$F6,$FF,$8A,$71,$FC,$E3,$FD,$56,$FF,$6E,$71,$E4,$7F,$7F,$F3,$3F
	.dc.b	$C7,$F3,$9E,$2F,$F3,$47,$D1,$BF,$28,$7C,$A7,$F1,$CC,$7F,$D6,$D3
	.dc.b	$98,$C4,$23,$19,$7F,$99,$9B,$1F,$E2,$0F,$A4,$FF,$26,$FB,$9E,$6C
	.dc.b	$7F,$6D,$18,$CF,$E9,$F2,$7F,$8F,$F8,$CD,$26,$5D,$74,$7C,$9F,$8B
	.dc.b	$4F,$43,$63,$33,$B8,$E9,$C5,$23,$FC,$F7,$D3,$F8,$58,$C6,$DF,$EB
	.dc.b	$CC,$3D,$9F,$19,$FC,$17,$E5,$3F,$CD,$51,$8C,$3F,$82,$C6,$3E,$07
	.dc.b	$E3,$C7,$C1,$6D,$F3,$6B,$FE,$CB,$AF,$F5,$F7,$17,$F8,$CB,$E5,$3F
	.dc.b	$89,$1F,$E6,$BE,$EF,$F3,$25,$7F,$9E,$23,$FC,$7B,$E4,$7F,$17,$F6
	.dc.b	$EE,$5F,$C5,$9F,$16,$FF,$0F,$3F,$83,$87,$86,$30,$7F,$90,$3D,$AF
	.dc.b	$E4,$07,$A6,$6D,$FF,$8D,$B4,$FE,$11,$AF,$F8,$03,$EB,$7F,$DE,$B3
	.dc.b	$9E,$7F,$E5,$79,$7E,$3C,$4E,$9F,$C9,$5B,$EB,$5C,$5F,$5D,$B8,$7E
	.dc.b	$1D,$B5,$F4,$5A,$FF,$9B,$7D,$2F,$F3,$03,$F9,$A6,$E3,$FC,$FF,$C2
	.dc.b	$FE,$75,$F2,$7F,$E1,$6F,$35,$CD,$FC,$9C,$99,$7F,$89,$23,$2F,$FA
	.dc.b	$F3,$BD,$E0,$FF,$9C,$F3,$46,$27,$FD,$59,$39,$8C,$47,$E2,$BB,$3F
	.dc.b	$C1,$D9,$FF,$51,$F2,$3F,$86,$F3,$7F,$9D,$19,$5F,$D5,$DC,$B4,$FF
	.dc.b	$23,$FD,$5F,$FA,$87,$2F,$DB,$FB,$0F,$9A,$FF,$5A,$CE,$71,$FE,$BD
	.dc.b	$1F,$E8,$02,$5F,$2F,$7E,$E1,$F5,$7C,$5F,$93,$F2,$3F,$BC,$81,$FE
	.dc.b	$81,$E2,$FF,$74,$46,$63,$FD,$09,$D7,$8F,$03,$71,$BC,$7F,$DE,$31
	.dc.b	$8C,$3E,$AC,$6A,$74,$7F,$09,$8C,$BB,$C3,$FD,$ED,$3B,$9F,$2D,$76
	.dc.b	$FA,$CF,$F5,$46,$3E,$39,$CE,$1E,$89,$CD,$7A,$B8,$9B,$7F,$AF,$F0
	.dc.b	$F7,$46,$3D,$9F,$56,$1F,$72,$9F,$1D,$8A,$D6,$7E,$CA,$D5,$F9,$9E
	.dc.b	$D7,$FE,$CB,$9F,$F6,$03,$63,$33,$B7,$CD,$E3,$BE,$47,$FC,$E7,$A7
	.dc.b	$F9,$2D,$B1,$99,$F9,$02,$74,$E3,$E5,$1F,$51,$FC,$BE,$30,$D7,$E8
	.dc.b	$46,$BC,$0F,$F4,$0B,$63,$AD,$7F,$94,$F4,$F1,$FA,$DF,$9E,$9C,$B5
	.dc.b	$FB,$53,$B7,$FC,$69,$D6,$F8,$FF,$C1,$F0,$FC,$E8,$C4,$FC,$9C,$3F
	.dc.b	$57,$5A,$E2,$FA,$D5,$FE,$C4,$AD,$35,$5F,$3A,$E1,$1F,$B2,$9E,$AE
	.dc.b	$AB,$9F,$C7,$EB,$07,$CC,$FD,$CA,$F1,$2E,$1D,$36,$33,$C9,$83,$33
	.dc.b	$23,$24,$3C,$07,$F0,$49,$EC,$E2,$BC,$35,$DB,$E6,$BC,$99,$7F,$A2
	.dc.b	$6B,$FD,$86,$FF,$A0,$DB,$B7,$EB,$3A,$DC,$DF,$1B,$93,$0D,$AE,$31
	.dc.b	$87,$E2,$37,$2F,$E3,$3F,$07,$9F,$39,$FA,$3B,$E4,$75,$A7,$FA,$F7
	.dc.b	$1F,$1F,$C1,$EE,$B5,$1F,$E2,$86,$BF,$E0,$BB,$E7,$7B,$39,$FB,$FD
	.dc.b	$3D,$FF,$F6,$7E,$7F,$CA,$9B,$91,$A6,$3D,$3E,$18,$FC,$48,$F2,$C7
	.dc.b	$AF,$E5,$5E,$07,$A3,$DC,$97,$FE,$EC,$9C,$BF,$EE,$EF,$C5,$DF,$27
	.dc.b	$E2,$6C,$7F,$3D,$6B,$FE,$8D,$6B,$FC,$73,$34,$3E,$47,$E6,$36,$7B
	.dc.b	$47,$CA,$F7,$7E,$5F,$C6,$7C,$CF,$76,$57,$DE,$8F,$9F,$F1,$9F,$08
	.dc.b	$91,$FD,$3F,$83,$FD,$38,$FF,$84,$B9,$3F,$17,$D3,$F8,$28,$72,$FD
	.dc.b	$AE,$CF,$AC,$7D,$AF,$88,$FB,$DF,$8B,$FF,$5E,$AF,$F2,$FB,$5F,$B0
	.dc.b	$DB,$E9,$C1,$FD,$7B,$DD,$F6,$FC,$37,$C7,$25,$FC,$77,$EC,$FD,$97
	.dc.b	$E2,$3D,$89,$3F,$9A,$8F,$C5,$FA,$9F,$89,$F0,$E4,$FE,$7A,$3F,$1F
	.dc.b	$E9,$FE,$08,$91,$F0,$5F,$C3,$FF,$00,$7F,$0F,$D3,$CA,$DF,$13,$F9
	.dc.b	$E3,$6B,$FB,$9A,$F9,$CF,$D0,$2F,$8C,$A4,$9F,$EB,$B7,$D3,$CB,$FB
	.dc.b	$7F,$01,$F8,$52,$4F,$E1,$5F,$1F,$D4,$74,$24,$72,$CB,$C7,$F4,$3E
	.dc.b	$0F,$84,$F0,$FB,$F2,$B4,$79,$A3,$CB,$F0,$9E,$E2,$47,$F8,$3F,$54
	.dc.b	$7F,$40,$8F,$BF,$C9,$FE,$60,$7F,$D3,$2D,$7F,$64,$DD,$7E,$3F,$AF
	.dc.b	$FB,$F8,$FF,$16,$CF,$0F,$83,$C1,$F1,$BE,$6B,$F8,$37,$E2,$DF,$C4
	.dc.b	$67,$FD,$DC,$D7,$F4,$4D,$7F,$B6,$35,$FA,$3E,$77,$D0,$23,$F0,$3A
	.dc.b	$DA,$3E,$B7,$F8,$B4,$BE,$B1,$F2,$98,$F8,$A7,$C9,$53,$73,$AA,$F9
	.dc.b	$67,$1F,$8F,$F8,$CD,$3E,$44,$7F,$AB,$5B,$1A,$FA,$C7,$CF,$F4,$89
	.dc.b	$3D,$27,$2F,$A7,$37,$CC,$FE,$33,$DC,$93,$98,$FC,$7F,$CF,$FE,$41
	.dc.b	$91,$C6,$3F,$27,$F5,$1F,$94,$F8,$A9,$3D,$63,$F2,$FE,$5E,$A3,$07
	.dc.b	$85,$FC,$DF,$C0,$48,$F2,$3F,$9E,$F1,$FF,$9F,$7F,$00,$FF,$8B,$7B
	.dc.b	$3F,$4F,$19,$AF,$A5,$39,$F0,$7F,$C4,$8F,$F8,$36,$3F,$D8,$2E,$5F
	.dc.b	$F2,$EB,$79,$CE,$EF,$F8,$94,$FD,$9E,$CF,$A3,$C8,$69,$23,$E9,$FC
	.dc.b	$C7,$DA,$89,$1F,$6A,$53,$EA,$E5,$F5,$FB,$72,$BD,$31,$EE,$F9,$6D
	.dc.b	$46,$0F,$6B,$F8,$6F,$96,$FC,$6F,$C9,$7D,$89,$FC,$3B,$5F,$F4,$A6
	.dc.b	$3F,$A7,$DB,$FC,$B2,$BE,$5F,$D6,$49,$FC,$5B,$F8,$EF,$B8,$72,$FE
	.dc.b	$52,$24,$FE,$52,$72,$FE,$87,$C1,$7B,$3F,$AE,$F6,$A4,$E6,$3F,$63
	.dc.b	$ED,$7E,$C9,$91,$C6,$3F,$43,$E0,$FC,$E2,$47,$E7,$9F,$A3,$F2,$7A
	.dc.b	$8F,$99,$1F,$A8,$8F,$D7,$FD,$6F,$D8,$E6,$FB,$3F,$8A,$FE,$FD,$BF
	.dc.b	$15,$AE,$1D,$FC,$67,$EB,$2F,$FC,$5B,$97,$F9,$E5,$AF,$FA,$01,$AF
	.dc.b	$FA,$23,$F1,$C7,$A7,$CA,$3F,$6F,$3F,$E3,$6C,$7F,$13,$5A,$9C,$FC
	.dc.b	$37,$72,$3F,$C8,$69,$1E,$3F,$82,$F5,$B2,$7A,$C7,$F0,$CF,$8C,$FA
	.dc.b	$31,$FE,$68,$6B,$FE,$91,$6B,$FE,$98,$6B,$FE,$A0,$7F,$B7,$E7,$9E
	.dc.b	$6E,$87,$5F,$0F,$F5,$FF,$0D,$FE,$1C,$F0,$7F,$C0,$F8,$7E,$B2,$30
	.dc.b	$67,$E1,$77,$1F,$B2,$8F,$DB,$C7,$EE,$3E,$7E,$4F,$ED,$9F,$DD,$F9
	.dc.b	$F2,$3F,$72,$D7,$F6,$FE,$77,$F0,$0F,$3F,$C4,$48,$FF,$6C,$63,$FC
	.dc.b	$A9,$BE,$36,$DF,$E0,$73,$FC,$22,$3F,$84,$F8,$DE,$B2,$4F,$27,$DD
	.dc.b	$E1,$EA,$3F,$85,$C7,$F1,$08,$FE,$23,$E0,$C9,$FE,$1E,$FF,$14,$F4
	.dc.b	$E4,$60,$87,$F1,$07,$F8,$CF,$C3,$7C,$B2,$47,$A9,$F9,$DD,$3A,$8F
	.dc.b	$E3,$71,$FC,$82,$3F,$90,$F7,$7F,$22,$EE,$95,$FC,$82,$3F,$92,$FC
	.dc.b	$B7,$D7,$24,$7C,$67,$ED,$7C,$B6,$A3,$F9,$3C,$7F,$2C,$8F,$F3,$57
	.dc.b	$E2,$9F,$E8,$A9,$FE,$B8,$BD,$0C,$9A,$F4,$B7,$E8,$FF,$98,$24,$D2
	.dc.b	$57,$FA,$F3,$CF,$4F,$E6,$3C,$2F,$F1,$9B,$7B,$7F,$DB,$4D,$A5,$3F
	.dc.b	$80,$7F,$79,$5F,$ED,$F7,$F5,$CD,$3F,$D0,$4E,$A7,$8B,$A7,$FC,$97
	.dc.b	$97,$F9,$CD,$FF,$0B,$FC,$4F,$B7,$E3,$BF,$D1,$13,$99,$FF,$40,$7D
	.dc.b	$63,$E3,$FF,$D1,$8F,$ED,$A3,$FD,$AF,$68,$C4,$7F,$C0,$18,$C7,$8B
	.dc.b	$FE,$1D,$6C,$7C,$3D,$3F,$D6,$4B,$5F,$87,$E1,$7D,$23,$E9,$FA,$DF
	.dc.b	$E0,$AF,$96,$FF,$13,$CA,$BF,$C8,$5E,$1D,$53,$FE,$61,$F8,$6F,$F3
	.dc.b	$C4,$63,$E0,$B5,$18,$9E,$73,$C8,$E8,$E2,$35,$3C,$2B,$B4,$F5,$6F
	.dc.b	$C0,$FA,$6F,$91,$E2,$71,$C9,$FE,$57,$8E,$B1,$8C,$38,$E3,$EB,$AD
	.dc.b	$57,$18,$C4,$FB,$63,$E0,$7C,$AE,$1C,$35,$FF,$79,$9F,$88,$8E,$0F
	.dc.b	$FB,$B0,$C1,$C0,$F3,$8F,$98,$FF,$48,$47,$C8,$8E,$E6,$A3,$38,$F4
	.dc.b	$AF,$A5,$5B,$9C,$D7,$58,$C7,$36,$AF,$F3,$BE,$E7,$FA,$77,$1F,$F4
	.dc.b	$9F,$A7,$F6,$EB,$F5,$F3,$97,$E2,$46,$70,$FF,$69,$7C,$97,$F9,$A2
	.dc.b	$7F,$22,$D8,$FC,$83,$5F,$C2,$71,$A3,$11,$F8,$98,$FD,$E7,$37,$78
	.dc.b	$FD,$2B,$63,$87,$C1,$DE,$8F,$D1,$61,$FE,$5F,$EA,$7C,$B7,$1E,$2F
	.dc.b	$DB,$7A,$16,$73,$5B,$8D,$6B,$F3,$E3,$FC,$D0,$DD,$3F,$45,$1F,$BD
	.dc.b	$9D,$F1,$70,$F7,$BF,$C6,$75,$F1,$1A,$FF,$06,$BF,$F2,$FD,$FF,$0F
	.dc.b	$6C,$7E,$27,$14,$7F,$BB,$27,$C3,$EF,$79,$3D,$FF,$2B,$E9,$F6,$3D
	.dc.b	$5F,$2D,$6B,$E0,$38,$7F,$8E,$6A,$8F,$1E,$3F,$C5,$DB,$19,$9D,$FC
	.dc.b	$B7,$C5,$F9,$6F,$8C,$D8,$F9,$9E,$1F,$C9,$E5,$F9,$4D,$8F,$A1,$1D
	.dc.b	$08,$F8,$9E,$FE,$17,$D1,$F6,$3E,$93,$67,$EC,$45,$E0,$F9,$E7,$F2
	.dc.b	$69,$C7,$57,$FB,$9B,$3F,$F1,$DD,$FD,$FF,$7B,$EE,$7B,$DF,$75,$91
	.dc.b	$F0,$9F,$79,$1F,$8B,$C7,$F8,$3F,$94,$FC,$23,$63,$F1,$51,$F1,$88
	.dc.b	$FC,$C7,$E4,$FC,$9F,$E2,$CF,$C6,$7C,$95,$9F,$C9,$7C,$9F,$E3,$CC
	.dc.b	$6B,$EC,$E4,$FE,$91,$1F,$93,$F9,$3F,$CA,$7C,$9D,$BF,$CD,$38,$7E
	.dc.b	$DB,$F6,$08,$FF,$B5,$7F,$10,$7C,$77,$E9,$F1,$47,$FA,$FE,$7F,$D6
	.dc.b	$EE,$9B,$F1,$DB,$8D,$4F,$F1,$46,$BF,$C7,$5A,$FF,$3B,$6B,$FC,$FF
	.dc.b	$E9,$DE,$AF,$85,$7A,$D6,$30,$CC,$3F,$DD,$93,$A6,$C7,$EA,$29,$39
	.dc.b	$7F,$93,$FD,$43,$E0,$FD,$DF,$4A,$ED,$F8,$13,$9C,$7F,$DA,$B3,$96
	.dc.b	$CF,$58,$C5,$FF,$32,$6B,$FC,$ED,$AF,$F5,$95,$DD,$F3,$7E,$1B,$F5
	.dc.b	$13,$F9,$61,$3A,$BF,$F7,$2C,$EB,$AF,$FD,$6F,$3F,$E6,$A9,$F2,$F6
	.dc.b	$FA,$BB,$7D,$6C,$8F,$E8,$B8,$74,$8F,$41,$23,$FC,$31,$F5,$CF,$1F
	.dc.b	$F8,$F2,$F0,$EC,$61,$93,$FD,$D7,$23,$FD,$76,$BC,$CD,$9E,$7E,$1F
	.dc.b	$CC,$F8,$5F,$9F,$DD,$6B,$E6,$B8,$7C,$FA,$A3,$E7,$12,$33,$DD,$F9
	.dc.b	$5F,$8F,$7C,$C7,$DA,$8C,$3F,$CD,$FD,$BF,$F1,$6D,$79,$CF,$EE,$4B
	.dc.b	$D3,$D7,$83,$B6,$4E,$F8,$BF,$C6,$B5,$DA,$31,$5F,$E4,$2A,$ED,$7F
	.dc.b	$8A,$F0,$BF,$17,$E1,$5B,$FC,$4B,$87,$DA,$FE,$24,$8F,$F9,$27,$07
	.dc.b	$CC,$7A,$7E,$63,$D4,$D9,$F0,$CA,$7D,$DF,$31,$9D,$3F,$82,$4A,$3E
	.dc.b	$17,$CC,$4B,$E3,$16,$B7,$ED,$FC,$7F,$6B,$E4,$36,$BB,$C7,$CB,$F9
	.dc.b	$8C,$57,$67,$E7,$7A,$DF,$3F,$C8,$B7,$E6,$70,$FA,$3C,$8B,$7E,$67
	.dc.b	$F8,$77,$07,$F9,$B2,$7F,$5F,$E4,$E3,$6F,$DA,$B2,$3F,$0A,$FE,$C4
	.dc.b	$8F,$E7,$BD,$5F,$BF,$87,$EA,$FC,$97,$C6,$FF,$1B,$9C,$F8,$1F,$98
	.dc.b	$AF,$F2,$75,$6A,$FE,$6F,$CB,$7E,$24,$FC,$57,$C1,$D8,$FC,$0F,$84
	.dc.b	$F4,$3F,$B1,$B6,$3F,$57,$18,$3F,$BC,$FB,$4F,$C6,$3F,$92,$E7,$E1
	.dc.b	$8C,$70,$FE,$5F,$DB,$FF,$0A,$56,$B0,$FF,$0D,$2C,$B2,$7E,$77,$86
	.dc.b	$CF,$E1,$3C,$AF,$EE,$B0,$69,$23,$F1,$DF,$05,$FA,$5F,$82,$FD,$33
	.dc.b	$63,$F1,$BE,$FF,$EA,$4F,$F1,$07,$F0,$0F,$F8,$D7,$6F,$91,$E2,$7F
	.dc.b	$A8,$A7,$FC,$49,$F0,$BF,$AB,$BF,$F5,$13,$79,$7F,$C2,$8D,$7F,$CF
	.dc.b	$55,$FE,$8B,$F1,$7F,$C1,$78,$FF,$89,$71,$EA,$DC,$3F,$CD,$0F,$F8
	.dc.b	$7A,$33,$F0,$1E,$4F,$7E,$57,$ED,$29,$EF,$E3,$2F,$EC,$D0,$F9,$1E
	.dc.b	$E4,$AF,$F5,$C4,$65,$BF,$07,$FE,$45,$FC,$5B,$FA,$15,$F2,$8C,$7B
	.dc.b	$3F,$10,$F8,$BD,$F6,$FC,$4F,$C4,$F1,$BF,$C4,$FC,$26,$92,$7F,$D5
	.dc.b	$BF,$79,$ED,$7F,$CB,$0D,$C7,$FD,$8B,$9B,$D2,$FF,$2C,$4E,$7B,$5A
	.dc.b	$E7,$0F,$F5,$FB,$70,$F8,$3F,$25,$EB,$C3,$ED,$FC,$97,$D0,$25,$FC
	.dc.b	$27,$E9,$FC,$97,$E0,$3E,$4B,$F0,$2D,$9F,$E4,$3B,$7E,$57,$1F,$F6
	.dc.b	$CB,$7A,$1B,$3E,$4B,$F0,$BE,$17,$E1,$BC,$5F,$86,$4C,$7F,$0E,$FF
	.dc.b	$93,$3E,$F7,$FD,$E5,$1F,$6F,$E4,$FF,$CE,$11,$8C,$FE,$DC,$7E,$53
	.dc.b	$E4,$FF,$2B,$F2,$76,$FF,$04,$E1,$FB,$AF,$E3,$88,$FF,$B7,$38,$D5
	.dc.b	$F7,$CF,$CF,$F9,$56,$7E,$5C,$7E,$8B,$E1,$71,$8F,$E3,$1F,$D2,$FC
	.dc.b	$9C,$AF,$C9,$7C,$9D,$AF,$C8,$CF,$FA,$27,$F0,$5F,$EC,$F5,$F9,$78
	.dc.b	$C3,$FA,$AF,$1F,$F8,$A7,$C9,$7F,$42,$BF,$9B,$18,$F6,$BF,$D9,$5C
	.dc.b	$CF,$23,$15,$AC,$3F,$D9,$35,$AB,$CF,$3F,$FA,$96,$AE,$09,$CF,$C9
	.dc.b	$7F,$AC,$72,$F0,$5F,$FB,$FE,$73,$7C,$1B,$8F,$68,$4B,$47,$CC,$78
	.dc.b	$1F,$F5,$2C,$B0,$EB,$1A,$BE,$57,$C7,$D3,$E0,$48,$CC,$EE,$F9,$69
	.dc.b	$FE,$EB,$8E,$B8,$7C,$79,$7D,$17,$F8,$6F,$CA,$FF,$10,$B6,$3F,$DC
	.dc.b	$ED,$7F,$DE,$0D,$7F,$E0,$0F,$B2,$FA,$3F,$F1,$4F,$CB,$23,$13,$FE
	.dc.b	$F0,$8C,$4E,$A7,$33,$D3,$E1,$FE,$2B,$E0,$6B,$9D,$BC,$B1,$88,$1F
	.dc.b	$33,$AF,$E5,$BF,$19,$AF,$C2,$35,$C1,$C1,$C5,$27,$FD,$B3,$1C,$67
	.dc.b	$5A,$7D,$E9,$E6,$D8,$CE,$DE,$2B,$E5,$C1,$8C,$3E,$FF,$A9,$F8,$97
	.dc.b	$EC,$B5,$F0,$B7,$0F,$C5,$AF,$4B,$95,$FF,$5B,$6C,$6F,$0F,$03,$6F
	.dc.b	$BC,$E8,$DF,$1C,$8F,$0C,$74,$CF,$FC,$B1,$1F,$E3,$6E,$2F,$F2,$E3
	.dc.b	$7E,$23,$E3,$7A,$5F,$D6,$AB,$73,$FE,$AB,$6C,$67,$7D,$F2,$FC,$39
	.dc.b	$CB,$6F,$37,$F5,$FE,$32,$ED,$FA,$3C,$2E,$5C,$2B,$47,$F9,$2B,$1F
	.dc.b	$F4,$83,$6F,$8C,$62,$3F,$DB,$3E,$5B,$E8,$BF,$D6,$B1,$ED,$F8,$AF
	.dc.b	$F3,$EE,$5F,$B1,$6E,$3F,$EF,$79,$FC,$21,$FA,$78,$DE,$0F,$96,$F2
	.dc.b	$C4,$9F,$24,$78,$BE,$37,$D3,$9C,$9F,$44,$7B,$BB,$1E,$BF,$89,$B3
	.dc.b	$E0,$3F,$82,$B8,$7C,$79,$A3,$11,$FE,$04,$FA,$3F,$EE,$CF,$F9,$B9
	.dc.b	$AF,$FB,$7D,$BC,$5F,$E1,$25,$EB,$7C,$8F,$7B,$E7,$FC,$7C,$BF,$9D
	.dc.b	$1F,$4B,$5F,$8A,$48,$F3,$65,$FC,$66,$3E,$4F,$CB,$FD,$82,$4F,$FA
	.dc.b	$0F,$8F,$E9,$73,$FD,$CD,$E4,$FE,$3D,$FC,$1C,$BE,$8F,$FC,$01,$3F
	.dc.b	$E8,$8D,$FC,$EF,$81,$AF,$C4,$6B,$FE,$26,$6B,$F9,$1F,$8B,$FC,$5F
	.dc.b	$C8,$C8,$FC,$54,$78,$39,$FF,$A2,$47,$D1,$C3,$F1,$1E,$3F,$FB,$09
	.dc.b	$B1,$99,$DB,$F5,$3A,$1F,$2F,$F8,$A2,$47,$C6,$8C,$B7,$87,$F3,$A4
	.dc.b	$7F,$D4,$BF,$45,$FC,$7B,$9B,$FC,$8F,$4F,$98,$FD,$19,$2B,$7F,$21
	.dc.b	$F8,$BF,$73,$F1,$44,$8F,$8C,$FE,$73,$E4,$FF,$56,$48,$EF,$F2,$3F
	.dc.b	$B0,$CB,$E7,$C3,$2F,$0F,$C9,$7E,$C8,$D7,$C0,$FE,$D3,$E3,$7F,$4E
	.dc.b	$4A,$F7,$3F,$AA,$E3,$FD,$C9,$23,$F0,$5F,$00,$F8,$0F,$DE,$FC,$C4
	.dc.b	$9F,$E2,$0F,$E0,$A7,$FC,$1B,$1F,$81,$F6,$3F,$4C,$47,$F0,$4F,$EE
	.dc.b	$3E,$3B,$F8,$39,$23,$E5,$FC,$7F,$F0,$E5,$ED,$B8,$B1,$FC,$03,$FE
	.dc.b	$DB,$9F,$F6,$C0,$70,$FF,$A8,$E3,$97,$D1,$FF,$A9,$23,$F6,$C1,$C3
	.dc.b	$FB,$96,$BF,$C2,$9A,$FF,$4F,$7F,$CA,$8F,$81,$E3,$C3,$FE,$AD,$8E
	.dc.b	$9F,$11,$E0,$F8,$7F,$AD,$E9,$7F,$26,$FC,$1B,$F0,$DE,$97,$F4,$F6
	.dc.b	$DF,$F8,$33,$9F,$FD,$40,$FE,$32,$B8,$B4,$FC,$7F,$C2,$73,$F8,$D9
	.dc.b	$1F,$EF,$6B,$F8,$39,$FF,$A6,$30,$FF,$52,$4F,$E5,$BE,$3B,$F1,$04
	.dc.b	$7F,$33,$F1,$FE,$5E,$E9,$1C,$DF,$C7,$7A,$F2,$3F,$3E,$D7,$F2,$1A
	.dc.b	$FF,$69,$C3,$F5,$B8,$FF,$AA,$3E,$7F,$F9,$9B,$E3,$F1,$FF,$30,$7F
	.dc.b	$48,$8C,$D7,$CD,$E3,$FE,$97,$E0,$FD,$FC,$7F,$C0,$B8,$7F,$AC,$A3
	.dc.b	$11,$E0,$3F,$C1,$3C,$98,$9F,$A4,$BA,$BE,$FC,$FD,$F6,$C6,$63,$F0
	.dc.b	$6D,$7F,$9B,$E3,$FC,$E6,$BF,$D8,$3C,$1F,$E2,$F6,$FB,$3E,$9F,$53
	.dc.b	$D5,$F3,$D7,$0F,$3E,$BF,$E0,$BC,$79,$36,$7F,$39,$F3,$2E,$6F,$CF
	.dc.b	$7E,$39,$F8,$9C,$FF,$14,$DB,$FF,$0F,$56,$F1,$FC,$87,$83,$EA,$23
	.dc.b	$EA,$7D,$BF,$1B,$FA,$06,$47,$E5,$E3,$FA,$D7,$C3,$C8,$FE,$6C,$FF
	.dc.b	$7A,$FA,$4F,$F7,$F4,$7A,$8F,$F2,$FB,$87,$E8,$7A,$E8,$C6,$7F,$EE
	.dc.b	$66,$E5,$FE,$2B,$7F,$94,$D7,$D7,$3F,$61,$F1,$9F,$BD,$FA,$6F,$EB
	.dc.b	$B5,$FE,$3B,$6C,$FD,$C9,$FD,$CF,$87,$F6,$09,$1F,$61,$FB,$BF,$4B
	.dc.b	$FE,$0F,$87,$E3,$9F,$AC,$6E,$7F,$C2,$E7,$5F,$4B,$FB,$79,$FF,$03
	.dc.b	$B5,$FF,$26,$35,$FF,$2A,$B5,$3F,$E5,$4D,$3F,$80,$3F,$5E,$37,$9F
	.dc.b	$F1,$98,$D3,$D3,$DB,$FE,$39,$F7,$BF,$B4,$AF,$88,$DB,$E7,$3C,$7D
	.dc.b	$BF,$CC,$E0,$69,$2F,$F8,$DF,$C2,$BC,$4F,$E5,$1F,$8E,$6E,$7F,$92
	.dc.b	$B5,$FA,$53,$B7,$F9,$4B,$E3,$7F,$C5,$2D,$7F,$CD,$CD,$7F,$CE,$ED
	.dc.b	$D5,$DA,$F5,$3F,$C6,$AD,$AF,$F0,$7B,$5F,$F1,$1B,$5F,$F3,$D4,$BB
	.dc.b	$3F,$02,$F4,$AF,$E5,$07,$F0,$7F,$99,$FE,$55,$F4,$12,$FF,$90,$3F
	.dc.b	$C2,$FE,$6A,$47,$E6,$6F,$A4,$62,$3F,$33,$87,$F3,$8E,$2F,$E6,$5F
	.dc.b	$43,$E8,$24,$FA,$1F,$F2,$7C,$FF,$0D,$9D,$CF,$F2,$F9,$FE,$35,$E3
	.dc.b	$7F,$3D,$F5,$FE,$09,$27,$E0,$BC,$BE,$6B,$51,$FC,$DE,$3F,$91,$47
	.dc.b	$F4,$5F,$A1,$93,$FC,$ED,$FE,$4D,$DF,$23,$04,$3D,$6F,$F4,$DF,$90
	.dc.b	$FA,$24,$8F,$A2,$FA,$BE,$33,$51,$FD,$1E,$3F,$99,$47,$F3,$3F,$97
	.dc.b	$FE,$AD,$F4,$32,$BE,$56,$C9,$FE,$71,$DF,$F1,$09,$3F,$E4,$C9,$FE
	.dc.b	$EE,$DF,$0F,$6F,$29,$FE,$81,$1F,$E8,$8F,$A7,$7D,$0F,$FA,$2B,$DC
	.dc.b	$7C,$65,$35,$1B,$F6,$3F,$D0,$35,$FD,$9B,$1F,$F3,$6B,$5F,$C5,$B5
	.dc.b	$FF,$5E,$3F,$9F,$7F,$8C,$B7,$C2,$73,$6B,$BF,$C9,$EB,$FF,$A7,$3E
	.dc.b	$9F,$F7,$9D,$6C,$3F,$6D,$7D,$2B,$4F,$F7,$DF,$76,$65,$FD,$CD,$FF
	.dc.b	$32,$E9,$FE,$91,$8F,$F3,$67,$CD,$57,$FD,$2F,$F1,$28,$C4,$7F,$B7
	.dc.b	$63,$1F,$1D,$FE,$5B,$9C,$CF,$B3,$E3,$BF,$CE,$30,$F8,$CA,$7C,$D8
	.dc.b	$FF,$15,$7C,$C3,$B7,$FC,$8D,$6F,$80,$FF,$39,$CF,$F9,$29,$BE,$07
	.dc.b	$FB,$B5,$CD,$70,$6F,$27,$FB,$E1,$DD,$DF,$90,$FF,$7A,$61,$CD,$B7
	.dc.b	$C5,$B7,$FE,$EF,$BB,$6E,$3F,$F1,$CE,$1E,$7A,$D4,$84,$65,$E1,$5F
	.dc.b	$EB,$1F,$95,$DC,$7F,$C0,$03,$27,$FB,$11,$CB,$CD,$A9,$F5,$3D,$C7
	.dc.b	$CD,$7E,$19,$F9,$18,$F7,$69,$B8,$D5,$7F,$A6,$AB,$9C,$FA,$67,$35
	.dc.b	$C6,$BF,$D4,$31,$89,$F8,$5F,$19,$EF,$E4,$FF,$59,$7C,$8B,$6F,$16
	.dc.b	$9E,$3D,$7F,$D1,$D8,$62,$FE,$37,$1F,$FB,$3E,$F1,$28,$FF,$69,$78
	.dc.b	$7F,$D7,$DB,$5F,$EA,$9C,$53,$96,$C7,$F3,$AF,$9E,$E7,$8F,$F4,$07
	.dc.b	$AD,$FE,$EB,$D2,$CF,$9C,$9D,$F6,$3C,$FE,$3B,$FC,$CC,$BF,$9B,$4E
	.dc.b	$5F,$E7,$11,$9D,$7F,$C1,$EB,$DD,$E2,$72,$E1,$B5,$C3,$DE,$FE,$F5
	.dc.b	$B7,$FA,$C5,$B7,$FE,$F8,$9F,$F7,$0B,$70,$FF,$7A,$AF,$A1,$F6,$F8
	.dc.b	$9F,$DB,$EB,$C7,$18,$F5,$3D,$7F,$2F,$EC,$F9,$7B,$5E,$A7,$0F,$FB
	.dc.b	$BB,$64,$79,$1B,$19,$8C,$78,$FF,$23,$E0,$7E,$3B,$67,$E2,$F2,$7C
	.dc.b	$CF,$23,$E5,$B6,$3F,$D0,$11,$F0,$39,$4D,$24,$78,$FC,$1F,$AB,$E2
	.dc.b	$7D,$46,$C7,$FA,$26,$2F,$0F,$F8,$FB,$4F,$E1,$93,$93,$EB,$FC,$9F
	.dc.b	$DE,$F2,$7E,$EB,$63,$FD,$25,$1E,$32,$3F,$E0,$27,$E1,$7C,$37,$E1
	.dc.b	$BC,$9F,$C2,$B6,$7F,$D3,$51,$EB,$23,$FE,$0A,$7E,$57,$C8,$7E,$30
	.dc.b	$FC,$6F,$C9,$59,$FF,$51,$79,$1F,$92,$31,$9F,$F9,$63,$74,$7E,$53
	.dc.b	$E4,$3F,$2F,$F1,$96,$FF,$D5,$4E,$1F,$F0,$CB,$F3,$48,$FF,$B2,$BE
	.dc.b	$9F,$F3,$CD,$F0,$7F,$C9,$11,$8C,$7F,$85,$1F,$C6,$5C,$38,$7F,$D0
	.dc.b	$53,$FE,$DA,$FA,$67,$C2,$7E,$E2,$95,$A7,$FC,$D5,$3A,$AF,$F7,$0F
	.dc.b	$D3,$FF,$75,$7F,$DB,$AD,$9F,$08,$70,$F9,$1A,$FF,$1F,$5E,$9F,$87
	.dc.b	$ED,$5D,$BE,$39,$CD,$FF,$BC,$27,$2F,$FB,$CB,$EA,$3F,$3B,$87,$F9
	.dc.b	$FD,$FE,$F7,$19,$EE,$F6,$46,$27,$5D,$BF,$E6,$F8,$C4,$7C,$86,$BF
	.dc.b	$CC,$DA,$FF,$43,$E5,$F0,$BF,$E8,$DF,$C5,$7F,$85,$3F,$E8,$F8,$CF
	.dc.b	$9B,$FE,$F5,$9C,$BF,$EF,$7A,$FC,$ED,$7C,$66,$BE,$1E,$DF,$21,$E5
	.dc.b	$F9,$AF,$A9,$87,$BA,$F9,$46,$32,$FE,$4B,$F5,$3E,$0D,$3F,$AC,$79
	.dc.b	$08,$C6,$5F,$D7,$D7,$CF,$F0,$E1,$FE,$23,$F3,$53,$0F,$F2,$D5,$70
	.dc.b	$C1,$E4,$2B,$E5,$F7,$FC,$CE,$EB,$87,$D3,$70,$FE,$0B,$37,$0A,$7E
	.dc.b	$81,$27,$5F,$20,$F9,$7F,$C9,$74,$BE,$13,$F2,$93,$9A,$FC,$C4,$7F
	.dc.b	$44,$F0,$FF,$C1,$91,$FE,$71,$75,$97,$FB,$21,$CB,$70,$E0,$7F,$48
	.dc.b	$F9,$6B,$5F,$E7,$9F,$0D,$1F,$88,$F1,$3F,$0C,$49,$F6,$6A,$8F,$C5
	.dc.b	$6F,$FD,$5B,$EB,$5E,$57,$E4,$23,$0F,$9F,$E2,$FD,$1F,$15,$71,$F4
	.dc.b	$3E,$BF,$89,$91,$83,$D8,$FF,$62,$E3,$FE,$ED,$BF,$FB,$91,$BD,$9F
	.dc.b	$DD,$78,$55,$BA,$F8,$DE,$C4,$C3,$E3,$7B,$D6,$79,$7A,$BD,$4C,$63
	.dc.b	$FD,$FF,$2F,$EC,$8F,$5F,$27,$E8,$1F,$47,$E0,$6E,$3F,$46,$3C,$DE
	.dc.b	$BE,$31,$F3,$3F,$B2,$F3,$7F,$87,$7D,$0F,$F8,$C6,$FF,$75,$18,$EF
	.dc.b	$91,$F8,$99,$D5,$FF,$0C,$6B,$FC,$79,$AF,$F8,$A7,$E1,$1F,$37,$F8
	.dc.b	$AF,$8B,$FE,$39,$18,$BF,$E4,$11,$F2,$FD,$B9,$1F,$89,$F6,$2E,$0D
	.dc.b	$FF,$C4,$1F,$4E,$F9,$8F,$EF,$ED,$BF,$E3,$FA,$A3,$17,$FC,$CA,$31
	.dc.b	$5F,$E0,$6E,$2F,$F0,$05,$FF,$B1,$DB,$5D,$E3,$FC,$57,$F5,$5F,$CE
	.dc.b	$F1,$FA,$5E,$EF,$E0,$7E,$42,$57,$F9,$56,$7F,$C8,$DF,$02,$BF,$EE
	.dc.b	$73,$D6,$BF,$11,$F0,$BE,$12,$47,$A3,$24,$7F,$86,$BB,$1E,$D7,$E3
	.dc.b	$FE,$43,$F5,$CD,$BF,$47,$2F,$E4,$CF,$D9,$78,$D6,$3F,$DC,$5C,$5F
	.dc.b	$E8,$17,$E2,$B7,$83,$E2,$35,$FF,$79,$B5,$FE,$33,$9B,$E7,$7F,$40
	.dc.b	$7F,$90,$BE,$A7,$F2,$77,$A8,$FD,$27,$97,$23,$FD,$E5,$3F,$E2,$FE
	.dc.b	$2F,$CD,$4F,$EA,$7E,$77,$E9,$12,$BE,$13,$FE,$55,$FA,$96,$F4,$93
	.dc.b	$FD,$73,$91,$3A,$8F,$C0,$E3,$F8,$DD,$3F,$69,$D5,$E7,$E0,$FD,$C3
	.dc.b	$FE,$5A,$DF,$FB,$9C,$FF,$36,$6B,$FE,$93,$6F,$47,$E2,$FB,$FF,$11
	.dc.b	$F3,$9F,$C0,$4F,$F3,$0F,$D5,$7F,$95,$73,$FC,$54,$7F,$05,$F9,$69
	.dc.b	$1F,$85,$3F,$CE,$BA,$FC,$B7,$F2,$11,$FC,$2F,$D4,$FC,$A1,$23,$F0
	.dc.b	$8F,$F9,$73,$65,$FF,$11,$F9,$6F,$A4,$DA,$ED,$AF,$FA,$AF,$E4,$7F
	.dc.b	$C1,$1C,$BF,$E8,$76,$CF,$F5,$47,$F3,$8D,$F0,$3E,$27,$C4,$7E,$B0
	.dc.b	$91,$F8,$87,$C5,$9F,$F2,$1F,$85,$FC,$13,$67,$EF,$C7,$F2,$5F,$13
	.dc.b	$F0,$CD,$8F,$D7,$47,$EC,$48,$FE,$85,$F7,$7C,$AF,$E3,$4F,$F4,$6F
	.dc.b	$07,$F6,$09,$FC,$5C,$65,$FD,$C3,$74,$FE,$64,$FF,$A5,$BE,$8F,$FD
	.dc.b	$FB,$3F,$13,$91,$18,$AF,$DF,$1F,$ED,$1E,$1F,$F0,$EB,$89,$DD,$73
	.dc.b	$3F,$A0,$7C,$35,$7F,$84,$3F,$EA,$4D,$FF,$6F,$1F,$EC,$DF,$1B,$FC
	.dc.b	$03,$F1,$FF,$D1,$BE,$4B,$F8,$C6,$3F,$57,$C3,$FF,$66,$46,$32,$56
	.dc.b	$EF,$EB,$56,$9E,$77,$4E,$58,$77,$7F,$DD,$55,$C5,$A6,$EF,$AB,$71
	.dc.b	$F3,$BF,$EE,$1B,$EB,$7C,$EB,$E3,$F9,$5C,$75,$37,$E9,$F0,$CA,$E1
	.dc.b	$3A,$6C,$E7,$0E,$C7,$FB,$EF,$EA,$7F,$D8,$9E,$87,$F8,$55,$B1,$FE
	.dc.b	$58,$6B,$FE,$6A,$6B,$FE,$F5,$F9,$5F,$F6,$BF,$D4,$FF,$87,$5B,$1F
	.dc.b	$E6,$06,$BF,$E7,$26,$BF,$EF,$D9,$FF,$18,$E7,$E1,$F8,$43,$49,$3F
	.dc.b	$D2,$7C,$97,$D5,$23,$E3,$7C,$13,$5C,$47,$BA,$D3,$98,$F9,$97,$C9
	.dc.b	$B7,$C3,$93,$F0,$55,$B9,$E4,$D8,$FC,$35,$71,$9E,$DE,$03,$7E,$5C
	.dc.b	$5F,$E5,$08,$C4,$FD,$B6,$BC,$38,$B6,$FC,$08,$CE,$1F,$93,$6D,$73
	.dc.b	$BD,$E6,$9F,$C1,$DF,$5C,$33,$1C,$23,$A5,$62,$78,$E5,$E2,$EF,$F8
	.dc.b	$98,$7C,$69,$F9,$31,$89,$EB,$7F,$B3,$78,$6B,$FE,$6C,$6C,$7D,$86
	.dc.b	$DF,$D0,$C3,$9D,$6F,$E5,$6D,$BC,$87,$AD,$FE,$F6,$8F,$E0,$D8,$FF
	.dc.b	$9D,$1B,$C3,$C2,$B7,$F0,$F6,$F8,$57,$F0,$A9,$CB,$FC,$C9,$AF,$A7
	.dc.b	$E1,$FC,$3A,$7E,$12,$FF,$CE,$8D,$9F,$F0,$CE,$3D,$1F,$C5,$C6,$5F
	.dc.b	$F0,$D4,$3E,$33,$9B,$E2,$DB,$E4,$5E,$27,$BF,$4F,$FA,$EA,$FD,$BF
	.dc.b	$31,$C0,$CE,$5E,$4E,$24,$79,$1B,$3F,$E2,$C6,$BF,$E2,$B6,$BE,$67
	.dc.b	$FC,$DE,$FF,$2F,$6B,$FC,$AD,$B8,$FD,$DF,$80,$F4,$FC,$AF,$AB,$E3
	.dc.b	$65,$FD,$28,$F6,$FC,$67,$C6,$D8,$D2,$4F,$F8,$0F,$85,$F2,$7F,$54
	.dc.b	$91,$F7,$1F,$B5,$F5,$0F,$8F,$A4,$9F,$C0,$FC,$BF,$C3,$3E,$11,$1F
	.dc.b	$84,$F8,$FE,$3F,$E8,$F7,$49,$FC,$4C,$7D,$4F,$A8,$FC,$33,$23,$F1
	.dc.b	$B1,$F6,$3E,$33,$D8,$48,$FC,$0B,$EF,$F8,$BF,$BA,$48,$FC,$AF,$C2
	.dc.b	$FE,$2B,$E4,$BF,$8B,$C7,$F5,$D6,$BF,$D6,$5B,$C9,$FA,$E5,$F3,$5F
	.dc.b	$7E,$BE,$F1,$2B,$EF,$3F,$94,$9F,$F6,$14,$75,$F9,$1F,$01,$E1,$F1
	.dc.b	$3C,$5F,$1D,$E3,$F8,$E9,$5E,$38,$F3,$7C,$77,$32,$47,$F9,$DB,$0F
	.dc.b	$F5,$2B,$FE,$3B,$FC,$5D,$F1,$DF,$AF,$6C,$7E,$AD,$AF,$B7,$E6,$F5
	.dc.b	$F3,$D2,$34,$FC,$5A,$FD,$7F,$D4,$70,$9F,$D5,$B5,$FF,$7A,$B5,$FE
	.dc.b	$82,$DF,$19,$8B,$E4,$BE,$81,$2B,$E8,$61,$FC,$31,$7D,$43,$E3,$FE
	.dc.b	$B7,$74,$9F,$F1,$A9,$A8,$E1,$3F,$E8,$08,$FF,$7B,$35,$FE,$96,$7F
	.dc.b	$4F,$70,$E1,$AF,$F5,$3C,$7C,$FF,$0B,$D3,$E3,$E4,$FE,$16,$3F,$11
	.dc.b	$F0,$7F,$89,$F9,$E9,$3F,$89,$8F,$C6,$7C,$26,$A3,$07,$E3,$5F,$76
	.dc.b	$AE,$BF,$C9,$7C,$84,$AF,$6C,$E5,$FC,$A7,$85,$C2,$A4,$60,$FC,$C7
	.dc.b	$CE,$FE,$6C,$8F,$E6,$FD,$FF,$CF,$7C,$8F,$E7,$D9,$1F,$9F,$8F,$E5
	.dc.b	$7F,$53,$F1,$BC,$1F,$CF,$47,$09,$FF,$20,$4E,$7D,$7F,$EE,$BC,$7F
	.dc.b	$E8,$C9,$F9,$FF,$1F,$F3,$9C,$B5,$FA,$0F,$F3,$6A,$7C,$07,$D8,$89
	.dc.b	$3D,$A3,$C7,$F2,$DE,$4F,$96,$93,$E2,$8F,$3F,$CB,$6A,$30,$79,$9F
	.dc.b	$C1,$FC,$97,$E2,$FE,$2D,$F2,$94,$CD,$7E,$2B,$4F,$F5,$A3,$F8,$8F
	.dc.b	$95,$F4,$3F,$92,$6C,$6A,$3D,$DF,$4E,$CB,$F2,$35,$27,$F2,$B3,$97
	.dc.b	$FC,$D1,$E1,$3A,$3F,$57,$CF,$27,$31,$FA,$EE,$6F,$D7,$B2,$38,$C7
	.dc.b	$E7,$BE,$5B,$C6,$48,$FC,$FB,$E6,$F9,$6D,$47,$B6,$3F,$4B,$1F,$A6
	.dc.b	$F9,$3F,$D7,$7C,$DB,$BA,$99,$BF,$89,$F2,$BF,$00,$8F,$EB,$1F,$8B
	.dc.b	$F5,$BF,$D0,$63,$64,$3F,$D0,$61,$D3,$8E,$17,$1F,$CC,$24,$7E,$E1
	.dc.b	$F9,$FF,$27,$FB,$C6,$47,$D7,$8F,$E0,$DF,$4A,$F7,$BE,$BB,$67,$F2
	.dc.b	$0D,$7F,$28,$D7,$F8,$1A,$F7,$F9,$CC,$9F,$E1,$B3,$97,$F4,$3E,$4F
	.dc.b	$E6,$48,$FE,$7F,$C9,$77,$FF,$17,$8C,$3F,$C4,$7C,$2F,$F1,$AB,$98
	.dc.b	$FF,$52,$36,$33,$E4,$FE,$CB,$C1,$95,$FB,$07,$F6,$BD,$D2,$3F,$66
	.dc.b	$D7,$9F,$67,$EE,$BB,$BF,$5C,$48,$F1,$BF,$B1,$E8,$D4,$7E,$F2,$3F
	.dc.b	$80,$C7,$F0,$2E,$8F,$DB,$12,$7D,$6F,$EE,$7A,$35,$1F,$C1,$23,$F8
	.dc.b	$4C,$7F,$0A,$E6,$93,$FC,$21,$FE,$1B,$F2,$F2,$30,$43,$E1,$BF,$C4
	.dc.b	$BE,$0B,$F8,$31,$23,$E5,$BF,$C2,$7E,$5F,$51,$FC,$52,$3F,$8C,$C7
	.dc.b	$F1,$AE,$7F,$E3,$7C,$F2,$BF,$8C,$C7,$F1,$EF,$95,$FE,$24,$48,$FA
	.dc.b	$EF,$F1,$5F,$90,$D4,$7F,$20,$8F,$E4,$B1,$FE,$50,$FA,$4F,$CC,$CF
	.dc.b	$FB,$8D,$5E,$4D,$5E,$EF,$FC,$AB,$26,$92,$BF,$15,$1F,$E9,$3E,$4F
	.dc.b	$F3,$DB,$7C,$8F,$EA,$8D,$9F,$F5,$6E,$5F,$E9,$DA,$FF,$40,$BF,$E7
	.dc.b	$83,$15,$FD,$6D,$B3,$C3,$E3,$FF,$CD,$3A,$FF,$98,$63,$FD,$F8,$D7
	.dc.b	$FD,$F0,$DB,$FF,$46,$FD,$23,$CF,$FF,$38,$BF,$C6,$EB,$F4,$4D,$7F
	.dc.b	$B7,$F5,$23,$11,$FE,$72,$6B,$FD,$BF,$E9,$3F,$D1,$E7,$9A,$4D,$3F
	.dc.b	$D2,$B0,$F9,$5F,$F0,$EE,$1F,$EC,$9F,$80,$FE,$FD,$0F,$67,$FD,$93
	.dc.b	$18,$6F,$6F,$FB,$7F,$D9,$5F,$18,$C3,$70,$FF,$6A,$61,$CD,$C1,$E1
	.dc.b	$BE,$6E,$AF,$9B,$FE,$44,$BE,$6F,$FB,$8B,$0B,$97,$F9,$22,$38,$63
	.dc.b	$FE,$0F,$9C,$9F,$DA,$7E,$53,$13,$FE,$18,$6B,$FE,$DD,$3A,$19,$29
	.dc.b	$98,$FF,$0E,$FA,$3F,$EE,$77,$FC,$3B,$3E,$A9,$FE,$E3,$C5,$54,$66
	.dc.b	$B8,$63,$E3,$9F,$61,$FE,$F4,$8C,$BF,$EF,$ED,$5F,$29,$FE,$F5,$CB
	.dc.b	$FD,$A7,$0F,$94,$C6,$5F,$7E,$AD,$F1,$9B,$A3,$E2,$3F,$B6,$5F,$F2
	.dc.b	$D6,$D7,$E3,$1A,$FF,$AF,$1A,$FF,$8E,$78,$FF,$A2,$4F,$E7,$3C,$7E
	.dc.b	$B4,$F4,$FB,$D6,$AF,$F0,$77,$FA,$7C,$BC,$7C,$FE,$98,$4E,$8F,$D8
	.dc.b	$CE,$B3,$F4,$E8,$F9,$4A,$66,$F8,$ED,$F1,$23,$7F,$20,$9F,$E9,$54
	.dc.b	$8C,$63,$FA,$A8,$C4,$EA,$73,$3F,$ED,$C9,$7C,$9F,$8B,$E4,$FC,$6D
	.dc.b	$BF,$09,$FC,$0F,$0F,$6E,$1E,$E6,$C6,$B0,$FE,$1B,$3D,$9F,$E1,$4D
	.dc.b	$73,$1F,$D8,$74,$F2,$C7,$F0,$07,$C7,$1C,$F8,$51,$FE,$EF,$8C,$3A
	.dc.b	$9F,$99,$3A,$C7,$E8,$FC,$97,$D2,$F9,$2B,$87,$07,$0F,$85,$F2,$57
	.dc.b	$66,$E0,$6F,$1F,$DC,$F3,$FE,$EB,$6F,$C7,$17,$2F,$1D,$FE,$17,$7F
	.dc.b	$E5,$B3,$F8,$4F,$91,$FC,$2F,$C8,$DC,$3D,$6E,$1F,$88,$F8,$88,$FC
	.dc.b	$7E,$6F,$C6,$FC,$7F,$E3,$9B,$1F,$12,$3D,$24,$7E,$73,$EC,$F9,$0F
	.dc.b	$CA,$9F,$96,$F9,$0B,$3F,$33,$E4,$3F,$32,$63,$5F,$5B,$F9,$CF,$83
	.dc.b	$FC,$EF,$C8,$5A,$FA,$6E,$1F,$B8,$FC,$A2,$3F,$E7,$BF,$A3,$FF,$1C
	.dc.b	$BF,$EB,$27,$FB,$0B,$7C,$AF,$DE,$E0,$9C,$BF,$E4,$46,$BF,$EB,$46
	.dc.b	$BF,$AE,$38,$79,$7F,$8F,$6D,$7F,$A2,$FE,$83,$EA,$3B,$7F,$C7,$4F
	.dc.b	$FB,$11,$CC,$7E,$E1,$CB,$FE,$54,$6B,$FE,$C4,$6B,$CF,$87,$C1,$18
	.dc.b	$8F,$E0,$3F,$28,$F5,$7D,$B2,$E0,$F7,$CE,$70,$F1,$46,$8E,$64,$7D
	.dc.b	$4F,$6E,$0F,$8F,$EF,$3E,$4F,$AC,$61,$EF,$C5,$E0,$E2,$B8,$7B,$DC
	.dc.b	$3F,$15,$F4,$91,$FF,$2B,$7C,$0F,$F6,$57,$F6,$8B,$CA,$FE,$62,$E7
	.dc.b	$6F,$D9,$B9,$9D,$CE,$BE,$27,$6D,$8F,$F7,$9E,$1F,$E9,$D7,$E9,$72
	.dc.b	$7E,$0A,$31,$87,$D1,$8F,$92,$7C,$E2,$4F,$DD,$7F,$01,$E2,$BB,$BF
	.dc.b	$17,$CC,$F6,$BF,$1B,$39,$BF,$E3,$18,$7F,$A3,$27,$FA,$5F,$D5,$66
	.dc.b	$3F,$86,$AF,$37,$38,$E9,$97,$FC,$5E,$7E,$93,$63,$9F,$07,$4F,$94
	.dc.b	$F0,$C5,$9F,$0E,$5D,$5B,$1E,$A2,$3F,$EE,$38,$D6,$3F,$E2,$6C,$1F
	.dc.b	$27,$E8,$F9,$4F,$4B,$6F,$C9,$1E,$CF,$94,$C5,$FB,$1F,$7F,$A5,$2B
	.dc.b	$D5,$16,$7E,$0F,$17,$ED,$3F,$2E,$6B,$F1,$A3,$0C,$BF,$85,$C9,$6F
	.dc.b	$E4,$BF,$33,$E4,$FE,$6F,$C9,$DA,$F9,$0E,$1F,$95,$F2,$76,$FE,$8B
	.dc.b	$F5,$3E,$2B,$EC,$F1,$DA,$F9,$2E,$1F,$8C,$FE,$08,$8F,$C9,$CB,$EE
	.dc.b	$FC,$9F,$E0,$9B,$1F,$67,$DB,$FC,$09,$F8,$2E,$BB,$1F,$A0,$F8,$DF
	.dc.b	$E8,$F0,$69,$23,$F8,$96,$7F,$E0,$0F,$98,$7E,$29,$FC,$2B,$CD,$FE
	.dc.b	$EB,$18,$F1,$FF,$1F,$87,$F8,$72,$31,$87,$F6,$B7,$8F,$99,$9C,$0F
	.dc.b	$E5,$BE,$3F,$0F,$E6,$3E,$57,$FA,$DF,$61,$A4,$C3,$FD,$A1,$0F,$95
	.dc.b	$FF,$14,$FC,$8F,$F8,$5A,$7F,$35,$18,$DB,$3E,$F4,$9F,$EE,$95,$FE
	.dc.b	$2C,$DF,$FC,$1F,$E1,$7F,$BD,$EB,$FC,$67,$F3,$CF,$90,$FE,$DA,$DC
	.dc.b	$3F,$DC,$0D,$C7,$FC,$FE,$D7,$F9,$D5,$FF,$C0,$1F,$96,$F2,$9E,$6F
	.dc.b	$81,$B7,$E2,$C3,$D2,$7A,$BE,$0A,$C7,$9B,$E4,$3F,$BE,$EA,$69,$23
	.dc.b	$FD,$E5,$1F,$C1,$5F,$A1,$CC,$F0,$FE,$2F,$93,$F5,$63,$13,$F0,$BD
	.dc.b	$3F,$93,$F2,$1F,$29,$B1,$FE,$19,$E0,$7C,$87,$C2,$24,$7C,$27,$FC
	.dc.b	$49,$D4,$F9,$F7,$C8,$FE,$0D,$B5,$F4,$2B,$13,$E3,$F2,$24,$FD,$48
	.dc.b	$BC,$BF,$E2,$4D,$DF,$23,$F0,$89,$5F,$01,$F8,$BF,$23,$F8,$2F,$89
	.dc.b	$FC,$1B,$67,$D5,$F2,$3F,$70,$FC,$39,$5F,$77,$C0,$FD,$E2,$3F,$79
	.dc.b	$FF,$16,$FB,$1F,$E0,$0C,$BC,$91,$FE,$78,$8C,$76,$FE,$2B,$3F,$F1
	.dc.b	$FD,$FE,$3B,$D5,$FC,$AF,$BB,$F9,$66,$C7,$E4,$A3,$EE,$91,$FB,$6F
	.dc.b	$CC,$E6,$79,$5F,$64,$FC,$FF,$C8,$5A,$FB,$F1,$FA,$2F,$91,$C6,$BF
	.dc.b	$25,$FF,$46,$FE,$23,$FE,$6F,$AF,$F2,$A3,$67,$FC,$A2,$BE,$47,$73
	.dc.b	$0D,$FB,$36,$B6,$FF,$A6,$3E,$73,$84,$FF,$A1,$BE,$43,$F9,$6D,$7F
	.dc.b	$59,$9F,$F5,$2C,$E7,$D1,$FF,$45,$F8,$E9,$E3,$5F,$EB,$29,$CC,$70
	.dc.b	$8C,$1C,$F0,$FF,$42,$E1,$DA,$F9,$46,$2F,$8E,$36,$F8,$63,$E1,$69
	.dc.b	$D2,$F8,$C6,$5F,$24,$F1,$73,$3B,$CB,$FD,$48,$DF,$8D,$D4,$63,$2F
	.dc.b	$23,$FE,$A3,$C3,$A3,$9C,$7E,$0B,$9B,$E9,$F2,$1C,$B0,$F4,$43,$E7
	.dc.b	$7F,$D9,$5E,$5F,$F3,$48,$FF,$59,$C6,$1C,$35,$FE,$72,$D7,$FC,$99
	.dc.b	$F2,$4F,$6D,$F3,$DC,$E5,$39,$8C,$36,$3C,$0D,$7A,$E1,$FE,$15,$7C
	.dc.b	$2D,$73,$B7,$5C,$3F,$D8,$8F,$A4,$75,$38,$FF,$CC,$1C,$1F,$6F,$E1
	.dc.b	$FF,$D7,$AD,$E1,$CF,$0F,$E2,$2F,$86,$5F,$5D,$AF,$DB,$BE,$33,$F8
	.dc.b	$E6,$C7,$F8,$AE,$77,$7F,$89,$ED,$FD,$04,$FE,$05,$AF,$E1,$5B,$1B
	.dc.b	$E4,$F1,$CF,$F8,$D6,$73,$87,$E7,$1B,$19,$E4,$DE,$89,$5E,$1D,$C6
	.dc.b	$67,$06,$67,$8E,$5C,$2F,$F7,$3F,$23,$FE,$87,$ED,$FE,$B3,$CD,$FE
	.dc.b	$E9,$9F,$F6,$85,$6D,$FF,$66,$B6,$7F,$AF,$B7,$6F,$8D,$87,$37,$E4
	.dc.b	$63,$6F,$FC,$A5,$5F,$EF,$D7,$F8,$4F,$C9,$67,$36,$5F,$EF,$29,$F0
	.dc.b	$F1,$66,$F8,$47,$8A,$30,$75,$F1,$DE,$F7,$19,$FE,$27,$F2,$5F,$95
	.dc.b	$C7,$AB,$71,$34,$D6,$3C,$76,$F8,$71,$E5,$9D,$DF,$C4,$CF,$C9,$F2
	.dc.b	$59,$24,$7C,$87,$D9,$D9,$F2,$E3,$FC,$47,$97,$F1,$7E,$1F,$24,$E6
	.dc.b	$7F,$BF,$B5,$FC,$15,$7C,$5E,$23,$4C,$7C,$97,$C5,$F0,$3C,$CC,$AF
	.dc.b	$B3,$1E,$9F,$92,$F8,$04,$8F,$D5,$68,$8F,$89,$3E,$F2,$4F,$F8,$C7
	.dc.b	$83,$FC,$5E,$FF,$40,$F9,$2F,$88,$47,$FC,$6D,$F3,$DF,$E4,$34,$38
	.dc.b	$7F,$C4,$BA,$FA,$A3,$F7,$4D,$7E,$2B,$5F,$87,$A7,$D4,$EA,$FA,$5E
	.dc.b	$E4,$9F,$CA,$C7,$E0,$72,$FF,$65,$47,$F1,$DE,$5F,$C1,$FC,$7F,$F1
	.dc.b	$56,$C6,$67,$6F,$D4,$DD,$F2,$7F,$85,$24,$7C,$67,$F1,$17,$F8,$32
	.dc.b	$47,$E9,$BE,$4F,$F0,$D8,$C8,$FE,$CA,$FE,$77,$E7,$7F,$8E,$6D,$F0
	.dc.b	$B8,$3F,$02,$4A,$F8,$6F,$E5,$7E,$43,$F2,$04,$8F,$97,$3F,$E4,$EA
	.dc.b	$FF,$6F,$46,$3B,$7F,$B1,$C3,$E3,$7F,$31,$F2,$32,$B4,$7F,$B0,$E7
	.dc.b	$4F,$F7,$AC,$FF,$48,$47,$F0,$AF,$E9,$FE,$47,$F4,$44,$8F,$F3,$A7
	.dc.b	$CF,$7F,$B8,$5E,$BD,$9F,$E8,$33,$F7,$9F,$11,$F0,$33,$FE,$EE,$D9
	.dc.b	$72,$2D,$F1,$DF,$A8,$DF,$F4,$A4,$AF,$C3,$BF,$BB,$BF,$F6,$94,$78
	.dc.b	$1A,$FE,$3F,$0E,$FF,$25,$F4,$7C,$CF,$A5,$CF,$E4,$F8,$D9,$7E,$8E
	.dc.b	$0F,$E1,$2F,$FB,$8B,$E6,$FF,$CF,$0F,$F5,$D6,$BF,$DA,$5A,$FF,$A7
	.dc.b	$9A,$FE,$80,$2F,$84,$FE,$45,$18,$F8,$5B,$3D,$39,$9A,$71,$8E,$15
	.dc.b	$E5,$F4,$FE,$5F,$99,$23,$04,$3E,$01,$F7,$63,$F2,$07,$E4,$7E,$3F
	.dc.b	$E9,$FC,$9C,$9F,$90,$FE,$57,$E3,$E4,$7C,$36,$BD,$0F,$F0,$07,$4F
	.dc.b	$DC,$EB,$91,$F4,$5F,$CF,$7C,$74,$8F,$98,$D7,$CB,$D4,$E1,$FF,$5A
	.dc.b	$47,$D1,$F8,$CF,$D3,$7C,$74,$8C,$1F,$77,$E2,$BF,$17,$F2,$72,$30
	.dc.b	$7D,$7F,$8C,$FF,$09,$3F,$41,$BC,$5F,$E9,$F8,$FB,$51,$FD,$0F,$E6
	.dc.b	$9F,$1F,$F9,$6F,$88,$95,$F9,$A9,$CB,$7E,$7B,$FA,$B4,$62,$3D,$6D
	.dc.b	$E8,$FF,$60,$A9,$FB,$FF,$06,$F8,$EF,$C0,$CF,$E5,$AB,$EF,$DF,$F2
	.dc.b	$68,$C7,$57,$E2,$6F,$84,$FE,$4F,$E3,$FD,$64,$7F,$28,$FB,$BE,$3F
	.dc.b	$F1,$6C,$8F,$CD,$47,$F6,$DF,$A5,$EB,$0E,$54,$61,$BE,$1F,$F7,$3F
	.dc.b	$9E,$FE,$75,$5F,$61,$BA,$7F,$B7,$9B,$9F,$F9,$C9,$B3,$D5,$75,$FE
	.dc.b	$7B,$6F,$EF,$DE,$77,$B2,$BF,$06,$DC,$3E,$1C,$7E,$A2,$7F,$4B,$F1
	.dc.b	$FF,$68,$91,$FB,$37,$EF,$FC,$C3,$C0,$FD,$6F,$CA,$72,$C3,$9B,$74
	.dc.b	$FD,$A5,$6F,$0F,$DF,$7C,$87,$E1,$88,$FE,$FD,$FC,$4F,$C7,$FE,$D9
	.dc.b	$91,$FC,$16,$3F,$73,$F1,$FF,$BC,$F9,$FF,$DD,$61,$FC,$05,$B5,$FD
	.dc.b	$8B,$B1,$5F,$C0,$39,$FF,$2A,$49,$FE,$1C,$FE,$6B,$E7,$DE,$FF,$F0
	.dc.b	$8F,$AA,$F9,$B4,$6C,$EA,$3F,$9E,$87,$FB,$E3,$FE,$75,$6B,$FE,$81
	.dc.b	$6B,$FC,$5C,$E1,$F0,$1F,$CB,$7D,$09,$1F,$E6,$FF,$47,$FD,$00,$D8
	.dc.b	$FD,$57,$9D,$FC,$F9,$91,$FA,$98,$FE,$83,$E9,$C8,$FE,$6C,$FE,$AB
	.dc.b	$C4,$91,$FA,$F6,$BF,$AC,$F0,$3F,$A4,$FA,$5F,$CF,$C9,$1E,$37,$F7
	.dc.b	$5D,$5A,$8F,$E8,$51,$FB,$68,$FE,$A3,$F3,$5F,$D1,$C9,$3E,$B7,$F8
	.dc.b	$17,$CA,$EA,$3F,$A5,$47,$EF,$A3,$FA,$CF,$5C,$9F,$E9,$CF,$F0,$2E
	.dc.b	$79,$18,$21,$FC,$11,$FE,$C3,$E8,$FF,$57,$24,$7C,$B7,$F8,$97,$C1
	.dc.b	$EA,$3F,$AD,$47,$F0,$E8,$FE,$1F,$F2,$7F,$DA,$BE,$6A,$57,$F1,$58
	.dc.b	$FE,$27,$CF,$FC,$6C,$91,$FE,$A5,$F2,$11,$FC,$7F,$E5,$37,$3F,$C7
	.dc.b	$63,$F8,$DC,$7F,$9C,$7E,$81,$7F,$C1,$F8,$E5,$FF,$7C,$9C,$B6,$FF
	.dc.b	$CD,$91,$DA,$3F,$D6,$8D,$7F,$CB,$ED,$7F,$B2,$79,$08,$61,$E1,$24
	.dc.b	$F7,$F9,$0F,$F4,$27,$6F,$EE,$63,$F8,$C5,$7F,$AD,$BC,$4F,$E5,$38
	.dc.b	$73,$BD,$F8,$3F,$E1,$0F,$1E,$61,$FE,$05,$7F,$D1,$DF,$2F,$FE,$7A
	.dc.b	$8F,$EA,$7F,$40,$E3,$FF,$43,$36,$3F,$D9,$CD,$7F,$CF,$2D,$7D,$7F
	.dc.b	$41,$E5,$C9,$3F,$C8,$9B,$DF,$FE,$9D,$9F,$55,$7F,$B9,$BD,$0F,$F3
	.dc.b	$22,$F3,$7F,$BF,$BF,$EE,$AF,$4B,$FC,$15,$18,$F9,$54,$63,$DD,$FF
	.dc.b	$79,$DF,$2A,$D3,$63,$FD,$EF,$3B,$BF,$F3,$9D,$F2,$7F,$CE,$B7,$C3
	.dc.b	$86,$DF,$47,$FC,$F3,$9F,$F8,$B2,$D8,$FA,$23,$0D,$BE,$18,$7F,$4D
	.dc.b	$7F,$86,$2F,$94,$E7,$18,$6C,$7F,$11,$6B,$FE,$FA,$39,$19,$3E,$36
	.dc.b	$3F,$C5,$FE,$5F,$FB,$4D,$7C,$89,$FF,$47,$D7,$A9,$FF,$19,$4E,$AB
	.dc.b	$C1,$1C,$8C,$F1,$6D,$F0,$5F,$3A,$E1,$1F,$D9,$F5,$7C,$C7,$F7,$8D
	.dc.b	$7F,$23,$F1,$2B,$C6,$3F,$22,$F6,$FF,$99,$27,$4D,$F6,$3F,$AF,$E3
	.dc.b	$F3,$A7,$FC,$27,$3F,$23,$6F,$F4,$9D,$6A,$BF,$BC,$56,$9F,$F6,$5F
	.dc.b	$08,$F0,$BF,$DF,$5C,$5F,$F9,$D3,$93,$FD,$7C,$FF,$99,$E3,$78,$7F
	.dc.b	$54,$D2,$E4,$F9,$4C,$F4,$FA,$B2,$FF,$1C,$CE,$71,$FF,$3F,$AF,$91
	.dc.b	$E2,$72,$BB,$3A,$F9,$2F,$56,$4B,$46,$1B,$3F,$05,$B8,$7F,$BA,$1F
	.dc.b	$F7,$BC,$7F,$A3,$BD,$9F,$F0,$03,$E2,$F9,$3F,$5C,$62,$BF,$C9,$F9
	.dc.b	$F9,$6B,$FD,$2A,$F4,$E2,$FE,$12,$E6,$77,$8F,$C0,$3C,$F1,$E2,$8F
	.dc.b	$E3,$2D,$7F,$4A,$D7,$FC,$58,$F9,$9F,$F2,$FE,$CF,$81,$F9,$79,$7D
	.dc.b	$08,$C6,$3E,$4F,$4F,$1E,$FF,$D0,$6C,$7F,$A7,$63,$E4,$C9,$A4,$8F
	.dc.b	$C4,$62,$F9,$3E,$DF,$03,$F6,$9B,$3F,$EA,$78,$B9,$FC,$69,$BA,$FC
	.dc.b	$4F,$90,$FC,$2F,$D6,$7F,$BE,$B0,$F9,$31,$8B,$91,$F9,$79,$D4,$7F
	.dc.b	$82,$F8,$91,$88,$FE,$82,$D7,$E4,$FB,$FD,$BA,$BF,$1E,$DB,$FF,$65
	.dc.b	$D7,$EA,$2B,$A3,$5E,$75,$F8,$C8,$C4,$7D,$4F,$8E,$FC,$C9,$F9,$AF
	.dc.b	$90,$B7,$D3,$E1,$3D,$A6,$2B,$FD,$2B,$7F,$E7,$99,$FC,$3F,$C7,$FC
	.dc.b	$2E,$EF,$D1,$B6,$3C,$B1,$E1,$23,$F0,$DF,$DC,$53,$E8,$7F,$B6,$CF
	.dc.b	$F2,$48,$C7,$9D,$A9,$CC,$7F,$28,$6B,$FE,$35,$7F,$B3,$CF,$F8,$FB
	.dc.b	$EA,$7F,$31,$C3,$FE,$68,$9C,$D7,$FA,$DE,$73,$5F,$ED,$3F,$A2,$FC
	.dc.b	$23,$E1,$6C,$FF,$36,$6B,$FE,$43,$6B,$C7,$E4,$FF,$86,$BF,$C3,$FB
	.dc.b	$2C,$A7,$3D,$3F,$AD,$BF,$07,$1E,$B8,$76,$D8,$FF,$2B,$DA,$7E,$3B
	.dc.b	$63,$CD,$18,$3E,$5F,$C2,$BE,$43,$C7,$F3,$1F,$EA,$DA,$FF,$30,$CE
	.dc.b	$5F,$F3,$1C,$67,$4F,$F6,$D6,$0E,$5E,$E4,$AF,$88,$F8,$BE,$3F,$F9
	.dc.b	$2F,$D4,$7F,$9C,$A7,$CF,$18,$EF,$FE,$53,$E8,$BE,$0E,$98,$F3,$AD
	.dc.b	$B3,$FF,$70,$EF,$F9,$BE,$6F,$F0,$CB,$67,$FB,$7E,$1C,$DE,$13,$AE
	.dc.b	$CF,$9F,$D5,$F4,$BE,$6F,$FD,$75,$5F,$2A,$31,$DD,$23,$F1,$D3,$AC
	.dc.b	$BF,$B9,$B5,$FF,$40,$35,$FF,$68,$7C,$17,$CC,$F7,$FE,$FF,$B2,$8C
	.dc.b	$5F,$90,$FF,$02,$FC,$5F,$F8,$65,$C5,$FF,$B3,$63,$E9,$3F,$F0,$01
	.dc.b	$D3,$7A,$3F,$A8,$C6,$B5,$FF,$1D,$F8,$2F,$A1,$CD,$7F,$12,$7E,$07
	.dc.b	$A5,$FA,$CD,$79,$E3,$EA,$8C,$1F,$C0,$FE,$1B,$F8,$CF,$A9,$FC,$69
	.dc.b	$BA,$79,$B8,$1A,$BE,$4B,$C7,$F2,$3E,$46,$D6,$E7,$FC,$62,$DF,$67
	.dc.b	$39,$FB,$5F,$E4,$FF,$11,$27,$47,$F1,$A7,$DF,$D5,$23,$FD,$6F,$F4
	.dc.b	$6F,$95,$C6,$1F,$C4,$B9,$7F,$98,$98,$E5,$F8,$AF,$C8,$F8,$89,$87
	.dc.b	$97,$E2,$AD,$FC,$F9,$FE,$73,$F2,$78,$C3,$E6,$BF,$89,$E7,$FC,$56
	.dc.b	$57,$1F,$F6,$4F,$37,$8D,$B1,$F3,$C8,$FE,$31,$FE,$F7,$C1,$FE,$69
	.dc.b	$8F,$E6,$FF,$17,$F7,$3E,$0B,$FA,$4B,$FD,$2B,$E4,$A4,$FC,$BF,$89
	.dc.b	$BB,$7E,$93,$A7,$FC,$0B,$E1,$1A,$4A,$FE,$9D,$DF,$FD,$50,$FF,$0F
	.dc.b	$FC,$C3,$E8,$FF,$8D,$F1,$7E,$2A,$7F,$AC,$FC,$A4,$BF,$CA,$DE,$FD
	.dc.b	$7C,$67,$FD,$03,$E3,$BF,$A5,$12,$BF,$AE,$3F,$E1,$8F,$A1,$FE,$BB
	.dc.b	$97,$E5,$BE,$17,$FC,$53,$18,$AF,$F3,$3F,$8A,$9F,$ED,$3F,$1F,$27
	.dc.b	$F2,$15,$FB,$FC,$5E,$1F,$F6,$6A,$FF,$19,$F0,$B8,$FE,$19,$30,$FF
	.dc.b	$22,$46,$5F,$EE,$1E,$F7,$F7,$63,$C7,$ED,$D8,$FE,$D3,$E4,$FF,$7A
	.dc.b	$3C,$FE,$3D,$8F,$ED,$BE,$DF,$F6,$92,$3F,$DE,$DF,$F2,$87,$1F,$F8
	.dc.b	$2F,$C8,$BE,$E9,$A4,$AF,$F2,$C7,$07,$F2,$FE,$3F,$DD,$FC,$6F,$C6
	.dc.b	$EC,$FF,$05,$BE,$EE,$4F,$BE,$4B,$F6,$BF,$DE,$FE,$33,$FC,$22,$7D
	.dc.b	$EF,$A2,$75,$E2,$BF,$75,$F6,$07,$FC,$B6,$D7,$EF,$EB,$FE,$98,$76
	.dc.b	$6B,$E8,$7F,$CA,$72,$69,$23,$D7,$8F,$4F,$13,$EF,$B6,$3F,$85,$7D
	.dc.b	$23,$9B,$FC,$D6,$D8,$CF,$4E,$27,$EE,$F3,$7E,$78,$93,$F7,$1F,$C7
	.dc.b	$47,$AB,$E3,$3F,$1A,$D9,$FC,$4F,$37,$BB,$E2,$FF,$22,$D8,$F8,$71
	.dc.b	$EC,$23,$FE,$8B,$D7,$FC,$F1,$DB,$FE,$35,$E6,$71,$FE,$68,$FD,$5F
	.dc.b	$0F,$EE,$E5,$CD,$FE,$EE,$AD,$E1,$F9,$AB,$F9,$FF,$19,$C6,$3E,$9B
	.dc.b	$FE,$8C,$F9,$F7,$1D,$31,$34,$FE,$65,$1A,$7E,$DF,$C2,$67,$2F,$0C
	.dc.b	$7F,$B4,$B4,$FF,$06,$47,$EA,$3E,$87,$83,$FE,$CD,$F2,$FF,$DA,$F1
	.dc.b	$8F,$90,$FE,$59,$3B,$E8,$FC,$33,$B3,$96,$78,$9D,$F8,$6B,$E5,$87
	.dc.b	$4A,$D3,$DB,$0B,$87,$2C,$BF,$DA,$33,$B6,$C6,$4F,$06,$1D,$6B,$8B
	.dc.b	$DE,$EE,$9F,$EA,$FC,$BC,$6F,$D8,$6F,$2E,$B8,$F6,$DC,$BB,$E0,$BE
	.dc.b	$59,$FF,$BE,$9E,$4F,$F8,$4F,$37,$D1,$FF,$B7,$3D,$BF,$F2,$A3,$63
	.dc.b	$FD,$F2,$D7,$81,$C3,$1F,$F7,$03,$FE,$FE,$F9,$2F,$F1,$17,$D2,$7F
	.dc.b	$97,$5B,$19,$F5,$D3,$96,$C7,$6D,$7F,$91,$B7,$4F,$1C,$62,$3C,$38
	.dc.b	$FF,$B5,$67,$27,$CE,$8F,$5C,$F2,$DB,$F0,$31,$89,$D5,$7F,$B5,$E2
	.dc.b	$D6,$F4,$F8,$93,$D6,$7F,$0C,$71,$2B,$C5,$B3,$C5,$FF,$1F,$47,$1B
	.dc.b	$F0,$B6,$BA,$F2,$7D,$19,$CD,$F1,$BE,$D1,$8C,$3C,$F5,$BB,$F2,$DE
	.dc.b	$F0,$E9,$87,$82,$31,$5B,$8F,$E9,$BC,$6F,$8F,$FC,$07,$47,$E8,$B2
	.dc.b	$4B,$E8,$F3,$C7,$FA,$F6,$E1,$CB,$0B,$7C,$1B,$7D,$67,$F9,$9C,$63
	.dc.b	$E0,$B8,$F8,$EF,$C6,$3F,$BE,$57,$F3,$8F,$9B,$E2,$6A,$33,$1F,$C0
	.dc.b	$04,$F1,$F4,$F4,$48,$FE,$83,$DF,$E6,$5F,$5A,$F8,$CF,$F5,$74,$61
	.dc.b	$B1,$98,$FF,$59,$35,$FF,$6C,$69,$FE,$9B,$7F,$CB,$67,$12,$C6,$F8
	.dc.b	$3E,$54,$EE,$7D,$1B,$7C,$CF,$4F,$E2,$92,$3F,$AB,$FC,$BF,$BF,$CD
	.dc.b	$FF,$61,$6B,$F1,$5B,$B1,$A7,$FB,$1F,$0F,$AF,$A7,$D0,$26,$1F,$D8
	.dc.b	$BD,$1F,$95,$EA,$C8,$FF,$0F,$E0,$8F,$AF,$F4,$9F,$3F,$D6,$EE,$4A
	.dc.b	$FD,$E9,$F4,$29,$1E,$1D,$BF,$D5,$13,$FD,$C3,$E2,$FF,$02,$7E,$0B
	.dc.b	$E2,$BE,$D7,$9D,$27,$C2,$FE,$1B,$E2,$3E,$F7,$9F,$EE,$24,$F8,$38
	.dc.b	$7F,$0D,$1F,$0B,$CE,$FC,$21,$27,$FC,$F1,$FB,$8F,$F0,$3E,$6F,$CD
	.dc.b	$2F,$88,$FC,$57,$9F,$2B,$71,$F9,$2F,$C5,$FF,$BF,$47,$76,$BF,$C7
	.dc.b	$32,$46,$23,$F9,$34,$63,$CE,$FA,$58,$7F,$B4,$67,$FC,$97,$5F,$D7
	.dc.b	$9B,$E7,$7E,$9A,$75,$1F,$84,$7F,$89,$E5,$E8,$FA,$6D,$1B,$F2,$3C
	.dc.b	$DE,$8F,$E1,$31,$34,$98,$7D,$6E,$B6,$3F,$E9,$98,$F8,$BF,$15,$E1
	.dc.b	$3C,$5E,$D7,$C1,$F1,$7C,$9F,$15,$2B,$D3,$1F,$17,$C4,$F0,$12,$3E
	.dc.b	$33,$E3,$F9,$3F,$94,$48,$FA,$7F,$1D,$F3,$7C,$4F,$81,$E8,$49,$F8
	.dc.b	$B1,$FB,$5F,$DB,$9F,$43,$FB,$46,$CF,$E4,$03,$87,$F2,$6D,$7E,$9F
	.dc.b	$0B,$C5,$FB,$84,$8F,$C2,$FD,$07,$D1,$F8,$C9,$1F,$88,$7F,$8D,$FA
	.dc.b	$AF,$D2,$FF,$CB,$53,$FB,$EC,$5F,$25,$E5,$24,$79,$63,$2F,$E9,$A9
	.dc.b	$FB,$7F,$D6,$8E,$55,$EF,$3B,$C6,$63,$F0,$5F,$1F,$F8,$5D,$64,$7E
	.dc.b	$72,$3F,$10,$47,$07,$E8,$3E,$33,$F1,$9F,$19,$23,$F4,$6F,$E3,$F0
	.dc.b	$91,$82,$1F,$91,$7F,$29,$F3,$1F,$95,$F9,$E9,$3F,$AC,$8F,$CC,$7C
	.dc.b	$F6,$A3,$07,$EC,$1F,$CE,$7C,$F6,$A3,$07,$ED,$1F,$E7,$FF,$A6,$7E
	.dc.b	$7B,$DC,$4F,$FA,$7B,$D3,$FF,$04,$F8,$BF,$D4,$5B,$3A,$F2,$17,$C6
	.dc.b	$F8,$7C,$1F,$4F,$1A,$47,$FB,$E7,$0F,$F0,$8B,$FC,$C1,$7C,$FF,$F5
	.dc.b	$65,$F1,$F4,$C4,$67,$E6,$65,$73,$78,$7C,$87,$F4,$BF,$E0,$1F,$C9
	.dc.b	$6D,$FE,$48,$A4,$63,$6F,$F1,$53,$63,$DF,$F8,$7F,$89,$FE,$92,$F8
	.dc.b	$9F,$CF,$35,$FE,$5E,$D7,$FC,$5A,$D7,$FC,$6D,$F8,$E3,$8F,$FA,$F4
	.dc.b	$F1,$C3,$C8,$DC,$37,$3F,$07,$E0,$66,$1F,$1D,$F8,$9F,$39,$23,$04
	.dc.b	$3E,$53,$F6,$BC,$1F,$95,$F3,$72,$7F,$DD,$3C,$89,$F9,$DF,$37,$B9
	.dc.b	$CC,$61,$FA,$DF,$05,$F4,$C8,$FF,$6C,$FC,$49,$5E,$CC,$BF,$BB,$EB
	.dc.b	$FE,$02,$AF,$87,$18,$3F,$83,$78,$5F,$B0,$F6,$F8,$4F,$CE,$9F,$D7
	.dc.b	$C7,$E4,$BC,$2F,$C1,$7C,$C4,$AF,$4C,$7E,$17,$E5,$B5,$18,$3F,$18
	.dc.b	$FE,$6B,$C1,$FC,$57,$C2,$49,$F8,$31,$F8,$DF,$83,$D4,$60,$FC,$A3
	.dc.b	$FE,$13,$FA,$97,$C4,$FF,$84,$9B,$3F,$C7,$DA,$FF,$70,$6B,$FE,$6E
	.dc.b	$5E,$9F,$39,$93,$FE,$14,$9C,$BF,$AE,$E9,$FF,$4E,$F8,$6F,$DB,$3A
	.dc.b	$C7,$EB,$FE,$63,$F4,$87,$F8,$D2,$33,$D9,$FB,$16,$DF,$F8,$C1,$B7
	.dc.b	$CF,$1F,$F2,$0F,$EC,$6F,$13,$FC,$61,$FB,$37,$F8,$13,$0F,$F0,$4B
	.dc.b	$5F,$F4,$3B,$5F,$F4,$83,$5F,$E0,$AB,$E1,$F9,$FC,$44,$9F,$F1,$8C
	.dc.b	$E5,$FC,$8F,$95,$FE,$28,$EF,$34,$91,$FD,$97,$07,$C4,$7F,$14,$99
	.dc.b	$3F,$C2,$63,$FC,$AB,$CD,$FC,$4A,$3F,$8B,$7C,$CF,$F0,$93,$2C,$3E
	.dc.b	$77,$07,$F1,$79,$E1,$B9,$E5,$F0,$3F,$17,$E2,$7A,$F9,$7F,$E3,$E9
	.dc.b	$FE,$09,$87,$F8,$3D,$AF,$F8,$39,$AF,$F2,$8A,$F9,$DF,$13,$CA,$3C
	.dc.b	$27,$F9,$92,$79,$F3,$7F,$2B,$9F,$E4,$1F,$39,$2B,$F9,$33,$FC,$8F
	.dc.b	$E2,$BF,$CF,$3F,$5F,$FC,$6E,$7F,$D8,$58,$A3,$11,$E1,$6A,$8D,$6B
	.dc.b	$FE,$53,$9F,$F1,$4E,$3F,$E8,$B7,$F9,$33,$63,$BF,$C5,$78,$3E,$2A
	.dc.b	$61,$EA,$C3,$C1,$1E,$4F,$98,$F2,$FC,$AF,$52,$57,$F8,$7F,$24,$78
	.dc.b	$7E,$57,$D6,$49,$E9,$3F,$DB,$F3,$F2,$7C,$67,$C0,$24,$EB,$E5,$3E
	.dc.b	$1F,$CA,$C8,$E2,$FC,$6F,$8A,$F8,$FE,$07,$BC,$93,$E3,$7E,$17,$89
	.dc.b	$F3,$09,$1E,$2F,$8A,$FF,$4A,$79,$3F,$BC,$8F,$F7,$83,$5F,$EC,$AD
	.dc.b	$7F,$C4,$B5,$FE,$98,$8F,$F0,$44,$7F,$BC,$1A,$F2,$CB,$47,$FA,$7B
	.dc.b	$E6,$1F,$16,$D7,$F1,$5F,$23,$FE,$FF,$9C,$C6,$2F,$FB,$AC,$61,$FB
	.dc.b	$D3,$BA,$E5,$18,$3F,$E0,$0F,$CC,$71,$7F,$DB,$46,$4B,$CB,$FE,$73
	.dc.b	$76,$67,$27,$2F,$E3,$1B,$CF,$FE,$41,$6C,$FF,$A4,$9A,$FC,$1E,$1E
	.dc.b	$04,$9E,$0F,$E6,$2F,$F3,$5F,$2C,$9F,$C1,$B6,$BF,$D6,$BC,$8F,$8D
	.dc.b	$C6,$FF,$99,$AB,$7F,$A1,$F9,$6E,$97,$FE,$CB,$8F,$C9,$46,$0C,$FC
	.dc.b	$6F,$0C,$7F,$4D,$5F,$E5,$5A,$FC,$9F,$8D,$9C,$FF,$5D,$76,$9E,$4D
	.dc.b	$31,$E5,$5C,$3F,$37,$E9,$EC,$C6,$5E,$A7,$A7,$DD,$FE,$D6,$7F,$D0
	.dc.b	$8D,$9F,$F1,$89,$FC,$2D,$C3,$FC,$01,$AE,$79,$79,$9D,$3E,$8A,$D7
	.dc.b	$ED,$E9,$F0,$BE,$0F,$96,$F0,$B6,$75,$1F,$4E,$4D,$24,$79,$BE,$57
	.dc.b	$D2,$7A,$BE,$56,$C7,$7F,$95,$F9,$9E,$6F,$F2,$DF,$88,$F7,$C6,$3C
	.dc.b	$C4,$53,$F8,$CC,$EE,$BF,$D9,$93,$A6,$CE,$A1,$EE,$73,$3E,$5F,$7D
	.dc.b	$AF,$E5,$F7,$FC,$F6,$7E,$7F,$CA,$E7,$87,$FD,$FF,$B2,$3E,$9F,$C1
	.dc.b	$F9,$7E,$12,$E3,$CB,$34,$66,$3C,$AF,$8C,$8F,$FA,$8F,$E2,$DF,$0B
	.dc.b	$F8,$DC,$90,$F8,$1C,$3A,$AF,$E8,$B3,$A9,$C9,$FC,$EA,$78,$4F,$58
	.dc.b	$C1,$F2,$3F,$69,$F8,$71,$F0,$1B,$5F,$C2,$E3,$FA,$A7,$EC,$2F,$2F
	.dc.b	$EA,$D6,$BC,$3F,$F0,$8B,$6B,$F2,$1F,$2D,$AE,$1F,$CB,$BE,$BF,$AB
	.dc.b	$7B,$37,$D8,$C3,$6F,$D7,$29,$FE,$A6,$D8,$FF,$64,$37,$4F,$CF,$9F
	.dc.b	$E9,$07,$0F,$E7,$FE,$53,$F1,$1C,$1F,$A8,$6D,$7D,$58,$FD,$5F,$97
	.dc.b	$8A,$FB,$AF,$EB,$BE,$53,$F2,$1F,$0F,$6F,$EE,$38,$7F,$4D,$F2,$96
	.dc.b	$FF,$0A,$FF,$05,$F8,$97,$5F,$FB,$F2,$BE,$BF,$E0,$AF,$13,$53,$9C
	.dc.b	$3E,$81,$FE,$AA,$9D,$57,$0A,$FE,$EB,$18,$9F,$82,$1C,$3F,$11,$AF
	.dc.b	$F0,$3F,$8A,$FE,$12,$7E,$37,$F0,$A7,$8F,$8C,$3F,$80,$7C,$87,$EB
	.dc.b	$FE,$13,$F8,$7B,$63,$F3,$91,$FC,$4B,$E5,$B1,$8F,$E9,$1F,$E2,$BF
	.dc.b	$2D,$FC,$5B,$E1,$ED,$FF,$0B,$70,$D3,$F8,$74,$62,$BE,$08,$F9,$E7
	.dc.b	$C0,$7E,$8F,$67,$99,$FA,$59,$CE,$1F,$32,$34,$7C,$62,$3F,$1A,$FE
	.dc.b	$34,$7F,$87,$A7,$F3,$93,$FA,$A6,$CF,$8A,$9E,$67,$7F,$07,$F9,$E7
	.dc.b	$D6,$7F,$B7,$B0,$FF,$0E,$36,$FE,$CF,$AA,$8F,$DC,$CE,$A7,$8F,$C4
	.dc.b	$72,$F9,$8B,$3C,$9E,$BE,$7F,$F2,$2F,$D4,$9F,$11,$8A,$FE,$2B,$18
	.dc.b	$6C,$79,$3E,$67,$83,$63,$CB,$E3,$FA,$3E,$5F,$D2,$D8,$F4,$C7,$B3
	.dc.b	$CE,$C6,$9E,$AD,$7F,$8C,$D7,$BA,$7E,$0F,$CB,$6E,$3E,$0B,$F1,$3C
	.dc.b	$5F,$8B,$E7,$59,$F8,$AE,$1A,$7C,$88,$C6,$1D,$FD,$0F,$9A,$7C,$EF
	.dc.b	$16,$D7,$CE,$8F,$A3,$F2,$B8,$AF,$33,$FA,$AF,$67,$F0,$DE,$4D,$BF
	.dc.b	$D7,$B8,$7F,$05,$D2,$8F,$D5,$92,$3F,$57,$CF,$F6,$7E,$53,$EF,$9F
	.dc.b	$BB,$FD,$4F,$FD,$6B,$3F,$80,$8F,$AD,$F1,$32,$3F,$07,$E1,$DD,$3F
	.dc.b	$15,$E3,$FE,$E0,$93,$E2,$7F,$AC,$F4,$FE,$43,$D3,$FF,$0C,$67,$F8
	.dc.b	$9F,$91,$FC,$17,$7F,$E4,$1F,$C2,$7C,$2C,$C7,$F2,$3E,$9D,$CF,$F8
	.dc.b	$4F,$9A,$CF,$FC,$29,$1F,$C9,$A3,$1F,$05,$E0,$25,$FF,$0A,$8E,$1F
	.dc.b	$07,$F9,$EF,$57,$F2,$4D,$9F,$CF,$FC,$8F,$E8,$8F,$D1,$FC,$8D,$8F
	.dc.b	$ED,$3F,$A8,$FF,$83,$B2,$FE,$42,$DC,$4D,$35,$C1,$FD,$12,$BF,$C5
	.dc.b	$1A,$FE,$76,$FF,$56,$FF,$8F,$63,$FD,$14,$E7,$87,$E3,$BF,$E4,$A7
	.dc.b	$FD,$C7,$7F,$E1,$89,$DF,$C5,$7F,$83,$23,$17,$F1,$9F,$E1,$73,$F0
	.dc.b	$5B,$1F,$1D,$AF,$F8,$B3,$1F,$4F,$99,$FE,$51,$C5,$F2,$5D,$7C,$CF
	.dc.b	$4B,$71,$F8,$7F,$05,$E1,$3D,$BF,$25,$63,$E3,$FC,$0F,$F8,$6B,$43
	.dc.b	$49,$1E,$5F,$7B,$D4,$7C,$6F,$92,$B1,$F1,$3E,$43,$E4,$9F,$2B,$D4
	.dc.b	$B1,$F5,$3D,$EE,$C4,$7E,$03,$E1,$F7,$BE,$31,$F4,$BE,$3A,$CF,$D0
	.dc.b	$F9,$2F,$AA,$7D,$6F,$0E,$C7,$E0,$3C,$2F,$51,$1F,$98,$FC,$7F,$92
	.dc.b	$FB,$A7,$DE,$F0,$EC,$FD,$CF,$90,$FE,$79,$87,$E0,$9B,$1F,$CD,$63
	.dc.b	$F9,$C1,$1F,$E7,$9E,$D2,$3E,$57,$B9,$F7,$CF,$C5,$F8,$B6,$BF,$2B
	.dc.b	$1F,$8E,$F9,$0C,$76,$FE,$39,$FC,$8F,$8B,$2B,$F1,$FE,$25,$AF,$C9
	.dc.b	$CF,$E5,$BE,$3F,$1E,$17,$E4,$9F,$F1,$C7,$D1,$FE,$13,$E2,$96,$F9
	.dc.b	$2F,$F3,$5B,$89,$D3,$9F,$90,$FF,$3C,$C7,$FA,$92,$77,$28,$DC,$61
	.dc.b	$B3,$C3,$C1,$FF,$23,$D7,$33,$B7,$8D,$FE,$A3,$BE,$13,$A6,$E1,$E6
	.dc.b	$6D,$71,$3C,$15,$C6,$FB,$56,$DB,$3B,$C3,$C6,$F8,$35,$D3,$EF,$C7
	.dc.b	$C1,$5C,$31,$EE,$7A,$F1,$ED,$5B,$C7,$8B,$F0,$6F,$DA,$FF,$90,$31
	.dc.b	$F9,$2D,$8C,$BE,$43,$FC,$8B,$F2,$7F,$E7,$7F,$89,$FF,$01,$36,$3F
	.dc.b	$C5,$2D,$77,$F1,$C9,$FF,$04,$5B,$C4,$79,$58,$6D,$FF,$8C,$A7,$53
	.dc.b	$92,$47,$C5,$9D,$E9,$F6,$5B,$AF,$E0,$3C,$2F,$F5,$CC,$E6,$FE,$0E
	.dc.b	$1D,$30,$E5,$18,$AD,$DF,$E2,$30,$E7,$5F,$8D,$6B,$F8,$96,$DA,$B8
	.dc.b	$CE,$EF,$BD,$68,$DC,$EA,$79,$4E,$88,$FE,$06,$39,$C3,$E8,$9D,$FF
	.dc.b	$ED,$4E,$3F,$EB,$B1,$8B,$FD,$2C,$62,$75,$C5,$FA,$8B,$E5,$E1,$62
	.dc.b	$FE,$8E,$1F,$E4,$47,$8D,$6E,$3A,$47,$F8,$C9,$B8,$78,$2B,$95,$F9
	.dc.b	$7C,$0B,$46,$9D,$FC,$B7,$D1,$FF,$B2,$31,$FF,$62,$B5,$ED,$3C,$1C
	.dc.b	$FC,$0F,$6F,$31,$F2,$9F,$E5,$1C,$BF,$D4,$2D,$8C,$D7,$F8,$41,$AF
	.dc.b	$F9,$F3,$C1,$77,$7F,$96,$63,$FC,$C8,$D8,$CC,$7F,$22,$8E,$9E,$CF
	.dc.b	$84,$91,$F1,$8F,$93,$E1,$FC,$0F,$0B,$C9,$B9,$A4,$9F,$F1,$1C,$76
	.dc.b	$7F,$C1,$BB,$7F,$B5,$9F,$87,$EE,$FC,$AC,$7E,$6E,$5F,$A4,$9F,$F3
	.dc.b	$1B,$78,$BF,$C3,$AD,$E3,$FE,$4B,$C5,$F3,$BD,$DE,$3F,$17,$F1,$EB
	.dc.b	$FD,$2B,$87,$F8,$D1,$AF,$FB,$03,$8B,$FD,$0D,$3E,$AF,$0B,$C3,$F1
	.dc.b	$B2,$BE,$05,$7E,$4E,$7C,$BD,$92,$73,$87,$F2,$ED,$1A,$79,$37,$7B
	.dc.b	$3E,$CF,$94,$F2,$13,$0F,$23,$FE,$8D,$DF,$F0,$FE,$C7,$C2,$F5,$A5
	.dc.b	$7E,$22,$3F,$19,$83,$D7,$E0,$D9,$FF,$19,$35,$FF,$1C,$35,$FE,$96
	.dc.b	$7F,$12,$D3,$FD,$2A,$FC,$5F,$5F,$E4,$11,$FF,$1E,$7D,$13,$CA,$F3
	.dc.b	$36,$33,$18,$3F,$2D,$F4,$6D,$7E,$DF,$CB,$4A,$F9,$73,$97,$E4,$FB
	.dc.b	$5F,$5C,$8F,$D7,$79,$ED,$F8,$4F,$2A,$4F,$28,$FD,$9D,$2F,$F1,$3E
	.dc.b	$CC,$AE,$F1,$F8,$CF,$67,$F0,$84,$8F,$C2,$3E,$CF,$53,$F2,$3F,$25
	.dc.b	$27,$F4,$1C,$8D,$D5,$E0,$D3,$F2,$FF,$03,$2F,$DB,$1F,$C3,$D7,$D1
	.dc.b	$31,$FD,$93,$63,$FD,$98,$D7,$FD,$E6,$D7,$FC,$91,$83,$D4,$FD,$2F
	.dc.b	$B7,$27,$CB,$39,$7F,$27,$ED,$FE,$A8,$8F,$D0,$D7,$F9,$17,$A6,$FA
	.dc.b	$3F,$C0,$CF,$E1,$0F,$EB,$0E,$1F,$C3,$C6,$F9,$FF,$DC,$4D,$8F,$4F
	.dc.b	$B9,$EC,$23,$83,$FD,$47,$C9,$FE,$A7,$7F,$4F,$1F,$BE,$3E,$CF,$07
	.dc.b	$DA,$DE,$CF,$E1,$A3,$0F,$E2,$33,$FE,$0B,$F2,$72,$30,$43,$FD,$D3
	.dc.b	$82,$7F,$0A,$7E,$1B,$0F,$C3,$FC,$9C,$AF,$C9,$BF,$8A,$B9,$1E,$87
	.dc.b	$F3,$64,$7D,$7F,$27,$A7,$BC,$7E,$4A,$3F,$A2,$FB,$9F,$2F,$C3,$FE
	.dc.b	$2F,$7F,$C7,$0C,$DF,$F4,$5F,$9A,$FF,$1A,$BF,$E8,$56,$BF,$C7,$1A
	.dc.b	$BA,$11,$8F,$1D,$83,$E4,$FF,$94,$E5,$30,$FE,$33,$39,$6F,$CA,$79
	.dc.b	$88,$EB,$E5,$3F,$98,$FC,$0C,$8F,$E3,$AF,$E6,$FE,$8B,$ED,$43,$6F
	.dc.b	$E3,$EF,$F9,$5B,$E8,$18,$7F,$4F,$6C,$7E,$34,$D4,$64,$D6,$72,$34
	.dc.b	$27,$D0,$F1,$6B,$FC,$6B,$4F,$83,$EC,$FF,$74,$C7,$E8,$79,$9F,$E0
	.dc.b	$16,$D7,$F8,$C3,$5F,$EE,$19,$FF,$B9,$9B,$5F,$E0,$76,$CE,$EF,$ED
	.dc.b	$FC,$A7,$DC,$F9,$3E,$DF,$0B,$2B,$E6,$C7,$DB,$F1,$BF,$9D,$71,$FF
	.dc.b	$AC,$A7,$FD,$C8,$D7,$FD,$D4,$D7,$D5,$A7,$C6,$7E,$66,$7E,$9E,$0F
	.dc.b	$61,$F8,$CF,$5B,$F8,$67,$D0,$BE,$53,$FB,$E3,$6F,$E4,$B5,$FA,$0D
	.dc.b	$7F,$7D,$C5,$FE,$11,$6E,$FF,$5C,$8F,$D7,$E2,$FC,$6F,$CA,$7C,$DE
	.dc.b	$49,$3F,$8C,$8F,$CD,$77,$7C,$02,$47,$C0,$7F,$C3,$1C,$8F,$63,$EB
	.dc.b	$74,$4A,$FC,$AC,$7E,$8F,$B7,$E4,$92,$3E,$4B,$FE,$19,$FA,$1F,$F5
	.dc.b	$B4,$7E,$19,$AF,$E3,$5A,$FF,$1B,$6B,$FD,$81,$FF,$1B,$3C,$7E,$07
	.dc.b	$A1,$1F,$81,$83,$E4,$7F,$76,$48,$FE,$87,$F4,$3C,$9F,$AE,$7E,$D5
	.dc.b	$C7,$67,$E3,$67,$84,$EB,$BD,$5A,$E7,$FE,$E7,$F4,$1F,$BA,$AF,$D0
	.dc.b	$B5,$FD,$43,$5F,$E6,$CD,$78,$FB,$7F,$E3,$D8,$34,$91,$F6,$FA,$9F
	.dc.b	$35,$FE,$4B,$6C,$66,$30,$7E,$6B,$C1,$91,$F9,$A6,$BF,$E2,$9D,$5F
	.dc.b	$29,$FC,$80,$91,$FD,$A3,$E6,$3F,$63,$A9,$FD,$9F,$47,$F1,$8A,$DD
	.dc.b	$7C,$59,$CE,$5F,$95,$8F,$DD,$74,$F9,$C8,$FF,$0B,$C3,$A7,$77,$F8
	.dc.b	$A6,$FF,$55,$3F,$E1,$E9,$CB,$B8,$CB,$FC,$17,$C3,$FE,$10,$47,$F9
	.dc.b	$3F,$CA,$7F,$3A,$F7,$3F,$32,$48,$FC,$2B,$FE,$73,$FA,$0F,$E4,$31
	.dc.b	$FE,$4B,$6B,$FC,$55,$AF,$FA,$4D,$AF,$FB,$0B,$37,$8B,$FC,$64,$91
	.dc.b	$FE,$50,$9F,$F1,$73,$F8,$ED,$90,$F9,$1F,$11,$27,$F7,$11,$96,$FC
	.dc.b	$6F,$FA,$62,$3F,$DA,$51,$D7,$E3,$B1,$39,$F9,$4B,$3E,$18,$C3,$FE
	.dc.b	$62,$CB,$F8,$67,$5F,$F4,$0C,$91,$8E,$9F,$E6,$1E,$37,$7F,$8F,$FE
	.dc.b	$69,$E4,$7E,$3C,$98,$FE,$4D,$F9,$5F,$29,$FC,$E7,$C6,$93,$F0,$63
	.dc.b	$FD,$55,$F7,$FF,$E1,$D8,$FF,$0B,$73,$A3,$11,$F7,$DA,$FF,$A1,$1A
	.dc.b	$FF,$69,$FC,$BF,$FD,$7A,$FE,$15,$AF,$FA,$29,$AE,$F1,$FF,$54,$7C
	.dc.b	$AB,$E0,$BF,$B0,$4F,$FB,$EB,$E2,$FF,$CD,$73,$9E,$1F,$F6,$84,$62
	.dc.b	$B7,$C3,$FE,$90,$AD,$5F,$13,$FC,$61,$F2,$BC,$1A,$64,$FF,$7D,$94
	.dc.b	$07,$86,$8F,$C1,$1E,$9F,$E4,$23,$0D,$8C,$F1,$FE,$A5,$B7,$F3,$3F
	.dc.b	$18,$FE,$6D,$3F,$AD,$6C,$67,$CD,$E9,$18,$8F,$CF,$7C,$97,$F8,$0B
	.dc.b	$7F,$C2,$73,$FE,$FF,$0F,$CD,$4E,$5F,$EE,$9E,$0D,$BF,$D1,$63,$FA
	.dc.b	$6D,$FF,$16,$DC,$3B,$4F,$F7,$B8,$C7,$CE,$F1,$AF,$C8,$B7,$B7,$FC
	.dc.b	$CD,$8F,$E5,$E3,$1E,$4E,$7E,$52,$CF,$EB,$5C,$3F,$EB,$EF,$A4,$61
	.dc.b	$8C,$BF,$61,$18,$9D,$7C,$97,$9E,$0D,$24,$7F,$2F,$F3,$FC,$DA,$FF
	.dc.b	$4B,$8E,$DD,$18,$9F,$55,$FF,$A9,$E7,$FB,$83,$70,$FF,$6B,$35,$FF
	.dc.b	$10,$35,$FC,$1E,$28,$7C,$0F,$94,$95,$D1,$F9,$9F,$23,$E4,$3C,$3F
	.dc.b	$19,$67,$F9,$C7,$AD,$F8,$9E,$33,$49,$1F,$D2,$FC,$EF,$AC,$47,$07
	.dc.b	$F2,$FF,$3B,$DB,$F1,$5F,$7F,$B3,$FA,$14,$62,$7E,$84,$DD,$BF,$A1
	.dc.b	$4D,$E9,$FF,$26,$DF,$F4,$9F,$2F,$C8,$4C,$3F,$9E,$3F,$87,$F9,$69
	.dc.b	$1F,$D7,$DB,$87,$F9,$47,$B9,$E3,$7F,$32,$6D,$7F,$94,$1B,$3B,$9D
	.dc.b	$6D,$2B,$4D,$9F,$E2,$7F,$46,$D3,$ED,$FC,$47,$BE,$31,$7F,$8D,$8C
	.dc.b	$B4,$FF,$66,$C6,$AB,$F1,$F3,$FA,$55,$F4,$6F,$94,$FE,$92,$DB,$FF
	.dc.b	$33,$B6,$7F,$CD,$CD,$7F,$D7,$8B,$D2,$E2,$49,$E2,$FD,$FF,$32,$47
	.dc.b	$F6,$76,$D7,$E4,$BD,$DF,$CD,$E1,$F9,$86,$D7,$F7,$FF,$0F,$F1,$04
	.dc.b	$7F,$B8,$3F,$A7,$F8,$6F,$D1,$E3,$FA,$06,$CF,$F1,$CE,$4F,$ED,$7F
	.dc.b	$41,$FA,$C8,$FF,$41,$B7,$2F,$F4,$53,$5F,$F6,$C3,$5F,$E4,$A3,$E8
	.dc.b	$DE,$07,$F7,$06,$D7,$FA,$39,$B3,$FE,$95,$6B,$FE,$74,$F5,$BF,$72
	.dc.b	$7E,$DB,$E0,$2D,$7F,$87,$7B,$BF,$4C,$47,$FC,$1C,$FF,$05,$F8,$7F
	.dc.b	$E2,$DF,$A5,$27,$F7,$F1,$87,$F8,$0F,$BB,$1F,$E2,$5F,$11,$72,$FE
	.dc.b	$23,$3C,$BE,$93,$1B,$FF,$71,$C5,$1F,$87,$DD,$E1,$FF,$83,$9B,$97
	.dc.b	$CA,$8C,$19,$F1,$24,$7C,$E6,$E5,$F2,$8F,$E4,$78,$3E,$5B,$D6,$7E
	.dc.b	$57,$E5,$DF,$2B,$8C,$3E,$D1,$FC,$53,$E9,$3F,$D8,$6E,$30,$FE,$5B
	.dc.b	$18,$7F,$C3,$8F,$F5,$C8,$D7,$94,$8D,$6D,$FE,$0F,$7F,$DB,$D3,$AC
	.dc.b	$BF,$C4,$4F,$08,$CD,$FF,$89,$E7,$2F,$F9,$C7,$E8,$FE,$3D,$7F,$92
	.dc.b	$5B,$1F,$E0,$C6,$BF,$E4,$96,$BE,$BF,$85,$FB,$7F,$03,$FD,$47,$0F
	.dc.b	$F3,$5D,$F8,$E3,$1F,$07,$9F,$26,$CF,$95,$C3,$FE,$AD,$AF,$F2,$0B
	.dc.b	$63,$FB,$14,$63,$0F,$A5,$A7,$C3,$F8,$0F,$01,$CA,$3E,$4B,$E3,$F2
	.dc.b	$26,$1F,$0F,$C3,$B5,$F3,$A7,$F2,$7F,$03,$8E,$0F,$C5,$3E,$BF,$27
	.dc.b	$D9,$EB,$5C,$7E,$63,$87,$E8,$7C,$15,$C7,$E7,$3F,$91,$F1,$BF,$20
	.dc.b	$49,$FD,$CB,$FE,$62,$F9,$27,$D2,$3E,$13,$0D,$AF,$C8,$60,$AD,$D6
	.dc.b	$8F,$F0,$37,$3B,$E8,$7F,$5A,$7D,$3F,$5E,$E1,$F8,$FF,$4D,$D4,$F8
	.dc.b	$BF,$D9,$FC,$27,$F5,$38,$C5,$7E,$2E,$3F,$55,$F2,$5F,$A9,$24,$7F
	.dc.b	$12,$7E,$AF,$B1,$FB,$CF,$A3,$FC,$13,$67,$C1,$6F,$A1,$FE,$03,$8F
	.dc.b	$76,$CF,$E6,$76,$F5,$91,$FF,$78,$6D,$F0,$F8,$3F,$03,$D9,$FC,$F7
	.dc.b	$A3,$F8,$07,$83,$F9,$1F,$4F,$F8,$6B,$F9,$3F,$4B,$F8,$7F,$C1,$5C
	.dc.b	$BF,$4C,$FF,$12,$D3,$E4,$92,$3B,$BF,$E3,$1F,$6B,$F4,$78,$FF,$0D
	.dc.b	$FA,$0F,$F1,$74,$62,$BF,$C1,$FA,$A7,$FB,$E7,$D2,$3E,$87,$FC,$A0
	.dc.b	$DA,$FF,$5B,$37,$7F,$F3,$CB,$5F,$E6,$FE,$23,$0F,$F1,$CC,$7E,$81
	.dc.b	$AE,$9A,$F8,$BE,$77,$BF,$99,$30,$EE,$D9,$EF,$F1,$7F,$CC,$BE,$43
	.dc.b	$CE,$DA,$FE,$15,$F0,$FF,$8C,$23,$F4,$9F,$C7,$FD,$1F,$F3,$A3,$DF
	.dc.b	$EF,$D9,$FE,$2F,$C0,$F1,$7F,$9F,$FC,$47,$C8,$6C,$FF,$1C,$F6,$3F
	.dc.b	$C9,$5A,$7E,$F1,$AF,$C2,$6B,$FE,$3C,$6B,$FE,$54,$7F,$AB,$F2,$7F
	.dc.b	$AC,$5B,$3F,$E8,$4F,$09,$C6,$F7,$BF,$CC,$D5,$AC,$3F,$CB,$B9,$BC
	.dc.b	$1F,$B3,$ED,$FD,$E6,$E5,$EC,$EC,$FC,$21,$1F,$DA,$F3,$AF,$FA,$0C
	.dc.b	$63,$1F,$EE,$DA,$F9,$7E,$23,$F1,$3E,$07,$89,$FE,$91,$97,$FB,$8B
	.dc.b	$D5,$F1,$61,$FC,$03,$54,$61,$FE,$06,$D7,$E3,$75,$FF,$A1,$F2,$FC
	.dc.b	$54,$7C,$78,$C1,$F9,$2F,$6B,$F3,$3D,$BF,$96,$6E,$1F,$62,$2D,$34
	.dc.b	$FF,$27,$F1,$2F,$31,$FE,$DF,$DB,$F2,$AD,$E8,$FD,$2F,$67,$E9,$99
	.dc.b	$97,$E1,$1C,$B4,$FE,$C5,$8B,$E9,$3F,$DF,$5F,$0B,$FE,$BC,$F8,$DF
	.dc.b	$D5,$4F,$FA,$FA,$72,$FF,$AF,$F6,$FF,$0E,$C7,$F7,$3B,$E1,$18,$F9
	.dc.b	$1F,$F6,$0D,$6C,$E4,$E2,$FA,$C7,$86,$31,$3D,$3D,$BF,$F0,$B4,$F1
	.dc.b	$35,$8F,$67,$9B,$5C,$9B,$C7,$C8,$DB,$EA,$F0,$31,$3C,$67,$78,$79
	.dc.b	$1F,$7B,$6B,$8D,$76,$8D,$5F,$99,$D5,$F3,$AF,$83,$5F,$E1,$9B,$F9
	.dc.b	$2D,$8C,$BF,$1E,$7F,$21,$F1,$EF,$89,$FE,$A1,$3F,$D4,$A3,$1E,$8E
	.dc.b	$A7,$33,$FE,$25,$9C,$BF,$D6,$30,$7B,$9F,$EC,$A8,$C3,$5F,$23,$5F
	.dc.b	$B6,$D7,$F0,$0F,$1C,$BE,$23,$74,$DF,$57,$C9,$F5,$FF,$CB,$F5,$A8
	.dc.b	$F9,$95,$BB,$E7,$1B,$72,$C7,$EC,$46,$DF,$CA,$57,$4B,$EF,$1A,$AE
	.dc.b	$17,$F5,$EB,$53,$D8,$D1,$23,$4F,$E2,$65,$CE,$FA,$2F,$F3,$BF,$27
	.dc.b	$F1,$BF,$7E,$F1,$FF,$21,$6E,$1F,$C8,$6A,$E1,$CB,$C8,$B8,$F3,$AF
	.dc.b	$E5,$D5,$89,$E3,$E1,$3E,$77,$F3,$B5,$FB,$FC,$BF,$97,$7D,$17,$23
	.dc.b	$73,$AD,$3F,$97,$79,$38,$A9,$7F,$CE,$E3,$C3,$C9,$FC,$67,$5F,$6B
	.dc.b	$75,$FC,$65,$7F,$25,$8C,$7C,$EF,$08,$FE,$89,$1E,$A8,$F5,$F9,$32
	.dc.b	$7F,$A4,$3E,$4E,$2F,$7E,$9D,$49,$3F,$CC,$5F,$07,$B5,$F1,$3E,$86
	.dc.b	$4F,$F5,$18,$F8,$FE,$4C,$8F,$EA,$8F,$CA,$F2,$7E,$5F,$AF,$E9,$24
	.dc.b	$FF,$3D,$7D,$9A,$FC,$6F,$56,$4F,$F5,$D8,$FA,$7F,$0D,$23,$FB,$03
	.dc.b	$F5,$BC,$5F,$B3,$F4,$FF,$CF,$E5,$C4,$8C,$36,$3E,$64,$60,$FD,$EF
	.dc.b	$2F,$F1,$6E,$7F,$B9,$E2,$FD,$83,$AF,$85,$DB,$E6,$3B,$FC,$E4,$C7
	.dc.b	$FD,$1B,$8A,$7C,$7F,$25,$D8,$93,$FD,$7A,$51,$E2,$F8,$8F,$4F,$C9
	.dc.b	$4A,$E7,$1E,$CF,$89,$F6,$FC,$4C,$9E,$F1,$F0,$3E,$27,$D6,$48,$EC
	.dc.b	$FB,$BE,$27,$E2,$FC,$4C,$9F,$3C,$7C,$8F,$89,$FC,$DF,$E9,$DF,$8C
	.dc.b	$9F,$CD,$0E,$24,$67,$8F,$F9,$DF,$27,$CD,$BF,$A3,$F2,$3F,$30,$98
	.dc.b	$7B,$DF,$9F,$F1,$7D,$BB,$A4,$FC,$88,$FB,$3F,$13,$E2,$EF,$93,$F3
	.dc.b	$23,$EE,$FC,$77,$D8,$24,$7C,$B7,$F0,$1C,$DF,$E1,$58,$FE,$C7,$F4
	.dc.b	$8C,$7F,$7F,$08,$C3,$FD,$A2,$78,$3F,$03,$4F,$C4,$7C,$57,$E3,$09
	.dc.b	$1F,$E8,$1D,$DD,$DF,$90,$F6,$3F,$16,$49,$FB,$CF,$E3,$BE,$23,$51
	.dc.b	$F8,$68,$FC,$BC,$7E,$63,$E2,$3F,$33,$F2,$12,$BF,$17,$1F,$C5,$3F
	.dc.b	$16,$FE,$11,$E7,$A7,$F3,$BA,$AB,$94,$7F,$B7,$AB,$6F,$E8,$9F,$F7
	.dc.b	$DB,$9F,$6F,$A1,$D7,$E4,$64,$71,$6B,$F0,$E5,$EE,$78,$FD,$99,$3D
	.dc.b	$DF,$37,$9F,$23,$FC,$77,$48,$FE,$34,$D8,$CC,$63,$E1,$78,$47,$78
	.dc.b	$F8,$11,$FC,$BB,$F4,$E7,$85,$49,$5E,$FF,$90,$F8,$C4,$70,$7C,$3F
	.dc.b	$90,$F9,$5F,$0B,$23,$E3,$3F,$37,$D0,$F9,$DF,$0F,$67,$E5,$46,$1F
	.dc.b	$F1,$6F,$CC,$7F,$87,$BC,$4F,$E7,$ED,$E5,$FA,$92,$8C,$3F,$5F,$B7
	.dc.b	$F0,$EF,$E5,$3D,$1F,$E2,$D1,$FE,$43,$8F,$F2,$B3,$5F,$F6,$BB,$5E
	.dc.b	$A7,$F1,$BF,$53,$F3,$71,$BC,$3F,$A8,$B7,$0F,$1F,$42,$BF,$15,$F1
	.dc.b	$3F,$8B,$F8,$F9,$8F,$E1,$E3,$2D,$21,$F8,$1F,$8F,$FA,$5E,$44,$8F
	.dc.b	$F7,$17,$22,$3F,$2D,$F1,$32,$7F,$24,$D7,$F1,$5E,$B7,$E6,$FE,$3A
	.dc.b	$47,$E5,$DF,$DA,$FD,$17,$DE,$8F,$F2,$DA,$F4,$BF,$9E,$36,$3F,$CF
	.dc.b	$8D,$7F,$DF,$AD,$7F,$D8,$3D,$8F,$8A,$EB,$52,$7F,$4D,$1F,$8D,$F2
	.dc.b	$E4,$7B,$1F,$D8,$7C,$7C,$8F,$D6,$35,$FD,$27,$C7,$FE,$57,$C8,$91
	.dc.b	$F1,$1F,$F0,$5F,$EA,$7F,$E3,$5E,$8F,$E4,$ED,$E7,$FD,$D7,$9A,$8C
	.dc.b	$5F,$F5,$86,$C6,$7E,$5F,$9F,$C8,$FF,$80,$E4,$D2,$57,$FA,$AA,$3F
	.dc.b	$23,$5D,$67,$8C,$FF,$09,$C3,$C7,$F1,$9F,$C9,$3E,$89,$E0,$7F,$69
	.dc.b	$6D,$7F,$AA,$5A,$FE,$48,$38,$7F,$2A,$BC,$1E,$93,$27,$FD,$2B,$39
	.dc.b	$7F,$D2,$FF,$42,$F1,$7F,$BC,$36,$3F,$D7,$0D,$7C,$0D,$7F,$DF,$BA
	.dc.b	$7F,$8D,$5F,$EA,$E2,$38,$65,$FE,$34,$6E,$DF,$42,$B8,$4F,$D4,$F8
	.dc.b	$DF,$98,$4B,$F9,$4F,$CF,$F8,$CF,$B6,$C8,$F9,$57,$CE,$7F,$C8,$7F
	.dc.b	$A7,$FF,$2E,$9F,$E9,$58,$3E,$37,$F0,$66,$BC,$FF,$C2,$FC,$6D,$AE
	.dc.b	$67,$F6,$FB,$E5,$9F,$E2,$27,$F1,$7F,$1B,$F8,$12,$57,$DF,$7F,$09
	.dc.b	$F1,$BA,$8E,$7C,$BF,$8D,$9F,$C9,$CF,$E5,$3E,$36,$56,$B4,$47,$E6
	.dc.b	$3E,$37,$F3,$3E,$E7,$E3,$C9,$5F,$8D,$7F,$25,$F1,$BA,$8E,$51,$FA
	.dc.b	$08,$FD,$0F,$BF,$27,$B3,$FA,$4F,$33,$F4,$BF,$1B,$F9,$B2,$4F,$E6
	.dc.b	$5F,$CF,$7C,$6F,$EA,$BD,$E9,$3E,$78,$FD,$6F,$A5,$FE,$78,$F1,$7F
	.dc.b	$92,$F1,$FF,$85,$1B,$D1,$F7,$E7,$5F,$1B,$30,$F8,$B5,$FE,$40,$AF
	.dc.b	$D9,$BF,$E3,$98,$CF,$C6,$FF,$9C,$60,$D2,$47,$E8,$F9,$7F,$87,$7C
	.dc.b	$67,$8F,$E2,$E4,$F6,$8F,$37,$C6,$73,$24,$71,$C1,$1D,$FD,$B9,$39
	.dc.b	$6B,$FE,$C9,$EA,$6D,$FC,$2B,$C7,$FE,$A3,$F2,$29,$CD,$7F,$98,$A7
	.dc.b	$31,$8D,$11,$8B,$E5,$5F,$D3,$BE,$42,$BF,$EC,$23,$91,$92,$99,$AE
	.dc.b	$6F,$F9,$C3,$3F,$E7,$1B,$FB,$7C,$EF,$F5,$2C,$FC,$E6,$BB,$6D,$7F
	.dc.b	$37,$F8,$EB,$66,$92,$FE,$47,$C6,$9F,$B7,$F1,$D7,$1F,$F1,$D5,$AB
	.dc.b	$19,$FE,$16,$AD,$FE,$03,$E3,$7B,$91,$FA,$8F,$8B,$D8,$DE,$3E,$D6
	.dc.b	$E7,$FD,$37,$E3,$6E,$7F,$D8,$F0,$FC,$6F,$C6,$E3,$D5,$CB,$63,$F2
	.dc.b	$3F,$1B,$C3,$D4,$E2,$D8,$FC,$AF,$C6,$7C,$22,$3F,$87,$7E,$2F,$C5
	.dc.b	$F8,$0F,$0F,$C6,$59,$FE,$03,$FA,$A7,$C5,$6F,$9D,$E8,$8C,$70,$7F
	.dc.b	$97,$D4,$64,$FA,$6E,$27,$F3,$CD,$9F,$F0,$9E,$E8,$C4,$7F,$88,$A3
	.dc.b	$1F,$1B,$DC,$F0,$7C,$6D,$BF,$2F,$C6,$F8,$FE,$37,$C8,$D8,$F4,$C7
	.dc.b	$88,$8F,$89,$F2,$FC,$6F,$D4,$F8,$F7,$A9,$FD,$2E,$BE,$0C,$E6,$31
	.dc.b	$9F,$FA,$72,$B7,$EC,$FC,$33,$45,$BF,$93,$F1,$DF,$1F,$E3,$BE,$43
	.dc.b	$63,$E6,$C7,$C6,$23,$F1,$9F,$93,$EE,$FE,$23,$07,$EC,$DF,$43,$BF
	.dc.b	$FD,$53,$C5,$FD,$C3,$0F,$E8,$35,$A7,$85,$7D,$68,$FF,$20,$B5,$FF
	.dc.b	$27,$B5,$FF,$42,$35,$7C,$EB,$A7,$C4,$BA,$2C,$61,$97,$FD,$F6,$B8
	.dc.b	$C7,$ED,$9B,$19,$AE,$4F,$E1,$3E,$03,$F1,$9F,$19,$F8,$D6,$C7,$E4
	.dc.b	$A3,$F1,$64,$7F,$16,$FE,$3B,$E3,$7F,$28,$7E,$57,$E3,$6C,$FF,$64
	.dc.b	$E2,$FB,$B5,$F6,$CC,$E3,$F6,$F8,$D1,$F9,$DF,$8C,$97,$FA,$2F,$8C
	.dc.b	$B7,$BC,$7F,$08,$63,$87,$FE,$00,$62,$8F,$D3,$7C,$67,$D9,$F9,$DF
	.dc.b	$F1,$7E,$3F,$AC,$8C,$7C,$4C,$8F,$91,$3A,$C7,$FC,$4C,$D7,$FC,$F2
	.dc.b	$D7,$FC,$83,$C9,$FE,$55,$9F,$DB,$F8,$BD,$FE,$27,$F6,$CF,$ED,$FE
	.dc.b	$2E,$4F,$EF,$3E,$2E,$E1,$F0,$73,$FC,$F1,$8D,$BD,$6F,$F1,$6F,$CA
	.dc.b	$5E,$1D,$31,$E9,$5C,$7F,$C0,$F3,$A6,$EF,$9C,$7F,$DC,$93,$ED,$C7
	.dc.b	$F9,$74,$3E,$65,$C7,$F9,$FF,$9E,$7C,$4F,$E8,$67,$39,$7E,$7A,$72
	.dc.b	$D9,$FC,$74,$62,$FF,$C9,$CD,$7F,$D4,$2D,$7F,$D5,$F4,$E9,$EC,$78
	.dc.b	$7E,$26,$E1,$EC,$E3,$F8,$5F,$2F,$F0,$DB,$1F,$03,$E3,$7E,$31,$8C
	.dc.b	$7D,$EF,$C9,$F8,$D9,$3F,$1B,$E3,$6D,$7F,$0B,$ED,$F8,$93,$FE,$F5
	.dc.b	$EE,$FC,$9F,$1E,$B8,$BF,$AA,$46,$27,$E7,$FB,$7F,$57,$8B,$EB,$36
	.dc.b	$3E,$9C,$7F,$64,$E4,$F8,$19,$7D,$A6,$F3,$1A,$38,$7F,$00,$73,$9F
	.dc.b	$B2,$FE,$0F,$E3,$26,$1F,$80,$F8,$CB,$5F,$7A,$7F,$11,$F1,$98,$D7
	.dc.b	$FB,$66,$5D,$E3,$F1,$9F,$19,$F8,$DF,$8C,$B9,$7E,$29,$C3,$F8,$CF
	.dc.b	$8C,$B9,$7E,$29,$FC,$A7,$C6,$7E,$57,$E2,$ED,$7E,$49,$C3,$FE,$2D
	.dc.b	$DD,$1F,$0C,$91,$9F,$8C,$FC,$F1,$FD,$0F,$97,$E1,$47,$E7,$23,$F9
	.dc.b	$97,$CA,$FE,$77,$2F,$F4,$3F,$77,$FB,$96,$BF,$DF,$BB,$FE,$67,$4F
	.dc.b	$D4,$7C,$6F,$D1,$24,$F8,$DF,$F0,$BF,$EB,$8C,$3C,$1B,$FF,$81,$23
	.dc.b	$15,$FC,$67,$44,$E7,$E3,$BC,$DC,$16,$B4,$F1,$F9,$8E,$A4,$8F,$F2
	.dc.b	$76,$08,$F0,$FC,$7F,$F6,$1F,$D5,$3B,$E7,$FE,$44,$BF,$F6,$6D,$FF
	.dc.b	$96,$1F,$E5,$53,$AE,$6F,$B7,$2F,$8F,$F4,$7C,$87,$A5,$B7,$FE,$8A
	.dc.b	$9F,$F2,$AB,$E2,$F9,$0F,$82,$47,$CB,$F2,$1F,$07,$E3,$BE,$13,$63
	.dc.b	$97,$C7,$7C,$5F,$8E,$F8,$CD,$8E,$DF,$1D,$F3,$08,$FC,$07,$E7,$FC
	.dc.b	$77,$CD,$F8,$EF,$9C,$D9,$F3,$7C,$77,$C3,$3E,$91,$5F,$57,$C6,$FD
	.dc.b	$82,$3F,$31,$FB,$7F,$1D,$F6,$4F,$B5,$F1,$D6,$7E,$17,$C0,$FC,$FF
	.dc.b	$8E,$FB,$CD,$8F,$8D,$F0,$5F,$84,$23,$F6,$1F,$C3,$CF,$E1,$4F,$C3
	.dc.b	$7C,$75,$9F,$F5,$BC,$FF,$81,$5F,$C6,$78,$3F,$DD,$FE,$3B,$FC,$9F
	.dc.b	$39,$EA,$FE,$76,$FF,$8C,$FA,$7F,$6A,$BE,$3B,$FC,$D6,$DC,$3F,$D2
	.dc.b	$CD,$AF,$F1,$CB,$5F,$F1,$C2,$F8,$EE,$E7,$F8,$B7,$F4,$EE,$18,$76
	.dc.b	$8F,$1F,$C7,$F9,$3E,$3E,$E3,$EA,$70,$F8,$1F,$19,$1F,$1F,$C7,$F5
	.dc.b	$F8,$FF,$6B,$63,$FC,$01,$8F,$F8,$E3,$1F,$6D,$67,$3F,$CF,$1F,$09
	.dc.b	$F8,$BF,$1F,$F1,$BE,$3E,$E1,$F3,$1C,$3F,$3B,$E3,$ED,$F2,$E1,$FF
	.dc.b	$12,$47,$F8,$DF,$E6,$FE,$17,$93,$FB,$DF,$91,$7C,$07,$EF,$F0,$56
	.dc.b	$94,$E9,$FE,$09,$1F,$3E,$31,$1C,$38,$97,$D6,$FA,$5F,$3B,$DB,$63
	.dc.b	$3C,$9F,$E9,$D9,$CD,$F2,$C3,$C3,$5C,$A3,$13,$AC,$3B,$57,$92,$F9
	.dc.b	$5E,$E7,$33,$E5,$6E,$1E,$18,$E1,$9E,$31,$EB,$7C,$67,$C2,$7C,$08
	.dc.b	$CB,$C0,$C9,$D4,$B5,$F2,$5F,$98,$64,$AE,$5E,$C6,$BE,$4B,$FC,$1B
	.dc.b	$F3,$9F,$E5,$76,$C7,$20,$E1,$E1,$EE,$27,$FC,$BD,$4E,$6F,$F1,$2D
	.dc.b	$71,$8C,$6F,$A9,$CC,$F9,$1A,$FD,$87,$8E,$DF,$61,$B9,$E6,$BE,$5F
	.dc.b	$37,$E2,$6B,$51,$F3,$AB,$77,$F8,$86,$BF,$8E,$AE,$17,$CE,$3E,$4C
	.dc.b	$EB,$0F,$3C,$7C,$CA,$DB,$CE,$FB,$11,$D1,$70,$F1,$BD,$23,$CE,$FE
	.dc.b	$21,$FE,$E3,$F0,$5F,$A1,$E5,$FE,$1F,$F2,$37,$83,$F8,$BF,$4A,$31
	.dc.b	$87,$F9,$43,$D8,$61,$8C,$7F,$BD,$FC,$25,$9C,$D7,$F6,$BC,$BF,$C1
	.dc.b	$F1,$88,$FD,$B2,$9E,$1F,$07,$CF,$2F,$F4,$0D,$7F,$2A,$F5,$72,$FF
	.dc.b	$A2,$7D,$FF,$F2,$BE,$3D,$2F,$33,$FE,$C3,$9C,$EB,$FE,$39,$7F,$91
	.dc.b	$47,$E2,$FE,$09,$18,$8F,$CB,$C6,$39,$FF,$D5,$0D,$8F,$C9,$6B,$FE
	.dc.b	$18,$E6,$F8,$73,$99,$F0,$57,$C1,$9F,$57,$C2,$C9,$FE,$86,$FC,$5F
	.dc.b	$81,$E3,$8B,$93,$BB,$6B,$CC,$D7,$CA,$D7,$FD,$E9,$C2,$F8,$0F,$A1
	.dc.b	$32,$7E,$86,$BF,$D4,$2B,$51,$F2,$7B,$FC,$24,$76,$7D,$DF,$6B,$CB
	.dc.b	$E0,$79,$09,$1E,$47,$CF,$E0,$7D,$7F,$8F,$93,$F5,$63,$DB,$CB,$23
	.dc.b	$EC,$3F,$85,$F8,$0F,$83,$5F,$00,$93,$F0,$1F,$87,$F1,$1F,$83,$F9
	.dc.b	$19,$3F,$81,$8F,$C8,$7C,$D3,$DD,$FC,$44,$49,$F5,$F4,$2B,$51,$F8
	.dc.b	$9F,$8D,$F9,$E4,$7E,$7B,$FE,$E7,$9E,$1E,$D7,$E4,$FE,$36,$57,$4B
	.dc.b	$4F,$8B,$DB,$FC,$C7,$C7,$CA,$D4,$7E,$1B,$E7,$3F,$41,$1F,$D3,$FB
	.dc.b	$91,$88,$FF,$47,$35,$46,$BC,$3F,$26,$2F,$6B,$F4,$DF,$21,$2F,$C9
	.dc.b	$1D,$FC,$0F,$D5,$6F,$27,$FD,$E1,$97,$FA,$E9,$BD,$FF,$9E,$83,$49
	.dc.b	$1F,$A7,$87,$C5,$7E,$FF,$F6,$2F,$E5,$3E,$2F,$FA,$29,$BD,$FE,$A9
	.dc.b	$F8,$1D,$BB,$F9,$19,$8F,$8A,$3E,$27,$C3,$F8,$49,$1D,$DF,$E8,$0F
	.dc.b	$F9,$CF,$E7,$7C,$2E,$1A,$FF,$45,$6B,$FD,$41,$AF,$F8,$C3,$17,$C8
	.dc.b	$7A,$49,$1F,$6F,$DC,$FA,$DD,$FF,$0C,$91,$F0,$5F,$A3,$EE,$6A,$3E
	.dc.b	$5C,$7E,$4A,$3F,$27,$ED,$C9,$FC,$33,$F9,$6F,$72,$47,$E2,$5A,$FF
	.dc.b	$85,$FE,$6F,$F8,$BB,$FE,$FC,$7F,$C8,$0F,$E6,$B2,$FF,$07,$F0,$FC
	.dc.b	$FF,$57,$F9,$85,$6A,$B9,$7B,$B8,$F6,$A4,$7E,$8A,$53,$F6,$7D,$39
	.dc.b	$3D,$5E,$7E,$BC,$8E,$0D,$7C,$DE,$87,$80,$8F,$57,$C7,$EE,$FF,$24
	.dc.b	$F9,$D7,$C3,$7E,$1F,$C2,$97,$FC,$3B,$C3,$56,$A3,$CD,$EF,$71,$B9
	.dc.b	$18,$3E,$17,$C9,$FE,$44,$8F,$B7,$D1,$F8,$7E,$8F,$C4,$64,$7C,$18
	.dc.b	$FC,$C7,$B1,$23,$E6,$3F,$D7,$17,$C8,$7C,$EF,$0B,$FC,$89,$DC,$EC
	.dc.b	$FC,$9D,$23,$0F,$E4,$E3,$31,$88,$FC,$A7,$C7,$75,$8C,$74,$AF,$AD
	.dc.b	$E6,$BE,$DF,$B3,$FA,$DF,$93,$98,$FD,$98,$FB,$FE,$B6,$A3,$07,$DD
	.dc.b	$7E,$87,$C4,$7E,$D7,$CB,$93,$F3,$F9,$13,$F8,$8F,$5B,$73,$98,$C3
	.dc.b	$F8,$FF,$93,$FD,$F1,$1F,$C5,$7A,$9F,$1F,$E7,$7E,$C4,$7F,$8D,$5A
	.dc.b	$FF,$90,$1A,$FF,$9B,$5A,$FF,$BB,$57,$A7,$CF,$A2,$4F,$F5,$21,$3A
	.dc.b	$7D,$5F,$0F,$FC,$47,$D0,$91,$83,$F3,$DE,$9F,$E4,$66,$46,$0F,$85
	.dc.b	$E5,$7F,$19,$F7,$64,$60,$FE,$8F,$F3,$7F,$C6,$9F,$E6,$3C,$0A,$FC
	.dc.b	$E1,$FE,$1F,$9D,$74,$3F,$0F,$7B,$FF,$C6,$23,$0D,$8C,$E1,$FE,$5B
	.dc.b	$9C,$BF,$E0,$CA,$FE,$2A,$FF,$72,$8C,$BF,$EA,$46,$9F,$89,$5E,$DA
	.dc.b	$78,$C7,$FA,$A8,$7C,$F3,$C1,$FC,$F7,$E1,$0F,$52,$9F,$A1,$9D,$57
	.dc.b	$D1,$5E,$1F,$E8,$1B,$1F,$CD,$1A,$FF,$9D,$5A,$FF,$A0,$3E,$7D,$E4
	.dc.b	$FE,$A1,$B1,$FC,$F1,$AF,$F9,$F5,$AF,$1F,$1F,$F6,$91,$AD,$7E,$04
	.dc.b	$A3,$1E,$1F,$FA,$E5,$57,$F1,$1F,$4F,$FC,$73,$C6,$69,$2F,$E8,$79
	.dc.b	$5B,$7F,$4D,$1F,$76,$3E,$F7,$91,$FD,$7F,$D8,$95,$F8,$68,$FF,$20
	.dc.b	$FC,$EF,$F2,$C8,$C1,$0F,$F6,$DF,$98,$8D,$46,$1F,$F2,$83,$FE,$03
	.dc.b	$6A,$3F,$C0,$2E,$1B,$F0,$BF,$E1,$5E,$5F,$EF,$51,$CE,$3C,$67,$E5
	.dc.b	$7C,$5F,$EE,$DF,$0B,$27,$FA,$53,$F9,$9E,$FF,$CD,$76,$F3,$24,$FE
	.dc.b	$4D,$ED,$A6,$A3,$F4,$B1,$FA,$18,$FF,$01,$7B,$52,$7F,$AB,$3F,$A5
	.dc.b	$3F,$4D,$F3,$1E,$72,$47,$E7,$DF,$57,$9B,$FA,$BE,$A9,$3F,$B2,$8F
	.dc.b	$F4,$67,$C0,$FF,$B2,$BA,$3F,$41,$DD,$F9,$FF,$03,$F6,$46,$AF,$FC
	.dc.b	$29,$8C,$7F,$B4,$3C,$BE,$37,$FC,$31,$97,$F8,$F7,$DA,$FD,$15,$7F
	.dc.b	$9F,$1A,$FF,$B6,$30,$46,$23,$EA,$1F,$B0,$C9,$F2,$DF,$E2,$3C,$A4
	.dc.b	$FF,$94,$27,$2F,$F7,$9D,$3C,$04,$7C,$0F,$8F,$D1,$91,$E3,$6B,$FE
	.dc.b	$E1,$F8,$E7,$A4,$DF,$F0,$DA,$3E,$03,$CB,$5A,$9C,$C6,$2B,$85,$FF
	.dc.b	$08,$8C,$4E,$A3,$FD,$EF,$F1,$95,$FF,$6E,$1B,$32,$53,$38,$78,$5F
	.dc.b	$E2,$7F,$BF,$FE,$67,$C3,$FF,$53,$B6,$3E,$63,$5F,$97,$28,$C4,$7B
	.dc.b	$17,$17,$F0,$08,$C3,$5E,$ED,$9E,$4D,$DB,$FC,$CD,$7F,$E8,$2B,$7C
	.dc.b	$CF,$F2,$9C,$BF,$17,$5A,$C7,$E6,$5F,$86,$B4,$F8,$FE,$2B,$B6,$1F
	.dc.b	$E8,$F8,$F6,$C6,$0C,$FC,$C7,$0B,$F8,$6D,$CD,$8F,$F7,$2B,$F2,$4E
	.dc.b	$61,$E7,$5F,$84,$C3,$6B,$F1,$9E,$7E,$FE,$3B,$83,$63,$F2,$1E,$3F
	.dc.b	$C1,$23,$FE,$D8,$C3,$F9,$13,$F5,$3A,$3C,$9B,$7E,$6A,$3F,$31,$F0
	.dc.b	$B8,$9F,$47,$CE,$FF,$92,$27,$FD,$56,$BC,$4C,$FC,$1E,$99,$3F,$EE
	.dc.b	$69,$D3,$63,$EB,$46,$38,$7F,$07,$F1,$9E,$23,$BF,$C3,$5B,$EF,$F0
	.dc.b	$7F,$2F,$CB,$FE,$54,$F9,$0C,$79,$7F,$EE,$99,$FF,$26,$37,$0F,$F7
	.dc.b	$DB,$5F,$E5,$D3,$F8,$9A,$E7,$5F,$BC,$9F,$E6,$11,$FC,$DF,$CF,$F8
	.dc.b	$27,$62,$CF,$C2,$7E,$4F,$B3,$8A,$FC,$49,$FA,$27,$E4,$FA,$DE,$6F
	.dc.b	$56,$DF,$99,$C3,$E1,$AF,$E7,$F1,$E5,$CB,$FB,$8C,$E7,$1F,$E9,$53
	.dc.b	$FA,$1A,$7D,$07,$F9,$2E,$BF,$C5,$0F,$D7,$8F,$F0,$13,$96,$F0,$7E
	.dc.b	$55,$E2,$F3,$8E,$11,$F0,$DA,$FF,$20,$6B,$FD,$57,$E8,$1D,$1E,$36
	.dc.b	$DF,$FA,$E1,$B3,$ED,$6B,$F3,$D7,$27,$8F,$F0,$4F,$B1,$E0,$7F,$7E
	.dc.b	$9C,$DF,$F9,$0E,$72,$FF,$20,$E4,$7C,$17,$E1,$7C,$67,$99,$F9,$88
	.dc.b	$C3,$2B,$E5,$4E,$9B,$1F,$83,$8C,$67,$FD,$06,$DF,$07,$F8,$EC,$3F
	.dc.b	$46,$DC,$3F,$1F,$1F,$AA,$F7,$F1,$7F,$82,$F0,$FF,$8E,$47,$EA,$BE
	.dc.b	$0F,$F2,$FF,$23,$70,$FC,$BB,$87,$F8,$B7,$CD,$5C,$39,$78,$08,$FB
	.dc.b	$1F,$43,$FE,$DA,$AF,$F3,$32,$F8,$DC,$FC,$76,$99,$3F,$EE,$A9,$D3
	.dc.b	$63,$F1,$51,$8C,$3F,$9A,$FC,$E7,$F0,$43,$F8,$0F,$8B,$6F,$F8,$0F
	.dc.b	$C2,$FE,$C3,$E1,$FF,$82,$B6,$3F,$63,$1F,$C3,$3E,$4B,$18,$7A,$9F
	.dc.b	$E4,$5F,$41,$FE,$58,$AF,$F0,$F7,$57,$F7,$C8,$F6,$57,$C9,$74,$DD
	.dc.b	$78,$9F,$C6,$3E,$76,$DF,$F3,$D7,$F3,$9F,$40,$F8,$1B,$F1,$98,$6D
	.dc.b	$7F,$A3,$65,$3F,$AB,$6C,$7E,$CF,$04,$61,$FD,$B3,$5F,$EB,$3C,$7F
	.dc.b	$CE,$DF,$AF,$F1,$BF,$AD,$FA,$1F,$C4,$DF,$F5,$18,$F8,$3F,$27,$F0
	.dc.b	$BE,$22,$CF,$C2,$3F,$BD,$EC,$8F,$E6,$BF,$3F,$27,$58,$FF,$A0,$2B
	.dc.b	$51,$FD,$76,$7E,$6F,$C9,$FF,$27,$EE,$F9,$7C,$5F,$36,$31,$C5,$2F
	.dc.b	$FB,$27,$CF,$DC,$BF,$1E,$FD,$6E,$29,$3E,$0E,$F6,$9F,$E1,$28,$C3
	.dc.b	$F3,$DB,$C7,$F7,$8E,$38,$7D,$B7,$FA,$7F,$AF,$27,$67,$F7,$57,$F0
	.dc.b	$BF,$11,$F8,$6F,$87,$B3,$D9,$C3,$FD,$EF,$DA,$B8,$7E,$99,$FC,$6F
	.dc.b	$C3,$C9,$F2,$FC,$4D,$AF,$C6,$79,$3F,$92,$31,$C5,$F9,$17,$F2,$BF
	.dc.b	$0D,$F2,$75,$B8,$7D,$0F,$4B,$F8,$7C,$66,$3F,$9A,$72,$FE,$0F,$1F
	.dc.b	$E9,$2D,$55,$C2,$7F,$C4,$7E,$EF,$F7,$03,$F4,$7F,$0D,$63,$F2,$99
	.dc.b	$7E,$4C,$8F,$E4,$FB,$BF,$8A,$E1,$FE,$52,$C3,$F9,$0F,$07,$F5,$BC
	.dc.b	$8D,$25,$F9,$B7,$FE,$FF,$E8,$FE,$15,$B1,$ED,$AF,$F6,$23,$FE,$27
	.dc.b	$F4,$7F,$15,$F1,$1F,$E0,$67,$F1,$9A,$F7,$24,$FC,$27,$FB,$67,$D0
	.dc.b	$FF,$8F,$F8,$BF,$CE,$FE,$17,$F8,$43,$6F,$F5,$1C,$E6,$FF,$C7,$DF
	.dc.b	$01,$FE,$25,$FA,$1F,$F0,$95,$7F,$29,$FD,$AF,$FA,$C6,$3F,$E8,$09
	.dc.b	$CB,$FE,$82,$8C,$DF,$F7,$8C,$1F,$07,$E2,$24,$7F,$69,$8C,$BF,$DB
	.dc.b	$FB,$58,$3D,$2F,$F1,$B6,$1F,$DA,$71,$F0,$46,$3D,$9F,$37,$9D,$AF
	.dc.b	$80,$B9,$7E,$97,$04,$7F,$85,$F3,$57,$C4,$25,$78,$DF,$F2,$C6,$5F
	.dc.b	$C2,$72,$F6,$7C,$17,$C3,$F2,$BC,$6F,$3F,$23,$CB,$F3,$17,$1F,$03
	.dc.b	$DF,$C4,$FA,$24,$8F,$80,$FD,$4E,$FF,$21,$F3,$DB,$1F,$9D,$D3,$FB
	.dc.b	$7F,$1F,$D1,$F8,$2F,$57,$47,$C4,$7F,$C6,$5F,$45,$FE,$0B,$9F,$C9
	.dc.b	$35,$FF,$14,$B5,$FD,$0B,$7B,$CF,$CD,$C3,$7C,$9F,$81,$C3,$AB,$FE
	.dc.b	$67,$EE,$D6,$0E,$7F,$F0,$7F,$2E,$6B,$4F,$90,$FF,$1E,$63,$F4,$B1
	.dc.b	$FF,$14,$36,$FF,$C7,$D9,$78,$AB,$FD,$11,$CC,$F9,$3F,$C0,$9E,$1E
	.dc.b	$3B,$8F,$F2,$1D,$3F,$C2,$38,$FB,$36,$F3,$1F,$10,$93,$E0,$7C,$64
	.dc.b	$7C,$7F,$09,$ED,$3E,$59,$5E,$CF,$E4,$8C,$4F,$E8,$DF,$89,$F0,$7F
	.dc.b	$17,$E0,$AD,$79,$5C,$3F,$93,$F9,$FB,$7F,$AB,$7F,$C8,$1F,$43,$FE
	.dc.b	$79,$F8,$B4,$BE,$33,$FC,$14,$E2,$74,$E7,$DD,$FF,$70,$C7,$CB,$9D
	.dc.b	$CA,$37,$18,$6C,$F0,$DB,$FD,$77,$5C,$CE,$DC,$2C,$BC,$78,$71,$AD
	.dc.b	$B7,$1C,$E1,$CD,$B8,$72,$BA,$F8,$EF,$8D,$6E,$3C,$06,$6B,$78,$FA
	.dc.b	$E7,$58,$77,$AD,$CF,$A9,$EC,$62,$FA,$BF,$1D,$F6,$BF,$E8,$B8,$F9
	.dc.b	$2D,$8C,$BE,$33,$FD,$21,$F0,$7F,$E1,$7F,$A2,$FF,$08,$36,$3F,$E0
	.dc.b	$07,$A6,$9C,$B6,$3F,$D1,$16,$ED,$FD,$34,$61,$AF,$8D,$AF,$DA,$6B
	.dc.b	$F7,$DE,$38,$FD,$96,$E5,$BE,$9F,$8F,$C7,$F8,$8B,$F8,$B5,$A7,$94
	.dc.b	$FC,$C8,$C4,$7E,$16,$78,$5F,$5A,$D1,$27,$E7,$CE,$AF,$C1,$1D,$A7
	.dc.b	$55,$F8,$DA,$D4,$F8,$0D,$12,$34,$FE,$1A,$5C,$CE,$3F,$57,$81,$FD
	.dc.b	$2F,$92,$F1,$7E,$92,$F7,$18,$C3,$F8,$96,$1C,$5F,$E8,$18,$66,$BA
	.dc.b	$B6,$75,$3F,$D6,$FE,$8B,$97,$4B,$F2,$DF,$D4,$E7,$FD,$6F,$CB,$E0
	.dc.b	$6A,$33,$A7,$FB,$51,$B8,$7F,$83,$1B,$7C,$F7,$F0,$7C,$2F,$9E,$9F
	.dc.b	$3D,$FE,$4B,$7F,$CB,$8D,$7F,$DA,$0D,$7F,$DC,$4D,$7E,$16,$8F,$85
	.dc.b	$E9,$3C,$74,$FD,$C5,$7F,$0A,$8C,$67,$C2,$BE,$17,$C3,$4A,$FE,$CC
	.dc.b	$FC,$5F,$A3,$F7,$7C,$C7,$F9,$BF,$5F,$88,$DD,$8D,$3F,$CE,$9C,$5F
	.dc.b	$6B,$E1,$3E,$71,$31,$FE,$E5,$F0,$DF,$27,$D8,$91,$FE,$39,$C3,$FC
	.dc.b	$FE,$FF,$77,$F8,$6F,$9D,$EE,$F7,$24,$7F,$BB,$E5,$1E,$0E,$27,$D1
	.dc.b	$7D,$D2,$57,$F9,$97,$D6,$FD,$7F,$27,$C4,$DF,$EF,$7B,$3F,$6C,$95
	.dc.b	$BF,$94,$FB,$DE,$D7,$BC,$91,$E1,$7E,$17,$B9,$F8,$32,$47,$7F,$82
	.dc.b	$FC,$86,$5F,$82,$D9,$EF,$7E,$24,$CF,$67,$E2,$FE,$5B,$E6,$13,$0F
	.dc.b	$73,$F3,$2D,$F0,$DF,$96,$F6,$3F,$7D,$4E,$0F,$64,$61,$FC,$0F,$42
	.dc.b	$F8,$D7,$C3,$F8,$6F,$2F,$C3,$48,$C1,$FE,$6B,$C3,$FC,$6E,$F6,$3F
	.dc.b	$CC,$EE,$1F,$EE,$ED,$7F,$CC,$0D,$7F,$6D,$5F,$86,$7E,$97,$5F,$C4
	.dc.b	$3E,$2F,$CB,$7A,$FE,$4B,$E3,$F9,$92,$BC,$51,$EF,$F8,$DF,$C9,$63
	.dc.b	$FE,$EF,$9F,$27,$42,$31,$1E,$49,$DF,$BF,$F1,$C9,$1F,$83,$EE,$F9
	.dc.b	$5F,$17,$F5,$BB,$A4,$FC,$18,$F9,$DE,$FF,$80,$91,$E0,$C5,$1E,$4F
	.dc.b	$8A,$FA,$84,$9F,$C8,$7C,$C7,$E0,$BB,$E4,$7E,$4D,$FE,$29,$F1,$9F
	.dc.b	$9A,$8F,$F4,$D7,$DA,$BC,$EF,$D8,$36,$3D,$ED,$7F,$C4,$ED,$7F,$A0
	.dc.b	$74,$7D,$FF,$8B,$FE,$37,$3B,$EA,$FD,$03,$7C,$CF,$AB,$3F,$96,$F0
	.dc.b	$FF,$2F,$E1,$CC,$7F,$51,$19,$69,$0F,$C8,$57,$DE,$F8,$AF,$CF,$F8
	.dc.b	$52,$7F,$60,$FE,$8B,$B6,$46,$08,$72,$3F,$C9,$4F,$E5,$DF,$DC,$78
	.dc.b	$3F,$CE,$7F,$76,$FC,$E6,$1F,$D0,$5B,$3C,$27,$5E,$DF,$EC,$19,$1F
	.dc.b	$EE,$59,$FF,$00,$ED,$FA,$45,$F3,$3C,$27,$57,$DE,$3F,$A1,$FE,$47
	.dc.b	$FE,$88,$9E,$5F,$35,$E3,$33,$EF,$F9,$7E,$32,$CF,$FB,$4E,$B8,$3E
	.dc.b	$DE,$46,$5F,$A9,$CF,$C5,$F3,$5E,$DF,$9A,$97,$E7,$8F,$81,$E2,$48
	.dc.b	$F5,$3F,$0F,$E3,$BF,$9F,$F6,$49,$FF,$23,$E1,$FD,$01,$FE,$9B,$1F
	.dc.b	$D3,$FE,$8C,$DE,$DF,$DE,$4F,$F9,$3A,$7F,$DA,$AE,$1A,$FF,$01,$6B
	.dc.b	$FC,$43,$E9,$9D,$3C,$E3,$C3,$CD,$FE,$11,$AE,$38,$F8,$EF,$75,$F4
	.dc.b	$FC,$6F,$B8,$47,$EB,$78,$DF,$7C,$C3,$F2,$78,$3F,$7B,$7F,$66,$7E
	.dc.b	$AF,$BB,$F8,$5F,$96,$97,$F4,$63,$F1,$1E,$36,$A3,$07,$E0,$DF,$C6
	.dc.b	$78,$D2,$3F,$0C,$FF,$75,$FA,$47,$2F,$C9,$F0,$BE,$3D,$37,$46,$1B
	.dc.b	$1F,$26,$3E,$6F,$C2,$BE,$F1,$F1,$DC,$EB,$E3,$9F,$46,$BF,$C3,$55
	.dc.b	$B9,$B5,$F8,$AD,$FF,$31,$E5,$7E,$8C,$CF,$77,$E9,$7C,$AB,$97,$4C
	.dc.b	$7C,$13,$98,$FC,$A7,$BF,$FA,$BF,$91,$91,$83,$F4,$DF,$2F,$FA,$09
	.dc.b	$91,$83,$CE,$3F,$2E,$7A,$BF,$11,$B1,$F8,$C6,$BF,$D2,$5A,$FF,$16
	.dc.b	$FD,$23,$FC,$63,$AF,$E5,$E7,$6D,$AF,$F1,$C4,$EB,$B3,$FD,$47,$E9
	.dc.b	$BE,$9B,$F1,$B1,$FE,$43,$FC,$87,$F4,$18,$7F,$60,$6C,$70,$9D,$E4
	.dc.b	$BE,$5E,$0F,$F9,$02,$73,$18,$9F,$AB,$E3,$7F,$4D,$F8,$CF,$F1,$C7
	.dc.b	$D3,$BE,$2F,$F0,$6D,$AF,$D5,$B5,$FE,$D0,$D7,$FB,$BB,$FC,$C5,$FF
	.dc.b	$5E,$BB,$D3,$F7,$0D,$CF,$FC,$F4,$EF,$17,$94,$F8,$6F,$F1,$A3,$70
	.dc.b	$FE,$C6,$D7,$FD,$D6,$D7,$FA,$2F,$88,$F9,$8F,$E3,$B5,$27,$F9,$34
	.dc.b	$7F,$20,$F8,$7F,$E6,$1F,$29,$27,$F8,$E4,$65,$A4,$3F,$6D,$F2,$FF
	.dc.b	$CA,$3E,$1F,$CA,$48,$F2,$BF,$E4,$6C,$5F,$35,$B9,$FE,$51,$3F,$CF
	.dc.b	$E3,$FC,$F1,$FB,$9F,$DA,$97,$1A,$30,$DE,$0D,$F8,$5C,$3B,$BF,$CA
	.dc.b	$75,$FC,$CA,$FF,$B2,$B5,$FF,$45,$35,$EA,$BE,$57,$94,$7E,$0A,$3F
	.dc.b	$65,$7C,$27,$CB,$CD,$2B,$4F,$A3,$E4,$FD,$3F,$31,$E2,$24,$FF,$8A
	.dc.b	$F3,$F5,$47,$9B,$E6,$3D,$E4,$9F,$07,$CC,$7C,$22,$38,$3C,$9E,$77
	.dc.b	$C6,$EB,$91,$E7,$7E,$4F,$C1,$7C,$AF,$95,$F8,$64,$9F,$16,$E8,$F8
	.dc.b	$FF,$2B,$F4,$09,$3F,$07,$C5,$FA,$7F,$29,$F3,$09,$1C,$9F,$9F,$F2
	.dc.b	$92,$3A,$B5,$FF,$25,$7C,$C3,$E3,$3F,$CB,$11,$F7,$3E,$4F,$FD,$B1
	.dc.b	$39,$F1,$3F,$D7,$D1,$8A,$DF,$1A,$31,$3A,$BF,$F3,$1F,$CB,$D7,$FD
	.dc.b	$B6,$75,$32,$53,$35,$CD,$FC,$0F,$85,$FD,$A5,$FC,$2B,$5C,$F4,$FC
	.dc.b	$88,$C4,$FE,$4F,$C0,$BC,$26,$92,$BE,$7F,$CB,$7F,$AD,$32,$FF,$3A
	.dc.b	$D2,$31,$3F,$94,$EF,$B9,$7E,$52,$6E,$2F,$DE,$FF,$47,$7F,$8B,$6D
	.dc.b	$7F,$A1,$27,$47,$9A,$33,$1F,$E4,$67,$E5,$4F,$F1,$E6,$C7,$F8,$5D
	.dc.b	$7C,$B7,$81,$CE,$5F,$E8,$C5,$F2,$B9,$35,$F2,$B7,$0E,$9F,$2B,$C4
	.dc.b	$C6,$3D,$5E,$9E,$87,$94,$93,$EB,$7D,$1F,$2B,$E1,$3C,$5C,$96,$7D
	.dc.b	$1F,$2B,$E5,$F9,$5F,$33,$63,$D9,$1F,$04,$8F,$C7,$7E,$27,$CA,$FB
	.dc.b	$4F,$77,$C3,$D9,$F8,$9E,$9F,$D2,$F4,$5F,$41,$F0,$A5,$B7,$F1,$4E
	.dc.b	$15,$EE,$B5,$3F,$D3,$DF,$9A,$F0,$31,$D7,$F8,$DC,$3A,$B6,$77,$3F
	.dc.b	$D8,$1F,$9F,$87,$F8,$22,$BC,$7C,$5F,$57,$DE,$F0,$32,$7E,$E7,$C4
	.dc.b	$5C,$3E,$E3,$F6,$FE,$53,$CD,$C9,$6B,$A6,$48,$CC,$7F,$AB,$A3,$0D
	.dc.b	$8F,$31,$FA,$5B,$7D,$1F,$F4,$D9,$FF,$03,$AF,$53,$F3,$73,$AF,$56
	.dc.b	$CE,$6B,$75,$CE,$31,$5F,$ED,$16,$BF,$EE,$06,$BF,$E0,$4F,$D3,$1E
	.dc.b	$B7,$D7,$6D,$FF,$7A,$C9,$F1,$39,$A9,$3F,$DF,$BD,$B7,$91,$8C,$7F
	.dc.b	$CF,$11,$89,$FF,$01,$FE,$68,$F0,$FD,$7F,$99,$FF,$A2,$A7,$31,$8F
	.dc.b	$93,$91,$FE,$6E,$9D,$61,$FE,$EB,$6B,$FE,$FA,$6B,$F8,$CF,$13,$F7
	.dc.b	$DF,$31,$F8,$BF,$0B,$F6,$4D,$BF,$AF,$1F,$B5,$F9,$8C,$5F,$EE,$5F
	.dc.b	$DC,$7C,$C7,$F0,$AF,$86,$FE,$D3,$E3,$63,$0F,$F2,$36,$48,$D3,$63
	.dc.b	$F8,$27,$CC,$62,$7F,$81,$3A,$EC,$FB,$7F,$47,$FE,$82,$AE,$71,$8E
	.dc.b	$D9,$1F,$E8,$E9,$D6,$9F,$A1,$3F,$52,$E1,$C3,$5F,$85,$E5,$FE,$BA
	.dc.b	$7F,$C6,$6D,$8F,$F6,$41,$FB,$4C,$7F,$92,$C6,$7C,$A5,$72,$F9,$9F
	.dc.b	$DA,$78,$FA,$6C,$7E,$82,$38,$F2,$63,$1F,$F5,$C6,$1F,$63,$0E,$71
	.dc.b	$FE,$12,$87,$CC,$F8,$3E,$67,$C2,$DC,$7C,$F1,$6B,$9C,$DB,$FF,$5B
	.dc.b	$7C,$F7,$F3,$87,$FD,$71,$AB,$C4,$FD,$3C,$E7,$1F,$E1,$13,$97,$FC
	.dc.b	$E6,$DC,$3E,$60,$70,$FC,$CA,$FD,$B4,$62,$3F,$89,$57,$DF,$7E,$5B
	.dc.b	$6F,$C7,$CF,$E2,$DF,$E6,$FC,$D4,$CB,$E9,$FC,$95,$C7,$E3,$FC,$D7
	.dc.b	$D2,$F9,$AF,$A6,$DA,$FB,$3F,$33,$FC,$97,$AF,$EB,$B6,$3F,$88,$C7
	.dc.b	$DA,$F0,$F1,$87,$D2,$7E,$EF,$CC,$FD,$EF,$99,$B7,$F8,$47,$0F,$E1
	.dc.b	$BE,$66,$DF,$DA,$7F,$0A,$7E,$1B,$E5,$AC,$FE,$31,$FC,$4F,$CB,$7E
	.dc.b	$2B,$E5,$AC,$FE,$41,$C3,$4F,$C9,$C6,$2F,$F0,$FF,$2D,$F9,$23,$F2
	.dc.b	$7E,$7D,$AF,$CC,$C7,$E5,$BC,$FC,$57,$E3,$9F,$CC,$F9,$9F,$D3,$FE
	.dc.b	$CD,$F0,$58,$C3,$FA,$04,$61,$B1,$FD,$26,$31,$7F,$AD,$F5,$11,$88
	.dc.b	$FE,$A1,$18,$9F,$F0,$CF,$81,$FE,$72,$BF,$D6,$FC,$8F,$F8,$3B,$EC
	.dc.b	$9F,$11,$FA,$C8,$FE,$C5,$F3,$6E,$9A,$7F,$85,$3F,$68,$FD,$7E,$1F
	.dc.b	$D5,$67,$4D,$EA,$FE,$BD,$06,$92,$3F,$33,$A2,$7F,$9B,$C7,$FA,$CA
	.dc.b	$BF,$DD,$2F,$5E,$9F,$07,$CA,$79,$3A,$64,$EF,$E5,$2F,$0B,$C2,$FD
	.dc.b	$CE,$7D,$7A,$BC,$64,$BF,$0E,$88,$F3,$F4,$FB,$7E,$53,$E0,$B6,$BC
	.dc.b	$9F,$29,$F0,$4F,$85,$F2,$96,$3E,$17,$57,$C0,$23,$D9,$FF,$24,$F0
	.dc.b	$BE,$53,$E5,$7C,$A7,$CB,$6D,$7C,$1F,$94,$FF,$09,$7C,$1F,$CE,$E1
	.dc.b	$F6,$4F,$F4,$7C,$15,$A8,$DF,$17,$F8,$36,$CD,$24,$7F,$A4,$E9,$F2
	.dc.b	$9D,$B7,$FB,$8D,$9F,$F0,$85,$B5,$F1,$9F,$81,$F9,$4B,$3A,$D7,$C0
	.dc.b	$47,$EE,$D2,$3C,$99,$FB,$0F,$C5,$75,$DA,$FC,$07,$CA,$7E,$34,$FC
	.dc.b	$77,$CA,$58,$F1,$CF,$B4,$8F,$67,$E0,$4F,$C8,$3F,$2D,$F2,$96,$7F
	.dc.b	$19,$F0,$5F,$99,$3F,$35,$E1,$58,$F7,$E1,$F2,$48,$FA,$9F,$B3,$F2
	.dc.b	$5F,$A1,$3F,$45,$F2,$56,$7F,$31,$E2,$7F,$3A,$F8,$AF,$D3,$36,$3F
	.dc.b	$C9,$3D,$DF,$D0,$A7,$FC,$55,$E9,$FF,$90,$3A,$3F,$C9,$AD,$4E,$7B
	.dc.b	$FF,$C9,$F9,$7E,$10,$FD,$97,$C1,$DC,$7E,$BC,$7F,$A2,$3F,$79,$FE
	.dc.b	$95,$82,$F8,$D7,$88,$FF,$08,$4E,$BE,$00,$FF,$96,$9B,$97,$3F,$92
	.dc.b	$E8,$CA,$E5,$87,$FB,$DA,$FF,$CA,$EF,$E5,$8C,$4F,$8F,$E0,$3F,$C8
	.dc.b	$7F,$47,$FE,$EF,$F8,$0F,$F3,$5F,$CA,$FF,$A3,$2F,$FC,$DB,$39,$51
	.dc.b	$95,$1F,$9C,$C7,$8D,$FF,$9B,$63,$85,$6D,$B1,$9E,$F6,$1C,$30,$EF
	.dc.b	$E0,$FF,$7E,$C3,$85,$6A,$7C,$37,$C6,$F8,$47,$8A,$75,$5D,$EF,$95
	.dc.b	$EE,$B8,$9E,$CA,$DC,$66,$B8,$63,$D3,$1F,$3D,$EF,$1C,$61,$C2,$3C
	.dc.b	$F5,$89,$91,$FE,$02,$AF,$90,$FF,$8B,$DA,$F8,$CF,$F1,$97,$CB,$3E
	.dc.b	$4D,$F0,$DF,$E0,$A9,$4E,$63,$0D,$8F,$F7,$DC,$F0,$7C,$1E,$3B,$87
	.dc.b	$96,$08,$C4,$EA,$73,$18,$8F,$33,$AD,$3E,$C3,$75,$FB,$8D,$9F,$C2
	.dc.b	$F8,$7F,$EA,$BA,$DD,$FC,$9A,$D5,$7E,$19,$B8,$7F,$19,$8D,$4C,$7E
	.dc.b	$A9,$B7,$E9,$5F,$5C,$3C,$31,$BC,$3B,$D7,$0B,$E8,$DC,$3B,$92,$B9
	.dc.b	$C7,$E3,$D7,$D1,$BC,$4F,$F6,$27,$4F,$F9,$3A,$73,$18,$E0,$FF,$42
	.dc.b	$56,$EF,$F5,$F5,$CB,$C5,$D4,$7D,$19,$FE,$76,$E9,$E0,$FE,$CE,$7A
	.dc.b	$B9,$9E,$31,$E4,$F0,$6D,$1A,$76,$E8,$7D,$1F,$EC,$F3,$FE,$DA,$DC
	.dc.b	$BC,$F5,$C3,$2F,$F6,$C4,$E5,$FE,$0D,$EF,$FE,$3F,$2F,$E1,$ED,$E9
	.dc.b	$FF,$4A,$4F,$F9,$72,$72,$FE,$57,$B7,$FD,$AC,$DF,$91,$FC,$C3,$FC
	.dc.b	$99,$B1,$A8,$FF,$48,$4F,$F1,$C9,$F7,$D6,$73,$FA,$0D,$CF,$E7,$F8
	.dc.b	$7F,$0B,$C3,$F3,$6F,$FE,$34,$CB,$C8,$D7,$AB,$5F,$F7,$F6,$5F,$7B
	.dc.b	$BF,$E4,$6C,$F9,$7A,$7C,$E8,$F1,$FC,$64,$BF,$BB,$A2,$38,$FC,$B7
	.dc.b	$A2,$FE,$31,$2B,$E3,$57,$57,$CB,$E1,$FB,$BE,$5E,$47,$E0,$E3,$E0
	.dc.b	$F6,$C8,$EE,$F9,$BC,$CF,$8B,$8F,$D3,$24,$F6,$7F,$83,$73,$FE,$0B
	.dc.b	$E1,$FE,$5F,$C5,$CA,$FB,$F1,$F0,$3E,$0F,$7C,$06,$92,$7D,$3F,$2F
	.dc.b	$F8,$12,$38,$3F,$C8,$BF,$07,$F7,$B1,$34,$91,$F8,$BF,$8F,$F9,$0F
	.dc.b	$CA,$D3,$EF,$9F,$9E,$8C,$C7,$E8,$F8,$1F,$2D,$F9,$12,$47,$CC,$7A
	.dc.b	$71,$7E,$23,$E3,$E4,$FE,$CA,$FE,$83,$FA,$D7,$F9,$63,$F6,$B6,$FC
	.dc.b	$79,$E1,$E1,$FC,$8F,$C5,$FE,$7C,$93,$F6,$9F,$9B,$A3,$DE,$FC,$BF
	.dc.b	$04,$AF,$ED,$FB,$FC,$29,$F6,$71,$7E,$73,$39,$5E,$58,$FC,$FF,$C8
	.dc.b	$7E,$B4,$91,$F8,$87,$BF,$B7,$FC,$0B,$F7,$9F,$F7,$B7,$81,$FC,$4D
	.dc.b	$BE,$02,$FF,$C1,$5A,$AF,$87,$9D,$27,$F8,$AD,$FF,$69,$7E,$27,$37
	.dc.b	$F0,$EF,$A4,$69,$FC,$05,$B3,$FB,$F6,$BF,$B8,$6B,$FE,$9A,$FA,$4F
	.dc.b	$F4,$AB,$FE,$88,$6B,$FE,$A1,$6B,$BF,$67,$F0,$9C,$BE,$DF,$93,$93
	.dc.b	$F6,$63,$F0,$9C,$FF,$85,$64,$7D,$D8,$FC,$47,$2F,$C5,$F9,$29,$3F
	.dc.b	$81,$8C,$B4,$87,$E0,$F9,$3F,$96,$FC,$E7,$CA,$8F,$EC,$CD,$F1,$0D
	.dc.b	$97,$37,$E5,$CF,$CE,$12,$BF,$92,$46,$9F,$E4,$D3,$F3,$F9,$BF,$A5
	.dc.b	$4E,$67,$F2,$95,$E2,$E3,$F1,$E5,$F7,$FE,$52,$57,$F4,$CC,$53,$FC
	.dc.b	$FF,$E7,$7F,$D1,$33,$FE,$B6,$6B,$FE,$CF,$6B,$FE,$CE,$6B,$F8,$7C
	.dc.b	$FD,$BF,$27,$F8,$DF,$94,$93,$DE,$3E,$2C,$48,$F1,$3F,$23,$E2,$FF
	.dc.b	$29,$32,$7F,$1A,$FF,$60,$7F,$CC,$BE,$1D,$FA,$33,$FD,$11,$B1,$FE
	.dc.b	$1D,$6B,$F9,$B2,$38,$70,$17,$4F,$32,$47,$37,$FA,$D6,$9F,$ED,$D8
	.dc.b	$FD,$0E,$1F,$AB,$3F,$03,$F1,$92,$3E,$73,$5F,$EE,$93,$FB,$F7,$E9
	.dc.b	$7C,$A7,$EC,$FE,$56,$46,$0F,$E0,$3B,$7F,$7E,$7E,$97,$AB,$FC,$D2
	.dc.b	$31,$1F,$4E,$3F,$B7,$77,$79,$63,$FC,$04,$DE,$6F,$DF,$65,$F9,$6F
	.dc.b	$03,$E6,$7C,$63,$87,$FB,$6B,$70,$FF,$23,$35,$F8,$E7,$FB,$E6,$3E
	.dc.b	$87,$91,$23,$E8,$35,$FE,$E7,$B7,$FB,$BD,$FE,$0F,$E2,$7F,$11,$3F
	.dc.b	$4F,$F1,$9F,$E1,$CF,$7D,$F4,$7F,$8C,$AF,$F2,$BB,$63,$3F,$05,$FE
	.dc.b	$59,$6C,$FE,$93,$E5,$BF,$8E,$72,$48,$FC,$F4,$7E,$77,$E2,$7E,$41
	.dc.b	$23,$F3,$CF,$F8,$57,$F3,$2F,$E7,$71,$FE,$0A,$6B,$FD,$F9,$AF,$E7
	.dc.b	$DA,$FF,$27,$FA,$4F,$E9,$EF,$E3,$1A,$FE,$45,$AE,$FD,$6F,$F1,$27
	.dc.b	$CE,$3C,$EF,$F1,$AB,$67,$FC,$82,$D7,$FD,$5D,$E7,$BE,$8F,$F2,$F1
	.dc.b	$FD,$19,$AF,$F9,$06,$78,$74,$7F,$93,$DB,$7F,$CF,$6F,$97,$77,$EB
	.dc.b	$7E,$8F,$F1,$78,$7E,$9D,$AF,$F4,$36,$A6,$9F,$E5,$07,$F5,$6D,$79
	.dc.b	$7C,$17,$F8,$E2,$31,$1F,$E8,$9E,$C7,$CA,$7E,$60,$91,$FE,$43,$8C
	.dc.b	$BF,$C4,$FD,$CF,$D0,$11,$C1,$FC,$93,$E5,$3F,$98,$91,$FD,$4F,$C7
	.dc.b	$FF,$1D,$30,$FE,$CA,$30,$FF,$5A,$EF,$FD,$49,$1F,$13,$FA,$FF,$76
	.dc.b	$47,$F2,$56,$BF,$E5,$0C,$3F,$3C,$FE,$C3,$AF,$FB,$71,$FE,$70,$FA
	.dc.b	$3F,$A3,$1F,$EA,$95,$F2,$DF,$E6,$56,$C7,$F9,$8D,$AF,$F9,$71,$AF
	.dc.b	$F9,$41,$FD,$37,$96,$69,$3B,$7F,$81,$FB,$3F,$DF,$C9,$5F,$E8,$C9
	.dc.b	$FF,$5C,$37,$E3,$3C,$44,$7C,$4F,$97,$E5,$3D,$5F,$2D,$27,$94,$7B
	.dc.b	$7E,$57,$DC,$C8,$ED,$1F,$E0,$FF,$02,$47,$F8,$21,$F8,$9C,$DF,$E1
	.dc.b	$3F,$9B,$FE,$32,$49,$FC,$13,$F1,$7E,$5B,$FC,$31,$F3,$92,$7D,$51
	.dc.b	$FE,$1B,$E5,$FE,$42,$48,$FC,$53,$FC,$93,$E7,$24,$7E,$35,$AF,$F1
	.dc.b	$CF,$94,$7C,$4B,$C1,$FE,$B9,$F2,$DF,$D2,$67,$31,$8A,$FF,$13,$46
	.dc.b	$1F,$F7,$24,$EE,$B9,$46,$0F,$E9,$FF,$29,$C5,$FF,$6E,$19,$2F,$32
	.dc.b	$BF,$97,$CF,$78,$74,$BE,$6B,$F8,$25,$FF,$71,$8C,$78,$7F,$A9,$9C
	.dc.b	$CF,$EB,$DB,$4D,$3F,$88,$57,$FA,$C5,$B5,$FE,$AA,$6B,$FE,$9E,$6B
	.dc.b	$FD,$FF,$9D,$F3,$3F,$C5,$F6,$FF,$6E,$57,$E7,$2F,$F1,$D3,$97,$FC
	.dc.b	$0D,$19,$BF,$F6,$B2,$EA,$F2,$B6,$7C,$91,$83,$3F,$31,$CE,$99,$7A
	.dc.b	$E3,$0B,$E8,$FE,$55,$FF,$BA,$B0,$56,$A5,$F3,$37,$E5,$30,$DB,$FF
	.dc.b	$78,$CA,$78,$79,$79,$CB,$FD,$DD,$3F,$EB,$9A,$DF,$1F,$CE,$8F,$C5
	.dc.b	$C7,$FB,$75,$B3,$FE,$D6,$6B,$F3,$7E,$63,$D3,$99,$F3,$7C,$7E,$1F
	.dc.b	$47,$C9,$75,$26,$1D,$5F,$07,$C0,$48,$F0,$37,$0F,$B9,$E1,$76,$3E
	.dc.b	$17,$C7,$5A,$FF,$47,$FD,$0F,$C8,$AF,$F1,$2F,$9E,$8C,$47,$F8,$AE
	.dc.b	$31,$D7,$F8,$E6,$D7,$E2,$FC,$5F,$B6,$47,$07,$F8,$2A,$3A,$71,$FD
	.dc.b	$3F,$8A,$FB,$91,$89,$FA,$31,$96,$9E,$69,$C6,$1F,$D9,$31,$FA,$D3
	.dc.b	$D5,$B1,$EA,$6B,$E9,$6B,$F9,$4F,$1B,$F2,$24,$9F,$D2,$3F,$AA,$87
	.dc.b	$D1,$BC,$1C,$36,$BF,$C4,$B3,$A9,$C9,$AE,$39,$3A,$AF,$D9,$36,$3E
	.dc.b	$57,$D1,$E2,$FF,$BD,$7C,$C7,$C4,$8C,$77,$7F,$7B,$8F,$CF,$FA,$86
	.dc.b	$9B,$BB,$87,$FA,$EE,$BF,$C1,$5F,$44,$F9,$0F,$63,$67,$D2,$DC,$BF
	.dc.b	$C7,$ED,$7F,$CD,$5D,$EF,$CB,$7F,$87,$C7,$F9,$AA,$31,$E2,$7F,$0E
	.dc.b	$9C,$CF,$06,$BF,$E1,$1E,$1F,$D4,$3F,$57,$BF,$F1,$51,$88,$FA,$D1
	.dc.b	$FA,$BF,$90,$FD,$2F,$9D,$FB,$96,$E7,$ED,$3F,$DB,$1E,$F7,$F0,$4F
	.dc.b	$9C,$FE,$42,$D8,$FF,$2C,$B7,$0F,$F4,$0B,$5F,$6F,$85,$FE,$B4,$9F
	.dc.b	$E1,$1F,$0F,$FC,$1F,$E4,$3F,$6A,$FF,$11,$A7,$D1,$BB,$3F,$0C,$DB
	.dc.b	$FC,$1B,$67,$FC,$D2,$D7,$FD,$23,$B3,$3F,$C7,$36,$BF,$16,$D9,$FF
	.dc.b	$38,$B5,$FE,$69,$E9,$7F,$08,$3F,$90,$F9,$7F,$EF,$9A,$FE,$41,$1F
	.dc.b	$AC,$AF,$F7,$9C,$7E,$97,$C2,$F9,$E4,$7F,$CF,$BC,$9F,$CF,$39,$BF
	.dc.b	$89,$FC,$A7,$46,$CF,$CF,$CB,$F9,$8E,$4F,$91,$F0,$9E,$2F,$91,$B5
	.dc.b	$BC,$7F,$A3,$CF,$F1,$EF,$9E,$CE,$1F,$D0,$9F,$99,$BB,$E5,$7E,$59
	.dc.b	$30,$F9,$6F,$CE,$F9,$89,$1F,$39,$B7,$FD,$67,$D6,$FE,$56,$7D,$7F
	.dc.b	$9E,$F9,$5D,$FF,$D5,$A3,$E3,$FC,$9C,$C3,$CD,$BB,$E5,$71,$8F,$F8
	.dc.b	$0E,$7F,$DB,$D2,$F9,$4F,$E6,$FF,$17,$F4,$1B,$3F,$CA,$BB,$BF,$36
	.dc.b	$47,$F8,$A3,$FD,$CA,$9F,$1D,$FD,$03,$C6,$FB,$2D,$AF,$1F,$C3,$FD
	.dc.b	$CF,$85,$FB,$AD,$8F,$3C,$73,$23,$F6,$AD,$1D,$FE,$13,$F0,$A7,$1D
	.dc.b	$3D,$87,$DA,$7F,$13,$F0,$52,$FE,$1F,$C1,$DA,$DF,$4F,$F8,$02,$BF
	.dc.b	$AB,$FC,$F6,$79,$3F,$C0,$2F,$E4,$BE,$13,$F2,$7E,$8D,$CB,$E6,$B8
	.dc.b	$7F,$BA,$7C,$45,$CB,$FC,$06,$FF,$42,$FA,$36,$1F,$D3,$5B,$5F,$D2
	.dc.b	$5B,$5F,$D5,$9A,$FF,$16,$FA,$37,$85,$FE,$20,$6D,$7E,$79,$B3,$FC
	.dc.b	$65,$AF,$F8,$7F,$E2,$BF,$B9,$9F,$AC,$F4,$2D,$7E,$87,$D8,$7C,$AB
	.dc.b	$B7,$FB,$CF,$37,$F5,$A8,$C5,$7F,$84,$63,$FB,$EF,$C5,$FF,$87,$89
	.dc.b	$1F,$E3,$17,$FC,$6B,$F4,$7F,$E2,$5E,$3F,$F6,$17,$2B,$F7,$CF,$F0
	.dc.b	$FD,$FE,$AE,$B4,$B6,$FF,$11,$CE,$9F,$F6,$3E,$5F,$EA,$E9,$FD,$94
	.dc.b	$FF,$7D,$C3,$FD,$9D,$A7,$AB,$AB,$FC,$53,$39,$AE,$99,$78,$7E,$93
	.dc.b	$FC,$93,$26,$92,$7F,$84,$E7,$FE,$2E,$7B,$FB,$BF,$E3,$48,$C6,$9F
	.dc.b	$8A,$BF,$F1,$8D,$7A,$FE,$0F,$C3,$E8,$6D,$B1,$E8,$F8,$6F,$09,$1F
	.dc.b	$E2,$F5,$FD,$EA,$FF,$C8,$78,$3D,$FF,$5F,$B3,$E1,$6D,$FD,$9F,$90
	.dc.b	$FF,$2D,$DB,$81,$18,$6C,$FF,$87,$63,$7B,$FD,$3F,$86,$F8,$64,$9F
	.dc.b	$1B,$BF,$88,$91,$FE,$1F,$AF,$83,$87,$F8,$A5,$AF,$FB,$CD,$7C,$2F
	.dc.b	$D2,$F8,$8F,$F1,$1C,$62,$BE,$CC,$7D,$6F,$A0,$FA,$64,$8F,$86,$FD
	.dc.b	$6F,$89,$FB,$67,$2F,$80,$B3,$F8,$9F,$57,$EF,$FB,$3F,$55,$B1,$F8
	.dc.b	$6F,$89,$FB,$E4,$7E,$7B,$F8,$2C,$BF,$0E,$79,$A6,$CF,$E5,$7C,$EF
	.dc.b	$4F,$A7,$F8,$16,$C7,$E6,$7C,$3F,$CF,$91,$FC,$4B,$F8,$EF,$98,$FC
	.dc.b	$99,$F8,$9F,$26,$CF,$F5,$1E,$17,$AB,$FE,$54,$7F,$44,$63,$E2,$BF
	.dc.b	$C7,$11,$98,$FD,$33,$79,$BF,$59,$EF,$FE,$70,$97,$F8,$C7,$F0,$18
	.dc.b	$7F,$A3,$3E,$87,$F5,$9E,$27,$EA,$38,$13,$9C,$26,$1F,$D7,$EF,$6F
	.dc.b	$DD,$9C,$E5,$FC,$AA,$7F,$5C,$FF,$A6,$9F,$D1,$B9,$CF,$C1,$E1,$4B
	.dc.b	$E8,$F8,$E3,$13,$FC,$E8,$7D,$27,$FA,$7B,$E7,$FF,$82,$7C,$8F,$FA
	.dc.b	$7A,$7F,$83,$4E,$54,$65,$E3,$F1,$EC,$78,$E3,$F3,$2F,$85,$ED,$B1
	.dc.b	$9E,$D6,$3C,$31,$EF,$8F,$37,$BD,$6E,$75,$1E,$8A,$E2,$E8,$F3,$B8
	.dc.b	$C3,$C1,$7C,$DF,$14,$66,$FA,$57,$13,$A5,$F6,$C3,$FD,$99,$19,$7D
	.dc.b	$2F,$68,$D5,$73,$8F,$4E,$38,$99,$1F,$8E,$38,$14,$EE,$5A,$F9,$CE
	.dc.b	$FE,$49,$D8,$F8,$9F,$F4,$3C,$61,$B1,$99,$E9,$39,$7A,$78,$FF,$EE
	.dc.b	$06,$F0,$FF,$A4,$30,$FF,$64,$36,$75,$3E,$79,$CB,$F0,$63,$96,$BF
	.dc.b	$6D,$BA,$FD,$FF,$03,$F5,$73,$9C,$3E,$55,$6A,$FF,$16,$D7,$F2,$35
	.dc.b	$C2,$FF,$11,$39,$C3,$C3,$87,$3B,$FA,$57,$B7,$9E,$1D,$88,$E8,$B3
	.dc.b	$C4,$98,$79,$E3,$F2,$13,$FD,$9F,$E0,$3F,$81,$F2,$FF,$6A,$F1,$DF
	.dc.b	$01,$8E,$1F,$EB,$B1,$8C,$3F,$B2,$DE,$5F,$ED,$6D,$C3,$A4,$68,$FE
	.dc.b	$D7,$EE,$E6,$9F,$41,$7E,$26,$57,$F7,$08,$D1,$1C,$1F,$E6,$5F,$E0
	.dc.b	$1F,$C6,$69,$FA,$78,$FE,$3B,$87,$FB,$31,$A7,$FB,$3B,$B5,$19,$6C
	.dc.b	$7F,$9E,$AB,$F5,$6F,$F0,$58,$CB,$67,$D9,$3A,$AF,$F0,$7B,$FE,$C5
	.dc.b	$9F,$27,$C3,$E8,$DF,$C3,$F0,$F8,$7F,$F6,$FE,$AB,$FF,$41,$5F,$86
	.dc.b	$73,$18,$8F,$F6,$33,$FE,$6D,$7E,$07,$C3,$78,$8F,$1F,$CF,$79,$3D
	.dc.b	$99,$3F,$15,$F3,$FC,$37,$A3,$E7,$FA,$92,$7E,$03,$E0,$F8,$7F,$69
	.dc.b	$23,$E7,$7C,$EF,$C0,$F6,$A4,$7D,$17,$E1,$FC,$47,$E0,$63,$F9,$5F
	.dc.b	$97,$F4,$7B,$8D,$24,$7C,$FF,$88,$F9,$64,$7E,$E7,$C4,$7C,$EF,$88
	.dc.b	$F9,$FF,$11,$27,$F0,$11,$F8,$FF,$65,$F2,$5F,$7D,$B3,$FC,$34,$FF
	.dc.b	$0B,$46,$4F,$C2,$73,$7D,$1F,$98,$FB,$7F,$15,$2B,$F1,$91,$F7,$BE
	.dc.b	$4F,$EF,$FC,$3C,$9F,$C8,$47,$87,$07,$5F,$7C,$BC,$6D,$8C,$D7,$F1
	.dc.b	$F9,$CB,$FE,$C8,$FE,$01,$FC,$C7,$77,$F9,$61,$BE,$2F,$F9,$29,$B2
	.dc.b	$D5,$0F,$9A,$FD,$72,$F4,$6F,$A3,$2F,$FD,$95,$39,$7F,$CE,$59,$FF
	.dc.b	$AC,$DF,$23,$5F,$E6,$D9,$FC,$CE,$CF,$51,$F9,$8A,$79,$BE,$02,$57
	.dc.b	$6F,$37,$C7,$F3,$1E,$4F,$82,$93,$E2,$8F,$91,$D9,$F1,$49,$1E,$27
	.dc.b	$B7,$BF,$EC,$24,$7A,$BE,$4F,$DF,$F2,$72,$3D,$AF,$F1,$1F,$55,$F9
	.dc.b	$FF,$FA,$6A,$7F,$C4,$1F,$47,$FC,$F1,$FE,$A4,$D7,$FC,$42,$D7,$FD
	.dc.b	$08,$D9,$7E,$A9,$FC,$5A,$79,$F3,$7E,$7E,$BF,$A3,$F4,$7F,$77,$BD
	.dc.b	$FA,$1F,$53,$E4,$7E,$AF,$C1,$CC,$7E,$9C,$65,$A4,$3F,$D8,$FC,$8F
	.dc.b	$83,$FB,$BF,$29,$27,$ED,$BF,$80,$F8,$49,$18,$21,$C9,$FC,$2F,$CA
	.dc.b	$7E,$1B,$E0,$A4,$FE,$12,$3F,$13,$E5,$6A,$30,$7E,$21,$FC,$6F,$C9
	.dc.b	$6A,$30,$7E,$31,$FD,$97,$EA,$7F,$96,$E2,$FC,$AF,$C9,$FE,$50,$CF
	.dc.b	$89,$FD,$0F,$DB,$F8,$EF,$FA,$C7,$C1,$5F,$E7,$9B,$1B,$BE,$1F,$09
	.dc.b	$C7,$E4,$A6,$7F,$EB,$9E,$7F,$F5,$FE,$4A,$ED,$F0,$32,$B5,$8F,$FA
	.dc.b	$D2,$BF,$C7,$17,$C3,$E4,$BF,$8C,$7B,$DF,$A5,$BF,$F2,$23,$67,$FA
	.dc.b	$63,$6F,$F0,$FF,$25,$F8,$1F,$91,$FE,$9D,$DA,$69,$30,$FE,$25,$F5
	.dc.b	$8F,$8D,$E1,$1F,$94,$9F,$E5,$13,$98,$F3,$9F,$BF,$D1,$87,$C9,$24
	.dc.b	$6B,$81,$3E,$18,$C3,$56,$9F,$1A,$B8,$C6,$33,$F8,$98,$FD,$AF,$1F
	.dc.b	$EB,$7C,$94,$AF,$F7,$C7,$27,$FB,$15,$EB,$F2,$5B,$73,$5F,$76,$3E
	.dc.b	$F7,$C5,$C9,$F1,$BF,$D1,$7E,$53,$FA,$D4,$7F,$9C,$78,$7F,$C7,$AF
	.dc.b	$F3,$76,$BF,$E4,$56,$BF,$B8,$9F,$E7,$9E,$27,$F9,$BF,$D3,$FD,$94
	.dc.b	$49,$F9,$33,$97,$D3,$E4,$7E,$47,$E4,$E4,$60,$F8,$FF,$27,$F9,$A9
	.dc.b	$91,$83,$E0,$F8,$9F,$99,$F9,$39,$18,$3E,$7F,$C9,$7E,$8A,$64,$60
	.dc.b	$F9,$3E,$1F,$F4,$5E,$13,$49,$1F,$E6,$3E,$8F,$B7,$D5,$FA,$C9,$93
	.dc.b	$F7,$23,$F5,$FE,$1F,$EA,$FE,$4E,$4F,$D6,$8C,$B4,$87,$DF,$3F,$19
	.dc.b	$F1,$DF,$DA,$72,$FE,$41,$E1,$FF,$0F,$CB,$FA,$D4,$60,$FE,$48,$FE
	.dc.b	$3B,$E3,$F8,$32,$FF,$1F,$1F,$E0,$1F,$A0,$70,$7F,$8C,$5B,$3F,$DC
	.dc.b	$DA,$FF,$A0,$5A,$FF,$A1,$5F,$E7,$0F,$FA,$D5,$DF,$77,$F8,$F5,$BD
	.dc.b	$BF,$ED,$87,$71,$FE,$41,$FC,$EF,$F8,$15,$7E,$8B,$C2,$FE,$20,$7F
	.dc.b	$85,$63,$31,$FE,$C1,$6C,$F1,$F1,$74,$7F,$83,$7E,$C3,$F6,$35,$FE
	.dc.b	$E9,$F5,$DE,$37,$F9,$35,$B1,$FE,$0F,$6B,$FE,$9D,$6B,$FE,$A1,$79
	.dc.b	$BF,$ED,$17,$77,$FE,$59,$6E,$7F,$EF,$47,$71,$FC,$7B,$D1,$FE,$54
	.dc.b	$4A,$FD,$FF,$8D,$FE,$6F,$FE,$01,$FF,$0E,$92,$3F,$C9,$EE,$3E,$5B
	.dc.b	$51,$0F,$E1,$51,$FE,$5B,$C5,$1F,$E0,$F8,$FF,$2E,$3F,$89,$E2,$46
	.dc.b	$31,$FF,$6B,$46,$3E,$4B,$C5,$E1,$78,$FE,$4B,$C2,$4C,$3F,$3D,$87
	.dc.b	$F9,$75,$EE,$77,$F9,$1F,$51,$1F,$2F,$C8,$FB,$48,$E0,$F4,$FC,$8F
	.dc.b	$C1,$F8,$09,$1E,$C7,$E2,$7C,$8F,$C5,$F9,$1F,$86,$49,$F1,$BF,$1B
	.dc.b	$E4,$7E,$51,$23,$E2,$FC,$8F,$CD,$F8,$2F,$96,$48,$F5,$BF,$3B,$E4
	.dc.b	$24,$7B,$9A,$FF,$B5,$3D,$0F,$E6,$6F,$F8,$EF,$AF,$E0,$FC,$B2,$73
	.dc.b	$C9,$F9,$68,$C7,$2F,$D4,$8C,$4E,$AB,$FD,$D5,$F2,$D9,$6E,$1D,$4F
	.dc.b	$F7,$84,$65,$A6,$6B,$ED,$FC,$97,$FA,$A6,$3B,$72,$23,$11,$F8,$36
	.dc.b	$CF,$0E,$8F,$B7,$F2,$38,$B3,$49,$1F,$80,$F9,$23,$F9,$06,$E1,$F8
	.dc.b	$DA,$7C,$8E,$35,$FC,$6C,$2B,$17,$E6,$F9,$2F,$01,$1F,$C6,$BE,$3F
	.dc.b	$91,$DE,$3F,$9B,$6E,$7F,$9A,$F9,$2B,$8A,$73,$FC,$03,$E6,$C7,$FB
	.dc.b	$73,$70,$FD,$13,$FE,$71,$8C,$9C,$3E,$4A,$38,$38,$BD,$3E,$4B,$E4
	.dc.b	$C9,$A4,$9F,$E8,$51,$FE,$6E,$7C,$3F,$23,$E4,$F1,$BC,$AF,$9B,$E4
	.dc.b	$7C,$FE,$FD,$AF,$3B,$87,$9B,$E0,$23,$DB,$E4,$7D,$E7,$C0,$F9,$1B
	.dc.b	$1F,$E1,$2D,$1F,$AC,$7F,$A7,$A3,$0B,$C6,$F9,$B3,$AF,$1E,$CE,$6B
	.dc.b	$75,$F8,$AA,$E6,$D9,$D4,$7F,$02,$3F,$88,$B8,$7F,$82,$F3,$3F,$33
	.dc.b	$7C,$67,$DC,$8C,$3F,$7B,$D6,$7D,$5B,$E0,$75,$7C,$33,$FF,$16,$E1
	.dc.b	$C1,$FB,$BE,$B6,$B1,$FC,$0B,$EC,$FA,$5F,$C9,$CF,$C3,$5D,$99,$EE
	.dc.b	$D3,$27,$FD,$A1,$3A,$6C,$7C,$18,$C6,$BE,$FE,$D7,$43,$E4,$BD,$5E
	.dc.b	$D3,$E3,$7F,$27,$18,$66,$5F,$EC,$B9,$D3,$63,$F6,$FC,$5F,$E8,$96
	.dc.b	$CF,$F5,$89,$FF,$26,$CF,$FA,$37,$D6,$FC,$FF,$C9,$F8,$19,$3F,$A0
	.dc.b	$AB,$9F,$E8,$1F,$D2,$FC,$97,$9B,$E2,$ED,$74,$94,$66,$3C,$CF,$88
	.dc.b	$8F,$E7,$BE,$4B,$DF,$F1,$5F,$B0,$6C,$73,$8F,$D9,$FC,$2E,$31,$FD
	.dc.b	$8B,$FB,$6E,$5F,$8F,$F1,$16,$FC,$4E,$1F,$DC,$FC,$25,$BF,$DC,$BF
	.dc.b	$C1,$3E,$29,$FB,$EF,$E4,$BA,$BF,$89,$D3,$E8,$7F,$CD,$B8,$7F,$93
	.dc.b	$2B,$4F,$F5,$88,$CC,$FF,$8B,$4E,$1F,$01,$F9,$F8,$CC,$7E,$75,$AF
	.dc.b	$18,$FE,$07,$F2,$9F,$C4,$0F,$D5,$FD,$17,$FA,$02,$FF,$88,$C7,$F1
	.dc.b	$4F,$95,$FC,$2F,$C5,$D9,$FA,$0E,$1F,$F4,$27,$EF,$CE,$DC,$65,$FD
	.dc.b	$E6,$31,$3A,$F5,$FF,$91,$C7,$E2,$A7,$51,$8E,$C9,$5D,$1D,$C6,$27
	.dc.b	$E0,$7C,$F3,$9B,$F8,$1F,$92,$F8,$5F,$E0,$B3,$9C,$7F,$86,$4E,$5B
	.dc.b	$3F,$C1,$63,$17,$FC,$0C,$38,$7F,$83,$35,$FF,$04,$D3,$4E,$E7,$8A
	.dc.b	$6D,$FC,$4F,$97,$F7,$9F,$03,$BA,$C7,$C8,$F9,$6F,$87,$F2,$DF,$11
	.dc.b	$B1,$F3,$3B,$BF,$87,$7E,$AC,$F8,$AF,$93,$18,$64,$FD,$2F,$7E,$ED
	.dc.b	$F4,$5F,$9F,$C1,$F4,$3C,$DB,$5F,$61,$C3,$F6,$BE,$56,$DF,$D9,$7E
	.dc.b	$B9,$F6,$3C,$0B,$3F,$80,$7E,$DF,$CA,$FD,$CF,$94,$B3,$F8,$47,$0D
	.dc.b	$3F,$0F,$18,$BF,$C3,$71,$FE,$10,$FC,$2F,$CA,$5A,$FC,$6C,$7E,$23
	.dc.b	$E5,$31,$5F,$8D,$7F,$59,$E0,$BE,$2F,$F0,$7F,$19,$86,$E1,$F8,$28
	.dc.b	$FC,$01,$1F,$C0,$5F,$E0,$23,$FC,$1F,$DA,$F9,$3F,$CB,$9F,$B9,$FD
	.dc.b	$B7,$F8,$7D,$FE,$82,$3F,$37,$F2,$52,$3F,$47,$E2,$DC,$7F,$D4,$5E
	.dc.b	$83,$A5,$F1,$BF,$CA,$1B,$7F,$91,$9D,$4E,$4F,$F1,$0F,$8B,$FE,$66
	.dc.b	$9F,$F7,$93,$F8,$36,$DF,$F8,$3A,$30,$FF,$9A,$5F,$CE,$F9,$3F,$E6
	.dc.b	$86,$CF,$F8,$F5,$FE,$DD,$3B,$F0,$FF,$1F,$F0,$FE,$52,$57,$97,$2F
	.dc.b	$F1,$23,$FE,$C9,$BF,$E2,$B7,$FC,$05,$7C,$97,$8F,$DD,$F2,$3E,$5F
	.dc.b	$92,$97,$EA,$F9,$2B,$87,$C6,$C7,$FC,$49,$3D,$FE,$23,$E3,$11,$F0
	.dc.b	$72,$FB,$F9,$BE,$03,$63,$E2,$FC,$97,$C3,$F9,$2F,$88,$D8,$FE,$A7
	.dc.b	$FA,$B7,$F8,$AF,$2F,$F0,$17,$22,$31,$1F,$E8,$68,$C7,$CF,$FF,$5E
	.dc.b	$9C,$CF,$C9,$E5,$FA,$1F,$29,$F4,$5B,$87,$D6,$F9,$4F,$A8,$7D,$5F
	.dc.b	$92,$B1,$FC,$53,$4F,$E2,$44,$7F,$27,$E7,$A3,$F0,$3F,$25,$F7,$4F
	.dc.b	$BD,$D1,$6B,$F0,$7F,$09,$F8,$13,$F0,$5F,$09,$63,$F0,$FC,$9F,$8A
	.dc.b	$23,$F6,$5F,$C6,$FC,$97,$E2,$4F,$F2,$BF,$EF,$FF,$E5,$1E,$EF,$C4
	.dc.b	$46,$38,$E5,$FF,$AE,$27,$FD,$67,$4F,$92,$CF,$1F,$FA,$E6,$3F,$CA
	.dc.b	$15,$FC,$B7,$D9,$46,$27,$FC,$28,$D7,$FD,$6B,$1A,$F9,$2F,$F4,$0C
	.dc.b	$FF,$95,$67,$A7,$C8,$F9,$BE,$4B,$CE,$FA,$3E,$4B,$FA,$0F,$A3,$72
	.dc.b	$FF,$26,$7B,$68,$CC,$7F,$00,$8C,$36,$3F,$C9,$11,$8C,$3C,$BF,$25
	.dc.b	$F1,$0F,$8B,$F2,$56,$FE,$11,$FE,$A7,$CD,$3F,$2B,$E4,$B5,$7F,$0D
	.dc.b	$F9,$BF,$1F,$2F,$8C,$FF,$69,$BF,$F1,$A4,$63,$7F,$8F,$5B,$AF,$99
	.dc.b	$F1,$97,$8B,$E8,$BF,$E6,$4F,$A1,$FF,$3F,$7B,$9F,$EE,$CF,$95,$7C
	.dc.b	$17,$FB,$82,$73,$82,$72,$FF,$BC,$A3,$ED,$67,$C6,$B8,$72,$FF,$51
	.dc.b	$AE,$11,$8C,$35,$1C,$67,$58,$74,$86,$1D,$1E,$26,$EB,$8B,$AA,$E3
	.dc.b	$7D,$63,$67,$2E,$6F,$F5,$CE,$1D,$70,$F5,$B6,$F8,$D7,$53,$8C,$F8
	.dc.b	$67,$58,$78,$23,$8E,$1D,$E7,$A6,$1D,$A3,$07,$8E,$FF,$D1,$15,$F2
	.dc.b	$9B,$87,$C5,$6E,$1F,$D3,$3E,$5B,$FC,$2D,$F1,$37,$DF,$34,$95,$FE
	.dc.b	$11,$C1,$D1,$FE,$2B,$8C,$35,$F0,$B5,$FB,$8D,$7E,$D3,$C7,$4F,$8C
	.dc.b	$DC,$F7,$C3,$F0,$7A,$57,$F0,$EB,$4F,$29,$C1,$B6,$9F,$32,$36,$FD
	.dc.b	$E6,$E0,$AE,$B3,$BA,$E3,$87,$91,$E9,$19,$24,$65,$FB,$33,$CA,$3F
	.dc.b	$2F,$2F,$A3,$7B,$5F,$EB,$8E,$1F,$5C,$E6,$31,$87,$F9,$56,$B6,$FE
	.dc.b	$B5,$7B,$9A,$AE,$8D,$8D,$CB,$91,$87,$69,$CC,$62,$7B,$57,$1F,$5A
	.dc.b	$47,$77,$A7,$AF,$23,$04,$3B,$61,$D6,$3C,$9D,$0F,$8F,$FF,$0B,$E7
	.dc.b	$FC,$A9,$B1,$9B,$FE,$5C,$D7,$FC,$B9,$DB,$FE,$1B,$7F,$90,$7B,$3F
	.dc.b	$E2,$05,$19,$77,$1D,$3E,$43,$C2,$48,$FE,$29,$F4,$7F,$E6,$F6,$F9
	.dc.b	$7E,$48,$FF,$34,$C7,$F3,$F6,$BF,$ED,$16,$BF,$CF,$9F,$F0,$3B,$D3
	.dc.b	$D9,$FA,$51,$FE,$FF,$8E,$FE,$CF,$C2,$F9,$19,$1F,$C9,$73,$47,$0F
	.dc.b	$67,$E3,$FC,$8F,$C4,$25,$7C,$47,$CD,$EB,$FC,$CF,$91,$93,$EC,$8F
	.dc.b	$9F,$F2,$32,$3B,$3E,$4F,$5B,$F1,$71,$FC,$72,$BE,$57,$69,$A4,$8F
	.dc.b	$85,$EB,$7D,$AF,$8F,$91,$F2,$A3,$DF,$EA,$7C,$92,$47,$C9,$7E,$EF
	.dc.b	$BF,$23,$EE,$B5,$F9,$9E,$A7,$D1,$23,$F4,$5F,$A9,$E9,$FE,$77,$F8
	.dc.b	$07,$F8,$57,$81,$F9,$7E,$C4,$63,$BB,$F5,$3C,$EB,$E1,$F1,$FF,$90
	.dc.b	$A3,$49,$97,$F7,$DF,$A1,$FF,$6D,$E5,$D6,$9F,$23,$E1,$3F,$5D,$3A
	.dc.b	$CB,$F6,$EB,$CE,$F3,$79,$53,$1F,$EC,$53,$97,$FB,$26,$0F,$91,$F2
	.dc.b	$12,$3F,$C1,$D1,$97,$C9,$E5,$FC,$1F,$90,$91,$D2,$3C,$7E,$AF,$C5
	.dc.b	$F9,$19,$3E,$08,$F9,$1F,$23,$F0,$49,$1D,$9F,$D9,$E4,$F6,$3E,$77
	.dc.b	$C8,$CA,$F4,$C7,$D1,$F4,$E4,$7A,$9F,$87,$EC,$7F,$16,$F5,$1F,$46
	.dc.b	$F9,$5F,$F4,$AC,$61,$B1,$9B,$FF,$6C,$4E,$5F,$F4,$DF,$D0,$7F,$43
	.dc.b	$7F,$9E,$B5,$34,$D7,$5B,$47,$B7,$F8,$7F,$94,$97,$F4,$A3,$EC,$E5
	.dc.b	$F6,$99,$1F,$5A,$3E,$EF,$9F,$F8,$FF,$7A,$4F,$DA,$8C,$B4,$87,$F2
	.dc.b	$01,$8F,$8F,$CC,$FC,$3F,$8D,$27,$F0,$6F,$E2,$BC,$B9,$1F,$91,$7E
	.dc.b	$99,$1F,$CA,$79,$5A,$7B,$5F,$BE,$7F,$27,$3F,$94,$F1,$7F,$4D,$F2
	.dc.b	$92,$FF,$23,$1F,$D2,$BE,$25,$F4,$1F,$CF,$E7,$FD,$12,$D7,$FD,$10
	.dc.b	$D7,$FB,$6C,$70,$F0,$BF,$9C,$FB,$1F,$A4,$9F,$CD,$FC,$A7,$ED,$4F
	.dc.b	$EA,$51,$98,$F3,$BF,$B0,$7F,$C8,$8D,$73,$E3,$F0,$F1,$3F,$7D,$F2
	.dc.b	$B2,$7F,$4E,$FF,$18,$F5,$FF,$CE,$D1,$FE,$50,$3F,$07,$E2,$7F,$43
	.dc.b	$EA,$34,$91,$F8,$CF,$0F,$44,$70,$7F,$9F,$72,$E2,$FE,$DB,$E5,$7F
	.dc.b	$88,$1E,$CA,$91,$EA,$6B,$EE,$AF,$83,$E1,$FF,$17,$F7,$24,$FC,$08
	.dc.b	$FA,$5C,$3F,$E6,$E8,$FF,$74,$35,$FA,$7E,$2F,$FB,$F3,$8B,$EB,$BF
	.dc.b	$C6,$7D,$EF,$E4,$44,$9F,$97,$E2,$7F,$3C,$FA,$1F,$EA,$71,$FE,$18
	.dc.b	$3F,$C4,$AE,$1C,$35,$FF,$0E,$B5,$FD,$7C,$7C,$B2,$3F,$B4,$7F,$03
	.dc.b	$E0,$FF,$33,$E1,$93,$F7,$23,$DB,$E2,$FA,$73,$34,$93,$F3,$FA,$7F
	.dc.b	$9E,$7B,$D2,$3F,$0B,$1F,$1B,$DE,$E4,$48,$F8,$EF,$F3,$BE,$49,$1F
	.dc.b	$CA,$1A,$FE,$73,$B7,$C4,$47,$E7,$BF,$9A,$E5,$FC,$B9,$23,$FD,$37
	.dc.b	$C3,$FA,$6E,$DF,$AE,$66,$3F,$37,$38,$7F,$D3,$DF,$C0,$3F,$E1,$3E
	.dc.b	$F3,$49,$1F,$6F,$2F,$E3,$98,$7F,$88,$EB,$04,$9F,$B0,$7F,$97,$3E
	.dc.b	$89,$F0,$FF,$E1,$F6,$C6,$83,$87,$8B,$5F,$C1,$3F,$49,$E0,$D7,$7B
	.dc.b	$F5,$3F,$7B,$C5,$FB,$A2,$47,$EB,$FE,$8B,$F7,$1E,$9F,$F0,$96,$F8
	.dc.b	$5F,$E5,$8E,$1F,$F1,$94,$FF,$75,$6C,$7F,$9A,$A7,$75,$FE,$78,$F8
	.dc.b	$5F,$F5,$73,$F9,$76,$BF,$E7,$06,$BB,$F0,$7F,$C8,$11,$FD,$96,$3E
	.dc.b	$3B,$5E,$5E,$96,$8F,$F2,$7F,$0F,$FA,$77,$2F,$F0,$43,$71,$FF,$37
	.dc.b	$B6,$BF,$9C,$E4,$E6,$FE,$36,$FF,$20,$24,$7F,$07,$7F,$CB,$7F,$43
	.dc.b	$F8,$12,$46,$0F,$0F,$CB,$FF,$9D,$6F,$FC,$A1,$1F,$CE,$9B,$8F,$F8
	.dc.b	$6E,$AE,$3F,$E8,$27,$FC,$9F,$57,$E1,$3B,$91,$FC,$3F,$1A,$7F,$80
	.dc.b	$DF,$3F,$84,$FF,$42,$47,$F9,$4E,$BF,$A4,$BA,$C3,$B7,$CB,$EB,$E1
	.dc.b	$65,$7F,$BE,$69,$3E,$9C,$A4,$FE,$81,$EB,$F2,$BE,$DC,$7E,$79,$27
	.dc.b	$CB,$A7,$85,$FF,$55,$E9,$E7,$F9,$2F,$89,$9C,$9F,$04,$7C,$7C,$BE
	.dc.b	$C9,$23,$9B,$F7,$3E,$1A,$47,$66,$BF,$AA,$F8,$B7,$B5,$FE,$A1,$8F
	.dc.b	$EE,$BE,$77,$FA,$F2,$73,$C3,$FE,$B0,$8C,$56,$F3,$FE,$AD,$5A,$BE
	.dc.b	$27,$F2,$DF,$8B,$E0,$D3,$27,$FA,$FC,$A7,$FB,$59,$DC,$FF,$06,$DF
	.dc.b	$FB,$73,$78,$BE,$3C,$63,$73,$7E,$5A,$4F,$CC,$6E,$1F,$CF,$EF,$F0
	.dc.b	$5F,$09,$FE,$BB,$D7,$F3,$FF,$13,$70,$FC,$64,$DB,$FF,$0B,$E4,$AC
	.dc.b	$61,$FC,$6B,$87,$C4,$4C,$3E,$DB,$E5,$F8,$9D,$D9,$E3,$CC,$F9,$1C
	.dc.b	$C7,$4F,$9D,$FF,$68,$E1,$FE,$F8,$D7,$FC,$F5,$0F,$90,$D4,$ED,$8E
	.dc.b	$CF,$F7,$2C,$7F,$9B,$DC,$6B,$FE,$30,$8E,$51,$8D,$11,$B9,$FF,$14
	.dc.b	$BF,$EF,$D9,$FD,$4B,$63,$E5,$35,$FC,$BB,$5F,$E0,$FF,$2D,$F0,$0F
	.dc.b	$67,$C2,$5A,$F0,$53,$D2,$FB,$AD,$8E,$F1,$83,$3E,$3C,$8F,$1B,$70
	.dc.b	$FB,$9B,$BE,$2F,$E7,$9F,$2F,$E1,$6D,$FF,$88,$FE,$85,$DF,$FE,$39
	.dc.b	$C1,$18,$9F,$E9,$0D,$8F,$E9,$ED,$7F,$C1,$0B,$BF,$93,$67,$8C,$60
	.dc.b	$FF,$23,$70,$BE,$83,$FD,$95,$3F,$0B,$CA,$FC,$74,$E7,$DF,$FC,$14
	.dc.b	$49,$D1,$F6,$A7,$4E,$31,$FF,$36,$0F,$A0,$7C,$77,$B5,$B3,$FD,$8D
	.dc.b	$B3,$FD,$A9,$AF,$F8,$8D,$FC,$A4,$6F,$A3,$91,$1F,$EE,$3E,$7B,$D0
	.dc.b	$FF,$60,$C6,$1B,$7F,$E2,$D9,$D7,$C6,$7F,$99,$DB,$5F,$C8,$3E,$67
	.dc.b	$F7,$44,$70,$73,$F3,$7F,$49,$87,$EC,$EB,$C1,$18,$9F,$C5,$CD,$C3
	.dc.b	$FD,$1D,$5A,$9C,$9F,$E1,$89,$7B,$9E,$36,$C7,$A6,$30,$7F,$0D,$F9
	.dc.b	$99,$1F,$CD,$9B,$7E,$DA,$7B,$7C,$C9,$5F,$0D,$FC,$EF,$B1,$FC,$0B
	.dc.b	$C5,$FD,$F3,$67,$C9,$EC,$F8,$C8,$FC,$B7,$F4,$BE,$CF,$F0,$AF,$99
	.dc.b	$FE,$0C,$D9,$FF,$3F,$7D,$03,$D4,$FC,$E3,$63,$FC,$44,$DD,$3F,$C5
	.dc.b	$2D,$7F,$CC,$5F,$41,$FE,$84,$9F,$F1,$43,$63,$FC,$5E,$D7,$7F,$07
	.dc.b	$FC,$C7,$9F,$F8,$E9,$FC,$67,$E5,$2D,$FC,$FF,$7B,$E7,$11,$FC,$33
	.dc.b	$FC,$F3,$E7,$FF,$D8,$71,$FE,$46,$E9,$FE,$5B,$18,$C7,$FC,$05,$18
	.dc.b	$E9,$95,$FC,$66,$9E,$AE,$2F,$FD,$83,$1F,$92,$EC,$FF,$0D,$FC,$A6
	.dc.b	$5B,$7F,$EA,$5F,$65,$ED,$7F,$03,$6C,$7F,$92,$5B,$87,$F9,$41,$AF
	.dc.b	$FA,$25,$1C,$49,$1C,$5F,$EB,$37,$FC,$F2,$7F,$DA,$AD,$BF,$F6,$77
	.dc.b	$CC,$35,$FE,$9B,$39,$C1,$7F,$EE,$8C,$BE,$A6,$09,$CF,$9E,$9C,$C6
	.dc.b	$17,$BD,$F3,$38,$BE,$6F,$C3,$F6,$F9,$CB,$7F,$D0,$F3,$47,$EB,$FE
	.dc.b	$79,$F3,$9F,$B9,$6D,$7F,$02,$6E,$BF,$BF,$6B,$FE,$1A,$CF,$FC,$1B
	.dc.b	$3D,$7D,$5F,$BB,$EF,$7B,$9F,$BF,$F0,$FF,$80,$F9,$1B,$5F,$75,$C3
	.dc.b	$FE,$3A,$B4,$7E,$24,$91,$9F,$03,$F1,$27,$0C,$7F,$06,$FE,$33,$E3
	.dc.b	$64,$FE,$27,$E3,$AD,$7E,$22,$7E,$8F,$CC,$63,$97,$E7,$4B,$95,$C5
	.dc.b	$FD,$9E,$73,$1F,$47,$9B,$F3,$3F,$1B,$F9,$A6,$E7,$F9,$78,$FB,$BE
	.dc.b	$DE,$31,$FC,$1B,$FE,$0F,$7F,$17,$F0,$5F,$93,$25,$7E,$4D,$F1,$C7
	.dc.b	$E1,$07,$CB,$3E,$7F,$F8,$DC,$7F,$B7,$63,$1E,$77,$08,$FC,$5C,$62
	.dc.b	$FF,$D5,$6D,$8F,$F1,$36,$BF,$DF,$F5,$F6,$B6,$BF,$19,$F1,$9F,$96
	.dc.b	$F8,$BF,$DB,$FC,$5F,$E6,$3E,$7A,$D7,$E2,$DF,$CD,$FC,$0F,$E8,$89
	.dc.b	$1F,$E7,$4C,$3F,$C0,$19,$7F,$8D,$7E,$4F,$F1,$79,$F8,$3D,$FF,$F1
	.dc.b	$7C,$62,$FF,$3F,$69,$CF,$8D,$FA,$8F,$82,$B5,$F9,$B7,$F5,$7E,$DF
	.dc.b	$EC,$89,$1E,$17,$FC,$9B,$87,$F7,$9C,$BF,$6B,$F2,$DE,$0F,$7F,$F6
	.dc.b	$6F,$8B,$E3,$BF,$6B,$F2,$F7,$1F,$D4,$3F,$B8,$F6,$BF,$7E,$48,$F7
	.dc.b	$3D,$BE,$33,$F7,$BF,$2F,$EC,$6C,$FF,$8C,$F3,$7C,$67,$90,$91,$FE
	.dc.b	$9A,$8C,$BF,$27,$E5,$FF,$86,$91,$C1,$BF,$8B,$F6,$91,$C1,$CF,$E2
	.dc.b	$7F,$87,$FB,$DF,$39,$B1,$FC,$97,$E3,$3F,$92,$11,$FE,$48,$FF,$1C
	.dc.b	$F8,$7F,$E2,$E7,$D6,$F8,$3B,$3E,$BF,$81,$FE,$37,$F0,$3F,$69,$B1
	.dc.b	$FC,$3F,$DD,$FE,$4A,$47,$F0,$8F,$F2,$8D,$7F,$91,$9F,$82,$EB,$B3
	.dc.b	$FE,$59,$F9,$DF,$E2,$ED,$F2,$FF,$CE,$B1,$8E,$97,$77,$ED,$A3,$F7
	.dc.b	$15,$FC,$A1,$FF,$2C,$4F,$EF,$5B,$E5,$9A,$63,$F6,$7F,$EA,$97,$FC
	.dc.b	$01,$B7,$CF,$CA,$FF,$42,$C6,$30,$FE,$73,$ED,$FF,$18,$AF,$DC,$47
	.dc.b	$EE,$63,$0E,$03,$87,$6D,$7F,$9F,$CF,$F6,$09,$FF,$6E,$B6,$3F,$D3
	.dc.b	$CD,$7F,$D9,$6D,$7E,$07,$87,$F1,$BC,$1F,$8E,$DA,$F9,$91,$67,$E4
	.dc.b	$C7,$03,$27,$D1,$F0,$7E,$7F,$77,$D0,$6C,$7D,$58,$B3,$C8,$F9,$D3
	.dc.b	$8C,$FF,$8D,$FC,$F3,$7F,$E9,$F4,$F9,$2F,$EB,$2F,$F6,$29,$D7,$B3
	.dc.b	$FE,$C7,$F7,$7F,$91,$47,$FA,$72,$78,$46,$2F,$9D,$F2,$95,$72,$76
	.dc.b	$75,$9D,$D7,$4F,$63,$FA,$0D,$F1,$BE,$0D,$AD,$DF,$1C,$3C,$8F,$9E
	.dc.b	$F8,$E1,$E3,$9C,$CF,$8A,$B8,$B6,$33,$87,$AF,$0E,$D1,$8A,$DC,$F8
	.dc.b	$AF,$CD,$A6,$6A,$61,$F8,$BA,$F7,$3F,$90,$6B,$E3,$3F,$21,$F2,$7F
	.dc.b	$E5,$AF,$3B,$F2,$50,$8C,$3A,$F8,$8F,$E9,$ED,$9D,$D3,$93,$FB,$34
	.dc.b	$F8,$5A,$F8,$5A,$FD,$86,$BF,$75,$E3,$A7,$CF,$6E,$7B,$E6,$F8,$BE
	.dc.b	$77,$F6,$1A,$D4,$7C,$8A,$DD,$F3,$8D,$B9,$63,$F8,$03,$F0,$71,$C1
	.dc.b	$FC,$2D,$75,$BF,$04,$6F,$0F,$75,$F2,$C3,$F0,$97,$BA,$F0,$1B,$C3
	.dc.b	$A5,$F0,$8E,$53,$BB,$72,$BE,$0B,$FB,$E7,$17,$F0,$DC,$EF,$07,$F0
	.dc.b	$06,$E3,$FE,$13,$BE,$0F,$FB,$ED,$7C,$56,$B1,$EB,$8D,$9D,$D7,$EC
	.dc.b	$E7,$C3,$59,$F8,$19,$3C,$DE,$1F,$03,$23,$04,$39,$5A,$3B,$7A,$9F
	.dc.b	$8A,$C3,$DA,$D9,$FF,$58,$D7,$F1,$88,$C7,$BF,$C2,$3D,$51,$E7,$8F
	.dc.b	$47,$1C,$9F,$73,$EB,$F7,$FD,$9E,$F7,$52,$4F,$F2,$47,$C1,$EF,$7C
	.dc.b	$1F,$7A,$4F,$C7,$8F,$89,$EF,$48,$F9,$4F,$C7,$F7,$BE,$47,$CC,$FA
	.dc.b	$49,$3F,$CC,$1F,$77,$CF,$3D,$FF,$A0,$BE,$5A,$98,$8F,$A5,$F0,$32
	.dc.b	$FE,$C3,$F5,$7E,$03,$8F,$A9,$D0,$93,$FD,$17,$91,$8A,$7F,$6F,$2F
	.dc.b	$A4,$6D,$F8,$68,$FE,$3F,$E8,$5E,$F3,$9A,$DC,$FE,$27,$E0,$7F,$0C
	.dc.b	$47,$F1,$9F,$19,$F8,$BF,$D2,$7B,$E8,$69,$27,$F0,$0F,$F3,$A7,$F1
	.dc.b	$CF,$FA,$B9,$FF,$4D,$F8,$3F,$E6,$88,$C4,$6B,$C7,$ED,$F1,$12,$37
	.dc.b	$E1,$FF,$A7,$AB,$FD,$5F,$18,$3F,$CE,$8B,$E5,$3C,$DF,$05,$E2,$24
	.dc.b	$FC,$99,$47,$9B,$E0,$7D,$84,$9E,$99,$FB,$79,$FE,$07,$3F,$C1,$F7
	.dc.b	$E5,$6B,$1F,$C0,$37,$E6,$3E,$09,$1F,$03,$F1,$3E,$03,$E5,$7B,$B2
	.dc.b	$79,$47,$CD,$E8,$F9,$DE,$F4,$9E,$D1,$F4,$7E,$4B,$E6,$92,$3D,$8F
	.dc.b	$D0,$F7,$BE,$B7,$C8,$C9,$F3,$47,$D9,$F7,$E4,$7A,$5F,$B9,$F1,$FF
	.dc.b	$77,$E3,$FE,$C9,$27,$E4,$3F,$73,$DF,$D4,$7C,$18,$FC,$24,$7F,$0F
	.dc.b	$FD,$EB,$FD,$2B,$D7,$FE,$A4,$7F,$D9,$75,$F3,$1A,$FF,$11,$77,$D3
	.dc.b	$FE,$DA,$6C,$7F,$AB,$DB,$CC,$68,$F6,$BF,$91,$FC,$9B,$EA,$1A,$61
	.dc.b	$B8,$7E,$1E,$75,$3E,$7F,$72,$4E,$6B,$6F,$F8,$5E,$B6,$FA,$67,$A6
	.dc.b	$9E,$FF,$66,$4F,$18,$F8,$5E,$CE,$A3,$07,$C1,$7E,$37,$C6,$EA,$30
	.dc.b	$7C,$57,$E5,$7C,$0C,$8F,$90,$FC,$DF,$8D,$91,$F2,$DA,$FC,$BF,$82
	.dc.b	$FA,$5F,$03,$F4,$FE,$32,$4F,$D1,$8F,$AD,$F0,$72,$3E,$A3,$F6,$7E
	.dc.b	$33,$F8,$47,$D5,$BC,$6A,$4A,$FD,$80,$FA,$0F,$F1,$67,$4F,$FB,$51
	.dc.b	$FE,$A3,$1F,$E0,$A6,$BF,$E5,$D6,$BF,$E6,$26,$BF,$ED,$BF,$39,$EC
	.dc.b	$7F,$19,$9E,$35,$D7,$E0,$7F,$8C,$4F,$E0,$3E,$95,$F1,$D4,$98,$FE
	.dc.b	$3F,$E3,$FF,$28,$61,$FB,$93,$C1,$D1,$F9,$59,$FC,$D1,$1F,$CC,$7C
	.dc.b	$7F,$E7,$48,$E0,$FC,$E7,$C7,$FE,$87,$D6,$91,$F9,$F7,$F9,$97,$D0
	.dc.b	$FF,$18,$8F,$F3,$10,$7F,$8C,$3F,$E3,$06,$BF,$E8,$36,$BF,$E8,$86
	.dc.b	$BF,$A7,$F9,$0E,$BE,$A4,$8F,$D7,$47,$EC,$FE,$42,$47,$EC,$5F,$F0
	.dc.b	$4F,$D3,$23,$FC,$6B,$C0,$F1,$F1,$39,$F1,$F7,$5A,$9C,$B4,$4F,$3F
	.dc.b	$97,$DF,$C6,$49,$FE,$6D,$3D,$E3,$9F,$C6,$C9,$C9,$F8,$AB,$EB,$1E
	.dc.b	$0E,$5F,$0F,$C6,$FF,$95,$30,$F4,$36,$CD,$31,$F1,$BC,$23,$AC,$7A
	.dc.b	$A3,$F8,$3F,$D1,$37,$A4,$AF,$1F,$C6,$FC,$03,$0F,$F0,$39,$E0,$E8
	.dc.b	$F4,$FC,$6F,$8C,$8F,$FA,$EF,$1F,$9D,$1E,$7F,$8D,$F9,$6C,$9F,$83
	.dc.b	$1F,$C6,$BE,$99,$E2,$7F,$9C,$BC,$D4,$62,$BF,$D8,$D1,$89,$D4,$E6
	.dc.b	$7F,$D9,$B8,$7D,$DF,$8E,$E9,$52,$7E,$7C,$7F,$81,$7E,$8D,$E3,$FF
	.dc.b	$82,$9B,$3F,$E8,$26,$BF,$ED,$56,$BF,$ED,$A7,$C1,$CF,$F8,$98,$FC
	.dc.b	$64,$70,$8C,$42,$31,$C9,$F9,$18,$C6,$1F,$E1,$7F,$C5,$DE,$0F,$F8
	.dc.b	$71,$B3,$FE,$94,$6B,$FE,$E7,$6B,$CF,$3F,$CE,$7C,$87,$3A,$93,$F8
	.dc.b	$E8,$FC,$DF,$C5,$48,$FC,$8B,$F9,$EF,$13,$F3,$FF,$21,$F8,$D2,$4F
	.dc.b	$DB,$0F,$E4,$7E,$43,$71,$F9,$A8,$FD,$3C,$7E,$A3,$E4,$24,$FE,$79
	.dc.b	$FD,$5F,$C8,$7E,$B3,$E4,$3F,$34,$49,$EC,$FE,$77,$E3,$BF,$63,$F1
	.dc.b	$D2,$7F,$4F,$1F,$BA,$F9,$3F,$95,$1F,$EE,$DD,$D1,$FE,$31,$E4,$FA
	.dc.b	$91,$8F,$8F,$E1,$5C,$7E,$5E,$3F,$B6,$BF,$F1,$9D,$F0,$7F,$8C,$37
	.dc.b	$C8,$FD,$94,$7F,$12,$9F,$E0,$2D,$C3,$C3,$3A,$F8,$FF,$37,$A5,$2F
	.dc.b	$B4,$7A,$7E,$3F,$99,$23,$F5,$18,$7F,$9A,$5B,$F1,$FE,$02,$38,$3F
	.dc.b	$DA,$FF,$2A,$F0,$9A,$7F,$19,$F2,$FF,$1D,$39,$8C,$57,$F2,$18,$C2
	.dc.b	$F3,$15,$C2,$F9,$CE,$63,$E7,$AF,$95,$E6,$FF,$B4,$CD,$1F,$51,$CB
	.dc.b	$FC,$BD,$F0,$EF,$F5,$BE,$47,$F4,$B3,$FC,$C3,$44,$62,$3E,$6B,$67
	.dc.b	$87,$81,$F1,$BE,$43,$16,$69,$23,$E5,$FC,$7F,$ED,$76,$FB,$9F,$21
	.dc.b	$67,$C1,$37,$2F,$B7,$37,$1F,$C5,$FC,$7F,$62,$5F,$D6,$7C,$3F,$21
	.dc.b	$BC,$CD,$F9,$0C,$39,$F8,$FD,$7C,$85,$C3,$F2,$0F,$1F,$90,$E5,$F1
	.dc.b	$B6,$7F,$28,$F5,$F9,$0F,$79,$23,$F0,$EF,$C2,$F8,$BF,$11,$E9,$E1
	.dc.b	$FE,$B9,$3F,$9D,$8C,$7C,$DE,$7E,$42,$CF,$F5,$6A,$FF,$26,$36,$3E
	.dc.b	$84,$63,$8B,$F9,$6B,$5F,$E2,$0D,$7F,$B4,$AF,$8F,$F1,$1E,$3F,$8F
	.dc.b	$B7,$DF,$E3,$FE,$85,$BE,$81,$C5,$E9,$9C,$C6,$2B,$FC,$62,$DA,$FC
	.dc.b	$9B,$5F,$E0,$6D,$7F,$97,$35,$FF,$00,$7C,$FB,$83,$C2,$FF,$9A,$2B
	.dc.b	$FC,$7A,$E5,$BC,$1A,$3F,$29,$49,$E0,$FE,$91,$AF,$F1,$46,$BF,$13
	.dc.b	$E4,$7F,$1D,$EC,$BE,$73,$E6,$6F,$FC,$58,$4E,$A3,$38,$FF,$85,$EB
	.dc.b	$87,$07,$E9,$5B,$1F,$D9,$27,$E9,$60,$9C,$BF,$63,$E4,$BF,$09,$F2
	.dc.b	$5F,$85,$6D,$FE,$06,$3E,$F9,$1F,$EF,$59,$23,$F0,$7F,$25,$EA,$F9
	.dc.b	$17,$C1,$FE,$3E,$30,$CB,$FF,$48,$4E,$9F,$C7,$4E,$71,$FE,$71,$87
	.dc.b	$F7,$29,$FC,$97,$B7,$F9,$9F,$77,$F3,$4C,$8F,$1B,$F9,$42,$3E,$6F
	.dc.b	$92,$FD,$01,$FA,$1C,$6C,$7E,$77,$DC,$FD,$21,$8C,$7C,$2F,$E9,$FE
	.dc.b	$3A,$4F,$E8,$FE,$3A,$D7,$9B,$E3,$BF,$58,$63,$4F,$B0,$FE,$BF,$E3
	.dc.b	$BF,$37,$F3,$DE,$FA,$F6,$C6,$38,$26,$3F,$CA,$A5,$5B,$7F,$AD,$B5
	.dc.b	$FF,$03,$35,$FF,$27,$E1,$FE,$39,$9F,$DC,$7C,$4F,$83,$E2,$7F,$80
	.dc.b	$3F,$C0,$7E,$3A,$4F,$F9,$4B,$9B,$F9,$F3,$FB,$F8,$FE,$3D,$1F,$D0
	.dc.b	$23,$5E,$87,$0C,$7F,$B6,$4E,$5B,$F1,$DF,$9A,$FD,$69,$97,$28,$C3
	.dc.b	$30,$FF,$68,$4E,$9F,$14,$E6,$BF,$D0,$7F,$38,$F7,$BF,$81,$FC,$CB
	.dc.b	$DD,$FE,$0B,$39,$C3,$FD,$63,$39,$6C,$FE,$D2,$31,$7F,$D5,$9A,$FF
	.dc.b	$81,$9A,$FF,$8C,$69,$F0,$7D,$8F,$0F,$BB,$70,$F9,$1F,$23,$F1,$8F
	.dc.b	$8F,$F2,$36,$3E,$67,$C8,$FC,$AE,$6F,$96,$D8,$FA,$1C,$1F,$38,$F9
	.dc.b	$FE,$E5,$8F,$A9,$F2,$1F,$48,$C6,$BF,$41,$FA,$BF,$1F,$F5,$BE,$36
	.dc.b	$D7,$DB,$70,$FF,$BF,$7E,$7D,$B6,$30,$FF,$7D,$46,$27,$53,$F8,$0F
	.dc.b	$8E,$FE,$27,$39,$AF,$BF,$1F,$84,$F8,$EC,$FC,$35,$C3,$B3,$86,$9F
	.dc.b	$8D,$8C,$57,$E2,$3E,$3B,$F1,$A7,$E3,$BE,$3A,$D7,$E4,$E3,$F2,$3C
	.dc.b	$98,$AF,$C7,$BF,$94,$F8,$EF,$CA,$FC,$75,$BF,$CD,$38,$7F,$D1,$DD
	.dc.b	$A8,$FF,$03,$36,$33,$18,$EB,$FC,$F9,$FA,$0F,$8E,$B3,$FE,$B6,$F8
	.dc.b	$9F,$F6,$23,$7E,$3B,$FC,$05,$18,$CF,$FC,$9C,$A3,$3F,$1D,$FE,$15
	.dc.b	$83,$49,$5F,$EA,$38,$77,$74,$EF,$EA,$D9,$F1,$72,$F8,$3E,$3B,$C2
	.dc.b	$D8,$F3,$72,$77,$23,$D5,$F1,$7C,$77,$A3,$E3,$3D,$2D,$9F,$77,$C6
	.dc.b	$7B,$0F,$6F,$C6,$58,$F8,$5F,$19,$EB,$23,$E9,$7D,$DF,$17,$F1,$0F
	.dc.b	$8B,$F1,$56,$7E,$57,$C1,$FC,$8F,$8A,$F9,$2D,$8F,$9D,$F1,$5F,$1C
	.dc.b	$8F,$C5,$7E,$57,$C1,$FD,$03,$E8,$FC,$1D,$9F,$F7,$C7,$A5,$FE,$AC
	.dc.b	$F9,$7F,$97,$1F,$C1,$35,$FF,$05,$C6,$3A,$7F,$CC,$73,$FE,$AF,$E5
	.dc.b	$57,$F8,$FF,$9D,$E9,$34,$FF,$28,$46,$1B,$19,$BF,$F3,$0C,$E5,$FB
	.dc.b	$9F,$17,$F8,$63,$F0,$FF,$17,$63,$CE,$32,$FF,$37,$7C,$F7,$F8,$D1
	.dc.b	$F2,$9F,$E7,$1C,$15,$E5,$9D,$6B,$FE,$CD,$C7,$CF,$3F,$E5,$5F,$8E
	.dc.b	$E5,$87,$FB,$05,$AF,$74,$61,$AD,$33,$F1,$BC,$CE,$9F,$0F,$73,$F4
	.dc.b	$D7,$0F,$8D,$F0,$7C,$6F,$85,$B3,$AC,$3B,$FC,$6F,$F9,$DD,$34,$91
	.dc.b	$FE,$8A,$8F,$2F,$C5,$FA,$CF,$67,$C5,$59,$E9,$3F,$E1,$16,$FC,$4F
	.dc.b	$C1,$F8,$AF,$84,$C8,$E9,$28,$F1,$12,$33,$F1,$5F,$27,$E2,$BE,$53
	.dc.b	$63,$BC,$5D,$7F,$D9,$97,$FE,$15,$6C,$7D,$1F,$8A,$FA,$5F,$15,$6F
	.dc.b	$CE,$E1,$F4,$FC,$55,$BE,$6F,$F9,$83,$E7,$3F,$CC,$DE,$6F,$FA,$97
	.dc.b	$E0,$7F,$0D,$49,$CB,$FE,$AF,$8C,$C7,$F9,$CF,$E4,$11,$8E,$3F,$E2
	.dc.b	$95,$B3,$F2,$15,$CA,$B7,$18,$9D,$C6,$23,$C1,$CA,$CB,$C9,$1B,$C6
	.dc.b	$BE,$5B,$E3,$8F,$86,$73,$8F,$96,$31,$5E,$33,$84,$EA,$B8,$C7,$9B
	.dc.b	$0F,$03,$BB,$F4,$5F,$2C,$7C,$D1,$89,$E1,$8F,$3C,$64,$7F,$B5,$A7
	.dc.b	$E4,$BF,$30,$F0,$99,$C7,$E3,$B9,$F9,$17,$BC,$F9,$1F,$F7,$6B,$67
	.dc.b	$FD,$C3,$3A,$9C,$9F,$E3,$BB,$74,$7F,$BA,$EB,$C4,$D7,$C5,$1A,$EB
	.dc.b	$E1,$39,$9F,$BA,$F1,$D3,$DE,$DD,$73,$5F,$3B,$A7,$F0,$F5,$A8,$FA
	.dc.b	$35,$BB,$FC,$0B,$5F,$C6,$57,$0B,$E7,$1F,$3E,$75,$87,$96,$3E,$95
	.dc.b	$6D,$E7,$7D,$88,$E8,$B8,$78,$DE,$91,$E7,$7F,$16,$BE,$7B,$F2,$1D
	.dc.b	$BF,$80,$E6,$FE,$2F,$F1,$FF,$A2,$8C,$42,$31,$C5,$FC,$AE,$F9,$6B
	.dc.b	$8C,$51,$D7,$E2,$30,$EA,$3F,$8D,$B5,$FE,$62,$ED,$CE,$4F,$85,$E9
	.dc.b	$97,$F2,$AF,$8F,$A9,$EB,$3B,$AE,$35,$82,$3D,$3E,$1F,$8F,$C7,$FF
	.dc.b	$A0,$70,$FE,$6A,$D4,$D3,$FD,$0D,$8F,$C7,$F8,$8E,$37,$FC,$92,$7C
	.dc.b	$31,$E4,$F8,$F9,$3F,$CA,$1F,$2F,$3F,$9B,$E2,$39,$E8,$69,$2B,$FA
	.dc.b	$EC,$3E,$17,$DB,$F1,$12,$7E,$2C,$7C,$2F,$8E,$91,$FC,$D9,$F8,$7F
	.dc.b	$0D,$F1,$BC,$CF,$61,$27,$F9,$AB,$EC,$F9,$E7,$BB,$F5,$BC,$EF,$BD
	.dc.b	$5F,$59,$AF,$D6,$9D,$BF,$D3,$FE,$4F,$F5,$9D,$7F,$89,$6F,$57,$F6
	.dc.b	$16,$CF,$E1,$36,$7C,$3F,$E1,$30,$7B,$9E,$32,$57,$F7,$E8,$CB,$FB
	.dc.b	$DE,$3F,$F5,$C3,$FC,$D5,$BE,$01,$A7,$DD,$C5,$F2,$1F,$8A,$6D,$7A
	.dc.b	$63,$F1,$3F,$0D,$9E,$EE,$04,$9F,$EC,$D4,$8F,$C1,$FC,$F7,$F9,$3A
	.dc.b	$7F,$36,$D7,$F3,$6D,$DF,$FC,$10,$FF,$1E,$7E,$37,$81,$E3,$24,$F2
	.dc.b	$F6,$3C,$DF,$1F,$E9,$24,$78,$9F,$0F,$C2,$FA,$FC,$29,$3E,$28,$FD
	.dc.b	$7E,$2F,$84,$F8,$1E,$1C,$AF,$3C,$7C,$3F,$03,$E3,$12,$3D,$EF,$BB
	.dc.b	$93,$99,$23,$DF,$ED,$F6,$F8,$1F,$99,$F1,$F2,$7E,$1C,$7C,$FF,$0F
	.dc.b	$E9,$12,$3E,$5B,$F2,$BD,$DF,$39,$23,$E5,$FC,$17,$D7,$F1,$64,$7E
	.dc.b	$3D,$F6,$FC,$0F,$BB,$E0,$7E,$99,$27,$FC,$8B,$C0,$8F,$AF,$F0,$1B
	.dc.b	$9F,$AF,$1F,$1A,$3F,$8A,$FC,$F3,$E0,$29,$2B,$F8,$BC,$7F,$64,$7F
	.dc.b	$8B,$B7,$A7,$FC,$88,$F1,$8D,$CF,$4F,$0F,$F9,$27,$C6,$3E,$7B,$C9
	.dc.b	$5F,$A9,$F3,$3F,$28,$6A,$33,$1F,$E5,$49,$E1,$9E,$49,$1E,$4F,$5F
	.dc.b	$D3,$EA,$C8,$F3,$BC,$FE,$0E,$47,$F7,$11,$7E,$38,$C4,$63,$0F,$F3
	.dc.b	$0B,$B9,$D3,$FE,$70,$D3,$FC,$96,$FF,$95,$23,$2F,$9E,$78,$BF,$1B
	.dc.b	$D6,$F9,$BF,$05,$FB,$1F,$90,$93,$F3,$23,$E8,$FC,$1E,$A3,$07,$D0
	.dc.b	$7E,$AF,$C0,$FE,$E7,$D4,$93,$F5,$23,$EC,$FC,$14,$8F,$AE,$FD,$CF
	.dc.b	$82,$FE,$01,$EA,$C9,$FE,$30,$FF,$57,$5F,$0A,$F1,$B1,$F3,$BF,$E3
	.dc.b	$39,$FF,$0C,$35,$FF,$22,$B5,$FF,$46,$35,$2F,$27,$9F,$97,$27,$FC
	.dc.b	$F6,$6A,$3F,$0B,$F3,$BF,$CB,$BB,$3F,$95,$7C,$97,$F1,$73,$3E,$FC
	.dc.b	$BF,$A1,$5B,$ED,$FF,$4E,$73,$A3,$13,$FD,$CB,$A1,$F0,$BF,$9E,$C3
	.dc.b	$F9,$17,$C9,$4A,$FC,$EC,$7E,$8B,$E0,$75,$18,$3F,$42,$FF,$29,$F6
	.dc.b	$64,$7F,$41,$7F,$53,$EB,$FC,$BF,$9D,$FF,$76,$CF,$E3,$1A,$FE,$45
	.dc.b	$AF,$F9,$0D,$AF,$F9,$C1,$7C,$87,$3F,$56,$4F,$C8,$9C,$BF,$B4,$F7
	.dc.b	$FF,$9E,$7C,$9C,$8C,$1F,$EA,$33,$FA,$8F,$C1,$B7,$FF,$67,$C6,$8F
	.dc.b	$F6,$0F,$C2,$7F,$86,$DE,$1E,$D7,$F4,$7F,$72,$46,$0E,$9D,$3F,$E1
	.dc.b	$47,$FA,$21,$93,$99,$FE,$2C,$E6,$FA,$BC,$BE,$08,$FF,$31,$73,$7D
	.dc.b	$7D,$7F,$62,$D8,$FD,$84,$ED,$FF,$49,$E5,$FE,$D9,$F6,$FF,$0F,$72
	.dc.b	$7D,$D1,$FC,$3B,$E7,$5F,$09,$FD,$7F,$E5,$25,$7E,$DE,$72,$D9,$E0
	.dc.b	$47,$E4,$7B,$9F,$C6,$3E,$5F,$FC,$C9,$1F,$8E,$6B,$FC,$29,$BA,$1C
	.dc.b	$E7,$FE,$EF,$F6,$F9,$71,$CB,$FA,$91,$FE,$47,$EF,$FF,$77,$F5,$7F
	.dc.b	$12,$6C,$FF,$17,$6B,$FE,$89,$6B,$FE,$B6,$C5,$F3,$AF,$23,$A7,$9F
	.dc.b	$FE,$0E,$AF,$F7,$4D,$F2,$3F,$D3,$B7,$C2,$BF,$81,$7C,$EF,$FB,$5E
	.dc.b	$BF,$6B,$F6,$BC,$55,$FF,$41,$8C,$3F,$D0,$E7,$84,$EB,$9B,$FD,$93
	.dc.b	$0F,$2F,$FC,$2E,$D9,$FE,$5E,$D7,$F9,$A3,$5F,$F5,$37,$CF,$3E,$47
	.dc.b	$FC,$3E,$D8,$FE,$6E,$D7,$F9,$E3,$5E,$3E,$37,$F1,$D8,$D6,$3F,$D6
	.dc.b	$5B,$8F,$FB,$79,$FE,$0B,$43,$F4,$7F,$59,$F3,$E7,$1C,$68,$C4,$A3
	.dc.b	$0F,$E8,$FD,$2F,$D2,$7A,$1F,$E6,$D8,$FF,$2D,$1F,$D4,$67,$B4,$64
	.dc.b	$D7,$9F,$C2,$7F,$71,$1F,$AD,$8F,$F4,$0D,$FF,$8A,$70,$F1,$B6,$BC
	.dc.b	$4D,$BF,$57,$95,$FB,$5E,$5F,$11,$2B,$C4,$7D,$D3,$F5,$A3,$B3,$F7
	.dc.b	$9E,$24,$8F,$E1,$51,$FE,$92,$F8,$FE,$D0,$F0,$FF,$81,$19,$F9,$1F
	.dc.b	$3F,$CD,$49,$F9,$9C,$5F,$EA,$DD,$7F,$C8,$9B,$7C,$CC,$6D,$7A,$BC
	.dc.b	$9F,$F4,$A5,$7F,$24,$8C,$1F,$25,$C6,$FF,$EE,$5F,$47,$E2,$F9,$72
	.dc.b	$BB,$E3,$D6,$7E,$4F,$CD,$74,$24,$F3,$7B,$C4,$8E,$CD,$7F,$DD,$FE
	.dc.b	$07,$E8,$9F,$F5,$EF,$C5,$FE,$1D,$7C,$9A,$75,$E1,$FF,$9B,$A7,$31
	.dc.b	$8A,$E1,$5B,$DF,$FD,$95,$1F,$E1,$5F,$93,$B7,$D4,$38,$7F,$DD,$E7
	.dc.b	$88,$D3,$C3,$5F,$CB,$7C,$97,$F8,$DE,$7F,$0E,$D7,$F0,$ED,$8D,$EB
	.dc.b	$F3,$7E,$6B,$1C,$06,$92,$3F,$35,$C5,$FE,$98,$7E,$7E,$1F,$92,$F8
	.dc.b	$D4,$63,$05,$6E,$BF,$5B,$5A,$7F,$5F,$F3,$5E,$02,$5F,$DF,$7C,$7F
	.dc.b	$1B,$BD,$BF,$14,$D9,$FC,$57,$C6,$DB,$FD,$1C,$DC,$BC,$3F,$19,$73
	.dc.b	$7C,$55,$8C,$39,$F8,$6D,$7C,$4D,$CF,$D5,$68,$E9,$A7,$BA,$7F,$78
	.dc.b	$DC,$FE,$C6,$5F,$C8,$E7,$27,$CC,$F8,$EF,$27,$0F,$C3,$BF,$85,$5F
	.dc.b	$1B,$CE,$CD,$76,$F9,$6E,$FF,$11,$71,$F9,$98,$23,$31,$E2,$7E,$E4
	.dc.b	$66,$30,$7E,$A7,$E6,$BC,$27,$AB,$E2,$AC,$7F,$8D,$3E,$49,$F3,$5F
	.dc.b	$2F,$64,$3E,$47,$0E,$AB,$84,$60,$FC,$74,$F0,$6C,$F1,$9F,$F3,$7B
	.dc.b	$67,$FC,$E9,$F3,$FF,$E9,$79,$FF,$39,$B6,$33,$C7,$D2,$31,$1F,$A5
	.dc.b	$7E,$A4,$FF,$93,$3F,$05,$7C,$77,$A6,$73,$5F,$D5,$4F,$E7,$D3,$AA
	.dc.b	$FF,$3D,$7D,$03,$83,$F1,$5B,$BE,$4D,$F3,$7F,$8D,$F6,$D7,$BC,$3F
	.dc.b	$B6,$D6,$9B,$5F,$AF,$BE,$4D,$9D,$65,$FD,$ED,$AF,$F8,$7A,$5F,$21
	.dc.b	$DF,$DE,$FF,$7C,$E1,$F8,$08,$C7,$8B,$23,$FD,$2D,$3A,$BF,$F1,$6B
	.dc.b	$5F,$F0,$33,$5F,$F3,$3F,$B2,$F9,$7F,$63,$6B,$F1,$BF,$39,$89,$91
	.dc.b	$FB,$BE,$D7,$87,$8C,$3F,$B3,$BF,$A4,$F7,$FD,$1F,$11,$6F,$FD,$4D
	.dc.b	$9A,$33,$1E,$87,$F8,$A7,$17,$F8,$A5,$B3,$B9,$FD,$6F,$C9,$7E,$C3
	.dc.b	$DE,$FD,$8B,$63,$F4,$71,$FA,$CE,$CC,$67,$FD,$E9,$FD,$BF,$C7,$7E
	.dc.b	$E3,$96,$DF,$EA,$9C,$3E,$DE,$DB,$7F,$E0,$27,$F4,$5F,$41,$F1,$63
	.dc.b	$F1,$41,$7C,$2E,$B8,$B6,$CB,$FF,$66,$CE,$9B,$1F,$A6,$8C,$57,$F0
	.dc.b	$8E,$27,$93,$FC,$18,$FE,$1F,$F0,$D7,$0F,$E0,$3F,$13,$FC,$77,$8F
	.dc.b	$F0,$1D,$5F,$C2,$84,$67,$A6,$57,$D9,$F2,$EE,$7F,$E1,$E7,$F1,$DA
	.dc.b	$C9,$FE,$71,$8F,$D1,$AF,$F7,$97,$2F,$9B,$2F,$F2,$4B,$6B,$FC,$27
	.dc.b	$E1,$71,$AF,$F1,$4B,$FC,$27,$D8,$FF,$4D,$CF,$FB,$01,$7A,$F9,$F6
	.dc.b	$34,$C9,$FF,$79,$CE,$9B,$1F,$C3,$23,$17,$F1,$FC,$DF,$E5,$67,$F1
	.dc.b	$7F,$A0,$7C,$06,$30,$F5,$FC,$57,$C7,$3E,$47,$C6,$58,$F7,$F4,$7F
	.dc.b	$34,$F4,$3E,$63,$63,$E1,$F1,$7C,$FF,$9A,$FA,$0D,$8F,$8F,$1F,$D3
	.dc.b	$DF,$E8,$5C,$5F,$CF,$38,$D3,$99,$FB,$1F,$31,$FC,$2F,$E8,$3D,$77
	.dc.b	$B8,$C7,$14,$AF,$F0,$17,$0B,$16,$1F,$0A,$3F,$D0,$4D,$7F,$DC,$6D
	.dc.b	$7F,$CA,$D0,$F8,$8F,$C4,$1D,$63,$FD,$83,$19,$AF,$EF,$6F,$6F,$6F
	.dc.b	$CC,$7E,$3F,$E2,$2C,$EB,$04,$FC,$1F,$9A,$CE,$1F,$EC,$CA,$47,$E5
	.dc.b	$FE,$27,$F3,$1C,$96,$7F,$1E,$E1,$A7,$E4,$A3,$15,$E6,$F8,$8F,$D0
	.dc.b	$7C,$47,$E8,$5B,$5F,$98,$8F,$F0,$E6,$9F,$E5,$B8,$FF,$38,$36,$BF
	.dc.b	$BA,$B5,$FF,$66,$C7,$F3,$B7,$D1,$E9,$7E,$B7,$E2,$3E,$E3,$FE,$0B
	.dc.b	$F8,$57,$C4,$53,$15,$6F,$FC,$40,$3D,$AF,$F1,$73,$4F,$F4,$11,$63
	.dc.b	$FC,$0E,$D7,$FD,$BF,$5F,$A8,$7E,$57,$B9,$E1,$F0,$3F,$1C,$F8,$FE
	.dc.b	$1E,$5F,$F8,$BE,$FF,$16,$FF,$A6,$14,$7E,$57,$DB,$F1,$92,$7C,$6F
	.dc.b	$F9,$0E,$5F,$07,$F9,$BF,$A0,$F8,$0D,$AE,$DF,$09,$F9,$FF,$9A,$F8
	.dc.b	$8D,$8F,$83,$ED,$FC,$02,$3F,$01,$F8,$7E,$D7,$E9,$CF,$99,$F0,$36
	.dc.b	$7D,$5E,$9F,$CF,$3E,$87,$AD,$63,$E6,$FC,$77,$CC,$23,$F3,$1F,$9F
	.dc.b	$ED,$FE,$CC,$FB,$1E,$85,$9F,$E2,$5F,$0D,$FC,$2C,$8F,$F1,$07,$FC
	.dc.b	$CD,$C1,$FC,$AB,$1B,$E9,$FF,$85,$F9,$BF,$D0,$13,$9B,$FD,$AB,$63
	.dc.b	$ED,$FB,$7F,$C0,$CF,$C4,$7A,$16,$BF,$CC,$F5,$FC,$DB,$CD,$FF,$2C
	.dc.b	$27,$F9,$CC,$6B,$0C,$3C,$23,$38,$7E,$B7,$DE,$FC,$69,$27,$F1,$AF
	.dc.b	$E4,$3D,$CF,$E2,$87,$F9,$8F,$E1,$7F,$67,$3F,$96,$8C,$75,$E7,$AA
	.dc.b	$CF,$F9,$E6,$BF,$C4,$D8,$7F,$8A,$5F,$F7,$07,$47,$FA,$25,$34,$93
	.dc.b	$F9,$EF,$77,$F9,$21,$E1,$EB,$B1,$D7,$A3,$F9,$47,$CB,$79,$5B,$1F
	.dc.b	$AA,$8F,$D4,$91,$FD,$4B,$E1,$E6,$F6,$1E,$DE,$7B,$3E,$98,$FE,$69
	.dc.b	$EF,$63,$4F,$E2,$8F,$C3,$E0,$F8,$9C,$16,$FE,$03,$87,$F7,$7F,$2D
	.dc.b	$6F,$F8,$E3,$FE,$65,$F9,$DF,$DF,$78,$69,$FF,$5B,$FA,$DF,$EC,$59
	.dc.b	$CC,$60,$D3,$FC,$3F,$E2,$D3,$BF,$17,$F8,$6D,$6C,$57,$2A,$DC,$72
	.dc.b	$9D,$C6,$23,$7E,$37,$FA,$B6,$31,$5C,$6F,$B6,$1E,$79,$CE,$1C,$A7
	.dc.b	$D1,$7C,$5D,$E1,$E3,$C3,$93,$E1,$C3,$94,$F0,$CA,$CE,$4E,$18,$75
	.dc.b	$71,$8F,$93,$1E,$D1,$8B,$EE,$FF,$B8,$71,$F9,$2D,$BF,$8A,$DC,$7F
	.dc.b	$B2,$BF,$EF,$2F,$57,$FD,$9E,$FF,$68,$F8,$CF,$EE,$3F,$17,$A6,$C6
	.dc.b	$67,$6F,$7E,$A7,$6F,$4C,$BF,$D5,$CD,$8C,$CF,$F7,$49,$CB,$63,$3A
	.dc.b	$FD,$06,$E9,$F8,$36,$CF,$E0,$BA,$FF,$10,$FC,$D8,$CB,$F8,$06,$DF
	.dc.b	$6B,$FB,$75,$C1,$ED,$87,$D0,$BE,$0D,$AE,$18,$7E,$4E,$F8,$3E,$39
	.dc.b	$D6,$1D,$8E,$F3,$AB,$FC,$BC,$66,$76,$75,$9D,$75,$FF,$C0,$0F,$80
	.dc.b	$F5,$FB,$1F,$ED,$CE,$2F,$F2,$9E,$3F,$D9,$3B,$F3,$1F,$A7,$BE,$13
	.dc.b	$FD,$A2,$F2,$A3,$2F,$F8,$2E,$33,$0F,$B9,$FA,$73,$FC,$55,$7C,$0E
	.dc.b	$9F,$F4,$84,$7E,$D7,$1F,$F7,$7C,$62,$38,$4E,$C8,$EF,$0F,$34,$F4
	.dc.b	$8C,$1E,$57,$F6,$14,$9F,$B7,$F3,$8F,$8E,$FC,$AD,$A3,$0F,$E6,$1A
	.dc.b	$FF,$8F,$9A,$FE,$13,$1F,$77,$CD,$FC,$48,$E5,$56,$B8,$E1,$EC,$AF
	.dc.b	$81,$F0,$12,$FC,$F8,$7C,$78,$F8,$9F,$01,$F1,$7E,$03,$F5,$33,$F2
	.dc.b	$DB,$A7,$FA,$72,$FF,$CA,$33,$96,$C3,$E4,$3E,$71,$23,$BC,$7F,$98
	.dc.b	$1B,$F3,$DF,$4C,$8E,$BE,$0B,$EB,$7C,$0F,$80,$91,$D9,$F1,$FC,$0F
	.dc.b	$DB,$24,$76,$F7,$7E,$F7,$CE,$48,$F0,$BF,$81,$F7,$FF,$05,$F2,$9E
	.dc.b	$C2,$4F,$A9,$F7,$FC,$07,$E1,$C9,$1E,$AF,$7B,$F1,$5F,$01,$F9,$0F
	.dc.b	$07,$F6,$F3,$FD,$8D,$BC,$DF,$8C,$9D,$78,$BF,$32,$7F,$8B,$D7,$DB
	.dc.b	$6B,$FA,$16,$BF,$9F,$5F,$29,$DF,$E1,$E4,$FD,$A9,$CB,$7E,$07,$C9
	.dc.b	$F0,$D2,$30,$6F,$E0,$BC,$44,$7E,$DD,$7F,$91,$1B,$F0,$BE,$D2,$3D
	.dc.b	$FB,$FE,$07,$C1,$FC,$1F,$83,$93,$E3,$8F,$89,$EF,$FB,$C9,$1D,$9F
	.dc.b	$85,$F2,$9F,$24,$91,$EB,$F8,$3F,$99,$F1,$FF,$37,$E1,$24,$FB,$E3
	.dc.b	$E8,$7C,$1F,$CB,$24,$7A,$9F,$9D,$F1,$9F,$54,$91,$F1,$FE,$17,$EC
	.dc.b	$7C,$5C,$8F,$94,$FF,$0E,$F4,$1F,$42,$F8,$0C,$7D,$07,$F9,$8A,$FF
	.dc.b	$C7,$AD,$7F,$03,$48,$C4,$7F,$91,$1A,$FD,$DF,$87,$E7,$C3,$23,$ED
	.dc.b	$47,$E2,$FE,$1F,$F8,$DF,$E9,$DF,$C2,$67,$F6,$ED,$7F,$1D,$1B,$7F
	.dc.b	$04,$D7,$FC,$69,$1F,$EA,$E7,$8F,$BD,$D0,$8F,$4F,$7B,$B9,$87,$8C
	.dc.b	$61,$F0,$FB,$3E,$4F,$66,$46,$0F,$2F,$C3,$7A,$26,$46,$0F,$4F,$C4
	.dc.b	$7B,$26,$46,$0F,$6F,$C5,$7C,$0F,$88,$91,$83,$E0,$FC,$47,$C4,$99
	.dc.b	$18,$3E,$2F,$C4,$7C,$8F,$8A,$91,$83,$E4,$FC,$47,$CC,$23,$F3,$3E
	.dc.b	$23,$F9,$57,$9D,$F1,$E1,$E6,$FC,$F8,$C3,$7E,$33,$F9,$87,$C2,$BE
	.dc.b	$2A,$FC,$57,$F3,$C8,$34,$98,$7F,$85,$3E,$8F,$FD,$E0,$FF,$48,$3F
	.dc.b	$C7,$71,$C9,$C4,$69,$FB,$7F,$19,$F8,$42,$3F,$84,$EE,$FC,$39,$87
	.dc.b	$66,$4F,$C4,$E1,$F8,$CB,$91,$83,$F1,$BF,$0F,$F9,$09,$91,$83,$F2
	.dc.b	$3F,$11,$F9,$42,$3E,$9F,$A0,$FE,$22,$FC,$8F,$88,$FF,$2F,$5F,$C3
	.dc.b	$24,$7F,$8A,$5F,$F1,$AF,$C6,$BE,$49,$EF,$FC,$C9,$95,$FD,$B2,$B5
	.dc.b	$3F,$E0,$27,$87,$BF,$FE,$00,$F6,$BF,$BF,$F1,$FE,$9F,$E1,$7F,$DE
	.dc.b	$7E,$46,$F3,$5E,$FE,$2F,$0C,$CF,$FB,$1E,$3F,$DE,$DF,$50,$F8,$F7
	.dc.b	$C9,$3A,$7C,$1F,$29,$2B,$FC,$03,$5A,$9F,$27,$C7,$FF,$86,$23,$13
	.dc.b	$AB,$E4,$F8,$27,$5E,$9F,$F8,$12,$8D,$24,$FF,$74,$C5,$F1,$DC,$22
	.dc.b	$1E,$A7,$9F,$C0,$FC,$5F,$8E,$93,$E6,$BF,$AF,$C5,$FC,$37,$1E,$F7
	.dc.b	$C2,$B4,$FC,$DF,$8E,$F2,$FC,$0F,$F9,$B2,$BC,$AD,$C0,$D3,$CD,$F1
	.dc.b	$DC,$30,$F0,$47,$D7,$8F,$F0,$C7,$E3,$4F,$4F,$FC,$46,$DA,$FF,$3C
	.dc.b	$B5,$FE,$DA,$1C,$3F,$E6,$6E,$86,$EF,$43,$AD,$CA,$FC,$54,$E5,$BE
	.dc.b	$7F,$F3,$FF,$D2,$9E,$E5,$3F,$1D,$49,$A4,$AE,$27,$FB,$E3,$F0,$07
	.dc.b	$C0,$FF,$8C,$FE,$79,$E8,$FF,$8F,$9B,$5F,$E9,$46,$BE,$C6,$BF,$E0
	.dc.b	$A5,$E8,$73,$99,$3F,$C1,$E7,$2F,$E6,$3E,$4B,$F4,$44,$7F,$37,$F2
	.dc.b	$9F,$A5,$F9,$2F,$C4,$12,$3F,$C7,$3A,$23,$F1,$9F,$25,$B9,$FD,$1C
	.dc.b	$7E,$B6,$3F,$5D,$F2,$52,$7F,$4E,$FE,$C7,$C7,$FD,$97,$87,$F9,$52
	.dc.b	$4F,$67,$F3,$1F,$25,$FB,$72,$47,$EC,$3E,$4B,$F7,$44,$70,$7E,$D3
	.dc.b	$E4,$BF,$7D,$E2,$C8,$FD,$BB,$FC,$07,$E4,$BF,$81,$7C,$97,$E9,$49
	.dc.b	$3E,$E7,$FC,$B7,$F8,$4F,$F8,$D3,$CF,$FF,$47,$36,$7F,$D1,$8D,$F3
	.dc.b	$8F,$FB,$33,$C6,$FE,$18,$FF,$1E,$9F,$F3,$93,$5E,$7E,$37,$63,$BF
	.dc.b	$C9,$C8,$FF,$33,$CA,$31,$1F,$ED,$3E,$6F,$E9,$EF,$FB,$3B,$CD,$FF
	.dc.b	$60,$FC,$B7,$F9,$AF,$14,$E4,$B1,$97,$01,$C7,$A1,$FC,$6E,$3F,$DF
	.dc.b	$86,$8A,$E5,$7C,$97,$68,$C3,$6F,$E3,$FC,$A7,$F8,$46,$AE,$DE,$C8
	.dc.b	$C4,$7C,$A6,$ED,$F2,$BC,$9B,$B7,$F0,$6B,$F8,$71,$83,$3F,$29,$FE
	.dc.b	$1F,$D7,$E7,$B6,$7E,$BF,$CA,$5C,$BE,$BC,$DC,$7E,$BF,$CA,$76,$25
	.dc.b	$FD,$07,$C3,$F1,$7B,$CC,$DF,$13,$0E,$7E,$52,$5F,$FA,$9A,$9F,$29
	.dc.b	$8C,$FF,$A2,$E5,$F8,$6F,$94,$E8,$67,$6F,$C3,$BD,$FD,$1F,$82,$4C
	.dc.b	$7F,$06,$FC,$4F,$2F,$C8,$79,$7E,$52,$CF,$87,$E5,$3D,$1F,$29,$E9
	.dc.b	$6C,$79,$63,$E6,$91,$FC,$7B,$F4,$3E,$53,$E0,$1F,$07,$D1,$B3,$FE
	.dc.b	$11,$F9,$D5,$7C,$DD,$90,$BC,$3A,$AE,$11,$83,$F0,$33,$C1,$B3,$C6
	.dc.b	$7F,$9C,$CE,$5F,$E7,$7E,$A0,$FE,$B6,$FF,$7F,$6C,$7F,$72,$EC,$7C
	.dc.b	$07,$F3,$3A,$D7,$B7,$F8,$3B,$E4,$FE,$16,$9F,$45,$FF,$00,$2B,$67
	.dc.b	$FB,$DA,$33,$DD,$C6,$DE,$FF,$C1,$FD,$31,$F1,$F9,$8C,$33,$0F,$5C
	.dc.b	$E9,$B1,$F0,$A3,$1A,$7F,$B1,$39,$FF,$85,$FC,$AF,$E6,$7D,$EF,$EF
	.dc.b	$BE,$0F,$E4,$E3,$1D,$08,$A7,$EB,$27,$79,$7F,$62,$6E,$1F,$ED,$06
	.dc.b	$BF,$C1,$8F,$F0,$C5,$7E,$7A,$7F,$C0,$1D,$DF,$A7,$FD,$F9,$1E,$08
	.dc.b	$8F,$F0,$EE,$CF,$F1,$54,$BE,$37,$38,$7E,$82,$31,$5B,$3F,$8A,$E8
	.dc.b	$8F,$EE,$51,$FE,$0F,$EE,$F0,$7B,$DE,$E6,$47,$AF,$3F,$D6,$B6,$3E
	.dc.b	$04,$63,$0F,$95,$6F,$95,$F8,$F3,$EB,$6D,$FF,$87,$63,$DD,$EF,$62
	.dc.b	$FD,$0F,$C1,$F9,$5F,$C4,$67,$FE,$00,$BF,$55,$23,$D5,$3A,$F0,$65
	.dc.b	$6A,$7D,$DF,$29,$8B,$F8,$8F,$4C,$7F,$55,$E7,$7F,$81,$2F,$FC,$B5
	.dc.b	$7F,$E3,$C7,$FD,$01,$19,$BF,$F2,$79,$FD,$17,$C7,$FD,$5C,$66,$3E
	.dc.b	$E1,$99,$FD,$5C,$66,$3E,$8E,$AF,$95,$FB,$C7,$EE,$FE,$75,$E8,$63
	.dc.b	$0F,$55,$3E,$5B,$F0,$9D,$7F,$85,$6C,$EA,$3F,$11,$F0,$38,$CB,$F0
	.dc.b	$4F,$E2,$FE,$0B,$F1,$9F,$2D,$6F,$AB,$86,$9E,$08,$C5,$7C,$0F,$9B
	.dc.b	$FE,$C9,$3F,$13,$CF,$FE,$25,$18,$9F,$8B,$BA,$7E,$69,$F1,$89,$5F
	.dc.b	$11,$FE,$27,$5F,$AB,$8F,$F5,$73,$6B,$9F,$47,$6E,$EF,$D2,$7C,$BC
	.dc.b	$BF,$8B,$F2,$F6,$FF,$41,$F2,$BF,$A9,$F9,$5F,$D5,$36,$BE,$5F,$CA
	.dc.b	$7E,$B7,$E5,$7F,$5C,$D8,$F9,$F1,$FB,$1F,$94,$C6,$1F,$AB,$7F,$69
	.dc.b	$F2,$BF,$D6,$FF,$4F,$79,$B8,$C3,$F2,$11,$86,$C6,$FC,$DE,$0D,$8D
	.dc.b	$C7,$3F,$33,$18,$FF,$58,$BF,$E3,$F5,$D3,$E5,$3C,$1E,$87,$85,$B1
	.dc.b	$E0,$8B,$5C,$66,$F9,$5F,$EA,$FB,$7C,$A7,$A4,$F5,$7C,$A5,$C7,$D3
	.dc.b	$1E,$DF,$94,$C6,$3E,$37,$E0,$78,$3F,$CE,$77,$FF,$42,$5F,$C0,$8C
	.dc.b	$7B,$12,$3F,$D0,$7C,$0A,$DE,$1F,$A6,$8C,$57,$F8,$51,$B5,$FE,$8A
	.dc.b	$F1,$9A,$3E,$43,$E8,$1F,$D2,$7D,$27,$26,$2F,$E8,$FC,$7F,$D5,$3E
	.dc.b	$B7,$C7,$D8,$FA,$BF,$1F,$FE,$20,$C4,$D2,$47,$86,$3F,$DB,$EF,$DC
	.dc.b	$F8,$EF,$C0,$7C,$7F,$E0,$5F,$C1,$7C,$7C,$9F,$C0,$FC,$7D,$E1,$FB
	.dc.b	$5F,$1F,$DC,$93,$D5,$F1,$7C,$7F,$E2,$FE,$3F,$F1,$8D,$9F,$C5,$FC
	.dc.b	$7F,$E3,$CF,$F1,$57,$7F,$F9,$3A,$32,$7F,$1F,$7F,$D5,$AF,$C9,$6E
	.dc.b	$1F,$D7,$5A,$FF,$51,$E4,$46,$23,$FA,$2A,$F8,$FE,$FE,$5F,$81,$B7
	.dc.b	$DF,$E3,$7C,$7F,$1B,$E4,$6C,$78,$FE,$37,$FC,$9B,$81,$A4,$8F,$37
	.dc.b	$97,$EC,$F8,$DF,$6B,$63,$D9,$F0,$3F,$00,$F8,$3E,$CD,$8F,$81,$E8
	.dc.b	$F6,$23,$F0,$5F,$0F,$A1,$F2,$0F,$93,$F1,$96,$7E,$47,$B7,$F3,$0F
	.dc.b	$9B,$EE,$D8,$F9,$9F,$19,$EA,$23,$F3,$5F,$6F,$C6,$7D,$43,$EA,$FC
	.dc.b	$65,$9F,$A9,$F1,$9F,$60,$FB,$3E,$45,$8F,$B1,$E3,$7C,$62,$3F,$65
	.dc.b	$FF,$3A,$FF,$00,$FF,$80,$A3,$FA,$E3,$7A,$7F,$CB,$AF,$F7,$E6,$C7
	.dc.b	$F9,$BE,$3F,$BE,$C7,$F3,$9E,$6C,$7C,$37,$F3,$86,$57,$F9,$03,$DA
	.dc.b	$47,$F3,$42,$47,$EC,$EF,$FB,$E3,$E2,$F8,$6E,$DF,$0D,$DD,$91,$E6
	.dc.b	$F8,$6B,$97,$99,$F2,$7C,$34,$9F,$5F,$C3,$5A,$F5,$CF,$A7,$E1,$31
	.dc.b	$EB,$FF,$88,$A7,$FC,$01,$9F,$F6,$3C,$3F,$C0,$70,$F9,$9F,$F0,$87
	.dc.b	$C9,$7F,$AC,$BE,$45,$ED,$FF,$8B,$E7,$35,$FE,$AA,$8C,$71,$75,$AD
	.dc.b	$CF,$FA,$EA,$B7,$18,$8E,$DB,$B0,$ED,$87,$58,$C4,$F5,$39,$61,$D6
	.dc.b	$75,$5C,$27,$B5,$70,$BE,$D8,$78,$2F,$CB,$8E,$EF,$C3,$85,$AD,$47
	.dc.b	$28,$C6,$3D,$E3,$07,$8C,$E2,$59,$E2,$FC,$98,$E0,$EB,$0F,$1E,$3E
	.dc.b	$C8,$FF,$0E,$D7,$CA,$6C,$65,$F6,$9F,$E2,$23,$FD,$BD,$EA,$FF,$A2
	.dc.b	$5C,$7C,$97,$F3,$BF,$8E,$FF,$47,$B6,$33,$3B,$7B,$F1,$FF,$6E,$7A
	.dc.b	$60,$8C,$46,$BD,$8F,$F4,$CC,$E5,$B3,$C7,$2F,$9A,$DD,$BF,$08,$DB
	.dc.b	$FC,$07,$27,$F9,$D6,$B7,$7F,$3E,$B5,$5F,$8F,$6E,$0B,$E1,$87,$4B
	.dc.b	$E0,$4B,$FA,$13,$A7,$F1,$D7,$C8,$93,$F7,$67,$53,$DE,$39,$61,$E4
	.dc.b	$8E,$67,$E2,$FC,$BF,$F4,$3F,$B3,$FE,$8D,$E3,$FD,$44,$62,$7F,$CC
	.dc.b	$9C,$5F,$DC,$DB,$EC,$EE,$3F,$D0,$A7,$F6,$C8,$DD,$E5,$FF,$38,$47
	.dc.b	$17,$FD,$DA,$ED,$FF,$77,$FC,$EF,$FA,$D9,$FF,$39,$35,$CF,$C0,$73
	.dc.b	$C1,$3F,$D4,$A7,$7E,$66,$63,$86,$5E,$A6,$E1,$FE,$9A,$78,$57,$F2
	.dc.b	$2F,$8A,$FF,$6C,$E0,$E8,$FF,$62,$CF,$F1,$99,$CC,$62,$3F,$2D,$DA
	.dc.b	$F3,$3F,$DC,$30,$8C,$3F,$C9,$63,$31,$88,$FF,$64,$4B,$E2,$BD,$8D
	.dc.b	$8F,$E6,$F1,$EB,$F8,$0F,$67,$C0,$F9,$B3,$34,$95,$FE,$21,$8F,$F4
	.dc.b	$1B,$F2,$BE,$07,$E2,$1F,$17,$E0,$7B,$FC,$4C,$9F,$E8,$AF,$C9,$F8
	.dc.b	$2F,$1F,$C5,$7C,$02,$4F,$F3,$B7,$C5,$F0,$1E,$8F,$8A,$93,$F6,$23
	.dc.b	$D7,$F1,$52,$3F,$A8,$BF,$57,$E0,$7D,$FF,$13,$F3,$09,$3F,$D1,$DF
	.dc.b	$77,$BF,$F1,$3E,$2A,$4F,$E1,$23,$F0,$5D,$DF,$43,$A0,$D2,$4F,$D8
	.dc.b	$F8,$A8,$FF,$53,$6B,$FE,$F9,$87,$C0,$7E,$09,$B1,$FE,$4B,$6B,$B8
	.dc.b	$D7,$BF,$F6,$3E,$2A,$4F,$FB,$26,$93,$F7,$3C,$2F,$C4,$92,$7F,$B2
	.dc.b	$3D,$FD,$FF,$C0,$F8,$72,$79,$47,$BB,$DE,$FD,$07,$EC,$5F,$C6,$3C
	.dc.b	$3F,$DA,$B7,$C3,$FF,$7A,$E6,$AD,$F8,$9F,$9C,$AF,$CF,$E3,$FE,$39
	.dc.b	$6B,$FE,$FC,$6B,$FE,$E5,$A7,$BD,$E0,$F1,$A4,$FB,$63,$E8,$FB,$9E
	.dc.b	$4F,$2E,$4F,$C1,$8F,$3F,$C5,$F8,$49,$1F,$E7,$5C,$51,$F3,$FD,$CF
	.dc.b	$6F,$8B,$2B,$E5,$47,$C0,$F1,$64,$7F,$BD,$6F,$F4,$2F,$CF,$F6,$FF
	.dc.b	$87,$79,$8F,$9E,$76,$FF,$4B,$AF,$F5,$7C,$61,$FF,$46,$B5,$FF,$46
	.dc.b	$35,$12,$5F,$3C,$F8,$3E,$91,$86,$C6,$67,$FD,$32,$D7,$FD,$FB,$DC
	.dc.b	$F9,$EF,$F0,$A4,$3A,$3E,$DF,$95,$F7,$8F,$E3,$D3,$A8,$FE,$67,$CC
	.dc.b	$F7,$B3,$D5,$30,$FC,$84,$70,$F6,$7F,$0F,$E6,$49,$F5,$E1,$FC,$F3
	.dc.b	$15,$6E,$73,$87,$C0,$CF,$CD,$39,$9F,$34,$E6,$9E,$CF,$9B,$D4,$95
	.dc.b	$C8,$61,$EF,$8F,$5F,$AB,$27,$FD,$6D,$7F,$DC,$DB,$1F,$CC,$A3,$11
	.dc.b	$F2,$17,$B9,$C6,$3C,$67,$68,$DE,$7E,$D9,$F9,$3E,$B7,$EA,$3D,$89
	.dc.b	$7E,$5C,$BE,$94,$FC,$EF,$73,$C0,$49,$F2,$BE,$3F,$47,$F6,$3E,$94
	.dc.b	$9E,$59,$A7,$EB,$FA,$72,$72,$FD,$AF,$4F,$F7,$3F,$1D,$27,$F3,$4F
	.dc.b	$F6,$55,$E8,$3E,$47,$1E,$DB,$D2,$FE,$9A,$DA,$FF,$0B,$C6,$0C,$FA
	.dc.b	$32,$3F,$C3,$6D,$7E,$57,$A7,$FC,$2B,$E3,$E4,$7A,$A3,$F1,$E4,$70
	.dc.b	$7B,$BC,$3F,$C9,$D7,$F1,$2F,$5E,$4F,$C2,$8F,$CB,$E1,$F6,$49,$1F
	.dc.b	$79,$FE,$25,$F1,$DF,$C6,$89,$1F,$2B,$D3,$F8,$5F,$2B,$FE,$C9,$8F
	.dc.b	$DC,$35,$FF,$22,$35,$FD,$A9,$FB,$67,$0F,$F9,$DF,$D5,$F8,$E4,$7F
	.dc.b	$C7,$AF,$F7,$EF,$81,$FE,$51,$1F,$96,$F2,$DE,$CF,$F2,$E3,$FC,$1D
	.dc.b	$3A,$F8,$17,$CE,$B3,$FE,$19,$18,$6C,$67,$0F,$F2,$D4,$E5,$FF,$2E
	.dc.b	$76,$FF,$A0,$7D,$6F,$E7,$7F,$21,$27,$F2,$98,$FF,$85,$97,$CA,$3E
	.dc.b	$77,$F0,$B1,$FE,$65,$8C,$74,$FF,$1D,$6C,$66,$76,$FF,$10,$D9,$F2
	.dc.b	$5F,$D0,$C9,$1F,$CC,$DE,$5F,$19,$FE,$1A,$B3,$49,$3F,$27,$D7,$F8
	.dc.b	$A4,$7F,$80,$4F,$0F,$90,$FF,$20,$36,$3D,$B3,$B7,$F7,$31,$97,$F7
	.dc.b	$7B,$31,$F4,$C7,$F5,$7C,$FF,$AD,$11,$F8,$58,$FF,$6E,$6F,$C0,$7A
	.dc.b	$88,$FD,$2F,$53,$F9,$E3,$FC,$F3,$87,$F9,$D7,$5F,$DA,$9C,$CA,$72
	.dc.b	$FF,$49,$F9,$D7,$C6,$7F,$8F,$9B,$19,$8C,$1F,$D1,$3E,$75,$EF,$7F
	.dc.b	$54,$6C,$FF,$25,$F6,$51,$88,$FE,$37,$87,$FB,$17,$D0,$FF,$34,$36
	.dc.b	$38,$4E,$FD,$0F,$F6,$E5,$F2,$C3,$7C,$FF,$C8,$BD,$3F,$D2,$FC,$E7
	.dc.b	$F9,$5F,$1F,$D3,$87,$0E,$BD,$0F,$F2,$CC,$62,$3F,$B0,$BF,$E5,$37
	.dc.b	$FD,$68,$D7,$9B,$5E,$FF,$11,$FE,$0F,$F9,$5F,$F0,$B9,$23,$F3,$BE
	.dc.b	$C7,$EB,$3D,$4F,$D9,$12,$3F,$4B,$92,$3F,$6D,$E9,$C9,$CB,$5F,$CE
	.dc.b	$7C,$AF,$F8,$90,$8F,$EA,$BC,$DF,$DC,$F9,$9F,$BF,$24,$76,$7F,$81
	.dc.b	$78,$FF,$BE,$F1,$E4,$FE,$CA,$3F,$80,$F8,$BF,$C0,$99,$1F,$B6,$8F
	.dc.b	$F4,$97,$EE,$5F,$33,$97,$FD,$4E,$DE,$3F,$ED,$4D,$AF,$EC,$7B,$7E
	.dc.b	$D2,$B6,$FF,$A4,$27,$FD,$C5,$3C,$23,$07,$FA,$53,$E7,$79,$78,$4A
	.dc.b	$FE,$49,$7E,$0F,$17,$FD,$49,$5A,$9F,$F2,$BE,$1F,$C9,$AF,$87,$C5
	.dc.b	$77,$24,$7F,$14,$97,$C9,$78,$C9,$19,$7F,$D3,$BF,$22,$E5,$FF,$0C
	.dc.b	$4B,$AF,$FD,$57,$39,$D3,$FC,$51,$F1,$5F,$CD,$A3,$15,$B8,$C5,$F2
	.dc.b	$BF,$F2,$9F,$C0,$F0,$AF,$CB,$B5,$FF,$76,$1C,$4C,$97,$4F,$89,$E2
	.dc.b	$FF,$8B,$23,$E8,$B5,$FF,$09,$E9,$FE,$74,$E1,$F8,$13,$9A,$F9,$9F
	.dc.b	$2F,$68,$D2,$57,$F3,$AE,$0C,$BF,$E0,$B9,$7C,$6A,$3F,$C1,$D3,$AC
	.dc.b	$7F,$8E,$D6,$A7,$87,$C9,$77,$23,$F4,$DF,$17,$8F,$CE,$B8,$B7,$87
	.dc.b	$FD,$EB,$87,$F9,$4E,$FC,$53,$97,$FC,$75,$0F,$97,$C5,$FF,$8D,$21
	.dc.b	$F2,$F8,$D5,$1B,$BF,$4C,$E6,$37,$1E,$CF,$2F,$13,$97,$5F,$2F,$F2
	.dc.b	$E8,$D2,$57,$E3,$FE,$47,$C1,$B7,$FA,$5E,$3E,$3F,$8F,$89,$F0,$7C
	.dc.b	$BF,$32,$4F,$D1,$7B,$79,$92,$3E,$A3,$75,$FF,$37,$E1,$EB,$7E,$8E
	.dc.b	$DE,$BF,$8D,$F0,$BF,$12,$1F,$3A,$9F,$C4,$46,$1B,$19,$C3,$FD,$18
	.dc.b	$D9,$FE,$5A,$72,$C3,$64,$9D,$BF,$93,$F3,$7F,$A7,$7C,$8F,$D4,$EA
	.dc.b	$7C,$EB,$B7,$EB,$D6,$B0,$FE,$91,$39,$6C,$F1,$8C,$63,$FD,$2A,$BF
	.dc.b	$D5,$B5,$A3,$FC,$E9,$0F,$9D,$79,$FE,$E6,$CF,$FA,$9D,$B3,$EC,$6B
	.dc.b	$FE,$30,$AF,$E0,$58,$FA,$FB,$FF,$81,$5F,$F8,$81,$B5,$FD,$42,$B7
	.dc.b	$3F,$DE,$3B,$3F,$69,$EB,$7F,$83,$9B,$D1,$FA,$5B,$FE,$AF,$39,$AF
	.dc.b	$E2,$DF,$2B,$FB,$52,$38,$3F,$C0,$7D,$9F,$9B,$F9,$CF,$F3,$54,$7F
	.dc.b	$85,$30,$FD,$36,$1F,$E0,$E9,$CB,$7E,$17,$81,$FA,$6C,$7F,$C4,$38
	.dc.b	$63,$0E,$58,$F9,$88,$FF,$84,$9F,$E1,$1D,$B2,$3F,$96,$37,$0F,$5E
	.dc.b	$5F,$E2,$C9,$FF,$0C,$BF,$00,$8F,$F9,$1B,$27,$AB,$FA,$69,$F6,$36
	.dc.b	$7F,$D5,$DA,$FF,$AB,$BB,$3F,$19,$DF,$FE,$10,$9F,$F2,$8B,$71,$FA
	.dc.b	$4D,$7F,$A2,$63,$FE,$AC,$9F,$E3,$55,$F2,$7E,$33,$F6,$AF,$E9,$7C
	.dc.b	$8F,$DB,$5F,$E7,$9A,$FF,$83,$1A,$FF,$9D,$1A,$FF,$BA,$BE,$77,$FA
	.dc.b	$45,$FF,$BC,$DB,$1F,$A8,$6B,$FE,$16,$6B,$FD,$AB,$C1,$FE,$0E,$7D
	.dc.b	$DF,$9C,$B5,$FE,$BB,$CB,$F2,$B5,$FE,$D9,$6B,$FE,$D2,$6E,$66,$8F
	.dc.b	$5B,$F8,$9F,$CB,$FE,$25,$B7,$D7,$E3,$7F,$18,$7E,$37,$E3,$6C,$64
	.dc.b	$FE,$85,$E2,$E3,$2F,$E7,$EF,$C9,$F1,$3F,$85,$EF,$FC,$4E,$7E,$44
	.dc.b	$60,$CF,$34,$BF,$E1,$72,$A4,$E6,$3F,$8E,$F8,$9F,$CA,$8F,$E3,$5F
	.dc.b	$39,$FA,$9F,$27,$FC,$27,$1F,$A3,$F8,$B9,$8F,$C7,$F8,$EB,$87,$E8
	.dc.b	$FE,$03,$F9,$9F,$91,$FA,$96,$D7,$CD,$F8,$CF,$EC,$DC,$86,$92,$3F
	.dc.b	$31,$F1,$3F,$CF,$3E,$53,$F6,$4D,$8F,$F7,$87,$CE,$FF,$04,$DB,$FA
	.dc.b	$E7,$42,$3F,$8C,$C6,$1D,$4B,$45,$6F,$CB,$FE,$B1,$5F,$C8,$27,$FC
	.dc.b	$3C,$D7,$84,$EE,$BF,$D6,$33,$99,$E6,$BE,$2B,$99,$D2,$ED,$7A,$3C
	.dc.b	$DE,$E7,$83,$E6,$AC,$64,$F3,$65,$EC,$F2,$3F,$C0,$51,$9C,$BF,$D3
	.dc.b	$B1,$87,$FD,$3F,$93,$E2,$7D,$7F,$2F,$EC,$6D,$FC,$88,$FE,$ED,$7F
	.dc.b	$C6,$30,$F8,$D1,$88,$FE,$95,$F1,$FF,$E9,$6A,$FE,$A7,$E7,$23,$19
	.dc.b	$7F,$A0,$A3,$1C,$BF,$E2,$56,$D7,$0E,$27,$16,$5B,$19,$8C,$1F,$E2
	.dc.b	$9C,$5F,$2F,$E6,$3E,$AF,$C3,$5A,$FC,$0F,$C3,$7A,$BE,$5F,$EC,$B6
	.dc.b	$3E,$49,$F6,$2F,$F9,$0B,$FE,$E8,$6E,$9F,$E9,$AE,$64,$62,$3F,$CB
	.dc.b	$CD,$7F,$03,$C5,$FE,$F7,$AE,$B1,$F8,$A9,$7C,$BF,$E2,$BE,$17,$E6
	.dc.b	$BF,$8C,$F8,$5F,$9F,$E6,$DC,$38,$E5,$E0,$8F,$A7,$E5,$7C,$12,$4F
	.dc.b	$9D,$FF,$26,$66,$F8,$9F,$B3,$F2,$FF,$98,$6D,$7E,$93,$E2,$BE,$EF
	.dc.b	$B5,$F9,$C6,$C7,$83,$6F,$E2,$CF,$C3,$F3,$7E,$81,$1F,$F0,$8E,$9F
	.dc.b	$E0,$9F,$27,$F0,$FE,$67,$F8,$F7,$C1,$FE,$15,$5C,$E3,$1F,$05,$FE
	.dc.b	$61,$9F,$F3,$27,$27,$CD,$6B,$FE,$89,$6B,$F5,$7C,$EF,$C0,$12,$FE
	.dc.b	$B3,$CF,$E0,$64,$79,$29,$18,$AE,$DE,$C7,$E7,$3C,$AF,$33,$6F,$E0
	.dc.b	$7C,$DF,$E3,$48,$FE,$0D,$FC,$87,$99,$FA,$43,$DD,$E7,$D9,$FA,$F9
	.dc.b	$7F,$6F,$C7,$F6,$3E,$6F,$C3,$F6,$3F,$56,$FE,$B3,$E5,$7F,$34,$4B
	.dc.b	$FC,$83,$F9,$DF,$3B,$F6,$27,$CB,$F5,$2C,$F6,$F4,$7E,$77,$8F,$F3
	.dc.b	$DB,$1F,$88,$F9,$DF,$F0,$7E,$7F,$EB,$36,$BF,$C5,$9A,$FF,$1E,$6F
	.dc.b	$3F,$E7,$BC,$8F,$E0,$27,$F9,$DE,$3F,$85,$79,$BF,$C3,$E3,$FA,$0F
	.dc.b	$C4,$7F,$96,$3D,$4F,$F5,$DE,$3F,$D2,$9B,$8F,$F9,$59,$AF,$EA,$3C
	.dc.b	$EF,$DD,$12,$7F,$60,$FF,$A9,$71,$79,$5E,$1F,$23,$C4,$DA,$F7,$45
	.dc.b	$B6,$7F,$CF,$AF,$8D,$68,$F8,$9E,$3F,$AF,$C5,$F6,$36,$3E,$44,$5B
	.dc.b	$E7,$1F,$1E,$71,$7F,$EA,$59,$7C,$BF,$9B,$E2,$FF,$A4,$7C,$4B,$D4
	.dc.b	$FE,$95,$5A,$D5,$5A,$9C,$DF,$F1,$AC,$B1,$7F,$D4,$63,$A2,$C3,$A3
	.dc.b	$9C,$39,$B9,$BE,$71,$89,$DC,$F9,$BD,$0F,$F5,$E4,$62,$FA,$B6,$B8
	.dc.b	$5F,$2C,$7C,$78,$71,$C6,$E1,$C2,$B5,$3E,$13,$C9,$5A,$C7,$BC,$EB
	.dc.b	$1E,$D3,$CB,$1E,$F7,$EF,$BD,$BC,$8E,$87,$C9,$9F,$F0,$AE,$1F,$12
	.dc.b	$3E,$69,$93,$FD,$46,$E5,$FA,$3F,$14,$F4,$5E,$F7,$F8,$7A,$30,$D8
	.dc.b	$CC,$FF,$AB,$27,$2F,$4F,$7D,$EA,$78,$9B,$19,$8C,$11,$F4,$E1,$FE
	.dc.b	$6C,$8F,$93,$3C,$34,$FA,$5C,$7F,$86,$7E,$A4,$67,$6F,$C2,$B7,$0F
	.dc.b	$C2,$35,$FC,$8D,$71,$9D,$DF,$62,$4F,$8A,$74,$F2,$8C,$DF,$81,$FB
	.dc.b	$B3,$A9,$FC,$ED,$F7,$C3,$D5,$1E,$C8,$C4,$EE,$73,$7C,$EB,$57,$D1
	.dc.b	$EF,$87,$99,$C9,$E5,$32,$79,$4C,$CF,$C6,$72,$79,$26,$D7,$F6,$C8
	.dc.b	$EC,$73,$24,$FE,$7F,$87,$CD,$84,$9D,$CE,$5F,$CE,$CF,$EE,$38,$3F
	.dc.b	$1B,$68,$C4,$FE,$E1,$AF,$ED,$DA,$FF,$04,$7E,$21,$1F,$D5,$FC,$97
	.dc.b	$FA,$05,$EE,$43,$F0,$F1,$F4,$63,$51,$FA,$DE,$97,$EC,$DF,$90,$8C
	.dc.b	$3F,$EF,$97,$F8,$4C,$6B,$C2,$FB,$EF,$0D,$2D,$71,$CF,$F8,$5C,$E6
	.dc.b	$F8,$CE,$75,$FF,$2D,$BF,$E5,$78,$CD,$F2,$23,$C3,$DB,$EB,$5D,$BD
	.dc.b	$B9,$3D,$23,$2D,$21,$EA,$F6,$F2,$FB,$22,$E5,$E6,$C7,$FC,$B1,$3E
	.dc.b	$6F,$9D,$F5,$91,$F4,$FB,$9E,$E2,$38,$3D,$BF,$0D,$F0,$88,$E0,$F8
	.dc.b	$3C,$DF,$1B,$DC,$91,$F1,$1F,$93,$F3,$5F,$2B,$E4,$A4,$FC,$87,$E6
	.dc.b	$FC,$97,$CE,$F7,$E4,$FC,$C7,$E8,$FB,$FF,$4B,$E4,$A4,$FD,$07,$EA
	.dc.b	$FC,$BF,$D6,$E8,$93,$F5,$1F,$B3,$EF,$FD,$AF,$7A,$4F,$D8,$7E,$FF
	.dc.b	$ED,$9F,$C9,$39,$3E,$EB,$FD,$1A,$7F,$14,$D8,$CF,$87,$C3,$E5,$38
	.dc.b	$FB,$D2,$FF,$17,$7F,$EC,$36,$F9,$BD,$BD,$D9,$1A,$7C,$3F,$29,$FC
	.dc.b	$0F,$DE,$FF,$65,$CF,$F6,$66,$F8,$86,$98,$F7,$64,$FF,$3E,$6B,$FE
	.dc.b	$F7,$E1,$FE,$D7,$E6,$FF,$71,$B5,$63,$C7,$FE,$D6,$DB,$FF,$5D,$7C
	.dc.b	$53,$0C,$78,$CD,$9F,$1D,$87,$FC,$8F,$7F,$E7,$06,$BF,$E0,$23,$FD
	.dc.b	$86,$F0,$75,$1C,$48,$F1,$F7,$FA,$7C,$9C,$8E,$8F,$7F,$8A,$91,$DD
	.dc.b	$AF,$7C,$3C,$DE,$FF,$AF,$CB,$53,$E0,$70,$DB,$FF,$47,$CE,$A7,$3F
	.dc.b	$03,$27,$FD,$6B,$48,$FF,$7F,$D7,$46,$C6,$67,$7F,$01,$C5,$F8,$55
	.dc.b	$B7,$BC,$A7,$E8,$43,$E8,$7F,$C2,$3C,$9F,$E2,$76,$F1,$FF,$89,$1B
	.dc.b	$3F,$EE,$F7,$F7,$AF,$CF,$39,$79,$BD,$23,$35,$CE,$31,$E5,$7D,$9F
	.dc.b	$8F,$97,$F6,$5F,$B9,$F0,$7F,$77,$E3,$EC,$FD,$D8,$C3,$FE,$B6,$3F
	.dc.b	$C7,$7B,$7E,$58,$8F,$F8,$41,$FC,$CE,$92,$3F,$C2,$6D,$7F,$C2,$67
	.dc.b	$C1,$F2,$BE,$1F,$57,$C5,$8E,$CD,$7D,$DA,$A7,$F2,$5F,$1F,$27,$F2
	.dc.b	$4F,$E5,$7E,$0E,$47,$E5,$5A,$FE,$1F,$A5,$F1,$FF,$4B,$F1,$87,$4D
	.dc.b	$25,$7B,$E3,$F1,$6D,$E5,$E0,$47,$F4,$5F,$09,$F7,$3E,$77,$F7,$70
	.dc.b	$69,$1F,$DF,$BF,$C3,$DF,$F1,$AB,$5F,$F3,$1B,$51,$CF,$C7,$F9,$2F
	.dc.b	$EB,$BE,$43,$C1,$12,$3F,$69,$1F,$B7,$F8,$79,$1F,$B7,$7F,$7B,$FB
	.dc.b	$5F,$F9,$D7,$A9,$D7,$DB,$BD,$39,$8C,$36,$D3,$A9,$E7,$F1,$DD,$3E
	.dc.b	$0E,$57,$FB,$D2,$D1,$E0,$F8,$49,$39,$6B,$FB,$B9,$7C,$17,$9B,$E0
	.dc.b	$BD,$F9,$79,$1B,$3F,$E1,$BA,$FF,$2E,$CE,$5F,$F7,$7C,$7F,$84,$DF
	.dc.b	$14,$F3,$7C,$9F,$05,$F0,$BE,$47,$E1,$FC,$34,$9F,$44,$65,$A4,$3C
	.dc.b	$DF,$21,$F2,$BE,$43,$C6,$48,$F0,$BE,$6F,$90,$D4,$7C,$38,$FA,$31
	.dc.b	$F4,$BD,$0F,$59,$27,$D2,$FB,$BD,$0D,$47,$CB,$8F,$B3,$1F,$6B,$CB
	.dc.b	$FB,$7E,$2C,$AF,$A1,$19,$69,$0F,$9D,$E2,$FE,$0B,$E3,$FE,$59,$23
	.dc.b	$E4,$BF,$3B,$E3,$F5,$1F,$6E,$3F,$13,$1F,$C6,$BE,$67,$FC,$63,$3F
	.dc.b	$E8,$E6,$BF,$EB,$C6,$BF,$9A,$6A,$59,$FF,$62,$6C,$7E,$79,$AF,$F0
	.dc.b	$86,$BF,$E9,$CA,$7B,$5E,$3A,$93,$FC,$5E,$72,$FF,$19,$E6,$FE,$BC
	.dc.b	$FE,$A9,$AF,$F1,$46,$BF,$EA,$86,$BF,$ED,$65,$EE,$7F,$78,$6C,$73
	.dc.b	$8C,$1F,$DF,$3C,$9C,$C7,$F1,$A8,$FE,$6C,$7E,$CB,$E4,$24,$7E,$9D
	.dc.b	$AF,$E8,$FE,$3F,$FA,$E7,$23,$C3,$A4,$9F,$E6,$9F,$AD,$BE,$1F,$FA
	.dc.b	$6A,$F8,$6F,$F0,$6B,$67,$F8,$4B,$5F,$E6,$6D,$7F,$DE,$7D,$AF,$7F
	.dc.b	$FC,$D5,$E1,$FF,$06,$9F,$E0,$31,$8F,$0F,$F8,$43,$63,$3F,$1B,$FD
	.dc.b	$7B,$F7,$CF,$EC,$D1,$FD,$01,$AF,$FB,$51,$AF,$F1,$E0,$E1,$FE,$42
	.dc.b	$B6,$FF,$13,$B6,$3F,$91,$B5,$ED,$1F,$E2,$CF,$8F,$F8,$98,$FF,$2A
	.dc.b	$F6,$7F,$D0,$54,$9C,$BF,$ED,$8C,$3F,$DC,$7B,$2F,$84,$61,$B3,$FE
	.dc.b	$49,$38,$16,$FA,$BF,$CE,$3E,$0F,$6F,$FB,$64,$D8,$F9,$67,$BB,$F3
	.dc.b	$63,$FD,$EF,$18,$DB,$F9,$FB,$63,$FB,$4B,$5D,$FB,$1C,$5F,$95,$08
	.dc.b	$C4,$7F,$4A,$8C,$4F,$49,$CF,$07,$DF,$F9,$0F,$BD,$9F,$DF,$8C,$57
	.dc.b	$F5,$28,$C7,$0B,$E3,$F7,$3F,$82,$BB,$97,$F5,$8F,$D6,$DF,$11,$E2
	.dc.b	$84,$62,$BF,$AD,$B6,$3F,$C0,$AD,$7F,$09,$BB,$F5,$AF,$E8,$B5,$FE
	.dc.b	$3A,$6C,$FF,$88,$FF,$5B,$FE,$CF,$39,$F3,$FC,$7B,$7E,$6F,$E4,$EE
	.dc.b	$C6,$FC,$1E,$1F,$4E,$2F,$CF,$7F,$B2,$E3,$FE,$21,$6C,$7F,$94,$6F
	.dc.b	$94,$E4,$E3,$FA,$5F,$F8,$5D,$FE,$EE,$D9,$FF,$10,$B5,$FD,$00,$70
	.dc.b	$FE,$8D,$7C,$97,$C9,$6C,$F4,$8C,$1F,$36,$FF,$95,$C7,$F8,$DF,$E7
	.dc.b	$BF,$C3,$EF,$A7,$C7,$F6,$C6,$30,$FF,$25,$35,$F1,$B5,$FD,$A3,$5F
	.dc.b	$F0,$9B,$53,$FC,$56,$7F,$C2,$AD,$8F,$F1,$FB,$5E,$BE,$0F,$E4,$32
	.dc.b	$7C,$AF,$D4,$F8,$8F,$AA,$DC,$3F,$20,$7F,$0E,$8E,$04,$7F,$C9,$7B
	.dc.b	$FF,$17,$8E,$9D,$DF,$78,$FB,$FF,$2B,$6B,$9E,$5F,$C6,$A7,$F0,$7E
	.dc.b	$26,$63,$EE,$3F,$86,$F8,$6F,$C3,$F0,$5B,$E3,$97,$F2,$58,$CC,$78
	.dc.b	$B4,$FA,$B8,$7F,$85,$1A,$FF,$2C,$6B,$FC,$A3,$E5,$7F,$27,$F2,$BF
	.dc.b	$94,$6D,$76,$A4,$FC,$72,$47,$77,$E5,$7C,$AF,$E7,$0F,$CE,$FC,$AD
	.dc.b	$9E,$51,$FA,$0F,$90,$C6,$3F,$99,$7F,$47,$E6,$FE,$93,$BE,$DF,$85
	.dc.b	$C3,$ED,$7E,$C1,$1F,$81,$E8,$FE,$B0,$FD,$7F,$EC,$7F,$B2,$C3,$F4
	.dc.b	$D1,$8F,$84,$95,$ED,$8B,$3F,$EB,$BA,$FE,$A1,$3F,$39,$B1,$9F,$96
	.dc.b	$EB,$B7,$6F,$96,$97,$B8,$B5,$CF,$E5,$BF,$77,$F2,$AE,$8F,$2C,$61
	.dc.b	$97,$FE,$6C,$9D,$36,$3F,$8C,$46,$2B,$FD,$32,$FF,$6D,$9E,$1C,$FA
	.dc.b	$23,$FE,$76,$F9,$AF,$F4,$6B,$FE,$A9,$6D,$7B,$FE,$5F,$F8,$7C,$63
	.dc.b	$CB,$FE,$84,$FF,$B6,$AB,$87,$C6,$6B,$E3,$B6,$FC,$EF,$97,$93,$F1
	.dc.b	$FE,$6E,$E5,$F2,$FD,$2F,$A6,$63,$2F,$E7,$9C,$7F,$D2,$DE,$FF,$29
	.dc.b	$FC,$83,$91,$F2,$FE,$48,$C3,$32,$FF,$48,$CE,$9B,$1F,$CF,$63,$18
	.dc.b	$FF,$AF,$9F,$E9,$ED,$F4,$F4,$4A,$FD,$CE,$3F,$05,$FF,$49,$3F,$06
	.dc.b	$7F,$09,$B7,$E3,$36,$FC,$6F,$CC,$4B,$FC,$34,$5A,$FF,$72,$70,$7F
	.dc.b	$1D,$D7,$E5,$FC,$A7,$CE,$F9,$6F,$9E,$FD,$0F,$26,$63,$F5,$62,$E1
	.dc.b	$FE,$FA,$FC,$EB,$F0,$18,$7F,$88,$7D,$44,$7F,$8B,$EF,$F8,$5C,$63
	.dc.b	$A7,$FB,$0B,$AF,$26,$D6,$F0,$FE,$7F,$39,$BF,$F7,$1B,$63,$33,$B7
	.dc.b	$FB,$0B,$F8,$F7,$C9,$D3,$D2,$FA,$BF,$B0,$E4,$98,$7E,$A7,$8E,$DF
	.dc.b	$EB,$7E,$67,$F6,$A6,$34,$FC,$1B,$FB,$8E,$69,$7F,$B1,$E0,$B7,$FB
	.dc.b	$59,$FD,$EE,$F8,$C3,$F1,$6F,$EB,$BC,$6F,$C5,$4D,$C3,$F9,$7D,$B9
	.dc.b	$D3,$A9,$FC,$5B,$F8,$92,$3F,$8E,$C7,$F1,$84,$70,$7E,$23,$8B,$FC
	.dc.b	$A9,$ED,$7F,$84,$F0,$FF,$6A,$F4,$7F,$51,$8C,$79,$48,$7C,$AF,$0F
	.dc.b	$6B,$F1,$F8,$5C,$38,$FC,$AD,$C3,$83,$7D,$AF,$F0,$0E,$5D,$32,$4E
	.dc.b	$7E,$56,$69,$FE,$41,$CF,$FD,$D9,$87,$FB,$8A,$31,$97,$FB,$AA,$15
	.dc.b	$D7,$E1,$A5,$6B,$3F,$F1,$3F,$B5,$8C,$1A,$7B,$31,$7C,$9F,$94,$F3
	.dc.b	$7C,$9D,$C7,$94,$BE,$1B,$D4,$7A,$FE,$4E,$CE,$BE,$4F,$DC,$7B,$FE
	.dc.b	$4E,$C7,$2F,$73,$E1,$1F,$0F,$E4,$EC,$76,$F9,$3F,$8C,$7C,$7F,$92
	.dc.b	$B1,$E2,$F9,$3F,$94,$7C,$BF,$93,$B1,$E6,$F9,$1F,$9C,$63,$8B,$DF
	.dc.b	$CF,$F2,$A3,$E9,$7C,$9C,$AE,$07,$B9,$FA,$DF,$23,$23,$E0,$FC,$7D
	.dc.b	$C3,$99,$F4,$67,$1C,$DF,$2E,$6E,$3F,$62,$7F,$05,$F1,$4F,$33,$38
	.dc.b	$3A,$93,$9C,$BE,$C2,$F8,$EE,$27,$2F,$8D,$B3,$A8,$B7,$E4,$BE,$75
	.dc.b	$8D,$3F,$33,$87,$3C,$3A,$E7,$91,$87,$8E,$7F,$CE,$93,$F8,$E8,$7C
	.dc.b	$CF,$E3,$A3,$0D,$8F,$86,$7F,$B7,$23,$F4,$4F,$0E,$FB,$3B,$C3,$F5
	.dc.b	$D3,$9A,$ED,$F0,$1F,$A4,$C7,$FD,$75,$7B,$6E,$1F,$EC,$9A,$FF,$2F
	.dc.b	$CF,$85,$B1,$9B,$DC,$7F,$82,$5F,$6C,$FE,$A5,$F3,$B7,$0F,$1B,$DE
	.dc.b	$37,$F2,$DE,$82,$57,$C0,$BB,$3C,$BC,$17,$C0,$7C,$46,$C7,$6B,$F8
	.dc.b	$31,$8C,$7B,$B7,$2F,$7B,$6B,$C3,$8F,$3C,$24,$77,$7E,$8E,$1D,$1C
	.dc.b	$E1,$D1,$CD,$63,$0E,$B1,$6F,$A9,$CC,$95,$C5,$F4,$56,$DF,$C0,$E1
	.dc.b	$E5,$6D,$76,$78,$DF,$66,$9D,$AF,$A1,$23,$4F,$D0,$C3,$D5,$3E,$33
	.dc.b	$EE,$DF,$CA,$6D,$FF,$A7,$9F,$B7,$19,$7F,$E0,$07,$C8,$FF,$8D,$7E
	.dc.b	$6F,$51,$FE,$F5,$F5,$51,$89,$D4,$F3,$6B,$F8,$CF,$17,$F6,$CF,$7D
	.dc.b	$91,$88,$D7,$27,$69,$CB,$67,$8E,$9F,$8C,$6E,$BF,$9C,$6D,$FE,$53
	.dc.b	$C5,$FD,$8B,$F9,$48,$CB,$FB,$06,$E1,$DA,$3F,$33,$3C,$2F,$B6,$1F
	.dc.b	$B4,$BD,$9C,$27,$58,$78,$67,$44,$BE,$65,$BE,$6F,$8E,$3F,$7F,$5C
	.dc.b	$A3,$F5,$77,$DA,$78,$5F,$EE,$2B,$55,$E9,$9C,$BC,$1E,$6E,$0D,$1D
	.dc.b	$CF,$19,$6F,$C6,$4A,$ED,$1C,$86,$FF,$C1,$2F,$E5,$B6,$3F,$81,$1F
	.dc.b	$C6,$A3,$2F,$23,$BD,$7F,$02,$8F,$4E,$2F,$16,$98,$89,$5F,$C8,$A3
	.dc.b	$2F,$69,$E0,$D8,$EE,$FF,$12,$6B,$FC,$B1,$E2,$FF,$14,$7E,$4F,$13
	.dc.b	$C1,$FA,$1C,$CF,$9B,$FC,$96,$5F,$E1,$47,$FA,$36,$1F,$DF,$DA,$FF
	.dc.b	$43,$77,$A7,$F3,$EA,$E5,$AF,$91,$B5,$FC,$9D,$AF,$CC,$38,$57,$E8
	.dc.b	$75,$FC,$BC,$EB,$0F,$15,$F9,$AB,$BD,$FF,$5D,$5C,$F9,$C7,$5F,$21
	.dc.b	$2B,$D1,$19,$7C,$55,$FD,$1B,$2E,$ED,$8F,$1C,$FA,$63,$FA,$5E,$3F
	.dc.b	$0E,$39,$4F,$C1,$F9,$1F,$27,$C7,$48,$F8,$CF,$9F,$E3,$A4,$78,$1E
	.dc.b	$84,$7C,$7E,$CF,$B4,$8F,$CF,$E7,$F8,$1E,$CC,$8F,$A4,$FC,$3F,$8E
	.dc.b	$F8,$9F,$1D,$E7,$24,$FB,$9F,$57,$AD,$F2,$49,$1F,$6F,$E3,$3E,$67
	.dc.b	$C6,$C8,$FB,$CF,$CF,$F8,$DF,$A1,$D1,$F0,$C9,$3F,$29,$F8,$DF,$19
	.dc.b	$F5,$49,$1F,$E4,$5F,$9B,$FE,$22,$FF,$82,$5F,$E6,$1C,$7F,$E6,$06
	.dc.b	$F1,$FE,$25,$B3,$FE,$19,$5F,$19,$F9,$67,$FC,$ED,$1F,$8B,$6B,$F8
	.dc.b	$B3,$87,$B9,$F9,$18,$CF,$3F,$E3,$17,$C6,$73,$25,$79,$FE,$2F,$F6
	.dc.b	$BE,$A4,$8F,$F0,$9B,$FB,$8F,$52,$47,$F3,$57,$FA,$1B,$FE,$1A,$6F
	.dc.b	$84,$6F,$A9,$FC,$CB,$E0,$3C,$D3,$9A,$FF,$48,$B5,$46,$A7,$86,$7F
	.dc.b	$17,$E2,$AD,$7F,$8A,$FC,$0B,$E1,$FF,$8F,$9B,$B2,$A9,$2B,$FC,$67
	.dc.b	$8F,$EC,$6D,$5B,$9C,$E1,$F4,$FE,$2F,$AF,$9E,$F0,$A9,$2B,$67,$D7
	.dc.b	$F8,$AF,$09,$97,$D7,$7F,$C1,$6B,$94,$70,$C7,$C3,$5E,$CF,$1B,$FD
	.dc.b	$B3,$DB,$E0,$F1,$25,$FF,$B6,$BE,$1D,$3B,$F8,$97,$B5,$C6,$73,$5B
	.dc.b	$AD,$4E,$FD,$AB,$3F,$EE,$4B,$E2,$D8,$F9,$AB,$E5,$DC,$98,$7F,$87
	.dc.b	$C7,$FB,$21,$AF,$B8,$FF,$2F,$BC,$1D,$76,$F1,$99,$18,$3E,$16,$EF
	.dc.b	$27,$E7,$F9,$32,$73,$1F,$73,$E2,$64,$7C,$9E,$64,$7E,$03,$E2,$7D
	.dc.b	$3F,$1A,$F8,$FA,$4B,$FC,$7F,$07,$BB,$B5,$C1,$F9,$82,$4F,$F8,$45
	.dc.b	$FC,$DF,$BF,$B6,$47,$E6,$23,$F0,$BE,$74,$8F,$5B,$F9,$2F,$89,$91
	.dc.b	$82,$1F,$1F,$AB,$F2,$31,$F4,$3E,$6D,$F1,$F7,$C1,$97,$FE,$F6,$9C
	.dc.b	$B7,$B7,$85,$48,$C1,$F3,$3D,$1F,$CC,$91,$FF,$7F,$F0,$3C,$BF,$BF
	.dc.b	$F3,$7F,$BA,$93,$48,$FE,$F9,$FF,$24,$78,$9F,$C3,$DF,$E3,$AD,$8E
	.dc.b	$7E,$E7,$EA,$3D,$3F,$D4,$FA,$72,$7E,$FC,$7E,$DB,$E2,$35,$18,$3E
	.dc.b	$E3,$FB,$0F,$92,$FD,$8F,$87,$27,$F0,$F1,$FC,$03,$E6,$DF,$01,$B9
	.dc.b	$C9,$FA,$D8,$CB,$F8,$EF,$5F,$FC,$B3,$CD,$FE,$C4,$6F,$84,$69,$AA
	.dc.b	$FF,$7C,$D7,$12,$1F,$65,$CB,$7E,$23,$F8,$3F,$B1,$23,$C5,$1E,$72
	.dc.b	$38,$3C,$DF,$13,$FC,$3B,$D9,$F1,$92,$3F,$CE,$99,$7F,$B6,$A3,$C9
	.dc.b	$EE,$6E,$7F,$53,$1F,$0E,$3F,$8B,$FB,$5E,$C2,$4F,$2D,$11,$ED,$F5
	.dc.b	$37,$3F,$B2,$8F,$99,$1F,$37,$C1,$FE,$43,$E2,$4A,$F9,$91,$96,$90
	.dc.b	$F8,$9E,$27,$F2,$8E,$AF,$9A,$48,$F1,$3F,$37,$E2,$35,$1F,$C0,$A3
	.dc.b	$EE,$47,$F1,$2F,$B4,$6B,$49,$5F,$CC,$07,$CE,$3E,$2F,$F8,$CB,$FE
	.dc.b	$3D,$8F,$F4,$BB,$5F,$F4,$F3,$5F,$EE,$AD,$7F,$D0,$EB,$D9,$E7,$C1
	.dc.b	$27,$C1,$39,$7C,$5F,$8A,$75,$EF,$FE,$E4,$DF,$5F,$FD,$22,$D9,$0F
	.dc.b	$87,$74,$7F,$30,$7D,$F1,$FE,$AE,$6B,$FE,$B5,$6B,$FE,$0C,$6B,$FE
	.dc.b	$9E,$7E,$EB,$F1,$5A,$FF,$35,$3C,$1C,$9F,$CE,$C8,$FF,$3B,$6B,$F9
	.dc.b	$DF,$8A,$FE,$DB,$F0,$7F,$BE,$C7,$FC,$BC,$D9,$74,$7F,$42,$7F,$A1
	.dc.b	$7B,$4F,$C4,$BF,$99,$47,$D7,$F9,$E7,$9F,$FE,$08,$6C,$7F,$B3,$1A
	.dc.b	$FF,$B4,$DA,$FF,$AE,$AB,$FA,$34,$FF,$67,$F5,$77,$C1,$FE,$6B,$9C
	.dc.b	$D7,$F6,$DF,$83,$FF,$66,$4F,$F8,$97,$E7,$BF,$AF,$CF,$FB,$7D,$AF
	.dc.b	$F8,$D9,$AF,$FB,$21,$AF,$D8,$5B,$7F,$88,$1B,$1F,$EE,$46,$BD,$63
	.dc.b	$FC,$4B,$F1,$4F,$67,$FC,$3B,$2C,$7F,$D1,$73,$9C,$3F,$0B,$18,$AD
	.dc.b	$B6,$33,$F1,$BF,$EF,$7F,$87,$B8,$7F,$93,$9A,$FF,$B6,$4E,$D1,$B6
	.dc.b	$9F,$92,$F9,$E7,$A9,$FE,$49,$8F,$F7,$3C,$63,$93,$FD,$FE,$D8,$FF
	.dc.b	$2D,$87,$0F,$F9,$79,$7C,$BF,$38,$EC,$D7,$83,$63,$73,$D3,$27,$EA
	.dc.b	$CF,$3F,$FD,$4D,$A7,$DD,$8C,$57,$DD,$E7,$CB,$F2,$23,$39,$7E,$7F
	.dc.b	$E4,$FF,$AF,$4F,$A2,$11,$87,$D6,$D7,$FC,$EE,$D7,$F3,$FE,$9F,$67
	.dc.b	$F3,$F6,$8C,$64,$FC,$39,$F0,$9B,$EB,$FF,$5A,$D7,$1C,$1F,$3D,$FE
	.dc.b	$EC,$CF,$F8,$D3,$63,$FC,$09,$D1,$FE,$30,$9C,$9C,$7E,$7D,$EB,$53
	.dc.b	$A5,$F1,$E0,$FF,$41,$4E,$AB,$F9,$32,$F5,$FE,$93,$63,$E2,$37,$1F
	.dc.b	$BA,$D7,$FD,$33,$FA,$03,$E6,$FD,$4D,$AE,$BE,$C6,$27,$DD,$F4,$0F
	.dc.b	$9D,$BF,$39,$86,$DF,$F3,$99,$4F,$E1,$5B,$1F,$5D,$B8,$FE,$41,$AF
	.dc.b	$FA,$B3,$ED,$DF,$2F,$F3,$5B,$5D,$7E,$7B,$13,$F4,$F8,$BF,$C9,$D3
	.dc.b	$FE,$71,$E8,$FF,$56,$BF,$C1,$FC,$5F,$B1,$18,$C7,$FC,$6C,$D7,$FD
	.dc.b	$EA,$D7,$F9,$1D,$7E,$7F,$9B,$FD,$E6,$D1,$F0,$5E,$13,$C5,$E1,$DC
	.dc.b	$7F,$8B,$47,$83,$E5,$B1,$87,$81,$F4,$72,$F9,$3E,$86,$DF,$A5,$C3
	.dc.b	$FE,$06,$CD,$1E,$36,$C6,$63,$C9,$F0,$5E,$EF,$97,$F8,$8D,$9F,$89
	.dc.b	$1C,$48,$F5,$7D,$9E,$B7,$CB,$3E,$67,$C1,$59,$FE,$61,$1F,$2B,$E5
	.dc.b	$F1,$A7,$CA,$7E,$97,$BD,$F3,$BE,$5A,$DF,$D3,$70,$FA,$DF,$51,$1F
	.dc.b	$F1,$1F,$CF,$FE,$C5,$FB,$5F,$17,$FA,$F8,$C6,$1F,$62,$3E,$D9,$F7
	.dc.b	$3E,$2A,$D7,$E0,$5E,$1B,$7E,$03,$CD,$B3,$F8,$57,$0F,$E1,$FC,$6B
	.dc.b	$87,$E1,$1F,$DE,$7C,$57,$F5,$C9,$FF,$36,$F2,$E3,$D4,$CB,$23,$F8
	.dc.b	$94,$FE,$11,$B1,$FC,$96,$31,$7F,$E0,$63,$FB,$7F,$D0,$FE,$50,$EF
	.dc.b	$E4,$5B,$FF,$11,$7D,$0F,$F1,$06,$CF,$F0,$F8,$C5,$7F,$80,$54,$64
	.dc.b	$FE,$29,$EC,$FF,$84,$EB,$89,$FE,$2D,$71,$5F,$E9,$7E,$0F,$F4,$AC
	.dc.b	$FE,$A7,$E4,$BE,$87,$07,$EA,$19,$19,$3F,$DB,$D3,$C2,$BF,$57,$3F
	.dc.b	$C8,$FA,$FF,$D7,$35,$FE,$3B,$5F,$25,$99,$D3,$27,$F9,$D4,$E9,$B1
	.dc.b	$FD,$36,$31,$5F,$D5,$B5,$FE,$39,$E5,$AF,$E6,$46,$1F,$9D,$5F,$D4
	.dc.b	$AB,$FC,$4D,$39,$7F,$C3,$B8,$FC,$8A,$F5,$E9,$F8,$FE,$1F,$C8,$3F
	.dc.b	$C2,$32,$FE,$13,$EA,$5A,$FE,$1C,$FD,$3F,$1F,$E7,$F8,$96,$7E,$73
	.dc.b	$87,$E8,$7C,$E5,$CF,$FC,$71,$C2,$8F,$E9,$FF,$31,$FE,$0B,$AF,$F2
	.dc.b	$5F,$07,$F8,$86,$3F,$12,$DF,$0F,$7C,$D2,$B6,$7F,$83,$BF,$30,$FF
	.dc.b	$29,$B8,$CB,$FC,$29,$87,$28,$D4,$7F,$5E,$9D,$BF,$C9,$3E,$4F,$0F
	.dc.b	$F2,$87,$F6,$5B,$7F,$2A,$F3,$6D,$FF,$2F,$70,$D3,$F0,$B1,$8A,$FC
	.dc.b	$77,$9B,$FB,$B3,$F7,$97,$6B,$F1,$71,$FC,$D8,$C4,$FF,$36,$7F,$19
	.dc.b	$3F,$8D,$BB,$5F,$8D,$70,$FE,$29,$FC,$41,$1F,$C6,$CF,$E1,$E9,$ED
	.dc.b	$3E,$63,$F0,$F3,$98,$C4,$FF,$9F,$3A,$3F,$B2,$3C,$2B,$77,$FE,$80
	.dc.b	$AD,$F9,$6E,$8F,$EB,$B1,$86,$D7,$FA,$1A,$75,$E6,$F7,$8C,$4E,$B0
	.dc.b	$FC,$AF,$B3,$E7,$C7,$F8,$46,$1D,$63,$13,$C1,$AF,$09,$FB,$6F,$F9
	.dc.b	$EB,$D8,$7D,$07,$78,$C3,$6F,$FC,$53,$F2,$76,$B7,$7E,$DC,$BD,$11
	.dc.b	$FC,$23,$4F,$7B,$63,$3F,$3F,$C7,$E1,$39,$75,$F9,$32,$FF,$1A,$B6
	.dc.b	$FF,$BD,$E9,$F4,$CF,$34,$67,$E4,$BC,$47,$8F,$E4,$EE,$9E,$3A,$F6
	.dc.b	$7C,$67,$9C,$F4,$7C,$5D,$9D,$65,$FE,$44,$F8,$CF,$01,$23,$E2,$C6
	.dc.b	$5B,$9B,$E7,$BE,$11,$F0,$FE,$46,$CF,$49,$7C,$97,$C7,$3E,$47,$CC
	.dc.b	$59,$D7,$C9,$7B,$C8,$F6,$7E,$17,$D1,$7C,$F3,$E8,$36,$3B,$7C,$67
	.dc.b	$D3,$3E,$A7,$D1,$58,$C9,$E8,$8F,$98,$47,$07,$C2,$F9,$7F,$F4,$A3
	.dc.b	$C3,$87,$F3,$47,$FA,$FA,$32,$E3,$64,$3E,$37,$F0,$CD,$AE,$9E,$FE
	.dc.b	$26,$47,$CC,$9D,$5F,$DD,$BE,$75,$F8,$39,$CD,$7E,$71,$DF,$57,$EA
	.dc.b	$9F,$DA,$47,$A6,$7C,$3D,$5F,$99,$C1,$F3,$7F,$96,$7D,$66,$3B,$75
	.dc.b	$E6,$7F,$9C,$72,$FD,$DC,$E7,$1F,$C5,$34,$FF,$53,$5F,$38,$FE,$ED
	.dc.b	$19,$6C,$66,$3F,$BD,$3E,$19,$EA,$BB,$BC,$CE,$71,$DC,$60,$97,$E2
	.dc.b	$9D,$58,$F6,$CE,$9F,$2B,$FB,$E7,$33,$FB,$E6,$9E,$AE,$DF,$F4,$D3
	.dc.b	$F1,$29,$18,$C3,$CF,$7E,$68,$C6,$1E,$C6,$DF,$8E,$F8,$FC,$57,$9E
	.dc.b	$F8,$3E,$9A,$DD,$7D,$46,$D7,$C3,$C7,$55,$2B,$C9,$5A,$BE,$71,$89
	.dc.b	$D6,$3E,$6B,$B8,$F9,$66,$D7,$96,$F8,$DF,$D2,$C3,$C2,$75,$C3,$75
	.dc.b	$E5,$BB,$8F,$96,$3C,$04,$BE,$53,$EF,$AD,$DF,$CC,$AF,$C1,$B5,$F2
	.dc.b	$1F,$EE,$AB,$F8,$EF,$FB,$B3,$E3,$DE,$C3,$E8,$BF,$DB,$11,$86,$C6
	.dc.b	$67,$83,$5F,$C8,$F8,$4F,$8E,$FF,$70,$60,$8C,$4F,$85,$AF,$09,$DC
	.dc.b	$6B,$4F,$CB,$B7,$6F,$CD,$B6,$BF,$01,$E1,$FF,$95,$EB,$77,$F9,$7A
	.dc.b	$D5,$7E,$C9,$B8,$7F,$AE,$23,$53,$1F,$C9,$4E,$AF,$A6,$1E,$0A,$D6
	.dc.b	$1D,$AF,$F3,$15,$A7,$F2,$57,$D8,$95,$FA,$99,$D4,$F8,$E3,$9D,$F5
	.dc.b	$8E,$66,$0E,$71,$5F,$E1,$73,$CB,$0F,$99,$39,$6C,$F2,$3D,$71,$9B
	.dc.b	$F6,$47,$8C,$C4,$57,$86,$BF,$C2,$AF,$FD,$04,$D8,$FD,$D5,$FC,$97
	.dc.b	$F8,$9F,$A3,$E3,$A9,$1F,$C1,$E3,$CC,$61,$F2,$65,$FC,$4A,$33,$1B
	.dc.b	$21,$FC,$9B,$85,$E0,$FC,$DE,$67,$D1,$7E,$23,$0F,$F0,$DB,$FD,$06
	.dc.b	$FF,$B5,$B5,$FE,$8C,$EF,$6F,$F0,$E3,$61,$5D,$39,$BB,$4F,$1B,$EB
	.dc.b	$B6,$09,$7D,$9D,$F4,$70,$C7,$8F,$CF,$CA,$F1,$C6,$5F,$2C,$65,$F2
	.dc.b	$C6,$5A,$70,$3D,$7F,$45,$E4,$FA,$2F,$A7,$1E,$A6,$BE,$AC,$91,$8D
	.dc.b	$3E,$3B,$61,$F1,$7C,$EB,$E3,$CF,$C0,$8F,$83,$EA,$EC,$93,$FE,$12
	.dc.b	$A4,$73,$F5,$7E,$39,$27,$E8,$FB,$FF,$2B,$E2,$BE,$C6,$CF,$66,$99
	.dc.b	$AF,$AF,$F1,$5F,$40,$8F,$DA,$F8,$9F,$A7,$F3,$12,$3E,$EB,$F5,$BE
	.dc.b	$2B,$EB,$F4,$76,$F5,$65,$7E,$0A,$3E,$DF,$CE,$48,$FC,$2B,$F8,$BF
	.dc.b	$A1,$FE,$D3,$D3,$FE,$64,$7F,$94,$37,$C5,$D1,$BF,$8E,$E1,$EC,$FB
	.dc.b	$3D,$89,$8F,$F8,$EB,$04,$F5,$F8,$E9,$39,$7C,$1F,$1D,$FB,$9F,$5F
	.dc.b	$F7,$5E,$14,$AF,$ED,$D1,$FB,$DF,$9E,$D4,$60,$FE,$E4,$E5,$FE,$5A
	.dc.b	$FF,$80,$9F,$EF,$ED,$7F,$B2,$37,$BB,$FC,$4F,$E0,$FF,$BF,$7E,$0F
	.dc.b	$FD,$1D,$0F,$94,$FF,$2F,$46,$1B,$E4,$FF,$A0,$FD,$15,$F1,$8F,$F4
	.dc.b	$36,$EF,$8E,$C6,$7F,$E3,$7A,$B9,$78,$23,$FD,$7C,$AB,$84,$63,$07
	.dc.b	$8B,$C3,$2F,$F4,$1C,$F0,$6A,$CF,$D3,$5C,$23,$95,$78,$23,$11,$BC
	.dc.b	$FD,$53,$E4,$C1,$F1,$BB,$F1,$A5,$F8,$DF,$5F,$90,$F8,$67,$81,$86
	.dc.b	$D7,$D1,$9D,$4F,$D1,$6C,$66,$76,$FF,$8E,$21,$F1,$2F,$77,$A5,$CA
	.dc.b	$FF,$72,$56,$A7,$E6,$C3,$E8,$9E,$2E,$1F,$E4,$1C,$1F,$E9,$76,$EF
	.dc.b	$FE,$37,$AE,$15,$BE,$3F,$C3,$47,$36,$D7,$DE,$8F,$C5,$35,$FF,$27
	.dc.b	$F1,$FF,$9C,$3C,$9F,$AD,$EB,$49,$FB,$51,$F4,$7C,$C9,$1F,$75,$FA
	.dc.b	$1F,$19,$E7,$F3,$24,$FF,$C0,$13,$D8,$FB,$BE,$31,$F2,$3F,$03,$D4
	.dc.b	$F8,$D5,$FE,$0C,$E1,$F4,$C6,$23,$FC,$39,$18,$F8,$09,$1F,$E1,$26
	.dc.b	$BE,$AE,$BF,$F0,$FB,$F8,$DF,$6B,$F2,$07,$E2,$7D,$89,$1F,$95,$6B
	.dc.b	$F8,$B3,$FA,$4F,$63,$E2,$7E,$AF,$D1,$7E,$72,$74,$17,$C9,$68,$DF
	.dc.b	$85,$C3,$E4,$3E,$9C,$9A,$4A,$FD,$F7,$37,$ED,$FD,$CF,$D3,$12,$3F
	.dc.b	$55,$C7,$F2,$7C,$FF,$BB,$C8,$69,$27,$F8,$37,$D1,$7E,$09,$34,$8F
	.dc.b	$F0,$9E,$17,$AF,$FB,$5F,$76,$4F,$EE,$63,$F6,$1E,$CE,$C9,$1E,$8A
	.dc.b	$47,$3F,$0F,$F7,$DE,$EC,$AF,$E0,$51,$FB,$9F,$96,$F0,$12,$39,$3F
	.dc.b	$C6,$FC,$B7,$C5,$D2,$4F,$C6,$8C,$BE,$1F,$6F,$D4,$47,$07,$97,$DB
	.dc.b	$FE,$10,$47,$F8,$BF,$D1,$3E,$33,$F8,$DB,$63,$F8,$C4,$ED,$FC,$27
	.dc.b	$2A,$75,$18,$C2,$BF,$B3,$9F,$E6,$17,$EB,$24,$7C,$7F,$77,$E8,$11
	.dc.b	$C1,$ED,$F8,$CF,$84,$47,$07,$CF,$F7,$7E,$C1,$1C,$1F,$4F,$C4,$FE
	.dc.b	$4C,$47,$DF,$E1,$7B,$8C,$3E,$CE,$3F,$D9,$BA,$3F,$9A,$7A,$DF,$4C
	.dc.b	$8F,$D1,$7F,$11,$D1,$23,$F0,$4D,$7F,$AC,$7E,$51,$FC,$37,$CF,$79
	.dc.b	$FF,$C5,$E3,$0D,$F3,$BF,$90,$FD,$1B,$7F,$E4,$7F,$46,$E0,$FE,$36
	.dc.b	$DC,$BF,$B1,$35,$FE,$CC,$D7,$FD,$D2,$FF,$93,$9D,$E9,$C8,$8F,$F6
	.dc.b	$DF,$9F,$7E,$69,$FB,$A8,$FF,$71,$6B,$E8,$53,$AF,$1F,$FB,$1B,$63
	.dc.b	$F8,$F7,$D2,$77,$7F,$82,$F9,$8E,$AF,$E6,$6D,$8F,$EF,$4D,$7F,$BF
	.dc.b	$35,$FF,$80,$2B,$9B,$FA,$2B,$F1,$63,$F9,$63,$5E,$F1,$FD,$1C,$D9
	.dc.b	$1F,$E6,$5E,$9F,$37,$FB,$87,$B9,$FD,$CC,$8F,$F7,$AF,$57,$FA,$63
	.dc.b	$FC,$7B,$E8,$FF,$C8,$AF,$F8,$45,$AF,$F8,$55,$AF,$06,$A3,$FD,$91
	.dc.b	$EA,$7F,$34,$7A,$79,$5F,$D8,$24,$FF,$74,$9D,$6F,$F4,$69,$F4,$6F
	.dc.b	$94,$FE,$BA,$D8,$FF,$12,$35,$FF,$14,$35,$F5,$AF,$92,$E2,$48,$E2
	.dc.b	$FF,$68,$F8,$EF,$F3,$1D,$FA,$7D,$BF,$F4,$84,$62,$BF,$DC,$5F,$1B
	.dc.b	$FD,$C6,$31,$3A,$F6,$FF,$C3,$55,$A3,$FC,$A6,$53,$FD,$B8,$53,$E3
	.dc.b	$7C,$4F,$7B,$FF,$20,$36,$3E,$57,$37,$F8,$33,$BF,$FC,$84,$D8,$FC
	.dc.b	$D7,$BD,$F1,$DF,$F0,$24,$65,$FB,$6F,$F5,$B9,$E3,$EC,$6E,$7A,$B5
	.dc.b	$FC,$FB,$9C,$9F,$46,$F8,$8F,$D2,$6D,$F5,$7E,$57,$FB,$DE,$3F,$E1
	.dc.b	$CA,$D3,$FE,$01,$8C,$E3,$FD,$A3,$C7,$FC,$24,$FF,$96,$DB,$1F,$E5
	.dc.b	$F6,$BF,$92,$6B,$F8,$BF,$2F,$FC,$B9,$F0,$DE,$2A,$C7,$9A,$8C,$57
	.dc.b	$FB,$0B,$A9,$F1,$78,$C3,$8E,$1F,$E1,$5F,$A2,$FE,$2A,$FF,$85,$F1
	.dc.b	$FE,$B6,$E4,$E7,$87,$F9,$AF,$D6,$B3,$A8,$FF,$01,$BF,$DA,$A7,$94
	.dc.b	$EB,$D7,$DF,$AB,$F3,$68,$D2,$57,$F1,$3F,$CB,$7E,$1F,$0F,$F1,$1E
	.dc.b	$54,$62,$7F,$C8,$1F,$46,$F9,$BF,$7D,$6B,$4F,$3F,$D1,$FF,$AD,$6F
	.dc.b	$FD,$10,$D8,$FF,$46,$35,$FD,$FB,$5F,$F2,$62,$F9,$1F,$C2,$36,$79
	.dc.b	$C6,$0F,$C4,$7A,$0F,$27,$E3,$B6,$7F,$C8,$5F,$46,$F7,$B3,$5F,$4F
	.dc.b	$99,$F2,$5F,$93,$E6,$FF,$30,$B6,$33,$87,$F8,$99,$BA,$7F,$96,$38
	.dc.b	$FF,$82,$4F,$F8,$17,$C6,$E5,$18,$9D,$47,$5B,$FE,$1F,$5F,$C7,$FE
	.dc.b	$4B,$F0,$91,$8A,$FE,$FF,$87,$FB,$02,$B4,$48,$FF,$01,$CE,$F9,$3F
	.dc.b	$D5,$F5,$B9,$E3,$F4,$7E,$92,$5F,$A5,$F2,$FC,$54,$8F,$F0,$E3,$67
	.dc.b	$FC,$E7,$4F,$93,$F8,$44,$AF,$84,$FB,$BD,$C9,$1F,$E2,$56,$D7,$F9
	.dc.b	$F3,$2F,$ED,$93,$FC,$F7,$E4,$FD,$FC,$7F,$45,$91,$93,$FD,$77,$3C
	.dc.b	$2B,$E1,$CF,$D7,$F5,$7E,$A1,$2B,$EA,$3F,$43,$E7,$24,$7F,$8D,$DB
	.dc.b	$5E,$CF,$8B,$FB,$84,$9F,$B8,$FD,$FF,$8A,$FD,$B7,$1B,$E7,$3F,$0D
	.dc.b	$18,$7F,$11,$F2,$9F,$85,$25,$FE,$15,$FC,$47,$99,$CB,$D1,$FC,$73
	.dc.b	$67,$FD,$29,$F1,$7F,$7F,$3F,$EB,$ED,$7F,$D5,$8D,$7F,$AD,$35,$3F
	.dc.b	$C0,$5E,$3E,$2F,$F2,$98,$C4,$72,$8F,$C9,$FC,$DF,$B0,$FD,$07,$D1
	.dc.b	$5B,$FF,$53,$7D,$1A,$3F,$99,$36,$3F,$E0,$03,$6F,$FB,$A8,$70,$FF
	.dc.b	$78,$F7,$3F,$60,$FF,$A5,$9F,$F3,$5C,$FF,$AC,$DC,$F5,$FF,$6D,$73
	.dc.b	$3F,$A0,$F9,$4F,$DA,$1F,$C9,$FE,$7F,$FD,$31,$7F,$E0,$58,$CB,$FE
	.dc.b	$F9,$F9,$FB,$5F,$E6,$37,$F9,$3F,$7B,$E8,$FF,$A6,$36,$BD,$34,$8C
	.dc.b	$3E,$C6,$BF,$E1,$25,$F4,$5C,$89,$3C,$A3,$2F,$4C,$BF,$CF,$8F,$BF
	.dc.b	$E3,$7F,$9C,$F0,$FF,$35,$C3,$E2,$47,$F1,$1C,$65,$7F,$90,$0F,$E9
	.dc.b	$DA,$7C,$FD,$7F,$7B,$3F,$C5,$F3,$FE,$95,$E4,$BD,$7C,$61,$C7,$27
	.dc.b	$8C,$AD,$4F,$D1,$F2,$3F,$B7,$6D,$FC,$8E,$3D,$11,$83,$3F,$25,$FD
	.dc.b	$42,$7F,$CC,$53,$F6,$63,$1F,$13,$FD,$7B,$27,$CD,$7F,$13,$AF,$F0
	.dc.b	$C7,$D1,$FF,$47,$D3,$CD,$39,$7D,$11,$9A,$3F,$DB,$A3,$27,$F8,$05
	.dc.b	$C4,$6A,$3F,$9A,$7B,$9F,$CE,$4F,$14,$F7,$D5,$E4,$FF,$83,$EB,$F9
	.dc.b	$5F,$A3,$FC,$FC,$FC,$67,$05,$CF,$FC,$A1,$1F,$33,$DE,$C5,$7F,$AC
	.dc.b	$38,$51,$F9,$3F,$A3,$FE,$90,$5B,$FF,$2C,$38,$7E,$96,$5F,$47,$2C
	.dc.b	$61,$CD,$FC,$AE,$3F,$94,$25,$7E,$2D,$B2,$F7,$DF,$43,$FE,$5B,$8F
	.dc.b	$CC,$63,$FE,$85,$9C,$C7,$F9,$7F,$AB,$FB,$A6,$1F,$E3,$CB,$E1,$0B
	.dc.b	$E1,$18,$6D,$71,$FA,$0A,$F6,$BE,$4E,$5B,$9B,$A7,$FB,$E6,$5E,$28
	.dc.b	$C5,$F4,$75,$EA,$DA,$DE,$5F,$9E,$F8,$1D,$70,$7B,$B2,$EB,$5F,$E3
	.dc.b	$AC,$FE,$15,$FF,$8C,$6B,$51,$83,$F3,$FF,$00,$FA,$3F,$03,$70,$F3
	.dc.b	$AF,$9F,$CC,$EB,$D8,$95,$AF,$9B,$B8,$75,$D9,$8F,$E9,$E3,$19,$3E
	.dc.b	$4F,$BF,$7F,$A9,$F0,$FC,$CC,$CB,$C3,$E1,$DC,$3A,$63,$FA,$C6,$F7
	.dc.b	$FA,$3E,$4B,$D2,$FA,$BE,$3B,$E2,$FD,$0D,$C7,$E8,$EA,$8F,$7F,$BD
	.dc.b	$27,$DF,$EE,$DC,$BB,$FB,$9F,$EC,$B9,$FC,$57,$0F,$B2,$31,$3F,$E5
	.dc.b	$DD,$51,$88,$F9,$AD,$9F,$87,$EA,$7C,$EF,$4F,$E7,$B7,$0F,$9F,$1B
	.dc.b	$C3,$FD,$41,$1F,$E6,$CA,$F9,$73,$93,$5E,$8F,$D8,$3E,$CF,$9D,$6B
	.dc.b	$EC,$C7,$E2,$FE,$23,$F8,$DE,$7F,$2A,$31,$E4,$EE,$3E,$C4,$EB,$C1
	.dc.b	$FC,$A6,$6B,$DE,$1F,$9A,$AF,$DE,$B7,$C6,$ED,$E3,$77,$64,$6F,$BE
	.dc.b	$E5,$D9,$B9,$7E,$87,$0F,$CA,$B6,$FD,$13,$FA,$CF,$A0,$FC,$A5,$FE
	.dc.b	$64,$38,$AF,$DD,$C7,$FB,$C7,$0E,$1D,$F1,$FD,$9C,$70,$F0,$30,$F3
	.dc.b	$BF,$CF,$4E,$9E,$98,$7F,$09,$73,$3B,$BF,$25,$CB,$FF,$4F,$D6,$9F
	.dc.b	$F7,$9B,$6B,$94,$FE,$15,$F1,$4F,$7F,$03,$FD,$3F,$1E,$39,$F8,$11
	.dc.b	$8C,$7D,$B7,$FE,$77,$F8,$8E,$07,$9E,$B6,$E2,$FD,$97,$C7,$2F,$0E
	.dc.b	$39,$C7,$9C,$E6,$BF,$E0,$06,$3D,$5F,$15,$6E,$FE,$15,$6E,$7E,$8B
	.dc.b	$6B,$E2,$DF,$4C,$38,$57,$4A,$D6,$3E,$68,$F2,$CE,$9F,$86,$60,$95
	.dc.b	$CD,$F1,$98,$C3,$C7,$3C,$70,$F0,$C6,$27,$47,$86,$7E,$36,$3B,$C7
	.dc.b	$E7,$CF,$D6,$6B,$C4,$C4,$FF,$88,$3E,$31,$F4,$0F,$A3,$C3,$6B,$FC
	.dc.b	$7B,$3A,$9C,$9F,$E2,$8C,$9E,$27,$F8,$46,$BF,$C8,$71,$8E,$4D,$4E
	.dc.b	$67,$F1,$2D,$7F,$24,$F1,$D7,$F0,$6D,$D7,$35,$F9,$9F,$13,$FB,$05
	.dc.b	$6A,$3F,$39,$5B,$BF,$BE,$D7,$F5,$75,$C2,$F9,$C7,$E9,$67,$58,$79
	.dc.b	$E3,$F5,$15,$B7,$9D,$F6,$23,$A2,$E1,$E3,$7A,$47,$9D,$CE,$1E,$09
	.dc.b	$D1,$FA,$0A,$ED,$3F,$C0,$30,$F1,$92,$F9,$47,$42,$57,$2B,$FD,$D5
	.dc.b	$F2,$C3,$D5,$39,$AF,$49,$E7,$23,$97,$C1,$7D,$A3,$C9,$1E,$83,$8B
	.dc.b	$87,$96,$DF,$C0,$AF,$FC,$24,$D8,$FE,$0F,$18,$7C,$A7,$5C,$2B,$FC
	.dc.b	$57,$D0,$F4,$4C,$8D,$1F,$B8,$8F,$A5,$D3,$FE,$59,$FA,$2B,$E0,$3E
	.dc.b	$46,$98,$89,$7F,$D9,$1A,$FF,$76,$F9,$0B,$A7,$F8,$56,$D5,$8C,$FF
	.dc.b	$C1,$95,$E0,$FA,$3F,$CC,$5F,$F8,$6D,$FF,$20,$E1,$FE,$20,$6B,$FE
	.dc.b	$05,$77,$48,$FF,$38,$C6,$1B,$B7,$53,$B7,$55,$8E,$91,$5E,$91,$F0
	.dc.b	$39,$3E,$0F,$C9,$C9,$FF,$00,$C6,$5A,$43,$FC,$2B,$EE,$FC,$83,$4F
	.dc.b	$F8,$62,$30,$FF,$82,$BE,$3F,$E6,$91,$F7,$6B,$F0,$F7,$91,$F0,$5F
	.dc.b	$8D,$B7,$C7,$B9,$3F,$15,$FA,$FF,$1F,$F6,$3E,$82,$4F,$F8,$61,$FB
	.dc.b	$7F,$41,$F7,$3E,$62,$4F,$F8,$71,$FB,$FF,$19,$F8,$0F,$76,$4F,$F8
	.dc.b	$81,$FC,$1F,$CF,$FE,$13,$E7,$E4,$FF,$89,$1F,$C4,$FE,$81,$FE,$67
	.dc.b	$8F,$AD,$5F,$74,$8F,$DB,$AD,$F8,$32,$3F,$CB,$B5,$FC,$B3,$A3,$FC
	.dc.b	$B8,$BE,$3F,$F6,$1E,$A7,$E7,$EB,$FA,$7B,$5F,$EF,$6D,$F0,$0D,$38
	.dc.b	$C7,$F7,$57,$F7,$5E,$9C,$8C,$10,$FE,$F8,$FB,$3D,$CF,$6F,$C8,$C9
	.dc.b	$F3,$C6,$5A,$43,$FC,$B5,$8B,$E4,$7E,$2F,$C8,$C9,$F7,$BF,$23,$E4
	.dc.b	$64,$60,$87,$27,$F9,$37,$7B,$E3,$1E,$CE,$17,$CB,$D3,$11,$2F,$F4
	.dc.b	$DE,$4F,$FA,$6E,$F9,$4F,$CB,$6C,$E8,$FD,$63,$85,$3C,$6F,$C3,$1F
	.dc.b	$E9,$69,$D7,$17,$FA,$91,$BE,$2F,$2A,$7B,$3E,$AF,$D8,$5F,$21,$7E
	.dc.b	$42,$65,$CE,$72,$D9,$E0,$47,$57,$F9,$0F,$92,$91,$F8,$A7,$F2,$73
	.dc.b	$23,$04,$3E,$AB,$FC,$EF,$6F,$CC,$47,$F8,$DD,$AF,$F8,$71,$AF,$F9
	.dc.b	$21,$BC,$FE,$C8,$FE,$ED,$96,$08,$F2,$F8,$E7,$77,$FB,$2F,$E3,$BE
	.dc.b	$5C,$FF,$8A,$2B,$1C,$3F,$E4,$76,$C7,$F1,$DB,$E2,$FF,$B7,$BE,$95
	.dc.b	$BE,$BC,$F4,$62,$75,$39,$8F,$C4,$47,$1D,$F0,$47,$A7,$C9,$78,$17
	.dc.b	$CE,$FF,$99,$34,$F3,$C6,$38,$F5,$39,$D7,$FD,$CC,$FF,$A9,$67,$F9
	.dc.b	$A5,$70,$8C,$42,$31,$F0,$7C,$8E,$7E,$5F,$87,$87,$FD,$47,$5F,$E2
	.dc.b	$1F,$82,$F9,$E4,$7F,$C4,$9F,$2B,$F4,$88,$E0,$FB,$38,$BD,$0E,$93
	.dc.b	$27,$CB,$9F,$A2,$7E,$07,$D2,$FC,$99,$F3,$4F,$FC,$01,$8C,$69,$F0
	.dc.b	$9B,$19,$DF,$89,$2B,$A9,$FD,$56,$7E,$8B,$D0,$71,$7D,$8F,$8C,$FA
	.dc.b	$DF,$17,$27,$EF,$3F,$77,$E9,$3E,$8E,$5F,$88,$B4,$63,$2F,$EC,$51
	.dc.b	$87,$81,$C7,$E0,$3F,$0D,$E2,$7B,$71,$FF,$69,$B5,$E2,$D7,$CA,$DA
	.dc.b	$3E,$AC,$3C,$4F,$C2,$9D,$F3,$F6,$D7,$BF,$37,$83,$E5,$65,$7C,$4C
	.dc.b	$79,$CE,$A3,$18,$23,$11,$F1,$E7,$C3,$F2,$BE,$AF,$92,$FD,$AC,$79
	.dc.b	$1B,$66,$98,$F9,$4E,$17,$DA,$3E,$1C,$7C,$4F,$93,$E2,$49,$FE,$07
	.dc.b	$7F,$02,$3A,$FC,$9E,$E7,$CF,$1F,$32,$3E,$6F,$1F,$CE,$F9,$39,$5E
	.dc.b	$C8,$CB,$48,$7B,$72,$72,$7E,$EF,$E1,$FF,$5F,$3F,$E1,$D3,$F0,$EE
	.dc.b	$1F,$BE,$D7,$3C,$9F,$12,$1C,$9C,$E2,$4F,$D1,$9C,$BF,$96,$1F,$A3
	.dc.b	$33,$FE,$3E,$D8,$FE,$46,$D7,$F9,$3B,$5F,$F0,$02,$CB,$89,$23,$8B
	.dc.b	$FC,$AB,$F3,$AF,$F5,$4F,$37,$F0,$CA,$E1,$5E,$36,$BF,$A7,$13,$C3
	.dc.b	$1F,$89,$F4,$8F,$93,$FE,$6A,$DA,$FE,$72,$D7,$F9,$EB,$5F,$F0,$B3
	.dc.b	$FD,$69,$CF,$9B,$C8,$8F,$F4,$2F,$92,$7B,$5F,$C9,$5B,$1F,$25,$AF
	.dc.b	$9F,$E7,$BF,$AC,$B7,$CC,$FE,$59,$18,$D7,$F9,$5B,$73,$FE,$4F,$FA
	.dc.b	$C3,$AB,$F9,$97,$D2,$7A,$6F,$F9,$8B,$6B,$F0,$8D,$C7,$FC,$C5,$F1
	.dc.b	$EF,$7F,$F9,$D2,$F7,$FF,$AC,$B6,$7F,$AE,$B5,$FE,$C2,$D7,$F2,$1D
	.dc.b	$9F,$C3,$BE,$4B,$F8,$91,$23,$F7,$5E,$23,$E5,$3F,$AB,$FC,$FB,$F3
	.dc.b	$27,$95,$FD,$AA,$30,$D8,$CD,$FF,$70,$9C,$BF,$DC,$3E,$93,$FA,$0B
	.dc.b	$FD,$D1,$AF,$F9,$35,$AF,$E9,$5A,$FF,$A0,$FE,$51,$F1,$5F,$D3,$E1
	.dc.b	$ED,$FF,$A2,$A7,$3C,$7F,$E1,$EF,$95,$FD,$74,$62,$B7,$18,$BE,$57
	.dc.b	$FE,$92,$9F,$F4,$87,$BB,$C5,$FF,$30,$35,$FF,$74,$9C,$8C,$94,$65
	.dc.b	$F7,$F9,$7E,$37,$37,$EE,$E3,$13,$F1,$63,$F8,$04,$3E,$FF,$FD,$57
	.dc.b	$5F,$25,$B5,$FE,$71,$9D,$7C,$6F,$09,$CC,$FD,$CF,$87,$7C,$E7,$F8
	.dc.b	$4B,$7F,$E9,$91,$8C,$7F,$9B,$2F,$67,$51,$FD,$3A,$35,$7F,$B5,$FA
	.dc.b	$1F,$ED,$D3,$FE,$28,$6C,$7F,$8B,$1A,$FF,$A1,$5A,$FF,$1D,$FA,$37
	.dc.b	$9F,$DA,$3F,$D7,$91,$8A,$FE,$2D,$4F,$8F,$D5,$2B,$4F,$F0,$7F,$A4
	.dc.b	$FF,$31,$DF,$F1,$D6,$C7,$F4,$BF,$34,$72,$8C,$E9,$FE,$1E,$7F,$CD
	.dc.b	$B5,$C7,$C7,$D7,$C1,$ED,$F0,$FC,$B7,$F2,$79,$FF,$23,$B7,$1F,$F2
	.dc.b	$5B,$5F,$F5,$23,$5F,$E7,$8F,$F2,$78,$E3,$E7,$7F,$3B,$8C,$4F,$28
	.dc.b	$FE,$7D,$83,$C0,$F8,$3E,$97,$F2,$1C,$3D,$D1,$8F,$36,$47,$F9,$6A
	.dc.b	$75,$87,$F9,$79,$AF,$FA,$DD,$AF,$F4,$F7,$E0,$BB,$CB,$FA,$74,$62
	.dc.b	$38,$47,$F5,$3F,$0F,$E5,$F8,$9F,$65,$F9,$BF,$33,$78,$13,$62,$63
	.dc.b	$FD,$BA,$51,$CD,$B1,$98,$C7,$CC,$F3,$3A,$7C,$CD,$9F,$EB,$91,$DF
	.dc.b	$E7,$F1,$9F,$DF,$7C,$5F,$23,$E3,$F9,$1B,$7F,$D8,$DC,$3E,$37,$C4
	.dc.b	$47,$CD,$BF,$AF,$6F,$63,$63,$FB,$64,$7A,$48,$FB,$5F,$67,$27,$C3
	.dc.b	$3E,$27,$C4,$D9,$FE,$E9,$1F,$1F,$E7,$F1,$9F,$E5,$DF,$95,$F0,$BF
	.dc.b	$2F,$CB,$B7,$FD,$ED,$C3,$F2,$DF,$94,$47,$FC,$37,$F4,$1F,$EE,$B7
	.dc.b	$FC,$64,$DB,$FF,$17,$2F,$92,$FB,$4E,$7A,$FF,$CA,$0B,$E4,$F2,$6B
	.dc.b	$C1,$B8,$7F,$82,$FE,$8F,$FD,$2A,$DF,$86,$E7,$F2,$7F,$B8,$9C,$E7
	.dc.b	$FD,$F5,$FF,$6E,$57,$1F,$87,$D5,$ED,$E1,$F0,$9F,$C0,$35,$7C,$3E
	.dc.b	$2F,$FC,$35,$18,$6C,$7F,$89,$63,$18,$FF,$AF,$9A,$FF,$8B,$4F,$81
	.dc.b	$1B,$F3,$65,$7F,$B0,$AB,$F9,$9D,$6D,$AF,$F8,$D7,$2F,$E0,$1B,$FF
	.dc.b	$A9,$9B,$59,$F2,$3E,$07,$C1,$7C,$1F,$16,$4F,$C7,$8B,$87,$D2,$CF
	.dc.b	$E8,$9F,$4B,$3B,$5F,$63,$E1,$BE,$A9,$8A,$FB,$2F,$F2,$4F,$8F,$7C
	.dc.b	$C7,$C7,$DF,$0D,$BF,$8F,$1F,$14,$8F,$BD,$F9,$1A,$48,$F8,$4D,$C3
	.dc.b	$E4,$78,$BF,$CA,$BE,$07,$FD,$B7,$E9,$BE,$3F,$FC,$AB,$F4,$38,$7F
	.dc.b	$CA,$F3,$6F,$FC,$B7,$36,$FF,$99,$B7,$C8,$36,$B0,$F1,$F5,$9F,$97
	.dc.b	$7F,$3C,$CB,$C1,$8F,$F7,$E8,$C6,$48,$C7,$85,$C6,$73,$1B,$F0,$B1
	.dc.b	$E1,$65,$BB,$FF,$7E,$C3,$96,$4F,$A4,$FE,$BB,$9F,$82,$31,$F1,$FE
	.dc.b	$59,$7C,$ED,$89,$87,$8C,$FF,$3E,$57,$07,$19,$F6,$BF,$ED,$31,$A6
	.dc.b	$C6,$6B,$B7,$A9,$D2,$18,$F8,$EB,$FB,$35,$FA,$67,$FD,$13,$E8,$FC
	.dc.b	$DF,$80,$F9,$CD,$9D,$72,$7F,$64,$9C,$57,$4D,$7F,$D1,$F7,$8C,$7E
	.dc.b	$7D,$BF,$1B,$FF,$16,$E3,$C9,$B8,$71,$3F,$DB,$EB,$DE,$FC,$03,$AB
	.dc.b	$E6,$DC,$7F,$DB,$6B,$E8,$37,$BF,$07,$F1,$58,$7F,$17,$F9,$0B,$E2
	.dc.b	$7F,$AA,$B3,$F8,$13,$96,$C6,$63,$C8,$7F,$1F,$F9,$0E,$27,$2A,$B5
	.dc.b	$C2,$0F,$D6,$9F,$C6,$57,$38,$FE,$09,$3A,$9F,$F5,$CC,$E5,$F5,$E5
	.dc.b	$F0,$0F,$83,$EE,$DB,$FA,$3F,$48,$F5,$7F,$48,$D8,$CF,$AD,$89,$F9
	.dc.b	$3E,$F4,$9F,$A1,$F0,$17,$2F,$A1,$C7,$F5,$7E,$5B,$E7,$9F,$43,$E0
	.dc.b	$ED,$FD,$7F,$A0,$FA,$67,$D4,$FA,$1B,$1F,$6F,$B3,$EB,$9F,$63,$B6
	.dc.b	$C7,$A3,$D6,$F3,$9F,$73,$BA,$C7,$B3,$C4,$F5,$9E,$CF,$5E,$C7,$E1
	.dc.b	$FC,$9F,$C1,$9F,$03,$C6,$B1,$F8,$BF,$A1,$FC,$3F,$C6,$7E,$21,$B1
	.dc.b	$F8,$F8,$BC,$5F,$03,$BD,$58,$F0,$3F,$1B,$5F,$90,$F9,$BC,$76,$7E
	.dc.b	$39,$FE,$25,$F4,$9F,$95,$CB,$F4,$6D,$CB,$E5,$BF,$93,$8C,$F8,$7C
	.dc.b	$08,$F0,$F0,$AD,$70,$9E,$B8,$63,$2F,$D4,$E1,$FC,$9F,$0F,$D4,$7C
	.dc.b	$77,$E9,$0F,$D2,$FD,$15,$8C,$93,$1F,$74,$65,$F8,$13,$FA,$CF,$94
	.dc.b	$C6,$1F,$AC,$7F,$4F,$F3,$9F,$E6,$0C,$3F,$4A,$DB,$FD,$1A,$F9,$9D
	.dc.b	$FA,$12,$B6,$F2,$8C,$57,$08,$FD,$04,$FF,$6B,$AE,$8D,$8C,$CF,$F3
	.dc.b	$31,$3A,$70,$F8,$F0,$FE,$77,$1A,$F8,$77,$D1,$78,$B0,$F4,$C6,$27
	.dc.b	$C5,$18,$9D,$47,$B2,$75,$5D,$2F,$FC,$A7,$5B,$AE,$26,$DC,$4E,$EB
	.dc.b	$8E,$DE,$6C,$7E,$5C,$62,$FF,$A7,$61,$D1,$E4,$74,$BE,$58,$74,$C3
	.dc.b	$DF,$7C,$6B,$E1,$E1,$D2,$B5,$1D,$27,$58,$7C,$5B,$E3,$5A,$AF,$8C
	.dc.b	$FC,$0C,$3B,$C7,$89,$A7,$C1,$C3,$B4,$66,$B9,$BE,$76,$CF,$06,$9F
	.dc.b	$07,$1F,$06,$1E,$FA,$FC,$06,$3F,$4D,$B8,$FF,$8B,$5F,$53,$57,$C8
	.dc.b	$3B,$DF,$49,$98,$C3,$63,$33,$CE,$72,$F5,$C3,$FD,$13,$E6,$F5,$6C
	.dc.b	$67,$E2,$BF,$E0,$0B,$67,$51,$8D,$7F,$08,$DD,$7F,$26,$DA,$FB,$BE
	.dc.b	$47,$E6,$6B,$77,$FA,$AA,$D5,$7E,$71,$B8,$7F,$3B,$8D,$4C,$7F,$5F
	.dc.b	$3A,$BE,$98,$78,$2B,$58,$76,$BF,$DA,$D6,$9F,$C7,$DF,$62,$57,$EF
	.dc.b	$27,$53,$E3,$8E,$77,$D6,$39,$98,$39,$C5,$7F,$6B,$3C,$B0,$F9,$78
	.dc.b	$78,$C9,$7F,$B8,$9D,$61,$E2,$BF,$6E,$1E,$37,$A6,$1C,$AF,$C5,$87
	.dc.b	$EC,$F0,$EC,$72,$BD,$B8,$BE,$38,$7C,$77,$A1,$FD,$BC,$FA,$A7,$F7
	.dc.b	$53,$D2,$5C,$3E,$93,$B3,$27,$C2,$25,$FF,$05,$F2,$7E,$77,$8E,$F8
	.dc.b	$D7,$93,$85,$E4,$D3,$11,$30,$FF,$03,$57,$18,$FE,$41,$C7,$FC,$86
	.dc.b	$31,$3A,$BE,$AD,$AE,$4F,$F7,$A8,$CB,$7C,$9F,$2F,$D2,$7E,$72,$3F
	.dc.b	$BF,$BF,$E2,$D7,$FC,$76,$D7,$3B,$F0,$3C,$6F,$F9,$97,$04,$F3,$C7
	.dc.b	$CC,$D9,$D3,$54,$F0,$AE,$91,$E0,$7C,$BD,$BF,$1F,$8F,$E4,$79,$52
	.dc.b	$BF,$C9,$F9,$FF,$01,$6C,$F9,$1E,$B1,$97,$EF,$61,$F4,$E7,$4D,$8F
	.dc.b	$95,$E5,$77,$79,$9D,$9F,$B1,$E5,$48,$F0,$BF,$2B,$6F,$97,$BF,$4D
	.dc.b	$7E,$ED,$7D,$A6,$BE,$26,$ED,$F1,$B2,$FA,$64,$9F,$6F,$93,$F8,$7F
	.dc.b	$1E,$47,$C0,$7F,$15,$E2,$FE,$2F,$C5,$F5,$EE,$69,$2B,$F0,$DE,$0F
	.dc.b	$E4,$48,$FC,$AF,$0B,$F2,$9E,$14,$8F,$9A,$FE,$6B,$E3,$7F,$DB,$DD
	.dc.b	$3F,$E7,$8D,$D8,$7B,$C9,$7E,$F7,$E6,$D7,$F3,$BA,$FF,$3F,$B7,$B7
	.dc.b	$F1,$0D,$AF,$C1,$AF,$86,$FE,$11,$F1,$9F,$C2,$7E,$36,$57,$F8,$42
	.dc.b	$32,$FE,$D5,$FC,$03,$FE,$04,$6F,$71,$A7,$08,$FF,$0B,$BF,$C5,$7E
	.dc.b	$37,$DD,$F1,$52,$7F,$D3,$78,$A3,$E1,$7A,$92,$73,$1B,$7F,$CC,$F2
	.dc.b	$9F,$91,$F1,$5F,$C0,$BE,$CF,$E4,$EB,$FA,$76,$DF,$C0,$9D,$BF,$EA
	.dc.b	$A8,$CF,$71,$B8,$2B,$F9,$D7,$82,$F8,$A7,$BB,$DF,$CA,$99,$7F,$AE
	.dc.b	$6B,$53,$C2,$B7,$39,$6B,$FE,$AF,$9D,$B9,$F8,$DE,$07,$1A,$91,$93
	.dc.b	$B7,$D1,$BE,$37,$CB,$F1,$EF,$8A,$A7,$A2,$75,$5C,$2B,$04,$79,$FC
	.dc.b	$DF,$C9,$E8,$91,$F1,$5F,$CA,$7D,$1C,$8F,$AE,$D7,$E9,$CF,$F5,$1C
	.dc.b	$E4,$7F,$96,$1F,$EA,$99,$C8,$FF,$00,$3F,$E0,$D7,$FC,$9A,$DE,$8F
	.dc.b	$8B,$1F,$E5,$5F,$87,$FF,$29,$7C,$87,$F4,$99,$F9,$2D,$AC,$F8,$FF
	.dc.b	$E5,$4A,$B7,$FC,$83,$D7,$7B,$F8,$E8,$46,$24,$31,$FF,$62,$4E,$AB
	.dc.b	$87,$D1,$FC,$AA,$D6,$3F,$CC,$E3,$17,$C3,$0E,$95,$A3,$C5,$7C,$B8
	.dc.b	$3C,$B0,$FA,$37,$C1,$E1,$FF,$75,$47,$93,$0F,$ED,$11,$88,$E8,$71
	.dc.b	$9E,$1D,$9C,$59,$1C,$73,$FF,$3D,$57,$F9,$DA,$FF,$B9,$7C,$6B,$C7
	.dc.b	$F1,$B6,$F8,$1F,$11,$C3,$E0,$E1,$7C,$57,$A9,$7C,$17,$D9,$24,$FF
	.dc.b	$8F,$5F,$B9,$F0,$9B,$64,$76,$3F,$D1,$B1,$F1,$3E,$8D,$CF,$7A,$25
	.dc.b	$6E,$72,$DE,$1E,$15,$23,$07,$AB,$97,$EF,$FA,$BF,$81,$E5,$7A,$B4
	.dc.b	$C4,$7D,$EF,$87,$B7,$C6,$93,$98,$E9,$E2,$FF,$11,$7F,$CD,$18,$7F
	.dc.b	$09,$F1,$BF,$CD,$0D,$8C,$CE,$DF,$F5,$61,$FC,$CB,$3F,$CA,$46,$79
	.dc.b	$3F,$22,$DE,$6F,$E7,$33,$F9,$9F,$1B,$F7,$D7,$E2,$F1,$F8,$99,$F0
	.dc.b	$A4,$7C,$19,$DB,$F9,$3C,$91,$8C,$BF,$3E,$BE,$8F,$C7,$F4,$BE,$AD
	.dc.b	$0D,$26,$1F,$02,$9F,$23,$C2,$3E,$2C,$F9,$E3,$DB,$E5,$73,$24,$FC
	.dc.b	$47,$E3,$D3,$E4,$37,$3F,$3A,$7E,$04,$7C,$1F,$8C,$F8,$FF,$4B,$2B
	.dc.b	$DD,$19,$69,$0F,$81,$6F,$95,$F9,$FD,$5E,$D2,$4F,$D6,$7E,$CF,$C3
	.dc.b	$FF,$95,$F7,$FA,$71,$C2,$7F,$19,$39,$E4,$FF,$73,$7D,$17,$C6,$9F
	.dc.b	$CE,$35,$FD,$BB,$6F,$FC,$C5,$F4,$BE,$F9,$FD,$12,$F9,$7F,$E4,$8D
	.dc.b	$8F,$F5,$93,$5F,$F5,$B3,$5F,$C5,$BE,$D7,$9C,$7F,$0E,$8E,$13,$E0
	.dc.b	$CB,$F8,$8F,$D5,$BE,$3A,$93,$0F,$DA,$7D,$2B,$D6,$FE,$2E,$BD,$3F
	.dc.b	$E7,$0D,$9F,$F6,$33,$5F,$F6,$53,$5F,$CF,$AF,$47,$99,$23,$F8,$17
	.dc.b	$C9,$3D,$DF,$E4,$AD,$8C,$C6,$0F,$E5,$3B,$FF,$02,$8F,$CE,$75,$5D
	.dc.b	$BF,$53,$E2,$FE,$B0,$FE,$87,$F4,$0F,$AB,$74,$FF,$32,$F9,$0F,$85
	.dc.b	$8F,$FB,$7D,$AF,$FB,$8D,$AF,$EE,$1A,$FF,$2C,$F7,$FF,$9C,$11,$FE
	.dc.b	$21,$F4,$8F,$97,$FE,$7F,$F4,$7F,$0E,$7F,$A7,$87,$0F,$F5,$26,$B9
	.dc.b	$E3,$FE,$CD,$1F,$CC,$23,$FD,$E8,$D7,$FD,$F0,$D7,$F8,$6B,$5E,$FF
	.dc.b	$24,$F8,$FF,$EB,$D0,$F7,$FF,$D2,$13,$9E,$0F,$EC,$71,$8A,$DB,$63
	.dc.b	$3F,$2B,$FC,$6A,$DF,$0B,$87,$FC,$B2,$75,$8D,$46,$63,$FD,$E8,$64
	.dc.b	$FE,$E3,$C3,$FC,$2B,$BF,$A4,$23,$0F,$76,$BF,$CA,$9A,$FC,$C7,$F7
	.dc.b	$D1,$C9,$B1,$C7,$B7,$FD,$98,$F8,$1A,$BE,$81,$52,$3F,$B7,$C6,$23
	.dc.b	$FA,$67,$C0,$FF,$78,$7F,$AD,$46,$67,$FB,$AF,$D1,$7F,$4E,$9F,$5B
	.dc.b	$63,$E4,$35,$FE,$86,$D7,$FC,$B1,$E7,$F8,$74,$FE,$86,$DA,$F8,$FD
	.dc.b	$5F,$E1,$87,$FA,$9C,$66,$BF,$C3,$78,$7E,$BE,$7C,$BD,$B9,$9F,$77
	.dc.b	$C1,$5A,$D4,$7F,$83,$3E,$2B,$F8,$95,$FF,$75,$6C,$7C,$2B,$E5,$18
	.dc.b	$CB,$FD,$F8,$FC,$E9,$FE,$BA,$D7,$F3,$ED,$7F,$C5,$ED,$5F,$52,$F9
	.dc.b	$6F,$33,$70,$F1,$79,$78,$9F,$57,$97,$FC,$4E,$7F,$C1,$4F,$AB,$1F
	.dc.b	$B1,$87,$F5,$EE,$9D,$E7,$2F,$F1,$B7,$F0,$2B,$FB,$9B,$5F,$DA,$B5
	.dc.b	$FF,$22,$F8,$CF,$95,$F9,$4D,$AF,$0E,$F8,$9F,$A5,$E9,$3C,$C7,$CB
	.dc.b	$63,$C6,$FF,$45,$61,$FD,$2A,$31,$F1,$32,$3F,$C0,$EF,$E5,$AF,$FD
	.dc.b	$28,$DA,$E3,$9F,$F8,$2A,$7B,$7C,$67,$83,$13,$78,$70,$CC,$7F,$C3
	.dc.b	$3E,$AD,$C3,$F9,$4B,$C7,$CD,$F3,$7A,$D6,$BF,$C3,$EE,$1F,$F0,$DB
	.dc.b	$E3,$23,$FE,$22,$C7,$E9,$3F,$EB,$1F,$A0,$F2,$71,$FC,$16,$47,$F8
	.dc.b	$91,$F6,$11,$E5,$F0,$5E,$B3,$D9,$C5,$63,$FB,$8C,$7C,$BF,$5F,$1A
	.dc.b	$7F,$42,$7E,$17,$C0,$7C,$FF,$A0,$B7,$FE,$38,$70,$FF,$8D,$1F,$98
	.dc.b	$47,$FC,$77,$F4,$5F,$ED,$37,$FC,$95,$F3,$3F,$EA,$F7,$FC,$99,$19
	.dc.b	$C3,$FC,$59,$39,$6F,$81,$F8,$28,$FA,$FD,$9F,$E8,$C7,$4D,$E7,$FD
	.dc.b	$A7,$D1,$6B,$1F,$E8,$AF,$F8,$CD,$FF,$33,$39,$C3,$FD,$2E,$E5,$FF
	.dc.b	$41,$B6,$BF,$43,$E7,$BD,$0E,$BC,$DF,$C0,$63,$18,$7F,$9C,$7D,$FF
	.dc.b	$F6,$43,$78,$FF,$83,$46,$32,$FF,$47,$A8,$C9,$FC,$A1,$C4,$FF,$38
	.dc.b	$6C,$FF,$45,$6B,$FE,$60,$F8,$7F,$E4,$59,$7F,$A4,$BE,$13,$BF,$1F
	.dc.b	$F1,$59,$CE,$1F,$0B,$CA,$F4,$76,$7C,$66,$C7,$FA,$27,$4F,$E0,$D9
	.dc.b	$7F,$A9,$1F,$DA,$36,$BE,$07,$81,$FB,$4E,$4F,$E9,$7A,$7D,$3F,$7B
	.dc.b	$1A,$7D,$76,$65,$F8,$07,$2D,$3E,$47,$37,$C6,$8F,$8E,$D8,$F9,$91
	.dc.b	$F1,$48,$FC,$57,$E4,$7B,$BF,$D3,$7C,$E7,$C5,$FF,$81,$38,$BF,$97
	.dc.b	$C6,$36,$FF,$6A,$C7,$F9,$97,$0F,$F6,$BC,$E6,$53,$C2,$BF,$DB,$7E
	.dc.b	$17,$F1,$1F,$9E,$ED,$18,$F0,$11,$8B,$57,$F1,$7F,$73,$0F,$1C,$BB
	.dc.b	$E1,$D3,$C5,$F1,$5F,$F8,$02,$31,$AF,$0C,$BA,$DF,$0A,$EB,$7C,$34
	.dc.b	$F6,$7A,$6F,$A2,$FF,$65,$DE,$E3,$1E,$17,$95,$D7,$87,$6B,$78,$F8
	.dc.b	$4F,$F0,$D3,$8C,$3A,$F8,$1F,$D1,$67,$57,$D3,$1E,$D3,$9A,$F1,$FD
	.dc.b	$17,$FA,$FA,$7C,$4B,$E6,$7C,$4E,$70,$E6,$E6,$77,$3A,$37,$F0,$F6
	.dc.b	$32,$7C,$3E,$16,$DF,$5F,$0F,$F1,$96,$1C,$DF,$64,$EB,$97,$E6,$7D
	.dc.b	$17,$D4,$BF,$F0,$DE,$7F,$85,$AF,$E9,$D3,$9E,$ED,$E5,$DE,$F7,$C9
	.dc.b	$F8,$1A,$D4,$F8,$FE,$87,$FC,$84,$DF,$1B,$F0,$31,$8D,$3C,$6A,$32
	.dc.b	$49,$E5,$C3,$6B,$EA,$1F,$94,$F8,$BF,$C7,$36,$7F,$08,$FE,$63,$D2
	.dc.b	$CF,$5A,$A9,$FE,$31,$AF,$F7,$3C,$65,$FF,$10,$57,$E9,$E7,$4D,$8C
	.dc.b	$BF,$A8,$8F,$0F,$C7,$7F,$8D,$78,$3F,$51,$7C,$EB,$CF,$39,$D7,$F5
	.dc.b	$3A,$7E,$D3,$1F,$8D,$1F,$13,$E4,$7E,$2B,$FA,$6F,$A2,$FE,$23,$3F
	.dc.b	$E7,$CE,$24,$63,$1F,$D9,$B5,$FD,$7B,$61,$83,$E4,$FE,$89,$F4,$BE
	.dc.b	$7E,$E1,$F5,$FE,$7F,$EA,$9F,$5B,$E0,$6C,$7D,$BF,$9F,$F8,$79,$FE
	.dc.b	$F3,$14,$63,$2E,$ED,$8C,$D7,$C8,$9F,$1F,$57,$E0,$8F,$C1,$F9,$96
	.dc.b	$7F,$35,$AB,$D2,$F8,$47,$E2,$3B,$6C,$EB,$C4,$F0,$51,$A4,$8F,$06
	.dc.b	$4F,$9F,$FC,$89,$F3,$7C,$6B,$3F,$97,$F9,$FF,$CA,$FC,$FF,$E5,$9B
	.dc.b	$1F,$9B,$8F,$61,$1E,$AD,$E0,$FC,$DF,$B7,$C1,$B3,$FB,$5F,$95,$7C
	.dc.b	$37,$F2,$26,$C6,$7C,$0C,$4C,$8F,$91,$3A,$AF,$F2,$8F,$1A,$31,$3F
	.dc.b	$85,$6C,$FF,$36,$78,$76,$FE,$ED,$FC,$84,$7F,$02,$73,$D5,$B3,$F6
	.dc.b	$BE,$94,$BF,$DD,$7C,$C5,$BF,$DB,$CF,$DD,$24,$FC,$E6,$C7,$EF,$3E
	.dc.b	$3F,$86,$3F,$C2,$F8,$BF,$DE,$57,$FB,$6E,$FF,$F2,$D3,$7C,$DF,$D5
	.dc.b	$FD,$17,$EB,$19,$1F,$B0,$C5,$5C,$1F,$F3,$99,$89,$FC,$AB,$67,$84
	.dc.b	$EF,$0F,$EB,$8F,$F8,$67,$E7,$BC,$13,$E3,$F6,$91,$8B,$FD,$CC,$61
	.dc.b	$FD,$DE,$1F,$C0,$23,$18,$74,$6E,$DF,$EF,$2C,$7B,$5F,$0C,$7B,$63
	.dc.b	$23,$AB,$DF,$0E,$4E,$5B,$1C,$4D,$91,$C9,$EC,$C7,$C9,$3C,$70,$F1
	.dc.b	$63,$E4,$B1,$F1,$4E,$51,$A9,$E1,$1A,$C7,$E4,$4E,$63,$66,$6B,$A3
	.dc.b	$CF,$0F,$1C,$ED,$B1,$CF,$0F,$2B,$C8,$FB,$75,$CA,$FB,$C7,$5C,$3C
	.dc.b	$6E,$9F,$8D,$87,$53,$C9,$1E,$C3,$81,$4A,$59,$FA,$6E,$7E,$2F,$F8
	.dc.b	$CF,$D1,$6E,$3E,$2C,$61,$C3,$5E,$1E,$02,$78,$47,$E5,$3E,$0D,$E5
	.dc.b	$63,$0F,$F5,$AB,$63,$33,$DE,$72,$D8,$CE,$BF,$96,$6E,$5F,$8B,$6C
	.dc.b	$FD,$5F,$6B,$E0,$56,$F0,$FD,$65,$6A,$BC,$91,$FA,$FA,$DD,$F3,$8D
	.dc.b	$92,$7F,$AB,$CE,$9F,$C1,$57,$3C,$3C,$F8,$77,$9D,$9C,$5C,$13,$0F
	.dc.b	$D1,$5F,$39,$F1,$9F,$D9,$2F,$A1,$D6,$35,$7D,$AB,$CD,$5B,$C3,$D4
	.dc.b	$F7,$8C,$E1,$DE,$3E,$21,$93,$D3,$35,$FE,$13,$7D,$B0,$F0,$CE,$70
	.dc.b	$4E,$63,$13,$E5,$AF,$09,$E3,$25,$FF,$0C,$AF,$49,$C6,$33,$87,$C2
	.dc.b	$78,$EF,$F9,$39,$FC,$C3,$63,$F8,$BC,$61,$FF,$37,$FC,$3B,$C7,$FA
	.dc.b	$9D,$08,$78,$F8,$6D,$7F,$90,$DE,$0D,$AD,$4C,$8C,$10,$FF,$22,$CF
	.dc.b	$06,$CF,$5F,$1F,$C3,$F4,$3F,$9A,$9F,$F2,$5B,$FD,$71,$FF,$29,$35
	.dc.b	$CE,$DC,$0E,$3D,$92,$37,$19,$69,$0E,$BD,$BE,$DF,$A2,$91,$ED,$7E
	.dc.b	$07,$24,$8C,$10,$F5,$B9,$69,$F1,$5F,$91,$E4,$FC,$9F,$26,$4F,$C9
	.dc.b	$7D,$FA,$7C,$0D,$AC,$FB,$A3,$0F,$C0,$BF,$8A,$47,$E9,$79,$1F,$57
	.dc.b	$C6,$91,$F5,$5F,$B1,$E2,$FD,$9F,$17,$EE,$FA,$0F,$3E,$99,$BF,$B9
	.dc.b	$E2,$7D,$F2,$3F,$7F,$C5,$FC,$17,$8B,$23,$F0,$4F,$E2,$3E,$1F,$F9
	.dc.b	$27,$6F,$F9,$F7,$9D,$87,$E3,$39,$7C,$18,$7F,$A2,$DA,$FD,$E8,$E5
	.dc.b	$E1,$48,$FF,$48,$61,$FD,$95,$FF,$46,$AF,$8D,$FD,$DF,$C4,$49,$FF
	.dc.b	$00,$3F,$BE,$FA,$29,$18,$21,$FE,$68,$72,$FF,$9A,$A3,$93,$93,$C3
	.dc.b	$F3,$9F,$23,$D4,$91,$82,$1E,$48,$F8,$47,$C3,$F5,$3E,$7C,$7E,$B3
	.dc.b	$F0,$D7,$89,$4C,$C7,$FA,$7C,$A7,$FA,$AB,$93,$E5,$F8,$BF,$EE,$4D
	.dc.b	$5F,$39,$F6,$B7,$97,$FE,$A7,$7F,$96,$78,$2F,$6D,$EE,$E1,$B5,$B2
	.dc.b	$1F,$EC,$E7,$2D,$8E,$04,$7F,$DA,$3F,$15,$DB,$E3,$5F,$19,$49,$3F
	.dc.b	$EE,$3F,$A3,$7C,$6F,$91,$B1,$FE,$E7,$AE,$2D,$AD,$57,$E5,$B9,$FE
	.dc.b	$47,$A9,$2B,$FD,$89,$19,$69,$0F,$C2,$FC,$EF,$F3,$EC,$24,$7F,$90
	.dc.b	$DF,$E8,$39,$48,$FF,$1C,$BF,$C8,$DF,$F0,$A3,$5F,$F2,$1B,$7D,$13
	.dc.b	$C6,$3F,$D6,$2F,$F4,$CD,$3F,$94,$FC,$9B,$2F,$F0,$47,$C4,$FE,$02
	.dc.b	$DE,$8B,$A1,$EB,$EB,$A7,$FC,$53,$39,$8C,$57,$0C,$7A,$3F,$E2,$EC
	.dc.b	$3A,$57,$F9,$DF,$DF,$DA,$AD,$D2,$B7,$18,$78,$FB,$5F,$EB,$99,$E9
	.dc.b	$8F,$F8,$7F,$E8,$7F,$97,$57,$A6,$51,$89,$D7,$A7,$E6,$9C,$F8,$96
	.dc.b	$BF,$DE,$11,$D0,$FF,$22,$B8,$7C,$95,$DE,$78,$7C,$0F,$3F,$8D,$FE
	.dc.b	$13,$5C,$43,$87,$89,$FE,$50,$71,$8F,$9F,$07,$AB,$CF,$CB,$FF,$15
	.dc.b	$DF,$26,$D9,$A6,$B9,$5F,$11,$C6,$BC,$55,$B7,$C9,$3A,$7C,$DF,$40
	.dc.b	$F9,$AF,$88,$D8,$CC,$63,$AA,$47,$06,$BE,$C1,$2F,$9C,$FB,$BC,$7F
	.dc.b	$E0,$A9,$F9,$9C,$9F,$7F,$DC,$C9,$F8,$2E,$57,$B3,$49,$5C,$23,$2D
	.dc.b	$F6,$BF,$13,$D4,$F8,$0F,$F2,$8E,$3F,$97,$6C,$66,$BF,$CB,$93,$96
	.dc.b	$C3,$DE,$E7,$52,$7C,$F3,$97,$D3,$E5,$BC,$DF,$0E,$92,$73,$1F,$9C
	.dc.b	$9D,$46,$08,$7E,$0B,$17,$97,$FC,$32,$31,$83,$E4,$3D,$24,$8F,$3B
	.dc.b	$EC,$F4,$3E,$0B,$23,$CF,$1F,$13,$D1,$F8,$24,$8F,$7B,$F1,$3E,$53
	.dc.b	$51,$EE,$8F,$97,$1E,$9E,$6F,$9B,$E3,$CA,$F8,$31,$96,$90,$F6,$65
	.dc.b	$FD,$9B,$BB,$F7,$F9,$FF,$12,$9F,$9F,$E3,$FD,$42,$3F,$4B,$C7,$FB
	.dc.b	$71,$23,$E8,$C7,$ED,$FE,$A7,$FD,$35,$7F,$BD,$6F,$89,$FE,$9F,$6D
	.dc.b	$7F,$A4,$3F,$0B,$FE,$11,$3F,$EA,$3F,$99,$7D,$17,$EB,$1B,$5F,$E9
	.dc.b	$86,$BF,$EA,$16,$BF,$EA,$9F,$A2,$7C,$9F,$F0,$99,$DE,$5F,$C3,$DB
	.dc.b	$87,$FA,$C6,$B7,$7F,$C0,$FE,$33,$F6,$4F,$E9,$DB,$7F,$C5,$DB,$5F
	.dc.b	$EB,$9F,$AE,$79,$5F,$C5,$FE,$8D,$F1,$BF,$C1,$1B,$8F,$FA,$C1,$AF
	.dc.b	$FA,$E5,$AF,$FB,$07,$E8,$DC,$DF,$C3,$1B,$1F,$EB,$86,$BF,$EC,$16
	.dc.b	$BC,$78,$FF,$94,$47,$DB,$8F,$E0,$0D,$D3,$D5,$F0,$3F,$B4,$FA,$69
	.dc.b	$5F,$AA,$7F,$6F,$E6,$7F,$31,$FA,$AE,$B3,$FE,$D7,$6B,$FE,$F4,$6A
	.dc.b	$6F,$87,$FC,$A3,$83,$FD,$BC,$D9,$FF,$7D,$B5,$FE,$83,$F0,$3F,$D5
	.dc.b	$3E,$87,$FC,$7B,$AF,$FB,$59,$B3,$FE,$DF,$6B,$FE,$E7,$6B,$FE,$E9
	.dc.b	$FA,$2F,$F0,$7B,$7A,$FF,$9A,$C6,$23,$FD,$E0,$D7,$FD,$D3,$F1,$DF
	.dc.b	$D6,$DF,$F7,$1F,$AD,$FE,$F5,$F9,$14,$E7,$E0,$3F,$D0,$D1,$8E,$25
	.dc.b	$F0,$39,$B8,$9F,$F4,$64,$6B,$E1,$F1,$87,$FB,$E5,$B1,$FE,$5E,$39
	.dc.b	$1F,$EE,$62,$E7,$F9,$CD,$FE,$27,$6F,$F6,$99,$F0,$89,$D3,$F1,$E3
	.dc.b	$4E,$7D,$87,$C5,$E2,$BF,$B8,$B6,$33,$7F,$3D,$B5,$D8,$F7,$FA,$4F
	.dc.b	$43,$FC,$25,$AF,$F7,$B8,$C5,$FF,$6D,$FA,$2F,$F2,$AB,$FD,$BE,$33
	.dc.b	$7F,$D4,$FE,$A3,$CE,$F5,$6B,$EC,$6B,$F3,$9B,$E7,$7F,$87,$FE,$9F
	.dc.b	$C3,$7F,$E1,$EF,$25,$18,$BF,$F1,$2C,$3E,$63,$17,$C6,$FF,$C0,$5F
	.dc.b	$28,$F9,$BE,$18,$7F,$B5,$DB,$A7,$F8,$DA,$B7,$1A,$A7,$CD,$FC,$EB
	.dc.b	$FF,$6F,$36,$33,$5F,$26,$7F,$0F,$18,$71,$F4,$9F,$EF,$A9,$D4,$23
	.dc.b	$0E,$BD,$7E,$D1,$88,$FA,$9F,$A4,$FF,$91,$6B,$FC,$38,$D8,$FF,$00
	.dc.b	$79,$1F,$DC,$67,$3F,$0D,$EE,$F4,$BF,$CF,$71,$FE,$23,$9F,$EB,$8F
	.dc.b	$F8,$96,$33,$5C,$BD,$7F,$8B,$5A,$8F,$63,$5F,$C3,$35,$FC,$AB,$5F
	.dc.b	$B5,$EF,$FC,$EE,$43,$F5,$FB,$9E,$9F,$F8,$9B,$2F,$F1,$B5,$69,$FE
	.dc.b	$FF,$E8,$7D,$AE,$67,$CD,$D8,$C3,$27,$FC,$12,$FE,$8B,$9B,$FC,$CC
	.dc.b	$DA,$E3,$3F,$E1,$0F,$69,$F4,$9F,$86,$6C,$7F,$89,$3E,$21,$EE,$66
	.dc.b	$A4,$FF,$84,$BE,$9A,$F0,$7F,$81,$9F,$37,$CF,$FE,$43,$E4,$DC,$38
	.dc.b	$BF,$F1,$7C,$61,$B1,$E7,$8C,$5F,$EC,$76,$46,$23,$F8,$0C,$62,$78
	.dc.b	$79,$1D,$FE,$2B,$C0,$DC,$3B,$C7,$52,$3F,$E3,$0C,$11,$F1,$F6,$7D
	.dc.b	$1F,$C3,$3F,$53,$BF,$D5,$CF,$FC,$4D,$3A,$C3,$FC,$63,$5E,$FF,$94
	.dc.b	$E1,$1E,$E7,$E1,$7B,$5F,$0F,$E3,$6C,$FC,$27,$0F,$8D,$F8,$44,$7F
	.dc.b	$C8,$B9,$7F,$B4,$DF,$F2,$BB,$71,$FF,$1C,$3F,$EC,$68,$CF,$C9,$7F
	.dc.b	$00,$F1,$FE,$FD,$7F,$8F,$5B,$E4,$7F,$A8,$2F,$93,$FE,$9C,$E7,$FE
	.dc.b	$07,$EA,$7E,$A6,$FF,$CD,$4D,$9F,$F4,$D7,$D3,$7E,$1A,$73,$EB,$70
	.dc.b	$E9,$9A,$7F,$A3,$F0,$E0,$FF,$38,$7F,$16,$FF,$A4,$F8,$FE,$55,$FF
	.dc.b	$9D,$FE,$37,$FC,$DF,$39,$E2,$FE,$2B,$18,$9D,$7C,$57,$F1,$18,$C4
	.dc.b	$7F,$37,$6B,$FD,$19,$AF,$FA,$77,$E6,$F9,$9E,$EF,$88,$B8,$79,$FC
	.dc.b	$4F,$82,$7C,$2F,$26,$C7,$AF,$E2,$3E,$29,$8C,$3E,$2B,$FC,$BB,$8F
	.dc.b	$FA,$9F,$17,$FB,$33,$CB,$76,$67,$D5,$FE,$67,$18,$C7,$FD,$22,$DC
	.dc.b	$7F,$CE,$5C,$2F,$9D,$DE,$1F,$AB,$7F,$9B,$C7,$F3,$9B,$FE,$DD,$D9
	.dc.b	$FD,$92,$31,$6F,$5D,$39,$CB,$FD,$D5,$18,$9F,$EF,$75,$C2,$31,$1F
	.dc.b	$B0,$F4,$3F,$DE,$18,$7F,$63,$AE,$AD,$8C,$CF,$85,$AF,$4F,$43,$BE
	.dc.b	$3E,$56,$C7,$82,$6B,$FB,$E8,$CF,$86,$F0,$7F,$BE,$E5,$E7,$8C,$67
	.dc.b	$AF,$0B,$F3,$B3,$9A,$DE,$5F,$E1,$6B,$DE,$7F,$1B,$24,$7F,$0C,$F9
	.dc.b	$EE,$F7,$C2,$73,$AA,$7C,$FE,$23,$E8,$FF,$81,$67,$F2,$61,$E1,$FF
	.dc.b	$96,$B3,$ED,$1B,$F8,$0E,$33,$9C,$BE,$A6,$BF,$E2,$78,$C6,$1C,$DB
	.dc.b	$1F,$E2,$BC,$7B,$BB,$F8,$4F,$1F,$D1,$BB,$69,$8F,$66,$EB,$E2,$9D
	.dc.b	$3F,$07,$87,$E6,$72,$EB,$5F,$19,$F1,$38,$BF,$05,$9F,$C9,$6C,$66
	.dc.b	$78,$6D,$F3,$DF,$93,$CB,$31,$ED,$C8,$F2,$B1,$87,$6C,$FE,$E5,$FD
	.dc.b	$EE,$5F,$F2,$7E,$1E,$5C,$FF,$C0,$D9,$FD,$F9,$CE,$1C,$7E,$8B,$F9
	.dc.b	$BD,$FA,$DB,$19,$E5,$FC,$4C,$62,$FF,$10,$DD,$BF,$09,$83,$E8,$77
	.dc.b	$DB,$E3,$7F,$27,$0F,$9D,$FC,$AF,$CA,$FF,$8F,$33,$F8,$E7,$F5,$48
	.dc.b	$CE,$9C,$1A,$FE,$3E,$34,$D9,$E5,$9F,$EB,$9B,$97,$F9,$71,$B5,$FB
	.dc.b	$4B,$E6,$FC,$19,$F6,$9E,$EF,$3A,$E1,$F7,$7E,$8F,$FC,$53,$8F,$E9
	.dc.b	$FC,$0F,$F5,$BC,$E7,$C3,$FD,$6C,$63,$1F,$99,$E7,$FC,$83,$E4,$F8
	.dc.b	$36,$FE,$77,$81,$F3,$0F,$9B,$F0,$76,$3E,$97,$0F,$DB,$F9,$8F,$A2
	.dc.b	$D8,$FA,$DD,$BF,$50,$FA,$BE,$55,$8F,$47,$BF,$FC,$03,$83,$FC,$B1
	.dc.b	$1F,$77,$C0,$C4,$E7,$D6,$D7,$A5,$6B,$E0,$3F,$80,$F8,$29,$18,$3F
	.dc.b	$0C,$FE,$13,$E8,$64,$7E,$27,$E8,$AF,$1F,$E3,$67,$F1,$3E,$EE,$39
	.dc.b	$FF,$12,$E2,$FF,$20,$E7,$A7,$F8,$77,$77,$FB,$8A,$7F,$5D,$F2,$FF
	.dc.b	$CE,$EF,$F8,$C3,$6B,$F8,$A3,$F9,$3C,$7F,$89,$B9,$6F,$9B,$FC,$8A
	.dc.b	$3F,$9C,$DF,$48,$C5,$65,$B1,$93,$C3,$1F,$A2,$F5,$24,$7E,$9F,$E6
	.dc.b	$AE,$1F,$AB,$9F,$D3,$FC,$76,$32,$FD,$3B,$70,$F6,$B6,$BF,$94,$E6
	.dc.b	$8F,$D2,$F9,$3F,$D0,$25,$E1,$23,$0D,$CF,$FD,$92,$7F,$AB,$5C,$4F
	.dc.b	$FA,$F5,$B8,$FE,$97,$D8,$7C,$FF,$EA,$70,$FD,$9C,$62,$FF,$67,$7C
	.dc.b	$33,$FD,$54,$E7,$0F,$D7,$5F,$0C,$3F,$65,$39,$AD,$C7,$5C,$72,$CC
	.dc.b	$7A,$D6,$DF,$F7,$0E,$3D,$DB,$19,$7E,$0B,$60,$DC,$3C,$53,$D2,$BB
	.dc.b	$63,$E6,$C2,$E3,$E4,$AD,$1E,$EB,$E3,$5F,$16,$F7,$3E,$49,$CD,$71
	.dc.b	$AD,$4F,$1B,$DB,$E6,$BB,$8F,$AE,$30,$78,$2B,$89,$E5,$71,$7D,$48
	.dc.b	$F8,$6F,$9E,$3F,$04,$F8,$0E,$30,$FF,$5F,$E1,$F1,$1B,$3F,$6E,$BF
	.dc.b	$08,$D7,$FD,$FF,$F1,$5F,$CD,$FE,$8F,$FD,$8E,$D8,$FF,$67,$35,$FF
	.dc.b	$7C,$35,$FE,$59,$F3,$5F,$E9,$27,$FD,$A9,$82,$31,$1F,$93,$6B,$C3
	.dc.b	$CA,$F3,$C7,$1D,$3F,$9C,$37,$4D,$57,$E4,$BC,$63,$FC,$FE,$74,$FE
	.dc.b	$21,$AF,$E9,$9B,$82,$78,$CF,$E7,$27,$24,$C3,$FA,$54,$EB,$0F,$1C
	.dc.b	$77,$25,$78,$CB,$1A,$7A,$47,$27,$38,$7A,$A7,$47,$F5,$5A,$ED,$19
	.dc.b	$24,$FA,$EB,$93,$D0,$95,$CA,$7F,$AE,$57,$29,$E0,$6A,$39,$BF,$C2
	.dc.b	$23,$4F,$C5,$AF,$2B,$C2,$BC,$8E,$0D,$1C,$CF,$DA,$CF,$2C,$3E,$64
	.dc.b	$6D,$FC,$05,$7A,$2B,$9C,$F1,$39,$B8,$25,$F9,$AB,$8C,$FC,$D3,$FC
	.dc.b	$39,$E3,$7F,$9D,$DF,$F1,$73,$FE,$49,$6F,$89,$F0,$7C,$D4,$EE,$78
	.dc.b	$F8,$D9,$8E,$11,$D2,$B1,$32,$3A,$B5,$F0,$F8,$DD,$7E,$8B,$F3,$79
	.dc.b	$FF,$99,$DF,$F4,$8C,$FF,$90,$5A,$FF,$8C,$DD,$E7,$FE,$67,$6C,$7F
	.dc.b	$99,$9D,$CF,$3E,$9E,$DA,$CA,$E8,$D7,$A7,$99,$EB,$F2,$24,$7C,$47
	.dc.b	$DD,$E4,$48,$C1,$0F,$96,$E5,$A7,$CC,$7E,$37,$91,$F1,$FC,$8F,$D3
	.dc.b	$F4,$FE,$A3,$8B,$E6,$C6,$2F,$FD,$05,$C4,$8C,$47,$FA,$0A,$30,$FF
	.dc.b	$54,$3E,$17,$17,$CB,$E2,$91,$F2,$63,$E7,$65,$CB,$07,$07,$F9,$2E
	.dc.b	$B9,$1F,$26,$75,$19,$86,$5E,$7C,$5E,$1F,$E0,$49,$3F,$8A,$F0,$FF
	.dc.b	$09,$E1,$FE,$27,$CD,$7D,$1D,$33,$5F,$90,$F0,$BF,$14,$47,$F2,$7E
	.dc.b	$1F,$E3,$7C,$29,$1F,$96,$7F,$29,$F0,$FF,$C4,$23,$F0,$95,$F9,$7E
	.dc.b	$6E,$F1,$FE,$89,$6B,$A3,$FD,$19,$3F,$CE,$7A,$7F,$D6,$CD,$EA,$FF
	.dc.b	$59,$36,$7F,$D6,$2B,$E7,$FF,$6D,$F0,$1F,$C0,$3E,$3B,$F2,$34,$6F
	.dc.b	$7A,$77,$3F,$38,$FF,$07,$BF,$71,$A6,$63,$83,$63,$8B,$FE,$08,$7F
	.dc.b	$83,$7B,$FF,$37,$D2,$93,$FE,$93,$7E,$2F,$CE,$E2,$3F,$D3,$5C,$5F
	.dc.b	$E8,$57,$47,$C2,$8F,$DF,$7C,$3F,$DE,$8F,$F5,$83,$5F,$F5,$13,$76
	.dc.b	$FF,$6C,$47,$DD,$38,$F4,$47,$FD,$48,$7E,$07,$C9,$FE,$15,$EA,$BC
	.dc.b	$1A,$66,$7F,$D5,$53,$F9,$5C,$BF,$B6,$5F,$E9,$5B,$3F,$EB,$D9,$D1
	.dc.b	$C5,$F6,$BF,$CD,$FC,$3F,$F6,$EF,$C9,$78,$3C,$39,$5F,$EC,$BC,$BF
	.dc.b	$52,$FF,$B2,$A9,$ED,$BE,$87,$93,$67,$85,$F2,$78,$46,$5F,$8B,$F1
	.dc.b	$BE,$EF,$5A,$47,$CA,$8F,$9A,$47,$07,$FB,$27,$9F,$F1,$5E,$24,$8F
	.dc.b	$F5,$97,$47,$EC,$78,$BF,$D6,$18,$FF,$45,$3F,$11,$D1,$E8,$AF,$C4
	.dc.b	$46,$08,$FF,$AE,$5E,$0F,$FA,$EF,$E5,$3E,$FF,$A3,$FD,$E3,$E0,$FF
	.dc.b	$01,$E4,$38,$BF,$7B,$7F,$E0,$49,$CC,$62,$11,$C3,$A3,$FC,$D1,$1B
	.dc.b	$E7,$FF,$05,$DA,$73,$3C,$27,$31,$8A,$E5,$87,$8E,$78,$FC,$4B,$87
	.dc.b	$18,$7F,$85,$BE,$83,$F6,$4F,$71,$97,$A7,$C7,$DB,$63,$42,$78,$BE
	.dc.b	$33,$FC,$43,$C3,$FD,$3B,$3F,$D0,$4E,$9B,$C7,$E2,$AF,$DF,$79,$DD
	.dc.b	$B7,$79,$14,$93,$E5,$CB,$E6,$7A,$FC,$7E,$21,$E3,$FF,$19,$AF,$F1
	.dc.b	$4C,$61,$C0,$70,$ED,$AF,$F2,$1D,$5E,$87,$5A,$93,$F0,$27,$2F,$8F
	.dc.b	$2E,$7A,$FD,$CF,$2B,$D4,$7A,$EF,$DD,$BB,$E7,$F1,$87,$19,$CC,$7C
	.dc.b	$1F,$2F,$E7,$F5,$7F,$33,$C7,$F1,$6D,$E6,$FC,$1E,$7F,$EB,$1C,$7E
	.dc.b	$27,$37,$F8,$79,$FA,$5C,$3F,$91,$C7,$A4,$6B,$D3,$EB,$E1,$FC,$4A
	.dc.b	$FF,$5B,$B7,$2F,$F3,$85,$7E,$41,$AF,$E1,$1F,$B5,$96,$67,$9C,$FD
	.dc.b	$59,$CB,$F8,$AE,$AF,$26,$9E,$AD,$3D,$A7,$E3,$23,$38,$7E,$6B,$2C
	.dc.b	$12,$FF,$01,$E2,$3C,$6F,$8A,$79,$C9,$1F,$81,$D3,$FD,$A3,$1E,$BF
	.dc.b	$9E,$DC,$F9,$63,$E6,$C7,$CE,$FA,$3F,$80,$49,$E5,$AA,$3E,$27,$0E
	.dc.b	$E7,$DD,$1F,$5A,$3D,$5E,$07,$D8,$F9,$29,$5F,$97,$DB,$F8,$63,$F5
	.dc.b	$1B,$C1,$FB,$FA,$E3,$3E,$EB,$FC,$0F,$D0,$FD,$32,$47,$91,$FA,$DF
	.dc.b	$45,$A8,$FA,$11,$F8,$88,$FC,$4F,$C5,$49,$FA,$8F,$E3,$3E,$4B,$F6
	.dc.b	$1F,$45,$FA,$49,$FF,$4B,$B5,$34,$D7,$98,$F9,$5F,$E1,$53,$FC,$0A
	.dc.b	$7F,$28,$F2,$3F,$86,$CE,$FC,$6F,$CD,$F9,$B2,$7F,$0C,$FE,$7B,$E9
	.dc.b	$BF,$81,$7C,$2B,$C1,$F2,$36,$BF,$72,$D7,$F7,$CD,$7F,$C3,$1F,$40
	.dc.b	$F1,$BF,$86,$7D,$0B,$E2,$E9,$FC,$3E,$75,$5F,$C6,$BE,$87,$EA,$BF
	.dc.b	$E2,$16,$BF,$EB,$26,$BF,$EB,$C6,$F6,$BA,$DF,$2D,$FB,$82,$57,$E8
	.dc.b	$FE,$97,$F9,$4F,$80,$E8,$FD,$83,$67,$F3,$0D,$7F,$C6,$58,$23,$11
	.dc.b	$FE,$35,$F8,$8F,$4B,$FC,$7D,$AE,$76,$FB,$91,$88,$FF,$69,$FC,$FB
	.dc.b	$F2,$D7,$CC,$7F,$24,$8C,$36,$33,$3F,$CA,$E7,$2F,$F2,$AF,$C1,$7F
	.dc.b	$9F,$BF,$E4,$76,$BF,$ED,$86,$BF,$EE,$26,$AF,$8B,$7C,$57,$F6,$08
	.dc.b	$75,$7F,$A5,$27,$3E,$17,$FB,$AA,$31,$5B,$6C,$67,$D4,$FE,$81,$EC
	.dc.b	$DC,$3F,$CB,$4D,$7F,$DC,$C7,$68,$DB,$4F,$F7,$BF,$DC,$FF,$BC,$7C
	.dc.b	$CF,$E8,$ED,$8F,$E9,$8D,$7F,$CC,$CD,$7E,$6F,$E5,$7F,$DF,$E7,$E4
	.dc.b	$36,$3F,$CD,$9A,$A3,$11,$FE,$FE,$8C,$79,$DF,$E4,$58,$FF,$80,$3E
	.dc.b	$8F,$FA,$5E,$FF,$AC,$B6,$3F,$AF,$35,$FF,$3C,$35,$FF,$03,$F2,$3C
	.dc.b	$7E,$91,$FD,$F2,$31,$5F,$90,$9C,$F3,$7E,$3D,$BD,$4F,$9B,$FF,$2D
	.dc.b	$70,$7F,$52,$6D,$FF,$45,$F9,$8B,$97,$F4,$59,$AF,$FB,$7B,$E5,$FC
	.dc.b	$2D,$AF,$EE,$7F,$18,$9C,$B6,$3F,$BB,$FD,$0F,$D1,$C7,$FC,$D1,$D8
	.dc.b	$EC,$EB,$39,$A6,$32,$B5,$1F,$E1,$6F,$A1,$FF,$10,$3F,$E0,$26,$CF
	.dc.b	$F8,$1D,$AF,$FA,$85,$AF,$F6,$1C,$BD,$BB,$FE,$FD,$BB,$7F,$AC,$AF
	.dc.b	$FB,$04,$7C,$2E,$8F,$E4,$DF,$3B,$F0,$BE,$85,$E9,$E7,$CE,$4E,$62
	.dc.b	$61,$FE,$16,$AD,$B6,$3E,$4C,$63,$3F,$E0,$B9,$3E,$27,$C6,$79,$BF
	.dc.b	$0E,$FF,$66,$E1,$FD,$CE,$31,$E7,$E7,$E1,$2C,$FF,$8A,$E9,$19,$8F
	.dc.b	$ED,$51,$86,$C7,$E1,$A3,$18,$7F,$16,$F5,$9F,$45,$F1,$9B,$7F,$DE
	.dc.b	$3D,$4C,$4C,$8F,$07,$35,$BF,$F0,$BE,$1F,$61,$BF,$2D,$EB,$F7,$3D
	.dc.b	$8D,$BF,$44,$78,$C8,$FA,$1F,$37,$D0,$FC,$3F,$8F,$F8,$8D,$9F,$81
	.dc.b	$1E,$B2,$3F,$01,$FD,$1F,$84,$F9,$1F,$C0,$B6,$7F,$C2,$7E,$86,$26
	.dc.b	$47,$CB,$F9,$0B,$9F,$B1,$FA,$9E,$7F,$D5,$F7,$6D,$7D,$17,$0F,$D0
	.dc.b	$7E,$71,$1F,$F0,$E7,$8E,$F9,$BF,$BD,$F1,$9F,$94,$8C,$4F,$DB,$8C
	.dc.b	$BF,$77,$E4,$2E,$5F,$39,$FD,$AF,$C3,$7F,$BA,$E7,$F0,$5F,$4F,$8E
	.dc.b	$0F,$C5,$32,$3F,$0D,$DF,$70,$FA,$EF,$2D,$BF,$7F,$F1,$6F,$90,$C5
	.dc.b	$FF,$A1,$A3,$0D,$8F,$E0,$71,$8B,$FF,$39,$46,$0F,$C7,$7D,$2F,$E6
	.dc.b	$CF,$CE,$7C,$F5,$BF,$F4,$8E,$9F,$D6,$9B,$3E,$CE,$2F,$6B,$23,$F3
	.dc.b	$DF,$01,$71,$FC,$7B,$F7,$3E,$6E,$4F,$FA,$37,$CE,$7A,$D8,$BF,$F3
	.dc.b	$5F,$4A,$FF,$5D,$CD,$9F,$3F,$FD,$7F,$DD,$F4,$AF,$FD,$44,$7F,$B0
	.dc.b	$7C,$B6,$7F,$C5,$BD,$7F,$C6,$4F,$F1,$8C,$BF,$90,$28,$CB,$FE,$C9
	.dc.b	$F4,$B8,$B9,$ED,$FE,$3E,$F5,$BE,$91,$8B,$E7,$1A,$9D,$1F,$D3,$63
	.dc.b	$2E,$F0,$EB,$3B,$E2,$75,$78,$32,$F1,$C6,$27,$BD,$FA,$6F,$58,$F9
	.dc.b	$2B,$53,$E7,$3F,$D9,$78,$BE,$6F,$71,$86,$F9,$FF,$E1,$FF,$4E,$D6
	.dc.b	$F2,$EF,$3F,$EE,$BA,$DE,$DF,$C3,$B0,$ED,$39,$AD,$CE,$74,$F9,$FE
	.dc.b	$83,$E1,$DE,$17,$2F,$A1,$F1,$CE,$63,$18,$72,$9B,$97,$AE,$B5,$39
	.dc.b	$3C,$BE,$BB,$8F,$D6,$D8,$F8,$F2,$FA,$2F,$D0,$67,$F8,$2D,$7F,$C7
	.dc.b	$D8,$70,$3E,$9E,$3F,$DE,$9A,$FF,$79,$6C,$70,$C3,$FB,$09,$F8,$47
	.dc.b	$13,$F1,$3E,$FF,$FC,$5B,$3E,$6A,$F9,$F1,$89,$F9,$39,$7E,$4B,$0F
	.dc.b	$27,$07,$F8,$DB,$0F,$6F,$27,$D4,$F7,$64,$7D,$49,$D6,$1C,$AB,$2F
	.dc.b	$D4,$87,$CB,$FD,$C3,$EF,$FD,$03,$E7,$31,$87,$39,$B7,$F7,$2B,$53
	.dc.b	$93,$F2,$31,$FE,$71,$7F,$16,$D7,$37,$FD,$95,$DF,$AD,$FD,$92,$7F
	.dc.b	$28,$6C,$FF,$18,$39,$8F,$D4,$36,$3F,$5E,$D7,$FC,$1E,$F3,$CF,$F4
	.dc.b	$BE,$47,$8A,$31,$D5,$FA,$38,$F7,$B2,$FF,$37,$8F,$FA,$55,$B1,$FA
	.dc.b	$2F,$A2,$FF,$28,$61,$F8,$5F,$07,$59,$7E,$7A,$73,$5F,$E2,$0E,$35
	.dc.b	$6A,$FF,$19,$8B,$E8,$FE,$79,$F4,$3C,$1B,$9F,$8B,$B3,$E9,$9F,$53
	.dc.b	$8E,$C7,$9B,$87,$F6,$7F,$31,$F6,$1B,$1E,$AF,$9D,$FB,$67,$DC,$F9
	.dc.b	$DB,$1F,$B8,$E0,$79,$3F,$98,$6C,$7E,$63,$8B,$D5,$18,$7F,$8F,$8E
	.dc.b	$14,$66,$7F,$3C,$BC,$7F,$91,$E3,$7C,$96,$CF,$CE,$FA,$0F,$D3,$F1
	.dc.b	$1A,$48,$F1,$63,$DE,$5F,$39,$F9,$63,$F2,$FF,$3D,$6B,$EC,$C7,$E6
	.dc.b	$BE,$0F,$17,$F9,$77,$13,$F7,$9C,$F3,$7D,$47,$F8,$FF,$D1,$FF,$4E
	.dc.b	$9F,$E3,$AD,$AF,$E3,$6F,$E8,$1B,$E6,$70,$38,$F5,$5A,$E2,$D9,$E2
	.dc.b	$DC,$3F,$87,$5F,$F8,$66,$7F,$59,$E5,$7E,$C8,$FD,$9F,$C9,$DC,$3F
	.dc.b	$2B,$1F,$B6,$F9,$CC,$57,$EC,$DB,$87,$E3,$9B,$8F,$E4,$F6,$47,$E8
	.dc.b	$FE,$03,$F4,$9E,$EB,$E7,$3F,$67,$18,$6E,$7F,$B6,$9D,$65,$FB,$6F
	.dc.b	$0B,$FC,$9F,$9F,$F6,$47,$FD,$05,$7C,$27,$55,$C2,$B5,$87,$59,$CD
	.dc.b	$75,$38,$4F,$3A,$E1,$9F,$EF,$9B,$5F,$BE,$6B,$FE,$FF,$BE,$95,$C6
	.dc.b	$FA,$4F,$9E,$73,$7C,$E7,$35,$E4,$C6,$CE,$6F,$CA,$71,$8C,$BE,$1A
	.dc.b	$AF,$86,$2E,$3F,$0A,$38,$DF,$56,$DF,$58,$F3,$47,$53,$27,$52,$9D
	.dc.b	$4B,$3D,$63,$89,$D4,$B3,$D4,$93,$FC,$93,$0F,$82,$F5,$3F,$95,$E3
	.dc.b	$F7,$1B,$8F,$F8,$55,$D1,$FE,$1C,$F8,$DF,$E8,$FF,$47,$98,$46,$1E
	.dc.b	$0D,$7F,$C3,$CD,$7F,$9B,$FC,$03,$CD,$E5,$A2,$31,$3E,$18,$C4,$EA
	.dc.b	$73,$3F,$D0,$9D,$E9,$F9,$16,$E9,$DA,$BF,$A6,$FC,$07,$F8,$D2,$B5
	.dc.b	$1F,$D4,$2B,$77,$FD,$21,$AB,$05,$F1,$C3,$A4,$FE,$86,$B7,$5E,$68
	.dc.b	$FD,$25,$6D,$E9,$7D,$C8,$E8,$B8,$79,$1E,$91,$E8,$73,$87,$82,$74
	.dc.b	$7E,$BE,$BB,$4F,$9F,$0F,$19,$2F,$94,$74,$25,$72,$BF,$ED,$37,$CB
	.dc.b	$0F,$54,$E6,$BD,$27,$9C,$E3,$5E,$32,$3F,$DC,$2B,$96,$1E,$18,$D9
	.dc.b	$93,$C7,$5D,$5F,$46,$18,$BF,$1C,$7C,$03,$C3,$3E,$F7,$46,$23,$C2
	.dc.b	$FD,$7B,$FA,$1D,$8F,$33,$FD,$0F,$3F,$E5,$67,$FC,$38,$DF,$2B,$71
	.dc.b	$A8,$FF,$1A,$73,$E0,$8F,$1C,$2B,$A3,$A4,$72,$8F,$0F,$91,$FE,$93
	.dc.b	$9F,$F3,$2B,$FE,$6D,$6F,$8D,$AF,$1E,$46,$9A,$FF,$9A,$25,$E3,$E3
	.dc.b	$5F,$F0,$85,$5C,$3F,$C1,$D5,$AB,$F7,$7C,$AF,$B3,$C9,$91,$82,$1F
	.dc.b	$E5,$18,$F8,$27,$C2,$BB,$1A,$C3,$FC,$C5,$A2,$B5,$3E,$CF,$27,$DE
	.dc.b	$47,$DF,$E4,$FC,$22,$38,$3E,$1F,$93,$F1,$88,$E0,$FA,$DB,$7D,$4D
	.dc.b	$A4,$7D,$97,$EB,$D7,$D8,$B9,$3F,$75,$FB,$77,$F7,$3C,$99,$3F,$81
	.dc.b	$7E,$FF,$B3,$F8,$0F,$1A,$4F,$E1,$5F,$C1,$F8,$DF,$6B,$C6,$93,$F6
	.dc.b	$9F,$BB,$E3,$7E,$21,$91,$F8,$D7,$F1,$71,$FC,$07,$A5,$F9,$BB,$C5
	.dc.b	$C3,$6F,$FD,$22,$D7,$FD,$34,$BC,$6D,$CC,$9D,$BC,$A2,$46,$08,$72
	.dc.b	$7F,$3B,$1F,$9E,$64,$7E,$8E,$32,$FE,$96,$32,$FF,$A5,$63,$F4,$BE
	.dc.b	$2F,$EB,$3A,$7F,$DB,$1C,$DF,$EA,$56,$9F,$E5,$FE,$6F,$F0,$6D,$FE
	.dc.b	$BE,$B8,$B4,$FF,$63,$CF,$F4,$FA,$FF,$60,$ED,$F0,$5B,$87,$C9,$8C
	.dc.b	$4A,$BF,$A4,$47,$67,$FD,$80,$E9,$CC,$F1,$39,$5C,$8F,$F5,$FC,$EF
	.dc.b	$3F,$F2,$85,$66,$7F,$DA,$F6,$FA,$1F,$81,$E7,$49,$F4,$C7,$C3,$30
	.dc.b	$F0,$8C,$3E,$EF,$9F,$F9,$04,$7E,$0F,$CF,$D7,$BC,$5D,$3E,$64,$FF
	.dc.b	$33,$F7,$E5,$7F,$A9,$9F,$E6,$DE,$F4,$8F,$F3,$29,$FD,$5D,$F8,$CF
	.dc.b	$FA,$A1,$BE,$5F,$FA,$A1,$B1,$FE,$A8,$79,$3F,$E3,$B7,$FA,$5F,$D0
	.dc.b	$FF,$19,$8F,$F7,$CF,$D1,$3E,$83,$F9,$6A,$FA,$1A,$62,$24,$FD,$7C
	.dc.b	$FF,$DD,$55,$C6,$31,$BF,$FA,$96,$31,$1F,$42,$FF,$51,$F5,$8F,$9B
	.dc.b	$A4,$9F,$F7,$BC,$3D,$FD,$99,$86,$2C,$7F,$11,$C8,$BE,$95,$CA,$FF
	.dc.b	$B0,$7B,$3F,$23,$E0,$FF,$D3,$BE,$7B,$E0,$F1,$C5,$FE,$CE,$6E,$1F
	.dc.b	$EC,$EB,$AB,$E1,$5D,$FF,$E2,$58,$C3,$63,$31,$FE,$29,$6B,$FE,$2F
	.dc.b	$EA,$7C,$37,$49,$CE,$7F,$EB,$06,$E5,$F7,$AB,$77,$FA,$5A,$FF,$5A
	.dc.b	$F2,$FF,$4F,$84,$63,$CC,$F3,$CF,$FB,$46,$78,$5F,$E2,$23,$85,$23
	.dc.b	$FD,$B5,$C3,$F5,$A3,$AF,$95,$C8,$F8,$7E,$87,$CF,$E8,$BF,$60,$F7
	.dc.b	$FE,$2F,$D2,$7D,$7D,$3B,$9F,$BB,$70,$F9,$B3,$FF,$74,$57,$18,$DE
	.dc.b	$1F,$6F,$E5,$3A,$E1,$27,$3A,$27,$C6,$FD,$E8,$FD,$19,$FE,$D7,$70
	.dc.b	$FE,$9A,$BF,$0D,$D5,$FE,$5B,$6C,$7F,$98,$5F,$C5,$7C,$C7,$49,$93
	.dc.b	$E0,$8E,$FD,$1E,$0F,$91,$93,$E4,$8C,$B4,$87,$BC,$F2,$7D,$1F,$A3
	.dc.b	$E8,$FC,$59,$1A,$49,$F2,$43,$E4,$78,$47,$C2,$9F,$83,$1F,$0B,$E8
	.dc.b	$39,$92,$78,$BD,$BE,$3B,$51,$F2,$A3,$E4,$C7,$CA,$F8,$AF,$97,$C1
	.dc.b	$2B,$E7,$C6,$5A,$43,$D9,$BF,$EC,$7E,$97,$E9,$FC,$C7,$B0,$93,$EA
	.dc.b	$7E,$87,$E4,$7F,$8F,$8F,$B7,$1C,$1F,$C0,$C6,$5F,$90,$3E,$B5,$C3
	.dc.b	$FB,$27,$F1,$7D,$5B,$6F,$3F,$F2,$47,$75,$F9,$2F,$15,$F9,$AB,$E4
	.dc.b	$BF,$D3,$91,$86,$C6,$67,$FD,$49,$39,$7F,$0D,$93,$E6,$FF,$82,$CE
	.dc.b	$67,$EF,$37,$2F,$76,$CB,$E1,$5F,$A4,$3D,$9C,$1E,$BE,$09,$19,$D3
	.dc.b	$F4,$8F,$C2,$FA,$8F,$F5,$C3,$F9,$E6,$BF,$D0,$9A,$FF,$69,$6A,$FA
	.dc.b	$FF,$E7,$4F,$85,$AF,$CE,$6E,$5F,$47,$C6,$7C,$67,$F2,$06,$D7,$A5
	.dc.b	$AF,$D6,$6C,$FC,$0C,$EF,$0F,$F5,$6F,$A7,$F9,$97,$FE,$CF,$6B,$FB
	.dc.b	$86,$BB,$E2,$FE,$59,$F2,$1F,$E0,$18,$FD,$EB,$5F,$EC,$2D,$7F,$C1
	.dc.b	$0D,$C1,$F6,$BF,$AD,$9F,$E2,$71,$8E,$BF,$E7,$AD,$8C,$CF,$EB,$DF
	.dc.b	$C9,$FC,$7F,$E6,$08,$FE,$09,$FF,$72,$7D,$23,$ED,$58,$E1,$B5,$FE
	.dc.b	$ED,$9D,$4E,$4F,$E9,$1F,$1F,$FE,$80,$C7,$FA,$CF,$C6,$7F,$BC,$7E
	.dc.b	$27,$F9,$DE,$1F,$EE,$98,$C4,$EA,$3F,$C4,$D5,$C5,$D1,$FE,$38,$29
	.dc.b	$FE,$E3,$29,$FD,$DC,$A7,$F6,$A3,$8E,$0F,$B5,$79,$DF,$B3,$9F,$F1
	.dc.b	$A4,$E6,$31,$08,$C7,$0F,$BE,$B8,$BF,$A8,$F9,$0D,$F5,$FE,$8D,$BC
	.dc.b	$86,$9A,$E6,$FC,$8D,$7F,$44,$6E,$3F,$C4,$2B,$FD,$EF,$39,$9F,$E1
	.dc.b	$9A,$7F,$09,$FB,$07,$CF,$F9,$B0,$46,$2F,$FA,$3C,$7F,$93,$E3,$0E
	.dc.b	$1A,$FF,$4C,$F9,$0F,$EB,$D3,$C2,$11,$87,$9B,$5E,$1D,$DE,$1A,$7C
	.dc.b	$BF,$9A,$FF,$C7,$31,$87,$F9,$0F,$D0,$E6,$FC,$95,$57,$E7,$DF,$E8
	.dc.b	$7A,$FE,$14,$D8,$FE,$4F,$7C,$DF,$E5,$31,$9C,$FE,$27,$D1,$F8,$C9
	.dc.b	$7E,$BB,$47,$C7,$F8,$0F,$E4,$D3,$FC,$5D,$F8,$B5,$FC,$D5,$CE,$1F
	.dc.b	$CD,$9C,$CF,$F9,$E7,$87,$F9,$07,$D1,$7A,$FC,$57,$EA,$DF,$63,$0F
	.dc.b	$B5,$39,$8C,$63,$FD,$16,$B7,$F3,$1F,$8E,$6E,$3F,$CD,$2B,$96,$1F
	.dc.b	$E6,$29,$CB,$F7,$32,$FD,$9D,$7A,$76,$F9,$51,$89,$F4,$47,$CF,$CF
	.dc.b	$F9,$B7,$C2,$DE,$57,$C2,$D8,$C3,$31,$FE,$86,$FE,$68,$8F,$F4,$6F
	.dc.b	$0F,$91,$CF,$9A,$C7,$F5,$E8,$ED,$F0,$38,$CF,$F3,$8F,$87,$E0,$7C
	.dc.b	$5F,$03,$6F,$FB,$2B,$87,$FA,$9B,$FA,$A2,$3F,$D5,$BE,$8F,$D5,$F3
	.dc.b	$FE,$B6,$C7,$F6,$E8,$FD,$89,$1F,$EB,$AF,$ED,$3C,$4F,$84,$7C,$3E
	.dc.b	$0B,$3F,$DD,$A3,$E3,$73,$E3,$3F,$DA,$3F,$27,$E8,$BE,$67,$9C,$E5
	.dc.b	$C6,$1F,$E3,$E8,$C3,$63,$F8,$2C,$62,$FF,$C7,$7C,$9F,$E0,$19,$C9
	.dc.b	$AF,$17,$EC,$7D,$AF,$FA,$57,$0E,$91,$8F,$89,$91,$FD,$BF,$EC,$DE
	.dc.b	$F6,$32,$FF,$06,$C6,$27,$C9,$F3,$BF,$AF,$87,$B5,$EA,$8C,$32,$7A
	.dc.b	$7A,$B6,$F9,$BC,$7D,$9F,$DC,$7C,$43,$E4,$31,$7F,$E0,$18,$C3,$63
	.dc.b	$F8,$B4,$62,$FF,$8A,$B5,$FF,$2E,$B5,$FF,$60,$FC,$7B,$DE,$FE,$0C
	.dc.b	$FA,$2F,$FC,$1C,$E5,$B9,$7E,$CF,$EC,$DD,$9F,$E3,$3F,$8E,$FF,$55
	.dc.b	$63,$F5,$E7,$FD,$49,$7F,$EB,$A8,$C5,$6F,$E0,$BF,$D5,$F5,$A9,$FF
	.dc.b	$59,$C6,$3B,$F5,$1C,$DB,$CC,$AF,$0E,$1D,$63,$13,$C3,$9F,$FC,$23
	.dc.b	$D5,$7C,$FF,$E2,$B7,$E3,$9F,$F2,$26,$59,$A9,$39,$A7,$5E,$33,$F2
	.dc.b	$8C,$3F,$C3,$B1,$E6,$86,$EF,$81,$C3,$70,$F8,$52,$F8,$CF,$E8,$FC
	.dc.b	$1F,$22,$38,$DF,$03,$FC,$BD,$5F,$E9,$4B,$E5,$3F,$DA,$DB,$19,$9D
	.dc.b	$CF,$DC,$E5,$F0,$9C,$E7,$C3,$48,$F8,$DD,$AF,$9D,$A6,$3D,$6B,$7F
	.dc.b	$1E,$74,$D9,$D3,$4D,$7C,$F7,$4F,$8D,$EA,$D8,$C9,$CE,$32,$D3,$ED
	.dc.b	$7E,$9A,$F9,$5F,$BE,$FF,$55,$DB,$F1,$BC,$C8,$D5,$70,$F9,$7D,$91
	.dc.b	$E7,$77,$7F,$B3,$7C,$6B,$18,$74,$C3,$9D,$69,$F0,$73,$7F,$53,$D5
	.dc.b	$EE,$FC,$83,$E4,$FC,$FD,$D3,$E3,$47,$CC,$F9,$FC,$63,$F8,$8E,$7F
	.dc.b	$F4,$94,$7D,$0F,$9F,$FA,$3F,$3F,$70,$F9,$CD,$BE,$19,$23,$1A,$FE
	.dc.b	$BB,$1F,$F4,$E4,$7E,$93,$F5,$3F,$C6,$61,$FA,$37,$D3,$87,$F2,$97
	.dc.b	$2F,$C2,$73,$C0,$69,$3A,$7A,$70,$CB,$E8,$F7,$8C,$5F,$9B,$E8,$3F
	.dc.b	$16,$7E,$33,$C4,$B7,$FA,$FF,$A3,$71,$7E,$D5,$B1,$9E,$1C,$4F,$E5
	.dc.b	$3E,$86,$4F,$E4,$7E,$86,$E5,$F1,$3E,$33,$F3,$27,$E6,$B4,$B5,$F9
	.dc.b	$7D,$7F,$3A,$7E,$7B,$A2,$C7,$F0,$1F,$8C,$61,$FC,$1D,$B1,$9F,$A3
	.dc.b	$C4,$F9,$F3,$93,$C7,$E8,$2C,$FD,$DD,$DA,$FB,$4C,$F3,$7D,$78,$CD
	.dc.b	$FC,$3D,$BE,$11,$B7,$07,$97,$C1,$F2,$1E,$5F,$A0,$B1,$92,$63,$F8
	.dc.b	$DD,$53,$A8,$C5,$7B,$7E,$83,$19,$7E,$26,$D1,$F4,$F8,$66,$3F,$5F
	.dc.b	$86,$E3,$ED,$9B,$87,$AE,$6F,$6F,$F8,$DF,$5F,$C6,$CF,$BF,$8B,$38
	.dc.b	$7F,$27,$9F,$D6,$CF,$87,$D2,$FA,$07,$D1,$F9,$DB,$9F,$D5,$8B,$7E
	.dc.b	$D9,$B5,$C6,$BE,$D7,$CD,$E3,$2F,$F2,$46,$C8,$FE,$69,$F3,$FF,$E8
	.dc.b	$0B,$FE,$5C,$BE,$45,$CF,$FC,$D2,$35,$87,$E1,$33,$46,$E7,$5E,$5F
	.dc.b	$4B,$E3,$5F,$D1,$0D,$C6,$67,$79,$F5,$AD,$E3,$E2,$AD,$5F,$1C,$3B
	.dc.b	$36,$B7,$8F,$8E,$7C,$B3,$F9,$87,$77,$EC,$AD,$CE,$5A,$F9,$30,$F1
	.dc.b	$63,$23,$B3,$E6,$69,$2B,$A4,$65,$B1,$9C,$3C,$97,$DB,$0F,$2C,$FC
	.dc.b	$88,$C1,$2F,$95,$70,$7C,$57,$D0,$F8,$98,$62,$75,$7D,$CF,$81,$8F
	.dc.b	$C8,$AE,$13,$A7,$8C,$EC,$F1,$5F,$37,$8D,$D7,$26,$F1,$F4,$9F,$32
	.dc.b	$FC,$31,$F8,$33,$65,$8F,$AE,$DC,$7F,$06,$EF,$E4,$3F,$8F,$FD,$1F
	.dc.b	$FA,$FD,$B1,$FE,$C8,$6B,$BE,$B4,$FF,$B1,$6F,$FD,$B9,$E0,$5F,$8F
	.dc.b	$FF,$31,$5F,$FB,$42,$72,$D8,$CB,$E5,$77,$AF,$E8,$1B,$9F,$E2,$DB
	.dc.b	$3F,$8B,$F8,$8F,$D0,$56,$F0,$FD,$25,$6A,$BF,$46,$DC,$17,$C3,$0E
	.dc.b	$97,$C0,$97,$F9,$D9,$D3,$F9,$9B,$E4,$49,$FD,$94,$EA,$7B,$C7,$2C
	.dc.b	$3C,$91,$CC,$C5,$F8,$1F,$D5,$5F,$96,$F9,$61,$F1,$AF,$89,$2F,$F5
	.dc.b	$93,$AA,$F1,$5F,$2C,$3C,$6F,$63,$D0,$D7,$F8,$2C,$F1,$8C,$61,$DE
	.dc.b	$FA,$D6,$2F,$AE,$1F,$C4,$B0,$E1,$3D,$C9,$19,$78,$61,$D2,$7D,$67
	.dc.b	$09,$F5,$4F,$D9,$3F,$CB,$9F,$44,$F7,$3E,$7F,$8C,$87,$B5,$86,$D7
	.dc.b	$0A,$C1,$FC,$87,$87,$F9,$15,$F1,$8C,$4F,$FB,$D6,$BB,$3F,$DA,$0C
	.dc.b	$46,$9F,$F3,$34,$69,$FF,$11,$37,$0F,$F0,$D3,$74,$E8,$7A,$3E,$41
	.dc.b	$E7,$5F,$3A,$61,$FE,$79,$9C,$B6,$7C,$19,$27,$32,$CF,$D7,$18,$6D
	.dc.b	$70,$8D,$E5,$FE,$2D,$6E,$3F,$E0,$6C,$3A,$57,$97,$CD,$F9,$5E,$7F
	.dc.b	$CB,$F4,$26,$7F,$4A,$32,$FC,$53,$E1,$BE,$6D,$7F,$CE,$11,$9C,$3E
	.dc.b	$B4,$E4,$E3,$1F,$21,$FB,$18,$7C,$5F,$26,$4F,$C3,$7E,$47,$34,$8E
	.dc.b	$9A,$7F,$8C,$1B,$1E,$F3,$E0,$79,$FF,$3C,$91,$F8,$AE,$0F,$C3,$6B
	.dc.b	$23,$F1,$AF,$E2,$6B,$F1,$58,$7A,$C9,$3E,$BD,$11,$EF,$F3,$BF,$20
	.dc.b	$49,$FC,$BF,$9B,$F9,$3F,$36,$47,$E6,$9F,$CB,$79,$BF,$83,$F3,$3E
	.dc.b	$B1,$27,$DA,$FC,$B3,$F3,$84,$7F,$47,$E6,$7F,$10,$EA,$7D,$1F,$FA
	.dc.b	$C6,$7F,$C8,$5F,$3F,$FA,$6F,$14,$EE,$75,$3F,$E8,$86,$F5,$7F,$81
	.dc.b	$6F,$9F,$97,$F9,$F6,$D7,$DF,$6B,$F9,$13,$FC,$9E,$E3,$AF,$FD,$37
	.dc.b	$7D,$9B,$5E,$1F,$2F,$F6,$C4,$BF,$DE,$BF,$B9,$8C,$47,$EB,$22,$CF
	.dc.b	$EB,$27,$F8,$0F,$67,$F5,$98,$FF,$44,$34,$FC,$E1,$6B,$F6,$FE,$5F
	.dc.b	$06,$1F,$EE,$78,$D7,$67,$F0,$B9,$F9,$9C,$2C,$69,$2F,$FD,$E9,$1E
	.dc.b	$DB,$F8,$2D,$86,$68,$E5,$1C,$2B,$AE,$52,$30,$43,$FD,$77,$3C,$F1
	.dc.b	$61,$FE,$F2,$9F,$F3,$F7,$BF,$F7,$4F,$87,$E7,$48,$F6,$35,$FF,$5B
	.dc.b	$4B,$D0,$FC,$27,$CF,$C9,$FF,$44,$3F,$33,$D1,$91,$82,$1C,$9F,$EA
	.dc.b	$3F,$15,$FD,$4B,$E3,$E4,$FF,$93,$E3,$2F,$F0,$53,$FC,$9E,$FF,$29
	.dc.b	$69,$9F,$25,$19,$6C,$73,$21,$FE,$4B,$F8,$BF,$E8,$3F,$1A,$F4,$FF
	.dc.b	$A1,$FD,$37,$DA,$3E,$DC,$66,$BF,$D4,$7C,$DF,$EC,$D8,$C4,$7F,$B6
	.dc.b	$67,$83,$D2,$BF,$59,$F4,$8E,$1F,$D7,$7C,$6B,$E8,$69,$A8,$97,$DB
	.dc.b	$8B,$F4,$77,$C6,$30,$7E,$11,$C3,$FE,$F8,$7F,$6A,$FF,$9D,$FE,$3F
	.dc.b	$FB,$E9,$1F,$F3,$D7,$C1,$7F,$80,$0F,$F1,$C3,$FE,$69,$6B,$FE,$69
	.dc.b	$6B,$FE,$7E,$6E,$EF,$84,$FF,$41,$7C,$97,$EE,$6B,$47,$FB,$21,$C4
	.dc.b	$AB,$7D,$FF,$E7,$7B,$B7,$FE,$CE,$FA,$4F,$E7,$EB,$0E,$91,$FD,$CA
	.dc.b	$31,$5F,$E9,$18,$C4,$EB,$D0,$E7,$49,$FC,$FB,$E3,$AE,$13,$A9,$EF
	.dc.b	$5D,$B4,$E3,$F4,$9F,$2B,$0E,$E7,$FB,$89,$FF,$54,$3C,$CC,$7B,$9C
	.dc.b	$3C,$4E,$3C,$16,$74,$2B,$A6,$1F,$1D,$FB,$2F,$F2,$1E,$E7,$8A,$FA
	.dc.b	$2F,$A7,$18,$6C,$66,$FE,$CC,$E5,$B1,$F8,$D7,$CD,$19,$7D,$6D,$AF
	.dc.b	$04,$EE,$72,$BE,$7B,$EA,$D7,$93,$E8,$EC,$9B,$32,$BB,$CF,$16,$C7
	.dc.b	$2D,$7D,$11,$DB,$E3,$3D,$64,$7D,$2F,$BB,$E3,$3B,$B2,$3A,$C7,$8B
	.dc.b	$E3,$BE,$19,$23,$E0,$BF,$1B,$E6,$F5,$1E,$48,$F4,$C7,$AB,$E2,$FD
	.dc.b	$7F,$17,$2B,$D3,$19,$69,$0F,$93,$8A,$BE,$2F,$C1,$FE,$86,$7F,$C8
	.dc.b	$6D,$7F,$C4,$8D,$5B,$7F,$BB,$79,$FE,$EC,$BE,$8B,$9C,$4A,$FB,$33
	.dc.b	$97,$F2,$1E,$73,$E5,$BF,$13,$5F,$DA,$1B,$3F,$D2,$1A,$BE,$32,$8F
	.dc.b	$2F,$F7,$AD,$8F,$E0,$2D,$7F,$82,$B5,$FF,$5F,$3F,$6B,$97,$F8,$03
	.dc.b	$B6,$CF,$EC,$A3,$56,$A9,$F4,$2F,$8C,$FD,$BB,$63,$F4,$ED,$7F,$62
	.dc.b	$7A,$78,$FA,$71,$FA,$09,$1E,$87,$D7,$C1,$E1,$24,$7E,$CB,$E8,$BF
	.dc.b	$DA,$3A,$7F,$17,$6E,$5E,$59,$46,$1F,$43,$56,$3F,$D0,$2B,$F9,$07
	.dc.b	$CB,$7F,$74,$9F,$3B,$63,$3E,$A7,$F1,$4F,$B4,$FE,$1F,$8F,$7E,$A7
	.dc.b	$93,$FC,$B9,$B8,$FF,$32,$6B,$FC,$D9,$AF,$FB,$BF,$17,$AD,$FC,$81
	.dc.b	$B1,$FC,$41,$AF,$CA,$CF,$F8,$8E,$1E,$FF,$B6,$7A,$1F,$CF,$9B,$5F
	.dc.b	$D0,$9A,$FF,$46,$6B,$FE,$FF,$7C,$6F,$3C,$B9,$11,$FF,$7E,$7C,$0F
	.dc.b	$F3,$B7,$FC,$37,$F0,$FF,$3F,$E3,$13,$9E,$9F,$F5,$14,$63,$87,$FC
	.dc.b	$87,$5B,$AD,$3F,$D6,$3D,$3C,$E1,$FE,$39,$3A,$97,$0F,$F5,$63,$FD
	.dc.b	$E0,$C8,$FA,$2F,$DD,$6B,$FD,$81,$B1,$FD,$91,$AF,$E1,$4E,$1C,$BF
	.dc.b	$BB,$87,$81,$49,$5C,$DF,$E0,$1F,$3E,$F4,$3F,$CB,$DA,$FF,$67,$8C
	.dc.b	$57,$F4,$67,$FB,$33,$67,$6F,$E5,$7E,$23,$C7,$95,$CB,$F8,$64,$62
	.dc.b	$3F,$5C,$D5,$F5,$FF,$C3,$6B,$FA,$9B,$63,$F5,$BF,$01,$FC,$4D,$FD
	.dc.b	$7C,$67,$2F,$37,$CB,$F8,$A7,$F7,$59,$A3,$10,$FA,$1F,$F2,$F6,$3E
	.dc.b	$B9,$CF,$B5,$F9,$E8,$FE,$C6,$DA,$FD,$ED,$F3,$9C,$F2,$F0,$DB,$F9
	.dc.b	$45,$FC,$6B,$FF,$0B,$B5,$FA,$8D,$7F,$75,$A3,$E8,$BC,$AD,$C3,$C3
	.dc.b	$EB,$62,$7C,$BD,$7E,$6F,$9E,$B5,$E6,$70,$FF,$89,$31,$47,$90,$91
	.dc.b	$FE,$2A,$E5,$7E,$8F,$FA,$C8,$FE,$6E,$DC,$FF,$8D,$DF,$BD,$FF,$06
	.dc.b	$D7,$0E,$CF,$9D,$E5,$BC,$A7,$CB,$61,$FA,$DC,$BF,$E0,$77,$39,$FF
	.dc.b	$46,$74,$7F,$51,$AE,$4D,$3E,$F6,$3F,$D5,$9C,$7C,$A7,$53,$C3,$F4
	.dc.b	$5F,$92,$9F,$F0,$3C,$A7,$57,$FD,$53,$E5,$6D,$69,$F3,$70,$F9,$FE
	.dc.b	$0E,$D7,$F5,$A7,$0F,$F8,$63,$04,$79,$49,$19,$F8,$3F,$81,$B7,$C1
	.dc.b	$6C,$7F,$64,$8F,$61,$1E,$8F,$BF,$9F,$E4,$1E,$CE,$9B,$3F,$DB,$23
	.dc.b	$E6,$74,$63,$4F,$7B,$F3,$F9,$FE,$87,$BD,$6F,$FB,$9B,$87,$D4,$FC
	.dc.b	$D2,$3F,$E6,$AF,$9F,$FE,$DC,$FF,$90,$9F,$F1,$E3,$7D,$1F,$D4,$C6
	.dc.b	$30,$FE,$BE,$DE,$4F,$F3,$67,$C4,$FE,$02,$FF,$AB,$4E,$5F,$EB,$11
	.dc.b	$9A,$FE,$DD,$3F,$C6,$AB,$8F,$17,$6F,$07,$F6,$F6,$8A,$63,$8E,$E3
	.dc.b	$FE,$58,$9F,$D4,$63,$FD,$CB,$E2,$1E,$C7,$F0,$05,$E9,$58,$C3,$2F
	.dc.b	$FD,$17,$3A,$6C,$7F,$88,$E3,$15,$FE,$92,$5E,$92,$BF,$D3,$F2,$9C
	.dc.b	$C6,$1E,$3E,$47,$F9,$FA,$75,$5C,$54,$7F,$A0,$BB,$5F,$3C,$FA,$2D
	.dc.b	$4E,$63,$18,$FF,$AB,$AB,$7C,$9A,$F6,$7F,$C3,$51,$CB,$0C,$63,$FE
	.dc.b	$F6,$C3,$8B,$E0,$AD,$F2,$7F,$1F,$CB,$B5,$F0,$C3,$FC,$65,$CF,$CB
	.dc.b	$BF,$F7,$FF,$3F,$7E,$27,$FA,$9E,$1F,$E9,$A9,$CE,$DF,$EC,$25,$3A
	.dc.b	$CF,$FC,$5D,$1A,$D3,$FD,$A9,$CD,$D7,$C2,$65,$D6,$3D,$33,$F3,$7D
	.dc.b	$CB,$C6,$6F,$9B,$86,$69,$FE,$F3,$9D,$36,$3E,$83,$FD,$9E,$33,$5F
	.dc.b	$E6,$56,$EB,$F5,$FE,$1B,$EF,$57,$DB,$BF,$1C,$62,$FE,$51,$FE,$68
	.dc.b	$F6,$DF,$0D,$F1,$FC,$5F,$9F,$8F,$C5,$9C,$E3,$F1,$67,$3C,$1F,$E7
	.dc.b	$2E,$4E,$07,$CF,$AD,$E3,$DB,$87,$F9,$84,$E7,$E5,$66,$3F,$3E,$DF
	.dc.b	$2F,$8C,$38,$42,$BC,$9E,$FE,$74,$FB,$1A,$7F,$9D,$F1,$FC,$67,$CF
	.dc.b	$FF,$66,$C7,$DB,$D0,$FD,$3F,$FC,$CD,$5A,$F7,$FF,$0D,$18,$AD,$DF
	.dc.b	$1C,$7F,$39,$39,$7F,$3D,$19,$E0,$FC,$BC,$E5,$B5,$D0,$FC,$94,$ED
	.dc.b	$63,$E3,$7E,$37,$CB,$4C,$BE,$17,$B5,$67,$E0,$CD,$CB,$E9,$4D,$C7
	.dc.b	$F2,$7B,$BE,$5B,$E8,$1F,$47,$E6,$6E,$5F,$36,$2E,$5F,$32,$6E,$3F
	.dc.b	$B9,$FA,$0F,$36,$86,$8F,$1B,$F5,$8E,$38,$3F,$27,$87,$EE,$23,$E8
	.dc.b	$65,$FE,$7F,$BE,$55,$AA,$F8,$35,$A7,$84,$7E,$3F,$E8,$FA,$B6,$FF
	.dc.b	$09,$F3,$3F,$8B,$3F,$19,$F2,$D6,$3F,$6D,$F4,$2F,$9C,$ED,$D6,$8C
	.dc.b	$4F,$DD,$CB,$FD,$8B,$18,$7F,$8D,$18,$AE,$3C,$2F,$9B,$FC,$D1,$F9
	.dc.b	$BF,$9B,$B8,$7E,$5F,$DE,$FC,$F1,$F9,$FE,$1B,$1F,$9B,$DB,$99,$1F
	.dc.b	$16,$9F,$EC,$78,$F3,$FC,$4F,$E2,$74,$FF,$6C,$B6,$4B,$E8,$3F,$8D
	.dc.b	$39,$DB,$CD,$87,$C2,$D3,$3F,$37,$FC,$7E,$73,$8F,$EC,$56,$BF,$19
	.dc.b	$B7,$E4,$6C,$FC,$4C,$BE,$55,$6F,$5E,$87,$56,$DF,$F2,$7C,$BF,$A5
	.dc.b	$37,$C5,$F3,$FC,$DF,$A1,$91,$F0,$A7,$78,$F7,$6C,$7F,$B1,$F1,$F8
	.dc.b	$35,$F1,$8F,$83,$F1,$FC,$AF,$F8,$65,$FF,$B7,$EF,$F0,$7C,$1F,$04
	.dc.b	$F8,$5C,$16,$32,$7B,$E3,$E1,$11,$C1,$F8,$CF,$8D,$E1,$7F,$25,$FA
	.dc.b	$3F,$37,$27,$E0,$7D,$0D,$AF,$1C,$FD,$C2,$4F,$BD,$B1,$E8,$F9,$9E
	.dc.b	$1D,$9F,$D4,$73,$47,$F3,$6F,$3F,$F9,$BF,$8B,$F8,$1B,$FE,$6E,$3C
	.dc.b	$BF,$F2,$AB,$87,$FC,$B9,$D1,$FD,$F1,$B1,$99,$E5,$8F,$F4,$5B,$EB
	.dc.b	$7B,$9C,$F8,$5F,$EF,$7C,$38,$DE,$F4,$D5,$CC,$7A,$3D,$F1,$E7,$18
	.dc.b	$25,$71,$6E,$1D,$71,$EF,$03,$D8,$E1,$F7,$57,$4A,$FE,$B9,$39,$C7
	.dc.b	$CB,$7E,$7A,$D3,$D7,$0F,$35,$F8,$4E,$2E,$23,$58,$7C,$9A,$E2,$F8
	.dc.b	$EF,$AB,$70,$F0,$4E,$A3,$CF,$7C,$DE,$11,$9A,$E3,$1A,$3B,$63,$B7
	.dc.b	$18,$F9,$1C,$9F,$18,$F0,$5E,$EB,$EF,$CE,$CE,$13,$FE,$E9,$8F,$BE
	.dc.b	$DC,$3F,$10,$DA,$FF,$77,$3F,$3F,$E3,$BF,$94,$7D,$0E,$21,$18,$76
	.dc.b	$D7,$FD,$EC,$D7,$F0,$DF,$09,$FD,$71,$EB,$87,$FA,$6F,$C3,$E8,$D8
	.dc.b	$CC,$ED,$F4,$3C,$72,$FF,$0C,$37,$5E,$B7,$F9,$BF,$73,$F0,$D7,$F9
	.dc.b	$FA,$D3,$F5,$DA,$FE,$95,$B8,$7F,$9C,$E3,$F4,$91,$82,$57,$58,$CB
	.dc.b	$1E,$CE,$5A,$4C,$3C,$F1,$96,$C6,$70,$F4,$BF,$B5,$C3,$D3,$3A,$24
	.dc.b	$E9,$FF,$1C,$46,$A3,$04,$9F,$75,$77,$8C,$47,$BE,$3D,$15,$CA,$2C
	.dc.b	$EB,$0F,$A5,$39,$9F,$31,$D0,$A7,$F0,$DB,$F1,$39,$8F,$3B,$6B,$F8
	.dc.b	$34,$EA,$3C,$47,$2A,$EF,$3A,$8F,$19,$C4,$87,$73,$FC,$E1,$F3,$EF
	.dc.b	$90,$F0,$F5,$21,$F4,$38,$6D,$70,$AC,$19,$99,$19,$21,$FE,$53,$9E
	.dc.b	$CE,$32,$FF,$41,$47,$FA,$7D,$FF,$42,$7C,$87,$F9,$52,$7F,$D0,$8F
	.dc.b	$F9,$4D,$BE,$87,$9A,$35,$1F,$E5,$F6,$CF,$29,$D1,$BF,$46,$57,$46
	.dc.b	$BC,$7C,$DF,$77,$A7,$EF,$FA,$19,$3F,$1A,$32,$FC,$88,$CB,$ED,$8C
	.dc.b	$B4,$E0,$7C,$EA,$F4,$7A,$1F,$AC,$E8,$FE,$E1,$9A,$31,$5F,$EA,$08
	.dc.b	$C3,$FE,$1A,$6B,$FE,$07,$7F,$C6,$4F,$C6,$8E,$AF,$C8,$F2,$BE,$37
	.dc.b	$9F,$D2,$1C,$2B,$4E,$63,$0D,$8C,$CF,$F8,$AB,$76,$BF,$83,$24,$7E
	.dc.b	$2B,$0F,$C3,$63,$F8,$AE,$A7,$C7,$53,$35,$F9,$0F,$47,$F1,$84,$7F
	.dc.b	$27,$E7,$FE,$3F,$CF,$91,$F9,$67,$F2,$5E,$77,$E0,$7C,$BF,$CB,$FB
	.dc.b	$4F,$62,$99,$BF,$CE,$CF,$E6,$48,$FE,$83,$CE,$FC,$FF,$8C,$F9,$DA
	.dc.b	$49,$FF,$74,$7D,$0F,$F9,$CD,$BE,$66,$5F,$F1,$1C,$FE,$E3,$C3,$56
	.dc.b	$E7,$FD,$46,$DE,$EF,$F0,$85,$71,$7F,$DF,$51,$9F,$07,$FD,$BF,$CE
	.dc.b	$F3,$7E,$57,$C8,$4C,$7F,$6F,$1F,$B3,$30,$FF,$AF,$F9,$93,$98,$FF
	.dc.b	$76,$5F,$F3,$A7,$5F,$15,$F5,$4F,$DF,$C6,$72,$FF,$7A,$34,$FF,$78
	.dc.b	$DA,$78,$57,$43,$E0,$72,$79,$DF,$F6,$B7,$BD,$F8,$13,$39,$48,$FF
	.dc.b	$42,$6C,$AE,$11,$FE,$84,$8C,$4F,$E5,$DB,$1C,$27,$79,$32,$FF,$0D
	.dc.b	$56,$AB,$FD,$A3,$5A,$9C,$BF,$36,$3F,$CF,$FF,$3D,$F8,$02,$3F,$1B
	.dc.b	$D3,$AF,$F9,$7A,$2E,$1F,$12,$7F,$92,$FE,$CB,$FD,$AE,$3C,$55,$FA
	.dc.b	$C8,$C6,$1F,$E2,$DF,$05,$18,$9F,$F4,$7F,$C1,$7E,$40,$E7,$9C,$8F
	.dc.b	$F3,$E4,$7F,$3D,$F2,$9F,$1F,$FC,$FF,$E8,$39,$51,$B3,$A8,$FF,$29
	.dc.b	$4F,$EF,$DF,$F2,$C3,$63,$51,$FE,$B2,$32,$7F,$AD,$CA,$7F,$97,$63
	.dc.b	$C4,$78,$FE,$43,$FA,$2F,$D8,$7F,$AE,$B5,$F8,$3A,$23,$1A,$FC,$16
	.dc.b	$E1,$FC,$43,$BB,$F1,$8D,$39,$F9,$11,$FF,$2A,$61,$D2,$77,$83,$83
	.dc.b	$FD,$E1,$5A,$C7,$E7,$46,$21,$7E,$48,$C3,$F5,$B7,$7C,$33,$E8,$FF
	.dc.b	$70,$DB,$E8,$7F,$26,$BE,$38,$76,$F1,$37,$7F,$4F,$C7,$97,$C0,$FA
	.dc.b	$8F,$D9,$C2,$47,$CD,$E1,$F0,$B6,$3E,$A1,$1F,$F4,$53,$C1,$FF,$47
	.dc.b	$7C,$8F,$F7,$FF,$A2,$7C,$07,$F8,$D3,$E4,$1C,$5F,$0B,$C8,$C3,$70
	.dc.b	$FF,$3F,$4F,$FB,$C2,$F8,$C6,$27,$FC,$7B,$CF,$FD,$D2,$15,$BC,$39
	.dc.b	$C7,$FA,$1E,$F8,$BA,$BE,$3E,$A6,$B9,$3E,$B7,$8B,$78,$FF,$7B,$97
	.dc.b	$9B,$0F,$99,$DD,$9A,$EB,$0C,$9E,$EE,$63,$CC,$7D,$39,$DC,$EA,$33
	.dc.b	$8F,$EB,$B2,$FF,$36,$36,$3F,$CE,$2D,$7F,$91,$35,$FF,$70,$76,$FF
	.dc.b	$87,$1F,$61,$FB,$B9,$F6,$47,$C5,$CB,$D5,$5F,$BC,$F5,$73,$CF,$AF
	.dc.b	$4E,$63,$E4,$C3,$F8,$2B,$D2,$FF,$81,$D7,$08,$D3,$F2,$7C,$97,$D0
	.dc.b	$F8,$3D,$5F,$5E,$06,$92,$FD,$B1,$FE,$22,$CF,$CA,$EE,$3C,$53,$EA
	.dc.b	$F5,$7D,$7E,$B7,$42,$4F,$27,$BF,$A9,$A8,$F5,$47,$C2,$8F,$87,$D9
	.dc.b	$F1,$3D,$49,$5E,$F8,$CB,$48,$7C,$1D,$1F,$15,$F3,$3D,$7F,$69,$27
	.dc.b	$D6,$FC,$0F,$07,$51,$F2,$E3,$E9,$C7,$D4,$F5,$64,$FC,$F7,$EB,$F9
	.dc.b	$7F,$63,$BF,$E4,$92,$7E,$3B,$FE,$B1,$F1,$1F,$53,$FE,$1E,$9F,$AD
	.dc.b	$C4,$AE,$51,$F9,$2A,$DE,$5F,$9F,$F1,$1E,$5D,$25,$7E,$52,$1F,$47
	.dc.b	$F9,$EF,$A1,$FF,$60,$4F,$FB,$C5,$A9,$A6,$3E,$81,$F0,$BF,$B2,$C9
	.dc.b	$F6,$4F,$A3,$C3,$70,$FD,$38,$BE,$2E,$3E,$29,$F2,$9E,$26,$C7,$FB
	.dc.b	$4D,$AF,$FB,$61,$AF,$F0,$FF,$D5,$5C,$9F,$C1,$3B,$91,$89,$FD,$BC
	.dc.b	$62,$75,$3F,$DF,$E7,$CB,$F0,$1F,$E0,$07,$FC,$28,$D7,$FD,$9E,$D4
	.dc.b	$7F,$92,$BC,$9F,$93,$D5,$FC,$AA,$3F,$DD,$AD,$7F,$C3,$CD,$7F,$DB
	.dc.b	$2D,$7F,$BB,$7B,$FF,$AC,$23,$FA,$A7,$FD,$E9,$F4,$4F,$B8,$7C,$EE
	.dc.b	$1B,$5F,$EF,$C9,$D4,$E4,$FB,$BF,$0F,$FE,$66,$D7,$FC,$2B,$E5,$FF
	.dc.b	$BA,$A3,$16,$E4,$FE,$61,$18,$6F,$87,$FC,$CA,$75,$5F,$E3,$97,$FA
	.dc.b	$E9,$83,$FC,$70,$53,$FD,$BE,$53,$FB,$59,$BC,$5F,$6D,$FD,$AF,$D1
	.dc.b	$F5,$36,$7F,$C8,$D3,$AF,$47,$FC,$0D,$08,$C5,$7F,$10,$F8,$4C,$F9
	.dc.b	$BF,$E1,$E6,$F8,$46,$9C,$AF,$FD,$21,$F2,$F7,$E1,$CD,$26,$1E,$F9
	.dc.b	$7D,$27,$E7,$E3,$0D,$8F,$88,$D7,$FC,$BC,$D7,$FC,$65,$83,$D4,$FE
	.dc.b	$A5,$C5,$FC,$99,$B7,$AE,$7B,$7F,$E3,$B9,$AB,$E8,$5E,$CF,$AD,$B7
	.dc.b	$E7,$6C,$FD,$76,$BF,$E7,$1D,$5F,$3D,$DB,$97,$F7,$7F,$35,$DA,$73
	.dc.b	$87,$F0,$FF,$09,$3E,$2F,$B5,$7C,$97,$CB,$6D,$7C,$76,$E5,$F8,$76
	.dc.b	$BF,$AA,$3F,$D8,$3E,$4F,$F3,$29,$FE,$2D,$E8,$7C,$0E,$67,$DA,$7F
	.dc.b	$83,$27,$27,$F2,$1F,$C7,$3F,$8F,$BF,$D7,$7E,$5B,$D5,$18,$C7,$F9
	.dc.b	$4C,$73,$E6,$3F,$84,$25,$7F,$58,$7F,$13,$FA,$DF,$EF,$A3,$FB,$DB
	.dc.b	$6B,$FB,$42,$F9,$5F,$33,$9E,$8F,$EE,$6F,$F8,$92,$B8,$F6,$EB,$E0
	.dc.b	$76,$C8,$FF,$01,$CF,$E7,$F0,$F8,$9A,$A3,$11,$FE,$79,$6B,$FE,$E8
	.dc.b	$5F,$31,$E2,$3E,$47,$C7,$3E,$53,$18,$7A,$A7,$CF,$E8,$E3,$0F,$5B
	.dc.b	$EA,$F9,$8F,$5F,$C2,$DB,$F8,$2E,$1F,$3B,$F1,$88,$FA,$B9,$FE,$1F
	.dc.b	$CA,$7C,$46,$C7,$CA,$8F,$9A,$47,$E0,$BF,$43,$E3,$FE,$59,$EA,$F0
	.dc.b	$2C,$FD,$28,$F9,$FE,$C6,$33,$FA,$6F,$D2,$F6,$3E,$9F,$C5,$5B,$FB
	.dc.b	$2E,$1F,$9E,$FD,$E2,$3F,$BD,$F1,$3F,$D3,$AF,$F8,$EB,$AD,$F4,$9F
	.dc.b	$B0,$9C,$E3,$FE,$29,$9C,$D6,$EB,$F3,$11,$FA,$68,$C1,$CF,$C0,$E5
	.dc.b	$E7,$F3,$F9,$FB,$3F,$D9,$B1,$46,$63,$C5,$EA,$5C,$BC,$0F,$F0,$2E
	.dc.b	$AF,$E2,$F3,$FE,$7F,$6C,$FF,$A4,$3E,$1F,$F5,$D1,$FD,$BA,$75,$E6
	.dc.b	$7F,$98,$7E,$37,$FC,$87,$3C,$27,$FD,$E8,$DB,$E3,$5F,$E6,$79,$CB
	.dc.b	$63,$F7,$59,$3E,$87,$A6,$38,$6C,$FF,$68,$C3,$CB,$8F,$F9,$4A,$F8
	.dc.b	$63,$FE,$9E,$F2,$BF,$C3,$3F,$1D,$C3,$C0,$FE,$03,$F2,$7C,$A7,$35
	.dc.b	$FE,$A4,$9C,$B7,$89,$5C,$5B,$1D,$79,$BF,$D5,$EE,$E7,$B7,$C8,$79
	.dc.b	$37,$FF,$5C,$FA,$D8,$E1,$F1,$6E,$8A,$4C,$3C,$57,$C7,$83,$FD,$6D
	.dc.b	$3F,$01,$79,$7F,$C6,$5F,$A8,$7F,$82,$9C,$70,$FD,$2F,$2F,$84,$63
	.dc.b	$3F,$E4,$CF,$1C,$7F,$94,$5F,$47,$E9,$3C,$DA,$E5,$FF,$5F,$1E,$3E
	.dc.b	$2F,$CB,$C7,$88,$D1,$5C,$E1,$FE,$EA,$F1,$3F,$35,$3E,$8E,$1F,$E2
	.dc.b	$9F,$57,$F5,$67,$35,$FE,$F0,$F1,$93,$F6,$5B,$8F,$D7,$6E,$1C,$EF
	.dc.b	$8E,$7F,$C7,$6F,$A7,$44,$C3,$EF,$7B,$CF,$8A,$C6,$3C,$6A,$E7,$FF
	.dc.b	$00,$2B,$57,$F8,$5C,$BF,$56,$FE,$37,$E8,$7F,$9D,$57,$FB,$C5,$B9
	.dc.b	$FF,$8F,$3A,$1F,$42,$F5,$BF,$21,$5A,$D3,$FC,$4C,$FF,$43,$C7,$FA
	.dc.b	$D5,$A3,$13,$CD,$AF,$03,$EB,$C6,$DF,$8D,$F1,$FD,$9B,$5F,$03,$E3
	.dc.b	$7E,$57,$D2,$7E,$8F,$64,$62,$74,$7C,$D9,$D4,$66,$BE,$C5,$3E,$37
	.dc.b	$EA,$7D,$27,$36,$E1,$F3,$9B,$5C,$1B,$BB,$7F,$8D,$F4,$7F,$16,$FE
	.dc.b	$77,$D1,$FE,$D6,$31,$87,$DD,$8C,$70,$7E,$1D,$B5,$C0,$93,$B7,$F8
	.dc.b	$15,$72,$C3,$F2,$55,$F0,$7E,$22,$57,$E1,$3E,$42,$E1,$F9,$5F,$A3
	.dc.b	$7C,$97,$EE,$38,$3F,$85,$46,$2F,$F8,$47,$83,$F2,$9F,$C6,$FC,$8F
	.dc.b	$CD,$3E,$77,$C5,$58,$FC,$8F,$C0,$FD,$13,$E9,$78,$16,$3F,$2B,$C3
	.dc.b	$F9,$FF,$8D,$FA,$CD,$8F,$E2,$DF,$19,$F8,$9F,$08,$D3,$FC,$15,$1A
	.dc.b	$DF,$CF,$9F,$EF,$29,$AF,$FB,$06,$BF,$C2,$54,$9C,$D6,$EB,$52,$AF
	.dc.b	$F0,$5C,$34,$F8,$7A,$7C,$46,$CF,$F2,$B9,$FF,$33,$3D,$B0,$F0,$D7
	.dc.b	$52,$46,$71,$FF,$3B,$71,$FE,$17,$1F,$24,$63,$97,$FD,$4F,$87,$C0
	.dc.b	$F6,$BF,$9E,$4E,$6B,$7C,$18,$F6,$B2,$DC,$BE,$4B,$67,$F8,$95,$71
	.dc.b	$C3,$E5,$DF,$DE,$E1,$98,$7A,$F8,$2E,$1F,$85,$9F,$C1,$FB,$F8,$D7
	.dc.b	$D1,$8F,$CE,$8C,$6B,$E8,$DF,$FC,$8F,$97,$C5,$C3,$8B,$F7,$3D,$BF
	.dc.b	$59,$EC,$F6,$EC,$64,$FC,$0F,$B7,$8D,$3D,$2D,$D3,$C3,$BF,$F9,$99
	.dc.b	$B8,$7C,$69,$FE,$87,$E2,$7F,$93,$EF,$F9,$C7,$92,$8C,$5F,$D1,$9F
	.dc.b	$E9,$FF,$47,$C3,$E0,$3F,$DA,$F3,$9B,$E1,$18,$DF,$FA,$6E,$1C,$9F
	.dc.b	$E9,$F7,$C7,$0E,$F3,$97,$FA,$C3,$5E,$FA,$73,$73,$75,$FE,$C3,$3B
	.dc.b	$C3,$C3,$1E,$AB,$ED,$8F,$96,$B4,$6E,$33,$8F,$9B,$0E,$98,$72,$BF
	.dc.b	$65,$6B,$0F,$09,$E8,$AD,$C6,$63,$C7,$1C,$B0,$ED,$8D,$89,$3C,$4F
	.dc.b	$34,$62,$FC,$D3,$71,$F8,$F3,$6B,$C3,$5A,$78,$9C,$DC,$12,$78,$C6
	.dc.b	$5F,$73,$4F,$F0,$36,$1E,$9C,$7E,$04,$EA,$BE,$0D,$F8,$6F,$DA,$7F
	.dc.b	$BD,$6B,$E7,$BF,$E0,$86,$BD,$8F,$F0,$5F,$C6,$7F,$28,$FA,$3F,$F6
	.dc.b	$F3,$63,$FD,$FC,$D7,$7D,$A9,$FF,$72,$53,$E8,$DE,$46,$1B,$7E,$36
	.dc.b	$BE,$36,$BF,$E1,$E7,$79,$7E,$5D,$B9,$E6,$BF,$03,$F0,$3F,$EC,$9A
	.dc.b	$D4,$7E,$12,$B7,$7F,$E2,$D6,$BF,$E3,$2A,$E1,$7C,$E3,$F5,$73,$AC
	.dc.b	$3D,$57,$DA,$7A,$31,$DB,$96,$93,$0F,$1B,$FB,$6C,$3B,$BE,$39,$E5
	.dc.b	$3A,$3F,$69,$3A,$9F,$F2,$3D,$F8,$C9,$5D,$EB,$86,$1E,$96,$D7,$F9
	.dc.b	$2E,$F9,$61,$EA,$8F,$25,$6C,$FF,$25,$CE,$B0,$EB,$3A,$39,$1E,$69
	.dc.b	$E0,$FA,$4F,$61,$93,$17,$E2,$7E,$71,$93,$D7,$85,$C3,$FD,$A9,$16
	.dc.b	$FC,$35,$F0,$A3,$15,$BC,$3C,$CE,$23,$C2,$79,$CE,$A5,$BE,$AD,$BE
	.dc.b	$B8,$7C,$08,$E5,$87,$9E,$B6,$DA,$DC,$FC,$23,$95,$F8,$DE,$37,$F1
	.dc.b	$1F,$79,$A3,$C5,$87,$A1,$C5,$FB,$5B,$3C,$70,$F0,$B7,$0F,$0C,$F4
	.dc.b	$AF,$26,$1E,$18,$E2,$7D,$0A,$F6,$38,$3E,$1C,$F7,$38,$BE,$43,$D2
	.dc.b	$71,$3D,$CE,$8F,$89,$87,$E2,$8F,$53,$88,$FD,$06,$9F,$C2,$E5,$1F
	.dc.b	$E1,$43,$E1,$C6,$5C,$63,$EE,$7B,$DF,$A2,$FF,$AA,$7C,$47,$79,$FC
	.dc.b	$F4,$E4,$FF,$50,$C6,$63,$F0,$33,$ED,$9C,$C7,$03,$A9,$0F,$F5,$8C
	.dc.b	$F1,$9E,$11,$9A,$4E,$6B,$C1,$39,$8C,$3B,$C3,$F5,$B1,$88,$FA,$D3
	.dc.b	$C1,$EE,$78,$CC,$9C,$CF,$EB,$31,$C6,$BE,$24,$70,$3F,$BE,$C6,$6B
	.dc.b	$BC,$70,$3A,$4F,$03,$D2,$E4,$F8,$B5,$E3,$8C,$C7,$43,$E9,$D7,$D8
	.dc.b	$7F,$C8,$C7,$CA,$29,$FE,$D5,$75,$87,$F4,$97,$C7,$1C,$CF,$F6,$47
	.dc.b	$B1,$8C,$7E,$DC,$78,$DF,$F7,$F4,$67,$0E,$D1,$A9,$E2,$7F,$81,$F9
	.dc.b	$95,$B8,$F1,$19,$7F,$93,$46,$A3,$04,$3F,$DB,$9F,$5B,$FE,$5F,$7F
	.dc.b	$C2,$4D,$7B,$92,$BF,$C5,$11,$96,$F8,$1F,$02,$7E,$34,$F1,$CB,$FD
	.dc.b	$E2,$D8,$CC,$ED,$FD,$1C,$73,$C7,$31,$E2,$9F,$E8,$7D,$6F,$49,$EB
	.dc.b	$61,$7D,$25,$89,$7F,$EE,$98,$D1,$1C,$1F,$EE,$DF,$57,$AF,$E0,$5F
	.dc.b	$6B,$5F,$F2,$EA,$F9,$5C,$9A,$DA,$5F,$F8,$E7,$35,$71,$D7,$FC,$F2
	.dc.b	$D9,$FF,$48,$2E,$9F,$EB,$BF,$20,$F4,$1A,$7E,$36,$FF,$1D,$39,$8C
	.dc.b	$47,$F0,$88,$C3,$86,$BC,$57,$E9,$5F,$94,$F8,$AF,$CB,$C6,$9B,$7E
	.dc.b	$0A,$FE,$27,$1A,$AE,$1A,$7A,$BB,$FF,$41,$3F,$A0,$D7,$DF,$D7,$8A
	.dc.b	$F7,$B7,$0F,$7D,$DA,$F8,$AD,$D7,$FC,$E5,$3E,$0B,$EC,$D3,$E3,$F5
	.dc.b	$6F,$A7,$E0,$C7,$CA,$6E,$1F,$41,$B7,$FB,$89,$FD,$B4,$FD,$4F,$A2
	.dc.b	$E5,$1D,$63,$83,$65,$3E,$5F,$A2,$E6,$49,$FA,$35,$F7,$A3,$BF,$C2
	.dc.b	$6E,$79,$47,$B6,$3D,$DD,$9E,$FF,$8A,$95,$DE,$32,$D2,$0D,$FE,$76
	.dc.b	$4F,$84,$F9,$1F,$0F,$EE,$25,$79,$25,$1F,$0B,$E1,$F7,$3E,$D8,$FA
	.dc.b	$11,$F5,$7C,$27,$AD,$FE,$4C,$C7,$F8,$DC,$60,$D4,$67,$8B,$FA,$BF
	.dc.b	$88,$FC,$89,$E8,$7D,$FC,$DF,$4C,$F8,$DF,$AF,$18,$6C,$66,$3E,$DB
	.dc.b	$5F,$D0,$5B,$C1,$7B,$9D,$6B,$5A,$FF,$2E,$6E,$7E,$4A,$DE,$1F,$89
	.dc.b	$D7,$ED,$E7,$EF,$E7,$F3,$CE,$5A,$FC,$FB,$F6,$EA,$F3,$7C,$1F,$1B
	.dc.b	$E8,$25,$7A,$1F,$C9,$ED,$E7,$F6,$3F,$40,$D9,$FC,$D3,$5E,$C7,$F9
	.dc.b	$FB,$F6,$D7,$C3,$7D,$36,$C7,$A5,$AF,$0D,$51,$8E,$0F,$AE,$BE,$0B
	.dc.b	$F5,$9F,$0B,$2F,$F3,$8F,$EB,$FE,$03,$E5,$7C,$57,$D2,$24,$FD,$27
	.dc.b	$F5,$3F,$19,$FB,$52,$47,$E9,$7E,$9F,$EF,$7B,$3F,$D2,$63,$F8,$9B
	.dc.b	$5F,$E3,$0D,$7F,$B3,$35,$FF,$0F,$F8,$CF,$80,$FC,$2B,$63,$F7,$AD
	.dc.b	$7F,$8F,$B5,$FF,$57,$FC,$57,$E9,$5F,$F5,$3F,$CC,$FF,$73,$F6,$93
	.dc.b	$9E,$14,$E6,$30,$DE,$1F,$F1,$8C,$EA,$FF,$98,$75,$E2,$BF,$97,$B6
	.dc.b	$32,$78,$0B,$3C,$57,$DD,$FE,$63,$AB,$F8,$BB,$63,$F9,$CB,$5F,$E7
	.dc.b	$CD,$7F,$C1,$9F,$33,$FD,$72,$7F,$8F,$B6,$3F,$A0,$B5,$FE,$8C,$D7
	.dc.b	$FA,$7F,$D4,$7F,$85,$32,$FE,$12,$DA,$FE,$A7,$F2,$16,$BF,$A9,$CD
	.dc.b	$5F,$20,$E2,$FF,$0D,$C7,$F8,$7E,$31,$7F,$CC,$9B,$1F,$D5,$1A,$FF
	.dc.b	$30,$F8,$7F,$F6,$DC,$FF,$2F,$6C,$7E,$8B,$CE,$47,$D6,$9D,$6B,$E5
	.dc.b	$1F,$B6,$3B,$B0,$D9,$FD,$F2,$F7,$73,$3E,$DE,$FF,$E5,$5D,$57,$80
	.dc.b	$EB,$0F,$DA,$C6,$1F,$EA,$1A,$7C,$2D,$FE,$0B,$FE,$76,$D9,$18,$8D
	.dc.b	$4F,$C3,$8D,$E8,$FC,$2B,$E2,$E1,$FC,$B1,$B5,$FC,$AB,$F5,$9F,$E4
	.dc.b	$CF,$F1,$78,$CE,$0C,$FF,$A1,$78,$3F,$57,$37,$1F,$AE,$30,$CC,$3E
	.dc.b	$A3,$E5,$9D,$C6,$0F,$F0,$4E,$6F,$C4,$7F,$8F,$47,$F2,$76,$E3,$FC
	.dc.b	$5B,$16,$6F,$97,$FA,$95,$AD,$7F,$C0,$5C,$6F,$A9,$FC,$67,$2F,$DC
	.dc.b	$F7,$AC,$E7,$4F,$C1,$46,$0F,$9B,$3B,$8F,$ED,$B2,$9C,$B6,$33,$F5
	.dc.b	$3D,$0E,$BF,$53,$63,$F1,$D1,$E0,$FA,$9C,$6F,$F8,$FB,$47,$93,$E6
	.dc.b	$7C,$BE,$05,$C3,$F2,$CE,$1F,$1B,$E2,$23,$E6,$F8,$AF,$6F,$C9,$FB
	.dc.b	$9B,$1F,$9F,$8F,$49,$1F,$6B,$EC,$FA,$8F,$8A,$7C,$6F,$4E,$CF,$E9
	.dc.b	$E3,$E4,$FA,$B8,$D3,$D4,$FE,$E7,$E0,$DF,$83,$FF,$B2,$AF,$FC,$51
	.dc.b	$18,$F9,$ED,$4E,$6F,$FA,$8C,$E6,$B7,$2A,$FC,$FC,$E6,$BF,$C9,$53
	.dc.b	$9A,$FF,$27,$36,$3F,$CA,$DC,$7F,$BA,$F9,$7F,$F5,$1F,$91,$F6,$65
	.dc.b	$39,$BF,$F3,$37,$17,$EE,$67,$FB,$F7,$2F,$FB,$8A,$B5,$18,$38,$CB
	.dc.b	$0E,$B5,$B8,$FE,$01,$7D,$27,$35,$C3,$C4,$B8,$7F,$0E,$F3,$2E,$1F
	.dc.b	$D4,$70,$F1,$FC,$76,$26,$47,$9B,$E5,$6E,$3E,$3C,$3F,$8B,$C7,$68
	.dc.b	$7C,$7F,$AF,$5F,$F5,$1F,$91,$DA,$B5,$39,$DB,$FD,$47,$3F,$EC,$8C
	.dc.b	$7F,$D4,$6D,$C3,$31,$F5,$DF,$F5,$1C,$65,$B1,$FE,$BF,$3A,$C7,$8F
	.dc.b	$34,$52,$1F,$EB,$A8,$B3,$F3,$44,$F0,$94,$67,$E4,$3F,$D4,$11,$DC
	.dc.b	$CE,$5F,$76,$53,$A7,$87,$C6,$FC,$58,$C4,$EE,$FB,$56,$AF,$9E,$3F
	.dc.b	$EE,$BB,$DB,$70,$F3,$E1,$DE,$73,$7E,$66,$BE,$29,$DD,$71,$BE,$B3
	.dc.b	$B9,$F9,$4E,$E3,$E7,$34,$FE,$BF,$5E,$B8,$FF,$59,$3D,$8F,$F5,$C3
	.dc.b	$D2,$38,$5F,$97,$E3,$7E,$05,$FC,$16,$D6,$EF,$8D,$FC,$13,$E7,$4E
	.dc.b	$9B,$3B,$AE,$4D,$39,$4E,$27,$67,$3B,$F1,$4E,$89,$19,$C3,$E7,$57
	.dc.b	$1A,$DC,$E4,$F9,$91,$99,$DD,$78,$67,$BD,$70,$24,$FC,$F8,$CB,$C1
	.dc.b	$A7,$D1,$9F,$1C,$F0,$8E,$F1,$D6,$31,$1F,$8C,$AF,$13,$C4,$E6,$66
	.dc.b	$79,$D7,$86,$B8,$E1,$F3,$A7,$EC,$DF,$E0,$23,$17,$E5,$9C,$9F,$57
	.dc.b	$D2,$7C,$8F,$49,$FC,$74,$61,$FD,$23,$70,$FC,$A3,$71,$F8,$11,$8A
	.dc.b	$E8,$4A,$F4,$E0,$BE,$11,$DA,$75,$5E,$A9,$CC,$FE,$B9,$B5,$C6,$BC
	.dc.b	$D3,$97,$8D,$FE,$AE,$B7,$8F,$EC,$DE,$24,$9E,$37,$CE,$FD,$B1,$C9
	.dc.b	$FA,$A6,$09,$87,$B5,$F5,$D7,$37,$38,$7C,$23,$C5,$5C,$E3,$CC,$70
	.dc.b	$71,$7E,$1B,$F8,$11,$8B,$F4,$E3,$63,$78,$FA,$DC,$61,$CC,$F0,$9F
	.dc.b	$2C,$C5,$F9,$C9,$8F,$96,$35,$18,$6C,$65,$F3,$C6,$5B,$19,$C7,$D0
	.dc.b	$FB,$A3,$C2,$7E,$54,$B1,$A7,$A9,$F3,$27,$A4,$66,$3A,$99,$9E,$93
	.dc.b	$AC,$7E,$1C,$F4,$7E,$84,$7C,$88,$C5,$FD,$C9,$CD,$72,$3D,$0E,$27
	.dc.b	$C6,$79,$EB,$B5,$7C,$EC,$7E,$6C,$79,$B1,$FA,$15,$D8,$F9,$B8,$75
	.dc.b	$79,$E3,$FC,$7A,$B9,$C7,$C2,$AE,$31,$83,$D0,$7E,$2E,$3F,$97,$E3
	.dc.b	$FA,$F6,$E3,$FD,$51,$A5,$3F,$AB,$FA,$6E,$CF,$F3,$44,$64,$46,$7C
	.dc.b	$7C,$46,$27,$6F,$39,$D3,$C7,$27,$C5,$BC,$9E,$31,$86,$C6,$67,$83
	.dc.b	$5F,$E9,$2E,$73,$FE,$B4,$DD,$7F,$B0,$36,$7F,$AE,$7C,$5F,$BE,$B7
	.dc.b	$7F,$D2,$6B,$55,$FD,$C1,$B8,$2F,$86,$1D,$2F,$81,$2F,$FB,$E4,$E9
	.dc.b	$FE,$D5,$7C,$89,$3F,$D5,$E7,$53,$DE,$39,$61,$E4,$8E,$66,$2F,$C0
	.dc.b	$FF,$82,$6F,$CB,$7C,$B0,$F9,$13,$9C,$38,$3F,$00,$A7,$73,$47,$62
	.dc.b	$E1,$E1,$26,$1F,$E0,$9B,$ED,$87,$AA,$B7,$39,$7A,$E1,$E2,$BE,$31
	.dc.b	$E1,$36,$7C,$4C,$37,$87,$47,$8E,$1E,$C9,$B1,$A6,$9C,$8F,$F1,$2C
	.dc.b	$F1,$71,$7E,$1C,$3C,$91,$8B,$F4,$61,$FE,$1F,$BE,$0F,$9A,$7E,$23
	.dc.b	$97,$BB,$9B,$FC,$14,$62,$FE,$9C,$EE,$3B,$61,$E7,$69,$E4,$BE,$F5
	.dc.b	$C0,$91,$F1,$6B,$B6,$1E,$69,$F4,$C6,$2F,$A3,$6B,$A1,$E3,$3E,$61
	.dc.b	$F6,$DD,$1E,$9B,$F0,$C6,$A3,$0F,$13,$E0,$D7,$80,$F2,$B8,$38,$39
	.dc.b	$69,$0D,$1F,$50,$E9,$87,$D3,$7D,$D1,$D2,$31,$1C,$1F,$9A,$FA,$63
	.dc.b	$58,$7E,$0E,$30,$77,$9E,$27,$0A,$FA,$47,$C0,$AF,$34,$F8,$CF,$A7
	.dc.b	$1A,$8C,$CF,$2A,$ED,$39,$AE,$D1,$8A,$DB,$F5,$63,$D6,$43,$D4,$ED
	.dc.b	$C5,$75,$C3,$F8,$04,$43,$C7,$5E,$2C,$3E,$AC,$7C,$37,$F7,$86,$2B
	.dc.b	$DE,$E2,$3E,$6B,$5F,$F3,$6B,$FC,$EF,$7F,$F3,$6B,$63,$F4,$71,$87
	.dc.b	$C5,$8F,$BD,$FF,$47,$FC,$5F,$38,$C5,$FC,$DC,$3F,$D1,$53,$97,$D1
	.dc.b	$5F,$E7,$18,$C3,$F0,$30,$FE,$29,$12,$3B,$CF,$D0,$8C,$47,$DC,$3E
	.dc.b	$B5,$FF,$A4,$E3,$38,$7F,$A1,$9F,$54,$FF,$A5,$5A,$FA,$0F,$F4,$7E
	.dc.b	$08,$EB,$5C,$E7,$FD,$19,$C4,$F8,$8F,$E6,$2D,$8F,$51,$0F,$F5,$D3
	.dc.b	$96,$C7,$02,$3F,$E9,$7E,$57,$57,$F4,$1E,$87,$C8,$BE,$43,$0B,$E4
	.dc.b	$69,$88,$98,$7F,$AB,$2B,$8C,$62,$3F,$DC,$D4,$F9,$1F,$E9,$DE,$0B
	.dc.b	$E4,$7A,$57,$F7,$58,$C6,$3F,$DD,$1B,$19,$9D,$BF,$DF,$E3,$97,$CB
	.dc.b	$FF,$80,$FC,$FF,$F6,$FC,$A7,$8A,$8C,$C6,$1B,$4A,$E5,$C7,$FE,$E0
	.dc.b	$D1,$E1,$3E,$5B,$39,$25,$FB,$BF,$F8,$47,$2F,$0C,$62,$76,$F2,$AB
	.dc.b	$7C,$E7,$FD,$DB,$5C,$23,$11,$E7,$C3,$C5,$18,$9E,$B8,$7F,$88,$DB
	.dc.b	$E3,$FF,$86,$B1,$F1,$B6,$4D,$31,$F3,$1E,$48,$CE,$3E,$2C,$7E,$04
	.dc.b	$62,$3C,$47,$BD,$4F,$1C,$7B,$12,$7F,$C5,$D8,$7C,$48,$F1,$7C,$5E
	.dc.b	$E7,$B1,$CE,$3C,$51,$E3,$F9,$6F,$27,$CB,$CA,$DD,$FC,$E9,$D4,$63
	.dc.b	$14,$62,$3C,$F3,$F5,$3E,$2F,$DD,$F1,$FF,$0C,$93,$E6,$B4,$7C,$7F
	.dc.b	$8C,$DC,$F5,$8F,$8D,$1F,$80,$F0,$5F,$15,$F4,$77,$5F,$F9,$0E,$30
	.dc.b	$DF,$1F,$F0,$FE,$3B,$E3,$7E,$67,$27,$E5,$A9,$F1,$ED,$BF,$CA,$71
	.dc.b	$86,$C6,$63,$FC,$AE,$D7,$FC,$C1,$6F,$0B,$FC,$C1,$1F,$38,$E4,$59
	.dc.b	$E1,$7F,$A6,$AE,$1E,$4F,$B7,$D2,$FF,$2B,$4F,$EA,$B6,$46,$36,$F7
	.dc.b	$C6,$0F,$6E,$8E,$5F,$D2,$E7,$EC,$E1,$F4,$91,$F4,$BF,$E7,$0C,$FD
	.dc.b	$1E,$A7,$C6,$D8,$D3,$E8,$7C,$73,$E7,$E9,$30,$FC,$5E,$69,$D4,$63
	.dc.b	$F1,$AF,$CB,$BF,$87,$F8,$7F,$D0,$11,$E3,$F8,$77,$B5,$FB,$7C,$6E
	.dc.b	$7F,$A2,$48,$FA,$2F,$FA,$23,$1F,$07,$C3,$79,$89,$3E,$4C,$9F,$23
	.dc.b	$F1,$3E,$77,$F8,$74,$FF,$A4,$1A,$FF,$A5,$5A,$FF,$A6,$DA,$FF,$A7
	.dc.b	$FE,$43,$F9,$E3,$FE,$9F,$6B,$9F,$37,$F1,$31,$88,$FF,$51,$FB,$FF
	.dc.b	$C8,$5F,$F4,$C7,$C8,$7F,$AA,$7D,$64,$E7,$44,$E6,$30,$DE,$BF,$C0
	.dc.b	$BF,$EA,$BE,$5B,$87,$FA,$C5,$AF,$FA,$5C,$EE,$64,$A3,$E3,$FF,$A6
	.dc.b	$75,$FF,$AC,$9B,$1F,$EB,$76,$BF,$EB,$D6,$BF,$EC,$1F,$A6,$7C,$CF
	.dc.b	$EA,$E3,$0D,$7F,$6B,$1A,$AF,$F5,$FB,$63,$FD,$91,$9B,$E4,$1E,$F7
	.dc.b	$F4,$AD,$3F,$45,$18,$C3,$F1,$EF,$F6,$86,$CE,$DF,$F6,$A7,$C6,$BE
	.dc.b	$43,$C5,$8A,$31,$3F,$ED,$06,$C7,$FB,$55,$AF,$FB,$77,$C8,$FE,$9B
	.dc.b	$F1,$BD,$63,$F9,$93,$73,$FE,$11,$F3,$79,$AD,$E1,$E6,$F8,$4D,$68
	.dc.b	$78,$7E,$1A,$F9,$BC,$C7,$5F,$13,$FB,$37,$6D,$E1,$3A,$C7,$F9,$CC
	.dc.b	$E9,$FE,$1D,$BB,$E4,$BD,$ED,$8F,$74,$60,$FA,$9B,$3E,$47,$F9,$04
	.dc.b	$FE,$E9,$BC,$3F,$D4,$30,$7C,$93,$8B,$E0,$56,$B4,$FE,$34,$7F,$82
	.dc.b	$F0,$FE,$41,$F6,$3F,$EF,$57,$CF,$19,$C7,$FA,$D8,$9D,$38,$6B,$FC
	.dc.b	$F0,$FE,$7A,$E1,$C3,$5C,$E5,$EC,$FB,$67,$CB,$7D,$56,$E1,$F2,$33
	.dc.b	$FE,$06,$FF,$18,$8C,$DF,$F5,$8F,$4F,$F8,$F4,$FF,$5D,$E0,$FB,$91
	.dc.b	$89,$D4,$60,$FC,$F7,$2F,$F8,$49,$B3,$F6,$2B,$FA,$F4,$63,$C3,$E6
	.dc.b	$74,$FA,$2B,$3F,$DA,$E3,$BF,$B9,$8C,$7C,$8F,$8B,$DC,$F1,$FB,$D6
	.dc.b	$FF,$B9,$38,$7F,$B6,$3E,$22,$3F,$DB,$FE,$07,$D7,$F0,$1E,$C6,$C7
	.dc.b	$F7,$B8,$F4,$91,$FE,$EC,$FB,$BF,$6C,$47,$E7,$DA,$FF,$81,$97,$C6
	.dc.b	$E6,$21,$9F,$8D,$B9,$FF,$81,$A9,$1F,$AD,$F8,$37,$C8,$7F,$87,$AF
	.dc.b	$FC,$2F,$18,$F6,$35,$39,$C3,$FC,$4B,$39,$AD,$CF,$F6,$37,$FC,$B8
	.dc.b	$D9,$FF,$2C,$35,$FC,$74,$F1,$7F,$20,$7F,$93,$7E,$07,$F7,$15,$FE
	.dc.b	$6D,$6C,$7F,$7F,$F8,$BF,$F2,$2F,$C4,$FF,$98,$3C,$C4,$7F,$92,$3C
	.dc.b	$3C,$37,$0E,$5F,$1F,$8C,$3B,$55,$E9,$FF,$2F,$CE,$65,$8F,$CE,$7F
	.dc.b	$D1,$B7,$9F,$8A,$FE,$49,$18,$8F,$F4,$B9,$C7,$E3,$3E,$9B,$83,$FC
	.dc.b	$4F,$9F,$9A,$7F,$C5,$33,$A7,$D8,$63,$8B,$91,$E3,$C3,$C5,$5C,$65
	.dc.b	$5C,$63,$0F,$3B,$F9,$6D,$9F,$DE,$35,$ED,$1F,$EB,$2B,$F4,$70,$F8
	.dc.b	$23,$CF,$AF,$08,$C5,$79,$8F,$0B,$8C,$3C,$2F,$2A,$E3,$3E,$E8,$D5
	.dc.b	$73,$9C,$CF,$26,$E1,$F9,$46,$E1,$CE,$7C,$13,$96,$DF,$47,$B5,$72
	.dc.b	$8E,$18,$7C,$4B,$B8,$78,$6F,$FB,$46,$1C,$F0,$F8,$F8,$7F,$AA,$9A
	.dc.b	$7F,$B5,$9F,$B9,$F1,$EF,$84,$ED,$3E,$88,$C5,$F9,$23,$13,$A8,$F4
	.dc.b	$CE,$B2,$F8,$F5,$A9,$E0,$6A,$B8,$4F,$DC,$AE,$18,$FA,$63,$13,$A8
	.dc.b	$FB,$B3,$AA,$F8,$67,$33,$AD,$70,$6C,$F0,$AE,$51,$F9,$4A,$E4,$FD
	.dc.b	$53,$95,$7A,$1B,$5E,$33,$93,$D0,$E5,$5D,$1F,$CB,$D7,$4C,$7E,$D5
	.dc.b	$F8,$CE,$85,$BE,$95,$C8,$E8,$72,$BE,$94,$BE,$98,$FD,$69,$CC,$63
	.dc.b	$1F,$9D,$8F,$5C,$3A,$3E,$0C,$7D,$2E,$63,$B1,$E0,$C7,$B6,$1C,$AB
	.dc.b	$67,$9A,$77,$5C,$6B,$71,$C7,$0F,$80,$4B,$F2,$9C,$E3,$C7,$7C,$EB
	.dc.b	$8D,$F3,$BF,$55,$7D,$03,$64,$3E,$B4,$6E,$FA,$46,$2F,$CE,$EE,$FA
	.dc.b	$C7,$7B,$F2,$4F,$53,$81,$B7,$37,$F1,$E3,$17,$DA,$FE,$DD,$F6,$9F
	.dc.b	$86,$D8,$CC,$70,$3B,$38,$76,$66,$BB,$38,$24,$7A,$1F,$CE,$C7,$BD
	.dc.b	$E7,$8F,$E0,$5E,$B8,$7C,$89,$D1,$F5,$67,$4E,$2F,$E1,$61,$EB,$AE
	.dc.b	$11,$E9,$7E,$0B,$AB,$EB,$3F,$69,$A7,$DF,$79,$19,$AE,$D1,$E8,$9E
	.dc.b	$93,$F2,$6F,$C7,$18,$9E,$47,$4A,$F8,$78,$7C,$18,$F8,$78,$79,$9F
	.dc.b	$91,$5F,$3E,$F9,$C6,$0F,$01,$4F,$96,$E9,$FC,$F9,$FD,$22,$7F,$8D
	.dc.b	$B7,$0F,$EE,$AF,$EC,$CF,$EE,$FE,$A3,$E0,$DE,$46,$A5,$39,$8C,$36
	.dc.b	$39,$B5,$FE,$AF,$F1,$EF,$17,$96,$88,$C4,$F8,$A3,$13,$A9,$CC,$F2
	.dc.b	$8C,$11,$E1,$1A,$7F,$B1,$B5,$FE,$E6,$D7,$FB,$6B,$67,$BC,$72,$9E
	.dc.b	$8F,$73,$04,$30,$43,$BB,$C1,$A6,$CA,$6C,$F1,$BB,$69,$B2,$9B,$3C
	.dc.b	$EE,$DA,$6C,$A6,$E3,$D6,$D9,$F7,$12,$79,$4E,$A7,$CA,$43,$C2,$53
	.dc.b	$A9,$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$27,$C9,$1E,$B8,$C4,$F9,$A7
	.dc.b	$33,$EB,$8C,$11,$E1,$1A,$7C,$CD,$7C,$6D,$7C,$4D,$9E,$F1,$CA,$7A
	.dc.b	$3D,$CC,$10,$C1,$0E,$EF,$06,$9B,$29,$B3,$C6,$ED,$A6,$CA,$6C,$F3
	.dc.b	$BB,$69,$B2,$9B,$8F,$5B,$67,$DC,$49,$E5,$3A,$9F,$29,$0F,$09,$4E
	.dc.b	$A5,$78,$3A,$21,$A2,$BA,$70,$47,$04,$9F,$24,$7A,$E3,$13,$E6,$9C
	.dc.b	$CF,$AE,$30,$47,$84,$69,$F3,$35,$F1,$B5,$F1,$36,$7B,$C7,$29,$E8
	.dc.b	$F7,$30,$43,$04,$3B,$BC,$1A,$6C,$A6,$CF,$1B,$B6,$9B,$29,$B3,$CE
	.dc.b	$ED,$A6,$CA,$6E,$3D,$6D,$9F,$71,$27,$94,$EA,$7C,$A4,$3C,$25,$3A
	.dc.b	$96,$BF,$9F,$66,$F6,$13,$27,$31,$88,$FE,$8F,$97,$F4,$A7,$FD,$AB
	.dc.b	$19,$9D,$C7,$19,$D5,$75,$8C,$65,$DA,$31,$1D,$23,$17,$CE,$B8,$BF
	.dc.b	$E5,$89,$FE,$BC,$7F,$4D,$D3,$FB,$0B,$92,$1F,$D8,$4A,$7F,$68,$2E
	.dc.b	$3F,$E4,$07,$99,$0C,$C7,$47,$91,$0E,$45,$8E,$4F,$0D,$3F,$B1,$CF
	.dc.b	$18,$D9,$27,$6F,$18,$D4,$62,$35,$18,$24,$F4,$24,$7C,$16,$BE,$93
	.dc.b	$A3,$87,$C8,$6C,$86,$CE,$EE,$DE,$26,$8E,$26,$E7,$CC,$49,$F2,$C6
	.dc.b	$CE,$26,$DC,$B6,$7B,$3E,$03,$81,$4D,$18,$9E,$D1,$C1,$B1,$99,$EA
	.dc.b	$F9,$88,$78,$8A,$75,$2C,$72,$7B,$90,$CC,$74,$79,$10,$E4,$58,$E4
	.dc.b	$F0,$9F,$34,$F1,$8D,$92,$76,$F1,$8D,$46,$23,$51,$82,$4F,$42,$47
	.dc.b	$C1,$6B,$E9,$3A,$38,$7C,$86,$C8,$6C,$EE,$ED,$E2,$68,$E2,$6E,$7C
	.dc.b	$C4,$9F,$2C,$6C,$E2,$6D,$CB,$67,$B3,$E0,$38,$14,$D1,$89,$ED,$1C
	.dc.b	$1B,$19,$9E,$AF,$98,$87,$88,$A7,$52,$C7,$27,$B9,$0C,$C7,$47,$91
	.dc.b	$0E,$45,$8E,$4F,$09,$F3,$4F,$18,$D9,$27,$6F,$18,$D4,$62,$35,$18
	.dc.b	$24,$F4,$24,$7C,$16,$BE,$93,$A3,$87,$C8,$7C,$EC,$BF,$CF,$4F,$F2
	.dc.b	$DC,$7F,$CF,$F8,$FF,$3B,$8F,$F4,$3B,$5E,$27,$53,$CB,$5D,$5E,$E7
	.dc.b	$8C,$86,$F0,$F6,$13,$0F,$5C,$68,$EA,$43,$99,$C5,$DB,$67,$C0,$F8
	.dc.b	$8E,$47,$03,$91,$93,$81,$A3,$06,$A7,$C9,$1D,$1B,$1B,$6B,$C8,$87
	.dc.b	$19,$F2,$B6,$79,$46,$23,$C8,$43,$C2,$59,$EA,$F1,$3C,$11,$C5,$D9
	.dc.b	$23,$6F,$18,$D4,$62,$35,$18,$24,$F9,$09,$1D,$9A,$FB,$0E,$8E,$1F
	.dc.b	$21,$93,$AB,$97,$67,$32,$1B,$9F,$21,$27,$C7,$1A,$3A,$90,$E6,$71
	.dc.b	$76,$D9,$F0,$3E,$23,$91,$FC,$D7,$0F,$F0,$8B,$FE,$0B,$3F,$C2,$A7
	.dc.b	$F8,$40,$A6,$72,$FF,$08,$C7,$02,$61,$FE,$10,$9E,$13,$FE,$16,$C3
	.dc.b	$FC,$5D,$1F,$E2,$C2,$1F,$E2,$C2,$BA,$70,$47,$04,$D3,$FC,$75,$1F
	.dc.b	$E4,$18,$C6,$9F,$E3,$C9,$CC,$FF,$90,$63,$04,$78,$46,$9F,$F1,$F4
	.dc.b	$F2,$73,$3B,$79,$18,$21,$82,$1C,$A7,$A3,$EC,$6B,$EB,$6C,$F7,$8E
	.dc.b	$26,$E7,$C0,$F3,$6B,$93,$C2,$64,$A6,$4E,$67,$77,$8B,$4D,$94,$DC
	.dc.b	$78,$DE,$86,$27,$CE,$49,$E7,$3B,$9F,$01,$C0,$86,$08,$77,$32,$53
	.dc.b	$2F,$37,$04,$70,$49,$F1,$47,$AA,$31,$3E,$49,$CC,$FA,$63,$04,$78
	.dc.b	$46,$9F,$1C,$F2,$7E,$2E,$1F,$D0,$1F,$D1,$1F,$CF,$B3,$FE,$82,$7E
	.dc.b	$B2,$59,$B3,$FF,$41,$EC,$DB,$FD,$05,$5A,$9C,$C6,$34,$6F,$8C,$BF
	.dc.b	$D1,$77,$B5,$87,$37,$38,$F7,$73,$87,$68,$C4,$EE,$54,$C3,$BC,$70
	.dc.b	$39,$E3,$E1,$77,$7C,$B0,$E9,$39,$BE,$58,$DA,$D5,$F2,$3A,$1C,$AB
	.dc.b	$86,$1E,$0C,$7C,$98,$7B,$5B,$7C,$AF,$81,$CA,$7B,$B6,$35,$1C,$71
	.dc.b	$F2,$CE,$EF,$B4,$62,$FA,$E1,$E7,$3C,$78,$DB,$E9,$16,$7C,$B7,$CA
	.dc.b	$7D,$91,$C9,$D9,$DE,$B8,$5F,$72,$47,$DA,$AE,$4F,$B2,$B9,$12,$7F
	.dc.b	$05,$5C,$A7,$BC,$72,$7E,$CD,$72,$C3,$E0,$46,$88,$6A,$B9,$3D,$EB
	.dc.b	$94,$FC,$AB,$F0,$4F,$08,$E6,$F1,$3D,$B1,$A7,$91,$EB,$AE,$B3,$C0
	.dc.b	$97,$E5,$C7,$EA,$93,$1F,$9F,$7D,$31,$F1,$CF,$DA,$AE,$15,$D2,$31
	.dc.b	$85,$8E,$F3,$AB,$E4,$7A,$5C,$63,$E7,$C7,$C7,$7C,$6B,$67,$B6,$75
	.dc.b	$87,$96,$3E,$D1,$AA,$F6,$47,$8A,$30,$EE,$7F,$0B,$5D,$5F,$05,$F8
	.dc.b	$B1,$F6,$E3,$F0,$27,$BB,$EA,$AE,$13,$F5,$2B,$86,$1F,$4F,$0F,$06
	.dc.b	$1C,$23,$C9,$3A,$7E,$99,$31,$FB,$97,$CF,$0E,$D3,$C1,$B7,$E6,$6C
	.dc.b	$EE,$74,$49,$D3,$63,$51,$82,$38,$24,$7D,$39,$EA,$D3,$91,$67,$8B
	.dc.b	$67,$8B,$EB,$BF,$75,$FB,$23,$95,$9F,$D3,$9F,$C6,$8F,$CA,$14,$FC
	.dc.b	$D3,$B7,$37,$F6,$1F,$E1,$9C,$1F,$AB,$B9,$5F,$5F,$0F,$E6,$F3,$8A
	.dc.b	$FC,$2D,$F1,$3F,$8C,$C6,$5C,$4F,$91,$CD,$73,$AF,$E7,$75,$BA,$EB
	.dc.b	$39,$9F,$E6,$B3,$97,$71,$9C,$13,$FC,$E2,$3D,$93,$C1,$F8,$86,$27
	.dc.b	$6E,$23,$B9,$F1,$63,$2F,$C2,$8F,$E8,$1C,$3F,$D4,$9F,$E7,$D5,$FC
	.dc.b	$62,$31,$3B,$3F,$9E,$B8,$C3,$DF,$18,$AE,$07,$F2,$0E,$27,$A5,$88
	.dc.b	$FE,$BD,$CA,$E7,$73,$FF,$2F,$8C,$73,$D9,$CD,$70,$51,$9A,$93,$9C
	.dc.b	$EE,$5F,$D7,$7A,$1D,$2E,$AF,$C3,$4A,$73,$18,$6D,$FC,$E9,$D5,$71
	.dc.b	$8C,$4F,$4F,$89,$FE,$D7,$E5,$7F,$6C,$8E,$36,$8C,$5A,$B7,$7F,$9B
	.dc.b	$BD,$D7,$F7,$4D,$5D,$BF,$42,$30,$FF,$62,$6D,$72,$C5,$5C,$39,$70
	.dc.b	$F2,$9B,$7C,$DF,$D5,$47,$6E,$3C,$C7,$0F,$19,$DF,$F1,$A9,$18,$9F
	.dc.b	$89,$18,$9D,$4E,$71,$FC,$AC,$F0,$F0,$3A,$57,$56,$DF,$EE,$F3,$F8
	.dc.b	$18,$3C,$1E,$64,$9E,$F1,$96,$F8,$5E,$79,$F4,$36,$3A,$1E,$E9,$78
	.dc.b	$9E,$DF,$13,$DC,$D9,$DE,$9F,$2A,$B7,$18,$D3,$E0,$D7,$C5,$F1,$78
	.dc.b	$CE,$69,$E3,$7C,$22,$47,$A1,$F8,$BF,$03,$F3,$5F,$9C,$D8,$F0,$F8
	.dc.b	$AF,$21,$E5,$7D,$D8,$C3,$F7,$FC,$57,$95,$E6,$DB,$F1,$D8,$7F,$25
	.dc.b	$BD,$BE,$C8,$79,$69,$FD,$0C,$61,$B1,$99,$FD,$04,$E5,$FD,$15,$FE
	.dc.b	$6F,$83,$F2,$13,$93,$C2,$23,$53,$E1,$AE,$93,$AB,$79,$8F,$86,$F7
	.dc.b	$36,$FE,$9E,$1E,$EF,$3B,$35,$E0,$F4,$1E,$87,$A9,$B7,$E8,$6E,$5F
	.dc.b	$B9,$6B,$FD,$0F,$0F,$DA,$70,$FC,$BE,$8F,$B3,$7F,$4B,$D3,$FE,$1F
	.dc.b	$7E,$CB,$DC,$E7,$C4,$FA,$95,$AB,$FF,$35,$EC,$F4,$DE,$9E,$AF,$F5
	.dc.b	$9B,$CC,$7E,$FE,$EF,$53,$1A,$FC,$A1,$BB,$D5,$F7,$12,$FD,$CF,$F1
	.dc.b	$AD,$1B,$7F,$2E,$9F,$57,$97,$E6,$F8,$DF,$3B,$F6,$3D,$77,$B1,$F8
	.dc.b	$26,$D7,$E0,$1B,$9F,$F2,$16,$BF,$D4,$3D,$9F,$D9,$3F,$C9,$5A,$E7
	.dc.b	$C3,$FC,$6C,$62,$3F,$95,$7A,$3F,$95,$F6,$AC,$7E,$4D,$BA,$9A,$7F
	.dc.b	$79,$F5,$DE,$DF,$F2,$BA,$E0,$7F,$80,$E3,$3D,$1D,$63,$13,$BB,$E9
	.dc.b	$18,$9D,$2F,$71,$EE,$7E,$82,$30,$D8,$CC,$7E,$91,$AF,$F4,$5F,$75
	.dc.b	$E5,$71,$6C,$7E,$A9,$AE,$E3,$F5,$B8,$3D,$EF,$E1,$33,$FC,$09,$AF
	.dc.b	$E9,$DB,$8F,$E3,$9F,$E3,$71,$FB,$DF,$80,$7C,$07,$A5,$B1,$FB,$A6
	.dc.b	$BF,$D6,$DA,$FF,$89,$7C,$DF,$E7,$BF,$05,$FE,$27,$C3,$F9,$9B,$79
	.dc.b	$7F,$95,$36,$BF,$85,$7C,$17,$F0,$B7,$D1,$E6,$67,$4F,$63,$74,$39
	.dc.b	$F8,$3F,$EC,$55,$FB,$EE,$57,$C2,$75,$F1,$BF,$DB,$ED,$8C,$CF,$63
	.dc.b	$E0,$7C,$2F,$BC,$8F,$BD,$F9,$BF,$0F,$E5,$8F,$E2,$0D,$DF,$F6,$ED
	.dc.b	$DF,$FC,$99,$8B,$E1,$FF,$C0,$D7,$E5,$F8,$7F,$80,$B8,$FF,$8C,$3F
	.dc.b	$47,$C6,$F3,$4E,$A7,$33,$F3,$E7,$EA,$11,$FF,$06,$70,$BE,$23,$FD
	.dc.b	$F5,$1F,$C9,$5B,$A7,$F0,$C6,$E7,$E6,$C9,$F1,$7F,$4F,$9B,$FB,$AF
	.dc.b	$3D,$F5,$5C,$34,$C4,$5C,$FF,$B5,$C7,$F3,$EA,$FE,$FA,$DA,$E3,$3F
	.dc.b	$DC,$3E,$2B,$A1,$1F,$EC,$1B,$57,$FA,$0D,$FF,$81,$BE,$2F,$C4,$67
	.dc.b	$6F,$EF,$BF,$05,$E5,$F8,$59,$7F,$DA,$62,$BF,$E0,$8C,$BF,$C8,$8F
	.dc.b	$F8,$02,$7C,$87,$97,$E0,$BD,$DB,$49,$FE,$E5,$15,$E5,$1E,$B2,$3F
	.dc.b	$E1,$89,$7C,$67,$C9,$6C,$7F,$7F,$6B,$EC,$BF,$A2,$FF,$8B,$A3,$38
	.dc.b	$FF,$86,$23,$E8,$1C,$EB,$F4,$DF,$06,$F6,$7F,$80,$D7,$F9,$15,$FF
	.dc.b	$0F,$57,$F8,$42,$75,$E4,$F0,$9D,$F2,$7F,$91,$5D,$D7,$FA,$02,$B6
	.dc.b	$FF,$94,$A3,$2F,$FA,$7E,$32,$FF,$8C,$BE,$35,$F0,$1F,$EA,$3E,$D7
	.dc.b	$C5,$7F,$90,$A7,$35,$FE,$74,$6A,$9D,$BF,$E4,$E8,$FF,$22,$BB,$AE
	.dc.b	$95,$CF,$2B,$3B,$6F,$1A,$FA,$E1,$D2,$9F,$0F,$FC,$B2,$30,$DC,$3F
	.dc.b	$D7,$4D,$9F,$F6,$17,$C4,$BD,$2E,$F1,$CB,$CB,$BF,$14,$F8,$EA,$6B
	.dc.b	$0E,$54,$9D,$3F,$EC,$DC,$BC,$91,$88,$F8,$59,$7F,$B2,$E3,$13,$E6
	.dc.b	$C3,$AD,$7F,$AB,$CF,$F6,$B4,$F1,$E1,$FF,$5A,$63,$FE,$D3,$6C,$EC
	.dc.b	$F1,$47,$A3,$E3,$E4,$78,$F1,$F0,$4F,$68,$F1,$9D,$A7,$6E,$0E,$67
	.dc.b	$13,$45,$3E,$03,$C3,$1F,$7C,$DA,$F7,$4D,$AF,$5C,$D7,$89,$CF,$04
	.dc.b	$F9,$6F,$E4,$BB,$8C,$CE,$EB,$CF,$87,$FB,$5E,$BE,$24,$F8,$0F,$89
	.dc.b	$5C,$1F,$F6,$C4,$7D,$CF,$8F,$FE,$A1,$87,$D6,$6E,$3C,$6B,$E1,$5F
	.dc.b	$1A,$D7,$85,$E7,$BD,$D7,$C4,$37,$87,$59,$D1,$ED,$7D,$53,$C2,$7F
	.dc.b	$10,$DB,$ED,$87,$87,$0E,$F5,$E2,$AD,$5F,$2C,$3A,$CF,$80,$EC,$62
	.dc.b	$B8,$9F,$29,$C9,$CE,$35,$7D,$C9,$1F,$90,$BF,$21,$E0,$78,$D7,$49
	.dc.b	$FC,$D4,$E6,$7C,$47,$52,$4F,$85,$CB,$C5,$ED,$5D,$2B,$C3,$1F,$12
	.dc.b	$3C,$AF,$CD,$9E,$0E,$4E,$B3,$C3,$05,$FA,$2F,$8D,$7E,$38,$D9,$E0
	.dc.b	$29,$9B,$EB,$5C,$A7,$8B,$F5,$27,$84,$EA,$BE,$03,$6B,$78,$7E,$06
	.dc.b	$BA,$DF,$37,$E0,$E1,$F2,$2B,$AE,$1C,$A3,$13,$C2,$30,$F8,$23,$38
	.dc.b	$79,$B0,$EC,$F8,$67,$A9,$B2,$1F,$1D,$FA,$46,$CE,$4F,$8A,$FD,$2D
	.dc.b	$3D,$33,$DA,$BC,$24,$8F,$A8,$D7,$B3,$70,$F9,$11,$83,$C6,$FC,$3B
	.dc.b	$F1,$CE,$F0,$FB,$15,$EA,$38,$12,$B8,$4F,$DF,$9D,$D7,$AA,$BE,$46
	.dc.b	$1E,$A6,$9E,$A3,$D9,$1B,$3E,$15,$79,$F0,$F7,$4F,$9A,$FE,$8B,$E1
	.dc.b	$7B,$CF,$86,$F9,$B6,$7B,$DF,$B1,$E5,$3E,$38,$EF,$1F,$0A,$FB,$BD
	.dc.b	$67,$F4,$4F,$03,$D0,$74,$AF,$33,$FC,$00,$1F,$88,$29,$F9,$D2,$9F
	.dc.b	$A7,$2F,$A2,$F1,$7F,$DD,$D1,$A1,$19,$F1,$BF,$DF,$2D,$8D,$35,$EB
	.dc.b	$9B,$E3,$BF,$CA,$31,$FD,$75,$AF,$F5,$A6,$BF,$D9,$DB,$1C,$7E,$3F
	.dc.b	$FA,$14,$1F,$E8,$9F,$1A,$DF,$31,$2F,$F9,$AD,$6D,$FE,$6F,$5B,$73
	.dc.b	$3B,$DB,$FB,$9D,$6D,$FE,$EB,$5B,$73,$3B,$D3,$D4,$F8,$B0,$C6,$1E
	.dc.b	$B8,$AF,$F8,$06,$F9,$57,$08,$F0,$92,$FB,$DF,$0C,$7E,$04,$E6,$BF
	.dc.b	$C0,$97,$CA,$3B,$19,$2D,$F2,$70,$7C,$BC,$3A,$4F,$18,$C1,$AB,$EF
	.dc.b	$5C,$4F,$7C,$67,$1F,$89,$87,$8B,$1F,$34,$62,$7C,$33,$97,$86,$3E
	.dc.b	$0A,$DB,$63,$27,$87,$1F,$26,$3E,$2B,$F8,$D8,$F8,$8F,$23,$82,$57
	.dc.b	$80,$FF,$11,$D7,$2C,$7C,$91,$EE,$AE,$55,$E5,$6C,$71,$35,$19,$3A
	.dc.b	$B8,$35,$6B,$ED,$87,$B2,$7F,$1B,$8D,$C3,$AC,$7E,$06,$31,$87,$4C
	.dc.b	$39,$46,$27,$AE,$1D,$1F,$8B,$7F,$01,$F0,$E3,$EC,$BF,$8D,$1F,$02
	.dc.b	$F9,$3E,$73,$C0,$72,$C3,$C1,$87,$5C,$3C,$B7,$CA,$B7,$18,$C3,$B1
	.dc.b	$DE,$B8,$5F,$A6,$31,$87,$83,$0E,$47,$B3,$0F,$B9,$1F,$7F,$0E,$F8
	.dc.b	$7E,$C6,$3F,$82,$B6,$7F,$11,$18,$3F,$C6,$B1,$F8,$43,$F0,$BF,$19
	.dc.b	$63,$F6,$13,$F4,$0F,$F1,$9C,$65,$C5,$7C,$97,$FC,$E5,$19,$8C,$5F
	.dc.b	$9D,$FF,$39,$C6,$71,$FC,$15,$7C,$D3,$AD,$F8,$2B,$8E,$1E,$48,$C5
	.dc.b	$70,$AF,$96,$FF,$9E,$EF,$A4,$73,$BF,$0C,$62,$36,$66,$78,$57,$48
	.dc.b	$FF,$3F,$CE,$F0,$34,$93,$B9,$F7,$B6,$76,$D9,$DC,$FC,$6A,$EE,$FF
	.dc.b	$25,$3B,$CF,$F9,$9A,$7E,$B4,$60,$D1,$D8,$F8,$A7,$FA,$17,$8F,$F9
	.dc.b	$A4,$FF,$9E,$67,$E3,$4E,$B0,$FF,$58,$CE,$1F,$A2,$EC,$FF,$45,$57
	.dc.b	$29,$E1,$7F,$EB,$69,$E0,$D7,$84,$78,$5F,$E3,$D3,$C6,$BF,$D8,$5C
	.dc.b	$AF,$6B,$35,$FC,$8E,$30,$F8,$0C,$46,$9F,$F4,$E4,$6A,$FE,$0B,$63
	.dc.b	$FB,$5B,$63,$FB,$7B,$5F,$F6,$5C,$F4,$9F,$F5,$47,$43,$BD,$F1,$F9
	.dc.b	$BF,$F5,$8C,$E7,$E0,$7F,$A0,$3B,$9B,$49,$B1,$86,$5F,$69,$FF,$70
	.dc.b	$B6,$38,$FC,$87,$2E,$BB,$3B,$6E,$3E,$2B,$57,$83,$2C,$E9,$FE,$F8
	.dc.b	$EF,$7C,$83,$C5,$FC,$2E,$29,$CC,$61,$B9,$FF,$83,$5B,$3F,$E1,$47
	.dc.b	$9B,$FD,$DF,$C3,$FE,$F1,$1A,$6D,$FF,$87,$AB,$FC,$3F,$18,$BE,$11
	.dc.b	$FE,$22,$DF,$E7,$7C,$8E,$31,$47,$D1,$F4,$70,$EA,$B9,$9C,$76,$4E
	.dc.b	$95,$D3,$1E,$95,$AE,$38,$76,$A7,$C9,$7C,$4C,$7F,$00,$D8,$CD,$7F
	.dc.b	$8B,$E7,$2D,$8F,$F1,$A3,$D2,$33,$4C,$FF,$C3,$91,$F8,$BE,$EF,$36
	.dc.b	$1E,$7F,$73,$B1,$30,$E8,$F8,$7E,$4F,$51,$D7,$24,$63,$2E,$D3,$F0
	.dc.b	$FE,$57,$E1,$7C,$AF,$C3,$6C,$71,$70,$F3,$F9,$6B,$5C,$8F,$89,$D9
	.dc.b	$F0,$49,$3E,$F7,$E2,$7C,$BF,$CF,$7E,$83,$63,$E6,$F8,$AF,$8F,$7A
	.dc.b	$9F,$76,$30,$FD,$FF,$1D,$F2,$2F,$8C,$D5,$F0,$C3,$EB,$E6,$F8,$B7
	.dc.b	$C8,$7E,$6A,$30,$D8,$CC,$FF,$99,$67,$2F,$F9,$9F,$1F,$A9,$F2,$1F
	.dc.b	$8F,$9C,$9F,$06,$33,$3E,$0A,$E7,$3F,$17,$47,$CC,$3E,$4F,$D8,$DB
	.dc.b	$FC,$6E,$1E,$CF,$77,$35,$DF,$17,$C7,$F9,$C9,$7E,$78,$CB,$E7,$EC
	.dc.b	$7C,$3F,$C7,$6C,$7C,$EF,$8F,$FE,$41,$1F,$E6,$9C,$BF,$21,$5F,$3E
	.dc.b	$73,$5F,$E9,$0D,$11,$8D,$51,$8F,$23,$87,$8F,$FA,$B8,$97,$F9,$1F
	.dc.b	$8C,$B7,$F8,$73,$EA,$6D,$F4,$49,$3E,$B7,$F5,$9F,$3F,$EC,$9F,$C7
	.dc.b	$78,$48,$C6,$DF,$A6,$7E,$7B,$78,$3E,$B7,$C7,$3D,$BF,$C0,$37,$2F
	.dc.b	$BC,$DC,$7F,$D4,$8D,$7F,$D5,$7F,$34,$F1,$BF,$10,$DA,$FC,$33,$67
	.dc.b	$FD,$58,$D7,$F1,$5E,$67,$C9,$DB,$FD,$60,$D8,$FF,$5B,$B5,$FF,$5D
	.dc.b	$B7,$D7,$78,$DF,$C0,$E7,$E7,$D7,$FB,$02,$73,$AE,$EB,$5C,$2A,$D4
	.dc.b	$61,$FF,$63,$FA,$DF,$D3,$70,$FC,$F3,$5F,$F6,$23,$5F,$F6,$5B,$5F
	.dc.b	$F6,$7F,$C7,$BD,$CF,$D3,$42,$30,$FE,$A5,$AF,$FB,$3D,$AF,$FB,$63
	.dc.b	$8F,$ED,$7C,$DF,$FB,$6B,$3F,$B4,$DD,$FF,$9F,$72,$23,$17,$FC,$FA
	.dc.b	$3F,$54,$FE,$FB,$E2,$BF,$3E,$FE,$E5,$AF,$FB,$7D,$AF,$FB,$99,$AF
	.dc.b	$6F,$8A,$EB,$8F,$E3,$5B,$46,$FC,$E7,$8B,$81,$F1,$CF,$0B,$0D,$C7
	.dc.b	$F8,$DB,$5F,$E0,$4B,$E6,$BD,$7E,$03,$9B,$FC,$35,$5F,$C0,$9A,$FF
	.dc.b	$0D,$6F,$07,$FB,$7F,$1F,$0E,$3F,$12,$9B,$3E,$33,$EA,$B6,$3E,$24
	.dc.b	$60,$FE,$3D,$4F,$9D,$7C,$5F,$B5,$B5,$FC,$D1,$AF,$F1,$FF,$8E,$FE
	.dc.b	$6A,$F9,$FB,$3C,$58,$BC,$6A,$7B,$B5,$FD,$55,$7D,$78,$F2,$91,$FB
	.dc.b	$0F,$93,$D7,$FE,$1B,$1F,$D1,$1A,$FF,$2B,$6E,$DF,$C6,$71,$F2,$BF
	.dc.b	$C0,$FE,$7B,$EC,$FC,$FC,$BF,$ED,$DC,$C9,$C4,$0F,$CA,$42,$7F,$A1
	.dc.b	$73,$FF,$0F,$6C,$66,$7F,$BE,$3F,$9E,$BF,$E2,$5D,$7F,$88,$3F,$13
	.dc.b	$F3,$36,$BC,$3F,$31,$F8,$C3,$0F,$F1,$CB,$7D,$07,$E4,$3E,$6A,$4F
	.dc.b	$A2,$2B,$E7,$3F,$4B,$1E,$42,$3E,$4E,$3F,$CC,$FC,$64,$8F,$83,$15
	.dc.b	$E4,$7E,$AE,$3D,$A4,$7D,$BE,$0F,$E8,$88,$FC,$AF,$8B,$AF,$A7,$E3
	.dc.b	$3F,$4E,$63,$6F,$F1,$EF,$D0,$FF,$A4,$5F,$F1,$FF,$C8,$3E,$83,$F7
	.dc.b	$7F,$0C,$8C,$61,$FE,$2E,$85,$F1,$F1,$BE,$B4,$6E,$FF,$BB,$36,$BF
	.dc.b	$CA,$17,$C9,$FF,$28,$35,$FB,$B5,$FC,$BB,$DC,$FE,$1A,$D8,$FE,$18
	.dc.b	$DE,$6F,$E9,$DF,$1D,$78,$FF,$D5,$5D,$8C,$3F,$8A,$FC,$6E,$4D,$61
	.dc.b	$C7,$E4,$1F,$23,$49,$9F,$FA,$1A,$3C,$D8,$73,$8F,$F2,$CC,$70,$75
	.dc.b	$E9,$73,$38,$BF,$C1,$F1,$FF,$49,$FA,$DE,$98,$7C,$72,$30,$D8,$E6
	.dc.b	$7F,$A8,$23,$F1,$59,$7F,$A7,$9B,$1D,$0F,$F3,$91,$4F,$F5,$2B,$DF
	.dc.b	$E3,$FF,$9E,$E5,$E9,$94,$63,$0E,$13,$98,$C4,$FE,$BD,$AF,$0E,$EF
	.dc.b	$3D,$F5,$7D,$9F,$20,$EA,$A6,$6B,$F7,$18,$78,$67,$51,$8C,$7F,$4D
	.dc.b	$18,$8F,$90,$7E,$FD,$C4,$F9,$6F,$AC,$F1,$35,$78,$7F,$DC,$51,$67
	.dc.b	$C1,$36,$FF,$69,$8F,$B3,$0F,$06,$5F,$AE,$6E,$3E,$4B,$F9,$EF,$63
	.dc.b	$E7,$94,$F3,$88,$74,$FC,$43,$C1,$1B,$C7,$E1,$46,$63,$D4,$71,$AD
	.dc.b	$BC,$E3,$4F,$6A,$F8,$78,$F9,$63,$8C,$78,$1B,$3C,$A3,$11,$C6,$39
	.dc.b	$DF,$C4,$8D,$90,$D1,$D2,$F7,$7F,$5A,$AC,$FA,$E3,$0F,$1A,$F6,$BC
	.dc.b	$4C,$46,$9E,$A7,$AE,$B9,$BE,$F3,$67,$53,$F0,$55,$AB,$F5,$DF,$A6
	.dc.b	$79,$BE,$06,$DF,$C5,$69,$C8,$EA,$7C,$F3,$E0,$61,$F8,$1B,$FB,$F1
	.dc.b	$E3,$11,$97,$F0,$26,$6B,$F3,$9D,$3F,$D8,$9B,$E4,$76,$C5,$39,$C7
	.dc.b	$F3,$ED,$BF,$CF,$35,$E9,$1A,$AF,$BB,$39,$AE,$71,$D6,$30,$47,$07
	.dc.b	$02,$57,$C2,$3F,$2D,$1B,$7F,$18,$F5,$C7,$F7,$6E,$70,$F5,$CF,$1B
	.dc.b	$F0,$BC,$2F,$AC,$FD,$13,$64,$C3,$D5,$3C,$6F,$A4,$62,$78,$34,$93
	.dc.b	$F1,$2B,$D6,$F7,$24,$F0,$74,$62,$78,$BF,$28,$CC,$78,$DB,$1F,$93
	.dc.b	$8C,$47,$C4,$9F,$5C,$E4,$E7,$7F,$02,$B6,$FE,$4A,$33,$7E,$CB,$FC
	.dc.b	$71,$E3,$3B,$90,$F2,$1D,$CD,$16,$BD,$8D,$9E,$B7,$F0,$A7,$C3,$5F
	.dc.b	$5E,$BC,$E7,$5B,$F3,$5F,$A6,$31,$3E,$57,$C3,$3A,$73,$7D,$6B,$BC
	.dc.b	$7D,$78,$EA,$F8,$CF,$6D,$F8,$A3,$35,$F0,$E7,$99,$EC,$70,$F1,$23
	.dc.b	$E1,$7C,$77,$E5,$8C,$1C,$0A,$E7,$0F,$CA,$D7,$82,$3E,$15,$77,$74
	.dc.b	$FE,$80,$FE,$78,$FF,$95,$1F,$C9,$98,$8F,$F1,$6B,$FE,$4D,$F5,$1F
	.dc.b	$14,$F3,$73,$29,$CC,$61,$B1,$C9,$AF,$F4,$BE,$E7,$C8,$F2,$CD,$18
	.dc.b	$9F,$0C,$62,$75,$39,$9E,$51,$FC,$FA,$38,$69,$FD,$AF,$A7,$F8,$AD
	.dc.b	$7F,$45,$9C,$ED,$FD,$B1,$B1,$FD,$AD,$AF,$F5,$79,$E3,$5F,$DD,$67
	.dc.b	$38,$78,$EB,$A6,$1F,$D3,$AF,$78,$75,$8F,$21,$27,$FA,$B4,$E9,$F0
	.dc.b	$46,$70,$EC,$F3,$DF,$17,$CE,$BB,$9C,$A3,$2F,$23,$35,$F0,$63,$31
	.dc.b	$FD,$C7,$1F,$49,$EA,$C3,$C5,$19,$C3,$78,$FB,$70,$E5,$3D,$31,$F1
	.dc.b	$47,$9A,$3C,$46,$4F,$13,$93,$89,$63,$89,$23,$FC,$35,$7E,$2A,$F0
	.dc.b	$4E,$67,$C8,$76,$AE,$4F,$68,$D4,$AB,$96,$1F,$73,$AB,$FD,$F9,$F2
	.dc.b	$DF,$3D,$FF,$7F,$7C,$2D,$F5,$93,$63,$0D,$D7,$FC,$55,$5C,$9B,$4A
	.dc.b	$DC,$62,$BA,$7C,$B6,$34,$FF,$1B,$E0,$FA,$2F,$F2,$6D,$F4,$5F,$2B
	.dc.b	$E5,$73,$5C,$1C,$CE,$FA,$35,$5B,$6C,$64,$95,$C7,$E5,$75,$1E,$8F
	.dc.b	$97,$F8,$1E,$A7,$C1,$6C,$64,$F6,$C6,$5A,$66,$BD,$7B,$2B,$E5,$7C
	.dc.b	$86,$E7,$D1,$F2,$DF,$B0,$BF,$F5,$F6,$BF,$06,$31,$3F,$ED,$78,$C4
	.dc.b	$FF,$92,$1B,$1F,$B8,$0F,$C9,$EF,$F8,$D5,$F1,$DB,$5F,$2F,$E1,$3A
	.dc.b	$DF,$0D,$D1,$8C,$3E,$93,$63,$35,$E1,$CF,$FD,$0F,$CF,$FE,$9B,$B4
	.dc.b	$63,$9F,$D0,$FF,$18,$AF,$C1,$E3,$F8,$DF,$9C,$70,$7E,$3F,$C4,$7C
	.dc.b	$E6,$63,$0E,$F7,$98,$7E,$3A,$FF,$03,$A7,$85,$4F,$5D,$A5,$6F,$7B
	.dc.b	$9F,$E1,$FB,$BF,$3D,$C4,$F9,$4F,$C6,$46,$1F,$37,$AF,$30,$F0,$7C
	.dc.b	$C5,$C3,$F2,$FD,$8F,$80,$FD,$43,$6B,$FD,$19,$E1,$BE,$8F,$DF,$1F
	.dc.b	$A6,$9D,$5F,$FA,$5A,$B5,$C7,$CB,$0F,$E4,$11,$8B,$E1,$18,$C3,$9D
	.dc.b	$F8,$A3,$17,$C6,$BE,$37,$BB,$F3,$7C,$FF,$A0,$D8,$C9,$FE,$D5,$8C
	.dc.b	$B4,$F9,$DA,$FD,$5F,$A2,$FB,$0D,$8C,$9F,$ED,$D8,$CB,$4F,$83,$F4
	.dc.b	$9F,$83,$CB,$FD,$BF,$F0,$5E,$C8,$C7,$63,$A9,$39,$AF,$CA,$EE,$BF
	.dc.b	$F6,$AE,$09,$93,$97,$73,$A7,$2F,$86,$35,$3F,$53,$E9,$7E,$4F,$57
	.dc.b	$FB,$CE,$73,$7F,$73,$E1,$7E,$D9,$F9,$7F,$99,$B8,$7D,$98,$B3,$FC
	.dc.b	$63,$B1,$58,$C3,$F4,$9A,$3E,$17,$F3,$C7,$E1,$BE,$06,$E5,$F8,$36
	.dc.b	$E3,$F8,$C6,$DF,$FB,$7E,$7F,$A8,$F8,$AF,$81,$FE,$85,$18,$6B,$FE
	.dc.b	$8C,$F3,$3F,$D1,$ED,$3F,$7F,$87,$FB,$35,$B1,$A8,$FF,$62,$47,$F5
	.dc.b	$EE,$E7,$B5,$D2,$7F,$9A,$C3,$E2,$7F,$52,$7F,$47,$F9,$07,$B1,$8C
	.dc.b	$3F,$A1,$C6,$1F,$E8,$B5,$C5,$CC,$EE,$FF,$A0,$FD,$33,$E4,$FF,$A9
	.dc.b	$36,$BF,$57,$F2,$4F,$6B,$35,$FA,$68,$C4,$FF,$52,$AE,$17,$FA,$D9
	.dc.b	$CB,$DA,$BE,$5F,$C3,$4B,$FF,$74,$E5,$E7,$AF,$F7,$9F,$02,$BE,$26
	.dc.b	$7F,$2A,$B5,$7F,$93,$BF,$CA,$46,$1C,$35,$CC,$F2,$9E,$FE,$5D,$9E
	.dc.b	$FD,$1F,$E3,$2F,$82,$FF,$19,$F7,$36,$7C,$66,$71,$F0,$6A,$C3,$18
	.dc.b	$FF,$42,$BD,$D7,$E9,$F1,$FE,$8D,$6F,$86,$DD,$A3,$95,$6C,$EB,$3A
	.dc.b	$8C,$CA,$BC,$B7,$F9,$3B,$E1,$3E,$1E,$8F,$1E,$7E,$9E,$7B,$69,$E9
	.dc.b	$F0,$B8,$9B,$87,$2B,$E0,$DE,$B7,$D0,$7F,$5F,$CB,$FB,$24,$62,$7F
	.dc.b	$CE,$35,$FC,$2E,$B5,$5F,$D9,$2B,$4F,$D7,$31,$F2,$DE,$68,$CE,$9F
	.dc.b	$41,$46,$6B,$5F,$1B,$BE,$7B,$97,$95,$B8,$FC,$58,$F2,$CF,$C2,$F3
	.dc.b	$1F,$15,$AC,$FB,$7C,$67,$9B,$5F,$E4,$4D,$D4,$D3,$1E,$EF,$9D,$F4
	.dc.b	$37,$2F,$8D,$EE,$71,$23,$F1,$29,$1D,$7D,$DF,$80,$FC,$16,$CF,$CF
	.dc.b	$F6,$FE,$27,$B9,$F1,$5B,$19,$26,$FE,$F3,$FC,$23,$B3,$E0,$F8,$4F
	.dc.b	$D9,$EE,$F6,$11,$F2,$3E,$FF,$9D,$FA,$4F,$D3,$6C,$7E,$06,$FE,$B1
	.dc.b	$F5,$FE,$06,$C7,$E1,$FE,$9F,$FC,$83,$3F,$77,$85,$18,$8F,$F1,$6B
	.dc.b	$5F,$F1,$63,$57,$9A,$FA,$87,$3F,$E6,$AB,$AF,$2F,$FA,$9A,$B5,$39
	.dc.b	$F8,$2F,$CF,$F9,$7F,$CA,$33,$FC,$D3,$72,$F8,$D9,$FE,$EE,$77,$E5
	.dc.b	$FE,$CB,$EA,$7F,$CC,$D1,$FE,$C6,$6D,$7F,$B3,$1A,$FE,$65,$AF,$F1
	.dc.b	$1F,$80,$7C,$47,$F0,$2D,$1F,$55,$FA,$FC,$3F,$6B,$7F,$E2,$6A,$D4
	.dc.b	$E6,$3E,$D0,$FA,$6F,$CB,$D7,$FB,$61,$B1,$FE,$DE,$6B,$FA,$B6,$BF
	.dc.b	$33,$C6,$F9,$07,$C9,$F7,$2D,$78,$BE,$AD,$F1,$DF,$B6,$B4,$62,$7F
	.dc.b	$DC,$ED,$8F,$F7,$6B,$5F,$DD,$7C,$C7,$F3,$76,$F9,$FC,$E3,$11,$FC
	.dc.b	$D5,$AF,$F9,$9B,$9B,$FC,$8F,$7F,$C2,$1B,$1F,$89,$6B,$F8,$8D,$FF
	.dc.b	$85,$FC,$8B,$EB,$3F,$CF,$97,$FF,$00,$23,$1D,$9A,$9C,$DF,$F0,$F9
	.dc.b	$CD,$7F,$11,$8F,$E2,$5E,$BF,$F8,$6B,$E6,$30,$D8,$C9,$F9,$48,$CB
	.dc.b	$4F,$F1,$17,$C7,$7F,$73,$EC,$E1,$08,$C3,$B3,$1E,$CF,$F3,$97,$11
	.dc.b	$AE,$75,$FF,$8B,$72,$7D,$6F,$F4,$7B,$F7,$35,$F6,$35,$FE,$C5,$1C
	.dc.b	$3D,$3F,$74,$7C,$FF,$8A,$BD,$BF,$B4,$6F,$11,$A7,$09,$FF,$45,$FB
	.dc.b	$57,$E3,$3F,$C0,$11,$D1,$AF,$CB,$6B,$FC,$F7,$D8,$FF,$46,$BE,$26
	.dc.b	$BF,$45,$AE,$FA,$BC,$FC,$2F,$AE,$7C,$5F,$F7,$D8,$FE,$E9,$18,$AF
	.dc.b	$EB,$F1,$B9,$D7,$2F,$F6,$FA,$7C,$5E,$2F,$FC,$B1,$F0,$5F,$E9,$E7
	.dc.b	$E2,$B5,$FC,$23,$5F,$EA,$CD,$7F,$C1,$DF,$5E,$F3,$FA,$73,$7E,$FB
	.dc.b	$C0,$BE,$1F,$EB,$2F,$F7,$F7,$B7,$F8,$0F,$D8,$7C,$D4,$D2,$3F,$5B
	.dc.b	$C8,$FE,$42,$FF,$1A,$F2,$BD,$5C,$E9,$CC,$FF,$1C,$7F,$B3,$BF,$E0
	.dc.b	$2C,$9F,$62,$FA,$9E,$3D,$3F,$EC,$EA,$FE,$71,$F2,$3E,$AA,$D4,$60
	.dc.b	$FD,$A4,$7E,$18,$8F,$ED,$F9,$39,$91,$FE,$85,$C5,$DB,$E3,$BB,$E3
	.dc.b	$27,$FA,$34,$57,$F7,$D1,$F9,$22,$3F,$C0,$FE,$47,$CE,$47,$FC,$29
	.dc.b	$D5,$FE,$64,$7F,$A8,$47,$B3,$E3,$A4,$60,$FE,$A9,$E1,$FC,$0F,$8A
	.dc.b	$91,$83,$FA,$C3,$FA,$22,$3F,$C4,$DF,$D2,$FD,$4F,$C8,$FA,$89,$3F
	.dc.b	$D7,$E2,$BF,$C5,$E3,$F5,$64,$7F,$8E,$7C,$A7,$D0,$23,$EA,$F9,$0A
	.dc.b	$FF,$54,$E2,$7D,$4F,$D6,$33,$E0,$7F,$94,$A1,$F5,$3F,$6B,$C9,$97
	.dc.b	$FD,$C6,$2B,$FC,$A6,$3F,$7A,$47,$F8,$1F,$D9,$7F,$84,$9F,$F1,$F7
	.dc.b	$9D,$FE,$51,$6F,$A7,$FA,$75,$3B,$7F,$B4,$CE,$DB,$E3,$7E,$52,$3F
	.dc.b	$7B,$97,$F9,$A1,$D2,$F1,$BF,$CE,$DF,$67,$FC,$7A,$FF,$B1,$E0,$F7
	.dc.b	$3C,$0D,$8F,$EC,$F3,$B7,$FB,$2C,$65,$F1,$F6,$78,$4E,$35,$E9,$EF
	.dc.b	$FE,$2B,$E8,$38,$FF,$D0,$D1,$86,$F9,$7F,$E8,$99,$D6,$1F,$ED,$DA
	.dc.b	$FF,$23,$BF,$ED,$F8,$CB,$FB,$19,$FF,$48,$67,$FE,$5B,$AF,$E4,$6D
	.dc.b	$8C,$CF,$63,$FD,$D1,$1F,$E6,$1C,$DD,$5E,$1C,$7E,$2B,$6B,$FD,$DD
	.dc.b	$3A,$7C,$53,$CD,$63,$FE,$E9,$C3,$FC,$EB,$1F,$EB,$EE,$DE,$5E,$13
	.dc.b	$81,$1D,$7C,$2F,$F5,$D7,$C3,$EA,$73,$7C,$48,$76,$CB,$FD,$FF,$8F
	.dc.b	$68,$C0,$C5,$F2,$3F,$DB,$29,$39,$FB,$1F,$C1,$3E,$A3,$1B,$F9,$AB
	.dc.b	$84,$60,$D5,$AB,$8C,$78,$23,$15,$C6,$3E,$41,$FC,$97,$E3,$2F,$09
	.dc.b	$A6,$3E,$8A,$CF,$FB,$DE,$3E,$BB,$F8,$16,$BC,$CD,$1E,$91,$C3,$FE
	.dc.b	$58,$70,$D7,$E2,$9C,$4F,$47,$87,$FC,$E3,$8B,$F1,$F9,$7F,$99,$B8
	.dc.b	$7E,$BF,$67,$19,$CD,$6F,$7F,$F3,$3D,$68,$98,$7C,$69,$D3,$F4,$D7
	.dc.b	$C5,$FE,$09,$B1,$B8,$FC,$2D,$7E,$16,$30,$43,$CF,$1A,$E2,$FC,$47
	.dc.b	$8D,$FD,$EB,$1F,$D3,$E3,$FD,$BB,$AF,$26,$BB,$2E,$3F,$A0,$3F,$07
	.dc.b	$3B,$70,$7F,$A0,$2F,$FC,$0B,$C5,$FE,$FB,$9D,$4F,$E0,$1B,$87,$C8
	.dc.b	$F8,$47,$C9,$FE,$79,$B5,$F8,$DA,$DE,$BF,$92,$E0,$FD,$CC,$FE,$26
	.dc.b	$9F,$4F,$C4,$F8,$71,$0F,$C8,$C6,$9B,$19,$7F,$25,$F4,$BF,$0E,$FB
	.dc.b	$53,$EA,$3C,$87,$EC,$A7,$55,$FA,$76,$E0,$AE,$13,$E0,$AE,$1B,$FE
	.dc.b	$DF,$ED,$3F,$98,$6B,$F8,$4F,$8F,$FB,$86,$78,$66,$1F,$83,$8C,$BF
	.dc.b	$0E,$B8,$E9,$F0,$F9,$3F,$DB,$B3,$B2,$1F,$BC,$DD,$5C,$23,$D1,$EC
	.dc.b	$FA,$C8,$FE,$DB,$77,$BF,$FC,$43,$E0,$3F,$C9,$93,$FB,$AE,$E7,$D2
	.dc.b	$7E,$FC,$D6,$1F,$17,$9E,$57,$37,$E4,$78,$BF,$27,$C3,$93,$DD,$F9
	.dc.b	$9E,$97,$CD,$F4,$7F,$05,$EE,$4A,$F2,$47,$D1,$F4,$A4,$79,$DF,$A9
	.dc.b	$E4,$79,$FD,$3F,$E0,$3D,$8E,$0F,$F1,$1D,$FF,$10,$8C,$4E,$A7,$3D
	.dc.b	$FF,$C7,$AB,$F9,$14,$7F,$37,$7F,$8B,$2C,$7E,$21,$F1,$70,$B3,$FA
	.dc.b	$2C,$3C,$13,$D4,$91,$9E,$3F,$F0,$2B,$F3,$FC,$FF,$C6,$1F,$8D,$F4
	.dc.b	$AC,$7D,$3F,$4B,$F2,$06,$1F,$CB,$CE,$1D,$1C,$CF,$E5,$F9,$2B,$ED
	.dc.b	$12,$33,$AD,$7D,$BE,$9F,$DF,$31,$DD,$FC,$72,$F9,$4F,$49,$D3,$DA
	.dc.b	$FF,$49,$E8,$C8,$FC,$53,$FA,$7E,$DC,$47,$EB,$BD,$2F,$D5,$18,$C7
	.dc.b	$F9,$4E,$BF,$E0,$56,$C6,$A7,$99,$FC,$DF,$CC,$7D,$1F,$F3,$7A,$FE
	.dc.b	$F7,$18,$EF,$D4,$E7,$1F,$F0,$3F,$C6,$7F,$4A,$8C,$61,$FE,$05,$6B
	.dc.b	$FE,$73,$6D,$A7,$85,$F4,$25,$7F,$44,$9D,$38,$C3,$C7,$7C,$8E,$77
	.dc.b	$CA,$78,$46,$6F,$D9,$39,$AE,$31,$CF,$0F,$4D,$D3,$D1,$15,$FF,$45
	.dc.b	$57,$19,$FF,$12,$DF,$59,$F3,$4E,$5E,$27,$8A,$BC,$2F,$3C,$3E,$53
	.dc.b	$93,$89,$9A,$E9,$87,$D0,$BE,$93,$F2,$6F,$94,$62,$F9,$D7,$B2,$B9
	.dc.b	$9D,$5C,$13,$0F,$7C,$78,$4F,$F1,$A6,$1F,$DC,$DF,$6E,$1F,$EE,$4B
	.dc.b	$F3,$C7,$EE,$4F,$F7,$C7,$23,$E8,$FF,$DD,$3F,$07,$FE,$36,$6C,$66
	.dc.b	$78,$1C,$23,$A7,$90,$F8,$DA,$49,$DE,$1F,$CF,$E3,$1A,$7E,$4A,$B8
	.dc.b	$57,$E1,$7E,$8F,$FC,$4F,$39,$AF,$CD,$56,$A3,$15,$BE,$AC,$D4,$8F
	.dc.b	$C7,$56,$F6,$FF,$30,$E1,$D1,$FF,$31,$C6,$67,$FC,$C0,$D7,$FD,$9C
	.dc.b	$DC,$55,$C6,$77,$7D,$F1,$F1,$9F,$AC,$9D,$4F,$29,$D6,$1E,$07,$F6
	.dc.b	$93,$AB,$F8,$95,$E8,$9D,$BF,$E7,$39,$D6,$1E,$32,$4F,$2C,$77,$5D
	.dc.b	$CE,$47,$C5,$2E,$3E,$6C,$7C,$51,$98,$C3,$EC,$37,$1F,$22,$7A,$E3
	.dc.b	$E6,$AD,$D7,$FB,$CF,$1F,$2E,$1A,$C7,$E4,$46,$27,$B6,$3E,$0B,$F8
	.dc.b	$D8,$FB,$BA,$9E,$33,$E4,$3E,$74,$63,$E4,$EC,$E6,$BF,$C8,$11,$8A
	.dc.b	$FF,$36,$C7,$F5,$3A,$E7,$18,$C7,$FC,$00,$D7,$FC,$8C,$D8,$DC,$FF
	.dc.b	$92,$E3,$BF,$C9,$75,$8F,$F3,$9F,$D1,$BC,$CF,$2A,$F9,$1B,$18,$6C
	.dc.b	$FF,$5D,$9E,$0D,$9F,$04,$63,$1F,$ED,$CD,$D3,$FC,$CA,$DD,$7F,$C7
	.dc.b	$4B,$CE,$EB,$5F,$E6,$CF,$4B,$E2,$FC,$2F,$C6,$6E,$5F,$E6,$57,$DD
	.dc.b	$C1,$FE,$A3,$8F,$EE,$0D,$AF,$EF,$CD,$8F,$94,$E7,$2F,$F3,$9D,$7D
	.dc.b	$2F,$93,$ED,$5F,$13,$CF,$F8,$67,$C4,$F3,$6C,$7C,$87,$DE,$47,$E2
	.dc.b	$3F,$0B,$CC,$F9,$67,$DE,$F7,$6C,$FF,$9F,$73,$FC,$09,$F8,$2E,$8B
	.dc.b	$1F,$EA,$6F,$87,$F8,$8F,$F8,$2D,$B3,$FE,$19,$6C,$7E,$15,$B1,$FE
	.dc.b	$65,$7F,$C7,$31,$CF,$1F,$C8,$C7,$D1,$9F,$F5,$F3,$A7,$FD,$64,$B0
	.dc.b	$D9,$C3,$1B,$7F,$88,$EC,$FB,$E4,$7F,$08,$FE,$46,$9E,$6F,$80,$F0
	.dc.b	$F9,$F6,$BF,$CF,$9F,$47,$EC,$E0,$FF,$41,$7D,$1F,$F1,$3F,$09,$39
	.dc.b	$E7,$FF,$39,$CE,$63,$DF,$B2,$B7,$CD,$FA,$36,$E3,$FE,$94,$6C,$FF
	.dc.b	$A4,$6B,$A3,$C5,$B1,$E0,$E8,$ED,$7F,$23,$D1,$98,$7D,$5F,$42,$E3
	.dc.b	$F2,$26,$E9,$FE,$C2,$DD,$58,$D7,$EA,$7A,$1F,$48,$FA,$7E,$75,$CB
	.dc.b	$F0,$11,$72,$FA,$73,$71,$FF,$5A,$FD,$0F,$FA,$C5,$B3,$F1,$FE,$87
	.dc.b	$D3,$39,$C7,$FD,$6B,$18,$9D,$6B,$F3,$6B,$51,$F2,$67,$59,$7F,$AB
	.dc.b	$DB,$1F,$84,$F3,$FF,$8C,$F8,$CF,$9E,$FC,$5C,$61,$B7,$F9,$98,$FC
	.dc.b	$7F,$A1,$8B,$FF,$5B,$7C,$EF,$23,$F2,$7E,$9D,$BF,$D0,$36,$7F,$26
	.dc.b	$DD,$9B,$7F,$9D,$7C,$17,$CC,$7F,$4F,$C3,$FA,$AC,$61,$FF,$23,$18
	.dc.b	$8F,$E4,$D8,$27,$4A,$7F,$C1,$11,$86,$CF,$08,$DD,$F8,$A8,$73,$9F
	.dc.b	$0F,$AF,$DE,$F3,$5B,$F2,$BF,$B7,$FB,$BF,$D0,$BD,$97,$CA,$7F,$44
	.dc.b	$F3,$FF,$DC,$38,$7F,$AB,$63,$15,$B6,$C6,$4F,$F7,$27,$DA,$B8,$3F
	.dc.b	$51,$F1,$8F,$87,$FD,$54,$E6,$BF,$DD,$53,$97,$8D,$68,$DF,$D0,$D8
	.dc.b	$F4,$F1,$7C,$FF,$A1,$F9,$A7,$CE,$FA,$4B,$3A,$E3,$F8,$7C,$89,$F9
	.dc.b	$91,$FE,$2C,$8C,$4F,$F8,$BA,$3E,$94,$E9,$FA,$2D,$7E,$F7,$D2,$72
	.dc.b	$F8,$ED,$5F,$29,$46,$27,$57,$F8,$5B,$B5,$F5,$BC,$4F,$F3,$D4,$FD
	.dc.b	$8B,$F0,$43,$E7,$37,$5F,$D8,$FD,$DB,$39,$BE,$07,$FB,$17,$ED,$BE
	.dc.b	$CB,$8A,$ED,$5C,$A7,$31,$E6,$9F,$F6,$65,$78,$A3,$07,$E2,$A5,$F4
	.dc.b	$5E,$0F,$03,$93,$6F,$B5,$29,$C5,$A9,$CC,$61,$B5,$F0,$BA,$58,$3E
	.dc.b	$8F,$E7,$62,$9C,$C6,$1B,$7F,$E6,$46,$CF,$F9,$B5,$E7,$C7,$D6,$FA
	.dc.b	$57,$C3,$9C,$B6,$33,$5F,$02,$7C,$7C,$DC,$7E,$8F,$93,$6F,$83,$87
	.dc.b	$A7,$DB,$BE,$33,$18,$73,$C5,$F3,$9E,$96,$DF,$A5,$B3,$E9,$6B,$FE
	.dc.b	$2D,$5F,$25,$EB,$AF,$63,$6B,$D5,$F2,$5D,$48,$F5,$7C,$1C,$9F,$0D
	.dc.b	$F8,$8D,$8F,$85,$F0,$DF,$1F,$97,$E4,$36,$32,$4C,$BE,$36,$8E,$0D
	.dc.b	$CF,$CB,$D3,$DC,$47,$DC,$FC,$1F,$03,$EA,$3F,$55,$B1,$F9,$7F,$45
	.dc.b	$F6,$DF,$13,$A7,$F6,$F1,$8E,$0F,$D1,$FD,$17,$CF,$BE,$17,$B7,$F9
	.dc.b	$0E,$89,$7C,$8F,$E3,$23,$0D,$8C,$CF,$F0,$16,$CF,$FA,$C1,$7C,$7F
	.dc.b	$82,$BC,$2D,$AF,$D0,$FB,$BF,$B3,$BF,$D4,$87,$19,$FF,$0B,$C3,$E4
	.dc.b	$57,$0E,$2F,$5F,$3F,$8C,$91,$E3,$7C,$DC,$FF,$00,$FE,$07,$F7,$1F
	.dc.b	$07,$2E,$F8,$7F,$29,$7C,$31,$9C,$7F,$8D,$37,$0F,$8B,$F7,$3F,$E3
	.dc.b	$89,$CF,$4F,$F0,$EF,$A4,$7D,$05,$3F,$88,$ED,$FB,$6C,$3F,$7F,$3A
	.dc.b	$7F,$80,$FC,$97,$F0,$CC,$FF,$DB,$CD,$7E,$9B,$5F,$EA,$AD,$59,$B6
	.dc.b	$F6,$9E,$EE,$7B,$7F,$C3,$BE,$89,$E7,$FD,$B6,$C7,$F3,$96,$E1,$FE
	.dc.b	$E8,$6B,$FE,$F1,$7C,$04,$F3,$39,$1F,$85,$8F,$EC,$A7,$F8,$E7,$E4
	.dc.b	$FF,$DE,$3E,$87,$F6,$CE,$A4,$62,$74,$7F,$BF,$67,$51,$F9,$2F,$91
	.dc.b	$7B,$5F,$D1,$2B,$62,$33,$C9,$C2,$3F,$BC,$C6,$1D,$B7,$0F,$0C,$7F
	.dc.b	$7E,$FB,$AF,$ED,$8F,$F5,$96,$BF,$A2,$35,$D1,$FD,$75,$46,$9D,$3F
	.dc.b	$C5,$9E,$5D,$BB,$8C,$47,$33,$8C,$73,$EB,$FF,$5A,$47,$F4,$26,$BF
	.dc.b	$A2,$6E,$3F,$DA,$FE,$85,$F3,$1F,$B6,$8F,$ED,$DF,$19,$E3,$BF,$DC
	.dc.b	$B5,$FF,$0D,$B5,$FF,$6C,$E5,$E1,$F7,$B9,$71,$FF,$00,$CB,$FD,$71
	.dc.b	$5E,$39,$CB,$FE,$24,$FB,$B7,$B5,$88,$7D,$13,$E3,$30,$D8,$FE,$BA
	.dc.b	$D7,$F7,$BB,$7F,$BA,$5F,$EC,$3F,$09,$FE,$28,$6F,$3F,$58,$C4,$7F
	.dc.b	$04,$AE,$5B,$FF,$83,$65,$EB,$FD,$56,$D7,$F8,$4D,$AF,$D1,$6B,$F7
	.dc.b	$70,$7C,$7F,$AD,$B1,$FD,$D1,$AF,$F1,$49,$FF,$11,$3E,$8F,$4B,$CD
	.dc.b	$EA,$48,$C1,$EC,$7C,$84,$7C,$8F,$F8,$9F,$9F,$FC,$E1,$F3,$7F,$39
	.dc.b	$B3,$FC,$EF,$E0,$93,$F0,$CF,$A7,$EA,$7F,$80,$3C,$7B,$E7,$FF,$8B
	.dc.b	$F7,$FF,$1A,$CF,$F0,$FC,$3F,$BB,$56,$B9,$FF,$C6,$D3,$99,$FE,$D1
	.dc.b	$18,$BF,$E7,$D2,$E9,$FE,$0F,$5F,$E0,$16,$BF,$DF,$DB,$87,$83,$DB
	.dc.b	$FA,$EF,$8A,$33,$87,$F9,$B2,$FA,$7C,$07,$AA,$7F,$CF,$F8,$75,$F1
	.dc.b	$31,$CD,$9E,$59,$3E,$18,$AF,$F8,$5E,$BE,$DB,$D5,$AF,$F8,$85,$B8
	.dc.b	$7A,$FD,$AF,$07,$B7,$2B,$D9,$15,$E9,$1F,$1A,$8D,$31,$F5,$7E,$92
	.dc.b	$4F,$F8,$97,$6F,$F1,$FB,$F1,$E3,$DD,$F4,$52,$30,$7C,$BF,$A1,$F8
	.dc.b	$5C,$F2,$30,$7C,$F7,$C4,$47,$FC,$71,$8F,$67,$F8,$67,$D1,$7E,$19
	.dc.b	$FF,$25,$BF,$C1,$9B,$E1,$FE,$DE,$35,$1F,$DB,$9D,$37,$E6,$74,$6E
	.dc.b	$7F,$7D,$F4,$3F,$EA,$59,$FE,$EA,$FF,$02,$7F,$CB,$2D,$73,$A7,$0F
	.dc.b	$07,$8F,$CE,$C9,$DC,$7F,$09,$F9,$CF,$F2,$C4,$7F,$98,$9F,$F1,$23
	.dc.b	$7C,$FD,$7C,$0E,$DF,$E6,$53,$A9,$FE,$58,$D7,$F9,$C3,$5F,$F0,$DB
	.dc.b	$FD,$95,$FF,$4E,$07,$0B,$CF,$F7,$46,$1B,$3F,$E2,$47,$FD,$8D,$E0
	.dc.b	$3E,$F3,$E5,$C6,$1B,$9F,$F5,$F6,$E0,$AE,$1A,$F3,$AD,$78,$16,$B5
	.dc.b	$E5,$B9,$31,$87,$FB,$05,$B8,$F4,$E9,$FF,$1B,$7C,$EF,$15,$F1,$5F
	.dc.b	$DA,$F3,$FD,$83,$72,$FF,$42,$7B,$7F,$CC,$5F,$81,$3F,$B5,$8C,$6D
	.dc.b	$FE,$DD,$AE,$17,$D2,$FB,$56,$AF,$E6,$FC,$27,$A6,$7E,$63,$78,$7F
	.dc.b	$CE,$31,$F4,$1F,$EC,$78,$FC,$E9,$FF,$27,$F8,$5F,$61,$B5,$D0,$FF
	.dc.b	$79,$C6,$5A,$64,$F4,$9F,$EF,$8A,$7C,$27,$84,$F1,$57,$C5,$C1,$F3
	.dc.b	$5F,$1F,$AF,$FD,$5F,$87,$FB,$7F,$4F,$E4,$1E,$4E,$EB,$4C,$9D,$57
	.dc.b	$07,$C3,$4B,$EB,$1C,$D7,$C4,$78,$49,$19,$F9,$AA,$F2,$BF,$B8,$7D
	.dc.b	$DA,$E3,$1F,$EF,$3E,$47,$CB,$3E,$2B,$F0,$AF,$F7,$89,$F8,$73,$AF
	.dc.b	$53,$35,$27,$95,$6F,$8B,$F3,$6D,$C3,$FC,$6C,$F5,$9D,$DF,$62,$3C
	.dc.b	$AF,$E6,$30,$F8,$07,$E4,$FE,$7F,$FC,$37,$87,$C5,$8F,$F6,$A6,$1F
	.dc.b	$8C,$23,$BF,$9D,$E1,$84,$8F,$A9,$7E,$57,$FD,$B8,$F9,$5F,$A2,$E7
	.dc.b	$63,$49,$8F,$EA,$BE,$87,$86,$5F,$BA,$6E,$5F,$E5,$56,$E1,$FB,$77
	.dc.b	$C1,$A1,$F0,$72,$FF,$A9,$A3,$18,$7E,$62,$B8,$36,$33,$1F,$98,$32
	.dc.b	$7E,$2B,$B7,$F2,$47,$96,$7E,$47,$CE,$49,$F8,$2F,$CB,$F8,$4F,$DF
	.dc.b	$7C,$1A,$3C,$1B,$23,$15,$DE,$31,$3A,$9C,$CB,$2F,$F5,$FD,$7C,$8E
	.dc.b	$1F,$B0,$49,$FA,$3F,$3B,$F6,$FE,$66,$47,$D4,$7E,$F7,$CE,$7D,$FF
	.dc.b	$99,$F1,$F1,$9A,$4A,$F2,$47,$F9,$F1,$FB,$BF,$07,$F8,$73,$F1,$1F
	.dc.b	$3B,$EE,$F9,$B9,$3F,$B5,$D9,$1F,$8D,$F9,$9F,$87,$5E,$D2,$57,$C1
	.dc.b	$A4,$7C,$1C,$7E,$49,$27,$A4,$7E,$5F,$CB,$91,$F9,$07,$F3,$5F,$33
	.dc.b	$88,$F9,$07,$F8,$6F,$3F,$EC,$1C,$3F,$29,$B1,$9A,$F8,$67,$E8,$FE
	.dc.b	$86,$47,$53,$D9,$7F,$70,$8F,$BF,$5A,$F8,$07,$17,$DF,$9F,$B8,$48
	.dc.b	$F8,$0F,$E1,$B5,$91,$D8,$FD,$5F,$C4,$FE,$D4,$8F,$EB,$BA,$AB,$DE
	.dc.b	$FF,$57,$3F,$70,$91,$F6,$9F,$DF,$79,$F2,$3A,$9F,$CF,$FE,$5B,$FA
	.dc.b	$17,$CB,$7F,$61,$8F,$E6,$E7,$F7,$17,$1D,$DF,$DC,$27,$7D,$BF,$D1
	.dc.b	$E7,$39,$7F,$7A,$6D,$FF,$3F,$AD,$BD,$3C,$84,$EA,$3B,$5F,$8D,$E8
	.dc.b	$C7,$84,$6E,$55,$D3,$0F,$F0,$1C,$E4,$D4,$67,$0F,$26,$1C,$EB,$D5
	.dc.b	$85,$9C,$BD,$A7,$C4,$D2,$4F,$68,$CB,$63,$33,$CA,$3A,$C6,$0F,$EC
	.dc.b	$17,$E5,$AE,$35,$B2,$4E,$A3,$2D,$39,$1F,$18,$F9,$57,$ED,$75,$7E
	.dc.b	$33,$C2,$5A,$EB,$1C,$63,$31,$F0,$23,$FC,$67,$7F,$81,$6D,$FF,$B8
	.dc.b	$DF,$01,$FE,$EF,$F8,$BF,$DB,$FD,$0E,$5B,$1F,$EE,$96,$BB,$E5,$4F
	.dc.b	$3C,$DF,$03,$FE,$CF,$AF,$F2,$0B,$5F,$F0,$33,$5F,$F2,$13,$63,$8F
	.dc.b	$C4,$7F,$85,$E3,$FD,$B5,$A7,$F9,$13,$E5,$BF,$80,$BF,$E4,$68,$CE
	.dc.b	$BD,$AB,$81,$FE,$1F,$AD,$C6,$70,$FC,$B5,$70,$7C,$75,$BE,$8E,$31
	.dc.b	$A8,$B8,$78,$66,$DF,$38,$F1,$4F,$1A,$E0,$48,$CB,$FA,$5C,$3C,$6F
	.dc.b	$80,$D5,$73,$3F,$CC,$F1,$A8,$F8,$B5,$E0,$78,$57,$8D,$E2,$62,$B9
	.dc.b	$3F,$E3,$3A,$E5,$8F,$CC,$8D,$9D,$0F,$26,$1D,$5C,$5F,$A0,$EC,$5B
	.dc.b	$F2,$CE,$E9,$7D,$B1,$F9,$98,$7F,$9F,$63,$18,$FC,$5C,$79,$4F,$97
	.dc.b	$0E,$55,$A8,$FF,$19,$7A,$1F,$1B,$8F,$FD,$C5,$EC,$5F,$25,$36,$30
	.dc.b	$CC,$3F,$C8,$BF,$43,$70,$E7,$1F,$E4,$89,$E5,$39,$CB,$FC,$49,$8B
	.dc.b	$E8,$7F,$C7,$97,$FE,$53,$5E,$5F,$8D,$CE,$3F,$E5,$07,$33,$BF,$23
	.dc.b	$46,$FC,$BB,$19,$25,$71,$8C,$BF,$E5,$1F,$37,$DD,$F1,$5E,$F6,$D7
	.dc.b	$F9,$81,$C3,$FE,$62,$F4,$2D,$71,$8D,$19,$2B,$E7,$F3,$3D,$47,$AF
	.dc.b	$CC,$B3,$EF,$70,$FC,$1F,$2E,$D7,$BD,$F8,$47,$D3,$F3,$EC,$FF,$9D
	.dc.b	$27,$EB,$1F,$5F,$9A,$C7,$F9,$F7,$E2,$5A,$FD,$D6,$C6,$78,$31,$39
	.dc.b	$CE,$4F,$1D,$EE,$3F,$E8,$0C,$19,$F3,$3A,$69,$6F,$ED,$76,$7D,$93
	.dc.b	$ED,$79,$36,$3C,$9E,$67,$8C,$F2,$79,$D6,$3F,$D0,$1F,$41,$E6,$CF
	.dc.b	$FC,$FA,$DE,$7F,$F6,$9B,$F8,$46,$EF,$FB,$EF,$09,$3B,$F0,$17,$FE
	.dc.b	$D3,$8C,$37,$83,$84,$62,$B8,$C6,$2F,$2D,$C7,$E6,$79,$DF,$2F,$CF
	.dc.b	$F9,$8D,$8C,$9F,$46,$32,$D3,$EA,$79,$7F,$4F,$7F,$A8,$D8,$C9,$33
	.dc.b	$FA,$1C,$1A,$8F,$F5,$27,$CF,$F9,$C9,$CD,$F0,$5C,$69,$4C,$45,$9E
	.dc.b	$2D,$BF,$F5,$4D,$7E,$B1,$B5,$F8,$7F,$33,$F1,$1E,$65,$C3,$F1,$8D
	.dc.b	$AF,$C2,$37,$8B,$F7,$BC,$E8,$FC,$87,$9D,$2F,$F2,$9E,$75,$C3,$F1
	.dc.b	$93,$70,$E3,$37,$97,$EC,$D7,$FB,$33,$E2,$3F,$9B,$77,$9B,$F1,$18
	.dc.b	$6E,$5F,$E9,$1A,$FF,$69,$CF,$F9,$6A,$39,$C6,$1D,$52,$75,$3E,$08
	.dc.b	$C4,$FE,$2B,$1F,$F5,$E4,$F1,$3F,$C7,$BC,$CF,$93,$F0,$C7,$FB,$9B
	.dc.b	$C9,$7C,$B7,$F2,$D5,$F3,$F6,$30,$DC,$7F,$DE,$93,$C1,$FD,$D6,$B9
	.dc.b	$BF,$F6,$FF,$CF,$BE,$0B,$F4,$0D,$AF,$F7,$42,$F8,$3C,$CE,$9B,$3A
	.dc.b	$6D,$7F,$95,$B6,$7C,$4F,$FA,$03,$E3,$7F,$CE,$50,$F8,$BC,$61,$F1
	.dc.b	$6A,$DF,$EB,$AB,$53,$FB,$0B,$FF,$3D,$7D,$6F,$FB,$EA,$3F,$C4,$4D
	.dc.b	$AF,$A6,$1C,$3F,$59,$AF,$E0,$BE,$7F,$F0,$13,$D4,$FC,$2C,$F3,$AF
	.dc.b	$F4,$44,$EA,$91,$E2,$F1,$7C,$F0,$9F,$3B,$AB,$ED,$F1,$FA,$9C,$DF
	.dc.b	$48,$F7,$FA,$0E,$C7,$B7,$BD,$B3,$18,$C7,$CD,$97,$F1,$2C,$B9,$C7
	.dc.b	$F1,$4C,$1F,$7B,$FD,$BA,$37,$5E,$5F,$8A,$EF,$39,$8F,$31,$CE,$3B
	.dc.b	$FB,$DA,$F5,$AD,$F8,$30,$EB,$1E,$4F,$2F,$FD,$59,$87,$87,$34,$62
	.dc.b	$3F,$D5,$F1,$8B,$FA,$B7,$EF,$AD,$2F,$A0,$FC,$DC,$F7,$AE,$17,$E6
	.dc.b	$AD,$75,$78,$6B,$C9,$6F,$97,$F9,$6F,$E7,$EF,$83,$99,$DF,$66,$AA
	.dc.b	$47,$2F,$9F,$B7,$C3,$E8,$FC,$B8,$1A,$4A,$F5,$C3,$E5,$BC,$33,$E2
	.dc.b	$6C,$F9,$BE,$8B,$99,$1E,$2F,$6E,$CF,$4B,$EA,$6C,$7B,$BE,$43,$DB
	.dc.b	$DD,$EE,$6C,$64,$99,$7C,$0F,$0B,$51,$F1,$BE,$6F,$D6,$47,$D2,$FB
	.dc.b	$BE,$33,$E4,$BF,$29,$B1,$F8,$AF,$8E,$F9,$57,$F8,$80,$B7,$FD,$1B
	.dc.b	$AE,$EF,$F7,$83,$9A,$FC,$CF,$33,$EF,$BF,$D8,$F1,$FB,$88,$C7,$C4
	.dc.b	$FE,$25,$B1,$99,$DE,$3F,$CC,$39,$FF,$11,$F1,$FF,$C2,$2F,$73,$F8
	.dc.b	$08,$C5,$70,$CB,$C7,$BF,$9B,$7F,$3B,$72,$FD,$3F,$CF,$FE,$27,$1F
	.dc.b	$E7,$AD,$7F,$A4,$B5,$FF,$01,$B5,$7D,$0F,$F8,$3F,$05,$7F,$50,$8C
	.dc.b	$3C,$B7,$FE,$13,$18,$AE,$71,$FC,$32,$FE,$44,$7E,$6B,$E8,$B1,$3F
	.dc.b	$C4,$3E,$3D,$F3,$7F,$AF,$A4,$63,$1F,$F6,$33,$63,$F9,$2B,$5F,$6E
	.dc.b	$9C,$71,$FE,$A1,$5F,$EC,$D6,$CF,$F2,$C6,$BF,$DA,$B8,$3F,$D1,$3E
	.dc.b	$6F,$F1,$A3,$FD,$17,$5F,$B4,$E2,$C4,$7E,$D3,$E3,$FF,$56,$49,$FD
	.dc.b	$4B,$FC,$1F,$E8,$5F,$7E,$F5,$70,$DA,$FE,$7B,$3A,$9C,$9F,$07,$2F
	.dc.b	$ED,$87,$F1,$2F,$17,$FA,$1F,$A7,$FE,$38,$E8,$46,$27,$47,$F7,$A9
	.dc.b	$D4,$7F,$81,$BE,$45,$EF,$7F,$40,$AD,$88,$CF,$3F,$08,$FF,$22,$C6
	.dc.b	$1D,$B7,$0F,$0C,$32,$7E,$01,$FC,$66,$3D,$AD,$7F,$C1,$B1,$AF,$17
	.dc.b	$FE,$00,$36,$3F,$88,$FC,$87,$F6,$05,$C7,$FC,$AE,$30,$D8,$D9,$FD
	.dc.b	$9B,$27,$D4,$BE,$7F,$FC,$4B,$1F,$DD,$E3,$13,$FD,$42,$37,$3A,$E1
	.dc.b	$FE,$E9,$4F,$9F,$C5,$FF,$99,$B3,$FF,$58,$BE,$56,$BF,$DD,$9A,$FF
	.dc.b	$8A,$9A,$FF,$8D,$BE,$4D,$DF,$ED,$84,$61,$F8,$0D,$7F,$C0,$2D,$7F
	.dc.b	$C6,$BF,$19,$FE,$DE,$7F,$B1,$B5,$FE,$34,$DC,$FF,$6F,$1F,$2B,$CB
	.dc.b	$C6,$7F,$E4,$8F,$A1,$FF,$6C,$CF,$F6,$C6,$BF,$C8,$DB,$5C,$A1,$7C
	.dc.b	$9F,$F4,$4F,$D1,$3E,$3F,$D8,$D8,$FB,$4D,$7F,$C3,$6D,$7F,$C9,$FE
	.dc.b	$8F,$F0,$07,$FB,$EA,$F6,$3E,$8C,$7F,$82,$DB,$1B,$8F,$EC,$7F,$45
	.dc.b	$F5,$B9,$DF,$80,$FC,$9C,$FE,$0C,$63,$83,$ED,$F2,$FF,$75,$BE,$17
	.dc.b	$F3,$70,$FD,$63,$5F,$F0,$0B,$6F,$94,$7E,$57,$17,$D1,$F9,$3E,$7A
	.dc.b	$4F,$AA,$3B,$F1,$7C,$9D,$3E,$9E,$6F,$82,$FF,$3B,$5F,$F8,$4D,$AF
	.dc.b	$F7,$EE,$3F,$3C,$62,$3F,$8B,$35,$FF,$13,$35,$FD,$3B,$67,$B6,$FC
	.dc.b	$89,$3D,$FE,$12,$BF,$E1,$3A,$F9,$98,$79,$9B,$87,$6B,$F1,$FB,$BE
	.dc.b	$4F,$82,$98,$7A,$62,$BF,$D6,$EF,$FC,$CE,$FF,$89,$9A,$E6,$7F,$77
	.dc.b	$1F,$AF,$8F,$85,$D7,$F0,$0F,$83,$D5,$63,$94,$70,$23,$D7,$7F,$8E
	.dc.b	$47,$E6,$7D,$0D,$7D,$DC,$1F,$30,$C6,$BE,$FF,$77,$E7,$FC,$F4,$AF
	.dc.b	$A9,$15,$F6,$47,$A0,$8F,$B7,$E6,$FF,$1B,$F8,$17,$C8,$8F,$E3,$2D
	.dc.b	$E3,$FC,$04,$EB,$D0,$FC,$EF,$D2,$35,$A6,$70,$FC,$27,$CE,$FE,$8B
	.dc.b	$E9,$3F,$D9,$11,$FE,$62,$6B,$F6,$63,$5F,$1D,$BF,$36,$4E,$E3,$F5
	.dc.b	$7F,$18,$F8,$0F,$D6,$FC,$57,$EF,$2B,$FC,$DA,$D7,$FC,$49,$3B,$73
	.dc.b	$E2,$FF,$19,$9D,$47,$FA,$4E,$30,$7F,$1D,$F9,$FF,$F5,$74,$7F,$9E
	.dc.b	$1F,$F5,$A3,$7E,$07,$F7,$B1,$A8,$FF,$3C,$F8,$4F,$A4,$F8,$91,$FE
	.dc.b	$CD,$EC,$FE,$C1,$0E,$6F,$F7,$0F,$A7,$FA,$BA,$FF,$3B,$CE,$5E,$06
	.dc.b	$38,$B7,$7C,$AE,$CF,$06,$E1,$E1,$6D,$F8,$A5,$87,$7E,$6F,$F2,$3F
	.dc.b	$17,$26,$CF,$81,$AF,$0C,$7F,$6F,$F1,$1B,$F8,$15,$D8,$95,$C6,$72
	.dc.b	$FA,$86,$AF,$83,$F4,$C6,$1B,$19,$8E,$1C,$CF,$91,$CD,$26,$78,$BF
	.dc.b	$F0,$02,$F8,$47,$FA,$89,$B1,$99,$E1,$39,$F8,$4E,$E4,$7C,$D5,$CB
	.dc.b	$6F,$8D,$AB,$E7,$BE,$33,$6B,$E3,$35,$F8,$C7,$F1,$AF,$9F,$DC,$C8
	.dc.b	$DB,$CA,$39,$FC,$FF,$4F,$91,$95,$D2,$32,$F6,$8C,$BD,$B8,$BF,$99
	.dc.b	$4F,$E0,$3E,$47,$FB,$DE,$5F,$8B,$E6,$43,$AF,$46,$FE,$5E,$E3,$F8
	.dc.b	$EA,$DC,$E2,$3D,$71,$CA,$71,$7D,$6F,$A5,$71,$6B,$C1,$F1,$EB,$FE
	.dc.b	$E3,$FA,$2B,$3F,$E2,$46,$E6,$6C,$FF,$6F,$BF,$CB,$46,$2B,$E8,$9F
	.dc.b	$97,$78,$9E,$08,$DF,$CB,$E4,$D6,$72,$BF,$39,$49,$D4,$62,$7F,$53
	.dc.b	$F4,$79,$8F,$88,$F5,$C3,$FC,$B0,$D8,$FF,$76,$BB,$7E,$3B,$9D,$0D
	.dc.b	$25,$F6,$B9,$5D,$9F,$0C,$7B,$7E,$82,$4F,$A1,$F8,$1D,$FF,$B7,$F8
	.dc.b	$94,$52,$4F,$EE,$63,$F9,$16,$9F,$2B,$6F,$90,$49,$F8,$7F,$43,$F2
	.dc.b	$FE,$7A,$47,$C6,$7E,$77,$D0,$FC,$FF,$9F,$93,$F2,$9F,$A5,$F4,$1F
	.dc.b	$4F,$E1,$E4,$FC,$E7,$EB,$7D,$07,$9E,$64,$F9,$5F,$B5,$E5,$7A,$E6
	.dc.b	$4F,$A5,$DB,$F9,$FD,$58,$7C,$12,$46,$CF,$BB,$F3,$FF,$85,$23,$F8
	.dc.b	$0F,$06,$BF,$A9,$E7,$4E,$38,$3F,$1F,$57,$9F,$F3,$1E,$67,$E3,$CC
	.dc.b	$61,$FE,$7C,$CB,$F8,$A3,$F3,$EB,$EA,$1F,$57,$6B,$19,$21,$EB,$72
	.dc.b	$FB,$75,$FB,$86,$31,$FD,$0F,$C5,$7E,$7C,$93,$F9,$AC,$AB,$EB,$8B
	.dc.b	$3F,$A8,$9A,$7E,$C7,$E7,$7F,$50,$62,$BF,$95,$78,$2F,$8F,$FE,$63
	.dc.b	$F1,$CE,$DF,$E4,$6D,$AF,$E3,$A7,$F3,$A9,$FE,$6E,$70,$9D,$9F,$CA
	.dc.b	$9C,$C6,$0E,$34,$AE,$98,$FF,$61,$9E,$86,$DC,$4E,$EF,$18,$FF,$3A
	.dc.b	$8F,$45,$60,$F5,$DF,$49,$D3,$63,$DF,$87,$9D,$F1,$C6,$5F,$EA,$AD
	.dc.b	$AE,$B1,$AB,$EC,$E9,$F7,$38,$8F,$33,$B3,$DF,$5D,$F0,$F5,$DF,$53
	.dc.b	$B4,$67,$0F,$3E,$1C,$E3,$17,$EC,$8C,$3D,$EB,$94,$78,$A3,$C1,$7D
	.dc.b	$A3,$71,$8B,$EA,$D3,$E1,$9F,$54,$FE,$D5,$5F,$7D,$B3,$FE,$DB,$C3
	.dc.b	$DB,$1F,$DA,$8F,$F7,$87,$1B,$E8,$FF,$DC,$1F,$1C,$8C,$4E,$CD,$CE
	.dc.b	$A3,$31,$FD,$E7,$E3,$3F,$D5,$8F,$E5,$1A,$FF,$82,$1B,$66,$9F,$EB
	.dc.b	$7C,$FF,$0E,$DC,$FF,$0D,$F2,$1F,$EC,$A9,$CF,$C4,$FF,$7E,$AD,$61
	.dc.b	$DE,$3F,$C0,$35,$BB,$FC,$75,$6A,$BC,$55,$C4,$DF,$97,$63,$F3,$B3
	.dc.b	$BC,$7C,$D5,$A9,$CB,$BC,$7C,$71,$FA,$4C,$7C,$0F,$8C,$D9,$23,$F4
	.dc.b	$71,$A8,$F1,$99,$C7,$CF,$5D,$EB,$A9,$23,$33,$DF,$1F,$14,$62,$75
	.dc.b	$3F,$C1,$2F,$B6,$1E,$A8,$EC,$79,$67,$C7,$7C,$A3,$17,$CC,$FA,$31
	.dc.b	$98,$FF,$20,$DF,$4C,$7E,$33,$71,$F4,$5F,$1A,$F1,$5F,$1C,$57,$B9
	.dc.b	$F5,$1E,$4F,$11,$F4,$7F,$E1,$FB,$FF,$2D,$2F,$1B,$E6,$CE,$BC,$5B
	.dc.b	$39,$AD,$C6,$2B,$2D,$8C,$92,$FF,$C6,$55,$FE,$0B,$8C,$6B,$FE,$6D
	.dc.b	$FC,$0D,$CD,$E0,$7F,$DD,$13,$FE,$6D,$72,$DD,$75,$3B,$6C,$64,$E1
	.dc.b	$18,$C3,$FC,$85,$4E,$6C,$61,$FE,$7D,$C5,$E2,$FB,$8F,$7F,$8B,$72
	.dc.b	$F8,$AD,$C7,$D6,$DB,$E4,$F0,$9F,$97,$E2,$5A,$F8,$3E,$2A,$7F,$CE
	.dc.b	$71,$FE,$75,$E0,$FA,$73,$98,$82,$33,$3C,$23,$19,$7F,$96,$23,$12
	.dc.b	$FC,$11,$F1,$FF,$69,$7C,$B5,$8C,$39,$F9,$DF,$05,$FF,$A0,$2F,$94
	.dc.b	$63,$1F,$36,$08,$C4,$79,$63,$15,$E2,$C5,$E3,$FE,$20,$FC,$4F,$93
	.dc.b	$67,$F1,$FD,$7F,$8E,$F3,$7F,$C2,$F1,$83,$FD,$58,$E3,$1F,$C4,$F8
	.dc.b	$5F,$92,$3F,$27,$E1,$DA,$FC,$CF,$8D,$F9,$63,$F2,$FE,$2D,$8F,$F5
	.dc.b	$B6,$BF,$DE,$5F,$F2,$07,$8F,$F1,$7C,$6F,$CF,$F8,$FF,$E1,$E9,$FD
	.dc.b	$1F,$6F,$E8,$0F,$D0,$D5,$8F,$F5,$09,$FC,$37,$B9,$F3,$9F,$C0,$A3
	.dc.b	$0D,$7E,$C6,$DF,$EB,$C7,$E4,$B5,$CF,$85,$F8,$56,$9C,$63,$FC,$17
	.dc.b	$C0,$FA,$3C,$76,$FF,$B5,$63,$FA,$55,$7C,$8E,$3F,$DB,$F7,$FC,$D6
	.dc.b	$46,$4F,$DF,$C6,$30,$FF,$53,$53,$8F,$EA,$1F,$C0,$BE,$99,$F0,$18
	.dc.b	$C3,$FD,$C1,$18,$7E,$BC,$EE,$75,$7F,$EB,$1A,$7C,$5F,$F0,$CF,$37
	.dc.b	$FC,$51,$87,$28,$C7,$99,$DD,$FC,$6F,$72,$F8,$D7,$E7,$9B,$87,$E9
	.dc.b	$A7,$55,$C2,$73,$B7,$FB,$E5,$BC,$9F,$EE,$3C,$5F,$82,$BD,$0F,$06
	.dc.b	$7F,$EF,$BF,$B6,$7D,$1F,$F2,$26,$C7,$FB,$07,$E0,$1F,$29,$9A,$FE
	.dc.b	$A7,$8A,$73,$5C,$E7,$31,$8D,$FF,$DA,$98,$74,$7E,$8F,$81,$DF,$E8
	.dc.b	$FF,$48,$DC,$3F,$46,$E1,$FA,$3F,$21,$6B,$FC,$9F,$D8,$F8,$4F,$CA
	.dc.b	$1F,$D8,$7A,$3E,$B4,$EA,$31,$CF,$F9,$29,$CE,$1F,$E1,$76,$D0,$8E
	.dc.b	$1B,$3C,$A7,$5C,$9F,$E5,$5A,$F9,$F5,$A9,$FD,$DF,$85,$FC,$77,$E2
	.dc.b	$FF,$8F,$74,$37,$FE,$2B,$5D,$E7,$38,$2F,$84,$FF,$27,$AD,$E0,$BD
	.dc.b	$D6,$A3,$F9,$45,$F1,$BE,$87,$F2,$FF,$07,$FA,$5C,$F1,$8F,$E6,$53
	.dc.b	$B8,$C1,$F0,$75,$7D,$1F,$5C,$FD,$1E,$5F,$7A,$56,$B0,$EF,$1F,$CE
	.dc.b	$74,$FE,$67,$18,$8D,$FC,$07,$1A,$E1,$1E,$03,$33,$B8,$CE,$BF,$22
	.dc.b	$EC,$FC,$3C,$BC,$91,$FC,$CA,$FD,$39,$7C,$D9,$F3,$7C,$3F,$E7,$EF
	.dc.b	$FC,$9D,$48,$C3,$FE,$51,$6B,$FE,$59,$6B,$F8,$33,$DF,$F0,$3D,$3A
	.dc.b	$BD,$4D,$AF,$83,$87,$DD,$F1,$7C,$BF,$11,$E6,$6C,$FB,$7E,$8F,$0D
	.dc.b	$38,$E9,$FC,$B1,$BF,$41,$EE,$E7,$F7,$B6,$3B,$FD,$DF,$E1,$38,$3E
	.dc.b	$34,$FD,$78,$C4,$FE,$26,$3E,$3F,$7F,$F6,$ED,$FF,$18,$D7,$EC,$37
	.dc.b	$5F,$CB,$37,$27,$1F,$2C,$3C,$ED,$BF,$A3,$CB,$FE,$1A,$7E,$9D,$38
	.dc.b	$3F,$15,$39,$D3,$EE,$46,$3C,$19,$8F,$CA,$F0,$9F,$27,$8C,$B9,$F0
	.dc.b	$F8,$C9,$5F,$E8,$9D,$BF,$59,$1E,$7F,$A2,$FC,$2B,$F8,$66,$CF,$CE
	.dc.b	$F3,$7D,$84,$79,$6D,$FD,$9F,$0F,$F4,$D6,$1F,$B7,$F8,$1F,$C7,$DF
	.dc.b	$E4,$1B,$7F,$80,$FB,$07,$87,$F9,$4F,$80,$FD,$44,$62,$7E,$EC,$62
	.dc.b	$64,$7B,$FE,$4E,$E5,$F2,$3E,$03,$E8,$12,$7C,$AF,$D3,$F9,$BF,$D0
	.dc.b	$BF,$A2,$6C,$7E,$4B,$D2,$FE,$D0,$DF,$4B,$F4,$EF,$EE,$E3,$37,$FB
	.dc.b	$1F,$B5,$7C,$9F,$E0,$E1,$18,$AF,$D4,$B6,$3F,$D7,$0D,$7F,$4D,$E7
	.dc.b	$3E,$C9,$E1,$7F,$07,$AE,$1D,$7F,$E3,$FA,$D4,$E7,$1F,$27,$97,$32
	.dc.b	$F5,$7C,$D5,$CF,$F8,$37,$D7,$B7,$F3,$36,$BF,$7A,$D9,$FF,$66,$35
	.dc.b	$FF,$6A,$7E,$0D,$F7,$74,$FF,$27,$FC,$27,$FB,$1E,$FF,$8B,$4E,$5F
	.dc.b	$E0,$11,$9A,$FE,$2D,$E6,$FF,$4C,$7F,$DB,$CD,$7F,$83,$35,$FE,$94
	.dc.b	$DE,$A7,$13,$D0,$FA,$47,$D3,$F5,$6E,$3F,$A1,$F9,$9F,$D9,$78,$46
	.dc.b	$92,$3F,$8D,$F9,$2F,$7B,$EC,$36,$33,$18,$3F,$7D,$F4,$8F,$B6,$79
	.dc.b	$38,$6D,$7F,$BC,$E7,$53,$93,$F0,$9F,$83,$FE,$97,$83,$FD,$FC,$D7
	.dc.b	$F9,$33,$5F,$ED,$4D,$7F,$95,$7A,$DF,$E1,$3F,$8A,$FF,$80,$0D,$8F
	.dc.b	$E5,$C1,$C3,$C0,$76,$7F,$AA,$1F,$EE,$DF,$1D,$FE,$39,$E7,$DB,$63
	.dc.b	$77,$FC,$E6,$31,$5C,$65,$18,$9F,$E7,$DF,$84,$3E,$4F,$FA,$3D,$23
	.dc.b	$0F,$F4,$A6,$BE,$83,$51,$FD,$3B,$47,$DC,$7F,$28,$9F,$98,$D7,$DF
	.dc.b	$1A,$F4,$78,$4E,$67,$E1,$F9,$1F,$EB,$67,$FA,$CB,$5F,$03,$5D,$F5
	.dc.b	$FF,$5E,$EA,$FF,$25,$7C,$E5,$E8,$34,$95,$FC,$A7,$AF,$F9,$7B,$FC
	.dc.b	$27,$D0,$FE,$C1,$C5,$FE,$18,$9C,$CF,$F6,$E5,$F4,$3A,$E7,$FE,$CD
	.dc.b	$F4,$FF,$E4,$6A,$FE,$DE,$D7,$E3,$35,$FB,$ED,$7F,$B7,$7E,$13,$ED
	.dc.b	$C3,$FD,$87,$F6,$8F,$89,$ED,$5F,$E3,$58,$C3,$FE,$0D,$AE,$18,$7F
	.dc.b	$77,$F8,$07,$CC,$7D,$A6,$D7,$F8,$09,$AF,$D9,$6B,$FE,$C3,$D1,$F4
	.dc.b	$7E,$0E,$1F,$DC,$B6,$33,$3F,$CB,$FE,$93,$CB,$D9,$FC,$87,$C9,$F3
	.dc.b	$CB,$EA,$3F,$3B,$C1,$FE,$DE,$6E,$BF,$E7,$5F,$43,$F8,$E5,$7F,$89
	.dc.b	$E7,$F2,$2F,$13,$F2,$6E,$0F,$CA,$E9,$FC,$33,$E1,$FC,$84,$9F,$E8
	.dc.b	$5E,$A7,$9F,$C8,$F4,$74,$49,$FE,$8D,$15,$FE,$BF,$BF,$E1,$9F,$C2
	.dc.b	$4F,$E2,$E3,$11,$FD,$93,$D6,$FB,$9A,$FF,$97,$1B,$AF,$FB,$F5,$B3
	.dc.b	$FE,$FB,$EA,$7D,$17,$CB,$9C,$EB,$FC,$93,$D2,$F9,$86,$7D,$FF,$9D
	.dc.b	$5F,$3F,$E8,$A6,$3F,$D7,$62,$BE,$88,$F3,$11,$FC,$E7,$CF,$FD,$7F
	.dc.b	$A0,$91,$FD,$96,$2B,$F0,$23,$DC,$47,$FC,$19,$93,$DD,$F4,$12,$3F
	.dc.b	$C2,$D3,$5F,$D2,$79,$BF,$85,$31,$BF,$EA,$7C,$8F,$C4,$7C,$EC,$AF
	.dc.b	$EE,$D1,$5F,$A3,$1F,$3C,$8F,$F8,$77,$D1,$F1,$3F,$E4,$EB,$7C,$EF
	.dc.b	$E5,$5B,$1F,$E4,$4C,$B1,$3F,$97,$88,$71,$7F,$2B,$F4,$9F,$52,$3A
	.dc.b	$7C,$67,$33,$3C,$16,$3C,$1E,$17,$73,$0F,$F9,$E7,$E8,$9F,$43,$FA
	.dc.b	$0F,$8D,$FF,$42,$CF,$FA,$05,$AE,$E3,$58,$7F,$2E,$9C,$CF,$F9,$C6
	.dc.b	$30,$7F,$1B,$D1,$F3,$FE,$F2,$4F,$C3,$F3,$7F,$5F,$8F,$F3,$DE,$27
	.dc.b	$5F,$F9,$E7,$6F,$E6,$F1,$8B,$FF,$5F,$46,$27,$FC,$FD,$5C,$5E,$1C
	.dc.b	$7F,$37,$E7,$E5,$FF,$A6,$EF,$82,$EB,$73,$BE,$13,$F3,$C4,$9F,$F3
	.dc.b	$BF,$8B,$5F,$F3,$3F,$CF,$FD,$93,$1A,$F9,$63,$FC,$0F,$DB,$FD,$A3
	.dc.b	$E2,$FF,$D2,$5E,$0F,$F2,$BF,$65,$18,$9D,$57,$FB,$66,$9F,$1B,$88
	.dc.b	$E6,$7F,$B7,$78,$7F,$B7,$CF,$F3,$56,$D7,$FA,$7D,$B3,$B3,$E2,$FC
	.dc.b	$CF,$3E,$33,$4E,$D7,$E5,$C7,$F8,$16,$BE,$4E,$0F,$8A,$DE,$0F,$E3
	.dc.b	$F8,$FF,$7F,$AF,$F7,$85,$3C,$4F,$ED,$BB,$7B,$AB,$E7,$5F,$08,$C5
	.dc.b	$7F,$86,$BE,$47,$B7,$41,$ED,$97,$C4,$F3,$1C,$1F,$CE,$32,$F0,$1F
	.dc.b	$EE,$8F,$7F,$89,$B8,$D6,$5D,$B1,$FF,$17,$57,$FC,$00,$D7,$FD,$D3
	.dc.b	$7C,$23,$07,$F8,$5A,$F8,$CF,$B3,$57,$87,$F0,$DB,$5C,$35,$F8,$13
	.dc.b	$E7,$F2,$DF,$21,$B8,$FA,$95,$CD,$AF,$F2,$D6,$D7,$28,$FB,$06,$4E
	.dc.b	$D9,$BE,$73,$E3,$9D,$63,$C5,$8F,$F8,$CB,$5F,$9B,$C3,$F9,$36,$F2
	.dc.b	$FA,$6B,$ED,$31,$F5,$4E,$F5,$F8,$11,$EF,$F8,$B8,$F6,$7F,$09,$F4
	.dc.b	$6F,$27,$33,$F5,$35,$FC,$15,$63,$2F,$BF,$7B,$AF,$9E,$7F,$9B,$3E
	.dc.b	$47,$F2,$24,$73,$CF,$FE,$D1,$6C,$7C,$87,$81,$F1,$8A,$7F,$B9,$61
	.dc.b	$1B,$6C,$78,$3A,$7F,$CA,$6D,$E6,$F0,$3E,$9B,$FB,$AD,$E2,$F7,$9F
	.dc.b	$03,$C0,$FD,$27,$E1,$5F,$C9,$33,$FC,$DF,$CE,$3E,$33,$F5,$0D,$C3
	.dc.b	$F0,$D5,$BE,$1F,$D3,$FC,$57,$F4,$09,$F9,$0D,$C3,$F4,$AD,$9F,$D6
	.dc.b	$E7,$F7,$5E,$3E,$3F,$ED,$67,$A5,$FF,$1B,$6F,$11,$FF,$13,$E1,$D3
	.dc.b	$C7,$FC,$D6,$92,$FF,$35,$5F,$CD,$9B,$E1,$F5,$23,$AF,$1B,$E0,$F8
	.dc.b	$5F,$98,$DB,$F8,$5C,$FE,$63,$34,$62,$3F,$92,$46,$27,$87,$CC,$7C
	.dc.b	$AF,$98,$95,$F1,$A3,$D9,$E1,$7E,$93,$3F,$DC,$CF,$26,$F6,$23,$E5
	.dc.b	$CE,$F0,$F8,$84,$8F,$6F,$75,$7E,$7F,$97,$F2,$4C,$70,$FD,$07,$11
	.dc.b	$F1,$1C,$F1,$FF,$1E,$DB,$F1,$33,$F9,$F6,$D7,$F1,$C6,$CE,$F9,$7F
	.dc.b	$0F,$D1,$F8,$8F,$87,$98,$7E,$12,$32,$FE,$1A,$32,$F5,$8F,$BD,$C1
	.dc.b	$23,$EB,$BF,$81,$F9,$CC,$47,$E2,$B8,$7F,$08,$63,$3F,$C5,$D6,$27
	.dc.b	$EF,$CE,$72,$FE,$AB,$E8,$FF,$42,$F9,$4F,$D2,$E1,$FC,$E6,$FF,$DC
	.dc.b	$31,$8A,$DC,$7E,$DA,$B8,$65,$FB,$8A,$D4,$FF,$5B,$9F,$E9,$A6,$9F
	.dc.b	$E7,$27,$49,$E4,$F1,$63,$AE,$E5,$74,$8F,$F7,$9D,$73,$C3,$CD,$5A
	.dc.b	$BE,$31,$89,$D6,$1F,$DC,$B0,$E5,$7B,$C3,$B5,$69,$FE,$E5,$3A,$37
	.dc.b	$19,$8F,$41,$E5,$8C,$BE,$0A,$ED,$85,$8F,$84,$53,$8E,$1E,$8A,$E3
	.dc.b	$1C,$E7,$58,$79,$6B,$D5,$38,$AF,$25,$7C,$88,$E0,$49,$E9,$19,$6C
	.dc.b	$66,$FC,$AF,$C0,$3E,$79,$FE,$0C,$AF,$CE,$BA,$31,$5F,$EE,$77,$FD
	.dc.b	$E3,$C2,$FA,$3C,$7C,$5F,$F8,$31,$B1,$99,$FF,$77,$4E,$5F,$F7,$86
	.dc.b	$4F,$87,$FE,$5F,$1F,$E2,$66,$BF,$E2,$36,$A8,$E1,$E0,$78,$DB,$1F
	.dc.b	$87,$9F,$EE,$BA,$7E,$13,$CF,$FF,$24,$3F,$86,$8C,$EB,$DF,$0F,$E9
	.dc.b	$F1,$8B,$FF,$19,$5F,$0A,$D4,$75,$AD,$9F,$D5,$A7,$51,$E7,$8C,$BB
	.dc.b	$3F,$35,$1A,$7D,$55,$DD,$B3,$C2,$FD,$1D,$17,$0F,$0C,$7A,$67,$C1
	.dc.b	$39,$AF,$01,$DE,$78,$B6,$B8,$91,$F6,$CF,$17,$C2,$6C,$C9,$4C,$CF
	.dc.b	$85,$F5,$C7,$1C,$7E,$8C,$E4,$F0,$96,$BF,$CC,$38,$FC,$19,$DC,$62
	.dc.b	$B9,$4E,$71,$F8,$95,$FE,$73,$9C,$9F,$DF,$DC,$61,$FD,$FE,$BF,$D0
	.dc.b	$93,$FE,$64,$6C,$7F,$95,$3E,$81,$E0,$7D,$6E,$C4,$7F,$AC,$AF,$FC
	.dc.b	$74,$E9,$BE,$B6,$E3,$17,$FE,$1D,$78,$36,$F5,$3B,$65,$74,$F5,$6D
	.dc.b	$7F,$89,$7E,$8D,$D3,$E3,$7F,$D0,$75,$FE,$4B,$72,$DD,$B5,$3B,$6C
	.dc.b	$64,$E1,$18,$AE,$DD,$57,$83,$FC,$DF,$A3,$D5,$F6,$1E,$DF,$12,$E3
	.dc.b	$F0,$9B,$87,$BD,$B7,$C9,$E1,$3F,$23,$C2,$B5,$F1,$7C,$3F,$90,$7C
	.dc.b	$FF,$A0,$7D,$1E,$2F,$FC,$D1,$0F,$85,$D4,$ED,$B3,$AF,$A3,$E4,$7D
	.dc.b	$3F,$5A,$C7,$D8,$F1,$BE,$B1,$F5,$FC,$7B,$1F,$73,$BB,$F0,$1E,$97
	.dc.b	$FB,$7A,$30,$7F,$9B,$3C,$8F,$BD,$E2,$C8,$FC,$1F,$8F,$78,$3F,$01
	.dc.b	$F4,$7F,$83,$3F,$09,$E4,$5A,$F5,$F9,$1F,$8D,$F2,$9F,$15,$F0,$63
	.dc.b	$0F,$E2,$FB,$65,$7E,$42,$2E,$3F,$D7,$E3,$F4,$C7,$F0,$7D,$5E,$8E
	.dc.b	$23,$FC,$D3,$18,$7F,$CC,$E6,$E3,$53,$FD,$BA,$74,$F5,$9E,$33,$AE
	.dc.b	$AF,$C2,$F6,$FC,$6E,$1F,$D0,$36,$BF,$4A,$E1,$FD,$3F,$95,$70,$FF
	.dc.b	$40,$7A,$8F,$98,$F9,$FC,$5F,$AF,$8C,$5F,$FA,$6A,$31,$32,$3F,$D5
	.dc.b	$5E,$3B,$E8,$71,$87,$4F,$96,$E4,$7F,$2E,$C9,$F4,$5F,$EC,$7C,$3F
	.dc.b	$D4,$33,$98,$C5,$71,$9F,$F5,$ED,$70,$EB,$E7,$18,$AD,$EF,$FE,$D7
	.dc.b	$AD,$4E,$7A,$FA,$5F,$87,$DA,$FF,$69,$CF,$46,$E3,$FE,$D4,$B7,$D9
	.dc.b	$BE,$73,$87,$BB,$E9,$8C,$61,$C6,$31,$3A,$8F,$F5,$DF,$CD,$7E,$63
	.dc.b	$BB,$E5,$36,$32,$4C,$7F,$D6,$1C,$3F,$AE,$8C,$6D,$FE,$39,$E8,$FA
	.dc.b	$35,$F8,$3E,$3F,$C6,$71,$7E,$35,$B7,$FE,$AB,$A4,$FF,$2A,$AF,$F3
	.dc.b	$74,$62,$75,$39,$94,$E5,$FC,$84,$7E,$67,$1E,$B1,$FE,$9E,$F2,$97
	.dc.b	$FE,$79,$F8,$1B,$2F,$43,$9F,$81,$89,$CC,$7F,$A6,$EF,$17,$FE,$81
	.dc.b	$8D,$BC,$CC,$75,$31,$DE,$3F,$EA,$5B,$B6,$AD,$5F,$1A,$D3,$5F,$E9
	.dc.b	$3F,$2D,$C9,$B1,$9F,$41,$F4,$19,$55,$62,$4E,$FC,$2B,$6D,$3E,$0A
	.dc.b	$F8,$5F,$CC,$C6,$1B,$5F,$E9,$F9,$FE,$B7,$F4,$3E,$39,$CC,$7F,$A8
	.dc.b	$D7,$91,$D2,$38,$46,$2B,$FC,$74,$D8,$E8,$71,$9E,$0F,$E0,$3E,$5F
	.dc.b	$B4,$7C,$1E,$34,$62,$FE,$17,$C0,$FC,$13,$E1,$7B,$16,$3E,$33,$87
	.dc.b	$E6,$7D,$9B,$E4,$31,$97,$C3,$9F,$B9,$F1,$7F,$9B,$8C,$1F,$85,$71
	.dc.b	$B7,$E1,$38,$DE,$0F,$DC,$9C,$DF,$E0,$A3,$EF,$CE,$A3,$EF,$72,$FE
	.dc.b	$07,$E4,$33,$5F,$84,$DD,$E7,$7C,$2A,$FC,$EB,$63,$37,$F2,$33,$FC
	.dc.b	$C4,$61,$F9,$75,$FA,$4F,$33,$A5,$F5,$6D,$F8,$BC,$AF,$96,$F9,$30
	.dc.b	$FF,$01,$E1,$F0,$30,$56,$A3,$13,$AD,$7F,$4B,$39,$6F,$17,$E7,$BC
	.dc.b	$0F,$E1,$F7,$CE,$B5,$8F,$EA,$70,$FD,$13,$EB,$C7,$F7,$7B,$7B,$C9
	.dc.b	$19,$9F,$F0,$AD,$F8,$A7,$E4,$75,$FC,$97,$E5,$78,$BF,$1C,$97,$C5
	.dc.b	$F9,$9F,$0F,$FC,$FE,$3E,$04,$7E,$D2,$31,$7E,$E8,$FD,$BE,$5F,$87
	.dc.b	$AF,$B3,$4F,$A1,$CF,$27,$16,$CF,$CD,$8F,$85,$CF,$8C,$7E,$65,$FE
	.dc.b	$35,$FD,$66,$A8,$C4,$7F,$0D,$6B,$FC,$55,$B1,$BF,$4F,$D1,$EA,$7A
	.dc.b	$5B,$8F,$F0,$9F,$A1,$7C,$8F,$F0,$96,$C7,$F0,$FD,$9F,$29,$F5,$B9
	.dc.b	$7F,$81,$D6,$B4,$FD,$09,$FE,$39,$F0,$7F,$6A,$DC,$3F,$D6,$9E,$E7
	.dc.b	$F1,$26,$CE,$BE,$21,$F1,$9F,$C6,$6B,$57,$D2,$73,$7F,$C7,$E7,$38
	.dc.b	$FF,$14,$FA,$17,$C7,$7F,$21,$6D,$6F,$C2,$C4,$FF,$25,$FA,$17,$8F
	.dc.b	$E6,$6D,$7F,$B2,$5B,$9F,$EA,$5A,$FF,$12,$F6,$FF,$AB,$B7,$CE,$F7
	.dc.b	$C6,$23,$F9,$13,$50,$7F,$A9,$67,$FC,$B1,$B7,$FC,$A7,$EC,$1F,$23
	.dc.b	$FC,$EE,$73,$7F,$77,$E9,$7F,$24,$7E,$4F,$E0,$2E,$1F,$6F,$D0,$FA
	.dc.b	$3C,$BF,$D2,$2F,$F8,$13,$5F,$E5,$6D,$7F,$9C,$79,$4F,$67,$FA,$53
	.dc.b	$6F,$FA,$76,$9F,$D9,$1F,$EA,$11,$9A,$FE,$43,$F4,$0F,$92,$7C,$27
	.dc.b	$F3,$68,$C3,$63,$37,$FC,$56,$72,$FF,$BE,$5F,$25,$7C,$86,$A6,$98
	.dc.b	$7E,$6F,$AF,$F3,$BB,$7F,$D3,$31,$FD,$22,$31,$E4,$7F,$23,$6C,$64
	.dc.b	$4F,$05,$F1,$AF,$80,$FE,$C5,$1A,$BF,$EB,$D3,$9E,$5E,$73,$C2,$B7
	.dc.b	$7C,$EB,$53,$C4,$FC,$1F,$DB,$7E,$81,$FE,$6C,$1C,$3F,$CE,$5A,$FF
	.dc.b	$62,$6B,$F1,$7D,$8F,$C3,$BF,$2D,$AF,$F6,$8C,$FF,$D9,$BC,$FF,$36
	.dc.b	$73,$3F,$DB,$3B,$3F,$DA,$2F,$F8,$2D,$AF,$EF,$1B,$9F,$F8,$0B,$7B
	.dc.b	$BF,$F7,$FF,$C2,$DE,$F7,$AB,$04,$61,$F6,$B5,$FE,$A8,$D7,$F8,$EF
	.dc.b	$92,$FA,$7F,$F0,$56,$3F,$E2,$C8,$C4,$FF,$2C,$9C,$E1,$FC,$87,$DF
	.dc.b	$B3,$FA,$EF,$A1,$F8,$38,$7F,$86,$1B,$5F,$E1,$56,$CF,$ED,$3C,$A7
	.dc.b	$C7,$F2,$84,$F9,$5E,$1A,$7B,$63,$11,$DE,$3E,$07,$D0,$7F,$B6,$63
	.dc.b	$ED,$B5,$FE,$E4,$D7,$FC,$36,$D7,$FC,$51,$6F,$90,$F1,$36,$3F,$C7
	.dc.b	$CD,$7F,$97,$FD,$0F,$FA,$E9,$FF,$21,$7D,$6F,$F9,$2D,$FF,$22,$7D
	.dc.b	$7B,$E7,$BD,$8D,$8D,$F8,$78,$9F,$B1,$39,$9F,$F0,$43,$5F,$F1,$8B
	.dc.b	$5F,$F1,$F7,$D0,$FF,$8D,$38,$3F,$C8,$AD,$E0,$CC,$FD,$4E,$CF,$E9
	.dc.b	$3E,$5F,$D6,$F5,$3F,$AD,$7A,$17,$9F,$E9,$C7,$F5,$DE,$CC,$4E,$5F
	.dc.b	$D2,$C6,$70,$FF,$01,$57,$E6,$9F,$EC,$53,$AF,$61,$F3,$FD,$DB,$3F
	.dc.b	$E1,$B8,$F1,$7A,$4F,$9C,$D4,$78,$FA,$25,$6C,$FE,$F7,$F0,$FE,$7F
	.dc.b	$87,$91,$83,$FB,$FB,$E6,$CC,$D2,$47,$A3,$DA,$F8,$1E,$BC,$8F,$F0
	.dc.b	$3C,$57,$FC,$6B,$5F,$E7,$87,$DF,$3D,$4F,$77,$CF,$7B,$49,$1F,$E4
	.dc.b	$8D,$1F,$47,$88,$FF,$05,$FA,$FF,$3C,$C7,$3F,$EB,$7E,$77,$E9,$78
	.dc.b	$F2,$BF,$C3,$31,$5F,$1C,$7B,$48,$FF,$8C,$FE,$6B,$FC,$F6,$FF,$96
	.dc.b	$1A,$FF,$91,$5B,$5F,$EA,$67,$FA,$74,$7F,$97,$BD,$1F,$EE,$6D,$C3
	.dc.b	$F0,$D1,$88,$E5,$F1,$5F,$B5,$FA,$0F,$F0,$BC,$7F,$A8,$5A,$FF,$84
	.dc.b	$9B,$87,$FA,$61,$7D,$17,$1E,$59,$5C,$5E,$9F,$01,$FC,$07,$E7,$7F
	.dc.b	$DC,$93,$FE,$AD,$F8,$EF,$D8,$63,$FE,$9F,$F2,$35,$E6,$7F,$07,$9F
	.dc.b	$DA,$57,$FA,$59,$AF,$FA,$72,$7F,$92,$46,$23,$FD,$45,$E8,$71,$23
	.dc.b	$FE,$BB,$F4,$1F,$33,$FC,$4D,$FF,$61,$46,$7E,$1F,$06,$7E,$2F,$5E
	.dc.b	$C4,$9E,$4E,$9F,$A9,$F1,$9F,$33,$C7,$91,$83,$E9,$3F,$B3,$F8,$B9
	.dc.b	$18,$3F,$D9,$DC,$1F,$D3,$3D,$B7,$CC,$FF,$AF,$E3,$0D,$F2,$3F,$D8
	.dc.b	$53,$AA,$FF,$10,$D7,$F5,$C8,$C4,$23,$FA,$BF,$99,$FE,$BE,$6C,$73
	.dc.b	$3F,$C5,$31,$9A,$FF,$67,$3B,$F1,$FB,$E1,$E3,$6D,$78,$1A,$F7,$AC
	.dc.b	$F9,$5F,$E9,$8F,$A1,$BE,$59,$A6,$BC,$4F,$C6,$E9,$EB,$AE,$18,$7F
	.dc.b	$84,$32,$FF,$39,$6D,$D3,$C7,$FF,$29,$CE,$45,$70,$57,$F1,$7C,$D7
	.dc.b	$D6,$3E,$47,$D3,$1F,$DB,$3C,$8B,$C1,$FB,$CC,$3A,$4F,$0A,$CD,$49
	.dc.b	$CD,$6D,$EF,$A7,$F8,$97,$0E,$D1,$CB,$E4,$7E,$A4,$FF,$8A,$5B,$1F
	.dc.b	$E2,$46,$E3,$FE,$DA,$E1,$F2,$C6,$B4,$FF,$60,$FD,$07,$CA,$33,$E8
	.dc.b	$DC,$3E,$14,$E9,$BC,$6A,$DC,$60,$FA,$64,$38,$F1,$3E,$8F,$11,$EC
	.dc.b	$3F,$D4,$1C,$BF,$44,$CF,$17,$D7,$A7,$9B,$F6,$C9,$5D,$3E,$4A,$BC
	.dc.b	$F3,$FF,$6D,$FC,$EF,$E1,$0C,$E1,$F9,$8E,$9F,$E7,$B5,$F7,$4F,$CD
	.dc.b	$7A,$3F,$8E,$83,$7D,$29,$3B,$9C,$B6,$72,$FA,$CE,$66,$48,$66,$3C
	.dc.b	$2F,$9F,$3F,$94,$FC,$6D,$BF,$30,$7E,$67,$AF,$F3,$FF,$47,$FE,$3C
	.dc.b	$CB,$F3,$5E,$0F,$F6,$1B,$FA,$AD,$7E,$14,$EF,$9F,$FD,$33,$B7,$FA
	.dc.b	$67,$DF,$FD,$4F,$C6,$7F,$8A,$2F,$EE,$B5,$F2,$1A,$F1,$A3,$A7,$8F
	.dc.b	$91,$CB,$E8,$7F,$63,$EC,$4A,$E7,$1D,$BE,$22,$47,$67,$C3,$F4,$3E
	.dc.b	$CF,$A0,$93,$F8,$ED,$91,$E6,$F9,$EF,$83,$F3,$BF,$0B,$E7,$25,$EE
	.dc.b	$3D,$7E,$BC,$8F,$5B,$A7,$F0,$9B,$F4,$7D,$F8,$7C,$23,$E1,$FC,$7D
	.dc.b	$8F,$87,$E8,$FC,$63,$0F,$27,$0E,$8E,$67,$A7,$C6,$FA,$E4,$7D,$9C
	.dc.b	$D5,$FA,$BE,$7F,$DC,$31,$D1,$FC,$57,$7F,$F4,$DC,$FC,$39,$D3,$D4
	.dc.b	$FA,$FB,$7D,$A2,$3F,$6B,$E3,$AB,$F0,$37,$FB,$E6,$30,$FE,$45,$B2
	.dc.b	$F3,$5F,$82,$AD,$74,$FF,$45,$F6,$FF,$9B,$CD,$3F,$9D,$94,$78,$9F
	.dc.b	$D0,$63,$71,$FD,$21,$FD,$67,$BD,$EC,$CB,$F6,$4D,$BE,$51,$FB,$59
	.dc.b	$DE,$FF,$AF,$AD,$5F,$33,$11,$1F,$D6,$46,$9F,$EA,$93,$C9,$CC,$EE
	.dc.b	$BB,$45,$BF,$EC,$57,$C7,$C0,$37,$0E,$57,$E8,$C3,$86,$1E,$33,$A9
	.dc.b	$23,$4F,$EF,$E7,$8E,$1D,$AB,$8C,$F9,$8B,$87,$82,$2D,$70,$9F,$EE
	.dc.b	$38,$79,$4F,$23,$C1,$C4,$FF,$15,$9E,$13,$FD,$D6,$72,$F8,$CE,$13
	.dc.b	$D9,$F9,$47,$F7,$98,$FC,$8B,$63,$FD,$BB,$5E,$58,$10,$BF,$1D,$FB
	.dc.b	$8F,$A3,$FF,$71,$B6,$34,$6B,$95,$3A,$6C,$68,$FF,$06,$7C,$07,$F2
	.dc.b	$47,$FC,$18,$D7,$FC,$30,$D7,$FC,$2A,$D7,$8F,$85,$FC,$DE,$3F,$96
	.dc.b	$EB,$FD,$07,$D4,$FC,$BD,$7F,$33,$9C,$ED,$D2,$F8,$1F,$CF,$EB,$77
	.dc.b	$FD,$26,$B7,$7E,$0A,$DB,$FC,$F6,$B7,$7C,$E3,$B4,$EA,$FC,$5B,$DA
	.dc.b	$D4,$75,$9F,$04,$E6,$BB,$1C,$8E,$CE,$5F,$79,$99,$EA,$DA,$EA,$47
	.dc.b	$FC,$77,$7E,$67,$AC,$F8,$5E,$06,$AB,$B6,$8C,$7E,$0C,$F8,$2B,$B9
	.dc.b	$C4,$B8,$77,$8F,$F2,$36,$1D,$DE,$11,$FE,$51,$8C,$5F,$5C,$7E,$0D
	.dc.b	$78,$8E,$38,$FF,$10,$7C,$87,$DA,$2B,$FE,$51,$AF,$BA,$D3,$9E,$1F
	.dc.b	$E0,$1F,$53,$BC,$FC,$1F,$0D,$F4,$6F,$0F,$0B,$C4,$B1,$86,$E3,$E3
	.dc.b	$AE,$33,$E0,$CB,$FB,$7D,$F3,$AD,$36,$3D,$15,$C2,$7F,$C5,$B3,$9C
	.dc.b	$BF,$CA,$CB,$C4,$EF,$7F,$E1,$DF,$81,$75,$FA,$5F,$F3,$DE,$1F,$E3
	.dc.b	$67,$2D,$E0,$D4,$F9,$6B,$FC,$B7,$5A,$7C,$6D,$AF,$DE,$64,$8C,$4F
	.dc.b	$F8,$29,$E1,$97,$F9,$AD,$B8,$67,$B7,$BE,$1E,$CF,$13,$E5,$FC,$C7
	.dc.b	$CC,$6C,$64,$F8,$AF,$97,$2F,$A6,$DD,$4D,$F1,$39,$57,$C8,$6E,$9F
	.dc.b	$E4,$1A,$78,$BF,$68,$FB,$DF,$46,$F9,$BC,$61,$FE,$5C,$87,$C6,$F8
	.dc.b	$5B,$1F,$75,$B8,$7D,$86,$C7,$C6,$5F,$2D,$D8,$FC,$4F,$93,$67,$EE
	.dc.b	$F9,$1F,$8C,$3F,$1B,$E4,$D8,$FC,$0F,$77,$BC,$8F,$E2,$1F,$85,$E2
	.dc.b	$7E,$50,$FC,$AF,$8F,$67,$F1,$3E,$E7,$E6,$0F,$CC,$F8,$F6,$3D,$DE
	.dc.b	$3F,$B4,$8F,$B5,$F9,$DD,$1F,$A0,$3F,$43,$EE,$D9,$FF,$33,$7A,$4F
	.dc.b	$4B,$F6,$F1,$FB,$B8,$C3,$FE,$65,$31,$1F,$8F,$E2,$4E,$9F,$DC,$18
	.dc.b	$8D,$46,$67,$FA,$3B,$BE,$DF,$F6,$7B,$70,$FE,$A0,$BC,$EE,$F5,$FA
	.dc.b	$AE,$DF,$DC,$9F,$BA,$F2,$AC,$7E,$B9,$B9,$7E,$1D,$B8,$FF,$B1,$7C
	.dc.b	$47,$D1,$FF,$05,$F5,$9F,$33,$FC,$1E,$73,$8F,$FA,$8A,$31,$5B,$6C
	.dc.b	$64,$FF,$5F,$FD,$1B,$A7,$A7,$D1,$7F,$31,$CD,$39,$AF,$F5,$6C,$E6
	.dc.b	$31,$A7,$FA,$9E,$B7,$79,$EA,$D3,$70,$EE,$DA,$E1,$7C,$B5,$FF,$6C
	.dc.b	$7D,$1B,$E1,$7C,$5E,$EB,$E0,$3C,$93,$9C,$BF,$D9,$D1,$8A,$91,$FE
	.dc.b	$DA,$97,$BD,$8C,$BF,$D6,$5C,$4F,$91,$F9,$47,$C9,$F1,$2E,$1F,$EC
	.dc.b	$CE,$4F,$DA,$36,$3E,$77,$89,$B8,$FC,$8F,$1F,$E3,$CF,$C8,$5D,$8F
	.dc.b	$CA,$BF,$8C,$C7,$FC,$2B,$D1,$E1,$9F,$F0,$64,$7E,$4A,$74,$D8,$CC
	.dc.b	$BE,$97,$FC,$0B,$F1,$1F,$D3,$63,$95,$23,$13,$AA,$FC,$9D,$DA,$79
	.dc.b	$6E,$AE,$38,$7F,$84,$23,$15,$B5,$F6,$AF,$A4,$C5,$F0,$C7,$FD,$73
	.dc.b	$39,$7F,$C6,$38,$73,$9D,$63,$E0,$AD,$3E,$7A,$FF,$60,$CE,$A3,$FA
	.dc.b	$F7,$83,$E7,$7F,$B0,$7C,$85,$F4,$D3,$63,$5F,$4B,$FA,$2C,$3C,$27
	.dc.b	$F9,$06,$B6,$E2,$FF,$C8,$B3,$B7,$C4,$7F,$92,$27,$6E,$07,$C6,$3C
	.dc.b	$BF,$23,$6F,$A4,$FF,$0A,$D7,$35,$E9,$E1,$7D,$37,$DC,$6D,$FD,$C8
	.dc.b	$C1,$9F,$8A,$E0,$F1,$6C,$76,$CD,$F4,$BF,$0F,$DD,$F8,$2D,$9D,$77
	.dc.b	$E1,$A7,$18,$FB,$3F,$19,$F6,$A6,$CF,$C5,$E5,$FC,$24,$7E,$07,$91
	.dc.b	$F1,$3A,$E0,$F8,$F5,$F8,$48,$C7,$37,$E2,$7B,$7F,$2D,$8F,$F9,$9E
	.dc.b	$91,$8B,$FF,$37,$C6,$3A,$7F,$48,$DA,$E1,$DB,$FD,$61,$F2,$5F,$1C
	.dc.b	$3F,$2D,$39,$C3,$C7,$39,$AF,$F4,$0D,$AB,$F8,$35,$6B,$93,$F3,$91
	.dc.b	$8C,$3D,$11,$8F,$02,$61,$F0,$EB,$F8,$4B,$63,$F4,$98,$FD,$AC,$3D
	.dc.b	$6D,$7F,$C2,$AD,$7C,$6D,$7F,$63,$4F,$8B,$E7,$C9,$E4,$6E,$1F,$03
	.dc.b	$93,$E2,$E0,$69,$23,$E3,$FC,$A7,$92,$3D,$AD,$8F,$57,$C5,$7A,$39
	.dc.b	$3E,$0B,$63,$24,$CF,$EE,$E7,$FA,$2D,$BE,$7D,$7C,$0E,$3F,$19,$2F
	.dc.b	$EF,$66,$8F,$B7,$7F,$ED,$69,$FB,$31,$F4,$3E,$17,$94,$7E,$A3,$DC
	.dc.b	$76,$FE,$FD,$B1,$FA,$CF,$8A,$FA,$3A,$7E,$CE,$72,$FE,$D6,$3E,$54
	.dc.b	$FF,$2A,$6C,$7F,$AD,$DA,$FF,$AF,$9A,$FF,$B1,$5F,$E9,$91,$DF,$83
	.dc.b	$F8,$84,$62,$7B,$47,$F1,$3F,$A3,$F8,$95,$FE,$CD,$6C,$FF,$B4,$3E
	.dc.b	$23,$FC,$01,$3F,$C1,$27,$32,$9F,$E2,$0F,$FB,$41,$AF,$FB,$5D,$AF
	.dc.b	$3F,$3F,$FA,$1C,$3C,$5F,$E4,$8D,$BF,$E2,$3F,$43,$F0,$38,$BF,$D2
	.dc.b	$13,$9C,$3F,$8D,$4E,$6B,$FD,$C8,$D8,$FF,$73,$BF,$3A,$BF,$D3,$0E
	.dc.b	$7A,$7F,$73,$DB,$FA,$03,$FA,$57,$D0,$3E,$67,$FD,$DF,$8F,$ED,$A7
	.dc.b	$F9,$C7,$CC,$7F,$25,$CB,$FD,$D0,$D7,$FD,$DE,$D7,$FD,$EC,$D7,$C7
	.dc.b	$3F,$D0,$71,$3F,$D4,$FE,$3D,$EC,$E6,$3F,$A3,$65,$F2,$2C,$D2,$6F
	.dc.b	$FD,$B1,$FE,$A3,$1F,$F0,$01,$A8,$8E,$1B,$DF,$FD,$D1,$D7,$A3,$BF
	.dc.b	$57,$FD,$3F,$1C,$4C,$87,$2E,$63,$87,$C3,$6E,$37,$3A,$F8,$D7,$BD
	.dc.b	$FD,$EE,$79,$57,$59,$CF,$1F,$29,$F4,$46,$1D,$98,$AE,$57,$E2,$FA
	.dc.b	$1F,$F1,$64,$FF,$84,$1A,$F8,$C3,$87,$CC,$D7,$DC,$FF,$8C,$DE,$5E
	.dc.b	$B7,$F8,$69,$E2,$D8,$D4,$75,$F9,$5F,$F6,$D3,$FD,$A1,$AF,$F6,$F6
	.dc.b	$E3,$FE,$13,$FC,$31,$F4,$FF,$36,$3F,$DD,$9E,$6B,$BF,$C0,$DC,$3F
	.dc.b	$C5,$6D,$7D,$8D,$7F,$9A,$FC,$D7,$3E,$4F,$E6,$AD,$8F,$F1,$2F,$8A
	.dc.b	$8C,$72,$3F,$07,$7C,$DE,$EF,$FD,$6B,$F2,$17,$DA,$FF,$7A,$47,$F8
	.dc.b	$11,$AF,$F8,$2F,$BF,$A3,$FE,$08,$FA,$2F,$E8,$4D,$F1,$7A,$46,$23
	.dc.b	$FC,$25,$5C,$B8,$3F,$47,$1F,$97,$9F,$9A,$FE,$2D,$AF,$E4,$DA,$EF
	.dc.b	$97,$EB,$7D,$17,$C5,$AE,$72,$8C,$70,$7E,$6D,$B1,$DF,$CE,$F8,$1F
	.dc.b	$43,$FE,$90,$8F,$F1,$0B,$FC,$55,$BE,$3E,$BC,$0D,$BF,$85,$9D,$4F
	.dc.b	$E3,$DA,$FE,$8D,$AF,$EA,$F6,$7D,$0F,$CF,$6C,$7F,$8C,$1A,$FF,$8D
	.dc.b	$E7,$FC,$F8,$FF,$79,$CB,$EE,$78,$FF,$E1,$2F,$9F,$BC,$E7,$3E,$7F
	.dc.b	$C2,$DF,$FC,$35,$1F,$E7,$4A,$E1,$87,$F5,$E8,$C5,$F1,$8F,$F0,$F7
	.dc.b	$99,$FE,$BF,$7F,$C5,$4D,$7F,$C4,$ED,$BF,$F3,$73,$F2,$63,$3F,$0F
	.dc.b	$E4,$E7,$91,$CA,$3D,$33,$EC,$8F,$F0,$FF,$CE,$FA,$8C,$FC,$0D,$8F
	.dc.b	$EE,$71,$E6,$23,$E6,$F2,$3E,$0F,$0C,$8F,$3C,$57,$FC,$37,$1E,$E2
	.dc.b	$3E,$EF,$57,$C6,$47,$DF,$EA,$D7,$FC,$67,$EC,$FC,$E3,$1A,$FC,$0F
	.dc.b	$43,$E8,$FB,$92,$BE,$3C,$57,$FC,$57,$1F,$38,$8F,$F9,$2F,$EC,$3F
	.dc.b	$DF,$CF,$F9,$5F,$E5,$9F,$13,$F7,$9B,$1F,$E5,$B9,$DB,$FE,$58,$8C
	.dc.b	$B7,$E0,$38,$1C,$6B,$F7,$9F,$42,$C6,$F8,$F2,$FF,$CF,$F3,$96,$C7
	.dc.b	$0E,$79,$18,$3F,$D0,$BF,$36,$F9,$3F,$E1,$0B,$E9,$E9,$88,$B3,$FE
	.dc.b	$C4,$9D,$36,$7F,$97,$46,$23,$F9,$C7,$C2,$BE,$6B,$F8,$A3,$63,$FD
	.dc.b	$2A,$D7,$FD,$39,$3F,$D0,$38,$7F,$80,$FA,$2F,$90,$FE,$38,$FF,$6E
	.dc.b	$AF,$E6,$DE,$CE,$27,$2E,$BE,$2A,$4E,$8F,$A1,$C1,$F4,$FE,$4E,$46
	.dc.b	$0F,$B0,$E1,$FA,$8F,$ED,$B9,$7E,$64,$79,$BE,$97,$E3,$FC,$74,$8F
	.dc.b	$97,$15,$FE,$B1,$AF,$FB,$F9,$F9,$13,$F1,$8F,$8F,$E8,$7F,$71,$D3
	.dc.b	$FB,$B5,$7F,$2D,$F7,$7E,$05,$FF,$80,$63,$13,$AB,$E1,$3F,$EE,$8F
	.dc.b	$0B,$F8,$02,$BE,$3E,$32,$73,$18,$6D,$FF,$A5,$1F,$E8,$35,$C6,$3F
	.dc.b	$D0,$BE,$F7,$6B,$FF,$53,$B7,$C7,$39,$BE,$77,$FE,$0D,$FA,$2F,$8F
	.dc.b	$8F,$AB,$DC,$46,$32,$FF,$0B,$36,$B8,$75,$FB,$2B,$7C,$57,$D2,$FF
	.dc.b	$47,$57,$F8,$7B,$0F,$F0,$F4,$62,$3D,$07,$86,$79,$7C,$37,$C8,$8F
	.dc.b	$F1,$03,$F4,$DD,$DF,$FA,$17,$7D,$F0,$F0,$64,$F7,$38,$BA,$71,$3C
	.dc.b	$5D,$70,$7F,$B3,$3E,$7B,$E3,$1C,$A3,$CB,$1F,$4F,$C9,$F9,$51,$0E
	.dc.b	$3C,$08,$F9,$86,$63,$8F,$17,$F8,$D7,$87,$ED,$4F,$D1,$F1,$3E,$97
	.dc.b	$8B,$2F,$9C,$57,$C3,$87,$FB,$A1,$EF,$3C,$5F,$70,$E5,$7D,$17,$E3
	.dc.b	$A3,$0D,$8C,$C7,$05,$EA,$66,$7F,$CA,$1E,$87,$F9,$2E,$77,$5F,$DF
	.dc.b	$0E,$67,$F9,$D8,$FD,$0C,$F0,$69,$D7,$E2,$7C,$07,$57,$E3,$EF,$F9
	.dc.b	$12,$3F,$2B,$A7,$F3,$1F,$43,$FE,$3A,$C7,$FC,$C5,$C2,$8C,$57,$FB
	.dc.b	$D1,$AA,$35,$F2,$1E,$38,$C7,$C4,$7F,$8A,$38,$BF,$5A,$DE,$AF,$F7
	.dc.b	$EB,$67,$F6,$87,$EA,$5F,$F2,$5C,$F0,$AC,$19,$F1,$75,$F3,$FD,$F1
	.dc.b	$FF,$5A,$36,$CD,$31,$F3,$BD,$49,$8F,$E4,$A5,$F3,$FF,$0B,$E7,$64
	.dc.b	$FE,$5B,$54,$79,$7E,$67,$E3,$FC,$EF,$86,$4D,$25,$7D,$73,$E1,$FA
	.dc.b	$3E,$D2,$3D,$7E,$77,$E0,$7A,$5F,$07,$19,$3E,$58,$AF,$F8,$FB,$6F
	.dc.b	$F6,$94,$FB,$C9,$1D,$E3,$2F,$B7,$E8,$BE,$61,$CB,$8B,$F8,$FE,$7F
	.dc.b	$E9,$D9,$F7,$78,$1F,$82,$3F,$07,$E5,$D8,$E9,$3B,$7C,$4D,$E6,$FC
	.dc.b	$3B,$63,$F0,$D5,$F8,$CF,$8E,$91,$D8,$FA,$1A,$FD,$F2,$3F,$4F,$E3
	.dc.b	$6B,$EF,$E3,$4F,$DD,$24,$7B,$E3,$2F,$DB,$E1,$FC,$51,$CB,$4F,$E8
	.dc.b	$B1,$FB,$7F,$99,$FD,$A1,$1F,$DA,$7C,$9D,$7F,$61,$DE,$9F,$CE,$47
	.dc.b	$13,$EE,$14,$DB,$FB,$FF,$94,$F9,$97,$E0,$A4,$63,$B3,$FA,$CD,$77
	.dc.b	$AD,$4E,$6F,$FA,$D3,$5E,$D1,$AD,$BF,$85,$4E,$70,$EB,$1F,$C3,$6F
	.dc.b	$C4,$E8,$97,$FD,$00,$E7,$87,$F6,$AB,$E4,$49,$E5,$3C,$30,$F0,$3C
	.dc.b	$70,$EE,$DC,$3B,$B9,$BE,$4E,$4E,$58,$77,$C2,$C7,$F7,$7A,$E3,$87
	.dc.b	$62,$57,$C6,$9D,$9F,$CB,$EF,$AC,$FC,$DA,$E3,$5D,$6B,$C5,$7E,$A6
	.dc.b	$92,$BA,$BE,$13,$E2,$9F,$CC,$6F,$F3,$CF,$33,$13,$FE,$ED,$5F,$23
	.dc.b	$FC,$13,$E8,$B0,$D8,$FF,$79,$35,$1F,$F0,$07,$99,$3F,$E4,$5F,$8E
	.dc.b	$78,$34,$95,$FE,$24,$8C,$B6,$3F,$A5,$46,$32,$FC,$77,$C2,$7F,$B3
	.dc.b	$1F,$C8,$46,$74,$FF,$13,$B6,$FF,$C4,$CD,$7F,$AD,$57,$19,$DD,$F6
	.dc.b	$24,$FE,$06,$74,$F2,$8C,$DF,$81,$FE,$BD,$3A,$9F,$EC,$D7,$DF,$63
	.dc.b	$4C,$4E,$E3,$35,$D3,$0F,$1D,$F8,$79,$B0,$F8,$4F,$CA,$61,$CA,$33
	.dc.b	$1E,$0A,$B6,$C3,$B5,$F8,$67,$31,$87,$D3,$78,$25,$F6,$8C,$B4,$F2
	.dc.b	$1F,$C2,$27,$FD,$FB,$3F,$DE,$DB,$1F,$E5,$86,$BF,$E5,$77,$E0,$1F
	.dc.b	$07,$E8,$AC,$F6,$CB,$EE,$BF,$E0,$98,$CD,$7F,$89,$BE,$89,$E1,$FD
	.dc.b	$4E,$C4,$3E,$3B,$0E,$B0,$FF,$18,$B9,$6D,$68,$DF,$89,$63,$24,$C3
	.dc.b	$C1,$19,$7F,$C6,$FF,$00,$EC,$F1,$3F,$E9,$3A,$FF,$1F,$B9,$6F,$0E
	.dc.b	$A6,$46,$BB,$AD,$F4,$9B,$B7,$F9,$DB,$05,$63,$0F,$D6,$EB,$F0,$3C
	.dc.b	$5F,$7F,$C8,$FC,$06,$E7,$F0,$1C,$3F,$0B,$C5,$BA,$7C,$B6,$E7,$FE
	.dc.b	$84,$DF,$FC,$0B,$5C,$4F,$93,$E3,$7C,$C3,$E8,$7D,$13,$E7,$B1,$3F
	.dc.b	$E5,$E8,$7C,$8E,$8D,$FC,$0D,$9D,$7C,$17,$23,$EB,$F9,$76,$3E,$BF
	.dc.b	$95,$F6,$8F,$B7,$E5,$D8,$FB,$7E,$07,$DE,$3E,$FF,$97,$63,$EF,$F8
	.dc.b	$FF,$82,$3F,$07,$E5,$58,$FC,$1F,$BD,$F8,$63,$F0,$FE,$5D,$8F,$57
	.dc.b	$93,$EE,$3D,$FE,$55,$8F,$C5,$F5,$FE,$38,$FC,$7F,$BF,$63,$FC,$CF
	.dc.b	$EA,$3E,$8B,$F4,$3D,$68,$C4,$FF,$6C,$6C,$7F,$98,$5B,$1F,$CA,$A3
	.dc.b	$F9,$5C,$6B,$AB,$FB,$EB,$70,$D2,$F8,$BF,$E2,$4E,$6B,$8C,$63,$0C
	.dc.b	$B7,$0F,$D2,$F9,$BF,$A8,$F0,$7F,$52,$D8,$C9,$2F,$F6,$5E,$1E,$A3
	.dc.b	$FD,$3B,$E4,$3E,$8B,$F6,$8D,$8C,$FC,$B6,$27,$2D,$8C,$9C,$A3,$F9
	.dc.b	$7E,$0F,$A2,$75,$61,$75,$D8,$C3,$6F,$FC,$A5,$5C,$67,$F8,$8B,$63
	.dc.b	$35,$D1,$B3,$C7,$3F,$F6,$0C,$BA,$F1,$C5,$FE,$D7,$FA,$27,$BB,$E6
	.dc.b	$5F,$3F,$63,$0D,$CB,$FD,$87,$3C,$1B,$F0,$B9,$CB,$FD,$83,$C1,$FA
	.dc.b	$B9,$F9,$7E,$27,$CA,$F0,$FE,$5B,$63,$27,$D0,$8C,$BF,$90,$F9,$0F
	.dc.b	$F5,$55,$FE,$21,$AF,$E4,$5B,$D0,$6F,$D0,$4A,$FF,$00,$4A,$31,$5F
	.dc.b	$93,$3F,$BE,$FC,$5B,$C8,$5F,$F8,$03,$E2,$ED,$7E,$62,$EC,$62,$17
	.dc.b	$C6,$B8,$1F,$9A,$AD,$BF,$C7,$BE,$85,$F2,$58,$BF,$F5,$94,$62,$78
	.dc.b	$C6,$3E,$47,$5E,$8A,$B5,$2C,$BC,$B3,$9A,$DD,$1B,$F4,$38,$7C,$16
	.dc.b	$FA,$27,$CC,$E1,$B9,$78,$A7,$5F,$33,$9A,$D3,$27,$55,$F4,$5B,$1D
	.dc.b	$B8,$18,$FD,$62,$4F,$D4,$7E,$CE,$32,$3E,$BB,$71,$FA,$C7,$9F,$6E
	.dc.b	$91,$E7,$6C,$F8,$BE,$33,$D4,$7A,$FE,$36,$C7,$99,$C3,$F3,$B6,$F9
	.dc.b	$9F,$17,$8C,$7E,$64,$FD,$7F,$8A,$F9,$67,$CC,$F8,$7B,$1C,$78,$51
	.dc.b	$98,$C6,$D6,$F9,$9F,$EE,$FE,$97,$A9,$F9,$CC,$FF,$C3,$8D,$8C,$DF
	.dc.b	$C0,$9C,$B6,$1E,$AF,$E1,$1D,$DF,$6A,$FF,$13,$4E,$6F,$5D,$FB,$EF
	.dc.b	$B8,$FC,$1C,$7F,$7E,$D9,$CE,$3C,$E5,$DD,$8E,$8F,$C7,$C3,$E3,$7D
	.dc.b	$04,$C3,$CC,$FA,$FE,$3B,$CC,$F9,$DB,$1E,$1F,$8C,$F8,$04,$7D,$CF
	.dc.b	$C3,$F8,$CF,$73,$EF,$6C,$7A,$7D,$3F,$85,$F1,$9F,$0D,$B1,$92,$69
	.dc.b	$EB,$F5,$35,$1F,$07,$E2,$BE,$81,$1F,$9C,$FD,$3F,$87,$F9,$CF,$CF
	.dc.b	$6C,$7E,$97,$E3,$7E,$13,$FE,$8D,$6D,$FE,$61,$AF,$F0,$F6,$BF,$D8
	.dc.b	$D6,$9F,$C7,$1B,$3C,$E3,$07,$F2,$0F,$A0,$7A,$5F,$BA,$AD,$DF,$E3
	.dc.b	$84,$EA,$FC,$D7,$FC,$03,$E3,$7F,$7B,$5F,$AC,$6B,$FC,$8D,$AF,$F7
	.dc.b	$76,$BF,$EA,$25,$E1,$FF,$32,$6D,$7F,$AB,$5B,$3D,$67,$D1,$C3,$FC
	.dc.b	$43,$0F,$8B,$C7,$FC,$A5,$FE,$95,$5E,$97,$38,$7F,$81,$DC,$CE,$F9
	.dc.b	$7F,$41,$C9,$CD,$FE,$37,$F4,$FF,$C4,$70,$F8,$4D,$AF,$E6,$DF,$45
	.dc.b	$FD,$BB,$1F,$E3,$93,$97,$FC,$29,$1F,$DA,$6B,$F8,$7B,$5F,$E8,$CD
	.dc.b	$77,$E9,$7E,$E7,$B9,$ED,$FF,$36,$6D,$FF,$0D,$5E,$EE,$67,$FA,$07
	.dc.b	$D0,$3E,$D5,$EF,$FF,$5E,$8C,$36,$33,$87,$F2,$29,$CB,$FE,$DB,$7F
	.dc.b	$96,$B7,$CB,$FE,$CB,$18,$CF,$84,$7F,$68,$7F,$DB,$7E,$AF,$FB,$73
	.dc.b	$D8,$FF,$40,$47,$F6,$38,$C7,$B7,$FC,$C1,$B1,$91,$3C,$1F,$F7,$5F
	.dc.b	$C0,$7F,$53,$73,$5F,$E3,$FF,$91,$4E,$6B,$84,$63,$8F,$AD,$F2,$6F
	.dc.b	$D1,$FF,$1E,$8F,$F7,$C3,$5F,$E8,$6D,$7F,$BD,$35,$FF,$7B,$E6,$FB
	.dc.b	$5F,$E5,$F1,$F2,$5A,$FF,$80,$63,$5E,$47,$CB,$9C,$CF,$F9,$5F,$CA
	.dc.b	$7D,$27,$EC,$23,$FA,$43,$5F,$E6,$0D,$C7,$FB,$9B,$FE,$49,$8F,$ED
	.dc.b	$5E,$17,$FA,$F9,$FE,$B6,$7F,$BA,$BE,$3E,$B4,$CC,$6D,$FE,$C3,$F5
	.dc.b	$3F,$C1,$9F,$EE,$8D,$7F,$80,$36,$FF,$C2,$92,$F8,$DF,$F3,$84,$7F
	.dc.b	$85,$FE,$89,$F3,$38,$E6,$FF,$3A,$46,$31,$FF,$00,$35,$FE,$B3,$4F
	.dc.b	$B5,$FE,$F7,$1F,$E0,$5F,$80,$FF,$6B,$37,$E5,$3A,$23,$FD,$52,$E1
	.dc.b	$5F,$5B,$FF,$15,$7D,$2F,$F4,$B9,$F8,$AD,$7F,$C0,$2D,$7F,$C7,$0D
	.dc.b	$7F,$D1,$8F,$C3,$73,$E4,$FC,$C8,$C4,$6A,$3E,$7F,$D4,$BE,$63,$C9
	.dc.b	$F4,$BF,$E2,$2A,$FF,$10,$35,$FE,$59,$3B,$73,$F0,$3F,$62,$75,$1F
	.dc.b	$E4,$B6,$BF,$E9,$F6,$BF,$0D,$FE,$52,$FE,$09,$AF,$F8,$79,$AF,$3F
	.dc.b	$33,$F0,$FC,$CF,$29,$F3,$3F,$3A,$31,$C5,$67,$FC,$A9,$B7,$F7,$6A
	.dc.b	$C4,$F2,$9F,$F3,$4D,$70,$C3,$FC,$19,$3D,$BC,$FF,$81,$87,$F9,$31
	.dc.b	$B1,$FE,$F6,$6C,$A9,$1F,$EC,$58,$FF,$16,$3A,$8C,$FC,$77,$9E,$3F
	.dc.b	$D7,$93,$AD,$31,$9E,$7E,$1E,$4F,$28,$AF,$F9,$DF,$0E,$F3,$E7,$24
	.dc.b	$7F,$89,$EF,$FC,$FC,$F8,$BE,$17,$C0,$78,$7E,$0E,$C6,$E3,$E0,$11
	.dc.b	$E5,$F4,$7E,$72,$3E,$CF,$22,$BF,$E2,$FB,$FF,$37,$3F,$03,$E1,$73
	.dc.b	$BF,$C9,$F8,$5F,$81,$F0,$B2,$BE,$2C,$57,$D9,$69,$FA,$44,$8F,$F2
	.dc.b	$17,$85,$F7,$38,$BF,$59,$BF,$F1,$47,$F0,$51,$99,$FA,$8F,$E1,$9F
	.dc.b	$F2,$F7,$D0,$7D,$57,$FC,$8C,$DC,$BF,$D1,$2D,$AE,$B5,$CE,$9F,$46
	.dc.b	$CF,$F7,$4D,$8F,$F4,$33,$5E,$BC,$7A,$CE,$46,$9E,$35,$FC,$0F,$E8
	.dc.b	$BF,$DC,$73,$FE,$8F,$F9,$EF,$E6,$BA,$7F,$AB,$5B,$9F,$F0,$B9,$57
	.dc.b	$07,$FC,$41,$3F,$CE,$1D,$35,$FE,$24,$D4,$D3,$FA,$43,$AF,$85,$FE
	.dc.b	$67,$18,$8D,$C7,$F3,$69,$FE,$F7,$1F,$EA,$56,$BF,$E3,$16,$DF,$FB
	.dc.b	$15,$7C,$B7,$BF,$E5,$A5,$71,$7E,$17,$AD,$FB,$BE,$0F,$95,$87,$D4
	.dc.b	$D3,$06,$79,$75,$B4,$C7,$EE,$BA,$7F,$D7,$1F,$1A,$F9,$CF,$E6,$1B
	.dc.b	$FF,$9E,$A7,$E7,$7C,$47,$F3,$3A,$D4,$60,$FE,$AB,$3B,$7F,$D6,$2D
	.dc.b	$7F,$DD,$0D,$7E,$7F,$BD,$F2,$08,$F2,$7F,$86,$79,$5F,$67,$D7,$70
	.dc.b	$7F,$0B,$FA,$3C,$36,$FF,$C2,$AF,$F0,$FA,$CC,$7F,$0F,$EA,$56,$3A
	.dc.b	$BF,$88,$B6,$33,$3C,$8E,$15,$CB,$D6,$7D,$1F,$F5,$DC,$BC,$75,$FE
	.dc.b	$1C,$8C,$65,$FE,$25,$6C,$67,$8B,$94,$62,$F9,$F0,$E0,$CE,$3D,$EB
	.dc.b	$FA,$D6,$3D,$AB,$F5,$5F,$05,$93,$5C,$37,$2F,$ED,$1D,$5F,$9F,$AD
	.dc.b	$C6,$0F,$F1,$C6,$1F,$D9,$67,$4D,$8C,$DF,$87,$82,$47,$C3,$BE,$6B
	.dc.b	$1F,$83,$C1,$FC,$EB,$E9,$7E,$57,$C8,$4A,$F9,$53,$96,$F0,$A7,$51
	.dc.b	$CF,$87,$FA,$34,$7D,$3B,$FA,$9F,$1D,$2B,$8B,$F5,$FE,$97,$11,$C4
	.dc.b	$FB,$DF,$23,$F6,$CC,$E7,$F5,$A3,$F7,$3F,$1F,$F8,$47,$F7,$7F,$1F
	.dc.b	$7A,$0D,$9C,$32,$7F,$08,$70,$2C,$A7,$50,$E3,$FC,$FD,$7F,$11,$6C
	.dc.b	$7D,$33,$F8,$6C,$65,$FE,$CD,$E8,$7E,$4C,$8F,$C8,$F8,$5F,$CC,$7D
	.dc.b	$27,$F8,$F7,$2F,$A2,$DE,$0F,$E7,$AD,$AF,$ED,$CB,$E8,$FF,$41,$F1
	.dc.b	$9F,$E3,$09,$F9,$1C,$88,$C6,$1F,$71,$B8,$7F,$79,$AE,$3E,$F4,$7E
	.dc.b	$EC,$F7,$BE,$58,$76,$F7,$25,$69,$F0,$F8,$3E,$2F,$95,$93,$C9,$F2
	.dc.b	$FD,$17,$BB,$E8,$24,$FE,$2B,$64,$7A,$BE,$73,$E1,$F9,$52,$B4,$FC
	.dc.b	$6F,$88,$F7,$FC,$DC,$9F,$3B,$F2,$BE,$33,$E1,$FC,$77,$C4,$C6,$57
	.dc.b	$B2,$2B,$F1,$B8,$7C,$2F,$C0,$9E,$2D,$3E,$1F,$D0,$7C,$E3,$8F,$2F
	.dc.b	$D6,$F0,$BE,$C9,$27,$D9,$F3,$F5,$FA,$7B,$FF,$A9,$9F,$81,$3F,$80
	.dc.b	$9D,$57,$E0,$3E,$3B,$F0,$66,$3A,$3F,$BE,$E7,$FE,$C4,$C3,$EA,$EB
	.dc.b	$F8,$03,$F0,$3B,$D8,$C9,$0F,$73,$97,$E0,$70,$7E,$20,$C6,$3F,$AB
	.dc.b	$EE,$7C,$BF,$EA,$49,$5F,$A4,$F9,$7A,$FE,$D4,$FE,$D3,$18,$AE,$53
	.dc.b	$AE,$FF,$E8,$B3,$FA,$3A,$FE,$E2,$D7,$89,$FA,$92,$9D,$6D,$F2,$FF
	.dc.b	$27,$0F,$33,$75,$FE,$FB,$7E,$17,$FB,$FC,$66,$FF,$BF,$B5,$EB,$5C
	.dc.b	$B8,$3F,$7F,$39,$C3,$BC,$7F,$02,$C3,$F8,$54,$E4,$91,$9B,$E2,$C7
	.dc.b	$D1,$1C,$E7,$8C,$E6,$FC,$15,$ED,$9C,$BC,$27,$58,$7B,$22,$9E,$A8
	.dc.b	$AF,$AA,$3D,$F1,$8C,$3D,$33,$D4,$C6,$1E,$87,$07,$72,$61,$E4,$3F
	.dc.b	$05,$87,$92,$3F,$BE,$CF,$E6,$5B,$7F,$E0,$48,$C4,$7F,$BF,$78,$DF
	.dc.b	$49,$FE,$DC,$F9,$2F,$DE,$B6,$33,$3F,$E3,$93,$FD,$89,$3A,$7F,$C1
	.dc.b	$BF,$1B,$FE,$AD,$7F,$C1,$AD,$7F,$C7,$6D,$7F,$C7,$6D,$7F,$C7,$6F
	.dc.b	$FA,$D7,$2F,$CE,$37,$4F,$F2,$1F,$C2,$7F,$3B,$AF,$C7,$4E,$5E,$7E
	.dc.b	$1F,$F4,$28,$C4,$7F,$91,$A7,$85,$F4,$AD,$12,$7F,$31,$3A,$BF,$3C
	.dc.b	$E6,$B8,$56,$B7,$C5,$6A,$3C,$75,$FD,$4A,$F9,$47,$90,$C9,$2F,$F7
	.dc.b	$11,$A8,$F8,$33,$D1,$E1,$3A,$BC,$46,$8E,$57,$F1,$23,$3B,$B1,$F8
	.dc.b	$2D,$C3,$C9,$3A,$C7,$C1,$39,$C3,$D8,$74,$3F,$CD,$F1,$F3,$E7,$FC
	.dc.b	$8E,$D8,$F0,$1F,$4C,$DC,$7F,$03,$9F,$F6,$EB,$E0,$8F,$E0,$9F,$0B
	.dc.b	$E0,$AF,$0B,$67,$FB,$CB,$67,$71,$83,$FB,$FF,$87,$FD,$ED,$C4,$7F
	.dc.b	$68,$FA,$47,$47,$D0,$F1,$11,$FE,$56,$AF,$F1,$0B,$A6,$EB,$B8,$C5
	.dc.b	$F3,$8C,$5E,$5B,$5C,$A3,$87,$56,$2B,$FB,$CF,$D1,$BE,$6B,$A7,$37
	.dc.b	$8E,$31,$87,$8A,$31,$3A,$8F,$F1,$AC,$EA,$BD,$95,$A9,$CD,$73,$FA
	.dc.b	$2C,$57,$83,$E8,$3E,$A3,$FE,$FB,$C1,$18,$D5,$18,$C3,$FC,$41,$18
	.dc.b	$9D,$4B,$E4,$7D,$11,$E9,$6D,$79,$BE,$07,$64,$79,$D2,$39,$FA,$FF
	.dc.b	$04,$F8,$5F,$3D,$6B,$DF,$EB,$FD,$7F,$0D,$F0,$3F,$1E,$30,$FC,$9F
	.dc.b	$46,$57,$C6,$F8,$FB,$BF,$D0,$F9,$EF,$9C,$7C,$FF,$4E,$D7,$CC,$F1
	.dc.b	$3B,$7C,$DF,$D3,$6C,$7D,$0F,$03,$EB,$1F,$5F,$C5,$B1,$F5,$3B,$BE
	.dc.b	$D1,$F6,$FB,$EC,$7F,$9B,$3E,$7D,$F4,$7E,$BE,$FF,$C9,$46,$27,$FC
	.dc.b	$6B,$18,$F8,$89,$1C,$7E,$46,$F0,$7F,$75,$F6,$DF,$23,$FC,$15,$5D
	.dc.b	$8C,$39,$F9,$1C,$61,$FE,$64,$A4,$62,$FA,$3F,$B7,$7F,$DE,$51,$A9
	.dc.b	$FE,$79,$01,$19,$72,$D7,$FD,$F5,$DE,$F0,$FC,$F5,$C7,$E2,$7F,$9B
	.dc.b	$4F,$F9,$6E,$7F,$9C,$46,$1D,$D2,$37,$D7,$CE,$F9,$1E,$3B,$6D,$C2
	.dc.b	$B9,$69,$D2,$B5,$39,$78,$63,$71,$61,$E7,$CF,$84,$63,$2F,$E0,$39
	.dc.b	$79,$67,$35,$FE,$73,$F1,$6E,$1F,$0E,$B5,$8F,$9B,$1E,$B9,$3E,$43
	.dc.b	$79,$70,$F9,$19,$7B,$F1,$DF,$86,$FF,$B0,$F2,$F8,$95,$E4,$9C,$E5
	.dc.b	$F5,$C3,$8B,$FA,$ED,$8F,$EE,$8D,$7F,$DA,$1F,$48,$F5,$FA,$63,$DE
	.dc.b	$FE,$F5,$AB,$51,$E5,$9D,$5F,$E0,$FC,$C7,$A7,$F2,$21,$1F,$EA,$FA
	.dc.b	$F8,$51,$89,$D4,$E6,$3E,$F4,$FE,$35,$E3,$DB,$ED,$6C,$7E,$49,$A7
	.dc.b	$E5,$27,$FC,$BB,$1F,$96,$E2,$FC,$86,$FF,$5B,$8F,$E2,$CF,$D3,$9C
	.dc.b	$B6,$33,$7E,$23,$F0,$7C,$4F,$93,$FB,$A7,$E0,$38,$DF,$1D,$8C,$3D
	.dc.b	$B9,$FF,$B5,$1B,$E0,$FC,$3F,$8E,$7B,$9F,$02,$3E,$37,$57,$E0,$EB
	.dc.b	$D8,$DD,$3F,$57,$BF,$C1,$8C,$47,$F8,$B5,$F2,$4F,$0D,$34,$47,$E1
	.dc.b	$6A,$F9,$3F,$CB,$FD,$03,$D1,$FC,$BC,$61,$B5,$F1,$6B,$83,$F1,$F8
	.dc.b	$1F,$1B,$AC,$FE,$31,$F7,$BC,$6F,$B7,$E4,$7D,$C6,$DF,$E0,$3E,$8F
	.dc.b	$E3,$BF,$65,$B5,$FC,$31,$AF,$F9,$25,$AF,$EF,$74,$FE,$C1,$DD,$FC
	.dc.b	$21,$B5,$FB,$AE,$5F,$6F,$A9,$F3,$FE,$03,$F1,$1C,$C6,$92,$BF,$15
	.dc.b	$1F,$9C,$C7,$F8,$EB,$5F,$F2,$D3,$5F,$CA,$35,$FC,$2F,$C7,$7C,$CF
	.dc.b	$5B,$F8,$0B,$6B,$E5,$7C,$7F,$ED,$7B,$3F,$91,$37,$8C,$D3,$1C,$5F
	.dc.b	$5B,$D7,$FE,$16,$DB,$FA,$9F,$23,$C4,$8F,$DE,$7B,$70,$3E,$7A,$92
	.dc.b	$7F,$49,$19,$7F,$0B,$F1,$DE,$32,$38,$3F,$13,$F1,$FF,$C7,$0F,$E3
	.dc.b	$FF,$1A,$F8,$BC,$4F,$EA,$E3,$0F,$E2,$67,$73,$AE,$1F,$D6,$6E,$C7
	.dc.b	$FC,$05,$7E,$4F,$73,$D2,$47,$F3,$3F,$1F,$F1,$C8,$E0,$F5,$FB,$9F
	.dc.b	$00,$8E,$0F,$D0,$FC,$7F,$CF,$23,$83,$F4,$BE,$17,$F3,$83,$DB,$E6
	.dc.b	$D8,$F6,$37,$3F,$4E,$DF,$87,$EC,$FE,$AD,$7F,$AC,$F7,$BE,$89,$30
	.dc.b	$F9,$EF,$DF,$A9,$1F,$B4,$6D,$FF,$6C,$DF,$FC,$E9,$3F,$C9,$9B,$1F
	.dc.b	$AE,$FA,$9F,$E3,$AF,$EC,$23,$38,$FF,$9B,$7C,$A7,$AB,$FC,$9D,$B8
	.dc.b	$7F,$27,$6D,$FF,$7A,$6B,$FE,$B7,$7F,$C1,$31,$BF,$57,$91,$1F,$F0
	.dc.b	$0F,$D0,$3E,$73,$FA,$0B,$6F,$FA,$07,$D1,$FF,$61,$AF,$F0,$34,$E5
	.dc.b	$FE,$83,$CA,$F8,$FF,$E7,$8D,$8F,$E7,$8D,$BF,$F0,$7B,$5F,$F6,$63
	.dc.b	$FB,$B8,$E7,$E7,$FF,$4F,$8C,$4F,$F1,$47,$F9,$7C,$FF,$50,$FA,$4F
	.dc.b	$F4,$24,$E7,$C5,$FE,$AD,$18,$DF,$FD,$2D,$7C,$9F,$E1,$DF,$2D,$FC
	.dc.b	$A0,$FE,$FB,$F1,$37,$1F,$EB,$7F,$11,$FE,$23,$7F,$AA,$37,$2F,$F1
	.dc.b	$3B,$5F,$F7,$23,$5F,$ED,$BF,$87,$7F,$90,$5F,$E5,$7F,$12,$F9,$6F
	.dc.b	$EF,$D3,$9C,$BF,$D5,$53,$9C,$3F,$B5,$F0,$3D,$DF,$EC,$CD,$8F,$EC
	.dc.b	$CD,$C7,$EC,$CE,$BC,$A9,$5A,$6C,$FF,$6D,$1F,$19,$FE,$02,$F8,$0F
	.dc.b	$F2,$0B,$63,$FD,$FC,$D7,$FB,$C3,$7E,$35,$F2,$3F,$E3,$99,$FF,$0A
	.dc.b	$D7,$F7,$D9,$CE,$9B,$AD,$7C,$0F,$FA,$DA,$FF,$C9,$D5,$A7,$89,$FE
	.dc.b	$51,$FA,$4F,$F4,$CB,$FE,$03,$6B,$FE,$56,$6B,$FE,$0B,$0E,$1F,$F0
	.dc.b	$72,$F0,$7E,$03,$63,$A4,$60,$FF,$08,$53,$E4,$9F,$23,$FE,$17,$CB
	.dc.b	$FD,$BD,$F8,$17,$F8,$AF,$0F,$F0,$FB,$5F,$F0,$65,$F1,$AE,$1B,$FF
	.dc.b	$87,$EB,$E3,$D7,$F8,$75,$AF,$F9,$C1,$AF,$AD,$AF,$F6,$5F,$91,$E6
	.dc.b	$FF,$96,$BA,$51,$8F,$33,$F9,$83,$7D,$4F,$89,$9F,$FB,$D7,$E7,$3F
	.dc.b	$C6,$F8,$7F,$8D,$BD,$BA,$FF,$7E,$6C,$EF,$E0,$FB,$CA,$73,$A7,$F8
	.dc.b	$E5,$FE,$8D,$1C,$FC,$8D,$C6,$A3,$FD,$14,$BE,$67,$F2,$0D,$8F,$F1
	.dc.b	$F3,$5F,$F4,$8B,$5F,$F2,$37,$53,$EA,$DE,$A7,$C2,$8C,$36,$33,$3F
	.dc.b	$E5,$2F,$A5,$7C,$9F,$AB,$E8,$EC,$FF,$93,$5A,$FF,$A8,$1A,$FE,$2D
	.dc.b	$AF,$F9,$5D,$7C,$AF,$EA,$5B,$1C,$E3,$07,$EB,$7E,$B5,$F2,$DF,$29
	.dc.b	$B3,$FE,$66,$6B,$FE,$2E,$FA,$CF,$96,$FF,$6A,$F1,$DF,$4D,$FE,$63
	.dc.b	$E8,$46,$27,$FC,$C4,$D7,$F4,$51,$A9,$E1,$D3,$EA,$9F,$C4,$69,$82
	.dc.b	$3F,$E6,$BF,$A4,$FF,$7C,$3F,$E6,$9C,$3F,$8B,$37,$C9,$FC,$74,$EA
	.dc.b	$3C,$D1,$F9,$19,$DC,$60,$FF,$39,$4E,$FA,$3F,$D8,$73,$F9,$09,$FF
	.dc.b	$64,$B5,$FF,$5F,$35,$FF,$5B,$B7,$2F,$F0,$66,$DE,$C2,$4F,$F8,$81
	.dc.b	$F8,$D3,$F2,$A3,$AF,$36,$08,$FF,$89,$BE,$3B,$E2,$91,$C1,$F0,$DF
	.dc.b	$67,$3C,$8F,$5C,$57,$EB,$45,$C7,$FD,$BF,$92,$BE,$27,$CF,$F9,$09
	.dc.b	$7F,$61,$FA,$BF,$15,$23,$FC,$72,$D7,$ED,$E5,$FD,$AD,$FB,$FE,$97
	.dc.b	$B8,$8F,$E0,$BE,$83,$F8,$17,$0F,$FA,$DA,$3F,$D4,$4D,$7F,$19,$1F
	.dc.b	$5F,$D2,$F8,$E4,$8F,$C5,$BF,$2B,$D3,$E7,$F2,$32,$7F,$09,$1F,$C1
	.dc.b	$3F,$0D,$FE,$95,$06,$91,$FE,$9D,$F1,$5F,$95,$7F,$CD,$FC,$F8,$23
	.dc.b	$FE,$58,$F9,$CF,$49,$1F,$C9,$7C,$07,$F1,$2F,$A5,$FF,$4A,$C7,$FB
	.dc.b	$15,$AF,$F4,$10,$E1,$FE,$88,$D7,$FD,$A0,$BE,$23,$A7,$85,$27,$FD
	.dc.b	$B3,$39,$7F,$45,$E5,$FE,$74,$8F,$FB,$37,$E9,$7F,$BA,$3A,$3E,$7F
	.dc.b	$8D,$F4,$8D,$BF,$EA,$AF,$A0,$FD,$43,$FD,$5D,$AF,$FB,$8D,$A8,$38
	.dc.b	$6F,$C3,$7F,$79,$83,$49,$1F,$EE,$9F,$2F,$FA,$A7,$67,$F4,$1D,$7F
	.dc.b	$9D,$C6,$3C,$AF,$F4,$FF,$85,$FB,$9F,$42,$61,$FC,$05,$FE,$99,$A7
	.dc.b	$FB,$36,$3F,$82,$7A,$9F,$8E,$F9,$E9,$18,$3F,$E0,$0E,$3F,$C7,$1F
	.dc.b	$EE,$8D,$7F,$E0,$03,$5F,$73,$52,$F3,$FC,$FE,$B4,$9E,$B1,$FD,$8B
	.dc.b	$D3,$7B,$1F,$0A,$7E,$46,$ED,$BE,$55,$3E,$69,$DF,$C5,$B7,$C2,$B7
	.dc.b	$87,$38,$FE,$F7,$3A,$7F,$C1,$D7,$CE,$FF,$C0,$F3,$99,$FB,$02,$74
	.dc.b	$FF,$36,$FA,$2F,$E4,$C4,$7C,$06,$E7,$B6,$3F,$C9,$BE,$87,$F9,$61
	.dc.b	$27,$F2,$1F,$23,$5F,$F5,$7F,$86,$FA,$1F,$E0,$26,$6F,$E1,$4B,$CE
	.dc.b	$72,$7E,$76,$3F,$A7,$7A,$5F,$CE,$4F,$BF,$DF,$63,$F8,$4F,$0F,$F0
	.dc.b	$C3,$0F,$E6,$FE,$0B,$F2,$44,$7F,$64,$7E,$61,$C3,$FE,$4E,$FA,$0F
	.dc.b	$CF,$7B,$5F,$E4,$A6,$C7,$E7,$FD,$F4,$7E,$8E,$B5,$3F,$A3,$EF,$FF
	.dc.b	$0F,$BB,$9F,$D4,$7A,$9A,$8F,$D3,$B6,$77,$1E,$13,$E1,$68,$F8,$EE
	.dc.b	$BB,$7C,$46,$C6,$67,$B1,$C3,$E4,$33,$F1,$DC,$3A,$DF,$43,$4C,$D7
	.dc.b	$0C,$7E,$7F,$C3,$F4,$24,$67,$17,$C7,$F8,$BE,$8F,$FB,$D7,$0F,$C2
	.dc.b	$AF,$B1,$18,$E0,$FE,$09,$8A,$31,$3E,$C3,$97,$C1,$79,$7E,$3E,$47
	.dc.b	$8E,$7F,$C7,$F8,$75,$BE,$5E,$DF,$84,$95,$BF,$73,$C9,$E1,$48,$E4
	.dc.b	$F9,$FD,$AF,$47,$AB,$F0,$FE,$3A,$57,$78,$F6,$79,$B2,$3C,$4F,$BF
	.dc.b	$D8,$F8,$1E,$67,$EA,$3B,$BF,$53,$AF,$B6,$31,$7F,$03,$7F,$F3,$3D
	.dc.b	$69,$B1,$9F,$37,$E5,$79,$12,$3E,$04,$57,$F4,$67,$CC,$8E,$AD,$3F
	.dc.b	$D7,$58,$7C,$D8,$CB,$F3,$F9,$3E,$B1,$27,$E6,$78,$F5,$FC,$74,$FD
	.dc.b	$5F,$0F,$EE,$19,$E4,$FA,$19,$7D,$DA,$DB,$FE,$42,$8C,$CE,$F0,$F8
	.dc.b	$1E,$4F,$E1,$49,$7F,$6F,$B2,$BC,$BC,$BF,$C5,$18,$C3,$BD,$62,$7F
	.dc.b	$05,$39,$BF,$EA,$1F,$3F,$FA,$72,$4F,$E8,$07,$57,$F9,$9B,$0F,$47
	.dc.b	$85,$FE,$0B,$6F,$B3,$FA,$F3,$8D,$7E,$BC,$FE,$B3,$79,$7A,$52,$77
	.dc.b	$5C,$A7,$FD,$D1,$39,$6C,$FD,$67,$FB,$1B,$63,$51,$DD,$7A,$DF,$CC
	.dc.b	$63,$0D,$AE,$D1,$8C,$FC,$73,$9C,$BF,$B9,$36,$7F,$88,$61,$FC,$2E
	.dc.b	$31,$1F,$DD,$1B,$87,$82,$FD,$B5,$B7,$A9,$2B,$CD,$19,$7C,$44,$C3
	.dc.b	$CA,$7C,$EB,$EA,$4A,$EB,$3C,$E7,$51,$70,$FF,$05,$5F,$59,$CD,$75
	.dc.b	$9C,$BB,$3F,$96,$5F,$57,$CA,$64,$F7,$5D,$C3,$E7,$DF,$49,$FF,$0D
	.dc.b	$57,$CF,$6D,$7F,$42,$AF,$89,$1F,$E3,$5F,$89,$FF,$75,$FD,$1F,$F4
	.dc.b	$66,$C6,$7D,$64,$E5,$B1,$FE,$3D,$AF,$F6,$97,$87,$D7,$D6,$34,$93
	.dc.b	$B8,$CB,$F8,$56,$E7,$F9,$AF,$53,$FB,$2B,$F9,$C8,$CE,$BF,$E3,$36
	.dc.b	$D7,$F5,$16,$BF,$E3,$4A,$E3,$3B,$BF,$CE,$D6,$B0,$F3,$DF,$E5,$6F
	.dc.b	$6F,$3A,$D9,$1D,$74,$DC,$7C,$93,$71,$FF,$23,$5F,$2B,$FF,$22,$E3
	.dc.b	$DC,$95,$FA,$E9,$D6,$3E,$2C,$39,$DF,$C3,$AD,$63,$D6,$3C,$86,$4F
	.dc.b	$19,$67,$89,$31,$F1,$CF,$19,$F8,$8F,$EE,$B0,$FF,$08,$BF,$E5,$F6
	.dc.b	$C7,$03,$F9,$99,$C0,$EF,$5F,$C3,$BC,$FF,$4C,$FA,$9B,$5F,$C2,$9B
	.dc.b	$5F,$C0,$5E,$AF,$F8,$1D,$AF,$F9,$A9,$B8,$FF,$89,$E7,$F8,$7C,$75
	.dc.b	$7F,$90,$71,$7F,$92,$5E,$C7,$F7,$BF,$75,$C9,$F6,$BA,$3F,$DE,$7A
	.dc.b	$59,$CD,$6E,$31,$5E,$4A,$FF,$7C,$B6,$B7,$97,$F5,$49,$CD,$FF,$52
	.dc.b	$9C,$E9,$FD,$DD,$B8,$7F,$B9,$B9,$BA,$D7,$F8,$3F,$E3,$1F,$33,$E3
	.dc.b	$E5,$F3,$46,$27,$D5,$18,$9F,$24,$E7,$3F,$13,$6B,$FB,$9B,$63,$FC
	.dc.b	$22,$EF,$2F,$F0,$53,$71,$CF,$CE,$F5,$BF,$27,$CD,$78,$8F,$1F,$CD
	.dc.b	$58,$F4,$3E,$22,$3F,$DC,$AD,$1E,$6F,$98,$F5,$9E,$CF,$99,$B5,$F0
	.dc.b	$7E,$43,$DE,$7C,$0F,$7E,$C7,$C5,$F5,$FD,$E4,$7B,$3F,$0B,$E2,$7E
	.dc.b	$39,$F2,$3D,$7B,$3F,$37,$E1,$FE,$59,$F3,$3E,$22,$C7,$D1,$F0,$FE
	.dc.b	$59,$1F,$53,$F3,$BB,$FE,$99,$F5,$3C,$3B,$3F,$67,$B7,$EB,$9F,$63
	.dc.b	$BE,$C7,$F9,$2F,$E7,$7F,$C9,$EF,$F8,$59,$BD,$1F,$89,$6C,$7F,$8A
	.dc.b	$9B,$1F,$E4,$85,$E7,$FE,$65,$B3,$D7,$D4,$C4,$FF,$0D,$F6,$1F,$47
	.dc.b	$4C,$7D,$1D,$BF,$F0,$34,$A3,$F7,$78,$FF,$B6,$63,$13,$FE,$84,$6C
	.dc.b	$7F,$BD,$DA,$FF,$AA,$02,$F9,$5F,$F4,$CF,$C0,$7F,$8E,$23,$8B,$63
	.dc.b	$35,$FE,$B0,$9C,$BE,$FC,$33,$7F,$EA,$1F,$0B,$D7,$3F,$EB,$0F,$11
	.dc.b	$AE,$29,$97,$FA,$A2,$7F,$DA,$58,$72,$76,$DA,$E1,$3B,$7D,$17,$D5
	.dc.b	$CB,$FE,$78,$C3,$CB,$87,$8F,$3E,$5F,$45,$D6,$73,$97,$92,$31,$3A
	.dc.b	$BF,$45,$72,$F1,$1F,$46,$C7,$96,$DF,$02,$73,$18,$BF,$F6,$7C,$7F
	.dc.b	$B8,$FD,$6D,$5F,$19,$CF,$C2,$A7,$37,$CE,$1E,$B5,$9C,$36,$BF,$DB
	.dc.b	$75,$FE,$48,$E8,$FA,$2F,$F9,$2B,$D1,$BE,$67,$F9,$06,$B6,$A3,$7F
	.dc.b	$43,$F9,$4A,$FA,$75,$FE,$EC,$AE,$11,$83,$E0,$4E,$F8,$FF,$CF,$B3
	.dc.b	$A7,$ED,$6E,$FA,$3F,$9D,$C5,$F3,$DB,$8F,$D7,$FA,$2F,$B8,$7D,$DF
	.dc.b	$89,$B1,$EC,$F8,$F7,$BB,$EC,$6C,$7C,$0E,$A6,$39,$AF,$83,$1E,$19
	.dc.b	$FD,$16,$3F,$A6,$DB,$F3,$ED,$9D,$47,$E9,$7E,$8F,$8C,$F2,$6D,$7E
	.dc.b	$09,$FC,$41,$F9,$3F,$A3,$B1,$F9,$16,$D7,$C7,$EA,$FC,$FC,$63,$7F
	.dc.b	$F5,$1F,$0B,$7F,$AB,$E1,$7D,$26,$E3,$F7,$25,$C1,$98,$C3,$72,$ED
	.dc.b	$C3,$CE,$7F,$2F,$7F,$1F,$DF,$FC,$F3,$F2,$79,$FE,$17,$1F,$E3,$EB
	.dc.b	$F8,$16,$1F,$9B,$8C,$38,$6B,$FC,$1D,$7B,$1C,$E7,$EB,$36,$7C,$7F
	.dc.b	$19,$FA,$DC,$FF,$5C,$DA,$FD,$B7,$D1,$E1,$A7,$EB,$A3,$F6,$9F,$19
	.dc.b	$FB,$5E,$1B,$3F,$B2,$7F,$81,$7B,$78,$C7,$E7,$78,$5F,$E5,$4A,$FD
	.dc.b	$BD,$FE,$F7,$E4,$7F,$92,$F2,$B8,$91,$86,$D7,$C2,$9D,$4F,$0F,$88
	.dc.b	$7C,$87,$3A,$FE,$35,$18,$7F,$82,$CF,$F1,$78,$7C,$47,$F0,$DF,$8C
	.dc.b	$C3,$6B,$8B,$87,$9F,$C4,$5D,$7E,$DF,$C6,$F9,$49,$3E,$5D,$11,$E5
	.dc.b	$F8,$AF,$E3,$4F,$85,$B3,$FC,$63,$8F,$DC,$47,$93,$ED,$F0,$3F,$91
	.dc.b	$3E,$96,$C7,$BB,$A3,$F9,$37,$81,$ED,$6C,$64,$99,$7E,$7F,$6F,$E2
	.dc.b	$F1,$8E,$EF,$E0,$71,$F3,$C9,$1E,$B7,$E7,$7C,$57,$F3,$37,$E5,$36
	.dc.b	$3F,$8E,$FD,$1F,$F0,$D7,$F9,$85,$A3,$0F,$F3,$3F,$A3,$7A,$1F,$C8
	.dc.b	$A7,$32,$9F,$E3,$CF,$F6,$16,$BD,$1A,$F3,$F4,$3F,$84,$47,$F4,$C8
	.dc.b	$C5,$FF,$30,$8C,$47,$59,$FE,$63,$F4,$5F,$D4,$20,$DF,$92,$7C,$BE
	.dc.b	$BD,$2F,$E6,$F1,$FC,$72,$BF,$66,$D7,$FB,$B3,$5F,$82,$D7,$FA,$92
	.dc.b	$F1,$79,$9F,$D0,$BE,$8B,$FC,$D9,$3F,$CE,$44,$EF,$3F,$E9,$33,$AE
	.dc.b	$04,$7C,$E8,$C3,$FD,$7F,$B5,$C2,$F8,$DE,$D3,$FD,$AF,$C1,$8F,$F6
	.dc.b	$0F,$C3,$FF,$8C,$BF,$DA,$A3,$FA,$AF,$C2,$3E,$63,$F9,$A3,$67,$F8
	.dc.b	$9E,$A8,$C3,$FC,$59,$AF,$E0,$BD,$67,$D1,$FF,$8D,$36,$F4,$CA,$73
	.dc.b	$F2,$DF,$B5,$F0,$3F,$BB,$5E,$F1,$FC,$5F,$D0,$FF,$4A,$7C,$2D,$C3
	.dc.b	$F9,$2B,$5F,$F1,$43,$5E,$3E,$57,$F4,$D8,$34,$93,$EC,$9F,$F0,$47
	.dc.b	$C6,$7F,$82,$72,$FF,$4E,$C7,$F8,$CE,$31,$F1,$FF,$CD,$1B,$19,$3D
	.dc.b	$91,$FA,$2F,$92,$FF,$02,$EB,$A9,$46,$1D,$9F,$E1,$77,$17,$E0,$8D
	.dc.b	$FB,$5F,$E3,$27,$63,$E8,$FF,$D3,$AF,$D4,$6B,$FD,$21,$AF,$F9,$2D
	.dc.b	$AF,$ED,$97,$C4,$7C,$16,$C7,$38,$C1,$F8,$0A,$78,$5F,$D3,$A7,$FC
	.dc.b	$52,$D7,$FC,$25,$1A,$F4,$BF,$81,$E5,$FE,$49,$E8,$FC,$05,$7F,$89
	.dc.b	$3E,$8F,$83,$FE,$5F,$6B,$FC,$29,$AF,$F8,$ED,$BD,$5F,$D1,$E3,$AF
	.dc.b	$07,$F9,$2D,$A7,$F9,$74,$B8,$BF,$0E,$FE,$8F,$1F,$E3,$4F,$9C,$7B
	.dc.b	$BC,$21,$18,$75,$1D,$30,$7C,$FF,$CE,$D3,$F3,$74,$F6,$F8,$3F,$BB
	.dc.b	$6C,$69,$B9,$FF,$26,$C6,$D7,$F9,$5E,$7F,$AA,$BF,$AE,$6B,$FE,$02
	.dc.b	$6B,$FE,$85,$6B,$FC,$C3,$E8,$DF,$17,$EB,$6C,$7F,$94,$9A,$F9,$7E
	.dc.b	$73,$E0,$7D,$1B,$E3,$ED,$CA,$FF,$2C,$CE,$5B,$1F,$A7,$8E,$71,$FE
	.dc.b	$97,$6B,$FD,$09,$AF,$F9,$CF,$E8,$BD,$7A,$7C,$86,$E9,$FD,$21,$B3
	.dc.b	$FE,$79,$A7,$C8,$7D,$1D,$BF,$DC,$33,$FE,$97,$F9,$9F,$E2,$CD,$F1
	.dc.b	$35,$F2,$BF,$5E,$73,$3F,$EA,$BE,$24,$62,$3F,$AB,$C6,$2B,$F8,$EB
	.dc.b	$63,$FC,$21,$6F,$95,$F1,$12,$3F,$D5,$11,$FA,$87,$FC,$98,$ED,$F3
	.dc.b	$7B,$DE,$33,$83,$AF,$A0,$F3,$73,$48,$E7,$15,$F4,$65,$F8,$C7,$CE
	.dc.b	$D7,$FA,$03,$56,$3F,$96,$E1,$F8,$5E,$EC,$9F,$3C,$57,$AC,$7C,$0A
	.dc.b	$FC,$5C,$72,$6D,$7E,$83,$E4,$FE,$91,$27,$E0,$FC,$A5,$7F,$CD,$94
	.dc.b	$E6,$FA,$46,$79,$7F,$D9,$5B,$3E,$37,$EB,$7C,$04,$C3,$E5,$C5,$7E
	.dc.b	$0C,$79,$48,$FD,$4F,$92,$FB,$DE,$B4,$8F,$A7,$15,$F1,$FC,$07,$E0
	.dc.b	$CC,$63,$FE,$DB,$FA,$07,$A1,$F8,$7F,$5E,$5F,$FB,$8E,$72,$FE,$3E
	.dc.b	$3F,$19,$F0,$12,$30,$7F,$AF,$BE,$89,$EA,$FF,$0B,$7F,$DF,$58,$FF
	.dc.b	$06,$6F,$81,$FE,$26,$77,$5C,$4F,$EA,$4F,$F5,$34,$D3,$FA,$B7,$2F
	.dc.b	$E6,$BE,$03,$F3,$C4,$9F,$F5,$A7,$6F,$F1,$BF,$A2,$FE,$BF,$1F,$ED
	.dc.b	$25,$F1,$39,$35,$EA,$59,$D7,$C4,$71,$30,$FF,$A9,$F8,$1F,$29,$FA
	.dc.b	$DF,$3E,$4F,$C7,$8A,$FE,$0A,$2E,$FF,$3E,$7F,$97,$7C,$F3,$E6,$AF
	.dc.b	$8B,$30,$FF,$73,$CE,$5B,$1F,$DD,$63,$F5,$91,$FE,$A4,$6B,$FE,$50
	.dc.b	$0E,$1F,$F2,$9F,$C7,$7F,$40,$7F,$C5,$CD,$C3,$F9,$CB,$78,$FF,$CB
	.dc.b	$35,$FB,$B9,$FC,$37,$C8,$FF,$41,$D0,$D3,$1F,$25,$FC,$3F,$D5,$FE
	.dc.b	$21,$EC,$4C,$FF,$8B,$47,$F4,$DF,$49,$F3,$B7,$E0,$A5,$7F,$53,$8A
	.dc.b	$FF,$90,$E7,$FA,$F3,$7E,$03,$3A,$7F,$8A,$BD,$7F,$F6,$8F,$0F,$F5
	.dc.b	$EF,$91,$F4,$70,$FF,$86,$5B,$87,$F8,$BD,$F9,$37,$C5,$CD,$BE,$59
	.dc.b	$E1,$FF,$68,$6D,$8F,$EF,$B5,$C2,$33,$1F,$C0,$3E,$8B,$F9,$81,$C6
	.dc.b	$BE,$47,$57,$F3,$3F,$A2,$95,$FA,$D8,$AF,$F0,$88,$B9,$7E,$D6,$7E
	.dc.b	$87,$5C,$AF,$DA,$3F,$4F,$E4,$F1,$1F,$C4,$38,$BF,$A0,$18,$BF,$EE
	.dc.b	$BD,$0F,$9D,$FC,$8F,$A7,$2F,$F1,$F1,$5F,$C6,$47,$E3,$C8,$FF,$98
	.dc.b	$27,$FB,$12,$F8,$4F,$B6,$FC,$58,$CC,$62,$3F,$27,$ED,$FF,$A9,$23
	.dc.b	$12,$AC,$73,$FF,$9E,$7B,$3F,$D9,$38,$7F,$94,$A3,$1F,$27,$A8,$FE
	.dc.b	$7B,$3A,$7C,$53,$C7,$E5,$35,$E3,$A9,$F2,$7F,$EB,$4E,$9F,$EB,$F1
	.dc.b	$8C,$3E,$E4,$62,$75,$1E,$36,$BE,$0F,$2B,$AD,$7F,$18,$8F,$09,$93
	.dc.b	$D0,$79,$BE,$0F,$C7,$F4,$5F,$DC,$B7,$F8,$F9,$23,$1A,$7F,$A6,$5B
	.dc.b	$7E,$E3,$8F,$B7,$E2,$AE,$0D,$9F,$F1,$7E,$9F,$DE,$9B,$5C,$FD,$6F
	.dc.b	$09,$1F,$2D,$F8,$7D,$AF,$2F,$5C,$9C,$DF,$E7,$33,$F3,$AF,$57,$D7
	.dc.b	$E8,$F8,$49,$5F,$28,$F5,$3E,$5F,$43,$E0,$92,$3A,$FA,$5F,$13,$C8
	.dc.b	$91,$E0,$7E,$3F,$9D,$F2,$3C,$9F,$71,$27,$B1,$FD,$FB,$2F,$F3,$9B
	.dc.b	$8D,$BE,$67,$87,$F4,$09,$1E,$CA,$FA,$7C,$72,$3D,$EF,$D6,$F0,$F1
	.dc.b	$1F,$E5,$CB,$FA,$D3,$F3,$C9,$1D,$F0,$FC,$F6,$FF,$1B,$C0,$FB,$E7
	.dc.b	$27,$E4,$F8,$DF,$82,$25,$7C,$EF,$22,$BD,$EF,$FD,$0A,$D9,$FB,$C4
	.dc.b	$7E,$AF,$C4,$7E,$30,$E3,$7F,$AC,$F9,$AF,$D3,$4F,$F4,$26,$F8,$27
	.dc.b	$3D,$5F,$E6,$2A,$E1,$5F,$30,$E0,$5A,$E0,$DA,$FD,$1B,$A3,$F4,$27
	.dc.b	$F5,$0F,$63,$B9,$E0,$8B,$1F,$A4,$34,$59,$FF,$11,$36,$36,$FF,$A3
	.dc.b	$0D,$96,$36,$D9,$F1,$C3,$E7,$BF,$D5,$B7,$FC,$39,$BB,$7F,$5C,$6D
	.dc.b	$FF,$17,$BF,$8B,$18,$8F,$ED,$2D,$BE,$99,$F9,$EB,$6F,$A6,$74,$6E
	.dc.b	$33,$5F,$12,$B8,$B6,$74,$F1,$A8,$E5,$EC,$E3,$0F,$54,$5C,$3F,$BA
	.dc.b	$E1,$DA,$73,$5C,$4F,$04,$66,$7E,$05,$71,$9F,$80,$53,$DE,$DC,$3E
	.dc.b	$6D,$F4,$9F,$F0,$64,$7A,$9B,$87,$E5,$9E,$A7,$F8,$44,$FF,$7E,$F1
	.dc.b	$BE,$8B,$11,$86,$FC,$5A,$75,$3C,$63,$13,$FD,$3F,$E1,$FF,$97,$4F
	.dc.b	$F8,$51,$AF,$F4,$F6,$BF,$E2,$06,$E1,$FE,$D3,$F0,$BF,$1D,$1F,$E4
	.dc.b	$EC,$BF,$A8,$FB,$6F,$07,$35,$30,$FB,$17,$BD,$BF,$C4,$77,$B9,$CD
	.dc.b	$6F,$0F,$EB,$77,$B9,$FE,$C5,$5B,$9C,$D6,$EF,$D4,$FE,$86,$75,$18
	.dc.b	$AE,$A4,$AF,$3E,$1C,$63,$E0,$5D,$7F,$C7,$F7,$C9,$C9,$C6,$FA,$5F
	.dc.b	$28,$C1,$A7,$CA,$FE,$5E,$79,$3F,$2E,$BB,$63,$F2,$E3,$58,$62,$B7
	.dc.b	$26,$FC,$FE,$19,$3E,$F3,$F8,$04,$62,$BF,$4C,$FC,$D8,$FC,$7B,$63
	.dc.b	$FC,$A6,$D7,$FC,$1C,$D7,$FC,$BE,$F8,$23,$F8,$77,$17,$DD,$7B,$92
	.dc.b	$7B,$BA,$38,$3F,$DE,$B9,$FF,$DC,$BC,$9F,$2D,$FF,$73,$E9,$7D,$F4
	.dc.b	$D8,$C3,$2F,$FB,$6C,$B9,$F1,$87,$58,$E3,$3A,$AE,$95,$AD,$BF,$C0
	.dc.b	$BF,$1E,$FA,$2F,$07,$37,$8E,$31,$87,$68,$C4,$C8,$FE,$FB,$2F,$A2
	.dc.b	$C6,$7E,$D8,$E2,$DA,$E9,$3A,$C3,$C3,$F4,$5D,$CF,$07,$C4,$5C,$7A
	.dc.b	$B8,$7B,$FA,$F7,$1E,$CF,$9C,$F4,$7B,$96,$7C,$BE,$97,$AC,$F6,$7C
	.dc.b	$7D,$8F,$4F,$A7,$EF,$3E,$07,$A7,$63,$DB,$F1,$5F,$0C,$F8,$9F,$13
	.dc.b	$63,$E0,$FC,$47,$C7,$3E,$47,$C4,$D8,$F8,$BE,$27,$CB,$3E,$67,$8B
	.dc.b	$63,$E4,$F8,$3F,$3C,$FA,$1E,$2D,$8F,$9B,$DF,$F4,$CF,$A9,$E0,$58
	.dc.b	$FF,$16,$FC,$EF,$FB,$B1,$BF,$11,$F6,$5F,$C6,$46,$79,$7F,$20,$BE
	.dc.b	$67,$26,$BE,$72,$D7,$F6,$9F,$65,$F3,$3F,$D9,$70,$FF,$4A,$46,$27
	.dc.b	$F9,$7B,$63,$F9,$A3,$5F,$F6,$33,$FE,$DA,$8D,$FC,$3F,$39,$FF,$33
	.dc.b	$B6,$75,$1F,$CF,$7E,$1F,$FD,$4B,$5F,$EA,$3F,$7F,$F9,$24,$F0,$8C
	.dc.b	$1F,$ED,$0F,$67,$F9,$44,$EA,$F9,$BF,$CA,$FC,$96,$B8,$BF,$F5,$74
	.dc.b	$63,$E8,$B8,$C6,$27,$58,$AB,$53,$E0,$8C,$4E,$A7,$AE,$7C,$A7,$3A
	.dc.b	$59,$CD,$F1,$7C,$7F,$43,$9C,$0D,$F8,$9C,$36,$B8,$4F,$FB,$91,$BF
	.dc.b	$13,$9C,$5A,$7F,$B5,$FE,$1D,$EE,$78,$FC,$BF,$F6,$83,$61,$ED,$D8
	.dc.b	$C3,$F7,$E7,$FD,$AC,$D9,$DE,$3F,$DC,$67,$2D,$8F,$ED,$CF,$C5,$8C
	.dc.b	$E1,$E4,$8F,$F6,$E4,$7F,$B8,$7D,$6E,$F1,$F2,$E5,$F4,$3F,$95,$8C
	.dc.b	$36,$39,$1F,$EE,$55,$C5,$F3,$E3,$8F,$3D,$9C,$D7,$2A,$C7,$16,$5F
	.dc.b	$A1,$F4,$32,$7E,$6F,$C5,$5C,$7B,$F2,$7E,$23,$E2,$BE,$B4,$FD,$76
	.dc.b	$D7,$A7,$9D,$C7,$F9,$B2,$47,$E9,$9F,$CF,$73,$48,$FD,$4B,$71,$FF
	.dc.b	$0E,$6B,$F9,$07,$FD,$97,$1E,$BE,$7F,$76,$1F,$81,$9F,$2F,$33,$E3
	.dc.b	$3F,$14,$7E,$2F,$E2,$AE,$5F,$88,$70,$FE,$2B,$E2,$ED,$FC,$C3,$E8
	.dc.b	$ED,$F2,$4F,$95,$B5,$AF,$A6,$E1,$FC,$9F,$C5,$5A,$E5,$DB,$F9,$09
	.dc.b	$FF,$19,$B6,$3F,$79,$D5,$FC,$06,$31,$02,$07,$F1,$6F,$F0,$3E,$7E
	.dc.b	$04,$AF,$F2,$17,$1F,$EF,$3E,$2F,$F0,$57,$C3,$B3,$F8,$07,$02,$73
	.dc.b	$87,$EB,$FE,$86,$63,$FA,$CF,$8B,$B9,$FC,$C7,$F6,$87,$ED,$78,$EC
	.dc.b	$FE,$CB,$DF,$FD,$C1,$FB,$D9,$7C,$7E,$27,$F7,$D1,$86,$C7,$EF,$EF
	.dc.b	$83,$C7,$E4,$FF,$97,$6E,$B4,$61,$B7,$F0,$1B,$7F,$20,$FE,$2C,$BE
	.dc.b	$03,$A4,$F8,$9B,$3F,$9C,$9F,$F7,$74,$F9,$BE,$33,$D5,$F1,$9E,$B6
	.dc.b	$C7,$AA,$3D,$DF,$1B,$8B,$E5,$E3,$F9,$49,$5E,$57,$D1,$F2,$5F,$15
	.dc.b	$F8,$CD,$8F,$8B,$F1,$DE,$D2,$3E,$D7,$E0,$7C,$77,$CD,$7E,$73,$63
	.dc.b	$E6,$F9,$3F,$47,$A7,$E9,$36,$32,$4C,$7E,$07,$2E,$A3,$EA,$FC,$6F
	.dc.b	$CD,$23,$F3,$5F,$A1,$F1,$DF,$75,$FB,$CD,$8F,$E3,$1F,$41,$FC,$56
	.dc.b	$FE,$93,$5F,$CD,$B5,$FC,$F3,$53,$FC,$59,$FF,$4E,$35,$E7,$E4,$7F
	.dc.b	$67,$8C,$47,$E9,$BE,$79,$EE,$7F,$21,$9E,$5D,$DF,$D9,$A7,$38,$79
	.dc.b	$B0,$FE,$4D,$F3,$EE,$0F,$EE,$4D,$9F,$F5,$4B,$72,$FF,$57,$35,$FC
	.dc.b	$72,$E0,$E2,$49,$E2,$FF,$7D,$C3,$FC,$C9,$5F,$C3,$5B,$8F,$F7,$3E
	.dc.b	$67,$C8,$7F,$39,$9C,$E1,$FD,$D2,$72,$FE,$AA,$7F,$5D,$F2,$3F,$AF
	.dc.b	$7F,$AE,$7E,$15,$FE,$F3,$9F,$EA,$0D,$C7,$F8,$E7,$C6,$FF,$0D,$9C
	.dc.b	$FB,$3F,$D1,$FE,$2F,$F5,$35,$FA,$96,$BF,$C4,$9A,$FF,$15,$6F,$7F
	.dc.b	$ED,$7E,$33,$F8,$01,$FC,$07,$E3,$AE,$1F,$D6,$FE,$7B,$FA,$1B,$FE
	.dc.b	$D4,$6E,$1F,$ED,$96,$BF,$B9,$6B,$FC,$89,$FE,$79,$1C,$7C,$FF,$F1
	.dc.b	$44,$62,$79,$47,$F8,$AD,$FE,$55,$F1,$9F,$DB,$BE,$3F,$FD,$CC,$D8
	.dc.b	$FE,$0C,$D7,$F9,$7B,$7E,$4B,$FC,$07,$C9,$7B,$BF,$C1,$74,$F6,$BF
	.dc.b	$D2,$6F,$F3,$7A,$DB,$9C,$3F,$9C,$C6,$27,$72,$F9,$EF,$F4,$33,$FE
	.dc.b	$F6,$6B,$FE,$FB,$6B,$FC,$6D,$AF,$F4,$45,$F1,$BF,$EF,$E8,$E8,$D8
	.dc.b	$CC,$75,$8F,$E4,$8F,$E2,$E7,$FA,$6E,$3F,$91,$6E,$B9,$9F,$E6,$F8
	.dc.b	$E1,$FE,$A9,$F3,$DF,$E2,$A7,$98,$70,$F6,$6B,$FC,$CD,$AF,$F8,$83
	.dc.b	$E2,$3F,$CE,$0F,$F8,$95,$AF,$F8,$72,$D1,$8C,$3F,$CB,$F5,$FE,$36
	.dc.b	$AF,$F2,$47,$CD,$7F,$AD,$DB,$F2,$7E,$7E,$44,$E6,$7F,$C3,$2F,$CE
	.dc.b	$BF,$F3,$4F,$CC,$7F,$45,$9F,$F1,$83,$5F,$F0,$13,$63,$FC,$D1,$3C
	.dc.b	$DF,$F3,$8F,$C4,$6E,$BF,$A8,$35,$FE,$F2,$D7,$FB,$E3,$52,$F3,$7E
	.dc.b	$D3,$63,$E6,$B5,$FB,$2D,$7F,$AF,$7C,$97,$F6,$E7,$FC,$91,$F3,$7F
	.dc.b	$C8,$9B,$E3,$FD,$95,$3B,$7F,$C3,$D3,$B7,$F0,$AD,$7F,$B5,$35,$FF
	.dc.b	$09,$B5,$FF,$0B,$2F,$8E,$FC,$93,$63,$F0,$AD,$7A,$C7,$E2,$34,$FA
	.dc.b	$BF,$29,$F5,$F4,$F2,$D7,$DC,$C3,$13,$97,$84,$EA,$7F,$BF,$57,$DE
	.dc.b	$9C,$9A,$F9,$2E,$84,$9F,$C3,$FC,$A5,$7F,$07,$CD,$E1,$31,$BF,$DE
	.dc.b	$F9,$4F,$27,$CA,$4A,$FC,$7C,$57,$C3,$9A,$7C,$A4,$8F,$F1,$9C,$7C
	.dc.b	$DA,$FC,$4E,$BE,$F3,$E0,$6F,$6B,$71,$EE,$23,$CB,$97,$E3,$11,$FD
	.dc.b	$0F,$CA,$57,$F3,$BF,$29,$F2,$CC,$61,$F9,$8F,$93,$F9,$DF,$29,$2B
	.dc.b	$F5,$31,$5F,$5C,$7C,$E2,$3E,$AE,$5F,$AC,$47,$F6,$3E,$37,$D9,$F9
	.dc.b	$39,$1F,$B4,$74,$FE,$C3,$BB,$EF,$11,$FD,$CF,$CA,$7E,$DF,$E6,$5F
	.dc.b	$2B,$49,$3F,$BE,$AD,$7C,$AF,$E6,$FE,$6D,$F0,$74,$CD,$7F,$AA,$63
	.dc.b	$2F,$F1,$66,$BF,$C8,$9A,$FF,$91,$DA,$FF,$9C,$1F,$E5,$89,$A4,$7F
	.dc.b	$98,$7C,$DB,$E3,$3F,$5C,$D8,$FF,$5D,$35,$FF,$54,$E5,$FC,$6F,$1F
	.dc.b	$F3,$EC,$BE,$33,$89,$2B,$8D,$FC,$5F,$94,$91,$FE,$E7,$C1,$1F,$27
	.dc.b	$DB,$F9,$5F,$21,$2B,$4F,$07,$F8,$DF,$B5,$FC,$0F,$E6,$5E,$0D,$3F
	.dc.b	$83,$79,$1F,$E1,$BA,$FF,$69,$4E,$9F,$E7,$ED,$7F,$A5,$B5,$FF,$38
	.dc.b	$35,$FF,$4C,$BF,$D4,$9D,$F8,$5F,$D5,$E3,$11,$C2,$3F,$82,$67,$FF
	.dc.b	$00,$23,$DD,$F2,$9F,$85,$3F,$0D,$F3,$1F,$87,$F9,$49,$3F,$0D,$D3
	.dc.b	$FC,$EB,$DC,$F5,$91,$FE,$7B,$F0,$7E,$EF,$62,$47,$F4,$07,$4F,$F5
	.dc.b	$1E,$2F,$77,$80,$E6,$FF,$7B,$C7,$AE,$31,$3F,$E0,$38,$C4,$EB,$83
	.dc.b	$FB,$47,$A3,$F8,$5D,$FF,$B3,$B6,$FF,$B9,$B5,$FF,$4C,$35,$FF,$60
	.dc.b	$BF,$8D,$F2,$B9,$E0,$69,$8A,$FD,$54,$4A,$FC,$33,$FA,$DF,$96,$FD
	.dc.b	$77,$44,$9F,$C5,$3B,$7F,$2B,$F2,$BF,$B3,$23,$F8,$FF,$66,$BF,$9A
	.dc.b	$F9,$5F,$DC,$18,$DB,$E1,$C7,$FB,$5F,$D3,$FC,$77,$9F,$23,$FD,$B9
	.dc.b	$3A,$F9,$4F,$EB,$8F,$0E,$95,$7E,$14,$95,$F9,$07,$F1,$1E,$84,$8F
	.dc.b	$C9,$B5,$FE,$87,$D5,$FE,$DD,$7F,$C8,$3C,$BF,$EE,$BF,$49,$39,$97
	.dc.b	$B6,$AD,$4E,$63,$12,$AE,$3F,$2B,$FE,$46,$BD,$FC,$95,$AD,$60,$8C
	.dc.b	$5F,$81,$B0,$F9,$3B,$E3,$CB,$F3,$9C,$9D,$FC,$2F,$2F,$9B,$FE,$F1
	.dc.b	$A7,$93,$24,$63,$3F,$73,$70,$F1,$9C,$7E,$4A,$3D,$AB,$8C,$75,$F9
	.dc.b	$3E,$DF,$25,$2B,$56,$8F,$17,$2F,$8F,$E4,$A5,$69,$F3,$7C,$9F,$9F
	.dc.b	$E3,$A4,$F2,$7D,$5F,$27,$EB,$F9,$39,$3D,$9F,$77,$99,$EF,$F8,$89
	.dc.b	$3E,$27,$E1,$7C,$9F,$C3,$F2,$64,$F9,$9D,$BF,$73,$8B,$E1,$FC,$9F
	.dc.b	$C9,$24,$6C,$F6,$C7,$CD,$23,$F0,$3C,$BA,$F3,$3F,$05,$C3,$F8,$47
	.dc.b	$1C,$9F,$4A,$BE,$B5,$70,$E5,$FA,$D7,$FE,$BC,$BF,$97,$F2,$19,$BF
	.dc.b	$AF,$E5,$7D,$D2,$4F,$D0,$F8,$DA,$F6,$BF,$F4,$83,$F5,$27,$F0,$93
	.dc.b	$AC,$3B,$FC,$57,$E1,$CC,$61,$FC,$CB,$DF,$FC,$B6,$0F,$0D,$18,$6C
	.dc.b	$7E,$48,$75,$FF,$8E,$70,$F8,$13,$A9,$F8,$87,$3A,$E4,$76,$AE,$17
	.dc.b	$E1,$9F,$F6,$81,$4F,$E0,$E7,$13,$20,$C1,$AA,$7B,$FE,$72,$47,$9E
	.dc.b	$3C,$98,$79,$A7,$33,$E6,$9E,$2E,$F0,$F8,$1E,$F7,$CC,$BF,$6B,$70
	.dc.b	$FE,$B2,$DA,$F2,$5F,$27,$A4,$65,$F9,$8D,$83,$70,$EF,$A7,$9E,$34
	.dc.b	$66,$B1,$11,$EA,$FB,$67,$8C,$60,$98,$7B,$DE,$0F,$AA,$6E,$1F,$DC
	.dc.b	$AB,$8C,$E6,$B8,$CE,$5D,$9E,$EB,$F2,$DF,$86,$32,$7B,$9B,$87,$C3
	.dc.b	$BE,$93,$FE,$02,$AF,$CB,$B6,$BF,$C1,$31,$88,$FF,$72,$F1,$BE,$6F
	.dc.b	$FD,$97,$F1,$A8,$C4,$FF,$B6,$23,$13,$A9,$CC,$FF,$84,$3E,$07,$FC
	.dc.b	$72,$FE,$75,$B9,$7F,$AB,$BC,$2F,$F0,$CC,$60,$FF,$22,$E7,$F9,$C6
	.dc.b	$E9,$FE,$18,$F8,$1F,$EF,$B3,$9A,$FC,$75,$6B,$0F,$C1,$B5,$FD,$65
	.dc.b	$70,$BF,$F1,$0C,$E7,$0F,$15,$7E,$4E,$31,$5B,$8E,$77,$D8,$8F,$E6
	.dc.b	$63,$4F,$B6,$33,$87,$96,$35,$EA,$E3,$1F,$EC,$1F,$88,$3E,$0F,$FC
	.dc.b	$2F,$E6,$FF,$86,$3A,$93,$9B,$FE,$EA,$70,$E4,$DB,$CA,$EE,$3F,$D8
	.dc.b	$EB,$F8,$4C,$F0,$8C,$4E,$AF,$AD,$DB,$ED,$4C,$3C,$51,$FB,$CC,$39
	.dc.b	$5F,$93,$93,$86,$DC,$5B,$3A,$AE,$18,$73,$8F,$7D,$6D,$E7,$F0,$6F
	.dc.b	$9C,$FE,$1F,$5F,$DE,$E3,$13,$FC,$3E,$26,$29,$CC,$7F,$B3,$9A,$EC
	.dc.b	$7C,$D7,$F7,$BA,$C4,$7F,$19,$8C,$38,$6B,$C5,$AF,$F8,$56,$5F,$1D
	.dc.b	$F2,$5B,$5E,$F3,$11,$F2,$E3,$D7,$7F,$69,$B9,$7A,$6B,$C7,$39,$6F
	.dc.b	$53,$E6,$FE,$B3,$FE,$B0,$8F,$A2,$E5,$BF,$0F,$E3,$F8,$E9,$1A,$F8
	.dc.b	$5B,$A7,$D1,$3E,$67,$C7,$79,$34,$34,$95,$E2,$F8,$3F,$77,$C0,$F8
	.dc.b	$9B,$1F,$63,$E0,$3E,$11,$E6,$F7,$EC,$7D,$CF,$81,$EC,$47,$DF,$8F
	.dc.b	$F9,$95,$F7,$FB,$DF,$2F,$DC,$F7,$B6,$3D,$FE,$F7,$CA,$E1,$34,$91
	.dc.b	$D3,$DD,$F6,$11,$C1,$F3,$3E,$6B,$91,$1C,$1F,$43,$DD,$FB,$9F,$1D
	.dc.b	$F4,$1B,$1F,$96,$F8,$EF,$90,$47,$D6,$FD,$EF,$C4,$5F,$37,$F8,$5F
	.dc.b	$8F,$FD,$74,$62,$BF,$3F,$18,$F7,$FF,$15,$E2,$D9,$FD,$1B,$F8,$DF
	.dc.b	$90,$FA,$C4,$8F,$92,$F7,$F7,$7F,$25,$E3,$F6,$6C,$FE,$DB,$17,$BD
	.dc.b	$F9,$7F,$1F,$C6,$D9,$FD,$8F,$A3,$F8,$32,$3F,$59,$F6,$7B,$9F,$9F
	.dc.b	$F2,$7D,$6D,$9F,$DC,$F9,$DF,$A3,$3D,$FE,$85,$8F,$E2,$1E,$9B,$C6
	.dc.b	$FD,$3F,$89,$FE,$AA,$6B,$FE,$AA,$6B,$FE,$0D,$8E,$1E,$57,$F0,$47
	.dc.b	$F8,$37,$C8,$5F,$34,$D2,$61,$FA,$CE,$07,$B7,$FB,$7F,$0B,$E9,$B6
	.dc.b	$7E,$37,$B5,$86,$9F,$26,$3E,$CF,$B5,$FB,$F8,$A7,$F1,$82,$E9,$FA
	.dc.b	$DF,$39,$8A,$BF,$03,$E9,$6B,$D2,$FE,$79,$2F,$67,$F0,$A7,$E1,$BC
	.dc.b	$6B,$97,$D9,$7F,$18,$47,$D8,$FE,$3F,$83,$84,$7D,$F6,$E7,$E3,$FC
	.dc.b	$4B,$E7,$E3,$FD,$2F,$F1,$3F,$E1,$53,$9E,$3F,$E8,$51,$89,$FE,$A1
	.dc.b	$4F,$4B,$F8,$CB,$71,$FE,$2E,$DC,$3F,$97,$6E,$8C,$47,$F3,$65,$E8
	.dc.b	$76,$E4,$FE,$A1,$18,$BF,$E9,$79,$7D,$A7,$FA,$6C,$3C,$FD,$9C,$3C
	.dc.b	$DB,$3C,$3D,$3F,$49,$FD,$83,$2F,$E7,$FF,$13,$89,$FE,$C1,$1F,$EB
	.dc.b	$FF,$47,$FA,$6B,$63,$FA,$1B,$4F,$E8,$BE,$5A,$76,$BE,$2E,$92,$B8
	.dc.b	$C6,$5F,$8B,$E7,$FC,$CF,$90,$F7,$B6,$3D,$AE,$1F,$7F,$9F,$71,$FB
	.dc.b	$1E,$7F,$D4,$F5,$BE,$3B,$6B,$E2,$F9,$FF,$50,$8F,$D4,$7F,$A7,$F1
	.dc.b	$3E,$43,$EE,$9F,$43,$CE,$B5,$F1,$3E,$31,$F0,$7F,$11,$B1,$F1,$1B
	.dc.b	$9F,$F6,$86,$BF,$DB,$77,$7C,$6F,$C6,$25,$7C,$47,$F1,$3E,$8F,$E3
	.dc.b	$BE,$3F,$F0,$2D,$9F,$F0,$A7,$C0,$FF,$BE,$67,$FC,$02,$D8,$FE,$EE
	.dc.b	$DC,$3F,$BE,$35,$FC,$77,$A1,$F9,$AF,$6F,$F1,$AD,$AF,$F1,$07,$CB
	.dc.b	$A6,$47,$F8,$0A,$31,$1F,$E1,$06,$C6,$63,$75,$F9,$EE,$07,$C8,$7C
	.dc.b	$E2,$57,$F7,$A8,$CB,$C7,$A7,$A5,$FE,$2E,$30,$4B,$FD,$4F,$77,$F8
	.dc.b	$0D,$FD,$C7,$9B,$FE,$3A,$CB,$D1,$EE,$7E,$A2,$31,$82,$31,$87,$BA
	.dc.b	$30,$FE,$AA,$7F,$7B,$8F,$E1,$09,$1E,$77,$CB,$DB,$23,$E0,$37,$1F
	.dc.b	$9F,$C8,$F1,$FE,$AB,$6B,$EA,$B6,$77,$3F,$AE,$FB,$47,$4D,$25,$FE
	.dc.b	$C2,$32,$FF,$8D,$25,$F7,$4F,$2B,$FC,$B7,$18,$7F,$CB,$BC,$9F,$E9
	.dc.b	$3F,$3B,$35,$BB,$FC,$5C,$FF,$31,$7F,$A8,$C6,$72,$FF,$11,$4F,$19
	.dc.b	$CD,$FF,$09,$F9,$FF,$E2,$55,$FE,$46,$6B,$FE,$4B,$6B,$FE,$53,$6B
	.dc.b	$FE,$E3,$7F,$C6,$8F,$1F,$13,$F8,$7C,$62,$39,$47,$ED,$BC,$4F,$E6
	.dc.b	$27,$F9,$CF,$F1,$4E,$3E,$67,$F3,$78,$C7,$93,$FE,$4A,$9E,$FE,$59
	.dc.b	$A4,$F2,$34,$4D,$3F,$C9,$B0,$8F,$E0,$BE,$07,$F4,$03,$C1,$DF,$6B
	.dc.b	$FC,$FB,$5F,$C2,$5E,$3E,$27,$84,$8F,$4E,$FF,$47,$5F,$A5,$B1,$CB
	.dc.b	$4C,$34,$EB,$1F,$D4,$3B,$24,$78,$1B,$C1,$FD,$AF,$BF,$FC,$7F,$3F
	.dc.b	$AD,$F8,$0F,$E3,$A4,$7D,$BF,$29,$FD,$6B,$B7,$E5,$36,$3F,$A6,$BF
	.dc.b	$C9,$08,$FE,$DD,$FE,$4F,$F0,$32,$3F,$74,$DC,$3F,$A2,$8F,$86,$FF
	.dc.b	$0A,$74,$7F,$A2,$9B,$1F,$E8,$F6,$BF,$E9,$76,$FC,$4F,$F2,$EE,$7B
	.dc.b	$E0,$7F,$AB,$6B,$F5,$97,$FE,$9D,$AD,$46,$00,$73,$DD,$FA,$58,$74
	.dc.b	$37,$7C,$9F,$F5,$C8,$C3,$63,$33,$FC,$5E,$31,$E4,$48,$FE,$34,$D7
	.dc.b	$FB,$EF,$3F,$FA,$AF,$4F,$F0,$57,$C9,$FF,$AC,$DB,$3F,$88,$31,$C1
	.dc.b	$E8,$CB,$C9,$F1,$5E,$12,$4F,$F2,$77,$C8,$F0,$3D,$79,$F7,$AE,$4D
	.dc.b	$8F,$7C,$BE,$33,$85,$A1,$F2,$78,$5F,$0B,$4C,$45,$BF,$F6,$DC,$FF
	.dc.b	$99,$9B,$F0,$99,$7F,$C7,$90,$F9,$1F,$98,$D8,$FF,$5C,$B5,$FF,$11
	.dc.b	$B5,$FF,$6B,$FE,$2B,$FE,$84,$7F,$DD,$2D,$7F,$C9,$0D,$D7,$FE,$00
	.dc.b	$4B,$E4,$FD,$5F,$0D,$E0,$25,$F8,$1F,$1F,$C2,$C8,$F1,$B5,$FF,$1A
	.dc.b	$E0,$F8,$AF,$8E,$D8,$F1,$C7,$C7,$E9,$FA,$50,$F9,$A7,$B1,$FA,$26
	.dc.b	$C7,$F8,$A1,$AF,$FB,$B1,$AF,$FB,$C5,$FE,$82,$EF,$D0,$E4,$47,$FB
	.dc.b	$EE,$8F,$57,$F0,$5D,$6F,$C4,$5E,$4E,$1B,$7F,$E0,$8B,$E0,$FF,$A3
	.dc.b	$A3,$2F,$FB,$7E,$B9,$57,$13,$F1,$DE,$7F,$E1,$F0,$F7,$FE,$01,$FE
	.dc.b	$8A,$BF,$F2,$1B,$5F,$D8,$91,$C3,$80,$E1,$FD,$BC,$FF,$75,$F8,$0F
	.dc.b	$F2,$6B,$63,$FC,$37,$E4,$7F,$99,$1B,$3F,$E5,$2F,$5F,$FC,$2D,$E3
	.dc.b	$5F,$30,$D2,$57,$84,$FE,$03,$C9,$FA,$37,$1D,$A8,$C4,$7E,$0B,$BB
	.dc.b	$F8,$24,$62,$3D,$B8,$7F,$89,$7C,$1F,$D4,$7C,$3C,$9F,$F0,$64,$57
	.dc.b	$F4,$71,$7B,$1E,$17,$F1,$6B,$E5,$5F,$E1,$33,$D4,$43,$FC,$58,$78
	.dc.b	$BE,$5E,$3E,$8C,$51,$88,$F9,$D2,$FC,$5B,$D1,$1D,$DA,$A3,$59,$FC
	.dc.b	$B9,$CE,$9F,$C1,$7E,$1B,$F4,$3E,$1C,$C3,$F8,$0C,$7E,$13,$F1,$7F
	.dc.b	$E0,$51,$FE,$4E,$6B,$F8,$2E,$44,$62,$3E,$FB,$5F,$E2,$BD,$1F,$AE
	.dc.b	$F1,$64,$7F,$0E,$8F,$D6,$FC,$B3,$E2,$3F,$A6,$B6,$7F,$D2,$6D,$7E
	.dc.b	$F3,$5F,$C8,$7C,$EB,$97,$FA,$AB,$63,$FD,$36,$D7,$F0,$CD,$7F,$89
	.dc.b	$F9,$1F,$C0,$88,$FF,$4E,$E5,$FB,$6F,$F4,$CF,$C6,$1F,$1D,$FC,$B9
	.dc.b	$7C,$B7,$F4,$D6,$CF,$E4,$1A,$FF,$B1,$1A,$FE,$AB,$89,$1F,$EC,$D8
	.dc.b	$FE,$47,$E5,$FF,$14,$3B,$7C,$67,$7F,$8C,$93,$E5,$EB,$FE,$99,$7F
	.dc.b	$EC,$58,$C1,$E7,$EF,$DB,$FC,$B3,$CD,$FE,$3E,$63,$5F,$6C,$38,$7C
	.dc.b	$4F,$CE,$CB,$E9,$91,$FA,$7A,$FD,$62,$38,$3F,$31,$F3,$7F,$CB,$08
	.dc.b	$FF,$B4,$3A,$DF,$0B,$F2,$7E,$2E,$4F,$A3,$87,$E1,$C7,$CD,$F8,$DF
	.dc.b	$9D,$F1,$BF,$CD,$BD,$29,$7B,$E1,$4F,$D3,$F9,$5F,$A4,$49,$FA,$4F
	.dc.b	$D5,$F3,$BF,$9F,$FA,$72,$79,$47,$DB,$F8,$BF,$B9,$E6,$C9,$ED,$15
	.dc.b	$FB,$51,$F7,$C8,$FD,$FF,$95,$FC,$21,$1F,$34,$7E,$1F,$E5,$24,$7A
	.dc.b	$5F,$C5,$7C,$AE,$23,$FB,$0F,$AD,$FD,$50,$C7,$47,$E4,$B8,$BF,$C6
	.dc.b	$1F,$2B,$F9,$1E,$C9,$7F,$27,$DF,$4E,$23,$FB,$5F,$CA,$FF,$5E,$31
	.dc.b	$87,$E7,$3C,$67,$E3,$3F,$EA,$9A,$FF,$48,$35,$FE,$79,$1A,$F5,$3F
	.dc.b	$97,$47,$FA,$86,$74,$FF,$44,$97,$E3,$5F,$CE,$E3,$8B,$5D,$35,$FF
	.dc.b	$1D,$C7,$0F,$0B,$F9,$9F,$CE,$FF,$74,$24,$FB,$7E,$37,$FC,$3B,$AB
	.dc.b	$E5,$BF,$BC,$FA,$D2,$BE,$14,$7F,$7C,$F9,$1F,$E7,$A4,$8F,$E5,$8F
	.dc.b	$F9,$0E,$BF,$B2,$ED,$F3,$7D,$77,$CC,$7F,$81,$0E,$FE,$E7,$83,$E5
	.dc.b	$3F,$C0,$FE,$B4,$CB,$E8,$47,$F8,$2B,$E2,$3F,$A5,$92,$3F,$9D,$3D
	.dc.b	$FE,$3F,$FC,$1F,$C5,$27,$EC,$47,$B3,$E2,$FF,$C7,$3F,$37,$FC,$3A
	.dc.b	$7F,$48,$D5,$AF,$E0,$F6,$57,$0F,$5F,$FC,$BD,$7F,$C9,$6F,$FC,$A0
	.dc.b	$D7,$FC,$CA,$DE,$CF,$F3,$D7,$BF,$E0,$F4,$25,$7E,$2E,$3E,$7F,$C4
	.dc.b	$FC,$72,$47,$F6,$CC,$91,$E0,$F4,$64,$FF,$70,$C7,$F8,$FB,$E2,$8F
	.dc.b	$A0,$C3,$91,$4F,$F0,$56,$2F,$C6,$FF,$BB,$EB,$FA,$1B,$FC,$63,$7A
	.dc.b	$FF,$CB,$F3,$AF,$07,$DF,$EA,$FC,$22,$63,$E3,$7E,$2F,$CE,$7C,$E2
	.dc.b	$47,$F1,$1A,$FF,$32,$B7,$D7,$FA,$7E,$8C,$8C,$1F,$53,$CF,$F9,$E4
	.dc.b	$7E,$03,$F4,$BD,$2F,$B6,$48,$ED,$E9,$7A,$3C,$9F,$4F,$91,$27,$C5
	.dc.b	$1E,$CF,$17,$FD,$31,$99,$A4,$9F,$9D,$F1,$1F,$88,$23,$EB,$F1,$7F
	.dc.b	$CB,$3F,$3A,$F9,$0F,$F5,$CB,$67,$F7,$6D,$7F,$C5,$4D,$7F,$AB,$72
	.dc.b	$FF,$96,$1F,$AA,$D7,$3E,$97,$E4,$0F,$3F,$83,$F5,$3C,$39,$3F,$36
	.dc.b	$2B,$F9,$AE,$6F,$F5,$26,$AA,$CF,$47,$F1,$F8,$FF,$5E,$56,$CF,$F5
	.dc.b	$ED,$7F,$80,$E7,$F9,$23,$FA,$56,$FB,$7F,$EA,$28,$D1,$FE,$B1,$34
	.dc.b	$57,$46,$FC,$0F,$F0,$9F,$A2,$F1,$3F,$D9,$FE,$07,$F0,$4C,$7F,$DA
	.dc.b	$3E,$D7,$F9,$D6,$73,$E4,$7F,$99,$23,$13,$FE,$E5,$8C,$7A,$7F,$EE
	.dc.b	$17,$58,$75,$9D,$1D,$1C,$61,$DE,$3F,$BA,$46,$27,$9C,$78,$0E,$18
	.dc.b	$57,$6D,$C3,$C7,$7C,$D6,$1E,$3A,$E2,$78,$8E,$25,$8E,$37,$E0,$8C
	.dc.b	$61,$E3,$78,$DF,$58,$C4,$7F,$82,$EF,$B3,$D0,$FF,$43,$5F,$C6,$6C
	.dc.b	$FC,$36,$E1,$FD,$3B,$E3,$DE,$B3,$E0,$B0,$DA,$FF,$40,$4E,$A7,$27
	.dc.b	$F9,$23,$D5,$FF,$71,$4F,$F9,$21,$AF,$89,$AF,$D9,$6B,$F7,$9E,$39
	.dc.b	$FC,$96,$E9,$F7,$FD,$6F,$F2,$2C,$E7,$E2,$BE,$85,$6A,$F9,$47,$D3
	.dc.b	$AD,$DF,$78,$D9,$27,$E8,$4E,$9E,$27,$D3,$AE,$8F,$F8,$DD,$BA,$A7
	.dc.b	$84,$E6,$BA,$61,$E5,$8E,$BC,$38,$C7,$FA,$2F,$B1,$FE,$5E,$E3,$FE
	.dc.b	$5D,$EC,$DA,$C9,$FD,$49,$C5,$FF,$4A,$AB,$E4,$BA,$91,$85,$F3,$76
	.dc.b	$31,$BE,$79,$6E,$5F,$D7,$9B,$8F,$EF,$F1,$E9,$5C,$F9,$73,$A7,$F9
	.dc.b	$63,$B1,$F2,$1F,$E5,$AC,$3F,$DA,$B1,$8C,$BF,$8E,$CE,$64,$27,$4F
	.dc.b	$F1,$CE,$2F,$6B,$99,$F5,$46,$27,$77,$FD,$4E,$FD,$C7,$BF,$DD,$AF
	.dc.b	$F9,$6F,$B7,$FC,$66,$FF,$B4,$9B,$8F,$F5,$E6,$BF,$E8,$36,$BE,$6D
	.dc.b	$10,$F6,$3E,$7B,$6B,$B7,$CB,$62,$73,$EC,$C9,$F1,$7C,$B5,$C7,$E6
	.dc.b	$60,$F9,$6F,$89,$97,$66,$C6,$6F,$D1,$18,$F8,$DF,$09,$E2,$F6,$AC
	.dc.b	$FB,$FD,$AF,$29,$E6,$F9,$5B,$1F,$0F,$E5,$BA,$91,$F8,$8F,$83,$E0
	.dc.b	$7D,$A7,$BB,$DD,$B3,$F2,$FD,$DF,$83,$E6,$7C,$26,$C7,$CF,$F9,$6F
	.dc.b	$49,$1F,$A0,$FB,$3A,$7E,$4F,$BB,$F2,$9B,$3F,$5F,$E2,$BE,$6F,$BD
	.dc.b	$F3,$9B,$1F,$6F,$C6,$F8,$A4,$7E,$E3,$F2,$BE,$71,$F3,$BF,$A4,$6C
	.dc.b	$FE,$0F,$9F,$13,$9F,$88,$D7,$CB,$5A,$FF,$51,$5A,$39,$7C,$9F,$52
	.dc.b	$4F,$E6,$B0,$FF,$65,$52,$7C,$5F,$1F,$E1,$3C,$5F,$23,$6F,$AF,$C0
	.dc.b	$F9,$7E,$3F,$CC,$D8,$F0,$FC,$7F,$AC,$8F,$7C,$BF,$29,$92,$7E,$07
	.dc.b	$C0,$7B,$CF,$81,$F1,$F6,$FD,$7E,$B7,$C3,$3F,$84,$FA,$FF,$E9,$6F
	.dc.b	$63,$13,$FE,$D8,$8F,$F7,$6F,$91,$FB,$76,$C7,$87,$3F,$EE,$9E,$CA
	.dc.b	$B4,$D9,$D7,$8C,$D5,$F0,$1E,$0A,$D6,$1F,$C3,$67,$37,$E1,$9C,$F9
	.dc.b	$9F,$EA,$F9,$7B,$DF,$6F,$E0,$7E,$E3,$6F,$EB,$B8,$7E,$CF,$C0,$DC
	.dc.b	$7B,$7C,$5F,$E0,$CF,$C2,$7B,$56,$BE,$FB,$67,$C8,$DC,$BC,$5E,$6F
	.dc.b	$E2,$FE,$03,$F1,$8D,$BF,$C3,$F0,$E1,$A7,$E2,$A3,$F9,$AF,$CE,$37
	.dc.b	$FC,$A5,$61,$B5,$FE,$F1,$97,$C5,$F0,$E4,$E2,$D9,$D7,$C5,$E1,$A7
	.dc.b	$18,$EF,$F1,$5E,$0F,$8B,$B3,$D5,$F1,$FC,$54,$8F,$07,$C5,$DE,$1F
	.dc.b	$F1,$F1,$FC,$D7,$CC,$FE,$60,$7F,$34,$F2,$5E,$96,$2F,$F9,$DF,$19
	.dc.b	$FD,$24,$BD,$1F,$E8,$18,$7F,$A5,$9B,$19,$9F,$F6,$DB,$67,$FC,$C2
	.dc.b	$F7,$9F,$E8,$0D,$9F,$8D,$F1,$7F,$30,$F9,$BF,$17,$63,$E5,$36,$FF
	.dc.b	$AD,$71,$23,$18,$FF,$BE,$E1,$F1,$5F,$5B,$E2,$FE,$BB,$70,$FA,$9F
	.dc.b	$0D,$86,$9F,$5E,$3E,$EF,$C2,$C8,$F8,$67,$F0,$5F,$9D,$FC,$EE,$7E
	.dc.b	$29,$F9,$5E,$BF,$C2,$25,$7C,$27,$FB,$37,$C5,$3C,$57,$B9,$C2,$BA
	.dc.b	$E1,$FE,$33,$9C,$C6,$23,$FD,$87,$7F,$E0,$6B,$E5,$EE,$7E,$4B,$E1
	.dc.b	$F1,$82,$31,$87,$E4,$7E,$1F,$31,$F9,$86,$C7,$83,$E5,$7C,$79,$FE
	.dc.b	$73,$E1,$FF,$C2,$71,$89,$FC,$D4,$7E,$83,$E2,$33,$5E,$6D,$7E,$9C
	.dc.b	$62,$33,$D3,$FE,$05,$C3,$DE,$D8,$FC,$7E,$1F,$C4,$67,$2F,$F0,$EE
	.dc.b	$4F,$F1,$83,$E5,$9D,$E1,$F8,$0F,$88,$EB,$39,$F8,$1F,$D9,$FC,$3F
	.dc.b	$91,$B1,$92,$57,$E2,$1D,$1D,$7B,$7C,$84,$7C,$98,$23,$FB,$DF,$CD
	.dc.b	$7F,$83,$BA,$BF,$49,$6D,$3E,$DD,$7F,$8E,$2B,$58,$7F,$89,$27,$2F
	.dc.b	$DB,$F9,$DF,$CA,$BF,$CC,$6D,$18,$7F,$9A,$35,$FE,$8C,$D7,$FC,$5F
	.dc.b	$CB,$FE,$79,$F2,$BF,$C5,$B3,$C7,$2F,$2C,$E6,$FC,$1D,$5F,$89,$87
	.dc.b	$99,$FD,$23,$E7,$7F,$C4,$75,$F9,$29,$FF,$42,$4E,$7C,$CE,$0C,$9F
	.dc.b	$E9,$9F,$2C,$E7,$C6,$BF,$C3,$BE,$77,$FC,$54,$FF,$4D,$6B,$FD,$7D
	.dc.b	$AF,$F6,$F6,$BF,$E4,$A5,$F1,$5F,$E4,$87,$9E,$BF,$42,$51,$AC,$3F
	.dc.b	$4C,$E9,$FF,$1F,$70,$D9,$FD,$83,$7B,$0F,$F5,$49,$DF,$C6,$7D,$9C
	.dc.b	$7E,$F9,$23,$FB,$0C,$F7,$BF,$EF,$17,$CB,$E3,$38,$57,$EE,$A3,$0D
	.dc.b	$AE,$19,$7E,$22,$7C,$3D,$DF,$87,$3F,$11,$F1,$96,$F8,$74,$7E,$40
	.dc.b	$8F,$89,$FC,$9F,$C6,$7E,$3C,$FC,$87,$95,$67,$C0,$E1,$F1,$77,$DD
	.dc.b	$7F,$D0,$5A,$3B,$3F,$30,$7E,$67,$C3,$B7,$E9,$F8,$CF,$D0,$91,$F8
	.dc.b	$6F,$E9,$3E,$33,$F4,$07,$E8,$7B,$EC,$FC,$1F,$8C,$FD,$49,$1F,$96
	.dc.b	$FE,$B3,$E3,$24,$7C,$E6,$E7,$FA,$51,$F1,$FF,$E8,$EF,$63,$FC,$F3
	.dc.b	$1F,$E2,$98,$C7,$6F,$F8,$6D,$B1,$91,$3C,$17,$C8,$FF,$86,$B9,$F7
	.dc.b	$18,$F7,$D1,$89,$D4,$F2,$C3,$FC,$8D,$18,$BF,$1F,$47,$E3,$3D,$2F
	.dc.b	$DE,$36,$FF,$13,$18,$E3,$37,$E3,$A4,$FE,$39,$AF,$E3,$A7,$FC,$8D
	.dc.b	$97,$E2,$5B,$B7,$E7,$A7,$FC,$B1,$39,$F8,$EB,$3F,$E9,$F7,$FC,$D3
	.dc.b	$3F,$E3,$A6,$BF,$9A,$6B,$FE,$61,$C5,$18,$8E,$5F,$1B,$D1,$FE,$B5
	.dc.b	$08,$C4,$F0,$6C,$7F,$9B,$E7,$F4,$8B,$E2,$B2,$48,$CF,$C6,$70,$F8
	.dc.b	$C9,$1C,$1D,$3F,$ED,$9A,$7C,$67,$CE,$6C,$7B,$63,$06,$7E,$2F,$C6
	.dc.b	$48,$F1,$FC,$67,$9B,$E2,$FC,$FF,$19,$27,$CF,$15,$F8,$D1,$DC,$8F
	.dc.b	$CB,$E6,$F7,$91,$F7,$FC,$67,$C7,$5F,$39,$FE,$17,$7F,$BE,$2F,$77
	.dc.b	$4D,$8C,$FB,$D8,$9F,$D1,$4E,$67,$FC,$F0,$D7,$FD,$14,$D7,$FA,$2E
	.dc.b	$BF,$CD,$5F,$E7,$8D,$7F,$3B,$C7,$AF,$37,$EE,$74,$BF,$1A,$78,$7F
	.dc.b	$7F,$D1,$FC,$1D,$6B,$E7,$7F,$C1,$D8,$FF,$A3,$5A,$F4,$13,$C1,$F0
	.dc.b	$FB,$3F,$0F,$E7,$7F,$D1,$D1,$FE,$1D,$6B,$C6,$7F,$23,$3B,$F9,$0F
	.dc.b	$E0,$55,$AA,$FF,$54,$35,$FF,$5C,$35,$FC,$AF,$C7,$FE,$5F,$DB,$FF
	.dc.b	$17,$F8,$F7,$CF,$3F,$96,$9C,$3F,$DC,$E7,$75,$FE,$02,$8C,$57,$08
	.dc.b	$FF,$02,$F8,$7C,$C8,$FE,$7B,$E3,$AB,$F9,$7F,$8F,$FD,$19,$8F,$13
	.dc.b	$FA,$17,$2F,$F7,$29,$FF,$0F,$1E,$53,$F5,$3E,$CD,$7F,$C2,$67,$88
	.dc.b	$87,$EE,$FE,$3F,$DD,$F3,$BF,$B3,$8F,$F1,$E3,$5F,$D4,$46,$BC,$6F
	.dc.b	$DA,$FB,$F2,$77,$1F,$B8,$F3,$3F,$01,$F3,$BF,$AC,$9F,$A8,$DC,$FD
	.dc.b	$33,$AF,$1B,$F7,$FD,$92,$FF,$C1,$57,$BE,$DF,$F0,$BA,$F1,$7F,$49
	.dc.b	$F2,$EF,$91,$A7,$82,$2D,$FF,$89,$67,$4D,$9F,$E8,$91,$88,$FE,$E7
	.dc.b	$F3,$BF,$EB,$37,$FB,$93,$5F,$CA,$79,$A8,$C6,$BF,$E0,$BF,$80,$73
	.dc.b	$FF,$1A,$9E,$93,$FD,$DD,$7C,$96,$4D,$65,$71,$EF,$F2,$5C,$4C,$3E
	.dc.b	$0A,$7C,$8F,$5F,$92,$93,$E6,$8A,$EE,$2E,$3E,$19,$FE,$15,$F4,$2F
	.dc.b	$43,$E4,$E3,$30,$FF,$43,$4E,$5F,$F4,$46,$FF,$ED,$07,$E5,$4F,$C8
	.dc.b	$23,$FE,$51,$E6,$7C,$4F,$F1,$76,$C7,$FA,$45,$AF,$FB,$2A,$BE,$1B
	.dc.b	$F3,$BE,$4F,$E4,$11,$FA,$3F,$27,$F2,$FC,$FF,$99,$F2,$72,$7E,$AC
	.dc.b	$7C,$FF,$43,$E8,$7C,$9C,$9F,$B3,$15,$F9,$5F,$19,$F5,$4C,$70,$FD
	.dc.b	$5F,$92,$FB,$1F,$25,$2B,$F0,$31,$5F,$A5,$E8,$FD,$D3,$15,$F7,$5F
	.dc.b	$C0,$7C,$94,$9F,$C4,$C7,$E0,$FE,$4B,$F0,$9E,$8C,$9F,$C6,$C5,$7E
	.dc.b	$F4,$5B,$FC,$44,$FF,$64,$F8,$17,$CF,$52,$5F,$E5,$B9,$6B,$F8,$7F
	.dc.b	$0F,$F2,$46,$30,$FF,$24,$5F,$F2,$67,$FD,$23,$E2,$3E,$7B,$FC,$75
	.dc.b	$7F,$EC,$A8,$C7,$C4,$78,$7D,$5F,$E7,$B5,$AA,$10,$FF,$12,$B9,$71
	.dc.b	$5C,$9D,$1F,$E3,$18,$D3,$CB,$C5,$FF,$1A,$FC,$87,$BC,$91,$E3,$F9
	.dc.b	$EF,$D0,$BF,$E6,$5C,$DF,$21,$F1,$49,$19,$F9,$1F,$91,$F2,$1F,$E2
	.dc.b	$AA,$34,$93,$E4,$3F,$DD,$91,$FE,$EC,$7B,$7C,$8F,$D1,$3E,$97,$AF
	.dc.b	$F4,$FE,$3E,$4F,$8D,$FA,$DF,$23,$F5,$FE,$47,$C0,$49,$ED,$B7,$F9
	.dc.b	$4A,$3C,$9C,$5F,$77,$E0,$65,$7B,$63,$F0,$1F,$1F,$23,$E0,$3F,$83
	.dc.b	$F9,$1F,$C2,$7C,$8F,$B4,$93,$DB,$64,$7C,$1F,$91,$FC,$57,$C0,$4A
	.dc.b	$F9,$51,$F8,$DF,$53,$F1,$DF,$21,$27,$E7,$47,$E4,$7D,$FF,$94,$48
	.dc.b	$EC,$FF,$95,$BE,$77,$FC,$87,$DB,$F8,$E9,$FC,$C1,$FE,$3C,$8C,$F7
	.dc.b	$7D,$F7,$FA,$64,$FF,$A6,$DB,$19,$F3,$F8,$7C,$D7,$1F,$90,$EC,$4C
	.dc.b	$3E,$5D,$A3,$C5,$F2,$1D,$C9,$3F,$EA,$B9,$FF,$24,$37,$DC,$F2,$11
	.dc.b	$D7,$BD,$E7,$F5,$FD,$84,$8E,$EF,$BF,$E3,$FD,$84,$8E,$DF,$21,$EF
	.dc.b	$F9,$09,$1E,$17,$E1,$7C,$87,$C3,$F8,$BF,$90,$49,$F5,$BF,$2F,$E2
	.dc.b	$7E,$41,$23,$D5,$F1,$5F,$2F,$E2,$BF,$C8,$9F,$38,$ED,$A6,$6B,$F6
	.dc.b	$BC,$DF,$ED,$F9,$D4,$7E,$7C,$38,$7F,$44,$D7,$F6,$ED,$7E,$C7,$C4
	.dc.b	$7D,$B2,$3F,$37,$E2,$AB,$FD,$9B,$8D,$F1,$5F,$81,$33,$87,$FB,$67
	.dc.b	$A7,$FB,$29,$FC,$04,$B4,$F1,$51,$85,$7F,$EB,$29,$FF,$73,$4E,$4F
	.dc.b	$EE,$67,$3F,$9A,$FF,$45,$BD,$21,$F0,$4F,$63,$97,$88,$D3,$F8,$D4
	.dc.b	$E7,$D3,$FE,$3B,$39,$BF,$E3,$F3,$97,$FD,$97,$5C,$3D,$9F,$E4,$55
	.dc.b	$B9,$EF,$87,$6A,$EA,$D8,$E6,$D7,$CB,$5C,$A7,$78,$78,$9B,$3E,$03
	.dc.b	$8C,$65,$F2,$CF,$17,$26,$CF,$1C,$6D,$A6,$70,$4F,$18,$CE,$1E,$B9
	.dc.b	$F8,$0D,$C3,$D1,$1C,$27,$A3,$F1,$5F,$E7,$ED,$7D,$26,$23,$ED,$3F
	.dc.b	$D1,$3E,$55,$CF,$FE,$80,$8F,$E6,$91,$8F,$90,$FC,$C3,$63,$33,$B7
	.dc.b	$BF,$B5,$FE,$D0,$7F,$9E,$60,$8C,$46,$BC,$4F,$F2,$24,$62,$3F,$02
	.dc.b	$F1,$CB,$E0,$37,$5F,$C2,$FB,$7F,$94,$AF,$91,$39,$7F,$0A,$D7,$F1
	.dc.b	$ED,$C1,$3C,$6F,$AE,$1F,$22,$B4,$6E,$75,$87,$92,$74,$4C,$3E,$C4
	.dc.b	$E9,$F0,$C6,$6F,$AC,$6B,$7C,$5F,$F6,$FE,$27,$C1,$FA,$F9,$3F,$69
	.dc.b	$F0,$B6,$F5,$37,$0F,$DA,$42,$71,$7F,$B7,$F8,$5A,$F4,$9F,$E2,$27
	.dc.b	$78,$CD,$EB,$E1,$38,$CF,$F9,$90,$DE,$1F,$C0,$6D,$3A,$F8,$BE,$15
	.dc.b	$A9,$FE,$0D,$3E,$0F,$84,$CD,$FF,$9D,$AB,$FC,$CE,$7F,$70,$2B,$FE
	.dc.b	$12,$AF,$E1,$BF,$11,$F6,$E3,$FC,$35,$18,$9F,$EE,$EB,$E1,$7C,$F5
	.dc.b	$EE,$6C,$FF,$12,$F8,$9F,$59,$F0,$BE,$1A,$C7,$9E,$DF,$0D,$DF,$1F
	.dc.b	$A4,$D7,$E0,$36,$3F,$91,$35,$FE,$31,$3C,$FA,$FE,$7F,$5F,$D0,$7E
	.dc.b	$5E,$F2,$7C,$1F,$0B,$71,$F8,$9C,$3D,$64,$D2,$3F,$E2,$3E,$D7,$77
	.dc.b	$F9,$42,$3F,$C8,$9F,$0D,$87,$FC,$8D,$19,$BF,$F7,$C5,$7F,$95,$A3
	.dc.b	$EE,$FC,$44,$AF,$B1,$2F,$87,$C5,$F0,$BF,$F1,$AD,$78,$5B,$1F,$E4
	.dc.b	$8A,$F1,$FC,$17,$C7,$6D,$7F,$5B,$F5,$DE,$F7,$CB,$B4,$62,$75,$39
	.dc.b	$8C,$4F,$E0,$E7,$F1,$E7,$F4,$CE,$8F,$CE,$DB,$E0,$FF,$2F,$1F,$98
	.dc.b	$E9,$F9,$44,$AF,$C7,$3F,$66,$BD,$3D,$9F,$C1,$F9,$51,$8A,$FF,$30
	.dc.b	$2F,$83,$EF,$AF,$E9,$3A,$A5,$7C,$8F,$81,$BA,$FE,$77,$E2,$3E,$D1
	.dc.b	$27,$F3,$CF,$87,$E0,$BF,$5B,$D7,$EA,$6C,$FD,$2F,$82,$FD,$8F,$6F
	.dc.b	$B9,$B1,$FD,$B3,$B3,$F0,$C4,$7F,$54,$FB,$7E,$07,$E2,$9F,$1B,$E0
	.dc.b	$6C,$FD,$EF,$81,$FD,$EF,$81,$F2,$9B,$1F,$DE,$3C,$9F,$C9,$11,$FD
	.dc.b	$B3,$F2,$7C,$8F,$A2,$7D,$2D,$EC,$FF,$80,$3C,$9F,$E1,$27,$F1,$2F
	.dc.b	$8F,$7C,$06,$27,$F8,$34,$3E,$73,$F5,$8D,$8F,$F3,$9B,$6F,$F8,$23
	.dc.b	$5F,$E0,$6B,$E0,$BB,$57,$76,$D7,$6F,$95,$F1,$1E,$3F,$80,$B1,$E2
	.dc.b	$6E,$0E,$DF,$C2,$62,$9C,$C7,$F4,$D8,$7E,$0A,$F8,$6F,$E4,$FE,$57
	.dc.b	$F3,$39,$CC,$63,$1F,$E5,$31,$86,$D7,$17,$E1,$FC,$1F,$18,$C6,$5E
	.dc.b	$39,$FF,$59,$53,$E0,$FE,$61,$BE,$5B,$3D,$1F,$E3,$13,$E6,$6D,$9A
	.dc.b	$7D,$0F,$82,$E5,$5F,$51,$B8,$78,$4F,$F2,$05,$FF,$80,$A7,$FC,$7A
	.dc.b	$7D,$DF,$5B,$FA,$3B,$F7,$FE,$07,$0F,$DE,$6D,$73,$F1,$56,$F8,$0F
	.dc.b	$EA,$FF,$23,$FC,$5F,$87,$E8,$C6,$3D,$0F,$C5,$7C,$14,$CF,$F1,$3E
	.dc.b	$C5,$CF,$A6,$3F,$C6,$67,$FA,$93,$63,$D4,$3E,$FD,$CF,$E0,$F4,$BF
	.dc.b	$B5,$4F,$F1,$4E,$2F,$E2,$44,$C3,$F8,$D3,$FC,$57,$BF,$7E,$87,$CB
	.dc.b	$9F,$EC,$B5,$FC,$72,$73,$5F,$83,$F9,$37,$1F,$BB,$E3,$B0,$DC,$3F
	.dc.b	$0B,$3A,$6C,$7E,$A7,$E0,$77,$1F,$A8,$F8,$2F,$8F,$EA,$7E,$B5,$B1
	.dc.b	$FD,$C7,$97,$FA,$16,$9F,$3F,$99,$5F,$E0,$49,$C9,$F9,$E7,$E8,$FC
	.dc.b	$6C,$BE,$5F,$25,$70,$F9,$3F,$3B,$E5,$7F,$DE,$8D,$44,$0E,$5C,$35
	.dc.b	$DB,$F8,$19,$F9,$2F,$5C,$3F,$C0,$0E,$67,$C3,$A7,$A7,$DF,$FC,$3F
	.dc.b	$A9,$70,$FB,$EF,$B7,$DE,$C5,$FE,$3F,$14,$61,$FE,$80,$7F,$88,$E3
	.dc.b	$F1,$3E,$B7,$C5,$F7,$BF,$26,$FC,$7F,$77,$18,$FC,$D6,$D7,$E9,$78
	.dc.b	$1C,$7F,$EA,$CA,$F8,$53,$F1,$27,$3C,$7F,$B4,$AD,$4E,$A7,$35,$C2
	.dc.b	$3F,$D5,$D9,$7F,$60,$7B,$57,$08,$F4,$FB,$7F,$8D,$C3,$F8,$87,$BD
	.dc.b	$F1,$E7,$FA,$E7,$6B,$E3,$F3,$E1,$7E,$B9,$B9,$FF,$14,$BE,$0F,$DA
	.dc.b	$F0,$FF,$91,$7B,$7A,$D7,$C3,$D3,$E9,$25,$79,$F7,$F8,$31,$FE,$25
	.dc.b	$D1,$ED,$7E,$86,$3F,$20,$DA,$FF,$25,$CB,$F1,$CF,$F2,$76,$1F,$22
	.dc.b	$31,$CF,$FE,$31,$6D,$7C,$56,$CF,$C9,$FC,$77,$FC,$F3,$3D,$2F,$9D
	.dc.b	$7F,$37,$9C,$D7,$F1,$B8,$E2,$7F,$93,$7E,$73,$F8,$33,$FE,$4F,$FB
	.dc.b	$7F,$F5,$7E,$1F,$BC,$9C,$BF,$9D,$8C,$D7,$F0,$C9,$78,$1F,$E1,$D6
	.dc.b	$D7,$FB,$ED,$B3,$F7,$5A,$FE,$1B,$F1,$EF,$F1,$5B,$7B,$3F,$C4,$D1
	.dc.b	$88,$FD,$63,$5F,$D7,$BF,$E1,$5A,$FE,$22,$FF,$86,$63,$3D,$5F,$C7
	.dc.b	$23,$1C,$4E,$67,$C0,$7F,$30,$3F,$82,$7B,$37,$1F,$E1,$5E,$D7,$DB
	.dc.b	$23,$F9,$5B,$FE,$CD,$1F,$7F,$D9,$FE,$18,$FF,$0D,$6C,$FF,$32,$F9
	.dc.b	$4F,$E7,$A7,$FA,$53,$E2,$FF,$9F,$47,$E3,$F2,$4F,$F1,$58,$FC,$51
	.dc.b	$0F,$07,$11,$FF,$39,$57,$CC,$8F,$F1,$D3,$F5,$B6,$64,$8D,$57,$E7
	.dc.b	$5D,$46,$7C,$6F,$27,$8F,$E5,$6D,$7F,$4A,$F8,$5F,$EA,$A7,$A7,$BA
	.dc.b	$C7,$F4,$EF,$82,$FF,$42,$6A,$69,$23,$E4,$65,$EE,$7F,$AA,$FC,$0F
	.dc.b	$C3,$DB,$FB,$0B,$FD,$8B,$E3,$B9,$92,$7E,$86,$08,$EF,$ED,$49,$CB
	.dc.b	$77,$FE,$14,$7F,$34,$F8,$AF,$E6,$BF,$0D,$FD,$0A,$3F,$8A,$C6,$3C
	.dc.b	$1F,$E2,$0D,$8C,$9F,$E6,$48,$FE,$79,$F1,$8F,$90,$FF,$13,$4E,$4F
	.dc.b	$E8,$31,$9E,$5E,$50,$8C,$3B,$31,$3C,$B0,$F2,$72,$FE,$57,$E1,$7D
	.dc.b	$2D,$AF,$83,$18,$D8,$DF,$4A,$4F,$C6,$6B,$FE,$CB,$E2,$DB,$FE,$04
	.dc.b	$F4,$FF,$AB,$69,$FC,$99,$BC,$1F,$85,$C2,$D7,$E2,$B1,$FF,$4B,$BF
	.dc.b	$4B,$D4,$F1,$11,$FA,$BF,$29,$C3,$4F,$D7,$36,$8E,$73,$B5,$FE,$E7
	.dc.b	$5F,$0D,$C6,$3F,$CC,$E7,$FB,$AC,$A7,$F8,$A6,$36,$E7,$E1,$78,$7C
	.dc.b	$2C,$8C,$1C,$7E,$53,$E5,$A6,$91,$FC,$6D,$F9,$C8,$F8,$7D,$BF,$27
	.dc.b	$C3,$79,$7E,$16,$4F,$96,$2B,$FE,$F9,$11,$E0,$23,$F9,$A3,$E2,$10
	.dc.b	$F7,$7C,$2F,$D0,$5F,$8F,$BD,$AC,$3F,$EE,$48,$FF,$17,$B5,$FF,$1E
	.dc.b	$BB,$73,$EA,$FE,$89,$E6,$D8,$E1,$1F,$DD,$9A,$FF,$86,$7E,$79,$F1
	.dc.b	$BC,$5B,$1F,$AC,$6B,$FD,$DD,$AF,$F8,$01,$7A,$FE,$BF,$B4,$FA,$13
	.dc.b	$FE,$46,$6B,$FE,$54,$9D,$BF,$E8,$6F,$C8,$1E,$E7,$E4,$BB,$1F,$0F
	.dc.b	$C1,$FE,$45,$5F,$E9,$78,$C7,$67,$F8,$3E,$F8,$CF,$F6,$E9,$EE,$FE
	.dc.b	$81,$AF,$F8,$E2,$BA,$F2,$7F,$08,$6C,$7F,$3F,$3F,$80,$1F,$E5,$6E
	.dc.b	$FF,$D2,$FB,$7F,$E2,$07,$F3,$1E,$0E,$3D,$EF,$F7,$5D,$7F,$81,$5A
	.dc.b	$FF,$87,$DA,$FF,$91,$EB,$93,$FE,$0B,$FA,$27,$C2,$76,$6C,$7F,$AA
	.dc.b	$9A,$FF,$A2,$CF,$F3,$69,$98,$FD,$97,$AB,$FE,$CD,$8F,$76,$DF,$E1
	.dc.b	$D8,$C4,$7F,$8E,$6B,$9E,$7F,$33,$C1,$FF,$2D,$FC,$47,$BB,$F2,$17
	.dc.b	$C9,$53,$F6,$D1,$70,$FF,$11,$4E,$9B,$3F,$CC,$B4,$4E,$67,$F4,$DF
	.dc.b	$15,$FB,$FF,$84,$91,$FB,$F8,$AF,$FA,$9F,$4F,$E0,$39,$F9,$5B,$AF
	.dc.b	$F8,$DE,$57,$FC,$27,$87,$F8,$C8,$FC,$8B,$C1,$87,$E2,$23,$1A,$6E
	.dc.b	$FE,$28,$8C,$B8,$D9,$97,$FB,$BE,$30,$D7,$FD,$8D,$3C,$1F,$F5,$7F
	.dc.b	$D0,$3C,$BF,$DE,$3F,$CA,$A3,$3E,$56,$27,$2F,$ED,$E7,$53,$FE,$6A
	.dc.b	$6B,$FE,$8D,$6B,$FD,$FA,$FF,$42,$DE,$6F,$C0,$9F,$0B,$47,$CB,$5F
	.dc.b	$96,$C7,$C5,$4B,$F8,$3A,$7C,$79,$C4,$7D,$1F,$1F,$F1,$66,$38,$3F
	.dc.b	$63,$CB,$F7,$30,$FC,$31,$2B,$9E,$5F,$C4,$DF,$F4,$37,$2F,$FA,$49
	.dc.b	$F9,$3F,$39,$F9,$63,$83,$FE,$D6,$FC,$39,$F3,$9F,$C8,$D7,$45,$31
	.dc.b	$16,$7F,$BF,$CE,$9B,$3F,$C8,$63,$11,$FE,$94,$E5,$FB,$7E,$57,$E9
	.dc.b	$89,$1F,$ED,$AF,$9C,$FE,$CE,$FD,$CC,$FF,$8F,$ED,$F8,$96,$F5,$7F
	.dc.b	$83,$1B,$3F,$89,$5F,$1D,$F4,$7C,$3F,$D9,$F9,$92,$BE,$2C,$7D,$5F
	.dc.b	$8D,$FA,$DF,$35,$27,$E4,$C5,$7F,$45,$E7,$7E,$F0,$C6,$7E,$3F,$99
	.dc.b	$FE,$27,$F8,$67,$F2,$6A,$FC,$DF,$3A,$31,$1E,$B9,$DF,$91,$FB,$59
	.dc.b	$CC,$FF,$13,$6B,$FE,$B9,$6B,$ED,$F8,$6F,$C6,$11,$EF,$1D,$B0,$FC
	.dc.b	$2B,$FE,$84,$8C,$FC,$67,$98,$FC,$A7,$C4,$58,$FD,$5F,$CB,$FF,$18
	.dc.b	$30,$FD,$1E,$3F,$CD,$7C,$CC,$8F,$C6,$C5,$7E,$B4,$5C,$7E,$B4,$FF
	.dc.b	$02,$F9,$17,$CB,$5C,$A6,$1F,$A0,$9C,$B6,$3F,$C9,$3E,$17,$FA,$D5
	.dc.b	$B3,$EB,$9F,$F6,$EC,$E5,$FF,$42,$78,$AF,$9E,$FE,$62,$48,$ED,$F1
	.dc.b	$9E,$3F,$8C,$91,$E1,$7C,$DF,$21,$FC,$E3,$E6,$7F,$C8,$38,$1A,$4A
	.dc.b	$FF,$73,$63,$FE,$C9,$F8,$DF,$E7,$FE,$AC,$9F,$64,$7C,$1F,$8C,$91
	.dc.b	$EF,$7E,$27,$C7,$7F,$46,$F9,$7E,$E4,$9F,$5B,$F0,$3E,$3B,$FA,$5F
	.dc.b	$CB,$C9,$F9,$11,$F3,$7E,$27,$FA,$7F,$B1,$27,$E6,$47,$F5,$2F,$33
	.dc.b	$D6,$48,$F8,$EF,$BB,$CC,$FE,$AE,$48,$FA,$9E,$AF,$D9,$F8,$89,$1F
	.dc.b	$5D,$FB,$9F,$0B,$FD,$7B,$DA,$F8,$E4,$9F,$A6,$FC,$AF,$8C,$FE,$C6
	.dc.b	$48,$FE,$3F,$F9,$1F,$F4,$A4,$3E,$7B,$F9,$3B,$FE,$8A,$6B,$A2,$38
	.dc.b	$70,$71,$F6,$7F,$13,$F4,$3F,$EB,$C8,$FA,$ED,$7F,$12,$D7,$A7,$3A
	.dc.b	$31,$1F,$DD,$E3,$FB,$D9,$1F,$4F,$B7,$FD,$7A,$5F,$23,$FD,$83,$E8
	.dc.b	$7F,$97,$57,$F5,$46,$F8,$BF,$D8,$6B,$73,$FD,$77,$D7,$FE,$E8,$FB
	.dc.b	$5F,$F6,$4F,$C0,$7F,$A4,$9C,$C7,$C6,$72,$FF,$B8,$DF,$F0,$4F,$8F
	.dc.b	$96,$C7,$FB,$49,$B9,$B6,$7B,$1F,$DD,$9F,$F3,$3D,$7D,$4F,$2B,$06
	.dc.b	$7D,$5D,$78,$FF,$B4,$AF,$D7,$57,$E8,$9A,$FE,$D1,$AF,$AF,$1F,$B3
	.dc.b	$AF,$F8,$75,$B9,$F9,$F8,$7C,$5E,$1F,$8F,$BF,$FD,$35,$4F,$9B,$A6
	.dc.b	$71,$F9,$0F,$C5,$23,$F0,$DF,$37,$77,$B4,$91,$F9,$0F,$02,$AC,$FC
	.dc.b	$9E,$FF,$80,$CB,$FE,$17,$D7,$FC,$3B,$3F,$98,$EE,$F9,$07,$C9,$EB
	.dc.b	$B1,$F6,$FD,$EF,$49,$87,$B7,$07,$E6,$3D,$1F,$F2,$AF,$C9,$FF,$64
	.dc.b	$ED,$46,$31,$D7,$47,$FB,$A5,$FC,$FF,$A9,$FA,$1B,$E1,$5B,$C7,$F5
	.dc.b	$38,$72,$C3,$84,$62,$74,$FE,$B6,$F9,$61,$DA,$18,$74,$BE,$31,$FB
	.dc.b	$3B,$E3,$83,$0E,$33,$D6,$B7,$3E,$18,$E7,$87,$5C,$3D,$D7,$C2,$7A
	.dc.b	$5F,$0C,$3A,$9D,$A7,$58,$73,$AF,$DE,$B5,$F4,$18,$9F,$F3,$DB,$FC
	.dc.b	$2B,$E4,$FF,$9C,$7C,$8F,$F0,$48,$C7,$32,$31,$09,$DB,$B6,$BE,$0F
	.dc.b	$53,$FD,$BE,$DF,$0F,$F8,$A4,$62,$3F,$89,$B5,$FA,$6F,$FB,$9B,$4F
	.dc.b	$BC,$DD,$7F,$8C,$FC,$B7,$FA,$AA,$73,$7F,$2A,$B5,$7C,$A3,$E6,$D6
	.dc.b	$EF,$BC,$6C,$93,$F4,$0F,$C1,$CE,$DE,$46,$4F,$E5,$D5,$D9,$B3,$DB
	.dc.b	$7E,$78,$79,$27,$86,$AB,$E0,$B0,$ED,$3C,$8F,$17,$40,$7C,$7F,$FB
	.dc.b	$23,$97,$F8,$17,$C8,$5C,$FF,$8C,$4D,$BF,$E3,$13,$6F,$97,$BF,$5E
	.dc.b	$77,$FC,$A6,$31,$79,$F7,$E4,$FF,$2D,$F8,$FB,$8F,$EC,$8F,$E4,$8E
	.dc.b	$32,$FE,$28,$D9,$FE,$55,$F1,$FF,$ED,$EA,$FE,$34,$D7,$31,$FC,$C6
	.dc.b	$31,$1F,$CB,$FE,$03,$CD,$1E,$76,$CF,$F3,$FF,$A0,$F5,$1E,$BF,$82
	.dc.b	$B1,$FD,$0F,$A7,$9E,$DF,$25,$AF,$EC,$63,$FA,$04,$63,$8B,$E2,$74
	.dc.b	$FC,$56,$CF,$F4,$DF,$7B,$E4,$1F,$27,$8E,$C7,$F5,$13,$A9,$0F,$E9
	.dc.b	$87,$B3,$AD,$E5,$A3,$EB,$47,$FB,$53,$E0,$2C,$E6,$7F,$D3,$7E,$5A
	.dc.b	$7E,$D5,$7F,$90,$A7,$FD,$78,$DE,$1F,$A9,$AF,$FB,$06,$B7,$EF,$F6
	.dc.b	$C1,$EE,$78,$DB,$8F,$F8,$C9,$B8,$7F,$91,$1A,$FF,$81,$3E,$11,$EA
	.dc.b	$7E,$36,$2F,$B1,$67,$37,$FE,$D2,$9C,$BE,$6F,$7B,$D2,$7A,$BE,$02
	.dc.b	$C7,$AB,$2F,$AF,$B7,$82,$FF,$BB,$35,$FD,$13,$76,$FF,$6F,$7C,$CF
	.dc.b	$C4,$F8,$0F,$8A,$DC,$3E,$2F,$CC,$FC,$83,$E4,$FB,$F6,$3E,$4F,$BF
	.dc.b	$E0,$23,$FC,$7F,$0F,$F7,$54,$79,$3E,$67,$E8,$9F,$4B,$82,$D7,$D2
	.dc.b	$F9,$9F,$AB,$F0,$3F,$59,$B1,$F5,$BD,$CF,$69,$1E,$DB,$7F,$88,$1B
	.dc.b	$F3,$3F,$7B,$E0,$BE,$FB,$63,$EF,$F3,$7E,$0B,$E3,$FF,$06,$D8,$FC
	.dc.b	$1F,$C1,$7C,$A2,$3D,$9F,$9B,$C9,$F8,$AF,$3B,$F1,$6D,$9F,$F2,$4F
	.dc.b	$91,$F9,$58,$FE,$21,$F1,$0F,$A0,$FE,$14,$DA,$D9,$FE,$2B,$5F,$33
	.dc.b	$A9,$DF,$CC,$FE,$CA,$4D,$25,$7C,$F8,$7C,$2F,$08,$EB,$87,$8A,$30
	.dc.b	$FE,$7A,$7F,$C6,$33,$AE,$74,$3E,$47,$CF,$F0,$BE,$86,$D6,$E7,$8D
	.dc.b	$F3,$F9,$0C,$BF,$C7,$61,$F4,$1F,$56,$30,$DD,$BE,$36,$5E,$0F,$86
	.dc.b	$FE,$61,$7C,$39,$FE,$3C,$62,$F8,$65,$F1,$E3,$13,$9E,$5F,$9B,$F1
	.dc.b	$F6,$BC,$CF,$D0,$F8,$8C,$69,$F0,$5B,$A7,$8B,$3F,$F5,$C5,$FB,$67
	.dc.b	$EB,$FC,$47,$18,$F8,$1E,$FF,$DA,$3E,$DF,$C3,$D8,$F8,$8E,$1F,$8D
	.dc.b	$F1,$97,$2F,$A1,$F0,$FF,$82,$F8,$7F,$C1,$B6,$BE,$67,$C0,$FE,$18
	.dc.b	$FC,$3F,$C6,$58,$FE,$09,$F4,$1F,$55,$BC,$FF,$C2,$FB,$DC,$7F,$C3
	.dc.b	$6B,$59,$FF,$4F,$3F,$88,$F8,$1F,$D1,$9B,$87,$F8,$4A,$D3,$96,$C7
	.dc.b	$F8,$5A,$BF,$31,$5C,$E7,$2F,$58,$CF,$1F,$F5,$79,$CB,$FA,$3E,$37
	.dc.b	$CB,$B8,$7E,$FB,$71,$FC,$67,$C4,$62,$7F,$4E,$D8,$C9,$F9,$CA,$E7
	.dc.b	$3F,$AC,$F8,$E9,$3F,$9B,$C1,$E9,$E3,$4F,$E7,$B1,$89,$D4,$FE,$D3
	.dc.b	$C6,$FE,$FD,$39,$AD,$FC,$BE,$0F,$E7,$1F,$41,$E7,$8C,$E1,$FC,$DE
	.dc.b	$1C,$DF,$35,$B1,$FE,$3E,$6C,$FF,$90,$DA,$FF,$AB,$ED,$F1,$FD,$AF
	.dc.b	$C6,$DB,$F6,$7C,$7F,$98,$F3,$FC,$6D,$8F,$80,$E1,$F8,$5F,$1F,$71
	.dc.b	$F8,$3F,$15,$EE,$3D,$FF,$15,$6B,$E4,$38,$7E,$57,$C5,$5A,$F9,$3F
	.dc.b	$15,$FE,$04,$D4,$D2,$57,$F4,$18,$7C,$57,$CC,$9F,$9A,$D9,$FB,$57
	.dc.b	$F1,$2B,$F0,$CD,$8F,$F6,$74,$3E,$2B,$F8,$2C,$E7,$0F,$F1,$E4,$E6
	.dc.b	$B9,$EF,$F6,$AF,$6F,$DC,$F8,$AF,$41,$27,$CC,$FA,$FE,$27,$51,$F8
	.dc.b	$96,$E1,$FC,$6F,$E8,$3F,$C9,$D3,$FE,$7B,$E9,$46,$27,$FE,$00,$C6
	.dc.b	$3E,$1B,$FC,$66,$DA,$F8,$9F,$17,$FC,$7A,$F8,$69,$FC,$88,$46,$6F
	.dc.b	$FC,$0D,$F3,$EF,$83,$FD,$C3,$70,$FF,$44,$B7,$2F,$F4,$73,$51,$02
	.dc.b	$F8,$4E,$B5,$D9,$B5,$FE,$11,$F9,$37,$C1,$FF,$07,$6C,$7F,$A6,$1B
	.dc.b	$87,$FA,$75,$AF,$E6,$FE,$37,$FC,$3D,$1A,$D3,$DD,$18,$BF,$D7,$FC
	.dc.b	$64,$C3,$F7,$1F,$1B,$71,$FD,$BE,$BF,$8F,$24,$FE,$35,$FC,$97,$C6
	.dc.b	$FE,$E1,$FD,$CB,$63,$FA,$E7,$C8,$FF,$90,$5B,$F1,$BF,$BE,$7F,$CF
	.dc.b	$91,$9B,$FF,$20,$BF,$E4,$37,$E4,$7B,$3F,$EB,$7F,$8D,$FF,$31,$CE
	.dc.b	$67,$E6,$CE,$6B,$FC,$41,$3F,$35,$F0,$F0,$F6,$E1,$EE,$F8,$3E,$2F
	.dc.b	$C3,$F1,$96,$BC,$EF,$93,$E2,$E4,$7A,$B8,$6E,$7E,$6F,$8C,$F1,$92
	.dc.b	$7C,$2F,$9B,$E2,$BD,$BF,$19,$EE,$6C,$FC,$4F,$8C,$F5,$91,$F4,$BE
	.dc.b	$EF,$8C,$91,$EC,$6E,$1F,$ED,$83,$FD,$EE,$3C,$D7,$9F,$FE,$DF,$F9
	.dc.b	$0F,$C8,$C6,$27,$F2,$47,$07,$0F,$E5,$63,$73,$F9,$7F,$91,$7B,$AE
	.dc.b	$DC,$DF,$FB,$B6,$30,$F8,$6B,$84,$63,$0E,$B7,$CA,$F8,$F3,$7F,$BC
	.dc.b	$7E,$3B,$FD,$E2,$DC,$7E,$CC,$63,$9C,$DF,$83,$93,$F7,$9A,$FB,$2B
	.dc.b	$FD,$F9,$ED,$FE,$BF,$0F,$F0,$EE,$0F,$8E,$C7,$07,$E2,$23,$18,$7E
	.dc.b	$5A,$F2,$FF,$B5,$E3,$FB,$9B,$F8,$CF,$73,$C8,$47,$F1,$FF,$1D,$C3
	.dc.b	$25,$23,$0D,$9F,$D1,$FB,$98,$73,$F1,$D1,$FD,$39,$CC,$FF,$2C,$1F
	.dc.b	$E4,$53,$FC,$AA,$5F,$8E,$D7,$C7,$6F,$E3,$BE,$5C,$FF,$95,$8F,$F4
	.dc.b	$11,$4F,$8C,$53,$F5,$A7,$F9,$3C,$ED,$F0,$5E,$13,$C5,$F1,$DE,$3F
	.dc.b	$8D,$F2,$7C,$74,$AF,$24,$57,$AC,$78,$88,$F8,$BE,$33,$D8,$47,$D9
	.dc.b	$F1,$BE,$FF,$8D,$FA,$30,$F9,$E7,$BD,$86,$C7,$C3,$9C,$4F,$E8,$E7
	.dc.b	$33,$F9,$E3,$94,$64,$FD,$47,$CE,$FF,$68,$8F,$F5,$F3,$FD,$99,$BE
	.dc.b	$EF,$8A,$35,$1F,$E3,$97,$4F,$56,$BF,$AC,$6B,$FC,$FD,$AB,$97,$FA
	.dc.b	$1F,$C7,$FE,$3B,$D9,$7C,$EF,$6E,$DF,$F1,$B3,$7B,$3F,$8E,$E0,$BE
	.dc.b	$13,$E4,$8F,$E9,$97,$CB,$E0,$3E,$04,$7F,$A2,$A7,$4F,$F5,$29,$D3
	.dc.b	$F8,$4F,$90,$FC,$77,$89,$23,$F1,$D1,$5F,$C3,$C7,$E3,$08,$FE,$33
	.dc.b	$E2,$BD,$5F,$3A,$F8,$AA,$7E,$66,$2C,$FF,$91,$E7,$4D,$9D,$3F,$9D
	.dc.b	$F8,$2F,$D0,$7C,$8C,$8C,$1F,$A1,$7F,$3A,$47,$F3,$AF,$E8,$3E,$47
	.dc.b	$F5,$04,$8F,$D4,$7C,$8F,$EA,$FE,$3E,$47,$EA,$DD,$3F,$C7,$78,$7E
	.dc.b	$7F,$99,$FB,$12,$47,$EC,$7C,$DF,$E0,$B0,$F9,$AF,$F2,$EC,$FF,$AF
	.dc.b	$A3,$1F,$21,$A9,$CC,$F5,$3F,$04,$7F,$37,$38,$3F,$11,$FF,$43,$7C
	.dc.b	$EF,$F1,$06,$F9,$7C,$F7,$FD,$2C,$E6,$7F,$9D,$B5,$FF,$07,$35,$FF
	.dc.b	$08,$B7,$0F,$47,$C8,$79,$49,$3F,$ED,$61,$2F,$2B,$D5,$F1,$D2,$7C
	.dc.b	$EF,$B7,$E4,$3D,$DF,$21,$27,$D6,$ED,$F8,$18,$FC,$7F,$07,$F4,$1F
	.dc.b	$3A,$AA,$7C,$78,$B5,$FE,$8D,$9D,$36,$74,$F9,$FE,$89,$E3,$7F,$08
	.dc.b	$F9,$E7,$35,$3F,$85,$CE,$AB,$DD,$3A,$7F,$4E,$D7,$F8,$43,$5F,$EC
	.dc.b	$ED,$7F,$C7,$5F,$31,$FD,$21,$FF,$5E,$37,$C3,$73,$EB,$D5,$F2,$75
	.dc.b	$7E,$0F,$E5,$26,$1F,$80,$7F,$0D,$F2,$9F,$87,$F2,$7F,$11,$E3,$4A
	.dc.b	$FC,$2C,$7E,$2F,$E4,$E4,$7E,$21,$FC,$77,$CA,$7E,$3F,$E5,$24,$FE
	.dc.b	$31,$DB,$F4,$BB,$BF,$CB,$AF,$E4,$3E,$43,$F2,$C7,$E5,$FE,$4E,$C7
	.dc.b	$FA,$F7,$E7,$9F,$1D,$FC,$5D,$B1,$C4,$D4,$7E,$7F,$E4,$64,$60,$FC
	.dc.b	$EF,$C9,$71,$F8,$E9,$18,$3B,$38,$7F,$4D,$E7,$B0,$FF,$38,$46,$CF
	.dc.b	$57,$2F,$EB,$48,$FE,$AB,$E4,$AB,$EF,$F9,$2F,$D9,$18,$E8,$FA,$B1
	.dc.b	$FE,$27,$F9,$EF,$5B,$F9,$65,$F2,$7F,$BE,$6C,$7E,$F5,$DC,$F2,$F0
	.dc.b	$D1,$A9,$FF,$16,$C6,$2B,$F9,$9B,$63,$FA,$A3,$5D,$CF,$FA,$0D,$ED
	.dc.b	$F2,$7E,$13,$C5,$AF,$8F,$E4,$FC,$9F,$27,$2B,$C9,$1E,$7F,$92,$EF
	.dc.b	$87,$FA,$22,$7B,$B5,$EF,$1A,$F9,$2F,$77,$C9,$C9,$F7,$47,$C1,$F9
	.dc.b	$2F,$85,$F2,$72,$7E,$14,$7C,$5F,$93,$EE,$48,$EE,$F8,$BE,$43,$E5
	.dc.b	$7C,$9C,$9F,$95,$1F,$37,$C6,$91,$F3,$5F,$A1,$F2,$3F,$47,$E4,$7D
	.dc.b	$64,$9F,$5B,$EE,$F4,$FE,$B1,$23,$EB,$7A,$FF,$67,$E3,$FE,$D7,$C8
	.dc.b	$C9,$FB,$51,$F7,$5F,$8E,$43,$E3,$9F,$63,$E7,$7F,$C6,$2F,$F1,$A6
	.dc.b	$F5,$7E,$05,$BD,$BF,$E6,$E7,$FC,$6B,$3F,$CF,$C7,$A0,$F8,$AF,$F0
	.dc.b	$EF,$CE,$BD,$BA,$7F,$16,$8F,$E9,$BF,$3C,$E7,$FC,$4B,$70,$FC,$4B
	.dc.b	$5F,$C7,$B5,$FF,$67,$77,$B2,$7C,$7F,$F5,$86,$CF,$F5,$76,$BE,$DF
	.dc.b	$91,$F0,$F3,$78,$B9,$64,$FF,$5D,$C7,$FC,$E6,$F1,$F9,$1F,$8C,$47
	.dc.b	$F3,$B8,$7F,$92,$DF,$D1,$AF,$90,$F1,$12,$3C,$5F,$23,$E5,$DB,$CD
	.dc.b	$ED,$C9,$F3,$47,$A7,$E4,$7E,$A1,$23,$BB,$F5,$FE,$33,$DC,$48,$F7
	.dc.b	$7C,$77,$C1,$F8,$DF,$85,$F1,$B2,$7E,$14,$57,$CF,$1F,$81,$23,$EC
	.dc.b	$F8,$CF,$95,$F1,$92,$3E,$54,$57,$F0,$D1,$7A,$FF,$0D,$3F,$47,$E0
	.dc.b	$BE,$97,$C6,$4B,$FA,$51,$5F,$C7,$7C,$0F,$D7,$31,$7F,$C6,$21,$D8
	.dc.b	$F5,$BF,$C1,$5F,$2C,$B5,$4A,$D7,$66,$2B,$53,$9B,$E3,$7C,$23,$AF
	.dc.b	$B5,$FE,$C3,$8C,$61,$CA,$57,$C9,$F1,$DF,$27,$A6,$1D,$EF,$8E,$1C
	.dc.b	$E7,$27,$28,$EB,$5C,$63,$9D,$71,$7D,$55,$C5,$C9,$C4,$E7,$7E,$3C
	.dc.b	$3C,$CD,$9E,$11,$8C,$3B,$DF,$5B,$E7,$7B,$71,$5E,$F7,$FD,$CE,$D7
	.dc.b	$89,$FE,$8D,$70,$FD,$0F,$96,$71,$7F,$B0,$A3,$FD,$25,$18,$F9,$6F
	.dc.b	$F3,$AB,$63,$33,$B7,$BF,$B4,$F1,$7F,$CF,$78,$23,$13,$E1,$6B,$C2
	.dc.b	$77,$1A,$C7,$E2,$37,$3F,$B7,$EC,$7F,$62,$9C,$D7,$C0,$AD,$5F,$E2
	.dc.b	$5A,$FE,$42,$B8,$5F,$E1,$27,$38,$79,$6F,$91,$F2,$AB,$77,$CE,$32
	.dc.b	$FD,$93,$F3,$75,$CA,$77,$18,$9D,$E1,$E6,$8E,$16,$AD,$BE,$18,$FC
	.dc.b	$26,$8F,$9E,$7C,$2F,$F8,$03,$87,$FB,$C4,$63,$C8,$FF,$3C,$2F,$83
	.dc.b	$D4,$7E,$AA,$F9,$5E,$30,$47,$5A,$C3,$AA,$E7,$18,$BE,$D5,$FE,$62
	.dc.b	$5F,$05,$93,$5F,$05,$6F,$C6,$DA,$E0,$DC,$BF,$A8,$7C,$87,$F0,$39
	.dc.b	$FE,$A0,$D7,$CB,$E1,$A3,$19,$FF,$43,$75,$2F,$9D,$78,$18,$6E,$5F
	.dc.b	$EC,$A6,$CF,$FB,$AD,$AF,$F4,$EC,$3E,$1C,$FF,$4B,$C3,$E0,$74,$FC
	.dc.b	$49,$CD,$7C,$1E,$9F,$1E,$C6,$93,$0F,$F1,$2C,$3D,$FF,$95,$CD,$F2
	.dc.b	$DB,$3F,$EA,$FF,$82,$7C,$8F,$D2,$6C,$7F,$AC,$E3,$FC,$B2,$ED,$D7
	.dc.b	$C8,$FC,$49,$FF,$6C,$4F,$F8,$D1,$B9,$7F,$8E,$9A,$FF,$6C,$A7,$C0
	.dc.b	$F7,$F4,$7F,$03,$18,$BF,$24,$78,$FE,$3E,$47,$8F,$E0,$2E,$1F,$E0
	.dc.b	$5F,$5D,$EE,$FD,$46,$D7,$FB,$E9,$B3,$FC,$D4,$FE,$7A,$E1,$F2,$7B
	.dc.b	$DF,$07,$E4,$3C,$6D,$AF,$85,$EB,$FE,$0B,$4F,$F0,$64,$7F,$94,$9B
	.dc.b	$7F,$E5,$86,$BF,$E0,$AD,$BF,$D6,$D3,$F1,$7E,$3B,$E4,$11,$E5,$F0
	.dc.b	$1D,$48,$E0,$ED,$EF,$FD,$5F,$90,$F8,$ED,$8F,$B5,$F1,$BF,$4C,$8F
	.dc.b	$D3,$7D,$1F,$03,$F7,$7C,$4F,$9E,$D9,$FB,$DF,$03,$F8,$1F,$8D,$FA
	.dc.b	$6D,$8F,$C3,$78,$DF,$7C,$8F,$DF,$7E,$27,$C0,$FE,$27,$C7,$FB,$6D
	.dc.b	$9F,$C5,$79,$7F,$8D,$F8,$DF,$BE,$D8,$FC,$97,$93,$F8,$B2,$3F,$8B
	.dc.b	$7F,$63,$F3,$6F,$98,$FC,$B7,$C7,$7F,$08,$8C,$57,$E6,$E3,$F3,$5E
	.dc.b	$2C,$8F,$EF,$7E,$67,$FA,$3E,$9F,$07,$9E,$94,$FF,$A7,$3D,$47,$E4
	.dc.b	$8F,$6B,$F8,$24,$61,$B1,$9C,$7F,$19,$6A,$D4,$7E,$47,$E1,$3F,$28
	.dc.b	$47,$07,$FA,$6B,$2F,$F2,$6B,$FB,$3F,$37,$D9,$F0,$BF,$B3,$7D,$DF
	.dc.b	$03,$89,$F5,$9F,$CB,$69,$19,$C7,$D7,$39,$F8,$2F,$DE,$79,$DF,$1D
	.dc.b	$B1,$CA,$7F,$99,$DF,$F9,$36,$31,$E8,$A3,$13,$D0,$DF,$C1,$73,$AD
	.dc.b	$CF,$F9,$D5,$BF,$05,$F5,$7D,$5F,$AC,$D8,$F8,$11,$F7,$08,$FC,$3F
	.dc.b	$82,$D6,$1D,$F1,$F1,$46,$38,$FE,$D7,$C1,$7F,$17,$F6,$BF,$04,$DC
	.dc.b	$3D,$9F,$1E,$F9,$FF,$F0,$14,$7D,$68,$C7,$6F,$F3,$F6,$C6,$67,$79
	.dc.b	$FF,$B4,$B6,$7C,$1F,$F2,$5F,$63,$F2,$0D,$BF,$83,$F0,$9F,$93,$3F
	.dc.b	$29,$F0,$16,$3F,$97,$1F,$CB,$4F,$EC,$58,$7F,$49,$E6,$C4,$7D,$38
	.dc.b	$7A,$92,$3F,$8D,$3F,$C7,$48,$FF,$1D,$F1,$3F,$90,$11,$C1,$F8,$DF
	.dc.b	$84,$FE,$DB,$F9,$2F,$F3,$AB,$F9,$9A,$A7,$3E,$97,$EA,$D9,$7F,$73
	.dc.b	$E0,$2D,$F0,$E7,$E7,$F2,$1F,$B1,$6D,$7D,$0F,$43,$F2,$6F,$CF,$6D
	.dc.b	$7F,$31,$E4,$46,$23,$FC,$D1,$18,$DF,$FB,$7E,$AF,$90,$F4,$9F,$E0
	.dc.b	$6F,$C9,$BF,$A6,$F5,$32,$FF,$07,$56,$DF,$95,$F4,$0E,$5F,$EE,$ED
	.dc.b	$8F,$DB,$E6,$8C,$3F,$BA,$6B,$FD,$1F,$F0,$F7,$CF,$FF,$80,$1B,$5C
	.dc.b	$9B,$3F,$C0,$9A,$FF,$B2,$B2,$FF,$6B,$4F,$EB,$31,$FE,$BD,$3F,$9D
	.dc.b	$9C,$BF,$9F,$8C,$E5,$ED,$86,$BF,$23,$C0,$FC,$C6,$FF,$AF,$9F,$98
	.dc.b	$D7,$E5,$B5,$FF,$19,$60,$F7,$7F,$07,$F2,$9F,$2D,$B9,$FD,$9F,$8A
	.dc.b	$EA,$47,$ED,$3F,$53,$E0,$64,$7F,$20,$E4,$6C,$9C,$E5,$E4,$F3,$7C
	.dc.b	$E4,$BF,$C2,$3F,$E2,$CC,$FF,$07,$1F,$C4,$1B,$3F,$C3,$F6,$FC,$3F
	.dc.b	$C6,$7E,$07,$6F,$F2,$54,$62,$F8,$46,$04,$EE,$3E,$87,$C9,$FE,$30
	.dc.b	$8F,$A7,$34,$7C,$CE,$A9,$3F,$99,$6E,$3F,$CF,$BE,$85,$D5,$FE,$42
	.dc.b	$6D,$7E,$31,$B3,$B9,$D7,$4C,$AD,$36,$7F,$C2,$9C,$0D,$BF,$CD,$55
	.dc.b	$F0,$FD,$8F,$F3,$47,$33,$F1,$E7,$C3,$7E,$8E,$73,$18,$AE,$07,$F3
	.dc.b	$DA,$E1,$5B,$8F,$E1,$DF,$44,$F7,$3F,$CB,$6D,$8F,$D2,$B7,$3F,$EB
	.dc.b	$AD,$7F,$C5,$4B,$E5,$78,$92,$78,$BF,$BA,$F8,$7F,$E2,$27,$EE,$7D
	.dc.b	$DB,$3F,$7F,$DE,$FC,$11,$1F,$C6,$6D,$FE,$7E,$6F,$BD,$F8,$72,$38
	.dc.b	$3E,$D7,$C3,$FF,$1E,$3F,$83,$7A,$F6,$3F,$19,$EE,$FE,$3C,$8F,$67
	.dc.b	$F2,$5E,$9F,$F0,$DF,$63,$F8,$73,$67,$F2,$86,$0A,$7E,$5D,$FE,$5B
	.dc.b	$F1,$12,$3F,$B4,$FE,$4F,$F8,$2E,$4F,$DD,$73,$FE,$9A,$BF,$99,$FC
	.dc.b	$3F,$F1,$E2,$5F,$E4,$DF,$F4,$7D,$BC,$6F,$E7,$3F,$2F,$FE,$A6,$8C
	.dc.b	$57,$F0,$B8,$FE,$7B,$F0,$FF,$CA,$09,$1F,$9D,$7F,$95,$FB,$F2,$3F
	.dc.b	$40,$DE,$8F,$E4,$63,$E4,$BF,$D7,$1E,$8F,$F8,$71,$B1,$FE,$69,$6A
	.dc.b	$23,$8F,$93,$7B,$9F,$E2,$C9,$FF,$5F,$D6,$D4,$EB,$83,$85,$6F,$83
	.dc.b	$FD,$61,$5D,$6B,$4F,$23,$FD,$93,$E3,$3E,$5B,$F9,$DD,$7F,$12,$6C
	.dc.b	$66,$7F,$7D,$18,$F8,$99,$1F,$C0,$5A,$FE,$CB,$81,$C5,$E7,$CF,$EC
	.dc.b	$E1,$FA,$7A,$FE,$75,$E0,$67,$83,$FD,$BD,$87,$09,$FE,$7D,$DF,$E2
	.dc.b	$23,$FC,$3D,$F2,$FB,$3C,$63,$D6,$DD,$7F,$95,$B6,$9C,$57,$5C,$65
	.dc.b	$30,$FC,$7C,$57,$FC,$51,$5F,$B5,$7F,$27,$8E,$74,$FC,$9C,$FF,$83
	.dc.b	$A7,$F9,$D3,$5F,$85,$8F,$CC,$8C,$47,$D9,$CB,$F9,$FF,$B3,$E7,$D4
	.dc.b	$D3,$D0,$7A,$67,$91,$0E,$E7,$D8,$F5,$3F,$B6,$BF,$D2,$DA,$FF,$24
	.dc.b	$6B,$E1,$6B,$FE,$00,$F7,$BC,$44,$7F,$CB,$31,$FE,$87,$7F,$C5,$9D
	.dc.b	$9F,$D1,$1F,$F0,$0B,$5F,$F3,$CB,$5F,$F6,$33,$5F,$CC,$2F,$6F,$F4
	.dc.b	$ED,$8E,$71,$83,$F5,$7F,$91,$7E,$6A,$3F,$CB,$55,$FE,$90,$7C,$BF
	.dc.b	$0F,$F8,$15,$3B,$DB,$F8,$3D,$6E,$78,$D7,$E1,$63,$13,$FE,$05,$9E
	.dc.b	$3D,$1C,$FE,$8D,$E5,$FE,$F1,$B3,$FE,$20,$6B,$FE,$98,$6B,$FE,$E5
	.dc.b	$5E,$37,$12,$47,$17,$FC,$53,$E9,$FF,$8E,$B8,$6F,$A6,$69,$2B,$FB
	.dc.b	$71,$FB,$42,$18,$3F,$CE,$9F,$42,$F9,$1F,$77,$BD,$27,$FC,$F7,$39
	.dc.b	$7E,$EC,$65,$F9,$AF,$EA,$F0,$4E,$E7,$FC,$14,$FF,$88,$5E,$FE,$D7
	.dc.b	$AA,$31,$1F,$E5,$56,$BF,$9F,$FC,$AB,$F1,$EF,$B5,$B8,$9A,$6A,$FC
	.dc.b	$39,$75,$25,$7B,$9F,$D0,$7B,$9F,$BF,$AF,$CB,$4F,$FA,$7D,$BC,$1F
	.dc.b	$82,$9F,$66,$3F,$5A,$BF,$98,$61,$EE,$DF,$D4,$48,$FD,$57,$C2,$F5
	.dc.b	$E3,$91,$FA,$48,$FC,$87,$D0,$3D,$EF,$E8,$2D,$9F,$F3,$2B,$5F,$F6
	.dc.b	$EB,$5F,$E7,$43,$E8,$BF,$92,$3F,$EE,$26,$BA,$6B,$FD,$05,$AF,$CB
	.dc.b	$F8,$BF,$84,$47,$F7,$35,$F0,$3D,$2F,$AC,$48,$FC,$3B,$FC,$27,$DF
	.dc.b	$F9,$5F,$37,$27,$F8,$0C,$7E,$9F,$E1,$24,$7F,$04,$7F,$55,$F0,$F8
	.dc.b	$8F,$A1,$CF,$F4,$CC,$71,$FE,$B7,$5F,$F6,$0C,$FC,$4A,$F8,$E4,$7E
	.dc.b	$3D,$FC,$A2,$38,$3F,$CC,$DC,$8F,$8C,$FC,$01,$23,$FD,$81,$C2,$F5
	.dc.b	$3F,$7D,$DB,$27,$F8,$DB,$FC,$07,$C0,$FE,$05,$F0,$3F,$89,$F8,$C9
	.dc.b	$5F,$C8,$23,$F0,$FE,$6F,$E7,$49,$1F,$AE,$7F,$B3,$FE,$52,$F9,$6F
	.dc.b	$C9,$76,$CA,$FF,$80,$33,$96,$F8,$3F,$D5,$63,$F4,$FC,$7F,$E1,$C6
	.dc.b	$E1,$FE,$03,$F0,$11,$FB,$19,$D7,$83,$E0,$26,$1F,$E8,$99,$FD,$7B
	.dc.b	$C5,$FF,$2E,$BE,$2F,$80,$F0,$1E,$19,$F1,$7C,$AC,$9F,$3B,$E5,$F9
	.dc.b	$6C,$47,$EA,$FE,$43,$F5,$C6,$31,$F4,$7C,$AF,$AF,$E5,$A5,$7C,$08
	.dc.b	$AF,$EC,$FC,$3F,$DB,$18,$AF,$C2,$6B,$FA,$82,$4F,$EA,$1F,$F6,$6F
	.dc.b	$C9,$BF,$2A,$61,$86,$D7,$EA,$84,$EE,$74,$7F,$87,$3A,$FF,$65,$87
	.dc.b	$F9,$29,$BC,$3F,$3D,$B7,$FC,$C7,$87,$F9,$ED,$7D,$5F,$85,$FE,$1D
	.dc.b	$EE,$4A,$FB,$B1,$FC,$47,$C2,$EC,$48,$FC,$BB,$F8,$6F,$13,$F8,$B7
	.dc.b	$83,$27,$F0,$B1,$F8,$3F,$83,$FE,$35,$E1,$49,$FC,$4C,$7F,$1D,$F8
	.dc.b	$DF,$51,$23,$F4,$6F,$E4,$BE,$13,$F9,$17,$81,$27,$F2,$31,$F8,$FE
	.dc.b	$0F,$E4,$DD,$F2,$7F,$2B,$1F,$CA,$7C,$1F,$8C,$48,$FD,$7B,$FE,$44
	.dc.b	$F9,$E7,$C9,$7F,$2F,$E8,$95,$FE,$59,$E3,$FF,$3A,$D6,$A3,$18,$7E
	.dc.b	$6E,$1F,$19,$CC,$93,$FE,$E1,$8C,$BF,$EE,$3F,$9B,$FD,$43,$FE,$BC
	.dc.b	$6B,$FE,$29,$6B,$FA,$76,$BF,$DA,$63,$FD,$00,$F5,$F8,$7F,$51,$1F
	.dc.b	$07,$C5,$FF,$91,$E9,$F3,$74,$FF,$25,$5A,$7F,$C1,$DF,$3D,$FA,$CB
	.dc.b	$FC,$83,$5F,$E7,$6D,$7F,$CE,$CD,$7F,$BE,$5F,$EE,$F2,$F4,$78,$DE
	.dc.b	$13,$BF,$C5,$F8,$3E,$2F,$C9,$E4,$4C,$3E,$59,$ED,$8F,$47,$91,$FE
	.dc.b	$78,$C7,$F8,$0C,$F6,$8C,$19,$F1,$FD,$C4,$8E,$7C,$4D,$BF,$CF,$D0
	.dc.b	$E2,$FB,$7F,$0B,$F0,$BE,$12,$4F,$0A,$FE,$AF,$5F,$80,$F8,$4F,$C3
	.dc.b	$12,$7E,$E3,$E5,$F2,$FE,$71,$23,$97,$C1,$7F,$8E,$7E,$7B,$FC,$6D
	.dc.b	$1F,$D9,$1A,$FF,$A7,$DA,$FF,$A9,$DB,$9B,$67,$BB,$F0,$C9,$5F,$8B
	.dc.b	$7F,$D3,$BF,$4B,$FE,$04,$E5,$FE,$5C,$FF,$99,$B4,$FF,$66,$B6,$33
	.dc.b	$F3,$BF,$E8,$B8,$C5,$FF,$8D,$5B,$89,$A7,$58,$D7,$B3,$C7,$DB,$91
	.dc.b	$F4,$25,$38,$8F,$A1,$87,$92,$B1,$8F,$92,$BF,$23,$E2,$7F,$91,$6F
	.dc.b	$FD,$9D,$E1,$FE,$3A,$31,$F1,$BF,$E6,$B9,$76,$27,$35,$F9,$4F,$85
	.dc.b	$FF,$65,$5E,$EB,$5A,$FF,$B4,$AB,$A3,$70,$FF,$6B,$4E,$9F,$09,$8B
	.dc.b	$E7,$5C,$23,$27,$8B,$0F,$0B,$C9,$AF,$9A,$78,$B9,$BF,$03,$C5,$B7
	.dc.b	$E1,$C3,$D1,$5C,$2B,$47,$82,$75,$87,$8E,$B7,$3C,$B0,$F9,$4F,$EE
	.dc.b	$1A,$FA,$4C,$61,$FD,$35,$FF,$77,$FB,$BF,$D3,$9F,$EA,$1F,$1B,$FE
	.dc.b	$41,$F4,$B4,$D8,$CC,$ED,$FE,$0B,$E4,$3E,$1F,$A6,$88,$C4,$FF,$0C
	.dc.b	$6B,$C2,$77,$1A,$CF,$EC,$37,$4F,$E1,$FE,$DB,$8B,$E3,$CE,$70,$FB
	.dc.b	$2D,$7F,$0A,$DC,$7F,$C5,$13,$C3,$19,$5F,$2E,$B5,$7D,$63,$13,$AC
	.dc.b	$7C,$6F,$D6,$9D,$46,$27,$64,$7F,$21,$1A,$C7,$F9,$3D,$F9,$F5,$FD
	.dc.b	$E7,$C0,$F7,$E2,$F0,$47,$EF,$FD,$8C,$69,$FD,$06,$AE,$7F,$CB,$A6
	.dc.b	$D7,$2A,$FE,$3F,$D1,$FE,$BC,$6F,$AD,$A8,$C5,$FE,$C8,$FE,$6F,$E8
	.dc.b	$67,$E1,$34,$DB,$E6,$E1,$EB,$EA,$5A,$E4,$3B,$5E,$D7,$FB,$3E,$3F
	.dc.b	$8E,$35,$FE,$38,$D7,$83,$9C,$FF,$8F,$7C,$EF,$BA,$73,$87,$B2,$31
	.dc.b	$7D,$A3,$DF,$D3,$ED,$3D,$DD,$56,$3E,$1F,$17,$5D,$BF,$25,$1F,$DC
	.dc.b	$63,$06,$7E,$6F,$E3,$1F,$1F,$96,$CF,$F4,$AF,$7B,$FC,$0A,$DF,$4B
	.dc.b	$E6,$3F,$4A,$33,$AF,$CD,$DB,$FC,$D1,$F2,$BE,$D9,$34,$95,$FE,$3D
	.dc.b	$9F,$F3,$77,$CF,$73,$3A,$7A,$96,$7C,$5F,$33,$DC,$F0,$7A,$B6,$3F
	.dc.b	$B4,$E4,$F2,$9F,$23,$F3,$DB,$3F,$BF,$6C,$FF,$93,$5A,$8F,$F3,$B6
	.dc.b	$9F,$E0,$59,$F8,$5E,$AF,$BF,$E6,$7E,$03,$F0,$7D,$5F,$85,$F2,$B6
	.dc.b	$BE,$43,$F9,$8F,$8B,$7D,$2B,$97,$F0,$91,$86,$C6,$6B,$FC,$C1,$39
	.dc.b	$7E,$36,$4F,$5F,$E8,$6D,$F4,$5B,$3F,$5F,$E5,$3E,$A7,$CB,$7D,$56
	.dc.b	$C7,$F8,$3B,$07,$CF,$7D,$48,$EE,$DE,$0F,$F6,$CC,$EB,$9F,$CF,$39
	.dc.b	$9F,$8C,$39,$FF,$D4,$DF,$2D,$F8,$33,$F0,$9F,$2F,$71,$FC,$57,$B3
	.dc.b	$F8,$7F,$6B,$F1,$0D,$8F,$26,$2E,$8E,$E4,$8F,$2B,$F8,$BF,$63,$F2
	.dc.b	$07,$E4,$7A,$6C,$FE,$5F,$4F,$CA,$79,$FF,$95,$6C,$76,$E8,$F5,$91
	.dc.b	$F8,$2F,$EC,$FE,$79,$EC,$7F,$0E,$6C,$FF,$8B,$17,$CA,$E6,$75,$F1
	.dc.b	$F2,$BF,$C7,$7E,$BB,$E7,$DF,$13,$9C,$FF,$84,$EA,$E8,$FC,$B1,$30
	.dc.b	$FC,$B3,$FC,$27,$07,$D3,$3A,$BF,$83,$5F,$1B,$EF,$39,$7F,$47,$19
	.dc.b	$C3,$F6,$9B,$3E,$53,$D9,$EE,$FB,$5B,$7E,$D7,$0F,$BF,$DD,$B7,$CB
	.dc.b	$E4,$FE,$21,$F1,$7E,$4E,$D7,$C5,$6C,$F8,$1B,$ED,$7F,$3E,$CB,$F8
	.dc.b	$D4,$F9,$3E,$4F,$E7,$7A,$BF,$3D,$FA,$1F,$27,$8C,$7E,$83,$6B,$D7
	.dc.b	$F0,$1A,$9F,$A8,$D8,$FE,$7D,$E2,$7E,$7B,$AF,$ED,$7C,$87,$EC,$FB
	.dc.b	$FF,$BB,$61,$F6,$E3,$EF,$74,$66,$AC,$FE,$51,$FC,$17,$C7,$FE,$0F
	.dc.b	$E4,$2C,$FE,$0D,$FC,$37,$C0,$C8,$FC,$37,$C8,$5D,$3F,$A3,$FB,$7F
	.dc.b	$46,$7F,$8B,$B6,$3F,$07,$D3,$FC,$FB,$A3,$EE,$F8,$7F,$C6,$49,$7F
	.dc.b	$C6,$5F,$E9,$56,$E1,$DC,$F4,$6E,$1F,$80,$F9,$0F,$CD,$1F,$9B,$F9
	.dc.b	$0B,$5F,$9B,$6C,$FE,$19,$BD,$1F,$E0,$3E,$4F,$F3,$8E,$7F,$A3,$F2
	.dc.b	$FF,$8E,$7C,$4F,$A6,$BE,$E4,$7F,$74,$F9,$B7,$A3,$AB,$C3,$73,$FF
	.dc.b	$4A,$CF,$FA,$D1,$BF,$23,$FD,$87,$E8,$1C,$DC,$E3,$FB,$27,$D0,$7E
	.dc.b	$C2,$BF,$C7,$0D,$DF,$F8,$99,$FC,$E5,$C3,$86,$BF,$CF,$7C,$6F,$D5
	.dc.b	$F9,$DD,$3C,$CF,$F0,$6C,$62,$FF,$A5,$DF,$5D,$FE,$17,$03,$C0,$F7
	.dc.b	$7C,$9F,$C2,$6D,$FB,$3E,$4F,$0D,$3F,$98,$57,$16,$F3,$FC,$7F,$93
	.dc.b	$F9,$6D,$8F,$8B,$F2,$78,$69,$C0,$E7,$E0,$FF,$4A,$BF,$E1,$B1,$FD
	.dc.b	$21,$BC,$1F,$CE,$9A,$FF,$38,$E7,$72,$BE,$4B,$ED,$E1,$F6,$5B,$87
	.dc.b	$A3,$BF,$EE,$63,$F8,$06,$C6,$49,$8F,$F8,$07,$07,$C6,$EE,$7E,$0F
	.dc.b	$6F,$9C,$8F,$9D,$FF,$11,$4B,$E3,$7F,$1B,$1F,$8B,$6C,$FF,$7E,$FA
	.dc.b	$07,$C1,$FF,$8A,$DB,$1F,$E7,$B6,$E3,$F8,$40,$E1,$FC,$3A,$F8,$4F
	.dc.b	$F1,$83,$C3,$9B,$FC,$C5,$3C,$27,$37,$FE,$52,$97,$D0,$39,$7F,$41
	.dc.b	$F2,$19,$8C,$65,$F9,$E8,$C4,$C8,$FE,$E5,$F9,$5B,$CD,$C6,$7F,$E4
	.dc.b	$CF,$A1,$77,$FF,$91,$DB,$7F,$E9,$76,$CF,$B5,$AF,$E8,$9F,$F1,$14
	.dc.b	$7F,$89,$1D,$E7,$EC,$73,$3D,$EB,$F7,$1F,$21,$2F,$F6,$BF,$25,$6B
	.dc.b	$EE,$7C,$97,$D6,$24,$FF,$8E,$79,$11,$F6,$BE,$4B,$F8,$13,$FC,$0D
	.dc.b	$B3,$FC,$03,$E4,$BF,$83,$1F,$C1,$FE,$4A,$C7,$F0,$4F,$91,$FC,$19
	.dc.b	$1E,$CF,$E1,$BE,$47,$F8,$71,$FC,$3E,$29,$FC,$30,$C1,$FC,$3B,$E3
	.dc.b	$2E,$7F,$E5,$5F,$A1,$FF,$0F,$70,$FF,$A4,$1B,$5F,$E4,$1F,$05,$1F
	.dc.b	$E5,$C9,$D7,$C5,$7F,$A4,$23,$F1,$57,$B6,$BF,$E2,$36,$D7,$2E,$9F
	.dc.b	$F3,$F4,$7F,$87,$70,$FF,$27,$61,$E1,$F8,$7E,$9D,$7D,$5E,$DF,$23
	.dc.b	$C0,$95,$FE,$67,$B4,$74,$F9,$19,$39,$6E,$7F,$ED,$D1,$F2,$9E,$DF
	.dc.b	$57,$F4,$8D,$8F,$E5,$2D,$7F,$C7,$AD,$F9,$67,$BA,$D7,$0D,$AF,$F2
	.dc.b	$15,$FF,$91,$2B,$75,$A3,$B6,$FF,$33,$E4,$FF,$DE,$0D,$BF,$64,$63
	.dc.b	$53,$7E,$4E,$4F,$C1,$6B,$FD,$02,$BE,$BF,$C9,$7F,$95,$30,$FB,$8D
	.dc.b	$DB,$EB,$36,$BE,$B7,$C7,$5B,$FB,$FF,$25,$E1,$24,$FC,$F7,$C9,$F1
	.dc.b	$9B,$D3,$FC,$71,$3F,$6B,$1F,$C1,$4E,$7E,$33,$5F,$25,$2B,$47,$E1
	.dc.b	$5E,$5F,$25,$CF,$E4,$A4,$EB,$3F,$C8,$C7,$18,$FC,$C7,$8B,$F1,$48
	.dc.b	$FE,$0D,$F9,$1F,$07,$E5,$24,$73,$D3,$FD,$48,$DF,$87,$F5,$30,$C1
	.dc.b	$B3,$E9,$10,$FC,$81,$F6,$3D,$DF,$F3,$2B,$E7,$6B,$FE,$C1,$6F,$17
	.dc.b	$FB,$F1,$FF,$33,$C7,$FA,$DB,$C2,$FE,$FA,$DF,$4F,$EA,$28,$46,$1D
	.dc.b	$C7,$FB,$C1,$CB,$FC,$F4,$E1,$1F,$9F,$8C,$C7,$F4,$16,$BF,$E0,$6F
	.dc.b	$A0,$74,$F8,$1B,$1F,$EC,$A6,$BF,$E4,$AF,$CB,$3C,$38,$FF,$44,$9C
	.dc.b	$BE,$DF,$93,$FC,$56,$47,$A6,$1E,$CF,$A1,$FE,$B9,$8F,$FB,$59,$7A
	.dc.b	$59,$35,$EA,$D9,$F6,$4E,$9B,$3F,$BB,$8C,$47,$E8,$70,$7C,$97,$E7
	.dc.b	$7E,$21,$ED,$7F,$AD,$6B,$FC,$23,$CD,$E6,$AD,$4E,$67,$FC,$0D,$3F
	.dc.b	$2E,$B8,$B6,$75,$3F,$EC,$09,$7D,$0B,$E0,$BC,$0D,$9F,$F6,$73,$5F
	.dc.b	$C2,$7C,$9E,$A7,$E3,$FC,$54,$9D,$C7,$CA,$F9,$3E,$7F,$09,$27,$F1
	.dc.b	$D1,$5F,$F6,$AE,$09,$F4,$36,$3F,$DD,$2D,$7E,$37,$C9,$F9,$B3,$91
	.dc.b	$F9,$98,$AF,$E6,$7E,$1B,$ED,$98,$E5,$FC,$D7,$C9,$7B,$B8,$E5,$7E
	.dc.b	$86,$2B,$FA,$1F,$77,$F0,$66,$29,$E2,$FF,$94,$4F,$D9,$CF,$F2,$B7
	.dc.b	$FD,$D7,$F4,$1F,$E7,$B8,$FE,$0E,$1C,$3B,$1F,$15,$D3,$E4,$3F,$53
	.dc.b	$3A,$9F,$F3,$1B,$5E,$BF,$15,$FD,$0E,$31,$1F,$EF,$9E,$8E,$B8,$FE
	.dc.b	$DE,$BF,$36,$7E,$D3,$E8,$5E,$37,$E7,$BD,$1F,$BF,$EF,$4C,$7F,$81
	.dc.b	$45,$7F,$81,$7C,$57,$E9,$0C,$63,$FC,$0F,$E4,$FF,$4F,$F1,$F2,$BF
	.dc.b	$85,$45,$7F,$85,$45,$9E,$11,$F1,$3E,$91,$D5,$F0,$2E,$4E,$62,$BF
	.dc.b	$E4,$1D,$D3,$F0,$49,$1F,$C2,$7C,$47,$BF,$FC,$49,$B1,$F4,$9A,$FD
	.dc.b	$13,$EA,$65,$FC,$03,$E5,$24,$7F,$23,$7F,$81,$F9,$5F,$C1,$3C,$7F
	.dc.b	$E0,$BF,$19,$2B,$F9,$44,$7F,$08,$F1,$7F,$53,$D3,$27,$F9,$64,$57
	.dc.b	$F0,$0B,$E8,$3F,$D5,$B1,$FD,$1D,$AF,$FB,$83,$C1,$FF,$2C,$F1,$7E
	.dc.b	$42,$15,$C7,$E4,$BF,$B0,$5F,$ED,$6B,$53,$D7,$C4,$CD,$69,$DF,$89
	.dc.b	$30,$F0,$45,$7C,$B8,$7F,$98,$DF,$27,$C6,$67,$1E,$B2,$F9,$1F,$0F
	.dc.b	$C8,$CB,$F5,$45,$79,$C5,$AE,$33,$E8,$F3,$65,$7C,$17,$D7,$E4,$62
	.dc.b	$3C,$7F,$23,$EE,$31,$7F,$0F,$C4,$7D,$67,$F5,$6E,$2F,$D8,$F0,$FE
	.dc.b	$A8,$FF,$03,$4E,$BE,$6F,$18,$FF,$4F,$EC,$FE,$46,$F0,$31,$5F,$E1
	.dc.b	$CD,$FF,$9A,$E5,$F0,$3C,$9F,$9E,$74,$FA,$3F,$E8,$77,$F2,$3E,$43
	.dc.b	$E9,$91,$FB,$9F,$21,$FC,$EF,$E9,$DF,$07,$FC,$D5,$B3,$FE,$9D,$F4
	.dc.b	$D1,$88,$FF,$71,$C6,$27,$C9,$F2,$1F,$80,$F2,$65,$7E,$26,$3F,$07
	.dc.b	$F2,$12,$3F,$18,$FE,$1B,$DA,$F1,$7C,$17,$84,$93,$E1,$CB,$FC,$46
	.dc.b	$FF,$8C,$F8,$3F,$4F,$C1,$F8,$0F,$90,$FC,$79,$F9,$0F,$23,$F2,$3C
	.dc.b	$F2,$FF,$34,$FE,$53,$E4,$3E,$07,$A7,$EF,$24,$F7,$7F,$C9,$3F,$46
	.dc.b	$F9,$1F,$F1,$D2,$FA,$0A,$62,$33,$F2,$3A,$F9,$29,$87,$F8,$BE,$FF
	.dc.b	$D8,$4F,$F8,$DB,$9B,$E5,$BF,$3C,$87,$BC,$F0,$FC,$8F,$83,$E4,$A4
	.dc.b	$73,$8F,$1F,$C7,$C8,$EC,$F9,$BE,$4B,$FC,$3F,$F4,$5F,$D9,$64,$40
	.dc.b	$E5,$C3,$5D,$B7,$C0,$64,$F9,$0F,$86,$4A,$F7,$76,$FC,$6F,$90,$91
	.dc.b	$F0,$5F,$93,$F2,$3F,$2B,$E4,$7E,$69,$27,$FD,$D5,$8F,$68,$FA,$3F
	.dc.b	$21,$F4,$09,$3F,$2F,$B7,$E9,$FB,$F2,$3E,$73,$F5,$BE,$47,$EB,$FC
	.dc.b	$8F,$DA,$24,$F6,$DB,$FD,$C2,$DF,$91,$FB,$A4,$7E,$C7,$57,$E0,$3E
	.dc.b	$43,$FC,$93,$F3,$9F,$E3,$19,$FE,$D2,$D7,$F9,$4D,$A7,$7E,$FF,$FA
	.dc.b	$96,$B5,$5F,$44,$38,$7E,$A3,$5F,$E5,$7F,$47,$FC,$25,$BE,$7F,$FA
	.dc.b	$2E,$31,$C3,$F5,$27,$13,$9F,$88,$91,$FE,$7F,$AD,$8F,$8E,$E4,$48
	.dc.b	$D6,$35,$FE,$87,$3F,$E4,$07,$97,$C7,$E7,$C7,$FE,$81,$1F,$8A,$F2
	.dc.b	$FF,$C8,$5F,$2A,$F8,$3F,$C5,$C6,$1B,$AF,$E4,$A7,$55,$F8,$EF,$37
	.dc.b	$FD,$87,$5A,$8F,$F6,$5B,$67,$71,$DE,$B9,$4F,$E6,$2B,$8C,$FE,$62
	.dc.b	$FC,$0D,$AD,$B4,$EA,$78,$4C,$9D,$4C,$96,$FC,$3D,$CA,$EB,$3F,$AA
	.dc.b	$BF,$0B,$BC,$3D,$04,$7B,$1D,$2B,$96,$1F,$E8,$D6,$C7,$E1,$4F,$11
	.dc.b	$F8,$77,$2F,$E2,$7E,$55,$F1,$9F,$E3,$68,$E1,$18,$F5,$FF,$D3,$0D
	.dc.b	$8C,$CE,$DE,$9E,$77,$E5,$DF,$05,$A3,$11,$AE,$5F,$34,$62,$3F,$83
	.dc.b	$BC,$75,$F8,$AD,$CF,$F8,$5F,$9D,$FE,$05,$9C,$DF,$CE,$C3,$B3,$6F
	.dc.b	$86,$1F,$C4,$2F,$86,$1D,$AB,$44,$9F,$95,$3A,$7F,$8B,$D7,$27,$F8
	.dc.b	$EB,$6F,$BC,$70,$F3,$B9,$63,$E4,$8F,$A9,$5B,$BE,$91,$CB,$0F,$C4
	.dc.b	$EA,$FC,$B7,$F3,$D8,$FF,$BA,$DB,$5F,$E2,$A7,$FD,$83,$19,$CF,$FC
	.dc.b	$4D,$C3,$FE,$22,$9C,$D6,$E7,$35,$FC,$DA,$FB,$7C,$26,$2B,$FC,$5B
	.dc.b	$7D,$B0,$E5,$59,$3C,$86,$FC,$B7,$C8,$BE,$07,$57,$FE,$36,$CB,$FA
	.dc.b	$94,$E5,$F5,$5F,$18,$C6,$3E,$B8,$ED,$CE,$F9,$2F,$E0,$33,$E7,$8C
	.dc.b	$5F,$F3,$A6,$C7,$F1,$FA,$46,$23,$F9,$17,$6F,$C2,$6E,$BD,$1F,$EA
	.dc.b	$F1,$97,$F9,$2C,$66,$B9,$FC,$5F,$90,$F8,$5F,$97,$7A,$37,$FF,$27
	.dc.b	$E3,$FE,$7A,$CB,$FB,$0C,$7F,$94,$A7,$FD,$99,$1F,$D4,$1B,$8F,$9D
	.dc.b	$AF,$F6,$7C,$11,$88,$FA,$CF,$F9,$3E,$7E,$5F,$C0,$F8,$BE,$0F,$E5
	.dc.b	$3F,$23,$69,$3F,$43,$E0,$AE,$1F,$CB,$BE,$4F,$D9,$A1,$A4,$AF,$35
	.dc.b	$7D,$2F,$93,$F7,$B6,$3F,$9F,$7C,$9F,$D6,$3E,$D7,$C0,$D8,$FE,$85
	.dc.b	$DB,$D8,$8F,$D7,$7E,$87,$3F,$E2,$3C,$0F,$BF,$F2,$5F,$83,$6D,$7E
	.dc.b	$23,$B7,$7B,$FF,$07,$8F,$F2,$33,$77,$FF,$22,$B5,$F9,$5B,$BE,$4B
	.dc.b	$DE,$4A,$FC,$43,$D1,$87,$42,$CF,$D5,$F8,$2F,$C9,$FC,$8F,$E0,$1B
	.dc.b	$3F,$D7,$FE,$47,$E5,$91,$FC,$A3,$F8,$9E,$6F,$E7,$51,$FD,$97,$E4
	.dc.b	$7F,$13,$E3,$FE,$79,$FC,$5E,$BF,$A0,$F0,$6D,$7F,$6B,$7F,$47,$F2
	.dc.b	$3F,$60,$91,$FA,$17,$E5,$FC,$9F,$F8,$9E,$3F,$B9,$77,$FE,$D6,$31
	.dc.b	$B7,$F6,$E8,$FC,$CF,$C2,$7E,$BB,$BA,$DF,$F7,$97,$F7,$DF,$06,$F9
	.dc.b	$FF,$D1,$36,$7F,$C5,$2D,$C3,$F6,$0D,$7F,$C8,$1D,$0C,$9E,$CF,$4F
	.dc.b	$92,$F2,$37,$0F,$D3,$9F,$C4,$F0,$FD,$2F,$87,$FE,$BA,$6C,$7F,$17
	.dc.b	$AF,$F3,$23,$C3,$1F,$F6,$1C,$E9,$FF,$62,$FD,$5B,$E4,$3F,$74,$DA
	.dc.b	$FF,$20,$B6,$7F,$84,$35,$F4,$FE,$5E,$F9,$4F,$CB,$B6,$BF,$D2,$CD
	.dc.b	$9F,$F4,$AB,$5F,$DD,$62,$F8,$8F,$97,$E6,$FF,$29,$8C,$5F,$A7,$87
	.dc.b	$FD,$01,$39,$F8,$8D,$5D,$AD,$5F,$0C,$3F,$DD,$2B,$E2,$3E,$BE,$7D
	.dc.b	$5B,$5C,$1C,$3C,$BE,$1E,$DF,$2F,$88,$F2,$1E,$5F,$87,$B5,$E0,$7F
	.dc.b	$91,$63,$FE,$B8,$8F,$F3,$4B,$6B,$F9,$83,$5F,$83,$F1,$1C,$E7,$D2
	.dc.b	$D8,$F4,$FC,$47,$E3,$BD,$0F,$88,$DA,$FF,$4C,$7D,$0B,$DD,$FE,$0E
	.dc.b	$D8,$FF,$63,$36,$FF,$D8,$AD,$7E,$8F,$C3,$FE,$6B,$E5,$3E,$83,$6B
	.dc.b	$E5,$F5,$7D,$33,$EA,$7C,$45,$8F,$A9,$D4,$F5,$7F,$46,$7F,$65,$C7
	.dc.b	$FC,$D7,$E2,$E2,$BF,$25,$0F,$A1,$7C,$0F,$F4,$D8,$C3,$63,$33,$FE
	.dc.b	$31,$9C,$B6,$3F,$CD,$0D,$F8,$2E,$04,$7E,$8F,$C4,$7E,$DB,$D7,$FC
	.dc.b	$63,$63,$F0,$EE,$1F,$C4,$FC,$3D,$C3,$F1,$3F,$11,$FB,$EF,$0F,$F2
	.dc.b	$8D,$AF,$A3,$ED,$7D,$02,$3F,$41,$FE,$07,$F2,$7F,$C1,$8F,$CE,$5D
	.dc.b	$9F,$87,$CA,$DF,$E1,$36,$3E,$03,$67,$FC,$A4,$D7,$FD,$31,$B7,$F7
	.dc.b	$97,$FD,$94,$D7,$FD,$9A,$D7,$E8,$7B,$FF,$DD,$FF,$30,$7C,$47,$F1
	.dc.b	$AF,$93,$FF,$0C,$46,$31,$FB,$B1,$FE,$1F,$F8,$97,$91,$E1,$C7,$FD
	.dc.b	$B3,$18,$D3,$FC,$0C,$DB,$FF,$34,$36,$7F,$C2,$18,$7E,$4F,$CB,$FE
	.dc.b	$31,$EF,$F8,$3B,$BF,$95,$BF,$E2,$7D,$DE,$CF,$E7,$09,$97,$E7,$1F
	.dc.b	$F0,$87,$E1,$DF,$C7,$B2,$FF,$0E,$35,$FF,$03,$35,$FF,$1D,$35,$FF
	.dc.b	$78,$FE,$36,$E5,$FC,$93,$6F,$F2,$0D,$9F,$F4,$13,$5F,$F0,$97,$CA
	.dc.b	$7F,$83,$D6,$8A,$FF,$24,$4E,$63,$12,$F4,$9F,$98,$BC,$0F,$F2,$45
	.dc.b	$EF,$05,$EE,$B5,$1F,$C8,$E7,$53,$9A,$E3,$E3,$7E,$C7,$E9,$77,$3F
	.dc.b	$E9,$96,$BF,$ED,$F6,$BD,$83,$87,$C2,$FF,$2B,$78,$79,$FF,$B7,$8C
	.dc.b	$47,$16,$BF,$D3,$BD,$EF,$EC,$64,$C7,$FB,$1B,$FC,$5B,$DE,$91,$FC
	.dc.b	$59,$B3,$F8,$6E,$BF,$C9,$9F,$83,$F6,$ED,$7F,$20,$6C,$FD,$08,$F8
	.dc.b	$FB,$39,$3F,$CE,$91,$FE,$79,$E7,$FF,$30,$F0,$FF,$67,$F9,$7F,$EE
	.dc.b	$E4,$9F,$EE,$EF,$F2,$AF,$6E,$47,$F2,$A6,$E3,$FD,$EB,$E0,$FF,$C0
	.dc.b	$27,$EA,$3C,$2B,$5F,$D5,$F1,$FF,$24,$3F,$EC,$9E,$E4,$62,$FF,$47
	.dc.b	$F0,$FF,$B3,$3F,$69,$F0,$F2,$F8,$52,$31,$5F,$E0,$AF,$81,$FF,$08
	.dc.b	$9F,$B8,$F8,$3B,$7E,$1F,$86,$FD,$F1,$1E,$6F,$F0,$1E,$4F,$DF,$F7
	.dc.b	$FF,$00,$6C,$FA,$79,$30,$D3,$D9,$1F,$E2,$0F,$83,$91,$EF,$E1,$B9
	.dc.b	$FD,$33,$FC,$1F,$F1,$9F,$E2,$42,$4F,$F8,$91,$FF,$44,$7E,$64,$F6
	.dc.b	$3F,$C5,$7F,$09,$FE,$A9,$8C,$57,$F4,$E8,$FF,$17,$BF,$E2,$E2,$1F
	.dc.b	$E2,$E3,$FD,$25,$F9,$9F,$EF,$79,$0D,$27,$CC,$7F,$4F,$CB,$F5,$BF
	.dc.b	$21,$FE,$BD,$D3,$FD,$FC,$D8,$FD,$20,$70,$FE,$9D,$BF,$24,$E2,$FE
	.dc.b	$03,$3F,$C9,$AB,$F5,$93,$9E,$BD,$D6,$B2,$56,$A3,$0F,$FB,$3F,$2F
	.dc.b	$E0,$98,$7E,$25,$AF,$FA,$E5,$AF,$FA,$8D,$AF,$1F,$07,$FB,$45,$FC
	.dc.b	$26,$CF,$F8,$BE,$7A,$46,$B6,$FE,$F5,$1F,$E2,$F9,$7C,$C7,$F0,$5C
	.dc.b	$BF,$8C,$E1,$F3,$23,$1B,$7F,$B9,$70,$E3,$87,$E0,$6B,$E8,$FC,$27
	.dc.b	$80,$91,$EB,$7C,$7F,$07,$CA,$3F,$A6,$E5,$FE,$37,$7C,$51,$9A,$5B
	.dc.b	$D3,$C2,$F3,$6F,$8B,$30,$FF,$24,$4E,$5B,$1C,$3C,$29,$18,$37,$E1
	.dc.b	$7C,$32,$3F,$B9,$A4,$7C,$AF,$83,$F2,$68,$69,$E4,$EF,$F2,$F5,$7C
	.dc.b	$D2,$61,$CD,$FA,$1D,$32,$3B,$35,$FB,$DF,$07,$D8,$8F,$B3,$8B,$E4
	.dc.b	$78,$CF,$CD,$3F,$D2,$53,$F9,$83,$FA,$73,$87,$F3,$71,$BF,$8A,$FF
	.dc.b	$16,$36,$3F,$1D,$F0,$BF,$86,$23,$83,$A4,$77,$F9,$EF,$A9,$F4,$FF
	.dc.b	$8B,$9F,$C0,$37,$AB,$FA,$FD,$6E,$7F,$CF,$53,$9C,$BF,$AE,$F3,$A3
	.dc.b	$13,$FE,$0E,$6D,$71,$8F,$E1,$8F,$BF,$E2,$7D,$3A,$1A,$49,$FD,$47
	.dc.b	$C4,$7E,$80,$8E,$0F,$F2,$67,$D1,$7E,$2D,$BD,$5F,$E4,$F5,$19,$7E
	.dc.b	$63,$5F,$E0,$AB,$DB,$E0,$D8,$EB,$18,$3E,$7F,$DB,$3E,$4B,$F3,$9E
	.dc.b	$B7,$F8,$12,$BF,$50,$DE,$0F,$93,$5B,$9F,$F0,$07,$07,$02,$3F,$EB
	.dc.b	$DD,$93,$B8,$FE,$64,$3F,$24,$FF,$15,$BF,$E6,$B6,$BF,$87,$6B,$FC
	.dc.b	$81,$AF,$D6,$FC,$4D,$F2,$BF,$C3,$1B,$1F,$D7,$5A,$FF,$28,$E3,$46
	.dc.b	$23,$FD,$BF,$C4,$F8,$DF,$B2,$48,$FC,$6C,$65,$FE,$D1,$C9,$FC,$0D
	.dc.b	$FE,$19,$DA,$F9,$FF,$D8,$9F,$92,$F1,$3F,$85,$7C,$0B,$EE,$9D,$FF
	.dc.b	$E8,$98,$C3,$63,$38,$7F,$95,$9A,$FA,$71,$77,$7E,$FB,$E4,$24,$FE
	.dc.b	$6A,$3F,$8C,$2F,$A3,$7C,$CF,$E2,$5B,$1F,$E9,$96,$BF,$B5,$6B,$FD
	.dc.b	$3B,$E5,$9F,$41,$F9,$16,$C7,$FA,$85,$AF,$EF,$5A,$FF,$41,$FA,$1F
	.dc.b	$E2,$64,$7F,$5F,$E1,$FF,$06,$E2,$FC,$A4,$7F,$02,$6B,$F5,$3B,$BF
	.dc.b	$56,$7F,$48,$D7,$E7,$47,$F4,$3F,$95,$FE,$3E,$7F,$00,$E3,$FE,$03
	.dc.b	$8C,$9F,$DE,$BF,$C1,$37,$C4,$7D,$8F,$95,$FE,$4E,$63,$BB,$F4,$1B
	.dc.b	$FE,$5E,$7F,$B7,$B5,$FF,$2A,$35,$FF,$31,$B5,$FA,$1E,$FF,$CF,$23
	.dc.b	$F3,$DF,$EB,$FC,$1F,$CD,$89,$1F,$ED,$2F,$9C,$4F,$F1,$AF,$96,$93
	.dc.b	$FC,$59,$FE,$3B,$8F,$F1,$EB,$FE,$7F,$12,$7F,$8D,$C7,$CD,$E6,$FE
	.dc.b	$26,$48,$FC,$DB,$F2,$FE,$47,$FA,$3F,$C9,$49,$FE,$47,$1F,$CA,$BE
	.dc.b	$7B,$FA,$F7,$D0,$7F,$AD,$27,$FD,$03,$B7,$F7,$08,$C4,$7F,$2E,$9D
	.dc.b	$FC,$97,$F7,$E9,$CC,$FF,$9E,$FC,$14,$62,$3F,$97,$B5,$FE,$5C,$7F
	.dc.b	$3A,$21,$E0,$F1,$B7,$5E,$4F,$91,$93,$93,$FD,$87,$48,$F4,$BA,$8F
	.dc.b	$DB,$E5,$FD,$98,$C6,$5E,$BF,$8A,$F7,$FC,$B4,$AE,$99,$FA,$27,$11
	.dc.b	$FC,$03,$E4,$7F,$B7,$98,$BF,$E9,$BF,$9A,$BE,$5B,$FA,$DB,$6B,$FA
	.dc.b	$BB,$5F,$EC,$2D,$7E,$37,$E2,$8F,$8E,$FF,$13,$B6,$3F,$CF,$0D,$7F
	.dc.b	$8E,$35,$FF,$18,$70,$FF,$80,$88,$FA,$B6,$FE,$65,$8F,$F6,$2F,$07
	.dc.b	$FC,$0D,$3F,$E2,$D3,$FC,$12,$7F,$7A,$72,$FF,$8A,$7E,$87,$EB,$6C
	.dc.b	$69,$38,$BF,$01,$9F,$F8,$96,$3E,$FD,$7C,$A6,$BF,$E3,$3C,$7D,$58
	.dc.b	$F8,$FB,$38,$71,$CA,$FE,$95,$9A,$7A,$7C,$D7,$F9,$2B,$13,$49,$5F
	.dc.b	$E9,$DC,$7F,$A0,$E3,$FE,$46,$A7,$1F,$94,$FC,$EF,$9B,$F9,$EF,$8E
	.dc.b	$F4,$72,$CC,$7E,$44,$7A,$FC,$7F,$01,$23,$FD,$55,$87,$F2,$87,$F3
	.dc.b	$7F,$1D,$F0,$7D,$49,$1F,$46,$3E,$63,$FE,$1E,$DF,$FB,$E7,$A6,$69
	.dc.b	$FD,$FF,$93,$E7,$78,$3E,$F2,$57,$81,$FF,$13,$FE,$01,$F9,$F8,$FF
	.dc.b	$11,$B5,$FF,$16,$B5,$FF,$76,$5A,$31,$1F,$ED,$67,$E2,$BF,$5F,$97
	.dc.b	$ED,$1F,$E3,$C8,$CC,$7E,$0F,$E4,$3E,$91,$1F,$86,$FD,$5E,$0F,$C1
	.dc.b	$12,$3F,$C0,$93,$F9,$2E,$EF,$F1,$7F,$E6,$DF,$E4,$3E,$4F,$AD,$F1
	.dc.b	$5F,$89,$3F,$CD,$D1,$99,$FF,$1A,$F0,$B8,$FF,$04,$4C,$7E,$9B,$F8
	.dc.b	$5E,$1F,$C9,$12,$3C,$3A,$74,$F7,$3D,$95,$27,$F7,$F8,$FF,$8E,$5B
	.dc.b	$F3,$DF,$E9,$88,$34,$91,$F8,$2F,$91,$FD,$01,$1E,$7F,$2D,$FD,$D3
	.dc.b	$CA,$4F,$F8,$ED,$B3,$FE,$65,$0E,$1F,$F3,$43,$5F,$E5,$5A,$3E,$6F
	.dc.b	$F4,$44,$8F,$C9,$BF,$84,$F9,$4F,$93,$F3,$B2,$7E,$8C,$57,$FD,$49
	.dc.b	$9A,$71,$C5,$FE,$66,$E3,$5F,$D2,$F4,$FE,$9F,$C0,$CC,$7E,$3C,$57
	.dc.b	$9C,$5B,$FF,$5A,$E3,$FD,$F9,$FF,$26,$7D,$23,$E1,$DD,$9F,$EE,$0C
	.dc.b	$98,$CA,$D4,$E6,$31,$5C,$1F,$F7,$15,$70,$55,$FE,$A4,$78,$18,$9E
	.dc.b	$1F,$27,$EC,$8E,$77,$AA,$EF,$1D,$0F,$17,$C5,$7F,$BB,$70,$ED,$3A
	.dc.b	$BE,$55,$D6,$F9,$61,$DA,$F7,$87,$1A,$EE,$75,$38,$17,$0E,$93,$C2
	.dc.b	$39,$DC,$75,$7D,$5F,$29,$93,$E1,$1D,$0B,$5E,$C6,$C7,$C9,$77,$F2
	.dc.b	$7F,$E6,$AF,$7B,$F2,$4D,$8F,$F5,$C3,$5D,$F6,$A7,$F3,$16,$F9,$67
	.dc.b	$99,$86,$DF,$FB,$06,$75,$39,$24,$7C,$19,$DE,$BF,$32,$BF,$CC,$FD
	.dc.b	$DF,$56,$73,$AF,$D2,$6D,$FE,$1D,$AF,$E0,$BD,$35,$71,$C6,$4F,$77
	.dc.b	$F0,$13,$A8,$C1,$31,$F1,$5F,$4C,$7C,$73,$9C,$7C,$58,$74,$BF,$46
	.dc.b	$1C,$DE,$31,$F9,$9D,$BF,$B4,$7C,$A7,$7E,$1F,$EA,$98,$5D,$3F,$8E
	.dc.b	$CD,$CB,$F4,$BD,$1F,$B2,$9F,$ED,$53,$9C,$39,$C7,$F6,$DF,$A2,$C5
	.dc.b	$2B,$97,$89,$FE,$BD,$F0,$33,$86,$3C,$7F,$F6,$B6,$3F,$DD,$63,$13
	.dc.b	$23,$FC,$43,$F4,$6F,$8D,$C6,$9F,$EB,$AE,$FE,$D5,$FC,$65,$B1,$FC
	.dc.b	$63,$3F,$F6,$53,$FB,$58,$CB,$FE,$43,$B7,$CC,$7B,$E3,$F9,$0D,$7B
	.dc.b	$67,$37,$FE,$65,$8C,$63,$23,$CB,$F1,$D6,$FF,$DE,$7A,$3E,$87,$C3
	.dc.b	$0D,$BF,$DB,$D1,$86,$DF,$EA,$27,$55,$C6,$31,$B3,$37,$89,$E5,$3C
	.dc.b	$DE,$2D,$AF,$F7,$05,$FF,$2B,$6F,$87,$EB,$F8,$9F,$A0,$FD,$1F,$63
	.dc.b	$C5,$5F,$E4,$7A,$F1,$69,$FE,$F3,$9C,$9E,$AC,$FC,$57,$F1,$4F,$8D
	.dc.b	$AD,$BF,$0F,$CA,$FD,$D3,$E5,$5D,$8F,$2F,$CA,$F8,$08,$F8,$32,$47
	.dc.b	$93,$C3,$FC,$31,$F4,$FC,$5B,$5F,$84,$F1,$BF,$15,$E0,$FD,$76,$C7
	.dc.b	$C2,$F0,$3D,$A4,$7B,$3E,$BE,$36,$9F,$93,$9F,$C0,$35,$FC,$8F,$CA
	.dc.b	$FE,$0F,$3F,$C2,$36,$3F,$D0,$1C,$2F,$88,$7C,$C7,$DD,$6C,$FC,$C6
	.dc.b	$F3,$FF,$26,$AE,$15,$E3,$F2,$3F,$20,$7E,$47,$82,$E1,$DB,$07,$CA
	.dc.b	$FE,$57,$83,$F5,$B1,$89,$FD,$3C,$63,$E2,$FC,$44,$8F,$13,$E5,$F9
	.dc.b	$4E,$27,$E7,$F8,$6C,$F7,$F7,$3F,$45,$F2,$BD,$9B,$1E,$3F,$89,$F7
	.dc.b	$11,$F7,$3F,$07,$E2,$3C,$A7,$EA,$E2,$9F,$E9,$CF,$0D,$F3,$3F,$C2
	.dc.b	$DB,$1F,$EA,$0F,$A3,$FC,$A7,$23,$CB,$FE,$4D,$E0,$7E,$E3,$0F,$E1
	.dc.b	$BF,$47,$FE,$24,$C3,$F3,$2D,$BF,$9C,$D7,$FC,$A2,$D7,$FB,$FD,$7F
	.dc.b	$02,$F9,$FE,$7E,$27,$8D,$B7,$E7,$F9,$BF,$37,$4F,$9D,$B1,$92,$61
	.dc.b	$E7,$7A,$EA,$F9,$CE,$11,$F0,$7E,$57,$E0,$1F,$07,$A6,$C7,$C5,$6D
	.dc.b	$F3,$3F,$C2,$F7,$F7,$DC,$69,$FC,$6A,$FE,$57,$C0,$FC,$BE,$8B,$8F
	.dc.b	$CF,$7E,$77,$46,$30,$E3,$8A,$31,$C5,$F3,$A1,$F2,$3C,$30,$FB,$0D
	.dc.b	$BF,$27,$73,$E7,$BF,$DC,$15,$F7,$7A,$7F,$A0,$C6,$2B,$BC,$7D,$FE
	.dc.b	$9C,$FC,$4D,$BF,$D5,$DE,$F3,$F1,$75,$7E,$17,$E4,$FF,$0C,$DA,$FC
	.dc.b	$57,$C9,$7E,$24,$FC,$57,$C9,$58,$FF,$7A,$69,$FE,$5C,$F8,$0F,$E7
	.dc.b	$DE,$A7,$F9,$66,$73,$F4,$7F,$EB,$A7,$71,$9A,$FA,$02,$7F,$D7,$33
	.dc.b	$FE,$67,$6C,$7F,$39,$6B,$FE,$22,$6B,$FD,$47,$9F,$FA,$74,$70,$D3
	.dc.b	$FC,$01,$3F,$E7,$09,$CF,$C5,$FE,$8F,$E4,$E6,$1F,$A7,$F9,$3B,$7F
	.dc.b	$61,$B8,$7D,$76,$CF,$F9,$6B,$93,$EA,$7C,$57,$EB,$7E,$4F,$E9,$B6
	.dc.b	$7F,$65,$EF,$E1,$A7,$D2,$8F,$DA,$63,$23,$E5,$F2,$7C,$5C,$BF,$0D
	.dc.b	$3E,$CF,$8E,$FE,$BE,$FC,$76,$EE,$69,$FD,$93,$E0,$FF,$CE,$3F,$03
	.dc.b	$C7,$1F,$F3,$8E,$1F,$E2,$39,$CC,$7D,$6A,$FF,$0A,$47,$F9,$E3,$8F
	.dc.b	$FC,$25,$F3,$EF,$92,$CF,$27,$97,$0F,$DE,$CE,$5E,$A6,$3E,$47,$77
	.dc.b	$C9,$B1,$F9,$2F,$8E,$EB,$D9,$F9,$A6,$C7,$2D,$FF,$D1,$38,$7E,$25
	.dc.b	$FD,$1F,$02,$BC,$BF,$25,$C6,$FF,$27,$F4,$4E,$5F,$8E,$D8,$FF,$59
	.dc.b	$37,$0F,$F0,$23,$5F,$F4,$76,$AF,$CD,$FF,$C4,$95,$D2,$FF,$C6,$91
	.dc.b	$8A,$FC,$CD,$F3,$E2,$FF,$27,$73,$BE,$8D,$E6,$FF,$8E,$FE,$83,$FC
	.dc.b	$1B,$9F,$F5,$C9,$CB,$EF,$D1,$3A,$CB,$F8,$0D,$69,$B7,$FE,$45,$DD
	.dc.b	$E2,$7F,$90,$AB,$59,$FE,$83,$E8,$FF,$B1,$E1,$FE,$88,$6B,$FE,$AB
	.dc.b	$6B,$FE,$B4,$6A,$7F,$96,$67,$FB,$E7,$FB,$6C,$F5,$F3,$7F,$C7,$91
	.dc.b	$89,$FD,$26,$3F,$EE,$27,$FC,$3E,$DB,$FE,$E8,$D7,$FD,$79,$AF,$F1
	.dc.b	$DF,$92,$FF,$11,$72,$7E,$E6,$F9,$C7,$E4,$BC,$1F,$E3,$EF,$F2,$0E
	.dc.b	$CF,$CA,$F6,$5A,$F9,$57,$FA,$37,$F8,$FC,$E7,$1F,$E7,$72,$F9,$7C
	.dc.b	$6D,$F6,$34,$FF,$2C,$D7,$AB,$E2,$BF,$C9,$51,$88,$FF,$27,$37,$91
	.dc.b	$83,$E0,$FF,$99,$DF,$EA,$1B,$9F,$93,$C1,$FD,$59,$FA,$CF,$97,$B1
	.dc.b	$E8,$F9,$6F,$82,$47,$FC,$AB,$82,$3E,$2F,$C9,$7E,$D1,$FD,$AB,$67
	.dc.b	$E0,$FC,$B7,$EE,$0F,$DC,$FC,$BD,$8F,$8B,$E0,$7C,$E2,$3D,$9F,$A3
	.dc.b	$C9,$FC,$00,$FE,$03,$C7,$67,$E6,$B8,$7E,$7F,$87,$78,$BE,$6F,$07
	.dc.b	$F0,$83,$F8,$4F,$C9,$5A,$FA,$BC,$5F,$78,$8F,$B9,$FF,$31,$7D,$1F
	.dc.b	$FA,$12,$FF,$B8,$B7,$97,$FC,$2A,$FE,$C5,$BE,$9F,$FA,$9E,$1F,$21
	.dc.b	$C8,$98,$7F,$9F,$A3,$2D,$F9,$1E,$E4,$70,$7F,$A1,$F0,$13,$E5,$3F
	.dc.b	$C3,$7A,$7F,$A2,$DB,$1F,$EF,$46,$BC,$03,$8F,$96,$71,$7F,$8B,$E5
	.dc.b	$AF,$FA,$ED,$EF,$3A,$D7,$85,$6E,$B5,$4A,$D3,$CB,$0F,$F1,$7E,$6F
	.dc.b	$7F,$D5,$68,$C5,$7C,$39,$CC,$62,$3D,$07,$91,$FF,$51,$BF,$27,$E4
	.dc.b	$F6,$47,$E6,$7C,$8F,$BF,$3F,$AA,$DD,$7C,$6D,$9E,$9F,$27,$72,$F2
	.dc.b	$7C,$9F,$88,$93,$F5,$DF,$2F,$C9,$6F,$33,$4C,$7C,$9E,$7E,$46,$5F
	.dc.b	$F2,$BA,$FF,$48,$B7,$E4,$F9,$7C,$94,$8D,$3C,$1F,$C7,$FC,$97,$C4
	.dc.b	$23,$F8,$97,$E3,$FC,$97,$8C,$91,$DF,$E4,$BC,$DF,$23,$E7,$F9,$19
	.dc.b	$3E,$38,$AF,$CE,$DF,$EA,$4F,$D2,$24,$7E,$5F,$8B,$E0,$11,$F6,$7C
	.dc.b	$8F,$C9,$F9,$CF,$F3,$EC,$7F,$A9,$BE,$77,$FD,$F4,$DF,$96,$D7,$C9
	.dc.b	$FD,$1F,$49,$5A,$AF,$F5,$33,$5F,$E6,$6D,$7F,$C3,$CD,$7F,$DD,$CF
	.dc.b	$EB,$9F,$07,$83,$D6,$31,$1F,$E6,$8C,$3F,$B0,$BA,$3F,$AF,$3F,$EB
	.dc.b	$A6,$FC,$97,$1A,$F6,$7A,$BC,$FC,$39,$5C,$22,$BE,$88,$FD,$B6,$3E
	.dc.b	$C6,$E2,$7F,$75,$2F,$A3,$74,$FE,$4E,$31,$D5,$67,$F1,$B1,$FE,$5C
	.dc.b	$76,$D8,$E7,$1A,$8F,$D3,$9C,$8F,$E1,$0E,$5F,$EB,$BF,$47,$FD,$15
	.dc.b	$FF,$6D,$B5,$FF,$3E,$B7,$BF,$FD,$63,$F2,$2E,$8F,$B1,$39,$E4,$FF
	.dc.b	$55,$FB,$9F,$E5,$CB,$FF,$5B,$B6,$3F,$CA,$93,$BC,$3F,$86,$FE,$70
	.dc.b	$ED,$F8,$1F,$49,$FE,$29,$9F,$F3,$CB,$5F,$F4,$FC,$ED,$CF,$9F,$F2
	.dc.b	$A7,$51,$FC,$19,$AF,$F1,$F6,$BF,$C7,$39,$9F,$2D,$F8,$5F,$12,$4F
	.dc.b	$E0,$A2,$BF,$87,$F9,$5F,$CE,$18,$DB,$FD,$A7,$7F,$EF,$49,$FF,$22
	.dc.b	$7C,$43,$E8,$9D,$FF,$C3,$63,$0F,$E9,$7E,$5F,$13,$AA,$FC,$8F,$CB
	.dc.b	$FE,$A4,$CD,$6B,$C1,$FC,$27,$D1,$7F,$81,$6B,$FD,$3A,$D7,$FD,$8D
	.dc.b	$8F,$86,$77,$EF,$FE,$3A,$B5,$5F,$CA,$9A,$FF,$43,$6B,$F1,$7A,$3F
	.dc.b	$74,$47,$FD,$57,$F3,$6F,$93,$FD,$F7,$CC,$FE,$B7,$9E,$57,$EA,$A2
	.dc.b	$BF,$AF,$F9,$9F,$E0,$86,$38,$BE,$4F,$C4,$7F,$07,$F9,$99,$5F,$B2
	.dc.b	$8A,$FE,$DE,$2C,$F0,$8F,$F5,$EF,$D6,$7F,$A3,$5F,$87,$F0,$BF,$18
	.dc.b	$FD,$54,$66,$3F,$D7,$7B,$7E,$75,$F8,$53,$F1,$48,$FF,$6D,$F9,$77
	.dc.b	$C2,$FE,$CD,$B1,$FE,$C6,$8F,$A6,$7F,$1F,$F9,$99,$1F,$C2,$9F,$E4
	.dc.b	$5F,$35,$FC,$8F,$E6,$BF,$92,$70,$CA,$FE,$1F,$1F,$C9,$FE,$6B,$F8
	.dc.b	$B7,$44,$9F,$E2,$71,$5F,$E3,$1F,$35,$FC,$B4,$C6,$FF,$92,$F9,$4F
	.dc.b	$E6,$1F,$35,$2B,$F8,$DC,$55,$2F,$1F,$FC,$13,$EA,$BE,$63,$FD,$55
	.dc.b	$5F,$E7,$F6,$BF,$C9,$F0,$FE,$C9,$18,$BF,$D0,$CF,$F3,$F3,$3F,$33
	.dc.b	$A9,$DF,$CC,$C9,$E9,$15,$F1,$E3,$E2,$84,$E6,$FC,$D5,$FE,$12,$9F
	.dc.b	$33,$E5,$F9,$2F,$19,$1F,$47,$CC,$D7,$A9,$C7,$E5,$3D,$26,$6F,$AE
	.dc.b	$9F,$07,$C1,$FE,$A5,$F4,$7F,$CB,$2F,$8F,$AE,$8C,$5F,$0C,$FF,$60
	.dc.b	$FF,$AB,$23,$85,$7F,$A2,$1A,$FF,$BC,$DA,$AB,$A7,$AB,$FE,$22,$BF
	.dc.b	$E4,$98,$FD,$3E,$BF,$9E,$7F,$84,$FE,$8D,$F2,$BF,$D9,$FC,$6F,$E2
	.dc.b	$19,$79,$9B,$A7,$FA,$F9,$BC,$5F,$37,$B7,$87,$93,$FD,$52,$4D,$25
	.dc.b	$FF,$48,$D9,$2F,$17,$F0,$39,$4A,$FF,$71,$56,$A7,$C9,$E2,$49,$CD
	.dc.b	$FF,$B0,$9F,$C6,$7C,$A7,$A7,$C5,$F1,$12,$3C,$12,$8F,$37,$85,$F9
	.dc.b	$0F,$94,$95,$D2,$3F,$27,$F2,$92,$3F,$30,$FE,$5B,$E3,$BE,$37,$83
	.dc.b	$EF,$24,$F6,$7E,$87,$CA,$7E,$73,$D6,$93,$FA,$18,$FC,$FF,$C9,$FC
	.dc.b	$F8,$87,$E9,$5F,$F0,$C7,$D1,$FE,$9A,$3F,$D8,$ED,$7F,$CE,$BC,$28
	.dc.b	$C4,$7E,$05,$BD,$3F,$B1,$F9,$5E,$7E,$14,$AF,$D9,$47,$6F,$02,$47
	.dc.b	$D0,$B4,$7E,$D3,$E5,$7C,$7E,$1F,$F8,$42,$7F,$90,$57,$92,$30,$67
	.dc.b	$BB,$D2,$48,$FE,$01,$F2,$5F,$E6,$1F,$A3,$FF,$19,$77,$FE,$99,$BD
	.dc.b	$FF,$C2,$1B,$5F,$B5,$DB,$E1,$F6,$F8,$49,$1F,$8B,$CF,$F3,$2D,$EB
	.dc.b	$F9,$24,$7F,$D3,$78,$F7,$F9,$6F,$17,$C2,$49,$CE,$7E,$18,$F3,$7C
	.dc.b	$3F,$D2,$EA,$F8,$04,$AF,$03,$FE,$8E,$D7,$FD,$0F,$1F,$C1,$35,$46
	.dc.b	$32,$F7,$7C,$9F,$C1,$F9,$4F,$85,$F0,$1F,$D0,$7A,$A6,$3E,$23,$FD
	.dc.b	$4B,$82,$73,$1F,$E3,$0F,$51,$5F,$E3,$E6,$C7,$F4,$D8,$FA,$DD,$1F
	.dc.b	$8A,$24,$75,$CF,$CB,$F2,$98,$8F,$46,$2F,$94,$FA,$C6,$75,$FA,$DE
	.dc.b	$D7,$D9,$25,$73,$AF,$E3,$3C,$5F,$66,$7E,$F7,$C4,$CA,$D4,$7F,$94
	.dc.b	$7E,$8F,$FC,$7D,$9F,$FB,$37,$4F,$F6,$44,$62,$BF,$D9,$31,$8F,$89
	.dc.b	$FF,$63,$C6,$0F,$F6,$AC,$BE,$47,$FD,$A1,$F2,$EB,$E1,$87,$5B,$DE
	.dc.b	$1C,$AB,$51,$89,$D2,$C3,$A3,$87,$C9,$3C,$23,$55,$CE,$FB,$C7,$94
	.dc.b	$E5,$87,$92,$33,$5C,$6F,$9C,$60,$F8,$33,$BA,$F2,$1C,$E7,$6E,$2B
	.dc.b	$94,$EE,$3C,$78,$7F,$BE,$6B,$E5,$47,$FB,$71,$AF,$98,$FF,$6F,$FC
	.dc.b	$BB,$C9,$7C,$1F,$FC,$01,$8C,$36,$33,$3F,$E8,$09,$CB,$FE,$82,$A7
	.dc.b	$C9,$BD,$9F,$13,$67,$4D,$7C,$8D,$7F,$02,$EF,$3F,$8C,$DD,$3F,$0B
	.dc.b	$EE,$7F,$78,$9C,$DF,$D5,$AD,$5F,$E1,$1A,$FD,$BA,$E1,$7D,$AB,$81
	.dc.b	$27,$E7,$4E,$9E,$51,$9B,$EF,$1A,$E1,$C5,$F3,$8F,$AD,$87,$83,$1F
	.dc.b	$F3,$AD,$6A,$5A,$BE,$47,$FD,$F7,$3F,$A9,$F9,$7F,$F0,$9F,$2F,$F8
	.dc.b	$22,$73,$7F,$B4,$9C,$D7,$F8,$B2,$7F,$D3,$8F,$EE,$DD,$D7,$1E,$3D
	.dc.b	$DF,$82,$73,$5C,$7E,$5B,$38,$FF,$16,$6C,$9C,$F3,$5A,$FE,$E7,$F9
	.dc.b	$CD,$C3,$F8,$F5,$DC,$BF,$C5,$7E,$62,$B1,$7D,$CE,$3C,$DF,$E3,$0C
	.dc.b	$07,$F5,$A9,$FD,$D4,$FF,$81,$5B,$1A,$35,$3C,$5F,$84,$FB,$AB,$C0
	.dc.b	$FA,$8C,$4F,$2B,$FE,$9D,$F2,$FE,$C3,$DF,$C1,$71,$F6,$F2,$75,$CB
	.dc.b	$FC,$55,$87,$DA,$7C,$46,$23,$59,$7B,$DE,$1B,$7D,$DF,$89,$F9,$71
	.dc.b	$F3,$1B,$87,$CA,$7C,$A7,$D2,$FA,$F7,$CF,$62,$7E,$79,$DB,$2F,$F3
	.dc.b	$63,$F1,$4C,$1E,$2F,$CE,$B8,$BF,$CC,$F7,$F8,$91,$8D,$3F,$9E,$C6
	.dc.b	$3E,$62,$57,$F4,$3C,$5F,$23,$8B,$FF,$2A,$E1,$8A,$FF,$1C,$36,$3E
	.dc.b	$1E,$C8,$C4,$7D,$78,$C5,$79,$7D,$DF,$C5,$F7,$F9,$1B,$87,$87,$DB
	.dc.b	$FC,$7F,$17,$C3,$6C,$7F,$90,$B2,$7B,$9E,$F2,$47,$91,$FA,$3A,$6A
	.dc.b	$3F,$AA,$71,$7C,$D9,$FC,$DB,$F3,$FE,$2B,$F3,$BC,$76,$BF,$AD,$3F
	.dc.b	$23,$DA,$F9,$84,$8F,$82,$F6,$F8,$9F,$D2,$C7,$D9,$6B,$E7,$E1,$FD
	.dc.b	$6F,$E7,$6F,$9A,$FA,$71,$87,$F3,$1E,$47,$FA,$96,$BF,$CC,$8D,$9F
	.dc.b	$9A,$D7,$E6,$35,$FF,$20,$AF,$88,$EF,$D3,$D1,$B5,$FE,$6A,$97,$B3
	.dc.b	$E4,$EE,$F0,$B6,$75,$EC,$FE,$72,$8D,$24,$7E,$8B,$1F,$DC,$FD,$1F
	.dc.b	$F9,$1A,$31,$08,$C3,$B6,$BF,$AA,$6B,$FE,$C6,$D1,$D7,$F2,$0F,$E2
	.dc.b	$BF,$9E,$7C,$3F,$07,$FD,$57,$3A,$C3,$D9,$EC,$79,$C8,$FC,$A7,$F6
	.dc.b	$75,$FE,$C0,$8D,$9F,$E2,$17,$18,$7F,$AC,$B3,$57,$83,$DD,$97,$BD
	.dc.b	$6D,$FF,$8F,$70,$F9,$73,$93,$A7,$C5,$7A,$2B,$D3,$B7,$E0,$BE,$5A
	.dc.b	$D7,$F8,$C7,$44,$7B,$BD,$DC,$E5,$D5,$BB,$7C,$0F,$07,$F1,$67,$C4
	.dc.b	$DA,$E3,$EB,$70,$FB,$7D,$0B,$87,$9D,$F9,$67,$CC,$F8,$3B,$3F,$85
	.dc.b	$F8,$0F,$CB,$FC,$AF,$E5,$1B,$1F,$1F,$E1,$30,$D3,$81,$EB,$F1,$7F
	.dc.b	$3E,$7D,$8F,$62,$C7,$CF,$F8,$4C,$34,$FA,$51,$FA,$6F,$1A,$47,$F9
	.dc.b	$87,$47,$C4,$BD,$AC,$C7,$FB,$C0,$FC,$76,$FF,$CD,$DB,$5F,$92,$EC
	.dc.b	$FB,$6D,$9D,$47,$F8,$7E,$3B,$3A,$9E,$8E,$5B,$3F,$92,$D7,$F7,$7F
	.dc.b	$11,$6F,$ED,$3F,$96,$F3,$F1,$7F,$85,$6F,$27,$F6,$4F,$CF,$5F,$07
	.dc.b	$FC,$13,$E5,$3F,$AC,$46,$31,$FC,$24,$7F,$74,$FC,$B1,$F0,$DA,$BC
	.dc.b	$36,$BF,$15,$E6,$FF,$0E,$38,$7C,$35,$8F,$95,$F4,$5F,$E6,$37,$F6
	.dc.b	$AD,$D7,$F7,$4D,$7F,$AE,$B5,$FE,$E0,$B7,$FF,$05,$C2,$39,$6B,$FD
	.dc.b	$ED,$D5,$78,$3D,$DF,$6F,$C2,$7F,$22,$F8,$CB,$3F,$A1,$7D,$DF,$0F
	.dc.b	$8C,$7F,$D5,$1B,$A3,$19,$7C,$9F,$27,$F9,$5F,$A5,$F2,$1B,$8F,$FA
	.dc.b	$6E,$FF,$0C,$FF,$AA,$7D,$87,$27,$FA,$6E,$35,$D3,$C5,$D7,$C1,$FF
	.dc.b	$12,$D1,$68,$C3,$7C,$CF,$EF,$F1,$BE,$86,$5D,$6D,$AF,$F8,$8A,$BF
	.dc.b	$90,$AC,$7F,$C3,$F7,$C3,$4F,$F0,$C5,$6A,$7F,$C3,$DF,$47,$FD,$DE
	.dc.b	$FF,$C4,$DF,$83,$7F,$86,$67,$3E,$97,$D2,$8C,$65,$FE,$2A,$BF,$F2
	.dc.b	$65,$7F,$32,$6C,$7F,$37,$6B,$FE,$1A,$6B,$FE,$2D,$FC,$29,$EB,$FE
	.dc.b	$AD,$B5,$FC,$F1,$B3,$FD,$05,$AF,$F8,$91,$FF,$0A,$BF,$B5,$77,$8F
	.dc.b	$F8,$FD,$CB,$6B,$F4,$3F,$11,$FE,$2A,$E2,$FF,$2F,$DF,$F8,$25,$AF
	.dc.b	$F9,$1B,$87,$ED,$CF,$F8,$BB,$83,$FC,$C3,$3F,$43,$2D,$5F,$F9,$7A
	.dc.b	$3F,$CD,$F3,$AC,$7F,$65,$7F,$AD,$F8,$2F,$EE,$BF,$1F,$71,$FE,$68
	.dc.b	$FE,$D2,$31,$87,$53,$F0,$B8,$FF,$97,$1E,$B1,$FB,$AF,$80,$F5,$12
	.dc.b	$FF,$B3,$67,$F9,$28,$F4,$1B,$8F,$CC,$36,$BB,$71,$BD,$17,$CB,$FF
	.dc.b	$82,$8F,$E0,$9E,$ED,$C3,$F4,$1F,$01,$F0,$C8,$FF,$75,$BF,$CA,$C7
	.dc.b	$C7,$E1,$FE,$18,$FF,$0D,$6C,$FF,$53,$F7,$7F,$C2,$E7,$FA,$13,$C2
	.dc.b	$FC,$94,$7F,$75,$8C,$7B,$BF,$E5,$E9,$FF,$0E,$4F,$76,$E9,$FE,$66
	.dc.b	$C5,$18,$8F,$F3,$D4,$63,$DB,$ED,$EE,$77,$6D,$FF,$A1,$39,$7F,$51
	.dc.b	$7F,$EA,$8C,$3F,$CA,$11,$8C,$3A,$FB,$1E,$7F,$5F,$D0,$FF,$8C,$3D
	.dc.b	$E9,$8F,$33,$F6,$38,$3D,$1C,$E5,$F4,$38,$DF,$2F,$FE,$7A,$6D,$FC
	.dc.b	$B6,$CF,$B5,$AF,$B2,$DE,$77,$F9,$03,$E5,$FE,$53,$6F,$FC,$01,$F1
	.dc.b	$7D,$48,$F5,$7C,$17,$23,$C0,$D3,$FC,$5A,$3E,$2F,$F8,$5F,$C8,$FF
	.dc.b	$8C,$DB,$1F,$E9,$66,$BF,$EA,$96,$F8,$AF,$73,$F8,$BC,$FF,$B5,$2B
	.dc.b	$FD,$5D,$39,$E3,$DD,$6B,$8D,$5A,$8C,$3F,$C7,$3F,$3E,$7C,$C7,$F9
	.dc.b	$0A,$91,$8C,$7F,$28,$D7,$F0,$CD,$7F,$0A,$BE,$5B,$93,$63,$8C,$60
	.dc.b	$ED,$C0,$EA,$FF,$63,$67,$F8,$5B,$7C,$AE,$37,$F3,$37,$1F,$C6,$5D
	.dc.b	$CB,$CF,$F0,$FE,$02,$4F,$A1,$F1,$F9,$BB,$C4,$D3,$96,$6E,$2B,$7D
	.dc.b	$7F,$2D,$F1,$BF,$93,$18,$A3,$31,$FC,$0A,$BF,$D9,$FA,$FE,$7E,$7E
	.dc.b	$14,$FF,$06,$7F,$D0,$0F,$3A,$E5,$97,$7D,$BC,$19,$7D,$3B,$F3,$DF
	.dc.b	$F0,$56,$BF,$B7,$6B,$F4,$BE,$1B,$DC,$47,$B4,$7F,$A2,$9B,$F0,$DE
	.dc.b	$DF,$97,$91,$83,$6F,$84,$8F,$85,$FB,$BF,$11,$F2,$C9,$1D,$BE,$23
	.dc.b	$E4,$7C,$47,$D7,$F2,$BF,$BD,$F8,$F7,$CC,$3A,$BF,$F0,$04,$61,$FE
	.dc.b	$62,$D7,$FD,$AE,$D7,$FA,$74,$7F,$A6,$27,$FC,$0F,$E4,$7D,$7F,$96
	.dc.b	$95,$D2,$2B,$FC,$DB,$9B,$FD,$4A,$FD,$16,$BF,$56,$7F,$01,$F0,$BE
	.dc.b	$9F,$CF,$FF,$D8,$53,$FE,$15,$6B,$F6,$E3,$5E,$5F,$8E,$73,$3F,$EE
	.dc.b	$B6,$BF,$D8,$DA,$87,$E8,$0F,$43,$F3,$D1,$8F,$91,$B2,$6C,$E1,$FE
	.dc.b	$1B,$19,$9F,$F5,$DB,$5F,$F6,$03,$5F,$01,$E1,$FA,$57,$A9,$FC,$59
	.dc.b	$B1,$FE,$C4,$6B,$FE,$C9,$6B,$C8,$2F,$94,$F4,$FD,$27,$E7,$A7,$FD
	.dc.b	$00,$DE,$EF,$E8,$B1,$C2,$7F,$5D,$F0,$9F,$BD,$97,$E5,$DF,$AC,$7F
	.dc.b	$A5,$98,$F1,$37,$7F,$B3,$8C,$3F,$ED,$D6,$BF,$B2,$9F,$DA,$BF,$C3
	.dc.b	$9A,$FF,$8F,$3B,$3F,$76,$FD,$DF,$4F,$F0,$07,$F0,$1F,$0A,$C7,$E1
	.dc.b	$FB,$7F,$85,$18,$7F,$C0,$3F,$8C,$7F,$A0,$1F,$C3,$7A,$9F,$89,$3F
	.dc.b	$8B,$46,$63,$FC,$29,$D4,$E5,$57,$F1,$5E,$E9,$5C,$8F,$F5,$67,$D1
	.dc.b	$3E,$0B,$F0,$AB,$DB,$A6,$22,$CF,$FB,$F2,$74,$D9,$FE,$A7,$18,$8F
	.dc.b	$E7,$7F,$19,$FC,$94,$8F,$F1,$5C,$FF,$D4,$2D,$F8,$4F,$D2,$7A,$F2
	.dc.b	$30,$7E,$95,$FE,$5D,$E2,$48,$C1,$CF,$E0,$FE,$1F,$C2,$48,$C1,$DD
	.dc.b	$FE,$6D,$E9,$C8,$C1,$FE,$CD,$FA,$1F,$EA,$5D,$1F,$8A,$7F,$96,$B7
	.dc.b	$CD,$FD,$4C,$6B,$D2,$FC,$FB,$FC,$CA,$BF,$30,$D7,$33,$FD,$DA,$31
	.dc.b	$1F,$CF,$BC,$77,$8F,$FD,$24,$91,$FE,$DF,$F9,$D7,$C4,$FD,$EF,$13
	.dc.b	$EF,$F8,$52,$BE,$54,$7F,$51,$F9,$39,$1F,$C4,$5D,$3F,$8A,$F2,$3F
	.dc.b	$88,$74,$48,$FA,$51,$5F,$E3,$5E,$57,$F5,$B3,$1C,$1F,$E2,$DE,$77
	.dc.b	$C1,$FE,$40,$95,$F6,$FC,$AF,$C9,$FC,$C4,$8F,$BC,$E9,$FE,$4D,$F2
	.dc.b	$7F,$D9,$C8,$E8,$FC,$27,$9D,$F9,$CF,$84,$91,$83,$F1,$0F,$F6,$FE
	.dc.b	$19,$18,$3F,$B0,$7D,$13,$E2,$FF,$BD,$36,$33,$1F,$A8,$E9,$CF,$A3
	.dc.b	$27,$8B,$C2,$78,$F9,$5F,$D3,$FF,$41,$FC,$1D,$FF,$A4,$5B,$7F,$DE
	.dc.b	$30,$56,$FB,$3F,$A5,$57,$E0,$3C,$93,$4F,$C1,$F9,$7F,$E2,$4F,$3F
	.dc.b	$FC,$14,$4C,$7D,$7F,$0D,$F9,$0F,$5F,$F1,$44,$8F,$E4,$C7,$F3,$FF
	.dc.b	$07,$FA,$7B,$FC,$CB,$1E,$FE,$D7,$88,$91,$FE,$92,$CB,$F9,$53,$B7
	.dc.b	$FB,$D3,$FD,$6F,$DF,$FF,$0E,$1F,$07,$DA,$F8,$58,$FF,$87,$F4,$95
	.dc.b	$FE,$ED,$BF,$EB,$FD,$7F,$E7,$6C,$BD,$FE,$77,$C2,$25,$78,$5F,$8B
	.dc.b	$E4,$FF,$8B,$FE,$1D,$F5,$54,$FA,$11,$F6,$BE,$13,$5F,$25,$30,$FB
	.dc.b	$F1,$F5,$7C,$2F,$9E,$48,$F7,$3F,$E5,$1E,$7F,$E3,$D1,$FE,$25,$C7
	.dc.b	$FD,$26,$DF,$80,$FC,$3F,$A7,$F8,$8F,$8A,$F2,$7C,$9C,$AF,$C6,$C7
	.dc.b	$F8,$47,$D8,$F3,$C7,$FA,$F1,$AF,$F5,$9E,$04,$62,$3E,$A3,$5F,$F8
	.dc.b	$03,$97,$FA,$39,$F2,$7D,$0F,$C3,$3F,$33,$F1,$7D,$DF,$8D,$EB,$49
	.dc.b	$FF,$24,$61,$DA,$7C,$9F,$25,$FE,$19,$C4,$D2,$57,$9F,$D6,$F5,$11
	.dc.b	$FF,$24,$7B,$DF,$9C,$C3,$D3,$E8,$FD,$5B,$95,$E0,$C3,$FB,$CF,$47
	.dc.b	$DF,$7F,$BB,$47,$C3,$F4,$BC,$44,$7E,$37,$CF,$FC,$92,$3E,$0E,$EF
	.dc.b	$F3,$86,$6F,$85,$B8,$4A,$F2,$C7,$D1,$F4,$3D,$C4,$8F,$A4,$FF,$A3
	.dc.b	$F9,$DE,$CF,$E3,$49,$3E,$EE,$EE,$FD,$BF,$6F,$E1,$24,$FC,$28,$AF
	.dc.b	$DB,$CB,$FC,$8C,$FF,$8B,$78,$BE,$ED,$7C,$F3,$E8,$79,$5E,$AF,$62
	.dc.b	$57,$C7,$8A,$FF,$BF,$33,$72,$FE,$84,$CF,$1F,$FB,$6F,$81,$E8,$FC
	.dc.b	$4E,$F9,$87,$D6,$8A,$F3,$FA,$3F,$D5,$18,$BF,$F1,$1F,$D0,$BD,$4F
	.dc.b	$F0,$9F,$C6,$FF,$4C,$BF,$F2,$34,$E7,$C8,$FF,$24,$4E,$65,$39,$7F
	.dc.b	$D9,$DE,$9F,$F8,$FA,$3F,$A1,$C7,$08,$C5,$F6,$6D,$E6,$7B,$CE,$5F
	.dc.b	$0F,$3F,$FB,$4E,$B7,$1D,$8D,$D7,$17,$C7,$5C,$5F,$2D,$71,$76,$71
	.dc.b	$37,$3C,$6F,$D5,$87,$B5,$F4,$CE,$DC,$E1,$E2,$9D,$61,$DE,$77,$18
	.dc.b	$BF,$E6,$37,$D8,$FE,$67,$3F,$1D,$FF,$2D,$35,$FF,$75,$F2,$BE,$8B
	.dc.b	$8F,$C7,$23,$13,$FC,$F2,$31,$3A,$9C,$CF,$4E,$5F,$F6,$63,$E0,$6B
	.dc.b	$9F,$8C,$F2,$E1,$FE,$A4,$6C,$F0,$CB,$E1,$37,$4F,$B9,$EB,$B9,$BE
	.dc.b	$44,$E7,$1F,$C3,$B5,$FC,$2B,$71,$57,$19,$F9,$93,$92,$63,$F4,$E7
	.dc.b	$58,$F8,$A3,$B9,$2B,$EA,$4E,$B1,$FC,$85,$F2,$9E,$B5,$CF,$1E,$18
	.dc.b	$7E,$77,$57,$37,$FA,$26,$7F,$A0,$78,$3F,$E8,$07,$FD,$4D,$3A,$F0
	.dc.b	$31,$8F,$FB,$CA,$FF,$D3,$35,$FD,$1F,$C0,$CD,$EE,$30,$7F,$4B,$BE
	.dc.b	$3D,$6F,$A1,$FF,$58,$4F,$F4,$FF,$73,$31,$8C,$BF,$81,$47,$4C,$BF
	.dc.b	$D6,$55,$FC,$F2,$B5,$CF,$F2,$F2,$F5,$FB,$7C,$B0,$E1,$D1,$FE,$A0
	.dc.b	$CB,$D7,$97,$F8,$DB,$2F,$F0,$2E,$1F,$0A,$B5,$5E,$D6,$C7,$C6,$E8
	.dc.b	$E7,$5F,$D9,$5B,$5F,$C5,$F8,$BC,$C7,$AB,$92,$D7,$F6,$B7,$9E,$3F
	.dc.b	$82,$8F,$0B,$6B,$FD,$79,$87,$D4,$AD,$D7,$C1,$F6,$FE,$13,$4F,$F6
	.dc.b	$D6,$BF,$07,$E8,$3F,$C7,$91,$83,$EC,$B8,$C1,$3B,$F7,$3E,$BC,$E6
	.dc.b	$39,$65,$F0,$23,$F0,$B7,$F1,$9B,$7E,$66,$BF,$ED,$CC,$91,$FE,$3E
	.dc.b	$9F,$EE,$3F,$43,$D5,$F1,$FB,$9F,$82,$8C,$5A,$F8,$FC,$3F,$90,$F2
	.dc.b	$FB,$77,$0F,$65,$6F,$17,$D0,$79,$C9,$3E,$77,$D5,$F4,$1D,$CF,$79
	.dc.b	$63,$9E,$3E,$8F,$9E,$F8,$7F,$11,$F1,$1B,$3A,$C1,$EC,$7C,$52,$47
	.dc.b	$C5,$7E,$47,$CE,$7C,$CF,$9E,$F9,$AD,$9E,$5F,$0D,$F4,$3D,$AF,$A2
	.dc.b	$D8,$ED,$EC,$FD,$12,$3F,$45,$FA,$9F,$33,$F6,$3E,$5B,$EC,$B6,$7C
	.dc.b	$DE,$D7,$CD,$E2,$FB,$AD,$8F,$57,$CC,$7D,$D2,$3F,$75,$FC,$07,$91
	.dc.b	$FB,$6F,$49,$F0,$9F,$86,$8C,$3F,$88,$C3,$5E,$CD,$C3,$FC,$5F,$68
	.dc.b	$FE,$07,$F4,$2F,$99,$FD,$5B,$6B,$F6,$2D,$BF,$D0,$B5,$10,$7C,$6F
	.dc.b	$E1,$E7,$F0,$70,$FA,$07,$CD,$7F,$16,$8D,$F5,$59,$CD,$7E,$03,$07
	.dc.b	$C6,$79,$8F,$81,$F2,$56,$75,$ED,$FA,$08,$FA,$1F,$5E,$FE,$E7,$DE
	.dc.b	$D8,$ED,$B7,$C2,$D3,$E1,$B6,$3C,$5F,$01,$86,$9C,$0F,$37,$BF,$F2
	.dc.b	$BE,$2B,$E5,$B6,$32,$4D,$FF,$7A,$FF,$38,$D7,$FC,$DA,$E7,$C8,$FE
	.dc.b	$4D,$0D,$7E,$A9,$F5,$BD,$9B,$5F,$15,$FA,$E4,$7E,$BB,$F6,$BE,$4B
	.dc.b	$84,$7C,$B6,$E1,$E2,$1E,$A7,$F8,$EB,$E3,$BF,$18,$7E,$13,$E4,$6D
	.dc.b	$FD,$37,$0F,$D5,$F8,$2B,$7C,$F9,$9A,$3E,$0B,$F1,$A7,$E3,$BE,$0A
	.dc.b	$E1,$F7,$7D,$EC,$34,$FC,$04,$7E,$17,$E8,$7F,$CC,$31,$F8,$A6,$E3
	.dc.b	$E9,$3F,$19,$C1,$BF,$85,$FC,$57,$0F,$F4,$69,$FC,$83,$7A,$FF,$C2
	.dc.b	$D8,$7F,$88,$1B,$3F,$ED,$2C,$1F,$0F,$CA,$BC,$0D,$9F,$76,$BF,$9D
	.dc.b	$6F,$0F,$EA,$B7,$FD,$5B,$6F,$F3,$51,$FA,$EF,$8F,$C5,$7D,$1E,$37
	.dc.b	$CD,$BE,$33,$FB,$8B,$70,$FE,$43,$F3,$AF,$8F,$CD,$7F,$7A,$8C,$4E
	.dc.b	$A7,$1F,$21,$97,$5F,$21,$27,$F6,$7F,$42,$E7,$C6,$9F,$2E,$1B,$FC
	.dc.b	$96,$D7,$F9,$19,$B3,$FE,$F8,$6B,$FD,$37,$2F,$43,$CE,$BC,$D9,$FE
	.dc.b	$FE,$B8,$F6,$7F,$BB,$75,$FF,$5E,$E3,$FE,$C2,$6E,$1F,$8B,$C5,$F2
	.dc.b	$1F,$13,$E4,$3E,$03,$2F,$E2,$F2,$5B,$F9,$98,$7F,$A5,$27,$A7,$93
	.dc.b	$F1,$FA,$7E,$6B,$63,$E6,$C7,$CB,$C7,$17,$E1,$AF,$F0,$6E,$5C,$E3
	.dc.b	$FC,$02,$D7,$E0,$46,$BE,$3F,$EA,$3F,$55,$B8,$7F,$22,$FA,$0F,$EE
	.dc.b	$75,$FC,$9D,$AF,$F2,$A6,$A6,$9F,$DD,$DF,$D0,$3F,$DF,$B5,$FF,$2B
	.dc.b	$46,$FD,$FF,$E5,$93,$9A,$FE,$69,$F4,$1F,$E6,$37,$FC,$FE,$DA,$FE
	.dc.b	$70,$7F,$3A,$70,$E1,$AF,$F9,$3B,$2F,$F2,$5B,$FB,$18,$E5,$87,$F8
	.dc.b	$F6,$72,$F9,$23,$35,$C7,$E2,$FE,$59,$1F,$F1,$ED,$7F,$97,$6F,$E6
	.dc.b	$FC,$67,$E7,$6B,$F3,$CD,$AF,$C3,$7C,$57,$E8,$4F,$D3,$95,$FD,$3F
	.dc.b	$C5,$E1,$A7,$03,$F5,$7F,$17,$F7,$08,$F8,$9F,$BF,$F1,$7F,$AD,$7F
	.dc.b	$5C,$D8,$FD,$9F,$C5,$7E,$C4,$FD,$97,$C5,$D8,$FD,$BF,$C5,$FE,$20
	.dc.b	$8F,$B9,$FC,$5B,$FB,$97,$F7,$4D,$3F,$C3,$FD,$BF,$E3,$0F,$08,$D1
	.dc.b	$F1,$7F,$E7,$A7,$1F,$17,$FE,$62,$7F,$17,$3C,$23,$06,$7E,$27,$87
	.dc.b	$37,$16,$E1,$E0,$F8,$9E,$9F,$15,$D5,$B1,$E4,$F8,$9F,$01,$E1,$F8
	.dc.b	$AB,$1E,$8F,$8A,$F1,$11,$FF,$3C,$E8,$8F,$37,$C4,$7A,$7E,$27,$D4
	.dc.b	$DA,$F8,$3F,$13,$EB,$23,$C9,$F7,$7C,$4C,$8E,$AD,$EB,$FF,$6D,$8F
	.dc.b	$8E,$FF,$52,$F9,$FF,$9E,$8C,$1F,$E5,$B7,$11,$A8,$FD,$1B,$57,$C8
	.dc.b	$BD,$AF,$F3,$AC,$71,$3F,$C4,$31,$9F,$0B,$83,$63,$47,$29,$E3,$CC
	.dc.b	$C3,$FD,$73,$BB,$E2,$FF,$DE,$73,$F8,$06,$C6,$6F,$E5,$C6,$3E,$2E
	.dc.b	$47,$CE,$6B,$FE,$96,$97,$C5,$7F,$8D,$73,$FF,$59,$43,$E2,$F1,$BF
	.dc.b	$89,$B7,$FE,$7F,$9E,$2F,$E5,$3E,$2F,$C6,$47,$EE,$BE,$6F,$8A,$E3
	.dc.b	$A7,$E6,$1B,$5F,$EF,$CC,$7C,$3F,$17,$AF,$9B,$FF,$28,$63,$FA,$03
	.dc.b	$66,$7E,$33,$5F,$15,$BF,$8B,$93,$FE,$84,$A4,$F3,$F8,$BF,$93,$1F
	.dc.b	$E6,$49,$FF,$64,$B5,$FF,$5C,$B5,$FF,$52,$3F,$E6,$37,$A7,$C5,$79
	.dc.b	$4F,$37,$C5,$F9,$FE,$2F,$D1,$F1,$72,$BC,$51,$5E,$B1,$E2,$23,$E2
	.dc.b	$F8,$BF,$80,$47,$D5,$F2,$7F,$0F,$E2,$3E,$27,$C4,$49,$F7,$45,$7D
	.dc.b	$31,$EE,$23,$EE,$EC,$FB,$3F,$0B,$FE,$0C,$F3,$6F,$BA,$7E,$84,$E1
	.dc.b	$FA,$5F,$17,$2F,$FB,$94,$7D,$5F,$8C,$FA,$DF,$19,$27,$E8,$45,$7E
	.dc.b	$5C,$7D,$02,$3F,$43,$E3,$3F,$16,$BE,$7D,$EC,$E1,$B1,$F6,$E7,$13
	.dc.b	$FB,$A9,$CC,$FE,$E5,$AF,$F5,$86,$BF,$E9,$1F,$9F,$7C,$07,$81,$7C
	.dc.b	$9D,$31,$16,$7F,$82,$09,$DB,$86,$BF,$C2,$5A,$FE,$23,$E2,$BF,$35
	.dc.b	$F1,$52,$3F,$29,$15,$FC,$7C,$7E,$50,$8F,$E5,$3C,$3F,$D1,$11,$FC
	.dc.b	$E7,$95,$F6,$BE,$75,$EC,$52,$4F,$E8,$6B,$F5,$FF,$37,$FE,$45,$8F
	.dc.b	$F7,$3B,$5F,$F3,$A3,$7C,$57,$95,$FE,$B8,$A4,$BE,$3B,$C1,$EC,$CC
	.dc.b	$3F,$DF,$7D,$4C,$D3,$E0,$F8,$EF,$0F,$95,$F7,$BE,$75,$EA,$53,$38
	.dc.b	$FF,$B0,$A3,$2F,$D0,$6B,$FC,$D5,$AF,$F8,$51,$AF,$9F,$3F,$F3,$33
	.dc.b	$EE,$F8,$FF,$82,$78,$7D,$3F,$87,$F1,$D2,$7E,$1B,$F1,$BE,$3B,$E6
	.dc.b	$7B,$B2,$7D,$BB,$23,$84,$7C,$8F,$8C,$F9,$A4,$7E,$6F,$C6,$D7,$E6
	.dc.b	$FC,$77,$D2,$31,$CD,$FE,$34,$F9,$9F,$F6,$FC,$FF,$B8,$BE,$6F,$F8
	.dc.b	$FB,$FE,$99,$6C,$EF,$CF,$FD,$3C,$6A,$7F,$C6,$AD,$7F,$DD,$CD,$7F
	.dc.b	$DE,$2D,$7F,$DE,$3D,$BF,$B0,$F9,$5F,$F3,$F4,$7F,$BC,$9A,$F8,$7D
	.dc.b	$5F,$C4,$F8,$3F,$8A,$F0,$A4,$FE,$2A,$3F,$1B,$F1,$D2,$3F,$1A,$E9
	.dc.b	$FC,$6F,$8B,$F9,$22,$3F,$92,$F8,$EF,$CD,$73,$48,$F9,$EE,$9F,$CA
	.dc.b	$FC,$77,$E6,$BE,$3A,$47,$E6,$A2,$BF,$E5,$6F,$0B,$F2,$F3,$FD,$9B
	.dc.b	$AC,$D3,$1F,$1D,$FA,$42,$57,$E9,$23,$F4,$FE,$6C,$8F,$D3,$BF,$AA
	.dc.b	$F8,$DC,$47,$EA,$3E,$3B,$F5,$A6,$39,$BF,$53,$0C,$BF,$53,$E3,$FF
	.dc.b	$60,$E8,$7C,$E5,$3F,$6B,$16,$87,$FB,$9E,$7F,$C7,$6E,$2F,$AD,$A7
	.dc.b	$33,$BB,$ED,$F1,$DD,$FC,$A9,$3C,$71,$FD,$E5,$AB,$33,$F2,$63,$F8
	.dc.b	$5B,$FE,$23,$C3,$B5,$FF,$84,$27,$33,$BF,$8E,$F6,$11,$FE,$07,$F3
	.dc.b	$8E,$9F,$E7,$B5,$FD,$A6,$7F,$39,$ED,$FE,$68,$CF,$BB,$F1,$7C,$CE
	.dc.b	$C4,$9F,$15,$7F,$0C,$6F,$C7,$FF,$5A,$C7,$A7,$07,$D0,$F3,$91,$8E
	.dc.b	$0F,$7D,$7D,$1F,$8F,$E1,$A4,$C7,$E2,$1F,$5F,$24,$FD,$7F,$1F,$E0
	.dc.b	$12,$7C,$6F,$C3,$F8,$FF,$17,$B9,$27,$9F,$1F,$D5,$9F,$C0,$7C,$77
	.dc.b	$E0,$7E,$3E,$56,$B2,$4F,$E1,$7E,$3F,$E7,$12,$7E,$0B,$F4,$7C,$0F
	.dc.b	$C5,$7A,$52,$79,$47,$E3,$7E,$2F,$F1,$DF,$1B,$27,$B4,$7E,$47,$E3
	.dc.b	$7E,$D1,$23,$E6,$BF,$77,$E2,$FF,$2D,$F1,$B2,$7C,$D1,$F9,$9F,$8D
	.dc.b	$FF,$04,$FC,$3B,$D6,$A6,$6B,$D8,$7E,$1C,$87,$DA,$3F,$CB,$3F,$19
	.dc.b	$F8,$57,$E3,$B5,$76,$7E,$7A,$30,$DF,$8D,$FF,$24,$C7,$F8,$F2,$7F
	.dc.b	$A8,$DF,$F7,$2A,$FC,$13,$7A,$FF,$AD,$56,$EB,$AC,$63,$E3,$7C,$64
	.dc.b	$8F,$D0,$4F,$F9,$05,$E3,$E0,$78,$08,$FA,$3D,$5F,$1F,$6F,$93,$E3
	.dc.b	$24,$FB,$23,$CF,$97,$C0,$24,$76,$7E,$1F,$C6,$7B,$09,$1E,$CF,$33
	.dc.b	$DF,$F1,$9F,$03,$BE,$4F,$C8,$8A,$FF,$21,$D1,$3F,$38,$91,$EC,$F8
	.dc.b	$BF,$B3,$BC,$8F,$A3,$15,$F8,$A7,$F1,$9A,$7C,$5F,$D0,$35,$C3,$F5
	.dc.b	$7E,$2F,$E9,$FC,$54,$BF,$B7,$15,$E7,$E1,$FF,$9C,$BE,$2F,$EC,$99
	.dc.b	$BF,$E4,$F2,$F9,$B7,$A9,$FC,$A6,$3F,$D6,$DE,$EF,$F2,$47,$51,$9F
	.dc.b	$90,$FE,$AF,$5C,$31,$FE,$53,$18,$E2,$E0,$EA,$B8,$61,$CE,$F8,$61
	.dc.b	$D8,$FE,$62,$E2,$B9,$47,$49,$E1,$5C,$A3,$6D,$7C,$B8,$6D,$E7,$87
	.dc.b	$8A,$B8,$CE,$67,$67,$84,$B7,$E0,$71,$87,$8E,$B8,$C6,$30,$F0,$4F
	.dc.b	$F5,$E8,$E8,$FB,$1F,$90,$76,$2B,$97,$99,$AF,$90,$FC,$4F,$CE,$7F
	.dc.b	$98,$1B,$1F,$EE,$56,$BB,$E3,$4F,$F9,$9A,$9E,$BB,$BF,$0D,$BF,$F7
	.dc.b	$74,$EA,$72,$48,$F7,$CE,$F3,$FB,$BE,$BF,$F6,$C7,$EF,$C6,$74,$FB
	.dc.b	$4D,$AF,$B2,$D7,$F1,$95,$C6,$77,$7F,$1A,$B5,$87,$8E,$FE,$7D,$ED
	.dc.b	$E7,$5B,$23,$AE,$6B,$8F,$92,$6E,$3F,$8D,$BE,$57,$F9,$DC,$7C,$DB
	.dc.b	$7F,$9E,$3E,$27,$FC,$0F,$C3,$FD,$F3,$D6,$46,$34,$FF,$39,$47,$F2
	.dc.b	$F7,$FA,$6B,$A6,$F8,$5B,$C7,$FC,$21,$39,$7F,$62,$F1,$6D,$6F,$0F
	.dc.b	$F4,$4F,$C3,$DA,$3C,$FD,$47,$E8,$4F,$17,$53,$98,$C6,$7E,$48,$C4
	.dc.b	$EA,$3F,$9E,$EF,$FB,$EC,$FF,$99,$B5,$FE,$1C,$DC,$BF,$DE,$EE,$61
	.dc.b	$F3,$2F,$81,$C3,$6F,$FD,$6A,$D9,$FC,$F9,$C2,$7A,$77,$3E,$3F,$E0
	.dc.b	$36,$FE,$5F,$C1,$E2,$7E,$17,$C4,$F8,$F8,$0D,$25,$7F,$89,$63,$C1
	.dc.b	$C7,$FE,$11,$E0,$79,$1F,$2F,$3F,$8C,$F8,$3E,$3E,$61,$FE,$A8,$A7
	.dc.b	$95,$8D,$BF,$93,$ED,$FE,$49,$AF,$EA,$7C,$3E,$A2,$3E,$1F,$83,$FB
	.dc.b	$5F,$65,$FE,$B5,$8F,$C2,$C6,$3E,$43,$3F,$21,$27,$C1,$E5,$DC,$BF
	.dc.b	$9A,$E1,$F8,$77,$F9,$76,$28,$C4,$7F,$B0,$DA,$FF,$97,$5B,$D3,$F8
	.dc.b	$6F,$2B,$C7,$F1,$FF,$85,$6E,$3E,$FF,$27,$F0,$15,$F8,$C6,$F0,$1A
	.dc.b	$7E,$0F,$0F,$CC,$63,$E4,$6B,$E3,$EE,$D1,$CB,$AE,$63,$F8,$2C,$7F
	.dc.b	$33,$18,$AF,$F4,$37,$3B,$E1,$BE,$4F,$77,$E8,$5B,$87,$F8,$EB,$27
	.dc.b	$C7,$78,$C9,$1F,$E6,$58,$CB,$FE,$67,$FC,$A3,$FC,$FA,$F1,$F8,$FF
	.dc.b	$D7,$46,$32,$F9,$F1,$8F,$88,$FB,$3D,$D6,$FC,$0F,$DC,$F0,$3E,$11
	.dc.b	$23,$C0,$FC,$AF,$B3,$F0,$E1,$F6,$BE,$3B,$F7,$11,$8F,$07,$FD,$0F
	.dc.b	$A2,$7C,$7F,$15,$F8,$8F,$0A,$E5,$EF,$7F,$85,$62,$F9,$D7,$B3,$FB
	.dc.b	$36,$D7,$EB,$A7,$53,$93,$C9,$87,$E5,$3E,$3B,$C6,$DA,$F9,$9F,$1F
	.dc.b	$F9,$83,$E2,$FC,$55,$8F,$F3,$9F,$A2,$FD,$0D,$C3,$FC,$1E,$30,$D8
	.dc.b	$CC,$FF,$00,$9C,$BE,$5D,$1E,$E7,$A4,$91,$97,$F8,$87,$37,$FA,$02
	.dc.b	$3F,$01,$E3,$7D,$8F,$8B,$FD,$6B,$F6,$BE,$2B,$13,$FE,$7E,$A4,$63
	.dc.b	$87,$FC,$3F,$0F,$8A,$FD,$AF,$B5,$F8,$26,$DF,$17,$0F,$3F,$89,$B7
	.dc.b	$CB,$E2,$7F,$12,$7E,$2B,$E2,$6D,$78,$5F,$49,$1F,$C9,$3E,$CF,$89
	.dc.b	$E1,$1E,$76,$D7,$9B,$E2,$3F,$84,$FB,$FF,$96,$6D,$7F,$A9,$3C,$77
	.dc.b	$7F,$E9,$E3,$F5,$8D,$BF,$F0,$6C,$EB,$C4,$FC,$A1,$F3,$3E,$1F,$F8
	.dc.b	$AF,$B5,$FA,$26,$E1,$F2,$7E,$1F,$F4,$A7,$E9,$B0,$B1,$FB,$1F,$73
	.dc.b	$F8,$E9,$F9,$4F,$6B,$F9,$9C,$64,$FF,$5C,$CF,$E5,$FC,$EF,$CB,$12
	.dc.b	$3F,$04,$FE,$67,$E0,$A4,$7E,$15,$B3,$FE,$26,$F4,$DF,$98,$FF,$5C
	.dc.b	$8F,$E4,$BA,$FF,$67,$8C,$79,$3F,$AC,$8C,$7B,$B3,$1F,$F6,$0D,$7F
	.dc.b	$B8,$5B,$1C,$7D,$9F,$E6,$F1,$4F,$E1,$C7,$F6,$FE,$C7,$C7,$7F,$26
	.dc.b	$6C,$7F,$1B,$6E,$5F,$41,$AF,$F3,$8F,$92,$7C,$D7,$F7,$D6,$D7,$93
	.dc.b	$DF,$C4,$FE,$FB,$E7,$9F,$01,$FD,$BD,$B5,$FB,$D6,$E1,$FD,$B1,$AF
	.dc.b	$F9,$33,$C6,$FE,$D4,$FF,$70,$6B,$9F,$81,$FE,$10,$7F,$7D,$DF,$FA
	.dc.b	$FD,$FF,$B9,$B3,$FE,$49,$C2,$85,$B8,$BE,$9F,$C0,$FA,$FE,$4F,$E3
	.dc.b	$7C,$8F,$C5,$25,$FC,$57,$B7,$95,$F0,$7D,$7F,$B0,$D9,$FC,$E7,$1F
	.dc.b	$FB,$C1,$F9,$9E,$37,$CF,$23,$FC,$77,$E7,$BF,$9C,$69,$FE,$6A,$6B
	.dc.b	$FE,$93,$6B,$E8,$6A,$58,$7F,$86,$DB,$3D,$63,$07,$C5,$F8,$2F,$C7
	.dc.b	$1F,$E3,$1F,$D1,$1F,$25,$8A,$FD,$A4,$61,$FC,$7C,$EE,$75,$87,$FA
	.dc.b	$E8,$FF,$39,$65,$FE,$2B,$F8,$8F,$E2,$F7,$FC,$E5,$B3,$FE,$28,$6B
	.dc.b	$FE,$88,$7F,$8D,$46,$FC,$CE,$44,$7F,$C5,$FC,$2E,$37,$C0,$7F,$8E
	.dc.b	$DB,$3A,$6E,$1F,$AF,$6B,$FE,$BD,$F1,$FF,$95,$F9,$9F,$A1,$25,$7E
	.dc.b	$85,$FC,$57,$C5,$FE,$CC,$FD,$46,$B6,$7E,$EB,$6F,$FD,$B7,$7F,$CD
	.dc.b	$A3,$1C,$FF,$8F,$9F,$C7,$62,$F7,$3D,$04,$C3,$D0,$FE,$CF,$E4,$A4
	.dc.b	$7E,$CD,$B5,$EB,$F0,$7F,$60,$7F,$07,$F1,$6D,$7F,$04,$F1,$BF,$44
	.dc.b	$47,$E4,$F6,$FF,$88,$5B,$F1,$DF,$A7,$23,$83,$E5,$7B,$3F,$BD,$3F
	.dc.b	$8A,$78,$76,$3F,$88,$F8,$3F,$AF,$23,$D9,$FF,$37,$7E,$5B,$FE,$47
	.dc.b	$D3,$FD,$EB,$8F,$F9,$3D,$FF,$7B,$C6,$78,$3E,$37,$37,$F2,$43,$FD
	.dc.b	$11,$1F,$E8,$D8,$CB,$BC,$3F,$73,$28,$D4,$61,$D5,$7D,$0F,$5F,$F8
	.dc.b	$B1,$D3,$E1,$AD,$7F,$05,$F0,$BF,$14,$47,$F1,$4F,$AB,$E0,$FF,$8F
	.dc.b	$76,$F9,$1B,$3F,$D6,$7F,$35,$57,$F9,$41,$B1,$FE,$45,$6E,$3F,$D6
	.dc.b	$9A,$FE,$4F,$F4,$5F,$F6,$D4,$FF,$92,$DB,$1F,$D8,$1A,$EF,$BF,$FD
	.dc.b	$30,$FF,$1A,$F8,$FF,$87,$C5,$FE,$94,$6C,$7F,$BD,$1A,$FF,$20,$0E
	.dc.b	$3E,$45,$F1,$8E,$2C,$36,$BF,$93,$5F,$1A,$DD,$68,$EF,$DC,$FD,$19
	.dc.b	$F0,$BF,$E3,$98,$C3,$63,$38,$7F,$9A,$5A,$FF,$7E,$FC,$DD,$C7,$C5
	.dc.b	$B1,$FE,$6F,$6B,$B8,$FF,$23,$73,$7E,$0F,$E1,$3D,$D9,$7E,$33,$C1
	.dc.b	$FF,$13,$5F,$F4,$3F,$1B,$3C,$7F,$87,$BB,$3F,$B6,$F2,$F9,$D9,$A4
	.dc.b	$8F,$E1,$DE,$FF,$0C,$BF,$8A,$36,$8E,$7A,$FD,$B9,$FE,$6B,$E0,$B1
	.dc.b	$F0,$33,$1F,$EA,$73,$96,$E9,$F9,$B9,$E5,$F0,$12,$72,$7F,$59,$C1
	.dc.b	$3D,$FD,$EF,$87,$06,$92,$7F,$93,$79,$7E,$72,$3D,$35,$FD,$67,$B9
	.dc.b	$E9,$CB,$99,$27,$F9,$9E,$1F,$EE,$9A,$7C,$7F,$F8,$12,$7F,$9B,$B5
	.dc.b	$FF,$10,$35,$FF,$60,$B5,$FD,$DE,$EF,$82,$F9,$44,$8F,$24,$BD,$5F
	.dc.b	$AF,$D3,$FE,$64,$F7,$FF,$1D,$E2,$1A,$4B,$FE,$89,$E7,$FD,$C2,$38
	.dc.b	$3F,$C4,$07,$DB,$C3,$F9,$B3,$E0,$F3,$FB,$DF,$EA,$63,$F4,$DB,$7F
	.dc.b	$12,$AD,$C6,$0F,$F0,$63,$8E,$4F,$E3,$AD,$9F,$F0,$8E,$5F,$E7,$CA
	.dc.b	$F5,$FE,$60,$F7,$BF,$50,$D8,$FF,$67,$35,$FF,$31,$B5,$FF,$1A,$7C
	.dc.b	$1F,$E6,$08,$FF,$63,$7F,$D5,$4B,$E8,$5F,$29,$86,$C7,$FA,$BD,$AF
	.dc.b	$BC,$E1,$1B,$CA,$46,$DA,$FF,$9D,$BF,$18,$FF,$61,$3F,$63,$D0,$F9
	.dc.b	$87,$D8,$8C,$C7,$C7,$E5,$FC,$4C,$EA,$38,$1F,$E8,$59,$DE,$DF,$DA
	.dc.b	$E1,$F4,$5C,$9F,$A8,$D7,$F8,$F3,$5F,$F1,$FB,$5E,$3F,$1D,$F1,$20
	.dc.b	$D2,$47,$D0,$F8,$3F,$A6,$47,$07,$E7,$BC,$FF,$E0,$1E,$C4,$8F,$CB
	.dc.b	$C7,$DC,$F7,$FE,$D9,$23,$ED,$BF,$C6,$35,$FE,$2B,$1F,$0F,$AD,$F0
	.dc.b	$FF,$C2,$CF,$CA,$78,$FF,$C5,$A5,$F8,$DF,$F8,$E1,$FF,$73,$35,$FB
	.dc.b	$0D,$7F,$30,$D7,$FA,$17,$D1,$FF,$16,$7F,$DD,$CD,$7F,$00,$D7,$F4
	.dc.b	$0D,$7F,$6F,$DB,$FC,$7C,$8F,$ED,$BC,$8F,$CD,$FC,$DF,$E6,$89,$1F
	.dc.b	$9A,$7E,$07,$89,$FC,$9F,$D3,$93,$FB,$E8,$FD,$97,$83,$23,$F8,$0B
	.dc.b	$FB,$5F,$0F,$11,$F1,$CF,$E6,$06,$39,$BF,$0D,$A7,$CA,$8F,$8B,$B7
	.dc.b	$E8,$A3,$11,$F1,$A3,$1A,$48,$F9,$6D,$7F,$D7,$1F,$15,$FB,$1B,$FF
	.dc.b	$0B,$B7,$C1,$FF,$09,$B6,$7F,$D1,$BC,$2E,$3F,$E1,$98,$49,$FE,$36
	.dc.b	$FF,$0F,$E5,$FE,$21,$97,$F4,$BC,$1F,$41,$4C,$DF,$FB,$9F,$C0,$7C
	.dc.b	$BF,$F0,$62,$47,$F0,$67,$F1,$8F,$3F,$8B,$8F,$F2,$A3,$FA,$C7,$E8
	.dc.b	$FD,$9F,$E8,$0B,$E5,$32,$6B,$0B,$3F,$E7,$29,$D3,$67,$FB,$8C,$62
	.dc.b	$3F,$DB,$CA,$BB,$92,$3B,$CF,$8B,$E2,$A4,$78,$9F,$2F,$6E,$23,$F8
	.dc.b	$36,$7D,$78,$BE,$8D,$E5,$FF,$24,$E2,$F9,$EF,$6F,$0C,$C3,$DB,$15
	.dc.b	$E7,$8F,$CF,$31,$7F,$C3,$7E,$8D,$1F,$CA,$1B,$3F,$C8,$C3,$87,$F9
	.dc.b	$33,$5F,$F1,$77,$C9,$BF,$48,$7C,$37,$18,$C3,$63,$33,$FC,$BE,$72
	.dc.b	$FF,$A8,$35,$6D,$F8,$52,$47,$D8,$F8,$BF,$EC,$7F,$6B,$FE,$24,$D3
	.dc.b	$F2,$BF,$45,$F8,$B3,$3E,$DF,$F0,$9D,$7E,$36,$1E,$B6,$BE,$AD,$FE
	.dc.b	$01,$F7,$3E,$33,$F2,$9C,$32,$7F,$0F,$1E,$0F,$9B,$FC,$C5,$49,$FF
	.dc.b	$0A,$FA,$4D,$15,$F9,$DD,$7F,$C6,$FB,$9A,$4B,$FC,$77,$D1,$7E,$8F
	.dc.b	$C5,$91,$F9,$68,$FC,$CF,$C6,$FE,$9F,$49,$3F,$9A,$8F,$D5,$7C,$2F
	.dc.b	$62,$47,$CB,$7F,$2F,$F1,$BF,$AF,$B9,$3F,$A2,$8F,$97,$ED,$FF,$96
	.dc.b	$75,$34,$93,$F5,$7A,$3D,$64,$70,$7F,$69,$DD,$FA,$4F,$F8,$7E,$11
	.dc.b	$8F,$96,$FF,$33,$4E,$63,$13,$C6,$3F,$DF,$AF,$F8,$D7,$81,$AF,$F0
	.dc.b	$73,$FC,$47,$3A,$9F,$D3,$E7,$F3,$88,$FE,$2D,$FA,$3E,$07,$F0,$F2
	.dc.b	$47,$D8,$97,$C8,$7F,$97,$7F,$34,$FF,$01,$F5,$7F,$20,$F9,$1F,$E2
	.dc.b	$C6,$7D,$9F,$E3,$1F,$17,$F7,$09,$8F,$E6,$5F,$BF,$ED,$7F,$1D,$24
	.dc.b	$77,$F5,$BB,$7D,$07,$AB,$D4,$93,$FE,$53,$CD,$3E,$3F,$A0,$FF,$3F
	.dc.b	$C1,$A4,$9F,$F0,$C6,$0F,$94,$7D,$1F,$F9,$DD,$B3,$FE,$5F,$6B,$FE
	.dc.b	$13,$B7,$C9,$FF,$6F,$E5,$F0,$CF,$F0,$16,$BF,$E2,$56,$BF,$E4,$56
	.dc.b	$BF,$ED,$6A,$7D,$1F,$FA,$59,$B1,$E6,$6B,$FE,$4D,$C7,$D1,$3F,$84
	.dc.b	$F5,$FE,$97,$D0,$CA,$FA,$D1,$5F,$44,$5E,$37,$AD,$FE,$4E,$E6,$FF
	.dc.b	$02,$47,$FB,$BB,$57,$D0,$FD,$DF,$92,$98,$FC,$F8,$AF,$C3,$8B,$5C
	.dc.b	$67,$FC,$25,$F4,$8F,$8E,$6D,$FE,$D3,$EB,$FF,$67,$4E,$63,$18,$84
	.dc.b	$65,$FF,$69,$CF,$0F,$8E,$FF,$6B,$7C,$AF,$F5,$E9,$E1,$5B,$8C,$47
	.dc.b	$2C,$3C,$2F,$23,$6E,$23,$73,$FE,$A6,$9F,$19,$C7,$0B,$1D,$70,$F4
	.dc.b	$DF,$17,$C9,$5A,$8C,$3D,$AB,$07,$2C,$3D,$35,$CA,$33,$3E,$53,$C0
	.dc.b	$5B,$F2,$36,$FD,$6F,$FA,$3E,$3E,$33,$63,$2F,$53,$FC,$ED,$F2,$AF
	.dc.b	$21,$E9,$7F,$57,$8C,$36,$33,$3C,$1A,$F7,$F9,$07,$C1,$7F,$5F,$C7
	.dc.b	$FD,$94,$D8,$CC,$FF,$A1,$67,$2D,$8C,$EB,$F4,$1B,$A7,$E1,$3C,$8F
	.dc.b	$CA,$D7,$C1,$9C,$BF,$79,$AF,$E1,$9B,$7F,$CC,$B9,$3E,$BC,$60,$98
	.dc.b	$FD,$AB,$EB,$8F,$86,$36,$4C,$3E,$84,$EB,$1F,$CB,$DF,$29,$F2,$57
	.dc.b	$3C,$38,$61,$F9,$3D,$BF,$86,$7A,$5E,$AE,$2F,$5C,$7F,$0F,$F0,$B1
	.dc.b	$7F,$C0,$6A,$E3,$FA,$6E,$AF,$F5,$8C,$FF,$11,$9C,$DF,$48,$FE,$29
	.dc.b	$F0,$B8,$A5,$72,$F2,$1F,$49,$FE,$C7,$BF,$E3,$5E,$16,$63,$19,$7F
	.dc.b	$AB,$70,$FE,$53,$5F,$DD,$27,$3A,$7D,$2C,$39,$FC,$6F,$7F,$0F,$93
	.dc.b	$E8,$C5,$E0,$FF,$A9,$B1,$FE,$45,$F1,$1F,$EB,$39,$FD,$E1,$A7,$0F
	.dc.b	$EF,$E7,$83,$FC,$77,$3F,$E5,$B5,$FC,$0F,$2F,$F5,$E5,$69,$F9,$73
	.dc.b	$AA,$FF,$76,$1D,$6B,$F9,$C7,$CA,$7C,$53,$E3,$7A,$96,$3D,$DE,$DF
	.dc.b	$8B,$5F,$8E,$DD,$BF,$C7,$B3,$FD,$D5,$B8,$79,$74,$FA,$07,$D1,$F0
	.dc.b	$ED,$7C,$9F,$6B,$E7,$1F,$57,$C2,$B1,$F3,$7D,$AE,$C4,$7B,$3F,$13
	.dc.b	$9B,$FA,$A7,$C9,$FD,$D3,$EF,$78,$96,$BE,$B7,$B1,$F6,$CF,$C1,$78
	.dc.b	$B6,$3E,$D7,$B1,$EB,$23,$EB,$7D,$DE,$C7,$E0,$CF,$C5,$78,$F6,$7F
	.dc.b	$B0,$78,$3F,$87,$F5,$BF,$1C,$D8,$FC,$37,$AB,$F1,$C8,$FC,$77,$D5
	.dc.b	$D3,$FB,$3E,$47,$57,$E2,$F9,$BF,$D2,$B7,$FD,$B7,$C5,$FC,$A7,$8B
	.dc.b	$F9,$B8,$FC,$E7,$D0,$57,$FB,$93,$F9,$9E,$4F,$AC,$48,$FA,$CF,$C9
	.dc.b	$FA,$27,$CA,$7E,$A1,$B3,$F8,$B3,$FA,$74,$A1,$EC,$EE,$BF,$55,$F4
	.dc.b	$4F,$1F,$97,$CF,$E1,$B8,$FF,$96,$A7,$F6,$2F,$F8,$ED,$BB,$FE,$80
	.dc.b	$FF,$1D,$B8,$7F,$47,$E0,$A3,$11,$FD,$86,$31,$E3,$78,$8F,$57,$15
	.dc.b	$BF,$07,$87,$ED,$F9,$6F,$3B,$63,$DD,$87,$F6,$BF,$A1,$FC,$CB,$E3
	.dc.b	$6B,$FB,$46,$BB,$F1,$7F,$49,$9B,$C4,$F8,$67,$CB,$F7,$2E,$1D,$B8
	.dc.b	$FE,$57,$CA,$7C,$F3,$F8,$DF,$A4,$FA,$37,$95,$9B,$F2,$E8,$D7,$FD
	.dc.b	$C5,$1F,$EA,$96,$C7,$C9,$35,$F0,$7F,$C2,$1B,$3F,$EA,$3E,$3F,$F1
	.dc.b	$CB,$D2,$B8,$F6,$FF,$4A,$CB,$FC,$D7,$ED,$FF,$17,$9C,$D7,$E0,$BA
	.dc.b	$9F,$31,$F8,$7F,$9F,$FC,$43,$67,$DD,$3A,$6C,$7E,$37,$DF,$DC,$6B
	.dc.b	$E5,$7F,$20,$7E,$47,$E0,$2C,$72,$6E,$DF,$A4,$8F,$14,$63,$5F,$E8
	.dc.b	$F8,$BE,$53,$F3,$5F,$37,$F9,$B6,$DF,$93,$E5,$70,$D3,$CF,$1F,$A1
	.dc.b	$F8,$1F,$D1,$4D,$9D,$65,$E1,$3F,$2B,$E2,$7E,$9F,$E6,$BF,$42,$D9
	.dc.b	$F8,$3F,$27,$86,$9F,$A1,$8F,$D7,$7C,$A7,$EB,$FE,$52,$CF,$C8,$7F
	.dc.b	$65,$84,$8F,$97,$F0,$16,$3F,$9C,$FB,$7F,$EC,$29,$FC,$2B,$63,$F9
	.dc.b	$EF,$37,$FB,$9A,$73,$E1,$FD,$BF,$3B,$F0,$93,$FE,$A0,$9F,$C5,$46
	.dc.b	$0F,$C4,$5F,$F9,$72,$3F,$0D,$1F,$D6,$63,$17,$DA,$3F,$B7,$FD,$07
	.dc.b	$F6,$2A,$FE,$0E,$D9,$FF,$2D,$AF,$94,$DF,$BD,$C1,$B5,$FA,$59,$D3
	.dc.b	$67,$FB,$B7,$57,$FA,$AA,$35,$F2,$98,$F1,$7C,$2F,$8B,$E5,$31,$3C
	.dc.b	$DB,$97,$D9,$F1,$99,$7F,$56,$EB,$FF,$07,$B6,$FF,$B8,$FD,$1B,$E3
	.dc.b	$7D,$BE,$5F,$E8,$FE,$3B,$E4,$92,$BF,$64,$FF,$76,$F1,$7F,$DA,$7E
	.dc.b	$FE,$E7,$F8,$6C,$61,$FD,$84,$FB,$3C,$4F,$93,$F0,$56,$7E,$13,$F3
	.dc.b	$3E,$07,$1B,$FA,$75,$46,$36,$FF,$08,$71,$BC,$5F,$A7,$E5,$7C,$86
	.dc.b	$E5,$F3,$7E,$4B,$0D,$3F,$90,$F9,$7F,$BC,$9F,$F0,$B7,$D1,$BD,$6F
	.dc.b	$E1,$4D,$AF,$E1,$DA,$A3,$0F,$F1,$16,$BF,$CE,$DF,$F6,$44,$7F,$1A
	.dc.b	$77,$9F,$F9,$62,$79,$7A,$9E,$0B,$FF,$1A,$7D,$1B,$C9,$FC,$6F,$C1
	.dc.b	$E1,$B9,$7A,$27,$4D,$8F,$F2,$0F,$E9,$5F,$A3,$CB,$FC,$69,$5A,$CB
	.dc.b	$FC,$43,$E2,$BC,$FF,$E4,$CD,$C7,$DA,$D9,$FE,$58,$D7,$FA,$B3,$F3
	.dc.b	$E3,$F9,$AB,$BB,$F6,$39,$9E,$F5,$F9,$AF,$37,$FC,$55,$C7,$FC,$5F
	.dc.b	$0F,$E7,$1C,$3F,$CB,$27,$27,$F8,$77,$95,$F0,$7C,$6B,$F4,$CD,$C7
	.dc.b	$F8,$A5,$FF,$34,$DB,$F6,$7F,$1B,$FA,$86,$DF,$DE,$CB,$E1,$6B,$F8
	.dc.b	$89,$FD,$CF,$C2,$6E,$B5,$E7,$79,$88,$F9,$AD,$1E,$5F,$95,$FD,$AB
	.dc.b	$FC,$09,$B3,$FB,$BF,$27,$F8,$29,$FC,$1B,$E1,$2C,$7E,$FF,$E5,$3D
	.dc.b	$C4,$7F,$AD,$78,$DF,$B6,$8F,$85,$F0,$5F,$C3,$9F,$E1,$ED,$9F,$E1
	.dc.b	$1E,$37,$F1,$23,$F8,$9F,$87,$63,$F8,$65,$FC,$B2,$3D,$B9,$3F,$DD
	.dc.b	$EF,$83,$E8,$3F,$B2,$F2,$7F,$07,$6F,$17,$FA,$51,$FF,$47,$C6,$7D
	.dc.b	$9F,$F4,$7C,$62,$75,$F1,$D8,$F9,$4C,$BA,$F8,$F9,$8F,$F8,$0E,$FF
	.dc.b	$C5,$0D,$8E,$7F,$1F,$FE,$5D,$9F,$F4,$7E,$5F,$E2,$5D,$11,$88,$FF
	.dc.b	$43,$C6,$27,$87,$CA,$F9,$0F,$2F,$C2,$DC,$7F,$C1,$7E,$43,$E3,$FF
	.dc.b	$C0,$CD,$8F,$DE,$B7,$1F,$F0,$63,$5F,$F1,$B3,$FE,$57,$6F,$C5,$72
	.dc.b	$24,$FF,$06,$80,$F9,$1F,$F2,$5F,$27,$F8,$5D,$B1,$FE,$41,$6B,$FE
	.dc.b	$A2,$6F,$C9,$BB,$7F,$CC,$B2,$DB,$F9,$B3,$DA,$75,$C7,$C2,$B7,$5A
	.dc.b	$A5,$69,$E5,$87,$FA,$C3,$37,$A3,$E9,$6C,$FC,$48,$C1,$9F,$87,$91
	.dc.b	$F2,$1A,$FD,$7E,$2F,$F1,$DB,$F7,$FE,$17,$FD,$7B,$A7,$E2,$DB,$B7
	.dc.b	$F9,$BA,$9E,$EE,$2F,$DD,$1F,$E4,$D7,$F2,$D8,$7F,$32,$D7,$F0,$6D
	.dc.b	$9D,$4F,$87,$87,$95,$7E,$21,$B3,$F8,$19,$FC,$97,$C9,$63,$49,$1F
	.dc.b	$95,$77,$F2,$5C,$3E,$4A,$4F,$E6,$5D,$BF,$A7,$F9,$2F,$11,$1F,$D6
	.dc.b	$3E,$5E,$5F,$09,$23,$F4,$5F,$25,$E4,$F8,$5F,$2F,$04,$9F,$D3,$45
	.dc.b	$7E,$64,$7B,$C8,$FF,$08,$F8,$FF,$F4,$DB,$FE,$6F,$6B,$FE,$01,$6F
	.dc.b	$1A,$7F,$D3,$B1,$FE,$9C,$9E,$1E,$9F,$D1,$F3,$3F,$CB,$FF,$0D,$7C
	.dc.b	$B3,$F3,$2F,$FB,$35,$FF,$2D,$8F,$EC,$B5,$C2,$3F,$94,$46,$2B,$8C
	.dc.b	$7F,$2B,$3A,$10,$F4,$7A,$B5,$F4,$7C,$7F,$73,$19,$FA,$63,$F4,$58
	.dc.b	$BE,$0B,$D1,$F3,$4F,$84,$A7,$9E,$2D,$7F,$50,$9D,$36,$7D,$31,$88
	.dc.b	$FF,$46,$7D,$03,$C8,$FC,$A4,$21,$E4,$61,$B1,$FE,$98,$78,$36,$3F
	.dc.b	$82,$46,$A3,$AC,$60,$F9,$3F,$41,$F9,$AC,$FD,$DC,$08,$C7,$17,$FB
	.dc.b	$05,$B7,$FC,$0F,$F4,$BF,$F6,$93,$FE,$1C,$8C,$F7,$7F,$B5,$7E,$4B
	.dc.b	$EC,$1F,$76,$38,$CB,$E8,$7F,$6E,$FF,$9F,$CC,$78,$3F,$02,$36,$6B
	.dc.b	$3F,$F0,$D3,$FB,$09,$FF,$6D,$35,$FE,$42,$72,$3F,$C5,$2E,$5F,$B3
	.dc.b	$F2,$5F,$63,$E1,$E4,$7E,$66,$2B,$F9,$9F,$86,$FC,$81,$8B,$FC,$16
	.dc.b	$7F,$EF,$1E,$8F,$C0,$F8,$F2,$FF,$45,$15,$FD,$17,$BB,$F9,$93,$17
	.dc.b	$FE,$C1,$FB,$A7,$A5,$F7,$BE,$87,$FC,$65,$7F,$ED,$26,$BF,$EA,$3F
	.dc.b	$89,$57,$08,$D7,$BB,$FD,$06,$B7,$1F,$EF,$D6,$BF,$11,$AF,$E6,$BE
	.dc.b	$1B,$F5,$84,$7F,$DC,$3F,$44,$F9,$4F,$D8,$1F,$9E,$F8,$29,$3F,$BF
	.dc.b	$8A,$FE,$FF,$E0,$FF,$6C,$63,$5F,$8F,$E1,$3E,$53,$F7,$5A,$4B,$FE
	.dc.b	$11,$15,$FE,$11,$16,$B8,$CF,$E8,$BE,$A1,$F1,$FF,$1F,$E5,$26,$1F
	.dc.b	$EF,$89,$CB,$F2,$63,$E2,$11,$FF,$35,$EE,$8F,$ED,$1F,$46,$E6,$FD
	.dc.b	$7B,$6B,$FB,$5E,$5F,$E6,$98,$C4,$7E,$5E,$FF,$86,$BF,$C7,$FE,$1F
	.dc.b	$F8,$B1,$1F,$E4,$9F,$2D,$FC,$67,$CC,$FE,$35,$94,$9F,$E5,$11,$FC
	.dc.b	$77,$E2,$BF,$85,$FC,$2C,$9F,$E5,$91,$5F,$E5,$87,$F2,$33,$19,$FE
	.dc.b	$53,$E8,$FF,$85,$37,$E5,$7F,$A9,$C7,$FB,$0E,$7D,$7E,$3F,$F8,$1B
	.dc.b	$BD,$3C,$23,$0E,$E7,$FD,$74,$FC,$C6,$F1,$74,$BE,$AF,$6F,$95,$93
	.dc.b	$D8,$FF,$04,$57,$1C,$7C,$DE,$1F,$93,$E5,$25,$6F,$3F,$1C,$E2,$3C
	.dc.b	$33,$6B,$8C,$6E,$DF,$27,$EE,$24,$67,$E5,$2B,$EA,$E6,$F8,$66,$2F
	.dc.b	$F6,$5E,$E7,$F2,$EF,$A2,$67,$4F,$F0,$27,$87,$FB,$09,$C6,$1F,$EB
	.dc.b	$4A,$DB,$6B,$FC,$65,$18,$8F,$F2,$56,$6F,$94,$FA,$64,$8F,$F6,$BF
	.dc.b	$D1,$BE,$43,$FB,$17,$D1,$BE,$3A,$92,$BE,$07,$8B,$E7,$DF,$B9,$23
	.dc.b	$C8,$FF,$64,$A7,$C6,$7E,$0B,$E5,$25,$7F,$B4,$A7,$2D,$F8,$8F,$09
	.dc.b	$1F,$9D,$F1,$9E,$4F,$8A,$E8,$48,$FF,$55,$D7,$F3,$A7,$F9,$8F,$B0
	.dc.b	$87,$C4,$7E,$4B,$E5,$25,$78,$63,$F2,$BF,$29,$23,$EE,$3F,$98,$E0
	.dc.b	$F8,$7C,$DE,$B2,$4F,$81,$FF,$1A,$72,$3C,$2F,$CF,$FC,$A4,$AF,$C3
	.dc.b	$47,$E8,$BE,$53,$E6,$FC,$4C,$9F,$C5,$47,$D0,$F8,$8F,$90,$48,$F6
	.dc.b	$BD,$BE,$3F,$FC,$2D,$F4,$7F,$E1,$99,$FF,$61,$B7,$BF,$F6,$53,$AF
	.dc.b	$8A,$95,$FF,$00,$2B,$FC,$32,$FE,$B3,$C4,$E7,$F1,$5F,$E1,$A9,$34
	.dc.b	$93,$FE,$D8,$8F,$F0,$83,$CF,$E2,$7C,$67,$F9,$43,$E8,$FF,$A2,$43
	.dc.b	$2F,$EA,$F1,$FE,$01,$6F,$C3,$6C,$E1,$F1,$1E,$CF,$82,$F1,$12,$FC
	.dc.b	$0F,$97,$E0,$BE,$09,$23,$F7,$18,$7F,$80,$9F,$07,$9F,$FC,$36,$1F
	.dc.b	$2B,$E5,$24,$F0,$F1,$7E,$5F,$BF,$F0,$49,$1E,$F7,$FC,$E1,$4F,$95
	.dc.b	$F7,$12,$7C,$3E,$2F,$C1,$F1,$3F,$C4,$FF,$37,$E0,$9F,$E2,$41,$C3
	.dc.b	$E6,$3F,$D3,$97,$FE,$82,$77,$1A,$7F,$C1,$67,$1C,$BF,$D5,$D1,$98
	.dc.b	$FF,$07,$B5,$78,$6E,$3F,$EF,$3C,$5F,$E2,$C6,$ED,$FE,$B6,$9D,$7B
	.dc.b	$BF,$67,$C5,$FC,$6B,$F6,$FE,$1E,$63,$F8,$98,$FB,$DF,$25,$23,$E8
	.dc.b	$BF,$81,$F1,$71,$1E,$DE,$A6,$FF,$E9,$18,$C3,$75,$FC,$2D,$FF,$93
	.dc.b	$BE,$8D,$F2,$5F,$EB,$FB,$FF,$65,$78,$5F,$E2,$C9,$CE,$3F,$EC,$EF
	.dc.b	$97,$FD,$85,$7F,$B4,$27,$2F,$03,$1F,$29,$FE,$CD,$BE,$58,$75,$BE
	.dc.b	$34,$BE,$31,$86,$D7,$86,$F9,$57,$7B,$E5,$3D,$30,$F2,$36,$7C,$47
	.dc.b	$5A,$E2,$F7,$AE,$2F,$3A,$E2,$D3,$8D,$AB,$8B,$E4,$8D,$61,$E7,$BE
	.dc.b	$71,$DC,$E6,$77,$BF,$C5,$BD,$4F,$C6,$D7,$C8,$6C,$FD,$17,$F2,$3F
	.dc.b	$2E,$F3,$9E,$AF,$29,$FF,$76,$C6,$1E,$B3,$B9,$D3,$63,$C3,$6F,$95
	.dc.b	$7B,$BC,$E3,$0D,$8C,$CF,$06,$BF,$E6,$EC,$9F,$9D,$73,$E7,$F9,$31
	.dc.b	$8D,$3F,$D5,$1A,$FE,$1A,$37,$86,$37,$FF,$09,$61,$F9,$0C,$38,$BC
	.dc.b	$F1,$FC,$55,$F3,$9C,$DF,$19,$4E,$5D,$FE,$98,$F2,$F5,$28,$C5,$F9
	.dc.b	$97,$95,$A7,$37,$CA,$F0,$66,$6D,$71,$8B,$3E,$09,$B9,$FB,$F2,$E7
	.dc.b	$C9,$FC,$CB,$CE,$FF,$65,$3F,$E8,$A8,$CE,$3F,$C9,$E7,$2F,$C4,$C6
	.dc.b	$E3,$EE,$AD,$B6,$73,$9F,$D2,$AB,$F9,$F3,$E9,$73,$D1,$F9,$9B,$FF
	.dc.b	$17,$4E,$5F,$E7,$3C,$FC,$2A,$47,$F8,$E7,$B5,$DA,$F7,$33,$A7,$F0
	.dc.b	$AE,$5F,$F7,$23,$F8,$3C,$FE,$D3,$FC,$A3,$5F,$E2,$33,$A8,$FE,$88
	.dc.b	$D7,$EC,$E8,$F3,$7C,$F3,$EA,$6E,$7E,$8F,$99,$F5,$F3,$7B,$9B,$1F
	.dc.b	$E4,$9F,$23,$FB,$4B,$FA,$EC,$BF,$C5,$5E,$77,$76,$C6,$67,$79,$F9
	.dc.b	$FF,$4D,$61,$F3,$1B,$5F,$E5,$57,$FE,$00,$36,$3E,$7F,$CB,$7C,$EE
	.dc.b	$4B,$3F,$D9,$1F,$A3,$D3,$E1,$E0,$FC,$7C,$F9,$5B,$8F,$C7,$6B,$F2
	.dc.b	$E3,$EC,$9F,$82,$E3,$FA,$7E,$CE,$23,$FC,$D1,$FA,$7F,$DB,$AF,$E8
	.dc.b	$0D,$EA,$FF,$7D,$FE,$9E,$F9,$6F,$C3,$CE,$6F,$FD,$05,$C0,$DB,$E3
	.dc.b	$36,$3E,$34,$60,$F9,$5C,$CF,$3F,$9F,$D4,$7E,$41,$B5,$F8,$9E,$AF
	.dc.b	$F2,$E3,$FC,$83,$7F,$C5,$FB,$5F,$81,$6C,$66,$77,$BF,$E2,$74,$FC
	.dc.b	$24,$FD,$B6,$C6,$7A,$7E,$E4,$62,$3F,$8A,$3F,$73,$83,$FD,$EE,$DA
	.dc.b	$FF,$50,$68,$F8,$BF,$87,$D3,$F1,$39,$E5,$7E,$9B,$E2,$EE,$5F,$DA
	.dc.b	$F8,$9F,$2F,$F1,$89,$5F,$8B,$7C,$7E,$CF,$CE,$3F,$67,$E3,$D9,$FD
	.dc.b	$8D,$FD,$2E,$5F,$DB,$B6,$3E,$DF,$AF,$EC,$23,$F0,$5F,$7F,$AF,$F6
	.dc.b	$8F,$C2,$7E,$75,$FE,$25,$9F,$C1,$C7,$DE,$E9,$91,$FB,$B9,$FD,$DB
	.dc.b	$F8,$88,$FE,$13,$A3,$D1,$FC,$09,$27,$F4,$CF,$CE,$D7,$F1,$27,$F1
	.dc.b	$0F,$D2,$9F,$15,$8A,$FF,$5D,$47,$F0,$77,$F2,$BD,$7B,$F1,$BF,$22
	.dc.b	$FF,$1C,$DB,$F9,$B7,$9B,$FC,$0D,$B5,$FD,$77,$C3,$FF,$3E,$36,$75
	.dc.b	$0F,$7B,$93,$AA,$FF,$1E,$1F,$D0,$63,$33,$E7,$F8,$AF,$D0,$FC,$B5
	.dc.b	$C3,$F5,$2D,$BF,$F1,$46,$7F,$EC,$C9,$F5,$FC,$4E,$7B,$BF,$17,$F2
	.dc.b	$DF,$AA,$3E,$17,$AD,$6B,$E0,$B8,$7F,$CA,$3F,$A8,$79,$69,$1F,$EF
	.dc.b	$1C,$3F,$BE,$3F,$BA,$F9,$6F,$DA,$9F,$37,$E2,$2C,$7C,$C7,$0F,$CE
	.dc.b	$F1,$EE,$9E,$8F,$88,$FA,$91,$F5,$5B,$3F,$C1,$BC,$3F,$E0,$3D,$1F
	.dc.b	$C0,$9B,$1F,$E6,$6E,$47,$C3,$7C,$E2,$47,$D3,$7F,$95,$69,$F4,$63
	.dc.b	$F8,$8F,$C9,$FE,$0F,$9B,$F8,$6B,$F8,$5A,$C4,$FE,$11,$BA,$7D,$6F
	.dc.b	$15,$C7,$F8,$3F,$CF,$BF,$8B,$5F,$F9,$E2,$72,$FF,$9E,$A3,$35,$F7
	.dc.b	$B3,$FF,$00,$CF,$E1,$FD,$2F,$C2,$91,$FF,$13,$7E,$7E,$F9,$CF,$C3
	.dc.b	$CF,$E4,$63,$13,$F0,$63,$F4,$7E,$34,$86,$1F,$E6,$BF,$92,$FA,$EF
	.dc.b	$F0,$39,$FE,$E3,$18,$E7,$FB,$B3,$FE,$14,$8C,$57,$2B,$43,$D2,$E3
	.dc.b	$D5,$E3,$F8,$AF,$EF,$7F,$A3,$FF,$BA,$30,$FF,$4C,$E8,$9C,$FC,$27
	.dc.b	$A7,$E3,$3F,$51,$26,$92,$BF,$C8,$DF,$59,$F5,$F5,$FD,$2F,$17,$F3
	.dc.b	$99,$FF,$52,$CE,$77,$FE,$F7,$39,$AF,$D5,$E0,$F2,$3B,$93,$0F,$9A
	.dc.b	$F8,$BE,$2B,$E6,$3F,$35,$B1,$EA,$F8,$BF,$E0,$1F,$2C,$F8,$3F,$A5
	.dc.b	$18,$7E,$A7,$B1,$EC,$25,$7D,$A7,$F8,$37,$D5,$3D,$BF,$E1,$5F,$82
	.dc.b	$7F,$87,$75,$FE,$03,$39,$BF,$F6,$0C,$E7,$7F,$F0,$BC,$FF,$47,$AF
	.dc.b	$BA,$1C,$3F,$80,$6B,$9F,$6F,$FA,$7F,$B1,$30,$FE,$A8,$D9,$FF,$10
	.dc.b	$7B,$3F,$8C,$7F,$B9,$C7,$F8,$DA,$31,$87,$F5,$48,$FE,$E7,$F3,$52
	.dc.b	$FF,$1F,$3F,$8A,$EC,$FF,$09,$F3,$67,$1F,$F5,$E6,$3F,$7E,$F6,$7E
	.dc.b	$35,$C4,$6B,$8F,$C1,$5F,$E0,$0E,$DF,$EC,$E4,$BF,$ED,$8F,$F8,$96
	.dc.b	$FF,$DE,$F1,$FE,$0C,$8F,$47,$27,$18,$FF,$07,$E1,$FD,$C0,$8F,$F7
	.dc.b	$57,$97,$C8,$FF,$83,$9F,$80,$D8,$F3,$C7,$F9,$E3,$95,$F4,$DF,$E5
	.dc.b	$A6,$CE,$B4,$FF,$19,$F2,$66,$BE,$53,$67,$59,$FF,$5C,$C3,$FD,$C1
	.dc.b	$87,$26,$FC,$DF,$F9,$06,$0D,$25,$7F,$91,$EB,$F6,$15,$C2,$3A,$F5
	.dc.b	$F9,$6B,$ED,$7C,$7F,$9F,$E6,$ED,$79,$F1,$47,$DF,$E9,$F9,$84,$9E
	.dc.b	$CF,$CF,$F8,$AD,$47,$A9,$B9,$7F,$BE,$7F,$51,$63,$FE,$7E,$FC,$F9
	.dc.b	$F3,$5F,$E8,$09,$CE,$1F,$1E,$72,$FC,$98,$FD,$73,$FE,$63,$6D,$FF
	.dc.b	$9D,$5A,$FF,$AE,$DA,$FF,$17,$7F,$51,$17,$0F,$07,$C6,$78,$5F,$CE
	.dc.b	$7C,$66,$27,$E9,$B6,$FE,$9E,$BF,$74,$FF,$4B,$FC,$23,$E5,$31,$7F
	.dc.b	$7F,$E5,$DD,$DF,$E7,$76,$C7,$F9,$ED,$B8,$7F,$A3,$5A,$FA,$71,$FF
	.dc.b	$51,$C7,$FB,$72,$7E,$14,$63,$AB,$FC,$21,$5F,$CF,$3E,$47,$F9,$EF
	.dc.b	$A3,$FE,$21,$8F,$E0,$73,$FC,$EA,$31,$1F,$C1,$A7,$7E,$B7,$F0,$88
	.dc.b	$EC,$7F,$68,$7F,$DF,$B1,$A8,$E7,$85,$79,$1B,$32,$53,$3D,$7F,$DC
	.dc.b	$9E,$D3,$FD,$26,$FA,$C7,$FB,$81,$B3,$B8,$7C,$A7,$F4,$DF,$07,$84
	.dc.b	$7F,$5A,$8C,$3C,$4C,$79,$C9,$DC,$72,$8C,$3B,$31,$1E,$B9,$7B,$5F
	.dc.b	$CA,$B0,$F8,$F1,$8A,$D5,$F2,$9C,$B7,$3F,$E6,$35,$BA,$F8,$4D,$AD
	.dc.b	$E1,$F2,$F1,$72,$FF,$B4,$23,$F9,$37,$EA,$4E,$4F,$E7,$71,$86,$ED
	.dc.b	$F9,$B2,$E4,$C1,$7F,$ED,$D8,$FF,$64,$35,$FF,$6D,$35,$FE,$8F,$1C
	.dc.b	$3E,$0F,$F0,$51,$FD,$F6,$FF,$BE,$BF,$C0,$FD,$2E,$F5,$E5,$6F,$37
	.dc.b	$EF,$27,$F8,$CE,$3A,$FD,$4F,$FD,$F1,$1F,$C8,$5A,$FF,$40,$8D,$79
	.dc.b	$92,$34,$F1,$F9,$3E,$5E,$BF,$92,$7D,$D6,$8C,$64,$8C,$4F,$FB,$9D
	.dc.b	$AF,$FB,$E5,$FF,$7A,$3F,$C3,$7E,$77,$D7,$DC,$7D,$1E,$6F,$BB,$F5
	.dc.b	$4F,$EE,$95,$FC,$E5,$E7,$C5,$FD,$6D,$BA,$FF,$88,$DD,$CF,$E4,$63
	.dc.b	$19,$F9,$A7,$C7,$3F,$41,$AE,$8F,$F7,$F3,$E6,$7F,$A0,$FC,$47,$D0
	.dc.b	$E1,$FF,$4A,$C7,$8F,$1F,$37,$C7,$67,$8E,$4F,$9E,$2B,$FB,$DD,$FF
	.dc.b	$19,$3F,$22,$FF,$05,$3F,$7E,$3C,$DF,$1D,$DF,$C4,$93,$F0,$62,$BF
	.dc.b	$D5,$22,$ED,$F0,$E7,$E4,$7C,$67,$9F,$E2,$65,$FC,$98,$AF,$8F,$14
	.dc.b	$F8,$C9,$1F,$0B,$E2,$FE,$07,$C4,$C8,$FA,$51,$5E,$51,$EB,$23,$F2
	.dc.b	$BE,$23,$ED,$11,$FB,$5F,$05,$F2,$F9,$64,$7F,$5F,$74,$FD,$93,$F5
	.dc.b	$9C,$7F,$EE,$D7,$FB,$27,$C0,$7D,$33,$F0,$DF,$0D,$63,$F5,$3C,$5F
	.dc.b	$B9,$9F,$9C,$48,$FC,$D7,$D3,$FF,$47,$6F,$97,$F9,$D5,$19,$7F,$66
	.dc.b	$D7,$F9,$B2,$F4,$BF,$80,$C3,$F0,$8F,$E1,$0F,$F8,$50,$C7,$0E,$AB
	.dc.b	$84,$67,$F4,$6F,$C8,$BF,$E8,$46,$A6,$93,$D5,$EF,$F5,$DF,$0E,$7F
	.dc.b	$DA,$AD,$E7,$FF,$43,$C7,$19,$F2,$D3,$9E,$FC,$94,$AF,$14,$57,$F8
	.dc.b	$8C,$A7,$E2,$36,$3F,$81,$35,$FE,$2F,$F0,$9F,$E0,$A7,$E0,$7C,$5F
	.dc.b	$F2,$BF,$07,$FC,$C6,$D8,$FF,$50,$CE,$DF,$F7,$6D,$3E,$47,$B9,$23
	.dc.b	$BB,$E2,$EA,$FE,$43,$EF,$1A,$7B,$8F,$E4,$BF,$AA,$FD,$2A,$FC,$57
	.dc.b	$32,$31,$BF,$F9,$8D,$BA,$7F,$AA,$DF,$E5,$F5,$C3,$C5,$FE,$C5,$5F
	.dc.b	$E6,$96,$D7,$FA,$C5,$B3,$FD,$06,$B9,$BF,$6F,$E4,$7E,$8F,$64,$8F
	.dc.b	$C0,$45,$7E,$7C,$7C,$E2,$3F,$CB,$FC,$6F,$73,$F8,$BE,$CE,$E9,$B1
	.dc.b	$A8,$F7,$C3,$E0,$FF,$95,$7B,$12,$7F,$95,$45,$7F,$CB,$BE,$9F,$F1
	.dc.b	$D3,$1A,$FE,$2B,$89,$E0,$FE,$08,$95,$F8,$27,$F8,$D7,$C8,$7F,$25
	.dc.b	$F5,$64,$FF,$96,$63,$FA,$BF,$E9,$3F,$E7,$48,$FF,$39,$3F,$E6,$86
	.dc.b	$FC,$06,$BE,$0B,$7F,$03,$27,$6F,$29,$FE,$65,$F2,$32,$7F,$CC,$CE
	.dc.b	$DF,$BF,$7F,$6B,$4F,$F5,$03,$79,$BF,$DD,$0D,$9F,$F7,$9F,$8D,$FE
	.dc.b	$E3,$7F,$CE,$1F,$45,$FC,$FC,$8F,$A3,$D6,$F5,$F9,$F2,$3D,$6F,$BB
	.dc.b	$CF,$C4,$7F,$A2,$3E,$53,$FA,$41,$8C,$FE,$67,$B9,$FB,$4F,$36,$57
	.dc.b	$F0,$38,$AF,$EC,$A3,$F6,$24,$7F,$8E,$E8,$F3,$FF,$AA,$7C,$9C,$9F
	.dc.b	$9B,$1F,$D5,$FD,$C9,$1F,$E8,$B7,$E9,$EF,$FE,$0A,$F9,$E7,$D5,$7F
	.dc.b	$7A,$6D,$7F,$2A,$6B,$FD,$65,$AF,$F8,$63,$CF,$FF,$10,$3F,$E9,$7F
	.dc.b	$17,$FB,$31,$FE,$24,$F9,$A7,$EA,$DF,$DF,$67,$FD,$65,$F4,$0F,$96
	.dc.b	$FF,$0D,$B6,$33,$1C,$65,$E0,$EB,$E5,$7F,$C1,$F7,$F5,$EF,$FB,$63
	.dc.b	$5F,$F1,$5B,$5F,$BF,$D7,$DB,$E3,$E4,$7E,$4A,$3C,$3B,$48,$F2,$BF
	.dc.b	$98,$E2,$F8,$DF,$27,$F1,$49,$3F,$15,$F1,$FC,$8F,$A8,$91,$F9,$ED
	.dc.b	$FD,$BE,$7C,$8F,$82,$FF,$8E,$BE,$57,$D5,$D5,$FE,$F5,$D5,$F1,$9E
	.dc.b	$E2,$5F,$F5,$C8,$CB,$FE,$46,$E3,$FC,$16,$1F,$E3,$56,$C6,$7A,$13
	.dc.b	$96,$C7,$F6,$6C,$59,$FD,$2F,$92,$93,$FE,$E7,$9C,$BF,$EE,$9E,$77
	.dc.b	$D9,$7E,$1C,$91,$F8,$77,$EA,$EF,$F7,$09,$1F,$EF,$6C,$DF,$AB,$FF
	.dc.b	$26,$E0,$6C,$F9,$4F,$C1,$92,$F8,$6C,$D7,$F0,$FF,$01,$F8,$22,$4F
	.dc.b	$C5,$7F,$CA,$7B,$73,$8F,$C7,$FC,$07,$E3,$8F,$A1,$87,$F9,$AB,$E1
	.dc.b	$3D,$72,$3F,$A3,$B8,$6E,$FB,$AF,$D4,$7C,$8F,$E9,$C9,$3F,$A7,$7C
	.dc.b	$33,$F9,$C2,$47,$CB,$7F,$01,$F2,$9F,$A0,$F9,$49,$3F,$3E,$2B,$F9
	.dc.b	$C8,$FC,$F9,$1F,$B1,$F2,$9F,$89,$23,$F1,$A6,$BF,$D8,$37,$47,$EB
	.dc.b	$8C,$F4,$3C,$CF,$81,$C4,$DD,$F3,$3F,$95,$26,$3F,$83,$FA,$1A,$F5
	.dc.b	$DB,$F7,$26,$31,$E9,$1F,$EA,$3E,$05,$7F,$B7,$FC,$AE,$F1,$89,$FF
	.dc.b	$74,$46,$27,$51,$C7,$0B,$92,$F8,$4B,$E5,$18,$71,$8E,$67,$87,$0E
	.dc.b	$35,$EC,$8C,$DF,$39,$C9,$23,$33,$B7,$B5,$7C,$6E,$BF,$F5,$4E,$1E
	.dc.b	$09,$FF,$60,$69,$FE,$30,$8C,$63,$F2,$F5,$F4,$DF,$0C,$39,$7C,$7F
	.dc.b	$B3,$0E,$11,$F4,$9B,$8F,$FB,$2E,$B8,$3F,$95,$FA,$37,$17,$F3,$08
	.dc.b	$FC,$C1,$97,$0E,$23,$79,$7F,$AE,$5B,$1F,$EC,$EF,$95,$79,$4F,$A4
	.dc.b	$FF,$5F,$46,$1B,$19,$9E,$0D,$7C,$22,$9F,$30,$F9,$8F,$03,$67,$FD
	.dc.b	$BB,$3A,$9C,$9F,$C7,$67,$F1,$3C,$1F,$2B,$9F,$DF,$C3,$F6,$1E,$AD
	.dc.b	$D7,$7C,$1F,$9A,$9C,$E5,$F6,$0F,$F7,$A3,$C9,$C6,$DF,$C4,$31,$E3
	.dc.b	$1A,$BB,$2F,$0B,$F4,$78,$FE,$21,$B5,$C9,$FC,$EC,$67,$97,$8C,$6B
	.dc.b	$3F,$C1,$B7,$3F,$D7,$BC,$2F,$95,$70,$E5,$E2,$DC,$3A,$1F,$04,$FD
	.dc.b	$6F,$37,$F4,$78,$B1,$F0,$4B,$5F,$02,$2E,$5F,$CB,$A6,$EB,$FC,$05
	.dc.b	$FE,$E7,$86,$1E,$3E,$0F,$F3,$3F,$4F,$8F,$D0,$3A,$33,$DE,$9C,$C6
	.dc.b	$19,$7F,$CA,$F5,$FE,$07,$9B,$07,$CB,$6B,$6F,$F3,$AC,$3E,$53,$FD
	.dc.b	$33,$1F,$C5,$23,$18,$7F,$AF,$1B,$1F,$EE,$B6,$BF,$A2,$BE,$DF,$07
	.dc.b	$E8,$BF,$4B,$6F,$C8,$7E,$ED,$F6,$FC,$77,$C6,$EC,$46,$27,$78,$30
	.dc.b	$7D,$23,$0F,$3C,$E6,$31,$5F,$ED,$26,$C7,$FC,$01,$6B,$E7,$7F,$8E
	.dc.b	$47,$6A,$EB,$DF,$F4,$0F,$B3,$3A,$AF,$8F,$F1,$FF,$53,$E6,$3E,$AB
	.dc.b	$70,$E5,$AF,$FA,$E5,$F9,$1F,$21,$E6,$CB,$E9,$36,$75,$87,$EE,$B9
	.dc.b	$3F,$B9,$7A,$9F,$78,$FC,$0F,$CB,$5C,$7F,$BD,$7D,$1A,$3F,$0E,$D8
	.dc.b	$F0,$FC,$86,$27,$D2,$F8,$76,$FF,$5F,$4F,$C5,$8C,$1E,$3F,$8D,$FC
	.dc.b	$7F,$C8,$7E,$41,$B3,$EC,$F8,$EF,$C8,$FC,$47,$E5,$1B,$1F,$03,$E2
	.dc.b	$BD,$84,$7F,$10,$FB,$FE,$0B,$F3,$67,$E7,$3E,$0E,$CF,$C8,$F8,$DF
	.dc.b	$CF,$9F,$A0,$FA,$0B,$1F,$57,$55,$FA,$C9,$1F,$7D,$F9,$9E,$57,$EA
	.dc.b	$0F,$D4,$FC,$75,$9F,$BD,$F2,$6F,$8D,$FD,$6F,$37,$ED,$A3,$13,$F6
	.dc.b	$A3,$1F,$19,$F6,$64,$D2,$4F,$F9,$0F,$E8,$BF,$C4,$CF,$F8,$BD,$B5
	.dc.b	$FE,$E2,$A7,$C7,$7E,$FA,$73,$E2,$FF,$78,$EA,$47,$F8,$CF,$0E,$5E
	.dc.b	$47,$97,$E2,$78,$3F,$C3,$3E,$35,$1F,$8D,$6C,$FF,$44,$CF,$F5,$11
	.dc.b	$87,$FA,$3F,$80,$8C,$47,$EE,$74,$7C,$D7,$C3,$3E,$24,$DB,$FD,$9D
	.dc.b	$7F,$85,$BD,$0F,$91,$E5,$FF,$20,$8C,$4F,$CE,$8C,$79,$EF,$85,$FD
	.dc.b	$BD,$7F,$87,$9B,$19,$AF,$44,$E5,$F5,$66,$CD,$F1,$DD,$30,$FB,$0D
	.dc.b	$AF,$BB,$F0,$5F,$6F,$E6,$BE,$E3,$63,$F0,$3F,$35,$86,$9C,$0F,$6F
	.dc.b	$D1,$7F,$87,$1F,$C3,$FB,$7E,$D8,$C6,$9F,$6E,$3F,$11,$E4,$E7,$82
	.dc.b	$DF,$27,$F1,$9F,$27,$23,$F2,$3E,$2D,$C3,$F0,$9F,$19,$F9,$18,$FC
	.dc.b	$93,$67,$F2,$FF,$27,$86,$9F,$9A,$8F,$9B,$F2,$5F,$28,$91,$F2,$B9
	.dc.b	$11,$F9,$9C,$B7,$3F,$A2,$6E,$3F,$95,$9F,$D1,$71,$7E,$8D,$B5,$FA
	.dc.b	$8E,$4C,$34,$FD,$5C,$7D,$E8,$FC,$17,$A8,$F2,$71,$5F,$88,$8C,$3F
	.dc.b	$8A,$FA,$1F,$F6,$93,$F8,$06,$E3,$F9,$96,$BF,$E1,$E6,$BF,$E3,$0F
	.dc.b	$8F,$FC,$D4,$F3,$F9,$3F,$C9,$46,$25,$E0,$FF,$99,$15,$F6,$8F,$EC
	.dc.b	$BD,$9F,$ED,$EC,$3F,$A1,$3F,$E3,$46,$F2,$7E,$BF,$97,$F6,$1C,$BF
	.dc.b	$B1,$8B,$3A,$3D,$9F,$23,$F9,$EC,$4D,$24,$7B,$F3,$FD,$44,$EA,$3E
	.dc.b	$24,$62,$73,$EC,$FE,$B7,$E6,$7F,$4D,$AF,$B2,$FF,$CA,$38,$7F,$B2
	.dc.b	$9F,$DA,$CE,$B5,$F8,$13,$FE,$D2,$9F,$9D,$F2,$3E,$52,$3F,$B0,$F8
	.dc.b	$87,$A5,$E4,$9F,$DE,$46,$27,$EC,$47,$A3,$C5,$F7,$12,$3E,$D3,$F0
	.dc.b	$7E,$3B,$51,$F7,$DB,$9F,$A4,$FF,$60,$7D,$07,$F8,$29,$FF,$35,$B5
	.dc.b	$CF,$BF,$FC,$22,$31,$07,$FC,$0F,$5F,$C5,$1F,$F0,$AE,$9F,$7D,$CC
	.dc.b	$F7,$38,$57,$F8,$D3,$DF,$FC,$5C,$7F,$83,$5B,$D3,$FE,$0D,$E7,$FE
	.dc.b	$F1,$18,$1F,$29,$C6,$39,$37,$2F,$F0,$54,$FD,$BB,$F7,$5F,$F8,$5F
	.dc.b	$8F,$C1,$C7,$E1,$6D,$F0,$8F,$27,$26,$2F,$8E,$1F,$65,$FE,$72,$D7
	.dc.b	$FC,$60,$D7,$FD,$96,$D7,$D9,$F2,$9E,$FE,$1F,$80,$DC,$3C,$FF,$25
	.dc.b	$EE,$23,$FC,$C2,$FF,$AC,$DF,$6F,$94,$F9,$15,$F2,$5B,$5F,$03,$8F
	.dc.b	$E6,$1F,$37,$8E,$C7,$C4,$6D,$FB,$1B,$C7,$FE,$56,$C5,$F2,$7F,$40
	.dc.b	$97,$E3,$7E,$9F,$C9,$FD,$97,$ED,$36,$3E,$8F,$C9,$7D,$DF,$93,$FB
	.dc.b	$CD,$8F,$AB,$F2,$7F,$70,$8F,$BD,$FB,$FF,$1F,$F8,$57,$F0,$CD,$8F
	.dc.b	$F2,$37,$D0,$7F,$BF,$1B,$7F,$E7,$E8,$C6,$BF,$80,$51,$9B,$FF,$21
	.dc.b	$41,$A4,$AF,$EB,$F1,$FE,$A7,$D3,$FC,$F2,$D8,$FB,$BF,$23,$DF,$1F
	.dc.b	$CD,$79,$3F,$9B,$8B,$3F,$92,$7F,$D2,$3F,$14,$E0,$A7,$FA,$53,$91
	.dc.b	$F0,$18,$BF,$F3,$14,$3D,$DF,$F2,$C3,$63,$FB,$9B,$77,$FF,$05,$B5
	.dc.b	$FF,$41,$69,$FE,$B6,$D7,$D3,$9F,$C5,$D3,$E3,$3F,$E0,$7A,$0F,$91
	.dc.b	$BE,$5B,$E0,$FF,$DB,$BF,$00,$D3,$31,$FE,$1D,$AB,$7F,$EF,$2A,$D3
	.dc.b	$FD,$82,$74,$D3,$55,$C1,$FF,$7C,$D5,$7F,$DD,$11,$5F,$21,$FE,$91
	.dc.b	$BE,$7D,$FF,$DB,$23,$FD,$E3,$5E,$A8,$C4,$74,$38,$1E,$EF,$94,$79
	.dc.b	$3F,$D5,$23,$22,$33,$AF,$3C,$3E,$2C,$E6,$B8,$5F,$6B,$DC,$F2,$6D
	.dc.b	$7C,$98,$F0,$79,$5C,$A3,$11,$DE,$31,$9F,$6A,$E2,$DC,$3C,$D7,$C2
	.dc.b	$55,$C2,$FA,$9F,$2E,$38,$BF,$EB,$7D,$DF,$21,$EA,$CF,$F3,$7A,$D3
	.dc.b	$E4,$4F,$F7,$58,$FC,$FC,$E6,$BF,$D8,$1F,$41,$FE,$57,$7F,$C8,$6D
	.dc.b	$7F,$CB,$6D,$7F,$DA,$4D,$7E,$A7,$37,$EC,$9F,$F5,$F3,$5F,$E6,$AD
	.dc.b	$AF,$F7,$3C,$FF,$98,$A7,$F9,$DD,$78,$F6,$FF,$64,$D7,$F1,$7C,$71
	.dc.b	$DB,$EE,$BF,$F6,$57,$D0,$7F,$BD,$DB,$F2,$5C,$63,$11,$FC,$F2,$B1
	.dc.b	$F2,$5F,$E6,$0C,$55,$FE,$F6,$AF,$F3,$0B,$5F,$F3,$CB,$5F,$F7,$83
	.dc.b	$FE,$E8,$7C,$3E,$57,$73,$C1,$E9,$7A,$7E,$6F,$C3,$A7,$E9,$B0,$47
	.dc.b	$58,$C3,$72,$FC,$E3,$C1,$B9,$79,$E3,$C5,$8E,$0F,$F3,$BB,$87,$FD
	.dc.b	$F0,$F9,$5F,$EC,$6B,$1F,$21,$E5,$C7,$CD,$A7,$9F,$E4,$E5,$FF,$BD
	.dc.b	$6F,$FE,$00,$36,$3C,$44,$7C,$2F,$97,$83,$E0,$7C,$9C,$9E,$51,$5F
	.dc.b	$1C,$5D,$3E,$0C,$FC,$7F,$93,$F9,$1F,$23,$2F,$C5,$15,$FF,$55,$E3
	.dc.b	$FE,$F3,$7E,$0C,$FC,$53,$E3,$7C,$8F,$D1,$F9,$19,$3E,$B8,$AF,$28
	.dc.b	$F9,$84,$7E,$5F,$B5,$F6,$48,$FC,$3F,$5B,$EE,$7C,$8C,$8F,$8C,$E9
	.dc.b	$F9,$FD,$DF,$81,$F8,$F9,$1F,$2E,$2B,$EA,$8F,$B8,$47,$FD,$67,$F3
	.dc.b	$EF,$3F,$F6,$90,$8F,$E9,$8F,$FB,$11,$AF,$F8,$BD,$E0,$E7,$C9,$FF
	.dc.b	$5E,$BC,$5B,$1C,$23,$F8,$F9,$93,$FD,$87,$F3,$9F,$87,$7F,$DA,$2F
	.dc.b	$E2,$F8,$BF,$CF,$8D,$E9,$FE,$4E,$EE,$7F,$C8,$0F,$E4,$5E,$AD,$77
	.dc.b	$CF,$E4,$8C,$47,$F5,$4F,$95,$FF,$76,$BE,$0E,$6F,$83,$27,$E1,$CE
	.dc.b	$A7,$BC,$73,$23,$EA,$7B,$72,$FC,$AF,$77,$FD,$B7,$3F,$E7,$8F,$89
	.dc.b	$EB,$3D,$B9,$BB,$F3,$4A,$F8,$31,$5F,$F7,$F6,$49,$F5,$12,$3F,$DD
	.dc.b	$3F,$3A,$F9,$1F,$E3,$4F,$F2,$9D,$FF,$81,$37,$7F,$EB,$4E,$EB,$FB
	.dc.b	$AC,$EB,$E0,$B8,$FC,$7C,$9C,$DF,$09,$FC,$2F,$CF,$FF,$82,$67,$EB
	.dc.b	$F4,$FC,$83,$3E,$8D,$8F,$BB,$1F,$44,$8F,$E0,$3A,$FF,$1F,$CA,$EC
	.dc.b	$A7,$D2,$8F,$E3,$9E,$97,$F1,$DC,$25,$FF,$07,$8A,$FF,$10,$F9,$6F
	.dc.b	$E4,$46,$36,$FE,$3F,$E8,$7D,$CE,$F9,$5F,$C3,$E2,$BF,$3A,$3F,$0C
	.dc.b	$47,$F8,$F9,$FD,$17,$E7,$9F,$2D,$4F,$E6,$11,$4F,$8E,$3C,$84,$BE
	.dc.b	$5B,$91,$B9,$FF,$44,$C7,$D3,$6F,$37,$73,$6F,$4A,$E7,$CB,$FB,$6F
	.dc.b	$88,$FF,$4B,$CF,$80,$38,$7A,$C6,$FE,$13,$D5,$CD,$27,$85,$6F,$6F
	.dc.b	$E9,$19,$7F,$65,$9F,$2F,$6F,$C1,$3E,$17,$CA,$58,$F2,$4B,$E5,$3E
	.dc.b	$31,$98,$F6,$FC,$9F,$EB,$3C,$89,$1C,$22,$BF,$9E,$8F,$DB,$11,$FE
	.dc.b	$0B,$E4,$7D,$1F,$93,$91,$FC,$1A,$3E,$A6,$92,$3C,$6F,$D7,$F9,$3F
	.dc.b	$D1,$FB,$3F,$DC,$7C,$1F,$F1,$AD,$7E,$A2,$30,$7F,$14,$7F,$CA,$FF
	.dc.b	$2C,$F9,$E5,$FF,$85,$E3,$0D,$8C,$CF,$C2,$AC,$11,$FC,$1D,$73,$8F
	.dc.b	$C5,$7B,$5F,$8B,$F9,$4F,$F0,$27,$01,$A4,$BF,$27,$3F,$E2,$BE,$53
	.dc.b	$F2,$5F,$29,$27,$E8,$C7,$E5,$7E,$4E,$47,$D4,$7F,$31,$F2,$9F,$E1
	.dc.b	$DF,$59,$F2,$3F,$E1,$36,$D7,$66,$BF,$CD,$B1,$46,$23,$FC,$AF,$1F
	.dc.b	$9C,$3F,$48,$47,$F0,$7F,$03,$FE,$2B,$F9,$3F,$C6,$79,$BF,$E6,$96
	.dc.b	$CF,$E9,$5B,$47,$C3,$DB,$FE,$19,$5F,$25,$FE,$29,$6D,$7C,$46,$BE
	.dc.b	$76,$BE,$65,$F2,$3F,$67,$A9,$F2,$34,$F3,$47,$9D,$F4,$7C,$94,$BF
	.dc.b	$D8,$5F,$07,$F6,$5A,$BE,$2B,$E5,$12,$3E,$53,$F8,$B5,$D0,$F9,$EF
	.dc.b	$E9,$5A,$FF,$A6,$A3,$1E,$37,$E3,$67,$3C,$1F,$1F,$0F,$47,$C9,$CC
	.dc.b	$7F,$17,$BA,$3E,$71,$F3,$FE,$4F,$E3,$12,$7E,$33,$F2,$7D,$EF,$A9
	.dc.b	$EF,$C9,$F6,$C7,$E1,$7E,$4E,$47,$B9,$FB,$47,$E2,$7E,$4F,$F1,$04
	.dc.b	$8F,$C4,$3F,$E5,$DF,$9C,$FF,$70,$47,$C8,$8F,$C1,$1F,$E6,$D8,$CC
	.dc.b	$38,$3E,$F7,$C9,$FE,$1F,$E5,$24,$FD,$08,$AF,$DD,$8F,$C1,$11,$FD
	.dc.b	$C7,$13,$64,$FD,$F2,$4F,$DA,$F9,$4A,$F8,$78,$7F,$29,$F2,$9F,$97
	.dc.b	$33,$97,$E4,$7E,$3B,$F3,$44,$AF,$C1,$FC,$9D,$79,$1F,$9B,$F9,$4F
	.dc.b	$D0,$19,$AF,$F3,$97,$42,$3F,$DA,$7E,$0F,$FB,$2E,$3F,$D1,$95,$FE
	.dc.b	$E5,$8C,$61,$A8,$FE,$95,$3C,$5F,$F7,$5F,$CA,$2B,$71,$86,$BC,$CE
	.dc.b	$35,$CF,$C0,$FD,$DE,$17,$0F,$0E,$1C,$EF,$67,$32,$9C,$EB,$8B,$95
	.dc.b	$DB,$FE,$A0,$9E,$51,$FB,$BD,$7C,$1D,$BF,$EF,$8B,$E5,$7B,$CF,$FD
	.dc.b	$D8,$D8,$FE,$09,$5C,$0F,$44,$66,$1F,$3A,$DF,$FB,$74,$7F,$39,$8C
	.dc.b	$63,$FB,$16,$C6,$67,$6F,$F9,$4B,$E5,$7F,$B3,$FC,$F7,$ED,$DB,$1F
	.dc.b	$D0,$DA,$EF,$D0,$4F,$5E,$27,$BB,$49,$5E,$38,$CB,$FC,$AF,$A7,$F0
	.dc.b	$5A,$FC,$96,$C7,$CD,$7F,$DD,$71,$9E,$23,$49,$3B,$E3,$CB,$C6,$6F
	.dc.b	$17,$D2,$FD,$3D,$38,$E2,$BC,$29,$CE,$3F,$6F,$F4,$EF,$F0,$B3,$F7
	.dc.b	$63,$37,$F8,$19,$CB,$F8,$38,$CE,$1F,$80,$9C,$BE,$19,$F0,$1F,$E2
	.dc.b	$38,$CE,$1E,$2A,$E2,$F9,$23,$33,$FE,$29,$9C,$BF,$E2,$BC,$FE,$1E
	.dc.b	$1F,$E2,$3E,$37,$CA,$7C,$2E,$7F,$F7,$05,$6A,$FF,$6B,$A7,$E5,$AB
	.dc.b	$58,$7C,$2A,$D5,$F0,$8C,$DF,$F8,$E7,$E7,$BA,$C7,$F8,$C6,$B6,$FC
	.dc.b	$A8,$FF,$6F,$3E,$3F,$95,$E3,$8A,$73,$7E,$2C,$AE,$3F,$EF,$59,$B8
	.dc.b	$73,$C3,$C3,$DA,$F9,$F7,$CA,$F4,$BD,$E5,$FE,$4D,$C9,$F1,$5F,$AD
	.dc.b	$8F,$D5,$B5,$D3,$74,$FD,$BB,$69,$F2,$DE,$8B,$E7,$A7,$F9,$3E,$72
	.dc.b	$FC,$FA,$F1,$63,$BB,$B3,$F4,$DF,$83,$F7,$CF,$80,$F0,$36,$7F,$C3
	.dc.b	$6D,$CF,$FC,$1C,$D7,$FC,$1B,$C6,$F8,$1F,$96,$DA,$FF,$31,$2F,$80
	.dc.b	$CC,$FC,$5F,$92,$F5,$78,$76,$FF,$0C,$FB,$74,$F8,$5B,$FC,$C9,$FF
	.dc.b	$16,$B7,$2F,$F1,$03,$5F,$F0,$FB,$FE,$C3,$8F,$C8,$7C,$A7,$D6,$E2
	.dc.b	$7C,$7F,$C9,$8F,$95,$94,$AF,$F2,$0F,$E1,$EF,$8E,$C6,$5F,$94,$E0
	.dc.b	$F5,$57,$C4,$DF,$E3,$91,$FF,$20,$F2,$3E,$3F,$EB,$7C,$A7,$E0,$DB
	.dc.b	$3F,$E4,$ED,$9F,$1D,$F8,$1E,$9F,$C5,$36,$33,$3F,$DA,$1B,$9F,$C0
	.dc.b	$F9,$3F,$8A,$49,$F8,$AF,$C8,$F9,$2F,$C3,$1F,$93,$F9,$1B,$3F,$B3
	.dc.b	$E3,$FC,$51,$F9,$7F,$1E,$C7,$FA,$67,$81,$F0,$DF,$84,$24,$7E,$11
	.dc.b	$FA,$BB,$7E,$4C,$FD,$5F,$C6,$FF,$AC,$27,$F8,$04,$63,$E4,$3F,$31
	.dc.b	$BD,$9F,$E0,$8F,$E6,$FE,$37,$EF,$92,$3E,$FB,$F8,$2F,$8D,$FD,$01
	.dc.b	$FB,$0F,$D6,$3F,$D3,$53,$FC,$36,$3F,$49,$F2,$72,$3F,$CF,$73,$F6
	.dc.b	$38,$BF,$79,$3F,$A0,$F2,$BF,$3E,$4A,$FC,$FB,$F9,$0E,$1F,$D7,$1E
	.dc.b	$CF,$02,$CF,$FA,$E7,$C4,$FB,$CF,$FA,$EF,$D4,$74,$FF,$14,$CB,$F7
	.dc.b	$FC,$16,$73,$5F,$EC,$1F,$37,$FD,$A0,$FF,$94,$5B,$5F,$E9,$D6,$BF
	.dc.b	$D1,$0F,$E8,$AE,$1C,$7C,$FF,$FA,$1A,$7A,$79,$1E,$18,$C4,$FF,$88
	.dc.b	$0F,$F0,$E4,$7D,$6E,$EF,$E1,$5D,$F6,$FF,$9B,$36,$BF,$99,$F3,$7C
	.dc.b	$9F,$2B,$FA,$CE,$08,$FF,$1D,$5F,$FB,$7D,$D3,$FE,$06,$EA,$FE,$2C
	.dc.b	$7E,$13,$D2,$B1,$B7,$0F,$1E,$AB,$8F,$F3,$13,$F8,$F1,$F8,$CF,$8B
	.dc.b	$B3,$DD,$C3,$E1,$F8,$BB,$3E,$8F,$8B,$FC,$9C,$7E,$51,$B3,$FE,$F1
	.dc.b	$DD,$F2,$5F,$D0,$BE,$9D,$F0,$1F,$26,$30,$FC,$BF,$8A,$FE,$6D,$92
	.dc.b	$FF,$D3,$F8,$7F,$A2,$5B,$1F,$ED,$56,$BF,$EB,$2D,$DF,$1B,$FD,$43
	.dc.b	$E9,$7F,$4D,$3F,$3A,$31,$F2,$DF,$AA,$F8,$EF,$D5,$C5,$C7,$E7,$3F
	.dc.b	$8D,$F9,$9E,$71,$F9,$1F,$27,$13,$F8,$F7,$F0,$5E,$8F,$A3,$4F,$F2
	.dc.b	$0D,$79,$E3,$06,$7E,$3B,$71,$F1,$5B,$9F,$F7,$4F,$4F,$E3,$CF,$F8
	.dc.b	$5B,$AB,$F0,$71,$89,$FE,$F3,$FA,$CB,$E1,$3F,$B6,$CE,$6B,$F1,$33
	.dc.b	$9B,$E4,$7F,$87,$9C,$5F,$2F,$83,$FE,$EB,$18,$AF,$F2,$A6,$6F,$8B
	.dc.b	$FD,$77,$C7,$FF,$A2,$B1,$FB,$91,$8F,$87,$FD,$17,$C6,$7C,$12,$4F
	.dc.b	$D7,$7E,$DD,$75,$8F,$D7,$7C,$93,$E1,$BA,$EF,$FB,$B8,$C4,$FE,$27
	.dc.b	$E2,$FE,$61,$1F,$C1,$3F,$3F,$1F,$D9,$3F,$B3,$6C,$7C,$0D,$FF,$C7
	.dc.b	$FF,$45,$FE,$18,$7F,$6D,$F5,$1E,$37,$3F,$0D,$FC,$2E,$31,$09,$FF
	.dc.b	$0B,$BF,$C1,$4C,$7C,$2F,$E5,$A3,$31,$FC,$25,$AF,$F9,$3B,$F4,$BF
	.dc.b	$EE,$CF,$F0,$5F,$37,$FC,$12,$FF,$07,$8C,$F1,$FF,$A0,$E7,$2F,$FA
	.dc.b	$16,$3F,$96,$4F,$85,$AF,$F1,$76,$BD,$FC,$CF,$F0,$E7,$7F,$F9,$52
	.dc.b	$3F,$8E,$79,$68,$C4,$7F,$98,$1A,$A3,$5E,$CF,$F8,$97,$E9,$3F,$8E
	.dc.b	$E3,$FC,$52,$FF,$CB,$EF,$F1,$68,$CE,$FC,$EB,$F6,$1F,$39,$23,$E9
	.dc.b	$60,$AF,$DC,$DF,$3A,$FF,$22,$C6,$2B,$FC,$E6,$DC,$7C,$8F,$EB,$1F
	.dc.b	$F2,$CB,$9B,$F6,$73,$7C,$02,$57,$D7,$AF,$66,$49,$F6,$76,$FF,$05
	.dc.b	$7C,$2D,$AF,$3C,$FF,$31,$79,$5F,$C9,$F1,$3E,$69,$23,$38,$BE,$4F
	.dc.b	$F8,$84,$FB,$1B,$3D,$3E,$23,$F8,$A1,$F0,$3A,$EC,$7C,$6B,$46,$34
	.dc.b	$F7,$B7,$6F,$A5,$F3,$9F,$78,$97,$E6,$D3,$FB,$BD,$FE,$5F,$4F,$A1
	.dc.b	$E4,$FF,$23,$BF,$9E,$DB,$F0,$75,$7F,$27,$F5,$FE,$9B,$63,$C9,$D7
	.dc.b	$F8,$32,$3F,$01,$FC,$37,$6F,$DA,$7E,$DB,$63,$F0,$3F,$2A,$F6,$FF
	.dc.b	$8B,$FC,$AF,$FA,$12,$31,$3F,$86,$F1,$93,$9F,$6F,$FC,$ED,$46,$92
	.dc.b	$BE,$AC,$7F,$92,$DE,$FE,$BF,$5F,$5F,$B3,$F8,$EF,$17,$13,$F3,$DB
	.dc.b	$C5,$FE,$A4,$E3,$FE,$DA,$FE,$23,$BB,$FD,$23,$D1,$FD,$62,$FF,$0B
	.dc.b	$1F,$E9,$4F,$8E,$7B,$1F,$E4,$D6,$CF,$E4,$7E,$1B,$FD,$99,$39,$3D
	.dc.b	$78,$71,$F9,$9F,$84,$DA,$FE,$30,$FF,$84,$FE,$43,$FD,$AF,$E9,$BE
	.dc.b	$0F,$F8,$97,$07,$FB,$7A,$31,$5F,$C5,$6B,$75,$AB,$B3,$91,$5C,$1F
	.dc.b	$F7,$1D,$F3,$AF,$D4,$5F,$09,$CD,$D7,$FD,$D3,$1F,$EE,$8B,$DC,$7F
	.dc.b	$09,$6F,$67,$68,$F0,$35,$7C,$9F,$F9,$83,$BB,$94,$7F,$BD,$23,$0F
	.dc.b	$33,$1E,$EA,$77,$1C,$A3,$0E,$CC,$47,$A6,$5A,$7F,$8F,$70,$F8,$11
	.dc.b	$8A,$D5,$F2,$9C,$B7,$3F,$F2,$25,$6E,$BE,$4C,$62,$FA,$3A,$AE,$13
	.dc.b	$FC,$AF,$E9,$1E,$3F,$F9,$37,$26,$AA,$70,$3E,$43,$F8,$0C,$7F,$59
	.dc.b	$8C,$78,$7C,$6B,$86,$EC,$BA,$DE,$F0,$7C,$BF,$F2,$18,$EB,$18,$7F
	.dc.b	$28,$D7,$FA,$E6,$3E,$19,$E1,$7F,$D1,$FE,$5F,$C1,$5F,$CA,$9B,$19
	.dc.b	$8F,$37,$3F,$3D,$3F,$A2,$67,$AC,$FF,$98,$57,$F4,$26,$BF,$CE,$A3
	.dc.b	$5E,$44,$8D,$3C,$7C,$AE,$5E,$3F,$92,$CD,$25,$7F,$36,$8F,$F7,$73
	.dc.b	$FC,$2F,$C5,$F3,$F8,$47,$C9,$9F,$AB,$E8,$DE,$95,$F8,$69,$87,$F5
	.dc.b	$89,$CB,$63,$E2,$C1,$A4,$8F,$E8,$F0,$F4,$BE,$3F,$82,$F9,$DA,$78
	.dc.b	$63,$C5,$F2,$9A,$F0,$66,$1E,$18,$AF,$F3,$BB,$FE,$1E,$FF,$3F,$9F
	.dc.b	$AF,$18,$8F,$E8,$18,$3E,$53,$C3,$D9,$27,$D9,$15,$FE,$41,$17,$97
	.dc.b	$FD,$23,$5C,$67,$E1,$7C,$4F,$AB,$BE,$46,$0F,$EC,$EF,$88,$8F,$F5
	.dc.b	$37,$E3,$FB,$DF,$0B,$E1,$E4,$FC,$D8,$AF,$B6,$3D,$64,$7E,$5F,$C6
	.dc.b	$FD,$72,$3F,$57,$E2,$BE,$6F,$C4,$C8,$FE,$F2,$E9,$FE,$11,$D9,$FA
	.dc.b	$8C,$DF,$1B,$F5,$3E,$2A,$57,$DF,$8A,$FC,$F8,$F9,$A4,$7F,$59,$E6
	.dc.b	$FE,$69,$FF,$22,$35,$FF,$65,$35,$FD,$F3,$5F,$F3,$07,$7F,$EB,$17
	.dc.b	$D6,$FF,$60,$7F,$A0,$B5,$FF,$66,$4F,$FB,$0D,$FE,$8B,$F4,$79,$7F
	.dc.b	$CA,$8D,$7F,$DB,$4D,$7F,$87,$35,$3F,$EB,$27,$FA,$7F,$D5,$7E,$D3
	.dc.b	$8F,$F8,$AC,$7F,$15,$8C,$69,$D3,$D9,$F4,$5B,$E6,$AC,$CB,$F2,$C5
	.dc.b	$7F,$C6,$53,$FE,$BC,$7C,$33,$FC,$55,$A7,$F1,$7F,$BB,$7C,$47,$C5
	.dc.b	$FA,$CF,$F2,$A4,$FF,$64,$6B,$FE,$EF,$9D,$BF,$DA,$3E,$97,$F6,$8F
	.dc.b	$F2,$16,$BF,$E4,$36,$BF,$E7,$B6,$BF,$F0,$05,$7C,$E7,$D5,$FA,$5F
	.dc.b	$C3,$CF,$F7,$46,$BF,$EF,$F9,$DB,$FE,$5E,$FC,$13,$FC,$84,$DF,$83
	.dc.b	$FE,$4F,$18,$8F,$C9,$9F,$E6,$F7,$FD,$5C,$FF,$99,$BD,$9F,$E1,$EE
	.dc.b	$BC,$E4,$63,$8B,$EA,$C6,$3C,$8E,$B8,$7E,$2F,$1F,$DC,$61,$E5,$F9
	.dc.b	$EC,$19,$F0,$EC,$7C,$58,$FA,$52,$69,$F5,$7E,$45,$F0,$9F,$97,$FA
	.dc.b	$BF,$C2,$DF,$A8,$38,$D3,$35,$C2,$7F,$95,$7B,$FF,$CB,$FE,$06,$46
	.dc.b	$0F,$E5,$CF,$F2,$6F,$98,$91,$83,$EE,$7C,$E7,$98,$8F,$DD,$7D,$3E
	.dc.b	$3F,$F5,$4F,$89,$34,$F8,$87,$F3,$F9,$FE,$B3,$F4,$EF,$96,$A6,$63
	.dc.b	$FD,$2F,$0F,$8C,$D9,$23,$33,$FC,$F3,$E5,$A4,$7F,$A2,$1D,$3F,$66
	.dc.b	$BF,$52,$FF,$AB,$5B,$CC,$69,$FB,$D8,$FF,$29,$7C,$27,$F4,$82,$3E
	.dc.b	$4F,$73,$D3,$EE,$48,$F3,$BE,$CE,$4C,$47,$F9,$6F,$0F,$EA,$26,$38
	.dc.b	$FE,$CF,$33,$E7,$7F,$11,$E7,$4B,$FD,$54,$57,$F5,$11,$F8,$62,$3F
	.dc.b	$E0,$DF,$67,$E7,$BF,$E5,$8F,$A9,$FF,$1F,$61,$F8,$66,$F8,$BF,$E1
	.dc.b	$F9,$FE,$D1,$C7,$FE,$0A,$FA,$27,$D3,$D2,$4F,$D2,$F8,$1F,$F0,$7F
	.dc.b	$84,$F8,$0F,$F0,$AB,$67,$FD,$6A,$D7,$FD,$40,$D7,$FD,$C3,$6F,$9C
	.dc.b	$F0,$12,$3F,$C5,$DD,$7C,$75,$FD,$6D,$3E,$73,$CC,$4A,$DF,$85,$E9
	.dc.b	$FA,$7F,$F1,$3D,$1A,$49,$FF,$64,$F5,$7F,$92,$70,$EB,$F3,$BF,$04
	.dc.b	$FC,$9F,$B7,$F0,$FC,$09,$7F,$EE,$8C,$51,$F9,$8F,$5B,$E1,$F1,$FC
	.dc.b	$22,$57,$C2,$7C,$7F,$21,$F3,$49,$1F,$9C,$F0,$3E,$87,$93,$23,$B3
	.dc.b	$FA,$4E,$DF,$A9,$F3,$7E,$C2,$4F,$FB,$8B,$3F,$E2,$EF,$F9,$23,$E9
	.dc.b	$FF,$54,$FF,$A7,$DA,$E7,$C4,$FF,$29,$7C,$52,$73,$3F,$83,$C1,$F0
	.dc.b	$9F,$3E,$3E,$07,$C6,$7E,$14,$FF,$27,$C6,$63,$F9,$27,$57,$FA,$65
	.dc.b	$E4,$D7,$FC,$CA,$D5,$5F,$CA,$BE,$3B,$EB,$FC,$9F,$E4,$49,$3F,$E6
	.dc.b	$FF,$D6,$BE,$B7,$83,$FE,$6E,$DB,$FB,$34,$FF,$33,$6F,$47,$E1,$BD
	.dc.b	$BF,$F3,$2E,$FE,$DF,$90,$FC,$E7,$C7,$FE,$10,$98,$7C,$97,$FC,$DB
	.dc.b	$BF,$F4,$3C,$3F,$1F,$F1,$DF,$A3,$3D,$FC,$BF,$E6,$EF,$D6,$FF,$A7
	.dc.b	$CF,$F4,$06,$EB,$FE,$82,$6E,$3E,$E7,$FD,$79,$3F,$D1,$B8,$BF,$DD
	.dc.b	$13,$FA,$6F,$4B,$F5,$24,$7F,$68,$7F,$A3,$32,$FF,$08,$57,$E8,$BE
	.dc.b	$63,$F6,$C7,$D2,$E2,$FA,$7A,$FE,$E7,$E6,$26,$1F,$E9,$BB,$FE,$AB
	.dc.b	$87,$FA,$0F,$4F,$95,$5F,$A6,$25,$7E,$2D,$F8,$78,$FD,$E2,$47,$FA
	.dc.b	$87,$9D,$E4,$62,$3E,$97,$BF,$FC,$24,$C6,$5E,$07,$F0,$E4,$8F,$26
	.dc.b	$3E,$DF,$83,$C4,$7D,$B7,$F8,$91,$8A,$FE,$DB,$E9,$7F,$AB,$25,$F2
	.dc.b	$AF,$5F,$FD,$63,$3F,$D1,$67,$FD,$C7,$18,$F5,$FF,$DF,$0E,$B9,$EC
	.dc.b	$FE,$4A,$BA,$5F,$38,$C5,$EB,$B7,$FD,$3A,$DA,$DE,$1D,$6B,$B9,$C0
	.dc.b	$B5,$C1,$B3,$C1,$FD,$0F,$0F,$F8,$DD,$DB,$5F,$6D,$7E,$07,$E0,$FC
	.dc.b	$52,$FA,$2B,$12,$B8,$BA,$8F,$6C,$3E,$A7,$FC,$73,$8F,$FB,$45,$B1
	.dc.b	$F8,$E3,$5C,$1F,$E3,$A7,$FA,$24,$69,$FC,$9F,$CB,$BE,$49,$F5,$3F
	.dc.b	$ED,$B8,$C3,$63,$33,$D6,$72,$F7,$A7,$13,$E6,$7C,$8D,$9D,$35,$FD
	.dc.b	$1C,$F0,$9F,$E4,$19,$FF,$99,$38,$7E,$C6,$8F,$0B,$E3,$72,$FD,$27
	.dc.b	$FC,$C8,$DA,$E1,$C3,$FC,$2F,$82,$E7,$FC,$52,$F8,$BF,$48,$C4,$6B
	.dc.b	$0E,$D7,$7E,$67,$FC,$29,$8F,$E0,$3B,$98,$67,$97,$84,$62,$7F,$0B
	.dc.b	$5C,$2B,$B6,$1F,$E4,$C7,$84,$F3,$69,$A9,$E1,$A7,$C0,$7F,$C5,$13
	.dc.b	$D3,$0F,$37,$D4,$7F,$BA,$A3,$FA,$C1,$FE,$8A,$E8,$FF,$2B,$C6,$70
	.dc.b	$FE,$55,$39,$3D,$38,$A7,$55,$C3,$E1,$75,$7F,$CE,$A5,$F5,$98,$BE
	.dc.b	$15,$AF,$15,$F5,$3F,$E0,$7C,$BF,$80,$E3,$C7,$E6,$75,$39,$8F,$F3
	.dc.b	$76,$5F,$EF,$7A,$DD,$F2,$F6,$F6,$77,$9F,$F8,$01,$3A,$BE,$5A,$7B
	.dc.b	$BE,$2F,$FC,$CB,$8F,$B2,$FF,$D1,$77,$FA,$1D,$3D,$D5,$AA,$F6,$36
	.dc.b	$3E,$17,$6F,$78,$F0,$36,$BD,$72,$FA,$97,$CE,$79,$74,$F8,$91,$8A
	.dc.b	$DE,$5E,$67,$E7,$F1,$7F,$9F,$1F,$23,$73,$FE,$87,$8A,$31,$1F,$DD
	.dc.b	$DA,$F2,$F9,$EF,$A5,$07,$E3,$7C,$66,$1F,$91,$F3,$5F,$27,$DE,$B8
	.dc.b	$F9,$1F,$99,$F0,$1F,$07,$33,$49,$3F,$C2,$78,$FE,$EF,$C3,$7D,$A3
	.dc.b	$EA,$7B,$56,$7E,$CF,$D5,$3E,$33,$F0,$8D,$8F,$87,$58,$9F,$4F,$BF
	.dc.b	$F4,$77,$34,$95,$F6,$73,$7C,$D7,$E0,$FE,$27,$F0,$8D,$9F,$9F,$ED
	.dc.b	$7E,$3A,$3F,$10,$D8,$FA,$7F,$31,$EF,$23,$D5,$F8,$5E,$B7,$E3,$CF
	.dc.b	$C8,$7D,$05,$9F,$B7,$E4,$FE,$4C,$FC,$A6,$B6,$3F,$0B,$AB,$E3,$FD
	.dc.b	$84,$8F,$BE,$FE,$3F,$E6,$7F,$77,$D3,$FE,$CF,$6F,$81,$FB,$88,$FD
	.dc.b	$74,$62,$BF,$6E,$BD,$3C,$F8,$DA,$6C,$FE,$5A,$5F,$21,$DC,$91,$FA
	.dc.b	$38,$CB,$F7,$7E,$A9,$EA,$FE,$ED,$B1,$CE,$71,$39,$EE,$93,$DF,$C3
	.dc.b	$B9,$FF,$BA,$7E,$A5,$EF,$7E,$B5,$B5,$FB,$36,$CF,$E8,$1A,$FF,$95
	.dc.b	$71,$FD,$26,$DF,$EF,$06,$D7,$C2,$CB,$E2,$FC,$97,$F1,$6B,$FE,$D1
	.dc.b	$E2,$5F,$91,$75,$E7,$D0,$FF,$29,$E3,$FC,$52,$D5,$A8,$C7,$D4,$FF
	.dc.b	$A9,$1F,$ED,$12,$8C,$3F,$DA,$DA,$FF,$8E,$9A,$FF,$AB,$FE,$39,$A7
	.dc.b	$96,$78,$EB,$FE,$FA,$9C,$DF,$83,$E4,$7E,$FF,$BB,$F8,$0E,$3B,$3F
	.dc.b	$6D,$FC,$1F,$D1,$63,$3F,$C4,$37,$1F,$8D,$F4,$3F,$88,$F7,$7F,$12
	.dc.b	$DC,$3F,$09,$E2,$E1,$A7,$E1,$E3,$F1,$FF,$43,$23,$F1,$5F,$43,$72
	.dc.b	$FC,$B7,$C8,$7E,$52,$3F,$2A,$D9,$FC,$87,$BB,$86,$9F,$93,$8F,$85
	.dc.b	$F2,$9F,$CC,$6F,$E6,$37,$8B,$FD,$81,$5F,$41,$B3,$F3,$F8,$B8,$CF
	.dc.b	$D6,$6C,$7E,$A7,$C1,$FD,$3F,$C8,$7E,$A1,$B5,$FA,$3E,$9C,$34,$FD
	.dc.b	$34,$7D,$5F,$93,$FC,$57,$CC,$2F,$15,$F8,$78,$C3,$F8,$6F,$94,$F4
	.dc.b	$11,$F4,$3E,$DF,$A6,$FF,$7F,$46,$CF,$C8,$38,$C9,$CA,$9F,$F7,$B6
	.dc.b	$BF,$ED,$07,$FA,$7F,$80,$F8,$8F,$EE,$0B,$E2,$EC,$61,$CF,$C3,$E2
	.dc.b	$B8,$B7,$1E,$F1,$DB,$E4,$BE,$81,$1F,$A0,$FD,$3F,$8E,$FC,$C3,$F9
	.dc.b	$96,$C7,$E9,$BE,$43,$F6,$1E,$57,$8E,$30,$7E,$B5,$C5,$F8,$0F,$F1
	.dc.b	$3E,$8F,$8E,$FC,$01,$2B,$F0,$0F,$C9,$F5,$7F,$50,$FE,$F7,$83,$FA
	.dc.b	$0C,$FE,$DA,$3F,$80,$7D,$3F,$F8,$27,$0F,$F2,$F3,$5F,$F5,$6B,$5F
	.dc.b	$F1,$B8,$70,$FF,$8E,$9F,$F0,$6B,$E7,$35,$F0,$BD,$BC,$7E,$EF,$F0
	.dc.b	$FF,$A7,$FE,$E1,$87,$F0,$66,$D7,$F9,$BD,$AF,$FA,$F9,$AF,$89,$FF
	.dc.b	$56,$C7,$F1,$37,$8F,$27,$E0,$E7,$FC,$9B,$1A,$7C,$8E,$A7,$FC,$73
	.dc.b	$F4,$FF,$C4,$27,$FD,$DE,$DE,$B3,$49,$F3,$5F,$E3,$97,$F8,$9D,$72
	.dc.b	$8C,$79,$39,$EF,$93,$FE,$20,$A7,$6E,$31,$FB,$98,$F6,$8E,$87,$5E
	.dc.b	$FB,$7C,$7B,$3F,$C5,$D8,$7E,$77,$0F,$F0,$FB,$5F,$AB,$87,$A2,$77
	.dc.b	$F3,$3E,$77,$D0,$DB,$F2,$7A,$1D,$08,$FF,$8F,$F0,$FE,$97,$87,$C6
	.dc.b	$F9,$9F,$81,$5F,$05,$B5,$ED,$EB,$F8,$87,$C5,$EB,$B1,$F0,$5B,$7F
	.dc.b	$03,$04,$63,$8B,$EC,$69,$F4,$3E,$5F,$DE,$4C,$3C,$8F,$C2,$F9,$6F
	.dc.b	$A4,$FD,$36,$C7,$CF,$E8,$FA,$DD,$BF,$5D,$B1,$F4,$FB,$FE,$59,$1F
	.dc.b	$80,$FC,$EF,$03,$EF,$3F,$7D,$B1,$FE,$5A,$E1,$7B,$9F,$E7,$B5,$EE
	.dc.b	$58,$C3,$9F,$6F,$FC,$CD,$3F,$E3,$BE,$1F,$F2,$BB,$5F,$B2,$D7,$FD
	.dc.b	$1A,$FF,$B0,$6F,$C1,$EE,$7E,$47,$C4,$F0,$3F,$93,$F5,$31,$3F,$8F
	.dc.b	$6D,$FF,$09,$F8,$3F,$F3,$7F,$0F,$FA,$3F,$D9,$3F,$E9,$1F,$90,$7A
	.dc.b	$99,$8C,$7A,$5F,$E7,$FB,$52,$30,$DC,$BF,$CB,$CD,$C7,$FC,$E7,$F4
	.dc.b	$FF,$E3,$15,$CE,$DF,$FD,$7D,$1F,$A2,$9D,$4E,$63,$FC,$28,$7F,$B1
	.dc.b	$BE,$47,$FD,$8F,$E8,$3E,$0B,$F8,$9C,$FC,$53,$FC,$90,$FF,$BD,$E7
	.dc.b	$6E,$B8,$7F,$82,$57,$1A,$E3,$5A,$7F,$8E,$CE,$AB,$94,$74,$33,$9E
	.dc.b	$DF,$E4,$75,$FE,$C2,$BD,$E1,$FE,$59,$9E,$A7,$C0,$8C,$9F,$CA,$BE
	.dc.b	$57,$E5,$76,$71,$9F,$F7,$1C,$61,$E0,$63,$6F,$EB,$EE,$04,$6E,$FC
	.dc.b	$11,$89,$DD,$FB,$E3,$7E,$BF,$F3,$A8,$C5,$FF,$A2,$27,$D8,$F8,$63
	.dc.b	$35,$CB,$3F,$45,$F1,$6D,$F0,$9D,$3F,$7B,$57,$C0,$FF,$48,$CF,$F7
	.dc.b	$B1,$FE,$71,$9F,$F6,$1B,$5F,$C1,$4E,$F5,$43,$87,$FD,$79,$DF,$FE
	.dc.b	$4C,$6C,$7F,$AB,$7E,$0B,$F8,$5B,$67,$5C,$CF,$03,$A4,$F3,$8C,$3F
	.dc.b	$EE,$86,$BF,$A7,$AF,$D2,$BF,$CE,$7E,$4F,$8F,$17,$F4,$4C,$AF,$85
	.dc.b	$FD,$3E,$BF,$8E,$56,$B5,$FF,$44,$C7,$BB,$D3,$C1,$C2,$F5,$EF,$7F
	.dc.b	$99,$B1,$5F,$F9,$82,$FF,$DE,$0D,$7F,$57,$3B,$8F,$F4,$7C,$7C,$AC
	.dc.b	$FC,$07,$87,$E0,$7C,$FF,$4C,$F9,$4B,$F2,$92,$FF,$DE,$73,$96,$C7
	.dc.b	$8E,$0D,$24,$7E,$C0,$FD,$DC,$7F,$A0,$9E,$7E,$9F,$80,$F0,$FA,$9E
	.dc.b	$2F,$93,$F1,$FC,$9C,$AF,$0C,$57,$FD,$F3,$69,$F1,$12,$3C,$1B,$25
	.dc.b	$F2,$7E,$EF,$92,$95,$EC,$8A,$FF,$6E,$94,$E3,$AB,$FA,$7C,$72,$9F
	.dc.b	$87,$F2,$3F,$23,$E4,$A4,$60,$F8,$EF,$C5,$23,$E4,$94,$7C,$9F,$92
	.dc.b	$FA,$3F,$25,$2B,$E7,$C5,$7D,$F1,$F3,$C8,$F3,$F8,$BF,$B2,$47,$EB
	.dc.b	$FC,$8F,$DC,$F8,$19,$1F,$69,$D3,$EF,$F0,$FF,$03,$F2,$12,$3E,$FC
	.dc.b	$57,$E7,$C7,$DF,$23,$FE,$B8,$F3,$FF,$69,$B7,$E7,$5F,$CD,$C7,$F8
	.dc.b	$F5,$B1,$9F,$90,$FD,$6C,$A7,$6F,$F3,$66,$BF,$E4,$A6,$BF,$BE,$6A
	.dc.b	$FA,$47,$B3,$D1,$FD,$EC,$7F,$93,$5B,$7F,$B1,$77,$1F,$ED,$57,$6F
	.dc.b	$F4,$66,$BF,$E5,$A6,$BF,$C3,$9A,$BE,$C1,$CF,$E1,$F0,$A4,$FF,$B8
	.dc.b	$67,$2F,$78,$FF,$6F,$BE,$96,$D1,$A6,$3C,$1F,$8F,$F4,$9F,$E9,$5A
	.dc.b	$FF,$76,$35,$F7,$47,$CC,$FA,$17,$6D,$ED,$95,$FE,$F3,$9C,$B6,$3C
	.dc.b	$71,$FB,$58,$FD,$C3,$5F,$ED,$2D,$7F,$97,$F6,$3E,$8F,$FB,$94,$7F
	.dc.b	$95,$97,$BD,$FC,$89,$B1,$FC,$DB,$7F,$F1,$CC,$70,$6C,$EF,$DF,$95
	.dc.b	$BC,$FB,$DF,$28,$FF,$44,$7D,$6F,$E3,$27,$FD,$1F,$97,$EE,$27,$DD
	.dc.b	$DF,$F8,$EA,$D4,$7F,$46,$9D,$3E,$E9,$F8,$24,$7F,$8C,$08,$77,$FE
	.dc.b	$63,$E3,$1F,$2B,$4F,$A1,$1F,$CA,$BD,$DF,$E5,$7E,$EC,$C3,$F9,$4C
	.dc.b	$57,$F8,$77,$BD,$FC,$C0,$C7,$07,$F4,$5C,$BF,$82,$CF,$D7,$F0,$BE
	.dc.b	$E1,$F7,$7C,$3B,$1E,$69,$4F,$DA,$24,$7F,$0F,$F6,$7F,$9E,$B0,$FE
	.dc.b	$76,$D7,$FD,$1B,$F4,$4F,$A2,$FE,$B2,$D8,$FE,$82,$63,$D6,$8F,$D5
	.dc.b	$3F,$C6,$F1,$FD,$E9,$C7,$3A,$39,$47,$6F,$07,$AF,$A1,$23,$73,$FC
	.dc.b	$2B,$9B,$F1,$AF,$F5,$BD,$BF,$CF,$EF,$DA,$C3,$A7,$A3,$C6,$FD,$3E
	.dc.b	$87,$A3,$96,$47,$48,$AF,$BB,$D0,$F6,$98,$C7,$FB,$A7,$D2,$B9,$BF
	.dc.b	$55,$E5,$4B,$F7,$09,$D3,$FA,$B8,$FD,$49,$1F,$A8,$FE,$B3,$CB,$FE
	.dc.b	$3F,$E0,$3C,$CA,$7E,$DE,$3E,$7F,$CC,$4B,$F8,$6F,$D2,$F9,$8F,$D5
	.dc.b	$7C,$27,$F8,$07,$C2,$FE,$01,$5F,$EB,$A6,$BF,$A6,$8F,$F1,$07,$57
	.dc.b	$F5,$78,$FF,$3E,$37,$6F,$F7,$07,$B5,$F8,$8F,$4B,$EF,$FC,$1F,$F8
	.dc.b	$82,$3F,$17,$E1,$5B,$FD,$0E,$1F,$7E,$72,$FD,$8F,$98,$FC,$3F,$BB
	.dc.b	$FE,$15,$EA,$FF,$40,$4F,$F7,$26,$BF,$E7,$2E,$AF,$C2,$4E,$FE,$63
	.dc.b	$F1,$FE,$FC,$AF,$C1,$C7,$E4,$BE,$5A,$47,$E1,$9F,$CA,$FC,$C7,$F8
	.dc.b	$BB,$E8,$BF,$96,$CF,$56,$BF,$ED,$E6,$BF,$C6,$1A,$FE,$63,$E6,$3F
	.dc.b	$3E,$47,$F2,$7F,$33,$FA,$2C,$E4,$7E,$59,$FD,$2F,$C2,$FF,$8F,$BD
	.dc.b	$E7,$C9,$FF,$8D,$5B,$5F,$E1,$26,$BF,$05,$AF,$FB,$C9,$7C,$A7,$D1
	.dc.b	$FA,$2F,$D3,$F4,$7D,$66,$F8,$FF,$CC,$2B,$77,$FA,$E5,$F2,$7E,$3F
	.dc.b	$95,$93,$98,$F8,$DD,$1F,$E4,$9E,$0F,$E0,$F3,$FE,$26,$6B,$F2,$63
	.dc.b	$FC,$D1,$F4,$7F,$E7,$F8,$FF,$62,$37,$3F,$C4,$B7,$2F,$27,$CB,$7A
	.dc.b	$7C,$D9,$7F,$87,$D7,$F7,$8F,$FA,$F3,$57,$CB,$7C,$9F,$90,$F9,$04
	.dc.b	$9F,$90,$FC,$BF,$6F,$E7,$7B,$72,$7E,$04,$7E,$07,$E1,$64,$7F,$10
	.dc.b	$E0,$FF,$4D,$6E,$DB,$E4,$7C,$1F,$E1,$FE,$17,$E1,$12,$FE,$13,$FE
	.dc.b	$6D,$F9,$1F,$F7,$BC,$7C,$DF,$8F,$FB,$C7,$F9,$F2,$33,$0C,$3B,$3E
	.dc.b	$DF,$92,$FC,$27,$C8,$48,$FA,$D1,$5F,$C1,$47,$E1,$C8,$FF,$99,$FC
	.dc.b	$67,$1B,$CD,$FC,$21,$27,$F0,$1F,$2D,$5F,$16,$DF,$85,$F9,$6F,$CA
	.dc.b	$19,$CF,$CB,$E0,$7F,$99,$FE,$2F,$F3,$04,$BF,$C5,$7C,$B5,$79,$9F
	.dc.b	$5F,$CF,$FC,$F1,$9B,$E7,$EC,$7F,$AC,$FE,$55,$E3,$7F,$81,$23,$0D
	.dc.b	$EE,$FF,$70,$B9,$F7,$EE,$3F,$EF,$EA,$DB,$7D,$FC,$EA,$9F,$E9,$92
	.dc.b	$C7,$BB,$C7,$1D,$CF,$2A,$E1,$78,$33,$3D,$3C,$07,$CB,$5F,$8B,$98
	.dc.b	$FF,$B2,$B2,$EB,$8F,$FA,$36,$B3,$3E,$1D,$3F,$65,$78,$BD,$E1,$D2
	.dc.b	$B9,$57,$FB,$51,$7D,$13,$76,$5F,$E0,$E8,$C3,$63,$33,$FE,$A9,$9C
	.dc.b	$BF,$EA,$BF,$97,$7A,$8F,$A3,$FF,$0B,$C6,$1B,$19,$9E,$0D,$7C,$87
	.dc.b	$93,$91,$F0,$BE,$36,$C7,$85,$AF,$0C,$3F,$CF,$13,$FE,$AE,$FD,$4F
	.dc.b	$FD,$F3,$D1,$FE,$F2,$3E,$06,$88,$CD,$FF,$B8,$63,$1C,$7F,$A8,$AD
	.dc.b	$D6,$9B,$1F,$AC,$E0,$FA,$73,$C9,$BC,$79,$E0,$FF,$29,$CE,$5F,$A1
	.dc.b	$F1,$3F,$ED,$BC,$3E,$9E,$43,$E9,$BC,$5C,$46,$DF,$BC,$6F,$4F,$E1
	.dc.b	$53,$CA,$13,$B8,$F9,$6D,$8F,$0F,$CB,$7B,$5B,$3F,$15,$A7,$63,$9C
	.dc.b	$78,$DE,$BF,$2D,$6B,$C9,$9F,$E5,$3E,$B7,$FA,$7E,$1F,$E2,$E6,$DF
	.dc.b	$EE,$36,$FA,$1F,$13,$FC,$53,$AF,$F8,$0D,$69,$FE,$07,$B2,$75,$87
	.dc.b	$F9,$F2,$B4,$FC,$C8,$CE,$1F,$98,$9C,$B7,$D3,$CE,$29,$FE,$B7,$7F
	.dc.b	$E0,$7F,$D5,$7F,$D2,$6F,$9E,$B9,$F8,$69,$CE,$1E,$9B,$E2,$FF,$12
	.dc.b	$9D,$7C,$97,$69,$CE,$9F,$C7,$3A,$FE,$AD,$FF,$6D,$CB,$F9,$9B,$FD
	.dc.b	$B5,$AF,$F6,$96,$BF,$83,$F9,$07,$BB,$88,$46,$27,$FD,$06,$D8,$FD
	.dc.b	$49,$A9,$E7,$AB,$F3,$3F,$C5,$63,$FC,$C1,$B8,$7C,$DF,$63,$E3,$C6
	.dc.b	$30,$FD,$D7,$EB,$2F,$84,$F2,$F7,$A3,$18,$78,$23,$13,$A9,$CD,$7E
	.dc.b	$F2,$7F,$AA,$7C,$C7,$D8,$87,$E0,$DF,$C4,$30,$FF,$3C,$56,$9F,$F2
	.dc.b	$EC,$66,$7F,$15,$8B,$A7,$CE,$7E,$0A,$DF,$07,$30,$FF,$3F,$D6,$DE
	.dc.b	$B0,$F8,$3F,$5B,$63,$D4,$DC,$3D,$0D,$7F,$84,$F1,$BD,$BF,$8F,$CB
	.dc.b	$F8,$16,$DF,$83,$E2,$FE,$59,$F8,$CF,$49,$F2,$38,$9F,$1F,$85,$E2
	.dc.b	$23,$E2,$7E,$7E,$DF,$77,$97,$EC,$71,$7B,$E3,$15,$FE,$E0,$E0,$FF
	.dc.b	$7F,$BF,$9D,$E4,$FC,$5F,$57,$E6,$DB,$D7,$F5,$38,$7F,$A1,$4E,$6B
	.dc.b	$ED,$7E,$B9,$EF,$9F,$E2,$98,$A3,$11,$FC,$F5,$B1,$9E,$4E,$BF,$11
	.dc.b	$2F,$FC,$11,$8F,$FB,$59,$B1,$F9,$4E,$47,$BB,$FB,$6F,$95,$79,$5F
	.dc.b	$AB,$8C,$3F,$77,$3F,$EF,$78,$79,$B8,$BF,$27,$18,$D7,$F1,$CD,$7F
	.dc.b	$9F,$F8,$BD,$2B,$FD,$82,$FF,$6E,$6D,$FF,$06,$8C,$4F,$EF,$FD,$BF
	.dc.b	$D1,$F8,$5F,$BD,$6D,$FE,$33,$8B,$F5,$BC,$7F,$2E,$FE,$94,$60,$F9
	.dc.b	$FF,$1F,$FA,$CF,$96,$FE,$1B,$C4,$8C,$61,$FC,$9F,$DC,$FD,$8F,$1F
	.dc.b	$F0,$B6,$C7,$FB,$63,$F2,$07,$87,$FB,$C6,$C7,$F0,$0F,$11,$18,$D3
	.dc.b	$FD,$38,$D7,$EF,$E4,$F7,$BF,$7F,$F2,$FE,$46,$E1,$FA,$8F,$86,$FE
	.dc.b	$06,$7F,$1D,$F0,$2C,$7F,$B4,$3E,$93,$F9,$1E,$3F,$DE,$1A,$FF,$83
	.dc.b	$5A,$FF,$5D,$6A,$7C,$8E,$F4,$FE,$3D,$18,$8E,$11,$F8,$7F,$80,$FE
	.dc.b	$2B,$F2,$7F,$CB,$DB,$7F,$EF,$CF,$D5,$BF,$C3,$4F,$F5,$18,$FE,$83
	.dc.b	$18,$BF,$E7,$51,$FC,$DF,$E0,$3F,$8F,$F1,$DB,$FE,$0E,$DA,$FE,$81
	.dc.b	$3C,$BD,$BC,$78,$BF,$CE,$BC,$F7,$C6,$FF,$1B,$6E,$3F,$17,$CA,$C4
	.dc.b	$FF,$1F,$DF,$F8,$44,$FF,$36,$CB,$E1,$57,$C0,$24,$66,$7F,$D2,$6F
	.dc.b	$33,$E8,$7C,$4F,$39,$FE,$A1,$F0,$5F,$CE,$3E,$52,$D7,$F2,$07,$FA
	.dc.b	$DF,$57,$F8,$7B,$1F,$DF,$B5,$FF,$08,$79,$48,$C4,$23,$1E,$DF,$ED
	.dc.b	$FE,$0F,$FA,$2F,$CF,$7F,$5F,$6E,$3F,$DF,$FE,$07,$FB,$11,$FD,$C3
	.dc.b	$F5,$DF,$E2,$78,$FF,$69,$8F,$EE,$7D,$EF,$8C,$A7,$F4,$3F,$6D,$BE
	.dc.b	$31,$FC,$77,$C6,$7E,$10,$8F,$C5,$7F,$0F,$E4,$48,$F9,$0D,$CB,$FD
	.dc.b	$F9,$F3,$5F,$E3,$EE,$8F,$F4,$7E,$6C,$3F,$C0,$B3,$9C,$3F,$C6,$5E
	.dc.b	$2F,$EA,$A7,$3C,$32,$7F,$C6,$D8,$7F,$A6,$3F,$0A,$7C,$76,$73,$FF
	.dc.b	$4A,$9F,$9D,$E2,$FF,$08,$F2,$FF,$96,$B0,$FD,$44,$63,$E1,$7F,$09
	.dc.b	$C7,$FA,$02,$4F,$E0,$5F,$F0,$07,$17,$F1,$9C,$3A,$B7,$3F,$F0,$5F
	.dc.b	$DA,$B1,$F9,$5E,$87,$F6,$DF,$A7,$FE,$C1,$87,$F8,$11,$AF,$F9,$B9
	.dc.b	$AF,$FA,$0D,$AF,$F9,$4D,$FF,$10,$BB,$F7,$3F,$C2,$51,$88,$E1,$18
	.dc.b	$DB,$FB,$E7,$D4,$78,$32,$FF,$0F,$DF,$C6,$D7,$E4,$9F,$E9,$19,$D3
	.dc.b	$E1,$F2,$FA,$91,$FD,$34,$FF,$17,$7F,$51,$E0,$FE,$19,$E7,$97,$FA
	.dc.b	$63,$F2,$D7,$9B,$FE,$0D,$AD,$4F,$F9,$46,$73,$5C,$23,$F7,$3B,$7F
	.dc.b	$8B,$6B,$E5,$4F,$68,$C1,$FC,$FB,$49,$1F,$CB,$FD,$2F,$65,$FF,$A6
	.dc.b	$3A,$3F,$C5,$D5,$FD,$C3,$95,$F3,$F4,$98,$7F,$BA,$E3,$2F,$F3,$9F
	.dc.b	$6B,$EE,$3F,$E5,$FE,$3F,$DC,$F5,$FF,$93,$5B,$1F,$E6,$D6,$E5,$FE
	.dc.b	$55,$7F,$C7,$3C,$3F,$D5,$09,$3E,$77,$C9,$F2,$12,$3C,$4D,$AF,$56
	.dc.b	$AF,$6F,$FA,$E9,$2B,$E0,$3E,$DF,$9B,$91,$EB,$6D,$7C,$3C,$FF,$B5
	.dc.b	$F3,$7F,$9F,$BE,$3F,$F8,$9F,$C1,$FF,$99,$DB,$87,$FA,$1D,$B7,$FE
	.dc.b	$6F,$7F,$CA,$7E,$DF,$F6,$E2,$4F,$D1,$7F,$C1,$5E,$74,$8F,$94,$DA
	.dc.b	$F5,$FC,$FF,$F9,$70,$FF,$41,$F9,$BF,$E5,$DE,$6F,$F3,$24,$7F,$9E
	.dc.b	$6F,$FC,$28,$43,$EA,$1F,$E6,$AF,$25,$14,$91,$FD,$C6,$32,$FF,$9C
	.dc.b	$BC,$B7,$EB,$DF,$C3,$71,$FE,$58,$D8,$7C,$AB,$E4,$3F,$D2,$33,$98
	.dc.b	$FE,$67,$3A,$F2,$3F,$C7,$71,$82,$7C,$37,$FA,$61,$FF,$54,$63,$FA
	.dc.b	$4A,$EB,$39,$F1,$BF,$D4,$7F,$92,$B9,$B8,$4E,$1B,$7F,$E6,$8F,$27
	.dc.b	$FD,$06,$7F,$9D,$FB,$7F,$4B,$B7,$F9,$CA,$3F,$CF,$4F,$E1,$E7,$FA
	.dc.b	$E3,$4F,$EB,$DD,$3F,$E9,$97,$07,$FB,$CB,$F5,$6F,$DE,$3F,$CA,$2D
	.dc.b	$EB,$7F,$AE,$27,$3D,$5F,$EA,$89,$CE,$1F,$B9,$BF,$F7,$47,$C8,$7F
	.dc.b	$BA,$7E,$73,$F9,$9C,$7F,$BA,$CF,$E6,$01,$CB,$98,$E1,$E2,$6E,$37
	.dc.b	$3A,$7A,$9C,$63,$5A,$2F,$8C,$F1,$BA,$FF,$BE,$E7,$4F,$13,$43,$A3
	.dc.b	$C8,$E1,$AF,$94,$D6,$1F,$EB,$CA,$FE,$AA,$FB,$7E,$57,$FD,$3F,$9E
	.dc.b	$28,$23,$2F,$23,$2E,$2F,$BB,$6B,$A7,$72,$FD,$73,$FE,$D2,$BB,$F2
	.dc.b	$1F,$36,$FE,$5C,$F6,$6D,$F1,$AF,$80,$78,$EB,$86,$9E,$96,$C7,$92
	.dc.b	$78,$57,$DA,$C5,$B7,$F1,$3A,$FE,$71,$F8,$B7,$D7,$CF,$FD,$D6,$DD
	.dc.b	$BF,$BC,$D6,$EB,$FB,$9E,$8F,$9F,$FC,$AB,$63,$FC,$64,$D7,$FD,$D2
	.dc.b	$D7,$F1,$58,$FF,$BA,$5F,$C6,$63,$FE,$E9,$6E,$1F,$EB,$16,$9F,$4B
	.dc.b	$81,$F0,$58,$B7,$33,$E7,$F7,$D2,$FA,$C7,$89,$86,$E5,$FD,$7B,$0E
	.dc.b	$37,$9A,$D7,$87,$2B,$F8,$05,$F0,$7E,$D6,$9F,$E3,$E8,$FC,$9C,$7E
	.dc.b	$7F,$1F,$04,$FF,$94,$DB,$1F,$C6,$67,$6F,$FB,$1B,$BF,$F8,$43,$FC
	.dc.b	$35,$AF,$63,$F8,$EF,$5F,$B3,$EB,$9F,$31,$F0,$7C,$D7,$CD,$FF,$98
	.dc.b	$77,$FE,$CA,$D8,$CE,$1F,$E9,$B6,$BF,$DB,$7C,$7F,$F8,$00,$FF,$A7
	.dc.b	$DA,$E7,$E5,$7A,$46,$23,$EC,$FE,$B0,$F9,$DF,$A4,$D8,$F9,$8D,$7D
	.dc.b	$58,$A9,$EC,$7D,$95,$F2,$FD,$89,$5E,$37,$D9,$6F,$26,$92,$7F,$D6
	.dc.b	$D1,$97,$FC,$05,$F2,$BD,$7E,$5E,$47,$F3,$9F,$55,$38,$8F,$2E,$2A
	.dc.b	$C7,$6F,$F0,$EE,$6F,$F3,$D5,$FE,$33,$0F,$C7,$61,$FE,$0F,$EA,$FB
	.dc.b	$D1,$88,$FC,$13,$5F,$8F,$E6,$FA,$88,$FA,$1E,$3F,$2B,$23,$FC,$34
	.dc.b	$D7,$FC,$45,$A3,$CD,$F8,$64,$8F,$82,$F8,$BC,$5F,$99,$F2,$92,$7E
	.dc.b	$04,$57,$DF,$17,$5F,$76,$5F,$C9,$9F,$81,$FA,$FB,$E8,$FF,$34,$DB
	.dc.b	$FE,$E4,$D7,$F4,$0D,$7F,$0B,$E8,$7D,$82,$3F,$59,$FE,$B7,$FA,$DF
	.dc.b	$EA,$E3,$FA,$33,$5F,$E9,$CD,$7F,$98,$B5,$2F,$0B,$F8,$17,$E6,$5F
	.dc.b	$DF,$E7,$F7,$8D,$7F,$D1,$33,$B7,$FC,$AD,$ED,$B8,$3F,$B6,$B6,$3F
	.dc.b	$D9,$4D,$7F,$AD,$07,$0F,$F5,$D7,$FD,$AE,$FF,$98,$7E,$3B,$F0,$5A
	.dc.b	$7F,$B6,$F8,$FF,$8C,$CF,$1F,$7A,$FD,$6F,$FB,$16,$3F,$DB,$0D,$7B
	.dc.b	$B5,$FE,$D4,$D7,$FB,$93,$FC,$89,$E3,$E5,$7F,$81,$A3,$11,$CA,$3F
	.dc.b	$C1,$38,$7F,$33,$C3,$FD,$DD,$E6,$F9,$23,$1E,$4F,$F7,$DF,$BC,$FC
	.dc.b	$76,$1F,$E7,$BC,$7E,$03,$E6,$D3,$B1,$1F,$4F,$97,$FC,$CF,$8F,$E0
	.dc.b	$E1,$FE,$C0,$6C,$FF,$6A,$BF,$E5,$59,$FF,$7A,$F0,$9E,$6F,$F2,$08
	.dc.b	$C3,$70,$FF,$64,$B6,$FF,$BF,$7C,$C7,$BA,$B7,$E5,$FF,$43,$F8,$AF
	.dc.b	$3D,$7F,$10,$6C,$EA,$3F,$D2,$DF,$59,$FC,$41,$FB,$A1,$C3,$F8,$06
	.dc.b	$BF,$E1,$B6,$BF,$E2,$37,$F8,$43,$C9,$AF,$1F,$0F,$FC,$73,$18,$8F
	.dc.b	$EC,$35,$F9,$37,$CD,$D3,$E9,$3F,$0F,$F5,$8F,$83,$FE,$4D,$F3,$72
	.dc.b	$BF,$AC,$78,$E8,$FE,$C5,$F8,$C7,$EC,$F4,$F9,$92,$F9,$DF,$E6,$BE
	.dc.b	$7C,$C3,$8C,$7F,$66,$E5,$7C,$AF,$EF,$89,$3F,$BE,$7F,$A0,$F1,$7F
	.dc.b	$3E,$F4,$64,$F8,$A3,$F3,$BC,$3E,$38,$F9,$B1,$83,$3A,$D8,$F8,$D1
	.dc.b	$F1,$08,$FD,$2E,$BF,$C1,$BF,$AB,$E5,$FC,$B1,$C5,$FE,$6F,$F1,$5F
	.dc.b	$D2,$48,$FF,$7B,$E0,$7C,$9E,$23,$FA,$8F,$97,$D1,$B7,$9E,$2F,$C9
	.dc.b	$5B,$E4,$7F,$AA,$7C,$E4,$BE,$91,$5E,$7F,$5B,$FD,$80,$C5,$FF,$80
	.dc.b	$B7,$5F,$E0,$09,$5F,$85,$7F,$07,$F3,$92,$3F,$10,$D7,$F9,$FF,$C6
	.dc.b	$FF,$6C,$23,$F8,$08,$FB,$1E,$6F,$F7,$0E,$E9,$3F,$E9,$5E,$6F,$E1
	.dc.b	$59,$7E,$B7,$F4,$AF,$ED,$F3,$FC,$6B,$C1,$46,$23,$FD,$DC,$D7,$FC
	.dc.b	$6A,$D7,$FC,$C3,$1F,$DE,$7E,$6F,$FB,$FF,$D4,$49,$FC,$9C,$7E,$2F
	.dc.b	$E6,$7F,$C0,$7F,$0B,$27,$C9,$1F,$E0,$6F,$86,$FE,$F6,$48,$FE,$F2
	.dc.b	$FF,$83,$BE,$6F,$FC,$17,$EC,$C9,$FC,$FC,$7F,$8F,$FE,$55,$F3,$57
	.dc.b	$86,$57,$F8,$EE,$72,$FF,$9E,$37,$FF,$43,$FD,$1B,$F6,$07,$D2,$7F
	.dc.b	$00,$8C,$36,$33,$5F,$D0,$27,$2F,$F8,$97,$E7,$BF,$CA,$1D,$3D,$A3
	.dc.b	$FC,$3A,$DC,$3F,$63,$3F,$AD,$F4,$3F,$B4,$92,$BF,$B4,$BD,$F3,$91
	.dc.b	$DF,$1F,$E7,$CB,$1F,$44,$FF,$72,$F3,$E4,$FF,$72,$6B,$FE,$AF,$97
	.dc.b	$C8,$7F,$0D,$CB,$F9,$77,$83,$FC,$CB,$9C,$FF,$34,$AD,$F9,$DF,$E6
	.dc.b	$E5,$CF,$F0,$B1,$E8,$D7,$FD,$85,$1A,$E0,$F9,$53,$99,$FC,$3D,$FF
	.dc.b	$4C,$7F,$D6,$8D,$C3,$C3,$EA,$7F,$9D,$BD,$5F,$ED,$D9,$FF,$18,$F9
	.dc.b	$6F,$F2,$5F,$B3,$31,$FE,$3F,$1E,$5C,$BE,$2E,$9E,$6E,$0F,$09,$2B
	.dc.b	$FC,$87,$EC,$7F,$95,$08,$FE,$0F,$87,$FD,$77,$5E,$EF,$17,$E0,$7C
	.dc.b	$8B,$BF,$FB,$BF,$E1,$8E,$CA,$6A,$3E,$37,$E7,$6F,$98,$FC,$53,$73
	.dc.b	$FC,$43,$5F,$A8,$D7,$FC,$D3,$F0,$7F,$E6,$82,$3F,$CF,$3A,$7F,$BA
	.dc.b	$78,$7F,$DE,$09,$1F,$C9,$5F,$F3,$1F,$3C,$8F,$F3,$0B,$5F,$F4,$97
	.dc.b	$DD,$FF,$5A,$E6,$FE,$87,$1F,$E6,$F3,$FD,$4B,$19,$E1,$FD,$EF,$AF
	.dc.b	$FE,$BF,$9F,$F7,$53,$6F,$FC,$F1,$1C,$3E,$27,$FD,$61,$F1,$F9,$FA
	.dc.b	$BF,$F3,$F6,$5F,$DE,$DB,$8F,$F0,$89,$D7,$C6,$FF,$AD,$E7,$38,$7F
	.dc.b	$BD,$DB,$8A,$BC,$52,$FA,$C7,$D2,$FF,$BC,$A1,$CD,$FE,$35,$9C,$C7
	.dc.b	$FB,$DA,$31,$3B,$75,$DD,$7B,$B9,$5F,$0B,$FF,$59,$B7,$4E,$D1,$E1
	.dc.b	$C3,$C1,$18,$9D,$4F,$29,$E8,$8F,$47,$CC,$E1,$A7,$96,$F8,$34,$66
	.dc.b	$FB,$CF,$F2,$C6,$7F,$E5,$38,$C7,$C7,$76,$6D,$71,$78,$7C,$86,$6F
	.dc.b	$FC,$C1,$A7,$FB,$B9,$B5,$FC,$E9,$AF,$81,$AF,$CF,$FA,$4F,$1B,$FE
	.dc.b	$60,$D5,$18,$8F,$84,$D7,$E5,$B5,$F8,$5F,$31,$FA,$BF,$03,$FC,$5A
	.dc.b	$D8,$FF,$35,$B5,$F9,$ED,$7E,$AF,$C4,$FF,$1A,$7C,$4D,$7E,$93,$5D
	.dc.b	$F5,$7F,$5D,$E4,$FF,$63,$67,$F1,$5B,$3F,$88,$E0,$FE,$B7,$0D,$D5
	.dc.b	$BD,$AD,$EC,$FC,$B3,$8E,$0F,$A9,$7D,$1B,$79,$C3,$AD,$5B,$FC,$27
	.dc.b	$8B,$FE,$F8,$C7,$F0,$C7,$E3,$9C,$47,$80,$DF,$CA,$F8,$A7,$4D,$8F
	.dc.b	$D1,$4F,$0C,$7F,$D2,$CD,$C3,$3F,$2B,$E1,$A4,$E5,$F1,$C6,$63,$11
	.dc.b	$A8,$C5,$FA,$78,$DA,$7F,$B4,$72,$FE,$55,$7F,$98,$F9,$FF,$F6,$34
	.dc.b	$62,$B7,$F2,$9F,$2B,$AF,$FC,$FD,$5A,$B4,$7F,$07,$9D,$5F,$D4,$6C
	.dc.b	$FF,$05,$F8,$8B,$87,$4A,$F0,$FC,$AB,$EA,$5F,$2F,$AC,$DF,$5B,$CA
	.dc.b	$B5,$39,$8C,$57,$0B,$E6,$DC,$BC,$B1,$F8,$6F,$84,$DD,$F9,$25,$F0
	.dc.b	$EF,$73,$EC,$47,$F1,$58,$C5,$FF,$96,$E7,$33,$FC,$16,$72,$FF,$0E
	.dc.b	$D3,$E6,$FC,$27,$59,$CF,$07,$F2,$B8,$C5,$FB,$7E,$AA,$63,$F4,$BB
	.dc.b	$5F,$5B,$FE,$7F,$8F,$A5,$C2,$AF,$47,$D6,$FF,$0E,$9F,$F4,$FB,$75
	.dc.b	$FF,$6E,$B5,$FE,$42,$D7,$FB,$3E,$6F,$93,$FA,$B9,$7D,$A7,$EA,$FE
	.dc.b	$21,$FE,$2C,$C3,$C9,$5B,$7C,$DE,$68,$E2,$57,$CF,$F6,$58,$8F,$9D
	.dc.b	$FA,$67,$F8,$FA,$7F,$D6,$AD,$CB,$FD,$E6,$D7,$F9,$BB,$5F,$F2,$BE
	.dc.b	$BF,$E9,$C9,$F8,$7F,$25,$F8,$AE,$B7,$C2,$FE,$1E,$3F,$0D,$E6,$3E
	.dc.b	$72,$9F,$8B,$F4,$6E,$1F,$A0,$C9,$9B,$DB,$FB,$B1,$86,$C6,$67,$D3
	.dc.b	$1F,$82,$E2,$FF,$1E,$73,$7C,$A7,$EF,$CF,$F1,$4F,$6F,$13,$F9,$BF
	.dc.b	$8F,$FC,$BF,$CA,$5A,$FC,$BB,$F9,$3F,$CC,$9F,$25,$E2,$6C,$FF,$B7
	.dc.b	$1B,$A7,$F5,$C0,$E1,$FE,$D9,$DC,$FB,$0F,$DB,$74,$7F,$22,$F1,$3F
	.dc.b	$5D,$F2,$9F,$B2,$9C,$E3,$FD,$7E,$72,$FE,$43,$E0,$7F,$65,$F0,$BF
	.dc.b	$B3,$6C,$7C,$9C,$FF,$D4,$AF,$97,$E4,$FD,$04,$7F,$D2,$DE,$6B,$D3
	.dc.b	$FE,$02,$D8,$FE,$1D,$93,$9F,$35,$FB,$CB,$FE,$05,$F1,$B6,$FF,$46
	.dc.b	$FC,$7F,$A7,$F8,$C4,$8F,$8C,$FE,$DB,$5F,$E1,$9F,$0D,$FC,$18,$F8
	.dc.b	$1D,$36,$BF,$83,$FC,$67,$F0,$F8,$FE,$16,$D8,$F7,$6B,$F4,$C8,$FD
	.dc.b	$37,$B7,$9B,$FE,$FD,$8F,$E2,$3F,$31,$F3,$7E,$43,$F8,$AB,$FC,$6F
	.dc.b	$C0,$93,$F9,$1F,$B0,$FE,$3A,$FE,$FA,$3F,$8A,$47,$EF,$E3,$EF,$74
	.dc.b	$A3,$0F,$E0,$5A,$FF,$87,$9A,$FF,$8B,$57,$8F,$D3,$CB,$EA,$DA,$FA
	.dc.b	$3E,$E7,$F2,$FF,$80,$FC,$23,$63,$24,$C3,$F9,$59,$FE,$1E,$E7,$7C
	.dc.b	$7F,$09,$FE,$2F,$F0,$3F,$CE,$4F,$C7,$7C,$75,$8F,$BE,$DB,$E7,$E0
	.dc.b	$7E,$5E,$31,$E0,$7F,$69,$DB,$F8,$05,$78,$0F,$E6,$FE,$2F,$C6,$F2
	.dc.b	$7F,$97,$3F,$23,$BB,$19,$7C,$D6,$D7,$1F,$AE,$FF,$51,$CE,$4F,$E0
	.dc.b	$4E,$23,$B6,$1F,$93,$9F,$EA,$BD,$3F,$E1,$0C,$BF,$B0,$F3,$7F,$53
	.dc.b	$8C,$4F,$F1,$68,$FE,$BB,$F5,$6F,$73,$5E,$FE,$1B,$5F,$CF,$39,$3F
	.dc.b	$AF,$7D,$6F,$F5,$F6,$C7,$F7,$1B,$FE,$61,$A7,$C6,$6E,$7F,$E1,$CE
	.dc.b	$24,$3E,$97,$C3,$EA,$FF,$5D,$F7,$9B,$53,$F1,$BE,$45,$CF,$F2,$5E
	.dc.b	$2F,$F2,$0E,$2F,$E6,$1A,$FF,$98,$5A,$FE,$46,$35,$C5,$F9,$67,$F2
	.dc.b	$ED,$AF,$F0,$47,$B9,$F1,$2B,$FC,$C6,$DD,$9F,$60,$FA,$CF,$F0,$1C
	.dc.b	$E6,$31,$7C,$B9,$7F,$C0,$F1,$8C,$3F,$0F,$5D,$5F,$D2,$71,$FE,$A3
	.dc.b	$93,$F5,$2D,$C3,$F5,$DE,$56,$1A,$7E,$C6,$3F,$C0,$5F,$60,$AA,$7F
	.dc.b	$81,$7E,$63,$ED,$57,$F4,$79,$D3,$FD,$26,$3F,$C1,$D8,$7F,$A3,$9A
	.dc.b	$FF,$A7,$5A,$FF,$A2,$DA,$97,$C6,$F7,$F8,$1F,$03,$6B,$B7,$8F,$E3
	.dc.b	$DF,$C8,$D8,$F1,$79,$B8,$69,$C0,$FE,$BD,$DF,$E8,$CB,$F5,$4F,$E1
	.dc.b	$23,$3A,$7F,$88,$E3,$13,$AD,$71,$39,$6D,$7E,$93,$E0,$FF,$97,$72
	.dc.b	$1A,$48,$FE,$AD,$C9,$F9,$6F,$9F,$FD,$44,$FE,$99,$B3,$FA,$EE,$3F
	.dc.b	$F2,$07,$1F,$EB,$5B,$19,$26,$7F,$A3,$F4,$35,$1F,$B6,$F9,$EF,$19
	.dc.b	$1F,$EC,$DE,$8F,$F7,$D8,$FF,$1C,$E7,$FA,$8F,$9E,$FD,$D4,$7E,$E1
	.dc.b	$B5,$FC,$03,$E7,$BD,$A4,$79,$EC,$8E,$7E,$13,$E5,$FF,$82,$47,$F0
	.dc.b	$26,$D7,$F0,$AF,$82,$FF,$2E,$79,$9F,$C2,$5B,$19,$26,$DF,$BF,$E7
	.dc.b	$D4,$7F,$90,$BC,$BF,$4F,$C1,$1A,$48,$F9,$F9,$BA,$7F,$CD,$53,$FC
	.dc.b	$51,$B3,$FE,$7D,$F2,$FF,$D2,$FE,$4F,$F0,$B6,$C7,$F8,$C0,$FF,$3B
	.dc.b	$B8,$CB,$FD,$BB,$83,$E5,$BE,$09,$2B,$C2,$FC,$4F,$8E,$FE,$46,$FF
	.dc.b	$24,$6C,$7F,$A0,$3E,$BB,$FD,$70,$DF,$8F,$FF,$3F,$47,$F0,$9E,$73
	.dc.b	$AC,$7F,$A8,$C6,$3C,$BF,$EF,$D3,$FE,$9B,$9F,$F7,$FC,$E6,$FF,$03
	.dc.b	$39,$9F,$F7,$E0,$9E,$BF,$09,$E0,$3F,$D2,$BF,$59,$F7,$6F,$CD,$18
	.dc.b	$E6,$91,$FE,$A5,$EF,$FE,$39,$5B,$C3,$FA,$B7,$6F,$EB,$F8,$7F,$D8
	.dc.b	$4D,$BF,$0B,$FE,$F2,$7F,$CA,$4E,$67,$FD,$87,$1F,$E0,$07,$FD,$7E
	.dc.b	$DF,$6B,$FC,$AB,$18,$3F,$D2,$1F,$5B,$FE,$11,$7F,$B5,$B6,$FF,$A3
	.dc.b	$61,$FE,$75,$9C,$F3,$FF,$A8,$50,$F9,$1B,$F3,$1F,$CD,$63,$E3,$F1
	.dc.b	$FF,$BA,$A3,$11,$F2,$A7,$7E,$3F,$CC,$85,$6E,$3F,$CA,$D3,$A9,$E7
	.dc.b	$A5,$79,$0B,$EA,$F2,$8D,$3E,$69,$F3,$F8,$3B,$8F,$04,$62,$BB,$C2
	.dc.b	$74,$F6,$6F,$C9,$FF,$A6,$74,$E1,$3F,$0A,$30,$F1,$AF,$C3,$C6,$2B
	.dc.b	$8F,$6F,$FB,$82,$B4,$71,$71,$5F,$1A,$BF,$D9,$93,$EF,$6B,$EF,$CB
	.dc.b	$AD,$70,$6E,$1B,$F8,$7F,$CB,$E1,$CD,$B3,$AB,$E4,$7B,$F7,$FE,$00
	.dc.b	$FF,$BE,$5A,$FE,$13,$B3,$F4,$38,$7E,$13,$E2,$AE,$DE,$EF,$A8,$FC
	.dc.b	$86,$A6,$93,$0F,$F2,$9F,$23,$E0,$38,$C7,$E2,$A3,$0F,$E7,$23,$2A
	.dc.b	$1E,$9F,$0E,$17,$2F,$7A,$FC,$9E,$9F,$98,$C7,$53,$99,$FF,$23,$F3
	.dc.b	$BB,$3C,$3D,$6F,$AD,$FF,$29,$D7,$2D,$FF,$D3,$91,$88,$FE,$12,$DC
	.dc.b	$BB,$46,$CF,$E0,$91,$FC,$B5,$FE,$92,$D7,$F8,$6B,$5F,$E1,$75,$E1
	.dc.b	$F0,$BF,$CE,$EB,$EB,$5F,$2B,$E6,$6D,$77,$9D,$4F,$AA,$D8,$FC,$5C
	.dc.b	$FF,$B6,$5F,$F0,$FF,$A6,$C1,$98,$B1,$FD,$B9,$47,$FA,$23,$7F,$EE
	.dc.b	$8D,$8C,$CF,$F7,$4F,$15,$E9,$FD,$36,$C7,$F4,$58,$C7,$BB,$E3,$F9
	.dc.b	$4D,$7A,$B2,$BF,$91,$45,$7F,$81,$65,$FE,$BB,$9F,$16,$1F,$E1,$E9
	.dc.b	$F0,$47,$C6,$F3,$7F,$DF,$31,$FE,$6E,$EA,$FB,$71,$88,$FB,$FE,$2E
	.dc.b	$27,$CD,$EE,$C9,$E0,$7F,$33,$F0,$B8,$F4,$C8,$C1,$FE,$7A,$7C,$04
	.dc.b	$7F,$BD,$BF,$16,$BC,$3F,$2F,$27,$F9,$F4,$57,$FA,$6E,$29,$F5,$12
	.dc.b	$3F,$C0,$BE,$1F,$A8,$8F,$F4,$7F,$A0,$AF,$FA,$83,$5F,$F3,$C3,$FD
	.dc.b	$CF,$E0,$33,$D7,$FE,$7D,$E4,$7B,$3F,$9E,$FA,$57,$D6,$D3,$F0,$D1
	.dc.b	$70,$FF,$0F,$CE,$9B,$3F,$94,$8C,$47,$F3,$71,$F5,$9F,$EF,$57,$F4
	.dc.b	$46,$3D,$4F,$DC,$C7,$E9,$FA,$FF,$B4,$BB,$9F,$F5,$A3,$97,$FA,$5B
	.dc.b	$FA,$7C,$BF,$37,$18,$F0,$39,$DF,$F0,$4F,$65,$BD,$E1,$97,$E2,$9C
	.dc.b	$B6,$38,$6B,$23,$07,$A3,$EA,$FF,$86,$B7,$8F,$F8,$8C,$1B,$E2,$E1
	.dc.b	$FE,$DD,$19,$9F,$F7,$7B,$5F,$8C,$D7,$FA,$F7,$D6,$7F,$11,$7E,$3B
	.dc.b	$5F,$F6,$B3,$75,$F8,$30,$9F,$E4,$AF,$FB,$6D,$7B,$BE,$6C,$E4,$FA
	.dc.b	$E2,$BB,$D7,$D6,$62,$FF,$D8,$5F,$25,$D1,$FF,$62,$79,$3F,$E5,$5A
	.dc.b	$FF,$59,$F6,$FF,$2F,$AD,$71,$FF,$6A,$AD,$57,$FA,$B2,$BF,$9A,$46
	.dc.b	$2F,$AF,$95,$FC,$EB,$EA,$3F,$82,$DF,$FB,$CD,$FD,$E3,$7C,$0D,$78
	.dc.b	$9B,$7F,$C3,$53,$A9,$FF,$03,$35,$FF,$19,$B5,$FF,$09,$FD,$5B,$D4
	.dc.b	$FE,$92,$D8,$FC,$AB,$5F,$17,$87,$FD,$2F,$EA,$FF,$3B,$9F,$FB,$95
	.dc.b	$BB,$7F,$82,$1B,$5F,$E0,$5E,$EF,$F0,$33,$FE,$63,$F8,$1F,$E4,$5F
	.dc.b	$2F,$23,$07,$F4,$5F,$1D,$6F,$90,$FD,$A7,$C6,$CA,$D4,$E5,$FE,$7D
	.dc.b	$E1,$27,$F9,$77,$34,$9C,$C7,$F6,$6F,$B7,$FC,$46,$BF,$13,$4F,$A0
	.dc.b	$67,$D2,$B7,$FB,$6E,$36,$3F,$54,$CC,$7D,$57,$FB,$8F,$D0,$7F,$71
	.dc.b	$8F,$DE,$B5,$D1,$E8,$F9,$AF,$E8,$5F,$35,$23,$07,$FA,$1D,$C3,$FC
	.dc.b	$65,$FA,$11,$F4,$7E,$3E,$4F,$C2,$7E,$A7,$87,$88,$FE,$41,$F1,$3F
	.dc.b	$D3,$CC,$73,$7E,$9B,$36,$3F,$8F,$99,$5E,$F8,$AF,$F1,$99,$4F,$E3
	.dc.b	$89,$1F,$D7,$7E,$23,$FA,$E1,$1F,$90,$FF,$2B,$F1,$64,$7D,$67,$F1
	.dc.b	$5F,$29,$FE,$18,$FC,$BB,$FC,$21,$3F,$E5,$D6,$BF,$E8,$A6,$BF,$E6
	.dc.b	$A6,$FB,$2E,$65,$7F,$6D,$F9,$99,$7F,$84,$8F,$E7,$9F,$27,$23,$F0
	.dc.b	$EF,$E6,$FC,$7F,$17,$9B,$FE,$83,$CB,$F9,$1D,$7F,$39,$6B,$E5,$8F
	.dc.b	$E7,$D3,$FD,$EF,$E3,$24,$FE,$4A,$3F,$C6,$1F,$1F,$FC,$9A,$1C,$5F
	.dc.b	$CD,$BB,$D3,$98,$C3,$7A,$FF,$CE,$3F,$4B,$FE,$1A,$BE,$31,$FE,$AC
	.dc.b	$F4,$1F,$31,$FE,$99,$6C,$FF,$4F,$3F,$B4,$46,$4F,$F0,$2F,$C2,$FF
	.dc.b	$84,$7E,$6A,$4F,$FB,$56,$3F,$C8,$D8,$7F,$0E,$CB,$FA,$5F,$27,$CD
	.dc.b	$F3,$FF,$B2,$1A,$F2,$FF,$C9,$F9,$FE,$C7,$1F,$EB,$CD,$7F,$D2,$CD
	.dc.b	$7D,$1E,$37,$F8,$5B,$CB,$FE,$DD,$B4,$9F,$F7,$34,$7F,$97,$BE,$77
	.dc.b	$FD,$3B,$EA,$DF,$8A,$FE,$BD,$7F,$EA,$A9,$CD,$FF,$07,$DB,$F6,$AF
	.dc.b	$F6,$CF,$AA,$FE,$52,$FF,$A9,$DA,$FF,$B5,$1A,$FF,$AD,$1B,$B7,$F9
	.dc.b	$BF,$4F,$F2,$FF,$A3,$FE,$11,$AF,$F2,$53,$5F,$F0,$67,$C4,$7F,$8E
	.dc.b	$FE,$32,$47,$FB,$FD,$F9,$7D,$DF,$E7,$4C,$7F,$98,$49,$A4,$E1,$FB
	.dc.b	$1A,$29,$E3,$7C,$36,$DF,$F9,$A6,$75,$3F,$E4,$BF,$90,$F9,$1F,$1A
	.dc.b	$FA,$AA,$4B,$FE,$59,$F9,$6A,$BC,$9D,$FF,$EE,$1D,$0D,$25,$7F,$96
	.dc.b	$79,$3F,$CB,$04,$7C,$5F,$57,$CD,$FA,$5C,$1F,$C8,$BE,$1A,$46,$0F
	.dc.b	$E4,$6F,$FB,$BF,$CE,$34,$91,$FE,$27,$E5,$7D,$5F,$E4,$FA,$3F,$DD
	.dc.b	$31,$8E,$4F,$F4,$9C,$E6,$BF,$DA,$5D,$1E,$37,$F4,$E2,$B8,$3F,$D3
	.dc.b	$7E,$13,$FD,$D5,$DD,$FB,$36,$E3,$DB,$0E,$9E,$1B,$1C,$E1,$D8,$FD
	.dc.b	$39,$FE,$AD,$FA,$A7,$D5,$E2,$3F,$AD,$7D,$36,$1B,$19,$7C,$A3,$EA
	.dc.b	$FF,$C2,$FD,$3F,$8D,$7F,$8D,$C7,$FB,$CF,$BC,$C8,$CA,$9E,$1D,$FA
	.dc.b	$8C,$55,$AD,$D7,$5F,$33,$4D,$AE,$C7,$5B,$CD,$F1,$AE,$9B,$75,$AF
	.dc.b	$44,$61,$D1,$E6,$8D,$61,$E6,$7C,$9A,$B4,$7C,$C7,$6B,$4E,$63,$0D
	.dc.b	$8F,$2B,$C7,$E7,$F3,$5C,$A7,$2D,$7E,$09,$D1,$C6,$9F,$BB,$AE,$53
	.dc.b	$F0,$0F,$A5,$E5,$FF,$85,$9F,$4B,$5F,$94,$D7,$E9,$35,$7C,$AF,$EC
	.dc.b	$7C,$2F,$E6,$4D,$8F,$F4,$13,$5F,$94,$D7,$EB,$7C,$2B,$B7,$F9,$D6
	.dc.b	$C8,$C4,$FF,$A2,$E3,$13,$A9,$CC,$FC,$EE,$36,$BF,$E8,$4D,$3E,$04
	.dc.b	$63,$7F,$96,$FD,$26,$CE,$F8,$7F,$1F,$B2,$B5,$87,$CC,$8C,$56,$E3
	.dc.b	$EC,$CE,$B0,$EF,$39,$97,$8B,$FE,$F6,$C7,$E8,$36,$3E,$B9,$AF,$90
	.dc.b	$E1,$19,$BF,$E6,$0A,$33,$F2,$7A,$6D,$F4,$70,$FD,$DE,$37,$B3,$8C
	.dc.b	$7D,$59,$BE,$0B,$FD,$53,$D9,$FD,$D6,$1E,$EE,$31,$F8,$71,$B7,$F7
	.dc.b	$B1,$9B,$F8,$37,$77,$F9,$0D,$B7,$CC,$58,$C7,$C3,$F1,$F0,$1F,$48
	.dc.b	$F8,$8C,$E9,$FC,$03,$5F,$EE,$73,$98,$FE,$1B,$7A,$D3,$F8,$FE,$9F
	.dc.b	$5F,$E5,$31,$87,$AF,$E5,$3F,$D4,$B5,$EA,$6B,$F6,$1B,$5F,$7D,$BB
	.dc.b	$3E,$77,$D5,$1E,$B6,$E5,$FE,$DA,$FA,$9F,$DC,$BF,$E1,$16,$E1,$FE
	.dc.b	$E1,$6B,$FC,$25,$AF,$F7,$BC,$5F,$17,$E4,$EF,$FF,$13,$B6,$33,$7F
	.dc.b	$EE,$B9,$CB,$FE,$ED,$CD,$F5,$2F,$A8,$FA,$4D,$9F,$E6,$7F,$1F,$89
	.dc.b	$FA,$9E,$64,$9F,$E6,$FE,$A5,$D3,$FA,$4E,$43,$FA,$BC,$F8,$5B,$1F
	.dc.b	$C8,$DD,$57,$A5,$CC,$7E,$CF,$E5,$FF,$04,$7E,$0F,$E7,$6D,$FF,$46
	.dc.b	$F8,$1F,$C3,$1F,$87,$F9,$9B,$1F,$DF,$F0,$47,$52,$47,$8D,$FA,$FF
	.dc.b	$2B,$F8,$F3,$F2,$1E,$2D,$9F,$EA,$7C,$1F,$93,$F9,$9F,$CA,$36,$3F
	.dc.b	$1C,$7F,$49,$FD,$25,$97,$E1,$B5,$46,$27,$FC,$9C,$D8,$FE,$88,$6A
	.dc.b	$7F,$17,$F5,$0F,$AC,$FD,$22,$F9,$BB,$18,$73,$F0,$7A,$F8,$7B,$97
	.dc.b	$FB,$0E,$FF,$CB,$EF,$9B,$1F,$EC,$79,$34,$F3,$36,$7C,$B1,$83,$C7
	.dc.b	$EC,$78,$8F,$1F,$CA,$D9,$ED,$CF,$E6,$F9,$2F,$3B,$63,$C5,$F0,$7E
	.dc.b	$C2,$3F,$91,$CF,$FC,$6E,$DF,$8E,$F7,$9F,$03,$CB,$B1,$EB,$F6,$3E
	.dc.b	$1F,$B3,$F1,$1B,$1F,$EE,$1F,$A6,$7C,$EF,$ED,$1B,$1F,$9F,$6F,$0F
	.dc.b	$F7,$3E,$74,$62,$3F,$B7,$E6,$F8,$FF,$A2,$7D,$2F,$A9,$B8,$7C,$BA
	.dc.b	$7D,$47,$CE,$7E,$DB,$5F,$F0,$03,$5F,$F0,$73,$5F,$F0,$AB,$FE,$F7
	.dc.b	$8F,$F4,$93,$63,$9F,$3F,$E2,$E3,$11,$E6,$FA,$7F,$E5,$4F,$8A,$7F
	.dc.b	$0F,$18,$9E,$5E,$37,$FA,$A7,$27,$C7,$7E,$33,$DF,$FC,$6B,$72,$FC
	.dc.b	$43,$87,$F1,$5E,$E5,$CF,$F9,$3D,$BE,$37,$F2,$A7,$F3,$9F,$A6,$FE
	.dc.b	$A9,$97,$78,$C7,$B7,$31,$FC,$A1,$70,$FF,$1D,$E8,$EC,$F4,$1E,$CF
	.dc.b	$A9,$FF,$3E,$57,$CF,$9D,$3F,$46,$33,$5F,$AA,$6B,$FE,$2F,$6B,$FE
	.dc.b	$3D,$6B,$FE,$44,$F9,$F7,$CB,$F7,$F9,$1F,$97,$18,$BE,$D1,$ED,$F7
	.dc.b	$E4,$7E,$AF,$A1,$BB,$E2,$B3,$97,$F4,$C8,$C5,$6D,$B1,$9F,$6F,$E4
	.dc.b	$57,$C9,$F5,$25,$EF,$B2,$DF,$3F,$9F,$F9,$D1,$F3,$DB,$3F,$17,$C0
	.dc.b	$FC,$31,$1F,$C5,$BF,$89,$F0,$64,$7E,$39,$B7,$FC,$4B,$F0,$9F,$E2
	.dc.b	$13,$F3,$E7,$F1,$B1,$89,$FA,$F1,$FD,$DF,$C0,$7C,$AD,$31,$F1,$F7
	.dc.b	$0F,$DA,$38,$7F,$6D,$F2,$B7,$0F,$C0,$F6,$FE,$5C,$FF,$01,$7B,$1F
	.dc.b	$E0,$78,$CD,$7C,$F8,$C7,$6A,$3F,$C5,$F1,$87,$FC,$63,$97,$F8,$A1
	.dc.b	$E1,$F0,$5F,$A6,$F9,$7F,$F0,$16,$1F,$CC,$1A,$FF,$9D,$5A,$FF,$A0
	.dc.b	$DA,$FF,$A2,$9F,$B2,$EF,$8F,$F9,$F4,$62,$38,$47,$F4,$1D,$FF,$C5
	.dc.b	$37,$FC,$1E,$BF,$03,$97,$A6,$FF,$D2,$B5,$A7,$FD,$3B,$3A,$AF,$16
	.dc.b	$BE,$93,$E9,$F8,$76,$3F,$89,$5A,$33,$18,$F1,$AE,$1F,$EA,$8F,$51
	.dc.b	$E2,$FF,$08,$6D,$7F,$1B,$FC,$17,$F9,$2E,$9E,$5B,$56,$9F,$F5,$8C
	.dc.b	$EA,$BF,$D5,$0D,$8F,$F4,$9F,$E6,$9F,$56,$7A,$74,$7F,$05,$FA,$0F
	.dc.b	$F1,$3F,$61,$A4,$9F,$F0,$A4,$3E,$83,$84,$7E,$A5,$B8,$7E,$B7,$E7
	.dc.b	$BF,$C8,$5D,$4F,$86,$FE,$25,$18,$6D,$FE,$C6,$3F,$8B,$7D,$06,$2F
	.dc.b	$F6,$DE,$2F,$90,$95,$E2,$7C,$FE,$27,$F1,$D7,$F8,$F3,$63,$F7,$9F
	.dc.b	$33,$EC,$23,$EA,$7D,$FF,$3B,$FC,$95,$FE,$4C,$D8,$FE,$09,$DF,$EA
	.dc.b	$F8,$2F,$E5,$4D,$8C,$93,$1F,$E1,$9E,$0E,$A3,$F8,$67,$5F,$CC,$23
	.dc.b	$F2,$9F,$9F,$F1,$FF,$CD,$5F,$E6,$CD,$8F,$E2,$9F,$39,$F5,$08,$FD
	.dc.b	$27,$EB,$FC,$9C,$8F,$AA,$DC,$3F,$B5,$72,$7F,$B6,$5F,$E4,$11,$FD
	.dc.b	$0D,$FE,$88,$FF,$9D,$FE,$AB,$FD,$0B,$5F,$E2,$0F,$8F,$46,$2B,$FD
	.dc.b	$CD,$18,$F9,$1F,$F3,$7B,$6B,$85,$7F,$A1,$1B,$F2,$3A,$8C,$5F,$26
	.dc.b	$BF,$EE,$AF,$A9,$79,$1F,$E8,$F6,$DF,$F3,$BF,$1F,$FD,$7D,$5F,$E8
	.dc.b	$49,$CB,$FE,$86,$8C,$F8,$3F,$B9,$F3,$BF,$D4,$5F,$2B,$FE,$99,$97
	.dc.b	$CC,$D8,$FE,$EB,$6E,$5A,$63,$92,$E7,$FE,$0B,$ED,$7B,$7F,$E4,$96
	.dc.b	$C7,$F9,$1A,$FF,$3D,$39,$3F,$D2,$3A,$FF,$9B,$35,$F2,$36,$BF,$DA
	.dc.b	$6B,$E3,$7E,$07,$4F,$93,$07,$CB,$1F,$F2,$27,$55,$71,$CF,$FC,$5F
	.dc.b	$E1,$3F,$9F,$C6,$4E,$27,$03,$84,$6E,$7F,$CB,$B5,$D6,$B8,$D7,$4C
	.dc.b	$B1,$5A,$8E,$46,$4A,$79,$6B,$FD,$DB,$5B,$E3,$FF,$2D,$4F,$8E,$72
	.dc.b	$D1,$1A,$5F,$30,$C9,$AF,$09,$CC,$61,$E3,$87,$C3,$AE,$15,$B6,$CE
	.dc.b	$8F,$94,$E2,$7D,$EB,$E3,$BE,$64,$61,$B9,$75,$C3,$C1,$3C,$1B,$1E
	.dc.b	$19,$DD,$F6,$6B,$E8,$C3,$E5,$BF,$03,$67,$C7,$FD,$B9,$FC,$2B,$75
	.dc.b	$FC,$2B,$6B,$85,$D5,$F6,$BF,$91,$7F,$0C,$D7,$26,$5C,$68,$F8,$EE
	.dc.b	$4D,$9F,$F3,$43,$5F,$A3,$87,$CD,$F8,$EF,$16,$EE,$97,$76,$1B,$7F
	.dc.b	$B7,$8C,$7C,$C7,$0D,$FF,$D7,$F3,$FE,$35,$6B,$FB,$BF,$90,$E5,$5E
	.dc.b	$FF,$01,$F5,$3F,$E6,$4A,$FE,$28,$D7,$F8,$9B,$5F,$D2,$47,$0F,$86
	.dc.b	$FD,$A3,$CB,$E1,$6D,$7F,$99,$6B,$9C,$60,$FD,$D6,$9E,$DF,$0F,$D9
	.dc.b	$AC,$AF,$E2,$51,$5F,$F4,$B4,$7C,$37,$FD,$F0,$D7,$2F,$41,$F3,$14
	.dc.b	$91,$C1,$D7,$77,$CB,$F9,$79,$3F,$C8,$62,$BE,$08,$E8,$47,$C3,$E2
	.dc.b	$7D,$32,$3F,$CA,$7E,$5E,$BF,$70,$6A,$F4,$3F,$27,$1F,$ED,$38,$C7
	.dc.b	$57,$0E,$BF,$E5,$9A,$7F,$17,$AF,$E6,$9D,$BF,$81,$3F,$05,$E0,$D8
	.dc.b	$F8,$71,$F0,$48,$FC,$6E,$DF,$C4,$61,$23,$F9,$F4,$57,$E6,$C7,$C9
	.dc.b	$23,$F3,$BA,$7F,$20,$47,$FA,$37,$3D,$7D,$D9,$FB,$7B,$FF,$19,$AF
	.dc.b	$F0,$FD,$7D,$78,$7E,$67,$B7,$F3,$5E,$B4,$C7,$FA,$84,$57,$EF,$C7
	.dc.b	$DB,$23,$FD,$07,$F2,$97,$C0,$7E,$D5,$B1,$FE,$D3,$6B,$F4,$7A,$3F
	.dc.b	$88,$70,$7E,$EF,$EA,$9F,$03,$49,$5C,$70,$FE,$01,$F4,$1F,$5A,$3E
	.dc.b	$D0,$71,$AF,$F9,$15,$BE,$17,$F4,$75,$D3,$FC,$27,$06,$F4,$98,$7F
	.dc.b	$BA,$6B,$8B,$FD,$47,$E3,$7F,$B4,$BF,$EE,$DF,$A9,$F5,$E5,$FE,$53
	.dc.b	$6E,$5F,$D5,$6B,$75,$E6,$9F,$82,$FF,$55,$6B,$FD,$6D,$AF,$F5,$F6
	.dc.b	$A7,$F2,$CF,$F6,$F6,$BF,$D7,$CD,$71,$7F,$23,$8C,$C7,$F6,$7D,$DF
	.dc.b	$B0,$FF,$4D,$8F,$F6,$33,$5F,$A8,$DC,$7F,$C1,$5E,$D3,$E4,$BF,$98
	.dc.b	$61,$FE,$76,$BF,$85,$F0,$18,$3F,$92,$E1,$C7,$4F,$EF,$19,$7F,$99
	.dc.b	$F3,$FE,$53,$97,$A3,$3E,$B1,$8E,$1C,$FC,$94,$CF,$FD,$B5,$68,$E3
	.dc.b	$5F,$D0,$7F,$11,$79,$FF,$D0,$FE,$A5,$F1,$94,$D4,$5C,$7F,$C2,$D3
	.dc.b	$A6,$CF,$F4,$98,$C4,$7C,$B7,$FD,$D0,$F0,$F8,$9D,$11,$FE,$3F,$DB
	.dc.b	$FD,$07,$69,$1F,$E3,$78,$FE,$8F,$F5,$6F,$9B,$A4,$9F,$F1,$E5,$6B
	.dc.b	$E6,$E4,$7F,$BE,$69,$1C,$BE,$73,$9F,$44,$AD,$3C,$1F,$D1,$F4,$BC
	.dc.b	$8F,$EC,$DF,$57,$FD,$52,$7F,$4E,$DE,$6F,$D3,$4E,$BD,$1F,$ED,$53
	.dc.b	$BA,$FF,$15,$35,$FF,$18,$35,$FF,$1A,$70,$7F,$A9,$DF,$F1,$B5,$FF
	.dc.b	$AD,$63,$11,$FB,$66,$C7,$FA,$E6,$7C,$5D,$1F,$86,$D5,$F2,$DF,$20
	.dc.b	$98,$7C,$3F,$9C,$F9,$7F,$39,$23,$E3,$3A,$7F,$86,$57,$EC,$9F,$F1
	.dc.b	$6B,$78,$4D,$1F,$39,$F5,$BB,$3E,$BF,$81,$2F,$E9,$47,$DA,$EB,$FC
	.dc.b	$77,$C4,$C9,$FA,$D1,$5F,$C4,$47,$E2,$C8,$FE,$43,$E2,$BF,$1C,$47
	.dc.b	$07,$EE,$72,$F2,$FC,$E7,$E2,$09,$1F,$83,$EF,$FF,$08,$FD,$4B,$E7
	.dc.b	$3F,$BB,$B6,$79,$35,$FF,$31,$5A,$31,$1F,$C0,$F6,$7C,$97,$E5,$BE
	.dc.b	$0A,$4F,$E4,$63,$F3,$3E,$34,$8F,$CA,$3F,$9C,$F9,$3F,$CE,$F9,$5F
	.dc.b	$E0,$8C,$3F,$9C,$D7,$AD,$AF,$89,$AF,$87,$DB,$FD,$2F,$BB,$23,$F4
	.dc.b	$11,$FE,$2E,$FA,$6F,$57,$67,$FA,$6D,$BD,$7F,$C3,$1B,$5F,$7A,$3F
	.dc.b	$C8,$DF,$40,$FD,$3F,$BC,$FF,$A2,$63,$19,$FF,$A5,$DB,$1F,$CA,$5A
	.dc.b	$FC,$8B,$FE,$07,$87,$87,$E7,$3C,$87,$97,$CA,$F3,$7C,$E7,$9F,$CB
	.dc.b	$98,$7E,$8B,$4F,$34,$F9,$34,$47,$85,$B3,$E1,$6B,$FE,$99,$CF,$D5
	.dc.b	$3F,$E3,$4C,$DF,$39,$F1,$7E,$5E,$5F,$73,$FD,$17,$3A,$71,$CE,$87
	.dc.b	$A1,$F3,$49,$3A,$8F,$DA,$E1,$F1,$FF,$4E,$FC,$3C,$DE,$AF,$91,$18
	.dc.b	$86,$68,$C6,$3D,$FB,$7E,$D1,$2B,$B7,$77,$DD,$F9,$69,$1E,$17,$F0
	.dc.b	$1F,$2F,$F6,$B0,$FE,$63,$3F,$0D,$AF,$FB,$07,$1F,$07,$17,$90,$F8
	.dc.b	$DF,$2D,$F8,$92,$4F,$B7,$E5,$FF,$53,$F2,$D2,$3E,$2F,$82,$8F,$C8
	.dc.b	$7C,$7F,$E2,$BE,$43,$FC,$7D,$5F,$4D,$AF,$85,$AF,$E3,$5A,$FF,$B5
	.dc.b	$FE,$AD,$87,$FA,$55,$B1,$FE,$E7,$6B,$FD,$E5,$AF,$F7,$7E,$4F,$F7
	.dc.b	$A3,$F5,$7E,$4B,$F4,$87,$E9,$7E,$4E,$C7,$F0,$99,$4F,$E1,$C9,$19
	.dc.b	$CB,$F5,$84,$7F,$01,$F3,$15,$FF,$23,$FD,$57,$D0,$F0,$3F,$2B,$F0
	.dc.b	$E8,$C7,$57,$EC,$70,$FF,$51,$FD,$57,$F8,$7A,$FF,$37,$D4,$8C,$57
	.dc.b	$F8,$09,$B8,$FE,$5B,$DE,$E3,$F3,$1F,$BF,$88,$7E,$62,$76,$BE,$AB
	.dc.b	$FC,$04,$FF,$BB,$FE,$28,$C8,$CA,$E4,$FF,$5D,$46,$1F,$F5,$E7,$E3
	.dc.b	$5F,$E4,$97,$3E,$87,$2A,$E1,$76,$DE,$8E,$6F,$9B,$67,$57,$6F,$C5
	.dc.b	$35,$F4,$36,$F9,$C7,$A4,$F4,$4F,$1D,$7E,$47,$CD,$5C,$DF,$31,$4C
	.dc.b	$45,$BE,$55,$F3,$E7,$11,$EA,$11,$D1,$F0,$1F,$2F,$D1,$CB,$FE,$49
	.dc.b	$C5,$18,$8F,$88,$6E,$7F,$C9,$CD,$3F,$CA,$1F,$32,$F9,$2F,$E0,$71
	.dc.b	$86,$BC,$A3,$5E,$0F,$26,$C7,$87,$CF,$FF,$1D,$B7,$8F,$BE,$5F,$E8
	.dc.b	$46,$CE,$A7,$FC,$B9,$F3,$0D,$7F,$95,$F1,$7F,$BE,$F4,$7C,$DF,$E1
	.dc.b	$B9,$BE,$3D,$ED,$F9,$31,$9C,$3E,$45,$7F,$BC,$30,$E7,$C9,$FE,$5A
	.dc.b	$AD,$63,$D7,$1E,$58,$70,$9C,$D7,$F9,$83,$F1,$47,$CC,$7F,$3E,$CF
	.dc.b	$FC,$E5,$18,$7C,$17,$F5,$A3,$0F,$F4,$2C,$D1,$89,$E3,$5E,$4A,$E0
	.dc.b	$DA,$DF,$C2,$5B,$E7,$9F,$C3,$7F,$C8,$98,$7E,$21,$CE,$3E,$7A,$E1
	.dc.b	$87,$3C,$FF,$23,$7F,$B3,$F8,$4F,$EA,$FD,$7F,$BE,$9C,$E9,$FE,$BB
	.dc.b	$E7,$56,$AF,$FD,$03,$5A,$BE,$1C,$F9,$C3,$F2,$B2,$F9,$AC,$6E,$8E
	.dc.b	$18,$72,$F9,$8C,$E7,$B8,$FE,$BB,$29,$CD,$F2,$8C,$4E,$A7,$C1,$E0
	.dc.b	$7F,$1B,$F1,$35,$1D,$0F,$EE,$39,$7F,$17,$6B,$FA,$AC,$79,$BF,$E2
	.dc.b	$C6,$BF,$92,$DB,$96,$1E,$46,$C7,$FB,$7F,$1F,$A2,$FC,$7F,$9A,$F3
	.dc.b	$70,$7B,$1F,$83,$F8,$BF,$CC,$9F,$F5,$C3,$5F,$F5,$5B,$5F,$DE,$B7
	.dc.b	$6F,$C3,$FC,$BF,$CA,$E4,$37,$E3,$F0,$FC,$5E,$EF,$97,$A5,$D7,$F9
	.dc.b	$83,$F3,$BE,$53,$E0,$E3,$FC,$ED,$BB,$FF,$BA,$AB,$FC,$B0,$DD,$3E
	.dc.b	$26,$DF,$58,$F9,$FF,$B1,$3C,$5C,$5F,$F3,$DF,$9C,$FB,$67,$D4,$FC
	.dc.b	$FF,$FC,$B9,$19,$3F,$C3,$3C,$3F,$3B,$1F,$83,$C9,$E4,$8C,$47,$C2
	.dc.b	$8F,$3F,$27,$FB,$F2,$BF,$A5,$70,$7E,$0F,$E5,$FF,$12,$F9,$3A,$7E
	.dc.b	$EF,$43,$6F,$98,$D8,$CC,$7C,$DE,$07,$CC,$FC,$22,$63,$F0,$9F,$8B
	.dc.b	$F3,$BF,$96,$3F,$37,$F6,$9F,$EC,$19,$FE,$B9,$1F,$9A,$F9,$E9,$1F
	.dc.b	$E3,$69,$FF,$5F,$BF,$D8,$63,$F0,$7C,$5E,$F2,$47,$FA,$1F,$85,$1F
	.dc.b	$53,$E7,$BF,$4A,$78,$3E,$46,$D7,$5F,$92,$FD,$4F,$0F,$91,$B1,$FD
	.dc.b	$BF,$C3,$FB,$A4,$7E,$EB,$F8,$0E,$CF,$D8,$9F,$72,$AC,$FF,$77,$ED
	.dc.b	$FD,$A9,$F8,$09,$B1,$F6,$A7,$EC,$11,$F5,$3F,$8C,$EC,$FD,$E9,$F2
	.dc.b	$3C,$DB,$3F,$E5,$DE,$97,$CA,$FE,$72,$31,$F5,$2F,$7B,$F8,$65,$6B
	.dc.b	$2B,$39,$AE,$5F,$54,$DF,$F7,$B6,$8C,$4F,$F5,$E6,$C7,$F6,$23,$53
	.dc.b	$C2,$3F,$DA,$F3,$D2,$3C,$31,$89,$78,$FF,$E7,$C8,$E6,$FF,$89,$B9
	.dc.b	$7F,$C1,$B7,$FE,$76,$F8,$6F,$C3,$5F,$08,$FF,$0E,$EC,$AD,$CE,$4F
	.dc.b	$EF,$B5,$C2,$FA,$7A,$7F,$90,$3F,$23,$F2,$17,$1E,$DF,$33,$FC,$BA
	.dc.b	$CF,$E5,$BE,$43,$0D,$9F,$CA,$47,$E6,$BD,$6C,$65,$E1,$D3,$FD,$27
	.dc.b	$F3,$5C,$CF,$CF,$FC,$85,$C3,$F3,$8E,$1F,$CF,$7C,$85,$BE,$5B,$FF
	.dc.b	$A9,$27,$27,$E8,$CF,$D4,$CF,$CC,$F9,$5F,$BB,$E9,$7F,$B6,$27,$ED
	.dc.b	$C6,$3E,$4F,$E8,$FC,$AF,$D5,$24,$FD,$B7,$F8,$DF,$CB,$BE,$53,$E1
	.dc.b	$7C,$D7,$F5,$08,$C5,$7D,$38,$C7,$CA,$E6,$AC,$FF,$AB,$2B,$7A,$FD
	.dc.b	$33,$FB,$0F,$17,$E4,$BF,$15,$7B,$BE,$88,$C3,$F8,$6E,$7F,$C5,$13
	.dc.b	$0F,$C8,$3F,$8D,$D7,$51,$F9,$86,$E1,$F4,$7F,$18,$FB,$33,$FE,$BE
	.dc.b	$6C,$7D,$5F,$C5,$FF,$40,$EF,$A3,$FB,$6C,$63,$AF,$FC,$EE,$DB,$FC
	.dc.b	$7A,$F9,$1D,$F8,$3F,$AC,$FA,$BF,$7D,$FF,$B5,$BE,$E9,$BE,$78,$BF
	.dc.b	$63,$18,$CB,$F7,$16,$E3,$FD,$7B,$63,$FC,$96,$DD,$7F,$5C,$D7,$FC
	.dc.b	$A3,$E1,$7F,$44,$7E,$07,$85,$FB,$CF,$96,$FF,$03,$F0,$FE,$B3,$0F
	.dc.b	$F3,$D6,$1F,$02,$FF,$CB,$2D,$73,$B7,$88,$F1,$EF,$ED,$7D,$CD,$AF
	.dc.b	$E2,$1F,$88,$7F,$8D,$38,$3F,$DE,$BA,$3E,$6F,$F8,$7C,$E6,$BF,$D1
	.dc.b	$53,$97,$8F,$BB,$FC,$3A,$7F,$B3,$65,$FE,$6C,$6B,$FE,$91,$6B,$FE
	.dc.b	$9C,$FD,$6B,$FC,$A9,$8F,$FA,$B3,$F5,$BF,$F3,$AE,$1F,$C8,$E7,$2F
	.dc.b	$F0,$68,$CD,$7E,$4F,$F5,$BF,$EE,$13,$FE,$78,$6C,$7F,$1B,$6B,$FE
	.dc.b	$7E,$6B,$FC,$43,$C7,$FE,$42,$FF,$A9,$7C,$64,$62,$35,$C7,$FE,$06
	.dc.b	$8C,$47,$FA,$43,$BB,$E1,$3F,$C1,$E3,$EB,$39,$DF,$F8,$95,$7F,$13
	.dc.b	$F4,$26,$5F,$AD,$BB,$7F,$C3,$F0,$F0,$CE,$4F,$D9,$7C,$E7,$EE,$6B
	.dc.b	$F8,$E4,$F8,$C9,$3F,$CE,$72,$F4,$C7,$F0,$2F,$9B,$DC,$FE,$ED,$B8
	.dc.b	$79,$FE,$87,$D8,$49,$E5,$82,$3F,$85,$F2,$FF,$2B,$7F,$85,$36,$7F
	.dc.b	$9E,$78,$BF,$CB,$CF,$E1,$DF,$2F,$63,$F8,$53,$70,$F7,$37,$8F,$FD
	.dc.b	$D7,$8B,$C6,$F9,$C4,$BF,$1B,$FC,$73,$E6,$3F,$9E,$3F,$C6,$DB,$1F
	.dc.b	$C5,$FE,$5B,$FA,$07,$D0,$FF,$1F,$6C,$7F,$1B,$F9,$7F,$E4,$A4,$7E
	.dc.b	$47,$02,$3F,$94,$FC,$B7,$F2,$77,$F9,$43,$67,$FD,$E7,$99,$FF,$7A
	.dc.b	$FC,$1F,$FA,$CB,$DC,$FF,$38,$FC,$33,$D9,$FF,$39,$FC,$87,$F8,$6A
	.dc.b	$F8,$9C,$27,$4E,$38,$98,$7F,$AF,$3E,$1B,$FC,$73,$7F,$ED,$1A,$FE
	.dc.b	$ED,$39,$6C,$66,$BF,$DB,$73,$FE,$C4,$F8,$3F,$09,$E2,$F8,$3B,$7F
	.dc.b	$CF,$6B,$0D,$3C,$55,$FE,$CC,$9F,$ED,$59,$49,$CF,$E2,$5F,$C8,$6F
	.dc.b	$F4,$13,$FE,$90,$F8,$E7,$B3,$FE,$50,$6D,$FA,$1B,$5E,$66,$BF,$B1
	.dc.b	$F8,$C7,$C1,$FF,$A9,$9F,$F2,$F4,$7F,$73,$73,$5F,$1F,$BD,$1A,$E2
	.dc.b	$FF,$4D,$FC,$8A,$DF,$08,$EA,$FF,$7F,$FC,$9F,$D9,$9C,$C7,$F8,$86
	.dc.b	$B8,$56,$E7,$90,$AE,$11,$98,$CC,$F0,$AE,$B9,$59,$E2,$27,$0F,$96
	.dc.b	$F9,$D7,$86,$19,$7A,$1E,$35,$FE,$26,$C3,$53,$99,$F0,$9C,$23,$CF
	.dc.b	$F2,$BF,$E7,$AC,$B8,$4F,$C3,$8C,$3B,$31,$C7,$FE,$3A,$70,$23,$77
	.dc.b	$DD,$E3,$7F,$05,$71,$7C,$68,$C3,$72,$F1,$E1,$D6,$78,$36,$3C,$F3
	.dc.b	$BC,$3C,$CD,$7C,$53,$B7,$E2,$EB,$F1,$9F,$C1,$7E,$2C,$FA,$6F,$AF
	.dc.b	$A7,$DD,$D1,$39,$AF,$C4,$5F,$17,$FC,$9D,$CA,$FD,$8D,$F3,$FF,$E6
	.dc.b	$39,$FD,$9C,$61,$FD,$8B,$5F,$D7,$B5,$FE,$95,$E1,$3E,$1B,$C7,$C9
	.dc.b	$FD,$3D,$B3,$FE,$5C,$6E,$0E,$9F,$D4,$CF,$E7,$57,$C8,$6B,$E4,$7F
	.dc.b	$93,$4F,$ED,$9A,$FF,$95,$E7,$6E,$7C,$4E,$04,$78,$78,$BC,$F1,$F7
	.dc.b	$62,$C3,$D1,$5F,$C3,$9A,$FF,$0C,$6B,$FA,$48,$E1,$DB,$EA,$F0,$37
	.dc.b	$C7,$FA,$D9,$46,$35,$FE,$4F,$87,$F2,$5C,$3D,$DF,$2F,$F1,$39,$3E
	.dc.b	$17,$75,$DC,$D9,$C3,$F5,$67,$75,$FD,$86,$B6,$D9,$D4,$78,$A3,$FD
	.dc.b	$25,$F0,$3D,$F8,$8F,$7C,$FC,$1E,$AE,$7E,$06,$5F,$D3,$8A,$FE,$F3
	.dc.b	$2F,$F0,$3C,$F8,$F0,$34,$F0,$F8,$BD,$FD,$F9,$7F,$76,$2B,$FE,$09
	.dc.b	$8B,$C3,$F9,$9D,$3F,$DC,$33,$F8,$3E,$CF,$47,$C4,$C8,$C1,$FE,$AD
	.dc.b	$7C,$44,$7F,$BD,$BF,$0B,$E4,$3E,$07,$C2,$49,$FC,$74,$57,$D9,$A7
	.dc.b	$F9,$F1,$FF,$01,$CF,$B4,$F9,$7E,$37,$D5,$24,$7E,$5F,$E3,$BE,$67
	.dc.b	$91,$23,$FD,$80,$E9,$FC,$B7,$8B,$F4,$7E,$26,$47,$E7,$E2,$BE,$28
	.dc.b	$F9,$84,$7F,$69,$FA,$DB,$CC,$FD,$B4,$23,$0E,$A3,$31,$88,$FF,$6F
	.dc.b	$2D,$BF,$21,$CF,$FE,$79,$9F,$E9,$8D,$7F,$D4,$13,$B7,$FD,$ED,$A3
	.dc.b	$E2,$BC,$2D,$8F,$E1,$0D,$7F,$DD,$8D,$7F,$A1,$AE,$AF,$E0,$70,$FC
	.dc.b	$71,$E1,$61,$B1,$E0,$9D,$FD,$67,$F1,$B8,$FF,$5A,$3F,$ED,$E7,$FC
	.dc.b	$14,$1C,$3A,$F2,$78,$FC,$57,$C8,$9C,$CF,$F5,$66,$BF,$D8,$1A,$FF
	.dc.b	$6B,$FC,$53,$F7,$4F,$F5,$6F,$1B,$E3,$75,$7F,$2F,$6F,$87,$FD,$A2
	.dc.b	$B7,$5E,$3D,$3C,$04,$7F,$97,$61,$EE,$7F,$B3,$9E,$9F,$17,$E2,$FE
	.dc.b	$05,$F2,$F2,$FE,$F1,$28,$C6,$3E,$7A,$FE,$15,$F8,$C3,$EA,$69,$2F
	.dc.b	$FC,$11,$1F,$E7,$E7,$FC,$15,$A6,$DB,$3F,$C4,$E3,$10,$C7,$FA,$65
	.dc.b	$61,$F1,$E4,$E7,$FA,$B9,$4B,$FA,$73,$96,$C7,$D1,$23,$F0,$9F,$CD
	.dc.b	$EA,$EE,$FC,$2F,$E3,$9F,$D5,$F1,$F8,$2D,$AF,$9D,$3F,$CA,$7F,$16
	.dc.b	$FB,$F5,$FD,$85,$FC,$33,$78,$B5,$C5,$B7,$F9,$15,$CC,$BF,$C5,$97
	.dc.b	$C2,$3F,$C4,$3F,$5E,$FA,$AF,$D7,$B6,$BF,$B3,$35,$FB,$2D,$7F,$C4
	.dc.b	$2F,$FB,$2D,$FC,$AF,$DE,$3E,$8F,$F4,$2D,$8F,$CC,$CE,$DF,$A1,$7F
	.dc.b	$E3,$BF,$25,$8B,$E5,$BE,$07,$C3,$CB,$F4,$C7,$C3,$F8,$79,$1E,$C7
	.dc.b	$4F,$43,$FC,$5F,$EE,$BE,$47,$FA,$93,$63,$FC,$9C,$D7,$86,$3E,$2A
	.dc.b	$FA,$1F,$0D,$27,$E4,$BF,$4F,$E1,$79,$FC,$A4,$9E,$6E,$DE,$7A,$7E
	.dc.b	$87,$93,$EC,$77,$FD,$B3,$EE,$77,$5A,$F1,$FC,$9F,$98,$C3,$EB,$1C
	.dc.b	$4F,$B8,$FE,$1A,$BE,$2E,$98,$8B,$3F,$6A,$3E,$E9,$1F,$F1,$17,$CE
	.dc.b	$3E,$8F,$F8,$BF,$E3,$9F,$B2,$D3,$F6,$6D,$E7,$FF,$33,$D6,$EF,$E4
	.dc.b	$F5,$7C,$C2,$3F,$8E,$F2,$FF,$5D,$F1,$BF,$DE,$FC,$53,$49,$3F,$ED
	.dc.b	$DE,$8F,$C2,$67,$FE,$3D,$6E,$5F,$85,$6C,$FF,$AA,$BF,$64,$FC,$63
	.dc.b	$F6,$23,$F7,$13,$BF,$1F,$EF,$9F,$E3,$2A,$FF,$61,$57,$E0,$3D,$4F
	.dc.b	$F0,$75,$FF,$07,$8F,$F0,$03,$5F,$F4,$2F,$5A,$31,$1F,$97,$E7,$FE
	.dc.b	$D4,$F8,$78,$FF,$1A,$7E,$DF,$CC,$FC,$7F,$24,$9F,$C7,$BF,$BB,$D3
	.dc.b	$F8,$97,$C6,$FC,$52,$4F,$FB,$FF,$81,$1E,$6F,$8F,$FC,$C9,$27,$FD
	.dc.b	$4D,$F8,$B3,$B3,$F3,$BF,$63,$27,$F3,$AF,$F8,$C6,$7F,$75,$C7,$F3
	.dc.b	$79,$3E,$09,$2B,$E9,$3F,$E4,$8F,$C5,$FF,$C2,$B1,$FE,$D0,$6B,$FE
	.dc.b	$45,$6B,$FE,$E6,$6B,$FE,$9E,$AF,$99,$5F,$6F,$E6,$7F,$62,$78,$F9
	.dc.b	$3F,$C8,$9F,$52,$FA,$9F,$F2,$03,$70,$FE,$8F,$CC,$8C,$47,$F8,$D5
	.dc.b	$AF,$D6,$E6,$7C,$AF,$EF,$89,$1F,$EE,$0F,$C3,$9F,$4D,$FC,$0B,$E3
	.dc.b	$64,$FF,$02,$7F,$CE,$FB,$FF,$80,$E3,$F1,$9F,$23,$F8,$62,$3F,$97
	.dc.b	$7F,$CA,$DC,$1F,$9C,$8F,$E1,$7E,$77,$F0,$E3,$E8,$71,$7F,$98,$7F
	.dc.b	$65,$FF,$5E,$71,$FF,$A6,$5B,$D3,$FC,$85,$B7,$FE,$C6,$5F,$4F,$F1
	.dc.b	$BD,$CF,$8A,$49,$F8,$BC,$5F,$E1,$17,$F9,$56,$7F,$E9,$9C,$7F,$55
	.dc.b	$D5,$FC,$80,$FC,$16,$FF,$83,$D7,$F9,$1F,$CC,$4C,$3F,$17,$15,$FD
	.dc.b	$1C,$7E,$98,$8F,$D4,$C3,$F5,$6F,$F2,$86,$E1,$FE,$B5,$6E,$BF,$85
	.dc.b	$F8,$1F,$C9,$5F,$F3,$2F,$98,$98,$7E,$6A,$3F,$2D,$F3,$12,$3F,$3A
	.dc.b	$FE,$65,$C1,$FB,$F7,$F9,$D9,$8C,$7F,$D8,$9D,$9F,$EA,$8F,$98,$7C
	.dc.b	$3F,$F4,$D8,$7A,$3F,$EF,$C8,$C3,$70,$FF,$7A,$CF,$19,$CC,$0F,$F7
	.dc.b	$F3,$86,$9C,$27,$97,$62,$B9,$4E,$4E,$5A,$FF,$0C,$8F,$1C,$60,$E9
	.dc.b	$19,$AF,$04,$7F,$89,$2B,$FD,$4F,$F2,$7D,$BE,$41,$F5,$94,$C4,$4A
	.dc.b	$F4,$67,$F0,$95,$FF,$AC,$2B,$84,$6B,$3F,$EE,$58,$7F,$B3,$A3,$DA
	.dc.b	$3E,$B7,$FC,$31,$7F,$EE,$96,$C7,$F8,$F1,$AF,$A0,$E1,$EF,$7F,$BC
	.dc.b	$DF,$5F,$CD,$7F,$9F,$7E,$BB,$FD,$E4,$D8,$FF,$22,$B5,$E4,$D7,$E1
	.dc.b	$D3,$CB,$71,$78,$DB,$3E,$36,$BE,$39,$E1,$3F,$C3,$3E,$57,$E4,$F0
	.dc.b	$FC,$1A,$FF,$03,$87,$19,$FF,$BA,$76,$C4,$FE,$02,$3F,$A2,$38,$DF
	.dc.b	$E7,$46,$EB,$8D,$F6,$AE,$0F,$F1,$F8,$CD,$F1,$DF,$FC,$45,$D6,$F9
	.dc.b	$9C,$61,$F6,$23,$FB,$B5,$F9,$1D,$37,$D6,$DE,$5F,$C9,$EE,$D7,$0B
	.dc.b	$F4,$3C,$1B,$5A,$99,$7D,$7E,$56,$D7,$A3,$E2,$BE,$0F,$47,$C2,$C9
	.dc.b	$AB,$9B,$38,$FF,$8E,$E3,$FD,$44,$FE,$A6,$B8,$46,$E3,$06,$EF,$8E
	.dc.b	$1F,$96,$9E,$2F,$F8,$D1,$FF,$5A,$4F,$08,$D4,$F0,$F9,$77,$D8,$7F
	.dc.b	$94,$6B,$59,$79,$BE,$63,$53,$9C,$3A,$46,$3C,$09,$7C,$32,$75,$62
	.dc.b	$FC,$5C,$1F,$E9,$67,$DC,$D7,$D0,$DC,$BE,$2B,$69,$BF,$BA,$73,$87
	.dc.b	$78,$C4,$FD,$7D,$BE,$77,$D6,$7C,$7C,$7E,$17,$0F,$18,$C4,$FC,$C8
	.dc.b	$E9,$C5,$FD,$23,$1F,$E8,$BE,$2F,$8E,$31,$87,$F6,$98,$C7,$83,$C0
	.dc.b	$F0,$FC,$C5,$BF,$85,$F2,$5E,$4F,$93,$F2,$B6,$3E,$37,$CE,$FB,$48
	.dc.b	$FE,$B5,$F8,$9C,$CF,$AE,$FD,$06,$7F,$E7,$58,$C7,$A7,$F0,$3E,$63
	.dc.b	$ED,$56,$B8,$3F,$5F,$4F,$AF,$FF,$3D,$65,$F8,$18,$C7,$C5,$7C,$8F
	.dc.b	$4F,$F0,$55,$AC,$FF,$CF,$7F,$B3,$3E,$87,$EA,$37,$1F,$F0,$6B,$67
	.dc.b	$F8,$C7,$4A,$31,$1F,$E4,$2E,$8F,$E9,$75,$F8,$5F,$9B,$FA,$FE,$17
	.dc.b	$D8,$7E,$CF,$3F,$AB,$C2,$B5,$E7,$E1,$FA,$EF,$E9,$7A,$3F,$08,$47
	.dc.b	$F8,$A3,$F8,$9F,$AD,$7C,$97,$E1,$DB,$3F,$95,$DB,$13,$9F,$AE,$93
	.dc.b	$FA,$0A,$7C,$A6,$37,$F8,$7F,$11,$F0,$89,$5F,$09,$F0,$7C,$9F,$73
	.dc.b	$C1,$F2,$76,$7A,$72,$78,$FE,$4F,$C8,$D8,$F0,$7C,$9F,$A4,$8F,$F8
	.dc.b	$EB,$34,$7E,$13,$81,$C7,$F8,$03,$DD,$EA,$5B,$F9,$1C,$9F,$84,$3E
	.dc.b	$17,$C3,$D8,$FD,$EF,$D6,$BE,$4B,$F6,$0D,$8F,$F2,$4B,$78,$7E,$EB
	.dc.b	$5F,$F3,$EE,$8F,$93,$F9,$C7,$CF,$F9,$4B,$7F,$C3,$E9,$F5,$EE,$3F
	.dc.b	$E2,$F0,$F9,$3F,$D6,$F8,$7F,$D1,$A3,$73,$A9,$CD,$FF,$7C,$B7,$27
	.dc.b	$F0,$9B,$EA,$D8,$CC,$FF,$6B,$9F,$F0,$94,$61,$C7,$D5,$BE,$33,$C7
	.dc.b	$39,$CB,$C5,$39,$BF,$54,$E5,$FF,$78,$6A,$E8,$FC,$77,$D2,$7E,$3D
	.dc.b	$B3,$F8,$BF,$A4,$C3,$4F,$C7,$47,$E5,$7D,$FF,$CB,$7D,$25,$9F,$C9
	.dc.b	$3F,$99,$F7,$B1,$B7,$FB,$BB,$2F,$EF,$55,$F9,$7B,$FC,$F7,$D1,$C8
	.dc.b	$FC,$D7,$CF,$DB,$E1,$1F,$A3,$8F,$D2,$34,$FE,$7F,$DD,$FC,$07,$83
	.dc.b	$FA,$B6,$3F,$DE,$E7,$3E,$5F,$07,$E1,$F5,$7C,$F2,$3F,$3F,$1F,$BF
	.dc.b	$1F,$4F,$9F,$E6,$BF,$39,$B3,$E0,$EA,$FE,$79,$97,$FA,$CE,$30,$79
	.dc.b	$3E,$4B,$3F,$33,$AA,$B3,$AC,$7F,$DF,$DC,$BF,$4E,$3F,$21,$F1,$6E
	.dc.b	$3F,$BB,$C9,$F9,$08,$C4,$FD,$CE,$5F,$C4,$11,$F5,$5A,$3F,$1D,$CE
	.dc.b	$F8,$AF,$C3,$C7,$E6,$A3,$15,$CC,$FD,$6F,$4F,$F5,$7E,$EF,$F7,$CC
	.dc.b	$60,$FD,$8F,$C6,$E7,$E4,$75,$56,$75,$B3,$9F,$F4,$9F,$59,$FB,$18
	.dc.b	$FF,$2E,$B5,$FF,$37,$35,$34,$71,$7E,$C1,$FD,$E4,$F4,$8C,$71,$67
	.dc.b	$2F,$67,$3D,$C7,$F7,$5A,$FE,$15,$F0,$E7,$FB,$1C,$BF,$4D,$9F,$EB
	.dc.b	$DA,$FF,$96,$DA,$FF,$9D,$3F,$02,$77,$78,$EB,$F8,$1C,$62,$FF,$84
	.dc.b	$5F,$4D,$7F,$82,$7D,$6F,$F8,$0A,$7F,$1C,$DC,$7F,$7C,$D7,$FC,$D4
	.dc.b	$D7,$FD,$06,$FF,$8E,$67,$9F,$BF,$FC,$4A,$31,$5A,$8F,$E2,$BF,$57
	.dc.b	$FD,$C2,$FF,$CE,$8D,$7F,$D1,$AD,$7F,$D3,$CD,$4F,$F9,$66,$FF,$8E
	.dc.b	$3F,$DD,$6B,$8F,$C3,$F6,$F6,$BF,$43,$F5,$7F,$E2,$1E,$D3,$49,$5F
	.dc.b	$EF,$C8,$7C,$F7,$08,$FD,$CB,$72,$FC,$44,$7F,$AC,$7D,$3F,$EF,$5F
	.dc.b	$33,$FA,$E3,$F5,$FF,$37,$67,$F8,$0B,$87,$F8,$1F,$CE,$5B,$E6,$7F
	.dc.b	$A8,$B1,$7C,$E7,$9C,$95,$E5,$7D,$5F,$27,$FB,$C7,$F7,$AD,$8F,$E1
	.dc.b	$FD,$3E,$F2,$3E,$D7,$E1,$73,$7F,$04,$7F,$82,$B6,$3F,$8B,$F6,$7F
	.dc.b	$08,$3F,$84,$F7,$D8,$FE,$36,$DC,$BD,$3B,$7F,$86,$E3,$1C,$9F,$E7
	.dc.b	$3D,$1F,$33,$F4,$49,$87,$CF,$7E,$A7,$C6,$7F,$16,$7F,$8B,$B6,$3F
	.dc.b	$94,$FC,$AF,$F1,$AF,$98,$FE,$36,$D8,$FE,$5B,$F2,$FF,$74,$8F,$DB
	.dc.b	$7F,$01,$F2,$9F,$C8,$9F,$E4,$6D,$8F,$F1,$C7,$CE,$3E,$2F,$FC,$F0
	.dc.b	$BE,$32,$C6,$1C,$FC,$67,$F9,$2A,$4D,$25,$7F,$65,$F8,$A4,$BE,$33
	.dc.b	$A9,$FC,$D3,$8E,$D7,$FA,$13,$EA,$7F,$35,$1F,$83,$3F,$CE,$8E,$39
	.dc.b	$3F,$C9,$11,$FE,$8D,$8F,$E5,$EF,$F9,$65,$BC,$7F,$E5,$6F,$92,$FE
	.dc.b	$B7,$39,$F4,$7F,$D2,$90,$F5,$BF,$CC,$6D,$C7,$D0,$DA,$FF,$74,$69
	.dc.b	$FE,$90,$AF,$F4,$57,$C8,$7F,$82,$27,$FD,$F2,$FF,$94,$9F,$C1,$B9
	.dc.b	$CF,$FD,$91,$F1,$5F,$E1,$49,$FF,$80,$23,$E2,$7F,$0A,$89,$C7,$FE
	.dc.b	$DC,$8C,$C6,$47,$CB,$6E,$33,$1F,$EF,$0D,$31,$D3,$FE,$15,$AB,$3C
	.dc.b	$6F,$A4,$E6,$3C,$4F,$57,$91,$83,$13,$FB,$FB,$E1,$8F,$A6,$72,$1C
	.dc.b	$3E,$79,$E0,$FB,$7E,$63,$F0,$FC,$18,$84,$61,$E5,$7E,$F8,$C5,$F4
	.dc.b	$6C,$67,$93,$FA,$D5,$79,$CE,$7C,$7F,$1E,$7E,$4D,$7E,$46,$7E,$19
	.dc.b	$E0,$AE,$33,$A6,$E5,$E1,$BE,$4D,$AE,$37,$E7,$BF,$B3,$83,$6F,$83
	.dc.b	$97,$FB,$CD,$F8,$97,$FE,$E0,$6B,$F8,$6A,$FB,$93,$FE,$AC,$FA,$A7
	.dc.b	$77,$F8,$7F,$89,$18,$BF,$D6,$35,$FC,$AB,$5F,$C3,$43,$95,$EE,$F6
	.dc.b	$E2,$FC,$74,$E6,$31,$48,$C3,$FE,$BB,$BF,$C8,$E2,$BF,$E8,$18,$3C
	.dc.b	$AE,$3A,$F7,$F8,$DB,$26,$CF,$6C,$9F,$1C,$F7,$B0,$DD,$3F,$D9,$4D
	.dc.b	$7F,$96,$E1,$CD,$CF,$BD,$F1,$64,$D2,$47,$F1,$18,$FF,$01,$BD,$3E
	.dc.b	$17,$B9,$E0,$F5,$7C,$FE,$EF,$A3,$E6,$25,$7F,$13,$8A,$FF,$B4,$B1
	.dc.b	$4F,$F9,$4F,$6F,$9C,$D8,$CC,$FF,$A1,$67,$E9,$77,$7C,$4F,$97,$93
	.dc.b	$FC,$7E,$3E,$3F,$CB,$FC,$8F,$5A,$4F,$F2,$38,$AF,$86,$3B,$11,$ED
	.dc.b	$87,$E4,$5F,$E5,$5E,$27,$D2,$3E,$9F,$55,$8C,$90,$F9,$AF,$D8,$F1
	.dc.b	$7E,$CF,$7C,$9F,$E6,$91,$5F,$85,$1E,$F2,$3E,$2E,$FF,$C0,$FC,$AC
	.dc.b	$8F,$E7,$91,$5F,$95,$1F,$1C,$8F,$AB,$E3,$3F,$12,$47,$FA,$27,$CA
	.dc.b	$FE,$33,$C5,$91,$FD,$1D,$D3,$F8,$6F,$8C,$FC,$8F,$85,$23,$FA,$64
	.dc.b	$57,$ED,$47,$D7,$23,$FE,$8D,$FA,$AF,$EC,$0F,$FB,$35,$FE,$C8,$DD
	.dc.b	$3F,$4E,$A3,$6F,$F8,$68,$C4,$7F,$A5,$DA,$75,$71,$1D,$23,$FD,$95
	.dc.b	$F1,$5F,$9D,$79,$2E,$7E,$ED,$8F,$F7,$03,$BF,$07,$FD,$B8,$E6,$39
	.dc.b	$CF,$86,$31,$1F,$CF,$5A,$FF,$4B,$FA,$6F,$F7,$73,$FE,$EB,$FA,$77
	.dc.b	$CB,$7B,$9B,$1F,$EE,$F9,$DB,$EF,$8C,$BF,$EE,$36,$BF,$D2,$DA,$FF
	.dc.b	$59,$6B,$FD,$7D,$7C,$AF,$CC,$9F,$DB,$CF,$CB,$6E,$3F,$EF,$9A,$DC
	.dc.b	$F9,$FC,$1F,$37,$83,$23,$07,$AD,$FA,$53,$FC,$1A,$3F,$D9,$4D,$7D
	.dc.b	$11,$FB,$BF,$A8,$FF,$32,$C7,$F7,$3B,$FE,$40,$FF,$5D,$87,$BB,$FC
	.dc.b	$81,$B1,$BF,$1F,$FC,$C9,$1C,$A3,$FC,$8A,$E9,$BE,$F7,$13,$95,$7E
	.dc.b	$33,$F1,$BE,$18,$7B,$FD,$1F,$A4,$67,$E0,$2E,$1E,$A8,$F6,$91,$F6
	.dc.b	$E5,$F2,$FE,$42,$47,$DD,$8F,$E6,$BE,$2F,$F3,$8F,$A4,$F9,$73,$FE
	.dc.b	$5A,$6E,$3F,$02,$7F,$A5,$1B,$E2,$E1,$C7,$2F,$E0,$1F,$E0,$1B,$E3
	.dc.b	$87,$F8,$5F,$E9,$DF,$17,$FA,$87,$FD,$8F,$5C,$03,$87,$F5,$4F,$07
	.dc.b	$4E,$5F,$F0,$B3,$5E,$DF,$19,$FA,$F8,$C4,$78,$FE,$6B,$DD,$F3,$52
	.dc.b	$3C,$FA,$32,$56,$67,$E1,$7C,$BF,$C5,$31,$B7,$C3,$A7,$CB,$7C,$92
	.dc.b	$57,$48,$57,$CD,$F8,$2F,$9D,$F3,$12,$B5,$15,$FA,$1F,$35,$F4,$CC
	.dc.b	$63,$DF,$B3,$EB,$7A,$72,$BB,$45,$7E,$C7,$67,$DB,$31,$5F,$98,$87
	.dc.b	$DD,$7F,$05,$8C,$70,$BE,$2F,$EA,$1F,$C7,$27,$5C,$BF,$66,$3E,$61
	.dc.b	$2F,$FC,$8C,$7F,$95,$EB,$93,$83,$ED,$FC,$67,$F2,$6F,$9A,$7A,$74
	.dc.b	$FC,$14,$7E,$47,$E6,$25,$7C,$97,$F2,$9E,$C7,$EB,$7E,$07,$F6,$84
	.dc.b	$9F,$ED,$9E,$27,$F9,$81,$FF,$0D,$FD,$3B,$E2,$BF,$C1,$CD,$8F,$CA
	.dc.b	$FB,$A8,$C4,$7E,$7F,$E6,$37,$F1,$52,$37,$28,$FD,$37,$C4,$7E,$9F
	.dc.b	$C5,$FF,$00,$76,$9A,$4B,$FF,$00,$47,$F8,$C1,$FC,$27,$9B,$FA,$F3
	.dc.b	$F6,$1F,$09,$FB,$1E,$69,$3F,$89,$7F,$69,$E4,$7F,$8B,$3E,$97,$F9
	.dc.b	$BC,$FF,$9B,$5B,$E2,$1A,$63,$E9,$5F,$39,$FE,$34,$6D,$7E,$81,$AF
	.dc.b	$FB,$61,$7C,$D6,$FC,$99,$3B,$79,$79,$5C,$FC,$DF,$F1,$05,$FF,$1D
	.dc.b	$AF,$F4,$A6,$48,$C4,$7F,$A6,$E3,$13,$C3,$D0,$FF,$21,$7D,$37,$F6
	.dc.b	$8F,$0F,$F9,$6F,$42,$31,$C1,$C7,$25,$FA,$FC,$2F,$67,$A3,$34,$F6
	.dc.b	$47,$C7,$F8,$0F,$F1,$A7,$17,$F7,$49,$F2,$B5,$F2,$35,$FF,$30,$7D
	.dc.b	$3B,$DC,$FF,$2B,$36,$3F,$12,$D7,$E4,$78,$7E,$AF,$80,$91,$F8,$ED
	.dc.b	$BF,$DC,$6F,$CF,$F9,$6F,$A5,$F2,$DF,$20,$91,$FE,$45,$DD,$1F,$33
	.dc.b	$E5,$7E,$C7,$CB,$4A,$FB,$11,$F8,$3F,$80,$91,$FC,$6B,$64,$7D,$FF
	.dc.b	$8F,$FC,$47,$C0,$FC,$22,$57,$F9,$4E,$BF,$D6,$AF,$D8,$FA,$3F,$DF
	.dc.b	$BF,$EB,$9C,$DF,$2F,$F8,$93,$FC,$FB,$3A,$EB,$FB,$DE,$5F,$E3,$7C
	.dc.b	$A9,$87,$E3,$62,$BF,$EC,$B9,$FC,$24,$FE,$20,$91,$FC,$2B,$AF,$F0
	.dc.b	$19,$FE,$10,$91,$F9,$7F,$97,$AF,$E5,$FC,$9F,$CE,$18,$EE,$FE,$AD
	.dc.b	$AF,$FB,$65,$B3,$FA,$13,$F4,$5F,$2F,$5F,$D1,$7C,$B7,$E9,$4C,$29
	.dc.b	$FE,$C1,$CB,$FE,$91,$F9,$67,$C4,$7E,$6E,$30,$DF,$63,$FD,$BC,$E7
	.dc.b	$DA,$B8,$7F,$B8,$EB,$6D,$F7,$33,$C5,$F8,$06,$CE,$A3,$F0,$78,$62
	.dc.b	$65,$0F,$F8,$03,$87,$85,$C6,$3E,$67,$FD,$E5,$F1,$FF,$E4,$B8,$C6
	.dc.b	$08,$C5,$7E,$D5,$B5,$C3,$E6,$3C,$B1,$8A,$E1,$F2,$98,$33,$F2,$BE
	.dc.b	$EB,$7D,$37,$F2,$3A,$FE,$84,$D8,$FE,$8E,$D7,$D4,$7A,$DC,$3F,$EF
	.dc.b	$FF,$8D,$7D,$3F,$F4,$BF,$99,$46,$27,$FA,$8C,$62,$75,$39,$9E,$A7
	.dc.b	$93,$D5,$72,$79,$1B,$1E,$46,$BC,$30,$46,$23,$F8,$D7,$E5,$FF,$ED
	.dc.b	$FF,$3F,$E7,$F0,$63,$6F,$F2,$2F,$0F,$CE,$AF,$F7,$05,$6A,$DB,$27
	.dc.b	$94,$7D,$A3,$2F,$07,$11,$BB,$EB,$87,$8B,$7B,$5C,$26,$D3,$F6,$17
	.dc.b	$87,$86,$E5,$E3,$9F,$F3,$FB,$63,$37,$E0,$F8,$5F,$C7,$4E,$6F,$AC
	.dc.b	$62,$7C,$B5,$C3,$4F,$C6,$CE,$5B,$3E,$4F,$D9,$DC,$BC,$A3,$CF,$B7
	.dc.b	$FA,$77,$2F,$F1,$9F,$2D,$9F,$D9,$47,$D0,$F9,$1F,$F3,$8F,$5F,$F9
	.dc.b	$4E,$B4,$FF,$95,$6F,$FD,$E7,$5F,$E3,$C7,$2F,$D8,$F9,$AE,$2F,$33
	.dc.b	$55,$8C,$3F,$4F,$37,$7F,$AD,$34,$EB,$3E,$41,$E1,$3E,$FD,$F0,$9C
	.dc.b	$72,$FD,$76,$5F,$E1,$09,$CB,$E0,$3D,$FF,$37,$FE,$35,$C3,$F6,$4D
	.dc.b	$7F,$72,$DC,$BF,$70,$DB,$7C,$DF,$96,$3C,$0D,$BF,$97,$9F,$A4,$F2
	.dc.b	$7A,$B6,$3F,$CC,$3C,$8F,$1B,$B3,$63,$FC,$7B,$97,$E2,$E7,$27,$8F
	.dc.b	$8D,$7F,$1D,$7C,$D5,$8C,$3F,$27,$E2,$26,$1F,$57,$E6,$AE,$FF,$2F
	.dc.b	$2F,$E2,$9A,$FB,$F5,$F6,$DF,$89,$B5,$FD,$2D,$AF,$F8,$43,$7F,$EF
	.dc.b	$13,$F8,$AC,$BE,$97,$B1,$F6,$5F,$BD,$C7,$F5,$B5,$FB,$86,$38,$FD
	.dc.b	$77,$F9,$28,$FC,$17,$0F,$F0,$88,$F8,$6D,$CB,$FA,$E3,$5F,$F1,$0B
	.dc.b	$5F,$E5,$9E,$0F,$E2,$4F,$C2,$79,$76,$BF,$CA,$7E,$9F,$F8,$D9,$BF
	.dc.b	$01,$F9,$1F,$6B,$FD,$E3,$18,$3F,$4B,$EC,$7C,$52,$3F,$15,$F9,$1C
	.dc.b	$9F,$97,$39,$7C,$95,$9F,$F4,$FD,$7F,$A5,$1F,$D7,$FC,$D7,$7F,$91
	.dc.b	$FC,$F3,$FA,$1F,$DA,$3F,$65,$B7,$F8,$F1,$AF,$F8,$ED,$AF,$F1,$C6
	.dc.b	$F4,$7F,$4D,$E5,$FD,$41,$EB,$F4,$2E,$3E,$2F,$3B,$F5,$9F,$37,$EF
	.dc.b	$6C,$7F,$02,$F2,$BF,$04,$47,$F0,$4F,$E9,$7A,$1C,$6F,$5B,$F6,$DE
	.dc.b	$D7,$DE,$8C,$5F,$F0,$C8,$C7,$83,$FB,$BE,$7B,$3F,$C4,$1F,$81,$D9
	.dc.b	$FA,$FE,$0F,$CD,$CF,$E8,$9B,$C3,$FE,$1B,$6B,$FE,$60,$EE,$FE,$7B
	.dc.b	$3F,$C6,$B8,$BE,$D7,$C6,$7F,$09,$7F,$86,$7C,$F7,$DC,$BF,$D0,$37
	.dc.b	$C7,$FF,$77,$74,$3E,$33,$7F,$2F,$FC,$56,$DF,$4C,$F9,$6F,$E0,$D1
	.dc.b	$86,$C6,$74,$FF,$59,$36,$7F,$D6,$EB,$A7,$A3,$9C,$BF,$D8,$CE,$6B
	.dc.b	$FD,$E8,$E5,$FF,$3D,$65,$FD,$91,$FF,$74,$E6,$F9,$9F,$CC,$7A,$9E
	.dc.b	$E7,$F3,$5F,$1D,$8A,$FC,$8B,$73,$FE,$54,$7B,$3A,$FD,$DE,$7B,$8B
	.dc.b	$CB,$18,$7C,$FE,$D7,$B8,$98,$7B,$9F,$E4,$DA,$23,$E7,$FC,$CF,$EA
	.dc.b	$DA,$7F,$C0,$0F,$D6,$BF,$D3,$EF,$E9,$0F,$E8,$0E,$32,$FE,$29,$5C
	.dc.b	$E7,$87,$C7,$E8,$DF,$C7,$58,$F8,$7E,$37,$DE,$3F,$87,$F9,$6F,$4B
	.dc.b	$13,$FE,$D2,$8C,$36,$3C,$3F,$1D,$C1,$FC,$3F,$93,$27,$D1,$F1,$D7
	.dc.b	$4F,$E5,$1A,$FF,$BE,$67,$27,$B7,$5F,$5D,$7C,$CF,$92,$93,$FE,$49
	.dc.b	$E9,$F8,$D8,$7F,$92,$FE,$9B,$F4,$EF,$D4,$E5,$FF,$6E,$57,$E0,$B9
	.dc.b	$BF,$1A,$47,$F1,$AF,$E4,$3E,$4B,$F1,$6F,$E3,$1B,$1D,$CF,$F1,$FF
	.dc.b	$D2,$FF,$1F,$BF,$C0,$FD,$BF,$E7,$E7,$51,$FE,$5D,$9D,$65,$E6,$85
	.dc.b	$6D,$FC,$4C,$B9,$3F,$46,$DB,$FF,$20,$CF,$39,$C9,$AF,$91,$FD,$65
	.dc.b	$1B,$F2,$18,$6D,$F9,$2D,$4A,$E3,$F1,$1A,$9F,$EB,$5E,$3F,$FA,$1A
	.dc.b	$FD,$7E,$32,$31,$3F,$A0,$E0,$F3,$3C,$3C,$FF,$1F,$C3,$7A,$89,$87
	.dc.b	$AA,$7C,$31,$EE,$77,$1E,$5C,$FF,$C6,$9B,$FC,$77,$FA,$B3,$F2,$7C
	.dc.b	$DF,$E1,$D3,$FE,$70,$6D,$7F,$A2,$DF,$DD,$47,$BD,$CC,$7F,$B3,$3C
	.dc.b	$2F,$E2,$51,$A8,$FD,$CA,$EF,$CC,$FF,$17,$FD,$8F,$F8,$45,$7F,$A5
	.dc.b	$DA,$9F,$E1,$CF,$F0,$27,$26,$30,$FC,$03,$70,$FF,$43,$AF,$0F,$B7
	.dc.b	$8F,$FC,$B3,$E7,$BF,$C4,$DE,$39,$A4,$BF,$F0,$DF,$77,$DF,$7F,$8C
	.dc.b	$C7,$F0,$0E,$6E,$51,$FC,$73,$C9,$FE,$6E,$7F,$04,$E6,$B1,$FC,$25
	.dc.b	$C3,$FC,$2F,$E4,$AE,$DF,$CF,$33,$F1,$92,$7F,$C5,$79,$FC,$09,$FF
	.dc.b	$50,$F3,$B4,$FF,$38,$EF,$FB,$48,$57,$C1,$AF,$F4,$4D,$6A,$FF,$D8
	.dc.b	$55,$AC,$3F,$A6,$ED,$F0,$89,$3E,$07,$FC,$95,$5F,$BE,$C7,$B6,$3F
	.dc.b	$2E,$BF,$92,$7A,$5C,$9F,$E4,$DE,$8F,$F5,$8F,$27,$F9,$2B,$63,$24
	.dc.b	$D3,$F8,$9F,$C6,$6A,$3F,$A2,$FC,$F7,$D4,$23,$F0,$9F,$1F,$4F,$F6
	.dc.b	$47,$F9,$8B,$63,$FA,$6F,$C3,$FF,$68,$3F,$CE,$9F,$AB,$7F,$AA,$35
	.dc.b	$F4,$C6,$3E,$1F,$FC,$CD,$3F,$E8,$EA,$FF,$2C,$B7,$2F,$F6,$1F,$12
	.dc.b	$31,$1F,$49,$FF,$58,$61,$F4,$5E,$0D,$F8,$BE,$D7,$FB,$CF,$DA,$7F
	.dc.b	$C5,$57,$FE,$59,$6D,$7D,$9F,$81,$7C,$37,$FA,$2E,$73,$7F,$62,$7F
	.dc.b	$D0,$B7,$E4,$6B,$F8,$46,$BB,$F6,$3F,$D0,$51,$FE,$12,$C3,$B7,$C3
	.dc.b	$7F,$A5,$E3,$17,$D6,$3E,$3F,$C4,$C8,$FE,$27,$A7,$82,$FD,$D3,$F3
	.dc.b	$33,$FF,$0A,$FC,$ED,$BF,$EF,$AF,$5F,$86,$F0,$92,$3F,$D0,$B9,$7F
	.dc.b	$9A,$34,$F5,$7C,$1F,$8C,$F2,$7C,$15,$AF,$E5,$52,$1F,$17,$7E,$07
	.dc.b	$FC,$75,$19,$3F,$CD,$CE,$1C,$46,$FA,$FF,$DF,$BC,$FF,$D6,$AB,$55
	.dc.b	$D6,$B4,$F3,$9D,$57,$F4,$49,$CC,$F8,$E7,$2F,$84,$F2,$9F,$EF,$8A
	.dc.b	$E0,$D3,$53,$FB,$28,$CF,$29,$F5,$7C,$83,$E1,$DC,$5B,$C3,$E1,$4E
	.dc.b	$63,$10,$8C,$3D,$CC,$4F,$27,$11,$AA,$F8,$CB,$E1,$FE,$54,$61,$B9
	.dc.b	$F9,$2F,$94,$F0,$6C,$7A,$27,$77,$E6,$6B,$E2,$C3,$E6,$E2,$DC,$FD
	.dc.b	$98,$FD,$13,$75,$FD,$03,$6B,$74,$F2,$F1,$2F,$C6,$FF,$C9,$95,$FE
	.dc.b	$D2,$6B,$FB,$A6,$BF,$CA,$9A,$FF,$24,$E6,$FF,$72,$BF,$E5,$6F,$8A
	.dc.b	$EB,$E1,$7F,$BD,$A3,$F4,$7D,$3E,$0A,$F5,$47,$E9,$B2,$C1,$9E,$6F
	.dc.b	$2E,$2F,$7E,$FB,$92,$FF,$BF,$CE,$5B,$E6,$7F,$79,$D7,$D8,$D9,$FE
	.dc.b	$AD,$3E,$D8,$D7,$77,$F9,$D6,$3F,$87,$79,$1F,$12,$31,$1F,$EE,$7A
	.dc.b	$FF,$41,$3F,$BC,$E3,$F5,$75,$C8,$C1,$FC,$01,$F1,$6C,$69,$23,$FA
	.dc.b	$AF,$C0,$FC,$62,$3F,$C2,$3E,$27,$E3,$7A,$7F,$2B,$E1,$A4,$FF,$0C
	.dc.b	$8A,$FF,$2E,$E4,$FF,$4D,$3F,$D8,$27,$E1,$46,$23,$F9,$9E,$3F,$A7
	.dc.b	$7F,$8A,$F5,$FD,$33,$EA,$70,$58,$FF,$47,$FB,$BF,$68,$C3,$FA,$CE
	.dc.b	$9F,$EC,$7F,$AA,$7E,$1B,$AF,$FA,$EF,$C2,$FE,$00,$FC,$6C,$66,$7F
	.dc.b	$B4,$E6,$9F,$29,$2F,$FB,$BB,$F6,$6B,$EE,$47,$FA,$E3,$DE,$FC,$59
	.dc.b	$D3,$A2,$C7,$F7,$E8,$F7,$91,$FF,$29,$E6,$AF,$B3,$1E,$93,$F5,$2E
	.dc.b	$32,$FF,$01,$37,$2F,$6F,$3F,$E6,$BD,$E9,$7F,$EB,$27,$DB,$EA,$7B
	.dc.b	$BD,$2F,$CF,$DC,$AF,$F2,$9E,$1F,$E2,$97,$FD,$73,$1F,$50,$8F,$F8
	.dc.b	$59,$FF,$28,$2F,$41,$F0,$1F,$B9,$8C,$36,$33,$1F,$E6,$75,$F0,$79
	.dc.b	$F9,$B9,$39,$78,$57,$3F,$D9,$7F,$DF,$73,$FE,$6F,$6B,$FE,$C6,$6E
	.dc.b	$DF,$E8,$35,$EC,$71,$F6,$25,$71,$7A,$7C,$5F,$9F,$97,$F9,$94,$FF
	.dc.b	$9E,$DF,$E4,$CD,$EF,$D7,$95,$B7,$E0,$CE,$A7,$FD,$16,$D7,$FA,$33
	.dc.b	$5F,$F1,$9F,$85,$F9,$DC,$FF,$C7,$4D,$E0,$CF,$6F,$C9,$FA,$17,$A7
	.dc.b	$F2,$FE,$97,$F1,$B8,$7F,$A7,$DB,$7F,$D8,$DB,$8F,$F6,$97,$FC,$E0
	.dc.b	$FF,$A5,$57,$8F,$CC,$91,$9E,$DE,$DE,$C7,$DC,$A3,$49,$3F,$E8,$F8
	.dc.b	$7B,$1F,$7F,$EE,$1F,$0F,$4F,$44,$7A,$7C,$8F,$56,$F3,$0F,$24,$57
	.dc.b	$FD,$67,$87,$F0,$17,$FB,$73,$5F,$74,$FE,$43,$E9,$FF,$9B,$47,$E9
	.dc.b	$43,$8D,$BF,$D8,$8D,$DB,$FC,$11,$F4,$EF,$8D,$FD,$FE,$DF,$C0,$34
	.dc.b	$98,$FE,$8A,$3F,$A7,$FE,$D5,$B8,$FF,$5B,$B5,$FD,$A7,$8A,$8C,$67
	.dc.b	$FB,$4B,$FE,$B5,$F2,$5F,$EB,$DC,$2F,$C4,$FC,$CC,$7F,$DB,$73,$97
	.dc.b	$F6,$3E,$AF,$83,$E4,$E4,$78,$1E,$37,$FC,$5B,$AA,$4F,$EE,$1F,$E3
	.dc.b	$3C,$38,$8F,$2F,$8A,$F3,$3F,$8E,$99,$D7,$F8,$4B,$EE,$F8,$C9,$5F
	.dc.b	$C0,$A3,$F7,$5F,$4E,$FA,$4F,$DF,$B6,$7F,$DA,$0D,$7F,$DD,$6D,$7F
	.dc.b	$C4,$CF,$EF,$9F,$F7,$37,$DC,$3E,$83,$F8,$33,$63,$FD,$D5,$3B,$7D
	.dc.b	$3C,$FF,$CD,$3C,$E9,$1F,$C5,$A2,$BF,$4B,$97,$C5,$A7,$FA,$95,$B3
	.dc.b	$FD,$02,$BD,$7E,$4F,$B3,$CA,$98,$78,$22,$BE,$A8,$F7,$91,$FE,$6B
	.dc.b	$7E,$5C,$3B,$79,$7F,$86,$3F,$C2,$BE,$C3,$97,$F1,$38,$CC,$3F,$12
	.dc.b	$EB,$7F,$F0,$DF,$4B,$8B,$FC,$04,$DA,$FE,$86,$D7,$FD,$2A,$D7,$FC
	.dc.b	$C5,$D0,$E9,$7A,$3F,$98,$DE,$57,$E3,$A7,$2D,$E1,$FC,$EF,$C9,$48
	.dc.b	$C1,$F9,$EF,$8B,$FF,$0A,$71,$1A,$48,$FE,$AF,$C0,$AF,$D3,$FC,$8C
	.dc.b	$9F,$F3,$D7,$1F,$FB,$5E,$B5,$1F,$AA,$F3,$FF,$5A,$47,$79,$7F,$B6
	.dc.b	$27,$F9,$CC,$60,$FF,$24,$EF,$E3,$7E,$2E,$5F,$20,$8F,$93,$CE,$FD
	.dc.b	$C9,$1E,$5E,$02,$FF,$A3,$E1,$FE,$E6,$D5,$E7,$FF,$00,$25,$F0,$A7
	.dc.b	$D1,$7F,$04,$F9,$CF,$8A,$49,$FC,$2B,$FE,$85,$7F,$D3,$5C,$8F,$3D
	.dc.b	$3F,$C1,$1B,$3F,$C0,$9A,$FF,$0C,$BF,$ED,$8F,$F1,$0B,$FE,$26,$47
	.dc.b	$C9,$F1,$9F,$8A,$BF,$C4,$92,$3F,$12,$F8,$D6,$0F,$D2,$3F,$4B,$1F
	.dc.b	$E7,$4C,$FF,$D5,$5C,$5F,$A9,$F8,$4F,$E3,$35,$AD,$FF,$C9,$B7,$FA
	.dc.b	$FF,$07,$F9,$29,$FA,$07,$F9,$3F,$AD,$F8,$82,$3F,$A8,$7F,$17,$F4
	.dc.b	$5F,$CB,$09,$1F,$53,$F5,$3F,$E9,$2F,$D6,$E2,$FE,$B2,$FF,$2E,$7F
	.dc.b	$AE,$E1,$F5,$BD,$2F,$F4,$1C,$6B,$47,$27,$FB,$26,$BF,$AA,$E7,$FA
	.dc.b	$C2,$61,$FA,$C7,$F3,$BC,$3F,$CF,$89,$1E,$3F,$98,$FF,$47,$E7,$FE
	.dc.b	$F6,$8F,$E8,$3F,$15,$FD,$10,$CF,$C6,$D8,$FE,$0B,$1F,$A8,$23,$F8
	.dc.b	$3E,$3F,$EC,$FF,$41,$D3,$E7,$64,$FF,$42,$EC,$4E,$23,$F8,$E1,$FD
	.dc.b	$48,$C6,$FE,$E7,$FD,$57,$97,$F8,$61,$FE,$A6,$FF,$A5,$9B,$F1,$3E
	.dc.b	$AF,$77,$D6,$FF,$AA,$BE,$79,$F2,$74,$CD,$7D,$1F,$53,$FD,$D7,$F0
	.dc.b	$FF,$2F,$9B,$FB,$7F,$C7,$A3,$13,$AA,$FF,$5E,$D3,$E4,$71,$CF,$FA
	.dc.b	$CA,$BC,$0E,$3A,$6A,$FA,$E3,$F6,$67,$58,$FF,$BD,$E3,$15,$C2,$30
	.dc.b	$78,$2F,$BE,$FF,$E9,$98,$CE,$28,$FF,$1B,$67,$E9,$9F,$93,$F0,$5C
	.dc.b	$3E,$2F,$8B,$23,$8D,$70,$CB,$F1,$8F,$58,$5F,$35,$F4,$3F,$DB,$F1
	.dc.b	$FD,$83,$63,$F8,$23,$51,$0E,$3E,$B7,$FB,$99,$FE,$15,$F2,$3F,$EE
	.dc.b	$9F,$A2,$FD,$D3,$63,$3C,$89,$CB,$63,$BF,$B7,$F2,$9E,$C7,$99,$C3
	.dc.b	$E1,$C3,$FD,$3D,$C1,$E8,$9C,$CF,$FB,$E3,$95,$F2,$BF,$91,$D7,$E5
	.dc.b	$6A,$9F,$CC,$F8,$1F,$41,$DB,$7A,$B8,$70,$FE,$D6,$73,$8F,$CA,$8C
	.dc.b	$4E,$A3,$C1,$8D,$C3,$B4,$DE,$5F,$93,$8C,$BF,$C4,$1F,$A5,$7E,$65
	.dc.b	$7F,$95,$A3,$06,$A3,$83,$FD,$30,$E8,$77,$FA,$1F,$D2,$C7,$E2,$AB
	.dc.b	$B6,$7F,$02,$73,$5E,$93,$95,$7E,$42,$33,$3F,$1A,$B8,$E5,$F0,$7E
	.dc.b	$5B,$F7,$5D,$9F,$BB,$CB,$E5,$7C,$2E,$32,$D1,$FD,$63,$6F,$F3,$6B
	.dc.b	$7C,$1D,$6B,$F5,$71,$B8,$7D,$39,$B6,$9D,$F8,$0F,$A2,$76,$6B,$E2
	.dc.b	$73,$18,$CF,$C1,$1C,$BC,$3C,$D5,$9F,$F2,$CD,$6F,$4F,$F3,$1E,$4F
	.dc.b	$95,$FE,$2D,$1F,$4A,$31,$5F,$ED,$E6,$C7,$E5,$35,$FC,$55,$70,$8F
	.dc.b	$C3,$78,$9E,$A8,$F5,$B6,$FE,$27,$CA,$FB,$8F,$83,$F4,$1F,$ED,$38
	.dc.b	$C9,$FE,$00,$E6,$57,$5C,$18,$A7,$35,$FE,$F2,$6C,$F9,$5A,$FA,$FB
	.dc.b	$3E,$7C,$31,$F9,$F1,$87,$E8,$F8,$32,$FE,$D7,$8D,$70,$F8,$FE,$1F
	.dc.b	$8F,$2F,$D5,$37,$63,$4C,$7C,$C7,$DC,$F9,$6F,$BA,$DC,$7F,$0D,$D5
	.dc.b	$F7,$B8,$FC,$91,$83,$FD,$1E,$E3,$2F,$F4,$CE,$EF,$98,$F1,$36,$BE
	.dc.b	$14,$60,$F2,$7C,$B7,$E3,$0F,$C6,$B6,$3F,$C5,$BE,$DB,$E8,$BF,$26
	.dc.b	$D8,$FC,$D6,$F8,$9C,$FC,$EF,$C1,$A3,$49,$5F,$E2,$BA,$FF,$27,$7A
	.dc.b	$1D,$0E,$BF,$3D,$67,$C3,$F3,$1E,$0F,$94,$F0,$B6,$3F,$C8,$3F,$A6
	.dc.b	$3E,$67,$E8,$36,$3B,$37,$83,$F2,$8D,$7F,$85,$E9,$FE,$49,$9F,$7F
	.dc.b	$BB,$EE,$BF,$7B,$F0,$3E,$77,$E0,$FC,$D5,$AF,$8A,$FC,$4F,$92,$F8
	.dc.b	$04,$8F,$76,$1F,$EF,$06,$F9,$5F,$B8,$F8,$27,$BD,$F3,$23,$0F,$CE
	.dc.b	$F6,$35,$EE,$DB,$FF,$66,$DF,$E8,$F6,$FF,$6B,$EF,$FE,$C3,$9F,$7F
	.dc.b	$AF,$F0,$88,$F8,$BE,$5F,$EC,$9F,$6B,$B2,$C7,$DE,$F9,$3F,$BA,$7F
	.dc.b	$12,$C1,$F4,$7F,$27,$D2,$FE,$39,$C0,$AD,$4F,$C9,$AF,$F3,$33,$5F
	.dc.b	$EA,$AD,$7F,$B2,$35,$FD,$BC,$FA,$5F,$F8,$01,$1C,$EB,$BE,$5D,$B9
	.dc.b	$3F,$23,$F1,$7F,$92,$F5,$EE,$7F,$96,$7F,$2B,$F2,$38,$CB,$FD,$F7
	.dc.b	$9A,$31,$97,$F1,$AC,$DC,$FF,$C2,$6F,$DE,$D8,$CE,$3F,$E6,$F6,$CF
	.dc.b	$BB,$9B,$E0,$D7,$C2,$6D,$7F,$A5,$3E,$7D,$F2,$DF,$A8,$8F,$E7,$71
	.dc.b	$87,$F5,$A6,$22,$3C,$7E,$3E,$E0,$CB,$FA,$65,$FF,$B3,$6B,$47,$7F
	.dc.b	$92,$EB,$7F,$D1,$3D,$17,$B1,$4F,$0F,$C9,$DB,$F3,$38,$7D,$1F,$25
	.dc.b	$73,$F0,$47,$F9,$C7,$27,$C9,$FA,$8F,$5F,$15,$9F,$80,$E1,$F8,$5F
	.dc.b	$27,$6F,$99,$FE,$26,$FA,$2F,$D0,$CF,$FB,$0E,$11,$87,$F0,$7C,$DF
	.dc.b	$93,$8C,$70,$7D,$9E,$5F,$AC,$4B,$FA,$CF,$DB,$FC,$0B,$F8,$B6,$1F
	.dc.b	$8C,$8F,$CE,$46,$2B,$95,$78,$7E,$8F,$E6,$5F,$FA,$FD,$AF,$F8,$B9
	.dc.b	$AF,$F8,$F5,$AF,$F9,$01,$6D,$FA,$97,$9D,$7F,$68,$73,$5D,$F8,$B5
	.dc.b	$F2,$B2,$3C,$5F,$2B,$71,$F8,$E2,$D1,$D4,$FF,$02,$FC,$EF,$FB,$EE
	.dc.b	$B8,$9F,$2F,$5F,$0B,$FE,$53,$6D,$F9,$FE,$1F,$E6,$65,$E7,$8F,$D1
	.dc.b	$63,$FE,$2A,$7D,$C6,$0F,$D3,$7D,$A7,$ED,$F4,$F6,$47,$F0,$78,$C6
	.dc.b	$3C,$0F,$B9,$E9,$7F,$42,$6F,$A3,$FC,$36,$1E,$7D,$8C,$32,$7F,$09
	.dc.b	$F8,$97,$A3,$0F,$EB,$93,$FC,$57,$E9,$7F,$83,$57,$F9,$BD,$B3,$FE
	.dc.b	$72,$5E,$76,$FC,$4E,$0C,$AF,$F3,$BE,$9F,$E3,$7E,$17,$C7,$E7,$1F
	.dc.b	$CB,$FA,$8E,$DF,$9D,$1F,$E1,$D6,$BF,$E5,$46,$BF,$A7,$F9,$9F,$F0
	.dc.b	$14,$39,$FF,$80,$C7,$F0,$26,$E5,$FC,$1F,$9F,$F8,$29,$FC,$1B,$1B
	.dc.b	$1F,$C2,$DC,$3F,$C3,$7C,$DB,$8F,$9B,$17,$CB,$F8,$09,$5F,$DB,$E3
	.dc.b	$2D,$F9,$8F,$E2,$73,$FC,$51,$B1,$FC,$67,$D2,$F4,$91,$E4,$FB,$3E
	.dc.b	$57,$F8,$DB,$FC,$71,$B1,$FC,$87,$E5,$7F,$8F,$FA,$7F,$C8,$1B,$19
	.dc.b	$26,$5E,$9F,$96,$D4,$7F,$29,$EE,$F9,$24,$7D,$4F,$CC,$9F,$E5,$6F
	.dc.b	$F2,$C6,$C7,$DC,$3F,$D6,$7C,$4F,$A4,$FF,$4D,$CF,$DC,$6F,$2F,$F9
	.dc.b	$DD,$FF,$40,$B7,$E4,$7F,$CE,$31,$FE,$9C,$9F,$F4,$8B,$5E,$AD,$7F
	.dc.b	$CA,$2D,$AF,$07,$C8,$72,$F9,$2E,$6D,$C7,$CD,$7F,$E5,$8F,$8A,$7C
	.dc.b	$77,$FA,$01,$B3,$AC,$10,$F8,$FD,$56,$FE,$3B,$FC,$A1,$3F,$E7,$2C
	.dc.b	$3F,$D4,$ED,$7E,$1B,$5F,$3F,$C8,$F7,$F8,$FF,$7B,$72,$F0,$6B,$DD
	.dc.b	$B9,$7C,$4F,$4B,$E2,$BF,$E9,$6F,$9A,$7C,$1F,$C8,$D7,$0D,$BF,$56
	.dc.b	$8F,$8E,$F3,$92,$3F,$C8,$91,$96,$FC,$5F,$C3,$3E,$27,$C6,$58,$FE
	.dc.b	$91,$BF,$FB,$C3,$E4,$FF,$DE,$3F,$17,$FE,$2C,$8F,$F7,$97,$42,$B3
	.dc.b	$DB,$FE,$E8,$9F,$F7,$F8,$AE,$0F,$FB,$EA,$B8,$39,$9E,$17,$CA,$3A
	.dc.b	$DD,$7F,$E0,$04,$55,$5C,$AF,$AC,$79,$CF,$F4,$BB,$D1,$A7,$FB,$F2
	.dc.b	$7F,$33,$19,$E5,$F3,$48,$7C,$B7,$CB,$EA,$DC,$66,$30,$F4,$37,$B7
	.dc.b	$F8,$2D,$C4,$6A,$BA,$C6,$AB,$D9,$2F,$91,$F0,$4F,$C3,$3B,$38,$C3
	.dc.b	$E4,$CF,$81,$B8,$77,$73,$3B,$D3,$9D,$75,$BE,$F7,$B9,$FC,$BF,$9F
	.dc.b	$E0,$C7,$FD,$D9,$C1,$FE,$9D,$BF,$C5,$CE,$76,$FC,$EF,$C7,$D9,$FF
	.dc.b	$61,$FD,$2F,$FA,$65,$FF,$48,$35,$FF,$6B,$35,$F6,$EC,$8C,$47,$F0
	.dc.b	$AE,$5F,$F4,$73,$FA,$4E,$BE,$D5,$FC,$35,$B3,$A9,$FE,$1D,$F2,$7A
	.dc.b	$EA,$FF,$66,$47,$EB,$23,$1D,$99,$D7,$9E,$4F,$8F,$A6,$6F,$F8,$14
	.dc.b	$3E,$3F,$F3,$51,$EE,$6E,$9E,$96,$CE,$E7,$E0,$FC,$7F,$62,$3D,$3A
	.dc.b	$FC,$DC,$9E,$7E,$69,$3F,$C0,$A2,$BF,$EF,$DD,$BE,$C3,$67,$C3,$18
	.dc.b	$8F,$E4,$FD,$5F,$04,$8F,$F0,$CF,$92,$F8,$9D,$9F,$17,$92,$4F,$F1
	.dc.b	$08,$AF,$78,$EA,$47,$FD,$95,$7F,$F0,$01,$FE,$2F,$B7,$CF,$3E,$87
	.dc.b	$35,$8F,$E3,$7D,$5F,$4C,$C3,$FD,$FB,$E7,$FF,$09,$D3,$FC,$02,$D1
	.dc.b	$FE,$DA,$9F,$E6,$4D,$78,$7C,$97,$E0,$A3,$11,$F8,$A6,$BF,$D4,$9A
	.dc.b	$A7,$AF,$B1,$F6,$27,$FC,$47,$3D,$B9,$7A,$1D,$7D,$7B,$1F,$EE,$99
	.dc.b	$FC,$33,$67,$91,$CF,$8F,$C8,$48,$F4,$F3,$79,$F9,$64,$7B,$1D,$3F
	.dc.b	$0F,$D6,$F6,$7A,$D2,$3E,$0C,$57,$F8,$19,$EB,$C7,$FC,$5E,$F9,$67
	.dc.b	$D8,$7E,$59,$79,$8F,$84,$FD,$B4,$61,$B1,$98,$FE,$C1,$F9,$73,$E1
	.dc.b	$F3,$F0,$D2,$7F,$D1,$15,$FE,$94,$6D,$F4,$FA,$5F,$F3,$3C,$7F,$A3
	.dc.b	$5A,$FF,$9B,$1B,$C9,$FC,$FA,$3F,$CA,$7E,$57,$1F,$5A,$47,$17,$A7
	.dc.b	$B1,$E8,$FA,$5F,$F5,$6C,$FF,$A6,$7E,$3B,$FA,$E3,$7C,$2D,$79,$FE
	.dc.b	$E9,$CC,$FF,$51,$6B,$F2,$9A,$FF,$2E,$6B,$FC,$FE,$BF,$C0,$0F,$FA
	.dc.b	$9D,$AF,$FA,$19,$B9,$7A,$E7,$F6,$AF,$B7,$E9,$7F,$1B,$5F,$E8,$C6
	.dc.b	$C6,$7A,$7F,$8D,$F1,$7C,$BA,$D7,$4F,$D6,$BF,$C7,$C6,$0F,$F5,$C3
	.dc.b	$CB,$CB,$FE,$8A,$FF,$BB,$23,$3F,$1B,$C4,$E5,$F1,$D6,$3F,$8C,$CA
	.dc.b	$7F,$0C,$D8,$CC,$7E,$27,$E8,$FF,$CB,$B1,$E0,$F9,$4F,$21,$F8,$A8
	.dc.b	$CC,$7B,$63,$99,$1E,$EF,$E4,$FE,$93,$FD,$EB,$1F,$ED,$16,$BF,$EB
	.dc.b	$06,$EB,$F2,$7E,$53,$F7,$D8,$4A,$FE,$06,$FF,$01,$CB,$F8,$17,$CC
	.dc.b	$49,$FE,$0E,$ED,$FE,$1D,$F2,$DF,$C1,$C8,$FF,$0C,$F9,$8F,$EB,$1F
	.dc.b	$38,$FA,$3F,$85,$3F,$EE,$88,$C6,$3A,$9C,$E7,$FC,$5F,$C1,$63,$FC
	.dc.b	$4F,$49,$8F,$F1,$87,$F8,$B6,$DF,$C5,$F9,$64,$FF,$1C,$76,$FF,$23
	.dc.b	$7F,$8E,$7C,$CC,$8D,$1F,$EF,$7F,$A4,$7C,$87,$EE,$9F,$F7,$0C,$67
	.dc.b	$E4,$B1,$39,$7F,$81,$4E,$A7,$FC,$64,$D7,$F8,$5B,$5F,$E5,$1F,$2B
	.dc.b	$FC,$B7,$9E,$47,$F3,$28,$AF,$F3,$9F,$95,$FE,$64,$63,$7F,$E7,$5E
	.dc.b	$47,$DA,$BF,$F0,$CE,$1F,$9D,$9F,$E0,$0D,$8F,$F4,$3C,$EE,$FF,$C5
	.dc.b	$D3,$FA,$37,$CB,$E3,$7A,$88,$FF,$92,$FC,$D5,$7F,$17,$8F,$F3,$C3
	.dc.b	$5F,$F6,$5B,$70,$F0,$FC,$B7,$F8,$2A,$3E,$24,$7A,$34,$C1,$9D,$7F
	.dc.b	$6B,$F0,$FF,$DF,$BB,$CD,$26,$3E,$3C,$3F,$95,$3F,$89,$E6,$70,$7E
	.dc.b	$24,$F0,$79,$1E,$1D,$BF,$19,$C1,$2F,$D3,$1E,$5D,$BF,$C1,$B8,$9A
	.dc.b	$49,$F0,$7C,$7F,$AF,$E5,$E4,$7C,$28,$F7,$78,$F2,$3E,$2B,$F0,$7E
	.dc.b	$43,$FC,$57,$F4,$FF,$C7,$27,$FC,$66,$D7,$FD,$76,$D7,$FB,$53,$79
	.dc.b	$3E,$8F,$C8,$FC,$BF,$2E,$57,$D2,$8F,$9D,$F1,$F2,$3E,$AB,$F4,$7E
	.dc.b	$43,$E9,$7C,$37,$B8,$93,$F2,$DF,$83,$E8,$FD,$72,$47,$DE,$F9,$0F
	.dc.b	$B5,$F2,$12,$3F,$02,$FD,$DE,$8F,$6F,$C0,$FF,$8D,$B6,$34,$95,$F6
	.dc.b	$7E,$2F,$F8,$1F,$D2,$7E,$13,$A3,$FD,$01,$E4,$A3,$15,$FE,$0B,$E8
	.dc.b	$FF,$47,$BF,$EC,$7F,$A3,$7D,$27,$E3,$62,$19,$7F,$CB,$30,$F9,$3F
	.dc.b	$F2,$4B,$67,$F3,$8D,$7F,$BF,$F3,$23,$11,$FE,$68,$5F,$25,$D0,$91
	.dc.b	$E3,$F9,$3E,$FF,$27,$FE,$68,$E8,$FE,$91,$3F,$E2,$36,$E1,$D6,$75
	.dc.b	$EB,$7F,$97,$27,$FA,$8D,$7F,$9D,$9A,$FF,$BF,$9A,$FF,$8B,$71,$FE
	.dc.b	$EA,$BC,$7F,$F1,$E4,$61,$BE,$CF,$F9,$EB,$D3,$74,$7C,$4F,$6F,$E2
	.dc.b	$FC,$9C,$C7,$E6,$45,$7E,$1E,$3F,$EF,$87,$E3,$CF,$CA,$3E,$5F,$C8
	.dc.b	$FF,$3C,$C3,$E5,$65,$FE,$CC,$E5,$46,$39,$BF,$D7,$9C,$1F,$36,$F8
	.dc.b	$71,$FD,$6F,$8F,$99,$FF,$B0,$72,$E5,$8F,$18,$F8,$1F,$21,$F7,$3D
	.dc.b	$F9,$1F,$83,$8A,$FC,$3C,$FB,$EC,$F9,$0F,$C1,$9A,$C7,$FC,$C5,$C6
	.dc.b	$F6,$BF,$DB,$BF,$19,$F3,$FD,$9B,$97,$D9,$6F,$0A,$F8,$4F,$FB,$D6
	.dc.b	$B8,$34,$E0,$59,$E1,$5C,$AB,$8C,$BE,$4D,$7C,$A3,$FC,$71,$7C,$1F
	.dc.b	$F7,$26,$9F,$85,$C3,$FD,$AE,$D8,$C9,$F0,$3E,$3F,$CF,$AF,$FA,$6B
	.dc.b	$1E,$B3,$F8,$9F,$90,$CD,$48,$F3,$56,$F2,$F9,$35,$99,$F3,$47,$FB
	.dc.b	$9B,$E8,$DE,$D7,$FA,$56,$7F,$CC,$F1,$8D,$3F,$D7,$2D,$8F,$F3,$58
	.dc.b	$70,$FF,$9B,$BE,$53,$EB,$7D,$27,$FB,$01,$B1,$C1,$AE,$F8,$93,$DB
	.dc.b	$DD,$FE,$0B,$3E,$26,$BE,$76,$A3,$C4,$E1,$FE,$9F,$F3,$EE,$0F,$97
	.dc.b	$1B,$F2,$93,$98,$FF,$36,$1C,$27,$6D,$33,$3C,$23,$F0,$91,$9E,$1F
	.dc.b	$A6,$FF,$5E,$F8,$AE,$91,$8A,$E1,$3F,$40,$F1,$7B,$1A,$E0,$F5,$47
	.dc.b	$F6,$6F,$77,$0E,$B1,$EF,$69,$CC,$FF,$6C,$BF,$4B,$B3,$EA,$CF,$FA
	.dc.b	$66,$BF,$DE,$2D,$8C,$FB,$3F,$07,$A7,$FD,$19,$39,$6C,$66,$BC,$B3
	.dc.b	$97,$F8,$0F,$E4,$6F,$57,$58,$7F,$B1,$E3,$13,$FC,$04,$FE,$0F,$F8
	.dc.b	$C3,$E2,$B7,$CC,$AD,$DF,$EF,$CF,$F6,$85,$6E,$F9,$47,$5C,$38,$CE
	.dc.b	$63,$85,$F3,$9E,$B7,$FB,$DB,$E3,$3D,$FB,$3C,$FE,$23,$E9,$1F,$1D
	.dc.b	$C2,$73,$18,$BE,$38,$73,$CB,$FC,$BD,$39,$7C,$31,$9D,$FF,$19,$CE
	.dc.b	$FA,$EF,$F0,$96,$1F,$A8,$8C,$7B,$FF,$61,$B1,$99,$DE,$1F,$AE,$FC
	.dc.b	$67,$FD,$E5,$3F,$B0,$84,$61,$D6,$DF,$82,$72,$FE,$10,$FE,$C5,$EB
	.dc.b	$7D,$0F,$07,$FB,$DD,$FB,$3D,$BC,$4F,$DA,$CB,$D1,$EE,$F9,$73,$34
	.dc.b	$97,$EB,$D5,$F0,$1F,$82,$D1,$F9,$13,$E5,$BA,$56,$AF,$ED,$D6,$A7
	.dc.b	$C9,$F3,$52,$7E,$FE,$6F,$33,$CF,$7F,$7B,$75,$F9,$E7,$F9,$7D,$FC
	.dc.b	$D6,$D7,$FC,$00,$BE,$5D,$FF,$3E,$73,$3F,$B0,$E1,$FC,$97,$C0,$7C
	.dc.b	$5A,$FC,$4B,$70,$FF,$0B,$7C,$CF,$C9,$3F,$13,$F1,$6F,$4F,$13,$E8
	.dc.b	$CB,$EA,$65,$FB,$CC,$FD,$59,$F8,$49,$5F,$83,$8F,$5B,$FE,$23,$F9
	.dc.b	$9F,$A5,$18,$AF,$81,$1F,$98,$FC,$1D,$F0,$9F,$C0,$3B,$D1,$8C,$3F
	.dc.b	$9A,$C6,$27,$53,$9B,$F8,$9F,$94,$7C,$27,$F9,$56,$28,$C4,$7F,$34
	.dc.b	$6C,$67,$83,$C1,$97,$F2,$DC,$BB,$B6,$FF,$8D,$8C,$91,$98,$FE,$3F
	.dc.b	$18,$9F,$CB,$76,$F9,$79,$7F,$C4,$3A,$FE,$EB,$5F,$CC,$3F,$95,$EB
	.dc.b	$FC,$65,$FE,$AD,$B8,$FD,$F6,$BF,$45,$AF,$F1,$8F,$85,$FE,$23,$F9
	.dc.b	$2F,$0C,$7F,$A1,$61,$FC,$4E,$73,$D1,$FC,$2B,$E1,$77,$F3,$97,$3F
	.dc.b	$E1,$78,$A3,$9F,$8F,$FA,$F8,$FE,$29,$9F,$E3,$5A,$FF,$8A,$1A,$FC
	.dc.b	$5F,$0F,$FD,$3F,$7D,$3E,$03,$F8,$8F,$C5,$F9,$9F,$E3,$75,$E8,$F9
	.dc.b	$BB,$5E,$27,$F9,$5F,$E5,$1F,$D3,$A7,$F3,$8D,$7F,$C6,$ED,$5B,$A3
	.dc.b	$1C,$8F,$AB,$7A,$5F,$1B,$E7,$BF,$93,$C6,$32,$F8,$31,$F2,$BE,$7A
	.dc.b	$47,$EF,$3E,$25,$E0,$E3,$4F,$E7,$83,$EA,$FF,$94,$3F,$86,$6B,$FD
	.dc.b	$89,$AF,$F7,$D6,$BC,$7E,$6B,$F8,$F4,$1A,$4B,$FF,$6F,$7C,$37,$FA
	.dc.b	$31,$FA,$FF,$1B,$FD,$42,$31,$7F,$D2,$39,$7F,$DC,$75,$FD,$1B,$F1
	.dc.b	$BF,$E7,$4F,$F9,$7A,$B7,$D9,$F8,$69,$CD,$FC,$A9,$CD,$6F,$E2,$3F
	.dc.b	$1B,$EA,$77,$6C,$7F,$52,$9F,$F4,$D3,$63,$FB,$1D,$BC,$AE,$13,$F2
	.dc.b	$FE,$67,$F9,$67,$01,$FE,$5D,$BB,$E1,$73,$18,$7E,$29,$2B,$FA,$44
	.dc.b	$ED,$BC,$3B,$7F,$99,$C7,$D8,$6C,$F6,$AF,$EE,$FF,$36,$F0,$31,$5F
	.dc.b	$DB,$23,$0F,$2F,$84,$FA,$A4,$7C,$2F,$F5,$3D,$9F,$1F,$FB,$2E,$5F
	.dc.b	$A8,$DA,$F8,$DA,$FD,$73,$E2,$F8,$96,$3F,$1D,$F0,$5F,$10,$FE,$A1
	.dc.b	$F2,$5F,$16,$32,$7F,$4E,$FA,$9F,$CC,$73,$7F,$96,$1A,$FF,$16,$6B
	.dc.b	$FE,$32,$6B,$FE,$4D,$5F,$33,$FD,$B5,$B3,$FD,$CD,$B3,$D6,$7F,$B8
	.dc.b	$7E,$17,$FE,$93,$AF,$F0,$8F,$83,$FD,$F3,$AB,$F8,$7C,$E6,$FF,$C3
	.dc.b	$33,$9A,$FF,$56,$CE,$6B,$7F,$19,$FE,$28,$FC,$6F,$FD,$67,$5F,$E1
	.dc.b	$CA,$D3,$E9,$E9,$FE,$E5,$F5,$9F,$03,$0F,$F2,$F6,$4D,$3F,$1D,$7F
	.dc.b	$DC,$AB,$4F,$C1,$F9,$3F,$71,$1F,$73,$FD,$3F,$E8,$7F,$D2,$B8,$7F
	.dc.b	$B1,$5B,$3F,$E5,$AA,$FF,$23,$61,$FD,$06,$72,$FF,$9D,$FF,$01,$73
	.dc.b	$7F,$85,$DB,$7F,$E1,$D6,$CF,$FA,$39,$AF,$F4,$C7,$FB,$84,$6F,$D9
	.dc.b	$E4,$47,$FC,$4D,$ED,$7E,$6E,$BE,$6F,$BF,$FE,$15,$8C,$67,$FD,$03
	.dc.b	$8B,$FC,$3B,$7F,$E1,$49,$FC,$55,$FF,$A7,$9A,$FF,$5C,$6B,$FE,$78
	.dc.b	$6B,$FE,$93,$E6,$7A,$7F,$CD,$3E,$17,$FC,$49,$18,$BF,$C1,$C6,$3E
	.dc.b	$6B,$FC,$71,$26,$92,$7F,$D5,$FE,$73,$C8,$78,$1F,$D3,$C9,$3A,$8C
	.dc.b	$BF,$E0,$2F,$33,$DB,$9F,$F7,$4F,$3B,$F9,$F7,$07,$F8,$26,$73,$3F
	.dc.b	$DD,$1A,$FF,$21,$F9,$8F,$E5,$07,$F4,$2D,$6E,$1F,$E7,$4C,$3F,$A4
	.dc.b	$46,$2B,$FC,$F5,$18,$9F,$E4,$15,$FE,$80,$FD,$AF,$FC,$D5,$5F,$E3
	.dc.b	$16,$BF,$EB,$BD,$11,$88,$FE,$AE,$DB,$FE,$45,$F3,$1F,$E2,$1F,$1C
	.dc.b	$D2,$63,$F0,$FE,$5F,$FB,$A1,$1C,$1F,$17,$E2,$BE,$11,$1C,$1F,$D2
	.dc.b	$BE,$67,$FA,$41,$FD,$8B,$8A,$C7,$C1,$DF,$EC,$3F,$CF,$67,$4D,$3F
	.dc.b	$D2,$3F,$AA,$7F,$91,$35,$F7,$35,$FE,$E6,$DD,$8D,$F8,$1F,$CF,$CF
	.dc.b	$F8,$B2,$7F,$DC,$FD,$5F,$E7,$D9,$CC,$FF,$96,$DA,$FF,$B9,$77,$7E
	.dc.b	$25,$FD,$42,$3E,$AB,$7A,$FF,$DC,$52,$FC,$0F,$F9,$AE,$7F,$76,$B4
	.dc.b	$F3,$8E,$57,$FE,$3A,$6B,$9F,$4F,$FC,$A9,$18,$8F,$F2,$12,$F2,$F2
	.dc.b	$4A,$FF,$2B,$C6,$5F,$F4,$F7,$80,$F3,$BF,$D1,$8D,$8F,$F4,$4D,$BC
	.dc.b	$4C,$D7,$FA,$46,$3F,$CB,$C7,$F9,$48,$A7,$F9,$53,$E0,$3F,$D4,$CE
	.dc.b	$01,$FC,$53,$D1,$E5,$F2,$1E,$06,$E1,$FE,$CA,$FC,$97,$C0,$FF,$A6
	.dc.b	$B2,$FE,$D9,$7F,$ED,$29,$CB,$FE,$20,$8C,$F0,$7F,$93,$BC,$0F,$E6
	.dc.b	$F3,$FE,$32,$6F,$0B,$F0,$3F,$74,$FF,$85,$7C,$EF,$67,$A9,$E8,$26
	.dc.b	$7F,$EC,$39,$FE,$E2,$5F,$96,$F5,$FE,$C5,$FE,$0B,$8F,$69,$A7,$0E
	.dc.b	$23,$7E,$7F,$F9,$E2,$D7,$8A,$FF,$3D,$D7,$C6,$F8,$7E,$31,$E0,$6A
	.dc.b	$8D,$5F,$3A,$B3,$C8,$FF,$41,$C6,$5A,$66,$BA,$DF,$8A,$32,$79,$CF
	.dc.b	$F2,$CC,$7D,$DB,$FE,$10,$D8,$C9,$DF,$AF,$FD,$22,$FD,$EF,$9A,$FF
	.dc.b	$66,$6F,$C2,$31,$F1,$88,$C4,$27,$6F,$43,$13,$B7,$11,$AB,$F9,$AB
	.dc.b	$E7,$3E,$8C,$61,$B9,$FD,$49,$E0,$DB,$E9,$5B,$BF,$92,$78,$EF,$8B
	.dc.b	$E8,$C3,$38,$79,$B5,$E1,$26,$9E,$4C,$3D,$0E,$A3,$FD,$6F,$3E,$77
	.dc.b	$F5,$07,$72,$B9,$7F,$89,$70,$FF,$94,$5F,$F5,$1F,$27,$E3,$B0,$FC
	.dc.b	$F3,$77,$FE,$25,$C2,$F6,$31,$7F,$E9,$2E,$5F,$F6,$DB,$FB,$2C,$3F
	.dc.b	$C2,$FF,$8F,$7F,$B2,$9B,$1C,$38,$55,$C2,$3F,$D1,$B5,$BC,$7E,$67
	.dc.b	$CA,$76,$9F,$B4,$D9,$D4,$7E,$0F,$E4,$3F,$55,$B1,$A4,$8F,$E2,$7E
	.dc.b	$EF,$3E,$9F,$F4,$84,$7F,$01,$8E,$D4,$F8,$EF,$6F,$53,$F0,$67,$D4
	.dc.b	$FC,$C8,$C3,$63,$38,$7F,$BE,$1A,$FF,$77,$7F,$C7,$2F,$FB,$A1,$AF
	.dc.b	$F3,$D6,$BF,$EE,$F6,$CF,$F4,$0F,$A1,$F1,$F8,$A6,$9E,$6F,$C2,$5F
	.dc.b	$3F,$F3,$7C,$8F,$56,$3F,$E1,$06,$BF,$D4,$5A,$FF,$4F,$CF,$F0,$FC
	.dc.b	$8F,$D6,$1F,$17,$86,$D7,$F8,$69,$AF,$F5,$B6,$BF,$E1,$1F,$1F,$EF
	.dc.b	$71,$7F,$7C,$8F,$F5,$47,$A5,$F3,$3C,$FF,$E8,$11,$82,$1F,$B9,$E8
	.dc.b	$FF,$5C,$4E,$EF,$87,$91,$F4,$6F,$F1,$17,$FE,$1F,$8C,$1E,$AE,$FF
	.dc.b	$F6,$CC,$7F,$8A,$FE,$1F,$04,$7F,$A6,$72,$7F,$AC,$FD,$CF,$B9,$EA
	.dc.b	$9A,$49,$FF,$1C,$FE,$D8,$F4,$7E,$3B,$63,$31,$83,$CD,$EE,$7C,$2E
	.dc.b	$09,$3F,$E0,$F8,$AF,$BC,$FF,$69,$46,$3C,$2F,$EF,$7B,$2B,$E4,$57
	.dc.b	$C1,$21,$FE,$45,$3E,$06,$51,$FF,$25,$35,$FE,$D5,$FB,$67,$F2,$A6
	.dc.b	$F9,$5F,$D4,$63,$11,$FD,$38,$38,$7F,$A8,$AE,$6F,$DD,$7E,$5A,$FA
	.dc.b	$8A,$7E,$DE,$75,$5F,$C0,$F9,$BF,$AC,$FE,$40,$F9,$4F,$EC,$0D,$9F
	.dc.b	$E2,$ED,$7F,$C0,$0D,$7C,$8F,$EB,$36,$FE,$99,$93,$2F,$F6,$9D,$77
	.dc.b	$3C,$CE,$1F,$9C,$2B,$96,$1F,$DB,$3C,$9F,$F0,$E4,$FF,$28,$6B,$FE
	.dc.b	$11,$6B,$F1,$5A,$F1,$F7,$BF,$BD,$41,$A4,$8F,$C7,$FD,$8F,$FA,$15
	.dc.b	$FF,$38,$35,$FF,$37,$B7,$2F,$F5,$FF,$E2,$DF,$E8,$69,$F6,$1F,$CC
	.dc.b	$7B,$FB,$91,$F4,$3E,$2E,$7F,$E5,$1F,$68,$F3,$29,$FC,$AF,$3F,$CF
	.dc.b	$5F,$C7,$DD,$E7,$BE,$93,$F9,$B3,$67,$4D,$7F,$31,$2F,$AD,$7C,$EF
	.dc.b	$0B,$FC,$2F,$C0,$4B,$E3,$1F,$CE,$FE,$D3,$D9,$1F,$E3,$D6,$BF,$91
	.dc.b	$6B,$FE,$20,$6B,$FE,$3E,$5F,$09,$FE,$29,$6C,$7F,$60,$6B,$D6,$3F
	.dc.b	$C5,$DA,$FF,$45,$8F,$96,$D7,$F0,$AD,$E1,$FF,$77,$FD,$D7,$E0,$9F
	.dc.b	$C7,$F7,$BE,$BF,$F8,$E9,$2F,$E5,$7D,$8B,$E6,$BF,$90,$F8,$7F,$D7
	.dc.b	$FC,$7F,$F3,$15,$7C,$F6,$BF,$2B,$0F,$E1,$4F,$E5,$B5,$FE,$DC,$FA
	.dc.b	$F9,$3D,$C4,$7D,$DD,$DF,$DA,$FF,$2A,$FB,$D1,$F6,$DA,$F3,$8F,$A1
	.dc.b	$EA,$7D,$1E,$E9,$3F,$8C,$C3,$F4,$6F,$CD,$8F,$98,$47,$E6,$3F,$E5
	.dc.b	$7F,$C1,$3D,$3C,$1F,$E3,$D9,$46,$27,$FC,$AE,$D8,$E8,$7E,$3F,$3F
	.dc.b	$EB,$4D,$F4,$FF,$A3,$9F,$E0,$5E,$0F,$B3,$87,$D0,$8F,$95,$AC,$AE
	.dc.b	$07,$D0,$EA,$FE,$9F,$F6,$12,$30,$77,$7E,$96,$D2,$30,$7E,$B3,$ED
	.dc.b	$9F,$63,$FE,$73,$6C,$7F,$8A,$9A,$FF,$A4,$DA,$FF,$27,$5F,$63,$C4
	.dc.b	$91,$C5,$FE,$A9,$C3,$F8,$32,$47,$EF,$3E,$5B,$FB,$1B,$FE,$F0,$E8
	.dc.b	$70,$7E,$27,$8A,$4F,$FB,$DE,$72,$FE,$E7,$C7,$6B,$FD,$E9,$B1,$FE
	.dc.b	$A9,$F0,$3F,$93,$4E,$4F,$EF,$F8,$3E,$0B,$FC,$69,$F5,$1F,$E0,$AB
	.dc.b	$FE,$08,$D7,$F2,$D1,$FE,$24,$C7,$AC,$7F,$02,$97,$D1,$F2,$35,$87
	.dc.b	$4F,$9D,$FF,$06,$C9,$A4,$AF,$36,$3F,$E5,$5E,$6F,$E7,$B5,$F4,$1B
	.dc.b	$C5,$FE,$0C,$F8,$6F,$F0,$6F,$D9,$FA,$38,$E6,$3F,$8D,$8F,$F2,$AF
	.dc.b	$B8,$FB,$A7,$D3,$7F,$99,$23,$0D,$8C,$D7,$FB,$3E,$72,$F9,$3C,$AF
	.dc.b	$8B,$F6,$72,$3F,$35,$1F,$E4,$6F,$8C,$FE,$2B,$ED,$7F,$AA,$7E,$6B
	.dc.b	$FC,$8D,$18,$3F,$8C,$E1,$E3,$F1,$FC,$04,$C3,$C1,$97,$F0,$97,$FC
	.dc.b	$2B,$FB,$33,$EC,$7F,$CA,$0D,$8C,$C6,$0F,$67,$C8,$7E,$BF,$AD,$CB
	.dc.b	$FC,$DF,$97,$E5,$F9,$DF,$12,$BF,$69,$A3,$E6,$7E,$0E,$5F,$E0,$18
	.dc.b	$C7,$1F,$C7,$9C,$CF,$FB,$4A,$72,$FE,$37,$1F,$23,$F8,$3C,$FF,$16
	.dc.b	$7F,$9A,$BF,$6D,$FE,$8B,$1F,$C0,$7D,$4F,$C7,$F9,$32,$30,$7F,$79
	.dc.b	$C3,$C9,$C3,$E1,$24,$7D,$5C,$BE,$33,$FE,$17,$DF,$DF,$F3,$BF,$9A
	.dc.b	$24,$73,$9F,$F2,$BF,$07,$FB,$4F,$0F,$F6,$F7,$9F,$FE,$8D,$FC,$7B
	.dc.b	$E1,$5F,$F6,$E6,$BF,$E3,$59,$DB,$E5,$FC,$AF,$F6,$EF,$C3,$6A,$10
	.dc.b	$8C,$C6,$1B,$97,$D6,$FB,$9F,$D8,$92,$7F,$93,$47,$97,$E8,$7F,$D7
	.dc.b	$70,$C1,$18,$6C,$7F,$B0,$1A,$FD,$BB,$FF,$25,$3F,$9D,$6B,$9F,$83
	.dc.b	$FD,$24,$62,$3D,$43,$AF,$F8,$6F,$B3,$FE,$B9,$FB,$8F,$B7,$3F,$E0
	.dc.b	$26,$ED,$FE,$FF,$AD,$CF,$F8,$36,$73,$E5,$FC,$5D,$11,$89,$FF,$26
	.dc.b	$36,$B8,$C7,$F8,$3B,$F4,$E7,$CF,$F8,$7E,$CE,$4F,$F8,$66,$72,$F7
	.dc.b	$8C,$BE,$57,$FD,$A1,$1A,$8F,$F2,$1F,$DD,$7E,$2B,$3F,$EB,$4B,$EC
	.dc.b	$3F,$DA,$0D,$9F,$F1,$8B,$BD,$3E,$2B,$98,$FF,$34,$07,$0F,$F9,$AD
	.dc.b	$AF,$09,$FB,$71,$C7,$C2,$FF,$5A,$BF,$ED,$49,$46,$3E,$5F,$E9,$BA
	.dc.b	$AB,$F0,$5E,$2F,$9E,$B7,$F9,$1C,$11,$88,$F2,$36,$3F,$C4,$B7,$E0
	.dc.b	$8F,$BF,$8F,$9D,$B5,$D4,$FC,$E4,$65,$A7,$C2,$FA,$0F,$F3,$FB,$EF
	.dc.b	$6B,$9E,$1F,$8A,$FC,$58,$CB,$FA,$7F,$99,$FE,$B3,$D3,$A6,$C7,$CC
	.dc.b	$6B,$FE,$95,$6B,$F4,$BD,$B7,$8B,$DB,$44,$62,$7E,$AC,$62,$75,$39
	.dc.b	$9F,$F5,$07,$D0,$7F,$BA,$B3,$FD,$D6,$BF,$EE,$88,$C4,$FD,$4D,$BF
	.dc.b	$4B,$1F,$D8,$2B,$71,$8F,$97,$46,$2B,$F8,$25,$6D,$E4,$62,$76,$E0
	.dc.b	$46,$EF,$C5,$3A,$7C,$9E,$7F,$F8,$C6,$BC,$6D,$7B,$B5,$E6,$DA,$79
	.dc.b	$DE,$0E,$4F,$F1,$D4,$7F,$86,$B1,$FF,$47,$BA,$6D,$6D,$F8,$35,$6F
	.dc.b	$DA,$D5,$F6,$0F,$96,$C6,$1C,$E7,$F8,$47,$DA,$71,$E4,$FE,$AF,$F5
	.dc.b	$17,$B5,$39,$F3,$13,$98,$9A,$7F,$0D,$F3,$6E,$3F,$EE,$BE,$F7,$DA
	.dc.b	$7E,$EE,$BF,$C3,$7E,$47,$9E,$73,$DB,$D6,$31,$5B,$BE,$4D,$E3,$FF
	.dc.b	$67,$EA,$FB,$3D,$E1,$EF,$FB,$AF,$F7,$26,$7F,$9E,$6E,$3F,$C1,$1A
	.dc.b	$FF,$BB,$5A,$FF,$47,$ED,$63,$F4,$63,$FD,$AF,$3C,$A3,$FA,$C1,$93
	.dc.b	$FC,$E0,$E6,$BF,$96,$70,$7F,$44,$7F,$C7,$3E,$47,$D5,$C1,$F7,$4F
	.dc.b	$33,$E6,$4F,$D7,$FB,$AF,$F5,$CD,$FE,$D9,$B9,$FF,$1C,$6B,$FD,$10
	.dc.b	$38,$7D,$59,$FF,$85,$E7,$FB,$D3,$63,$FD,$8F,$FB,$63,$EA,$FC,$93
	.dc.b	$93,$FB,$A7,$C1,$7F,$B5,$EF,$FC,$04,$DB,$FF,$27,$7D,$6F,$AE,$31
	.dc.b	$5F,$E9,$1C,$5E,$D7,$E7,$DB,$87,$AE,$30,$7E,$97,$AB,$FC,$4B,$F5
	.dc.b	$7F,$8D,$3F,$21,$E4,$DA,$FE,$F5,$87,$F5,$BF,$9F,$7A,$9F,$89,$6C
	.dc.b	$FF,$99,$BE,$87,$13,$F2,$FE,$53,$C4,$49,$F9,$8F,$ED,$79,$7F,$C0
	.dc.b	$BF,$4F,$F5,$53,$FA,$1F,$3B,$0F,$E9,$7E,$C1,$8F,$E6,$7E,$E9,$F4
	.dc.b	$99,$8C,$3F,$B1,$FC,$C1,$ED,$FF,$01,$6E,$9F,$CC,$9B,$CF,$FD,$61
	.dc.b	$AF,$F8,$03,$8D,$F4,$4F,$9F,$FD,$43,$FC,$72,$FE,$29,$FD,$ED,$D3
	.dc.b	$8A,$FF,$04,$46,$DF,$F4,$37,$95,$FB,$BC,$BF,$78,$DA,$E1,$2C,$7F
	.dc.b	$7F,$1E,$7F,$2B,$D4,$47,$FC,$A1,$EE,$7F,$05,$3F,$85,$79,$BF,$E8
	.dc.b	$D8,$FF,$4C,$C7,$EF,$3E,$8F,$F8,$57,$AB,$67,$FD,$3A,$FC,$7C,$BE
	.dc.b	$31,$23,$E3,$3F,$CA,$FC,$CF,$E1,$7F,$29,$FC,$18,$EF,$F0,$96,$BE
	.dc.b	$04,$FF,$10,$7F,$8C,$6B,$FC,$31,$A7,$BB,$56,$7F,$54,$91,$F5,$5F
	.dc.b	$07,$D2,$7F,$47,$ED,$FF,$16,$36,$7F,$C5,$5B,$7C,$6C,$3F,$90,$4E
	.dc.b	$5F,$F2,$97,$88,$EE,$FE,$68,$DB,$F9,$FF,$73,$FC,$75,$FF,$1F,$C6
	.dc.b	$6B,$F8,$9C,$7F,$90,$A5,$EE,$7F,$4B,$6C,$7F,$7C,$6C,$FF,$88,$9A
	.dc.b	$FF,$07,$7F,$66,$F4,$3F,$CC,$CF,$0C,$BF,$B0,$C2,$37,$18,$75,$3F
	.dc.b	$03,$D1,$FE,$91,$F1,$CF,$A5,$FC,$0C,$61,$B3,$F9,$3E,$AF,$EA,$73
	.dc.b	$F8,$6F,$27,$39,$78,$E7,$FD,$DF,$F0,$3D,$0F,$C5,$F8,$F7,$0E,$5C
	.dc.b	$28,$C5,$F1,$8C,$4F,$7C,$FF,$CA,$EF,$CE,$F9,$8F,$A3,$D5,$FD,$97
	.dc.b	$64,$62,$7E,$9F,$16,$63,$4D,$8F,$D2,$FD,$A3,$6C,$C6,$1B,$9F,$CD
	.dc.b	$3E,$57,$8B,$DF,$2F,$03,$6B,$F0,$3E,$FE,$1A,$7E,$12,$3E,$6F,$2F
	.dc.b	$CE,$FB,$9B,$3E,$D7,$F5,$FD,$12,$3E,$2F,$75,$CB,$C5,$F7,$2C,$FF
	.dc.b	$C0,$8D,$AF,$DF,$FD,$BF,$E5,$E9,$39,$7D,$51,$FB,$B7,$FC,$DA,$D7
	.dc.b	$FA,$93,$5F,$F4,$BB,$5F,$F5,$92,$F9,$4F,$F0,$A3,$CF,$AB,$F1,$EE
	.dc.b	$6F,$C7,$F2,$BF,$E1,$3F,$B8,$FF,$34,$C7,$E2,$5B,$87,$E1,$FE,$E5
	.dc.b	$E7,$EF,$D1,$FF,$0D,$46,$2F,$84,$62,$75,$1F,$03,$CC,$F1,$F6,$79
	.dc.b	$1B,$1F,$13,$C9,$C3,$4E,$07,$E0,$BE,$0F,$FC,$5D,$D3,$FE,$03,$8F
	.dc.b	$F5,$BF,$9E,$EE,$CD,$7F,$8C,$E3,$FA,$BE,$BF,$E9,$66,$BF,$DE,$DA
	.dc.b	$F2,$F4,$BF,$C9,$10,$FA,$4E,$51,$CD,$B8,$7F,$38,$F9,$1F,$B1,$C9
	.dc.b	$FB,$A7,$F8,$53,$79,$FF,$C8,$91,$89,$FF,$66,$AE,$CC,$CE,$9B,$3F
	.dc.b	$D1,$3A,$BF,$40,$DF,$67,$FC,$9D,$18,$D3,$F4,$0F,$F8,$85,$BC,$5A
	.dc.b	$EF,$DB,$6F,$F9,$BF,$93,$FE,$23,$7F,$D7,$3E,$5A,$31,$7F,$E1,$F8
	.dc.b	$C7,$B5,$FC,$D0,$FF,$05,$7E,$53,$FC,$46,$F9,$C7,$F9,$6E,$31,$5F
	.dc.b	$CD,$23,$FC,$A1,$7F,$CD,$F7,$FF,$3F,$F9,$3F,$5F,$D3,$CF,$A5,$FC
	.dc.b	$71,$B9,$79,$A3,$A7,$8B,$8C,$BF,$CB,$DC,$5F,$E6,$DA,$F4,$B6,$3D
	.dc.b	$0D,$7F,$DF,$7A,$FE,$CF,$E1,$7F,$C4,$5F,$B7,$3E,$57,$FC,$C9,$7C
	.dc.b	$75,$F2,$CE,$71,$FE,$BB,$39,$58,$B1,$F9,$1E,$7F,$F5,$82,$4F,$CC
	.dc.b	$DD,$1F,$23,$EA,$3E,$1B,$F1,$DB,$3F,$53,$93,$F9,$EC,$7C,$86,$C6
	.dc.b	$49,$87,$F3,$9E,$3F,$CE,$E9,$FD,$9A,$BF,$D1,$9E,$DF,$E7,$29,$C6
	.dc.b	$8C,$37,$AF,$FC,$89,$18,$3F,$41,$8F,$F4,$97,$AB,$F0,$27,$97,$5F
	.dc.b	$6A,$FF,$46,$7D,$C3,$CC,$A6,$3D,$2B,$87,$F8,$1A,$7F,$CC,$4F,$E4
	.dc.b	$1B,$97,$E3,$DA,$FD,$86,$BF,$43,$85,$1F,$E3,$08,$FF,$50,$3C,$B5
	.dc.b	$FF,$65,$CE,$BA,$3C,$BC,$FF,$EB,$5B,$FF,$25,$E1,$FE,$30,$69,$FD
	.dc.b	$24,$FF,$26,$B9,$F1,$38,$7B,$1F,$EC,$B9,$FF,$6A,$CF,$FB,$F1,$A7
	.dc.b	$FB,$FA,$3F,$DA,$1E,$1F,$F5,$E9,$FF,$1E,$B6,$7F,$B8,$7D,$B7,$83
	.dc.b	$3F,$F0,$0C,$E5,$FD,$2C,$66,$BE,$83,$6B,$ED,$F2,$3E,$03,$FD,$95
	.dc.b	$39,$D5,$F6,$FF,$C3,$70,$FF,$1A,$B7,$0F,$EC,$D9,$3D,$2F,$F6,$24
	.dc.b	$E7,$3F,$F7,$57,$C6,$FE,$7B,$DD,$FF,$00,$61,$FA,$01,$EA,$7F,$BE
	.dc.b	$23,$31,$99,$E1,$E2,$7F,$A9,$EB,$4F,$FA,$A6,$53,$FA,$AF,$B8,$E3
	.dc.b	$1D,$34,$C5,$6E,$AC,$F8,$6D,$58,$8F,$01,$FE,$78,$AF,$04,$78,$A3
	.dc.b	$FD,$71,$55,$FD,$CD,$79,$1C,$9C,$4E,$D3,$B7,$F7,$F3,$67,$84,$78
	.dc.b	$3A,$BF,$D0,$EF,$26,$B9,$C7,$D4,$FE,$C9,$F9,$5F,$25,$F6,$73,$FF
	.dc.b	$45,$4F,$F0,$C8,$C3,$B3,$1C,$3F,$EE,$67,$02,$37,$7E,$07,$8D,$FD
	.dc.b	$45,$F3,$FF,$66,$30,$DD,$3C,$78,$75,$9E,$0D,$8F,$3C,$EF,$0F,$33
	.dc.b	$5F,$14,$ED,$F3,$E3,$F5,$70,$F0,$C5,$AF,$C5,$6B,$DA,$B5,$7C,$6B
	.dc.b	$D9,$5D,$6F,$8C,$7D,$53,$BD,$7D,$DE,$1F,$CF,$F0,$7E,$56,$3F,$80
	.dc.b	$F0,$7E,$76,$BF,$43,$39,$E0,$FC,$D7,$0B,$E2,$F1,$5F,$C9,$B8,$FF
	.dc.b	$D9,$AF,$F9,$A5,$AF,$EE,$3D,$0D,$F1,$39,$5F,$3F,$CA,$9F,$71,$FA
	.dc.b	$77,$FC,$A8,$DD,$BF,$CF,$ED,$D3,$F7,$B8,$F1,$E0,$7D,$07,$F0,$D6
	.dc.b	$E1,$FA,$93,$F0,$11,$93,$5F,$2B,$C4,$91,$FE,$CC,$9F,$F7,$6B,$FE
	.dc.b	$B6,$FB,$87,$C3,$7C,$0E,$44,$3E,$9F,$0D,$AE,$15,$89,$F1,$57,$06
	.dc.b	$D6,$EB,$FC,$10,$D7,$FB,$8F,$DB,$BE,$63,$C0,$D8,$FF,$64,$B5,$FF
	.dc.b	$53,$5F,$93,$5F,$9F,$E4,$7A,$A7,$FD,$30,$B4,$C9,$AF,$8C,$B3,$FC
	.dc.b	$3F,$93,$FB,$04,$FC,$96,$C6,$6B,$93,$FD,$DF,$C2,$F8,$9E,$69,$A7
	.dc.b	$E0,$39,$9E,$BE,$2B,$FD,$41,$E9,$7E,$13,$AC,$E6,$3F,$BB,$65,$D3
	.dc.b	$9A,$47,$FA,$D7,$0F,$EE,$AF,$F8,$D2,$3E,$6E,$5F,$90,$8F,$F0,$CB
	.dc.b	$5F,$EC,$4D,$7F,$BE,$F4,$BC,$EF,$B9,$E8,$79,$A7,$FC,$33,$A7,$E6
	.dc.b	$2B,$51,$83,$FC,$53,$1F,$72,$78,$47,$F8,$AA,$3C,$7E,$AF,$C4,$24
	.dc.b	$7F,$B3,$7E,$81,$F3,$18,$8F,$F6,$66,$FF,$E7,$37,$FD,$88,$E7,$BB
	.dc.b	$FC,$83,$4F,$98,$F9,$F5,$0E,$8D,$7F,$CA,$5A,$FF,$6F,$E7,$FF,$2B
	.dc.b	$47,$C6,$3F,$B9,$7D,$A7,$A3,$0F,$F7,$5B,$FC,$C9,$BE,$4F,$ED,$63
	.dc.b	$5D,$3F,$D6,$DD,$37,$93,$46,$E7,$F7,$9F,$69,$FC,$2F,$0F,$F6,$73
	.dc.b	$FA,$A7,$FC,$87,$87,$FB,$EA,$31,$1A,$F1,$78,$F8,$1C,$BC,$09,$3C
	.dc.b	$32,$FE,$D9,$7C,$30,$FE,$BB,$3F,$E4,$EC,$3F,$C5,$8E,$CC,$79,$DF
	.dc.b	$C3,$E0,$DF,$4B,$0F,$F8,$56,$33,$3F,$C2,$5A,$FF,$73,$6B,$FC,$5F
	.dc.b	$EC,$DF,$27,$FC,$6D,$B1,$E1,$0E,$1F,$F7,$84,$BD,$2F,$27,$85,$E5
	.dc.b	$F0,$A5,$71,$AF,$DD,$67,$EE,$8F,$F0,$94,$FA,$FA,$A4,$6E,$3F,$97
	.dc.b	$FD,$9B,$E8,$7F,$98,$FC,$6F,$F9,$E6,$BE,$27,$2A,$31,$0C,$7F,$D1
	.dc.b	$F7,$C6,$37,$3A,$9F,$F4,$9B,$78,$BF,$CB,$CD,$9F,$F3,$4D,$71,$7A
	.dc.b	$4F,$F8,$8F,$EC,$1F,$25,$FD,$05,$F8,$98,$7B,$5B,$3B,$9D,$70,$FF
	.dc.b	$92,$67,$E5,$5F,$F4,$76,$BD,$63,$FC,$9D,$F5,$7F,$63,$0F,$90,$DC
	.dc.b	$3F,$C6,$8D,$AF,$F4,$AD,$FF,$BD,$9F,$DB,$7D,$37,$F5,$8E,$E3,$4C
	.dc.b	$7A,$3F,$C6,$7D,$DF,$E3,$5E,$EC,$BF,$9B,$E1,$7F,$B8,$DF,$C1,$63
	.dc.b	$FE,$D9,$6F,$2F,$C1,$3E,$17,$C2,$D8,$DF,$77,$C5,$30,$FC,$A3,$FD
	.dc.b	$57,$F6,$8F,$88,$F9,$5E,$04,$9F,$C8,$4E,$5F,$0C,$7C,$FE,$19,$18
	.dc.b	$3C,$AF,$F7,$3F,$B1,$F3,$47,$70,$E3,$8F,$EB,$4F,$C5,$F8,$BF,$E7
	.dc.b	$5F,$63,$2F,$F8,$F4,$57,$F8,$EC,$5A,$FE,$37,$3F,$37,$A6,$57,$F2
	.dc.b	$57,$E8,$7C,$A6,$23,$F9,$2F,$D6,$FF,$47,$31,$7F,$A3,$FB,$17,$C9
	.dc.b	$FF,$82,$D7,$CA,$D3,$11,$6F,$FC,$09,$3A,$6C,$FF,$A5,$63,$11,$FE
	.dc.b	$71,$FB,$1F,$EB,$8F,$EE,$DA,$FC,$56,$DF,$8F,$87,$FC,$0D,$D9,$FC
	.dc.b	$EE,$BF,$BA,$B6,$7F,$D5,$ED,$DF,$FD,$8D,$D9,$F3,$5F,$F4,$47,$E6
	.dc.b	$1F,$E2,$6B,$FF,$0D,$35,$FF,$42,$35,$FF,$14,$35,$FF,$60,$57,$E4
	.dc.b	$5F,$E9,$BF,$07,$FD,$BC,$F2,$F6,$79,$B7,$93,$FC,$9F,$CE,$5A,$7F
	.dc.b	$45,$8F,$F8,$DF,$1F,$F1,$6D,$F8,$BD,$FF,$F5,$E4,$62,$3C,$71,$FE
	.dc.b	$20,$BF,$46,$9F,$CD,$BC,$47,$77,$F9,$5A,$30,$DC,$7D,$3E,$AF,$F8
	.dc.b	$14,$98,$FE,$B1,$FF,$25,$FE,$4D,$F9,$D8,$FC,$B7,$90,$FA,$CF,$F6
	.dc.b	$73,$63,$FC,$84,$D7,$FD,$58,$D7,$FD,$6D,$97,$93,$CB,$FF,$0A,$6B
	.dc.b	$27,$F7,$71,$FE,$5A,$F8,$9F,$F0,$BC,$7C,$77,$F8,$BA,$E1,$FF,$5B
	.dc.b	$F9,$1B,$F7,$3F,$CB,$D9,$7F,$80,$2B,$C7,$EB,$F9,$88,$FF,$8C,$7F
	.dc.b	$6B,$F6,$71,$7F,$BD,$3E,$C5,$F4,$3F,$E5,$16,$D7,$F3,$AA,$DF,$1A
	.dc.b	$1F,$03,$FE,$D0,$F2,$7F,$CE,$ED,$8F,$EC,$8D,$7F,$86,$71,$39,$1F
	.dc.b	$29,$FE,$75,$FB,$07,$D9,$FF,$BE,$5B,$7F,$E6,$D6,$BF,$ED,$76,$BF
	.dc.b	$EE,$1E,$37,$D9,$7F,$43,$E0,$FE,$EF,$18,$F5,$7F,$CD,$93,$9D,$BD
	.dc.b	$5A,$7E,$1B,$8F,$FC,$8D,$EB,$4C,$BF,$96,$BF,$82,$F8,$FF,$37,$91
	.dc.b	$E5,$24,$F9,$5F,$F4,$17,$0B,$E5,$5F,$75,$F1,$1B,$5F,$11,$AF,$F8
	.dc.b	$D3,$DF,$F9,$3D,$8F,$5A,$92,$7F,$B5,$FE,$DA,$F2,$7E,$46,$1F,$1C
	.dc.b	$93,$F1,$DF,$F3,$67,$89,$FE,$5A,$8F,$F2,$0F,$07,$F9,$8C,$FA,$FF
	.dc.b	$69,$FC,$33,$8B,$EE,$6C,$D3,$F6,$97,$FE,$0D,$E7,$7C,$E7,$D7,$3C
	.dc.b	$FE,$4D,$9F,$AD,$1F,$64,$8F,$FA,$BC,$FC,$C4,$7D,$16,$F3,$FF,$A7
	.dc.b	$E3,$19,$FE,$16,$71,$3F,$EB,$CF,$B0,$C9,$C7,$A7,$F8,$44,$7F,$B0
	.dc.b	$3E,$D1,$F3,$54,$97,$FE,$12,$8F,$F1,$BB,$F9,$27,$6F,$F8,$5A,$FF
	.dc.b	$71,$D9,$FC,$25,$BB,$FE,$1A,$B7,$5F,$D5,$FE,$D3,$FB,$BB,$FE,$31
	.dc.b	$6B,$9F,$7F,$FD,$A1,$F5,$1F,$E0,$18,$4E,$B1,$FF,$17,$2F,$B2,$FF
	.dc.b	$6B,$36,$7A,$5E,$27,$FD,$9D,$1C,$9B,$3B,$8F,$9C,$D7,$FD,$BC,$FF
	.dc.b	$83,$DF,$B9,$DF,$FB,$97,$C7,$F4,$97,$93,$FD,$5F,$18,$3F,$7C,$E2
	.dc.b	$B9,$36,$78,$57,$39,$46,$2B,$F8,$34,$61,$E7,$3B,$F6,$B5,$1F,$13
	.dc.b	$C6,$FC,$B3,$CD,$AF,$26,$BB,$E3,$FF,$78,$7C,$B7,$F4,$9E,$8B,$CD
	.dc.b	$FC,$52,$BF,$21,$39,$6C,$65,$F8,$BF,$02,$F9,$CF,$F7,$D6,$88,$C4
	.dc.b	$FC,$C6,$BC,$27,$FE,$00,$7C,$EF,$F1,$8E,$5F,$6C,$FF,$25,$8F,$F7
	.dc.b	$64,$62,$11,$B7,$E9,$9B,$F9,$7F,$F4,$6F,$0F,$D2,$9D,$37,$C9,$CF
	.dc.b	$0F,$D9,$F8,$CF,$F2,$EE,$3D,$6F,$73,$F7,$57,$DA,$7F,$9E,$F2,$C3
	.dc.b	$5F,$33,$5F,$2B,$7A,$3F,$DC,$F7,$FE,$8D,$AE,$6F,$FA,$3E,$33,$5F
	.dc.b	$E3,$99,$CB,$FE,$3B,$8C,$F4,$7F,$A2,$DB,$5F,$91,$FD,$2B,$E2,$D6
	.dc.b	$FD,$3F,$F5,$6D,$FE,$C7,$E4,$AD,$FF,$36,$ED,$FF,$6D,$BF,$B7,$8C
	.dc.b	$DB,$9D,$5A,$8F,$E9,$EF,$FB,$6E,$B8,$1F,$BF,$FC,$07,$8D,$70,$BB
	.dc.b	$4F,$8B,$EB,$E1,$3E,$DB,$F6,$D8,$FF,$07,$9C,$C6,$32,$EF,$1C,$8E
	.dc.b	$7E,$6D,$AF,$EC,$7A,$7F,$9A,$DF,$E4,$9D,$9F,$EA,$3F,$81,$FA,$8D
	.dc.b	$8C,$CE,$F0,$FD,$C1,$EF,$CF,$FD,$0F,$3D,$2F,$D3,$E1,$7C,$4C,$39
	.dc.b	$5F,$F9,$B7,$EC,$BF,$D6,$D7,$FC,$CE,$11,$87,$F9,$B3,$5E,$ED,$7F
	.dc.b	$A2,$E3,$FE,$06,$9F,$2F,$B1,$E8,$8C,$4F,$9A,$FF,$D7,$D3,$F8,$FE
	.dc.b	$07,$D9,$FF,$B9,$30,$FF,$62,$B6,$FE,$BF,$8B,$F5,$23,$18,$FF,$33
	.dc.b	$F6,$FE,$A9,$F5,$BE,$3E,$E1,$FE,$0E,$CD,$EF,$7C,$A6,$C7,$CE,$6E
	.dc.b	$5E,$56,$BF,$E3,$4E,$0F,$E3,$33,$FD,$0F,$C8,$FC,$1F,$91,$F8,$47
	.dc.b	$F0,$BF,$11,$27,$FA,$47,$C4,$DC,$3F,$D5,$18,$FF,$89,$3D,$9F,$11
	.dc.b	$2B,$C4,$FC,$3C,$9E,$5F,$E4,$0F,$C8,$FC,$C5,$AF,$EA,$9F,$1F,$F9
	.dc.b	$4F,$2B,$F2,$AD,$8F,$F8,$01,$F9,$8B,$DA,$FD,$D3,$63,$EF,$37,$4F
	.dc.b	$EE,$4D,$7E,$07,$C7,$FE,$8A,$9F,$27,$FA,$28,$C3,$FA,$6F,$1B,$F6
	.dc.b	$FC,$FF,$D9,$71,$C7,$9B,$FA,$87,$F6,$78,$BF,$67,$FF,$33,$47,$E3
	.dc.b	$DB,$1F,$E0,$36,$BF,$9C,$6B,$FE,$34,$BF,$F5,$7C,$FF,$93,$EF,$FC
	.dc.b	$F9,$18,$9F,$F2,$5B,$FC,$AE,$33,$EF,$79,$FC,$1F,$47,$C7,$DC,$39
	.dc.b	$1D,$5F,$67,$99,$F9,$5C,$3F,$8C,$E3,$FE,$76,$6B,$E3,$8D,$78,$3F
	.dc.b	$0C,$F8,$9F,$53,$70,$EB,$5F,$C4,$ED,$F5,$5F,$27,$DF,$F9,$4D,$AD
	.dc.b	$F8,$8F,$AD,$FD,$76,$1F,$C6,$DB,$19,$F6,$FF,$55,$18,$C7,$FC,$D3
	.dc.b	$7C,$9B,$7F,$13,$ED,$3E,$C7,$C3,$7D,$4F,$B2,$FE,$8F,$3F,$E3,$36
	.dc.b	$BF,$C3,$1A,$FF,$81,$5A,$FF,$9F,$D7,$8F,$FD,$1D,$B8,$7F,$0B,$6C
	.dc.b	$F5,$9F,$E9,$9F,$87,$FE,$03,$95,$7D,$67,$F0,$F3,$98,$C1,$EB,$FA
	.dc.b	$DF,$C8,$7C,$2F,$E4,$5B,$19,$26,$1F,$82,$F8,$2D,$47,$C3,$F8,$8F
	.dc.b	$CC,$1F,$99,$FA,$3B,$1F,$1D,$C3,$F2,$7E,$96,$DB,$4F,$3B,$F0,$7C
	.dc.b	$AF,$94,$7E,$7B,$EC,$19,$E3,$0F,$C0,$D2,$73,$1F,$CB,$DB,$1F,$E6
	.dc.b	$36,$BF,$CF,$1A,$FF,$8D,$7B,$3F,$39,$3F,$E3,$79,$E3,$87,$92,$95
	.dc.b	$A7,$FD,$69,$3A,$AF,$B5,$E2,$FE,$04,$FE,$BF,$F3,$4F,$A5,$C4,$FE
	.dc.b	$4A,$30,$D8,$FE,$01,$EF,$70,$7F,$81,$7D,$2C,$9F,$C8,$7C,$25,$CB
	.dc.b	$F6,$67,$F0,$77,$F8,$43,$63,$E2,$FD,$97,$EE,$63,$FA,$DB,$5F,$F2
	.dc.b	$A3,$5F,$F6,$73,$52,$F0,$FF,$C3,$0D,$9F,$EB,$8D,$9F,$F4,$83,$5F
	.dc.b	$ED,$3C,$DF,$A2,$DF,$E5,$4E,$DF,$99,$19,$AF,$1D,$FF,$81,$BB,$1E
	.dc.b	$9D,$31,$E9,$5A,$FD,$DC,$A3,$31,$8F,$4E,$E7,$F8,$D8,$F7,$F3,$FF
	.dc.b	$28,$D0,$FE,$93,$CF,$C3,$67,$F1,$BA,$27,$F5,$3C,$F9,$C3,$A1,$FE
	.dc.b	$F5,$FB,$1F,$F6,$55,$FF,$7F,$6C,$7F,$AC,$DA,$FF,$83,$1A,$FF,$A1
	.dc.b	$57,$CD,$FF,$91,$5B,$3C,$E3,$07,$EF,$9F,$E8,$31,$A9,$F1,$8C,$3F
	.dc.b	$80,$CF,$E9,$A7,$5C,$1F,$9B,$F5,$3F,$B1,$CF,$EA,$9B,$7C,$BF,$0F
	.dc.b	$FF,$26,$D6,$FD,$4F,$E9,$F1,$8B,$FF,$4E,$36,$BF,$95,$46,$BC,$CF
	.dc.b	$EA,$BC,$CE,$FA,$7F,$90,$BE,$B7,$FA,$66,$5F,$AD,$9D,$5F,$F3,$69
	.dc.b	$CA,$C3,$FC,$A1,$18,$6F,$C4,$73,$F3,$FF,$C9,$BF,$54,$FB,$5F,$F1
	.dc.b	$13,$70,$FF,$72,$79,$A8,$C6,$1F,$E3,$E8,$C5,$70,$F9,$0E,$6F,$46
	.dc.b	$DF,$F4,$AB,$FF,$13,$BE,$5F,$85,$7D,$5F,$F4,$4D,$BF,$8D,$E3,$DF
	.dc.b	$EA,$7F,$CB,$B3,$9B,$F6,$4E,$5E,$93,$EA,$F8,$DF,$5B,$FE,$68,$FB
	.dc.b	$37,$AD,$FE,$3D,$6D,$7B,$1B,$B7,$AD,$AF,$FB,$47,$0F,$81,$ED,$73
	.dc.b	$BF,$8E,$DB,$F6,$78,$DF,$20,$8F,$C8,$DB,$F8,$BE,$1F,$1D,$FF,$46
	.dc.b	$BF,$E8,$37,$3A,$FC,$07,$38,$FC,$FB,$FA,$BF,$31,$F0,$BC,$DB,$8F
	.dc.b	$F7,$AE,$0F,$93,$18,$CB,$F5,$B8,$7D,$7B,$FB,$BF,$2F,$9C,$FE,$6F
	.dc.b	$C1,$7E,$86,$FD,$CD,$7E,$6B,$5F,$F4,$C3,$6F,$F0,$5E,$07,$FA,$52
	.dc.b	$3F,$4B,$9F,$D4,$D6,$DF,$0A,$FF,$4A,$FD,$9F,$F9,$A6,$3F,$DF,$9A
	.dc.b	$A3,$11,$F8,$36,$A8,$D7,$17,$FB,$0D,$E1,$C1,$6B,$FC,$87,$8F,$FA
	.dc.b	$42,$F8,$46,$1F,$F4,$9D,$70,$9F,$A9,$E6,$F9,$2F,$F4,$2D,$8F,$EC
	.dc.b	$B5,$FE,$E2,$7F,$D2,$4D,$3F,$DD,$6E,$3C,$DF,$F6,$A6,$FF,$CA,$E3
	.dc.b	$FD,$30,$D9,$F2,$35,$FF,$52,$B5,$FC,$DF,$5B,$D1,$FD,$F1,$FB,$4F
	.dc.b	$A4,$B7,$FA,$97,$0F,$EA,$FE,$56,$DF,$F1,$A1,$DF,$FE,$5E,$6C,$FB
	.dc.b	$3D,$0C,$4E,$65,$F2,$9F,$C1,$FE,$0F,$F8,$84,$7F,$09,$15,$FE,$0E
	.dc.b	$8C,$C6,$67,$87,$87,$FE,$C0,$AD,$3F,$4E,$3F,$89,$7D,$9F,$0A,$E7
	.dc.b	$8E,$2B,$50,$9D,$5F,$F8,$71,$E1,$1E,$63,$FD,$15,$3D,$EB,$C3,$1F
	.dc.b	$ED,$0B,$AF,$F2,$1B,$F1,$B9,$38,$9E,$D7,$2F,$AC,$C9,$C6,$79,$BF
	.dc.b	$CA,$E7,$F8,$8E,$1E,$8E,$8F,$F4,$E4,$F5,$6B,$CD,$BF,$2E,$D7,$FC
	.dc.b	$71,$29,$CC,$61,$BD,$5D,$A7,$F9,$D5,$71,$8C,$42,$30,$FA,$17,$D9
	.dc.b	$FD,$9C,$3E,$BC,$62,$7E,$2B,$C2,$33,$87,$91,$DD,$71,$D3,$F0,$55
	.dc.b	$C6,$FB,$5E,$E7,$2F,$B3,$5E,$33,$F7,$4F,$4D,$FE,$2E,$3B,$E3,$99
	.dc.b	$F5,$57,$37,$67,$9A,$BC,$4F,$C2,$8D,$3F,$07,$E9,$3F,$2D,$5F,$E4
	.dc.b	$FC,$51,$8D,$7D,$2D,$AE,$A7,$F2,$6E,$BD,$BF,$D5,$63,$5A,$FF,$5F
	.dc.b	$6C,$FC,$1C,$3F,$AA,$F1,$FF,$88,$AF,$FB,$87,$A1,$C2,$31,$1D,$6F
	.dc.b	$C0,$BE,$CB,$F8,$5B,$4F,$E1,$AE,$0C,$E6,$F8,$CF,$E2,$2D,$9F,$E3
	.dc.b	$C7,$E0,$63,$26,$33,$FE,$EC,$FB,$FA,$DF,$64,$F8,$CC,$36,$FF,$BA
	.dc.b	$56,$27,$AD,$70,$6D,$6E,$BF,$95,$B5,$FE,$76,$BE,$2F,$64,$8F,$E6
	.dc.b	$95,$E2,$8F,$36,$EF,$A5,$F9,$FE,$1D,$BF,$F6,$7B,$5F,$AD,$1A,$3D
	.dc.b	$D1,$98,$FE,$B3,$18,$3A,$FA,$5F,$60,$91,$FE,$11,$F9,$BA,$FF,$94
	.dc.b	$B1,$FF,$08,$F7,$7C,$7F,$33,$F9,$94,$FF,$85,$EF,$F0,$3D,$7F,$82
	.dc.b	$E0,$93,$FE,$1B,$8A,$FF,$7D,$8F,$69,$1F,$F1,$4F,$C5,$3E,$27,$F1
	.dc.b	$BC,$BF,$E9,$D9,$FF,$76,$B5,$FE,$F3,$3B,$73,$E5,$FD,$D9,$D4,$7F
	.dc.b	$85,$5A,$FF,$89,$1A,$FF,$0D,$97,$D2,$FE,$71,$B1,$FE,$FC,$6B,$E0
	.dc.b	$FA,$6D,$4F,$C3,$F0,$E4,$EE,$39,$7C,$DF,$3F,$9B,$93,$CE,$2B,$FE
	.dc.b	$45,$C1,$3F,$95,$6C,$7C,$48,$FC,$C7,$82,$F2,$7F,$62,$D9,$FF,$2D
	.dc.b	$78,$1F,$E4,$48,$C4,$7F,$87,$9F,$F4,$0B,$FE,$DC,$FB,$17,$17,$EC
	.dc.b	$1B,$19,$8E,$B7,$FB,$4F,$CC,$1F,$2D,$49,$5E,$0F,$13,$FC,$9F,$8F
	.dc.b	$F3,$98,$FF,$74,$FD,$93,$E2,$3F,$7A,$D8,$CC,$72,$7F,$D1,$BF,$96
	.dc.b	$3D,$8F,$E0,$73,$67,$FD,$E6,$D7,$71,$A3,$FC,$23,$19,$8F,$EE,$87
	.dc.b	$86,$32,$6B,$D3,$91,$FC,$1A,$51,$88,$FF,$13,$7E,$03,$FC,$81,$FF
	.dc.b	$50,$B7,$83,$FD,$F0,$DC,$FF,$C0,$4B,$D6,$F4,$F8,$32,$FD,$2F,$F1
	.dc.b	$E9,$E1,$1F,$D7,$75,$FE,$BC,$FC,$B3,$1D,$9F,$78,$FB,$FD,$BF,$CC
	.dc.b	$3E,$CF,$F4,$13,$FE,$5A,$6E,$7F,$E8,$46,$E7,$F5,$2F,$F9,$6F,$47
	.dc.b	$FA,$66,$FB,$0F,$B3,$FB,$AE,$1A,$FF,$B0,$23,$83,$FE,$C2,$8C,$BF
	.dc.b	$39,$AF,$F8,$8D,$AF,$FA,$B5,$AF,$7F,$91,$FF,$23,$C1,$A4,$8F,$E0
	.dc.b	$7F,$41,$FB,$3F,$A1,$91,$FB,$38,$FD,$B7,$D0,$C8,$FD,$B3,$A7,$FC
	.dc.b	$FB,$F8,$AF,$E4,$DB,$E9,$FF,$58,$8C,$6D,$FE,$AA,$EF,$7D,$0F,$F5
	.dc.b	$2D,$1F,$11,$49,$8F,$D5,$DD,$8F,$2E,$0F,$F2,$17,$D8,$BE,$07,$FA
	.dc.b	$DF,$DA,$3E,$5A,$9F,$D7,$A7,$58,$FD,$29,$D3,$FD,$61,$AF,$F9,$DD
	.dc.b	$AF,$E7,$1A,$FC,$FF,$4B,$D0,$47,$A5,$7F,$1B,$E4,$91,$FC,$6D,$FE
	.dc.b	$3D,$E0,$FF,$1F,$F9,$49,$3F,$C7,$DD,$BF,$CB,$BE,$3F,$F9,$21,$1F
	.dc.b	$E4,$9F,$25,$5F,$E6,$7F,$27,$FC,$A4,$C7,$2B,$7F,$F3,$37,$E6,$1F
	.dc.b	$E6,$6A,$FC,$8F,$DA,$3E,$3B,$FC,$02,$D8,$FC,$AC,$ED,$5F,$37,$FD
	.dc.b	$F1,$3B,$7F,$D3,$8D,$7F,$80,$B5,$FF,$3B,$35,$E3,$F2,$FF,$E9,$88
	.dc.b	$34,$91,$FD,$53,$9B,$FA,$21,$1F,$E8,$9F,$2B,$FB,$3D,$BF,$C1,$9E
	.dc.b	$01,$A4,$9F,$EB,$BE,$EF,$F4,$DF,$22,$47,$F4,$D8,$FE,$A1,$B4,$8F
	.dc.b	$EA,$0F,$F5,$3E,$5F,$F1,$7F,$D8,$3E,$8F,$FC,$34,$DA,$FF,$5E,$76
	.dc.b	$23,$11,$FC,$8A,$31,$3C,$38,$B9,$FC,$9C,$AF,$CE,$5F,$CB,$E2,$F9
	.dc.b	$D8,$73,$E9,$F0,$92,$B7,$C5,$FE,$38,$F0,$3F,$27,$00,$E1,$B3,$FE
	.dc.b	$C6,$8D,$D7,$9F,$E4,$3D,$BF,$29,$2F,$C7,$1F,$03,$E8,$BF,$C9,$FF
	.dc.b	$16,$FA,$9A,$66,$BF,$94,$7C,$1B,$F2,$FF,$E9,$D1,$FD,$0A,$31,$CD
	.dc.b	$FE,$44,$6C,$7F,$B8,$DA,$FC,$7E,$07,$D2,$7D,$0D,$A4,$FC,$68,$FA
	.dc.b	$7F,$45,$23,$E4,$BF,$5B,$E8,$FF,$CD,$58,$7F,$B1,$3A,$7C,$1F,$23
	.dc.b	$FE,$43,$BF,$8D,$39,$7C,$51,$97,$FD,$41,$93,$E8,$7F,$02,$48,$FA
	.dc.b	$FF,$45,$F8,$4F,$93,$91,$F6,$9F,$C3,$F9,$7E,$2F,$CB,$FF,$07,$3F
	.dc.b	$ED,$7C,$91,$88,$F1,$B5,$46,$BB,$7F,$1F,$F2,$7F,$92,$24,$FE,$23
	.dc.b	$E2,$FF,$D0,$7F,$63,$F8,$08,$FE,$4C,$D7,$F7,$3E,$17,$E1,$7B,$13
	.dc.b	$9E,$FF,$C7,$F0,$3E,$63,$E4,$F5,$FF,$6B,$C7,$EB,$35,$FB,$8D,$7D
	.dc.b	$8D,$7F,$DE,$9F,$60,$E2,$FF,$41,$C2,$30,$FF,$A1,$9A,$FF,$82,$0D
	.dc.b	$47,$EB,$BE,$27,$C7,$F3,$12,$7F,$4F,$15,$F1,$C7,$E3,$6C,$D3,$F6
	.dc.b	$F1,$FB,$A2,$1F,$B0,$E6,$AF,$F9,$BE,$3F,$BC,$AF,$90,$E7,$F1,$FA
	.dc.b	$D7,$F6,$BF,$61,$FE,$23,$6F,$C2,$FF,$AD,$23,$1C,$5F,$3D,$FE,$F3
	.dc.b	$3A,$C3,$FA,$D4,$3E,$7F,$81,$C6,$FF,$D6,$7F,$51,$FE,$24,$AF,$F0
	.dc.b	$3B,$5F,$B6,$7E,$FD,$C5,$FF,$35,$8F,$F1,$D7,$C0,$72,$F8,$7F,$F5
	.dc.b	$EF,$D8,$3E,$06,$99,$AF,$F0,$9C,$65,$FF,$5C,$B5,$FA,$8D,$7F,$C7
	.dc.b	$AD,$7F,$88,$07,$F9,$93,$FC,$4F,$85,$F2,$BA,$9C,$CA,$72,$FF,$19
	.dc.b	$8C,$C6,$21,$F1,$5F,$E4,$38,$C4,$EA,$BA,$7B,$5E,$29,$CB,$63,$C1
	.dc.b	$EA,$5C,$3B,$7A,$FF,$E4,$07,$D6,$D7,$CC,$D7,$7D,$BF,$ED,$CF,$99
	.dc.b	$FC,$6F,$85,$7A,$3F,$6F,$5F,$99,$9C,$B6,$32,$FC,$8F,$33,$FB,$63
	.dc.b	$F0,$DA,$FC,$16,$BF,$2D,$CD,$BF,$4E,$7C,$47,$F7,$3D,$BE,$4C,$BA
	.dc.b	$3F,$CD,$7D,$3F,$47,$7F,$E0,$18,$70,$8C,$64,$BE,$1C,$9C,$EF,$79
	.dc.b	$5A,$D6,$1C,$5F,$B1,$3D,$2F,$47,$96,$AE,$2F,$9F,$FC,$A5,$FE,$8D
	.dc.b	$AF,$86,$35,$97,$F6,$A6,$C7,$F6,$9F,$D3,$BF,$DA,$93,$FA,$8A,$FF
	.dc.b	$09,$74,$7F,$66,$9C,$D7,$E9,$31,$F8,$55,$AB,$FE,$41,$5A,$BE,$1D
	.dc.b	$99,$C3,$E1,$4F,$FB,$4A,$BF,$C9,$CE,$DF,$DD,$7C,$C6,$F1,$F9,$CD
	.dc.b	$97,$C5,$E7,$34,$E6,$30,$DB,$FE,$02,$EF,$D4,$91,$FE,$91,$F0,$3F
	.dc.b	$8C,$F3,$E3,$3F,$E2,$8D,$D7,$FC,$65,$3A,$AF,$E1,$77,$E1,$FA,$0F
	.dc.b	$9B,$AF,$E0,$DB,$A1,$A6,$B5,$77,$71,$C3,$E0,$B7,$4F,$99,$3F,$E7
	.dc.b	$69,$FE,$33,$F2,$1F,$07,$E2,$FE,$03,$F0,$FE,$1B,$A1,$27,$FC,$4B
	.dc.b	$5F,$ED,$A6,$FC,$3F,$C9,$3E,$57,$7D,$8F,$E4,$BF,$45,$F2,$CF,$9D
	.dc.b	$F1,$16,$3F,$94,$FC,$A7,$B0,$8F,$E2,$5F,$67,$27,$D3,$EF,$FA,$CD
	.dc.b	$9F,$BB,$EB,$FD,$7F,$8E,$FB,$4D,$8F,$E6,$BF,$21,$F2,$08,$FE,$55
	.dc.b	$F9,$7F,$35,$F7,$CF,$C2,$7E,$9C,$F8,$6C,$57,$DB,$FC,$17,$FC,$66
	.dc.b	$DF,$81,$DF,$C3,$FE,$47,$B5,$5A,$E8,$FC,$D7,$5B,$E8,$3E,$C9,$31
	.dc.b	$FD,$33,$F7,$B3,$7C,$3F,$E4,$7E,$5B,$C6,$DA,$FF,$54,$4B,$E1,$FF
	.dc.b	$2D,$E4,$7A,$1B,$3F,$D5,$3E,$67,$F0,$E4,$7F,$68,$FE,$2B,$C9,$FC
	.dc.b	$F1,$F0,$3E,$1A,$CF,$6F,$6B,$F4,$5E,$2F,$C4,$6C,$7F,$AE,$7F,$53
	.dc.b	$FD,$BF,$07,$F4,$8A,$46,$23,$FC,$2C,$D7,$F8,$B3,$78,$9F,$89,$BD
	.dc.b	$9F,$DA,$4B,$E7,$FE,$64,$E7,$1F,$04,$63,$CF,$DF,$CF,$D9,$F2,$3C
	.dc.b	$BC,$FF,$D4,$E1,$FB,$C9,$FF,$3E,$B7,$3F,$F3,$E3,$5F,$52,$F9,$FF
	.dc.b	$19,$F8,$4F,$7E,$D7,$D2,$F7,$FC,$FF,$37,$F8,$86,$C7,$F9,$E7,$D2
	.dc.b	$FE,$F5,$F5,$FF,$E9,$B8,$FF,$12,$64,$8C,$3F,$E2,$76,$BF,$DC,$8D
	.dc.b	$7B,$9F,$CB,$9E,$1E,$67,$F1,$F7,$35,$DC,$FE,$F3,$49,$D3,$DA,$FF
	.dc.b	$D2,$7E,$83,$EA,$7E,$7F,$A7,$FC,$8E,$31,$87,$DD,$8C,$4C,$8F,$F7
	.dc.b	$C4,$FD,$3C,$7F,$9B,$3C,$1F,$E7,$1B,$FF,$85,$67,$F0,$DF,$01,$D7
	.dc.b	$DA,$FB,$AD,$8E,$71,$E1,$F6,$31,$7C,$BD,$5F,$C1,$FA,$DF,$84,$6D
	.dc.b	$F8,$BD,$8F,$E4,$B9,$FF,$A6,$E3,$FD,$72,$DB,$FF,$5C,$35,$FC,$B3
	.dc.b	$FC,$A6,$3D,$7E,$C7,$C1,$F5,$BE,$13,$F9,$2F,$5E,$4F,$F3,$BE,$D7
	.dc.b	$89,$8C,$7F,$C7,$3E,$07,$E6,$A3,$F2,$E7,$CB,$F5,$AD,$FD,$5A,$FF
	.dc.b	$5E,$7E,$04,$F9,$7F,$B4,$D9,$D4,$BE,$BB,$F8,$94,$62,$75,$3F,$E4
	.dc.b	$83,$5F,$2D,$F8,$5A,$34,$95,$F9,$0F,$C6,$8F,$EB,$BB,$7F,$15,$18
	.dc.b	$8F,$DE,$47,$E3,$69,$F2,$DD,$49,$3F,$9A,$8C,$B7,$CC,$E1,$3C,$5B
	.dc.b	$1F,$C1,$FB,$FF,$7A,$7E,$3F,$DA,$B1,$F8,$C7,$0F,$E3,$BD,$8B,$A7
	.dc.b	$BF,$EB,$BF,$1D,$E0,$FF,$99,$5B,$5F,$E5,$87,$FC,$E4,$FC,$A7,$33
	.dc.b	$FE,$EB,$72,$F3,$F8,$9F,$D5,$47,$E6,$EB,$A4,$7E,$B7,$EC,$7F,$99
	.dc.b	$DF,$F1,$FF,$9E,$C7,$AB,$FC,$59,$E3,$EA,$7E,$D7,$1F,$F3,$2D,$7E
	.dc.b	$F9,$B9,$7F,$A8,$DA,$FE,$F5,$FC,$5C,$79,$E7,$C7,$CF,$D2,$3F,$07
	.dc.b	$F8,$2F,$F3,$F6,$F0,$7F,$01,$8F,$E9,$DA,$FF,$A1,$DD,$63,$F8,$57
	.dc.b	$35,$FE,$7C,$6C,$03,$FE,$F6,$8E,$FE,$E7,$F0,$DF,$BC,$FF,$23,$5F
	.dc.b	$FA,$4D,$B3,$FE,$86,$7F,$CA,$0F,$E6,$1C,$DF,$57,$2F,$E5,$DB,$3F
	.dc.b	$E9,$DF,$DB,$3F,$C6,$93,$BF,$57,$FA,$0F,$C2,$7F,$38,$26,$1F,$CE
	.dc.b	$1F,$DF,$EB,$A8,$FE,$26,$DA,$FE,$23,$D9,$FD,$28,$FE,$39,$EC,$DA
	.dc.b	$FE,$32,$E1,$FD,$A6,$0F,$DB,$FC,$6B,$C5,$F5,$BF,$D7,$3E,$13,$FA
	.dc.b	$59,$2B,$FA,$5B,$FE,$36,$AF,$EB,$11,$FD,$96,$3F,$C9,$D1,$8C,$7F
	.dc.b	$92,$C7,$F5,$EF,$8A,$FE,$AA,$4A,$FE,$AA,$F3,$F5,$B5,$1E,$6B,$FE
	.dc.b	$B1,$BF,$EE,$B0,$FE,$D7,$7D,$BD,$4F,$ED,$3F,$0B,$70,$FE,$EC,$DA
	.dc.b	$FF,$36,$60,$AF,$2F,$A1,$9C,$7A,$69,$FD,$EB,$E7,$FF,$B4,$93,$2F
	.dc.b	$ED,$2F,$BF,$CF,$FE,$EE,$FB,$9B,$1E,$7F,$2F,$FB,$DB,$F0,$9A,$FA
	.dc.b	$FC,$5F,$90,$47,$FB,$06,$9F,$E9,$BD,$3F,$B1,$BF,$97,$C7,$EB,$78
	.dc.b	$7F,$32,$FE,$6B,$6F,$E3,$73,$7F,$9F,$72,$7C,$BF,$D2,$8C,$3F,$53
	.dc.b	$9B,$EC,$12,$BC,$38,$FE,$0E,$3E,$E7,$27,$DA,$3E,$DF,$1D,$AF,$A8
	.dc.b	$E1,$FA,$DB,$DD,$7F,$9B,$1F,$EE,$91,$F6,$0F,$9B,$FE,$FD,$EC,$FD
	.dc.b	$B8,$C4,$FF,$80,$7E,$3B,$F6,$D1,$88,$FD,$C4,$EE,$BE,$3D,$7E,$EF
	.dc.b	$AB,$54,$AF,$C3,$46,$23,$FD,$EC,$7F,$B4,$2B,$C3,$8F,$17,$9C,$79
	.dc.b	$9C,$B4,$F2,$14,$F1,$C7,$9A,$38,$CF,$8A,$BC,$31,$83,$9C,$78,$EF
	.dc.b	$C1,$7F,$C4,$CF,$45,$E1,$C4,$57,$27,$A4,$F1,$EB,$FE,$EA,$6D,$7A
	.dc.b	$A3,$11,$C8,$FF,$73,$47,$7F,$8F,$79,$2E,$7F,$C7,$4E,$63,$15,$C0
	.dc.b	$FE,$51,$5C,$2B,$71,$89,$DD,$F8,$2B,$53,$E7,$0B,$E2,$FE,$06,$1F
	.dc.b	$76,$31,$3F,$05,$E3,$19,$BF,$1B,$BA,$E3,$9F,$AD,$B7,$E3,$C7,$95
	.dc.b	$FA,$B5,$E1,$3F,$5C,$F5,$D7,$D1,$8E,$D5,$99,$F3,$D7,$82,$FB,$47
	.dc.b	$96,$FB,$BA,$AE,$E7,$1C,$7D,$71,$F8,$76,$E1,$DA,$3F,$19,$93,$F5
	.dc.b	$1F,$F5,$84,$7F,$34,$6B,$F5,$75,$FE,$05,$18,$C3,$F8,$BB,$FB,$E9
	.dc.b	$E1,$E7,$FE,$2E,$72,$DA,$E2,$D9,$5F,$87,$E3,$30,$F8,$FE,$87,$FC
	.dc.b	$EE,$D9,$FE,$0B,$8A,$72,$D8,$CF,$C6,$F2,$E9,$F3,$65,$FC,$65,$B3
	.dc.b	$E0,$A5,$78,$1F,$F1,$7C,$7C,$1F,$1B,$06,$7D,$0B,$1E,$28,$F0,$91
	.dc.b	$F2,$F9,$3E,$1D,$64,$7C,$B8,$ED,$F3,$FE,$5C,$64,$FC,$F8,$AF,$FA
	.dc.b	$47,$56,$7E,$68,$FE,$4A,$D7,$4D,$7F,$D8,$59,$BE,$23,$D8,$48,$FB
	.dc.b	$9F,$3F,$5F,$B8,$7D,$2A,$F9,$4E,$2F,$C3,$BE,$F8,$7E,$FF,$2F,$CD
	.dc.b	$F4,$25,$FE,$1E,$2B,$E4,$8F,$B3,$87,$FB,$B2,$3D,$31,$F6,$3E,$77
	.dc.b	$EB,$71,$49,$FC,$84,$57,$DB,$1C,$88,$FE,$0B,$AF,$EC,$11,$FC,$C7
	.dc.b	$C2,$FE,$07,$8A,$47,$F8,$3B,$1F,$F7,$C3,$F9,$DA,$FC,$7F,$EA,$3F
	.dc.b	$EB,$38,$D7,$2F,$E2,$4C,$FC,$25,$8F,$A3,$1E,$C2,$3F,$EC,$CF,$DC
	.dc.b	$1C,$5F,$9D,$84,$61,$FF,$46,$B5,$FF,$45,$3F,$5D,$CF,$D0,$F1,$83
	.dc.b	$4C,$4F,$5F,$AC,$FF,$73,$57,$FA,$61,$AF,$FA,$55,$BC,$5F,$E1,$78
	.dc.b	$43,$CF,$E5,$E7,$49,$E4,$F5,$F8,$AF,$57,$E3,$7F,$D8,$70,$E4,$BE
	.dc.b	$8B,$26,$B9,$EE,$1F,$EE,$E9,$D3,$67,$FD,$E2,$A3,$2F,$F8,$43,$F5
	.dc.b	$37,$9F,$FB,$9F,$A5,$93,$FE,$B4,$9C,$BF,$EB,$6F,$2D,$F0,$7F,$3D
	.dc.b	$B1,$98,$C7,$C3,$E7,$E6,$64,$F1,$78,$4F,$E0,$BE,$2E,$4F,$F9,$53
	.dc.b	$A1,$1F,$86,$F9,$1F,$E1,$9E,$2C,$AF,$C6,$3C,$1F,$F6,$27,$D6,$FF
	.dc.b	$8F,$5F,$F6,$5A,$F9,$0F,$C2,$36,$33,$CB,$89,$F8,$B3,$99,$FF,$19
	.dc.b	$B5,$FE,$EC,$1C,$3F,$DE,$7E,$B1,$C1,$FC,$87,$E5,$A4,$FF,$B7,$27
	.dc.b	$2F,$93,$DB,$EE,$47,$FC,$82,$75,$7E,$BF,$E0,$AF,$92,$FC,$E3,$67
	.dc.b	$FD,$CE,$D7,$E0,$F2,$FF,$3D,$F8,$EF,$E2,$5E,$7D,$F3,$7F,$DC,$B5
	.dc.b	$FE,$F3,$9C,$BF,$3F,$E4,$3F,$9C,$4C,$8F,$DC,$3C,$6B,$F6,$FC,$B2
	.dc.b	$7E,$B3,$FB,$AE,$7F,$DD,$F3,$FF,$48,$E7,$34,$95,$FC,$AB,$C2,$FE
	.dc.b	$8D,$E1,$48,$FE,$0D,$1F,$C1,$3E,$32,$47,$E0,$DF,$E0,$FE,$B7,$F4
	.dc.b	$DF,$02,$4F,$F0,$D7,$6F,$F2,$9C,$5F,$33,$FC,$38,$91,$F8,$DF,$85
	.dc.b	$AF,$F2,$FF,$2F,$F4,$7E,$27,$FA,$09,$BE,$AF,$F9,$C7,$0E,$75,$FE
	.dc.b	$C9,$AF,$DE,$7D,$5F,$F7,$18,$FE,$76,$1F,$EB,$ED,$F5,$FF,$8A,$46
	.dc.b	$E7,$8C,$7E,$D2,$3F,$BF,$CF,$E3,$9A,$FF,$91,$5A,$FF,$78,$AF,$F1
	.dc.b	$ED,$79,$FD,$2F,$EE,$07,$F2,$CF,$6F,$FB,$97,$D3,$4B,$FE,$64,$FF
	.dc.b	$7E,$F0,$1F,$41,$FD,$DF,$D4,$FE,$8E,$4A,$FE,$BE,$FE,$53,$B3,$FA
	.dc.b	$E7,$2F,$F7,$DF,$4E,$57,$E2,$FB,$3F,$93,$B7,$CD,$FF,$01,$30,$C1
	.dc.b	$FE,$D0,$FA,$1F,$E8,$0F,$F6,$F6,$BF,$E6,$3F,$65,$18,$8F,$F5,$13
	.dc.b	$5F,$F0,$CE,$9F,$BF,$7F,$B4,$43,$E4,$7F,$C3,$6D,$8F,$F6,$FC,$EF
	.dc.b	$87,$FD,$7F,$2F,$9C,$D9,$2B,$7C,$FC,$BC,$5F,$F1,$1D,$7E,$BE,$7C
	.dc.b	$CD,$7C,$B1,$FE,$28,$C1,$E3,$78,$BC,$D9,$5E,$28,$FA,$5F,$0F,$E6
	.dc.b	$F3,$E4,$F9,$A3,$D3,$F2,$FE,$02,$47,$FA,$DB,$24,$7D,$5F,$C0,$FF
	.dc.b	$98,$CF,$F8,$33,$95,$18,$DF,$FD,$D1,$18,$F1,$FF,$05,$F0,$DF,$1B
	.dc.b	$CA,$98,$7D,$57,$F0,$DF,$19,$F2,$BE,$37,$E0,$92,$7C,$4F,$BF,$C3
	.dc.b	$F9,$FE,$5C,$9F,$9F,$1F,$E4,$CF,$D4,$3F,$C8,$3D,$3F,$73,$3F,$AA
	.dc.b	$67,$C5,$FA,$FE,$67,$CE,$26,$1E,$F7,$E8,$FC,$3F,$DC,$24,$7E,$CF
	.dc.b	$1F,$F7,$36,$3F,$E1,$F8,$FF,$15,$2F,$7B,$C4,$48,$E0,$7F,$98,$A3
	.dc.b	$FC,$2A,$F0,$F5,$FF,$CC,$30,$69,$23,$FD,$CF,$1F,$CF,$1F,$C8,$F9
	.dc.b	$BF,$90,$3E,$07,$5F,$C1,$F0,$7F,$27,$E7,$4A,$F0,$EE,$C3,$FC,$25
	.dc.b	$1F,$9A,$9F,$C7,$91,$FC,$03,$FB,$3F,$D2,$9F,$3D,$F9,$EF,$A1,$95
	.dc.b	$FC,$F6,$72,$FE,$BE,$32,$FE,$99,$FE,$8D,$1A,$8F,$8F,$E8,$FE,$A3
	.dc.b	$D1,$91,$E6,$C7,$FC,$25,$C1,$FB,$29,$CC,$FE,$C1,$FB,$D5,$F7,$FC
	.dc.b	$C9,$3C,$5F,$C1,$79,$D8,$8F,$DB,$FA,$3F,$B6,$31,$DD,$FB,$47,$F5
	.dc.b	$E4,$8F,$D0,$3F,$53,$CB,$FF,$72,$BF,$80,$6C,$7D,$B6,$F0,$7C,$DF
	.dc.b	$36,$EA,$9C,$F8,$6F,$95,$A6,$A3,$9E,$0F,$A6,$78,$5F,$EA,$88,$FF
	.dc.b	$01,$99,$70,$FF,$81,$A3,$79,$7F,$7C,$6C,$70,$F8,$CF,$EC,$BF,$4D
	.dc.b	$FE,$00,$6C,$67,$B9,$39,$6C,$78,$29,$EF,$FA,$E3,$0D,$7C,$4D,$44
	.dc.b	$8E,$39,$7F,$0B,$F5,$1C,$DA,$F8,$E7,$17,$F3,$09,$CE,$09,$CB,$FE
	.dc.b	$CA,$E0,$FF,$65,$DF,$1B,$E0,$D9,$D6,$DC,$AF,$8B,$FC,$FB,$0C,$5F
	.dc.b	$F3,$E8,$B5,$D6,$6C,$3E,$9B,$FA,$CE,$1F,$4A,$31,$EB,$7F,$9B,$EF
	.dc.b	$DF,$18,$9D,$57,$09,$FF,$5F,$3F,$AA,$8F,$D9,$B9,$E5,$F7,$D7,$7E
	.dc.b	$CF,$83,$18,$9D,$57,$67,$95,$FF,$AE,$5B,$3B,$FA,$3E,$35,$BB,$B7
	.dc.b	$FE,$F2,$BF,$F4,$24,$F8,$BE,$1B,$38,$FB,$FE,$1A,$D6,$AF,$B4,$62
	.dc.b	$B2,$DB,$FF,$1F,$71,$FE,$61,$BC,$DC,$BE,$13,$C3,$E7,$7F,$8D,$A3
	.dc.b	$13,$A3,$E1,$CE,$AB,$96,$3E,$DF,$84,$F9,$CF,$A5,$BA,$7D,$56,$E9
	.dc.b	$FE,$71,$C3,$E8,$FC,$B7,$AA,$7D,$6D,$8F,$A3,$F4,$3E,$EF,$67,$DE
	.dc.b	$D8,$FA,$BE,$BF,$42,$3F,$71,$EF,$EB,$7C,$63,$E3,$FC,$15,$9F,$BB
	.dc.b	$F2,$DF,$28,$F9,$7F,$39,$63,$F0,$3F,$2D,$E8,$23,$F8,$87,$D7,$E0
	.dc.b	$7D,$2E,$3F,$A6,$D9,$FC,$4F,$CB,$7D,$6F,$63,$EB,$B6,$3F,$1A,$7C
	.dc.b	$42,$1F,$94,$3E,$4F,$D2,$3D,$6F,$C0,$B6,$3F,$A3,$AF,$98,$CC,$EB
	.dc.b	$E5,$E5,$6B,$E2,$6F,$43,$C0,$FC,$DE,$2F,$96,$EE,$4B,$FD,$54,$65
	.dc.b	$BF,$2B,$E2,$F9,$6F,$1B,$63,$C7,$ED,$F9,$BE,$5B,$CE,$D8,$F3,$FB
	.dc.b	$5E,$C2,$3D,$9F,$7F,$C8,$7B,$8F,$7F,$BB,$67,$DF,$F1,$DF,$0B,$DD
	.dc.b	$F8,$6D,$8F,$87,$F1,$9F,$20,$8F,$A9,$F9,$7E,$D7,$E7,$BE,$8B,$FC
	.dc.b	$7D,$3F,$2A,$31,$F1,$7F,$3B,$E3,$75,$F1,$76,$BF,$D4,$D6,$8F,$A7
	.dc.b	$EE,$FD,$62,$4F,$C8,$7E,$CF,$BB,$F6,$0F,$B3,$EF,$59,$FF,$07,$F4
	.dc.b	$3E,$93,$FD,$DF,$1F,$73,$DA,$FE,$07,$39,$F1,$BF,$03,$18,$A7,$8F
	.dc.b	$FB,$0A,$E4,$D8,$CD,$FF,$B2,$9B,$3F,$ED,$8F,$39,$D7,$FD,$0A,$BC
	.dc.b	$75,$AE,$DF,$C2,$DF,$7C,$3F,$88,$FC,$47,$C8,$8F,$C5,$5F,$F2,$C8
	.dc.b	$7C,$C6,$B8,$F6,$D9,$FF,$0E,$CF,$1C,$3F,$95,$BF,$E7,$58,$CD,$76
	.dc.b	$F8,$E9,$1C,$3E,$3A,$E3,$E0,$1A,$B0,$4F,$98,$F3,$F1,$DB,$F0,$FC
	.dc.b	$77,$A8,$F5,$FC,$6D,$8F,$2B,$87,$CF,$F1,$B7,$2F,$06,$5D,$E3,$E1
	.dc.b	$9F,$13,$2B,$5E,$EF,$8D,$FC,$77,$C3,$7C,$86,$C7,$C2,$F6,$30,$D3
	.dc.b	$81,$F8,$4F,$A1,$79,$3F,$75,$B1,$FE,$0E,$CD,$18,$7F,$C2,$2D,$7F
	.dc.b	$BE,$9F,$DF,$97,$95,$F8,$C7,$87,$1F,$F7,$A9,$E7,$39,$BD,$79,$1F
	.dc.b	$92,$FA,$2F,$F0,$1E,$3E,$58,$C3,$CF,$C9,$F3,$11,$F3,$64,$8F,$CB
	.dc.b	$7D,$EF,$FA,$8A,$7D,$D1,$F9,$E8,$C6,$5F,$EE,$8A,$E1,$5F,$88,$F9
	.dc.b	$0F,$C8,$9F,$92,$F8,$FB,$1F,$8C,$70,$FE,$3B,$E4,$2D,$FF,$BC,$BE
	.dc.b	$8B,$F2,$13,$F1,$1F,$C8,$46,$7A,$FF,$5F,$18,$E0,$FC,$AF,$C8,$E2
	.dc.b	$7F,$42,$DB,$FD,$CD,$7F,$89,$1F,$DD,$FD,$1B,$D1,$E7,$F1,$5F,$BA
	.dc.b	$8C,$4F,$0E,$7F,$DC,$5F,$F9,$3E,$3E,$E3,$73,$FF,$25,$B5,$FB,$63
	.dc.b	$EF,$7F,$A1,$3E,$DA,$FE,$0F,$18,$D3,$EE,$57,$37,$EF,$7D,$23,$D3
	.dc.b	$FE,$1A,$FF,$A8,$F1,$FE,$26,$E5,$BE,$A7,$43,$73,$63,$6F,$F1,$6F
	.dc.b	$A3,$FF,$10,$E7,$FE,$4C,$FA,$4F,$F2,$0B,$FB,$71,$1A,$51,$AC,$7F
	.dc.b	$81,$FC,$AF,$3F,$8E,$FE,$1E,$DC,$7F,$84,$F0,$7F,$83,$3C,$0F,$A1
	.dc.b	$8F,$F1,$B6,$BF,$A0,$F0,$BF,$82,$4E,$F8,$7F,$8C,$BF,$C6,$9B,$7F
	.dc.b	$C5,$3E,$3F,$F8,$E9,$FE,$2E,$FA,$4F,$5D,$FF,$0D,$8C,$67,$2B,$F8
	.dc.b	$DC,$DC,$7F,$82,$DB,$E4,$3F,$C4,$2D,$AF,$F2,$A3,$67,$F9,$93,$5F
	.dc.b	$43,$FE,$46,$8F,$F2,$BB,$BD,$7F,$CA,$2E,$67,$F9,$0B,$99,$DF,$C7
	.dc.b	$F8,$08,$F8,$1F,$1F,$C7,$FB,$5F,$73,$63,$FD,$A3,$2F,$8B,$F8,$5F
	.dc.b	$1D,$F0,$DB,$3A,$70,$F0,$F8,$DB,$8F,$C2,$F8,$DF,$82,$49,$F8,$2F
	.dc.b	$C4,$F8,$DF,$9C,$FC,$F6,$C7,$8B,$D7,$FA,$5F,$1B,$F4,$DB,$1E,$6F
	.dc.b	$57,$E6,$91,$F9,$AF,$D0,$F8,$CF,$B4,$FD,$B6,$C7,$BB,$E3,$3E,$F7
	.dc.b	$C6,$FD,$F6,$C7,$C2,$F8,$DF,$B2,$47,$EC,$BF,$73,$E3,$7F,$0C,$7E
	.dc.b	$1F,$E3,$2C,$FC,$A7,$0F,$CC,$F8,$DB,$BF,$E8,$CF,$F5,$60,$FA,$47
	.dc.b	$C3,$7E,$FB,$E0,$7F,$D5,$B1,$89,$FE,$03,$F2,$5F,$EB,$88,$C4,$7F
	.dc.b	$AE,$A7,$75,$FE,$D2,$AF,$F5,$EF,$76,$AB,$FD,$BD,$8E,$2B,$50,$9D
	.dc.b	$5F,$FA,$F1,$E1,$1F,$EB,$D3,$CD,$3E,$5A,$F1,$3C,$B0,$E3,$18,$9F
	.dc.b	$39,$D6,$78,$34,$CD,$70,$70,$7F,$58,$C3,$CC,$FF,$5A,$C3,$C9,$5E
	.dc.b	$8A,$E0,$79,$AB,$FD,$7A,$FB,$3B,$7D,$97,$E2,$6B,$D4,$FF,$04,$C7
	.dc.b	$F6,$6F,$96,$7A,$3F,$E6,$59,$C9,$FE,$E6,$8C,$F7,$76,$8C,$4E,$EF
	.dc.b	$AD,$6A,$78,$87,$17,$F6,$17,$A1,$FD,$DA,$30,$DC,$BC,$98,$78,$E7
	.dc.b	$83,$63,$D1,$3B,$BF,$33,$5F,$04,$ED,$F3,$E3,$F5,$5F,$24,$FF,$81
	.dc.b	$B5,$F0,$C6,$71,$F5,$47,$82,$BE,$9C,$F0,$2C,$FA,$9B,$7D,$DF,$0C
	.dc.b	$F3,$BE,$D8,$7A,$27,$B5,$7C,$FC,$3F,$3D,$93,$E9,$3F,$D3,$94,$3F
	.dc.b	$C6,$2E,$1B,$F2,$3F,$E1,$EC,$BF,$92,$EF,$4F,$F7,$FD,$7F,$AA,$6B
	.dc.b	$94,$FF,$BE,$A3,$11,$D3,$E4,$78,$3F,$E4,$C8,$ED,$3B,$F6,$BF,$C6
	.dc.b	$18,$F9,$5B,$3F,$E3,$6A,$FE,$EA,$71,$8C,$FC,$8F,$4A,$F6,$7B,$96
	.dc.b	$FF,$C6,$77,$E4,$CB,$CF,$79,$F5,$35,$ED,$CA,$D1,$E5,$8C,$3F,$E7
	.dc.b	$BB,$4F,$90,$91,$E4,$C7,$E6,$7A,$3E,$1F,$6E,$4F,$49,$CB,$6F,$FC
	.dc.b	$0C,$DF,$90,$F4,$1E,$9F,$4E,$C7,$59,$4F,$C4,$24,$67,$E3,$FE,$01
	.dc.b	$1E,$DF,$23,$5F,$BB,$F2,$3F,$14,$C7,$17,$D7,$F9,$1F,$91,$D1,$2B
	.dc.b	$CD,$15,$F8,$51,$F5,$08,$F9,$7E,$47,$E8,$7C,$84,$8F,$74,$57,$E5
	.dc.b	$47,$DC,$23,$ED,$F9,$0F,$B0,$47,$E3,$7C,$87,$DB,$F0,$64,$7C,$97
	.dc.b	$4F,$E6,$0F,$C1,$7C,$E7,$FA,$99,$FA,$0D,$74,$63,$E3,$FF,$1F,$19
	.dc.b	$8F,$F0,$E3,$5F,$D2,$71,$A3,$11,$FE,$A4,$5F,$39,$F9,$C8,$7D,$23
	.dc.b	$83,$0D,$8F,$1C,$E2,$73,$C5,$27,$FD,$AF,$5B,$7F,$C7,$9D,$BF,$E5
	.dc.b	$F7,$FD,$B7,$F4,$8F,$63,$C0,$D8,$FF,$6F,$CE,$DF,$F2,$24,$65,$BE
	.dc.b	$C7,$03,$8D,$7A,$7E,$8D,$E9,$FA,$FE,$93,$FB,$0F,$47,$FA,$65,$BE
	.dc.b	$17,$FA,$82,$B7,$8F,$FB,$BA,$7F,$AD,$BF,$C2,$DA,$9A,$4E,$7F,$07
	.dc.b	$A9,$27,$94,$7D,$AF,$88,$F9,$9E,$83,$E3,$E9,$9A,$FF,$7E,$C6,$5B
	.dc.b	$F1,$FC,$0D,$BF,$82,$F9,$2F,$C2,$91,$FC,$2F,$CA,$7E,$23,$E5,$24
	.dc.b	$7E,$21,$D3,$FE,$6D,$FA,$37,$91,$F7,$FE,$91,$F0,$34,$FC,$14,$EA
	.dc.b	$BF,$85,$89,$DB,$FE,$CD,$6B,$FC,$71,$AF,$FA,$49,$AF,$FB,$39,$7C
	.dc.b	$07,$6F,$2A,$4F,$69,$CB,$7D,$FF,$37,$3C,$8C,$1B,$7F,$DB,$CF,$95
	.dc.b	$B3,$F7,$23,$C9,$1F,$A6,$F9,$6F,$D6,$FC,$B4,$9F,$D6,$C5,$5D,$2E
	.dc.b	$9F,$E9,$9F,$37,$FA,$3E,$EF,$E8,$2D,$AF,$DA,$E1,$FB,$8E,$69,$1F
	.dc.b	$B8,$7F,$77,$F1,$5F,$BC,$E6,$93,$FB,$C7,$6F,$83,$1F,$DB,$F8,$8F
	.dc.b	$97,$FE,$06,$49,$E0,$7F,$04,$F6,$BF,$83,$91,$FE,$0F,$ED,$7F,$0A
	.dc.b	$F9,$49,$1F,$C2,$9D,$3D,$CF,$E0,$5F,$19,$FC,$40,$8F,$F1,$0F,$94
	.dc.b	$AE,$FC,$BF,$E1,$5F,$1D,$FC,$58,$CF,$1F,$C7,$9F,$E1,$9F,$1D,$FC
	.dc.b	$2F,$E8,$FF,$B2,$EF,$FD,$D6,$51,$87,$67,$E3,$9F,$EF,$0F,$FB,$69
	.dc.b	$AF,$16,$BF,$DF,$DA,$F4,$F7,$BF,$94,$FC,$14,$8E,$91,$FD,$B7,$B1
	.dc.b	$F1,$F4,$93,$FC,$BB,$4D,$7C,$7F,$F7,$AA,$FF,$5B,$4F,$F6,$56,$BF
	.dc.b	$EE,$E9,$FF,$73,$4E,$FE,$77,$F0,$BE,$93,$DA,$A7,$86,$3C,$5E,$7E
	.dc.b	$BC,$C9,$8F,$F6,$DC,$BF,$8F,$70,$FC,$BC,$FF,$DE,$5E,$87,$8C,$95
	.dc.b	$E3,$7C,$DE,$2F,$B7,$C7,$93,$CE,$3E,$07,$C7,$48,$EC,$FC,$3F,$8F
	.dc.b	$F8,$9E,$47,$BC,$93,$EF,$7E,$17,$A1,$F2,$7C,$89,$3E,$78,$F9,$9F
	.dc.b	$1D,$F3,$7E,$3E,$4F,$AE,$3E,$87,$C7,$FC,$B2,$47,$CB,$7E,$77,$C7
	.dc.b	$7D,$52,$47,$C3,$F8,$FF,$B1,$F1,$D2,$3E,$33,$F6,$FD,$5F,$B9,$F1
	.dc.b	$FF,$5C,$93,$F5,$DF,$F1,$9F,$D0,$FE,$3F,$91,$CD,$F9,$7A,$FC,$E3
	.dc.b	$79,$BE,$85,$6C,$FF,$1C,$4F,$E8,$5F,$F2,$1C,$6B,$BB,$FC,$71,$1A
	.dc.b	$8E,$3F,$19,$E0,$25,$7F,$8B,$67,$FC,$76,$FE,$25,$FE,$BD,$3E,$0F
	.dc.b	$8B,$E6,$74,$F8,$CE,$BF,$17,$2F,$FC,$65,$68,$F0,$FC,$67,$8B,$D9
	.dc.b	$F7,$92,$BC,$0F,$C2,$F8,$BF,$3E,$B2,$79,$C7,$AB,$E2,$FD,$7E,$24
	.dc.b	$9E,$F1,$5F,$44,$7C,$C2,$3E,$DF,$84,$FE,$4B,$F4,$4F,$16,$9F,$1A
	.dc.b	$2C,$FF,$95,$BD,$95,$6E,$30,$D7,$E8,$4F,$CE,$F1,$FE,$7F,$C6,$49
	.dc.b	$F8,$71,$5F,$AD,$17,$8F,$ED,$CF,$D6,$F8,$C9,$5F,$29,$FB,$3E,$46
	.dc.b	$23,$EF,$7C,$5F,$DC,$31,$7F,$84,$7F,$16,$48,$FB,$8F,$E3,$BE,$33
	.dc.b	$FD,$A9,$EA,$7F,$8C,$39,$9F,$19,$8B,$76,$27,$31,$FE,$99,$F9,$15
	.dc.b	$F1,$C3,$85,$EF,$0B,$87,$0C,$18,$70,$7A,$C6,$63,$15,$DE,$FC,$54
	.dc.b	$FA,$17,$87,$E4,$8F,$F4,$AC,$62,$BF,$D2,$F1,$B9,$D3,$FC,$EC,$C4
	.dc.b	$6B,$E4,$BF,$AC,$7D,$15,$F7,$4D,$25,$76,$F1,$3D,$4F,$F4,$5C,$51
	.dc.b	$88,$FF,$5E,$46,$35,$F1,$B6,$3F,$99,$7E,$50,$F8,$3F,$EA,$5B,$7C
	.dc.b	$08,$C7,$B3,$F3,$78,$7F,$CC,$71,$FE,$33,$8C,$5F,$D4,$8D,$BF,$32
	.dc.b	$77,$F0,$7D,$1E,$38,$59,$E1,$37,$75,$F2,$C7,$E8,$F1,$7F,$B5,$72
	.dc.b	$FE,$05,$87,$F1,$13,$F0,$C5,$3E,$C9,$6F,$9F,$C1,$7F,$6D,$E5,$FE
	.dc.b	$2B,$36,$BD,$73,$63,$8C,$7F,$AD,$F1,$4E,$6F,$9C,$E7,$E0,$31,$5A
	.dc.b	$8E,$1F,$03,$BF,$CD,$5F,$25,$A8,$C3,$67,$67,$7F,$82,$C6,$7F,$E0
	.dc.b	$7C,$1F,$05,$CF,$E4,$BB,$B6,$FC,$CE,$1F,$E0,$3D,$56,$FF,$C2,$19
	.dc.b	$BE,$3B,$E2,$36,$BF,$DB,$0D,$9F,$EE,$8D,$7F,$C2,$1B,$7D,$4F,$84
	.dc.b	$F6,$5F,$C0,$6D,$FC,$79,$F8,$1F,$1B,$F1,$1B,$1F,$EE,$1F,$CD,$3F
	.dc.b	$C1,$2F,$FB,$9D,$B8,$7F,$80,$DA,$F8,$1A,$FF,$AC,$97,$C1,$FD,$36
	.dc.b	$CF,$5F,$83,$C4,$E7,$E0,$FE,$97,$C7,$DA,$FF,$79,$6A,$8F,$AD,$F1
	.dc.b	$FF,$10,$93,$F1,$1F,$8F,$F1,$BF,$6C,$FC,$37,$9A,$FC,$D5,$EC,$E6
	.dc.b	$FA,$F8,$1F,$81,$37,$2F,$63,$13,$E2,$F8,$DF,$A6,$47,$E9,$BF,$99
	.dc.b	$E0,$FF,$59,$C7,$A3,$C2,$F8,$1F,$07,$F8,$D7,$AF,$AF,$FA,$1D,$3C
	.dc.b	$8D,$CB,$FB,$BD,$FF,$0F,$E4,$56,$A3,$F0,$3E,$0F,$E1,$08,$E0,$FC
	.dc.b	$2F,$95,$F9,$B3,$E8,$7B,$F6,$3E,$5F,$87,$F9,$F3,$EA,$79,$56,$3E
	.dc.b	$7F,$89,$F9,$02,$3F,$90,$7F,$27,$F1,$9F,$A7,$3E,$E4,$51,$9F,$F9
	.dc.b	$A7,$F3,$47,$CB,$FE,$B7,$C6,$FD,$A4,$62,$BE,$EC,$7E,$C7,$C4,$FC
	.dc.b	$F9,$23,$F3,$EF,$E8,$BC,$BF,$DB,$1F,$73,$E1,$5E,$1E,$2B,$F0,$FE
	.dc.b	$27,$EE,$CD,$77,$58,$FC,$5F,$C6,$FE,$B0,$8F,$EB,$1F,$D7,$F8,$BF
	.dc.b	$C0,$CF,$3F,$C0,$59,$FF,$62,$E0,$E4,$FE,$25,$C0,$FA,$2F,$E7,$75
	.dc.b	$A9,$CC,$62,$7F,$D2,$43,$E8,$DF,$11,$FC,$8E,$30,$D8,$CD,$FF,$54
	.dc.b	$6C,$FE,$15,$7C,$47,$89,$CF,$A1,$FE,$3F,$73,$5F,$ED,$C7,$2F,$D6
	.dc.b	$D1,$C9,$FE,$E6,$8F,$B3,$5F,$E7,$3F,$CE,$BF,$7E,$FF,$9D,$67,$58
	.dc.b	$FF,$A9,$67,$35,$CE,$73,$5B,$F8,$4F,$BF,$E0,$7E,$01,$FE,$69,$7F
	.dc.b	$C0,$B0,$E5,$E6,$7E,$13,$E0,$F2,$DC,$BC,$67,$58,$FC,$4F,$C1,$E6
	.dc.b	$BC,$5F,$07,$D0,$FC,$6F,$C0,$5B,$E3,$3F,$E1,$2D,$7F,$D2,$D1,$8D
	.dc.b	$BE,$0D,$7E,$53,$E0,$78,$CE,$69,$F0,$19,$8C,$37,$0F,$B7,$F4,$6F
	.dc.b	$88,$FC,$1F,$D5,$FE,$16,$FC,$34,$AD,$46,$27,$53,$CB,$E5,$7F,$05
	.dc.b	$46,$92,$BF,$82,$69,$F6,$3E,$13,$E5,$FB,$1F,$AC,$6C,$FA,$7E,$13
	.dc.b	$0D,$3D,$91,$F9,$8F,$A4,$FD,$96,$5F,$6F,$DA,$C7,$91,$F5,$9F,$CE
	.dc.b	$E8,$E5,$F2,$93,$1F,$F1,$3C,$65,$BC,$BC,$27,$8B,$63,$E8,$73,$BE
	.dc.b	$4F,$F0,$67,$F0,$4C,$AC,$FD,$17,$0F,$D7,$F4,$BF,$BA,$B9,$3E,$6F
	.dc.b	$D1,$FF,$86,$BB,$7F,$D3,$DB,$7F,$39,$9C,$FB,$1E,$FF,$63,$F6,$F6
	.dc.b	$F6,$3F,$60,$DA,$FF,$20,$B7,$2F,$E6,$4D,$7F,$B6,$E5,$FD,$C7,$2F
	.dc.b	$D3,$7C,$A7,$F8,$3E,$31,$7F,$AD,$D5,$3F,$C1,$BC,$BF,$81,$3F,$E4
	.dc.b	$E6,$E1,$FC,$FD,$AF,$F9,$3D,$AF,$F2,$C0,$F4,$8F,$C6,$BC,$B0,$FF
	.dc.b	$5F,$39,$9F,$E1,$EE,$67,$97,$D1,$FF,$8D,$65,$5F,$EF,$1F,$C4,$FF
	.dc.b	$CC,$B5,$F9,$8A,$D4,$E7,$E0,$7F,$8F,$FE,$6B,$F4,$B0,$FE,$AC,$D7
	.dc.b	$FC,$CC,$D7,$D0,$D7,$FA,$2F,$E1,$EF,$92,$FE,$07,$F2,$FF,$CA,$A3
	.dc.b	$19,$7E,$1A,$31,$F2,$7F,$C9,$64,$D2,$4F,$E9,$37,$7C,$C7,$F9,$2D
	.dc.b	$B1,$98,$C1,$FE,$4F,$E1,$FE,$93,$1F,$94,$8F,$B5,$E8,$FF,$16,$7F
	.dc.b	$8B,$FC,$C4,$9F,$C9,$7A,$B7,$5F,$93,$1F,$CA,$F0,$4F,$E0,$CF,$C2
	.dc.b	$74,$59,$FD,$0F,$C1,$79,$48,$FA,$5F,$27,$47,$F2,$67,$F1,$8D,$8F
	.dc.b	$D4,$FC,$C7,$B4,$8F,$C1,$7D,$9F,$03,$FC,$B9,$FC,$A3,$63,$F6,$3E
	.dc.b	$CF,$F3,$23,$F3,$1F,$01,$63,$F5,$CD,$D3,$DA,$7F,$AE,$B4,$46,$79
	.dc.b	$FF,$A3,$7A,$DF,$3C,$97,$F4,$DF,$9D,$EF,$7F,$40,$7F,$4C,$D8,$FF
	.dc.b	$72,$7E,$0B,$FE,$73,$C7,$BF,$C1,$FC,$AD,$7F,$C7,$33,$9C,$57,$FE
	.dc.b	$0E,$AD,$3C,$BE,$0F,$FC,$59,$46,$92,$3F,$C5,$92,$9F,$5C,$71,$3C
	.dc.b	$79,$F8,$2B,$FC,$93,$F0,$FF,$DE,$70,$F0,$E5,$FE,$34,$FC,$53,$F9
	.dc.b	$CC,$63,$25,$6F,$1F,$35,$69,$F4,$43,$E6,$BF,$C8,$4D,$8F,$2B,$70
	.dc.b	$FF,$68,$5F,$4A,$DF,$CD,$79,$0F,$E2,$1F,$09,$70,$FE,$4D,$F0,$98
	.dc.b	$69,$FC,$A6,$3F,$C2,$F8,$7F,$BC,$47,$D2,$BD,$AF,$F6,$F7,$CA,$FB
	.dc.b	$A3,$13,$FE,$E1,$F9,$4F,$F1,$EC,$62,$3F,$C7,$F3,$BA,$FF,$2C,$57
	.dc.b	$F9,$0B,$E1,$F5,$4A,$F9,$31,$88,$FF,$7D,$87,$71,$AA,$E7,$3E,$67
	.dc.b	$CB,$8F,$89,$CE,$1C,$23,$13,$E7,$3A,$4F,$06,$99,$BE,$0E,$0F,$E5
	.dc.b	$D8,$7A,$67,$84,$F9,$E3,$13,$FC,$CF,$0F,$33,$DF,$D8,$F5,$36,$BD
	.dc.b	$31,$88,$EA,$1C,$7C,$BF,$EB,$38,$BF,$C6,$95,$FC,$FE,$31,$5B,$8C
	.dc.b	$72,$7F,$38,$8C,$1F,$E7,$09,$DD,$F4,$9D,$57,$D7,$97,$C5,$7F,$4B
	.dc.b	$8C,$37,$3E,$B8,$FB,$B1,$5F,$8A,$F7,$19,$8E,$6D,$C3,$BD,$F2,$8F
	.dc.b	$B0,$79,$70,$FE,$B5,$AF,$28,$C9,$DA,$BD,$91,$E7,$79,$C7,$13,$C1
	.dc.b	$3E,$77,$E0,$46,$9E,$58,$FE,$09,$F4,$CF,$82,$FE,$63,$63,$8C,$FE
	.dc.b	$23,$27,$57,$D5,$9F,$F2,$37,$E7,$4F,$9D,$FE,$E5,$7F,$C0,$AD,$39
	.dc.b	$AF,$F0,$B4,$6B,$83,$F8,$B7,$E5,$CF,$9C,$C5,$7F,$AF,$23,$0F,$F9
	.dc.b	$5B,$9D,$18,$8F,$E6,$37,$DB,$2E,$35,$AF,$84,$FF,$63,$57,$36,$C6
	.dc.b	$63,$F8,$FD,$A7,$2F,$EE,$73,$FE,$31,$D5,$FC,$4E,$BC,$4D,$7F,$99
	.dc.b	$CE,$F3,$FE,$61,$F8,$33,$C4,$F1,$AF,$0A,$98,$8B,$7F,$CC,$27,$4D
	.dc.b	$9F,$3C,$62,$3F,$8F,$AF,$03,$DF,$E3,$67,$3F,$E5,$CD,$C7,$C7,$3F
	.dc.b	$33,$7F,$A7,$F0,$D7,$CB,$FF,$52,$E5,$FE,$50,$F8,$BD,$72,$2B,$51
	.dc.b	$F2,$AD,$7C,$2F,$FC,$60,$7E,$DD,$C3,$FE,$33,$8D,$CF,$1B,$7B,$DE
	.dc.b	$1F,$8D,$97,$F8,$A8,$AF,$FA,$A6,$2E,$BF,$8B,$AE,$33,$F9,$0F,$63
	.dc.b	$D5,$EA,$C8,$C1,$FE,$6A,$7D,$19,$9A,$48,$FF,$43,$D7,$F8,$91,$FF
	.dc.b	$38,$FC,$87,$C5,$3E,$1F,$A1,$63,$FC,$55,$1D,$08,$F2,$DF,$E4,$91
	.dc.b	$FD,$2F,$27,$CF,$F1,$E4,$7F,$9F,$9D,$3F,$9A,$C3,$FD,$BE,$FF,$A0
	.dc.b	$BC,$7F,$AC,$7D,$4F,$32,$C7,$F8,$EE,$3D,$A4,$7F,$3F,$E3,$FF,$9B
	.dc.b	$1B,$B7,$E8,$94,$65,$FB,$26,$23,$50,$F5,$3F,$0D,$CD,$2B,$FD,$6F
	.dc.b	$39,$7F,$D7,$3F,$9B,$7F,$82,$DF,$D7,$B5,$EA,$D7,$F0,$ED,$7F,$CF
	.dc.b	$0F,$F8,$A1,$FF,$68,$63,$FE,$BD,$6F,$C3,$7F,$04,$53,$B7,$FC,$62
	.dc.b	$3E,$97,$FC,$C2,$FF,$B3,$17,$C8,$F1,$6C,$67,$D7,$C4,$FC,$59,$CC
	.dc.b	$FC,$46,$BF,$9E,$6A,$3F,$AB,$6D,$FA,$3E,$6F,$E4,$1E,$B3,$E1,$7F
	.dc.b	$60,$DA,$FE,$4F,$D7,$EB,$E6,$7B,$1F,$55,$B3,$FC,$61,$AF,$D6,$6B
	.dc.b	$FB,$05,$F2,$DF,$17,$F1,$6F,$E2,$13,$FE,$EC,$6B,$F1,$A7,$6F,$F3
	.dc.b	$8E,$BF,$DF,$91,$FE,$77,$1E,$8F,$8C,$91,$EF,$D3,$DB,$1B,$8F,$A3
	.dc.b	$F4,$BF,$CF,$5B,$F2,$1F,$5E,$0D,$F2,$F0,$FE,$4E,$33,$3F,$CD,$9A
	.dc.b	$FE,$51,$AF,$F0,$EF,$C5,$BF,$8E,$3F,$E0,$5D,$FF,$0C,$FF,$42,$6E
	.dc.b	$3F,$E5,$28,$4F,$0F,$13,$FA,$9F,$57,$F1,$EE,$D9,$5F,$D6,$A2,$BC
	.dc.b	$FE,$67,$EC,$98,$C3,$A6,$5F,$32,$7F,$90,$B5,$FF,$69,$35,$FC,$33
	.dc.b	$53,$FC,$E5,$F9,$95,$FE,$D9,$FB,$6F,$BF,$5E,$96,$C7,$4F,$0B,$F9
	.dc.b	$D7,$E6,$AE,$0B,$DB,$31,$E6,$7F,$8A,$67,$4E,$35,$43,$4E,$64,$9D
	.dc.b	$47,$CA,$C7,$F3,$D4,$DD,$F1,$3F,$9C,$25,$F0,$C3,$CB,$F0,$92,$39
	.dc.b	$BE,$8E,$EF,$D1,$B2,$3C,$97,$FE,$DF,$76,$78,$BE,$57,$DE,$47,$CB
	.dc.b	$F2,$D5,$FC,$05,$FE,$B8,$C7,$0F,$D6,$E3,$FC,$96,$B2,$BE,$29,$F4
	.dc.b	$A3,$0F,$E3,$E3,$F2,$84,$7F,$C1,$3C,$9F,$D2,$5E,$3F,$9D,$7F,$88
	.dc.b	$5F,$E2,$2D,$BF,$C8,$4E,$E7,$F2,$9E,$0F,$F0,$22,$3F,$3B,$E4,$3F
	.dc.b	$82,$E9,$FD,$B7,$53,$49,$3F,$CD,$B8,$7E,$96,$DF,$C3,$09,$1F,$E2
	.dc.b	$5F,$D4,$DF,$23,$FC,$43,$EA,$24,$FC,$E7,$FC,$4B,$3F,$E1,$F8,$FA
	.dc.b	$39,$78,$48,$F7,$C9,$1F,$E0,$1E,$57,$CC,$7F,$1C,$C6,$5F,$E2,$23
	.dc.b	$F8,$FF,$B5,$23,$EE,$BF,$98,$F9,$6F,$E4,$7E,$D7,$BC,$93,$DD,$F8
	.dc.b	$5E,$F7,$F2,$82,$47,$E6,$7F,$34,$7D,$27,$F2,$CC,$64,$FE,$29,$FF
	.dc.b	$0F,$4B,$E8,$BF,$98,$FD,$47,$CD,$25,$7B,$5F,$F2,$1F,$E3,$AF,$90
	.dc.b	$A4,$9F,$F7,$1C,$65,$FA,$56,$F7,$7F,$9E,$FC,$DC,$9F,$34,$7F,$40
	.dc.b	$DB,$EE,$12,$3E,$63,$F7,$FE,$2B,$FA,$29,$23,$FC,$BD,$87,$F5,$DE
	.dc.b	$CF,$E9,$71,$FD,$28,$86,$BE,$43,$FC,$79,$F4,$DF,$12,$3F,$0C,$D7
	.dc.b	$FC,$6B,$D6,$8C,$47,$F8,$D5,$AF,$EF,$B4,$FE,$F1,$7F,$E7,$99,$7D
	.dc.b	$37,$63,$E6,$FC,$8F,$CE,$FA,$39,$87,$D5,$7E,$8E,$5F,$4B,$E4,$7F
	.dc.b	$C9,$19,$1A,$4A,$F9,$7E,$E7,$AB,$92,$47,$DF,$8F,$B7,$F2,$1E,$EF
	.dc.b	$5E,$4F,$E0,$E2,$BE,$ED,$7F,$B2,$BF,$C6,$1A,$FF,$B2,$DB,$BB,$C1
	.dc.b	$FE,$B1,$C1,$FE,$97,$6F,$8F,$FC,$81,$B6,$F8,$B7,$C7,$7F,$99,$63
	.dc.b	$F9,$64,$EB,$0F,$9D,$EC,$FD,$43,$FC,$E1,$19,$38,$4F,$FA,$61,$B8
	.dc.b	$F8,$7E,$23,$7F,$3F,$C1,$FB,$BD,$52,$FB,$C7,$4F,$83,$91,$E2,$7B
	.dc.b	$FC,$F6,$23,$FD,$4B,$BB,$E5,$3F,$0E,$67,$2F,$BC,$FF,$9F,$63,$53
	.dc.b	$F8,$A8,$C1,$FE,$81,$CF,$FA,$14,$7F,$88,$7E,$53,$EE,$3F,$EF,$78
	.dc.b	$FF,$35,$28,$CE,$DF,$3F,$E2,$7F,$C5,$13,$9E,$7F,$9D,$E5,$E2,$F7
	.dc.b	$87,$26,$DE,$6B,$EA,$36,$FC,$17,$57,$D4,$7F,$9F,$F1,$FF,$60,$B6
	.dc.b	$3F,$CB,$ED,$7F,$D3,$2D,$7C,$BF,$2A,$EE,$7D,$4F,$FB,$2A,$30,$D8
	.dc.b	$CC,$FF,$9A,$A7,$2F,$6F,$81,$78,$7E,$26,$C6,$63,$07,$EE,$3E,$51
	.dc.b	$D7,$F3,$76,$FC,$D7,$CC,$E2,$FE,$41,$F9,$BA,$C7,$07,$CE,$E3,$56
	.dc.b	$2F,$F4,$55,$6F,$86,$1D,$5F,$D2,$79,$19,$E0,$FE,$0D,$57,$17,$CC
	.dc.b	$FF,$8A,$EF,$F5,$8D,$7F,$58,$DA,$FE,$1C,$D9,$FD,$3E,$1F,$E1,$17
	.dc.b	$9B,$7B,$7D,$9C,$FF,$C3,$EA,$CF,$EE,$66,$E1,$FB,$96,$E1,$FC,$6F
	.dc.b	$86,$E3,$DA,$BF,$3F,$E2,$E3,$E5,$F2,$DC,$7C,$4E,$1F,$27,$8D,$69
	.dc.b	$A7,$CF,$F0,$5F,$55,$FE,$C9,$BF,$57,$8B,$E4,$9C,$FC,$3F,$8A,$31
	.dc.b	$5B,$C3,$9D,$FF,$A1,$EF,$6F,$C3,$E3,$DE,$1E,$AF,$9F,$F9,$04,$9F
	.dc.b	$E7,$6F,$CB,$F1,$B5,$1F,$2D,$B8,$76,$1C,$2F,$A8,$7C,$4F,$C6,$8F
	.dc.b	$93,$18,$BF,$A9,$18,$F8,$9F,$B5,$3F,$E7,$89,$FF,$08,$36,$FF,$A5
	.dc.b	$B5,$FF,$7F,$2F,$23,$A9,$DB,$C7,$B5,$FE,$51,$97,$8D,$E2,$F9,$6F
	.dc.b	$1B,$67,$5E,$47,$88,$8F,$89,$FC,$87,$0B,$EA,$3F,$CB,$53,$FE,$60
	.dc.b	$8C,$7C,$2F,$BB,$E4,$FE,$B5,$6B,$5F,$8D,$5C,$7E,$1F,$27,$C4,$F2
	.dc.b	$AE,$3E,$7F,$2B,$E1,$91,$F8,$6F,$C6,$F8,$AF,$97,$F2,$BF,$31,$B3
	.dc.b	$EF,$F9,$9F,$0F,$C1,$FD,$06,$C7,$C3,$F9,$5F,$9E,$47,$E7,$BF,$4B
	.dc.b	$E6,$3E,$B9,$F6,$3E,$4A,$CF,$CB,$F3,$BE,$D9,$F7,$3B,$AC,$7C,$FF
	.dc.b	$97,$FB,$64,$7E,$DB,$F7,$9F,$C1,$9F,$84,$F8,$CA,$FF,$A1,$FE,$9B
	.dc.b	$FD,$1C,$DF,$95,$FD,$8C,$63,$93,$F2,$1D,$E8,$FF,$70,$57,$1F,$91
	.dc.b	$FE,$09,$06,$92,$3F,$A1,$D7,$E9,$77,$FF,$71,$C7,$DB,$6C,$76,$AF
	.dc.b	$B5,$E3,$FE,$D6,$72,$DF,$67,$F6,$91,$FE,$35,$BF,$57,$3E,$27,$D3
	.dc.b	$F2,$3E,$42,$4F,$F9,$B7,$04,$7A,$3E,$3B,$DE,$7C,$0F,$91,$B5,$FE
	.dc.b	$8A,$FA,$77,$37,$F1,$6D,$7F,$DF,$7C,$D6,$73,$5F,$E9,$0F,$7D,$39
	.dc.b	$E7,$FE,$55,$6B,$46,$1B,$7F,$1A,$46,$E7,$8F,$D2,$BE,$2F,$A5,$7F
	.dc.b	$A7,$A3,$0F,$9E,$7E,$CF,$57,$DA,$F9,$1B,$3F,$7D,$FB,$BF,$01,$8C
	.dc.b	$7E,$83,$79,$7F,$A9,$73,$FD,$FF,$8E,$FC,$1F,$C8,$7E,$11,$B8,$FE
	.dc.b	$27,$E3,$F0,$D3,$F1,$91,$F8,$AD,$A4,$7E,$3F,$E3,$EE,$7F,$7F,$E3
	.dc.b	$BF,$1F,$1F,$90,$6C,$FE,$57,$E3,$B0,$D3,$F3,$11,$F4,$FE,$9B,$F2
	.dc.b	$11,$F2,$6D,$18,$7E,$63,$5F,$C9,$B5,$FF,$0B,$BF,$90,$8B,$DD,$D7
	.dc.b	$CB,$EC,$FE,$93,$E3,$BF,$4B,$F1,$F6,$BF,$54,$FE,$A3,$E1,$B1,$8F
	.dc.b	$E6,$DB,$C5,$FD,$F6,$3C,$BC,$3F,$8B,$E7,$FD,$5C,$7F,$3D,$33,$87
	.dc.b	$DD,$F0,$FD,$DC,$5F,$8A,$DB,$FD,$83,$3E,$F6,$D7,$BA,$7F,$1F,$F6
	.dc.b	$CF,$13,$8D,$FF,$B9,$23,$0F,$FB,$9A,$F8,$E7,$F8,$0F,$84,$FF,$58
	.dc.b	$CF,$F0,$5C,$7F,$B9,$C6,$1F,$DB,$18,$F8,$FC,$FC,$8D,$AF,$CC,$DF
	.dc.b	$38,$FD,$47,$D3,$FF,$AC,$67,$EB,$36,$FF,$40,$D7,$F8,$3B,$5F,$F2
	.dc.b	$5A,$C3,$9F,$C4,$74,$6D,$7F,$C0,$0F,$7B,$FA,$8B,$E2,$3F,$57,$D9
	.dc.b	$C7,$E4,$3F,$75,$3A,$C7,$F8,$ED,$FF,$B2,$EF,$97,$AF,$FB,$8A,$34
	.dc.b	$91,$F8,$FF,$B6,$FC,$45,$FE,$E2,$B7,$3E,$18,$FD,$DC,$E9,$4E,$9F
	.dc.b	$C0,$7D,$37,$CD,$7F,$93,$B5,$FF,$35,$B5,$FE,$0A,$D4,$FF,$AE,$AB
	.dc.b	$C1,$18,$D7,$F8,$C4,$E6,$3B,$C7,$F1,$BF,$A6,$79,$5F,$C5,$9B,$7F
	.dc.b	$AB,$6D,$7F,$15,$6B,$FC,$F1,$79,$7C,$EB,$A3,$6B,$E1,$6D,$F6,$7C
	.dc.b	$AF,$F2,$56,$DF,$D0,$31,$FF,$43,$CE,$6B,$F9,$13,$63,$FD,$BB,$2F
	.dc.b	$93,$FC,$9D,$FE,$51,$B7,$DA,$7F,$A2,$57,$DC,$F3,$FF,$33,$F1,$DF
	.dc.b	$9A,$6D,$6E,$3F,$3B,$F2,$78,$AF,$C1,$F2,$79,$89,$5E,$4C,$3F,$29
	.dc.b	$3F,$E3,$27,$F2,$91,$E1,$9F,$D3,$C7,$EA,$1F,$D4,$FC,$7F,$C0,$24
	.dc.b	$F3,$7E,$1F,$C7,$6A,$3D,$4D,$BF,$CA,$FC,$7F,$EC,$8F,$D9,$FC,$75
	.dc.b	$AF,$80,$D9,$FC,$DB,$7A,$BF,$79,$F1,$BF,$40,$95,$EB,$7E,$9F,$C7
	.dc.b	$7E,$F9,$FD,$FB,$63,$FC,$61,$F4,$DF,$EA,$76,$E7,$FE,$65,$8C,$63
	.dc.b	$FE,$30,$7F,$D0,$2D,$CB,$FC,$77,$1F,$E7,$98,$FF,$62,$FC,$12,$31
	.dc.b	$1F,$E6,$96,$BF,$EF,$35,$9F,$7A,$F0,$37,$0F,$F2,$26,$EF,$73,$FC
	.dc.b	$F0,$BD,$EB,$18,$73,$EE,$FF,$93,$A4,$D2,$57,$DF,$F8,$FF,$E4,$27
	.dc.b	$F2,$2F,$8E,$B1,$F8,$FE,$FC,$34,$FC,$94,$7F,$81,$F5,$FF,$75,$0F
	.dc.b	$A6,$7A,$9F,$E2,$4F,$3B,$FD,$81,$18,$9F,$F1,$3F,$C9,$7F,$80,$A3
	.dc.b	$11,$FE,$04,$9D,$D7,$F8,$1E,$BF,$C0,$FF,$21,$AA,$FF,$69,$E3,$8A
	.dc.b	$D4,$27,$57,$FE,$CF,$78,$47,$FB,$EC,$F3,$4F,$96,$BC,$4F,$2C,$38
	.dc.b	$C6,$27,$CE,$75,$9E,$0D,$33,$5C,$1C,$1F,$E1,$E3,$C0,$E4,$F1,$E1
	.dc.b	$E9,$8D,$61,$E7,$AF,$5F,$7F,$AE,$7C,$2D,$7A,$57,$F4,$0B,$D9,$FF
	.dc.b	$00,$38,$BF,$D8,$33,$FE,$AA,$8C,$74,$6E,$3F,$C6,$3F,$26,$AD,$E1
	.dc.b	$D6,$31,$5B,$A4,$61,$F6,$BF,$98,$7D,$46,$B9,$3E,$06,$1F,$72,$78
	.dc.b	$37,$0F,$2B,$99,$DE,$5F,$1E,$BA,$5F,$6B,$DC,$FA,$A3,$EC,$9E,$AA
	.dc.b	$F9,$DA,$F3,$8F,$B0,$7A,$2B,$E1,$C7,$87,$1F,$3E,$1D,$B1,$E1,$7D
	.dc.b	$E7,$37,$C3,$1F,$26,$3D,$9B,$5D,$B0,$F8,$98,$FE,$77,$2F,$EF,$59
	.dc.b	$7E,$07,$2F,$F0,$84,$5D,$7F,$BB,$F6,$7F,$B7,$B1,$FE,$79,$79,$7F
	.dc.b	$DC,$3F,$41,$FE,$E0,$7F,$A6,$B5,$FF,$4B,$B7,$0E,$B1,$FE,$C9,$FA
	.dc.b	$57,$B7,$FE,$47,$EA,$46,$27,$FC,$E5,$18,$9D,$4E,$67,$FD,$37,$3C
	.dc.b	$3D,$BF,$23,$CB,$7B,$5F,$B8,$C1,$A7,$F9,$F2,$B8,$61,$FE,$DB,$FA
	.dc.b	$57,$C5,$72,$5F,$15,$4C,$45,$9F,$2C,$E9,$B3,$EB,$8C,$47,$99,$FF
	.dc.b	$21,$3D,$FE,$2B,$B1,$DF,$D5,$F9,$9C,$4E,$07,$C9,$E1,$7C,$95,$31
	.dc.b	$17,$0F,$EE,$F3,$A6,$CF,$F9,$CE,$BA,$36,$5F,$27,$D7,$E4,$A5,$FD
	.dc.b	$A8,$AF,$DA,$8B,$C1,$FE,$94,$AE,$33,$F8,$0F,$92,$F3,$7A,$B2,$30
	.dc.b	$7E,$11,$F0,$11,$F3,$BE,$3F,$92,$F7,$7C,$8C,$9F,$C5,$C5,$7F,$DB
	.dc.b	$3A,$7C,$59,$F5,$12,$3E,$0F,$0F,$C7,$23,$F9,$4E,$2F,$95,$C3,$23
	.dc.b	$F2,$EE,$9F,$CB,$F0,$FC,$FF,$8D,$91,$F9,$C8,$AF,$8B,$5F,$F6,$9B
	.dc.b	$F4,$27,$E6,$1F,$EA,$15,$E2,$36,$FD,$44,$63,$6B,$3F,$E5,$C8,$FF
	.dc.b	$23,$BB,$6C,$73,$8D,$47,$FB,$55,$AF,$FB,$F1,$AF,$F3,$BF,$A4,$79
	.dc.b	$7E,$05,$E7,$53,$11,$67,$FD,$71,$3A,$6C,$FA,$E3,$11,$FE,$E5,$FD
	.dc.b	$53,$FA,$43,$FE,$A7,$FA,$5F,$F2,$CB,$FE,$60,$6B,$FE,$AB,$8D,$F2
	.dc.b	$7F,$AC,$67,$33,$C9,$AF,$F5,$36,$BF,$EF,$26,$A7,$FC,$BE,$FF,$10
	.dc.b	$F8,$EF,$D3,$1F,$3A,$33,$1F,$EB,$BF,$35,$EE,$FD,$25,$EB,$D3,$11
	.dc.b	$67,$FD,$B5,$3A,$6C,$FD,$D8,$C4,$7E,$0F,$E4,$3F,$6D,$12,$3F,$91
	.dc.b	$C5,$7F,$91,$FC,$A7,$EE,$CC,$65,$FE,$D3,$FA,$07,$37,$E1,$BE,$93
	.dc.b	$F9,$47,$87,$FD,$D5,$B8,$05,$70,$7F,$D1,$F1,$FC,$C5,$FA,$AD,$79
	.dc.b	$7C,$CF,$E5,$63,$11,$CD,$FE,$1E,$FF,$A6,$27,$F8,$BB,$FE,$E3,$6E
	.dc.b	$1F,$EE,$99,$DC,$FF,$3E,$F5,$7F,$8B,$61,$23,$FA,$1C,$57,$FA,$1F
	.dc.b	$A9,$FC,$6C,$C5,$F3,$EC,$F8,$1D,$B2,$BD,$71,$5F,$C7,$8C,$7F,$96
	.dc.b	$BE,$E9,$F6,$1F,$C7,$FE,$81,$F1,$3F,$CB,$17,$D1,$D3,$11,$6B,$FD
	.dc.b	$4B,$2F,$8F,$E0,$66,$19,$7F,$4A,$5E,$1A,$9F,$21,$DC,$93,$D2,$FC
	.dc.b	$5F,$33,$E3,$F8,$49,$3E,$08,$AF,$6E,$1F,$F7,$3E,$0F,$98,$F5,$1A
	.dc.b	$CF,$C5,$9F,$87,$E1,$7D,$DF,$2F,$31,$F5,$C5,$7A,$1E,$8C,$1F,$2F
	.dc.b	$F1,$4D,$61,$ED,$97,$E1,$2D,$3F,$29,$CB,$27,$FD,$BB,$39,$7F,$23
	.dc.b	$1F,$8D,$23,$F6,$0F,$B0,$FE,$47,$5F,$E4,$7C,$EE,$AA,$7E,$86,$3E
	.dc.b	$C7,$C9,$CB,$FA,$AF,$DB,$F9,$8F,$B9,$F2,$7F,$7C,$93,$FE,$D8,$D9
	.dc.b	$1F,$E0,$DF,$A3,$B3,$FA,$9D,$BE,$D7,$AC,$9C,$CF,$C6,$8C,$57,$E2
	.dc.b	$7D,$AF,$F0,$6D,$3D,$AA,$4B,$FC,$4B,$F8,$FF,$6B,$F2,$44,$8F,$23
	.dc.b	$F9,$5F,$99,$FC,$A7,$CC,$C9,$FC,$8C,$7E,$5F,$E6,$64,$7E,$51,$FC
	.dc.b	$D7,$CC,$FE,$6F,$E1,$BF,$3C,$49,$F8,$0F,$F7,$FF,$A2,$7C,$9F,$F8
	.dc.b	$31,$B3,$F0,$79,$3F,$85,$76,$A7,$33,$F8,$68,$FC,$E3,$C7,$CD,$8F
	.dc.b	$13,$F4,$33,$FC,$B1,$C4,$F3,$F9,$7F,$07,$9D,$E3,$24,$FE,$EA,$7C
	.dc.b	$11,$E7,$F9,$3F,$F1,$CF,$D1,$BE,$4A,$9E,$98,$F5,$7C,$7F,$AF,$E5
	.dc.b	$26,$3E,$0C,$D3,$EF,$F9,$4F,$84,$49,$F0,$4A,$3F,$CA,$18,$FF,$A3
	.dc.b	$67,$D9,$D4,$8C,$47,$E0,$DB,$D7,$FB,$0B,$F3,$4F,$09,$7C,$A7,$A8
	.dc.b	$91,$E5,$F8,$BF,$A5,$EE,$7D,$52,$47,$91,$FB,$1F,$1F,$F5,$FD,$D9
	.dc.b	$3E,$D8,$FB,$57,$23,$E0,$3F,$77,$E3,$FE,$F7,$CA,$7E,$04,$93,$F0
	.dc.b	$1F,$C2,$7C,$97,$E0,$FE,$06,$4F,$C9,$8F,$C3,$7C,$9F,$E1,$FE,$42
	.dc.b	$4F,$CD,$8A,$FF,$8C,$B7,$FF,$5F,$3F,$EC,$9E,$8F,$C5,$57,$E4,$4F
	.dc.b	$F3,$E6,$9F,$E5,$39,$FD,$63,$78,$7F,$62,$DE,$BF,$C9,$E1,$FE,$8F
	.dc.b	$AD,$FC,$6F,$F9,$93,$C9,$FE,$C3,$91,$FE,$C7,$CB,$FE,$C5,$9D,$7C
	.dc.b	$6C,$BF,$C9,$D7,$F9,$59,$FC,$E3,$FD,$63,$3F,$E5,$2E,$5B,$5D,$CE
	.dc.b	$53,$C3,$E1,$FC,$44,$7A,$FC,$6D,$7A,$FC,$47,$9C,$C5,$F6,$7F,$CE
	.dc.b	$58,$7F,$8D,$A7,$FA,$4B,$73,$F6,$4E,$BD,$FF,$F7,$4F,$17,$ED,$AB
	.dc.b	$FD,$2B,$7F,$E0,$1F,$93,$4E,$6B,$73,$D3,$E3,$AE,$3F,$1E,$6E,$0F
	.dc.b	$8E,$A6,$23,$8D,$BE,$8F,$F9,$A5,$FF,$39,$6C,$7F,$72,$6B,$FE,$42
	.dc.b	$6B,$FE,$E5,$EB,$7D,$1F,$F4,$1F,$95,$FF,$28,$B6,$33,$3F,$DF,$27
	.dc.b	$2F,$5F,$8A,$79,$7E,$16,$C6,$63,$07,$F0,$BF,$CE,$DC,$3F,$DD,$78
	.dc.b	$7E,$2D,$7C,$EF,$86,$C7,$17,$F2,$0B,$DF,$57,$CB,$8F,$9D,$3A,$AF
	.dc.b	$EF,$B5,$A5,$7D,$37,$4E,$63,$0F,$CD,$C3,$A5,$72,$C7,$A3,$FE,$52
	.dc.b	$F8,$DF,$ED,$CF,$F8,$EC,$C4,$FF,$67,$F3,$53,$A9,$FF,$20,$46,$27
	.dc.b	$5F,$19,$FE,$13,$E7,$FE,$F1,$39,$AF,$EF,$E7,$F9,$82,$75,$5D,$E7
	.dc.b	$55,$FD,$FF,$F1,$BF,$F1,$04,$FF,$3D,$BE,$0D,$8C,$C7,$4A,$5F,$48
	.dc.b	$FC,$DF,$C6,$DD,$7F,$6C,$DD,$9F,$09,$9E,$9F,$F2,$FE,$7F,$D2,$27
	.dc.b	$2D,$9E,$5B,$FF,$BD,$23,$07,$EF,$3E,$4F,$91,$D6,$EC,$73,$70,$FF
	.dc.b	$3C,$FD,$C7,$FD,$02,$E4,$FE,$0B,$F2,$7F,$37,$3F,$E1,$8D,$A3,$4F
	.dc.b	$A5,$A3,$F0,$EF,$D8,$DF,$23,$E4,$38,$8F,$E1,$F3,$F0,$7E,$07,$E0
	.dc.b	$71,$DB,$FE,$2A,$FB,$BE,$0B,$C5,$AF,$F8,$8E,$7F,$39,$18,$33,$D3
	.dc.b	$F3,$F1,$FF,$7C,$F1,$FC,$68,$C7,$89,$EE,$9F,$E3,$FC,$9F,$3C,$FA
	.dc.b	$5F,$03,$71,$FE,$4B,$8F,$BC,$8F,$F7,$5D,$FC,$57,$FE,$F0,$58,$FD
	.dc.b	$93,$EE,$7C,$0D,$AF,$E5,$FF,$21,$F7,$4F,$C0,$7C,$0D,$8F,$E6,$7F
	.dc.b	$0B,$F3,$48,$F7,$7E,$83,$F8,$53,$F1,$1F,$05,$5F,$B5,$E8,$FE,$37
	.dc.b	$F4,$7F,$F3,$EC,$60,$FF,$08,$B8,$D3,$FC,$34,$FF,$A0,$23,$FC,$20
	.dc.b	$D8,$FF,$6C,$F3,$A3,$11,$FD,$6D,$AF,$FA,$BF,$CD,$74,$BD,$0F,$17
	.dc.b	$C6,$FE,$86,$73,$18,$C7,$F3,$9F,$09,$E7,$F8,$EF,$D1,$36,$3C,$FF
	.dc.b	$15,$F9,$1C,$FF,$CD,$D1,$E6,$8C,$19,$F8,$DF,$80,$7E,$AF,$E1,$EC
	.dc.b	$FE,$A7,$E1,$BE,$27,$C7,$7E,$BD,$B1,$F1,$3E,$3B,$B1,$1F,$86,$F8
	.dc.b	$7E,$3B,$E6,$1E,$BF,$CB,$5E,$46,$2B,$F6,$DF,$17,$F4,$7C,$9F,$DE
	.dc.b	$36,$3E,$8F,$77,$AC,$8F,$D0,$7D,$DE,$37,$D9,$3B,$78,$16,$7F,$81
	.dc.b	$7A,$1F,$77,$C2,$FE,$10,$D8,$FE,$29,$F4,$9F,$C4,$1F,$F4,$63,$7A
	.dc.b	$FF,$D8,$AD,$7F,$96,$07,$0F,$F2,$F7,$F3,$BE,$77,$FA,$7D,$E4,$DC
	.dc.b	$3F,$5D,$18,$AF,$F3,$3E,$BF,$EF,$17,$F1,$FE,$1F,$F1,$DC,$3F,$28
	.dc.b	$FF,$21,$F8,$2F,$F2,$9C,$FF,$1D,$9C,$7C,$36,$E3,$F2,$FE,$27,$E6
	.dc.b	$CF,$E5,$1F,$93,$7F,$9E,$23,$27,$F9,$B2,$DF,$0B,$AF,$8D,$EC,$D9
	.dc.b	$FE,$71,$F1,$9F,$B6,$6F,$95,$FC,$1F,$EC,$BD,$5A,$05,$6D,$FE,$8F
	.dc.b	$F4,$9F,$A5,$E2,$FF,$61,$57,$09,$4E,$70,$FE,$B1,$39,$8F,$E8,$B3
	.dc.b	$F4,$1A,$FF,$5D,$6B,$E1,$6F,$90,$D5,$F1,$FF,$BD,$F2,$FE,$1B,$71
	.dc.b	$FF,$3A,$F1,$7F,$BB,$9F,$F4,$7D,$70,$73,$E3,$70,$73,$E8,$7F,$07
	.dc.b	$F1,$FE,$6B,$63,$E6,$7C,$56,$1A,$70,$3F,$B1,$7D,$37,$BF,$6F,$F7
	.dc.b	$FF,$92,$F1,$FF,$99,$4E,$77,$FE,$15,$1F,$CD,$78,$1C,$1F,$D2,$1B
	.dc.b	$8F,$FB,$8D,$B8,$FE,$58,$F9,$B3,$BD,$39,$3C,$DB,$5F,$C7,$BC,$6F
	.dc.b	$E4,$47,$E2,$39,$2C,$7E,$1F,$DC,$C3,$4F,$8F,$1F,$D2,$7E,$97,$FD
	.dc.b	$2F,$1F,$CA,$3C,$4C,$70,$7F,$2D,$75,$E8,$62,$7F,$C0,$94,$8C,$69
	.dc.b	$FE,$C7,$FC,$B7,$FB,$CB,$FC,$C3,$C3,$FF,$01,$46,$3A,$98,$7F,$B4
	.dc.b	$39,$3F,$C7,$AF,$F8,$3F,$E2,$FF,$C3,$18,$7E,$E1,$AF,$F8,$95,$AF
	.dc.b	$E6,$5A,$9F,$CB,$1F,$E1,$C8,$E1,$EA,$FF,$67,$8C,$4F,$E8,$7E,$95
	.dc.b	$F1,$DF,$D2,$E3,$FB,$64,$62,$FF,$A5,$C6,$37,$CD,$59,$FF,$13,$56
	.dc.b	$EB,$FC,$5F,$EB,$3E,$0F,$FC,$45,$FB,$97,$EE,$E7,$31,$8F,$13,$FB
	.dc.b	$D5,$3E,$3A,$CE,$1B,$8F,$FB,$B3,$0F,$CF,$61,$FD,$61,$AF,$FA,$C5
	.dc.b	$AF,$FB,$F1,$AB,$D7,$78,$4F,$37,$CB,$E2,$7F,$59,$8C,$61,$DE,$3D
	.dc.b	$3E,$14,$8F,$D7,$FC,$9D,$C5,$CB,$F7,$17,$85,$FB,$B9,$FE,$36,$75
	.dc.b	$F2,$AD,$FA,$FB,$7F,$C6,$1B,$7F,$12,$8F,$D1,$5A,$31,$87,$BE,$BF
	.dc.b	$CC,$5B,$7A,$BC,$0F,$9B,$1E,$96,$E1,$F1,$F2,$EA,$47,$AE,$08,$E7
	.dc.b	$B7,$EE,$E7,$81,$FC,$3B,$C3,$FB,$33,$F6,$BE,$4E,$4F,$F1,$06,$E7
	.dc.b	$F0,$A3,$F9,$66,$9F,$C6,$E7,$F9,$D9,$FE,$63,$FC,$F3,$FB,$5C,$F1
	.dc.b	$3E,$E7,$9B,$F0,$88,$F9,$9F,$F2,$A7,$93,$FD,$2B,$0E,$C7,$CE,$D3
	.dc.b	$1C,$3F,$47,$2F,$E5,$57,$F9,$03,$FC,$29,$F9,$6F,$BF,$1F,$C5,$7C
	.dc.b	$AE,$35,$FC,$AB,$FE,$19,$FA,$7F,$E0,$79,$FF,$68,$E3,$46,$23,$FD
	.dc.b	$2D,$18,$EE,$FA,$73,$99,$FF,$13,$3F,$C1,$70,$FE,$9D,$A7,$E1,$6F
	.dc.b	$84,$60,$DF,$89,$E1,$C3,$FC,$55,$A7,$F0,$69,$FD,$07,$CB,$E1,$FD
	.dc.b	$5B,$FE,$2F,$FA,$6F,$F1,$F6,$7E,$46,$BF,$E3,$76,$BF,$A1,$6E,$2F
	.dc.b	$A6,$7D,$0F,$3F,$A0,$FF,$1D,$46,$31,$F3,$61,$C6,$7E,$D7,$B1,$27
	.dc.b	$3C,$6E,$6F,$ED,$5C,$1F,$4D,$FF,$06,$CF,$F3,$DF,$8F,$FE,$E1,$F4
	.dc.b	$FF,$D1,$32,$FF,$60,$79,$9F,$A9,$8C,$47,$FB,$0E,$77,$E8,$FF,$86
	.dc.b	$AB,$4F,$39,$FF,$29,$39,$BF,$EF,$9F,$09,$CA,$7F,$BF,$5E,$2B,$51
	.dc.b	$C0,$FF,$12,$C7,$67,$C7,$55,$F1,$0C,$3A,$61,$E2,$8E,$27,$33,$FC
	.dc.b	$18,$FF,$03,$C3,$C5,$1F,$EE,$83,$47,$F4,$02,$99,$7A,$D7,$9E,$3A
	.dc.b	$C6,$1D,$DF,$EB,$A3,$E1,$67,$FE,$73,$9F,$1B,$5F,$4B,$7E,$47,$F9
	.dc.b	$76,$9C,$67,$FC,$45,$18,$7F,$8C,$D7,$FB,$CE,$31,$5C,$7C,$6F,$E3
	.dc.b	$F5,$A3,$8B,$8A,$FB,$35,$FD,$32,$7D,$41,$C3,$F1,$DF,$04,$70,$9E
	.dc.b	$5A,$78,$7C,$BF,$F1,$B5,$ED,$B5,$BC,$7A,$E1,$E9,$D7,$C3,$5C,$1B
	.dc.b	$7F,$6E,$75,$87,$B7,$1E,$F3,$F0,$AF,$AC,$F8,$63,$AC,$E9,$F5,$3F
	.dc.b	$0A,$FA,$C7,$E1,$67,$83,$4F,$8B,$87,$E3,$72,$7E,$76,$E1,$F6,$CF
	.dc.b	$F1,$98,$C5,$7E,$B1,$AF,$F2,$0A,$E3,$F0,$FC,$37,$FF,$5A,$71,$BE
	.dc.b	$A3,$17,$FB,$FF,$81,$BC,$EE,$84,$62,$F8,$57,$EF,$AB,$5E,$5F,$F8
	.dc.b	$CE,$3F,$84,$36,$33,$1F,$E3,$76,$BF,$C9,$ED,$F0,$3E,$2D,$3F,$AD
	.dc.b	$7B,$D7,$C4,$FF,$3C,$DF,$7B,$DB,$E1,$C9,$F2,$FE,$5F,$98,$93,$98
	.dc.b	$AF,$F3,$59,$4F,$95,$B1,$E5,$6B,$FC,$EF,$C6,$78,$BE,$3F,$81,$93
	.dc.b	$F0,$67,$2F,$F9,$B2,$7D,$1E,$8C,$8C,$1F,$E1,$37,$E3,$6B,$FD,$DD
	.dc.b	$BA,$7F,$88,$67,$FB,$7C,$6B,$C9,$78,$9F,$15,$B3,$FE,$7A,$8F,$91
	.dc.b	$CE,$F8,$6F,$91,$DB,$2B,$FC,$44,$F1,$AF,$BB,$EC,$BD,$5F,$9D,$E0
	.dc.b	$4B,$FB,$93,$97,$FC,$55,$1E,$62,$3F,$E0,$47,$EF,$43,$D3,$FB,$1E
	.dc.b	$54,$AE,$71,$5F,$F1,$94,$7C,$02,$3F,$E1,$AD,$3E,$FB,$F9,$96,$F7
	.dc.b	$FF,$77,$6F,$77,$FA,$CF,$AD,$8F,$F8,$56,$FF,$C8,$1E,$5F,$E2,$48
	.dc.b	$FF,$A8,$3A,$FE,$0F,$97,$F8,$DF,$9D,$93,$F0,$22,$BF,$E4,$78,$FA
	.dc.b	$A4,$7F,$3D,$FB,$9F,$FB,$9B,$4F,$F1,$D4,$3F,$74,$F9,$0F,$F8,$40
	.dc.b	$C7,$0F,$EB,$95,$71,$D7,$F8,$BD,$70,$BF,$F0,$A7,$3B,$DF,$FD,$11
	.dc.b	$C6,$75,$5F,$C7,$BD,$37,$BF,$D5,$79,$D4,$C4,$5C,$FF,$CC,$9E,$E7
	.dc.b	$EB,$4C,$3E,$7F,$C0,$5F,$21,$E7,$5E,$ED,$31,$16,$7F,$80,$4E,$9B
	.dc.b	$3F,$C0,$D4,$65,$FE,$6A,$7F,$98,$3F,$2D,$FC,$43,$FE,$77,$F4,$BF
	.dc.b	$80,$1F,$C0,$E3,$31,$F2,$FE,$A5,$F0,$9F,$2D,$F8,$31,$FE,$10,$0E
	.dc.b	$1E,$11,$BC,$FF,$86,$78,$92,$78,$56,$F8,$FF,$96,$47,$EB,$FC,$78
	.dc.b	$FF,$2D,$3F,$8B,$7E,$8F,$FE,$22,$7F,$C3,$6D,$7F,$98,$36,$7F,$31
	.dc.b	$3F,$81,$F8,$AF,$B9,$5E,$77,$FD,$64,$DE,$0D,$72,$ED,$F9,$53,$A9
	.dc.b	$FF,$70,$35,$F9,$6D,$7F,$B0,$FC,$C7,$FB,$3D,$F2,$FA,$5E,$83,$F2
	.dc.b	$51,$98,$FE,$8D,$F8,$D3,$E0,$BF,$87,$79,$72,$7F,$C6,$F3,$97,$FC
	.dc.b	$73,$E2,$7F,$17,$7D,$FF,$47,$F1,$0E,$0F,$F9,$C3,$C5,$FD,$AF,$95
	.dc.b	$FC,$64,$91,$FC,$39,$AF,$C8,$F9,$7F,$9A,$63,$87,$F5,$BF,$56,$F9
	.dc.b	$0F,$C6,$36,$BF,$A8,$35,$FC,$68,$70,$FE,$41,$7C,$EF,$EC,$FE,$E3
	.dc.b	$D9,$3F,$B7,$6D,$F8,$27,$E3,$F8,$F2,$BF,$1D,$C2,$8E,$5C,$3F,$D6
	.dc.b	$78,$3F,$A1,$D7,$F1,$9F,$3B,$F9,$C1,$9F,$4B,$F1,$3F,$49,$27,$F1
	.dc.b	$2E,$9F,$9F,$D3,$FC,$FF,$CF,$91,$83,$E8,$BF,$92,$FA,$59,$18,$3F
	.dc.b	$45,$C0,$E9,$FE,$20,$FE,$F2,$33,$E2,$E2,$72,$FC,$C2,$47,$CC,$7F
	.dc.b	$A9,$79,$FF,$A0,$FA,$29,$3F,$82,$8F,$D1,$FB,$12,$3F,$46,$FF,$4E
	.dc.b	$EE,$FD,$3F,$B3,$FD,$83,$C1,$FE,$EF,$5F,$ED,$86,$BF,$C1,$3C,$34
	.dc.b	$62,$3F,$C7,$70,$EF,$FA,$11,$FB,$0F,$A5,$FD,$99,$FD,$97,$9F,$F6
	.dc.b	$BD,$92,$7F,$6A,$FF,$87,$3C,$77,$77,$EE,$BB,$3C,$84,$AF,$23,$FD
	.dc.b	$C7,$C4,$FD,$FF,$D1,$49,$FD,$0C,$7F,$02,$FA,$29,$1F,$C0,$9F,$EE
	.dc.b	$DC,$1F,$C1,$BE,$8B,$E0,$12,$7E,$03,$FD,$F7,$CB,$FE,$17,$F2,$72
	.dc.b	$7F,$5B,$1E,$78,$FE,$1F,$F1,$52,$7F,$C6,$1A,$A7,$F8,$9F,$4F,$CD
	.dc.b	$24,$FC,$D7,$FC,$65,$3F,$CC,$7C,$4F,$F7,$07,$2F,$E0,$EB,$F8,$BF
	.dc.b	$57,$C5,$F3,$FF,$8E,$92,$3F,$8E,$BF,$E1,$1E,$7F,$E4,$3E,$97,$DB
	.dc.b	$24,$FD,$B7,$FC,$3B,$CA,$FA,$CF,$F1,$E3,$67,$FD,$82,$D7,$FC,$AD
	.dc.b	$F4,$7F,$C9,$E9,$0B,$FE,$01,$F0,$FC,$89,$1F,$EC,$AA,$FF,$29,$BF
	.dc.b	$E5,$FB,$7C,$3F,$88,$91,$FC,$12,$FF,$85,$47,$9B,$E6,$BC,$A4,$9E
	.dc.b	$9E,$7F,$F9,$4F,$F3,$BF,$F3,$5F,$07,$F7,$EF,$C1,$3F,$B6,$D7,$F4
	.dc.b	$F6,$BF,$E6,$F9,$DE,$BE,$AC,$91,$E4,$9F,$13,$5F,$D0,$35,$FC,$FB
	.dc.b	$5E,$5A,$7E,$61,$F9,$3F,$49,$F3,$0F,$F3,$44,$66,$3F,$C2,$1E,$7F
	.dc.b	$F3,$5C,$98,$F8,$F6,$7D,$27,$D4,$3F,$A1,$FA,$97,$0F,$8D,$3A,$6C
	.dc.b	$FD,$C3,$FC,$B3,$8B,$CD,$F0,$92,$7F,$D0,$9F,$7F,$FE,$76,$7E,$A7
	.dc.b	$D2,$FE,$08,$8F,$D9,$9F,$4F,$9B,$23,$EE,$3F,$47,$E8,$FF,$13,$F4
	.dc.b	$B2,$7F,$12,$ED,$FF,$17,$7E,$27,$FD,$B1,$BE,$4F,$F8,$BE,$3F,$83
	.dc.b	$C6,$BC,$6C,$4E,$7C,$59,$3F,$E0,$9A,$DF,$17,$F8,$F1,$FE,$D9,$1F
	.dc.b	$94,$6A,$69,$AE,$6F,$BF,$F1,$FF,$ED,$2F,$95,$FE,$FF,$18,$AF,$CE
	.dc.b	$C6,$3A,$3E,$33,$7E,$47,$86,$9F,$89,$9C,$BF,$67,$E3,$73,$7F,$A5
	.dc.b	$E6,$FF,$40,$3D,$27,$86,$5A,$87,$D5,$3C,$8F,$D6,$47,$F6,$68,$C6
	.dc.b	$9F,$EB,$96,$C6,$67,$6F,$F8,$E3,$E5,$9D,$4F,$AA,$FF,$60,$46,$1B
	.dc.b	$19,$9F,$EE,$73,$97,$C1,$E4,$FE,$A1,$FF,$7B,$E1,$FD,$5F,$E1,$BC
	.dc.b	$53,$99,$FE,$07,$39,$7F,$A3,$76,$3D,$7F,$CE,$C3,$8F,$E7,$CE,$7B
	.dc.b	$FE,$24,$FF,$0B,$8F,$F6,$3C,$63,$64,$6E,$79,$70,$7E,$C3,$D5,$EB
	.dc.b	$87,$4C,$6C,$E7,$1E,$58,$71,$AF,$C9,$7C,$7F,$F3,$59,$FF,$72,$37
	.dc.b	$1F,$F0,$AB,$5F,$F3,$1B,$5F,$E4,$23,$EA,$BF,$8D,$CF,$F9,$76,$79
	.dc.b	$74,$FF,$A4,$27,$35,$FE,$8F,$9C,$D7,$29,$78,$FF,$1A,$F8,$79,$16
	.dc.b	$B5,$37,$0F,$E6,$37,$D2,$FE,$6E,$DF,$E3,$1E,$FD,$37,$0F,$E6,$5E
	.dc.b	$B6,$3A,$55,$AD,$93,$4C,$7A,$B6,$FF,$A1,$FC,$B7,$13,$C5,$F0,$8F
	.dc.b	$AA,$FF,$66,$4E,$8F,$E8,$BA,$7E,$02,$75,$C3,$C2,$B7,$BF,$F4,$CF
	.dc.b	$5F,$13,$ED,$87,$97,$FE,$AD,$C7,$FC,$8F,$18,$E5,$FF,$5D,$BF,$E6
	.dc.b	$A6,$BF,$D5,$27,$83,$F9,$2F,$5F,$A4,$63,$3F,$8F,$BA,$73,$7F,$0F
	.dc.b	$BF,$E3,$78,$17,$2F,$EC,$6F,$97,$D9,$F4,$E7,$FE,$A6,$6B,$F8,$46
	.dc.b	$BF,$D9,$9B,$CB,$F9,$AF,$53,$E8,$77,$7D,$16,$E1,$FD,$C3,$C5,$FA
	.dc.b	$C7,$DA,$FD,$07,$FB,$B4,$64,$F9,$A7,$FB,$B7,$8D,$F0,$3F,$0B,$97
	.dc.b	$EA,$36,$33,$3F,$E7,$09,$CB,$FD,$2B,$E8,$3F,$08,$77,$ED,$B1,$FE
	.dc.b	$02,$FA,$2F,$C4,$1E,$3F,$4A,$C7,$8B,$81,$87,$E0,$FC,$7F,$01,$27
	.dc.b	$C1,$97,$F1,$47,$FA,$CF,$CD,$FE,$4C,$F7,$79,$56,$3A,$71,$37,$FC
	.dc.b	$BC,$77,$C7,$F2,$1F,$03,$F1,$7C,$1F,$C6,$3F,$8D,$E8,$F8,$04,$AF
	.dc.b	$0B,$F8,$6C,$3A,$71,$FE,$4F,$DE,$F9,$CD,$AF,$0E,$4E,$1F,$D4,$7D
	.dc.b	$17,$D3,$6C,$F8,$FE,$AD,$ED,$7D,$16,$C7,$CF,$6F,$1F,$FB,$B3,$74
	.dc.b	$62,$3F,$CD,$3E,$33,$E4,$7F,$6C,$7E,$07,$E4,$2E,$1D,$FC,$CF,$DD
	.dc.b	$47,$EC,$B0,$FE,$49,$E1,$FF,$02,$C3,$FD,$E3,$7E,$5F,$37,$C8,$47
	.dc.b	$C8,$F9,$E3,$91,$FC,$23,$E2,$1F,$2B,$8A,$F7,$7C,$2F,$F0,$9F,$91
	.dc.b	$F0,$36,$3E,$17,$8F,$F0,$48,$FC,$17,$E2,$78,$FE,$73,$B7,$D1,$59
	.dc.b	$FC,$57,$0B,$EA,$FF,$92,$BF,$DA,$7D,$5F,$E0,$91,$8E,$BF,$F6,$74
	.dc.b	$63,$0F,$E0,$35,$C5,$B5,$FD,$AE,$FF,$84,$B6,$7F,$DA,$2F,$FA,$C6
	.dc.b	$3C,$2E,$78,$3F,$AB,$CF,$8F,$C8,$ED,$5F,$94,$DD,$F3,$1F,$CA,$EB
	.dc.b	$E4,$36,$75,$E4,$61,$A7,$18,$F9,$DF,$33,$FC,$DB,$E2,$FE,$7C,$F5
	.dc.b	$8C,$71,$FD,$3C,$77,$3E,$16,$E2,$CF,$CF,$F5,$2F,$87,$FE,$55,$F5
	.dc.b	$7F,$2A,$93,$9C,$3E,$7C,$E6,$3F,$AA,$CF,$F5,$C3,$7E,$CF,$CC,$8C
	.dc.b	$C7,$F6,$06,$BF,$E8,$27,$FA,$CD,$7D,$A7,$FA,$A3,$BF,$37,$F9,$E4
	.dc.b	$62,$7F,$83,$F9,$DC,$1F,$55,$7F,$9C,$F6,$7D,$3F,$F4,$EA,$D6,$9F
	.dc.b	$D7,$A7,$35,$F4,$FE,$4B,$59,$7E,$65,$B1,$F5,$FE,$4B,$E7,$7A,$FF
	.dc.b	$9D,$6C,$7D,$BF,$33,$0D,$38,$1E,$BF,$3F,$FA,$A7,$D5,$FC,$9D,$BC
	.dc.b	$E7,$C1,$8C,$F1,$7B,$BC,$AF,$90,$4B,$FE,$B3,$C2,$8F,$99,$E3,$EE
	.dc.b	$7E,$43,$70,$EB,$DC,$E6,$FE,$B1,$5F,$17,$E5,$7F,$6D,$E7,$FE,$25
	.dc.b	$FD,$C7,$C2,$E2,$7F,$2C,$DC,$3C,$E3,$EA,$DC,$1F,$DB,$D7,$0D,$8C
	.dc.b	$39,$F8,$7C,$61,$FB,$6A,$46,$2F,$AE,$1F,$DA,$EB,$F9,$CF,$CF,$FF
	.dc.b	$38,$9C,$F8,$DF,$E1,$48,$C7,$4F,$EB,$AF,$FC,$59,$5F,$EC,$C6,$A6
	.dc.b	$93,$C4,$FF,$03,$3F,$9D,$3F,$71,$87,$F8,$D1,$FD,$07,$D5,$7B,$23
	.dc.b	$F8,$6A,$E9,$FF,$07,$39,$BF,$8A,$E6,$77,$B6,$B6,$DB,$63,$F8,$BF
	.dc.b	$C9,$3E,$6B,$FC,$30,$FF,$5D,$C3,$E7,$B9,$9D,$FB,$5A,$CB,$77,$67
	.dc.b	$9B,$F9,$FF,$5B,$FB,$D7,$80,$69,$27,$FD,$CD,$F4,$FF,$D3,$38,$3F
	.dc.b	$C5,$D5,$B9,$FE,$D3,$39,$7F,$DE,$FE,$A2,$75,$A7,$E3,$BC,$27,$9B
	.dc.b	$FC,$4C,$FD,$87,$AD,$70,$FE,$E1,$7F,$D8,$DB,$1F,$E4,$EC,$5F,$1D
	.dc.b	$C2,$77,$DF,$EA,$23,$FE,$11,$BF,$D1,$3E,$AF,$5B,$F7,$D3,$CD,$B1
	.dc.b	$E0,$F0,$3E,$19,$1E,$4F,$F7,$BC,$3F,$D0,$FA,$7E,$D3,$5F,$74,$F9
	.dc.b	$7D,$4F,$E1,$4F,$F2,$FF,$AA,$98,$7C,$6F,$56,$E1,$FB,$0C,$51,$ED
	.dc.b	$3D,$DE,$A5,$AF,$F0,$AF,$0F,$F4,$27,$F9,$45,$BC,$44,$E6,$3F,$96
	.dc.b	$46,$3D,$0F,$F0,$94,$E6,$7F,$D1,$72,$C5,$F0,$9F,$D1,$6F,$E7,$37
	.dc.b	$0F,$F1,$17,$D5,$7E,$67,$9B,$FC,$DD,$F4,$DF,$6A,$73,$DD,$FE,$76
	.dc.b	$8C,$67,$F1,$6F,$F8,$0D,$7F,$90,$DB,$1F,$E1,$56,$BF,$E9,$EE,$24
	.dc.b	$62,$3F,$16,$FF,$02,$9F,$1F,$77,$FA,$12,$31,$3D,$E3,$FD,$0F,$F5
	.dc.b	$1F,$E5,$A8,$FA,$DA,$2F,$3F,$31,$FE,$8C,$8C,$47,$F8,$FE,$BF,$98
	.dc.b	$57,$8D,$AF,$E7,$1A,$FF,$94,$9B,$BF,$CB,$F0,$FF,$D2,$FF,$50,$FA
	.dc.b	$AF,$CC,$46,$1F,$CD,$7B,$33,$1F,$D4,$F8,$97,$2F,$E5,$20,$FF,$6D
	.dc.b	$0F,$AA,$7A,$3F,$DA,$7D,$CF,$F6,$FC,$62,$7F,$B6,$7C,$87,$18,$C4
	.dc.b	$73,$9D,$D7,$E2,$AB,$B7,$C1,$EA,$95,$FE,$EA,$8C,$47,$89,$D4,$66
	.dc.b	$B9,$4F,$95,$F2,$63,$E1,$73,$87,$08,$C4,$F9,$85,$E5,$B5,$D2,$30
	.dc.b	$78,$B0,$F3,$C7,$69,$D9,$CE,$77,$1E,$4D,$BA,$CF,$76,$BF,$EE,$CA
	.dc.b	$F9,$D1,$88,$FA,$11,$8A,$FF,$78,$FC,$97,$F3,$1E,$8C,$57,$D5,$8C
	.dc.b	$3F,$B3,$9D,$F9,$1F,$E2,$78,$C1,$B7,$18,$78,$AB,$8D,$7D,$67,$FA
	.dc.b	$45,$7B,$8E,$DE,$4F,$E0,$DF,$78,$8E,$39,$78,$31,$E1,$39,$AE,$17
	.dc.b	$D5,$B5,$B9,$E7,$7E,$AD,$35,$1F,$58,$E4,$58,$D4,$F1,$BD,$CF,$2B
	.dc.b	$EA,$FB,$F0,$F1,$DF,$08,$F5,$9C,$E7,$D9,$87,$C1,$CB,$E4,$3F,$9B
	.dc.b	$C9,$F5,$4F,$6F,$E1,$65,$FB,$BC,$7F,$91,$4E,$5F,$F5,$DB,$5F,$F0
	.dc.b	$EA,$E7,$E3,$BF,$E0,$23,$FC,$B5,$3F,$EB,$FE,$57,$D2,$6A,$3F,$D9
	.dc.b	$F1,$87,$AF,$B1,$9E,$C7,$4B,$E4,$3F,$9A,$D7,$F0,$56,$CE,$9B,$3B
	.dc.b	$9F,$F7,$33,$FD,$C5,$FE,$25,$F1,$DD,$60,$DF,$3F,$C1,$8B,$CA,$BE
	.dc.b	$54,$C3,$D5,$15,$FF,$61,$4A,$7C,$6D,$8F,$F7,$B3,$5D,$FD,$0F,$B7
	.dc.b	$C3,$FF,$7F,$C7,$FB,$8D,$AF,$F6,$19,$4F,$C7,$E4,$FA,$FF,$1B,$7C
	.dc.b	$8D,$F2,$FF,$9A,$FE,$13,$5C,$8A,$D4,$7C,$8B,$FE,$B6,$DB,$EF,$97
	.dc.b	$F5,$86,$BF,$67,$E3,$39,$91,$FC,$0F,$CC,$F8,$7E,$AA,$47,$F6,$A7
	.dc.b	$4F,$E1,$39,$3E,$D7,$B3,$FE,$D2,$8F,$F8,$00,$D7,$46,$3E,$1B,$ED
	.dc.b	$C6,$63,$FA,$C3,$5F,$F1,$0B,$5F,$CA,$F2,$76,$F8,$39,$1F,$96,$8A
	.dc.b	$FF,$BD,$E5,$3D,$89,$1F,$E0,$0E,$84,$FF,$8A,$A3,$CF,$EC,$7A,$8F
	.dc.b	$9B,$EC,$7D,$1F,$87,$93,$FE,$0A,$74,$FE,$9B,$D7,$F8,$3E,$74,$8C
	.dc.b	$1F,$09,$F7,$91,$F1,$61,$FE,$D4,$7F,$3F,$F5,$5F,$D2,$1F,$F7,$6B
	.dc.b	$FE,$BD,$E7,$FF,$6E,$B7,$C1,$FE,$34,$EE,$7F,$D0,$EE,$5F,$EC,$1F
	.dc.b	$53,$F4,$6B,$FD,$CA,$DA,$E3,$3F,$B5,$1E,$CB,$E3,$70,$BE,$32,$98
	.dc.b	$8B,$3F,$DB,$70,$FF,$78,$D7,$8D,$BE,$7E,$A7,$FD,$37,$F5,$2F,$4B
	.dc.b	$C0,$BE,$4E,$98,$8B,$3F,$C0,$A7,$4D,$9F,$E0,$AA,$32,$FF,$8F,$17
	.dc.b	$A7,$FA,$BF,$A4,$FC,$1C,$F9,$F5,$FE,$AF,$18,$8F,$2C,$FC,$3F,$A9
	.dc.b	$F8,$71,$B3,$FC,$14,$FC,$57,$01,$C3,$FE,$B8,$78,$B7,$5E,$9B,$48
	.dc.b	$E3,$3C,$35,$FE,$65,$E0,$BE,$A3,$EB,$57,$F3,$43,$F7,$E7,$F0,$57
	.dc.b	$31,$FC,$F2,$3F,$83,$7C,$5C,$7F,$9C,$9F,$77,$7F,$F6,$0B,$E8,$5F
	.dc.b	$17,$93,$5F,$13,$67,$F9,$8C,$E9,$B3,$F4,$23,$11,$FE,$6A,$7F,$D8
	.dc.b	$0F,$C4,$F9,$9F,$F6,$BB,$7D,$DF,$17,$C4,$FA,$7C,$0F,$57,$C6,$4A
	.dc.b	$F9,$DA,$7C,$19,$C4,$7F,$A7,$FE,$27,$F9,$11,$8D,$FE,$3E,$EF,$87
	.dc.b	$FE,$23,$ED,$4B,$E7,$39,$7F,$89,$C7,$F2,$B2,$38,$3E,$DF,$D4,$3E
	.dc.b	$73,$F4,$8F,$FA,$32,$33,$F1,$38,$9C,$BF,$92,$9D,$4F,$FA,$31,$AF
	.dc.b	$F8,$3C,$38,$7F,$C2,$4B,$E7,$3F,$65,$F5,$2D,$E9,$E2,$8F,$C0,$79
	.dc.b	$BF,$81,$F4,$66,$1E,$E8,$AF,$FB,$17,$E0,$3F,$A4,$18,$DF,$E3,$F0
	.dc.b	$7E,$27,$A6,$57,$C6,$8A,$FF,$B3,$7E,$23,$FA,$81,$8A,$FF,$03,$47
	.dc.b	$AB,$E1,$FE,$2F,$4C,$8F,$93,$15,$F8,$51,$F1,$48,$FF,$7C,$DF,$F6
	.dc.b	$F5,$FA,$0F,$53,$F0,$F3,$FE,$DD,$E9,$FA,$3E,$27,$F4,$CE,$A9,$3F
	.dc.b	$D3,$1F,$D1,$67,$FD,$43,$AB,$FA,$59,$27,$F9,$6B,$FE,$08,$F0,$BF
	.dc.b	$C2,$11,$FE,$E9,$EE,$FE,$D8,$7F,$84,$63,$39,$FF,$88,$FC,$F4,$3E
	.dc.b	$73,$44,$BF,$C6,$E5,$C7,$E3,$BB,$12,$3F,$A7,$57,$FB,$B9,$BE,$CF
	.dc.b	$F7,$BF,$0A,$47,$F0,$DC,$7B,$CF,$F7,$FC,$24,$FF,$B6,$1F,$F0,$27
	.dc.b	$BD,$E6,$F2,$3F,$D3,$97,$E6,$EB,$F7,$12,$FC,$2F,$F8,$0E,$BD,$51
	.dc.b	$FE,$E2,$F8,$6F,$F0,$61,$D3,$5E,$BA,$FC,$3F,$A9,$95,$FB,$4B,$F5
	.dc.b	$3F,$EF,$7D,$BF,$9D,$3E,$EE,$7F,$9C,$47,$FD,$DB,$C5,$FE,$30,$F7
	.dc.b	$FE,$77,$97,$27,$A4,$7F,$87,$66,$47,$FB,$D9,$F7,$EB,$FE,$45,$FA
	.dc.b	$6F,$F2,$6C,$FE,$AF,$74,$62,$3F,$DC,$0D,$7E,$2B,$5F,$F7,$F7,$CF
	.dc.b	$FF,$B2,$DF,$E1,$ED,$5C,$BF,$CC,$A3,$0F,$F3,$4D,$5F,$27,$FE,$35
	.dc.b	$9F,$A5,$5D,$3E,$1B,$07,$D6,$EE,$F4,$7C,$07,$F9,$63,$2F,$F1,$B5
	.dc.b	$7F,$84,$5A,$FA,$DA,$FF,$A5,$75,$60,$F9,$CF,$01,$27,$EA,$70,$7E
	.dc.b	$13,$9F,$FC,$91,$F4,$7F,$1B,$7F,$E9,$6D,$8F,$F5,$AE,$FF,$BA,$9F
	.dc.b	$DD,$FD,$47,$F8,$27,$0F,$E2,$CD,$7F,$D4,$CD,$7F,$DD,$4D,$CD,$87
	.dc.b	$FA,$16,$BE,$47,$47,$C5,$3C,$7F,$17,$E4,$FA,$8F,$F0,$77,$A4,$F1
	.dc.b	$A9,$9C,$FF,$A4,$3F,$DF,$23,$51,$FE,$82,$6B,$FC,$AC,$38,$7F,$03
	.dc.b	$A7,$E1,$C8,$FF,$4C,$C9,$F5,$1F,$67,$EA,$24,$FE,$09,$DB,$F8,$2D
	.dc.b	$FF,$DB,$74,$FA,$3F,$9A,$49,$E9,$F1,$1F,$83,$F0,$24,$7E,$31,$D3
	.dc.b	$CC,$0F,$A8,$FE,$3F,$C3,$FC,$B1,$FE,$D6,$DF,$17,$FC,$69,$1A,$CB
	.dc.b	$FD,$B0,$E9,$BE,$36,$8D,$CD,$BF,$F5,$4D,$7F,$80,$1F,$90,$D7,$FA
	.dc.b	$A3,$5F,$7F,$C3,$7F,$BE,$E7,$E0,$EF,$65,$CB,$F8,$08,$C3,$F8,$2A
	.dc.b	$BB,$7F,$AB,$AB,$FE,$00,$DF,$26,$9C,$A3,$4B,$E5,$3F,$C1,$D7,$CE
	.dc.b	$36,$3E,$9D,$DD,$E2,$8F,$C1,$46,$27,$FD,$8E,$D8,$CC,$ED,$FE,$E5
	.dc.b	$F2,$9F,$91,$FA,$8F,$F6,$6B,$63,$F1,$8D,$77,$CE,$9E,$BE,$37,$E8
	.dc.b	$F0,$FF,$2D,$7C,$0E,$09,$7F,$DF,$F2,$FD,$CF,$8F,$FE,$ED,$D3,$DB
	.dc.b	$E0,$FF,$B3,$F9,$3E,$1D,$69,$B1,$9B,$F8,$9C,$1F,$88,$AD,$46,$0F
	.dc.b	$E3,$78,$75,$9D,$57,$FB,$2A,$F9,$CE,$6B,$F2,$3F,$51,$FF,$00,$67
	.dc.b	$FD,$DA,$D8,$FD,$63,$5F,$F0,$C3,$5F,$CC,$85,$F4,$49,$FF,$31,$C7
	.dc.b	$3E,$7F,$F0,$3B,$A8,$CF,$7E,$EF,$F4,$87,$3E,$FC,$5F,$0C,$2D,$F8
	.dc.b	$30,$56,$2F,$D7,$57,$4F,$E4,$FE,$DD,$D3,$FC,$75,$8F,$F3,$EF,$2B
	.dc.b	$13,$96,$47,$F4,$2F,$6E,$E3,$FB,$7F,$2D,$F5,$1F,$EC,$AA,$FE,$EC
	.dc.b	$D9,$FD,$E6,$5F,$CD,$BE,$73,$C9,$7C,$3E,$07,$C5,$13,$0F,$E9,$FF
	.dc.b	$3D,$70,$E3,$AF,$F4,$C9,$FE,$CE,$D8,$FC,$53,$5F,$E8,$8D,$7E,$FF
	.dc.b	$A3,$EF,$F4,$3E,$03,$6B,$FD,$2B,$E5,$79,$F9,$3F,$AE,$77,$FC,$9C
	.dc.b	$3F,$C3,$9A,$7C,$78,$F9,$1D,$9F,$E3,$E9,$FC,$5B,$5F,$ED,$2D,$DF
	.dc.b	$FA,$AB,$74,$FF,$57,$E5,$FD,$B3,$D6,$FA,$5E,$6F,$D3,$7E,$B7,$D4
	.dc.b	$3E,$6A,$9F,$5E,$A9,$FE,$82,$87,$CD,$77,$6C,$7F,$83,$1B,$9F,$EB
	.dc.b	$DA,$FF,$69,$7F,$20,$FD,$66,$ED,$FE,$AE,$7F,$2D,$3F,$AD,$72,$DF
	.dc.b	$4F,$CE,$7A,$3A,$EC,$79,$7C,$FF,$21,$1F,$23,$E7,$F3,$7D,$E7,$C0
	.dc.b	$F9,$5B,$3E,$DF,$95,$F8,$7E,$87,$C4,$6C,$7C,$1F,$37,$E0,$11,$F8
	.dc.b	$0F,$C3,$FA,$0F,$97,$E7,$7C,$C6,$CF,$C9,$F9,$DF,$9F,$F2,$BF,$41
	.dc.b	$B1,$F3,$7E,$53,$E6,$11,$F9,$8F,$CF,$F8,$EF,$AE,$7D,$8F,$90,$B3
	.dc.b	$F5,$7E,$3F,$F3,$FF,$51,$FC,$5F,$97,$E6,$46,$3E,$73,$EF,$FA,$3F
	.dc.b	$64,$97,$F6,$5F,$C5,$7C,$43,$E6,$7F,$74,$D9,$FC,$17,$56,$27,$3F
	.dc.b	$4D,$E0,$F2,$2D,$7F,$9A,$74,$47,$2F,$17,$A9,$27,$FC,$A7,$8F,$EE
	.dc.b	$71,$FF,$17,$2F,$9A,$F0,$9E,$2F,$8F,$B5,$FE,$FE,$FA,$8F,$EA,$2D
	.dc.b	$F2,$BF,$55,$18,$C3,$F8,$DA,$8C,$F8,$FF,$A9,$94,$A3,$0D,$AF,$F0
	.dc.b	$E4,$EA,$7E,$2F,$CA,$72,$9F,$F7,$2E,$5F,$ED,$F7,$C9,$3F,$23,$E4
	.dc.b	$7E,$4F,$CD,$59,$DB,$F3,$3E,$4B,$18,$FF,$18,$C1,$18,$E6,$FE,$69
	.dc.b	$E6,$26,$61,$E0,$F9,$BF,$AB,$1F,$59,$FA,$FE,$FE,$27,$CA,$DC,$7F
	.dc.b	$9E,$FD,$37,$F4,$97,$FC,$42,$D4,$D2,$7B,$3F,$C5,$5F,$DA,$C7,$08
	.dc.b	$C7,$B7,$96,$5F,$F8,$B3,$3F,$F6,$AF,$DB,$FF,$40,$BF,$0D,$7F,$34
	.dc.b	$F0,$FF,$7B,$0F,$5D,$18,$DB,$E0,$7C,$AE,$EF,$F2,$6C,$CB,$E8,$FC
	.dc.b	$BD,$C7,$E5,$7C,$9F,$E6,$0F,$CC,$FC,$A5,$AF,$AF,$F1,$38,$69,$F6
	.dc.b	$A3,$DD,$E3,$FB,$FC,$8B,$3E,$B3,$FB,$7F,$12,$3E,$27,$8F,$F8,$B8
	.dc.b	$34,$93,$F8,$AE,$1F,$C9,$3C,$F3,$FA,$33,$FD,$06,$73,$1F,$5A,$3F
	.dc.b	$63,$F2,$7F,$B2,$F8,$4B,$7F,$90,$7F,$6B,$F2,$78,$BF,$C5,$37,$97
	.dc.b	$FB,$E4,$3E,$9D,$B7,$F4,$F6,$E1,$E3,$9D,$6F,$9A,$D3,$27,$F3,$FC
	.dc.b	$96,$BF,$C0,$F9,$FA,$31,$EE,$FF,$67,$6F,$83,$FD,$76,$31,$7F,$E1
	.dc.b	$87,$F6,$D5,$FE,$1A,$0E,$1D,$7B,$BF,$B7,$8C,$47,$F9,$E3,$ED,$DE
	.dc.b	$C7,$5A,$FE,$D5,$18,$BF,$65,$FF,$03,$C7,$E1,$FD,$43,$9B,$FB,$83
	.dc.b	$6B,$E1,$3F,$13,$2F,$F1,$1B,$96,$ED,$C0,$FF,$0F,$7D,$43,$E9,$F1
	.dc.b	$7F,$2F,$E0,$1D,$5F,$DF,$5B,$19,$EB,$C4,$E5,$91,$F8,$5E,$26,$0F
	.dc.b	$97,$CE,$7F,$E1,$E8,$FF,$69,$57,$FB,$F1,$B1,$FE,$5F,$6B,$FE,$D9
	.dc.b	$6B,$FE,$59,$5F,$07,$EA,$F9,$6F,$5B,$6B,$F2,$DE,$0F,$ED,$3D,$7F
	.dc.b	$DC,$1F,$03,$E4,$EC,$F9,$F0,$46,$63,$1F,$29,$72,$F4,$7A,$FF,$E1
	.dc.b	$7B,$FF,$36,$47,$96,$30,$67,$E6,$7E,$6B,$F3,$9B,$7D,$FE,$3F,$91
	.dc.b	$1F,$84,$F5,$F8,$EF,$AA,$FD,$66,$C7,$9F,$E3,$7E,$C9,$F6,$BD,$7B
	.dc.b	$1E,$B7,$0F,$B7,$E3,$AE,$5F,$47,$E3,$7D,$44,$9F,$A4,$FB,$7E,$37
	.dc.b	$F0,$AF,$E1,$9B,$1F,$33,$EA,$3F,$1D,$DD,$FE,$71,$E3,$47,$E3,$E3
	.dc.b	$0E,$A7,$FD,$AF,$3A,$EE,$FF,$2E,$CB,$C4,$FF,$22,$4F,$26,$C6,$6B
	.dc.b	$FD,$ED,$39,$7F,$11,$E2,$72,$AF,$CD,$B6,$3F,$03,$F4,$FF,$E3,$CC
	.dc.b	$7C,$CF,$F9,$1E,$33,$F1,$9F,$E6,$E8,$C6,$BF,$E7,$7F,$A6,$FD,$25
	.dc.b	$78,$1B,$1F,$BF,$C9,$18,$8F,$F4,$DC,$63,$C1,$FC,$37,$B3,$FA,$EF
	.dc.b	$1F,$F5,$ED,$C3,$F2,$1E,$1F,$EC,$8F,$D9,$FC,$7D,$8F,$E3,$98,$FF
	.dc.b	$BF,$BE,$4F,$FD,$FF,$F5,$1F,$CE,$63,$FD,$53,$F0,$BF,$EA,$18,$C4
	.dc.b	$7F,$AB,$27,$7E,$9F,$F1,$5A,$D3,$FC,$5E,$7F,$DC,$C3,$CF,$E2,$A7
	.dc.b	$FD,$59,$18,$8D,$BA,$8C,$CF,$3A,$EF,$5E,$4C,$2B,$E3,$18,$71,$8C
	.dc.b	$C7,$B4,$F1,$57,$9E,$11,$83,$F9,$75,$63,$0E,$51,$93,$B3,$83,$D1
	.dc.b	$1B,$AE,$93,$B3,$C1,$E3,$FC,$17,$A8,$70,$FA,$5A,$F0,$AF,$F6,$CF
	.dc.b	$CB,$FF,$26,$E4,$FF,$76,$46,$3C,$A4,$62,$BD,$75,$B7,$B5,$6E,$75
	.dc.b	$09,$D4,$FD,$C5,$E5,$7F,$BA,$23,$0D,$C7,$95,$7C,$3A,$E3,$1E,$69
	.dc.b	$D4,$F8,$9A,$F7,$C3,$CE,$F1,$6E,$9D,$A3,$ED,$DF,$89,$B3,$AB,$F0
	.dc.b	$E1,$B9,$F1,$61,$DA,$FC,$71,$F0,$AF,$83,$E3,$3A,$CF,$E0,$1F,$C3
	.dc.b	$DF,$B3,$0F,$D0,$64,$FA,$7F,$F6,$05,$7F,$4F,$6B,$FE,$EA,$6E,$7F
	.dc.b	$E5,$2F,$17,$FC,$16,$FF,$95,$21,$EC,$F7,$8C,$74,$DE,$0F,$C8,$E3
	.dc.b	$DE,$31,$8F,$E1,$70,$E2,$FF,$4E,$E8,$FE,$BC,$FF,$07,$F3,$3F,$CE
	.dc.b	$3D,$3F,$C1,$9B,$3A,$9F,$E1,$5D,$4F,$8C,$EB,$68,$7C,$5E,$1B,$5F
	.dc.b	$C7,$2B,$C7,$9F,$C3,$F6,$B8,$E2,$F2,$A9,$AD,$13,$8C,$3F,$8E,$4F
	.dc.b	$90,$8F,$F1,$FF,$27,$9F,$95,$23,$DD,$1C,$FE,$47,$A7,$AB,$27,$E9
	.dc.b	$45,$7F,$C4,$99,$FD,$79,$FF,$49,$B6,$3F,$89,$35,$DE,$4F,$8F,$F4
	.dc.b	$12,$3E,$F7,$C8,$FA,$FE,$46,$47,$E0,$5D,$3F,$81,$F9,$1F,$81,$C1
	.dc.b	$23,$F0,$D1,$5E,$F1,$E4,$C3,$E9,$47,$91,$AF,$93,$E4,$7E,$57,$1C
	.dc.b	$8F,$C8,$45,$7D,$31,$CC,$8F,$3F,$53,$E9,$11,$FC,$C7,$C0,$FE,$03
	.dc.b	$E9,$FF,$D9,$11,$FE,$18,$6B,$FC,$2B,$AB,$FB,$4C,$63,$AF,$F3,$57
	.dc.b	$F7,$7B,$FE,$F7,$C1,$4C,$3F,$49,$15,$F9,$B1,$EF,$23,$FE,$C6,$FA
	.dc.b	$7F,$E8,$0F,$F9,$AD,$AF,$84,$FE,$66,$E2,$FF,$55,$F0,$3F,$A4,$94
	.dc.b	$FF,$98,$E7,$FC,$F2,$DC,$BF,$78,$D9,$E7,$E4,$71,$B7,$C9,$D2,$4F
	.dc.b	$EF,$5D,$7C,$9C,$8F,$F4,$04,$A3,$71,$E7,$1F,$8E,$BD,$AF,$37,$D3
	.dc.b	$FF,$B9,$E7,$FD,$10,$D7,$84,$ED,$CF,$A1,$FE,$EA,$53,$B7,$FD,$00
	.dc.b	$D7,$F8,$C3,$5F,$EB,$2B,$CF,$F0,$FA,$12,$7C,$31,$F2,$7E,$9F,$FD
	.dc.b	$63,$1F,$E9,$B6,$BE,$B6,$F0,$7F,$19,$CF,$86,$B2,$B8,$3C,$F5,$FE
	.dc.b	$01,$B7,$F0,$6F,$A7,$7C,$BD,$33,$7F,$E7,$78,$CB,$FC,$AC,$FE,$10
	.dc.b	$6C,$FF,$57,$7D,$0F,$FA,$45,$FF,$58,$6F,$FE,$39,$6F,$B1,$F8,$39
	.dc.b	$D4,$7D,$39,$D3,$FE,$11,$6B,$FE,$A9,$6B,$FD,$05,$AF,$F8,$09,$7C
	.dc.b	$4F,$E3,$DF,$F5,$7C,$78,$BE,$1F,$13,$EF,$F9,$8F,$81,$F2,$F2,$BD
	.dc.b	$D1,$F0,$FE,$5A,$47,$C1,$74,$FE,$7F,$C7,$68,$F9,$5F,$94,$49,$F9
	.dc.b	$1F,$2F,$FB,$9F,$66,$47,$EE,$5D,$3F,$93,$FA,$6F,$C6,$37,$B3,$F5
	.dc.b	$11,$8E,$7F,$F6,$FF,$C5,$7D,$2A,$D1,$FE,$DD,$97,$C4,$70,$38,$DD
	.dc.b	$9F,$F1,$E0,$9D,$B8,$6B,$FE,$41,$6B,$FD,$BF,$4F,$7F,$CB,$7E,$1F
	.dc.b	$E5,$A4,$FE,$16,$2B,$F1,$76,$7C,$BF,$E3,$4C,$E8,$F0,$FF,$B4,$CB
	.dc.b	$E5,$BF,$25,$F2,$D3,$0F,$C8,$45,$7A,$7C,$BF,$E5,$CC,$61,$CE,$3F
	.dc.b	$C6,$97,$FE,$D4,$7E,$27,$C3,$7C,$A3,$F3,$DC,$56,$3E,$04,$7E,$E7
	.dc.b	$73,$4C,$75,$7D,$2F,$0E,$57,$CF,$8F,$D4,$7C,$54,$8F,$D3,$3F,$AB
	.dc.b	$F8,$4F,$D6,$7C,$97,$F5,$CC,$CD,$25,$7F,$5E,$8F,$D0,$E9,$FE,$A7
	.dc.b	$FA,$5F,$F0,$13,$7E,$0B,$F6,$DE,$17,$ED,$FC,$19,$87,$ED,$5F,$DD
	.dc.b	$78,$3F,$BB,$F8,$8F,$01,$27,$B3,$E3,$EA,$FF,$05,$E4,$F7,$A9,$FC
	.dc.b	$0E,$3F,$C1,$3E,$AF,$8B,$A7,$FC,$3E,$DC,$3F,$CC,$4D,$92,$F8,$AE
	.dc.b	$3F,$27,$FD,$F2,$0D,$25,$7F,$7E,$87,$C9,$78,$3C,$29,$3F,$E7,$EE
	.dc.b	$1F,$BD,$8F,$6A,$F2,$FC,$17,$9B,$E4,$E5,$EF,$4F,$3C,$FA,$B5,$F1
	.dc.b	$12,$7C,$0F,$97,$C3,$F7,$FC,$1C,$9E,$47,$BA,$3E,$1F,$C9,$48,$CE
	.dc.b	$7F,$02,$3E,$3F,$8D,$F2,$3E,$4F,$E0,$12,$BD,$CF,$C3,$F9,$2F,$9A
	.dc.b	$48,$E7,$C2,$F9,$3F,$A3,$E0,$C9,$CB,$F5,$3E,$4B,$EA,$F8,$3F,$34
	.dc.b	$93,$F2,$DF,$A1,$EB,$FD,$A2,$47,$6F,$0F,$FC,$8B,$F4,$5F,$E6,$2E
	.dc.b	$4F,$BA,$7E,$00,$CF,$CE,$EB,$E3,$FF,$C6,$B2,$69,$2F,$F8,$6F,$D2
	.dc.b	$7F,$80,$6F,$FC,$CB,$5F,$E6,$29,$FE,$50,$DE,$AF,$DC,$42,$B8,$3F
	.dc.b	$EA,$D9,$FF,$29,$BF,$39,$7C,$67,$A0,$F4,$FC,$6D,$9F,$F2,$3C,$E9
	.dc.b	$B3,$E8,$3D,$3F,$17,$F0,$09,$1D,$7E,$37,$FC,$DB,$F4,$BF,$9A,$8F
	.dc.b	$C0,$FC,$8A,$31,$1F,$E1,$F8,$C7,$2F,$28,$F6,$FC,$77,$CC,$23,$E9
	.dc.b	$F8,$FF,$9F,$F1,$F2,$3D,$8E,$9F,$F0,$A7,$D2,$FF,$AA,$5B,$E7,$FF
	.dc.b	$8A,$63,$F8,$4C,$7E,$50,$3B,$E1,$F9,$6E,$A3,$9F,$97,$C4,$E5,$56
	.dc.b	$3F,$A1,$CF,$F6,$86,$CF,$F8,$6E,$31,$1F,$E7,$6E,$0F,$F4,$6A,$F9
	.dc.b	$7E,$1E,$0A,$73,$18,$6D,$FE,$0B,$E4,$7F,$D7,$13,$9F,$90,$B3,$F8
	.dc.b	$6D,$EE,$3F,$81,$6E,$3D,$6B,$32,$FA,$4F,$E6,$F3,$FD,$71,$B1,$FE
	.dc.b	$C5,$06,$FB,$FF,$DF,$0E,$4C,$4F,$F6,$4F,$98,$7B,$6F,$A5,$C3,$6B
	.dc.b	$84,$EA,$72,$78,$7C,$1F,$F3,$1C,$FF,$B5,$71,$46,$23,$C7,$18,$F8
	.dc.b	$7E,$07,$FA,$7B,$C8,$F3,$F7,$FA,$38,$3E,$57,$C9,$FF,$5D,$7E,$7C
	.dc.b	$EB,$0F,$A5,$C3,$FA,$EA,$D3,$FE,$D4,$9D,$57,$0A,$C8,$D9,$F9,$57
	.dc.b	$F9,$03,$1F,$06,$1C,$6B,$75,$F7,$7E,$4F,$F6,$F1,$FB,$C6,$BF,$B7
	.dc.b	$6C,$6E,$BF,$D7,$36,$F8,$7F,$EE,$DD,$3F,$CE,$B9,$D8,$62,$FF,$DD
	.dc.b	$77,$B7,$F8,$04,$67,$1F,$44,$BE,$23,$19,$7A,$B2,$B3,$9C,$3F,$9A
	.dc.b	$BC,$EF,$FC,$70,$BE,$1B,$8E,$4F,$93,$E3,$18,$79,$FC,$46,$30,$E6
	.dc.b	$DC,$39,$FC,$37,$2F,$93,$F0,$B6,$BF,$84,$F9,$FF,$E4,$57,$C7,$1E
	.dc.b	$BF,$13,$F9,$F6,$1C,$27,$F8,$F7,$C6,$FC,$AD,$0D,$24,$7C,$CF,$88
	.dc.b	$F4,$C7,$C0,$6C,$7F,$27,$F8,$EF,$69,$F1,$3E,$22,$C7,$C4,$EF,$FA
	.dc.b	$FC,$3F,$A0,$8F,$AF,$18,$33,$F0,$FF,$33,$E1,$FE,$6B,$67,$F9,$A7
	.dc.b	$C7,$FC,$A3,$E8,$F5,$58,$FE,$71,$DD,$EB,$23,$FC,$D9,$F0,$FC,$27
	.dc.b	$D2,$E5,$FB,$2D,$9F,$93,$87,$DE,$FC,$DF,$F2,$D5,$FC,$C7,$95,$39
	.dc.b	$F9,$0D,$75,$7E,$4A,$FF,$01,$7F,$E5,$96,$BF,$C1,$3A,$D1,$88,$FD
	.dc.b	$23,$F8,$97,$FC,$FC,$DE,$EF,$F0,$8F,$E4,$2F,$84,$F2,$77,$FD,$BD
	.dc.b	$3F,$59,$AF,$E0,$32,$FC,$AD,$BE,$0F,$51,$EB,$F9,$0F,$31,$2B,$F8
	.dc.b	$A4,$79,$23,$D5,$E0,$FC,$13,$9F,$C1,$5A,$FF,$08,$E1,$FE,$D9,$7F
	.dc.b	$C2,$7A,$FE,$3F,$6F,$F4,$B4,$E5,$F5,$F7,$7C,$BF,$8F,$F8,$C4,$BF
	.dc.b	$0D,$A3,$E5,$78,$3F,$44,$F5,$ED,$6B,$C3,$F0,$1F,$57,$E4,$3D,$ED
	.dc.b	$8F,$1F,$81,$F4,$C8,$F6,$7F,$84,$F8,$6F,$CA,$9F,$05,$F7,$FE,$47
	.dc.b	$F4,$D1,$8B,$F6,$C6,$3C,$EF,$C2,$78,$56,$7E,$0B,$F8,$7F,$13,$F0
	.dc.b	$44,$8F,$6B,$F8,$5F,$0F,$F1,$87,$F0,$AF,$AB,$7C,$16,$2B,$FB,$77
	.dc.b	$4F,$FA,$11,$BE,$9F,$F1,$3E,$77,$D7,$7F,$09,$9C,$9F,$DD,$7E,$9B
	.dc.b	$FD,$F8,$FF,$AC,$1B,$3F,$D4,$5A,$FF,$57,$6B,$FD,$60,$2F,$86,$F0
	.dc.b	$3C,$BC,$2F,$F1,$4B,$9A,$F0,$7C,$37,$AF,$E1,$BF,$51,$E3,$D9,$FC
	.dc.b	$2B,$EF,$F9,$EC,$65,$FE,$DB,$DD,$18,$CB,$FC,$6B,$F4,$DF,$E3,$FC
	.dc.b	$3F,$C2,$1F,$5F,$F4,$F0,$56,$B2,$F8,$93,$98,$E3,$87,$F4,$6A,$61
	.dc.b	$DB,$0F,$F6,$9B,$63,$FB,$A3,$5F,$F6,$86,$3F,$E4,$09,$FC,$EF,$C0
	.dc.b	$7D,$7F,$87,$FE,$08,$FD,$9F,$86,$C4,$FF,$9D,$29,$18,$BF,$73,$B9
	.dc.b	$D7,$C2,$59,$F5,$C7,$D1,$C5,$F0,$BF,$7D,$B1,$FE,$31,$5F,$0B,$99
	.dc.b	$D3,$27,$5F,$09,$6F,$F8,$5F,$CA,$7F,$1B,$3F,$21,$56,$BD,$1F,$0B
	.dc.b	$86,$9E,$B8,$FE,$45,$EB,$48,$FF,$21,$74,$7C,$9E,$2F,$F5,$2F,$94
	.dc.b	$AF,$47,$AD,$F4,$76,$34,$98,$7D,$2D,$36,$FB,$5A,$7C,$43,$FC,$57
	.dc.b	$F4,$DF,$EF,$C8,$F0,$F2,$7E,$8E,$31,$3B,$3F,$4B,$19,$F2,$7F,$05
	.dc.b	$26,$92,$BF,$C6,$94,$F9,$5F,$F6,$54,$7F,$97,$DB,$1D,$57,$CB,$7E
	.dc.b	$CD,$CE,$BF,$E8,$C7,$35,$C6,$75,$F2,$9B,$7F,$A6,$4C,$9F,$97,$CB
	.dc.b	$70,$F5,$7E,$0C,$F8,$EF,$CD,$36,$BC,$ED,$9F,$F2,$63,$5F,$F2,$93
	.dc.b	$F9,$97,$9C,$FF,$09,$8C,$65,$FE,$A2,$79,$3F,$EA,$4F,$CF,$1F,$2B
	.dc.b	$FC,$3D,$B8,$7F,$A0,$5F,$F7,$C3,$63,$5F,$2B,$23,$5E,$BD,$CB,$EA
	.dc.b	$7E,$E5,$FC,$26,$7F,$87,$36,$3F,$CC,$EF,$F0,$B7,$7E,$27,$F1,$F8
	.dc.b	$C6,$5F,$EB,$86,$D7,$23,$5E,$37,$F7,$7F,$80,$FE,$E4,$48,$FE,$9A
	.dc.b	$FF,$91,$BC,$1F,$F2,$0C,$7F,$58,$8F,$E6,$5F,$11,$C6,$3F,$AD,$FB
	.dc.b	$9F,$E0,$13,$F9,$B7,$C3,$D8,$FD,$F3,$87,$F8,$07,$C2,$DD,$7F,$31
	.dc.b	$4F,$CF,$3F,$13,$1F,$CB,$63,$1F,$33,$FE,$58,$9C,$CF,$F8,$15,$AF
	.dc.b	$F9,$9F,$37,$C3,$7F,$94,$27,$F1,$F8,$FF,$82,$E7,$2F,$FA,$43,$73
	.dc.b	$CF,$0E,$1D,$3D,$C9,$1F,$E0,$0B,$FC,$BE,$3F,$E9,$9C,$98,$FF,$9F
	.dc.b	$F1,$E7,$3F,$E7,$9A,$DC,$62,$B7,$3D,$AF,$C4,$7F,$87,$FD,$4F,$F3
	.dc.b	$15,$6E,$DD,$7F,$E6,$A9,$CD,$7C,$18,$C7,$C2,$C8,$E7,$87,$FA,$26
	.dc.b	$9E,$96,$63,$E7,$E1,$FE,$25,$AF,$67,$C2,$FC,$D2,$3F,$E1,$EB,$FF
	.dc.b	$39,$3D,$33,$FA,$D3,$E3,$F8,$3F,$9C,$FC,$FF,$99,$95,$BB,$FE,$1F
	.dc.b	$AF,$AE,$7D,$5F,$05,$FE,$92,$9F,$F0,$D5,$77,$8C,$19,$F0,$B7,$1C
	.dc.b	$DB,$A7,$AB,$E1,$7E,$F1,$F1,$6E,$D7,$87,$DC,$EC,$47,$D8,$F8,$7D
	.dc.b	$AF,$99,$ED,$FC,$D6,$CF,$A7,$D8,$FA,$07,$D1,$F4,$6C,$7F,$BF,$72
	.dc.b	$FF,$61,$0F,$A7,$FF,$4B,$3F,$E0,$5F,$33,$F0,$71,$88,$FF,$76,$9B
	.dc.b	$F8,$CF,$F7,$14,$E9,$FF,$70,$7C,$6F,$FB,$93,$C1,$E1,$5F,$D1,$30
	.dc.b	$C5,$6A,$13,$AB,$FC,$8B,$C2,$17,$FE,$EE,$C3,$58,$5C,$39,$46,$2B
	.dc.b	$C8,$7F,$BD,$6B,$BB,$EB,$AA,$FA,$AF,$81,$DE,$73,$1E,$73,$CB,$7E
	.dc.b	$79,$DB,$E9,$E6,$F4,$CF,$73,$FD,$C1,$39,$22,$1F,$24,$DB,$FC,$3B
	.dc.b	$1C,$0D,$46,$79,$7F,$D1,$55,$FE,$37,$8C,$4E,$CE,$75,$CA,$FC,$71
	.dc.b	$8B,$F7,$D3,$7F,$F2,$06,$1F,$6E,$31,$5A,$C3,$C7,$39,$6E,$3E,$1B
	.dc.b	$E2,$DA,$E1,$84,$7E,$89,$EE,$9F,$F2,$76,$9D,$DB,$5E,$BB,$F8,$33
	.dc.b	$DE,$FC,$F5,$DA,$FE,$1B,$5F,$61,$F5,$0F,$8F,$5C,$AF,$C9,$1E,$2C
	.dc.b	$7D,$AD,$9D,$3F,$C3,$72,$FF,$31,$61,$FC,$77,$85,$18,$D7,$F8,$E3
	.dc.b	$6B,$F8,$A7,$9B,$61,$5F,$42,$FF,$93,$7E,$38,$F0,$34,$DC,$3F,$86
	.dc.b	$6F,$8F,$03,$FC,$27,$8A,$BF,$C1,$F5,$FE,$B6,$6B,$FC,$DD,$AF,$F4
	.dc.b	$DE,$CF,$EF,$2F,$F1,$D6,$BF,$C9,$BC,$AF,$0F,$81,$E2,$EC,$F1,$F5
	.dc.b	$4A,$FE,$49,$82,$71,$1D,$E7,$CB,$3F,$EC,$D8,$F2,$47,$A3,$07,$95
	.dc.b	$F0,$DB,$3E,$38,$F8,$38,$FC,$2F,$2B,$44,$9F,$0B,$5F,$EA,$94,$8E
	.dc.b	$C4,$8F,$89,$F3,$5F,$30,$8F,$AB,$E4,$6B,$FD,$1B,$E6,$FC,$86,$38
	.dc.b	$BE,$AF,$C5,$79,$EE,$57,$C2,$8A,$F8,$A3,$DC,$47,$EA,$7C,$3F,$BF
	.dc.b	$09,$1F,$2A,$2B,$EA,$8F,$8C,$47,$EE,$79,$FF,$84,$23,$F4,$BE,$1B
	.dc.b	$E5,$7C,$DC,$8F,$B8,$E9,$FF,$07,$F0,$FF,$76,$7F,$B1,$E5,$F4,$4F
	.dc.b	$C7,$F5,$58,$FF,$00,$47,$D7,$23,$FE,$62,$FD,$D5,$F1,$9F,$9D,$84
	.dc.b	$61,$FE,$E6,$62,$35,$09,$BA,$4A,$DC,$FF,$BB,$9B,$5D,$7E,$4F,$FD
	.dc.b	$4F,$1D,$1A,$FF,$9E,$DB,$BF,$B5,$7B,$5C,$7E,$6E,$57,$17,$A7,$C2
	.dc.b	$7A,$7E,$8B,$F1,$13,$E8,$5C,$99,$35,$CB,$67,$FD,$07,$3A,$6C,$FF
	.dc.b	$A1,$94,$65,$FC,$AE,$7F,$9F,$23,$F1,$BF,$5C,$7C,$87,$CB,$FA,$37
	.dc.b	$C3,$52,$57,$CC,$87,$C3,$EF,$E7,$24,$E5,$DB,$FE,$9F,$E8,$FF,$11
	.dc.b	$BF,$E9,$A6,$BF,$EF,$B3,$FA,$73,$87,$0D,$7F,$E0,$02,$F5,$7F,$01
	.dc.b	$F8,$53,$B6,$9F,$83,$9D,$57,$E0,$9F,$EB,$0D,$F3,$74,$6F,$CB,$B1
	.dc.b	$E5,$8F,$95,$3F,$CC,$A3,$C7,$1F,$C2,$7A,$7F,$84,$12,$3F,$1F,$F4
	.dc.b	$8F,$5B,$F9,$77,$E3,$3F,$EA,$59,$FF,$1D,$3F,$84,$6F,$4E,$BE,$13
	.dc.b	$F8,$97,$C9,$49,$FE,$48,$F2,$9F,$CF,$7D,$2B,$DB,$A4,$AF,$CF,$C7
	.dc.b	$F9,$61,$BE,$5F,$02,$3A,$9F,$2F,$C6,$C8,$F2,$BE,$8F,$88,$C4,$7F
	.dc.b	$A1,$B3,$FF,$7F,$BE,$AF,$8E,$CF,$87,$FE,$6D,$9C,$CF,$F9,$7B,$8B
	.dc.b	$F9,$61,$F0,$AB,$D3,$C9,$F9,$5D,$3E,$C5,$7F,$B7,$1A,$FE,$0C,$38
	.dc.b	$7F,$0C,$BC,$AF,$E0,$1F,$03,$F8,$AC,$BF,$86,$37,$C8,$FF,$0C,$5F
	.dc.b	$2A,$FA,$7D,$5F,$53,$B6,$4F,$D4,$8A,$FF,$9D,$BC,$BF,$E8,$86,$31
	.dc.b	$FE,$31,$97,$F0,$D9,$FF,$7D,$3F,$C1,$DB,$D1,$F5,$BC,$1F,$AE,$FB
	.dc.b	$C9,$1E,$F7,$F6,$DF,$80,$7F,$13,$8F,$E2,$ED,$57,$FC,$9E,$30,$FF
	.dc.b	$29,$FC,$05,$F2,$DF,$D6,$7E,$26,$5F,$F6,$B7,$F2,$3F,$13,$FD,$77
	.dc.b	$83,$FB,$01,$27,$FD,$58,$FC,$F5,$F3,$0F,$CA,$BF,$C0,$93,$FE,$41
	.dc.b	$F6,$3E,$86,$3F,$E0,$9E,$3F,$0E,$1F,$EA,$0C,$38,$7B,$5F,$E0,$F7
	.dc.b	$FC,$71,$ED,$EF,$E5,$E5,$6F,$C5,$FF,$06,$3F,$ED,$2C,$1F,$2D,$FE
	.dc.b	$0C,$6C,$7E,$0F,$3F,$C0,$4E,$4F,$F0,$8F,$6F,$FB,$42,$7F,$D4,$D5
	.dc.b	$FA,$F7,$A7,$89,$FB,$2F,$8B,$F4,$78,$92,$76,$FF,$88,$FE,$5D,$F3
	.dc.b	$DE,$DF,$53,$C4,$4A,$F1,$63,$FD,$01,$FF,$73,$53,$E7,$BE,$27,$AD
	.dc.b	$27,$FD,$71,$39,$7C,$BE,$37,$C9,$23,$AF,$9D,$F9,$9E,$37,$C1,$24
	.dc.b	$78,$1F,$AD,$CB,$FE,$63,$F2,$FE,$77,$21,$A7,$AF,$E3,$7E,$B7,$95
	.dc.b	$30,$F8,$AF,$BF,$E3,$7E,$D7,$A7,$F3,$C9,$3E,$E7,$BF,$E4,$DF,$7A
	.dc.b	$3E,$B7,$95,$F8,$13,$EF,$46,$7C,$4F,$F7,$77,$07,$F9,$61,$FF,$78
	.dc.b	$71,$BC,$AF,$C4,$1F,$3F,$6F,$C5,$7D,$17,$DF,$26,$5F,$39,$FC,$17
	.dc.b	$89,$F8,$F2,$47,$CC,$F9,$9F,$B1,$F1,$3F,$67,$C0,$93,$F4,$23,$FC
	.dc.b	$BD,$8F,$E3,$A3,$F8,$1C,$F8,$FC,$2F,$F2,$CC,$E6,$7F,$3C,$D7,$FC
	.dc.b	$D7,$9F,$FB,$27,$4F,$5F,$C5,$FE,$84,$EB,$F1,$3D,$BD,$3F,$D2,$7C
	.dc.b	$A4,$C3,$FD,$F1,$8F,$F3,$77,$F2,$B1,$F9,$D2,$3F,$8C,$7E,$7F,$1F
	.dc.b	$A3,$E2,$64,$FE,$B7,$93,$FB,$9B,$D2,$3F,$65,$F0,$52,$30,$78,$3E
	.dc.b	$0B,$E0,$FC,$4C,$8C,$1E,$47,$F7,$5F,$29,$23,$07,$A3,$D8,$F9,$3C
	.dc.b	$32,$30,$7B,$1F,$E0,$5E,$44,$8C,$1F,$B2,$7F,$74,$47,$F5,$2F,$D0
	.dc.b	$F8,$6F,$E6,$BF,$1E,$E8,$FA,$31,$86,$D7,$FB,$46,$74,$FF,$A7,$B8
	.dc.b	$5F,$13,$B8,$C5,$DE,$8F,$E8,$DF,$21,$73,$39,$C3,$F9,$76,$3D,$30
	.dc.b	$7D,$1F,$EA,$F0,$FE,$48,$D8,$FE,$A8,$D7,$FC,$30,$D7,$FD,$BD,$F1
	.dc.b	$EF,$A4,$FE,$59,$F2,$3F,$EA,$E6,$C6,$67,$FA,$F4,$E5,$EB,$F0,$3E
	.dc.b	$66,$E7,$E2,$8C,$47,$81,$AF,$F3,$5D,$3F,$DE,$DE,$C7,$C7,$E0,$C4
	.dc.b	$A3,$0F,$FB,$1B,$E8,$FF,$8F,$C6,$35,$FE,$D9,$3B,$1B,$7D,$08,$57
	.dc.b	$09,$CD,$F6,$BE,$53,$9B,$E1,$87,$92,$F6,$78,$7E,$5F,$FD,$04,$FF
	.dc.b	$4D,$A4,$62,$35,$7F,$D0,$67,$2D,$F4,$7C,$DE,$0F,$9F,$3F,$2E,$18
	.dc.b	$AE,$95,$70,$E6,$7F,$6C,$8F,$07,$CB,$E6,$FD,$B5,$71,$4D,$8C,$7A
	.dc.b	$19,$F8,$AB,$83,$1E,$87,$CD,$8E,$5F,$0F,$A8,$CE,$5F,$2A,$31,$A7
	.dc.b	$CF,$F4,$7F,$D2,$98,$F8,$7E,$8B,$D1,$18,$CB,$B5,$7B,$F5,$57,$9F
	.dc.b	$D1,$DF,$53,$1C,$61,$FD,$D6,$30,$F2,$3E,$87,$C6,$FC,$FD,$3E,$EB
	.dc.b	$71,$FF,$41,$CF,$D1,$8C,$7A,$3F,$15,$F8,$CD,$8E,$9B,$7E,$37,$E2
	.dc.b	$FE,$51,$F2,$FA,$AC,$EB,$1F,$C5,$FA,$1D,$C9,$1F,$7D,$F1,$7D,$07
	.dc.b	$D3,$F9,$9F,$A8,$D9,$E5,$A7,$F0,$96,$F5,$7D,$9F,$99,$FB,$4F,$DB
	.dc.b	$F4,$7D,$84,$9F,$C6,$3E,$FC,$BF,$01,$F0,$5F,$81,$6C,$F9,$FE,$8D
	.dc.b	$E9,$FC,$ED,$3F,$11,$18,$9F,$9D,$18,$F4,$BE,$64,$9A,$49,$FF,$0E
	.dc.b	$47,$F9,$ED,$ED,$E9,$74,$F9,$9E,$AF,$6F,$42,$4F,$8F,$E3,$EF,$2F
	.dc.b	$E3,$30,$7C,$CF,$80,$95,$D9,$F1,$FA,$3E,$83,$D3,$E8,$D9,$F7,$7A
	.dc.b	$1E,$CF,$96,$F6,$B6,$3E,$17,$91,$EC,$23,$EA,$7D,$FD,$DF,$10,$F8
	.dc.b	$BE,$85,$9F,$95,$F1,$BF,$23,$E3,$BE,$4B,$63,$E7,$7B,$FF,$20,$8F
	.dc.b	$C6,$7E,$5F,$9D,$FA,$4F,$A3,$57,$D2,$8C,$3F,$53,$E4,$75,$16,$FF
	.dc.b	$CA,$D4,$8F,$B3,$F1,$DF,$58,$93,$F5,$1F,$D2,$7C,$97,$F9,$BA,$3F
	.dc.b	$30,$DE,$4F,$A1,$1F,$C2,$63,$1E,$5F,$F2,$99,$7D,$13,$DF,$FE,$25
	.dc.b	$18,$6F,$D1,$B4,$CD,$6F,$1F,$D5,$DA,$99,$7F,$25,$8C,$36,$33,$3F
	.dc.b	$E7,$B6,$CF,$FA,$AF,$F1,$0F,$CB,$3F,$C5,$8E,$33,$DA,$33,$97,$A6
	.dc.b	$B8,$BF,$CB,$73,$FF,$3A,$D7,$C2,$F9,$0F,$81,$F2,$1F,$05,$F8,$5F
	.dc.b	$1F,$89,$F8,$ED,$CB,$C1,$F4,$7C,$9F,$F5,$F3,$5F,$F5,$B8,$70,$FF
	.dc.b	$AE,$9A,$FF,$2F,$51,$F3,$1E,$98,$FD,$17,$35,$F2,$9C,$BF,$E9,$BC
	.dc.b	$5F,$1D,$F6,$7E,$43,$ED,$36,$7E,$FB,$87,$F0,$3F,$1D,$72,$FE,$77
	.dc.b	$68,$FC,$11,$F8,$3F,$8F,$B5,$E1,$F8,$EF,$31,$E7,$F8,$EB,$1E,$57
	.dc.b	$0F,$E3,$BE,$3E,$E3,$DA,$BF,$1C,$7E,$3F,$E2,$ED,$7E,$51,$C3,$F9
	.dc.b	$6B,$B5,$F7,$F9,$7E,$0B,$FD,$5D,$AF,$CF,$F8,$3F,$C8,$F1,$FC,$B8
	.dc.b	$FC,$77,$07,$F7,$97,$EA,$B5,$FA,$87,$E4,$FE,$D7,$FD,$25,$8F,$CF
	.dc.b	$8F,$CF,$C6,$1F,$C0,$9F,$A2,$FA,$47,$BB,$F9,$B6,$D7,$FB,$4D,$B3
	.dc.b	$F8,$40,$E1,$FC,$3B,$FD,$59,$FC,$E3,$C3,$93,$F6,$4E,$5B,$5F,$B2
	.dc.b	$F4,$FF,$9A,$5F,$EB,$9B,$7F,$A5,$5E,$76,$FC,$FE,$0C,$AF,$DA,$FD
	.dc.b	$AF,$CA,$AD,$C7,$EE,$37,$FC,$23,$79,$7F,$80,$FD,$2F,$E5,$32,$FA
	.dc.b	$91,$8E,$9F,$DE,$49,$A4,$AF,$F7,$B6,$3D,$2F,$F7,$55,$B9,$F1,$DF
	.dc.b	$37,$F8,$4F,$D2,$BC,$CF,$E2,$2D,$8C,$F9,$B8,$9C,$B2,$3F,$8F,$FC
	.dc.b	$65,$C7,$F8,$47,$D2,$FF,$04,$9F,$F1,$04,$23,$0F,$F8,$8D,$FF,$28
	.dc.b	$47,$F1,$A7,$31,$B8,$FE,$47,$1A,$8F,$D5,$FC,$A6,$27,$F9,$67,$CA
	.dc.b	$7F,$32,$23,$F8,$3F,$65,$E0,$27,$F9,$BF,$91,$C2,$BF,$9B,$B7,$0F
	.dc.b	$DB,$FC,$A7,$F3,$73,$F9,$C7,$C9,$5A,$FE,$7A,$E1,$FE,$7F,$F2,$76
	.dc.b	$BF,$7F,$F2,$7F,$D1,$09,$3E,$3C,$91,$FD,$23,$C0,$FE,$88,$FF,$45
	.dc.b	$6C,$FF,$4A,$AF,$E9,$A4,$79,$3F,$E3,$3F,$84,$FF,$37,$76,$FF,$31
	.dc.b	$8F,$F2,$B4,$62,$7F,$8D,$46,$3E,$3B,$3F,$1F,$67,$FC,$5D,$5F,$E6
	.dc.b	$8A,$FF,$28,$BF,$E5,$C8,$CD,$74,$F8,$FE,$15,$AF,$8F,$F1,$12,$BE
	.dc.b	$85,$A3,$CD,$F1,$9E,$47,$CA,$D9,$F0,$FC,$6F,$A0,$F4,$FC,$7D,$8F
	.dc.b	$2F,$89,$EF,$23,$D9,$F8,$5F,$1D,$F0,$1F,$82,$D8,$F6,$E3,$F1,$0F
	.dc.b	$8B,$F1,$D6,$3E,$0F,$C7,$7C,$B2,$3E,$A7,$E7,$7C,$3F,$CC,$F8,$EF
	.dc.b	$9A,$D9,$F9,$3F,$19,$F4,$0F,$A3,$F1,$D6,$3F,$BA,$F0,$7F,$B5,$47
	.dc.b	$D2,$FF,$7F,$7F,$4B,$CD,$FE,$AB,$8C,$47,$FB,$80,$DF,$67,$F6,$09
	.dc.b	$D3,$FC,$7F,$E4,$7F,$D1,$9F,$1D,$C2,$BF,$DB,$38,$62,$B5,$09,$D5
	.dc.b	$FF,$AE,$5E,$11,$FE,$BA,$3B,$9E,$4A,$FF,$4C,$61,$AC,$2E,$1E,$39
	.dc.b	$B5,$FD,$D2,$7C,$0F,$13,$C1,$3E,$2B,$F0,$D6,$A7,$DD,$18,$76,$73
	.dc.b	$29,$FE,$A9,$C3,$D1,$5E,$CF,$03,$D9,$3B,$6B,$F2,$5A,$FD,$17,$3F
	.dc.b	$24,$DD,$D7,$FE,$B5,$9C,$C6,$29,$18,$9D,$63,$FC,$72,$F9,$46,$2B
	.dc.b	$FD,$7D,$18,$7D,$2B,$83,$F9,$1D,$7D,$E8,$C4,$FC,$37,$84,$67,$0F
	.dc.b	$13,$BA,$E3,$8F,$85,$B7,$E0,$AD,$CF,$95,$F5,$6B,$C6,$7E,$B8,$C3
	.dc.b	$C2,$FE,$00,$C4,$F0,$AE,$58,$F9,$70,$EF,$5C,$5B,$7E,$8A,$DC,$F3
	.dc.b	$8E,$B8,$78,$63,$18,$75,$7E,$6F,$13,$CD,$FE,$8F,$97,$FB,$DE,$31
	.dc.b	$3F,$C5,$A7,$3B,$7D,$79,$7C,$86,$2B,$9E,$D4,$4F,$E1,$30,$E9,$B6
	.dc.b	$63,$EE,$D7,$FB,$3F,$E9,$3F,$D7,$2F,$F4,$96,$BF,$E5,$0B,$46,$23
	.dc.b	$FA,$EC,$63,$E4,$7F,$C2,$51,$F8,$89,$FF,$02,$BD,$7D,$1F,$24,$62
	.dc.b	$3F,$DF,$EF,$FB,$05,$F1,$7C,$8E,$BE,$42,$46,$0F,$33,$DB,$0F,$1B
	.dc.b	$68,$D3,$1F,$1F,$E0,$25,$7B,$FE,$47,$C7,$E7,$F9,$3C,$F9,$3F,$0E
	.dc.b	$2B,$F0,$70,$4F,$72,$47,$C7,$F8,$EF,$69,$1F,$99,$F2,$15,$F7,$FC
	.dc.b	$87,$C2,$31,$C4,$E4,$FD,$96,$2F,$8F,$F8,$FF,$21,$30,$FA,$D1,$5F
	.dc.b	$24,$7C,$12,$3F,$67,$E4,$3E,$7F,$C7,$C8,$FB,$D1,$5F,$64,$7C,$92
	.dc.b	$3F,$81,$F8,$EF,$AE,$47,$F0,$DF,$21,$F6,$BE,$3E,$47,$E2,$5D,$3F
	.dc.b	$73,$E3,$FE,$FE,$92,$3F,$1D,$15,$F9,$91,$F6,$48,$FE,$2B,$E9,$19
	.dc.b	$FE,$76,$11,$87,$FD,$82,$D7,$F8,$DC,$FF,$1B,$F9,$A7,$CD,$F1,$A6
	.dc.b	$B4,$97,$CB,$0E,$DE,$BF,$F9,$36,$3F,$D9,$AD,$7F,$D8,$8D,$E1,$FB
	.dc.b	$11,$FC,$9B,$17,$C8,$72,$F8,$F9,$3C,$9E,$BF,$09,$EB,$F8,$A7,$B9
	.dc.b	$ED,$FA,$4F,$C6,$DF,$FB,$71,$AF,$FB,$4E,$76,$E7,$AF,$F5,$EA,$76
	.dc.b	$FF,$B1,$9A,$FF,$94,$9A,$87,$D2,$3D,$4F,$9C,$D8,$FF,$76,$B5,$FF
	.dc.b	$72,$4B,$B3,$EA,$7D,$23,$E3,$A9,$2B,$FD,$EB,$E3,$3D,$8D,$FC,$7C
	.dc.b	$9F,$F7,$CD,$A3,$84,$7F,$BC,$7B,$5E,$EF,$E0,$57,$BD,$4C,$45,$9F
	.dc.b	$E2,$33,$A6,$CF,$C6,$8C,$47,$FA,$9F,$E8,$DE,$F7,$5F,$4A,$4F,$C5
	.dc.b	$E9,$FB,$AA,$B6,$E7,$BF,$FC,$FB,$87,$F6,$B8,$F0,$DF,$72,$3F,$C9
	.dc.b	$2B,$E1,$3C,$4F,$A4,$E1,$FC,$C9,$8D,$7E,$07,$AD,$FC,$EF,$AF,$FD
	.dc.b	$F7,$3C,$7C,$D4,$62,$38,$C6,$3E,$53,$87,$4F,$1E,$B9,$5C,$6F,$F3
	.dc.b	$5A,$FE,$9E,$FF,$5D,$F4,$7F,$CD,$77,$F8,$2D,$9F,$F4,$DB,$6F,$F4
	.dc.b	$1D,$BC,$3D,$09,$1E,$D1,$28,$EB,$CF,$EB,$F8,$B9,$5E,$37,$DD,$F2
	.dc.b	$BE,$FF,$94,$93,$E7,$76,$FB,$B5,$47,$C5,$F9,$59,$3B,$3C,$59,$7E
	.dc.b	$5D,$BE,$7F,$B8,$F7,$F7,$58,$FE,$B3,$29,$FD,$6B,$63,$31,$FD,$47
	.dc.b	$0F,$D0,$47,$CB,$F8,$DF,$AC,$7D,$7E,$FB,$1F,$3F,$BB,$ED,$18,$7E
	.dc.b	$F8,$FA,$A7,$C6,$FD,$2F,$2A,$4F,$CB,$9C,$BE,$C8,$F2,$91,$F4,$BF
	.dc.b	$C1,$3E,$8F,$F9,$0C,$7D,$36,$BE,$06,$F3,$7D,$EF,$47,$F1,$DF,$1D
	.dc.b	$2B,$F0,$CF,$E4,$71,$FC,$97,$C0,$FC,$12,$4F,$FB,$3F,$6F,$F3,$5B
	.dc.b	$FC,$8F,$E3,$BF,$C2,$9F,$3A,$FA,$3F,$F0,$97,$67,$F1,$18,$C7,$C0
	.dc.b	$7E,$76,$BF,$3D,$E8,$F0,$F0,$FF,$C2,$3B,$FF,$2E,$BF,$F2,$57,$BF
	.dc.b	$AC,$FF,$35,$E8,$FD,$32,$3E,$37,$EB,$7C,$2F,$EA,$7E,$3E,$4F,$E8
	.dc.b	$63,$FC,$4D,$F4,$4F,$A3,$A4,$9F,$D3,$1A,$F9,$1F,$F0,$84,$9A,$48
	.dc.b	$FF,$78,$47,$F0,$EC,$FF,$A2,$F9,$48,$FE,$37,$3F,$B9,$6B,$FB,$87
	.dc.b	$87,$0F,$F9,$CD,$FE,$1F,$19,$F3,$FC,$E7,$A3,$E4,$7D,$3E,$6F,$90
	.dc.b	$93,$E2,$7C,$FF,$21,$EE,$24,$78,$7E,$47,$E0,$FC,$8C,$8F,$23,$F1
	.dc.b	$3D,$9F,$8B,$F1,$3F,$00,$93,$EE,$7E,$1F,$A9,$F7,$BE,$8F,$F1,$7B
	.dc.b	$7D,$D6,$D7,$FA,$2B,$6F,$E4,$CF,$BF,$E4,$BE,$8F,$4C,$8F,$89,$1F
	.dc.b	$53,$E4,$7E,$79,$23,$E6,$BF,$4B,$E3,$7E,$C9,$23,$E6,$7C,$97,$DC
	.dc.b	$EF,$FB,$BF,$1F,$27,$E8,$47,$E0,$39,$BE,$D9,$23,$EC,$BF,$E4,$DF
	.dc.b	$33,$FC,$A9,$1F,$CB,$1B,$D1,$F7,$1B,$BF,$FA,$A5,$FC,$BB,$7D,$4F
	.dc.b	$F3,$94,$7F,$21,$AF,$E7,$0D,$7A,$46,$BD,$5F,$01,$23,$FB,$B1,$FE
	.dc.b	$6E,$9E,$0D,$F8,$FF,$A1,$E7,$48,$EB,$15,$F4,$7C,$77,$88,$C6,$5E
	.dc.b	$9F,$8F,$F2,$FC,$24,$AF,$2C,$57,$E0,$7C,$7F,$A8,$C5,$7C,$1F,$5B
	.dc.b	$DB,$EA,$CA,$F6,$C5,$7E,$47,$C7,$FC,$23,$15,$F2,$5F,$9C,$48,$F7
	.dc.b	$9F,$2D,$FA,$5F,$1D,$F0,$BE,$4B,$FD,$3F,$3F,$ED,$4D,$FF,$C8,$11
	.dc.b	$8E,$34,$63,$9F,$EA,$5F,$17,$EB,$7C,$55,$C3,$93,$6F,$8B,$6F,$FC
	.dc.b	$E5,$3F,$E2,$7C,$3C,$2B,$E8,$DF,$21,$FE,$AA,$8F,$4C,$63,$0F,$F5
	.dc.b	$1B,$63,$DA,$71,$9E,$0F,$FB,$AF,$C1,$7D,$1F,$FA,$AB,$E5,$3F,$D7
	.dc.b	$2D,$8C,$CF,$29,$CB,$D2,$DF,$10,$F3,$BC,$2D,$9D,$35,$F3,$B5,$FF
	.dc.b	$22,$65,$FE,$91,$F8,$EF,$F2,$EE,$9F,$37,$83,$FC,$BB,$DB,$F0,$F8
	.dc.b	$77,$0F,$86,$DC,$62,$EE,$2C,$7F,$D3,$95,$87,$B6,$1C,$E7,$58,$73
	.dc.b	$8C,$4E,$A5,$8F,$DF,$E1,$FF,$3D,$D7,$CC,$AF,$C0,$FB,$DF,$E3,$8B
	.dc.b	$FD,$A3,$5F,$C4,$36,$3F,$D5,$DF,$0F,$FE,$7A,$E6,$FD,$E4,$DB,$F6
	.dc.b	$CD,$BE,$31,$FE,$31,$C9,$F0,$F9,$C7,$9C,$E6,$31,$2B,$D6,$4F,$7E
	.dc.b	$9A,$F8,$7B,$87,$57,$0F,$7F,$86,$B9,$FF,$13,$E8,$E6,$76,$F8,$4B
	.dc.b	$5E,$57,$0F,$F9,$3F,$DB,$7E,$49,$FD,$7E,$34,$7F,$14,$E6,$F8,$9A
	.dc.b	$7E,$2A,$3E,$63,$71,$FD,$AF,$07,$EB,$AB,$E5,$6F,$FB,$0B,$E4,$7C
	.dc.b	$5B,$F7,$CF,$C0,$F9,$39,$3F,$27,$E0,$AE,$3F,$E5,$8C,$1F,$09,$D0
	.dc.b	$95,$F7,$63,$2D,$F8,$0F,$9B,$F0,$3F,$39,$B1,$F5,$7D,$FF,$9F,$F2
	.dc.b	$7F,$49,$B1,$FC,$C8,$F7,$90,$F7,$9F,$85,$FC,$7F,$FD,$92,$FF,$9B
	.dc.b	$DB,$97,$F9,$BB,$F6,$DF,$F0,$14,$E7,$E5,$7F,$05,$5F,$A6,$AF,$F3
	.dc.b	$9B,$7B,$3F,$C0,$77,$FE,$9A,$BE,$9D,$FC,$3E,$3B,$C0,$48,$FC,$E5
	.dc.b	$A3,$F3,$15,$FE,$98,$9F,$27,$81,$F9,$18,$C6,$3E,$08,$FC,$C7,$3F
	.dc.b	$E0,$3D,$5B,$D0,$78,$7B,$F8,$7D,$3D,$FF,$04,$99,$78,$9F,$0F,$BD
	.dc.b	$EE,$F0,$7C,$0D,$9F,$8D,$F1,$DF,$08,$F2,$7B,$76,$3E,$57,$85,$F3
	.dc.b	$48,$FB,$9F,$A1,$F1,$DF,$28,$F6,$65,$67,$FD,$41,$C0,$F1,$FE,$7F
	.dc.b	$89,$F0,$5B,$3F,$5F,$C3,$FB,$44,$7E,$5B,$F7,$7D,$6F,$AE,$7C,$6F
	.dc.b	$DD,$9E,$86,$2B,$F0,$1E,$2F,$DC,$35,$E8,$58,$FC,$27,$C6,$7E,$1C
	.dc.b	$8F,$D8,$7F,$15,$E2,$7E,$10,$FE,$0D,$F8,$3F,$FA,$8E,$7F,$19,$1F
	.dc.b	$88,$F1,$A4,$7F,$AF,$65,$EC,$E3,$B7,$FD,$0F,$F4,$8A,$FE,$39,$08
	.dc.b	$C4,$7F,$4E,$8C,$3A,$8C,$C7,$F5,$2F,$A4,$FC,$DF,$87,$FE,$DE,$9C
	.dc.b	$BF,$EE,$08,$CD,$7F,$84,$A7,$2F,$F8,$52,$3F,$38,$FF,$85,$BE,$0B
	.dc.b	$F8,$E3,$FE,$C1,$8C,$D8,$FF,$31,$79,$FF,$CA,$D4,$7E,$76,$7D,$0D
	.dc.b	$8F,$37,$87,$EE,$8C,$47,$F8,$93,$CE,$68,$F6,$BF,$63,$E6,$7C,$E6
	.dc.b	$E3,$F2,$FD,$BC,$34,$F9,$D1,$F5,$3D,$9F,$DC,$F9,$D6,$7F,$80,$37
	.dc.b	$2F,$F7,$6E,$2A,$FB,$7E,$D6,$78,$7F,$CF,$F3,$FE,$C9,$C3,$ED,$4F
	.dc.b	$E0,$3A,$37,$1F,$77,$D0,$FC,$19,$FB,$6E,$6F,$E9,$71,$93,$F9,$C3
	.dc.b	$8C,$98,$FF,$A8,$E3,$0F,$83,$A3,$97,$8D,$F8,$EF,$52,$61,$F8,$BE
	.dc.b	$5B,$8F,$A9,$FC,$99,$F9,$4F,$4E,$CF,$F1,$FF,$8C,$FE,$36,$7D,$7F
	.dc.b	$A1,$FE,$E9,$7F,$C7,$23,$ED,$7D,$1F,$F1,$B8,$FF,$0E,$B5,$FF,$1D
	.dc.b	$B5,$FF,$2D,$B5,$FC,$22,$F8,$BE,$7F,$2D,$D1,$B8,$7E,$8F,$D8,$FE
	.dc.b	$5B,$EB,$FE,$A9,$B1,$FA,$73,$05,$38,$1F,$91,$E0,$FD,$46,$3F,$93
	.dc.b	$F8,$57,$C0,$FE,$4D,$B5,$B3,$FC,$E2,$BA,$F5,$3B,$64,$EE,$BA,$74
	.dc.b	$E1,$FE,$83,$EE,$FF,$44,$3F,$C0,$DF,$83,$FE,$B2,$3F,$C4,$9E,$CF
	.dc.b	$F8,$DE,$F9,$43,$E8,$BF,$B6,$3F,$E4,$86,$BF,$E6,$66,$BF,$E8,$66
	.dc.b	$BF,$AF,$5E,$47,$EB,$DE,$7A,$FC,$29,$ED,$39,$F2,$7F,$69,$F4,$5F
	.dc.b	$BE,$9F,$DC,$35,$70,$FA,$54,$66,$30,$4F,$73,$F7,$CF,$BF,$C5,$6B
	.dc.b	$FC,$11,$B3,$FA,$1F,$17,$E2,$B7,$5F,$D0,$C7,$C8,$F1,$F1,$97,$F6
	.dc.b	$C3,$FB,$B1,$27,$FB,$3B,$FE,$3D,$C3,$F2,$FA,$FF,$BC,$B7,$FE,$25
	.dc.b	$5F,$5F,$2F,$1F,$B9,$FD,$E3,$0F,$F1,$D7,$E2,$1F,$E2,$57,$F5,$F5
	.dc.b	$FE,$64,$8C,$61,$FE,$FC,$C7,$F6,$33,$9F,$93,$FB,$BE,$0C,$8E,$D2
	.dc.b	$F5,$F1,$8F,$43,$FC,$C9,$8F,$F8,$29,$FF,$16,$B5,$FE,$3C,$D7,$C5
	.dc.b	$F2,$7E,$DF,$63,$CE,$FE,$31,$B8,$7B,$69,$F3,$7F,$E5,$F6,$C7,$F9
	.dc.b	$9D,$B7,$E0,$9F,$91,$EA,$FC,$27,$F2,$ED,$AF,$5F,$37,$C6,$F0,$3F
	.dc.b	$36,$D8,$F7,$F4,$61,$A7,$03,$C9,$E0,$78,$C8,$F8,$DF,$B1,$E9,$FD
	.dc.b	$27,$F4,$ED,$8F,$67,$CA,$7D,$63,$F5,$7E,$25,$8F,$9F,$E7,$FE,$04
	.dc.b	$8F,$FA,$17,$89,$1F,$85,$EC,$FD,$9B,$FB,$46,$CF,$C9,$F9,$4F,$C1
	.dc.b	$9F,$E7,$DF,$92,$FC,$0C,$7E,$0E,$3F,$77,$C3,$F8,$F2,$47,$77,$FD
	.dc.b	$1D,$3F,$3B,$A3,$F8,$8F,$52,$3F,$01,$18,$75,$5F,$EB,$27,$2D,$DB
	.dc.b	$C5,$C3,$E3,$6C,$7F,$08,$C3,$FD,$AC,$3E,$85,$C1,$FE,$EC,$F0,$BF
	.dc.b	$90,$C6,$27,$FD,$DF,$F1,$1F,$DD,$63,$11,$FD,$DE,$77,$5F,$ED,$8A
	.dc.b	$FE,$F5,$F2,$1A,$AF,$E3,$01,$C3,$FE,$B6,$31,$3E,$1C,$38,$BC,$E3
	.dc.b	$CC,$E5,$A7,$90,$A7,$8E,$3C,$D1,$C6,$7C,$57,$FC,$F1,$F0,$DF,$6B
	.dc.b	$FF,$08,$1E,$68,$CB,$E4,$33,$3C,$6B,$A1,$E3,$9D,$EB,$ED,$AF,$23
	.dc.b	$5E,$B7,$FE,$E8,$9F,$64,$E7,$E2,$FE,$F6,$D9,$BF,$F1,$14,$61,$FF
	.dc.b	$72,$57,$9A,$31,$5C,$76,$FE,$FF,$5A,$38,$B8,$AF,$AF,$4F,$81,$F7
	.dc.b	$C6,$1B,$97,$5A,$E0,$DC,$7E,$36,$0C,$39,$38,$6B,$F1,$CD,$57,$27
	.dc.b	$C7,$3F,$51,$F6,$DF,$D0,$D7,$C5,$19,$38,$CD,$BF,$85,$8E,$A7,$B4
	.dc.b	$78,$23,$DB,$87,$79,$E2,$7E,$12,$7C,$B7,$C4,$DD,$F8,$F2,$FD,$77
	.dc.b	$07,$E0,$A3,$F5,$DF,$29,$77,$FD,$73,$71,$5F,$07,$FA,$36,$4F,$86
	.dc.b	$D4,$FE,$E7,$E1,$6C,$9B,$38,$F8,$DF,$F3,$95,$7E,$FD,$B4,$69,$AE
	.dc.b	$E7,$CA,$F6,$6E,$3F,$CE,$E3,$BD,$78,$3E,$53,$C3,$F2,$92,$BC,$F1
	.dc.b	$5F,$F7,$BC,$A7,$C2,$48,$FE,$85,$97,$72,$3F,$03,$E2,$3D,$BE,$97
	.dc.b	$8F,$69,$3F,$12,$2B,$FD,$62,$3D,$44,$7F,$62,$7F,$A0,$75,$7E,$EE
	.dc.b	$F9,$7F,$93,$12,$7F,$BA,$4E,$5F,$A3,$1F,$0F,$82,$46,$0D,$FC,$B7
	.dc.b	$C7,$E1,$91,$83,$FB,$13,$F3,$48,$FF,$61,$7E,$AD,$FF,$AB,$A3,$FB
	.dc.b	$46,$DF,$44,$FB,$F1,$98,$FE,$F1,$83,$87,$F9,$D4,$7F,$6B,$AF,$B4
	.dc.b	$47,$E0,$9F,$8C,$AF,$F2,$7B,$EE,$F5,$7A,$9D,$BE,$0B,$F0,$1D,$52
	.dc.b	$7F,$BD,$3A,$7E,$2F,$AB,$F8,$5E,$49,$18,$3F,$C0,$0F,$E2,$C8,$FF
	.dc.b	$7F,$7F,$D8,$9F,$1E,$FC,$FB,$FD,$D5,$3F,$E5,$98,$C7,$C5,$FE,$8D
	.dc.b	$B1,$F5,$3E,$13,$E1,$FC,$24,$8F,$AE,$EE,$F9,$C7,$F8,$74,$7E,$EE
	.dc.b	$F8,$5F,$D7,$46,$1B,$19,$8F,$F3,$0B,$FE,$F2,$6F,$C2,$68,$8E,$A7
	.dc.b	$9F,$E1,$DF,$B8,$8F,$F3,$4F,$D1,$7E,$25,$F1,$37,$C1,$FF,$30,$4E
	.dc.b	$E7,$3D,$5F,$EE,$75,$5C,$5F,$CC,$B5,$FE,$96,$D7,$FC,$73,$DF,$FB
	.dc.b	$72,$3F,$9A,$C3,$DF,$F4,$4F,$93,$A4,$9F,$F3,$FC,$3D,$BD,$FB,$52
	.dc.b	$72,$ED,$FC,$8F,$4B,$E1,$FF,$2B,$F1,$12,$7F,$53,$15,$FC,$BF,$6B
	.dc.b	$C4,$FE,$18,$67,$87,$FD,$05,$F1,$1F,$C3,$AB,$EC,$3F,$C7,$DF,$F4
	.dc.b	$9B,$5C,$F8,$FF,$6A,$31,$1F,$BE,$6B,$FD,$C1,$AF,$FC,$01,$6F,$83
	.dc.b	$FA,$9E,$BF,$E3,$C4,$9F,$F5,$0E,$CF,$67,$F6,$3F,$2D,$27,$F8,$2B
	.dc.b	$FB,$4C,$3F,$6B,$F0,$F2,$7F,$84,$BB,$7F,$CD,$DD,$CF,$CF,$5F,$39
	.dc.b	$F9,$D6,$CF,$E8,$5A,$FC,$EF,$37,$F9,$FF,$E0,$AF,$99,$A4,$9E,$17
	.dc.b	$AF,$86,$91,$FE,$8D,$A4,$72,$F8,$8E,$7B,$CA,$D3,$C1,$FE,$1B,$5F
	.dc.b	$83,$7F,$CF,$5E,$87,$F4,$33,$83,$AF,$98,$F2,$F5,$C8,$F0,$45,$7F
	.dc.b	$03,$EB,$7F,$4B,$31,$CF,$FE,$C1,$FA,$37,$93,$FA,$16,$D7,$F1,$80
	.dc.b	$E1,$FE,$36,$D7,$F9,$EB,$FB,$77,$F8,$07,$D4,$3D,$DF,$DF,$B6,$3F
	.dc.b	$DB,$B3,$BB,$F1,$78,$7F,$C1,$21,$E4,$53,$11,$6B,$C9,$3A,$6C,$FA
	.dc.b	$4F,$ED,$5D,$9F,$DA,$49,$1F,$4F,$E1,$BF,$BC,$FE,$27,$FC,$0F,$9F
	.dc.b	$FA,$A3,$67,$FC,$02,$DA,$71,$7E,$4E,$7F,$D9,$12,$FA,$CF,$EF,$AD
	.dc.b	$8F,$E7,$F4,$8C,$47,$F6,$28,$C4,$F0,$F5,$BC,$64,$9F,$C9,$65,$E6
	.dc.b	$E8,$91,$E0,$7F,$31,$E2,$7A,$BE,$63,$C2,$49,$F0,$E5,$FD,$5D,$FF
	.dc.b	$30,$E5,$EA,$F7,$BE,$0F,$D0,$C9,$FC,$B4,$7C,$4F,$5A,$47,$B9,$FD
	.dc.b	$0F,$C3,$7C,$8F,$63,$DE,$49,$EE,$FC,$2F,$92,$F9,$A4,$8F,$D1,$1F
	.dc.b	$E2,$8F,$C1,$1F,$1F,$F4,$BD,$A9,$3F,$EF,$39,$CB,$F3,$BB,$3E,$69
	.dc.b	$1F,$6B,$FD,$6B,$F7,$8F,$A9,$1F,$BE,$6F,$5B,$A1,$2F,$1D,$5A,$9F
	.dc.b	$F0,$A6,$0F,$87,$F3,$7C,$5F,$E1,$BD,$C9,$8F,$96,$3F,$13,$DF,$F8
	.dc.b	$02,$47,$D0,$7F,$07,$D5,$F8,$E2,$47,$B7,$E0,$BF,$23,$EA,$48,$FC
	.dc.b	$5B,$F1,$7E,$07,$F2,$BD,$9F,$8C,$24,$FD,$C7,$FC,$97,$A7,$F9,$D3
	.dc.b	$83,$FD,$97,$D6,$FA,$1F,$CE,$9F,$3F,$E0,$7E,$87,$B7,$F4,$7D,$59
	.dc.b	$97,$CF,$8F,$A9,$E6,$74,$24,$7E,$6B,$24,$7E,$7B,$D3,$FD,$49,$27
	.dc.b	$EC,$79,$7F,$E6,$4F,$8D,$FD,$46,$8F,$53,$F5,$E6,$7D,$EB,$7F,$96
	.dc.b	$8F,$D4,$91,$FD,$4D,$3C,$8E,$84,$8F,$F0,$B7,$1B,$9F,$F7,$7E,$B4
	.dc.b	$9F,$DA,$BB,$7F,$C6,$F6,$F1,$BC,$C4,$8F,$C2,$F2,$7F,$BD,$27,$FC
	.dc.b	$7C,$FC,$CE,$CF,$E0,$E4,$9D,$BF,$E5,$B9,$E9,$18,$3F,$D1,$2F,$F8
	.dc.b	$F9,$F0,$F8,$5F,$14,$F8,$DE,$1D,$8E,$1E,$3F,$F8,$97,$D5,$FF,$46
	.dc.b	$F7,$7F,$4F,$9C,$E0,$C3,$35,$FE,$8C,$AD,$57,$F0,$D6,$BF,$C0,$1A
	.dc.b	$F9,$FB,$FF,$DA,$FD,$3E,$9F,$0E,$CF,$A1,$B9,$32,$FE,$79,$F1,$3F
	.dc.b	$E4,$68,$C4,$EA,$7F,$A2,$4E,$5B,$7D,$EF,$F6,$33,$FE,$44,$FA,$2F
	.dc.b	$CB,$61,$FD,$35,$B3,$FE,$66,$6B,$FE,$88,$6B,$FE,$E5,$F0,$1F,$47
	.dc.b	$FD,$57,$E2,$FF,$D5,$4D,$8C,$CF,$F9,$C2,$72,$F5,$EC,$7C,$67,$85
	.dc.b	$B1,$98,$C1,$FC,$EB,$7F,$F1,$A7,$1F,$AB,$5F,$F4,$D4,$3D,$9F,$E1
	.dc.b	$D6,$DB,$E3,$4F,$08,$C5,$70,$AD,$D6,$BD,$9B,$3D,$1B,$5F,$ED,$2B
	.dc.b	$FF,$04,$E1,$E1,$8C,$4E,$B7,$FF,$36,$7C,$17,$8A,$7F,$00,$D7,$CE
	.dc.b	$D7,$A3,$63,$FC,$E9,$ED,$7A,$BA,$3D,$71,$D3,$0F,$CA,$E2,$AD,$47
	.dc.b	$43,$3E,$D6,$31,$F4,$4D,$BF,$F4,$04,$DB,$E5,$7D,$3E,$4F,$06,$7E
	.dc.b	$86,$C7,$81,$C3,$E2,$F6,$6E,$5F,$EB,$2B,$FF,$23,$4F,$9B,$DF,$F1
	.dc.b	$7A,$2F,$A3,$FE,$9F,$39,$3F,$D3,$CE,$2F,$FC,$8A,$DD,$BE,$0E,$0F
	.dc.b	$A2,$E7,$39,$BF,$C0,$FC,$A7,$C9,$A3,$49,$1F,$89,$EC,$F8,$11,$F0
	.dc.b	$5B,$1F,$33,$D9,$F8,$87,$C5,$F9,$DB,$1F,$43,$E6,$7B,$11,$FC,$A3
	.dc.b	$E1,$F6,$BE,$67,$CA,$FC,$D6,$CF,$D8,$EC,$FA,$1F,$2B,$F4,$5B,$1F
	.dc.b	$73,$83,$D4,$47,$F4,$0F,$E1,$7E,$8D,$F3,$3F,$6D,$B3,$FC,$61,$F8
	.dc.b	$92,$E9,$46,$A3,$FD,$CE,$FF,$90,$E7,$87,$C3,$73,$F9,$9E,$8F,$5F
	.dc.b	$77,$C0,$49,$FF,$15,$D7,$F7,$56,$FC,$BF,$8F,$D0,$F2,$36,$3C,$9F
	.dc.b	$2F,$E7,$3D,$1F,$19,$63,$D1,$EF,$FB,$48,$F6,$7E,$07,$95,$EF,$F9
	.dc.b	$7F,$80,$D9,$F8,$1E,$F7,$C3,$3E,$27,$CB,$D8,$F8,$9F,$2F,$F2,$48
	.dc.b	$FA,$9F,$99,$F2,$1F,$2C,$F9,$9F,$03,$67,$E6,$7C,$07,$CF,$F3,$7E
	.dc.b	$83,$63,$E8,$7C,$07,$D5,$23,$F1,$9F,$DF,$78,$CF,$A3,$7A,$FF,$A4
	.dc.b	$6D,$7D,$9E,$AC,$4E,$7E,$16,$4F,$F5,$4F,$A0,$7C,$16,$3B,$5E,$5F
	.dc.b	$F0,$69,$FF,$31,$57,$F9,$81,$B3,$FE,$5F,$6B,$FE,$52,$6B,$FE,$34
	.dc.b	$C5,$1F,$EB,$49,$F0,$7C,$67,$F0,$98,$FF,$01,$46,$0F,$27,$C6,$64
	.dc.b	$FE,$21,$6F,$A3,$FF,$20,$4E,$8F,$EE,$90,$FA,$2F,$F5,$D4,$7F,$A7
	.dc.b	$5B,$1F,$E4,$62,$38,$70,$71,$F7,$FF,$94,$61,$FE,$0B,$AE,$B5,$CB
	.dc.b	$3F,$89,$F0,$5E,$48,$C4,$F4,$F9,$3F,$E6,$58,$F8,$DB,$3F,$CF,$A3
	.dc.b	$B9,$99,$F3,$7B,$5C,$EB,$A3,$63,$EB,$7C,$17,$C7,$F8,$FF,$90,$D8
	.dc.b	$C9,$2F,$EC,$BF,$47,$37,$C2,$70,$8F,$C0,$7C,$77,$D1,$3E,$97,$C2
	.dc.b	$D8,$FC,$23,$F0,$88,$FE,$05,$F8,$BC,$DC,$23,$F1,$4D,$C3,$FD,$E7
	.dc.b	$E9,$7F,$42,$7C,$67,$F0,$9C,$5E,$B7,$F3,$A9,$D7,$0F,$E3,$E3,$F0
	.dc.b	$9D,$19,$F7,$AE,$5F,$E3,$1E,$B6,$1C,$23,$DD,$F1,$BF,$8B,$F8,$8F
	.dc.b	$C6,$36,$3F,$3D,$F0,$9F,$8F,$3F,$21,$F1,$B6,$3F,$13,$F4,$5F,$DE
	.dc.b	$F5,$FE,$DC,$D8,$FC,$47,$D1,$7F,$1D,$E0,$F5,$89,$D3,$FE,$92,$8F
	.dc.b	$E3,$4F,$DE,$6B,$F8,$E6,$BD,$3D,$2F,$C3,$FD,$7F,$F9,$AE,$FF,$37
	.dc.b	$5C,$B0,$FF,$00,$4E,$6F,$C3,$39,$AF,$BC,$7E,$AF,$E3,$BF,$58,$D3
	.dc.b	$F7,$C6,$0F,$E0,$1F,$1B,$6B,$D7,$E3,$FF,$94,$67,$F5,$6F,$E2,$1B
	.dc.b	$D5,$FB,$28,$C4,$F0,$E9,$C4,$E5,$B5,$FE,$60,$FA,$0F,$D2,$D7,$F4
	.dc.b	$96,$AF,$A3,$FE,$DB,$3B,$F0,$BF,$7F,$1F,$E5,$58,$FD,$3C,$EF,$E0
	.dc.b	$7F,$5B,$39,$8F,$D5,$B5,$FD,$DF,$D7,$BE,$23,$9D,$7F,$0A,$8C,$5F
	.dc.b	$F0,$8B,$E5,$C3,$FE,$2D,$F7,$FC,$BF,$49,$FC,$3B,$2F,$BF,$18,$BF
	.dc.b	$E2,$6F,$F9,$15,$FE,$22,$FB,$0C,$47,$0C,$FF,$3D,$F0,$7F,$E2,$7D
	.dc.b	$73,$8F,$F9,$33,$B3,$F3,$73,$9F,$8E,$D7,$8F,$C3,$3F,$94,$DC,$BF
	.dc.b	$57,$E6,$B2,$FF,$29,$47,$F1,$D8,$C7,$91,$FE,$41,$9C,$CF,$F3,$36
	.dc.b	$BF,$4F,$E3,$7C,$93,$E5,$6E,$1F,$67,$C2,$F4,$1E,$9F,$8C,$B1,$F7
	.dc.b	$5C,$3F,$7F,$E3,$6E,$1E,$6C,$5F,$1B,$E3,$25,$7F,$22,$CF,$D5,$1E
	.dc.b	$7E,$7F,$8C,$FC,$76,$CF,$E2,$FE,$37,$D8,$47,$95,$FF,$91,$B0,$FE
	.dc.b	$77,$87,$C5,$F8,$DF,$A1,$7F,$45,$B7,$F9,$5E,$9F,$A9,$F1,$9F,$55
	.dc.b	$B1,$F9,$9F,$23,$0D,$38,$1F,$9D,$F8,$CF,$9C,$47,$C6,$FD,$1E,$5F
	.dc.b	$C0,$3F,$81,$6C,$7E,$97,$E2,$FF,$08,$7E,$17,$E3,$2C,$7E,$A7,$AF
	.dc.b	$ED,$11,$F7,$BF,$77,$E3,$3F,$18,$FE,$35,$B1,$FE,$D0,$F8,$0F,$F1
	.dc.b	$CE,$E6,$F0,$B8,$7F,$CF,$51,$A3,$FC,$78,$FE,$2D,$AF,$F7,$B6,$BF
	.dc.b	$E4,$36,$CF,$5F,$8A,$E7,$F1,$7D,$1B,$87,$8F,$E2,$BB,$F9,$7E,$06
	.dc.b	$C7,$F7,$AC,$7F,$D9,$63,$E8,$DC,$3F,$DD,$B5,$FF,$59,$C6,$27,$FB
	.dc.b	$CF,$C7,$FA,$E3,$11,$ED,$9D,$D7,$F7,$0A,$F8,$1F,$03,$AA,$FF,$63
	.dc.b	$E1,$8A,$D4,$27,$57,$FE,$C2,$78,$47,$FB,$0C,$F3,$4F,$96,$BC,$4F
	.dc.b	$2C,$38,$C6,$27,$CE,$75,$9E,$0D,$33,$5C,$1C,$1F,$E1,$9C,$3C,$CF
	.dc.b	$F8,$6F,$0F,$25,$7A,$2B,$81,$E6,$AF,$F1,$93,$EC,$D7,$D9,$7E,$26
	.dc.b	$BD,$4F,$F3,$F4,$7F,$89,$FE,$43,$F1,$1E,$27,$29,$FC,$44,$61,$D9
	.dc.b	$8E,$0F,$F2,$4B,$81,$1B,$BE,$EF,$1B,$FA,$EB,$E1,$FF,$C6,$D1,$86
	.dc.b	$E5,$E3,$C3,$AC,$F0,$6C,$79,$E7,$78,$79,$9A,$F8,$A7,$6F,$9F,$1F
	.dc.b	$A8,$F9,$A7,$FC,$8D,$AF,$86,$33,$8F,$BE,$3C,$15,$F4,$A7,$81,$67
	.dc.b	$D4,$DB,$EE,$F8,$67,$9D,$F6,$C3,$D1,$3D,$AB,$E4,$E1,$F9,$DC,$BF
	.dc.b	$71,$F1,$7F,$E5,$9C,$7F,$DF,$34,$F8,$DC,$6F,$FB,$56,$E3,$FB,$5B
	.dc.b	$B8,$7F,$01,$F8,$CB,$83,$E3,$69,$88,$CF,$C6,$FF,$70,$AF,$E1,$0D
	.dc.b	$A3,$4C,$3D,$09,$87,$F1,$1F,$8D,$EF,$BF,$83,$E3,$64,$FF,$15,$8A
	.dc.b	$FF,$38,$A4,$F7,$24,$7F,$47,$87,$C6,$FA,$89,$1F,$C7,$FE,$37,$DB
	.dc.b	$F1,$BE,$EF,$5A,$4F,$F2,$38,$AF,$78,$F6,$11,$ED,$CD,$F2,$3E,$7B
	.dc.b	$FD,$4B,$1F,$C9,$F9,$1C,$9F,$E9,$D9,$CC,$FF,$74,$CF,$F5,$75,$A8
	.dc.b	$FF,$74,$4E,$A3,$EC,$F0,$7D,$23,$B7,$C6,$F7,$F9,$9F,$07,$CA,$CC
	.dc.b	$3C,$19,$77,$9C,$47,$F5,$EC,$55,$F5,$DB,$1F,$DB,$5A,$EF,$E3,$FD
	.dc.b	$9F,$1F,$23,$A9,$E6,$8C,$3D,$63,$A1,$1E,$9F,$19,$F1,$48,$F2,$CD
	.dc.b	$6F,$8E,$F9,$5F,$1B,$2B,$53,$FE,$94,$6D,$FC,$FF,$7F,$E8,$7B,$F2
	.dc.b	$77,$15,$F8,$11,$EF,$23,$EF,$F8,$DF,$D0,$FB,$8F,$A3,$73,$E1,$B5
	.dc.b	$E2,$AC,$4F,$DF,$F8,$E9,$3F,$E8,$FA,$DB,$FE,$80,$FA,$3F,$F7,$7B
	.dc.b	$FE,$D0,$7F,$D3,$6D,$F8,$EF,$D8,$A8,$DB,$C6,$33,$0F,$8E,$A6,$23
	.dc.b	$8F,$EC,$4F,$87,$E7,$F4,$8E,$4A,$6A,$2E,$3F,$E6,$69,$D3,$67,$FC
	.dc.b	$D2,$A3,$2F,$E2,$6F,$F3,$C4,$7F,$DD,$7F,$49,$F9,$7E,$AF,$F7,$47
	.dc.b	$D2,$3E,$13,$DA,$BE,$12,$98,$8B,$7F,$EF,$49,$7C,$27,$03,$31,$FE
	.dc.b	$5A,$E8,$7C,$8F,$EB,$C9,$1F,$86,$E0,$AE,$FE,$4B,$F6,$86,$3C,$2F
	.dc.b	$D7,$47,$97,$E1,$3E,$8F,$C9,$3D,$0A,$7D,$4D,$7F,$CA,$75,$FA,$01
	.dc.b	$3B,$7F,$D1,$ED,$7F,$DE,$ED,$7F,$C5,$07,$F8,$A5,$C3,$88,$7C,$27
	.dc.b	$72,$47,$79,$FE,$60,$FE,$89,$FE,$1C,$47,$F4,$BF,$25,$FC,$47,$9B
	.dc.b	$F8,$97,$C7,$C9,$FD,$4C,$7F,$15,$E3,$FC,$FF,$D1,$BC,$6A,$66,$BF
	.dc.b	$DD,$11,$97,$F6,$DF,$42,$F8,$EF,$E8,$AD,$8D,$D6,$3E,$3B,$3F,$1D
	.dc.b	$27,$8C,$57,$FC,$DF,$49,$FE,$90,$D8,$FF,$3F,$35,$DE,$FE,$02,$3E
	.dc.b	$49,$F1,$FC,$6C,$8F,$3B,$E6,$EA,$C4,$7F,$A8,$AF,$F1,$AF,$B3,$A7
	.dc.b	$3C,$BF,$84,$FA,$27,$8F,$FB,$45,$E3,$53,$11,$6F,$FD,$9F,$3A,$6C
	.dc.b	$FE,$AA,$31,$1F,$03,$E3,$FA,$78,$12,$3E,$6C,$57,$F0,$E7,$F9,$BE
	.dc.b	$7F,$83,$B8,$0E,$1D,$B5,$FF,$3B,$78,$4C,$FF,$33,$1F,$07,$AF,$EC
	.dc.b	$9E,$DE,$DF,$B7,$D7,$27,$EF,$BF,$7B,$E3,$FE,$FF,$C7,$FD,$62,$4F
	.dc.b	$D6,$7F,$DD,$FE,$53,$A7,$F8,$B4,$FE,$2B,$0F,$C7,$9F,$88,$F8,$FF
	.dc.b	$C8,$F8,$52,$7F,$1E,$FE,$33,$E3,$FF,$1B,$F1,$FF,$84,$24,$FE,$11
	.dc.b	$FC,$3F,$C7,$FE,$4B,$E1,$24,$FE,$5A,$3F,$2B,$F1,$F2,$3F,$32,$FE
	.dc.b	$63,$E3,$FF,$33,$F1,$FF,$90,$24,$FE,$41,$FC,$9B,$F9,$E2,$1F,$A2
	.dc.b	$F8,$AF,$F1,$2F,$D0,$3E,$72,$FD,$14,$AF,$EB,$93,$96,$FC,$8F,$F8
	.dc.b	$96,$0D,$24,$7F,$B7,$61,$F2,$1D,$89,$1D,$3E,$47,$C3,$F1,$BE,$2F
	.dc.b	$91,$93,$E0,$8F,$2F,$A3,$DC,$91,$D5,$F1,$7C,$6F,$A8,$91,$E8,$F8
	.dc.b	$EF,$6F,$C6,$C8,$F5,$BF,$03,$E3,$BE,$0F,$C4,$7A,$C9,$3E,$97,$DD
	.dc.b	$F1,$BF,$1B,$E3,$A4,$FC,$48,$F9,$3F,$1D,$F2,$BE,$3A,$4F,$C8,$8F
	.dc.b	$9B,$F0,$9F,$1C,$91,$F1,$5F,$95,$F1,$BF,$48,$91,$F4,$3E,$3B,$EA
	.dc.b	$F9,$5F,$E5,$AF,$03,$FA,$A4,$FF,$63,$6C,$39,$BF,$AE,$FA,$1F,$E3
	.dc.b	$18,$4F,$F8,$DA,$BF,$9C,$C6,$3E,$3B,$FC,$BD,$39,$9F,$D7,$B5,$78
	.dc.b	$EF,$7D,$D1,$FE,$0E,$6C,$FF,$8A,$A7,$53,$E7,$F8,$EF,$47,$C7,$4A
	.dc.b	$FF,$13,$61,$FE,$34,$CF,$FC,$8D,$87,$09,$CF,$C7,$7B,$C8,$F0,$F7
	.dc.b	$3E,$17,$C7,$48,$E6,$E9,$5F,$C1,$C1,$F1,$BF,$27,$E3,$A5,$78,$A2
	.dc.b	$BC,$3E,$07,$E0,$FC,$77,$CF,$33,$87,$37,$E3,$92,$7F,$8D,$71,$7C
	.dc.b	$E8,$F9,$7F,$1D,$FE,$D7,$E7,$F9,$6D,$AF,$B0,$DA,$FA,$BE,$E5,$D3
	.dc.b	$FC,$4F,$F2,$3F,$E3,$69,$FB,$3E,$EE,$4D,$5F,$3C,$1F,$40,$F8,$8F
	.dc.b	$F0,$EC,$7F,$76,$8C,$5F,$F2,$F8,$DC,$E9,$18,$8D,$7C,$93,$C4,$7D
	.dc.b	$06,$1B,$5F,$CE,$67,$53,$93,$B6,$1F,$E0,$CF,$3E,$FC,$59,$A4,$BF
	.dc.b	$E8,$5E,$9F,$F8,$AF,$93,$CC,$D8,$CF,$C2,$7F,$87,$9F,$F0,$8C,$65
	.dc.b	$FE,$9A,$FF,$8B,$A7,$FC,$38,$F1,$E5,$FF,$12,$B9,$9F,$BC,$7C,$07
	.dc.b	$53,$D3,$E1,$7B,$4E,$5A,$A5,$18,$6C,$FA,$3E,$0F,$04,$70,$7B,$3E
	.dc.b	$07,$FD,$49,$87,$92,$6C,$FB,$26,$E1,$E5,$9B,$C0,$FB,$B7,$89,$86
	.dc.b	$E1,$FE,$91,$F7,$BE,$73,$FE,$93,$8C,$E1,$F0,$2F,$FC,$E1,$B7,$1C
	.dc.b	$3A,$E1,$C2,$B5,$7C,$33,$FE,$14,$BC,$3F,$01,$E6,$F0,$9F,$9C,$FF
	.dc.b	$9E,$F4,$F0,$4E,$63,$18,$74,$8C,$63,$CA,$BF,$D6,$3C,$8F,$96,$7C
	.dc.b	$1F,$D2,$8C,$36,$33,$3F,$E0,$99,$CB,$F9,$DF,$CE,$7D,$F8,$FF,$84
	.dc.b	$1A,$FF,$96,$1A,$FE,$81,$BC,$1F,$C9,$BE,$23,$E7,$F7,$9F,$9D,$E4
	.dc.b	$61,$FA,$B8,$F9,$FE,$0F,$C8,$4C,$BC,$99,$A3,$D5,$C5,$FE,$64,$9F
	.dc.b	$F3,$86,$7F,$2D,$F8,$31,$9D,$3F,$2A,$27,$4F,$D9,$CF,$E5,$CF,$F1
	.dc.b	$DD,$BC,$7E,$9B,$CE,$F4,$C6,$37,$F9,$53,$FE,$A9,$CB,$E4,$37,$1F
	.dc.b	$F1,$AE,$A8,$C4,$7F,$46,$E4,$79,$7F,$13,$CD,$FC,$83,$70,$FE,$77
	.dc.b	$F0,$7F,$20,$FC,$04,$D8,$FB,$BF,$21,$FE,$75,$7E,$C3,$70,$F0,$B5
	.dc.b	$FF,$46,$E5,$FE,$EA,$F3,$BF,$05,$1B,$DF,$F9,$E7,$8B,$F4,$A3,$18
	.dc.b	$7F,$58,$8F,$CA,$7C,$CF,$E6,$E7,$FD,$2F,$D6,$8C,$47,$F2,$08,$C7
	.dc.b	$6F,$F0,$06,$DF,$E3,$79,$3F,$0F,$9F,$F6,$78,$C7,$6F,$E5,$67,$F1
	.dc.b	$74,$FC,$C3,$FD,$05,$5F,$E5,$D8,$C7,$C6,$F5,$9C,$E5,$FE,$64,$9C
	.dc.b	$D7,$F4,$AE,$3F,$EA,$EF,$F6,$6F,$17,$F7,$DE,$7F,$E5,$1F,$CA,$FC
	.dc.b	$97,$F1,$6E,$8F,$EF,$AD,$D0,$D3,$58,$3D,$0F,$CE,$FC,$97,$F0,$96
	.dc.b	$E5,$FC,$1F,$D0,$FD,$0B,$FC,$35,$AF,$F4,$6E,$5F,$F0,$67,$E0,$7F
	.dc.b	$EA,$98,$FE,$C2,$DD,$FF,$CF,$FA,$23,$11,$FE,$10,$6B,$F4,$7F,$61
	.dc.b	$FF,$76,$57,$53,$FC,$FF,$2F,$93,$FE,$1D,$3A,$AF,$E3,$53,$9A,$FE
	.dc.b	$3D,$E9,$7F,$26,$8F,$9B,$8F,$F1,$88,$C7,$26,$7E,$4F,$E4,$57,$3C
	.dc.b	$3C,$91,$83,$F0,$58,$BC,$BE,$9E,$7F,$89,$B9,$FF,$BA,$7E,$13,$FD
	.dc.b	$D2,$DE,$5F,$E7,$38,$BE,$B3,$FD,$FB,$39,$3F,$DE,$1F,$47,$E9,$9F
	.dc.b	$F5,$13,$5F,$F5,$3B,$5F,$F7,$7B,$5F,$EE,$EF,$E2,$9D,$EB,$FD,$5A
	.dc.b	$31,$1C,$23,$F0,$3E,$B7,$C5,$F9,$5F,$E9,$0D,$BF,$A1,$4E,$37,$E6
	.dc.b	$FF,$87,$9F,$E9,$55,$FD,$4A,$31,$5F,$D1,$A3,$FA,$97,$9D,$F4,$7D
	.dc.b	$5B,$7D,$30,$FC,$14,$62,$FF,$A6,$D7,$3B,$C7,$2B,$0F,$F0,$9D,$3E
	.dc.b	$8F,$F1,$93,$98,$FF,$0B,$C6,$27,$E2,$FC,$A7,$D0,$8C,$4F,$C6,$8F
	.dc.b	$88,$49,$FF,$11,$6C,$8F,$91,$E8,$7D,$88,$C4,$FD,$68,$C4,$E7,$E4
	.dc.b	$BB,$75,$5A,$F9,$38,$79,$E3,$97,$C9,$7D,$32,$4F,$81,$FD,$9F,$A1
	.dc.b	$F9,$7E,$7F,$EF,$4D,$9F,$EF,$1E,$6F,$F7,$D3,$E8,$15,$FF,$37,$7D
	.dc.b	$97,$F8,$E1,$B8,$FC,$1E,$AF,$EE,$93,$98,$FE,$97,$F4,$6F,$8D,$FF
	.dc.b	$0D,$B6,$FF,$C8,$8D,$9F,$F7,$A3,$5F,$F7,$CA,$F8,$CE,$24,$9E,$2F
	.dc.b	$F0,$EF,$4B,$F6,$5E,$E7,$F8,$55,$B3,$FE,$13,$70,$FE,$17,$AB,$F5
	.dc.b	$38,$7F,$84,$27,$F5,$3E,$B7,$FA,$3E,$BF,$15,$E1,$FF,$80,$2B,$FB
	.dc.b	$2C,$E5,$FE,$07,$ED,$70,$BF,$F1,$BC,$FE,$C2,$7F,$16,$7F,$97,$9C
	.dc.b	$38,$6B,$FC,$43,$E5,$FF,$85,$1F,$E2,$DF,$52,$E3,$F7,$BD,$4F,$99
	.dc.b	$5D,$1A,$FE,$61,$AF,$F9,$B1,$AB,$F0,$B7,$C9,$FF,$93,$1B,$5F,$E7
	.dc.b	$56,$CF,$A0,$E1,$3F,$E5,$2F,$95,$FF,$5D,$D7,$C4,$EB,$FE,$1F,$18
	.dc.b	$AF,$E4,$B1,$8F,$97,$E0,$4A,$FF,$41,$52,$3F,$A2,$F9,$1F,$CA,$7B
	.dc.b	$BF,$C9,$8D,$AF,$F5,$E7,$95,$FE,$61,$7F,$AE,$37,$1F,$EB,$4D,$7F
	.dc.b	$B7,$B5,$FE,$77,$E9,$7F,$30,$24,$FF,$48,$7F,$CD,$BE,$A3,$E5,$E9
	.dc.b	$27,$FD,$31,$19,$7F,$B3,$7A,$9F,$E6,$13,$FA,$2D,$D8,$F3,$B4,$FF
	.dc.b	$3F,$CB,$0F,$D2,$78,$DF,$E5,$CB,$FE,$91,$A7,$F4,$42,$61,$FD,$71
	.dc.b	$FF,$02,$7C,$77,$FA,$B6,$3F,$BE,$37,$1F,$EF,$4D,$7F,$C1,$ED,$7F
	.dc.b	$AD,$F8,$BF,$D4,$09,$3F,$DA,$1F,$F4,$6E,$1F,$E3,$18,$FF,$5B,$36
	.dc.b	$7F,$04,$D7,$F3,$6D,$7F,$D7,$3E,$5F,$FA,$17,$D4,$FF,$42,$1F,$E7
	.dc.b	$5F,$CD,$DF,$31,$8C,$3F,$DE,$D1,$87,$FD,$0D,$3B,$9D,$5F,$F6,$EF
	.dc.b	$23,$FB,$39,$2B,$FB,$FB,$FE,$27,$FD,$A3,$FB,$54,$FF,$8A,$DA,$B8
	.dc.b	$11,$86,$FA,$9F,$EA,$CF,$17,$E5,$6B,$FC,$F6,$FF,$8B,$76,$7F,$6A
	.dc.b	$58,$FF,$6C,$AD,$E5,$FB,$CE,$4F,$F5,$AD,$7E,$14,$FF,$6F,$38,$33
	.dc.b	$F0,$F2,$7F,$DC,$53,$F9,$53,$FD,$96,$FF,$B9,$6B,$F8,$9E,$3E,$36
	.dc.b	$C7,$F0,$AE,$9F,$17,$47,$97,$DE,$F0,$92,$7A,$BE,$4C,$1F,$17,$FE
	.dc.b	$EE,$FA,$47,$37,$F1,$4A,$FE,$B9,$18,$AF,$E2,$F5,$BA,$D5,$D9,$C8
	.dc.b	$AE,$0F,$FB,$EA,$F9,$D7,$F0,$9B,$E1,$39,$BA,$FF,$C0,$08,$B5,$C2
	.dc.b	$3A,$DF,$47,$FD,$88,$D7,$D6,$3E,$4B,$53,$FD,$D5,$D5,$27,$58,$7A
	.dc.b	$9A,$76,$71,$7C,$23,$D9,$18,$3D,$4F,$03,$D4,$53,$D4,$5B,$F1,$47
	.dc.b	$0B,$F2,$E1,$E3,$7B,$18,$AE,$47,$B3,$9F,$D8,$FC,$E6,$BF,$E1,$4C
	.dc.b	$7F,$5F,$3E,$49,$CF,$C6,$7F,$AA,$76,$CD,$FC,$18,$C3,$FE,$2D,$AF
	.dc.b	$F0,$EC,$62,$B8,$FC,$32,$F8,$DA,$B8,$DF,$DE,$5F,$2B,$FE,$26,$8C
	.dc.b	$37,$3E,$B8,$73,$BF,$91,$7C,$A7,$C3,$5B,$9F,$33,$5F,$05,$71,$BF
	.dc.b	$03,$CF,$5E,$F3,$F8,$CC,$3C,$6D,$9D,$61,$E2,$C7,$E8,$3E,$09,$A7
	.dc.b	$80,$A7,$82,$2D,$7C,$8B,$EF,$3F,$84,$38,$14,$E0,$58,$E1,$5D,$E3
	.dc.b	$18,$FF,$93,$2B,$D7,$7E,$73,$D5,$19,$AE,$13,$AA,$E2,$7F,$73,$3E
	.dc.b	$A4,$FF,$0A,$BF,$83,$1F,$DD,$DA,$FF,$84,$38,$1F,$03,$FE,$5C,$C7
	.dc.b	$F8,$63,$76,$FF,$2F,$CE,$AB,$FC,$97,$6F,$95,$D4,$7E,$2F,$E5,$7C
	.dc.b	$04,$7C,$92,$8F,$27,$74,$9C,$B5,$FF,$06,$5B,$E0,$7A,$C7,$F9,$15
	.dc.b	$B1,$98,$F1,$FC,$8F,$8B,$B3,$F9,$DF,$57,$F7,$4E,$DF,$E6,$B5,$FE
	.dc.b	$DF,$6B,$FE,$DD,$6B,$FE,$C1,$EB,$F9,$5C,$DF,$EC,$56,$F9,$DF,$E0
	.dc.b	$19,$FF,$65,$C7,$F8,$03,$C1,$7C,$0F,$C2,$B7,$C2,$7F,$47,$E5,$FF
	.dc.b	$42,$46,$39,$3C,$D7,$C3,$1F,$F4,$95,$EE,$7F,$B3,$B6,$3F,$D4,$0D
	.dc.b	$D5,$5C,$67,$F6,$95,$FE,$94,$E8,$FE,$93,$1F,$C8,$CC,$1F,$D3,$1F
	.dc.b	$E9,$B8,$7F,$B8,$7B,$FD,$A4,$9F,$F7,$3B,$FE,$1C,$FC,$9F,$FD,$9D
	.dc.b	$1F,$E3,$FC,$51,$88,$FF,$2C,$C6,$3D,$AE,$07,$C8,$F0,$64,$FF,$BC
	.dc.b	$9A,$F0,$CD,$F2,$BF,$27,$DB,$7C,$FD,$3E,$3C,$5A,$FF,$5C,$4E,$9B
	.dc.b	$3A,$7F,$D3,$FF,$BC,$BA,$BF,$B5,$B6,$3F,$4C,$D7,$FC,$B4,$D7,$FD
	.dc.b	$7B,$B3,$95,$E2,$7E,$08,$93,$FE,$61,$9D,$47,$E0,$FD,$99,$3F,$1F
	.dc.b	$4F,$F4,$A3,$63,$F1,$B9,$BD,$EF,$C4,$63,$FE,$28,$E4,$FF,$13,$B6
	.dc.b	$FF,$DC,$15,$B7,$FD,$77,$F9,$BF,$F4,$07,$FD,$12,$D7,$FD,$AA,$D7
	.dc.b	$FA,$F3,$53,$FE,$EB,$7F,$DC,$5F,$84,$7F,$A0,$1B,$EA,$7E,$C2,$75
	.dc.b	$1F,$8E,$3F,$B4,$7D,$2B,$E6,$3F,$C2,$ED,$8F,$E2,$4D,$7F,$D0,$CD
	.dc.b	$7F,$D3,$EB,$C9,$E2,$48,$E2,$FE,$CB,$0F,$F2,$2C,$79,$BC,$FF,$D7
	.dc.b	$46,$23,$C9,$C2,$CD,$F1,$3F,$C2,$3E,$95,$EA,$7F,$8B,$1B,$7F,$C9
	.dc.b	$DA,$FF,$A6,$DA,$FF,$AD,$17,$A7,$E0,$FA,$19,$3F,$EE,$C9,$CB,$F9
	.dc.b	$4F,$53,$F3,$44,$7F,$C1,$DE,$1F,$F0,$2F,$3F,$F9,$14,$6C,$FA,$3E
	.dc.b	$2F,$B0,$E0,$FC,$F1,$E9,$FE,$FD,$FF,$63,$B5,$FF,$0D,$1A,$8F,$F2
	.dc.b	$4C,$66,$3F,$C3,$CB,$E8,$72,$48,$FE,$91,$19,$7F,$7D,$F7,$3F,$EB
	.dc.b	$17,$D3,$D3,$FE,$27,$F1,$13,$99,$FF,$4C,$B7,$2F,$F3,$37,$E0,$8F
	.dc.b	$90,$FD,$E7,$C7,$CB,$FD,$CB,$FC,$9F,$57,$C1,$7F,$01,$F8,$9F,$BB
	.dc.b	$B4,$BF,$DF,$C7,$F4,$2D,$9E,$8F,$B8,$93,$EA,$7E,$0D,$FF,$48,$FB
	.dc.b	$BF,$9F,$3F,$2D,$BA,$FF,$9E,$EF,$C3,$3E,$E3,$FC,$89,$F4,$EF,$95
	.dc.b	$FE,$A2,$BE,$76,$98,$8B,$3F,$DC,$67,$4D,$9F,$F3,$DC,$62,$3F,$DD
	.dc.b	$1F,$82,$3E,$4F,$FA,$EB,$63,$C4,$1C,$3F,$83,$9E,$58,$FF,$4C,$ED
	.dc.b	$7C,$A7,$F2,$92,$57,$AF,$BD,$F7,$5E,$B8,$FE,$53,$3F,$AA,$F4,$24
	.dc.b	$60,$E1,$2F,$9F,$FE,$69,$E9,$C9,$CC,$57,$F6,$3E,$23,$E9,$FF,$6A
	.dc.b	$67,$4F,$F5,$46,$BF,$83,$AF,$8D,$D5,$F1,$08,$FC,$9F,$97,$F8,$E4
	.dc.b	$70,$7F,$87,$79,$7F,$9C,$5F,$FA,$CD,$BD,$FF,$A2,$6C,$FF,$9F,$7E
	.dc.b	$E3,$FB,$35,$2B,$F3,$5E,$8F,$E0,$2B,$D3,$EF,$7F,$0C,$3F,$BA,$F5
	.dc.b	$7F,$0E,$F4,$BF,$89,$12,$7F,$A0,$6F,$FC,$F1,$FD,$89,$FC,$DB,$E4
	.dc.b	$FF,$8A,$FD,$44,$8F,$EA,$F1,$FE,$15,$F8,$4F,$B3,$9F,$FB,$92,$DF
	.dc.b	$51,$FC,$70,$D7,$6F,$F1,$EF,$4B,$F9,$11,$31,$F0,$BF,$D2,$BE,$53
	.dc.b	$F9,$1F,$A5,$27,$FB,$2C,$7F,$6B,$F9,$3F,$E4,$FE,$3C,$9F,$ED,$31
	.dc.b	$FC,$AB,$4F,$E5,$C4,$8F,$6B,$FD,$5B,$E4,$7F,$97,$FC,$3C,$9F,$EE
	.dc.b	$31,$FE,$2F,$F8,$C7,$CE,$7F,$94,$7C,$FF,$F6,$5B,$63,$35,$FE,$09
	.dc.b	$6B,$F2,$3E,$79,$F7,$EF,$A2,$FF,$2D,$46,$1B,$19,$9F,$9E,$27,$4F
	.dc.b	$F8,$57,$F1,$E7,$A3,$E3,$6C,$66,$30,$76,$F8,$1F,$1F,$77,$F8,$DB
	.dc.b	$9F,$F1,$3E,$51,$FC,$5E,$3F,$E0,$2D,$7F,$CB,$EB,$E5,$BF,$59,$97
	.dc.b	$AD,$AF,$F3,$48,$D7,$9F,$EC,$9C,$CF,$E0,$7D,$1F,$EB,$64,$7F,$89
	.dc.b	$70,$FF,$9B,$7C,$6F,$F1,$27,$CA,$7F,$60,$F0,$A5,$7F,$88,$A3,$C7
	.dc.b	$87,$C5,$8F,$17,$1F,$72,$3F,$98,$F9,$3F,$ED,$24,$7F,$90,$69,$EB
	.dc.b	$F2,$FD,$BC,$DF,$E3,$49,$FE,$27,$F9,$A3,$A3,$FC,$E6,$D8,$E3,$3B
	.dc.b	$7F,$C0,$D6,$FB,$D7,$CC,$7D,$C6,$CF,$DA,$9D,$4E,$4F,$F0,$07,$2F
	.dc.b	$F7,$E2,$4F,$F9,$03,$E4,$BC,$3B,$7F,$A7,$E1,$97,$D3,$7F,$D4,$4D
	.dc.b	$7F,$0E,$D7,$EA,$63,$FE,$80,$47,$E2,$DC,$36,$BD,$07,$E7,$7B,$7F
	.dc.b	$99,$F4,$7F,$31,$F9,$1F,$F2,$C1,$FB,$DF,$5E,$DF,$F9,$5A,$2B,$FE
	.dc.b	$29,$E6,$FF,$54,$4F,$FA,$1F,$EE,$BC,$D7,$FC,$EF,$E4,$3F,$CC,$46
	.dc.b	$75,$B8,$7F,$98,$22,$BF,$E2,$0A,$E3,$E4,$FF,$80,$BE,$AB,$F0,$58
	.dc.b	$7F,$1F,$AF,$6C,$3E,$B3,$FD,$56,$D8,$FE,$9E,$F0,$AE,$53,$F3,$63
	.dc.b	$53,$FB,$06,$BF,$C8,$1A,$F8,$7E,$63,$81,$1F,$F6,$A4,$7F,$B2,$5E
	.dc.b	$5F,$59,$E2,$F0,$24,$78,$A2,$BD,$E3,$C4,$47,$9F,$CD,$FF,$95,$1F
	.dc.b	$80,$D7,$DE,$D5,$48,$C7,$CD,$7F,$96,$7A,$3F,$DA,$BE,$6F,$F0,$BB
	.dc.b	$FF,$73,$35,$FE,$7B,$3B,$7E,$17,$C0,$7F,$A1,$1B,$E3,$7B,$A3,$11
	.dc.b	$F1,$5A,$FF,$8D,$B8,$7F,$8A,$EF,$FD,$F9,$7C,$0F,$D7,$E1,$FB,$14
	.dc.b	$9D,$BF,$E4,$37,$75,$C7,$C8,$E0,$F5,$6C,$FE,$73,$F3,$4F,$EE,$93
	.dc.b	$B3,$C1,$EE,$F6,$9F,$23,$5F,$D1,$DA,$31,$1E,$36,$C3,$D4,$EF,$39
	.dc.b	$AF,$64,$62,$79,$46,$09,$3F,$E5,$DF,$C5,$BD,$B9,$7C,$0A,$E9,$5F
	.dc.b	$EB,$ED,$7D,$7E,$07,$FB,$B3,$CB,$BF,$05,$FE,$B4,$AF,$F7,$4C,$62
	.dc.b	$A4,$7F,$88,$7F,$06,$7C,$96,$34,$FD,$97,$D7,$FF,$A1,$EB,$FD,$E3
	.dc.b	$EC,$7C,$8F,$05,$39,$AF,$F4,$6F,$CE,$7F,$90,$6F,$F9,$AC,$E7,$7F
	.dc.b	$D7,$5F,$0D,$7C,$F9,$BE,$3B,$A5,$EF,$0F,$F3,$BF,$D7,$7F,$93,$5F
	.dc.b	$EB,$79,$A3,$0F,$F5,$E6,$BF,$E0,$E6,$BF,$C6,$3A,$D9,$FC,$E8,$FF
	.dc.b	$44,$4F,$28,$FE,$05,$3A,$E5,$FA,$F3,$9B,$FE,$43,$C4,$DB,$FC,$81
	.dc.b	$1F,$E4,$96,$E3,$FC,$F3,$EB,$FF,$B0,$4E,$7C,$EE,$7C,$3F,$86,$C7
	.dc.b	$E3,$5F,$A6,$30,$7E,$37,$3F,$3F,$B3,$F9,$66,$E1,$F5,$FE,$23,$1B
	.dc.b	$2B,$5E,$EF,$FA,$1A,$BF,$AB,$B7,$1F,$AB,$F5,$DA,$F5,$BD,$BF,$0D
	.dc.b	$E3,$25,$FC,$ED,$51,$F9,$FE,$1F,$81,$EB,$7E,$3B,$8B,$FC,$25,$7F
	.dc.b	$89,$F8,$1C,$4F,$DA,$FA,$A9,$3F,$17,$CC,$BA,$FD,$0F,$3F,$E1,$12
	.dc.b	$7C,$0F,$C1,$86,$DF,$05,$B3,$F9,$7F,$80,$C4,$FE,$83,$EA,$5F,$3F
	.dc.b	$F9,$5F,$A7,$FD,$06,$1F,$EE,$09,$CE,$1F,$E0,$89,$CF,$2F,$F0,$AF
	.dc.b	$CF,$FF,$BC,$4F,$E3,$9B,$7F,$E2,$A6,$BF,$E5,$A6,$BF,$84,$F7,$5F
	.dc.b	$0F,$FA,$5F,$AF,$FF,$76,$DF,$FA,$36,$72,$FF,$86,$23,$35,$FA,$8F
	.dc.b	$53,$E6,$7A,$BF,$35,$B5,$FE,$9E,$C7,$F1,$75,$E3,$F5,$FF,$87,$C6
	.dc.b	$23,$C9,$1F,$5F,$9B,$FC,$9F,$5F,$95,$F5,$BF,$5F,$D7,$F6,$5F,$DD
	.dc.b	$7D,$43,$E8,$FE,$E7,$D3,$E1,$B7,$E1,$E4,$EE,$47,$E3,$3F,$C8,$FC
	.dc.b	$B5,$BE,$67,$A9,$FC,$53,$EA,$FF,$4B,$CB,$FC,$0A,$75,$87,$F8,$E9
	.dc.b	$FE,$3A,$FE,$BB,$5F,$F7,$23,$9F,$57,$F2,$1F,$0D,$FC,$AB,$8B,$EE
	.dc.b	$F1,$7D,$3F,$47,$E1,$92,$FF,$CC,$3A,$F9,$3C,$CF,$79,$E3,$E5,$B3
	.dc.b	$FE,$CA,$FA,$F7,$D6,$7F,$36,$6C,$7F,$B3,$BE,$B9,$F4,$D9,$AF,$E7
	.dc.b	$71,$F6,$B3,$FF,$4E,$B5,$FF,$13,$87,$0F,$F8,$A9,$AE,$78,$F9,$FD
	.dc.b	$17,$46,$E1,$FE,$D5,$E4,$FB,$75,$FC,$BB,$C9,$F9,$9D,$3F,$B3,$6D
	.dc.b	$6E,$3E,$87,$16,$2B,$F9,$AF,$89,$FB,$93,$EA,$5D,$BE,$CD,$9F,$E4
	.dc.b	$F1,$FC,$83,$0F,$E0,$6F,$F1,$FA,$F8,$3F,$15,$FD,$73,$EB,$FF,$15
	.dc.b	$3F,$1E,$31,$D7,$F1,$BB,$B7,$3F,$2D,$B8,$F6,$3F,$C9,$BF,$75,$FE
	.dc.b	$FF,$9D,$1F,$EF,$E7,$11,$B3,$E6,$4F,$D2,$F8,$8F,$A7,$F5,$96,$FE
	.dc.b	$C3,$F5,$BE,$23,$17,$C7,$8D,$18,$EF,$FF,$21,$79,$FF,$CF,$7C,$DF
	.dc.b	$E3,$DE,$DF,$F1,$F6,$E5,$F4,$DF,$CE,$1F,$9D,$E8,$B1,$FE,$66,$BF
	.dc.b	$F0,$AF,$0F,$F8,$4B,$3F,$43,$63,$FC,$33,$E5,$BE,$5B,$E5,$7C,$1F
	.dc.b	$F8,$1A,$3F,$83,$DF,$FA,$E9,$AF,$FB,$A9,$AF,$E3,$9A,$FF,$9D,$97
	.dc.b	$A5,$CF,$CD,$E8,$DC,$BF,$AC,$67,$FC,$EC,$FD,$B7,$C2,$58,$FC,$83
	.dc.b	$70,$FA,$DB,$BA,$53,$9D,$7F,$C6,$7E,$5F,$EA,$2F,$EB,$2E,$0F,$F1
	.dc.b	$0B,$99,$DC,$63,$E1,$F3,$9D,$CF,$E5,$1F,$AA,$BE,$0F,$F0,$59,$FE
	.dc.b	$A4,$7F,$8B,$3D,$0F,$F4,$AC,$7F,$73,$8F,$F1,$7F,$D7,$FF,$87,$23
	.dc.b	$FC,$12,$DC,$BF,$DC,$AD,$7F,$6E,$1C,$3F,$BA,$5F,$4F,$D3,$E0,$BA
	.dc.b	$B6,$BC,$F5,$FE,$A6,$F3,$DF,$23,$FE,$40,$7F,$D7,$B3,$A9,$FE,$AF
	.dc.b	$39,$F9,$0D,$76,$6E,$2D,$6E,$F9,$BF,$C9,$FE,$1B,$FB,$97,$17,$F3
	.dc.b	$A6,$C7,$8B,$91,$F8,$17,$F2,$EC,$FF,$D5,$9E,$9F,$F9,$9E,$B5,$39
	.dc.b	$E5,$FE,$69,$87,$F9,$5B,$CF,$FD,$2D,$7F,$8C,$1B,$87,$F2,$CE,$FF
	.dc.b	$F2,$FC,$62,$3F,$8F,$E5,$E0,$FA,$0E,$55,$FE,$54,$C9,$18,$C3,$FA
	.dc.b	$05,$FF,$82,$FE,$A3,$FA,$9B,$63,$24,$7F,$23,$4F,$9A,$DC,$FD,$0C
	.dc.b	$9F,$5B,$FE,$99,$8F,$E8,$2D,$BF,$BB,$3A,$E7,$FE,$B1,$39,$9F,$4F
	.dc.b	$EF,$4E,$4F,$F2,$F4,$F1,$E0,$FE,$0B,$39,$C7,$C1,$39,$7F,$9F,$CB
	.dc.b	$E0,$BF,$D0,$6D,$8F,$F2,$CB,$70,$FC,$79,$FE,$54,$9F,$F1,$76,$9F
	.dc.b	$E1,$09,$FF,$4E,$9B,$3F,$8A,$B9,$AE,$F3,$FE,$73,$F9,$EF,$CA,$C7
	.dc.b	$F9,$4F,$EA,$B1,$D3,$FE,$37,$6C,$7C,$79,$DE,$3F,$33,$A3,$2F,$FB
	.dc.b	$C7,$EC,$3F,$10,$FF,$5A,$E0,$46,$20,$23,$3E,$9F,$F8,$0E,$75,$39
	.dc.b	$7F,$90,$61,$C2,$B6,$FD,$98,$CC,$EE,$BF,$D4,$EB,$CB,$A6,$39,$AD
	.dc.b	$FE,$AE,$7F,$38,$7F,$A1,$BA,$3F,$D6,$6E,$0F,$9E,$FF,$AD,$BC,$8F
	.dc.b	$F4,$74,$7F,$B3,$E7,$F6,$91,$88,$FF,$5E,$C7,$5E,$CF,$19,$E4,$F6
	.dc.b	$6C,$78,$65,$F3,$5F,$13,$E0,$7F,$15,$1F,$E0,$2A,$56,$7A,$3F,$03
	.dc.b	$76,$B5,$35,$FF,$04,$C7,$CE,$FB,$1E,$71,$F4,$2F,$9A,$8C,$BF,$52
	.dc.b	$78,$5F,$F1,$CA,$DD,$6A,$BC,$E7,$19,$D5,$F0,$8F,$E4,$77,$5F,$04
	.dc.b	$57,$B1,$C4,$F4,$94,$F4,$94,$FF,$39,$D7,$86,$7D,$D1,$FE,$76,$C3
	.dc.b	$09,$F5,$E1,$EE,$73,$58,$BF,$15,$7A,$A3,$0F,$73,$DB,$38,$3A,$97
	.dc.b	$0F,$A3,$1D,$DF,$82,$D3,$E0,$94,$F8,$25,$3E,$0F,$77,$9D,$DB,$5F
	.dc.b	$82,$D7,$73,$FE,$44,$F9,$5F,$E0,$DD,$3F,$E8,$99,$FD,$5C,$61,$57
	.dc.b	$38,$C3,$7A,$3F,$DB,$73,$AB,$E9,$18,$7F,$0F,$4F,$85,$FF,$29,$C6
	.dc.b	$1B,$8F,$0C,$7F,$21,$5E,$1A,$D6,$18,$BD,$BE,$6C,$3B,$61,$C6,$3F
	.dc.b	$01,$7D,$DF,$C4,$35,$CC,$FA,$32,$F2,$BC,$A3,$E0,$DF,$65,$5C,$70
	.dc.b	$DE,$3E,$6B,$E5,$8F,$CB,$C3,$FC,$DF,$8F,$A3,$1E,$F7,$E2,$BD,$C7
	.dc.b	$63,$CB,$5C,$2F,$D5,$7D,$A7,$D7,$39,$78,$1D,$09,$87,$72,$57,$C3
	.dc.b	$7E,$8F,$CC,$FF,$44,$D3,$F0,$0D,$AF,$BE,$D7,$DF,$77,$F3,$BF,$F0
	.dc.b	$9E,$7F,$BE,$9F,$87,$F4,$3B,$D4,$EE,$30,$FF,$4C,$7F,$16,$D7,$89
	.dc.b	$98,$E2,$DF,$9E,$E8,$47,$FC,$2D,$1F,$DB,$1F,$E6,$DA,$7A,$7A,$7F
	.dc.b	$A1,$F3,$DE,$9F,$E6,$D5,$A3,$F9,$E4,$7F,$82,$63,$31,$FD,$C9,$AF
	.dc.b	$F4,$D6,$BF,$D1,$7A,$DF,$59,$DD,$B1,$FD,$66,$3E,$57,$27,$F8,$7B
	.dc.b	$97,$51,$8F,$86,$FE,$D3,$C7,$FE,$CB,$33,$7F,$E2,$5E,$27,$95,$F4
	.dc.b	$76,$F9,$F5,$FE,$B8,$6D,$7E,$CA,$B6,$FF,$8E,$63,$E6,$3F,$CA,$CE
	.dc.b	$07,$CF,$7E,$86,$4F,$DE,$DC,$BF,$27,$EC,$DF,$0D,$4F,$5C,$EB,$0F
	.dc.b	$F0,$E4,$E9,$FF,$6D,$B5,$FF,$0D,$B5,$FE,$EC,$D7,$E5,$7C,$8F,$C9
	.dc.b	$23,$F0,$76,$AF,$F9,$F3,$DA,$7B,$DF,$31,$B1,$FE,$07,$AF,$9F,$A9
	.dc.b	$FC,$C7,$D9,$FE,$16,$7F,$D6,$8D,$EA,$FF,$55,$D6,$E7,$FC,$C1,$3B
	.dc.b	$7F,$CD,$ED,$7F,$DC,$ED,$7F,$C1,$AD,$DB,$F8,$D7,$B3,$F8,$32,$4F
	.dc.b	$DB,$6B,$EB,$F8,$6F,$51,$8C,$3F,$55,$F6,$4F,$5F,$F6,$0F,$F3,$78
	.dc.b	$CF,$81,$89,$CB,$AC,$66,$1F,$E7,$9B,$E1,$CB,$FE,$FE,$FB,$27,$D4
	.dc.b	$7E,$ED,$7D,$0D,$31,$16,$FF,$6D,$3A,$6C,$FE,$DE,$31,$1F,$E9,$4C
	.dc.b	$1C,$2F,$8A,$FA,$C4,$9F,$F4,$65,$7F,$94,$5F,$F4,$77,$D9,$7F,$AD
	.dc.b	$9B,$EB,$7F,$0F,$83,$7D,$9C,$39,$F4,$65,$70,$3F,$51,$E3,$FE,$17
	.dc.b	$D0,$91,$83,$FD,$21,$9A,$33,$1F,$8D,$7F,$13,$3F,$C7,$F0,$7C,$ED
	.dc.b	$EB,$97,$FE,$A5,$9C,$B6,$3F,$8D,$47,$F0,$28,$FF,$34,$B5,$FF,$5C
	.dc.b	$B5,$FF,$2A,$8F,$B1,$7C,$67,$F2,$F6,$C7,$FA,$CD,$AB,$9F,$F9,$85
	.dc.b	$3E,$C7,$F4,$3D,$12,$BF,$85,$3F,$A4,$F3,$3F,$51,$F2,$5F,$A9,$F8
	.dc.b	$79,$5F,$12,$3E,$07,$B5,$F0,$6A,$4F,$BE,$2B,$E1,$EA,$F4,$E5,$F9
	.dc.b	$78,$F8,$98,$EB,$3F,$E8,$9F,$62,$F8,$FF,$E9,$5F,$5E,$FA,$3A,$7F
	.dc.b	$4D,$9D,$61,$FE,$D1,$9D,$37,$E7,$F8,$9C,$1C,$1F,$ED,$79,$7C,$97
	.dc.b	$33,$71,$FE,$CF,$FB,$1E,$6F,$F0,$BF,$85,$FE,$14,$7F,$36,$8C,$E7
	.dc.b	$FE,$58,$F0,$3F,$37,$3A,$AF,$C4,$BB,$CF,$87,$D0,$C9,$EC,$F3,$F9
	.dc.b	$CC,$47,$F1,$9F,$3F,$F8,$C1,$8C,$7F,$B4,$6C,$F9,$AF,$8B,$D5,$2F
	.dc.b	$E1,$45,$7F,$D6,$73,$F8,$27,$F1,$21,$C3,$F8,$C6,$BF,$EA,$2F,$AF
	.dc.b	$FC,$A5,$FF,$77,$6F,$2F,$E4,$9B,$3F,$EA,$8A,$FA,$6F,$DF,$3F,$D5
	.dc.b	$5E,$BF,$DF,$71,$ED,$23,$18,$7F,$A0,$23,$1C,$DF,$37,$E7,$BF,$C1
	.dc.b	$5F,$A0,$7F,$8B,$F0,$F4,$B5,$FD,$1B,$5F,$F5,$CB,$6B,$B7,$1F,$F4
	.dc.b	$1F,$A8,$95,$FE,$C7,$8F,$F0,$07,$AE,$E9,$A4,$9F,$C1,$70,$F8,$A7
	.dc.b	$FC,$21,$8F,$F5,$B9,$FF,$6F,$B5,$F2,$C7,$F8,$5F,$1F,$EC,$BF,$43
	.dc.b	$FD,$47,$E3,$65,$7F,$BF,$69,$3F,$D5,$7E,$8E,$4F,$F1,$07,$F1,$DF
	.dc.b	$27,$FE,$2B,$F0,$7F,$27,$3F,$69,$AF,$F0,$46,$BF,$ED,$A6,$EB,$FC
	.dc.b	$BB,$E6,$BF,$B1,$7C,$E4,$AF,$E3,$B1,$FE,$10,$F8,$8F,$B7,$AF,$FB
	.dc.b	$03,$E7,$7F,$C4,$90,$75,$E5,$78,$F8,$BF,$C6,$59,$FC,$9C,$7F,$C6
	.dc.b	$CD,$7F,$1E,$1C,$3F,$C6,$B3,$7E,$67,$FD,$D2,$3F,$0B,$F6,$0F,$27
	.dc.b	$FC,$68,$D8,$FF,$7A,$F4,$FE,$BE,$B8,$3F,$EC,$8F,$A6,$7C,$37,$9E
	.dc.b	$B7,$DF,$FE,$FC,$9C,$ED,$FE,$57,$6C,$FE,$FE,$7F,$CB,$1F,$5F,$E3
	.dc.b	$8F,$D3,$B5,$FE,$62,$D4,$1C,$70,$FF,$89,$3A,$BF,$1A,$49,$FE,$97
	.dc.b	$F3,$1F,$E1,$3F,$93,$91,$FD,$39,$FF,$20,$FD,$77,$F4,$F8,$F4,$9D
	.dc.b	$B0,$7C,$07,$F9,$85,$B1,$FE,$65,$6B,$F4,$1A,$FF,$48,$CD,$CE,$F2
	.dc.b	$3E,$13,$67,$BC,$EA,$7F,$C5,$1F,$11,$F1,$FA,$7F,$C4,$15,$FB,$DA
	.dc.b	$46,$38,$BE,$93,$78,$3E,$2F,$33,$5F,$F1,$F5,$7F,$A0,$74,$75,$7A
	.dc.b	$C9,$97,$AD,$FF,$40,$5F,$F8,$1B,$E3,$3F,$C7,$44,$9F,$3F,$D7,$B8
	.dc.b	$BC,$DF,$09,$F6,$B8,$65,$7B,$36,$4E,$23,$E3,$4F,$CB,$23,$F9,$8E
	.dc.b	$5F,$F2,$EF,$A1,$FC,$18,$91,$F5,$33,$F6,$78,$18,$8F,$F2,$87,$CA
	.dc.b	$FF,$93,$CC,$70,$7F,$45,$E1,$FA,$33,$C5,$AF,$F2,$7B,$F8,$D9,$7E
	.dc.b	$0B,$5F,$C1,$BF,$E5,$9F,$95,$95,$C6,$3F,$D5,$DF,$5F,$FC,$2A,$3F
	.dc.b	$0E,$1C,$3F,$CC,$CC,$78,$9F,$EB,$5F,$C4,$9F,$43,$CA,$39,$F9,$72
	.dc.b	$B9,$DF,$2D,$7F,$D7,$0A,$7F,$D6,$4D,$9F,$F5,$B3,$5F,$E0,$2D,$7F
	.dc.b	$A8,$AF,$A7,$F4,$92,$3D,$3F,$61,$FD,$31,$F5,$47,$BB,$C1,$91,$E4
	.dc.b	$8F,$83,$F3,$CE,$0F,$0B,$67,$C2,$D7,$C2,$D7,$FC,$57,$F2,$DF,$E5
	.dc.b	$CE,$EF,$F6,$57,$C5,$1A,$49,$F5,$FB,$BF,$C3,$9B,$E3,$7C,$08,$C4
	.dc.b	$7B,$9A,$FF,$8E,$34,$7D,$A3,$C2,$FA,$FC,$DF,$C7,$DF,$83,$19,$AF
	.dc.b	$B1,$39,$7E,$D5,$72,$AD,$9F,$80,$FB,$1F,$F4,$FB,$89,$E3,$E9,$FF
	.dc.b	$BB,$DF,$F2,$4D,$A3,$11,$FC,$6A,$31,$3F,$EE,$F6,$C7,$F3,$AA,$F1
	.dc.b	$72,$7F,$BE,$67,$2F,$FB,$EE,$33,$5D,$AF,$D5,$39,$F3,$2C,$FF,$96
	.dc.b	$AB,$78,$7D,$9F,$87,$7C,$E6,$A3,$F9,$6C,$EB,$8D,$F6,$3F,$ED,$EC
	.dc.b	$3F,$CE,$BC,$1B,$9C,$FC,$C5,$CB,$FA,$1D,$7E,$DE,$F8,$63,$FA,$F6
	.dc.b	$E9,$FE,$39,$AF,$E7,$75,$BB,$E9,$5F,$E7,$C8,$FE,$67,$E1,$BE,$C7
	.dc.b	$FD,$33,$1D,$7E,$27,$D1,$39,$C3,$8C,$49,$F1,$7D,$05,$CF,$F9,$56
	.dc.b	$7F,$EE,$89,$FE,$CF,$CD,$FE,$77,$F8,$0F,$AA,$D8,$CC,$EE,$BF,$D5
	.dc.b	$5A,$3D,$7F,$91,$7B,$9F,$74,$62,$B8,$1F,$8D,$87,$D8,$7F,$88,$70
	.dc.b	$FE,$EF,$18,$F6,$3F,$0A,$D8,$CC,$EE,$3F,$C9,$78,$7F,$A3,$7C,$9F
	.dc.b	$3E,$06,$92,$FF,$C0,$7C,$5F,$6B,$E6,$3E,$69,$F3,$BE,$2E,$CF,$D7
	.dc.b	$FA,$1F,$A2,$7D,$2E,$DB,$1F,$D5,$B2,$FF,$70,$3F,$21,$B9,$F9,$1A
	.dc.b	$FF,$61,$6B,$FD,$53,$7F,$E4,$33,$F8,$5F,$1F,$EF,$F9,$DF,$80,$7F
	.dc.b	$03,$E0,$C9,$FC,$57,$CA,$DC,$3F,$B1,$1F,$C8,$E3,$EC,$B4,$F8,$A5
	.dc.b	$3F,$B9,$14,$F1,$7C,$CF,$E3,$8F,$C7,$FD,$0D,$9F,$CC,$7C,$07,$E7
	.dc.b	$7E,$A7,$F3,$77,$F9,$98,$C7,$7F,$E9,$E1,$F9,$F7,$D5,$7F,$CF,$ED
	.dc.b	$7F,$D9,$2D,$7F,$C3,$2D,$E4,$7C,$DB,$8F,$F5,$8F,$F6,$8C,$BF,$B0
	.dc.b	$39,$6F,$07,$E9,$BE,$47,$F4,$FF,$29,$67,$F6,$0F,$97,$1F,$81,$97
	.dc.b	$83,$D0,$FC,$6C,$62,$3D,$31,$FC,$53,$3F,$DF,$FC,$47,$EC,$8F,$D9
	.dc.b	$FC,$2D,$C7,$F7,$93,$FB,$6D,$3F,$6E,$D8,$FE,$ED,$F5,$5F,$EC,$97
	.dc.b	$F2,$2D,$C7,$FD,$47,$28,$C4,$7F,$95,$5A,$FF,$86,$9F,$F1,$44,$F4
	.dc.b	$F8,$1F,$E2,$11,$FA,$3E,$EF,$F6,$13,$AA,$FE,$F8,$FF,$9F,$AB,$C7
	.dc.b	$F1,$DC,$3D,$6E,$2F,$2D,$7E,$05,$7C,$6C,$BC,$91,$83,$3E,$A7,$80
	.dc.b	$F0,$FD,$85,$C7,$FC,$03,$F5,$AF,$9D,$FD,$CB,$63,$F7,$AD,$D3,$FC
	.dc.b	$EA,$D7,$FD,$8A,$BE,$67,$89,$27,$8B,$FC,$13,$EB,$BF,$DA,$51,$E0
	.dc.b	$8F,$C9,$46,$2B,$FC,$21,$5C,$EB,$E9,$FC,$57,$C8,$F6,$FE,$4B,$63
	.dc.b	$24,$AF,$93,$F2,$7A,$8F,$B7,$F3,$1F,$40,$FA,$3E,$8D,$8F,$BE,$E1
	.dc.b	$FC,$0F,$C2,$5B,$F1,$FC,$0F,$E9,$F3,$FF,$31,$D7,$F8,$A3,$45,$E6
	.dc.b	$BD,$6F,$BC,$93,$FC,$9A,$76,$DC,$36,$F3,$6E,$1F,$0B,$E1,$3C,$07
	.dc.b	$C6,$F8,$FB,$5E,$17,$0F,$8F,$EB,$6D,$7E,$29,$FC,$79,$F9,$09,$B3
	.dc.b	$F9,$EF,$82,$C3,$4F,$D0,$C7,$F0,$5F,$1B,$F7,$57,$FE,$B4,$6B,$FE
	.dc.b	$9A,$3F,$D3,$6E,$1C,$35,$CF,$97,$FD,$12,$3F,$94,$C7,$38,$C1,$F8
	.dc.b	$5A,$F0,$3F,$84,$69,$FB,$4F,$81,$FD,$49,$FA,$AF,$9C,$B1,$FB,$87
	.dc.b	$0F,$EE,$BE,$76,$DF,$F9,$6F,$C3,$67,$FC,$85,$B5,$FE,$62,$FC,$0B
	.dc.b	$FC,$1B,$49,$CB,$FE,$63,$8C,$9F,$E0,$DE,$EF,$F1,$A5,$71,$F8,$1F
	.dc.b	$E5,$9F,$5B,$FE,$08,$BF,$8C,$1C,$3F,$25,$AF,$FB,$11,$AE,$FD,$CF
	.dc.b	$ED,$70,$69,$2B,$FB,$2F,$D5,$FF,$8E,$AB,$FC,$F1,$5B,$9F,$F4,$84
	.dc.b	$E7,$C5,$FE,$F5,$39,$AF,$F3,$F7,$C1,$BE,$47,$FA,$23,$70,$E3,$E0
	.dc.b	$E2,$7F,$C5,$18,$FF,$03,$C7,$FC,$A1,$E2,$A3,$17,$FE,$BA,$7F,$DA
	.dc.b	$2F,$23,$C7,$E1,$72,$39,$EB,$63,$CE,$E1,$F4,$FC,$FD,$D3,$FD,$41
	.dc.b	$6F,$9D,$FE,$80,$DA,$F8,$71,$83,$3F,$03,$E8,$7D,$2D,$8F,$8B,$F3
	.dc.b	$1F,$E3,$2F,$AD,$FF,$1D,$E1,$DA,$30,$FC,$06,$4F,$CC,$F8,$9B,$3F
	.dc.b	$EB,$3F,$AD,$F9,$0F,$F4,$66,$BF,$A7,$D9,$18,$8F,$F6,$FB,$5F,$F4
	.dc.b	$A6,$7F,$E2,$3B,$FF,$64,$4F,$4D,$FE,$7C,$E6,$BC,$1C,$1F,$53,$E5
	.dc.b	$7E,$19,$30,$FC,$2B,$F1,$BC,$5D,$47,$E1,$1B,$5F,$33,$E6,$3F,$CA
	.dc.b	$7F,$39,$F8,$5C,$BB,$C6,$3A,$3F,$02,$CC,$3F,$1B,$F2,$77,$05,$FF
	.dc.b	$A4,$FE,$2D,$F5,$6E,$6F,$F4,$4C,$61,$B1,$9C,$3F,$DD,$4D,$BF,$ED
	.dc.b	$FB,$FC,$19,$42,$F9,$47,$E7,$B7,$FF,$35,$DF,$1C,$7F,$D3,$5F,$21
	.dc.b	$FE,$54,$C0,$D2,$5F,$F9,$6B,$DD,$FF,$3A,$CF,$EE,$0F,$F0,$E7,$95
	.dc.b	$FE,$5C,$8C,$C7,$EF,$1A,$FE,$7F,$E6,$3F,$C5,$91,$CB,$0F,$AF,$18
	.dc.b	$AF,$F1,$85,$7A,$BE,$5F,$D7,$ED,$7E,$C6,$3F,$96,$7C,$07,$FA,$EE
	.dc.b	$31,$1F,$CB,$E7,$7A,$FF,$0C,$AD,$3E,$78,$CC,$F2,$CE,$BC,$4F,$F1
	.dc.b	$BC,$F1,$8C,$C7,$8C,$7D,$6F,$81,$FF,$24,$B5,$57,$D2,$8D,$5F,$FB
	.dc.b	$45,$D4,$66,$FB,$57,$47,$CC,$D3,$CC,$53,$CD,$3D,$2F,$94,$7F,$2A
	.dc.b	$AA,$FE,$0A,$FC,$17,$D3,$0E,$13,$EE,$38,$14,$E1,$7F,$01,$F5,$9E
	.dc.b	$13,$DF,$3E,$6C,$39,$9E,$C8,$E4,$7F,$9B,$2B,$07,$80,$B3,$F0,$1B
	.dc.b	$1E,$E7,$97,$4F,$F9,$A2,$3C,$AD,$7E,$B3,$5F,$A8,$E7,$E5,$5F,$09
	.dc.b	$FA,$E9,$4E,$63,$0D,$DB,$F4,$ED,$9D,$DF,$F7,$9B,$E7,$5F,$86,$5F
	.dc.b	$0B,$FD,$F6,$7F,$23,$18,$AD,$61,$E1,$9C,$B7,$1E,$97,$C5,$B5,$C3
	.dc.b	$08,$FD,$FC,$3A,$36,$BF,$C1,$99,$74,$AF,$BF,$5E,$6A,$FB,$4F,$0C
	.dc.b	$7C,$B7,$DE,$7F,$82,$CE,$E7,$94,$75,$39,$CF,$6C,$3E,$16,$3F,$2F
	.dc.b	$1D,$9D,$2B,$67,$8A,$35,$3C,$31,$CC,$F5,$3D,$93,$A7,$C6,$FC,$57
	.dc.b	$CB,$7E,$75,$8F,$AE,$FA,$47,$BA,$B8,$3A,$AE,$15,$B8,$EB,$7F,$C0
	.dc.b	$32,$FE,$E5,$E9,$7F,$8D,$63,$FD,$F1,$F5,$BF,$E1,$26,$D6,$75,$FE
	.dc.b	$ED,$56,$3F,$96,$D7,$FB,$F1,$A5,$70,$9B,$83,$E0,$F8,$36,$7F,$95
	.dc.b	$1E,$E8,$C9,$8C,$7F,$C1,$1C,$9E,$1E,$6F,$E8,$FB,$5E,$6F,$E4,$3A
	.dc.b	$FF,$60,$AC,$4F,$4A,$DE,$3F,$D2,$23,$17,$FC,$5E,$3F,$B5,$F1,$7B
	.dc.b	$71,$FF,$0B,$37,$9C,$DF,$23,$0F,$8F,$B5,$F2,$34,$CE,$1F,$E2,$C3
	.dc.b	$A7,$1B,$E5,$29,$D6,$29,$FD,$2F,$2F,$D6,$C7,$C1,$9E,$13,$96,$BF
	.dc.b	$EC,$E7,$FC,$80,$FF,$60,$DF,$FC,$5F,$F5,$8F,$89,$F6,$36,$75,$CD
	.dc.b	$FC,$02,$F8,$C7,$F7,$48,$FE,$07,$3B,$8F,$F6,$DB,$5F,$F0,$B3,$5D
	.dc.b	$D7,$9F,$2F,$79,$23,$B9,$FC,$4A,$DF,$29,$98,$F0,$1F,$DD,$F4,$7C
	.dc.b	$8F,$CD,$35,$9F,$F9,$2A,$DF,$23,$F3,$26,$5F,$6D,$BF,$8A,$CE,$23
	.dc.b	$E3,$4F,$CB,$23,$F2,$FC,$0F,$AA,$47,$95,$BE,$47,$EC,$FC,$6C,$9C
	.dc.b	$BB,$7D,$38,$FF,$24,$F8,$EF,$BE,$48,$E7,$ED,$FE,$A5,$7D,$4F,$FA
	.dc.b	$25,$F8,$A6,$3D,$8D,$46,$63,$AE,$4B,$FC,$5F,$C8,$73,$FA,$9C,$57
	.dc.b	$26,$BC,$23,$53,$FB,$79,$CC,$FF,$A4,$1A,$FF,$AD,$DA,$FF,$8A,$94
	.dc.b	$F1,$24,$7E,$E2,$32,$FE,$EB,$F1,$57,$C3,$FB,$DB,$1E,$E6,$BE,$CF
	.dc.b	$17,$E2,$7D,$53,$E0,$69,$27,$85,$FE,$AB,$A6,$47,$D6,$7F,$5B,$A7
	.dc.b	$EB,$BC,$49,$3F,$69,$DB,$F8,$E1,$F3,$BF,$92,$7E,$8E,$15,$D3,$7A
	.dc.b	$1D,$BF,$ED,$BA,$C3,$FE,$DD,$9E,$15,$FE,$BA,$6B,$9E,$BF,$E4,$D1
	.dc.b	$88,$FF,$2A,$BF,$C2,$1F,$E4,$2D,$7F,$C9,$8D,$7A,$76,$7F,$03,$F9
	.dc.b	$4F,$E1,$7F,$27,$27,$F2,$F1,$FC,$3B,$9A,$47,$E6,$9D,$3F,$07,$ED
	.dc.b	$FF,$79,$A7,$B2,$3F,$CC,$F1,$8C,$FF,$DF,$0D,$8C,$F5,$7E,$67,$36
	.dc.b	$34,$93,$F9,$AA,$E0,$BE,$A9,$F1,$DF,$A1,$5F,$1D,$4C,$45,$9F,$E8
	.dc.b	$73,$A6,$CF,$F4,$F8,$C4,$7F,$9C,$7E,$9D,$F5,$3F,$AD,$6C,$7E,$B1
	.dc.b	$AF,$E0,$67,$8E,$BF,$C2,$FB,$FF,$9A,$92,$7F,$80,$FA,$5F,$D0,$FE
	.dc.b	$AB,$8D,$7F,$B6,$9B,$C1,$FE,$28,$6D,$7F,$1D,$BE,$1F,$2D,$23,$F8
	.dc.b	$A4,$A3,$A7,$CF,$66,$7F,$84,$CB,$E6,$7C,$06,$71,$FA,$D3,$C7,$CB
	.dc.b	$F7,$76,$C8,$FE,$13,$8F,$F4,$4A,$DE,$DF,$AE,$9F,$F5,$19,$FE,$A4
	.dc.b	$70,$E1,$AE,$6B,$FB,$E4,$7D,$88,$FE,$EA,$D7,$E8,$70,$E8,$FE,$ED
	.dc.b	$93,$E6,$3E,$41,$27,$E0,$FC,$CF,$CB,$F9,$8D,$92,$3F,$B0,$52,$39
	.dc.b	$FC,$C7,$D1,$F9,$99,$5F,$2A,$3E,$A6,$32,$3E,$6B,$F5,$F3,$FB,$1C
	.dc.b	$DE,$42,$4F,$67,$CF,$EE,$FD,$DF,$98,$93,$F5,$A3,$F0,$1F,$2D,$23
	.dc.b	$EC,$BF,$83,$F9,$7F,$F1,$17,$CE,$3E,$D1,$1F,$E1,$A8,$C3,$63,$35
	.dc.b	$FE,$C7,$9C,$BF,$03,$BF,$F1,$DF,$2D,$23,$F0,$F1,$FE,$13,$F8,$3F
	.dc.b	$E1,$F1,$FE,$7A,$7F,$83,$73,$7E,$E7,$AF,$FD,$5F,$E2,$EF,$C6,$FE
	.dc.b	$9D,$F1,$9F,$EE,$BC,$3F,$C5,$AD,$7F,$DB,$0D,$7F,$0A,$1C,$3F,$9D
	.dc.b	$F9,$7F,$F1,$27,$D4,$3E,$A2,$9F,$A5,$8F,$D3,$78,$52,$BF,$3E,$FE
	.dc.b	$A7,$E6,$BE,$3F,$BD,$F8,$99,$FC,$9F,$0F,$F8,$B3,$A7,$E4,$4E,$67
	.dc.b	$FD,$F3,$39,$7F,$C4,$9F,$4D,$E1,$7F,$BF,$35,$FE,$E2,$DE,$0F,$88
	.dc.b	$BE,$BB,$FC,$78,$BE,$62,$92,$B9,$FC,$CF,$9B,$5F,$19,$23,$EC,$64
	.dc.b	$8F,$F2,$A5,$BE,$67,$A9,$2B,$CB,$F1,$FE,$0F,$98,$91,$E8,$7C,$7F
	.dc.b	$33,$F8,$AE,$8F,$F0,$FC,$FE,$15,$AF,$D4,$DF,$FD,$D7,$F0,$7F,$8D
	.dc.b	$9C,$CF,$D8,$FA,$97,$CB,$FF,$9C,$DB,$1F,$E6,$C6,$BF,$78,$38,$7F
	.dc.b	$82,$F5,$B5,$FC,$14,$7F,$8E,$31,$FF,$7D,$FA,$BE,$73,$E5,$F5,$59
	.dc.b	$F4,$47,$94,$8F,$D8,$FA,$3F,$A2,$47,$EC,$FC,$D5,$7F,$C8,$3F,$1F
	.dc.b	$FE,$57,$F1,$BF,$CF,$DE,$F1,$D7,$57,$DC,$F9,$6F,$B6,$4B,$FC,$1F
	.dc.b	$05,$7F,$07,$F3,$3F,$80,$31,$5F,$CA,$BD,$E7,$C4,$7F,$8C,$DB,$5F
	.dc.b	$E3,$26,$BF,$8C,$F9,$8D,$4F,$F8,$E7,$EA,$1E,$A5,$33,$5F,$E3,$A8
	.dc.b	$CB,$FE,$9C,$6B,$F4,$5A,$FE,$C1,$AF,$F3,$6F,$A8,$FD,$EF,$8B,$7C
	.dc.b	$9F,$F2,$44,$62,$47,$F9,$AB,$D4,$CC,$F9,$7D,$49,$3F,$E4,$CA,$E0
	.dc.b	$B0,$79,$AE,$BF,$33,$67,$CC,$D7,$CC,$D7,$FA,$5F,$CB,$7F,$91,$3A
	.dc.b	$FF,$D7,$2D,$8F,$C7,$35,$FE,$1A,$D7,$FA,$9D,$FF,$8F,$FC,$5F,$8D
	.dc.b	$E7,$7F,$AB,$A7,$E3,$35,$F8,$CD,$7E,$0E,$3F,$97,$F9,$AF,$F0,$D7
	.dc.b	$27,$F0,$0F,$96,$BC,$3A,$F0,$15,$8C,$3E,$04,$2B,$18,$7E,$87,$E6
	.dc.b	$6C,$F0,$C3,$F8,$7C,$63,$0C,$FC,$C7,$F7,$4C,$7B,$35,$EA,$DC,$3F
	.dc.b	$58,$DC,$1F,$2D,$CB,$1F,$E3,$6D,$BF,$E3,$7A,$F8,$A3,$C6,$D9,$FF
	.dc.b	$4F,$7E,$0D,$FE,$CB,$7F,$91,$D7,$AA,$30,$FF,$24,$3E,$07,$81,$FE
	.dc.b	$C3,$F5,$B5,$F0,$18,$F0,$BF,$AC,$E1,$D2,$39,$7A,$DC,$FE,$72,$CF
	.dc.b	$81,$E1,$D1,$8C,$BF,$73,$87,$E3,$2B,$CB,$7D,$7D,$0F,$89,$9F,$FB
	.dc.b	$4D,$B5,$FC,$AE,$BF,$8A,$4E,$5F,$F7,$16,$9F,$EE,$C7,$FD,$93,$19
	.dc.b	$C3,$FD,$5F,$39,$7C,$37,$F0,$7C,$C9,$7F,$1B,$E5,$AD,$7F,$0E,$FD
	.dc.b	$8F,$FA,$24,$F8,$1B,$1F,$D8,$B1,$46,$23,$FC,$53,$18,$F5,$7E,$E6
	.dc.b	$4F,$97,$FA,$1F,$2F,$F4,$5B,$8F,$A3,$1F,$ED,$0D,$F6,$BE,$B7,$9F
	.dc.b	$F5,$27,$39,$F9,$7D,$7F,$8F,$06,$92,$FD,$1E,$C7,$DF,$F2,$79,$36
	.dc.b	$3C,$1E,$C7,$53,$F0,$9F,$2D,$63,$C9,$EB,$F6,$23,$F0,$DF,$0F,$AB
	.dc.b	$F8,$BF,$96,$F3,$36,$7F,$23,$F2,$5F,$8F,$F9,$4F,$C8,$36,$3E,$07
	.dc.b	$A9,$EA,$23,$F3,$DF,$93,$F2,$7F,$97,$EB,$FC,$C3,$67,$F3,$BF,$27
	.dc.b	$F9,$BF,$92,$FC,$E3,$63,$E6,$79,$1F,$18,$8F,$DB,$7E,$AF,$C7,$FE
	.dc.b	$8F,$E4,$FF,$48,$D9,$FA,$9E,$37,$E9,$FC,$AF,$A4,$D8,$FB,$1D,$1F
	.dc.b	$48,$8F,$E1,$DF,$CD,$FA,$0F,$1B,$F8,$4F,$AF,$FE,$92,$AF,$CB,$CE
	.dc.b	$5F,$E6,$FE,$7F,$F8,$DE,$75,$7F,$D8,$2B,$F3,$53,$9F,$90,$FD,$BC
	.dc.b	$62,$FF,$8F,$DA,$73,$E2,$F8,$89,$3F,$AA,$A4,$7F,$13,$C1,$F4,$BF
	.dc.b	$8B,$AF,$F0,$14,$63,$BB,$53,$98,$FF,$03,$35,$FF,$02,$E3,$FE,$99
	.dc.b	$AF,$59,$E4,$F9,$0F,$83,$F2,$3F,$09,$F4,$F8,$9F,$18,$93,$E5,$7F
	.dc.b	$92,$57,$F6,$A8,$F8,$F1,$F2,$BD,$BE,$31,$DB,$0F,$EA,$D5,$FE,$8D
	.dc.b	$8C,$70,$A3,$12,$DF,$E7,$7B,$3F,$53,$E4,$3E,$AB,$30,$E3,$7F,$E8
	.dc.b	$F6,$C7,$DA,$F8,$EF,$B7,$F1,$F6,$76,$FD,$EF,$89,$C6,$5D,$E7,$EA
	.dc.b	$C6,$32,$FB,$DF,$1F,$F8,$43,$F0,$BF,$09,$6B,$C2,$E1,$F1,$FA,$D7
	.dc.b	$0E,$87,$C0,$7F,$1A,$7E,$3B,$E3,$EC,$FE,$3B,$BF,$F2,$3F,$1F,$F8
	.dc.b	$86,$C7,$BB,$D7,$C3,$4E,$07,$C2,$F7,$BF,$32,$7E,$6B,$D7,$B1,$F1
	.dc.b	$BE,$37,$0D,$3E,$4C,$7E,$81,$87,$D7,$CD,$FB,$37,$F3,$D7,$D7,$CD
	.dc.b	$F5,$E7,$FB,$94,$6A,$FF,$2D,$46,$92,$7F,$2F,$07,$8B,$B7,$C0,$E5
	.dc.b	$B1,$FD,$F3,$F0,$5F,$E5,$BC,$3F,$EA,$77,$F9,$FB,$7D,$5D,$7B,$1B
	.dc.b	$EF,$E1,$17,$0F,$C7,$DF,$27,$FC,$63,$EC,$BF,$7C,$FF,$7C,$46,$BD
	.dc.b	$5F,$ED,$F1,$8D,$7E,$0D,$FF,$A4,$6B,$9F,$C8,$6B,$C2,$F3,$36,$3B
	.dc.b	$FA,$DE,$93,$E3,$9F,$E7,$E7,$27,$F8,$F7,$87,$F7,$DF,$0D,$FE,$3B
	.dc.b	$8F,$DE,$B7,$1F,$F4,$03,$5F,$F5,$FD,$3E,$2B,$E9,$CE,$F8,$7E,$75
	.dc.b	$7F,$93,$AB,$57,$DC,$FE,$E7,$A7,$EC,$2B,$E6,$7A,$FF,$03,$C3,$FA
	.dc.b	$0F,$D1,$F9,$0C,$5F,$CA,$6E,$3F,$E8,$0C,$1E,$0F,$D7,$F5,$FE,$4B
	.dc.b	$6F,$E8,$FC,$6E,$1A,$78,$A3,$E7,$F8,$7F,$E0,$EB,$34,$93,$FE,$A1
	.dc.b	$E0,$71,$BE,$3B,$8C,$FD,$E6,$E7,$F6,$38,$BF,$15,$EB,$FE,$2D,$B5
	.dc.b	$FE,$14,$F0,$5E,$FF,$F8,$79,$B1,$FE,$1C,$6D,$FF,$83,$5A,$FF,$83
	.dc.b	$34,$78,$7E,$92,$57,$F8,$92,$32,$FA,$73,$7C,$77,$E3,$AB,$F3,$CD
	.dc.b	$9F,$CB,$F9,$5F,$E5,$1A,$FD,$86,$1F,$EF,$ED,$BE,$0C,$E4,$D7,$C4
	.dc.b	$FE,$9D,$FD,$43,$6F,$F4,$3F,$13,$FA,$BF,$63,$F5,$8D,$8F,$D2,$FC
	.dc.b	$4E,$1A,$70,$3F,$53,$F1,$1E,$32,$3E,$B7,$CD,$EC,$7E,$DD,$FD,$C3
	.dc.b	$63,$F8,$66,$7F,$EA,$0F,$90,$FF,$50,$FD,$0F,$F9,$E6,$3F,$D4,$62
	.dc.b	$BF,$33,$19,$8C,$CF,$0C,$F7,$1B,$BC,$38,$9F,$F5,$6F,$B7,$C6,$34
	.dc.b	$1C,$28,$DD,$74,$C3,$F5,$35,$A9,$EB,$5F,$EB,$89,$F2,$D7,$FB,$CE
	.dc.b	$73,$3D,$B0,$E0,$9F,$15,$70,$8F,$31,$9A,$F0,$61,$CA,$3C,$87,$9E
	.dc.b	$32,$D3,$8E,$1D,$27,$D3,$1D,$2F,$0F,$53,$15,$CA,$3C,$A6,$8A,$E4
	.dc.b	$F6,$47,$13,$E0,$4F,$97,$0F,$7E,$1E,$B9,$FA,$A7,$94,$D9,$4F,$C0
	.dc.b	$15,$F9,$B5,$DA,$FC,$4D,$3C,$53,$EE,$7E,$0E,$1F,$30,$F6,$95,$C9
	.dc.b	$E8,$72,$FA,$B6,$FC,$04,$FD,$06,$BE,$86,$BE,$67,$3E,$CB,$E0,$7F
	.dc.b	$C2,$93,$FE,$EF,$BF,$F7,$4C,$62,$B6,$D8,$CF,$83,$FE,$DE,$6D,$7C
	.dc.b	$5B,$7C,$0F,$E8,$27,$F3,$D1,$89,$FB,$61,$F1,$56,$1B,$1C,$EB,$96
	.dc.b	$3F,$EF,$C6,$C0,$E5,$87,$AB,$0E,$EF,$B3,$2F,$0D,$7E,$C9,$B5,$FA
	.dc.b	$0C,$3B,$C7,$0C,$7C,$55,$DA,$BC,$CD,$AF,$2E,$1B,$9F,$C5,$47,$B5
	.dc.b	$B3,$DF,$1F,$8B,$86,$CF,$0C,$E8,$F1,$63,$E7,$9D,$9E,$F8,$F3,$57
	.dc.b	$37,$E0,$5F,$A2,$7E,$8E,$1D,$08,$E0,$E7,$3C,$2B,$93,$EC,$3C,$38
	.dc.b	$C7,$C8,$4C,$7D,$15,$CE,$39,$1F,$18,$FB,$18,$FD,$18,$EF,$1D,$4E
	.dc.b	$4F,$8B,$2F,$EB,$B1,$FC,$92,$FF,$BD,$C3,$F0,$27,$C5,$E7,$7F,$BD
	.dc.b	$39,$8C,$5A,$39,$BF,$E7,$0A,$E2,$FF,$6B,$6B,$FD,$35,$AF,$F3,$48
	.dc.b	$E8,$EF,$DF,$EB,$1E,$3F,$8F,$B4,$6C,$E1,$CF,$C6,$4B,$F1,$47,$0F
	.dc.b	$80,$E3,$F1,$92,$7C,$D1,$5F,$24,$7F,$A8,$31,$F8,$4D,$C4,$DE,$FF
	.dc.b	$83,$E7,$FF,$8F,$7C,$0B,$E0,$1F,$3C,$E1,$FF,$55,$CE,$F1,$FE,$F1
	.dc.b	$18,$AE,$11,$FE,$03,$F8,$3D,$C9,$A4,$8F,$99,$F1,$7F,$13,$E0,$64
	.dc.b	$7D,$18,$F8,$FF,$15,$E4,$F6,$64,$FD,$58,$AF,$F5,$EC,$BF,$CA,$D3
	.dc.b	$E3,$24,$7D,$9F,$8A,$F7,$F3,$BD,$8A,$7D,$48,$B3,$F8,$1F,$8A,$FA
	.dc.b	$E6,$1F,$F3,$EF,$0F,$F1,$F7,$FD,$71,$8B,$07,$AF,$D8,$FB,$FD,$76
	.dc.b	$BF,$19,$DB,$F8,$23,$0F,$FA,$88,$7E,$24,$F8,$3F,$CA,$C3,$E9,$5E
	.dc.b	$CE,$1B,$1F,$E9,$3A,$E0,$FF,$A5,$63,$2F,$F8,$08,$E5,$3F,$EE,$46
	.dc.b	$9F,$EE,$68,$DB,$F9,$BD,$1F,$19,$F9,$82,$47,$FA,$87,$E1,$DB,$FE
	.dc.b	$73,$E3,$3E,$6F,$B3,$2B,$F5,$51,$5F,$93,$F0,$9F,$A2,$31,$E2,$FE
	.dc.b	$E3,$27,$C6,$7E,$9B,$C8,$97,$FB,$38,$AF,$D2,$EC,$FD,$59,$8B,$FF
	.dc.b	$3D,$C3,$F6,$6F,$F2,$7B,$FE,$B7,$5F,$0F,$F3,$1B,$19,$F5,$71,$3F
	.dc.b	$16,$73,$3F,$25,$AF,$F8,$B1,$AF,$F0,$CC,$BF,$85,$3F,$C2,$7C,$4F
	.dc.b	$DF,$9D,$FC,$1F,$E0,$7F,$87,$3E,$72,$FC,$7C,$BF,$F7,$FC,$E5,$BF
	.dc.b	$09,$C0,$8E,$A7,$9F,$C7,$7E,$73,$DA,$93,$CA,$2B,$FE,$D4,$B7,$3F
	.dc.b	$88,$CF,$07,$F2,$88,$7C,$AF,$98,$95,$E4,$F8,$4F,$4F,$C7,$C8,$F3
	.dc.b	$BA,$7B,$7C,$17,$B8,$8E,$8F,$6F,$81,$F0,$BE,$3E,$46,$0F,$82,$FC
	.dc.b	$6F,$83,$91,$83,$E2,$F3,$FC,$A2,$3F,$23,$E3,$FE,$6F,$44,$8F,$96
	.dc.b	$E9,$F8,$5C,$BF,$C8,$3E,$DF,$F9,$DF,$07,$FB,$0B,$C9,$46,$3D,$8F
	.dc.b	$E8,$FC,$5F,$C7,$30,$5E,$EB,$FA,$0C,$7F,$B9,$9A,$AF,$9F,$B1,$FC
	.dc.b	$A2,$73,$D4,$C1,$F1,$FF,$84,$26,$1F,$82,$F9,$0F,$E0,$3F,$39,$FE
	.dc.b	$BE,$8F,$DD,$07,$0F,$EF,$5A,$FF,$A0,$1A,$FF,$87,$73,$7E,$42,$F6
	.dc.b	$F9,$C6,$1B,$19,$9F,$E1,$13,$97,$F8,$56,$BF,$1A,$BF,$37,$E9,$C9
	.dc.b	$FE,$47,$AF,$17,$E3,$EF,$F1,$7D,$AF,$D0,$FC,$8C,$9D,$C7,$D3,$F6
	.dc.b	$FE,$31,$23,$BD,$FF,$58,$6F,$B3,$FA,$AC,$24,$78,$23,$FB,$7F,$2B
	.dc.b	$E4,$2F,$B1,$2B,$E1,$5F,$F8,$0A,$B5,$1E,$8F,$63,$EB,$11,$F0,$BF
	.dc.b	$DA,$E3,$F9,$EC,$72,$3F,$C1,$0B,$E4,$7C,$27,$8B,$D8,$F1,$FC,$87
	.dc.b	$93,$D6,$97,$C7,$AB,$F8,$9C,$FE,$4F,$D2,$FC,$31,$27,$E0,$BE,$1F
	.dc.b	$90,$F6,$FA,$92,$79,$66,$9F,$CE,$7A,$7F,$0B,$E4,$25,$7D,$48,$FD
	.dc.b	$06,$5F,$93,$24,$7C,$E7,$FB,$EF,$EF,$8F,$3F,$FC,$0A,$D9,$FE,$BE
	.dc.b	$D7,$EE,$9F,$0B,$E4,$77,$E8,$FD,$2F,$91,$93,$C2,$39,$FA,$3D,$89
	.dc.b	$1F,$8C,$B4,$78,$BD,$1F,$09,$27,$DF,$F2,$3F,$7B,$D3,$FB,$FF,$23
	.dc.b	$27,$E1,$C7,$A3,$D2,$F5,$92,$3B,$3F,$6F,$E0,$7F,$0E,$48,$F9,$7F
	.dc.b	$21,$F8,$AF,$90,$F8,$3E,$A4,$9F,$9F,$15,$FF,$08,$6B,$FD,$C1,$B3
	.dc.b	$F2,$8F,$97,$EA,$7E,$53,$E3,$E4,$FC,$A8,$AF,$CA,$E4,$FE,$AB,$F2
	.dc.b	$1F,$9A,$33,$E6,$7F,$15,$C7,$FA,$DD,$6F,$A9,$F1,$DF,$A0,$3E,$BF
	.dc.b	$AD,$67,$F0,$BF,$21,$FA,$43,$0F,$13,$FC,$AB,$F6,$7F,$C7,$DF,$F3
	.dc.b	$13,$73,$FE,$BF,$97,$F9,$B2,$30,$3E,$3B,$89,$CB,$D7,$B7,$FE,$0F
	.dc.b	$B5,$6E,$30,$D7,$FD,$60,$7F,$9D,$1C,$3F,$E6,$B7,$F8,$55,$F8,$7C
	.dc.b	$EF,$39,$E8,$F4,$6D,$7E,$AE,$BF,$CD,$8D,$F8,$EC,$BF,$E2,$DF,$99
	.dc.b	$7A,$FF,$DD,$A3,$C2,$23,$39,$7F,$2E,$6C,$7F,$8A,$5A,$F0,$F9,$1F
	.dc.b	$F1,$0F,$CD,$7F,$34,$6C,$7F,$8B,$DA,$EF,$E2,$13,$D7,$DB,$FB,$33
	.dc.b	$D5,$AF,$92,$FF,$A2,$F1,$EC,$F9,$3E,$A3,$D2,$F8,$1E,$B5,$BF,$C5
	.dc.b	$6E,$AC,$71,$7D,$9A,$B7,$F3,$E6,$DF,$E0,$1B,$B7,$D8,$F9,$1B,$8F
	.dc.b	$3B,$FE,$31,$F0,$DF,$81,$7E,$F3,$70,$F9,$77,$FE,$31,$E6,$FF,$19
	.dc.b	$C7,$FA,$C2,$74,$FF,$A9,$DF,$F5,$15,$74,$9F,$0C,$62,$7F,$48,$F5
	.dc.b	$CF,$FD,$7A,$E5,$BC,$7C,$39,$38,$BE,$8F,$57,$D3,$F2,$17,$2F,$C3
	.dc.b	$37,$1F,$8F,$3E,$EF,$90,$C6,$3F,$AF,$F7,$3E,$0F,$37,$C6,$F7,$5D
	.dc.b	$E9,$CE,$7F,$C9,$1D,$46,$63,$F2,$9F,$21,$89,$F9,$F9,$BE,$FF,$FA
	.dc.b	$B5,$7F,$7C,$6C,$7F,$9D,$DA,$FE,$59,$AF,$EA,$3F,$7A,$79,$1D,$27
	.dc.b	$8E,$9F,$E6,$29,$CD,$FF,$57,$BF,$07,$BD,$F7,$3E,$4A,$4F,$F5,$BE
	.dc.b	$E7,$E1,$3F,$E9,$B8,$FD,$64,$F5,$F7,$FE,$DE,$03,$D4,$E1,$B3,$EB
	.dc.b	$8C,$62,$F2,$3E,$0B,$67,$F7,$1F,$23,$89,$FC,$9E,$4F,$43,$FA,$95
	.dc.b	$79,$5B,$8F,$E7,$BD,$DD,$C6,$36,$FF,$40,$EE,$F4,$BC,$8D,$C3,$E7
	.dc.b	$46,$0F,$5F,$77,$F1,$6F,$90,$FC,$F9,$FA,$0F,$91,$B5,$FD,$83,$E1
	.dc.b	$7E,$E9,$FA,$4F,$91,$B1,$F3,$71,$F4,$3F,$6F,$3F,$09,$1F,$F1,$CF
	.dc.b	$DA,$7F,$9B,$9F,$A8,$DD,$BF,$73,$EF,$FE,$1A,$31,$B7,$F3,$1D,$7F
	.dc.b	$25,$FB,$4F,$DE,$C3,$F9,$44,$63,$A3,$F1,$BD,$5F,$EB,$49,$FF,$23
	.dc.b	$B7,$2F,$F6,$13,$5F,$E4,$2D,$7E,$E5,$7F,$B3,$FE,$4B,$BF,$C3,$FF
	.dc.b	$04,$F9,$4F,$F4,$C6,$1F,$EE,$0C,$BF,$DB,$EF,$F8,$6E,$1F,$01,$BF
	.dc.b	$83,$E0,$D9,$FC,$96,$6F,$03,$F1,$0D,$8F,$3C,$60,$FA,$B7,$FE,$98
	.dc.b	$8E,$DF,$01,$E3,$F8,$1F,$23,$F1,$3E,$0F,$CD,$F2,$56,$BC,$8F,$EC
	.dc.b	$3F,$10,$FF,$76,$47,$F9,$A9,$BC,$3F,$EE,$56,$BF,$D1,$1A,$FE,$7F
	.dc.b	$C7,$6E,$F8,$4F,$A6,$7C,$5F,$94,$B8,$7C,$5E,$9F,$91,$F0,$BF,$61
	.dc.b	$B1,$FB,$3F,$9D,$FF,$24,$3F,$E7,$C6,$E5,$FE,$F7,$6B,$FD,$59,$AF
	.dc.b	$8F,$1F,$DB,$3F,$E3,$2C,$BC,$BF,$0D,$F5,$FE,$57,$F0,$EF,$D9,$F9
	.dc.b	$59,$8F,$D7,$F8,$7B,$87,$F8,$E8,$7C,$EF,$FA,$52,$7F,$D1,$ED,$8F
	.dc.b	$EC,$E6,$BD,$FF,$E0,$11,$98,$FE,$F7,$DA,$E7,$7C,$3F,$F2,$0A,$E5
	.dc.b	$E0,$7F,$57,$AD,$4E,$70,$FC,$77,$C3,$FE,$3F,$E5,$AE,$7F,$8D,$6E
	.dc.b	$7F,$E2,$7C,$FF,$DF,$73,$F9,$1F,$88,$CE,$7D,$65,$F2,$DF,$99,$3F
	.dc.b	$35,$F0,$D6,$FF,$30,$DC,$3A,$36,$F9,$78,$1F,$A1,$D3,$F4,$4D,$BF
	.dc.b	$D0,$79,$58,$69,$FA,$38,$FD,$43,$FA,$98,$AE,$B0,$FB,$18,$FF,$AC
	.dc.b	$E3,$FD,$09,$F5,$8F,$93,$FE,$A1,$39,$AF,$F2,$34,$E6,$B7,$E4,$6B
	.dc.b	$DF,$FE,$F3,$3F,$C1,$6B,$FD,$A5,$7C,$74,$FA,$13,$FB,$89,$E7,$EB
	.dc.b	$70,$F2,$78,$BC,$BE,$4A,$67,$E4,$F9,$3B,$7F,$E5,$7F,$8F,$76,$FF
	.dc.b	$82,$1B,$5F,$D3,$FE,$25,$E4,$66,$BF,$B0,$C3,$E4,$3F,$A4,$36,$BF
	.dc.b	$DB,$AD,$C3,$F1,$ED,$7F,$C6,$EF,$EC,$63,$B4,$F4,$F1,$FA,$BE,$3E
	.dc.b	$A9,$3F,$EA,$FC,$9D,$58,$C7,$D4,$DE,$1F,$CA,$4E,$AB,$85,$FF,$80
	.dc.b	$7D,$4F,$E4,$38,$7C,$1E,$8C,$7A,$1F,$05,$F8,$5D,$38,$C3,$C0,$DA
	.dc.b	$FF,$3F,$7D,$8F,$D1,$CF,$F4,$4D,$AF,$F3,$1F,$E8,$6F,$73,$FC,$67
	.dc.b	$39,$BF,$EE,$12,$F6,$BF,$B8,$36,$FF,$6B,$28,$C3,$FB,$86,$BF,$E6
	.dc.b	$05,$A7,$F8,$51,$DE,$3F,$E4,$B9,$E5,$39,$BF,$F0,$CF,$9F,$FE,$33
	.dc.b	$C3,$FA,$3E,$2F,$95,$CF,$C4,$FA,$1F,$4F,$CA,$E2,$BF,$DC,$15,$FE
	.dc.b	$D7,$7F,$C2,$97,$CF,$27,$99,$E4,$F1,$FE,$0F,$3F,$D1,$6D,$F2,$F7
	.dc.b	$70,$D3,$FD,$E9,$5C,$5B,$E0,$BE,$33,$FC,$33,$3F,$C1,$1B,$19,$9F
	.dc.b	$E3,$B3,$97,$FD,$35,$E1,$7B,$5E,$75,$FA,$ED,$BF,$6A,$FE,$DB,$E1
	.dc.b	$FF,$63,$BD,$AF,$DD,$B7,$5F,$DF,$CF,$ED,$BE,$57,$1C,$9F,$E7,$9C
	.dc.b	$FF,$8B,$FC,$47,$90,$98,$F9,$35,$FE,$ED,$8B,$CD,$6D,$FA,$EA,$F7
	.dc.b	$DF,$F0,$2F,$0B,$F8,$4B,$FC,$13,$C0,$FF,$0F,$D7,$EA,$F0,$FE,$FE
	.dc.b	$D7,$A3,$70,$EB,$C7,$C7,$1F,$E2,$AD,$AF,$E2,$7F,$0F,$FC,$43,$8F
	.dc.b	$F8,$8B,$6B,$F8,$DB,$87,$F8,$EF,$AB,$6B,$F8,$DF,$A7,$E4,$24,$FD
	.dc.b	$07,$CF,$E1,$7F,$1C,$7F,$8E,$B6,$3F,$CB,$93,$FE,$53,$F9,$EF,$F2
	.dc.b	$A7,$CF,$2F,$F8,$3D,$61,$FE,$B3,$36,$BF,$DA,$3E,$37,$FA,$A6,$F8
	.dc.b	$47,$FB,$12,$7F,$CB,$DF,$03,$88,$FF,$30,$57,$FA,$BE,$31,$1E,$10
	.dc.b	$E2,$35,$7E,$09,$5E,$70,$F1,$47,$FB,$8B,$0A,$FF,$BF,$22,$A8,$CD
	.dc.b	$F9,$EF,$A5,$F3,$AE,$B1,$DF,$0C,$1E,$68,$E4,$79,$A3,$E2,$B4,$F8
	.dc.b	$07,$23,$E2,$E1,$E2,$C3,$C2,$D8,$D1,$F1,$1C,$5F,$8C,$F0,$4F,$09
	.dc.b	$D9,$F1,$E3,$55,$E7,$3C,$77,$C2,$7C,$87,$52,$9C,$63,$D3,$3B,$39
	.dc.b	$C7,$AA,$7E,$0C,$F5,$8F,$59,$C8,$F8,$71,$A7,$84,$78,$8F,$B1,$B7
	.dc.b	$DC,$8F,$8A,$D7,$EC,$1F,$B3,$9D,$9F,$EA,$55,$F2,$7F,$E5,$FE,$6F
	.dc.b	$F3,$6C,$E6,$31,$3C,$4F,$F5,$63,$8A,$EC,$D9,$E1,$EF,$AF,$F2,$B8
	.dc.b	$73,$F8,$9F,$05,$7E,$5E,$D1,$89,$E3,$5C,$31,$EF,$87,$37,$E7,$DF
	.dc.b	$07,$47,$AA,$F9,$3C,$2B,$33,$EA,$CB,$C1,$5F,$05,$B5,$F9,$E9,$D6
	.dc.b	$1E,$28,$EF,$5D,$1E,$C7,$4C,$3C,$98,$FA,$DF,$36,$1D,$27,$F2,$38
	.dc.b	$79,$DB,$3A,$C3,$B5,$6C,$8E,$09,$3F,$43,$0F,$44,$7C,$B8,$F4,$57
	.dc.b	$37,$E0,$1E,$83,$E0,$CF,$A6,$FE,$DE,$1E,$7C,$7E,$04,$F4,$AE,$C7
	.dc.b	$48,$8F,$AA,$34,$FC,$23,$D7,$1A,$9F,$80,$4C,$3E,$53,$DE,$79,$1E
	.dc.b	$03,$AE,$3F,$C7,$32,$FE,$8F,$F1,$1F,$95,$C3,$FC,$87,$F2,$F7,$6F
	.dc.b	$E8,$6D,$C3,$FA,$1D,$5B,$FE,$9F,$3F,$E0,$E6,$CF,$F6,$68,$C7,$29
	.dc.b	$B1,$C6,$3F,$48,$FF,$96,$8C,$7C,$BE,$5D,$70,$C7,$FC,$C6,$78,$7C
	.dc.b	$87,$CB,$DF,$1A,$48,$FD,$26,$9F,$ED,$AA,$DC,$67,$5F,$F4,$84,$2B
	.dc.b	$D4,$D9,$FF,$4E,$E9,$FE,$31,$9C,$9A,$F5,$3E,$27,$7F,$FA,$33,$8E
	.dc.b	$F7,$FE,$47,$2E,$B5,$89,$FF,$19,$D6,$F1,$FE,$F5,$18,$BF,$CC,$47
	.dc.b	$F0,$79,$FF,$4B,$F9,$3D,$09,$1E,$9D,$3D,$9F,$23,$88,$F7,$9F,$E0
	.dc.b	$3D,$5F,$2F,$F1,$CD,$69,$F0,$3E,$5B,$CF,$F1,$92,$FA,$1C,$63,$0F
	.dc.b	$D6,$CF,$FD,$53,$3E,$72,$47,$C6,$FA,$F7,$C7,$7C,$4F,$C0,$1E,$0D
	.dc.b	$3F,$11,$3A,$AF,$F1,$AC,$E9,$F0,$F2,$25,$87,$03,$94,$EB,$F7,$E7
	.dc.b	$C7,$7E,$56,$1F,$B6,$73,$7F,$DB,$4D,$DB,$FD,$9D,$5C,$27,$DE,$70
	.dc.b	$CF,$F3,$90,$F8,$FA,$75,$BE,$5C,$7F,$CE,$FF,$7E,$77,$F8,$5F,$CF
	.dc.b	$63,$D1,$AF,$FB,$AD,$DB,$9E,$FF,$F3,$AA,$8D,$BF,$A4,$6B,$DE,$3F
	.dc.b	$5D,$8B,$D3,$F3,$92,$7C,$1D,$FE,$AF,$89,$FC,$AF,$D9,$BB,$A9,$9A
	.dc.b	$FF,$32,$C6,$5F,$99,$EA,$FC,$23,$6F,$CF,$F8,$6F,$8B,$F0,$D2,$30
	.dc.b	$7D,$37,$0F,$E4,$B9,$DA,$27,$E6,$D7,$D7,$F9,$AF,$A5,$7F,$98,$0F
	.dc.b	$D1,$7E,$AB,$5D,$3C,$39,$7F,$D2,$8E,$63,$F8,$5B,$5F,$EE,$0D,$7F
	.dc.b	$82,$37,$05,$7F,$BB,$67,$F2,$2D,$7D,$4D,$AE,$FF,$25,$FC,$47,$E7
	.dc.b	$25,$7E,$63,$C0,$47,$F1,$5F,$67,$F8,$B7,$B1,$FC,$5F,$E6,$E5,$EE
	.dc.b	$3F,$8D,$7C,$DF,$F0,$AF,$DB,$3D,$D9,$FF,$A0,$5B,$D5,$FC,$66,$75
	.dc.b	$D9,$FC,$87,$D9,$98,$FF,$A7,$AF,$6F,$F0,$1F,$4F,$F9,$37,$97,$23
	.dc.b	$CD,$15,$FA,$DB,$FF,$B2,$5F,$57,$BF,$9D,$3F,$92,$BF,$CB,$E6,$4E
	.dc.b	$8F,$81,$E5,$FF,$33,$F1,$E4,$60,$F8,$8E,$1E,$EF,$F3,$98,$FE,$75
	.dc.b	$CB,$27,$E5,$3F,$CF,$7C,$5C,$47,$9B,$D2,$FE,$80,$63,$27,$0F,$F3
	.dc.b	$FB,$7D,$43,$E1,$3F,$1D,$D3,$31,$FE,$1B,$39,$7F,$1B,$1F,$B2,$EA
	.dc.b	$FC,$1C,$7F,$BA,$1A,$FF,$56,$3F,$2B,$1F,$D7,$E7,$E7,$DF,$83,$93
	.dc.b	$FB,$24,$60,$FC,$B6,$3F,$D4,$FC,$B7,$23,$C1,$FC,$01,$32,$F2,$F6
	.dc.b	$7F,$1B,$FA,$2D,$47,$FB,$E9,$AF,$8D,$AF,$F6,$A6,$BD,$E9,$F0,$DE
	.dc.b	$4F,$98,$93,$F8,$78,$FC,$87,$C8,$C8,$FC,$53,$EA,$F1,$FF,$96,$7E
	.dc.b	$78,$F8,$FF,$E2,$ED,$AF,$88,$1C,$3F,$1D,$AF,$F5,$BC,$7E,$5F,$2F
	.dc.b	$D8,$8F,$F6,$C7,$03,$E6,$3E,$49,$FA,$3E,$4F,$97,$F1,$B2,$BE,$CB
	.dc.b	$F3,$BE,$63,$CB,$1E,$42,$1E,$4C,$BF,$96,$F3,$7E,$13,$8F,$EA,$FC
	.dc.b	$B4,$AF,$C1,$C7,$D8,$F2,$BF,$BD,$FE,$FE,$F9,$4A,$66,$BF,$A3,$FE
	.dc.b	$09,$FE,$0A,$7F,$08,$D7,$FC,$36,$D7,$FD,$00,$D7,$DF,$CB,$DB,$8E
	.dc.b	$47,$E5,$23,$FC,$5D,$F3,$CF,$B1,$6F,$86,$D7,$F7,$69,$D4,$FE,$43
	.dc.b	$E5,$3F,$B1,$62,$DF,$FA,$6B,$6F,$EB,$B5,$F4,$C7,$F8,$53,$47,$CB
	.dc.b	$7C,$22,$4F,$EA,$3A,$7F,$3B,$C9,$23,$F5,$6F,$E8,$3E,$03,$FB,$97
	.dc.b	$8D,$FA,$D9,$FD,$43,$5F,$F1,$FB,$5F,$F5,$23,$56,$AF,$2F,$FC,$03
	.dc.b	$1E,$F6,$BF,$E1,$E8,$D7,$8D,$FB,$0F,$47,$E4,$FE,$FC,$F8,$4A,$4B
	.dc.b	$FE,$03,$F2,$DE,$6E,$9F,$EE,$39,$FE,$A2,$7F,$C9,$0D,$7F,$C3,$71
	.dc.b	$FB,$FF,$96,$FD,$E9,$23,$E9,$7E,$38,$F9,$1F,$B3,$F3,$9F,$E4,$99
	.dc.b	$FE,$B9,$83,$B7,$E2,$9A,$F8,$8B,$3F,$AA,$9D,$36,$7C,$A7,$D5,$F9
	.dc.b	$4F,$CA,$92,$3F,$8D,$F9,$95,$FE,$E1,$F1,$DF,$D5,$E3,$F9,$23,$61
	.dc.b	$E6,$BA,$97,$FE,$7B,$AF,$DC,$F9,$1F,$B5,$EC,$FF,$6B,$F8,$59,$35
	.dc.b	$E1,$DC,$BF,$C6,$F6,$F9,$AE,$46,$63,$FC,$05,$F8,$F3,$DC,$FF,$44
	.dc.b	$36,$3F,$C7,$ED,$77,$F1,$7A,$9D,$FC,$5C,$9D,$C5,$7F,$D1,$39,$AD
	.dc.b	$5F,$E9,$36,$CF,$F9,$E9,$AF,$0C,$5F,$31,$F0,$49,$1E,$CF,$9B,$AF
	.dc.b	$CC,$F1,$7E,$31,$8D,$FF,$8D,$43,$E9,$FF,$C1,$79,$FF,$24,$6C,$7F
	.dc.b	$49,$6B,$FE,$A0,$6B,$FF,$00,$7C,$3F,$D0,$BF,$D3,$F2,$FD,$17,$87
	.dc.b	$FC,$B9,$B1,$99,$DB,$CF,$C2,$FD,$03,$DD,$AE,$7C,$1F,$24,$62,$3F
	.dc.b	$0B,$F3,$3F,$CC,$F3,$FC,$06,$0F,$9A,$C7,$56,$8F,$BE,$E2,$FF,$87
	.dc.b	$4A,$F1,$AA,$31,$7F,$86,$E8,$CC,$62,$FF,$2F,$DF,$F2,$9F,$CB,$B5
	.dc.b	$EA,$DB,$E8,$D9,$7C,$CF,$48,$EA,$DC,$3F,$2A,$E1,$FC,$BF,$CD,$5C
	.dc.b	$3F,$4F,$F2,$FE,$4F,$90,$F2,$B6,$BF,$3A,$E1,$FC,$FF,$CD,$5A,$FF
	.dc.b	$57,$F8,$1F,$AB,$AF,$6F,$CD,$7C,$58,$C5,$FE,$9A,$31,$E0,$AA,$9F
	.dc.b	$13,$E4,$6D,$7F,$A5,$3E,$93,$FD,$52,$FF,$6E,$6E,$3F,$E1,$36,$BF
	.dc.b	$EE,$A6,$BF,$C0,$3E,$47,$EB,$28,$E3,$CB,$F7,$A7,$31,$8A,$FE,$0F
	.dc.b	$CF,$F5,$BE,$47,$EB,$B6,$3C,$9A,$7D,$5C,$7A,$36,$CD,$3E,$C7,$D6
	.dc.b	$FF,$BD,$AB,$FD,$7A,$DA,$FE,$2E,$7E,$21,$C6,$5F,$C4,$5F,$F4,$C4
	.dc.b	$F0,$F2,$75,$F2,$5B,$7D,$BF,$48,$F8,$4F,$23,$6B,$FC,$2A,$DC,$7F
	.dc.b	$C7,$4D,$47,$F9,$27,$0F,$F7,$3C,$F9,$BC,$7F,$27,$CC,$F9,$5F,$B9
	.dc.b	$FC,$01,$E6,$9E,$1E,$46,$3E,$6F,$D4,$F9,$3D,$CF,$19,$27,$C6,$FE
	.dc.b	$77,$A1,$7E,$EC,$FF,$16,$DE,$0F,$85,$E5,$F8,$38,$7E,$37,$C9,$FC
	.dc.b	$7F,$94,$B9,$FC,$C7,$CB,$A7,$5C,$7F,$91,$CF,$F9,$67,$3E,$F3,$93
	.dc.b	$EA,$FC,$9F,$D1,$F9,$4F,$A4,$DA,$FA,$FF,$27,$F5,$4F,$AD,$E5,$58
	.dc.b	$F9,$3B,$7F,$8E,$F1,$F5,$DF,$84,$93,$E1,$C9,$1F,$84,$F9,$2F,$C0
	.dc.b	$1F,$81,$F9,$4B,$5D,$0F,$F3,$45,$27,$E8,$9F,$87,$F9,$4B,$3B,$E2
	.dc.b	$FE,$BB,$49,$F8,$44,$9F,$13,$F1,$7D,$8F,$B6,$7E,$47,$E5,$2C,$F3
	.dc.b	$B7,$D2,$7F,$5D,$8F,$F2,$FB,$63,$FD,$08,$D7,$F1,$67,$E3,$1C,$3E
	.dc.b	$AF,$81,$FA,$84,$9F,$88,$FD,$7F,$37,$F1,$A7,$E8,$F0,$B3,$FA,$8F
	.dc.b	$95,$FD,$35,$7E,$9D,$B1,$F0,$77,$F8,$1F,$27,$F8,$DD,$0D,$24,$FF
	.dc.b	$A6,$F1,$77,$7F,$91,$E3,$95,$70,$7F,$6F,$3F,$C6,$E3,$19,$FF,$BF
	.dc.b	$7C,$8B,$C7,$FE,$07,$7F,$D6,$17,$F4,$EF,$F8,$F6,$3F,$D1,$27,$37
	.dc.b	$C6,$30,$67,$E6,$AD,$F1,$AF,$DC,$61,$D7,$C2,$E5,$F2,$5C,$DB,$5B
	.dc.b	$8E,$DF,$27,$8C,$BC,$DF,$23,$E1,$3D,$5F,$1F,$6F,$DE,$FC,$72,$3F
	.dc.b	$A7,$D9,$1F,$D4,$73,$FF,$39,$CF,$03,$CD,$E1,$74,$7E,$4F,$C7,$FB
	.dc.b	$FE,$47,$E4,$36,$3D,$97,$86,$9C,$0F,$81,$F1,$1F,$1C,$F9,$1F,$0B
	.dc.b	$63,$E2,$56,$1A,$7C,$78,$FE,$7F,$E7,$7F,$1C,$8F,$95,$F1,$F8,$DB
	.dc.b	$E9,$3F,$D1,$3E,$87,$F9,$94,$7E,$60,$38,$7F,$36,$D7,$F9,$3B,$5D
	.dc.b	$F9,$DC,$FC,$BE,$8D,$BF,$B5,$F1,$FF,$81,$3F,$05,$F1,$F6,$3E,$F3
	.dc.b	$87,$F0,$1F,$1F,$70,$FE,$DB,$C6,$F8,$DF,$C5,$7C,$17,$E4,$DB,$7F
	.dc.b	$87,$F8,$EC,$34,$FC,$DC,$7F,$62,$FD,$93,$FC,$39,$1F,$8D,$8C,$7C
	.dc.b	$27,$E5,$5F,$EC,$FF,$A6,$FE,$5B,$2F,$88,$D7,$F7,$AD,$7F,$A4,$B5
	.dc.b	$5F,$F9,$B2,$FF,$85,$4F,$E8,$3B,$BF,$C8,$F3,$9A,$47,$F9,$2A,$75
	.dc.b	$5F,$D5,$3C,$37,$CE,$7F,$B4,$E3,$F4,$87,$F8,$35,$C5,$FE,$CE,$3F
	.dc.b	$C6,$3F,$58,$FA,$5D,$7C,$86,$1B,$3F,$C1,$ED,$F1,$BC,$2B,$A3,$67
	.dc.b	$F7,$57,$FC,$7F,$3F,$E7,$D7,$C2,$73,$F2,$1C,$1E,$9F,$1F,$FD,$76
	.dc.b	$7F,$CD,$B1,$FD,$45,$B9,$FE,$3F,$4F,$6D,$7F,$8F,$71,$7C,$8F,$B6
	.dc.b	$3D,$6D,$C3,$D1,$F1,$BE,$ED,$3E,$13,$63,$24,$C7,$E6,$BF,$DF,$BC
	.dc.b	$B6,$3F,$ED,$E8,$D6,$BE,$CF,$89,$F4,$12,$3D,$0F,$97,$3F,$A0,$FC
	.dc.b	$E6,$C7,$C6,$E4,$F5,$11,$FE,$5D,$7F,$90,$C7,$FD,$2E,$FE,$07,$0F
	.dc.b	$F5,$73,$96,$FC,$6F,$DA,$BF,$B7,$8F,$DC,$CE,$E3,$F5,$1B,$8F,$B8
	.dc.b	$FF,$80,$18,$AB,$F0,$7F,$19,$AC,$BB,$FC,$67,$CF,$26,$3E,$67,$E9
	.dc.b	$7C,$67,$E2,$9F,$C5,$B6,$3F,$B2,$63,$FE,$28,$FA,$4F,$F1,$4F,$D2
	.dc.b	$7F,$96,$E3,$FB,$0F,$8A,$AC,$F9,$9F,$C3,$65,$78,$77,$35,$FE,$D3
	.dc.b	$1F,$E3,$4F,$53,$A4,$7F,$8D,$AF,$A2,$8C,$BE,$33,$17,$FC,$6E,$77
	.dc.b	$3A,$9F,$08,$C3,$8C,$F4,$75,$7E,$39,$4E,$6F,$D5,$39,$8C,$4F,$F2
	.dc.b	$98,$C6,$1C,$63,$BE,$1D,$A3,$D9,$7C,$A7,$CE,$7C,$22,$9F,$08,$F3
	.dc.b	$3E,$B8,$E9,$5D,$A7,$27,$84,$A7,$C7,$8D,$B8,$3D,$6E,$30,$F2,$46
	.dc.b	$4F,$8A,$E0,$F8,$6F,$90,$E2,$53,$DB,$5E,$57,$15,$E5,$9E,$27,$39
	.dc.b	$F8,$17,$E0,$3A,$95,$C9,$CD,$CB,$4F,$91,$C7,$F8,$1B,$F8,$0D,$7E
	.dc.b	$43,$5F,$C2,$39,$F8,$F7,$BF,$FC,$C2,$53,$98,$C3,$76,$FE,$11,$87
	.dc.b	$F8,$F2,$F8,$C6,$2F,$FD,$13,$18,$7E,$CA,$F7,$7F,$C8,$75,$F9,$A8
	.dc.b	$C4,$FE,$08,$E3,$87,$8E,$74,$DC,$7C,$17,$C9,$B5,$C4,$FF,$77,$4E
	.dc.b	$DE,$78,$7E,$AF,$0E,$F1,$F8,$46,$E5,$DA,$75,$5C,$6B,$FD,$F1,$87
	.dc.b	$77,$DA,$D7,$BC,$F8,$6F,$E0,$5F,$B6,$BF,$CC,$2D,$AE,$B5,$1F,$14
	.dc.b	$EB,$0F,$5C,$79,$70,$FA,$97,$E5,$8D,$9E,$BA,$E0,$FB,$8E,$17,$E3
	.dc.b	$8F,$36,$1F,$03,$1E,$47,$3B,$F2,$5F,$43,$89,$1C,$93,$1F,$C0,$63
	.dc.b	$E4,$6E,$3D,$EA,$1F,$80,$9F,$95,$87,$C4,$AF,$6D,$F3,$AF,$43,$FC
	.dc.b	$9F,$2F,$EA,$7F,$1F,$FE,$88,$C7,$FA,$1D,$DD,$7F,$AC,$71,$FF,$22
	.dc.b	$9F,$E8,$F5,$9C,$7F,$9F,$DF,$F2,$D6,$9F,$D4,$1C,$61,$FD,$81,$B6
	.dc.b	$F8,$DE,$53,$FD,$91,$B1,$9A,$FF,$00,$FC,$5E,$27,$3F,$1D,$27,$FC
	.dc.b	$0B,$1C,$3E,$3B,$8F,$1C,$9F,$F0,$4C,$57,$FB,$0C,$78,$88,$FF,$66
	.dc.b	$3D,$9D,$DF,$E2,$0E,$5B,$DE,$7C,$B3,$87,$E0,$4E,$E3,$FC,$1D,$18
	.dc.b	$9E,$11,$FE,$11,$CB,$99,$1F,$F4,$B7,$D5,$BC,$EC,$47,$F6,$BD,$92
	.dc.b	$F3,$BE,$3B,$67,$FC,$55,$78,$F8,$AF,$3E,$D2,$7A,$C5,$7F,$C4,$19
	.dc.b	$7C,$D6,$CF,$9C,$FA,$F6,$F8,$DF,$83,$30,$F3,$B5,$F3,$43,$E3,$7C
	.dc.b	$A6,$79,$7F,$B3,$E9,$EA,$F8,$AF,$9B,$FC,$02,$D2,$9E,$C8,$B8,$FC
	.dc.b	$4F,$8E,$F7,$98,$7F,$C1,$C3,$EF,$5E,$EF,$E4,$E1,$F4,$FF,$2D,$FF
	.dc.b	$75,$35,$E3,$3C,$1C,$F1,$FE,$65,$E2,$D9,$E1,$1F,$E9,$56,$BF,$EE
	.dc.b	$E9,$7A,$FF,$53,$E3,$E4,$FF,$BE,$67,$2F,$DE,$F9,$0F,$83,$9C,$8C
	.dc.b	$1F,$EE,$9D,$51,$F7,$BE,$42,$4E,$63,$F0,$3F,$21,$23,$F1,$2F,$E1
	.dc.b	$3E,$43,$F0,$BE,$2C,$9F,$C6,$BB,$7E,$4E,$09,$F9,$0A,$A9,$88,$B3
	.dc.b	$FB,$D9,$D3,$67,$E4,$46,$23,$FC,$5C,$FF,$5E,$7F,$3B,$E3,$FE,$60
	.dc.b	$EF,$7F,$9A,$F9,$0F,$CD,$FB,$72,$BF,$47,$1F,$9E,$F8,$F9,$1F,$A6
	.dc.b	$74,$FE,$43,$E3,$BF,$17,$72,$3F,$57,$15,$FC,$AF,$C7,$7E,$A0,$C7
	.dc.b	$4F,$E5,$3A,$FF,$BC,$3F,$E3,$CF,$09,$AB,$4F,$CB,$9F,$B0,$F3,$2E
	.dc.b	$3F,$E4,$BA,$DC,$FF,$02,$AE,$13,$96,$BF,$ED,$DE,$CF,$DD,$9F,$BA
	.dc.b	$23,$F8,$19,$FE,$47,$F5,$4F,$8D,$A4,$91,$FC,$16,$1F,$47,$C3,$E3
	.dc.b	$E4,$EB,$4F,$D8,$D2,$79,$4F,$1F,$8E,$EE,$47,$87,$9F,$FE,$97,$A7
	.dc.b	$C7,$E6,$7C,$1F,$1F,$E6,$31,$8F,$6F,$8F,$FE,$53,$F4,$7E,$9A,$E6
	.dc.b	$38,$51,$98,$F5,$4F,$0D,$BF,$95,$CE,$67,$DE,$71,$70,$FF,$B5,$5A
	.dc.b	$FF,$8A,$B2,$7C,$87,$CA,$F1,$64,$F8,$B8,$95,$A8,$F2,$E5,$E4,$D7
	.dc.b	$D2,$48,$E1,$2F,$90,$FE,$1D,$F8,$57,$F8,$E2,$7F,$B0,$35,$FF,$47
	.dc.b	$35,$F4,$87,$0F,$B3,$E9,$1F,$25,$FC,$45,$B1,$FD,$AD,$AF,$FA,$65
	.dc.b	$AF,$C4,$C7,$E5,$F1,$FD,$02,$47,$A2,$BF,$92,$7B,$9F,$CF,$E3,$C6
	.dc.b	$D7,$E6,$B5,$FC,$6B,$6B,$EB,$EB,$F7,$09,$3F,$13,$E4,$7F,$2F,$F2
	.dc.b	$32,$3E,$3B,$F9,$AF,$92,$FC,$1E,$3F,$70,$93,$FA,$9E,$2F,$5B,$F8
	.dc.b	$FF,$90,$FD,$09,$1F,$A3,$E6,$7E,$93,$E4,$3F,$4B,$F2,$32,$7E,$AC
	.dc.b	$7E,$42,$BF,$12,$48,$EC,$FF,$84,$B8,$DF,$21,$FA,$EF,$91,$95,$F7
	.dc.b	$A3,$F3,$24,$7F,$0B,$97,$07,$F0,$7D,$BF,$E2,$1F,$99,$7D,$0F,$E2
	.dc.b	$E7,$EF,$C6,$3A,$FF,$C4,$8D,$8F,$F6,$93,$5F,$C2,$F2,$3D,$B7,$C8
	.dc.b	$FF,$89,$36,$B5,$F9,$96,$BF,$EE,$FA,$E1,$3B,$F9,$09,$3B,$79,$7C
	.dc.b	$8F,$9F,$37,$2F,$91,$B5,$E4,$6B,$EA,$6B,$FB,$2E,$F6,$4F,$90,$F3
	.dc.b	$7C,$8C,$AF,$34,$7A,$7E,$46,$47,$A5,$F6,$7C,$E7,$C9,$FA,$07,$C3
	.dc.b	$FC,$66,$D7,$C6,$6B,$F3,$1A,$F8,$F1,$7C,$9F,$C7,$F9,$39,$3F,$1E
	.dc.b	$3E,$57,$C9,$FD,$7F,$9F,$FF,$01,$79,$9F,$EF,$AE,$74,$63,$A7,$F9
	.dc.b	$6F,$0F,$BA,$F8,$7C,$17,$DA,$AF,$F0,$0E,$4C,$38,$CE,$4F,$F2,$1F
	.dc.b	$4F,$92,$7F,$86,$B5,$FB,$0D,$CB,$FC,$4B,$AB,$E4,$FF,$CA,$33,$9D
	.dc.b	$3C,$72,$F9,$4F,$C2,$1A,$DB,$F0,$DF,$23,$32,$FC,$33,$F8,$9F,$27
	.dc.b	$11,$FE,$4B,$BF,$F2,$9B,$7E,$37,$3A,$78,$B3,$FC,$81,$27,$F2,$1F
	.dc.b	$23,$5E,$5F,$21,$F9,$53,$15,$EA,$7F,$CF,$F8,$1A,$49,$FE,$46,$7F
	.dc.b	$B8,$21,$F2,$1F,$9F,$24,$7E,$7F,$E3,$AB,$E9,$E8,$FD,$21,$8C,$1C
	.dc.b	$1F,$D1,$BE,$79,$E5,$7F,$2E,$B4,$63,$A3,$FD,$2C,$D8,$F1,$B5,$FE
	.dc.b	$6F,$F2,$9F,$E2,$3E,$7F,$F4,$E3,$63,$D2,$D7,$7F,$0B,$F8,$37,$9F
	.dc.b	$99,$EC,$6E,$DE,$08,$C4,$7F,$44,$C1,$18,$8F,$C8,$FD,$27,$FA,$0F
	.dc.b	$4F,$E0,$11,$EE,$3F,$D0,$91,$97,$EE,$19,$D7,$FC,$DF,$1F,$E8,$79
	.dc.b	$FE,$C0,$EE,$33,$F2,$7B,$9E,$FD,$76,$DB,$30,$FF,$3B,$57,$D8,$8C
	.dc.b	$7C,$87,$8E,$BF,$82,$56,$9F,$F0,$F3,$5C,$FC,$97,$5A,$E3,$7E,$48
	.dc.b	$C5,$71,$CB,$D3,$F2,$58,$9C,$B6,$BD,$7F,$25,$C3,$E3,$78,$B6,$3F
	.dc.b	$5F,$D9,$86,$9C,$0F,$87,$F1,$FE,$0F,$91,$F0,$B6,$3F,$6F,$DF,$ED
	.dc.b	$DB,$FC,$5F,$E8,$7F,$35,$8C,$47,$F8,$6F,$BD,$18,$9D,$57,$F3,$EF
	.dc.b	$AE,$CB,$67,$8F,$52,$73,$87,$F4,$86,$BF,$33,$E4,$7E,$A4,$EE,$BE
	.dc.b	$3C,$62,$B8,$F0,$FE,$03,$AB,$E5,$9F,$33,$E4,$6E,$5F,$E3,$2C,$7C
	.dc.b	$1E,$0F,$8B,$0F,$DB,$4F,$9E,$30,$7B,$B9,$BF,$05,$F2,$3F,$5B,$E3
	.dc.b	$FE,$BB,$6B,$F8,$EF,$85,$F8,$9E,$0F,$EB,$4F,$DC,$F9,$1C,$71,$7F
	.dc.b	$BF,$EB,$C5,$E3,$3E,$1F,$E1,$B6,$3F,$B2,$B7,$93,$FB,$5B,$5F,$F1
	.dc.b	$05,$7F,$48,$9F,$CE,$FC,$5F,$6F,$1B,$F1,$6F,$82,$A4,$FE,$8B,$E4
	.dc.b	$6C,$39,$7F,$A1,$63,$FE,$32,$AF,$0F,$C0,$79,$08,$FF,$9F,$FE,$AF
	.dc.b	$F5,$97,$FB,$2F,$92,$F8,$11,$8C,$3F,$63,$1F,$9C,$F9,$0F,$CE,$FC
	.dc.b	$95,$C3,$F6,$6F,$0F,$0B,$F1,$18,$7F,$BA,$67,$B4,$60,$FA,$5F,$43
	.dc.b	$FE,$7E,$8F,$DD,$F3,$FE,$AE,$31,$9F,$F0,$48,$C7,$C8,$67,$E3,$ED
	.dc.b	$FF,$98,$6F,$FD,$18,$FF,$98,$31,$77,$F8,$89,$1E,$F7,$C3,$F2,$1D
	.dc.b	$FE,$43,$C0,$D9,$E5,$D7,$E3,$F8,$DF,$86,$D8,$FF,$40,$F1,$3F,$12
	.dc.b	$FF,$49,$47,$F4,$28,$C6,$BF,$AC,$6C,$66,$77,$BF,$F4,$8F,$9F,$7C
	.dc.b	$2F,$EC,$A1,$18,$AF,$E9,$B1,$89,$D4,$E6,$3F,$C5,$FB,$7F,$AF,$6B
	.dc.b	$E2,$76,$7D,$AE,$AF,$98,$FD,$CE,$AF,$AA,$49,$FC,$13,$F6,$39,$BF
	.dc.b	$02,$7D,$3F,$8F,$B3,$F3,$FE,$23,$EB,$1F,$CB,$72,$FE,$BF,$F0,$98
	.dc.b	$9F,$F5,$2C,$7F,$62,$F8,$4F,$67,$AF,$FE,$2D,$9F,$5B,$67,$FB,$4B
	.dc.b	$5F,$F4,$6A,$F8,$3E,$CE,$F8,$BF,$9B,$39,$AE,$EE,$5F,$F0,$FF,$4F
	.dc.b	$E8,$6F,$F4,$91,$8F,$92,$FC,$77,$81,$F8,$F7,$F2,$1F,$21,$89,$FC
	.dc.b	$63,$6B,$FC,$A9,$F5,$7F,$63,$2F,$EB,$1D,$EC,$BF,$CA,$38,$7F,$AC
	.dc.b	$2B,$4F,$8B,$E2,$7F,$85,$46,$30,$FF,$64,$36,$F9,$7C,$8E,$BA,$FF
	.dc.b	$45,$F2,$33,$1E,$9D,$D7,$1F,$1B,$FC,$D7,$13,$49,$3F,$D7,$73,$FF
	.dc.b	$2E,$BC,$8F,$D4,$7C,$0F,$92,$7C,$BC,$BF,$B1,$F8,$1B,$5F,$AD,$7F
	.dc.b	$69,$EF,$E3,$2F,$96,$7F,$6C,$C7,$FB,$FD,$7C,$1B,$FD,$D7,$C8,$48
	.dc.b	$F8,$9C,$56,$FE,$5B,$F1,$0F,$E0,$1F,$21,$67,$F7,$87,$F0,$33,$F1
	.dc.b	$DF,$3F,$FD,$AD,$FB,$A6,$3D,$EF,$D4,$41,$A4,$8F,$D4,$C7,$F1,$3B
	.dc.b	$F4,$87,$84,$7E,$61,$D7,$2F,$F8,$0D,$CC,$FE,$0B,$87,$FC,$01,$F4
	.dc.b	$3F,$E9,$CB,$FC,$04,$63,$E4,$7E,$FB,$23,$FD,$ED,$2F,$91,$C5,$FE
	.dc.b	$33,$83,$9F,$8D,$F8,$66,$DF,$0E,$FC,$34,$FC,$74,$78,$BB,$A4,$76
	.dc.b	$EF,$B8,$7B,$BE,$7F,$F9,$2B,$FE,$56,$6B,$FE,$FA,$6A,$23,$87,$FC
	.dc.b	$53,$9B,$E4,$7F,$80,$4F,$4E,$0F,$D1,$CE,$70,$FE,$03,$39,$7F,$C1
	.dc.b	$FF,$53,$F0,$5F,$87,$F2,$3F,$DB,$63,$18,$FE,$33,$BD,$82,$BE,$7F
	.dc.b	$C6,$EF,$86,$E3,$E7,$CF,$FC,$95,$ED,$33,$E7,$18,$7F,$82,$67,$EA
	.dc.b	$F9,$1F,$06,$FF,$69,$FB,$7F,$21,$8C,$7E,$BB,$6B,$FD,$A9,$7E,$D7
	.dc.b	$FD,$BA,$D7,$E4,$0C,$7C,$5C,$1F,$E2,$99,$D4,$FF,$AC,$B6,$FF,$50
	.dc.b	$61,$E9,$3E,$2D,$7E,$33,$E4,$3E,$23,$F1,$72,$FC,$7F,$C3,$DA,$FC
	.dc.b	$5B,$70,$F2,$CF,$E5,$3D,$AC,$6D,$F0,$71,$F1,$13,$0F,$F4,$BE,$9F
	.dc.b	$1E,$3F,$C4,$99,$FF,$8C,$27,$EB,$C7,$E7,$BE,$17,$EB,$3F,$A0,$F9
	.dc.b	$0D,$92,$7F,$CD,$54,$8E,$7E,$F6,$E7,$F4,$2D,$D3,$EF,$57,$F8,$A5
	.dc.b	$BF,$05,$FA,$CF,$81,$FD,$6B,$FA,$EF,$80,$C6,$1F,$AA,$6D,$72,$F4
	.dc.b	$3D,$24,$9F,$1B,$EC,$8F,$DB,$BF,$B8,$6C,$7F,$56,$7F,$41,$F4,$1F
	.dc.b	$A1,$FA,$0F,$54,$7D,$CA,$FF,$4B,$46,$23,$EF,$CE,$F4,$FE,$AF,$76
	.dc.b	$7F,$06,$27,$F5,$9F,$03,$C6,$3F,$D8,$B3,$8A,$FF,$64,$D5,$9F,$04
	.dc.b	$D7,$B8,$BE,$58,$79,$2F,$8C,$E5,$F1,$B4,$F1,$DF,$95,$FD,$F4,$F1
	.dc.b	$72,$6C,$CD,$ED,$F8,$A3,$0F,$0C,$79,$63,$31,$87,$57,$E4,$C3,$07
	.dc.b	$C1,$8F,$88,$77,$3E,$19,$4C,$BC,$23,$91,$EC,$79,$61,$E5,$C3,$D5
	.dc.b	$5F,$1A,$78,$3E,$CB,$F2,$1C,$1F,$29,$F3,$8A,$71,$BE,$C7,$81,$D4
	.dc.b	$F1,$9E,$73,$F4,$1B,$3D,$E3,$E8,$C6,$DA,$7D,$6E,$CF,$95,$3F,$41
	.dc.b	$AF,$DE,$6B,$F4,$1C,$FC,$93,$E1,$7F,$CF,$D3,$FE,$DC,$BF,$E2,$33
	.dc.b	$9E,$4F,$E7,$57,$C2,$F6,$D8,$CD,$70,$9E,$6F,$E7,$69,$EF,$7F,$1C
	.dc.b	$8C,$37,$1E,$17,$F9,$5A,$F0,$63,$F2,$9E,$2E,$6B,$83,$FD,$36,$37
	.dc.b	$7D,$F0,$F5,$E5,$DA,$BF,$2D,$5C,$9F,$CA,$3D,$A3,$86,$1D,$66,$1E
	.dc.b	$1A,$F5,$E3,$E6,$8E,$38,$FF,$36,$BE,$D1,$D2,$BC,$35,$B9,$C9,$1C
	.dc.b	$12,$7D,$55,$CE,$3E,$8B,$63,$DD,$5D,$1B,$7E,$6C,$7C,$27,$AA,$3D
	.dc.b	$11,$F0,$2B,$94,$31,$F8,$58,$EC,$98,$F8,$27,$81,$F2,$31,$F0,$3C
	.dc.b	$5F,$86,$FA,$0F,$14,$7C,$5C,$79,$C6,$4F,$8E,$73,$C7,$F8,$E6,$4F
	.dc.b	$6D,$F2,$1F,$D8,$27,$FC,$C7,$0F,$91,$C7,$07,$E4,$A3,$18,$7F,$5F
	.dc.b	$18,$73,$C3,$71,$89,$DB,$FC,$AA,$7C,$17,$FE,$A2,$6C,$7F,$2E,$38
	.dc.b	$46,$65,$F2,$3D,$2B,$D1,$F0,$D6,$4D,$9C,$39,$F3,$66,$1E,$7B,$E2
	.dc.b	$DB,$7C,$87,$3F,$90,$93,$98,$AF,$A6,$53,$E5,$24,$7A,$6B,$E6,$7C
	.dc.b	$87,$C2,$EE,$FF,$0D,$77,$DF,$0C,$FA,$67,$0F,$FB,$12,$77,$7F,$DF
	.dc.b	$2B,$6D,$9D,$43,$BB,$A1,$27,$D9,$F2,$15,$FF,$19,$67,$FE,$EA,$7E
	.dc.b	$07,$C4,$E7,$93,$EF,$FC,$7F,$97,$C0,$95,$F1,$62,$BF,$E2,$68,$F0
	.dc.b	$91,$F8,$16,$F9,$0F,$74,$43,$E7,$35,$F1,$F5,$7D,$93,$18,$7E,$43
	.dc.b	$B7,$E4,$7B,$5F,$80,$AF,$C7,$70,$23,$11,$EC,$9D,$F3,$FE,$0F,$E3
	.dc.b	$A4,$F0,$AD,$E1,$F9,$CC,$FF,$2B,$2F,$D4,$1E,$36,$1B,$1F,$85,$9C
	.dc.b	$4F,$FB,$E2,$38,$36,$77,$1F,$99,$6B,$FE,$4A,$E6,$FD,$77,$C7,$FE
	.dc.b	$5C,$91,$F9,$1F,$88,$FC,$D7,$C4,$7C,$0E,$99,$3F,$95,$8A,$FD,$AF
	.dc.b	$07,$F4,$06,$34,$FF,$33,$EA,$E5,$F9,$24,$AF,$CF,$67,$5F,$F3,$A7
	.dc.b	$D0,$3E,$73,$F5,$8D,$8F,$91,$5F,$1B,$E7,$5A,$5E,$A9,$5F,$C0,$44
	.dc.b	$E9,$C2,$8F,$90,$D8,$FD,$F1,$C2,$32,$7F,$03,$B7,$C9,$79,$09,$3D
	.dc.b	$3A,$FF,$0B,$CF,$23,$EF,$3E,$AF,$87,$F5,$FC,$94,$9F,$E0,$18,$23
	.dc.b	$84,$7C,$2F,$92,$FC,$7F,$24,$8D,$C5,$7E,$3F,$C3,$FC,$63,$1C,$6F
	.dc.b	$27,$FB,$AE,$2F,$8F,$FC,$CF,$34,$C3,$C3,$15,$FF,$25,$F7,$7E,$6E
	.dc.b	$7F,$80,$73,$FF,$81,$A7,$FD,$71,$39,$7F,$86,$5B,$E3,$FE,$B9,$23
	.dc.b	$F3,$D7,$FC,$3F,$E9,$BE,$4C,$7E,$A1,$AF,$FB,$FA,$D1,$8C,$FD,$53
	.dc.b	$FC,$EB,$07,$C4,$F3,$E4,$FC,$2F,$C7,$CA,$E9,$1F,$88,$F9,$0F,$C4
	.dc.b	$FC,$84,$9F,$9F,$15,$FC,$7F,$C5,$FE,$38,$C6,$5E,$08,$7C,$DF,$B5
	.dc.b	$FC,$63,$E2,$6E,$5F,$B8,$8D,$47,$FC,$00,$74,$FF,$49,$6B,$FE,$78
	.dc.b	$6B,$FE,$BC,$6B,$EF,$F9,$1F,$D0,$91,$F0,$E7,$F2,$F9,$24,$7C,$B7
	.dc.b	$F4,$FF,$15,$FC,$2B,$4F,$E2,$73,$FE,$86,$6B,$FE,$CA,$6B,$FE,$F3
	.dc.b	$6F,$07,$D2,$E2,$FD,$8F,$C8,$4A,$FB,$11,$FC,$5B,$E7,$5C,$1F,$C4
	.dc.b	$5B,$3F,$AA,$6B,$FD,$A5,$AF,$FA,$63,$E7,$9E,$2F,$F1,$96,$C7,$EC
	.dc.b	$9A,$FF,$71,$6B,$FE,$E2,$C7,$F0,$FA,$FF,$06,$E7,$93,$F8,$D8,$FE
	.dc.b	$13,$F2,$3F,$C2,$BE,$32,$4F,$EC,$23,$F2,$9B,$7F,$5D,$E8,$34,$93
	.dc.b	$F9,$EE,$BF,$E2,$7F,$23,$23,$F7,$31,$FE,$01,$F1,$1F,$13,$7E,$42
	.dc.b	$57,$F7,$D9,$CB,$FA,$1C,$FC,$04,$7F,$DC,$F9,$23,$FC,$25,$2F,$8F
	.dc.b	$ED,$F2,$12,$FA,$47,$F7,$BF,$CA,$D8,$F8,$FE,$36,$57,$F5,$69,$CB
	.dc.b	$FE,$25,$F9,$9F,$EC,$9E,$BF,$F2,$06,$C7,$F8,$7D,$AF,$FB,$39,$AF
	.dc.b	$D1,$96,$3E,$12,$47,$BF,$E4,$BC,$99,$7C,$5F,$8E,$93,$E3,$8F,$F1
	.dc.b	$9E,$1F,$DD,$7D,$6F,$16,$5F,$E4,$1A,$F1,$35,$F1,$35,$F1,$67,$F0
	.dc.b	$48,$FC,$FF,$91,$FA,$9F,$1D,$F5,$7E,$46,$4F,$D3,$8F,$0F,$D4,$34
	.dc.b	$F8,$4D,$9F,$84,$D7,$E1,$35,$FE,$D5,$C3,$FD,$BB,$C7,$FE,$2F,$EB
	.dc.b	$FF,$50,$F6,$BF,$08,$7E,$17,$E4,$BF,$0D,$F2,$5E,$0C,$E6,$3F,$85
	.dc.b	$8A,$F6,$8F,$19,$1F,$F2,$97,$80,$E6,$4F,$F8,$D5,$B3,$A6,$BF,$CF
	.dc.b	$1A,$FA,$3E,$4B,$F3,$1F,$09,$23,$F2,$D1,$F8,$BC,$BF,$39,$F0,$72
	.dc.b	$7F,$15,$15,$6B,$F0,$36,$7C,$5F,$E8,$CD,$46,$7E,$47,$F4,$DF,$0B
	.dc.b	$2F,$F4,$91,$5E,$7E,$0F,$FB,$33,$07,$C8,$FE,$B4,$D6,$1D,$DF,$D2
	.dc.b	$92,$BF,$CB,$D3,$C1,$BF,$21,$FE,$75,$F4,$BF,$D1,$71,$FE,$6F,$6B
	.dc.b	$E0,$3F,$D0,$F2,$9E,$1F,$1D,$FB,$D9,$87,$17,$FC,$6F,$F3,$9F,$E4
	.dc.b	$1E,$7F,$F0,$4B,$67,$FD,$10,$D4,$40,$E5,$FF,$76,$F6,$BB,$FF,$C2
	.dc.b	$7F,$23,$FE,$59,$6C,$66,$7F,$D2,$93,$97,$BF,$8B,$E9,$6E,$5E,$48
	.dc.b	$C4,$78,$4F,$0B,$87,$FD,$F7,$A3,$E3,$BF,$B8,$EF,$F9,$C8,$FF,$58
	.dc.b	$70,$7C,$99,$CB,$C3,$E4,$2D,$FF,$AC,$2B,$94,$EC,$FA,$1F,$17,$AD
	.dc.b	$FF,$B6,$E1,$6B,$FD,$4F,$5F,$E9,$76,$DF,$FA,$D5,$AF,$F9,$65,$AF
	.dc.b	$A3,$E3,$B9,$7C,$87,$36,$D7,$DF,$70,$FE,$07,$E4,$2D,$7E,$07,$1F
	.dc.b	$EF,$FF,$0F,$E3,$F8,$9F,$23,$6F,$F1,$0E,$1F,$C5,$7C,$4D,$BF,$F0
	.dc.b	$D7,$81,$FE,$A1,$9F,$C7,$7C,$57,$C4,$F0,$FF,$71,$E0,$7B,$E3,$15
	.dc.b	$BB,$E3,$3F,$0F,$9F,$2F,$CC,$87,$D8,$FF,$30,$C7,$FC,$AE,$D8,$FF
	.dc.b	$6B,$35,$FB,$CD,$7F,$21,$F5,$2E,$FE,$93,$C7,$9F,$FC,$41,$39,$BF
	.dc.b	$F3,$54,$E5,$FC,$2F,$C8,$FD,$73,$EF,$F6,$3E,$75,$E1,$66,$B8,$FA
	.dc.b	$1F,$E2,$BD,$3F,$41,$8F,$F5,$B9,$F4,$CE,$67,$F4,$B3,$97,$FD,$E1
	.dc.b	$97,$EC,$EB,$FC,$F8,$D8,$FB,$1D,$2F,$8E,$F8,$D3,$9D,$7F,$C9,$3A
	.dc.b	$7F,$B7,$3E,$0F,$CF,$E4,$9F,$C8,$7C,$96,$1F,$87,$1E,$3C,$BE,$7B
	.dc.b	$74,$FF,$72,$E9,$F4,$23,$D1,$D5,$FE,$46,$F8,$DF,$CD,$1F,$9B,$F5
	.dc.b	$2D,$7F,$71,$E4,$FB,$E7,$E7,$FE,$46,$C7,$CE,$F3,$7B,$11,$FB,$AF
	.dc.b	$8F,$EA,$7F,$C5,$F1,$FE,$9D,$6E,$9F,$0B,$E0,$BF,$0D,$18,$E1,$FD
	.dc.b	$C7,$C6,$FE,$3B,$C9,$FD,$73,$70,$FF,$53,$7C,$03,$AF,$E0,$36,$3F
	.dc.b	$D4,$6D,$CB,$F8,$40,$70,$FE,$CF,$83,$F3,$DF,$99,$BC,$EF,$DE,$46
	.dc.b	$1F,$E0,$1D,$DF,$EA,$AC,$3F,$91,$E9,$FD,$A2,$3F,$D8,$11,$83,$F9
	.dc.b	$3D,$23,$86,$1F,$B7,$76,$FF,$39,$3C,$9D,$5E,$A2,$57,$A9,$FA,$F3
	.dc.b	$FE,$AF,$8E,$D2,$AE,$DF,$23,$E6,$F9,$2F,$3B,$EC,$F2,$7D,$3F,$25
	.dc.b	$70,$E3,$7E,$58,$FD,$B7,$CF,$7E,$BF,$1F,$E7,$ED,$4D,$35,$E3,$B7
	.dc.b	$FF,$6A,$61,$E4,$8C,$79,$1F,$23,$BB,$E6,$BF,$3B,$C8,$F9,$7F,$27
	.dc.b	$6F,$C7,$C1,$FE,$BF,$7F,$DC,$5F,$85,$74,$D3,$EC,$B5,$FE,$84,$D7
	.dc.b	$FA,$A3,$56,$5E,$5E,$6F,$BA,$7D,$AF,$8D,$B7,$E0,$F9,$1F,$BA,$7F
	.dc.b	$17,$EA,$FF,$0A,$C6,$4F,$F7,$A2,$F7,$7F,$5C,$D8,$FF,$78,$B7,$1F
	.dc.b	$C7,$B5,$FE,$B0,$BD,$BE,$D5,$DD,$B5,$F0,$BE,$4B,$F1,$FF,$17,$F9
	.dc.b	$06,$C6,$49,$87,$E2,$FE,$3F,$51,$F2,$BE,$33,$F2,$E7,$E6,$3E,$4A
	.dc.b	$C7,$CE,$6D,$7E,$4D,$BD,$DF,$9B,$F9,$1F,$CF,$FC,$97,$E8,$1B,$7F
	.dc.b	$5B,$E4,$30,$D3,$EC,$C7,$E9,$BD,$EF,$D3,$CD,$9D,$63,$FE,$E7,$F3
	.dc.b	$3F,$43,$8F,$EA,$7C,$5F,$E9,$D1,$8A,$FD,$4F,$47,$ED,$DF,$C2,$FC
	.dc.b	$86,$BE,$57,$F6,$6D,$8F,$CD,$FC,$FF,$FA,$29,$FC,$59,$FD,$DD,$C6
	.dc.b	$5F,$B1,$8C,$7C,$7C,$78,$FC,$95,$9F,$ED,$95,$FC,$EF,$4F,$F3,$2B
	.dc.b	$6B,$FC,$45,$F3,$AF,$57,$B7,$B3,$FD,$7E,$1E,$A7,$F3,$36,$D7,$F1
	.dc.b	$2C,$51,$87,$F8,$AB,$5F,$F0,$F2,$DF,$A5,$75,$6D,$7C,$0F,$90,$F7
	.dc.b	$1E,$FF,$91,$B1,$F1,$1B,$87,$B1,$BA,$27,$FC,$79,$E3,$FF,$89,$AB
	.dc.b	$E0,$FC,$7F,$F7,$98,$C5,$7C,$28,$F9,$9F,$23,$9F,$88,$B8,$7F,$14
	.dc.b	$C3,$FA,$B4,$7D,$1F,$8E,$93,$F5,$BE,$46,$D7,$F2,$1F,$9B,$78,$DF
	.dc.b	$D6,$5B,$5F,$1D,$B3,$FC,$E9,$AF,$F9,$4F,$D1,$72,$FE,$C7,$5F,$64
	.dc.b	$FF,$96,$2B,$8C,$7D,$AF,$1B,$FC,$29,$18,$C3,$FC,$BB,$E5,$7C,$FF
	.dc.b	$17,$F1,$1E,$3F,$E5,$1B,$3F,$8E,$F9,$1C,$34,$FE,$41,$CC,$9F,$CD
	.dc.b	$F8,$9F,$91,$F7,$AD,$7E,$59,$FC,$A7,$B9,$8D,$BF,$1E,$DD,$BD,$3A
	.dc.b	$7E,$07,$C4,$FF,$07,$37,$0F,$F5,$5E,$69,$CB,$63,$FA,$FF,$0F,$C4
	.dc.b	$F9,$1F,$D6,$D7,$E8,$DB,$3E,$8F,$0B,$F4,$DE,$27,$EC,$9B,$19,$26
	.dc.b	$9F,$A4,$F7,$F5,$1F,$CF,$FE,$C7,$F5,$B3,$FE,$85,$D1,$18,$8F,$9A
	.dc.b	$D7,$F9,$AB,$73,$FB,$3F,$0B,$FE,$4F,$8F,$DB,$ED,$FB,$08,$C7,$CA
	.dc.b	$7E,$EF,$E4,$FF,$C3,$3D,$3F,$EA,$EC,$FD,$ED,$CB,$FD,$0D,$97,$E1
	.dc.b	$2B,$7E,$8F,$1B,$FE,$0E,$DA,$FF,$35,$7D,$07,$E7,$A7,$FB,$CE,$1F
	.dc.b	$EA,$07,$FB,$3C,$66,$7F,$B8,$4E,$5E,$3D,$FF,$E8,$59,$D6,$27,$FC
	.dc.b	$19,$7C,$A7,$5F,$23,$FE,$32,$93,$49,$3F,$6A,$3F,$D2,$77,$E6,$9E
	.dc.b	$DF,$23,$CA,$BF,$61,$3F,$9F,$1F,$3B,$FE,$DC,$7F,$D7,$19,$FE,$92
	.dc.b	$31,$1F,$EB,$BE,$EF,$F5,$EC,$62,$3F,$CA,$CD,$51,$FE,$C2,$F8,$4D
	.dc.b	$57,$EB,$9D,$46,$79,$FF,$AF,$E1,$6F,$FD,$71,$3D,$A7,$31,$E6,$7F
	.dc.b	$78,$D3,$CF,$3B,$78,$99,$3F,$DA,$77,$EB,$9E,$38,$79,$2B,$85,$7B
	.dc.b	$46,$19,$6D,$78,$23,$07,$AA,$3B,$1F,$0A,$3A,$9E,$8A,$D1,$CC,$AE
	.dc.b	$4D,$C7,$A8,$E6,$7C,$A3,$AD,$FA,$5F,$6D,$78,$A7,$C8,$7C,$67,$AC
	.dc.b	$73,$AA,$FC,$2B,$F7,$BF,$81,$9E,$4F,$4A,$E4,$74,$7C,$91,$EF,$36
	.dc.b	$FC,$7D,$7E,$3B,$F4,$1A,$E6,$3E,$BC,$63,$E4,$FF,$CE,$5C,$7F,$E6
	.dc.b	$09,$54,$9C,$BF,$E5,$78,$CA,$8C,$D7,$3F,$65,$7C,$EB,$94,$61,$FB
	.dc.b	$27,$5F,$84,$E7,$87,$E7,$30,$FF,$56,$4F,$E4,$4E,$D5,$C6,$74,$DC
	.dc.b	$7C,$75,$C2,$BA,$9E,$5B,$EB,$3A,$AE,$13,$E4,$C3,$F1,$F7,$C7,$3F
	.dc.b	$16,$1E,$E7,$BC,$EA,$F8,$DF,$53,$CD,$7E,$1C,$77,$1E,$28,$F9,$37
	.dc.b	$D2,$3B,$E1,$D7,$1F,$26,$3D,$4C,$CC,$3F,$51,$7C,$A7,$71,$8A,$E4
	.dc.b	$FC,$B3,$91,$C3,$1F,$3D,$79,$6B,$E3,$D7,$C4,$AF,$84,$79,$EB,$EB
	.dc.b	$61,$E7,$8E,$C7,$47,$DA,$64,$95,$E4,$7E,$0E,$3C,$49,$1E,$87,$8B
	.dc.b	$E2,$9F,$74,$E1,$FF,$52,$70,$3E,$4F,$F1,$F9,$FF,$81,$34,$FE,$69
	.dc.b	$F2,$78,$DB,$FD,$6F,$68,$C6,$1C,$B0,$B7,$C2,$3F,$9A,$4F,$F8,$EE
	.dc.b	$31,$1F,$D4,$2B,$93,$FC,$E7,$E0,$38,$C1,$A6,$3E,$12,$CF,$F4,$EF
	.dc.b	$93,$D9,$87,$FA,$AF,$C9,$F2,$F8,$39,$1F,$D5,$22,$BE,$18,$F2,$60
	.dc.b	$69,$8F,$92,$F6,$F7,$FF,$76,$F6,$6F,$80,$7C,$F3,$87,$FC,$B1,$3B
	.dc.b	$C3,$FD,$D6,$D7,$F7,$67,$F7,$C8,$DD,$3D,$3E,$A4,$9F,$ED,$7C,$D5
	.dc.b	$FE,$E3,$CF,$F1,$8C,$72,$7D,$7D,$3F,$CD,$D3,$FE,$0F,$F4,$7C,$E7
	.dc.b	$CC,$F6,$2C,$79,$32,$4F,$88,$91,$FD,$AF,$D6,$F7,$FA,$12,$3F,$BE
	.dc.b	$C5,$7F,$C0,$7E,$57,$D8,$31,$97,$8F,$E3,$7E,$3F,$9B,$2B,$FC,$0B
	.dc.b	$15,$FF,$05,$FC,$6F,$E0,$0C,$57,$AF,$BF,$F1,$B2,$F9,$B7,$B3,$86
	.dc.b	$C7,$F8,$42,$71,$3F,$E8,$98,$E0,$D9,$DC,$7D,$23,$27,$F9,$0B,$E7
	.dc.b	$5E,$F7,$E6,$9B,$1F,$CD,$B0,$FD,$24,$62,$3F,$C4,$7D,$DB,$F9,$CF
	.dc.b	$3F,$D8,$B6,$A6,$6B,$FD,$01,$19,$7A,$78,$7F,$D0,$AD,$83,$E3,$FC
	.dc.b	$47,$2A,$D8,$9D,$FC,$7F,$9F,$E3,$E4,$EA,$72,$DB,$4E,$A3,$FA,$4E
	.dc.b	$1E,$19,$CC,$6E,$FE,$77,$A7,$F3,$FD,$39,$3F,$E9,$9E,$04,$E2,$3E
	.dc.b	$64,$27,$E8,$12,$3E,$77,$C0,$7C,$D2,$3C,$FD,$3F,$A1,$F1,$72,$3D
	.dc.b	$0F,$D3,$F8,$DF,$A9,$F1,$9F,$57,$E0,$25,$7B,$63,$EC,$7C,$67,$EE
	.dc.b	$3E,$C7,$FC,$8F,$3F,$23,$C8,$46,$23,$E0,$9F,$2F,$81,$5C,$7D,$7F
	.dc.b	$C1,$7C,$04,$AE,$57,$C7,$AB,$F0,$7F,$1B,$F8,$7F,$8D,$95,$F3,$E2
	.dc.b	$BD,$BA,$3E,$EF,$C6,$FE,$34,$CD,$FE,$31,$FC,$87,$C1,$CA,$FB,$11
	.dc.b	$F9,$3F,$82,$FC,$A7,$C6,$C9,$FB,$91,$5E,$DB,$A7,$18,$7E,$62,$BF
	.dc.b	$39,$D7,$2F,$F0,$8F,$E7,$FE,$37,$11,$CB,$E0,$FF,$44,$62,$FF,$D8
	.dc.b	$F0,$F9,$67,$87,$FB,$67,$F3,$11,$FE,$DD,$6B,$F3,$8E,$11,$BF,$07
	.dc.b	$F6,$D3,$A9,$FE,$4A,$D7,$FC,$42,$D7,$F4,$19,$3E,$13,$FA,$7B,$63
	.dc.b	$F4,$0D,$7C,$12,$F0,$7F,$09,$EB,$49,$FB,$6F,$2F,$8E,$FE,$1F,$BB
	.dc.b	$E0,$FF,$85,$36,$BF,$84,$1F,$36,$32,$7F,$40,$E7,$7A,$BE,$4F,$5E
	.dc.b	$57,$E2,$23,$F2,$9E,$A4,$8F,$C5,$BE,$AF,$8E,$FE,$3F,$F3,$DF,$92
	.dc.b	$9E,$ED,$7F,$8C,$35,$FE,$96,$D7,$FD,$EB,$83,$D2,$F8,$BE,$7C,$9F
	.dc.b	$CC,$C7,$C8,$F8,$E9,$1F,$21,$F9,$7E,$D7,$F3,$0F,$B0,$7C,$37,$F2
	.dc.b	$B6,$D7,$C2,$6B,$FC,$AD,$AF,$F9,$63,$C0,$FF,$78,$7C,$77,$D6,$F6
	.dc.b	$A4,$FD,$68,$FB,$3F,$1F,$FD,$BB,$F1,$FF,$C5,$CF,$ED,$DA,$FE,$A7
	.dc.b	$A9,$3E,$58,$F2,$10,$F2,$1F,$E0,$F8,$7B,$9D,$7A,$A4,$FE,$1A,$3F
	.dc.b	$C3,$3F,$30,$FA,$9F,$C3,$4F,$F2,$E8,$C5,$6A,$73,$3E,$3F,$92,$FE
	.dc.b	$AB,$F6,$3F,$E8,$08,$FE,$FA,$D7,$FC,$F2,$D7,$FC,$FE,$DF,$3D,$A3
	.dc.b	$D5,$F8,$44,$AF,$CD,$FC,$9F,$E7,$BE,$4E,$47,$E7,$9F,$D0,$FB,$BE
	.dc.b	$2E,$3F,$EC,$19,$7F,$45,$AF,$CB,$B5,$FE,$C0,$D7,$FD,$1B,$B3,$E1
	.dc.b	$BF,$C1,$31,$FC,$92,$31,$DD,$F1,$A7,$3E,$4F,$F9,$17,$8F,$FD,$CA
	.dc.b	$BC,$CF,$F4,$DD,$BE,$AB,$FA,$2E,$3F,$2D,$AF,$56,$D7,$3F,$94,$FE
	.dc.b	$FF,$F0,$8F,$63,$FB,$D3,$6F,$F6,$2D,$7F,$BD,$35,$FC,$86,$5D,$32
	.dc.b	$FE,$D8,$D8,$FE,$13,$EF,$FE,$03,$E8,$7F,$C4,$31,$E5,$EF,$F0,$FB
	.dc.b	$52,$30,$7F,$70,$18,$FE,$BD,$FF,$5C,$B5,$CC,$7A,$0F,$F2,$1F,$3B
	.dc.b	$DB,$FC,$F1,$27,$F8,$BF,$BF,$5F,$CF,$7B,$FF,$C7,$0C,$70,$7F,$9C
	.dc.b	$FE,$89,$F3,$7F,$E6,$E6,$D7,$F7,$F6,$BF,$A7,$D3,$53,$FC,$9A,$21
	.dc.b	$B7,$97,$BF,$23,$93,$D7,$E5,$B1,$1F,$D6,$6F,$F9,$13,$7D,$1C,$E3
	.dc.b	$E5,$7F,$CF,$D3,$A9,$FE,$04,$D7,$FC,$3C,$D7,$FC,$34,$BE,$3F,$D8
	.dc.b	$48,$F6,$7C,$95,$7C,$5F,$25,$F0,$4C,$61,$FD,$53,$E8,$5C,$DF,$9D
	.dc.b	$AF,$F7,$6C,$63,$BF,$F9,$CB,$63,$FC,$84,$D7,$FC,$EB,$F2,$DF,$A3
	.dc.b	$ED,$FE,$7E,$D8,$FF,$24,$35,$DF,$72,$7A,$FA,$FF,$C8,$A7,$C2,$D5
	.dc.b	$7F,$DD,$78,$F6,$7F,$94,$F1,$6E,$F9,$2F,$D6,$5F,$D1,$F9,$3B,$B7
	.dc.b	$E0,$E6,$F7,$7C,$C9,$B5,$F3,$26,$E1,$D1,$B8,$7F,$AB,$6F,$F1,$11
	.dc.b	$8F,$6F,$FB,$ED,$FD,$66,$BF,$59,$B7,$E6,$6D,$FF,$60,$BF,$C9,$47
	.dc.b	$6F,$67,$A4,$7E,$55,$C3,$F9,$7D,$EE,$1F,$8D,$F9,$0F,$2F,$C9,$F9
	.dc.b	$9B,$5F,$9D,$70,$FE,$7F,$C6,$B5,$FE,$3A,$F3,$3F,$95,$FC,$2F,$BB
	.dc.b	$D8,$F8,$71,$8B,$FD,$34,$63,$C2,$55,$3E,$2F,$89,$6B,$FA,$97,$37
	.dc.b	$C2,$C3,$F9,$BB,$6F,$FC,$5D,$3F,$EB,$E6,$E3,$FE,$BE,$E1,$FB,$7F
	.dc.b	$23,$DE,$FE,$8B,$6F,$F7,$13,$F5,$3E,$47,$EA,$B6,$3F,$B1,$1D,$C8
	.dc.b	$77,$3E,$9F,$D0,$FF,$8C,$9F,$F4,$03,$70,$FF,$28,$3F,$E9,$C8,$CF
	.dc.b	$C2,$FE,$12,$3F,$C4,$91,$C3,$E1,$71,$EB,$65,$F9,$3E,$63,$E0,$FE
	.dc.b	$13,$6B,$FD,$BA,$DC,$7F,$DB,$BB,$FF,$A1,$2B,$CB,$EA,$F7,$3C,$1F
	.dc.b	$21,$70,$FF,$3A,$61,$FE,$BC,$7D,$3F,$1D,$E0,$F8,$1F,$33,$E2,$F8
	.dc.b	$0F,$09,$27,$C2,$F9,$3D,$FF,$39,$ED,$F8,$FB,$3F,$13,$D5,$F8,$1E
	.dc.b	$9F,$C1,$6C,$78,$BD,$EF,$69,$1F,$6B,$F3,$BE,$3F,$E4,$7A,$DF,$25
	.dc.b	$B3,$F4,$3E,$3F,$E6,$1F,$37,$E3,$EC,$7B,$BE,$5F,$E4,$91,$F9,$2F
	.dc.b	$DA,$F0,$3E,$A1,$F5,$7E,$42,$CF,$F5,$1F,$9D,$7C,$67,$ED,$3A,$51
	.dc.b	$89,$FF,$2A,$B6,$3F,$DB,$CD,$44,$E9,$FF,$5C,$FB,$9F,$70,$95,$F7
	.dc.b	$1F,$BF,$E8,$7E,$08,$FC,$47,$A9,$67,$FD,$59,$F8,$2F,$F5,$47,$F2
	.dc.b	$3E,$2F,$EC,$23,$18,$FE,$4A,$31,$ED,$7E,$AE,$4D,$25,$7F,$BD,$72
	.dc.b	$4F,$F9,$CE,$3B,$FB,$5F,$99,$F9,$1E,$CF,$F2,$0C,$7F,$CA,$FC,$3F
	.dc.b	$9D,$F9,$1C,$6A,$8C,$61,$FD,$EE,$1F,$03,$F2,$DB,$1F,$E7,$E6,$DF
	.dc.b	$C8,$23,$87,$07,$F9,$9B,$C0,$EF,$5E,$06,$D7,$ED,$0F,$D8,$1F,$CD
	.dc.b	$3E,$7B,$FD,$34,$E4,$FF,$7C,$38,$AF,$DC,$C6,$FE,$4B,$8B,$AF,$91
	.dc.b	$E5,$E3,$73,$6C,$64,$95,$FE,$D1,$BF,$C2,$7B,$5F,$D3,$3C,$C4,$2F
	.dc.b	$C9,$E3,$4C,$3C,$7E,$2D,$C7,$93,$E9,$3D,$5E,$2D,$9F,$4B,$87,$D7
	.dc.b	$F1,$F6,$BF,$9B,$72,$FF,$47,$7C,$9F,$1D,$F8,$7F,$63,$E2,$3F,$17
	.dc.b	$E3,$B1,$87,$C4,$6D,$7A,$FC,$6F,$94,$7C,$BF,$8D,$B5,$F2,$BE,$37
	.dc.b	$0D,$3E,$6C,$7E,$1F,$E6,$FF,$CE,$11,$F4,$3E,$37,$1E,$87,$D5,$75
	.dc.b	$E7,$FE,$BE,$8D,$25,$7F,$B1,$21,$E7,$F0,$8F,$F2,$44,$61,$B3,$F8
	.dc.b	$1E,$BF,$C1,$7C,$75,$9F,$C0,$B7,$5F,$B1,$3F,$87,$EE,$C7,$57,$F8
	.dc.b	$13,$C2,$FD,$2F,$AD,$F8,$BE,$FF,$D8,$37,$2F,$C5,$F8,$F8,$69,$FB
	.dc.b	$88,$FC,$97,$83,$F9,$3F,$8B,$B3,$F9,$27,$F5,$9F,$4E,$F5,$7F,$88
	.dc.b	$36,$7F,$DC,$6D,$DB,$F4,$61,$C3,$FE,$C3,$F0,$7F,$88,$CF,$F9,$AE
	.dc.b	$78,$6B,$F9,$39,$CD,$23,$FC,$42,$F3,$BF,$EF,$7F,$82,$FF,$58,$8D
	.dc.b	$61,$FC,$73,$C3,$FF,$70,$7C,$6E,$BD,$8F,$F1,$0F,$BD,$FC,$4B,$2F
	.dc.b	$D0,$3F,$C4,$2F,$F9,$1C,$7C,$BC,$7A,$3A,$F3,$39,$4E,$8F,$E5,$13
	.dc.b	$F8,$27,$19,$32,$FF,$78,$7C,$6F,$9B,$A3,$C2,$F8,$BC,$CC,$5F,$0A
	.dc.b	$FF,$7B,$46,$2F,$FC,$35,$87,$FA,$1E,$BF,$91,$36,$3F,$B6,$B5,$FE
	.dc.b	$29,$CC,$8C,$47,$C3,$D1,$F1,$9F,$23,$DA,$F8,$6D,$C3,$D3,$87,$1F
	.dc.b	$63,$E3,$FC,$67,$CE,$6C,$EB,$24,$66,$3D,$58,$FF,$9E,$5D,$47,$D6
	.dc.b	$F6,$7F,$C5,$14,$69,$27,$FC,$C1,$4F,$89,$E1,$1D,$1B,$A7,$DE,$3E
	.dc.b	$74,$FF,$96,$5F,$F1,$73,$5F,$E6,$DB,$23,$11,$F0,$FE,$73,$FC,$B3
	.dc.b	$1F,$E6,$DB,$F4,$C7,$E1,$BE,$2B,$F0,$EF,$E2,$3C,$BF,$C4,$FC,$4D
	.dc.b	$AF,$7B,$70,$F2,$63,$DE,$BF,$1F,$F1,$39,$E7,$FC,$7F,$C4,$F8,$09
	.dc.b	$8F,$C0,$7C,$7F,$11,$F9,$67,$F2,$ED,$8F,$F6,$86,$3F,$DE,$3C,$7F
	.dc.b	$EF,$3F,$3B,$FD,$6E,$3F,$A9,$CE,$3C,$6F,$D8,$D5,$9C,$8A,$E0,$FF
	.dc.b	$5B,$8F,$F0,$1F,$89,$88,$78,$9F,$E9,$08,$C4,$27,$77,$FD,$82,$B4
	.dc.b	$F8,$27,$B3,$9C,$3C,$51,$DC,$C9,$E4,$C3,$AD,$EA,$3C,$33,$5F,$01
	.dc.b	$C0,$FF,$0C,$DF,$FA,$92,$B8,$C3,$0F,$54,$F9,$23,$FB,$AD,$FA,$1F
	.dc.b	$87,$3C,$2B,$77,$DE,$BC,$57,$D7,$0F,$65,$6B,$0F,$11,$F4,$CB,$3B
	.dc.b	$7B,$98,$BE,$8D,$38,$9E,$CC,$3A,$1E,$A2,$DF,$80,$F9,$51,$C3,$0F
	.dc.b	$2D,$74,$AD,$CF,$81,$AF,$CF,$6B,$E7,$6B,$F3,$36,$F9,$0F,$A1,$AF
	.dc.b	$D7,$6B,$EF,$73,$F1,$CD,$FF,$9B,$4F,$F2,$1A,$FF,$13,$4E,$76,$FE
	.dc.b	$69,$18,$AD,$DF,$26,$C6,$67,$EB,$5A,$DB,$7F,$8C,$6B,$F3,$31,$89
	.dc.b	$DB,$DD,$FC,$16,$7C,$DA,$7F,$98,$23,$83,$83,$B5,$75,$9E,$15,$DB
	.dc.b	$0F,$6E,$BE,$0A,$FF,$21,$36,$FF,$3B,$87,$86,$38,$63,$E4,$AE,$95
	.dc.b	$E1,$6B,$ED,$BF,$93,$87,$08,$FC,$F6,$1E,$6C,$3B,$46,$2B,$C7,$8F
	.dc.b	$B6,$72,$47,$04,$9F,$01,$E6,$C3,$D3,$19,$8F,$16,$3E,$76,$1F,$17
	.dc.b	$0F,$99,$8F,$97,$0E,$D3,$D0,$F2,$57,$C2,$C7,$CB,$19,$9D,$BD,$8D
	.dc.b	$9D,$71,$F8,$78,$7D,$4C,$3E,$11,$DF,$1F,$C5,$CF,$A2,$BA,$47,$B7
	.dc.b	$2F,$E7,$DF,$17,$FE,$75,$CF,$FB,$EF,$25,$D7,$FB,$13,$70,$FE,$93
	.dc.b	$56,$FF,$A3,$4F,$F8,$62,$7F,$A6,$B6,$33,$1F,$D5,$A3,$07,$E9,$AF
	.dc.b	$FC,$17,$97,$38,$C6,$1F,$DC,$7E,$2F,$13,$9F,$8A,$98,$7F,$68,$8E
	.dc.b	$18,$F8,$FE,$71,$F3,$94,$CD,$7F,$81,$E3,$2F,$F8,$BA,$51,$88,$FF
	.dc.b	$49,$C6,$27,$D5,$39,$9F,$EC,$AB,$E3,$3E,$27,$7F,$F8,$27,$E1,$EF
	.dc.b	$71,$F2,$4E,$1F,$8D,$3B,$BF,$F0,$AC,$62,$B8,$47,$F8,$63,$E0,$B9
	.dc.b	$91,$F5,$79,$55,$FE,$D3,$8B,$C9,$F8,$46,$78,$BF,$D2,$D0,$F8,$BF
	.dc.b	$2F,$14,$BF,$89,$15,$FF,$12,$C7,$8C,$8F,$C7,$97,$AB,$F0,$3F,$1E
	.dc.b	$FF,$75,$CF,$F8,$89,$AF,$92,$35,$F1,$1F,$57,$CC,$93,$B8,$FB,$1E
	.dc.b	$B7,$3F,$80,$93,$F5,$E2,$BE,$EF,$33,$EF,$18,$D7,$DB,$F1,$9F,$95
	.dc.b	$97,$D4,$3D,$DC,$36,$3F,$05,$38,$9F,$CC,$47,$06,$CE,$E3,$FC,$E6
	.dc.b	$D7,$F0,$3E,$6F,$E3,$FE,$36,$47,$E3,$63,$F2,$5F,$1B,$F1,$6E,$4F
	.dc.b	$E4,$62,$BF,$63,$CF,$FC,$C1,$8C,$BE,$BF,$C5,$FE,$6C,$93,$F9,$9F
	.dc.b	$8D,$AF,$E0,$3A,$FF,$40,$62,$BF,$CF,$FF,$3A,$E1,$F8,$8B,$E1,$A9
	.dc.b	$88,$B7,$FB,$99,$D3,$67,$E2,$C6,$23,$FD,$2A,$BE,$1F,$F6,$1F,$1B
	.dc.b	$27,$FC,$F6,$27,$4F,$EC,$7B,$BF,$6D,$F1,$B2,$30,$7E,$D7,$B3,$F7
	.dc.b	$44,$7F,$71,$1F,$C3,$FE,$69,$F1,$DF,$91,$F8,$39,$5F,$E8,$39,$CB
	.dc.b	$F9,$38,$E1,$F1,$D2,$30,$6F,$E3,$BA,$11,$E5,$DB,$DF,$E3,$A4,$75
	.dc.b	$74,$FF,$96,$30,$F9,$CF,$87,$87,$CC,$70,$75,$F1,$BE,$92,$3E,$7E
	.dc.b	$2A,$F3,$E2,$F7,$18,$E3,$70,$7F,$AD,$71,$7C,$6F,$C3,$EA,$98,$7C
	.dc.b	$18,$AF,$96,$2D,$F2,$8F,$F5,$75,$72,$7E,$BF,$D3,$FF,$9F,$DB,$E8
	.dc.b	$7E,$AE,$0D,$F4,$70,$FE,$A2,$33,$3F,$D0,$5A,$FF,$B6,$9A,$88,$17
	.dc.b	$A3,$F7,$7E,$36,$4F,$92,$3F,$96,$F5,$3E,$0E,$92,$7F,$01,$F1,$DA
	.dc.b	$F8,$4F,$E5,$B5,$FE,$B8,$9F,$F7,$5F,$A2,$8C,$47,$88,$46,$67,$8F
	.dc.b	$CB,$7E,$3F,$E3,$E5,$7F,$AF,$AD,$3E,$4F,$83,$93,$97,$F2,$DF,$1F
	.dc.b	$E8,$F8,$5F,$19,$27,$C7,$68,$FE,$23,$F3,$8F,$2E,$92,$BF,$B6,$46
	.dc.b	$5F,$EB,$1C,$8F,$90,$FD,$21,$23,$F4,$5E,$F7,$C8,$F8,$2F,$86,$48
	.dc.b	$F0,$BF,$DD,$69,$F2,$1F,$AD,$24,$FE,$AF,$C5,$FD,$87,$C2,$FE,$C7
	.dc.b	$E4,$24,$FE,$BE,$3E,$AF,$57,$D0,$24,$7B,$9F,$F0,$17,$CA,$3E,$3F
	.dc.b	$F7,$53,$0F,$F7,$D4,$65,$FF,$75,$ED,$FD,$CF,$BB,$F3,$1D,$08,$C7
	.dc.b	$83,$F8,$B8,$C7,$C0,$FF,$84,$F0,$7C,$7F,$F7,$4B,$FE,$ED,$8F,$E0
	.dc.b	$9A,$FE,$06,$FA,$7C,$17,$F7,$49,$CC,$FE,$5A,$7C,$1F,$01,$D7,$E3
	.dc.b	$E4,$F9,$A3,$C1,$DD,$23,$D6,$F8,$FC,$0F,$17,$BD,$D4,$93,$E3,$7E
	.dc.b	$2F,$7F,$A7,$E3,$E4,$FC,$48,$F6,$78,$12,$3E,$3B,$EF,$F8,$FF,$77
	.dc.b	$BB,$E9,$24,$FB,$DF,$A3,$F1,$FF,$17,$C1,$93,$F4,$23,$E4,$7C,$7F
	.dc.b	$C7,$F7,$A4,$FD,$48,$AF,$C7,$E2,$4F,$C7,$24,$7C,$DF,$6F,$E8,$11
	.dc.b	$FB,$BE,$0D,$7F,$DF,$FC,$FF,$E8,$AE,$AF,$E2,$9F,$10,$8C,$75,$FD
	.dc.b	$DC,$98,$FD,$DF,$36,$63,$F8,$87,$F0,$1F,$1F,$88,$EB,$F1,$7F,$83
	.dc.b	$31,$9F,$6C,$BF,$87,$CF,$E4,$3C,$DF,$C4,$1F,$89,$F0,$6C,$7E,$53
	.dc.b	$E2,$FF,$18,$61,$F4,$BF,$9C,$23,$FE,$4B,$AF,$F5,$FB,$FC,$7F,$67
	.dc.b	$8B,$F9,$0C,$7D,$ED,$DF,$4D,$A5,$7C,$33,$FC,$DD,$19,$79,$F8,$58
	.dc.b	$33,$E0,$FF,$88,$7E,$59,$F0,$3C,$B0,$FE,$3C,$23,$37,$FE,$77,$6C
	.dc.b	$69,$AF,$F2,$5F,$8E,$73,$BE,$5F,$FC,$FD,$18,$6C,$66,$7B,$CE,$5E
	.dc.b	$9E,$7B,$C9,$F0,$36,$33,$18,$3F,$CE,$9A,$FF,$96,$F6,$FE,$91,$BF
	.dc.b	$C8,$F8,$DF,$EB,$4F,$CA,$F1,$AC,$FC,$A9,$B8,$7C,$AB,$FF,$30,$4F
	.dc.b	$23,$E7,$D6,$B6,$FA,$BF,$1B,$6B,$C1,$6C,$7E,$BF,$8B,$FD,$A6,$3F
	.dc.b	$AC,$DF,$DC,$8C,$47,$DA,$6B,$E1,$6C,$7E,$0B,$E2,$FB,$47,$76,$E5
	.dc.b	$E5,$AF,$47,$C5,$E6,$31,$7F,$8A,$8C,$4F,$96,$7F,$13,$E1,$FA,$7E
	.dc.b	$2F,$D4,$D8,$FE,$57,$F0,$D8,$69,$D3,$44,$FC,$1F,$8A,$F8,$5E,$B5
	.dc.b	$AD,$37,$1F,$01,$F3,$67,$E4,$78,$B9,$DF,$F9,$BF,$CE,$3E,$1B,$E1
	.dc.b	$FA,$28,$C6,$5F,$DD,$1B,$8F,$F3,$36,$BF,$3B,$47,$C5,$FD,$CB,$E5
	.dc.b	$87,$D7,$8C,$5F,$2C,$7F,$5B,$F1,$7F,$6C,$FB,$9E,$45,$CF,$E7,$63
	.dc.b	$E3,$F8,$6F,$2E,$06,$92,$7F,$27,$D7,$F8,$5F,$03,$F0,$CD,$8E,$75
	.dc.b	$FE,$6F,$6F,$C3,$7E,$2F,$93,$CA,$F9,$BE,$17,$B9,$27,$E1,$3E,$2F
	.dc.b	$85,$FC,$97,$C5,$7B,$1B,$3D,$BE,$1B,$DE,$7E,$5D,$A7,$8B,$E1,$BD
	.dc.b	$64,$7E,$73,$EE,$F8,$4F,$D0,$7C,$AF,$F9,$AA,$7D,$31,$8F,$8B,$CF
	.dc.b	$C5,$6B,$9E,$D7,$E4,$AF,$E7,$3F,$DC,$77,$7C,$27,$CA,$24,$7D,$D7
	.dc.b	$F4,$33,$F3,$63,$97,$C5,$F8,$B9,$3C,$6F,$93,$83,$CB,$C3,$6B,$C0
	.dc.b	$FE,$07,$F0,$A7,$B9,$F8,$C6,$CF,$F8,$F9,$BD,$BF,$DF,$5A,$FF,$8F
	.dc.b	$FE,$5D,$BF,$ED,$A1,$18,$AF,$E4,$F1,$89,$D4,$E6,$3F,$C9,$1E,$2B
	.dc.b	$57,$C6,$FC,$CE,$6F,$9A,$DC,$7F,$0B,$7F,$8E,$23,$FA,$17,$E2,$F2
	.dc.b	$FD,$43,$EA,$F3,$59,$FA,$1C,$DF,$63,$E3,$3E,$CB,$63,$F3,$7F,$38
	.dc.b	$F8,$4F,$D7,$B6,$3F,$CC,$2D,$C7,$FC,$40,$D7,$F8,$DE,$8F,$8C,$FC
	.dc.b	$29,$F8,$6F,$8D,$B7,$F8,$1F,$8D,$FC,$49,$F8,$AE,$9B,$1E,$2F,$83
	.dc.b	$F0,$91,$F0,$BF,$82,$F1,$FF,$75,$9F,$F9,$86,$7F,$19,$18,$9F,$CA
	.dc.b	$FC,$64,$9F,$C8,$7C,$65,$D7,$F8,$8F,$2F,$D1,$8C,$79,$1F,$E0,$AD
	.dc.b	$D1,$8B,$FE,$E5,$1F,$E9,$17,$F3,$5E,$AF,$E8,$B0,$FD,$1B,$FD,$0F
	.dc.b	$E4,$FF,$CA,$D3,$F9,$F9,$C7,$C6,$EE,$3F,$99,$CF,$F9,$39,$BF,$1B
	.dc.b	$C7,$E3,$79,$36,$37,$1D,$79,$71,$87,$AA,$FE,$A6,$9F,$E2,$26,$DF
	.dc.b	$F8,$25,$FF,$50,$3E,$0A,$F9,$3F,$07,$E7,$F8,$CB,$3E,$47,$D5,$F1
	.dc.b	$98,$C7,$B3,$71,$ED,$E1,$7B,$FE,$33,$E0,$37,$0F,$6F,$C6,$7C,$33
	.dc.b	$E2,$72,$D8,$F8,$3F,$01,$F1,$CF,$ED,$F9,$3E,$5B,$FC,$6F,$3A,$3F
	.dc.b	$AE,$46,$7D,$4F,$D8,$49,$A4,$9F,$D8,$C7,$F8,$69,$E2,$ED,$BE,$97
	.dc.b	$82,$31,$5F,$2F,$E5,$7E,$D7,$C7,$7D,$B6,$DF,$D8,$F8,$CC,$34,$FB
	.dc.b	$71,$FD,$37,$E2,$FF,$C0,$F7,$F5,$7E,$33,$1F,$07,$F8,$57,$F0,$DF
	.dc.b	$1B,$8C,$3F,$06,$DA,$FB,$97,$F8,$B3,$E2,$FE,$69,$FE,$DE,$AF,$C4
	.dc.b	$C7,$C8,$F9,$9F,$F0,$7D,$7F,$B4,$DA,$FF,$B9,$9A,$FF,$BA,$5A,$A9
	.dc.b	$F1,$FC,$FE,$23,$BB,$70,$FE,$BF,$83,$E1,$BF,$3D,$D7,$EF,$6C,$FE
	.dc.b	$72,$B0,$D3,$8C,$7C,$5F,$84,$FD,$37,$8B,$67,$F4,$6F,$F8,$AF,$E5
	.dc.b	$9F,$1F,$9D,$7F,$6F,$18,$AF,$1D,$FC,$FF,$86,$FF,$02,$E8,$69,$27
	.dc.b	$FC,$B3,$8A,$DF,$1F,$C6,$7F,$C4,$B1,$87,$DF,$5E,$3E,$3F,$27,$C6
	.dc.b	$59,$DB,$73,$F2,$CF,$A7,$83,$1E,$07,$FB,$43,$4F,$C9,$57,$FB,$1D
	.dc.b	$AF,$FB,$15,$AF,$FB,$4D,$AF,$F9,$EB,$27,$C6,$7E,$37,$AF,$E3,$37
	.dc.b	$3F,$4F,$C6,$7F,$91,$A8,$D2,$47,$C2,$8F,$B4,$F0,$3E,$0F,$17,$D1
	.dc.b	$9F,$A5,$F1,$7F,$4F,$E3,$2D,$7C,$67,$EB,$71,$E3,$3F,$AA,$DD,$7F
	.dc.b	$D6,$79,$7F,$AF,$AF,$C1,$F1,$9E,$52,$3E,$3E,$2F,$C0,$47,$E0,$5B
	.dc.b	$1F,$C0,$30,$1F,$C1,$47,$CC,$7F,$82,$1C,$9F,$E8,$58,$DC,$66,$33
	.dc.b	$E0,$63,$D4,$DC,$6C,$E6,$FF,$A7,$A3,$BF,$4F,$19,$D3,$5F,$13,$9F
	.dc.b	$3D,$87,$0B,$E5,$5D,$0F,$2C,$E9,$C1,$E4,$29,$E3,$C3,$B6,$1E,$78
	.dc.b	$E3,$75,$E1,$3E,$18,$C4,$7C,$69,$F1,$3C,$0F,$DE,$DF,$93,$0E,$11
	.dc.b	$83,$E1,$CF,$2C,$3C,$D8,$78,$A7,$CB,$18,$BF,$6C,$F2,$34,$57,$27
	.dc.b	$CE,$8E,$13,$D8,$E7,$87,$A1,$F9,$73,$B8,$F4,$C7,$C5,$C3,$CE,$76
	.dc.b	$9D,$47,$C8,$3F,$09,$7E,$17,$B4,$F0,$9F,$80,$70,$7D,$86,$63,$71
	.dc.b	$93,$EC,$F4,$FF,$BC,$B1,$FF,$79,$56,$09,$5F,$94,$F8,$E7,$87,$FE
	.dc.b	$BF,$8F,$F7,$8C,$FE,$52,$73,$BF,$29,$E1,$5B,$C3,$F9,$8C,$62,$74
	.dc.b	$AB,$8B,$F5,$CC,$78,$5F,$A0,$C3,$F3,$D2,$8C,$65,$E1,$C7,$C9,$5A
	.dc.b	$6C,$79,$67,$77,$EB,$6B,$E1,$9D,$BE,$53,$F3,$37,$DD,$FC,$A3,$5C
	.dc.b	$C7,$C3,$CF,$C1,$8F,$DC,$AD,$63,$E8,$C3,$BC,$7B,$A3,$78,$76,$9F
	.dc.b	$DE,$DE,$E3,$BE,$1D,$71,$F7,$63,$E2,$BD,$C7,$B6,$B7,$31,$E8,$6C
	.dc.b	$C5,$F8,$30,$ED,$1F,$3F,$1F,$3D,$F6,$8C,$63,$E1,$7E,$41,$CC,$F8
	.dc.b	$38,$74,$9F,$14,$78,$0C,$92,$FB,$92,$3B,$3F,$0B,$0F,$19,$23,$C4
	.dc.b	$FF,$21,$CB,$FA,$E7,$C7,$7F,$22,$C3,$F9,$8F,$C7,$5D,$BF,$94,$D7
	.dc.b	$F6,$18,$C6,$5F,$D3,$6E,$CE,$63,$FA,$34,$FF,$A6,$A3,$11,$FE,$98
	.dc.b	$E2,$4E,$5F,$F1,$94,$7F,$6A,$F8,$DE,$70,$6F,$8F,$87,$3D,$72,$B8
	.dc.b	$1F,$DC,$BE,$37,$8F,$C6,$C8,$C1,$FD,$D9,$F0,$11,$FE,$B8,$F8,$F4
	.dc.b	$F6,$FC,$73,$BD,$78,$6D,$FF,$80,$2B,$13,$FE,$B8,$AE,$0D,$AD,$D7
	.dc.b	$F7,$98,$F8,$98,$FF,$B0,$A3,$FA,$AE,$FF,$CF,$76,$F8,$1F,$0D,$F3
	.dc.b	$2B,$55,$FE,$90,$5F,$1D,$99,$D7,$C7,$4A,$E7,$A7,$D8,$BE,$38,$F5
	.dc.b	$5F,$19,$EE,$FC,$11,$85,$3C,$11,$70,$FE,$BE,$36,$57,$08,$CC,$62
	.dc.b	$BE,$14,$E6,$7F,$DA,$F6,$F8,$EF,$01,$23,$CE,$35,$FC,$04,$66,$3E
	.dc.b	$07,$C6,$FC,$33,$19,$FA,$FE,$3B,$D5,$F1,$92,$B9,$E1,$F8,$99,$C4
	.dc.b	$7C,$99,$F9,$86,$2B,$FB,$64,$3E,$FF,$FC,$4E,$FF,$A2,$3E,$65,$F1
	.dc.b	$9F,$A0,$84,$61,$FF,$46,$CF,$05,$5D,$27,$51,$F2,$9A,$FF,$82,$9A
	.dc.b	$FF,$84,$9A,$FC,$3D,$3F,$DF,$AF,$B1,$AF,$E0,$FE,$3B,$C9,$E7,$3D
	.dc.b	$EF,$C4,$65,$F7,$7E,$3A,$5F,$18,$AF,$E2,$E3,$F1,$C6,$32,$FC,$37
	.dc.b	$C7,$7E,$47,$E3,$25,$78,$62,$BF,$93,$F8,$EF,$CB,$18,$AF,$97,$ED
	.dc.b	$FE,$E6,$7F,$D6,$2F,$E8,$A3,$3D,$9F,$2E,$35,$3F,$23,$AB,$FC,$67
	.dc.b	$1F,$A8,$8F,$F1,$7B,$5C,$CF,$E5,$1F,$F1,$EC,$65,$F9,$1F,$1D,$FA
	.dc.b	$92,$3E,$2E,$FF,$D6,$7C,$74,$8F,$98,$FE,$BF,$C7,$FD,$87,$C6,$7E
	.dc.b	$C7,$E3,$65,$7D,$18,$FD,$A7,$C6,$FE,$A6,$64,$FD,$58,$AF,$ED,$FD
	.dc.b	$0F,$DD,$18,$DB,$F7,$DF,$2A,$F8,$DF,$E5,$2D,$AF,$CC,$B5,$FE,$00
	.dc.b	$6B,$E3,$63,$A7,$8C,$F2,$F8,$D9,$3C,$9D,$BF,$34,$FD,$75,$3E,$5B
	.dc.b	$C3,$D3,$27,$C3,$15,$E1,$E0,$A9,$F1,$BE,$83,$59,$FB,$3E,$37,$D7
	.dc.b	$F1,$B3,$0F,$5C,$57,$9D,$A7,$15,$C6,$7E,$0F,$C6,$FF,$32,$FD,$BB
	.dc.b	$BC,$7F,$06,$F2,$D1,$88,$E7,$3B,$CB,$F9,$94,$E6,$7F,$CF,$B5,$C2
	.dc.b	$72,$7F,$55,$C5,$F1,$DF,$48,$95,$F4,$BE,$3B,$EA,$FB,$32,$3E,$AB
	.dc.b	$F6,$3E,$3B,$CD,$AF,$80,$93,$FC,$93,$3F,$73,$FE,$CA,$C5,$F1,$DF
	.dc.b	$80,$F7,$64,$FE,$02,$3F,$07,$F1,$D2,$3F,$06,$FE,$1B,$C0,$F8,$99
	.dc.b	$FB,$89,$3D,$DF,$67,$C6,$7E,$33,$D7,$93,$F8,$C8,$FE,$35,$93,$E0
	.dc.b	$FF,$55,$3F,$75,$B1,$98,$FD,$73,$5F,$F5,$0F,$23,$F3,$5F,$E2,$71
	.dc.b	$FE,$09,$6B,$FD,$71,$AF,$F2,$46,$F3,$FE,$2C,$FD,$0F,$C0,$49,$FD
	.dc.b	$0C,$7F,$83,$3E,$57,$FD,$DF,$1F,$CF,$1F,$F7,$EB,$7E,$3B,$5F,$1D
	.dc.b	$F8,$7D,$3F,$02,$49,$FA,$4F,$F8,$4E,$9F,$1B,$D7,$E3,$A5,$75,$8F
	.dc.b	$F0,$0F,$CB,$F8,$23,$F4,$34,$F4,$FC,$86,$B1,$DE,$1F,$E0,$1B,$34
	.dc.b	$97,$FD,$9A,$39,$3F,$9B,$6B,$FB,$5E,$D4,$63,$7F,$F0,$26,$3F,$D3
	.dc.b	$5F,$F6,$14,$EB,$E3,$FE,$37,$C7,$4C,$3A,$47,$AA,$3C,$84,$8F,$13
	.dc.b	$E7,$F8,$FF,$79,$23,$E7,$3F,$47,$E3,$FE,$97,$C7,$C9,$FA,$51,$F1
	.dc.b	$B9,$BE,$01,$23,$DC,$FC,$FF,$8F,$FB,$44,$8F,$B5,$1F,$77,$E3,$FE
	.dc.b	$79,$23,$EF,$35,$FF,$77,$E7,$F1,$23,$E7,$12,$3F,$C2,$BC,$AD,$BF
	.dc.b	$C9,$71,$F6,$3E,$27,$F1,$47,$E2,$FE,$32,$C7,$E2,$FE,$33,$F1,$C6
	.dc.b	$1F,$A5,$C4,$F8,$BF,$C9,$7C,$04,$9F,$B1,$15,$FC,$2B,$F9,$73,$1C
	.dc.b	$9F,$63,$D9,$FC,$D7,$B1,$2B,$F3,$51,$5F,$C6,$FB,$5F,$9F,$31,$5F
	.dc.b	$80,$7F,$30,$48,$F7,$EB,$FE,$18,$47,$F9,$DB,$E5,$3F,$C8,$AF,$F8
	.dc.b	$89,$FF,$20,$CE,$7E,$2F,$FC,$EF,$1A,$9F,$E7,$AE,$9F,$F1,$93,$FE
	.dc.b	$39,$BF,$F2,$A3,$6B,$94,$FF,$8E,$0F,$CB,$E5,$FE,$39,$7F,$C8,$FA
	.dc.b	$7C,$9E,$0F,$EE,$CD,$8C,$CE,$DF,$49,$FE,$E3,$F2,$DF,$2F,$FD,$F7
	.dc.b	$E4,$11,$89,$FF,$29,$C6,$27,$53,$99,$E6,$7E,$AB,$81,$EC,$78,$9B
	.dc.b	$1E,$66,$BC,$31,$FF,$48,$4F,$E3,$34,$7C,$4B,$5F,$99,$F1,$B7,$0F
	.dc.b	$93,$37,$7F,$8F,$36,$BE,$A4,$DA,$F9,$AD,$DD,$87,$F9,$96,$BE,$C4
	.dc.b	$E7,$F0,$77,$C0,$7D,$FC,$3F,$CC,$D1,$8A,$FD,$0B,$63,$E0,$E0,$8C
	.dc.b	$47,$CE,$8C,$7B,$FD,$6F,$B3,$6F,$FC,$D3,$68,$CC,$63,$DF,$B8,$7E
	.dc.b	$23,$E1,$FD,$59,$BE,$2E,$C6,$1B,$7F,$93,$8D,$FC,$5E,$2F,$F9,$06
	.dc.b	$3F,$E9,$CA,$FC,$C7,$C4,$FC,$7F,$03,$C5,$F1,$FD,$A3,$15,$BB,$E3
	.dc.b	$3E,$2F,$8A,$CB,$F3,$E1,$F5,$9F,$E7,$AC,$7F,$C9,$0D,$8F,$F4,$EB
	.dc.b	$5F,$C4,$B5,$FC,$AF,$EA,$2F,$63,$A4,$F1,$E6,$FF,$44,$CE,$6F,$F4
	.dc.b	$93,$97,$F6,$9F,$17,$F1,$8F,$C1,$76,$3F,$22,$FF,$67,$4F,$ED,$E7
	.dc.b	$AE,$9F,$7F,$3F,$4B,$70,$34,$D6,$2F,$57,$E0,$37,$0F,$E0,$5F,$19
	.dc.b	$89,$F9,$3F,$25,$27,$F8,$37,$C7,$5C,$BF,$C8,$C7,$C8,$EE,$F0,$12
	.dc.b	$7E,$4B,$F9,$3E,$CF,$CC,$FC,$6F,$E0,$8F,$C1,$FC,$75,$AF,$F1,$C7
	.dc.b	$2F,$FB,$41,$BE,$07,$E2,$3B,$FE,$DB,$F3,$32,$F7,$12,$7E,$BB,$FA
	.dc.b	$1E,$97,$E2,$DF,$95,$9F,$F2,$23,$67,$F8,$DF,$7F,$E2,$63,$1D,$1F
	.dc.b	$C6,$B2,$FC,$4F,$77,$E6,$1B,$87,$F1,$BF,$C0,$5E,$97,$D1,$6C,$7F
	.dc.b	$A3,$DB,$97,$FB,$91,$AF,$EB,$7C,$17,$C8,$7E,$67,$CC,$FF,$33,$DF
	.dc.b	$F3,$38,$C7,$C4,$FE,$A2,$FF,$53,$CB,$6B,$F9,$C3,$F9,$9F,$9C,$79
	.dc.b	$BF,$83,$6C,$FF,$A8,$5B,$5F,$EF,$66,$BF,$BF,$EA,$7C,$27,$F0,$F6
	.dc.b	$D7,$FB,$9B,$E0,$FF,$D2,$4F,$FB,$A2,$33,$5F,$B4,$F1,$BF,$8B,$53
	.dc.b	$E4,$7F,$63,$18,$7F,$5D,$F2,$FF,$C3,$AB,$F9,$DB,$5F,$EA,$AD,$4E
	.dc.b	$7A,$3F,$87,$78,$9F,$BC,$3B,$FC,$8D,$CF,$F1,$3E,$07,$8B,$E4,$3F
	.dc.b	$80,$B6,$3F,$A9,$6A,$F8,$5F,$D4,$36,$3F,$D8,$CD,$CF,$FA,$29,$FA
	.dc.b	$79,$FC,$67,$4B,$C1,$FE,$1A,$7C,$0E,$5B,$87,$C7,$F9,$0F,$86,$7C
	.dc.b	$4F,$91,$B1,$FD,$8F,$E6,$1E,$DF,$EE,$5B,$1F,$ED,$76,$E7,$E2,$6B
	.dc.b	$FC,$D7,$97,$FD,$49,$3C,$EB,$E8,$FC,$7F,$F1,$BB,$E3,$E0,$FF,$51
	.dc.b	$F1,$D8,$7F,$A6,$B0,$FB,$9F,$25,$F6,$3E,$4B,$EC,$B6,$3F,$01,$1F
	.dc.b	$73,$E4,$B1,$A7,$FA,$FA,$FF,$B7,$D7,$E1,$7D,$AF,$C0,$FC,$7F,$E0
	.dc.b	$9B,$19,$3F,$15,$19,$7F,$19,$F2,$1F,$88,$F9,$1F,$C4,$B6,$3F,$21
	.dc.b	$E0,$61,$A7,$03,$F2,$9E,$B7,$E4,$23,$F2,$2D,$8F,$F6,$17,$EA,$8E
	.dc.b	$DF,$CA,$E7,$FD,$1E,$31,$3F,$9D,$8F,$CC,$FC,$86,$6A,$CF,$27,$F3
	.dc.b	$BC,$3F,$9E,$EC,$B3,$FA,$57,$FB,$B7,$CC,$3E,$3E,$98,$EE,$B5,$FE
	.dc.b	$D5,$FC,$F7,$FB,$13,$7C,$FF,$EA,$5F,$B8,$7F,$82,$A3,$07,$F8,$A5
	.dc.b	$DF,$2F,$E5,$1A,$FF,$86,$9A,$FF,$89,$9A,$FF,$90,$17,$BD,$DA,$BB
	.dc.b	$B6,$BF,$C8,$BD,$5F,$87,$F4,$BF,$D4,$31,$83,$FC,$93,$B7,$F2,$CF
	.dc.b	$93,$FF,$0A,$4F,$FA,$7E,$75,$92,$B5,$18,$7F,$99,$FC,$9F,$0B,$E2
	.dc.b	$D9,$FF,$7C,$7E,$5B,$E0,$B6,$3E,$3F,$57,$FB,$F4,$E6,$31,$5D,$72
	.dc.b	$FF,$6B,$3C,$B1,$F2,$FC,$77,$07,$FC,$0D,$BB,$E0,$E9,$FE,$01,$AF
	.dc.b	$D2,$67,$FE,$AD,$9D,$47,$F9,$64,$FC,$FC,$7F,$04,$6B,$FE,$4A,$6B
	.dc.b	$FE,$56,$6B,$FE,$70,$5D,$BD,$FB,$BC,$0D,$8F,$F3,$37,$E6,$AF,$88
	.dc.b	$F9,$A7,$F8,$55,$C4,$7B,$E3,$E8,$FA,$F9,$6C,$72,$7E,$3F,$8F,$FE
	.dc.b	$49,$F8,$CF,$F1,$17,$07,$FB,$13,$87,$F8,$2D,$FD,$C9,$CB,$7D,$FE
	.dc.b	$2F,$D9,$84,$63,$5F,$F6,$97,$80,$F2,$FF,$B8,$36,$FF,$94,$FA,$48
	.dc.b	$C5,$7F,$AD,$9A,$FF,$93,$FB,$BF,$DC,$75,$FB,$8B,$FD,$5F,$CA,$FE
	.dc.b	$DD,$FD,$C7,$91,$FA,$EE,$6B,$5F,$8D,$BF,$C2,$3F,$E9,$EC,$BC,$15
	.dc.b	$AC,$7F,$17,$3F,$9A,$AF,$CC,$47,$AA,$7F,$82,$72,$7A,$89,$5E,$AC
	.dc.b	$FE,$DF,$5B,$07,$CA,$F0,$AF,$13,$71,$EE,$7E,$8E,$5F,$2B,$FE,$23
	.dc.b	$6D,$7F,$B5,$1B,$3F,$E7,$46,$BF,$17,$E5,$7F,$86,$D7,$F0,$E6,$D7
	.dc.b	$C1,$EB,$FE,$23,$F2,$BF,$C4,$9B,$19,$25,$F9,$32,$F5,$77,$6E,$7E
	.dc.b	$57,$CA,$F9,$48,$F9,$5F,$47,$CA,$FF,$1E,$7F,$8F,$B6,$3F,$D7,$D7
	.dc.b	$FE,$CE,$F4,$7F,$BD,$FA,$BF,$EA,$EF,$7F,$FD,$45,$ED,$FF,$69,$AF
	.dc.b	$F5,$8D,$69,$FF,$5E,$CE,$AB,$FB,$04,$7F,$95,$DE,$57,$FE,$DA,$F2
	.dc.b	$F9,$4F,$FB,$12,$F1,$5A,$8E,$07,$63,$FC,$16,$E7,$0E,$31,$E7,$32
	.dc.b	$78,$AF,$C8,$F5,$32,$7F,$BA,$B0,$F3,$3D,$67,$9C,$62,$38,$61,$8A
	.dc.b	$F2,$C7,$8E,$30,$F9,$8E,$D5,$83,$CA,$B0,$F1,$BF,$ED,$6A,$AE,$4F
	.dc.b	$29,$92,$99,$9E,$98,$7C,$C6,$C7,$A6,$3E,$21,$F4,$2F,$CA,$78,$AF
	.dc.b	$D3,$3F,$06,$EB,$F2,$6F,$9B,$EA,$BE,$6F,$BD,$FA,$97,$CE,$3E,$01
	.dc.b	$E7,$70,$78,$23,$C4,$70,$3F,$07,$A7,$FA,$66,$7E,$E3,$5F,$AC,$D7
	.dc.b	$EE,$39,$F9,$1F,$E6,$3D,$DF,$D6,$AD,$52,$72,$FF,$59,$8C,$A8,$CD
	.dc.b	$73,$F0,$FF,$AA,$CE,$CE,$4E,$2B,$F2,$B4,$F4,$F9,$C6,$1B,$8F,$69
	.dc.b	$F0,$CF,$0A,$E5,$39,$AD,$DF,$9D,$AF,$96,$76,$F8,$B1,$FC,$CE,$1E
	.dc.b	$18,$B5,$FE,$70,$BF,$05,$FA,$B2,$F0,$61,$F7,$B0,$E1,$87,$E5,$5B
	.dc.b	$8E,$A3,$17,$C7,$0F,$34,$76,$3C,$31,$97,$C8,$7E,$AA,$BA,$E1,$F0
	.dc.b	$E3,$78,$47,$58,$F9,$DF,$80,$66,$7C,$58,$F1,$8F,$34,$FC,$4B,$F9
	.dc.b	$78,$FB,$B0,$F2,$57,$43,$94,$65,$C1,$30,$F7,$1D,$A3,$DB,$1D,$8F
	.dc.b	$8D,$8F,$D0,$7E,$95,$76,$AF,$21,$CA,$BF,$A8,$78,$3F,$8C,$9F,$F3
	.dc.b	$A7,$33,$E5,$31,$B7,$F5,$56,$E3,$FD,$26,$EE,$DF,$D4,$67,$FD,$2B
	.dc.b	$3F,$D3,$5B,$19,$8F,$E9,$91,$8B,$FF,$06,$47,$F6,$7F,$0F,$9C,$62
	.dc.b	$3F,$A2,$2F,$95,$CC,$EB,$E5,$25,$71,$3F,$B8,$78,$DC,$BC,$59,$18
	.dc.b	$3F,$BA,$3E,$12,$3F,$D7,$DF,$26,$9E,$EF,$07,$FC,$0D,$C7,$7B,$FF
	.dc.b	$23,$C1,$F1,$AB,$13,$F0,$2B,$78,$7F,$75,$8C,$5F,$E6,$23,$F7,$33
	.dc.b	$FE,$11,$F5,$3A,$12,$3F,$CA,$9F,$30,$F9,$2C,$47,$F6,$58,$FF,$0F
	.dc.b	$BA,$F2,$33,$9F,$FA,$AE,$7F,$DD,$F3,$FE,$8E,$F5,$7D,$07,$D6,$8C
	.dc.b	$C7,$F8,$9A,$3C,$A4,$7F,$C1,$8F,$D8,$F2,$BF,$C5,$51,$FE,$6A,$CB
	.dc.b	$FA,$7D,$B5,$FF,$0E,$74,$7F,$60,$8F,$F5,$E5,$F0,$BF,$F0,$7D,$F0
	.dc.b	$BF,$F3,$BC,$BD,$BF,$C4,$B6,$BB,$61,$CE,$B7,$ED,$7E,$42,$5F,$B9
	.dc.b	$BE,$37,$0D,$8F,$F3,$E6,$0A,$CD,$76,$F2,$A5,$7F,$A1,$0E,$71,$D7
	.dc.b	$EE,$FF,$96,$C7,$F8,$98,$7D,$DB,$E7,$3C,$4D,$8E,$53,$C3,$7F,$89
	.dc.b	$1F,$CE,$A7,$4F,$FB,$E1,$AF,$66,$BF,$CF,$1A,$B7,$7E,$95,$D2,$3E
	.dc.b	$5B,$6B,$BC,$EB,$6F,$2F,$C4,$7C,$CE,$2F,$CD,$7F,$00,$BE,$22,$99
	.dc.b	$CB,$EE,$46,$5F,$93,$E3,$FD,$43,$6F,$FA,$9F,$91,$E9,$FF,$00,$5E
	.dc.b	$D5,$31,$16,$78,$45,$7F,$C7,$D3,$C9,$FA,$5F,$05,$FE,$58,$6F,$A7
	.dc.b	$F7,$E3,$FC,$AB,$1A,$F5,$31,$3F,$C3,$A1,$3A,$7E,$C9,$88,$D4,$66
	.dc.b	$3F,$63,$F8,$C7,$EA,$9F,$CA,$B7,$87,$E5,$B7,$83,$FB,$7D,$3D,$5F
	.dc.b	$D5,$7D,$14,$C3,$F6,$EF,$EB,$7E,$4F,$F5,$DF,$01,$FA,$FF,$9C,$95
	.dc.b	$FB,$C8,$FD,$97,$C0,$FF,$0D,$FC,$75,$EF,$53,$35,$FA,$F8,$CB,$FC
	.dc.b	$13,$E0,$7F,$74,$6D,$FE,$0F,$F1,$FF,$BD,$F9,$99,$18,$3F,$85,$BF
	.dc.b	$C0,$7E,$62,$46,$0F,$E1,$EF,$F0,$5F,$91,$91,$83,$F8,$9F,$C8,$FF
	.dc.b	$09,$F1,$24,$60,$FE,$2E,$E1,$FE,$2C,$FF,$0F,$8F,$E2,$1F,$2F,$27
	.dc.b	$F8,$EB,$FC,$4F,$DA,$C4,$7F,$1D,$F9,$8F,$E2,$C6,$3A,$3F,$8E,$7C
	.dc.b	$9F,$E0,$FA,$25,$7F,$26,$8A,$FE,$03,$A3,$F4,$CF,$EE,$9A,$FD,$A9
	.dc.b	$FE,$A7,$E7,$3E,$03,$F9,$21,$27,$F9,$8F,$CB,$7D,$5F,$85,$91,$F4
	.dc.b	$DF,$E5,$4F,$F1,$3F,$B6,$71,$7E,$4D,$B1,$FD,$89,$AF,$E5,$1A,$FF
	.dc.b	$42,$E6,$7C,$27,$F6,$AF,$9F,$FE,$0F,$BF,$F0,$AF,$81,$46,$36,$FF
	.dc.b	$65,$5F,$0C,$3F,$0D,$F0,$3F,$8B,$23,$FE,$4A,$C3,$95,$BC,$CF,$E5
	.dc.b	$5E,$F3,$E4,$BF,$54,$DB,$FE,$F2,$D7,$F5,$6D,$7D,$0F,$9F,$E4,$A4
	.dc.b	$7E,$5E,$3F,$C0,$7F,$74,$E9,$A4,$9F,$26,$BF,$E5,$ED,$BD,$BE,$7F
	.dc.b	$F3,$DF,$A3,$72,$7E,$F5,$B7,$FE,$0C,$6B,$FB,$E6,$BF,$E6,$4A,$F8
	.dc.b	$FF,$2F,$F2,$FE,$62,$4F,$80,$4E,$9C,$3F,$3B,$E4,$3F,$C0,$1E,$A7
	.dc.b	$F8,$3E,$7F,$B0,$35,$FF,$6B,$71,$7F,$5D,$E3,$FE,$69,$F4,$4F,$2F
	.dc.b	$F8,$A3,$6B,$FC,$46,$D7,$F8,$AB,$5F,$45,$7F,$61,$7F,$B3,$F3,$BE
	.dc.b	$17,$F8,$31,$F8,$5F,$80,$FF,$0A,$7A,$CF,$90,$A4,$BF,$5E,$0F,$99
	.dc.b	$FE,$D1,$F3,$8F,$8E,$FE,$50,$DA,$FF,$1B,$B5,$FE,$52,$D7,$FC,$2F
	.dc.b	$F3,$FF,$E1,$37,$FC,$04,$D7,$FB,$EB,$77,$FC,$7F,$CA,$7E,$6F,$E0
	.dc.b	$E5,$79,$1F,$CF,$7C,$27,$93,$E2,$FC,$64,$9F,$1B,$FE,$2B,$C3,$FC
	.dc.b	$2F,$87,$F5,$E6,$F6,$7F,$A9,$BE,$05,$18,$BF,$BB,$B7,$AE,$FF,$B9
	.dc.b	$FD,$0B,$ED,$BF,$C9,$91,$86,$E3,$FE,$B0,$9D,$7C,$7F,$EC,$F3,$FF
	.dc.b	$08,$78,$BF,$CC,$B2,$FE,$9A,$D7,$FC,$0E,$D7,$FD,$6A,$D7,$C7,$F3
	.dc.b	$BF,$BD,$23,$F9,$3B,$7E,$4D,$FD,$36,$31,$97,$F7,$DC,$B9,$FC,$2F
	.dc.b	$88,$FF,$38,$DF,$AE,$31,$5F,$E7,$0C,$7F,$B6,$3E,$6A,$FE,$CD,$18
	.dc.b	$8F,$B7,$EF,$7F,$97,$77,$34,$FE,$25,$E9,$59,$FC,$5F,$C7,$FF,$15
	.dc.b	$30,$FE,$8B,$AB,$F8,$C1,$1F,$C8,$E1,$5F,$D3,$F8,$1F,$C7,$4C,$71
	.dc.b	$7F,$1D,$F8,$77,$B1,$FE,$8A,$6D,$7E,$D1,$AA,$78,$7B,$5B,$AE,$1E
	.dc.b	$9C,$AE,$11,$5F,$F1,$DF,$8C,$C3,$F9,$8C,$7F,$11,$6B,$FC,$11,$B0
	.dc.b	$F5,$DF,$3D,$FE,$9E,$6D,$7F,$08,$6B,$FC,$12,$75,$3B,$F9,$E9,$3B
	.dc.b	$8F,$F2,$D7,$5B,$D1,$FE,$27,$3F,$E1,$88,$C5,$FF,$AA,$9B,$19,$9D
	.dc.b	$BF,$E5,$6F,$97,$78,$0E,$9F,$F5,$8C,$61,$B1,$99,$FF,$12,$CE,$5F
	.dc.b	$07,$77,$F6,$36,$FB,$3E,$18,$C4,$7F,$99,$31,$46,$23,$F9,$D6,$BF
	.dc.b	$E4,$2E,$BF,$89,$F0,$3F,$E6,$2E,$4F,$EB,$53,$9A,$FB,$55,$C2,$B4
	.dc.b	$F2,$F8,$AB,$3F,$CD,$AB,$85,$F2,$ED,$CF,$07,$D5,$BB,$4F,$89,$FE
	.dc.b	$FF,$3F,$65,$B1,$9C,$7F,$C8,$B3,$96,$C7,$F8,$F5,$FB,$93,$D3,$E1
	.dc.b	$FA,$B6,$3F,$01,$1E,$0F,$89,$C6,$7E,$C3,$D1,$EA,$F9,$3C,$CF,$2B
	.dc.b	$6F,$F1,$0E,$1F,$C5,$7C,$D5,$AE,$59,$FE,$47,$E6,$3D,$BF,$35,$EE
	.dc.b	$6D,$FE,$45,$C3,$F9,$3F,$7E,$DF,$F3,$4E,$DF,$F1,$0F,$97,$FC,$99
	.dc.b	$BE,$3F,$EC,$2F,$FC,$CF,$39,$7E,$1F,$BF,$F4,$A3,$17,$E4,$E2,$4E
	.dc.b	$6F,$94,$E7,$B7,$E9,$FC,$C7,$D4,$6E,$3F,$D2,$3E,$13,$C2,$47,$C3
	.dc.b	$68,$FA,$FF,$31,$F7,$0F,$BB,$D7,$6B,$F6,$07,$E0,$3D,$6F,$C0,$B4
	.dc.b	$FE,$AB,$A7,$B8,$8F,$67,$F0,$BF,$30,$F9,$7F,$C5,$B6,$7F,$C0,$6F
	.dc.b	$FB,$C5,$B1,$EF,$D7,$7F,$41,$67,$6F,$C7,$CB,$FD,$8D,$1F,$31,$B3
	.dc.b	$FE,$CC,$E0,$46,$20,$7F,$5B,$70,$DF,$84,$E8,$79,$BD,$DB,$7F,$EF
	.dc.b	$3A,$7B,$FE,$AF,$90,$F1,$36,$75,$F0,$3F,$82,$A3,$49,$1F,$A4,$D5
	.dc.b	$F2,$1F,$13,$D5,$F8,$AD,$9F,$F7,$47,$D0,$37,$FF,$36,$EC,$8C,$4F
	.dc.b	$FA,$CA,$31,$3A,$9C,$C7,$C3,$F5,$FF,$12,$FC,$23,$FD,$C0,$E1,$FE
	.dc.b	$FC,$D7,$3D,$5F,$6B,$E4,$3E,$C9,$F6,$BE,$46,$E3,$F6,$BE,$43,$EE
	.dc.b	$FC,$8F,$DE,$6C,$7C,$9F,$82,$F8,$84,$7E,$C3,$F8,$6F,$5B,$F0,$A7
	.dc.b	$E1,$BD,$7B,$3F,$86,$F9,$1F,$C4,$E7,$F8,$A6,$C7,$D5,$F6,$BE,$81
	.dc.b	$1F,$C2,$3F,$92,$EF,$FC,$89,$F9,$2F,$91,$B3,$F9,$2E,$EF,$CA,$FC
	.dc.b	$67,$E5,$9B,$1F,$E5,$BE,$E7,$D0,$FF,$56,$8F,$DD,$B7,$8F,$F2,$0D
	.dc.b	$7E,$74,$F0,$F6,$BF,$47,$F2,$3F,$BF,$AE,$1E,$AF,$FA,$47,$B1,$39
	.dc.b	$AF,$E6,$3F,$15,$F9,$CA,$D7,$B3,$8F,$98,$FD,$5B,$74,$FF,$36,$D6
	.dc.b	$DF,$F4,$06,$3F,$A0,$73,$5D,$3E,$46,$47,$0F,$90,$BB,$7F,$24,$C1
	.dc.b	$1E,$23,$C7,$E3,$DA,$EF,$E2,$F9,$8F,$67,$AB,$63,$E4,$70,$7F,$9A
	.dc.b	$5F,$8F,$1F,$C9,$A3,$18,$FC,$88,$FE,$97,$8F,$EA,$2B,$8C,$7C,$2F
	.dc.b	$17,$AC,$74,$F4,$BE,$2F,$C7,$FC,$B6,$C7,$C2,$F8,$EC,$34,$E0,$7C
	.dc.b	$6F,$8C,$F9,$BF,$1D,$F3,$9B,$1F,$2B,$E0,$BE,$89,$F4,$BE,$36,$C7
	.dc.b	$FA,$77,$57,$D0,$FF,$AC,$63,$E8,$F8,$9F,$DB,$27,$3D,$BF,$6A,$31
	.dc.b	$5F,$E1,$FA,$FD,$C3,$6B,$F6,$87,$F1,$67,$0E,$1A,$FE,$E5,$FC,$3C
	.dc.b	$78,$DE,$5D,$28,$D7,$5F,$83,$B7,$F1,$7F,$1D,$27,$F0,$FF,$19,$71
	.dc.b	$FC,$03,$F9,$03,$F2,$3F,$1D,$67,$ED,$FA,$3F,$94,$F8,$DF,$CD,$36
	.dc.b	$3F,$23,$E2,$E1,$A7,$03,$E4,$7E,$76,$AF,$CE,$FC,$6F,$C7,$8C,$4F
	.dc.b	$E6,$A3,$E5,$7D,$5F,$F8,$6F,$3F,$8A,$D7,$F8,$D3,$5F,$E7,$AD,$55
	.dc.b	$FE,$F1,$AF,$B1,$3F,$E1,$E7,$F9,$14,$66,$7F,$C1,$82,$74,$BC,$6F
	.dc.b	$E6,$1F,$3D,$FB,$4F,$90,$FE,$F1,$EA,$7F,$59,$AF,$D4,$47,$EE,$3E
	.dc.b	$0F,$35,$72,$FE,$15,$7B,$E4,$FA,$D8,$75,$8F,$F6,$E4,$FF,$B9,$BD
	.dc.b	$E7,$C7,$F6,$F7,$3F,$C6,$71,$89,$D7,$9B,$96,$9D,$A7,$CB,$ED,$FF
	.dc.b	$85,$30,$FF,$48,$CF,$F8,$49,$B8,$FF,$67,$6B,$FE,$FD,$C1,$F1,$9C
	.dc.b	$AB,$C8,$D9,$FF,$14,$CF,$F8,$FD,$BF,$19,$E6,$F8,$BF,$3B,$6F,$D9
	.dc.b	$1E,$AF,$8C,$C5,$7F,$8E,$F8,$BF,$AE,$7B,$5E,$72,$5F,$9D,$FF,$27
	.dc.b	$E2,$F8,$BF,$A3,$1F,$11,$B3,$F2,$7E,$33,$64,$7F,$CA,$75,$FE,$51
	.dc.b	$C3,$FD,$2E,$FF,$12,$E2,$FA,$7C,$9F,$ED,$EB,$F9,$FE,$67,$D0,$7E
	.dc.b	$8F,$95,$F4,$BE,$2E,$E5,$F5,$DB,$97,$C4,$9F,$AF,$F1,$98,$CB,$AE
	.dc.b	$6F,$8B,$F5,$93,$1F,$2B,$EE,$F8,$CF,$BE,$FE,$01,$B1,$FC,$47,$1F
	.dc.b	$F0,$DF,$D0,$7F,$87,$3E,$69,$EE,$7F,$0A,$F7,$3F,$1B,$18,$AF,$E1
	.dc.b	$B5,$BA,$D5,$D9,$C8,$EF,$E0,$FF,$8A,$30,$FC,$BC,$62,$3C,$26,$EB
	.dc.b	$9B,$4E,$78,$6E,$BC,$73,$FC,$7A,$58,$78,$5E,$D8,$57,$AD,$F1,$8C
	.dc.b	$47,$01,$19,$6D,$75,$AF,$F1,$FE,$19,$3D,$13,$C5,$C4,$EE,$38,$9F
	.dc.b	$E2,$5C,$3C,$EF,$AA,$3D,$B7,$CA,$FC,$D5,$AA,$FF,$16,$46,$30,$F3
	.dc.b	$B6,$3A,$9A,$28,$9E,$4F,$CC,$C3,$CA,$FD,$46,$BC,$6F,$E0,$1E,$37
	.dc.b	$9C,$F2,$7D,$D8,$7D,$4A,$D1,$CA,$3B,$46,$4F,$85,$19,$8F,$9D,$C5
	.dc.b	$F8,$37,$F0,$AD,$7D,$4D,$7F,$06,$E7,$E4,$1E,$6B,$7F,$F3,$8C,$E6
	.dc.b	$31,$48,$C4,$EA,$FF,$CA,$D7,$CA,$31,$5F,$E7,$A8,$C3,$F8,$F0,$BC
	.dc.b	$AF,$F2,$F5,$7E,$6A,$31,$3F,$84,$C7,$AE,$3E,$2A,$D3,$63,$C3,$7C
	.dc.b	$9B,$5C,$4F,$14,$ED,$B8,$7E,$6F,$0E,$F1,$F8,$06,$E7,$DA,$75,$5C
	.dc.b	$6B,$D3,$87,$77,$DC,$D7,$DB,$1B,$BF,$83,$7E,$DA,$FF,$3E,$36,$BA
	.dc.b	$DC,$7C,$53,$AC,$3D,$71,$E5,$C3,$E9,$DF,$96,$36,$7A,$EB,$83,$EE
	.dc.b	$38,$5F,$C3,$7D,$25,$C3,$E0,$E3,$CC,$E9,$7E,$59,$EA,$72,$23,$92
	.dc.b	$63,$F5,$AF,$D5,$5C,$49,$19,$AF,$C1,$5F,$C4,$8F,$6D,$F2,$9F,$43
	.dc.b	$FC,$8F,$27,$07,$FB,$BA,$7F,$B8,$7C,$77,$FA,$97,$2F,$E6,$ED,$DF
	.dc.b	$F3,$35,$BA,$FE,$C3,$3F,$EB,$1B,$E6,$D9,$34,$C7,$93,$CE,$BF,$41
	.dc.b	$3F,$ED,$67,$FC,$01,$19,$AF,$2F,$C7,$5A,$D5,$7F,$70,$F8,$EC,$19
	.dc.b	$BD,$7C,$76,$F2,$95,$FE,$0B,$8A,$FF,$60,$8F,$31,$1F,$E6,$98,$FF
	.dc.b	$BA,$3C,$7F,$37,$77,$FB,$3F,$E1,$6F,$88,$7C,$D9,$2B,$31,$FE,$D3
	.dc.b	$AE,$18,$7F,$83,$E3,$17,$C6,$3B,$DB,$C6,$E6,$48,$FF,$7B,$F0,$7F
	.dc.b	$74,$97,$C7,$66,$7F,$86,$69,$FC,$32,$7F,$DC,$D1,$C2,$31,$F1,$9B
	.dc.b	$BE,$5F,$1D,$EA,$E5,$97,$B3,$F8,$96,$09,$CC,$7F,$89,$67,$D2,$47
	.dc.b	$D3,$E4,$FC,$4F,$4E,$47,$3E,$16,$0A,$CC,$FA,$0E,$FF,$1D,$F2,$8C
	.dc.b	$E5,$EF,$F8,$CF,$9F,$30,$E9,$68,$C4,$73,$C3,$BF,$C6,$FD,$43,$37
	.dc.b	$FE,$4E,$97,$E0,$EF,$33,$F2,$50,$FA,$95,$E1,$B1,$FE,$9F,$AE,$0F
	.dc.b	$FA,$86,$32,$FF,$7B,$6B,$FE,$9A,$6A,$23,$87,$11,$FE,$23,$7D,$BE
	.dc.b	$67,$E3,$0F,$2C,$66,$3D,$D9,$BE,$3F,$F2,$BF,$7C,$F8,$8A,$7E,$52
	.dc.b	$2D,$7C,$8E,$9F,$CB,$98,$79,$FC,$7F,$E6,$88,$FC,$EF,$90,$FC,$EF
	.dc.b	$C8,$48,$FA,$2E,$9F,$F4,$3F,$CF,$BD,$EF,$96,$BD,$FA,$62,$2C,$FF
	.dc.b	$88,$44,$ED,$C3,$5F,$F1,$33,$5F,$F6,$2B,$FE,$A9,$7F,$07,$D3,$FB
	.dc.b	$23,$C3,$EF,$7E,$D3,$E0,$A4,$FE,$21,$FD,$BF,$C8,$FE,$E3,$E0,$BF
	.dc.b	$73,$DB,$2B,$F1,$B1,$FB,$CF,$43,$E6,$FC,$34,$9F,$C8,$C5,$7E,$99
	.dc.b	$FC,$08,$C7,$5F,$F9,$AB,$E3,$1F,$21,$F5,$BD,$39,$5F,$A5,$9C,$BF
	.dc.b	$B1,$F2,$5D,$3F,$E0,$88,$FF,$23,$46,$2B,$F8,$64,$E6,$7F,$DC,$F3
	.dc.b	$97,$F8,$CD,$3E,$43,$F9,$53,$63,$C1,$1E,$27,$F5,$10,$F9,$0D,$92
	.dc.b	$3C,$DD,$7C,$BE,$42,$47,$A5,$D3,$CB,$E3,$7B,$91,$F7,$7C,$7D,$7B
	.dc.b	$FC,$7F,$90,$C6,$FF,$1F,$E3,$FF,$37,$EC,$CA,$F8,$D1,$5F,$7E,$FF
	.dc.b	$E2,$F7,$F5,$9B,$FC,$8A,$F8,$A7,$F5,$3D,$9F,$1D,$F0,$B9,$25,$FD
	.dc.b	$18,$F8,$BC,$52,$3E,$A3,$F2,$3E,$3F,$FA,$0F,$C6,$BE,$2B,$F9,$E3
	.dc.b	$6B,$EA,$B5,$FA,$6D,$7F,$80,$1F,$E3,$DC,$7F,$4C,$FE,$07,$DE,$FA
	.dc.b	$A7,$D6,$F0,$3E,$BF,$74,$9F,$C3,$3F,$6B,$C1,$F1,$FC,$3F,$88,$93
	.dc.b	$E2,$7E,$06,$3F,$80,$F8,$E9,$3F,$8F,$8F,$C1,$F9,$F2,$3F,$24,$FE
	.dc.b	$1B,$87,$FB,$A7,$2F,$F2,$5F,$86,$FD,$B3,$6B,$FD,$30,$D7,$DA,$7F
	.dc.b	$97,$7D,$04,$FE,$F6,$1F,$34,$F8,$6C,$36,$3F,$81,$35,$FF,$52,$35
	.dc.b	$F9,$38,$BD,$BF,$9D,$8F,$FB,$59,$7C,$6F,$F8,$1D,$B5,$FD,$45,$DF
	.dc.b	$B5,$F8,$6D,$5F,$0D,$F6,$89,$5F,$69,$FB,$BF,$07,$CF,$E3,$64,$FE
	.dc.b	$D6,$BF,$C3,$6D,$F8,$2F,$C2,$91,$F0,$7C,$6F,$E2,$3D,$6F,$C3,$92
	.dc.b	$3F,$0E,$FE,$2B,$2F,$47,$C6,$C9,$ED,$1E,$BE,$69,$1E,$B7,$DD,$C9
	.dc.b	$FE,$14,$FC,$93,$FC,$07,$3F,$C1,$9A,$FF,$04,$6B,$F9,$2F,$09,$18
	.dc.b	$8F,$F4,$9F,$03,$E3,$BE,$57,$C7,$49,$F9,$51,$F3,$79,$E4,$7C,$D7
	.dc.b	$E8,$74,$78,$FE,$07,$FC,$41,$8F,$F7,$FA,$F2,$46,$0F,$F2,$36,$3F
	.dc.b	$D6,$F9,$FE,$D7,$C7,$7D,$93,$ED,$7C,$77,$DB,$E8,$F7,$7C,$14,$C7
	.dc.b	$EE,$45,$7D,$D1,$E0,$23,$FE,$3A,$C1,$F1,$7F,$85,$24,$7E,$17,$E5
	.dc.b	$6B,$FE,$09,$9F,$BA,$DE,$BC,$E7,$F8,$FF,$8B,$FC,$71,$27,$F1,$DD
	.dc.b	$55,$E5,$D3,$F9,$33,$14,$E8,$FF,$60,$F9,$9F,$E4,$7A,$FF,$13,$35
	.dc.b	$FF,$0D,$36,$7F,$3D,$3F,$E6,$38,$C3,$CB,$E2,$70,$67,$E2,$ED,$FF
	.dc.b	$96,$AB,$EF,$3F,$EC,$D6,$F8,$3F,$D3,$AB,$FC,$F1,$1F,$E3,$F9,$F1
	.dc.b	$FC,$4F,$73,$C1,$F1,$56,$3C,$FF,$17,$E3,$30,$FF,$8E,$C7,$CB,$7D
	.dc.b	$0C,$7F,$BF,$36,$3F,$83,$9A,$E3,$FF,$70,$38,$8D,$3F,$C3,$3E,$3D
	.dc.b	$E8,$3E,$5F,$FC,$11,$18,$6C,$66,$7A,$CE,$5E,$9C,$4F,$7B,$C6,$D8
	.dc.b	$CC,$60,$F9,$1F,$1D,$F5,$E3,$E9,$F6,$DE,$4C,$6B,$FD,$DA,$B1,$C9
	.dc.b	$F7,$2A,$DF,$CF,$EA,$B7,$BA,$FB,$F1,$8E,$36,$17,$E3,$BF,$CA,$D8
	.dc.b	$7E,$19,$AF,$D9,$6E,$1F,$E0,$66,$D3,$E3,$79,$63,$F8,$C6,$DF,$E3
	.dc.b	$3E,$33,$C1,$1E,$16,$CF,$E4,$BE,$2B,$C9,$F1,$9E,$56,$C7,$F8,$57
	.dc.b	$C3,$7C,$66,$63,$0D,$C7,$F3,$3F,$15,$ED,$8F,$73,$67,$F3,$EE,$1F
	.dc.b	$D0,$FC,$65,$AF,$F4,$27,$97,$FC,$5B,$E2,$FE,$0B,$6B,$F8,$F4,$63
	.dc.b	$13,$F0,$BE,$2F,$E8,$D6,$EF,$F5,$71,$8A,$DD,$FE,$C3,$E2,$FE,$91
	.dc.b	$F4,$F0,$B1,$EF,$E7,$EC,$47,$F9,$73,$F4,$0F,$B5,$F1,$7F,$6D,$B1
	.dc.b	$F1,$F1,$F0,$1F,$7C,$AF,$F8,$FF,$EC,$7F,$C8,$AF,$F8,$F9,$B1,$FE
	.dc.b	$3C,$6B,$FC,$E5,$AF,$F8,$81,$FF,$1F,$47,$FA,$BF,$D5,$E3,$D5,$F9
	.dc.b	$9D,$55,$AB,$FF,$53,$56,$AF,$97,$C5,$7C,$8A,$FF,$21,$47,$3F,$8B
	.dc.b	$C4,$E7,$E2,$BE,$9D,$1A,$4A,$F5,$7C,$47,$E8,$30,$FF,$4A,$47,$BE
	.dc.b	$31,$F1,$79,$AD,$61,$6B,$D4,$FD,$AF,$33,$8E,$7F,$E5,$E6,$BF,$D6
	.dc.b	$DA,$FF,$90,$9A,$FF,$AB,$FC,$E7,$93,$FA,$A8,$FF,$05,$36,$33,$7F
	.dc.b	$E9,$36,$CF,$F8,$73,$BD,$A3,$4F,$67,$C5,$7B,$5B,$87,$F9,$E7,$E6
	.dc.b	$1D,$5F,$98,$6C,$7F,$9E,$9B,$8F,$F9,$E1,$AF,$F8,$7B,$07,$C5,$FC
	.dc.b	$93,$E5,$7C,$65,$BF,$BD,$F1,$7F,$37,$E3,$3E,$73,$63,$F1,$9F,$11
	.dc.b	$F5,$88,$FE,$35,$FE,$13,$83,$E3,$3E,$B1,$F5,$FC,$DB,$5F,$8B,$AF
	.dc.b	$B5,$F1,$5F,$6D,$B1,$F9,$9F,$87,$FC,$19,$1F,$CD,$3D,$BE,$2B,$F0
	.dc.b	$47,$E0,$FE,$2A,$CF,$E5,$FE,$2B,$F0,$C7,$F2,$8E,$07,$C9,$FF,$67
	.dc.b	$9D,$1F,$EE,$9F,$03,$FD,$F5,$F2,$CF,$6F,$EE,$36,$7F,$D9,$8D,$BF
	.dc.b	$E9,$ED,$7F,$CA,$EB,$DB,$F0,$3B,$BE,$B1,$8B,$FC,$D7,$0C,$C3,$F5
	.dc.b	$DA,$DC,$7F,$56,$DC,$3F,$A2,$D2,$31,$5F,$E9,$6F,$CB,$5F,$0F,$FA
	.dc.b	$5E,$CF,$E2,$F1,$89,$FD,$D4,$62,$64,$7C,$11,$8A,$7F,$C7,$7F,$19
	.dc.b	$A8,$65,$FD,$5B,$CB,$FF,$4C,$5F,$19,$E1,$18,$7F,$97,$E3,$FE,$41
	.dc.b	$9D,$61,$FE,$46,$F8,$FF,$2F,$C4,$7F,$32,$AD,$5F,$FB,$41,$AF,$F9
	.dc.b	$35,$AF,$9A,$BF,$C9,$FF,$1B,$CB,$0E,$B8,$A3,$15,$E2,$F8,$7F,$EC
	.dc.b	$F2,$F8,$DF,$8B,$18,$73,$F1,$B8,$C3,$9B,$6F,$E0,$C7,$86,$FE,$64
	.dc.b	$60,$F8,$3B,$F6,$F8,$DE,$EF,$83,$E3,$71,$7E,$66,$D7,$1D,$3F,$CD
	.dc.b	$3B,$79,$8F,$C4,$FA,$16,$FE,$75,$39,$FE,$BC,$7E,$52,$3F,$2E,$47
	.dc.b	$FC,$46,$FF,$91,$7E,$43,$F2,$6F,$E8,$E3,$F9,$B3,$F8,$09,$E6,$FD
	.dc.b	$8F,$85,$F8,$FF,$1B,$F8,$A6,$D7,$DF,$F4,$DC,$9F,$E6,$28,$C3,$6F
	.dc.b	$D9,$C1,$FE,$B5,$9C,$DF,$E5,$3C,$EF,$A7,$C9,$67,$EC,$BF,$5B,$E3
	.dc.b	$71,$87,$26,$F4,$7E,$47,$E3,$3E,$D9,$F7,$3E,$36,$E1,$F1,$BE,$5B
	.dc.b	$EA,$E1,$FE,$CA,$6B,$FE,$E8,$6B,$FA,$06,$F2,$7F,$A6,$BE,$3B,$F8
	.dc.b	$44,$7D,$EC,$7F,$41,$18,$E7,$FF,$01,$FB,$2F,$8C,$FC,$6F,$06,$1B
	.dc.b	$9F,$E5,$3E,$27,$0D,$3E,$55,$71,$6F,$C3,$FE,$5B,$E3,$BA,$B6,$3F
	.dc.b	$3B,$C7,$86,$9C,$0E,$7F,$11,$FE,$1C,$B3,$49,$1F,$E6,$6D,$FF,$D6
	.dc.b	$EF,$AA,$3F,$4A,$D3,$F6,$FE,$EF,$F5,$8D,$FF,$CA,$AF,$F1,$B8,$FD
	.dc.b	$93,$96,$F8,$FF,$E5,$48,$C7,$17,$F0,$B6,$E1,$FE,$08,$F2,$35,$ED
	.dc.b	$7C,$AF,$84,$F4,$13,$1F,$F3,$DF,$63,$54,$FF,$95,$B1,$F9,$19,$7F
	.dc.b	$8F,$27,$37,$E8,$E2,$E8,$FA,$BE,$3B,$81,$1F,$85,$68,$E9,$F1,$7E
	.dc.b	$97,$D4,$D9,$F8,$7F,$17,$ED,$E3,$F7,$36,$3E,$3F,$16,$1A,$70,$3E
	.dc.b	$5F,$C5,$7A,$48,$F8,$9F,$67,$C5,$7C,$97,$E5,$36,$3F,$80,$EB,$FC
	.dc.b	$DB,$D0,$FE,$6F,$F2,$DF,$E0,$48,$FE,$72,$7F,$A4,$E7,$3E,$F7,$FA
	.dc.b	$92,$AC,$EB,$C8,$FF,$49,$5F,$18,$E3,$3F,$D0,$BD,$1C,$47,$33,$FA
	.dc.b	$85,$F3,$69,$CF,$0E,$11,$9B,$E3,$87,$68,$C1,$E6,$AE,$93,$D9,$FE
	.dc.b	$EA,$1E,$35,$E8,$8F,$29,$99,$E5,$5D,$63,$A6,$1E,$27,$C8,$79,$8E
	.dc.b	$38,$74,$70,$7C,$B3,$A1,$C3,$0F,$24,$F8,$63,$CF,$7E,$A7,$4D,$7F
	.dc.b	$B4,$DF,$8D,$F3,$D7,$73,$9B,$D8,$DE,$1D,$5F,$83,$87,$94,$D9,$6F
	.dc.b	$B4,$78,$8F,$79,$CC,$C1,$EE,$7E,$2C,$FC,$38,$C1,$F1,$27,$AC,$EA
	.dc.b	$3E,$19,$E2,$71,$1E,$8E,$7F,$F0,$84,$F8,$1A,$FD,$46,$BF,$2D,$CF
	.dc.b	$C6,$FF,$90,$FA,$BF,$BC,$57,$E5,$A3,$15,$FE,$0F,$8C,$7B,$5F,$E0
	.dc.b	$E8,$C1,$FE,$14,$9D,$DF,$49,$D5,$7E,$66,$5E,$CF,$E8,$A3,$0D,$C7
	.dc.b	$AE,$3F,$0E,$BC,$35,$AC,$71,$7B,$7C,$18,$76,$C3,$8C,$7D,$9B,$EE
	.dc.b	$FE,$55,$AE,$67,$D5,$9F,$95,$E5,$1E,$BB,$EC,$AB,$8E,$1B,$C7,$CD
	.dc.b	$7C,$B1,$F9,$18,$7E,$FB,$1F,$46,$3D,$EF,$C5,$7B,$8E,$C7,$96,$B8
	.dc.b	$5F,$AA,$FB,$4F,$AE,$72,$F0,$3A,$13,$0E,$E4,$AF,$86,$F2,$AF,$19
	.dc.b	$23,$B1,$27,$A4,$65,$F9,$B8,$7A,$0F,$A2,$C3,$F7,$18,$FD,$6A,$E5
	.dc.b	$1F,$1F,$0F,$3B,$4E,$47,$EE,$B2,$FF,$6A,$FC,$67,$F2,$4C,$3F,$AC
	.dc.b	$F4,$DD,$7F,$AC,$B6,$FF,$AB,$55,$BF,$F6,$EC,$FF,$8A,$9B,$3F,$D2
	.dc.b	$23,$1C,$FF,$D5,$23,$FA,$5F,$C6,$71,$8F,$E5,$98,$FF,$68,$CF,$13
	.dc.b	$97,$5F,$19,$2F,$FB,$7C,$71,$F8,$CE,$5F,$17,$27,$FB,$9C,$57,$FD
	.dc.b	$95,$8A,$7C,$64,$8F,$EB,$BF,$19,$ED,$EF,$FE,$D5,$8D,$EF,$3E,$79
	.dc.b	$C3,$FE,$48,$9D,$D7,$F7,$2A,$DB,$67,$E0,$BE,$F2,$1F,$E0,$88,$AF
	.dc.b	$F9,$77,$07,$37,$C5,$33,$AB,$87,$FC,$F9,$E7,$BD,$2F,$36,$93,$1F
	.dc.b	$F4,$4C,$E5,$FF,$14,$C7,$9F,$23,$49,$1F,$E1,$F9,$7C,$67,$C1,$88
	.dc.b	$74,$6B,$FE,$73,$87,$C6,$7E,$0D,$B1,$FE,$38,$31,$F1,$7F,$37,$EB
	.dc.b	$33,$1F,$E8,$3E,$04,$63,$5F,$74,$FF,$AB,$6D,$7C,$70,$ED,$F1,$B2
	.dc.b	$BF,$CD,$D8,$71,$DB,$E2,$FC,$67,$E5,$E5,$F2,$CF,$13,$0D,$8F,$8F
	.dc.b	$38,$9F,$F4,$1C,$70,$6C,$EE,$3F,$BA,$19,$3E,$47,$C5,$7C,$12,$3F
	.dc.b	$4B,$E3,$7E,$27,$C6,$7C,$19,$93,$F5,$A2,$BF,$2F,$E3,$3E,$51,$8E
	.dc.b	$0F,$87,$C3,$F1,$BE,$33,$E5,$12,$BE,$FB,$5F,$A9,$F1,$9F,$4C,$C5
	.dc.b	$7C,$C8,$7C,$AA,$7E,$32,$F1,$E9,$88,$B3,$FE,$A6,$9D,$36,$7E,$3C
	.dc.b	$62,$3F,$C3,$1E,$17,$FA,$82,$7F,$08,$48,$FF,$5A,$CF,$F4,$2B,$FB
	.dc.b	$F5,$FE,$A6,$F1,$3F,$15,$F1,$B2,$FF,$37,$1F,$8D,$F8,$C9,$1F,$9E
	.dc.b	$74,$FE,$4F,$E3,$7F,$1D,$12,$3F,$47,$15,$FC,$C7,$C6,$FE,$5C,$C6
	.dc.b	$3E,$2E,$B7,$C6,$FE,$5C,$95,$FA,$C6,$BF,$E3,$5E,$0F,$CC,$C7,$F0
	.dc.b	$3E,$9E,$33,$FE,$74,$6B,$F1,$2B,$FC,$07,$7D,$FC,$CF,$D4,$1D,$1F
	.dc.b	$BF,$FA,$73,$E3,$BF,$8A,$36,$FF,$D9,$8D,$7F,$65,$C1,$A9,$FE,$69
	.dc.b	$E1,$66,$FE,$6B,$7B,$56,$D5,$8E,$61,$8F,$FA,$2F,$E3,$B9,$FC,$74
	.dc.b	$9E,$39,$74,$9C,$47,$FB,$D7,$25,$62,$B5,$3F,$EF,$4E,$27,$C3,$7E
	.dc.b	$D5,$7C,$35,$31,$16,$7F,$DE,$B3,$A6,$CF,$ED,$A3,$11,$F9,$0C,$1F
	.dc.b	$1D,$F0,$89,$1E,$1F,$86,$F8,$BF,$09,$23,$D7,$C1,$F1,$36,$FF,$55
	.dc.b	$4F,$CA,$F8,$EF,$A2,$FF,$3F,$E0,$34,$97,$FE,$C7,$C9,$F1,$BF,$4B
	.dc.b	$E3,$A4,$F6,$3E,$6C,$E9,$C6,$3F,$EB,$9F,$8C,$FB,$24,$8C,$CB,$E3
	.dc.b	$BF,$07,$F2,$9E,$32,$4F,$FB,$0B,$34,$7F,$73,$AF,$9F,$3C,$A3,$F0
	.dc.b	$A7,$E1,$B8,$BF,$0F,$BC,$9E,$EF,$E2,$B8,$BF,$B2,$7D,$03,$B7,$FB
	.dc.b	$03,$6B,$FD,$A2,$D7,$F4,$8D,$7F,$D5,$FA,$FF,$B5,$7E,$33,$F2,$BF
	.dc.b	$1D,$27,$DB,$1F,$CD,$FE,$5B,$D3,$1F,$CB,$83,$87,$F9,$8B,$5F,$F5
	.dc.b	$33,$5F,$CE,$F5,$BB,$7F,$07,$1F,$EE,$66,$BF,$B9,$6B,$FE,$C4,$8E
	.dc.b	$1D,$1F,$27,$E3,$BF,$C4,$1D,$6F,$AE,$7C,$26,$1B,$87,$D6,$AF,$DB
	.dc.b	$FC,$7F,$EC,$89,$1F,$B8,$7F,$77,$F1,$FF,$8B,$24,$7D,$97,$F1,$DD
	.dc.b	$DF,$C0,$09,$1F,$82,$8F,$F0,$B7,$D6,$3E,$3F,$F7,$FF,$21,$2B,$F1
	.dc.b	$DB,$2B,$51,$85,$F2,$1F,$E1,$96,$C7,$F5,$C9,$E1,$39,$35,$F2,$1E
	.dc.b	$02,$4F,$E5,$32,$FF,$21,$79,$3C,$A1,$C1,$FC,$F2,$3C,$31,$89,$FD
	.dc.b	$74,$62,$75,$7E,$CF,$8E,$F2,$12,$FC,$99,$23,$FC,$9D,$6F,$90,$F0
	.dc.b	$92,$BF,$57,$0C,$BC,$BF,$21,$F2,$3D,$A9,$5A,$8A,$FF,$3B,$C7,$FB
	.dc.b	$F3,$F8,$36,$BF,$8C,$6B,$FC,$9A,$FF,$C0,$99,$FB,$EF,$E1,$9F,$13
	.dc.b	$2B,$5E,$4F,$90,$F8,$E6,$1F,$F1,$16,$0F,$8B,$F9,$7C,$D2,$7E,$C4
	.dc.b	$57,$E5,$1E,$1F,$90,$FA,$26,$72,$FA,$1F,$21,$F5,$2A,$57,$C4,$8A
	.dc.b	$F2,$EE,$FF,$18,$FC,$6F,$DA,$33,$7F,$65,$F9,$E4,$9F,$F1,$0C,$F0
	.dc.b	$6E,$CF,$8D,$FC,$1F,$C6,$C9,$FA,$71,$5F,$04,$FF,$99,$9B,$F1,$79
	.dc.b	$C7,$FC,$2D,$F3,$0E,$DF,$EA,$75,$FE,$30,$8C,$65,$FC,$15,$B1,$99
	.dc.b	$DB,$FD,$78,$FF,$72,$7B,$AF,$98,$FE,$19,$F2,$1F,$E2,$16,$C6,$67
	.dc.b	$FC,$7B,$39,$7C,$1E,$07,$F9,$D1,$BF,$1B,$E1,$8C,$47,$71,$8A,$33
	.dc.b	$1F,$A0,$F8,$CF,$F2,$EE,$9F,$A5,$F9,$0B,$DF,$93,$F3,$2E,$2B,$F1
	.dc.b	$D5,$76,$FF,$33,$B7,$83,$FD,$39,$7F,$A0,$8C,$7A,$1F,$E3,$FB,$FD
	.dc.b	$7B,$5F,$CC,$36,$FF,$CE,$2D,$BF,$F0,$2D,$FE,$AA,$3B,$57,$48,$FD
	.dc.b	$6B,$87,$F5,$FA,$58,$FC,$67,$A3,$FE,$84,$9F,$EB,$AF,$FA,$5D,$BF
	.dc.b	$17,$EB,$8C,$4F,$08,$C1,$9F,$8D,$B5,$FC,$67,$4E,$1F,$1B,$F1,$7C
	.dc.b	$5F,$E3,$D1,$89,$D1,$30,$F1,$FC,$6E,$FB,$D5,$F3,$65,$F3,$6E,$DF
	.dc.b	$8D,$18,$6C,$67,$0F,$EA,$8D,$9F,$F4,$1F,$D6,$FF,$4F,$9F,$EF,$D1
	.dc.b	$C5,$B1,$98,$FD,$C9,$93,$F7,$7D,$2F,$AE,$FF,$18,$D7,$F8,$1E,$31
	.dc.b	$C9,$F8,$37,$F5,$DF,$05,$6B,$77,$D3,$1F,$E1,$58,$79,$BC,$1F,$49
	.dc.b	$1F,$ED,$9F,$81,$3E,$0F,$E1,$36,$3E,$BF,$C8,$E2,$7E,$47,$CA,$49
	.dc.b	$FB,$7F,$23,$72,$F0,$1F,$18,$91,$F1,$9F,$C7,$74,$3E,$43,$F0,$07
	.dc.b	$E0,$7E,$46,$D7,$F5,$5E,$BF,$AC,$7E,$17,$E4,$6C,$7C,$CF,$57,$F0
	.dc.b	$F8,$FE,$BD,$AF,$EF,$DA,$FF,$97,$9B,$D4,$F5,$3F,$1F,$E1,$7F,$AB
	.dc.b	$1B,$7F,$E4,$1D,$1F,$23,$F8,$D9,$CE,$3F,$E6,$49,$FF,$54,$57,$F6
	.dc.b	$E9,$FC,$F4,$FE,$83,$C7,$E1,$E0,$FE,$37,$E6,$FF,$A9,$5F,$F8,$6D
	.dc.b	$B5,$FE,$CA,$6B,$FC,$25,$AF,$8F,$E4,$BF,$3D,$F3,$7F,$72,$3F,$5B
	.dc.b	$18,$E8,$FD,$57,$C9,$7E,$AF,$E4,$AE,$3F,$B0,$7F,$3D,$F3,$9F,$46
	.dc.b	$3F,$C5,$8D,$C3,$FD,$B6,$D7,$F8,$D3,$5F,$89,$F2,$7F,$B7,$3F,$71
	.dc.b	$F2,$16,$BE,$17,$2F,$EE,$FE,$1B,$E4,$36,$3F,$47,$F0,$5F,$E9,$B7
	.dc.b	$FC,$78,$DC,$3F,$DD,$6D,$7F,$95,$35,$FF,$1D,$F1,$7D,$07,$FB,$0F
	.dc.b	$37,$DB,$E3,$FE,$15,$F2,$7F,$5D,$FB,$1B,$7F,$0D,$C2,$E3,$FC,$49
	.dc.b	$FE,$00,$3E,$1B,$F7,$ED,$8F,$2C,$60,$FE,$51,$87,$E2,$75,$FC,$21
	.dc.b	$FC,$67,$C6,$B5,$FE,$30,$FA,$F7,$B9,$FC,$55,$B1,$FC,$13,$E7,$7F
	.dc.b	$C3,$4E,$E7,$F8,$94,$6B,$6F,$E9,$41,$C3,$FD,$35,$AF,$F5,$D6,$BF
	.dc.b	$EB,$C5,$EF,$F4,$9E,$AD,$AF,$F4,$7F,$9D,$FE,$B6,$6F,$95,$FD,$36
	.dc.b	$31,$87,$F8,$15,$FA,$DD,$FF,$EF,$67,$2D,$F9,$4E,$1E,$3F,$17,$97
	.dc.b	$C3,$CC,$7B,$FC,$A5,$C7,$A3,$70,$E8,$D9,$FF,$59,$62,$F9,$2F,$27
	.dc.b	$CA,$79,$5B,$3E,$AF,$94,$C3,$4F,$6C,$7A,$FE,$4F,$D9,$36,$75,$97
	.dc.b	$E2,$BE,$B3,$FD,$8F,$1F,$13,$E5,$7F,$A6,$46,$27,$E1,$47,$C5,$F9
	.dc.b	$5C,$ED,$6F,$FC,$DF,$7B,$DB,$E3,$7C,$9F,$CB,$F9,$5F,$98,$DA,$FA
	.dc.b	$3D,$7F,$3C,$FA,$1F,$2B,$63,$F1,$5F,$15,$F8,$93,$FA,$B3,$FC,$50
	.dc.b	$C9,$FD,$38,$FD,$26,$BF,$E1,$16,$BF,$E2,$16,$BF,$DF,$3C,$0F,$EB
	.dc.b	$B0,$F2,$7A,$C7,$66,$C7,$E1,$BE,$53,$F0,$87,$E1,$7E,$1E,$C7,$E2
	.dc.b	$9B,$7F,$87,$6E,$A9,$FF,$19,$7D,$7F,$F8,$8D,$FC,$8F,$C3,$7F,$71
	.dc.b	$8C,$5F,$E3,$E3,$F2,$5E,$9E,$5B,$5F,$DB,$2B,$77,$F9,$3F,$23,$F3
	.dc.b	$07,$F8,$0F,$8D,$EA,$E2,$FF,$B9,$FE,$B4,$ED,$FE,$FA,$D8,$FF,$67
	.dc.b	$37,$4F,$D8,$B5,$FF,$6C,$F6,$FA,$E7,$9D,$7E,$9D,$FF,$05,$57,$1C
	.dc.b	$BE,$05,$BC,$77,$43,$97,$FC,$75,$AF,$FA,$E7,$1F,$C4,$C6,$3E,$53
	.dc.b	$3F,$29,$F9,$28,$B3,$A3,$F1,$FC,$7C,$BE,$4F,$F2,$ED,$86,$7F,$E7
	.dc.b	$DC,$7F,$D4,$D1,$C0,$CE,$9E,$6F,$27,$FB,$27,$CA,$72,$9F,$CD,$F3
	.dc.b	$7F,$20,$D8,$D2,$47,$FA,$0B,$07,$CA,$7E,$A2,$7D,$AD,$9F,$F6,$9F
	.dc.b	$CE,$FB,$F3,$F2,$7C,$AF,$F8,$E6,$31,$A7,$8F,$1F,$E4,$B5,$FC,$FA
	.dc.b	$73,$7F,$EF,$C6,$C3,$4F,$EF,$9E,$5F,$3E,$2E,$9C,$1F,$CE,$2F,$FB
	.dc.b	$5D,$FD,$36,$F1,$9C,$E1,$FE,$05,$F2,$F8,$E5,$F1,$8F,$97,$19,$7D
	.dc.b	$78,$7D,$AE,$FF,$F3,$56,$86,$92,$7E,$37,$CA,$ED,$E9,$9F,$A6,$93
	.dc.b	$98,$FB,$1F,$2B,$F8,$6F,$89,$FC,$3B,$6F,$8E,$5F,$E1,$B8,$CC,$63
	.dc.b	$0F,$F0,$37,$07,$CB,$9F,$C7,$FC,$A7,$98,$95,$F6,$5F,$4F,$CA,$6A
	.dc.b	$39,$37,$2F,$F2,$B9,$F8,$CF,$9E,$FC,$6F,$03,$E5,$BF,$D4,$1F,$09
	.dc.b	$FA,$08,$C5,$7F,$A8,$EB,$75,$AC,$2C,$E4,$7A,$FC,$1F,$CC,$E1,$FA
	.dc.b	$A8,$C4,$78,$4D,$D7,$36,$9C,$F0,$DD,$78,$E7,$FB,$5E,$1E,$0A,$EB
	.dc.b	$85,$7A,$5F,$08,$C4,$27,$57,$CE,$3D,$18,$6D,$F0,$9F,$EC,$F9,$FD
	.dc.b	$8C,$64,$F4,$C7,$33,$F2,$18,$7A,$70,$CD,$76,$C3,$C2,$DA,$D9,$ED
	.dc.b	$AE,$07,$B1,$C9,$EC,$29,$EC,$76,$FA,$1F,$91,$7E,$08,$F4,$8C,$3E
	.dc.b	$84,$F5,$C3,$DC,$6C,$AF,$86,$BD,$CD,$3A,$96,$BF,$DE,$55,$D1,$F1
	.dc.b	$18,$C3,$E3,$CF,$AF,$6F,$C1,$CF,$E1,$5A,$FC,$D3,$F8,$05,$EC,$FF
	.dc.b	$7F,$7C,$B7,$F5,$BD,$BF,$6F,$18,$F2,$11,$8A,$FE,$4B,$5B,$79,$98
	.dc.b	$9D,$B8,$11,$BA,$FC,$B2,$F6,$7F,$97,$CF,$E0,$A3,$0F,$B0,$E9,$75
	.dc.b	$E7,$7D,$AF,$9D,$F2,$6C,$EE,$FC,$18,$FA,$F0,$FC,$D4,$76,$9F,$C0
	.dc.b	$B7,$4F,$2E,$3E,$A6,$D7,$3C,$7B,$61,$F2,$1B,$87,$98,$F5,$5F,$19
	.dc.b	$D4,$78,$27,$DF,$7D,$23,$C3,$5B,$7E,$6C,$EA,$B8,$C6,$AF,$D3,$5B
	.dc.b	$8F,$9B,$5B,$78,$9A,$98,$FC,$D8,$F3,$BC,$0F,$8D,$3F,$2E,$BE,$34
	.dc.b	$7C,$49,$E8,$EC,$F1,$12,$FE,$7B,$B3,$BB,$ED,$C7,$24,$AE,$AF,$C1
	.dc.b	$8E,$24,$8F,$0B,$FC,$8B,$2F,$F6,$84,$7E,$1B,$1F,$EB,$DF,$1F,$74
	.dc.b	$FE,$9B,$C4,$8C,$65,$FD,$16,$EE,$1F,$ED,$99,$FF,$42,$CF,$F4,$96
	.dc.b	$C6,$63,$FB,$A4,$62,$9F,$27,$C9,$B3,$FD,$1E,$31,$F1,$59,$75,$F2
	.dc.b	$B2,$74,$7F,$4D,$E1,$E4,$C3,$07,$F5,$13,$C0,$43,$FC,$0C,$79,$BC
	.dc.b	$17,$CE,$FB,$BB,$7F,$B9,$78,$37,$87,$FC,$17,$7F,$D7,$AF,$6F,$F8
	.dc.b	$2A,$7E,$0D,$70,$8F,$F5,$3B,$5F,$D9,$9F,$DF,$E3,$75,$FE,$A4,$8F
	.dc.b	$F0,$9F,$77,$73,$E6,$DB,$E5,$71,$3E,$1F,$89,$C9,$87,$F8,$5F,$1F
	.dc.b	$F8,$12,$9F,$2B,$E9,$EF,$95,$E9,$8A,$F9,$34,$FF,$08,$4F,$90,$91
	.dc.b	$FE,$C0,$E4,$FF,$23,$FC,$AF,$C3,$EF,$93,$F0,$E2,$BE,$23,$FC,$45
	.dc.b	$8B,$E4,$3D,$66,$BA,$7F,$C7,$D0,$F9,$0F,$9D,$D5,$2F,$E7,$45,$7A
	.dc.b	$65,$FE,$58,$F4,$3E,$29,$9C,$39,$FC,$87,$E5,$A9,$F5,$DF,$E2,$17
	.dc.b	$FC,$3A,$63,$C6,$F9,$B3,$B9,$D4,$FF,$B7,$9A,$FD,$86,$BF,$CB,$1A
	.dc.b	$FB,$BE,$23,$EB,$91,$FF,$15,$F0,$BF,$2B,$FF,$78,$C7,$E2,$7E,$4B
	.dc.b	$CD,$F6,$0F,$22,$9F,$80,$8B,$3F,$8F,$F6,$FF,$06,$61,$F8,$7C,$5F
	.dc.b	$EA,$77,$F2,$9E,$37,$E2,$0F,$C4,$F8,$F6,$32,$43,$8B,$F0,$5F,$91
	.dc.b	$F3,$FF,$9C,$7E,$D0,$71,$C3,$FE,$8A,$8F,$9A,$70,$F0,$7F,$80,$BB
	.dc.b	$6C,$F3,$8F,$D0,$B5,$FF,$41,$EC,$FB,$0E,$97,$FE,$96,$8C,$76,$FE
	.dc.b	$CE,$72,$FF,$5C,$BF,$D2,$5F,$EC,$FE,$33,$F4,$E4,$7F,$6D,$F2,$3F
	.dc.b	$AA,$F8,$FF,$D5,$FC,$84,$9F,$DD,$47,$EB,$BE,$43,$F8,$07,$D5,$7E
	.dc.b	$A2,$7E,$CB,$5F,$9F,$1A,$F0,$FF,$6B,$E7,$C9,$DC,$7E,$E3,$BB,$F9
	.dc.b	$2F,$90,$F8,$FA,$66,$BF,$CC,$F1,$96,$FC,$77,$03,$6E,$0D,$FC,$7F
	.dc.b	$43,$4F,$2F,$83,$EF,$F0,$52,$30,$76,$70,$FE,$AA,$FF,$DA,$0E,$8C
	.dc.b	$7C,$67,$A0,$8F,$9B,$E3,$EB,$C7,$E3,$FD,$A6,$3A,$7F,$09,$CA,$AF
	.dc.b	$D9,$29,$A6,$22,$D7,$F4,$89,$D3,$67,$F1,$71,$88,$FE,$2D,$F6,$0F
	.dc.b	$85,$FE,$6C,$D8,$FD,$7B,$5F,$CC,$CB,$1F,$BB,$E0,$49,$FB,$AF,$D6
	.dc.b	$F9,$0F,$E1,$1F,$81,$BD,$AF,$CD,$B6,$BF,$AE,$B5,$FE,$76,$1C,$3F
	.dc.b	$E3,$0E,$F7,$4F,$F8,$8F,$7F,$F6,$E7,$03,$E4,$7F,$0A,$7E,$43,$C3
	.dc.b	$FC,$8E,$D3,$1F,$C8,$BF,$8A,$F9,$1F,$E3,$9F,$60,$F5,$3F,$5E,$DA
	.dc.b	$FE,$E2,$D7,$CA,$D7,$C9,$9B,$BF,$F2,$BF,$25,$27,$F3,$F1,$F9,$8F
	.dc.b	$8B,$91,$F9,$67,$F3,$7F,$25,$FC,$B7,$E4,$3F,$49,$3F,$D7,$1A,$FF
	.dc.b	$94,$1A,$FF,$C0,$06,$F8,$1F,$ED,$DF,$91,$FD,$2F,$C9,$4A,$FD,$1C
	.dc.b	$7E,$A3,$E3,$FF,$B9,$7D,$1B,$C7,$A6,$6B,$F9,$DF,$D1,$B8,$BF,$88
	.dc.b	$B6,$3F,$C2,$6D,$7E,$BB,$5F,$F3,$2E,$5F,$EE,$FC,$FF,$C0,$1E,$43
	.dc.b	$C4,$A7,$EE,$A3,$3E,$1C,$BF,$E7,$B5,$F9,$F7,$F7,$7F,$21,$FD,$47
	.dc.b	$EB,$DE,$0F,$F2,$46,$CF,$F8,$99,$AF,$E2,$9A,$FE,$26,$BF,$30,$F6
	.dc.b	$F0,$BC,$A7,$F0,$CF,$07,$FC,$63,$D0,$F8,$FA,$4A,$FE,$1B,$E7,$F8
	.dc.b	$FD,$2F,$11,$23,$C4,$FF,$86,$31,$FC,$A4,$7B,$3C,$0F,$86,$76,$F1
	.dc.b	$FF,$B6,$7B,$7F,$E2,$29,$FE,$8A,$DF,$13,$FA,$23,$71,$4B,$93,$F4
	.dc.b	$7F,$1F,$FD,$D7,$BD,$E6,$7F,$48,$6D,$7F,$92,$9A,$FE,$B9,$AF,$FA
	.dc.b	$73,$E8,$1F,$5F,$F5,$E3,$BE,$6C,$BC,$35,$AE,$9F,$F2,$CB,$6B,$73
	.dc.b	$AB,$FE,$47,$1F,$E1,$99,$7C,$E3,$AB,$D1,$E8,$4A,$E2,$2B,$73,$98
	.dc.b	$F4,$4F,$B4,$8F,$F9,$13,$89,$E1,$FE,$40,$91,$D3,$83,$F8,$4F,$CA
	.dc.b	$E2,$3F,$C4,$3C,$FF,$35,$F9,$8D,$F3,$18,$74,$9E,$53,$FE,$4D,$78
	.dc.b	$FC,$4F,$E2,$0F,$C4,$FC,$AD,$8D,$FC,$AF,$E3,$0C,$3E,$0F,$89,$FC
	.dc.b	$79,$1E,$BD,$15,$F2,$FC,$AF,$E5,$0C,$73,$7D,$38,$F3,$C7,$FA,$13
	.dc.b	$23,$49,$1F,$E4,$F8,$7C,$AF,$E7,$89,$1E,$CF,$9C,$AF,$C4,$F4,$3F
	.dc.b	$46,$63,$0F,$F0,$B7,$D2,$3C,$CF,$C9,$57,$F3,$88,$C7,$1F,$FA,$61
	.dc.b	$B1,$99,$DB,$FE,$78,$F9,$6F,$DD,$F4,$7F,$A7,$9B,$1F,$D0,$9A,$EF
	.dc.b	$C7,$FC,$23,$DF,$A5,$F2,$5D,$DB,$19,$8C,$1F,$07,$E5,$BE,$46,$9F
	.dc.b	$27,$4F,$E9,$7F,$0D,$8E,$FF,$D6,$C7,$F4,$EA,$C6,$1F,$84,$AB,$39
	.dc.b	$AF,$C1,$D6,$38,$9F,$1D,$63,$1F,$2D,$FE,$3D,$C7,$AD,$23,$11,$F9
	.dc.b	$06,$E1,$FD,$79,$B3,$FE,$BC,$C3,$F2,$31,$DF,$E5,$BA,$C7,$E5,$1C
	.dc.b	$3F,$96,$F9,$6B,$87,$F9,$FB,$B7,$FD,$67,$3F,$9A,$DB,$D7,$18,$9F
	.dc.b	$D1,$C6,$3E,$22,$57,$E8,$26,$9C,$FE,$13,$8F,$CA,$F2,$6C,$FE,$BF
	.dc.b	$0F,$F5,$3B,$63,$BF,$CA,$EE,$3F,$9F,$9F,$0B,$C2,$FD,$7B,$6B,$F8
	.dc.b	$83,$FB,$D3,$E2,$AF,$95,$FA,$B3,$C2,$7C,$11,$89,$E3,$8F,$C1,$F9
	.dc.b	$4F,$7F,$CA,$FC,$06,$DF,$FA,$33,$C9,$79,$DE,$36,$C7,$FB,$41,$B7
	.dc.b	$FE,$BD,$6B,$FD,$83,$07,$CA,$7C,$C3,$E6,$FC,$AD,$BF,$89,$F2,$3F
	.dc.b	$43,$E5,$3E,$8B,$63,$C5,$F2,$DE,$12,$3E,$17,$E0,$FA,$DF,$60,$FB
	.dc.b	$3E,$05,$9F,$57,$9B,$F7,$3E,$1F,$E2,$EC,$8C,$4F,$BF,$CB,$F7,$11
	.dc.b	$F7,$3F,$07,$CB,$E2,$7E,$1B,$E4,$2C,$FF,$B0,$3E,$7B,$E0,$F1,$7F
	.dc.b	$54,$0E,$1F,$EA,$ED,$73,$C7,$F3,$7C,$C9,$5F,$35,$B3,$FE,$FC,$F0
	.dc.b	$BF,$98,$FC,$8F,$E5,$8F,$CB,$FC,$55,$BF,$C7,$7C,$27,$E6,$BD,$9F
	.dc.b	$CD,$B6,$3E,$E7,$8F,$F6,$C8,$FD,$B7,$F1,$78,$7E,$88,$FD,$1F,$55
	.dc.b	$9F,$E9,$1C,$5F,$DB,$1B,$F2,$3F,$C0,$A3,$1A,$7F,$BA,$9F,$F4,$A4
	.dc.b	$FE,$2E,$3F,$14,$47,$F1,$4F,$EF,$A3,$FC,$49,$1D,$BD,$AE,$9E,$CF
	.dc.b	$57,$B7,$C1,$F7,$F9,$1B,$5E,$47,$F4,$3F,$80,$FE,$8E,$3F,$CD,$3E
	.dc.b	$22,$31,$97,$D5,$11,$9F,$13,$F6,$13,$A9,$FF,$1C,$E7,$FE,$7D,$BF
	.dc.b	$83,$F0,$7E,$FF,$92,$F8,$0F,$F2,$4A,$7C,$EB,$9F,$E2,$7C,$96,$63
	.dc.b	$18,$7C,$A8,$C7,$B5,$FC,$82,$F8,$B6,$BF,$CE,$55,$FE,$82,$9C,$B6
	.dc.b	$3F,$AC,$BF,$E0,$68,$C9,$FE,$EC,$8C,$B6,$7E,$B7,$C7,$FD,$7F,$85
	.dc.b	$B7,$F7,$1B,$5F,$66,$7E,$E7,$C5,$E3,$D8,$FE,$8D,$B3,$E3,$BF,$01
	.dc.b	$F0,$9F,$81,$6E,$5F,$86,$70,$FE,$23,$E3,$EE,$5F,$85,$F8,$FF,$C4
	.dc.b	$1F,$89,$F8,$0B,$5F,$8E,$70,$FE,$43,$E3,$6D,$7E,$35,$FC,$81,$F9
	.dc.b	$1F,$7E,$CF,$F8,$A3,$A3,$F0,$1E,$1F,$E5,$7E,$33,$F0,$AD,$9F,$CD
	.dc.b	$98,$29,$C7,$A5,$1F,$D7,$9F,$F3,$14,$7F,$8A,$5B,$8B,$83,$FB,$54
	.dc.b	$7F,$B2,$67,$5F,$19,$C7,$C7,$E5,$F1,$96,$79,$3F,$89,$F9,$BF,$D4
	.dc.b	$E1,$CB,$E2,$F1,$F1,$BE,$17,$5E,$27,$E1,$E8,$D2,$61,$F8,$78,$FD
	.dc.b	$5E,$9E,$58,$FE,$29,$1C,$1F,$C4,$3A,$C3,$FC,$55,$28,$D3,$C4,$FF
	.dc.b	$37,$68,$F8,$BF,$89,$EA,$FC,$56,$E1,$F1,$5C,$3F,$1F,$E3,$2D,$F3
	.dc.b	$E2,$7E,$15,$F0,$2B,$FC,$1F,$8F,$F3,$57,$ED,$46,$6B,$C3,$EA,$7D
	.dc.b	$3F,$53,$FA,$FC,$E7,$CD,$FD,$F4,$63,$87,$EC,$7A,$9F,$67,$E3,$6E
	.dc.b	$5F,$65,$FD,$C7,$D5,$BC,$AF,$D9,$36,$7F,$60,$DD,$3F,$DE,$6D,$7F
	.dc.b	$D7,$BA,$FE,$F1,$F8,$C2,$38,$71,$FE,$13,$C9,$EF,$87,$A0,$FC,$67
	.dc.b	$8F,$71,$FC,$63,$F8,$FF,$8E,$C6,$3F,$47,$7F,$E3,$51,$8C,$7F,$CC
	.dc.b	$1C,$5F,$2B,$C8,$FC,$B7,$C5,$FC,$46,$E5,$F9,$7F,$8B,$C3,$4F,$9B
	.dc.b	$1F,$2B,$CA,$FF,$08,$72,$9A,$49,$FE,$51,$A7,$E0,$BD,$4E,$11,$FA
	.dc.b	$36,$C7,$9F,$8B,$FC,$6C,$DF,$8A,$FD,$57,$C5,$7E,$AD,$B7,$FA,$B8
	.dc.b	$FD,$77,$C5,$62,$BF,$0D,$93,$C9,$F2,$92,$FF,$9E,$C6,$5F,$2F,$17
	.dc.b	$F3,$FF,$8A,$FC,$2D,$7E,$E5,$B3,$FB,$9B,$FF,$26,$E7,$FE,$CA,$8E
	.dc.b	$2D,$8F,$EE,$0D,$7F,$DF,$78,$7F,$91,$78,$BF,$7F,$1F,$E6,$28,$C5
	.dc.b	$7E,$6A,$31,$F1,$19,$F8,$9B,$3F,$EB,$1A,$46,$2B,$EF,$47,$F9,$19
	.dc.b	$BF,$13,$9C,$3F,$CE,$D3,$FD,$56,$7F,$CE,$AD,$7F,$DC,$4D,$D3,$F5
	.dc.b	$BE,$B7,$F9,$AA,$1F,$13,$E7,$8F,$43,$72,$FE,$95,$7F,$EE,$CF,$9D
	.dc.b	$FF,$76,$FC,$AF,$F9,$EA,$3F,$C3,$9B,$FF,$AB,$E3,$11,$FE,$1F,$9D
	.dc.b	$F3,$FF,$6D,$AD,$3F,$88,$8F,$F7,$BF,$BD,$C2,$BF,$D2,$F8,$62,$B5
	.dc.b	$1F,$EC,$E7,$94,$66,$3F,$BC,$57,$67,$C7,$15,$F1,$0C,$3B,$61,$E2
	.dc.b	$8E,$26,$6F,$AB,$4C,$CF,$0A,$E9,$1F,$0C,$F2,$1E,$02,$9E,$0C,$38
	.dc.b	$38,$38,$4E,$0F,$80,$B0,$EB,$87,$96,$3D,$33,$BC,$3E,$1B,$63,$CB
	.dc.b	$1E,$88,$E7,$5C,$B0,$F3,$D7,$53,$C5,$7F,$EE,$5C,$37,$18,$9F,$98
	.dc.b	$F8,$30,$E8,$78,$0A,$FC,$89,$F6,$9C,$8B,$1F,$61,$B1,$B8,$F9,$B1
	.dc.b	$E0,$D0,$FD,$E6,$CF,$D4,$6B,$F6,$9B,$F1,$CF,$87,$FE,$DD,$29,$CC
	.dc.b	$61,$BB,$FF,$B6,$2F,$FC,$E5,$7C,$63,$16,$11,$97,$ED,$AF,$86,$FF
	.dc.b	$3D,$57,$E7,$A3,$13,$F2,$1E,$11,$9C,$3C,$AE,$EB,$8E,$3E,$26,$DF
	.dc.b	$86,$F8,$CF,$9A,$FD,$BA,$F1,$AF,$CE,$57,$81,$FC,$B1,$E8,$9D,$D7
	.dc.b	$1C,$7C,$98,$76,$AF,$6B,$6F,$BC,$78,$6F,$8B,$E3,$9F,$F4,$C6,$1E
	.dc.b	$18,$EF,$5F,$16,$3D,$D7,$E4,$AD,$E1,$ED,$BF,$3E,$1D,$1F,$35,$6D
	.dc.b	$87,$48,$D6,$1E,$48,$F3,$9E,$A3,$E7,$47,$CC,$AF,$74,$79,$2F,$C2
	.dc.b	$FA,$4F,$0D,$C7,$47,$83,$0F,$4B,$E1,$8E,$CF,$E0,$27,$E3,$5F,$78
	.dc.b	$F8,$D3,$C6,$FD,$D1,$C4,$FE,$11,$97,$F8,$6B,$E3,$3F,$D8,$15,$FC
	.dc.b	$DB,$E2,$EE,$BF,$CD,$9B,$7F,$CD,$AA,$DF,$F8,$8A,$7F,$CF,$CD,$9F
	.dc.b	$EB,$7C,$49,$CB,$63,$3F,$15,$CA,$0D,$31,$F1,$96,$7F,$A8,$7C,$66
	.dc.b	$CC,$3F,$D4,$3E,$33,$96,$D2,$3F,$AB,$45,$7F,$DA,$58,$7E,$CD,$F2
	.dc.b	$CF,$90,$F2,$EF,$EE,$F0,$FF,$C7,$BE,$CD,$B3,$E7,$9C,$3F,$E9,$D9
	.dc.b	$DD,$FF,$7C,$8C,$57,$08,$F8,$7F,$15,$EF,$23,$FD,$AF,$E3,$2B,$FD
	.dc.b	$AB,$DA,$F8,$86,$38,$FF,$B4,$FC,$5F,$97,$A5,$ED,$D3,$E5,$45,$BF
	.dc.b	$F6,$4C,$E9,$B3,$E7,$8C,$47,$C4,$F8,$BE,$9E,$D4,$8F,$EF,$D1,$5F
	.dc.b	$EF,$BD,$9F,$5C,$C6,$3F,$DF,$39,$FC,$9E,$BC,$AF,$F0,$34,$57,$FC
	.dc.b	$85,$8B,$E2,$FF,$12,$D8,$FF,$50,$3F,$E7,$57,$F8,$F5,$BA,$FF,$90
	.dc.b	$5A,$9E,$AF,$E5,$2D,$F5,$5C,$5F,$F2,$DB,$70,$E5,$1F,$CC,$6B,$8B
	.dc.b	$FE,$D1,$9E,$0F,$FB,$B5,$AF,$F2,$F3,$AE,$5F,$ED,$B6,$9F,$E3,$89
	.dc.b	$7C,$67,$B3,$DD,$93,$EA,$CF,$D3,$5A,$8C,$1E,$3F,$8B,$FB,$9E,$B4
	.dc.b	$8C,$EB,$EF,$9C,$47,$CA,$F8,$CF,$92,$63,$5F,$6D,$7C,$5F,$8B,$F9
	.dc.b	$A4,$AE,$78,$FF,$A1,$DB,$1F,$4B,$E3,$24,$60,$FF,$7A,$7C,$F3,$DC
	.dc.b	$F9,$CB,$DD,$A6,$22,$CF,$FB,$FE,$74,$D9,$F9,$11,$88,$FF,$39,$AD
	.dc.b	$7F,$07,$EC,$49,$F0,$C7,$E1,$BE,$33,$F0,$FE,$CC,$9F,$5C,$7E,$2B
	.dc.b	$E2,$E4,$7B,$9D,$3F,$90,$F8,$AF,$D4,$7B,$12,$3E,$1C,$57,$F2,$BF
	.dc.b	$17,$F9,$43,$1C,$5F,$33,$85,$F1,$7F,$98,$25,$7C,$CF,$6B,$F3,$7E
	.dc.b	$3C,$8F,$9E,$E9,$FD,$0F,$C5,$FE,$80,$8E,$8F,$BF,$F2,$DF,$B4,$C7
	.dc.b	$ED,$B7,$AB,$F4,$2D,$9F,$CD,$7C,$67,$F3,$28,$C7,$4F,$F9,$52,$7E
	.dc.b	$FD,$67,$C5,$D7,$8B,$27,$BB,$C7,$E2,$F1,$1F,$EB,$7C,$1F,$19,$D4
	.dc.b	$CE,$7F,$EE,$EF,$9A,$FF,$0D,$57,$E6,$87,$95,$FB,$B7,$FD,$96,$D9
	.dc.b	$E1,$A7,$ED,$E3,$53,$FA,$D3,$8F,$8F,$FA,$D8,$CC,$7E,$C5,$AF,$F8
	.dc.b	$85,$FD,$CB,$E7,$F8,$CF,$84,$75,$F2,$3E,$27,$8F,$27,$E6,$BF,$1F
	.dc.b	$E3,$3F,$9F,$6B,$FE,$8C,$9F,$F6,$2F,$A0,$8C,$47,$96,$31,$EE,$FF
	.dc.b	$3F,$9C,$CF,$F0,$5D,$38,$7C,$5F,$D3,$F8,$C9,$3F,$72,$3E,$B7,$8B
	.dc.b	$23,$EF,$BF,$67,$C6,$F1,$FB,$3D,$49,$3E,$4C,$7F,$B2,$BF,$EC,$65
	.dc.b	$F1,$5F,$81,$F1,$A4,$FE,$2A,$3F,$09,$F1,$32,$3F,$1A,$FE,$1F,$E2
	.dc.b	$BE,$0F,$27,$AC,$93,$DD,$FE,$F7,$83,$E2,$BF,$1D,$F0,$12,$BF,$2F
	.dc.b	$1F,$E0,$4E,$87,$9E,$F3,$F0,$DA,$FE,$3F,$3A,$9F,$CC,$7C,$57,$D2
	.dc.b	$F5,$BE,$61,$2B,$E0,$3F,$CB,$FE,$5D,$E7,$52,$4F,$FB,$F0,$46,$9F
	.dc.b	$91,$CF,$F7,$88,$FE,$BB,$CD,$FE,$FD,$F0,$BF,$8C,$EB,$FF,$4E,$78
	.dc.b	$1F,$DE,$E3,$04,$F5,$7F,$5D,$F1,$73,$0F,$E7,$37,$FE,$A3,$9F,$D9
	.dc.b	$44,$3F,$76,$7E,$D7,$CE,$EF,$EB,$75,$24,$7E,$43,$04,$7F,$8A,$9F
	.dc.b	$F7,$6C,$7F,$42,$5F,$25,$FE,$30,$6C,$67,$E2,$B1,$F1,$72,$38,$3B
	.dc.b	$F8,$BF,$7F,$27,$B4,$93,$E1,$7C,$1E,$27,$5F,$8B,$93,$E1,$8F,$07
	.dc.b	$C4,$FC,$8E,$59,$3E,$58,$AF,$E9,$F3,$FA,$4D,$9F,$98,$7B,$7E,$1F
	.dc.b	$D4,$48,$F7,$7C,$4D,$5D,$5F,$D7,$2D,$F1,$3F,$04,$D4,$67,$E2,$3E
	.dc.b	$21,$2B,$E4,$78,$D5,$E7,$EA,$7F,$82,$B0,$7C,$4F,$CB,$35,$87,$7F
	.dc.b	$86,$F9,$C4,$BF,$A5,$F1,$35,$E6,$7F,$AE,$EF,$F9,$0B,$F5,$FC,$6D
	.dc.b	$5F,$A5,$FB,$84,$8F,$E4,$07,$17,$F0,$1E,$0F,$DD,$24,$7E,$07,$E2
	.dc.b	$2B,$E0,$F0,$7F,$DE,$5F,$11,$F8,$33,$38,$7F,$87,$A5,$F2,$BF,$D6
	.dc.b	$2B,$F4,$CD,$8F,$0B,$5F,$E6,$4D,$7F,$DA,$3A,$BE,$57,$F5,$DF,$1B
	.dc.b	$F2,$1B,$19,$9F,$5C,$E5,$EB,$EF,$FF,$8D,$DB,$E0,$F8,$A3,$11,$E0
	.dc.b	$6B,$F5,$71,$FF,$1B,$E5,$F8,$1F,$8B,$BC,$79,$3F,$48,$E2,$BF,$DD
	.dc.b	$9F,$27,$FE,$4D,$9F,$D4,$5E,$44,$66,$7E,$57,$17,$FB,$CE,$BF,$C5
	.dc.b	$F3,$FE,$4C,$7E,$AD,$70,$F6,$7E,$65,$FF,$94,$AB,$4F,$F5,$99,$D5
	.dc.b	$70,$97,$8B,$E1,$77,$97,$F8,$D9,$CC,$F8,$30,$D7,$C5,$EE,$EC,$F9
	.dc.b	$1E,$5E,$6F,$3F,$1E,$CF,$A1,$F8,$5E,$43,$5F,$F0,$BC,$FB,$2B,$31
	.dc.b	$8D,$7F,$15,$9F,$F3,$8A,$FE,$37,$FA,$D2,$7D,$1F,$23,$D2,$31,$3F
	.dc.b	$1A,$31,$32,$3E,$4F,$C6,$D8,$FD,$2E,$5E,$3F,$84,$F9,$30,$8C,$57
	.dc.b	$F2,$19,$CC,$62,$7E,$66,$3F,$D2,$EB,$F9,$7E,$4B,$F9,$9E,$47,$B1
	.dc.b	$F9,$DE,$2C,$AF,$B7,$F1,$97,$1F,$1E,$3F,$EA,$89,$FE,$FE,$D8,$FE
	.dc.b	$FC,$D7,$FD,$01,$97,$E2,$3E,$17,$D3,$CB,$F8,$AF,$8B,$FB,$9F,$19
	.dc.b	$F7,$5B,$8F,$F4,$4C,$5F,$0D,$F5,$1B,$1F,$8E,$F8,$CC,$4E,$7E,$1F
	.dc.b	$E0,$D7,$F8,$C2,$BC,$B1,$83,$E4,$F5,$3F,$71,$FF,$3A,$4F,$E6,$3E
	.dc.b	$37,$F4,$53,$9F,$82,$FC,$84,$78,$7E,$22,$57,$E7,$BE,$32,$F1,$7F
	.dc.b	$1A,$C1,$F5,$3F,$E9,$AA,$FE,$2B,$18,$EF,$F0,$CE,$6B,$FC,$71,$39
	.dc.b	$AF,$EE,$5F,$2F,$E9,$9F,$E3,$AD,$7F,$C7,$CD,$7F,$D5,$8D,$7F,$73
	.dc.b	$E1,$FE,$9C,$FD,$47,$C7,$DA,$F9,$9F,$15,$F1,$88,$FD,$A7,$E4,$FC
	.dc.b	$57,$CF,$3F,$61,$F1,$F6,$7F,$7A,$7E,$CC,$FD,$A7,$6D,$7E,$C7,$C5
	.dc.b	$7D,$22,$3F,$86,$7F,$02,$FD,$AC,$BF,$99,$3F,$89,$D3,$FD,$62,$E5
	.dc.b	$BD,$7F,$81,$F8,$9F,$E2,$53,$F1,$67,$FD,$A7,$DB,$FC,$52,$73,$CF
	.dc.b	$F8,$7F,$87,$FC,$19,$2F,$F2,$EF,$E1,$BE,$1F,$F1,$87,$4F,$8F,$B3
	.dc.b	$F9,$1F,$86,$EF,$D1,$F9,$16,$C7,$E5,$7E,$17,$F1,$E4,$7F,$46,$FE
	.dc.b	$4B,$CF,$FC,$C1,$E8,$F8,$DB,$3F,$ED,$41,$CB,$EC,$3D,$BF,$1D,$63
	.dc.b	$FC,$4A,$7F,$86,$0F,$F3,$69,$78,$BF,$CD,$8D,$7F,$5E,$D7,$FD,$FE
	.dc.b	$B9,$BF,$8E,$BC,$78,$BE,$7C,$63,$E1,$FE,$6F,$7C,$AF,$37,$C7,$5C
	.dc.b	$3E,$BB,$70,$FA,$ED,$9F,$E4,$1C,$A5,$F0,$FF,$63,$A3,$C2,$D8,$F6
	.dc.b	$B8,$7E,$27,$C7,$5C,$7F,$0B,$F1,$DF,$80,$3D,$3F,$0F,$6B,$E2,$BF
	.dc.b	$C8,$13,$48,$FF,$52,$1D,$5C,$9F,$A4,$DA,$FC,$9F,$47,$E3,$BA,$7F
	.dc.b	$1E,$DA,$FB,$1F,$1B,$86,$9F,$6E,$3F,$2B,$CF,$F9,$6E,$8B,$3F,$7D
	.dc.b	$FC,$CE,$12,$3F,$05,$D1,$63,$FC,$29,$F2,$DF,$56,$7F,$A5,$79,$A8
	.dc.b	$C4,$FF,$4D,$F5,$1D,$9F,$DD,$67,$32,$97,$67,$F4,$88,$C3,$63,$38
	.dc.b	$7F,$A0,$5B,$3F,$E2,$C5,$D7,$FE,$05,$8E,$33,$FE,$D7,$9C,$B6,$BF
	.dc.b	$94,$F8,$AF,$0B,$C9,$F1,$98,$6D,$7F,$A6,$A7,$4D,$8F,$57,$C6,$6E
	.dc.b	$3F,$B5,$76,$7E,$2F,$26,$DF,$0F,$E3,$3E,$07,$57,$C1,$6D,$F0,$3E
	.dc.b	$4E,$FF,$84,$71,$7C,$A1,$3F,$27,$A3,$95,$6B,$0F,$A1,$F1,$9F,$34
	.dc.b	$F8,$5F,$B0,$B4,$C5,$7F,$BE,$7E,$5D,$EC,$FC,$F6,$C7,$FA,$B5,$BA
	.dc.b	$7F,$96,$1A,$FF,$BE,$6F,$F4,$F3,$CE,$BB,$EB,$F4,$6F,$8E,$DF,$0F
	.dc.b	$3F,$C2,$7C,$67,$E0,$7E,$37,$F0,$4D,$C7,$D5,$C0,$F8,$EC,$C6,$1B
	.dc.b	$AF,$E4,$3E,$13,$F1,$5F,$1B,$F8,$B6,$D7,$2F,$8E,$C3,$4E,$B1,$FC
	.dc.b	$33,$E3,$BF,$64,$48,$FE,$2E,$FF,$80,$31,$7C,$76,$E7,$CA,$DC,$7F
	.dc.b	$C3,$FA,$FF,$93,$76,$F7,$5B,$0F,$F1,$AC,$E7,$0F,$BB,$18,$F8,$F9
	.dc.b	$1F,$03,$E3,$EE,$3F,$E2,$A8,$7C,$6F,$F0,$72,$3F,$CB,$1F,$37,$C7
	.dc.b	$7F,$26,$7F,$58,$D8,$F9,$71,$EB,$23,$FE,$5E,$CB,$AB,$7E,$3F,$F6
	.dc.b	$93,$FB,$56,$C7,$D4,$CF,$F7,$1F,$19,$FB,$96,$C7,$E4,$7E,$3B,$FD
	.dc.b	$6E,$FF,$9A,$9B,$97,$F9,$9D,$AF,$F9,$3C,$FD,$2B,$87,$F8,$4F,$85
	.dc.b	$FE,$77,$75,$5F,$C1,$75,$FF,$19,$CE,$67,$71,$8F,$8D,$FF,$40,$4F
	.dc.b	$CF,$9F,$D8,$BF,$E7,$2C,$36,$7E,$B2,$3C,$5F,$13,$FE,$61,$F9,$67
	.dc.b	$9B,$E5,$97,$93,$FB,$B9,$CC,$66,$B8,$56,$EB,$4F,$FA,$FA,$3F,$CD
	.dc.b	$9A,$F0,$AF,$F3,$75,$7F,$05,$8C,$47,$F9,$C6,$77,$5C,$F0,$B3,$DC
	.dc.b	$61,$C3,$0E,$F1,$98,$C4,$F9,$23,$15,$CD,$E2,$F9,$5A,$7D,$3A,$E6
	.dc.b	$FC,$41,$87,$46,$CF,$18,$C1,$EA,$8F,$F4,$91,$D0,$F7,$D7,$8D,$E9
	.dc.b	$7C,$CE,$18,$66,$3E,$7C,$D7,$9C,$EE,$FD,$0D,$8F,$69,$E9,$9E,$15
	.dc.b	$F2,$23,$C2,$7A,$0F,$95,$7E,$8B,$E6,$7C,$5B,$F1,$CF,$A4,$F1,$94
	.dc.b	$F6,$CF,$17,$9C,$FC,$D8,$E2,$66,$3E,$53,$5F,$BC,$7F,$44,$7E,$96
	.dc.b	$9F,$3A,$3E,$F3,$5F,$A4,$DF,$8F,$74,$39,$BF,$B7,$DF,$EA,$E7,$31
	.dc.b	$8B,$FF,$5E,$46,$1F,$01,$8A,$E4,$E0,$46,$EB,$F3,$0F,$F2,$28,$F9
	.dc.b	$8D,$7E,$61,$C2,$79,$CF,$06,$E1,$E5,$E7,$FE,$01,$3D,$1B,$1E,$99
	.dc.b	$DD,$79,$70,$FC,$46,$1D,$F2,$E1,$38,$AF,$4C,$77,$AF,$2E,$1D,$2B
	.dc.b	$E0,$B5,$F3,$47,$AA,$B8,$E1,$E5,$9E,$F5,$FC,$56,$FA,$47,$E6,$F0
	.dc.b	$F9,$51,$E2,$8C,$4E,$EB,$D5,$87,$94,$E7,$30,$E5,$87,$81,$F7,$1F
	.dc.b	$2B,$0F,$99,$8F,$53,$C2,$78,$B1,$F0,$13,$0E,$F5,$D3,$0E,$D3,$98
	.dc.b	$EA,$F2,$9F,$74,$77,$C7,$E2,$61,$F7,$E7,$9D,$71,$AF,$C1,$D7,$B5
	.dc.b	$E7,$3F,$CD,$BE,$3F,$F9,$D6,$1F,$DF,$77,$7C,$0E,$37,$FE,$76,$DC
	.dc.b	$3F,$99,$DD,$DB,$FC,$81,$3F,$CB,$E7,$65,$8C,$B6,$FA,$BF,$D3,$79
	.dc.b	$78,$3F,$D2,$9B,$2E,$4F,$ED,$33,$B3,$97,$C7,$78,$6D,$F1,$F4,$C4
	.dc.b	$67,$E3,$F5,$E2,$4B,$FF,$17,$5A,$71,$1F,$D1,$F1,$FF,$2D,$57,$8A
	.dc.b	$BF,$C5,$B3,$E9,$8C,$7C,$7F,$C3,$F6,$7F,$D2,$1C,$B7,$83,$F8,$07
	.dc.b	$0F,$CB,$AC,$4F,$BA,$B7,$87,$FA,$B2,$31,$7F,$C1,$E3,$FB,$64,$FF
	.dc.b	$94,$38,$3B,$67,$F2,$27,$FC,$2F,$1F,$07,$E3,$B2,$FC,$3F,$8F,$93
	.dc.b	$FE,$50,$BE,$4F,$C0,$F8,$FF,$47,$7B,$B6,$9F,$2E,$2C,$FD,$BA,$FF
	.dc.b	$41,$4F,$A6,$B8,$4E,$5A,$FC,$9E,$07,$C7,$FD,$5F,$8B,$93,$F4,$22
	.dc.b	$BF,$C8,$76,$FF,$03,$FC,$7F,$DB,$33,$C5,$F9,$3F,$8C,$FB,$DF,$17
	.dc.b	$2F,$EC,$C5,$79,$1F,$E0,$6F,$8B,$FC,$21,$9A,$FF,$3D,$43,$EE,$3C
	.dc.b	$F1,$8F,$C5,$3F,$DC,$73,$FE,$0B,$6B,$D0,$E1,$BF,$E8,$54,$69,$FF
	.dc.b	$07,$35,$FF,$12,$35,$1C,$3E,$27,$F1,$9E,$C7,$E6,$49,$1F,$E1,$BF
	.dc.b	$CB,$3F,$C3,$18,$7E,$CF,$E3,$BF,$5D,$FB,$F3,$E3,$E9,$FA,$18,$B7
	.dc.b	$F9,$CE,$1F,$D2,$18,$7E,$2D,$BE,$B1,$F1,$7F,$BB,$8C,$36,$33,$3F
	.dc.b	$E9,$3C,$5F,$29,$9A,$F1,$FC,$8C,$AF,$F4,$EE,$1C,$9F,$F1,$A7,$E2
	.dc.b	$1F,$AD,$6F,$87,$F2,$63,$F5,$71,$F3,$5D,$36,$3F,$81,$A8,$DB,$FE
	.dc.b	$3B,$6B,$FE,$51,$6B,$FE,$B6,$C5,$F2,$5F,$20,$91,$E4,$F1,$3E,$5F
	.dc.b	$C2,$C8,$F7,$3F,$3B,$E1,$FE,$7F,$27,$D0,$F8,$79,$5F,$0E,$3E,$9F
	.dc.b	$C9,$7D,$4D,$E4,$FC,$78,$AF,$E5,$3C,$76,$8F,$91,$FB,$66,$B8,$3F
	.dc.b	$D5,$90,$F9,$1F,$DA,$F8,$F3,$0F,$A5,$15,$FF,$2E,$78,$DF,$EF,$59
	.dc.b	$FD,$A1,$23,$F9,$85,$3E,$47,$F1,$24,$8F,$E0,$9E,$3A,$BF,$81,$F8
	.dc.b	$B2,$72,$FE,$43,$E4,$71,$1E,$BC,$1F,$23,$F9,$43,$3A,$7A,$BD,$4F
	.dc.b	$CB,$FC,$84,$BF,$C5,$45,$79,$45,$9E,$11,$F8,$FF,$91,$FE,$2B,$E4
	.dc.b	$48,$C1,$E8,$7F,$89,$11,$FE,$24,$FF,$48,$C5,$F2,$3F,$A8,$24,$FE
	.dc.b	$6F,$E4,$7F,$90,$F9,$52,$3E,$1B,$FA,$EF,$7B,$F9,$2F,$99,$FC,$7C
	.dc.b	$93,$FC,$7D,$FE,$45,$E3,$7E,$D2,$21,$F3,$1F,$E5,$BE,$34,$8F,$9E
	.dc.b	$FF,$67,$E9,$7C,$97,$F4,$9F,$43,$F7,$95,$FC,$BD,$AF,$DE,$6B,$FD
	.dc.b	$11,$AF,$F0,$CE,$94,$3C,$6E,$DF,$25,$2B,$FA,$14,$E5,$FE,$27,$C7
	.dc.b	$FE,$FB,$C5,$F1,$9E,$5F,$8E,$95,$AD,$3F,$8C,$D0,$DC,$74,$F1,$3F
	.dc.b	$AE,$62,$69,$23,$FC,$05,$2F,$90,$F8,$44,$8E,$C7,$58,$7C,$8F,$C5
	.dc.b	$F1,$64,$FA,$9F,$95,$F2,$3F,$27,$C5,$F0,$92,$7E,$53,$FE,$0A,$B7
	.dc.b	$C7,$7D,$12,$4F,$82,$3D,$9F,$1F,$FC,$F7,$E5,$FF,$49,$3F,$E1,$46
	.dc.b	$BF,$E1,$D6,$BF,$EC,$D6,$BF,$33,$D2,$F8,$AF,$F3,$A6,$BF,$E2,$E6
	.dc.b	$BB,$E1,$FF,$10,$F1,$3F,$43,$FF,$0F,$CF,$F3,$86,$C3,$8B,$EC,$C6
	.dc.b	$1B,$F1,$9F,$D5,$7C,$9F,$F6,$B5,$7D,$DF,$91,$FC,$81,$9F,$17,$F2
	.dc.b	$1E,$5F,$D8,$26,$5F,$91,$7E,$DD,$FE,$54,$91,$F1,$F7,$7C,$8F,$E6
	.dc.b	$FE,$2F,$F3,$9F,$23,$2B,$51,$F9,$DF,$2B,$F0,$A4,$8F,$CF,$3F,$57
	.dc.b	$E4,$7F,$48,$48,$ED,$F1,$7F,$A7,$F9,$1F,$D3,$79,$B2,$7C,$51,$5F
	.dc.b	$D3,$47,$E4,$88,$FF,$86,$7C,$3F,$CA,$3F,$E2,$36,$BF,$E0,$84,$63
	.dc.b	$87,$F4,$9C,$5F,$E3,$D8,$C7,$1F,$E8,$B7,$FA,$17,$F9,$7F,$8B,$D7
	.dc.b	$81,$FB,$BF,$2F,$F7,$DF,$0F,$2B,$F6,$31,$5F,$E8,$96,$F9,$1F,$E0
	.dc.b	$86,$74,$FD,$A4,$7F,$9B,$73,$7C,$8D,$3F,$85,$45,$BF,$CD,$4F,$F8
	.dc.b	$73,$4F,$DB,$B6,$3F,$70,$F2,$77,$F0,$7F,$E6,$E8,$FF,$1C,$C7,$F8
	.dc.b	$58,$DC,$64,$D7,$C7,$78,$FE,$36,$4F,$82,$2B,$FD,$7E,$BF,$CA,$D9
	.dc.b	$7E,$5E,$3C,$3F,$15,$A8,$FF,$19,$FC,$C3,$D2,$FE,$19,$5F,$E8,$98
	.dc.b	$C6,$9F,$E8,$46,$C6,$67,$6F,$9C,$FF,$6F,$F5,$7F,$A1,$DF,$F4,$97
	.dc.b	$C8,$A3,$11,$B8,$C7,$C8,$7F,$A4,$1B,$1E,$1F,$4D,$EA,$78,$5B,$19
	.dc.b	$8C,$1F,$E9,$DD,$5F,$09,$EF,$D7,$E0,$6B,$F8,$CF,$90,$C7,$1F,$D4
	.dc.b	$AB,$8F,$D4,$9B,$5F,$92,$9B,$87,$4B,$FB,$BF,$21,$89,$E3,$3F,$E3
	.dc.b	$4F,$83,$FF,$2E,$E3,$FD,$0F,$E1,$D1,$88,$FC,$EC,$63,$0F,$F3,$E3
	.dc.b	$63,$FC,$16,$BD,$1E,$AE,$74,$FB,$58,$7F,$90,$A7,$38,$69,$98,$7F
	.dc.b	$A3,$ED,$F0,$98,$C7,$F0,$9F,$03,$EC,$F8,$5F,$6B,$6F,$F3,$2E,$1F
	.dc.b	$CD,$FC,$85,$AF,$DB,$66,$F5,$3E,$2F,$C8,$7C,$28,$C5,$FE,$8A,$31
	.dc.b	$32,$3F,$4B,$F0,$96,$FF,$CF,$9F,$19,$FD,$12,$7F,$DA,$2D,$8F,$F6
	.dc.b	$D3,$5F,$C7,$35,$FD,$6F,$C7,$FD,$A5,$1C,$78,$7F,$03,$92,$B5,$39
	.dc.b	$BF,$DA,$FC,$27,$DE,$3E,$FF,$C8,$58,$FE,$C7,$8F,$4F,$53,$ED,$36
	.dc.b	$3C,$F1,$8B,$3F,$7F,$1F,$F3,$BF,$97,$F6,$8F,$C5,$F7,$DB,$FF,$3B
	.dc.b	$FC,$BF,$F9,$4D,$BF,$15,$F9,$0E,$EF,$CA,$EE,$AD,$69,$FE,$A3,$FB
	.dc.b	$67,$27,$91,$B8,$FF,$BE,$9B,$3F,$B5,$0E,$1F,$F6,$EF,$CB,$BC,$EF
	.dc.b	$CD,$B6,$BF,$8D,$F0,$62,$7C,$72,$F1,$A9,$FA,$2F,$DF,$BE,$8D,$FF
	.dc.b	$A4,$A7,$4F,$CA,$F9,$9F,$F1,$23,$E3,$C5,$18,$7C,$CD,$7F,$86,$35
	.dc.b	$FF,$13,$FE,$DD,$D1,$FF,$55,$B7,$6F,$D2,$FE,$ED,$FD,$92,$FF,$D4
	.dc.b	$73,$97,$F9,$9F,$C5,$FE,$E3,$E3,$5F,$15,$FB,$E8,$C3,$F8,$BF,$A1
	.dc.b	$75,$FE,$01,$B7,$E9,$6E,$7F,$35,$AF,$F3,$9F,$23,$FA,$DC,$F7,$F8
	.dc.b	$BF,$E1,$6D,$3F,$C4,$1A,$FF,$AF,$1B,$F0,$9F,$C4,$A3,$17,$FE,$99
	.dc.b	$7F,$DA,$13,$F9,$4C,$3F,$9C,$3E,$3E,$AF,$31,$1F,$C0,$62,$AF,$F5
	.dc.b	$9C,$78,$3E,$4F,$B7,$C6,$F7,$7E,$27,$C4,$78,$7D,$BB,$5E,$67,$F5
	.dc.b	$DF,$34,$F1,$FF,$46,$D9,$FB,$8D,$D7,$F2,$0D,$7F,$AE,$78,$5F,$EB
	.dc.b	$2F,$8A,$FA,$47,$BF,$CE,$B7,$FE,$52,$9F,$F5,$BF,$C2,$7C,$3F,$37
	.dc.b	$EC,$36,$7F,$6B,$8B,$CA,$F1,$92,$3C,$6F,$EC,$FA,$7F,$A9,$47,$56
	.dc.b	$E7,$F3,$7C,$FF,$C2,$7A,$DF,$47,$C8,$97,$F5,$BC,$DB,$7F,$2E,$3F
	.dc.b	$BB,$7C,$77,$D6,$8F,$EC,$4D,$73,$E8,$FF,$08,$8C,$47,$F9,$E7,$9F
	.dc.b	$FE,$00,$D7,$F0,$E9,$FB,$F1,$FD,$60,$CC,$F7,$07,$FA,$3A,$BF,$0D
	.dc.b	$F2,$5F,$87,$F8,$8B,$87,$E3,$1B,$5F,$89,$9F,$C6,$7C,$46,$30,$ED
	.dc.b	$F0,$FF,$8F,$3F,$21,$F1,$37,$1F,$CA,$B7,$0F,$C9,$36,$BC,$DF,$25
	.dc.b	$F9,$7F,$4B,$F3,$0D,$BF,$CE,$FC,$96,$1A,$7E,$82,$3F,$3D,$F0,$BF
	.dc.b	$9F,$9B,$3A,$C7,$FC,$89,$C9,$FC,$4D,$FD,$17,$BB,$FB,$9F,$43,$F4
	.dc.b	$CF,$E9,$E3,$15,$FA,$C6,$9F,$F0,$07,$DF,$FD,$FC,$7F,$A3,$DF,$F6
	.dc.b	$0B,$7C,$AF,$F0,$0C,$62,$78,$7C,$6E,$0C,$F9,$56,$FF,$2D,$FB,$CB
	.dc.b	$E2,$FF,$C0,$6D,$8F,$B9,$C1,$89,$FE,$AD,$D0,$E1,$FC,$21,$2B,$F9
	.dc.b	$24,$65,$BC,$1C,$27,$8B,$63,$C3,$BF,$80,$F0,$FC,$45,$8F,$2B,$75
	.dc.b	$FF,$37,$67,$FE,$5F,$9E,$27,$FB,$03,$8D,$FB,$FF,$FB,$4A,$3F,$65
	.dc.b	$FB,$D3,$E1,$FF,$A9,$CE,$70,$FF,$3C,$4E,$6B,$7E,$C6,$B8,$BF,$C4
	.dc.b	$B3,$FC,$C6,$BF,$89,$5F,$1C,$7F,$C3,$72,$F8,$2D,$9F,$DE,$FC,$FF
	.dc.b	$F3,$7E,$5E,$28,$C7,$C4,$FF,$72,$93,$49,$3F,$D3,$63,$C9,$8F,$EA
	.dc.b	$4F,$F7,$44,$F5,$D3,$9D,$7F,$7E,$FD,$9B,$FB,$3E,$5E,$4F,$37,$FC
	.dc.b	$19,$18,$9F,$F7,$F6,$7F,$C7,$6B,$D3,$E3,$EB,$DE,$F7,$B6,$BE,$66
	.dc.b	$DF,$E0,$47,$E5,$FA,$7F,$2B,$E0,$7E,$2B,$F1,$BE,$07,$13,$F4,$1B
	.dc.b	$BF,$E8,$B8,$3F,$8D,$12,$7F,$BE,$79,$EF,$05,$3F,$E4,$F1,$DF,$FE
	.dc.b	$FA,$AF,$E4,$11,$F9,$EF,$94,$FE,$C4,$FF,$96,$A7,$DD,$87,$25,$F2
	.dc.b	$B9,$DF,$F4,$AD,$C3,$F9,$B4,$F0,$BE,$FF,$2B,$9C,$3F,$D7,$91,$F5
	.dc.b	$AB,$CE,$D8,$FF,$44,$67,$F5,$27,$3C,$DE,$9C,$1E,$B7,$F9,$BB,$0E
	.dc.b	$1C,$7F,$E7,$C9,$CD,$FF,$47,$9C,$BF,$C9,$9F,$83,$EA,$F3,$24,$7F
	.dc.b	$64,$9F,$F2,$56,$5E,$FE,$2F,$E0,$35,$FC,$09,$B5,$FD,$4F,$F8,$07
	.dc.b	$E1,$BF,$C2,$7D,$4F,$F4,$0C,$62,$FF,$87,$46,$37,$95,$C7,$7B,$5F
	.dc.b	$C4,$FD,$2F,$71,$27,$C8,$FC,$1F,$94,$FE,$26,$FF,$14,$6C,$7F,$95
	.dc.b	$27,$FA,$1F,$AB,$FD,$13,$E7,$DE,$77,$FA,$CF,$CD,$FF,$60,$46,$2B
	.dc.b	$FD,$6D,$5B,$AD,$5D,$9C,$8F,$92,$E0,$FF,$50,$C3,$FD,$95,$18,$8F
	.dc.b	$09,$BA,$E6,$D3,$9E,$1B,$AF,$1C,$FF,$81,$25,$39,$BF,$14,$7F,$B8
	.dc.b	$70,$AF,$FB,$0A,$2A,$8C,$DF,$48,$F4,$E1,$CD,$F1,$1C,$CA,$73,$29
	.dc.b	$CC,$F1,$E1,$EA,$C3,$DD,$19,$7A,$18,$BF,$0E,$1E,$46,$CF,$23,$E0
	.dc.b	$D7,$33,$E0,$57,$27,$E1,$5F,$9C,$E9,$3E,$6C,$3E,$7C,$F0,$39,$47
	.dc.b	$43,$FC,$06,$7A,$EB,$8D,$73,$9E,$A7,$D4,$72,$D3,$33,$CA,$BB,$BF
	.dc.b	$82,$E2,$FC,$13,$6B,$EC,$46,$23,$E6,$37,$E3,$D9,$FF,$93,$E7,$F8
	.dc.b	$9D,$79,$A7,$3C,$9F,$B5,$8C,$56,$EF,$93,$63,$33,$FE,$20,$7F,$27
	.dc.b	$5F,$DB,$EF,$18,$F5,$F3,$3E,$1D,$6B,$0F,$13,$70,$DD,$FF,$8A,$B0
	.dc.b	$E4,$FE,$19,$B3,$B8,$C5,$7A,$F3,$F0,$E1,$F4,$1B,$3F,$98,$BF,$6B
	.dc.b	$F8,$10,$E2,$F7,$1D,$EB,$CF,$8F,$BA,$3B,$E3,$CE,$7B,$4F,$B6,$7B
	.dc.b	$63,$E3,$8E,$18,$F4,$23,$82,$4F,$BE,$B9,$B4,$E7,$71,$F2,$1F,$18
	.dc.b	$F4,$C7,$99,$F9,$17,$F3,$2B,$AC,$68,$F4,$1C,$EF,$E5,$1E,$47,$C0
	.dc.b	$7C,$5C,$7E,$3C,$7C,$FC,$7E,$29,$2B,$E2,$3C,$9F,$85,$5C,$70,$F5
	.dc.b	$BF,$3B,$0F,$88,$FF,$59,$8F,$D2,$E3,$FD,$1F,$E5,$6E,$DF,$D1,$79
	.dc.b	$91,$8B,$E0,$7F,$91,$5F,$F0,$54,$6A,$3F,$A4,$46,$0F,$F6,$D7,$4B
	.dc.b	$95,$E4,$F3,$8C,$36,$33,$5F,$D2,$FC,$AC,$4F,$87,$87,$FB,$BF,$CA
	.dc.b	$DC,$8E,$AF,$FC,$B1,$F2,$B9,$C2,$46,$0F,$F2,$E3,$E8,$23,$FD,$7D
	.dc.b	$F5,$E1,$E5,$C3,$FC,$79,$3F,$DA,$7C,$8F,$04,$E5,$F6,$CE,$A7,$FB
	.dc.b	$BB,$5F,$F5,$DB,$5F,$F4,$AE,$7F,$E1,$0F,$0B,$E1,$66,$69,$ED,$E3
	.dc.b	$B5,$FE,$73,$BF,$80,$61,$FF,$50,$3F,$EA,$0F,$95,$F4,$7C,$AC,$9F
	.dc.b	$F5,$1F,$13,$8B,$FC,$01,$0A,$D3,$E8,$AF,$59,$1F,$F1,$4F,$82,$F5
	.dc.b	$3F,$05,$87,$1F,$93,$B8,$1F,$B9,$5B,$BF,$F3,$44,$E5,$FF,$6C,$43
	.dc.b	$D5,$D9,$CA,$B8,$F7,$FE,$3E,$BF,$CB,$3D,$16,$FF,$89,$E3,$EB,$97
	.dc.b	$0F,$F9,$7A,$F5,$7E,$2F,$12,$57,$C2,$DB,$FA,$1E,$3C,$9F,$E8,$BF
	.dc.b	$8D,$3E,$6F,$C0,$BD,$CA,$62,$2D,$FF,$4C,$38,$C5,$69,$44,$F3,$7E
	.dc.b	$5F,$8B,$F1,$88,$F9,$4F,$F6,$CC,$3D,$BF,$97,$D9,$F3,$BC,$E9,$5F
	.dc.b	$2A,$2B,$FE,$E1,$DF,$FC,$FF,$DB,$F4,$8C,$F0,$FC,$78,$7E,$F7,$F8
	.dc.b	$1A,$F0,$46,$37,$FD,$A4,$FF,$A2,$27,$2E,$DE,$CE,$5B,$1C,$0E,$3C
	.dc.b	$BF,$67,$E2,$7F,$6B,$3E,$D0,$B1,$D1,$BC,$AC,$FC,$99,$D3,$67,$E6
	.dc.b	$3C,$9F,$E1,$B8,$3E,$5F,$D2,$49,$F0,$53,$C0,$FE,$15,$1F,$B7,$D3
	.dc.b	$F1,$5D,$68,$C7,$27,$E9,$BB,$F7,$B7,$0E,$CE,$3E,$BC,$BF,$B1,$87
	.dc.b	$19,$EB,$A7,$F0,$9F,$5A,$56,$8F,$31,$FE,$73,$7C,$7D,$92,$30,$6F
	.dc.b	$E5,$FC,$F3,$23,$07,$37,$D7,$F2,$F2,$30,$77,$7D,$FE,$8C,$8C,$1E
	.dc.b	$3F,$63,$E1,$F4,$48,$C1,$E7,$70,$FF,$0E,$0F,$CA,$8F,$97,$C7,$27
	.dc.b	$DC,$FC,$EF,$97,$C4,$71,$F9,$7F,$A2,$63,$93,$E8,$C3,$D7,$FE,$01
	.dc.b	$F3,$AC,$69,$F6,$22,$CF,$EB,$67,$4D,$9F,$C2,$C6,$23,$F6,$DC,$5F
	.dc.b	$77,$E5,$BF,$02,$48,$FA,$9E,$4F,$F3,$FC,$E4,$7F,$A9,$DF,$C3,$FC
	.dc.b	$8F,$F1,$3F,$C3,$DF,$25,$F9,$26,$D7,$EF,$DA,$FF,$03,$6B,$FD,$0B
	.dc.b	$7F,$C8,$F5,$C8,$FF,$5A,$47,$F4,$FD,$A4,$7F,$AD,$DF,$EC,$DF,$8C
	.dc.b	$74,$E0,$FE,$41,$2C,$1F,$15,$E3,$24,$78,$DF,$EA,$BB,$72,$64,$7F
	.dc.b	$B0,$E3,$AF,$C8,$C8,$FE,$07,$82,$3C,$3F,$25,$FC,$CF,$F1,$6F,$EA
	.dc.b	$D5,$FC,$75,$AF,$F2,$16,$BF,$C9,$5A,$FF,$60,$A7,$C9,$7B,$3E,$4E
	.dc.b	$4F,$82,$3D,$FF,$27,$FE,$05,$F2,$7D,$B3,$FC,$59,$AF,$F1,$FB,$4F
	.dc.b	$97,$3F,$E0,$B8,$9A,$4A,$FE,$6D,$6B,$7C,$9F,$7D,$25,$7C,$28,$FE
	.dc.b	$A5,$F8,$D7,$F9,$EF,$83,$E8,$B7,$8F,$FD,$1A,$DA,$FE,$A7,$FA,$17
	.dc.b	$F2,$05,$5F,$63,$E8,$1E,$07,$F1,$A6,$CF,$F4,$76,$BF,$D2,$DA,$FF
	.dc.b	$8D,$B9,$DF,$19,$FD,$C9,$B1,$FD,$26,$FF,$2B,$3F,$E2,$DF,$DF,$1E
	.dc.b	$17,$F8,$BF,$CF,$7A,$F4,$98,$7E,$03,$E5,$BF,$B8,$7E,$17,$F8,$18
	.dc.b	$FE,$B4,$D7,$FA,$F3,$5F,$EC,$4D,$7F,$D5,$FE,$9F,$FB,$93,$3F,$CB
	.dc.b	$B7,$87,$3F,$05,$F9,$CF,$98,$FF,$21,$7D,$D7,$DC,$BF,$EA,$CD,$AF
	.dc.b	$F1,$3B,$70,$FE,$FF,$CF,$FD,$DD,$FE,$99,$F6,$BF,$E0,$B7,$FB,$83
	.dc.b	$5F,$EE,$8D,$7F,$BC,$35,$FC,$A6,$5F,$AE,$23,$FA,$2F,$6B,$F6,$3F
	.dc.b	$35,$FE,$13,$AF,$9F,$3F,$73,$1F,$F6,$6C,$63,$2F,$ED,$D8,$AF,$87
	.dc.b	$1F,$F5,$8A,$E3,$87,$8A,$30,$7F,$81,$EF,$F2,$9E,$B7,$F8,$BB,$0F
	.dc.b	$EB,$15,$F7,$5A,$FE,$87,$E6,$B5,$3F,$C1,$B7,$93,$B8,$FE,$13,$F2
	.dc.b	$D2,$3F,$7A,$FF,$0C,$F9,$9C,$47,$EA,$FE,$63,$F8,$79,$8C,$BE,$9F
	.dc.b	$07,$D6,$24,$FF,$82,$38,$7E,$93,$FE,$11,$7F,$C3,$0E,$78,$7E,$E9
	.dc.b	$C1,$D3,$FE,$84,$8F,$F1,$24,$7F,$8A,$FC,$04,$62,$3F,$DF,$31,$89
	.dc.b	$FF,$45,$ED,$FE,$31,$C3,$E7,$D7,$F2,$63,$F9,$3F,$CD,$57,$F8,$DF
	.dc.b	$C7,$7F,$2A,$30,$FD,$93,$FC,$01,$F4,$7F,$D4,$7D,$1F,$F1,$C3,$63
	.dc.b	$FC,$78,$D7,$FC,$80,$D7,$FD,$F1,$F2,$2F,$8A,$FF,$20,$7C,$82,$31
	.dc.b	$3F,$E4,$58,$C4,$EA,$73,$3D,$3D,$46,$3E,$26,$C7,$89,$AE,$E3,$F8
	.dc.b	$DF,$D0,$7F,$9A,$FA,$FF,$C7,$6B,$E9,$7F,$CC,$DB,$7F,$90,$63,$57
	.dc.b	$FE,$47,$75,$5C,$3E,$2F,$6F,$FA,$5A,$37,$5C,$7E,$33,$37,$FE,$70
	.dc.b	$9C,$D6,$FB,$3F,$AA,$CD,$8F,$2C,$62,$38,$37,$91,$4F,$94,$EB,$18
	.dc.b	$9B,$39,$BF,$6C,$62,$A4,$7C,$0F,$97,$B8,$7B,$7E,$5F,$8F,$8B,$C9
	.dc.b	$B5,$F1,$9C,$3F,$23,$E3,$AD,$7F,$9A,$F3,$FF,$42,$4F,$CC,$F9,$7F
	.dc.b	$2C,$62,$7E,$54,$63,$E3,$E5,$7D,$2F,$8F,$B5,$CF,$BB,$E0,$61,$FE
	.dc.b	$7D,$AF,$EA,$6D,$9F,$D1,$35,$FD,$0E,$FF,$8B,$9F,$EC,$55,$F1,$39
	.dc.b	$BC,$6F,$C6,$F9,$0F,$8F,$F2,$B6,$BF,$06,$FD,$8E,$4F,$F7,$97,$BF
	.dc.b	$F4,$DB,$3F,$E9,$96,$E5,$FA,$DF,$94,$FF,$3B,$CF,$E3,$BE,$47,$E9
	.dc.b	$9F,$53,$E5,$6C,$7E,$43,$C8,$FA,$FF,$25,$F6,$1B,$1F,$D5,$7E,$EF
	.dc.b	$FA,$CB,$DD,$B8,$FE,$ED,$AF,$F9,$CC,$FE,$D2,$E1,$FC,$A7,$CA,$FE
	.dc.b	$10,$FC,$2F,$CB,$5A,$FF,$47,$7D,$13,$E1,$3F,$21,$68,$C4,$FF,$AD
	.dc.b	$5B,$1F,$EB,$96,$BE,$1D,$9F,$42,$F9,$4F,$90,$DB,$F9,$0D,$AF,$E2
	.dc.b	$8D,$79,$70,$7F,$BF,$67,$FB,$B3,$63,$F6,$5E,$4F,$D5,$8C,$61,$FE
	.dc.b	$B7,$FA,$8F,$F7,$B5,$FF,$88,$7C,$FF,$D2,$C6,$30,$FC,$B4,$7E,$53
	.dc.b	$D1,$FC,$69,$2B,$F1,$AF,$EB,$E9,$F2,$DF,$9A,$F2,$79,$B6,$BF,$D8
	.dc.b	$73,$EB,$C7,$C5,$F2,$FE,$0D,$BC,$2F,$E2,$FF,$80,$7F,$CC,$77,$F8
	.dc.b	$96,$E9,$FC,$D5,$AF,$CB,$6B,$FE,$AF,$EF,$63,$FE,$B9,$AF,$0F,$CB
	.dc.b	$FB,$BE,$63,$DE,$FF,$04,$CD,$E4,$7C,$2F,$98,$C3,$6F,$FD,$B7,$3F
	.dc.b	$C1,$F2,$FE,$98,$D7,$FA,$5B,$5F,$13,$56,$5F,$E3,$8F,$1F,$C0,$7D
	.dc.b	$0F,$98,$B7,$DB,$C8,$FA,$7F,$31,$E5,$6C,$7F,$8C,$F7,$FE,$0D,$8F
	.dc.b	$C1,$F2,$7F,$61,$18,$8F,$F6,$86,$68,$C4,$78,$A3,$F8,$03,$FE,$FD
	.dc.b	$69,$F8,$5F,$82,$F8,$9E,$47,$E1,$3E,$5B,$F0,$BF,$0F,$72,$FB,$6F
	.dc.b	$C9,$F2,$7F,$98,$60,$69,$27,$FC,$CB,$CB,$FE,$BE,$7E,$5C,$7E,$43
	.dc.b	$E6,$39,$47,$E1,$BC,$9F,$C9,$FC,$B7,$E5,$1B,$19,$26,$BF,$DB,$F9
	.dc.b	$3E,$FF,$CC,$6E,$7F,$1F,$F1,$FF,$9C,$F9,$6F,$CE,$B6,$32,$4B,$E5
	.dc.b	$A7,$E9,$A3,$F1,$4F,$F9,$FD,$FC,$CF,$CB,$FE,$98,$FD,$3B,$4F,$CE
	.dc.b	$98,$3E,$89,$F9,$3F,$8C,$FF,$55,$BA,$33,$5F,$84,$8C,$BE,$B6,$EE
	.dc.b	$9E,$51,$FD,$68,$CB,$D7,$E4,$71,$E4,$E5,$B1,$DB,$5F,$F6,$6F,$CA
	.dc.b	$E6,$31,$68,$C6,$3E,$1F,$94,$ED,$F2,$BD,$DB,$7C,$3E,$3B,$C4,$78
	.dc.b	$FE,$52,$C7,$B2,$DF,$3B,$FE,$84,$8E,$FF,$1F,$FD,$FA,$73,$EE,$FA
	.dc.b	$E3,$14,$F0,$FF,$96,$B6,$FF,$9C,$5F,$F3,$98,$C3,$83,$BC,$F0,$F7
	.dc.b	$79,$BD,$1B,$5F,$03,$DB,$F9,$5E,$17,$CB,$6C,$64,$97,$F4,$FE,$27
	.dc.b	$51,$F2,$3E,$37,$E9,$7C,$7F,$D3,$6C,$7C,$CF,$8E,$C3,$4E,$07,$C5
	.dc.b	$FD,$AF,$FC,$3B,$87,$E0,$7E,$3F,$E2,$46,$39,$DC,$E9,$F9,$1F,$5C
	.dc.b	$F2,$FF,$A8,$B7,$1F,$85,$8A,$30,$FC,$56,$BF,$2B,$1F,$EC,$BD,$9F
	.dc.b	$F0,$02,$76,$FE,$3E,$42,$77,$5F,$DA,$6B,$6F,$E7,$7D,$5F,$CA,$7A
	.dc.b	$DF,$95,$6C,$7E,$3B,$E3,$B0,$D3,$F2,$31,$F9,$BC,$3F,$39,$EB,$D9
	.dc.b	$FC,$AB,$FE,$38,$F6,$3F,$C9,$73,$FA,$53,$1F,$15,$B7,$F8,$17,$C6
	.dc.b	$7B,$9F,$E8,$4D,$7F,$98,$B7,$87,$FC,$D9,$BF,$F9,$1E,$DF,$15,$FE
	.dc.b	$46,$AF,$6D,$FF,$38,$9C,$E5,$FA,$B9,$CE,$1F,$E7,$5F,$3F,$CD,$F1
	.dc.b	$7C,$5B,$5B,$70,$FE,$CE,$FE,$5E,$1E,$49,$F0,$79,$BF,$E5,$9C,$4D
	.dc.b	$25,$FE,$7F,$16,$1F,$D6,$23,$C9,$3F,$1F,$D6,$E7,$1D,$FE,$2B,$DC
	.dc.b	$47,$95,$7F,$55,$BE,$7E,$17,$CF,$AF,$A0,$DA,$F4,$7A,$DF,$4F,$C1
	.dc.b	$FA,$8D,$8C,$93,$5F,$E6,$98,$BC,$1D,$CF,$C1,$F8,$AF,$A0,$47,$C6
	.dc.b	$FD,$3F,$57,$F0,$0F,$E0,$5B,$1F,$C3,$6F,$F2,$BF,$3F,$F9,$6F,$9F
	.dc.b	$78,$1F,$C1,$BA,$3F,$35,$18,$AF,$E1,$35,$BA,$D5,$D9,$C8,$F8,$5E
	.dc.b	$0F,$E8,$B0,$FD,$24,$62,$3C,$26,$EB,$9B,$4E,$78,$6E,$BC,$73,$FC
	.dc.b	$5F,$0F,$05,$75,$C2,$BD,$2F,$84,$62,$13,$AB,$E7,$1E,$8C,$36,$F8
	.dc.b	$4F,$F2,$F4,$FE,$EA,$32,$7A,$63,$98,$BE,$26,$67,$B5,$F8,$5B,$3B
	.dc.b	$C3,$D4,$D3,$FC,$D9,$87,$91,$F0,$99,$3D,$C5,$3D,$D1,$F2,$2F,$C1
	.dc.b	$1E,$93,$FD,$AF,$7E,$99,$EF,$87,$A9,$F0,$D7,$A1,$A7,$33,$67,$D4
	.dc.b	$8E,$A7,$1A,$F9,$11,$93,$6F,$AA,$BC,$0F,$E1,$3E,$0B,$F0,$8D,$AF
	.dc.b	$83,$18,$8F,$7B,$7E,$3D,$F1,$8F,$0B,$F9,$5C,$E6,$31,$48,$C4,$EA
	.dc.b	$FF,$D2,$D7,$CA,$31,$5F,$CD,$23,$0F,$DF,$7F,$C6,$0F,$A0,$38,$75
	.dc.b	$7E,$0A,$FC,$EC,$F2,$6F,$B5,$FE,$A5,$6D,$73,$9D,$65,$E4,$BF,$4E
	.dc.b	$7A,$AF,$BD,$5D,$E3,$AC,$EA,$B8,$63,$CF,$1E,$D1,$EE,$6E,$1E,$DA
	.dc.b	$F3,$4F,$08,$EB,$3F,$EB,$9B,$F0,$1D,$AF,$F5,$78,$7C,$48,$CD,$F5
	.dc.b	$BC,$E1,$E3,$7B,$5F,$0C,$3A,$56,$E7,$D7,$5B,$7E,$5B,$E4,$7E,$94
	.dc.b	$FA,$1F,$9E,$DC,$3E,$3C,$76,$3D,$27,$73,$C7,$71,$C9,$30,$F6,$3E
	.dc.b	$28,$F9,$0C,$3E,$DB,$F5,$6F,$CD,$5F,$47,$0F,$55,$F2,$3F,$91,$73
	.dc.b	$7F,$48,$AF,$F6,$96,$9F,$EA,$7E,$2F,$C3,$63,$FE,$01,$AF,$E9,$B9
	.dc.b	$EB,$3F,$EC,$58,$59,$CC,$7F,$4A,$9F,$F0,$8C,$62,$3F,$C6,$ED,$7F
	.dc.b	$DE,$0F,$F5,$8F,$8A,$E5,$81,$A6,$3E,$2E,$D7,$F6,$7F,$8C,$D9,$87
	.dc.b	$FA,$6F,$C6,$72,$F8,$B9,$1F,$DB,$62,$BE,$2C,$13,$E6,$6C,$7F,$62
	.dc.b	$8C,$70,$FB,$7B,$9E,$F5,$3C,$F1,$6B,$FB,$94,$E9,$B3,$F0,$A3,$11
	.dc.b	$F0,$97,$B9,$D4,$91,$FE,$03,$E5,$AF,$FA,$6F,$E0,$9F,$2B,$F3,$DB
	.dc.b	$1F,$E1,$1A,$C7,$C5,$79,$F9,$64,$FF,$9F,$EB,$FD,$E6,$FF,$85,$E3
	.dc.b	$C8,$47,$D2,$FD,$88,$7C,$4F,$7F,$89,$95,$CE,$2B,$E5,$F8,$AF,$21
	.dc.b	$8E,$8F,$C4,$5B,$E2,$BD,$1F,$13,$2F,$C9,$15,$F7,$7C,$57,$B4,$C5
	.dc.b	$FF,$94,$61,$F8,$53,$1F,$C9,$43,$EB,$9E,$F6,$1B,$1F,$E9,$1A,$E0
	.dc.b	$FF,$A4,$E3,$2F,$FA,$1D,$AF,$FA,$31,$AF,$F9,$80,$ED,$3F,$E6,$C7
	.dc.b	$C7,$F1,$7F,$4F,$D1,$91,$DE,$3E,$B7,$C6,$7D,$7F,$42,$4F,$D3,$8A
	.dc.b	$FD,$D9,$FB,$86,$33,$F9,$3F,$17,$F7,$C9,$3F,$6F,$E3,$2B,$F8,$5F
	.dc.b	$4B,$F0,$86,$2B,$E4,$75,$BE,$47,$E3,$AF,$4E,$98,$8B,$7F,$E8,$89
	.dc.b	$D3,$67,$E0,$46,$23,$FD,$5C,$BD,$3F,$CA,$32,$3F,$D8,$0E,$5F,$C9
	.dc.b	$7C,$97,$E6,$48,$FE,$57,$E3,$BF,$39,$F1,$BF,$9D,$EE,$93,$F9,$98
	.dc.b	$FD,$07,$6F,$F0,$FF,$A2,$7C,$F5,$33,$5F,$ED,$D8,$CB,$7E,$0B,$81
	.dc.b	$B7,$5F,$1D,$CC,$8F,$1F,$8F,$ED,$F1,$F2,$3A,$3A,$7F,$BD,$63,$FE
	.dc.b	$A4,$F3,$D8,$BE,$3F,$CC,$4A,$E2,$79,$7E,$3B,$D4,$47,$D1,$F1,$F5
	.dc.b	$EA,$2F,$AF,$C7,$FC,$13,$5C,$FF,$CD,$71,$7C,$67,$C5,$25,$FC,$3F
	.dc.b	$02,$BD,$30,$F3,$FC,$77,$CB,$33,$87,$3F,$8E,$FE,$79,$5F,$E6,$08
	.dc.b	$FE,$84,$DD,$BE,$06,$4A,$DF,$B5,$FD,$02,$BF,$C9,$87,$A4,$B8,$FF
	.dc.b	$B2,$A3,$FA,$2D,$3E,$37,$EE,$7C,$74,$BF,$B5,$1E,$2F,$8B,$91,$FE
	.dc.b	$D2,$CF,$C7,$1F,$83,$C3,$FA,$A4,$3E,$A5,$C1,$FE,$E2,$8C,$36,$33
	.dc.b	$7F,$BD,$3A,$38,$7F,$AB,$8F,$8C,$FC,$87,$C7,$C8,$F1,$6F,$F8,$17
	.dc.b	$46,$0F,$80,$7C,$A2,$1F,$95,$CB,$CF,$F2,$EF,$8D,$F3,$36,$7C,$87
	.dc.b	$F9,$1D,$C3,$E7,$CD,$18,$8F,$E2,$D1,$8F,$8C,$FD,$37,$C7,$49,$F2
	.dc.b	$4E,$5F,$D3,$7C,$37,$EA,$C8,$FE,$A7,$4F,$97,$AB,$3F,$90,$49,$F2
	.dc.b	$DF,$CF,$8F,$EF,$FE,$CF,$9A,$3F,$29,$3F,$B7,$3F,$C0,$D1,$9D,$7F
	.dc.b	$50,$FF,$BB,$A3,$CB,$A7,$FB,$15,$BF,$15,$C3,$E3,$BE,$3F,$C6,$7C
	.dc.b	$62,$4F,$8A,$91,$E2,$DF,$E6,$FC,$5C,$AF,$A9,$D1,$FD,$AA,$7F,$0F
	.dc.b	$FB,$B3,$8A,$92,$BE,$3C,$7C,$17,$EE,$0F,$9F,$FF,$3F,$3F,$89,$6B
	.dc.b	$FD,$1D,$A8,$E1,$F0,$5F,$D2,$7E,$79,$EA,$52,$4F,$E0,$04,$69,$FF
	.dc.b	$7F,$70,$7F,$31,$E2,$FF,$1A,$37,$8B,$E5,$FA,$1F,$3B,$D1,$FE,$CB
	.dc.b	$E0,$7A,$70,$FC,$5B,$5F,$D0,$35,$FE,$BC,$D7,$F8,$85,$BE,$47,$EC
	.dc.b	$FA,$52,$7E,$5C,$E5,$BC,$FF,$78,$8E,$BE,$47,$F6,$BF,$C0,$3F,$97
	.dc.b	$8F,$DD,$35,$FF,$52,$37,$53,$9E,$5F,$F1,$0F,$C9,$7E,$2B,$A2,$5F
	.dc.b	$8A,$3F,$1B,$D1,$FA,$CF,$8B,$93,$E6,$8A,$FE,$A7,$A9,$3F,$B0,$24
	.dc.b	$7E,$2F,$E2,$7F,$6A,$47,$DF,$F1,$F5,$FE,$55,$BF,$F9,$9B,$97,$F3
	.dc.b	$7F,$1F,$9D,$FE,$AF,$C7,$FE,$80,$97,$F2,$3E,$3E,$BC,$B3,$FD,$29
	.dc.b	$8A,$FB,$B9,$FE,$A0,$93,$F4,$3E,$3E,$BE,$2E,$CF,$D6,$98,$AF,$C2
	.dc.b	$BF,$B4,$24,$7E,$29,$FF,$3A,$F8,$7F,$E4,$48,$FE,$88,$D9,$FF,$40
	.dc.b	$37,$4F,$DB,$FC,$6F,$F9,$FE,$31,$5F,$D5,$5F,$F2,$9B,$CB,$6D,$7C
	.dc.b	$6F,$F8,$DB,$E6,$DE,$EF,$FB,$76,$BF,$8E,$C6,$30,$FD,$03,$63,$33
	.dc.b	$B7,$F9,$B7,$C9,$39,$5F,$21,$FA,$78,$C3,$63,$33,$FC,$AA,$72,$F8
	.dc.b	$3E,$01,$E6,$78,$1B,$19,$8C,$1F,$CC,$BA,$FF,$A6,$7C,$53,$83,$F8
	.dc.b	$5F,$07,$F9,$99,$B1,$9B,$F9,$7F,$1F,$78,$BE,$5F,$C8,$5B,$FF,$04
	.dc.b	$61,$FE,$5B,$79,$C6,$6F,$E6,$61,$71,$FE,$B9,$F1,$DF,$E6,$39,$E2
	.dc.b	$D7,$E9,$36,$37,$7E,$9C,$1F,$1D,$CD,$FF,$3A,$47,$AA,$31,$59,$F8
	.dc.b	$FD,$45,$9D,$1F,$03,$E2,$F9,$7C,$6F,$36,$C7,$F4,$EF,$91,$FE,$3A
	.dc.b	$FF,$46,$6E,$7F,$D6,$5A,$FF,$58,$6B,$F1,$3E,$33,$CD,$E7,$F9,$DB
	.dc.b	$5F,$41,$C3,$F4,$BE,$3A,$D7,$F8,$A3,$85,$F3,$6E,$CF,$8D,$18,$6C
	.dc.b	$66,$FF,$AD,$B6,$BF,$DA,$FC,$0F,$8E,$FA,$98,$71,$BF,$C0,$C6,$2F
	.dc.b	$84,$FE,$A3,$E3,$BE,$71,$F3,$FE,$3E,$DF,$F7,$2F,$8C,$79,$FE,$46
	.dc.b	$C7,$F6,$C6,$E7,$FD,$FD,$AF,$FB,$A3,$37,$5F,$DB,$EF,$FB,$8D,$BF
	.dc.b	$EF,$FA,$7E,$36,$FE,$23,$74,$FD,$EF,$C7,$FA,$7A,$BD,$BF,$A2,$7F
	.dc.b	$3F,$C3,$FB,$EB,$70,$FF,$08,$35,$FF,$07,$B5,$FF,$53,$7C,$DB,$E3
	.dc.b	$BF,$29,$08,$C5,$7F,$5D,$6C,$7F,$87,$9A,$FF,$86,$B2,$77,$7C,$D3
	.dc.b	$F3,$1F,$23,$6F,$FC,$37,$F3,$6E,$CF,$BA,$D8,$FF,$0A,$B7,$0F,$F1
	.dc.b	$43,$5F,$F5,$E7,$80,$FA,$C7,$A3,$FA,$F6,$DF,$EF,$8C,$47,$EC,$FD
	.dc.b	$5F,$F4,$8C,$7F,$83,$1B,$7F,$E6,$D5,$CD,$BF,$93,$FC,$37,$85,$F9
	.dc.b	$02,$5F,$E4,$1F,$DC,$53,$E4,$FB,$1D,$F8,$AD,$77,$F9,$2F,$11,$E3
	.dc.b	$F9,$3B,$1F,$80,$F1,$7F,$3A,$47,$F3,$AF,$6F,$8A,$F5,$7C,$97,$AD
	.dc.b	$B3,$F9,$5D,$FD,$DF,$27,$FA,$0D,$FE,$1F,$07,$B6,$3F,$41,$E3,$7E
	.dc.b	$A8,$97,$FA,$A7,$D7,$F2,$3F,$1C,$F9,$1F,$25,$67,$F4,$5E,$DF,$CB
	.dc.b	$F9,$3E,$CD,$8F,$CC,$F8,$FF,$B6,$23,$FB,$67,$E3,$FC,$97,$D3,$3E
	.dc.b	$A7,$C2,$59,$FA,$9F,$25,$F5,$CF,$E5,$9A,$BE,$AD,$A6,$6B,$FC,$7B
	.dc.b	$1F,$EA,$A7,$FC,$64,$D9,$FF,$39,$B5,$FF,$44,$35,$FF,$1F,$1D,$6F
	.dc.b	$BC,$71,$F1,$BF,$A2,$4E,$6B,$BE,$9F,$8B,$F9,$39,$1F,$8B,$F9,$0B
	.dc.b	$5F,$8B,$7F,$20,$7E,$47,$E4,$EC,$FE,$47,$E3,$FF,$28,$7E,$57,$D8
	.dc.b	$B1,$F9,$57,$0F,$E5,$FE,$4A,$E1,$F9,$73,$F3,$87,$F4,$FF,$E0,$1F
	.dc.b	$F4,$A4,$FE,$76,$31,$F2,$92,$3F,$CD,$72,$F9,$3C,$57,$FB,$36,$31
	.dc.b	$3A,$9E,$9F,$29,$D5,$ED,$F2,$3D,$FE,$46,$D7,$47,$C5,$F2,$18,$C7
	.dc.b	$A3,$78,$BF,$B7,$79,$4C,$FF,$8B,$46,$1B,$87,$FB,$D5,$7C,$87,$B7
	.dc.b	$E4,$7D,$DE,$CF,$BE,$2D,$7B,$1F,$85,$F2,$32,$3F,$D0,$3F,$3B,$FE
	.dc.b	$8E,$7E,$0C,$7E,$32,$1F,$25,$FC,$81,$B3,$FE,$7B,$6E,$BF,$EB,$C6
	.dc.b	$BF,$ED,$17,$FC,$39,$1D,$A7,$A7,$C3,$F5,$7E,$A1,$23,$E8,$F8,$D6
	.dc.b	$FE,$83,$F6,$4F,$B5,$56,$7E,$BF,$C8,$7D,$DF,$0F,$EF,$36,$3E,$DF
	.dc.b	$C8,$61,$A7,$03,$EF,$FC,$7F,$E1,$4F,$C3,$7C,$8D,$8F,$F5,$2F,$CE
	.dc.b	$3C,$EF,$E8,$2D,$8F,$F4,$E3,$74,$FF,$6E,$B5,$FE,$67,$C6,$F9,$2F
	.dc.b	$DB,$CE,$F9,$7E,$E4,$63,$0F,$CB,$7C,$87,$E5,$FE,$4A,$E5,$F9,$47
	.dc.b	$FC,$51,$DC,$F9,$2C,$E8,$CB,$F1,$33,$9B,$FC,$DE,$1C,$3E,$3F,$8F
	.dc.b	$C5,$59,$FF,$64,$D7,$F7,$E7,$FD,$79,$D9,$FE,$F9,$F8,$4D,$65,$D6
	.dc.b	$73,$5F,$DB,$DB,$87,$F8,$2B,$E7,$BF,$AF,$D7,$FB,$17,$34,$62,$3C
	.dc.b	$CD,$8C,$FB,$3F,$E3,$DC,$3F,$B2,$E1,$FE,$FC,$AF,$ED,$75,$EC,$F8
	.dc.b	$2E,$D7,$F0,$BE,$3F,$E1,$FC,$85,$C3,$D6,$DC,$BD,$03,$56,$6B,$F9
	.dc.b	$7E,$FE,$F3,$FF,$2E,$69,$FD,$5F,$2F,$E4,$38,$FA,$E3,$0E,$27,$DB
	.dc.b	$EF,$7F,$8C,$A7,$33,$FD,$CB,$1F,$EE,$D5,$D2,$BE,$CF,$C1,$FC,$77
	.dc.b	$ED,$F2,$FC,$12,$63,$FE,$75,$BF,$C5,$61,$F4,$3E,$3F,$73,$F6,$1B
	.dc.b	$1E,$ED,$5F,$BF,$FF,$06,$8F,$81,$6F,$90,$FF,$38,$4E,$6B,$E1,$C6
	.dc.b	$3C,$89,$8F,$D9,$3F,$07,$5F,$C3,$9C,$61,$F2,$E3,$96,$0F,$73,$ED
	.dc.b	$12,$BD,$4F,$DD,$F9,$0F,$CC,$BF,$9A,$6C,$74,$BF,$E9,$FF,$3C,$5F
	.dc.b	$3C,$F5,$FE,$9F,$A9,$FE,$93,$8C,$4F,$D6,$AE,$11,$88,$E5,$3B,$AF
	.dc.b	$B7,$0A,$ED,$DF,$B8,$FF,$54,$E1,$8A,$DD,$59,$FE,$C7,$7F,$EA,$3B
	.dc.b	$E1,$1C,$CF,$3D,$79,$AF,$C6,$F2,$8C,$9C,$A3,$2F,$A4,$C9,$E8,$72
	.dc.b	$FC,$43,$27,$98,$A7,$86,$7D,$55,$E5,$AF,$04,$62,$BD,$55,$CF,$0C
	.dc.b	$3C,$B0,$EE,$F5,$8F,$6C,$63,$0F,$95,$3E,$79,$E2,$69,$F2,$C6,$9E
	.dc.b	$D8,$7C,$37,$B0,$C3,$C2,$71,$2C,$7C,$7A,$EA,$D3,$93,$B7,$A4,$59
	.dc.b	$F9,$87,$58,$F8,$26,$8A,$EA,$3D,$1D,$DE,$38,$FA,$4D,$7E,$83,$5F
	.dc.b	$CA,$61,$F9,$58,$D7,$CA,$7F,$9B,$B9,$B1,$7F,$E2,$88,$C5,$AB,$7E
	.dc.b	$57,$F3,$78,$C5,$6E,$31,$3A,$BF,$AF,$6F,$27,$FC,$67,$7F,$9D,$8C
	.dc.b	$61,$AC,$3A,$4E,$5B,$97,$8E,$F8,$B6,$B8,$61,$1F,$C8,$E1,$D1,$B5
	.dc.b	$FE,$42,$CF,$A5,$7E,$52,$BD,$B8,$F8,$EB,$AD,$FE,$E6,$79,$4F,$08
	.dc.b	$FA,$86,$71,$F8,$58,$71,$8E,$97,$DF,$1F,$89,$8F,$7B,$DC,$7A,$6B
	.dc.b	$CA,$66,$63,$82,$5F,$84,$F1,$61,$DE,$3C,$63,$1F,$3D,$F8,$23,$31
	.dc.b	$F6,$AB,$81,$27,$4F,$57,$C8,$E6,$7D,$D7,$E4,$BF,$01,$2B,$B3,$C6
	.dc.b	$7E,$46,$3F,$05,$A7,$B2,$FF,$8F,$F2,$7F,$5A,$F9,$3F,$F3,$AE,$3F
	.dc.b	$E0,$0F,$80,$BC,$1F,$DD,$9B,$87,$F4,$2A,$B7,$FD,$8A,$7F,$91,$CF
	.dc.b	$F4,$66,$C6,$63,$FA,$94,$63,$93,$F4,$98,$7F,$95,$9B,$87,$F4,$6C
	.dc.b	$FA,$C7,$F8,$7E,$FF,$A3,$E1,$FD,$52,$7C,$3C,$8F,$92,$A7,$08,$CF
	.dc.b	$C9,$6B,$E4,$A6,$1F,$E1,$68,$AF,$F7,$4C,$3F,$BD,$CF,$A3,$03,$4C
	.dc.b	$7C,$7F,$C2,$D3,$FC,$27,$F1,$57,$C0,$3E,$49,$C3,$F1,$27,$78,$FF
	.dc.b	$48,$6B,$FE,$01,$AE,$2F,$F4,$CC,$5F,$C0,$2F,$8E,$F9,$ED,$9F,$86
	.dc.b	$D7,$FB,$C6,$7F,$EA,$AC,$3F,$D0,$33,$96,$FC,$57,$23,$9D,$F4,$F1
	.dc.b	$FD,$5F,$1B,$31,$EB,$15,$F3,$47,$AF,$33,$4F,$BF,$8F,$FB,$D6,$7C
	.dc.b	$B1,$E8,$3D,$3E,$25,$8F,$4E,$BE,$C3,$0F,$F8,$46,$9F,$1B,$F0,$3E
	.dc.b	$32,$4F,$C0,$8A,$EF,$E3,$7E,$29,$8D,$BE,$0F,$C6,$FE,$62,$5F,$77
	.dc.b	$E8,$7F,$C0,$CD,$7C,$11,$B8,$C4,$7F,$8E,$1A,$FF,$B6,$9A,$FF,$83
	.dc.b	$5B,$C5,$F6,$3E,$27,$EB,$92,$7E,$BF,$C6,$7D,$AF,$8C,$FB,$7F,$1B
	.dc.b	$27,$ED,$C5,$7E,$17,$C7,$7E,$00,$C6,$7F,$73,$E3,$3F,$06,$49,$FC
	.dc.b	$1F,$89,$F8,$6F,$8D,$91,$F8,$67,$4F,$F9,$68,$78,$5F,$E5,$A7,$FC
	.dc.b	$50,$BE,$2F,$E4,$36,$33,$D9,$89,$FD,$FC,$72,$6C,$EE,$3F,$C4,$AD
	.dc.b	$7F,$E0,$0E,$FF,$9C,$F8,$DF,$CD,$92,$3C,$3F,$19,$F9,$EF,$8D,$FC
	.dc.b	$FF,$9D,$27,$F3,$F1,$FA,$26,$1F,$A2,$34,$7F,$96,$7F,$80,$7D,$8D
	.dc.b	$AF,$E2,$F1,$FE,$F3,$8F,$E1,$6E,$B7,$FF,$73,$B9,$8E,$59,$F0,$38
	.dc.b	$D7,$2F,$8E,$93,$C9,$EB,$F1,$DD,$BE,$3A,$4F,$67,$6F,$97,$D5,$F1
	.dc.b	$91,$D1,$E4,$F8,$EF,$39,$1F,$3F,$C7,$7A,$BE,$3A,$47,$A9,$D3,$F0
	.dc.b	$7E,$33,$DE,$47,$DF,$F1,$BF,$0B,$D6,$91,$F0,$9D,$3F,$85,$FA,$BF
	.dc.b	$F6,$23,$FE,$6B,$5F,$0D,$FA,$B6,$C6,$7E,$1F,$13,$F8,$49,$CC,$FF
	.dc.b	$99,$9A,$FF,$A8,$9A,$FF,$B1,$B8,$BF,$C0,$DC,$7F,$4B,$D8,$FA,$E7
	.dc.b	$7F,$86,$FB,$3F,$1B,$2F,$EC,$BF,$73,$E3,$BF,$99,$78,$3F,$C5,$49
	.dc.b	$3F,$CB,$FC,$F4,$7F,$33,$F9,$87,$C5,$7E,$13,$E3,$A5,$FF,$A5,$34
	.dc.b	$56,$A3,$F0,$FF,$19,$F8,$A2,$3F,$8A,$F8,$FF,$E8,$5E,$3F,$F2,$52
	.dc.b	$47,$85,$F0,$7C,$6F,$E4,$BD,$C9,$3D,$F0,$4F,$F6,$CF,$A8,$FA,$33
	.dc.b	$F9,$BF,$84,$FC,$C9,$9E,$7F,$EA,$1E,$27,$F3,$32,$4F,$BD,$FE,$E3
	.dc.b	$4E,$4E,$84,$9F,$ED,$71,$F5,$B6,$7C,$0F,$FB,$E2,$3C,$73,$AC,$BC
	.dc.b	$87,$97,$87,$FB,$17,$B9,$FA,$E9,$FF,$04,$B5,$FF,$7C,$35,$60,$8C
	.dc.b	$79,$2C,$1F,$19,$F0,$49,$7E,$6F,$8E,$FF,$07,$FD,$4F,$E1,$B4,$FF
	.dc.b	$5E,$60,$8C,$57,$FB,$D5,$B1,$D7,$E3,$FF,$A1,$7C,$5F,$F1,$3A,$FF
	.dc.b	$62,$1F,$EC,$FF,$33,$F8,$DC,$66,$3F,$D9,$0D,$D7,$D1,$E6,$7C,$02
	.dc.b	$4F,$CA,$F8,$FF,$E9,$DF,$3C,$F8,$FF,$E4,$AD,$9F,$F1,$2B,$5F,$F6
	.dc.b	$4B,$5F,$18,$F9,$F7,$67,$F2,$E6,$C7,$F8,$B9,$AF,$FB,$49,$AF,$D2
	.dc.b	$F3,$FE,$A9,$1F,$BF,$F2,$3F,$8D,$F9,$1F,$C7,$63,$27,$F0,$71,$FE
	.dc.b	$1D,$ED,$67,$FA,$72,$4F,$F8,$66,$32,$FF,$85,$F9,$9F,$23,$F9,$92
	.dc.b	$47,$E3,$B8,$3F,$39,$F1,$DF,$88,$F4,$E4,$FE,$4A,$2B,$F0,$B8,$BF
	.dc.b	$0D,$3F,$E2,$9E,$A3,$4F,$CE,$FC,$67,$E9,$C9,$5F,$9C,$F6,$2B,$F9
	.dc.b	$0E,$AF,$CE,$7C,$6F,$EB,$4C,$E3,$FA,$09,$FC,$AF,$CB,$3D,$8F,$F1
	.dc.b	$C3,$63,$FC,$70,$D7,$CC,$D7,$FA,$EE,$EF,$8D,$FD,$D7,$C7,$49,$FD
	.dc.b	$7C,$7E,$F7,$2F,$F3,$4F,$C7,$7E,$8E,$7F,$CB,$AD,$7D,$03,$BF,$F4
	.dc.b	$D3,$C3,$83,$8D,$4A,$E3,$15,$FF,$2E,$47,$F9,$15,$FE,$E2,$D7,$F5
	.dc.b	$2D,$7B,$2F,$8E,$F2,$FC,$64,$9F,$2C,$57,$CB,$87,$F9,$57,$E3,$7D
	.dc.b	$66,$77,$F9,$92,$F9,$EF,$F3,$B7,$99,$FE,$86,$6C,$7E,$80,$D7,$17
	.dc.b	$FB,$75,$C4,$69,$FD,$2F,$C9,$3D,$C7,$CF,$7F,$A4,$23,$0D,$8C,$CF
	.dc.b	$59,$CB,$D3,$9D,$F0,$5E,$36,$C6,$63,$07,$F9,$CF,$81,$F1,$BF,$87
	.dc.b	$D7,$F1,$1B,$FF,$9D,$7E,$47,$1C,$9F,$32,$AE,$3F,$32,$6D,$7F,$9E
	.dc.b	$E6,$E1,$D2,$FF,$03,$F1,$F8,$9C,$FC,$87,$CA,$C7,$EB,$B5,$FA,$ED
	.dc.b	$C3,$FD,$0C,$DC,$1F,$1B,$CB,$1F,$C6,$B6,$FE,$CC,$EA,$72,$7E,$4B
	.dc.b	$E4,$3C,$B5,$E6,$F6,$FC,$FF,$1F,$6B,$F2,$ED,$C3,$9C,$FB,$3E,$3B
	.dc.b	$1B,$FC,$7F,$1B,$F8,$2E,$1F,$9E,$F8,$FF,$8B,$3F,$DB,$BC,$7F,$87
	.dc.b	$18,$AD,$E3,$FA,$2A,$D5,$FE,$9B,$E3,$F5,$87,$CE,$AF,$F5,$BF,$C4
	.dc.b	$FC,$56,$D7,$FB,$53,$C2,$4E,$5B,$1F,$77,$DE,$E5,$18,$AF,$8B,$18
	.dc.b	$AF,$AF,$F1,$B2,$FF,$6F,$F1,$F6,$FF,$CD,$98,$3E,$2F,$B1,$2B,$C4
	.dc.b	$FD,$5E,$AF,$C0,$9F,$82,$F7,$EC,$FF,$84,$7E,$75,$F1,$FF,$19,$B1
	.dc.b	$FC,$1B,$1C,$4F,$C1,$F9,$DF,$DC,$4F,$FB,$C9,$B3,$FA,$D6,$BF,$D3
	.dc.b	$9A,$FF,$6A,$F9,$BF,$81,$C5,$F9,$C6,$D7,$F6,$EF,$B0,$7C,$9F,$E6
	.dc.b	$A7,$37,$FD,$F3,$CC,$B8,$7F,$7D,$E2,$FC,$FE,$1F,$E8,$BC,$7F,$C5
	.dc.b	$95,$C2,$DF,$84,$BE,$53,$EC,$36,$BF,$82,$4E,$A7,$27,$E8,$71,$7C
	.dc.b	$F3,$DB,$FC,$5B,$6D,$7F,$C5,$B2,$7C,$06,$AF,$CB,$EE,$FE,$4B,$4F
	.dc.b	$B1,$87,$0C,$BF,$7B,$82,$B5,$CB,$F8,$FF,$B2,$79,$FF,$8F,$6E,$1E
	.dc.b	$86,$E9,$FC,$75,$AF,$FA,$8B,$0F,$E9,$D3,$FE,$4F,$FA,$7F,$F7,$F3
	.dc.b	$E1,$3E,$87,$B1,$FA,$68,$CD,$7A,$F9,$BE,$6F,$B5,$F3,$9B,$7C,$67
	.dc.b	$E0,$E2,$F3,$BC,$C4,$9F,$33,$FC,$27,$2F,$A1,$5F,$1F,$8F,$EC,$FC
	.dc.b	$BF,$DA,$7C,$5E,$5F,$DC,$E7,$B8,$F7,$F3,$18,$FD,$DF,$9C,$7B,$DF
	.dc.b	$B3,$6D,$7D,$C6,$E7,$FD,$05,$AF,$F8,$D3,$95,$C2,$F2,$FE,$11,$F8
	.dc.b	$CE,$5B,$87,$C9,$F2,$FF,$1F,$E9,$FC,$76,$C7,$C3,$C5,$97,$F0,$C7
	.dc.b	$FC,$5E,$D7,$FC,$86,$D7,$F1,$2D,$CB,$C7,$E3,$FD,$33,$F3,$BF,$31
	.dc.b	$6F,$E1,$6C,$F9,$5F,$D0,$9F,$C5,$7E,$81,$78,$AF,$24,$3C,$9F,$E3
	.dc.b	$0D,$8F,$D0,$B7,$3F,$EC,$CD,$7F,$C8,$EB,$CA,$ED,$5D,$DB,$5E,$4F
	.dc.b	$93,$FD,$9F,$CA,$FE,$D1,$B1,$92,$61,$FA,$7F,$96,$D4,$7B,$3E,$27
	.dc.b	$F7,$67,$EF,$3E,$56,$C7,$C0,$6D,$7E,$BD,$BC,$1F,$AA,$F7,$BF,$81
	.dc.b	$FC,$AF,$F0,$46,$DF,$E2,$7E,$C5,$F2,$D9,$8C,$37,$1F,$F0,$7C,$FF
	.dc.b	$B4,$A7,$27,$1F,$9A,$FE,$DB,$1C,$A3,$FD,$75,$F4,$DF,$B9,$7B,$7C
	.dc.b	$C7,$F5,$08,$C5,$F4,$8E,$3F,$2F,$9F,$2E,$DF,$27,$AF,$CB,$76,$F9
	.dc.b	$8B,$3E,$97,$C3,$F2,$D2,$3D,$9F,$31,$72,$F4,$FC,$97,$98,$FE,$B1
	.dc.b	$C1,$E7,$8C,$D7,$E5,$E3,$DD,$1F,$E2,$56,$BF,$E8,$66,$BF,$E9,$A6
	.dc.b	$BF,$C8,$1F,$F6,$04,$73,$DF,$BE,$BE,$07,$E3,$FA,$52,$7E,$97,$CA
	.dc.b	$5C,$3E,$7B,$79,$3F,$C3,$FC,$5F,$E9,$DA,$FB,$17,$FD,$C7,$EC,$BF
	.dc.b	$98,$61,$F5,$BE,$57,$13,$F5,$1F,$AB,$E4,$E2,$FF,$00,$DA,$FB,$BF
	.dc.b	$2B,$F6,$8F,$B7,$F0,$36,$BE,$27,$CA,$FF,$7D,$E0,$34,$91,$FE,$4B
	.dc.b	$F3,$5A,$EA,$38,$9F,$8F,$9F,$A2,$7E,$0F,$CE,$FC,$4F,$AF,$67,$F2
	.dc.b	$8F,$E3,$3D,$3C,$69,$F9,$26,$F0,$FF,$90,$FB,$3F,$CA,$97,$F9,$BF
	.dc.b	$82,$FC,$77,$9D,$F9,$47,$F2,$BF,$27,$89,$FD,$03,$71,$FF,$0F,$5F
	.dc.b	$F0,$99,$FF,$32,$60,$8C,$4F,$FA,$95,$AF,$F2,$76,$BF,$EE,$4D,$7F
	.dc.b	$DF,$75,$FA,$2F,$2F,$F4,$AF,$E8,$5F,$D1,$78,$BF,$E5,$4F,$CE,$BF
	.dc.b	$DE,$93,$FA,$88,$C7,$B7,$A8,$C1,$FC,$FE,$7F,$A9,$E5,$F0,$6B,$F5
	.dc.b	$9E,$3F,$F7,$5C,$FF,$DB,$37,$EA,$6E,$7E,$99,$FF,$2C,$61,$FE,$16
	.dc.b	$C3,$E5,$47,$8F,$E4,$F9,$E1,$DF,$E4,$F6,$47,$EF,$5A,$39,$FC,$A7
	.dc.b	$AD,$F6,$36,$7D,$1F,$1F,$EF,$F9,$1F,$80,$D8,$C9,$33,$FF,$4E,$E0
	.dc.b	$F9,$1D,$CF,$C1,$F6,$FD,$24,$7C,$6F,$B3,$E4,$3E,$63,$F3,$5B,$1F
	.dc.b	$D4,$EF,$FA,$68,$FA,$17,$C6,$FF,$86,$23,$FC,$33,$F1,$BF,$ED,$59
	.dc.b	$CC,$7F,$58,$38,$4E,$DF,$F4,$DD,$7F,$88,$27,$71,$FD,$70,$7A,$38
	.dc.b	$76,$62,$BF,$B9,$34,$F1,$5F,$FB,$96,$78,$47,$F6,$A3,$C2,$6A,$FC
	.dc.b	$15,$CA,$FB,$5F,$A6,$30,$79,$DC,$57,$AE,$77,$5C,$A3,$DD,$78,$7C
	.dc.b	$87,$AD,$C1,$E4,$F6,$F8,$47,$C8,$9A,$FC,$EA,$E4,$6D,$F4,$9B,$29
	.dc.b	$BC,$3C,$D3,$CA,$3E,$04,$F2,$9F,$5C,$7C,$13,$89,$F1,$6F,$CB,$1F
	.dc.b	$31,$B1,$EF,$8E,$AF,$0B,$F3,$E1,$F1,$23,$DD,$3A,$AF,$41,$CA,$32
	.dc.b	$78,$5C,$C6,$CF,$B7,$E0,$FE,$02,$7F,$10,$D7,$F0,$8D,$7F,$26,$E7
	.dc.b	$E4,$DE,$E3,$9B,$FA,$66,$1F,$95,$9C,$C6,$2F,$FC,$DF,$18,$57,$CE
	.dc.b	$F9,$46,$2F,$94,$61,$FC,$BD,$7F,$8A,$67,$FC,$F3,$18,$EA,$FD,$23
	.dc.b	$AA,$E1,$8F,$2C,$76,$F9,$6B,$6D,$BC,$9E,$6A,$E2,$DA,$FC,$4E,$1E
	.dc.b	$18,$FC,$2B,$74,$F0,$4E,$AB,$8D,$2B,$86,$3E,$78,$EE,$66,$77,$3F
	.dc.b	$16,$F5,$3D,$B1,$FE,$01,$1D,$A3,$C3,$5F,$22,$36,$4C,$7E,$04,$EB
	.dc.b	$1F,$4D,$F2,$9F,$81,$39,$26,$3E,$02,$47,$27,$C1,$7D,$67,$ED,$D7
	.dc.b	$37,$E3,$1C,$E3,$4F,$C1,$8D,$3E,$FB,$F8,$C7,$C2,$C7,$E2,$E1,$E4
	.dc.b	$C3,$DC,$78,$1F,$3E,$1D,$27,$E3,$47,$FA,$EB,$C3,$71,$3E,$4B,$FD
	.dc.b	$0D,$C1,$98,$FF,$43,$67,$FD,$57,$E0,$32,$AF,$8D,$E1,$E7,$3F,$EC
	.dc.b	$E6,$CF,$F4,$C8,$C5,$74,$8F,$E5,$D3,$E2,$F9,$1E,$52,$69,$8F,$92
	.dc.b	$B5,$FD,$9F,$E0,$76,$61,$FE,$9B,$9F,$2F,$80,$91,$FD,$B6,$2B,$E0
	.dc.b	$C3,$FD,$12,$FF,$60,$9F,$29,$E6,$F7,$FD,$FC,$3F,$E0,$7E,$4B,$E0
	.dc.b	$7E,$CA,$FF,$BC,$56,$27,$FD,$DB,$5B,$9F,$EE,$D1,$8A,$E1,$1F,$D3
	.dc.b	$67,$FC,$EB,$E9,$76,$F6,$CD,$3E,$1F,$BF,$6B,$FC,$0D,$F2,$3F,$18
	.dc.b	$C3,$FD,$B3,$AB,$97,$BB,$23,$FC,$19,$15,$F0,$60,$9F,$3B,$63,$FC
	.dc.b	$4A,$D7,$E3,$1E,$5F,$6E,$3F,$E1,$72,$9F,$C6,$F9,$FF,$19,$93,$E3
	.dc.b	$FF,$08,$D9,$FF,$75,$69,$FE,$94,$BC,$FC,$D7,$C1,$F7,$25,$7F,$B8
	.dc.b	$6D,$38,$8E,$C7,$E4,$61,$F1,$BD,$8D,$60,$CB,$C7,$F0,$1F,$98,$B7
	.dc.b	$E5,$0F,$47,$0D,$8F,$04,$E2,$7F,$C0,$51,$FE,$09,$9D,$38,$6B,$FE
	.dc.b	$07,$6B,$F0,$BE,$37,$E0,$91,$F6,$7C,$77,$C4,$F8,$EF,$26,$9F,$C9
	.dc.b	$73,$F4,$50,$FC,$1F,$8B,$B7,$FD,$E1,$3B,$D3,$E7,$FC,$6F,$CE,$39
	.dc.b	$D7,$D0,$F8,$DF,$A3,$ED,$4C,$3E,$64,$57,$E5,$F8,$7F,$90,$9C,$65
	.dc.b	$E2,$8F,$C9,$D7,$C9,$FB,$4F,$E0,$D7,$FF,$00,$5F,$DD,$57,$A2,$11
	.dc.b	$8C,$BF,$C6,$0F,$08,$FB,$26,$BC,$5F,$9A,$FB,$E3,$F8,$23,$FE,$31
	.dc.b	$6B,$CA,$3F,$86,$68,$F8,$DF,$C6,$92,$7C,$9E,$2F,$E4,$3C,$B9,$1F
	.dc.b	$8B,$7F,$27,$ED,$7E,$53,$E2,$FF,$2B,$C1,$2B,$F2,$11,$F9,$8F,$63
	.dc.b	$F6,$9F,$03,$27,$F2,$91,$5F,$93,$9B,$E2,$FF,$3E,$67,$83,$E5,$43
	.dc.b	$DA,$FD,$1F,$C5,$CB,$FC,$F4,$57,$9F,$91,$FA,$83,$17,$C9,$FD,$5F
	.dc.b	$A5,$27,$F4,$D1,$FC,$C7,$ED,$3F,$DF,$F1,$F8,$9A,$FB,$51,$FC,$32
	.dc.b	$31,$CB,$FD,$06,$31,$1F,$82,$78,$B7,$2E,$84,$78,$9F,$D0,$63,$75
	.dc.b	$DB,$C3,$F1,$12,$77,$D0,$A7,$C4,$E6,$7C,$1E,$3F,$A4,$C6,$1F,$8B
	.dc.b	$F8,$37,$0F,$EE,$5F,$9D,$5E,$23,$FD,$AF,$C3,$99,$D3,$F8,$C8,$C9
	.dc.b	$F8,$E9,$E2,$FE,$F1,$AF,$FA,$17,$2F,$37,$C6,$7C,$B2,$47,$79,$FB
	.dc.b	$4F,$F1,$AF,$A9,$7C,$0F,$F4,$66,$C6,$63,$EA,$79,$3F,$C4,$72,$FD
	.dc.b	$34,$FF,$A3,$5A,$FF,$A5,$5A,$FF,$01,$6E,$6B,$FF,$5A,$4F,$CC,$D9
	.dc.b	$F1,$BF,$81,$3C,$5F,$03,$E3,$F8,$09,$5E,$77,$F0,$DF,$1B,$FC,$8F
	.dc.b	$ED,$3C,$F3,$FC,$79,$AF,$FA,$85,$AF,$FA,$A5,$AF,$F9,$67,$17,$C0
	.dc.b	$7E,$47,$E3,$24,$FC,$58,$FC,$A7,$C6,$48,$F4,$3F,$97,$F8,$EF,$E6
	.dc.b	$9F,$68,$F8,$CF,$B0,$DA,$FE,$0A,$D7,$F9,$9B,$5F,$B1,$BB,$E3,$7F
	.dc.b	$45,$E1,$49,$F8,$D1,$FA,$5F,$8E,$FA,$DF,$01,$27,$E5,$47,$8B,$E3
	.dc.b	$BC,$24,$8F,$0B,$FE,$1B,$C1,$D3,$FE,$04,$F9,$97,$4D,$3F,$65,$1F
	.dc.b	$D3,$F3,$7A,$F4,$98,$FD,$9D,$BF,$AA,$65,$A8,$FE,$46,$D7,$FA,$8B
	.dc.b	$5F,$F6,$A3,$5F,$DF,$63,$F8,$27,$FB,$6B,$5F,$EC,$FD,$28,$C3,$FC
	.dc.b	$12,$7F,$C3,$BF,$5B,$FE,$DC,$8F,$EA,$21,$C7,$89,$FC,$2D,$B7,$FE
	.dc.b	$99,$7F,$C6,$CF,$EC,$FE,$85,$C7,$FB,$16,$C7,$F3,$E6,$BF,$DB,$1A
	.dc.b	$FF,$BB,$FE,$87,$F5,$AF,$F4,$66,$E1,$FD,$35,$B8,$7F,$1D,$FB,$77
	.dc.b	$99,$FC,$87,$E4,$E6,$1F,$A5,$7F,$92,$7B,$9F,$E0,$1F,$B4,$FF,$05
	.dc.b	$CF,$F5,$06,$BF,$DF,$9A,$FF,$C0,$16,$BF,$E6,$7F,$B4,$FD,$3E,$7F
	.dc.b	$D9,$5B,$96,$7C,$5F,$E6,$4F,$F9,$8E,$1F,$41,$E1,$F8,$09,$87,$F2
	.dc.b	$49,$CB,$FC,$6F,$1F,$F0,$DB,$E5,$9F,$E3,$51,$88,$F3,$47,$F9,$3B
	.dc.b	$E9,$31,$97,$D8,$6E,$3F,$D0,$5B,$5F,$E7,$7C,$1F,$2D,$FE,$5C,$9C
	.dc.b	$F9,$69,$FE,$1B,$5F,$E0,$46,$E1,$D7,$1F,$67,$CA,$4A,$F5,$3E,$FF
	.dc.b	$96,$C4,$71,$F2,$7E,$11,$8B,$F9,$1E,$2F,$C5,$24,$FC,$3F,$96,$AF
	.dc.b	$87,$E5,$BE,$51,$8A,$FA,$0F,$FA,$37,$33,$49,$3F,$E3,$8D,$3D,$9E
	.dc.b	$E7,$D3,$24,$7D,$1F,$93,$AF,$BF,$DF,$FB,$06,$31,$FF,$0E,$7D,$1F
	.dc.b	$F1,$5F,$67,$FC,$2E,$DA,$FF,$22,$B5,$FF,$1C,$1F,$E3,$97,$0F,$FB
	.dc.b	$F7,$DE,$7D,$1F,$F8,$83,$E5,$91,$89,$FF,$27,$C6,$27,$53,$99,$EA
	.dc.b	$29,$E1,$BE,$3B,$CA,$D9,$F2,$B5,$E3,$3C,$27,$E3,$70,$3C,$DF,$E0
	.dc.b	$3D,$5F,$9F,$F9,$5B,$87,$E2,$66,$F1,$7E,$1E,$6D,$F1,$BF,$CE,$56
	.dc.b	$3E,$5B,$2E,$BE,$57,$FA,$FE,$1F,$8F,$6D,$9A,$63,$E4,$F8,$61,$FA
	.dc.b	$76,$E5,$F9,$57,$C2,$78,$BE,$56,$CF,$F2,$2F,$3B,$CB,$F2,$9E,$AE
	.dc.b	$5F,$E4,$51,$89,$D1,$30,$FD,$06,$FB,$F3,$55,$F0,$3E,$87,$FD,$27
	.dc.b	$7F,$B9,$BC,$7C,$9F,$C2,$6C,$7F,$88,$27,$71,$FE,$6E,$7F,$D4,$91
	.dc.b	$99,$E9,$E1,$7C,$49,$FF,$4D,$CF,$E6,$5B,$5F,$EE,$66,$BF,$E1,$77
	.dc.b	$FA,$6C,$77,$9F,$37,$A3,$D2,$3C,$5F,$25,$E9,$F9,$2F,$53,$63,$CD
	.dc.b	$F2,$5E,$12,$3E,$17,$C9,$F2,$1F,$04,$F8,$5F,$25,$67,$FC,$65,$F4
	.dc.b	$5F,$C3,$1F,$F3,$DB,$74,$FF,$54,$B5,$FC,$CB,$5F,$F4,$F7,$3B,$57
	.dc.b	$C9,$FC,$32,$5F,$C3,$7F,$C4,$FC,$6F,$A1,$FF,$6D,$4F,$FB,$31,$B8
	.dc.b	$7C,$FF,$93,$FC,$04,$62,$BF,$AE,$AF,$85,$C9,$F7,$7C,$1B,$87,$D9
	.dc.b	$F9,$3F,$A8,$47,$EA,$3F,$5F,$D6,$FC,$21,$F8,$5F,$93,$B3,$F8,$1F
	.dc.b	$8A,$FC,$53,$F8,$96,$BF,$85,$F8,$CF,$C0,$11,$FC,$03,$F8,$3F,$93
	.dc.b	$FC,$87,$C9,$FE,$45,$B3,$F8,$DF,$2F,$F2,$9F,$27,$F9,$56,$C7,$F9
	.dc.b	$C3,$B5,$F2,$FF,$B0,$6C,$7F,$25,$6F,$1F,$F0,$CE,$6F,$F7,$7D,$6F
	.dc.b	$E4,$BF,$7F,$5F,$E7,$8B,$E9,$F2,$98,$9F,$D6,$7D,$07,$F3,$29,$F0
	.dc.b	$7C,$A6,$3D,$5E,$2F,$2F,$95,$F1,$12,$7C,$4F,$97,$E2,$BB,$9E,$0F
	.dc.b	$95,$B3,$FE,$90,$EB,$62,$F4,$3C,$BF,$2B,$E2,$6D,$7C,$1E,$CF,$80
	.dc.b	$47,$E0,$3F,$0F,$E2,$7D,$A7,$BB,$E4,$AC,$FF,$A3,$7E,$BB,$F6,$FC
	.dc.b	$BF,$EE,$6E,$37,$C9,$FF,$24,$84,$EB,$5F,$F7,$45,$6A,$F9,$3F,$E7
	.dc.b	$C9,$FE,$07,$5F,$CE,$9A,$FF,$48,$6B,$FF,$00,$5F,$E0,$35,$FE,$ED
	.dc.b	$7F,$85,$B7,$CC,$F1,$46,$2F,$F5,$76,$F9,$1F,$B3,$E7,$7D,$A6,$CF
	.dc.b	$E1,$5C,$3F,$87,$F9,$2B,$97,$BF,$E4,$3F,$03,$F2,$3F,$82,$6D,$7E
	.dc.b	$37,$DC,$C3,$4F,$C8,$47,$E2,$3E,$32,$47,$D9,$F9,$AF,$5B,$E6,$8F
	.dc.b	$E6,$91,$FC,$73,$0F,$F6,$C3,$56,$E8,$C3,$77,$FC,$E7,$C7,$72,$7F
	.dc.b	$2E,$DC,$3F,$DA,$3F,$92,$FF,$AD,$9F,$D2,$1F,$D6,$9C,$61,$F7,$63
	.dc.b	$1F,$1D,$9F,$8F,$B1,$C9,$E1,$F1,$D2,$3A,$FC,$C5,$BF,$3C,$7F,$1C
	.dc.b	$C1,$F1,$BD,$EB,$C0,$D9,$FF,$6D,$78,$FF,$D3,$9F,$3F,$C6,$FE,$36
	.dc.b	$31,$87,$41,$E4,$31,$56,$F9,$7F,$D9,$37,$FE,$A3,$C3,$FC,$0F,$80
	.dc.b	$46,$5C,$B5,$FF,$06,$BF,$89,$8F,$2C,$F8,$F9,$BB,$47,$CC,$F8,$CF
	.dc.b	$95,$F1,$DF,$2D,$B1,$F4,$3E,$3B,$0D,$38,$1F,$53,$E3,$7E,$97,$93
	.dc.b	$F4,$DB,$1F,$63,$E3,$70,$D3,$81,$F7,$3E,$2B,$ED,$7C,$77,$DB,$6C
	.dc.b	$7B,$FA,$4F,$F8,$D3,$EB,$BF,$D2,$AF,$C4,$E0,$46,$1F,$8E,$D7,$FC
	.dc.b	$5E,$D7,$EE,$78,$FF,$C0,$63,$85,$76,$9F,$AB,$C7,$FC,$56,$FC,$47
	.dc.b	$87,$C6,$FC,$87,$C5,$7E,$47,$E3,$2C,$FE,$59,$FC,$A7,$C5,$63,$0F
	.dc.b	$C0,$74,$7D,$F8,$C6,$5D,$E3,$F0,$5F,$15,$F9,$BF,$8B,$FC,$E3,$71
	.dc.b	$FD,$0F,$C5,$E1,$A7,$E9,$23,$F4,$58,$FF,$8D,$35,$FC,$34,$FF,$9A
	.dc.b	$5B,$1F,$D1,$5A,$FF,$97,$B0,$7D,$77,$F2,$1A,$FD,$04,$FF,$91,$A3
	.dc.b	$0F,$C4,$31,$F1,$79,$F2,$ED,$7C,$69,$FC,$ED,$7F,$1B,$6D,$69,$F8
	.dc.b	$71,$FC,$29,$CC,$F1,$BF,$1F,$C5,$FF,$93,$A8,$D2,$4F,$FA,$16,$1F
	.dc.b	$17,$E7,$9F,$43,$67,$A4,$BE,$2F,$A1,$23,$93,$DF,$E2,$BE,$03,$F0
	.dc.b	$5B,$1C,$BE,$1F,$E2,$7C,$5F,$C5,$6C,$64,$99,$FB,$BE,$27,$51,$E2
	.dc.b	$F8,$8F,$61,$1F,$53,$EF,$F8,$8F,$A0,$FD,$16,$C7,$CD,$AF,$F4,$AF
	.dc.b	$D0,$7F,$03,$FA,$0F,$F4,$04,$7F,$A6,$39,$7F,$35,$18,$8D,$87,$0F
	.dc.b	$FA,$3E,$B9,$74,$FF,$0E,$BF,$E1,$DF,$09,$AC,$3F,$48,$F0,$B4,$EA
	.dc.b	$BF,$D3,$75,$A7,$94,$F7,$73,$7D,$A3,$FD,$43,$55,$FE,$39,$5E,$8A
	.dc.b	$E3,$87,$86,$F5,$5F,$C8,$6E,$DE,$E3,$98,$C3,$83,$70,$F1,$C6,$0F
	.dc.b	$55,$72,$BE,$B5,$EC,$3D,$07,$CB,$73,$1B,$8C,$9F,$33,$0E,$E7,$47
	.dc.b	$27,$CD,$29,$D5,$DD,$F9,$0F,$8B,$1E,$73,$D7,$7E,$78,$F8,$66,$A7
	.dc.b	$B1,$CA,$77,$1C,$A7,$C4,$71,$8C,$9E,$B3,$B1,$F8,$47,$4F,$E1,$BC
	.dc.b	$9F,$94,$FC,$C6,$B9,$8F,$BB,$18,$F9,$26,$7F,$DF,$AB,$FA,$AD,$7F
	.dc.b	$BA,$A7,$3D,$7F,$D6,$6F,$85,$ED,$B1,$9A,$E1,$3C,$DF,$CD,$53,$E2
	.dc.b	$BF,$DE,$B1,$86,$E3,$C2,$FF,$29,$8F,$F6,$7B,$F0,$BF,$01,$E4,$E8
	.dc.b	$38,$75,$87,$86,$BA,$47,$E0,$27,$59,$78,$63,$F3,$51,$DA,$3C,$F5
	.dc.b	$C6,$7C,$58,$79,$9A,$78,$6B,$CB,$8F,$0C,$79,$D7,$F7,$FA,$F0,$37
	.dc.b	$0D,$47,$8F,$0F,$A1,$1E,$38,$D9,$E4,$8C,$E3,$EA,$9D,$11,$C1,$EE
	.dc.b	$C3,$91,$1C,$1A,$AF,$49,$D8,$F1,$61,$DC,$F4,$B0,$C9,$27,$A3,$C9
	.dc.b	$F5,$9E,$CA,$F2,$47,$67,$15,$FE,$CE,$E4,$FA,$D7,$FD,$A5,$7C,$5F
	.dc.b	$C7,$8C,$37,$6E,$19,$DC,$7F,$AA,$CF,$F8,$27,$A3,$FB,$0B,$70,$E7
	.dc.b	$3F,$D3,$63,$1F,$13,$C6,$BF,$B3,$B6,$8D,$F8,$BC,$39,$F8,$B9,$87
	.dc.b	$F5,$28,$E1,$F1,$7C,$77,$93,$FD,$5A,$2B,$FE,$BB,$C3,$F6,$2F,$F7
	.dc.b	$76,$BF,$0E,$7D,$1C,$5E,$5F,$D2,$3F,$D8,$F3,$FD,$AD,$AF,$F7,$08
	.dc.b	$D7,$C6,$7C,$19,$CC,$FF,$80,$DA,$FF,$BF,$1A,$FF,$96,$17,$C5,$77
	.dc.b	$24,$77,$F8,$CA,$FF,$88,$25,$F1,$9E,$53,$3C,$7F,$E2,$18,$7C,$67
	.dc.b	$A7,$E2,$E5,$FA,$62,$BF,$ED,$D9,$4F,$A4,$91,$FE,$66,$E2,$4B,$E3
	.dc.b	$3E,$27,$C5,$4A,$F8,$91,$5F,$37,$C5,$FC,$93,$19,$79,$7E,$2F,$E6
	.dc.b	$7C,$5C,$AF,$99,$15,$F7,$7A,$1F,$44,$C5,$7B,$7E,$2B,$F2,$12,$FA
	.dc.b	$EF,$F4,$5B,$FD,$ED,$AF,$C7,$8D,$C6,$23,$F8,$11,$C7,$E0,$3F,$CC
	.dc.b	$0D,$3F,$CC,$51,$B7,$EC,$7B,$DF,$83,$F7,$E4,$7E,$0E,$3F,$0D,$F1
	.dc.b	$7F,$87,$F8,$B9,$3F,$87,$8A,$FD,$9F,$8B,$FC,$61,$8C,$FE,$C7,$C5
	.dc.b	$FE,$3F,$C4,$95,$F8,$F8,$AF,$E0,$7E,$27,$F2,$86,$2B,$FC,$DD,$0F
	.dc.b	$A0,$4F,$C6,$73,$3A,$6D,$FF,$23,$77,$5F,$E2,$37,$9B,$F1,$9A,$EE
	.dc.b	$3E,$3C,$62,$3F,$D5,$5F,$40,$F8,$AF,$DE,$7C,$55,$9F,$F2,$93,$5F
	.dc.b	$F5,$C0,$9C,$61,$FE,$79,$97,$C5,$F2,$65,$72,$8E,$BF,$17,$23,$AB
	.dc.b	$A7,$FD,$0B,$3F,$82,$AE,$9F,$15,$E4,$3C,$BC,$96,$BC,$BF,$17,$E8
	.dc.b	$30,$F2,$F8,$BF,$E3,$D6,$F1,$5C,$58,$6D,$7B,$AB,$13,$FE,$AB,$AE
	.dc.b	$0D,$AD,$D7,$E4,$A7,$2F,$EB,$F2,$7C,$5F,$CA,$24,$7F,$A0,$E7,$6F
	.dc.b	$BB,$E2,$FE,$7F,$C5,$48,$F9,$F1,$5F,$87,$F1,$7F,$50,$C5,$7C,$FF
	.dc.b	$89,$FA,$E4,$9F,$AF,$F1,$75,$F9,$7F,$17,$F7,$0C,$57,$E4,$3A,$9E
	.dc.b	$EF,$ED,$17,$BD,$4C,$45,$BF,$F6,$74,$E9,$B3,$FB,$28,$C4,$7F,$BD
	.dc.b	$97,$BD,$F8,$CF,$66,$4F,$F9,$E6,$72,$FE,$3B,$E2,$7F,$22,$47,$F2
	.dc.b	$3F,$15,$F9,$4F,$8A,$FE,$AD,$C6,$69,$27,$F8,$28,$D1,$F3,$FF,$8B
	.dc.b	$F0,$12,$F3,$D8,$3E,$2F,$F4,$27,$E2,$2F,$85,$6E,$7F,$01,$7C,$AF
	.dc.b	$8C,$FE,$A0,$FD,$4F,$C1,$FF,$65,$FA,$B7,$D1,$7F,$6D,$6C,$FF,$35
	.dc.b	$6B,$FC,$72,$91,$88,$FE,$F5,$8B,$A3,$A7,$C6,$49,$FC,$64,$E5,$FF
	.dc.b	$64,$53,$E3,$3C,$44,$8C,$FC,$7F,$97,$E3,$7C,$84,$8F,$23,$E7,$F8
	.dc.b	$CF,$57,$C6,$C9,$EB,$1E,$DF,$8C,$F7,$7C,$6C,$9F,$0C,$7C,$1E,$CF
	.dc.b	$80,$48,$F8,$0F,$C3,$EC,$F9,$DC,$B2,$7D,$31,$F2,$7E,$2E,$47,$B1
	.dc.b	$F9,$9D,$BF,$37,$E2,$BE,$61,$27,$E6,$3F,$D1,$3D,$67,$1F,$F5,$06
	.dc.b	$CF,$FB,$25,$AF,$E1,$FA,$3E,$E7,$1F,$F8,$82,$3F,$8F,$FC,$EF,$FA
	.dc.b	$03,$AF,$F9,$BB,$6B,$FC,$A8,$D9,$FC,$65,$3E,$2B,$F8,$8F,$CF,$FF
	.dc.b	$B1,$E7,$F2,$8D,$C0,$D3,$50,$F9,$F7,$B7,$FD,$79,$B1,$F9,$E3,$51
	.dc.b	$93,$F1,$DF,$17,$F9,$12,$47,$E1,$3E,$33,$F2,$9F,$19,$F9,$32,$47
	.dc.b	$E4,$DF,$CB,$7C,$57,$E6,$7E,$2E,$4F,$E3,$23,$FC,$81,$F6,$9F,$E2
	.dc.b	$AD,$1C,$1F,$A3,$87,$C5,$E8,$DF,$A5,$70,$FF,$0A,$4E,$9B,$3F,$E5
	.dc.b	$18,$C4,$7E,$8D,$FF,$11,$BF,$B8,$BE,$EF,$EE,$63,$2F,$D9,$31,$F1
	.dc.b	$5E,$6F,$8B,$F3,$BE,$8F,$8C,$93,$CF,$2F,$44,$FF,$46,$F9,$E7,$85
	.dc.b	$49,$5B,$3C,$1F,$1B,$87,$C1,$E4,$B3,$FF,$17,$46,$E3,$1F,$17,$BD
	.dc.b	$7D,$31,$BF,$8C,$F9,$64,$AE,$16,$F8,$CC,$47,$68,$7C,$5F,$D2,$33
	.dc.b	$87,$48,$CC,$FC,$E9,$F9,$84,$7E,$8F,$77,$DA,$23,$E0,$E9,$AF,$7F
	.dc.b	$8C,$FB,$E6,$30,$FF,$0B,$FC,$EF,$F6,$AE,$7F,$C9,$36,$BF,$98,$B5
	.dc.b	$FF,$1B,$B5,$FF,$71,$78,$6F,$9E,$FC,$D7,$C9,$FF,$46,$6C,$66,$7F
	.dc.b	$9D,$CE,$5E,$BF,$18,$F5,$3C,$2D,$8C,$C6,$0F,$C7,$FC,$6F,$E2,$33
	.dc.b	$F8,$DF,$1D,$7A,$BF,$31,$36,$FF,$19,$37,$95,$F8,$43,$DA,$FF,$15
	.dc.b	$C6,$1B,$19,$C3,$FC,$E0,$DB,$FC,$D7,$B9,$C3,$98,$D2,$57,$E9,$3E
	.dc.b	$13,$C1,$3A,$8F,$C1,$C6,$27,$C9,$F6,$FD,$6F,$FD,$C9,$92,$31,$1F
	.dc.b	$D5,$5A,$FF,$5F,$6C,$7B,$3D,$EE,$FC,$1E,$FC,$D1,$8C,$BF,$35,$18
	.dc.b	$6C,$7C,$5F,$D0,$1F,$13,$C6,$39,$7C,$07,$AB,$E4,$6D,$FE,$71,$BA
	.dc.b	$7E,$A3,$1F,$EF,$53,$FE,$25,$FB,$47,$1E,$B9,$3F,$CC,$BE,$F6,$2B
	.dc.b	$F9,$03,$70,$FE,$FE,$D7,$FD,$0A,$D7,$FB,$F8,$FB,$A7,$85,$F7,$E3
	.dc.b	$FB,$36,$16,$73,$7F,$BF,$9C,$B7,$EF,$BF,$CE,$AD,$F9,$4E,$58,$B8
	.dc.b	$BA,$56,$AF,$FC,$CF,$FA,$27,$DB,$7F,$C2,$AD,$7F,$D3,$6D,$7C,$ED
	.dc.b	$7F,$C2,$FF,$47,$FE,$16,$C7,$FA,$03,$63,$FC,$3E,$D7,$FD,$4A,$D7
	.dc.b	$F9,$A4,$7E,$30,$FC,$6E,$35,$FE,$2D,$F8,$7F,$FA,$05,$FF,$64,$36
	.dc.b	$7F,$C2,$8D,$7F,$1E,$D7,$FC,$44,$FF,$A0,$23,$9F,$CA,$7E,$2F,$BB
	.dc.b	$F5,$53,$9A,$FE,$4E,$BE,$53,$26,$BE,$2A,$E1,$E4,$E8,$FD,$7D,$7E
	.dc.b	$3A,$3C,$B1,$83,$3F,$27,$D8,$FD,$D7,$23,$ED,$3F,$1B,$97,$82,$7F
	.dc.b	$01,$F3,$DF,$D6,$30,$FB,$CD,$9F,$F4,$EF,$E0,$5F,$07,$0F,$F5,$AC
	.dc.b	$E5,$FE,$73,$93,$E5,$FC,$A4,$C3,$E4,$BF,$EA,$DB,$7C,$AF,$E3,$DB
	.dc.b	$3F,$8D,$5F,$27,$99,$F8,$FF,$2F,$F9,$5F,$C7,$9E,$EE,$30,$FC,$A4
	.dc.b	$61,$FB,$3F,$31,$F1,$88,$FD,$A7,$FD,$7D,$F8,$CB,$E0,$BF,$40,$D9
	.dc.b	$FE,$AD,$F9,$CB,$E4,$F3,$5F,$63,$E6,$BE,$CF,$CB,$5B,$FC,$3B,$F7
	.dc.b	$35,$FA,$A4,$8F,$C4,$BF,$63,$6F,$C0,$9F,$53,$E5,$FF,$05,$C7,$F2
	.dc.b	$A3,$F0,$BF,$35,$F8,$6F,$9A,$B8,$7E,$51,$FC,$4F,$CD,$7E,$08,$91
	.dc.b	$F9,$67,$F0,$BE,$A7,$E3,$8E,$DC,$D6,$7F,$3B,$E6,$FE,$48,$FC,$9F
	.dc.b	$9D,$63,$E0,$FC,$C7,$F3,$CE,$33,$49,$1F,$E8,$CF,$45,$3F,$E0,$68
	.dc.b	$E2,$7C,$AF,$83,$FE,$91,$1C,$A3,$F5,$BF,$2B,$FA,$23,$3F,$31,$63
	.dc.b	$E8,$B8,$7F,$C1,$1F,$C0,$2F,$8E,$C7,$85,$FA,$83,$F5,$7E,$47,$EB
	.dc.b	$1B,$5F,$6B,$E6,$70,$D3,$F7,$B1,$FD,$6F,$E3,$3F,$D2,$71,$FE,$15
	.dc.b	$6C,$7F,$86,$B3,$FF,$61,$CE,$7A,$3D,$7F,$33,$23,$F0,$FF,$33,$6F
	.dc.b	$83,$F0,$0F,$ED,$1F,$A2,$7F,$90,$A7,$67,$7F,$98,$F1,$1F,$1B,$E0
	.dc.b	$EC,$7E,$49,$C3,$FE,$E7,$C5,$D3,$8C,$7F,$C5,$FA,$7E,$5E,$BD,$5E
	.dc.b	$17,$D0,$F9,$AF,$73,$63,$F3,$B1,$F5,$3E,$6B,$17,$FE,$47,$1F,$B8
	.dc.b	$FF,$35,$F0,$7F,$C9,$B1,$8F,$82,$FE,$8D,$D5,$FC,$5F,$1F,$E8,$97
	.dc.b	$B7,$FB,$8F,$A3,$C2,$EC,$7F,$9D,$6B,$F8,$C4,$7E,$26,$74,$FF,$B2
	.dc.b	$7D,$0E,$B1,$D9,$B5,$FE,$49,$F9,$3F,$C1,$3F,$5F,$DB,$FF,$0A,$46
	.dc.b	$2F,$E1,$C7,$F8,$5C,$7C,$AE,$BA,$F0,$DB,$E9,$E5,$F2,$38,$7C,$8D
	.dc.b	$8F,$F2,$C5,$A3,$31,$FE,$61,$DD,$F7,$1D,$A3,$73,$FD,$AF,$F7,$1F
	.dc.b	$DE,$C7,$F7,$06,$E7,$FE,$F0,$7F,$DE,$31,$FD,$B5,$CC,$7F,$BC,$FC
	.dc.b	$DF,$8A,$D9,$F9,$51,$8E,$0F,$F7,$14,$7C,$DF,$3B,$EA,$CF,$D6,$6C
	.dc.b	$7C,$9F,$8E,$F8,$E7,$F7,$FF,$DC,$7F,$94,$61,$FE,$78,$E1,$4E,$7D
	.dc.b	$49,$7F,$E0,$1F,$1B,$DD,$7F,$87,$C3,$8E,$FF,$E7,$AF,$C1,$FF,$3D
	.dc.b	$3F,$E8,$26,$C7,$F9,$FB,$85,$F1,$7F,$E4,$B9,$CE,$2B,$6F,$FE,$F7
	.dc.b	$AF,$B4,$7F,$5A,$E2,$FF,$09,$CF,$FA,$3B,$0E,$77,$FB,$9F,$A0,$FF
	.dc.b	$29,$F7,$1A,$48,$FE,$31,$1F,$C5,$EB,$F5,$33,$FC,$0F,$CB,$FE,$08
	.dc.b	$FF,$18,$FA,$8F,$E0,$DE,$6D,$C3,$F8,$33,$6F,$F8,$34,$FF,$3B,$71
	.dc.b	$B7,$FA,$73,$17,$07,$90,$98,$78,$9F,$3F,$95,$FC,$4D,$FE,$28,$D8
	.dc.b	$FE,$29,$E4,$FB,$08,$FA,$9F,$7F,$93,$FC,$6D,$FE,$38,$D8,$FE,$B7
	.dc.b	$DB,$FC,$7F,$C7,$FE,$66,$D8,$C9,$33,$FF,$5C,$FB,$EF,$03,$73,$FC
	.dc.b	$97,$C5,$F9,$A4,$7E,$5B,$F4,$3C,$5F,$E5,$8F,$F2,$D6,$C7,$F9,$26
	.dc.b	$FE,$F7,$BB,$F7,$FE,$A7,$F8,$DC,$7F,$95,$3C,$EF,$F6,$DC,$62,$3F
	.dc.b	$CA,$F3,$BF,$2B,$FC,$B3,$1F,$E9,$FB,$C3,$89,$FC,$57,$77,$18,$D1
	.dc.b	$FE,$08,$AE,$2A,$77,$1D,$30,$FF,$75,$D7,$1A,$D0,$C3,$A4,$F9,$AB
	.dc.b	$9D,$F8,$AB,$F3,$53,$97,$C8,$71,$9E,$D8,$6E,$3E,$19,$93,$E1,$39
	.dc.b	$7C,$38,$78,$23,$CF,$1D,$23,$0F,$B4,$C4,$F6,$70,$7C,$A2,$9F,$12
	.dc.b	$3D,$D7,$E3,$BF,$44,$EF,$0E,$A6,$9F,$31,$EE,$C3,$CD,$5C,$5B,$1D
	.dc.b	$AB,$E5,$E1,$F1,$A3,$13,$EA,$7B,$47,$9E,$FB,$1E,$07,$86,$1E,$E3
	.dc.b	$C4,$57,$27,$BD,$CB,$4F,$07,$7F,$F9,$22,$BE,$9B,$5F,$88,$D7,$E9
	.dc.b	$B9,$F9,$16,$8E,$6F,$E2,$D3,$98,$C5,$7F,$28,$C3,$FD,$5B,$7C,$2F
	.dc.b	$71,$8A,$DD,$F1,$9F,$85,$4D,$7F,$D6,$B7,$FA,$18,$C6,$1A,$C3,$BC
	.dc.b	$E5,$B8,$F8,$2F,$8B,$6B,$86,$11,$FB,$38,$74,$6D,$7F,$B1,$32,$E9
	.dc.b	$5F,$9B,$AF,$55,$7E,$29,$E1,$8F,$96,$FB,$CF,$F3,$99,$DC,$F2,$8E
	.dc.b	$A7,$39,$ED,$87,$C3,$BE,$D8,$FC,$AC,$7A,$5E,$CC,$CC,$7C,$57,$E0
	.dc.b	$58,$F9,$30,$E9,$1E,$7C,$3E,$2B,$E5,$AE,$15,$B8,$F4,$4E,$9E,$6F
	.dc.b	$8E,$BE,$54,$7A,$8F,$93,$19,$78,$DF,$9D,$D6,$1F,$04,$F8,$58,$FB
	.dc.b	$23,$D2,$7D,$37,$57,$DA,$74,$62,$7F,$B1,$F3,$39,$FF,$DF,$38,$7F
	.dc.b	$4D,$FA,$0B,$B7,$F6,$9E,$F5,$A7,$38,$72,$C2,$D6,$A7,$FC,$69,$5B
	.dc.b	$AF,$EC,$B1,$83,$FC,$71,$E8,$7E,$9F,$16,$3D,$E3,$0D,$D3,$AF,$23
	.dc.b	$D7,$EF,$5A,$C7,$D7,$CD,$E2,$CF,$FC,$F9,$1F,$EB,$0C,$31,$E4,$E7
	.dc.b	$E7,$A4,$FF,$9D,$62,$BF,$D0,$B6,$F7,$CF,$A8,$91,$FD,$FA,$BB,$72
	.dc.b	$C8,$FF,$3F,$45,$79,$61,$FE,$20,$7F,$C0,$B3,$F1,$0F,$F2,$C7,$9F
	.dc.b	$FE,$4A,$8F,$F0,$B6,$FF,$09,$FF,$0B,$C6,$5F,$E0,$DB,$FD,$38,$C4
	.dc.b	$EC,$7E,$04,$F0,$BE,$9C,$EA,$BE,$D7,$E2,$97,$0F,$F1,$1B,$75,$FF
	.dc.b	$6A,$79,$DF,$6A,$1F,$41,$75,$97,$E6,$C3,$9C,$E2,$3E,$45,$7E,$19
	.dc.b	$FF,$5D,$35,$CD,$7F,$9D,$FF,$76,$FF,$71,$47,$33,$FC,$6A,$E1,$F7
	.dc.b	$98,$CB,$F1,$FA,$BE,$F1,$E7,$7E,$43,$CC,$7C,$ED,$26,$1D,$77,$65
	.dc.b	$D7,$1F,$F4,$8F,$DE,$3E,$03,$F3,$76,$8C,$3E,$53,$11,$A8,$FC,$94
	.dc.b	$75,$C5,$F2,$5E,$32,$47,$19,$E6,$FF,$A8,$FE,$AF,$E2,$61,$FE,$A2
	.dc.b	$5E,$9F,$89,$B3,$E1,$77,$C7,$FE,$5B,$84,$69,$FF,$08,$B5,$FF,$0D
	.dc.b	$B5,$E0,$D7,$FC,$C9,$F8,$5B,$E0,$FE,$0B,$63,$E0,$35,$FF,$5D,$2F
	.dc.b	$81,$FD,$FE,$DF,$D6,$A7,$E7,$35,$FB,$3B,$FD,$0F,$1B,$EC,$F1,$BE
	.dc.b	$0B,$EC,$7C,$9C,$BF,$94,$F4,$AE,$BD,$FF,$77,$E6,$25,$76,$8A,$F6
	.dc.b	$EA,$F8,$5C,$5F,$A9,$CD,$18,$E1,$FE,$35,$CA,$C7,$E7,$7D,$17,$CF
	.dc.b	$F9,$79,$97,$CF,$8A,$FC,$C8,$FF,$1E,$BF,$E8,$BE,$44,$62,$3E,$9B
	.dc.b	$7F,$73,$7C,$3F,$E3,$1B,$3F,$8B,$6B,$F8,$3F,$A3,$FD,$BF,$81,$23
	.dc.b	$F3,$1D,$FF,$B2,$6F,$81,$FB,$5F,$A6,$91,$F9,$DE,$04,$71,$8F,$A3
	.dc.b	$5F,$05,$FE,$17,$D9,$FC,$10,$FD,$5F,$88,$F5,$7F,$80,$7D,$2C,$AE
	.dc.b	$8E,$DF,$1B,$FC,$30,$8F,$EB,$7F,$43,$79,$5F,$C4,$32,$93,$EA,$76
	.dc.b	$FD,$7F,$84,$FE,$1C,$47,$47,$C0,$7F,$89,$71,$48,$C1,$EF,$7F,$8B
	.dc.b	$72,$48,$C1,$F2,$1F,$E4,$24,$7F,$81,$7E,$87,$F2,$DF,$C4,$71,$7F
	.dc.b	$1F,$23,$C4,$FA,$35,$FC,$AA,$24,$60,$FA,$AF,$F3,$A2,$3F,$E8,$2E
	.dc.b	$97,$02,$7F,$13,$1F,$94,$FA,$6F,$E5,$C4,$AF,$B5,$F4,$FF,$CC,$7C
	.dc.b	$99,$1F,$75,$FA,$FE,$6F,$F3,$5F,$A5,$E7,$CC,$69,$2B,$F9,$8F,$C5
	.dc.b	$FF,$3B,$F2,$E4,$7E,$1E,$3F,$9F,$7C,$34,$8F,$C5,$3F,$CB,$BF,$07
	.dc.b	$FC,$5C,$7E,$37,$C6,$EE,$47,$F9,$EB,$E2,$F9,$DF,$E9,$1E,$14,$9F
	.dc.b	$CA,$47,$F2,$CF,$A6,$75,$D2,$4F,$E7,$B7,$FE,$A1,$E2,$FB,$09,$1F
	.dc.b	$D2,$9F,$93,$F4,$8F,$B0,$FE,$6F,$F4,$7D,$2B,$F9,$B3,$5F,$E2,$33
	.dc.b	$B7,$F9,$D7,$B9,$F8,$1F,$2B,$FB,$74,$7F,$86,$FC,$6F,$E8,$33,$9C
	.dc.b	$7F,$97,$3F,$D9,$6D,$F4,$3E,$22,$5F,$DF,$8C,$BF,$C1,$A7,$EA,$3E
	.dc.b	$0E,$5E,$A7,$6D,$3F,$C4,$F4,$FA,$1A,$4A,$FE,$C9,$9F,$F5,$3F,$6B
	.dc.b	$7E,$F7,$C2,$24,$F8,$5F,$8B,$F0,$DD,$09,$1F,$D8,$70,$7C,$97,$83
	.dc.b	$E6,$24,$E5,$FF,$03,$F4,$F9,$3E,$7B,$E7,$92,$7D,$CF,$D2,$F9,$3F
	.dc.b	$49,$23,$B7,$CE,$FF,$6A,$FA,$B7,$85,$FC,$95,$B3,$FC,$A9,$AF,$FB
	.dc.b	$A1,$AF,$FB,$CB,$F4,$97,$CA,$7F,$2D,$6C,$7F,$32,$6B,$FE,$F0,$6B
	.dc.b	$FE,$FA,$F4,$9F,$35,$FD,$F1,$B1,$FD,$ED,$AF,$F8,$38,$7A,$3F,$E4
	.dc.b	$87,$1F,$87,$7C,$87,$FC,$05,$FA,$6F,$F6,$3E,$6F,$C4,$B7,$B3,$FB
	.dc.b	$A5,$6E,$BA,$4F,$F6,$97,$FB,$C0,$70,$FF,$7B,$6A,$35,$F0,$5F,$E4
	.dc.b	$6F,$13,$FC,$76,$48,$FC,$2F,$8D,$FE,$20,$7F,$A3,$C7,$E3,$3E,$16
	.dc.b	$46,$0F,$E9,$7E,$37,$F9,$10,$8F,$F8,$97,$0F,$F0,$AE,$7F,$DA,$7C
	.dc.b	$F5,$7E,$5B,$E8,$25,$F2,$3F,$C5,$9D,$5F,$99,$7F,$90,$35,$FF,$35
	.dc.b	$35,$FF,$0C,$79,$FF,$92,$F0,$3F,$77,$F4,$1F,$A2,$F8,$C9,$5F,$97
	.dc.b	$8A,$FF,$9C,$FB,$FF,$C5,$ED,$9F,$E0,$67,$F9,$AB,$8E,$47,$F3,$1D
	.dc.b	$3F,$7A,$F8,$1A,$FF,$A6,$E9,$F1,$FF,$B2,$F8,$99,$3F,$A7,$8F,$F2
	.dc.b	$1F,$D6,$BC,$1F,$F3,$D4,$FD,$A8,$C7,$A7,$FE,$04,$6C,$7D,$E0,$E1
	.dc.b	$FC,$0F,$C8,$FF,$6F,$FA,$DF,$F0,$5B,$63,$83,$5D,$FB,$C9,$ED,$E9
	.dc.b	$BE,$52,$92,$BC,$18,$27,$51,$FE,$9B,$FA,$CF,$E5,$1C,$1F,$C0,$FB
	.dc.b	$7F,$D1,$0F,$FA,$F6,$33,$3F,$CA,$27,$2B,$95,$3A,$E9,$FF,$17,$36
	.dc.b	$F9,$73,$70,$AD,$FA,$77,$EB,$7F,$C9,$D5,$FD,$95,$B8,$FF,$AB,$DA
	.dc.b	$FE,$19,$AF,$FB,$45,$FF,$24,$47,$F9,$ED,$E3,$8F,$F8,$FD,$CC,$FF
	.dc.b	$94,$1C,$CF,$1F,$8A,$F1,$1E,$3F,$8D,$B1,$FE,$85,$CF,$F9,$2B,$FC
	.dc.b	$9E,$3D,$51,$8B,$FE,$97,$18,$EB,$95,$C7,$CD,$B5,$FE,$98,$C9,$F5
	.dc.b	$BF,$E8,$4A,$FE,$65,$97,$CC,$9C,$FB,$7F,$1F,$AF,$D3,$1F,$EB,$BB
	.dc.b	$E1,$87,$36,$C7,$D0,$F3,$79,$46,$31,$FF,$42,$7D,$6F,$F8,$3D,$FF
	.dc.b	$2B,$B6,$FF,$28,$D7,$F8,$9B,$5E,$3E,$77,$E0,$14,$71,$C3,$FB,$3B
	.dc.b	$FC,$F2,$BF,$D4,$8E,$5F,$F4,$A7,$D6,$BD,$DE,$4D,$8F,$F5,$57,$D6
	.dc.b	$FF,$6A,$7F,$D5,$91,$9C,$3F,$99,$4F,$F6,$39,$FF,$37,$B6,$3F,$58
	.dc.b	$D7,$F9,$5B,$5F,$F5,$22,$F5,$7D,$4D,$9F,$F5,$D7,$E9,$3E,$7A,$FE
	.dc.b	$5D,$39,$7E,$76,$1F,$30,$FA,$9F,$19,$6F,$FD,$85,$EC,$79,$EB,$F9
	.dc.b	$DB,$5F,$F5,$A3,$5F,$E7,$4D,$E9,$70,$FE,$E6,$BF,$BA,$77,$7E,$3B
	.dc.b	$C7,$FC,$13,$F8,$5F,$3B,$E2,$70,$FF,$BF,$5B,$5F,$C6,$72,$D6,$9F
	.dc.b	$5F,$3F,$E2,$D3,$FE,$96,$6C,$FF,$13,$6B,$E8,$F1,$7F,$91,$E0,$FA
	.dc.b	$EF,$C8,$DF,$F5,$18,$C7,$47,$E8,$8F,$CC,$4E,$AF,$FA,$B7,$D6,$FD
	.dc.b	$B7,$FB,$53,$FC,$0D,$CF,$BD,$F8,$B8,$C4,$17,$5F,$F3,$48,$C3,$63
	.dc.b	$38,$FF,$AC,$1B,$3F,$CB,$56,$FD,$7C,$CE,$CD,$AF,$F8,$01,$F3,$5F
	.dc.b	$E0,$26,$FD,$4F,$F1,$58,$C6,$1F,$ED,$65,$1F,$A1,$9F,$F5,$21,$FE
	.dc.b	$C0,$2C,$FF,$3D,$6B,$FD,$F1,$AF,$F7,$9E,$45,$7F,$A0,$A7,$C7,$E7
	.dc.b	$7F,$25,$FA,$BF,$F0,$EC,$60,$FF,$1F,$F7,$7F,$70,$7D,$3E,$D7,$EF
	.dc.b	$BD,$2F,$43,$FB,$5F,$C4,$7F,$65,$9F,$F4,$F6,$BF,$E0,$E6,$BF,$E0
	.dc.b	$B0,$E1,$72,$7C,$0E,$2F,$E2,$9B,$7D,$B8,$FF,$75,$66,$FA,$4C,$D6
	.dc.b	$BB,$FE,$57,$D3,$DD,$7E,$7B,$E5,$F8,$AF,$B9,$A1,$A4,$9F,$EC,$D9
	.dc.b	$7D,$0E,$6F,$09,$F1,$7E,$8A,$CF,$DA,$FA,$7F,$B0,$7D,$9F,$A0,$B1
	.dc.b	$F9,$73,$FC,$1B,$E7,$FE,$2F,$9F,$F8,$27,$17,$D3,$8C,$1F,$E0,$28
	.dc.b	$70,$BE,$8B,$F9,$04,$7F,$0F,$D3,$E0,$C6,$2B,$EC,$7D,$17,$E2,$7E
	.dc.b	$9E,$E5,$F8,$E6,$E5,$F3,$23,$F0,$59,$2B,$F0,$1F,$3B,$AC,$95,$F8
	.dc.b	$7E,$BF,$9C,$FE,$63,$9B,$F0,$DE,$17,$E5,$DB,$19,$3E,$06,$5F,$DD
	.dc.b	$72,$56,$E7,$F4,$3F,$4D,$F9,$FF,$A7,$FD,$03,$63,$8F,$03,$EB,$1F
	.dc.b	$21,$A9,$CC,$61,$E6,$63,$E7,$7F,$BF,$46,$73,$F8,$07,$F8,$1D,$46
	.dc.b	$AB,$F1,$BF,$01,$FB,$1F,$7A,$E1,$FB,$67,$9F,$CD,$E3,$1E,$79,$79
	.dc.b	$63,$1C,$3F,$22,$7F,$C2,$9C,$2E,$9F,$DE,$7D,$0F,$EF,$5B,$97,$AB
	.dc.b	$E6,$FC,$E7,$A3,$E6,$AC,$7C,$AC,$3F,$C3,$DE,$D7,$E0,$8F,$C1,$FC
	.dc.b	$95,$9D,$6B,$FC,$4A,$BF,$CF,$6D,$7F,$CB,$BD,$FF,$91,$3F,$25,$F2
	.dc.b	$5F,$8C,$7F,$1A,$DA,$FA,$3E,$3B,$E9,$BF,$C5,$0D,$8C,$F2,$62,$72
	.dc.b	$D8,$C9,$CA,$3F,$C1,$9F,$53,$C7,$1F,$93,$6F,$9E,$CD,$76,$7B,$EF
	.dc.b	$8C,$3C,$6D,$BF,$F5,$26,$8F,$95,$FD,$29,$FA,$6F,$8E,$B7,$F1,$3E
	.dc.b	$37,$F6,$04,$7F,$CC,$57,$FE,$49,$8F,$DA,$7A,$5F,$AE,$7F,$5E,$D9
	.dc.b	$FA,$5E,$27,$A7,$A3,$FD,$EB,$F5,$5F,$9C,$9C,$ED,$FE,$15,$8C,$63
	.dc.b	$FE,$B9,$5F,$1B,$FE,$19,$C3,$95,$FF,$21,$9C,$BC,$8F,$F7,$6F,$0B
	.dc.b	$9F,$E8,$FB,$1F,$29,$B3,$F5,$BE,$2B,$0D,$38,$C7,$F2,$8F,$7B,$FB
	.dc.b	$5F,$A6,$69,$27,$FD,$09,$C7,$FA,$FF,$90,$E1,$1F,$AE,$6E,$DF,$E4
	.dc.b	$9B,$FF,$61,$4F,$F5,$AF,$37,$F9,$AC,$FF,$59,$6C,$7F,$02,$E1,$FE
	.dc.b	$B3,$3F,$E7,$1F,$4F,$FB,$94,$62,$75,$F4,$3F,$D8,$2B,$FA,$24,$A3
	.dc.b	$11,$FE,$01,$6B,$FD,$19,$B9,$7F,$8B,$7D,$5F,$F2,$D4,$7F,$6A,$AF
	.dc.b	$DA,$4E,$6F,$BF,$CA,$FF,$6F,$F9,$AF,$E6,$1E,$29,$A4,$9F,$8B,$9F
	.dc.b	$F8,$CF,$0E,$75,$B9,$FE,$2F,$39,$7F,$94,$57,$BF,$C2,$FE,$F9,$E8
	.dc.b	$D9,$FF,$00,$B7,$5F,$E1,$76,$AF,$8D,$E1,$67,$2E,$FF,$43,$E6,$26
	.dc.b	$3F,$63,$34,$7F,$49,$C1,$F1,$3F,$E0,$B8,$FA,$0D,$AF,$E1,$99,$7F
	.dc.b	$BE,$FE,$AB,$FD,$F9,$F5,$5F,$ED,$08,$FE,$E9,$F0,$BF,$C6,$E3,$11
	.dc.b	$FD,$DE,$77,$E7,$7F,$79,$8F,$F2,$40,$9D,$BD,$4F,$E8,$31,$BA,$FE
	.dc.b	$4D,$E9,$EA,$57,$FE,$54,$AD,$61,$67,$B4,$F8,$1F,$2C,$57,$C8,$78
	.dc.b	$67,$8B,$4E,$25,$07,$A3,$0F,$15,$73,$8C,$4F,$FB,$9E,$3C,$51,$C3
	.dc.b	$0F,$4C,$E6,$3D,$A7,$90,$A7,$92,$BE,$1C,$FC,$B9,$E1,$18,$BF,$05
	.dc.b	$7C,$23,$CC,$78,$8F,$4C,$6D,$C9,$B7,$07,$A9,$E3,$5D,$DF,$83,$87
	.dc.b	$13,$45,$BF,$19,$F2,$2B,$91,$D1,$F7,$46,$E7,$9C,$7C,$E8,$E6,$D3
	.dc.b	$65,$8F,$A3,$E0,$7F,$80,$9B,$3F,$26,$31,$1F,$69,$BF,$28,$F3,$BF
	.dc.b	$C8,$73,$FC,$AA,$BF,$C9,$13,$9E,$0F,$B7,$7C,$2F,$6D,$8C,$D7,$09
	.dc.b	$E6,$FE,$72,$BF,$BB,$5F,$DE,$6B,$F6,$9E,$CD,$9E,$38,$78,$BC,$EF
	.dc.b	$F5,$6B,$67,$94,$E9,$FC,$7D,$72,$CB,$C1,$7F,$E6,$16,$CF,$E6,$E7
	.dc.b	$57,$E7,$C7,$C9,$7D,$AB,$9B,$55,$79,$B1,$F6,$63,$F2,$31,$EB,$5F
	.dc.b	$A1,$AF,$6C,$6B,$0F,$05,$6F,$1F,$55,$EF,$1F,$34,$78,$E3,$94,$ED
	.dc.b	$F1,$9E,$99,$DB,$F0,$23,$4F,$99,$F8,$8F,$A2,$3D,$58,$7D,$0C,$7C
	.dc.b	$93,$F3,$F0,$E8,$C3,$C2,$FA,$0C,$63,$EE,$38,$9C,$A7,$E0,$93,$1F
	.dc.b	$53,$A7,$DF,$1C,$5F,$73,$FC,$8F,$9F,$FD,$F1,$E3,$7F,$54,$DB,$FC
	.dc.b	$C9,$EB,$DC,$BF,$C7,$ED,$AD,$C7,$FA,$FA,$3F,$C1,$F1,$FE,$B0,$6F
	.dc.b	$21,$A6,$B9,$3F,$D9,$35,$FD,$33,$E0,$7A,$46,$23,$FA,$3F,$CA,$E2
	.dc.b	$7C,$17,$FA,$AD,$BF,$AA,$ED,$9B,$FE,$AB,$F2,$7A,$E3,$91,$83,$FD
	.dc.b	$A6,$F9,$08,$FF,$4D,$7C,$9F,$0D,$EF,$BF,$EE,$9E,$85,$EF,$3E,$69
	.dc.b	$C3,$FE,$2E,$9D,$DF,$FA,$B2,$B6,$D9,$D4,$3D,$DE,$84,$9F,$ED,$BD
	.dc.b	$15,$FF,$38,$FC,$27,$F9,$55,$FF,$43,$B5,$FF,$3F,$E1,$FE,$09,$AF
	.dc.b	$A9,$F3,$1F,$E1,$29,$CE,$0F,$37,$D4,$6B,$C6,$B3,$E7,$8F,$29,$1F
	.dc.b	$F1,$3F,$D4,$3C,$5F,$89,$F5,$0F,$4A,$9F,$83,$9D,$57,$F8,$8E,$74
	.dc.b	$FF,$A7,$A1,$AE,$CE,$31,$FE,$81,$FA,$A7,$A7,$F8,$F8,$47,$E0,$F7
	.dc.b	$FF,$48,$46,$39,$F5,$3C,$23,$94,$63,$A2,$46,$0F,$F4,$77,$D4,$7C
	.dc.b	$9B,$E6,$FF,$AC,$23,$3E,$C7,$86,$35,$DD,$FE,$F9,$75,$3F,$E0,$30
	.dc.b	$E1,$FF,$03,$35,$E1,$3F,$E7,$27,$7F,$3B,$D4,$8F,$93,$DE,$F9,$5E
	.dc.b	$F7,$CE,$EA,$93,$F2,$62,$BF,$EB,$4E,$0F,$E9,$0F,$CD,$F1,$F3,$A7
	.dc.b	$F9,$0A,$7F,$C4,$F3,$F4,$7C,$1F,$AC,$7D,$7F,$2A,$C6,$48,$75,$7E
	.dc.b	$17,$D1,$FC,$3F,$72,$4F,$CA,$8A,$FC,$3D,$FF,$D3,$CF,$FA,$41,$AE
	.dc.b	$67,$E1,$47,$F9,$FE,$3E,$9F,$AB,$F3,$CF,$A1,$EE,$7E,$6B,$59,$3F
	.dc.b	$EE,$6E,$E4,$7E,$7B,$E0,$7F,$89,$74,$FF,$B7,$AB,$F4,$BC,$98,$33
	.dc.b	$C7,$FA,$4F,$57,$F7,$58,$4A,$E9,$15,$FD,$7F,$2F,$43,$1C,$7F,$E6
	.dc.b	$CC,$DE,$9F,$EB,$B8,$A5,$F9,$62,$BF,$B8,$E2,$F2,$98,$BF,$F3,$AE
	.dc.b	$A8,$F4,$FC,$04,$AF,$DE,$C7,$EE,$BC,$6F,$6F,$24,$9F,$85,$15,$FE
	.dc.b	$0B,$17,$0F,$34,$FF,$35,$FA,$7F,$CB,$C7,$FA,$11,$BD,$5F,$E4,$4F
	.dc.b	$31,$18,$C1,$18,$E2,$E5,$C7,$F3,$B9,$65,$7F,$93,$31,$E9,$3D,$F5
	.dc.b	$FD,$47,$CB,$4A,$D6,$7D,$F3,$56,$67,$F1,$15,$F8,$42,$3F,$EB,$6D
	.dc.b	$DF,$4F,$F7,$89,$1F,$C8,$BA,$3F,$9A,$6F,$23,$FD,$5A,$FC,$1E,$0F
	.dc.b	$E1,$5F,$8A,$BD,$AF,$D9,$36,$BF,$DC,$AD,$7F,$78,$D7,$FD,$6B,$F2
	.dc.b	$7F,$8D,$7F,$99,$E7,$F9,$03,$FA,$34,$67,$81,$BF,$C5,$8F,$7D,$7F
	.dc.b	$B1,$5A,$FF,$38,$6C,$FD,$5F,$8C,$F3,$FC,$5F,$98,$95,$E6,$7F,$A7
	.dc.b	$53,$8F,$F4,$35,$2B,$B4,$7E,$93,$E9,$24,$7F,$48,$7F,$09,$F0,$5F
	.dc.b	$CB,$7F,$15,$FE,$8D,$3F,$E9,$D6,$BF,$E8,$80,$E1,$FF,$45,$B7,$A3
	.dc.b	$F9,$DF,$CD,$FE,$CB,$E9,$A5,$7F,$56,$8E,$DC,$7F,$CE,$FE,$0B,$F9
	.dc.b	$5C,$FF,$A5,$1A,$F6,$8D,$72,$7F,$27,$9F,$E1,$33,$FE,$B1,$6B,$E5
	.dc.b	$6B,$CF,$5E,$DA,$48,$FC,$89,$93,$FD,$ED,$E5,$FF,$82,$78,$BF,$34
	.dc.b	$D8,$71,$7F,$50,$F5,$FF,$C3,$DF,$46,$E4,$FE,$A9,$F8,$DB,$E1,$FF
	.dc.b	$9E,$37,$0F,$E8,$81,$C3,$FD,$1D,$AF,$FC,$00,$F9,$87,$BD,$FE,$0F
	.dc.b	$6C,$7A,$32,$46,$23,$FC,$07,$F3,$3F,$C8,$39,$E4,$7F,$81,$1F,$3F
	.dc.b	$C3,$79,$BE,$23,$CA,$49,$F2,$E4,$8F,$C5,$F6,$7E,$68,$93,$FE,$0E
	.dc.b	$D3,$FB,$97,$D3,$FE,$CA,$3E,$58,$70,$FC,$8E,$2F,$C4,$46,$35,$FE
	.dc.b	$EF,$E5,$FF,$84,$DF,$EF,$30,$C3,$FB,$23,$6F,$E4,$B5,$FE,$DC,$D7
	.dc.b	$FD,$E5,$F4,$FF,$E2,$C7,$C0,$7F,$B7,$9C,$57,$F8,$19,$B8,$79,$39
	.dc.b	$BF,$C1,$9F,$4F,$F6,$EF,$D6,$BE,$13,$26,$B9,$2C,$FD,$C9,$D3,$67
	.dc.b	$FB,$E4,$62,$3F,$50,$7E,$2B,$C1,$74,$FF,$06,$24,$7E,$53,$BD,$D1
	.dc.b	$FE,$19,$8F,$E0,$93,$FD,$73,$8A,$47,$F0,$98,$FC,$9F,$AF,$23,$95
	.dc.b	$23,$F2,$FF,$0D,$99,$FF,$24,$FC,$D7,$F6,$63,$1C,$3F,$C5,$A3,$F9
	.dc.b	$FF,$C9,$FA,$63,$19,$A3,$1C,$3F,$D1,$DB,$5C,$3C,$7F,$F4,$14,$62
	.dc.b	$B8,$70,$E0,$CF,$0D,$BF,$F1,$44,$E9,$B5,$F4,$DF,$93,$19,$6E,$AF
	.dc.b	$6B,$F6,$64,$8F,$83,$F1,$D5,$FF,$33,$E7,$FE,$04,$31,$A7,$F9,$23
	.dc.b	$E9,$9D,$AF,$17,$F6,$F1,$86,$C6,$6F,$FC,$7D,$39,$7F,$C8,$1F,$17
	.dc.b	$FE,$12,$FA,$7F,$DF,$B6,$33,$EC,$A7,$2D,$8E,$7D,$EF,$88,$F0,$B6
	.dc.b	$33,$18,$3E,$3F,$D2,$FF,$11,$D7,$F1,$0B,$E9,$FE,$D6,$DF,$62,$35
	.dc.b	$7F,$A4,$E2,$46,$EF,$8F,$2F,$07,$FC,$05,$1C,$2F,$97,$36,$70,$FF
	.dc.b	$01,$CE,$6B,$FC,$11,$F4,$FF,$4D,$FB,$6D,$9F,$B2,$D7,$FA,$56,$28
	.dc.b	$C4,$72,$FA,$7E,$D3,$9E,$5B,$49,$B1,$86,$5F,$C2,$E8,$B8,$7B,$F1
	.dc.b	$7C,$CF,$47,$F3,$55,$F2,$23,$1D,$39,$64,$7C,$CF,$95,$B7,$CF,$E5
	.dc.b	$7E,$16,$EF,$A1,$7A,$5D,$67,$31,$8C,$3E,$AC,$65,$FA,$FE,$5D,$C3
	.dc.b	$F8,$B8,$FA,$7F,$E4,$53,$FD,$21,$B1,$FE,$8D,$6B,$FC,$31,$AF,$F7
	.dc.b	$5F,$A6,$7B,$DF,$5E,$3A,$79,$96,$73,$7F,$CE,$2F,$AE,$9F,$EA,$AF
	.dc.b	$A6,$FC,$3E,$1F,$C6,$1A,$FF,$80,$5A,$83,$86,$F1,$7F,$AD,$1E,$0D
	.dc.b	$78,$47,$F2,$DB,$FE,$CA,$D8,$FF,$55,$61,$FE,$3E,$9F,$E9,$AD,$8F
	.dc.b	$8E,$BD,$BF,$67,$B9,$F7,$7E,$3A,$5F,$E9,$3E,$76,$E1,$F8,$3F,$5B
	.dc.b	$C5,$D7,$FA,$26,$F4,$7F,$70,$BF,$D6,$C6,$3E,$37,$F1,$07,$E2,$7E
	.dc.b	$3A,$CF,$EC,$3C,$CF,$C6,$1F,$8D,$F3,$6C,$7E,$D3,$E5,$BA,$91,$FD
	.dc.b	$CB,$E0,$EF,$FC,$A1,$F9,$5E,$FB,$3F,$BC,$F1,$3F,$31,$E6,$7E,$65
	.dc.b	$B1,$FD,$C3,$E9,$5D,$9F,$37,$85,$18,$9F,$F0,$8B,$63,$FD,$AC,$D7
	.dc.b	$F6,$7C,$1F,$6E,$7A,$6F,$FA,$79,$FC,$55,$23,$F6,$BF,$43,$FF,$00
	.dc.b	$6B,$FA,$C2,$F9,$5C,$FC,$A6,$9F,$C9,$4B,$E4,$FC,$CD,$3F,$C4,$A5
	.dc.b	$C7,$FD,$D4,$D7,$F8,$3D,$BE,$63,$C2,$78,$BE,$22,$D7,$8B,$E6,$3C
	.dc.b	$BF,$4B,$E6,$6C,$79,$BE,$93,$F0,$38,$1A,$48,$FE,$A7,$0F,$A4,$F7
	.dc.b	$9F,$03,$E6,$AC,$FC,$0F,$96,$F8,$7F,$49,$F1,$1B,$1F,$13,$D5,$EE
	.dc.b	$47,$AB,$E2,$F9,$FF,$96,$7C,$CE,$0B,$3F,$E2,$CF,$03,$F5,$84,$F5
	.dc.b	$3E,$F3,$FB,$21,$A3,$E6,$3D,$A4,$AF,$5B,$F0,$3E,$33,$EC,$BF,$69
	.dc.b	$B1,$F6,$BE,$6B,$EE,$FC,$BF,$DE,$6C,$64,$9C,$DF,$D2,$F0,$7C,$C6
	.dc.b	$E7,$F0,$7F,$1D,$F8,$63,$F9,$DF,$D2,$3E,$93,$13,$FE,$1F,$8C,$36
	.dc.b	$3F,$1B,$F0,$7C,$1E,$3E,$AC,$9E,$1F,$33,$73,$FE,$B9,$82,$3F,$7A
	.dc.b	$D9,$FF,$06,$B6,$7F,$C1,$6D,$7F,$A9,$AF,$8E,$F3,$D7,$A1,$B5,$E7
	.dc.b	$F9,$8F,$59,$EC,$F9,$4B,$1E,$B7,$0F,$B7,$E6,$AE,$3E,$6F,$98,$F8
	.dc.b	$67,$C4,$F5,$6D,$7C,$3F,$57,$0D,$3E,$34,$7F,$3B,$E7,$FE,$B7,$1F
	.dc.b	$E8,$C6,$DF,$F8,$AB,$E0,$3F,$0A,$7E,$1B,$CB,$95,$B6,$CF,$F8,$C3
	.dc.b	$DD,$E9,$1D,$5B,$3F,$E5,$3F,$1F,$FA,$3D,$7F,$82,$5B,$ED,$FF,$91
	.dc.b	$38,$9F,$21,$FD,$E2,$73,$7F,$D1,$73,$56,$A3,$F0,$15,$F8,$4F,$7B
	.dc.b	$F0,$AF,$1D,$E4,$FE,$17,$E5,$EE,$7F,$79,$FC,$59,$F8,$CF,$8E,$B3
	.dc.b	$FA,$7F,$6B,$F4,$A4,$7E,$33,$FA,$8F,$6A,$47,$C9,$6D,$FF,$9D,$87
	.dc.b	$D2,$7F,$1B,$9F,$D9,$4F,$F6,$B8,$C4,$FF,$78,$D5,$39,$EC,$97,$F9
	.dc.b	$DF,$96,$B7,$F9,$67,$F4,$67,$E9,$3E,$12,$CF,$CF,$F8,$AF,$94,$7E
	.dc.b	$A0,$AF,$E9,$DC,$3F,$A9,$F8,$EB,$7F,$A2,$F9,$1F,$D0,$92,$7F,$42
	.dc.b	$FE,$93,$E4,$7F,$66,$FE,$D1,$B1,$FE,$91,$F8,$6F,$F2,$25,$7F,$AA
	.dc.b	$B9,$FF,$9E,$3F,$E3,$48,$CC,$7F,$93,$BD,$7F,$E9,$13,$FE,$4D,$9C
	.dc.b	$D7,$E0,$5B,$0D,$FF,$07,$3C,$FE,$37,$A7,$C6,$FF,$74,$B3,$49,$3F
	.dc.b	$E8,$68,$FF,$28,$5F,$37,$71,$E9,$F3,$B9,$BE,$B8,$E8,$47,$93,$DF
	.dc.b	$E3,$3D,$8F,$B5,$B1,$F0,$FB,$BE,$07,$C6,$7C,$16,$C6,$49,$87,$C1
	.dc.b	$F0,$35,$1F,$2F,$C0,$F6,$11,$F5,$3E,$FF,$87,$F9,$8F,$CD,$6C,$7F
	.dc.b	$BB,$2B,$FC,$5D,$F4,$7F,$E2,$FF,$31,$F1,$1F,$EC,$CF,$83,$FF,$5C
	.dc.b	$46,$2B,$FD,$9F,$5B,$AD,$5D,$9C,$8A,$E0,$FF,$8E,$EF,$9D,$7E,$8E
	.dc.b	$F8,$4E,$6E,$BE,$41,$F4,$7C,$9F,$F4,$D4,$7F,$91,$63,$10,$9C,$57
	.dc.b	$92,$38,$BD,$DF,$23,$9C,$38,$CF,$83,$0F,$2D,$7A,$A7,$C1,$18,$8F
	.dc.b	$F2,$A1,$DC,$E6,$61,$D3,$E9,$BE,$38,$79,$23,$E2,$C6,$1F,$29,$8C
	.dc.b	$3D,$11,$E4,$8C,$9F,$14,$A7,$58,$D9,$ED,$8E,$87,$1B,$EC,$58,$F8
	.dc.b	$06,$CE,$33,$E1,$3D,$55,$E7,$7D,$F8,$7C,$58,$FA,$B1,$F0,$6B,$AD
	.dc.b	$7C,$D6,$C7,$C9,$8E,$EF,$E1,$3B,$3F,$08,$D9,$FA,$91,$88,$EE,$7F
	.dc.b	$B9,$E3,$E7,$FC,$8F,$F8,$F7,$9F,$FD,$09,$5B,$8C,$56,$CF,$F4,$43
	.dc.b	$88,$D7,$AE,$AE,$37,$D2,$31,$7F,$68,$5B,$D8,$FB,$F5,$F9,$A8,$C4
	.dc.b	$FE,$61,$F0,$C6,$6B,$9B,$BA,$E3,$8F,$A9,$B7,$E6,$AD,$CF,$8F,$0F
	.dc.b	$A1,$87,$6A,$FC,$0B,$74,$ED,$3A,$BE,$35,$F0,$30,$EA,$FA,$9B,$7E
	.dc.b	$6C,$3B,$4F,$49,$FC,$DE,$1E,$78,$E9,$84,$74,$79,$27,$75,$CA,$BC
	.dc.b	$73,$A2,$38,$33,$8F,$58,$F9,$D8,$7B,$63,$35,$E4,$74,$47,$07,$73
	.dc.b	$E5,$4F,$8C,$F3,$57,$27,$E1,$1F,$56,$FE,$0E,$3F,$1A,$3E,$EB,$70
	.dc.b	$F1,$5F,$49,$F2,$CF,$CC,$E8,$72,$BE,$2F,$F8,$0E,$8D,$BF,$A9,$78
	.dc.b	$78,$6E,$3F,$D7,$EB,$0F,$F4,$B9,$FF,$16,$59,$A4,$8F,$EA,$5F,$17
	.dc.b	$C3,$1F,$0B,$68,$DF,$0F,$0E,$7E,$16,$5F,$F6,$F8,$E1,$AF,$1F,$8B
	.dc.b	$93,$FD,$CE,$2B,$FC,$70,$FE,$C1,$1E,$62,$3F,$D9,$7E,$2B,$C5,$F1
	.dc.b	$72,$3F,$BE,$45,$79,$67,$F2,$27,$DE,$48,$FE,$DD,$CF,$EB,$23,$FE
	.dc.b	$05,$F8,$BA,$FF,$81,$7E,$89,$F0,$FF,$2B,$8D,$18,$9F,$F0,$65,$FC
	.dc.b	$57,$E4,$7C,$FF,$F8,$31,$FE,$01,$C9,$FE,$47,$8C,$5F,$D9,$9F,$EF
	.dc.b	$C7,$0F,$17,$FC,$C1,$3C,$63,$F9,$FE,$88,$C4,$73,$9D,$FC,$57,$80
	.dc.b	$90,$3C,$27,$F8,$56,$5F,$17,$F7,$7E,$86,$CE,$F1,$F2,$62,$9C,$CF
	.dc.b	$28,$C5,$76,$8C,$5B,$E3,$3E,$09,$23,$A6,$7F,$07,$D7,$FC,$9C,$3E
	.dc.b	$77,$FC,$9A,$E8,$F4,$38,$7D,$26,$FA,$3A,$38,$FB,$C7,$37,$4F,$A2
	.dc.b	$FF,$59,$4F,$30,$E1,$E1,$1B,$F8,$3F,$F3,$24,$7C,$79,$D3,$F9,$36
	.dc.b	$BF,$DC,$1A,$EF,$E0,$F8,$BD,$9B,$06,$9F,$73,$E3,$FF,$0A,$49,$F8
	.dc.b	$1C,$1F,$77,$9F,$EA,$7C,$5C,$9C,$F1,$A3,$84,$7D,$0F,$8F,$FC,$79
	.dc.b	$1E,$7C,$6F,$8F,$FB,$DF,$11,$27,$9B,$B7,$E6,$EA,$F8,$6F,$95,$F0
	.dc.b	$B2,$7E,$54,$E5,$B1,$EE,$23,$F2,$A1,$1F,$0B,$E1,$A4,$E5,$AF,$ED
	.dc.b	$B1,$7C,$7F,$EA,$3E,$3E,$4F,$AA,$3F,$57,$AC,$8F,$6B,$A7,$F2,$3E
	.dc.b	$67,$EC,$08,$FC,$2F,$2F,$F6,$7F,$1F,$23,$E2,$BA,$7F,$33,$F1,$FF
	.dc.b	$B8,$F8,$F9,$1F,$2A,$2B,$E7,$8F,$DA,$91,$FF,$56,$72,$FE,$D7,$87
	.dc.b	$F8,$09,$23,$E9,$CF,$F0,$4F,$46,$47,$D6,$7F,$83,$FC,$7F,$F0,$8F
	.dc.b	$4E,$4F,$DA,$76,$FE,$C7,$E8,$9F,$1F,$FC,$D5,$B1,$F8,$A6,$BB,$ED
	.dc.b	$D4,$EF,$B6,$4E,$E3,$97,$C7,$F3,$E9,$93,$CE,$2B,$FC,$B3,$0F,$C1
	.dc.b	$36,$7F,$DE,$D1,$88,$FD,$47,$C6,$7E,$88,$8F,$E1,$5D,$F4,$7A,$49
	.dc.b	$1E,$9E,$9F,$67,$C5,$F4,$24,$7C,$EB,$47,$83,$E3,$FE,$17,$C7,$CA
	.dc.b	$F8,$51,$F1,$7E,$3A,$47,$C5,$7E,$46,$5F,$27,$E3,$7D,$24,$9E,$CF
	.dc.b	$B3,$E2,$FE,$77,$C7,$49,$F9,$D1,$F4,$7E,$2E,$47,$D1,$7E,$A7,$C6
	.dc.b	$FF,$1E,$FA,$2F,$E0,$B2,$69,$1F,$93,$F1,$BF,$6F,$E3,$A4,$7D,$B8
	.dc.b	$FB,$DF,$1D,$FC,$A3,$E8,$9D,$54,$CD,$7F,$1F,$F0,$3F,$62,$FF,$1C
	.dc.b	$6B,$FF,$00,$5A,$FF,$6F,$6B,$FD,$1F,$C0,$7C,$E3,$E4,$3F,$1D,$EA
	.dc.b	$CA,$FE,$2F,$39,$6F,$C7,$FF,$81,$E0,$D3,$1F,$21,$9F,$85,$FE,$5F
	.dc.b	$8F,$66,$D7,$F8,$7A,$FF,$56,$D7,$FB,$DA,$F8,$FE,$C4,$8F,$1F,$C8
	.dc.b	$78,$7E,$42,$47,$99,$F2,$7C,$87,$97,$E4,$3B,$92,$7B,$BE,$2F,$8C
	.dc.b	$F5,$12,$3D,$FF,$21,$FD,$63,$C6,$FE,$F3,$C1,$FE,$B2,$6C,$FF,$6F
	.dc.b	$6B,$F2,$BD,$DF,$EB,$FF,$40,$F0,$BF,$A3,$B6,$7F,$A8,$B5,$FF,$11
	.dc.b	$07,$0F,$F8,$95,$78,$3C,$89,$1C,$9F,$EA,$7F,$B9,$3E,$1F,$FB,$7F
	.dc.b	$D0,$7D,$4A,$FE,$C8,$D7,$FB,$6C,$ED,$CF,$3F,$F5,$C9,$D4,$7F,$8D
	.dc.b	$5A,$FF,$B4,$9A,$B9,$BE,$03,$F8,$6C,$FF,$08,$7E,$17,$E4,$2C,$7F
	.dc.b	$7F,$F8,$77,$D0,$3C,$7F,$F1,$73,$67,$91,$F9,$0F,$8F,$FC,$71,$1F
	.dc.b	$C9,$FC,$8D,$7F,$C2,$DD,$7F,$E5,$67,$F2,$FF,$23,$9C,$7F,$C3,$BD
	.dc.b	$6E,$EF,$F2,$F3,$6B,$FB,$9B,$5E,$7F,$0D,$A9,$DF,$7C,$9D,$C5,$7F
	.dc.b	$CA,$35,$F9,$97,$FC,$72,$D7,$F5,$CD,$7E,$46,$2F,$8F,$FD,$51,$23
	.dc.b	$F5,$FF,$23,$5F,$6F,$C8,$FE,$BC,$C6,$7F,$E4,$B8,$7D,$03,$CD,$FE
	.dc.b	$0F,$1F,$8C,$8C,$78,$FF,$E2,$06,$C6,$67,$6F,$F9,$CF,$E4,$FF,$C2
	.dc.b	$1F,$41,$FE,$26,$6C,$7E,$59,$AE,$FE,$0D,$3D,$7C,$EF,$C6,$4F,$F9
	.dc.b	$F7,$04,$62,$3F,$CA,$C7,$A1,$C3,$FE,$59,$8D,$F7,$7F,$6E,$FA,$0F
	.dc.b	$F1,$FF,$89,$79,$7E,$05,$FF,$19,$9C,$BF,$E8,$26,$BF,$E6,$A5,$BF
	.dc.b	$F9,$C7,$7F,$F2,$FC,$70,$6C,$66,$FA,$5D,$BF,$AF,$87,$81,$B3,$FC
	.dc.b	$62,$31,$DB,$FC,$A9,$B8,$FD,$D9,$F5,$71,$60,$8F,$B6,$FB,$5A,$B8
	.dc.b	$47,$F9,$C5,$FF,$3E,$4F,$F9,$21,$CE,$5F,$E7,$8F,$D9,$5F,$2B,$C9
	.dc.b	$B5,$FE,$4E,$5F,$2B,$99,$D3,$67,$4D,$C3,$FD,$01,$4F,$93,$E8,$7C
	.dc.b	$3C,$1F,$49,$F2,$70,$FE,$95,$5B,$7F,$7D,$19,$BF,$E0,$D3,$97,$FD
	.dc.b	$2B,$5C,$B6,$EB,$7F,$E7,$9F,$A0,$FD,$B3,$FE,$76,$6D,$7F,$26,$6B
	.dc.b	$FE,$C9,$6B,$F8,$C7,$FA,$9C,$71,$F1,$7E,$C4,$A7,$2F,$38,$FB,$3D
	.dc.b	$5F,$3A,$3F,$85,$9F,$81,$1F,$60,$EC,$FA,$F3,$BC,$BF,$65,$F8,$7F
	.dc.b	$F9,$C7,$2F,$F4,$6B,$70,$FE,$7A,$D7,$FD,$C0,$D7,$F4,$6F,$F9,$92
	.dc.b	$3F,$D3,$4B,$E4,$BC,$4E,$67,$FD,$4A,$E6,$7F,$D1,$7E,$6F,$E0,$27
	.dc.b	$FD,$70,$DB,$FD,$06,$EE,$FF,$C7,$CE,$6F,$FC,$04,$FF,$B1,$9F,$F2
	.dc.b	$EB,$6F,$FD,$DA,$D7,$FD,$AA,$D7,$F0,$3F,$25,$D3,$8F,$F2,$4D,$AF
	.dc.b	$2F,$1F,$80,$FC,$56,$3F,$D5,$E3,$27,$AB,$E6,$3F,$57,$8F,$FB,$36
	.dc.b	$3C,$71,$83,$F6,$59,$FF,$3B,$E5,$F5,$FC,$AF,$E5,$7C,$3F,$66,$BF
	.dc.b	$D5,$71,$FF,$7D,$E3,$FE,$C5,$FA,$0F,$BD,$3F,$C4,$83,$87,$F8,$AB
	.dc.b	$5F,$F7,$63,$54,$3E,$4F,$F1,$AD,$3F,$D9,$A5,$8F,$EF,$AD,$7C,$1F
	.dc.b	$55,$FE,$7C,$8F,$F6,$9B,$63,$FC,$06,$D7,$7F,$15,$FC,$3B,$AF,$A6
	.dc.b	$BF,$EF,$8C,$9F,$31,$FC,$07,$2F,$BA,$FD,$EE,$2F,$96,$4C,$FE,$EB
	.dc.b	$CF,$E5,$3F,$07,$CF,$FC,$25,$B3,$FC,$27,$E4,$FF,$0F,$F2,$BF,$C3
	.dc.b	$5B,$1F,$90,$C7,$EB,$91,$FC,$4B,$F5,$B8,$7F,$2D,$BF,$E5,$3C,$8F
	.dc.b	$AD,$97,$E4,$9F,$E3,$1F,$27,$F9,$4F,$95,$B8,$FE,$71,$FC,$BE,$9F
	.dc.b	$84,$24,$7E,$59,$ED,$1F,$9C,$CF,$F0,$4D,$96,$2E,$0F,$ED,$1E,$23
	.dc.b	$CE,$FD,$F3,$6B,$73,$AF,$03,$35,$AF,$80,$FE,$17,$7F,$E5,$5B,$FF
	.dc.b	$2E,$C6,$0C,$FB,$F2,$3F,$CC,$2D,$C7,$F2,$14,$E6,$7C,$9F,$ED,$3E
	.dc.b	$0F,$F9,$D3,$70,$FE,$75,$E9,$61,$A7,$F3,$D8,$FE,$93,$E0,$BE,$4F
	.dc.b	$F7,$BC,$78,$6D,$7B,$0F,$F4,$04,$ED,$C6,$7F,$E3,$B8,$C3,$F1,$E3
	.dc.b	$7F,$27,$CD,$DE,$9F,$D1,$7B,$7C,$27,$8B,$B6,$CE,$B0,$47,$F6,$66
	.dc.b	$C7,$FB,$A9,$B8,$7C,$46,$BF,$E0,$6B,$FE,$A5,$F2,$7D,$6B,$A3,$67
	.dc.b	$CC,$7A,$7E,$4F,$F8,$EF,$CA,$FC,$46,$D7,$F9,$47,$EA,$BD,$B5,$FE
	.dc.b	$12,$E7,$46,$23,$F0,$2D,$7F,$10,$DC,$7F,$B2,$7C,$97,$F2,$AE,$1F
	.dc.b	$A2,$DC,$3E,$07,$C2,$FD,$43,$EA,$FC,$95,$8F,$9F,$8B,$EC,$BF,$A5
	.dc.b	$47,$F8,$91,$B1,$FE,$B5,$F8,$8F,$EA,$51,$88,$FF,$38,$7C,$4F,$E4
	.dc.b	$89,$5F,$EA,$E7,$F2,$BF,$09,$23,$FD,$62,$D9,$F9,$DF,$09,$FD,$0B
	.dc.b	$E5,$FF,$12,$DA,$FB,$0E,$1F,$B5,$F2,$96,$BE,$78,$FA,$15,$FF,$4C
	.dc.b	$EC,$FE,$D1,$18,$BF,$EA,$B1,$FD,$AF,$CB,$F9,$96,$69,$3B,$3F,$C5
	.dc.b	$7F,$65,$F9,$2E,$6F,$E5,$F5,$FE,$8F,$8C,$61,$F9,$AF,$96,$FF,$14
	.dc.b	$C6,$27,$F7,$FF,$80,$38,$FF,$C0,$2D,$BF,$F5,$7C,$A3,$0F,$FA,$C9
	.dc.b	$AF,$F4,$57,$FD,$2D,$1C,$1B,$1B,$DF,$A9,$F9,$DF,$86,$F1,$F8,$1F
	.dc.b	$B4,$6D,$FE,$47,$9F,$FA,$CF,$51,$A4,$8F,$D2,$7C,$AF,$B3,$C1,$FD
	.dc.b	$EB,$63,$F3,$5A,$72,$23,$FE,$CA,$DF,$F3,$71,$F2,$BC,$19,$3F,$EC
	.dc.b	$EE,$D7,$07,$EB,$B7,$FE,$4B,$18,$AF,$2F,$3C,$BE,$9A,$A3,$17,$FC
	.dc.b	$6B,$E3,$9F,$27,$FE,$5A,$6D,$FE,$BF,$83,$13,$FE,$49,$C3,$FD,$2D
	.dc.b	$3F,$C2,$AB,$77,$FE,$44,$9C,$BF,$BC,$C9,$3A,$D7,$F0,$71,$EA,$F0
	.dc.b	$BF,$0E,$4C,$3D,$DA,$A3,$E7,$EF,$27,$D6,$DC,$7F,$99,$D3,$E6,$BF
	.dc.b	$22,$4A,$E6,$FF,$9F,$B2,$FE,$8D,$5F,$07,$CA,$C1,$2B,$E7,$4F,$F4
	.dc.b	$59,$F5,$37,$3F,$4A,$F9,$4F,$91,$E5,$FC,$96,$DF,$C5,$6D,$FC,$3C
	.dc.b	$9E,$2F,$E9,$89,$5F,$01,$F3,$7C,$A4,$8F,$1B,$6B,$F4,$23,$E8,$FF
	.dc.b	$B2,$FD,$1B,$E4,$7F,$C4,$9F,$15,$FB,$08,$C5,$7F,$89,$EB,$75,$AC
	.dc.b	$2C,$E4,$FE,$07,$19,$69,$9F,$13,$9B,$FD,$D3,$0F,$DE,$C6,$23,$FB
	.dc.b	$BD,$FF,$78,$8C,$4E,$AB,$FC,$71,$5E,$93,$A9,$CE,$F8,$46,$AF,$CB
	.dc.b	$39,$18,$77,$79,$3D,$AB,$71,$E2,$3A,$9F,$E5,$72,$9E,$1C,$38,$38
	.dc.b	$AE,$38,$7C,$9C,$3C,$53,$E0,$7E,$26,$1E,$1A,$DE,$1E,$0A,$F1,$D6
	.dc.b	$8E,$35,$D8,$E1,$1C,$CF,$F7,$5D,$F7,$7E,$74,$FD,$1C,$36,$79,$8E
	.dc.b	$35,$CE,$3A,$9D,$E7,$89,$D2,$FE,$5B,$83,$E5,$57,$49,$EC,$7B,$9E
	.dc.b	$CF,$C0,$F0,$FF,$A2,$CF,$E1,$5A,$FF,$8D,$0F,$B4,$E1,$FA,$0E,$BE
	.dc.b	$3D,$E9,$FF,$8E,$65,$39,$8C,$37,$8B,$FC,$7E,$D9,$DE,$1D,$EF,$9D
	.dc.b	$7B,$25,$0F,$57,$C9,$5F,$96,$8C,$4E,$DE,$2F,$E4,$1B,$1B,$9E,$38
	.dc.b	$F7,$6D,$7F,$94,$AB,$73,$E3,$C7,$D9,$9F,$1A,$FC,$BD,$78,$1F,$BC
	.dc.b	$79,$A7,$75,$C7,$1F,$26,$1D,$AB,$D6,$DB,$EF,$1E,$1B,$E2,$F8,$E7
	.dc.b	$E9,$61,$E1,$8E,$F5,$EE,$8F,$85,$5C,$27,$A9,$E6,$AE,$38,$78,$E3
	.dc.b	$8C,$EE,$FD,$57,$D4,$E7,$87,$BA,$7E,$E6,$3E,$12,$61,$CD,$DE,$1E
	.dc.b	$4C,$7B,$BE,$93,$95,$C7,$EA,$47,$C1,$3E,$15,$F6,$BE,$87,$D5,$9F
	.dc.b	$92,$7D,$D8,$D3,$F1,$EB,$DC,$C3,$DA,$FF,$20,$E8,$FE,$AB,$E4,$7F
	.dc.b	$83,$1B,$3F,$D0,$A3,$1C,$A7,$5C,$6F,$9B,$A4,$AF,$EB,$2F,$0F,$91
	.dc.b	$FF,$2B,$70,$FF,$76,$C1,$F2,$58,$B6,$5F,$D9,$A3,$0F,$F6,$8B,$B5
	.dc.b	$C2,$BF,$B8,$56,$0C,$CE,$BE,$33,$CE,$49,$FE,$B6,$FA,$3C,$59,$1F
	.dc.b	$DB,$9A,$FF,$BA,$F5,$FF,$35,$E5,$FE,$2E,$6C,$7F,$82,$A3,$1D,$FF
	.dc.b	$EA,$AD,$5E,$87,$C4,$25,$7F,$7A,$79,$7D,$63,$E3,$3D,$7F,$1F,$2B
	.dc.b	$FD,$33,$92,$B5,$1F,$E0,$D9,$D3,$FD,$C9,$F9,$13,$BA,$B4,$F6,$FB
	.dc.b	$92,$FF,$C5,$91,$97,$FC,$1B,$F2,$FF,$68,$8E,$0D,$F8,$1E,$82,$38
	.dc.b	$3F,$C6,$5F,$40,$F8,$8F,$AF,$AF,$F9,$5E,$7F,$D5,$AD,$7F,$C4,$53
	.dc.b	$B7,$DF,$83,$D5,$F8,$A4,$8F,$3B,$F9,$39,$7C,$5F,$6F,$46,$57,$FA
	.dc.b	$EA,$72,$FF,$A5,$3E,$FB,$E4,$BF,$DB,$9A,$FF,$37,$0E,$1F,$E7,$4D
	.dc.b	$7F,$DD,$EF,$FB,$B5,$FF,$37,$B5,$E9,$E4,$7F,$84,$A3,$11,$FE,$BF
	.dc.b	$CB,$FC,$9C,$FC,$2F,$3B,$E2,$9F,$9C,$B7,$C5,$FE,$86,$5F,$47,$F9
	.dc.b	$D7,$FC,$0A,$D7,$D0,$D7,$FA,$83,$51,$CB,$AB,$E9,$FA,$D2,$3F,$D6
	.dc.b	$F2,$9F,$17,$C4,$C9,$F9,$4F,$97,$E3,$31,$1F,$E3,$7F,$9F,$FB,$A6
	.dc.b	$3A,$3F,$D3,$70,$FA,$27,$B3,$FE,$34,$6C,$7F,$A5,$9A,$FF,$88,$1A
	.dc.b	$FD,$35,$EC,$F1,$24,$71,$7F,$C7,$FF,$26,$F9,$1F,$90,$D9,$FF,$69
	.dc.b	$B5,$FF,$5E,$D7,$FB,$E1,$FC,$37,$C6,$FD,$12,$3F,$89,$F8,$CF,$A9
	.dc.b	$F1,$DF,$9A,$F7,$24,$FE,$36,$3F,$5B,$F2,$DF,$BA,$24,$7E,$B1,$FC
	.dc.b	$B7,$9D,$FA,$2F,$98,$93,$F9,$58,$FC,$07,$C7,$7E,$07,$E2,$E4,$FE
	.dc.b	$66,$2B,$F9,$5F,$3D,$D4,$AF,$D1,$12,$7F,$63,$F1,$9F,$8B,$23,$FA
	.dc.b	$38,$FC,$77,$C5,$48,$FD,$33,$F9,$1F,$8B,$C4,$7F,$9C,$FE,$07,$F7
	.dc.b	$26,$39,$FF,$D2,$BF,$BB,$BE,$23,$F3,$1E,$EC,$BF,$C7,$D6,$A7,$11
	.dc.b	$FE,$7D,$FA,$0F,$E0,$66,$27,$F0,$FC,$FF,$E8,$09,$FC,$27,$C1,$7E
	.dc.b	$1C,$8F,$E1,$FD,$9F,$C5,$11,$C1,$F8,$BF,$4F,$F5,$15,$FC,$C2,$3C
	.dc.b	$7E,$AF,$F8,$CE,$31,$1F,$82,$F5,$BF,$93,$F8,$BF,$B0,$8F,$E2,$B5
	.dc.b	$FC,$60,$CF,$8F,$FD,$43,$E6,$7F,$D8,$73,$FE,$CF,$6B,$F5,$5A,$FF
	.dc.b	$90,$79,$D1,$88,$FF,$69,$F0,$FE,$71,$F2,$9F,$D7,$A1,$F2,$7D,$CF
	.dc.b	$EB,$F3,$B9,$6E,$C7,$E7,$E4,$EF,$F3,$7A,$D2,$FF,$A4,$CE,$5E,$F9
	.dc.b	$F9,$3C,$EF,$EF,$9A,$7F,$16,$9F,$F7,$3B,$5F,$24,$7F,$80,$73,$FF
	.dc.b	$68,$47,$29,$FE,$90,$FF,$B6,$DB,$9F,$E1,$BE,$1F,$D3,$C7,$F2,$7E
	.dc.b	$1A,$5F,$8B,$0F,$C0,$3C,$7E,$4B,$C6,$47,$E7,$3E,$1E,$5F,$A5,$EB
	.dc.b	$C9,$E5,$1F,$C9,$FE,$87,$FA,$44,$7F,$AA,$83,$87,$FD,$5A,$D7,$FA
	.dc.b	$03,$5F,$F3,$C7,$D0,$BE,$97,$FD,$96,$D8,$E8,$D7,$FD,$6E,$D7,$FA
	.dc.b	$53,$FE,$7C,$77,$E2,$72,$23,$FE,$BE,$DB,$F9,$EF,$83,$F8,$DF,$8A
	.dc.b	$93,$F6,$23,$F2,$1F,$27,$F5,$89,$1F,$91,$7E,$CF,$B5,$F9,$52,$47
	.dc.b	$E0,$3C,$4F,$C4,$FB,$3F,$99,$F8,$49,$3F,$46,$3F,$39,$F4,$1F,$82
	.dc.b	$24,$7E,$75,$FC,$2F,$C2,$7E,$84,$91,$F8,$CF,$1B,$FA,$B7,$99,$FD
	.dc.b	$BA,$3F,$2B,$E0,$7F,$A8,$61,$F2,$BA,$37,$EF,$70,$F1,$3F,$DE,$F3
	.dc.b	$FB,$1B,$FF,$70,$B5,$FE,$E8,$D7,$F2,$BE,$D7,$EC,$C8,$FE,$A8,$7D
	.dc.b	$3B,$E0,$BF,$B5,$7D,$1D,$9F,$C2,$35,$FF,$77,$B5,$FF,$01,$35,$FF
	.dc.b	$5D,$3F,$E1,$77,$3F,$27,$F8,$D5,$19,$77,$1F,$E4,$5F,$6F,$F7,$1D
	.dc.b	$1F,$CA,$3D,$CF,$DD,$9F,$E3,$68,$CE,$FF,$C1,$23,$37,$FA,$57,$5E
	.dc.b	$E7,$F0,$2F,$4B,$F8,$1E,$D2,$BF,$55,$15,$FA,$27,$F9,$A3,$C3,$4E
	.dc.b	$72,$FE,$73,$5F,$C3,$31,$FF,$35,$FC,$2F,$F1,$A9,$FE,$19,$87,$F9
	.dc.b	$C2,$31,$3F,$C6,$0D,$FC,$97,$FA,$0E,$3F,$8E,$4E,$E3,$07,$F0,$CF
	.dc.b	$8B,$91,$FC,$9D,$AF,$F5,$B9,$FF,$14,$5F,$83,$CC,$FF,$2B,$79,$DE
	.dc.b	$43,$CB,$E6,$DB,$FF,$1B,$1E,$58,$F1,$11,$FE,$01,$F4,$0F,$77,$FD
	.dc.b	$D7,$1C,$23,$1D,$7A,$9C,$CF,$F9,$06,$72,$F6,$F9,$07,$CD,$3E,$87
	.dc.b	$F8,$64,$61,$B1,$99,$E0,$D7,$A7,$13,$E4,$FC,$AD,$8F,$23,$5D,$C7
	.dc.b	$F4,$DF,$80,$FA,$7B,$FC,$5F,$51,$F1,$F8,$E5,$FE,$6F,$57,$0F,$E9
	.dc.b	$53,$78,$BF,$A8,$CD,$B3,$7E,$17,$0E,$7E,$3B,$FC,$73,$87,$F4,$26
	.dc.b	$BF,$D5,$5B,$7F,$D7,$1B,$7F,$E4,$EC,$7F,$A2,$47,$7A,$E9,$1F,$D1
	.dc.b	$DC,$3F,$D2,$73,$B1,$F9,$8B,$7C,$DF,$F9,$F2,$BF,$AB,$BF,$E8,$B6
	.dc.b	$F8,$9E,$D8,$C5,$71,$78,$36,$35,$E8,$ED,$F8,$DB,$3E,$7B,$E5,$E1
	.dc.b	$DB,$C7,$CC,$63,$0E,$91,$8E,$09,$5E,$4F,$1E,$D7,$EC,$07,$D0,$FF
	.dc.b	$AE,$E7,$FA,$93,$63,$F3,$2D,$7F,$D2,$0D,$7F,$CE,$E7,$2E,$BF,$B8
	.dc.b	$F4,$EB,$B2,$9B,$1A,$F8,$9F,$F5,$F5,$FF,$98,$3B,$78,$FD,$03,$E7
	.dc.b	$35,$18,$F7,$FE,$EE,$6A,$FF,$05,$5F,$F6,$96,$C7,$EC,$1A,$FF,$A9
	.dc.b	$D7,$C5,$76,$3C,$FF,$55,$FE,$15,$AF,$E6,$93,$A7,$FD,$B7,$C3,$FE
	.dc.b	$77,$6F,$5F,$DB,$F8,$DF,$A2,$FC,$0F,$98,$FD,$D4,$FF,$7B,$6E,$5F
	.dc.b	$C0,$9A,$FF,$AE,$5A,$FF,$A8,$3A,$7F,$17,$B7,$E1,$CF,$C4,$7C,$7D
	.dc.b	$BF,$1F,$5F,$E0,$0F,$C6,$6F,$63,$FA,$66,$3F,$CA,$9F,$17,$7F,$C8
	.dc.b	$8C,$65,$E3,$8F,$D3,$F4,$FF,$8E,$BE,$57,$F1,$FF,$0D,$F9,$68,$FE
	.dc.b	$3D,$DD,$F0,$EF,$F9,$BC,$6F,$1F,$EB,$46,$CF,$C1,$FD,$0F,$EB,$67
	.dc.b	$F1,$D1,$97,$FD,$7B,$19,$9F,$F5,$8C,$FF,$18,$3F,$C4,$45,$9F,$E5
	.dc.b	$2D,$7F,$DC,$0D,$7F,$23,$F4,$2F,$9E,$FC,$8B,$FE,$B2,$9F,$ED,$CE
	.dc.b	$5B,$E7,$FE,$CF,$C6,$FD,$A7,$C9,$D9,$FD,$CB,$CB,$E2,$7F,$1F,$66
	.dc.b	$92,$7F,$9E,$F2,$FF,$02,$F9,$BF,$DF,$1F,$BF,$F8,$3B,$3F,$E6,$8D
	.dc.b	$7F,$CC,$FC,$DF,$C0,$FE,$53,$E7,$B6,$7C,$FE,$BF,$80,$8F,$F9,$17
	.dc.b	$34,$7D,$8E,$3F,$CF,$E1,$FC,$C1,$B5,$F8,$BE,$37,$C9,$FD,$CE,$8F
	.dc.b	$BB,$C1,$FC,$4F,$E1,$2D,$FF,$18,$7E,$4F,$C0,$7C,$02,$47,$85,$F0
	.dc.b	$6F,$F8,$63,$F8,$E7,$C9,$D9,$FF,$0E,$71,$7E,$F3,$9F,$F8,$73,$76
	.dc.b	$FF,$77,$AF,$8C,$FE,$0E,$E6,$BF,$D2,$5F,$19,$FB,$E8,$34,$91,$FE
	.dc.b	$67,$87,$C6,$72,$9E,$6D,$9F,$E6,$7E,$C7,$F3,$1F,$86,$FE,$64,$D8
	.dc.b	$C9,$32,$FE,$69,$E0,$6A,$3F,$D2,$3E,$87,$FB,$25,$BF,$27,$FD,$5E
	.dc.b	$1F,$01,$63,$0D,$9E,$2D,$BF,$E9,$15,$F9,$36,$F7,$F3,$F0,$3A,$3D
	.dc.b	$7B,$F1,$87,$89,$B5,$CB,$E7,$B5,$3E,$66,$C7,$ED,$BE,$6F,$D0,$F8
	.dc.b	$3E,$43,$FA,$5C,$63,$93,$FC,$35,$9A,$7D,$9F,$21,$AD,$AE,$5F,$99
	.dc.b	$C3,$83,$FC,$5E,$DF,$29,$F0,$FE,$4F,$E2,$36,$7E,$4F,$83,$F1,$CF
	.dc.b	$91,$F1,$16,$3F,$0F,$F1,$FF,$82,$3F,$07,$F1,$F6,$3F,$17,$F1,$FF
	.dc.b	$83,$23,$F8,$17,$F0,$DF,$1F,$F8,$A7,$F1,$6D,$8F,$B3,$E0,$7D,$73
	.dc.b	$EC,$7C,$C5,$8F,$BA,$DE,$3F,$AA,$DE,$3F,$F2,$6C,$7F,$53,$E1,$FE
	.dc.b	$BB,$F4,$0E,$AF,$EB,$FE,$97,$C4,$A0,$9D,$63,$FE,$7F,$5D,$9C,$30
	.dc.b	$E6,$D9,$FC,$7F,$CC,$FE,$34,$FE,$D7,$F4,$1F,$C0,$23,$27,$FA,$25
	.dc.b	$6B,$FD,$91,$B1,$FE,$CB,$6E,$3F,$E5,$26,$BF,$EF,$FE,$67,$C7,$F6
	.dc.b	$BF,$CE,$36,$FF,$01,$CF,$F9,$FF,$8B,$FC,$4B,$63,$F4,$BE,$AE,$1A
	.dc.b	$70,$3F,$11,$D1,$FA,$7F,$8E,$FC,$8B,$63,$F5,$BF,$33,$86,$9C,$0F
	.dc.b	$C8,$72,$FF,$56,$EC,$34,$91,$FC,$CB,$5F,$EF,$5B,$7E,$86,$7F,$6D
	.dc.b	$3F,$A0,$7F,$43,$A7,$F9,$3F,$E7,$9F,$21,$8A,$FB,$3D,$1F,$E7,$D9
	.dc.b	$CC,$7E,$86,$B5,$87,$F9,$06,$5E,$96,$2F,$FC,$C3,$9B,$6F,$41,$31
	.dc.b	$FF,$75,$46,$5F,$E9,$37,$FE,$2F,$AE,$35,$E5,$F8,$3F,$DB,$BE,$7F
	.dc.b	$84,$FF,$2D,$D7,$F8,$AE,$F8,$C6,$0C,$FC,$1E,$E3,$D6,$7F,$83,$67
	.dc.b	$F9,$76,$6F,$35,$1C,$E3,$D9,$E6,$7C,$5F,$3B,$E3,$36,$F9,$CF,$1A
	.dc.b	$FE,$63,$18,$BE,$75,$F3,$7C,$EE,$33,$9A,$7C,$B7,$A8,$91,$F0,$9F
	.dc.b	$6F,$8B,$F5,$5F,$AC,$D8,$FF,$74,$63,$FA,$EF,$A0,$FD,$7F,$D0,$3E
	.dc.b	$4F,$FC,$A7,$3F,$E9,$9A,$CC,$7F,$AC,$AA,$CF,$F9,$63,$E0,$7F,$D9
	.dc.b	$37,$C6,$38,$CF,$EE,$BC,$6C,$47,$EF,$2B,$FD,$9F,$18,$8F,$11,$FC
	.dc.b	$0E,$FC,$33,$A7,$C1,$0B,$E1,$3E,$09,$CC,$63,$0F,$2D,$7F,$9B,$30
	.dc.b	$AF,$AC,$E5,$5C,$0E,$F3,$A8,$4F,$A6,$73,$1E,$53,$DA,$53,$D9,$3B
	.dc.b	$73,$5E,$37,$B9,$89,$E7,$1F,$06,$3B,$9A,$39,$94,$F8,$AF,$33,$85
	.dc.b	$7B,$2F,$E1,$4F,$53,$EA,$5F,$AE,$7C,$91,$D3,$0E,$2F,$A4,$F8,$8E
	.dc.b	$2F,$B6,$1F,$16,$79,$9E,$33,$CF,$3E,$D3,$99,$63,$EB,$B6,$3E,$3B
	.dc.b	$5F,$A8,$7F,$3A,$7E,$EF,$87,$FD,$0E,$3E,$6B,$5D,$B7,$E4,$9E,$B3
	.dc.b	$9F,$FA,$3C,$E6,$31,$5D,$4F,$E9,$75,$C2,$B7,$7F,$EB,$08,$C4,$EC
	.dc.b	$38,$8F,$CC,$AF,$7F,$FD,$6F,$18,$6E,$3C,$F0,$F0,$CF,$06,$C7,$79
	.dc.b	$DD,$F9,$1A,$F8,$27,$6F,$A1,$F5,$E5,$DE,$3F,$3B,$87,$8D,$B3,$AB
	.dc.b	$F1,$63,$F8,$0A,$D5,$7E,$5D,$B8,$F9,$67,$57,$C6,$BC,$87,$6C,$7F
	.dc.b	$DA,$B8,$EE,$3F,$2D,$87,$C8,$8F,$24,$62,$77,$8F,$8A,$37,$7D,$5F
	.dc.b	$9B,$3A,$7B,$9A,$C3,$D6,$7A,$63,$E2,$E1,$E8,$BE,$98,$F8,$F1,$F0
	.dc.b	$92,$3C,$0E,$5F,$A1,$87,$AC,$EA,$77,$3E,$25,$F6,$3B,$E3,$F4,$A3
	.dc.b	$EA,$63,$C5,$F7,$1C,$AF,$F9,$07,$1F,$F4,$2F,$33,$FD,$F7,$5F,$D8
	.dc.b	$7C,$CB,$BF,$F4,$96,$E1,$FD,$06,$AD,$FF,$4A,$9F,$F3,$A7,$19,$A4
	.dc.b	$AF,$EB,$FF,$2F,$C2,$4D,$31,$E6,$D9,$FE,$91,$F3,$1B,$30,$FF,$71
	.dc.b	$DF,$97,$C7,$48,$FE,$9D,$15,$FE,$87,$97,$F5,$E9,$F2,$92,$3F,$B8
	.dc.b	$FC,$37,$B3,$67,$81,$FE,$40,$7F,$BD,$EF,$E7,$9D,$4F,$F9,$12,$BE
	.dc.b	$3B,$5F,$F0,$4B,$5F,$DF,$36,$9F,$40,$F0,$7C,$0D,$AF,$F6,$33,$5F
	.dc.b	$F0,$77,$CA,$6A,$7E,$47,$C9,$C9,$DC,$7C,$BE,$BE,$7E,$0C,$9F,$EE
	.dc.b	$11,$5F,$EC,$B8,$F8,$A7,$FB,$3C,$7F,$88,$31,$46,$23,$F8,$B4,$61
	.dc.b	$FF,$0B,$9E,$8F,$02,$3F,$DF,$4A,$7F,$1D,$D7,$FC,$C7,$BB,$E7,$9F
	.dc.b	$15,$F8,$86,$D7,$C0,$BC,$7C,$A7,$C4,$E9,$93,$FE,$0D,$A4,$E2,$3B
	.dc.b	$1F,$E1,$B8,$7C,$A7,$53,$5A,$B3,$F1,$CB,$CC,$FC,$DD,$BE,$83,$E8
	.dc.b	$5F,$FA,$01,$B7,$C2,$B7,$1F,$56,$BF,$C1,$ED,$7F,$D0,$27,$3E,$1F
	.dc.b	$94,$D3,$E7,$ED,$F0,$FE,$57,$E1,$12,$3F,$C5,$A7,$8F,$C9,$F2,$FC
	.dc.b	$F3,$B2,$F6,$4A,$DC,$E5,$B1,$C3,$B6,$46,$0F,$95,$F2,$1F,$47,$E3
	.dc.b	$64,$60,$F9,$CE,$1F,$9B,$9F,$D9,$8F,$B1,$3F,$23,$E7,$DE,$DD,$F8
	.dc.b	$69,$87,$AA,$72,$D8,$F8,$71,$FE,$E8,$8F,$C7,$35,$FC,$9B,$5F,$F1
	.dc.b	$DA,$F8,$6F,$01,$23,$B7,$C3,$7E,$2F,$E2,$E4,$7E,$1D,$FC,$77,$C0
	.dc.b	$FE,$3F,$E2,$FF,$21,$F2,$B2,$BF,$19,$1F,$93,$F9,$1F,$D9,$F8,$B2
	.dc.b	$7F,$21,$15,$F8,$7D,$9F,$DE,$9F,$CA,$79,$B9,$ED,$FC,$74,$7E,$4F
	.dc.b	$CE,$FC,$F7,$C8,$C8,$FC,$D4,$57,$9F,$C7,$7E,$8C,$C5,$72,$E6,$FD
	.dc.b	$44,$7E,$9F,$E3,$E5,$7E,$8E,$3F,$55,$F1,$FF,$AB,$F8,$F9,$3F,$A7
	.dc.b	$8A,$FA,$22,$E1,$D7,$45,$7F,$3A,$F9,$BF,$AF,$7F,$80,$6D,$78,$DB
	.dc.b	$E2,$7E,$AB,$D5,$E1,$D5,$FB,$BF,$5E,$61,$FE,$83,$BD,$BF,$EE,$4E
	.dc.b	$A7,$C7,$FE,$27,$E3,$E4,$FE,$1E,$2B,$F8,$28,$FC,$39,$1F,$C3,$FC
	.dc.b	$3F,$E4,$48,$FE,$3F,$E3,$FF,$9D,$78,$8F,$1E,$92,$7C,$9E,$0F,$F0
	.dc.b	$0F,$9F,$FF,$60,$C7,$F1,$F6,$BF,$C9,$1A,$FF,$A3,$1A,$FF,$B1,$A9
	.dc.b	$E2,$7C,$9F,$0A,$4F,$FA,$9B,$8D,$5A,$8C,$2F,$0B,$CE,$48,$F9,$1E
	.dc.b	$1F,$97,$E0,$FF,$84,$E5,$FE,$CE,$9F,$34,$60,$FE,$6D,$87,$6F,$0B
	.dc.b	$ED,$7C,$8C,$AE,$F1,$F7,$78,$64,$7D,$57,$F0,$1C,$5F,$0F,$E0,$7B
	.dc.b	$92,$7E,$23,$FD,$0F,$BD,$CC,$E5,$FC,$47,$2C,$BF,$C0,$47,$CD,$F8
	.dc.b	$1F,$9D,$E9,$49,$FE,$3B,$C0,$9F,$E6,$3F,$3F,$FD,$0A,$7F,$A9,$B5
	.dc.b	$FF,$5D,$B5,$FE,$5A,$D4,$17,$8D,$D3,$C6,$93,$FC,$B2,$72,$FD,$2F
	.dc.b	$0F,$F0,$E4,$7F,$31,$F2,$5F,$D1,$3E,$B9,$F1,$1F,$CE,$9B,$3F,$D9
	.dc.b	$5A,$FF,$6A,$6B,$FE,$2A,$EA,$78,$DF,$94,$24,$7E,$97,$E4,$FF,$61
	.dc.b	$F2,$12,$3F,$50,$FE,$CF,$E3,$BC,$5F,$1F,$E1,$24,$F8,$5F,$C0,$E0
	.dc.b	$F1,$7F,$42,$49,$FD,$97,$C7,$7F,$5E,$AF,$EC,$B1,$FD,$39,$AF,$F6
	.dc.b	$A0,$E3,$BB,$F8,$DC,$3E,$BF,$1E,$57,$F5,$EA,$FE,$D5,$3F,$E0,$76
	.dc.b	$BF,$E0,$C6,$BF,$EF,$46,$BF,$D9,$BE,$85,$AF,$F7,$26,$C7,$F6,$26
	.dc.b	$BE,$59,$FD,$E3,$FD,$9B,$E8,$5C,$5F,$DE,$9D,$E5,$E8,$6E,$3F,$DD
	.dc.b	$1D,$D7,$1C,$7F,$BA,$C7,$68,$FE,$F2,$1C,$3C,$63,$F5,$FE,$43,$C7
	.dc.b	$FF,$24,$36,$7F,$80,$35,$FE,$55,$F2,$BA,$9F,$E5,$DF,$29,$27,$71
	.dc.b	$FC,$C7,$E4,$A4,$7F,$29,$7F,$9A,$73,$E2,$3F,$9A,$6D,$FC,$DC,$C6
	.dc.b	$BF,$CD,$75,$FE,$74,$49,$FE,$65,$F2,$95,$FE,$F7,$F0,$DF,$C7,$AB
	.dc.b	$FA,$73,$6B,$FC,$6F,$E1,$FF,$98,$5F,$F1,$CC,$7F,$20,$A7,$89,$FE
	.dc.b	$46,$F9,$4F,$E0,$93,$98,$C4,$FF,$87,$A1,$EF,$FF,$A1,$9B,$1F,$B8
	.dc.b	$C7,$C3,$39,$F3,$34,$EF,$E0,$25,$6F,$2F,$1C,$FF,$8E,$3E,$87,$F6
	.dc.b	$BD,$7F,$E5,$76,$D7,$F9,$71,$AF,$DC,$6B,$FE,$E9,$F1,$1F,$43,$FE
	.dc.b	$5F,$F9,$3F,$E1,$2D,$8C,$CF,$F9,$9A,$72,$F5,$E2,$7C,$5F,$85,$B1
	.dc.b	$98,$C1,$FE,$40,$FB,$27,$81,$FE,$06,$E0,$FA,$3C,$37,$7F,$F3,$3D
	.dc.b	$69,$F8,$18,$7D,$09,$FF,$3A,$C6,$38,$94,$EA,$E5,$7F,$5A,$39,$3F
	.dc.b	$E3,$A7,$55,$C5,$CE,$1F,$82,$F9,$7F,$F4,$D6,$1F,$E8,$26,$E2,$69
	.dc.b	$AA,$FE,$4B,$7E,$39,$D3,$D7,$A3,$C7,$7C,$E3,$C1,$AE,$2F,$8E,$5E
	.dc.b	$58,$C5,$E7,$A3,$8E,$2F,$97,$C6,$3D,$DF,$7E,$8F,$C2,$FE,$2E,$5E
	.dc.b	$FC,$33,$18,$9F,$F1,$66,$69,$F8,$F4,$FA,$27,$4E,$3E,$CD,$DF,$A9
	.dc.b	$CE,$9F,$E7,$69,$CB,$FE,$78,$8F,$F5,$E3,$FE,$A9,$6D,$FF,$AB,$9A
	.dc.b	$FF,$0E,$6B,$FA,$3C,$FD,$D3,$FE,$B4,$CF,$D9,$A7,$FB,$35,$B1,$99
	.dc.b	$DD,$7F,$B7,$BE,$8D,$DB,$F0,$30,$FA,$31,$8B,$FF,$0A,$5F,$0C,$3F
	.dc.b	$80,$FC,$87,$CE,$37,$DD,$6B,$7E,$87,$86,$FF,$BC,$47,$86,$30,$7E
	.dc.b	$23,$0F,$9D,$1F,$EC,$07,$F8,$0B,$F9,$77,$39,$FF,$0A,$72,$FF,$AE
	.dc.b	$B1,$FB,$5E,$97,$D8,$3D,$39,$5B,$F4,$FC,$CF,$8C,$8F,$C3,$7C,$DF
	.dc.b	$19,$F8,$8F,$5F,$F0,$2D,$9F,$83,$F3,$1F,$84,$3E,$2D,$58,$F8,$BF
	.dc.b	$2F,$EF,$23,$F3,$DF,$85,$F2,$DF,$94,$DB,$F1,$AD,$9F,$77,$C7,$FE
	.dc.b	$63,$4F,$A2,$D8,$FF,$22,$7E,$13,$FE,$DA,$7E,$F3,$77,$FF,$7A,$B5
	.dc.b	$FF,$65,$EE,$8C,$47,$CD,$CB,$F4,$B9,$FD,$E6,$DF,$CF,$CB,$F5,$3F
	.dc.b	$23,$F8,$26,$C7,$8B,$E4,$3F,$19,$89,$A4,$8F,$F6,$BF,$C1,$7F,$BE
	.dc.b	$1F,$DF,$70,$7E,$F2,$31,$9F,$F4,$C8,$FD,$9D,$FF,$8C,$AB,$F8,$26
	.dc.b	$B8,$F1,$FF,$76,$EB,$E3,$7C,$24,$9F,$86,$F9,$3E,$36,$47,$C6,$6E
	.dc.b	$3F,$1B,$E3,$3F,$82,$F2,$78,$1B,$5F,$3B,$E3,$3D,$A4,$7E,$7B,$FA
	.dc.b	$BF,$C2,$3F,$DB,$31,$FE,$77,$6D,$FF,$B8,$36,$F8,$93,$9F,$8D,$FE
	.dc.b	$27,$F3,$52,$3F,$99,$4F,$F9,$ED,$FE,$45,$1F,$C3,$3C,$93,$FE,$2A
	.dc.b	$BF,$F2,$B9,$FD,$A9,$C3,$FE,$5A,$8D,$CF,$F9,$BE,$7F,$99,$D3,$E4
	.dc.b	$BA,$47,$D1,$6C,$F8,$B8,$3F,$93,$F3,$7D,$56,$C7,$9B,$83,$0D,$38
	.dc.b	$1F,$07,$E2,$7F,$97,$FC,$97,$F2,$CD,$91,$89,$D1,$35,$F8,$DF,$25
	.dc.b	$B9,$FC,$27,$1C,$AF,$99,$EE,$DB,$FD,$F6,$1F,$E3,$89,$FE,$3C,$D8
	.dc.b	$FF,$18,$F6,$23,$11,$FA,$26,$BF,$1F,$81,$F0,$BE,$7C,$3F,$28,$DC
	.dc.b	$3E,$E7,$C1,$7F,$4A,$EC,$F6,$36,$3E,$CF,$36,$1A,$70,$3E,$EF,$B1
	.dc.b	$F9,$F3,$E2,$7A,$76,$3F,$10,$E1,$FE,$E9,$F4,$AF,$7B,$1A,$FF,$8C
	.dc.b	$3D,$7F,$EB,$67,$F5,$0F,$07,$57,$FD,$A6,$3F,$11,$A7,$E1,$2B,$FC
	.dc.b	$5E,$DE,$B3,$4C,$4F,$0A,$FF,$03,$D7,$F8,$DA,$30,$FD,$0F,$53,$FB
	.dc.b	$5B,$F7,$5B,$3F,$A1,$70,$FA,$BD,$EB,$3F,$63,$DE,$FE,$F7,$FC,$03
	.dc.b	$FC,$03,$9B,$FC,$35,$CC,$9C,$FC,$D7,$F8,$03,$81,$CF,$F8,$B9,$C3
	.dc.b	$73,$FF,$0D,$63,$FC,$87,$14,$7E,$8E,$30,$FE,$96,$DF,$1D,$E2,$F9
	.dc.b	$AF,$EF,$3F,$41,$FE,$1E,$D7,$FC,$F2,$D7,$F5,$ED,$7F,$D8,$ED,$4F
	.dc.b	$F8,$91,$FF,$51,$3F,$E2,$4B,$E3,$CB,$DB,$DE,$FF,$07,$DF,$F6,$BB
	.dc.b	$67,$E8,$7F,$CC,$B9,$BE,$6F,$7E,$B7,$C8,$7E,$27,$C8,$63,$3F,$C3
	.dc.b	$B6,$FF,$B1,$63,$FE,$08,$C7,$A7,$3B,$27,$CD,$F8,$BD,$7F,$A6,$FC
	.dc.b	$FF,$87,$C6,$7F,$92,$6E,$5D,$3E,$03,$FC,$15,$9F,$F8,$9A,$3F,$A6
	.dc.b	$B5,$FF,$3A,$35,$FF,$3F,$F1,$BC,$0F,$83,$5F,$75,$B8,$7F,$95,$B5
	.dc.b	$61,$FE,$0D,$8F,$B5,$F2,$FF,$23,$DB,$FC,$4B,$63,$84,$EF,$2E,$FC
	.dc.b	$79,$7E,$1F,$B1,$E8,$25,$7F,$89,$69,$1F,$E5,$8D,$3E,$D6,$5F,$E6
	.dc.b	$D9,$CD,$7E,$D3,$97,$EC,$DF,$E6,$BE,$6B,$FC,$BF,$C4,$69,$2B,$FD
	.dc.b	$87,$97,$C2,$EE,$E1,$1F,$BE,$6C,$7E,$0A,$3F,$2D,$D7,$FA,$6F,$98
	.dc.b	$FD,$13,$63,$F8,$2F,$C1,$61,$A7,$F0,$88,$FD,$6F,$87,$E4,$24,$78
	.dc.b	$BA,$3F,$05,$1F,$E7,$6F,$D3,$7F,$25,$C1,$F8,$6D,$7F,$AB,$E7,$E3
	.dc.b	$3F,$D6,$F3,$AC,$3F,$59,$7F,$E6,$09,$0F,$A1,$BF,$42,$F5,$FF,$DC
	.dc.b	$7E,$37,$E7,$63,$15,$FE,$E6,$AD,$D6,$B0,$B3,$91,$E5,$70,$7B,$E1
	.dc.b	$FA,$78,$C4,$78,$AF,$C7,$18,$9D,$5F,$D2,$9E,$B5,$CA,$EB,$C4,$61
	.dc.b	$C6,$33,$1E,$D3,$35,$E7,$69,$E7,$9E,$0E,$0F,$EE,$F5,$E4,$9E,$18
	.dc.b	$7A,$27,$A3,$C0,$C4,$F9,$30,$F9,$15,$E9,$AE,$87,$FB,$23,$0F,$2D
	.dc.b	$6D,$B1,$AA,$E2,$F9,$6B,$D1,$87,$68,$D5,$70,$C3,$C5,$5D,$DB,$3C
	.dc.b	$4F,$F0,$8D,$F4,$7D,$A7,$23,$CA,$7B,$27,$DE,$73,$2B,$F7,$67,$AB
	.dc.b	$C9,$AF,$D3,$F2,$BF,$02,$FE,$0D,$AF,$81,$AE,$E3,$FC,$4B,$F1,$2F
	.dc.b	$96,$FE,$5B,$2A,$E3,$39,$DF,$F3,$31,$8A,$DB,$63,$37,$D2,$79,$BF
	.dc.b	$93,$A7,$A7,$FE,$34,$8C,$37,$1E,$17,$F8,$1C,$7F,$4F,$7E,$17,$F3
	.dc.b	$B3,$E8,$AE,$11,$A6,$A9,$E1,$7E,$1C,$3D,$BA,$78,$2B,$F3,$B5,$E4
	.dc.b	$6C,$EB,$0F,$1C,$78,$0E,$C7,$94,$ED,$87,$82,$BB,$CF,$0B,$ED,$5F
	.dc.b	$9D,$9E,$CD,$BE,$CF,$B9,$AF,$43,$CB,$7F,$E5,$AA,$F5,$63,$E6,$33
	.dc.b	$30,$E8,$E5,$F4,$1C,$CB,$7E,$2C,$79,$C6,$30,$F0,$CF,$9A,$FD,$26
	.dc.b	$89,$8F,$D1,$AF,$04,$61,$F5,$61,$F1,$6B,$B5,$6E,$7B,$4E,$6F,$BE
	.dc.b	$3E,$DA,$D6,$1C,$F8,$5C,$0F,$8D,$FF,$5A,$EF,$FE,$17,$F9,$BB,$A7
	.dc.b	$F8,$09,$B7,$C2,$AF,$1A,$7F,$A6,$D6,$CF,$E7,$D1,$96,$C6,$AB,$C3
	.dc.b	$DF,$D3,$17,$C6,$53,$11,$6F,$FC,$5B,$EC,$F8,$0C,$3F,$D1,$6F,$D7
	.dc.b	$D9,$E4,$F9,$A9,$3F,$D1,$38,$BF,$C2,$39,$7F,$6E,$8F,$F1,$BC,$F8
	.dc.b	$F4,$34,$91,$FD,$5E,$5F,$07,$F1,$2F,$FC,$0A,$DF,$03,$35,$DF,$83
	.dc.b	$FD,$45,$78,$8F,$F1,$1D,$7C,$6B,$E3,$1F,$E0,$18,$C5,$EE,$3F,$C3
	.dc.b	$7F,$05,$D4,$8F,$53,$FB,$1F,$C1,$E1,$F8,$59,$BD,$CF,$AA,$67,$1F
	.dc.b	$83,$F0,$7E,$7F,$82,$97,$CF,$83,$F0,$73,$88,$FF,$4C,$E6,$AF,$29
	.dc.b	$23,$E2,$73,$FC,$1F,$80,$91,$CF,$7F,$EC,$D9,$7F,$86,$67,$F0,$FD
	.dc.b	$52,$BF,$C4,$F1,$F8,$1F,$80,$F9,$5F,$01,$27,$94,$57,$DB,$D9,$F9
	.dc.b	$03,$18,$7F,$95,$E1,$F3,$CF,$53,$F3,$31,$8F,$8C,$B2,$6C,$E1,$FC
	.dc.b	$CC,$EC,$FC,$DC,$ED,$B1,$F8,$7F,$82,$FD,$09,$23,$E0,$7B,$BF,$9B
	.dc.b	$7F,$08,$43,$E2,$14,$FA,$FF,$03,$FA,$A3,$18,$FF,$9C,$47,$D8,$7F
	.dc.b	$94,$A7,$FC,$A8,$D7,$FC,$49,$1A,$E3,$FD,$AF,$F0,$0B,$DE,$FC,$B5
	.dc.b	$EE,$BF,$C5,$9F,$05,$FB,$83,$37,$F1,$BF,$40,$FE,$53,$3F,$2D,$69
	.dc.b	$93,$5C,$17,$3F,$DF,$CE,$9B,$3F,$2A,$31,$1F,$E3,$43,$FD,$45,$F0
	.dc.b	$FF,$C2,$08,$FF,$AF,$BE,$B5,$EE,$FA,$3D,$99,$3F,$01,$F6,$FB,$FE
	.dc.b	$EF,$7A,$4F,$E6,$78,$7F,$D1,$6E,$CC,$7B,$1F,$0B,$D7,$91,$83,$E5
	.dc.b	$BF,$C6,$BE,$6E,$46,$0E,$9E,$B7,$CA,$F3,$24,$60,$FA,$6F,$EE,$F9
	.dc.b	$4D,$24,$7F,$97,$F1,$FE,$44,$FF,$17,$6F,$1F,$FA,$91,$BE,$57,$F1
	.dc.b	$7B,$E7,$5F,$80,$F1,$BE,$D9,$1F,$81,$85,$7F,$80,$FA,$7F,$CD,$0C
	.dc.b	$63,$D7,$C2,$FC,$2D,$CA,$F9,$11,$5F,$E1,$3E,$A7,$F3,$C3,$15,$E5
	.dc.b	$E7,$FC,$37,$34,$AF,$A1,$15,$FF,$46,$F3,$A7,$F0,$44,$8F,$F4,$07
	.dc.b	$C9,$3E,$2B,$FA,$5B,$63,$F8,$93,$5F,$8D,$DD,$F9,$AA,$91,$FA,$27
	.dc.b	$F3,$FD,$BF,$9E,$DB,$E9,$12,$7F,$D3,$F9,$FE,$AA,$3F,$98,$7E,$11
	.dc.b	$D6,$3F,$DA,$78,$BE,$27,$C6,$7F,$3D,$9D,$71,$FF,$49,$5F,$13,$C8
	.dc.b	$98,$F6,$8C,$B7,$E2,$3B,$91,$C1,$BF,$8B,$F8,$35,$23,$FD,$A1,$49
	.dc.b	$FE,$DF,$DD,$E5,$9F,$8B,$F4,$3F,$20,$CE,$5E,$7F,$87,$F6,$92,$7C
	.dc.b	$8F,$F6,$FA,$65,$F4,$09,$3F,$D0,$63,$2D,$C3,$EA,$FC,$54,$8C,$1E
	.dc.b	$6F,$83,$F9,$64,$7E,$13,$DF,$E3,$3E,$ED,$49,$ED,$1F,$37,$6F,$C0
	.dc.b	$FD,$0C,$9F,$83,$1F,$DB,$7F,$08,$FE,$61,$1F,$C2,$47,$12,$33,$1F
	.dc.b	$C3,$63,$1E,$67,$F6,$D9,$CC,$FF,$80,$5F,$F0,$F3,$FE,$FD,$6B,$FC
	.dc.b	$A9,$B2,$AF,$F6,$5F,$AC,$D5,$F0,$FE,$5F,$7F,$F3,$3F,$19,$32,$E9
	.dc.b	$1E,$9F,$80,$F1,$12,$3F,$27,$9A,$3C,$DF,$23,$EF,$24,$FD,$3F,$88
	.dc.b	$FC,$0F,$C4,$FF,$8C,$3F,$80,$7F,$AE,$4F,$F7,$76,$BF,$E2,$1D,$13
	.dc.b	$F2,$7E,$0B,$E1,$12,$7C,$2F,$F8,$C7,$B5,$C9,$FB,$22,$4F,$5E,$3F
	.dc.b	$07,$C3,$7F,$6A,$FC,$6D,$EE,$D3,$35,$FE,$13,$11,$A7,$FC,$86,$D7
	.dc.b	$ED,$1A,$8F,$ED,$F1,$98,$FB,$DF,$3A,$ED,$FE,$E6,$D8,$FF,$26,$B5
	.dc.b	$E6,$D7,$8F,$C7,$FF,$80,$3E,$F3,$FA,$1C,$7E,$23,$A7,$FA,$94,$63
	.dc.b	$AB,$FC,$11,$6B,$FC,$87,$C1,$FC,$FE,$89,$8F,$E1,$E2,$BF,$E4,$FE
	.dc.b	$0F,$16,$1F,$C2,$B3,$CB,$F8,$99,$75,$B9,$5C,$FF,$6F,$E2,$65,$7D
	.dc.b	$B8,$AF,$5D,$3F,$92,$98,$C3,$F8,$BB,$F6,$49,$1F,$CB,$1F,$F3,$65
	.dc.b	$BE,$1F,$F1,$1C,$D2,$BF,$11,$15,$F5,$7D,$17,$F3,$23,$18,$7F,$8A
	.dc.b	$FE,$79,$E1,$7F,$B5,$EB,$EC,$C6,$3D,$0F,$F3,$FB,$63,$33,$B7,$F7
	.dc.b	$1F,$1D,$FE,$53,$F9,$EF,$F4,$2B,$63,$F0,$6D,$77,$EA,$27,$AF,$CF
	.dc.b	$3E,$46,$92,$BC,$71,$97,$F8,$37,$1B,$97,$F6,$FD,$5F,$D3,$3E,$53
	.dc.b	$FC,$58,$FE,$F2,$33,$5F,$6E,$72,$FD,$88,$7C,$F6,$2F,$A7,$91,$67
	.dc.b	$38,$7F,$58,$8C,$6D,$FE,$53,$C3,$B3,$5E,$8D,$BF,$33,$6D,$97,$2C
	.dc.b	$3F,$B1,$B6,$FF,$21,$E2,$F8,$63,$C4,$D9,$FF,$42,$71,$3E,$71,$F2
	.dc.b	$FE,$7F,$1B,$DB,$39,$8C,$57,$F7,$03,$35,$C1,$C1,$D1,$C6,$3F,$AB
	.dc.b	$03,$E7,$7F,$C7,$33,$FE,$C2,$6C,$7F,$00,$6B,$FC,$ED,$AF,$F9,$15
	.dc.b	$FF,$69,$47,$F0,$B7,$8D,$FF,$9F,$5C,$CF,$84,$F0,$7C,$2F,$D5,$F9
	.dc.b	$DF,$F4,$05,$FF,$0E,$DD,$18,$AF,$C2,$B5,$FF,$2A,$46,$BE,$37,$EA
	.dc.b	$4F,$FA,$16,$BF,$D4,$8F,$1C,$7F,$79,$5D,$3C,$9F,$05,$F8,$71,$98
	.dc.b	$7C,$7F,$9D,$B9,$7F,$B5,$72,$F4,$BE,$06,$BF,$D3,$9A,$FF,$08,$6E
	.dc.b	$3E,$AF,$17,$DD,$B7,$BD,$B8,$79,$BC,$2F,$84,$7C,$3F,$36,$C7,$AB
	.dc.b	$C4,$F6,$91,$F8,$D8,$A3,$E0,$E7,$F8,$2F,$93,$F9,$8D,$AF,$87,$E1
	.dc.b	$7C,$F3,$E8,$78,$76,$3E,$3F,$CA,$FC,$A2,$3D,$9F,$9B,$7F,$8E,$F8
	.dc.b	$FF,$B0,$D9,$F9,$FF,$29,$F6,$FC,$AF,$B8,$D8,$FA,$7E,$77,$D6,$23
	.dc.b	$EA,$7E,$CF,$C9,$FE,$0F,$E4,$FF,$08,$D9,$FB,$7E,$27,$E1,$F2,$FC
	.dc.b	$43,$63,$EF,$F3,$7E,$08,$8F,$C6,$7F,$71,$EC,$BE,$47,$F8,$23,$FE
	.dc.b	$AF,$9F,$F7,$53,$96,$FC,$96,$BE,$43,$F6,$3F,$3B,$FE,$F5,$9D,$C6
	.dc.b	$3C,$BE,$7F,$23,$D8,$93,$F9,$BC,$11,$E2,$F6,$79,$FC,$37,$89,$B5
	.dc.b	$E2,$F9,$1E,$FF,$0F,$E6,$6C,$79,$BE,$47,$D6,$47,$B3,$EE,$F9,$0F
	.dc.b	$E0,$FF,$3B,$FE,$77,$9F,$6C,$63,$E2,$FE,$0F,$CC,$C9,$FF,$01,$78
	.dc.b	$DF,$E0,$67,$E1,$C7,$C6,$F9,$0F,$92,$48,$F6,$3F,$33,$E6,$75,$1F
	.dc.b	$29,$BC,$DF,$C3,$EF,$FC,$B3,$F2,$DF,$43,$8F,$E8,$B6,$FE,$8B,$87
	.dc.b	$E9,$FC,$95,$BF,$F7,$3E,$BF,$E6,$D9,$FE,$4E,$D8,$ED,$C9,$FE,$E0
	.dc.b	$9C,$FC,$4F,$DE,$F9,$2F,$86,$D8,$C9,$CB,$E1,$FF,$07,$CD,$F2,$1B
	.dc.b	$1F,$84,$F9,$1C,$34,$E0,$7E,$23,$9B,$F1,$71,$F8,$C6,$C7,$9B,$E0
	.dc.b	$FE,$99,$F9,$0F,$91,$B1,$F9,$07,$0F,$E4,$BE,$56,$EB,$F0,$3D,$DF
	.dc.b	$CB,$FC,$1F,$DC,$6D,$7F,$3A,$F5,$FF,$8E,$FC,$E7,$FA,$8E,$30,$7F
	.dc.b	$98,$9C,$5F,$D3,$7F,$08,$DF,$90,$FE,$D2,$BB,$37,$7F,$E2,$38,$C4
	.dc.b	$EB,$A1,$E3,$70,$BD,$E3,$FC,$D3,$C6,$FE,$D1,$87,$F8,$A9,$B7,$FE
	.dc.b	$03,$9F,$F2,$5C,$E5,$B4,$F9,$1E,$D7,$D5,$B3,$F8,$0F,$81,$FC,$E9
	.dc.b	$FD,$F7,$4F,$F5,$EC,$64,$F8,$38,$7F,$8E,$A3,$2F,$E2,$A7,$87,$C9
	.dc.b	$F2,$AF,$C5,$FC,$07,$EA,$0E,$9D,$56,$39,$E8,$C7,$FD,$67,$18,$6E
	.dc.b	$3F,$37,$2E,$33,$9F,$92,$F3,$1F,$3B,$E7,$5F,$0D,$89,$FF,$39,$47
	.dc.b	$F8,$72,$FF,$D5,$0D,$7F,$BB,$87,$0F,$F7,$A6,$E3,$E5,$EC,$ED,$C9
	.dc.b	$F1,$5B,$7F,$E0,$EF,$37,$FB,$23,$F8,$96,$E1,$F1,$FE,$4B,$FC,$1F
	.dc.b	$18,$C3,$E0,$47,$CF,$3E,$87,$C9,$DB,$E9,$F0,$1C,$0F,$A9,$F1,$D6
	.dc.b	$3E,$33,$87,$E4,$78,$17,$2F,$CA,$77,$7F,$8B,$78,$BF,$CF,$95,$FE
	.dc.b	$03,$6C,$FF,$91,$9A,$FF,$80,$97,$C6,$FE,$0E,$BF,$08,$DA,$F8,$1E
	.dc.b	$FF,$B0,$FC,$47,$15,$8F,$B4,$DC,$3F,$4C,$DE,$DF,$D6,$F1,$79,$49
	.dc.b	$5E,$57,$D1,$F1,$5F,$93,$7F,$28,$D8,$FC,$37,$57,$B4,$8F,$B5,$F8
	.dc.b	$1F,$13,$F9,$B7,$F3,$8D,$8F,$C7,$7C,$4F,$E7,$FA,$FF,$40,$D8,$FC
	.dc.b	$91,$82,$9C,$0F,$CC,$FC,$D7,$F8,$A7,$4F,$F7,$2B,$5F,$F4,$FB,$79
	.dc.b	$8D,$3F,$C5,$91,$FC,$ED,$FC,$21,$8D,$3F,$C5,$AF,$E2,$A3,$F2,$FF
	.dc.b	$3B,$FE,$4E,$F9,$D7,$BB,$F9,$A9,$FE,$04,$23,$2E,$67,$84,$EF,$C1
	.dc.b	$FF,$5A,$56,$AB,$A8,$F1,$F0,$F4,$6A,$BF,$F2,$FC,$6A,$BF,$D7,$CE
	.dc.b	$A3,$BC,$FE,$BA,$7B,$DF,$FB,$82,$73,$3F,$E6,$89,$CB,$E7,$3D,$53
	.dc.b	$E7,$CF,$71,$FE,$6E,$AB,$5C,$23,$D3,$7E,$3A,$F2,$DF,$79,$F6,$3C
	.dc.b	$DA,$77,$C3,$83,$E0,$31,$1B,$8F,$64,$64,$F4,$4F,$91,$F8,$86,$4F
	.dc.b	$A7,$1C,$0E,$A5,$9E,$87,$96,$BC,$27,$62,$BE,$98,$E2,$D3,$8C,$FC
	.dc.b	$79,$DB,$63,$DD,$1C,$DA,$77,$2C,$7C,$FF,$1B,$F0,$0D,$9F,$91,$18
	.dc.b	$8F,$C2,$B7,$E3,$DE,$73,$5F,$F5,$0C,$E6,$31,$48,$C4,$EA,$FF,$98
	.dc.b	$DF,$28,$C5,$7F,$AA,$A3,$0F,$D6,$0B,$CF,$FE,$73,$5F,$99,$8C,$4F
	.dc.b	$E0,$0E,$58,$78,$A7,$4D,$C7,$C3,$7C,$9B,$5C,$4E,$F3,$B7,$9E,$1F
	.dc.b	$98,$C3,$BC,$7E,$0D,$B9,$76,$9D,$57,$1A,$F4,$E1,$DD,$F7,$35,$EF
	.dc.b	$3E,$1B,$F8,$37,$ED,$AF,$EA,$ED,$AE,$B5,$1F,$14,$EB,$0F,$5C,$79
	.dc.b	$70,$FA,$77,$E5,$8D,$9E,$BA,$E0,$FB,$8E,$17,$E3,$8F,$36,$1F,$03
	.dc.b	$1E,$47,$3B,$F2,$5F,$43,$89,$1C,$93,$1F,$C0,$E3,$E4,$6E,$3D,$EA
	.dc.b	$1F,$81,$C3,$F0,$78,$77,$AF,$6E,$1C,$EB,$D0,$FF,$20,$DF,$F9,$7F
	.dc.b	$6F,$F7,$EC,$7F,$95,$FC,$AD,$DB,$F9,$BB,$70,$FE,$6F,$56,$FF,$9B
	.dc.b	$CF,$F8,$B1,$B3,$FE,$EB,$E1,$4E,$5B,$19,$F8,$AE,$50,$EA,$A6,$22
	.dc.b	$CF,$F4,$AF,$8A,$D9,$87,$FA,$27,$C5,$72,$F8,$99,$1F,$D4,$22,$BF
	.dc.b	$D0,$72,$FE,$9B,$3E,$32,$47,$9F,$E2,$7C,$7F,$11,$23,$FA,$CC,$57
	.dc.b	$96,$69,$F6,$12,$3E,$07,$C4,$FC,$7E,$0F,$F6,$07,$C4,$DF,$24,$FC
	.dc.b	$09,$C3,$2B,$FC,$4C,$E5,$FE,$D2,$FC,$6F,$89,$91,$83,$FB,$6B,$F3
	.dc.b	$48,$FD,$17,$E8,$72,$FC,$E2,$47,$F7,$4F,$89,$F2,$7C,$4C,$8F,$F1
	.dc.b	$B3,$A7,$FB,$87,$C4,$FD,$62,$3F,$DF,$23,$D9,$F1,$32,$3F,$C7,$CE
	.dc.b	$9F,$EF,$0F,$E2,$57,$CE,$3E,$33,$0D,$8F,$F0,$3C,$E2,$7F,$CE,$F1
	.dc.b	$C1,$B3,$B8,$FE,$F8,$64,$FC,$87,$B1,$F8,$C2,$3F,$E7,$43,$C3,$F1
	.dc.b	$BF,$90,$61,$FE,$16,$3F,$27,$E1,$FD,$76,$1F,$E5,$93,$67,$FB,$37
	.dc.b	$E6,$DF,$1B,$F7,$FE,$6F,$FD,$FD,$3F,$E8,$2E,$44,$62,$38,$4E,$F9
	.dc.b	$24,$6E,$75,$4F,$01,$2E,$A6,$3E,$08,$F6,$7D,$0F,$0A,$FF,$3B,$FC
	.dc.b	$E7,$F3,$0C,$3F,$D5,$4D,$8F,$F3,$DC,$EE,$BA,$CF,$F3,$67,$FC,$4C
	.dc.b	$D7,$9F,$8F,$F4,$A3,$11,$E4,$F9,$D7,$89,$F7,$5B,$1F,$3D,$AF,$F6
	.dc.b	$1C,$3C,$BF,$0B,$FC,$27,$37,$C8,$52,$57,$FA,$82,$3C,$DA,$7C,$AF
	.dc.b	$8F,$E1,$F2,$12,$BE,$44,$73,$F9,$09,$1C,$5E,$D1,$F5,$7E,$42,$4F
	.dc.b	$CF,$76,$FD,$B7,$C9,$F0,$B2,$3C,$51,$5F,$51,$FD,$EF,$8D,$F1,$DF
	.dc.b	$81,$35,$97,$D9,$9D,$47,$BF,$E3,$E4,$7D,$F8,$F8,$47,$E2,$7E,$3A
	.dc.b	$47,$C1,$8A,$F8,$2B,$F1,$73,$87,$B7,$37,$E1,$2B,$E6,$7A,$52,$FE
	.dc.b	$53,$F3,$FE,$3F,$11,$F9,$9C,$3E,$91,$8C,$39,$F0,$7E,$1E,$7F,$20
	.dc.b	$B8,$BE,$C1,$F6,$79,$EC,$FE,$CB,$3F,$CA,$CF,$F4,$3A,$E1,$39,$6B
	.dc.b	$FC,$CD,$7C,$77,$E0,$C9,$1F,$B7,$E8,$7C,$6F,$E1,$FE,$3B,$F5,$CC
	.dc.b	$9F,$C3,$47,$E2,$E3,$C0,$48,$F0,$3E,$3F,$8E,$FE,$69,$F3,$AF,$91
	.dc.b	$A7,$E4,$E3,$F2,$9E,$C6,$BE,$3E,$61,$F9,$38,$FC,$C7,$B7,$ED,$24
	.dc.b	$7B,$5F,$EC,$5D,$5D,$63,$F8,$16,$2F,$8B,$F0,$9F,$DA,$67,$53,$F8
	.dc.b	$CE,$14,$31,$D7,$C8,$4A,$FE,$07,$C9,$F8,$0A,$D4,$63,$C0,$FE,$0F
	.dc.b	$87,$2F,$8F,$F1,$92,$7E,$B5,$23,$CF,$F1,$FE,$3F,$90,$95,$CC,$FE
	.dc.b	$21,$49,$F4,$7C,$77,$A7,$59,$5B,$C7,$D3,$3E,$DF,$8F,$F8,$64,$9F
	.dc.b	$0B,$F1,$BE,$3F,$E1,$7C,$7C,$9E,$59,$27,$E3,$7C,$77,$D8,$DA,$56
	.dc.b	$A3,$E5,$7C,$77,$D0,$24,$7B,$9F,$A7,$F1,$BF,$3C,$91,$DB,$E3,$BE
	.dc.b	$97,$C7,$7F,$77,$F9,$D7,$C2,$D3,$35,$E4,$F8,$FF,$BA,$47,$E5,$BF
	.dc.b	$80,$F8,$EF,$B8,$48,$F5,$FC,$7F,$DF,$D7,$F0,$1F,$1F,$27,$DF,$1F
	.dc.b	$E2,$BF,$9A,$FE,$B9,$1F,$D0,$0F,$E8,$0E,$1F,$BA,$D7,$3E,$EF,$F8
	.dc.b	$C6,$1E,$DF,$32,$4F,$F5,$3D,$3F,$C3,$0F,$D7,$9F,$E6,$BD,$9F,$C2
	.dc.b	$9F,$F1,$AC,$7E,$4D,$B1,$9F,$5F,$CD,$F1,$BD,$C9,$5F,$C0,$6B,$8B
	.dc.b	$FE,$EF,$F9,$B7,$47,$F8,$05,$FB,$91,$FD,$CD,$AF,$F2,$C0,$F0,$75
	.dc.b	$CF,$FE,$02,$8D,$C7,$58,$C1,$F0,$7E,$37,$ED,$7C,$24,$9F,$74,$57
	.dc.b	$D9,$C9,$F3,$CC,$69,$F0,$7E,$3B,$E9,$7C,$74,$AF,$8D,$15,$F8,$9F
	.dc.b	$1D,$F5,$CC,$57,$C9,$7E,$91,$23,$E2,$3F,$57,$E3,$BE,$F7,$C6,$49
	.dc.b	$FA,$51,$5F,$A1,$F1,$BF,$83,$31,$7F,$D2,$E1,$F3,$6E,$BF,$F6,$44
	.dc.b	$7F,$88,$A3,$1E,$67,$F8,$92,$37,$3A,$7F,$18,$62,$35,$F2,$2F,$61
	.dc.b	$F3,$78,$6D,$7F,$8B,$A7,$53,$93,$B6,$0E,$E7,$C4,$61,$B7,$E5,$6B
	.dc.b	$E2,$9E,$13,$FD,$27,$F8,$03,$97,$7F,$E5,$7E,$47,$FB,$56,$DF,$D7
	.dc.b	$A7,$38,$7E,$56,$73,$5F,$C8,$25,$E1,$62,$BF,$C5,$11,$C2,$F9,$4E
	.dc.b	$7C,$B4,$E6,$FF,$5B,$F2,$3F,$3F,$0B,$1F,$9D,$8C,$47,$ED,$1B,$5C
	.dc.b	$A7,$A7,$72,$7B,$5F,$F8,$EE,$7F,$61,$1E,$1E,$9F,$F3,$14,$7F,$29
	.dc.b	$9D,$61,$FD,$0E,$72,$FF,$45,$8C,$E5,$F8,$3F,$9A,$FD,$FD,$FF,$97
	.dc.b	$7E,$81,$E3,$FC,$1C,$15,$A7,$F5,$D3,$AA,$FF,$32,$4F,$F8,$E9,$FC
	.dc.b	$43,$67,$FC,$DA,$D7,$F1,$2D,$7F,$A7,$E3,$BC,$3F,$85,$77,$7B,$7C
	.dc.b	$3F,$86,$DC,$3F,$A3,$56,$DE,$3E,$1E,$70,$FF,$0F,$78,$9F,$E6,$09
	.dc.b	$FF,$3E,$E2,$8C,$4F,$E6,$A3,$0F,$D4,$6C,$FF,$84,$ED,$E0,$FE,$0E
	.dc.b	$73,$5F,$36,$31,$7D,$27,$F9,$17,$C9,$FD,$A3,$EB,$78,$37,$1F,$F3
	.dc.b	$E6,$9F,$E5,$87,$CA,$DC,$7C,$8D,$7F,$DD,$8D,$7E,$3F,$83,$F8,$1F
	.dc.b	$0B,$F0,$4D,$AE,$BF,$13,$E7,$B3,$F8,$7F,$0F,$0E,$7C,$EF,$81,$3F
	.dc.b	$E0,$8A,$F1,$C6,$0C,$FC,$3C,$8F,$3B,$72,$F3,$FC,$37,$73,$F2,$9A
	.dc.b	$DA,$F7,$FC,$37,$62,$3F,$01,$FD,$0F,$13,$E4,$FF,$38,$7E,$77,$E4
	.dc.b	$AD,$7F,$4F,$F9,$3F,$D2,$9F,$A1,$F0,$6C,$7C,$CF,$84,$F5,$91,$F9
	.dc.b	$AF,$EB,$BF,$0F,$FF,$57,$C7,$F8,$3D,$BA,$FF,$8A,$57,$C8,$FE,$F5
	.dc.b	$CF,$3F,$F5,$DC,$FF,$6A,$67,$8E,$E1,$FE,$B7,$FC,$27,$FC,$88,$FE
	.dc.b	$41,$B8,$FE,$3D,$AF,$F7,$5B,$46,$23,$FC,$63,$83,$8B,$F8,$1E,$DE
	.dc.b	$36,$E1,$DF,$DD,$FE,$0F,$F2,$9F,$95,$C5,$18,$9F,$27,$BB,$F9,$D9
	.dc.b	$FF,$25,$47,$96,$30,$7C,$5E,$3F,$80,$7C,$1F,$8C,$B3,$ED,$F4,$FF
	.dc.b	$8A,$F1,$F7,$6C,$7C,$1F,$6F,$B1,$1F,$84,$FD,$1F,$8C,$F9,$87,$CD
	.dc.b	$F8,$CB,$3F,$E0,$BE,$1F,$E4,$27,$D1,$E6,$B1,$FE,$F1,$E3,$FE,$9C
	.dc.b	$FF,$55,$6C,$FE,$FB,$D2,$FE,$5A,$A3,$2F,$F8,$8F,$CC,$72,$BE,$2B
	.dc.b	$F8,$F5,$72,$3F,$D4,$71,$9C,$3C,$18,$7E,$0F,$E2,$24,$7E,$27,$E2
	.dc.b	$2E,$5F,$89,$7F,$10,$7E,$27,$D6,$B3,$FE,$5D,$FC,$27,$F0,$79,$7F
	.dc.b	$AC,$1B,$1F,$E7,$05,$F0,$FF,$D3,$5C,$E1,$FE,$F9,$73,$3C,$E7,$5F
	.dc.b	$11,$B7,$FA,$A7,$F0,$0F,$E6,$27,$FC,$F6,$DB,$E5,$F1,$3A,$F9,$6E
	.dc.b	$CC,$AF,$37,$65,$AF,$F4,$66,$5F,$8C,$9C,$9E,$BF,$8A,$F3,$57,$9F
	.dc.b	$B3,$C3,$E1,$DA,$FF,$3E,$67,$FA,$C7,$E1,$7C,$4E,$72,$EA,$DD,$FF
	.dc.b	$DA,$75,$FC,$8D,$BF,$29,$F0,$FE,$13,$DA,$FC,$5E,$DC,$63,$F3,$5B
	.dc.b	$8F,$FA,$8F,$0F,$E4,$33,$E2,$F1,$7F,$AE,$7C,$9F,$FA,$4E,$30,$7E
	.dc.b	$37,$B7,$F1,$39,$FF,$B5,$E3,$F1,$D1,$83,$3D,$BF,$88,$7F,$12,$D8
	.dc.b	$FB,$DD,$FF,$64,$FB,$5F,$13,$63,$F0,$4E,$1F,$C2,$7C,$55,$D3,$E7
	.dc.b	$FB,$5F,$6B,$CF,$FC,$13,$6B,$F1,$5F,$03,$86,$9F,$8D,$8F,$C0,$FC
	.dc.b	$A4,$8F,$C8,$7C,$05,$C3,$FC,$CD,$F3,$9F,$DE,$5F,$F1,$D8,$70,$FF
	.dc.b	$8F,$9A,$FF,$A2,$DA,$8E,$1F,$0D,$FE,$07,$8E,$9C,$1F,$E4,$F7,$FC
	.dc.b	$9D,$7F,$EC,$3A,$ED,$8F,$E7,$63,$F2,$FE,$75,$7F,$0E,$FF,$8B,$7F
	.dc.b	$44,$FF,$08,$61,$FA,$79,$C7,$C5,$EF,$0F,$F7,$05,$7F,$80,$D7,$94
	.dc.b	$E7,$73,$73,$F5,$FF,$4A,$F5,$F7,$71,$97,$FB,$47,$1F,$EF,$ED,$AF
	.dc.b	$D5,$FB,$1F,$E3,$6C,$0D,$24,$FC,$AC,$57,$FE,$C2,$8F,$D0,$CF,$AF
	.dc.b	$D7,$E7,$1D,$FD,$BF,$DB,$FA,$7E,$F6,$C6,$49,$AF,$BB,$DB,$D4,$7E
	.dc.b	$B3,$D4,$F2,$11,$F1,$68,$8F,$F2,$6E,$6F,$8A,$FE,$0B,$1F,$35,$B5
	.dc.b	$FB,$6F,$47,$DC,$47,$9B,$CB,$DC,$FE,$1A,$FD,$56,$C7,$C9,$F8,$4F
	.dc.b	$E2,$3F,$2B,$F6,$5B,$19,$26,$9F,$63,$E0,$B5,$1F,$C1,$BD,$8F,$9C
	.dc.b	$47,$D6,$FA,$BD,$FF,$E3,$AF,$E1,$5B,$1F,$8D,$AF,$CA,$7C,$E7,$E5
	.dc.b	$7E,$71,$5F,$E9,$8F,$71,$79,$8F,$D0,$D5,$BF,$F7,$45,$69,$5F,$27
	.dc.b	$FD,$43,$3F,$A1,$F6,$71,$1F,$A3,$AE,$11,$88,$F1,$87,$11,$AB,$F0
	.dc.b	$CF,$67,$CD,$87,$8E,$3F,$DF,$57,$5F,$F5,$2C,$55,$19,$AF,$46,$1E
	.dc.b	$19,$C9,$D4,$F6,$D7,$1A,$F5,$5F,$82,$7C,$CF,$93,$0F,$3D,$6A,$7E
	.dc.b	$09,$CC,$F8,$0F,$37,$07,$98,$A7,$8A,$F9,$B8,$BF,$44,$F1,$9C,$9E
	.dc.b	$98,$CC,$EC,$F2,$9E,$A2,$9F,$18,$A7,$86,$B9,$3E,$4C,$3E,$2E,$1F
	.dc.b	$42,$FB,$BE,$F3,$E1,$57,$03,$CE,$57,$26,$EB,$C5,$1D,$BD,$9F,$BD
	.dc.b	$1F,$19,$AF,$95,$AF,$E1,$1C,$FC,$63,$6F,$F0,$C4,$A7,$31,$86,$F7
	.dc.b	$FF,$31,$6C,$EE,$FF,$9A,$2B,$E9,$5F,$66,$31,$AE,$AB,$F3,$34,$8C
	.dc.b	$63,$CA,$38,$D7,$CE,$6D,$70,$AD,$CF,$F4,$39,$E4,$F7,$77,$5C,$F1
	.dc.b	$F6,$E5,$C6,$BF,$3D,$5C,$1F,$C6,$3E,$08,$F2,$CF,$C4,$C3,$85,$78
	.dc.b	$F0,$F0,$BE,$F9,$D6,$1D,$F1,$EB,$5F,$13,$0E,$27,$48,$F0,$D7,$92
	.dc.b	$BE,$05,$6D,$B7,$AA,$8F,$9A,$74,$FC,$16,$BE,$38,$DE,$1E,$82,$57
	.dc.b	$96,$7E,$65,$F8,$5E,$47,$A0,$EF,$87,$9D,$F0,$19,$3E,$3C,$71,$7C
	.dc.b	$07,$C5,$3E,$96,$1F,$1B,$1F,$A7,$1F,$5E,$78,$47,$D1,$AF,$8B,$7E
	.dc.b	$57,$F9,$07,$5F,$F5,$CF,$80,$FF,$4C,$6F,$FE,$4C,$F3,$6E,$5F,$DC
	.dc.b	$3A,$11,$8B,$E1,$3F,$E4,$28,$FF,$07,$BB,$6C,$7F,$77,$6C,$F2,$9F
	.dc.b	$ED,$9D,$0E,$1E,$8D,$8F,$EF,$11,$8F,$80,$CB,$E1,$A7,$9D,$4C,$D7
	.dc.b	$F8,$67,$8B,$F9,$F6,$4F,$5D,$ED,$71,$F7,$E5,$EF,$6F,$F1,$05,$EE
	.dc.b	$B3,$3F,$B0,$3E,$15,$7A,$88,$FF,$81,$6D,$EA,$79,$7E,$56,$4F,$53
	.dc.b	$FC,$33,$3A,$7F,$CD,$31,$F1,$08,$FF,$7E,$7E,$4C,$3D,$0F,$A0,$D9
	.dc.b	$E9,$8F,$B6,$73,$EA,$69,$F8,$DE,$FC,$AD,$DF,$17,$FD,$55,$A7,$C0
	.dc.b	$8F,$F6,$A4,$7F,$9F,$3A,$BE,$71,$E0,$F3,$2C,$7F,$84,$A3,$ED,$91
	.dc.b	$F8,$5E,$47,$9B,$C7,$91,$E0,$8A,$FF,$9E,$7B,$BE,$D1,$8E,$2F,$AF
	.dc.b	$8B,$C3,$F7,$F7,$4B,$F4,$C5,$7D,$27,$F9,$13,$E2,$5F,$2B,$F9,$56
	.dc.b	$C7,$C0,$AF,$CD,$BF,$E4,$8F,$92,$BE,$77,$F8,$0E,$39,$9F,$E4,$CE
	.dc.b	$0C,$CE,$9F,$CE,$C6,$4F,$CF,$CF,$FB,$51,$C1,$F8,$7E,$46,$1F,$EC
	.dc.b	$08,$FA,$ED,$7D,$31,$AF,$81,$FB,$BD,$FF,$7B,$A7,$EF,$F4,$4B,$FB
	.dc.b	$31,$5E,$38,$3E,$0B,$E0,$19,$E1,$F9,$7C,$5F,$88,$25,$7E,$0B,$9F
	.dc.b	$F1,$7C,$D2,$3F,$0A,$E9,$F8,$7E,$1B,$C2,$F9,$2B,$C1,$A6,$22,$CF
	.dc.b	$F0,$29,$D3,$67,$E4,$C6,$23,$E3,$66,$F8,$2F,$B0,$48,$FF,$46,$71
	.dc.b	$7B,$5F,$CF,$FF,$00,$FC,$9D,$7F,$7A,$E6,$3E,$5B,$63,$3C,$5F,$83
	.dc.b	$F8,$29,$5F,$49,$FE,$35,$5B,$E5,$FE,$02,$A3,$3D,$52,$BC,$31,$5F
	.dc.b	$A3,$1F,$8F,$31,$87,$93,$E5,$7A,$92,$7D,$3F,$07,$E0,$F8,$29,$1E
	.dc.b	$C7,$4F,$DC,$8F,$CD,$11,$D1,$F0,$BE,$57,$D2,$47,$E2,$FC,$1F,$B3
	.dc.b	$E0,$E4,$7C,$87,$4F,$C6,$F8,$1F,$82,$47,$E6,$FC,$1F,$C4,$F8,$39
	.dc.b	$1F,$41,$D3,$F3,$BE,$03,$F5,$7F,$2F,$FC,$B2,$3F,$D8,$4D,$74,$63
	.dc.b	$CC,$FD,$4C,$66,$3F,$96,$35,$FF,$0E,$B5,$FD,$E6,$DF,$BB,$23,$F8
	.dc.b	$2F,$84,$FB,$3F,$07,$23,$F0,$AF,$DC,$F1,$7F,$AC,$F1,$3F,$05,$7C
	.dc.b	$57,$F5,$68,$C3,$63,$37,$FE,$E8,$9C,$B6,$3F,$31,$1F,$D3,$3B,$1C
	.dc.b	$BF,$C3,$CE,$F3,$AE,$9F,$E9,$7F,$C0,$2F,$84,$FE,$D0,$DC,$7F,$9C
	.dc.b	$B5,$FE,$06,$BE,$0F,$D3,$F8,$63,$E2,$7D,$0D,$AF,$33,$5F,$EA,$DA
	.dc.b	$23,$11,$FE,$48,$8C,$7A,$9F,$C9,$7A,$A4,$FF,$55,$9C,$BE,$4F,$8E
	.dc.b	$FE,$55,$D7,$23,$06,$F0,$FF,$78,$BE,$2F,$0B,$CA,$47,$FD,$D1,$C2
	.dc.b	$ED,$FE,$68,$48,$F8,$3F,$05,$F1,$3E,$5E,$47,$C4,$7E,$3F,$C1,$FC
	.dc.b	$2F,$80,$F8,$24,$9F,$82,$FF,$83,$B0,$78,$1F,$D0,$49,$3F,$3B,$E0
	.dc.b	$BE,$8F,$93,$23,$E8,$BF,$53,$E3,$FE,$7F,$9D,$F3,$89,$3F,$39,$FA
	.dc.b	$3E,$7F,$D4,$24,$7D,$AF,$37,$F1,$FE,$87,$F3,$D8,$FA,$0D,$7F,$C2
	.dc.b	$8D,$5C,$BF,$CE,$7C,$DF,$CA,$4F,$F1,$B9,$E2,$E0,$8F,$E5,$5F,$E5
	.dc.b	$0F,$D9,$6B,$FD,$92,$7F,$A2,$79,$DF,$E3,$7F,$E0,$1F,$C3,$47,$F8
	.dc.b	$91,$FF,$09,$4F,$F5,$27,$FC,$09,$19,$F7,$38,$7B,$BE,$CF,$3B,$CA
	.dc.b	$4B,$F1,$3F,$47,$CD,$FE,$E0,$48,$FF,$0A,$E0,$F8,$DF,$17,$57,$C6
	.dc.b	$F2,$E5,$6A,$2B,$F9,$AD,$53,$F1,$09,$1F,$E3,$31,$EC,$30,$7B,$BE
	.dc.b	$F2,$4F,$8B,$DD,$AF,$F6,$8F,$4B,$FC,$0C,$63,$83,$FC,$05,$F3,$2F
	.dc.b	$0B,$FB,$FB,$6B,$FC,$BA,$D7,$C4,$D7,$FC,$8C,$FE,$7D,$FF,$20,$0F
	.dc.b	$A8,$78,$DF,$E1,$26,$C6,$A7,$83,$F4,$FD,$FF,$B1,$EF,$48,$F9,$71
	.dc.b	$5F,$F0,$0F,$C7,$7F,$87,$CC,$65,$FC,$55,$FD,$21,$23,$F6,$6F,$F9
	.dc.b	$D3,$6F,$F3,$4C,$7D,$6F,$7F,$F0,$C7,$E1,$FE,$0A,$C7,$F8,$63,$D9
	.dc.b	$FF,$16,$18,$7F,$C4,$FF,$32,$8F,$F6,$E4,$7D,$58,$C7,$7F,$FA,$11
	.dc.b	$B1,$99,$DB,$FE,$6D,$F8,$8F,$F2,$07,$CD,$7F,$A2,$9B,$1F,$7D,$AE
	.dc.b	$F5,$4F,$5F,$6B,$FD,$11,$3F,$E7,$7C,$11,$88,$FF,$74,$C6,$3E,$1F
	.dc.b	$81,$FC,$5B,$DB,$7A,$FF,$E3,$6B,$B7,$FE,$2E,$9B,$C3,$FD,$06,$6D
	.dc.b	$7F,$A2,$3C,$B7,$B7,$8C,$3F,$CA,$98,$5E,$CF,$F2,$1C,$DA,$4D,$BF
	.dc.b	$F2,$95,$F3,$6B,$FE,$5D,$6D,$FF,$93,$1B,$F1,$EF,$9B,$E5,$1F,$E8
	.dc.b	$48,$C3,$FE,$50,$3C,$4F,$F5,$37,$51,$8F,$6B,$2F,$A6,$DF,$31,$FD
	.dc.b	$4E,$BF,$D7,$AD,$8F,$D9,$35,$FF,$53,$35,$11,$7C,$C7,$57,$96,$DF
	.dc.b	$E7,$17,$35,$FA,$87,$2F,$FA,$D3,$1B,$7F,$C0,$1B,$91,$E3,$1F,$E6
	.dc.b	$C7,$26,$2F,$FD,$86,$E3,$0F,$87,$ED,$FC,$4F,$93,$F9,$64,$C7,$F8
	.dc.b	$76,$3D,$63,$E7,$DE,$E7,$E7,$37,$1F,$07,$B5,$F3,$8F,$C1,$65,$FD
	.dc.b	$EF,$CB,$C5,$FD,$5F,$93,$FB,$24,$7B,$E0,$8F,$BB,$EC,$FD,$8F,$93
	.dc.b	$FB,$2D,$AF,$BD,$EC,$FD,$C3,$EE,$FB,$56,$3F,$05,$F3,$7F,$86,$23
	.dc.b	$D9,$FC,$4F,$C9,$F9,$B8,$5F,$1F,$F8,$68,$C3,$F8,$8F,$6A,$5F,$E3
	.dc.b	$7D,$BB,$C5,$F3,$FD,$BF,$C9,$92,$7D,$6F,$E5,$BE,$2F,$F2,$3F,$0D
	.dc.b	$F9,$26,$CF,$E5,$FC,$0F,$CA,$9F,$96,$F7,$2C,$7E,$6F,$0F,$CF,$91
	.dc.b	$F8,$EF,$E9,$3E,$5F,$FC,$4D,$1F,$EE,$47,$FD,$FE,$DF,$84,$FD,$94
	.dc.b	$62,$7F,$D6,$2F,$F3,$49,$E1,$E8,$EB,$A3,$6F,$0F,$73,$B1,$27,$F0
	.dc.b	$B5,$FE,$F5,$6F,$07,$63,$BF,$B5,$63,$C9,$ED,$7E,$13,$E6,$3F,$D1
	.dc.b	$D1,$83,$FC,$64,$E3,$3F,$2F,$B1,$ED,$24,$F8,$1F,$81,$F2,$DE,$CF
	.dc.b	$6B,$DA,$D9,$F8,$5E,$77,$7F,$67,$E0,$B6,$3E,$37,$AF,$F2,$48,$FB
	.dc.b	$1F,$99,$EA,$FC,$8F,$83,$F9,$2D,$9F,$9D,$EB,$7C,$C3,$F8,$47,$8D
	.dc.b	$FE,$F9,$8C,$9F,$4F,$E4,$3E,$B1,$1F,$92,$FD,$9F,$91,$FA,$AF,$D6
	.dc.b	$6C,$7D,$BF,$91,$FB,$3F,$0B,$F6,$9B,$1F,$7F,$DA,$C3,$4E,$07,$F1
	.dc.b	$2C,$3F,$9D,$F9,$CF,$98,$7C,$37,$F3,$55,$F2,$76,$30,$E7,$E1,$31
	.dc.b	$87,$33,$F8,$ED,$F3,$8C,$F3,$FF,$BF,$30,$FF,$6E,$56,$E3,$FC,$2F
	.dc.b	$18,$75,$F2,$5E,$1F,$88,$F1,$3F,$CF,$A7,$F0,$53,$FC,$A8,$E3,$19
	.dc.b	$35,$F2,$5C,$2F,$B6,$3E,$D8,$C5,$78,$FE,$27,$DC,$7B,$FE,$42,$DF
	.dc.b	$0C,$7F,$D3,$6D,$8F,$89,$F1,$3B,$8D,$7C,$77,$C7,$3E,$47,$5D,$8E
	.dc.b	$5F,$1B,$86,$9D,$63,$FA,$67,$82,$D2,$9F,$88,$EA,$B5,$FE,$28,$E4
	.dc.b	$75,$7F,$62,$DF,$FC,$CE,$D8,$CC,$FF,$AE,$E3,$58,$7A,$AB,$FC,$DB
	.dc.b	$3F,$EA,$19,$F0,$EB,$F9,$67,$F0,$1F,$19,$F8,$1F,$8D,$B5,$F0,$DF
	.dc.b	$C2,$7C,$6E,$30,$FB,$8D,$CB,$F1,$7C,$2F,$8B,$FC,$51,$F8,$BF,$86
	.dc.b	$B8,$FC,$D7,$0F,$CF,$F8,$AB,$8F,$E1,$FE,$17,$F2,$47,$F7,$1F,$98
	.dc.b	$E9,$5F,$56,$31,$5F,$0E,$7F,$C6,$13,$FE,$63,$A4,$61,$FF,$33,$35
	.dc.b	$FC,$57,$47,$E8,$7E,$23,$9D,$7E,$05,$B3,$F9,$98,$FD,$21,$FE,$19
	.dc.b	$F9,$6F,$F0,$44,$66,$3F,$1F,$18,$D7,$F5,$AD,$8F,$D2,$BF,$E1,$29
	.dc.b	$FA,$67,$F9,$E1,$D3,$7D,$BE,$C7,$7F,$84,$B5,$D9,$C3,$E0,$F8,$5B
	.dc.b	$5D,$32,$61,$FE,$03,$AF,$EF,$2D,$9F,$D5,$1E,$1F,$84,$FF,$15,$C6
	.dc.b	$63,$E2,$AF,$85,$F8,$13,$F0,$5B,$5F,$03,$E1,$7E,$21,$F1,$7E,$16
	.dc.b	$C7,$C4,$6E,$1E,$5C,$11,$8D,$7E,$3F,$C2,$F8,$89,$7E,$87,$CB,$F0
	.dc.b	$9F,$45,$FA,$4D,$8F,$A3,$F0,$BE,$A2,$3F,$01,$F6,$FC,$2F,$D9,$7E
	.dc.b	$D3,$63,$EC,$FC,$2F,$DD,$3E,$F7,$C2,$58,$FB,$AD,$D3,$D4,$DC,$BE
	.dc.b	$FF,$C2,$FC,$A2,$57,$D0,$7E,$6F,$C2,$FE,$25,$FC,$53,$63,$FD,$4B
	.dc.b	$7F,$E4,$6F,$98,$FF,$23,$FC,$C2,$FE,$BF,$C2,$FF,$9C,$A3,$15,$F6
	.dc.b	$AB,$75,$AB,$B3,$91,$E9,$F0,$7F,$CA,$78,$7F,$9F,$23,$11,$E1,$37
	.dc.b	$5C,$DA,$73,$C3,$75,$E3,$9F,$F6,$96,$1E,$0A,$EB,$85,$7A,$5F,$08
	.dc.b	$C4,$27,$57,$CE,$3D,$18,$6D,$F0,$9F,$D0,$67,$FC,$D7,$19,$3D,$31
	.dc.b	$CC,$5F,$13,$33,$DA,$FC,$2D,$9D,$E1,$EA,$69,$FD,$3F,$0F,$23,$E1
	.dc.b	$32,$7B,$8A,$7B,$A3,$E4,$5F,$82,$3D,$22,$7E,$84,$F7,$C3,$D4,$F8
	.dc.b	$6B,$D0,$D3,$99,$B3,$EA,$47,$53,$8D,$7C,$88,$C9,$B7,$D5,$87,$C8
	.dc.b	$8F,$77,$A7,$F8,$49,$FC,$33,$5F,$73,$5F,$C2,$39,$F8,$E7,$B2,$D7
	.dc.b	$FD,$3B,$39,$8C,$52,$31,$3A,$C7,$FB,$D5,$F2,$8C,$57,$FA,$9A,$30
	.dc.b	$FE,$01,$7B,$3F,$E0,$1A,$FC,$D4,$62,$7D,$87,$1C,$3C,$33,$A6,$E3
	.dc.b	$E0,$BE,$4D,$AE,$22,$FC,$37,$C7,$0F,$CE,$47,$79,$FC,$0B,$73,$D4
	.dc.b	$63,$1E,$EF,$AF,$1E,$ED,$3B,$D7,$6C,$7E,$1D,$7C,$8B,$F6,$DF,$F8
	.dc.b	$7D,$B7,$E2,$9D,$57,$0C,$3C,$B7,$B8,$F6,$57,$8E,$78,$1A,$61,$92
	.dc.b	$63,$D9,$FA,$15,$E8,$7E,$25,$FB,$1B,$5C,$B0,$E0,$7C,$22,$61,$E2
	.dc.b	$72,$F9,$E7,$BB,$E6,$C3,$C0,$7C,$A3,$8C,$7C,$73,$E7,$E1,$ED,$7E
	.dc.b	$F5,$FF,$1E,$EA,$FE,$BF,$F1,$1F,$E4,$58,$FE,$E3,$F1,$17,$6F,$EE
	.dc.b	$2D,$BF,$EE,$15,$6F,$FB,$3C,$FF,$82,$5B,$3F,$DA,$23,$1D,$26,$FC
	.dc.b	$47,$18,$FD,$E4,$FF,$80,$B8,$71,$39,$64,$64,$FF,$03,$C7,$17,$C5
	.dc.b	$E7,$BE,$5A,$99,$8F,$F2,$DC,$65,$FF,$79,$B5,$FE,$F7,$8F,$F9,$2E
	.dc.b	$7D,$53,$99,$FF,$0F,$AF,$89,$EF,$F1,$32,$7A,$C5,$79,$CA,$79,$92
	.dc.b	$3C,$3F,$13,$E8,$F8,$99,$1E,$68,$AF,$F8,$4E,$93,$8E,$0F,$F4,$8C
	.dc.b	$2B,$E0,$FC,$3F,$C2,$F8,$89,$87,$C0,$8A,$F9,$23,$DA,$47,$E1,$FC
	.dc.b	$3F,$CA,$23,$F2,$3D,$6F,$9B,$F0,$F2,$3E,$5B,$A7,$CD,$F0,$DF,$48
	.dc.b	$8F,$D0,$F8,$7F,$AB,$D9,$23,$E9,$BA,$7D,$DD,$3F,$8D,$58,$3C,$EC
	.dc.b	$36,$3E,$DC,$E2,$7F,$CB,$11,$C1,$B3,$B8,$FF,$1B,$99,$3F,$03,$ED
	.dc.b	$7E,$20,$8F,$FA,$B4,$F0,$F5,$FE,$33,$E1,$E4,$7E,$29,$FC,$7F,$B5
	.dc.b	$F9,$0F,$87,$93,$F8,$E7,$6F,$D9,$F8,$7F,$CA,$F8,$92,$3F,$27,$15
	.dc.b	$FB,$DE,$0F,$E6,$8C,$73,$B2,$FF,$5C,$7D,$07,$E5,$E3,$14,$3F,$C3
	.dc.b	$DE,$0F,$C9,$F0,$FF,$5D,$1E,$BA,$E1,$7F,$EC,$19,$CC,$FC,$46,$BF
	.dc.b	$E5,$C6,$BB,$AF,$97,$F0,$DF,$48,$91,$F1,$E5,$E0,$FF,$03,$F9,$8F
	.dc.b	$F7,$0C,$FF,$B2,$9F,$F7,$2B,$7E,$27,$5F,$13,$BC,$24,$ED,$E5,$3C
	.dc.b	$FE,$22,$4F,$37,$B7,$C4,$F7,$F0,$64,$F7,$76,$FF,$9E,$F1,$7C,$4F
	.dc.b	$97,$E2,$64,$F9,$62,$BF,$C0,$A5,$3F,$C6,$DB,$1F,$8A,$F2,$11,$88
	.dc.b	$E1,$F0,$DF,$04,$8F,$C1,$AF,$89,$F1,$3F,$17,$E2,$64,$FC,$58,$AF
	.dc.b	$9F,$E2,$7E,$51,$8E,$6F,$BF,$97,$83,$E2,$3E,$77,$C4,$4B,$F9,$D1
	.dc.b	$5F,$81,$F1,$3F,$4C,$C5,$F2,$D7,$FD,$9D,$3F,$57,$E1,$7E,$C1,$F6
	.dc.b	$7E,$22,$C7,$BE,$53,$F0,$C9,$19,$F8,$8F,$C0,$11,$FD,$57,$CC,$7E
	.dc.b	$41,$69,$F2,$3C,$6F,$C3,$7C,$4C,$9F,$C3,$47,$E2,$7B,$BE,$81,$23
	.dc.b	$C4,$FF,$4D,$F9,$77,$C7,$FF,$53,$D9,$F3,$1F,$92,$BF,$D2,$47,$E9
	.dc.b	$8C,$F7,$FF,$57,$BE,$33,$AC,$3F,$55,$9A,$5F,$17,$BF,$91,$97,$B3
	.dc.b	$D7,$3A,$71,$5C,$7E,$2F,$CC,$48,$FC,$44,$A3,$D5,$F1,$9E,$9F,$98
	.dc.b	$FE,$75,$5D,$74,$46,$23,$FD,$A9,$6A,$F5,$7C,$64,$9F,$0C,$FF,$B3
	.dc.b	$DB,$F1,$BE,$FF,$8D,$F9,$04,$8F,$1B,$F2,$FE,$33,$E2,$12,$39,$FC
	.dc.b	$6F,$C7,$F8,$CF,$91,$F1,$B2,$7B,$C7,$CB,$F8,$DF,$A8,$48,$F7,$BF
	.dc.b	$5F,$E3,$3E,$81,$23,$CF,$F1,$BF,$4F,$E3,$7E,$A7,$A1,$27,$D7,$1F
	.dc.b	$5F,$E2,$BF,$00,$48,$F9,$6F,$E0,$FA,$7E,$E1,$23,$E1,$FA,$5F,$7F
	.dc.b	$E2,$7F,$0D,$E9,$7F,$12,$9F,$F3,$87,$67,$E1,$A3,$1D,$1F,$AE,$E5
	.dc.b	$FF,$09,$D7,$D0,$3F,$20,$43,$ED,$9F,$E1,$FF,$96,$FF,$02,$3F,$82
	.dc.b	$6B,$D8,$FD,$0B,$89,$FD,$DB,$FA,$5E,$B3,$F7,$E3,$3F,$11,$FE,$2D
	.dc.b	$8F,$F0,$55,$75,$F8,$7C,$7C,$47,$84,$91,$FC,$56,$7F,$C0,$ED,$F8
	.dc.b	$8E,$44,$7F,$1D,$9F,$F2,$47,$D1,$F1,$1D,$FE,$0A,$46,$0F,$63,$EF
	.dc.b	$23,$DD,$F8,$5F,$11,$E7,$F8,$89,$3F,$0E,$2B,$F0,$F6,$F7,$CE,$36
	.dc.b	$F8,$75,$FE,$01,$F9,$77,$A9,$7D,$49,$8F,$F7,$59,$CB,$63,$FC,$03
	.dc.b	$06,$92,$3E,$27,$C4,$7C,$BF,$89,$91,$F5,$E2,$BE,$5E,$0F,$F5,$AB
	.dc.b	$F6,$FE,$27,$39,$FD,$87,$F0,$04,$8F,$A0,$FE,$0F,$E2,$7E,$C7,$C3
	.dc.b	$C9,$FC,$24,$57,$B7,$C4,$FD,$D3,$17,$FD,$F2,$1F,$2D,$FC,$C3,$E2
	.dc.b	$3F,$C3,$31,$8F,$5F,$FD,$88,$FF,$87,$27,$4D,$8C,$BF,$E0,$53,$FD
	.dc.b	$89,$AB,$E5,$F9,$FC,$82,$31,$3F,$E2,$68,$C4,$EA,$73,$3C,$CF,$2F
	.dc.b	$B3,$FA,$67,$B6,$08,$C4,$7F,$85,$63,$1E,$07,$A9,$B1,$FE,$34,$FC
	.dc.b	$19,$F0,$5F,$CD,$FC,$2F,$47,$80,$F8,$0C,$5F,$F8,$7E,$F7,$5F,$E3
	.dc.b	$7A,$D3,$FE,$39,$8F,$C4,$BF,$21,$B3,$DF,$A7,$E7,$3F,$16,$32,$FB
	.dc.b	$7E,$71,$EF,$EA,$3F,$C5,$D1,$8B,$FA,$6D,$8F,$F2,$6B,$5F,$8D,$F3
	.dc.b	$0F,$67,$D3,$09,$E1,$CB,$F7,$EB,$57,$FE,$51,$9C,$BF,$66,$3F,$A9
	.dc.b	$BF,$82,$6D,$7F,$97,$1A,$F8,$3D,$AF,$47,$82,$FE,$00,$7C,$3F,$CA
	.dc.b	$8F,$E1,$FF,$21,$67,$38,$FF,$19,$8C,$54,$8F,$EB,$B3,$FE,$24,$6C
	.dc.b	$71,$E2,$E9,$DE,$F9,$0C,$57,$F2,$36,$E7,$FC,$07,$0F,$B2,$DF,$0B
	.dc.b	$3B,$FF,$24,$F9,$0F,$4E,$26,$93,$1F,$F1,$97,$F0,$0B,$D9,$F0,$CF
	.dc.b	$B6,$31,$3F,$CC,$E3,$CB,$F2,$32,$3F,$BB,$CB,$DB,$C6,$7F,$C5,$7E
	.dc.b	$31,$E1,$FC,$A6,$DF,$F9,$05,$B3,$FE,$C8,$6B,$FA,$5F,$31,$F0,$DF
	.dc.b	$89,$8F,$F4,$DB,$63,$35,$FB,$C6,$CF,$FB,$4B,$5E,$BB,$7F,$46,$C7
	.dc.b	$D9,$EE,$FC,$E7,$ED,$7A,$2F,$90,$FB,$ED,$CB,$FC,$AC,$DC,$7F,$DB
	.dc.b	$AD,$7F,$82,$74,$37,$7B,$BF,$24,$FC,$26,$D7,$0F,$F0,$EE,$0F,$90
	.dc.b	$F0,$9F,$3F,$DD,$B3,$FE,$22,$FD,$C9,$EB,$7E,$49,$B1,$FE,$69,$6E
	.dc.b	$9F,$EF,$06,$BE,$4C,$DE,$FF,$DB,$3F,$31,$F0,$7F,$A0,$E2,$FF,$31
	.dc.b	$CE,$B0,$F2,$E3,$F1,$0D,$FC,$35,$8F,$4F,$C8,$7C,$62,$3F,$19,$F9
	.dc.b	$3F,$23,$F3,$0E,$F8,$D9,$FC,$3F,$C1,$7D,$0F,$0B,$F1,$CD,$8F,$81
	.dc.b	$D5,$EF,$23,$EF,$7F,$67,$9B,$E0,$BF,$2C,$7B,$39,$6D,$7E,$53,$E0
	.dc.b	$7E,$ED,$FE,$6D,$B1,$F4,$BE,$47,$EF,$91,$FB,$EF,$6F,$47,$F4,$47
	.dc.b	$C8,$C2,$CF,$E8,$38,$FF,$13,$E3,$7E,$9D,$B1,$F6,$7E,$47,$EC,$11
	.dc.b	$FB,$0F,$F0,$CF,$C4,$1F,$2B,$FC,$2D,$B3,$FE,$62,$FC,$42,$DF,$F7
	.dc.b	$F9,$BC,$FF,$E9,$AE,$CF,$E6,$6D,$F4,$B9,$7C,$B7,$E7,$1F,$E2,$D8
	.dc.b	$3E,$1B,$F8,$DB,$72,$FF,$59,$37,$2F,$EA,$9E,$0A,$31,$1F,$9E,$58
	.dc.b	$74,$BE,$AD,$BF,$F3,$CE,$7C,$F2,$FF,$57,$B6,$3F,$D0,$B6,$CF,$E0
	.dc.b	$61,$E6,$64,$FC,$1F,$1A,$E5,$F0,$7D,$5F,$DB,$F9,$9F,$1D,$B5,$F1
	.dc.b	$FE,$0F,$0D,$3C,$71,$FB,$EE,$19,$1F,$37,$E0,$AE,$1F,$E2,$BD,$11
	.dc.b	$F2,$0F,$93,$ED,$DA,$FA,$8E,$1F,$AD,$F0,$56,$FE,$B7,$C3,$7D,$B3
	.dc.b	$E8,$FC,$0D,$AF,$B8,$E1,$FB,$DF,$07,$6B,$EF,$70,$7E,$0C,$FB,$3E
	.dc.b	$AD,$AF,$C2,$38,$7F,$0D,$E9,$DA,$FF,$33,$0E,$2F,$EC,$28,$C7,$B7
	.dc.b	$FD,$AA,$3F,$DA,$75,$FE,$B3,$9C,$B5,$78,$7F,$CC,$5B,$1F,$E2,$A6
	.dc.b	$9F,$E2,$C8,$D7,$6F,$F7,$68,$CF,$47,$F8,$1A,$31,$3F,$DA,$7E,$1B
	.dc.b	$FC,$1F,$5F,$E5,$BF,$32,$DF,$FB,$DA,$6E,$1F,$EC,$C9,$B5,$FE,$E2
	.dc.b	$9F,$4B,$FE,$63,$73,$D8,$73,$7C,$FD,$EF,$06,$10,$FF,$67,$FE,$CF
	.dc.b	$FE,$E2,$7F,$DA,$34,$7F,$BF,$46,$70,$FF,$38,$36,$F5,$1F,$EF,$48
	.dc.b	$FF,$08,$BA,$F9,$1F,$F0,$36,$5E,$2B,$F5,$7C,$2F,$D1,$F8,$3C,$BF
	.dc.b	$1F,$E2,$BE,$71,$27,$E7,$3F,$47,$C6,$D4,$7D,$56,$E3,$F8,$03,$EF
	.dc.b	$7C,$D7,$DF,$F9,$A7,$C6,$FF,$93,$EB,$CF,$18,$AF,$F2,$95,$6E,$B5
	.dc.b	$76,$72,$2B,$83,$F8,$BB,$E7,$5F,$E5,$B8,$5F,$17,$57,$5F,$29,$CE
	.dc.b	$7B,$B4,$EF,$7D,$95,$F6,$73,$3C,$27,$51,$E0,$39,$1E,$02,$9E,$0F
	.dc.b	$7B,$A5,$7D,$87,$8E,$1F,$A5,$73,$85,$C3,$D9,$3E,$9A,$F1,$5F,$FA
	.dc.b	$0E,$31,$5E,$D8,$F1,$61,$C9,$F0,$98,$BF,$41,$D1,$F2,$DF,$47,$E2
	.dc.b	$CF,$13,$CA,$B0,$E8,$60,$B5,$CE,$78,$34,$EF,$3F,$6A,$76,$D8,$FA
	.dc.b	$31,$DA,$78,$CF,$CE,$3F,$D4,$15,$E5,$7D,$46,$2F,$E0,$61,$F0,$A3
	.dc.b	$DE,$DA,$E9,$87,$8A,$7E,$41,$EB,$7C,$27,$C3,$C3,$CB,$1A,$9F,$1B
	.dc.b	$67,$93,$F1,$9A,$F9,$0F,$C0,$47,$C5,$8E,$B1,$EF,$3D,$07,$84,$F7
	.dc.b	$1C,$1F,$90,$EA,$3D,$2D,$8C,$BB,$3A,$1F,$8F,$9F,$03,$D9,$AE,$63
	.dc.b	$DC,$79,$1F,$61,$D4,$F1,$9F,$A0,$BF,$9A,$FE,$19,$B5,$F1,$23,$F3
	.dc.b	$93,$CD,$A7,$85,$F3,$BC,$CF,$AA,$73,$2D,$7D,$07,$89,$E7,$3D,$A7
	.dc.b	$17,$D6,$D8,$FC,$D7,$B7,$F9,$58,$FE,$0E,$D7,$EF,$B5,$FE,$02,$E7
	.dc.b	$E2,$5D,$EE,$2F,$EA,$D3,$98,$C5,$7F,$A2,$B0,$FF,$79,$5F,$0B,$DC
	.dc.b	$62,$B7,$7C,$67,$F7,$57,$FD,$86,$DF,$15,$FE,$FB,$C3,$F8,$DC,$62
	.dc.b	$76,$1E,$CF,$E1,$CC,$61,$E4,$9E,$35,$CF,$1F,$2B,$6B,$C1,$5B,$9F
	.dc.b	$36,$1F,$C9,$70,$F0,$57,$F1,$86,$E5,$E4,$BF,$82,$DA,$F2,$CF,$0C
	.dc.b	$7C,$37,$8B,$EC,$F8,$2B,$B6,$3F,$E0,$18,$EC,$DC,$3A,$47,$86,$BB
	.dc.b	$E3,$EA,$8E,$27,$9E,$32,$F5,$3E,$26,$3E,$39,$E1,$18,$9F,$1B,$F2
	.dc.b	$30,$F4,$E1,$F1,$63,$E7,$36,$FE,$7E,$3E,$4C,$7D,$B5,$E8,$3C,$57
	.dc.b	$1C,$12,$38,$63,$F3,$A3,$CB,$87,$53,$B6,$3E,$4C,$7E,$ED,$F6,$25
	.dc.b	$F1,$79,$DF,$E2,$6F,$E4,$4F,$BA,$73,$3E,$D3,$B9,$EE,$3E,$A6,$11
	.dc.b	$F3,$4F,$13,$B1,$1E,$C7,$D0,$3E,$A4,$7B,$E7,$CB,$1A,$32,$4C,$7E
	.dc.b	$39,$2B,$E1,$3E,$6C,$3F,$1B,$1E,$19,$FB,$55,$F8,$2A,$E1,$1D,$5C
	.dc.b	$4F,$96,$3E,$B9,$E5,$9E,$98,$FE,$12,$32,$7B,$8E,$B3,$C6,$FC,$31
	.dc.b	$E9,$AE,$38,$7C,$78,$F9,$27,$E2,$5F,$91,$1F,$02,$3A,$BC,$1E,$C6
	.dc.b	$2B,$B5,$F8,$EF,$BC,$E4,$F1,$90,$C1,$1F,$C1,$CF,$E6,$AF,$FC,$7B
	.dc.b	$3F,$C0,$FB,$FF,$55,$18,$A4,$63,$4E,$07,$F1,$67,$FC,$CD,$1A,$AF
	.dc.b	$F1,$31,$CC,$AE,$5F,$F7,$D7,$0F,$35,$CE,$E3,$4E,$63,$0E,$E7,$FD
	.dc.b	$01,$E8,$67,$99,$E8,$D2,$4F,$F9,$83,$DE,$D7,$C7,$48,$FF,$31,$BA
	.dc.b	$7F,$CB,$FE,$9F,$3E,$E9,$1F,$E6,$88,$AF,$F9,$E7,$E5,$3C,$9A,$9C
	.dc.b	$F0,$FC,$7F,$85,$F3,$70,$3C,$4A,$62,$2E,$1F,$12,$3E,$49,$1F,$93
	.dc.b	$E1,$3C,$BF,$80,$D8,$FF,$4D,$35,$FF,$51,$AF,$4B,$E4,$7E,$60,$F6
	.dc.b	$A9,$2B,$FD,$4D,$9F,$F9,$F3,$99,$4F,$67,$E8,$7A,$9F,$E8,$DB,$FF
	.dc.b	$6F,$F8,$D8,$39,$3F,$D6,$FA,$FF,$D6,$18,$72,$F6,$3C,$24,$9E,$F3
	.dc.b	$AC,$5E,$C7,$6D,$A5,$7F,$BB,$1E,$0F,$53,$F0,$B2,$F9,$9F,$69,$23
	.dc.b	$A9,$AF,$5B,$E0,$F0,$C8,$CE,$7E,$E8,$E1,$1D,$63,$FD,$ED,$5F,$8F
	.dc.b	$F3,$FB,$4F,$06,$BB,$C3,$E8,$70,$7C,$95,$E6,$7C,$8F,$5A,$57,$82
	.dc.b	$72,$F8,$B8,$53,$DF,$E6,$64,$E6,$3E,$CF,$9B,$23,$93,$F7,$3C,$3F
	.dc.b	$AB,$F2,$32,$7A,$3B,$7C,$DC,$7F,$A9,$FD,$F9,$E8,$53,$F0,$B1,$67
	.dc.b	$FD,$39,$C2,$AD,$C6,$1A,$FF,$6D,$3F,$6E,$E1,$FE,$E1,$0E,$0F,$7C
	.dc.b	$7C,$D6,$F0,$7E,$3D,$B7,$F8,$8D,$9E,$BF,$E4,$3D,$79,$87,$C7,$7D
	.dc.b	$5F,$17,$FC,$0F,$F8,$05,$F1,$FF,$C3,$B4,$46,$2F,$D6,$D7,$D1,$1A
	.dc.b	$9E,$18,$FF,$8D,$27,$EF,$7A,$FF,$81,$3F,$43,$E7,$FE,$8B,$D2,$93
	.dc.b	$F6,$5F,$99,$F2,$BF,$87,$E5,$F1,$70,$1A,$4A,$F9,$B8,$BD,$AF,$C7
	.dc.b	$73,$49,$FC,$24,$7E,$C3,$C6,$91,$F8,$77,$ED,$79,$DF,$07,$E0,$BE
	.dc.b	$07,$09,$A4,$AE,$FE,$47,$E6,$FE,$26,$47,$E4,$63,$F3,$DE,$5C,$8F
	.dc.b	$CF,$3F,$87,$F6,$FF,$45,$D3,$EC,$24,$FE,$1B,$83,$FA,$3B,$F8,$0F
	.dc.b	$83,$FC,$2F,$5F,$F0,$B6,$D7,$F0,$46,$CF,$E2,$3D,$CF,$E1,$BE,$CC
	.dc.b	$8F,$D2,$3F,$CF,$FD,$67,$C1,$FD,$FF,$5B,$DE,$4A,$FB,$AF,$E0,$70
	.dc.b	$FB,$D9,$78,$7B,$3F,$A4,$46,$33,$FC,$77,$B7,$FD,$2F,$E3,$9F,$17
	.dc.b	$49,$97,$F0,$6D,$3F,$34,$FF,$08,$C1,$E0,$F7,$24,$7E,$A2,$32,$DF
	.dc.b	$87,$F0,$91,$FE,$23,$83,$E2,$3F,$9D,$F9,$9F,$D9,$A5,$E8,$7E,$A1
	.dc.b	$FD,$73,$5F,$F2,$E3,$5F,$EC,$FD,$4F,$8A,$FC,$19,$23,$F4,$2F,$C5
	.dc.b	$F9,$7F,$86,$48,$F1,$FC,$57,$C3,$F6,$3E,$3F,$CB,$C9,$F3,$C5,$7F
	.dc.b	$53,$B2,$7E,$81,$23,$F8,$4F,$07,$F9,$FB,$97,$FA,$63,$79,$3F,$A2
	.dc.b	$B6,$7F,$BF,$BF,$E0,$39,$DF,$8F,$F6,$3C,$09,$5F,$4E,$3F,$09,$E1
	.dc.b	$48,$FA,$CF,$E1,$FC,$0C,$47,$F3,$BC,$DE,$3F,$E0,$8C,$E3,$FE,$1E
	.dc.b	$F9,$67,$AD,$F4,$5B,$7F,$E1,$86,$BF,$C7,$C3,$87,$F9,$13,$FD,$0B
	.dc.b	$5F,$F4,$87,$D1,$BE,$2F,$FC,$2C,$DA,$FE,$AA,$72,$F6,$33,$3F,$81
	.dc.b	$9D,$4F,$76,$BF,$CB,$DA,$FF,$A7,$7F,$0A,$76,$7E,$25,$B1,$FE,$30
	.dc.b	$6B,$E9,$6B,$FC,$E5,$73,$7F,$8A,$FF,$09,$78,$F4,$FF,$17,$4E,$AB
	.dc.b	$D3,$F3,$1F,$C7,$7C,$2F,$F1,$EB,$63,$E3,$35,$FE,$8C,$D7,$FD,$B7
	.dc.b	$E9,$BE,$67,$FC,$86,$D8,$F9,$DF,$04,$8C,$47,$F4,$F8,$C4,$F8,$3D
	.dc.b	$07,$C6,$52,$57,$8A,$32,$FD,$9F,$99,$FF,$12,$71,$FF,$95,$1B,$19
	.dc.b	$F8,$4F,$E3,$68,$46,$9F,$EC,$0F,$E2,$27,$ED,$BC,$B9,$FF,$CD,$2E
	.dc.b	$67,$FC,$30,$72,$A5,$7E,$0B,$E6,$3F,$AC,$BF,$E6,$66,$CF,$E3,$DA
	.dc.b	$FF,$6F,$6B,$FE,$DD,$7F,$89,$C7,$1F,$27,$D3,$29,$CB,$CE,$33,$1F
	.dc.b	$88,$9F,$14,$79,$27,$CF,$39,$6C,$67,$0F,$F2,$37,$93,$FE,$45,$BF
	.dc.b	$E3,$7F,$0B,$97,$FC,$83,$0F,$84,$C6,$1F,$E7,$C8,$E0,$FF,$8B,$DD
	.dc.b	$4F,$23,$F9,$2C,$F0,$F6,$D3,$C3,$0F,$0F,$C2,$66,$3F,$C9,$37,$FC
	.dc.b	$B3,$BB,$13,$96,$C6,$4F,$E6,$31,$D7,$C0,$7C,$C7,$E0,$31,$46,$2B
	.dc.b	$F9,$E4,$E6,$31,$9F,$F0,$3F,$33,$FC,$A5,$5F,$E7,$6F,$85,$FA,$31
	.dc.b	$89,$D4,$71,$F9,$8F,$01,$ED,$F2,$5D,$78,$C3,$FC,$AB,$E7,$7E,$71
	.dc.b	$FF,$0C,$F4,$79,$63,$18,$FF,$86,$9A,$FF,$8C,$38,$1F,$0B,$F4,$23
	.dc.b	$FA,$5F,$B5,$F1,$E3,$1A,$A3,$E1,$7A,$B3,$2F,$EA,$BF,$01,$70,$FE
	.dc.b	$55,$F1,$7F,$04,$93,$FD,$7D,$F9,$19,$73,$8F,$EC,$DE,$EF,$D5,$C7
	.dc.b	$E7,$3F,$3F,$D4,$93,$FD,$88,$B8,$7F,$9C,$7F,$80,$5F,$13,$F4,$DB
	.dc.b	$3F,$E5,$46,$CF,$F8,$E9,$AF,$F9,$43,$D3,$FC,$FF,$5F,$F8,$47,$E5
	.dc.b	$FF,$15,$18,$BF,$F0,$94,$63,$CC,$FC,$04,$FE,$7E,$F9,$46,$0F,$C0
	.dc.b	$7C,$84,$BF,$EF,$CD,$9F,$F2,$5D,$7F,$8C,$67,$FC,$D2,$D8,$CB,$D8
	.dc.b	$EF,$D7,$F9,$1F,$2F,$CA,$DA,$FF,$4A,$E7,$FE,$45,$7E,$07,$C1,$7C
	.dc.b	$32,$3D,$B7,$FC,$D7,$C7,$7B,$9B,$1F,$E9,$EC,$7F,$C0,$8F,$F8,$67
	.dc.b	$E3,$BE,$1F,$0F,$E8,$5F,$D1,$7C,$17,$E7,$49,$3F,$E1,$B7,$E2,$E1
	.dc.b	$FA,$7E,$BF,$96,$D9,$ED,$5F,$AB,$DB,$E7,$B6,$3F,$D5,$FC,$6F,$73
	.dc.b	$F5,$CD,$8F,$F4,$7B,$6B,$FC,$EA,$D7,$FD,$17,$F2,$BF,$E4,$29,$FF
	.dc.b	$54,$7C,$17,$F0,$58,$C4,$FF,$AE,$7E,$1A,$F3,$FF,$B6,$D7,$CB,$FF
	.dc.b	$0B,$6E,$1F,$EC,$2B,$7C,$BE,$6B,$A1,$FC,$0E,$37,$87,$FA,$21,$AF
	.dc.b	$92,$35,$EC,$CB,$FF,$54,$D7,$F8,$05,$B1,$CF,$E1,$FC,$58,$7A,$5B
	.dc.b	$3E,$3F,$07,$0D,$3C,$D1,$EF,$F8,$FF,$55,$D9,$FF,$65,$37,$2F,$F6
	.dc.b	$D6,$3F,$EC,$C9,$F7,$7C,$7E,$75,$FF,$7F,$4F,$FA,$BA,$7E,$17,$C6
	.dc.b	$FC,$BF,$63,$E6,$36,$3D,$71,$F0,$C8,$FB,$BE,$37,$58,$F8,$1B,$87
	.dc.b	$6F,$93,$FA,$1F,$2F,$F5,$DB,$7F,$47,$B7,$E7,$11,$F9,$2F,$D1,$F8
	.dc.b	$4F,$B1,$C1,$F7,$DB,$3F,$67,$E5,$FF,$04,$7E,$0F,$CD,$B1,$FE,$2E
	.dc.b	$E3,$FF,$59,$37,$E4,$3F,$92,$6D,$FE,$1E,$8C,$1F,$E3,$45,$D9,$FD
	.dc.b	$A9,$B1,$FE,$75,$8C,$19,$EC,$91,$FE,$79,$6C,$FE,$6A,$BF,$AD,$75
	.dc.b	$BE,$2E,$C6,$1B,$7F,$A6,$8D,$F9,$98,$B7,$17,$F7,$5F,$51,$F2,$BF
	.dc.b	$EF,$AC,$3F,$5A,$7F,$67,$71,$5C,$23,$FC,$35,$3F,$CA,$9F,$F0,$E5
	.dc.b	$73,$AF,$F0,$5D,$2B,$AF,$C7,$FF,$63,$9C,$C6,$2F,$FC,$C5,$5F,$E9
	.dc.b	$6F,$7F,$FA,$59,$F3,$3C,$4B,$3F,$2F,$C4,$F9,$44,$7F,$D3,$CF,$CD
	.dc.b	$A7,$CB,$FF,$8B,$FE,$45,$E6,$FF,$8C,$67,$FC,$65,$5F,$E1,$59,$CC
	.dc.b	$64,$E2,$1C,$BC,$2F,$A5,$7F,$9E,$6F,$FC,$79,$F2,$FA,$AF,$ED,$6D
	.dc.b	$79,$B9,$C3,$FC,$48,$E6,$79,$CF,$96,$7C,$95,$D6,$3C,$4F,$96,$33
	.dc.b	$1E,$79,$F0,$CE,$6E,$CE,$45,$F8,$63,$31,$F0,$EB,$C2,$FB,$4E,$E7
	.dc.b	$1C,$38,$38,$38,$14,$FF,$04,$BE,$B7,$18,$79,$9D,$C7,$23,$45,$C3
	.dc.b	$DD,$87,$63,$91,$5C,$9E,$D9,$F5,$B8,$7E,$46,$1E,$C7,$FC,$DD,$87
	.dc.b	$B1,$F2,$CF,$CB,$BF,$04,$7B,$63,$A4,$79,$1A,$6C,$B1,$B3,$B9,$F0
	.dc.b	$4F,$C3,$C7,$47,$DF,$1C,$1F,$8F,$7E,$B7,$EA,$35,$CC,$7B,$63,$E0
	.dc.b	$B4,$F1,$47,$82,$7C,$CD,$8C,$BF,$66,$32,$D8,$F8,$CF,$DD,$9E,$AF
	.dc.b	$A9,$DB,$F1,$8F,$71,$C4,$A6,$4F,$69,$E1,$3D,$71,$E7,$69,$D5,$F2
	.dc.b	$BC,$4F,$AA,$7E,$1A,$FE,$B3,$FA,$68,$E4,$FC,$43,$91,$D0,$F5,$3E
	.dc.b	$C7,$F5,$ED,$C3,$F5,$AD,$9F,$C2,$4F,$D6,$8F,$A8,$F8,$AF,$EC,$3E
	.dc.b	$4B,$FA,$8D,$9F,$1B,$B7,$8B,$CC,$F8,$47,$C4,$3E,$41,$4F,$80,$7F
	.dc.b	$0C,$E6,$FE,$18,$FE,$D5,$AF,$EB,$1A,$FC,$67,$3F,$08,$EE,$FE,$73
	.dc.b	$3F,$E7,$EC,$3F,$8E,$4E,$78,$7F,$D9,$D7,$C2,$F6,$D8,$CD,$70,$9E
	.dc.b	$6F,$F1,$EA,$7C,$77,$F2,$78,$C3,$71,$E1,$7F,$C2,$2B,$C1,$8F,$F0
	.dc.b	$57,$8B,$9A,$E0,$FF,$B8,$63,$77,$DF,$0F,$5E,$5D,$AB,$F8,$AD,$72
	.dc.b	$7F,$7E,$71,$9D,$E3,$E2,$8F,$06,$1D,$95,$71,$C3,$B6,$3E,$7C,$38
	.dc.b	$D7,$8E,$FF,$A2,$63,$E1,$8E,$E7,$29,$DB,$6B,$CB,$5C,$31,$FA,$2D
	.dc.b	$9F,$34,$FA,$AB,$8E,$3E,$BC,$3B,$1E,$4A,$EF,$1F,$13,$0E,$71,$F2
	.dc.b	$4F,$45,$C7,$78,$77,$7D,$ED,$AF,$A7,$7D,$71,$F6,$13,$0F,$90,$E5
	.dc.b	$EF,$7F,$28,$ED,$1E,$59,$F3,$CE,$9C,$3C,$8E,$31,$D6,$7E,$09,$F8
	.dc.b	$3C,$7F,$1B,$3A,$24,$6A,$79,$BF,$4C,$F9,$A7,$C9,$C6,$3F,$00,$C9
	.dc.b	$F2,$2B,$E6,$9D,$9E,$47,$AC,$F1,$47,$B2,$BC,$31,$F2,$0C,$1E,$37
	.dc.b	$38,$7E,$25,$F7,$46,$4E,$B7,$EF,$7F,$08,$76,$3C,$C4,$C3,$DA,$E9
	.dc.b	$F9,$6F,$99,$CD,$7E,$06,$3A,$D7,$8B,$0F,$64,$76,$BF,$05,$79,$63
	.dc.b	$C8,$66,$BD,$38,$FE,$8E,$7A,$4F,$86,$7E,$D1,$1F,$AF,$87,$E0,$DF
	.dc.b	$09,$0E,$86,$A3,$83,$F2,$4F,$A5,$8F,$D6,$C3,$CB,$87,$87,$1F,$4E
	.dc.b	$3F,$BB,$BF,$F1,$73,$5F,$C8,$98,$9F,$F2,$1A,$E9,$7B,$3F,$E3,$9C
	.dc.b	$5A,$FF,$8E,$E7,$35,$FE,$75,$AD,$4F,$F2,$4E,$FB,$3F,$E5,$3E,$47
	.dc.b	$AF,$FE,$50,$E5,$FF,$30,$B6,$33,$3F,$E4,$33,$84,$67,$CF,$E7,$5F
	.dc.b	$E6,$8F,$5E,$D9,$B3,$8E,$D9,$87,$F9,$9B,$2F,$F3,$4D,$A7,$8F,$A7
	.dc.b	$AA,$FF,$37,$1E,$7E,$FE,$A6,$77,$F8,$31,$FC,$AF,$C3,$F8,$FC,$46
	.dc.b	$92,$3F,$CD,$B0,$F8,$ED,$92,$32,$FD,$0E,$27,$C2,$7C,$89,$FA,$4D
	.dc.b	$8C,$C7,$AC,$FE,$79,$E2,$7C,$4F,$96,$7C,$1D,$24,$EF,$8D,$3F,$EF
	.dc.b	$98,$C2,$F9,$1F,$9B,$E9,$3C,$7A,$75,$8E,$DF,$05,$F6,$68,$D2,$5F
	.dc.b	$D9,$C3,$BF,$C2,$74,$D2,$4F,$82,$3D,$1F,$09,$E9,$F8,$39,$3E,$48
	.dc.b	$AF,$FA,$16,$FD,$33,$E1,$24,$77,$C5,$F1,$DF,$0C,$91,$EE,$9F,$8D
	.dc.b	$E6,$48,$F8,$2F,$C9,$F8,$2C,$47,$E0,$B0,$6E,$F5,$BE,$29,$AE,$7F
	.dc.b	$F5,$D6,$6F,$2F,$E9,$79,$D3,$1F,$9B,$15,$E9,$83,$D5,$F9,$E6,$70
	.dc.b	$E7,$F2,$9F,$34,$95,$ED,$7E,$87,$81,$23,$E0,$35,$F8,$D6,$F5,$7E
	.dc.b	$D9,$23,$F0,$5F,$2A,$F9,$6F,$D5,$AF,$89,$FC,$0F,$B1,$2B,$71,$FB
	.dc.b	$0F,$CA,$FF,$82,$47,$DD,$6B,$F5,$5A,$FF,$AF,$F8,$91,$88,$FF,$6C
	.dc.b	$DB,$E4,$7F,$1F,$E7,$49,$F2,$C7,$93,$E5,$3F,$27,$E7,$C9,$F4,$C7
	.dc.b	$E5,$BC,$EF,$E1,$5B,$9A,$49,$FC,$96,$2F,$8B,$FC,$EF,$91,$27,$E1
	.dc.b	$C7,$C2,$F3,$7F,$43,$E4,$49,$F8,$F1,$F0,$63,$F8,$C6,$DF,$BB,$9F
	.dc.b	$85,$18,$3F,$8E,$60,$F8,$8F,$D5,$F9,$12,$BE,$8C,$7D,$0F,$33,$64
	.dc.b	$8F,$3B,$EC,$F6,$E4,$7E,$AF,$64,$62,$3F,$49,$83,$D9,$F0,$92,$3E
	.dc.b	$0B,$FC,$BF,$D3,$7C,$17,$EF,$BD,$59,$5F,$C1,$67,$2F,$F0,$7F,$E0
	.dc.b	$1F,$E0,$8F,$EE,$DA,$FE,$CD,$AF,$DF,$6A,$1D,$1F,$AA,$F6,$7F,$86
	.dc.b	$92,$3C,$1F,$03,$FC,$F2,$DE,$FF,$F3,$EF,$8B,$7C,$C5,$35,$19,$F9
	.dc.b	$AF,$E9,$52,$69,$30,$FE,$25,$0F,$99,$E8,$48,$F8,$1E,$5F,$87,$C4
	.dc.b	$FE,$5D,$E6,$C9,$F8,$91,$FD,$43,$77,$9D,$FD,$65,$B3,$F1,$63,$07
	.dc.b	$93,$E2,$3D,$44,$8F,$E6,$5F,$1E,$F8,$AF,$E7,$9E,$9F,$F4,$12,$4F
	.dc.b	$F6,$13,$FB,$93,$EE,$F5,$E4,$7B,$1A,$FF,$25,$E2,$FD,$13,$F6,$FC
	.dc.b	$6F,$98,$7C,$DF,$16,$C6,$48,$7F,$1F,$9E,$2D,$F0,$3E,$A7,$83,$23
	.dc.b	$F0,$D1,$5E,$11,$7C,$EF,$EC,$BD,$5F,$E0,$8A,$FE,$B5,$F3,$1F,$D6
	.dc.b	$2F,$FB,$68,$5A,$68,$DF,$B7,$67,$E1,$4E,$9B,$3F,$E1,$68,$C4,$7F
	.dc.b	$81,$DF,$E8,$2F,$0F,$5B,$44,$7E,$67,$CB,$7F,$1F,$7F,$B5,$05,$F2
	.dc.b	$1F,$DC,$5B,$1A,$78,$3E,$A7,$2F,$F8,$5D,$AF,$E8,$4F,$D1,$38,$70
	.dc.b	$D7,$F4,$DF,$2C,$F5,$DE,$0F,$E0,$63,$0D,$8C,$CF,$F8,$96,$72,$FF
	.dc.b	$89,$FD,$F7,$A6,$F9,$6F,$F0,$8B,$67,$4D,$7F,$C8,$6D,$7D,$15,$FD
	.dc.b	$6F,$E2,$3C,$5E,$D1,$A4,$9F,$F1,$B7,$CB,$7D,$AF,$67,$E5,$BF,$E1
	.dc.b	$A6,$F0,$FF,$8A,$39,$7F,$C5,$34,$AD,$3F,$84,$79,$36,$78,$57,$CB
	.dc.b	$FB,$2F,$DF,$3D,$0F,$B3,$F2,$FF,$D6,$5F,$D4,$36,$BF,$C9,$4D,$7E
	.dc.b	$A0,$70,$FD,$77,$F0,$31,$D3,$DA,$F0,$CF,$F1,$B9,$CB,$F9,$A7,$B6
	.dc.b	$3E,$C7,$2D,$E3,$E3,$2C,$3D,$F3,$9C,$3F,$C6,$9E,$3B,$E0,$3B,$F5
	.dc.b	$23,$FC,$77,$3A,$9C,$C6,$30,$FF,$0B,$3C,$67,$47,$4E,$CD,$D7,$0D
	.dc.b	$78,$FC,$15,$AF,$0B,$70,$E7,$3D,$BA,$F1,$9F,$D8,$F8,$2F,$E4,$38
	.dc.b	$7F,$15,$6D,$FF,$9D,$DA,$FC,$E6,$BF,$D0,$F4,$F6,$75,$FC,$49,$E3
	.dc.b	$7F,$C2,$35,$FF,$1C,$E1,$FB,$49,$CB,$C7,$5F,$D2,$BF,$D5,$FA,$BE
	.dc.b	$3F,$87,$F9,$58,$C1,$C7,$E5,$BF,$1F,$A7,$FA,$5D,$AF,$F5,$5C,$11
	.dc.b	$88,$FF,$50,$3A,$D3,$E9,$7C,$1F,$D0,$3E,$8F,$8F,$70,$FB,$3F,$1F
	.dc.b	$F5,$0F,$AB,$E5,$D8,$FB,$BE,$07,$3D,$FF,$7D,$1F,$D5,$23,$06,$72
	.dc.b	$FB,$A7,$DE,$EC,$B3,$F8,$63,$F0,$27,$E0,$BE,$0A,$BF,$E7,$DF,$94
	.dc.b	$FF,$1E,$3F,$E7,$16,$ED,$FE,$69,$6B,$FE,$35,$6B,$F5,$5F,$CD,$C5
	.dc.b	$EC,$FC,$9C,$3E,$42,$C6,$3E,$0B,$3D,$B6,$FC,$4F,$0F,$9A,$F9,$58
	.dc.b	$1A,$49,$F3,$7C,$1F,$69,$EE,$D8,$F6,$7C,$17,$8B,$E5,$FC,$6D,$8F
	.dc.b	$81,$F1,$FD,$88,$FC,$17,$C3,$F0,$9E,$A3,$D7,$F1,$F6,$7E,$47,$C2
	.dc.b	$7B,$BD,$4F,$7B,$63,$E6,$7C,$87,$A8,$8F,$CD,$7D,$BF,$1D,$F1,$8F
	.dc.b	$8F,$F0,$D6,$7E,$A7,$C1,$FC,$AF,$8F,$F9,$6D,$8F,$B1,$E7,$7C,$62
	.dc.b	$3F,$65,$FC,$97,$CA,$7F,$8E,$23,$FD,$74,$DE,$8F,$F4,$B7,$1F,$F8
	.dc.b	$BE,$73,$F0,$BF,$C2,$A9,$F2,$B4,$C7,$CB,$5A,$E7,$F0,$DD,$88,$FF
	.dc.b	$AE,$69,$1E,$2F,$96,$E8,$F5,$6C,$F9,$3C,$8F,$07,$C3,$78,$5B,$19
	.dc.b	$26,$9F,$EA,$DC,$3F,$D9,$D1,$8F,$37,$FC,$13,$1F,$EA,$F8,$F7,$FC
	.dc.b	$27,$AF,$E4,$3D,$8D,$8C,$9F,$12,$32,$FC,$6F,$67,$E1,$7C,$2F,$C3
	.dc.b	$6C,$7C,$AF,$8F,$C3,$4E,$07,$CE,$F8,$3F,$95,$1F,$2D,$B1,$F4,$BE
	.dc.b	$3F,$E7,$7C,$27,$CF,$6C,$7D,$6F,$84,$C3,$4E,$07,$DA,$F8,$DF,$AD
	.dc.b	$F0,$BF,$5D,$B1,$F7,$BD,$7F,$B4,$7D,$BF,$8C,$B1,$FE,$18,$E2,$77
	.dc.b	$7F,$4E,$6C,$67,$BF,$13,$9E,$FF,$E3,$D4,$69,$2B,$F9,$E7,$AD,$B7
	.dc.b	$F9,$93,$67,$FD,$C9,$F2,$5F,$C6,$9B,$E9,$FF,$58,$87,$AD,$63,$0D
	.dc.b	$BF,$F4,$27,$4F,$F9,$1A,$B7,$18,$3F,$A3,$3D,$63,$FC,$97,$F2,$9F
	.dc.b	$C4,$67,$58,$7F,$A3,$23,$F8,$EC,$EB,$D2,$FE,$D3,$18,$AF,$EA,$1F
	.dc.b	$27,$FE,$8A,$78,$FB,$49,$E9,$87,$FB,$FE,$B4,$FF,$C0,$18,$CE,$1F
	.dc.b	$D2,$BC,$7F,$9C,$7C,$FF,$1E,$E3,$F4,$ED,$F2,$DF,$EA,$1E,$37,$93
	.dc.b	$FE,$10,$B5,$E6,$3F,$C2,$77,$C0,$57,$07,$FC,$95,$5C,$1C,$CF,$0B
	.dc.b	$FF,$57,$7C,$B6,$E7,$FD,$61,$5B,$8C,$47,$FA,$CE,$77,$5E,$3C,$2C
	.dc.b	$FF,$89,$26,$BC,$A2,$BC,$45,$74,$BE,$D7,$E3,$9C,$BD,$5A,$75,$29
	.dc.b	$D4,$A7,$59,$E9,$5E,$08,$F8,$91,$87,$66,$27,$B4,$7B,$A3,$07,$C8
	.dc.b	$9E,$57,$F3,$E3,$07,$CA,$3C,$98,$75,$39,$47,$C0,$C3,$DB,$3E,$9B
	.dc.b	$F1,$1A,$7D,$31,$A7,$FD,$01,$3C,$6F,$C3,$1A,$6D,$72,$C3,$E4,$61
	.dc.b	$E1,$69,$F2,$5E,$D1,$BA,$F1,$4F,$C0,$36,$FB,$27,$C8,$73,$76,$7C
	.dc.b	$09,$E7,$3C,$63,$27,$D3,$8C,$BB,$8F,$B9,$1E,$18,$F6,$9E,$B3,$E9
	.dc.b	$DF,$CD,$7E,$E5,$78,$AB,$D4,$72,$3C,$11,$CA,$BA,$3F,$81,$8C,$B4
	.dc.b	$F1,$1F,$59,$E6,$D3,$E0,$3D,$5D,$9D,$5E,$6F,$E1,$0F,$73,$B6,$9C
	.dc.b	$CF,$4B,$F3,$1A,$7C,$63,$FC,$9C,$7C,$03,$E1,$47,$C8,$79,$9E,$13
	.dc.b	$C6,$7A,$CC,$94,$E6,$FC,$77,$D6,$78,$CE,$F3,$D5,$DC,$FA,$5A,$6F
	.dc.b	$0F,$E0,$D3,$CE,$3D,$27,$D8,$7F,$08,$D9,$F9,$6F,$DC,$E6,$FE,$22
	.dc.b	$FE,$DD,$AE,$67,$F7,$D1,$8F,$89,$77,$35,$5F,$FB,$6A,$FF,$DC,$55
	.dc.b	$A9,$CD,$F1,$BE,$17,$B3,$8D,$6D,$C0,$71,$7F,$C7,$D7,$77,$FB,$BE
	.dc.b	$30,$DC,$7C,$37,$DA,$78,$36,$3C,$B3,$BB,$F0,$B5,$F0,$4E,$DF,$2B
	.dc.b	$E8,$CF,$BC,$7F,$28,$C3,$C6,$D9,$D5,$F8,$B2,$FD,$35,$6A,$BF,$89
	.dc.b	$1E,$DC,$3C,$11,$EE,$AE,$13,$FC,$A2,$38,$61,$E0,$8C,$61,$D9,$F2
	.dc.b	$E1,$E3,$C3,$E9,$C7,$16,$D7,$C2,$9F,$31,$B9,$87,$47,$2F,$9A,$FD
	.dc.b	$18,$F8,$4E,$51,$9C,$3C,$4F,$AC,$E7,$7A,$26,$3F,$42,$32,$FB,$A3
	.dc.b	$9D,$75,$8D,$4F,$5C,$3E,$4D,$F6,$C3,$E1,$D6,$B0,$EB,$7F,$08,$F8
	.dc.b	$07,$63,$0F,$68,$EB,$87,$79,$E0,$F2,$7D,$47,$23,$E2,$12,$7E,$91
	.dc.b	$27,$D4,$48,$FC,$59,$E1,$3F,$17,$87,$CC,$AF,$04,$76,$7C,$8F,$9A
	.dc.b	$7D,$84,$70,$79,$9F,$49,$1C,$1A,$3D,$35,$E5,$71,$7F,$AD,$8C,$46
	.dc.b	$A7,$D0,$F9,$8F,$81,$1A,$9E,$13,$F9,$4A,$F4,$C7,$E1,$23,$07,$27
	.dc.b	$EA,$D7,$C2,$3E,$44,$67,$1F,$C9,$BE,$D8,$F0,$5F,$D1,$AE,$31,$E5
	.dc.b	$3D,$38,$FE,$06,$F8,$C7,$5A,$E1,$39,$8F,$A1,$5E,$87,$E2,$9F,$3E
	.dc.b	$FE,$43,$EF,$C3,$F8,$67,$77,$EA,$32,$FF,$18,$4F,$F1,$18,$C6,$DF
	.dc.b	$C7,$1B,$1E,$D9,$FF,$1B,$71,$38,$9F,$21,$A8,$FF,$28,$7A,$37,$90
	.dc.b	$E6,$B7,$87,$F9,$5A,$72,$DF,$47,$81,$C6,$B9,$76,$73,$EC,$95,$C6
	.dc.b	$2B,$FE,$5B,$8B,$9F,$F9,$B2,$7D,$32,$FC,$53,$F8,$3B,$FC,$EF,$6F
	.dc.b	$E7,$3E,$16,$B8,$8F,$0A,$DE,$1F,$1A,$73,$7F,$E5,$C3,$F9,$C5,$56
	.dc.b	$99,$8D,$BF,$E6,$25,$E0,$F7,$24,$7F,$A0,$0F,$E8,$5E,$17,$C7,$F8
	.dc.b	$F9,$1F,$E7,$CE,$04,$7C,$BF,$91,$F9,$9F,$2F,$F3,$7E,$46,$5E,$E3
	.dc.b	$E8,$7C,$2F,$D1,$F8,$99,$3C,$E2,$BF,$E8,$5F,$8A,$FA,$C6,$37,$FB
	.dc.b	$F9,$BE,$27,$ED,$7C,$54,$BF,$24,$57,$FD,$1D,$F1,$1F,$7C,$C5,$F2
	.dc.b	$7F,$05,$F1,$52,$7D,$91,$F8,$5F,$2F,$F0,$DF,$13,$27,$E0,$45,$7F
	.dc.b	$D3,$51,$6F,$CB,$3F,$8D,$F8,$79,$5F,$1D,$FC,$87,$99,$88,$FF,$51
	.dc.b	$79,$BF,$93,$31,$7E,$EF,$88,$FC,$B7,$C4,$4A,$FA,$11,$5F,$C4,$F8
	.dc.b	$5F,$A1,$9F,$C1,$36,$3F,$D3,$ED,$77,$F2,$8F,$97,$FD,$33,$63,$FD
	.dc.b	$6E,$D7,$FD,$6D,$DF,$AF,$8A,$91,$A7,$8F,$A3,$CB,$E2,$7C,$36,$69
	.dc.b	$2B,$C7,$0F,$89,$F0,$FC,$54,$9F,$0C,$79,$3E,$2A,$47,$91,$F3,$FC
	.dc.b	$2F,$A3,$E2,$BB,$92,$7A,$BE,$2F,$2B,$DB,$F1,$52,$7D,$B1,$F0,$3C
	.dc.b	$C9,$1F,$01,$F8,$7F,$15,$F1,$3E,$2B,$D6,$49,$F4,$BE,$EF,$4F,$E4
	.dc.b	$FC,$54,$9F,$93,$1F,$33,$DD,$F9,$BE,$84,$9F,$9B,$1F,$43,$D1,$F8
	.dc.b	$E4,$8F,$8A,$FC,$AF,$6B,$EA,$FB,$92,$7E,$AC,$7F,$1D,$F2,$5F,$15
	.dc.b	$49,$3F,$68,$D7,$C4,$FE,$12,$4D,$24,$7E,$1A,$1F,$13,$D4,$91,$FE
	.dc.b	$FA,$87,$C5,$78,$7E,$26,$4E,$5F,$27,$C5,$79,$7E,$0F,$C0,$49,$EC
	.dc.b	$F8,$FE,$23,$D4,$48,$ED,$F1,$5E,$DF,$88,$91,$E1,$7E,$07,$C4,$FF
	.dc.b	$2A,$97,$9D,$FC,$7D,$B5,$FC,$8D,$AF,$FC,$00,$6B,$FE,$36,$F1,$DF
	.dc.b	$18,$E9,$FE,$90,$D9,$FF,$36,$F9,$7F,$D3,$A3,$11,$FC,$DC,$53,$E5
	.dc.b	$1E,$8F,$D5,$F8,$BF,$37,$AB,$2F,$E5,$45,$7C,$91,$E9,$23,$F4,$FE
	.dc.b	$13,$EF,$11,$FA,$5F,$0D,$5F,$16,$2E,$6F,$F2,$F6,$9F,$D9,$1B,$E7
	.dc.b	$7F,$A1,$A7,$58,$7E,$27,$E0,$26,$1F,$75,$FC,$67,$C1,$62,$3A,$E0
	.dc.b	$F8,$1F,$C8,$19,$CB,$FB,$BF,$CA,$7F,$33,$AF,$E9,$0D,$7E,$30,$71
	.dc.b	$1F,$DF,$BE,$0B,$F3,$33,$C3,$DC,$FE,$CB,$4D,$E9,$88,$B3,$FD,$7E
	.dc.b	$74,$D9,$FE,$C1,$18,$8F,$ED,$A3,$E5,$1E,$AF,$F7,$26,$C7,$F5,$F6
	.dc.b	$BE,$8F,$C1,$5E,$26,$EB,$87,$AB,$2B,$84,$7F,$7C,$F9,$47,$B6,$F8
	.dc.b	$DF,$EE,$91,$86,$C6,$6B,$CB,$39,$7F,$BA,$74,$3E,$57,$FB,$C3,$63
	.dc.b	$31,$83,$AF,$BF,$E6,$8E,$6D,$7F,$C1,$B4,$8C,$47,$FA,$5A,$31,$9F
	.dc.b	$F8,$27,$E5,$1E,$E3,$AF,$FC,$05,$F3,$EF,$81,$C6,$3F,$72,$F7,$5A
	.dc.b	$9C,$F3,$FF,$8E,$A7,$E7,$CF,$5E,$0F,$F1,$ED,$FF,$A9,$A7,$2F,$D6
	.dc.b	$6B,$FE,$24,$FB,$77,$BD,$C2,$3F,$DB,$F1,$8B,$FF,$10,$36,$3E,$FB
	.dc.b	$5F,$F1,$6F,$E1,$6F,$1B,$CB,$5F,$EB,$6A,$E1,$97,$F9,$42,$B5,$3F
	.dc.b	$E2,$19,$CB,$FE,$BC,$87,$95,$E3,$C1,$18,$9F,$F1,$93,$63,$F2,$2D
	.dc.b	$7F,$88,$78,$CF,$57,$C7,$1F,$C2,$BC,$07,$CA,$63,$C5,$FE,$2B,$56
	.dc.b	$73,$8F,$F1,$43,$F7,$98,$73,$71,$87,$F1,$48,$C3,$FC,$19,$7A,$DD
	.dc.b	$5C,$EB,$FC,$77,$D6,$ED,$F2,$39,$6E,$1D,$33,$FE,$4F,$19,$8C,$5B
	.dc.b	$D7,$C6,$5F,$C2,$3B,$BE,$C6,$1F,$C5,$5B,$A7,$E9,$70,$FE,$A1,$1A
	.dc.b	$C3,$F5,$F1,$FD,$3E,$3C,$DD,$3F,$05,$F8,$9F,$40,$F9,$AA,$7C,$7F
	.dc.b	$9D,$FE,$C7,$7E,$79,$D3,$E6,$CF,$B1,$1F,$F1,$2E,$28,$F3,$D7,$F5
	.dc.b	$E9,$FF,$14,$B6,$3F,$C5,$ED,$7F,$A5,$35,$FE,$5D,$5F,$E9,$29,$F1
	.dc.b	$7C,$A7,$97,$D7,$F4,$BE,$AF,$91,$F8,$B4,$69,$2B,$C3,$D3,$ED,$3D
	.dc.b	$DD,$16,$3D,$DE,$5F,$C4,$3E,$17,$AD,$63,$E1,$7C,$8F,$62,$3F,$E4
	.dc.b	$4C,$51,$F0,$3D,$2F,$94,$7C,$BF,$96,$B5,$F2,$FE,$4B,$E8,$9F,$3F
	.dc.b	$D4,$B1,$F3,$F9,$BD,$64,$7B,$3E,$EF,$91,$FB,$27,$D7,$F4,$EC,$FE
	.dc.b	$37,$93,$FD,$26,$FD,$9F,$92,$FB,$9E,$B7,$E0,$1F,$C0,$FA,$BF,$20
	.dc.b	$93,$EB,$7E,$CF,$57,$E0,$CF,$D5,$7B,$9F,$82,$E9,$FE,$CF,$8F,$E4
	.dc.b	$5F,$ED,$31,$F5,$DF,$ED,$7C,$5F,$D4,$70,$FD,$04,$E7,$C1,$FF,$67
	.dc.b	$CE,$6B,$7F,$23,$F6,$89,$7F,$35,$FD,$0D,$FC,$E8,$E7,$C9,$E4,$E2
	.dc.b	$F2,$BE,$6E,$1F,$CE,$FB,$36,$BF,$34,$FE,$83,$E4,$7F,$0E,$48,$FB
	.dc.b	$4F,$6D,$FD,$E7,$C0,$F6,$6C,$FF,$AE,$FA,$FF,$9D,$37,$CE,$FE,$23
	.dc.b	$B7,$F2,$5F,$56,$D6,$A3,$FB,$C9,$FC,$28,$FE,$D7,$E4,$FF,$21,$70
	.dc.b	$7F,$B6,$9C,$79,$3D,$23,$F8,$6E,$1F,$ED,$E8,$C4,$FD,$4E,$19,$3F
	.dc.b	$3B,$5B,$7F,$51,$FB,$27,$DA,$D6,$CF,$D4,$F7,$7F,$81,$9F,$78,$AF
	.dc.b	$D8,$CF,$0D,$38,$1F,$C1,$74,$FE,$16,$7E,$1B,$D4,$B1,$F8,$07,$0F
	.dc.b	$E0,$BE,$46,$E5,$F8,$87,$F1,$A7,$E3,$BE,$46,$CF,$E2,$0F,$C8,$9F
	.dc.b	$CE,$FE,$57,$FD,$60,$E4,$FE,$58,$E2,$BF,$C7,$6D,$E8,$FE,$DF,$E5
	.dc.b	$A3,$FB,$7C,$EB,$E4,$BA,$7A,$B3,$0E,$7E,$DD,$AE,$6F,$84,$F1,$7C
	.dc.b	$8D,$9F,$E4,$DB,$FF,$2C,$3F,$03,$B5,$8F,$E9,$77,$FC,$97,$D0,$FC
	.dc.b	$01,$23,$FC,$5B,$19,$6B,$FC,$E8,$FC,$4C,$7E,$0F,$D0,$FC,$23,$F8
	.dc.b	$58,$7C,$BF,$F3,$1F,$97,$7C,$5F,$F3,$39,$FE,$69,$7F,$E1,$09,$CC
	.dc.b	$64,$E2,$1C,$BC,$30,$E9,$5F,$E4,$2B,$B3,$D6,$FA,$4E,$63,$93,$D6
	.dc.b	$33,$1D,$E7,$8B,$D4,$D9,$92,$99,$BF,$1B,$FD,$36,$38,$BF,$16,$32
	.dc.b	$FA,$9D,$CF,$82,$3E,$14,$F0,$8C,$47,$33,$B9,$CE,$BD,$5E,$AF,$2C
	.dc.b	$3F,$AE,$3D,$AE,$CF,$B6,$6D,$7A,$70,$E0,$F4,$3C,$93,$B7,$15,$E3
	.dc.b	$AF,$5C,$68,$F9,$8E,$23,$51,$C0,$F3,$E1,$EA,$78,$1F,$1E,$BE,$95
	.dc.b	$F8,$8F,$44,$70,$BE,$8D,$3A,$61,$E8,$36,$5A,$E6,$7A,$6B,$99,$D5
	.dc.b	$CC,$FA,$0F,$34,$65,$F8,$91,$C4,$DB,$E7,$3F,$01,$3D,$DB,$1F,$1E
	.dc.b	$3B,$BC,$63,$99,$DE,$7C,$F1,$B3,$EC,$96,$36,$77,$3F,$06,$7C,$18
	.dc.b	$EA,$F3,$8F,$2B,$F8,$33,$89,$4E,$31,$ED,$79,$CF,$A5,$B3,$E1,$76
	.dc.b	$FA,$5A,$77,$3A,$DF,$DB,$C3,$EF,$DF,$B2,$7E,$65,$D7,$F5,$0F,$A1
	.dc.b	$F6,$1F,$9B,$8F,$23,$6F,$C0,$D3,$C0,$79,$8E,$AF,$C8,$69,$F4,$8A
	.dc.b	$6C,$F0,$BF,$1A,$FD,$11,$F2,$E3,$18,$7E,$3A,$BC,$73,$FC,$3A,$FE
	.dc.b	$33,$FB,$16,$BF,$9A,$3E,$71,$E7,$3D,$A7,$1A,$F9,$8F,$16,$B9,$3C
	.dc.b	$E6,$4A,$7F,$1E,$F1,$53,$FC,$5D,$B1,$FC,$45,$AF,$E1,$9B,$F2,$CF
	.dc.b	$35,$CB,$FE,$5F,$9C,$C6,$2B,$FD,$B1,$7F,$CC,$6F,$85,$EE,$31,$5B
	.dc.b	$BE,$33,$FC,$2A,$D6,$F3,$7F,$9D,$61,$FC,$D2,$31,$3B,$7A,$BE,$CC
	.dc.b	$7A,$DE,$6F,$8C,$78,$DB,$3F,$E7,$7A,$E3,$7D,$71,$F5,$65,$DE,$FF
	.dc.b	$A5,$36,$FF,$9C,$DF,$82,$38,$63,$DE,$61,$E3,$3B,$5F,$C5,$C3,$71
	.dc.b	$FC,$FA,$FD,$6D,$C3,$BE,$3E,$A8,$D5,$F9,$23,$D5,$1E,$29,$DB,$DB
	.dc.b	$0F,$2D,$77,$AD,$CF,$8C,$F9,$B5,$F0,$9F,$9F,$3B,$79,$1B,$BF,$2E
	.dc.b	$1F,$06,$31,$3B,$78,$E1,$E3,$61,$E8,$7C,$66,$31,$E4,$7C,$23,$E1
	.dc.b	$D7,$52,$63,$F0,$5D,$3E,$A8,$E2,$FC,$57,$35,$F6,$F1,$F8,$6F,$B8
	.dc.b	$F0,$9E,$33,$D8,$FC,$1C,$7D,$B5,$DC,$E8,$77,$24,$FC,$4A,$EE,$F5
	.dc.b	$3E,$C1,$D9,$CC,$FD,$C7,$CF,$1C,$67,$B3,$E7,$3F,$0F,$7F,$0A,$3C
	.dc.b	$31,$93,$A9,$E9,$9D,$39,$9F,$8E,$FD,$87,$C3,$7F,$0A,$FC,$57,$D2
	.dc.b	$BE,$B9,$D2,$FC,$AE,$1E,$2F,$D7,$8E,$13,$FA,$67,$F1,$77,$E0,$9E
	.dc.b	$6F,$07,$A4,$70,$C7,$F3,$11,$EB,$BE,$F7,$FA,$9A,$EB,$87,$9C,$F4
	.dc.b	$5F,$0B,$F2,$4F,$D2,$7E,$85,$FC,$E9,$D5,$FB,$A7,$B5,$78,$71,$F9
	.dc.b	$AF,$A0,$FA,$31,$D0,$D9,$0C,$12,$7A,$3E,$08,$E8,$FB,$8E,$11,$EB
	.dc.b	$7E,$F1,$8C,$3D,$75,$F8,$5A,$F5,$57,$43,$85,$7E,$2E,$59,$FF,$29
	.dc.b	$F3,$7F,$CB,$51,$89,$FE,$31,$18,$9F,$DE,$CE,$67,$F9,$4C,$FF,$9A
	.dc.b	$A8,$D2,$47,$FB,$E7,$83,$92,$E7,$79,$09,$CC,$61,$DC,$BC,$27,$8B
	.dc.b	$AE,$A7,$C8,$D2,$57,$FA,$5B,$E5,$35,$F2,$72,$3F,$D3,$2E,$9F,$F4
	.dc.b	$CF,$C9,$73,$23,$FE,$9E,$F9,$3E,$DF,$27,$23,$FD,$42,$E9,$FF,$50
	.dc.b	$FC,$9F,$8C,$8F,$FA,$9B,$E3,$7C,$DC,$B2,$3F,$D5,$2E,$9F,$F7,$CF
	.dc.b	$89,$F0,$38,$8D,$38,$47,$FA,$BB,$DE,$F8,$1E,$64,$8F,$F5,$9C,$57
	.dc.b	$FB,$3F,$41,$F9,$D6,$F1,$BB,$1B,$E9,$FF,$5B,$F1,$FF,$A5,$5B,$5F
	.dc.b	$34,$F9,$DE,$0F,$CF,$F0,$24,$7F,$B0,$9D,$3F,$EB,$0E,$B7,$AB,$F6
	.dc.b	$17,$AB,$4C,$45,$9F,$F5,$9C,$E9,$B3,$F8,$B3,$FD,$6B,$D4,$F3,$7F
	.dc.b	$06,$D9,$FF,$5F,$B5,$F8,$27,$C1,$5E,$87,$E3,$3E,$75,$F2,$94,$95
	.dc.b	$F1,$F5,$FF,$6C,$CB,$E5,$3E,$A7,$17,$FB,$23,$C1,$FC,$BB,$78,$FF
	.dc.b	$45,$87,$FB,$06,$72,$FF,$AF,$CF,$CF,$5B,$E5,$3F,$3B,$F0,$2F,$3E
	.dc.b	$9E,$58,$F3,$7C,$0C,$BF,$26,$5F,$6F,$16,$2A,$F6,$7B,$3D,$3E,$27
	.dc.b	$CD,$99,$A4,$C7,$C9,$87,$2F,$80,$F1,$FA,$72,$7B,$9E,$88,$F9,$1E
	.dc.b	$FC,$8C,$D7,$8A,$3E,$67,$C6,$7B,$3D,$5F,$09,$2B,$BE,$48,$F9,$5A
	.dc.b	$FF,$84,$B8,$FF,$87,$62,$A7,$C4,$7C,$63,$EC,$79,$FF,$23,$D6,$99
	.dc.b	$78,$A9,$1F,$77,$E4,$3E,$6F,$CA,$FC,$22,$57,$89,$F8,$BE,$77,$D2
	.dc.b	$24,$7B,$34,$FC,$93,$F8,$AC,$FE,$7F,$C4,$7E,$24,$91,$BF,$5B,$ED
	.dc.b	$FA,$1F,$40,$91,$F0,$1F,$A7,$E8,$FE,$00,$91,$F2,$BE,$49,$F1,$1F
	.dc.b	$C8,$17,$C4,$52,$4E,$FE,$23,$F1,$3E,$A7,$DE,$24,$7C,$E7,$F0,$3E
	.dc.b	$A7,$E3,$89,$1D,$FD,$6E,$DF,$11,$F9,$BC,$4D,$24,$FE,$1A,$1F,$09
	.dc.b	$B2,$46,$5F,$C7,$FB,$1F,$9C,$24,$7B,$7C,$BF,$67,$C1,$FE,$83,$D8
	.dc.b	$93,$F0,$62,$BF,$C5,$CF,$CF,$75,$27,$F3,$64,$8F,$27,$C1,$7C,$72
	.dc.b	$3F,$2F,$E0,$EB,$F6,$69,$ED,$FE,$C0,$CF,$81,$F9,$0D,$5F,$05,$F4
	.dc.b	$49,$7F,$57,$E0,$EB,$CF,$DD,$FD,$C9,$8B,$FE,$C5,$EF,$7E,$F6,$7F
	.dc.b	$80,$1F,$D6,$FC,$EA,$D3,$31,$B7,$F8,$32,$F9,$4F,$EB,$FF,$7D,$F4
	.dc.b	$27,$F6,$61,$C3,$FC,$78,$FE,$D1,$1C,$39,$3B,$CE,$A7,$E6,$1C,$63
	.dc.b	$27,$93,$E5,$35,$1F,$45,$AF,$F4,$96,$A6,$9F,$CB,$F1,$FF,$01,$7D
	.dc.b	$0F,$AD,$B9,$7F,$78,$9D,$7A,$5F,$77,$E4,$BE,$7F,$C3,$FC,$66,$E3
	.dc.b	$FC,$9D,$AF,$A1,$AF,$FA,$F7,$CE,$7C,$9F,$F4,$86,$C7,$F2,$FF,$75
	.dc.b	$18,$8F,$83,$18,$9F,$07,$03,$DF,$A4,$AF,$14,$65,$FE,$75,$F2,$4F
	.dc.b	$57,$F9,$FC,$7F,$6D,$8C,$69,$F8,$06,$C7,$09,$DA,$8F,$E4,$53,$F7
	.dc.b	$1D,$EF,$FE,$2B,$73,$3F,$75,$CC,$FE,$97,$E4,$7F,$C4,$4F,$E3,$9B
	.dc.b	$3F,$D3,$DA,$FE,$01,$AF,$1F,$5B,$C7,$06,$92,$7A,$7C,$B7,$F9,$07
	.dc.b	$0F,$F1,$F5,$6E,$7F,$26,$A7,$51,$D4,$FE,$69,$F1,$B8,$C3,$FC,$7D
	.dc.b	$DA,$8C,$79,$FF,$CF,$FE,$22,$CE,$6F,$F9,$E4,$62,$B2,$D8,$FE,$81
	.dc.b	$F1,$1C,$0E,$3F,$0D,$63,$FA,$23,$71,$FE,$04,$DD,$3F,$A7,$E3,$F8
	.dc.b	$1A,$FE,$AB,$C3,$F7,$33,$FE,$A4,$D9,$D6,$1F,$DE,$FE,$13,$FC,$6B
	.dc.b	$7F,$BE,$6C,$67,$E2,$7E,$FC,$62,$3F,$C0,$9F,$21,$FE,$10,$AF,$ED
	.dc.b	$2D,$8F,$E2,$C7,$C4,$71,$97,$F6,$47,$FC,$25,$3F,$DE,$1B,$3F,$BC
	.dc.b	$C9,$18,$8E,$FF,$09,$EB,$8F,$F0,$85,$78,$3E,$5F,$B7,$6F,$77,$C1
	.dc.b	$F0,$72,$7C,$BE,$15,$C7,$FB,$16,$0F,$83,$E8,$4A,$F2,$3D,$F9,$FD
	.dc.b	$5B,$7A,$DB,$3F,$03,$E0,$BD,$C7,$BF,$9E,$C7,$C4,$EE,$F4,$11,$F8
	.dc.b	$0F,$AF,$E0,$3E,$37,$C6,$FC,$76,$CF,$CC,$F8,$DF,$80,$47,$E4,$3F
	.dc.b	$0F,$E0,$64,$7C,$B6,$ED,$F2,$F6,$FF,$2C,$CF,$CF,$F8,$BF,$94,$47
	.dc.b	$E9,$7B,$FF,$5F,$E2,$FE,$C3,$63,$EE,$FC,$07,$D1,$23,$F6,$5F,$C7
	.dc.b	$71,$3E,$4F,$F0,$1E,$4F,$E9,$23,$15,$F8,$68,$C7,$BD,$9F,$7E,$CF
	.dc.b	$F8,$CA,$BF,$C1,$2D,$F7,$FA,$91,$FF,$1F,$C2,$3C,$3F,$25,$DF,$E4
	.dc.b	$BC,$0D,$AE,$9E,$F7,$8F,$DF,$F2,$36,$3C,$1F,$27,$EA,$23,$D9,$F6
	.dc.b	$FA,$9E,$BF,$8F,$F6,$36,$7F,$CE,$FE,$CF,$F0,$5E,$6F,$F5,$E3,$5F
	.dc.b	$E0,$47,$F5,$DE,$FF,$EE,$B3,$FD,$8A,$1E,$67,$F2,$C6,$C7,$F6,$E8
	.dc.b	$C1,$9F,$32,$47,$F7,$26,$CF,$EA,$78,$1E,$FF,$D1,$F8,$0F,$A4,$DB
	.dc.b	$F9,$DE,$F6,$1A,$7D,$18,$FB,$19,$FD,$9F,$7E,$CF,$D5,$6E,$3F,$E4
	.dc.b	$BC,$7D,$6D,$F7,$73,$E6,$FF,$98,$A7,$FD,$C5,$7F,$77,$D6,$FC,$27
	.dc.b	$9F,$F8,$56,$C7,$DC,$8F,$C5,$91,$FB,$FF,$23,$AC,$7C,$0D,$C3,$FD
	.dc.b	$7F,$F1,$DF,$8F,$AF,$E6,$7F,$23,$FE,$93,$7F,$18,$63,$E3,$FF,$92
	.dc.b	$47,$FA,$2E,$3F,$D1,$2D,$C3,$F8,$93,$5F,$F5,$13,$FD,$6A,$FF,$D9
	.dc.b	$BE,$D7,$6F,$8F,$EE,$F8,$3E,$3A,$61,$FE,$8C,$BF,$F4,$0B,$63,$F0
	.dc.b	$DC,$FF,$13,$1F,$C2,$6D,$F8,$7F,$67,$2F,$E2,$7D,$9F,$C2,$13,$0F
	.dc.b	$F1,$8B,$F8,$7A,$7C,$9F,$FB,$97,$C7,$7C,$27,$F8,$9A,$97,$98,$FF
	.dc.b	$5D,$55,$BF,$F2,$B5,$69,$5F,$27,$FC,$5F,$3F,$EF,$1F,$93,$C4,$7F
	.dc.b	$BC,$EB,$84,$62,$3F,$DE,$D3,$BB,$F1,$61,$67,$98,$AE,$17,$CA,$BC
	.dc.b	$B3,$97,$8B,$4E,$33,$C2,$BA,$47,$F9,$BE,$EB,$FD,$7A,$7C,$8F,$C6
	.dc.b	$3D,$B5,$D3,$0C,$D7,$C3,$9F,$65,$D7,$D6,$78,$4C,$94,$CD,$F8,$DC
	.dc.b	$1D,$EB,$07,$67,$C2,$7D,$09,$EC,$EE,$FC,$87,$C2,$9E,$2F,$C4,$AE
	.dc.b	$C7,$32,$BB,$38,$CF,$CE,$9E,$2F,$C8,$8F,$97,$3C,$5A,$71,$8F,$21
	.dc.b	$E6,$3E,$61,$4F,$F0,$4C,$FA,$CE,$AE,$9A,$FC,$86,$BE,$D3,$AB,$E1
	.dc.b	$7B,$CD,$7E,$6B,$5F,$B5,$3F,$04,$F3,$BE,$D7,$DE,$F8,$5A,$F8,$5A
	.dc.b	$F8,$4F,$B3,$3E,$98,$CD,$FD,$FC,$3F,$45,$39,$AF,$3D,$7C,$57,$8B
	.dc.b	$E3,$7D,$2D,$7D,$2D,$7D,$27,$D1,$36,$7A,$4E,$A5,$3A,$94,$F9,$EF
	.dc.b	$85,$A7,$33,$B9,$4E,$E7,$C3,$3A,$94,$EA,$53,$99,$67,$E0,$C7,$A5
	.dc.b	$FD,$A4,$75,$6B,$BC,$3F,$33,$7F,$25,$B1,$FA,$59,$DB,$FA,$6A,$FB
	.dc.b	$2F,$E3,$30,$FC,$CB,$D6,$BF,$72,$FE,$DE,$32,$EE,$78,$BF,$74,$F6
	.dc.b	$94,$F6,$9D,$CF,$9E,$71,$29,$B3,$E0,$96,$BE,$0B,$F4,$5A,$FB,$CF
	.dc.b	$D1,$1F,$C2,$FC,$8F,$E4,$EF,$F1,$C6,$BF,$A8,$6B,$FC,$29,$CF,$C0
	.dc.b	$BD,$27,$1A,$95,$A9,$CC,$63,$4D,$2B,$E5,$7C,$6F,$83,$6B,$73,$AA
	.dc.b	$FD,$E2,$F4,$B9,$CF,$F3,$C8,$C4,$FF,$23,$78,$46,$70,$F1,$BB,$AE
	.dc.b	$39,$76,$6D,$FA,$EF,$8C,$F8,$AF,$D7,$9F,$1A,$FE,$73,$5E,$07,$F9
	.dc.b	$73,$E7,$8F,$06,$3E,$3C,$3A,$D7,$C1,$6D,$76,$8F,$05,$F6,$C3,$C9
	.dc.b	$1D,$AB,$E4,$E1,$D2,$3B,$D6,$DF,$44,$EA,$B8,$C6,$B0,$F3,$D6,$E3
	.dc.b	$C4,$75,$33,$11,$F9,$F5,$E6,$3C,$F3,$F2,$AB,$E2,$57,$1A,$F8,$35
	.dc.b	$E1,$7D,$71,$CF,$1F,$91,$1D,$0F,$AD,$7C,$DD,$99,$24,$7B,$CF,$81
	.dc.b	$87,$22,$61,$E2,$9F,$71,$C1,$F8,$33,$C5,$D3,$E2,$23,$83,$94,$F4
	.dc.b	$9F,$96,$7C,$37,$C3,$1F,$31,$D4,$7E,$6E,$BE,$66,$3F,$6E,$3E,$69
	.dc.b	$E3,$3D,$B7,$E3,$7A,$99,$39,$61,$F1,$0F,$C4,$30,$FD,$0D,$7C,$77
	.dc.b	$67,$1A,$FC,$95,$75,$8C,$DF,$9A,$7F,$2C,$FA,$A3,$E7,$B9,$C7,$EC
	.dc.b	$47,$0C,$7E,$23,$E6,$BF,$8E,$FA,$A3,$2F,$BA,$32,$F9,$1F,$15,$F9
	.dc.b	$27,$E9,$9C,$1F,$D3,$3F,$2C,$93,$E4,$75,$5F,$91,$8F,$88,$7C,$2C
	.dc.b	$7F,$21,$87,$B7,$1F,$A2,$7C,$5C,$3C,$15,$E9,$C3,$C1,$39,$C3,$A4
	.dc.b	$F3,$25,$FC,$37,$F0,$15,$E2,$7E,$96,$3E,$CB,$FB,$F7,$F2,$30,$F9
	.dc.b	$B7,$F8,$A2,$5F,$DE,$C3,$CD,$8F,$48,$C4,$FC,$07,$E8,$1F,$34,$F4
	.dc.b	$10,$F9,$8F,$EF,$1F,$E5,$67,$E3,$8B,$5F,$C4,$DB,$7A,$7F,$98,$F0
	.dc.b	$3C,$6F,$E6,$78,$7F,$9A,$65,$39,$7F,$9E,$46,$63,$11,$FD,$03,$6F
	.dc.b	$E6,$BE,$97,$27,$FA,$17,$B8,$E5,$7C,$7E,$6D,$39,$8C,$36,$3F,$D2
	.dc.b	$1F,$21,$8A,$7C,$75,$F8,$F9,$5F,$E9,$58,$D7,$C7,$FB,$63,$AD,$23
	.dc.b	$18,$78,$DB,$19,$9F,$F5,$4F,$C7,$77,$F8,$F9,$1F,$EA,$18,$F1,$7C
	.dc.b	$7F,$8F,$E3,$E4,$FF,$A9,$62,$BF,$EB,$08,$EA,$47,$FD,$6B,$CB,$EB
	.dc.b	$23,$FE,$AD,$8F,$77,$99,$23,$FD,$60,$FC,$1F,$87,$C4,$7C,$F3,$DB
	.dc.b	$E7,$7C,$53,$3C,$BF,$EA,$1D,$BE,$87,$81,$F2,$6A,$1F,$EB,$B6,$BC
	.dc.b	$E3,$EB,$77,$7C,$F3,$35,$F8,$3E,$A7,$BB,$F3,$6F,$FB,$C3,$63,$31
	.dc.b	$F4,$5A,$FE,$3D,$7B,$9B,$24,$6D,$E9,$D4,$F4,$BF,$0D,$3F,$92,$6C
	.dc.b	$66,$3F,$27,$D5,$FE,$04,$F7,$3F,$13,$F2,$0F,$87,$BF,$27,$2F,$67
	.dc.b	$E4,$27,$4E,$37,$43,$E1,$FF,$0E,$D9,$FD,$44,$EA,$72,$6B,$E1,$BC
	.dc.b	$3F,$11,$2B,$A5,$A7,$CB,$E9,$79,$BE,$22,$56,$A3,$D3,$B7,$80,$91
	.dc.b	$E0,$BF,$D9,$B7,$E1,$BD,$FF,$11,$23,$BC,$7C,$2F,$87,$F8,$7F,$11
	.dc.b	$27,$C7,$1F,$1B,$CC,$F6,$92,$3B,$3F,$03,$C9,$F9,$7E,$5C,$9F,$5C
	.dc.b	$7C,$EE,$9F,$8A,$48,$F3,$3F,$23,$A2,$47,$A5,$AF,$F1,$6E,$3F,$9F
	.dc.b	$CD,$F3,$89,$1E,$F7,$E8,$F9,$BF,$73,$E1,$E4,$FC,$C8,$FB,$FE,$7F
	.dc.b	$D6,$24,$7C,$77,$F7,$FF,$22,$F8,$6F,$E1,$AD,$9F,$C0,$35,$FD,$12
	.dc.b	$F8,$8D,$FC,$3F,$0F,$88,$95,$C2,$39,$FC,$47,$62,$47,$E1,$AD,$1E
	.dc.b	$2F,$53,$C2,$49,$F0,$FC,$47,$93,$E1,$FC,$BF,$11,$27,$CB,$1E,$8F
	.dc.b	$17,$D6,$48,$EC,$FB,$BE,$1F,$DA,$48,$F6,$FC,$47,$EA,$3E,$17,$F5
	.dc.b	$A4,$8F,$E4,$DE,$02,$3E,$3F,$C3,$C9,$F5,$35,$FB,$1B,$3E,$23,$E5
	.dc.b	$D4,$9F,$97,$15,$F9,$71,$7C,$CF,$E2,$95,$C6,$7E,$8F,$C3,$FD,$4F
	.dc.b	$88,$91,$83,$EA,$B8,$7E,$AB,$F6,$A3,$F8,$2F,$C0,$FE,$D6,$7F,$92
	.dc.b	$05,$E5,$68,$DF,$97,$67,$FA,$34,$E9,$B3,$FC,$76,$31,$1F,$D9,$DF
	.dc.b	$DD,$3C,$3C,$BD,$11,$FE,$A1,$F1,$FF,$C5,$5F,$1A,$E0,$FE,$36,$D8
	.dc.b	$FE,$26,$1E,$1A,$7F,$71,$75,$1F,$DD,$DA,$FF,$7E,$6B,$FE,$AA,$6B
	.dc.b	$FE,$B1,$F8,$F6,$9F,$EB,$28,$FF,$00,$46,$3B,$7F,$B3,$36,$33,$3B
	.dc.b	$7F,$C1,$3E,$3B,$E3,$FF,$9E,$36,$3F,$9E,$35,$DC,$68,$FF,$61,$7C
	.dc.b	$5F,$AB,$D3,$57,$18,$C3,$74,$FF,$5E,$B6,$FF,$C3,$7F,$0B,$FD,$3D
	.dc.b	$34,$8F,$F5,$36,$BF,$E0,$16,$BF,$E1,$F6,$BD,$0E,$04,$38,$14,$E0
	.dc.b	$53,$81,$A2,$1A,$29,$A3,$04,$30,$43,$FD,$C2,$F9,$1A,$77,$29,$D0
	.dc.b	$A7,$32,$9C,$CF,$F7,$83,$CD,$A7,$32,$9C,$CA,$73,$29,$CC,$F9,$E7
	.dc.b	$12,$9C,$4A,$71,$29,$C5,$F1,$38,$3C,$E4,$3B,$94,$D1,$82,$18,$23
	.dc.b	$D0,$8F,$95,$AF,$95,$AF,$95,$AF,$94,$E0,$43,$81,$4E,$05,$38,$1A
	.dc.b	$21,$A2,$9A,$30,$43,$04,$3E,$03,$E4,$69,$DC,$A7,$42,$9C,$CA,$73
	.dc.b	$3E,$3B,$CD,$A7,$32,$9C,$CA,$73,$29,$CC,$F9,$E7,$12,$9C,$4A,$71
	.dc.b	$29,$C5,$F1,$38,$3C,$E4,$3B,$94,$D1,$82,$18,$23,$D0,$8F,$95,$AF
	.dc.b	$95,$AF,$95,$AF,$94,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$3E,$03,$E4,$69,$DC,$A7,$42,$9C,$CA,$73,$3E,$3B,$CD,$A7
	.dc.b	$32,$9C,$CA,$73,$29,$CC,$F9,$E7,$13,$F7,$56,$F2,$DB,$7F,$9F,$DB
	.dc.b	$3F,$E7,$F6,$BF,$ED,$26,$A2,$4F,$39,$E4,$48,$E4,$D7,$93,$5E,$42
	.dc.b	$39,$3C,$08,$70,$29,$C1,$DB,$92,$19,$9F,$3B,$E6,$6B,$DD,$B1,$CA
	.dc.b	$38,$4F,$95,$E0,$D7,$83,$63,$84,$68,$F2,$94,$F1,$94,$EA,$68,$E6
	.dc.b	$57,$9B,$83,$89,$4E,$26,$1D,$92,$7D,$64,$9F,$33,$67,$B3,$5E,$4D
	.dc.b	$7A,$90,$F4,$10,$F1,$16,$7C,$13,$D8,$91,$D9,$AF,$66,$BD,$4E,$64
	.dc.b	$8E,$6F,$12,$1C,$4A,$71,$76,$E4,$86,$67,$D8,$F9,$DA,$F7,$6C,$72
	.dc.b	$8E,$13,$E6,$78,$35,$E0,$D8,$E1,$1A,$3C,$A5,$3C,$65,$3A,$9A,$39
	.dc.b	$95,$E6,$E0,$E2,$53,$89,$87,$64,$9F,$59,$27,$CC,$D9,$EC,$D7,$93
	.dc.b	$5E,$A4,$3D,$04,$3C,$45,$9F,$04,$F6,$24,$76,$6B,$D9,$AF,$53,$99
	.dc.b	$23,$9B,$C4,$87,$12,$9C,$5D,$B9,$21,$99,$F6,$3E,$76,$BD,$DB,$1C
	.dc.b	$A3,$84,$F9,$9E,$0D,$78,$36,$38,$46,$8F,$29,$4F,$19,$4E,$A6,$8E
	.dc.b	$65,$79,$B8,$38,$94,$E2,$61,$D9,$27,$D6,$49,$FF,$3D,$4B,$1F,$F3
	.dc.b	$BE,$5F,$E7,$CC,$BF,$CE,$ED,$9F,$E7,$79,$7F,$9E,$5F,$E6,$8F,$FA
	.dc.b	$15,$D9,$FE,$88,$8E,$AF,$FA,$20,$C1,$0C,$10,$E2,$53,$BB,$C9,$C9
	.dc.b	$0C,$9C,$CC,$1E,$47,$83,$A3,$A1,$93,$81,$93,$8E,$3F,$E9,$A8,$E3
	.dc.b	$1B,$38,$C6,$5E,$26,$4E,$33,$F0,$DE,$84,$34,$58,$E4,$F4,$8E,$0E
	.dc.b	$89,$1A,$7A,$CF,$92,$31,$1C,$E7,$8C,$62,$38,$C6,$09,$3C,$5B,$1E
	.dc.b	$A9,$E0,$48,$CC,$79,$67,$81,$1E,$11,$A7,$CA,$62,$34,$F1,$30,$43
	.dc.b	$04,$38,$94,$F6,$3C,$9C,$90,$C9,$CC,$C1,$F0,$1E,$0E,$8E,$86,$4E
	.dc.b	$06,$4E,$33,$F0,$63,$8C,$6C,$E3,$19,$78,$99,$38,$CF,$C3,$7A,$10
	.dc.b	$D1,$63,$93,$D2,$38,$3A,$24,$69,$EB,$3E,$48,$C4,$73,$9E,$31,$88
	.dc.b	$E3,$18,$24,$F1,$6C,$7A,$A7,$81,$23,$31,$E5,$9E,$04,$78,$46,$9F
	.dc.b	$29,$88,$D3,$C4,$C1,$0C,$10,$E2,$53,$D8,$F2,$72,$43,$27,$33,$07
	.dc.b	$C0,$78,$3A,$3A,$19,$38,$19,$38,$CF,$C1,$8E,$31,$B3,$8C,$65,$E2
	.dc.b	$64,$E3,$3F,$0D,$E8,$43,$45,$8E,$4F,$48,$E0,$E8,$91,$A7,$AC,$F9
	.dc.b	$23,$11,$CE,$78,$C6,$23,$8C,$60,$93,$C5,$B1,$EA,$9E,$04,$8C,$C7
	.dc.b	$96,$78,$11,$E1,$1A,$7C,$A6,$23,$4F,$13,$04,$30,$43,$89,$4F,$63
	.dc.b	$C9,$C9,$0C,$9C,$CC,$1F,$01,$E0,$E8,$E8,$64,$E0,$64,$E3,$3F,$06
	.dc.b	$38,$C6,$CE,$31,$97,$89,$93,$8C,$FC,$37,$A1,$0D,$16,$39,$3D,$23
	.dc.b	$83,$A2,$46,$9E,$B3,$E4,$8C,$47,$39,$E3,$18,$8E,$31,$82,$4F,$16
	.dc.b	$C7,$AA,$78,$12,$33,$1E,$59,$E0,$47,$84,$69,$F2,$98,$8D,$3C,$4F
	.dc.b	$DB,$7E,$56,$20

;--------------------------------------------------------------------------------
;	FDMOVE.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fdmove_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FDMOVE.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FDMOVE.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fdmove_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdmove.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDMOVE.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdmove.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDMOVE.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fdmove_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$BB,$F6,$7F,$79,$A5,$3F,$9A,$B5,$A5,$70,$9A,$47
	.dc.b	$44,$3A,$39,$68,$FA,$EF,$F3,$0F,$DC,$FF,$9E,$3E,$3F,$FA,$2C,$7F
	.dc.b	$99,$27,$C2,$EB,$3F,$F0,$9B,$98,$F1,$1C,$FD,$3F,$E7,$37,$CE,$30
	.dc.b	$7B,$A7,$95,$F6,$AF,$85,$85,$AF,$EA,$B7,$C6,$FB,$4E,$65,$39,$AE
	.dc.b	$31,$D8,$CE,$1E,$68,$CD,$F8,$27,$47,$B2,$3C,$F3,$B8,$F3,$46,$30
	.dc.b	$F0,$46,$CF,$2C,$ED,$FE,$AF,$87,$97,$0F,$24,$7D,$CF,$88,$F0,$1E
	.dc.b	$A7,$FA,$D6,$1D,$5A,$6D,$C4,$6C,$E2,$FB,$9B,$1B,$3C,$78,$7F,$7C
	.dc.b	$8E,$D8,$FB,$CB,$5D,$DB,$5C,$0F,$86,$7D,$A7,$84,$7C,$83,$81,$63
	.dc.b	$84,$72,$7E,$7D,$7C,$67,$94,$F0,$7E,$59,$C1,$E4,$F0,$3C,$38,$7C
	.dc.b	$9C,$7E,$64,$7C,$F3,$57,$DF,$1F,$9B,$1F,$40,$FC,$03,$87,$83,$5F
	.dc.b	$19,$E1,$3C,$67,$4B,$F1,$C6,$EF,$C5,$87,$AD,$B5,$E6,$3B,$E1,$E6
	.dc.b	$7A,$9C,$CE,$B5,$CD,$DD,$73,$69,$CF,$0F,$9E,$F0,$AE,$97,$E0,$9F
	.dc.b	$81,$18,$9E,$31,$8A,$E8,$F2,$8F,$88,$D7,$E1,$CF,$18,$C4,$7A,$2F
	.dc.b	$CD,$18,$BF,$06,$1F,$12,$F8,$3F,$6E,$FA,$B5,$EB,$87,$8A,$31,$1D
	.dc.b	$67,$8B,$F4,$4E,$25,$38,$9F,$0E,$38,$D7,$8A,$3E,$25,$75,$78,$1D
	.dc.b	$5C,$3F,$44,$EA,$53,$AD,$FC,$77,$B1,$CA,$7B,$3C,$30,$F8,$77,$E2
	.dc.b	$8F,$9E,$FC,$3B,$EF,$3A,$9E,$46,$8F,$39,$D8,$F1,$1E,$73,$0F,$17
	.dc.b	$35,$E2,$7C,$F5,$DD,$DE,$1F,$29,$D9,$DC,$F4,$39,$AE,$84,$70,$4B
	.dc.b	$F4,$B6,$7A,$C6,$08,$E0,$98,$7C,$92,$5F,$95,$B1,$F0,$1A,$F9,$27
	.dc.b	$9C,$62,$3D,$A6,$E7,$91,$1C,$10,$E4,$78,$9D,$38,$8E,$8F,$CD,$6E
	.dc.b	$1F,$2E,$31,$3C,$1F,$09,$C5,$C3,$A3,$C0,$68,$86,$4E,$06,$8C,$1A
	.dc.b	$8F,$01,$1F,$AC,$64,$A7,$02,$3B,$9F,$D9,$46,$A3,$13,$F9,$39,$C9
	.dc.b	$1C,$12,$7F,$27,$19,$6C,$65,$FC,$4C,$65,$B1,$9C,$3F,$11,$7F,$62
	.dc.b	$3E,$2D,$F8,$AB,$99,$0E,$25,$AE,$2F,$D7,$3E,$34,$71,$76,$48,$DD
	.dc.b	$FD,$88,$F8,$04,$7C,$D1,$AA,$F1,$CE,$5F,$46,$1F,$9B,$AE,$47,$07
	.dc.b	$11,$C2,$31,$87,$B6,$BC,$A7,$33,$D5,$5C,$1C,$CF,$07,$D8,$7E,$08
	.dc.b	$87,$DD,$74,$78,$AB,$CB,$5D,$E7,$67,$E0,$0F,$01,$F0,$EB,$C0,$EB
	.dc.b	$0F,$91,$87,$C1,$8F,$1D,$F4,$7C,$67,$85,$C6,$1E,$18,$F5,$E1,$E1
	.dc.b	$74,$60,$E2,$78,$E3,$D3,$3F,$1C,$E6,$58,$E1,$1D,$5F,$44,$F5,$79
	.dc.b	$4F,$07,$CE,$70,$79,$3C,$0F,$0D,$FB,$B0,$F8,$31,$E6,$35,$7D,$F0
	.dc.b	$F8,$51,$F1,$4F,$E4,$4F,$F9,$3C,$61,$FE,$2D,$6C,$7F,$34,$6B,$FE
	.dc.b	$11,$69,$C6,$3F,$98,$38,$AE,$2D,$8F,$F2,$A9,$A3,$FC,$75,$3F,$EB
	.dc.b	$88,$FF,$30,$36,$7B,$BA,$21,$FE,$31,$35,$3F,$E8,$19,$E3,$1A,$3F
	.dc.b	$C4,$C6,$8F,$F4,$43,$A7,$FC,$CF,$AF,$66,$CF,$38,$C4,$7F,$9A,$1B
	.dc.b	$26,$92,$78,$12,$7B,$C6,$5B,$1E,$7D,$79,$91,$C1,$23,$CF,$19,$6C
	.dc.b	$65,$F9,$11,$96,$C7,$C3,$D7,$CE,$47,$04,$8F,$87,$19,$6C,$65,$FA
	.dc.b	$91,$96,$C7,$CF,$F8,$4F,$86,$47,$04,$8F,$9F,$19,$6C,$65,$FC,$04
	.dc.b	$65,$B1,$FE,$BD,$FB,$87,$CC,$FF,$B8,$63,$F1,$32,$9C,$F8,$DF,$8E
	.dc.b	$94,$E6,$B8,$3F,$EA,$08,$CD,$70,$79,$BD,$BC,$7E,$53,$C6,$EC,$E6
	.dc.b	$B7,$69,$CB,$DB,$4F,$D3,$D7,$5B,$F0,$1E,$3C,$37,$7C,$1B,$87,$46
	.dc.b	$CF,$0A,$F0,$37,$0F,$05,$6C,$F1,$DF,$36,$99,$9E,$11,$E6,$F1,$E4
	.dc.b	$FA,$BC,$7B,$5D,$66,$E5,$EA,$C1,$58,$BF,$93,$0A,$C5,$FC,$FA,$B8
	.dc.b	$FD,$89,$B5,$CA,$FE,$4C,$60,$F9,$35,$63,$27,$94,$B7,$F0,$7C,$7F
	.dc.b	$9F,$E3,$FD,$06,$CF,$D4,$6C,$EE,$3E,$BC,$ED,$A6,$CB,$1B,$3E,$74
	.dc.b	$6D,$B1,$B6,$9B,$CB,$E3,$4E,$DB,$3B,$6C,$FD,$9F,$1F,$F0,$FE,$3F
	.dc.b	$E2,$1B,$5F,$8C,$6C,$EE,$3F,$1F,$3B,$69,$B2,$C6,$CF,$C3,$46,$DB
	.dc.b	$1B,$69,$BA,$FB,$D3,$B6,$CE,$DB,$3F,$91,$F1,$FF,$3E,$7E,$83,$C7
	.dc.b	$B5,$FA,$48,$B3,$F9,$79,$B5,$F8,$B9,$B5,$F8,$A9,$B5,$FB,$79,$BB
	.dc.b	$7F,$00,$9B,$B2,$6C,$61,$B7,$E1,$3F,$82,$DB,$EE,$7F,$81,$C7,$F0
	.dc.b	$48,$86,$4F,$E7,$B1,$A7,$26,$62,$B4,$CC,$6D,$E6,$D7,$B3,$5F,$E2
	.dc.b	$6D,$79,$1C,$0F,$09,$E2,$F7,$3F,$88,$BE,$96,$C7,$97,$2F,$EE,$2A
	.dc.b	$32,$4A,$F8,$31,$B7,$C7,$19,$AF,$E2,$67,$58,$DB,$9A,$EB,$3E,$39
	.dc.b	$F8,$35,$DD,$B1,$9A,$F8,$6D,$7E,$1B,$5F,$0C,$70,$23,$F2,$BD,$CF
	.dc.b	$A2,$48,$EE,$D8,$D4,$60,$8E,$08,$7D,$28,$E0,$FD,$4F,$73,$EE,$92
	.dc.b	$7D,$6D,$8F,$B6,$D7,$84,$68,$8E,$9A,$FE,$03,$DC,$FC,$49,$23,$E3
	.dc.b	$B6,$3F,$0E,$D7,$84,$68,$8E,$9A,$FE,$33,$B7,$F2,$A4,$8F,$A6,$D8
	.dc.b	$FC,$9B,$5E,$11,$A2,$3A,$6B,$FA,$3D,$BF,$B6,$F9,$9F,$D6,$BC,$CF
	.dc.b	$EB,$6D,$AF,$EB,$ED,$AF,$EA,$4D,$D7,$FA,$33,$67,$FA,$2B,$67,$FB
	.dc.b	$93,$6F,$BC,$FF,$6A,$BE,$ED,$3B,$96,$BB,$B6,$FB,$B4,$FD,$7F,$B3
	.dc.b	$FA,$63,$FB,$CC,$3F,$51,$7A,$79,$AF,$D2,$D5,$BF,$EE,$B7,$BA,$C3
	.dc.b	$FA,$E9,$B5,$FE,$0C,$8D,$3C,$89,$5F,$C1,$27,$F8,$59,$23,$F5,$2D
	.dc.b	$8D,$46,$08,$E0,$87,$F0,$C8,$E1,$3F,$E6,$AB,$7A,$7E,$4A,$4F,$BE
	.dc.b	$30,$DB,$F2,$BC,$63,$35,$CA,$72,$7B,$5C,$57,$F0,$6E,$16,$1F,$32
	.dc.b	$73,$87,$C5,$BF,$F1,$9B,$DE,$F8,$56,$F0,$EE,$4A,$EF,$3E,$8B,$E7
	.dc.b	$1F,$CA,$3D,$6F,$E3,$E7,$DA,$FD,$4D,$ED,$62,$7F,$8E,$CD,$BF,$B3
	.dc.b	$5A,$9C,$92,$3D,$53,$CE,$B8,$65,$FC,$AA,$2D,$7D,$DB,$E5,$7C,$23
	.dc.b	$95,$6D,$F7,$CE,$A3,$13,$B2,$3F,$EA,$08,$D4,$FB,$4F,$24,$FF,$48
	.dc.b	$AF,$EA,$24,$8F,$E5,$ED,$8F,$E9,$ED,$78,$46,$88,$E9,$B5,$E1,$F6
	.dc.b	$BD,$B3,$F8,$6B,$46,$25,$18,$AF,$79,$F0,$E7,$51,$9A,$F6,$CE,$4D
	.dc.b	$BD,$67,$64,$8C,$D7,$68,$C4,$F2,$38,$4E,$AB,$B5,$7C,$C3,$D5,$19
	.dc.b	$AE,$31,$98,$F1,$57,$38,$ED,$1F,$D8,$B2,$FE,$FA,$47,$FA,$DB,$67
	.dc.b	$FB,$DB,$5F,$EF,$6D,$7F,$BD,$B5,$FE,$F6,$FF,$82,$3D,$79,$3F,$D9
	.dc.b	$9B,$3F,$E0,$66,$BF,$E0,$66,$BF,$E0,$66,$BF,$E0,$66,$BF,$E1,$9F
	.dc.b	$5F,$FC,$2A,$49,$FE,$E6,$D8,$FF,$09,$B5,$FF,$09,$B5,$FF,$09,$B5
	.dc.b	$FF,$09,$BF,$ED,$CF,$8E,$FF,$7F,$CF,$F8,$79,$B3,$FE,$22,$6B,$FE
	.dc.b	$28,$6B,$FE,$2B,$6B,$FE,$B0,$6A,$69,$23,$C0,$43,$C0,$78,$FE,$4A
	.dc.b	$3E,$36,$CF,$8D,$AF,$8D,$AF,$8D,$AF,$8D,$AF,$8F,$E3,$7E,$01,$27
	.dc.b	$E0,$36,$3E,$03,$5F,$80,$D7,$E0,$35,$F8,$0F,$CB,$F1,$64,$FC,$B6
	.dc.b	$CF,$CB,$6B,$F2,$DA,$FC,$B6,$BF,$2D,$AF,$CB,$F1,$7E,$C1,$27,$EC
	.dc.b	$36,$3E,$C3,$5F,$B0,$D7,$EC,$35,$FB,$0F,$E0,$FC,$59,$3F,$83,$6C
	.dc.b	$FE,$0D,$AF,$E0,$DA,$FE,$0D,$AF,$E0,$DA,$BF,$3A,$78,$1F,$E5,$EC
	.dc.b	$DE,$B7,$F9,$82,$73,$87,$F9,$EA,$72,$FF,$9F,$23,$37,$FE,$5A,$9C
	.dc.b	$BF,$E5,$C8,$CF,$02,$BF,$2D,$7D,$70,$F0,$36,$3B,$38,$69,$DE,$31
	.dc.b	$7D,$E6,$D7,$58,$1E,$D9,$CD,$78,$0E,$71,$E5,$8C,$BF,$A7,$AA,$F1
	.dc.b	$9D,$D7,$58,$FD,$54,$EA,$FD,$B5,$D6,$7C,$8F,$03,$FD,$3D,$1A,$8E
	.dc.b	$D1,$96,$9E,$37,$0F,$66,$C7,$18,$C1,$D8,$95,$E5,$8D,$61,$F4,$A3
	.dc.b	$D0,$DA,$F3,$46,$F0,$F7,$CE,$6F,$C7,$39,$AF,$1C,$E6,$7C,$33,$97
	.dc.b	$C3,$3D,$08,$FD,$2B,$F1,$E1,$D8,$FB,$D1,$C0,$EF,$3C,$1F,$A0,$64
	.dc.b	$EE,$E6,$79,$BE,$49,$E7,$5E,$E7,$B4,$7E,$1C,$D1,$F2,$B0,$CD,$F5
	.dc.b	$BE,$6F,$7C,$3D,$D1,$8C,$3A,$5F,$13,$9D,$71,$BE,$73,$A8,$FB,$47
	.dc.b	$3C,$3F,$0D,$3A,$AF,$53,$67,$E0,$9F,$1D,$C3,$F0,$0C,$47,$82,$BB
	.dc.b	$1F,$44,$AF,$C9,$C3,$D3,$5C,$9B,$1C,$4F,$09,$6B,$C2,$7D,$0A,$E6
	.dc.b	$E2,$78,$BE,$B9,$F8,$4D,$3E,$13,$F1,$1F,$EA,$97,$FD,$6C,$91,$FD
	.dc.b	$6D,$B1,$FD,$6D,$AF,$08,$D1,$1D,$36,$BF,$65,$D4,$F2,$3F,$6E,$D9
	.dc.b	$FD,$4B,$6B,$F1,$4D,$AF,$6C,$23,$13,$F9,$16,$C7,$C4,$6B,$F4,$CE
	.dc.b	$15,$E0,$24,$F8,$1B,$1E,$06,$D7,$82,$B9,$E1,$FA,$B6,$E1,$FC,$66
	.dc.b	$7D,$B5,$C3,$0F,$23,$70,$F2,$35,$F2,$5F,$29,$FF,$52,$1F,$EC,$B9
	.dc.b	$D6,$1D,$DB,$87,$79,$D1,$E9,$25,$FA,$5B,$87,$A6,$B8,$C7,$69,$FE
	.dc.b	$5F,$87,$85,$FF,$50,$C7,$46,$9A,$9F,$03,$63,$A9,$BA,$F8,$2E,$23
	.dc.b	$47,$23,$E8,$B9,$69,$9A,$F1,$4F,$28,$C5,$78,$1D,$C7,$AE,$BD,$51
	.dc.b	$87,$C4,$7D,$03,$35,$F8,$23,$AB,$A8,$F2,$CF,$73,$A3,$E0,$8E,$35
	.dc.b	$F5,$EB,$93,$4F,$23,$D2,$79,$CF,$57,$C8,$7E,$01,$E6,$47,$EB,$1D
	.dc.b	$C9,$5E,$23,$C9,$1C,$F0,$FC,$13,$C6,$7F,$B3,$9F,$38,$A7,$CC,$2D
	.dc.b	$7E,$AD,$B7,$FD,$8D,$FB,$C6,$27,$8B,$CC,$C9,$0C,$10,$F6,$9F,$E1
	.dc.b	$3E,$CF,$87,$3F,$E1,$16,$CF,$F8,$3D,$B3,$FE,$0D,$6E,$1F,$D8,$5B
	.dc.b	$3F,$DC,$1F,$9B,$F1,$72,$3E,$FF,$E8,$6F,$73,$18,$7E,$26,$AE,$1F
	.dc.b	$DF,$EB,$57,$F8,$29,$CB,$F2,$32,$47,$E3,$67,$77,$CE,$B8,$1F,$11
	.dc.b	$FE,$21,$3F,$25,$FD,$4B,$5C,$C6,$EB,$AB,$67,$51,$89,$F9,$ED,$7E
	.dc.b	$BC,$68,$93,$C4,$91,$FA,$BF,$BD,$78,$7E,$C8,$FF,$14,$C6,$27,$D3
	.dc.b	$1B,$9D,$29,$D3,$EF,$35,$3C,$31,$FF,$14,$4E,$9F,$E7,$37,$FB,$38
	.dc.b	$C4,$FC,$B8,$C5,$F6,$9C,$DF,$C6,$AF,$DB,$C6,$23,$E3,$9E,$03,$C5
	.dc.b	$7E,$0A,$E4,$4A,$F2,$1F,$A0,$AE,$58,$7C,$BA,$E1,$3A,$BF,$87,$7C
	.dc.b	$5E,$B1,$C6,$31,$3E,$C8,$F2,$C6,$1F,$61,$83,$A4,$FC,$98,$D3,$ED
	.dc.b	$3D,$53,$CD,$B3,$C9,$A6,$E3,$CE,$F9,$0F,$3C,$77,$76,$7C,$68,$D9
	.dc.b	$23,$D4,$79,$2F,$E6,$12,$BC,$B5,$E7,$9E,$95,$E5,$39,$47,$80,$EF
	.dc.b	$7F,$2E,$7A,$C7,$8A,$FE,$7E,$1F,$7A,$79,$9A,$BE,$8D,$25,$75,$8C
	.dc.b	$B6,$32,$F9,$4F,$84,$58,$F2,$B6,$3E,$54,$62,$BD,$D1,$F0,$70,$F2
	.dc.b	$57,$AA,$ED,$EE,$7E,$1D,$F0,$AE,$ED,$9E,$F8,$7C,$19,$F3,$10,$F1
	.dc.b	$1F,$67,$0F,$4D,$71,$7C,$A6,$2F,$C2,$E2,$79,$4F,$EC,$1F,$E6,$ED
	.dc.b	$7E,$79,$89,$FE,$19,$19,$8F,$AC,$66,$7F,$8A,$46,$63,$E5,$19,$3F
	.dc.b	$9D,$B9,$24,$FE,$E7,$BB,$CB,$3F,$C2,$1A,$FF,$05,$6C,$9A,$4A,$FC
	.dc.b	$0C,$F1,$6B,$C5,$B3,$C5,$B3,$C6,$7F,$18,$DC,$3E,$E3,$67,$E6,$B6
	.dc.b	$7E,$4B,$E9,$8F,$E1,$6E,$63,$C0,$7F,$25,$71,$5E,$18,$C4,$F8,$8E
	.dc.b	$24,$7C,$6D,$AF,$E4,$35,$C8,$F0,$4E,$8E,$67,$CE,$7E,$B9,$27,$EA
	.dc.b	$B6,$35,$18,$23,$82,$1F,$84,$8E,$13,$FC,$CB,$E0,$3C,$2D,$AF,$31
	.dc.b	$FC,$8E,$32,$D3,$35,$FD,$12,$32,$D8,$CB,$FD,$0E,$32,$D8,$CD,$A7
	.dc.b	$2D,$8C,$C6,$0F,$E8,$84,$72,$7F,$40,$71,$3F,$22,$BD,$4F,$F5,$B8
	.dc.b	$CC,$7D,$09,$E0,$49,$D3,$D1,$A4,$8E,$0D,$8F,$84,$D3,$F2,$C5,$3D
	.dc.b	$C5,$AF,$C3,$4F,$90,$93,$C9,$EA,$62,$7E,$51,$C6,$FD,$33,$E0,$8F
	.dc.b	$DE,$DF,$EE,$89,$1F,$B8,$6C,$7E,$D9,$AF,$08,$D1,$1D,$36,$8F,$F9
	.dc.b	$0B,$C2,$7A,$FE,$2A,$08,$CC,$FC,$BA,$E3,$3A,$9C,$BF,$15,$B1,$EF
	.dc.b	$8E,$55,$D5,$B5,$FE,$8B,$CD,$E4,$FF,$8A,$AF,$AC,$63,$7F,$F4,$4B
	.dc.b	$AF,$7A,$D6,$EF,$FD,$26,$DC,$3F,$D7,$CD,$AE,$13,$DE,$30,$4B,$E7
	.dc.b	$7D,$19,$3B,$72,$D2,$5F,$FA,$7A,$FC,$D1,$D6,$B8,$CE,$6F,$B4,$E4
	.dc.b	$E3,$87,$F8,$9B,$B7,$F9,$B9,$23,$F9,$AB,$6B,$F9,$9B,$5F,$E8,$2D
	.dc.b	$7F,$C3,$8D,$7F,$C7,$6F,$F4,$2E,$D9,$3F,$D0,$1B,$3F,$CF,$9A,$FF
	.dc.b	$4C,$6B,$FE,$27,$6B,$FE,$46,$6A,$FB,$DF,$C5,$CF,$F9,$7D,$B1,$FE
	.dc.b	$84,$6B,$FE,$8E,$6B,$FE,$92,$6B,$F9,$06,$BD,$63,$FC,$CD,$18,$23
	.dc.b	$82,$1E,$1E,$ED,$91,$C1,$27,$C9,$19,$6C,$65,$F3,$46,$5B,$1E,$DE
	.dc.b	$EF,$19,$1C,$12,$3E,$04,$65,$B1,$97,$E1,$46,$5B,$1F,$27,$BB,$DE
	.dc.b	$47,$04,$8F,$99,$19,$6C,$65,$F9,$D1,$96,$C7,$D5,$EE,$F9,$64,$70
	.dc.b	$48,$FB,$11,$96,$C6,$5F,$B5,$19,$6C,$7E,$07,$BB,$EB,$91,$C1,$23
	.dc.b	$F0,$91,$96,$C6,$5F,$C3,$46,$5B,$1F,$8D,$F4,$3F,$06,$47,$04,$8F
	.dc.b	$C8,$46,$5B,$19,$7F,$25,$19,$6C,$7E,$67,$D0,$FC,$79,$1C,$12,$3F
	.dc.b	$39,$19,$6C,$65,$FC,$F4,$65,$B1,$FA,$5F,$43,$F3,$64,$70,$48,$FD
	.dc.b	$44,$65,$B1,$97,$F5,$51,$96,$C7,$EC,$7D,$0F,$D3,$91,$C1,$23,$F6
	.dc.b	$91,$96,$C6,$5F,$DB,$46,$5B,$1F,$BD,$F4,$3F,$66,$47,$04,$8F,$E0
	.dc.b	$11,$96,$C6,$5F,$E0,$51,$96,$C7,$F0,$9F,$43,$F7,$E4,$70,$48,$FE
	.dc.b	$19,$19,$6C,$65,$FE,$1D,$19,$6C,$7F,$15,$F0,$FF,$85,$91,$C1,$23
	.dc.b	$F8,$C4,$65,$B1,$97,$F8,$D4,$65,$B1,$FC,$87,$C0,$FE,$2E,$47,$04
	.dc.b	$8F,$E4,$91,$96,$C6,$5F,$E4,$D1,$96,$C7,$F2,$DE,$AF,$E4,$64,$70
	.dc.b	$48,$FE,$61,$19,$6C,$65,$FE,$65,$19,$6C,$7F,$39,$E9,$FE,$5E,$47
	.dc.b	$04,$8F,$E7,$91,$96,$C6,$5F,$E7,$D1,$96,$C7,$F4,$5E,$8F,$E7,$64
	.dc.b	$70,$48,$FE,$91,$19,$6C,$65,$FE,$95,$19,$6C,$7F,$51,$F7,$BF,$C2
	.dc.b	$BD,$1F,$E1,$69,$FE,$AD,$18,$9F,$F0,$EC,$BE,$31,$F0,$99,$BF,$EA
	.dc.b	$57,$5E,$37,$C0,$FE,$B1,$1B,$BE,$55,$BE,$84,$D2,$57,$1B,$5F,$08
	.dc.b	$C4,$76,$32,$F6,$3D,$84,$7F,$B6,$4F,$F6,$42,$46,$63,$FB,$74,$61
	.dc.b	$DC,$65,$E8,$60,$EE,$79,$30,$79,$9E,$88,$CD,$FC,$78,$C3,$F3,$A7
	.dc.b	$75,$EA,$C3,$E0,$BE,$F9,$E0,$E7,$0F,$91,$5C,$30,$F0,$9F,$06,$BA
	.dc.b	$61,$F4,$6D,$E3,$53,$FC,$83,$F0,$36,$3E,$49,$A2,$C6,$9A,$4B,$EE
	.dc.b	$ED,$FF,$1B,$4F,$37,$33,$F8,$97,$FC,$C6,$7F,$9A,$1C,$3F,$E6,$69
	.dc.b	$E4,$E8,$D3,$5D,$47,$07,$26,$63,$8B,$D9,$AF,$07,$93,$C0,$EC,$E9
	.dc.b	$C1,$C6,$3F,$CC,$B8,$FA,$F0,$FF,$35,$F7,$7D,$4A,$F6,$DF,$F9,$8E
	.dc.b	$B4,$FF,$9A,$9B,$1F,$E6,$A6,$F7,$FF,$9D,$9F,$A9,$EF,$4A,$FB,$DA
	.dc.b	$3E,$0F,$18,$7C,$EA,$B9,$7E,$1E,$6D,$FB,$E6,$D7,$E4,$66,$BF,$E9
	.dc.b	$68,$F9,$12,$69,$1F,$8A,$64,$AE,$63,$E7,$35,$F9,$2D,$7C,$51,$C0
	.dc.b	$F5,$FA,$5E,$4B,$FE,$0C,$DC,$0D,$24,$FE,$4A,$FF,$3B,$1F,$99,$F6
	.dc.b	$A4,$FF,$A4,$7C,$CB,$8F,$E4,$A6,$E3,$F5,$A6,$D7,$E0,$E6,$D7,$FA
	.dc.b	$72,$6B,$FE,$BA,$8F,$8A,$49,$F8,$6D,$8F,$C7,$35,$FC,$4B,$5E,$D1
	.dc.b	$C0,$8F,$07,$FD,$51,$F0,$5E,$2D,$7F,$D6,$ED,$D8,$D2,$4F,$ED,$2F
	.dc.b	$FD,$6F,$1F,$BB,$F6,$64,$FF,$05,$FD,$59,$E2,$E3,$2F,$DA,$D5,$D3
	.dc.b	$F9,$DD,$6A,$FF,$7F,$39,$7F,$50,$FF,$33,$8D,$57,$29,$D9,$F8,$69
	.dc.b	$FC,$59,$1C,$12,$3F,$69,$19,$6C,$65,$FD,$8C,$65,$B1,$EB,$F1,$7C
	.dc.b	$91,$FC,$FA,$91,$8B,$46,$2B,$CB,$5F,$CB,$A3,$0F,$A4,$C5,$71,$8C
	.dc.b	$D7,$F5,$DA,$7C,$87,$F2,$73,$FD,$CF,$F4,$3F,$C7,$67,$51,$99,$FE
	.dc.b	$4E,$EB,$97,$18,$FF,$24,$8B,$5F,$EF,$2B,$E5,$7F,$C5,$67,$75,$FC
	.dc.b	$49,$FD,$A9,$23,$F6,$6D,$8F,$E3,$CD,$78,$46,$88,$E9,$B5,$E1,$F8
	.dc.b	$1F,$1C,$25,$C0,$C1,$5A,$9C,$C6,$2B,$DE,$7F,$71,$AE,$15,$B6,$C6
	.dc.b	$AB,$FD,$F9,$5C,$5F,$E8,$FF,$21,$FD,$4C,$FC,$77,$8B,$6B,$FA,$8C
	.dc.b	$59,$FB,$53,$71,$FA,$D3,$6B,$FC,$23,$A2,$B1,$7F,$D2,$EB,$F8,$F9
	.dc.b	$2B,$F8,$EB,$63,$FA,$43,$5F,$E8,$6D,$7F,$A5,$35,$E9,$1C,$8F,$D8
	.dc.b	$78,$BE,$0C,$3F,$26,$DC,$0D,$24,$FB,$8E,$27,$F7,$0F,$8F,$FE,$EC
	.dc.b	$7E,$43,$E1,$AC,$64,$98,$7C,$C8,$CB,$FD,$CA,$6E,$1F,$8F,$9B,$5F
	.dc.b	$D8,$E7,$FA,$09,$1F,$E7,$ED,$AF,$ED,$4D,$7F,$B2,$B5,$E5,$1C,$08
	.dc.b	$F0,$7F,$09,$9F,$F4,$98,$F2,$46,$3A,$3C,$71,$F8,$C6,$EB,$F9,$16
	.dc.b	$D7,$1C,$7F,$69,$87,$EC,$A7,$FC,$95,$F7,$DF,$E0,$7C,$3F,$C3,$ED
	.dc.b	$AF,$E3,$BE,$26,$A6,$65,$FE,$44,$AD,$3F,$88,$9C,$CF,$22,$D7,$29
	.dc.b	$E1,$1A,$3F,$C2,$95,$E1,$69,$FE,$13,$70,$FF,$87,$9A,$FE,$BD,$A9
	.dc.b	$A4,$79,$12,$3F,$C2,$F1,$97,$F9,$F7,$89,$E6,$7F,$42,$D9,$FE,$4C
	.dc.b	$DA,$FE,$4A,$DD,$8D,$3E,$5F,$C0,$4B,$FF,$27,$78,$8F,$37,$18,$FC
	.dc.b	$EA,$B8,$FF,$A1,$2B,$57,$F3,$67,$2F,$C1,$7F,$CF,$91,$AA,$E5,$3B
	.dc.b	$3D,$B3,$F2,$48,$E0,$91,$FE,$41,$C9,$3A,$8C,$35,$FA,$AD,$7C,$53
	.dc.b	$EC,$F3,$7C,$B1,$FE,$88,$B4,$63,$04,$62,$BC,$D5,$FE,$78,$8C,$3E
	.dc.b	$A3,$15,$C6,$33,$5F,$EA,$7C,$1E,$FF,$E6,$0F,$E9,$7E,$8D,$C7,$F2
	.dc.b	$91,$71,$FE,$4F,$36,$F9,$4D,$AF,$E7,$1C,$5F,$CD,$EB,$17,$F8,$DA
	.dc.b	$FC,$91,$2B,$F1,$2D,$8F,$94,$D7,$F3,$6D,$7F,$30,$D7,$A4,$72,$3F
	.dc.b	$7B,$F0,$FE,$0B,$FE,$90,$DE,$33,$49,$3E,$E3,$89,$FB,$3F,$8B,$FD
	.dc.b	$D9,$F8,$6F,$3E,$C6,$49,$87,$D1,$8C,$BF,$C0,$26,$E7,$F9,$39,$B5
	.dc.b	$FB,$19,$FD,$A1,$1F,$D5,$B6,$BF,$18,$D7,$F7,$AD,$79,$47,$02,$3C
	.dc.b	$1F,$D3,$F9,$FE,$08,$FC,$63,$6F,$F1,$6D,$D8,$D2,$57,$F1,$5F,$2B
	.dc.b	$F8,$E9,$FE,$10,$F0,$EC,$64,$98,$7F,$6C,$8C,$BF,$C8,$A6,$E1,$FE
	.dc.b	$0F,$9B,$5F,$C4,$E7,$F8,$B1,$1F,$E1,$AD,$AF,$D7,$35,$FE,$3E,$D7
	.dc.b	$94,$70,$23,$C1,$FE,$83,$F0,$BE,$08,$FF,$04,$B7,$6F,$F0,$73,$76
	.dc.b	$34,$95,$FC,$DF,$E0,$3F,$9F,$9F,$C2,$BC,$FB,$19,$26,$1F,$B5,$8C
	.dc.b	$BF,$D1,$26,$E7,$FC,$5E,$6D,$7F,$35,$9F,$E7,$04,$7F,$97,$B6,$BF
	.dc.b	$88,$35,$FE,$82,$D7,$94,$70,$23,$C1,$FE,$57,$AF,$82,$3F,$88,$36
	.dc.b	$FF,$87,$B4,$FE,$23,$3E,$C8,$FE,$5D,$19,$3F,$B0,$F9,$32,$7F,$B7
	.dc.b	$FC,$2D,$BF,$EC,$73,$6F,$FC,$A1,$36,$BF,$AA,$4D,$AF,$F1,$EC,$D7
	.dc.b	$FA,$A4,$7F,$5B,$24,$FF,$52,$6C,$7F,$33,$6B,$FD,$91,$AF,$68,$E0
	.dc.b	$47,$83,$FE,$06,$F1,$BC,$5A,$FF,$94,$5B,$A9,$A4,$9F,$EF,$57,$FE
	.dc.b	$74,$8F,$F0,$1F,$AB,$27,$FC,$B9,$91,$BF,$AC,$FF,$9C,$32,$FE,$E5
	.dc.b	$5B,$7F,$C1,$2E,$D5,$51,$3C,$63,$FB,$C4,$F1,$9F,$E7,$CE,$EB,$BF
	.dc.b	$8F,$97,$FC,$05,$1F,$E0,$52,$3F,$DF,$1B,$5F,$D7,$5A,$FF,$83,$9A
	.dc.b	$F2,$8E,$04,$78,$3F,$00,$F7,$F8,$DE,$18,$F5,$E1,$FE,$71,$C3,$CF
	.dc.b	$5B,$9C,$D6,$E7,$DE,$DA,$E1,$5E,$1A,$F8,$AF,$63,$92,$F8,$8F,$F5
	.dc.b	$5C,$7F,$A3,$63,$1E,$4F,$FA,$AB,$DF,$F8,$D1,$FE,$91,$AD,$B6,$72
	.dc.b	$F0,$AF,$F2,$5D,$62,$B7,$39,$3C,$67,$19,$D9,$E4,$8D,$9F,$EC,$28
	.dc.b	$F2,$1C,$4A,$E5,$F4,$35,$E8,$70,$38,$1A,$AF,$C4,$3C,$9A,$E6,$78
	.dc.b	$3C,$0E,$66,$DC,$CF,$C2,$DF,$FD,$65,$F2,$9E,$FB,$FF,$68,$B6,$BF
	.dc.b	$CD,$AD,$BF,$F5,$6B,$73,$FF,$67,$36,$7F,$C0,$FC,$08,$F9,$5E,$CC
	.dc.b	$9F,$C1,$57,$E7,$BA,$31,$87,$CC,$3E,$E5,$62,$7F,$49,$AA,$B1,$7F
	.dc.b	$22,$AD,$FE,$42,$6B,$CA,$7D,$F3,$F1,$A3,$F0,$4D,$74,$7C,$FD,$11
	.dc.b	$96,$CF,$22,$4E,$E7,$44,$8C,$C7,$DE,$8F,$C8,$E1,$FB,$CE,$7F,$3B
	.dc.b	$6B,$F1,$13,$AC,$7F,$AA,$C6,$0F,$8E,$E2,$FE,$36,$1E,$47,$F2,$77
	.dc.b	$CD,$CD,$F3,$C3,$F3,$5E,$5F,$AC,$98,$FE,$09,$B5,$F0,$5A,$FE,$59
	.dc.b	$AF,$E5,$9A,$FE,$AD,$F8,$DE,$54,$9F,$C6,$B6,$7E,$63,$5F,$D0,$B5
	.dc.b	$FD,$0B,$5F,$DB,$35,$FD,$F7,$6F,$D3,$24,$FE,$69,$B1,$F6,$5A,$FE
	.dc.b	$B9,$AF,$EB,$9A,$FF,$03,$7E,$F7,$64,$9F,$D2,$B6,$7F,$08,$D7,$F7
	.dc.b	$2D,$7F,$72,$D7,$F8,$73,$5F,$E2,$DE,$C7,$E2,$C9,$3F,$B2,$6C,$7E
	.dc.b	$45,$AF,$F0,$66,$BF,$C1,$9A,$FF,$1B,$7F,$2D,$EC,$49,$FD,$EB,$67
	.dc.b	$F3,$8D,$7F,$88,$B5,$FE,$22,$D7,$F9,$33,$5F,$E5,$DE,$3F,$F2,$33
	.dc.b	$F3,$7F,$AD,$BD,$7C,$5F,$F2,$19,$B5,$F9,$4A,$D4,$E4,$91,$F9,$A9
	.dc.b	$E7,$5C,$37,$FE,$3B,$C1,$FE,$C8,$BF,$E2,$F5,$FB,$32,$38,$25,$FE
	.dc.b	$E6,$32,$D8,$CB,$FC,$C2,$32,$D8,$FE,$27,$83,$D7,$F1,$47,$F7,$C8
	.dc.b	$C3,$E8,$C3,$FB,$B4,$62,$BD,$53,$CA,$B8,$C7,$F7,$C9,$D5,$7B,$2F
	.dc.b	$E4,$D7,$F4,$DC,$66,$1F,$D9,$79,$DC,$CF,$47,$39,$FF,$00,$BB,$97
	.dc.b	$F8,$26,$F7,$58,$C3,$FA,$5D,$5B,$56,$A7,$FA,$B4,$F4,$9F,$EA,$6F
	.dc.b	$F1,$F2,$47,$F4,$66,$C7,$F2,$56,$BC,$23,$44,$74,$DB,$F3,$60,$F8
	.dc.b	$0F,$26,$1E,$E6,$C6,$6B,$CA,$7C,$37,$17,$F0,$63,$13,$C0,$FE,$ED
	.dc.b	$EA,$48,$FE,$9B,$EC,$3D,$3C,$5F,$F2,$8A,$B9,$7F,$8B,$EB,$57,$FD
	.dc.b	$E6,$72,$FF,$74,$7F,$9C,$C6,$AB,$94,$EC,$FE,$D7,$3F,$D1,$C8,$E0
	.dc.b	$91,$FD,$36,$32,$D8,$CB,$FE,$04,$8C,$B6,$3D,$7E,$97,$92,$3F,$C8
	.dc.b	$F4,$8C,$5A,$31,$5E,$5A,$FE,$8B,$18,$7D,$26,$2B,$8C,$66,$BF,$C1
	.dc.b	$BF,$80,$7B,$DB,$8F,$F8,$AB,$4F,$CA,$FC,$B7,$F8,$A2,$B5,$97,$F8
	.dc.b	$C5,$B9,$03,$8A,$E3,$3F,$E3,$A8,$DC,$EA,$FA,$1D,$F5,$61,$8B,$E1
	.dc.b	$3F,$DE,$6F,$EA,$5F,$F8,$B9,$AF,$8E,$34,$49,$FF,$12,$CE,$A7,$CC
	.dc.b	$D7,$C5,$1A,$9E,$11,$FE,$33,$9D,$1E,$BB,$7A,$DE,$59,$FF,$22,$E3
	.dc.b	$FB,$A8,$C6,$5F,$A2,$6E,$3F,$E2,$66,$D7,$ED,$AF,$9C,$FB,$8F,$D1
	.dc.b	$4F,$CF,$F3,$64,$7E,$7B,$C8,$B7,$FE,$29,$C9,$58,$C7,$F3,$D5,$6F
	.dc.b	$EB,$4D,$BF,$C5,$CD,$7E,$BC,$7C,$D2,$4F,$C8,$6C,$7C,$36,$BF,$45
	.dc.b	$AF,$68,$E0,$47,$83,$FB,$6F,$1F,$C5,$7F,$99,$6E,$DF,$EB,$39,$FC
	.dc.b	$35,$FE,$CA,$72,$48,$FC,$56,$1F,$93,$3F,$11,$ED,$58,$C9,$31,$FA
	.dc.b	$BA,$27,$51,$F2,$6A,$E5,$FC,$62,$6D,$FE,$76,$7F,$20,$47,$F1,$6D
	.dc.b	$AF,$C3,$35,$FC,$A3,$5E,$51,$C0,$8F,$07,$F9,$07,$CB,$78,$23,$F0
	.dc.b	$8D,$DB,$FA,$2B,$71,$34,$95,$FA,$DF,$13,$F6,$C7,$F1,$AE,$1B,$19
	.dc.b	$26,$1F,$C1,$A3,$2F,$E2,$E6,$E7,$FC,$9E,$6D,$7E,$FE,$7F,$64,$47
	.dc.b	$F5,$AD,$AF,$D4,$35,$FD,$B3,$5E,$51,$C0,$8F,$07,$F9,$DF,$07,$82
	.dc.b	$3F,$90,$B6,$FF,$90,$37,$7F,$D4,$CF,$A6,$B8,$9E,$EA,$FE,$81,$5F
	.dc.b	$C6,$70,$93,$FD,$3F,$D8,$B7,$FA,$B9,$B7,$FD,$6A,$6D,$7F,$1F,$9B
	.dc.b	$5F,$B9,$9A,$FF,$20,$8F,$E2,$C4,$9F,$E2,$2D,$8F,$E1,$8D,$7F,$8D
	.dc.b	$35,$ED,$1C,$08,$F0,$7F,$B3,$7C,$B7,$8B,$5F,$F0,$4B,$75,$34,$93
	.dc.b	$FC,$C6,$FF,$97,$47,$F3,$BD,$64,$FF,$6A,$F9,$6B,$8F,$F0,$D9,$B9
	.dc.b	$FF,$69,$9B,$5F,$D0,$66,$D7,$F5,$69,$AF,$F4,$28,$FE,$70,$49,$FE
	.dc.b	$66,$D8,$FE,$5C,$D7,$F9,$E3,$5E,$D1,$C0,$8F,$07,$FD,$9B,$FA,$E7
	.dc.b	$F5,$17,$FA,$74,$FF,$AC,$A3,$13,$EB,$9D,$BE,$8B,$4E,$9F,$EC,$75
	.dc.b	$C6,$75,$39,$AF,$EC,$17,$FD,$8C,$FF,$5F,$FE,$BC,$F2,$F1,$3F,$D7
	.dc.b	$E6,$DF,$FA,$E2,$B5,$39,$24,$7F,$3B,$9E,$75,$C3,$1F,$EE,$B1,$FD
	.dc.b	$84,$95,$FD,$71,$B5,$FD,$5D,$AF,$F6,$56,$BE,$88,$E0,$47,$83,$FE
	.dc.b	$F6,$BF,$F7,$D3,$E4,$A3,$FD,$F5,$46,$A7,$D1,$3C,$78,$7D,$0E,$E7
	.dc.b	$55,$EA,$8D,$1D,$4D,$4F,$6A,$4D,$5F,$30,$F9,$1D,$4E,$63,$17,$FE
	.dc.b	$49,$C7,$2F,$F8,$4E,$F8,$DE,$1F,$F0,$B5,$70,$7B,$CE,$E3,$17,$FE
	.dc.b	$5A,$BE,$15,$CA,$BE,$11,$BB,$CD,$F0,$8F,$F1,$17,$07,$F9,$F2,$F3
	.dc.b	$52,$7C,$97,$C5,$AF,$92,$3E,$03,$93,$37,$E7,$78,$B5,$D1,$88,$E2
	.dc.b	$F3,$9E,$AF,$A4,$E7,$1A,$70,$71,$8F,$F4,$A7,$B1,$F9,$F8,$E4,$7F
	.dc.b	$87,$9C,$3C,$F1,$FF,$4A,$46,$2F,$E1,$4F,$FA,$4A,$31,$7F,$E4,$B6
	.dc.b	$DF,$69,$F9,$3F,$23,$23,$F2,$5C,$F6,$7E,$A9,$F6,$27,$18,$FE,$0B
	.dc.b	$5F,$EF,$97,$F5,$2B,$4D,$7F,$13,$5C,$23,$F2,$97,$67,$67,$CE,$9F
	.dc.b	$89,$91,$A4,$AF,$97,$19,$6C,$66,$79,$46,$08,$E0,$98,$7A,$78,$3C
	.dc.b	$4F,$EB,$1B,$8F,$F7,$EA,$F5,$61,$DE,$FF,$85,$56,$89,$3A,$F9,$7F
	.dc.b	$C9,$9F,$3E,$AC,$64,$95,$F6,$E3,$2F,$CE,$3F,$5F,$38,$C3,$EE,$CD
	.dc.b	$AF,$D3,$4F,$C6,$23,$F1,$1B,$5F,$98,$6B,$F9,$86,$BC,$A3,$81,$1E
	.dc.b	$0F,$E4,$B4,$F0,$47,$E2,$DB,$B3,$5F,$C7,$CF,$A2,$B8,$1E,$EC,$7F
	.dc.b	$2B,$7C,$2B,$78,$BE,$17,$F7,$27,$F0,$5A,$FF,$16,$C6,$2F,$E8,$CD
	.dc.b	$C3,$F2,$75,$A9,$C9,$23,$E5,$57,$58,$FC,$A4,$EE,$73,$5C,$DF,$DD
	.dc.b	$BF,$90,$24,$FE,$39,$B1,$FC,$09,$AF,$08,$D1,$1D,$36,$7F,$93,$E7
	.dc.b	$F0,$2B,$DC,$D9,$F2,$B6,$BD,$AD,$AE,$38,$FC,$16,$D7,$E5,$3E,$5F
	.dc.b	$F9,$09,$FB,$DF,$7E,$C6,$49,$87,$F0,$98,$CB,$F1,$A6,$D7,$F0,$A9
	.dc.b	$B5,$FC,$CE,$7F,$66,$47,$F6,$2D,$AF,$E4,$AD,$7F,$92,$B5,$E5,$1C
	.dc.b	$08,$F0,$7F,$8F,$FC,$67,$82,$3F,$8C,$37,$5F,$C7,$37,$53,$49,$5F
	.dc.b	$B6,$EB,$FE,$86,$7C,$EF,$98,$B1,$92,$61,$F3,$A3,$2F,$E2,$26,$E7
	.dc.b	$F6,$26,$D7,$F5,$59,$FE,$2A,$47,$F8,$9B,$6B,$FA,$3B,$5F,$E8,$ED
	.dc.b	$79,$47,$02,$3C,$1F,$C1,$F0,$78,$23,$FB,$9E,$C8,$C6,$1F,$E9,$26
	.dc.b	$C6,$6B,$DB,$3F,$90,$9C,$9F,$D9,$BE,$2A,$4F,$F4,$EE,$3B,$87,$EA
	.dc.b	$26,$DF,$F4,$C9,$B5,$FD,$D6,$6D,$7F,$4D,$9A,$FF,$81,$A3,$F9,$C9
	.dc.b	$27,$F9,$BB,$63,$FB,$5B,$5F,$ED,$6D,$7B,$47,$02,$3C,$1F,$EC,$BF
	.dc.b	$01,$E2,$DB,$F8,$93,$76,$34,$93,$FC,$EA,$FF,$B1,$47,$F8,$33,$F0
	.dc.b	$27,$C3,$D3,$F6,$BF,$25,$6F,$FA,$D4,$E9,$B3,$A6,$93,$2F,$E7,$53
	.dc.b	$B7,$F7,$75,$CD,$CD,$7F,$80,$67,$51,$FE,$14,$6B,$FE,$29,$6A,$0E
	.dc.b	$1B,$1C,$C9,$5F,$E1,$09,$D1,$23,$33,$FE,$20,$8F,$36,$9F,$C5,$B8
	.dc.b	$BC,$CD,$BF,$63,$6B,$D3,$1F,$C6,$27,$6F,$F9,$57,$74,$62,$7F,$32
	.dc.b	$FC,$FF,$86,$F8,$04,$AF,$33,$67,$C8,$D7,$E2,$35,$F8,$8D,$7E,$26
	.dc.b	$08,$F9,$9F,$0D,$2B,$E1,$36,$BE,$03,$5F,$A0,$D7,$E8,$35,$EA,$D7
	.dc.b	$E8,$FB,$3F,$64,$93,$F3,$DB,$1F,$35,$AF,$DD,$6B,$F7,$5A,$FA,$9F
	.dc.b	$C2,$7C,$64,$9F,$B4,$D9,$FB,$0D,$7F,$10,$D7,$F1,$0D,$7E,$2B,$78
	.dc.b	$7F,$BC,$72,$FF,$90,$CF,$F2,$27,$25,$BF,$F2,$8C,$5B,$FF,$28,$4D
	.dc.b	$DB,$FC,$9F,$36,$BF,$CD,$FC,$7F,$C4,$6F,$FC,$AD,$39,$7F,$CA,$4F
	.dc.b	$69,$EA,$48,$DB,$FE,$60,$BE,$18,$79,$A3,$58,$7F,$97,$A7,$A5,$7E
	.dc.b	$8B,$D7,$FD,$51,$2B,$F4,$0D,$8F,$CF,$35,$FD,$73,$5F,$D7,$35,$FC
	.dc.b	$5B,$FB,$5F,$5E,$4F,$EA,$DB,$3F,$A9,$6B,$FB,$96,$BF,$B9,$6B,$F9
	.dc.b	$66,$BF,$BA,$F5,$BF,$89,$1F,$E7,$5F,$86,$B5,$FC,$42,$2C,$FF,$A0
	.dc.b	$66,$E3,$FE,$B5,$E1,$56,$2F,$E9,$D5,$BF,$D2,$CF,$F0,$F2,$4F,$F0
	.dc.b	$56,$C7,$F0,$36,$BF,$C4,$DA,$FF,$13,$6B,$D2,$39,$1F,$E9,$8F,$8E
	.dc.b	$F0,$69,$FE,$45,$6E,$C6,$92,$7D,$C7,$13,$F9,$7F,$C3,$7F,$35,$3F
	.dc.b	$D3,$1D,$F6,$32,$4C,$3F,$D4,$91,$97,$F9,$8C,$DD,$3F,$D4,$B3,$6B
	.dc.b	$F9,$64,$FF,$31,$23,$FC,$9D,$B5,$FC,$95,$AF,$F3,$56,$BC,$A3,$81
	.dc.b	$1E,$0F,$FA,$EB,$E1,$FC,$11,$FE,$B3,$6E,$5F,$E6,$96,$E4,$69,$2B
	.dc.b	$FA,$8F,$AF,$FD,$5C,$FF,$5D,$75,$58,$C9,$30,$FF,$63,$46,$5F,$EA
	.dc.b	$73,$73,$FF,$63,$CD,$AF,$E9,$D3,$FD,$4C,$8F,$F4,$96,$D7,$F4,$76
	.dc.b	$BF,$D5,$DA,$F2,$8E,$04,$78,$3F,$ED,$CE,$8F,$04,$7F,$B5,$5B,$97
	.dc.b	$FA,$2D,$B9,$7F,$4A,$AF,$66,$3F,$ED,$E9,$C8,$8F,$F6,$77,$E0,$4F
	.dc.b	$85,$C6,$3F,$EA,$B8,$C4,$FF,$BB,$3E,$23,$38,$6A,$95,$A7,$85,$EA
	.dc.b	$FF,$B8,$46,$A3,$FD,$D5,$79,$8F,$EE,$95,$C6,$FA,$36,$F8,$B6,$7F
	.dc.b	$DE,$95,$C5,$E9,$3E,$AB,$F2,$47,$06,$BD,$9A,$F0,$3C,$A6,$62,$B4
	.dc.b	$B5,$ED,$76,$D7,$B4,$70,$7A,$1A,$9E,$F3,$B8,$E9,$7F,$3B,$2F,$F6
	.dc.b	$17,$C7,$FA,$EB,$DD,$18,$C3,$D8,$7F,$86,$6B,$86,$6B,$E1,$8F,$CB
	.dc.b	$BE,$18,$FD,$2C,$38,$1F,$D3,$AB,$6E,$21,$3C,$30,$FA,$3E,$BF,$8C
	.dc.b	$E9,$8B,$C3,$C6,$1F,$88,$9B,$82,$FC,$B5,$87,$9D,$6C,$F0,$61,$FE
	.dc.b	$E1,$9D,$47,$DE,$BF,$8E,$47,$E6,$66,$8C,$61,$F5,$1A,$FD,$46,$BC
	.dc.b	$E7,$8C,$62,$3F,$1E,$FB,$3E,$23,$C3,$1F,$0B,$24,$63,$1F,$F7,$B3
	.dc.b	$63,$37,$EB,$AF,$80,$FF,$7A,$9D,$1F,$6F,$E1,$64,$7C,$0A,$7C,$9E
	.dc.b	$2F,$F5,$D5,$74,$FB,$B3,$6F,$F4,$33,$6B,$CD,$35,$FD,$8C,$7E,$38
	.dc.b	$93,$F2,$1B,$1F,$9A,$6B,$F9,$A6,$BD,$A3,$81,$1E,$0F,$A7,$C5,$F1
	.dc.b	$64,$C5,$18,$6C,$FF,$0C,$BF,$C4,$CE,$49,$1F,$C4,$3D,$7F,$D0,$1F
	.dc.b	$8E,$F8,$9B,$19,$26,$3F,$38,$FA,$D1,$A7,$F8,$B4,$DC,$FF,$4B,$36
	.dc.b	$BF,$84,$CF,$EC,$88,$FE,$31,$B5,$FB,$E6,$BF,$BE,$6B,$CA,$38,$11
	.dc.b	$E0,$FE,$73,$EA,$7F,$96,$3E,$8B,$FD,$F4,$62,$7C,$F3,$B3,$CC,$FE
	.dc.b	$CA,$35,$3F,$C8,$AB,$F7,$91,$8A,$F1,$57,$C3,$CD,$39,$8F,$E5,$7F
	.dc.b	$01,$FC,$2C,$E3,$E4,$5B,$FE,$7F,$16,$79,$57,$FA,$6A,$7F,$97,$CE
	.dc.b	$71,$FE,$66,$F2,$8E,$24,$AF,$E4,$B1,$FC,$78,$91,$FB,$86,$C6,$A3
	.dc.b	$04,$70,$43,$F9,$6C,$70,$9F,$93,$F1,$3E,$3A,$FF,$5D,$63,$F0,$23
	.dc.b	$17,$F2,$5B,$19,$9F,$35,$72,$69,$CB,$05,$7F,$31,$F6,$A4,$FD,$BF
	.dc.b	$26,$E3,$FD,$8E,$6E,$1F,$72,$6D,$7F,$56,$9B,$5F,$6E,$6B,$FD,$9A
	.dc.b	$3F,$A1,$92,$7F,$8E,$B6,$3F,$A7,$B5,$FE,$9E,$D7,$B4,$70,$23,$C1
	.dc.b	$FC,$97,$93,$E2,$CB,$F0,$AD,$C8,$D2,$4F,$F8,$0A,$FF,$1F,$1F,$D4
	.dc.b	$FD,$A9,$3F,$9D,$F2,$AE,$3F,$E0,$59,$B9,$FE,$7A,$6D,$7F,$78,$9B
	.dc.b	$5F,$9D,$9A,$FF,$82,$23,$FB,$29,$27,$F9,$FB,$63,$FB,$8B,$5F,$EE
	.dc.b	$2D,$7B,$47,$02,$3C,$1F,$DA,$79,$5E,$2C,$FF,$4E,$DC,$CD,$24,$FF
	.dc.b	$89,$2F,$F6,$31,$FE,$0F,$FC,$09,$85,$3F,$6B,$F1,$16,$FF,$B6,$42
	.dc.b	$77,$18,$6C,$7E,$C2,$78,$3F,$C7,$4E,$06,$33,$FE,$13,$16,$BF,$C5
	.dc.b	$75,$CA,$BF,$C3,$2D,$72,$7F,$87,$9C,$3E,$86,$BD,$A3,$FC,$31,$18
	.dc.b	$24,$71,$24,$69,$A7,$F0,$8F,$88,$F2,$3F,$C6,$5B,$9F,$F1,$46,$E6
	.dc.b	$69,$2B,$E2,$7A,$5F,$34,$FE,$3F,$E6,$58,$C9,$30,$FE,$2F,$19,$7E
	.dc.b	$46,$2A,$C6,$7F,$CC,$2A,$DF,$C8,$9F,$88,$47,$E0,$B6,$FF,$C8,$37
	.dc.b	$CE,$72,$48,$CC,$ED,$FA,$2D,$7B,$47,$F4,$3F,$33,$C3,$1F,$CF,$5B
	.dc.b	$9F,$F3,$86,$E6,$69,$2B,$F1,$1E,$87,$E4,$4F,$E8,$7F,$11,$63,$24
	.dc.b	$C3,$F9,$DC,$65,$F8,$B3,$74,$FE,$A7,$36,$BF,$17,$3F,$87,$23,$F8
	.dc.b	$46,$D7,$C4,$6B,$F9,$06,$BC,$A3,$81,$1E,$0F,$F6,$5F,$33,$C1,$1F
	.dc.b	$D7,$DB,$97,$F5,$A6,$E5,$FE,$F4,$AF,$66,$3F,$DA,$A7,$27,$EC,$FC
	.dc.b	$D9,$3F,$DC,$7E,$3E,$E3,$F8,$49,$B7,$FD,$CA,$6D,$7E,$AA,$6D,$7F
	.dc.b	$71,$9A,$FE,$E6,$3F,$54,$49,$FD,$3B,$63,$F0,$ED,$7F,$66,$D7,$B4
	.dc.b	$70,$23,$C1,$FF,$04,$7C,$3F,$8B,$2F,$EF,$6D,$C8,$D2,$4F,$F0,$BB
	.dc.b	$FF,$02,$C7,$F1,$5F,$32,$4F,$F8,$4F,$CD,$B8,$FE,$9A,$6E,$7F,$E1
	.dc.b	$49,$B5,$FC,$32,$6D,$7F,$84,$E6,$BF,$C6,$E3,$F8,$61,$27,$F8,$4B
	.dc.b	$63,$F5,$2D,$7F,$8A,$B5,$ED,$1C,$08,$F0,$7F,$C5,$5F,$1B,$E2,$CF
	.dc.b	$FC,$3A,$DC,$CD,$24,$FF,$2D,$BF,$F1,$3C,$7F,$37,$F2,$E4,$FF,$8D
	.dc.b	$7E,$22,$E3,$FC,$22,$6E,$7F,$E3,$69,$B5,$FC,$BA,$6D,$7F,$8D,$66
	.dc.b	$BF,$CF,$63,$F9,$71,$27,$F9,$5B,$63,$F8,$5B,$5F,$E6,$ED,$7B,$47
	.dc.b	$02,$3C,$1F,$F2,$67,$9D,$E2,$CF,$FC,$7E,$DC,$CD,$24,$FF,$4F,$BF
	.dc.b	$F2,$4C,$7F,$59,$F2,$24,$FF,$95,$FE,$46,$E3,$FC,$AA,$6E,$7F,$E5
	.dc.b	$89,$B5,$FD,$42,$6D,$7F,$95,$E6,$BF,$D7,$E3,$FA,$81,$27,$FA,$6B
	.dc.b	$63,$F9,$6B,$5F,$EB,$2D,$7B,$47,$02,$3C,$1F,$F3,$77,$C4,$78,$B3
	.dc.b	$FF,$30,$B7,$33,$49,$3F,$DC,$6F,$FC,$D7,$1F,$DE,$FC,$69,$3F,$ED
	.dc.b	$4F,$AA,$7A,$18,$CB,$FA,$6D,$5D,$3F,$DF,$15,$AB,$FE,$F9,$39,$7F
	.dc.b	$BA,$BF,$EF,$38,$D5,$72,$9D,$9F,$DC,$27,$FB,$B1,$1C,$12,$3F,$A9
	.dc.b	$C6,$5B,$19,$7F,$C0,$31,$96,$C7,$AF,$E3,$FC,$91,$FE,$FE,$A4,$63
	.dc.b	$44,$62,$BC,$B5,$FE,$8D,$8C,$3E,$93,$15,$C6,$33,$82,$BF,$C4,$B5
	.dc.b	$DB,$C8,$C3,$83,$FC,$53,$68,$CC,$66,$31,$3B,$8D,$CE,$9E,$B3,$A7
	.dc.b	$4D,$79,$46,$DE,$33,$CD,$F2,$36,$78,$47,$49,$E0,$E8,$EB,$E3,$E2
	.dc.b	$7F,$D0,$F5,$C6,$30,$47,$04,$9E,$2D,$7D,$D1,$B7,$27,$AB,$C7,$EC
	.dc.b	$49,$F3,$36,$35,$18,$23,$82,$1F,$2E,$38,$3F,$37,$C6,$F5,$12,$7E
	.dc.b	$13,$63,$E5,$B5,$E1,$1A,$23,$A6,$BF,$67,$C6,$F8,$C4,$8F,$9C,$D8
	.dc.b	$FA,$ED,$78,$46,$88,$E9,$AF,$E1,$7C,$5F,$A4,$48,$FB,$4D,$8F,$C1
	.dc.b	$B5,$E1,$1A,$23,$A6,$BF,$91,$F1,$3F,$2F,$A7,$E5,$33,$FF,$2C,$57
	.dc.b	$F0,$5F,$8A,$CB,$F9,$2A,$B7,$F9,$C6,$BF,$AE,$7F,$81,$4E,$2B,$9C
	.dc.b	$FF,$9F,$DB,$7D,$EB,$77,$FE,$8A,$9C,$BD,$CF,$CC,$47,$E4,$C8,$E0
	.dc.b	$95,$FA,$48,$CB,$63,$2F,$BD,$AF,$53,$F8,$07,$9F,$E7,$9F,$E3,$4D
	.dc.b	$9F,$E4,$CD,$9D,$C7,$9E,$76,$D3,$67,$AF,$C2,$F8,$47,$F2,$5F,$3A
	.dc.b	$C6,$49,$5F,$C4,$63,$2F,$C7,$9B,$8F,$F3,$29,$B5,$FC,$12,$7F,$6C
	.dc.b	$47,$F7,$CD,$AF,$E0,$ED,$7F,$88,$35,$E5,$1C,$08,$F0,$7F,$A3,$7C
	.dc.b	$4F,$82,$3F,$96,$37,$8F,$F9,$5B,$78,$CD,$25,$7D,$DF,$03,$F0,$87
	.dc.b	$F4,$7F,$3E,$C6,$49,$87,$F3,$38,$CB,$F8,$99,$B9,$FF,$54,$9B,$5F
	.dc.b	$C9,$A7,$F8,$C1,$1F,$E4,$0D,$AF,$E5,$2D,$7F,$99,$35,$E5,$1C,$08
	.dc.b	$F0,$7F,$B4,$7C,$77,$82,$3F,$A7,$37,$2F,$E9,$AD,$C8,$D2,$57,$E7
	.dc.b	$FB,$FF,$4C,$7F,$69,$F8,$AB,$19,$26,$1F,$D5,$63,$2F,$EA,$E6,$E7
	.dc.b	$FD,$DA,$6D,$7F,$48,$9F,$E7,$44,$7F,$A1,$36,$BF,$A5,$B5,$FE,$A8
	.dc.b	$D7,$94,$70,$23,$C1,$FF,$05,$7A,$1E,$08,$FE,$E0,$DC,$BF,$B7,$B7
	.dc.b	$23,$49,$5F,$C0,$7B,$3F,$84,$1F,$E0,$BF,$86,$B1,$92,$61,$FD,$DE
	.dc.b	$32,$FF,$0D,$9B,$9F,$F8,$62,$6D,$7F,$6A,$9F,$EB,$84,$7F,$B2,$36
	.dc.b	$BF,$B6,$B5,$FE,$EC,$D7,$94,$70,$23,$C1,$FF,$16,$7C,$37,$82,$3F
	.dc.b	$C2,$4D,$CB,$FC,$22,$DC,$8D,$25,$29,$8F,$F8,$23,$F0,$27,$3F,$F3
	.dc.b	$06,$D7,$F8,$EF,$D3,$C4,$C8,$E5,$5C,$63,$FC,$6D,$39,$BE,$B5,$FE
	.dc.b	$1A,$BE,$71,$83,$C4,$E2,$FF,$C8,$75,$C5,$E9,$5F,$E0,$F9,$E3,$5F
	.dc.b	$E2,$16,$BD,$6F,$FC,$1F,$18,$8E,$B1,$82,$4F,$36,$C7,$F8,$96,$78
	.dc.b	$26,$9E,$59,$F9,$19,$7F,$8F,$BE,$37,$CD,$18,$CD,$18,$BF,$F2,$33
	.dc.b	$6B,$85,$7C,$83,$D0,$7F,$99,$E3,$55,$CA,$35,$5F,$1B,$93,$EA,$1F
	.dc.b	$E6,$3F,$8B,$B3,$A9,$F4,$4E,$1F,$F2,$84,$DC,$FE,$F4,$DA,$F4,$4D
	.dc.b	$AE,$D1,$F3,$E7,$E5,$91,$C1,$23,$E4,$46,$5B,$19,$7E,$B4,$65,$B1
	.dc.b	$FE,$85,$F5,$7C,$19,$7F,$9A,$1B,$89,$A4,$9F,$76,$7F,$D0,$9E,$46
	.dc.b	$38,$FF,$30,$7F,$A1,$FE,$22,$C6,$49,$8F,$F9,$CE,$32,$FE,$4C,$FC
	.dc.b	$04,$E3,$4F,$F4,$AC,$DA,$FC,$D4,$FE,$38,$8F,$E5,$5B,$5F,$8B,$6B
	.dc.b	$F9,$76,$BC,$A3,$81,$1E,$0F,$FA,$A7,$E0,$FC,$11,$FE,$8F,$6E,$7F
	.dc.b	$E8,$E6,$E6,$69,$2B,$E3,$F0,$FE,$EC,$FF,$55,$7A,$F6,$32,$4C,$3F
	.dc.b	$D2,$D1,$97,$E2,$69,$FC,$2A,$75,$A7,$FA,$E2,$B4,$FF,$09,$BF,$D8
	.dc.b	$92,$7F,$70,$DB,$FD,$73,$5F,$DD,$35,$FE,$0C,$D7,$F8,$2B,$C6,$7F
	.dc.b	$D8,$FF,$17,$E0,$AF,$D8,$36,$BF,$D6,$8D,$CF,$FD,$66,$DC,$CD,$3F
	.dc.b	$8F,$70,$4B,$FF,$6C,$7C,$55,$C3,$E0,$CD,$C7,$FD,$AF,$36,$BF,$8E
	.dc.b	$4D,$AF,$F6,$24,$D7,$F8,$54,$7F,$14,$24,$FF,$1B,$6C,$7F,$11,$6B
	.dc.b	$FC,$75,$AF,$68,$E0,$47,$83,$FC,$1B,$17,$D5,$FF,$3D,$8F,$D3,$D7
	.dc.b	$F9,$B2,$73,$5E,$76,$CE,$8E,$55,$C6,$7E,$0C,$61,$E1,$19,$C3,$FA
	.dc.b	$46,$DF,$D1,$CE,$D0,$F8,$AC,$5F,$EE,$26,$DF,$F4,$1C,$3C,$F5,$8C
	.dc.b	$39,$DE,$CF,$1C,$65,$FE,$95,$7F,$CF,$88,$FF,$45,$6D,$FF,$3B,$6B
	.dc.b	$FD,$1D,$AF,$28,$E0,$47,$83,$F0,$BE,$2B,$C1,$0C,$3E,$6C,$61,$B3
	.dc.b	$E7,$9F,$66,$2A,$D3,$67,$58,$FF,$BE,$AF,$8D,$FE,$BB,$3F,$EE,$07
	.dc.b	$CA,$F6,$6D,$FF,$20,$8B,$3D,$A6,$DF,$F7,$99,$B5,$FD,$5B,$05,$62
	.dc.b	$FF,$B7,$D7,$F6,$62,$57,$F6,$D6,$C7,$F6,$36,$BF,$DB,$DA,$FF,$7B
	.dc.b	$6B,$D2,$39,$1F,$8A,$F8,$AF,$06,$3F,$94,$6E,$06,$92,$7D,$C7,$13
	.dc.b	$FA,$E7,$CE,$7C,$9B,$FE,$6C,$DC,$BE,$55,$BE,$2F,$FD,$8F,$39,$BF
	.dc.b	$F0,$F5,$A3,$0F,$F8,$85,$B8,$FE,$A2,$B8,$E9,$2F,$C2,$FE,$72,$31
	.dc.b	$7E,$28,$4F,$68,$EF,$7E,$77,$0D,$73,$3C,$1E,$07,$95,$C3,$C1,$AF
	.dc.b	$09,$DB,$B6,$CF,$58,$F6,$4F,$06,$CE,$A3,$11,$C4,$FC,$BF,$B7,$EC
	.dc.b	$CB,$E5,$B6,$BF,$14,$D9,$FD,$16,$48,$C6,$5F,$C4,$1B,$19,$AF,$B3
	.dc.b	$97,$D4,$3F,$82,$FC,$5D,$8C,$93,$0F,$D4,$C6,$5F,$BF,$37,$1F,$BF
	.dc.b	$36,$BE,$7C,$FB,$33,$FC,$14,$7C,$48,$C1,$1C,$12,$FE,$8C,$F0,$6C
	.dc.b	$ED,$B3,$B7,$ED,$C7,$F1,$EF,$8B,$F1,$47,$C4,$6E,$7F,$CC,$9B,$91
	.dc.b	$A4,$BF,$CE,$BF,$98,$3F,$94,$FB,$76,$32,$4B,$FE,$1B,$19,$7F,$47
	.dc.b	$37,$3F,$D1,$CD,$AF,$C9,$CF,$C5,$23,$F8,$86,$D7,$C2,$6B,$F9,$86
	.dc.b	$BC,$A3,$81,$1E,$0F,$F4,$8F,$8A,$F0,$47,$E1,$9B,$8F,$F5,$46,$E2
	.dc.b	$69,$2B,$F7,$FF,$15,$FB,$A3,$FA,$59,$5C,$92,$FF,$97,$C6,$5F,$E0
	.dc.b	$AD,$CB,$F8,$2B,$67,$F6,$AF,$E1,$C8,$FE,$A9,$B3,$F8,$46,$BF,$BA
	.dc.b	$6B,$CA,$38,$11,$E0,$FF,$6A,$F7,$3C,$11,$FA,$86,$E3,$FD,$D9,$B8
	.dc.b	$9A,$4A,$FE,$43,$F1,$3F,$C7,$0F,$ED,$BF,$15,$63,$24,$C3,$FA,$8C
	.dc.b	$65,$FE,$4F,$37,$2F,$E4,$F3,$6B,$F8,$BC,$FE,$A4,$8F,$F0,$C6,$D7
	.dc.b	$E9,$9A,$FF,$1C,$6B,$CA,$38,$11,$E0,$FF,$83,$23,$C1,$1F,$C2,$9B
	.dc.b	$8F,$F8,$61,$B8,$9A,$4A,$FE,$87,$F1,$3F,$CF,$8F,$F0,$7F,$BD,$63
	.dc.b	$24,$C3,$FB,$9C,$65,$FE,$93,$37,$2F,$E9,$33,$6B,$F9,$CC,$FF,$0B
	.dc.b	$23,$FC,$B9,$B5,$FC,$21,$AF,$F3,$E6,$BC,$A3,$81,$1E,$0F,$F8,$BB
	.dc.b	$E2,$BC,$11,$FC,$B1,$B9,$7F,$8E,$9B,$91,$A4,$AF,$EC,$BF,$13,$FC
	.dc.b	$84,$8F,$F4,$46,$DF,$F1,$E6,$BF,$D5,$1A,$FF,$58,$6B,$FD,$95,$FE
	.dc.b	$59,$F1,$32,$7F,$A7,$B6,$7F,$94,$B5,$FE,$BE,$D7,$FB,$1B,$5F,$EE
	.dc.b	$0D,$7F,$BA,$7B,$9F,$CE,$49,$3F,$D7,$1B,$1F,$CD,$9A,$FF,$6C,$6B
	.dc.b	$FD,$C1,$AF,$F7,$D7,$FA,$FF,$E0,$4F,$77,$FB,$86,$DF,$DB,$A3,$17
	.dc.b	$FE,$EE,$F9,$D7,$E1,$FF,$B3,$8F,$F7,$64,$EB,$1F,$F6,$33,$6F,$96
	.dc.b	$1D,$29,$4C,$39,$F1,$E6,$33,$87,$1A,$FF,$0A,$B6,$33,$7B,$BE,$35
	.dc.b	$E2,$8E,$52,$AF,$23,$FE,$11,$6B,$D2,$3E,$09,$27,$FC,$13,$3A,$9E
	.dc.b	$F1,$89,$D4,$7F,$85,$27,$4F,$78,$CB,$DD,$F3,$7E,$23,$FE,$2C,$8D
	.dc.b	$57,$CB,$8C,$57,$B6,$D1,$8B,$E3,$DB,$EC,$AD,$4E,$6F,$94,$FF,$A3
	.dc.b	$38,$A4,$7F,$A5,$BE,$0A,$E5,$FE,$70,$9B,$A7,$F9,$12,$6D,$7F,$98
	.dc.b	$E6,$D7,$F8,$6A,$6F,$1F,$FA,$A2,$6F,$1A,$6C,$61,$B7,$FE,$A6,$8F
	.dc.b	$F4,$FC,$70,$E2,$FF,$50,$F1,$7F,$A8,$9B,$5F,$EA,$C6,$E3,$FE,$86
	.dc.b	$6E,$5F,$E5,$A6,$CF,$F9,$E5,$B3,$FE,$30,$6E,$5F,$EB,$E6,$E4,$69
	.dc.b	$2B,$FD,$77,$87,$8D,$A7,$E7,$3D,$AF,$A0,$7F,$AE,$38,$EC,$64,$98
	.dc.b	$FF,$A7,$23,$2F,$E8,$A6,$E5,$FE,$72,$9B,$5F,$9C,$9F,$D8,$91,$FC
	.dc.b	$8B,$6B,$F1,$CD,$7F,$34,$D7,$94,$70,$23,$C1,$FF,$70,$7B,$FF,$E7
	.dc.b	$27,$CF,$7F,$EA,$D8,$C4,$F9,$E7,$67,$99,$FF,$5E,$46,$A7,$F8,$25
	.dc.b	$7F,$B9,$E3,$15,$E2,$AF,$F6,$C4,$61,$E3,$C3,$B3,$87,$C0,$DA,$F2
	.dc.b	$C5,$9E,$33,$70,$F3,$4D,$AF,$E1,$3A,$B0,$5E,$70,$EF,$5D,$0F,$83
	.dc.b	$19,$AE,$17,$6F,$8D,$6E,$FD,$33,$9B,$D7,$C0,$FB,$EF,$E0,$36,$FE
	.dc.b	$6B,$6F,$E0,$B7,$1F,$9C,$D9,$F4,$D2,$31,$5F,$39,$B8,$FB,$5B,$3B
	.dc.b	$AF,$B6,$F8,$9A,$78,$8B,$5E,$26,$DF,$F3,$BF,$87,$FE,$14,$7D,$9F
	.dc.b	$82,$B3,$FC,$EA,$2D,$7D,$A9,$B5,$C6,$B4,$7E,$32,$32,$FF,$36,$AF
	.dc.b	$E9,$64,$7F,$95,$B6,$BF,$94,$35,$FE,$68,$D7,$94,$70,$23,$C1,$FC
	.dc.b	$DF,$C1,$78,$23,$F2,$AD,$CB,$F0,$8D,$C4,$D2,$57,$E9,$3E,$37,$F3
	.dc.b	$47,$E6,$F6,$B1,$92,$61,$F9,$C8,$CB,$F9,$E9,$B9,$FE,$AE,$6D,$7D
	.dc.b	$E9,$B5,$FA,$A9,$B8,$FE,$66,$6E,$29,$B1,$87,$3F,$1B,$FA,$FF,$8C
	.dc.b	$FD,$83,$71,$FD,$EB,$71,$FD,$8B,$72,$FD,$F3,$67,$F2,$2D,$9F,$DE
	.dc.b	$B7,$2F,$D6,$37,$23,$49,$5F,$C3,$30,$F1,$B4,$FE,$FB,$87,$FB,$DB
	.dc.b	$81,$F0,$BF,$DD,$1B,$3F,$C5,$3E,$0F,$13,$23,$89,$FE,$06,$CF,$FC
	.dc.b	$29,$1F,$C3,$E7,$50,$BE,$D3,$87,$F8,$E4,$DA,$E5,$7E,$0B,$E8,$7F
	.dc.b	$84,$9C,$38,$6B,$C2,$D3,$C2,$38,$46,$1C,$4F,$07,$33,$B7,$D2,$D8
	.dc.b	$D4,$60,$E3,$3C,$DF,$13,$5F,$14,$FF,$3A,$F8,$CF,$64,$7F,$34,$6E
	.dc.b	$9F,$C8,$DB,$A1,$A4,$AF,$E8,$7F,$19,$FC,$E4,$FE,$75,$F0,$96,$32
	.dc.b	$4C,$3F,$9D,$C6,$5F,$E7,$B3,$74,$FE,$95,$36,$BF,$8E,$CD,$AF,$E9
	.dc.b	$33,$71,$FE,$71,$37,$14,$D8,$C3,$9E,$2F,$E9,$DF,$19,$FD,$3D,$B8
	.dc.b	$FF,$5C,$6E,$3F,$D4,$1B,$97,$F5,$D6,$CF,$F3,$06,$CF,$F5,$C6,$E5
	.dc.b	$FD,$2D,$B9,$1A,$4A,$FE,$D7,$87,$8D,$A7,$E6,$2F,$F1,$9E,$D7,$E7
	.dc.b	$CF,$F2,$B7,$94,$F2,$71,$5B,$AF,$F3,$85,$61,$FC,$E4,$7F,$99,$BC
	.dc.b	$35,$F1,$8E,$1C,$2B,$E1,$AF,$EB,$DE,$F8,$E2,$BF,$43,$1F,$8B,$C7
	.dc.b	$CB,$1F,$93,$8C,$47,$8A,$30,$47,$F3,$B1,$AB,$F0,$4E,$67,$F5,$13
	.dc.b	$92,$47,$1B,$FD,$36,$2F,$C5,$9E,$7F,$97,$0F,$F4,$14,$62,$FD,$B8
	.dc.b	$7F,$9F,$63,$13,$F1,$0F,$49,$6F,$A5,$F3,$6D,$F2,$3F,$D3,$15,$CB
	.dc.b	$0F,$F1,$07,$C2,$FF,$86,$4F,$F0,$D7,$C6,$DA,$FF,$13,$C5,$C7,$FC
	.dc.b	$37,$36,$B8,$D6,$8F,$F1,$6C,$65,$FE,$FB,$56,$BF,$C5,$33,$78,$7F
	.dc.b	$C3,$13,$78,$53,$63,$0E,$7E,$37,$FC,$5D,$F0,$BF,$E2,$F6,$E3,$FE
	.dc.b	$3C,$6E,$3F,$E3,$06,$E5,$FE,$3D,$6C,$FF,$84,$1B,$3F,$E3,$C6,$E5
	.dc.b	$FE,$2B,$6E,$46,$92,$BF,$C9,$78,$78,$DA,$7E,$53,$E0,$7F,$9F,$1F
	.dc.b	$EB,$BB,$7D,$7B,$E3,$73,$5B,$26,$5F,$76,$B7,$3A,$8F,$C0,$1F,$ED
	.dc.b	$5B,$C6,$7F,$CC,$6B,$77,$CA,$3F,$A8,$4F,$E5,$C9,$1F,$CC,$5B,$5F
	.dc.b	$CE,$1A,$FF,$41,$6B,$FD,$21,$AF,$F4,$77,$8C,$F8,$70,$FE,$98,$F8
	.dc.b	$2B,$F8,$15,$FC,$0E,$AF,$3C,$7F,$96,$2B,$6F,$53,$6E,$30,$ED,$18
	.dc.b	$9E,$30,$AE,$93,$C3,$E1,$BE,$75,$9B,$F1,$B8,$64,$EC,$EF,$18,$BF
	.dc.b	$A1,$5C,$67,$51,$89,$D6,$3E,$FB,$B5,$EB,$9B,$5E,$5B,$E4,$F4,$AF
	.dc.b	$1D,$74,$BE,$4D,$39,$16,$79,$7C,$6F,$B3,$E1,$FD,$AD,$9F,$8E,$D9
	.dc.b	$F4,$B6,$BE,$A3,$67,$E8,$36,$7A,$C7,$D0,$AE,$AD,$3A,$96,$3A,$B6
	.dc.b	$BA,$B4,$EA,$58,$0F,$C1,$5F,$03,$FE,$07,$AF,$EF,$D1,$8A,$FC,$0F
	.dc.b	$C7,$65,$FF,$10,$46,$6F,$F4,$2D,$AE,$37,$F8,$0B,$E1,$5E,$0A,$D5
	.dc.b	$F2,$AC,$CF,$8D,$E1,$1F,$D4,$67,$73,$C6,$7F,$A1,$C7,$86,$3F,$C4
	.dc.b	$ED,$7A,$4F,$F8,$76,$31,$1C,$63,$13,$C8,$8E,$09,$1F,$E2,$B9,$E2
	.dc.b	$F9,$63,$F4,$7F,$1D,$E8,$9F,$2E,$DF,$C2,$A3,$17,$FA,$06,$D7,$03
	.dc.b	$D9,$19,$69,$9B,$FC,$FF,$57,$EA,$0F,$D4,$FC,$75,$9F,$D9,$C5,$9F
	.dc.b	$D2,$CD,$CF,$F8,$04,$DA,$FD,$E4,$DA,$FD,$14,$DC,$7F,$90,$CD,$DF
	.dc.b	$FB,$4D,$78,$AF,$F8,$0C,$E6,$F4,$EF,$E3,$BF,$7B,$F1,$3F,$BE,$6D
	.dc.b	$7F,$08,$6D,$FE,$E9,$B9,$7F,$15,$6C,$FF,$11,$6C,$FE,$DD,$B9,$7F
	.dc.b	$36,$6E,$46,$92,$BF,$8B,$E1,$E3,$69,$F8,$BF,$03,$EC,$1F,$C5,$3E
	.dc.b	$3A,$C6,$49,$8F,$F1,$08,$CB,$F9,$E9,$B9,$7F,$2D,$9B,$5F,$A4,$9F
	.dc.b	$CA,$91,$FC,$93,$6B,$F1,$8D,$7F,$32,$D7,$94,$70,$23,$C1,$FE,$67
	.dc.b	$F1,$DE,$08,$FE,$38,$DC,$FF,$AD,$B7,$33,$49,$5F,$CF,$3E,$27,$F9
	.dc.b	$A9,$FC,$DB,$A2,$C6,$49,$87,$F3,$28,$CB,$FC,$D2,$6E,$7F,$D3,$E6
	.dc.b	$D7,$F4,$B9,$B5,$FC,$C6,$6E,$3F,$DC,$26,$E2,$9B,$18,$73,$F1,$DF
	.dc.b	$D2,$BE,$2B,$FA,$5B,$71,$FE,$AA,$DC,$7F,$A3,$B7,$2F,$EC,$4D,$9F
	.dc.b	$EB,$CD,$9F,$E8,$8D,$CB,$FC,$06,$DC,$8D,$25,$7F,$64,$C3,$C6,$D3
	.dc.b	$F8,$C7,$C0,$7F,$05,$3F,$90,$F8,$7F,$EE,$9E,$DC,$4E,$89,$97,$F9
	.dc.b	$1F,$CF,$FF,$09,$46,$34,$FE,$67,$87,$0A,$D5,$D9,$CD,$7F,$1F,$AF
	.dc.b	$E4,$A4,$70,$4A,$FE,$39,$19,$6C,$65,$FE,$57,$19,$6C,$7F,$A2,$70
	.dc.b	$7E,$37,$FE,$B9,$8F,$31,$FE,$83,$71,$3F,$D1,$2B,$E0,$C6,$2F,$CF
	.dc.b	$3F,$E6,$58,$C5,$7F,$4C,$AD,$F9,$BE,$D8,$E8,$FF,$82,$BE,$3B,$FC
	.dc.b	$17,$F1,$B6,$7F,$C2,$AD,$C3,$FC,$0E,$DC,$7F,$C4,$4D,$9F,$F0,$F3
	.dc.b	$67,$FC,$08,$DE,$0F,$F2,$1B,$78,$0D,$25,$7F,$89,$30,$F1,$B4,$F1
	.dc.b	$9F,$E2,$8F,$8C,$FF,$10,$9F,$E4,$FF,$B1,$FE,$E3,$19,$26,$3F,$E1
	.dc.b	$E9,$D3,$C2,$33,$97,$F8,$EE,$73,$18,$3F,$C7,$0E,$E7,$FC,$AD,$5D
	.dc.b	$9E,$93,$B7,$FC,$45,$3C,$9C,$92,$BB,$4F,$F9,$1A,$B9,$46,$27,$FC
	.dc.b	$BF,$F8,$32,$BF,$BC,$B5,$E1,$DB,$86,$1F,$EC,$B8,$D4,$F9,$A7,$2F
	.dc.b	$49,$B5,$FE,$69,$BF,$F0,$0C,$AF,$8C,$F8,$63,$FC,$37,$1D,$1F,$F0
	.dc.b	$E9,$8A,$FF,$64,$C6,$63,$2D,$75,$1C,$1C,$BC,$08,$65,$D9,$63,$93
	.dc.b	$E5,$8D,$B9,$7F,$01,$F0,$3E,$C7,$FC,$63,$AB,$04,$E6,$BF,$CD,$51
	.dc.b	$8A,$F9,$07,$22,$9F,$56,$78,$57,$37,$F1,$DF,$1F,$F8,$FF,$8F,$B5
	.dc.b	$F6,$DB,$87,$E1,$9B,$97,$CF,$6C,$F9,$B1,$46,$2B,$CA,$DC,$7D,$4D
	.dc.b	$9D,$D7,$E2,$70,$F1,$34,$F1,$16,$FC,$4D,$C3,$C4,$FE,$47,$E3,$7F
	.dc.b	$50,$7E,$A7,$E3,$EC,$64,$95,$FA,$08,$CB,$FA,$29,$B5,$F8,$B9,$B5
	.dc.b	$ED,$9B,$5F,$42,$6E,$5F,$56,$6E,$29,$B1,$87,$8D,$7E,$57,$DB,$FD
	.dc.b	$19,$FB,$FF,$4A,$E3,$FA,$18,$B3,$FB,$69,$B9,$7E,$6E,$39,$56,$CF
	.dc.b	$D9,$4E,$9F,$CC,$D7,$E7,$08,$FE,$56,$78,$9F,$95,$8C,$B6,$35,$B2
	.dc.b	$73,$3F,$A1,$8C,$4E,$A7,$33,$E0,$8F,$E0,$9D,$FE,$38,$FC,$BB,$74
	.dc.b	$FC,$E3,$74,$34,$95,$FC,$37,$E3,$BF,$92,$7C,$77,$F2,$56,$C6,$4F
	.dc.b	$E2,$51,$96,$99,$C3,$F8,$FC,$65,$B1,$9D,$3F,$85,$4E,$5B,$19,$3F
	.dc.b	$3D,$19,$69,$99,$FD,$9C,$65,$B1,$FC,$97,$D6,$FE,$78,$7F,$3D,$F8
	.dc.b	$AB,$19,$24,$FF,$34,$8C,$BF,$CD,$A6,$DF,$F1,$F9,$B5,$FB,$19,$B5
	.dc.b	$FC,$12,$6E,$7F,$C2,$66,$E6,$9B,$18,$78,$D7,$F2,$9F,$6B,$F9,$B9
	.dc.b	$FD,$4F,$E3,$AE,$3F,$CD,$62,$CF,$F4,$A9,$B9,$7E,$57,$75,$62,$FF
	.dc.b	$83,$55,$BF,$E6,$73,$FC,$C0,$93,$F9,$06,$C7,$F2,$96,$BF,$91,$6B
	.dc.b	$FC,$D9,$AF,$48,$E4,$7F,$55,$F8,$EF,$06,$9F,$CA,$DB,$A1,$A4,$9F
	.dc.b	$71,$C4,$FE,$BD,$DB,$FD,$D7,$E3,$BF,$BB,$36,$32,$7F,$63,$8C,$B4
	.dc.b	$CE,$1F,$DC,$23,$2D,$8C,$E9,$FD,$6E,$72,$D8,$C9,$FC,$BE,$32,$D3
	.dc.b	$33,$FD,$12,$32,$D8,$FE,$ED,$F1,$7F,$E0,$D3,$FC,$1D,$F1,$76,$32
	.dc.b	$49,$FF,$02,$C6,$5F,$F0,$3C,$DB,$FE,$E1,$36,$BF,$A1,$4D,$AF,$EA
	.dc.b	$B3,$73,$FE,$B5,$37,$34,$D8,$C3,$C6,$9F,$83,$7F,$97,$96,$3F,$C8
	.dc.b	$63,$0F,$F7,$DF,$0D,$F8,$EB,$DA,$FF,$00,$4E,$FC,$6B,$AF,$F8,$03
	.dc.b	$0E,$2D,$8F,$F3,$8D,$F1,$86,$3E,$07,$46,$2F,$A3,$4C,$C6,$C6,$3E
	.dc.b	$4C,$7C,$31,$D3,$1F,$0B,$71,$FE,$78,$F2,$3F,$C1,$D1,$F0,$5B,$1A
	.dc.b	$78,$5F,$F3,$B9,$D4,$71,$21,$A3,$FC,$28,$E9,$C1,$D2,$39,$4F,$FA
	.dc.b	$4B,$D0,$FF,$37,$CF,$1C,$7E,$74,$62,$3E,$AC,$EF,$C6,$F6,$C6,$2F
	.dc.b	$FD,$05,$5B,$C3,$DF,$92,$31,$87,$1A,$FF,$25,$7C,$67,$F9,$87,$E3
	.dc.b	$3F,$CC,$4D,$8C,$9F,$E4,$F8,$CB,$4C,$E3,$FE,$5C,$8C,$B6,$33,$9F
	.dc.b	$F9,$1E,$72,$D8,$C9,$FE,$1B,$8C,$B4,$CC,$FF,$8C,$23,$2D,$8F,$F3
	.dc.b	$17,$C6,$7F,$9E,$4F,$BF,$C4,$FB,$4F,$F2,$9C,$FF,$89,$E7,$4F,$03
	.dc.b	$F1,$54,$AE,$11,$A8,$FF,$3B,$CD,$C3,$FC,$B9,$5A,$AE,$55,$A6,$CE
	.dc.b	$AF,$C3,$3F,$E4,$8B,$E5,$5E,$33,$C3,$7C,$E7,$91,$2F,$FC,$DD,$5E
	.dc.b	$CB,$E7,$5F,$E5,$58,$FC,$0E,$AF,$5A,$9B,$F8,$3B,$97,$BB,$2F,$DE
	.dc.b	$CE,$1F,$D4,$61,$F1,$EB,$19,$7E,$FE,$AD,$FF,$99,$32,$56,$30,$FD
	.dc.b	$65,$7E,$95,$EB,$5F,$A5,$8C,$63,$E5,$8C,$12,$7F,$4B,$3A,$24,$66
	.dc.b	$3F,$67,$18,$77,$1F,$22,$DF,$1B,$E3,$CB,$F0,$0D,$8C,$CF,$C1,$C3
	.dc.b	$C5,$7F,$02,$72,$FF,$09,$3D,$26,$6B,$91,$F4,$7E,$37,$E9,$7C,$15
	.dc.b	$C3,$C4,$DA,$E8,$DD,$3F,$04,$D9,$F9,$4D,$9F,$C8,$60,$8C,$65,$F9
	.dc.b	$96,$C6,$6B,$E0,$61,$E2,$69,$E2,$2D,$F8,$9B,$87,$89,$F9,$1F,$1B
	.dc.b	$F8,$B3,$F1,$9F,$1B,$63,$24,$AF,$8D,$19,$7E,$44,$DA,$FC,$F4,$DA
	.dc.b	$FC,$34,$DA,$F6,$CD,$CB,$F6,$93,$71,$4D,$8C,$3C,$6B,$FA,$9F,$AF
	.dc.b	$F9,$E3,$F4,$5E,$3D,$C7,$EE,$45,$9F,$B9,$37,$2F,$EB,$13,$6B,$F3
	.dc.b	$53,$6B,$FA,$24,$FE,$58,$93,$FC,$F5,$B1,$FC,$E9,$AF,$F4,$66,$BF
	.dc.b	$D7,$5A,$F4,$8E,$47,$E5,$FD,$4F,$06,$7F,$C5,$5B,$99,$A4,$9F,$69
	.dc.b	$C4,$FD,$07,$C6,$FF,$03,$F8,$DF,$E0,$8D,$8C,$9F,$A5,$8C,$B4,$CE
	.dc.b	$1F,$A1,$8C,$B6,$33,$9F,$F1,$99,$CB,$63,$27,$F0,$38,$CB,$4C,$CF
	.dc.b	$E7,$23,$2D,$8F,$DC,$7C,$6F,$F1,$B3,$F8,$E7,$5D,$8C,$92,$7F,$6D
	.dc.b	$19,$7F,$71,$36,$FF,$97,$4D,$AF,$E2,$D3,$6B,$F5,$B3,$73,$FE,$89
	.dc.b	$37,$3F,$ED,$F1,$8B,$F0,$61,$FC,$E6,$BC,$B1,$83,$FC,$2B,$83,$D1
	.dc.b	$FF,$0B,$C7,$F2,$D6,$C7,$F0,$DF,$8D,$FE,$EB,$18,$BF,$F6,$53,$6B
	.dc.b	$9B,$6B,$F8,$1D,$70,$64,$F5,$7F,$A1,$46,$1F,$E3,$13,$E0,$7C,$A7
	.dc.b	$94,$FE,$68,$E5,$EE,$60,$93,$FE,$D5,$8D,$47,$F8,$C9,$AF,$38,$D3
	.dc.b	$C0,$EE,$64,$B1,$EB,$C3,$FB,$97,$8D,$E7,$C3,$FB,$23,$6F,$FA,$63
	.dc.b	$6F,$F9,$1B,$6B,$8C,$FC,$13,$FA,$04,$7F,$63,$F8,$DF,$EC,$9F,$1B
	.dc.b	$67,$FA,$23,$6F,$F9,$F3,$70,$FE,$E8,$D9,$FE,$B8,$D9,$FE,$66,$DC
	.dc.b	$3F,$C1,$4D,$D0,$D2,$57,$F5,$3C,$3C,$6D,$3C,$67,$F5,$7F,$8C,$FE
	.dc.b	$F4,$7F,$7B,$F8,$CB,$19,$26,$3F,$D5,$63,$2F,$F5,$79,$B9,$7F,$83
	.dc.b	$66,$D7,$F7,$59,$B5,$FD,$32,$6E,$5F,$E2,$39,$B9,$26,$C6,$1E,$35
	.dc.b	$F9,$DF,$4B,$F3,$27,$F9,$AB,$C3,$7E,$16,$F5,$73,$9F,$E8,$6A,$DF
	.dc.b	$F9,$CE,$F7,$58,$D3,$F4,$35,$6F,$FC,$FF,$5A,$79,$12,$FF,$17,$3F
	.dc.b	$9A,$24,$7E,$5D,$B1,$A8,$C1,$1C,$10,$FD,$2F,$02,$78,$DF,$FA,$6F
	.dc.b	$F0,$CF,$BB,$7E,$18,$C7,$07,$9F,$0F,$F4,$6D,$69,$FD,$8D,$FC,$58
	.dc.b	$C5,$7B,$AF,$FC,$9B,$8A,$73,$1F,$C0,$2B,$86,$1F,$E3,$DF,$1F,$FC
	.dc.b	$7D,$F1,$D6,$BF,$C3,$CD,$CB,$FC,$30,$DC,$BF,$C9,$CD,$9F,$F1,$B3
	.dc.b	$67,$FC,$1A,$DE,$1F,$F3,$23,$78,$4D,$25,$7F,$8B,$70,$F1,$B4,$F1
	.dc.b	$9F,$E5,$8D,$BF,$D5,$77,$F8,$3F,$1F,$FC,$E3,$0F,$8E,$B1,$2B,$83
	.dc.b	$A8,$C6,$5F,$E5,$2A,$E6,$48,$CE,$7F,$E7,$08,$CD,$71,$BE,$64,$BE
	.dc.b	$51,$96,$C6,$6B,$94,$E5,$B1,$FC,$E7,$D0,$FE,$6E,$7A,$BE,$36,$C6
	.dc.b	$49,$7E,$78,$CB,$FC,$CA,$6D,$79,$E6,$D7,$F3,$E9,$FE,$66,$47,$F9
	.dc.b	$83,$6B,$F9,$4B,$5F,$CB,$B5,$E5,$1C,$08,$F0,$7E,$A7,$C7,$78,$23
	.dc.b	$FD,$39,$8A,$31,$9F,$EE,$5B,$19,$AF,$6E,$3F,$3E,$72,$7D,$CF,$8E
	.dc.b	$FB,$BF,$1D,$6B,$F0,$4D,$C7,$ED,$36,$FE,$C3,$67,$FD,$FB,$8F,$F0
	.dc.b	$98,$C5,$79,$9B,$19,$9F,$C7,$5F,$81,$A7,$80,$B5,$E0,$6D,$F8,$1A
	.dc.b	$78,$31,$61,$F9,$6F,$8E,$93,$F9,$BF,$2A,$E1,$F9,$39,$B8,$7E,$46
	.dc.b	$6D,$73,$9B,$5F,$3A,$6E,$5F,$C9,$66,$E3,$FE,$90,$C3,$F3,$51,$83
	.dc.b	$F4,$F5,$70,$FD,$64,$61,$A7,$F8,$37,$CC,$FF,$05,$9F,$CC,$F9,$5E
	.dc.b	$E6,$27,$44,$CF,$F9,$94,$EA,$31,$8F,$F8,$4A,$F8,$56,$AE,$CE,$6B
	.dc.b	$FC,$09,$5F,$E0,$A2,$38,$25,$7F,$80,$23,$2D,$8C,$BF,$CC,$23,$2D
	.dc.b	$8F,$E4,$B6,$F7,$3C,$31,$FD,$12,$30,$FA,$0C,$46,$A5,$D3,$FC,$EE
	.dc.b	$30,$D9,$E1,$3A,$AE,$6F,$F0,$FF,$8D,$FE,$2F,$E9,$7F,$18,$6D,$7F
	.dc.b	$1D,$6C,$EE,$3F,$90,$4E,$DA,$6C,$B1,$B3,$F8,$E4,$6D,$B1,$B6,$9B
	.dc.b	$CB,$F8,$F4,$ED,$B3,$B6,$CF,$F2,$7F,$8D,$FE,$5E,$7F,$30,$F8,$DB
	.dc.b	$5F,$CD,$62,$CF,$F2,$D9,$B5,$FC,$AE,$6D,$7E,$96,$6D,$7F,$08,$9B
	.dc.b	$C3,$FD,$FE,$6F,$0A,$6C,$61,$B7,$E1,$17,$FE,$ED,$FC,$27,$FD,$7F
	.dc.b	$18,$C4,$FF,$9B,$23,$33,$FD,$0F,$E3,$B2,$F1,$F3,$AD,$E2,$15,$C2
	.dc.b	$30,$DB,$FE,$01,$7C,$1E,$D7,$6F,$CB,$5B,$6C,$7F,$5A,$AD,$47,$68
	.dc.b	$B3,$E6,$8F,$F3,$7B,$5E,$B5,$FE,$64,$78,$36,$35,$1D,$A3,$13,$EB
	.dc.b	$8E,$27,$F9,$B2,$76,$F2,$31,$1B,$79,$18,$9F,$EE,$9F,$1D,$E6,$9F
	.dc.b	$EE,$6D,$8C,$9F,$44,$EF,$87,$F6,$D8,$D5,$7F,$73,$6C,$66,$7F,$C0
	.dc.b	$1E,$77,$F7,$EF,$8E,$FE,$FE,$D8,$C9,$FE,$05,$8C,$B4,$CD,$FF,$7F
	.dc.b	$8C,$B6,$33,$5F,$DF,$E7,$2D,$8C,$9F,$CE,$E3,$2D,$33,$3F,$D8,$23
	.dc.b	$2D,$8F,$F0,$C7,$C7,$7F,$85,$8F,$F0,$BF,$C7,$58,$C9,$27,$FC,$29
	.dc.b	$19,$7F,$C2,$D3,$6F,$FC,$29,$36,$BF,$A7,$4D,$AF,$EE,$33,$78,$3F
	.dc.b	$CD,$13,$78,$13,$63,$0F,$1A,$FC,$EF,$99,$F9,$03,$FC,$E5,$F2,$8F
	.dc.b	$82,$C6,$5F,$6E,$BF,$37,$58,$7F,$CE,$55,$B9,$D4,$74,$6F,$02,$B8
	.dc.b	$69,$FB,$0E,$76,$1C,$70,$FD,$D5,$75,$7F,$42,$FE,$4D,$E8,$7E,$65
	.dc.b	$C5,$F5,$C3,$F4,$51,$82,$38,$24,$F1,$6B,$D9,$FF,$35,$FC,$9B,$E2
	.dc.b	$7C,$F1,$FE,$8A,$8C,$3E,$CA,$FF,$4C,$46,$2B,$D3,$3C,$AB,$8C,$7F
	.dc.b	$4A,$9D,$65,$FC,$2A,$F8,$D7,$F9,$73,$E3,$7F,$CB,$BF,$1D,$71,$FF
	.dc.b	$31,$37,$1F,$F2,$D3,$70,$FF,$2C,$36,$7F,$C2,$4D,$9F,$F1,$AB,$78
	.dc.b	$7F,$D5,$8D,$E1,$34,$95,$FE,$71,$C3,$C6,$D3,$C6,$7F,$9D,$7E,$3B
	.dc.b	$FC,$E6,$7F,$1F,$B7,$DB,$FF,$BD,$70,$4E,$AB,$81,$F8,$BA,$DE,$39
	.dc.b	$3F,$CE,$D3,$72,$FE,$6B,$3A,$AE,$35,$A6,$CE,$B0,$F0,$5F,$19,$FF
	.dc.b	$1A,$3E,$58,$CC,$70,$32,$78,$DC,$D7,$F4,$BF,$2F,$F9,$99,$C7,$E3
	.dc.b	$ED,$7E,$72,$2C,$B0,$F1,$56,$1F,$E8,$F3,$6B,$FC,$AB,$6A,$C5,$FF
	.dc.b	$46,$AF,$CD,$92,$BF,$9E,$36,$3F,$9C,$35,$FE,$8A,$D7,$FA,$F3,$5E
	.dc.b	$91,$C8,$F7,$7A,$1E,$0C,$BE,$BB,$70,$34,$93,$EF,$38,$9F,$1F,$6F
	.dc.b	$99,$F1,$FF,$35,$B1,$93,$E6,$46,$5A,$67,$0F,$90,$FC,$48,$D4,$60
	.dc.b	$61,$F8,$C8,$CE,$7F,$7D,$B5,$C4,$E9,$19,$69,$9A,$E9,$5F,$96,$BF
	.dc.b	$3C,$63,$1F,$DA,$D6,$EF,$F1,$9A,$FE,$24,$FC,$57,$C7,$D9,$FC,$44
	.dc.b	$5B,$E5,$36,$B8,$D6,$8F,$C7,$46,$5F,$C1,$D5,$AF,$7D,$AB,$18,$FE
	.dc.b	$B2,$AC,$E6,$FA,$5F,$A6,$73,$7C,$E1,$F8,$4B,$C7,$FF,$07,$B6,$FF
	.dc.b	$9E,$73,$FF,$9F,$BE,$3B,$35,$27,$F9,$DD,$6E,$FE,$A5,$EE,$73,$5B
	.dc.b	$9F,$B0,$78,$1F,$F1,$14,$EE,$33,$5C,$F4,$FF,$17,$3D,$63,$FC,$76
	.dc.b	$7F,$8F,$1C,$38,$6B,$9A,$FE,$89,$18,$8F,$05,$FF,$89,$63,$04,$9D
	.dc.b	$92,$34,$D9,$F6,$FC,$A7,$FA,$6D,$F3,$C7,$AA,$31,$7E,$88,$C7,$93
	.dc.b	$F0,$63,$FA,$E4,$EA,$FD,$15,$A7,$E3,$9F,$42,$35,$7F,$80,$F1,$FF
	.dc.b	$8B,$61,$FC,$5D,$B5,$FC,$55,$B3,$F9,$76,$E1,$FC,$85,$B3,$FC,$39
	.dc.b	$B3,$D6,$3F,$8F,$57,$56,$9D,$4B,$1D,$5B,$5D,$5A,$75,$2C,$7E,$63
	.dc.b	$C8,$FE,$5C,$7F,$2F,$F9,$1B,$3F,$CB,$62,$CF,$EB,$E6,$D7,$F3,$99
	.dc.b	$B5,$FC,$9A,$6D,$75,$8E,$47,$F3,$78,$CF,$0F,$F5,$0A,$BC,$29,$B1
	.dc.b	$87,$9D,$FE,$93,$C3,$FD,$19,$FD,$17,$E4,$6E,$3F,$A0,$8B,$3F,$C1
	.dc.b	$66,$E5,$F9,$89,$B5,$FC,$F2,$6D,$7E,$66,$7F,$2A,$49,$FC,$A3,$63
	.dc.b	$F3,$E6,$E3,$24,$72,$79,$DC,$3C,$37,$4F,$38,$C3,$FD,$7F,$87,$C7
	.dc.b	$A7,$F7,$46,$E8,$69,$30,$F8,$07,$F6,$37,$FB,$47,$81,$FD,$A7,$D0
	.dc.b	$B3,$FD,$9D,$B8,$FF,$29,$6E,$9F,$DC,$DB,$3F,$D8,$1B,$3F,$C7,$9B
	.dc.b	$8F,$F8,$21,$B8,$9A,$4A,$FE,$ED,$87,$8D,$A7,$8C,$FE,$F5,$F2,$1F
	.dc.b	$DF,$CF,$F0,$07,$C8,$58,$C9,$31,$FE,$81,$19,$7F,$A1,$4D,$CB,$FC
	.dc.b	$19,$36,$BF,$BC,$CD,$AF,$E6,$F3,$72,$FF,$10,$CD,$C9,$36,$30,$F1
	.dc.b	$AF,$E6,$BD,$BF,$CD,$0F,$F6,$46,$2F,$AD,$79,$F9,$CF,$F9,$9D,$5B
	.dc.b	$FE,$CD,$7B,$AC,$3F,$CB,$26,$D7,$FB,$8E,$F9,$BC,$89,$7F,$8E,$E0
	.dc.b	$FF,$1B,$C7,$F2,$9A,$FE,$5A,$47,$04,$9E,$11,$96,$C6,$5F,$CB,$46
	.dc.b	$6F,$FD,$DB,$FA,$C7,$FB,$D2,$BC,$31,$87,$CF,$5F,$E2,$B9,$D3,$FD
	.dc.b	$22,$BE,$2C,$62,$BD,$D5,$FE,$95,$51,$98,$FE,$91,$3C,$1F,$F2,$17
	.dc.b	$C8,$C7,$FC,$81,$F2,$37,$0F,$F0,$14,$DC,$3F,$C9,$93,$6B,$FC,$73
	.dc.b	$36,$BF,$BB,$CD,$DF,$FC,$C3,$37,$74,$D8,$C3,$6F,$FC,$A1,$1F,$E5
	.dc.b	$58,$E9,$83,$F6,$07,$B3,$F8,$45,$ED,$58,$C3,$70,$F8,$53,$A6,$CE
	.dc.b	$9A,$6A,$78,$67,$F8,$AA,$E0,$DA,$E0,$DA,$E0,$72,$78,$57,$2B,$4A
	.dc.b	$7F,$CC,$97,$D2,$73,$18,$BE,$B1,$83,$E0,$38,$AF,$F0,$1F,$E1,$7F
	.dc.b	$E1,$FA,$FF,$06,$B6,$BF,$D0,$1F,$25,$C6,$B7,$E0,$DB,$F1,$5E,$E7
	.dc.b	$27,$F9,$FE,$B5,$29,$D3,$C2,$B3,$18,$DB,$FD,$3B,$58,$99,$7F,$CD
	.dc.b	$AF,$94,$7F,$B5,$EF,$D1,$1F,$CF,$9A,$F8,$8F,$E7,$AE,$1F,$04,$F3
	.dc.b	$8C,$47,$48,$C4,$F5,$23,$83,$FD,$35,$F2,$5E,$49,$FF,$2E,$37,$1F
	.dc.b	$F5,$33,$67,$FD,$26,$D9,$FF,$28,$B7,$1F,$F5,$DB,$FB,$0F,$92,$95
	.dc.b	$FA,$FE,$27,$C8,$E3,$0F,$E2,$70,$AC,$63,$FC,$02,$AD,$FE,$16,$3F
	.dc.b	$6F,$38,$BF,$E0,$15,$67,$35,$DE,$BF,$80,$CE,$6F,$B4,$E4,$F2,$46
	.dc.b	$67,$4F,$5C,$7F,$DA,$BC,$DF,$ED,$7F,$BB,$FE,$0D,$97,$F9,$C2,$31
	.dc.b	$87,$FB,$1D,$AF,$FB,$15,$B1,$FE,$87,$6E,$5F,$ED,$16,$CF,$FA,$F9
	.dc.b	$B3,$D6,$3F,$DC,$15,$E0,$BF,$1C,$60,$F0,$14,$F0,$16,$7C,$0D,$BF
	.dc.b	$D5,$C7,$E6,$7A,$BF,$2E,$7F,$07,$F8,$0B,$5F,$3C,$FD,$5C,$62,$7F
	.dc.b	$80,$4D,$AE,$7B,$AB,$17,$F5,$F8,$97,$8C,$3F,$5B,$5F,$9F,$25,$7E
	.dc.b	$65,$B1,$F9,$E6,$BF,$A3,$6B,$E4,$8E,$04,$78,$3F,$CF,$7F,$61,$78
	.dc.b	$5E,$1C,$91,$89,$F7,$57,$91,$E1,$5B,$9D,$4F,$F0,$AB,$FE,$DF,$18
	.dc.b	$BF,$83,$18,$AF,$C3,$6F,$30,$FB,$3B,$DC,$7F,$D9,$7B,$7E,$49,$B5
	.dc.b	$F8,$D9,$D6,$3F,$C8,$36,$5E,$E3,$0D,$C3,$F9,$44,$EA,$B8,$C6,$27
	.dc.b	$58,$7F,$B9,$1B,$5B,$9F,$C6,$E1,$C2,$B4,$F8,$E3,$24,$8F,$3B,$EC
	.dc.b	$31,$87,$F7,$6F,$92,$FE,$DF,$F2,$7F,$DC,$1B,$3F,$C0,$5B,$5F,$C0
	.dc.b	$1B,$5C,$67,$F8,$35,$71,$69,$C4,$B5,$C5,$B3,$C6,$3F,$85,$CF,$16
	.dc.b	$CF,$16,$9C,$4B,$3F,$9A,$E3,$FE,$72,$7E,$FE,$ED,$7E,$1A,$BF,$A0
	.dc.b	$4E,$1F,$B7,$36,$B9,$56,$F8,$57,$CE,$30,$F1,$BF,$E7,$CF,$F4,$12
	.dc.b	$47,$F3,$B6,$C7,$F4,$06,$BC,$23,$44,$74,$DA,$FC,$3E,$1E,$58,$FC
	.dc.b	$23,$67,$F0,$8D,$CB,$F1,$ED,$CC,$D3,$FC,$19,$F0,$72,$FF,$C1,$3F
	.dc.b	$07,$70,$FF,$02,$4D,$C7,$FB,$B7,$1A,$B1,$7F,$C1,$6A,$DF,$F8,$02
	.dc.b	$6D,$FF,$86,$E6,$DF,$1B,$FF,$0B,$C6,$0F,$F0,$F5,$58,$C9,$E4,$2D
	.dc.b	$FE,$C3,$E4,$FF,$9D,$7B,$BF,$CE,$DB,$3F,$CB,$9B,$3B,$8F,$E6,$33
	.dc.b	$B6,$9B,$2C,$6C,$FD,$EC,$6D,$B1,$B6,$9B,$C7,$F7,$73,$B6,$CE,$DB
	.dc.b	$27,$F9,$CF,$2A,$FF,$C2,$6D,$AF,$ED,$D1,$82,$3F,$99,$8D,$57,$F7
	.dc.b	$09,$CB,$D6,$7F,$C6,$4F,$F3,$88,$CD,$7F,$9C,$9B,$3A,$AE,$75,$FE
	.dc.b	$D4,$6B,$99,$FE,$9E,$D9,$FF,$10,$4E,$A3,$FC,$80,$D7,$C1,$1A,$9E
	.dc.b	$31,$A6,$BA,$3F,$DB,$F3,$C4,$91,$FE,$80,$BF,$3C,$FF,$A1,$5B,$7F
	.dc.b	$E8,$06,$F3,$9A,$4C,$3F,$B8,$FC,$9F,$F8,$04,$FF,$01,$7C,$9D,$8C
	.dc.b	$93,$0F,$E7,$B1,$97,$F9,$FC,$DC,$3F,$9C,$CD,$AF,$E6,$D3,$6B,$F9
	.dc.b	$B4,$DE,$0F,$E8,$33,$78,$13,$63,$0F,$1C,$3F,$D0,$5F,$0D,$FE,$81
	.dc.b	$F3,$FF,$D0,$4D,$CB,$FC,$FE,$DC,$BF,$CF,$4D,$AF,$DB,$B6,$7F,$BD
	.dc.b	$36,$7F,$CE,$AD,$9E,$F3,$FE,$8A,$AE,$ED,$3B,$96,$BB,$B6,$7B,$B4
	.dc.b	$FC,$67,$C8,$FE,$70,$FF,$3A,$7A,$2F,$41,$E7,$E6,$BF,$39,$57,$0F
	.dc.b	$CD,$0B,$E1,$59,$C3,$F5,$35,$6F,$FD,$21,$5A,$79,$92,$FF,$1B,$3F
	.dc.b	$90,$E2,$34,$93,$F9,$C6,$CF,$08,$C1,$1C,$10,$FD,$44,$7B,$B1,$79
	.dc.b	$FE,$59,$F5,$46,$2B,$C9,$8F,$FA,$93,$1F,$89,$7B,$6C,$E8,$F6,$D7
	.dc.b	$1B,$F8,$A7,$F8,$D3,$0F,$05,$FF,$B9,$B6,$97,$FE,$77,$DA,$DF,$F9
	.dc.b	$DA,$6E,$5F,$EF,$49,$B5,$FE,$75,$E2,$56,$2F,$FD,$E9,$56,$FF,$DD
	.dc.b	$F3,$6B,$F3,$77,$FE,$85,$8C,$1F,$EF,$8A,$B1,$93,$C8,$5B,$FF,$15
	.dc.b	$FC,$9F,$F9,$D7,$E4,$FF,$CE,$CD,$9F,$F3,$4B,$67,$71,$FE,$6C,$9D
	.dc.b	$B4,$D9,$63,$67,$F8,$E2,$36,$D8,$DB,$4D,$E3,$FE,$36,$9D,$B6,$76
	.dc.b	$D9,$FA,$7B,$3E,$2B,$F9,$D9,$FE,$F9,$F8,$AB,$7F,$CE,$62,$D7,$FB
	.dc.b	$C6,$6D,$FF,$44,$9B,$5F,$E7,$19,$B5,$FC,$BA,$7F,$34,$49,$FE,$66
	.dc.b	$D8,$FE,$56,$D7,$F9,$E3,$5F,$E8,$0D,$7A,$47,$23,$FC,$E1,$C3,$E0
	.dc.b	$BF,$F3,$7B,$7C,$3F,$DC,$CF,$AC,$D9,$6B,$F7,$35,$C2,$30,$FF,$BF
	.dc.b	$E7,$8C,$FF,$9A,$BE,$4F,$FD,$68,$7F,$AD,$78,$6D,$FF,$AA,$22,$CF
	.dc.b	$F9,$C2,$6D,$FF,$9A,$A6,$D7,$F9,$9E,$6D,$7F,$99,$E6,$F3,$FF,$9C
	.dc.b	$A6,$F3,$A6,$C6,$1B,$7E,$13,$FD,$C1,$B9,$FF,$71,$6F,$71,$FF,$70
	.dc.b	$CD,$CA,$71,$2B,$17,$FA,$DA,$B7,$FE,$03,$AD,$4F,$63,$A9,$63,$FD
	.dc.b	$DB,$5C,$2D,$4B,$C3,$CA,$BC,$77,$C9,$A7,$27,$A5,$79,$AF,$8D,$F2
	.dc.b	$6C,$F1,$9F,$5C,$E6,$3F,$CD,$5F,$86,$3A,$3D,$78,$23,$13,$FE,$B4
	.dc.b	$E7,$7A,$DF,$EC,$D8,$E3,$7C,$2B,$77,$FC,$71,$B8,$FF,$25,$BE,$13
	.dc.b	$9A,$E0,$4B,$ED,$3A,$AF,$F7,$2D,$F2,$9F,$F1,$1B,$DA,$FF,$C4,$A7
	.dc.b	$9E,$FF,$C2,$71,$98,$DC,$60,$93,$C2,$07,$17,$0D,$35,$3C,$DD,$B5
	.dc.b	$D1,$E2,$70,$79,$DF,$EC,$1E,$BF,$FB,$56,$7C,$67,$CB,$71,$1D,$2B
	.dc.b	$FB,$6C,$62,$3C,$57,$FD,$36,$31,$7D,$71,$46,$2F,$8F,$8D,$F8,$DF
	.dc.b	$01,$F1,$94,$C7,$C6,$5D,$3F,$DD,$F9,$AB,$71,$86,$E9,$FC,$09,$57
	.dc.b	$18,$C5,$DB,$FF,$46,$CD,$BF,$9B,$1D,$67,$15,$FE,$69,$BE,$75,$DE
	.dc.b	$73,$5B,$8F,$15,$E2,$BC,$73,$FE,$9A,$AF,$5D,$71,$3B,$9F,$EE,$FF
	.dc.b	$82,$FD,$AF,$2F,$FB,$E3,$E5,$2F,$32,$6C,$61,$B5,$CB,$3F,$F7,$B4
	.dc.b	$61,$FF,$0D,$DA,$F9,$56,$E3,$B3,$63,$33,$C1,$5F,$79,$D3,$67,$4D
	.dc.b	$C7,$F8,$A5,$F4,$9E,$58,$78,$E7,$38,$76,$6D,$F1,$C3,$F3,$B8,$71
	.dc.b	$9F,$D5,$FC,$5F,$ED,$CF,$E1,$FC,$D6,$BF,$81,$C5,$CB,$FA,$0F,$37
	.dc.b	$F2,$D9,$C9,$23,$F7,$D3,$CA,$31,$7D,$9F,$D5,$F1,$A7,$F6,$A4,$8C
	.dc.b	$B6,$BF,$5E,$D7,$F5,$8D,$7F,$80,$B5,$FD,$9B,$FC,$A3,$9F,$D1,$7F
	.dc.b	$CE,$1B,$5F,$E2,$36,$D7,$F3,$76,$CF,$F3,$26,$E1,$FE,$4C,$AF,$B5
	.dc.b	$F2,$9F,$88,$F9,$0F,$C4,$B6,$32,$7D,$F8,$CB,$4C,$E1,$F8,$18,$CB
	.dc.b	$63,$38,$7D,$A9,$CB,$63,$27,$DA,$D9,$3A,$8C,$36,$9A,$FE,$56,$30
	.dc.b	$D9,$FD,$E7,$C7,$7F,$05,$3F,$83,$7C,$7D,$8C,$92,$FF,$80,$46,$5F
	.dc.b	$E0,$53,$71,$FE,$CD,$C6,$AC,$5F,$F0,$08,$56,$2F,$F8,$05,$5B,$FE
	.dc.b	$21,$36,$B9,$5F,$86,$7A,$CE,$6B,$95,$FF,$55,$F8,$8F,$EA,$07,$FA
	.dc.b	$53,$E5,$9F,$0F,$8C,$7F,$AE,$CD,$AF,$F4,$9F,$5A,$F7,$5A,$24,$FF
	.dc.b	$59,$AE,$97,$C7,$2F,$EA,$71,$FD,$38,$95,$F9,$E6,$D7,$F4,$86,$BF
	.dc.b	$D1,$5A,$FF,$5A,$6B,$FD,$31,$E3,$3F,$E8,$BE,$8F,$DE,$56,$CF,$EE
	.dc.b	$4E,$2B,$84,$63,$9F,$CC,$FF,$AA,$A7,$53,$F1,$E3,$47,$BA,$BB,$57
	.dc.b	$F8,$40,$FF,$08,$F3,$D9,$FE,$8B,$C7,$FC,$61,$B3,$A6,$9A,$9E,$17
	.dc.b	$FC,$21,$B7,$FE,$2A,$6C,$FF,$A5,$5B,$3F,$E2,$96,$CF,$F8,$91,$B7
	.dc.b	$E1,$9F,$E2,$77,$E1,$6D,$F8,$4F,$E6,$1F,$2B,$FC,$F8,$FE,$7F,$F2
	.dc.b	$B6,$32,$49,$FE,$69,$19,$7F,$9B,$4D,$C7,$F9,$6C,$DA,$FD,$94,$DA
	.dc.b	$FD,$7F,$22,$B1,$C7,$FD,$16,$AC,$E6,$FC,$18,$7F,$38,$9C,$DF,$3C
	.dc.b	$1F,$86,$3E,$1F,$FC,$66,$DC,$3F,$BA,$74,$E2,$64,$7F,$57,$AE,$31
	.dc.b	$FE,$01,$9C,$C7,$FB,$3A,$B9,$36,$75,$5C,$67,$38,$FF,$98,$EB,$9C
	.dc.b	$FF,$89,$27,$8C,$7F,$8C,$1A,$FF,$8A,$8D,$9F,$E2,$D7,$2F,$F8,$B9
	.dc.b	$AE,$8E,$55,$DD,$E5,$1A,$24,$66,$39,$35,$EF,$5F,$E9,$3F,$2B,$DE
	.dc.b	$DB,$F6,$57,$F9,$BF,$C4,$FF,$78,$4F,$B7,$0F,$7D,$7F,$85,$A7,$2F
	.dc.b	$8A,$B8,$5F,$48,$D3,$FE,$0A,$F1,$BF,$C1,$7C,$96,$7F,$C0,$0D,$CB
	.dc.b	$FC,$00,$DC,$BF,$BB,$B6,$7F,$9E,$36,$7F,$16,$DD,$FF,$C1,$CD,$DF
	.dc.b	$F8,$BC,$FF,$82,$F0,$F0,$DF,$F8,$3D,$B1,$98,$F1,$99,$2B,$FD,$F3
	.dc.b	$AF,$FD,$1E,$7F,$A4,$3A,$EC,$FF,$00,$8B,$3F,$E0,$29,$B9,$7F,$A9
	.dc.b	$26,$D7,$F8,$FB,$8F,$FA,$AB,$C4,$97,$DD,$FE,$17,$7C,$23,$FD,$51
	.dc.b	$56,$FF,$D4,$73,$6D,$36,$38,$57,$E9,$BE,$1B,$F4,$A7,$FA,$CB,$A5
	.dc.b	$F1,$EF,$85,$CE,$7F,$9D,$AB,$7F,$D0,$AB,$53,$92,$47,$E9,$E7,$A4
	.dc.b	$EA,$31,$3A,$7F,$45,$3F,$93,$E3,$34,$93,$FA,$96,$CF,$E9,$5A,$F2
	.dc.b	$8D,$11,$D3,$FE,$B1,$F8,$E7,$C5,$FA,$B6,$FF,$6C,$66,$9C,$D7,$AA
	.dc.b	$73,$18,$BF,$49,$FD,$89,$C4,$7E,$8F,$0F,$95,$18,$AF,$D9,$C6,$30
	.dc.b	$FF,$46,$7C,$57,$FA,$37,$E4,$ED,$7F,$A1,$DB,$8F,$FA,$1D,$B8,$FF
	.dc.b	$81,$9B,$3F,$E0,$56,$CF,$FA,$0D,$B3,$FE,$93,$6F,$84,$69,$2B,$FD
	.dc.b	$2D,$7E,$36,$9E,$33,$FC,$E9,$F7,$CF,$83,$FF,$41,$B6,$33,$F0,$98
	.dc.b	$99,$1C,$1D,$C6,$32,$FF,$40,$D7,$16,$CE,$AB,$81,$23,$2F,$63,$19
	.dc.b	$7F,$9F,$2B,$93,$6B,$95,$70,$24,$66,$B9,$46,$1F,$E4,$FF,$09,$FD
	.dc.b	$18,$3F,$DE,$3B,$F0,$FF,$3D,$8B,$3F,$E7,$C9,$B5,$FC,$8E,$6D,$7E
	.dc.b	$EE,$6D,$7F,$39,$9F,$CF,$12,$7F,$A2,$36,$3F,$A0,$35,$FE,$8C,$D7
	.dc.b	$CF,$1C,$08,$F0,$7F,$DA,$7F,$8E,$FD,$F7,$D0,$63,$4F,$1E,$09,$D5
	.dc.b	$7F,$5D,$BF,$82,$A6,$9B,$AE,$35,$C2,$3E,$05,$F5,$3F,$D8,$1C,$2F
	.dc.b	$C5,$1F,$05,$FC,$35,$B3,$A5,$F0,$79,$99,$39,$78,$46,$32,$FF,$28
	.dc.b	$D7,$26,$CE,$AB,$81,$23,$2F,$7A,$E5,$87,$F0,$0B,$E4,$DB,$E5,$7C
	.dc.b	$09,$19,$BE,$47,$57,$FC,$F3,$F2,$DF,$1A,$3F,$CF,$1F,$2D,$6B,$FD
	.dc.b	$ED,$36,$9C,$6A,$C2,$9B,$18,$6D,$FF,$BA,$EB,$8C,$62,$FE,$B4,$FF
	.dc.b	$A1,$E3,$13,$E4,$8E,$17,$DD,$B3,$A7,$81,$AB,$F0,$34,$D9,$6F,$B9
	.dc.b	$EC,$71,$7F,$E1,$FF,$C3,$9F,$07,$F0,$69,$18,$AF,$C2,$FE,$37,$F8
	.dc.b	$07,$FC,$61,$19,$9E,$A7,$0A,$E5,$19,$BF,$E6,$F7,$C6,$B7,$39,$BF
	.dc.b	$F1,$53,$60,$73,$8F,$0E,$BD,$30,$91,$FE,$31,$9E,$31,$FE,$25,$AF
	.dc.b	$24,$65,$AE,$A7,$83,$DA,$32,$D9,$E4,$49,$DC,$78,$CE,$33,$C1,$C1
	.dc.b	$DE,$7F,$BA,$5B,$E1,$FC,$F3,$FE,$B2,$8C,$5F,$CA,$AD,$D6,$B1,$F8
	.dc.b	$ED,$9F,$87,$BA,$F8,$E1,$F2,$B0,$E3,$5F,$E9,$8F,$27,$F3,$1E,$0F
	.dc.b	$F9,$C6,$BF,$03,$92,$1E,$16,$AA,$5F,$13,$FD,$43,$18,$BE,$35,$CE
	.dc.b	$FF,$86,$B6,$BF,$C4,$55,$B9,$CD,$D9,$C9,$D6,$F9,$C6,$5F,$E3,$17
	.dc.b	$E4,$9D,$57,$26,$9C,$A7,$8B,$FC,$63,$F2,$1F,$F7,$7B,$7D,$0F,$F1
	.dc.b	$24,$62,$7F,$DC,$2F,$FB,$B9,$BE,$8E,$A7,$6D,$AE,$8D,$9D,$C7,$53
	.dc.b	$25,$72,$76,$9E,$71,$9B,$FF,$0B,$DF,$8D,$B3,$AA,$E1,$5D,$1A,$7E
	.dc.b	$9B,$D4,$FD,$59,$FC,$43,$E6,$6C,$FE,$CA,$2C,$FD,$EE,$55,$63,$0F
	.dc.b	$DF,$55,$BF,$D6,$C7,$F1,$C9,$C5,$FE,$8E,$BF,$54,$4A,$FD,$0F,$12
	.dc.b	$31,$3F,$A2,$6B,$FA,$F6,$BF,$BC,$6B,$D6,$79,$C7,$FA,$57,$C1,$F0
	.dc.b	$E0,$F3,$7F,$C9,$75,$ED,$3F,$95,$38,$BF,$89,$18,$AE,$33,$9A,$F7
	.dc.b	$CE,$4F,$EB,$BE,$EF,$F0,$C6,$FC,$BF,$F9,$A6,$31,$87,$F2,$6E,$54
	.dc.b	$27,$53,$B6,$D7,$23,$AC,$65,$A6,$5E,$93,$DA,$FF,$CB,$2F,$59,$DC
	.dc.b	$EA,$BA,$B6,$33,$5C,$27,$2D,$8C,$9F,$C1,$AB,$A5,$7E,$E3,$CC,$FE
	.dc.b	$06,$7F,$5B,$F2,$7F,$C6,$35,$FE,$99,$8C,$BF,$D5,$F8,$FF,$D5,$F8
	.dc.b	$7F,$82,$AB,$77,$FB,$6B,$DB,$D6,$FF,$AF,$4E,$AB,$F8,$75,$69,$FE
	.dc.b	$0E,$F8,$7C,$2C,$57,$84,$F0,$46,$5A,$4B,$EC,$7F,$1A,$BE,$D8,$79
	.dc.b	$A7,$FA,$4F,$2F,$F5,$03,$FD,$49,$F9,$7F,$F9,$EE,$FF,$9F,$46,$67
	.dc.b	$FA,$CB,$AF,$1F,$13,$FD,$52,$2E,$7F,$E8,$EB,$E5,$3F,$D9,$27,$6F
	.dc.b	$F4,$D7,$FA,$81,$2F,$F4,$ED,$8F,$E8,$ED,$78,$46,$88,$E9,$B5,$FE
	.dc.b	$9E,$FC,$C1,$F0,$7E,$49,$FE,$87,$18,$7D,$67,$F8,$E9,$C5,$7A,$E3
	.dc.b	$13,$C6,$1C,$0A,$DD,$FF,$80,$F0,$FF,$66,$46,$30,$FE,$0F,$CC,$F9
	.dc.b	$AF,$F1,$09,$FE,$22,$F6,$2C,$FF,$86,$22,$E5,$FC,$4E,$6E,$3F,$E5
	.dc.b	$4B,$E5,$7B,$3F,$C3,$F3,$A7,$BD,$DB,$FF,$2E,$52,$B1,$7F,$E3,$4A
	.dc.b	$B3,$9B,$E9,$7F,$C8,$E7,$37,$CE,$3F,$CB,$3E,$57,$F9,$A1,$FF,$05
	.dc.b	$78,$28,$C4,$7F,$84,$8D,$DF,$F8,$0A,$74,$D9,$DC,$FF,$86,$27,$2F
	.dc.b	$F8,$39,$AF,$68,$DB,$D0,$D3,$88,$DC,$78,$5A,$FD,$8F,$CD,$DF,$0D
	.dc.b	$FF,$00,$5B,$7F,$E7,$A7,$EC,$4E,$E7,$44,$9D,$6F,$67,$F9,$05,$72
	.dc.b	$6D,$75,$9D,$69,$FE,$F0,$AD,$36,$75,$7E,$17,$6F,$F8,$62,$76,$DA
	.dc.b	$CD,$F9,$63,$81,$E4,$2B,$FE,$4C,$C1,$F1,$7F,$14,$93,$F4,$DB,$3F
	.dc.b	$3D,$AF,$DC,$6B,$F2,$DA,$FD,$77,$E8,$7C,$0C,$9F,$BC,$D9,$FB,$4D
	.dc.b	$7F,$0E,$D7,$EB,$35,$FC,$13,$5F,$F5,$3F,$C4,$FD,$D2,$4F,$E2,$DB
	.dc.b	$1F,$87,$6B,$F9,$46,$BF,$83,$6B,$F8,$F7,$F1,$1F,$0F,$27,$F2,$CD
	.dc.b	$9F,$C9,$35,$FC,$FB,$5F,$C7,$35,$FC,$D3,$5F,$F5,$FF,$C1,$7E,$54
	.dc.b	$93,$FA,$36,$C7,$E7,$DA,$FE,$B1,$AF,$E6,$DA,$FE,$9D,$FD,$07,$BF
	.dc.b	$27,$F5,$CD,$9F,$D5,$35,$FD,$BB,$5F,$D3,$35,$FD,$93,$5F,$F6,$BF
	.dc.b	$8B,$FA,$D2,$4F,$EE,$DB,$1F,$B7,$6B,$FC,$11,$AF,$EC,$DA,$FE,$FD
	.dc.b	$FD,$C7,$89,$27,$F8,$33,$67,$F8,$13,$5F,$E1,$ED,$7F,$7C,$D7,$F8
	.dc.b	$53,$5F,$F7,$7F,$BB,$FC,$14,$93,$FC,$4D,$B1,$FC,$3D,$AF,$F1,$C6
	.dc.b	$BF,$C2,$DA,$FF,$17,$7F,$88,$7A,$92,$7F,$8F,$36,$7F,$8D,$35,$FE
	.dc.b	$4E,$D7,$F8,$B3,$5F,$E4,$4D,$5B,$3D,$1F,$E3,$C4,$AF,$E5,$8D,$9F
	.dc.b	$E5,$0D,$7F,$9A,$B5,$FE,$48,$D7,$F9,$83,$FC,$A7,$CE,$93,$FC,$DD
	.dc.b	$B3,$FC,$CD,$AF,$F4,$06,$BF,$CB,$DA,$FF,$3B,$6B,$EA,$F3,$BF,$9B
	.dc.b	$12,$7F,$A2,$36,$3F,$A0,$35,$FE,$9A,$D7,$F9,$E3,$5F,$E9,$0F,$F4
	.dc.b	$1F,$26,$4F,$F4,$F6,$CF,$F4,$B6,$BF,$D6,$1A,$FF,$47,$6B,$FD,$4D
	.dc.b	$AF,$D2,$E9,$FE,$9C,$49,$FE,$B8,$D8,$FE,$B0,$D7,$FB,$2B,$5F,$EA
	.dc.b	$8D,$7F,$B0,$3F,$D6,$7A,$64,$FF,$67,$6C,$FF,$63,$6B,$FD,$C1,$AF
	.dc.b	$F5,$F6,$BF,$DA,$DA,$FE,$2B,$E5,$7F,$B3,$12,$7F,$BA,$36,$3F,$B8
	.dc.b	$35,$FE,$FA,$D7,$FB,$63,$5F,$EF,$0F,$F7,$4F,$C4,$FF,$D0,$71,$FE
	.dc.b	$01,$0E,$1F,$F0,$23,$51,$A3,$91,$27,$93,$A8,$E0,$EE,$32,$D3,$71
	.dc.b	$A2,$3E,$5F,$1E,$4F,$09,$FF,$3F,$CF,$9D,$AF,$9A,$76,$F4,$6B,$B6
	.dc.b	$BB,$9E,$91,$F0,$FE,$72,$47,$76,$CF,$99,$AF,$85,$AF,$C6,$6B,$F0
	.dc.b	$5A,$FC,$CF,$9C,$FA,$04,$9F,$63,$63,$E1,$B5,$F7,$35,$FA,$6D,$7E
	.dc.b	$73,$F6,$FC,$49,$3F,$1D,$B3,$F3,$9A,$FC,$96,$BF,$79,$AF,$D9,$6B
	.dc.b	$F8,$4F,$81,$FC,$41,$27,$EA,$36,$3E,$DB,$5F,$AC,$D7,$F1,$6D,$7F
	.dc.b	$0C,$FE,$4F,$E5,$64,$FD,$F6,$CF,$E1,$9A,$FE,$05,$AF,$E5,$9A,$FE
	.dc.b	$45,$AF,$E7,$3E,$3B,$F4,$04,$9F,$C6,$36,$3F,$26,$D7,$F1,$CD,$7F
	.dc.b	$46,$D7,$F3,$CF,$EA,$FE,$2E,$4F,$E5,$DB,$3F,$9E,$6B,$F9,$96,$BF
	.dc.b	$AE,$6B,$FA,$96,$BF,$B4,$F8,$AF,$DC,$12,$7F,$48,$D8,$FD,$5B,$5F
	.dc.b	$D3,$35,$FD,$DB,$5F,$DB,$3F,$C0,$FC,$E9,$3F,$AF,$6C,$FE,$D9,$AF
	.dc.b	$EC,$5A,$FF,$06,$6B,$FC,$05,$AF,$F0,$CF,$85,$FE,$20,$49,$FD,$E3
	.dc.b	$63,$F8,$1B,$5F,$DF,$35,$FE,$26,$D7,$F8,$73,$FC,$6F,$D1,$93,$FC
	.dc.b	$1D,$B3,$FC,$39,$AF,$F0,$96,$BF,$C7,$9A,$FF,$19,$6B,$FC,$93,$E5
	.dc.b	$3F,$94,$12,$7F,$8A,$36,$3F,$8D,$B5,$FE,$2C,$D7,$F9,$5B,$5F,$E4
	.dc.b	$CF,$F3,$3F,$9C,$93,$FC,$7D,$B3,$FC,$99,$AF,$F2,$16,$BF,$CD,$9A
	.dc.b	$FF,$31,$6B,$FC,$F3,$E7,$3F,$A0,$12,$7F,$96,$36,$3F,$99,$B5,$FE
	.dc.b	$5C,$D7,$FA,$1B,$5F,$E7,$CF,$F4,$BF,$46,$4F,$F3,$76,$CF,$F3,$E6
	.dc.b	$BF,$CE,$5A,$FF,$4E,$6B,$FD,$25,$AF,$F5,$4F,$99,$FE,$B0,$49,$FE
	.dc.b	$88,$D8,$FE,$96,$D7,$FA,$33,$5F,$EB,$6D,$7F,$AB,$3F,$D8,$FE,$66
	.dc.b	$4F,$F4,$F6,$CF,$F5,$66,$BF,$D4,$5A,$FF,$66,$6B,$FD,$85,$AF,$F6
	.dc.b	$CF,$81,$FE,$E0,$49,$FE,$B8,$D8,$FE,$C6,$D7,$FA,$F3,$5F,$EE,$6D
	.dc.b	$7F,$B7,$3F,$DE,$FE,$5E,$4F,$F6,$76,$CF,$F6,$E6,$BF,$DA,$5A,$FF
	.dc.b	$7E,$6B,$FD,$E5,$AF,$F8,$13,$E6,$7F,$C1,$04,$9F,$EE,$8D,$8F,$EF
	.dc.b	$6D,$7F,$BB,$35,$FF,$05,$B5,$FF,$03,$28,$FF,$02,$FE,$27,$5A,$66
	.dc.b	$7F,$C1,$B1,$96,$C6,$5F,$F0,$D4,$65,$B1,$97,$FC,$32,$23,$4E,$1A
	.dc.b	$FF,$87,$9A,$EE,$39,$39,$6B,$97,$D3,$F0,$58,$93,$49,$1A,$24,$FB
	.dc.b	$23,$2D,$8C,$BE,$C8,$CB,$EE,$F0,$FB,$12,$7E,$4B,$63,$E3,$B5,$E1
	.dc.b	$1A,$23,$A6,$BF,$2B,$E0,$BD,$44,$8F,$AA,$D8,$FA,$6D,$78,$46,$88
	.dc.b	$E9,$AF,$D6,$F8,$1F,$8C,$48,$FC,$0B,$63,$EF,$B5,$E1,$1A,$23,$A6
	.dc.b	$BF,$82,$F9,$8F,$A4,$48,$FC,$6B,$63,$F1,$6D,$78,$46,$88,$E9,$AF
	.dc.b	$E3,$BE,$63,$EF,$12,$3F,$32,$D8,$FC,$BB,$5E,$11,$A2,$3A,$6B,$F9
	.dc.b	$AF,$53,$F1,$44,$8F,$D2,$B6,$3F,$46,$D7,$84,$68,$8E,$9A,$FE,$9B
	.dc.b	$D3,$FC,$B1,$23,$F6,$2D,$8F,$D7,$B5,$E1,$1A,$23,$A6,$BF,$B2,$F9
	.dc.b	$6F,$D1,$12,$3F,$7A,$D8,$FD,$DB,$5E,$11,$A2,$3A,$6B,$FB,$EF,$97
	.dc.b	$FD,$71,$23,$F8,$4B,$63,$F8,$3B,$5E,$11,$A2,$3A,$6B,$FC,$2B,$E0
	.dc.b	$BF,$74,$48,$FE,$2A,$D8,$FE,$26,$D7,$84,$68,$8E,$9A,$FF,$16,$F9
	.dc.b	$5F,$E0,$C4,$8F,$E4,$2D,$8F,$E3,$ED,$78,$46,$88,$E9,$AF,$F2,$2F
	.dc.b	$93,$FE,$24,$48,$FE,$5A,$D8,$FE,$56,$D7,$84,$68,$8E,$9A,$FF,$2E
	.dc.b	$F9,$4F,$E3,$C4,$8F,$E7,$2D,$8F,$E6,$ED,$78,$46,$88,$E9,$AF,$F3
	.dc.b	$AF,$92,$FE,$54,$48,$FE,$8A,$D8,$FE,$86,$D7,$84,$68,$8E,$9A,$FF
	.dc.b	$46,$F9,$2F,$E6,$C4,$8F,$EA,$2D,$8F,$E9,$ED,$78,$46,$88,$E9,$AF
	.dc.b	$F5,$2F,$8E,$FE,$84,$48,$FE,$BA,$D8,$FE,$B6,$D7,$84,$68,$8E,$9A
	.dc.b	$FF,$5E,$F8,$2F,$E9,$C4,$8F,$ED,$2D,$8F,$EC,$ED,$78,$46,$88,$E9
	.dc.b	$AF,$F6,$AF,$8B,$FE,$B4,$48,$FE,$EA,$D8,$FE,$E6,$D7,$84,$68,$8E
	.dc.b	$9A,$FF,$76,$F8,$7F,$EC,$C4,$8F,$F0,$0B,$63,$FB,$FB,$5E,$11,$A2
	.dc.b	$3A,$6B,$FD,$F8,$7E,$26,$F8,$9F,$EE,$ED,$9D,$48,$DC,$E8,$91,$A8
	.dc.b	$CB,$67,$94,$72,$6B,$E1,$6B,$FE,$10,$6B,$DD,$AF,$43,$97,$C2,$79
	.dc.b	$89,$3D,$DB,$1A,$8C,$11,$C1,$0F,$14,$70,$7C,$9F,$33,$F0,$89,$3E
	.dc.b	$B6,$C7,$AD,$AF,$08,$D1,$1D,$35,$F8,$1F,$35,$F3,$89,$1F,$1D,$B1
	.dc.b	$F1,$DA,$F0,$8D,$11,$D3,$5F,$99,$F0,$9F,$68,$91,$F4,$DB,$1F,$4D
	.dc.b	$AF,$08,$D1,$1D,$35,$FB,$1F,$09,$F8,$62,$47,$DF,$6C,$7D,$F6,$BC
	.dc.b	$23,$44,$74,$D7,$43,$E1,$BF,$CB,$19,$3E,$13,$FC,$BD,$18,$BB,$7F
	.dc.b	$89,$9B,$7F,$89,$9A,$FE,$7A,$2D,$A6,$C6,$1B,$5F,$9F,$BE,$CD,$3B
	.dc.b	$16,$FF,$2D,$5F,$A8,$24,$7E,$85,$B1,$FA,$16,$BF,$A1,$6B,$D0,$D4
	.dc.b	$71,$3F,$55,$F0,$9F,$B4,$23,$82,$47,$EB,$63,$2D,$8C,$BF,$AF,$8C
	.dc.b	$B6,$3F,$6D,$F0,$7F,$10,$F8,$B3,$63,$DD,$3F,$C1,$A7,$0F,$D5,$8F
	.dc.b	$DC,$4E,$2F,$EB,$D5,$9C,$9C,$B1,$47,$2B,$FE,$04,$FF,$10,$24,$F9
	.dc.b	$09,$1F,$C2,$63,$2D,$8C,$BC,$9A,$FF,$0B,$F8,$3F,$E3,$64,$8F,$E2
	.dc.b	$8D,$8F,$E2,$8D,$7F,$8A,$35,$E5,$1C,$08,$FF,$1B,$F8,$3F,$E4,$E4
	.dc.b	$8F,$E4,$0D,$8D,$46,$08,$E0,$87,$F2,$38,$E0,$FF,$25,$F8,$3F,$E6
	.dc.b	$64,$9F,$E5,$8D,$8F,$E5,$8D,$78,$46,$88,$E9,$AF,$F3,$1F,$83,$FE
	.dc.b	$7E,$48,$FE,$70,$D8,$FE,$70,$D7,$84,$68,$8E,$9A,$FF,$3D,$F8,$3F
	.dc.b	$E9,$64,$8F,$E8,$8D,$8F,$E8,$8D,$78,$46,$88,$E9,$AF,$F4,$9F,$83
	.dc.b	$FE,$AE,$48,$FE,$A0,$D8,$FE,$A0,$D7,$84,$68,$8E,$9A,$E8,$FC,$F1
	.dc.b	$FD,$4F,$E0,$BF,$84,$CF,$F0,$A6,$CF,$E7,$5B,$3C,$0F,$ED,$AE,$35
	.dc.b	$FE,$36,$D9,$FE,$C0,$F3,$7F,$B4,$3F,$DC,$88,$E0,$93,$FD,$AE,$32
	.dc.b	$D8,$CB,$FD,$B6,$32,$FF,$6F,$F8,$2F,$EF,$A4,$9F,$EE,$CD,$8F,$EE
	.dc.b	$CD,$78,$46,$88,$E9,$AF,$F7,$BF,$82,$FF,$04,$92,$3F,$C0,$4D,$8F
	.dc.b	$F0,$13,$5E,$11,$A2,$3A,$6B,$FE,$17,$F7,$FF,$C2,$51,$FE,$0F,$6C
	.dc.b	$7F,$85,$9A,$EC,$46,$DA,$6C,$A6,$DD,$11,$D3,$5D,$1D,$FE,$07,$8B
	.dc.b	$FE,$25,$12,$8C,$C7,$F8,$A6,$31,$3E,$99,$CC,$F0,$6C,$EA,$3B,$4E
	.dc.b	$C8,$E0,$87,$C5,$9E,$3F,$03,$27,$E3,$36,$7E,$33,$5F,$98,$D7,$D5
	.dc.b	$48,$C4,$7A,$5A,$FB,$9F,$97,$F6,$EE,$8E,$3F,$03,$87,$9C,$DA,$F8
	.dc.b	$53,$5E,$58,$7F,$93,$6F,$94,$6A,$31,$7C,$9B,$19,$13,$C1,$E6,$62
	.dc.b	$7C,$73,$C2,$33,$3B,$C1,$7E,$36,$D6,$DB,$5C,$CF,$85,$19,$9E,$04
	.dc.b	$78,$61,$67,$81,$F8,$49,$C5,$F8,$66,$E1,$E6,$AF,$C5,$E1,$E0,$C3
	.dc.b	$F4,$18,$7E,$61,$A7,$E7,$A7,$A3,$67,$84,$73,$8D,$BD,$27,$93,$92
	.dc.b	$1F,$2E,$3F,$5D,$F0,$3A,$8F,$CC,$35,$FC,$EB,$68,$D2,$4F,$8D,$B3
	.dc.b	$EC,$6C,$76,$34,$7A,$7E,$07,$A1,$1C,$12,$7E,$2C,$65,$B1,$93,$F8
	.dc.b	$34,$65,$A7,$C5,$F8,$1F,$41,$1C,$12,$7E,$8C,$65,$B1,$93,$F8,$94
	.dc.b	$65,$A7,$D1,$F8,$1F,$88,$47,$04,$9F,$BB,$19,$6C,$64,$FE,$3D,$19
	.dc.b	$69,$F7,$7E,$07,$E8,$11,$C1,$27,$F1,$31,$96,$C6,$4F,$E5,$51,$96
	.dc.b	$9F,$89,$F8,$1F,$B8,$47,$04,$9F,$CA,$C6,$5B,$19,$3F,$9B,$46,$5A
	.dc.b	$71,$DB,$F6,$DF,$89,$BD,$5F,$E4,$D7,$FA,$6D,$58,$7F,$05,$AD,$70
	.dc.b	$FE,$CE,$31,$7C,$30,$FD,$D4,$62,$75,$8F,$EF,$EF,$8E,$3D,$6B,$4F
	.dc.b	$ED,$23,$37,$E4,$C3,$A7,$07,$F2,$2B,$DC,$E7,$1E,$95,$AC,$3F,$9E
	.dc.b	$56,$AF,$FA,$16,$1D,$1B,$5B,$AF,$E8,$B5,$B7,$FD,$D9,$5B,$6C,$EF
	.dc.b	$2F,$87,$97,$83,$19,$5E,$9A,$D6,$1C,$E1,$87,$36,$CE,$AF,$8E,$5E
	.dc.b	$B9,$E5,$87,$C8,$69,$23,$47,$CE,$9E,$39,$7B,$2A,$D7,$C2,$9A,$F8
	.dc.b	$22,$E5,$EA,$8E,$95,$C5,$A7,$12,$9C,$4B,$1C,$6F,$C7,$3C,$5B,$3C
	.dc.b	$5A,$F1,$6C,$F1,$9E,$04,$9F,$9F,$5F,$06,$B8,$E3,$F2,$4F,$95,$7C
	.dc.b	$6B,$67,$9A,$33,$97,$D1,$78,$65,$F3,$E7,$4F,$E0,$63,$79,$7C,$67
	.dc.b	$78,$F8,$72,$F2,$5F,$A5,$B8,$F8,$63,$07,$8B,$2F,$6E,$5E,$1C,$BB
	.dc.b	$1E,$18,$F5,$B8,$C3,$B5,$71,$C3,$BE,$3E,$48,$F3,$CF,$D8,$C3,$A3
	.dc.b	$72,$F4,$61,$CE,$BF,$59,$18,$9D,$47,$23,$EC,$D7,$26,$CE,$DA,$6F
	.dc.b	$1F,$5C,$65,$B1,$9C,$7E,$1D,$73,$9C,$9E,$EB,$FF,$19,$57,$E3,$4B
	.dc.b	$8F,$F0,$26,$E3,$FC,$41,$B3,$C1,$FB,$A6,$0C,$4F,$17,$A1,$0D,$94
	.dc.b	$FD,$8F,$AA,$F1,$7A,$46,$1B,$29,$D4,$F6,$6B,$E8,$6B,$FC,$31,$AF
	.dc.b	$F0,$86,$BD,$4E,$53,$DA,$33,$3D,$E3,$04,$9F,$E4,$53,$E2,$C3,$F9
	.dc.b	$AB,$72,$FE,$37,$E4,$7F,$97,$67,$37,$FC,$B6,$B5,$7E,$38,$FE,$61
	.dc.b	$5B,$BF,$44,$6C,$93,$FC,$C6,$74,$F1,$3C,$71,$AB,$FE,$8F,$3A,$9F
	.dc.b	$74,$78,$9D,$9C,$CE,$15,$E0,$73,$1C,$9C,$9D,$DE,$46,$5D,$C7,$18
	.dc.b	$CB,$F0,$A3,$89,$E2,$74,$65,$E2,$75,$72,$FF,$3E,$9F,$56,$1F,$4E
	.dc.b	$36,$73,$34,$76,$75,$5E,$28,$C9,$C2,$33,$5D,$A3,$FA,$0D,$76,$C3
	.dc.b	$EA,$C7,$52,$52,$75,$87,$7A,$DC,$E6,$B1,$3E,$3A,$DC,$78,$6B,$71
	.dc.b	$8C,$3C,$D3,$DA,$31,$87,$97,$0F,$3C,$FA,$AB,$71,$88,$E0,$78,$0F
	.dc.b	$75,$72,$7E,$61,$0F,$79,$B9,$E7,$1B,$8F,$7D,$F9,$9D,$5F,$8E,$75
	.dc.b	$7E,$12,$38,$AF,$86,$71,$AF,$81,$87,$C3,$79,$9E,$3A,$F9,$0E,$8F
	.dc.b	$49,$A3,$DA,$67,$0F,$3B,$98,$F0,$B6,$BC,$13,$F1,$9F,$1D,$7E,$BA
	.dc.b	$74,$D7,$F4,$47,$5C,$3F,$29,$5F,$9D,$39,$4E,$DC,$47,$D5,$AE,$EF
	.dc.b	$E6,$F4,$FE,$53,$8F,$FB,$75,$B8,$7F,$A4,$71,$F7,$C6,$23,$FD,$C7
	.dc.b	$18,$79,$10,$FC,$CC,$6C,$8E,$4F,$F0,$CF,$91,$DD,$FF,$15,$B5,$FF
	.dc.b	$13,$B5,$D9,$E3,$71,$26,$92,$34,$43,$CE,$E5,$F4,$F9,$1C,$88,$FB
	.dc.b	$DA,$E8,$C1,$0C,$12,$BF,$A8,$C7,$07,$FA,$AF,$E2,$7F,$CF,$9B,$A7
	.dc.b	$AE,$31,$42,$38,$6C,$6C,$8F,$FB,$0E,$34,$9A,$48,$DC,$27,$C7,$E4
	.dc.b	$F8,$49,$1C,$DA,$EC,$F2,$1D,$5C,$C7,$28,$E0,$43,$81,$EA,$8D,$BC
	.dc.b	$A9,$E4,$FC,$12,$47,$AD,$AF,$23,$D8,$E2,$BE,$5B,$63,$A3,$5F,$92
	.dc.b	$FC,$DD,$24,$7C,$76,$BD,$69,$18,$AF,$AE,$D9,$F4,$35,$FA,$AD,$7E
	.dc.b	$83,$FD,$5E,$79,$C7,$DB,$DA,$BA,$31,$3F,$0F,$E2,$38,$32,$32,$76
	.dc.b	$29,$F7,$63,$4A,$7A,$D7,$66,$D7,$8E,$A5,$72,$AD,$4F,$B2,$31,$3A
	.dc.b	$7A,$46,$5A,$78,$CA,$78,$CE,$27,$57,$CB,$5E,$49,$E9,$3F,$A2,$7F
	.dc.b	$C4,$AF,$13,$FC,$65,$48,$CB,$63,$FC,$56,$D7,$71,$A2,$3A,$6B,$A3
	.dc.b	$8F,$91,$D0,$8E,$08,$78,$9C,$B4,$C9,$FB,$A7,$2D,$3C,$BE,$47,$A0
	.dc.b	$8E,$08,$7B,$9C,$B4,$C9,$FC,$19,$CB,$4F,$83,$F1,$1F,$10,$8E,$08
	.dc.b	$7C,$A7,$2D,$32,$7F,$12,$72,$D3,$E6,$F9,$1F,$40,$8E,$08,$7D,$67
	.dc.b	$2D,$32,$7F,$1E,$72,$D3,$EC,$F9,$1F,$70,$8E,$08,$7E,$09,$CB,$4C
	.dc.b	$9F,$CA,$9C,$B4,$FC,$2F,$8F,$FC,$5B,$C4,$FE,$23,$9A,$B1,$B7,$F1
	.dc.b	$16,$DF,$F2,$EB,$AF,$F2,$96,$D7,$18,$FE,$69,$7D,$2B,$35,$27,$37
	.dc.b	$D5,$A7,$F3,$A9,$D2,$AF,$2D,$7E,$78,$91,$EA,$21,$FA,$47,$2D,$32
	.dc.b	$7F,$4C,$72,$D3,$F4,$FE,$3F,$EA,$88,$E0,$87,$EC,$1C,$B4,$C9,$FD
	.dc.b	$61,$CB,$4F,$D9,$F8,$FF,$B6,$23,$82,$1F,$BC,$72,$D3,$27,$F6,$47
	.dc.b	$2D,$3F,$7F,$E3,$FF,$02,$23,$82,$1F,$C2,$1C,$B4,$C9,$FD,$C1,$CB
	.dc.b	$4F,$E1,$7F,$11,$FC,$38,$8E,$08,$7F,$14,$72,$D3,$27,$F7,$C7,$2D
	.dc.b	$3F,$8B,$F9,$1F,$C6,$88,$E0,$87,$F2,$07,$2D,$32,$7F,$82,$1C,$B4
	.dc.b	$FE,$47,$E4,$7F,$26,$23,$82,$1F,$CB,$1C,$B4,$C9,$FE,$14,$72,$D3
	.dc.b	$8E,$5F,$AA,$8F,$D1,$9F,$A6,$FC,$4F,$FD,$DF,$19,$9F,$CF,$9F,$E1
	.dc.b	$19,$DB,$89,$DE,$BF,$E0,$19,$CC,$FE,$8E,$30,$EC,$EC,$7F,$46,$9F
	.dc.b	$E8,$E9,$A4,$74,$43,$FD,$E2,$E5,$A6,$4F,$F8,$00,$F2,$EE,$EC,$48
	.dc.b	$E2,$D7,$6D,$7F,$C4,$67,$8D,$C3,$FE,$26,$84,$62,$3F,$C7,$91,$87
	.dc.b	$CB,$E0,$FC,$B9,$91,$F1,$E3,$E7,$18,$7E,$31,$CA,$32,$EC,$F9,$CE
	.dc.b	$F0,$FF,$18,$D2,$B7,$5F,$01,$F9,$84,$7C,$64,$34,$53,$DF,$1C,$DC
	.dc.b	$1E,$F8,$E4,$FC,$CF,$88,$FB,$04,$7E,$9B,$5D,$18,$21,$82,$1E,$07
	.dc.b	$83,$E2,$F2,$3F,$08,$47,$EF,$B5,$FB,$AD,$78,$1A,$21,$A2,$9E,$EF
	.dc.b	$23,$F2,$04,$7F,$16,$D7,$F1,$2D,$78,$1A,$21,$A2,$9F,$2B,$C8,$FC
	.dc.b	$E1,$1F,$CB,$B5,$FC,$AB,$5E,$06,$88,$68,$A7,$D6,$F2,$3F,$50,$47
	.dc.b	$F4,$6D,$7F,$42,$D7,$81,$A2,$1A,$29,$F8,$2F,$23,$F6,$84,$7F,$5E
	.dc.b	$D7,$F5,$AD,$78,$1A,$21,$A2,$9F,$8E,$F2,$3F,$80,$11,$FD,$DB,$5F
	.dc.b	$DC,$B5,$E0,$68,$86,$8A,$7E,$6B,$C8,$FE,$18,$47,$F8,$3B,$5F,$E0
	.dc.b	$AD,$78,$1A,$21,$A2,$9F,$A6,$F8,$8F,$E3,$04,$7F,$89,$B5,$FE,$22
	.dc.b	$D7,$81,$A2,$1A,$29,$FB,$2F,$88,$FE,$48,$47,$F8,$FB,$5F,$E3,$AD
	.dc.b	$78,$1A,$21,$A2,$9F,$BE,$F8,$8F,$E6,$04,$7F,$95,$B5,$FE,$52,$D7
	.dc.b	$81,$A2,$1A,$29,$FC,$2B,$E2,$3F,$9E,$11,$FE,$6E,$D7,$F9,$AB,$5E
	.dc.b	$06,$88,$68,$A7,$F1,$6F,$88,$FE,$90,$47,$FA,$1B,$5F,$E8,$2D,$78
	.dc.b	$1A,$21,$A2,$9F,$C8,$BE,$23,$FA,$A1,$1F,$E9,$ED,$7F,$A6,$B5,$E0
	.dc.b	$68,$86,$8A,$7F,$2E,$F8,$8F,$EC,$04,$7F,$AD,$B5,$FE,$B2,$D7,$81
	.dc.b	$A2,$1A,$29,$FC,$EB,$E2,$3F,$B6,$11,$FE,$CE,$D7,$FB,$2B,$5E,$06
	.dc.b	$88,$68,$A7,$F4,$6F,$88,$FE,$F0,$47,$FB,$9B,$5F,$EE,$2D,$78,$1A
	.dc.b	$21,$A2,$9F,$D4,$BE,$23,$FC,$08,$47,$FB,$FB,$5F,$EF,$AD,$78,$1A
	.dc.b	$21,$A2,$9F,$DF,$3C,$97,$57,$FB,$59,$B1,$FE,$00,$01,$CB,$96,$BC
	.dc.b	$E3,$91,$1E,$47,$17,$0F,$16,$BC,$4D,$C3,$E1,$FC,$84,$8C,$B5,$F1
	.dc.b	$35,$F6,$35,$F4,$9E,$97,$0F,$B2,$3B,$3E,$0F,$87,$F8,$44,$70,$43
	.dc.b	$E0,$39,$69,$93,$E6,$39,$69,$EC,$F5,$3E,$71,$1C,$10,$F9,$8E,$5A
	.dc.b	$64,$FB,$0E,$5A,$7C,$8F,$53,$ED,$11,$C1,$0F,$B0,$E5,$A6,$4F,$C2
	.dc.b	$39,$69,$F5,$3E,$1F,$F0,$C4,$70,$43,$F0,$8E,$5A,$64,$FC,$83,$96
	.dc.b	$9F,$80,$F8,$7F,$C9,$11,$C1,$0F,$C8,$39,$69,$93,$F3,$8E,$5A,$7E
	.dc.b	$33,$E1,$FF,$3C,$47,$04,$3F,$38,$E5,$A6,$4F,$D4,$39,$69,$F9,$8F
	.dc.b	$87,$FD,$51,$1C,$10,$FD,$43,$96,$99,$3F,$68,$E5,$A7,$E9,$3E,$1F
	.dc.b	$F6,$C4,$70,$43,$F6,$8E,$5A,$64,$FE,$00,$E5,$A7,$EC,$3E,$1F,$F8
	.dc.b	$11,$1C,$10,$FE,$00,$E5,$A6,$4F,$E1,$8E,$5A,$7E,$F3,$E1,$FF,$87
	.dc.b	$11,$C1,$0F,$E1,$8E,$5A,$64,$FE,$30,$E5,$A7,$F0,$8F,$87,$FE,$34
	.dc.b	$47,$04,$3F,$8C,$39,$69,$93,$F9,$23,$96,$9F,$C5,$3E,$1F,$F9,$31
	.dc.b	$1C,$10,$FE,$48,$E5,$A6,$4F,$E6,$0E,$5A,$7F,$20,$F8,$7F,$E6,$44
	.dc.b	$70,$43,$F9,$83,$96,$99,$3F,$9E,$39,$69,$FC,$B3,$E1,$FF,$9F,$11
	.dc.b	$C1,$0F,$E7,$8E,$5A,$64,$FE,$90,$E5,$A7,$F3,$8F,$87,$FE,$94,$47
	.dc.b	$04,$3F,$A4,$39,$69,$93,$FA,$A3,$96,$9F,$D1,$3E,$1F,$FA,$B1,$1C
	.dc.b	$10,$FE,$A8,$E5,$A6,$4F,$EC,$0E,$5A,$7F,$50,$F8,$1F,$EC,$44,$70
	.dc.b	$43,$FB,$03,$96,$99,$3F,$B6,$39,$69,$FD,$73,$E0,$3F,$B7,$11,$C1
	.dc.b	$0F,$ED,$8E,$5A,$64,$FE,$F0,$E5,$A7,$F6,$8F,$80,$FE,$F4,$47,$04
	.dc.b	$3F,$BC,$39,$69,$93,$FC,$08,$E5,$A7,$F7,$AF,$C4,$DF,$03,$FD,$F2
	.dc.b	$11,$87,$51,$98,$C4,$7F,$82,$DA,$FF,$74,$38,$C6,$C8,$F2,$21,$C5
	.dc.b	$CB,$CF,$E0,$7B,$11,$F3,$35,$D1,$82,$18,$21,$E3,$94,$71,$8C,$7C
	.dc.b	$0F,$AC,$8F,$C3,$6B,$ED,$6B,$C0,$D1,$0D,$14,$F0,$75,$7C,$72,$3F
	.dc.b	$3D,$AF,$C9,$6B,$C0,$D1,$0D,$14,$F6,$7C,$47,$D3,$23,$F6,$DA,$FD
	.dc.b	$56,$BC,$0D,$10,$D1,$4F,$91,$F1,$1F,$7C,$8F,$E1,$DA,$FE,$05,$AF
	.dc.b	$03,$44,$34,$53,$EA,$7C,$17,$E2,$C8,$FE,$4D,$AF,$E3,$5A,$F0,$34
	.dc.b	$43,$45,$3F,$01,$F0,$5F,$97,$23,$F9,$F6,$BF,$99,$6B,$C0,$D1,$0D
	.dc.b	$14,$FC,$67,$C1,$7E,$8C,$8F,$EA,$DA,$FE,$95,$AF,$03,$44,$34,$53
	.dc.b	$F3,$1F,$05,$FA,$F2,$3F,$B7,$6B,$FB,$16,$BC,$0D,$10,$D1,$4F,$D2
	.dc.b	$7C,$17,$EE,$C8,$FF,$03,$6B,$FB,$D6,$BC,$0D,$10,$D1,$4F,$D8,$7C
	.dc.b	$0F,$F0,$72,$3F,$C3,$DA,$FF,$09,$6B,$C0,$D1,$0D,$14,$FD,$E7,$C3
	.dc.b	$FF,$13,$23,$FC,$6D,$AF,$F1,$56,$BC,$0D,$10,$D1,$4F,$E1,$1F,$05
	.dc.b	$FC,$7C,$8F,$F2,$76,$BF,$C8,$5A,$F0,$34,$43,$45,$3F,$8A,$7C,$17
	.dc.b	$F2,$B2,$3F,$CC,$DA,$FF,$2D,$6B,$C0,$D1,$0D,$14,$FE,$41,$F0,$5F
	.dc.b	$CD,$C8,$FF,$3F,$6B,$FC,$E5,$AF,$03,$44,$34,$53,$F9,$67,$C1,$7F
	.dc.b	$43,$23,$FD,$2D,$AF,$F4,$56,$BC,$0D,$10,$D1,$4F,$E7,$1F,$05,$FD
	.dc.b	$3C,$8F,$F5,$76,$BF,$D4,$5A,$F0,$34,$43,$45,$3F,$A2,$7C,$17,$F5
	.dc.b	$B2,$3F,$D8,$DA,$FF,$5D,$6B,$C0,$D1,$0D,$14,$FE,$A1,$F0,$5F,$D9
	.dc.b	$C8,$FF,$6F,$6B,$FD,$A5,$AF,$03,$44,$34,$53,$FA,$E7,$C3,$FF,$73
	.dc.b	$23,$FD,$ED,$AF,$F7,$56,$BC,$0D,$10,$D1,$4F,$EF,$5F,$89,$FF,$80
	.dc.b	$D3,$48,$FF,$81,$DA,$FF,$80,$41,$C0,$D1,$0D,$19,$2B,$C9,$E4,$BE
	.dc.b	$13,$B1,$23,$6D,$76,$D7,$D2,$D7,$C6,$7A,$9C,$3E,$68,$EC,$F8,$BE
	.dc.b	$23,$D8,$47,$04,$3C,$AE,$5A,$64,$F9,$2E,$5A,$7B,$BE,$0F,$E4,$11
	.dc.b	$C1,$0F,$82,$E5,$A6,$4F,$AA,$E5,$A7,$CA,$F8,$3F,$A8,$47,$04,$3E
	.dc.b	$6B,$96,$99,$3F,$02,$E5,$A7,$D8,$F9,$E7,$A1,$F3,$3E,$19,$EA,$7C
	.dc.b	$E9,$CD,$7E,$0A,$72,$FE,$12,$32,$A3,$2A,$1E,$97,$5A,$E9,$76,$B9
	.dc.b	$35,$E0,$DA,$E3,$1E,$19,$E2,$D3,$89,$4E,$25,$8E,$2F,$96,$38,$B4
	.dc.b	$E2,$53,$89,$63,$8D,$FB,$1F,$7B,$6B,$CC,$79,$EE,$D7,$86,$2B,$CE
	.dc.b	$3D,$D9,$BE,$0F,$F6,$04,$9F,$D4,$B5,$FD,$4B,$5F,$DD,$B5,$FD,$1C
	.dc.b	$73,$78,$90,$FD,$D7,$C1,$3D,$7F,$9C,$D8,$F9,$0D,$7E,$06,$7A,$99
	.dc.b	$1A,$7E,$A4,$62,$3E,$44,$F1,$CB,$F0,$2D,$9F,$C5,$1B,$8C,$D5,$8C
	.dc.b	$D7,$52,$1D,$23,$BB,$89,$E2,$D9,$E2,$FD,$9B,$7C,$0F,$BB,$DB,$93
	.dc.b	$F8,$59,$CB,$67,$2F,$AA,$7E,$AC,$48,$F8,$B1,$AB,$FC,$ED,$72,$9F
	.dc.b	$05,$7F,$29,$23,$FC,$89,$AF,$F2,$26,$BF,$CD,$1A,$FD,$A6,$BF,$C6
	.dc.b	$DE,$33,$E0,$7C,$91,$F4,$3D,$DB,$39,$8F,$61,$F7,$23,$38,$75,$8C
	.dc.b	$9C,$4F,$C0,$FC,$07,$E1,$08,$FE,$13,$E0,$2B,$F7,$A2,$DF,$DE,$9B
	.dc.b	$5F,$CD,$34,$56,$2F,$FA,$4C,$2B,$10,$8D,$BE,$23,$C0,$53,$C0,$5B
	.dc.b	$F0,$36,$FC,$17,$D2,$BB,$4F,$E7,$1F,$CF,$FC,$14,$8C,$10,$FC,$CB
	.dc.b	$97,$F3,$71,$FA,$63,$17,$F9,$C9,$C4,$F1,$8C,$3F,$AA,$8C,$BC,$8C
	.dc.b	$10,$C1,$23,$8B,$E2,$3A,$5F,$EF,$34,$FD,$BF,$C2,$49,$FC,$D4,$7F
	.dc.b	$0A,$30,$FE,$CA,$30,$E8,$FD,$EB,$AA,$FC,$CC,$EA,$7F,$BA,$BF,$E0
	.dc.b	$12,$3F,$DE,$9A,$E8,$C1,$0C,$10,$FF,$07,$BC,$23,$BC,$78,$CF,$E0
	.dc.b	$FC,$7E,$18,$CC,$7E,$E6,$7F,$8A,$CE,$6F,$E1,$4F,$1A,$FE,$1B,$3F
	.dc.b	$C6,$B9,$BF,$8D,$F3,$4A,$FE,$37,$19,$7F,$8E,$C6,$5F,$E2,$F1,$97
	.dc.b	$F8,$CC,$66,$BF,$8C,$CE,$5F,$E3,$71,$9B,$FC,$F4,$E5,$FD,$04,$66
	.dc.b	$BF,$3F,$39,$7F,$43,$19,$7F,$41,$1F,$D7,$FF,$24,$75,$52,$4F,$F5
	.dc.b	$B8,$75,$62,$3F,$AB,$47,$07,$53,$CE,$75,$8F,$F7,$2A,$D5,$70,$AD
	.dc.b	$3C,$08,$64,$F0,$57,$2C,$7F,$C5,$F5,$CA,$FA,$5F,$2B,$E0,$48,$CB
	.dc.b	$C8,$FE,$E9,$F8,$A7,$F6,$8B,$E7,$DD,$F0,$4F,$F0,$04,$66,$7F,$C9
	.dc.b	$ED,$58,$27,$84,$74,$E9,$E2,$61,$F8,$B1,$87,$47,$F9,$5A,$BF,$C2
	.dc.b	$EF,$F9,$28,$38,$7F,$C9,$AD,$7C,$73,$FE,$5E,$76,$D8,$D4,$60,$ED
	.dc.b	$19,$23,$82,$1F,$E0,$F8,$F7,$1E,$FA,$78,$1E,$99,$D5,$5C,$7E,$E4
	.dc.b	$DC,$7C,$F3,$6B,$FC,$EF,$7C,$9F,$C5,$78,$7F,$8F,$F0,$E4,$60,$87
	.dc.b	$E3,$1C,$BF,$8E,$8F,$CB,$18,$C7,$F1,$93,$89,$E3,$18,$7F,$31,$19
	.dc.b	$79,$18,$21,$82,$47,$17,$C4,$74,$C3,$F2,$3E,$3F,$E9,$C9,$1F,$A5
	.dc.b	$F1,$EB,$FA,$48,$B5,$FA,$19,$B5,$CA,$BF,$59,$5F,$B1,$9C,$C7,$C8
	.dc.b	$8A,$FE,$42,$2E,$29,$B1,$86,$CF,$E7,$BD,$9F,$DE,$79,$72,$30,$43
	.dc.b	$F7,$2E,$5F,$C1,$47,$E8,$8C,$63,$FB,$99,$C4,$F1,$3F,$31,$B2,$BF
	.dc.b	$24,$48,$FC,$F3,$5F,$F5,$D4,$6D,$C9,$0C,$C6,$DE,$AF,$1A,$FF,$3D
	.dc.b	$CF,$19,$F3,$9F,$C5,$7C,$DF,$2B,$82,$4F,$F1,$18,$CB,$CA,$AE,$7F
	.dc.b	$C3,$27,$FD,$01,$F5,$EF,$4A,$93,$0F,$F3,$E4,$3D,$2C,$47,$F9,$DE
	.dc.b	$38,$3A,$9E,$73,$AC,$7F,$D2,$75,$AA,$E1,$5A,$78,$10,$C9,$E0,$AE
	.dc.b	$58,$7F,$B4,$2B,$95,$F4,$BE,$57,$C0,$91,$97,$91,$EA,$F5,$BD,$FE
	.dc.b	$B4,$8C,$10,$F6,$39,$7D,$D1,$F1,$8C,$4F,$B2,$71,$3C,$63,$0F,$C8
	.dc.b	$8C,$BC,$8C,$10,$C1,$23,$8B,$E2,$3A,$61,$FC,$93,$D6,$FA,$FE,$CC
	.dc.b	$9F,$E7,$B1,$FC,$E8,$C3,$F5,$63,$0E,$8F,$B6,$EA,$BF,$1B,$3A,$9F
	.dc.b	$CB,$BF,$CB,$48,$FF,$39,$6B,$A3,$04,$30,$43,$F4,$EF,$08,$EF,$1E
	.dc.b	$33,$F1,$5A,$F8,$63,$31,$FC,$D6,$7F,$2B,$39,$BF,$85,$3C,$6B,$F1
	.dc.b	$B3,$F9,$AF,$76,$4F,$E6,$1F,$CE,$FB,$B8,$8F,$CC,$45,$AF,$CA,$CD
	.dc.b	$BF,$C7,$4D,$A0,$D5,$59,$8C,$4D,$9C,$C7,$84,$87,$80,$FD,$5C,$EE
	.dc.b	$78,$3F,$B2,$E7,$FD,$9F,$95,$2B,$F6,$31,$97,$F6,$71,$97,$F6,$51
	.dc.b	$97,$F6,$91,$9C,$7F,$63,$39,$7F,$67,$19,$C7,$F5,$93,$97,$F5,$D1
	.dc.b	$9A,$F8,$38,$FF,$80,$C9,$19,$8F,$89,$5F,$19,$FF,$7D,$2E,$3F,$E1
	.dc.b	$B7,$27,$E5,$E0,$8E,$15,$B9,$CC,$62,$B8,$47,$F0,$F8,$ED,$5C,$63
	.dc.b	$EA,$D7,$17,$C3,$5C,$27,$F8,$8D,$70,$3A,$4F,$63,$8C,$FF,$83,$6B
	.dc.b	$94,$7F,$85,$5A,$F0,$3F,$C3,$01,$CB,$C8,$C9,$98,$E8,$47,$31,$C1
	.dc.b	$D1,$88,$D3,$EA,$3C,$C6,$85,$7C,$15,$F9,$23,$E0,$FD,$D1,$E3,$39
	.dc.b	$12,$78,$4E,$E3,$AD,$FC,$99,$E8,$DC,$3E,$04,$70,$BE,$B5,$E0,$F8
	.dc.b	$5B,$81,$A6,$3C,$3B,$3E,$A9,$D3,$67,$4D,$21,$EE,$C3,$F1,$93,$C2
	.dc.b	$30,$7E,$D1,$57,$68,$C9,$E5,$39,$E1,$F0,$23,$4F,$C2,$BB,$7D,$A6
	.dc.b	$D7,$59,$AF,$CE,$8F,$7F,$8F,$F0,$3C,$89,$3E,$D8,$CB,$EF,$8C,$BE
	.dc.b	$4C,$7F,$DD,$ED,$9C,$E3,$F5,$2B,$8C,$E4,$E4,$43,$04,$8E,$0E,$5F
	.dc.b	$BB,$3C,$9C,$C7,$86,$71,$8F,$EC,$3B,$BE,$CF,$8F,$2B,$F0,$71,$F7
	.dc.b	$B5,$39,$8F,$54,$E6,$24,$7D,$D8,$D6,$1F,$8A,$9C,$CF,$EA,$2B,$9B
	.dc.b	$FA,$28,$FD,$81,$1C,$10,$E0,$E5,$A6,$4F,$DB,$39,$9F,$04,$7A,$8F
	.dc.b	$C8,$7C,$47,$8A,$33,$1F,$C5,$A7,$F3,$33,$9C,$3E,$24,$F1,$3E,$C4
	.dc.b	$7E,$7B,$E2,$64,$7E,$6D,$FD,$0F,$93,$88,$FB,$D1,$70,$FD,$7C,$DB
	.dc.b	$FC,$DC,$DA,$FC,$D4,$D7,$F7,$71,$75,$4D,$8C,$36,$7C,$47,$EB,$5F
	.dc.b	$D9,$F8,$FF,$B4,$F2,$24,$FE,$C6,$32,$FE,$CE,$32,$FE,$56,$32,$FE
	.dc.b	$5E,$33,$8F,$F0,$89,$CB,$FC,$2A,$33,$8F,$ED,$A7,$2F,$EE,$23,$35
	.dc.b	$FB,$79,$CB,$FB,$98,$CB,$FC,$6A,$3F,$21,$E9,$FF,$0E,$E7,$FF,$78
	.dc.b	$CF,$F2,$26,$BF,$D3,$A3,$6E,$7C,$9E,$06,$1D,$19,$21,$82,$1F,$CF
	.dc.b	$E3,$FA,$11,$1F,$E6,$ED,$7F,$34,$D7,$81,$A2,$1A,$2E,$3F,$CB,$23
	.dc.b	$C8,$FF,$39,$A4,$63,$1F,$EC,$6D,$AF,$F0,$8C,$E9,$FE,$6F,$19,$7F
	.dc.b	$A9,$3F,$1C,$F9,$1E,$AF,$C9,$F2,$64,$FC,$98,$CB,$F2,$E3,$2F,$F2
	.dc.b	$8A,$4E,$A3,$13,$AC,$3D,$13,$97,$D5,$19,$C3,$FC,$05,$7C,$AF,$C0
	.dc.b	$77,$21,$DE,$F9,$4F,$0A,$E5,$3F,$02,$7F,$05,$C1,$23,$F0,$4F,$E1
	.dc.b	$7C,$8C,$47,$A6,$2E,$3F,$36,$6E,$1E,$49,$B5,$F3,$26,$BF,$E3,$C8
	.dc.b	$B8,$A6,$C6,$1B,$3E,$23,$F2,$8F,$F6,$4F,$C5,$5E,$A7,$E6,$7C,$99
	.dc.b	$5F,$DE,$A7,$2D,$9C,$BF,$46,$79,$44,$8F,$D0,$46,$B2,$FF,$00,$57
	.dc.b	$29,$FE,$BE,$D7,$FC,$64,$1C,$3F,$E3,$56,$BF,$E5,$C6,$BB,$AF,$F0
	.dc.b	$CB,$4F,$F0,$DC,$68,$91,$A6,$C6,$A3,$07,$A5,$F2,$BF,$BB,$F6,$B1
	.dc.b	$5E,$A8,$B9,$FE,$22,$6D,$79,$66,$BF,$E7,$08,$FE,$09,$EE,$7F,$05
	.dc.b	$EE,$93,$FC,$16,$32,$FF,$07,$8C,$BF,$9D,$8C,$BF,$9F,$8C,$E1,$FB
	.dc.b	$39,$CB,$FB,$58,$CE,$1F,$9C,$9C,$BF,$9E,$8C,$D7,$ED,$A7,$2F,$EE
	.dc.b	$23,$2F,$FA,$22,$3F,$C5,$3F,$92,$7F,$B6,$AB,$FC,$D0,$D9,$FF,$47
	.dc.b	$B5,$3F,$EF,$17,$FC,$86,$E4,$9E,$47,$4F,$27,$AB,$27,$6E,$5A,$43
	.dc.b	$FC,$4F,$3C,$E7,$84,$62,$6E,$3F,$E4,$FA,$F1,$4E,$4E,$51,$99,$ED
	.dc.b	$1F,$94,$EC,$F7,$F9,$12,$3F,$45,$1F,$A7,$30,$F8,$A3,$0E,$8F,$7B
	.dc.b	$AA,$FD,$0C,$EA,$7F,$CE,$7B,$23,$F2,$84,$8F,$CC,$B5,$DB,$5F,$D3
	.dc.b	$C7,$17,$64,$36,$59,$F0,$BE,$53,$E4,$FC,$0F,$8E,$33,$1F,$04,$FA
	.dc.b	$F1,$9C,$FF,$65,$76,$7E,$FC,$FE,$13,$E0,$FF,$0B,$E5,$49,$FB,$F1
	.dc.b	$97,$F0,$31,$97,$EC,$C6,$5F,$B7,$19,$C3,$EC,$4E,$5F,$B5,$19,$C3
	.dc.b	$F2,$33,$97,$F2,$71,$9A,$FC,$9C,$E5,$FC,$AC,$65,$FC,$AC,$7E,$73
	.dc.b	$C8,$91,$F9,$67,$F3,$FC,$D8,$8F,$C7,$C5,$C3,$F1,$93,$70,$FD,$1C
	.dc.b	$DA,$FD,$1C,$D7,$F4,$71,$76,$FF,$77,$C7,$43,$C0,$43,$F3,$B3,$B9
	.dc.b	$E0,$FF,$37,$F1,$7F,$67,$E3,$CA,$FE,$6D,$19,$7F,$5F,$19,$7F,$97
	.dc.b	$C6,$5F,$D4,$C6,$71,$FE,$6B,$39,$9F,$C6,$BF,$C9,$C8,$FF,$2C,$6B
	.dc.b	$F9,$A6,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$82,$FC,$3F,$96,$33,$1E
	.dc.b	$83,$26,$CC,$11,$FD,$F4,$7F,$AF,$BE,$79,$EC,$FF,$18,$D2,$5F,$FB
	.dc.b	$52,$72,$D9,$CB,$FC,$32,$79,$44,$8F,$E3,$31,$AB,$FF,$6D,$57,$29
	.dc.b	$FF,$5C,$B6,$3F,$88,$4E,$C4,$C9,$D4,$62,$11,$87,$F9,$5C,$71,$78
	.dc.b	$57,$8D,$F3,$1E,$43,$D4,$7B,$70,$F8,$44,$7E,$07,$93,$5F,$3C,$5C
	.dc.b	$3D,$73,$6F,$C9,$87,$FB,$66,$7D,$F5,$F3,$6B,$57,$EB,$8A,$FA,$A2
	.dc.b	$D7,$1B,$E7,$5C,$6B,$CC,$43,$C6,$79,$8F,$8F,$69,$F8,$DF,$8B,$32
	.dc.b	$FB,$DE,$4C,$BD,$D6,$A7,$FD,$FF,$5B,$9F,$C0,$56,$DF,$F7,$7C,$65
	.dc.b	$B3,$9B,$FB,$B5,$CA,$24,$FF,$85,$8F,$B7,$5C,$5F,$F0,$CB,$5F,$11
	.dc.b	$B0,$E5,$E3,$1F,$E1,$88,$C4,$6A,$30,$47,$94,$68,$91,$E4,$8F,$49
	.dc.b	$F9,$DF,$8D,$C5,$FC,$98,$EE,$7C,$62,$9F,$98,$8D,$E1,$E8,$8E,$66
	.dc.b	$4F,$D4,$7B,$1F,$AD,$E5,$91,$82,$1F,$A2,$72,$FE,$92,$3F,$68,$62
	.dc.b	$FF,$53,$38,$9E,$31,$87,$F6,$D1,$97,$91,$82,$18,$24,$71,$7C,$47
	.dc.b	$4A,$FC,$DF,$91,$FC,$14,$91,$FC,$0F,$C8,$AF,$EE,$A2,$D7,$EF,$A6
	.dc.b	$D7,$2A,$FE,$15,$5F,$C3,$E7,$31,$FB,$D8,$AF,$EF,$22,$EC,$9B,$18
	.dc.b	$6C,$FD,$53,$F0,$FF,$11,$FC,$6F,$C9,$91,$9D,$93,$F9,$E3,$31,$FC
	.dc.b	$4A,$70,$EE,$33,$97,$F1,$89,$CC,$FE,$8A,$7F,$24,$47,$F1,$ED,$7F
	.dc.b	$41,$5E,$07,$24,$33,$1B,$7F,$46,$D7,$F7,$2F,$82,$BF,$29,$1E,$E3
	.dc.b	$F9,$A1,$9C,$B6,$FA,$4C,$94,$F4,$16,$3F,$99,$BF,$CA,$DE,$BF,$92
	.dc.b	$3D,$9C,$57,$F2,$F8,$FF,$79,$CE,$BC,$AC,$DF,$F3,$69,$E0,$FD,$36
	.dc.b	$A9,$E0,$E6,$78,$A9,$DC,$F8,$AF,$31,$E0,$BF,$15,$E2,$65,$F3,$21
	.dc.b	$DE,$7D,$11,$A8,$C1,$E7,$9E,$A6,$6F,$C1,$1C,$9F,$3B,$C0,$F3,$4F
	.dc.b	$93,$C8,$91,$EF,$7C,$FE,$4E,$23,$86,$3F,$EE,$56,$CE,$9E,$04,$32
	.dc.b	$7C,$99,$E5,$7F,$52,$6E,$3F,$4E,$6D,$7D,$19,$AF,$CB,$8B,$7E,$1B
	.dc.b	$E5,$3D,$4F,$D2,$78,$1F,$3B,$C8,$91,$82,$1E,$A7,$2F,$F4,$28,$FE
	.dc.b	$8E,$62,$7F,$13,$38,$9E,$27,$F3,$E9,$FE,$70,$47,$F9,$AB,$5F,$D3
	.dc.b	$B5,$E0,$68,$86,$8B,$3F,$A4,$D5,$3E,$19,$FA,$3F,$07,$6B,$53,$73
	.dc.b	$FC,$E4,$DA,$FE,$AB,$35,$FC,$A4,$7E,$4B,$D8,$FC,$9F,$95,$27,$F3
	.dc.b	$D1,$97,$F4,$11,$97,$F0,$31,$97,$F0,$71,$9C,$3F,$5F,$39,$7F,$63
	.dc.b	$19,$C3,$F6,$13,$97,$F6,$51,$9A,$FD,$84,$E5,$FD,$94,$65,$FD,$64
	.dc.b	$7E,$AB,$C7,$91,$FB,$57,$F5,$BE,$46,$23,$F2,$F1,$70,$FE,$07,$37
	.dc.b	$0F,$E0,$53,$6B,$F8,$04,$D7,$F7,$51,$75,$4D,$8C,$36,$7C,$47,$F0
	.dc.b	$77,$F8,$1F,$E2,$E9,$FE,$0B,$F0,$BF,$ED,$2C,$7F,$90,$B7,$1F,$F2
	.dc.b	$15,$6F,$0F,$E8,$4D,$7F,$71,$3B,$7F,$C9,$6D,$51,$BA,$FE,$5A,$DA
	.dc.b	$FF,$27,$56,$CE,$11,$9A,$FF,$2A,$BD,$DF,$F2,$A1,$D2,$FF,$C3,$4D
	.dc.b	$3F,$C3,$91,$B7,$D3,$1A,$8C,$47,$48,$C1,$1E,$2D,$7F,$C5,$91,$C1
	.dc.b	$EC,$7F,$17,$7C,$CE,$2F,$E4,$FC,$4F,$CE,$A9,$3F,$21,$F9,$F1,$88
	.dc.b	$E5,$16,$78,$C7,$F6,$FB,$7B,$3F,$5B,$79,$39,$8F,$B2,$61,$FA,$71
	.dc.b	$87,$47,$37,$55,$F7,$A7,$53,$C0,$D1,$C0,$FB,$AF,$0C,$BE,$BD,$70
	.dc.b	$AE,$55,$C0,$E4,$70,$3D,$5E,$3F,$E3,$88,$FB,$7C,$8A,$FE,$2A,$2C
	.dc.b	$FE,$4A,$6D,$FE,$1A,$6D,$7E,$3A,$6A,$C5,$38,$53,$63,$0D,$9F,$29
	.dc.b	$E3,$3F,$3B,$C3,$FA,$4F,$1A,$46,$08,$7E,$81,$CB,$F8,$38,$FD,$69
	.dc.b	$8C,$7F,$57,$38,$9E,$27,$E6,$B6,$57,$E5,$49,$1F,$9A,$6B,$F9,$26
	.dc.b	$BC,$5A,$FE,$D9,$AF,$EA,$1F,$03,$E5,$3F,$19,$F1,$78,$BF,$D9,$CD
	.dc.b	$CF,$F6,$73,$6B,$91,$F0,$2B,$F8,$2C,$7F,$08,$F5,$64,$FE,$5D,$FE
	.dc.b	$17,$E4,$62,$3F,$82,$C5,$AF,$E1,$B3,$6F,$F8,$0C,$DA,$FE,$13,$35
	.dc.b	$FB,$91,$75,$4D,$8C,$36,$7C,$47,$EA,$5F,$F1,$0F,$D4,$FF,$B6,$63
	.dc.b	$F0,$D7,$FC,$80,$CF,$93,$23,$F6,$11,$97,$FC,$C1,$3C,$4E,$58,$7F
	.dc.b	$91,$B1,$FE,$4F,$5C,$6B,$FC,$35,$5A,$8E,$51,$8B,$D5,$70,$8E,$E7
	.dc.b	$68,$CC,$FF,$24,$38,$CF,$F9,$52,$BD,$5A,$3E,$1F,$DB,$E5,$4C,$7F
	.dc.b	$9A,$C7,$F3,$E3,$0F,$92,$30,$E8,$F6,$BA,$BF,$E6,$B3,$A9,$FE,$42
	.dc.b	$FF,$29,$23,$FC,$B9,$AE,$8C,$10,$C1,$0F,$D0,$3C,$23,$BC,$78,$CF
	.dc.b	$93,$F0,$5E,$18,$CC,$7F,$2F,$9F,$B7,$39,$C7,$FA,$AC,$59,$0C,$55
	.dc.b	$F8,$2F,$66,$4F,$DC,$7F,$0B,$E5,$E2,$3E,$CC,$5C,$7E,$84,$DC,$7A
	.dc.b	$D7,$5A,$C3,$E0,$C1,$58,$8D,$1D,$CF,$C5,$B9,$84,$D8,$C3,$67,$F2
	.dc.b	$3E,$4F,$E6,$BC,$99,$18,$21,$F9,$07,$2F,$E4,$A3,$F4,$26,$31,$FC
	.dc.b	$54,$E2,$78,$C6,$1F,$CF,$C6,$5E,$46,$08,$60,$91,$C5,$F1,$1D,$31
	.dc.b	$F8,$FF,$04,$F1,$FF,$67,$E4,$CA,$FF,$66,$4E,$5B,$39,$7F,$57,$3F
	.dc.b	$E9,$A9,$D4,$7F,$A1,$DB,$7F,$A5,$9E,$0F,$FA,$EA,$33,$7E,$2B,$FF
	.dc.b	$03,$C7,$2B,$FF,$05,$B5,$D1,$FE,$0E,$3F,$16,$E5,$B1,$C8,$8E,$C4
	.dc.b	$6D,$B1,$A7,$A4,$64,$F2,$BD,$0F,$82,$7F,$10,$7E,$04,$48,$C9,$FC
	.dc.b	$3A,$30,$FC,$38,$B5,$FC,$1E,$7F,$90,$FB,$12,$BF,$8D,$BF,$C9,$2B
	.dc.b	$11,$FC,$62,$2D,$7F,$0F,$9B,$7F,$A0,$D5,$58,$BF,$D0,$55,$7F,$61
	.dc.b	$16,$7D,$B1,$83,$C2,$43,$F9,$3C,$EE,$78,$3F,$ED,$8F,$C9,$3F,$DA
	.dc.b	$F7,$FB,$07,$FD,$AD,$5F,$BE,$6A,$76,$E7,$C8,$E5,$E4,$C8,$FE,$0B
	.dc.b	$97,$FB,$FA,$B7,$1C,$6B,$4F,$FB,$4E,$3A,$9A,$25,$6A,$B9,$57,$9C
	.dc.b	$C5,$F9,$E7,$17,$E0,$8F,$54,$EA,$3D,$53,$A7,$F3,$9B,$7C,$1F,$22
	.dc.b	$47,$E9,$A3,$F4,$46,$1F,$34,$61,$D1,$DC,$F8,$CE,$F0,$FD,$24,$EE
	.dc.b	$7F,$26,$FE,$90,$8F,$E6,$F5,$46,$23,$66,$5C,$3F,$AA,$6B,$FC,$01
	.dc.b	$E1,$3E,$08,$F2,$9F,$37,$E0,$FC,$51,$98,$F8,$27,$A4,$FB,$91,$AC
	.dc.b	$BA,$47,$29,$FB,$53,$F8,$7F,$66,$4F,$1D,$11,$F8,$BF,$2B,$33,$F7
	.dc.b	$62,$DF,$E0,$E6,$E1,$F8,$F9,$B5,$F8,$F9,$AF,$E3,$E2,$DA,$6C,$61
	.dc.b	$B3,$E3,$8F,$13,$E5,$F2,$3F,$3D,$E4,$C8,$C1,$0F,$C9,$39,$7F,$29
	.dc.b	$1F,$A5,$31,$8F,$E6,$E7,$13,$C6,$30,$FC,$38,$CB,$C8,$C1,$0C,$12
	.dc.b	$38,$BE,$23,$A6,$3F,$CF,$39,$FF,$6D,$E4,$49,$FE,$77,$1F,$CD,$CC
	.dc.b	$3F,$AB,$8C,$3A,$3E,$B3,$AA,$FE,$71,$3A,$9F,$E4,$EF,$F3,$82,$3F
	.dc.b	$95,$6B,$A3,$04,$30,$43,$F9,$FB,$C2,$3B,$C7,$8C,$FE,$01,$F0,$3E
	.dc.b	$18,$CC,$7F,$4F,$9F,$E2,$93,$9B,$F8,$53,$C6,$BF,$86,$4F,$F8,$19
	.dc.b	$4F,$F8,$23,$E8,$9E,$C6,$63,$0D,$BF,$F0,$CC,$E9,$FD,$13,$FE,$EB
	.dc.b	$30,$AB,$AB,$C2,$79,$29,$DC,$EA,$31,$EC,$EA,$73,$7E,$29,$EF,$79
	.dc.b	$AD,$32,$32,$79,$23,$0F,$8E,$79,$34,$F2,$1B,$38,$99,$3C,$D3,$CA
	.dc.b	$FC,$D7,$EF,$F2,$BC,$DE,$54,$BF,$87,$1E,$93,$0F,$39,$4E,$63,$67
	.dc.b	$CA,$77,$5D,$AB,$53,$C6,$73,$87,$D7,$9C,$D7,$92,$73,$5F,$5A,$73
	.dc.b	$5F,$76,$73,$3F,$46,$7E,$9F,$E3,$0F,$91,$FA,$1E,$5C,$AF,$E3,$D3
	.dc.b	$96,$CE,$5F,$64,$F2,$89,$1F,$84,$8D,$63,$F7,$6B,$94,$FC,$D6,$BF
	.dc.b	$E5,$80,$E1,$FF,$2D,$B5,$FF,$34,$B5,$DD,$7F,$87,$5A,$7F,$87,$E3
	.dc.b	$44,$8D,$36,$35,$18,$3D,$2F,$95,$FC,$97,$C1,$E2,$BD,$51,$72,$FD
	.dc.b	$44,$DA,$F2,$CD,$7F,$3D,$1F,$9B,$F6,$64,$7E,$A1,$FC,$F7,$9B,$88
	.dc.b	$FC,$34,$5C,$3F,$1F,$37,$0F,$DB,$4D,$AF,$D9,$CD,$7F,$59,$17,$54
	.dc.b	$D8,$C3,$67,$C4,$7E,$E9,$FD,$87,$97,$FB,$1F,$2E,$4F,$EF,$A3,$2F
	.dc.b	$F0,$08,$CB,$F9,$D8,$CB,$F9,$F8,$CE,$3F,$A8,$9C,$BF,$AA,$8C,$E3
	.dc.b	$FC,$46,$72,$FF,$13,$8C,$D7,$F1,$09,$CB,$FC,$4A,$32,$FF,$0B,$8F
	.dc.b	$CD,$DF,$F0,$9A,$93,$F9,$A8,$CB,$FC,$62,$32,$FE,$4E,$32,$FE,$DE
	.dc.b	$33,$7F,$99,$9C,$CF,$E5,$F4,$F9,$AF,$E3,$E3,$F1,$A4,$70,$43,$FC
	.dc.b	$F1,$1C,$1B,$1B,$69,$B3,$B3,$E4,$9F,$14,$7F,$84,$7E,$05,$F0,$59
	.dc.b	$C3,$F4,$F7,$6B,$FD,$9F,$5A,$7E,$11,$0F,$80,$73,$AF,$E6,$B0,$F6
	.dc.b	$E9,$27,$C1,$E5,$D7,$F9,$34,$ED,$F2,$4F,$29,$D6,$3F,$CF,$6B,$55
	.dc.b	$C2,$B4,$F0,$21,$93,$B9,$89,$E5,$1D,$31,$FF,$6F,$D2,$F1,$7F,$02
	.dc.b	$AA,$9D,$3F,$27,$C9,$F7,$F9,$52,$30,$43,$DA,$E5,$F7,$C7,$C6,$31
	.dc.b	$8F,$B6,$71,$3C,$63,$0F,$D7,$8C,$BC,$8C,$10,$C1,$23,$8B,$E2,$3A
	.dc.b	$63,$FC,$93,$93,$EB,$F9,$52,$7F,$9B,$C7,$F3,$C3,$0F,$D6,$8C,$3A
	.dc.b	$3F,$14,$EA,$BF,$9B,$4E,$A7,$F2,$EF,$F2,$82,$3F,$CB,$5A,$E8,$C1
	.dc.b	$0C,$10,$FD,$43,$C2,$3B,$C7,$8C,$FC,$5E,$3E,$18,$CC,$7F,$47,$9F
	.dc.b	$CA,$CE,$6F,$E1,$4F,$1A,$FC,$74,$FE,$6B,$D8,$93,$FA,$77,$F3,$BE
	.dc.b	$5E,$23,$F3,$31,$6B,$F2,$D3,$6F,$F1,$B3,$6B,$F5,$73,$5F,$C6,$45
	.dc.b	$B4,$D8,$C3,$67,$C4,$7E,$E5,$FD,$7F,$95,$FB,$0F,$2A,$4F,$EF,$63
	.dc.b	$2F,$EF,$E3,$2F,$EC,$23,$2F,$EC,$A3,$38,$FE,$BE,$72,$FE,$C6,$33
	.dc.b	$8F,$EA,$27,$2F,$EA,$A3,$35,$FC,$22,$72,$FF,$0A,$8C,$BF,$AB,$8F
	.dc.b	$F1,$DF,$E1,$4E,$EF,$F1,$EF,$E4,$9F,$EE,$03,$FD,$C3,$19,$AF,$F5
	.dc.b	$43,$5F,$F1,$74,$6F,$CB,$E7,$52,$7F,$D5,$D3,$97,$8C,$65,$FF,$04
	.dc.b	$0A,$E8,$FF,$82,$DA,$88,$E1,$B1,$FE,$09,$8C,$47,$38,$C1,$1E,$E6
	.dc.b	$4B,$1C,$09,$1E,$5C,$BF,$CB,$91,$EF,$7F,$CB,$D4,$F8,$3D,$67,$E5
	.dc.b	$C3,$A1,$FE,$51,$BE,$11,$9A,$E1,$3F,$08,$86,$F0,$FA,$37,$F4,$BD
	.dc.b	$A9,$3F,$29,$FA,$BE,$6E,$23,$E8,$C5,$CB,$CB,$37,$1F,$04,$DA,$F9
	.dc.b	$58,$AB,$10,$84,$E1,$D4,$78,$48,$78,$0F,$BB,$3B,$9E,$0F,$E2,$BC
	.dc.b	$8F,$C5,$F9,$32,$BF,$07,$19,$7F,$0B,$19,$7F,$17,$19,$7F,$1B,$19
	.dc.b	$C7,$E8,$CE,$5F,$A7,$19,$C7,$E7,$4E,$5F,$A1,$19,$AF,$83,$39,$7E
	.dc.b	$1C,$65,$F8,$51,$FA,$4F,$4F,$F4,$7E,$4C,$9F,$D0,$C6,$5F,$CE,$C6
	.dc.b	$5F,$C0,$46,$5F,$D2,$C6,$70,$FC,$5C,$E6,$7F,$3C,$FE,$64,$8F,$E5
	.dc.b	$1A,$FE,$9B,$64,$62,$38,$9B,$21,$B7,$8C,$F8,$23,$D2,$7E,$D3,$E0
	.dc.b	$BC,$D1,$98,$F4,$99,$36,$60,$8F,$E8,$A3,$F8,$3F,$B1,$27,$F7,$EF
	.dc.b	$F0,$AF,$2F,$11,$FC,$1A,$2D,$7E,$AA,$6E,$9F,$A5,$9B,$5F,$96,$9A
	.dc.b	$FE,$56,$2E,$E9,$B1,$86,$CF,$88,$FE,$22,$FF,$B6,$3F,$24,$FF,$69
	.dc.b	$C7,$F0,$B5,$E5,$E4,$D7,$97,$27,$F8,$86,$5F,$ED,$DA,$DC,$63,$2F
	.dc.b	$E0,$11,$88,$FF,$68,$3D,$93,$4D,$4F,$3B,$ED,$3C,$2F,$C4,$F5,$3C
	.dc.b	$D1,$A9,$F1,$4E,$A7,$BC,$E6,$FC,$73,$9A,$ED,$5C,$CF,$1C,$7F,$40
	.dc.b	$E1,$F8,$9E,$5C,$9F,$E7,$B1,$97,$86,$3F,$EE,$06,$CE,$5F,$E3,$D3
	.dc.b	$C5,$F8,$31,$9B,$FE,$53,$3F,$CD,$49,$3F,$CC,$1A,$FE,$75,$AF,$F4
	.dc.b	$16,$BF,$CB,$77,$FE,$26,$EC,$E4,$43,$C6,$7A,$8F,$4F,$C1,$78,$F2
	.dc.b	$E5,$56,$7F,$05,$35,$FE,$A5,$17,$14,$FE,$27,$DA,$95,$F1,$1F,$C6
	.dc.b	$79,$B8,$8F,$C0,$C5,$CB,$F1,$13,$71,$FC,$3C,$DA,$FC,$94,$D7,$F0
	.dc.b	$F1,$70,$4D,$8C,$36,$7C,$47,$D6,$7F,$3D,$E5,$7E,$7F,$C7,$93,$F6
	.dc.b	$A3,$2F,$DC,$8C,$BF,$98,$8C,$BF,$9A,$8C,$E3,$FA,$19,$CB,$FA,$38
	.dc.b	$CE,$3F,$A2,$9C,$BF,$A4,$8C,$D7,$EB,$E7,$2F,$EC,$63,$2F,$E9,$A3
	.dc.b	$FB,$55,$2B,$FB,$67,$E0,$EE,$DF,$DD,$79,$73,$0F,$CE,$4E,$5B,$39
	.dc.b	$7F,$67,$3D,$2B,$9C,$D9,$FF,$06,$57,$17,$67,$E8,$2B,$4A,$35,$7E
	.dc.b	$58,$E3,$5F,$E0,$C7,$4D,$9E,$91,$FE,$05,$32,$7F,$84,$9C,$BE,$43
	.dc.b	$11,$C4,$8E,$08,$7F,$8A,$23,$8B,$E2,$3E,$09,$0F,$E2,$7F,$05,$EA
	.dc.b	$8F,$4E,$1D,$E5,$7E,$B7,$CA,$70,$39,$39,$9E,$31,$DC,$D1,$CA,$7F
	.dc.b	$C4,$5F,$54,$F6,$FE,$4B,$67,$F8,$93,$5F,$E4,$F3,$A9,$91,$A7,$F9
	.dc.b	$C4,$62,$3B,$1F,$E4,$39,$E5,$7E,$96,$CF,$F3,$43,$84,$66,$AC,$66
	.dc.b	$BB,$10,$EA,$2B,$9B,$FD,$1E,$38,$BE,$19,$AF,$F3,$E8,$F7,$F9,$5F
	.dc.b	$03,$CB,$93,$F0,$23,$2F,$C2,$8C,$BE,$0B,$FF,$76,$E1,$EB,$AC,$BE
	.dc.b	$DA,$E3,$5B,$9C,$DF,$D0,$9C,$D7,$1A,$EC,$43,$81,$E3,$30,$F5,$70
	.dc.b	$7C,$2C,$7F,$DE,$0F,$D8,$F2,$BE,$D9,$F8,$BF,$22,$46,$48,$71,$72
	.dc.b	$FB,$63,$8B,$FB,$88,$CB,$FA,$59,$EE,$FE,$76,$3F,$54,$47,$F5,$0D
	.dc.b	$74,$60,$86,$08,$75,$7C,$75,$F9,$B7,$F2,$BF,$07,$8C,$7F,$87,$C5
	.dc.b	$AF,$C7,$4D,$7F,$81,$45,$CD,$3F,$9E,$F5,$E5,$7E,$79,$FD,$0F,$99
	.dc.b	$88,$FB,$D1,$71,$FC,$C4,$DC,$7F,$53,$36,$BF,$3D,$35,$FC,$A4,$5C
	.dc.b	$13,$63,$0D,$9F,$11,$FB,$07,$F6,$7E,$5F,$ED,$3C,$B9,$3F,$B4,$8C
	.dc.b	$BF,$B6,$8C,$BF,$95,$8C,$BF,$97,$8C,$E3,$FB,$19,$CB,$FB,$38,$CE
	.dc.b	$3F,$C1,$27,$2F,$F0,$68,$CD,$7E,$EE,$72,$FE,$F6,$32,$FE,$CE,$3F
	.dc.b	$8D,$7C,$C7,$F0,$CF,$32,$4F,$F4,$28,$CB,$FC,$3A,$32,$FE,$6E,$32
	.dc.b	$FE,$BE,$33,$87,$F1,$D9,$CC,$FE,$4D,$FE,$74,$47,$F9,$C3,$5F,$E6
	.dc.b	$4D,$78,$1A,$21,$A2,$C7,$77,$D0,$7F,$24,$3C,$AE,$5F,$41,$93,$66
	.dc.b	$08,$FF,$21,$7F,$AB,$F3,$3D,$AA,$48,$FF,$21,$43,$CD,$C4,$7F,$20
	.dc.b	$8E,$0E,$A7,$9C,$EA,$BF,$A3,$56,$AB,$85,$69,$E0,$43,$27,$82,$B9
	.dc.b	$67,$FE,$E7,$AE,$57,$D2,$F9,$5F,$02,$46,$5E,$47,$AB,$CB,$F7,$F7
	.dc.b	$C8,$C1,$0F,$63,$97,$DD,$1F,$18,$C4,$FB,$27,$13,$C6,$30,$FC,$88
	.dc.b	$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$06,$7C,$9F,$F7,$19,$F9,$BE
	.dc.b	$67,$DA,$3F,$C2,$31,$9A,$E1,$48,$C3,$CF,$25,$71,$8F,$F4,$04,$E9
	.dc.b	$E1,$59,$7F,$01,$3D,$A2,$47,$F8,$5C,$F2,$BF,$E1,$D9,$FF,$03,$47
	.dc.b	$F8,$6D,$AF,$91,$AF,$03,$91,$E0,$2B,$A7,$07,$77,$04,$8F,$01,$E4
	.dc.b	$8F,$EE,$B1,$EB,$A7,$C0,$67,$0F,$F0,$D4,$FB,$CD,$46,$6F,$CC,$5B
	.dc.b	$F9,$11,$FA,$8F,$66,$47,$E9,$9F,$D5,$F9,$B8,$8F,$D3,$45,$C3,$F4
	.dc.b	$73,$6F,$F3,$93,$6B,$EA,$4D,$58,$A7,$0A,$6C,$61,$B3,$E3,$3F,$72
	.dc.b	$FE,$FB,$CA,$FD,$FF,$95,$27,$F7,$B1,$97,$F7,$F1,$97,$F7,$D1,$97
	.dc.b	$F8,$04,$67,$1F,$DE,$CE,$5F,$DF,$C6,$71,$FD,$C4,$E5,$FD,$D4,$66
	.dc.b	$BF,$25,$39,$7F,$29,$19,$7E,$0C,$7E,$73,$1F,$F3,$6F,$E4,$9F,$EE
	.dc.b	$4B,$FC,$EB,$6B,$F4,$B3,$B7,$FD,$B5,$19,$4D,$D3,$E6,$C7,$E8,$27
	.dc.b	$AF,$97,$23,$04,$3F,$21,$1F,$9D,$23,$F9,$66,$BA,$30,$43,$04,$3F
	.dc.b	$48,$F0,$CB,$F2,$6F,$A6,$3F,$C9,$B6,$F8,$2C,$56,$A7,$C8,$FF,$9C
	.dc.b	$E7,$51,$8C,$7F,$5D,$7C,$1F,$8F,$7F,$2B,$D3,$95,$F1,$9F,$9B,$E6
	.dc.b	$62,$3C,$31,$71,$ED,$36,$FE,$8C,$DA,$FE,$2D,$8F,$FC,$01,$78,$9D
	.dc.b	$48,$7B,$2B,$83,$F5,$26,$CF,$1B,$F6,$46,$0F,$2C,$FD,$CF,$27,$F0
	.dc.b	$7E,$4C,$8C,$10,$F9,$2E,$5F,$97,$1F,$8A,$31,$5F,$26,$71,$3C,$63
	.dc.b	$0F,$E2,$A3,$2F,$23,$04,$30,$48,$E2,$F8,$8E,$98,$FF,$21,$F2,$FF
	.dc.b	$37,$E5,$49,$FE,$63,$1F,$CE,$8C,$3F,$82,$8C,$3A,$3F,$3A,$EA,$BF
	.dc.b	$8D,$CE,$A7,$F9,$33,$FC,$CC,$8F,$F2,$C6,$BA,$30,$43,$04,$3F,$9D
	.dc.b	$3C,$23,$BC,$78,$CF,$CB,$FC,$17,$86,$33,$1F,$CC,$A7,$F7,$33,$9B
	.dc.b	$F8,$53,$C6,$BF,$35,$3F,$BE,$F6,$3F,$7F,$E6,$4A,$FD,$DC,$65,$FD
	.dc.b	$EC,$65,$FD,$3C,$65,$FD,$4C,$66,$BF,$4D,$39,$7F,$51,$19,$BF,$E1
	.dc.b	$73,$97,$F8,$6C,$66,$BF,$3B,$39,$7F,$3F,$19,$7F,$86,$C7,$F7,$0A
	.dc.b	$FF,$73,$7D,$6F,$FB,$12,$4D,$31,$B7,$F6,$F7,$1E,$5F,$1C,$7F,$BC
	.dc.b	$C6,$9C,$57,$F8,$0E,$36,$FF,$81,$6F,$BD,$7F,$26,$AD,$CF,$82,$2D
	.dc.b	$70,$8F,$E1,$B3,$89,$91,$CE,$3C,$31,$CE,$30,$78,$5D,$BF,$EE,$29
	.dc.b	$F3,$3A,$1F,$81,$BE,$63,$C3,$E6,$49,$D4,$E5,$B3,$97,$A4,$FA,$A7
	.dc.b	$51,$FE,$F0,$6E,$3F,$3A,$78,$2C,$7F,$C4,$33,$A6,$DF,$F8,$3A,$FC
	.dc.b	$AD,$9E,$B1,$FE,$28,$32,$7F,$85,$1C,$BC,$8C,$47,$12,$38,$21,$FE
	.dc.b	$25,$8E,$2F,$F8,$ED,$F2,$36,$3D,$C7,$D6,$F7,$EC,$7C,$42,$1F,$3D
	.dc.b	$CB,$F0,$E2,$DF,$E4,$27,$EF,$FB,$5F,$80,$F3,$A5,$FD,$5C,$D3,$A8
	.dc.b	$C4,$E9,$FB,$91,$97,$EF,$46,$6F,$F3,$B3,$97,$F3,$F1,$9C,$BF,$33
	.dc.b	$39,$7F,$37,$19,$AF,$CD,$4E,$5F,$CE,$46,$5F,$D3,$C7,$E6,$3C,$A9
	.dc.b	$1E,$97,$F3,$7E,$56,$23,$F2,$51,$70,$FD,$8C,$DC,$3F,$57,$36,$BF
	.dc.b	$55,$35,$FD,$AC,$5C,$7F,$DC,$F1,$DC,$87,$63,$E5,$4E,$E7,$83,$F9
	.dc.b	$7E,$FF,$D7,$F9,$52,$BF,$3F,$19,$7E,$94,$65,$FC,$E4,$65,$FD,$64
	.dc.b	$67,$1F,$CE,$47,$E2,$67,$4F,$E5,$F6,$FF,$22,$3F,$97,$8F,$CE,$11
	.dc.b	$C1,$0F,$CA,$39,$69,$93,$A3,$99,$F0,$C7,$EF,$23,$CD,$1F,$C7,$BE
	.dc.b	$0A,$E7,$FB,$09,$B5,$FC,$5E,$6B,$C4,$F8,$67,$E4,$DF,$E2,$BE,$97
	.dc.b	$F1,$6F,$2E,$4F,$E5,$E3,$2F,$E6,$63,$2F,$F1,$18,$CB,$FC,$4E,$33
	.dc.b	$87,$F3,$29,$CB,$FC,$D2,$33,$87,$F2,$E9,$CB,$FC,$C2,$33,$5F,$CB
	.dc.b	$E7,$2F,$F3,$18,$CB,$FC,$F2,$3F,$97,$FE,$49,$FE,$D6,$AF,$F5,$8B
	.dc.b	$67,$FD,$74,$D4,$FF,$BB,$5F,$E5,$0E,$49,$E5,$74,$F2,$FA,$B2,$76
	.dc.b	$E5,$A4,$18,$FD,$8A,$E9,$5C,$67,$37,$E6,$BC,$5E,$EA,$CE,$6B,$A4
	.dc.b	$E5,$FE,$73,$E5,$FC,$0F,$2E,$46,$08,$76,$3E,$0B,$A7,$F9,$D4,$7F
	.dc.b	$3D,$31,$3E,$79,$C4,$F2,$3F,$3D,$3F,$CD,$C8,$FF,$35,$6B,$FC,$BD
	.dc.b	$AF,$03,$44,$34,$59,$FC,$EE,$C9,$F0,$CF,$A4,$FA,$FF,$07,$8A,$DC
	.dc.b	$DC,$3E,$FC,$DA,$FE,$7D,$35,$E2,$7D,$EF,$6F,$F0,$FE,$6C,$8C,$10
	.dc.b	$F0,$6B,$FE,$F6,$7F,$0B,$1F,$8E,$99,$18,$21,$C1,$CB,$CA,$3B,$39
	.dc.b	$3B,$63,$F8,$58,$CB,$F8,$79,$E2,$E6,$7A,$B9,$7F,$2F,$E6,$7E,$80
	.dc.b	$8F,$E6,$B8,$AB,$EA,$8B,$3F,$93,$9B,$9F,$E9,$A6,$D7,$E9,$A6,$BF
	.dc.b	$A6,$8B,$8F,$F1,$E8,$E8,$78,$88,$7C,$89,$DC,$F0,$7F,$C4,$7F,$81
	.dc.b	$FF,$7D,$C3,$F4,$9E,$6F,$EE,$0C,$DC,$3F,$66,$E5,$FF,$0E,$4F,$1C
	.dc.b	$51,$23,$0F,$EF,$A7,$83,$F8,$98,$CA,$7F,$C2,$4D,$3F,$C2,$8E,$08
	.dc.b	$E0,$87,$F8,$54,$F0,$18,$30,$F0,$74,$75,$2B,$E2,$23,$FC,$63,$DE
	.dc.b	$F8,$0F,$9C,$B7,$FC,$26,$2D,$FC,$09,$E8,$6C,$FE,$33,$1F,$E5,$0C
	.dc.b	$1E,$D5,$F3,$65,$7F,$91,$67,$2D,$9C,$BF,$AF,$8F,$F7,$C1,$A7,$11
	.dc.b	$AC,$3F,$A9,$CE,$A3,$C0,$67,$0F,$34,$66,$BB,$1E,$1B,$D3,$F6,$A3
	.dc.b	$B5,$F4,$BC,$CA,$79,$3E,$5F,$2B,$DC,$47,$D9,$E5,$D7,$47,$BE,$3E
	.dc.b	$24,$66,$FD,$31,$71,$F3,$CD,$AF,$24,$D7,$C5,$16,$D3,$63,$0D,$9F
	.dc.b	$CB,$FC,$B7,$D3,$F2,$E4,$60,$87,$CF,$72,$FE,$5E,$3F,$4E,$63,$1F
	.dc.b	$9F,$38,$9E,$27,$F9,$9F,$55,$7E,$68,$91,$F9,$06,$BF,$96,$6B,$C5
	.dc.b	$AF,$E3,$63,$9B,$C4,$87,$94,$F4,$D7,$E0,$BE,$0F,$C9,$19,$24,$65
	.dc.b	$DE,$7D,$6F,$73,$99,$FC,$87,$9F,$F9,$9F,$36,$47,$E5,$E3,$F3,$86
	.dc.b	$1F,$CB,$46,$1D,$1F,$9D,$75,$3F,$96,$9D,$4F,$03,$47,$03,$C2,$60
	.dc.b	$D1,$82,$47,$E8,$27,$AD,$FE,$7A,$BF,$61,$E5,$4B,$FD,$73,$FB,$3F
	.dc.b	$2F,$11,$FA,$D8,$B8,$7E,$AA,$6D,$FE,$A2,$6D,$7E,$96,$6B,$FA,$38
	.dc.b	$BB,$A6,$C6,$1B,$3E,$23,$F8,$03,$FC,$CB,$C6,$FE,$0D,$72,$7F,$96
	.dc.b	$C6,$5F,$E0,$D1,$97,$F9,$54,$65,$FE,$0D,$19,$C3,$FC,$AB,$AA,$B5
	.dc.b	$5F,$95,$7F,$96,$91,$FE,$46,$D7,$F9,$4B,$5E,$06,$88,$68,$B3,$E0
	.dc.b	$8F,$23,$FE,$66,$FC,$91,$F0,$58,$C3,$F3,$95,$6F,$FA,$4D,$69,$F8
	.dc.b	$24,$3D,$E7,$FB,$AD,$D3,$FE,$13,$5E,$DF,$FB,$F9,$79,$94,$C4,$49
	.dc.b	$E5,$19,$7A,$46,$5E,$91,$97,$B4,$67,$1F,$64,$E5,$F7,$46,$70,$FE
	.dc.b	$89,$7E,$2B,$EE,$76,$21,$D8,$C9,$D8,$F7,$79,$5E,$B2,$3F,$07,$1A
	.dc.b	$FC,$18,$B8,$7C,$F9,$B7,$E2,$9B,$5F,$3A,$6A,$BF,$F7,$0B,$E0,$3B
	.dc.b	$14,$EC,$5B,$EC,$DB,$EC,$D3,$B1,$FE,$06,$F3,$7E,$A7,$93,$23,$04
	.dc.b	$3E,$D3,$97,$FC,$0D,$1F,$E1,$63,$15,$F8,$B9,$C4,$F1,$3F,$98,$4F
	.dc.b	$F8,$20,$8F,$F7,$E6,$BF,$E0,$46,$BC,$0D,$10,$D1,$67,$F5,$71,$E0
	.dc.b	$8F,$C8,$7B,$56,$73,$1E,$B7,$F4,$11,$9B,$FE,$75,$56,$BF,$3F,$3F
	.dc.b	$97,$F6,$3F,$31,$E5,$C9,$FC,$F4,$65,$FD,$04,$65,$FD,$04,$65,$FD
	.dc.b	$14,$67,$0F,$D8,$4E,$5F,$D9,$46,$70,$FC,$AC,$E5,$FC,$BC,$66,$BF
	.dc.b	$67,$39,$7F,$6B,$19,$7F,$11,$1F,$AF,$F2,$64,$7E,$D5,$FD,$97,$95
	.dc.b	$88,$FD,$B4,$5C,$3F,$82,$4D,$C3,$F4,$F3,$6B,$F8,$1C,$D7,$F2,$D1
	.dc.b	$74,$4D,$8C,$36,$7C,$47,$F0,$77,$FC,$6D,$7F,$EF,$8F,$C0,$9E,$96
	.dc.b	$31,$FF,$1D,$7E,$49,$FE,$D1,$C3,$FD,$3C,$DB,$E1,$5B,$9F,$F2,$FD
	.dc.b	$F2,$7F,$CC,$32,$BE,$7E,$66,$E5,$87,$39,$FF,$24,$5F,$19,$FF,$56
	.dc.b	$36,$F9,$47,$F9,$B2,$B8,$36,$FA,$56,$EF,$9E,$1F,$E5,$EA,$4F,$57
	.dc.b	$FC,$AC,$62,$35,$19,$8E,$27,$F9,$99,$C3,$EE,$8D,$91,$D3,$5F,$F3
	.dc.b	$94,$70,$7B,$1F,$12,$7A,$BE,$E9,$F8,$96,$F8,$4F,$5C,$49,$E0,$78
	.dc.b	$4A,$7C,$A8,$D3,$CC,$CD,$FC,$37,$E4,$47,$C6,$EC,$F7,$92,$3D,$CD
	.dc.b	$7D,$2D,$7E,$3C,$A3,$11,$F6,$1A,$FE,$09,$F9,$7F,$1D,$23,$E5,$35
	.dc.b	$F8,$AD,$7A,$B5,$FC,$1B,$5F,$C6,$B5,$FB,$7D,$FF,$60,$8F,$D7,$6B
	.dc.b	$F4,$9A,$FA,$9A,$FE,$41,$AF,$E6,$9F,$C1,$F8,$1F,$53,$CE,$93,$F3
	.dc.b	$23,$F1,$E6,$1F,$D2,$C6,$1D,$1F,$5B,$34,$6F,$6F,$D8,$56,$EB,$F0
	.dc.b	$EF,$81,$FC,$83,$5C,$98,$21,$82,$1F,$59,$E0,$D3,$C6,$7E,$B1,$FC
	.dc.b	$7F,$9B,$E1,$9F,$1C,$E9,$FD,$7B,$FB,$68,$D6,$DF,$B6,$AB,$3F,$B3
	.dc.b	$9F,$D8,$F8,$7F,$9A,$F3,$24,$FE,$4A,$32,$FC,$38,$CB,$FB,$E8,$CB
	.dc.b	$FC,$3A,$33,$87,$F1,$59,$CC,$FE,$49,$FD,$B9,$1F,$DB,$35,$FD,$8B
	.dc.b	$5E,$06,$88,$68,$B1,$DD,$F4,$1F,$67,$DB,$F2,$C6,$63,$D0,$64,$D9
	.dc.b	$82,$3F,$2B,$34,$FF,$16,$F6,$7F,$7F,$E2,$4B,$FD,$BC,$65,$FC,$44
	.dc.b	$65,$FE,$47,$19,$7F,$99,$C6,$6F,$F9,$D4,$E6,$7F,$6E,$FF,$1C,$23
	.dc.b	$FC,$6D,$AF,$F1,$66,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$3D,$F0,$9E
	.dc.b	$58,$CC,$7A,$0C,$9B,$30,$47,$E3,$47,$F4,$5F,$11,$EB,$FF,$24,$F3
	.dc.b	$65,$FF,$20,$9C,$B6,$72,$FF,$56,$9E,$51,$23,$F6,$51,$AB,$FC,$ED
	.dc.b	$72,$94,$62,$BF,$A4,$4E,$67,$FA,$3B,$51,$1C,$36,$38,$91,$FE,$9E
	.dc.b	$64,$A7,$FA,$95,$D3,$A8,$E6,$E4,$DB,$E7,$9F,$34,$7C,$13,$F8,$77
	.dc.b	$9B,$EF,$7D,$7C,$3E,$67,$91,$83,$C4,$7F,$7C,$8E,$13,$D2,$34,$70
	.dc.b	$7F,$35,$1F,$57,$0F,$EF,$5F,$62,$F3,$A9,$9B,$FA,$D3,$87,$FA,$BC
	.dc.b	$ED,$C9,$0E,$EE,$5B,$3C,$DF,$1E,$2A,$FA,$A4,$8F,$8A,$D7,$E5,$35
	.dc.b	$FB,$2D,$7E,$24,$73,$78,$90,$E2,$79,$70,$FC,$3C,$FC,$0D,$3C,$91
	.dc.b	$89,$D6,$5F,$3F,$15,$6A,$3B,$DD,$AF,$C4,$D6,$9F,$CE,$79,$FF,$33
	.dc.b	$CE,$91,$82,$1F,$19,$CB,$F1,$23,$F5,$06,$31,$FE,$73,$4A,$CD,$72
	.dc.b	$3F,$37,$5F,$9E,$23,$F8,$E6,$BF,$93,$6B,$C0,$D1,$0D,$16,$BF,$5B
	.dc.b	$1E,$08,$FB,$BE,$D5,$9C,$C7,$B1,$F3,$C6,$71,$FA,$B5,$6B,$E3,$CF
	.dc.b	$F0,$5E,$1F,$C8,$F3,$49,$FE,$09,$19,$7F,$1F,$19,$7F,$13,$19,$7F
	.dc.b	$1F,$19,$C3,$F8,$1C,$E6,$7F,$7E,$FF,$02,$23,$FB,$26,$BF,$B7,$6B
	.dc.b	$C0,$D1,$0D,$16,$3B,$BE,$83,$F4,$BD,$1E,$58,$CC,$7A,$0C,$9B,$30
	.dc.b	$47,$EF,$C7,$F2,$9F,$9D,$FD,$AF,$5C,$AF,$E5,$11,$97,$F6,$71,$97
	.dc.b	$F5,$B1,$97,$F6,$71,$9A,$FE,$4F,$39,$9F,$E4,$6F,$F2,$62,$3F,$C5
	.dc.b	$9A,$FF,$1B,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$F8,$4F,$77,$96,$33
	.dc.b	$1E,$83,$26,$CC,$11,$FD,$1C,$7F,$4D,$EB,$FE,$33,$C5,$2B,$FA,$64
	.dc.b	$65,$FE,$2F,$19,$7F,$88,$C6,$5F,$E2,$F1,$9A,$FE,$97,$39,$9F,$E8
	.dc.b	$EF,$F4,$A2,$3F,$CE,$9A,$FF,$3F,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83
	.dc.b	$F9,$6F,$C1,$F9,$63,$31,$E8,$32,$6C,$C1,$1F,$E0,$F1,$FD,$C7,$D7
	.dc.b	$FE,$7B,$C1,$2B,$FB,$84,$65,$FE,$77,$19,$7F,$9A,$C6,$5F,$E7,$71
	.dc.b	$9A,$FE,$DF,$39,$9F,$ED,$6F,$F6,$E2,$3F,$D7,$9A,$FF,$63,$6B,$C0
	.dc.b	$D1,$0D,$16,$3B,$BE,$83,$FD,$4F,$E9,$3C,$FF,$34,$EA,$7D,$26,$A3
	.dc.b	$37,$FD,$1A,$EC,$FF,$80,$2F,$93,$FE,$B2,$FC,$0B,$FD,$97,$87,$F9
	.dc.b	$A2,$5E,$87,$FB,$7D,$B5,$B3,$FD,$C5,$5C,$30,$FE,$ED,$3A,$9F,$EE
	.dc.b	$2D,$BF,$F3,$9D,$6E,$7F,$C2,$2D,$53,$BA,$E7,$7D,$A6,$1F,$E1,$0A
	.dc.b	$FF,$72,$26,$91,$FF,$74,$35,$E0,$60,$86,$08,$7F,$87,$9F,$21,$1F
	.dc.b	$85,$EE,$78,$9F,$F0,$6C,$65,$FA,$B1,$9C,$3D,$F1,$6F,$E2,$CF,$C8
	.dc.b	$F9,$5F,$99,$E8,$4A,$FA,$51,$97,$EF,$C6,$5E,$F1,$97,$E6,$C6,$6F
	.dc.b	$E6,$CE,$67,$FC,$50,$7C,$57,$E1,$91,$F9,$4D,$7E,$03,$5E,$38,$23
	.dc.b	$11,$FE,$33,$8C,$3C,$88,$79,$8F,$59,$F9,$8F,$84,$F4,$57,$9A,$32
	.dc.b	$7E,$1A,$33,$8F,$5B,$B3,$F9,$79,$FC,$7F,$9D,$F9,$3F,$46,$4F,$E6
	.dc.b	$63,$2F,$E9,$A3,$2F,$D6,$8C,$BF,$94,$8C,$E1,$F9,$49,$CC,$FC,$07
	.dc.b	$F1,$24,$7F,$20,$D7,$F0,$CD,$78,$1A,$21,$A2,$C7,$77,$D0,$7E,$F3
	.dc.b	$D0,$F2,$C6,$63,$D0,$64,$D9,$82,$3F,$B7,$8F,$D9,$FB,$BF,$B7,$A9
	.dc.b	$5F,$BD,$8C,$BF,$C2,$A3,$2F,$E6,$A3,$2F,$EE,$23,$35,$FB,$89,$CC
	.dc.b	$FE,$11,$FD,$69,$1F,$DA,$35,$FD,$53,$5E,$06,$88,$68,$B1,$DD,$F4
	.dc.b	$1F,$C8,$3D,$DF,$2C,$66,$3D,$06,$4D,$98,$23,$FC,$6E,$3F,$8B,$E3
	.dc.b	$FC,$6F,$CF,$95,$FC,$86,$32,$FF,$2E,$8C,$BF,$BE,$8C,$BF,$C7,$23
	.dc.b	$35,$FC,$72,$73,$3F,$A8,$7F,$88,$91,$FE,$30,$D7,$F8,$73,$5E,$06
	.dc.b	$88,$68,$B1,$DD,$F4,$1F,$D1,$0F,$2B,$97,$D0,$64,$D9,$82,$3F,$CF
	.dc.b	$DF,$E7,$7F,$07,$FE,$AF,$86,$BF,$EB,$1F,$C9,$1E,$8E,$63,$0F,$0B
	.dc.b	$FE,$95,$79,$7F,$D9,$B5,$C1,$FF,$67,$DF,$47,$35,$C5,$FE,$57,$5C
	.dc.b	$55,$71,$7F,$D7,$6D,$AF,$E9,$73,$C6,$7F,$8B,$BF,$D0,$08,$FF,$47
	.dc.b	$6B,$FC,$F5,$AF,$03,$44,$34,$58,$EE,$FB,$8F,$5D,$FC,$5F,$47,$D5
	.dc.b	$3A,$9F,$59,$A3,$E1,$C6,$A7,$F9,$55,$59,$C9,$FD,$CF,$93,$E1,$60
	.dc.b	$E9,$A7,$F7,$08,$93,$D4,$F8,$F1,$A7,$91,$0F,$92,$E5,$FE,$B3,$1C
	.dc.b	$9F,$EC,$0F,$F6,$82,$3F,$DB,$DA,$E8,$C1,$0C,$10,$FE,$92,$F0,$C7
	.dc.b	$FC,$02,$F7,$7F,$05,$1F,$7B,$CF,$C6,$3F,$D3,$6A,$CF,$E2,$A6,$BF
	.dc.b	$DF,$E2,$E0,$9F,$F0,$9F,$AB,$F2,$3D,$09,$7F,$E1,$D8,$CB,$F1,$E9
	.dc.b	$3A,$8F,$EF,$73,$A7,$F1,$F1,$9C,$7F,$C3,$33,$99,$FE,$CA,$FF,$83
	.dc.b	$88,$FF,$85,$5A,$FF,$82,$DA,$F0,$34,$43,$45,$8E,$EF,$A4,$F4,$78
	.dc.b	$5E,$58,$CC,$7A,$0C,$9B,$30,$47,$E7,$DA,$7F,$5D,$F8,$13,$E3,$BF
	.dc.b	$C0,$F6,$24,$65,$C8,$70,$F0,$6B,$FE,$10,$6B,$CC,$E1,$1C,$9A,$74
	.dc.b	$8D,$BE,$08,$D4,$62,$36,$77,$8C,$F7,$C8,$E4,$D7,$C8,$D7,$C8,$D7
	.dc.b	$C9,$28,$C4,$7C,$06,$BE,$FF,$6F,$C6,$47,$D0,$D7,$E1,$35,$F8,$4D
	.dc.b	$7B,$B5,$F9,$AF,$BB,$C1,$91,$F0,$DA,$FC,$D6,$BF,$35,$AF,$A9,$AF
	.dc.b	$D8,$6B,$F5,$FE,$8F,$E5,$91,$FA,$0D,$7E,$D3,$5F,$B4,$D7,$E3,$B5
	.dc.b	$FC,$2B,$F5,$BC,$39,$1F,$6D,$AF,$E1,$5A,$FE,$15,$AF,$D2,$6B,$F9
	.dc.b	$06,$BF,$8F,$F1,$7F,$91,$FA,$52,$3F,$90,$64,$9F,$CD,$99,$8F,$E8
	.dc.b	$F3,$87,$71,$9D,$BF,$47,$39,$94,$7F,$14,$AE,$91,$F9,$5A,$FC,$99
	.dc.b	$1C,$10,$FD,$13,$96,$99,$3F,$48,$E5,$A7,$90,$87,$F5,$EF,$4B,$C9
	.dc.b	$5E,$39,$BC,$9F,$DC,$E6,$D7,$ED,$A6,$BF,$DA,$23,$F7,$BF,$0B,$FD
	.dc.b	$2B,$CF,$93,$F2,$23,$2F,$F2,$38,$CB,$F0,$AF,$F4,$B3,$A8,$FE,$E3
	.dc.b	$3A,$C3,$E3,$4E,$67,$F8,$03,$FB,$82,$3F,$BD,$6B,$FC,$1D,$AF,$03
	.dc.b	$44,$34,$58,$EE,$FA,$4F,$E7,$5E,$97,$96,$33,$1E,$93,$F8,$CC,$FF
	.dc.b	$6A,$9D,$5F,$B2,$79,$1F,$E0,$C8,$FE,$47,$E6,$FF,$84,$7F,$24,$7A
	.dc.b	$74,$CD,$7E,$42,$70,$FF,$71,$9D,$B9,$21,$E3,$0E,$9C,$37,$1F,$C8
	.dc.b	$4F,$69,$FE,$58,$FF,$22,$23,$FC,$9D,$AE,$8C,$10,$C1,$0F,$E6,$8F
	.dc.b	$08,$EF,$1E,$67,$FB,$9C,$BD,$3F,$14,$6A,$7D,$5B,$2F,$E3,$56,$B2
	.dc.b	$FE,$57,$16,$BF,$A5,$61,$CD,$FD,$BF,$CE,$7C,$DF,$4E,$46,$08,$7C
	.dc.b	$47,$2F,$82,$3F,$AB,$98,$CB,$CF,$38,$9E,$27,$F5,$19,$FE,$92,$47
	.dc.b	$FA,$73,$5F,$EB,$0D,$78,$1A,$21,$A2,$CF,$F0,$58,$F0,$47,$DA,$BB
	.dc.b	$39,$8F,$5B,$F3,$23,$38,$7F,$50,$AB,$5F,$1A,$7F,$B9,$E7,$F9,$1F
	.dc.b	$4A,$4F,$F1,$E8,$CB,$F8,$F8,$CB,$F5,$23,$2F,$E3,$23,$38,$7F,$1B
	.dc.b	$9C,$CF,$F7,$57,$FB,$69,$1F,$EE,$4D,$7F,$BE,$35,$E0,$68,$86,$8B
	.dc.b	$1D,$DF,$41,$FA,$5F,$4B,$CB,$19,$8F,$41,$93,$66,$08,$FE,$52,$3F
	.dc.b	$D0,$9F,$80,$3E,$2B,$F6,$DC,$92,$FF,$CC,$53,$96,$CE,$5F,$D7,$CF
	.dc.b	$28,$91,$FB,$98,$D5,$FF,$95,$2B,$94,$FE,$E5,$AF,$F8,$6C,$38,$7F
	.dc.b	$C3,$CD,$7F,$C5,$4D,$77,$5F,$E2,$18,$C1,$1C,$12,$3F,$C5,$F1,$96
	.dc.b	$C7,$A1,$F2,$3F,$C5,$F9,$B1,$5E,$98,$BA,$FF,$8C,$77,$56,$2F,$CB
	.dc.b	$55,$FE,$2B,$1F,$3B,$97,$F9,$17,$44,$9F,$9B,$19,$7F,$90,$46,$5F
	.dc.b	$8D,$19,$7F,$8E,$C6,$71,$F9,$93,$99,$F9,$59,$23,$E2,$12,$3E,$0B
	.dc.b	$5F,$F2,$AC,$6D,$C9,$0C,$C6,$DE,$AF,$19,$F0,$C7,$A8,$FE,$71,$D5
	.dc.b	$E7,$8C,$C7,$A8,$C9,$B3,$04,$7F,$93,$C7,$E5,$3E,$43,$FA,$46,$92
	.dc.b	$BF,$27,$19,$7F,$A3,$46,$5F,$C6,$46,$5F,$E8,$71,$9A,$FC,$94,$E6
	.dc.b	$7E,$3B,$F8,$92,$3F,$86,$6B,$F1,$9A,$F0,$34,$43,$45,$8E,$EF,$A0
	.dc.b	$FF,$66,$7E,$48,$F5,$7C,$D3,$A9,$F4,$9A,$8C,$DF,$E8,$AE,$CF,$FA
	.dc.b	$2E,$F9,$3F,$B6,$F0,$BF,$DF,$D0,$F1,$29,$FB,$E8,$93,$E2,$8C,$BC
	.dc.b	$48,$7F,$6D,$9E,$4E,$CC,$E3,$FB,$BA,$FD,$81,$23,$F5,$AD,$7F,$1A
	.dc.b	$D7,$F2,$AD,$7F,$7A,$D7,$F7,$AF,$18,$ED,$1E,$0C,$3F,$D4,$14,$F4
	.dc.b	$FD,$73,$FC,$2E,$ED,$FF,$B8,$AB,$53,$C4,$EA,$50,$E9,$FE,$29,$F1
	.dc.b	$1F,$2B,$D4,$91,$FC,$8A,$3F,$93,$98,$7E,$34,$61,$D1,$F7,$9D,$57
	.dc.b	$F2,$19,$D4,$FE,$C1,$FE,$30,$47,$F8,$AB,$5D,$18,$21,$82,$1F,$C0
	.dc.b	$1E,$11,$DE,$3C,$67,$DB,$F3,$3C,$31,$98,$F7,$9F,$94,$8C,$E5,$FC
	.dc.b	$C2,$AC,$FE,$3E,$7F,$A3,$F9,$7F,$8E,$F4,$A4,$FF,$46,$8C,$BF,$9E
	.dc.b	$8C,$BF,$CF,$E3,$2F,$E3,$63,$38,$7F,$45,$9C,$CF,$F1,$A7,$F9,$E1
	.dc.b	$1F,$E7,$2D,$7F,$8C,$B5,$E0,$68,$86,$8B,$1D,$DF,$41,$FB,$4F,$4B
	.dc.b	$CB,$19,$8F,$41,$93,$66,$08,$FE,$3E,$3F,$B5,$FD,$27,$EC,$BD,$39
	.dc.b	$5F,$DA,$A3,$2F,$F0,$28,$CB,$FD,$8E,$32,$FE,$C6,$33,$5F,$DA,$67
	.dc.b	$33,$FC,$F9,$FE,$C0,$47,$FA,$EB,$5F,$E7,$AD,$78,$1A,$21,$A2,$C7
	.dc.b	$77,$D0,$7F,$18,$F4,$BC,$B1,$98,$F4,$19,$36,$60,$8F,$EC,$E1,$12
	.dc.b	$1A,$62,$1F,$04,$F4,$FF,$8D,$92,$38,$4E,$E7,$51,$83,$AB,$87,$F8
	.dc.b	$D4,$F1,$7F,$D1,$11,$9A,$F0,$C7,$09,$D4,$7F,$09,$7C,$8E,$A3,$CB
	.dc.b	$18,$AF,$13,$FE,$19,$8A,$53,$65,$AF,$23,$FE,$1F,$6B,$C0,$E4,$E1
	.dc.b	$F2,$1C,$63,$A3,$D0,$DB,$88,$E1,$19,$AF,$44,$7B,$EB,$B4,$7F,$25
	.dc.b	$9F,$51,$D8,$F8,$B5,$87,$94,$FF,$30,$71,$5C,$DE,$15,$DA,$BF,$A8
	.dc.b	$C6,$A7,$97,$53,$D3,$A6,$67,$EC,$1F,$80,$9C,$B8,$9E,$0E,$A3,$F0
	.dc.b	$AD,$76,$FF,$B0,$E3,$6F,$DE,$8E,$CF,$CC,$7E,$31,$1F,$A8,$D7,$46
	.dc.b	$08,$60,$87,$D9,$78,$5F,$CE,$7B,$BF,$87,$96,$B8,$C7,$F3,$15,$6F
	.dc.b	$F4,$75,$A7,$D8,$43,$D4,$7C,$3A,$FA,$3A,$FC,$CF,$4A,$4F,$C5,$8C
	.dc.b	$BF,$26,$32,$FE,$82,$32,$FC,$78,$CE,$3F,$86,$9C,$CF,$E5,$5F,$C6
	.dc.b	$11,$FC,$D3,$5F,$CB,$B5,$E0,$68,$86,$8B,$1D,$DF,$41,$F7,$FD,$2F
	.dc.b	$2C,$66,$3D,$06,$4D,$98,$23,$F3,$E3,$F2,$BF,$45,$F9,$0E,$69,$5F
	.dc.b	$89,$8C,$BF,$8D,$8C,$BF,$C1,$23,$2F,$E2,$E3,$35,$FA,$A9,$CC,$FE
	.dc.b	$E5,$FD,$81,$1F,$DF,$35,$FD,$DB,$5E,$06,$88,$68,$B1,$DD,$F4,$1F
	.dc.b	$A3,$F4,$7C,$B1,$98,$F4,$19,$36,$60,$8F,$E4,$E3,$F7,$3C,$DF,$B4
	.dc.b	$F3,$A5,$7E,$B6,$32,$FE,$C6,$32,$FF,$28,$8C,$BF,$AF,$8C,$D7,$F0
	.dc.b	$E9,$CC,$FF,$1D,$7F,$8A,$11,$FE,$44,$D7,$F8,$FB,$5E,$06,$88,$68
	.dc.b	$B1,$DD,$F4,$1F,$C1,$FB,$BC,$B1,$98,$F4,$19,$36,$60,$8F,$ED,$E3
	.dc.b	$F8,$EF,$C2,$FF,$18,$F3,$E5,$7F,$11,$8C,$BF,$C5,$63,$2F,$F4,$C8
	.dc.b	$CB,$FC,$4E,$33,$5F,$CC,$A7,$33,$FD,$05,$FE,$70,$47,$FA,$33,$5F
	.dc.b	$E8,$6D,$78,$1A,$21,$A2,$C7,$77,$D0,$7F,$2B,$F3,$BC,$B1,$98,$F4
	.dc.b	$19,$36,$60,$8F,$F1,$B8,$FE,$83,$F4,$FF,$D3,$C9,$3F,$D7,$5A,$FF
	.dc.b	$5A,$6B,$FD,$7D,$AF,$F7,$96,$BF,$DE,$1F,$EB,$5E,$0C,$8F,$ED,$0D
	.dc.b	$7F,$B2,$B5,$FE,$D4,$D7,$FC,$08,$D7,$FC,$06,$D5,$F8,$13,$E2,$BF
	.dc.b	$AF,$36,$3F,$A9,$B5,$61,$FE,$01,$87,$97,$C2,$A5,$70,$79,$C6,$23
	.dc.b	$8F,$47,$F8,$76,$BA,$5F,$3B,$DD,$71,$21,$C0,$FF,$13,$CF,$0A,$E5
	.dc.b	$7D,$6F,$95,$70,$7C,$D1,$99,$F0,$5D,$AE,$73,$FD,$F7,$EF,$DE,$B7
	.dc.b	$F8,$31,$B3,$FE,$19,$6B,$FE,$1D,$6B,$FE,$37,$0E,$1F,$F1,$D3,$5F
	.dc.b	$F1,$FB,$5D,$C7,$72,$3D,$DA,$F7,$F1,$FC,$84,$7C,$8D,$74,$60,$86
	.dc.b	$08,$79,$E5,$1C,$63,$1E,$B7,$C1,$23,$F0,$5A,$FC,$16,$BC,$0D,$10
	.dc.b	$D1,$4F,$07,$95,$F3,$48,$FC,$D6,$BF,$35,$AF,$03,$44,$34,$53,$D9
	.dc.b	$D9,$F9,$AF,$4A,$47,$F8,$B3,$04,$FD,$F3,$31,$F8,$D3,$F4,$73,$97
	.dc.b	$84,$6B,$5F,$D0,$4E,$67,$85,$74,$7F,$0D,$5F,$89,$23,$82,$1F,$89
	.dc.b	$72,$D3,$27,$A8,$A7,$E2,$9F,$05,$7F,$9C,$67,$C1,$3E,$9D,$9E,$8E
	.dc.b	$B7,$E1,$3E,$A3,$F7,$71,$97,$F2,$78,$57,$F8,$1C,$7E,$8F,$E1,$7F
	.dc.b	$81,$FD,$4C,$9F,$95,$19,$7F,$82,$C6,$5F,$87,$19,$7C,$91,$9C,$BD
	.dc.b	$13,$99,$FC,$93,$FA,$F2,$3F,$AF,$6B,$FA,$F6,$BC,$0D,$10,$D1,$63
	.dc.b	$BB,$E8,$3F,$8D,$FA,$7E,$58,$CC,$7A,$0C,$9B,$30,$47,$F9,$04,$7F
	.dc.b	$07,$3F,$93,$FA,$F2,$7F,$1D,$19,$7F,$94,$C6,$5F,$C3,$C6,$5F,$B1
	.dc.b	$19,$9F,$B9,$39,$9F,$DB,$3F,$C4,$C8,$FF,$13,$6B,$FC,$4D,$AF,$03
	.dc.b	$44,$34,$58,$EE,$FA,$0F,$E7,$FE,$A7,$96,$33,$1E,$83,$26,$CC,$11
	.dc.b	$FE,$89,$1F,$CA,$FE,$1B,$FA,$5E,$12,$BF,$65,$19,$7F,$A6,$C6,$5F
	.dc.b	$D5,$C6,$5F,$CE,$46,$6B,$F4,$13,$99,$FE,$34,$FF,$37,$23,$FC,$DD
	.dc.b	$AF,$F3,$76,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$B1,$FA,$BE,$58,$CC
	.dc.b	$7A,$0C,$9B,$30,$47,$FB,$44,$7F,$4F,$F5,$FF,$B7,$FC,$64,$AF,$E2
	.dc.b	$D1,$97,$FB,$8C,$65,$FE,$1F,$19,$7F,$80,$46,$6B,$F8,$24,$E6,$7F
	.dc.b	$9F,$3F,$D6,$C8,$FF,$5B,$6B,$FD,$6D,$AF,$03,$44,$34,$58,$EE,$FA
	.dc.b	$0F,$F0,$3E,$DE,$58,$CC,$7A,$0C,$9B,$30,$47,$FC,$19,$1F,$EE,$FF
	.dc.b	$BC,$7C,$3F,$F8,$53,$DC,$97,$FE,$E6,$9C,$B6,$72,$FF,$23,$9E,$51
	.dc.b	$23,$FC,$37,$1A,$B7,$12,$65,$F4,$7F,$A2,$35,$FE,$FC,$D7,$80,$70
	.dc.b	$F2,$BF,$F0,$23,$A8,$CB,$B8,$CB,$B2,$C6,$89,$1A,$8E,$06,$9C,$47
	.dc.b	$9E,$BD,$73,$CC,$F4,$CF,$F8,$F3,$9B,$E1,$46,$0F,$90,$7F,$8F,$9C
	.dc.b	$CF,$28,$EE,$6C,$FE,$9D,$1F,$57,$C0,$FF,$23,$F4,$49,$FA,$91,$97
	.dc.b	$FC,$93,$19,$7E,$5C,$65,$FE,$B1,$19,$AF,$06,$0A,$D5,$7D,$07,$DE
	.dc.b	$47,$E3,$B5,$F9,$F1,$C5,$C9,$0C,$C6,$DE,$86,$67,$C3,$1E,$57,$FC
	.dc.b	$BD,$D3,$72,$FB,$53,$6B,$FC,$A7,$35,$FC,$14,$5E,$7F,$F1,$6C,$7E
	.dc.b	$5F,$E1,$BF,$CD,$BD,$72,$30,$43,$FB,$D3,$97,$F2,$31,$FA,$33,$18
	.dc.b	$FF,$9C,$67,$13,$C4,$FC,$AC,$FE,$10,$8F,$E3,$1A,$FC,$66,$BC,$0D
	.dc.b	$10,$D1,$67,$EA,$C7,$82,$3F,$C2,$18,$D9,$CC,$7A,$DF,$F4,$34,$64
	.dc.b	$FD,$14,$D9,$FF,$3D,$C7,$EF,$FC,$5F,$D5,$D7,$E4,$7F,$03,$7C,$06
	.dc.b	$2B,$F5,$F1,$8B,$FE,$05,$7C,$5F,$DD,$5F,$1B,$E1,$1F,$BD,$AE,$0F
	.dc.b	$EF,$2B,$83,$FE,$41,$8C,$D7,$18,$EA,$43,$F1,$B3,$FB,$02,$3F,$B8
	.dc.b	$6B,$A3,$04,$30,$43,$F8,$5B,$C2,$3F,$42,$F9,$1F,$85,$E0,$DC,$3F
	.dc.b	$85,$4D,$AF,$64,$D7,$F8,$94,$5B,$4F,$F2,$BF,$87,$F8,$74,$97,$B5
	.dc.b	$86,$C7,$F2,$D9,$E0,$FC,$CC,$3E,$8D,$6C,$CC,$7F,$26,$9C,$3B,$3E
	.dc.b	$84,$6B,$1F,$DE,$CF,$F1,$82,$4F,$F2,$06,$BF,$BA,$6B,$FC,$C1,$AF
	.dc.b	$F2,$B6,$BF,$CE,$1E,$31,$DA,$3D,$47,$8C,$F8,$1E,$D7,$8E,$F9,$46
	.dc.b	$8F,$C1,$46,$5E,$86,$67,$F2,$91,$FD,$57,$CF,$F9,$1F,$05,$2B,$FA
	.dc.b	$A4,$65,$F0,$64,$9D,$47,$F4,$C9,$D3,$F9,$18,$CD,$FF,$22,$9C,$CF
	.dc.b	$F5,$07,$F9,$F9,$1F,$E8,$ED,$7F,$90,$B5,$E0,$68,$86,$8B,$1D,$DF
	.dc.b	$49,$E8,$F2,$FC,$B1,$98,$F4,$19,$36,$60,$8F,$EE,$63,$FB,$D7,$C3
	.dc.b	$FE,$37,$E3,$65,$7F,$79,$8C,$BF,$1A,$32,$FF,$70,$8C,$BF,$B5,$8C
	.dc.b	$D7,$F4,$69,$CC,$FF,$74,$7F,$B1,$91,$FE,$D6,$D7,$FA,$2B,$5E,$06
	.dc.b	$88,$68,$B1,$DD,$F4,$1F,$73,$DB,$F2,$C6,$63,$D0,$64,$D9,$82,$3F
	.dc.b	$C7,$63,$F6,$7F,$79,$FE,$0A,$9F,$E2,$5E,$8F,$EC,$CC,$B0,$FC,$5B
	.dc.b	$97,$F3,$91,$C6,$BF,$BC,$37,$6F,$E4,$F5,$BA,$FF,$6F,$4E,$5F,$F1
	.dc.b	$03,$FE,$0D,$4D,$23,$FE,$11,$6B,$FE,$18,$6B,$FE,$1E,$6B,$EB,$6B
	.dc.b	$FE,$26,$7C,$0F,$63,$D8,$7F,$31,$F7,$2D,$FA,$A2,$DF,$F4,$09,$B1
	.dc.b	$E8,$29,$FC,$6D,$F9,$7F,$0F,$FC,$6F,$E1,$E4,$FC,$68,$CB,$FB,$48
	.dc.b	$CB,$F1,$23,$2F,$F4,$38,$CD,$FC,$C9,$CC,$FB,$5F,$88,$47,$E2,$35
	.dc.b	$F8,$8D,$78,$1A,$21,$A2,$C7,$77,$D0,$7F,$0A,$F5,$FC,$B1,$98,$F4
	.dc.b	$19,$36,$60,$8F,$F6,$B8,$FC,$7F,$BD,$FC,$FF,$E2,$65,$7E,$2A,$32
	.dc.b	$FF,$18,$8C,$BF,$88,$8C,$BF,$D9,$E3,$35,$F9,$09,$CC,$FE,$05,$FC
	.dc.b	$41,$1F,$C4,$35,$FC,$43,$5E,$06,$88,$68,$B1,$DD,$F4,$1F,$CB,$A3
	.dc.b	$CB,$19,$8F,$41,$93,$66,$08,$FF,$83,$E3,$F6,$78,$7F,$63,$F6,$E5
	.dc.b	$7E,$BA,$32,$FF,$3C,$8C,$BF,$AC,$8C,$BF,$E0,$B8,$CD,$7E,$D2,$73
	.dc.b	$3F,$A5,$7F,$58,$47,$F5,$8D,$7F,$58,$D7,$81,$A2,$1A,$2C,$77,$7D
	.dc.b	$07,$F5,$2F,$7B,$CB,$19,$8F,$41,$93,$66,$08,$FF,$8C,$E3,$F8,$BE
	.dc.b	$FF,$E0,$7C,$E5,$7F,$12,$8C,$BF,$D8,$23,$2F,$F1,$08,$CB,$FE,$2F
	.dc.b	$8C,$D7,$F1,$89,$CC,$FF,$09,$7F,$88,$11,$FE,$20,$D7,$F8,$83,$5E
	.dc.b	$06,$88,$68,$B1,$DD,$F4,$1F,$DD,$BD,$EF,$2C,$66,$3D,$06,$4D,$98
	.dc.b	$23,$FE,$57,$8F,$E7,$7C,$BF,$E2,$BF,$8D,$95,$FC,$DA,$32,$FF,$81
	.dc.b	$23,$2F,$F3,$48,$CB,$FE,$53,$8C,$D7,$F3,$C9,$CC,$FF,$2D,$7F,$9A
	.dc.b	$11,$FE,$68,$D7,$F9,$A3,$5E,$06,$88,$68,$B1,$DD,$F4,$1F,$E1,$A3
	.dc.b	$CA,$E5,$F4,$19,$36,$60,$8F,$F9,$ED,$FE,$BF,$EC,$FF,$4B,$23,$FD
	.dc.b	$2D,$AF,$F4,$B6,$BF,$D2,$DA,$FF,$4B,$6B,$FD,$69,$FE,$AD,$D7,$23
	.dc.b	$FA,$B3,$5F,$EA,$CD,$7F,$AB,$35,$FE,$AC,$D7,$FB,$2B,$5F,$EC,$BE
	.dc.b	$C7,$F6,$32,$3F,$D8,$DA,$FF,$63,$6B,$FD,$8D,$AF,$F6,$36,$BF,$DC
	.dc.b	$9F,$ED,$DD,$B2,$3F,$B7,$35,$FE,$DC,$D7,$FB,$73,$5F,$ED,$CD,$7F
	.dc.b	$BE,$B5,$FE,$FB,$DF,$FD,$EC,$8F,$F7,$B6,$BF,$DE,$DA,$FF,$7B,$6B
	.dc.b	$FD,$ED,$AF,$F8,$29,$FF,$0D,$22,$7B,$4E,$6C,$C6,$1A,$A3,$4E,$9A
	.dc.b	$F4,$6B,$FE,$27,$6B,$DC,$E5,$1D,$23,$31,$DC,$E9,$19,$23,$DF,$C5
	.dc.b	$FF,$52,$FE,$48,$F8,$0A,$7F,$AA,$2F,$FD,$F7,$4D,$7C,$B1,$87,$CF
	.dc.b	$5C,$67,$FD,$20,$D7,$86,$1E,$7A,$E0,$DC,$3B,$CF,$0C,$3C,$11,$23
	.dc.b	$CC,$43,$E2,$3E,$F2,$39,$3E,$33,$87,$66,$4F,$01,$CF,$0F,$5C,$77
	.dc.b	$8F,$39,$F6,$F0,$76,$66,$B7,$3F,$EA,$DC,$7D,$6E,$E7,$58,$F6,$8E
	.dc.b	$67,$DF,$C3,$E7,$CF,$CD,$F3,$3E,$8F,$C0,$4A,$E9,$19,$7E,$5C,$65
	.dc.b	$E7,$82,$75,$1D,$B2,$56,$F2,$FA,$F5,$AA,$FC,$53,$F8,$72,$3F,$91
	.dc.b	$6B,$F9,$16,$BC,$0D,$10,$D1,$63,$BB,$EA,$3F,$07,$F4,$1E,$68,$CC
	.dc.b	$7A,$A7,$F2,$33,$9C,$FF,$0D,$16,$7E,$94,$FE,$4F,$D2,$FC,$BF,$87
	.dc.b	$27,$E9,$46,$5F,$C8,$C6,$5F,$8F,$19,$7E,$34,$67,$0F,$CD,$CE,$67
	.dc.b	$F5,$CF,$EA,$C8,$FE,$D5,$AF,$ED,$5A,$F0,$34,$43,$45,$8E,$EF,$A0
	.dc.b	$FD,$3F,$C1,$F9,$63,$31,$E8,$32,$6C,$C1,$1F,$D8,$C7,$ED,$FD,$8F
	.dc.b	$DD,$FC,$14,$AF,$CB,$46,$5F,$DA,$C6,$5F,$C5,$C6,$5F,$C5,$46,$6B
	.dc.b	$F7,$F3,$99,$FE,$24,$FF,$0F,$23,$FC,$65,$AF,$F1,$96,$BC,$0D,$10
	.dc.b	$D1,$63,$BB,$E8,$3F,$85,$F9,$3E,$58,$CC,$7A,$0C,$9B,$30,$47,$F8
	.dc.b	$AC,$7F,$1B,$F6,$FF,$8F,$FC,$14,$AF,$DD,$46,$5F,$E3,$31,$97,$F5
	.dc.b	$F1,$97,$F5,$D1,$9A,$FE,$47,$39,$9F,$E6,$CF,$F3,$32,$3F,$CF,$5A
	.dc.b	$FF,$3D,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$F9,$7F,$99,$E5,$8C,$C7
	.dc.b	$A0,$C9,$B3,$04,$7F,$9C,$C7,$F3,$FF,$89,$FE,$87,$F0,$D2,$BF,$8F
	.dc.b	$46,$5F,$E7,$B1,$97,$F8,$9C,$65,$FE,$25,$19,$AF,$E8,$F3,$99,$FE
	.dc.b	$B4,$FF,$57,$23,$FD,$85,$AF,$F6,$16,$BC,$0D,$10,$D1,$63,$BB,$E8
	.dc.b	$3F,$A9,$FC,$27,$96,$33,$1E,$83,$26,$CC,$11,$FE,$BB,$1F,$D8,$FE
	.dc.b	$0B,$FB,$3F,$A3,$2B,$FA,$14,$65,$FE,$C3,$19,$7F,$9B,$C6,$5F,$E6
	.dc.b	$D1,$9A,$FE,$D7,$39,$9F,$EF,$CF,$F7,$B2,$3F,$E0,$56,$BF,$E0,$56
	.dc.b	$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$BB,$F3,$F9,$63,$31,$E8,$32,$6C
	.dc.b	$C1,$1F,$F0,$0C,$7F,$84,$DF,$F0,$CF,$E4,$4F,$89,$FF,$06,$FC,$C4
	.dc.b	$BF,$F1,$35,$6A,$78,$74,$7F,$86,$EF,$85,$7F,$82,$AF,$83,$FD,$AA
	.dc.b	$32,$DA,$CD,$FF,$84,$AB,$9C,$FF,$8C,$EB,$73,$DE,$3F,$C6,$71,$C9
	.dc.b	$FF,$15,$07,$0F,$8E,$BF,$C6,$F1,$88,$E9,$18,$23,$D1,$AF,$F9,$12
	.dc.b	$38,$3D,$63,$33,$E3,$8F,$51,$FE,$28,$F5,$FD,$2E,$CF,$8F,$87,$8D
	.dc.b	$A7,$88,$E0,$60,$E0,$74,$24,$FF,$8D,$A7,$AD,$FF,$94,$67,$FC,$6F
	.dc.b	$7F,$E4,$23,$38,$C8,$C9,$0F,$A4,$E5,$FF,$08,$47,$17,$C5,$3D,$5D
	.dc.b	$9F,$42,$7E,$59,$1C,$10,$F9,$0E,$5A,$64,$FA,$F1,$D5,$B1,$BB,$FC
	.dc.b	$8B,$E5,$7F,$CA,$BF,$0F,$71,$FC,$74,$DA,$FF,$1F,$4D,$7F,$05,$17
	.dc.b	$B1,$3F,$23,$83,$EA,$EC,$DA,$99,$C3,$E0,$CE,$1F,$A9,$3B,$72,$43
	.dc.b	$FC,$8A,$E5,$B3,$CD,$FD,$04,$FE,$48,$8F,$E6,$9A,$FE,$3D,$AF,$C6
	.dc.b	$6B,$FA,$D6,$BF,$A5,$78,$E5,$FB,$87,$C0,$FF,$A4,$A9,$F0,$F8,$CF
	.dc.b	$95,$FA,$6B,$53,$98,$E8,$66,$7F,$D2,$D5,$CA,$3F,$0B,$C9,$F2,$BD
	.dc.b	$F9,$18,$21,$F1,$1C,$BF,$C0,$63,$F3,$26,$30,$F3,$4E,$27,$89,$F8
	.dc.b	$69,$FD,$C9,$1F,$E0,$2D,$7F,$6C,$D7,$81,$A2,$1A,$2C,$FF,$1E,$8F
	.dc.b	$04,$7D,$AE,$3B,$39,$8F,$5B,$F2,$E3,$38,$7F,$1F,$AB,$5F,$EC,$9C
	.dc.b	$15,$FB,$6E,$9F,$C7,$FC,$3C,$AF,$D6,$C6,$5F,$C6,$C6,$5F,$E4,$D1
	.dc.b	$97,$F1,$B1,$9C,$7F,$9A,$CE,$67,$F5,$6F,$F1,$E2,$3F,$C9,$9A,$FF
	.dc.b	$1B,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$F4,$7F,$05,$E5,$8C,$C7,$A0
	.dc.b	$C9,$B3,$04,$7F,$29,$1F,$C6,$BC,$0F,$D9,$F4,$CA,$FE,$23,$19,$7F
	.dc.b	$63,$19,$7F,$A5,$46,$5F,$D8,$C6,$6B,$FA,$CC,$E6,$7F,$87,$BF,$D0
	.dc.b	$88,$FF,$4A,$6B,$FC,$FD,$AF,$03,$44,$34,$58,$EE,$FA,$0F,$E0,$FF
	.dc.b	$0D,$E5,$8C,$C7,$A0,$C9,$B3,$04,$7F,$71,$1F,$CF,$BE,$23,$F8,$BF
	.dc.b	$C3,$CA,$FE,$6B,$19,$7F,$8A,$C6,$5F,$ED,$D1,$97,$F8,$AC,$66,$BF
	.dc.b	$BE,$CE,$67,$F9,$9B,$FD,$98,$8F,$F6,$E6,$BF,$D8,$DA,$F0,$34,$43
	.dc.b	$45,$8E,$EF,$A0,$FE,$57,$F0,$FE,$58,$CC,$7A,$0C,$9B,$30,$47,$F8
	.dc.b	$E4,$7F,$6E,$FC,$89,$E8,$7F,$3C,$F8,$89,$7F,$DA,$E7,$2D,$9C,$BF
	.dc.b	$CE,$27,$94,$48,$FE,$7F,$1A,$B7,$1A,$65,$F4,$53,$A7,$37,$FE,$14
	.dc.b	$78,$09,$E0,$FF,$86,$1A,$E9,$AF,$23,$91,$C6,$2B,$4C,$C7,$07,$07
	.dc.b	$85,$C1,$23,$8B,$E9,$9F,$44,$F0,$6C,$66,$BF,$A0,$F8,$9F,$16,$BD
	.dc.b	$F1,$E7,$BC,$5F,$F4,$FA,$B3,$9A,$FA,$07,$F6,$A8,$FB,$1E,$B7,$F7
	.dc.b	$2F,$88,$93,$F5,$E3,$2F,$F7,$08,$CB,$E0,$3E,$8C,$69,$FE,$E3,$1A
	.dc.b	$BF,$4C,$E6,$7E,$73,$F0,$08,$FC,$86,$BF,$1D,$AF,$03,$44,$34,$58
	.dc.b	$EE,$FA,$4F,$F0,$3F,$C4,$79,$63,$31,$E9,$3F,$05,$3F,$E1,$29,$D6
	.dc.b	$1F,$81,$8B,$3F,$D2,$E7,$F3,$DE,$DF,$F8,$63,$CB,$93,$F9,$D8,$CB
	.dc.b	$FE,$16,$8C,$BD,$F4,$4E,$A3,$FC,$2F,$3A,$C7,$EF,$CE,$67,$F2,$8F
	.dc.b	$E1,$88,$FE,$39,$AF,$E3,$5A,$F0,$34,$43,$45,$8E,$EF,$A4,$FF,$16
	.dc.b	$FC,$3F,$96,$33,$1E,$99,$FF,$1B,$CE,$72,$FD,$34,$59,$FE,$E1,$3F
	.dc.b	$C0,$FD,$FF,$F1,$EF,$C3,$C9,$FE,$05,$19,$7F,$C7,$71,$97,$E3,$46
	.dc.b	$5F,$F1,$DC,$67,$0F,$D1,$CE,$67,$F7,$0F,$EA,$88,$FE,$C9,$AF,$EC
	.dc.b	$5A,$F0,$34,$43,$45,$8E,$EF,$A0,$FF,$28,$FA,$1E,$58,$CC,$7A,$0C
	.dc.b	$9B,$30,$47,$FC,$85,$1F,$C9,$FE,$07,$F8,$09,$27,$F8,$4B,$5F,$E1
	.dc.b	$0D,$7F,$8B,$35,$FE,$2C,$D7,$F9,$1B,$FC,$33,$E0,$A4,$7F,$14,$6B
	.dc.b	$FC,$4D,$AF,$F2,$16,$BF,$C8,$5A,$FF,$2E,$6B,$FC,$A7,$D9,$FE,$32
	.dc.b	$47,$F9,$0B,$5F,$E4,$0D,$7F,$97,$35,$FE,$5C,$D7,$F9,$DB,$FC,$93
	.dc.b	$E1,$24,$7F,$2C,$6B,$FC,$AD,$AF,$F3,$96,$BF,$CE,$5A,$FF,$46,$6B
	.dc.b	$FD,$07,$DA,$FE,$62,$47,$F9,$CB,$5F,$E7,$0D,$7F,$A3,$35,$FE,$8C
	.dc.b	$D7,$FA,$9B,$FC,$F3,$DA,$91,$FD,$11,$AF,$F4,$36,$BF,$D4,$5A,$FF
	.dc.b	$51,$6B,$FD,$79,$AF,$F5,$9F,$6B,$FA,$49,$1F,$EA,$2D,$7F,$A8,$35
	.dc.b	$FE,$BC,$D7,$FA,$F3,$5F,$ED,$6F,$F5,$4F,$6A,$47,$F5,$C6,$BF,$D6
	.dc.b	$DA,$FF,$69,$6B,$FD,$A5,$AF,$F7,$66,$BF,$DC,$7D,$AF,$EC,$24,$7F
	.dc.b	$B4,$B5,$FE,$D0,$D7,$FB,$B3,$5F,$EE,$CD,$7F,$C0,$6F,$F7,$4F,$C8
	.dc.b	$BF,$B5,$C7,$F8,$00,$38,$7F,$C0,$6D,$7F,$C1,$8D,$77,$C7,$CC,$8F
	.dc.b	$F8,$0A,$34,$48,$CC,$7F,$83,$E3,$3F,$0D,$E3,$24,$71,$6B,$B6,$BE
	.dc.b	$78,$46,$23,$CE,$D7,$CE,$FB,$F7,$91,$E5,$6B,$E3,$6B,$D5,$AF,$C2
	.dc.b	$6B,$F0,$9A,$FC,$8E,$2F,$98,$47,$E1,$35,$F8,$0D,$7D,$4D,$7E,$7B
	.dc.b	$5F,$9E,$FD,$7E,$39,$1F,$35,$AF,$CB,$6B,$F1,$5A,$FD,$A6,$BF,$69
	.dc.b	$AF,$E0,$39,$7F,$08,$47,$ED,$35,$FB,$0D,$7E,$93,$5F,$C3,$B5,$FC
	.dc.b	$3B,$F8,$FE,$79,$1F,$85,$6B,$F8,$36,$BF,$75,$AF,$E4,$9A,$FE,$49
	.dc.b	$AF,$E6,$3A,$7F,$38,$47,$F2,$4D,$7F,$20,$D7,$F1,$4D,$7F,$3E,$D7
	.dc.b	$F3,$EF,$E9,$FA,$A4,$7E,$75,$AF,$E6,$DA,$FE,$55,$AF,$EA,$9A,$FE
	.dc.b	$A9,$AF,$EC,$3B,$3F,$68,$47,$F5,$4D,$7F,$50,$D7,$F4,$4D,$7F,$6E
	.dc.b	$D7,$F6,$EF,$F9,$CB,$F2,$46,$FF,$E7,$EE,$8F,$F3,$5C,$62,$BF,$D1
	.dc.b	$11,$89,$D7,$07,$FA,$13,$C8,$FF,$38,$CE,$6B,$73,$FE,$88,$BE,$51
	.dc.b	$FE,$49,$6C,$FF,$9E,$2B,$8D,$6E,$FA,$DF,$FA,$5E,$F7,$39,$8E,$A7
	.dc.b	$FA,$6E,$B8,$BB,$AE,$2A,$B8,$B9,$38,$89,$E2,$EE,$78,$CF,$08,$C1
	.dc.b	$C2,$77,$19,$24,$F8,$23,$9F,$0B,$C2,$E7,$CB,$30,$F0,$CE,$5B,$39
	.dc.b	$7E,$34,$FA,$A7,$51,$FE,$71,$6E,$5F,$5A,$78,$3F,$EA,$28,$CE,$5F
	.dc.b	$CD,$AF,$F9,$E1,$2F,$F9,$93,$5F,$E6,$0D,$7F,$95,$35,$E4,$70,$21
	.dc.b	$C0,$F4,$90,$F3,$79,$7E,$98,$91,$9B,$F3,$4E,$23,$D9,$36,$73,$1B
	.dc.b	$AF,$0B,$A3,$E4,$D6,$E3,$33,$F0,$6B,$E4,$4E,$6B,$FA,$AC,$57,$FA
	.dc.b	$DC,$5B,$F8,$F1,$A3,$A9,$E1,$30,$78,$4F,$69,$93,$FB,$2C,$F3,$9F
	.dc.b	$0C,$66,$7A,$BE,$03,$A9,$CE,$BD,$0E,$1F,$82,$E6,$FD,$A7,$F7,$0A
	.dc.b	$EB,$19,$3C,$B3,$C0,$93,$C2,$32,$F1,$7B,$39,$7D,$26,$0F,$EF,$73
	.dc.b	$FE,$03,$23,$FD,$E5,$AF,$F7,$76,$BF,$DC,$5A,$F7,$34,$71,$21,$F2
	.dc.b	$89,$3F,$EF,$AE,$07,$A9,$EE,$8F,$5C,$F9,$0F,$DC,$CE,$9F,$41,$F3
	.dc.b	$4E,$76,$64,$6D,$C9,$CC,$A1,$D4,$F8,$23,$A3,$63,$33,$ED,$8D,$60
	.dc.b	$C3,$F8,$8D,$6A,$FC,$38,$7A,$F0,$E0,$79,$63,$C1,$1C,$0D,$3E,$5C
	.dc.b	$3E,$24,$66,$3D,$F8,$77,$7E,$1C,$65,$F5,$C6,$5F,$F7,$C5,$F8,$DE
	.dc.b	$E7,$2C,$38,$5F,$68,$C1,$D0,$87,$48,$DE,$1E,$9A,$E9,$7D,$EF,$E3
	.dc.b	$61,$AC,$3D,$57,$F0,$1F,$86,$74,$3F,$C9,$D7,$F5,$5B,$3F,$E5,$17
	.dc.b	$B3,$5E,$26,$63,$83,$A3,$4E,$1D,$9D,$0C,$C7,$37,$E1,$9C,$DF,$43
	.dc.b	$8A,$FC,$A7,$B7,$8A,$F9,$44,$71,$59,$6D,$7C,$8A,$E2,$DB,$F3,$3F
	.dc.b	$52,$BF,$0F,$5C,$23,$07,$E7,$9C,$57,$D5,$8F,$B7,$18,$AE,$73,$DA
	.dc.b	$78,$91,$E0,$64,$B1,$C0,$91,$FA,$69,$E8,$FC,$E9,$E1,$3A,$8F,$88
	.dc.b	$DB,$57,$09,$C9,$CC,$F8,$07,$39,$E1,$1B,$9F,$95,$3A,$38,$91,$EA
	.dc.b	$E5,$FA,$37,$F2,$23,$0F,$D1,$7E,$41,$82,$9F,$20,$C1,$49,$7F,$11
	.dc.b	$F3,$1E,$93,$37,$EF,$9F,$04,$60,$D9,$E7,$BC,$D7,$38,$C1,$E2,$71
	.dc.b	$7F,$02,$39,$1E,$FA,$E3,$7F,$AD,$BF,$8C,$62,$7B,$D7,$1B,$FA,$B1
	.dc.b	$8B,$D3,$FC,$6E,$B8,$BE,$1B,$F5,$34,$F7,$57,$92,$B8,$DF,$D0,$9F
	.dc.b	$C5,$57,$D8,$7E,$59,$FB,$8B,$FD,$54,$EE,$30,$7E,$CA,$32,$47,$04
	.dc.b	$3F,$74,$E5,$A6,$6F,$DA,$FA,$CF,$F1,$CE,$CE,$AF,$3D,$7C,$08,$C4
	.dc.b	$FB,$4E,$07,$CE,$72,$FF,$24,$F7,$7F,$C7,$FD,$B2,$3F,$A6,$C7,$F3
	.dc.b	$A3,$0F,$F9,$66,$30,$E8,$F1,$BA,$C3,$FA,$D4,$EA,$7F,$8B,$3F,$18
	.dc.b	$8F,$F1,$66,$BA,$30,$43,$04,$3F,$93,$BC,$23,$BC,$78,$C2,$CB,$FA
	.dc.b	$77,$E4,$6F,$86,$F2,$4F,$0A,$F8,$47,$F9,$92,$74,$F3,$8D,$12,$B9
	.dc.b	$D7,$27,$FA,$15,$DA,$FE,$E1,$7C,$9E,$A4,$BE,$4E,$AF,$83,$FD,$C6
	.dc.b	$AD,$71,$AD,$CE,$4F,$25,$F8,$9A,$7F,$7E,$AE,$37,$D3,$44,$62,$76
	.dc.b	$73,$3C,$95,$CD,$A7,$33,$78,$78,$E3,$DF,$1C,$AB,$AB,$F2,$6F,$C1
	.dc.b	$7D,$E3,$99,$DF,$0F,$5E,$1C,$E3,$B1,$AB,$E6,$FD,$AC,$3C,$95,$F6
	.dc.b	$E3,$E4,$1E,$1B,$FF,$4F,$DF,$38,$36,$F9,$D7,$83,$0D,$9E,$6A,$F4
	.dc.b	$10,$F1,$5E,$9C,$4F,$D2,$AE,$78,$79,$09,$3E,$A9,$EC,$F0,$8F,$8E
	.dc.b	$FC,$2C,$3C,$C7,$96,$FE,$35,$7B,$67,$D3,$87,$95,$F8,$47,$C0,$29
	.dc.b	$A3,$E1,$1D,$CD,$47,$9E,$7D,$51,$C2,$BC,$0F,$DC,$AC,$D6,$E3,$91
	.dc.b	$D4,$F8,$66,$27,$8B,$EB,$9E,$37,$EE,$AF,$83,$3D,$E7,$D5,$1E,$DA
	.dc.b	$F8,$31,$8A,$F9,$35,$DC,$FF,$22,$14,$FC,$A1,$89,$FD,$63,$67,$FC
	.dc.b	$35,$18,$8D,$90,$F9,$06,$CF,$F1,$53,$B7,$27,$89,$F2,$BF,$CB,$BE
	.dc.b	$07,$DA,$FC,$CA,$F3,$E6,$9C,$CF,$C7,$C3,$DD,$87,$FB,$AA,$73,$87
	.dc.b	$4A,$FB,$37,$C5,$F9,$95,$CA,$31,$1F,$24,$C9,$E3,$29,$93,$9D,$76
	.dc.b	$9E,$4D,$8F,$B7,$3D,$2F,$CB,$1C,$9F,$C8,$46,$70,$F8,$57,$CA,$B5
	.dc.b	$1F,$02,$FA,$47,$86,$F8,$C6,$1E,$B1,$E4,$38,$9F,$24,$FF,$4A,$5F
	.dc.b	$77,$07,$C9,$AE,$CF,$03,$07,$0A,$ED,$1A,$8F,$13,$E8,$7E,$31,$83
	.dc.b	$AE,$1F,$C1,$6B,$C1,$1D,$09,$19,$AE,$B1,$E4,$9D,$CF,$D9,$BF,$69
	.dc.b	$F0,$27,$CD,$1C,$5B,$87,$E0,$5A,$79,$27,$8E,$1E,$78,$F8,$C7,$9C
	.dc.b	$F1,$CF,$17,$15,$E1,$9F,$F5,$45,$F9,$2B,$CF,$48,$C3,$F2,$5A,$F2
	.dc.b	$8E,$97,$E2,$8F,$8B,$18,$AF,$24,$E6,$7A,$61,$DB,$0F,$74,$73,$3F
	.dc.b	$1B,$1C,$5E,$67,$BE,$BC,$F1,$E6,$C3,$E8,$9E,$4B,$EC,$FD,$4F,$C8
	.dc.b	$DD,$9F,$E9,$56,$E1,$FE,$A1,$6B,$FE,$A5,$6B,$FC,$21,$AF,$FB,$35
	.dc.b	$AF,$FB,$8D,$A9,$A4,$8F,$01,$23,$C1,$D9,$E3,$23,$82,$1E,$37,$2D
	.dc.b	$32,$79,$5C,$B4,$F6,$76,$7B,$C8,$E0,$87,$BD,$CB,$4C,$9F,$05,$CB
	.dc.b	$4F,$91,$B7,$CB,$23,$82,$1F,$2D,$CB,$4C,$9F,$35,$CB,$4F,$A9,$D9
	.dc.b	$F5,$C8,$E0,$87,$D7,$72,$D3,$27,$D9,$72,$D3,$F0,$1D,$9F,$83,$23
	.dc.b	$82,$1F,$83,$72,$D3,$27,$E1,$5C,$B4,$FC,$67,$67,$E3,$C8,$E0,$87
	.dc.b	$E3,$DC,$B4,$C9,$F9,$17,$2D,$3F,$31,$D9,$F9,$B2,$38,$21,$F9,$B7
	.dc.b	$2D,$32,$7E,$75,$CB,$4F,$D2,$79,$1F,$A7,$23,$82,$1F,$A7,$72,$D3
	.dc.b	$27,$EA,$5C,$B4,$FD,$87,$98,$7F,$D2,$3C,$75,$FE,$B6,$DE,$0F,$F0
	.dc.b	$D6,$EA,$C7,$17,$EC,$DB,$7F,$D0,$AE,$BF,$C6,$5B,$5C,$63,$C1,$7D
	.dc.b	$2B,$7C,$3F,$CE,$EF,$71,$2B,$CA,$43,$FC,$4F,$7C,$DB,$1C,$A6,$47
	.dc.b	$6C,$3C,$CD,$3F,$84,$4E,$9F,$79,$93,$F8,$53,$E1,$72,$4C,$3D,$0E
	.dc.b	$EF,$C5,$1F,$DC,$2F,$A3,$A9,$F2,$9E,$97,$4F,$1A,$EC,$FF,$1F,$7F
	.dc.b	$93,$11,$FA,$E4,$3F,$93,$39,$69,$93,$91,$4F,$82,$78,$89,$1F,$E0
	.dc.b	$6F,$1B,$DC,$F9,$C9,$87,$F5,$48,$CB,$CA,$33,$3F,$E0,$09,$CB,$F3
	.dc.b	$0F,$61,$FE,$DC,$72,$FC,$C3,$27,$F8,$A1,$CE,$1F,$2A,$BB,$1E,$19
	.dc.b	$F9,$DD,$7F,$86,$F8,$69,$18,$21,$FE,$32,$72,$FE,$1A,$3F,$22,$63
	.dc.b	$0F,$F1,$2C,$E2,$78,$C6,$1F,$F7,$9C,$65,$E4,$60,$86,$09,$1C,$5F
	.dc.b	$11,$D2,$BF,$AC,$57,$F6,$12,$47,$F6,$16,$BF,$D8,$5A,$FF,$61,$6B
	.dc.b	$C8,$E0,$43,$81,$DC,$FE,$C8,$FE,$B3,$E1,$7D,$55,$FB,$A8,$B5,$FE
	.dc.b	$10,$9B,$5C,$0E,$47,$F8,$45,$FD,$AF,$C2,$7E,$DB,$CC,$93,$FA,$7D
	.dc.b	$93,$A8,$C4,$E9,$FF,$32,$C6,$5F,$F3,$3C,$67,$1F,$F2,$BC,$E5,$FF
	.dc.b	$2C,$C6,$70,$FF,$12,$CE,$5F,$F1,$3C,$66,$BF,$CA,$33,$97,$FC,$A7
	.dc.b	$19,$7F,$C5,$51,$FC,$53,$F2,$37,$BB,$FE,$94,$9F,$E4,$11,$87,$01
	.dc.b	$C3,$B6,$BF,$E9,$C6,$BC,$A3,$8B,$C5,$B1,$A8,$FE,$60,$D7,$47,$81
	.dc.b	$C1,$3D,$EE,$24,$8D,$35,$F2,$35,$F2,$35,$F6,$B5,$F0,$BE,$4F,$06
	.dc.b	$47,$85,$AF,$BD,$AF,$BD,$AF,$C8,$6B,$EC,$6B,$F2,$7D,$DF,$82,$47
	.dc.b	$DC,$D7,$E6,$35,$F9,$8D,$7E,$AB,$5F,$92,$FC,$CF,$1A,$47,$C9,$6B
	.dc.b	$F5,$DA,$FD,$76,$BF,$80,$6B,$F5,$1A,$FE,$05,$FC,$41,$1F,$E8,$D6
	.dc.b	$F6,$B1,$1F,$EF,$A8,$4E,$35,$6B,$F8,$EB,$C3,$E1,$BE,$75,$F9,$59
	.dc.b	$91,$F8,$37,$F1,$24,$70,$43,$F2,$8E,$5A,$64,$FC,$B3,$97,$FD,$E7
	.dc.b	$F9,$13,$E1,$BF,$BB,$B6,$3F,$B1,$35,$61,$EA,$7F,$C1,$10,$F0,$F8
	.dc.b	$D4,$9E,$2F,$48,$C4,$72,$C7,$FC,$1D,$5D,$6F,$A4,$A3,$13,$AA,$E3
	.dc.b	$5A,$9C,$D7,$8A,$FB,$3E,$02,$1D,$CD,$1E,$CA,$E7,$7E,$29,$B3,$EC
	.dc.b	$9F,$87,$E2,$FC,$4F,$5A,$4F,$C4,$8C,$BF,$1A,$32,$F7,$3D,$51,$A7
	.dc.b	$18,$27,$78,$FA,$EB,$4F,$B6,$33,$8F,$C6,$9C,$BF,$22,$33,$7F,$4A
	.dc.b	$72,$FD,$48,$CB,$E2,$3E,$DF,$B1,$F7,$BE,$1A,$46,$08,$79,$9C,$BE
	.dc.b	$88,$FC,$29,$8C,$3E,$94,$E2,$78,$C6,$1F,$C4,$C6,$5E,$46,$08,$60
	.dc.b	$91,$C5,$F1,$1D,$30,$FC,$3F,$C2,$7E,$58,$91,$F9,$6F,$86,$AF,$CE
	.dc.b	$8B,$5F,$87,$9B,$5C,$AB,$F3,$B5,$FA,$09,$CC,$7E,$72,$2B,$F9,$38
	.dc.b	$B9,$A6,$C6,$1B,$3F,$A8,$F9,$CF,$D5,$FB,$F2,$30,$43,$F0,$6E,$5F
	.dc.b	$C2,$C7,$EC,$8C,$63,$F9,$F9,$C4,$F1,$8C,$3F,$B6,$8C,$BC,$8C,$10
	.dc.b	$C1,$23,$8B,$E2,$3A,$63,$FD,$8B,$6F,$EA,$E4,$8F,$EA,$8D,$7F,$B1
	.dc.b	$35,$FE,$C4,$D7,$91,$C0,$87,$F6,$9F,$6F,$FB,$19,$1F,$EC,$0D,$74
	.dc.b	$60,$86,$08,$7F,$71,$78,$3F,$E9,$91,$DB,$FD,$C9,$A7,$F7,$47,$04
	.dc.b	$70,$43,$FC,$02,$E5,$A6,$4F,$F0,$43,$CB,$A7,$A1,$23,$8B,$5D,$B5
	.dc.b	$F3,$47,$F8,$A1,$D9,$0D,$94,$DF,$57,$A4,$8F,$99,$AF,$91,$AF,$03
	.dc.b	$44,$34,$53,$47,$F1,$98,$47,$FA,$E9,$B1,$FD,$13,$93,$FA,$1A,$FC
	.dc.b	$47,$E6,$BF,$42,$B8,$11,$E1,$9F,$FB,$3E,$77,$7D,$27,$C1,$5B,$7F
	.dc.b	$93,$DF,$6A,$EB,$C1,$FD,$BF,$0E,$58,$7F,$B5,$CE,$53,$D3,$0F,$F6
	.dc.b	$D3,$C2,$F9,$4F,$F5,$06,$E1,$FE,$E1,$8E,$27,$3A,$E2,$EE,$FB,$CF
	.dc.b	$F6,$0A,$E2,$FB,$CC,$94,$F8,$93,$C5,$A7,$19,$DC,$78,$23,$C5,$5E
	.dc.b	$07,$47,$03,$FB,$25,$78,$27,$84,$77,$BF,$75,$F8,$B0,$F8,$D3,$C6
	.dc.b	$B6,$70,$8C,$D7,$09,$D6,$1E,$17,$A5,$7A,$67,$47,$CD,$8F,$4C,$EE
	.dc.b	$B9,$D6,$CF,$95,$3A,$3C,$98,$7C,$89,$DC,$7C,$03,$47,$BD,$CB,$F0
	.dc.b	$30,$CE,$1E,$18,$D5,$73,$3E,$26,$1E,$9C,$3D,$D1,$9B,$E3,$1E,$4A
	.dc.b	$E3,$1E,$43,$2F,$8C,$C9,$2F,$C0,$49,$F7,$47,$D6,$8C,$3C,$5F,$8F
	.dc.b	$1A,$8F,$96,$65,$C6,$1E,$E7,$EA,$E1,$F2,$F0,$F4,$57,$D4,$3C,$A7
	.dc.b	$C3,$3C,$93,$EB,$BF,$0C,$65,$F0,$BD,$C8,$66,$3C,$87,$81,$F8,$47
	.dc.b	$43,$6F,$CA,$C3,$33,$DA,$31,$3B,$3D,$A4,$C3,$EC,$9E,$DC,$3E,$A1
	.dc.b	$ED,$BE,$75,$E9,$25,$7C,$B7,$CB,$3C,$23,$66,$4F,$2B,$CD,$F1,$47
	.dc.b	$AD,$F2,$57,$85,$F9,$2F,$93,$0F,$C3,$C6,$1F,$9F,$87,$B6,$FA,$3F
	.dc.b	$2A,$7D,$B7,$F0,$23,$0F,$77,$57,$DE,$7E,$C1,$E2,$BE,$EE,$27,$71
	.dc.b	$87,$EC,$CF,$97,$0F,$BF,$7D,$CF,$87,$7E,$78,$F8,$6F,$93,$0F,$3C
	.dc.b	$EA,$3B,$3F,$8E,$C3,$B3,$F3,$A3,$F0,$47,$CB,$39,$1D,$2F,$D5,$87
	.dc.b	$92,$BE,$2D,$71,$3C,$95,$C5,$F8,$47,$7C,$3D,$38,$79,$27,$4F,$DB
	.dc.b	$7E,$24,$69,$F0,$57,$99,$FC,$BE,$1F,$0D,$F5,$98,$8F,$1B,$F8,$23
	.dc.b	$8C,$F3,$7A,$9C,$0F,$CA,$47,$07,$DE,$6B,$0F,$C0,$3C,$8E,$18,$7B
	.dc.b	$9F,$8E,$70,$8F,$B3,$5D,$DB,$1C,$30,$F8,$0F,$0A,$F8,$AF,$CF,$C3
	.dc.b	$E6,$C7,$D3,$7D,$D5,$E2,$9F,$24,$FC,$FA,$E7,$5C,$27,$E1,$5F,$E0
	.dc.b	$AB,$AC,$79,$A7,$4F,$C0,$9D,$57,$D1,$BF,$39,$DC,$B7,$DE,$BE,$14
	.dc.b	$66,$3E,$E4,$7B,$4C,$BB,$7C,$E7,$D3,$29,$F1,$DE,$31,$CE,$3E,$16
	.dc.b	$1F,$7D,$F3,$1F,$BA,$AF,$35,$70,$3B,$57,$9A,$7B,$35,$EC,$74,$BF
	.dc.b	$A9,$3C,$63,$6D,$76,$7C,$7B,$F1,$47,$DF,$C3,$E4,$B6,$33,$87,$D2
	.dc.b	$6D,$7C,$23,$A9,$E1,$3E,$31,$F0,$A7,$EA,$D7,$DE,$7D,$67,$8C,$CC
	.dc.b	$FE,$55,$B7,$F6,$2B,$F8,$01,$F1,$4B,$87,$F1,$2C,$3E,$3C,$78,$63
	.dc.b	$E7,$3E,$D9,$F8,$6E,$6B,$C2,$D3,$D2,$6C,$F4,$CF,$CD,$8F,$1C,$78
	.dc.b	$4F,$D6,$61,$EB,$78,$B5,$F9,$AD,$8E,$33,$E7,$9F,$4C,$7F,$17,$8F
	.dc.b	$E1,$D5,$EF,$9E,$B8,$7C,$3C,$3D,$B1,$E5,$AF,$6E,$1D,$67,$2D,$9E
	.dc.b	$A7,$F1,$02,$9F,$B3,$8D,$D7,$58,$CC,$FE,$02,$7E,$F5,$FC,$89,$F0
	.dc.b	$61,$F2,$1A,$7C,$0B,$E4,$FA,$27,$D5,$87,$B2,$3E,$43,$4F,$84,$EC
	.dc.b	$CD,$7F,$03,$AE,$73,$B8,$F8,$C6,$6B,$AB,$6B,$89,$FC,$10,$A7,$EB
	.dc.b	$27,$FC,$7D,$F0,$7F,$E2,$92,$4F,$F9,$55,$CB,$FE,$25,$73,$3F,$E5
	.dc.b	$27,$2F,$F8,$95,$CB,$FE,$13,$7D,$73,$E7,$8F,$A0,$7B,$5C,$CF,$9D
	.dc.b	$B1,$E7,$6B,$E7,$8F,$7B,$97,$FC,$D4,$FF,$AF,$71,$FF,$2E,$B7,$0F
	.dc.b	$F4,$9B,$70,$FF,$61,$3F,$E6,$A3,$06,$27,$8B,$D0,$86,$CA,$7F,$8B
	.dc.b	$CF,$07,$C3,$BF,$24,$EB,$09,$4E,$63,$B3,$5E,$33,$C2,$76,$D9,$D4
	.dc.b	$89,$3A,$79,$CE,$C8,$E0,$87,$1F,$C8,$C4,$00

;--------------------------------------------------------------------------------
;	FDMUL.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fdmul_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FDMUL.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FDMUL.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fdmul_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdmul.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDMUL.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdmul.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDMUL.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fdmul_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3A,$F9,$AF,$EC,$ED,$2D,$FC,$E8,$CC,$76,$B4,$62
	.dc.b	$75,$2F,$A0,$F5,$BF,$DD,$32,$F7,$46,$23,$DF,$F1,$7F,$DD,$E3,$15
	.dc.b	$C9,$B8,$76,$AE,$97,$F2,$2F,$83,$63,$FC,$03,$5C,$30,$B3,$F1,$AF
	.dc.b	$97,$CF,$F8,$63,$E8,$37,$0F,$F0,$43,$C7,$03,$70,$D4,$74,$3D,$93
	.dc.b	$E2,$7E,$44,$6E,$BC,$91,$E8,$36,$71,$3E,$56,$17,$1F,$81,$3C,$2B
	.dc.b	$BD,$71,$36,$59,$DB,$D0,$D4,$FC,$E3,$E0,$57,$9F,$E6,$FF,$5F,$E5
	.dc.b	$FF,$87,$BE,$13,$FC,$4F,$18,$BF,$F1,$4D,$7F,$15,$8C,$47,$DE,$9D
	.dc.b	$F8,$9F,$94,$8C,$09,$DB,$FE,$1C,$8F,$E2,$0F,$E3,$23,$2F,$2B,$FC
	.dc.b	$5C,$7F,$02,$ED,$B5,$F9,$EA,$FF,$1C,$CE,$5E,$51,$97,$B4,$FA,$DF
	.dc.b	$14,$7F,$6F,$C3,$D4,$DC,$3D,$35,$BF,$9C,$7E,$2D,$FE,$46,$BF,$E6
	.dc.b	$91,$8C,$35,$39,$BF,$F2,$6C,$F0,$F9,$8C,$D7,$E9,$23,$83,$FA,$A6
	.dc.b	$BF,$E5,$26,$BF,$E4,$BA,$E8,$A1,$EA,$7F,$1C,$8F,$A7,$7E,$08,$C4
	.dc.b	$7D,$09,$EB,$87,$BB,$DF,$F8,$5A,$7F,$2C,$6E,$9F,$C1,$67,$E1,$F3
	.dc.b	$E3,$93,$E7,$93,$1F,$9A,$FD,$2E,$6E,$24,$8F,$19,$F3,$BE,$47,$AF
	.dc.b	$CB,$48,$CE,$7F,$CF,$23,$C3,$E6,$7D,$5C,$9F,$B7,$FF,$96,$6F,$F8
	.dc.b	$7C,$63,$BB,$F8,$6B,$63,$33,$B7,$F0,$7A,$BD,$3F,$69,$23,$C3,$0F
	.dc.b	$D0,$FE,$0F,$E9,$B2,$72,$FC,$4F,$27,$F1,$11,$FC,$F3,$C2,$FE,$3A
	.dc.b	$D3,$F8,$FF,$8C,$8C,$B6,$3F,$27,$EA,$FC,$E2,$3E,$2F,$57,$E8,$F9
	.dc.b	$72,$3C,$AF,$D4,$F4,$3C,$36,$F6,$BA,$92,$BC,$2F,$E9,$BC,$7F,$C3
	.dc.b	$F2,$FF,$92,$E7,$F6,$2D,$E2,$FD,$C4,$FF,$9A,$A7,$2F,$F4,$D7,$FD
	.dc.b	$1E,$D7,$F8,$0B,$5F,$DD,$35,$7A,$1F,$DF,$9F,$E8,$2D,$7F,$A3,$B5
	.dc.b	$FF,$57,$B5,$EB,$F2,$BE,$92,$3F,$C0,$E7,$FC,$04,$F9,$3E,$57,$DC
	.dc.b	$47,$4F,$D4,$A7,$C6,$52,$4F,$91,$F8,$DF,$2B,$23,$B3,$5F,$EB,$BF
	.dc.b	$63,$FE,$15,$7F,$8C,$87,$19,$7F,$7F,$6F,$87,$FE,$CB,$EF,$F8,$13
	.dc.b	$9E,$FF,$C3,$D7,$3B,$FF,$67,$B5,$FF,$69,$35,$FF,$6A,$BD,$1F,$F6
	.dc.b	$BD,$3E,$37,$F2,$6D,$8F,$D7,$9E,$5A,$FD,$EE,$BF,$DA,$70,$FB,$FF
	.dc.b	$81,$E0,$CF,$95,$2F,$F6,$74,$9D,$46,$27,$9F,$C5,$48,$DB,$DB,$E2
	.dc.b	$71,$1F,$8A,$8B,$87,$E3,$A7,$F5,$3F,$63,$FE,$3B,$8F,$F1,$33,$76
	.dc.b	$FF,$79,$36,$FF,$DE,$BA,$FF,$33,$7F,$DE,$F1,$F8,$27,$7C,$7F,$B2
	.dc.b	$8C,$46,$A3,$07,$EE,$27,$FC,$3D,$1F,$69,$B7,$E2,$9F,$8F,$EB,$CA
	.dc.b	$F8,$6F,$CA,$F9,$8C,$47,$E9,$A2,$D7,$EA,$A7,$E7,$FC,$C7,$D0,$F8
	.dc.b	$C9,$7F,$32,$32,$FC,$E8,$CB,$FB,$08,$FE,$1B,$F1,$1F,$BD,$8F,$B6
	.dc.b	$D7,$F2,$2D,$BF,$F2,$77,$E2,$BF,$E1,$09,$DF,$5F,$F1,$4A,$E1,$03
	.dc.b	$FD,$06,$7F,$8D,$1C,$B8,$3F,$3B,$3F,$DD,$EB,$AF,$C3,$79,$2F,$F1
	.dc.b	$7F,$1B,$2F,$F0,$EF,$E3,$BE,$13,$11,$FC,$2A,$2D,$FF,$0E,$9F,$ED
	.dc.b	$5F,$90,$34,$A4,$BF,$E8,$FE,$92,$58,$66,$7F,$AD,$47,$17,$6D,$73
	.dc.b	$1F,$CD,$BE,$3F,$FC,$F9,$1F,$B8,$6E,$3F,$E7,$56,$F2,$7C,$59,$FF
	.dc.b	$5A,$FC,$D7,$1E,$1F,$E7,$B5,$CA,$74,$1C,$3F,$EB,$D8,$D9,$FD,$1B
	.dc.b	$99,$F1,$F4,$E9,$3E,$EF,$CA,$9E,$47,$F8,$35,$B1,$98,$C7,$91,$96
	.dc.b	$46,$4E,$51,$FB,$EC,$A4,$7E,$E9,$FE,$03,$F2,$38,$8F,$E7,$B1,$75
	.dc.b	$FE,$83,$3F,$C1,$B1,$FE,$90,$4A,$FE,$60,$D7,$FA,$2F,$C8,$FF,$0C
	.dc.b	$38,$3F,$C2,$3E,$6F,$F8,$79,$1F,$E1,$7F,$37,$5F,$F1,$AF,$84,$9F
	.dc.b	$EA,$C4,$8F,$F2,$07,$C6,$FF,$AA,$5F,$A9,$E5,$A3,$11,$FE,$52,$6B
	.dc.b	$FE,$4B,$6C,$6F,$8F,$FC,$3B,$1F,$EE,$7D,$3F,$CA,$CD,$8E,$47,$F7
	.dc.b	$CF,$9B,$FE,$8F,$F3,$BF,$A7,$FA,$3F,$F2,$C4,$63,$14,$63,$CC,$56
	.dc.b	$FB,$BF,$15,$7F,$E1,$1A,$D4,$E6,$FF,$65,$3C,$E7,$31,$D2,$B8,$FB
	.dc.b	$3F,$93,$9C,$92,$3A,$E1,$E0,$AE,$07,$4A,$E1,$3A,$C3,$C1,$2C,$3C
	.dc.b	$13,$98,$E2,$66,$B8,$BE,$6B,$ED,$87,$86,$31,$87,$4C,$3D,$D8,$73
	.dc.b	$AD,$57,$7B,$E3,$3E,$C8,$E1,$18,$79,$9E,$2A,$F2,$BE,$B3,$77,$1D
	.dc.b	$4F,$17,$D0,$66,$FC,$35,$CA,$32,$70,$AD,$5F,$9C,$F2,$46,$49,$5C
	.dc.b	$B0,$F8,$71,$EB,$AE,$2F,$A2,$7F,$0B,$3A,$BF,$01,$30,$F3,$3F,$1A
	.dc.b	$B8,$61,$E1,$69,$EF,$3C,$2E,$5F,$01,$F1,$6B,$B5,$74,$26,$1F,$4E
	.dc.b	$32,$F9,$A7,$8E,$1E,$67,$18,$79,$67,$84,$7C,$D3,$F8,$EC,$7E,$7D
	.dc.b	$B7,$F8,$F6,$CF,$FB,$F2,$75,$1F,$8B,$32,$7F,$26,$3D,$26,$2B,$EA
	.dc.b	$D7,$C4,$BF,$3C,$FC,$03,$E1,$9F,$62,$77,$7E,$89,$DC,$62,$7E,$21
	.dc.b	$EC,$AF,$86,$E0,$E8,$7C,$2C,$3E,$66,$1F,$4E,$30,$78,$1F,$B3,$7D
	.dc.b	$A7,$75,$E7,$AE,$D8,$78,$DA,$76,$BE,$87,$CD,$71,$7C,$DE,$D5,$CD
	.dc.b	$E0,$FA,$30,$FA,$D1,$8B,$EC,$D3,$99,$E0,$8F,$CB,$37,$0F,$6D,$F7
	.dc.b	$8F,$69,$30,$FC,$96,$9F,$C9,$63,$FD,$63,$9F,$F2,$37,$F9,$57,$CE
	.dc.b	$FF,$28,$D2,$47,$F3,$47,$FD,$E3,$3F,$CB,$8F,$E5,$FB,$FF,$3B,$9F
	.dc.b	$F3,$B6,$FF,$CE,$DF,$E7,$ED,$DF,$FD,$1E,$FF,$43,$7F,$D2,$3C,$1F
	.dc.b	$D2,$A7,$FA,$69,$FE,$52,$E1,$C4,$FF,$A5,$B8,$BF,$A9,$4F,$FA,$51
	.dc.b	$AF,$FA,$53,$93,$FA,$E6,$76,$3F,$CB,$7C,$98,$7F,$D2,$F3,$76,$68
	.dc.b	$AF,$ED,$B5,$FE,$E7,$E7,$FE,$E6,$DE,$7F,$EE,$6F,$F7,$13,$05,$CD
	.dc.b	$97,$FC,$00,$A7,$3F,$F8,$02,$BF,$D1,$3D,$1F,$E0,$47,$FC,$EE,$D7
	.dc.b	$FC,$2D,$D3,$B7,$FC,$1F,$D5,$FE,$0E,$7F,$C7,$4D,$7F,$C2,$51,$FE
	.dc.b	$10,$7F,$D0,$31,$FE,$1C,$77,$1F,$E1,$3D,$BF,$C5,$59,$79,$A5,$DB
	.dc.b	$FE,$2F,$EE,$CE,$9F,$E4,$78,$FF,$19,$76,$E2,$B7,$DD,$FE,$2F,$C3
	.dc.b	$95,$71,$9F,$F1,$9D,$7F,$90,$2F,$FD,$8E,$D8,$FF,$63,$F7,$DE,$3F
	.dc.b	$A3,$37,$8D,$3F,$E4,$FA,$FF,$4D,$F8,$3F,$E5,$3C,$3F,$C6,$CD,$8F
	.dc.b	$F2,$E7,$85,$FE,$63,$9F,$F2,$D4,$7F,$8D,$A3,$FC,$BA,$EB,$C2,$FF
	.dc.b	$32,$11,$FF,$33,$90,$FF,$30,$FC,$47,$F9,$D9,$FF,$71,$EF,$FE,$76
	.dc.b	$DB,$FD,$CF,$E2,$65,$FF,$74,$C6,$71,$FF,$72,$CF,$FA,$11,$FF,$73
	.dc.b	$F8,$B6,$FF,$DC,$FC,$77,$83,$F2,$33,$77,$FB,$14,$F2,$3F,$D4,$4F
	.dc.b	$FA,$FF,$DF,$FF,$4D,$1F,$EA,$AF,$22,$47,$F2,$F3,$83,$FE,$A0,$71
	.dc.b	$3F,$EA,$47,$27,$FB,$19,$EA,$76,$F2,$7F,$D5,$D8,$FE,$93,$CA,$FF
	.dc.b	$5C,$E9,$FF,$00,$1F,$F5,$F5,$FF,$B0,$08,$7F,$AF,$CF,$F6,$4F,$57
	.dc.b	$FB,$2D,$70,$7F,$B4,$1F,$F6,$6F,$99,$98,$FC,$96,$1F,$B2,$F3,$3F
	.dc.b	$D9,$F5,$C3,$1F,$F4,$C4,$FE,$DA,$F8,$47,$FB,$9B,$CD,$FF,$70,$E7
	.dc.b	$2F,$FD,$C3,$E7,$FF,$B9,$DF,$F7,$11,$F1,$5F,$F7,$63,$83,$FD,$FB
	.dc.b	$3F,$EF,$43,$FD,$05,$97,$FB,$DE,$39,$C7,$FB,$BB,$E4,$FF,$DD,$7E
	.dc.b	$8F,$CF,$CF,$F8,$4B,$78,$8D,$3E,$A5,$BD,$27,$A3,$86,$C7,$D9,$6B
	.dc.b	$FD,$A5,$AF,$F9,$8A,$9E,$9F,$72,$47,$F6,$FC,$3F,$8D,$C7,$8F,$92
	.dc.b	$4E,$6B,$FD,$CB,$5F,$C8,$27,$C5,$94,$AF,$13,$5F,$F3,$BF,$C4,$FB
	.dc.b	$48,$F3,$7E,$03,$F2,$38,$3E,$BE,$1F,$EA,$6D,$BE,$EB,$63,$3F,$0B
	.dc.b	$BA,$FF,$6A,$7A,$DF,$B6,$4D,$27,$77,$CF,$F7,$7F,$95,$CF,$FB,$6F
	.dc.b	$1F,$C4,$C6,$23,$F1,$03,$54,$66,$77,$EC,$78,$67,$3E,$CD,$A3,$4C
	.dc.b	$79,$9E,$7C,$3C,$BE,$D7,$DC,$CC,$D2,$47,$D3,$B7,$B9,$D4,$91,$E9
	.dc.b	$DB,$F3,$B2,$F7,$3C,$5E,$F4,$AD,$1E,$B9,$47,$9F,$DE,$F9,$5B,$BD
	.dc.b	$F7,$B7,$FC,$0E,$30,$D8,$CE,$1F,$E3,$59,$CB,$E7,$F8,$1F,$84,$47
	.dc.b	$C3,$C3,$F2,$BE,$07,$E3,$77,$49,$CE,$88,$F9,$5F,$05,$F2,$FE,$0B
	.dc.b	$EC,$12,$BE,$23,$F6,$FE,$0F,$E8,$12,$39,$F9,$3F,$85,$F4,$DF,$0B
	.dc.b	$F8,$99,$FE,$AF,$F0,$AC,$34,$D8,$F8,$F5,$C1,$E5,$5F,$B2,$8C,$47
	.dc.b	$F9,$55,$AF,$FB,$F1,$A8,$FE,$C5,$1E,$18,$DE,$FF,$D8,$23,$03,$B7
	.dc.b	$F1,$5F,$0D,$F8,$BE,$F9,$87,$C6,$8C,$BF,$85,$7F,$38,$47,$F1,$1F
	.dc.b	$0F,$F9,$F2,$38,$21,$F9,$EE,$2F,$CE,$C7,$E6,$4F,$CD,$7C,$2F,$E5
	.dc.b	$FE,$23,$E3,$67,$EB,$F8,$97,$9D,$ED,$C6,$F8,$B6,$73,$1F,$06,$53
	.dc.b	$97,$85,$F4,$95,$70,$51,$98,$C4,$F2,$9F,$13,$E1,$30,$F2,$7A,$F0
	.dc.b	$FF,$48,$8C,$1F,$4E,$BA,$4F,$1B,$F2,$4F,$9D,$AF,$8A,$B9,$5F,$96
	.dc.b	$F8,$CE,$CF,$19,$4E,$E7,$1C,$32,$D1,$EB,$7B,$30,$F9,$4D,$8F,$C4
	.dc.b	$3D,$F0,$F5,$36,$B9,$E3,$F0,$31,$E1,$7E,$0C,$7C,$AF,$78,$C9,$27
	.dc.b	$E2,$D7,$92,$BE,$95,$7D,$B3,$CC,$57,$2F,$03,$FD,$68,$FE,$DF,$E9
	.dc.b	$FF,$C1,$0D,$8C,$F3,$A7,$2D,$8E,$39,$7D,$57,$B3,$5E,$C6,$2F,$AC
	.dc.b	$66,$3E,$D3,$5F,$39,$C0,$86,$A3,$89,$1F,$0B,$5E,$C6,$08,$60,$87
	.dc.b	$43,$B9,$D4,$D9,$9B,$F8,$EE,$4E,$25,$35,$3E,$98,$DC,$74,$23,$82
	.dc.b	$46,$C9,$3E,$99,$E6,$D7,$B1,$0E,$71,$C1,$F3,$99,$9E,$B3,$C6,$36
	.dc.b	$D8,$CC,$70,$32,$57,$2F,$B8,$C9,$5C,$BE,$63,$84,$F8,$DD,$35,$D1
	.dc.b	$89,$F1,$46,$63,$D0,$D7,$B9,$C0,$86,$A3,$89,$1F,$0B,$5E,$C6,$08
	.dc.b	$60,$87,$43,$B9,$D4,$D9,$99,$F7,$B9,$38,$94,$D4,$FA,$63,$71,$D0
	.dc.b	$8E,$09,$1B,$24,$FA,$67,$9B,$5E,$C4,$39,$C7,$07,$CE,$66,$7A,$CF
	.dc.b	$18,$DB,$63,$31,$C0,$C9,$5C,$BE,$E3,$25,$72,$F9,$8E,$13,$E3,$74
	.dc.b	$D7,$46,$27,$C5,$19,$8F,$43,$5E,$E7,$02,$1A,$8E,$24,$7C,$2D,$7B
	.dc.b	$18,$21,$82,$1D,$0E,$E7,$53,$66,$67,$DE,$E4,$E2,$53,$53,$E9,$8D
	.dc.b	$C7,$42,$38,$24,$6C,$93,$E9,$9E,$6D,$7B,$1F,$BD,$AF,$E6,$1F,$3C
	.dc.b	$F2,$3F,$B8,$C6,$1B,$8F,$F8,$5E,$75,$5F,$D9,$EB,$4D,$9D,$57,$08
	.dc.b	$57,$07,$FB,$CD,$70,$6D,$70,$7F,$A7,$7C,$57,$FA,$71,$34,$8F,$FA
	.dc.b	$81,$AF,$FB,$55,$AF,$F5,$73,$89,$B2,$1B,$29,$B7,$8B,$E2,$32,$53
	.dc.b	$81,$0E,$06,$8F,$F1,$03,$C9,$A6,$4F,$F2,$03,$96,$99,$3C,$E7,$53
	.dc.b	$C0,$76,$30,$74,$29,$C8,$A7,$23,$81,$0E,$64,$39,$19,$29,$93,$A1
	.dc.b	$92,$99,$3B,$15,$F6,$38,$39,$90,$E2,$53,$46,$08,$60,$8F,$42,$3F
	.dc.b	$1D,$AF,$C0,$6B,$E7,$38,$9B,$21,$B2,$9B,$78,$BE,$B3,$25,$38,$10
	.dc.b	$E0,$68,$F6,$3C,$9A,$64,$F6,$B9,$69,$93,$CE,$75,$3C,$07,$63,$07
	.dc.b	$42,$9C,$8A,$72,$38,$10,$E6,$43,$91,$92,$99,$3A,$19,$29,$93,$B1
	.dc.b	$5F,$63,$83,$99,$0E,$25,$34,$60,$86,$08,$F4,$23,$F1,$DA,$FC,$06
	.dc.b	$BE,$73,$89,$B2,$1B,$29,$B7,$8B,$EB,$32,$53,$81,$0E,$06,$8F,$63
	.dc.b	$C9,$A6,$4F,$6B,$96,$99,$3C,$E7,$53,$C0,$76,$30,$74,$29,$C8,$A7
	.dc.b	$23,$81,$0E,$64,$39,$19,$29,$93,$A1,$FC,$17,$2F,$F3,$D3,$FE,$81
	.dc.b	$AF,$F3,$C3,$6B,$FC,$F6,$7F,$4F,$7F,$C8,$7A,$7F,$A3,$5F,$F5,$13
	.dc.b	$FE,$97,$7F,$D2,$99,$FF,$AA,$B4,$FF,$58,$47,$FA,$6B,$6F,$F5,$7E
	.dc.b	$3F,$E9,$83,$FD,$33,$3F,$E9,$D3,$73,$CB,$5F,$F4,$C5,$F3,$8E,$F3
	.dc.b	$AC,$3C,$04,$70,$48,$FF,$3C,$46,$5E,$24,$3F,$D5,$11,$FE,$9F,$8F
	.dc.b	$64,$7F,$B7,$8F,$0D,$47,$FD,$7F,$3E,$77,$95,$FA,$EF,$CD,$1C,$4D
	.dc.b	$E1,$F0,$23,$33,$F0,$23,$FD,$A3,$5C,$EB,$C6,$7A,$1F,$53,$CF,$0F
	.dc.b	$53,$A3,$11,$AA,$F1,$61,$F3,$E6,$1F,$2C,$95,$EE,$AF,$05,$70,$9F
	.dc.b	$39,$C2,$7B,$C7,$9E,$31,$3F,$1C,$8F,$CB,$79,$E1,$EE,$3E,$1B,$ED
	.dc.b	$33,$7F,$2E,$FC,$58,$7C,$58,$C4,$FC,$03,$77,$E8,$71,$3B,$8C,$1E
	.dc.b	$17,$15,$F1,$A7,$C1,$18,$7C,$F3,$F2,$0F,$B3,$7E,$7B,$F6,$C6,$1F
	.dc.b	$1B,$D4,$A7,$D1,$9E,$AE,$30,$F8,$CF,$CA,$3C,$A6,$EF,$EF,$9E,$37
	.dc.b	$EF,$1D,$4D,$C7,$2C,$3F,$26,$FD,$57,$E3,$C7,$7C,$3E,$6D,$74,$8C
	.dc.b	$61,$EA,$7D,$C7,$E1,$EF,$A6,$1F,$87,$BF,$81,$3E,$7C,$3B,$C6,$1E
	.dc.b	$F1,$E4,$8E,$EF,$5A,$F2,$57,$D4,$8F,$09,$CD,$ED,$7F,$21,$F8,$F1
	.dc.b	$CE,$3C,$D1,$F7,$AF,$EC,$61,$EB,$AD,$1C,$CF,$A8,$7C,$C9,$E7,$3F
	.dc.b	$96,$C3,$CB,$3B,$3E,$03,$88,$F6,$DF,$8E,$FB,$47,$18,$F1,$C7,$D5
	.dc.b	$AF,$B0,$7B,$EB,$E2,$5F,$BE,$3C,$B7,$F1,$8E,$4F,$86,$3D,$87,$E4
	.dc.b	$0F,$39,$F9,$2C,$3E,$9C,$7B,$27,$E0,$BD,$F0,$F3,$34,$F8,$18,$7A
	.dc.b	$2F,$96,$1E,$D7,$E7,$CE,$8E,$17,$D1,$F8,$98,$7D,$87,$F2,$11,$E8
	.dc.b	$3D,$53,$D3,$0F,$87,$3F,$72,$72,$FE,$60,$F8,$A7,$4B,$F8,$AF,$4A
	.dc.b	$E4,$FC,$3C,$3E,$66,$1F,$02,$3F,$2A,$F8,$DF,$CB,$CF,$C8,$7B,$3F
	.dc.b	$21,$F0,$47,$18,$D4,$3D,$06,$68,$99,$3C,$4C,$46,$E6,$CE,$63,$85
	.dc.b	$72,$78,$D7,$0C,$9E,$5F,$08,$E7,$5B,$A5,$F3,$56,$F4,$78,$47,$96
	.dc.b	$30,$F6,$3C,$38,$E9,$F3,$19,$39,$CE,$67,$8C,$78,$30,$F1,$D7,$1F
	.dc.b	$97,$79,$94,$95,$CF,$0C,$BD,$66,$BC,$CF,$95,$0F,$82,$F2,$36,$38
	.dc.b	$57,$47,$81,$EF,$9D,$99,$CB,$E2,$5F,$67,$E3,$5F,$3A,$E2,$D8,$DD
	.dc.b	$FC,$3F,$73,$FD,$A1,$1F,$35,$B9,$7D,$C6,$DF,$E0,$7D,$D7,$3F,$1E
	.dc.b	$7F,$99,$0E,$8A,$66,$BE,$75,$F2,$53,$B7,$E9,$35,$F7,$C6,$FE,$69
	.dc.b	$EE,$7D,$EE,$8E,$3E,$F5,$C0,$FD,$DA,$DE,$1F,$75,$7B,$BB,$24,$65
	.dc.b	$FC,$27,$C6,$3D,$97,$C6,$FF,$B9,$69,$39,$8C,$36,$3F,$DD,$4D,$7F
	.dc.b	$46,$70,$E4,$78,$9C,$1B,$1C,$1A,$F0,$94,$62,$3F,$5B,$F6,$6F,$57
	.dc.b	$F5,$9C,$BF,$AD,$DD,$F8,$9F,$C6,$E6,$FC,$E5,$ED,$FD,$14,$EB,$8B
	.dc.b	$8C,$74,$F4,$AE,$4A,$E7,$F0,$18,$86,$0A,$DD,$6A,$FF,$55,$F4,$6F
	.dc.b	$8C,$E7,$A3,$F1,$47,$AF,$CE,$B5,$7F,$B5,$AD,$4E,$6F,$8D,$F0,$AE
	.dc.b	$B1,$89,$D4,$60,$E5,$FA,$9B,$E8,$73,$AF,$EA,$3E,$99,$EC,$79,$29
	.dc.b	$F3,$3E,$D9,$CE,$1C,$A3,$15,$B6,$C6,$4F,$E1,$13,$C2,$FF,$79,$A3
	.dc.b	$0E,$B5,$AC,$39,$34,$95,$C2,$5F,$43,$DA,$3F,$7F,$F6,$8F,$83,$F4
	.dc.b	$5F,$D3,$8C,$3F,$00,$C4,$69,$FE,$39,$1A,$9E,$D1,$A3,$33,$F5,$BF
	.dc.b	$53,$7C,$AE,$6B,$F8,$79,$FC,$FF,$D2,$FE,$81,$E9,$3D,$0F,$E7,$BE
	.dc.b	$E7,$F3,$D8,$C5,$7E,$82,$B7,$5A,$CA,$CE,$45,$70,$7F,$9E,$DF,$3A
	.dc.b	$FD,$4C,$2F,$8B,$AB,$AF,$94,$E7,$3D,$DA,$77,$BE,$DB,$B0,$ED,$39
	.dc.b	$AE,$15,$E3,$8E,$8F,$09,$DC,$7F,$5E,$C3,$C7,$7C,$F0,$F1,$CF,$9C
	.dc.b	$F0,$9E,$72,$9D,$30,$EE,$7B,$AB,$9D,$F6,$9D,$1F,$DB,$9C,$3D,$70
	.dc.b	$F2,$5F,$82,$B8,$B6,$7C,$38,$79,$5F,$09,$C7,$0F,$0B,$F0,$4E,$D3
	.dc.b	$C6,$FA,$C7,$0B,$EA,$D3,$AE,$1E,$48,$CC,$FC,$99,$EC,$74,$38,$3D
	.dc.b	$0D,$47,$0B,$F8,$53,$D2,$FC,$51,$E6,$8F,$A0,$64,$E6,$E5,$F8,$E7
	.dc.b	$CF,$AE,$AE,$DA,$78,$63,$BB,$63,$E1,$1C,$63,$27,$57,$D6,$64,$DE
	.dc.b	$1F,$36,$7A,$CE,$EF,$D9,$5A,$C3,$E3,$3E,$39,$EA,$FC,$5A,$E2,$D8
	.dc.b	$CB,$DC,$F8,$31,$E7,$6C,$77,$7A,$1D,$F0,$F9,$2F,$33,$07,$C8,$30
	.dc.b	$72,$29,$F2,$B0,$F6,$4F,$89,$B1,$CA,$BC,$EF,$94,$E1,$7E,$E8,$F3
	.dc.b	$46,$2F,$C6,$EC,$F9,$75,$D1,$F1,$19,$C3,$E4,$B8,$8D,$3C,$CF,$51
	.dc.b	$B3,$E0,$9B,$BF,$16,$1F,$17,$0F,$1C,$F8,$27,$2F,$C0,$31,$87,$9A
	.dc.b	$FB,$C6,$0E,$EE,$30,$F2,$C7,$8E,$FE,$1C,$E6,$3C,$66,$70,$F9,$58
	.dc.b	$78,$23,$D1,$5E,$4B,$E7,$1D,$E3,$9C,$EE,$FA,$46,$E7,$A9,$BB,$EB
	.dc.b	$87,$A6,$F8,$11,$C1,$27,$D7,$19,$6C,$66,$BC,$2D,$AE,$B3,$A2,$4E
	.dc.b	$9B,$1A,$8C,$11,$C1,$23,$F1,$4D,$7E,$C4,$F7,$72,$43,$27,$63,$07
	.dc.b	$CD,$29,$F0,$CB,$87,$CC,$23,$F2,$1B,$5F,$2C,$95,$F8,$72,$4F,$E1
	.dc.b	$67,$71,$88,$FB,$2D,$C3,$EA,$B7,$0F,$99,$7F,$06,$BF,$B0,$1F,$E4
	.dc.b	$BB,$FF,$26,$3F,$D6,$4C,$1F,$D8,$3D,$2F,$F5,$7F,$47,$F9,$45,$B1
	.dc.b	$9F,$3F,$FC,$6C,$D9,$D7,$A1,$FE,$A6,$8E,$4D,$7F,$CA,$75,$FE,$A8
	.dc.b	$CF,$64,$8F,$F6,$51,$FE,$D3,$CF,$FD,$E5,$97,$FB,$65,$A7,$FB,$41
	.dc.b	$FF,$6B,$C2,$BF,$D9,$D3,$FE,$CB,$8C,$5F,$FB,$61,$FF,$68,$FA,$7F
	.dc.b	$ED,$68,$FF,$76,$47,$FB,$D6,$7F,$DF,$53,$FE,$F1,$8F,$F6,$9C,$E9
	.dc.b	$FF,$6D,$BC,$E7,$B5,$78,$2B,$85,$78,$30,$FF,$7E,$BF,$EE,$96,$E1
	.dc.b	$FE,$0F,$BF,$F6,$ED,$FF,$B7,$EF,$C9,$5F,$EF,$4B,$FE,$F3,$3D,$5F
	.dc.b	$F7,$FD,$79,$9A,$7F,$BA,$AB,$C3,$3B,$8F,$01,$CB,$0F,$55,$78,$1B
	.dc.b	$5E,$09,$FA,$26,$EF,$CF,$5F,$EE,$C3,$83,$E5,$C3,$EA,$4F,$FB,$CE
	.dc.b	$7E,$1D,$FC,$5B,$F1,$B6,$7C,$E7,$C5,$3B,$D7,$27,$EA,$9F,$81,$3A
	.dc.b	$4F,$2B,$F1,$C7,$D7,$3E,$01,$D3,$0F,$94,$F1,$39,$9E,$B2,$57,$E2
	.dc.b	$23,$E3,$3F,$4E,$BE,$13,$9B,$F4,$57,$CA,$8F,$F2,$35,$76,$8E,$78
	.dc.b	$7D,$37,$4D,$9E,$91,$E3,$8F,$8B,$87,$AA,$3F,$16,$D8,$F9,$B1,$F1
	.dc.b	$1F,$CF,$1F,$A0,$7F,$27,$1F,$46,$33,$1F,$18,$F8,$57,$F7,$1F,$06
	.dc.b	$1F,$25,$F3,$9B,$3C,$E7,$CA,$29,$F5,$6F,$E1,$5F,$96,$7E,$2D,$F8
	.dc.b	$EB,$4F,$BC,$FC,$34,$6A,$3B,$57,$C2,$7C,$27,$03,$33,$C1,$F3,$9E
	.dc.b	$AA,$F7,$3F,$00,$F8,$85,$9F,$2B,$D2,$7E,$6B,$6B,$E0,$4F,$D8,$BF
	.dc.b	$89,$3E,$2C,$3E,$D4,$FE,$46,$F8,$BB,$C3,$C3,$5E,$72,$4E,$8F,$80
	.dc.b	$FE,$4C,$CE,$1E,$08,$CC,$7D,$E3,$FC,$E7,$E9,$33,$44,$B6,$D8,$86
	.dc.b	$6F,$47,$15,$A9,$E3,$39,$79,$57,$1C,$3B,$46,$0E,$EE,$15,$73,$7B
	.dc.b	$DF,$6A,$F1,$5F,$59,$DC,$75,$C3,$51,$D1,$AF,$23,$C0,$53,$89,$B3
	.dc.b	$3E,$23,$D1,$E3,$83,$D2,$A7,$49,$D5,$70,$AC,$BD,$AB,$83,$92,$1E
	.dc.b	$03,$E5,$57,$07,$15,$D1,$B5,$9C,$BE,$25,$F4,$9C,$9E,$CF,$3D,$E8
	.dc.b	$FC,$06,$C6,$63,$E1,$FA,$0F,$4B,$51,$8F,$4A,$56,$C8,$79,$E9,$1A
	.dc.b	$8F,$55,$F6,$7D,$9E,$8B,$CA,$FA,$7C,$EC,$1E,$96,$1B,$7F,$52,$F8
	.dc.b	$37,$D2,$D1,$1C,$1F,$6B,$CF,$7A,$4F,$4B,$FD,$CB,$29,$CC,$61,$B1
	.dc.b	$FE,$EA,$6B,$F9,$D3,$87,$1B,$D2,$E0,$D8,$E0,$D7,$84,$A3,$11,$FA
	.dc.b	$7F,$3D,$E9,$7E,$97,$97,$F4,$DB,$BD,$2F,$8D,$D5,$F9,$4B,$DB,$FA
	.dc.b	$59,$D7,$17,$18,$E9,$E8,$DC,$95,$CF,$D2,$C4,$30,$56,$EB,$57,$FA
	.dc.b	$2F,$49,$E8,$F3,$D1,$E9,$3D,$2E,$75,$AB,$FD,$C5,$6A,$73,$7C,$6F
	.dc.b	$85,$75,$8C,$4E,$A3,$07,$2F,$39,$E9,$67,$5F,$D2,$7A,$2F,$47,$C9
	.dc.b	$4F,$4B,$DB,$39,$C3,$94,$62,$B6,$D8,$C9,$FC,$02,$78,$5F,$F0,$0D
	.dc.b	$18,$75,$AD,$61,$C9,$A7,$2A,$F2,$C6,$5F,$DD,$7A,$2F,$47,$CF,$5F
	.dc.b	$4A,$30,$FB,$CC,$46,$9F,$E2,$71,$A9,$F2,$5F,$F3,$09,$CD,$6E,$7B
	.dc.b	$B7,$0F,$E4,$D1,$FE,$7F,$3F,$D0,$7E,$67,$F9,$B5,$BE,$5F,$FA,$2D
	.dc.b	$D1,$0F,$F4,$21,$A3,$FD,$0A,$7F,$A0,$C9,$1F,$E8,$27,$FD,$00,$62
	.dc.b	$3F,$CF,$97,$FE,$95,$8E,$44,$3B,$79,$DF,$E8,$67,$8D,$73,$7F,$D0
	.dc.b	$E6,$2B,$9B,$FE,$9B,$39,$1E,$63,$99,$E1,$38,$10,$E6,$4B,$FF,$48
	.dc.b	$B9,$6B,$9F,$3F,$FD,$27,$3F,$16,$3B,$12,$FD,$0E,$F0,$F7,$3E,$2C
	.dc.b	$3D,$58,$74,$26,$1D,$27,$E5,$CE,$A3,$11,$EC,$BE,$B5,$C2,$FA,$CF
	.dc.b	$84,$F7,$1E,$13,$E2,$1E,$1C,$3E,$3E,$1F,$7E,$30,$7B,$F0,$EE,$E0
	.dc.b	$E5,$1B,$7B,$3E,$5A,$F0,$E1,$E4,$AE,$35,$AB,$EE,$7C,$79,$D5,$70
	.dc.b	$9F,$3E,$1E,$B7,$CE,$71,$3C,$91,$BC,$3C,$11,$AA,$E6,$D3,$9E,$1E
	.dc.b	$28,$CC,$FA,$5B,$87,$8E,$B7,$7D,$63,$CF,$1F,$49,$AF,$D1,$36,$77
	.dc.b	$AF,$19,$D5,$CB,$4C,$C7,$76,$C7,$53,$F4,$A7,$FB,$60,$87,$F4,$D1
	.dc.b	$5F,$ED,$9F,$2F,$FD,$AF,$1F,$ED,$26,$BF,$E2,$FF,$3F,$FD,$BA,$E7
	.dc.b	$9B,$FD,$A8,$FF,$B6,$3C,$DF,$F6,$D3,$B6,$BF,$E3,$A9,$FF,$6E,$3F
	.dc.b	$EE,$0D,$3F,$C9,$31,$FE,$8E,$CB,$FC,$BD,$3F,$E5,$1A,$FF,$73,$79
	.dc.b	$5F,$E6,$52,$4F,$5B,$FF,$22,$C7,$F9,$4E,$BF,$CC,$27,$F9,$4E,$FF
	.dc.b	$CB,$35,$FE,$82,$8F,$F4,$FE,$7B,$38,$57,$48,$FF,$40,$38,$9F,$F2
	.dc.b	$A9,$FE,$90,$75,$08,$79,$FF,$E5,$39,$E1,$3E,$FC,$7D,$35,$EC,$9E
	.dc.b	$38,$7C,$27,$C5,$1F,$E8,$78,$F8,$55,$FE,$67,$75,$1C,$49,$1D,$71
	.dc.b	$F8,$31,$E6,$C3,$FD,$19,$3E,$E8,$F8,$CF,$FA,$F6,$FD,$31,$E8,$8E
	.dc.b	$27,$33,$FC,$64,$F0,$8D,$CF,$B6,$30,$49,$F8,$EF,$91,$FF,$6C,$4F
	.dc.b	$C2,$7E,$34,$F1,$76,$77,$8D,$4F,$8E,$3C,$26,$8E,$6F,$77,$99,$E8
	.dc.b	$29,$C0,$D9,$D5,$F2,$1D,$63,$C9,$7F,$32,$7B,$4E,$4E,$D1,$98,$EB
	.dc.b	$3F,$63,$0F,$B7,$1D,$4E,$37,$ED,$7C,$24,$BE,$8F,$AE,$FA,$1D,$63
	.dc.b	$31,$D9,$F0,$D7,$68,$C1,$C5,$C4,$F1,$AF,$05,$72,$8E,$B3,$C6,$3A
	.dc.b	$99,$78,$99,$9E,$27,$A6,$78,$E1,$F3,$6B,$8C,$E6,$B8,$CF,$EE,$1F
	.dc.b	$F5,$2D,$FE,$36,$BE,$7D,$6E,$BE,$B5,$FE,$CA,$BF,$DA,$95,$F9,$FB
	.dc.b	$E3,$8F,$E8,$A7,$FD,$BE,$7E,$86,$3F,$DA,$71,$FE,$D1,$C3,$FD,$91
	.dc.b	$3F,$EF,$38,$FF,$74,$1C,$5F,$A1,$1E,$C7,$C3,$3E,$87,$F8,$2D,$FE
	.dc.b	$B9,$FF,$68,$57,$1B,$FF,$6C,$47,$57,$F8,$9C,$78,$63,$FD,$E0,$72
	.dc.b	$AF,$4C,$FE,$54,$F8,$F1,$E0,$C2,$3B,$7F,$DB,$75,$DF,$0F,$5C,$FF
	.dc.b	$B9,$0E,$B1,$FE,$F8,$34,$E2,$BE,$45,$72,$7C,$15,$EF,$9F,$D1,$61
	.dc.b	$F6,$9F,$01,$FE,$FF,$39,$1F,$38,$F8,$33,$FC,$25,$E8,$7D,$63,$4F
	.dc.b	$F2,$13,$EE,$BE,$CC,$7C,$AB,$0F,$A7,$3F,$CB,$CF,$97,$1F,$50,$FC
	.dc.b	$65,$FE,$15,$F3,$E3,$F9,$37,$27,$59,$F2,$3F,$D0,$F1,$CB,$FB,$F8
	.dc.b	$FC,$E4,$7C,$EB,$F9,$11,$C1,$FC,$26,$3F,$A7,$AE,$38,$FF,$02,$C3
	.dc.b	$C1,$1C,$F1,$F2,$4F,$B2,$FD,$73,$FC,$95,$FE,$80,$8F,$F5,$77,$9B
	.dc.b	$FE,$F1,$23,$FE,$EF,$6B,$FD,$BF,$C1,$FF,$75,$3F,$EB,$D6,$BC,$CE
	.dc.b	$67,$FB,$21,$FF,$78,$67,$FE,$EE,$F3,$BF,$DE,$1C,$1B,$9C,$BC,$7C
	.dc.b	$BF,$F7,$05,$23,$1E,$6F,$FB,$B3,$0F,$44,$E5,$F5,$79,$7F,$EF,$17
	.dc.b	$FD,$DA,$7F,$B4,$6F,$FD,$D4,$DA,$FF,$6D,$A8,$CB,$D8,$EB,$7E,$BA
	.dc.b	$FF,$7A,$C6,$0E,$FE,$6F,$8D,$FF,$7B,$DF,$FB,$E7,$0F,$24,$7F,$BD
	.dc.b	$67,$51,$8F,$92,$7C,$AF,$C7,$8C,$37,$1F,$95,$3A,$BF,$00,$F8,$D7
	.dc.b	$C5,$6E,$3E,$D4,$ED,$FF,$12,$35,$FF,$79,$35,$EB,$7F,$0A,$38,$61
	.dc.b	$E8,$BF,$C2,$D2,$31,$CB,$DE,$15,$C3,$1F,$C0,$7A,$0F,$43,$E6,$B6
	.dc.b	$BF,$C6,$C1,$C3,$FE,$39,$6B,$F8,$93,$58,$F2,$23,$CA,$3E,$13,$D3
	.dc.b	$CD,$7A,$0F,$43,$0D,$9F,$F2,$34,$EA,$72,$71,$E3,$7A,$14,$95,$C2
	.dc.b	$32,$FF,$93,$FC,$C7,$D1,$FF,$B1,$A3,$F5,$51,$8D,$7D,$ED,$8E,$13
	.dc.b	$B5,$0F,$47,$0D,$AE,$31,$83,$87,$D4,$3D,$8F,$D0,$71,$CF,$D8,$9F
	.dc.b	$25,$8C,$3F,$65,$7B,$7F,$31,$4A,$DC,$7E,$32,$75,$97,$F9,$AD,$AF
	.dc.b	$D3,$6B,$D3,$0F,$19,$DD,$FE,$15,$1C,$AF,$CB,$FB,$13,$BB,$54,$AD
	.dc.b	$61,$FC,$1F,$D2,$7A,$3C,$F1,$7A,$4F,$4B,$8D,$6A,$FD,$73,$97,$F4
	.dc.b	$B1,$9C,$7F,$31,$1F,$C1,$A7,$51,$83,$F9,$04,$EE,$FE,$5C,$E9,$FD
	.dc.b	$CC,$73,$9F,$CA,$35,$F4,$61,$F6,$E3,$11,$D2,$3C,$E7,$AF,$CD,$7A
	.dc.b	$58,$C7,$F7,$7E,$8B,$D1,$D7,$A5,$F1,$A3,$13,$F6,$2D,$39,$AF,$C5
	.dc.b	$56,$B1,$FF,$51,$35,$FD,$53,$5F,$E6,$8E,$F3,$FE,$2A,$DC,$3F,$95
	.dc.b	$E1,$F5,$2B,$FA,$24,$E5,$FF,$59,$1F,$C0,$FD,$0F,$E6,$1E,$4F,$F0
	.dc.b	$68,$FE,$95,$E8,$7F,$96,$F1,$57,$F3,$5F,$2F,$55,$FE,$B6,$3F,$CB
	.dc.b	$D3,$AC,$BF,$CC,$51,$C8,$CD,$71,$F1,$3F,$D7,$D8,$75,$8C,$4F,$92
	.dc.b	$FF,$DC,$98,$FF,$B3,$71,$FE,$33,$1D,$67,$FC,$D8,$75,$8C,$D7,$28
	.dc.b	$FF,$6C,$56,$31,$FF,$04,$C7,$F5,$BC,$3C,$11,$E6,$8F,$55,$ED,$EA
	.dc.b	$63,$1F,$E7,$98,$75,$8F,$04,$62,$B9,$5F,$FB,$BA,$7D,$55,$F3,$2B
	.dc.b	$DF,$1F,$0A,$B6,$D3,$BD,$71,$7D,$33,$FE,$F2,$7E,$56,$1F,$EE,$A7
	.dc.b	$27,$F6,$8B,$FF,$80,$0F,$3C,$3E,$AD,$FB,$A3,$A9,$FE,$F6,$BF,$F7
	.dc.b	$A1,$D5,$FB,$57,$D5,$EF,$19,$9F,$24,$60,$F0,$61,$F3,$1F,$86,$60
	.dc.b	$EF,$7F,$2C,$F1,$BE,$13,$E2,$C7,$17,$C3,$5D,$DD,$1E,$E3,$F2,$D1
	.dc.b	$BA,$F1,$C7,$D5,$3C,$15,$D8,$F0,$57,$BA,$3C,$97,$E2,$3D,$CF,$80
	.dc.b	$CC,$6D,$EE,$7F,$92,$5C,$63,$F9,$46,$E3,$F7,$67,$FA,$F6,$1F,$3F
	.dc.b	$0F,$15,$7E,$16,$7E,$8E,$1E,$0A,$D4,$FE,$12,$BC,$F5,$C8,$FB,$51
	.dc.b	$93,$94,$74,$8F,$7C,$65,$A7,$18,$E6,$FE,$7C,$E6,$66,$39,$9C,$5F
	.dc.b	$7C,$75,$6C,$67,$0F,$D8,$4E,$5B,$19,$C3,$F4,$95,$F4,$CE,$27,$73
	.dc.b	$CA,$78,$EB,$CA,$FE,$20,$D9,$99,$F2,$BC,$4C,$9B,$AF,$0B,$96,$BC
	.dc.b	$4E,$E6,$EB,$CB,$7F,$42,$38,$46,$1F,$8A,$71,$3B,$9E,$B9,$EA,$F3
	.dc.b	$C7,$F7,$38,$7A,$A3,$55,$CE,$31,$7D,$E7,$F3,$57,$F8,$97,$EF,$63
	.dc.b	$FC,$3C,$F9,$97,$C7,$1F,$83,$3C,$23,$F7,$B3,$B9,$FA,$18,$F8,$9D
	.dc.b	$D7,$94,$FD,$8B,$87,$F3,$C7,$E7,$27,$F9,$64,$67,$0F,$34,$FE,$F4
	.dc.b	$D4,$7E,$7A,$BF,$61,$1F,$B4,$35,$87,$F3,$3B,$F4,$CF,$8E,$7E,$89
	.dc.b	$E0,$9F,$80,$FA,$E3,$9B,$FE,$4A,$BF,$F3,$CB,$FD,$2C,$C1,$FE,$49
	.dc.b	$1E,$73,$D0,$7A,$1F,$E5,$28,$C3,$63,$33,$C6,$72,$DF,$43,$FC,$C4
	.dc.b	$DC,$79,$46,$23,$FC,$B1,$28,$C4,$7F,$AE,$73,$FF,$65,$7A,$1F,$EE
	.dc.b	$6A,$D5,$FF,$4E,$9F,$F6,$6C,$62,$BF,$DB,$FA,$7F,$AE,$67,$FD,$F2
	.dc.b	$72,$8F,$F7,$08,$78,$E1,$FE,$FA,$8F,$F6,$FE,$1F,$EB,$3A,$FF,$80
	.dc.b	$11,$FC,$54,$FF,$74,$5F,$78,$ED,$18,$AE,$4F,$FB,$EC,$E0,$7F,$BB
	.dc.b	$8F,$F6,$55,$FF,$BE,$9F,$F7,$A1,$A5,$7E,$17,$D1,$19,$8C,$3F,$EF
	.dc.b	$53,$D4,$FC,$59,$F0,$CF,$B3,$0F,$8D,$39,$6D,$7C,$29,$D3,$F2,$1F
	.dc.b	$A0,$F5,$73,$7F,$EE,$C8,$F4,$9F,$18,$FE,$FF,$87,$C3,$9E,$31,$F8
	.dc.b	$AC,$3E,$1C,$E6,$38,$1F,$EF,$B3,$C0,$74,$3D,$67,$97,$0F,$43,$A3
	.dc.b	$E6,$5F,$FB,$D1,$ED,$87,$8A,$FE,$0D,$70,$AF,$45,$70,$C3,$F0,$6F
	.dc.b	$B6,$75,$1F,$4E,$7E,$DC,$FD,$73,$81,$F7,$6F,$B1,$F2,$9F,$A8,$60
	.dc.b	$FB,$71,$E5,$C3,$E0,$47,$23,$A3,$8A,$F9,$A7,$84,$F9,$2F,$03,$57
	.dc.b	$E6,$9F,$93,$5F,$94,$3B,$61,$F3,$8F,$83,$1F,$24,$87,$98,$FA,$11
	.dc.b	$F1,$2B,$AC,$6A,$39,$E1,$EB,$75,$5D,$27,$94,$78,$70,$F9,$71,$F4
	.dc.b	$8F,$82,$E2,$FD,$55,$C6,$BA,$D6,$8F,$98,$53,$E2,$C6,$EB,$93,$ED
	.dc.b	$BF,$A1,$5C,$8F,$AF,$3F,$A8,$AF,$43,$F2,$AB,$84,$7E,$26,$FD,$EF
	.dc.b	$AA,$7F,$5B,$5C,$DE,$17,$CA,$7E,$7C,$68,$F1,$C6,$63,$F2,$E7,$33
	.dc.b	$F1,$87,$9A,$0F,$E7,$7C,$1F,$F5,$E3,$89,$6B,$A8,$4E,$9F,$F5,$81
	.dc.b	$8E,$AD,$57,$08,$EB,$EC,$6E,$31,$1B,$3B,$5F,$59,$7A,$3C,$A3,$0D
	.dc.b	$9F,$3E,$1C,$E7,$C8,$D8,$EF,$3B,$55,$D1,$EE,$72,$C3,$5E,$8B,$D0
	.dc.b	$A4,$9D,$8F,$45,$CB,$98,$F8,$30,$F4,$7C,$0D,$8E,$37,$D5,$B7,$AA
	.dc.b	$91,$82,$1D,$1C,$BF,$3A,$79,$3F,$43,$CF,$7A,$1E,$F6,$C7,$A1,$AF
	.dc.b	$D6,$F3,$DE,$8E,$EB,$E3,$47,$FA,$B3,$83,$EF,$36,$BF,$0B,$7C,$B8
	.dc.b	$7E,$CB,$6B,$7C,$5C,$8E,$95,$F2,$3D,$07,$A3,$F6,$7A,$BE,$53,$78
	.dc.b	$33,$5B,$8F,$B1,$4C,$F5,$F2,$52,$BA,$47,$DD,$F4,$5E,$93,$D2,$FF
	.dc.b	$03,$52,$73,$18,$6C,$7F,$82,$DA,$FE,$74,$E1,$D2,$F4,$B8,$36,$38
	.dc.b	$35,$E1,$3F,$A4,$9F,$0F,$A2,$F4,$BF,$45,$C7,$FA,$3E,$07,$B0,$F9
	.dc.b	$5C,$F2,$7E,$B7,$0E,$0F,$E5,$AB,$73,$AC,$FF,$6F,$4F,$53,$15,$D3
	.dc.b	$D9,$D5,$2B,$58,$73,$C5,$5A,$F4,$5E,$9F,$5D,$1E,$A3,$D5,$D4,$E6
	.dc.b	$31,$8F,$E9,$2F,$8B,$FA,$CB,$E3,$7C,$23,$15,$C3,$3F,$E0,$F7,$C1
	.dc.b	$B8,$78,$7D,$57,$87,$C6,$7F,$86,$FA,$6F,$4F,$CD,$EA,$3D,$5F,$7C
	.dc.b	$E7,$4F,$14,$E5,$FD,$CC,$66,$DB,$7D,$4A,$D4,$76,$C3,$AE,$1D,$1B
	.dc.b	$87,$26,$CE,$E3,$07,$C1,$F4,$9E,$AF,$18,$FE,$13,$E9,$BD,$4F,$9D
	.dc.b	$4F,$57,$EB,$CE,$6B,$94,$62,$B6,$D8,$C9,$D2,$31,$87,$B2,$3F,$9C
	.dc.b	$5F,$CD,$D3,$F9,$7E,$1D,$6B,$F9,$B9,$FE,$78,$F5,$7F,$D1,$7E,$37
	.dc.b	$FB,$BF,$D0,$FF,$6A,$4F,$F9,$F4,$FF,$3C,$4E,$9F,$F7,$29,$FE,$F8
	.dc.b	$3F,$C9,$27,$3C,$7F,$D0,$B3,$C8,$95,$D4,$93,$C8,$FF,$41,$1C,$8E
	.dc.b	$F3,$CA,$37,$1F,$E9,$23,$C1,$3F,$EF,$98,$CE,$1F,$E8,$C9,$F1,$1C
	.dc.b	$FD,$5F,$EB,$B5,$E2,$8F,$01,$A7,$93,$83,$C2,$71,$3E,$14,$72,$AE
	.dc.b	$F8,$7C,$39,$F2,$1E,$AA,$E6,$E0,$ED,$3F,$EA,$97,$A4,$78,$DD,$1E
	.dc.b	$07,$8D,$FA,$E3,$13,$C6,$73,$5E,$1A,$D5,$F9,$6F,$D4,$D7,$AB,$63
	.dc.b	$71,$EC,$25,$78,$E3,$2D,$AE,$C4,$C3,$EA,$57,$63,$8E,$1F,$12,$33
	.dc.b	$1E,$03,$26,$EB,$A4,$6A,$FB,$D7,$1A,$F8,$AD,$AF,$7D,$6E,$FC,$55
	.dc.b	$A3,$A5,$6C,$F8,$C4,$7F,$0E,$D7,$EB,$35,$FA,$A6,$88,$68,$A6,$8C
	.dc.b	$10,$C1,$0F,$F6,$C1,$FE,$D7,$C3,$FD,$B6,$47,$FD,$A5,$EA,$FF,$B5
	.dc.b	$1F,$F6,$6B,$5F,$F1,$77,$A7,$FE,$D6,$73,$DB,$FE,$D4,$CE,$47,$FB
	.dc.b	$5C,$87,$FB,$5B,$D4,$FF,$2A,$3F,$E5,$8D,$3F,$CA,$3E,$AF,$FA,$D2
	.dc.b	$3F,$CC,$51,$87,$FC,$8D,$87,$F9,$42,$3F,$C9,$D1,$FE,$9F,$CF,$FD
	.dc.b	$02,$FF,$93,$8F,$F2,$49,$FE,$4B,$39,$BF,$E5,$47,$FC,$CC,$FF,$97
	.dc.b	$DF,$F3,$DB,$FE,$89,$7A,$57,$FA,$1D,$FF,$2A,$CF,$81,$FF,$49,$4F
	.dc.b	$18,$FF,$3D,$47,$AE,$3B,$47,$A6,$3A,$9E,$53,$C6,$76,$3F,$D1,$11
	.dc.b	$FE,$64,$7C,$B3,$F0,$EB,$FD,$03,$87,$F9,$8E,$7B,$1C,$4F,$74,$69
	.dc.b	$FF,$3B,$61,$F3,$9F,$F4,$1C,$F1,$69,$A7,$17,$EB,$3F,$CE,$D1,$D8
	.dc.b	$F7,$DF,$8E,$BF,$D0,$78,$7F,$A7,$2F,$8C,$E5,$FF,$3E,$5F,$D4,$BF
	.dc.b	$92,$FB,$6B,$E2,$1F,$22,$3C,$6F,$37,$C8,$47,$04,$8F,$AD,$1E,$43
	.dc.b	$F0,$B7,$EF,$8F,$54,$7E,$0A,$7E,$C6,$1F,$89,$9F,$11,$F5,$89,$3F
	.dc.b	$2E,$77,$5C,$F0,$F6,$4F,$D4,$3B,$E1,$F0,$1A,$79,$EF,$CB,$87,$78
	.dc.b	$F9,$4D,$7D,$F1,$AB,$ED,$1C,$B0,$F7,$3F,$87,$AF,$B9,$18,$8F,$96
	.dc.b	$74,$9E,$2F,$E3,$8C,$9D,$8E,$4F,$D8,$AF,$05,$FA,$67,$84,$7D,$72
	.dc.b	$61,$FA,$16,$C7,$E5,$8F,$C2,$78,$BF,$E9,$B7,$FD,$33,$EA,$FF,$A6
	.dc.b	$FD,$5F,$F4,$D4,$7F,$AC,$27,$46,$E3,$38,$7F,$AB,$9B,$C1,$FE,$A6
	.dc.b	$EA,$FF,$52,$56,$F0,$E8,$62,$3F,$D3,$97,$FE,$B6,$F3,$BF,$D3,$58
	.dc.b	$FF,$A4,$CF,$F4,$AC,$7F,$AE,$BD,$2F,$F4,$9B,$BF,$4F,$FD,$2E,$E7
	.dc.b	$D4,$E7,$9F,$FB,$27,$0E,$73,$BC,$39,$CF,$FB,$37,$0E,$6E,$F0,$E6
	.dc.b	$E7,$0F,$F4,$6D,$7F,$A8,$7D,$3B,$87,$FB,$33,$CF,$FF,$4E,$CF,$FA
	.dc.b	$86,$7F,$D2,$8F,$FA,$EA,$3F,$D3,$8B,$D3,$FF,$4D,$3F,$E9,$E9,$DB
	.dc.b	$FE,$A7,$E0,$EB,$8F,$FB,$87,$D2,$FF,$4B,$CF,$F8,$BC,$73,$7F,$A7
	.dc.b	$1F,$F3,$C6,$DA,$CF,$FD,$BF,$56,$F9,$67,$FE,$9B,$BE,$58,$7F,$B3
	.dc.b	$7D,$2F,$F5,$0C,$FF,$A7,$7D,$2F,$F4,$E5,$7F,$A5,$DA,$FF,$A4,$1A
	.dc.b	$FF,$BD,$7D,$1F,$F4,$DB,$79,$3F,$D2,$CD,$F4,$7F,$D3,$64,$7F,$D4
	.dc.b	$07,$FA,$1F,$D1,$FF,$45,$F9,$DF,$E9,$5D,$3F,$DC,$3E,$7F,$FA,$52
	.dc.b	$BC,$F3,$97,$FE,$00,$C7,$FA,$39,$CF,$9F,$FE,$92,$8C,$70,$7F,$A8
	.dc.b	$6B,$FD,$23,$7F,$E8,$9F,$33,$FD,$19,$E7,$FF,$9F,$EB,$FD,$4F,$E7
	.dc.b	$FF,$A1,$D5,$32,$9E,$7F,$F9,$76,$31,$3F,$E9,$07,$A1,$FE,$90,$BE
	.dc.b	$8F,$1A,$57,$17,$FD,$07,$E7,$FF,$A4,$27,$8C,$7F,$9F,$4F,$BD,$6F
	.dc.b	$3B,$FD,$0E,$F9,$23,$C9,$E7,$FF,$A3,$63,$FD,$19,$E7,$FF,$A0,$67
	.dc.b	$8D,$70,$BF,$15,$ED,$F4,$C6,$67,$9F,$9F,$FE,$8E,$C3,$B4,$2F,$83
	.dc.b	$FE,$9A,$AE,$0E,$6B,$85,$F6,$6C,$7C,$9F,$2F,$FD,$05,$87,$FB,$53
	.dc.b	$CC,$FF,$49,$5F,$FA,$2C,$FF,$46,$D7,$E7,$45,$70,$8F,$F4,$6F,$9D
	.dc.b	$FE,$90,$BE,$4D,$8C,$9C,$EB,$E4,$DF,$67,$FD,$22,$7F,$97,$0F,$F1
	.dc.b	$69,$4F,$F1,$61,$4F,$F1,$67,$99,$FE,$59,$F3,$7F,$C8,$3E,$5F,$F9
	.dc.b	$12,$7F,$C8,$47,$F9,$3D,$FF,$26,$B8,$C7,$FC,$9A,$FF,$91,$88,$6F
	.dc.b	$C1,$FF,$24,$3E,$2B,$ED,$8F,$F9,$3C,$91,$CF,$0F,$0D,$7F,$92,$63
	.dc.b	$CC,$4A,$E2,$EE,$B8,$E1,$E7,$25,$71,$AD,$C6,$2B,$C4,$7F,$9A,$0F
	.dc.b	$31,$E8,$C3,$D6,$E0,$8E,$09,$87,$9A,$39,$CF,$C4,$3B,$3F,$3A,$BC
	.dc.b	$37,$E3,$8F,$05,$F2,$C3,$D3,$39,$BE,$4D,$3E,$4D,$73,$38,$E1,$E1
	.dc.b	$71,$3C,$1F,$1C,$F0,$7E,$E9,$C2,$BB,$BA,$3E,$05,$73,$77,$5C,$DE
	.dc.b	$B5,$CD,$FA,$47,$33,$E9,$12,$3E,$0B,$CA,$38,$61,$F0,$A3,$C0,$7C
	.dc.b	$CA,$F0,$3F,$E7,$C3,$F8,$51,$4F,$E1,$91,$FE,$7D,$7F,$CE,$E7,$F9
	.dc.b	$FB,$BF,$F9,$6A,$F3,$3F,$CF,$8F,$F9,$FB,$CD,$FF,$40,$3F,$C7,$61
	.dc.b	$E6,$B4,$FF,$3F,$CF,$F9,$F5,$AF,$F8,$B0,$FF,$1B,$1F,$E0,$EF,$37
	.dc.b	$FC,$18,$47,$FC,$18,$7F,$85,$CF,$F0,$CE,$7F,$E3,$A7,$FC,$CE,$72
	.dc.b	$D3,$FC,$91,$19,$C3,$FC,$4A,$FF,$87,$27,$FC,$64,$7F,$88,$2F,$FC
	.dc.b	$A7,$7D,$E3,$FC,$CF,$87,$F8,$8D,$D1,$FE,$3D,$3C,$C7,$92,$BB,$46
	.dc.b	$E3,$CB,$3F,$E4,$4C,$3F,$CB,$75,$FE,$68,$9F,$53,$FE,$28,$7B,$DF
	.dc.b	$89,$E9,$5C,$67,$E3,$61,$FE,$45,$7B,$D6,$EB,$89,$E5,$BF,$83,$7C
	.dc.b	$63,$FC,$B1,$5F,$E5,$C6,$9F,$E4,$5A,$FF,$1F,$57,$08,$FF,$27,$4F
	.dc.b	$C7,$8F,$F2,$63,$E7,$7F,$CC,$33,$C2,$3D,$87,$92,$7C,$4E,$2B,$C5
	.dc.b	$1F,$56,$30,$F6,$3E,$6D,$7C,$67,$FD,$31,$5F,$D2,$27,$FD,$64,$DA
	.dc.b	$FE,$C4,$FF,$4C,$BF,$C7,$BF,$DB,$6D,$5F,$D9,$CE,$3A,$DB,$E2,$DC
	.dc.b	$3F,$B8,$BF,$ED,$8C,$7F,$31,$0B,$E6,$FF,$79,$BE,$6D,$BE,$6D,$39
	.dc.b	$9F,$A3,$AE,$75,$CB,$E0,$5F,$36,$F8,$1E,$31,$8F,$0A,$C9,$B3,$30
	.dc.b	$EB,$5B,$6D,$6D,$B3,$E4,$D1,$FB,$6B,$B3,$0B,$DE,$A6,$22,$5F,$AA
	.dc.b	$B8,$CE,$A3,$15,$CD,$F7,$8F,$9F,$7B,$9F,$12,$30,$D8,$CC,$70,$E0
	.dc.b	$FE,$27,$87,$F8,$1A,$73,$3F,$DE,$5B,$19,$AE,$95,$C3,$87,$FC,$49
	.dc.b	$5A,$6C,$F7,$37,$7F,$5B,$E2,$1F,$8D,$3E,$2B,$E7,$F9,$72,$BE,$CD
	.dc.b	$6A,$78,$56,$E7,$E7,$D6,$DF,$F3,$55,$6D,$E3,$F1,$8F,$9F,$FB,$7F
	.dc.b	$49,$27,$31,$E3,$9D,$47,$DA,$9D,$3D,$23,$2F,$F9,$DF,$93,$F1,$3F
	.dc.b	$13,$F8,$62,$47,$E0,$BE,$AE,$BB,$E6,$4E,$38,$BF,$C8,$17,$F9,$2B
	.dc.b	$FC,$7F,$B9,$30,$FC,$5B,$F9,$2D,$F1,$1C,$E2,$DF,$2C,$7E,$24,$F7
	.dc.b	$F6,$3F,$32,$47,$F2,$BE,$65,$74,$7E,$62,$3F,$3F,$1F,$87,$EC,$61
	.dc.b	$49,$5F,$8E,$F8,$87,$81,$88,$FF,$1F,$57,$07,$F2,$6D,$7F,$C8,$53
	.dc.b	$C1,$FD,$3F,$8B,$FA,$E2,$3F,$AA,$FA,$6A,$FC,$98,$BA,$7B,$E7,$F7
	.dc.b	$FE,$1B,$F1,$D7,$C7,$61,$B8,$7F,$02,$9D,$4F,$E8,$FD,$DF,$F3,$1E
	.dc.b	$1F,$E7,$19,$CD,$6E,$F8,$BF,$EA,$C6,$CF,$F9,$AA,$B8,$F6,$7F,$A9
	.dc.b	$FA,$BF,$D5,$BE,$3F,$FB,$7B,$D7,$46,$2B,$5F,$0B,$FE,$91,$AD,$7C
	.dc.b	$57,$F8,$02,$73,$9A,$79,$5F,$18,$FF,$4C,$61,$D3,$0E,$57,$FE,$B7
	.dc.b	$C3,$95,$6A,$F8,$2B,$E0,$F5,$AE,$F8,$78,$A3,$99,$C7,$0A,$F1,$BE
	.dc.b	$77,$CA,$B8,$4F,$33,$81,$4E,$15,$CA,$78,$5F,$9E,$71,$7D,$09,$5B
	.dc.b	$C3,$D7,$3C,$8E,$97,$C9,$EF,$87,$88,$F8,$F5,$D0,$91,$C2,$B9,$D7
	.dc.b	$18,$E9,$5C,$1D,$1C,$1C,$13,$0F,$44,$78,$B0,$F3,$B7,$0F,$43,$C6
	.dc.b	$7E,$69,$1F,$B3,$87,$8A,$BF,$DD,$F1,$F2,$0F,$F7,$95,$7F,$8F,$B0
	.dc.b	$FF,$78,$C7,$E0,$3C,$7F,$F7,$8B,$7D,$9F,$F7,$83,$FE,$4A,$9F,$EA
	.dc.b	$D1,$FE,$F3,$CB,$F2,$54,$8F,$B7,$F3,$5F,$EE,$8B,$F9,$73,$96,$CF
	.dc.b	$D7,$8C,$53,$3F,$F7,$05,$F1,$8E,$B5,$B2,$5F,$E0,$67,$55,$FE,$E7
	.dc.b	$C3,$CB,$3D,$89,$7D,$5F,$F7,$A5,$70,$7D,$C6,$AB,$96,$1E,$47,$C7
	.dc.b	$EC,$FF,$BA,$6F,$C7,$5B,$38,$BD,$55,$F5,$7F,$49,$7C,$DD,$DF,$6C
	.dc.b	$7E,$3E,$1D,$63,$13,$D4,$E3,$3A,$D7,$E0,$E3,$E2,$25,$79,$A3,$2F
	.dc.b	$28,$CD,$7C,$0B,$EC,$F8,$88,$76,$3E,$63,$A7,$11,$C9,$B3,$DD,$F8
	.dc.b	$5C,$0F,$C6,$DF,$5A,$C7,$F9,$24,$61,$FF,$04,$FB,$6E,$FD,$61,$FE
	.dc.b	$89,$92,$C3,$A6,$1A,$E3,$FE,$55,$8F,$3A,$E0,$6D,$C4,$74,$86,$3D
	.dc.b	$F0,$F1,$DF,$37,$BF,$81,$FE,$36,$FA,$8C,$5A,$31,$3F,$E3,$0F,$A8
	.dc.b	$CC,$62,$7F,$B8,$56,$3C,$13,$6B,$0F,$4C,$9E,$A7,$C5,$C7,$FC,$60
	.dc.b	$DC,$3D,$63,$E3,$FF,$C5,$B3,$A3,$94,$EE,$BE,$23,$63,$E4,$D7,$1F
	.dc.b	$2F,$71,$E5,$D2,$3F,$E4,$AF,$8A,$B5,$FE,$09,$9B,$87,$F7,$F9,$B5
	.dc.b	$F1,$B4,$75,$FB,$CF,$81,$F1,$76,$7F,$CA,$B1,$71,$FF,$0A,$4D,$C3
	.dc.b	$FC,$15,$F0,$3F,$1C,$F9,$7F,$54,$ED,$C6,$3F,$E5,$E9,$B8,$7F,$88
	.dc.b	$EB,$53,$93,$FC,$63,$E5,$72,$F9,$0F,$B3,$E7,$BC,$DF,$F1,$1D,$7F
	.dc.b	$9A,$AD,$5A,$9C,$E1,$D6,$31,$87,$3B,$FF,$22,$FA,$1F,$71,$E5,$F2
	.dc.b	$9D,$A3,$17,$E5,$8C,$73,$7B,$1B,$5C,$0F,$F3,$4E,$DF,$E1,$C6,$F9
	.dc.b	$7F,$90,$BF,$F7,$57,$97,$6B,$51,$F8,$46,$9F,$E0,$C9,$D4,$ED,$95
	.dc.b	$D0,$EF,$18,$AF,$D3,$D3,$F6,$EF,$C4,$47,$FB,$E5,$FF,$2D,$57,$FA
	.dc.b	$0D,$CC,$EF,$D2,$FC,$D4,$60,$DD,$74,$71,$1F,$73,$C1,$FB,$BE,$55
	.dc.b	$C3,$E9,$36,$FE,$E3,$70,$FF,$2B,$F0,$FF,$8B,$AB,$E3,$FB,$9F,$0C
	.dc.b	$F8,$D7,$FE,$1D,$3F,$D7,$1E,$AF,$FB,$1A,$35,$3C,$0F,$F4,$BC,$EF
	.dc.b	$DA,$FF,$5E,$AC,$15,$C2,$3E,$73,$D9,$FF,$13,$1D,$8F,$E2,$B1,$C2
	.dc.b	$BA,$C6,$AB,$AC,$E6,$38,$1F,$B4,$F1,$BF,$DB,$18,$79,$EE,$CF,$EE
	.dc.b	$0E,$51,$97,$89,$E9,$71,$7E,$A8,$C1,$E4,$C3,$F9,$5C,$66,$39,$99
	.dc.b	$AE,$EF,$8C,$F2,$E1,$E7,$9E,$18,$6A,$31,$87,$96,$B9,$9F,$27,$0E
	.dc.b	$F1,$9A,$E1,$18,$9F,$64,$63,$0F,$0C,$7A,$0E,$4E,$1E,$B7,$E0,$78
	.dc.b	$9E,$6A,$E8,$FA,$70,$F9,$F1,$88,$E9,$3C,$5D,$61,$F2,$0F,$33,$CE
	.dc.b	$FC,$18,$79,$63,$E7,$9E,$2B,$E8,$D9,$DB,$C4,$DC,$F1,$C3,$E4,$C6
	.dc.b	$5A,$79,$A7,$8B,$63,$89,$BB,$F2,$C6,$4E,$2F,$A6,$3E,$1C,$66,$36
	.dc.b	$66,$3D,$E7,$DF,$3A,$C6,$DE,$31,$B7,$E3,$DF,$BA,$FD,$53,$D2,$79
	.dc.b	$57,$89,$B1,$97,$EC,$9D,$9E,$8F,$C7,$30,$70,$3C,$25,$3C,$E7,$28
	.dc.b	$EC,$D8,$ED,$87,$DB,$8F,$3E,$1F,$2A,$7B,$DF,$A2,$7A,$9C,$4B,$5C
	.dc.b	$5F,$25,$F6,$8C,$4F,$19,$F1,$C6,$1F,$A8,$72,$C3,$E3,$38,$36,$E2
	.dc.b	$BE,$74,$63,$0F,$33,$EF,$3D,$27,$C4,$3C,$C6,$CC,$57,$86,$79,$C7
	.dc.b	$07,$D1,$5D,$F0,$F8,$33,$E3,$34,$FC,$88,$D4,$78,$4F,$83,$19,$BF
	.dc.b	$2C,$78,$8E,$27,$BD,$CD,$F8,$E7,$BD,$F9,$2B,$99,$B3,$D3,$7E,$07
	.dc.b	$A1,$A3,$DD,$1E,$AB,$EB,$3B,$BE,$D5,$E7,$C3,$F8,$83,$FC,$5C,$AE
	.dc.b	$1F,$E8,$47,$F2,$9F,$71,$C1,$FE,$49,$8F,$E3,$11,$8F,$1B,$9B,$63
	.dc.b	$33,$B7,$97,$0F,$F4,$57,$8B,$5C,$E1,$CA,$31,$1F,$E3,$0A,$47,$1C
	.dc.b	$FF,$A7,$E8,$F7,$3F,$C6,$B1,$86,$C7,$19,$E0,$DC,$3F,$9B,$BC,$70
	.dc.b	$B5,$C1,$FF,$33,$E1,$E0,$AF,$E7,$AD,$8F,$F0,$FD,$7F,$91,$0E,$4E
	.dc.b	$0F,$F3,$7B,$8E,$7F,$EC,$B3,$C5,$B8,$79,$2B,$78,$78,$55,$F4,$C7
	.dc.b	$78,$F8,$AF,$85,$6A,$F8,$5E,$CF,$74,$7F,$66,$C7,$D1,$87,$C5,$8F
	.dc.b	$79,$CF,$08,$FF,$6C,$AE,$4F,$90,$CC,$F8,$9D,$5F,$80,$F4,$5F,$6A
	.dc.b	$F0,$D7,$7A,$D9,$CA,$3C,$B8,$E4,$8E,$0E,$27,$CD,$BE,$D3,$C4,$FA
	.dc.b	$B8,$FC,$29,$F3,$C7,$C3,$34,$30,$FA,$11,$CE,$3F,$D4,$77,$E0,$79
	.dc.b	$E3,$89,$F0,$C7,$C1,$3D,$27,$CB,$BF,$A3,$7E,$F8,$D6,$3E,$F3,$E3
	.dc.b	$1F,$1F,$0E,$D1,$E5,$3F,$0C,$79,$6B,$F4,$91,$C2,$FC,$D3,$E0,$AC
	.dc.b	$3E,$63,$11,$A8,$CC,$73,$C3,$D4,$FA,$6B,$C3,$87,$C2,$8E,$55,$E7
	.dc.b	$BE,$47,$C0,$8E,$47,$63,$DA,$E6,$79,$47,$AE,$BA,$E1,$EA,$AE,$47
	.dc.b	$D4,$3E,$AD,$7C,$7A,$EB,$7E,$17,$A9,$E4,$9F,$C1,$3E,$97,$E1,$DF
	.dc.b	$CB,$7A,$E3,$F8,$49,$FA,$98,$FE,$02,$7C,$13,$F9,$4C,$7E,$B9,$E5
	.dc.b	$AE,$84,$3D,$F7,$F2,$AB,$8C,$7E,$2B,$1F,$44,$62,$3B,$92,$7E,$EC
	.dc.b	$7E,$FC,$F4,$C7,$B9,$F9,$AF,$B8,$C1,$EF,$30,$FC,$6A,$F3,$E1,$F6
	.dc.b	$A7,$E7,$4F,$C2,$3A,$DF,$93,$0F,$85,$8F,$E0,$A3,$15,$FB,$E3,$CF
	.dc.b	$1E,$D9,$DD,$7D,$BC,$3F,$55,$18,$AF,$21,$E6,$3F,$31,$5F,$5D,$C5
	.dc.b	$F6,$7F,$0C,$62,$B9,$BE,$19,$F2,$46,$23,$73,$F8,$33,$94,$F5,$71
	.dc.b	$3B,$8C,$3F,$12,$7B,$DF,$C7,$C3,$F0,$18,$7A,$67,$D1,$3F,$CD,$25
	.dc.b	$B7,$EB,$1B,$3F,$BD,$6B,$C2,$BF,$B1,$57,$F3,$C8,$FF,$45,$35,$FD
	.dc.b	$A9,$88,$3F,$E9,$AA,$FE,$89,$E5,$62,$3F,$D4,$91,$87,$47,$85,$E3
	.dc.b	$E3,$3C,$07,$B9,$C7,$2F,$3C,$E7,$CC,$C5,$F0,$B4,$63,$05,$73,$C3
	.dc.b	$0D,$8F,$0B,$CD,$B5,$C9,$B3,$CF,$07,$B4,$F7,$30,$BD,$CA,$62,$25
	.dc.b	$FB,$AB,$8C,$EA,$31,$5C,$DF,$86,$3D,$C7,$93,$EB,$8C,$36,$33,$1C
	.dc.b	$38,$3F,$D7,$D8,$7D,$A9,$CC,$FF,$AB,$5A,$FF,$B0,$DA,$FD,$56,$BF
	.dc.b	$69,$E3,$1C,$FE,$3B,$C0,$78,$6F,$E7,$78,$8F,$09,$EE,$7C,$ED,$DE
	.dc.b	$AD,$37,$3A,$CF,$F0,$37,$C6,$F5,$1F,$4A,$B7,$1D,$BC,$67,$85,$D7
	.dc.b	$DD,$97,$CA,$72,$D9,$CB,$F9,$03,$F0,$79,$2B,$A4,$F9,$FC,$7F,$F2
	.dc.b	$1F,$2F,$E4,$5B,$46,$92,$FF,$13,$E1,$3E,$5B,$F1,$AD,$8F,$C5,$CF
	.dc.b	$C3,$F7,$24,$7E,$39,$F8,$DA,$62,$3C,$D8,$FE,$86,$71,$C5,$FA,$0C
	.dc.b	$BE,$45,$69,$7B,$DE,$62,$47,$FB,$AE,$FF,$55,$18,$8E,$B9,$3C,$5F
	.dc.b	$39,$F5,$E7,$F1,$DD,$EC,$BF,$CC,$4D,$BF,$F3,$3B,$5F,$F7,$C3,$5F
	.dc.b	$DF,$1F,$E5,$68,$FC,$17,$B7,$23,$F3,$B3,$C7,$3F,$67,$B9,$F8,$72
	.dc.b	$4F,$EC,$AA,$BE,$F8,$B3,$F0,$E7,$F7,$DE,$33,$DC,$79,$3F,$E7,$C8
	.dc.b	$C3,$63,$38,$79,$84,$E9,$C3,$5F,$8F,$F3,$1F,$70,$EF,$3C,$A3,$F1
	.dc.b	$83,$DD,$FC,$FF,$A3,$FC,$8B,$CC,$79,$BF,$C8,$67,$34,$9C,$BF,$CA
	.dc.b	$23,$31,$88,$E9,$E9,$FF,$29,$8C,$4E,$AB,$A7,$63,$0F,$04,$61,$B8
	.dc.b	$75,$C3,$94,$E6,$3C,$4D,$C3,$AD,$F3,$9C,$9C,$4E,$67,$BD,$D4,$70
	.dc.b	$BC,$E1,$E5,$AE,$07,$23,$84,$6A,$B8,$DF,$84,$F8,$D8,$79,$AB,$E2
	.dc.b	$4E,$9B,$5C,$4E,$F3,$B8,$EA,$77,$BF,$13,$8C,$3E,$4C,$EE,$72,$4B
	.dc.b	$ED,$5F,$40,$EC,$73,$73,$1C,$27,$94,$60,$97,$E4,$AE,$37,$F1,$5E
	.dc.b	$51,$96,$92,$34,$79,$9F,$09,$FA,$CB,$FC,$4B,$6B,$EB,$36,$BF,$DB
	.dc.b	$73,$FD,$8B,$C9,$BE,$41,$A4,$AF,$D1,$4E,$DF,$91,$3B,$7E,$C7,$BA
	.dc.b	$F7,$7F,$C0,$F9,$7E,$4A,$B4,$DC,$3F,$03,$18,$D3,$FC,$01,$7C,$7C
	.dc.b	$BC,$D4,$C7,$BB,$F8,$C9,$D4,$60,$98,$75,$7C,$78,$F8,$B0,$FD,$66
	.dc.b	$3D,$E7,$F5,$4D,$7F,$68,$DB,$EB,$7C,$C9,$5C,$B0,$F1,$61,$D7,$0E
	.dc.b	$71,$E7,$AD,$C7,$F8,$7A,$F9,$DF,$18,$F3,$B6,$7F,$75,$3A,$9E,$66
	.dc.b	$8B,$1A,$BF,$04,$7A,$2F,$C1,$7C,$CF,$97,$3A,$CB,$EC,$5F,$92,$F8
	.dc.b	$1F,$1B,$2E,$2F,$6B,$E4,$FA,$6F,$96,$5F,$27,$81,$EE,$3D,$8F,$F2
	.dc.b	$2F,$2F,$F8,$C3,$AD,$D1,$8D,$3F,$C7,$F8,$70,$BC,$67,$FE,$30,$BD
	.dc.b	$D7,$F8,$FA,$B4,$FF,$2E,$C7,$B3,$6F,$FC,$A3,$7D,$27,$2F,$F4,$CE
	.dc.b	$CF,$F2,$27,$AB,$D9,$B7,$FE,$38,$F6,$F1,$DE,$AA,$49,$D1,$FE,$3E
	.dc.b	$7C,$FA,$BD,$AF,$F3,$2C,$EB,$1F,$F0,$44,$E6,$3A,$E1,$C2,$B7,$5F
	.dc.b	$E3,$06,$BF,$E5,$76,$BF,$E6,$4A,$E5,$5F,$DD,$70,$F6,$C7,$CD,$F5
	.dc.b	$7D,$3A,$7B,$76,$46,$2F,$84,$FC,$6C,$3F,$C3,$6D,$D3,$FC,$D0,$DC
	.dc.b	$5D,$3F,$10,$F8,$BE,$CD,$C3,$FC,$BF,$17,$0F,$F0,$64,$DC,$FF,$C2
	.dc.b	$1E,$CF,$8F,$C9,$F9,$AD,$C3,$FC,$CE,$F7,$23,$ED,$7C,$DE,$52,$7F
	.dc.b	$C5,$53,$FE,$28,$F5,$FF,$0F,$39,$F2,$3E,$04,$62,$B7,$97,$F8,$62
	.dc.b	$31,$3A,$BF,$F2,$4F,$99,$F7,$DF,$F1,$9F,$B3,$D2,$31,$7F,$E2,$08
	.dc.b	$C7,$2F,$CD,$6D,$70,$3F,$CE,$3A,$7D,$06,$E5,$F9,$CB,$4E,$5F,$F0
	.dc.b	$0B,$ED,$9E,$3C,$5F,$56,$76,$D9,$E1,$1F,$E7,$68,$C1,$D6,$3A,$1F
	.dc.b	$EB,$23,$F5,$31,$DA,$3F,$2F,$EC,$FE,$42,$7F,$6A,$FF,$99,$2B,$FD
	.dc.b	$1C,$E5,$FC,$83,$9A,$FC,$DB,$97,$D6,$E6,$79,$CF,$6F,$37,$F2,$5D
	.dc.b	$BF,$93,$F1,$AC,$FF,$98,$1B,$87,$F9,$31,$B7,$FC,$1D,$F6,$79,$7E
	.dc.b	$DF,$1E,$DF,$CE,$6D,$FF,$8C,$EF,$FC,$A9,$3F,$E3,$58,$F9,$F5,$F2
	.dc.b	$49,$1F,$59,$FA,$91,$FE,$2F,$1E,$C3,$D3,$FF,$64,$4E,$A7,$FC,$60
	.dc.b	$6A,$33,$5F,$ED,$8A,$E1,$3F,$E5,$AF,$0F,$15,$A8,$E0,$64,$A7,$9A
	.dc.b	$B6,$E6,$B9,$B9,$36,$79,$DC,$E1,$F0,$23,$27,$9A,$33,$87,$FB,$D9
	.dc.b	$E7,$1C,$6F,$C9,$18,$8E,$11,$83,$FC,$BD,$87,$92,$3C,$53,$B8,$C5
	.dc.b	$F8,$63,$13,$EC,$8C,$57,$28,$C5,$F8,$E3,$15,$E6,$C3,$B5,$79,$4F
	.dc.b	$E4,$15,$B7,$E2,$57,$95,$A6,$A7,$8B,$EA,$30,$7C,$4C,$3C,$F5,$CE
	.dc.b	$3F,$CC,$87,$92,$78,$CE,$9B,$19,$39,$16,$FD,$E7,$A5,$CB,$4F,$87
	.dc.b	$C9,$92,$E1,$F8,$37,$8B,$EB,$9F,$1C,$F1,$8F,$8A,$66,$3E,$21,$F4
	.dc.b	$9F,$99,$3D,$5A,$75,$38,$C7,$C2,$9E,$ED,$8C,$BC,$CF,$96,$53,$E2
	.dc.b	$CF,$57,$F3,$0D,$7C,$F1,$E5,$7E,$81,$C4,$F3,$E3,$F8,$27,$AB,$5E
	.dc.b	$B1,$E7,$7A,$B5,$EA,$79,$CE,$38,$FD,$97,$AB,$5F,$0C,$F5,$7F,$79
	.dc.b	$5E,$76,$D7,$57,$8E,$1F,$7B,$0F,$A1,$1D,$A3,$13,$ED,$C5,$2B,$E8
	.dc.b	$FF,$04,$6C,$7E,$26,$3F,$07,$3C,$5F,$C4,$3F,$43,$0F,$34,$71,$8F
	.dc.b	$23,$4F,$CC,$D7,$D8,$7E,$61,$EE,$C7,$E0,$5F,$7C,$75,$87,$48,$F3
	.dc.b	$DF,$92,$FD,$18,$76,$9E,$0F,$8F,$0E,$73,$E6,$3E,$11,$FA,$0A,$E4
	.dc.b	$73,$C3,$E3,$C7,$D6,$3C,$B7,$D1,$E3,$1E,$4A,$EA,$F0,$8F,$A2,$7C
	.dc.b	$63,$EF,$1C,$F0,$F5,$3E,$33,$13,$CB,$0F,$99,$87,$91,$F6,$4E,$CE
	.dc.b	$B7,$E1,$C7,$ED,$C7,$63,$FA,$01,$FD,$1F,$1F,$E9,$0E,$0F,$E3,$07
	.dc.b	$E0,$0F,$C0,$FB,$3F,$E2,$CF,$07,$FA,$73,$63,$46,$B8,$53,$A6,$C6
	.dc.b	$84,$FF,$99,$34,$E7,$E1,$7F,$81,$A7,$53,$96,$C6,$5E,$F1,$9C,$BF
	.dc.b	$A3,$E6,$F6,$7D,$11,$86,$C7,$AE,$78,$36,$FC,$15,$BB,$FE,$81,$3C
	.dc.b	$DF,$E8,$55,$C1,$F8,$75,$C2,$7E,$31,$B8,$FE,$BA,$DA,$FE,$B7,$5C
	.dc.b	$67,$E0,$D7,$17,$F5,$B5,$B3,$9D,$FA,$23,$BE,$1E,$2A,$E1,$1F,$DD
	.dc.b	$E7,$4D,$35,$31,$F1,$C6,$AB,$B4,$74,$3F,$B9,$DF,$96,$33,$1C,$CC
	.dc.b	$DF,$9F,$BD,$F3,$2E,$7F,$C3,$64,$DF,$F8,$3D,$6A,$73,$18,$AE,$18
	.dc.b	$7F,$9F,$63,$1C,$9A,$85,$F1,$74,$62,$B8,$BF,$E4,$CA,$E9,$7C,$6B
	.dc.b	$CB,$77,$C1,$7C,$D5,$26,$1D,$29,$2F,$93,$CC,$F5,$3F,$D6,$95,$8C
	.dc.b	$FF,$D7,$93,$D7,$E7,$25,$73,$B4,$78,$7E,$23,$33,$CF,$1E,$F3,$8C
	.dc.b	$3F,$D8,$D5,$74,$F9,$38,$3E,$7B,$D4,$67,$1F,$F6,$37,$CF,$7B,$48
	.dc.b	$F8,$7E,$5B,$E0,$7D,$04,$8F,$23,$A7,$C5,$6F,$77,$E3,$12,$36,$7A
	.dc.b	$FD,$DF,$94,$47,$E4,$7D,$0F,$F0,$0D,$7F,$01,$1F,$62,$1E,$9F,$D3
	.dc.b	$6C,$70,$8C,$74,$E7,$D4,$93,$D1,$E1,$F3,$7C,$7E,$8B,$97,$5C,$AF
	.dc.b	$0C,$75,$F9,$59,$1E,$47,$C1,$F3,$BE,$1F,$96,$93,$E8,$7C,$9F,$3B
	.dc.b	$E5,$F6,$7C,$DF,$11,$2B,$DB,$1E,$9F,$88,$91,$F0,$1F,$67,$CE,$FB
	.dc.b	$7E,$92,$4F,$C4,$7E,$07,$D3,$7C,$1F,$A8,$93,$F2,$1F,$89,$F3,$DF
	.dc.b	$17,$E5,$A4,$FC,$C7,$E4,$7D,$4F,$C9,$FA,$29,$3F,$41,$FC,$F7,$C7
	.dc.b	$7F,$A7,$FC,$2F,$D7,$AF,$57,$24,$AF,$CF,$43,$E4,$78,$FC,$2C,$9C
	.dc.b	$BD,$3E,$23,$AD,$49,$E2,$F8,$3D,$DF,$DB,$43,$F4,$16,$98,$6C,$7E
	.dc.b	$E0,$D4,$7E,$F7,$C1,$70,$BE,$43,$84,$CA,$E1,$7C,$9F,$D9,$5F,$0A
	.dc.b	$D4,$7E,$D4,$EA,$70,$AE,$D7,$D6,$3E,$0F,$CB,$7C,$82,$3F,$13,$E3
	.dc.b	$EB,$FB,$5C,$D3,$8C,$7F,$3B,$E7,$25,$7F,$4B,$E4,$66,$1F,$39,$FA
	.dc.b	$BF,$23,$88,$E9,$17,$0E,$76,$AF,$89,$E4,$FE,$AF,$93,$FC,$5F,$5F
	.dc.b	$B5,$C3,$26,$BC,$7A,$FF,$24,$8D,$36,$B9,$3F,$EF,$07,$F6,$2F,$CA
	.dc.b	$73,$0F,$86,$72,$7D,$56,$D7,$33,$F1,$FF,$31,$1F,$C5,$BF,$92,$F9
	.dc.b	$2C,$47,$C9,$8B,$7E,$E9,$FC,$BF,$C9,$4A,$FC,$9B,$F9,$AF,$86,$C4
	.dc.b	$7D,$18,$B5,$F1,$A7,$F1,$FF,$26,$F8,$3A,$4B,$FC,$74,$3D,$DC,$47
	.dc.b	$E1,$63,$83,$A9,$E7,$3A,$BF,$E3,$32,$AD,$BC,$6B,$6F,$EA,$BE,$3F
	.dc.b	$F6,$24,$7F,$5B,$F1,$D5,$FC,$54,$5B,$FB,$F3,$FA,$1F,$DC,$3F,$A9
	.dc.b	$D7,$D9,$6B,$F9,$E6,$CF,$F2,$C0,$FF,$4C,$8E,$1E,$EF,$32,$3C,$FD
	.dc.b	$CF,$8F,$EF,$48,$F8,$AF,$87,$2C,$47,$C5,$8B,$87,$7C,$3E,$0B,$F1
	.dc.b	$33,$4E,$B9,$7F,$C8,$91,$8B,$F5,$61,$FC,$53,$E5,$A4,$7F,$10,$7F
	.dc.b	$8B,$FC,$7E,$23,$F5,$D1,$71,$FD,$1C,$FF,$1C,$E4,$95,$FC,$61,$FE
	.dc.b	$3F,$F1,$F8,$8F,$DB,$45,$AF,$D5,$CF,$F1,$4E,$1F,$DC,$D7,$F1,$16
	.dc.b	$BF,$C1,$DB,$3F,$BC,$FA,$D7,$CA,$F1,$F0,$25,$F1,$7F,$C1,$93,$A8
	.dc.b	$FF,$04,$3F,$EA,$36,$FC,$5E,$8D,$FC,$2D,$8C,$90,$FF,$01,$4F,$F8
	.dc.b	$6B,$0F,$ED,$3F,$14,$FC,$8F,$F9,$04,$61,$BC,$5F,$EB,$29,$D7,$8D
	.dc.b	$E4,$E4,$FF,$14,$46,$31,$FF,$13,$B5,$FF,$0D,$37,$0F,$81,$3E,$CF
	.dc.b	$8E,$95,$EC,$7D,$FE,$46,$23,$CF,$16,$BE,$44,$FC,$4E,$59,$5F,$11
	.dc.b	$F8,$FF,$1D,$88,$F7,$C5,$AF,$A1,$3F,$33,$8A,$57,$CC,$7E,$7F,$C6
	.dc.b	$E2,$3E,$3C,$5A,$FB,$13,$FD,$4B,$E5,$9E,$77,$D5,$F8,$FF,$CA,$60
	.dc.b	$9C,$F3,$7F,$B6,$A7,$31,$89,$46,$3A,$7F,$D4,$0E,$AF,$A6,$FF,$EA
	.dc.b	$3B,$E1,$39,$C3,$A4,$E6,$F8,$BB,$BE,$38,$7F,$BA,$B0,$E3,$39,$BE
	.dc.b	$33,$9B,$E2,$EE,$F8,$BE,$B6,$DF,$28,$E0,$72,$35,$56,$35,$18,$3D
	.dc.b	$F7,$E7,$75,$5D,$09,$5C,$6F,$C7,$3C,$70,$F8,$0F,$23,$CD,$3B,$79
	.dc.b	$B1,$D1,$C6,$FC,$0E,$30,$F2,$93,$0F,$6E,$1E,$08,$C5,$F3,$C3,$EB
	.dc.b	$5D,$C3,$E5,$3C,$F0,$E2,$E4,$F2,$99,$AF,$33,$4F,$F6,$A5,$FE,$0D
	.dc.b	$FF,$6D,$B5,$EA,$63,$C5,$FE,$D6,$FE,$0D,$AF,$DD,$6B,$F9,$16,$C7
	.dc.b	$F6,$EF,$1F,$EA,$47,$F9,$93,$1F,$CB,$FC,$13,$DA,$FF,$35,$C7,$E5
	.dc.b	$A7,$59,$F7,$BE,$27,$DF,$BE,$1C,$D8,$CB,$E6,$5F,$07,$C9,$5B,$C7
	.dc.b	$AE,$1C,$E7,$73,$C7,$1F,$1D,$6A,$78,$9A,$24,$6B,$0E,$17,$CE,$3D
	.dc.b	$17,$DE,$73,$5C,$A7,$2F,$E1,$31,$F4,$BD,$2F,$91,$10,$8D,$CF,$22
	.dc.b	$4E,$E7,$58,$7A,$5F,$A5,$87,$89,$A7,$86,$35,$97,$D5,$8F,$11,$E7
	.dc.b	$AF,$0B,$D3,$1E,$53,$BC,$3C,$33,$E6,$C7,$CB,$7E,$9F,$39,$D2,$F9
	.dc.b	$1F,$EF,$99,$7F,$B9,$A3,$0F,$F4,$CF,$13,$38,$FF,$B8,$A7,$2F,$F4
	.dc.b	$8A,$CD,$F4,$6C,$FF,$B7,$AF,$FA,$AC,$63,$E3,$64,$7F,$58,$F1,$6E
	.dc.b	$1F,$D4,$A6,$F2,$FF,$63,$9B,$5E,$5F,$8D,$EA,$76,$F8,$DB,$1F,$D8
	.dc.b	$A2,$E1,$FD,$72,$6D,$FF,$5C,$FC,$71,$F1,$5E,$CE,$F4,$63,$1F,$ED
	.dc.b	$7F,$1D,$98,$C6,$3F,$D6,$1B,$87,$3C,$3D,$F5,$B7,$E0,$FC,$74,$C3
	.dc.b	$FB,$AF,$C7,$5C,$7F,$B7,$CD,$C3,$FC,$01,$36,$BE,$97,$A7,$F2,$8F
	.dc.b	$97,$85,$8F,$EF,$F1,$70,$FE,$F3,$36,$FF,$D0,$BF,$57,$F9,$6B,$FF
	.dc.b	$3C,$9F,$98,$E3,$46,$6D,$5C,$3A,$BE,$DC,$E6,$39,$CE,$0C,$CD,$BE
	.dc.b	$B0,$BE,$BF,$1B,$C8,$E7,$D1,$6B,$73,$A6,$E2,$AD,$C6,$1F,$C8,$DB
	.dc.b	$E5,$5F,$19,$86,$D7,$FA,$47,$F7,$17,$85,$9A,$93,$FE,$C1,$AD,$DF
	.dc.b	$FA,$EA,$F7,$39,$AD,$CF,$FA,$EE,$B7,$AF,$F9,$86,$F7,$39,$AF,$D2
	.dc.b	$CA,$5E,$C6,$1B,$87,$97,$D8,$C4,$C8,$F4,$3B,$8C,$61,$F5,$27,$EC
	.dc.b	$68,$F8,$FF,$AA,$7D,$6C,$AC,$FC,$A8,$B5,$C4,$FA,$95,$EA,$9E,$73
	.dc.b	$F2,$75,$91,$F1,$6A,$9F,$06,$2E,$7F,$E9,$4D,$3F,$09,$5F,$C4,$9F
	.dc.b	$B0,$FC,$D9,$DB,$9A,$A7,$8B,$E3,$FF,$1A,$7E,$3B,$E3,$EC,$FE,$0A
	.dc.b	$2C,$FA,$78,$3F,$2F,$58,$C7,$F3,$BF,$1F,$F9,$63,$F2,$FF,$1F,$71
	.dc.b	$FC,$5C,$5C,$39,$6A,$AC,$5F,$DB,$F8,$5F,$B3,$EB,$FF,$B8,$71,$FC
	.dc.b	$54,$63,$DF,$FB,$4C,$8F,$C7,$79,$57,$2F,$C5,$4D,$C3,$F0,$D3,$6B
	.dc.b	$EC,$FC,$7F,$EB,$0F,$D6,$F9,$76,$3F,$43,$17,$0F,$3C,$DB,$FD,$DF
	.dc.b	$C7,$FE,$D0,$FD,$AF,$C7,$DC,$3F,$53,$16,$FD,$F3,$6B,$FD,$97,$FA
	.dc.b	$13,$DB,$FD,$5B,$70,$FF,$7C,$7D,$5B,$E2,$33,$52,$7F,$7B,$8A,$F8
	.dc.b	$3F,$8D,$C1,$87,$1A,$DD,$6A,$7C,$37,$CE,$F8,$5E,$DF,$CB,$1D,$63
	.dc.b	$F8,$8F,$C8,$CC,$3F,$82,$FC,$8D,$AF,$B7,$97,$F1,$63,$F8,$BF,$C8
	.dc.b	$DA,$FE,$17,$F2,$3F,$C6,$8F,$E3,$7F,$23,$63,$F8,$7F,$BF,$FC,$78
	.dc.b	$FE,$3F,$F2,$36,$3F,$8A,$FC,$31,$FE,$BA,$D7,$3F,$23,$87,$3C,$F1
	.dc.b	$E3,$CF,$75,$FC,$D5,$FF,$A1,$27,$AF,$B9,$D3,$E3,$7A,$BD,$BE,$46
	.dc.b	$4F,$8B,$8E,$E1,$CB,$E4,$3C,$7F,$21,$E4,$6D,$7A,$3B,$7C,$E7,$A3
	.dc.b	$CA,$B1,$EC,$F2,$BD,$67,$B3,$E4,$2C,$7C,$0F,$90,$F7,$FC,$87,$C0
	.dc.b	$6C,$7C,$4F,$90,$F8,$67,$C4,$F8,$FB,$1F,$23,$D1,$F8,$E7,$C8,$F1
	.dc.b	$AC,$7C,$CF,$8F,$F9,$67,$F8,$23,$CD,$FF,$6F,$C6,$4F,$E8,$AB,$E3
	.dc.b	$F4,$6F,$E3,$EC,$E9,$C3,$C3,$E1,$6F,$27,$F0,$9B,$47,$73,$C1,$F1
	.dc.b	$D6,$BB,$FC,$77,$8F,$E3,$BC,$8D,$8C,$93,$0E,$9F,$1D,$A8,$C7,$95
	.dc.b	$F8,$79,$FE,$C5,$92,$FD,$D7,$B9,$F2,$7C,$77,$C1,$3E,$17,$C7,$5C
	.dc.b	$7E,$0C,$59,$F5,$4D,$C7,$B7,$AC,$CB,$FC,$67,$18,$6E,$1F,$DF,$DF
	.dc.b	$F1,$95,$FC,$3F,$89,$FA,$04,$9F,$9F,$F1,$76,$7F,$66,$7F,$B5,$47
	.dc.b	$CB,$77,$7F,$CC,$38,$36,$F8,$13,$99,$FF,$6B,$9C,$31,$FF,$2B,$D6
	.dc.b	$F0,$FF,$34,$D6,$EB,$B3,$C6,$75,$1C,$4F,$F2,$F9,$FE,$E4,$30,$7F
	.dc.b	$86,$2F,$C3,$7A,$8D,$57,$39,$CC,$72,$F8,$DF,$0D,$79,$27,$33,$E8
	.dc.b	$AE,$18,$6A,$38,$1E,$0C,$3C,$B3,$E9,$8E,$F7,$CE,$B8,$47,$38,$C3
	.dc.b	$E8,$3D,$EE,$27,$96,$1C,$2F,$AC,$FA,$E3,$18,$79,$AF,$89,$F4,$9C
	.dc.b	$3F,$14,$F8,$45,$39,$18,$30,$7F,$A6,$BE,$59,$AB,$E2,$78,$C6,$1B
	.dc.b	$19,$9F,$9D,$39,$6E,$FF,$E4,$A6,$E5,$CA,$31,$1F,$6A,$51,$88,$FF
	.dc.b	$4D,$C6,$27,$E4,$E5,$FE,$43,$9B,$EE,$FF,$A8,$34,$FC,$75,$EE,$7F
	.dc.b	$06,$D8,$FC,$95,$71,$9F,$C3,$B6,$38,$E1,$E7,$6E,$1F,$EB,$CE,$6E
	.dc.b	$38,$FF,$AB,$9B,$86,$70,$F3,$E1,$EF,$9E,$33,$DA,$BC,$D5,$FE,$B6
	.dc.b	$3C,$07,$0B,$FC,$2E,$3E,$78,$D4,$F0,$AF,$F6,$6C,$70,$7E,$B6,$1E
	.dc.b	$78,$E5,$1D,$CF,$8A,$74,$8D,$3C,$2F,$2E,$CF,$8F,$1A,$E1,$7C,$B3
	.dc.b	$9F,$F6,$FC,$3F,$E8,$78,$C5,$FF,$9E,$1A,$FF,$8D,$5F,$E5,$AE,$67
	.dc.b	$F2,$35,$D1,$B3,$C2,$7F,$DB,$33,$97,$F7,$0B,$E5,$9C,$7C,$9B,$3A
	.dc.b	$6B,$FC,$12,$78,$4E,$CF,$4C,$7F,$32,$F8,$FB,$3F,$C9,$DB,$5C,$B3
	.dc.b	$FE,$56,$BE,$3F,$24,$BF,$F4,$BC,$3E,$3F,$11,$E1,$8E,$0F,$96,$7D
	.dc.b	$04,$78,$FC,$7F,$80,$E3,$7E,$0F,$8E,$F1,$FC,$6C,$AE,$F1,$5E,$D1
	.dc.b	$F0,$C8,$FC,$3F,$8D,$F5,$91,$F3,$C7,$BB,$E3,$7E,$0F,$CB,$7E,$8F
	.dc.b	$0F,$E9,$2D,$E2,$FF,$62,$69,$FE,$E4,$C3,$FD,$6B,$2F,$8E,$FB,$6D
	.dc.b	$8F,$AE,$D7,$E9,$35,$FB,$AB,$E3,$BC,$24,$8E,$BF,$19,$E4,$F8,$DF
	.dc.b	$2F,$44,$9F,$0C,$7A,$3E,$33,$C2,$48,$F0,$BE,$4F,$8A,$F6,$92,$3D
	.dc.b	$3F,$15,$F0,$3E,$2B,$E0,$FC,$54,$9F,$6C,$7C,$4F,$8A,$F6,$92,$3D
	.dc.b	$AF,$C0,$F8,$7F,$92,$48,$F8,$BF,$0F,$F3,$3E,$1F,$E2,$92,$3E,$2B
	.dc.b	$F2,$3E,$1A,$47,$C8,$6B,$F2,$3E,$17,$E6,$91,$F9,$AF,$EA,$BE,$05
	.dc.b	$F0,$B4,$93,$FA,$88,$CB,$7E,$13,$81,$1C,$1B,$8E,$84,$70,$7F,$48
	.dc.b	$F9,$67,$C3,$FE,$AA,$11,$87,$FD,$06,$D7,$FD,$05,$9F,$F8,$7D,$CC
	.dc.b	$FF,$4F,$9D,$47,$F7,$A6,$A2,$38,$6C,$7E,$EB,$AD,$DD,$F9,$79,$46
	.dc.b	$3C,$9F,$35,$F8,$E3,$FC,$4B,$8F,$FA,$7E,$BF,$C2,$FD,$BD,$FC,$54
	.dc.b	$E6,$7F,$D9,$CD,$AE,$3F,$03,$FC,$22,$31,$7F,$E2,$BF,$13,$F8,$5F
	.dc.b	$5F,$F0,$0C,$38,$FC,$1E,$FE,$0E,$47,$80,$67,$E0,$8E,$9F,$07,$A9
	.dc.b	$FF,$2B,$79,$1D,$0C,$6B,$FE,$4B,$3E,$0C,$7C,$3E,$0F,$F1,$1C,$65
	.dc.b	$BA,$7F,$8D,$1F,$DA,$46,$67,$FC,$06,$F2,$7F,$C4,$CD,$79,$C7,$F8
	.dc.b	$16,$30,$78,$7D,$B7,$27,$D3,$6D,$7B,$67,$77,$F3,$77,$7C,$1F,$C9
	.dc.b	$25,$7C,$3D,$7E,$8F,$C1,$E2,$3F,$CE,$DE,$5F,$CC,$31,$8F,$DA,$F8
	.dc.b	$3F,$A4,$49,$E5,$8B,$E0,$B1,$1F,$E7,$FF,$37,$EA,$98,$BF,$F3,$A4
	.dc.b	$3E,$1B,$FC,$B0,$FF,$B7,$1A,$FF,$B4,$9B,$1C,$1F,$F0,$BC,$6F,$D9
	.dc.b	$FF,$15,$47,$08,$FF,$2A,$B5,$FC,$E1,$DE,$36,$7E,$6F,$F5,$A7,$C5
	.dc.b	$53,$F5,$D5,$E0,$E0,$F9,$5F,$0D,$F9,$12,$61,$F0,$3E,$1A,$BF,$E9
	.dc.b	$DF,$91,$FC,$99,$8C,$FF,$11,$F2,$FF,$D7,$70,$FF,$3B,$B6,$BF,$52
	.dc.b	$D9,$3F,$9B,$78,$FA,$FF,$B1,$87,$AF,$4C,$47,$A3,$E1,$65,$FC,$B7
	.dc.b	$86,$F8,$8F,$F3,$A6,$BF,$E0,$66,$CF,$CA,$23,$F8,$BC,$3F,$85,$3F
	.dc.b	$BE,$CB,$D3,$F0,$9C,$F3,$FD,$EF,$C4,$7E,$D4,$95,$F8,$4F,$88,$AF
	.dc.b	$FB,$07,$DB,$FD,$B9,$8A,$FE,$0B,$D9,$FB,$D2,$4F,$E3,$3E,$22,$BF
	.dc.b	$EC,$9E,$FF,$DF,$98,$AF,$C5,$FB,$FF,$C2,$67,$FD,$2A,$D7,$F0,$2D
	.dc.b	$7F,$8F,$35,$FF,$30,$F8,$9F,$C2,$BD,$D9,$1F,$C1,$CC,$9F,$A7,$F9
	.dc.b	$4F,$F7,$13,$7E,$0B,$FC,$37,$18,$8F,$8C,$7F,$0E,$7F,$C6,$B8,$7F
	.dc.b	$B1,$9D,$47,$F8,$C1,$AF,$F2,$36,$BC,$67,$FD,$13,$F0,$AF,$97,$FE
	.dc.b	$68,$DF,$3F,$31,$82,$3F,$E2,$EF,$C4,$7F,$89,$DF,$F2,$06,$D7,$18
	.dc.b	$F6,$7C,$5F,$F9,$1B,$13,$49,$1F,$E4,$48,$CB,$F0,$FE,$2F,$E0,$9F
	.dc.b	$0A,$7E,$1F,$54,$9F,$90,$E9,$A4,$3E,$3F,$90,$BF,$D8,$47,$C9,$BF
	.dc.b	$9B,$C1,$E5,$24,$78,$DA,$F2,$CB,$F9,$7F,$47,$F4,$D6,$C7,$D3,$9F
	.dc.b	$ED,$5F,$2E,$F8,$E7,$CC,$7F,$61,$F8,$DF,$D2,$4E,$63,$1D,$DF,$C2
	.dc.b	$A3,$18,$2B,$7D,$7F,$C3,$9E,$38,$74,$9D,$5F,$18,$C4,$EB,$0E,$B5
	.dc.b	$FD,$C7,$06,$1C,$F0,$E1,$5A,$A5,$78,$AF,$94,$76,$39,$5F,$17,$A6
	.dc.b	$1E,$0C,$39,$C6,$CE,$71,$97,$89,$DE,$FB,$4F,$0B,$EE,$7C,$09,$D5
	.dc.b	$71,$79,$E1,$83,$96,$1E,$38,$C9,$E1,$C3,$AB,$E6,$9D,$BC,$4D,$57
	.dc.b	$59,$DC,$7C,$2A,$E9,$18,$24,$EE,$32,$F1,$25,$76,$AE,$AF,$7B,$F6
	.dc.b	$E1,$F7,$EF,$8C,$E6,$3E,$11,$9A,$F3,$1F,$D3,$8F,$69,$70,$FC,$3B
	.dc.b	$70,$FC,$23,$9F,$3B,$FC,$EC,$FE,$05,$AF,$DE,$6C,$FF,$80,$3E,$1F
	.dc.b	$B4,$ED,$FA,$D3,$FE,$83,$CB,$F9,$E7,$C3,$3E,$01,$BE,$AE,$61,$F4
	.dc.b	$70,$E1,$9F,$F6,$6C,$38,$5E,$B1,$FD,$1D,$6A,$FC,$07,$E2,$6B,$6F
	.dc.b	$3C,$38,$61,$FA,$7A,$E1,$3F,$E9,$1C,$7C,$6E,$C9,$87,$6B,$FC,$F6
	.dc.b	$1D,$23,$13,$AC,$3A,$56,$AF,$9D,$68,$93,$AA,$F3,$4E,$A3,$13,$A9
	.dc.b	$F5,$BF,$22,$75,$18,$9D,$91,$FF,$33,$5F,$96,$31,$1F,$5E,$FE,$0B
	.dc.b	$AA,$E6,$E4,$F5,$1D,$4D,$63,$C1,$D6,$1E,$57,$E4,$47,$4E,$2F,$E4
	.dc.b	$DF,$1D,$FE,$53,$8F,$E3,$DE,$D5,$AF,$EA,$13,$69,$C4,$AC,$72,$2B
	.dc.b	$FC,$67,$7F,$D4,$5A,$F2,$C3,$FA,$04,$63,$0F,$E6,$DE,$3F,$23,$9F
	.dc.b	$B1,$6B,$FB,$34,$5B,$FE,$7B,$87,$FB,$EE,$72,$7F,$6B,$F5,$78,$B4
	.dc.b	$FE,$DD,$7E,$68,$F0,$4F,$FB,$FE,$7F,$A9,$46,$0F,$3F,$53,$E6,$1E
	.dc.b	$07,$C1,$DB,$FD,$C1,$7F,$E8,$1A,$D4,$E7,$8F,$E4,$46,$27,$F5,$AD
	.dc.b	$CF,$F7,$8D,$7F,$C9,$6D,$7F,$BC,$79,$7E,$62,$4F,$F6,$76,$C7,$C2
	.dc.b	$7E,$8F,$8D,$F4,$9F,$A7,$E2,$79,$FE,$22,$DF,$F9,$D7,$34,$63,$0F
	.dc.b	$E1,$DC,$3F,$D9,$EF,$E2,$92,$FF,$82,$C3,$E5,$7F,$D7,$91,$CF,$E2
	.dc.b	$7F,$09,$18,$BF,$ED,$11,$86,$4F,$FA,$16,$5D,$18,$9F,$EE,$F1,$F8
	.dc.b	$B9,$F3,$7B,$9D,$39,$FE,$83,$63,$94,$78,$38,$31,$7F,$E2,$61,$F2
	.dc.b	$EC,$BF,$35,$E4,$3E,$61,$E6,$7E,$76,$F7,$5A,$9C,$E5,$FE,$37,$C3
	.dc.b	$93,$FC,$9A,$B6,$D8,$CC,$7F,$2A,$6B,$FC,$E5,$CE,$1F,$E9,$66,$CF
	.dc.b	$FB,$96,$3C,$9E,$1F,$8E,$73,$5F,$EE,$2B,$FF,$57,$DF,$FA,$CD,$7B
	.dc.b	$3E,$7F,$73,$D0,$D9,$D4,$7A,$FD,$CC,$5F,$CE,$E4,$FC,$D7,$C5,$FD
	.dc.b	$86,$DF,$D5,$70,$FD,$7D,$6D,$7C,$DC,$FE,$B3,$FE,$54,$D5,$F0,$4F
	.dc.b	$2F,$D5,$C5,$F4,$23,$18,$FC,$B8,$FE,$07,$7F,$59,$B8,$74,$8F,$91
	.dc.b	$18,$39,$CF,$D3,$8A,$7E,$23,$8F,$F6,$9F,$31,$F9,$26,$CF,$E3,$DC
	.dc.b	$3F,$91,$D6,$D7,$E4,$2B,$F7,$9F,$19,$F9,$A6,$D7,$E5,$DC,$3F,$99
	.dc.b	$8B,$3F,$07,$89,$F2,$BF,$8E,$E5,$7C,$17,$E2,$E3,$0D,$C3,$FD,$EB
	.dc.b	$5C,$1B,$3F,$C9,$B8,$3F,$2C,$D9,$D6,$1D,$E3,$F2,$DE,$57,$38,$FD
	.dc.b	$4F,$C6,$FF,$15,$3F,$66,$5C,$3F,$60,$E1,$FD,$97,$95,$67,$F6,$3F
	.dc.b	$1B,$FC,$76,$3F,$76,$D9,$FD,$C3,$87,$F7,$5F,$1B,$6B,$F6,$1F,$30
	.dc.b	$F6,$3F,$55,$F1,$5F,$AD,$8C,$5F,$E6,$63,$13,$F8,$29,$CD,$7F,$84
	.dc.b	$9A,$FF,$88,$5A,$FF,$BC,$1D,$ED,$FF,$00,$1B,$7F,$AB,$F9,$8F,$C9
	.dc.b	$BF,$EA,$F8,$CD,$3C,$3F,$E3,$5F,$07,$FD,$47,$C3,$FF,$7B,$DF,$F1
	.dc.b	$E6,$D7,$F2,$46,$BF,$F0,$02,$BF,$91,$75,$F1,$25,$71,$7A,$7C,$C4
	.dc.b	$8E,$8D,$AE,$9E,$D7,$84,$93,$E1,$7F,$B2,$7C,$0F,$F8,$BA,$3F,$98
	.dc.b	$B6,$7F,$9B,$B5,$F2,$35,$1E,$F7,$F1,$2E,$B9,$7A,$92,$7A,$FA,$FF
	.dc.b	$DF,$6B,$FD,$35,$1F,$C5,$4E,$79,$21,$F1,$1B,$AE,$1F,$31,$E8,$25
	.dc.b	$FA,$1F,$5F,$C6,$75,$F8,$CE,$CD,$9E,$D9,$F9,$FB,$3C,$47,$8F,$C8
	.dc.b	$B3,$AD,$3C,$5E,$47,$C6,$24,$7C,$67,$E4,$FC,$5F,$AF,$E2,$7D,$8D
	.dc.b	$9E,$5B,$BE,$27,$E0,$1F,$07,$E1,$AC,$EB,$CE,$FA,$64,$7E,$9B,$F5
	.dc.b	$BE,$1B,$E4,$78,$BF,$25,$B3,$E7,$F2,$5F,$0D,$FE,$29,$6C,$67,$E5
	.dc.b	$71,$39,$7F,$C4,$92,$69,$27,$FB,$9C,$3E,$07,$84,$77,$6F,$1F,$FB
	.dc.b	$8A,$5F,$11,$E1,$F4,$3C,$4D,$BF,$33,$E0,$23,$D9,$B5,$CD,$F5,$7C
	.dc.b	$37,$29,$F7,$7C,$37,$F8,$97,$C9,$F7,$36,$32,$7A,$1F,$69,$1F,$5F
	.dc.b	$C1,$FC,$02,$38,$26,$1E,$B8,$F9,$5F,$15,$F2,$DF,$EF,$51,$FE,$EF
	.dc.b	$1F,$2E,$D3,$F4,$70,$F2,$7F,$A9,$BA,$8F,$F5,$0F,$C3,$7E,$A2,$B8
	.dc.b	$4F,$F7,$5D,$BF,$D1,$B1,$88,$FF,$7A,$1C,$4D,$65,$C6,$F9,$C6,$4F
	.dc.b	$05,$72,$8C,$D7,$3C,$3C,$6F,$5A,$F5,$39,$AE,$11,$FE,$C3,$C3,$B3
	.dc.b	$E0,$3B,$1C,$2F,$DF,$7D,$1F,$1F,$0F,$3C,$7F,$82,$BD,$E9,$78,$AE
	.dc.b	$38,$7F,$AB,$2B,$53,$E0,$AF,$5D,$D7,$C0,$78,$F1,$DB,$F0,$71,$F1
	.dc.b	$3E,$93,$18,$7B,$1F,$96,$7F,$8B,$6F,$FC,$5C,$FD,$B3,$07,$FB,$5B
	.dc.b	$E5,$DF,$30,$F8,$2F,$F1,$A4,$61,$B1,$99,$E0,$D7,$FD,$D9,$2D,$FF
	.dc.b	$09,$1D,$5A,$E9,$AF,$E2,$A3,$86,$3E,$09,$E3,$97,$E1,$75,$FF,$52
	.dc.b	$5F,$C7,$6B,$F8,$3B,$E2,$DA,$E3,$3F,$96,$37,$5C,$FD,$8F,$CC,$BF
	.dc.b	$91,$BE,$2D,$AF,$4E,$DB,$C3,$ED,$1F,$D0,$B0,$F0,$C6,$F1,$CC,$B0
	.dc.b	$E7,$5E,$1C,$7B,$B6,$7F,$01,$3A,$7F,$A6,$B5,$FD,$74,$F0,$7C,$0D
	.dc.b	$7B,$61,$EE,$78,$B5,$E2,$7C,$78,$E1,$3D,$4F,$8F,$87,$B6,$7F,$95
	.dc.b	$DF,$95,$A7,$4C,$3E,$0F,$7B,$E5,$DC,$5F,$A4,$E1,$FE,$B7,$18,$BF
	.dc.b	$E9,$8D,$7F,$83,$46,$BE,$27,$84,$7F,$82,$67,$4F,$F8,$0D,$AF,$4A
	.dc.b	$FF,$43,$5F,$85,$BF,$30,$F8,$DA,$4A,$FF,$56,$C6,$5B,$1E,$0A,$4F
	.dc.b	$F3,$A9,$FE,$C4,$D7,$CE,$7F,$87,$5E,$4F,$FA,$8F,$C0,$B9,$B3,$FE
	.dc.b	$47,$18,$6E,$1E,$EF,$75,$F8,$9B,$E2,$BE,$0E,$3F,$E0,$1A,$FD,$AE
	.dc.b	$1D,$A3,$F9,$69,$93,$FD,$4C,$E5,$EA,$7F,$7E,$B7,$97,$E1,$8F,$F5
	.dc.b	$BB,$5E,$47,$B3,$47,$C3,$3D,$9F,$88,$D9,$FF,$6E,$46,$3D,$5C,$BF
	.dc.b	$4A,$75,$3F,$ED,$E6,$BF,$E3,$B6,$BF,$E1,$E5,$F3,$1C,$49,$1F,$EE
	.dc.b	$28,$CB,$7D,$5E,$C4,$70,$6F,$E6,$3C,$44,$70,$73,$F5,$BC,$C4,$70
	.dc.b	$77,$F9,$8F,$51,$1C,$1E,$3F,$5F,$DC,$47,$07,$9F,$E6,$3E,$11,$1C
	.dc.b	$1F,$EF,$CE,$CF,$F5,$9E,$DF,$C0,$F9,$7F,$33,$18,$27,$BB,$82,$57
	.dc.b	$17,$7E,$CC,$8E,$8D,$7F,$0B,$7F,$C3,$FE,$03,$A9,$23,$C4,$F8,$3D
	.dc.b	$99,$1E,$56,$BC,$B0,$FF,$28,$BE,$9F,$6B,$CE,$47,$D9,$CB,$EA,$23
	.dc.b	$83,$E0,$7B,$DE,$E2,$38,$3E,$27,$B5,$F0,$88,$E0,$FC,$F9,$FE,$6E
	.dc.b	$F9,$87,$0F,$E9,$9B,$1F,$E5,$C0,$E1,$FF,$74,$7E,$EA,$F8,$2E,$15
	.dc.b	$FD,$82,$B5,$5F,$AC,$6B,$FE,$1A,$6B,$E1,$8F,$E2,$BE,$07,$FB,$CB
	.dc.b	$BA,$F4,$7E,$AE,$FC,$13,$FC,$06,$FF,$D4,$5B,$3D,$EF,$E1,$CD,$8F
	.dc.b	$F3,$73,$5F,$D9,$3F,$ED,$C7,$31,$C0,$8F,$0F,$7B,$9F,$BD,$D3,$DE
	.dc.b	$93,$D2,$32,$F6,$8C,$BD,$A3,$F8,$7F,$EE,$AF,$86,$A4,$9F,$F4,$3E
	.dc.b	$4F,$73,$C9,$1F,$C8,$85,$71,$7E,$B9,$8A,$FF,$06,$C6,$63,$ED,$B5
	.dc.b	$FE,$36,$E6,$BE,$41,$DF,$2F,$9F,$7F,$16,$3E,$65,$75,$3E,$4F,$C5
	.dc.b	$7C,$C2,$3F,$33,$E0,$2B,$F3,$22,$CF,$D5,$EA,$FE,$E9,$5F,$53,$E0
	.dc.b	$7E,$AF,$C0,$CC,$3E,$AC,$65,$FA,$F1,$97,$EB,$C7,$F3,$8F,$82,$7C
	.dc.b	$0D,$24,$FE,$57,$F1,$57,$C5,$E2,$3F,$3B,$5C,$1F,$92,$D7,$35,$FE
	.dc.b	$36,$8C,$67,$F0,$AF,$FC,$B9,$E1,$FF,$9C,$1B,$7F,$ED,$4A,$DC,$FF
	.dc.b	$B3,$39,$5F,$09,$F9,$2E,$09,$7F,$92,$8F,$CA,$98,$7F,$27,$18,$7F
	.dc.b	$07,$D8,$F8,$DF,$D1,$B6,$33,$18,$F6,$32,$C8,$C9,$CA,$3E,$47,$0C
	.dc.b	$8F,$DD,$66,$8F,$EC,$7A,$7E,$17,$4F,$F4,$6C,$E6,$BE,$3D,$7C,$52
	.dc.b	$38,$3E,$87,$77,$23,$B3,$FD,$EB,$67,$EF,$2F,$82,$FD,$EB,$67,$F4
	.dc.b	$2D,$78,$3F,$E5,$B7,$31,$C0,$8F,$0F,$6F,$F7,$FC,$52,$3F,$7E,$FF
	.dc.b	$02,$F8,$6C,$47,$F0,$08,$B8,$7E,$57,$8D,$5F,$E0,$4F,$AF,$FE,$35
	.dc.b	$7F,$AE,$6B,$FC,$99,$AF,$F8,$19,$AF,$FA,$D1,$DD,$7F,$15,$F8,$07
	.dc.b	$D8,$63,$81,$7F,$E7,$F8,$C3,$8B,$BD,$4E,$27,$CC,$FF,$8A,$5B,$5C
	.dc.b	$B2,$FD,$74,$62,$3F,$C1,$CF,$FB,$D9,$D6,$3E,$0B,$E9,$7F,$E3,$CA
	.dc.b	$E1,$5F,$EC,$C9,$CB,$FD,$11,$AF,$F6,$08,$FF,$7D,$39,$6B,$9F,$89
	.dc.b	$E8,$47,$C2,$27,$D9,$F1,$3E,$B3,$D9,$3E,$DF,$1F,$C8,$4A,$EC,$D7
	.dc.b	$DB,$F1,$3F,$0C,$E0,$F6,$BF,$91,$F0,$FF,$1C,$91,$9C,$7F,$C9,$DA
	.dc.b	$7C,$89,$F8,$44,$9F,$63,$F1,$79,$24,$7B,$DA,$FC,$37,$EA,$72,$48
	.dc.b	$ED,$1F,$E4,$EF,$99,$7A,$1F,$63,$E2,$1B,$FF,$BE,$A7,$3C,$1F,$84
	.dc.b	$9C,$E1,$F8,$69,$CB,$FE,$DC,$AE,$1D,$FF,$8A,$8C,$1B,$C3,$B4,$F5
	.dc.b	$6C,$67,$0E,$D3,$CA,$29,$93,$98,$BF,$0E,$1E,$48,$C5,$F4,$75,$87
	.dc.b	$8E,$75,$5C,$2F,$C5,$87,$6B,$E2,$D9,$FE,$C1,$17,$0E,$91,$C6,$3C
	.dc.b	$97,$D3,$0E,$D8,$74,$9F,$3D,$6D,$EA,$62,$76,$F2,$9E,$15,$E6,$BE
	.dc.b	$F5,$AA,$E3,$39,$C3,$C8,$79,$E7,$58,$77,$25,$F2,$BD,$9E,$47,$D6
	.dc.b	$65,$C5,$6A,$FB,$47,$9C,$C6,$1E,$62,$3F,$83,$C3,$C3,$7F,$B8,$8F
	.dc.b	$B0,$7F,$A6,$30,$FC,$3B,$63,$4F,$DB,$F4,$BF,$A5,$3B,$6B,$F9,$36
	.dc.b	$BF,$91,$6B,$C7,$BF,$EC,$47,$F5,$3C,$7F,$09,$F2,$4F,$8E,$FF,$1A
	.dc.b	$47,$E0,$A7,$59,$F5,$BE,$27,$E0,$EF,$87,$36,$31,$FB,$77,$C1,$F1
	.dc.b	$56,$F1,$EB,$87,$39,$DC,$F1,$C7,$C7,$5A,$9E,$26,$89,$1A,$C3,$85
	.dc.b	$F3,$8F,$45,$F7,$9C,$D7,$29,$CB,$FA,$6C,$7D,$2F,$4B,$E4,$47,$F9
	.dc.b	$FC,$6A,$78,$92,$75,$29,$D4,$63,$36,$5F,$0E,$BE,$91,$E1,$CB,$DC
	.dc.b	$E3,$0E,$8F,$9E,$FF,$83,$61,$D2,$30,$43,$84,$6A,$B9,$E1,$E0,$CB
	.dc.b	$F8,$2F,$1F,$F5,$3F,$8A,$FF,$24,$E9,$FD,$2F,$91,$F1,$78,$CF,$F9
	.dc.b	$F4,$2B,$17,$FD,$6E,$AC,$E6,$FF,$A5,$57,$F9,$42,$FA,$35,$E6,$DB
	.dc.b	$FE,$C9,$F0,$D6,$CD,$25,$FF,$55,$7B,$9E,$0F,$88,$B1,$FD,$0F,$E6
	.dc.b	$70,$D3,$FA,$34,$79,$A3,$B1,$23,$FB,$03,$E1,$F8,$6F,$5F,$CF,$7B
	.dc.b	$1B,$3F,$D3,$DD,$3F,$D9,$DB,$9F,$F6,$BF,$9E,$F5,$12,$7F,$B6,$36
	.dc.b	$3F,$BD,$BF,$23,$E5,$FE,$4B,$F3,$E1,$F5,$1F,$E6,$4A,$FE,$DF,$E1
	.dc.b	$BE,$C3,$F2,$BD,$FF,$E7,$CA,$D5,$A3,$FC,$2D,$3A,$BF,$E9,$95,$AA
	.dc.b	$FE,$9C,$76,$F1,$FE,$A3,$67,$FD,$27,$3B,$8F,$C2,$7C,$D7,$DE,$3E
	.dc.b	$FF,$CD,$5A,$FE,$E5,$16,$7F,$C2,$F3,$71,$FF,$4A,$FC,$DF,$FA,$9E
	.dc.b	$FF,$D4,$AD,$AF,$F2,$62,$F9,$9F,$C8,$52,$34,$F2,$F6,$73,$18,$BF
	.dc.b	$F1,$43,$C5,$EB,$1F,$2F,$E3,$9F,$19,$D3,$E8,$1F,$1D,$98,$C3,$6F
	.dc.b	$FD,$67,$5C,$1B,$3F,$1A,$31,$9F,$FB,$B9,$B9,$7F,$AB,$9B,$5F,$E3
	.dc.b	$BC,$1F,$2F,$D6,$FF,$CD,$5F,$18,$F3,$BD,$ED,$BF,$F6,$4B,$FE,$0C
	.dc.b	$6C,$68,$8E,$BC,$3B,$5F,$E4,$6D,$BF,$6B,$E6,$7F,$95,$2B,$FC,$4D
	.dc.b	$18,$9F,$DB,$B6,$32,$7F,$8B,$E7,$84,$FF,$B5,$F3,$7A,$9F,$53,$E6
	.dc.b	$5F,$0D,$F3,$63,$0D,$C3,$E4,$C7,$C7,$23,$83,$FC,$93,$6B,$EB,$1F
	.dc.b	$EE,$8B,$7C,$77,$3A,$FC,$E4,$FF,$97,$F1,$FF,$55,$3E,$7F,$6B,$DF
	.dc.b	$E3,$FC,$06,$C6,$4F,$64,$65,$F7,$79,$5F,$18,$F8,$FE,$ED,$8C,$93
	.dc.b	$0F,$09,$FE,$5F,$C3,$FD,$A7,$19,$E4,$FF,$76,$F8,$7F,$E9,$08,$FF
	.dc.b	$80,$3F,$0E,$F4,$FF,$79,$39,$C3,$FD,$55,$39,$AF,$87,$39,$9F,$DA
	.dc.b	$B5,$FF,$22,$35,$FF,$47,$B5,$3D,$6F,$C7,$EC,$7E,$AA,$FF,$D6,$D7
	.dc.b	$DA,$3F,$0B,$F3,$9F,$86,$F9,$7B,$3F,$81,$6E,$3F,$03,$CC,$68,$9C
	.dc.b	$E4,$BF,$A9,$1F,$90,$F9,$C9,$1F,$8B,$EF,$B8,$F3,$C1,$58,$C3,$94
	.dc.b	$FC,$DA,$C5,$FF,$98,$FE,$6D,$F1,$9F,$91,$F9,$3F,$E4,$D1,$89,$FC
	.dc.b	$44,$62,$7F,$8D,$CE,$6F,$F8,$F8,$70,$FF,$22,$6B,$DE,$39,$57,$FA
	.dc.b	$7B,$07,$EE,$EF,$86,$CD,$FA,$FC,$5F,$EB,$5A,$FF,$11,$8F,$E5,$BB
	.dc.b	$FC,$89,$C9,$AE,$C9,$97,$FA,$CE,$FF,$43,$7C,$1A,$FF,$AE,$5F,$F0
	.dc.b	$9B,$7D,$6E,$E4,$9E,$5E,$1F,$88,$8E,$0E,$DE,$37,$98,$8E,$0F,$17
	.dc.b	$CB,$7A,$88,$E0,$F3,$78,$FE,$E2,$38,$3D,$5D,$1F,$08,$8E,$0F,$77
	.dc.b	$1F,$C6,$23,$83,$E1,$77,$FC,$A2,$38,$3E,$37,$85,$F3,$88,$E0,$F9
	.dc.b	$5E,$37,$D2,$23,$83,$E7,$7C,$AF,$D6,$23,$83,$FD,$C9,$F3,$1F,$9B
	.dc.b	$69,$FE,$15,$71,$BF,$DC,$6B,$FE,$ED,$5F,$2D,$B2,$4F,$47,$97,$CB
	.dc.b	$48,$EE,$DA,$E9,$EE,$F7,$24,$F9,$1B,$1F,$B3,$BF,$F2,$4C,$FF,$B9
	.dc.b	$5F,$F6,$9B,$67,$FB,$07,$87,$FE,$39,$F2,$1F,$2B,$FC,$0B,$0F,$E7
	.dc.b	$93,$98,$C7,$67,$F5,$98,$FE,$BC,$4C,$3F,$B1,$B6,$B8,$C1,$FE,$4B
	.dc.b	$6F,$C6,$1E,$4F,$8A,$BC,$DF,$2F,$FE,$48,$D3,$C7,$18,$3D,$59,$E6
	.dc.b	$7E,$81,$23,$EA,$BF,$E4,$7F,$13,$F9,$46,$5F,$AE,$6B,$FE,$E9,$EE
	.dc.b	$FE,$2D,$E4,$6C,$97,$E7,$DB,$F8,$7C,$EA,$30,$7A,$DF,$E3,$DE,$EF
	.dc.b	$F9,$7B,$0E,$F5,$FE,$59,$F8,$0F,$81,$F3,$9F,$C8,$F2,$E9,$E0,$7F
	.dc.b	$1A,$8C,$47,$FB,$A4,$38,$AF,$E5,$79,$72,$C3,$F8,$AE,$1F,$2F,$E7
	.dc.b	$75,$8F,$F2,$47,$9E,$3F,$3E,$75,$5F,$46,$3F,$C8,$15,$B8,$E7,$0A
	.dc.b	$E7,$7C,$23,$FC,$83,$75,$E3,$15,$EC,$71,$AE,$D8,$7A,$23,$07,$22
	.dc.b	$9C,$8A,$72,$C7,$C2,$FC,$29,$F2,$C6,$0F,$6C,$2F,$C9,$5D,$F0,$CC
	.dc.b	$63,$1F,$75,$78,$67,$A9,$A7,$C7,$5C,$70,$F8,$71,$FD,$08,$EC,$52
	.dc.b	$99,$BD,$7C,$5F,$FA,$7F,$DA,$BA,$1A,$4A,$FE,$9B,$3F,$DF,$F0,$E3
	.dc.b	$B7,$F9,$CE,$72,$D7,$FC,$E3,$3B,$73,$3F,$4F,$1F,$D4,$C5,$E2,$FB
	.dc.b	$D9,$FE,$32,$F8,$37,$0E,$CF,$E4,$2F,$8D,$FD,$46,$C6,$F0,$FC,$05
	.dc.b	$F2,$9F,$C5,$1A,$2E,$1E,$07,$F1,$D1,$BB,$FE,$00,$D7,$FC,$09,$3B
	.dc.b	$7F,$82,$4F,$37,$78,$79,$A3,$9B,$63,$73,$C5,$E6,$79,$A3,$4A,$BC
	.dc.b	$15,$D6,$36,$7E,$4A,$3C,$86,$AB,$A4,$EE,$3D,$5B,$BE,$77,$FD,$9F
	.dc.b	$7F,$C0,$84,$67,$E0,$7F,$43,$C3,$FC,$16,$B7,$5F,$E2,$55,$F0,$3A
	.dc.b	$8E,$31,$AD,$3F,$87,$E5,$A9,$ED,$8F,$F8,$AE,$F8,$3D,$E3,$3F,$3A
	.dc.b	$F8,$BC,$AF,$8B,$A6,$22,$4F,$08,$CB,$CA,$3C,$8B,$E7,$7A,$3F,$E7
	.dc.b	$26,$BF,$E5,$E3,$51,$EB,$EB,$75,$3E,$13,$9D,$F0,$D7,$FC,$73,$5F
	.dc.b	$D3,$AF,$87,$5E,$78,$3F,$8B,$5D,$CB,$FA,$43,$6F,$84,$7D,$16,$BE
	.dc.b	$38,$E1,$AB,$E7,$7F,$D7,$33,$F2,$9A,$FF,$95,$DB,$97,$F3,$E5,$E4
	.dc.b	$FC,$E7,$38,$FF,$3F,$7F,$D8,$B5,$CF,$C9,$D7,$95,$B7,$EA,$EE,$F9
	.dc.b	$DA,$66,$BC,$5E,$5F,$5F,$2E,$46,$0F,$33,$CB,$D2,$91,$83,$D5,$E6
	.dc.b	$76,$F9,$89,$18,$3D,$DE,$6F,$8B,$E7,$24,$60,$F8,$5E,$77,$9B,$E6
	.dc.b	$E4,$60,$F8,$DE,$7F,$AB,$E7,$24,$60,$FC,$9F,$85,$FE,$72,$7F,$C1
	.dc.b	$DF,$37,$F0,$0F,$CA,$C6,$79,$FF,$B4,$BF,$E2,$09,$FF,$09,$7C,$BF
	.dc.b	$C6,$37,$E9,$70,$F4,$BE,$4F,$AF,$2B,$BC,$74,$F4,$FE,$6F,$CD,$C9
	.dc.b	$F1,$C7,$83,$E4,$BE,$8F,$CD,$C9,$F3,$C7,$93,$D4,$FA,$BF,$29,$27
	.dc.b	$D7,$1E,$8F,$17,$EC,$FC,$6C,$9F,$7C,$7B,$3B,$7E,$EF,$C1,$49,$F8
	.dc.b	$71,$F0,$3B,$3F,$6B,$EC,$3E,$56,$99,$AF,$F6,$DF,$CE,$FF,$B9,$1B
	.dc.b	$DB,$FB,$08,$C4,$7F,$B2,$B9,$DF,$39,$FE,$EC,$8D,$9F,$EC,$8F,$33
	.dc.b	$F8,$54,$BC,$3F,$D4,$4A,$F9,$CF,$D7,$6F,$91,$FE,$1D,$AF,$D7,$4F
	.dc.b	$1B,$F1,$63,$F6,$BC,$29,$8F,$E0,$1F,$BB,$E1,$62,$3F,$D8,$1D,$A9
	.dc.b	$C6,$1F,$80,$AF,$C1,$F8,$3F,$84,$F0,$26,$1F,$8A,$8C,$BF,$8C,$8C
	.dc.b	$BD,$23,$F8,$9F,$CE,$BE,$12,$92,$7F,$D6,$1D,$2C,$31,$1F,$26,$3C
	.dc.b	$8E,$B2,$CE,$5F,$99,$C3,$A4,$E4,$87,$F1,$D7,$2F,$BB,$C2,$FC,$E9
	.dc.b	$1F,$D1,$F8,$55,$D1,$FA,$68,$FD,$1C,$7E,$93,$C1,$FD,$2F,$81,$2B
	.dc.b	$F5,$71,$97,$F5,$B1,$97,$E6,$C7,$F3,$4F,$9D,$76,$D2,$4F,$F5,$31
	.dc.b	$F2,$EF,$80,$CC,$7F,$37,$8F,$F2,$67,$3F,$F9,$0D,$B1,$FE,$75,$BE
	.dc.b	$AF,$36,$BC,$7D,$6F,$E8,$B3,$C2,$13,$AD,$BF,$7F,$E1,$FF,$02,$E1
	.dc.b	$95,$FC,$22,$3F,$82,$98,$7F,$13,$18,$7F,$DD,$5D,$6F,$8B,$FC,$EB
	.dc.b	$63,$31,$8F,$6B,$2C,$8C,$9C,$A3,$E4,$70,$C8,$F6,$3E,$AE,$6C,$47
	.dc.b	$C5,$F8,$4F,$69,$8C,$FF,$8F,$68,$BC,$D7,$CF,$D7,$F6,$8F,$F6,$39
	.dc.b	$D3,$FE,$05,$A7,$CE,$BE,$6F,$F6,$ED,$9D,$35,$FE,$00,$BD,$ED,$CC
	.dc.b	$9D,$BC,$A3,$F9,$7F,$24,$9F,$E6,$AF,$F3,$2F,$1B,$11,$FB,$68,$B8
	.dc.b	$7F,$34,$9F,$EF,$FF,$39,$EA,$87,$2F,$FA,$DF,$E0,$5F,$3B,$98,$C3
	.dc.b	$71,$FF,$5E,$5D,$4E,$6B,$8F,$07,$F8,$7D,$B3,$F9,$79,$E5,$EC,$FE
	.dc.b	$63,$C7,$FF,$0E,$F0,$FF,$29,$AF,$F1,$0B,$56,$7F,$EB,$48,$FF,$53
	.dc.b	$C6,$A7,$F9,$8F,$C3,$48,$C1,$CB,$E4,$FA,$FC,$9C,$8C,$10,$EB,$B7
	.dc.b	$FA,$F7,$04,$EE,$72,$7F,$90,$21,$5E,$8F,$23,$D3,$F0,$12,$BD,$31
	.dc.b	$97,$D7,$19,$79,$E9,$FB,$89,$F8,$3F,$11,$27,$E0,$BB,$D7,$E0,$3F
	.dc.b	$D3,$B0,$93,$A3,$E3,$F2,$FC,$AE,$59,$18,$21,$E0,$CF,$F1,$CF,$CE
	.dc.b	$8F,$A3,$3F,$E4,$2F,$9D,$7A,$7F,$4F,$E2,$7E,$1D,$FF,$BE,$E7,$3C
	.dc.b	$5F,$EC,$FE,$0F,$F6,$74,$62,$7F,$D6,$D1,$8F,$03,$FD,$A5,$7C,$A5
	.dc.b	$5B,$8C,$36,$B9,$E1,$E1,$79,$57,$73,$C5,$19,$6E,$19,$7A,$C6,$5E
	.dc.b	$F8,$78,$AB,$AD,$6E,$FC,$15,$C8,$F4,$9E,$FA,$E0,$E7,$0F,$3E,$1D
	.dc.b	$6B,$8C,$F6,$3A,$46,$49,$7D,$EF,$9C,$F9,$A3,$64,$7C,$27,$32,$D7
	.dc.b	$62,$57,$86,$B9,$C6,$2B,$95,$FC,$C9,$CD,$78,$AF,$9B,$C5,$DE,$1F
	.dc.b	$11,$CD,$F9,$DC,$9F,$06,$35,$7D,$63,$D0,$0E,$45,$3E,$61,$4F,$80
	.dc.b	$53,$F0,$7E,$AF,$F3,$F7,$58,$A3,$11,$F8,$E6,$BF,$93,$6B,$CB,$C0
	.dc.b	$FB,$91,$FD,$26,$BF,$21,$F2,$4F,$8C,$FE,$3F,$7F,$32,$B5,$8F,$5C
	.dc.b	$38,$9F,$8B,$BE,$1C,$D8,$BE,$13,$71,$EB,$8F,$47,$F3,$78,$F4,$AF
	.dc.b	$35,$6F,$1F,$89,$8F,$49,$E6,$4B,$E3,$5E,$1C,$3A,$5F,$96,$B4,$DA
	.dc.b	$F0,$E3,$D6,$7B,$9A,$3E,$A6,$1D,$E7,$AC,$64,$F9,$07,$E4,$23,$55
	.dc.b	$D4,$91,$D2,$7B,$BC,$72,$FA,$D8,$F8,$61,$7C,$9E,$17,$C9,$EF,$7C
	.dc.b	$8F,$19,$23,$53,$C6,$77,$1F,$CD,$7E,$2F,$EC,$F2,$7F,$43,$F9,$DB
	.dc.b	$9E,$4F,$E9,$2E,$2F,$F9,$C5,$5E,$7F,$F1,$3C,$DE,$44,$D8,$C7,$C5
	.dc.b	$EF,$D6,$FE,$87,$8F,$F5,$48,$C6,$7A,$8C,$1F,$D6,$3E,$2B,$19,$FF
	.dc.b	$2F,$F9,$3E,$D1,$DD,$B3,$FD,$79,$B3,$B8,$E4,$7F,$62,$F8,$AF,$37
	.dc.b	$CE,$F9,$DB,$19,$3F,$B3,$C6,$5A,$7F,$6A,$F9,$CF,$57,$77,$BD,$B1
	.dc.b	$92,$5F,$F4,$5F,$9D,$D4,$7F,$93,$7E,$73,$FC,$A2,$FE,$F7,$8F,$F5
	.dc.b	$3E,$23,$E7,$7F,$35,$1F,$EC,$5A,$DE,$1F,$DA,$2B,$5D,$5F,$4E,$13
	.dc.b	$AA,$E7,$7C,$B0,$FE,$11,$39,$7F,$C2,$91,$9A,$FE,$E5,$F3,$BF,$64
	.dc.b	$FB,$5F,$3B,$6B,$FC,$13,$16,$7F,$B0,$4D,$C3,$F1,$3F,$39,$FE,$8D
	.dc.b	$BF,$F1,$C3,$6B,$FD,$64,$BE,$5F,$F1,$54,$8D,$3C,$BD,$1C,$C6,$2F
	.dc.b	$E3,$3C,$5E,$B1,$F0,$7E,$4D,$F1,$FD,$3E,$75,$F2,$39,$8C,$36,$FF
	.dc.b	$C9,$93,$A7,$FD,$01,$5C,$9C,$CE,$F3,$FF,$0B,$89,$D3,$FE,$CD,$A7
	.dc.b	$CC,$6F,$3F,$D5,$FC,$7F,$F8,$73,$1F,$F2,$D3,$6F,$FD,$A8,$FE,$89
	.dc.b	$BE,$67,$C4,$8D,$57,$28,$C7,$66,$5F,$D7,$6D,$FB,$1F,$37,$FA,$36
	.dc.b	$3F,$E3,$E8,$C6,$3A,$8F,$C7,$1F,$D2,$FC,$35,$70,$9F,$F0,$D5,$70
	.dc.b	$AF,$87,$E8,$FD,$6F,$9D,$7C,$3F,$CF,$8C,$36,$7F,$CE,$D3,$A7,$89
	.dc.b	$88,$D5,$FC,$AE,$05,$F5,$F9,$1C,$ED,$FD,$2B,$4F,$2F,$95,$ED,$F6
	.dc.b	$FD,$CD,$CF,$D7,$92,$B7,$18,$7D,$B5,$C2,$31,$B2,$3E,$47,$95,$27
	.dc.b	$3E,$FD,$C3,$FC,$97,$7F,$CC,$27,$8D,$7C,$FA,$D6,$5F,$D8,$2D,$E2
	.dc.b	$BD,$8F,$CE,$FC,$4F,$FA,$47,$1F,$F6,$64,$E5,$FF,$4E,$46,$6B,$F1
	.dc.b	$B3,$97,$FC,$CB,$69,$D3,$DE,$74,$FF,$38,$8C,$D7,$F8,$16,$72,$DB
	.dc.b	$F1,$FA,$9F,$C1,$E7,$38,$7C,$9C,$DF,$3B,$F8,$BF,$81,$FC,$63,$71
	.dc.b	$F9,$2D,$AF,$8A,$DC,$FE,$C5,$BE,$77,$F2,$87,$E5,$7B,$EE,$1F,$4A
	.dc.b	$2E,$1F,$3A,$6D,$FC,$BF,$9D,$7C,$77,$E4,$7E,$57,$F8,$F4,$63,$1F
	.dc.b	$C7,$C6,$27,$51,$FE,$F9,$9D,$61,$F2,$BA,$97,$BA,$D5,$FF,$12,$F9
	.dc.b	$D7,$C3,$E7,$0F,$C0,$F8,$BF,$D5,$FD,$EF,$D7,$37,$0F,$C5,$F6,$7F
	.dc.b	$5B,$E0,$FD,$93,$63,$F1,$FE,$5F,$F5,$FE,$5F,$DB,$36,$3F,$27,$E2
	.dc.b	$7F,$63,$F7,$BF,$74,$D8,$FC,$BF,$8F,$FD,$9F,$D7,$FD,$F3,$63,$F3
	.dc.b	$7F,$33,$FD,$AF,$D8,$FE,$04,$D8,$FC,$FF,$97,$FD,$BF,$D7,$FE,$0C
	.dc.b	$D8,$FD,$1F,$7F,$F7,$3F,$5F,$F8,$53,$63,$F4,$FC,$FF,$DD,$FB,$FF
	.dc.b	$87,$36,$3F,$57,$E3,$7F,$7B,$F5,$FF,$89,$36,$3F,$5F,$E4,$FF,$7F
	.dc.b	$F5,$FF,$8B,$36,$3F,$67,$E7,$7F,$80,$FD,$6F,$E3,$4D,$8F,$DB,$FC
	.dc.b	$BF,$F8,$1F,$D7,$FE,$3C,$D8,$FD,$F7,$CC,$7F,$BB,$9F,$E0,$1E,$D3
	.dc.b	$D6,$FF,$0A,$4E,$7C,$0F,$F6,$2C,$E6,$B8,$FC,$C6,$9D,$B5,$EB,$F3
	.dc.b	$1F,$E1,$7F,$63,$9B,$63,$C3,$F0,$18,$69,$C0,$B8,$7F,$B5,$31,$FF
	.dc.b	$38,$47,$A3,$E0,$BC,$B5,$FD,$67,$D1,$FF,$16,$7A,$CF,$97,$78,$99
	.dc.b	$C3,$F8,$FC,$E6,$31,$97,$FB,$AA,$31,$3F,$D6,$AF,$9D,$F2,$8C,$12
	.dc.b	$B8,$6D,$FC,$8E,$31,$E7,$7F,$31,$D1,$F3,$1F,$2B,$0F,$96,$DA,$FA
	.dc.b	$2D,$C3,$DC,$DB,$FE,$75,$E9,$FF,$14,$9F,$E6,$DF,$33,$FD,$9A,$FF
	.dc.b	$9B,$4E,$7D,$84,$E6,$BF,$DF,$73,$92,$78,$DF,$E4,$F7,$A4,$7F,$B8
	.dc.b	$87,$CE,$A3,$FD,$4D,$8F,$F5,$EC,$7F,$D5,$53,$9A,$FF,$54,$7C,$26
	.dc.b	$2F,$7E,$6F,$FA,$6A,$B5,$5D,$5B,$87,$38,$EF,$5F,$EA,$FB,$E5,$1F
	.dc.b	$E3,$BB,$E5,$39,$8D,$99,$C3,$6F,$58,$F0,$C6,$23,$FD,$81,$5A,$BF
	.dc.b	$06,$1E,$49,$C9,$D0,$46,$AF,$C1,$3E,$8C,$3C,$75,$F0,$0E,$15,$B7
	.dc.b	$13,$BC,$3C,$50,$C3,$C5,$7C,$A3,$D0,$7A,$CF,$11,$E4,$AE,$4E,$0E
	.dc.b	$47,$FB,$54,$F4,$96,$70,$E0,$FF,$6F,$7C,$C7,$FA,$2B,$E0,$F8,$36
	.dc.b	$3F,$DB,$CD,$7F,$DC,$4D,$78,$F9,$BF,$11,$DB,$5D,$B5,$DE,$5D,$A3
	.dc.b	$86,$1F,$82,$CF,$FC,$11,$5F,$61,$AF,$D1,$B5,$70,$9E,$73,$F5,$5D
	.dc.b	$D7,$3F,$73,$FD,$F4,$FE,$3A,$F8,$E1,$F8,$46,$DF,$1A,$DD,$FF,$C0
	.dc.b	$13,$99,$DA,$F9,$61,$E1,$A1,$FA,$3A,$E2,$E2,$B8,$BA,$BE,$F7,$CB
	.dc.b	$0F,$75,$F0,$AE,$47,$97,$0E,$CA,$F8,$E1,$E5,$AE,$0F,$69,$D5,$70
	.dc.b	$AD,$C7,$8E,$B6,$77,$8C,$EE,$F9,$DF,$F1,$6D,$7F,$8F,$5A,$FE,$83
	.dc.b	$6F,$F2,$E4,$63,$2F,$F2,$52,$F1,$7F,$81,$CE,$BC,$7B,$7F,$E4,$9A
	.dc.b	$E1,$39,$8E,$D5,$A5,$F3,$8F,$86,$C3,$63,$FC,$A6,$6A,$3A,$F5,$3A
	.dc.b	$3C,$53,$AA,$E0,$FF,$A3,$1B,$9F,$F9,$2D,$B8,$FD,$AB,$E5,$8F,$4F
	.dc.b	$B5,$FB,$B8,$7F,$44,$6B,$FD,$0D,$B8,$2F,$94,$3E,$07,$95,$43,$91
	.dc.b	$D5,$C3,$FD,$37,$E6,$3F,$3A,$FF,$52,$6B,$DC,$F9,$7B,$FF,$3A,$F3
	.dc.b	$6F,$57,$F8,$C2,$7F,$CE,$8D,$7E,$C1,$FD,$12,$37,$E6,$C8,$DB,$CA
	.dc.b	$3A,$EF,$FE,$44,$9F,$F0,$0F,$B7,$83,$83,$FD,$FB,$E0,$FC,$7E,$07
	.dc.b	$FB,$36,$3F,$C0,$B1,$E6,$FC,$67,$E2,$EF,$FE,$A3,$7E,$24,$FF,$A2
	.dc.b	$5A,$F9,$B4,$46,$38,$3F,$D2,$D8,$75,$AD,$4F,$F0,$49,$FF,$4D,$36
	.dc.b	$FB,$74,$78,$7E,$13,$C5,$E1,$CC,$3C,$0F,$97,$E0,$FE,$5F,$5B,$CA
	.dc.b	$A4,$AF,$37,$77,$AF,$E0,$E4,$7A,$5F,$77,$C1,$FB,$FD,$E9,$3E,$D7
	.dc.b	$E1,$7B,$DF,$0F,$DE,$93,$F0,$5F,$8D,$EF,$7C,$7F,$7A,$4F,$C5,$7E
	.dc.b	$57,$BB,$F2,$FD,$C9,$3F,$25,$F9,$DD,$7E,$AF,$9E,$93,$FE,$B5,$E0
	.dc.b	$47,$D3,$ED,$FA,$9E,$E4,$AF,$A4,$FD,$7F,$6F,$EC,$7B,$52,$7E,$B3
	.dc.b	$F6,$FD,$9F,$C8,$FA,$7F,$EB,$3E,$DF,$F6,$53,$7B,$3F,$DC,$0D,$BF
	.dc.b	$F6,$43,$F3,$5F,$F6,$42,$F5,$F9,$92,$33,$EB,$7F,$03,$F9,$47,$C9
	.dc.b	$D2,$4F,$F0,$2F,$8D,$77,$E2,$3F,$3E,$2B,$8B,$FC,$21,$AE,$A7,$8B
	.dc.b	$9A,$E5,$9F,$FB,$3E,$B6,$FF,$B4,$6B,$75,$F9,$7C,$9E,$D7,$C4,$24
	.dc.b	$7B,$BD,$9A,$EE,$2D,$FC,$89,$FC,$8F,$CB,$FE,$4B,$E4,$65,$FD,$98
	.dc.b	$CB,$F6,$E3,$2F,$E9,$F9,$93,$FC,$BA,$DF,$2E,$F1,$70,$DB,$FF,$7D
	.dc.b	$CE,$A7,$3E,$2E,$99,$3A,$AE,$0F,$FB,$99,$FF,$6C,$B9,$8D,$B9,$79
	.dc.b	$BC,$7C,$FF,$09,$1E,$9E,$3D,$7A,$45,$C3,$F9,$54,$7F,$8B,$AF,$F7
	.dc.b	$8F,$E2,$B6,$FC,$8F,$5C,$8F,$C5,$C7,$E5,$0C,$3F,$56,$30,$FE,$07
	.dc.b	$E5,$FF,$18,$47,$F1,$5E,$2D,$7F,$13,$C6,$9C,$63,$F8,$1A,$BD,$1F
	.dc.b	$D1,$2B,$57,$D4,$E6,$7E,$7B,$CD,$FD,$3F,$A1,$23,$04,$3F,$14,$E5
	.dc.b	$FC,$64,$7D,$BF,$97,$FF,$4A,$C7,$D8,$6B,$FC,$F3,$14,$61,$F9,$26
	.dc.b	$3D,$9E,$5E,$FC,$AE,$4F,$58,$FD,$77,$CD,$C9,$FD,$33,$E1,$F8,$FC
	.dc.b	$47,$C4,$8B,$7D,$B0,$56,$34,$F8,$95,$6F,$FC,$CF,$5A,$7F,$8A,$76
	.dc.b	$7F,$30,$23,$FC,$5F,$E3,$EB,$F9,$D8,$B8,$FE,$13,$D3,$FF,$11,$57
	.dc.b	$F1,$2C,$25,$FF,$0C,$7F,$8A,$FC,$E6,$23,$F6,$D1,$6F,$F8,$BC,$FF
	.dc.b	$83,$FE,$61,$E8,$52,$5F,$F8,$F2,$BF,$27,$F0,$3F,$E7,$5F,$6F,$33
	.dc.b	$FC,$3A,$F8,$BF,$E1,$46,$BF,$DC,$27,$83,$FC,$83,$E7,$7F,$94,$91
	.dc.b	$FE,$47,$F3,$95,$FE,$19,$17,$0F,$E5,$B3,$FC,$C3,$A2,$57,$F2,$A7
	.dc.b	$F9,$9F,$C8,$62,$3F,$8A,$45,$AF,$E6,$B3,$FC,$E3,$E7,$25,$7F,$32
	.dc.b	$7F,$9D,$F8,$58,$8F,$E3,$91,$6B,$F9,$EC,$FF,$40,$F1,$25,$7F,$3A
	.dc.b	$7F,$A1,$F1,$E2,$3F,$92,$45,$AF,$E8,$B3,$FD,$23,$8A,$57,$F4,$27
	.dc.b	$FC,$37,$F8,$D3,$E6,$F3,$3F,$CB,$66,$DF,$E0,$6B,$53,$FE,$42,$F9
	.dc.b	$BF,$E6,$37,$FE,$BA,$F9,$8F,$F1,$BC,$62,$7F,$D5,$91,$8F,$2B,$8F
	.dc.b	$C5,$23,$7A,$7F,$8F,$A3,$15,$FE,$7C,$BE,$51,$88,$46,$2B,$A6,$1E
	.dc.b	$19,$E1,$7C,$E7,$31,$E2,$3A,$61,$E2,$8C,$E1,$D7,$05,$F1,$BD,$E1
	.dc.b	$D5,$F0,$E1,$C5,$C9,$D4,$F2,$CF,$17,$73,$C6,$BD,$11,$C2,$FC,$15
	.dc.b	$B9,$CC,$F0,$3A,$38,$3C,$AF,$3B,$F1,$3B,$AF,$54,$62,$77,$3C,$A3
	.dc.b	$04,$BF,$25,$F8,$63,$13,$B9,$CB,$DD,$8E,$4F,$2C,$F1,$8C,$DF,$89
	.dc.b	$E3,$3B,$8C,$37,$0F,$C0,$BD,$0F,$F3,$EE,$1F,$89,$6E,$1F,$90,$6D
	.dc.b	$FF,$BB,$65,$E7,$7F,$96,$A7,$F2,$2D,$7F,$14,$D7,$F1,$2D,$8E,$3E
	.dc.b	$3F,$D5,$8F,$F3,$2E,$3F,$8F,$F2,$5E,$D7,$F6,$48,$FC,$34,$EB,$3E
	.dc.b	$F7,$C4,$FC,$35,$F0,$EF,$C6,$5F,$6A,$F8,$3E,$4A,$DE,$3D,$70,$E7
	.dc.b	$3B,$9E,$38,$F8,$EB,$53,$C4,$D1,$23,$58,$70,$BE,$71,$E8,$BE,$F3
	.dc.b	$9A,$E5,$39,$7F,$35,$8F,$A5,$E9,$7C,$88,$FF,$BC,$23,$53,$C4,$93
	.dc.b	$A9,$FF,$11,$61,$E8,$7E,$8E,$1E,$16,$9E,$08,$D6,$5F,$52,$3C,$27
	.dc.b	$9A,$BF,$D8,$78,$78,$1E,$D3,$BC,$3C,$35,$E5,$BE,$79,$7D,$0E,$E6
	.dc.b	$EF,$9C,$FE,$EF,$AB,$05,$6A,$73,$18,$AF,$ED,$F5,$FD,$54,$E1,$5B
	.dc.b	$71,$3B,$AE,$33,$98,$FF,$57,$57,$26,$CE,$A3,$F9,$25,$72,$AF,$3E
	.dc.b	$79,$F8,$8D,$32,$75,$5C,$30,$EB,$59,$7B,$FC,$B7,$53,$B7,$C6,$DB
	.dc.b	$E3,$16,$7C,$B3,$79,$3F,$A0,$78,$DE,$53,$CD,$E3,$5C,$3C,$31,$6F
	.dc.b	$DB,$36,$BF,$A4,$7C,$77,$B4,$F7,$7C,$B5,$C3,$D3,$16,$FE,$2C,$DA
	.dc.b	$FF,$29,$C3,$E7,$1D,$5F,$52,$30,$FD,$6F,$09,$4F,$77,$57,$C3,$2F
	.dc.b	$AB,$87,$06,$C6,$63,$FC,$11,$5C,$67,$38,$7D,$5F,$97,$91,$F4,$3E
	.dc.b	$5E,$DF,$DC,$9B,$7F,$D9,$E6,$D7,$E1,$BD,$87,$CA,$7E,$11,$B1,$F7
	.dc.b	$D7,$CA,$66,$64,$E5,$E1,$18,$C7,$FD,$5F,$F2,$0F,$80,$F8,$8D,$BF
	.dc.b	$F4,$7F,$C6,$BE,$17,$35,$27,$E1,$D6,$EF,$FC,$B9,$7B,$9C,$D6,$E7
	.dc.b	$BB,$FA,$4D,$15,$B9,$CD,$78,$3D,$7D,$49,$BE,$86,$19,$3B,$3C,$F1
	.dc.b	$8A,$FD,$3F,$C9,$FE,$97,$D1,$FD,$33,$6B,$F4,$2D,$9D,$C7,$E9,$27
	.dc.b	$E1,$FC,$34,$8F,$F5,$8F,$CD,$BE,$27,$17,$F2,$AA,$E9,$FE,$7A,$AD
	.dc.b	$5F,$C6,$9C,$BF,$C1,$E9,$B2,$7E,$25,$70,$9E,$95,$FB,$6F,$98,$FD
	.dc.b	$F7,$CC,$7E,$FD,$B1,$93,$F7,$71,$96,$9F,$EC,$8F,$88,$FE,$11,$97
	.dc.b	$FA,$39,$B1,$FE,$67,$7F,$81,$46,$7D,$BF,$AD,$1A,$C3,$F4,$AD,$93
	.dc.b	$C2,$B9,$57,$DD,$F9,$9F,$D2,$57,$FA,$71,$B3,$DF,$CF,$D7,$5C,$8F
	.dc.b	$F6,$EC,$EB,$1E,$F5,$97,$FD,$51,$19,$B3,$71,$FF,$29,$5F,$2C,$3F
	.dc.b	$4D,$C9,$FC,$AB,$C3,$FE,$6D,$3F,$A7,$8C,$63,$A8,$FE,$51,$3A,$8F
	.dc.b	$E0,$71,$FB,$19,$DB,$FC,$A3,$E6,$7F,$98,$1F,$CC,$7A,$AC,$64,$95
	.dc.b	$FC,$E2,$32,$FE,$C3,$E3,$BF,$5E,$7E,$8F,$E6,$9F,$1B,$8B,$FD,$6C
	.dc.b	$DA,$FF,$64,$56,$A7,$24,$8E,$8F,$F4,$AD,$93,$B9,$C9,$FD,$2B,$B7
	.dc.b	$FA,$41,$FD,$27,$3B,$7F,$D2,$7F,$1B,$F8,$58,$7F,$AE,$5B,$1F,$D6
	.dc.b	$17,$A9,$FD,$59,$CE,$1F,$D8,$97,$C4,$64,$D7,$AB,$6F,$B7,$C6,$71
	.dc.b	$39,$7C,$55,$8F,$A7,$B7,$D9,$3B,$7C,$95,$8F,$37,$C9,$78,$4F,$17
	.dc.b	$87,$63,$D5,$E1,$F9,$4F,$35,$D8,$F7,$7A,$FE,$93,$D5,$E1,$58,$F8
	.dc.b	$5F,$25,$ED,$3D,$DE,$3D,$8F,$8D,$E1,$FC,$13,$E1,$79,$96,$3F,$16
	.dc.b	$FE,$44,$F8,$DF,$25,$63,$E7,$75,$7C,$93,$E5,$79,$36,$3E,$97,$81
	.dc.b	$F3,$4F,$9D,$F2,$56,$3F,$BF,$7E,$38,$7F,$C0,$ED,$73,$F2,$98,$72
	.dc.b	$C7,$8F,$CA,$5E,$BF,$CF,$9F,$E1,$AF,$96,$FE,$9A,$E3,$DC,$46,$3B
	.dc.b	$FF,$DF,$9F,$2F,$DE,$B5,$39,$C3,$FC,$3B,$39,$AD,$C7,$F8,$72,$18
	.dc.b	$D8,$C3,$5F,$F0,$D4,$69,$A7,$F8,$B3,$CE,$7C,$CF,$F2,$7B,$F8,$58
	.dc.b	$FF,$2C,$F9,$17,$AB,$BA,$D6,$1F,$E3,$7C,$15,$A8,$FF,$18,$C7,$4E
	.dc.b	$8E,$07,$1E,$8B,$3A,$BE,$78,$FF,$3A,$6D,$F9,$2B,$73,$FD,$BB,$C0
	.dc.b	$FE,$DA,$7A,$76,$FD,$2E,$1F,$DA,$A3,$2D,$3C,$71,$A9,$E5,$1F,$E3
	.dc.b	$DE,$F7,$B9,$FE,$46,$6C,$67,$DC,$C4,$C8,$E0,$EE,$31,$5F,$80,$3F
	.dc.b	$DD,$43,$E7,$17,$FD,$5F,$81,$E6,$FC,$E9,$CD,$7F,$5B,$1F,$07,$FE
	.dc.b	$63,$8C,$C7,$F8,$B2,$78,$5E,$E3,$67,$77,$B4,$6A,$3A,$57,$39,$CE
	.dc.b	$56,$72,$71,$3F,$CE,$58,$79,$A3,$51,$EC,$C3,$C7,$39,$8E,$06,$6B
	.dc.b	$AB,$83,$FC,$F9,$7C,$A7,$85,$7B,$27,$27,$28,$CD,$7B,$6B,$86,$1E
	.dc.b	$BA,$EC,$79,$9C,$3C,$AF,$CB,$18,$8F,$8B,$18,$C3,$D7,$1B,$BF,$3C
	.dc.b	$F8,$E3,$E5,$9F,$34,$A7,$8C,$C1,$83,$FD,$31,$F2,$EF,$79,$EE,$F1
	.dc.b	$8C,$36,$33,$3F,$66,$72,$DF,$7F,$FC,$D2,$DD,$39,$46,$23,$F0,$D2
	.dc.b	$8C,$47,$F8,$8E,$31,$3F,$63,$1F,$E9,$33,$7C,$BF,$F1,$46,$5F,$8C
	.dc.b	$BD,$CF,$DA,$6C,$7E,$42,$B8,$CF,$DE,$6C,$71,$C3,$CE,$DC,$3F,$D7
	.dc.b	$5F,$07,$C7,$1F,$F1,$B3,$70,$CE,$1E,$7C,$3D,$F3,$C6,$7B,$57,$9A
	.dc.b	$BF,$C7,$C7,$80,$E1,$7F,$94,$C7,$CF,$1A,$9E,$15,$FE,$CB,$8E,$0F
	.dc.b	$D6,$C3,$CF,$1C,$A3,$B9,$F1,$4E,$91,$A7,$85,$FA,$1E,$95,$EC,$AF
	.dc.b	$04,$64,$5F,$CC,$C3,$F7,$BD,$6F,$9C,$7B,$BF,$C0,$F7,$FF,$6C,$C6
	.dc.b	$30,$FF,$68,$35,$FE,$90,$BD,$DE,$11,$FE,$E6,$8D,$3F,$E6,$3B,$ED
	.dc.b	$7F,$99,$6C,$7F,$07,$6B,$FC,$27,$E7,$1F,$07,$F9,$C6,$C7,$E8,$5A
	.dc.b	$EE,$34,$78,$BA,$9F,$31,$AE,$0F,$E9,$31,$89,$F3,$D2,$F5,$A7,$F4
	.dc.b	$9E,$CF,$E4,$18,$70,$6C,$66,$BB,$47,$F8,$99,$FE,$90,$7E,$D7,$0F
	.dc.b	$EB,$2D,$8F,$DA,$C6,$23,$F9,$33,$5F,$EE,$AF,$F5,$8E,$2F,$57,$77
	.dc.b	$EA,$E7,$A9,$FF,$00,$7E,$71,$E4,$FD,$0D,$91,$89,$D4,$E6,$3F,$DC
	.dc.b	$11,$83,$E8,$3B,$7F,$9F,$4F,$29,$FA,$D3,$99,$FB,$5E,$13,$E5,$FE
	.dc.b	$D3,$63,$ED,$35,$E6,$D7,$F0,$5D,$FF,$EC,$47,$F0,$0D,$7C,$2D,$77
	.dc.b	$EB,$FE,$1F,$E7,$3E,$74,$79,$9A,$FE,$35,$A8,$64,$8C,$C6,$1F,$E1
	.dc.b	$2F,$1E,$AE,$C4,$7F,$29,$0F,$0B,$C4,$48,$CB,$F9,$BE,$AF,$F0,$FC
	.dc.b	$7E,$6D,$AF,$CC,$6B,$FA,$36,$BE,$7B,$F3,$7A,$9E,$F2,$47,$89,$F8
	.dc.b	$5E,$9C,$8F,$2B,$5F,$89,$E9,$7C,$72,$3E,$A7,$E5,$78,$D2,$3D,$AD
	.dc.b	$7E,$67,$A3,$F3,$C8,$FC,$27,$DD,$F3,$52,$3D,$CD,$7E,$A7,$A3,$F5
	.dc.b	$C8,$FC,$A7,$ED,$7A,$12,$3E,$6B,$5F,$C8,$F7,$7F,$8F,$9B,$D7,$FC
	.dc.b	$1E,$31,$1F,$C2,$DA,$FC,$95,$E7,$6C,$91,$FC,$36,$32,$FF,$69,$F9
	.dc.b	$A7,$C8,$7F,$02,$6C,$7F,$83,$AF,$FD,$F5,$18,$8F,$F1,$57,$C3,$FE
	.dc.b	$0D,$D1,$F8,$F8,$D3,$FE,$27,$6B,$CA,$BF,$BA,$C6,$23,$84,$7F,$12
	.dc.b	$9F,$F0,$CD,$7E,$66,$51,$87,$67,$83,$87,$F8,$EF,$C7,$FF,$01,$C4
	.dc.b	$D3,$1F,$39,$6B,$84,$E9,$B3,$A7,$A7,$23,$E4,$BF,$86,$2F,$92,$A6
	.dc.b	$22,$4F,$08,$CB,$CA,$3F,$A1,$F8,$8F,$9C,$7B,$F8,$6D,$7F,$B6,$DA
	.dc.b	$FF,$A5,$A7,$87,$BF,$BA,$FF,$10,$47,$36,$D7,$0A,$FF,$2D,$35,$FF
	.dc.b	$23,$BD,$27,$FD,$12,$78,$FA,$FC,$47,$59,$F1,$7C,$0F,$94,$8F,$97
	.dc.b	$BA,$BE,$5B,$FE,$27,$7D,$1B,$C5,$E0,$AF,$A9,$BF,$D5,$E8,$98,$7D
	.dc.b	$28,$CB,$F5,$23,$2F,$FA,$CE,$3F,$AC,$BF,$8E,$1C,$AF,$96,$FC,$C1
	.dc.b	$23,$5C,$AF,$81,$C4,$7F,$9C,$EB,$83,$FA,$1A,$FC,$99,$1E,$3F,$03
	.dc.b	$F9,$33,$8E,$BF,$EB,$0E,$7F,$ED,$7E,$2F,$E3,$FC,$89,$7F,$8C,$8F
	.dc.b	$C9,$18,$7D,$51,$87,$F4,$F4,$F9,$C7,$B3,$F8,$66,$CE,$9A,$FE,$0D
	.dc.b	$FE,$C4,$D8,$E0,$47,$87,$97,$FB,$BF,$9C,$91,$F2,$75,$47,$7F,$8C
	.dc.b	$CC,$FC,$58,$FD,$A9,$1F,$E0,$CF,$E6,$BF,$7A,$7C,$D5,$38,$E7,$F1
	.dc.b	$7B,$BF,$83,$37,$C1,$FD,$49,$FA,$AF,$8C,$B8,$FE,$BA,$2B,$FE,$F1
	.dc.b	$A4,$FE,$EF,$19,$3F,$B8,$7F,$7D,$F3,$B8,$8F,$C3,$45,$C3,$EB,$73
	.dc.b	$AB,$FC,$3B,$F3,$AF,$2F,$FD,$0A,$DC,$3F,$D1,$4D,$7F,$D7,$0D,$7F
	.dc.b	$D8,$0E,$6F,$FD,$E3,$EF,$3E,$81,$E7,$7F,$90,$23,$0F,$83,$0F,$E2
	.dc.b	$9F,$3D,$31,$FE,$22,$FF,$17,$E0,$C4,$7F,$40,$8F,$E7,$C4,$7F,$9C
	.dc.b	$75,$FF,$41,$23,$82,$0C,$BF,$2D,$F0,$EF,$9B,$FF,$12,$36,$7F,$90
	.dc.b	$D7,$F2,$7F,$9D,$FE,$98,$48,$FE,$8C,$D7,$F6,$5C,$9F,$CB,$4E,$0F
	.dc.b	$F2,$AF,$2B,$F9,$81,$1F,$E5,$B1,$5F,$EB,$71,$FD,$64,$8F,$F5,$3E
	.dc.b	$4F,$EB,$84,$70,$43,$D9,$E2,$7F,$81,$71,$FE,$71,$18,$3F,$C3,$AE
	.dc.b	$33,$FC,$0D,$7F,$97,$3E,$77,$FA,$1F,$CA,$FF,$B7,$BE,$77,$FA,$64
	.dc.b	$63,$0F,$F7,$DC,$63,$CE,$56,$E7,$FC,$55,$5F,$D0,$5C,$07,$97,$C4
	.dc.b	$2F,$95,$6A,$73,$1E,$0D,$3F,$AE,$56,$EF,$AE,$1E,$07,$C6,$63,$0E
	.dc.b	$51,$99,$E1,$3E,$1A,$E7,$7E,$08,$E2,$79,$6B,$84,$66,$B8,$E1,$EB
	.dc.b	$AD,$BE,$67,$C0,$62,$B8,$BF,$0A,$FB,$BA,$AE,$73,$A8,$F0,$61,$F0
	.dc.b	$6B,$71,$83,$A4,$7B,$0D,$34,$C9,$2B,$CF,$3C,$AF,$CD,$3C,$E7,$C3
	.dc.b	$5C,$A3,$86,$1E,$F2,$3E,$93,$B4,$F0,$76,$64,$A7,$58,$E2,$FC,$F3
	.dc.b	$CC,$7F,$85,$4F,$61,$6B,$F1,$8D,$C3,$EE,$39,$E0,$FF,$38,$CF,$ED
	.dc.b	$7C,$BE,$53,$F8,$E9,$C9,$23,$F1,$4F,$FA,$B7,$0F,$C0,$F9,$AF,$59
	.dc.b	$C1,$AB,$98,$7D,$CC,$38,$67,$FE,$2B,$C3,$85,$EB,$1F,$C1,$D6,$AF
	.dc.b	$C5,$87,$E7,$AF,$71,$CE,$38,$61,$FA,$AA,$E1,$3F,$E1,$CC,$7C,$6E
	.dc.b	$C9,$87,$6B,$E9,$3A,$8C,$4E,$AF,$A5,$6B,$1E,$F3,$E4,$34,$7F,$A7
	.dc.b	$71,$F2,$DF,$22,$5F,$29,$FD,$1C,$EA,$31,$3A,$9E,$8F,$CC,$C3,$C9
	.dc.b	$3A,$24,$E9,$F9,$51,$A8,$C1,$0F,$E0,$98,$FC,$4B,$F0,$46,$72,$FC
	.dc.b	$03,$EA,$3E,$46,$3E,$EF,$1D,$F3,$AF,$85,$FF,$1B,$CF,$F2,$1F,$4E
	.dc.b	$DF,$FA,$F2,$B5,$3F,$E6,$3E,$45,$6B,$0E,$78,$7F,$A1,$9B,$5F,$90
	.dc.b	$6E,$1F,$EE,$0C,$39,$B7,$5E,$11,$FC,$76,$B8,$36,$75,$18,$9F,$04
	.dc.b	$E7,$93,$FC,$50,$BE,$0B,$33,$AF,$8A,$DC,$5B,$FF,$3E,$56,$DB,$3E
	.dc.b	$22,$3E,$27,$CB,$E4,$FB,$F5,$78,$7E,$28,$C3,$6B,$D3,$1C,$1A,$7B
	.dc.b	$3E,$27,$DE,$47,$04,$D3,$FA,$86,$BF,$5A,$7D,$FF,$3B,$F0,$1F,$AB
	.dc.b	$2F,$9D,$FF,$0D,$DF,$FA,$9B,$C3,$7D,$03,$E5,$37,$5A,$BF,$EB,$95
	.dc.b	$A9,$CF,$77,$FA,$FA,$7A,$36,$B8,$47,$F1,$36,$BF,$EE,$BE,$07,$C8
	.dc.b	$7D,$43,$EA,$FC,$F5,$CF,$E7,$C5,$AF,$6D,$2B,$1C,$BF,$E6,$1F,$59
	.dc.b	$F3,$3F,$8D,$6E,$3F,$EB,$35,$F3,$19,$99,$39,$78,$46,$31,$FE,$F4
	.dc.b	$3E,$29,$F2,$DF,$9E,$C5,$E9,$7F,$AC,$71,$FF,$41,$56,$9F,$F3,$5C
	.dc.b	$66,$F9,$CE,$5F,$E5,$AD,$7F,$9C,$35,$FF,$67,$35,$1E,$3B,$FB,$71
	.dc.b	$CE,$FF,$D5,$3F,$0F,$F6,$70,$F1,$D7,$9E,$3F,$D2,$11,$FE,$7E,$6D
	.dc.b	$FF,$AE,$DF,$F2,$94,$67,$DB,$E0,$47,$87,$9D,$6B,$E9,$FC,$77,$EB
	.dc.b	$3E,$43,$F5,$AD,$AF,$D3,$B6,$77,$1F,$AA,$9F,$4F,$C3,$FC,$BF,$9E
	.dc.b	$7C,$4E,$2B,$FD,$27,$18,$7E,$6C,$FE,$3F,$63,$49,$1F,$D6,$30,$FB
	.dc.b	$11,$CD,$FF,$6C,$53,$E5,$71,$BF,$F0,$2F,$9E,$FE,$0B,$F3,$DF,$C1
	.dc.b	$9B,$8F,$F0,$06,$E1,$F9,$26,$D7,$FB,$63,$E2,$DE,$F7,$E1,$5B,$7F
	.dc.b	$EA,$C7,$F8,$83,$F6,$A5,$1A,$BF,$F7,$63,$63,$A5,$73,$F2,$3A,$3F
	.dc.b	$91,$F9,$EF,$DA,$4F,$87,$CA,$C7,$7F,$85,$B9,$7F,$7A,$95,$7E,$CB
	.dc.b	$2F,$F3,$A3,$63,$37,$FD,$9A,$7F,$1E,$69,$E6,$F9,$BD,$3B,$3F,$B1
	.dc.b	$E7,$FE,$79,$F3,$1F,$96,$C7,$F6,$11,$8B,$FD,$A3,$63,$27,$C3,$3F
	.dc.b	$3B,$E3,$AB,$95,$7F,$00,$F9,$EF,$E7,$27,$F3,$AE,$EB,$7F,$CD,$22
	.dc.b	$CF,$F0,$69,$B8,$FA,$F9,$9F,$0F,$FB,$9F,$9E,$FD,$46,$3F,$AE,$8C
	.dc.b	$77,$7E,$B1,$FC,$71,$23,$F1,$95,$C9,$B1,$B7,$FD,$D3,$4F,$90,$C6
	.dc.b	$5F,$C6,$3C,$1F,$EA,$E4,$AF,$EA,$EF,$F6,$1F,$9E,$F7,$CF,$D0,$DD
	.dc.b	$18,$8F,$8A,$D5,$1A,$79,$7C,$64,$BF,$ED,$54,$8C,$4F,$F9,$16,$BA
	.dc.b	$FC,$77,$84,$97,$C1,$FE,$E9,$C3,$EF,$AE,$0D,$7C,$CD,$7E,$CB,$5F
	.dc.b	$2E,$1F,$ED,$2A,$F1,$FC,$17,$BC,$8F,$9B,$C3,$F8,$44,$70,$7A,$BD
	.dc.b	$9F,$8C,$47,$07,$BB,$C2,$F9,$44,$70,$7C,$2F,$0B,$E7,$11,$C1,$FE
	.dc.b	$0F,$C9,$F3,$DF,$00,$91,$97,$EB,$70,$48,$F9,$AD,$D3,$E6,$69,$FB
	.dc.b	$C9,$FA,$3D,$FF,$74,$8F,$F8,$77,$D4,$FF,$32,$B4,$FF,$11,$38,$AF
	.dc.b	$AC,$D7,$F0,$8B,$E7,$76,$49,$E8,$FF,$8A,$FE,$57,$FB,$0D,$7D,$F6
	.dc.b	$BF,$8D,$6B,$FB,$06,$BF,$8C,$F9,$6F,$03,$FB,$46,$B9,$AF,$F1,$C4
	.dc.b	$63,$8C,$FA,$7A,$BF,$CD,$F1,$89,$FF,$20,$4F,$A5,$FE,$01,$5C,$DB
	.dc.b	$D2,$F9,$9F,$E6,$D1,$8F,$90,$FF,$03,$BA,$8C,$D7,$FB,$BE,$95,$A8
	.dc.b	$C4,$E9,$FE,$D6,$D7,$FB,$5C,$F7,$F2,$3F,$90,$46,$27,$4F,$F8,$05
	.dc.b	$EF,$97,$FA,$2A,$B8,$61,$E3,$E4,$FF,$30,$C6,$2F,$FB,$FF,$46,$3D
	.dc.b	$97,$B3,$FD,$E2,$30,$DC,$3F,$86,$C7,$F5,$D2,$3F,$E6,$3F,$9C,$7B
	.dc.b	$7F,$E6,$76,$C6,$63,$04,$7E,$6F,$02,$77,$8F,$F9,$19,$F1,$CE,$9C
	.dc.b	$E9,$F0,$A5,$F0,$3F,$E1,$0F,$27,$94,$7F,$59,$EA,$FF,$55,$C6,$23
	.dc.b	$FA,$3C,$EE,$FF,$A4,$DF,$F3,$BC,$31,$5F,$ED,$5A,$FF,$0D,$FC,$EF
	.dc.b	$1A,$D4,$A3,$A3,$FE,$20,$9D,$DF,$4C,$7F,$D7,$75,$AA,$F3,$9C,$67
	.dc.b	$58,$70,$8F,$01,$93,$D4,$E5,$F4,$99,$18,$79,$2B,$B6,$1B,$8C,$5F
	.dc.b	$FB,$9A,$3D,$B5,$C3,$0F,$8B,$39,$8F,$F5,$43,$E4,$69,$E4,$BF,$85
	.dc.b	$3E,$C9,$E1,$18,$C3,$CD,$5F,$0C,$F3,$61,$E2,$79,$57,$5A,$F3,$C7
	.dc.b	$F8,$8C,$F8,$A7,$F8,$97,$0C,$38,$3F,$BF,$FC,$7B,$DD,$7B,$9F,$E2
	.dc.b	$B8,$C3,$63,$33,$FE,$06,$3F,$CA,$B3,$A7,$01,$7C,$8F,$CC,$8C,$C6
	.dc.b	$32,$ED,$39,$9E,$93,$97,$71,$C7,$1F,$C8,$65,$FD,$76,$7E,$D3,$5F
	.dc.b	$B5,$5C,$67,$F2,$CD,$9F,$F0,$E9,$B8,$E7,$E2,$7F,$88,$1D,$DF,$79
	.dc.b	$F8,$AD,$BF,$F3,$9C,$71,$39,$D7,$17,$78,$7E,$37,$0F,$05,$70,$AD
	.dc.b	$34,$D5,$F5,$7E,$43,$6F,$B4,$6C,$F6,$C6,$EB,$AC,$7E,$FE,$BA,$B4
	.dc.b	$EB,$71,$FC,$F5,$F4,$74,$74,$8F,$D4,$6E,$F9,$D7,$AD,$FB,$CD,$FF
	.dc.b	$C6,$F0,$F4,$F1,$87,$F8,$52,$30,$FF,$12,$78,$B6,$B7,$7D,$6A,$D7
	.dc.b	$0A,$FF,$6D,$BC,$1F,$F5,$9F,$CE,$FF,$BD,$DF,$E8,$FC,$F5,$FE,$84
	.dc.b	$DC,$BF,$CB,$53,$72,$FF,$2D,$B6,$BF,$D7,$95,$7A,$3F,$DE,$BF,$05
	.dc.b	$EA,$95,$67,$AF,$F3,$D8,$FF,$41,$ED,$FC,$DD,$7F,$55,$E2,$F8,$38
	.dc.b	$FF,$59,$8B,$7C,$33,$FE,$9F,$7C,$27,$FA,$C3,$70,$FE,$99,$5B,$9F
	.dc.b	$EA,$07,$49,$D5,$72,$C3,$AC,$FC,$AF,$9C,$7C,$ED,$24,$7F,$42,$8F
	.dc.b	$F0,$EB,$63,$E8,$F8,$0F,$97,$A7,$09,$FE,$43,$87,$F8,$06,$3B,$78
	.dc.b	$F2,$7F,$A6,$D7,$17,$FA,$7F,$86,$F4,$FE,$EF,$CE,$BE,$2E,$9F,$80
	.dc.b	$9D,$57,$F5,$69,$D3,$7E,$2B,$89,$C1,$D7,$5F,$E2,$3C,$97,$CC,$D3
	.dc.b	$BC,$78,$3C,$6D,$6D,$2F,$B4,$7E,$43,$BF,$FC,$55,$1E,$AF,$91,$F8
	.dc.b	$66,$7C,$BF,$4F,$A5,$DF,$9E,$57,$9E,$3D,$BE,$97,$8F,$E5,$E4,$FA
	.dc.b	$E3,$E0,$FA,$7E,$7F,$97,$93,$EF,$8F,$8B,$E7,$FA,$FE,$52,$4F,$C3
	.dc.b	$8F,$93,$EA,$FB,$FE,$56,$4F,$C7,$8F,$57,$1F,$DA,$F7,$A4,$FF,$80
	.dc.b	$38,$7E,$2B,$F3,$7B,$3E,$47,$14,$8C,$1F,$47,$D5,$F9,$9F,$3F,$23
	.dc.b	$07,$F8,$2B,$C8,$FF,$62,$78,$BF,$E3,$CE,$A7,$CB,$7D,$33,$F7,$73
	.dc.b	$AF,$0B,$FD,$76,$BE,$7B,$EC,$7C,$F4,$C7,$A4,$7E,$BF,$E6,$DF,$33
	.dc.b	$FB,$1F,$8D,$FF,$02,$63,$FE,$D2,$6D,$85,$70,$7F,$C9,$66,$2B,$97
	.dc.b	$A7,$23,$33,$B7,$F3,$A7,$37,$1A,$FF,$B8,$6F,$BC,$9B,$3A,$C3,$DB
	.dc.b	$F3,$1F,$00,$95,$F1,$FE,$5A,$BD,$BD,$46,$CA,$CE,$5F,$E6,$68,$5F
	.dc.b	$E5,$76,$FC,$B7,$CC,$CC,$7E,$E4,$65,$FB,$D1,$97,$B7,$32,$7F,$99
	.dc.b	$E0,$F9,$D7,$91,$85,$E3,$53,$11,$30,$FF,$3A,$D7,$19,$D4,$62,$B9
	.dc.b	$B6,$B2,$FC,$CA,$E0,$E4,$E1,$E5,$F8,$48,$F3,$F2,$6B,$C6,$FF,$81
	.dc.b	$B6,$7D,$13,$F8,$ED,$7F,$1F,$DB,$2B,$F2,$D1,$97,$F3,$11,$97,$E9
	.dc.b	$47,$D2,$7F,$07,$D7,$F8,$FF,$85,$91,$82,$1F,$86,$3F,$04,$E9,$C7
	.dc.b	$22,$7F,$34,$67,$3F,$C5,$56,$2B,$F5,$5E,$57,$E9,$BC,$99,$18,$21
	.dc.b	$E6,$72,$FA,$23,$ED,$FC,$EF,$FA,$2A,$3E,$AB,$5F,$A0,$DB,$73,$FE
	.dc.b	$9D,$47,$F4,$A9,$FF,$2E,$BB,$9D,$36,$33,$F0,$3E,$0F,$4F,$F6,$DE
	.dc.b	$5F,$E3,$3E,$49,$E3,$D3,$37,$FA,$E9,$C3,$F8,$A9,$DB,$92,$1F,$87
	.dc.b	$72,$D9,$E6,$FE,$73,$BF,$F9,$C1,$1F,$E3,$DE,$35,$74,$7F,$21,$8F
	.dc.b	$E7,$D1,$FC,$52,$E4,$FF,$1B,$7F,$8B,$F9,$78,$8F,$C2,$C5,$D7,$F6
	.dc.b	$93,$FE,$03,$F9,$D7,$9B,$FE,$1E,$F7,$DF,$3A,$F4,$30,$DC,$BF,$89
	.dc.b	$DF,$06,$DE,$AA,$46,$0F,$E5,$DE,$7F,$F5,$DF,$0A,$46,$08,$7F,$0B
	.dc.b	$F2,$91,$A8,$FC,$EC,$FF,$32,$E9,$FE,$CD,$F3,$B2,$7F,$9D,$46,$5F
	.dc.b	$F6,$2D,$A7,$51,$FA,$69,$FE,$77,$E5,$FF,$3C,$E2,$93,$FD,$0E,$32
	.dc.b	$FF,$45,$8C,$BF,$AE,$8F,$E8,$78,$FF,$77,$99,$3F,$D2,$E3,$2F,$8A
	.dc.b	$32,$FE,$DA,$3F,$A5,$F1,$7F,$7F,$F1,$E4,$FF,$53,$8C,$BE,$A8,$CB
	.dc.b	$FB,$E8,$FF,$1D,$7C,$EF,$F2,$FF,$8F,$FF,$65,$7C,$97,$F4,$9A,$4E
	.dc.b	$5F,$F6,$FC,$66,$31,$F0,$A8,$C6,$3F,$ED,$E7,$58,$B0,$EB,$87,$F5
	.dc.b	$5C,$38,$D6,$A7,$33,$E0,$BE,$4E,$EF,$93,$E2,$31,$3C,$2F,$9C,$75
	.dc.b	$AE,$2F,$AA,$B8,$E1,$E7,$8E,$27,$33,$57,$CE,$B8,$DF,$49,$C9,$B7
	.dc.b	$07,$E3,$9F,$15,$F6,$C3,$D9,$1B,$8C,$3E,$53,$77,$1D,$3E,$17,$18
	.dc.b	$79,$DF,$0B,$6F,$C3,$87,$9A,$FA,$D6,$B0,$F0,$DE,$EB,$58,$74,$9C
	.dc.b	$9F,$62,$B0,$7C,$4A,$ED,$7D,$30,$F7,$4F,$98,$93,$AB,$F4,$46,$4F
	.dc.b	$F7,$04,$FE,$25,$B5,$F6,$9B,$3F,$EE,$27,$F1,$FA,$FF,$5C,$7F,$1C
	.dc.b	$D7,$F1,$0D,$77,$E2,$F1,$BF,$DB,$3F,$83,$9F,$F3,$3E,$1F,$27,$C7
	.dc.b	$7A,$BF,$D0,$30,$FB,$55,$AC,$FB,$E1,$C4,$FC,$9D,$F0,$F0,$31,$7C
	.dc.b	$26,$E3,$D7,$1E,$8F,$D5,$C7,$A5,$7A,$2B,$78,$FE,$B7,$1E,$93,$CC
	.dc.b	$97,$C6,$BC,$38,$74,$BF,$2D,$69,$B5,$E1,$C7,$AC,$F7,$34,$7F,$90
	.dc.b	$F0,$EF,$3D,$63,$27,$CA,$3F,$45,$1A,$AE,$A4,$8E,$93,$DD,$E3,$97
	.dc.b	$D0,$79,$D7,$17,$75,$C5,$ED,$5C,$4F,$11,$23,$53,$C6,$77,$1F,$E5
	.dc.b	$BF,$35,$F3,$0F,$88,$FB,$5D,$BF,$E6,$FF,$3E,$E9,$BA,$D3,$FC,$EA
	.dc.b	$33,$8B,$0E,$78,$61,$73,$FF,$93,$AF,$FC,$08,$73,$9D,$46,$6B,$FC
	.dc.b	$B1,$87,$F2,$1A,$FE,$8F,$C9,$CE,$31,$3F,$D4,$E3,$19,$F2,$6D,$70
	.dc.b	$3B,$CF,$96,$9F,$13,$CF,$E0,$FA,$36,$B7,$9F,$C4,$8C,$5F,$91,$BA
	.dc.b	$FE,$7E,$7E,$4F,$8B,$EA,$F2,$FC,$CD,$C3,$9E,$48,$C6,$3E,$E6,$E1
	.dc.b	$CE,$5F,$11,$F2,$7B,$BF,$9F,$61,$CA,$31,$D1,$F0,$DB,$19,$3D,$33
	.dc.b	$C3,$1F,$F2,$57,$CD,$37,$FA,$FE,$1B,$E7,$1F,$39,$F6,$AD,$7B,$C7
	.dc.b	$E4,$DE,$EB,$53,$9A,$FF,$4D,$CE,$6B,$FD,$47,$39,$D7,$F8,$24,$E6
	.dc.b	$BE,$EF,$CE,$7D,$B3,$EE,$7C,$E5,$C3,$E3,$C5,$9F,$C1,$CD,$BF,$A1
	.dc.b	$EC,$3E,$57,$F1,$CD,$C3,$EC,$AF,$93,$CC,$C9,$CB,$C2,$31,$87,$F9
	.dc.b	$BF,$E2,$9F,$2B,$F9,$CC,$5E,$87,$F8,$83,$1F,$F5,$FD,$69,$FF,$27
	.dc.b	$78,$DF,$26,$31,$87,$F9,$36,$31,$87,$3B,$E5,$3D,$E3,$13,$AB,$E6
	.dc.b	$7F,$A8,$BE,$1D,$F3,$9E,$36,$C6,$7C,$DC,$4C,$8E,$0E,$E3,$18,$7E
	.dc.b	$B3,$E4,$3F,$55,$F3,$9F,$AB,$6D,$FE,$59,$B3,$B8,$FC,$CC,$FB,$7E
	.dc.b	$1B,$E2,$7C,$E3,$E2,$31,$5F,$EB,$88,$C3,$F1,$A7,$73,$AB,$FF,$5B
	.dc.b	$56,$9F,$A1,$3A,$AE,$07,$3A,$FE,$01,$F2,$5F,$C0,$7E,$72,$CF,$EB
	.dc.b	$1B,$7F,$C2,$1B,$C5,$FA,$B7,$ED,$7C,$4B,$E7,$29,$8F,$22,$E1,$FE
	.dc.b	$D0,$9D,$36,$74,$D3,$FD,$AB,$F3,$9F,$9D,$7F,$D3,$D3,$FC,$93,$1F
	.dc.b	$F6,$94,$E5,$62,$9D,$3F,$CB,$A7,$4E,$7C,$BF,$C8,$46,$27,$77,$FE
	.dc.b	$AB,$8C,$4E,$BC,$1F,$6B,$FA,$2E,$4F,$E6,$1F,$2E,$CB,$18,$7F,$1C
	.dc.b	$8C,$3F,$A6,$9D,$CE,$8F,$3C,$7F,$45,$9D,$C7,$F3,$4F,$9C,$FE,$64
	.dc.b	$7F,$33,$EC,$B5,$FC,$7A,$2C,$FF,$3A,$9B,$A7,$F8,$16,$1D,$6F,$83
	.dc.b	$FD,$77,$CE,$7E,$A2,$B5,$39,$C7,$F9,$CC,$27,$51,$FE,$6B,$9D,$5F
	.dc.b	$FC,$01,$9C,$BF,$06,$33,$5F,$D7,$27,$33,$E2,$BF,$0F,$67,$EE,$5B
	.dc.b	$87,$F3,$8E,$EF,$E6,$5B,$7F,$51,$6C,$7F,$46,$F9,$BF,$8B,$97,$F4
	.dc.b	$3F,$1F,$F1,$39,$7C,$A3,$FB,$24,$EA,$BF,$B8,$4E,$AB,$FA,$54,$BE
	.dc.b	$23,$FA,$07,$37,$F6,$98,$C5,$FF,$3C,$8C,$7C,$57,$F4,$5F,$46,$CF
	.dc.b	$F4,$07,$F0,$3A,$FF,$71,$F9,$37,$26,$2B,$FA,$EC,$61,$F5,$69,$FD
	.dc.b	$3F,$D2,$F0,$36,$3D,$DE,$17,$F5,$3B,$F2,$36,$3E,$17,$B1,$FD,$5F
	.dc.b	$D1,$F4,$36,$3E,$37,$81,$FD,$6F,$E1,$7D,$8D,$8F,$95,$E1,$7F,$5F
	.dc.b	$F4,$3E,$03,$63,$F3,$1F,$39,$EB,$F3,$3F,$3A,$D8,$FA,$5F,$25,$FD
	.dc.b	$9F,$E7,$7E,$43,$63,$EB,$78,$9F,$DA,$FC,$FF,$98,$D8,$FF,$69,$9F
	.dc.b	$3F,$E7,$FF,$C2,$6D,$3F,$DA,$FF,$2B,$86,$9C,$0F,$97,$E1,$FE,$C2
	.dc.b	$7E,$83,$63,$FC,$17,$F2,$E7,$26,$0B,$DF,$F8,$7F,$69,$3B,$9D,$1C
	.dc.b	$B0,$FF,$16,$46,$6B,$F1,$DE,$2B,$E6,$9F,$3D,$FE,$34,$87,$C8,$36
	.dc.b	$C5,$6B,$2F,$F1,$2D,$6A,$73,$7C,$6F,$85,$6A,$F8,$61,$E0,$9C,$BE
	.dc.b	$28,$CF,$89,$FE,$83,$C7,$A5,$F1,$C3,$A6,$3F,$D9,$FC,$1F,$37,$BB
	.dc.b	$FE,$25,$AF,$ED,$B1,$8E,$BD,$47,$9A,$75,$9F,$F7,$7A,$D4,$FF,$8E
	.dc.b	$3E,$5B,$FC,$7A,$DE,$EF,$F2,$94,$62,$BE,$73,$FA,$16,$F8,$1A,$9D
	.dc.b	$B6,$BF,$C9,$33,$FE,$F9,$F8,$2F,$E5,$9F,$3C,$C3,$FD,$F9,$BF,$38
	.dc.b	$C5,$7F,$BF,$C5,$70,$AD,$E3,$FE,$A7,$AD,$D7,$50,$E2,$FA,$57,$39
	.dc.b	$E6,$69,$E7,$87,$58,$F0,$61,$5F,$EA,$38,$77,$AE,$53,$9B,$EB,$5C
	.dc.b	$0E,$87,$F4,$97,$2F,$97,$0F,$25,$F3,$8E,$11,$8C,$3C,$11,$FD,$46
	.dc.b	$FB,$56,$E3,$96,$1E,$87,$B9,$83,$C1,$1B,$79,$1A,$C3,$C6,$E3,$0F
	.dc.b	$03,$C8,$D9,$FE,$E1,$BF,$F7,$13,$83,$E0,$9F,$2F,$E5,$FF,$CD,$FE
	.dc.b	$E7,$FB,$9D,$B1,$9E,$F4,$E5,$B1,$C7,$C1,$FE,$FA,$DC,$79,$46,$23
	.dc.b	$F0,$56,$8C,$47,$F7,$51,$19,$9F,$B1,$9F,$F4,$39,$BE,$56,$F3,$FC
	.dc.b	$3D,$EE,$7E,$4B,$63,$F1,$75,$C6,$7E,$6B,$63,$8E,$1E,$86,$E1,$DA
	.dc.b	$A1,$D2,$75,$5E,$7B,$EC,$FA,$4E,$85,$C3,$CC,$EE,$78,$D7,$F8,$4E
	.dc.b	$B8,$BF,$E1,$59,$E2,$DA,$E2,$EE,$B8,$BF,$19,$B3,$C2,$3F,$65,$3C
	.dc.b	$1D,$5F,$E8,$6F,$BC,$68,$F6,$E1,$F0,$1F,$CF,$F4,$BE,$79,$EA,$FE
	.dc.b	$F3,$83,$FD,$CD,$18,$AF,$F6,$2B,$5F,$E7,$0B,$D1,$E1,$1F,$EE,$38
	.dc.b	$D6,$5F,$EC,$77,$FD,$31,$1C,$BB,$77,$8F,$87,$1B,$F3,$BF,$E6,$DA
	.dc.b	$FF,$69,$B5,$FE,$8C,$DC,$3F,$DA,$4F,$FA,$EA,$37,$E3,$72,$23,$CB
	.dc.b	$E0,$3A,$F8,$BD,$BE,$76,$4F,$68,$AF,$68,$B8,$76,$9B,$B7,$F9,$37
	.dc.b	$36,$4B,$CE,$09,$CD,$FB,$BE,$8F,$FC,$27,$3F,$EE,$E6,$BF,$D8,$5B
	.dc.b	$8F,$FB,$B5,$FF,$74,$46,$F9,$39,$11,$E5,$D1,$5E,$5E,$3F,$73,$17
	.dc.b	$FE,$35,$F2,$3E,$7F,$98,$F7,$69,$88,$CF,$BB,$30,$FF,$75,$BA,$7F
	.dc.b	$DD,$BE,$E7,$2E,$49,$18,$3F,$DE,$3C,$FD,$B9,$A4,$60,$FF,$7A,$FA
	.dc.b	$BE,$22,$3F,$EF,$8F,$67,$F0,$7F,$8F,$7F,$49,$DB,$FC,$11,$3F,$2E
	.dc.b	$1F,$17,$A3,$71,$C3,$BB,$8F,$C5,$48,$FE,$75,$D5,$F5,$EF,$F8,$26
	.dc.b	$1D,$3A,$10,$F1,$7C,$24,$BE,$27,$87,$E8,$7C,$BE,$17,$9B,$9A,$4E
	.dc.b	$B0,$E7,$3E,$AF,$82,$F5,$FA,$D2,$B5,$97,$FB,$71,$BE,$C7,$C0,$23
	.dc.b	$AF,$5F,$E1,$F9,$DF,$13,$C4,$93,$CA,$3E,$3F,$89,$F2,$3D,$79,$3D
	.dc.b	$A3,$E5,$F8,$72,$3C,$2F,$CE,$F1,$3E,$7F,$AF,$F4,$3D,$D9,$5E,$68
	.dc.b	$FD,$DF,$2A,$DE,$DF,$EF,$A9,$F3,$4F,$9E,$C3,$63,$FD,$33,$5C,$1F
	.dc.b	$F0,$5C,$65,$FF,$43,$1C,$A7,$89,$E0,$9E,$11,$A9,$E1,$7F,$EA,$B6
	.dc.b	$F4,$FF,$4B,$6C,$FF,$B4,$2B,$8D,$7D,$FE,$2F,$F6,$94,$FD,$2F,$6F
	.dc.b	$F1,$E7,$E4,$3D,$BB,$19,$21,$FE,$55,$94,$7F,$18,$FC,$7F,$FD,$79
	.dc.b	$7F,$8B,$5E,$E6,$4D,$67,$70,$FC,$34,$E9,$B3,$A7,$FC,$5F,$F2,$4F
	.dc.b	$1F,$E7,$3F,$C0,$A3,$E3,$B7,$0F,$F6,$63,$B8,$E3,$3C,$3D,$89,$18
	.dc.b	$21,$C9,$CB,$D2,$3D,$9B,$C8,$F3,$BE,$FF,$91,$C4,$7F,$9D,$F2,$7C
	.dc.b	$97,$C3,$33,$97,$C3,$F9,$1F,$8C,$4A,$F8,$3E,$DD,$79,$45,$9F,$95
	.dc.b	$3F,$8D,$F8,$CF,$F3,$DD,$7F,$94,$BE,$A3,$FD,$D0,$DE,$FD,$76,$FE
	.dc.b	$4E,$73,$3F,$8C,$6E,$1F,$EF,$16,$F1,$7F,$84,$70,$EB,$5B,$C3,$91
	.dc.b	$1F,$F3,$0C,$6A,$FE,$F7,$CF,$7E,$07,$E7,$A5,$7D,$C8,$AF,$C5,$8B
	.dc.b	$3C,$23,$FD,$F7,$F1,$5F,$21,$BB,$7E,$BA,$31,$1F,$EE,$99,$43,$5D
	.dc.b	$56,$F6,$97,$C8,$FF,$3B,$BF,$DC,$34,$FF,$17,$2F,$87,$FD,$C3,$67
	.dc.b	$7E,$96,$26,$47,$27,$71,$88,$FD,$BC,$F1,$6C,$E9,$FF,$55,$9F,$DD
	.dc.b	$68,$BE,$0E,$F8,$9F,$E0,$66,$C7,$0E,$2C,$3C,$4E,$1F,$24,$F9,$BE
	.dc.b	$0D,$8E,$F7,$FB,$3F,$87,$91,$FA,$E7,$F6,$DF,$39,$88,$FC,$C5,$FE
	.dc.b	$E4,$C7,$0F,$EE,$7E,$6B,$F7,$84,$9F,$DB,$FC,$ED,$7F,$41,$16,$7F
	.dc.b	$80,$4F,$F0,$3F,$9B,$95,$FB,$E7,$F8,$37,$D0,$E2,$3F,$51,$F3,$9F
	.dc.b	$C2,$4C,$5F,$F0,$9F,$A1,$FE,$18,$49,$FE,$0F,$F2,$F5,$FD,$84,$59
	.dc.b	$FE,$21,$3F,$C4,$FC,$39,$5F,$C3,$9F,$E2,$DD,$38,$8F,$DC,$45,$AF
	.dc.b	$E3,$13,$FE,$32,$FC,$81,$C5,$49,$7C,$CF,$EB,$D1,$A6,$FC,$A7,$13
	.dc.b	$6F,$F3,$CF,$C1,$5F,$1B,$FD,$1D,$B1,$FE,$CC,$6B,$BF,$86,$D4,$C8
	.dc.b	$D3,$C6,$3F,$A8,$7C,$FF,$F1,$0C,$7F,$D5,$DE,$87,$FA,$2E,$31,$7F
	.dc.b	$ED,$08,$C7,$7F,$9F,$E4,$3F,$A5,$D3,$83,$FD,$B9,$5A,$C3,$9E,$1F
	.dc.b	$98,$AE,$77,$C3,$0E,$57,$DF,$0E,$37,$B9,$CB,$70,$E6,$B0,$E7,$7C
	.dc.b	$63,$31,$87,$81,$E9,$2E,$1E,$8C,$3B,$C7,$3C,$3B,$B7,$0E,$71,$8B
	.dc.b	$F7,$4E,$26,$3D,$4E,$05,$8D,$B6,$79,$12,$7C,$F5,$E2,$AF,$45,$72
	.dc.b	$BF,$8D,$5A,$8E,$F1,$87,$A4,$65,$A4,$8D,$1E,$F7,$8E,$1F,$55,$D3
	.dc.b	$F2,$5A,$7C,$67,$51,$C6,$3A,$9F,$EC,$F3,$DE,$5A,$FC,$3B,$70,$FC
	.dc.b	$E3,$9F,$23,$FC,$9E,$F1,$3E,$D3,$87,$5F,$03,$97,$2D,$A4,$F2,$BF
	.dc.b	$C8,$B7,$0F,$F7,$17,$C0,$3D,$9F,$F2,$FD,$6B,$0F,$C2,$DE,$F1,$FC
	.dc.b	$A3,$55,$F1,$F0,$77,$52,$7E,$E5,$6E,$B5,$18,$9D,$5F,$78,$FC,$24
	.dc.b	$EA,$30,$4C,$7F,$5D,$7C,$F1,$F3,$C6,$30,$E6,$DC,$39,$CE,$9F,$4C
	.dc.b	$E9,$A6,$B1,$C3,$F2,$8F,$31,$92,$E1,$E3,$78,$BF,$E4,$6C,$7D,$18
	.dc.b	$78,$63,$53,$CD,$D6,$1E,$38,$C4,$EC,$8F,$F5,$8A,$E6,$F6,$8C,$BC
	.dc.b	$0D,$4F,$3C,$BE,$FE,$3F,$C0,$3F,$06,$7C,$1F,$FA,$6B,$9B,$FD,$37
	.dc.b	$82,$73,$A7,$FA,$3A,$31,$F0,$9B,$8F,$ED,$31,$BB,$C3,$FD,$B2,$B8
	.dc.b	$BF,$E6,$A6,$DF,$65,$7C,$27,$95,$FF,$3A,$F9,$AE,$7F,$02,$E8,$FF
	.dc.b	$3E,$CE,$4F,$E7,$FB,$6A,$15,$4C,$78,$F6,$BF,$A2,$B8,$7F,$A3,$FC
	.dc.b	$9D,$C5,$A7,$FA,$C7,$0F,$F0,$A5,$70,$36,$59,$DB,$FE,$EF,$87,$C6
	.dc.b	$78,$CF,$27,$AB,$67,$FA,$AC,$5A,$E2,$7F,$BC,$E9,$59,$BF,$17,$C6
	.dc.b	$FB,$7A,$FD,$CD,$C7,$FA,$FB,$87,$FB,$0F,$5D,$C7,$AE,$FF,$27,$E5
	.dc.b	$3E,$31,$F1,$FE,$36,$DF,$F6,$88,$B5,$C3,$45,$63,$3F,$89,$F1,$DF
	.dc.b	$3F,$E3,$BE,$83,$71,$FE,$E2,$E1,$FE,$E7,$ED,$5C,$7D,$FE,$87,$CE
	.dc.b	$E4,$FC,$3A,$D5,$C7,$8C,$FE,$DE,$3F,$ED,$EA,$D4,$E6,$E4,$E6,$9E
	.dc.b	$1E,$31,$FF,$5C,$FE,$17,$FE,$E9,$AF,$CD,$36,$3F,$D7,$B9,$34,$FC
	.dc.b	$94,$E7,$0F,$D1,$B6,$FF,$3E,$DB,$FF,$24,$DF,$19,$EB,$F1,$9F,$9B
	.dc.b	$8C,$4F,$8A,$31,$87,$67,$F4,$F8,$3F,$7D,$7C,$7E,$1F,$D3,$E3,$FE
	.dc.b	$56,$9D,$4E,$7B,$74,$C9,$D1,$BA,$FD,$6B,$8C,$7F,$3B,$1D,$A3,$FC
	.dc.b	$4F,$EB,$7E,$A4,$FD,$57,$C8,$58,$FF,$15,$B6,$BF,$08,$DC,$7F,$83
	.dc.b	$F8,$EF,$2F,$F7,$4D,$BF,$E1,$1F,$52,$F4,$B3,$52,$7F,$86,$56,$F1
	.dc.b	$FF,$77,$DE,$E7,$35,$FC,$16,$31,$3F,$EF,$4A,$E1,$87,$F1,$89,$CB
	.dc.b	$E4,$AF,$E0,$9A,$4C,$3F,$C8,$BE,$15,$C7,$F3,$D3,$6F,$F8,$04,$DA
	.dc.b	$FD,$E6,$DF,$C4,$BB,$FF,$D0,$F1,$FE,$4E,$8C,$61,$FC,$45,$91,$F3
	.dc.b	$F8,$BF,$77,$E2,$62,$BF,$7E,$3F,$4F,$7A,$DF,$C8,$5F,$F4,$55,$FF
	.dc.b	$2F,$72,$DF,$57,$53,$B6,$C6,$4E,$11,$96,$99,$AF,$E0,$79,$3E,$4B
	.dc.b	$D8,$7B,$7C,$BB,$5E,$99,$FE,$F2,$D9,$D3,$4D,$73,$FC,$53,$E3,$7C
	.dc.b	$95,$8C,$93,$0F,$0C,$65,$FB,$5E,$DF,$E2,$0F,$D1,$FC,$EB,$E2,$31
	.dc.b	$7F,$86,$9B,$5F,$ED,$1A,$D4,$E4,$FC,$F7,$1A,$33,$3D,$A7,$31,$89
	.dc.b	$FC,$CD,$7D,$DF,$92,$FB,$DF,$27,$6F,$E1,$3F,$81,$F9,$39,$1F,$17
	.dc.b	$E4,$EE,$5F,$33,$E1,$BF,$0D,$EA,$7E,$1D,$B5,$F2,$FE,$0B,$F1,$5F
	.dc.b	$05,$F8,$B6,$C7,$CF,$F9,$2F,$C7,$1F,$8F,$F9,$3B,$1F,$4F,$E3,$FF
	.dc.b	$25,$F2,$7F,$93,$6C,$7D,$7F,$92,$FC,$B7,$C9,$FE,$5D,$B1,$F6,$F9
	.dc.b	$7F,$34,$7E,$6F,$E4,$6C,$7E,$5F,$F2,$33,$FD,$C1,$7C,$95,$8C,$39
	.dc.b	$F9,$2D,$7C,$15,$9E,$4F,$1F,$83,$91,$D7,$A6,$F2,$FF,$54,$C5,$F1
	.dc.b	$BD,$FE,$3F,$C0,$DB,$F2,$7C,$7F,$8F,$C4,$F2,$36,$3D,$1F,$1F,$E7
	.dc.b	$3D,$1E,$35,$8F,$67,$A1,$86,$9E,$F8,$F7,$7C,$7F,$BF,$E3,$EC,$FC
	.dc.b	$36,$EB,$E7,$9F,$89,$F1,$F8,$F1,$3E,$3C,$FF,$59,$DD,$3A,$C3,$E6
	.dc.b	$7C,$7D,$AF,$EE,$7F,$39,$F3,$A7,$F8,$33,$63,$3C,$DF,$E2,$08,$C6
	.dc.b	$3D,$BC,$F4,$62,$FF,$A8,$B6,$FF,$B5,$3F,$62,$3A,$9C,$EF,$A3,$4C
	.dc.b	$9E,$2E,$6F,$01,$E1,$E6,$B1,$92,$5F,$AA,$32,$FF,$83,$BC,$7F,$F1
	.dc.b	$AB,$7D,$CF,$F1,$CC,$63,$3F,$53,$FD,$F2,$78,$7C,$4E,$8D,$FC,$4D
	.dc.b	$C7,$FC,$A1,$21,$F1,$97,$E6,$58,$7F,$94,$B1,$FF,$18,$DE,$63,$FD
	.dc.b	$49,$56,$FF,$B5,$7C,$5F,$D8,$C3,$8C,$7F,$AC,$67,$4F,$F9,$0A,$75
	.dc.b	$5C,$A1,$5C,$1D,$5F,$07,$CD,$7C,$B0,$F1,$C7,$0B,$E5,$7C,$67,$2F
	.dc.b	$39,$E2,$6C,$F3,$DF,$57,$E0,$99,$3C,$25,$38,$C7,$5C,$3D,$31,$ED
	.dc.b	$38,$57,$82,$FC,$67,$C9,$9D,$CF,$F9,$F3,$0F,$54,$62,$3A,$CE,$F0
	.dc.b	$F2,$56,$9B,$3F,$EC,$F3,$CE,$7F,$B4,$63,$0E,$0F,$F4,$4F,$CD,$3E
	.dc.b	$31,$F3,$7F,$ED,$78,$C3,$63,$33,$FE,$DF,$9C,$B7,$E3,$7F,$CF,$2D
	.dc.b	$CF,$94,$62,$3F,$DD,$D2,$8C,$47,$89,$E3,$A7,$DD,$CF,$FB,$34,$FE
	.dc.b	$11,$AF,$D1,$AE,$57,$F8,$86,$C7,$D5,$9D,$D7,$3F,$57,$FD,$4E,$FE
	.dc.b	$26,$B6,$DC,$3D,$3F,$0B,$BC,$3F,$40,$D7,$FC,$BB,$87,$9F,$1C,$B6
	.dc.b	$3C,$18,$F6,$8F,$F5,$A9,$E2,$35,$87,$7A,$F6,$4E,$EB,$F5,$D5,$B7
	.dc.b	$F6,$18,$F9,$A3,$89,$D2,$78,$4F,$EC,$9B,$3B,$8E,$38,$FB,$2B,$84
	.dc.b	$7A,$0E,$F5,$FE,$CA,$C7,$C9,$1E,$E3,$FD,$09,$F3,$7F,$E6,$47,$FC
	.dc.b	$58,$D7,$FD,$3A,$DC,$7F,$93,$2F,$83,$FE,$09,$BF,$F9,$DA,$38,$57
	.dc.b	$FB,$63,$A1,$18,$8D,$7B,$FF,$ED,$C7,$51,$97,$FD,$2D,$F3,$6F,$4B
	.dc.b	$51,$FE,$82,$8C,$38,$B4,$62,$3A,$46,$27,$8C,$74,$7C,$EF,$6F,$42
	.dc.b	$CF,$26,$D7,$3C,$FF,$A2,$2F,$43,$24,$BF,$F2,$94,$3D,$0C,$47,$8A
	.dc.b	$38,$3E,$76,$B9,$9F,$2E,$1E,$A7,$C7,$DC,$F1,$E9,$D3,$1F,$EA,$DF
	.dc.b	$37,$FA,$DC,$3F,$CC,$4D,$7F,$DA,$6D,$8D,$F5,$7B,$A3,$53,$F5,$5A
	.dc.b	$FF,$AB,$9A,$FF,$B1,$DB,$89,$7C,$87,$A1,$B3,$F4,$DA,$FD,$26,$BF
	.dc.b	$83,$F6,$7F,$D8,$4F,$D9,$6B,$F6,$1A,$EF,$E2,$3E,$2C,$3E,$36,$92
	.dc.b	$79,$3A,$E5,$91,$D5,$AF,$5E,$4E,$64,$7C,$4F,$6F,$84,$91,$E5,$6B
	.dc.b	$F8,$CF,$9A,$7C,$27,$D8,$6C,$7E,$4D,$AE,$E3,$5F,$07,$23,$4D,$7A
	.dc.b	$7C,$17,$32,$3C,$DE,$D9,$C8,$EC,$D7,$C9,$EF,$F8,$C8,$F8,$DF,$37
	.dc.b	$BD,$23,$CC,$D7,$D9,$D5,$EB,$23,$EB,$7D,$DD,$52,$3D,$CD,$7E,$27
	.dc.b	$57,$C3,$23,$F0,$DF,$8D,$EE,$48,$F8,$CD,$7F,$7B,$F3,$6F,$6B,$F6
	.dc.b	$B0,$8C,$3F,$EE,$D3,$11,$FB,$AE,$9F,$D8,$79,$5C,$2B,$75,$D2,$BF
	.dc.b	$B3,$46,$23,$F8,$83,$5F,$F1,$01,$D0,$ED,$F5,$BF,$BB,$6E,$3D,$6B
	.dc.b	$F8,$3F,$A4,$F9,$9A,$4A,$FF,$6A,$72,$7F,$AB,$5F,$F4,$04,$65,$FF
	.dc.b	$3A,$3F,$C4,$23,$50,$CF,$FC,$A5,$F3,$9F,$E8,$48,$FF,$44,$7C,$C7
	.dc.b	$80,$FE,$2F,$19,$8E,$1E,$2F,$F8,$E7,$BD,$39,$BF,$F2,$E6,$4C,$39
	.dc.b	$61,$FE,$E9,$AD,$61,$F0,$BE,$2D,$F1,$3F,$49,$B5,$F4,$5A,$FF,$BB
	.dc.b	$97,$C9,$EE,$7F,$C2,$91,$A9,$FE,$7C,$D7,$B4,$68,$ED,$F2,$52,$3B
	.dc.b	$47,$BB,$D5,$91,$E7,$7E,$0F,$06,$23,$E0,$8D,$BE,$0C,$E7,$6F,$97
	.dc.b	$5F,$23,$E0,$BE,$CE,$53,$0F,$85,$19,$7F,$D6,$D1,$97,$AE,$1F,$32
	.dc.b	$7F,$2B,$F3,$8E,$AA,$4A,$FC,$B7,$CE,$3E,$27,$11,$FD,$6A,$B8,$3F
	.dc.b	$D5,$1A,$E6,$BF,$C7,$91,$8C,$FF,$D4,$32,$C1,$E1,$7E,$99,$B1,$FA
	.dc.b	$5A,$E1,$C3,$F8,$3E,$2F,$E9,$53,$FE,$D1,$F9,$CF,$CC,$9F,$8A,$F2
	.dc.b	$E4,$64,$87,$C0,$9E,$2F,$CE,$C9,$5E,$38,$FF,$44,$47,$FA,$D8,$CE
	.dc.b	$3F,$B5,$9D,$4F,$15,$C3,$99,$93,$CB,$1E,$75,$C2,$FA,$3C,$27,$F8
	.dc.b	$15,$3D,$9F,$84,$48,$CF,$8F,$5F,$F0,$4C,$ED,$FF,$01,$B5,$CC,$FF
	.dc.b	$9A,$E3,$17,$F4,$3E,$C9,$C7,$FC,$51,$B7,$E2,$BF,$F0,$5F,$CB,$BE
	.dc.b	$51,$F0,$98,$6D,$7B,$44,$EE,$78,$C7,$F8,$3F,$5F,$F5,$94,$E6,$38
	.dc.b	$35,$FC,$FB,$72,$FF,$72,$BF,$EB,$9F,$17,$F8,$2D,$CA,$FD,$F3,$FC
	.dc.b	$23,$E1,$F1,$1F,$76,$2D,$FF,$10,$9F,$E1,$BF,$0D,$FE,$14,$E2,$34
	.dc.b	$97,$FE,$31,$8C,$BE,$DE,$97,$E4,$9F,$0E,$30,$D7,$F1,$95,$FC,$6B
	.dc.b	$E1,$24,$7F,$13,$7F,$8E,$FC,$3E,$23,$FC,$77,$82,$7C,$84,$8F,$17
	.dc.b	$89,$E7,$23,$82,$1F,$BA,$C1,$ED,$7F,$90,$1B,$1F,$C8,$E7,$F9,$6F
	.dc.b	$27,$C1,$24,$7B,$DA,$FF,$04,$F9,$AF,$66,$CE,$2F,$E6,$51,$87,$F6
	.dc.b	$55,$FE,$4E,$8E,$F5,$C7,$1C,$19,$E3,$B1,$FE,$5D,$9D,$BF,$E6,$B6
	.dc.b	$BD,$A7,$FD,$85,$F3,$DF,$EB,$D8,$FF,$2F,$B5,$FF,$34,$B5,$FF,$38
	.dc.b	$35,$FE,$C2,$EF,$1F,$E1,$EF,$F9,$27,$1C,$FB,$FF,$DD,$23,$51,$FD
	.dc.b	$A3,$E7,$BF,$77,$97,$FB,$D3,$CD,$C7,$6F,$E0,$7E,$15,$18,$9D,$66
	.dc.b	$17,$FD,$55,$C9,$8B,$E5,$5C,$61,$7D,$1E,$55,$E2,$73,$3E,$5F,$27
	.dc.b	$FD,$DC,$F1,$AE,$13,$D7,$0E,$F7,$C6,$72,$F8,$E7,$99,$DB,$0F,$1B
	.dc.b	$6B,$8C,$62,$3D,$78,$78,$1D,$1C,$8F,$01,$6B,$C1,$87,$9E,$39,$DC
	.dc.b	$75,$7D,$6B,$C9,$87,$9A,$30,$4A,$F5,$E1,$E1,$BE,$F5,$B9,$E8,$FB
	.dc.b	$A7,$57,$E2,$AF,$3C,$ED,$F3,$4E,$AF,$B4,$79,$2F,$BB,$D3,$0F,$9C
	.dc.b	$47,$EA,$D7,$19,$D4,$76,$9F,$80,$F3,$3F,$D7,$F8,$7D,$B7,$FD,$8C
	.dc.b	$D7,$F0,$46,$3E,$1F,$FC,$DE,$FE,$4D,$AF,$E1,$1A,$FE,$39,$B0,$7F
	.dc.b	$CF,$38,$FE,$01,$B5,$FE,$44,$F5,$5F,$25,$FB,$DC,$3E,$C5,$69,$ED
	.dc.b	$EB,$FE,$2A,$31,$0A,$E3,$AE,$6A,$4F,$27,$F3,$F3,$A8,$C1,$31,$F2
	.dc.b	$4E,$B0,$F1,$46,$27,$58,$FC,$1A,$D6,$1D,$AB,$44,$9F,$65,$F2,$78
	.dc.b	$5F,$23,$D6,$7E,$5F,$0F,$25,$71,$8C,$B5,$F7,$57,$37,$35,$D2,$75
	.dc.b	$5C,$A3,$04,$9D,$9F,$C0,$E3,$4E,$0F,$B1,$87,$AE,$7D,$38,$59,$D6
	.dc.b	$5F,$6E,$BE,$0C,$7A,$C7,$CF,$3E,$63,$FC,$2D,$7F,$76,$69,$FC,$4E
	.dc.b	$75,$1F,$E8,$FC,$7F,$8B,$3F,$E9,$18,$CC,$FF,$A2,$0E,$15,$FE,$36
	.dc.b	$7F,$D1,$71,$A7,$4D,$7F,$A2,$E3,$B7,$FC,$B1,$C9,$FC,$9A,$FF,$A0
	.dc.b	$4E,$7B,$13,$9A,$F0,$46,$27,$D2,$DA,$FE,$E5,$4F,$8A,$CB,$C7,$97
	.dc.b	$F3,$55,$D3,$C2,$CD,$69,$B7,$C9,$F1,$5F,$F8,$BE,$3F,$AA,$46,$0C
	.dc.b	$FC,$07,$8B,$AF,$C1,$3E,$28,$C5,$F8,$9D,$C7,$F5,$D6,$E7,$EC,$C1
	.dc.b	$F1,$3E,$02,$57,$F6,$36,$CF,$F7,$17,$DB,$D3,$EE,$7D,$FF,$3D,$E3
	.dc.b	$DA,$DF,$F8,$49,$B7,$DE,$7D,$A4,$9F,$EE,$1E,$FF,$C8,$3E,$4F,$C0
	.dc.b	$58,$F9,$B1,$6B,$FC,$05,$37,$83,$FC,$67,$C0,$F5,$3F,$1B,$E4,$7F
	.dc.b	$A5,$5B,$3F,$1A,$FF,$C7,$75,$BC,$7F,$C7,$ED,$8C,$D7,$0E,$0E,$73
	.dc.b	$9B,$FF,$42,$FC,$EB,$E0,$3F,$37,$82,$3F,$D2,$D8,$FF,$BA,$5D,$37
	.dc.b	$D2,$DC,$62,$7F,$38,$71,$AF,$CB,$36,$B2,$D3,$73,$C6,$FC,$75,$E2
	.dc.b	$C1,$A7,$68,$FD,$7C,$BE,$77,$F5,$6F,$FA,$25,$B8,$FF,$A3,$17,$C1
	.dc.b	$70,$73,$97,$FB,$9D,$B5,$CF,$A7,$55,$B7,$F4,$3E,$87,$E5,$7A,$6E
	.dc.b	$3F,$E5,$46,$E3,$FE,$45,$6E,$1F,$A6,$E9,$FD,$04,$FF,$B3,$8F,$F3
	.dc.b	$F7,$8D,$ED,$F8,$4F,$9F,$E1,$FF,$A1,$B1,$FF,$3F,$57,$FB,$4E,$73
	.dc.b	$5F,$2A,$B5,$3F,$39,$AF,$F0,$B6,$BF,$E2,$53,$BD,$74,$8E,$38,$7F
	.dc.b	$B1,$3E,$73,$FC,$EB,$39,$BF,$DF,$EB,$FB,$3E,$5B,$3F,$E7,$26,$E3
	.dc.b	$FE,$63,$6E,$1E,$F7,$F8,$5F,$B5,$FC,$82,$7F,$DC,$2D,$9F,$E1,$3D
	.dc.b	$BF,$4B,$91,$39,$C3,$FC,$79,$18,$BE,$67,$F1,$2E,$1B,$3F,$1B,$E7
	.dc.b	$3F,$D4,$EF,$C8,$FD,$4F,$F1,$4F,$C1,$8C,$D7,$D1,$9C,$BF,$E4,$98
	.dc.b	$CB,$FC,$CD,$AF,$1F,$4B,$F3,$31,$88,$E1,$1F,$A6,$8E,$B9,$FF,$BB
	.dc.b	$97,$A5,$99,$F8,$7F,$19,$FC,$77,$C3,$B9,$7F,$A7,$1B,$8F,$FA,$35
	.dc.b	$B8,$FE,$25,$FE,$49,$AC,$AF,$F5,$33,$EA,$D7,$FD,$4E,$E6,$FF,$D2
	.dc.b	$CF,$D2,$E9,$C6,$1F,$E1,$19,$FF,$67,$7A,$DF,$BE,$E4,$F4,$E3,$F8
	.dc.b	$7E,$54,$E7,$5F,$CA,$BF,$91,$24,$FE,$EA,$76,$D8,$DB,$FE,$D7,$C1
	.dc.b	$F3,$8F,$99,$CE,$5F,$84,$F8,$9F,$EA,$DA,$9A,$4C,$3F,$85,$ED,$F8
	.dc.b	$6F,$9C,$D9,$23,$34,$8E,$7E,$74,$9C,$B7,$0F,$EC,$34,$F6,$7C,$24
	.dc.b	$AE,$6F,$93,$E7,$24,$76,$6D,$72,$F8,$EF,$49,$27,$C6,$FB,$3E,$2A
	.dc.b	$47,$99,$B3,$E2,$D7,$E0,$92,$7D,$6F,$F7,$9F,$86,$7C,$AD,$24,$FC
	.dc.b	$48,$CB,$7D,$5E,$04,$70,$6F,$D7,$E8,$47,$07,$3F,$9A,$F0,$11,$C1
	.dc.b	$DF,$B3,$C8,$47,$07,$8F,$E6,$7D,$04,$70,$79,$FA,$7D,$84,$70,$7A
	.dc.b	$FD,$FF,$80,$47,$07,$BF,$E6,$7E,$21,$1C,$13,$67,$4F,$F8,$5B,$55
	.dc.b	$7F,$87,$97,$CC,$D8,$FF,$10,$FC,$9F,$FB,$C7,$0F,$F6,$65,$FF,$87
	.dc.b	$BE,$87,$FA,$4C,$60,$FE,$2B,$3B,$C7,$BC,$7F,$18,$9D,$C6,$07,$52
	.dc.b	$B8,$47,$F3,$58,$C5,$7F,$8D,$E7,$33,$DA,$30,$79,$8F,$3F,$15,$7C
	.dc.b	$91,$DA,$7F,$D2,$91,$8A,$F0,$7C,$67,$BB,$8F,$DE,$D9,$FD,$6D,$70
	.dc.b	$BF,$F2,$2B,$5F,$33,$5F,$E6,$BF,$2F,$FE,$11,$7D,$4D,$73,$EA,$7F
	.dc.b	$94,$63,$11,$F0,$9F,$F0,$C6,$9F,$E4,$A6,$DF,$4F,$63,$FC,$A7,$18
	.dc.b	$AF,$F3,$37,$BF,$FE,$FF,$F8,$24,$FF,$5D,$C1,$E3,$FF,$AB,$67,$37
	.dc.b	$C0,$38,$7F,$0B,$1B,$AF,$F0,$BD,$A5,$F3,$3B,$C3,$AE,$59,$8E,$53
	.dc.b	$BA,$C5,$F8,$F0,$FF,$5F,$56,$A9,$5A,$7F,$1B,$3A,$AF,$0C,$78,$1E
	.dc.b	$57,$B8,$FE,$F5,$57,$0F,$1C,$6C,$F7,$61,$E7,$C3,$A4,$7C,$1C,$71
	.dc.b	$87,$C9,$C7,$BC,$62,$7E,$2C,$78,$AF,$07,$A7,$1F,$06,$1D,$2B,$C6
	.dc.b	$78,$4F,$97,$87,$9D,$F8,$86,$71,$F3,$DF,$38,$FF,$20,$9F,$10,$B7
	.dc.b	$87,$07,$FA,$97,$E6,$7F,$CF,$7F,$35,$FE,$48,$6C,$7F,$93,$9A,$EF
	.dc.b	$DE,$4E,$FD,$FF,$E0,$73,$B6,$BD,$5A,$FF,$89,$E9,$18,$8F,$C2,$46
	.dc.b	$31,$FA,$9A,$3C,$DF,$F1,$76,$7F,$8D,$8C,$5F,$D9,$BE,$0F,$C7,$6C
	.dc.b	$EE,$B8,$D7,$F8,$DC,$E3,$5B,$C3,$D3,$5F,$99,$BE,$7B,$71,$AF,$B8
	.dc.b	$7E,$66,$F8,$C6,$EF,$E0,$B7,$1F,$4E,$3E,$0B,$E0,$D5,$87,$9B,$0F
	.dc.b	$89,$87,$89,$F3,$56,$DF,$3E,$3D,$6F,$D1,$87,$66,$18,$24,$7A,$5F
	.dc.b	$C5,$4F,$23,$BC,$FE,$9F,$91,$F3,$4F,$8F,$FE,$01,$C1,$FE,$46,$8C
	.dc.b	$61,$FC,$7D,$AF,$F3,$1A,$E3,$EA,$F0,$8F,$E6,$73,$AC,$3F,$DB,$6D
	.dc.b	$BE,$98,$6F,$0E,$37,$FC,$AF,$E6,$9F,$31,$FE,$9A,$94,$62,$7F,$C2
	.dc.b	$ED,$7F,$DD,$ED,$7F,$A0,$39,$C7,$FD,$D7,$CE,$F0,$FC,$0D,$AF,$06
	.dc.b	$18,$F9,$09,$1F,$CF,$65,$1E,$7C,$1E,$46,$AB,$F9,$DC,$F9,$EB,$71
	.dc.b	$FD,$02,$30,$71,$F9,$D9,$1D,$63,$E1,$7C,$DF,$E9,$23,$FA,$62,$F7
	.dc.b	$32,$6B,$96,$CF,$F8,$E2,$74,$D9,$F9,$11,$88,$FF,$4B,$1F,$D9,$9F
	.dc.b	$F7,$CB,$DB,$A7,$C2,$74,$7E,$E7,$A2,$F8,$5F,$AC,$D8,$FE,$C0,$D7
	.dc.b	$7F,$05,$AF,$96,$FB,$12,$69,$27,$FC,$9F,$F0,$1E,$F7,$F1,$4D,$E3
	.dc.b	$FF,$5E,$69,$E0,$F8,$2F,$C3,$4F,$A2,$B7,$9E,$0F,$2F,$4E,$B6,$FC
	.dc.b	$45,$7C,$1A,$FF,$01,$35,$FF,$50,$B5,$FF,$04,$AF,$87,$EE,$48,$FE
	.dc.b	$FD,$F2,$EF,$8B,$F1,$FD,$04,$9F,$F0,$15,$A3,$F3,$94,$F8,$CF,$4F
	.dc.b	$0F,$90,$97,$F0,$9F,$3F,$17,$B8,$91,$E0,$E2,$F8,$3C,$52,$3B,$BD
	.dc.b	$BE,$73,$E2,$F0,$FC,$02,$4F,$CE,$7E,$1F,$D0,$7C,$A2,$47,$B3,$DA
	.dc.b	$F9,$BC,$52,$3D,$6F,$AB,$B3,$E8,$F7,$FC,$C2,$4F,$DA,$7E,$7F,$C2
	.dc.b	$FD,$62,$47,$C8,$F6,$BF,$82,$2F,$9C,$7C,$E6,$1B,$1F,$EF,$46,$BF
	.dc.b	$EF,$BF,$9D,$7A,$3B,$AF,$EB,$B5,$AA,$FC,$EB,$5F,$F7,$A3,$5F,$F0
	.dc.b	$83,$D3,$B5,$E3,$7F,$94,$BF,$80,$3F,$83,$37,$5E,$D7,$FE,$3D,$FC
	.dc.b	$9F,$F5,$37,$C2,$3F,$C9,$F1,$8A,$FF,$75,$B6,$3F,$CA,$7E,$CB,$E5
	.dc.b	$FF,$8F,$4E,$7C,$A3,$4C,$4E,$EB,$FC,$98,$BE,$5B,$8F,$81,$27,$31
	.dc.b	$FC,$A6,$75,$1E,$79,$D3,$FC,$D6,$32,$FF,$99,$C7,$C7,$BE,$3F,$EA
	.dc.b	$B6,$3F,$CB,$E6,$A3,$3F,$0B,$A6,$46,$A7,$83,$D5,$4E,$A3,$13,$75
	.dc.b	$FF,$73,$56,$A7,$FC,$D5,$F3,$2F,$84,$F8,$FE,$A4,$9C,$C7,$F4,$79
	.dc.b	$D4,$7D,$59,$D3,$FD,$4E,$32,$FC,$1A,$F8,$C4,$7B,$FC,$55,$74,$7D
	.dc.b	$08,$ED,$1F,$98,$3F,$AC,$47,$F9,$65,$AF,$FB,$EB,$AB,$F3,$BF,$3A
	.dc.b	$F9,$5E,$11,$FE,$32,$AF,$EC,$F5,$FC,$C9,$AF,$F9,$8C,$E4,$7F,$36
	.dc.b	$3B,$47,$EA,$3E,$7D,$F1,$54,$EB,$3F,$86,$F9,$EF,$C6,$F9,$32,$30
	.dc.b	$43,$FB,$B7,$2F,$F1,$37,$F3,$91,$F8,$99,$FE,$F1,$C0,$59,$63,$A9
	.dc.b	$F3,$F4,$DC,$EB,$0F,$F5,$2D,$F1,$BD,$46,$2F,$8B,$AA,$E0,$FE,$4F
	.dc.b	$E7,$FF,$26,$FE,$41,$AE,$78,$3E,$97,$CB,$FF,$0A,$9C,$CF,$F5,$C6
	.dc.b	$BF,$2D,$AF,$FA,$3D,$A9,$EB,$8F,$EE,$E3,$1F,$11,$D4,$ED,$7F,$E1
	.dc.b	$BF,$9A,$7C,$9B,$E3,$71,$F8,$93,$E3,$E9,$B9,$D6,$3F,$EC,$8B,$E3
	.dc.b	$7A,$8C,$5F,$17,$55,$C1,$FC,$D4,$FF,$08,$F9,$69,$1F,$A6,$8F,$D1
	.dc.b	$18,$7E,$EC,$61,$FE,$A7,$3F,$D8,$88,$FF,$70,$6B,$F8,$4F,$A3,$FE
	.dc.b	$2A,$7F,$6E,$8F,$E2,$FF,$09,$27,$F9,$13,$FB,$1F,$A0,$C4,$7E,$32
	.dc.b	$2E,$DF,$C9,$7C,$47,$6A,$FF,$BC,$FC,$47,$F7,$42,$61,$FE,$02,$6B
	.dc.b	$F9,$6F,$57,$F9,$39,$C1,$FF,$77,$7C,$DF,$D2,$7F,$4D,$82,$31,$1F
	.dc.b	$E1,$E8,$C6,$7F,$E4,$99,$CC,$FE,$BE,$72,$F4,$BF,$F2,$94,$FF,$B6
	.dc.b	$A3,$BD,$7F,$38,$F9,$79,$1F,$D0,$DD,$D6,$23,$F5,$B1,$73,$F8,$57
	.dc.b	$CE,$7F,$98,$7D,$07,$F8,$7D,$FF,$19,$B5,$FE,$04,$D7,$FD,$CC,$D4
	.dc.b	$FF,$89,$7C,$BF,$EB,$BD,$BF,$DA,$4F,$ED,$B1,$99,$FF,$77,$7D,$07
	.dc.b	$F2,$BC,$FF,$DD,$5F,$21,$FC,$3E,$31,$3F,$EE,$88,$C7,$8F,$F3,$7E
	.dc.b	$3D,$1B,$F3,$7F,$DC,$AE,$B0,$E7,$7D,$AF,$84,$E6,$3A,$1E,$2A,$E2
	.dc.b	$EE,$B8,$AC,$3C,$78,$78,$23,$47,$23,$55,$C9,$E1,$87,$81,$DC,$F1
	.dc.b	$73,$7D,$AF,$FB,$0C,$E5,$EF,$87,$9E,$31,$7C,$EF,$B9,$C5,$C5,$70
	.dc.b	$8E,$F5,$F1,$5D,$5C,$7D,$13,$B7,$99,$8A,$E6,$49,$F3,$61,$ED,$AE
	.dc.b	$04,$9F,$0C,$65,$E1,$5C,$EB,$8B,$E1,$C3,$D7,$17,$0F,$34,$EC,$93
	.dc.b	$A7,$CF,$7D,$6B,$FA,$94,$7C,$A3,$FA,$A6,$1F,$88,$7F,$D1,$6D,$7F
	.dc.b	$19,$F2,$5F,$E6,$57,$6D,$76,$D7,$7E,$77,$69,$4F,$1C,$3E,$B3,$71
	.dc.b	$FE,$CD,$A3,$E6,$7F,$C6,$55,$AC,$3E,$F5,$EF,$1F,$D0,$B5,$FF,$39
	.dc.b	$57,$0C,$3C,$31,$F5,$A7,$58,$F8,$AF,$F2,$B8,$F7,$DB,$AC,$4B,$E0
	.dc.b	$EB,$1E,$91,$E0,$8D,$65,$E5,$8F,$CD,$5F,$37,$89,$23,$47,$E6,$63
	.dc.b	$4E,$32,$F3,$B6,$47,$B3,$0E,$F3,$A2,$46,$AB,$9C,$62,$3F,$61,$97
	.dc.b	$C0,$3E,$34,$EA,$3D,$47,$96,$78,$46,$63,$8E,$1E,$07,$78,$7A,$71
	.dc.b	$F8,$71,$A3,$91,$AB,$FB,$71,$AC,$BE,$84,$FC,$63,$FC,$DD,$C1,$FD
	.dc.b	$D7,$F4,$DF,$F2,$5E,$9F,$E5,$4C,$FE,$0B,$FE,$56,$8C,$D7,$F9,$76
	.dc.b	$72,$FF,$8A,$DA,$FF,$B8,$9A,$FE,$30,$FF,$01,$1F,$E7,$AF,$77,$C3
	.dc.b	$54,$FE,$A3,$96,$27,$FD,$03,$37,$97,$FC,$F6,$DF,$20,$E7,$93,$F9
	.dc.b	$5E,$1F,$E8,$B8,$C7,$CA,$67,$1B,$3D,$8F,$EB,$F1,$E4,$C5,$8F,$F5
	.dc.b	$F8,$C3,$67,$FB,$1E,$09,$CF,$C2,$59,$FF,$4D,$79,$DE,$78,$C6,$1F
	.dc.b	$E4,$48,$C5,$70,$25,$7F,$BD,$9E,$F9,$FE,$67,$6F,$EF,$CD,$9F,$C9
	.dc.b	$E5,$FD,$CF,$E1,$FE,$17,$0C,$9F,$F5,$53,$F0,$E3,$19,$7F,$C0,$0D
	.dc.b	$7E,$47,$99,$E5,$6D,$FF,$BB,$1F,$59,$1F,$F0,$0E,$BF,$32,$BF,$C0
	.dc.b	$1B,$3E,$0B,$E9,$7B,$FF,$45,$B7,$FE,$0A,$6D,$36,$FE,$EF,$18,$7F
	.dc.b	$D7,$DE,$37,$D0,$F5,$3F,$00,$FF,$B2,$3E,$0A,$F0,$7F,$AA,$71,$FF
	.dc.b	$1F,$08,$CD,$70,$8C,$1F,$E4,$39,$E1,$8F,$86,$3F,$77,$7D,$5F,$F5
	.dc.b	$8F,$D0,$7F,$A2,$DB,$EE,$7E,$46,$93,$97,$FD,$78,$FF,$BB,$27,$8F
	.dc.b	$2F,$E2,$67,$6D,$9E,$11,$FE,$91,$33,$3F,$D0,$23,$A6,$7F,$E6,$56
	.dc.b	$C7,$FA,$5D,$79,$1E,$37,$35,$F2,$FE,$81,$F0,$3F,$A7,$C1,$FC,$02
	.dc.b	$F9,$8C,$56,$B2,$FF,$44,$CE,$5F,$F4,$5C,$67,$0F,$8C,$7D,$79,$D3
	.dc.b	$8B,$4E,$DF,$E1,$33,$A7,$FC,$A9,$C8,$F4,$FF,$6D,$EB,$FE,$8D,$B8
	.dc.b	$FF,$2C,$AD,$C6,$31,$F0,$F0,$B0,$E5,$7B,$3F,$E0,$0E,$1F,$56,$5F
	.dc.b	$17,$F5,$FE,$7D,$EA,$E6,$30,$DB,$FE,$DF,$5C,$1B,$3F,$66,$31,$8F
	.dc.b	$F9,$82,$38,$3F,$EF,$BF,$1F,$FD,$F7,$5F,$EF,$8A,$DE,$9F,$03,$9B
	.dc.b	$F8,$01,$FC,$4F,$E1,$AD,$7A,$70,$FE,$1B,$AF,$E5,$2F,$84,$FF,$9C
	.dc.b	$2B,$75,$F0,$BD,$FF,$E5,$B5,$FC,$0B,$E0,$BF,$91,$46,$2F,$F8,$1F
	.dc.b	$84,$9C,$DF,$F2,$56,$CF,$87,$35,$6E,$3C,$94,$FA,$1F,$C7,$C6,$1B
	.dc.b	$1F,$84,$1F,$43,$FE,$B5,$9D,$74,$FE,$9E,$B7,$9F,$FA,$FD,$B3,$FE
	.dc.b	$F1,$AE,$17,$CE,$72,$FF,$6E,$6B,$E3,$BF,$D9,$4F,$0B,$E9,$1F,$CB
	.dc.b	$7E,$6F,$FA,$17,$65,$C7,$DD,$97,$ED,$71,$FC,$5E,$EB,$DD,$7C,$38
	.dc.b	$C3,$C2,$FF,$A6,$7C,$AC,$8D,$6F,$70,$FF,$7E,$CD,$CF,$F3,$F3,$6B
	.dc.b	$FC,$0F,$4E,$5F,$44,$7E,$7F,$DF,$7D,$07,$E5,$FD,$BF,$DC,$46,$27
	.dc.b	$E9,$D7,$C6,$BF,$E1,$0D,$7A,$B5,$FF,$0A,$B5,$F0,$57,$59,$FB,$BC
	.dc.b	$CF,$81,$C5,$7E,$BB,$81,$F2,$BF,$DA,$CF,$C6,$FB,$37,$0F,$97,$F1
	.dc.b	$9F,$D5,$08,$FF,$B1,$29,$1F,$D6,$35,$FE,$E8,$7E,$5B,$CD,$B5,$F5
	.dc.b	$3E,$97,$FB,$C1,$F9,$AF,$2A,$C7,$E5,$7E,$64,$FF,$70,$6B,$FE,$D2
	.dc.b	$6E,$9F,$ED,$16,$BF,$EC,$F7,$F8,$94,$7F,$66,$6C,$7F,$7C,$E2,$72
	.dc.b	$7F,$82,$E7,$3D,$65,$F1,$B9,$3F,$5F,$AD,$C7,$CD,$EC,$79,$48,$F9
	.dc.b	$5F,$47,$B5,$D4,$FF,$0C,$7C,$5F,$6C,$11,$E1,$E4,$97,$FB,$5A,$FF
	.dc.b	$3A,$BF,$0E,$3C,$DF,$4B,$F0,$C9,$1F,$0D,$F8,$D7,$EB,$3B,$96,$3E
	.dc.b	$67,$CC,$7B,$CF,$1F,$D1,$58,$F0,$47,$F8,$7B,$03,$49,$1F,$D6,$78
	.dc.b	$3F,$C4,$B2,$F8,$7F,$69,$EE,$EA,$B5,$F6,$F0,$F9,$C7,$C2,$FA,$2B
	.dc.b	$1E,$F7,$C0,$47,$B7,$0B,$04,$F9,$7C,$9E,$55,$F1,$DB,$97,$F4,$8F
	.dc.b	$A1,$79,$1F,$E4,$7F,$11,$D3,$FE,$49,$9C,$E1,$FC,$8E,$72,$FF,$25
	.dc.b	$8C,$DF,$F7,$7C,$D1,$88,$FD,$EC,$63,$0F,$F2,$2C,$E6,$7F,$C6,$AF
	.dc.b	$2F,$0B,$F3,$71,$FE,$4C,$8C,$5F,$F8,$EB,$E7,$3F,$37,$5A,$C3,$A4
	.dc.b	$FF,$37,$9F,$C9,$BE,$5F,$8F,$FF,$31,$46,$30,$F8,$11,$8E,$9F,$F2
	.dc.b	$BC,$E6,$7C,$ED,$7C,$11,$FE,$5C,$F8,$97,$C3,$FF,$9B,$5B,$87,$F4
	.dc.b	$77,$FC,$B3,$97,$95,$CD,$FF,$76,$6C,$66,$B9,$7B,$DC,$DF,$F1,$55
	.dc.b	$7F,$AA,$BE,$0F,$FD,$57,$F3,$EC,$3F,$93,$E1,$FC,$CA,$93,$97,$FC
	.dc.b	$51,$5C,$2B,$7C,$1F,$32,$B5,$5F,$CC,$2B,$4F,$F8,$36,$75,$5F,$99
	.dc.b	$9C,$CF,$F9,$4E,$72,$F6,$3C,$33,$DB,$2D,$C7,$A4,$C9,$D5,$CA,$AE
	.dc.b	$B8,$7A,$F0,$F2,$CF,$17,$CF,$19,$BF,$24,$FD,$27,$C0,$D3,$C9,$8F
	.dc.b	$0C,$7F,$16,$F4,$8F,$01,$E4,$3D,$98,$F9,$31,$F1,$D7,$BA,$7E,$31
	.dc.b	$BC,$3C,$4F,$C1,$C7,$B5,$F1,$8F,$E0,$07,$C1,$2D,$E1,$C1,$FD,$83
	.dc.b	$E8,$1F,$02,$FA,$1E,$31,$86,$C6,$67,$F8,$64,$E5,$BF,$3D,$FE,$87
	.dc.b	$6E,$5C,$A3,$11,$FE,$6B,$94,$62,$3C,$4F,$1C,$BF,$0B,$AB,$E5,$7F
	.dc.b	$BA,$C6,$1B,$1C,$30,$FC,$2E,$1D,$1F,$C3,$C7,$1C,$2D,$70,$42,$B8
	.dc.b	$3F,$4D,$BB,$77,$8D,$B5,$FB,$C7,$1C,$7E,$04,$7E,$72,$79,$E3,$B2
	.dc.b	$61,$E7,$BE,$D7,$EF,$8F,$F0,$89,$E7,$8F,$C2,$36,$7A,$4E,$E7,$D3
	.dc.b	$5D,$AB,$8E,$1E,$D9,$FD,$B3,$67,$51,$C2,$B9,$3A,$3E,$29,$F2,$9E
	.dc.b	$DC,$6F,$A1,$FF,$4B,$CF,$F9,$4D,$AF,$FA,$01,$B9,$FF,$23,$F9,$3F
	.dc.b	$F5,$43,$FC,$96,$33,$C1,$FE,$E2,$9F,$74,$EF,$2F,$E1,$73,$FE,$E6
	.dc.b	$AD,$DF,$FB,$F3,$E8,$5E,$97,$FA,$09,$B1,$FD,$41,$AF,$F4,$16,$BF
	.dc.b	$D0,$1C,$E1,$FC,$DD,$FF,$05,$C7,$F9,$B5,$AF,$FA,$69,$DD,$FF,$44
	.dc.b	$F7,$FC,$B3,$FD,$0D,$B9,$FF,$4E,$6F,$D9,$FC,$0A,$E1,$5F,$D9,$3A
	.dc.b	$BE,$3B,$FD,$49,$AE,$8F,$EB,$0F,$C0,$B5,$3E,$37,$A9,$E0,$9D,$7A
	.dc.b	$2F,$A0,$F1,$AF,$9F,$A6,$22,$E3,$FD,$DA,$74,$D9,$FF,$56,$C6,$23
	.dc.b	$FB,$52,$F8,$8E,$A4,$8F,$F4,$A4,$6F,$6F,$EC,$5F,$44,$FA,$2F,$8A
	.dc.b	$DA,$FF,$4D,$35,$FF,$63,$BE,$97,$FD,$5D,$2F,$C2,$DF,$35,$FC,$6A
	.dc.b	$30,$D8,$CC,$FF,$87,$67,$2D,$87,$CA,$7D,$6F,$3B,$FC,$41,$3F,$E5
	.dc.b	$7F,$37,$ED,$DF,$7A,$D7,$9B,$2B,$FC,$B8,$F1,$FA,$7F,$BF,$F4,$8F
	.dc.b	$9B,$FF,$69,$B6,$BF,$C4,$2D,$7F,$C2,$4D,$7F,$CB,$DE,$6F,$8F,$CE
	.dc.b	$91,$FE,$69,$8F,$37,$D2,$C8,$FF,$36,$3E,$9F,$9B,$F5,$7D,$0F,$FB
	.dc.b	$8B,$73,$49,$5F,$E6,$FD,$7E,$07,$0C,$8F,$F3,$B4,$7C,$3F,$9F,$91
	.dc.b	$FE,$79,$7F,$37,$F9,$53,$E0,$FD,$FE,$EF,$B8,$95,$EE,$7E,$97,$81
	.dc.b	$F2,$BF,$79,$FF,$1B,$CF,$DA,$6A,$E5,$FF,$15,$9F,$76,$7B,$F9,$92
	.dc.b	$7F,$DB,$D8,$FF,$87,$1B,$E8,$F9,$3E,$63,$FE,$00,$C9,$A4,$9F,$F2
	.dc.b	$8C,$3E,$43,$D7,$F1,$D2,$7A,$C7,$EE,$97,$D1,$3E,$6B,$0D,$8F,$F7
	.dc.b	$6B,$5F,$56,$7F,$D2,$34,$FF,$2B,$FB,$7C,$2F,$FB,$65,$EE,$FF,$D9
	.dc.b	$67,$68,$C9,$1F,$F7,$CF,$D1,$BE,$63,$C8,$D8,$F0,$CF,$D4,$F6,$64
	.dc.b	$7C,$67,$EB,$FC,$36,$23,$CA,$3C,$D6,$FF,$EE,$98,$E0,$7C,$E9,$FE
	.dc.b	$33,$F4,$5F,$EA,$98,$F6,$6D,$FD,$92,$D3,$99,$EB,$18,$EB,$FC,$3F
	.dc.b	$AB,$C9,$93,$C8,$F1,$57,$1E,$0F,$F3,$36,$5F,$E0,$09,$D7,$CD,$7F
	.dc.b	$72,$F8,$8F,$D9,$F5,$EF,$4F,$9C,$DA,$56,$DE,$27,$FB,$68,$FE,$13
	.dc.b	$1B,$C0,$F1,$9E,$13,$F1,$A7,$39,$78,$09,$1D,$30,$F1,$FC,$2F,$93
	.dc.b	$E1,$E4,$F6,$8C,$BE,$08,$CB,$D7,$14,$FE,$B3,$DD,$FD,$6F,$CA,$CA
	.dc.b	$FC,$FC,$65,$FD,$0C,$65,$F9,$FA,$FF,$8D,$1F,$98,$3E,$21,$F4,$FF
	.dc.b	$9B,$6C,$69,$AF,$E1,$0F,$CF,$FB,$5B,$9F,$F2,$14,$EA,$7F,$2E,$D7
	.dc.b	$FC,$4A,$77,$8D,$9D,$FA,$E4,$77,$8F,$E1,$3E,$8C,$8F,$DD,$BF,$C3
	.dc.b	$3D,$5C,$47,$B2,$2F,$0F,$E1,$27,$FB,$6F,$17,$F8,$D3,$A6,$DB,$E2
	.dc.b	$E9,$88,$91,$F8,$28,$CB,$F8,$48,$E3,$F0,$12,$3F,$0E,$F4,$F5,$71
	.dc.b	$1F,$B6,$BF,$F2,$1C,$FF,$9A,$1B,$1F,$EE,$96,$BB,$3F,$D0,$7C,$FF
	.dc.b	$1B,$DF,$7C,$4F,$F0,$D6,$CF,$80,$FD,$27,$67,$EB,$3D,$5F,$D8,$D4
	.dc.b	$9C,$F6,$7F,$56,$7F,$5E,$DD,$FF,$D9,$CF,$48,$CC,$FE,$EA,$BF,$9C
	.dc.b	$FB,$3F,$CE,$BD,$99,$5F,$CB,$23,$2F,$F2,$E8,$CB,$F8,$E8,$FE,$83
	.dc.b	$F3,$9F,$CB,$89,$1F,$CB,$1A,$FE,$53,$93,$FA,$39,$C1,$FE,$73,$F4
	.dc.b	$FF,$D2,$88,$FF,$3C,$F8,$AA,$FE,$73,$E2,$BF,$A7,$98,$D7,$F8,$2F
	.dc.b	$C0,$FF,$3E,$24,$FF,$3C,$6B,$FC,$4B,$EE,$BF,$C5,$CF,$F1,$F6,$F0
	.dc.b	$7F,$20,$9F,$F1,$AC,$7F,$8A,$67,$F9,$99,$BF,$81,$FF,$52,$D2,$75
	.dc.b	$3F,$CE,$23,$13,$A9,$CD,$FF,$BF,$DF,$F2,$CF,$1F,$F2,$76,$DF,$53
	.dc.b	$EF,$B8,$BE,$DF,$0B,$E4,$3C,$B8,$5A,$DF,$D0,$FA,$0C,$3F,$B9,$FA
	.dc.b	$47,$95,$FE,$B9,$6C,$7F,$AF,$9A,$FB,$9A,$FF,$52,$73,$9B,$3F,$DE
	.dc.b	$C7,$F4,$59,$ED,$E5,$FF,$65,$8D,$CF,$F0,$2F,$A4,$79,$3F,$E0,$5B
	.dc.b	$FF,$74,$7A,$59,$8C,$7C,$67,$F7,$59,$FF,$56,$D7,$0F,$63,$FD,$5F
	.dc.b	$39,$94,$FF,$8D,$9C,$DF,$3A,$B1,$CA,$FA,$CE,$63,$C2,$76,$F6,$7F
	.dc.b	$8B,$E1,$D7,$0F,$04,$E8,$EF,$7C,$9E,$98,$76,$7C,$77,$CB,$0E,$D8
	.dc.b	$73,$BE,$15,$CA,$F8,$61,$E9,$8E,$CF,$82,$B8,$4E,$AB,$8C,$EE,$73
	.dc.b	$5C,$27,$26,$B0,$F1,$47,$98,$D5,$F3,$3C,$31,$A8,$F2,$1A,$25,$7C
	.dc.b	$38,$D5,$F9,$70,$F6,$92,$BE,$51,$D4,$E2,$E5,$F8,$A6,$6B,$D2,$48
	.dc.b	$FC,$35,$76,$C3,$D7,$86,$E7,$37,$D4,$FF,$12,$E1,$F5,$9B,$5F,$8D
	.dc.b	$6D,$7E,$E1,$FC,$3F,$C8,$FE,$D9,$FC,$A3,$5F,$C7,$B5,$DF,$07,$1B
	.dc.b	$4F,$F0,$3C,$BF,$0C,$DC,$7F,$AF,$F0,$3E,$4B,$F4,$D8,$FD,$CA,$D3
	.dc.b	$DB,$1F,$C0,$C6,$21,$5C,$75,$CD,$49,$E4,$FE,$3E,$75,$18,$26,$3E
	.dc.b	$49,$D6,$1E,$28,$C4,$EB,$1F,$83,$5A,$C3,$B5,$68,$93,$EC,$BE,$4F
	.dc.b	$0B,$E4,$7A,$CF,$D3,$61,$E4,$AE,$31,$96,$BE,$EA,$E6,$E6,$BA,$4E
	.dc.b	$AB,$94,$60,$93,$B0,$73,$BF,$4C,$6A,$3E,$C9,$E4,$BE,$EE,$2F,$C7
	.dc.b	$97,$DC,$C7,$E1,$4F,$B0,$FA,$6F,$FA,$43,$0F,$BB,$C5,$67,$C3,$97
	.dc.b	$FB,$B6,$7F,$C5,$AA,$75,$3C,$A7,$F2,$53,$B8,$FE,$96,$D7,$FC,$FE
	.dc.b	$D7,$FA,$27,$CC,$F7,$E4,$FE,$9D,$F0,$97,$99,$36,$30,$CC,$3F,$A4
	.dc.b	$F8,$97,$0F,$EC,$0D,$E6,$FE,$A4,$D9,$FE,$A4,$FE,$62,$7F,$C7,$38
	.dc.b	$FF,$5D,$8C,$46,$A7,$FB,$1C,$E5,$B8,$BC,$5F,$37,$D2,$F9,$DB,$3F
	.dc.b	$D7,$1E,$77,$FD,$C5,$B7,$FD,$AD,$B7,$CE,$BA,$FD,$27,$C1,$F6,$3E
	.dc.b	$13,$63,$FB,$3B,$87,$FB,$4F,$1D,$C7,$FC,$01,$E4,$FC,$9F,$2B,$E5
	.dc.b	$36,$BF,$B7,$BE,$02,$3E,$C7,$C7,$F3,$FF,$47,$E7,$FE,$93,$67,$FB
	.dc.b	$BB,$6F,$FC,$1C,$DE,$0F,$E0,$7B,$3E,$91,$E3,$7D,$D5,$E3,$D8,$C3
	.dc.b	$71,$FF,$20,$CE,$9B,$3A,$7F,$D4,$77,$FD,$F9,$FF,$3F,$FD,$23,$E6
	.dc.b	$BF,$31,$E7,$FF,$BA,$FC,$5F,$CA,$CE,$6B,$8C,$7D,$49,$D4,$7F,$96
	.dc.b	$6F,$A6,$BF,$E5,$F6,$BF,$E7,$E9,$DB,$FE,$CA,$6C,$7F,$22,$7F,$DF
	.dc.b	$5E,$E7,$8A,$39,$99,$1F,$05,$9F,$82,$D3,$70,$F6,$B7,$0E,$D5,$C2
	.dc.b	$7E,$0D,$7E,$97,$D7,$91,$FE,$2A,$F8,$4B,$7F,$E4,$19,$B7,$FE,$33
	.dc.b	$9B,$5F,$A0,$E0,$F8,$BD,$BF,$BA,$E2,$FA,$18,$FF,$A8,$F7,$FD,$8D
	.dc.b	$FF,$AB,$27,$33,$FE,$41,$BF,$CE,$BF,$EE,$B6,$C6,$A3,$B4,$F1,$9F
	.dc.b	$E0,$FF,$AA,$BE,$27,$C1,$1F,$BF,$F9,$DF,$E0,$1F,$03,$6F,$FC,$8A
	.dc.b	$DC,$3F,$CB,$8D,$D7,$FC,$9C,$FF,$17,$5F,$48,$F4,$70,$BD,$2B,$18
	.dc.b	$6E,$1F,$EF,$3A,$E3,$18,$C3,$FD,$C5,$7C,$6B,$4F,$FB,$31,$BE,$A7
	.dc.b	$E4,$3E,$65,$F1,$9E,$08,$C3,$67,$FD,$F7,$3A,$78,$98,$8D,$5F,$E3
	.dc.b	$67,$4F,$8A,$BF,$55,$EB,$4B,$FF,$34,$F5,$DA,$67,$FB,$2E,$D5,$E6
	.dc.b	$3F,$CE,$95,$6F,$FC,$E9,$F1,$1F,$CB,$CF,$E6,$5F,$96,$FF,$98,$63
	.dc.b	$FC,$EB,$19,$69,$FE,$B6,$8D,$63,$FE,$8F,$9F,$E8,$7F,$40,$F9,$2F
	.dc.b	$E8,$BE,$AB,$E3,$B3,$18,$6E,$3F,$E1,$9A,$E0,$D9,$FC,$6C,$63,$0F
	.dc.b	$F4,$AB,$73,$FE,$74,$DA,$FE,$CF,$0F,$D9,$1F,$21,$DA,$FE,$2D,$7F
	.dc.b	$6C,$EB,$6A,$FA,$7F,$83,$F3,$DF,$09,$B5,$EF,$F6,$BF,$9C,$D7,$F8
	.dc.b	$9A,$3F,$A2,$B7,$0F,$F5,$23,$5F,$E9,$AF,$FA,$12,$3E,$47,$B7,$F3
	.dc.b	$BE,$23,$E7,$BF,$99,$FD,$EF,$FB,$64,$FE,$4D,$7A,$F9,$F6,$74,$FF
	.dc.b	$67,$F3,$1E,$9F,$F4,$C6,$DF,$F5,$86,$E7,$FE,$B7,$6B,$E4,$F8,$DF
	.dc.b	$BE,$7E,$03,$E3,$ED,$7D,$DF,$A2,$FC,$1F,$C3,$FE,$11,$B1,$F8,$1F
	.dc.b	$87,$F8,$64,$7E,$1B,$F1,$BE,$63,$F1,$67,$E3,$3A,$AC,$FE,$27,$E5
	.dc.b	$BF,$1F,$F2,$FF,$90,$6C,$7D,$A3,$EB,$6A,$EF,$F7,$12,$3D,$CF,$D4
	.dc.b	$F3,$3F,$32,$7E,$6B,$C6,$B3,$FB,$DF,$DE,$DF,$2F,$F9,$EE,$EF,$E1
	.dc.b	$11,$89,$FC,$E4,$63,$E2,$BF,$BC,$F8,$2A,$46,$1B,$5F,$EE,$89,$D4
	.dc.b	$FE,$1B,$E2,$7F,$56,$7E,$B3,$C4,$B5,$ED,$E7,$78,$4F,$A3,$72,$FF
	.dc.b	$88,$DB,$5B,$9D,$72,$E6,$B4,$FF,$83,$28,$D2,$57,$F8,$6A,$3F,$4D
	.dc.b	$7E,$07,$71,$D5,$CD,$F6,$8F,$07,$DC,$7E,$95,$A7,$F8,$DD,$C5,$7E
	.dc.b	$B0,$F5,$BF,$C1,$34,$FF,$01,$BA,$8F,$F1,$7B,$5F,$D9,$35,$E3,$E1
	.dc.b	$FF,$89,$A3,$AC,$F2,$8F,$07,$89,$E1,$7F,$C9,$1F,$49,$AC,$3F,$78
	.dc.b	$F8,$5B,$1A,$76,$FF,$8C,$A3,$C1,$89,$A6,$3E,$F5,$F3,$3F,$E5,$57
	.dc.b	$F8,$D5,$FD,$37,$2D,$DF,$53,$23,$EC,$7C,$4B,$CB,$C6,$DF,$1C,$FF
	.dc.b	$79,$8F,$A5,$4F,$FA,$02,$3F,$9D,$F2,$7F,$AB,$A7,$31,$91,$F1,$DF
	.dc.b	$E8,$68,$CC,$70,$C3,$17,$B8,$D9,$DD,$ED,$C1,$F9,$A8,$C1,$D6,$F8
	.dc.b	$BF,$CB,$30,$F1,$57,$39,$CC,$F7,$38,$9E,$CC,$3C,$77,$D5,$F2,$C7
	.dc.b	$FA,$72,$78,$CE,$6B,$A4,$F2,$8C,$3E,$33,$FC,$53,$1E,$8B,$FF,$68
	.dc.b	$FB,$FD,$27,$DA,$F8,$23,$33,$D3,$0C,$63,$E8,$BF,$11,$DE,$BD,$AF
	.dc.b	$79,$DB,$EC,$36,$E0,$F0,$63,$FE,$F1,$8E,$67,$FB,$CD,$C3,$F4,$0F
	.dc.b	$B1,$F4,$AF,$9B,$7A,$BF,$EF,$98,$C3,$63,$33,$F6,$67,$2D,$F8,$2F
	.dc.b	$EF,$ED,$D7,$94,$62,$3B,$B5,$FC,$64,$A3,$8E,$3F,$88,$D3,$FD,$79
	.dc.b	$5F,$71,$AF,$DC,$AE,$17,$F8,$06,$C7,$E0,$27,$75,$CF,$CB,$FC,$D3
	.dc.b	$F9,$4A,$DB,$70,$F4,$ED,$BC,$3F,$DB,$ED,$7E,$F9,$FE,$E2,$AE,$F1
	.dc.b	$C7,$1F,$36,$1D,$B1,$FD,$41,$E2,$3D,$F8,$7B,$9B,$3C,$30,$87,$B6
	.dc.b	$74,$F7,$6B,$FB,$4A,$EC,$FF,$BD,$2B,$B3,$BA,$F8,$95,$EC,$9F,$1D
	.dc.b	$7E,$EE,$FD,$87,$A6,$FD,$BC,$8F,$96,$7C,$87,$ED,$38,$7F,$C2,$D1
	.dc.b	$8A,$FE,$84,$D7,$F9,$85,$71,$C3,$84,$7F,$4C,$9D,$3F,$ED,$E6,$BD
	.dc.b	$2F,$F8,$23,$B6,$D3,$E9,$1F,$07,$86,$D7,$FB,$91,$AF,$F1,$39,$E1
	.dc.b	$3B,$3C,$6F,$F2,$DF,$2A,$F4,$9C,$DF,$2D,$FF,$D9,$3F,$49,$ED,$6E
	.dc.b	$5C,$A3,$1A,$FF,$8E,$7E,$91,$F2,$F9,$AF,$4D,$7B,$B5,$FE,$DA,$DD
	.dc.b	$0D,$39,$C6,$BE,$53,$B1,$CD,$FE,$E1,$F2,$6F,$98,$F5,$B6,$33,$18
	.dc.b	$F8,$6C,$BF,$2A,$75,$3F,$15,$AF,$F3,$DC,$11,$88,$FF,$2A,$3F,$EE
	.dc.b	$87,$F9,$C4,$6B,$E3,$7A,$12,$3F,$C2,$1F,$47,$FE,$6D,$6F,$83,$F7
	.dc.b	$38,$13,$99,$FE,$91,$18,$9D,$4B,$C9,$A4,$AE,$AF,$E1,$3D,$CF,$EC
	.dc.b	$71,$FE,$60,$F0,$F0,$47,$CF,$F2,$DF,$D7,$DB,$E2,$7E,$32,$31,$1F
	.dc.b	$87,$6B,$F0,$FE,$8D,$F1,$BF,$92,$6C,$7C,$26,$BF,$8F,$6B,$F2,$E7
	.dc.b	$FD,$DA,$F8,$BB,$BC,$04,$7C,$BF,$1F,$E3,$23,$83,$D3,$F3,$3E,$72
	.dc.b	$38,$3D,$BF,$3D,$EB,$23,$83,$E0,$FA,$7E,$F2,$38,$3D,$5F,$23,$ED
	.dc.b	$23,$83,$FC,$95,$F4,$6F,$88,$FB,$4D,$8C,$C6,$0F,$99,$E6,$C8,$FA
	.dc.b	$2D,$7E,$77,$B5,$F4,$48,$FD,$67,$EA,$79,$32,$3E,$CB,$5F,$F0,$47
	.dc.b	$D1,$3E,$6B,$F8,$6C,$23,$0F,$F9,$6C,$C4,$6B,$E8,$DF,$33,$C2,$B7
	.dc.b	$5D,$2B,$FB,$44,$62,$3F,$C3,$8D,$7F,$C4,$4F,$0E,$B7,$8D,$FE,$25
	.dc.b	$F8,$9F,$F7,$F5,$78,$5A,$FF,$8B,$6B,$84,$FF,$AF,$F9,$7F,$B8,$4F
	.dc.b	$F9,$AD,$AF,$F2,$A6,$DF,$68,$FF,$3A,$C3,$E8,$FF,$8D,$BA,$AE,$9E
	.dc.b	$96,$0C,$CF,$F2,$0F,$A0,$FE,$3D,$7F,$E6,$DE,$CF,$E4,$7E,$D7,$EA
	.dc.b	$A3,$F9,$CC,$E9,$BE,$6E,$AA,$47,$FA,$06,$75,$9F,$94,$FF,$2A,$D7
	.dc.b	$F4,$06,$CF,$FA,$35,$AF,$F9,$D5,$FF,$2D,$FC,$36,$EF,$C1,$3F,$E4
	.dc.b	$26,$BF,$EE,$A6,$A9,$E6,$EE,$BC,$1F,$C0,$3F,$C2,$63,$C0,$D7,$C1
	.dc.b	$5E,$8D,$3A,$6F,$E1,$24,$F9,$E5,$F1,$98,$8E,$33,$71,$F8,$15,$A9
	.dc.b	$FF,$52,$7D,$27,$D5,$7E,$E7,$D2,$7D,$63,$FA,$8C,$66,$3F,$B3,$B6
	.dc.b	$BE,$F7,$92,$E2,$5F,$1B,$EB,$C3,$FA,$0C,$9D,$D4,$93,$F9,$1F,$94
	.dc.b	$AF,$F9,$BE,$76,$FF,$6E,$6B,$99,$FF,$24,$46,$32,$FC,$84,$FF,$80
	.dc.b	$FC,$9F,$D7,$FA,$DF,$CE,$67,$35,$BE,$0F,$CE,$7D,$27,$E5,$3C,$F9
	.dc.b	$8F,$E7,$63,$F2,$E6,$1F,$CF,$46,$1F,$F6,$97,$CE,$FF,$A5,$7E,$0F
	.dc.b	$FC,$12,$D8,$CF,$97,$FE,$CC,$76,$3E,$83,$71,$98,$FF,$61,$4F,$16
	.dc.b	$C6,$73,$FD,$64,$A7,$91,$FC,$27,$47,$C7,$7C,$02,$47,$83,$DD,$AF
	.dc.b	$FA,$6E,$76,$FF,$86,$9A,$E6,$7F,$CE,$91,$8B,$F0,$E1,$F9,$3A,$61
	.dc.b	$FE,$B5,$EA,$EB,$5F,$84,$EC,$F4,$6F,$E7,$E7,$46,$30,$FF,$7C,$B6
	.dc.b	$33,$9F,$F0,$1D,$A5,$7E,$E5,$FE,$09,$1F,$C2,$BD,$49,$3F,$C4,$DF
	.dc.b	$CB,$98,$7F,$8A,$37,$1F,$CB,$F4,$FF,$83,$1B,$CD,$FE,$38,$83,$49
	.dc.b	$5F,$E0,$A8,$CB,$FC,$87,$D3,$FE,$36,$7E,$9A,$7F,$8E,$92,$3F,$93
	.dc.b	$3F,$AA,$F9,$AC,$47,$F7,$98,$FE,$F8,$47,$FC,$2F,$F4,$5F,$EA,$3D
	.dc.b	$0D,$24,$7F,$93,$9F,$F0,$9E,$DF,$CD,$7E,$63,$F9,$69,$FE,$48,$8C
	.dc.b	$CA,$B9,$FC,$07,$12,$47,$81,$AF,$F9,$92,$5E,$CF,$33,$A4,$FF,$3A
	.dc.b	$F9,$B9,$5F,$D0,$DD,$34,$FF,$0D,$BF,$E1,$92,$3F,$E6,$8F,$A4,$79
	.dc.b	$DF,$EC,$16,$C6,$63,$04,$7F,$DC,$1E,$9F,$FB,$F6,$4F,$FB,$FF,$E1
	.dc.b	$AC,$FE,$96,$6B,$FE,$FE,$9E,$15,$B7,$F5,$50,$FA,$4F,$E5,$3D,$5D
	.dc.b	$A3,$F5,$D1,$87,$91,$8F,$1F,$FC,$6D,$8F,$69,$FE,$70,$D8,$CD,$F6
	.dc.b	$9E,$53,$9B,$ED,$7E,$1F,$6F,$C8,$F5,$8C,$D7,$0B,$E5,$87,$4A,$F2
	.dc.b	$DF,$07,$D5,$19,$6D,$F2,$8C,$5F,$9A,$F9,$CE,$49,$5E,$E3,$C8,$63
	.dc.b	$0F,$43,$9A,$F2,$3E,$23,$55,$D1,$D5,$F8,$30,$F3,$57,$09,$F2,$9A
	.dc.b	$C3,$DF,$3A,$AF,$39,$23,$33,$D8,$F8,$C7,$13,$C9,$3C,$63,$51,$DB
	.dc.b	$0F,$AD,$39,$8E,$11,$EC,$32,$FC,$23,$F0,$A7,$98,$B7,$F8,$36,$DF
	.dc.b	$F9,$C9,$FC,$3F,$CC,$FF,$AA,$64,$D2,$7C,$07,$FB,$22,$32,$FD,$E8
	.dc.b	$FF,$6B,$5F,$E6,$37,$7D,$27,$F8,$AE,$3F,$1D,$3A,$D3,$F1,$8D,$C7
	.dc.b	$C3,$87,$E3,$30,$E3,$C9,$D6,$24,$FD,$9A,$E0,$FE,$FB,$1E,$D8,$74
	.dc.b	$AD,$E3,$E0,$7A,$4E,$AB,$B9,$CE,$37,$5F,$AC,$C7,$C3,$39,$25,$F3
	.dc.b	$7F,$1D,$1A,$8C,$12,$7C,$53,$AB,$F0,$3F,$B7,$C3,$D3,$97,$CA,$8E
	.dc.b	$B5,$E1,$AD,$CE,$6F,$A5,$7A,$1F,$9D,$3A,$BE,$64,$C7,$DF,$97,$B2
	.dc.b	$F8,$63,$E4,$C3,$8C,$E6,$7C,$95,$72,$F6,$5E,$CF,$89,$C3,$EA,$87
	.dc.b	$CC,$D8,$C3,$70,$FC,$8E,$1F,$DF,$E9,$5F,$E1,$E8,$E7,$3A,$7F,$9B
	.dc.b	$C6,$5F,$F3,$8F,$3D,$E5,$FF,$1B,$37,$CB,$34,$97,$FE,$7D,$7F,$CC
	.dc.b	$CF,$F7,$9E,$6F,$3B,$93,$FC,$25,$D7,$FE,$42,$8F,$F6,$8C,$60,$E1
	.dc.b	$9F,$2C,$7F,$A9,$4E,$9F,$E9,$D1,$9C,$7F,$AF,$FB,$DE,$7A,$D5,$7F
	.dc.b	$BF,$DA,$F8,$E3,$59,$71,$1F,$07,$CE,$BF,$CA,$98,$FF,$A5,$E3,$13
	.dc.b	$C2,$3F,$B7,$B6,$FF,$7F,$83,$F3,$2F,$F1,$75,$7D,$28,$C4,$FC,$B9
	.dc.b	$CB,$C1,$AF,$F4,$C8,$DC,$FC,$38,$FE,$E5,$3C,$33,$F9,$1E,$AF,$C1
	.dc.b	$8C,$57,$CF,$F2,$31,$C5,$EB,$6C,$73,$C7,$FC,$0B,$1F,$1C,$91,$FE
	.dc.b	$0A,$7C,$DC,$1D,$4F,$89,$F4,$B6,$79,$61,$F8,$C6,$CE,$9A,$77,$F9
	.dc.b	$F7,$87,$F5,$DB,$19,$F2,$F1,$32,$38,$3B,$8C,$6D,$FE,$A3,$9F,$EE
	.dc.b	$3F,$49,$FE,$F7,$7F,$CD,$7A,$3D,$BF,$C4,$52,$B4,$FF,$99,$A7,$55
	.dc.b	$FE,$E9,$9C,$CF,$F3,$86,$BD,$BD,$DF,$D2,$3C,$DB,$1D,$38,$7F,$D1
	.dc.b	$2D,$AF,$F3,$77,$99,$FE,$B9,$F7,$7C,$73,$9C,$3F,$CC,$38,$BE,$03
	.dc.b	$F3,$D7,$FE,$2F,$6E,$2F,$81,$CD,$6F,$2F,$7D,$60,$DC,$D8,$FA,$BD
	.dc.b	$9F,$85,$EA,$FC,$33,$63,$27,$DA,$8C,$B4,$F8,$8F,$A3,$9F,$D2,$D8
	.dc.b	$C9,$FC,$27,$4F,$D6,$E9,$FE,$8D,$6B,$F4,$67,$E3,$9C,$A7,$F3,$91
	.dc.b	$FE,$4D,$E0,$C3,$FE,$49,$FA,$2F,$CE,$76,$7E,$CD,$B5,$F9,$9E,$C4
	.dc.b	$62,$B8,$4F,$E5,$2B,$F8,$46,$EF,$A2,$FD,$EC,$FF,$BD,$5B,$5F,$EF
	.dc.b	$55,$E4,$F0,$73,$5F,$EF,$F6,$D7,$F1,$86,$D7,$FB,$D5,$E3,$1D,$7B
	.dc.b	$B9,$79,$5E,$2F,$9A,$FF,$2B,$CF,$86,$31,$E5,$6A,$3F,$9B,$E3,$E7
	.dc.b	$9F,$E6,$4D,$8C,$E0,$9C,$B6,$3A,$1E,$DF,$33,$A6,$5F,$A7,$3F,$CF
	.dc.b	$5D,$9F,$E9,$69,$FE,$05,$F5,$5F,$EE,$C7,$F6,$D5,$FC,$AE,$31,$5C
	.dc.b	$7A,$17,$C6,$31,$3A,$CB,$FD,$F7,$39,$7F,$DF,$B1,$9C,$3F,$D7,$5D
	.dc.b	$EF,$57,$F8,$C1,$FC,$67,$E2,$AC,$FF,$0C,$8B,$8F,$86,$95,$8C,$3F
	.dc.b	$DE,$FE,$73,$D9,$FD,$FE,$8F,$6F,$F8,$0C,$E7,$3F,$C7,$CE,$5F,$C8
	.dc.b	$C6,$72,$FE,$CE,$D7,$FC,$72,$D7,$FC,$FE,$D7,$FD,$D4,$F1,$C3,$C7
	.dc.b	$87,$F1,$0F,$A3,$7C,$FF,$8A,$3F,$9B,$79,$5F,$D0,$49,$8F,$F4,$5D
	.dc.b	$7F,$A7,$E5,$FD,$0F,$E8,$7F,$A3,$12,$74,$FF,$49,$EF,$91,$C5,$B3
	.dc.b	$FE,$50,$CD,$F4,$5F,$B4,$6F,$7F,$E3,$23,$07,$EE,$37,$78,$BE,$2B
	.dc.b	$FE,$A4,$D7,$FD,$E4,$D7,$FD,$88,$D8,$FE,$C1,$E8,$FF,$5D,$26,$5E
	.dc.b	$D7,$FB,$07,$C3,$C8,$F8,$0D,$BF,$EC,$5E,$6F,$F6,$52,$4F,$C5,$7F
	.dc.b	$B4,$7D,$34,$8F,$90,$D9,$FE,$D5,$E7,$7F,$6D,$24,$FC,$D7,$FC,$35
	.dc.b	$97,$F3,$C9,$F4,$F9,$78,$25,$7D,$3F,$8B,$FE,$EC,$47,$07,$D7,$F9
	.dc.b	$4F,$EF,$44,$70,$7E,$EB,$E7,$7F,$9B,$5F,$D3,$6B,$FE,$C8,$E1,$46
	.dc.b	$23,$FC,$46,$D5,$1A,$F0,$BF,$81,$CF,$F0,$A9,$E9,$18,$3F,$C6,$DF
	.dc.b	$12,$F8,$1A,$4C,$3F,$C2,$51,$96,$FC,$EF,$02,$38,$25,$FF,$01,$94
	.dc.b	$FF,$8B,$D7,$C0,$D8,$FF,$19,$7C,$EA,$FF,$C6,$9F,$6A,$F2,$73,$18
	.dc.b	$6B,$FE,$25,$9E,$0D,$3F,$CA,$0E,$31,$FF,$16,$B7,$5E,$D0,$E7,$4E
	.dc.b	$69,$39,$87,$CE,$F8,$23,$F8,$E7,$CC,$ED,$E5,$F3,$1F,$CE,$9C,$DF
	.dc.b	$F9,$36,$31,$8F,$83,$15,$F9,$6F,$FC,$70,$D7,$C3,$1A,$BF,$F3,$54
	.dc.b	$E6,$7D,$1F,$3A,$F8,$3F,$F3,$73,$6B,$FC,$90,$D9,$FF,$26,$B5,$FF
	.dc.b	$23,$39,$C3,$C8,$D9,$F0,$BF,$C5,$5B,$5F,$E2,$4A,$F8,$BF,$01,$F1
	.dc.b	$BE,$57,$F9,$A4,$7F,$9D,$39,$3E,$6C,$62,$3E,$58,$9E,$1E,$5F,$E6
	.dc.b	$EB,$6F,$FA,$42,$3E,$8F,$CE,$71,$CF,$F3,$0F,$3B,$F0,$87,$57,$F7
	.dc.b	$1D,$47,$73,$57,$DF,$2E,$AF,$78,$C9,$E9,$8C,$BF,$EF,$2C,$2B,$FA
	.dc.b	$CB,$ED,$5C,$6F,$C3,$7A,$AF,$2E,$38,$4F,$93,$0F,$83,$1D,$CC,$DF
	.dc.b	$33,$CC,$F9,$0F,$93,$1E,$2C,$7E,$1C,$EE,$79,$61,$D5,$E1,$3A,$7D
	.dc.b	$47,$FA,$DC,$F8,$87,$FA,$E6,$B0,$E0,$FA,$FF,$38,$F9,$47,$CE,$FF
	.dc.b	$AF,$E3,$0D,$8C,$CF,$FB,$22,$72,$DF,$80,$FE,$26,$DC,$39,$46,$23
	.dc.b	$F0,$A2,$7F,$CE,$53,$A3,$E8,$E5,$FE,$D2,$9B,$F0,$5F,$8C,$CF,$F1
	.dc.b	$D7,$C2,$7E,$43,$63,$F2,$55,$C6,$7E,$63,$63,$8E,$1E,$66,$E1,$F5
	.dc.b	$AB,$FC,$21,$87,$E6,$B4,$E6,$DC,$3E,$3E,$1E,$E9,$E3,$5D,$AB,$CD
	.dc.b	$5F,$A2,$3C,$07,$0B,$FC,$6E,$3E,$78,$D4,$F0,$AF,$E2,$71,$C1,$FE
	.dc.b	$2B,$3B,$6E,$1E,$99,$E3,$1F,$18,$EB,$19,$77,$79,$7C,$C6,$9F,$D0
	.dc.b	$F1,$3E,$39,$EF,$FF,$51,$E0,$FE,$2B,$18,$BF,$F6,$23,$5F,$F2,$02
	.dc.b	$F2,$78,$47,$F2,$F8,$D6,$1F,$C8,$DB,$5D,$2F,$78,$78,$70,$BF,$3A
	.dc.b	$FA,$6F,$C3,$E5,$FC,$9A,$30,$FF,$B9,$DA,$FF,$BB,$1A,$FF,$B9,$5C
	.dc.b	$E5,$FC,$CF,$9D,$F1,$FE,$06,$D7,$FB,$7E,$B7,$9F,$F3,$4F,$9E,$FC
	.dc.b	$56,$1F,$CE,$DB,$19,$E1,$F4,$C7,$B2,$FF,$A7,$3F,$91,$7F,$DE,$CD
	.dc.b	$78,$1F,$EF,$F7,$0F,$08,$F8,$7D,$8F,$98,$A7,$68,$F8,$FE,$C3,$E4
	.dc.b	$2F,$B9,$30,$FF,$7C,$4E,$5B,$1F,$29,$FE,$C5,$08,$C7,$0F,$F9,$A1
	.dc.b	$B1,$99,$FA,$32,$F6,$FF,$DD,$11,$FE,$84,$E5,$FB,$31,$88,$FF,$41
	.dc.b	$57,$17,$FD,$19,$B3,$E7,$BF,$AD,$47,$F7,$08,$FE,$CD,$A7,$F8,$22
	.dc.b	$31,$BF,$F6,$48,$C4,$72,$AE,$2F,$FA,$3F,$E7,$9F,$33,$EA,$6C,$7A
	.dc.b	$1A,$FF,$10,$D1,$18,$8F,$EA,$F0,$FA,$87,$C4,$7E,$51,$B3,$E4,$3F
	.dc.b	$C1,$4E,$1F,$F5,$02,$F8,$DF,$CD,$FB,$7F,$8A,$9F,$F0,$83,$FE,$D2
	.dc.b	$79,$5A,$31,$1F,$E9,$9E,$9E,$3E,$97,$D4,$93,$49,$3F,$E8,$3F,$2B
	.dc.b	$B9,$1E,$BF,$99,$FF,$88,$1F,$F5,$A3,$FE,$1C,$73,$F4,$3F,$B0,$83
	.dc.b	$7C,$CC,$7C,$9F,$E1,$24,$D2,$57,$E1,$21,$F2,$BC,$C9,$1E,$2F,$9A
	.dc.b	$ED,$F4,$FD,$F9,$64,$F9,$A3,$C5,$EA,$F7,$24,$75,$7F,$31,$D0,$FA
	.dc.b	$1F,$98,$DE,$1F,$E1,$ED,$AF,$AC,$EF,$E3,$FD,$DE,$2F,$BF,$D5,$95
	.dc.b	$F1,$A3,$E1,$7A,$DE,$F2,$47,$B5,$F8,$5F,$0D,$F1,$C9,$1F,$3B,$E9
	.dc.b	$BF,$8F,$AF,$A2,$7C,$06,$1B,$1F,$89,$35,$1C,$38,$5F,$31,$B8,$FE
	.dc.b	$F1,$5C,$2B,$F9,$0B,$5F,$ED,$2D,$7F,$C5,$EE,$F5,$FC,$07,$D2,$F4
	.dc.b	$9F,$03,$5F,$F7,$F1,$E2,$3F,$0F,$F4,$FF,$83,$FA,$79,$18,$3F,$16
	.dc.b	$E1,$FC,$5B,$F8,$98,$FE,$73,$F5,$7F,$C8,$7B,$CD,$31,$F0,$1F,$B2
	.dc.b	$C7,$FC,$C9,$1B,$EA,$65,$F9,$33,$9F,$A7,$F7,$3E,$97,$F7,$F7,$FE
	.dc.b	$7A,$F9,$C7,$D8,$F0,$DB,$F6,$91,$8C,$38,$7C,$07,$F5,$28,$C4,$EA
	.dc.b	$B8,$3F,$EA,$8F,$33,$FD,$03,$3A,$8F,$E2,$F8,$F8,$27,$35,$C3,$1F
	.dc.b	$13,$63,$38,$F8,$2F,$F2,$3F,$5A,$F8,$BF,$CA,$B6,$33,$1F,$98,$F3
	.dc.b	$5C,$FA,$8F,$DA,$7D,$54,$AD,$9F,$C2,$3E,$45,$F5,$1F,$B9,$E0,$93
	.dc.b	$98,$AF,$F0,$08,$BB,$7F,$A4,$EB,$8C,$FE,$4F,$DD,$76,$7E,$69,$B1
	.dc.b	$98,$C7,$C9,$E5,$FF,$26,$4E,$A7,$FB,$AB,$5F,$F1,$EB,$5F,$F4,$1B
	.dc.b	$BB,$FE,$29,$D7,$FC,$96,$7F,$51,$E1,$F5,$7F,$8C,$E1,$FC,$54,$8F
	.dc.b	$C1,$F4,$6B,$FC,$73,$83,$F8,$D1,$8C,$3E,$AF,$87,$F8,$1E,$5F,$E9
	.dc.b	$31,$8A,$FC,$3F,$CA,$FF,$B5,$1F,$C2,$56,$FC,$39,$18,$3F,$11,$C9
	.dc.b	$CC,$8F,$2F,$1E,$BF,$E5,$CE,$AF,$CA,$CF,$F6,$F6,$C7,$F8,$69,$AE
	.dc.b	$F2,$FC,$0B,$F9,$5F,$23,$D0,$74,$CF,$F4,$FE,$FF,$EB,$89,$3F,$D2
	.dc.b	$BE,$9B,$D4,$FE,$6C,$FF,$6F,$78,$1F,$E3,$37,$FA,$7F,$E1,$2D,$FF
	.dc.b	$C5,$ED,$8F,$EA,$73,$B7,$97,$86,$9F,$F4,$DC,$72,$F8,$4E,$27,$F8
	.dc.b	$E3,$A7,$FB,$04,$64,$EF,$F4,$3D,$8D,$BE,$2F,$A4,$FF,$07,$D9,$A4
	.dc.b	$8F,$F0,$7C,$65,$F4,$FD,$4F,$A0,$F4,$CF,$AB,$E9,$24,$FB,$9F,$6F
	.dc.b	$C6,$E2,$3D,$76,$9C,$70,$7F,$BD,$61,$5F,$13,$E5,$3C,$A4,$BF,$1B
	.dc.b	$5E,$7F,$03,$F2,$8E,$0F,$ED,$7E,$87,$FC,$57,$A9,$A4,$8F,$F1,$5C
	.dc.b	$65,$FA,$BF,$1B,$F5,$0F,$99,$3F,$5B,$E7,$64,$FD,$A7,$4D,$21,$EE
	.dc.b	$72,$FF,$3E,$FA,$7F,$94,$FF,$80,$C3,$8F,$2F,$FD,$66,$D8,$FF,$7C
	.dc.b	$B5,$1C,$37,$FF,$6A,$9A,$E1,$B1,$A9,$46,$23,$81,$CB,$E4,$7C,$7F
	.dc.b	$4F,$FE,$08,$E7,$C6,$1D,$A9,$39,$55,$C3,$DB,$FC,$C3,$CB,$C9,$ED
	.dc.b	$5B,$8D,$61,$D2,$FD,$33,$9C,$3B,$3F,$E4,$AA,$DC,$F6,$9C,$E1,$D9
	.dc.b	$B7,$CE,$7C,$B8,$78,$2B,$51,$D9,$AF,$A9,$B8,$77,$BE,$73,$DC,$E1
	.dc.b	$7D,$67,$CD,$1E,$8B,$ED,$39,$BE,$67,$48,$CC,$F3,$3C,$13,$B3,$A4
	.dc.b	$6B,$0F,$79,$30,$F8,$71,$97,$CB,$19,$AF,$11,$CC,$ED,$87,$99,$EE
	.dc.b	$62,$B9,$3E,$C3,$78,$79,$DE,$55,$D5,$E2,$FD,$B7,$B1,$F8,$33,$A9
	.dc.b	$6B,$F0,$8D,$9F,$E3,$4E,$7D,$C7,$D3,$FF,$A7,$EB,$FC,$B2,$D8,$CC
	.dc.b	$FF,$B3,$27,$2D,$8C,$DF,$E5,$9B,$8F,$E5,$7D,$27,$CE,$37,$FB,$15
	.dc.b	$AC,$3F,$1E,$D7,$F3,$6D,$AE,$33,$C3,$29,$5F,$32,$B5,$7D,$63,$13
	.dc.b	$AC,$7C,$8F,$E4,$F0,$E7,$3A,$24,$E9,$FC,$24,$6A,$30,$4C,$7F,$47
	.dc.b	$87,$83,$1E,$D1,$8C,$2E,$1F,$C0,$B0,$E5,$39,$8E,$ED,$C3,$F8,$7C
	.dc.b	$EA,$7E,$1D,$72,$7E,$1B,$E9,$C7,$DD,$5E,$39,$D9,$D0,$F0,$46,$AF
	.dc.b	$B1,$23,$73,$E0,$9D,$46,$27,$4F,$80,$F8,$71,$A7,$13,$E8,$FA,$5C
	.dc.b	$60,$F9,$2B,$18,$6E,$5F,$C9,$0F,$CB,$4A,$7E,$C4,$61,$B7,$F8,$96
	.dc.b	$AE,$EF,$ED,$DE,$87,$DF,$EC,$FE,$A8,$DB,$FE,$63,$3F,$DC,$FE,$A7
	.dc.b	$17,$C7,$B7,$FC,$87,$5F,$DF,$AF,$37,$D2,$3F,$C0,$37,$8A,$DD,$FF
	.dc.b	$7E,$EA,$FF,$48,$D6,$8F,$F2,$77,$B9,$BB,$E7,$3A,$AF,$F7,$53,$5F
	.dc.b	$F7,$4B,$5F,$3B,$5F,$37,$9B,$E1,$F8,$7F,$13,$70,$FF,$0A,$BF,$D5
	.dc.b	$67,$71,$FE,$02,$8C,$65,$FE,$13,$6E,$4F,$8E,$F6,$1E,$DF,$8E,$B7
	.dc.b	$FE,$20,$8B,$87,$F8,$0A,$6E,$DF,$E0,$1F,$A5,$F8,$9D,$FF,$15,$B8
	.dc.b	$7F,$8A,$1E,$E4,$7F,$C2,$AF,$8B,$7F,$99,$F4,$DF,$35,$B3,$FE,$30
	.dc.b	$70,$FF,$C0,$0F,$8B,$7C,$8E,$33,$FF,$17,$F9,$4F,$A0,$FB,$3E,$37
	.dc.b	$E3,$67,$31,$87,$EE,$CE,$F0,$FF,$1C,$CF,$E0,$3E,$95,$C9,$A8,$FC
	.dc.b	$3F,$A3,$FC,$5F,$5B,$F8,$F7,$5E,$1F,$F5,$66,$29,$CE,$7C,$23,$07
	.dc.b	$F9,$64,$AF,$F9,$7D,$AF,$F7,$9B,$ED,$1B,$9E,$47,$8B,$99,$AE,$27
	.dc.b	$FD,$72,$7D,$B2,$F1,$F8,$79,$3C,$4D,$8D,$C7,$3F,$33,$19,$7F,$96
	.dc.b	$2F,$FA,$0D,$7F,$86,$33,$FE,$55,$8F,$83,$87,$E6,$F8,$3F,$39,$B5
	.dc.b	$B3,$F5,$F7,$CB,$0F,$F3,$1F,$1B,$D0,$DC,$7F,$BA,$BE,$A5,$F4,$5F
	.dc.b	$B4,$87,$C3,$3D,$1C,$56,$B2,$ED,$5F,$E2,$CA,$D4,$62,$13,$B7,$FB
	.dc.b	$11,$8B,$FD,$33,$67,$2D,$37,$3C,$76,$FE,$95,$19,$6C,$3E,$97,$CB
	.dc.b	$7F,$90,$F0,$E6,$1F,$97,$F0,$EE,$3F,$09,$B8,$FC,$CD,$7F,$90,$57
	.dc.b	$B3,$87,$EE,$E1,$FC,$5A,$9F,$64,$F9,$0F,$CE,$F0,$66,$31,$87,$1E
	.dc.b	$1F,$9C,$FC,$6B,$E3,$7B,$3A,$7D,$4F,$E4,$A3,$F9,$45,$F2,$D7,$FD
	.dc.b	$45,$D3,$FD,$D7,$D2,$FC,$D1,$E3,$F9,$3F,$F3,$FE,$5F,$90,$8C,$B4
	.dc.b	$F0,$46,$AB,$9C,$27,$97,$AD,$C2,$7F,$D6,$18,$AF,$83,$FA,$CA,$FD
	.dc.b	$BF,$C6,$FE,$F3,$8A,$E3,$EF,$6E,$1F,$C0,$A7,$F9,$CF,$CC,$FE,$12
	.dc.b	$31,$5F,$17,$CC,$FE,$0A,$7F,$06,$F9,$9B,$1F,$25,$B3,$FA,$36,$E5
	.dc.b	$F8,$2F,$A9,$FF,$7F,$D7,$E3,$FE,$C9,$F3,$1F,$99,$9C,$E1,$FE,$CB
	.dc.b	$13,$AA,$FC,$B4,$EA,$7F,$C6,$CD,$7F,$CF,$4D,$7F,$D1,$2D,$4F,$6B
	.dc.b	$F2,$7C,$57,$EA,$7A,$3E,$75,$F6,$8F,$E4,$7E,$7C,$8F,$C3,$7D,$55
	.dc.b	$C7,$F8,$07,$6F,$F4,$6E,$9F,$CE,$57,$EA,$31,$FE,$1F,$39,$35,$EA
	.dc.b	$7F,$2E,$F8,$DF,$E5,$ED,$AF,$C8,$7D,$57,$F8,$05,$BC,$5F,$9B,$87
	.dc.b	$15,$8F,$E6,$FE,$5F,$84,$95,$E1,$CF,$F9,$B6,$9F,$E5,$BD,$3F,$90
	.dc.b	$F2,$FF,$3E,$FA,$AF,$E7,$ED,$BF,$D1,$78,$5F,$D0,$8F,$E8,$7F,$55
	.dc.b	$63,$F4,$DF,$1B,$EF,$23,$DD,$F8,$5F,$33,$FD,$2B,$A7,$FA,$5B,$67
	.dc.b	$F5,$DD,$BF,$D3,$8F,$F0,$3F,$DE,$FF,$96,$E3,$27,$EE,$DF,$D6,$6B
	.dc.b	$FE,$B5,$D5,$18,$8F,$F0,$AB,$5F,$F0,$E3,$63,$7F,$3B,$DB,$DE,$EE
	.dc.b	$DB,$ED,$F0,$1E,$23,$C7,$F1,$76,$3F,$B4,$6D,$FE,$02,$F2,$8D,$24
	.dc.b	$7F,$26,$EE,$7D,$5F,$FB,$FE,$3F,$8A,$7D,$03,$E5,$7F,$8A,$4E,$77
	.dc.b	$5F,$F4,$6A,$D5,$FF,$1B,$F9,$F7,$C8,$7F,$8B,$1B,$1F,$C7,$FE,$6D
	.dc.b	$F3,$D9,$AF,$F1,$8C,$62,$7F,$24,$DC,$7F,$04,$D7,$E6,$35,$FF,$1F
	.dc.b	$BD,$9F,$2C,$62,$BA,$D7,$67,$FC,$7F,$F5,$7F,$B5,$AD,$9F,$E4,$DE
	.dc.b	$FF,$F1,$FE,$7F,$D8,$A7,$4F,$F6,$48,$CE,$1F,$E4,$E9,$E5,$5F,$7D
	.dc.b	$AE,$7A,$3F,$C6,$71,$88,$FF,$2D,$3D,$27,$47,$F4,$37,$F9,$23,$67
	.dc.b	$FC,$AB,$E6,$62,$BC,$FC,$CF,$AB,$7C,$16,$BA,$BF,$CB,$71,$8C,$7F
	.dc.b	$A7,$61,$C7,$0F,$CA,$35,$FF,$33,$B5,$FF,$37,$37,$07,$D5,$BE,$77
	.dc.b	$FC,$CE,$DB,$FE,$B4,$FE,$0A,$FC,$AE,$5B,$3C,$0E,$3C,$B6,$BF,$BD
	.dc.b	$48,$7C,$55,$FA,$AF,$F4,$64,$7F,$85,$72,$FF,$5F,$46,$23,$81,$C0
	.dc.b	$39,$E1,$FE,$A8,$EE,$3B,$57,$5F,$1B,$8D,$69,$FF,$04,$C6,$67,$C1
	.dc.b	$85,$7B,$C5,$AE,$11,$FE,$F1,$8C,$61,$E6,$BE,$73,$E4,$7B,$C6,$63
	.dc.b	$81,$93,$B9,$4E,$E3,$E5,$3D,$91,$FE,$DF,$7D,$91,$F0,$5C,$E1,$F2
	.dc.b	$DC,$DF,$92,$3D,$53,$F0,$8E,$D8,$7A,$9E,$F8,$F2,$7C,$46,$DC,$1D
	.dc.b	$1F,$F1,$59,$F4,$4F,$F1,$6D,$61,$C1,$F5,$BE,$A9,$F2,$8F,$AB,$FF
	.dc.b	$19,$C6,$1B,$19,$9F,$F1,$DC,$E5,$BE,$4F,$F1,$26,$E3,$CA,$31,$1F
	.dc.b	$84,$13,$FE,$6A,$9D,$1F,$6F,$5F,$F4,$6C,$DF,$80,$FC,$5E,$7F,$8D
	.dc.b	$BE,$13,$F4,$DB,$1F,$91,$AE,$33,$F5,$DB,$1C,$70,$F3,$37,$0F,$A9
	.dc.b	$3F,$98,$38,$EB,$CF,$1F,$CD,$BB,$C3,$F0,$13,$F2,$70,$E6,$F1,$AF
	.dc.b	$3B,$67,$F2,$D8,$F9,$A3,$86,$1E,$D9,$87,$AE,$F9,$BF,$B1,$AE,$0F
	.dc.b	$EC,$EB,$D6,$75,$78,$57,$6A,$E3,$1F,$F0,$06,$FC,$6F,$16,$BF,$4A
	.dc.b	$3B,$F1,$3E,$7D,$EB,$FE,$CB,$83,$FD,$F3,$1F,$B9,$6E,$59,$C1,$7C
	.dc.b	$6F,$0D,$8F,$F3,$65,$F1,$76,$D8,$D4,$7F,$2A,$FA,$A7,$93,$FE,$84
	.dc.b	$94,$62,$7F,$C4,$8D,$7F,$C2,$8D,$7F,$BF,$39,$D3,$FD,$53,$F5,$6F
	.dc.b	$9E,$F2,$B6,$BF,$D0,$15,$BD,$3F,$90,$60,$FA,$BF,$F5,$45,$7F,$A1
	.dc.b	$5A,$FF,$40,$6E,$3F,$EB,$67,$E4,$C6,$FE,$07,$D5,$1C,$23,$FC,$24
	.dc.b	$D7,$BB,$5F,$96,$F4,$BF,$1F,$D5,$FF,$7A,$9F,$F4,$B3,$5F,$15,$7D
	.dc.b	$0F,$07,$FA,$8C,$7F,$A6,$D7,$C8,$E4,$D7,$81,$67,$FC,$9D,$3A,$6C
	.dc.b	$FD,$18,$C4,$7F,$90,$7D,$EF,$E0,$EF,$FA,$9D,$AF,$F6,$26,$E7,$FD
	.dc.b	$D6,$BF,$07,$E3,$BE,$82,$92,$FF,$C2,$DF,$51,$F8,$8F,$75,$F3,$7F
	.dc.b	$81,$6C,$FF,$98,$DA,$FF,$94,$DA,$FF,$8F,$7C,$F7,$AB,$F9,$27,$EE
	.dc.b	$47,$E5,$9B,$C6,$77,$1F,$EC,$1F,$AA,$7C,$77,$E6,$9B,$19,$8C,$7C
	.dc.b	$7F,$E8,$61,$EE,$FF,$8D,$5F,$F6,$DB,$5F,$F7,$0B,$5F,$E4,$4D,$7F
	.dc.b	$B3,$FD,$B3,$E9,$7A,$7D,$37,$E9,$E4,$D9,$D4,$7F,$97,$7E,$93,$F5
	.dc.b	$FE,$E9,$A6,$38,$FC,$DE,$C7,$FB,$FA,$BF,$DD,$ED,$7F,$DE,$AD,$7F
	.dc.b	$9B,$35,$FE,$F7,$FB,$2B,$E1,$3F,$80,$74,$3F,$6C,$79,$18,$6D,$FF
	.dc.b	$82,$67,$FD,$F3,$7C,$63,$FC,$E9,$EC,$FA,$08,$FF,$9D,$FD,$9F,$5F
	.dc.b	$8D,$FC,$07,$6F,$DB,$4F,$FA,$45,$AF,$F9,$ED,$AF,$F9,$EB,$B3,$E1
	.dc.b	$F6,$48,$FF,$41,$47,$C6,$F8,$89,$1F,$E8,$57,$E4,$FC,$97,$CA,$FA
	.dc.b	$CF,$01,$27,$FD,$65,$82,$3C,$9E,$BF,$D0,$F5,$E5,$7F,$A3,$E3,$F0
	.dc.b	$9F,$70,$F8,$7A,$49,$F8,$9E,$57,$F8,$B7,$E3,$DF,$2C,$F8,$9D,$57
	.dc.b	$F6,$71,$87,$37,$06,$30,$FE,$FF,$19,$8E,$0D,$E5,$63,$DF,$07,$C5
	.dc.b	$7F,$39,$CB,$C2,$F8,$2B,$78,$7F,$A8,$BE,$93,$F1,$BE,$04,$8F,$F5
	.dc.b	$64,$7E,$40,$C3,$FE,$62,$E0,$FC,$AB,$63,$F2,$BF,$55,$23,$FD,$68
	.dc.b	$FF,$46,$FC,$4B,$F7,$78,$78,$3C,$8F,$F5,$AC,$EE,$BA,$3B,$F2,$FD
	.dc.b	$1A,$BE,$3E,$92,$BE,$68,$DB,$E6,$F8,$19,$8F,$F0,$94,$7D,$09,$F9
	.dc.b	$6D,$8F,$A0,$D7,$75,$DD,$EB,$1F,$63,$E3,$FC,$67,$5C,$FD,$5A,$7F
	.dc.b	$91,$6B,$C7,$A2,$1F,$4F,$F0,$4F,$85,$EF,$4A,$DC,$E9,$B3,$97,$FD
	.dc.b	$1F,$0F,$A6,$F9,$44,$8E,$3F,$37,$5D,$C5,$C7,$CB,$3F,$C6,$7E,$DD
	.dc.b	$9D,$25,$FC,$A1,$F0,$CF,$91,$CC,$7E,$56,$B8,$BF,$94,$6B,$9A,$FF
	.dc.b	$29,$C6,$30,$FC,$6C,$F4,$F9,$2F,$E5,$58,$1E,$B3,$F2,$BE,$77,$F0
	.dc.b	$C4,$AF,$8B,$55,$F8,$71,$67,$E8,$CF,$E1,$B8,$7F,$0F,$F0,$92,$FE
	.dc.b	$CC,$7E,$C1,$AE,$5F,$F6,$8F,$17,$ED,$27,$94,$7C,$3C,$1C,$DF,$D0
	.dc.b	$5B,$19,$8F,$E8,$9F,$00,$9D,$47,$F9,$D1,$DC,$FE,$F5,$AF,$F9,$95
	.dc.b	$AF,$FA,$A9,$DE,$5F,$22,$BE,$9C,$FD,$56,$BF,$EE,$83,$C1,$5F,$AD
	.dc.b	$FA,$AF,$D7,$79,$B2,$7F,$1B,$19,$7F,$1F,$19,$7F,$1B,$1F,$B5,$F8
	.dc.b	$8F,$DB,$7C,$CC,$9F,$CA,$C6,$5F,$CB,$C6,$5F,$CA,$C7,$EF,$7E,$A3
	.dc.b	$FC,$75,$CC,$69,$27,$F8,$F7,$4F,$E9,$63,$F8,$27,$D4,$49,$FD,$1C
	.dc.b	$7F,$07,$FA,$8F,$E1,$1F,$49,$27,$F4,$F1,$97,$F5,$31,$97,$F4,$F1
	.dc.b	$FC,$3F,$DB,$F1,$92,$3C,$38,$FF,$7E,$7F,$D2,$D8,$FE,$DE,$7F,$8B
	.dc.b	$7B,$72,$7F,$6B,$1F,$C6,$7E,$0F,$DA,$48,$E6,$D7,$F6,$FF,$03,$FC
	.dc.b	$7C,$E0,$FE,$F9,$C3,$4F,$E0,$31,$87,$F7,$FF,$4F,$FC,$9C,$C7,$27
	.dc.b	$F0,$EF,$A4,$FF,$77,$F9,$5F,$EF,$0C,$FF,$87,$46,$2F,$FC,$B7,$A2
	.dc.b	$31,$1F,$E7,$88,$C3,$FE,$FF,$AE,$51,$88,$FF,$80,$3E,$0F,$F9,$93
	.dc.b	$F0,$BF,$F7,$63,$63,$FD,$E2,$F6,$BB,$01,$F5,$4F,$9D,$FF,$1D,$FC
	.dc.b	$A7,$F1,$99,$D4,$E7,$CA,$E1,$39,$9E,$35,$C3,$0E,$91,$D3,$CB,$EE
	.dc.b	$EA,$F8,$CE,$B0,$E9,$3E,$A3,$BD,$F3,$7C,$37,$CA,$77,$87,$85,$B7
	.dc.b	$CE,$F8,$5F,$84,$EE,$74,$C2,$BE,$33,$C0,$5B,$E9,$87,$7A,$E7,$87
	.dc.b	$6A,$E4,$70,$53,$BB,$EB,$1B,$9E,$98,$7B,$EB,$51,$C6,$FA,$9D,$A3
	.dc.b	$38,$7B,$63,$38,$74,$9C,$DF,$BE,$3B,$CF,$1B,$B1,$1E,$CF,$17,$17
	.dc.b	$D6,$2D,$79,$66,$9E,$E9,$D3,$EF,$BF,$4B,$F7,$8F,$94,$5C,$3E,$DB
	.dc.b	$70,$FF,$40,$BF,$8A,$F0,$FF,$9E,$3B,$6B,$F8,$36,$BB,$F0,$BF,$C5
	.dc.b	$75,$FD,$45,$F9,$B3,$FD,$17,$0F,$C9,$7D,$2F,$F6,$6D,$F3,$53,$0F
	.dc.b	$AD,$7B,$CF,$F2,$B7,$B9,$CD,$6F,$0F,$C6,$5E,$E7,$F4,$D5,$B9,$CD
	.dc.b	$6E,$FD,$0F,$E3,$27,$51,$8A,$EA,$4A,$F3,$F8,$DF,$E3,$A7,$9C,$66
	.dc.b	$78,$46,$B1,$ED,$3F,$0C,$D1,$E4,$AE,$55,$C0,$91,$BA,$E4,$F0,$3F
	.dc.b	$7F,$8F,$B5,$B8,$FA,$6F,$9C,$71,$BE,$6E,$8F,$D3,$E3,$EE,$74,$43
	.dc.b	$58,$78,$0F,$69,$27,$9C,$F1,$9D,$C7,$F1,$9B,$F0,$F0,$B8,$1E,$1E
	.dc.b	$1F,$F3,$9E,$9F,$BC,$9D,$4E,$6B,$FC,$CB,$3F,$CB,$EB,$78,$7F,$87
	.dc.b	$2F,$A5,$6A,$3C,$15,$FC,$B5,$FE,$55,$8F,$F9,$BB,$8F,$FC,$C1,$83
	.dc.b	$E2,$F1,$7A,$3F,$C9,$31,$96,$D6,$AD,$8F,$AE,$30,$FB,$7E,$46,$DA
	.dc.b	$6C,$61,$B7,$D0,$EF,$1F,$E3,$6B,$FE,$80,$D7,$BD,$23,$17,$DE,$BF
	.dc.b	$BD,$CB,$E4,$BC,$3F,$1B,$E2,$6C,$FB,$1E,$33,$B8,$E3,$18,$33,$F1
	.dc.b	$9E,$AF,$A0,$F5,$B6,$7E,$2B,$72,$6B,$F3,$23,$0F,$CE,$F4,$7E,$1F
	.dc.b	$1F,$C4,$6D,$FC,$E7,$C6,$47,$D6,$F9,$BD,$2F,$97,$EA,$FC,$C6,$CF
	.dc.b	$D6,$6E,$3E,$28,$FE,$D1,$18,$C3,$EF,$FB,$1F,$53,$D8,$FA,$AD,$AF
	.dc.b	$BE,$DB,$E4,$7D,$07,$15,$FE,$6A,$FA,$76,$BF,$83,$87,$B4,$F8,$1F
	.dc.b	$C3,$60,$BD,$E3,$FE,$A6,$BD,$D6,$A7,$35,$FE,$D1,$6B,$FC,$CD,$AF
	.dc.b	$F7,$26,$BF,$DF,$DE,$33,$E5,$C3,$C9,$83,$E0,$BB,$47,$E6,$7D,$37
	.dc.b	$E4,$2F,$47,$0B,$CF,$B1,$86,$EB,$FE,$B2,$9D,$36,$74,$FE,$A3,$E8
	.dc.b	$BF,$42,$7E,$8B,$E8,$6C,$64,$9B,$7C,$EE,$AF,$E0,$93,$A8,$FF,$36
	.dc.b	$7D,$77,$F7,$2A,$F9,$1E,$73,$97,$E6,$CE,$6B,$EB,$F0,$AB,$57,$FE
	.dc.b	$AB,$AD,$57,$F5,$D6,$BF,$E1,$C6,$BF,$E2,$96,$A7,$B5,$7F,$AA,$9B
	.dc.b	$3F,$E2,$F8,$F0,$74,$79,$3C,$7F,$E0,$3F,$25,$2B,$F8,$5F,$CB,$DC
	.dc.b	$7E,$07,$72,$B1,$8F,$F1,$4A,$B7,$FE,$41,$F9,$FF,$E1,$7F,$3F,$FC
	.dc.b	$31,$B1,$FC,$59,$B8,$F4,$6D,$A7,$FC,$89,$CA,$FC,$81,$F0,$9F,$80
	.dc.b	$F3,$BF,$09,$18,$AF,$BD,$18,$9F,$E2,$55,$C5,$B5,$BC,$7F,$DF,$AD
	.dc.b	$7F,$C8,$F3,$BB,$FF,$7C,$CE,$F0,$F2,$52,$BC,$9E,$DF,$7B,$FF,$7C
	.dc.b	$FC,$67,$A5,$A7,$F2,$E5,$F5,$36,$21,$A3,$8B,$8B,$FC,$74,$FF,$BA
	.dc.b	$E7,$AF,$9F,$CF,$E0,$BA,$36,$32,$78,$A3,$2F,$FB,$D7,$E9,$9F,$17
	.dc.b	$F2,$39,$7F,$7F,$18,$9F,$7C,$62,$7F,$9D,$62,$AD,$3F,$E2,$F3,$11
	.dc.b	$A8,$CC,$7F,$8C,$DD,$EB,$FE,$FA,$6D,$FD,$CD,$9F,$15,$DE,$3E,$4F
	.dc.b	$AB,$F1,$FE,$9B,$E4,$36,$3E,$6F,$CA,$7F,$4A,$DC,$D2,$47,$FB,$1A
	.dc.b	$1F,$29,$F4,$4F,$A4,$D8,$FA,$FE,$37,$D5,$F2,$7E,$B3,$63,$ED,$FC
	.dc.b	$87,$72,$3D,$5F,$17,$C8,$FD,$D3,$EF,$7C,$8D,$9F,$D7,$1F,$E9,$25
	.dc.b	$F0,$5F,$CC,$7E,$E5,$1F,$CC,$0C,$CF,$F2,$F8,$CB,$CB,$8F,$FD,$4D
	.dc.b	$FA,$CF,$C8,$8F,$A5,$08,$C3,$F5,$5A,$FF,$99,$5A,$FF,$AB,$FC,$86
	.dc.b	$BF,$52,$3F,$C0,$AD,$8C,$FC,$9F,$E6,$79,$FF,$A2,$F6,$BD,$4F,$E9
	.dc.b	$5F,$7C,$F1,$33,$18,$7F,$A5,$FE,$B1,$FC,$DF,$5F,$3B,$74,$FC,$53
	.dc.b	$5F,$F3,$D3,$5F,$36,$9F,$D5,$A7,$F5,$BE,$97,$EA,$FA,$BF,$58,$FF
	.dc.b	$86,$BE,$53,$F4,$D7,$FC,$D5,$78,$39,$F0,$74,$FF,$88,$3E,$75,$E8
	.dc.b	$7F,$85,$1B,$97,$E7,$78,$D1,$8C,$7F,$DF,$ED,$7F,$DD,$5F,$54,$E4
	.dc.b	$FF,$15,$B6,$FE,$C5,$A3,$D5,$5F,$D2,$5D,$3F,$D2,$FB,$7F,$2F,$5C
	.dc.b	$1B,$1F,$D3,$FE,$AA,$F8,$9F,$C4,$B0,$FC,$ED,$7F,$8E,$A7,$75,$FE
	.dc.b	$FD,$6B,$F7,$DA,$FF,$90,$9B,$27,$F5,$77,$E8,$7B,$BC,$7C,$0E,$D7
	.dc.b	$FE,$48,$FA,$AF,$E8,$53,$EE,$F7,$9F,$21,$9F,$3F,$FC,$9B,$18,$9F
	.dc.b	$87,$5C,$33,$FF,$5C,$66,$8C,$47,$FB,$E2,$31,$7F,$E4,$89,$CC,$FE
	.dc.b	$65,$E5,$A7,$C3,$6D,$E7,$A3,$C4,$FD,$2F,$94,$79,$FF,$E6,$0F,$AA
	.dc.b	$FE,$E1,$39,$8C,$7A,$BF,$E6,$4A,$DC,$FE,$0D,$B9,$7A,$1A,$F9,$DA
	.dc.b	$FE,$B7,$EA,$BF,$54,$7D,$EC,$FE,$C3,$8E,$4F,$F3,$8B,$6F,$F0,$ED
	.dc.b	$AF,$F1,$D8,$27,$C4,$DF,$A9,$FF,$34,$47,$F7,$BC,$7F,$AA,$C6,$23
	.dc.b	$80,$9E,$1F,$09,$8C,$97,$C2,$3F,$9E,$4E,$AB,$A6,$35,$E7,$16,$F8
	.dc.b	$CE,$67,$CA,$30,$F1,$4E,$A3,$17,$D9,$A7,$62,$9D,$8B,$1C,$67,$D9
	.dc.b	$1D,$9E,$B8,$7A,$63,$DE,$7B,$C7,$CD,$EB,$0F,$94,$F2,$8D,$39,$77
	.dc.b	$87,$96,$F0,$6F,$1F,$81,$87,$7A,$F3,$4E,$AB,$81,$F3,$67,$53,$F1
	.dc.b	$0F,$E6,$F5,$FC,$E1,$C1,$F0,$CF,$4F,$D4,$FF,$96,$FE,$AB,$F9,$EB
	.dc.b	$63,$FB,$83,$5D,$FC,$0A,$77,$F0,$1F,$E9,$C9,$DB,$5E,$AD,$7F,$0B
	.dc.b	$5F,$E6,$BD,$78,$9F,$7F,$2F,$F6,$FC,$DF,$8C,$FC,$5E,$9F,$8D,$BE
	.dc.b	$13,$F6,$1B,$1F,$91,$AE,$33,$F7,$1B,$1C,$70,$F3,$37,$0F,$AB,$3F
	.dc.b	$98,$38,$EB,$CF,$1F,$CD,$BB,$C3,$EB,$9F,$A5,$C7,$C3,$5C,$6B,$CE
	.dc.b	$D9,$FC,$D4,$E9,$E1,$8F,$8A,$75,$1E,$7A,$E2,$EB,$0E,$35,$B8,$FD
	.dc.b	$A5,$7B,$0E,$38,$F9,$4F,$36,$3E,$99,$F6,$9C,$30,$F5,$D7,$D2,$BF
	.dc.b	$B1,$5F,$BC,$E7,$7D,$53,$A7,$F7,$1C,$3F,$DE,$63,$19,$7F,$4E,$6B
	.dc.b	$FE,$95,$5F,$13,$C2,$3F,$E0,$0C,$69,$FE,$5C,$D7,$A5,$7F,$83,$2F
	.dc.b	$C2,$DF,$AA,$FE,$53,$3F,$D9,$5A,$FF,$3A,$6B,$FC,$E5,$AF,$F8,$3D
	.dc.b	$DE,$DF,$D6,$7E,$6D,$EA,$79,$70,$46,$2F,$FB,$4D,$F0,$CF,$FC,$31
	.dc.b	$C3,$FE,$6A,$9F,$ED,$AD,$7F,$D6,$8D,$C7,$FB,$37,$D5,$7F,$12,$AE
	.dc.b	$3E,$EF,$B6,$B8,$5F,$F4,$F6,$BA,$3F,$C0,$4F,$39,$FF,$14,$63,$FE
	.dc.b	$6A,$9F,$EF,$AD,$7C,$35,$EB,$D1,$F2,$B4,$93,$FD,$9F,$E5,$2B,$FD
	.dc.b	$8A,$76,$FF,$84,$5A,$E6,$7E,$A7,$D4,$FF,$6E,$8F,$F0,$B3,$5F,$ED
	.dc.b	$6D,$7F,$B5,$35,$FF,$20,$7D,$E3,$E4,$7E,$F3,$63,$FC,$40,$D7,$FB
	.dc.b	$AB,$5F,$EE,$8B,$E4,$F8,$92,$38,$BF,$89,$F2,$FF,$03,$0E,$2F,$F1
	.dc.b	$5F,$85,$92,$47,$F9,$6B,$EA,$7F,$D0,$6D,$F4,$BF,$93,$46,$23,$FD
	.dc.b	$20,$D7,$F9,$FF,$D4,$7F,$B1,$1F,$F3,$1B,$5F,$F4,$B3,$5D,$FC,$3F
	.dc.b	$F3,$1F,$D9,$1D,$B4,$93,$FE,$6C,$8C,$BE,$5F,$1B,$C6,$47,$07,$83
	.dc.b	$DB,$FE,$79,$87,$E0,$E3,$CB,$18,$3F,$A0,$7E,$C8,$F8,$4A,$49,$FF
	.dc.b	$3D,$C6,$5F,$F4,$D7,$C4,$7E,$CF,$DA,$34,$91,$F0,$2B,$C2,$47,$07
	.dc.b	$FA,$8F,$4E,$44,$70,$7F,$A9,$FE,$27,$B1,$1C,$1F,$33,$D8,$F6,$91
	.dc.b	$C1,$FE,$67,$FC,$E9,$F1,$7F,$C2,$5B,$19,$8C,$1F,$D8,$7E,$95,$F3
	.dc.b	$1F,$C5,$65,$18,$7F,$D4,$06,$23,$F8,$D7,$C1,$FF,$83,$7A,$F8,$56
	.dc.b	$EB,$FD,$A8,$D7,$FC,$06,$D7,$FC,$46,$D7,$FC,$58,$F1,$F1,$7F,$DF
	.dc.b	$A3,$F7,$D7,$C9,$FF,$29,$8C,$3E,$0C,$7F,$8C,$67,$FE,$0E,$C3,$FC
	.dc.b	$FE,$D5,$5F,$E7,$78,$C3,$7E,$A3,$8F,$9B,$2B,$8B,$D2,$3A,$FC,$EF
	.dc.b	$6F,$A8,$95,$DA,$32,$F8,$23,$2F,$83,$C9,$FF,$5F,$62,$AF,$9D,$F5
	.dc.b	$9F,$D1,$B2,$57,$F9,$1A,$30,$DF,$1F,$E7,$FA,$7F,$BF,$C7,$E9,$D7
	.dc.b	$F4,$77,$FC,$A8,$D4,$D2,$4F,$2F,$9E,$93,$CA,$3E,$4F,$D4,$C8,$F9
	.dc.b	$2F,$CC,$F6,$71,$1E,$B3,$F8,$56,$89,$CE,$7F,$36,$BE,$9F,$D4,$FD
	.dc.b	$4F,$0E,$61,$F5,$23,$2F,$D6,$8C,$BD,$7C,$76,$2A,$FC,$A7,$E2,$8F
	.dc.b	$3A,$92,$FF,$2B,$8B,$E5,$31,$1F,$E5,$21,$5C,$5F,$F7,$BB,$5D,$35
	.dc.b	$FF,$2C,$B9,$CF,$EA,$C7,$FA,$06,$9F,$27,$FA,$76,$C7,$E9,$AB,$87
	.dc.b	$17,$E0,$3D,$CF,$CA,$FC,$24,$AF,$CA,$C7,$E6,$0C,$3F,$07,$5F,$E9
	.dc.b	$59,$7E,$5E,$7F,$5F,$F9,$4B,$EA,$29,$2F,$FD,$29,$5E,$6F,$01,$4F
	.dc.b	$82,$D5,$7D,$99,$CD,$E2,$F8,$5E,$EB,$F7,$FC,$9F,$DA,$E7,$E0,$C7
	.dc.b	$DF,$F2,$BF,$83,$4E,$75,$FF,$4E,$FA,$ED,$33,$5F,$E6,$DA,$FE,$65
	.dc.b	$7F,$EA,$26,$BF,$EA,$B6,$BD,$4F,$E1,$F4,$FA,$AF,$E2,$8D,$8F,$2D
	.dc.b	$7F,$12,$FC,$2D,$F2,$14,$93,$F1,$32,$FF,$5C,$6C,$AF,$EB,$5A,$7C
	.dc.b	$77,$8D,$6C,$4E,$EF,$9C,$7C,$99,$E3,$7F,$BC,$CB,$F8,$89,$27,$F8
	.dc.b	$8F,$D6,$57,$EE,$45,$C7,$F8,$94,$FF,$18,$F9,$8F,$E6,$C4,$AF,$E7
	.dc.b	$4D,$7F,$09,$F5,$9F,$C7,$8E,$0F,$F1,$EE,$1F,$E4,$24,$7F,$90,$F1
	.dc.b	$57,$FD,$FB,$C7,$F8,$97,$FA,$34,$FF,$45,$3F,$A3,$7D,$5C,$8F,$E9
	.dc.b	$4D,$78,$BF,$CB,$BE,$B2,$47,$F2,$E8,$FE,$63,$EF,$7F,$4F,$24,$7F
	.dc.b	$53,$6B,$F9,$DE,$CF,$E6,$E7,$07,$E8,$77,$3F,$3C,$FF,$64,$EF,$FB
	.dc.b	$08,$C4,$FF,$95,$67,$3C,$3F,$CB,$2F,$13,$FE,$67,$9F,$F5,$E5,$70
	.dc.b	$8F,$E0,$4D,$7F,$DA,$AB,$E8,$BF,$DA,$F7,$FE,$F6,$8C,$4F,$FB,$E2
	.dc.b	$31,$3A,$81,$FE,$2B,$70,$D3,$FD,$EB,$E3,$6A,$BD,$9F,$57,$89,$FF
	.dc.b	$19,$60,$8C,$4F,$82,$7A,$FC,$67,$F4,$77,$BF,$D6,$7F,$85,$FD,$2C
	.dc.b	$E3,$E2,$94,$E5,$57,$17,$3E,$B7,$F8,$0A,$31,$3F,$E1,$A9,$E6,$76
	.dc.b	$9D,$57,$59,$E9,$5C,$27,$35,$E4,$F5,$FD,$8F,$6C,$39,$B4,$F1,$E1
	.dc.b	$CE,$B8,$C7,$A2,$B8,$E1,$DA,$B5,$7C,$A3,$15,$DC,$C5,$F2,$8C,$C7
	.dc.b	$7C,$3D,$B8,$74,$AD,$12,$FA,$46,$5E,$64,$AF,$39,$DC,$A7,$5C,$3C
	.dc.b	$AE,$30,$F0,$9F,$2D,$C5,$F7,$6D,$78,$1E,$D8,$7C,$B8,$B7,$F3,$F0
	.dc.b	$F2,$57,$02,$57,$B1,$D9,$F5,$EA,$BF,$2C,$F5,$1F,$85,$3E,$91,$70
	.dc.b	$FC,$33,$70,$FF,$22,$B9,$F5,$7F,$28,$FF,$69,$94,$62,$35,$9F,$F6
	.dc.b	$D7,$F2,$31,$97,$F1,$0F,$2C,$3F,$1E,$DC,$BF,$31,$E5,$BE,$97,$FB
	.dc.b	$6D,$6B,$0F,$C0,$5E,$F0,$FD,$0B,$55,$F1,$E0,$DD,$49,$FA,$95,$BA
	.dc.b	$D4,$62,$75,$7D,$E3,$EE,$CE,$A3,$04,$C7,$D7,$5F,$9C,$9D,$46,$2B
	.dc.b	$A1,$2B,$F6,$13,$A8,$F4,$D7,$82,$B8,$56,$B1,$F8,$31,$82,$57,$03
	.dc.b	$CB,$1A,$71,$3E,$56,$CF,$CC,$9D,$47,$D4,$AF,$03,$A3,$F8,$CD,$F9
	.dc.b	$5D,$1E,$9A,$F0,$BF,$3A,$B9,$E5,$F8,$2B,$E7,$3C,$30,$F5,$65,$DB
	.dc.b	$6F,$E5,$9F,$1B,$E7,$7F,$9E,$37,$84,$E7,$86,$D7,$F1,$2A,$E2,$FF
	.dc.b	$2E,$3F,$98,$CF,$F8,$C9,$A6,$70,$FE,$5D,$7C,$9F,$F0,$BD,$7F,$4B
	.dc.b	$D7,$FC,$33,$1F,$CC,$F0,$FC,$4D,$FF,$44,$6C,$7F,$44,$6B,$FD,$0D
	.dc.b	$CF,$1F,$F5,$36,$CF,$3F,$83,$FF,$4B,$BD,$6E,$DF,$F3,$EF,$62,$E3
	.dc.b	$FD,$76,$EE,$1F,$F0,$07,$0F,$EC,$2F,$FB,$ED,$B5,$E8,$BE,$51,$E7
	.dc.b	$8C,$5F,$F6,$76,$D3,$E3,$7A,$FA,$1D,$9B,$7D,$9D,$91,$F8,$4F,$2F
	.dc.b	$8D,$91,$F1,$5B,$8F,$C5,$7D,$27,$AB,$EA,$6C,$FA,$BE,$77,$DB,$C9
	.dc.b	$EE,$6C,$64,$FA,$0F,$A0,$8F,$D3,$F6,$BE,$29,$F1,$BD,$AB,$1F,$1A
	.dc.b	$2E,$3F,$E8,$7D,$3F,$DB,$33,$93,$FC,$75,$E4,$FF,$06,$9F,$A6,$D8
	.dc.b	$CF,$0F,$F7,$16,$CE,$A3,$F1,$9C,$5F,$E6,$B6,$CF,$F9,$29,$E1,$3F
	.dc.b	$E6,$B7,$2F,$F8,$13,$EA,$9F,$59,$F8,$2B,$FF,$2A,$F9,$7F,$84,$A5
	.dc.b	$6B,$0F,$F2,$85,$69,$FF,$77,$C6,$67,$FD,$34,$D7,$F9,$E3,$5F,$F0
	.dc.b	$13,$5F,$15,$F7,$8E,$55,$FE,$EC,$B7,$C2,$66,$BE,$2F,$E5,$6F,$81
	.dc.b	$A7,$C8,$F8,$D7,$85,$8D,$3E,$6D,$6D,$B5,$B9,$D6,$9F,$57,$EB,$7F
	.dc.b	$3F,$8F,$E8,$1B,$7F,$A0,$6C,$EE,$3F,$47,$3F,$1D,$FF,$46,$1F,$1B
	.dc.b	$F6,$B7,$D0,$67,$D6,$F9,$D1,$89,$FF,$3D,$5F,$FB,$B2,$7F,$54,$D8
	.dc.b	$FE,$B0,$74,$7F,$D8,$CD,$7F,$C8,$71,$D6,$3C,$1A,$59,$FC,$87,$D0
	.dc.b	$7E,$EB,$EB,$7F,$76,$DA,$FD,$DB,$67,$71,$FB,$E9,$FE,$03,$F3,$B2
	.dc.b	$3F,$80,$FC,$E5,$AF,$CE,$4D,$E3,$FE,$21,$36,$BF,$C8,$1F,$B2,$FF
	.dc.b	$74,$7E,$9F,$C0,$7E,$36,$31,$87,$FA,$FF,$8F,$FD,$AF,$5F,$12,$73
	.dc.b	$3F,$E0,$56,$BD,$3C,$BF,$8F,$18,$8F,$F2,$FB,$D3,$1F,$F6,$1B,$6B
	.dc.b	$FD,$5B,$6F,$99,$F1,$61,$FC,$7B,$E4,$FF,$8F,$FD,$6D,$BF,$E3,$ED
	.dc.b	$CB,$F7,$4D,$CB,$F9,$73,$FC,$A3,$F1,$25,$53,$1C,$35,$FF,$44,$C6
	.dc.b	$9A,$68,$A7,$E1,$3E,$BF,$FD,$8E,$FE,$C3,$F7,$EF,$56,$3F,$E7,$C9
	.dc.b	$CA,$C7,$FE,$00,$89,$DC,$27,$6F,$F8,$BD,$AF,$F9,$D5,$AF,$5E,$CF
	.dc.b	$C9,$C7,$6A,$FF,$76,$B6,$FE,$5F,$C3,$FF,$07,$8C,$5B,$CB,$65,$F9
	.dc.b	$F8,$FF,$4D,$B5,$FF,$4D,$35,$FF,$49,$35,$FF,$4C,$F6,$3D,$3F,$D2
	.dc.b	$B7,$1F,$F1,$E3,$67,$FC,$94,$7F,$02,$9F,$EA,$35,$F8,$5B,$F1,$F0
	.dc.b	$60,$98,$7F,$AC,$27,$95,$F9,$2F,$89,$27,$BD,$78,$88,$E0,$E5,$F5
	.dc.b	$DE,$02,$38,$3F,$CA,$5F,$5A,$F7,$7F,$7E,$D8,$CC,$60,$FE,$07,$E2
	.dc.b	$FF,$27,$C7,$FD,$FF,$E0,$60,$95,$FE,$5D,$FA,$C7,$C2,$7F,$0A,$6C
	.dc.b	$66,$30,$7F,$0E,$FD,$7F,$FC,$9F,$7F,$E7,$16,$BF,$EB,$76,$BF,$EC
	.dc.b	$A6,$AC,$1E,$67,$98,$97,$FE,$6F,$8C,$BF,$E7,$1F,$AC,$7C,$77,$F1
	.dc.b	$E6,$C6,$63,$07,$F8,$23,$EB,$1D,$54,$93,$FE,$E6,$8C,$BF,$EE,$8F
	.dc.b	$8F,$76,$FF,$84,$1B,$19,$8C,$19,$ED,$91,$C5,$BA,$7F,$BC,$3E,$A5
	.dc.b	$F5,$5F,$E1,$A6,$D7,$FB,$D9,$B3,$B9,$FF,$0F,$FD,$5F,$E2,$A3,$FD
	.dc.b	$D8,$D7,$F4,$8D,$EA,$34,$FC,$D7,$77,$F0,$A9,$DF,$57,$F8,$D2,$5D
	.dc.b	$36,$3F,$C6,$FF,$24,$F2,$7F,$C7,$3F,$54,$F8,$CC,$C6,$1B,$B7,$E9
	.dc.b	$B4,$5F,$18,$C3,$73,$FF,$5A,$4E,$AF,$FC,$77,$39,$D3,$CD,$39,$B5
	.dc.b	$FF,$13,$6C,$3E,$AD,$E6,$78,$AB,$47,$F0,$7C,$1E,$5F,$AB,$E2,$BD
	.dc.b	$AD,$C7,$AE,$5E,$0A,$DC,$7A,$27,$CA,$47,$DD,$F5,$7F,$D3,$30,$F5
	.dc.b	$B5,$F3,$9F,$C5,$DC,$3E,$86,$B9,$F3,$7F,$CC,$B1,$CF,$0F,$82,$BC
	.dc.b	$3C,$CF,$F8,$A6,$BE,$1F,$C1,$FC,$4F,$9A,$FF,$30,$47,$AB,$3F,$BD
	.dc.b	$18,$8F,$1C,$EF,$DF,$F2,$E5,$F5,$B3,$C5,$7F,$B9,$EB,$E7,$FC,$C7
	.dc.b	$1A,$D4,$A3,$A3,$E2,$31,$7D,$31,$FC,$74,$EA,$57,$D2,$B5,$85,$9F
	.dc.b	$01,$C8,$F5,$39,$7D,$26,$4F,$EA,$97,$DA,$7F,$9A,$D6,$F0,$F3,$E1
	.dc.b	$E1,$8F,$6D,$F0,$C2,$BE,$9C,$7D,$D8,$FC,$8C,$79,$CF,$7C,$79,$C6
	.dc.b	$31,$DE,$1E,$5A,$F8,$51,$87,$CC,$62,$7D,$EE,$0F,$86,$7F,$4C,$3E
	.dc.b	$51,$71,$C3,$83,$F1,$7F,$4E,$F5,$1F,$57,$C6,$30,$D8,$CC,$FF,$58
	.dc.b	$9C,$B7,$C2,$FE,$F8,$DC,$F9,$46,$23,$F3,$22,$7F,$96,$CE,$8F,$A9
	.dc.b	$8F,$FB,$6A,$6F,$C3,$FE,$83,$3F,$C8,$5F,$09,$FB,$0D,$8F,$CA,$57
	.dc.b	$19,$FB,$8D,$8E,$38,$79,$9B,$87,$DA,$9E,$1B,$72,$C7,$F5,$EE,$B0
	.dc.b	$36,$F5,$1E,$63,$F1,$98,$F8,$EF,$E6,$E3,$DD,$FD,$C9,$E3,$2E,$1E
	.dc.b	$D7,$78,$F2,$9E,$17,$FC,$5A,$78,$36,$B8,$3F,$C6,$AF,$C0,$DA,$E2
	.dc.b	$FF,$B6,$67,$6D,$3B,$5F,$E9,$F9,$9F,$56,$F8,$3F,$DD,$F0,$7F,$24
	.dc.b	$8C,$5F,$F1,$C8,$C4,$EA,$3F,$C5,$53,$FD,$B9,$D1,$88,$D5,$76,$6C
	.dc.b	$F4,$9F,$E5,$B3,$E2,$F2,$37,$F5,$6F,$0F,$F3,$92,$8C,$57,$F4,$56
	.dc.b	$BF,$D0,$9A,$FF,$28,$8E,$13,$E2,$F9,$77,$35,$3B,$C7,$9B,$EA,$DF
	.dc.b	$11,$49,$87,$B3,$04,$BE,$23,$33,$FD,$4A,$38,$BF,$D6,$5A,$E6,$BD
	.dc.b	$71,$8D,$FF,$AA,$FD,$CB,$EC,$3E,$3B,$6F,$C1,$5B,$E1,$FE,$AF,$F5
	.dc.b	$9F,$AC,$C3,$FA,$EA,$F8,$FF,$9E,$DB,$F4,$3B,$AE,$33,$F3,$23,$53
	.dc.b	$FD,$A9,$AF,$C9,$6B,$FD,$9B,$EB,$5F,$37,$F7,$1B,$1F,$15,$AF,$DC
	.dc.b	$9E,$B8,$7F,$72,$F3,$7F,$A9,$4F,$CB,$7F,$C8,$6D,$F5,$7F,$0F,$1A
	.dc.b	$8F,$C2,$EF,$FB,$27,$FC,$08,$D7,$FC,$02,$D7,$F0,$0D,$7F,$BE,$FD
	.dc.b	$6F,$E1,$AF,$FC,$06,$DD,$33,$C5,$F9,$6F,$C1,$7F,$C3,$97,$FE,$00
	.dc.b	$1F,$56,$FA,$9F,$D0,$43,$EE,$7F,$AE,$61,$FD,$D9,$B1,$FE,$0E,$AE
	.dc.b	$0F,$FB,$AB,$EB,$7F,$C9,$AF,$F3,$76,$BF,$E8,$46,$BF,$CA,$1A,$FF
	.dc.b	$85,$BE,$E9,$F3,$DF,$AC,$6C,$7E,$55,$AF,$EB,$17,$CF,$FA,$3E,$A3
	.dc.b	$FC,$65,$3F,$B6,$DD,$FB,$EB,$E0,$BF,$79,$18,$78,$7A,$9F,$03,$EE
	.dc.b	$9F,$31,$FD,$31,$B8,$7F,$51,$6B,$FE,$9E,$6B,$FE,$AA,$F7,$3F,$81
	.dc.b	$FC,$3F,$F9,$62,$3F,$D9,$F1,$8E,$8C,$FD,$7E,$BC,$19,$5F,$ED,$48
	.dc.b	$E3,$F1,$7F,$C1,$30,$34,$93,$FE,$A7,$F8,$EE,$E4,$7F,$DB,$BE,$37
	.dc.b	$8B,$D6,$F1,$F7,$C9,$FF,$70,$C7,$9B,$E5,$BB,$12,$3F,$D6,$4F,$87
	.dc.b	$E3,$BD,$64,$8F,$F7,$4F,$D9,$7E,$1B,$EE,$9F,$2F,$49,$3F,$C9,$BE
	.dc.b	$2D,$F9,$9F,$D8,$E8,$FC,$8D,$F0,$AF,$EF,$8D,$7F,$D9,$2D,$7F,$B6
	.dc.b	$B5,$FF,$23,$BB,$8F,$F3,$B5,$76,$F8,$FF,$E7,$0D,$AF,$E6,$E2,$B8
	.dc.b	$78,$5F,$ED,$EF,$A4,$FA,$E4,$9F,$E4,$5C,$7F,$9C,$7F,$9C,$A8,$FB
	.dc.b	$BF,$35,$27,$71,$FD,$0B,$D6,$7C,$9D,$24,$FF,$2D,$8F,$DE,$69,$FE
	.dc.b	$78,$77,$19,$FB,$0E,$47,$07,$FC,$EB,$A7,$CA,$FA,$AF,$6E,$E6,$98
	.dc.b	$F8,$6B,$5F,$EF,$2A,$E0,$DA,$E0,$DA,$FF,$63,$46,$23,$BA,$FA,$AE
	.dc.b	$E7,$48,$FE,$B7,$8B,$E9,$FE,$01,$23,$DD,$F6,$15,$DF,$D6,$FC,$53
	.dc.b	$1A,$FC,$1F,$88,$F9,$04,$9F,$8D,$F2,$75,$EF,$F0,$5F,$34,$C5,$7F
	.dc.b	$1A,$FB,$0F,$E1,$AD,$8F,$E3,$31,$8E,$6F,$F4,$6E,$1F,$6E,$B5,$7F
	.dc.b	$E4,$BA,$FE,$22,$FF,$AA,$1A,$9A,$73,$7F,$D5,$B2,$E5,$FE,$4E,$D8
	.dc.b	$FC,$7C,$EF,$DE,$FF,$08,$7C,$47,$E1,$49,$3F,$83,$FB,$0A,$FC,$CF
	.dc.b	$73,$F1,$46,$32,$FC,$EF,$1B,$9F,$F0,$24,$AF,$D7,$43,$87,$11,$F6
	.dc.b	$B8,$BF,$0E,$62,$FF,$A0,$F1,$3C,$6F,$E7,$ED,$AF,$D3,$B5,$FE,$8D
	.dc.b	$F0,$3F,$83,$73,$3F,$E7,$09,$D4,$7F,$98,$9A,$FF,$AF,$1A,$FF,$B5
	.dc.b	$1E,$95,$FD,$4A,$9D,$3F,$D4,$1B,$5D,$8F,$D4,$7C,$87,$EB,$48,$FE
	.dc.b	$AF,$A6,$BF,$94,$E9,$FD,$91,$8C,$3F,$5D,$87,$ED,$49,$3F,$B3,$EF
	.dc.b	$AF,$E7,$3D,$4F,$DD,$18,$AF,$DB,$7A,$7F,$E3,$6E,$CF,$E9,$F5,$FE
	.dc.b	$F9,$6B,$FE,$F6,$6B,$FE,$35,$73,$3F,$BF,$FA,$DF,$E0,$C7,$F0,$7A
	.dc.b	$FE,$11,$F5,$B2,$FF,$83,$3F,$E4,$CF,$C7,$FF,$C3,$90,$E9,$FF,$2D
	.dc.b	$4E,$E3,$F6,$6E,$DE,$1F,$05,$FE,$4C,$AF,$F0,$EC,$71,$6D,$1A,$7F
	.dc.b	$86,$BB,$93,$BF,$89,$F1,$1D,$1F,$F2,$34,$BE,$9F,$B1,$23,$FD,$46
	.dc.b	$77,$70,$FF,$98,$A1,$F5,$9E,$D3,$DD,$3E,$FF,$AC,$95,$D9,$D3,$48
	.dc.b	$73,$72,$FF,$9B,$BE,$AB,$FC,$7F,$8F,$F9,$95,$B1,$EC,$C9,$18,$8F
	.dc.b	$98,$D8,$CE,$7F,$0A,$7F,$DE,$7D,$5C,$73,$7C,$7E,$7E,$4F,$F9,$15
	.dc.b	$7C,$D8,$C5,$77,$C7,$FC,$A3,$5A,$A7,$D7,$3E,$6B,$FA,$0C,$E6,$31
	.dc.b	$E5,$FD,$28,$C4,$EB,$0F,$25,$7F,$53,$BD,$46,$27,$51,$F6,$A7,$9B
	.dc.b	$EA,$31,$1A,$7C,$7E,$67,$A1,$CE,$1D,$1A,$71,$3D,$33,$C6,$BC,$51
	.dc.b	$97,$BE,$1E,$8A,$E5,$3C,$1B,$3C,$27,$53,$C6,$BA,$1D,$6F,$C9,$1C
	.dc.b	$4D,$5F,$29,$D4,$58,$8E,$8F,$24,$F2,$BF,$2E,$1F,$1E,$B7,$87,$D1
	.dc.b	$AD,$61,$E2,$9C,$D7,$D0,$9C,$C6,$30,$F8,$B7,$C5,$C1,$E7,$7D,$A6
	.dc.b	$5F,$23,$8A,$F1,$47,$98,$F4,$9F,$91,$3E,$01,$70,$FB,$6D,$C3,$FC
	.dc.b	$12,$E7,$E8,$FF,$7F,$26,$93,$5F,$DA,$CF,$E4,$A3,$FD,$55,$87,$E4
	.dc.b	$5B,$87,$E8,$78,$5F,$5A,$E1,$FB,$35,$AC,$3F,$1C,$D7,$A6,$3F,$80
	.dc.b	$BE,$35,$C2,$24,$FE,$1E,$B8,$56,$A3,$13,$AC,$7C,$8F,$3C,$3C,$53
	.dc.b	$A2,$4E,$9F,$CB,$57,$38,$C1,$31,$FD,$4E,$1E,$0B,$E7,$39,$BF,$5E
	.dc.b	$1E,$1B,$F8,$55,$A7,$8C,$66,$76,$78,$27,$4E,$2B,$A3,$F1,$0D,$63
	.dc.b	$C5,$C4,$6B,$1F,$46,$1D,$A7,$DA,$D9,$FE,$15,$3A,$9E,$D3,$92,$3F
	.dc.b	$CE,$A3,$4F,$68,$FE,$29,$C9,$E6,$E2,$FE,$85,$F2,$17,$85,$39,$E3
	.dc.b	$4E,$63,$E8,$E7,$FC,$B5,$B8,$F4,$8F,$E9,$26,$4A,$FF,$29,$F8,$CF
	.dc.b	$B3,$D5,$F6,$B2,$FE,$A5,$18,$AF,$E5,$DD,$76,$FF,$AC,$4D,$AD,$D7
	.dc.b	$F5,$28,$C6,$4B,$B1,$8F,$92,$CC,$5A,$FE,$B2,$DC,$3F,$5D,$87,$F6
	.dc.b	$3A,$FF,$72,$C7,$F3,$F6,$BF,$DA,$5B,$1F,$DA,$78,$3C,$3F,$03,$E2
	.dc.b	$6E,$1F,$D9,$9D,$E5,$F9,$DC,$3F,$A6,$5F,$1A,$FE,$CE,$D8,$F1,$CB
	.dc.b	$EB,$7D,$A7,$BB,$D6,$B7,$FD,$CA,$2D,$7F,$6D,$9B,$AF,$CD,$BF,$F7
	.dc.b	$05,$FF,$78,$F8,$3F,$8F,$F0,$9F,$21,$B1,$E2,$9D,$E1,$E2,$AF,$49
	.dc.b	$EA,$F9,$BF,$9F,$F5,$5F,$41,$B7,$FE,$04,$70,$FF,$81,$BE,$72,$DF
	.dc.b	$F8,$07,$E8,$7E,$BF,$C8,$FD,$86,$D7,$F8,$31,$C3,$FE,$12,$D3,$F0
	.dc.b	$DF,$A1,$7F,$11,$8D,$1F,$EB,$CF,$AE,$5F,$E1,$E9,$F7,$CF,$9E,$FC
	.dc.b	$56,$4B,$DE,$1D,$4E,$D5,$B7,$A6,$1F,$80,$AE,$0D,$AE,$31,$FE,$1E
	.dc.b	$6B,$FC,$DD,$CE,$1F,$C5,$A3,$8D,$77,$51,$E6,$F0,$FC,$B5,$C2,$BF
	.dc.b	$C6,$3E,$FB,$F3,$57,$C4,$7E,$8E,$31,$F1,$16,$73,$5F,$EE,$F5,$19
	.dc.b	$A9,$39,$F8,$8B,$9F,$8A,$3F,$C4,$9B,$BE,$9F,$F5,$CB,$EC,$2C,$61
	.dc.b	$B3,$EB,$EB,$57,$18,$C3,$FD,$45,$B5,$EE,$BF,$AD,$F6,$1E,$69,$FB
	.dc.b	$51,$8C,$F5,$1F,$D5,$B6,$FE,$13,$3F,$34,$FA,$44,$8E,$27,$CD,$E0
	.dc.b	$5B,$4E,$75,$EF,$3F,$CD,$59,$7F,$8C,$6F,$CF,$F4,$FF,$03,$CC,$F8
	.dc.b	$2D,$8F,$5C,$7C,$4F,$91,$C5,$F3,$DB,$E0,$FC,$8F,$C8,$F7,$BE,$4B
	.dc.b	$70,$F8,$6E,$1F,$8B,$F2,$36,$FF,$DF,$BC,$9F,$25,$BE,$7F,$DD,$E3
	.dc.b	$FB,$CD,$BF,$F7,$E5,$6E,$BF,$CA,$DC,$DF,$87,$BF,$E2,$ED,$AD,$C7
	.dc.b	$19,$EF,$CB,$78,$FF,$5B,$F2,$3F,$80,$F9,$1F,$C0,$B7,$0F,$B6,$E1
	.dc.b	$FB,$BD,$17,$0F,$85,$C9,$F3,$E7,$F0,$3D,$1F,$8A,$E6,$FC,$5B,$63
	.dc.b	$F0,$B1,$F8,$EF,$91,$C5,$FD,$51,$F7,$A8,$FC,$F5,$7E,$EA,$30,$FE
	.dc.b	$7C,$C4,$7E,$DE,$33,$3F,$E9,$26,$BF,$E2,$B6,$BF,$E7,$D7,$7A,$7E
	.dc.b	$DD,$B5,$FB,$2F,$37,$F6,$95,$FE,$A4,$9C,$BF,$8A,$F9,$2F,$D3,$9F
	.dc.b	$A8,$F9,$2B,$7F,$A2,$F5,$FF,$A3,$EA,$69,$23,$FA,$0C,$3E,$6B,$F6
	.dc.b	$3F,$1F,$FB,$26,$CF,$EB,$FF,$43,$78,$5F,$D8,$DB,$19,$F1,$71,$39
	.dc.b	$F8,$BF,$EA,$F4,$69,$2B,$FA,$8C,$3D,$7E,$9F,$15,$D5,$B3,$E1,$F8
	.dc.b	$AF,$01,$E1,$F9,$0B,$1F,$C7,$78,$7F,$61,$F1,$3F,$CA,$79,$CD,$24
	.dc.b	$FF,$94,$BE,$B9,$C7,$FC,$E7,$EF,$5E,$2F,$F3,$89,$CD,$7E,$F6,$72
	.dc.b	$FC,$3F,$94,$F8,$47,$C3,$F9,$4B,$1E,$3F,$AF,$FE,$E1,$A7,$FA,$ED
	.dc.b	$AF,$FB,$79,$A9,$BE,$23,$DC,$FE,$99,$F7,$AF,$53,$FA,$5C,$E7,$0F
	.dc.b	$E2,$13,$97,$F9,$26,$3F,$88,$F3,$0D,$FA,$FF,$EB,$59,$FF,$09,$8D
	.dc.b	$7D,$17,$F7,$08,$FF,$02,$B5,$FF,$6C,$35,$52,$30,$BE,$F7,$F6,$53
	.dc.b	$FC,$55,$FE,$5C,$DE,$4F,$F1,$54,$63,$C4,$FF,$19,$79,$EF,$4B,$FC
	.dc.b	$5C,$DD,$3F,$9F,$7D,$87,$F3,$BC,$7F,$90,$4E,$5F,$E4,$51,$97,$FC
	.dc.b	$3A,$D7,$FD,$E0,$D7,$D2,$DC,$8F,$F4,$FA,$EF,$E8,$7F,$8F,$E3,$15
	.dc.b	$DA,$3F,$C8,$5F,$5F,$FE,$16,$CB,$FC,$95,$F0,$AF,$9A,$CF,$BD,$FE
	.dc.b	$47,$8C,$65,$FD,$5E,$FE,$39,$27,$FC,$76,$D8,$D4,$76,$F8,$0E,$71
	.dc.b	$FD,$7F,$BD,$EF,$FF,$94,$FE,$C1,$F2,$DF,$E5,$59,$CD,$7F,$65,$9C
	.dc.b	$BC,$6B,$FC,$93,$3A,$DB,$FC,$A2,$D7,$FC,$8A,$D4,$7E,$03,$EC,$1E
	.dc.b	$AF,$F9,$8D,$B7,$FD,$CD,$7A,$D9,$99,$39,$78,$46,$2B,$F9,$81,$DB
	.dc.b	$E3,$3B,$FD,$7B,$E4,$3C,$33,$FE,$18,$CB,$FD,$63,$39,$8F,$F6,$B5
	.dc.b	$70,$AD,$F1,$AF,$71,$FE,$DB,$A5,$6D,$FF,$1F,$4E,$9E,$62,$FB,$3E
	.dc.b	$1C,$B8,$3E,$4A,$B8,$77,$9C,$CF,$AC,$E5,$87,$93,$0E,$D1,$C4,$C9
	.dc.b	$E2,$9E,$2F,$98,$C9,$E1,$29,$D7,$CF,$F2,$DF,$86,$3F,$CA,$17,$EF
	.dc.b	$C3,$EC,$D6,$B0,$FF,$6C,$BE,$CA,$F3,$63,$9C,$7C,$37,$EC,$BF,$3D
	.dc.b	$6B,$0F,$55,$6A,$3C,$B8,$F8,$70,$E9,$1E,$93,$FD,$FF,$87,$FC,$00
	.dc.b	$7E,$21,$83,$EA,$7D,$7F,$EA,$7E,$C3,$EB,$42,$30,$EB,$99,$3A,$9C
	.dc.b	$CF,$E1,$8E,$1D,$FF,$E5,$E7,$AB,$5D,$47,$F2,$4C,$79,$4E,$67,$F1
	.dc.b	$F3,$97,$E8,$E7,$FE,$D4,$7F,$02,$D7,$EA,$C6,$F2,$FC,$3B,$67,$F3
	.dc.b	$26,$E3,$F1,$6D,$3B,$D7,$9E,$38,$BE,$9D,$77,$EC,$7E,$89,$B8,$7D
	.dc.b	$CB,$E3,$8D,$8F,$DB,$E3,$E1,$7F,$73,$3D,$DB,$7F,$AC,$C7,$BD,$FE
	.dc.b	$72,$F8,$57,$EC,$6B,$57,$EA,$6C,$76,$C7,$D1,$3E,$28,$F8,$F3,$A7
	.dc.b	$91,$D6,$76,$F1,$BF,$97,$3B,$AF,$34,$FD,$A9,$FD,$B7,$5B,$EC,$1F
	.dc.b	$19,$FB,$EE,$0F,$E2,$B3,$9F,$8D,$C6,$1F,$D5,$AD,$18,$9E,$57,$67
	.dc.b	$9D,$7F,$A2,$DD,$B6,$B5,$3C,$74,$FF,$5A,$5F,$E5,$70,$FC,$BB,$6F
	.dc.b	$FD,$D7,$3A,$3F,$D6,$58,$AB,$CF,$F3,$BF,$E8,$5A,$FF,$4D,$35,$FF
	.dc.b	$3E,$57,$09,$F6,$FD,$73,$E4,$3D,$4D,$AF,$F5,$13,$5F,$F4,$23,$FE
	.dc.b	$09,$73,$1E,$68,$D4,$7F,$2C,$C5,$18,$8E,$F3,$B3,$E4,$7D,$7F,$F4
	.dc.b	$08,$FF,$58,$35,$EF,$1F,$37,$EB,$BF,$49,$9F,$BE,$7F,$C0,$F0,$F9
	.dc.b	$5D,$1B,$E1,$B3,$FC,$D7,$25,$6E,$30,$D7,$F7,$D1,$AA,$FA,$3F,$5E
	.dc.b	$FA,$1F,$BC,$DA,$FF,$63,$35,$FF,$54,$5F,$3C,$7F,$C2,$3F,$09,$F8
	.dc.b	$5F,$23,$FD,$01,$5F,$EC,$D6,$BF,$E1,$0F,$83,$FC,$57,$CB,$7F,$96
	.dc.b	$E3,$D0,$D7,$CA,$D7,$E7,$35,$FC,$BF,$D7,$FF,$B7,$DF,$F6,$D3,$FE
	.dc.b	$DF,$6F,$C4,$FE,$5E,$35,$1F,$EB,$FF,$45,$F1,$BF,$9B,$6C,$7F,$B8
	.dc.b	$DA,$EF,$1F,$D1,$AF,$BD,$F5,$BF,$C9,$DA,$FF,$6E,$6B,$FE,$00,$6A
	.dc.b	$FA,$C7,$97,$FA,$ED,$9F,$58,$F3,$70,$DB,$FF,$3A,$5F,$07,$FD,$F1
	.dc.b	$F5,$AF,$A3,$FD,$A3,$63,$31,$8F,$A2,$FD,$C7,$C3,$FF,$8B,$67,$FD
	.dc.b	$0F,$E4,$FE,$C2,$31,$1F,$84,$6B,$FD,$0F,$EB,$FF,$2D,$A7,$E4,$C7
	.dc.b	$E3,$1F,$AF,$70,$DF,$07,$FC,$1F,$3C,$1F,$E1,$B6,$FA,$BF,$19,$30
	.dc.b	$FF,$52,$7D,$5F,$9B,$B7,$C8,$48,$FF,$4E,$BE,$7F,$88,$F5,$FA,$D2
	.dc.b	$7F,$D5,$91,$EE,$FA,$A9,$1F,$EA,$F7,$E0,$FD,$57,$C2,$FA,$AF,$80
	.dc.b	$49,$FF,$54,$BF,$0F,$EA,$BE,$3F,$1C,$9F,$F5,$C4,$7E,$0B,$EB,$9F
	.dc.b	$4D,$49,$3F,$E3,$EE,$3F,$8D,$4F,$A3,$CC,$FF,$8A,$6F,$8B,$FD,$D9
	.dc.b	$AE,$6F,$FB,$F4,$62,$F9,$F5,$FE,$1F,$0F,$C1,$5F,$FB,$BE,$B5,$C7
	.dc.b	$F3,$B9,$FE,$F9,$27,$C9,$1F,$CF,$B9,$FF,$BF,$63,$FC,$37,$C8,$73
	.dc.b	$30,$FE,$7D,$4F,$A8,$A7,$3C,$58,$71,$9F,$E2,$F8,$F5,$F2,$B8,$9C
	.dc.b	$BC,$AB,$7E,$23,$FD,$43,$4F,$AB,$EE,$6A,$3A,$DF,$F9,$A3,$E2,$5F
	.dc.b	$93,$38,$3F,$AA,$C6,$1B,$19,$AE,$D3,$C3,$5C,$D7,$F8,$92,$38,$4F
	.dc.b	$F9,$6D,$AF,$F8,$85,$AF,$F9,$A1,$CE,$9F,$CC,$DF,$1F,$27,$80,$EB
	.dc.b	$3E,$DE,$0F,$F4,$D7,$D6,$7C,$A2,$46,$75,$FF,$27,$E9,$F4,$E7,$E7
	.dc.b	$FC,$24,$AD,$47,$D2,$F2,$E4,$71,$7E,$AF,$A1,$88,$F8,$E7,$CC,$F8
	.dc.b	$DF,$B2,$67,$5F,$E9,$50,$FA,$E7,$CA,$FF,$1C,$6C,$7F,$C0,$16,$BD
	.dc.b	$07,$AE,$F7,$B8,$57,$F9,$4E,$B7,$5F,$E4,$96,$BF,$E9,$06,$BF,$EA
	.dc.b	$27,$AD,$F9,$2F,$FA,$03,$7E,$63,$AD,$7C,$7F,$6B,$F2,$64,$7E,$57
	.dc.b	$B7,$5F,$83,$C5,$FE,$C9,$EF,$FC,$C9,$9C,$3E,$C7,$47,$FB,$EA,$BF
	.dc.b	$0A,$D7,$E4,$F3,$7E,$3B,$DE,$FC,$5F,$93,$F8,$C7,$F1,$BC,$F2,$B8
	.dc.b	$C7,$E4,$38,$64,$74,$7F,$27,$F2,$F8,$8F,$C6,$63,$F2,$5F,$C8,$35
	.dc.b	$CC,$FE,$56,$31,$97,$A3,$7F,$CE,$9C,$DF,$D8,$7D,$5F,$ED,$89,$7F
	.dc.b	$8D,$F4,$6B,$F1,$33,$FE,$0F,$5F,$B7,$F1,$32,$FE,$E7,$EA,$BF,$7E
	.dc.b	$4A,$FC,$B7,$C4,$D7,$97,$C9,$FF,$04,$31,$5F,$C0,$7C,$8F,$F1,$CE
	.dc.b	$3F,$C7,$EB,$FC,$24,$D7,$FC,$1E,$D7,$FC,$50,$E6,$7F,$85,$7D,$4F
	.dc.b	$F1,$03,$F8,$8D,$7F,$93,$FF,$20,$FF,$1D,$5B,$99,$1F,$DF,$2F,$F8
	.dc.b	$D3,$7C,$1F,$F2,$DB,$C2,$77,$1C,$FE,$96,$5F,$ED,$EB,$47,$F8,$25
	.dc.b	$FF,$12,$B5,$FF,$1F,$35,$EB,$EE,$FF,$96,$23,$86,$1E,$2F,$90,$F4
	.dc.b	$1E,$99,$F5,$7D,$46,$C9,$7F,$E3,$7B,$46,$23,$F3,$70,$F9,$3F,$84
	.dc.b	$72,$8F,$43,$86,$9E,$B9,$C3,$CB,$DF,$F9,$46,$31,$F8,$FE,$F2,$66
	.dc.b	$9E,$E8,$C4,$7F,$08,$D7,$E7,$C6,$27,$53,$C0,$F9,$F1,$FE,$FB,$EA
	.dc.b	$FA,$3F,$39,$67,$EC,$68,$8C,$61,$D2,$7C,$5F,$1A,$BF,$0F,$D7,$7C
	.dc.b	$4F,$A0,$CC,$63,$64,$62,$FD,$55,$B5,$7C,$DC,$FD,$27,$8D,$4E,$E7
	.dc.b	$51,$8B,$EA,$F2,$C3,$C1,$87,$78,$E3,$87,$3C,$39,$56,$A7,$81,$E4
	.dc.b	$BE,$78,$76,$A5,$F1,$9C,$DF,$1A,$D4,$62,$FD,$27,$86,$FB,$BD,$70
	.dc.b	$C4,$EE,$3C,$A6,$22,$3E,$43,$DC,$F3,$C3,$C4,$EA,$FC,$2E,$67,$18
	.dc.b	$79,$48,$F9,$0E,$98,$79,$1E,$A6,$30,$F2,$47,$C0,$8E,$67,$A4,$D5
	.dc.b	$77,$69,$2F,$C6,$F6,$C3,$CE,$E7,$0F,$1B,$F0,$0E,$27,$E6,$8F,$19
	.dc.b	$6F,$F0,$CD,$C3,$EF,$B9,$E2,$FF,$10,$3C,$4F,$C3,$B8,$75,$EA,$65
	.dc.b	$CB,$69,$3C,$B1,$FC,$03,$70,$FD,$6F,$A7,$FE,$F2,$E0,$FB,$93,$9C
	.dc.b	$7F,$36,$D7,$A9,$F6,$67,$83,$F4,$23,$58,$78,$31,$E6,$7E,$5E,$B7
	.dc.b	$8F,$8A,$33,$7C,$5D,$78,$FF,$EC,$F9,$D4,$E4,$8F,$E9,$AB,$9E,$5F
	.dc.b	$B6,$C7,$C2,$4C,$39,$E5,$DA,$F9,$92,$BE,$0E,$3D,$DE,$55,$B6,$E3
	.dc.b	$EF,$9D,$3F,$12,$74,$D3,$57,$87,$E3,$DF,$77,$F8,$F6,$3E,$87,$84
	.dc.b	$65,$DE,$1E,$4A,$E5,$97,$CB,$9F,$83,$39,$69,$FC,$73,$EA,$7C,$BB
	.dc.b	$FF,$54,$F9,$4B,$C4,$9B,$18,$7F,$DF,$98,$FF,$4C,$6D,$FF,$58,$6B
	.dc.b	$D6,$3F,$97,$3F,$CA,$BE,$53,$FD,$65,$1F,$D7,$BE,$A2,$CE,$63,$FA
	.dc.b	$E4,$63,$1F,$EC,$F5,$71,$37,$E6,$F0,$DB,$FE,$C8,$DC,$3F,$A9,$B5
	.dc.b	$FE,$C7,$3F,$E4,$C7,$FB,$7B,$5F,$E7,$ED,$BF,$EE,$FF,$51,$E1,$F8
	.dc.b	$DF,$13,$70,$FE,$E4,$EF,$13,$49,$1F,$D9,$FE,$37,$D5,$F4,$7E,$B6
	.dc.b	$C7,$F7,$B6,$E5,$FD,$A9,$BD,$7F,$E0,$6E,$8F,$85,$F4,$BF,$0D,$B7
	.dc.b	$FE,$03,$7C,$44,$7F,$BB,$3E,$5F,$A6,$F9,$9F,$8C,$7F,$52,$9F,$F0
	.dc.b	$54,$63,$B3,$E7,$32,$3F,$D0,$93,$FE,$69,$76,$63,$CC,$E7,$CD,$D1
	.dc.b	$B1,$92,$69,$FE,$5E,$A7,$D3,$EE,$7F,$D3,$FF,$94,$B4,$FC,$1F,$A6
	.dc.b	$F1,$7C,$D8,$FF,$97,$AB,$57,$FE,$A6,$9C,$EF,$FE,$9D,$9C,$D7,$FA
	.dc.b	$8D,$AF,$FA,$8D,$AF,$FA,$D5,$AA,$D1,$CA,$7A,$61,$E4,$86,$B6,$31
	.dc.b	$D1,$9E,$3B,$9F,$B1,$B9,$7A,$67,$97,$3E,$32,$FF,$6D,$6D,$FE,$9C
	.dc.b	$C3,$E7,$FC,$EF,$D6,$F6,$7E,$BB,$63,$E9,$C7,$DA,$FA,$6C,$65,$FE
	.dc.b	$9F,$FA,$D7,$83,$FA,$CC,$9D,$1F,$E3,$6C,$7F,$DB,$56,$BD,$C7,$FB
	.dc.b	$76,$B7,$E2,$FE,$D2,$B5,$87,$E9,$5B,$5F,$D2,$6F,$8C,$7F,$4F,$6B
	.dc.b	$FD,$65,$CD,$FF,$AE,$9B,$3F,$E3,$98,$F2,$F8,$BD,$E7,$35,$FE,$67
	.dc.b	$D5,$EB,$7E,$63,$CE,$FC,$CB,$71,$FC,$AB,$87,$F2,$FE,$BD,$BF,$C8
	.dc.b	$7C,$77,$E8,$3E,$3B,$F4,$2D,$AF,$CE,$B8,$7F,$3F,$F1,$B6,$BF,$21
	.dc.b	$D7,$F8,$1E,$0F,$C1,$36,$BF,$32,$FF,$09,$EB,$FC,$4B,$7A,$FF,$35
	.dc.b	$3F,$EF,$76,$DF,$E5,$67,$9C,$FE,$8B,$82,$CF,$E9,$FE,$3B,$F6,$FF
	.dc.b	$1D,$FB,$86,$D7,$ED,$1C,$3F,$B6,$F8,$EB,$5F,$AF,$F8,$EF,$DF,$FC
	.dc.b	$4F,$F0,$06,$D7,$EF,$1C,$3F,$BE,$F8,$9B,$5F,$EE,$9F,$AB,$FC,$4C
	.dc.b	$FE,$17,$0F,$F3,$47,$47,$FB,$D2,$73,$5F,$EF,$A8,$4E,$AB,$FC,$E7
	.dc.b	$5A,$AF,$F0,$E3,$5C,$F5,$7E,$36,$31,$1F,$E7,$D7,$AD,$F7,$9F,$CE
	.dc.b	$7E,$DA,$F8,$CF,$CB,$56,$B0,$FF,$42,$79,$AE,$BD,$47,$E3,$CC,$FE
	.dc.b	$8B,$E5,$CF,$F1,$D8,$7F,$4F,$8C,$59,$7A,$9F,$D0,$DB,$7F,$E2,$D9
	.dc.b	$D4,$E4,$D7,$C7,$F3,$3A,$78,$16,$BC,$58,$FF,$B0,$F0,$7C,$7F,$87
	.dc.b	$E3,$BC,$4D,$AD,$E5,$E3,$FA,$0F,$11,$23,$C5,$68,$F3,$7C,$87,$AC
	.dc.b	$F6,$7C,$85,$AF,$E0,$3D,$AF,$83,$FE,$0D,$F2,$9F,$E7,$C9,$CC,$63
	.dc.b	$D4,$F8,$9F,$23,$F0,$49,$3E,$07,$E2,$7C,$8F,$C9,$3E,$57,$BF,$67
	.dc.b	$F8,$7F,$D6,$3D,$9F,$E2,$9F,$59,$FE,$AB,$9C,$C6,$3E,$0B,$F8,$85
	.dc.b	$7F,$6A,$9F,$F2,$B3,$79,$7F,$CB,$8D,$7F,$CC,$CD,$7F,$03,$3B,$EE
	.dc.b	$56,$3D,$5F,$F0,$64,$BD,$DB,$1F,$E0,$FF,$3F,$E2,$57,$D0,$5F,$0F
	.dc.b	$9E,$1D,$3F,$E1,$5F,$AA,$7C,$77,$F8,$1D,$B7,$F8,$56,$E9,$FE,$7B
	.dc.b	$EA,$46,$23,$FD,$AF,$0F,$AA,$FF,$0E,$D7,$E0,$1B,$2D,$BD,$6A,$B8
	.dc.b	$7B,$5F,$E2,$7A,$DD,$6A,$5F,$0F,$FD,$06,$F9,$61,$FA,$68,$C5,$6F
	.dc.b	$EA,$9F,$17,$FE,$31,$9C,$D7,$EB,$27,$2F,$E8,$5B,$7E,$D6,$BE,$C6
	.dc.b	$BC,$63,$FC,$79,$1C,$2F,$C1,$18,$BF,$F2,$07,$D5,$3E,$53,$FC,$85
	.dc.b	$DF,$FD,$0A,$BF,$47,$39,$7F,$4B,$19,$C7,$DA,$7C,$59,$D3,$F4,$5A
	.dc.b	$FD,$06,$BF,$C2,$1A,$F1,$AF,$19,$E4,$DE,$DF,$ED,$3D,$E7,$C8,$7F
	.dc.b	$95,$FE,$A9,$EC,$FF,$96,$67,$35,$FB,$A9,$CB,$C6,$BF,$C9,$D3,$AC
	.dc.b	$7F,$0C,$D7,$F0,$AD,$7F,$8E,$FD,$57,$EA,$5F,$7F,$0A,$3F,$CF,$17
	.dc.b	$FC,$21,$D6,$FF,$83,$73,$5C,$1C,$CF,$4F,$8B,$FF,$38,$46,$27,$F1
	.dc.b	$CF,$0A,$FF,$33,$7C,$72,$5F,$53,$FC,$45,$FE,$F9,$97,$E5,$A3,$11
	.dc.b	$B1,$3C,$63,$34,$AE,$1E,$47,$FB,$66,$B4,$F1,$9D,$65,$8C,$3F,$DC
	.dc.b	$D5,$5F,$D3,$5F,$27,$27,$9A,$F9,$46,$63,$C9,$3C,$6F,$AC,$6C,$E3
	.dc.b	$87,$9A,$B8,$47,$C1,$AE,$11,$88,$F6,$8F,$97,$E1,$8F,$C7,$7E,$1C
	.dc.b	$71,$35,$8F,$92,$74,$FC,$18,$F1,$1E,$C8,$F7,$3C,$70,$EA,$E8,$C4
	.dc.b	$F3,$70,$79,$23,$67,$89,$F3,$9F,$54,$F8,$65,$29,$4D,$9F,$6F,$EA
	.dc.b	$1F,$06,$FA,$AD,$46,$1B,$19,$9E,$93,$96,$F7,$FF,$1D,$6E,$BC,$A3
	.dc.b	$11,$E5,$9F,$F2,$1C,$E4,$FC,$06,$2F,$47,$F1,$D9,$7E,$42,$31,$7F
	.dc.b	$85,$BE,$2F,$CC,$6C,$EE,$B8,$D7,$E5,$CE,$35,$BC,$3C,$D5,$F9,$CB
	.dc.b	$E7,$A7,$1A,$FB,$27,$E9,$6F,$8C,$6E,$FD,$CD,$C7,$D3,$8F,$82,$F8
	.dc.b	$35,$61,$E6,$C3,$E8,$61,$E2,$7C,$D5,$B7,$CF,$8F,$5B,$F4,$61,$D9
	.dc.b	$86,$09,$8F,$A5,$FC,$99,$E2,$3D,$67,$8B,$0F,$4D,$FF,$8B,$BE,$AB
	.dc.b	$F9,$B4,$FF,$85,$9A,$FF,$14,$6D,$FF,$53,$7F,$B3,$C6,$F9,$FF,$19
	.dc.b	$C6,$9E,$33,$FD,$91,$AF,$79,$DE,$9F,$D5,$A3,$FC,$64,$D7,$FB,$73
	.dc.b	$5F,$ED,$AD,$7F,$C8,$AE,$EF,$FB,$37,$CE,$7F,$BE,$27,$FC,$5E,$D7
	.dc.b	$F9,$6C,$EF,$83,$FC,$9B,$F5,$4F,$2B,$F8,$A4,$61,$B5,$FD,$D6,$74
	.dc.b	$FF,$8E,$97,$CE,$EB,$44,$6A,$7F,$C0,$CD,$7F,$C0,$AD,$7B,$1F,$E5
	.dc.b	$17,$86,$5F,$DF,$FE,$A9,$F2,$BD,$5B,$5E,$1C,$7C,$1F,$53,$FE,$50
	.dc.b	$8F,$F2,$6B,$5F,$E9,$6D,$BF,$F0,$52,$F5,$F8,$FA,$FF,$62,$38,$57
	.dc.b	$F8,$69,$AF,$F9,$C9,$AF,$F8,$5F,$EA,$BF,$B2,$BF,$E6,$96,$BF,$E2
	.dc.b	$56,$BF,$E2,$46,$BF,$E1,$7F,$43,$F0,$FE,$3F,$F9,$2E,$3F,$CE,$4D
	.dc.b	$7F,$C5,$CD,$7F,$C5,$BE,$BB,$E5,$7F,$1E,$D8,$FF,$3D,$35,$FF,$1B
	.dc.b	$35,$FF,$1A,$FD,$53,$E2,$3F,$2E,$D8,$FF,$41,$35,$FF,$1F,$35,$FE
	.dc.b	$59,$F4,$EF,$A0,$FE,$55,$08,$C3,$FC,$B5,$AF,$F3,$16,$BF,$CD,$7E
	.dc.b	$9B,$F2,$8D,$F5,$39,$C6,$23,$FB,$73,$5F,$EE,$DF,$6B,$F9,$07,$F9
	.dc.b	$D3,$5F,$E7,$CD,$7F,$A1,$35,$FF,$55,$FD,$37,$F3,$E7,$FD,$40,$D7
	.dc.b	$FC,$B8,$D7,$FC,$B6,$D7,$C7,$F3,$DE,$62,$3F,$D2,$E9,$1F,$C0,$BE
	.dc.b	$07,$C1,$3F,$EA,$5E,$BC,$11,$FF,$4F,$F7,$EC,$8E,$0E,$9B,$7C,$42
	.dc.b	$38,$3F,$D4,$FE,$7F,$72,$38,$3F,$D5,$7F,$49,$E3,$23,$83,$D1,$F2
	.dc.b	$1F,$40,$8E,$0F,$F5,$9F,$07,$AC,$8E,$0F,$ED,$5F,$4D,$FE,$D9,$7F
	.dc.b	$D8,$61,$C3,$FE,$5A,$31,$F0,$FF,$C3,$1F,$F1,$36,$B6,$77,$1F,$DE
	.dc.b	$5A,$FF,$8B,$DA,$FF,$8D,$DA,$F4,$3C,$3E,$6B,$D9,$EC,$D8,$F2,$CF
	.dc.b	$93,$BB,$FD,$19,$C1,$FE,$60,$9F,$F6,$AB,$5F,$F2,$9B,$63,$7F,$4B
	.dc.b	$FC,$7E,$31,$E2,$FF,$BC,$BD,$3C,$19,$9F,$E4,$9F,$7F,$F8,$DB,$FF
	.dc.b	$4D,$2F,$98,$C9,$AF,$46,$65,$FE,$E6,$9D,$36,$72,$FF,$A1,$3E,$21
	.dc.b	$F8,$0B,$E4,$BF,$9B,$C6,$1B,$19,$9F,$F7,$BC,$F0,$F9,$2C,$D7,$F8
	.dc.b	$8E,$38,$4F,$F8,$9D,$AF,$F8,$AD,$AF,$F9,$85,$CE,$5F,$EB,$A7,$C7
	.dc.b	$F2,$7E,$03,$AC,$F8,$30,$7C,$AF,$90,$91,$9F,$94,$AF,$B6,$76,$F1
	.dc.b	$9F,$5C,$FD,$1F,$86,$FA,$5F,$0D,$2B,$EC,$C6,$5F,$B7,$19,$7E,$DC
	.dc.b	$7E,$47,$E7,$3F,$B2,$67,$FE,$D2,$94,$63,$8F,$FC,$FC,$D8,$FF,$43
	.dc.b	$7A,$BF,$6E,$73,$E2,$FF,$22,$AF,$E2,$D5,$CA,$30,$47,$F4,$B4,$F9
	.dc.b	$9F,$D1,$36,$3F,$53,$5D,$FD,$57,$43,$E9,$FF,$B7,$7C,$0F,$9F,$4F
	.dc.b	$F7,$23,$74,$FF,$4D,$5F,$0A,$FC,$B1,$C2,$26,$3F,$9E,$8F,$C3,$FE
	.dc.b	$02,$F9,$8A,$49,$FC,$3C,$7F,$53,$C7,$EF,$3B,$8C,$CE,$DF,$CA,$46
	.dc.b	$EF,$E4,$9F,$3C,$91,$FE,$AF,$AF,$F8,$02,$3F,$80,$1C,$1A,$9F,$F3
	.dc.b	$C6,$BF,$55,$FF,$3E,$35,$EA,$7F,$9E,$DF,$84,$A9,$F8,$93,$D0,$FF
	.dc.b	$0D,$C6,$1E,$CF,$ED,$38,$7F,$6E,$4A,$FD,$97,$CF,$57,$FD,$AB,$DB
	.dc.b	$F5,$DB,$3F,$BE,$9F,$AB,$F1,$72,$7D,$DA,$FD,$4E,$2F,$F1,$84,$38
	.dc.b	$90,$F2,$38,$4F,$FB,$8A,$37,$3A,$BF,$C0,$D7,$E0,$BC,$BF,$F2,$7D
	.dc.b	$7F,$75,$BE,$D1,$82,$3E,$23,$C1,$0F,$9E,$FC,$59,$CE,$35,$87,$93
	.dc.b	$EA,$BF,$1E,$48,$CD,$3E,$AB,$11,$FE,$AF,$AE,$0F,$FA,$BD,$AE,$65
	.dc.b	$5E,$8F,$5E,$4E,$E5,$18,$8E,$0F,$E8,$3E,$4E,$47,$8A,$3F,$47,$3F
	.dc.b	$08,$91,$CD,$AF,$CB,$F3,$3F,$52,$70,$7D,$7F,$01,$FA,$C2,$3E,$E8
	.dc.b	$AF,$A6,$3E,$79,$1F,$77,$D4,$B0,$44,$E0,$AF,$DF,$9D,$39,$3F,$6E
	.dc.b	$E1,$FE,$25,$9A,$31,$3C,$A3,$15,$FE,$FE,$F6,$BF,$19,$E7,$DC,$3E
	.dc.b	$EB,$6B,$71,$FE,$FD,$9F,$1F,$D5,$7E,$07,$EA,$B1,$3E,$88,$C2,$F8
	.dc.b	$DF,$F2,$54,$61,$BF,$37,$DE,$53,$9B,$EB,$C6,$C3,$AD,$6A,$30,$F5
	.dc.b	$9F,$1D,$F3,$BE,$53,$E3,$AD,$E1,$E2,$6D,$F0,$9E,$B5,$C1,$ED,$86
	.dc.b	$2B,$71,$ED,$C3,$B4,$62,$36,$78,$CF,$04,$F4,$9D,$BC,$C9,$1B,$3E
	.dc.b	$0C,$69,$C5,$78,$C9,$19,$C3,$D5,$5C,$9F,$97,$57,$0F,$93,$36,$3D
	.dc.b	$05,$3C,$F3,$C9,$EF,$5D,$6B,$8C,$78,$B0,$F7,$60,$AD,$4F,$B8,$D6
	.dc.b	$1E,$F7,$DA,$7E,$60,$F8,$A5,$9F,$C2,$B6,$BE,$DB,$9E,$47,$CF,$71
	.dc.b	$8F,$C6,$46,$1C,$57,$FA,$46,$75,$1F,$8A,$9D,$3C,$E3,$EE,$C7,$1C
	.dc.b	$BF,$D3,$9F,$3D,$FE,$D7,$DF,$35,$2F,$E8,$DE,$F4,$FD,$8D,$EE,$73
	.dc.b	$5B,$C3,$F3,$B7,$B9,$FD,$CD,$6E,$73,$5B,$BF,$D4,$DE,$ED,$7B,$9C
	.dc.b	$D6,$E7,$E0,$BF,$A2,$9D,$46,$0F,$19,$2F,$DB,$F0,$CC,$79,$4F,$C5
	.dc.b	$BB,$7C,$A3,$07,$1C,$3C,$2F,$E5,$63,$59,$7C,$F8,$D1,$E1,$C7,$BE
	.dc.b	$1D,$A7,$51,$F0,$27,$96,$3E,$37,$89,$E0,$8D,$1C,$9C,$CF,$2C,$3C
	.dc.b	$73,$93,$6F,$F2,$0E,$3E,$91,$FE,$D9,$DF,$FB,$1F,$D3,$62,$B5,$97
	.dc.b	$F1,$69,$CB,$78,$FF,$D5,$EF,$F2,$BC,$BF,$97,$63,$E7,$8F,$E5,$13
	.dc.b	$A9,$FE,$5A,$D8,$FE,$59,$CA,$F2,$7F,$C1,$98,$7F,$74,$6C,$66,$7F
	.dc.b	$AB,$46,$0F,$F0,$CB,$FD,$F5,$B7,$FD,$BB,$E6,$AD,$F3,$AF,$ED,$3E
	.dc.b	$3E,$30,$55,$8C,$36,$7F,$B2,$47,$88,$8F,$F6,$7F,$84,$E5,$CD,$CD
	.dc.b	$B1,$FD,$D9,$F4,$91,$E0,$7F,$6F,$FA,$7F,$69,$1C,$13,$2F,$F0,$94
	.dc.b	$79,$FE,$A3,$D0,$FB,$3F,$47,$FE,$7D,$5F,$E0,$85,$F2,$59,$EA,$D3
	.dc.b	$67,$FC,$09,$3C,$30,$FF,$04,$35,$FE,$BC,$D7,$FC,$B6,$BE,$AF,$C0
	.dc.b	$78,$78,$ED,$78,$62,$D7,$F9,$06,$38,$CE,$3C,$7F,$CD,$70,$BE,$E5
	.dc.b	$F0,$BF,$59,$B7,$FE,$22,$5F,$0B,$99,$93,$97,$84,$63,$1F,$F6,$F7
	.dc.b	$4F,$F8,$6A,$BF,$C6,$18,$BF,$0C,$FF,$10,$C7,$FB,$9A,$75,$5E,$73
	.dc.b	$FC,$B1,$3A,$79,$F8,$9F,$E5,$F7,$FB,$A3,$67,$8C,$7F,$37,$6B,$FE
	.dc.b	$10,$7A,$DF,$FA,$75,$B7,$FE,$71,$FE,$01,$7C,$AF,$E6,$2F,$AC,$FF
	.dc.b	$B1,$BC,$0F,$F0,$B6,$1F,$EB,$D6,$C7,$F8,$7B,$DC,$FE,$43,$E1,$7E
	.dc.b	$81,$FE,$47,$76,$B6,$A3,$35,$D6,$B5,$3F,$8B,$FA,$EF,$C6,$7A,$56
	.dc.b	$BF,$18,$D9,$F8,$DC,$4E,$5F,$F5,$BD,$7E,$46,$F8,$BF,$E5,$FA,$7D
	.dc.b	$6B,$E2,$7F,$6B,$AF,$FA,$7F,$BB,$E0,$4E,$6F,$EB,$C7,$FA,$8D,$E5
	.dc.b	$19,$BF,$EB,$8D,$7F,$B3,$35,$FF,$1F,$B5,$FF,$21,$BC,$6B,$FD,$9C
	.dc.b	$D9,$FF,$51,$57,$F0,$27,$BD,$FE,$B7,$E1,$FF,$5D,$F5,$D6,$7F,$5C
	.dc.b	$DC,$7E,$1E,$68,$C6,$5F,$9E,$8C,$57,$FC,$00,$FC,$11,$DF,$F5,$DB
	.dc.b	$1F,$E8,$5E,$17,$A5,$FB,$5C,$3F,$DF,$95,$AC,$BF,$2D,$48,$C4,$EB
	.dc.b	$0F,$C1,$5F,$1C,$3F,$0D,$5A,$6F,$83,$E1,$F0,$7C,$4D,$8F,$33,$FE
	.dc.b	$BD,$C7,$B3,$68,$D2,$4F,$03,$DD,$E1,$74,$7F,$75,$F1,$9F,$BD,$E0
	.dc.b	$FD,$F3,$63,$27,$EF,$E3,$2D,$3F,$8E,$7C,$D7,$F1,$F3,$F9,$07,$C4
	.dc.b	$58,$C9,$37,$FA,$71,$97,$F7,$87,$F7,$B7,$F2,$3F,$75,$FE,$BF,$9F
	.dc.b	$CB,$08,$CF,$AD,$FB,$38,$C1,$FE,$26,$2D,$7F,$8B,$5A,$FF,$8D,$DA
	.dc.b	$FF,$A4,$1D,$F0,$FF,$79,$8F,$14,$7F,$14,$BF,$BE,$F8,$AF,$F9,$E7
	.dc.b	$C4,$FF,$62,$E3,$34,$93,$FE,$38,$F2,$3F,$8B,$61,$FE,$47,$6B,$FE
	.dc.b	$A0,$6B,$FE,$A4,$6F,$57,$EE,$7D,$9E,$C4,$AF,$E8,$AF,$87,$D9,$91
	.dc.b	$FD,$21,$B7,$FE,$4D,$BF,$F6,$D4,$F8,$FD,$4F,$E4,$D1,$89,$D4,$7A
	.dc.b	$FE,$86,$57,$F5,$26,$CF,$F5,$2F,$AF,$F8,$04,$9F,$EA,$EF,$F3,$1E
	.dc.b	$77,$BF,$49,$3F,$E5,$A8,$CB,$FE,$E1,$F8,$37,$D5,$FF,$83,$1B,$19
	.dc.b	$8C,$1F,$CF,$7D,$66,$54,$93,$FE,$66,$8C,$BD,$F9,$FC,$84,$70,$7F
	.dc.b	$9B,$3E,$A7,$F4,$ED,$EB,$FE,$95,$18,$D7,$FC,$EC,$D7,$FD,$9D,$D2
	.dc.b	$F8,$EF,$F1,$1B,$6F,$FD,$F8,$D9,$FF,$80,$0D,$7F,$DA,$DF,$53,$FD
	.dc.b	$A2,$7F,$D0,$5C,$A8,$C4,$FF,$A3,$63,$1E,$7F,$F8,$C2,$73,$3F,$ED
	.dc.b	$FF,$A9,$FF,$1C,$37,$DA,$FF,$1A,$C6,$23,$FD,$DF,$68,$C4,$21,$E8
	.dc.b	$3D,$DF,$F1,$EB,$72,$F2,$B6,$BC,$8D,$7F,$DE,$1E,$3B,$E2,$BF,$C8
	.dc.b	$AD,$AF,$6B,$67,$D8,$D7,$8C,$FF,$90,$A3,$5E,$0F,$F8,$49,$7C,$4E
	.dc.b	$67,$FC,$8F,$F4,$FF,$E9,$7A,$FF,$0B,$FD,$47,$F3,$AC,$3F,$B0,$CE
	.dc.b	$71,$F8,$33,$9A,$FF,$0E,$CF,$FA,$7E,$87,$C6,$70,$DF,$81,$FF,$2F
	.dc.b	$C6,$23,$E4,$3D,$5B,$1E,$4F,$63,$CA,$FF,$97,$3C,$E7,$CB,$7F,$97
	.dc.b	$BE,$A1,$E2,$66,$30,$DD,$3F,$07,$5C,$1B,$3F,$E7,$08,$C6,$3F,$85
	.dc.b	$FA,$87,$C4,$FF,$9D,$1B,$7F,$8C,$6C,$EE,$74,$7F,$9C,$23,$3A,$FF
	.dc.b	$90,$1F,$F1,$0C,$FE,$15,$CD,$7F,$B5,$FD,$BF,$8F,$F0,$1F,$E2,$4A
	.dc.b	$F9,$36,$C7,$FD,$45,$39,$9F,$F1,$71,$FE,$60,$F6,$FE,$7A,$AE,$18
	.dc.b	$74,$9E,$D2,$FA,$6E,$55,$A8,$C4,$EA,$73,$87,$FA,$C5,$F0,$4F,$DC
	.dc.b	$C7,$EE,$C6,$B0,$B7,$C4,$F1,$E1,$D6,$78,$DA,$F6,$FB,$B0,$ED,$7C
	.dc.b	$63,$18,$69,$F7,$E3,$DD,$F8,$07,$F9,$44,$A7,$C1,$AF,$8F,$5C,$E3
	.dc.b	$15,$CD,$F5,$1F,$00,$EE,$53,$E6,$E1,$DC,$F4,$E1,$E3,$C7,$D9,$7F
	.dc.b	$32,$BE,$89,$F5,$0A,$76,$29,$F0,$4C,$FD,$2F,$F7,$1F,$A8,$CB,$63
	.dc.b	$3E,$82,$72,$D8,$E3,$E0,$7D,$F6,$E1,$CA,$31,$1E,$4B,$46,$23,$F0
	.dc.b	$F1,$89,$F9,$D8,$3E,$A3,$F1,$79,$FE,$3E,$31,$7F,$4E,$F8,$3F,$1D
	.dc.b	$B3,$BA,$E3,$5F,$94,$38,$D6,$F0,$F4,$57,$E6,$EF,$9E,$DC,$6B,$F3
	.dc.b	$07,$F0,$AB,$E3,$1B,$BF,$80,$DC,$7D,$38,$F8,$2F,$83,$56,$1E,$6C
	.dc.b	$3E,$0E,$1E,$27,$CD,$5B,$7C,$F8,$F5,$BF,$46,$1D,$98,$60,$91,$E9
	.dc.b	$7F,$15,$3C,$8E,$F3,$FA,$0E,$D7,$D4,$7F,$A1,$6B,$F7,$FB,$FF,$4C
	.dc.b	$8C,$47,$FA,$41,$B5,$FE,$CA,$7F,$90,$C7,$03,$FA,$83,$A9,$E5,$E3
	.dc.b	$7E,$C9,$B3,$BB,$F0,$F2,$66,$3F,$AC,$6A,$F5,$3F,$2F,$5F,$21,$B1
	.dc.b	$99,$FE,$DA,$D7,$FB,$63,$72,$FF,$6F,$FC,$A3,$E1,$FB,$60,$8C,$5F
	.dc.b	$8E,$F8,$65,$FD,$B3,$EA,$3F,$DC,$33,$E7,$6B,$FE,$B2,$6E,$3F,$EE
	.dc.b	$C5,$EC,$76,$77,$3F,$C9,$31,$46,$23,$B7,$91,$EE,$8C,$47,$F7,$97
	.dc.b	$AE,$1F,$DF,$BE,$A1,$F1,$DE,$A6,$D7,$96,$FD,$BF,$4D,$FC,$B2,$FF
	.dc.b	$C0,$AB,$C3,$C9,$AE,$4B,$7F,$E1,$69,$D3,$67,$E8,$C6,$23,$FC,$31
	.dc.b	$F4,$CF,$5B,$E5,$B6,$3E,$E3,$5F,$F6,$BC,$F1,$DD,$8F,$F8,$63,$C9
	.dc.b	$7B,$BF,$51,$7B,$F4,$C4,$7D,$8F,$95,$7C,$2F,$E1,$1B,$97,$F8,$B9
	.dc.b	$AF,$F8,$B5,$AF,$D1,$FA,$67,$B9,$F8,$26,$C7,$E4,$DA,$FF,$BB,$97
	.dc.b	$CA,$FE,$56,$1F,$4A,$F3,$7F,$31,$B3,$EE,$FF,$8E,$D7,$E8,$E3,$1E
	.dc.b	$27,$1B,$E1,$3F,$EF,$5F,$C2,$7B,$BF,$CE,$1A,$FF,$6A,$6B,$FE,$1A
	.dc.b	$6F,$5B,$EF,$FF,$C0,$B3,$FA,$B6,$BF,$DC,$78,$FF,$C8,$B1,$8C,$7E
	.dc.b	$8C,$FF,$91,$63,$FC,$8A,$FE,$E4,$7D,$3B,$D8,$F6,$FD,$7B,$DE,$A7
	.dc.b	$C3,$9D,$57,$EE,$FE,$07,$DA,$FF,$97,$5A,$FE,$85,$AF,$F9,$61,$AF
	.dc.b	$F9,$5F,$E3,$BF,$AC,$E5,$E6,$D1,$18,$D3,$FC,$BE,$DA,$E1,$F1,$FF
	.dc.b	$BC,$F8,$97,$D0,$DF,$A1,$95,$FC,$3E,$72,$DF,$A4,$FE,$1B,$06,$92
	.dc.b	$3F,$CF,$30,$FA,$4E,$C4,$8E,$9F,$4B,$E1,$FA,$8F,$E1,$BF,$A4,$F9
	.dc.b	$76,$37,$04,$EB,$4F,$E3,$B2,$69,$30,$FF,$44,$C7,$F5,$8D,$FE,$23
	.dc.b	$77,$FF,$40,$B6,$7F,$D0,$0B,$E5,$FF,$31,$F6,$DF,$52,$FF,$DC,$4D
	.dc.b	$BF,$F1,$BB,$6F,$FC,$71,$F9,$87,$F9,$FD,$FC,$24,$7D,$64,$D2,$76
	.dc.b	$FF,$83,$23,$2F,$E9,$27,$F8,$94,$74,$F2,$BC,$73,$F1,$BE,$BB,$E3
	.dc.b	$FD,$7C,$AF,$E8,$9A,$27,$51,$89,$D3,$F9,$D3,$DF,$1F,$42,$E4,$72
	.dc.b	$7E,$9F,$99,$88,$E1,$C3,$F5,$27,$1A,$FC,$CA,$FE,$A5,$F2,$2F,$C9
	.dc.b	$1F,$0D,$86,$E3,$FD,$7E,$75,$39,$F4,$34,$FF,$B8,$EB,$75,$FE,$34
	.dc.b	$AD,$BF,$E3,$46,$B9,$8D,$E1,$FB,$69,$F1,$6D,$E1,$3A,$D7,$C6,$F2
	.dc.b	$7F,$1F,$F2,$52,$30,$43,$DF,$C1,$FE,$F2,$7E,$6C,$7E,$5A,$7F,$2F
	.dc.b	$F6,$72,$7E,$8B,$F9,$AC,$F1,$1C,$A2,$E5,$F9,$29,$FE,$49,$DD,$FD
	.dc.b	$A2,$BF,$D5,$ED,$7F,$73,$ED,$7F,$29,$F6,$7F,$55,$1F,$C3,$E7,$FB
	.dc.b	$C3,$FE,$58,$6B,$FE,$8A,$6B,$D2,$7F,$1D,$1D,$65,$D9,$D4,$F1,$CF
	.dc.b	$EC,$B5,$97,$F8,$C7,$F6,$BC,$18,$8F,$8B,$17,$2F,$D6,$CF,$DF,$F7
	.dc.b	$FF,$01,$D5,$2F,$F0,$1C,$5F,$AF,$E2,$FC,$35,$66,$7F,$29,$1F,$89
	.dc.b	$AE,$51,$F8,$7F,$C5,$9E,$0F,$E3,$9B,$19,$8C,$78,$19,$7F,$CE,$F6
	.dc.b	$AD,$D7,$E2,$DA,$FF,$AC,$23,$53,$C2,$77,$9F,$E2,$DF,$E3,$D5,$E3
	.dc.b	$F0,$7B,$3F,$B0,$F6,$BF,$8C,$FB,$D2,$30,$43,$F1,$6E,$5F,$C6,$C7
	.dc.b	$F2,$0F,$7B,$F9,$0F,$D7,$C9,$FD,$DC,$65,$FD,$EC,$65,$FC,$AC,$7F
	.dc.b	$28,$F9,$2F,$E4,$64,$8F,$E4,$2D,$7F,$37,$F2,$DF,$CB,$8E,$0F,$F0
	.dc.b	$8F,$8E,$FE,$63,$F1,$F2,$30,$7F,$0C,$70,$FE,$91,$FE,$6F,$1F,$E4
	.dc.b	$4F,$D2,$BF,$B2,$CF,$FA,$AB,$C2,$46,$23,$FC,$74,$D4,$23,$7E,$3F
	.dc.b	$F9,$62,$79,$F0,$7F,$5B,$5F,$3B,$DC,$F0,$4F,$87,$AB,$81,$30,$FF
	.dc.b	$5C,$E1,$E2,$8C,$47,$F1,$AB,$7C,$9F,$A4,$E5,$1F,$E6,$D8,$FF,$0D
	.dc.b	$E3,$E1,$F6,$BD,$C6,$E7,$4E,$1A,$71,$AF,$F3,$3F,$D3,$7F,$93,$A3
	.dc.b	$E2,$E2,$8C,$65,$FE,$5D,$6C,$FC,$49,$D7,$C2,$7C,$68,$E3,$3C,$FE
	.dc.b	$35,$F4,$BB,$D3,$FC,$5F,$5F,$EF,$D8,$C7,$BB,$C6,$B7,$3F,$3E,$F9
	.dc.b	$57,$5F,$A9,$FF,$1E,$FC,$E7,$1A,$FA,$B1,$87,$FE,$00,$4E,$F6,$56
	.dc.b	$E5,$5F,$E5,$47,$0F,$1C,$3C,$31,$AA,$E2,$F4,$3B,$B8,$8D,$CF,$39
	.dc.b	$D1,$C6,$BA,$D6,$EF,$C9,$1E,$0F,$82,$F8,$11,$8A,$E3,$7E,$38,$C6
	.dc.b	$1D,$AB,$75,$E2,$BE,$A4,$78,$C6,$9F,$11,$D3,$0F,$13,$C4,$D9,$C6
	.dc.b	$BD,$71,$A8,$E3,$1E,$17,$57,$E4,$23,$F4,$AB,$C1,$87,$A6,$3B,$57
	.dc.b	$4C,$3E,$04,$72,$33,$7C,$9F,$AB,$7C,$B0,$F1,$61,$CE,$7C,$75,$F8
	.dc.b	$3A,$D5,$F3,$9C,$9E,$33,$E2,$1F,$95,$3E,$19,$6B,$F1,$0D,$BF,$BE
	.dc.b	$E7,$7F,$F1,$C3,$C4,$FC,$4B,$87,$5E,$9E,$5C,$B6,$BF,$72,$FE,$3A
	.dc.b	$79,$E5,$F8,$EF,$07,$FC,$23,$C3,$9A,$98,$7E,$0A,$F7,$A7,$EB,$6F
	.dc.b	$73,$9A,$DE,$1F,$92,$BD,$CF,$F1,$8A,$DC,$E6,$B7,$7F,$AB,$BD,$DF
	.dc.b	$F9,$2F,$C9,$FF,$44,$DE,$E3,$B1,$1F,$44,$6A,$B9,$5F,$4A,$FD,$9D
	.dc.b	$F4,$C3,$C3,$5A,$9E,$4D,$C3,$9E,$3C,$67,$E4,$9A,$3D,$58,$F7,$7A
	.dc.b	$E3,$BC,$BE,$6C,$7E,$C3,$1E,$EE,$8E,$47,$4C,$3B,$B8,$9E,$31,$EF
	.dc.b	$32,$5C,$7C,$C4,$9F,$1E,$3E,$FE,$2F,$E4,$DE,$CF,$95,$FE,$97,$F2
	.dc.b	$77,$87,$FD,$59,$5F,$C5,$E7,$2D,$8C,$DF,$F2,$8B,$FF,$5C,$CF,$F2
	.dc.b	$F6,$C7,$F2,$F9,$DC,$FF,$56,$F8,$5F,$F5,$FF,$37,$F5,$16,$C7,$F2
	.dc.b	$FF,$AA,$B7,$FD,$67,$99,$58,$CB,$FB,$0B,$63,$37,$FD,$36,$F1,$39
	.dc.b	$C7,$51,$67,$47,$F5,$38,$FE,$D1,$7F,$DB,$1B,$7F,$D2,$E7,$FA,$FB
	.dc.b	$70,$FE,$E4,$FF,$B4,$E3,$FA,$E7,$D5,$F8,$FD,$FF,$23,$63,$8C,$74
	.dc.b	$23,$D3,$A7,$D4,$7A,$F5,$B1,$FD,$9D,$B8,$7F,$6D,$6E,$BF,$DE,$7E
	.dc.b	$AB,$E1,$7C,$7F,$C3,$6D,$FF,$6F,$7D,$64,$7D,$6F,$BB,$E9,$7E,$57
	.dc.b	$CF,$FC,$B6,$CF,$F7,$76,$E3,$FD,$F5,$B8,$7F,$82,$7B,$3E,$97,$C5
	.dc.b	$FD,$36,$DF,$F7,$F6,$DF,$F8,$15,$B5,$FE,$8B,$FD,$01,$F4,$5F,$80
	.dc.b	$F3,$90,$FA,$2C,$3A,$C7,$FD,$3C,$E5,$B5,$FE,$21,$AE,$2D,$9E,$18
	.dc.b	$FF,$A7,$DA,$FF,$95,$5C,$E1,$E3,$9E,$FF,$39,$E0,$7C,$3E,$C7,$8B
	.dc.b	$E6,$EE,$5E,$26,$E3,$FE,$6E,$C5,$18,$C7,$DC,$FE,$5F,$E9,$A5,$FF
	.dc.b	$8D,$A4,$FF,$85,$7D,$BF,$F0,$0B,$FE,$27,$9D,$E1,$FE,$2D,$7F,$45
	.dc.b	$3A,$8C,$4F,$33,$FC,$29,$E3,$FE,$AA,$DF,$1E,$FA,$0F,$D6,$DB,$C2
	.dc.b	$5E,$EB,$58,$73,$AF,$F2,$5D,$6A,$3F,$A4,$36,$3F,$A7,$35,$FE,$B0
	.dc.b	$D7,$C7,$1C,$DF,$05,$77,$CF,$C7,$5C,$F1,$F8,$75,$FE,$59,$C9,$F3
	.dc.b	$7F,$00,$F8,$3F,$57,$72,$F3,$78,$0A,$DC,$61,$B7,$FE,$EE,$55,$C6
	.dc.b	$31,$E2,$7C,$CF,$94,$F9,$AD,$BE,$4D,$9D,$C7,$59,$FC,$C4,$BA,$6F
	.dc.b	$46,$1B,$5F,$EF,$59,$D3,$FE,$EF,$6D,$FF,$BC,$DA,$FF,$AA,$5A,$FF
	.dc.b	$9F,$39,$FF,$CA,$B5,$D6,$BA,$75,$5A,$F7,$73,$7E,$18,$FC,$3F,$0D
	.dc.b	$AF,$89,$16,$7E,$24,$DE,$0F,$A3,$BF,$CB,$E4,$FC,$7F,$C9,$FE,$41
	.dc.b	$B8,$FC,$D6,$E1,$F3,$5B,$7F,$E9,$8F,$B8,$67,$F9,$3F,$DC,$3F,$1B
	.dc.b	$6A,$D4,$E6,$FF,$2B,$7C,$30,$FF,$0B,$36,$FF,$C3,$8D,$7F,$C4,$CD
	.dc.b	$7F,$CF,$8E,$EF,$AC,$FE,$6D,$61,$FB,$68,$FC,$A4,$ED,$FD,$97,$CF
	.dc.b	$FE,$B8,$FD,$7F,$D3,$D8,$FC,$67,$67,$F4,$8E,$6F,$E1,$79,$7E,$E6
	.dc.b	$30,$67,$B7,$F6,$FD,$DF,$B8,$6D,$FE,$57,$E6,$FF,$76,$7E,$F3,$E4
	.dc.b	$AC,$7E,$67,$E0,$3B,$11,$FF,$58,$E5,$FE,$F8,$6F,$BF,$FC,$13,$E0
	.dc.b	$7F,$82,$B6,$3F,$45,$F0,$DF,$C2,$0F,$E1,$3F,$23,$63,$F5,$3F,$24
	.dc.b	$CF,$FB,$1B,$63,$FD,$7E,$DD,$FF,$DA,$4D,$7F,$CA,$DD,$8F,$7B,$F8
	.dc.b	$87,$CA,$BC,$0F,$E1,$F3,$9C,$3E,$94,$E5,$FD,$CF,$DA,$BB,$FF,$BF
	.dc.b	$36,$33,$DF,$89,$FE,$33,$E6,$FC,$1C,$FF,$CB,$CD,$7F,$CD,$0D,$7F
	.dc.b	$D7,$CD,$7F,$DC,$AF,$F4,$97,$F2,$55,$FC,$CF,$EE,$3F,$C3,$BD,$BF
	.dc.b	$EF,$0A,$D3,$FE,$18,$8C,$CF,$FB,$2F,$C0,$7C,$1F,$F8,$55,$B1,$FC
	.dc.b	$51,$7C,$26,$67,$FC,$35,$F5,$0F,$8C,$FF,$08,$B6,$FF,$DD,$AD,$CF
	.dc.b	$FE,$00,$35,$FF,$6C,$78,$AF,$A8,$7B,$3F,$E2,$57,$FC,$2D,$AF,$EE
	.dc.b	$9C,$E1,$FC,$99,$CB,$FE,$E3,$F6,$9F,$07,$FE,$2D,$6C,$7F,$2C,$F2
	.dc.b	$3E,$C6,$5F,$18,$65,$FE,$34,$BF,$F1,$A4,$EA,$7C,$AD,$7F,$DE,$0D
	.dc.b	$78,$B6,$7F,$AA,$7C,$3F,$2E,$9E,$6D,$C3,$F9,$EF,$D4,$3E,$47,$FC
	.dc.b	$85,$F2,$7F,$B9,$CB,$C9,$8A,$B5,$7F,$D1,$2B,$5B,$7F,$1C,$9C,$DF
	.dc.b	$FA,$B6,$F8,$61,$FE,$42,$9C,$CF,$CE,$6A,$3E,$0B,$CA,$BA,$4F,$F9
	.dc.b	$56,$3F,$A9,$61,$E8,$8F,$F2,$C7,$43,$E3,$3F,$CB,$5F,$51,$FE,$10
	.dc.b	$AE,$99,$7F,$82,$27,$3E,$57,$07,$FC,$AB,$7C,$6B,$FD,$79,$5B,$6C
	.dc.b	$67,$3F,$67,$D4,$3E,$27,$FC,$D0,$DA,$FE,$CE,$FF,$6F,$7C,$2E,$67
	.dc.b	$E5,$B7,$5F,$E8,$6D,$9D,$FB,$1D,$2F,$FC,$17,$7F,$E0,$61,$F4,$EF
	.dc.b	$97,$F9,$11,$FC,$FF,$2F,$EF,$B3,$98,$FF,$03,$9C,$27,$79,$7D,$D9
	.dc.b	$D4,$F2,$F6,$BE,$94,$62,$75,$5F,$E1,$2A,$E4,$7F,$8B,$6F,$BC,$66
	.dc.b	$B8,$46,$27,$C3,$18,$CB,$AD,$6C,$F1,$D6,$EB,$C4,$76,$3D,$93,$C1
	.dc.b	$F5,$FB,$5C,$DF,$F0,$CD,$79,$63,$07,$B2,$3E,$45,$EC,$C3,$FE,$35
	.dc.b	$3B,$61,$EE,$7C,$F5,$E4,$BF,$04,$79,$2B,$94,$FC,$38,$FF,$1B,$9F
	.dc.b	$EF,$F3,$A1,$83,$EA,$8F,$A6,$7A,$8F,$A8,$CC,$FE,$86,$73,$18,$8E
	.dc.b	$13,$B7,$73,$A6,$F8,$5F,$DF,$DB,$97,$19,$4E,$67,$C7,$18,$9D,$4F
	.dc.b	$1C,$7F,$07,$8B,$E5,$3F,$1B,$18,$6C,$7E,$42,$F9,$61,$F5,$DB,$5B
	.dc.b	$8F,$93,$3C,$DF,$97,$3A,$8E,$57,$E9,$D3,$85,$7D,$C3,$F8,$24,$F1
	.dc.b	$70,$7C,$B9,$E1,$1C,$B1,$EF,$8F,$9A,$B8,$B5,$E1,$1E,$2A,$F8,$15
	.dc.b	$FA,$A3,$B6,$3E,$23,$F4,$B8,$F9,$EF,$EB,$57,$07,$A9,$E0,$3A,$5C
	.dc.b	$7F,$35,$7D,$0E,$D8,$78,$EF,$A7,$13,$EA,$3F,$AD,$DF,$F9,$C1,$AF
	.dc.b	$F3,$26,$E1,$FE,$4A,$5D,$5F,$BC,$DF,$F8,$FC,$70,$AF,$EB,$6D,$73
	.dc.b	$E2,$7E,$37,$0F,$03,$71,$7C,$7D,$25,$7F,$24,$8C,$B6,$3C,$0F,$F5
	.dc.b	$8F,$98,$FF,$3B,$3F,$E8,$86,$BF,$D2,$A7,$8B,$FD,$33,$EA,$3F,$9C
	.dc.b	$3F,$E8,$E6,$B9,$E9,$EB,$8A,$7F,$C0,$73,$B9,$F5,$47,$08,$FE,$F2
	.dc.b	$D7,$FA,$0B,$5F,$ED,$AF,$6D,$7F,$CE,$7F,$50,$F2,$BB,$B6,$BB,$1F
	.dc.b	$EA,$2F,$A7,$7C,$E7,$91,$FE,$D1,$19,$F0,$31,$39,$7E,$74,$EA,$7F
	.dc.b	$B2,$35,$FF,$07,$B5,$FE,$AF,$F5,$0F,$6F,$E0,$B6,$3F,$D6,$6D,$7F
	.dc.b	$BA,$CF,$5C,$BF,$D1,$BE,$43,$E1,$FE,$4B,$6B,$FD,$76,$D7,$FB,$EF
	.dc.b	$E3,$EF,$97,$F9,$FA,$7F,$91,$E7,$FB,$1B,$5F,$EF,$8D,$7F,$C4,$ED
	.dc.b	$7F,$B7,$FD,$4B,$E5,$3E,$CA,$C2,$98,$8F,$BB,$F8,$0F,$F8,$BA,$BF
	.dc.b	$D9,$ED,$7F,$C2,$2D,$CF,$FD,$52,$BE,$57,$F0,$BE,$8F,$F9,$8E,$BF
	.dc.b	$BC,$B5,$FF,$09,$35,$FF,$1E,$B5,$FF,$02,$FD,$47,$5E,$DB,$5F,$90
	.dc.b	$8E,$D3,$FE,$E2,$74,$FF,$8D,$BD,$87,$C3,$7E,$59,$B1,$FE,$70,$8C
	.dc.b	$79,$BF,$9A,$FA,$77,$D1,$7E,$C5,$B5,$FE,$10,$6B,$FE,$29,$6B,$FE
	.dc.b	$54,$F7,$FF,$01,$B7,$FB,$29,$7C,$37,$E9,$9B,$3F,$E9,$0E,$9F,$D4
	.dc.b	$FB,$BF,$CA,$63,$FC,$BA,$D7,$FC,$4A,$D7,$FC,$A2,$DF,$01,$FA,$67
	.dc.b	$F5,$D6,$CF,$F0,$98,$C1,$3E,$73,$04,$8F,$F1,$5C,$23,$87,$93,$27
	.dc.b	$2D,$7F,$C9,$12,$F9,$CE,$C4,$8E,$6F,$87,$E9,$64,$76,$6B,$FD,$B3
	.dc.b	$EA,$1F,$3F,$FC,$5A,$11,$87,$FC,$A1,$BB,$9F,$FC,$57,$39,$9F,$F3
	.dc.b	$A7,$C2,$BF,$37,$FE,$11,$5A,$C3,$FC,$01,$7B,$BF,$EF,$ED,$74,$7F
	.dc.b	$7B,$7A,$61,$DE,$18,$75,$FA,$4F,$E5,$6D,$8F,$2E,$1C,$9F,$F3,$9E
	.dc.b	$8F,$A8,$FE,$4A,$FF,$8D,$E7,$FC,$BE,$D9,$FF,$45,$BB,$79,$98,$9E
	.dc.b	$3E,$94,$8C,$F0,$FE,$02,$7A,$FB,$F2,$73,$68,$F0,$F8,$79,$9F,$2C
	.dc.b	$5D,$3D,$D8,$7E,$2A,$BD,$BB,$FC,$2C,$FF,$D0,$1A,$7C,$DD,$F2,$6B
	.dc.b	$A6,$E7,$FE,$26,$9D,$36,$FE,$73,$F2,$9B,$1A,$8F,$B3,$1F,$EC,$37
	.dc.b	$E0,$C7,$DB,$F8,$5E,$B1,$F3,$3C,$AF,$97,$F1,$F2,$3D,$F1,$F3,$8C
	.dc.b	$3F,$4A,$30,$FC,$4C,$7E,$99,$1F,$8F,$55,$FB,$31,$75,$F8,$BB,$F9
	.dc.b	$67,$FD,$84,$3F,$01,$FC,$7B,$FE,$A5,$6B,$9F,$43,$F3,$9F,$03,$FE
	.dc.b	$4E,$BF,$C6,$4E,$4F,$F2,$55,$7F,$61,$70,$7F,$94,$1C,$1D,$27,$F6
	.dc.b	$D1,$F8,$60,$E1,$F2,$C7,$93,$87,$F8,$2E,$FF,$92,$24,$FE,$1B,$96
	.dc.b	$BF,$98,$8B,$87,$C0,$9F,$C1,$F4,$FE,$13,$A6,$5F,$D6,$E2,$FF,$20
	.dc.b	$5F,$E9,$F8,$BF,$65,$7A,$AF,$BB,$7C,$9F,$47,$E7,$6F,$98,$FE,$00
	.dc.b	$D8,$CC,$7F,$03,$F4,$FF,$DE,$B1,$83,$FC,$E1,$3B,$7F,$CE,$AD,$7F
	.dc.b	$CF,$6D,$7F,$D6,$2E,$F1,$FA,$33,$FE,$1E,$6F,$AD,$E2,$BF,$DB,$F0
	.dc.b	$FE,$E3,$E2,$A6,$3F,$AC,$8C,$BF,$AE,$8C,$BF,$C0,$23,$F7,$FE,$C7
	.dc.b	$F0,$0F,$6A,$4F,$ED,$23,$2F,$ED,$A3,$2F,$F0,$88,$FE,$0F,$F0,$DF
	.dc.b	$CE,$49,$1F,$CD,$9A,$FF,$0E,$FA,$6F,$E1,$A7,$07,$F8,$0F,$D3,$FF
	.dc.b	$10,$23,$FC,$23,$F4,$DF,$C8,$EB,$F9,$C9,$FE,$8C,$49,$FE,$86,$FF
	.dc.b	$94,$65,$F5,$14,$93,$FE,$F2,$8C,$B6,$3F,$C8,$71,$F6,$21,$5F,$C3
	.dc.b	$A3,$0D,$9D,$ED,$DC,$95,$D5,$AF,$2E,$F2,$E8,$61,$FE,$FF,$8F,$77
	.dc.b	$C3,$F4,$69,$0E,$47,$18,$E0,$E2,$79,$DF,$AA,$BD,$7F,$11,$2B,$96
	.dc.b	$5E,$BF,$57,$FD,$FB,$97,$FB,$FB,$1F,$1C,$62,$BC,$F1,$87,$D3,$3B
	.dc.b	$CB,$FC,$65,$18,$36,$E1,$7C,$47,$E9,$6B,$DF,$EB,$FC,$29,$CF,$2F
	.dc.b	$08,$C5,$6E,$F9,$3F,$A7,$FA,$86,$BF,$E4,$BC,$38,$9F,$32,$33,$ED
	.dc.b	$7C,$EA,$E1,$5B,$79,$7C,$97,$6C,$BC,$57,$C3,$0E,$61,$C5,$72,$C3
	.dc.b	$A5,$70,$BE,$57,$E1,$C3,$B4,$F4,$6C,$66,$B8,$BE,$7A,$E2,$F7,$9D
	.dc.b	$B6,$76,$D3,$CB,$87,$F9,$47,$0F,$24,$62,$B8,$CF,$A3,$0F,$34,$F2
	.dc.b	$8C,$3E,$33,$85,$E7,$0F,$1C,$EA,$7C,$18,$76,$3E,$4B,$CF,$0E,$E4
	.dc.b	$BE,$51,$96,$3D,$8F,$4D,$F6,$9E,$11,$9A,$EA,$DC,$3E,$55,$F2,$9C
	.dc.b	$9E,$38,$CC,$EC,$F9,$87,$80,$FC,$71,$F2,$4B,$87,$DF,$6D,$FE,$2D
	.dc.b	$CF,$07,$F5,$37,$F0,$AD,$7F,$1C,$D7,$F0,$CD,$9F,$F6,$EF,$9D,$F3
	.dc.b	$E3,$FA,$FE,$1F,$CF,$FB,$3F,$C8,$FB,$E6,$A6,$1F,$72,$F7,$97,$E9
	.dc.b	$6F,$73,$9A,$DE,$1F,$A3,$BD,$CF,$F0,$CA,$DC,$E6,$B7,$7E,$A7,$F1
	.dc.b	$73,$A8,$C5,$75,$25,$79,$FC,$5F,$F2,$53,$CE,$33,$3C,$23,$58,$F6
	.dc.b	$9F,$8B,$85,$7F,$CD,$D7,$CA,$F8,$12,$36,$7E,$C3,$0F,$23,$8C,$3C
	.dc.b	$04,$9F,$89,$5C,$E7,$51,$CD,$FA,$77,$EE,$9D,$46,$0F,$49,$C0,$F9
	.dc.b	$0E,$B2,$FB,$31,$A3,$85,$FB,$9E,$3C,$1F,$CE,$3D,$9F,$13,$FC,$EB
	.dc.b	$D9,$BC,$29,$B1,$87,$FC,$E9,$AF,$F2,$96,$DF,$F2,$86,$BF,$CF,$5B
	.dc.b	$7F,$D1,$BE,$9B,$FC,$F9,$CB,$FD,$4F,$E9,$EE,$1F,$CC,$26,$E1,$FD
	.dc.b	$16,$6D,$F1,$AF,$ED,$55,$8C,$9F,$4F,$63,$0D,$9F,$EE,$0D,$C3,$FA
	.dc.b	$75,$70,$9F,$ED,$53,$FE,$44,$9F,$E8,$2D,$7F,$AB,$36,$FF,$AF,$FC
	.dc.b	$27,$87,$E9,$BC,$4D,$BF,$EF,$CE,$F0,$34,$91,$FD,$E7,$E9,$F4,$47
	.dc.b	$04,$C7,$CF,$1E,$FF,$A2,$F8,$0F,$C1,$FA,$2F,$85,$D5,$6F,$FC,$1C
	.dc.b	$DB,$FE,$13,$9F,$F7,$6A,$F2,$12,$BF,$C1,$5F,$4F,$F2,$FE,$1B,$E6
	.dc.b	$36,$3F,$C3,$0D,$BF,$92,$DE,$1F,$F4,$CF,$E7,$4E,$DF,$AE,$BE,$42
	.dc.b	$C6,$1B,$87,$FA,$AA,$74,$D9,$D3,$FE,$6D,$EA,$FF,$5D,$3F,$EB,$4B
	.dc.b	$7E,$EC,$FA,$0F,$C3,$E7,$FE,$E9,$F6,$FF,$13,$5A,$9C,$E1,$CA,$BF
	.dc.b	$95,$9B,$C2,$B4,$CC,$6D,$FE,$62,$F1,$AF,$F4,$4B,$67,$FD,$CB,$83
	.dc.b	$E0,$BB,$E3,$F9,$AF,$45,$F9,$BB,$E1,$70,$BE,$8E,$C6,$1B,$8F,$FA
	.dc.b	$CA,$74,$D9,$D3,$F2,$7E,$A3,$E7,$9F,$43,$C4,$B1,$92,$6D,$FE,$2D
	.dc.b	$E1,$FF,$40,$46,$37,$FF,$51,$F8,$4E,$DF,$93,$FB,$A3,$AB,$E6,$CE
	.dc.b	$6F,$E9,$4E,$5F,$F5,$14,$67,$1F,$9A,$D7,$FA,$CB,$5F,$F1,$2D,$FE
	.dc.b	$B5,$FF,$16,$BD,$8F,$0F,$4D,$AF,$C3,$7D,$4F,$E3,$0F,$C6,$FC,$35
	.dc.b	$8C,$92,$FD,$D1,$97,$E0,$7C,$D7,$E5,$3E,$77,$F2,$AD,$AF,$C9,$36
	.dc.b	$77,$1F,$95,$9F,$AF,$D7,$F6,$38,$2C,$FD,$56,$D7,$F7,$6E,$5F,$BF
	.dc.b	$5F,$C2,$5B,$5F,$EF,$5E,$4F,$CF,$D6,$DB,$C3,$FE,$B9,$6C,$FF,$C0
	.dc.b	$05,$C7,$DA,$7F,$5F,$F3,$1F,$AE,$3F,$5F,$F3,$56,$FF,$59,$17,$0F
	.dc.b	$C0,$4D,$AF,$DB,$FC,$77,$ED,$B9,$FF,$6E,$DC,$3F,$68,$DB,$FC,$43
	.dc.b	$6B,$FD,$D3,$F5,$1F,$81,$9F,$F3,$DE,$9F,$9E,$EF,$FD,$9C,$E6,$BD
	.dc.b	$91,$8E,$8F,$C0,$4F,$1B,$FF,$0E,$B5,$FF,$36,$35,$FF,$47,$BC,$F1
	.dc.b	$CC,$FE,$B7,$F4,$67,$85,$D6,$7F,$86,$76,$7F,$37,$EC,$FD,$4E,$3F
	.dc.b	$C8,$9A,$FF,$A1,$DA,$FF,$A1,$97,$D4,$F1,$25,$7F,$08,$F1,$13,$A8
	.dc.b	$C2,$FA,$6F,$01,$23,$2F,$8F,$E4,$24,$70,$6E,$1E,$5F,$0B,$D0,$49
	.dc.b	$EA,$FA,$FC,$59,$1E,$06,$CF,$F3,$EF,$8A,$FE,$49,$3F,$EB,$6E,$23
	.dc.b	$FE,$A2,$87,$D4,$FF,$6C,$77,$5F,$EB,$17,$33,$C4,$FF,$54,$FC,$97
	.dc.b	$BE,$7F,$A5,$B5,$FF,$58,$35,$DE,$DF,$D9,$FE,$4D,$F2,$1F,$DE,$9B
	.dc.b	$7D,$6B,$FD,$95,$0F,$90,$D5,$7F,$80,$3E,$53,$F8,$BD,$7F,$B1,$57
	.dc.b	$8B,$9F,$0B,$4F,$F8,$1B,$E5,$5E,$FF,$F7,$76,$E7,$F6,$5B,$87,$F6
	.dc.b	$46,$BF,$EC,$9F,$5F,$FC,$08,$FF,$B4,$DB,$C3,$F2,$A3,$FD,$C1,$1F
	.dc.b	$DF,$2B,$D5,$CF,$FE,$27,$F5,$7F,$C2,$FE,$1B,$E3,$2F,$C3,$E1,$B7
	.dc.b	$C3,$E9,$7F,$CC,$B9,$7F,$B6,$5A,$FF,$B6,$1A,$FF,$BB,$DB,$8B,$E9
	.dc.b	$5F,$07,$FE,$26,$6D,$F9,$17,$C1,$E6,$7F,$C5,$6B,$E1,$29,$FE,$2E
	.dc.b	$F3,$9F,$11,$8C,$3D,$B1,$87,$F4,$4D,$CF,$FC,$1A,$D7,$FD,$E4,$D7
	.dc.b	$FD,$E2,$BE,$2B,$B5,$77,$6D,$7C,$AF,$A4,$FC,$2D,$B8,$FE,$86,$BF
	.dc.b	$8A,$F4,$FF,$C8,$75,$A9,$CF,$52,$72,$FF,$8D,$E7,$9D,$FF,$BD,$ED
	.dc.b	$18,$80,$8C,$CF,$13,$9E,$1E,$5E,$BB,$5E,$58,$E3,$4F,$4F,$F0,$11
	.dc.b	$F8,$4F,$92,$79,$FF,$E5,$A9,$CD,$F1,$3C,$13,$AA,$FC,$5C,$EA,$7D
	.dc.b	$0D,$7B,$B5,$F4,$36,$DF,$48,$FA,$5F,$F2,$C3,$6F,$F8,$D3,$6F,$FC
	.dc.b	$8A,$D7,$E3,$39,$DB,$E8,$B6,$7D,$EB,$E2,$7F,$CE,$AE,$6B,$FB,$F8
	.dc.b	$FA,$47,$CC,$FC,$E8,$FB,$99,$FE,$AE,$73,$1C,$1A,$EB,$0F,$AB,$18
	.dc.b	$8F,$AF,$EE,$62,$FA,$65,$F4,$6F,$15,$CE,$73,$87,$77,$FC,$2B,$1E
	.dc.b	$57,$33,$C2,$3C,$D7,$6F,$A4,$6C,$E8,$7F,$2D,$BE,$8E,$0F,$41,$B3
	.dc.b	$99,$FE,$63,$BE,$ED,$3E,$3F,$B5,$F0,$2B,$BB,$EF,$85,$F8,$EF,$BC
	.dc.b	$74,$C7,$84,$62,$3E,$29,$E1,$BF,$4C,$F3,$AE,$58,$FB,$EB,$67,$F9
	.dc.b	$8A,$3F,$CC,$6E,$0F,$8E,$7D,$91,$F4,$4F,$65,$F4,$BF,$E6,$A8,$C3
	.dc.b	$63,$33,$FE,$70,$9C,$B7,$C1,$FD,$F3,$72,$E5,$18,$8F,$04,$A3,$11
	.dc.b	$F8,$B8,$C4,$FD,$9D,$5F,$2B,$F8,$FD,$3F,$17,$18,$BF,$BD,$7C,$1F
	.dc.b	$AC,$D9,$DD,$71,$AF,$CC,$1C,$6B,$78,$7A,$2B,$F2,$F7,$CF,$6E,$35
	.dc.b	$FC,$31,$AF,$DF,$3E,$4D,$7C,$08,$E2,$F3,$BE,$F8,$78,$6B,$F8,$AC
	.dc.b	$78,$70,$DE,$3E,$AB,$E1,$5A,$BE,$0F,$EC,$1B,$8F,$B7,$0F,$91,$5C
	.dc.b	$2B,$47,$C9,$AD,$C6,$6B,$D9,$5E,$4A,$F6,$35,$F6,$46,$9F,$E5,$5F
	.dc.b	$4A,$F9,$3F,$DA,$EF,$FC,$9A,$31,$3F,$DA,$1A,$FF,$23,$E4,$43,$E4
	.dc.b	$F8,$CF,$4B,$FF,$5B,$D6,$9B,$5C,$A7,$F9,$F3,$5F,$E6,$B8,$3E,$4B
	.dc.b	$8B,$63,$FC,$78,$D7,$FD,$CE,$D7,$FA,$43,$9D,$7F,$A9,$FD,$2B,$E0
	.dc.b	$FC,$B9,$23,$17,$FE,$04,$BE,$1A,$78,$BC,$3D,$C7,$F7,$49,$D4,$72
	.dc.b	$F9,$6F,$EE,$31,$83,$CB,$1E,$82,$38,$3F,$C2,$37,$FE,$62,$6F,$B9
	.dc.b	$DA,$7C,$BF,$46,$F5,$A9,$2B,$FC,$33,$0F,$94,$C4,$7E,$FF,$44,$F1
	.dc.b	$8F,$F3,$13,$5D,$35,$FE,$DD,$F4,$9F,$EE,$37,$FC,$D0,$D7,$A3,$5D
	.dc.b	$FB,$7F,$72,$7E,$0C,$74,$F2,$30,$47,$C3,$E4,$6C,$8E,$0F,$F3,$B7
	.dc.b	$D1,$BD,$2F,$C3,$B6,$33,$18,$3F,$17,$F4,$6F,$94,$A4,$9F,$F4,$04
	.dc.b	$65,$FE,$44,$7F,$82,$3E,$CB,$F0,$CF,$F5,$D6,$B9,$E8,$FE,$57,$18
	.dc.b	$D9,$EB,$73,$AF,$F2,$96,$1F,$9C,$8C,$47,$F8,$5D,$AF,$F8,$3B,$DA
	.dc.b	$7C,$1F,$E8,$9B,$1F,$E9,$56,$BF,$65,$AF,$F3,$9C,$3F,$BC,$3E,$A0
	.dc.b	$E1,$F6,$B5,$E1,$E4,$FE,$C3,$DE,$7C,$05,$24,$FF,$A9,$63,$2F,$FA
	.dc.b	$9F,$E8,$3F,$A5,$E8,$69,$23,$FA,$7F,$81,$FE,$5A,$7C,$8D,$7E,$7B
	.dc.b	$5F,$EA,$4D,$7F,$D7,$1F,$03,$FC,$13,$E1,$3F,$D3,$B1,$FE,$B4,$8C
	.dc.b	$19,$FA,$A9,$1F,$EB,$86,$BD,$BD,$FF,$19,$1F,$1B,$D7,$EA,$A4,$7F
	.dc.b	$B0,$1A,$FE,$13,$E8,$BF,$33,$E3,$7F,$A1,$E1,$18,$F1,$7F,$C4,$4D
	.dc.b	$87,$C6,$3F,$49,$FF,$53,$C7,$FA,$94,$E3,$DB,$FD,$D2,$BF,$4B,$1F
	.dc.b	$DC,$DA,$F4,$8D,$1F,$CA,$2B,$A7,$97,$6F,$FC,$1D,$86,$EF,$FD,$33
	.dc.b	$EB,$FF,$7D,$CB,$FC,$DE,$D9,$FF,$2E,$B6,$7F,$CC,$CB,$E5,$7F,$90
	.dc.b	$B9,$E3,$FF,$2F,$CE,$0C,$CE,$BC,$FD,$FC,$F4,$9F,$0C,$65,$F1,$C6
	.dc.b	$5F,$E8,$97,$FD,$C7,$1F,$99,$F9,$B7,$FB,$1E,$7E,$33,$5F,$F3,$60
	.dc.b	$71,$1F,$EB,$9D,$7E,$77,$CB,$7F,$4C,$AD,$7B,$9F,$E2,$0B,$E5,$1F
	.dc.b	$E1,$F6,$BF,$E5,$96,$BF,$E8,$87,$AE,$BF,$E8,$AB,$FB,$5F,$51,$F3
	.dc.b	$CE,$D3,$AA,$F9,$31,$F2,$08,$FD,$2C,$2B,$F0,$22,$E3,$F5,$27,$E7
	.dc.b	$F0,$7D,$0F,$7A,$5F,$DA,$8C,$BF,$72,$32,$FC,$A8,$F8,$1C,$6D,$A9
	.dc.b	$27,$F2,$1E,$F7,$9D,$FC,$CC,$7E,$44,$8F,$F0,$99,$E4,$D8,$F9,$92
	.dc.b	$FA,$7A,$75,$9F,$B7,$BF,$E2,$22,$47,$E4,$A3,$F1,$66,$1F,$C0,$C6
	.dc.b	$1F,$F5,$BE,$7F,$6F,$F4,$9F,$E9,$8D,$F3,$3F,$41,$18,$E1,$FA,$8F
	.dc.b	$F1,$4A,$FF,$58,$70,$A3,$51,$FB,$26,$C6,$6B,$9F,$81,$F8,$7F,$D0
	.dc.b	$9F,$27,$49,$87,$ED,$FD,$5F,$E3,$19,$FF,$01,$AD,$CF,$F2,$F6,$BF
	.dc.b	$D2,$9A,$FF,$9B,$5A,$FF,$A9,$DD,$C7,$F0,$B9,$FF,$04,$4F,$E7,$9B
	.dc.b	$1D,$EB,$F8,$07,$82,$F9,$EA,$4B,$F9,$3A,$FE,$5E,$7F,$C2,$F1,$F1
	.dc.b	$1E,$0E,$8F,$E2,$AE,$1D,$4F,$6F,$63,$F8,$0F,$A9,$27,$F8,$5B,$FC
	.dc.b	$13,$D4,$C4,$7E,$EA,$2E,$5F,$C3,$A7,$F8,$4F,$D4,$FF,$8F,$3B,$0D
	.dc.b	$25,$FF,$70,$EF,$FF,$75,$C7,$F1,$0F,$A9,$93,$FC,$6A,$3F,$89,$FC
	.dc.b	$6C,$8F,$E3,$AF,$F1,$6F,$8E,$C4,$7F,$87,$70,$4F,$84,$91,$FD,$97
	.dc.b	$ED,$3F,$D6,$48,$38,$6F,$6F,$FA,$A2,$30,$7F,$2E,$8F,$F5,$2F,$DF
	.dc.b	$BC,$9F,$F2,$DF,$C2,$A3,$13,$FC,$FE,$B8,$6A,$C9,$EC,$FC,$52,$57
	.dc.b	$C3,$8C,$B6,$3F,$99,$F9,$32,$3F,$9E,$C7,$F3,$6F,$A9,$91,$FD,$01
	.dc.b	$D3,$FC,$B9,$FA,$44,$7B,$3F,$E7,$0F,$A5,$FC,$8F,$47,$FB,$EF,$14
	.dc.b	$62,$BE,$34,$63,$DE,$FF,$80,$15,$BA,$D5,$2B,$4F,$FB,$F3,$E5,$7F
	.dc.b	$A7,$FC,$1F,$FB,$C9,$B1,$93,$85,$7E,$8B,$E2,$53,$F5,$65,$F5,$1F
	.dc.b	$E7,$FF,$9D,$E1,$18,$F7,$91,$89,$D4,$FF,$BB,$AB,$94,$7E,$DA,$FA
	.dc.b	$CF,$E0,$63,$71,$88,$E9,$5C,$EF,$AC,$73,$38,$78,$1E,$B7,$94,$6A
	.dc.b	$B9,$5F,$4C,$3C,$33,$97,$C7,$7C,$6F,$B9,$D4,$B7,$CA,$B8,$47,$80
	.dc.b	$F2,$61,$1E,$18,$77,$74,$62,$78,$BD,$30,$F1,$D7,$C7,$C3,$C5,$7C
	.dc.b	$6B,$47,$B2,$74,$F7,$C3,$C9,$16,$FD,$73,$4E,$07,$86,$38,$B4,$93
	.dc.b	$C9,$F4,$DF,$8B,$0F,$3E,$1E,$32,$46,$6F,$B5,$79,$8F,$C2,$C6,$5F
	.dc.b	$F3,$DC,$FE,$09,$FC,$51,$F3,$8A,$E5,$FC,$2F,$B3,$FE,$21,$78,$B5
	.dc.b	$DB,$5D,$FA,$5F,$E6,$F9,$4F,$1C,$7F,$02,$DC,$BF,$35,$E6,$7F,$AC
	.dc.b	$F7,$FC,$B4,$E7,$1F,$C9,$35,$FC,$FB,$67,$84,$7D,$E8,$D6,$1E,$0C
	.dc.b	$79,$9F,$A4,$AD,$E3,$E2,$8C,$DF,$17,$5D,$CC,$73,$5B,$AD,$1F,$93
	.dc.b	$BE,$79,$7C,$08,$CC,$EC,$FD,$44,$EA,$3D,$78,$74,$AE,$15,$AC,$3C
	.dc.b	$15,$CA,$72,$47,$F9,$A5,$F3,$7E,$1D,$F3,$7A,$9E,$5A,$E7,$7E,$3B
	.dc.b	$E5,$3C,$0D,$61,$C5,$C4,$ED,$EF,$5C,$AB,$8E,$3F,$36,$7C,$C6,$AF
	.dc.b	$C5,$C1,$FC,$9B,$EA,$38,$4F,$F9,$AB,$E2,$2C,$E6,$BF,$8F,$3F,$D0
	.dc.b	$63,$3C,$2B,$F8,$75,$FC,$AD,$B7,$FC,$C2,$75,$3F,$D5,$CD,$16,$3F
	.dc.b	$CF,$7F,$53,$FE,$95,$6C,$7F,$4B,$8C,$47,$F9,$F9,$B8,$7F,$53,$7E
	.dc.b	$9F,$85,$FE,$86,$6D,$7F,$40,$FD,$43,$F0,$CF,$FA,$32,$33,$7F,$CF
	.dc.b	$E5,$F2,$B8,$C1,$1C,$25,$87,$E0,$27,$35,$BA,$D4,$E5,$E7,$87,$F8
	.dc.b	$15,$B7,$FB,$AB,$E3,$39,$BF,$6C,$BF,$32,$7D,$1F,$9B,$C0,$7C,$33
	.dc.b	$4F,$4D,$EE,$B5,$39,$C7,$FD,$6D,$87,$27,$FD,$D8,$DA,$FE,$16,$71
	.dc.b	$8C,$D7,$CC,$CD,$F2,$94,$93,$EB,$6D,$7C,$48,$F9,$D1,$F0,$F9,$FE
	.dc.b	$23,$F1,$7A,$BE,$37,$BD,$6B,$FC,$75,$BF,$CE,$8C,$61,$F1,$71,$5F
	.dc.b	$8C,$97,$F2,$FE,$63,$E8,$1F,$47,$BA,$C7,$4A,$4E,$2F,$CC,$7D,$FA
	.dc.b	$C7,$17,$FA,$C7,$F2,$FF,$C9,$36,$7F,$0B,$EA,$27,$FD,$99,$8F,$F9
	.dc.b	$46,$31,$3A,$9E,$33,$FE,$AF,$6C,$7F,$BA,$9E,$2F,$F9,$F9,$CB,$FE
	.dc.b	$D7,$EA,$FC,$9D,$BD,$37,$BD,$F7,$BB,$BF,$94,$CF,$E7,$A7,$38,$FF
	.dc.b	$BC,$E7,$F9,$6E,$1F,$D1,$9A,$FF,$84,$5A,$FF,$91,$DA,$F8,$63,$96
	.dc.b	$1E,$2A,$F3,$DF,$F1,$EA,$D1,$FE,$96,$FD,$2F,$F1,$51,$FE,$D9,$6D
	.dc.b	$FF,$B3,$DF,$F1,$0B,$F1,$DC,$CB,$34,$61,$BC,$9F,$42,$38,$CF,$78
	.dc.b	$FD,$77,$C5,$FE,$BF,$B2,$E1,$F7,$DB,$5F,$7B,$93,$F7,$31,$8C,$BF
	.dc.b	$68,$FE,$A3,$F4,$57,$CD,$7D,$0F,$25,$F2,$F9,$8C,$37,$2F,$D3,$E4
	.dc.b	$BE,$31,$86,$C7,$EB,$1A,$FF,$78,$73,$8F,$CB,$6C,$FF,$B2,$30,$7C
	.dc.b	$AF,$7B,$F5,$73,$BE,$3F,$F8,$77,$D7,$7F,$0F,$6E,$1F,$9F,$6D,$70
	.dc.b	$6E,$5F,$C4,$FE,$7F,$F8,$B1,$FC,$5F,$EB,$AD,$FE,$9E,$2D,$F8,$26
	.dc.b	$D7,$E0,$7E,$BB,$FD,$0D,$7F,$A1,$FA,$24,$7F,$12,$33,$7F,$ED,$A8
	.dc.b	$CC,$FF,$AD,$E3,$7E,$AF,$F1,$EA,$DC,$7F,$87,$1A,$FF,$99,$DA,$FF
	.dc.b	$A7,$5E,$B7,$E3,$9F,$DA,$DB,$EA,$FC,$51,$FC,$CD,$FE,$69,$57,$0F
	.dc.b	$E0,$8D,$C3,$E6,$B7,$0F,$E7,$0F,$F3,$BF,$7A,$57,$F0,$BF,$7A,$DF
	.dc.b	$D4,$9B,$7F,$CF,$E6,$D7,$F9,$FF,$CF,$6B,$E6,$D7,$F9,$BC,$62,$7E
	.dc.b	$54,$62,$7F,$A0,$CE,$71,$FE,$6A,$D7,$FC,$82,$D7,$B4,$71,$AF,$A8
	.dc.b	$D8,$FC,$4F,$A3,$FC,$EC,$93,$FD,$4D,$FE,$9F,$F8,$F7,$F3,$E8,$FC
	.dc.b	$EB,$67,$FA,$2B,$5F,$F2,$8B,$5F,$ED,$9F,$5F,$FD,$1C,$93,$FD,$7D
	.dc.b	$FE,$95,$F3,$72,$3F,$B1,$36,$7A,$F9,$FE,$22,$4F,$89,$FE,$FB,$F2
	.dc.b	$4F,$86,$A4,$9F,$F7,$84,$65,$FF,$79,$7C,$1D,$BF,$ED,$CD,$7F,$D8
	.dc.b	$8D,$7F,$D8,$6D,$7F,$DD,$5E,$C7,$F5,$09,$FD,$FB,$63,$FB,$3B,$5F
	.dc.b	$F3,$9B,$5F,$F6,$6F,$D3,$3E,$03,$FC,$2A,$DA,$FE,$DF,$E1,$23,$15
	.dc.b	$FE,$03,$6B,$FE,$D9,$F9,$7F,$EE,$CD,$F1,$7F,$C3,$F1,$88,$FF,$6D
	.dc.b	$B5,$F4,$FA,$5F,$E0,$AC,$BF,$CF,$78,$FA,$A3,$07,$F8,$A3,$C6,$77
	.dc.b	$D2,$4F,$9A,$32,$FA,$3C,$17,$C2,$FF,$8C,$1B,$19,$8C,$1F,$E3,$4F
	.dc.b	$A4,$7C,$65,$24,$FC,$28,$CB,$7E,$33,$81,$1C,$1F,$23,$D2,$FF,$19
	.dc.b	$6B,$FE,$FD,$6B,$FE,$FC,$6B,$FE,$F9,$6F,$0A,$DF,$0D,$E5,$25,$F8
	.dc.b	$DB,$7F,$D2,$67,$73,$FE,$16,$5F,$0F,$FE,$45,$73,$3F,$D3,$FE,$89
	.dc.b	$CD,$FE,$49,$F8,$E4,$7F,$92,$CC,$E1,$FE,$20,$8C,$BF,$E2,$28,$CD
	.dc.b	$7F,$91,$5A,$89,$97,$4B,$FF,$30,$CE,$A7,$2F,$2F,$0F,$FC,$AC,$DB
	.dc.b	$E3,$F4,$9F,$90,$9C,$DF,$97,$83,$D1,$E7,$5B,$F3,$37,$1F,$24,$F8
	.dc.b	$09,$3F,$9C,$FA,$4F,$9E,$47,$FC,$EA,$D7,$FC,$7E,$FF,$99,$9F,$5B
	.dc.b	$FE,$64,$C3,$FC,$D8,$E6,$BF,$8F,$8F,$A4,$FD,$63,$F0,$E7,$E4,$B9
	.dc.b	$EC,$FE,$4E,$D3,$5A,$7F,$AD,$AB,$8C,$7F,$35,$1F,$49,$8B,$E6,$EA
	.dc.b	$3E,$87,$8D,$F4,$63,$57,$FE,$C0,$75,$29,$D6,$1F,$5E,$3F,$D0,$F5
	.dc.b	$B8,$F2,$D7,$2C,$B7,$5E,$E8,$C3,$E4,$9D,$9C,$F0,$F2,$BE,$EC,$76
	.dc.b	$78,$B1,$FC,$33,$88,$E3,$5F,$0B,$0E,$EF,$C1,$C3,$B9,$C5,$F0,$1B
	.dc.b	$29,$B9,$F1,$63,$EB,$8F,$AA,$7F,$89,$30,$FF,$12,$BF,$20,$AF,$F8
	.dc.b	$A3,$E9,$3F,$C5,$BF,$07,$76,$34,$95,$FE,$94,$9F,$F1,$1E,$9C,$7A
	.dc.b	$7F,$D6,$33,$96,$BF,$3C,$4F,$07,$53,$F0,$32,$FF,$35,$C5,$F6,$FE
	.dc.b	$B6,$9F,$84,$BE,$0D,$C3,$BB,$F8,$8B,$E3,$7F,$11,$B1,$BC,$3E,$6D
	.dc.b	$F8,$76,$E1,$8F,$E0,$DB,$87,$D6,$C3,$F0,$D5,$B6,$E3,$E0,$7F,$14
	.dc.b	$76,$2D,$FA,$1D,$E1,$E0,$9E,$17,$FC,$1E,$78,$36,$B8,$3F,$C2,$EF
	.dc.b	$DF,$3C,$A3,$D9,$3A,$7F,$AD,$36,$77,$1C,$6B,$BB,$A3,$E6,$1E,$3B
	.dc.b	$FC,$EF,$1B,$E4,$3F,$DB,$13,$FE,$C0,$6B,$FE,$77,$6C,$FF,$BB,$7E
	.dc.b	$03,$FD,$20,$FF,$BB,$E3,$73,$FA,$DA,$E9,$F0,$9F,$E7,$46,$C7,$F6
	.dc.b	$E9,$DB,$FC,$A3,$16,$1B,$6C,$7F,$74,$6B,$FE,$F5,$6B,$DD,$CE,$1F
	.dc.b	$F0,$07,$E9,$1E,$67,$ED,$79,$3F,$9A,$C6,$2F,$FD,$BB,$7C,$27,$F9
	.dc.b	$CF,$C9,$3E,$37,$85,$34,$4E,$63,$FD,$CA,$D7,$FD,$38,$BE,$37,$8C
	.dc.b	$F0,$AF,$85,$3B,$AF,$8C,$D7,$C3,$1A,$38,$FD,$A7,$F9,$A6,$3F,$DE
	.dc.b	$6D,$7B,$C7,$6F,$A4,$7C,$A5,$EC,$95,$FE,$FA,$9C,$B6,$3E,$A4,$7F
	.dc.b	$82,$23,$FC,$42,$D7,$F7,$18,$A3,$11,$F7,$7E,$97,$FD,$C4,$FF,$3F
	.dc.b	$D3,$FB,$0C,$62,$3F,$D7,$CD,$E0,$4F,$F6,$6A,$ED,$18,$A2,$FC,$C3
	.dc.b	$F1,$B5,$FD,$2F,$F3,$0F,$C8,$B7,$97,$EA,$D6,$E2,$CB,$E4,$3F,$D8
	.dc.b	$72,$8C,$61,$FC,$4C,$EE,$E1,$FE,$2D,$1B,$9E,$33,$F7,$63,$CC,$73
	.dc.b	$DB,$B9,$E0,$E1,$FB,$FF,$1F,$FE,$38,$C3,$E2,$1F,$89,$7F,$D1,$EE
	.dc.b	$36,$46,$23,$FD,$C8,$F1,$7F,$DC,$DE,$5B,$D6,$FD,$53,$63,$FC,$85
	.dc.b	$18,$3F,$5D,$F4,$DC,$72,$F9,$98,$23,$1A,$FF,$85,$9B,$7F,$9D,$7C
	.dc.b	$55,$C3,$D1,$7D,$57,$E6,$97,$D6,$53,$11,$F9,$EF,$57,$FC,$73,$87
	.dc.b	$F9,$A9,$AF,$F9,$A7,$CE,$FE,$05,$18,$8F,$E0,$DF,$4C,$FA,$5F,$D3
	.dc.b	$B6,$3E,$33,$5F,$C2,$E9,$FE,$2E,$87,$C1,$FE,$BB,$E9,$9F,$21,$FC
	.dc.b	$3D,$B7,$FE,$7C,$6B,$F9,$C6,$BF,$C5,$FD,$BE,$CD,$C7,$F8,$C4,$62
	.dc.b	$3F,$C7,$AD,$7F,$82,$7C,$5B,$E6,$7F,$8F,$36,$3F,$D1,$6D,$7F,$56
	.dc.b	$D7,$F9,$37,$EA,$2F,$77,$F9,$33,$63,$FD,$26,$D7,$F6,$6D,$78,$F9
	.dc.b	$BC,$08,$F0,$79,$FB,$D2,$39,$B5,$FD,$17,$DB,$7F,$99,$1F,$CB,$87
	.dc.b	$0F,$D0,$3F,$CB,$CE,$39,$BF,$71,$F2,$2F,$96,$FD,$F7,$1F,$F8,$B6
	.dc.b	$B8,$DF,$F8,$05,$AE,$7D,$8F,$F0,$9C,$62,$3A,$C7,$F3,$0B,$E9,$F0
	.dc.b	$57,$0F,$F6,$75,$79,$AF,$C5,$87,$F9,$C3,$47,$CD,$FF,$30,$6D,$7E
	.dc.b	$BD,$AF,$AF,$B5,$0F,$8D,$E1,$5C,$78,$65,$F1,$8E,$9E,$D4,$8E,$8F
	.dc.b	$7F,$7B,$11,$D7,$07,$AB,$E3,$33,$C3,$F2,$36,$FD,$6D,$7F,$06,$E0
	.dc.b	$F9,$5D,$DE,$F3,$E0,$7D,$75,$9F,$E3,$13,$A6,$D7,$27,$FC,$5D,$19
	.dc.b	$6C,$66,$FC,$1E,$07,$CB,$3A,$4F,$9F,$DB,$F9,$04,$7E,$47,$CF,$57
	.dc.b	$D9,$F6,$3F,$34,$C5,$FF,$0E,$D5,$F5,$9F,$44,$95,$F4,$75,$AF,$C5
	.dc.b	$CB,$EB,$18,$BF,$E8,$50,$FB,$5E,$6F,$E5,$1A,$FE,$4C,$D5,$FF,$76
	.dc.b	$5F,$06,$FA,$1C,$57,$09,$FF,$29,$35,$FF,$41,$35,$FF,$5D,$35,$EA
	.dc.b	$7E,$92,$7F,$A1,$74,$7F,$A2,$5B,$1C,$2F,$F2,$1C,$32,$3F,$20,$FE
	.dc.b	$4F,$AB,$11,$F8,$5F,$9A,$FC,$B1,$8C,$FF,$09,$DD,$F9,$F2,$4F,$E1
	.dc.b	$78,$FF,$33,$3A,$8C,$77,$7E,$94,$DB,$F1,$7E,$93,$F9,$DE,$BF,$E9
	.dc.b	$F6,$C7,$FB,$5D,$AF,$F8,$73,$E5,$BE,$54,$7F,$53,$8F,$F3,$C3,$5E
	.dc.b	$5F,$23,$FE,$8A,$8C,$47,$38,$FA,$5E,$E7,$F4,$C8,$FF,$13,$FC,$6F
	.dc.b	$49,$FD,$A7,$8F,$27,$F6,$8F,$ED,$FC,$AC,$47,$EA,$7E,$B3,$F7,$46
	.dc.b	$35,$FE,$B7,$FA,$97,$F7,$6D,$7F,$DD,$ED,$AF,$F1,$B3,$7C,$53,$FD
	.dc.b	$02,$B8,$CF,$F8,$D2,$3F,$D0,$F1,$D3,$E1,$71,$39,$F2,$BF,$CB,$15
	.dc.b	$FC,$DE,$3B,$C6,$08,$FF,$7B,$87,$D5,$F5,$39,$46,$7E,$AF,$C0,$47
	.dc.b	$87,$C4,$57,$FC,$5F,$49,$F0,$92,$3F,$C3,$03,$E9,$F4,$E1,$AF,$F8
	.dc.b	$75,$AF,$FA,$C5,$AF,$FB,$85,$DE,$5F,$E3,$1F,$8B,$7D,$47,$F6,$F8
	.dc.b	$C3,$76,$E1,$87,$C9,$FA,$BF,$8A,$4C,$7F,$C5,$9B,$FD,$08,$C4,$7F
	.dc.b	$2F,$CD,$F2,$9F,$44,$E5,$1F,$27,$E5,$3E,$A1,$1F,$99,$15,$E5,$17
	.dc.b	$3F,$C1,$CF,$F9,$BB,$C8,$FF,$38,$50,$FE,$17,$8A,$33,$18,$6E,$7F
	.dc.b	$E9,$76,$C7,$19,$E1,$EF,$F3,$87,$C8,$7F,$4C,$9F,$C6,$FA,$99,$8C
	.dc.b	$69,$FD,$A1,$D6,$1E,$07,$FA,$8F,$D4,$FF,$67,$F5,$BC,$D1,$8F,$07
	.dc.b	$FC,$00,$D8,$CF,$A7,$E1,$AF,$F4,$C6,$7E,$28,$C5,$F0,$6D,$70,$9F
	.dc.b	$85,$5D,$DE,$06,$3E,$3B,$D1,$5C,$23,$22,$78,$61,$E3,$8C,$C7,$8A
	.dc.b	$30,$49,$E4,$FB,$AB,$AD,$78,$B0,$F3,$37,$0F,$33,$DE,$32,$F8,$67
	.dc.b	$8B,$99,$DE,$1E,$4C,$3B,$9F,$16,$B7,$19,$AE,$C7,$A2,$FA,$C7,$A6
	.dc.b	$B8,$CE,$B0,$F5,$3F,$14,$EF,$5C,$09,$87,$8B,$0E,$F1,$89,$ED,$39
	.dc.b	$78,$B1,$C9,$F2,$A7,$8C,$66,$BC,$33,$F0,$4E,$07,$E4,$4F,$84,$7F
	.dc.b	$A6,$70,$FB,$8E,$5A,$FE,$1F,$E5,$BF,$CB,$AF,$E0,$9A,$FD,$D6,$C1
	.dc.b	$A7,$F9,$A2,$7F,$2E,$DC,$7F,$45,$D4,$FA,$4F,$F5,$C5,$6B,$DC,$FA
	.dc.b	$B7,$BC,$7C,$11,$F7,$AF,$87,$16,$A2,$4F,$E3,$2B,$83,$FE,$6B,$C3
	.dc.b	$AE,$3E,$BA,$DE,$3D,$EB,$F2,$F8,$F4,$79,$4E,$AB,$84,$6A,$F9,$BE
	.dc.b	$43,$64,$BF,$F2,$2D,$F3,$7B,$12,$B9,$BA,$AE,$75,$F2,$EF,$C0,$F5
	.dc.b	$9D,$63,$88,$D1,$DE,$FE,$5D,$F2,$CB,$E9,$61,$B6,$D7,$A6,$F1,$97
	.dc.b	$C6,$9C,$E3,$DA,$BA,$1E,$1C,$7C,$E7,$6B,$F2,$F0,$FF,$23,$F8,$7F
	.dc.b	$86,$FF,$66,$CE,$F0,$7F,$BA,$6B,$F8,$54,$E5,$B1,$9C,$3F,$95,$5F
	.dc.b	$FB,$E6,$7F,$95,$36,$3F,$95,$4E,$E7,$F9,$55,$7F,$74,$F8,$DF,$F7
	.dc.b	$5C,$7F,$7A,$6D,$7F,$7A,$F8,$DB,$5B,$AF,$ED,$D5,$8E,$33,$6B,$0E
	.dc.b	$7E,$0B,$15,$FD,$D5,$B8,$7F,$53,$7F,$C9,$EF,$F8,$29,$AF,$F8,$29
	.dc.b	$B8,$7F,$82,$9C,$C3,$EA,$3F,$D1,$F5,$FE,$FE,$EE,$7E,$54,$8F,$5D
	.dc.b	$6F,$7F,$6D,$6A,$73,$7C,$AB,$F7,$C6,$E3,$FB,$FC,$66,$3F,$81,$35
	.dc.b	$FF,$10,$FC,$C7,$9B,$53,$49,$5F,$DD,$23,$2F,$F8,$5E,$3E,$57,$AB
	.dc.b	$F2,$DF,$99,$F4,$7F,$37,$E3,$2D,$7F,$87,$5B,$87,$F8,$5A,$7C,$04
	.dc.b	$9E,$DD,$BF,$57,$8F,$EB,$36,$3F,$C4,$CE,$1F,$F1,$47,$C7,$5B,$FF
	.dc.b	$5A,$7A,$1F,$26,$7F,$BD,$78,$3F,$2E,$5C,$6A,$7C,$DF,$E1,$9E,$13
	.dc.b	$CA,$31,$70,$E6,$7F,$7D,$97,$D4,$E6,$BF,$D1,$9D,$AD,$7F,$27,$93
	.dc.b	$D5,$FF,$0C,$E3,$FE,$E6,$AD,$3D,$2D,$5B,$F1,$3F,$03,$5A,$9F,$BA
	.dc.b	$D7,$F9,$E3,$5F,$F0,$73,$5F,$F0,$D3,$C6,$7F,$D4,$AD,$AF,$8B,$6F
	.dc.b	$1B,$BE,$3F,$AA,$F8,$07,$EA,$2F,$A3,$C2,$FA,$2B,$18,$6E,$3F,$ED
	.dc.b	$39,$E0,$DF,$A3,$CE,$BF,$5B,$C5,$FA,$1E,$4F,$D1,$6D,$7C,$3E,$4F
	.dc.b	$AD,$19,$8C,$71,$BE,$6F,$18,$7F,$AE,$FE,$9B,$E7,$57,$CB,$E3,$7E
	.dc.b	$EB,$FE,$3B,$8F,$9D,$4A,$DB,$F4,$23,$F0,$F3,$B8,$E9,$C7,$FB,$8A
	.dc.b	$DC,$7F,$6B,$6B,$FE,$3D,$6B,$FE,$49,$7A,$DF,$FB,$51,$B7,$F9,$4E
	.dc.b	$27,$AB,$FB,$E9,$CD,$FE,$5B,$37,$D5,$FE,$57,$E8,$FF,$2C,$DC,$7E
	.dc.b	$1B,$87,$E2,$FC,$BD,$BF,$D1,$FD,$4F,$E7,$70,$FC,$F3,$6B,$E5,$B8
	.dc.b	$7E,$6F,$BB,$6B,$E2,$0E,$8F,$C1,$F1,$7E,$0B,$8B,$F0,$6D,$BF,$C0
	.dc.b	$3F,$C8,$A7,$71,$FC,$51,$B5,$FE,$13,$6B,$FE,$65,$73,$C1,$FB,$76
	.dc.b	$CF,$F9,$CD,$FC,$9F,$27,$82,$3F,$07,$EE,$FE,$E7,$E9,$FF,$74,$D8
	.dc.b	$C9,$30,$FE,$0B,$F5,$1A,$8F,$C5,$F5,$FF,$01,$E9,$FE,$04,$D8,$C9
	.dc.b	$2B,$F8,$6F,$5E,$A3,$F1,$7F,$6C,$F9,$BF,$D2,$F8,$3F,$CE,$23,$13
	.dc.b	$FA,$38,$C4,$FF,$42,$9C,$DF,$F3,$86,$BF,$E3,$76,$BF,$E8,$97,$7B
	.dc.b	$FF,$1B,$84,$62,$FE,$37,$77,$F1,$C7,$CB,$7F,$C7,$7D,$7F,$EA,$7C
	.dc.b	$C6,$92,$7F,$D6,$31,$E0,$C3,$F2,$55,$C0,$93,$AF,$A8,$E6,$47,$07
	.dc.b	$FA,$EB,$F5,$47,$B3,$FD,$89,$B1,$98,$C1,$FD,$9B,$EA,$D5,$49,$3F
	.dc.b	$EC,$48,$CB,$FE,$C6,$F9,$97,$C5,$FF,$6E,$6C,$66,$30,$7F,$72,$F7
	.dc.b	$5F,$1F,$FE,$04,$6C,$FF,$0F,$78,$E9,$FA,$0F,$1D,$1A,$8C,$7C,$DB
	.dc.b	$E3,$BF,$BE,$B6,$FF,$DB,$2D,$9F,$F3,$85,$A3,$11,$FE,$1A,$F9,$CF
	.dc.b	$F7,$FD,$7F,$1F,$7F,$DF,$0D,$F8,$7F,$F0,$CC,$62,$7C,$9F,$3B,$F2
	.dc.b	$9B,$C1,$FE,$1E,$87,$C7,$D8,$FF,$10,$FC,$F3,$C1,$FF,$09,$B7,$2F
	.dc.b	$F7,$7B,$70,$F3,$B5,$FF,$17,$F8,$EF,$1F,$FC,$56,$FF,$69,$AF,$E6
	.dc.b	$CE,$5B,$D1,$FD,$83,$F1,$C7,$C5,$D3,$1E,$8D,$AF,$E6,$D3,$F0,$B1
	.dc.b	$FF,$4B,$77,$A3,$11,$FE,$E9,$8C,$70,$FF,$8B,$9B,$5F,$E2,$07,$87
	.dc.b	$C6,$78,$7E,$2B,$C4,$D9,$FE,$99,$F9,$93,$D8,$FF,$22,$B6,$3F,$A8
	.dc.b	$7C,$DB,$E1,$B3,$5F,$E4,$A8,$C4,$FF,$8B,$BA,$51,$8D,$3E,$FB,$5F
	.dc.b	$F2,$BB,$51,$FE,$12,$7F,$CA,$33,$E0,$BE,$DF,$0D,$DD,$FF,$2C,$FD
	.dc.b	$67,$C5,$D7,$FC,$35,$8F,$F7,$7F,$9F,$CE,$FF,$E5,$EE,$14,$63,$2F
	.dc.b	$36,$1F,$2B,$4F,$0E,$08,$C4,$7F,$89,$1B,$19,$C3,$94,$FE,$8E,$7F
	.dc.b	$25,$8F,$89,$F2,$5F,$4F,$9E,$F5,$D7,$EA,$7E,$61,$F3,$DF,$E7,$29
	.dc.b	$CE,$1E,$A9,$CD,$FF,$7F,$9E,$37,$FE,$7E,$6B,$D7,$6F,$F3,$5C,$62
	.dc.b	$3F,$C7,$7F,$6B,$FE,$8B,$7F,$D0,$AD,$7D,$ED,$77,$F0,$7F,$E7,$48
	.dc.b	$FF,$49,$63,$F3,$1B,$87,$0E,$2F,$F4,$4C,$62,$FF,$DD,$5F,$1F,$F6
	.dc.b	$BE,$B3,$F3,$B1,$FE,$4E,$CB,$FC,$9D,$18,$8F,$BE,$27,$87,$4F,$E6
	.dc.b	$EB,$6F,$F9,$E2,$33,$3C,$F1,$AF,$28,$B9,$2F,$8C,$F9,$06,$1E,$19
	.dc.b	$D5,$F2,$9F,$C8,$DF,$2B,$E3,$1D,$4E,$38,$79,$6B,$84,$7A,$4C,$9E
	.dc.b	$52,$9C,$7C,$0F,$05,$7F,$9B,$DF,$85,$1E,$03,$F8,$AD,$75,$69,$E9
	.dc.b	$C7,$8D,$F8,$30,$F3,$36,$7B,$E3,$EC,$7C,$27,$5A,$E2,$E0,$F4,$1F
	.dc.b	$EA,$42,$9F,$3C,$C1,$83,$F6,$BF,$54,$FA,$17,$C6,$F1,$8C,$36,$33
	.dc.b	$3F,$EB,$09,$CB,$78,$BF,$CF,$6D,$C3,$94,$62,$3F,$0A,$27,$F4,$F3
	.dc.b	$A3,$F0,$39,$BA,$FF,$17,$97,$E3,$63,$17,$F8,$6B,$E2,$FD,$46,$CE
	.dc.b	$EB,$8D,$7E,$50,$E3,$5B,$C3,$CF,$5F,$99,$BE,$7A,$F1,$AF,$E3,$2D
	.dc.b	$7E,$C9,$E9,$AF,$7C,$71,$79,$DF,$7C,$3C,$35,$FC,$8A,$3C,$38,$6F
	.dc.b	$1F,$55,$F0,$AD,$5F,$07,$FA,$5B,$71,$F6,$E1,$FA,$E6,$CF,$03,$E4
	.dc.b	$CE,$EB,$A4,$79,$EF,$2C,$3E,$13,$FB,$2E,$6F,$F4,$77,$C3,$FF,$51
	.dc.b	$DB,$F9,$4F,$BB,$FE,$C8,$BF,$F4,$B3,$5F,$F0,$43,$FE,$C4,$73,$3F
	.dc.b	$91,$AE,$71,$FD,$26,$78,$3F,$C1,$A3,$2F,$F8,$23,$2F,$F4,$0B,$FD
	.dc.b	$3D,$AE,$7D,$FE,$91,$88,$E4,$EB,$1F,$E8,$FF,$57,$FE,$55,$9F,$F5
	.dc.b	$63,$5F,$F0,$F4,$EF,$2F,$E8,$BF,$36,$E7,$F8,$15,$FE,$E8,$8C,$3F
	.dc.b	$EB,$56,$BF,$E2,$77,$FB,$63,$98,$F8,$11,$A8,$F7,$35,$EF,$1A,$3B
	.dc.b	$E4,$FA,$BF,$F6,$84,$FF,$B0,$DA,$F7,$9D,$FD,$53,$E5,$2F,$C7,$CA
	.dc.b	$FF,$65,$4E,$5B,$1F,$4E,$0D,$24,$7F,$7D,$FA,$B7,$CF,$79,$1B,$1F
	.dc.b	$ED,$56,$BF,$E4,$7A,$E6,$FF,$B6,$7E,$FB,$FC,$66,$DE,$8F,$6C,$62
	.dc.b	$3F,$C9,$CB,$E8,$7E,$0F,$C8,$FF,$3D,$9F,$F0,$73,$5F,$F2,$5B,$5F
	.dc.b	$C4,$35,$FE,$FB,$F6,$FF,$E3,$77,$FD,$D4,$BC,$DF,$9A,$D8,$CF,$9F
	.dc.b	$F4,$3E,$AD,$F1,$34,$93,$FE,$F3,$87,$C4,$7D,$6F,$4B,$FC,$EF,$3F
	.dc.b	$E2,$36,$BF,$E6,$06,$BF,$E4,$AF,$7F,$F4,$5E,$A7,$68,$FF,$7F,$3D
	.dc.b	$1B,$5F,$82,$8D,$47,$F9,$83,$DD,$FC,$2F,$B1,$FE,$80,$8F,$F1,$5F
	.dc.b	$77,$F8,$92,$31,$1F,$E6,$6F,$A4,$FC,$5F,$B5,$F5,$23,$FC,$E6,$D7
	.dc.b	$FC,$E5,$D1,$FB,$58,$C4,$7F,$89,$BD,$BF,$F7,$AB,$DB,$BF,$FC,$53
	.dc.b	$F8,$37,$F9,$59,$B1,$F9,$89,$DB,$FE,$89,$FA,$CF,$E5,$AF,$EF,$5A
	.dc.b	$FF,$8C,$9A,$BC,$04,$61,$7C,$0B,$E9,$7F,$85,$36,$BF,$D4,$51,$83
	.dc.b	$F8,$77,$D4,$FF,$AC,$23,$FC,$AC,$D7,$FD,$26,$D7,$F8,$83,$5F,$F2
	.dc.b	$2F,$D5,$BD,$8F,$E3,$0D,$8F,$F3,$03,$5F,$F4,$EB,$5F,$E3,$5F,$0D
	.dc.b	$F9,$D6,$F8,$1F,$C8,$63,$11,$FE,$52,$6B,$F0,$5F,$D1,$3F,$EC,$9E
	.dc.b	$FE,$04,$7F,$BD,$79,$3F,$3B,$97,$FD,$C4,$DE,$0F,$F1,$7B,$65,$3F
	.dc.b	$BA,$F9,$BF,$E3,$70,$9F,$AE,$FF,$72,$6B,$CB,$E5,$BF,$C0,$B1,$88
	.dc.b	$FF,$3C,$BD,$6B,$F8,$2F,$8F,$FC,$99,$B3,$F8,$73,$75,$FC,$73,$27
	.dc.b	$CF,$D2,$4E,$F8,$93,$FE,$7A,$8C,$3C,$1B,$B7,$33,$8B,$FE,$D7,$8F
	.dc.b	$E0,$D8,$7F,$A1,$9B,$A7,$CF,$F5,$FC,$5F,$01,$E3,$7C,$9E,$94,$C3
	.dc.b	$BC,$7F,$40,$F9,$37,$D4,$FE,$B3,$26,$3F,$E9,$A9,$CF,$7E,$EB,$53
	.dc.b	$95,$F3,$59,$9F,$F1,$95,$70,$C7,$FC,$42,$1C,$3F,$E2,$46,$BF,$E6
	.dc.b	$A7,$38,$78,$A3,$85,$BB,$FC,$47,$78,$F9,$3E,$CF,$D1,$23,$F3,$3E
	.dc.b	$8A,$BF,$27,$EA,$FE,$B1,$8C,$3E,$57,$C5,$FD,$92,$4F,$D4,$FA,$BA
	.dc.b	$FD,$1F,$8E,$FB,$C6,$2B,$CD,$1F,$2B,$1F,$C8,$E7,$FE,$4D,$8F,$C3
	.dc.b	$19,$2F,$C2,$7F,$6C,$AE,$33,$B9,$F8,$ED,$7F,$C9,$2D,$7F,$D1,$ED
	.dc.b	$7F,$D9,$EF,$18,$F1,$CF,$F7,$69,$FD,$2F,$07,$FA,$5A,$31,$3E,$0B
	.dc.b	$FC,$C6,$F2,$FF,$26,$FE,$6F,$6C,$47,$E4,$78,$3F,$3C,$62,$FE,$B7
	.dc.b	$07,$E2,$B9,$7F,$0C,$4A,$FD,$34,$7E,$3B,$93,$11,$E4,$3E,$EF,$1F
	.dc.b	$E3,$4C,$DF,$4F,$AB,$FF,$0D,$37,$E4,$7F,$31,$18,$DF,$F2,$5F,$1F
	.dc.b	$89,$FC,$8C,$FF,$88,$70,$FC,$63,$5F,$CE,$35,$FC,$A3,$CD,$BE,$6F
	.dc.b	$63,$A3,$FB,$6F,$63,$F7,$C4,$7F,$73,$E2,$D7,$F6,$DD,$DF,$C0,$CC
	.dc.b	$67,$FE,$FC,$FD,$57,$FA,$64,$FF,$43,$6F,$71,$A3,$E8,$7F,$C4,$EE
	.dc.b	$3E,$0A,$98,$AC,$FD,$4F,$F9,$46,$4D,$24,$7E,$CA,$FB,$47,$4F,$22
	.dc.b	$4F,$F1,$5A,$E2,$DF,$90,$F0,$91,$D7,$C2,$D7,$FD,$3B,$87,$A6,$7C
	.dc.b	$44,$8F,$F5,$A7,$D6,$7F,$49,$7F,$DC,$ED,$41,$C3,$70,$FF,$5C,$C7
	.dc.b	$F5,$3D,$3F,$82,$7C,$43,$EB,$3F,$88,$DF,$F3,$38,$C7,$5F,$F4,$2B
	.dc.b	$E1,$CF,$DB,$3F,$9D,$F2,$1F,$1C,$98,$7C,$5D,$91,$88,$FC,$B4,$FD
	.dc.b	$5F,$7F,$E9,$9F,$52,$6C,$FC,$98,$AF,$0A,$E6,$E4,$F9,$DF,$59,$F7
	.dc.b	$0D,$EB,$F8,$5F,$96,$F8,$73,$FC,$33,$1F,$C2,$46,$32,$FF,$45,$60
	.dc.b	$8C,$57,$B6,$30,$FC,$C9,$DC,$F4,$F8,$DF,$E9,$D7,$F8,$FF,$AB,$C7
	.dc.b	$9D,$C5,$B8,$7E,$56,$EA,$FA,$C7,$CC,$7F,$84,$71,$FE,$B1,$39,$8C
	.dc.b	$61,$EA,$AD,$F8,$1F,$7A,$F7,$0F,$9A,$EA,$EB,$04,$EB,$0E,$71,$F8
	.dc.b	$69,$D4,$F2,$AE,$35,$AA,$F1,$5F,$7C,$3C,$53,$9C,$3C,$15,$93,$A4
	.dc.b	$64,$95,$98,$F8,$17,$DA,$31,$5E,$2B,$EB,$18,$9D,$57,$AA,$72,$D3
	.dc.b	$BC,$ED,$E7,$7E,$07,$47,$A8,$B7,$E8,$6C,$6F,$0F,$8D,$5C,$E7,$2F
	.dc.b	$42,$3C,$B0,$F2,$CF,$08,$D9,$93,$6E,$49,$87,$DB,$9E,$2F,$B2,$33
	.dc.b	$3F,$56,$7E,$0B,$F9,$A3,$D4,$5C,$3E,$F3,$70,$FF,$2A,$3F,$8D,$FA
	.dc.b	$2F,$E4,$0F,$F8,$22,$51,$88,$E4,$D7,$F1,$CD,$7F,$0C,$F2,$BF,$C7
	.dc.b	$B7,$0F,$D6,$F5,$34,$FF,$07,$56,$BB,$3E,$ED,$EF,$0E,$71,$F8,$3B
	.dc.b	$E1,$BE,$A2,$4F,$D6,$AE,$0F,$F3,$2C,$3A,$E3,$E6,$AD,$E3,$DE,$BE
	.dc.b	$FE,$3D,$1E,$53,$AA,$E1,$1A,$BE,$6F,$90,$D9,$2F,$F6,$77,$CD,$EC
	.dc.b	$4A,$E6,$EA,$B9,$D7,$D1,$BF,$03,$D6,$75,$8E,$23,$47,$7B,$F9,$77
	.dc.b	$CB,$2F,$AF,$86,$DB,$5E,$9B,$C6,$5F,$1A,$73,$8F,$6A,$E8,$78,$71
	.dc.b	$F3,$9D,$AF,$CB,$BF,$F6,$7F,$95,$F8,$6F,$F6,$9F,$98,$BC,$49,$B1
	.dc.b	$87,$FC,$99,$5F,$CA,$9B,$87,$F2,$96,$BF,$DD,$5B,$7F,$DD,$7D,$5F
	.dc.b	$F2,$A7,$27,$F7,$BF,$80,$B8,$7F,$31,$9B,$87,$F7,$D9,$B7,$C6,$BF
	.dc.b	$B7,$56,$32,$7D,$4D,$8C,$36,$7F,$BA,$37,$0F,$E9,$8F,$F9,$41,$FE
	.dc.b	$80,$D7,$FC,$18,$DB,$FF,$06,$39,$87,$D5,$BE,$6B,$CF,$DC,$FA,$C7
	.dc.b	$CB,$FA,$AF,$75,$A9,$CE,$5F,$E5,$CC,$39,$3F,$E7,$EA,$DB,$63,$31
	.dc.b	$FE,$82,$6B,$FD,$6B,$EA,$7C,$DA,$9A,$4A,$FE,$E8,$FC,$73,$E4,$7B
	.dc.b	$D6,$3F,$C3,$1F,$57,$F2,$FE,$53,$E6,$36,$32,$7F,$7D,$7B,$91,$FF
	.dc.b	$00,$7C,$E7,$D6,$FD,$5B,$FD,$43,$1F,$E2,$38,$C3,$F5,$D8,$FF,$AE
	.dc.b	$E3,$E1,$E5,$FD,$FE,$79,$7D,$27,$3E,$BB,$7B,$6D,$7F,$83,$2D,$5E
	.dc.b	$1E,$FC,$F3,$FF,$9B,$3D,$6F,$EC,$D8,$FF,$9A,$97,$A3,$F8,$EF,$59
	.dc.b	$1A,$CB,$FC,$D9,$18,$9D,$4E,$67,$FD,$5C,$D7,$FD,$7C,$D7,$FD,$6C
	.dc.b	$D9,$EB,$E1,$F2,$39,$FA,$97,$0F,$27,$D4,$F6,$F9,$2E,$ED,$8C,$93
	.dc.b	$1F,$2E,$2F,$A9,$DC,$FC,$BF,$9E,$FA,$5F,$3D,$F4,$DB,$19,$25,$FC
	.dc.b	$9F,$AA,$D4,$7F,$AB,$BE,$AD,$EC,$7E,$CB,$85,$F8,$03,$E1,$FF,$39
	.dc.b	$5A,$BF,$F7,$D5,$6A,$73,$7F,$A7,$AD,$63,$FE,$DA,$6B,$FD,$75,$AF
	.dc.b	$F8,$E1,$DD,$7F,$B1,$1B,$7F,$E5,$E8,$F1,$FC,$47,$68,$FC,$57,$CB
	.dc.b	$FE,$47,$E6,$7F,$24,$D8,$C9,$30,$FC,$47,$CD,$EA,$3F,$25,$C5,$F9
	.dc.b	$9E,$1F,$CD,$36,$32,$4A,$FC,$87,$1E,$A3,$F1,$9E,$2F,$E1,$BB,$3F
	.dc.b	$0E,$D8,$C9,$FE,$FA,$E8,$FC,$7C,$7F,$0F,$6C,$67,$9B,$F4,$78,$7E
	.dc.b	$3A,$B9,$CF,$E6,$7A,$EC,$FE,$87,$E6,$7F,$61,$F5,$7F,$B1,$6D,$7E
	.dc.b	$B5,$C3,$FA,$FF,$AC,$B5,$FA,$9F,$AC,$FD,$C7,$6F,$EE,$5B,$5F,$B5
	.dc.b	$70,$FE,$DF,$B6,$D7,$FB,$44,$7D,$67,$EB,$27,$F0,$7F,$80,$3B,$FF
	.dc.b	$91,$4E,$6B,$FD,$B1,$D2,$AD,$5F,$FB,$FA,$B5,$5F,$E2,$16,$BF,$E7
	.dc.b	$C6,$BF,$EA,$06,$BE,$2F,$03,$B9,$F9,$9A,$78,$18,$BF,$E2,$13,$FC
	.dc.b	$F7,$C2,$E7,$E7,$1A,$63,$D7,$7B,$3F,$D0,$DB,$7F,$EC,$36,$E5,$FE
	.dc.b	$BA,$6B,$FE,$92,$BF,$F1,$37,$CB,$71,$25,$71,$7F,$A7,$76,$F8,$89
	.dc.b	$96,$0F,$EA,$5F,$58,$E1,$A4,$BF,$F6,$A4,$65,$FF,$6B,$7C,$CB,$E4
	.dc.b	$BF,$AD,$36,$33,$18,$3F,$AF,$7B,$CF,$94,$FE,$DE,$BE,$52,$C6,$1F
	.dc.b	$EE,$7F,$37,$FC,$0E,$FF,$DC,$8D,$DF,$FD,$B6,$D7,$E1,$35,$FF,$65
	.dc.b	$7C,$EB,$E3,$3F,$BE,$B6,$BF,$7C,$6B,$E0,$F3,$3F,$E0,$2F,$C8,$BD
	.dc.b	$D7,$FE,$CC,$7F,$19,$D4,$BF,$F6,$B4,$6A,$31,$2E,$44,$F0,$AF,$F6
	.dc.b	$E7,$D0,$7F,$99,$9F,$F7,$F3,$6B,$FD,$EA,$D7,$EE,$35,$FF,$72,$79
	.dc.b	$4F,$47,$FC,$2E,$DA,$F5,$CF,$86,$06,$A7,$FC,$3B,$F3,$3F,$E5,$9C
	.dc.b	$BC,$F9,$23,$FC,$65,$87,$28,$C7,$4F,$0F,$81,$FF,$12,$FD,$FF,$FB
	.dc.b	$6E,$BF,$D1,$19,$FB,$E3,$0F,$FA,$2E,$D1,$88,$FC,$B3,$5F,$F7,$F7
	.dc.b	$DF,$FF,$B5,$AB,$C0,$D8,$FF,$7D,$BF,$AC,$8C,$FA,$DF,$E3,$6F,$AE
	.dc.b	$7B,$54,$FF,$1C,$FC,$BF,$B6,$98,$7D,$3A,$DC,$7B,$67,$FC,$F5,$3F
	.dc.b	$A9,$6B,$C3,$24,$62,$3E,$E3,$63,$3D,$3E,$0D,$7C,$2D,$BF,$95,$F5
	.dc.b	$DF,$8F,$C1,$C1,$F4,$3E,$39,$F1,$3F,$E5,$0A,$D4,$E6,$31,$CF,$FE
	.dc.b	$45,$9E,$57,$FB,$F6,$BE,$76,$BF,$91,$79,$E1,$99,$ED,$D1,$D2,$3E
	.dc.b	$F7,$CC,$7C,$7B,$FC,$C7,$CE,$7C,$09,$CE,$BF,$E6,$48,$C4,$FA,$A7
	.dc.b	$E3,$57,$C3,$25,$7F,$96,$DF,$8D,$F5,$CF,$82,$FF,$2F,$36,$BF,$13
	.dc.b	$C0,$8C,$61,$F8,$96,$BF,$C7,$5C,$ED,$F9,$86,$D7,$D7,$5F,$05,$FE
	.dc.b	$7A,$73,$5F,$ED,$DF,$8D,$FC,$1F,$D7,$3E,$63,$F0,$B3,$F5,$34,$FD
	.dc.b	$CC,$E6,$38,$07,$0E,$F1,$E7,$5F,$A2,$BB,$3F,$E0,$9B,$E7,$39,$8E
	.dc.b	$4F,$83,$24,$EA,$7B,$E1,$DA,$73,$87,$49,$F3,$E1,$67,$31,$D0,$E3
	.dc.b	$87,$5C,$3A,$47,$03,$47,$A4,$A7,$E9,$AB,$F4,$FE,$5F,$86,$7F,$8D
	.dc.b	$BF,$02,$74,$1D,$61,$C2,$FD,$D8,$E2,$BA,$8C,$3D,$18,$79,$23,$9C
	.dc.b	$EC,$EC,$EC,$F3,$E3,$EE,$8C,$D7,$FA,$66,$7C,$67,$FA,$6D,$C3,$C4
	.dc.b	$FB,$A3,$EB,$9F,$3E,$F8,$DF,$F5,$14,$61,$B1,$99,$FE,$3B,$39,$6F
	.dc.b	$81,$FB,$A6,$E3,$CA,$31,$1E,$09,$FE,$59,$39,$3E,$D6,$7F,$D4,$E6
	.dc.b	$F2,$FE,$4B,$3F,$C6,$DF,$09,$FA,$ED,$8F,$C8,$D7,$19,$FB,$6D,$8E
	.dc.b	$38,$79,$5B,$87,$D5,$AF,$E4,$75,$C7,$6E,$78,$FE,$89,$DC,$FD,$D9
	.dc.b	$FD,$2E,$3E,$17,$F4,$E7,$72,$D7,$A5,$DE,$3E,$16,$E1,$EE,$B8,$7B
	.dc.b	$27,$4F,$F3,$6B,$EC,$F9,$AF,$A3,$4F,$D1,$46,$9F,$B9,$3F,$4F,$0F
	.dc.b	$5C,$7B,$CF,$DF,$57,$B7,$0F,$1C,$7B,$4F,$EE,$DE,$8B,$DE,$FD,$A7
	.dc.b	$07,$F4,$18,$C4,$FF,$63,$6B,$FC,$84,$E3,$DF,$C1,$E7,$5F,$EE,$DE
	.dc.b	$4F,$F2,$1B,$6B,$8C,$FF,$8D,$DA,$FF,$4F,$D3,$FB,$E3,$FC,$91,$AF
	.dc.b	$F6,$B6,$BF,$E3,$F6,$BC,$7C,$3F,$15,$BE,$8B,$FD,$A3,$3F,$DE,$DA
	.dc.b	$FA,$63,$F4,$93,$C3,$D6,$FF,$0C,$CE,$5B,$C1,$FE,$06,$6E,$9F,$D9
	.dc.b	$57,$BB,$C6,$3E,$04,$72,$AF,$F2,$A3,$5E,$11,$A3,$CF,$F5,$AF,$6E
	.dc.b	$9C,$E7,$FC,$25,$F5,$8F,$91,$F6,$B6,$33,$18,$F8,$DC,$BF,$46,$75
	.dc.b	$3F,$E1,$36,$BF,$E6,$76,$BF,$DF,$FE,$B5,$F3,$FF,$65,$B1,$FD,$59
	.dc.b	$AF,$F8,$75,$AF,$F9,$C7,$EB,$7F,$BD,$BF,$D7,$1A,$FF,$88,$DA,$EF
	.dc.b	$C6,$FC,$37,$CD,$FF,$83,$A3,$FB,$1B,$5F,$F1,$53,$5F,$F3,$F3,$5F
	.dc.b	$F0,$B7,$D6,$BE,$2B,$F2,$0D,$8F,$ED,$AD,$7F,$C6,$8D,$7F,$D1,$1F
	.dc.b	$5A,$F8,$7F,$CC,$36,$3F,$BA,$B5,$FF,$1E,$35,$FE,$5B,$E2,$FE,$69
	.dc.b	$FF,$17,$B5,$FE,$F9,$D7,$FC,$DE,$35,$D0,$F5,$79,$57,$F9,$93,$0F
	.dc.b	$D4,$46,$23,$FC,$D4,$D9,$7D,$6B,$E8,$3F,$5C,$DA,$FF,$05,$B5,$FF
	.dc.b	$29,$35,$F3,$FC,$57,$8C,$8F,$F9,$68,$FF,$09,$3F,$B9,$FA,$D7,$CF
	.dc.b	$D2,$4F,$F8,$62,$32,$FF,$86,$BE,$B5,$E8,$FF,$01,$6C,$66,$30,$7F
	.dc.b	$59,$FB,$6F,$F6,$3C,$78,$5A,$FF,$9B,$1A,$FF,$5A,$6B,$FE,$55,$FA
	.dc.b	$EF,$EE,$4F,$F8,$A9,$AF,$F9,$C5,$AF,$FB,$15,$AF,$F9,$77,$EB,$6B
	.dc.b	$FE,$97,$6E,$A6,$9A,$85,$3C,$7F,$51,$23,$D4,$FF,$21,$F4,$DF,$5D
	.dc.b	$49,$3F,$ED,$B7,$FB,$D7,$DE,$FC,$68,$FF,$7F,$37,$C6,$FE,$26,$DF
	.dc.b	$13,$FC,$6F,$EB,$3B,$3F,$98,$5A,$B9,$3F,$E0,$06,$BF,$E3,$E6,$BF
	.dc.b	$E8,$46,$BC,$77,$E7,$4F,$A6,$A7,$6A,$F1,$57,$FA,$F7,$3F,$DE,$E5
	.dc.b	$FE,$AA,$6E,$5F,$E8,$96,$CE,$FE,$8B,$F9,$AC,$63,$80,$D2,$47,$F3
	.dc.b	$AF,$9A,$F7,$57,$FA,$CD,$F2,$37,$2D,$6B,$23,$04,$3F,$9F,$4F,$27
	.dc.b	$FA,$C7,$CB,$BE,$F5,$F5,$9F,$D6,$63,$0D,$8C,$CF,$FB,$32,$78,$7C
	.dc.b	$F6,$6B,$F5,$11,$C2,$7F,$D8,$6D,$7F,$C6,$2D,$7F,$CD,$EE,$76,$FF
	.dc.b	$73,$BE,$3F,$1F,$C0,$75,$9F,$06,$0F,$AE,$F2,$12,$33,$DD,$5F,$6C
	.dc.b	$ED,$E3,$3E,$B9,$FA,$3F,$37,$F4,$BD,$D9,$5F,$ED,$FE,$14,$EA,$31
	.dc.b	$3A,$7F,$B4,$4E,$9F,$F7,$78,$FA,$E7,$CD,$FE,$51,$B1,$A6,$BF,$E0
	.dc.b	$91,$F6,$2F,$4F,$85,$7F,$97,$6B,$75,$FE,$5A,$6B,$FE,$98,$6B,$FE
	.dc.b	$DE,$7A,$EF,$FD,$EA,$FF,$C1,$2D,$F5,$7A,$D7,$BF,$E1,$BF,$27,$EE
	.dc.b	$C8,$C1,$0F,$7B,$97,$E0,$C7,$E1,$FC,$0F,$EF,$11,$F8,$56,$BF,$8E
	.dc.b	$6D,$FD,$E5,$F3,$9C,$7D,$69,$5C,$5E,$91,$F2,$7C,$29,$3F,$25,$FF
	.dc.b	$17,$5B,$F1,$6F,$D1,$CF,$CB,$D7,$F8,$A3,$FC,$ED,$AF,$FA,$08,$FA
	.dc.b	$D5,$F3,$48,$8F,$F1,$E5,$3F,$16,$FF,$37,$2F,$03,$4E,$7D,$FF,$2D
	.dc.b	$7E,$BB,$E2,$7F,$70,$47,$F7,$1C,$95,$D1,$FB,$B8,$FD,$F4,$7F,$04
	.dc.b	$F8,$39,$3F,$93,$7F,$83,$F7,$E2,$3F,$23,$17,$0F,$E2,$13,$FC,$33
	.dc.b	$E9,$3F,$98,$12,$BF,$DD,$B6,$8C,$47,$E6,$BE,$57,$F8,$99,$C5,$FD
	.dc.b	$17,$CA,$FF,$16,$23,$FA,$5F,$96,$AF,$FA,$B3,$C7,$FE,$33,$3F,$CF
	.dc.b	$C9,$1F,$DA,$FE,$BB,$F9,$69,$3C,$BF,$F5,$C3,$63,$F5,$2F,$F6,$A7
	.dc.b	$F4,$8D,$D3,$FC,$A6,$FF,$30,$BF,$E4,$FF,$65,$FD,$34,$95,$EE,$6B
	.dc.b	$FB,$6F,$0B,$F9,$79,$FC,$C2,$3F,$98,$F8,$52,$7F,$80,$3A,$69,$F2
	.dc.b	$1F,$EB,$44,$7F,$15,$4F,$A9,$FE,$73,$1F,$89,$E4,$FC,$52,$9D,$70
	.dc.b	$62,$73,$18,$9E,$07,$E4,$5C,$3F,$E7,$26,$FC,$67,$F9,$31,$FE,$8F
	.dc.b	$E4,$76,$8C,$5F,$F3,$6A,$FF,$77,$7B,$DC,$F1,$F2,$C6,$0F,$CF,$7D
	.dc.b	$7F,$F7,$9F,$A0,$FF,$80,$15,$E8,$8C,$4E,$C3,$8C,$3F,$CB,$6F,$F1
	.dc.b	$A6,$F5,$AF,$9D,$76,$8C,$3C,$EB,$95,$F7,$F1,$7D,$4F,$86,$B7,$87
	.dc.b	$4C,$39,$E1,$E2,$9C,$D7,$18,$CE,$1D,$1F,$05,$62,$B7,$1F,$03,$0E
	.dc.b	$EE,$8F,$24,$69,$ED,$3D,$27,$71,$83,$71,$D4,$D3,$4C,$93,$0F,$8F
	.dc.b	$3C,$AF,$CD,$3C,$E7,$DC,$FC,$59,$D5,$F8,$EB,$64,$8C,$DF,$7B,$E2
	.dc.b	$76,$BC,$61,$F2,$E6,$C7,$C1,$29,$E1,$8D,$3E,$13,$D4,$7E,$68,$F8
	.dc.b	$05,$BF,$C2,$B7,$0F,$BA,$E7,$DD,$FD,$D3,$F8,$B6,$BF,$90,$6B,$F8
	.dc.b	$C6,$CF,$FB,$C7,$D8,$FB,$71,$FE,$6C,$C7,$FD,$53,$CB,$FE,$26,$E0
	.dc.b	$CD,$4C,$3F,$19,$7B,$CB,$F6,$37,$B9,$CD,$6F,$0F,$CC,$DE,$E7,$FA
	.dc.b	$05,$6E,$73,$5B,$BF,$53,$F9,$29,$D4,$62,$BA,$92,$BC,$FB,$7F,$47
	.dc.b	$8C,$4E,$E7,$AC,$7E,$B6,$B9,$BF,$16,$FC,$0D,$35,$7C,$EB,$89,$2F
	.dc.b	$A1,$F0,$23,$4E,$27,$94,$ED,$F7,$CE,$9C,$CE,$DF,$A9,$5E,$F9,$D4
	.dc.b	$60,$F5,$9C,$0E,$78,$FA,$1D,$5F,$77,$FD,$47,$3C,$B2,$FB,$D1,$FC
	.dc.b	$4F,$83,$F9,$57,$D7,$F8,$DF,$F3,$1F,$9B,$78,$53,$9C,$3F,$DE,$B9
	.dc.b	$7F,$0E,$BE,$2F,$F2,$AC,$3F,$98,$5F,$08,$C4,$EA,$7F,$96,$4E,$5F
	.dc.b	$E5,$D4,$F3,$FE,$C4,$7F,$36,$8C,$5F,$F6,$A9,$CC,$FF,$9F,$5A,$FF
	.dc.b	$A0,$DF,$C0,$74,$7F,$A0,$DB,$3F,$CF,$7F,$57,$FE,$7F,$39,$8F,$F0
	.dc.b	$59,$9D,$B7,$7F,$D2,$6B,$73,$93,$71,$93,$79,$27,$FD,$AF,$3F,$D5
	.dc.b	$9B,$3F,$DD,$DA,$FF,$52,$6B,$FD,$63,$EB,$7F,$CC,$33,$FE,$5C,$5E
	.dc.b	$4F,$92,$51,$A7,$FC,$B4,$EA,$78,$79,$FF,$12,$76,$D9,$E1,$1F,$C3
	.dc.b	$5A,$FC,$5F,$13,$E5,$E3,$EB,$6E,$26,$93,$0F,$87,$3B,$6C,$FF,$BC
	.dc.b	$9B,$3F,$1A,$1F,$65,$7E,$CF,$0D,$AE,$2F,$80,$8F,$CB,$7C,$7F,$21
	.dc.b	$CC,$E9,$F2,$F6,$7C,$31,$72,$6F,$FC,$0F,$3F,$AD,$79,$8C,$7D,$17
	.dc.b	$FA,$07,$0F,$5B,$6F,$8C,$7F,$81,$61,$E8,$FE,$01,$D5,$72,$F4,$F0
	.dc.b	$66,$6D,$7F,$A5,$BF,$30,$76,$FE,$3A,$DF,$8B,$3E,$2B,$EF,$56,$AF
	.dc.b	$F1,$F5,$A9,$CD,$FD,$DA,$D6,$7F,$E6,$D6,$BF,$D2,$5A,$FF,$89,$1D
	.dc.b	$D7,$EB,$9B,$7F,$E6,$CF,$C5,$9F,$1F,$DA,$3F,$DA,$1F,$81,$7E,$8B
	.dc.b	$07,$B8,$8C,$37,$AB,$F5,$2E,$66,$CE,$89,$38,$8F,$C2,$7B,$BF,$85
	.dc.b	$F7,$AE,$1F,$8A,$6C,$FD,$23,$F6,$8E,$34,$FB,$9D,$08,$FD,$36,$1F
	.dc.b	$D4,$67,$F8,$1F,$95,$F9,$9C,$3F,$D8,$F3,$97,$FD,$17,$19,$C3,$FD
	.dc.b	$89,$3F,$D6,$5F,$EE,$4D,$7F,$C9,$AD,$7F,$D3,$0D,$4F,$57,$3D,$FF
	.dc.b	$7E,$7F,$81,$DF,$F1,$EA,$D7,$3F,$F0,$9F,$53,$F4,$D7,$FA,$76,$E3
	.dc.b	$FA,$C6,$E3,$F0,$EB,$84,$FE,$C2,$BE,$8F,$E0,$4F,$83,$A6,$3C,$8B
	.dc.b	$5F,$C6,$B8,$55,$B8,$C3,$71,$FA,$13,$AA,$E3,$18,$F8,$17,$D4,$F8
	.dc.b	$3E,$67,$F0,$31,$8C,$79,$C6,$27,$F9,$BE,$0A,$D5,$FF,$87,$E3,$13
	.dc.b	$A9,$CC,$FF,$9F,$27,$37,$E9,$9E,$3B,$FF,$7F,$9F,$5F,$9D,$DE,$BF
	.dc.b	$77,$F4,$9F,$B7,$FB,$1F,$DC,$36,$32,$7F,$00,$8C,$B4,$FE,$37,$F4
	.dc.b	$BF,$C6,$4F,$E3,$5E,$65,$8C,$93,$3F,$A9,$CF,$FD,$77,$3F,$CA,$DF
	.dc.b	$F3,$9F,$B2,$FE,$71,$1F,$D5,$BE,$61,$EF,$67,$CC,$FE,$09,$18,$CB
	.dc.b	$FA,$7E,$4C,$3F,$9B,$BF,$EE,$76,$B9,$AF,$E8,$11,$83,$BC,$FF,$A9
	.dc.b	$1B,$3F,$C9,$29,$F0,$BE,$1A,$FE,$73,$F6,$DF,$D0,$08,$FF,$5B,$7F
	.dc.b	$A6,$7E,$05,$FC,$32,$FF,$DF,$0D,$4D,$35,$D4,$EA,$FF,$4B,$5F,$FB
	.dc.b	$93,$8F,$FA,$7C,$62,$FF,$DB,$8D,$7F,$D6,$FF,$5C,$FA,$EF,$EA,$AD
	.dc.b	$BF,$F6,$1B,$67,$FD,$C8,$D7,$FC,$D5,$8B,$E4,$FD,$24,$AF,$F3,$74
	.dc.b	$FA,$A3,$FB,$0F,$D7,$3E,$06,$92,$BF,$DA,$31,$97,$FD,$A7,$E0,$7F
	.dc.b	$BF,$5B,$F0,$9F,$DA,$A3,$18,$7F,$BD,$DA,$FF,$9F,$75,$7D,$17,$F7
	.dc.b	$76,$DF,$92,$7F,$D7,$93,$93,$FB,$AF,$D6,$FF,$82,$23,$FD,$BC,$DE
	.dc.b	$03,$4C,$7D,$6B,$E0,$7F,$BE,$B7,$0F,$F7,$5B,$6F,$FD,$21,$68,$C4
	.dc.b	$79,$FE,$07,$CC,$4A,$F3,$64,$8F,$F0,$4F,$7F,$F6,$9A,$FF,$77,$B5
	.dc.b	$FF,$77,$35,$FF,$7B,$37,$0F,$C9,$E9,$FE,$12,$E6,$34,$98,$FD,$89
	.dc.b	$7D,$6F,$F4,$18,$C3,$70,$F5,$B5,$FF,$80,$2D,$7F,$E0,$07,$D6,$FE
	.dc.b	$58,$9E,$8F,$F8,$89,$FC,$CF,$D6,$3E,$AB,$FC,$4A,$DC,$3C,$4D,$BF
	.dc.b	$92,$D4,$4C,$7F,$C7,$7E,$7F,$F8,$BB,$DF,$34,$97,$ED,$8C,$BF,$E2
	.dc.b	$E6,$D7,$F4,$17,$DD,$E1,$7F,$87,$5C,$BF,$E1,$FF,$AC,$FE,$F1,$83
	.dc.b	$DB,$FE,$79,$39,$BF,$F2,$CC,$63,$0F,$35,$FF,$9E,$E6,$9F,$E4,$47
	.dc.b	$0A,$D1,$86,$C7,$F9,$92,$3C,$27,$F6,$09,$E6,$7F,$92,$3E,$57,$EE
	.dc.b	$BF,$E4,$AF,$98,$47,$F9,$B0,$CD,$7F,$5F,$8C,$CF,$1E,$1F,$F2,$BD
	.dc.b	$6A,$BE,$79,$E5,$70,$FD,$16,$BE,$9F,$A0,$FF,$0F,$D7,$E2,$DB,$1F
	.dc.b	$A2,$6B,$D3,$14,$62,$3E,$E3,$C3,$3F,$F2,$73,$6B,$3E,$27,$FA,$1A
	.dc.b	$31,$5F,$EF,$DF,$88,$FC,$2F,$A0,$F8,$0F,$4E,$3E,$EA,$4E,$5E,$06
	.dc.b	$27,$79,$FF,$B7,$E7,$53,$F0,$AF,$F2,$3F,$55,$8C,$3A,$3A,$9F,$CA
	.dc.b	$CE,$B0,$E7,$9D,$9E,$F5,$C7,$1F,$F7,$63,$C6,$17,$B7,$FC,$DF,$3A
	.dc.b	$BF,$E5,$B3,$9C,$3C,$31,$8A,$5F,$5C,$39,$47,$90,$F7,$1F,$AA,$AF
	.dc.b	$89,$8F,$87,$0F,$93,$87,$A2,$B6,$7C,$13,$AB,$9C,$38,$46,$0E,$98
	.dc.b	$F9,$DF,$7C,$EC,$F3,$4F,$96,$38,$1D,$4F,$E6,$D5,$FC,$DD,$F4,$18
	.dc.b	$3F,$A6,$FD,$57,$F7,$EF,$AB,$FE,$78,$D8,$CF,$9A,$9C,$B6,$38,$F5
	.dc.b	$FF,$A6,$9B,$A7,$28,$C4,$7E,$80,$5F,$F7,$C9,$D1,$F5,$72,$FF,$5D
	.dc.b	$4D,$F8,$6F,$D3,$E7,$F8,$8B,$E1,$3E,$E6,$C7,$E3,$2B,$8C,$FC,$26
	.dc.b	$C7,$1C,$3C,$CD,$C3,$EA,$D7,$F2,$2A,$E3,$AF,$3C,$7F,$6E,$EE,$7E
	.dc.b	$CE,$3E,$0A,$FD,$D9,$D8,$B5,$E8,$7F,$7F,$1B,$9F,$E5,$ED,$7F,$59
	.dc.b	$3B,$7C,$4D,$7E,$3E,$3E,$8A,$E6,$D8,$D9,$F1,$CF,$CF,$57,$53,$B4
	.dc.b	$F9,$63,$67,$C9,$9E,$5C,$3F,$DF,$BC,$3F,$DE,$F0,$7F,$43,$7F,$C0
	.dc.b	$35,$FE,$EA,$6B,$FE,$7D,$5F,$07,$C2,$3F,$98,$46,$9F,$E2,$8D,$7A
	.dc.b	$3F,$E0,$9A,$F0,$47,$F6,$7C,$5D,$F9,$6C,$66,$30,$47,$F9,$A7,$83
	.dc.b	$E1,$A7,$3A,$31,$7C,$AB,$FD,$29,$8F,$F8,$BD,$FD,$C6,$5F,$D8,$63
	.dc.b	$11,$AF,$23,$A4,$7F,$B9,$23,$67,$F4,$D7,$6F,$F3,$16,$BC,$E3,$DD
	.dc.b	$18,$8E,$11,$E7,$FA,$97,$CC,$53,$BC,$7A,$FE,$A5,$F5,$97,$C2,$98
	.dc.b	$78,$B6,$56,$A3,$0F,$F8,$0D,$FF,$09,$35,$FF,$14,$B5,$FE,$90,$D7
	.dc.b	$FC,$05,$F5,$4F,$8D,$F9,$8D,$8F,$03,$5F,$F5,$F4,$F5,$E7,$6D,$FE
	.dc.b	$0B,$FA,$AF,$F5,$55,$79,$D7,$AD,$F5,$DB,$19,$F7,$3E,$D7,$8D,$FE
	.dc.b	$80,$8F,$F1,$53,$5F,$F2,$0B,$5F,$EC,$0D,$7F,$C3,$5F,$54,$FB,$4F
	.dc.b	$C2,$36,$3E,$43,$5F,$F6,$B2,$FB,$4F,$C5,$7D,$63,$E8,$6F,$2C,$BF
	.dc.b	$A1,$39,$7E,$97,$E3,$1F,$8F,$7F,$C8,$0D,$7F,$CA,$ED,$7F,$BB,$35
	.dc.b	$FF,$18,$7D,$67,$37,$FD,$D0,$DE,$0F,$97,$1F,$E1,$B8,$C7,$07,$0C
	.dc.b	$3F,$41,$F5,$8F,$9B,$BF,$33,$30,$FC,$34,$E5,$FC,$47,$B0,$F8,$FF
	.dc.b	$D9,$36,$3F,$CB,$0D,$7F,$CD,$ED,$7F,$C1,$DF,$59,$FC,$BB,$4F,$2F
	.dc.b	$52,$31,$A7,$F9,$2D,$B5,$FE,$33,$5F,$39,$FB,$7F,$AC,$7C,$CF,$F0
	.dc.b	$36,$D7,$F9,$A5,$AF,$FA,$01,$AF,$F8,$7F,$DF,$FC,$93,$7B,$BF,$86
	.dc.b	$46,$23,$FC,$B8,$D7,$FC,$89,$F5,$8F,$90,$FE,$24,$D8,$FF,$3C,$35
	.dc.b	$FF,$47,$B5,$FF,$17,$7D,$65,$7F,$CD,$0D,$7F,$CA,$0D,$D4,$DF,$87
	.dc.b	$97,$FC,$85,$B5,$FE,$87,$F0,$1F,$39,$FC,$95,$B1,$98,$C1,$FD,$EF
	.dc.b	$CB,$7B,$1F,$CC,$67,$F9,$0B,$63,$31,$FC,$63,$8D,$F8,$8F,$DD,$7F
	.dc.b	$1B,$1C,$0F,$F5,$E4,$F0,$7F,$DA,$2D,$7F,$BD,$35,$FF,$2D,$BB,$8F
	.dc.b	$CB,$9D,$BD,$2F,$E8,$8D,$9F,$E5,$27,$70,$FE,$2B,$E3,$5F,$75,$FC
	.dc.b	$ED,$F3,$73,$FF,$A5,$5B,$5F,$E7,$D7,$6F,$F2,$36,$BD,$23,$87,$4F
	.dc.b	$33,$A5,$75,$CE,$57,$17,$C1,$E5,$E2,$39,$1D,$E3,$18,$F9,$E7,$FB
	.dc.b	$07,$C8,$FF,$99,$6E,$FA,$BE,$2A,$FD,$CC,$E7,$0F,$E9,$2F,$F9,$AF
	.dc.b	$0D,$DF,$FB,$AE,$35,$3F,$E2,$76,$BF,$E8,$86,$BF,$EB,$87,$75,$DA
	.dc.b	$FE,$2C,$7F,$9D,$F1,$EB,$5F,$17,$CD,$F9,$9F,$65,$23,$04,$3D,$FC
	.dc.b	$CD,$93,$B9,$CC,$7F,$3F,$DB,$EA,$7D,$CF,$D7,$24,$FF,$55,$FD,$63
	.dc.b	$F1,$BA,$FD,$D9,$CC,$F2,$7F,$BC,$4E,$AB,$FD,$4B,$F5,$9F,$E0,$CA
	.dc.b	$FC,$3B,$5F,$A6,$1C,$5F,$E6,$BE,$BD,$F7,$3F,$A3,$9E,$15,$FE,$51
	.dc.b	$6B,$DB,$D6,$FF,$32,$46,$23,$FD,$C6,$F4,$C3,$FD,$6F,$7F,$CC,$27
	.dc.b	$FD,$CE,$D7,$C5,$5E,$AF,$B4,$91,$F0,$5F,$6F,$AD,$88,$F9,$B9,$A7
	.dc.b	$19,$FF,$2A,$8F,$F0,$8B,$F6,$F8,$FF,$17,$F0,$D2,$FF,$13,$1F,$EA
	.dc.b	$7C,$FF,$04,$D8,$72,$7F,$AD,$EB,$84,$7E,$6A,$3F,$D7,$53,$F9,$8F
	.dc.b	$C1,$7F,$BD,$61,$F9,$66,$E1,$F8,$D6,$DF,$4A,$7C,$EF,$1F,$13,$F3
	.dc.b	$15,$FC,$B7,$13,$49,$1F,$A4,$B8,$7E,$8A,$3A,$F2,$B4,$FF,$17,$4F
	.dc.b	$C6,$F8,$EF,$C2,$9F,$86,$F8,$D9,$1F,$8D,$8C,$BF,$8F,$8F,$C5,$FC
	.dc.b	$07,$E3,$38,$A4,$FE,$4E,$32,$FE,$56,$32,$FD,$48,$FF,$25,$78,$6F
	.dc.b	$92,$FE,$6C,$49,$FF,$4D,$0C,$93,$B8,$CC,$62,$7F,$3B,$F4,$B2,$7F
	.dc.b	$C7,$15,$B7,$F0,$7F,$05,$FA,$22,$3F,$A9,$F9,$BA,$FF,$AA,$70,$F8
	.dc.b	$D3,$FE,$69,$D7,$FC,$B9,$3F,$F0,$05,$A8,$9F,$89,$3D,$3F,$F6,$13
	.dc.b	$63,$D6,$1C,$3E,$E6,$BF,$EB,$97,$3C,$DF,$E1,$4F,$89,$7D,$F7,$AE
	.dc.b	$30,$DE,$8F,$15,$F5,$C3,$F7,$FF,$31,$E9,$24,$7C,$8C,$91,$88,$FD
	.dc.b	$3F,$D3,$7F,$06,$38,$BF,$C3,$7E,$03,$F8,$49,$1F,$E2,$11,$5F,$D7
	.dc.b	$C5,$CB,$F5,$D3,$F7,$BD,$9F,$2E,$5E,$6D,$FE,$DC,$63,$0F,$95,$18
	.dc.b	$7E,$64,$EF,$87,$FC,$59,$1F,$DA,$A7,$4B,$E3,$5F,$41,$CB,$5F,$F2
	.dc.b	$74,$FD,$59,$CB,$F8,$B8,$CB,$C6,$7C,$37,$DA,$B7,$2F,$AB,$75,$7F
	.dc.b	$94,$67,$31,$8F,$8B,$ED,$87,$D2,$8C,$5F,$53,$F0,$CB,$E6,$7E,$BD
	.dc.b	$F2,$C3,$A5,$F1,$6D,$F0,$6B,$E3,$AE,$33,$BC,$3C,$33,$B6,$BE,$9A
	.dc.b	$E3,$87,$86,$32,$F9,$B0,$F1,$56,$2B,$78,$76,$BE,$87,$AC,$E9,$3B
	.dc.b	$39,$C6,$E3,$26,$EF,$B6,$1E,$78,$C4,$EE,$73,$1D,$B0,$F1,$BA,$3B
	.dc.b	$1C,$F0,$F5,$9E,$67,$31,$C0,$97,$F4,$EF,$B5,$70,$C3,$73,$9B,$A7
	.dc.b	$85,$F3,$5F,$DE,$BE,$35,$AA,$F6,$CE,$5F,$29,$FB,$23,$E0,$97,$0F
	.dc.b	$C1,$B7,$0F,$F4,$A3,$F8,$DE,$CF,$E7,$EF,$F2,$09,$46,$23,$93,$5F
	.dc.b	$C7,$35,$FC,$6B,$CA,$BF,$2A,$DC,$7F,$82,$FB,$7F,$CA,$B6,$FC,$0C
	.dc.b	$E6,$F8,$F8,$38,$3F,$13,$5B,$8F,$B5,$3B,$7D,$13,$B7,$F0,$93,$BC
	.dc.b	$3C,$33,$96,$CE,$B5,$FF,$57,$C6,$09,$8F,$B2,$BC,$33,$A8,$C5,$78
	.dc.b	$A7,$A4,$7B,$27,$53,$EA,$34,$4A,$FD,$DD,$F9,$6B,$9C,$78,$27,$83
	.dc.b	$AA,$E7,$1E,$E3,$94,$F0,$3C,$18,$F8,$A3,$8C,$65,$FE,$89,$7D,$DC
	.dc.b	$90,$E9,$7D,$EB,$94,$62,$B9,$3F,$EB,$AB,$F1,$E5,$F8,$4C,$7F,$7F
	.dc.b	$BF,$F6,$4E,$5F,$82,$FF,$7D,$F2,$2D,$FF,$53,$CB,$7C,$3F,$E3,$87
	.dc.b	$8C,$67,$2F,$E5,$8D,$8F,$E6,$18,$7F,$B9,$1A,$7F,$34,$9F,$E5,$DF
	.dc.b	$6F,$FE,$92,$8A,$7F,$94,$FE,$13,$13,$FC,$BE,$6E,$5F,$E1,$16,$E2
	.dc.b	$6D,$61,$CF,$C0,$E3,$0F,$F2,$F3,$70,$FF,$1E,$BF,$EA,$C7,$FA,$03
	.dc.b	$5F,$E7,$CD,$C3,$FC,$40,$E6,$1F,$56,$FB,$2F,$3F,$6B,$EF,$9E,$27
	.dc.b	$AA,$F7,$5A,$9C,$E5,$FE,$AE,$C3,$93,$FD,$F9,$B5,$FE,$1A,$6B,$FC
	.dc.b	$15,$AF,$F5,$7E,$9F,$36,$DF,$D1,$6B,$FB,$B4,$60,$8F,$F9,$FD,$F9
	.dc.b	$3F,$41,$F2,$9F,$9B,$E8,$BD,$8F,$9B,$F4,$B8,$6D,$FF,$AE,$5F,$89
	.dc.b	$3B,$8F,$F1,$5B,$71,$FF,$11,$35,$FE,$F1,$FA,$CB,$BB,$ED,$AF,$95
	.dc.b	$B1,$86,$DF,$F8,$97,$C2,$3F,$DF,$AB,$83,$93,$15,$C5,$C5,$FF,$90
	.dc.b	$65,$F0,$1E,$0E,$DF,$0B,$6B,$7C,$CA,$DC,$61,$FF,$02,$B6,$78,$5F
	.dc.b	$F9,$DF,$E5,$3F,$CA,$53,$FE,$C8,$5F,$61,$F9,$AF,$45,$1A,$AE,$13
	.dc.b	$98,$C4,$7F,$B2,$5A,$FF,$B2,$5A,$FF,$A2,$1B,$3D,$3C,$EE,$47,$3F
	.dc.b	$B3,$B8,$78,$FE,$7B,$B7,$CC,$77,$6C,$64,$F4,$46,$5A,$7C,$1F,$94
	.dc.b	$FA,$E7,$D8,$F8,$6B,$19,$26,$7F,$ED,$4D,$93,$A8,$FF,$61,$FD,$77
	.dc.b	$FA,$AE,$BE,$AF,$1B,$EA,$DF,$4F,$FA,$7A,$D5,$FF,$BF,$2B,$53,$9F
	.dc.b	$73,$F6,$55,$AA,$FE,$D0,$D7,$FC,$B6,$D7,$FC,$DC,$F4,$AF,$C5,$C6
	.dc.b	$27,$F8,$7C,$78,$FE,$AB,$B6,$3F,$81,$F9,$FF,$CB,$9F,$98,$DE,$C6
	.dc.b	$49,$87,$C1,$8C,$BF,$0F,$7F,$CF,$FC,$FF,$E8,$1B,$5F,$8D,$6C,$EE
	.dc.b	$3F,$21,$3F,$89,$FC,$B3,$F7,$93,$FE,$C5,$5D,$D9,$EF,$D3,$6F,$FD
	.dc.b	$43,$CC,$AE,$37,$FE,$7E,$8C,$5F,$F1,$F6,$D7,$0A,$EF,$AF,$C6,$9C
	.dc.b	$DF,$87,$C0,$EC,$FD,$FF,$13,$F7,$67,$EF,$3E,$A2,$E3,$FA,$98,$B3
	.dc.b	$F8,$59,$B8,$7D,$2E,$DF,$D8,$7D,$AF,$F0,$4E,$BF,$E0,$AD,$C7,$F6
	.dc.b	$4D,$C3,$F1,$CD,$BF,$B7,$0F,$CF,$BF,$87,$57,$E3,$BF,$05,$FC,$6B
	.dc.b	$FC,$12,$75,$E1,$7E,$02,$73,$0B,$E7,$7C,$A3,$FD,$96,$D7,$FC,$6C
	.dc.b	$D7,$FD,$52,$D7,$FD,$78,$F1,$8E,$75,$DF,$9B,$FA,$24,$E7,$0F,$E3
	.dc.b	$9D,$DF,$D3,$FB,$4D,$24,$7F,$2E,$EB,$FE,$62,$47,$07,$FB,$73,$EF
	.dc.b	$FF,$0A,$DE,$6F,$EB,$51,$8D,$3F,$DA,$CD,$7F,$D5,$9E,$DF,$E1,$9F
	.dc.b	$F5,$5B,$5F,$F7,$53,$78,$7F,$D8,$3C,$FF,$D9,$7E,$D7,$04,$70,$7F
	.dc.b	$69,$F9,$67,$1D,$26,$7F,$EF,$28,$CB,$FE,$F4,$FA,$87,$5F,$F8,$01
	.dc.b	$B1,$FC,$D9,$FE,$CB,$C9,$F8,$97,$3A,$7D,$99,$FF,$61,$FD,$FF,$F6
	.dc.b	$68,$68,$8C,$37,$A7,$FB,$F4,$63,$F2,$27,$C4,$7F,$84,$3E,$77,$FC
	.dc.b	$31,$9F,$E5,$E7,$39,$7F,$48,$9C,$BF,$ED,$2F,$AF,$7A,$FF,$E1,$86
	.dc.b	$C7,$F4,$E5,$EB,$66,$7F,$C3,$BF,$40,$F8,$EF,$F0,$83,$6F,$FD,$09
	.dc.b	$BF,$AA,$30,$FF,$A1,$F7,$46,$23,$FD,$CD,$F1,$1F,$ED,$0A,$FE,$BA
	.dc.b	$FF,$84,$23,$E3,$1F,$11,$D3,$8A,$46,$F5,$FD,$B7,$8C,$FA,$5F,$F1
	.dc.b	$9B,$70,$FE,$D2,$BE,$D3,$33,$FE,$38,$FB,$0D,$65,$E7,$6B,$FE,$FB
	.dc.b	$6B,$FE,$FA,$6B,$F6,$3E,$C3,$F9,$C6,$1F,$EF,$E7,$E0,$E3,$FC,$25
	.dc.b	$CC,$EF,$5F,$F2,$3F,$D8,$3E,$72,$9F,$E4,$AF,$35,$F2,$98,$CB,$FC
	.dc.b	$03,$18,$7E,$7F,$02,$31,$8F,$CF,$6A,$24,$6E,$BF,$C9,$12,$F9,$4F
	.dc.b	$04,$78,$5B,$7F,$E1,$0F,$D0,$7F,$CD,$5A,$FF,$23,$AF,$F0,$EE,$3F
	.dc.b	$4E,$73,$5D,$A7,$35,$FC,$A6,$7F,$CE,$55,$E3,$3F,$3E,$E1,$F2,$B5
	.dc.b	$CF,$C2,$7F,$97,$63,$AC,$62,$FC,$9D,$5F,$E7,$09,$CE,$6F,$4B,$EF
	.dc.b	$57,$F8,$A3,$E7,$BF,$BF,$BF,$E2,$A8,$CE,$1F,$8B,$9C,$BF,$8D,$8C
	.dc.b	$BE,$7A,$E9,$C9,$FE,$6F,$9C,$CF,$CC,$6B,$FB,$AF,$B0,$FE,$6E,$F1
	.dc.b	$6B,$F4,$9A,$EF,$C9,$FF,$3D,$47,$F3,$ED,$7F,$A8,$37,$2E,$1F,$3D
	.dc.b	$FE,$8E,$8C,$5F,$FB,$CF,$E3,$3E,$F7,$D8,$7D,$B8,$FF,$2A,$65,$FE
	.dc.b	$0A,$8C,$47,$D1,$9D,$F1,$7D,$3C,$BE,$46,$58,$AE,$82,$F8,$CE,$A3
	.dc.b	$15,$FC,$02,$F8,$47,$F5,$5C,$3C,$13,$98,$D9,$E4,$C3,$6F,$E5,$6B
	.dc.b	$8D,$F5,$C3,$CF,$39,$38,$9D,$6B,$8A,$AE,$38,$78,$23,$D3,$F0,$3E
	.dc.b	$88,$FF,$15,$BD,$23,$81,$FE,$7D,$C7,$E1,$4E,$9F,$46,$1C,$70,$EF
	.dc.b	$87,$66,$DF,$93,$1F,$45,$F0,$39,$CE,$A7,$F9,$29,$E0,$2D,$61,$F7
	.dc.b	$1F,$A5,$3F,$CE,$7E,$33,$E2,$38,$FB,$FF,$D4,$1B,$19,$9F,$E6,$33
	.dc.b	$96,$F7,$7F,$81,$DB,$97,$28,$C4,$77,$94,$62,$3F,$6D,$18,$9F,$BB
	.dc.b	$97,$FA,$C6,$6F,$C2,$7E,$EF,$4F,$C6,$DE,$E7,$EB,$B6,$3F,$23,$5C
	.dc.b	$67,$ED,$B6,$38,$E1,$E7,$6E,$1D,$AA,$1D,$27,$55,$E6,$6C,$7F,$1F
	.dc.b	$8E,$87,$F0,$F2,$BF,$C8,$EF,$C9,$1C,$6B,$AD,$F9,$B0,$F6,$4E,$6F
	.dc.b	$C1,$5A,$8E,$93,$A9,$E9,$87,$89,$FF,$55,$B6,$7C,$31,$FC,$BE,$78
	.dc.b	$3A,$AF,$D4,$FE,$12,$F9,$EF,$DA,$6D,$FC,$9E,$31,$5F,$E4,$56,$BF
	.dc.b	$E9,$1F,$A0,$FF,$4D,$B9,$AE,$2F,$F4,$19,$DD,$FF,$94,$9B,$7F,$ED
	.dc.b	$9B,$E3,$5F,$DC,$04,$BE,$83,$9C,$61,$B1,$98,$FF,$53,$FE,$86,$F3
	.dc.b	$B3,$52,$7D,$0F,$93,$15,$6E,$7F,$CC,$1C,$AF,$7B,$BB,$63,$31,$E2
	.dc.b	$FC,$BD,$F5,$3A,$89,$3A,$3F,$CD,$53,$A7,$07,$B5,$FE,$37,$AF,$F5
	.dc.b	$36,$C7,$F6,$7B,$F0,$3F,$37,$E2,$BC,$7E,$34,$8F,$F6,$2E,$E9,$C4
	.dc.b	$7C,$9A,$FF,$13,$3E,$CA,$F8,$15,$B8,$FB,$4F,$CA,$23,$B3,$D3,$F6
	.dc.b	$FE,$F2,$3F,$ED,$4F,$99,$7D,$27,$D4,$FB,$9F,$95,$3F,$65,$BC,$1F
	.dc.b	$E9,$06,$EB,$EF,$C3,$E5,$78,$12,$FE,$3B,$F7,$BE,$4D,$E3,$FD,$EF
	.dc.b	$0A,$57,$E4,$5F,$C6,$FC,$3B,$E5,$FF,$DC,$B3,$FE,$99,$FC,$D5,$F0
	.dc.b	$BF,$88,$6C,$66,$38,$F8,$0F,$A9,$FC,$5E,$92,$BF,$3C,$F6,$FC,$BF
	.dc.b	$FD,$E5,$1F,$A1,$BF,$C8,$91,$F8,$9D,$FF,$96,$FC,$97,$FC,$CB,$1F
	.dc.b	$EA,$B6,$BF,$9D,$6F,$1F,$FA,$D3,$C8,$7C,$1F,$E7,$38,$E5,$FE,$BD
	.dc.b	$FC,$C7,$D5,$FE,$A2,$3F,$59,$F3,$7F,$A2,$23,$FE,$2F,$C3,$FD,$A9
	.dc.b	$F8,$B3,$93,$F5,$3F,$83,$BE,$B6,$92,$F8,$74,$FF,$8D,$F9,$DF,$8A
	.dc.b	$7F,$90,$A7,$F8,$07,$8F,$FB,$02,$3C,$04,$BE,$93,$E3,$C7,$EE,$FF
	.dc.b	$09,$F7,$E1,$FE,$31,$6C,$7F,$91,$5B,$AB,$C0,$FF,$2A,$4F,$FB,$4F
	.dc.b	$47,$D2,$7A,$89,$1F,$ED,$7F,$5B,$FC,$E0,$DD,$BF,$83,$C7,$BE,$30
	.dc.b	$79,$F0,$71,$3C,$3F,$E1,$BF,$8B,$7F,$AB,$2B,$FD,$C4,$D7,$F8,$A3
	.dc.b	$6B,$FD,$D7,$E8,$3E,$B7,$8F,$A7,$2F,$8B,$FC,$62,$75,$18,$21,$D2
	.dc.b	$35,$3F,$C7,$BE,$DD,$F4,$14,$95,$FE,$F5,$8F,$DD,$3F,$C9,$DD,$B9
	.dc.b	$8D,$B9,$F3,$3A,$70,$F5,$F1,$E4,$F3,$8C,$BD,$63,$2F,$58,$FE,$63
	.dc.b	$F8,$B7,$F0,$C8,$FE,$6C,$DD,$1C,$BF,$E6,$D8,$C3,$FE,$6F,$FA,$57
	.dc.b	$77,$2F,$36,$61,$F8,$91,$7C,$1F,$F4,$45,$70,$75,$19,$77,$3C,$9F
	.dc.b	$E9,$13,$B7,$FA,$57,$E6,$8F,$B0,$FE,$92,$D8,$F0,$35,$DF,$C4,$6A
	.dc.b	$64,$69,$E3,$1F,$D4,$8F,$9F,$1F,$D5,$DF,$A7,$9F,$C2,$8F,$F5,$04
	.dc.b	$63,$3F,$F4,$83,$C2,$75,$3F,$EC,$AD,$FA,$6F,$23,$E2,$56,$EF,$FD
	.dc.b	$31,$F5,$7F,$AC,$C3,$8B,$63,$3F,$45,$E5,$3F,$B3,$C6,$6F,$FD,$64
	.dc.b	$D4,$DB,$C3,$27,$89,$CE,$7E,$B7,$DC,$FF,$2A,$7F,$BB,$3F,$CC,$1B
	.dc.b	$E6,$7F,$76,$8D,$63,$EE,$6E,$3F,$EB,$17,$84,$66,$78,$79,$7F,$DF
	.dc.b	$BE,$BF,$EA,$4F,$C6,$6D,$FF,$AE,$5B,$87,$FB,$35,$FF,$79,$BF,$8B
	.dc.b	$72,$9A,$4C,$FF,$C1,$7F,$A6,$FF,$30,$9F,$C9,$35,$FF,$09,$36,$FE
	.dc.b	$BF,$C7,$3E,$0F,$8F,$99,$2F,$8B,$FE,$16,$9D,$46,$08,$74,$8D,$4E
	.dc.b	$FA,$24,$FE,$8E,$BF,$91,$37,$A7,$2F,$F8,$96,$32,$D9,$F0,$4F,$F8
	.dc.b	$9B,$E1,$3E,$7C,$FF,$BC,$5A,$9A,$7D,$2D,$7F,$DC,$8E,$36,$E4,$73
	.dc.b	$DF,$FC,$63,$F8,$4F,$F8,$5A,$CD,$3F,$C3,$3F,$09,$FE,$A3,$FA,$0D
	.dc.b	$FC,$24,$AF,$DF,$56,$BE,$19,$5F,$E3,$AF,$8A,$95,$FE,$4D,$7F,$C2
	.dc.b	$FD,$58,$8F,$F2,$27,$C2,$FF,$53,$6C,$E9,$FE,$15,$F6,$4F,$9F,$FF
	.dc.b	$23,$B6,$33,$1F,$E5,$4D,$1F,$92,$7E,$8A,$72,$43,$FC,$A5,$5C,$57
	.dc.b	$3F,$EF,$6B,$8F,$6F,$6F,$1E,$57,$89,$F0,$FC,$D6,$23,$9C,$5B,$FE
	.dc.b	$03,$82,$BF,$CB,$7E,$7C,$BF,$F3,$73,$EA,$F2,$B1,$1F,$0A,$2D,$F2
	.dc.b	$9F,$F3,$2F,$D8,$FF,$92,$38,$BF,$A2,$B7,$C3,$61,$23,$0F,$FA,$F7
	.dc.b	$F0,$4F,$75,$70,$F8,$8F,$F7,$B5,$F0,$A5,$F0,$6C,$E9,$E5,$C3,$5D
	.dc.b	$FA,$BF,$79,$FF,$73,$5F,$3B,$FF,$03,$73,$7F,$BC,$63,$15,$BC,$7D
	.dc.b	$3F,$3F,$CE,$F7,$87,$23,$D5,$5B,$BE,$38,$73,$9F,$75,$F1,$AD,$47
	.dc.b	$09,$D3,$E6,$6D,$AB,$84,$62,$B8,$D6,$CF,$11,$F1,$27,$6E,$2B,$CF
	.dc.b	$87,$D2,$AE,$31,$D6,$7A,$5C,$78,$4E,$AF,$9D,$F1,$AD,$3F,$12,$AE
	.dc.b	$1F,$01,$E3,$3E,$AA,$E3,$39,$AE,$55,$E5,$9C,$BF,$0C,$E2,$64,$EA
	.dc.b	$66,$38,$9E,$17,$6E,$27,$B1,$F7,$67,$E1,$CF,$E2,$4F,$90,$5C,$3E
	.dc.b	$DB,$70,$FC,$33,$9E,$2F,$F3,$E3,$C4,$FC,$8B,$87,$5F,$09,$97,$2D
	.dc.b	$AF,$EB,$AF,$DC,$9E,$78,$FF,$8A,$38,$1E,$EF,$F7,$78,$FD,$3C,$EB
	.dc.b	$3F,$CB,$37,$1F,$D4,$35,$FF,$29,$D7,$19,$DE,$1D,$C9,$3F,$86,$C3
	.dc.b	$BF,$7E,$31,$EF,$56,$FF,$5D,$5A,$BF,$C5,$63,$DF,$2F,$5C,$62,$FF
	.dc.b	$33,$5B,$C3,$9C,$E4,$8F,$BF,$2F,$34,$7A,$AB,$C0,$D3,$CD,$1A,$7E
	.dc.b	$59,$CB,$1F,$6B,$AC,$B8,$BA,$BF,$09,$DE,$33,$8F,$A2,$33,$5E,$EB
	.dc.b	$ED,$39,$8E,$B3,$AB,$FE,$2F,$8F,$B6,$72,$48,$F4,$4F,$77,$CC,$70
	.dc.b	$F6,$BD,$54,$7E,$0F,$3B,$E3,$FF,$D4,$99,$FC,$A9,$D9,$1F,$E1,$53
	.dc.b	$BF,$77,$18,$7F,$00,$FA,$47,$C9,$FF,$8F,$FA,$3F,$C7,$71,$8C,$3F
	.dc.b	$CB,$99,$FF,$30,$7F,$CB,$D1,$9B,$FE,$97,$1F,$CB,$63,$15,$CA,$FA
	.dc.b	$4F,$E1,$2A,$9F,$CE,$AB,$71,$FC,$D7,$EC,$1F,$29,$AA,$46,$27,$FC
	.dc.b	$DC,$FF,$BB,$1B,$3B,$8D,$4F,$90,$FE,$8E,$E3,$0F,$F2,$BB,$C6,$34
	.dc.b	$76,$9E,$5F,$5D,$2F,$FD,$89,$A7,$F7,$EF,$AD,$C5,$F0,$C1,$78,$C7
	.dc.b	$FB,$67,$42,$F1,$8F,$C5,$E2,$F6,$68,$7E,$1D,$FF,$AE,$7C,$CC,$C4
	.dc.b	$AF,$9E,$EB,$DC,$FF,$1C,$46,$0F,$87,$E2,$FC,$13,$E6,$ED,$FE,$DA
	.dc.b	$F5,$31,$3A,$26,$FF,$16,$75,$18,$CB,$F0,$5F,$A4,$3E,$9F,$CA,$D9
	.dc.b	$FF,$68,$65,$FE,$20,$7F,$DA,$51,$9C,$3E,$85,$A3,$0F,$4B,$F5,$6F
	.dc.b	$D1,$EB,$DD,$30,$E7,$E0,$5C,$79,$E3,$FD,$6E,$7B,$D7,$92,$B5,$8F
	.dc.b	$FA,$4B,$F4,$B7,$CB,$7E,$25,$B7,$E6,$FA,$0C,$4C,$8F,$F5,$C4,$EB
	.dc.b	$2E,$F5,$97,$F0,$B1,$9B,$ED,$7F,$12,$B5,$1F,$23,$EC,$FE,$39,$F2
	.dc.b	$3E,$86,$E3,$F1,$A2,$E3,$EC,$9B,$5F,$43,$F0,$0F,$9F,$F6,$BF,$41
	.dc.b	$B8,$7C,$E6,$DF,$C4,$6D,$7F,$AC,$BF,$2E,$7D,$57,$EA,$5B,$7F,$5B
	.dc.b	$ED,$F1,$32,$3F,$DB,$13,$AC,$3E,$7D,$65,$FD,$2C,$66,$FB,$5F,$E1
	.dc.b	$2B,$51,$F8,$8F,$B9,$FC,$3F,$B9,$F8,$86,$E3,$F8,$66,$E3,$F7,$1B
	.dc.b	$5F,$90,$E8,$FC,$79,$F9,$0F,$7A,$DF,$E3,$A2,$DF,$E1,$26,$D7,$FB
	.dc.b	$4B,$A9,$F3,$3F,$C2,$5B,$87,$E5,$A3,$04,$7F,$DF,$11,$AB,$FC,$7C
	.dc.b	$E5,$FE,$0B,$19,$AE,$D3,$FA,$43,$4F,$EA,$3C,$CF,$D3,$F9,$BF,$A8
	.dc.b	$6E,$1F,$A6,$6E,$1F,$9C,$6D,$7E,$C3,$ED,$3F,$5E,$7E,$C3,$E1,$AD
	.dc.b	$FE,$BA,$2D,$FE,$92,$6D,$7F,$BC,$BF,$4B,$FE,$3D,$7F,$E5,$F6,$D7
	.dc.b	$FA,$03,$F3,$17,$CF,$7F,$2D,$9C,$C9,$A7,$0B,$57,$0B,$FF,$38,$C7
	.dc.b	$7F,$27,$91,$23,$26,$A7,$85,$FF,$2E,$BE,$FE,$37,$F0,$98,$FE,$62
	.dc.b	$DC,$3F,$A9,$7D,$9B,$E0,$3F,$93,$F8,$9F,$D0,$FF,$0A,$FE,$59,$9F
	.dc.b	$6C,$7F,$D0,$71,$87,$F8,$8F,$C9,$BD,$9E,$1E,$67,$F4,$BF,$C5,$7F
	.dc.b	$89,$E1,$FE,$83,$6D,$FF,$A9,$3F,$15,$FB,$D3,$9F,$2B,$87,$91,$23
	.dc.b	$C1,$F3,$0F,$C6,$5D,$19,$C7,$FA,$5F,$B0,$FB,$FF,$EB,$4D,$C3,$8F
	.dc.b	$DE,$DA,$DD,$79,$7C,$E9,$5D,$BE,$4F,$CB,$4D,$FF,$91,$3F,$D6,$17
	.dc.b	$9B,$ED,$3F,$A5,$FE,$42,$F9,$FC,$61,$F1,$EC,$BE,$17,$FB,$73,$4F
	.dc.b	$99,$D6,$FC,$A3,$DB,$1C,$4D,$57,$87,$D0,$F9,$87,$F5,$AF,$CB,$FF
	.dc.b	$17,$A2,$3E,$87,$A5,$2F,$E4,$FD,$6F,$E1,$32,$F1,$E3,$FE,$BC,$C1
	.dc.b	$7F,$E0,$1F,$86,$E5,$87,$A5,$B5,$FD,$F7,$9D,$F9,$53,$D0,$E1,$A7
	.dc.b	$DF,$F5,$66,$3F,$63,$F0,$1F,$8E,$FC,$3D,$13,$F8,$5F,$5A,$4F,$C7
	.dc.b	$F8,$4F,$F6,$E3,$B8,$FF,$0A,$78,$AE,$FA,$7E,$37,$AE,$D7,$57,$0F
	.dc.b	$CC,$FC,$8F,$C3,$C1,$F6,$67,$F2,$9E,$B4,$AF,$AD,$87,$E4,$AB,$E2
	.dc.b	$4F,$83,$D2,$95,$F5,$BC,$D7,$D3,$E2,$FF,$B8,$C6,$27,$9D,$7F,$83
	.dc.b	$A3,$13,$FE,$20,$C3,$B2,$E4,$FF,$27,$5F,$5F,$6B,$FC,$65,$18,$9E
	.dc.b	$B1,$8F,$86,$95,$E3,$FB,$3B,$5F,$E2,$CB,$E1,$D5,$F7,$A3,$B5,$F8
	.dc.b	$EB,$73,$9E,$47,$B1,$FB,$73,$FC,$4B,$FA,$2F,$E1,$32,$FB,$11,$96
	.dc.b	$9F,$98,$8D,$63,$FA,$59,$FE,$01,$ED,$4A,$F9,$1F,$01,$6B,$F5,$53
	.dc.b	$71,$FF,$31,$6E,$AC,$5F,$FA,$E7,$C9,$60,$FA,$C7,$CA,$62,$18,$FE
	.dc.b	$C6,$F7,$5A,$9F,$F6,$55,$E2,$F8,$5E,$E7,$A4,$77,$69,$FC,$C3,$C6
	.dc.b	$FD,$3D,$F3,$8F,$05,$F3,$75,$85,$72,$72,$3C,$B5,$CA,$FB,$5F,$3C
	.dc.b	$3C,$4F,$5A,$F7,$E1,$CE,$73,$5C,$0F,$4D,$F3,$AF,$69,$DE,$76,$E2
	.dc.b	$FB,$4E,$E3,$CB,$18,$C3,$D1,$1E,$03,$94,$ED,$C4,$ED,$F6,$18,$BF
	.dc.b	$1D,$72,$86,$1F,$EE,$D7,$A1,$FE,$EF,$70,$E8,$FF,$7B,$7D,$9F,$F2
	.dc.b	$5F,$84,$FF,$7A,$36,$33,$E4,$27,$2D,$8E,$3E,$9F,$E0,$9B,$A7,$28
	.dc.b	$C4,$7E,$1C,$5F,$F5,$79,$D1,$F6,$72,$78,$FF,$8D,$D3,$F0,$11,$8B
	.dc.b	$FB,$D7,$C5,$FB,$CD,$9D,$D7,$1A,$FC,$B1,$C6,$B7,$87,$9E,$BF,$19
	.dc.b	$7C,$FE,$1B,$8D,$7E,$81,$A9,$DD,$7B,$DB,$8F,$A7,$1F,$05,$7E,$A3
	.dc.b	$1F,$06,$3A,$8F,$0D,$6D,$86,$09,$1E,$02,$4F,$EB,$F1,$F7,$37,$1F
	.dc.b	$13,$EE,$C3,$F6,$98,$FA,$0F,$01,$FA,$9F,$94,$FE,$53,$8F,$F6,$56
	.dc.b	$BF,$EC,$46,$E5,$FE,$E5,$E0,$FF,$5E,$4E,$FB,$3F,$19,$5C,$23,$EE
	.dc.b	$35,$FE,$08,$D7,$F8,$4B,$FE,$96,$7F,$04,$63,$AB,$99,$1C,$88,$E0
	.dc.b	$FF,$AB,$A9,$E8,$7F,$0A,$E2,$FF,$1A,$C6,$27,$D9,$79,$7F,$8C,$FD
	.dc.b	$87,$F5,$97,$FC,$08,$D7,$FD,$C6,$DC,$BF,$9C,$7D,$8B,$EC,$B9,$52
	.dc.b	$B8,$3F,$9B,$6B,$FC,$8D,$AF,$83,$C4,$FD,$94,$75,$CB,$F9,$45,$FF
	.dc.b	$11,$6F,$07,$69,$D5,$FA,$3E,$07,$04,$7F,$DA,$7F,$15,$5F,$97,$29
	.dc.b	$EE,$48,$F9,$7B,$FF,$3D,$6E,$DF,$68,$8F,$D0,$C9,$F4,$9F,$C2,$5B
	.dc.b	$19,$8C,$1F,$C3,$7E,$C9,$E1,$52,$4F,$DA,$8C,$BD,$7E,$6F,$A1,$1C
	.dc.b	$1E,$6F,$92,$FC,$69,$1C,$1F,$84,$FB,$2F,$F3,$33,$7E,$0B,$F8,$F4
	.dc.b	$62,$3F,$1A,$D7,$F2,$B5,$F9,$82,$3F,$11,$FC,$DC,$C8,$F8,$ED,$7F
	.dc.b	$2F,$F6,$0F,$87,$FE,$58,$D8,$FC,$BB,$5D,$C7,$E9,$B3,$91,$F4,$9A
	.dc.b	$FE,$8B,$67,$CB,$7E,$B0,$91,$F6,$1F,$E7,$5F,$5A,$F8,$5A,$49,$FD
	.dc.b	$3C,$65,$FB,$FB,$FE,$D8,$8E,$0F,$D6,$FD,$77,$9D,$FE,$68,$D7,$FB
	.dc.b	$E3,$5F,$F0,$0B,$54,$BE,$5F,$CA,$48,$F1,$BF,$D4,$7E,$BB,$C9,$1F
	.dc.b	$C0,$1A,$FF,$43,$6B,$FE,$0B,$6B,$FE,$10,$FA,$DF,$F3,$C3,$FE,$7B
	.dc.b	$6B,$FE,$10,$EA,$FF,$09,$46,$3B,$7F,$C6,$1E,$1B,$EB,$7F,$73,$3C
	.dc.b	$2F,$FA,$93,$5E,$3E,$3F,$F6,$78,$C4,$78,$63,$F8,$0D,$7F,$8D,$A3
	.dc.b	$FD,$18,$D7,$C4,$6C,$FF,$48,$7D,$6B,$E4,$FF,$84,$B6,$33,$18,$EF
	.dc.b	$CB,$23,$27,$28,$E3,$F4,$DC,$BE,$9A,$4F,$68,$CB,$E0,$8C,$BE,$41
	.dc.b	$69,$FE,$37,$F1,$1F,$C1,$E7,$FD,$4C,$D7,$83,$75,$FF,$2D,$7D,$6F
	.dc.b	$F0,$29,$CF,$85,$FE,$0D,$AE,$11,$FD,$F5,$AF,$F9,$15,$AF,$F9,$29
	.dc.b	$E9,$C1,$FE,$49,$A7,$6F,$F2,$E6,$D7,$92,$FF,$98,$FD,$6B,$D3,$A4
	.dc.b	$9F,$F5,$F4,$7D,$3C,$BF,$CE,$0E,$E3,$33,$B7,$E2,$46,$F1,$FF,$2D
	.dc.b	$7A,$9F,$56,$7E,$42,$DF,$FA,$0B,$63,$FD,$98,$EF,$0F,$F3,$F3,$6B
	.dc.b	$9D,$70,$38,$F0,$7F,$4A,$F9,$67,$D6,$D2,$63,$D1,$AF,$F6,$68,$DB
	.dc.b	$FD,$91,$AE,$63,$FC,$7D,$18,$C7,$C7,$C5,$E8,$39,$BF,$E9,$3F,$AD
	.dc.b	$FE,$4C,$DE,$0F,$EB,$B1,$F4,$30,$FC,$43,$AC,$FF,$CF,$CD,$8F,$F4
	.dc.b	$1B,$BA,$E5,$1F,$EE,$F8,$D2,$FA,$DF,$A4,$FE,$4D,$B3,$FE,$A5,$6D
	.dc.b	$7F,$A2,$DF,$A9,$3F,$E9,$07,$2D,$DF,$FB,$9C,$62,$3B,$6D,$FD,$DB
	.dc.b	$EB,$7F,$DC,$91,$FF,$00,$5A,$FE,$8D,$B8,$7F,$AD,$7E,$B7,$FC,$79
	.dc.b	$3B,$F8,$9F,$F3,$45,$70,$8F,$F2,$C3,$5F,$F5,$03,$5F,$F5,$2B,$D3
	.dc.b	$2F,$F5,$25,$7E,$1E,$7F,$64,$DB,$FF,$62,$56,$E7,$FD,$5B,$3F,$B5
	.dc.b	$7F,$D5,$F3,$E5,$E6,$FF,$08,$CE,$67,$9E,$D8,$99,$1C,$30,$F0,$C7
	.dc.b	$09,$FF,$0C,$FD,$6F,$D4,$27,$91,$8E,$0C,$B3,$0E,$47,$FB,$02,$78
	.dc.b	$4F,$F8,$8B,$EB,$7F,$D4,$D3,$FE,$C1,$6B,$FE,$C6,$6A,$69,$FE,$1F
	.dc.b	$7F,$D1,$5B,$B0,$FF,$1B,$7B,$5F,$87,$7B,$7A,$3F,$E3,$0F,$21,$F5
	.dc.b	$B4,$98,$79,$B8,$3F,$DB,$1D,$18,$8F,$ED,$93,$FE,$56,$23,$FE,$4C
	.dc.b	$FA,$C7,$D5,$FF,$A9,$5B,$19,$8C,$11,$F8,$79,$FE,$6E,$FC,$FA,$7E
	.dc.b	$FE,$7F,$DC,$B7,$C9,$FE,$60,$FF,$4A,$4D,$27,$85,$E1,$3F,$C9,$AB
	.dc.b	$C2,$A6,$23,$E1,$78,$D2,$BF,$CF,$0E,$9A,$7F,$8F,$9F,$F3,$A9,$1F
	.dc.b	$F3,$5F,$D5,$7F,$97,$F9,$BF,$D6,$4D,$8F,$F7,$D3,$5F,$F7,$FB,$53
	.dc.b	$FE,$67,$F7,$FF,$38,$BE,$5F,$FD,$E7,$E0,$F2,$6D,$72,$F6,$2B,$CB
	.dc.b	$E8,$9F,$11,$FE,$EE,$BF,$0C,$62,$7F,$DB,$31,$89,$D4,$79,$2F,$9E
	.dc.b	$0B,$E1,$8F,$91,$B8,$71,$BE,$67,$13,$99,$E3,$C2,$DF,$36,$D7,$29
	.dc.b	$E0,$72,$38,$16,$78,$36,$38,$3F,$23,$0F,$3C,$62,$77,$87,$9E,$FC
	.dc.b	$35,$A7,$C2,$4B,$ED,$19,$7B,$46,$67,$E9,$CE,$5E,$93,$A3,$89,$BC
	.dc.b	$3E,$24,$68,$E3,$19,$3F,$03,$87,$9C,$F4,$57,$5A,$E2,$49,$F8,$B5
	.dc.b	$CE,$31,$1F,$1A,$3A,$BE,$33,$F2,$35,$F7,$DB,$7F,$96,$8C,$47,$E5
	.dc.b	$3C,$37,$D5,$FE,$69,$B1,$FE,$CA,$C9,$18,$8F,$CE,$B5,$E5,$3F,$7A
	.dc.b	$38,$E1,$FC,$D7,$5F,$F3,$8E,$B9,$A9,$7F,$76,$F7,$9A,$25,$F0,$75
	.dc.b	$5C,$30,$FD,$2D,$F0,$9D,$FA,$4C,$75,$8F,$92,$78,$9F,$8D,$C7,$A6
	.dc.b	$1C,$B1,$DC,$FE,$D3,$1E,$8F,$F9,$C7,$1E,$8F,$2A,$E0,$DC,$3D,$0E
	.dc.b	$F1,$E6,$FE,$F2,$F9,$E5,$F2,$B1,$FD,$2E,$5E,$2C,$3C,$D8,$F4,$23
	.dc.b	$D2,$35,$87,$96,$BA,$DF,$38,$F6,$65,$E6,$C3,$8C,$6C,$F3,$1F,$1B
	.dc.b	$2E,$39,$79,$0F,$7E,$3D,$A7,$D1,$8E,$8F,$3E,$1D,$72,$F3,$9E,$5F
	.dc.b	$47,$15,$D4,$F9,$CE,$1F,$88,$63,$1F,$BC,$7C,$87,$FD,$3D,$1F,$3A
	.dc.b	$31,$87,$FB,$81,$DC,$F7,$6B,$F8,$7F,$A1,$7D,$5F,$D6,$8E,$91,$89
	.dc.b	$ED,$1B,$9D,$3F,$62,$74,$E9,$BD,$5C,$DB,$5F,$4E,$7F,$07,$39,$AB
	.dc.b	$3F,$82,$FA,$BF,$F6,$D3,$FD,$FA,$BF,$D7,$46,$8F,$CF,$46,$9F,$F4
	.dc.b	$56,$1E,$EB,$F5,$4E,$67,$E3,$35,$ED,$EB,$FA,$9E,$6D,$50,$FA,$BF
	.dc.b	$07,$5B,$CF,$7C,$B7,$8E,$F7,$5A,$9C,$E7,$E9,$C3,$93,$F2,$AB,$6D
	.dc.b	$8C,$C7,$CE,$6B,$F4,$B7,$F2,$6D,$FE,$7F,$AE,$4D,$9F,$A1,$3A,$3C
	.dc.b	$91,$9D,$3E,$7F,$BE,$F8,$5C,$D7,$A7,$1F,$F6,$16,$3F,$E0,$AF,$51
	.dc.b	$F4,$D9,$F0,$7E,$0C,$62,$7D,$91,$F1,$EB,$8E,$9E,$98,$C1,$F3,$73
	.dc.b	$7D,$5B,$E8,$BE,$D4,$61,$BA,$FF,$70,$C7,$FC,$51,$5B,$6C,$F0,$8C
	.dc.b	$61,$FE,$DC,$B5,$F1,$8C,$4E,$A7,$F1,$EF,$F8,$0E,$FA,$F7,$73,$EF
	.dc.b	$E8,$C8,$C9,$E2,$8C,$5F,$FB,$5F,$EA,$DF,$0F,$EA,$E3,$FF,$1B,$DF
	.dc.b	$E5,$E7,$2F,$F9,$06,$33,$5C,$6B,$F1,$D5,$A8,$FD,$6B,$5F,$DE,$35
	.dc.b	$FE,$4A,$EF,$2F,$DB,$36,$FF,$CE,$79,$3E,$87,$F2,$93,$9B,$5F,$D2
	.dc.b	$F9,$EF,$94,$7C,$BE,$7B,$8F,$D0,$8B,$5F,$2A,$6D,$7C,$AF,$99,$FA
	.dc.b	$5F,$37,$F4,$DB,$87,$D8,$6D,$FD,$26,$D7,$F0,$DF,$AB,$7D,$5F,$DB
	.dc.b	$F9,$BF,$D5,$46,$30,$FC,$0C,$62,$7F,$69,$39,$C3,$F8,$9B,$5F,$E4
	.dc.b	$4D,$7F,$A5,$BB,$BF,$04,$FE,$27,$E6,$FA,$61,$F8,$EF,$AB,$FC,$67
	.dc.b	$B7,$F8,$D6,$C6,$4F,$C9,$C6,$5A,$7E,$5B,$DD,$FC,$A1,$F9,$5F,$76
	.dc.b	$C6,$49,$9F,$E5,$23,$2F,$D4,$8F,$EC,$3D,$4F,$9C,$FC,$F7,$C2,$FF
	.dc.b	$0D,$8C,$4F,$E9,$23,$13,$FC,$5E,$73,$5F,$CD,$9A,$FF,$45,$6B,$DA
	.dc.b	$38,$D7,$EB,$9B,$1F,$A8,$F2,$FF,$5D,$E7,$7E,$BD,$B5,$FB,$46,$CE
	.dc.b	$E3,$F6,$F3,$FB,$5F,$A5,$91,$FB,$AF,$84,$B5,$FB,$49,$BC,$3F,$B4
	.dc.b	$9B,$5F,$E0,$2F,$AB,$FE,$79,$3F,$B8,$8F,$E5,$D1,$87,$F3,$E6,$3C
	.dc.b	$AF,$E7,$11,$FC,$F7,$0F,$F1,$13,$53,$4E,$4E,$A3,$F8,$9F,$CD,$7F
	.dc.b	$1D,$8F,$A8,$66,$BF,$A9,$79,$3F,$C3,$63,$FB,$2F,$A7,$FC,$D0,$93
	.dc.b	$FC,$F7,$EA,$9F,$55,$FD,$2D,$B1,$98,$C1,$FD,$3F,$E7,$BF,$8D,$DF
	.dc.b	$F7,$B6,$BF,$E2,$46,$BF,$E3,$D6,$A5,$E2,$F3,$25,$7F,$4F,$8C,$BF
	.dc.b	$D4,$7D,$57,$6F,$F5,$F6,$C6,$63,$07,$97,$C7,$91,$CD,$B8,$7F,$81
	.dc.b	$2F,$F8,$84,$F7,$F1,$3D,$84,$7F,$B0,$0F,$5F,$FB,$73,$4C,$B8,$3E
	.dc.b	$27,$87,$23,$D4,$DB,$F0,$78,$9F,$24,$93,$EF,$7F,$BD,$7D,$2F,$F7
	.dc.b	$29,$FF,$0C,$B5,$FF,$14,$35,$FF,$28,$B5,$F9,$DE,$2F,$D5,$25,$7C
	.dc.b	$B7,$EC,$78,$B2,$3E,$73,$67,$FC,$4F,$C5,$F4,$1B,$C1,$FE,$0D,$8C
	.dc.b	$47,$F8,$E5,$AF,$DF,$DF,$F0,$84,$BF,$B4,$FE,$1F,$C3,$91,$F7,$5B
	.dc.b	$3E,$0E,$DE,$C4,$9F,$F0,$85,$FF,$02,$9F,$F4,$33,$FC,$23,$A7,$FC
	.dc.b	$80,$D9,$FF,$45,$43,$9D,$F3,$5F,$E1,$FA,$D5,$FF,$A3,$FE,$87,$FC
	.dc.b	$9C,$FF,$73,$AF,$EA,$0A,$33,$7F,$E4,$09,$CC,$70,$9F,$F2,$D3,$5F
	.dc.b	$F4,$1B,$5E,$DE,$BF,$F8,$D6,$39,$E3,$A8,$FD,$CF,$D0,$FF,$8C,$E7
	.dc.b	$35,$FA,$7F,$13,$F7,$BF,$27,$FE,$37,$E8,$46,$31,$FF,$27,$46,$1F
	.dc.b	$D4,$47,$ED,$48,$FE,$AF,$C6,$FD,$C1,$1C,$12,$FF,$84,$9F,$B5,$8F
	.dc.b	$E1,$BE,$7F,$85,$FF,$24,$47,$FA,$DF,$B3,$FD,$71,$EA,$BE,$0F,$FD
	.dc.b	$47,$C0,$FA,$0F,$F5,$4C,$E6,$FF,$D0,$17,$C2,$F7,$5A,$7F,$CE,$F1
	.dc.b	$FE,$C3,$FA,$0E,$15,$FE,$02,$AF,$F5,$9C,$62,$3C,$41,$DE,$1E,$1C
	.dc.b	$F9,$61,$D2,$B4,$F8,$23,$31,$E4,$32,$7E,$2A,$FB,$CF,$1B,$EB,$18
	.dc.b	$9E,$11,$8A,$E7,$1F,$EC,$4C,$3D,$CF,$38,$AE,$4F,$F2,$E5,$F8,$DA
	.dc.b	$74,$71,$5E,$F8,$57,$6C,$3D,$D1,$E8,$3A,$4F,$8D,$C1,$F1,$9C,$1C
	.dc.b	$E7,$FC,$F9,$1F,$28,$FE,$5C,$61,$FF,$40,$7D,$07,$F2,$1F,$6E,$F8
	.dc.b	$06,$92,$BF,$D5,$F2,$F8,$1F,$F1,$CC,$61,$AF,$FB,$06,$35,$8F,$69
	.dc.b	$CC,$FF,$BB,$27,$2F,$B7,$27,$D0,$7F,$BC,$33,$FC,$2C,$62,$7E,$8D
	.dc.b	$70,$9F,$C4,$B6,$7F,$DE,$E6,$E7,$A5,$73,$8F,$05,$FA,$2F,$8B,$63
	.dc.b	$8C,$EE,$87,$E8,$5C,$2B,$F0,$3A,$6B,$B3,$F0,$F3,$E3,$BE,$F8,$7A
	.dc.b	$B0,$F0,$57,$6A,$D1,$E2,$75,$31,$F5,$9D,$30,$F1,$E1,$E0,$3C,$35
	.dc.b	$CA,$7A,$4E,$AF,$BC,$3E,$1F,$F9,$4B,$FE,$DB,$6B,$FC,$19,$B8,$7F
	.dc.b	$AD,$BC,$6F,$E0,$3B,$7C,$4A,$E5,$D7,$87,$FC,$F7,$F1,$1F,$46,$FF
	.dc.b	$DD,$6D,$AE,$18,$74,$6F,$B1,$C6,$BF,$D4,$FF,$41,$79,$8F,$5C,$7D
	.dc.b	$15,$70,$FF,$4F,$C7,$F4,$4C,$3F,$A2,$DF,$08,$C1,$FE,$8E,$70,$FB
	.dc.b	$6B,$A7,$CF,$FF,$85,$A3,$FD,$F8,$D7,$F9,$3B,$71,$5F,$C1,$F9,$F7
	.dc.b	$D0,$72,$F1,$7C,$D1,$FC,$5A,$75,$C5,$9C,$7A,$D6,$A2,$46,$0F,$0F
	.dc.b	$A3,$E4,$23,$EA,$C7,$FD,$97,$F0,$58,$8F,$AD,$AF,$C3,$31,$3E,$68
	.dc.b	$FC,$CE,$DF,$ED,$27,$97,$E3,$2F,$A0,$FA,$0D,$8F,$C1,$5F,$AB,$E7
	.dc.b	$BE,$87,$D0,$BC,$6A,$66,$BE,$68,$8D,$38,$F2,$5C,$DF,$65,$B1,$C2
	.dc.b	$76,$FF,$54,$E2,$7C,$97,$DF,$F4,$77,$3F,$D6,$1B,$BF,$B6,$15,$C1
	.dc.b	$FC,$D6,$DF,$93,$F2,$1F,$18,$F9,$4C,$36,$B6,$7A,$BE,$87,$0F,$FB
	.dc.b	$E7,$F3,$A7,$0F,$E5,$5B,$1B,$BF,$C9,$6B,$DF,$EC,$DF,$8D,$53,$35
	.dc.b	$FA,$DA,$FC,$F7,$CC,$FD,$38,$FF,$3D,$EE,$8C,$3E,$DE,$7F,$1B,$EF
	.dc.b	$FA,$47,$BD,$FA,$3F,$33,$FC,$CF,$87,$D2,$6B,$FD,$FA,$76,$FE,$F7
	.dc.b	$6F,$D7,$7D,$97,$E5,$63,$FD,$E2,$D7,$F8,$24,$7E,$B7,$E5,$3E,$57
	.dc.b	$DA,$7C,$D9,$FF,$48,$F2,$23,$19,$F9,$EB,$F6,$DF,$4A,$F5,$7E,$97
	.dc.b	$DA,$77,$C0,$F7,$9D,$63,$FC,$47,$BB,$EC,$7C,$37,$CA,$8F,$DC,$DA
	.dc.b	$31,$1F,$C5,$67,$F8,$27,$D3,$FB,$A3,$E5,$F1,$F9,$3C,$7F,$F2,$3B
	.dc.b	$63,$F9,$DC,$EF,$5F,$C5,$64,$F9,$7F,$E1,$DF,$4E,$FA,$0A,$7B,$23
	.dc.b	$3B,$FC,$AF,$4A,$65,$FA,$2B,$F5,$BF,$E2,$EF,$A7,$FF,$46,$3F,$A6
	.dc.b	$6B,$9F,$5F,$F8,$E7,$A8,$F8,$EF,$E3,$D5,$AD,$7F,$3D,$3F,$E4,$4B
	.dc.b	$FD,$7B,$5E,$47,$5B,$E7,$EE,$48,$C1,$FB,$6F,$98,$7C,$7F,$F2,$B6
	.dc.b	$C6,$63,$1E,$7E,$59,$19,$39,$47,$1E,$3E,$5E,$6C,$9E,$D1,$97,$C1
	.dc.b	$19,$7F,$CA,$37,$F9,$F7,$F8,$4F,$D3,$BD,$1F,$E7,$CD,$8C,$C7,$F4
	.dc.b	$1F,$97,$FD,$8C,$7F,$96,$A7,$4F,$F9,$72,$3E,$F2,$6C,$6B,$6F,$E1
	.dc.b	$0D,$E8,$FE,$B5,$3C,$0D,$DF,$F4,$FF,$A7,$7C,$0D,$26,$3F,$C8,$23
	.dc.b	$E9,$63,$FC,$4D,$DC,$66,$76,$FF,$9C,$23,$78,$7F,$69,$DF,$E8,$CF
	.dc.b	$F9,$D1,$6D,$FD,$7D,$B1,$FC,$B9,$DE,$1F,$C8,$5B,$5C,$EB,$81,$C7
	.dc.b	$7F,$ED,$1F,$24,$FA,$7A,$4C,$7A,$35,$F1,$70,$78,$4C,$65,$FC,$BE
	.dc.b	$B9,$4F,$8E,$74,$FF,$42,$8C,$BF,$E8,$EF,$A7,$FE,$33,$97,$F8,$1D
	.dc.b	$FA,$A4,$F2,$BF,$BE,$3A,$9F,$F1,$7B,$AF,$37,$33,$FE,$01,$FA,$2F
	.dc.b	$9B,$5F,$E9,$A6,$E5,$FD,$0D,$B8,$7F,$85,$5F,$9F,$3F,$E3,$A7,$2D
	.dc.b	$E0,$E8,$75,$DF,$FC,$1D,$F4,$EF,$97,$A4,$AF,$EB,$FF,$4E,$F8,$8C
	.dc.b	$47,$FA,$B2,$3E,$BB,$78,$F3,$A7,$F5,$7C,$3A,$4F,$29,$FF,$0E,$57
	.dc.b	$0A,$D6,$18,$E9,$97,$E0,$77,$C5,$88,$F0,$45,$AF,$EC,$97,$FD,$9B
	.dc.b	$6F,$F1,$B7,$85,$F3,$6F,$F8,$E7,$07,$F8,$AE,$30,$7E,$A0,$AE,$4D
	.dc.b	$70,$FF,$8C,$7E,$9F,$FA,$47,$C4,$38,$FB,$61,$FE,$87,$6A,$9F,$E9
	.dc.b	$87,$08,$DC,$F1,$D7,$A7,$BF,$FE,$3B,$FA,$79,$87,$F9,$4D,$FF,$21
	.dc.b	$74,$E2,$3F,$A8,$FE,$C7,$FE,$17,$6F,$B7,$9E,$1F,$F0,$27,$D3,$BE
	.dc.b	$A3,$FC,$A9,$A7,$9F,$1B,$AF,$EA,$31,$FF,$3E,$DE,$DF,$F0,$DB,$56
	.dc.b	$1F,$E2,$BB,$E5,$1F,$E4,$C6,$BF,$EE,$C9,$E1,$3E,$2F,$0F,$FC,$B1
	.dc.b	$1F,$EA,$79,$EB,$E2,$62,$64,$60,$87,$37,$FC,$BD,$E5,$48,$FF,$3B
	.dc.b	$3D,$27,$11,$FD,$D3,$F4,$67,$CC,$FF,$88,$5B,$1C,$0F,$F1,$BF,$D4
	.dc.b	$FF,$97,$7E,$43,$F2,$EB,$E6,$FF,$DB,$9D,$DF,$E7,$D8,$C5,$7F,$A8
	.dc.b	$AB,$7F,$13,$A8,$FF,$52,$CF,$3A,$E1,$79,$94,$F2,$7F,$DB,$B3,$D6
	.dc.b	$B5,$3F,$DF,$7E,$57,$B4,$F8,$7D,$54,$62,$FB,$C6,$1E,$67,$8B,$0E
	.dc.b	$71,$9C,$7D,$AD,$BE,$97,$CE,$FC,$B7,$C2,$31,$87,$8E,$7A,$9E,$AB
	.dc.b	$E4,$F7,$36,$E1,$D4,$E0,$E7,$5D,$63,$9D,$75,$C3,$CB,$5A,$9E,$06
	.dc.b	$AF,$A5,$6F,$0F,$9B,$3A,$BE,$2E,$EF,$93,$49,$1A,$3E,$FE,$1E,$37
	.dc.b	$15,$F1,$27,$CB,$39,$8C,$4E,$A7,$E0,$1D,$23,$24,$73,$1B,$7E,$19
	.dc.b	$B3,$31,$DD,$FF,$65,$1A,$2D,$FD,$16,$E1,$F7,$DC,$F8,$FF,$E7,$C7
	.dc.b	$89,$F8,$E7,$0E,$BE,$13,$2E,$5B,$49,$E5,$87,$E8,$9B,$87,$FB,$77
	.dc.b	$77,$B3,$FE,$AB,$AD,$61,$F8,$6B,$DE,$3F,$A2,$6B,$FE,$4B,$AE,$18
	.dc.b	$7E,$AE,$73,$8F,$86,$F9,$1F,$93,$C7,$BF,$83,$8C,$7F,$0F,$7C,$0E
	.dc.b	$B3,$A8,$CD,$6F,$2F,$2C,$7E,$62,$75,$18,$26,$3E,$3A,$F0,$65,$E6
	.dc.b	$BF,$54,$73,$9D,$47,$39,$D3,$F2,$CF,$09,$2B,$FB,$0D,$F8,$2B,$8C
	.dc.b	$78,$27,$A9,$CF,$1E,$CE,$23,$4F,$C9,$8C,$BD,$E3,$35,$D6,$38,$57
	.dc.b	$62,$23,$91,$E5,$7E,$E7,$AF,$EA,$9E,$53,$FE,$81,$8C,$61,$96,$E1
	.dc.b	$F5,$5B,$3F,$C1,$3E,$75,$F5,$3F,$EE,$FE,$7F,$F7,$87,$B4,$FD,$A5
	.dc.b	$F2,$79,$CB,$C9,$8B,$1E,$31,$FE,$9E,$AD,$BF,$EE,$C8,$CC,$FF,$52
	.dc.b	$8E,$31,$8C,$3F,$DC,$5A,$7B,$E7,$86,$5D,$EF,$E1,$7D,$47,$FB,$22
	.dc.b	$7F,$DE,$F5,$B9,$E8,$6B,$E1,$38,$3F,$EA,$32,$D7,$17,$66,$6B,$FE
	.dc.b	$00,$C6,$70,$F1,$CF,$0F,$AA,$F1,$77,$7F,$9C,$FE,$33,$18,$7F,$69
	.dc.b	$F5,$D1,$8B,$F1,$D6,$EB,$AD,$F2,$BD,$C7,$C5,$E3,$F7,$70,$BF,$65
	.dc.b	$CE,$3E,$6E,$0F,$F3,$5D,$FF,$B2,$A7,$3A,$23,$E5,$E1,$F0,$2B,$77
	.dc.b	$FE,$F4,$F9,$0F,$F1,$F5,$77,$3D,$FE,$27,$B3,$DB,$7D,$5F,$C6,$8C
	.dc.b	$36,$7F,$C0,$D3,$A7,$89,$88,$FA,$D9,$3E,$A9,$F5,$78,$6D,$7E,$6E
	.dc.b	$7F,$1D,$8F,$F6,$48,$C1,$99,$DE,$5F,$72,$BA,$5F,$33,$BD,$F1,$9D
	.dc.b	$57,$0A,$FC,$D3,$FE,$03,$BE,$B3,$FE,$22,$8C,$3C,$3C,$2F,$37,$83
	.dc.b	$2B,$77,$EB,$8F,$54,$E6,$BF,$13,$F5,$6F,$98,$F8,$3C,$7F,$E3,$7C
	.dc.b	$7B,$D2,$B5,$1D,$27,$53,$CA,$B8,$F9,$7F,$98,$BD,$CF,$C1,$AD,$E5
	.dc.b	$F7,$E3,$F8,$24,$EA,$31,$1D,$EF,$E8,$78,$FC,$DF,$A5,$F4,$B3,$2F
	.dc.b	$6E,$BF,$8A,$7E,$84,$66,$FE,$5B,$F6,$7A,$71,$8F,$CC,$FA,$3F,$B9
	.dc.b	$F4,$9F,$75,$B8,$74,$6E,$1F,$51,$B5,$E6,$FA,$B7,$D5,$FE,$17,$E9
	.dc.b	$BF,$6F,$18,$C3,$CB,$18,$9D,$47,$CC,$9D,$61,$F8,$0A,$D4,$E6,$BF
	.dc.b	$21,$F4,$D8,$AF,$C2,$FD,$5F,$E4,$FD,$FF,$CA,$36,$FE,$1B,$67,$71
	.dc.b	$F1,$A7,$F3,$5F,$07,$23,$E4,$FC,$1D,$AF,$C8,$CD,$D3,$F2,$33,$6B
	.dc.b	$F9,$3F,$6B,$E7,$7F,$49,$F1,$7F,$C5,$A3,$13,$F5,$23,$13,$A8,$FC
	.dc.b	$5C,$EB,$1F,$CF,$D6,$A7,$35,$FA,$F3,$13,$FA,$4F,$43,$F6,$5E,$97
	.dc.b	$EC,$DB,$5F,$82,$6C,$EE,$3F,$0B,$3F,$B9,$FA,$B9,$1F,$88,$F8,$AB
	.dc.b	$5F,$B0,$9B,$A7,$EC,$26,$D7,$F4,$6F,$AB,$FE,$58,$FE,$F4,$FE,$6A
	.dc.b	$E2,$FF,$69,$18,$F3,$3F,$84,$34,$FC,$B9,$63,$F9,$CC,$F1,$B7,$D4
	.dc.b	$FF,$15,$9C,$DF,$F2,$3F,$5F,$F8,$AF,$DB,$3E,$13,$19,$7E,$FA,$30
	.dc.b	$FE,$E3,$EB,$3F,$AC,$E1,$FC,$47,$E4,$DE,$F7,$F4,$C9,$CE,$9F,$D9
	.dc.b	$E7,$37,$FC,$03,$EA,$BF,$89,$3F,$C7,$FE,$D9,$CB,$FD,$52,$73,$97
	.dc.b	$FA,$D2,$73,$5C,$7C,$9F,$E8,$7F,$6C,$E3,$E1,$18,$7F,$AE,$FB,$AF
	.dc.b	$03,$FA,$4F,$D4,$BC,$3C,$C6,$1F,$07,$9B,$FD,$3B,$E5,$7F,$94,$70
	.dc.b	$FF,$08,$DF,$F9,$1B,$FE,$00,$8C,$EB,$FD,$FA,$7F,$D6,$6F,$F2,$4F
	.dc.b	$B6,$FE,$F0,$E7,$CC,$FE,$83,$18,$E1,$FE,$F5,$CB,$FD,$13,$ED,$BD
	.dc.b	$57,$FC,$90,$F8,$F3,$FC,$F6,$32,$FF,$74,$C3,$F8,$A3,$FE,$D8,$FA
	.dc.b	$3F,$EF,$D9,$F9,$FC,$BF,$E9,$F1,$8C,$3C,$DE,$67,$F6,$AF,$CE,$7F
	.dc.b	$B2,$47,$D3,$8C,$3F,$E0,$4F,$B1,$EB,$83,$F3,$DF,$F3,$54,$FF,$38
	.dc.b	$9C,$C6,$32,$F2,$79,$FF,$DD,$BE,$DB,$E3,$DF,$C5,$8C,$71,$FD,$DE
	.dc.b	$4F,$F0,$8F,$E8,$0F,$B1,$C5,$FF,$44,$8C,$3F,$D1,$BC,$37,$67,$F8
	.dc.b	$63,$ED,$DF,$75,$FE,$1A,$9C,$D7,$F4,$C9,$CB,$F8,$FE,$BD,$5F,$E3
	.dc.b	$5B,$1F,$93,$F4,$BF,$BA,$FD,$8F,$83,$83,$E7,$46,$34,$ED,$E7,$7E
	.dc.b	$56,$2D,$F1,$3C,$D1,$8A,$FE,$E9,$E5,$73,$E5,$FE,$A3,$AF,$F1,$57
	.dc.b	$FA,$9F,$C1,$3E,$2B,$FC,$0B,$5F,$D5,$F1,$FF,$24,$CE,$5F,$ED,$F1
	.dc.b	$9C,$3F,$CC,$13,$97,$BF,$B3,$FE,$3C,$8C,$61,$FE,$57,$AE,$14,$9C
	.dc.b	$DF,$6C,$3C,$3F,$6D,$C5,$FD,$97,$A9,$E4,$F2,$ED,$78,$9B,$87,$B9
	.dc.b	$B9,$77,$7F,$75,$EA,$FF,$88,$3F,$4C,$7A,$78,$C3,$FC,$13,$18,$7F
	.dc.b	$82,$4E,$70,$F9,$3E,$FD,$C7,$FB,$CC,$7F,$9F,$FC,$BF,$F4,$07,$DB
	.dc.b	$BE,$63,$FC,$DF,$4F,$AE,$FF,$01,$4E,$6F,$FD,$A3,$7C,$2F,$75,$A7
	.dc.b	$FD,$9B,$19,$9F,$F1,$94,$E5,$E2,$2B,$A2,$BF,$06,$5C,$63,$FD,$03
	.dc.b	$76,$73,$1C,$70,$EE,$D3,$BD,$71,$7D,$C6,$4E,$D8,$78,$DE,$A6,$6B
	.dc.b	$AE,$1E,$48,$E2,$F9,$A3,$37,$E0,$8C,$57,$3A,$DC,$FC,$4B,$F1,$43
	.dc.b	$0F,$55,$6E,$30,$78,$5C,$46,$A7,$8B,$FE,$F4,$9F,$59,$FE,$F6,$70
	.dc.b	$F2,$3F,$D4,$E3,$ED,$FF,$A5,$FC,$5F,$FB,$F1,$B1,$A3,$5E,$62,$74
	.dc.b	$D8,$D4,$BE,$57,$EB,$46,$1A,$F8,$23,$5A,$77,$9C,$CF,$FA,$EA,$72
	.dc.b	$FD,$3C,$DF,$51,$FE,$BE,$D7,$F0,$71,$89,$FC,$15,$70,$9F,$C3,$B6
	.dc.b	$7F,$D8,$A6,$E7,$A5,$73,$8F,$05,$FA,$2F,$8B,$63,$8C,$EE,$BF,$DC
	.dc.b	$51,$CE,$B7,$87,$8A,$F8,$56,$AF,$83,$FA,$96,$E1,$E5,$8F,$F7,$63
	.dc.b	$5D,$1F,$9D,$C3,$B1,$E6,$BE,$C6,$6B,$AC,$6A,$79,$DF,$89,$FB,$CD
	.dc.b	$8D,$47,$63,$F4,$17,$E3,$F0,$3F,$C8,$9C,$1F,$BE,$DB,$FE,$00,$46
	.dc.b	$2B,$FD,$32,$D7,$FD,$BE,$FF,$92,$F9,$38,$42,$35,$9F,$FA,$1D,$B5
	.dc.b	$D3,$AD,$87,$09,$E5,$7D,$B1,$7C,$06,$5B,$19,$8C,$7C,$06,$59,$19
	.dc.b	$39,$47,$5F,$61,$EF,$52,$4F,$B2,$3F,$CC,$8D,$8C,$BE,$E8,$E2,$EF
	.dc.b	$2C,$E9,$ED,$9C,$CE,$EB,$C1,$5E,$DF,$C6,$FF,$B9,$4F,$CA,$6B,$F2
	.dc.b	$9B,$7F,$1D,$FF,$71,$C7,$D0,$FC,$E5,$EB,$7C,$76,$C6,$63,$B7,$89
	.dc.b	$F2,$FF,$02,$FD,$FE,$1F,$45,$BC,$3F,$71,$B7,$F5,$F0,$E6,$FD,$CF
	.dc.b	$C7,$3F,$9F,$3F,$80,$6B,$9E,$BF,$B1,$0F,$94,$FB,$5F,$83,$7F,$95
	.dc.b	$AB,$F0,$ED,$7F,$0E,$DC,$BF,$08,$FF,$83,$A3,$F1,$BF,$8E,$3D,$1F
	.dc.b	$C2,$2F,$56,$98,$8F,$C4,$7C,$83,$D3,$BE,$84,$C3,$F2,$F3,$97,$F3
	.dc.b	$3F,$54,$FA,$AF,$C8,$36,$33,$18,$F8,$AF,$CA,$7D,$47,$19,$FC,$FB
	.dc.b	$76,$FC,$E3,$79,$3F,$4E,$FF,$8A,$A7,$F4,$FF,$8E,$3E,$2B,$F3,$CD
	.dc.b	$8C,$C7,$E8,$7C,$7F,$A7,$E0,$BF,$0A,$79,$DF,$A5,$6D,$FE,$C9,$AF
	.dc.b	$EC,$97,$A1,$FA,$BF,$C7,$1F,$7F,$7F,$06,$97,$FB,$A9,$CB,$FB,$CF
	.dc.b	$C7,$13,$FB,$37,$FA,$E6,$FF,$C0,$9B,$3F,$BD,$76,$FF,$04,$F7,$3D
	.dc.b	$24,$7F,$78,$BE,$8F,$F7,$DF,$8E,$3E,$76,$92,$BF,$86,$C3,$E3,$3F
	.dc.b	$82,$7E,$39,$B9,$FE,$22,$D7,$F8,$8B,$79,$BF,$87,$3C,$63,$F8,$BF
	.dc.b	$D2,$E7,$4F,$E3,$6D,$8F,$E2,$CD,$E3,$3C,$63,$F8,$EF,$E3,$0F,$AB
	.dc.b	$FE,$2C,$BD,$AA,$62,$3F,$90,$CB,$B9,$D9,$86,$C0,$FF,$0E,$65,$5E
	.dc.b	$13,$BD,$F8,$32,$38,$6A,$AE,$17,$D2,$79,$9D,$5F,$73,$D0,$E3,$1C
	.dc.b	$2B,$A4,$70,$3D,$17,$FC,$CF,$E9,$FE,$FE,$A5,$F5,$B9,$35,$EB,$CB
	.dc.b	$FE,$8B,$3A,$6C,$E5,$FE,$8F,$F8,$AB,$E9,$BF,$A0,$36,$3F,$A6,$B5
	.dc.b	$DF,$C7,$EA,$64,$69,$E3,$1F,$43,$EF,$3F,$47,$BF,$C5,$AF,$9D,$18
	.dc.b	$DB,$E7,$36,$3E,$97,$DB,$FF,$B7,$5F,$8A,$70,$F0,$3A,$11,$C8,$8E
	.dc.b	$0F,$C9,$9E,$8E,$7D,$6F,$C1,$CF,$F4,$4C,$3C,$92,$AF,$16,$ED,$1F
	.dc.b	$6F,$F1,$3E,$DA,$5F,$C4,$8F,$8E,$61,$F9,$71,$87,$E4,$FD,$9F,$CC
	.dc.b	$23,$F3,$3C,$6A,$FD,$28,$B9,$7F,$75,$C3,$F8,$A4,$FE,$0F,$CA,$FD
	.dc.b	$EB,$7C,$2F,$CE,$C6,$2B,$D6,$FF,$04,$93,$9E,$0F,$CE,$36,$3F,$40
	.dc.b	$D7,$7C,$BD,$2E,$47,$CE,$9D,$1F,$92,$FC,$5F,$F9,$29,$23,$FC,$10
	.dc.b	$FF,$0D,$7F,$C2,$AE,$5A,$6B,$C1,$FF,$0A,$FC,$4B,$F1,$FF,$E9,$93
	.dc.b	$F8,$08,$C7,$CB,$EA,$4D,$F4,$A5,$6E,$72,$D9,$CB,$FA,$BF,$3B,$FB
	.dc.b	$3E,$DF,$80,$6E,$DF,$BF,$6C,$FF,$03,$7F,$B5,$3F,$8E,$72,$9A,$4F
	.dc.b	$57,$F8,$7B,$8F,$5E,$98,$8F,$8B,$F0,$BF,$1B,$EB,$E5,$FC,$68,$CB
	.dc.b	$F2,$23,$2F,$F8,$F3,$7F,$F3,$46,$5F,$E3,$97,$FA,$3C,$67,$87,$E8
	.dc.b	$9F,$4B,$CE,$91,$F4,$A3,$2F,$D4,$8F,$AF,$C7,$23,$EB,$BF,$6B,$C0
	.dc.b	$C4,$78,$62,$E1,$E1,$9F,$BF,$EC,$7E,$03,$EE,$25,$FE,$02,$32,$FE
	.dc.b	$0A,$32,$FA,$E3,$F0,$FF,$1B,$23,$F0,$EF,$E2,$BC,$BC,$47,$C1,$8B
	.dc.b	$87,$C1,$9F,$C7,$FD,$FC,$AF,$C7,$BF,$92,$FB,$9C,$47,$C9,$8B,$5F
	.dc.b	$26,$7F,$2F,$F4,$BF,$98,$F3,$25,$FE,$62,$32,$FE,$6A,$32,$FD,$38
	.dc.b	$FE,$9B,$F6,$9F,$E0,$D8,$FF,$64,$79,$7F,$D2,$63,$FC,$2D,$F1,$FF
	.dc.b	$D9,$5B,$19,$F7,$3F,$DC,$ED,$AF,$F7,$F3,$C7,$DD,$CC,$7F,$55,$AE
	.dc.b	$2D,$AC,$A8,$FF,$35,$FC,$A7,$28,$E1,$E5,$A3,$11,$B8,$C3,$E8,$3B
	.dc.b	$5F,$4A,$E7,$19,$38,$57,$28,$CE,$3E,$DA,$E1,$7D,$23,$E0,$37,$1F
	.dc.b	$88,$DC,$3B,$9C,$DC,$3D,$8F,$05,$F6,$9F,$89,$1B,$C2,$3A,$AE,$35
	.dc.b	$F1,$A3,$95,$F8,$AB,$6F,$9E,$2D,$F9,$26,$C7,$90,$E8,$E1,$E7,$1E
	.dc.b	$27,$24,$9E,$CF,$A3,$0F,$87,$5C,$6F,$D5,$3D,$6B,$64,$BF,$14,$71
	.dc.b	$38,$61,$EC,$3D,$B8,$74,$8F,$F0,$B1,$F7,$B0,$AD,$2E,$1F,$69,$DA
	.dc.b	$FA,$8F,$E6,$D1,$86,$BF,$99,$6B,$F9,$16,$BF,$8F,$7F,$9F,$63,$F9
	.dc.b	$D6,$E3,$FE,$29,$E9,$FF,$1C,$F7,$7E,$1E,$73,$7C,$FC,$3C,$1F,$94
	.dc.b	$AD,$C7,$E0,$27,$6F,$79,$DB,$F9,$39,$DE,$1E,$19,$CB,$67,$5C,$7F
	.dc.b	$E1,$18,$C1,$31,$F7,$57,$86,$75,$18,$AF,$14,$F4,$8F,$64,$EA,$7D
	.dc.b	$46,$89,$5F,$E6,$3B,$F2,$D7,$38,$F0,$4F,$07,$55,$CE,$3D,$C7,$29
	.dc.b	$E0,$78,$31,$F1,$47,$18,$CB,$FE,$59,$BE,$EE,$48,$74,$BE,$F5,$CA
	.dc.b	$31,$5C,$95,$72,$72,$75,$E4,$F3,$46,$43,$8C,$32,$DB,$FB,$8D,$AF
	.dc.b	$E1,$DE,$53,$B3,$FC,$F1,$5F,$E0,$6F,$37,$FC,$8D,$CD,$FD,$77,$D3
	.dc.b	$B3,$FE,$75,$BE,$15,$8C,$11,$CE,$B0,$EA,$BF,$C2,$31,$8C,$3F,$D7
	.dc.b	$31,$FE,$14,$F6,$F0,$67,$E5,$6C,$64,$98,$7F,$71,$8C,$BF,$DC,$FE
	.dc.b	$1B,$A1,$D7,$EC,$2D,$7F,$88,$22,$CF,$F7,$99,$BA,$7F,$98,$F3,$7E
	.dc.b	$A3,$FE,$89,$C3,$FC,$BF,$18,$F9,$6F,$5C,$E7,$0F,$F4,$5C,$E6,$B7
	.dc.b	$39,$BE,$8A,$F8,$CF,$27,$9F,$D2,$C8,$E3,$CD,$71,$EB,$6A,$C6,$1D
	.dc.b	$2A,$DF,$D3,$A7,$57,$9B,$E3,$7C,$ED,$9F,$2B,$71,$EA,$DC,$3C,$FF
	.dc.b	$1D,$EE,$F9,$2F,$7B,$6F,$DA,$DB,$F2,$B6,$BF,$D1,$9E,$83,$E3,$FE
	.dc.b	$05,$FF,$92,$5B,$7F,$EB,$29,$D6,$BF,$89,$78,$6D,$6B,$78,$74,$6E
	.dc.b	$1F,$EC,$5A,$E2,$F5,$AE,$04,$7C,$17,$70,$F0,$78,$EE,$3F,$02,$E3
	.dc.b	$B1,$86,$DF,$F9,$86,$74,$D9,$D3,$F8,$0F,$93,$FC,$29,$F8,$6F,$92
	.dc.b	$B1,$92,$61,$F7,$23,$2F,$FB,$53,$57,$EA,$0F,$86,$FD,$64,$61,$B1
	.dc.b	$FB,$7B,$FF,$59,$CE,$5B,$1C,$5B,$3C,$3A,$78,$B2,$32,$70,$8C,$5F
	.dc.b	$E4,$3E,$4F,$F3,$E7,$E8,$3E,$6A,$D7,$E7,$A2,$CF,$E5,$A6,$E3,$FA
	.dc.b	$0F,$9C,$FD,$3F,$CF,$7E,$A1,$B8,$7E,$99,$B7,$F9,$E6,$D7,$FB,$0B
	.dc.b	$E1,$51,$F5,$BC,$2F,$E1,$D4,$9C,$DF,$D5,$9C,$C6,$27,$FD,$F0,$DB
	.dc.b	$FC,$AF,$0F,$F2,$6B,$FF,$6B,$CE,$55,$48,$C3,$70,$FC,$3F,$85,$D9
	.dc.b	$FE,$07,$F2,$32,$FF,$80,$FB,$76,$FF,$73,$37,$2F,$E0,$73,$6B,$F7
	.dc.b	$1E,$D7,$F0,$CF,$96,$FE,$1A,$D8,$FE,$16,$DC,$3F,$81,$B6,$FF,$9A
	.dc.b	$FD,$77,$E9,$25,$08,$C3,$7A,$7F,$17,$18,$27,$56,$23,$2C,$C3,$F1
	.dc.b	$B3,$B9,$C5,$FF,$20,$F7,$DF,$5B,$FC,$F9,$7E,$35,$63,$0F,$F4,$2F
	.dc.b	$8C,$7A,$F4,$C7,$AF,$73,$FE,$5F,$F8,$D3,$83,$FA,$4B,$63,$3E,$B6
	.dc.b	$27,$FA,$6F,$5B,$8E,$98,$8B,$5F,$CE,$BF,$1A,$FF,$6B,$B7,$93,$FA
	.dc.b	$AC,$75,$E0,$FE,$76,$E9,$FE,$7B,$F8,$D7,$EA,$DB,$DD,$FD,$72,$31
	.dc.b	$CD,$FD,$25,$FD,$A4,$FF,$44,$E9,$7C,$0F,$F6,$25,$F3,$D6,$30,$FF
	.dc.b	$66,$FC,$6B,$3A,$7F,$50,$6D,$FF,$4E,$5F,$31,$BA,$FE,$DB,$F8,$D3
	.dc.b	$E3,$69,$8F,$8B,$B8,$FF,$5A,$FC,$69,$E2,$FF,$74,$6C,$67,$93,$13
	.dc.b	$FD,$E3,$E2,$BF,$AD,$4F,$F7,$F6,$EB,$FD,$F9,$FC,$87,$AD,$FE,$F1
	.dc.b	$9F,$EF,$FE,$67,$F5,$77,$FB,$4B,$FE,$5D,$6F,$83,$AF,$5B,$FC,$0B
	.dc.b	$BB,$C8,$BF,$0D,$86,$DF,$F7,$2F,$C6,$93,$FE,$0C,$6C,$66,$71,$3F
	.dc.b	$E1,$0F,$D2,$9E,$D5,$3F,$C2,$7F,$8D,$FA,$B5,$4E,$B0,$FE,$FF,$F4
	.dc.b	$7F,$E1,$4E,$0F,$EF,$AD,$EC,$FF,$03,$7E,$43,$F7,$2B,$7F,$01,$FE
	.dc.b	$1E,$8C,$5F,$F8,$11,$B5,$FE,$0D,$EC,$7C,$66,$DF,$F1,$29,$8C,$3F
	.dc.b	$C1,$ED,$BF,$F0,$6A,$F8,$DD,$D7,$F8,$B3,$F1,$CF,$C0,$5F,$F8,$59
	.dc.b	$B7,$FE,$14,$5F,$29,$BE,$D9,$5B,$79,$46,$2B,$84,$5C,$BF,$C4,$97
	.dc.b	$CA,$F8,$1F,$E2,$4F,$91,$FB,$0D,$E4,$FF,$1F,$C3,$E3,$EC,$49,$D1
	.dc.b	$C6,$31,$8F,$F8,$AB,$F1,$6F,$E0,$B3,$FE,$30,$6C,$67,$CD,$FF,$25
	.dc.b	$C6,$32,$FF,$1C,$36,$BF,$CA,$8D,$AF,$F2,$9A,$AE,$3E,$77,$F4,$A9
	.dc.b	$FE,$7F,$F1,$5F,$D0,$3E,$A3,$F1,$B1,$FD,$0B,$16,$8C,$B5,$D5,$FE
	.dc.b	$BC,$C2,$DE,$EB,$4B,$0E,$6F,$F4,$C1,$3C,$1F,$E6,$F3,$C2,$77,$1E
	.dc.b	$43,$95,$71,$C3,$C7,$87,$19,$D5,$76,$9C,$C7,$03,$FD,$A7,$87,$78
	.dc.b	$CC,$F6,$9F,$19,$B3,$C7,$5D,$5C,$C7,$5C,$3C,$D8,$7C,$19,$E1,$7D
	.dc.b	$23,$15,$BB,$E5,$87,$87,$0E,$38,$79,$0E,$2E,$1F,$8D,$1B,$9E,$31
	.dc.b	$A6,$BF,$EE,$C3,$D2,$58,$C3,$F2,$4F,$F7,$9F,$D4,$7F,$28,$FA,$9E
	.dc.b	$0D,$8F,$F7,$AB,$5D,$F9,$09,$DF,$7F,$F6,$C9,$DB,$5E,$AD,$7E,$F5
	.dc.b	$7F,$3E,$D3,$89,$F7,$F4,$7C,$AF,$E0,$F5,$FC,$2C,$62,$FF,$09,$7C
	.dc.b	$5F,$A8,$D9,$DD,$71,$AF,$C6,$1C,$6B,$78,$79,$EB,$F2,$37,$C6,$B7
	.dc.b	$7C,$6F,$FC,$3F,$1C,$6B,$77,$C6,$3D,$6D,$9D,$D7,$17,$F9,$9B,$6F
	.dc.b	$B4,$7F,$8A,$9A,$E8,$F8,$51,$B9,$E9,$09,$DD,$78,$CF,$3C,$7C,$49
	.dc.b	$DC,$EA,$BC,$55,$BA,$F0,$57,$9F,$B5,$F1,$2F,$88,$FD,$56,$FF,$ED
	.dc.b	$98,$C5,$FF,$AF,$9A,$FF,$AF,$17,$8B,$C2,$3F,$DB,$B1,$AD,$7F,$D9
	.dc.b	$19,$7F,$88,$27,$FD,$8E,$D7,$96,$1C,$71,$F7,$7C,$1D,$D4,$D3,$1E
	.dc.b	$A5,$BF,$F6,$8C,$E9,$B3,$A7,$FD,$AB,$F1,$1F,$EC,$06,$FC,$4F,$78
	.dc.b	$C4,$7F,$B5,$9F,$86,$E8,$86,$63,$86,$9C,$59,$1A,$D3,$FD,$CD,$5C
	.dc.b	$A3,$FD,$C3,$F1,$3F,$EA,$97,$FD,$CE,$D7,$3F,$11,$F0,$63,$FB,$9C
	.dc.b	$7F,$BB,$3E,$A7,$FC,$7C,$DF,$67,$E3,$46,$23,$89,$FD,$D0,$F9,$5E
	.dc.b	$3F,$F2,$B7,$FB,$A9,$C2,$7E,$77,$D4,$FF,$38,$8F,$F7,$CB,$5F,$F7
	.dc.b	$BB,$72,$FF,$80,$0F,$F8,$DA,$3F,$DF,$FD,$FF,$E2,$E6,$F5,$7D,$98
	.dc.b	$C4,$7F,$C0,$07,$FD,$6C,$FF,$47,$E6,$FF,$44,$F1,$BE,$A7,$FC,$A9
	.dc.b	$39,$11,$9F,$4B,$F0,$93,$A9,$FF,$3C,$D7,$28,$7C,$27,$E2,$7E,$85
	.dc.b	$F2,$D7,$DF,$98,$75,$C9,$5A,$8C,$7C,$DB,$EA,$7F,$24,$D8,$CC,$63
	.dc.b	$E2,$BF,$2D,$EC,$71,$9F,$19,$FE,$3F,$FB,$FC,$9A,$CF,$F3,$9F,$54
	.dc.b	$F9,$FF,$CF,$71,$BE,$66,$9A,$8F,$D1,$7D,$53,$D7,$A4,$CB,$D5,$0F
	.dc.b	$97,$FD,$47,$D5,$7F,$8A,$27,$DC,$FF,$88,$DB,$EB,$EB,$D7,$FD,$77
	.dc.b	$D5,$66,$7D,$3C,$A8,$FE,$CD,$D3,$F5,$9B,$3F,$51,$E1,$DB,$FD,$EA
	.dc.b	$BE,$EC,$FC,$DF,$95,$FC,$84,$E5,$B7,$F8,$07,$E0,$4F,$D3,$FA,$A7
	.dc.b	$CE,$FF,$01,$5F,$51,$4C,$47,$F0,$5F,$A9,$DD,$7D,$B7,$6D,$9D,$4F
	.dc.b	$F0,$BF,$98,$DC,$FB,$9B,$97,$A5,$B9,$7E,$51,$7D,$57,$F1,$3F,$93
	.dc.b	$7D,$25,$FA,$09,$8F,$E2,$E7,$2F,$E6,$4F,$C6,$F7,$7F,$23,$69,$9F
	.dc.b	$07,$FD,$E8,$DD,$35,$18,$8F,$C5,$9C,$0A,$E7,$4F,$F1,$4C,$7E,$6A
	.dc.b	$78,$D4,$8E,$CF,$48,$FE,$59,$AF,$DA,$E1,$F9,$BE,$FE,$0C,$FA,$B3
	.dc.b	$0F,$D3,$57,$FA,$3D,$B3,$97,$FD,$B9,$F5,$5F,$CD,$FE,$64,$D3,$1F
	.dc.b	$55,$67,$F6,$13,$A6,$CE,$9F,$91,$F5,$3F,$9F,$6F,$77,$C9,$8C,$47
	.dc.b	$C4,$FA,$A7,$C7,$E6,$BE,$74,$7F,$3A,$E0,$F8,$AD,$E7,$FF,$26,$61
	.dc.b	$CA,$7E,$4B,$63,$FC,$A1,$3B,$C0,$1C,$29,$E0,$FF,$16,$CB,$FD,$98
	.dc.b	$FB,$BE,$23,$E2,$1F,$17,$E2,$24,$64,$87,$EC,$27,$8B,$7E,$4B,$E6
	.dc.b	$11,$F9,$3E,$25,$74,$7C,$C8,$FA,$51,$F8,$FF,$33,$F6,$B9,$FF,$9A
	.dc.b	$DB,$5E,$C6,$A7,$F7,$AD,$F0,$B9,$11,$FC,$E4,$68,$E2,$48,$F0,$72
	.dc.b	$7F,$79,$F9,$7F,$DF,$D7,$F3,$A6,$F3,$1A,$7F,$02,$8F,$E2,$2E,$1E
	.dc.b	$47,$3F,$1F,$FC,$19,$F4,$6F,$AA,$FE,$85,$C7,$FE,$8A,$8C,$7C,$DE
	.dc.b	$A4,$DF,$52,$61,$FD,$36,$B6,$DA,$DB,$67,$F7,$1E,$87,$F6,$1D,$FF
	.dc.b	$D2,$8D,$9F,$C2,$B5,$3F,$D9,$5F,$DC,$B9,$27,$C4,$F4,$F8,$EE,$AC
	.dc.b	$9F,$E1,$13,$B9,$F9,$5F,$0F,$F2,$FE,$9A,$57,$F5,$9D,$FF,$CB,$5B
	.dc.b	$7F,$31,$6E,$5F,$DB,$15,$70,$AF,$9B,$7F,$57,$CD,$FA,$DC,$72,$7A
	.dc.b	$46,$5E,$D1,$97,$EA,$C7,$DD,$E3,$91,$E3,$7F,$01,$E0,$62,$3E,$DC
	.dc.b	$5C,$FC,$D3,$F8,$5F,$67,$F0,$DF,$55,$2F,$D9,$19,$7D,$D1,$97,$F0
	.dc.b	$B1,$F8,$DF,$9B,$91,$F0,$DF,$C8,$79,$98,$8F,$C5,$C5,$C3,$E3,$4F
	.dc.b	$E5,$7E,$3A,$57,$CB,$7F,$31,$F1,$F8,$8F,$C9,$C5,$AF,$9D,$3F,$9D
	.dc.b	$FA,$AF,$CF,$79,$D2,$FE,$A4,$57,$F3,$71,$6B,$EB,$4F,$F4,$3F,$A9
	.dc.b	$7B,$BF,$EB,$AF,$33,$FD,$C1,$1F,$E0,$BF,$91,$FE,$AE,$D8,$CC,$EF
	.dc.b	$1F,$E2,$B3,$C9,$FF,$7A,$CF,$2F,$73,$50,$9E,$4F,$0A,$CC,$3E,$21
	.dc.b	$F5,$5C,$AB,$AC,$61,$FF,$79,$E1,$E6,$8C,$63,$E6,$C7,$9C,$62,$75
	.dc.b	$7C,$EF,$C5,$39,$7D,$35,$C2,$75,$87,$8D,$B5,$BA,$F5,$56,$DF,$19
	.dc.b	$E6,$9D,$9C,$A3,$58,$7C,$AA,$E5,$7D,$6F,$89,$D0,$E2,$61,$D4,$F1
	.dc.b	$BE,$93,$A8,$C3,$BA,$EB,$7E,$0C,$3E,$0C,$E6,$F8,$B4,$97,$C8,$F9
	.dc.b	$4F,$59,$E3,$19,$7E,$E4,$64,$F8,$07,$27,$2D,$33,$1C,$5F,$9E,$7B
	.dc.b	$8F,$F4,$26,$1F,$0D,$B8,$7D,$F6,$DF,$FA,$25,$FC,$7F,$CD,$7F,$37
	.dc.b	$7F,$2E,$D7,$3E,$1E,$DF,$C8,$46,$5B,$1F,$7A,$3F,$C8,$38,$7F,$8D
	.dc.b	$3A,$9C,$4E,$CD,$5C,$C3,$EA,$61,$C3,$4F,$EE,$B1,$FC,$7A,$F8,$3A
	.dc.b	$AE,$18,$FE,$AA,$F8,$4F,$F7,$CC,$7B,$E3,$AC,$7D,$78,$5A,$D6,$3D
	.dc.b	$31,$E5,$39,$9F,$59,$F8,$AC,$39,$BE,$12,$4F,$07,$55,$CE,$FB,$C6
	.dc.b	$9B,$3F,$0E,$75,$3F,$30,$D1,$63,$58,$78,$F4,$65,$EE,$AF,$2D,$69
	.dc.b	$E2,$63,$0E,$E7,$CA,$C7,$D3,$3E,$72,$46,$A7,$E1,$E1,$DE,$FB,$61
	.dc.b	$E0,$CB,$F7,$D8,$F2,$C3,$86,$1F,$D4,$B9,$3D,$6F,$D7,$C7,$71,$8A
	.dc.b	$E1,$3F,$56,$31,$9E,$5B,$5F,$C5,$7C,$4F,$F0,$57,$9B,$F8,$0E,$3F
	.dc.b	$EC,$4D,$9F,$EC,$5E,$AD,$C7,$FC,$25,$36,$FF,$C2,$53,$7B,$FF,$D5
	.dc.b	$D7,$FD,$B6,$31,$59,$F7,$24,$FF,$71,$FA,$9B,$8F,$F8,$62,$6E,$5F
	.dc.b	$DC,$E6,$D7,$F7,$3F,$8A,$EC,$77,$F9,$AB,$1F,$DE,$E2,$E1,$FE,$22
	.dc.b	$9B,$7F,$EA,$DF,$AA,$FD,$25,$FF,$B6,$57,$D3,$FA,$DC,$DF,$FA,$E9
	.dc.b	$CC,$EC,$D4,$ED,$A7,$3F,$57,$91,$CF,$9A,$C6,$49,$7D,$63,$2F,$7E
	.dc.b	$9F,$17,$C7,$F8,$DB,$5E,$76,$CE,$E3,$D5,$3E,$9F,$8F,$F5,$7C,$9D
	.dc.b	$9F,$73,$6B,$D8,$DD,$3E,$23,$F7,$7C,$F7,$C8,$7C,$0A,$FF,$79,$C6
	.dc.b	$35,$D4,$FF,$A2,$30,$FF,$21,$3B,$C3,$FC,$94,$D8,$FF,$40,$D7,$17
	.dc.b	$7A,$F2,$B9,$1C,$9E,$B1,$8C,$3C,$3E,$3B,$8B,$CC,$B8,$AC,$61,$B8
	.dc.b	$7F,$95,$67,$4D,$9D,$3F,$81,$F9,$4F,$C0,$1F,$81,$F9,$2B,$19,$26
	.dc.b	$1F,$83,$8C,$AD,$7F,$51,$F5,$4F,$86,$FD,$34,$23,$17,$FE,$A2,$8D
	.dc.b	$F5,$6A,$73,$7C,$EB,$51,$8A,$EB,$7B,$BE,$0C,$AE,$87,$07,$4E,$2F
	.dc.b	$CD,$7C,$E3,$FD,$67,$D6,$E7,$7C,$AF,$E8,$4F,$D1,$7D,$55,$AF,$D3
	.dc.b	$45,$C3,$F4,$93,$73,$FD,$67,$D5,$7E,$A7,$E8,$7F,$54,$DC,$3F,$5C
	.dc.b	$DB,$FD,$63,$6B,$FD,$E3,$F0,$A8,$FB,$BE,$17,$F0,$FA,$4E,$6F,$EE
	.dc.b	$4E,$63,$13,$F8,$68,$C5,$72,$AF,$85,$3C,$A3,$31,$F8,$5F,$0A,$CF
	.dc.b	$EF,$FE,$4B,$F8,$11,$FC,$0F,$D9,$B5,$FC,$1E,$2C,$FF,$05,$9B,$97
	.dc.b	$F0,$DF,$6F,$F8,$57,$CA,$FF,$0B,$6E,$1F,$C3,$DB,$7F,$C3,$5B,$5F
	.dc.b	$CC,$7E,$AB,$F5,$13,$F8,$D6,$C1,$FD,$84,$7E,$1C,$D7,$67,$E4,$23
	.dc.b	$64,$EC,$C4,$4C,$3F,$09,$7C,$E3,$84,$E6,$FF,$DD,$5F,$55,$FC,$F7
	.dc.b	$E0,$9F,$55,$63,$0F,$F4,$1F,$8C,$7A,$F4,$C7,$AF,$74,$FC,$97,$53
	.dc.b	$EA,$9C,$3F,$D2,$57,$AB,$63,$0F,$F4,$DE,$B7,$25,$E4,$C3,$71,$EB
	.dc.b	$F5,$5F,$ED,$66,$CF,$F5,$58,$EF,$18,$3C,$7F,$55,$FA,$3E,$26,$88
	.dc.b	$FF,$76,$4E,$BC,$0F,$EB,$D1,$8A,$FA,$14,$E9,$7C,$0F,$F6,$35,$F3
	.dc.b	$F6,$30,$FF,$67,$FA,$AC,$E5,$F1,$1B,$87,$45,$F3,$1B,$AF,$ED,$DF
	.dc.b	$54,$F8,$CB,$F1,$B8,$6E,$3F,$3B,$EA,$9E,$2F,$F7,$55,$E1,$58,$C3
	.dc.b	$FD,$E7,$E2,$7F,$AF,$F8,$5F,$E4,$0E,$1F,$F7,$8E,$19,$F0,$B4,$FF
	.dc.b	$7F,$F3,$3F,$23,$5B,$7F,$C4,$F1,$F7,$9C,$BF,$C3,$9C,$F0,$FC,$8D
	.dc.b	$DF,$0D,$FE,$09,$6D,$7E,$19,$79,$F9,$9F,$F0,$77,$D5,$6A,$FE,$8B
	.dc.b	$71,$F4,$3C,$30,$FC,$7B,$97,$F8,$FF,$D4,$BD,$AF,$F0,$C3,$63,$F2
	.dc.b	$BF,$55,$F0,$9D,$CF,$F8,$77,$E8,$FF,$C3,$78,$FE,$01,$B8,$FC,$7F
	.dc.b	$AA,$FC,$0D,$FE,$7A,$72,$FF,$2E,$8C,$D7,$E2,$1B,$5F,$3F,$A9,$F1
	.dc.b	$9F,$E2,$B9,$CD,$FF,$36,$9C,$BF,$AA,$6D,$FE,$49,$7C,$6F,$F8,$CD
	.dc.b	$CE,$1F,$6F,$EA,$BF,$11,$7F,$B2,$6D,$FE,$69,$75,$7F,$8E,$DC,$D7
	.dc.b	$1E,$AC,$19,$9B,$7C,$E2,$DF,$39,$B5,$F8,$BF,$90,$FA,$93,$FA,$86
	.dc.b	$C6,$7E,$3B,$FC,$95,$18,$C3,$FA,$AB,$6B,$F6,$AD,$E8,$2A,$E3,$CB
	.dc.b	$FE,$53,$FA,$AF,$E2,$78,$7F,$10,$FC,$77,$FD,$17,$97,$E7,$67,$39
	.dc.b	$7E,$EA,$73,$5F,$D9,$27,$35,$BF,$37,$A7,$9B,$D5,$FE,$A9,$80,$7C
	.dc.b	$65,$FA,$AF,$CE,$C7,$F7,$FD,$7F,$D5,$71,$88,$FF,$57,$CE,$FA,$FF
	.dc.b	$D6,$38,$FB,$33,$C5,$7F,$AD,$0E,$A2,$F9,$4E,$E3,$15,$FE,$CD,$BE
	.dc.b	$31,$FE,$15,$C3,$C3,$39,$8D,$9E,$5C,$36,$FA,$EB,$8D,$F5,$C3,$D1
	.dc.b	$39,$38,$9D,$6B,$8B,$E6,$8F,$2E,$1E,$08,$F5,$5F,$48,$C4,$7C,$23
	.dc.b	$9D,$F7,$C3,$CD,$7C,$B0,$EB,$87,$16,$DF,$0C,$39,$BE,$5A,$E0,$F9
	.dc.b	$8F,$41,$FE,$EB,$AF,$F7,$63,$83,$E2,$1F,$3F,$EA,$BF,$98,$FD,$5F
	.dc.b	$FB,$C5,$B1,$FE,$F9,$6B,$BF,$19,$3B,$EF,$FF,$22,$4E,$DA,$F5,$6B
	.dc.b	$FE,$F0,$AF,$E8,$FA,$71,$3E,$CE,$8F,$95,$FF,$7A,$E7,$F8,$98,$C5
	.dc.b	$FD,$EB,$E2,$FD,$46,$CE,$EB,$8D,$7F,$BF,$8E,$35,$BC,$3C,$F5,$F9
	.dc.b	$5B,$E3,$5B,$BE,$37,$F9,$F8,$E3,$5B,$BE,$31,$EB,$6C,$EE,$B8,$BF
	.dc.b	$D0,$1B,$7D,$A3,$F5,$4D,$74,$7C,$28,$DC,$F4,$84,$EE,$BC,$67,$9E
	.dc.b	$3E,$24,$EE,$75,$5E,$2A,$DD,$78,$2B,$CF,$C8,$F8,$77,$C0,$FE,$D3
	.dc.b	$83,$FC,$CF,$18,$BF,$F0,$DB,$5F,$F0,$1B,$FE,$6A,$73,$3C,$5F,$F0
	.dc.b	$6C,$6F,$4F,$ED,$CD,$AE,$BE,$8E,$F0,$F6,$FB,$D7,$5F,$F3,$26,$3F
	.dc.b	$E2,$A6,$BF,$E1,$25,$1B,$38,$7B,$92,$78,$47,$3F,$57,$A7,$AB,$27
	.dc.b	$A4,$65,$ED,$19,$7B,$46,$5F,$04,$67,$4F,$F0,$46,$1F,$E3,$C9,$FF
	.dc.b	$1F,$FE,$92,$D3,$D8,$FF,$BB,$A3,$3E,$56,$27,$2F,$C1,$FA,$CF,$F2
	.dc.b	$1C,$FF,$92,$1A,$FF,$8C,$2B,$93,$FE,$32,$8C,$E1,$FE,$19,$F1,$9F
	.dc.b	$23,$F2,$DB,$5C,$F3,$F1,$FC,$4C,$8F,$3B,$F4,$7E,$B3,$FC,$73,$1E
	.dc.b	$AF,$3B,$C9,$F5,$52,$30,$7F,$7F,$F9,$2E,$C4,$7D,$11,$F6,$FE,$8C
	.dc.b	$D2,$47,$F9,$93,$F0,$07,$CB,$FE,$07,$E7,$5F,$2F,$7E,$96,$5F,$F9
	.dc.b	$AA,$72,$FF,$9B,$3E,$77,$8B,$6F,$F1,$51,$88,$FF,$2B,$AF,$AC,$FC
	.dc.b	$77,$8B,$FC,$0B,$93,$FC,$80,$DE,$8F,$F6,$2B,$6F,$FC,$CA,$B7,$F3
	.dc.b	$3E,$7F,$81,$93,$E5,$8F,$CD,$7D,$57,$F5,$6B,$FF,$67,$B6,$7F,$CD
	.dc.b	$ED,$EF,$3F,$E4,$08,$FF,$43,$FD,$55,$7F,$CD,$ED,$CB,$FC,$AC,$DB
	.dc.b	$FF,$6D,$BF,$E0,$39,$FF,$47,$E3,$E9,$C3,$FD,$E9,$97,$EB,$63,$D5
	.dc.b	$3A,$F1,$BF,$61,$F5,$5F,$E0,$C8,$FF,$4B,$B5,$FF,$3F,$B7,$2F,$F3
	.dc.b	$33,$FE,$80,$8F,$F4,$EF,$D5,$67,$3F,$F7,$6B,$63,$FD,$1A,$DC,$8F
	.dc.b	$18,$FF,$45,$E3,$E9,$6E,$FE,$73,$D1,$E9,$7F,$05,$F8,$A7,$CE,$D2
	.dc.b	$57,$FA,$BA,$1F,$3B,$FC,$46,$1D,$5F,$E3,$E7,$FD,$C6,$BE,$97,$4D
	.dc.b	$8C,$FA,$F8,$33,$32,$32,$43,$FD,$A4,$76,$74,$E2,$7C,$0F,$FA,$A3
	.dc.b	$E4,$1F,$51,$FC,$7D,$B1,$FE,$C6,$6B,$BF,$A9,$D4,$C8,$D3,$C6,$3F
	.dc.b	$94,$7D,$57,$C9,$E4,$FF,$4D,$37,$8F,$FC,$6D,$C6,$8C,$61,$FE,$C0
	.dc.b	$5C,$3C,$B8,$64,$FF,$B6,$27,$75,$C6,$BE,$67,$D4,$FF,$B0,$6B,$FC
	.dc.b	$85,$8F,$F1,$F7,$FC,$89,$19,$7E,$2B,$63,$50,$FA,$AF,$F5,$AB,$CB
	.dc.b	$2F,$F2,$62,$AE,$11,$86,$C7,$6F,$8E,$FB,$D5,$CA,$3B,$C6,$3E,$46
	.dc.b	$47,$53,$81,$FE,$F4,$F9,$AF,$E5,$38,$7C,$65,$F4,$DF,$D4,$DB,$3B
	.dc.b	$F2,$B0,$66,$64,$64,$E9,$E3,$F1,$23,$E1,$D9,$FA,$EF,$F1,$5D,$BF
	.dc.b	$DA,$18,$FF,$26,$AE,$24,$3F,$DD,$F3,$BC,$BE,$2F,$8A,$F7,$FF,$3C
	.dc.b	$BE,$A2,$98,$89,$7F,$04,$4E,$DC,$35,$FC,$DB,$D5,$CC,$74,$77,$3D
	.dc.b	$1F,$2F,$8D,$FD,$E7,$0F,$E5,$D1,$EF,$6B,$EA,$F9,$DD,$4C,$8D,$3C
	.dc.b	$63,$FC,$11,$F4,$4F,$49,$E4,$E1,$78,$74,$C4,$4C,$3E,$F5,$71,$9D
	.dc.b	$46,$2B,$9B,$F0,$FE,$43,$9B,$7E,$83,$F8,$44,$1B,$52,$74,$43,$F8
	.dc.b	$0C,$F2,$6C,$72,$99,$1D,$E3,$DB,$E5,$BD,$0B,$EE,$CA,$F4,$CE,$5B
	.dc.b	$39,$5C,$9F,$AF,$CF,$FA,$4D,$F4,$AF,$A5,$E8,$7D,$3F,$42,$61,$F4
	.dc.b	$E3,$2F,$D5,$8C,$BF,$8E,$8F,$B5,$E0,$7D,$BF,$83,$93,$F6,$E3,$2F
	.dc.b	$DD,$8C,$BF,$96,$8F,$C1,$7C,$54,$8F,$C1,$3F,$85,$F8,$0C,$47,$E7
	.dc.b	$22,$E9,$FC,$27,$AF,$F4,$B5,$F8,$BE,$3F,$C6,$70,$4C,$3F,$19,$19
	.dc.b	$7F,$1D,$19,$7F,$4F,$1F,$93,$E1,$FC,$A7,$CE,$49,$FC,$A4,$65,$FC
	.dc.b	$B4,$65,$FD,$7C,$7E,$6F,$E7,$24,$7E,$6D,$D6,$5E,$7E,$84,$6A,$7F
	.dc.b	$6E,$EB,$B3,$F8,$DF,$D4,$7F,$0A,$9F,$F7,$3F,$0E,$3C,$1F,$F7,$1B
	.dc.b	$63,$FD,$A4,$F0,$F0,$73,$1F,$EE,$FA,$E5,$F0,$FF,$E2,$28,$47,$F9
	.dc.b	$1A,$30,$DA,$E5,$3D,$0E,$5F,$4D,$FE,$28,$AD,$B6,$7C,$ED,$9F,$06
	.dc.b	$1E,$27,$BC,$65,$BE,$C7,$56,$9E,$E3,$89,$9C,$13,$C5,$B3,$98,$C5
	.dc.b	$77,$9C,$D7,$33,$AE,$1E,$8B,$E9,$5B,$BE,$84,$8C,$D7,$56,$3C,$EF
	.dc.b	$AD,$FC,$8B,$E5,$5A,$8E,$D5,$B9,$CB,$C6,$FB,$39,$38,$61,$F2,$AE
	.dc.b	$C7,$C2,$38,$38,$7D,$71,$F8,$17,$24,$BF,$09,$E5,$78,$61,$E8,$9E
	.dc.b	$D1,$E1,$38,$9F,$DA,$8F,$A6,$58,$F9,$6D,$BF,$F3,$33,$F9,$6F,$A7
	.dc.b	$FF,$43,$BB,$6B,$9F,$7B,$A3,$F9,$38,$CB,$63,$EF,$47,$FA,$43,$0F
	.dc.b	$CB,$7B,$2F,$A2,$FF,$03,$47,$E5,$A7,$59,$75,$C3,$EA,$46,$11,$FE
	.dc.b	$71,$AE,$33,$BA,$E5,$87,$80,$FC,$96,$1D,$F6,$C6,$1D,$26,$D7,$93
	.dc.b	$0E,$73,$E9,$34,$48,$D5,$70,$C3,$C0,$49,$F2,$E1,$DD,$E1,$19,$7F
	.dc.b	$6B,$8F,$AD,$F5,$57,$37,$C1,$7E,$A7,$F3,$D7,$CE,$30,$4B,$E9,$19
	.dc.b	$7B,$66,$C3,$B3,$72,$F8,$79,$7B,$B1,$F1,$C7,$85,$A7,$82,$35,$3C
	.dc.b	$63,$D6,$79,$AF,$FC,$6B,$87,$86,$30,$7C,$2F,$2B,$B4,$66,$3F,$D5
	.dc.b	$11,$8B,$CB,$73,$FA,$ED,$9F,$E1,$1D,$1F,$E2,$6F,$C0,$BF,$C9,$9C
	.dc.b	$9F,$CB,$DB,$7F,$E5,$CE,$4F,$F3,$5B,$FE,$0F,$8C,$DA,$F9,$4E,$4F
	.dc.b	$F0,$A7,$A3,$A8,$CC,$FF,$8B,$BD,$1C,$61,$D2,$31,$7C,$0F,$F0,$E4
	.dc.b	$EA,$F6,$D8,$FF,$10,$7C,$87,$2F,$5B,$9B,$63,$27,$F8,$96,$32,$D3
	.dc.b	$FC,$51,$EB,$78,$8F,$1F,$D6,$58,$C9,$31,$FF,$22,$C6,$5F,$F0,$2C
	.dc.b	$7F,$A0,$3F,$5E,$FD,$4B,$FE,$51,$5C,$9F,$05,$CC,$F0,$3E,$43,$A9
	.dc.b	$E1,$1B,$9E,$0D,$8E,$9E,$F7,$3E,$3E,$8D,$8C,$9E,$18,$CB,$4F,$F1
	.dc.b	$BF,$E7,$0D,$3E,$CF,$92,$86,$B8,$75,$5C,$23,$15,$96,$C6,$4F,$F4
	.dc.b	$07,$1B,$E9,$FE,$E3,$F8,$C5,$C1,$F8,$88,$D7,$0D,$9C,$D7,$27,$8B
	.dc.b	$FE,$A0,$73,$AF,$C9,$AE,$33,$BB,$E5,$7C,$DB,$19,$3F,$CA,$FF,$2F
	.dc.b	$FE,$DC,$6F,$BB,$F0,$A3,$18,$7D,$5D,$51,$FE,$DE,$AE,$3E,$FE,$EF
	.dc.b	$83,$63,$F0,$1F,$31,$F7,$CF,$C0,$7C,$C5,$8C,$93,$1F,$C1,$74,$39
	.dc.b	$55,$B9,$F7,$7D,$3B,$E7,$FE,$42,$F9,$FB,$18,$6E,$3F,$4A,$74,$D9
	.dc.b	$D3,$4D,$4F,$0A,$EE,$FE,$5F,$E5,$24,$FE,$6F,$E6,$AD,$7A,$32,$56
	.dc.b	$38,$FF,$C9,$90,$AC,$5F,$E9,$7E,$0F,$F4,$5F,$07,$FA,$36,$C7,$E9
	.dc.b	$DB,$8F,$56,$E1,$FE,$DA,$F6,$BF,$D8,$F5,$F8,$EF,$C5,$1E,$1F,$F1
	.dc.b	$2C,$15,$A7,$F0,$D3,$AA,$FF,$5A,$4E,$6F,$A7,$99,$F9,$1F,$4F,$94
	.dc.b	$E6,$BF,$DC,$AD,$C3,$FB,$2B,$6F,$F3,$D5,$C2,$FF,$7D,$8C,$C3,$F8
	.dc.b	$17,$8D,$71,$F9,$13,$70,$F8,$F3,$6B,$FD,$C9,$C1,$FC,$9F,$E0,$3E
	.dc.b	$CE,$1F,$86,$5F,$4D,$99,$95,$FE,$ED,$78,$46,$2B,$F1,$3E,$5F,$E2
	.dc.b	$3C,$DF,$C4,$B6,$FF,$0C,$D9,$DC,$7E,$26,$76,$D3,$65,$8F,$CC,$FC
	.dc.b	$37,$FB,$1A,$3F,$DE,$0D,$8C,$FA,$5F,$D0,$63,$1A,$7F,$A4,$7E,$9B
	.dc.b	$FD,$69,$7F,$EF,$75,$EE,$7F,$47,$73,$87,$E9,$9C,$CF,$FB,$F7,$E9
	.dc.b	$FE,$2D,$FF,$A6,$DB,$1F,$E1,$AE,$D4,$3E,$1B,$FA,$93,$AC,$3F,$AC
	.dc.b	$7C,$C3,$E6,$3F,$AB,$37,$0F,$F1,$1E,$3F,$EA,$78,$E1,$89,$D5,$F9
	.dc.b	$78,$7D,$1C,$3E,$96,$E1,$C8,$F2,$7D,$3B,$DD,$FE,$C8,$D8,$D5,$FF
	.dc.b	$6B,$8F,$F7,$9C,$63,$14,$6E,$7F,$84,$F3,$3D,$EF,$ED,$CD,$BE,$0B
	.dc.b	$DF,$CC,$FF,$73,$FC,$31,$F0,$F7,$EB,$F0,$DC,$3C,$1F,$50,$F6,$BF
	.dc.b	$BD,$36,$33,$E7,$E2,$7F,$BF,$7D,$47,$C5,$9F,$43,$75,$FE,$3E,$FC
	.dc.b	$8E,$4F,$F6,$93,$97,$FD,$A9,$9B,$EB,$3C,$DF,$4D,$E7,$6E,$1F,$0F
	.dc.b	$E2,$0F,$F8,$35,$AE,$7E,$73,$0F,$F8,$47,$E6,$FD,$EF,$CB,$6E,$7F
	.dc.b	$CC,$BD,$47,$C5,$6F,$87,$FC,$31,$FA,$57,$F5,$0D,$BF,$DD,$15,$AC
	.dc.b	$7E,$A4,$E5,$FC,$84,$67,$0F,$E7,$B3,$FD,$A9,$BE,$DF,$F8,$8F,$EA
	.dc.b	$7D,$D7,$CA,$3F,$C4,$D7,$F5,$E3,$CD,$F6,$18,$E8,$F5,$3F,$E2,$BF
	.dc.b	$31,$F0,$19,$AF,$C3,$46,$1F,$C4,$5F,$3C,$BF,$A7,$CB,$E0,$39,$5F
	.dc.b	$36,$DF,$E3,$FF,$1B,$7C,$1F,$F8,$E7,$3F,$F0,$CE,$FF,$43,$2F,$F2
	.dc.b	$05,$6A,$FF,$2D,$5A,$7F,$65,$19,$9F,$EB,$F3,$97,$8D,$F5,$7F,$C8
	.dc.b	$D3,$BB,$ED,$E0,$7E,$66,$73,$5B,$C3,$BF,$B5,$E5,$F6,$3C,$CD,$8C
	.dc.b	$9E,$48,$CB,$4F,$D6,$7D,$47,$F4,$5B,$FE,$E6,$DC,$7C,$2F,$FA,$3D
	.dc.b	$BE,$17,$F9,$6E,$35,$5F,$C3,$DB,$5F,$E6,$37,$84,$66,$7F,$AB,$4F
	.dc.b	$FA,$BB,$E2,$7F,$D5,$FF,$4D,$FD,$AA,$3F,$87,$ED,$F3,$63,$11,$FE
	.dc.b	$B4,$13,$C3,$2C,$65,$F7,$6B,$75,$D4,$3C,$A3,$53,$FE,$CC,$AE,$55
	.dc.b	$E0,$C6,$BD,$E2,$D7,$08,$FF,$6F,$46,$30,$F3,$5F,$39,$EE,$F9,$23
	.dc.b	$31,$C0,$C9,$DC,$A7,$72,$D7,$77,$18,$7A,$67,$9C,$62,$BB,$61,$E8
	.dc.b	$8F,$81,$5B,$8C,$47,$C8,$3A,$D7,$27,$75,$C9,$F1,$9F,$24,$FF,$78
	.dc.b	$9C,$4B,$86,$1D,$1F,$57,$E8,$DF,$20,$FA,$7E,$31,$86,$C6,$67,$FD
	.dc.b	$FB,$39,$6F,$92,$D3,$8B,$63,$31,$82,$3F,$2B,$1F,$E0,$6F,$DA,$6B
	.dc.b	$F3,$AB,$FA,$84,$FD,$E6,$C7,$1E,$4F,$C7,$6B,$F9,$0B,$E2,$FE,$4A
	.dc.b	$B6,$DC,$3C,$8E,$EB,$8D,$7E,$5C,$E2,$7C,$59,$DB,$70,$F3,$4F,$0C
	.dc.b	$07,$F9,$31,$C2,$C3,$C3,$5C,$63,$D0,$72,$3D,$97,$D9,$E7,$3C,$1E
	.dc.b	$B3,$E4,$6D,$76,$9D,$47,$7A,$E4,$F8,$CE,$C7,$CB,$7B,$71,$3E,$9F
	.dc.b	$FC,$33,$3F,$DD,$DA,$FF,$9F,$9B,$8F,$F1,$C7,$FC,$15,$1B,$F4,$BF
	.dc.b	$77,$C1,$FE,$19,$CF,$FB,$84,$FF,$80,$5A,$F3,$9D,$BF,$EF,$1C,$9F
	.dc.b	$4F,$8C,$BE,$47,$A1,$67,$FC,$3D,$5F,$E0,$B7,$6D,$AD,$4C,$8C,$1B
	.dc.b	$FA,$7E,$5F,$25,$23,$04,$3F,$D3,$76,$8D,$46,$27,$C4,$63,$5F,$E7
	.dc.b	$B5,$B9,$E5,$1F,$D0,$3E,$9D,$E2,$FA,$DB,$3F,$E2,$36,$BF,$EA,$A7
	.dc.b	$FA,$A3,$63,$E1,$7C,$CF,$58,$FF,$57,$B6,$BB,$4E,$B1,$F9,$1E,$37
	.dc.b	$FB,$E2,$FF,$C0,$DF,$1F,$97,$C1,$F4,$FE,$32,$3F,$E3,$5F,$A7,$FF
	.dc.b	$1E,$3E,$5F,$2B,$C3,$F3,$F2,$30,$7F,$84,$7C,$3F,$AD,$F6,$BF,$E5
	.dc.b	$58,$FF,$1F,$35,$FF,$62,$E5,$FD,$DD,$FF,$21,$7D,$1A,$FE,$F3,$63
	.dc.b	$31,$F8,$1F,$A8,$7C,$D7,$E0,$FE,$89,$F3,$57,$D1,$98,$7F,$93,$67
	.dc.b	$2F,$F9,$3F,$E3,$38,$B7,$E2,$BF,$19,$1F,$EA,$E8,$D6,$1F,$8F,$F1
	.dc.b	$3F,$87,$F3,$7F,$78,$5F,$23,$93,$5E,$27,$8E,$3C,$9F,$3D,$30,$F1
	.dc.b	$47,$E6,$7E,$7F,$FA,$C7,$89,$FE,$7B,$6C,$BC,$24,$61,$BF,$21,$FA
	.dc.b	$1F,$A8,$FF,$69,$57,$F9,$A5,$AF,$FB,$C1,$BA,$7F,$83,$97,$C8,$FE
	.dc.b	$A7,$3F,$5E,$5F,$E8,$B7,$E9,$5F,$FB,$E1,$B7,$FE,$16,$77,$1E,$5F
	.dc.b	$A8,$FF,$0C,$3F,$EF,$E7,$F0,$AD,$F2,$BF,$6D,$1A,$AF,$F0,$FF,$D4
	.dc.b	$5C,$3F,$D3,$4F,$08,$CD,$7E,$F6,$35,$7F,$0F,$0F,$4B,$FE,$87,$5C
	.dc.b	$5E,$B3,$D9,$97,$F0,$5F,$A8,$7C,$C5,$25,$FF,$A3,$E1,$F3,$3F,$C3
	.dc.b	$A1,$D8,$F9,$DC,$2F,$9F,$A6,$22,$4F,$FA,$70,$E3,$1A,$71,$3C,$DB
	.dc.b	$39,$7F,$B1,$6F,$FE,$21,$9E,$D1,$A3,$FD,$9F,$F3,$BF,$C7,$FE,$48
	.dc.b	$D3,$1F,$11,$67,$64,$3F,$DB,$2F,$F2,$7F,$8B,$F9,$31,$E0,$3F,$1C
	.dc.b	$E3,$8F,$FC,$A6,$DC,$7F,$D9,$CF,$CE,$9F,$F5,$B3,$97,$8E,$FD,$4E
	.dc.b	$DC,$1F,$2F,$EA,$1E,$D5,$25,$7C,$7F,$A8,$7C,$7E,$23,$C8,$23,$F8
	.dc.b	$36,$7F,$C8,$1B,$9F,$F9,$1F,$0E,$B5,$F4,$5A,$F2,$9D,$BF,$DC,$1A
	.dc.b	$A3,$75,$FE,$72,$F9,$8F,$E5,$17,$FE,$ED,$5F,$35,$FD,$3D,$B7,$FE
	.dc.b	$D7,$76,$FE,$DD,$BD,$A7,$71,$F3,$E3,$13,$CF,$D0,$E9,$E4,$CA,$E8
	.dc.b	$FF,$5D,$FA,$2F,$E2,$B1,$C7,$0F,$E4,$4D,$9E,$0F,$59,$DE,$5F,$D1
	.dc.b	$BC,$77,$D4,$7E,$75,$FF,$71,$D7,$E5,$DA,$FD,$07,$6E,$63,$87,$C5
	.dc.b	$C8,$C1,$0E,$0E,$5E,$51,$E9,$FA,$89,$1F,$EF,$DC,$11,$ED,$F2,$B3
	.dc.b	$3E,$98,$B8,$FF,$A9,$2B,$F9,$94,$FE,$5F,$E8,$9E,$BF,$F9,$4A,$3F
	.dc.b	$C5,$75,$FE,$B7,$9C,$F8,$5F,$E0,$86,$CE,$B8,$BF,$C6,$31,$CB,$C9
	.dc.b	$D6,$1C,$EB,$B2,$BC,$9A,$9F,$E0,$DF,$51,$9A,$FE,$00,$BE,$73,$26
	.dc.b	$AA,$4E,$A3,$2F,$08,$CB,$FA,$09,$E4,$EC,$D1,$E9,$18,$7F,$44,$F5
	.dc.b	$7D,$DE,$67,$F8,$C3,$3F,$6B,$6B,$6A,$76,$70,$F7,$E4,$F5,$BE,$8F
	.dc.b	$D7,$F4,$BF,$25,$E9,$48,$C1,$0F,$C8,$39,$7F,$25,$1F,$99,$F0,$7F
	.dc.b	$35,$F0,$B2,$7F,$05,$19,$7F,$09,$19,$7F,$35,$1F,$A1,$F8,$A9,$1F
	.dc.b	$89,$7F,$49,$E7,$62,$3F,$41,$17,$8F,$F1,$13,$FA,$9E,$2F,$D5,$70
	.dc.b	$CB,$FC,$94,$65,$FC,$A4,$65,$FD,$54,$7E,$C7,$8B,$F6,$5F,$51,$27
	.dc.b	$F3,$51,$97,$F3,$91,$97,$F6,$51,$FB,$9F,$A8,$91,$FA,$17,$58,$FE
	.dc.b	$75,$FF,$2D,$7D,$1B,$EA,$69,$C2,$BF,$91,$FD,$4F,$F0,$AF,$8B,$FF
	.dc.b	$0C,$C2,$3F,$C8,$11,$86,$F0,$FF,$BA,$DE,$0D,$F0,$FF,$DC,$0E,$7C
	.dc.b	$3B,$7F,$E9,$4A,$E2,$DA,$CB,$DF,$E8,$BF,$C4,$F8,$77,$6E,$1D,$DB
	.dc.b	$3C,$F0,$E3,$82,$73,$3C,$FD,$BF,$F5,$1E,$1C,$DB,$87,$96,$69,$E5
	.dc.b	$AE,$37,$CE,$B3,$3C,$C5,$FC,$3C,$3C,$37,$C0,$95,$C2,$FD,$F7,$C6
	.dc.b	$FC,$4F,$2C,$3D,$95,$BC,$39,$BC,$70,$E7,$5B,$3E,$01,$98,$8E,$4E
	.dc.b	$D5,$ED,$71,$87,$AD,$B5,$D3,$0F,$54,$72,$7C,$A4,$AE,$4D,$8F,$8D
	.dc.b	$7B,$9C,$9E,$A8,$EB,$19,$7C,$E6,$09,$3C,$DF,$C7,$61,$F0,$1F,$F6
	.dc.b	$C9,$CC,$A7,$D1,$2E,$1E,$B6,$FD,$1F,$FA,$3D,$BF,$07,$C1,$FC,$84
	.dc.b	$65,$FC,$C3,$5C,$C7,$D7,$8F,$F4,$C5,$FE,$8B,$E0,$5F,$48,$EF,$D5
	.dc.b	$CC,$3E,$6E,$1C,$34,$F4,$E1,$C2,$F5,$8F,$EA,$6B,$57,$E0,$3E,$C5
	.dc.b	$6D,$E7,$87,$0C,$3C,$EF,$E4,$A7,$57,$CE,$35,$5C,$E3,$D1,$8F,$9E
	.dc.b	$39,$92,$38,$3E,$08,$D4,$60,$97,$E0,$9D,$63,$D9,$FE,$09,$8F,$A9
	.dc.b	$B8,$79,$70,$E7,$7C,$70,$E7,$5E,$58,$F8,$B3,$A8,$C1,$0E,$67,$80
	.dc.b	$F4,$47,$32,$21,$8F,$C4,$8C,$C7,$2A,$F0,$BA,$AF,$03,$99,$FA,$FE
	.dc.b	$76,$2F,$94,$F5,$7F,$C9,$B9,$7B,$4F,$A5,$3B,$BE,$0D,$8F,$E1,$7E
	.dc.b	$5B,$C4,$7C,$E7,$D3,$E6,$FA,$92,$E5,$FF,$14,$DF,$F8,$66,$B4,$FB
	.dc.b	$67,$5E,$46,$34,$FF,$5C,$63,$D2,$B8,$1F,$E2,$1F,$4F,$57,$E1,$C2
	.dc.b	$D3,$D3,$B5,$86,$DF,$F8,$A5,$B8,$7F,$8F,$DB,$9F,$F8,$9F,$E5,$B9
	.dc.b	$FA,$FD,$1B,$7F,$E3,$16,$DF,$F9,$1D,$B5,$FE,$2F,$F6,$3C,$67,$93
	.dc.b	$E4,$ED,$FF,$8E,$62,$DF,$F9,$3E,$6D,$7C,$8D,$FF,$01,$7F,$E1,$7D
	.dc.b	$BE,$57,$2F,$C0,$9C,$E1,$FE,$63,$51,$98,$D3,$6B,$94,$E9,$BF,$03
	.dc.b	$CF,$93,$A3,$63,$27,$08,$CB,$4F,$F0,$F7,$D4,$B5,$FB,$1E,$4A,$3F
	.dc.b	$D8,$13,$A9,$CC,$63,$2F,$F2,$93,$C6,$79,$EF,$AA,$DB,$F8,$3D,$5F
	.dc.b	$53,$FE,$B8,$AF,$F1,$D4,$63,$87,$53,$9C,$7F,$D6,$93,$9A,$DC,$E6
	.dc.b	$3F,$2C,$FE,$1E,$33,$C5,$CA,$F9,$B6,$32,$72,$8F,$7F,$CB,$FF,$B9
	.dc.b	$23,$FC,$80,$FF,$B8,$9B,$F0,$1A,$F7,$F6,$D8,$C9,$C2,$3E,$E7,$C2
	.dc.b	$C8,$FA,$FF,$2D,$74,$FB,$D3,$74,$FC,$34,$DA,$FF,$3C,$FD,$4B,$EA
	.dc.b	$7E,$43,$FE,$C4,$8F,$8E,$E5,$BE,$FE,$A7,$6D,$8C,$9C,$23,$2D,$33
	.dc.b	$8F,$E4,$7E,$53,$F2,$87,$E5,$7E,$26,$CF,$E4,$22,$CF,$E5,$E6,$E5
	.dc.b	$F9,$FF,$83,$FC,$E7,$C1,$FE,$75,$B8,$7E,$61,$B7,$F9,$F6,$D7,$F4
	.dc.b	$18,$7B,$BF,$EC,$EA,$FC,$2F,$D4,$FF,$1C,$9F,$F7,$1C,$E5,$FC,$2C
	.dc.b	$66,$B9,$4E,$5F,$E3,$51,$9A,$FE,$EF,$39,$7F,$D4,$F1,$9A,$FE,$6D
	.dc.b	$39,$6F,$9B,$F9,$4F,$57,$F8,$E4,$79,$A3,$F7,$5E,$34,$C3,$F6,$BE
	.dc.b	$3D,$AF,$DE,$4D,$CB,$F8,$24,$DA,$FF,$76,$70,$7F,$22,$F8,$2F,$BD
	.dc.b	$87,$FB,$77,$E9,$71,$32,$BF,$DD,$CE,$E3,$13,$FA,$6E,$37,$99,$F8
	.dc.b	$8F,$43,$F1,$2D,$C3,$F1,$2D,$AE,$0D,$CF,$FA,$05,$F2,$AF,$C7,$DF
	.dc.b	$26,$9F,$C6,$FE,$1B,$FD,$88,$DF,$5F,$F9,$DC,$7F,$B0,$63,$07,$F2
	.dc.b	$1F,$A9,$FF,$58,$37,$DD,$FE,$83,$18,$CD,$97,$F5,$08,$4F,$F4,$6F
	.dc.b	$A8,$F8,$B3,$FD,$51,$6B,$9D,$B4,$FF,$4C,$FA,$87,$CC,$5E,$CC,$37
	.dc.b	$4F,$25,$F8,$AF,$C6,$D8,$FE,$71,$F5,$0F,$7B,$FA,$BB,$63,$3E,$F6
	.dc.b	$27,$FA,$DF,$3B,$DF,$A6,$3E,$02,$DF,$F4,$3F,$A5,$7C,$47,$F6,$25
	.dc.b	$F5,$16,$30,$FF,$66,$FA,$8F,$ED,$5C,$DF,$1A,$FF,$B8,$47,$F2,$36
	.dc.b	$FB,$5B,$F3,$7F,$B6,$FD,$43,$96,$FC,$F6,$1B,$7E,$3F,$A7,$F1,$7D
	.dc.b	$47,$8D,$B1,$FD,$67,$E2,$BF,$C9,$8D,$3F,$BC,$AF,$9D,$B1,$FD,$F3
	.dc.b	$E6,$BD,$AF,$F6,$1E,$57,$9F,$9E,$2F,$F0,$0C,$62,$7F,$B3,$7D,$47
	.dc.b	$F3,$BE,$2F,$70,$F7,$DE,$AB,$E3,$7F,$C1,$35,$BF,$0B,$FC,$2F,$39
	.dc.b	$C3,$9F,$B5,$A3,$FC,$1F,$F5,$0E,$1C,$4F,$F7,$5A,$75,$FA,$7E,$A3
	.dc.b	$FC,$2B,$18,$9F,$EF,$5C,$29,$FF,$0C,$FA,$0F,$80,$DE,$18,$6D,$FF
	.dc.b	$80,$67,$4D,$8E,$3F,$01,$B8,$FF,$03,$7D,$43,$E1,$3F,$C4,$B8,$7F
	.dc.b	$81,$B7,$FF,$19,$CE,$6B,$FC,$19,$39,$7F,$C6,$51,$9E,$3F,$93,$C7
	.dc.b	$EC,$7F,$C5,$F1,$9B,$FF,$1E,$DF,$19,$F9,$D3,$9B,$FF,$1E,$4E,$6D
	.dc.b	$39,$AF,$A9,$39,$3E,$D3,$E0,$5E,$D7,$AB,$9F,$D6,$DC,$3D,$2D,$AE
	.dc.b	$0F,$F9,$13,$EA,$3F,$9A,$62,$FA,$AB,$18,$6F,$8D,$B3,$87,$AD,$71
	.dc.b	$FF,$38,$6B,$FD,$37,$2F,$F7,$68,$FA,$8F,$2B,$AD,$7F,$D2,$51,$88
	.dc.b	$FF,$5E,$FC,$67,$F9,$E6,$31,$5F,$83,$9E,$5A,$FF,$AE,$6B,$4F,$FB
	.dc.b	$12,$75,$5C,$EB,$C2,$F2,$BE,$F5,$C6,$79,$9E,$3C,$39,$BF,$EC,$CC
	.dc.b	$2B,$FC,$9A,$B8,$B9,$38,$99,$9D,$BE,$0A,$F2,$4F,$08,$F6,$9C,$30
	.dc.b	$F3,$CF,$8F,$0F,$35,$F2,$A6,$1E,$2B,$DC,$78,$6B,$6F,$88,$F0,$1E
	.dc.b	$B3,$07,$23,$C1,$87,$8D,$FA,$A7,$FB,$8C,$D1,$6B,$0E,$8F,$F7,$8F
	.dc.b	$D3,$3C,$B7,$D4,$F1,$8C,$36,$33,$3F,$EE,$F9,$CB,$7B,$FF,$A8,$B7
	.dc.b	$2E,$51,$88,$FF,$7E,$CF,$F4,$39,$C9,$F5,$34,$FE,$05,$37,$9B,$FD
	.dc.b	$80,$FD,$4A,$E0,$DB,$EF,$A7,$E3,$EF,$8C,$FD,$A6,$C7,$E4,$EB,$8B
	.dc.b	$6B,$8C,$F6,$36,$5B,$F0,$CE,$F0,$FC,$4D,$F2,$AD,$61,$E7,$33,$87
	.dc.b	$47,$FD,$B1,$3B,$7F,$DB,$53,$D9,$B8,$76,$8F,$A7,$5C,$65,$5C,$5F
	.dc.b	$E0,$75,$B7,$89,$E0,$B8,$FC,$18,$D3,$C0,$CE,$1F,$B8,$97,$C5,$FC
	.dc.b	$E9,$FF,$43,$B6,$3F,$DB,$4D,$DB,$F8,$1B,$FE,$AE,$9F,$F4,$8B,$97
	.dc.b	$FC,$22,$D9,$EB,$1F,$BB,$9E,$3F,$17,$63,$FD,$F9,$FA,$83,$E0,$71
	.dc.b	$08,$C3,$FC,$4A,$BF,$07,$18,$8F,$F4,$EF,$C8,$FF,$75,$74,$74,$AE
	.dc.b	$73,$BC,$3F,$CC,$CD,$BF,$E3,$B7,$C6,$FF,$D4,$FF,$91,$BB,$3C,$D8
	.dc.b	$23,$13,$FE,$B0,$6B,$C3,$AB,$71,$89,$61,$FC,$AA,$30,$DB,$DD,$70
	.dc.b	$78,$FC,$34,$C3,$B4,$7C,$4F,$82,$79,$F7,$C0,$95,$F0,$04,$E9,$C2
	.dc.b	$8D,$C7,$F8,$10,$DB,$D6,$7B,$7A,$92,$78,$BC,$34,$F5,$C3,$E3,$4B
	.dc.b	$EC,$BE,$3F,$CF,$3E,$A6,$9F,$7E,$B7,$3A,$F4,$25,$FF,$B6,$AD,$1C
	.dc.b	$23,$FD,$BB,$FA,$6B,$8F,$97,$DF,$BE,$BE,$9F,$5A,$75,$5F,$82,$C9
	.dc.b	$D0,$F1,$3C,$4D,$9D,$35,$FF,$13,$CF,$0F,$9A,$FC,$9F,$E8,$4F,$8F
	.dc.b	$A4,$AE,$DD,$CE,$8F,$77,$E3,$AF,$9D,$A4,$AF,$24,$7D,$57,$BF,$E1
	.dc.b	$8F,$97,$F8,$CD,$8F,$53,$5D,$FA,$1F,$2F,$F4,$27,$AF,$49,$3F,$06
	.dc.b	$3F,$62,$FB,$3F,$16,$79,$FF,$4D,$B1,$F1,$DA,$EF,$E7,$7E,$C7,$D4
	.dc.b	$63,$BB,$F0,$3C,$88,$C7,$8B,$F8,$28,$C6,$3C,$2F,$F7,$7F,$9D,$7E
	.dc.b	$BF,$1F,$C3,$B5,$39,$A7,$E3,$BF,$E5,$E9,$FB,$5F,$27,$F2,$DB,$C7
	.dc.b	$F8,$B8,$C4,$7D,$87,$E6,$69,$E6,$FC,$D6,$E9,$E5,$6D,$7E,$65,$B8
	.dc.b	$9F,$97,$1F,$89,$FC,$D3,$E8,$5F,$C0,$6D,$FB,$DB,$3F,$A3,$7F,$CA
	.dc.b	$0F,$E4,$7D,$2F,$F4,$9B,$7E,$2B,$F4,$51,$88,$FC,$7A,$F9,$DF,$E4
	.dc.b	$3F,$8E,$3E,$FA,$92,$BF,$7D,$F6,$AF,$03,$11,$FB,$BA,$E0,$FF,$03
	.dc.b	$6B,$FB,$F9,$E0,$E6,$B8,$E9,$FB,$C8,$57,$08,$C5,$70,$9F,$E1,$1C
	.dc.b	$BE,$77,$F8,$83,$5C,$E3,$FC,$EB,$EC,$39,$6D,$FD,$EA,$B5,$87,$76
	.dc.b	$C6,$6B,$86,$5F,$D0,$A7,$F6,$95,$BF,$0F,$07,$5F,$92,$B9,$78,$A2
	.dc.b	$BF,$C7,$72,$4F,$3F,$46,$4F,$A1,$FE,$9F,$E5,$7F,$B3,$E3,$04,$3D
	.dc.b	$F1,$AF,$3B,$8C,$7F,$29,$9F,$E5,$8F,$F3,$16,$BC,$0F,$E6,$9F,$39
	.dc.b	$F8,$C7,$F9,$83,$72,$FA,$0D,$DB,$FC,$3F,$56,$BF,$96,$CD,$4D,$31
	.dc.b	$17,$E6,$FF,$B4,$C1,$A7,$F6,$CF,$B0,$FD,$84,$66,$BC,$B3,$A7,$FA
	.dc.b	$54,$EB,$EA,$39,$92,$33,$F1,$35,$F5,$C5,$CB,$F1,$55,$FE,$C5,$9F
	.dc.b	$83,$39,$CB,$FC,$77,$8A,$3C,$98,$7F,$80,$3F,$1C,$7D,$D5,$26,$1F
	.dc.b	$D7,$7E,$5F,$FB,$B6,$7F,$97,$AD,$CF,$F6,$16,$BF,$D7,$5B,$1F,$9D
	.dc.b	$9E,$39,$7F,$93,$26,$D7,$F5,$A6,$A6,$FD,$83,$EC,$BC,$DF,$2B,$2B
	.dc.b	$C0,$FF,$86,$27,$51,$E3,$D7,$D8,$47,$FB,$B4,$7F,$18,$CB,$D5,$39
	.dc.b	$9F,$43,$CF,$C2,$C4,$EA,$7A,$1C,$A3,$FC,$4F,$FA,$9B,$EA,$29,$27
	.dc.b	$FC,$09,$F5,$1F,$D5,$5F,$F0,$34,$ED,$B3,$B7,$FC,$09,$3B,$7F,$C0
	.dc.b	$D3,$FC,$F9,$FE,$FE,$D4,$D8,$FF,$7E,$47,$6F,$27,$CA,$47,$C7,$F2
	.dc.b	$1F,$E0,$3F,$D8,$7F,$25,$1F,$DC,$1B,$97,$F4,$E6,$F0,$7F,$4D,$5F
	.dc.b	$65,$C7,$9E,$5F,$F7,$F9,$DB,$FE,$15,$76,$FF,$80,$FF,$1E,$79,$BF
	.dc.b	$E1,$36,$C6,$63,$1F,$33,$96,$46,$4E,$51,$FE,$1B,$FC,$3D,$F4,$D4
	.dc.b	$93,$FE,$11,$87,$CF,$62,$3F,$C1,$B1,$C1,$D4,$F3,$9F,$F1,$37,$E8
	.dc.b	$5F,$E2,$99,$FF,$0D,$35,$FF,$0A,$B7,$6F,$EE,$6F,$38,$CF,$C2,$F2
	.dc.b	$23,$CA,$FA,$F8,$1F,$E3,$3F,$C7,$BF,$C5,$53,$FE,$27,$6B,$DA,$36
	.dc.b	$E7,$E0,$B8,$18,$7F,$C9,$5F,$69,$FE,$50,$7F,$C8,$7F,$2B,$FB,$D7
	.dc.b	$FC,$8B,$1F,$E3,$47,$FD,$C3,$E1,$FF,$AF,$23,$13,$B9,$47,$37,$67
	.dc.b	$2F,$8B,$FF,$61,$CE,$E7,$8D,$73,$F8,$7C,$BD,$6A,$DF,$8B,$D7,$7D
	.dc.b	$AF,$9A,$BC,$51,$89,$EE,$78,$9C,$63,$DF,$0E,$73,$9B,$E4,$F5,$6B
	.dc.b	$93,$9E,$3E,$67,$46,$2F,$AE,$1E,$2A,$E3,$5B,$AE,$58,$7B,$EF,$71
	.dc.b	$CA,$3C,$8F,$08,$CB,$C2,$FA,$3E,$2C,$3D,$36,$AD,$B7,$0E,$47,$8A
	.dc.b	$77,$3A,$24,$66,$B6,$7A,$DB,$87,$92,$30,$47,$D5,$1C,$63,$18,$79
	.dc.b	$A7,$31,$E7,$39,$C6,$48,$FE,$32,$FC,$B5,$A8,$F8,$93,$F8,$37,$CA
	.dc.b	$7F,$B7,$70,$F7,$3F,$ED,$F3,$EB,$19,$C3,$F0,$4E,$FC,$0F,$F6,$F3
	.dc.b	$7E,$1B,$83,$F9,$68,$CB,$F9,$B6,$BF,$A0,$7F,$DC,$D9,$7E,$5D,$B8
	.dc.b	$7F,$AA,$27,$FD,$9D,$DB,$FE,$AC,$AD,$5F,$E0,$AF,$75,$CE,$3F,$1B
	.dc.b	$E5,$FF,$9E,$BB,$75,$12,$7F,$1F,$5C,$1E,$B3,$9B,$EB,$1A,$AE,$78
	.dc.b	$78,$71,$F5,$46,$DB,$3F,$A4,$C7,$C0,$F1,$25,$F4,$3F,$75,$1A,$71
	.dc.b	$3C,$9B,$3F,$EC,$89,$D4,$E4,$91,$EF,$3C,$17,$E5,$71,$5C,$9F,$A5
	.dc.b	$7E,$AC,$3D,$11,$AB,$E6,$FC,$23,$CE,$E3,$3F,$F1,$5D,$71,$9C,$9C
	.dc.b	$48,$71,$BF,$06,$3F,$EC,$BC,$38,$47,$C0,$F3,$BC,$D5,$BC,$7E,$F5
	.dc.b	$EE,$3D,$77,$F6,$AB,$73,$F0,$A7,$FD,$D9,$F5,$3F,$EB,$CA,$FF,$6D
	.dc.b	$7E,$3C,$F9,$9F,$F7,$8F,$4F,$FA,$CA,$B5,$7F,$ED,$8A,$D5,$FF,$B8
	.dc.b	$67,$35,$FE,$B8,$9C,$D7,$FB,$7A,$73,$5F,$EB,$99,$CD,$75,$F8,$EE
	.dc.b	$F3,$FE,$BD,$A4,$62,$FC,$FF,$65,$7D,$E5,$6A,$FC,$57,$98,$C5,$F8
	.dc.b	$6F,$13,$C7,$F1,$C7,$DC,$53,$1E,$35,$C7,$FD,$F9,$3A,$6C,$E9,$A7
	.dc.b	$FC,$01,$FC,$77,$FC,$96,$DF,$4B,$CD,$0E,$AB,$12,$74,$71,$8C,$66
	.dc.b	$DB,$FD,$C3,$EE,$BC,$AF,$80,$DB,$56,$DF,$F7,$24,$7F,$AF,$5C,$D7
	.dc.b	$2A,$E2,$48,$C9,$C3,$DB,$C5,$F3,$3A,$FE,$8B,$E9,$9D,$C6,$31,$FA
	.dc.b	$5F,$7E,$F8,$DF,$AD,$2B,$E0,$F3,$31,$5A,$8C,$E5,$FE,$FE,$BE,$58
	.dc.b	$76,$BD,$B6,$BF,$C3,$AD,$F2,$3F,$09,$6F,$B8,$FC,$6D,$FA,$27,$B1
	.dc.b	$2B,$D3,$5B,$F8,$AC,$57,$97,$0F,$EA,$D5,$DF,$C2,$FC,$77,$E1,$1F
	.dc.b	$EE,$58,$C1,$2B,$84,$EB,$E0,$7D,$8C,$9D,$62,$7E,$BF,$53,$0E,$98
	.dc.b	$72,$9F,$81,$F8,$F3,$E5,$FE,$5B,$63,$D3,$F6,$FF,$8E,$AF,$99,$39
	.dc.b	$6D,$70,$3E,$2C,$EF,$1F,$F7,$85,$6D,$BC,$BD,$8F,$D6,$7D,$4F,$E3
	.dc.b	$70,$EB,$19,$C7,$CA,$EB,$9B,$13,$E2,$F9,$BF,$39,$F5,$38,$FF,$CD
	.dc.b	$37,$EC,$8C,$B4,$FA,$71,$A9,$E5,$19,$69,$F9,$3C,$7E,$F7,$67,$28
	.dc.b	$FC,$67,$C9,$3E,$7F,$F2,$2D,$BF,$CB,$F0,$BF,$48,$7C,$E6,$AE,$57
	.dc.b	$E3,$EF,$86,$1F,$A1,$C3,$85,$EA,$3B,$E3,$D8,$F9,$BD,$17,$0E,$51
	.dc.b	$67,$D1,$C5,$F9,$87,$8C,$F7,$F8,$7F,$E3,$5F,$A7,$BD,$EF,$4C,$61
	.dc.b	$B3,$B3,$D7,$F4,$3F,$E8,$C7,$17,$F0,$AB,$6F,$A6,$74,$E6,$7F,$59
	.dc.b	$F7,$6F,$53,$F6,$ED,$8F,$A7,$F4,$18,$99,$1C,$1F,$B9,$18,$AF,$DE
	.dc.b	$53,$ED,$F9,$9F,$CD,$7E,$7B,$FA,$97,$2F,$E8,$6B,$F8,$BD,$6F,$1F
	.dc.b	$1D,$F0,$7E,$BD,$F0,$7F,$C3,$13,$BA,$EF,$F8,$4F,$A0,$FC,$E9,$67
	.dc.b	$DA,$D8,$FA,$2D,$C3,$C5,$0A,$4E,$6B,$F9,$3E,$F8,$79,$9F,$8A,$FA
	.dc.b	$AF,$E2,$EF,$EE,$9B,$6E,$44,$7C,$B8,$FC,$53,$A8,$7D,$8F,$E4,$AB
	.dc.b	$F2,$AD,$AF,$C8,$7E,$6B,$FE,$88,$9C,$F3,$7F,$5F,$E2,$7D,$9D,$31
	.dc.b	$E2,$5A,$FC,$FF,$E3,$AF,$73,$FB,$33,$63,$3E,$36,$27,$F9,$BF,$DD
	.dc.b	$7E,$83,$8F,$E2,$B6,$BF,$8B,$EC,$F9,$5D,$F7,$FF,$71,$FC,$75,$F2
	.dc.b	$54,$C7,$3D,$C3,$F6,$3E,$E3,$E0,$BF,$BC,$36,$33,$E9,$E2,$7F,$BE
	.dc.b	$7E,$5D,$FC,$5E,$7F,$93,$B7,$9F,$E7,$3F,$2C,$C7,$47,$F4,$FE,$4F
	.dc.b	$F0,$27,$9F,$FC,$B3,$8B,$F2,$13,$FD,$4F,$8F,$35,$FE,$09,$FC,$73
	.dc.b	$E5,$E5,$FC,$19,$B9,$7F,$01,$79,$47,$F3,$87,$38,$7E,$2B,$F4,$CF
	.dc.b	$99,$5F,$D7,$5B,$1E,$0C,$FF,$80,$3B,$C7,$FC,$31,$F8,$E7,$E1,$63
	.dc.b	$CA,$DC,$7F,$A2,$BF,$83,$8C,$F8,$7F,$D7,$BE,$EF,$FC,$4F,$5F,$D2
	.dc.b	$87,$C9,$67,$E0,$7F,$C4,$91,$83,$FA,$FF,$E9,$DF,$D9,$30,$34,$7C
	.dc.b	$07,$F6,$A7,$FB,$4F,$0B,$ED,$BE,$B7,$07,$F3,$19,$DE,$1F,$CB,$27
	.dc.b	$38,$7F,$91,$A7,$35,$FD,$FF,$F1,$DF,$EF,$6D,$F7,$7F,$C0,$BF,$1B
	.dc.b	$F1,$31,$FF,$4A,$56,$B0,$FD,$AC,$E6,$FF,$C3,$AD,$4B,$E3,$B9,$FA
	.dc.b	$5D,$1B,$87,$F8,$26,$74,$D9,$FF,$03,$FE,$5F,$F3,$63,$FC,$4B,$F3
	.dc.b	$DF,$9A,$7F,$C0,$B1,$9C,$3F,$C6,$53,$97,$FA,$94,$67,$13,$67,$F7
	.dc.b	$13,$AF,$27,$B9,$2F,$FC,$23,$3A,$BF,$F3,$6D,$F7,$B7,$CD,$53,$B7
	.dc.b	$B4,$F4,$BF,$C5,$B1,$9C,$A9,$FE,$22,$9D,$65,$FE,$29,$9F,$37,$E3
	.dc.b	$9F,$9A,$C3,$FC,$98,$DC,$8F,$E6,$2B,$FC,$8E,$E4,$9D,$DF,$E2,$E7
	.dc.b	$55,$CA,$3F,$C7,$EF,$F9,$4B,$E3,$7F,$CA,$7F,$60,$FA,$6F,$F6,$86
	.dc.b	$3F,$E8,$28,$C5,$7F,$B4,$C5,$70,$AD,$F4,$FF,$47,$AD,$D7,$FB,$66
	.dc.b	$B4,$FF,$97,$27,$55,$FE,$ED,$9C,$CF,$F9,$A6,$72,$F7,$8C,$DF,$8E
	.dc.b	$B8,$4A,$FB,$CE,$63,$81,$99,$F2,$E5,$DA,$3E,$01,$93,$C8,$E5,$F6
	.dc.b	$99,$3C,$8E,$6F,$C7,$87,$9A,$7D,$C6,$EF,$C4,$E3,$0F,$2C,$EE,$3C
	.dc.b	$31,$8B,$E9,$1C,$67,$4F,$88,$FA,$07,$A0,$A6,$4F,$86,$7D,$5F,$B7
	.dc.b	$7D,$CF,$F0,$38,$E3,$18,$E9,$E2,$D8,$CC,$ED,$E5,$F7,$3F,$E9,$B7
	.dc.b	$8B,$5C,$E5,$CA,$31,$1F,$E9,$C1,$48,$E5,$97,$E1,$34,$7D,$CE,$E3
	.dc.b	$0D,$8E,$55,$C7,$0F,$C2,$B7,$0F,$A9,$3C,$9F,$C4,$B5,$FB,$11,$BB
	.dc.b	$FC,$9B,$C7,$CB,$EF,$3F,$EB,$26,$BF,$D9,$2B,$C3,$8F,$C1,$9E,$2F
	.dc.b	$AA,$3D,$C7,$7B,$E2,$6B,$0F,$3C,$EE,$FF,$D7,$F5,$B3,$8E,$3E,$F8
	.dc.b	$FF,$61,$E3,$E7,$9F,$0D,$FB,$2B,$55,$FB,$56,$D6,$E3,$8D,$F7,$74
	.dc.b	$71,$3C,$75,$FB,$23,$FC,$E0,$FE,$FB,$5F,$F6,$7F,$9B,$FB,$D9,$FF
	.dc.b	$20,$1A,$9F,$F5,$8C,$3E,$53,$84,$7F,$B7,$E3,$F2,$6D,$3F,$2B,$3C
	.dc.b	$5E,$91,$A5,$F7,$2F,$77,$8B,$63,$F3,$AD,$7F,$42,$D7,$F8,$93,$9A
	.dc.b	$F3,$FC,$67,$F4,$D9,$FF,$63,$B5,$FF,$6D,$4E,$F1,$FF,$76,$FD,$CB
	.dc.b	$E5,$7D,$78,$23,$17,$FE,$D0,$6B,$FE,$9B,$5F,$0F,$C1,$CC,$7F,$28
	.dc.b	$6B,$DA,$BE,$1C,$62,$15,$DE,$79,$78,$72,$7C,$11,$F3,$9F,$F4,$07
	.dc.b	$C9,$DF,$2B,$FA,$64,$7F,$BB,$DA,$FF,$BD,$D4,$6F,$BF,$EA,$C6,$A7
	.dc.b	$F9,$D3,$5F,$DC,$35,$1F,$E8,$FF,$73,$C0,$47,$C1,$3F,$82,$F2,$37
	.dc.b	$F1,$9F,$2B,$C2,$7C,$DD,$35,$7F,$EE,$53,$FD,$F7,$0F,$9B,$FC,$3B
	.dc.b	$63,$F8,$FE,$DF,$EC,$09,$C9,$FC,$83,$07,$DC,$FF,$95,$6B,$FD,$8E
	.dc.b	$D7,$FD,$84,$DA,$FC,$DF,$DC,$B9,$7D,$1D,$0F,$9A,$E8,$DC,$3F,$97
	.dc.b	$E3,$F9,$F9,$C9,$F9,$EF,$B9,$7C,$45,$25,$7F,$34,$8C,$BC,$BE,$E7
	.dc.b	$FC,$0A,$DF,$27,$D5,$18,$8F,$ED,$98,$23,$11,$FC,$AE,$1F,$1C,$FB
	.dc.b	$9F,$84,$D9,$F3,$B5,$E6,$D7,$97,$DC,$3E,$5F,$E5,$36,$3D,$ED,$7C
	.dc.b	$6D,$7C,$5F,$70,$FA,$1F,$A4,$D8,$F8,$ED,$7D,$6D,$7D,$5F,$63,$5F
	.dc.b	$C6,$E4,$8C,$47,$FA,$CD,$B1,$9A,$FE,$1B,$F7,$1F,$C0,$6B,$FA,$73
	.dc.b	$5F,$CB,$B5,$39,$97,$C5,$FC,$C9,$F0,$B5,$FF,$61,$36,$37,$B7,$E3
	.dc.b	$FE,$DF,$F5,$75,$F8,$46,$BF,$61,$AF,$D7,$6B,$EE,$FB,$7F,$A4,$DD
	.dc.b	$FF,$35,$18,$8F,$F6,$93,$6F,$FC,$65,$F6,$FF,$9F,$F3,$4D,$25,$7F
	.dc.b	$80,$BE,$DD,$EA,$FF,$07,$6C,$7E,$EB,$AF,$FD,$D3,$18,$8F,$E2,$3E
	.dc.b	$4F,$F7,$7F,$B7,$7C,$EF,$1C,$3F,$9D,$DE,$EF,$FC,$80,$D7,$FB,$F3
	.dc.b	$5F,$E0,$EF,$5C,$3F,$DE,$F6,$FA,$CF,$E3,$4D,$AF,$26,$3C,$B0,$FE
	.dc.b	$3D,$C8,$FA,$3A,$4A,$D9,$FC,$53,$ED,$FB,$5F,$F8,$3E,$B8,$4E,$84
	.dc.b	$FE,$01,$C3,$5F,$82,$7F,$1E,$8E,$2E,$0F,$CA,$77,$79,$08,$F2,$F9
	.dc.b	$2A,$EE,$BF,$87,$5F,$F8,$85,$B1,$D6,$BD,$BE,$64,$8F,$36,$1F,$D3
	.dc.b	$35,$61,$FE,$29,$F1,$5E,$56,$EF,$FC,$83,$7F,$D2,$5F,$D4,$61,$FD
	.dc.b	$C2,$30,$73,$BE,$5E,$57,$F4,$9C,$3F,$A2,$E3,$DB,$0F,$81,$87,$F4
	.dc.b	$DF,$89,$FE,$01,$39,$94,$63,$0F,$F1,$E3,$63,$FA,$2D,$D7,$7E,$07
	.dc.b	$7F,$4E,$47,$D3,$AD,$1F,$EE,$07,$1F,$0F,$E1,$3F,$AF,$7D,$BF,$F7
	.dc.b	$58,$FE,$A3,$3B,$7E,$A4,$7F,$35,$6E,$FC,$5E,$DF,$23,$EB,$23,$F8
	.dc.b	$39,$7C,$1E,$23,$FC,$D7,$6A,$FF,$0C,$B6,$3F,$CD,$AD,$77,$1F,$DC
	.dc.b	$FE,$DF,$E3,$6D,$F1,$DB,$B7,$C6,$9F,$EF,$1F,$6E,$F9,$3A,$4C,$3F
	.dc.b	$2D,$6F,$B7,$C4,$7E,$46,$1F,$3F,$DC,$CE,$5F,$57,$0E,$15,$F9,$F6
	.dc.b	$BF,$99,$6C,$7D,$BF,$B7,$7C,$17,$93,$1F,$F0,$64,$7B,$67,$59,$E7
	.dc.b	$0E,$9A,$FA,$88,$FE,$B6,$BF,$AC,$57,$8E,$7F,$C8,$38,$FE,$22,$BF
	.dc.b	$D2,$AD,$7F,$D3,$2D,$7F,$D9,$4E,$67,$CD,$EC,$F9,$0E,$F3,$FB,$3F
	.dc.b	$B6,$7D,$BF,$F8,$8D,$B1,$FB,$F6,$BF,$B9,$FA,$37,$BB,$BA,$FF,$15
	.dc.b	$47,$FB,$1B,$8B,$FD,$2C,$DA,$FF,$0F,$DF,$2C,$7F,$44,$FF,$5B,$6F
	.dc.b	$9F,$E1,$39,$FC,$5D,$AF,$E2,$17,$C7,$0E,$B8,$E6,$7F,$87,$F9,$4F
	.dc.b	$B7,$FC,$E4,$7F,$7D,$6B,$FD,$91,$AA,$38,$75,$7F,$88,$67,$F4,$99
	.dc.b	$FE,$EE,$7A,$7D,$77,$F8,$CE,$4D,$AF,$F1,$B7,$DB,$BE,$DE,$93,$1F
	.dc.b	$E5,$B8,$27,$F9,$B4,$7A,$1E,$0F,$F8,$91,$AF,$F8,$61,$AF,$F8,$5D
	.dc.b	$AF,$F7,$EF,$B7,$FF,$39,$3F,$E1,$B6,$B9,$F9,$4F,$F1,$E4,$62,$3E
	.dc.b	$03,$FE,$79,$CF,$DC,$FE,$73,$5E,$9D,$1F,$E5,$18,$D4,$7F,$83,$3E
	.dc.b	$DF,$FC,$A3,$7F,$E3,$A6,$BF,$E2,$E6,$BF,$E2,$D6,$BC,$7E,$0F,$FC
	.dc.b	$B9,$1F,$E5,$38,$FB,$2D,$C3,$E8,$3D,$1B,$E2,$7F,$98,$FE,$57,$FA
	.dc.b	$0E,$A6,$D7,$1C,$7F,$C4,$31,$A2,$7C,$FF,$82,$FF,$D0,$1F,$6C,$F8
	.dc.b	$8F,$F7,$67,$CA,$7F,$13,$8F,$F2,$D6,$9F,$E7,$68,$C7,$C2,$EB,$C1
	.dc.b	$FF,$79,$57,$1A,$D4,$62,$75,$1F,$E7,$E9,$E6,$AB,$9B,$C0,$D5,$F0
	.dc.b	$53,$FE,$2C,$FB,$7D,$46,$3D,$E4,$62,$75,$3D,$71,$F1,$39,$AD,$B9
	.dc.b	$3C,$8E,$30,$F8,$0D,$9E,$33,$B3,$C1,$7C,$C8,$E6,$FD,$B7,$CA,$72
	.dc.b	$D9,$D6,$3E,$DA,$F3,$CE,$49,$7E,$06,$D7,$9A,$F9,$CE,$4F,$A3,$87
	.dc.b	$9E,$74,$78,$DC,$3F,$2B,$0F,$04,$63,$0F,$0D,$72,$BF,$95,$5A,$7B
	.dc.b	$DF,$4C,$3E,$BB,$D6,$32,$FC,$43,$DA,$62,$FA,$BC,$6F,$D0,$D2,$1C
	.dc.b	$6B,$A9,$F1,$DF,$F6,$29,$EC,$29,$DC,$B3,$F8,$06,$F8,$4F,$86,$DC
	.dc.b	$7E,$46,$30,$E2,$7F,$C2,$33,$A8,$FC,$7C,$E9,$E7,$1F,$8D,$8E,$35
	.dc.b	$FD,$DF,$CE,$FF,$4B,$70,$66,$A5,$FE,$0E,$F7,$8F,$AA,$F7,$39,$AD
	.dc.b	$E1,$FA,$FB,$DC,$FF,$B9,$2B,$73,$9A,$DD,$FE,$C2,$F7,$6B,$DC,$E6
	.dc.b	$B7,$3F,$05,$FD,$4C,$EA,$30,$78,$C9,$7E,$DF,$03,$FC,$74,$F8,$63
	.dc.b	$37,$D5,$FD,$25,$73,$7E,$4D,$F8,$1A,$6A,$F9,$D7,$1B,$E8,$47,$E3
	.dc.b	$46,$A7,$CB,$39,$6C,$EB,$1E,$D9,$7D,$A9,$F2,$9F,$26,$F9,$3E,$2B
	.dc.b	$E4,$DB,$E4,$F0,$3E,$54,$F0,$C7,$D8,$7F,$BE,$70,$F9,$31,$9F,$33
	.dc.b	$4D,$8C,$9F,$C6,$F9,$1F,$6E,$F9,$6F,$F3,$27,$3F,$F6,$B8,$C5,$7F
	.dc.b	$64,$FB,$7F,$F7,$43,$FD,$9A,$33,$9F,$F7,$C9,$CB,$FD,$FA,$33,$5F
	.dc.b	$DD,$67,$2F,$5B,$EF,$7F,$EE,$29,$78,$F8,$C3,$FC,$01,$1C,$27,$53
	.dc.b	$9C,$53,$9A,$DC,$FF,$9E,$5D,$D7,$EB,$AB,$73,$99,$DC,$E7,$ED,$FF
	.dc.b	$C0,$EF,$F8,$09,$7C,$1F,$77,$33,$FE,$1A,$73,$3B,$F0,$FA,$60,$9D
	.dc.b	$D7,$F0,$68,$C4,$EA,$73,$23,$FC,$3D,$F6,$FF,$E6,$7B,$F4,$36,$3F
	.dc.b	$86,$DA,$31,$1D,$5A,$F3,$8D,$7A,$1F,$1E,$79,$6B,$FE,$27,$5D,$19
	.dc.b	$9F,$99,$EF,$BC,$AB,$E4,$E1,$B9,$FF,$89,$A7,$4F,$CD,$6E,$7E,$66
	.dc.b	$BF,$05,$AF,$F2,$8F,$A5,$F0,$CF,$F8,$C3,$F3,$A7,$81,$F8,$1A,$56
	.dc.b	$9F,$F2,$4C,$EA,$BF,$DB,$D3,$9D,$7F,$D7,$0D,$8F,$85,$F7,$1F,$E6
	.dc.b	$27,$7A,$F5,$97,$CF,$FE,$3E,$DF,$84,$B4,$D4,$E6,$31,$87,$82,$32
	.dc.b	$FF,$BD,$29,$F6,$DF,$93,$D3,$D3,$5D,$FC,$3F,$CC,$7D,$CF,$FB,$33
	.dc.b	$0F,$4C,$63,$C3,$D4,$7F,$3F,$D1,$3F,$97,$C3,$E1,$B6,$33,$3B,$8E
	.dc.b	$33,$FE,$6E,$B7,$C1,$E3,$3F,$F6,$E7,$DC,$BE,$5F,$E0,$7C,$33,$F0
	.dc.b	$3F,$A1,$39,$CF,$D9,$39,$7F,$2D,$7C,$AF,$67,$C9,$FB,$5F,$C6,$CF
	.dc.b	$F6,$4B,$E5,$9A,$FC,$3C,$5F,$B9,$FC,$53,$EA,$4E,$6F,$84,$E7,$DA
	.dc.b	$F1,$B2,$3F,$CE,$BF,$77,$F8,$EC,$39,$C7,$90,$F0,$7C,$EF,$D7,$F1
	.dc.b	$FD,$6D,$8D,$63,$FA,$FC,$D3,$F3,$5B,$3F,$DB,$DB,$3F,$DD,$9A,$FF
	.dc.b	$7E,$73,$7F,$32,$BF,$15,$CC,$FB,$AF,$0E,$1E,$0C,$1F,$2E,$F8,$6F
	.dc.b	$E3,$5F,$4B,$FC,$6A,$B5,$39,$7F,$D2,$71,$9A,$F8,$C7,$F0,$79,$D3
	.dc.b	$83,$AD,$7F,$25,$FB,$B7,$CF,$FE,$AE,$B7,$A3,$1F,$F0,$CC,$79,$DF
	.dc.b	$CD,$71,$FF,$BC,$1F,$1E,$7F,$77,$E1,$B9,$32,$37,$DB,$74,$F1,$60
	.dc.b	$F8,$DF,$E6,$DF,$76,$F8,$4F,$14,$61,$B8,$7F,$87,$6B,$83,$67,$F4
	.dc.b	$11,$8B,$FF,$13,$B7,$4F,$54,$7C,$39,$FE,$89,$F4,$7F,$EB,$DB,$F9
	.dc.b	$71,$FC,$2F,$EC,$FF,$D1,$51,$FB,$7F,$0F,$FB,$45,$FC,$99,$CB,$F3
	.dc.b	$B0,$FE,$07,$EA,$71,$F6,$E6,$1F,$BF,$E6,$7C,$7E,$30,$E5,$C5,$FD
	.dc.b	$57,$E8,$3F,$C5,$79,$79,$A5,$9F,$CD,$7F,$DC,$73,$AC,$7F,$C9,$73
	.dc.b	$9A,$E0,$D8,$FE,$79,$3E,$89,$CD,$FB,$7E,$EF,$F6,$5F,$8D,$5B,$FB
	.dc.b	$6D,$C7,$FC,$9F,$8A,$BF,$86,$36,$BF,$C9,$CD,$9D,$CF,$1C,$3F,$C0
	.dc.b	$BC,$D9,$7C,$B1,$FD,$CF,$EF,$3F,$9B,$57,$F5,$F6,$BF,$D4,$DA,$FF
	.dc.b	$BE,$DB,$E9,$BE,$E3,$F2,$F9,$7F,$77,$6D,$7F,$60,$6B,$FD,$AD,$AF
	.dc.b	$F5,$FE,$37,$DD,$FF,$72,$6D,$7F,$80,$5B,$3F,$DA,$9A,$FF,$75,$FB
	.dc.b	$BF,$E6,$8F,$F6,$E6,$B9,$F3,$BF,$BE,$C6,$23,$FD,$03,$F7,$7F,$B2
	.dc.b	$7F,$D0,$2D,$73,$E0,$7F,$39,$8C,$43,$37,$DD,$BE,$53,$FC,$12,$DC
	.dc.b	$FF,$C3,$0D,$AF,$F0,$0B,$5F,$F0,$67,$C1,$3E,$1F,$FC,$22,$DA,$FF
	.dc.b	$10,$36,$7F,$C1,$2D,$7F,$C2,$9F,$73,$FE,$08,$7C,$8D,$C4,$D3,$1E
	.dc.b	$A7,$F4,$CB,$FF,$16,$36,$BF,$C2,$AD,$7F,$C4,$0D,$7F,$C2,$9F,$77
	.dc.b	$F4,$1B,$C3,$FE,$28,$8C,$47,$C5,$6A,$FB,$BF,$A0,$FF,$89,$DA,$E7
	.dc.b	$87,$FB,$34,$6B,$27,$DD,$FE,$F2,$BF,$C6,$4D,$7F,$C4,$CD,$7E,$A3
	.dc.b	$63,$F9,$37,$CA,$7F,$93,$F3,$3F,$E0,$7F,$BB,$FF,$2E,$C7,$F8,$BC
	.dc.b	$C9,$FE,$03,$FB,$AF,$F3,$1E,$1F,$EC,$6E,$A4,$62,$3F,$B5,$B5,$FE
	.dc.b	$CC,$DC,$1F,$6F,$F6,$6B,$FC,$EB,$2F,$BB,$FF,$31,$4E,$70,$E7,$1F
	.dc.b	$E2,$9F,$BB,$FE,$DB,$7F,$C4,$5B,$3F,$ED,$8F,$92,$FF,$10,$BF,$D5
	.dc.b	$63,$39,$7F,$94,$E7,$2F,$FB,$59,$AF,$FB,$79,$AF,$3F,$37,$FC,$2D
	.dc.b	$0F,$97,$FF,$1D,$3C,$B0,$FE,$44,$E7,$0F,$EC,$8E,$5F,$F0,$6F,$DD
	.dc.b	$FF,$00,$9F,$F4,$9F,$D0,$7E,$85,$FF,$4A,$C6,$67,$FC,$27,$39,$78
	.dc.b	$FA,$5F,$E7,$DA,$57,$F0,$6C,$BF,$C2,$4D,$8C,$CE,$E3,$A4,$EA,$3F
	.dc.b	$C8,$96,$F9,$5F,$F2,$35,$6A,$BF,$D0,$B5,$FE,$49,$F8,$77,$DB,$FF
	.dc.b	$96,$7E,$3B,$9E,$5F,$E6,$B8,$C5,$7F,$8E,$E7,$C9,$7C,$B3,$FF,$7F
	.dc.b	$47,$97,$EE,$FF,$51,$3F,$E2,$96,$E0,$69,$3E,$13,$FC,$B4,$FE,$CE
	.dc.b	$3F,$D7,$6F,$F1,$C6,$C7,$F9,$B6,$7F,$D2,$1F,$2B,$FE,$91,$FB,$B7
	.dc.b	$D6,$FE,$5F,$E5,$7F,$B6,$C6,$2B,$FD,$2C,$70,$0E,$5F,$F1,$F4,$70
	.dc.b	$9D,$C6,$E7,$58,$7F,$B2,$1E,$11,$99,$F0,$67,$E3,$78,$C3,$0C,$3F
	.dc.b	$E5,$2C,$3B,$3D,$A3,$2F,$8A,$AB,$D4,$E0,$E1,$E8,$7A,$AB,$C2,$D3
	.dc.b	$A1,$FC,$D5,$CF,$8D,$E8,$8E,$2F,$B2,$FF,$C2,$6E,$6F,$C7,$8F,$BE
	.dc.b	$FD,$71,$8A,$F4,$DF,$CC,$8C,$61,$E5,$6C,$78,$CD,$1F,$10,$FA,$45
	.dc.b	$70,$E0,$FA,$FF,$74,$F7,$FF,$85,$C7,$CC,$8C,$75,$73,$6C,$66,$76
	.dc.b	$FD,$A8,$FC,$EE,$5B,$E1,$FF,$3E,$4F,$36,$BB,$6B,$E2,$77,$AF,$E1
	.dc.b	$F4,$7D,$DF,$E3,$23,$0D,$8E,$17,$F7,$30,$F0,$E1,$F0,$E7,$8B,$F8
	.dc.b	$06,$BF,$1E,$37,$7F,$95,$6C,$FE,$68,$DC,$78,$27,$B6,$1E,$69,$F0
	.dc.b	$C7,$7C,$3C,$8F,$FB,$DE,$38,$9A,$AE,$D5,$C6,$75,$5C,$1F,$D6,$B6
	.dc.b	$FC,$71,$FF,$00,$5A,$E8,$F6,$46,$E7,$A4,$27,$77,$E2,$8E,$E7,$AB
	.dc.b	$0F,$D3,$9F,$E8,$F7,$FA,$96,$9F,$EB,$AF,$61,$E1,$E2,$7F,$D6,$D1
	.dc.b	$87,$FD,$B9,$E6,$7E,$EA,$FF,$CD,$E7,$38,$FF,$70,$4E,$DF,$F3,$43
	.dc.b	$63,$73,$C6,$FF,$84,$FD,$DB,$DC,$F0,$53,$E3,$BF,$A4,$63,$FE,$EB
	.dc.b	$6B,$FE,$9D,$AE,$0F,$FA,$AD,$AF,$F9,$DE,$37,$3E,$66,$DF,$82,$FB
	.dc.b	$FD,$DB,$E6,$E9,$30,$FF,$7D,$47,$FB,$11,$B1,$C3,$E2,$A4,$60,$87
	.dc.b	$FA,$1E,$51,$A8,$C4,$FC,$D5,$A3,$E6,$70,$FF,$A2,$A3,$FD,$8C,$D7
	.dc.b	$FD,$78,$ED,$CF,$BD,$C3,$D6,$91,$83,$8B,$87,$8B,$DA,$3E,$7F,$73
	.dc.b	$E7,$E9,$2B,$FD,$CB,$E5,$17,$C7,$70,$F9,$5F,$AF,$F7,$5F,$5E,$7F
	.dc.b	$DD,$7C,$68,$C4,$7F,$AA,$71,$57,$E3,$70,$7D,$DF,$F9,$5E,$BB,$B5
	.dc.b	$FF,$6E,$C6,$BD,$8E,$FC,$BC,$7E,$81,$F0,$D7,$E3,$E6,$1E,$89,$CB
	.dc.b	$FE,$EF,$FB,$B7,$C3,$F8,$7C,$2F,$C0,$4F,$B9,$AF,$FB,$A6,$76,$FC
	.dc.b	$2F,$BB,$79,$7E,$95,$F2,$D4,$C4,$7B,$7E,$31,$F6,$F7,$C2,$98,$7C
	.dc.b	$C9,$CB,$F3,$BE,$ED,$F6,$1F,$15,$CC,$7C,$26,$F4,$7F,$B7,$B9,$11
	.dc.b	$C2,$7E,$BF,$DD,$B8,$BE,$72,$F9,$8A,$62,$3E,$97,$D9,$6E,$BE,$FB
	.dc.b	$B6,$CE,$A7,$F3,$1F,$77,$FA,$EC,$BC,$5F,$7B,$F9,$EB,$FC,$3B,$54
	.dc.b	$7F,$B7,$E7,$87,$91,$F8,$BD,$1F,$3F,$F3,$2F,$F1,$EF,$CC,$6F,$1F
	.dc.b	$E1,$E3,$51,$F8,$4F,$BC,$FD,$13,$7E,$03,$F1,$91,$FE,$63,$8F,$CC
	.dc.b	$BA,$CB,$E4,$7D,$E7,$CA,$CB,$C4,$FF,$19,$8C,$F2,$7E,$56,$35,$7F
	.dc.b	$EB,$5F,$3D,$F7,$9F,$99,$5E,$DD,$31,$1F,$BD,$FB,$C7,$B7,$49,$7F
	.dc.b	$AB,$C9,$9F,$F8,$AB,$E8,$BE,$84,$FC,$88,$FF,$16,$DF,$28,$E1,$1B
	.dc.b	$FB,$CF,$E1,$54,$7A,$CF,$13,$F7,$15,$C5,$C0,$9E,$0F,$43,$F7,$CE
	.dc.b	$1D,$C7,$1B,$F3,$C7,$F0,$3E,$77,$DE,$7F,$1B,$FB,$BF,$B7,$3C,$C3
	.dc.b	$87,$C3,$3C,$1D,$35,$F2,$C6,$EB,$BC,$72,$E3,$FE,$51,$5B,$9C,$BF
	.dc.b	$BE,$9E,$3E,$3F,$8C,$DF,$8D,$23,$C9,$87,$F0,$DB,$EA,$DD,$7F,$CB
	.dc.b	$78,$73,$9E,$BE,$97,$F3,$7F,$2B,$F9,$94,$79,$1B,$5C,$27,$CA,$6F
	.dc.b	$CF,$98,$FB,$A3,$2F,$C5,$AE,$2F,$F2,$D6,$AC,$3F,$89,$E1,$F0,$BD
	.dc.b	$7F,$DF,$4F,$DA,$AF,$CC,$63,$FE,$72,$6C,$7F,$37,$9D,$D7,$1F,$0B
	.dc.b	$D5,$D1,$2B,$EB,$56,$B0,$FF,$3D,$BF,$ED,$C8,$FB,$5F,$77,$F0,$E7
	.dc.b	$FC,$FE,$D8,$ED,$F1,$7F,$D6,$EB,$84,$7E,$3A,$74,$9B,$1A,$C7,$D9
	.dc.b	$F2,$5C,$49,$5E,$0F,$86,$AF,$E6,$3E,$EF,$F0,$75,$F8,$46,$E5,$E8
	.dc.b	$BD,$CE,$63,$FB,$74,$EA,$BE,$25,$7F,$71,$FB,$A7,$D6,$EB,$E7,$E5
	.dc.b	$7E,$66,$72,$D9,$CB,$E3,$9E,$51,$CF,$ED,$BF,$BE,$7D,$DB,$0A,$66
	.dc.b	$FA,$4E,$1F,$D3,$4E,$DC,$9E,$1D,$7C,$C4,$77,$E7,$D7,$F7,$73,$FC
	.dc.b	$77,$4F,$F5,$9B,$63,$FB,$C6,$7F,$E3,$19,$FF,$5B,$B6,$3F,$BE,$4E
	.dc.b	$D3,$63,$50,$FB,$B7,$D8,$7F,$86,$7E,$9B,$FD,$73,$87,$FB,$09,$B5
	.dc.b	$FE,$07,$AD,$BF,$C0,$DA,$FF,$12,$8D,$A6,$C6,$B8,$3F,$80,$BF,$D5
	.dc.b	$67,$85,$F8,$4F,$F1,$54,$66,$77,$E6,$E6,$BF,$89,$5F,$3C,$3F,$DA
	.dc.b	$F7,$B7,$FA,$97,$DD,$FE,$4D,$BF,$69,$FE,$15,$83,$79,$A6,$1F,$E2
	.dc.b	$49,$CB,$67,$2F,$F8,$63,$EE,$9F,$77,$FE,$1F,$FD,$9D,$F7,$B4,$FF
	.dc.b	$11,$4E,$AB,$FC,$5E,$D7,$FC,$43,$3C,$1F,$F1,$9B,$5E,$D5,$FE,$2B
	.dc.b	$FB,$DF,$F2,$EC,$3D,$24,$39,$B4,$6F,$E6,$25,$7F,$8F,$A7,$4D,$9C
	.dc.b	$BF,$E3,$2F,$BD,$FF,$23,$3F,$E4,$47,$FC,$76,$DF,$07,$FC,$79,$1A
	.dc.b	$8E,$1F,$0D,$89,$91,$C6,$3F,$C8,$5F,$33,$FC,$C6,$3D,$0D,$C8,$DB
	.dc.b	$EF,$3F,$E4,$48,$D1,$3C,$9E,$B1,$FD,$7B,$EF,$7F,$CA,$3F,$2B,$FA
	.dc.b	$D8,$46,$3B,$FF,$CB,$91,$8F,$85,$FF,$78,$46,$27,$FC,$EB,$3F,$E6
	.dc.b	$1A,$E5,$1C,$63,$FC,$C8,$D7,$91,$B9,$E5,$F1,$7D,$23,$7E,$73,$E6
	.dc.b	$B5,$18,$6B,$E6,$8D,$2C,$3C,$38,$F3,$8D,$57,$0C,$7F,$C5,$6D,$BE
	.dc.b	$37,$C2,$B1,$31,$E9,$8F,$BE,$B6,$FC,$1C,$3B,$61,$D2,$B6,$79,$CA
	.dc.b	$78,$8A,$FC,$33,$86,$1F,$22,$FA,$CF,$18,$F2,$3C,$23,$2F,$38,$E1
	.dc.b	$87,$B6,$FC,$75,$AB,$EB,$87,$A6,$B5,$87,$B7,$0E,$2F,$58,$CB,$49
	.dc.b	$1A,$3C,$65,$9E,$91,$A7,$E2,$3D,$5C,$C6,$E3,$F0,$78,$79,$AB,$51
	.dc.b	$EF,$AE,$2F,$84,$FC,$89,$F2,$4B,$7F,$3D,$B8,$7E,$09,$CF,$86,$F8
	.dc.b	$7E,$31,$F9,$28,$C3,$8A,$FF,$7D,$CE,$A3,$F4,$33,$A7,$9C,$7E,$26
	.dc.b	$38,$E5,$FA,$7F,$37,$FD,$15,$DB,$9A,$97,$F7,$AF,$79,$7E,$AA,$F7
	.dc.b	$39,$AD,$E1,$FA,$8B,$DC,$FF,$B2,$EB,$73,$9A,$DD,$FE,$C6,$F7,$7F
	.dc.b	$EF,$5F,$03,$FC,$A3,$7B,$8E,$C4,$7D,$11,$AA,$E5,$7D,$2B,$E3,$3F
	.dc.b	$AE,$C3,$C3,$3A,$9E,$4D,$BE,$78,$F1,$9F,$93,$76,$F8,$C7,$43,$89
	.dc.b	$6B,$8B,$83,$89,$63,$8B,$88,$F6,$12,$FE,$66,$1E,$B9,$E9,$1E,$52
	.dc.b	$3F,$2E,$34,$F3,$8E,$38,$7B,$E7,$37,$FE,$E8,$8C,$12,$BF,$77,$1E
	.dc.b	$43,$4E,$27,$C5,$C4,$FB,$D7,$CC,$7F,$A1,$3C,$3F,$F6,$A4,$63,$1F
	.dc.b	$F4,$65,$3E,$F7,$FD,$BB,$18,$9D,$63,$FE,$FE,$6D,$F2,$C3,$A5,$7F
	.dc.b	$B1,$EF,$8D,$EE,$72,$FF,$5F,$87,$C8,$63,$1F,$07,$A1,$67,$38,$79
	.dc.b	$1E,$78,$FF,$BA,$DC,$E1,$FE,$A9,$BE,$35,$BC,$39,$61,$D7,$EF,$7F
	.dc.b	$C0,$B3,$EE,$5E,$5E,$7C,$4D,$36,$74,$DC,$7B,$57,$09,$F8,$FF,$7B
	.dc.b	$FE,$54,$7F,$BF,$F4,$FA,$E3,$15,$FE,$04,$8C,$79,$BE,$FF,$11,$18
	.dc.b	$BE,$27,$D2,$F8,$5F,$F6,$BB,$7C,$CF,$8D,$18,$C3,$E0,$3F,$E3,$E9
	.dc.b	$E1,$EF,$6A,$B6,$D8,$FF,$5F,$7D,$3B,$BF,$E8,$7D,$EB,$C5,$FA,$B9
	.dc.b	$2B,$57,$F8,$4A,$D3,$CA,$F8,$DF,$0A,$DD,$FD,$3B,$DB,$FE,$DA,$87
	.dc.b	$D0,$FE,$0B,$9D,$F6,$8D,$79,$4E,$63,$19,$FA,$23,$13,$A8,$FF,$71
	.dc.b	$57,$28,$F1,$78,$5F,$8B,$FB,$C7,$85,$8B,$FC,$EC,$61,$F8,$73,$B9
	.dc.b	$D6,$7F,$8B,$AD,$3E,$99,$CE,$1F,$EE,$CE,$A5,$3E,$0B,$38,$FF,$85
	.dc.b	$BE,$E9,$F2,$FE,$FF,$83,$7C,$57,$CE,$9C,$D7,$AA,$72,$FF,$BC,$E3
	.dc.b	$3A,$33,$FD,$C5,$6A,$31,$5D,$F0,$ED,$87,$5D,$BF,$59,$F7,$8F,$9A
	.dc.b	$C5,$FE,$FA,$30,$F7,$9D,$CE,$B0,$FD,$65,$69,$EE,$06,$6F,$BB,$FB
	.dc.b	$1E,$37,$B5,$B8,$FA,$DB,$5C,$8E,$2E,$35,$4F,$C9,$8E,$4F,$DE,$FB
	.dc.b	$87,$DD,$62,$F9,$7C,$A3,$E1,$3F,$85,$7D,$27,$F1,$9C,$3C,$D9,$74
	.dc.b	$BF,$45,$EE,$7F,$8F,$D6,$E7,$9D,$F2,$F2,$7F,$47,$7B,$9F,$E1,$D5
	.dc.b	$BC,$3A,$42,$7C,$B8,$7E,$53,$ED,$DF,$07,$CB,$3F,$D0,$46,$27,$CA
	.dc.b	$7F,$97,$2B,$8D,$FC,$EC,$15,$B8,$F7,$DF,$2C,$3C,$BF,$3F,$FC,$A7
	.dc.b	$EE,$BF,$D6,$99,$74,$BF,$7C,$E7,$E0,$F7,$3F,$CF,$AB,$75,$F2,$AB
	.dc.b	$53,$C6,$3D,$91,$87,$51,$C6,$BF,$83,$FD,$CF,$F1,$A8,$FD,$9F,$81
	.dc.b	$FD,$4B,$4F,$8B,$39,$BF,$64,$E6,$BF,$CF,$73,$9A,$FE,$A9,$39,$AF
	.dc.b	$C7,$4E,$6B,$ED,$CE,$65,$3F,$16,$BF,$69,$F3,$DF,$E1,$C7,$C3,$9F
	.dc.b	$F4,$D8,$C6,$1E,$08,$C7,$27,$A9,$B1,$F0,$6D,$F7,$3F,$63,$EE,$BD
	.dc.b	$ED,$9D,$37,$2F,$CE,$57,$09,$FC,$15,$70,$76,$7E,$16,$B1,$5F,$D6
	.dc.b	$7E,$EB,$F9,$4E,$7F,$C3,$57,$D1,$FF,$65,$73,$7F,$B1,$F0,$91,$A8
	.dc.b	$FE,$F3,$F7,$57,$0F,$EC,$9F,$7C,$FA,$EF,$ED,$D3,$9C,$BF,$8D,$4E
	.dc.b	$6F,$FC,$03,$C6,$FB,$7F,$EE,$8F,$F2,$2A,$FE,$40,$E6,$77,$E4,$7F
	.dc.b	$79,$FB,$BD,$47,$95,$F8,$59,$FF,$75,$72,$FF,$B7,$DC,$DF,$EA,$BE
	.dc.b	$EF,$F5,$CF,$F2,$DC,$D1,$FD,$50,$CD,$FC,$18,$CB,$FE,$18,$FB,$BF
	.dc.b	$56,$1F,$E0,$35,$EF,$FF,$83,$5C,$E1,$FC,$E5,$CD,$7F,$88,$25,$F0
	.dc.b	$9F,$E1,$37,$E7,$DF,$FB,$D5,$CC,$EF,$D2,$FF,$0C,$7D,$DF,$F7,$EC
	.dc.b	$7F,$A2,$3F,$97,$BF,$F7,$E3,$9A,$FB,$0E,$78,$FF,$B7,$79,$BF,$E2
	.dc.b	$1F,$45,$E8,$58,$C3,$FE,$25,$FB,$BF,$A1,$8F,$F8,$7D,$F9,$76,$73
	.dc.b	$97,$F5,$57,$E7,$CB,$EE,$F7,$97,$E9,$F3,$FF,$75,$C6,$3B,$7F,$C6
	.dc.b	$5F,$77,$E5,$FD,$3F,$FD,$C1,$5F,$E4,$48,$C7,$C4,$FF,$8D,$E1,$F2
	.dc.b	$94,$FE,$F9,$F7,$9F,$E4,$48,$FC,$B1,$A3,$FB,$BF,$DE,$7F,$8E,$9F
	.dc.b	$E6,$B1,$BD,$3F,$AF,$4E,$5F,$EC,$11,$9D,$92,$FA,$8F,$F0,$53,$6F
	.dc.b	$FC,$B3,$F7,$3F,$5B,$0F,$E5,$53,$97,$FA,$87,$DE,$7F,$5B,$C3,$F9
	.dc.b	$CB,$6F,$FA,$E4,$BE,$43,$FC,$2F,$1F,$C8,$E7,$53,$FD,$86,$72,$F4
	.dc.b	$F5,$FF,$C5,$31,$8B,$EB,$18,$F9,$5C,$FB,$D7,$1F,$E5,$D7,$B7,$FA
	.dc.b	$BF,$DE,$7E,$D5,$FE,$B3,$F3,$FF,$98,$73,$E4,$FF,$8A,$63,$19,$FF
	.dc.b	$81,$6B,$83,$FE,$07,$8C,$E6,$9C,$A8,$CC,$60,$CB,$F8,$09,$FE,$A5
	.dc.b	$1D,$1F,$F3,$3D,$7F,$8E,$7E,$0D,$F7,$5F,$E4,$8F,$8B,$EB,$18,$CB
	.dc.b	$FA,$54,$62,$75,$5C,$5F,$F2,$17,$DE,$7F,$8A,$6B,$FC,$77,$F1,$0F
	.dc.b	$98,$CF,$83,$FE,$47,$8C,$4F,$F9,$02,$B8,$67,$FD,$9E,$72,$FF,$69
	.dc.b	$8C,$ED,$FE,$9F,$9C,$BF,$DE,$27,$FC,$35,$F2,$9F,$E1,$BF,$B5,$7D
	.dc.b	$67,$E0,$AB,$FD,$3F,$18,$AF,$F8,$03,$5B,$AD,$7C,$A7,$F8,$36,$B5
	.dc.b	$5F,$8A,$AD,$3F,$E1,$29,$D5,$7E,$3A,$3C,$22,$B8,$DF,$82,$75,$2B
	.dc.b	$E6,$FF,$90,$EF,$9B,$E4,$39,$4F,$93,$2D,$C7,$6C,$3A,$D7,$08,$F3
	.dc.b	$1C,$0F,$34,$F0,$7C,$06,$4F,$03,$9B,$F1,$E1,$E7,$9F,$15,$F6,$AD
	.dc.b	$47,$19,$D3,$E0,$9D,$3D,$4F,$36,$1E,$97,$07,$88,$F8,$65,$3E,$2D
	.dc.b	$E1,$F9,$67,$CD,$FB,$B7,$C0,$3A,$F8,$C6,$1B,$19,$9F,$9F,$39,$6F
	.dc.b	$07,$F9,$C5,$B9,$72,$8C,$47,$76,$BF,$C2,$A7,$F4,$B3,$F6,$31,$7D
	.dc.b	$E7,$F0,$EC,$FF,$0F,$18,$BF,$BB,$7C,$1F,$A0,$D9,$DD,$71,$AF,$E2
	.dc.b	$E7,$1A,$DE,$1E,$7A,$FC,$9D,$F1,$AD,$DF,$1B,$FF,$3D,$47,$1A,$DD
	.dc.b	$F1,$8F,$53,$67,$75,$C5,$FD,$3B,$6F,$B4,$7F,$A1,$5A,$E8,$F8,$51
	.dc.b	$B9,$E9,$09,$DD,$78,$CF,$3C,$7C,$49,$DC,$EA,$BC,$55,$BA,$F0,$57
	.dc.b	$9F,$F0,$2F,$F1,$CF,$4F,$EE,$F6,$FF,$4C,$46,$2B,$FC,$CC,$D7,$FD
	.dc.b	$3A,$FF,$90,$3C,$4E,$10,$8D,$69,$FE,$8B,$6D,$74,$F6,$3F,$DB,$B5
	.dc.b	$D1,$FF,$40,$BD,$B4,$7C,$1F,$4B,$F8,$0D,$8C,$CE,$FE,$87,$13,$23
	.dc.b	$93,$B8,$F2,$E2,$FB,$B7,$CE,$61,$B7,$FE,$E6,$6B,$FE,$8D,$7F,$11
	.dc.b	$3F,$EA,$53,$94,$F1,$24,$F1,$E0,$AF,$18,$B3,$C6,$B9,$E9,$F3,$BF
	.dc.b	$18,$FF,$72,$5F,$CE,$6B,$F3,$9B,$5F,$2D,$FF,$7B,$C6,$FC,$FF,$8D
	.dc.b	$1F,$E2,$48,$C1,$CF,$AB,$33,$F2,$FF,$05,$F0,$CF,$D7,$6E,$EB,$E5
	.dc.b	$3F,$EA,$DE,$D7,$BF,$F2,$9C,$E5,$F0,$DB,$9F,$FB,$29,$DD,$7F,$AB
	.dc.b	$67,$FC,$EB,$C5,$C4,$91,$99,$E2,$FF,$BE,$7F,$21,$7C,$7F,$D8,$6C
	.dc.b	$73,$6B,$E5,$AF,$C3,$7C,$97,$F9,$E6,$3F,$2E,$FF,$A8,$1B,$F0,$7A
	.dc.b	$F8,$2F,$C6,$7D,$5B,$E0,$69,$2B,$F3,$F0,$F0,$BF,$25,$F1,$1F,$E6
	.dc.b	$69,$FD,$1B,$5F,$D1,$B7,$6F,$D0,$2F,$7F,$F3,$5E,$17,$FC,$01,$AF
	.dc.b	$D5,$B5,$FD,$5B,$6B,$F5,$0F,$F5,$EE,$1F,$F4,$D7,$DC,$BE,$7F,$F5
	.dc.b	$FE,$E2,$31,$87,$FB,$6E,$31,$3A,$9F,$F2,$14,$FE,$EB,$F1,$37,$E0
	.dc.b	$1F,$AD,$5A,$53,$11,$FB,$1F,$C6,$33,$5F,$C0,$97,$D9,$E4,$D5,$7E
	.dc.b	$DF,$DA,$7A,$D4,$97,$FC,$2A,$3D,$DA,$FF,$0E,$FC,$63,$FA,$8C,$FF
	.dc.b	$0E,$6B,$FC,$39,$BC,$1F,$C2,$DF,$F3,$5C,$7F,$14,$FC,$1B,$35,$FC
	.dc.b	$61,$B1,$FC,$4D,$BD,$E7,$8C,$2D,$5F,$EC,$7F,$BA,$F8,$F3,$98,$C7
	.dc.b	$B9,$F4,$27,$35,$C2,$9C,$69,$7D,$37,$F1,$9F,$C6,$1F,$3D,$7D,$B9
	.dc.b	$87,$F2,$C9,$CA,$3C,$05,$FB,$77,$D6,$FF,$2E,$84,$61,$E1,$9F,$8D
	.dc.b	$CC,$EE,$5F,$47,$62,$4F,$0D,$BF,$CF,$75,$CE,$3C,$27,$6A,$DB,$88
	.dc.b	$D5,$F7,$6D,$72,$AF,$E7,$1F,$89,$7C,$2B,$FE,$92,$BF,$01,$C9,$AD
	.dc.b	$64,$EA,$32,$F0,$8F,$E8,$5F,$89,$BE,$3A,$FB,$B2,$BF,$AA,$CE,$5B
	.dc.b	$39,$7F,$AB,$C2,$79,$CF,$CB,$F9,$4F,$35,$7C,$06,$EB,$F2,$DB,$A7
	.dc.b	$FA,$83,$4B,$5F,$5E,$6A,$69,$88,$B9,$7D,$95,$F3,$F4,$F1,$47,$F5
	.dc.b	$FF,$1B,$F5,$B7,$FD,$C1,$AF,$F7,$06,$E3,$FD,$B5,$7C,$CF,$1E,$E9
	.dc.b	$5C,$5E,$91,$FD,$AF,$F1,$57,$DF,$5F,$BF,$97,$FD,$FA,$72,$D9,$CB
	.dc.b	$FE,$00,$84,$F3,$9F,$CA,$7D,$DF,$E0,$AB,$F3,$4D,$4D,$3F,$26,$FE
	.dc.b	$44,$DF,$37,$22,$39,$38,$ED,$87,$F2,$B3,$CE,$7C,$33,$FC,$93,$F1
	.dc.b	$7F,$E0,$3D,$87,$F7,$FB,$7E,$DE,$31,$E0,$EF,$C9,$99,$7F,$29,$BD
	.dc.b	$F2,$7E,$CE,$F7,$39,$AF,$E6,$3E,$7F,$EE,$23,$F6,$F2,$8C,$63,$DD
	.dc.b	$B7,$CD,$FF,$66,$57,$F3,$77,$2F,$F0,$A6,$D7,$3E,$EE,$FD,$DF,$C0
	.dc.b	$7E,$F3,$F5,$57,$F1,$5A,$FE,$E1,$B8,$FF,$03,$7F,$1F,$1B,$EC,$E4
	.dc.b	$47,$97,$45,$79,$45,$AE,$53,$FD,$43,$F1,$37,$DC,$D2,$5F,$F5,$18
	.dc.b	$FC,$4B,$F1,$1D,$B9,$8D,$B9,$E4,$FE,$B7,$E0,$BE,$7E,$92,$7F,$AE
	.dc.b	$47,$F8,$8B,$83,$EA,$3B,$8C,$CE,$DF,$A7,$1B,$7F,$CA,$9D,$0F,$8D
	.dc.b	$FD,$11,$23,$F4,$1F,$80,$57,$F3,$D1,$75,$FF,$2D,$52,$BF,$55,$F6
	.dc.b	$1F,$AB,$E2,$98,$7E,$A6,$32,$FE,$AE,$32,$FE,$A6,$3F,$65,$C5,$23
	.dc.b	$F6,$0F,$ED,$7E,$57,$11,$FA,$F8,$B8,$F8,$A7,$FC,$C9,$F1,$1F,$CF
	.dc.b	$E3,$FC,$DC,$DD,$BF,$CF,$2D,$BF,$F3,$A3,$F1,$DF,$F3,$CB,$9F,$2C
	.dc.b	$D2,$6F,$DB,$E0,$65,$7F,$A0,$A7,$E6,$57,$0B,$CC,$F4,$AB,$87,$92
	.dc.b	$7F,$AC,$7D,$C7,$F3,$5A,$FF,$7B,$7C,$F7,$F5,$F5,$19,$F8,$FF,$F1
	.dc.b	$63,$67,$4F,$F3,$99,$FE,$CD,$3A,$F2,$39,$36,$33,$5C,$27,$2F,$2F
	.dc.b	$26,$CF,$F6,$FB,$7C,$23,$E6,$38,$CE,$63,$0F,$A0,$C4,$69,$F4,$E1
	.dc.b	$D1,$F5,$CF,$0C,$7D,$6D,$C7,$D6,$D9,$F1,$57,$1C,$3C,$78,$75,$6C
	.dc.b	$F8,$4E,$B1,$99,$E1,$5C,$E7,$26,$EB,$8E,$1F,$16,$31,$3B,$9C,$E1
	.dc.b	$E4,$C3,$A5,$78,$6B,$44,$8C,$C7,$27,$44,$7C,$18,$79,$23,$18,$7C
	.dc.b	$27,$8B,$67,$91,$1F,$31,$92,$5F,$96,$3A,$9D,$1C,$46,$A3,$E5,$99
	.dc.b	$3E,$B9,$D6,$3F,$08,$F9,$CF,$C9,$1F,$20,$B8,$7D,$06,$E1,$F8,$27
	.dc.b	$3E,$5F,$FB,$19,$FC,$AB,$5F,$C6,$35,$FC,$EB,$67,$FC,$0D,$F0,$FF
	.dc.b	$76,$3F,$DA,$38,$7F,$A2,$FE,$E3,$FB,$DE,$B9,$A9,$87,$DB,$BD,$E5
	.dc.b	$FA,$9B,$DC,$E6,$B7,$87,$EA,$AF,$73,$FE,$4D,$AD,$CE,$6B,$77,$EA
	.dc.b	$7F,$39,$3A,$8C,$57,$52,$57,$9F,$CA,$FF,$31,$3C,$E3,$33,$C2,$35
	.dc.b	$8F,$69,$F8,$B8,$55,$87,$77,$27,$22,$D7,$27,$07,$CC,$9E,$53,$C6
	.dc.b	$30,$6B,$2F,$9E,$FE,$A6,$35,$18,$AF,$41,$30,$F9,$98,$FB,$1F,$51
	.dc.b	$2F,$C6,$EB,$1F,$0C,$EE,$7E,$73,$5F,$6F,$97,$71,$F3,$12,$71,$19
	.dc.b	$79,$71,$7F,$34,$D3,$FD,$1D,$C3,$FC,$F9,$B8,$FF,$3E,$DA,$E5,$FC
	.dc.b	$FA,$6D,$F1,$AF,$E7,$11,$8E,$3F,$E3,$F1,$83,$F9,$E5,$E4,$D7,$C0
	.dc.b	$D8,$C9,$30,$FE,$93,$19,$7F,$A5,$FC,$2F,$53,$B7,$97,$6B,$FA,$44
	.dc.b	$59,$FE,$A3,$37,$3F,$EB,$3E,$5F,$97,$E5,$FC,$CD,$C3,$FA,$83,$6F
	.dc.b	$FA,$CB,$6B,$FB,$0F,$CA,$7B,$4F,$76,$96,$FF,$AC,$45,$BF,$EC,$33
	.dc.b	$6B,$FC,$89,$F9,$B7,$FB,$DE,$FF,$DD,$4D,$AF,$F1,$EA,$F4,$7E,$7F
	.dc.b	$96,$8D,$5F,$2F,$81,$CC,$63,$0F,$F2,$53,$C5,$EB,$1C,$BD,$07,$D1
	.dc.b	$D3,$1F,$45,$6B,$FD,$F9,$3A,$6C,$E9,$A7,$F9,$AF,$07,$D5,$FF,$A9
	.dc.b	$E3,$0D,$D3,$FD,$9B,$3A,$F0,$3F,$12,$F0,$F0,$2D,$6E,$3F,$22,$DB
	.dc.b	$FF,$67,$BF,$EA,$B7,$AC,$F0,$23,$E0,$AB,$7F,$E6,$6E,$1F,$84,$DD
	.dc.b	$FE,$5C,$3E,$16,$C4,$BF,$F4,$37,$07,$FB,$6A,$37,$39,$BF,$F3,$8F
	.dc.b	$E1,$BF,$E3,$5B,$DA,$F1,$FE,$BB,$99,$DC,$63,$C9,$CB,$6B,$8B,$F0
	.dc.b	$B7,$91,$F2,$B8,$6D,$FC,$AD,$FE,$A3,$6B,$E4,$CE,$B8,$3E,$4C,$5B
	.dc.b	$F9,$33,$6B,$FC,$6B,$F2,$9F,$B5,$F9,$3F,$DB,$36,$BF,$C6,$0D,$9F
	.dc.b	$F1,$CB,$70,$FF,$20,$FC,$9F,$EF,$4F,$DF,$63,$6F,$FC,$71,$16,$FF
	.dc.b	$C8,$33,$69,$5F,$E3,$3F,$AD,$7D,$6F,$E0,$B1,$FE,$1B,$82,$73,$5F
	.dc.b	$8C,$9C,$C6,$25,$7D,$11,$C6,$FF,$1B,$D3,$C1,$98,$7F,$AF,$EF,$95
	.dc.b	$FE,$42,$B3,$7F,$ED,$1F,$88,$FE,$3B,$E8,$3E,$1E,$C6,$1B,$8F,$FB
	.dc.b	$6D,$B5,$FE,$D9,$7F,$92,$FC,$03,$4B,$BE,$1B,$7F,$ED,$5F,$15,$5B
	.dc.b	$8C,$2A,$E1,$18,$7F,$53,$E0,$C8,$FD,$8F,$55,$CB,$F1,$13,$76,$FD
	.dc.b	$6C,$DA,$FD,$6C,$DA,$F9,$7C,$8A,$C6,$4F,$81,$7C,$BF,$F4,$05,$EF
	.dc.b	$58,$C3,$FC,$63,$EA,$5F,$21,$4C,$79,$D7,$5F,$81,$F8,$AF,$F9,$F7
	.dc.b	$5F,$B7,$1F,$6B,$34,$E7,$7F,$BD,$18,$E0,$FE,$A3,$39,$AE,$1E,$3B
	.dc.b	$C8,$FE,$AA,$D8,$F9,$CE,$F1,$60,$8D,$47,$88,$CF,$E2,$DF,$4D,$C4
	.dc.b	$A3,$19,$79,$1F,$AD,$7F,$6D,$CB,$EB,$F7,$BF,$88,$3F,$7D,$7E,$09
	.dc.b	$FD,$A9,$CE,$1D,$9C,$CF,$E1,$7D,$16,$DF,$DC,$1F,$E8,$15,$EA,$73
	.dc.b	$3B,$F2,$FF,$A4,$FD,$43,$E5,$E9,$FD,$33,$DA,$FF,$49,$57,$F2,$09
	.dc.b	$D3,$F9,$3F,$C5,$9F,$7B,$FD,$F9,$B1,$9F,$A3,$C4,$FF,$80,$BF,$16
	.dc.b	$77,$53,$1C,$B6,$FF,$3D,$F1,$7F,$C9,$DB,$E5,$FF,$82,$A1,$F0,$56
	.dc.b	$3F,$C1,$BF,$1C,$F8,$3B,$F0,$78,$6E,$1F,$D4,$FF,$3C,$FF,$7E,$37
	.dc.b	$87,$FA,$7F,$D4,$BF,$13,$FE,$A3,$5A,$E7,$FE,$AB,$39,$BF,$EB,$8D
	.dc.b	$8C,$FC,$67,$2B,$E7,$F1,$9D,$22,$DF,$07,$FC,$47,$D9,$FC,$4A,$3F
	.dc.b	$79,$F4,$0F,$91,$CE,$3F,$E2,$88,$C4,$FF,$01,$AE,$13,$F8,$E6,$F8
	.dc.b	$1F,$A0,$FC,$5F,$C8,$FE,$8A,$33,$5F,$A7,$9C,$B7,$3F,$1F,$C4,$FF
	.dc.b	$8D,$A7,$38,$7F,$28,$E0,$FD,$AD,$FF,$7B,$6C,$7F,$78,$7F,$16,$DF
	.dc.b	$1F,$FB,$CC,$6B,$2F,$F2,$3B,$6B,$F9,$8C,$7D,$7F,$07,$75,$27,$B3
	.dc.b	$CA,$31,$7F,$AC,$F6,$FF,$CE,$B5,$FC,$8D,$B1,$9F,$98,$FF,$27,$C6
	.dc.b	$30,$FE,$06,$DA,$FE,$10,$DA,$FD,$EA,$AE,$3F,$11,$FE,$58,$FC,$55
	.dc.b	$EA,$D3,$1E,$AD,$C7,$F9,$A4,$E9,$B3,$A6,$9F,$00,$FF,$5E,$E2,$F8
	.dc.b	$AF,$EA,$7F,$4E,$FA,$7F,$EA,$B3,$FD,$63,$4F,$EB,$13,$98,$FF,$62
	.dc.b	$D7,$0A,$DE,$BF,$C1,$2B,$55,$CA,$B4,$FD,$0B,$E7,$7E,$29,$CC,$F7
	.dc.b	$C2,$BF,$D9,$A3,$38,$75,$AE,$12,$C3,$BC,$E6,$3C,$A6,$67,$CB,$87
	.dc.b	$82,$33,$5C,$23,$11,$E7,$38,$1E,$7C,$3C,$6F,$C7,$32,$79,$CA,$79
	.dc.b	$CC,$DF,$8D,$E9,$5D,$E7,$9E,$1E,$CB,$F0,$D7,$1A,$D3,$D0,$F0,$1F
	.dc.b	$60,$A7,$C2,$31,$39,$FA,$7F,$EB,$FF,$51,$7D,$43,$49,$5F,$82,$9F
	.dc.b	$EC,$F9,$71,$F5,$7F,$C2,$F3,$96,$BF,$84,$13,$C1,$D4,$FD,$AD,$1F
	.dc.b	$51,$F8,$AC,$FF,$07,$18,$AF,$BF,$7C,$67,$F0,$ED,$9F,$C8,$9B,$9F
	.dc.b	$9E,$D9,$DE,$1E,$66,$E1,$E1,$C3,$C1,$5C,$9B,$7F,$E3,$0C,$3B,$37
	.dc.b	$0E,$55,$B8,$E4,$79,$A3,$2E,$EF,$2F,$92,$FA,$BF,$12,$76,$D9,$DE
	.dc.b	$1F,$0A,$B7,$87,$8A,$B4,$73,$37,$87,$EC,$A3,$CF,$E9,$3E,$31,$EF
	.dc.b	$7E,$C3,$83,$FC,$25,$18,$C3,$FC,$6C,$D7,$FC,$06,$FF,$8F,$5C,$CF
	.dc.b	$17,$FC,$1B,$83,$E8,$B1,$87,$F0,$36,$DF,$3D,$3F,$C3,$B9,$7F,$8F
	.dc.b	$F0,$FF,$22,$35,$FF,$08,$B6,$37,$F0,$7D,$23,$19,$FF,$8A,$5B,$5F
	.dc.b	$E1,$E6,$E3,$F8,$EA,$F3,$5B,$EA,$1D,$B8,$6E,$1F,$EB,$F6,$BF,$EB
	.dc.b	$C7,$F1,$53,$FE,$71,$39,$4F,$12,$4F,$1D,$AB,$C6,$2C,$F1,$AE,$79
	.dc.b	$7F,$90,$7E,$A1,$F2,$BF,$21,$7C,$8D,$31,$17,$1F,$F3,$54,$E9,$B3
	.dc.b	$A6,$9F,$DD,$32,$FA,$7E,$1F,$82,$3F,$C8,$75,$C5,$FF,$6C,$FA,$CF
	.dc.b	$91,$F8,$EE,$74,$F8,$0D,$CF,$FB,$0B,$BA,$FF,$73,$3F,$8D,$6F,$0F
	.dc.b	$23,$9E,$7F,$4B,$E9,$DF,$51,$F8,$6F,$C6,$DF,$80,$D3,$51,$F8,$AF
	.dc.b	$9F,$7D,$55,$F1,$E6,$7F,$E8,$B9,$CB,$FE,$8D,$FA,$A7,$C9,$FE,$49
	.dc.b	$B1,$98,$C7,$87,$F9,$6F,$1B,$FC,$9F,$3F,$E9,$57,$6F,$F9,$89,$BC
	.dc.b	$5F,$E7,$47,$71,$FE,$66,$F0,$DF,$77,$FA,$05,$F7,$14,$C4,$7E,$9F
	.dc.b	$E5,$9F,$59,$FA,$7B,$FF,$79,$B5,$FF,$3A,$46,$BC,$2E,$12,$FA,$8F
	.dc.b	$D6,$FD,$67,$F7,$CB,$FF,$56,$3F,$EB,$96,$E7,$AF,$2F,$F6,$7F,$59
	.dc.b	$99,$FF,$3E,$AF,$88,$C9,$AA,$FD,$CF,$CE,$FF,$82,$EB,$FD,$70,$D7
	.dc.b	$FD,$0A,$DE,$0F,$F4,$BB,$F0,$A3,$FD,$7D,$F5,$9F,$EA,$86,$FC,$37
	.dc.b	$F0,$48,$FF,$37,$C6,$B8,$BF,$83,$FD,$66,$E3,$FD,$22,$E9,$B5,$A9
	.dc.b	$F9,$5F,$59,$F0,$E7,$FC,$28,$38,$91,$98,$F9,$6D,$CB,$FD,$D5,$3A
	.dc.b	$FA,$CF,$E2,$FF,$51,$FE,$74,$BF,$F6,$B3,$FE,$73,$6F,$B3,$AF,$86
	.dc.b	$FE,$41,$0F,$85,$FF,$41,$BE,$23,$1F,$13,$A8,$CC,$7F,$AB,$CE,$1F
	.dc.b	$2D,$99,$91,$C6,$77,$9F,$FB,$FE,$36,$66,$76,$F3,$9D,$F9,$3F,$E3
	.dc.b	$07,$FD,$B9,$1F,$EE,$FF,$96,$69,$FC,$D1,$7C,$1D,$31,$12,$78,$46
	.dc.b	$5E,$51,$FC,$F3,$E6,$1F,$45,$7E,$8A,$57,$FB,$EA,$34,$61,$FF,$65
	.dc.b	$CA,$73,$1F,$1F,$E9,$3C,$AD,$BF,$9B,$18,$8F,$91,$B5,$DB,$E2,$4D
	.dc.b	$D7,$FC,$9F,$2F,$88,$FA,$AD,$BE,$84,$39,$C6,$8E,$F8,$7F,$58,$F1
	.dc.b	$7F,$49,$3F,$8C,$F6,$7F,$85,$F6,$E4,$D7,$D2,$D9,$D5,$70,$7F,$DD
	.dc.b	$2D,$D3,$FD,$F6,$A7,$FB,$2F,$CF,$3E,$A6,$FC,$5C,$C3,$C1,$39,$6C
	.dc.b	$E5,$FF,$7A,$C7,$FA,$BD,$C9,$F9,$1F,$91,$FC,$1A,$0E,$1B,$CF,$F9
	.dc.b	$A8,$FC,$3C,$7E,$55,$D1,$23,$5B,$ED,$93,$DD,$CB,$FB,$2E,$B7,$D6
	.dc.b	$7E,$EF,$5F,$D7,$AF,$0B,$F9,$13,$6B,$83,$BE,$3F,$45,$F2,$EF,$DD
	.dc.b	$CB,$DC,$7F,$2B,$F3,$BF,$71,$1F,$CA,$1B,$8F,$EF,$9B,$87,$47,$F7
	.dc.b	$93,$BE,$DE,$44,$7D,$51,$AA,$FD,$97,$D6,$7E,$A2,$7F,$7C,$D7,$F7
	.dc.b	$8D,$8D,$F3,$FF,$06,$8C,$61,$FE,$A4,$6D,$7D,$56,$A7,$F3,$0F,$4E
	.dc.b	$CE,$44,$7C,$35,$FD,$2F,$E6,$9F,$59,$49,$3F,$4E,$3F,$14,$FF,$4C
	.dc.b	$76,$E6,$36,$E7,$93,$FA,$B7,$84,$F9,$3A,$49,$FB,$D1,$FE,$18,$6C
	.dc.b	$65,$FE,$B3,$1C,$5D,$FC,$07,$E8,$3E,$B2,$47,$E7,$9F,$D1,$FD,$46
	.dc.b	$23,$F3,$B1,$77,$6F,$F9,$BE,$E7,$53,$0F,$D5,$FC,$BF,$EB,$39,$26
	.dc.b	$5F,$AA,$8C,$BF,$AC,$8C,$BF,$AA,$8F,$D9,$F8,$F2,$3F,$62,$FE,$DB
	.dc.b	$E5,$F1,$1F,$B0,$8B,$9F,$92,$19,$79,$2B,$FC,$C3,$F1,$5F,$CE,$6B
	.dc.b	$53,$FE,$70,$D3,$26,$B8,$2E,$1F,$C3,$EF,$84,$FB,$DB,$1D,$9F,$98
	.dc.b	$AF,$AC,$7D,$38,$C4,$F1,$DF,$C6,$76,$8C,$9C,$DF,$3F,$D4,$48,$F1
	.dc.b	$47,$F5,$EF,$AB,$FE,$7F,$1F,$EE,$BF,$94,$FF,$7E,$C6,$21,$F2,$3F
	.dc.b	$E2,$48,$C3,$7C,$9F,$F7,$D5,$70,$F2,$ED,$FF,$15,$BD,$CE,$63,$B1
	.dc.b	$9A,$DA,$9F,$F3,$2F,$D6,$6A,$31,$F0,$88,$C4,$EA,$7C,$B7,$D5,$A7
	.dc.b	$86,$B8,$38,$39,$D7,$28,$CE,$3E,$DC,$7C,$78,$75,$8F,$81,$3A,$C3
	.dc.b	$E2,$B7,$0F,$75,$F0,$8C,$47,$63,$C1,$7C,$EB,$CD,$7D,$30,$F1,$12
	.dc.b	$B7,$87,$38,$C4,$EE,$72,$62,$BC,$73,$63,$C6,$53,$C5,$3C,$4F,$73
	.dc.b	$9B,$F2,$3F,$1E,$7E,$6D,$F8,$63,$D3,$87,$B2,$B7,$1F,$5E,$75,$7D
	.dc.b	$70,$F1,$CF,$98,$E3,$87,$7C,$39,$5F,$59,$FE,$B4,$78,$4A,$52,$E1
	.dc.b	$F7,$DB,$E6,$7F,$AE,$9F,$CB,$B5,$FC,$83,$5F,$C6,$36,$0F,$FB,$13
	.dc.b	$1F,$CE,$37,$1F,$1F,$CA,$7F,$7B,$DB,$F1,$F3,$9C,$3A,$FC,$6E,$0F
	.dc.b	$CB,$56,$E3,$EF,$CE,$DF,$0C,$ED,$FC,$94,$EF,$0F,$0C,$E5,$B3,$AF
	.dc.b	$2F,$F8,$94,$60,$98,$FC,$0A,$F0,$CE,$A3,$15,$E2,$9E,$91,$EC,$9D
	.dc.b	$4F,$A8,$D1,$2B,$FD,$43,$7E,$5A,$E7,$1E,$09,$E0,$EA,$B9,$C7,$B8
	.dc.b	$E5,$3C,$0F,$06,$3E,$28,$E3,$19,$7F,$C3,$57,$DD,$C9,$0E,$97,$DE
	.dc.b	$B9,$46,$2B,$92,$AE,$4E,$4E,$B7,$E6,$8C,$9F,$EA,$5A,$F9,$24,$9F
	.dc.b	$A0,$FD,$CF,$37,$85,$7C,$8E,$57,$1F,$E2,$AB,$F9,$E7,$C2,$7F,$47
	.dc.b	$E5,$FE,$6F,$B5,$9F,$E7,$D3,$72,$FE,$7D,$37,$0E,$6D,$C1,$C3,$63
	.dc.b	$0C,$C3,$FA,$57,$C3,$DC,$3F,$9F,$E1,$E4,$AC,$65,$FD,$22,$AD,$FF
	.dc.b	$4F,$F8,$9E,$C7,$7F,$32,$C7,$F5,$58,$B8,$F4,$CD,$58,$C3,$FA,$9F
	.dc.b	$9B,$E7,$F9,$BF,$43,$71,$FE,$BC,$DC,$39,$36,$BF,$AD,$FD,$17,$BC
	.dc.b	$F8,$18,$DB,$FE,$CD,$16,$FC,$53,$6B,$FD,$7B,$F5,$8F,$A2,$F9,$9E
	.dc.b	$82,$31,$8F,$F9,$5D,$7A,$9A,$73,$8F,$FB,$C9,$B8,$74,$C3,$55,$B7
	.dc.b	$AF,$A2,$FA,$BA,$63,$C1,$B9,$7F,$99,$67,$4D,$9D,$34,$FF,$6D,$E0
	.dc.b	$FA,$8F,$F5,$7C,$61,$B9,$FF,$7A,$9D,$78,$3F,$89,$78,$78,$16,$B7
	.dc.b	$1F,$91,$6D,$FF,$A1,$63,$8B,$D6,$78,$11,$F0,$55,$BF,$F0,$4E,$CE
	.dc.b	$1F,$93,$5F,$E8,$16,$CE,$97,$C4,$7D,$27,$38,$FF,$BC,$9B,$5F,$E8
	.dc.b	$06,$D7,$FB,$ED,$E3,$1D,$77,$FB,$7F,$59,$FE,$41,$8F,$F4,$82,$F2
	.dc.b	$33,$E5,$69,$B3,$A6,$E1,$EA,$AE,$13,$F3,$77,$F9,$07,$C9,$E1,$B1
	.dc.b	$92,$57,$D1,$E0,$4E,$A3,$15,$74,$FA,$13,$6F,$FC,$79,$F5,$5F,$BB
	.dc.b	$FA,$9F,$DE,$36,$3F,$C8,$4D,$C3,$F5,$CD,$BF,$F1,$FF,$CB,$FF,$03
	.dc.b	$3F,$82,$63,$6F,$FC,$95,$16,$FF,$6D,$36,$BF,$CB,$10,$FA,$B7,$D6
	.dc.b	$7E,$1F,$1F,$E2,$53,$98,$C6,$3F,$87,$3F,$B9,$D7,$07,$F2,$95,$C2
	.dc.b	$31,$5F,$16,$31,$52,$BF,$35,$D3,$6B,$FD,$B7,$F1,$1F,$C7,$FD,$17
	.dc.b	$C3,$D8,$C3,$6F,$FD,$CA,$DA,$FF,$71,$BF,$C9,$FE,$05,$95,$DF,$0D
	.dc.b	$BF,$F7,$E4,$E9,$B3,$A6,$9F,$AF,$F0,$3F,$54,$7E,$AF,$9E,$C6,$49
	.dc.b	$8F,$E0,$E3,$2F,$E1,$66,$EF,$FB,$19,$B5,$FE,$9B,$F8,$27,$CE,$FF
	.dc.b	$3F,$5E,$FD,$8C,$3F,$C6,$BE,$B1,$F2,$54,$C7,$9F,$70,$FA,$DA,$7F
	.dc.b	$A5,$3E,$B3,$FD,$20,$FF,$42,$7E,$03,$79,$3F,$A7,$46,$35,$FF,$4B
	.dc.b	$F8,$CF,$1F,$FA,$92,$2B,$87,$FA,$B7,$D6,$71,$C7,$FD,$6C,$DC,$7F
	.dc.b	$D4,$AB,$8B,$7C,$9F,$D7,$7D,$FF,$DE,$D7,$F5,$65,$F4,$D9,$F1,$34
	.dc.b	$FF,$63,$F4,$9C,$17,$CA,$C3,$74,$F8,$3E,$3B,$EB,$1F,$37,$FD,$09
	.dc.b	$B3,$AF,$6F,$FD,$10,$F0,$8C,$EB,$FD,$A3,$EB,$1F,$1D,$FD,$CD,$FC
	.dc.b	$EF,$2F,$FA,$F1,$CD,$FD,$0F,$51,$1A,$8F,$F6,$C7,$D5,$FF,$66,$74
	.dc.b	$FF,$29,$6F,$81,$FD,$FE,$31,$8F,$F7,$8F,$8B,$FE,$2A,$DF,$37,$FC
	.dc.b	$09,$0F,$82,$B1,$FE,$07,$F8,$E7,$C1,$DF,$83,$C3,$73,$FE,$9B,$F5
	.dc.b	$7F,$EF,$D7,$D1,$E2,$BE,$A3,$FA,$54,$E7,$4F,$F0,$4C,$E6,$BF,$AA
	.dc.b	$FD,$5B,$EA,$FF,$A8,$36,$33,$F1,$B8,$9C,$FC,$6E,$A2,$CE,$8F,$91
	.dc.b	$DC,$C7,$FC,$41,$F4,$6F,$92,$FF,$11,$4E,$6B,$E7,$4E,$5E,$35,$A3
	.dc.b	$FC,$51,$F5,$7E,$37,$FC,$43,$5C,$5F,$F1,$1C,$67,$A7,$FD,$EB,$39
	.dc.b	$7F,$DE,$F1,$9A,$F2,$74,$FE,$52,$FF,$B7,$70,$FE,$CA,$7F,$BB,$B6
	.dc.b	$3F,$BA,$3F,$96,$8C,$F9,$7C,$08,$F0,$F0,$2D,$70,$9B,$8F,$F9,$1E
	.dc.b	$F9,$4E,$6F,$F1,$DE,$1F,$F9,$6D,$BF,$0F,$FE,$47,$8F,$AD,$3F,$E4
	.dc.b	$07,$4D,$8D,$1B,$E5,$B5,$F9,$9F,$AB,$7A,$9F,$E5,$1F,$B8,$FE,$67
	.dc.b	$39,$8C,$7A,$BF,$E5,$5A,$DC,$E4,$E5,$19,$9F,$E9,$78,$FD,$DF,$8A
	.dc.b	$FB,$DF,$56,$FA,$BF,$F5,$EE,$FF,$C0,$A3,$15,$FE,$BE,$AD,$D6,$B6
	.dc.b	$65,$8A,$D0,$BE,$2B,$0E,$97,$FE,$C8,$86,$1C,$5D,$61,$D1,$E5,$3B
	.dc.b	$9D,$57,$29,$EF,$7D,$A3,$11,$FE,$D6,$9D,$DF,$3A,$F1,$CF,$C1,$3A
	.dc.b	$9F,$E3,$18,$D5,$F7,$8C,$9E,$4C,$3C,$CF,$1A,$F0,$57,$08,$F1,$D7
	.dc.b	$C5,$AD,$C7,$B8,$E7,$87,$99,$E5,$3B,$7B,$98,$3C,$93,$CB,$0F,$91
	.dc.b	$1E,$D2,$9D,$4C,$BA,$FA,$BF,$EB,$3F,$59,$7D,$73,$49,$5F,$4E,$7F
	.dc.b	$B0,$E3,$C7,$BB,$EC,$4E,$5A,$FF,$C0,$09,$DB,$99,$FB,$59,$3E,$AF
	.dc.b	$F8,$54,$FC,$A8,$C3,$F7,$EB,$8E,$BF,$86,$6C,$F0,$37,$3C,$6B,$9D
	.dc.b	$6F,$0F,$2B,$70,$F0,$CE,$AF,$93,$6F,$F8,$E6,$1D,$9B,$87,$2A,$DC
	.dc.b	$72,$3C,$D1,$97,$77,$97,$C8,$69,$8E,$09,$87,$95,$FC,$5D,$F8,$5F
	.dc.b	$D8,$61,$E5,$69,$CA,$7F,$63,$CE,$F8,$BF,$EF,$13,$FD,$75,$AF,$FB
	.dc.b	$31,$B8,$7E,$A6,$78,$7C,$0F,$EF,$37,$FE,$53,$5A,$6F,$A9,$9C,$17
	.dc.b	$DE,$30,$F2,$8F,$ED,$99,$FF,$AB,$5B,$F0,$5A,$8F,$F3,$8C,$6B,$E0
	.dc.b	$B0,$66,$64,$64,$FE,$ED,$F1,$63,$B9,$FE,$DA,$73,$E6,$E2,$24,$60
	.dc.b	$87,$FB,$B1,$E4,$F4,$79,$C4,$8C,$1F,$E0,$5F,$AB,$F9,$8D,$D3,$DF
	.dc.b	$18,$8F,$F7,$BB,$D1,$B1,$F1,$3E,$6D,$F1,$54,$E3,$16,$7F,$C2,$53
	.dc.b	$A6,$CE,$9F,$F0,$B7,$DE,$FF,$93,$9B,$F1,$BF,$42,$0D,$E8,$C3,$F5
	.dc.b	$3A,$FF,$C4,$75,$FE,$1F,$6B,$FC,$BF,$8B,$FB,$3D,$F2,$8D,$7A,$9F
	.dc.b	$73,$EA,$DF,$1D,$4E,$91,$F8,$0F,$9D,$FF,$1E,$D7,$F8,$AD,$AF,$53
	.dc.b	$F0,$EE,$3A,$3F,$20,$FF,$BF,$E3,$FC,$65,$F5,$3F,$DD,$DE,$1C,$E8
	.dc.b	$FF,$37,$C6,$BB,$3F,$1F,$1B,$BF,$CD,$78,$3F,$ED,$8A,$FF,$1E,$35
	.dc.b	$E6,$D8,$DF,$95,$F9,$8E,$1F,$ED,$35,$FE,$43,$6B,$E2,$6C,$FE,$9D
	.dc.b	$FF,$4F,$C7,$F9,$27,$F2,$1F,$F1,$E3,$7C,$8F,$D1,$C7,$F8,$F2,$35
	.dc.b	$EB,$7E,$9B,$E9,$5F,$01,$49,$3F,$E5,$38,$7C,$BF,$EB,$3E,$96,$CF
	.dc.b	$F9,$5D,$AF,$C6,$6E,$7F,$BB,$5F,$59,$FB,$4E,$9F,$EC,$1D,$1F,$E2
	.dc.b	$45,$F5,$79,$35,$E4,$7E,$EB,$E9,$FF,$D9,$B9,$7F,$99,$5A,$FD,$46
	.dc.b	$E5,$FC,$25,$7D,$7F,$F0,$2F,$A8,$CE,$5F,$E3,$26,$DF,$F0,$86,$F4
	.dc.b	$1E,$31,$FE,$73,$D9,$F3,$3F,$C3,$1F,$B1,$7F,$E3,$C6,$CF,$F1,$27
	.dc.b	$6F,$F9,$EF,$E8,$DE,$FF,$F1,$46,$C6,$63,$F8,$BF,$CD,$FF,$BB,$35
	.dc.b	$FF,$76,$FC,$A7,$FA,$DA,$BF,$D6,$61,$FC,$C6,$3F,$D5,$9B,$1F,$E6
	.dc.b	$17,$83,$F9,$F3,$87,$B1,$FC,$9D,$E5,$EC,$D9,$E3,$1A,$3B,$BC,$AF
	.dc.b	$C7,$3B,$38,$57,$F2,$EF,$94,$7B,$F4,$98,$7F,$A6,$E3,$FC,$FC,$D8
	.dc.b	$CB,$FA,$C8,$E2,$EF,$E0,$3F,$9C,$FC,$AF,$FA,$C2,$3F,$DD,$FF,$17
	.dc.b	$83,$F9,$EF,$E3,$6F,$A2,$C4,$E8,$87,$EE,$23,$4F,$CA,$F9,$FF,$E1
	.dc.b	$8F,$C5,$6E,$BF,$E3,$98,$46,$37,$FF,$01,$F8,$DF,$B0,$7F,$C0,$B1
	.dc.b	$99,$FF,$48,$35,$E3,$D9,$F3,$EB,$84,$6A,$31,$E0,$48,$F2,$47,$F5
	.dc.b	$8F,$99,$7D,$2D,$F8,$69,$5F,$EC,$59,$CB,$67,$2F,$F1,$68,$FF,$50
	.dc.b	$39,$3F,$D9,$9D,$5F,$B2,$6F,$D3,$7F,$66,$83,$7C,$29,$3A,$21,$FC
	.dc.b	$8A,$79,$3F,$8D,$F8,$DF,$ED,$8D,$F6,$7F,$1D,$1F,$DB,$23,$5E,$D6
	.dc.b	$0C,$CC,$8F,$CE,$CF,$2E,$1F,$A4,$DE,$1F,$F1,$9B,$CE,$33,$5C,$EB
	.dc.b	$FD,$D1,$F3,$5F,$D1,$1B,$E9,$7F,$80,$A3,$11,$FC,$E9,$FE,$B8,$D8
	.dc.b	$D1,$1E,$7E,$B7,$F8,$37,$E4,$DE,$8B,$E8,$30,$BE,$42,$98,$89,$7F
	.dc.b	$ED,$3C,$FF,$D0,$97,$CA,$B7,$39,$D7,$FD,$B1,$3D,$AF,$F6,$D8,$3E
	.dc.b	$8E,$FC,$24,$C3,$F8,$1C,$E5,$B3,$97,$F8,$04,$7F,$94,$B2,$FC,$E3
	.dc.b	$B8,$CC,$EE,$73,$5E,$08,$F9,$1D,$BF,$E2,$AF,$D8,$7F,$53,$7F,$D8
	.dc.b	$9B,$7F,$E3,$7D,$FF,$6D,$3F,$36,$B8,$77,$48,$FA,$B1,$CF,$EB,$FA
	.dc.b	$78,$92,$7E,$11,$FE,$3A,$C5,$3B,$8C,$F9,$89,$FE,$29,$C1,$FE,$EC
	.dc.b	$6C,$F1,$AF,$47,$8D,$E9,$F9,$89,$3C,$AB,$F9,$3E,$1C,$EB,$33,$ED
	.dc.b	$AE,$2E,$FA,$7E,$11,$1E,$5F,$15,$5F,$89,$17,$7F,$F2,$D6,$3F,$D1
	.dc.b	$31,$F4,$7B,$3F,$2F,$DA,$98,$78,$E3,$E7,$18,$7E,$74,$61,$F4,$7B
	.dc.b	$5F,$4F,$C3,$91,$82,$1F,$55,$CB,$F5,$E2,$E3,$E4,$F7,$3E,$E1,$8E
	.dc.b	$0F,$95,$78,$BF,$8D,$39,$43,$E9,$BF,$57,$F1,$7F,$E2,$C6,$F0,$FF
	.dc.b	$AA,$1B,$3C,$1F,$D7,$CE,$FB,$BF,$DB,$D5,$B6,$D6,$DA,$F7,$EE,$B6
	.dc.b	$FA,$5F,$E7,$78,$78,$5B,$87,$85,$B7,$CC,$F4,$CE,$FA,$2D,$6D,$B3
	.dc.b	$E7,$AE,$2F,$BF,$1F,$0C,$63,$0F,$15,$70,$AF,$89,$39,$AD,$CE,$6B
	.dc.b	$74,$C3,$C4,$DC,$3E,$13,$BC,$3B,$12,$F9,$61,$D2,$78,$1E,$78,$E0
	.dc.b	$69,$A6,$8F,$2B,$8B,$FA,$51,$6F,$E1,$3B,$C3,$CF,$3A,$8C,$61,$E3
	.dc.b	$6D,$F7,$69,$27,$DA,$71,$2D,$F6,$6C,$7D,$53,$84,$64,$8E,$63,$E3
	.dc.b	$BF,$EF,$43,$D6,$52,$99,$BF,$9E,$DF,$2B,$F9,$03,$F9,$56,$BF,$95
	.dc.b	$6B,$F9,$D6,$C7,$F8,$0B,$AF,$ED,$47,$F2,$FA,$FF,$45,$FC,$87,$F9
	.dc.b	$93,$7C,$D4,$C3,$E5,$5E,$F1,$FD,$4C,$2F,$83,$AA,$E1,$87,$EA,$EF
	.dc.b	$84,$FE,$C2,$B7,$39,$AD,$DF,$AD,$FC,$94,$EA,$31,$5D,$49,$5E,$8F
	.dc.b	$55,$8F,$29,$F8,$58,$57,$CD,$3A,$78,$DF,$4B,$F1,$47,$E3,$70,$F0
	.dc.b	$3C,$8D,$96,$3C,$18,$75,$CB,$E6,$4E,$8E,$F1,$9C,$BE,$0C,$7D,$3C
	.dc.b	$BD,$B9,$78,$F0,$F0,$93,$2F,$21,$F3,$B2,$F1,$DF,$13,$B6,$5E,$0A
	.dc.b	$F1,$C7,$86,$76,$4F,$2F,$B5,$F3,$C2,$D7,$D3,$6B,$E8,$3E,$14,$E2
	.dc.b	$BE,$1F,$37,$F5,$7F,$A2,$F9,$5C,$9F,$D1,$3D,$8B,$BF,$F5,$99,$B8
	.dc.b	$FF,$61,$9B,$7C,$6B,$FA,$C4,$63,$A5,$76,$31,$F8,$0E,$62,$D7,$F5
	.dc.b	$F6,$E1,$FD,$E5,$B8,$FF,$56,$7A,$7D,$7F,$5F,$AE,$B7,$FD,$9D,$B7
	.dc.b	$FE,$01,$6D,$FF,$5E,$7C,$9F,$5B,$2B,$FB,$77,$CC,$5B,$FF,$04,$4D
	.dc.b	$BF,$EC,$B3,$6B,$FB,$D7,$B1,$ED,$F5,$FD,$CD,$8F,$EF,$0D,$C3,$FC
	.dc.b	$24,$DB,$FC,$B6,$EF,$D8,$BF,$AB,$DF,$F8,$21,$FF,$15,$36,$FE,$6C
	.dc.b	$61,$FF,$7E,$BC,$1B,$CF,$A7,$6F,$0F,$27,$8F,$1D,$BE,$CD,$AE,$6D
	.dc.b	$C7,$CB,$87,$F0,$1A,$FF,$0F,$F9,$8F,$9D,$78,$3F,$82,$8C,$78,$36
	.dc.b	$73,$5C,$14,$66,$B4,$D9,$D3,$72,$63,$FE,$52,$BE,$35,$CE,$F8,$BD
	.dc.b	$7C,$7F,$07,$8D,$E1,$6C,$64,$F1,$46,$5A,$7C,$FF,$98,$F9,$A7,$CE
	.dc.b	$FB,$1B,$19,$26,$5F,$42,$32,$F9,$A3,$FC,$D1,$F8,$0F,$FB,$E9,$FF
	.dc.b	$22,$E6,$87,$57,$EA,$23,$1D,$76,$73,$5C,$1E,$57,$D4,$F8,$6E,$9C
	.dc.b	$57,$29,$E1,$7C,$EF,$A3,$63,$27,$E3,$7E,$13,$F1,$7F,$0B,$F8,$C6
	.dc.b	$C6,$4F,$C9,$46,$5A,$7F,$A6,$3E,$93,$F7,$FE,$8B,$E5,$AC,$61,$B3
	.dc.b	$FE,$9E,$6E,$7F,$E9,$B7,$F8,$4A,$FA,$AF,$F7,$9C,$FE,$3B,$CD,$E8
	.dc.b	$FF,$9B,$A7,$5D,$9F,$8F,$9C,$C7,$C5,$33,$3B,$AE,$71,$F2,$79,$DF
	.dc.b	$8D,$3D,$4F,$0D,$6B,$87,$FD,$D4,$DC,$3F,$B0,$B7,$06,$1C,$AF,$8E
	.dc.b	$5F,$C8,$3D,$D7,$05,$31,$C1,$72,$FF,$71,$45,$C7,$FD,$C1,$37,$3F
	.dc.b	$E0,$7F,$04,$FA,$8F,$E5,$6D,$C3,$FD,$16,$B8,$B3,$32,$72,$F0,$8C
	.dc.b	$61,$F8,$6F,$AC,$7D,$57,$E4,$5E,$91,$98,$C5,$65,$B7,$F9,$AA,$E2
	.dc.b	$D9,$D3,$4D,$4F,$0A,$FC,$C7,$86,$F8,$67,$B1,$FD,$19,$B7,$FE,$A6
	.dc.b	$5E,$C6,$67,$FA,$5F,$C5,$3C,$FB,$E8,$61,$B8,$7F,$AB,$BE,$AD,$F1
	.dc.b	$3F,$D4,$9B,$19,$DB,$13,$FD,$59,$7C,$DD,$31,$16,$7F,$D7,$3F,$57
	.dc.b	$F1,$B9,$3F,$9B,$BF,$46,$33,$7F,$D8,$23,$1C,$1F,$03,$EA,$DE,$3F
	.dc.b	$F6,$46,$DF,$FB,$11,$78,$F9,$9F,$ED,$3A,$FF,$B2,$AB,$FD,$92,$DB
	.dc.b	$F9,$6B,$E0,$37,$F0,$1F,$DC,$3F,$21,$F9,$75,$F4,$1F,$97,$7F,$ED
	.dc.b	$27,$3A,$FF,$50,$73,$3F,$5B,$EA,$9F,$05,$FD,$E9,$B7,$FE,$DA,$7F
	.dc.b	$9D,$CE,$E7,$FB,$FF,$9D,$FE,$9D,$AF,$F6,$FB,$67,$F0,$2F,$CE,$8C
	.dc.b	$FC,$07,$F8,$1B,$E1,$DF,$63,$7E,$1B,$0D,$AF,$F7,$4F,$D5,$BE,$23
	.dc.b	$FC,$1A,$D8,$CF,$CB,$E2,$7F,$C2,$3F,$82,$BD,$4F,$F0,$9E,$DF,$D6
	.dc.b	$76,$FC,$9C,$E7,$0F,$F7,$AC,$E7,$87,$FB,$9F,$E3,$AC,$B9,$E3,$FE
	.dc.b	$1C,$8C,$5F,$FB,$F3,$0E,$58,$FE,$7A,$7C,$8D,$BE,$19,$71,$7F,$C4
	.dc.b	$BF,$63,$F8,$0C,$3F,$DA,$B9,$7F,$80,$A3,$17,$FA,$8F,$AB,$7C,$07
	.dc.b	$0E,$AF,$F1,$6C,$63,$0E,$27,$FB,$7B,$C6,$61,$CB,$1F,$F0,$6D,$6F
	.dc.b	$1F,$D9,$CE,$6F,$84,$E6,$B7,$8F,$6F,$3B,$BB,$FE,$3C,$FC,$81,$F0
	.dc.b	$97,$E1,$30,$DC,$7C,$79,$B4,$5F,$09,$CC,$62,$F8,$C6,$25,$18,$AF
	.dc.b	$E1,$31,$89,$D5,$F7,$9F,$E1,$BE,$47,$F8,$8F,$3F,$F1,$5A,$F8,$1F
	.dc.b	$F2,$9B,$9B,$F8,$CE,$5F,$1B,$9A,$F4,$39,$7A,$FC,$17,$2A,$FF,$2E
	.dc.b	$FD,$5F,$F3,$EB,$F5,$B6,$BF,$8F,$AF,$4B,$7E,$9C,$AD,$BC,$A3,$15
	.dc.b	$FD,$64,$FF,$7F,$8F,$AA,$FE,$96,$FE,$1B,$2F,$F4,$E4,$62,$3F,$D6
	.dc.b	$BF,$14,$78,$5F,$FA,$E3,$4F,$E0,$75,$B9,$EA,$F1,$8C,$CF,$FB,$4E
	.dc.b	$B9,$4E,$6F,$BC,$E6,$FA,$9E,$3C,$77,$19,$AE,$11,$89,$FF,$70,$5D
	.dc.b	$7F,$C7,$77,$D1,$C9,$D0,$CC,$ED,$F3,$DF,$AE,$78,$47,$C0,$38,$61
	.dc.b	$EA,$79,$E1,$E2,$7A,$1C,$8F,$41,$81,$87,$96,$79,$47,$59,$E4,$7C
	.dc.b	$97,$67,$FB,$E4,$FA,$E5,$9C,$38,$3F,$DD,$3F,$52,$F7,$1F,$57,$C6
	.dc.b	$30,$D8,$CC,$FF,$BC,$0F,$B9,$3A,$70,$1F,$E6,$F8,$73,$EE,$FF,$33
	.dc.b	$CE,$A7,$2D,$8C,$BD,$63,$38,$FE,$2F,$3F,$D4,$4F,$D7,$6B,$F5,$67
	.dc.b	$71,$EC,$AE,$F5,$C7,$7F,$C9,$5F,$DB,$AE,$33,$F9,$7B,$E3,$3E,$CB
	.dc.b	$E3,$5C,$0D,$96,$F9,$DF,$19,$DD,$71,$C3,$CD,$5A,$BE,$13,$F8,$FC
	.dc.b	$3B,$BF,$1E,$F8,$BF,$E3,$6A,$E0,$FE,$A6,$BC,$2F,$53,$E0,$4F,$08
	.dc.b	$CC,$71,$3B,$47,$4A,$EF,$3A,$36,$FC,$0E,$47,$D4,$7F,$99,$EF,$F9
	.dc.b	$1B,$5F,$F1,$3B,$71,$FF,$62,$AF,$AB,$FD,$E7,$07,$F3,$F8,$D3,$FD
	.dc.b	$07,$1F,$E5,$37,$FC,$B9,$AF,$4F,$80,$E2,$FF,$35,$C9,$F2,$D8,$CF
	.dc.b	$F1,$9F,$2D,$64,$D9,$95,$FC,$EA,$72,$D9,$CB,$FE,$40,$F5,$3F,$C5
	.dc.b	$AF,$F4,$05,$F5,$7D,$DB,$19,$70,$66,$24,$64,$87,$57,$2F,$78,$F5
	.dc.b	$7D,$5F,$CE,$8F,$E9,$ED,$7F,$CA,$6D,$DF,$FD,$E2,$F9,$A3,$3B,$7C
	.dc.b	$3F,$AB,$7C,$C5,$3A,$C5,$9F,$EB,$73,$A6,$CE,$9F,$EB,$FF,$55,$FC
	.dc.b	$05,$BF,$1B,$F3,$E3,$FD,$D3,$1A,$E7,$C4,$FD,$3F,$AB,$FF,$73,$CF
	.dc.b	$F6,$86,$BF,$E6,$CA,$E2,$FF,$9B,$BE,$AD,$F3,$7F,$6D,$B1,$FD,$C1
	.dc.b	$AF,$F9,$D2,$B9,$EA,$E6,$FF,$05,$7D,$5B,$E6,$7F,$08,$DB,$FE,$F2
	.dc.b	$D7,$FC,$FA,$FF,$97,$9F,$EF,$BF,$45,$FE,$7A,$6F,$5F,$E3,$63,$FC
	.dc.b	$9B,$1A,$F8,$4F,$C8,$7C,$F7,$FA,$12,$3F,$C0,$AE,$9F,$F4,$4B,$72
	.dc.b	$F3,$47,$F8,$36,$38,$3F,$E8,$EE,$06,$9F,$9B,$7F,$D1,$11,$FE,$10
	.dc.b	$6B,$B8,$FD,$07,$D5,$7F,$8F,$E3,$FC,$2A,$BE,$AF,$26,$BD,$0F,$D3
	.dc.b	$7D,$5F,$2C,$BC,$7D,$C8,$C7,$0F,$F8,$CD,$B8,$FF,$88,$D7,$C3,$FE
	.dc.b	$BF,$EA,$EC,$FF,$89,$1A,$FF,$AA,$1B,$87,$65,$F4,$7F,$B7,$E8,$FE
	.dc.b	$F5,$8F,$F9,$05,$FF,$08,$5F,$F8,$B9,$AF,$FA,$C9,$DB,$FE,$35,$FA
	.dc.b	$B7,$D0,$7F,$01,$5F,$31,$4C,$47,$F0,$5F,$AB,$DD,$7F,$8E,$9D,$B6
	.dc.b	$75,$3F,$C2,$F4,$FF,$4E,$CF,$F9,$01,$FA,$FA,$7C,$86,$E5,$FE,$59
	.dc.b	$77,$3F,$EC,$8F,$8E,$7D,$5F,$F1,$56,$C7,$F9,$29,$AF,$FB,$3B,$EA
	.dc.b	$FF,$0A,$B0,$BF,$1F,$FE,$A7,$8F,$F2,$A9,$FE,$C6,$ED,$FE,$46,$D3
	.dc.b	$31,$B7,$96,$5F,$D7,$23,$15,$D2,$B8,$C7,$E8,$E3,$BF,$C5,$60,$CC
	.dc.b	$CB,$EB,$1F,$CB,$FE,$45,$ED,$D2,$4F,$FB,$8F,$E8,$AB,$FE,$EA,$9D
	.dc.b	$BF,$EE,$79,$E5,$3F,$CE,$7E,$59,$F3,$14,$95,$FE,$70,$FA,$B7,$D4
	.dc.b	$62,$3F,$DE,$B5,$C1,$FF,$3A,$B5,$FF,$7C,$CF,$07,$E4,$FD,$47,$F1
	.dc.b	$0D,$7E,$7B,$77,$FF,$00,$B6,$7F,$CA,$9E,$27,$ED,$1E,$3D,$3F,$42
	.dc.b	$3F,$4B,$06,$CF,$18,$36,$36,$6A,$7F,$AC,$7C,$DB,$EA,$AF,$D5,$CC
	.dc.b	$3F,$D1,$F3,$96,$CE,$5F,$BA,$7F,$A5,$E1,$3D,$27,$FB,$27,$C3,$FE
	.dc.b	$C2,$BF,$D3,$2B,$EA,$F2,$6B,$BE,$4E,$A3,$2F,$08,$FC,$7F,$C8,$7F
	.dc.b	$70,$8F,$C9,$BF,$DA,$9F,$18,$71,$B7,$E7,$1E,$11,$A6,$C6,$7D,$6E
	.dc.b	$A7,$6F,$62,$57,$68,$CB,$E0,$8F,$F0,$07,$CF,$7F,$44,$AF,$E3,$6D
	.dc.b	$E1,$FF,$66,$36,$BF,$D7,$8F,$F5,$F7,$E4,$F8,$C8,$D3,$7D,$1E,$A7
	.dc.b	$6F,$4F,$FC,$23,$F3,$EF,$3D,$F2,$58,$7F,$DF,$D8,$FF,$B2,$A7,$5F
	.dc.b	$1B,$9A,$93,$9A,$DB,$F6,$AB,$6E,$67,$F7,$B2,$F8,$7B,$F0,$32,$BF
	.dc.b	$81,$CE,$5B,$39,$7E,$9C,$7F,$98,$B6,$FE,$1A,$EE,$33,$3B,$9C,$D7
	.dc.b	$82,$3E,$6F,$77,$F8,$B7,$EA,$BF,$55,$7F,$21,$B7,$F4,$27,$73,$FE
	.dc.b	$E6,$9C,$F0,$7F,$BF,$5B,$1F,$4A,$7A,$7C,$C7,$5F,$12,$4F,$18,$CB
	.dc.b	$CE,$32,$F8,$63,$C9,$E2,$F9,$7E,$6A,$4F,$86,$32,$F8,$E3,$2F,$A6
	.dc.b	$3D,$9D,$52,$3D,$0F,$BF,$E2,$31,$1E,$F8,$BA,$7F,$68,$CF,$FA,$0C
	.dc.b	$B9,$51,$FE,$09,$8F,$8D,$EC,$FC,$93,$E5,$7B,$32,$3E,$34,$65,$F9
	.dc.b	$11,$F3,$FD,$9F,$A1,$E1,$49,$F9,$91,$97,$E7,$46,$5F,$A9,$17,$1F
	.dc.b	$37,$D3,$FD,$93,$1C,$5F,$4D,$C5,$7D,$77,$2F,$FB,$CF,$EA,$FF,$C5
	.dc.b	$FF,$15,$FE,$27,$6F,$07,$FA,$99,$B3,$BE,$BF,$F7,$DC,$E7,$B2,$CF
	.dc.b	$36,$AA,$E2,$DA,$CC,$3E,$63,$F9,$C6,$1E,$16,$E1,$E1,$6D,$F5,$07
	.dc.b	$FB,$92,$35,$E5,$E2,$B9,$36,$33,$3B,$39,$14,$F3,$57,$66,$CF,$17
	.dc.b	$E3,$46,$6F,$D6,$7A,$67,$53,$C0,$E9,$19,$6E,$3F,$02,$B9,$12,$B8
	.dc.b	$D7,$48,$C5,$72,$7C,$15,$C6,$B7,$1A,$C3,$EC,$4E,$6B,$85,$68,$EB
	.dc.b	$19,$9F,$8B,$3A,$3C,$A6,$CC,$3A,$3A,$B9,$71,$87,$CC,$73,$1C,$23
	.dc.b	$04,$7E,$7C,$78,$1F,$F7,$89,$F1,$CB,$87,$C7,$6E,$1F,$C4,$9F,$C9
	.dc.b	$79,$3F,$C6,$DB,$D5,$C9,$FC,$D4,$65,$B1,$97,$F2,$51,$FC,$A6,$FF
	.dc.b	$3C,$DC,$3F,$17,$2F,$AA,$FF,$30,$ED,$F5,$2B,$53,$D0,$FB,$AE,$1E
	.dc.b	$1E,$A7,$FA,$03,$0E,$B8,$F1,$24,$FE,$3F,$0E,$AF,$28,$CC,$F0,$8D
	.dc.b	$57,$3C,$7D,$F1,$B2,$4F,$FA,$76,$F9,$BC,$49,$3C,$1D,$57,$3A,$F5
	.dc.b	$46,$9B,$3F,$0A,$B9,$BC,$49,$1A,$3F,$63,$1A,$70,$7C,$EB,$F5,$CF
	.dc.b	$3B,$F0,$4E,$4F,$4C,$74,$7E,$4E,$3C,$1E,$79,$FF,$8B,$2F,$F5,$B8
	.dc.b	$77,$9E,$58,$FA,$5B,$87,$7C,$39,$47,$43,$E5,$BC,$BE,$77,$C7,$7C
	.dc.b	$C9,$96,$23,$2F,$C7,$E7,$FE,$7F,$F4,$FF,$EB,$CE,$3F,$E8,$8D,$C3
	.dc.b	$FB,$07,$AF,$72,$FE,$8D,$D2,$AC,$61,$FD,$22,$AC,$E6,$FF,$B5,$46
	.dc.b	$2B,$3E,$C6,$A2,$CE,$8F,$EE,$11,$96,$99,$C3,$FB,$C7,$C8,$F6,$F9
	.dc.b	$3E,$ED,$9F,$EF,$0D,$9D,$C7,$F7,$B9,$F2,$FD,$5C,$8F,$EF,$DF,$2B
	.dc.b	$6B,$FC,$09,$37,$5F,$EC,$D3,$6B,$FC,$05,$EB,$FB,$BD,$7F,$7B,$63
	.dc.b	$FC,$10,$DC,$3F,$C1,$CD,$BF,$F6,$BF,$D5,$FF,$3E,$AF,$E7,$F1,$8E
	.dc.b	$6F,$97,$39,$BF,$F1,$EB,$FE,$29,$AD,$9F,$D1,$B8,$51,$BB,$E5,$3A
	.dc.b	$AF,$EF,$93,$98,$E0,$F3,$8E,$BE,$47,$83,$97,$C2,$DA,$E5,$87,$68
	.dc.b	$C5,$70,$9C,$DF,$F0,$D7,$FC,$91,$F3,$CF,$07,$F1,$1E,$62,$1E,$0E
	.dc.b	$1D,$57,$08,$C5,$65,$B1,$93,$AC,$65,$A6,$70,$FF,$29,$53,$C7,$ED
	.dc.b	$E3,$77,$6D,$79,$9B,$5C,$1B,$B7,$D6,$FA,$BF,$A6,$7D,$4F,$9D,$B7
	.dc.b	$F0,$22,$DF,$D6,$9B,$5F,$AB,$FA,$BF,$D1,$4F,$FA,$7A,$11,$FE,$FE
	.dc.b	$BF,$AD,$18,$EB,$FD,$73,$AE,$AB,$5C,$A3,$A4,$62,$72,$DA,$E7,$5D
	.dc.b	$6F,$C0,$F1,$8F,$C7,$B9,$BF,$C9,$FC,$A7,$E4,$3E,$63,$F2,$2D,$AF
	.dc.b	$BA,$DB,$FC,$9B,$6B,$FD,$2F,$F5,$7F,$C0,$BD,$47,$CB,$58,$C3,$70
	.dc.b	$FC,$5C,$E9,$B3,$A7,$F3,$18,$3E,$AB,$FD,$EB,$1F,$E9,$3F,$37,$84
	.dc.b	$FF,$AD,$A7,$2F,$FA,$46,$33,$8A,$BF,$B0,$D6,$A3,$14,$E6,$5F,$0B
	.dc.b	$DF,$67,$E6,$23,$15,$FE,$34,$C3,$95,$68,$FF,$53,$7D,$5F,$FC,$00
	.dc.b	$7D,$31,$A9,$FF,$6E,$6B,$FC,$9B,$DF,$37,$7C,$32,$FF,$DC,$EE,$5A
	.dc.b	$7E,$CB,$E0,$9F,$51,$FC,$C1,$B1,$9E,$0C,$4C,$8E,$0E,$E3,$18,$FE
	.dc.b	$4F,$EA,$DF,$57,$F9,$07,$9C,$66,$31,$59,$6D,$FE,$62,$B8,$B6,$74
	.dc.b	$D3,$53,$C2,$BF,$DA,$FF,$0C,$F5,$FF,$A4,$B6,$BF,$87,$AF,$5B,$33
	.dc.b	$FD,$3B,$E2,$5E,$7D,$F3,$F0,$DC,$3F,$8B,$FD,$5B,$E2,$7F,$AA,$BE
	.dc.b	$6C,$FF,$DC,$EE,$67,$7A,$7F,$5A,$D5,$F3,$57,$E6,$F0,$DA,$FE,$47
	.dc.b	$F5,$7F,$1B,$B3,$F1,$7E,$5A,$3F,$DE,$53,$AC,$3F,$B2,$C6,$33,$FF
	.dc.b	$7B,$FD,$5F,$F1,$5A,$FE,$5A,$D8,$CF,$8F,$FD,$B2,$31,$8F,$6D,$DE
	.dc.b	$FF,$F7,$06,$E1,$FC,$D9,$7C,$06,$67,$FB,$AF,$D5,$FC,$DC,$98,$7A
	.dc.b	$23,$E7,$61,$FC,$F9,$D4,$67,$9B,$FB,$EF,$D5,$BE,$0A,$F8,$58,$6E
	.dc.b	$1F,$D1,$BC,$DF,$F5,$03,$7E,$7B,$FC,$0B,$18,$CF,$C3,$3F,$D6,$61
	.dc.b	$3F,$E0,$AF,$8A,$7D,$5D,$F8,$8C,$36,$FF,$A9,$FD,$5B,$E2,$BF,$C2
	.dc.b	$2D,$8C,$FC,$CE,$27,$FC,$2B,$A7,$F5,$5D,$3C,$27,$FB,$7D,$C7,$07
	.dc.b	$DA,$FA,$B7,$A7,$C3,$DA,$FF,$0E,$FD,$5B,$2F,$F0,$FD,$6B,$3F,$EC
	.dc.b	$B3,$97,$FB,$3C,$67,$2E,$FD,$AA,$F3,$46,$1E,$0D,$D3,$BE,$1F,$E2
	.dc.b	$CF,$AB,$76,$53,$FC,$5F,$F5,$6F,$80,$C6,$9F,$DD,$A3,$0F,$F7,$8A
	.dc.b	$E2,$E6,$77,$8F,$92,$B6,$D8,$EB,$E9,$76,$7F,$C7,$BF,$56,$F8,$4B
	.dc.b	$F0,$98,$6E,$3F,$E0,$69,$D3,$67,$4D,$35,$3C,$30,$F8,$B5,$C1,$B5
	.dc.b	$FE,$4C,$F3,$3F,$C2,$DC,$9F,$AE,$5F,$03,$9F,$82,$D3,$72,$FF,$0B
	.dc.b	$CE,$9B,$3A,$7F,$C3,$7F,$57,$FC,$C9,$FF,$0E,$FD,$63,$E8,$FF,$CB
	.dc.b	$73,$9C,$FE,$CC,$E6,$B8,$FA,$DA,$37,$EB,$D8,$FE,$B3,$5F,$EF,$71
	.dc.b	$F5,$7F,$D0,$1F,$F1,$FE,$7F,$D4,$A3,$11,$FE,$B6,$F8,$CF,$F8,$03
	.dc.b	$18,$DB,$F7,$0E,$A7,$A3,$9A,$E5,$8F,$FB,$1E,$F8,$BF,$EC,$0B,$E2
	.dc.b	$E6,$FC,$6E,$6B,$8C,$7F,$B6,$2B,$8B,$E6,$C2,$BE,$53,$89,$92,$99
	.dc.b	$C3,$C9,$1C,$23,$CF,$7D,$E3,$10,$AE,$B8,$79,$A3,$C4,$74,$70,$F5
	.dc.b	$AE,$4F,$80,$EA,$78,$EF,$E3,$3D,$4E,$58,$7B,$9F,$51,$FE,$E9,$BF
	.dc.b	$F7,$53,$F5,$CC,$1F,$EF,$1F,$AA,$FF,$02,$FD,$67,$FB,$BD,$B1,$9F
	.dc.b	$1D,$39,$6C,$71,$F0,$7F,$97,$B7,$3E,$51,$88,$FF,$7A,$8B,$FE,$5D
	.dc.b	$3A,$3E,$8E,$7F,$A0,$9B,$C5,$FE,$FC,$D3,$F1,$37,$C2,$7E,$9B,$63
	.dc.b	$F1,$B5,$C6,$7E,$BB,$63,$8E,$1E,$66,$E1,$DA,$A1,$D2,$75,$48,$97
	.dc.b	$DE,$31,$1F,$E1,$EC,$3C,$D1,$83,$C0,$6A,$78,$4F,$2A,$EA,$F5,$9D
	.dc.b	$B6,$B9,$BB,$BF,$29,$FE,$3C,$7C,$07,$F9,$2A,$7E,$04,$66,$FB,$CE
	.dc.b	$8F,$E8,$65,$8E,$53,$C3,$5F,$36,$1E,$4A,$7E,$75,$FE,$19,$7F,$CA
	.dc.b	$ED,$79,$B6,$FC,$8F,$F6,$98,$DE,$DF,$B5,$BF,$8F,$3B,$6C,$66,$7E
	.dc.b	$74,$F1,$7C,$FC,$2F,$95,$D4,$23,$0F,$F9,$A9,$AF,$BB,$F1,$F7,$C4
	.dc.b	$70,$9C,$E5,$F2,$5B,$97,$DA,$6D,$FD,$FB,$E2,$F5,$FA,$EE,$44,$7F
	.dc.b	$CF,$31,$FC,$AD,$D1,$0F,$4C,$7A,$23,$51,$83,$8F,$A5,$93,$A3,$E9
	.dc.b	$F9,$CF,$7F,$E6,$FF,$E4,$28,$FF,$43,$B5,$D1,$89,$D9,$1F,$2C,$FC
	.dc.b	$8F,$39,$E0,$52,$4F,$13,$FD,$27,$0E,$DE,$04,$8F,$7F,$99,$CF,$E3
	.dc.b	$24,$70,$74,$FE,$3B,$F4,$0F,$F5,$2B,$FE,$9E,$6B,$9F,$8A,$FB,$71
	.dc.b	$F4,$21,$F9,$C3,$E0,$7E,$FB,$63,$FD,$4E,$D7,$F3,$7C,$EF,$B9,$FC
	.dc.b	$2F,$E6,$CF,$94,$A4,$AF,$F5,$7D,$7F,$A3,$1F,$F5,$8C,$FE,$3F,$9F
	.dc.b	$F5,$8D,$E7,$FD,$EB,$67,$F8,$15,$79,$E3,$FC,$2F,$1C,$BC,$CF,$09
	.dc.b	$F9,$6F,$CD,$BF,$CB,$31,$FE,$BE,$87,$47,$E6,$9B,$19,$8E,$1E,$16
	.dc.b	$74,$93,$C5,$E9,$E5,$FE,$8B,$F3,$67,$D3,$52,$57,$FB,$33,$F3,$6F
	.dc.b	$E5,$FB,$FF,$07,$A7,$AD,$C4,$95,$BF,$33,$F5,$BF,$8F,$3E,$6E,$92
	.dc.b	$7A,$75,$6B,$69,$1F,$ED,$AA,$FF,$66,$BE,$1F,$B1,$FD,$BF,$E7,$AF
	.dc.b	$A3,$A4,$9F,$F7,$0F,$E7,$BF,$E9,$27,$FD,$C7,$35,$FE,$2E,$DD,$BF
	.dc.b	$90,$37,$A3,$F9,$8D,$7C,$AA,$FE,$3B,$F4,$6F,$A0,$FE,$0E,$D8,$FF
	.dc.b	$77,$B5,$FE,$47,$F2,$17,$1F,$E5,$93,$70,$FE,$7A,$DA,$FE,$84,$D7
	.dc.b	$FD,$61,$7D,$88,$F6,$7F,$8B,$DE,$FE,$B3,$FC,$DF,$3F,$F0,$01,$AF
	.dc.b	$F3,$46,$E1,$FC,$F5,$FF,$12,$61,$FD,$39,$CB,$7E,$07,$F9,$1C,$62
	.dc.b	$BF,$AB,$3C,$1B,$F5,$7A,$FA,$C9,$1E,$18,$E2,$61,$F1,$46,$1F,$8F
	.dc.b	$F4,$BF,$28,$8F,$A7,$82,$BF,$42,$2E,$5E,$99,$B9,$7C,$A9,$B5,$E6
	.dc.b	$9F,$E7,$7F,$0E,$F9,$2A,$49,$FE,$E1,$87,$F5,$F9,$FF,$6F,$46,$30
	.dc.b	$FE,$E3,$1C,$63,$07,$F7,$77,$0F,$F6,$06,$CF,$3D,$7C,$3F,$57,$E2
	.dc.b	$FA,$C9,$7C,$85,$A7,$71,$98,$4F,$08,$ED,$3F,$D5,$BF,$1E,$FE,$1D
	.dc.b	$1F,$DC,$1B,$87,$F7,$96,$E3,$FE,$0A,$FC,$0B,$F8,$3D,$6F,$E5,$79
	.dc.b	$11,$F2,$46,$AF,$94,$E6,$78,$1A,$F9,$2F,$EC,$F4,$FA,$6B,$F4,$F2
	.dc.b	$B9,$C6,$8C,$3F,$E0,$69,$4E,$63,$C3,$E2,$74,$23,$E4,$F4,$EB,$C6
	.dc.b	$2E,$5F,$E1,$7B,$FE,$A3,$3F,$37,$F1,$EF,$F6,$E3,$7E,$33,$FB,$EC
	.dc.b	$62,$3F,$C2,$FF,$5F,$FD,$C9,$FF,$0C,$C6,$6F,$FC,$44,$DD,$BF,$C6
	.dc.b	$6D,$9F,$F1,$C5,$71,$AD,$DF,$62,$4F,$DF,$9D,$5F,$F8,$F1,$74,$7F
	.dc.b	$84,$1B,$3B,$8C,$74,$65,$91,$93,$A4,$71,$FB,$29,$1D,$DF,$F0,$DF
	.dc.b	$D2,$FF,$8E,$F2,$FF,$1E,$60,$9D,$D7,$C2,$77,$8F,$F8,$8B,$F2,$3F
	.dc.b	$E7,$73,$F9,$76,$BF,$E3,$E6,$DA,$7F,$CB,$5F,$0E,$FC,$F7,$F3,$97
	.dc.b	$C7,$CA,$E8,$47,$F5,$73,$BE,$8F,$F1,$8F,$E7,$3F,$6A,$7F,$C9,$6D
	.dc.b	$C7,$FC,$A8,$DC,$7F,$CC,$B9,$7F,$52,$7F,$CC,$F1,$9A,$FF,$36,$35
	.dc.b	$36,$78,$46,$BE,$B7,$C2,$47,$CB,$9F,$C9,$BE,$B3,$99,$F2,$BF,$E4
	.dc.b	$DC,$BE,$04,$FA,$BD,$59,$5F,$BF,$C7,$FC,$A8,$F9,$3E,$6F,$2F,$BA
	.dc.b	$6E,$7E,$E9,$F8,$7E,$17,$FB,$6C,$F6,$96,$BD,$AD,$7D,$B5,$6D,$34
	.dc.b	$93,$FE,$8F,$F8,$CC,$BF,$1A,$B9,$7D,$87,$F6,$B8,$D5,$7F,$A4,$A3
	.dc.b	$11,$FE,$CF,$8C,$79,$7C,$3E,$1F,$8B,$27,$89,$FE,$9D,$AE,$38,$78
	.dc.b	$21,$1F,$E6,$C8,$C2,$FB,$17,$D9,$72,$AD,$C6,$1F,$11,$D5,$C5,$72
	.dc.b	$BF,$5C,$63,$1F,$53,$6B,$85,$F5,$C3,$CB,$5A,$8C,$56,$EB,$C3,$1B
	.dc.b	$BE,$AE,$89,$1E,$43,$DB,$1A,$71,$7E,$D9,$DE,$1E,$79,$B8,$78,$5D
	.dc.b	$E1,$E1,$BF,$2C,$7A,$0C,$DF,$07,$E7,$61,$DF,$0F,$97,$7C,$4E,$51
	.dc.b	$9C,$3B,$4E,$AF,$D9,$7C,$CF,$79,$D8,$F4,$D7,$17,$C5,$3C,$5C,$9C
	.dc.b	$70,$F7,$3C,$63,$DC,$F3,$3F,$DE,$F1,$A6,$E1,$F7,$5B,$7F,$EF,$C7
	.dc.b	$F1,$5E,$63,$E4,$B1,$1F,$8A,$8C,$3B,$B4,$62,$3B,$B5,$FC,$E3,$C3
	.dc.b	$0F,$C8,$B7,$2F,$F7,$8F,$37,$FA,$1F,$87,$F0,$D3,$9B,$F1,$E3,$83
	.dc.b	$F2,$D5,$B8,$FB,$F3,$B7,$CD,$3B,$7F,$0D,$3B,$C3,$C3,$39,$6C,$EB
	.dc.b	$DF,$FE,$5C,$F4,$9D,$63,$EA,$7C,$33,$AA,$E2,$49,$D5,$73,$C7,$8B
	.dc.b	$EA,$3A,$19,$2C,$F1,$74,$FE,$E7,$0F,$45,$74,$8D,$4F,$B8,$E4,$E3
	.dc.b	$2F,$A7,$8F,$82,$7F,$D0,$F8,$F9,$EF,$91,$23,$BB,$F4,$A3,$51,$82
	.dc.b	$1D,$EF,$C9,$1A,$8C,$4F,$1C,$7E,$29,$F5,$BE,$1B,$59,$7A,$A8,$79
	.dc.b	$27,$6F,$D6,$9D,$DF,$0F,$3A,$57,$E0,$3E,$3E,$DF,$DC,$9B,$97,$C7
	.dc.b	$9B,$97,$E3,$1F,$B7,$CB,$F7,$2F,$FD,$53,$18,$C3,$F2,$3A,$59,$F2
	.dc.b	$E8,$AC,$47,$0A,$FF,$26,$4E,$5B,$5A,$8C,$4A,$FC,$1F,$9C,$FF,$BC
	.dc.b	$EF,$FC,$A8,$D9,$D4,$7F,$7C,$E0,$72,$78,$2B,$57,$CB,$E1,$B0,$79
	.dc.b	$FA,$3F,$9C,$D6,$30,$EB,$E4,$71,$39,$7C,$75,$8C,$93,$3F,$09,$F5
	.dc.b	$32,$4E,$E3,$22,$1F,$01,$E5,$3C,$5F,$5C,$FA,$FC,$61,$EE,$9B,$7D
	.dc.b	$B3,$5E,$EB,$47,$FA,$03,$57,$D8,$3C,$BF,$A9,$18,$6C,$3C,$CB,$12
	.dc.b	$B9,$4F,$16,$CE,$71,$F7,$DB,$EB,$1F,$69,$D9,$B7,$E3,$F7,$BF,$B2
	.dc.b	$BE,$58,$CD,$F2,$6D,$FE,$39,$B3,$FE,$A0,$AE,$35,$BB,$FC,$5F,$E2
	.dc.b	$8F,$AF,$A7,$B3,$DC,$B8,$7F,$A6,$67,$4D,$9D,$34,$97,$F5,$A7,$6F
	.dc.b	$BB,$4F,$F7,$54,$EB,$3F,$37,$87,$EF,$3E,$07,$89,$67,$DD,$16,$FD
	.dc.b	$D3,$6B,$F5,$76,$F9,$CF,$A7,$D3,$F5,$1B,$8F,$D2,$6E,$1E,$36,$DF
	.dc.b	$EC,$FF,$19,$7D,$07,$EB,$9B,$7F,$3F,$F0,$9C,$4C,$8F,$F5,$E4,$EB
	.dc.b	$0F,$A5,$59,$7F,$2F,$19,$BE,$D7,$F7,$6B,$51,$EA,$F2,$FF,$17,$F6
	.dc.b	$BF,$8C,$6E,$3F,$8A,$6E,$3F,$2D,$B5,$F7,$FC,$FF,$C4,$1F,$89,$F7
	.dc.b	$2D,$FE,$1E,$2D,$FE,$1E,$6D,$7C,$1F,$1F,$E0,$1F,$9C,$F8,$6B,$87
	.dc.b	$BA,$2D,$FD,$B9,$B5,$C6,$B4,$7E,$82,$32,$FF,$BC,$3A,$1F,$25,$F9
	.dc.b	$DF,$8D,$FC,$F3,$67,$F3,$8D,$C3,$F3,$8D,$C3,$EF,$F8,$7F,$B0,$3F
	.dc.b	$63,$EE,$5B,$FD,$7C,$5B,$FC,$84,$DA,$FE,$61,$F6,$AC,$FF,$95,$B7
	.dc.b	$0F,$D6,$75,$62,$64,$7F,$BE,$67,$58,$7E,$BE,$B2,$FF,$14,$8C,$DF
	.dc.b	$6B,$FD,$D5,$6A,$3F,$A0,$7E,$24,$FC,$27,$F9,$DB,$71,$FE,$21,$DB
	.dc.b	$FE,$C6,$7F,$AA,$D7,$0A,$FE,$83,$E5,$FF,$83,$E3,$FA,$53,$67,$75
	.dc.b	$FE,$27,$78,$9F,$D7,$B3,$FF,$15,$79,$FF,$D2,$BF,$37,$7C,$A5,$3F
	.dc.b	$8E,$FE,$83,$FE,$FD,$BE,$37,$FD,$9A,$B7,$18,$84,$7E,$F1,$FE,$6A
	.dc.b	$DC,$BF,$AF,$FD,$67,$09,$CE,$5E,$1F,$2A,$61,$FC,$BE,$BF,$9F,$BD
	.dc.b	$70,$FE,$7A,$F7,$F4,$DC,$DF,$D8,$9B,$1F,$CD,$FF,$38,$77,$66,$B5
	.dc.b	$F3,$72,$BF,$9E,$7E,$91,$FD,$71,$E9,$1F,$DB,$3D,$D7,$CB,$53,$F9
	.dc.b	$FF,$7B,$C0,$C5,$FC,$1A,$7A,$DC,$4D,$FB,$76,$75,$C7,$D4,$FE,$93
	.dc.b	$F9,$C7,$E8,$A3,$27,$CB,$1F,$29,$FD,$F9,$A7,$F5,$38,$FC,$34,$FF
	.dc.b	$62,$8C,$BF,$D9,$3F,$2E,$79,$DF,$E0,$66,$CF,$F5,$BF,$33,$F0,$D5
	.dc.b	$FE,$20,$9C,$BF,$DB,$3F,$52,$7A,$5F,$E0,$E6,$DF,$DF,$5E,$F6,$67
	.dc.b	$5F,$49,$2B,$F0,$9F,$5E,$F3,$71,$8F,$F7,$88,$C4,$EA,$7F,$C3,$7F
	.dc.b	$A9,$FC,$D8,$C1,$CB,$CB,$FF,$0F,$D7,$E4,$EF,$FC,$03,$EF,$BD,$5C
	.dc.b	$D7,$F8,$92,$31,$7F,$95,$6D,$7F,$82,$3F,$28,$FE,$FC,$F2,$BF,$F1
	.dc.b	$5C,$63,$C4,$FF,$1A,$57,$07,$FC,$6D,$19,$BF,$D5,$4E,$5F,$D6,$7B
	.dc.b	$7F,$88,$7F,$C1,$1F,$AA,$7E,$E7,$1F,$F0,$BC,$E5,$FE,$FF,$19,$C3
	.dc.b	$FC,$85,$39,$6F,$67,$F8,$3A,$5F,$4B,$FE,$40,$AD,$61,$E0,$AC,$1B
	.dc.b	$9B,$1F,$B5,$9F,$AE,$DE,$2F,$F0,$CF,$E7,$4F,$90,$E9,$12,$74,$7F
	.dc.b	$88,$67,$4E,$33,$F5,$FC,$2F,$8C,$F2,$36,$3C,$11,$63,$E0,$97,$4F
	.dc.b	$F3,$01,$FB,$17,$FC,$76,$D3,$FC,$C8,$FF,$9D,$3C,$6F,$F3,$AF,$D6
	.dc.b	$BE,$B3,$F8,$D5,$6F,$0B,$F0,$FF,$E6,$48,$C6,$1F,$C7,$C6,$1C,$AF
	.dc.b	$77,$FC,$66,$7B,$0C,$3B,$3F,$CC,$30,$FF,$58,$C6,$23,$F9,$55,$F3
	.dc.b	$75,$55,$C9,$CE,$FB,$B4,$F2,$4E,$F0,$F8,$57,$D6,$7C,$6F,$78,$CD
	.dc.b	$7F,$B8,$EB,$86,$1E,$48,$C5,$F5,$3F,$CF,$D1,$97,$B9,$9A,$E6,$E2
	.dc.b	$FB,$4E,$E3,$D3,$18,$C3,$D9,$1F,$18,$FF,$78,$E1,$FE,$F2,$70,$78
	.dc.b	$8F,$A1,$F5,$BF,$C6,$BE,$17,$FD,$EE,$D8,$FF,$7E,$35,$DF,$82,$9D
	.dc.b	$FC,$47,$FA,$5E,$76,$D7,$AB,$5F,$EB,$95,$FE,$86,$CF,$89,$F4,$F4
	.dc.b	$FF,$54,$4D,$D7,$FB,$16,$9F,$6E,$F8,$4F,$C9,$6C,$7D,$FA,$E3,$3F
	.dc.b	$35,$B1,$C7,$0F,$33,$70,$FF,$65,$FC,$B7,$1C,$7F,$B8,$37,$0C,$E1
	.dc.b	$E8,$C3,$DD,$3C,$67,$B5,$79,$AB,$FB,$C1,$E0,$38,$5F,$E3,$F1,$F3
	.dc.b	$C6,$A7,$85,$7F,$B7,$23,$83,$FE,$DE,$9D,$B9,$9F,$D5,$EB,$C6,$3F
	.dc.b	$BE,$57,$C0,$BF,$F6,$B5,$F0,$9F,$D9,$5F,$E0,$6B,$E2,$C7,$EB,$DA
	.dc.b	$F3,$3F,$64,$3E,$B7,$FD,$96,$FF,$82,$5A,$FF,$AF,$1B,$7F,$EE,$57
	.dc.b	$FD,$83,$1B,$DF,$F2,$11,$C2,$3F,$7C,$D7,$FC,$3C,$D7,$FD,$FB,$F5
	.dc.b	$BF,$D1,$DF,$18,$70,$F9,$9A,$FF,$09,$6B,$FE,$29,$77,$5F,$EF,$9F
	.dc.b	$AD,$FD,$B6,$7F,$E2,$D6,$D7,$83,$AF,$E6,$BE,$2F,$3A,$BE,$16,$E1
	.dc.b	$FE,$1C,$8F,$F1,$65,$3E,$93,$35,$FA,$FA,$D5,$79,$E3,$04,$78,$F3
	.dc.b	$76,$BF,$EE,$F1,$FE,$37,$AE,$39,$32,$FF,$22,$7D,$6E,$AB,$FD,$DA
	.dc.b	$D9,$D4,$04,$66,$7E,$07,$57,$B2,$39,$4F,$F3,$16,$BD,$B1,$FE,$11
	.dc.b	$18,$8E,$9E,$7F,$FB,$A5,$FA,$AD,$7F,$12,$D7,$F9,$E3,$5F,$F2,$E2
	.dc.b	$F2,$3F,$8A,$36,$39,$C6,$0C,$FD,$6C,$8E,$ED,$7F,$2F,$F5,$BF,$E9
	.dc.b	$67,$F0,$ED,$73,$F0,$7F,$C8,$63,$11,$FD,$3F,$EB,$5E,$EF,$F2,$56
	.dc.b	$C7,$E4,$1A,$FE,$91,$AF,$F5,$7F,$AD,$FF,$39,$BF,$96,$6B,$FA,$A6
	.dc.b	$BB,$F7,$BF,$99,$47,$E9,$A3,$5D,$98,$23,$CF,$DC,$D9,$1C,$1F,$A4
	.dc.b	$7F,$1A,$FF,$69,$6B,$FE,$89,$6B,$F8,$36,$A5,$F5,$9F,$D1,$5B,$1D
	.dc.b	$63,$07,$F4,$9F,$AB,$FF,$37,$C7,$ED,$1A,$FF,$08,$6B,$FD,$ED,$AF
	.dc.b	$FA,$6D,$FF,$38,$BC,$7D,$6E,$84,$7F,$7F,$F5,$7F,$EB,$F6,$FC,$37
	.dc.b	$F5,$98,$C4,$7F,$16,$6B,$FE,$09,$7F,$CE,$6F,$2F,$17,$81,$1F,$E1
	.dc.b	$7F,$57,$FE,$70,$6F,$AF,$FD,$96,$31,$1F,$C8,$9A,$FF,$85,$57,$DE
	.dc.b	$6C,$91,$D1,$FE,$DD,$ED,$3E,$0F,$F8,$4B,$67,$FD,$64,$D5,$DB,$FB
	.dc.b	$85,$F5,$19,$8C,$4E,$E7,$FC,$4C,$D7,$FD,$8C,$D7,$F8,$03,$53,$CF
	.dc.b	$BB,$FD,$99,$5F,$B7,$AF,$03,$63,$9D,$FF,$20,$FA,$B7,$C7,$52,$63
	.dc.b	$FE,$CB,$8F,$E5,$29,$DB,$98,$DB,$EE,$AF,$F6,$E4,$F0,$1F,$49,$DF
	.dc.b	$E9,$25,$75,$8F,$11,$87,$91,$FC,$76,$53,$A8,$F9,$BF,$39,$F4,$08
	.dc.b	$F9,$FE,$AE,$BF,$D7,$B5,$4F,$D8,$24,$7F,$50,$8C,$BF,$EE,$9F,$AA
	.dc.b	$FF,$31,$E5,$FE,$F5,$6C,$7D,$0A,$FE,$8F,$0F,$AB,$A4,$9E,$5B,$57
	.dc.b	$D6,$27,$83,$DE,$79,$CF,$4F,$86,$EB,$F0,$F2,$BC,$91,$97,$CD,$19
	.dc.b	$7A,$DF,$FB,$8D,$FE,$89,$C3,$F8,$FF,$AB,$7D,$5F,$F5,$E6,$CE,$B4
	.dc.b	$FE,$94,$A3,$2F,$A3,$EA,$DE,$B7,$1B,$FF,$27,$56,$AB,$FC,$F8,$D7
	.dc.b	$9B,$5F,$EA,$CF,$4C,$3F,$45,$7F,$42,$7B,$6D,$D5,$F2,$1F,$D7,$61
	.dc.b	$F0,$3F,$DD,$97,$CB,$53,$11,$27,$88,$9D,$B8,$6B,$ED,$8E,$9F,$41
	.dc.b	$23,$4F,$7F,$A0,$C4,$7C,$38,$B8,$FA,$A7,$FC,$13,$F5,$7F,$EB,$9A
	.dc.b	$F0,$B5,$FC,$0B,$67,$F8,$17,$D5,$3E,$2B,$8F,$C1,$7F,$9E,$6B,$85
	.dc.b	$FF,$A9,$DA,$FD,$66,$BF,$E0,$17,$A7,$0F,$DA,$BF,$DB,$9F,$B8,$8C
	.dc.b	$CF,$F8,$1E,$3F,$66,$FC,$66,$BF,$96,$6E,$1E,$6E,$9F,$F1,$2C,$62
	.dc.b	$FB,$75,$E0,$CC,$EB,$EA,$A6,$1D,$5F,$F1,$6F,$D5,$FE,$AA,$18,$27
	.dc.b	$71,$E1,$77,$87,$F8,$D3,$EA,$FF,$A5,$CF,$F9,$61,$AF,$FB,$2D,$AF
	.dc.b	$E6,$9A,$FF,$89,$5E,$39,$7F,$31,$5F,$33,$D4,$ED,$E4,$7F,$90,$BE
	.dc.b	$AB,$F9,$35,$FF,$8B,$5A,$9B,$5F,$ED,$BB,$FE,$79,$1F,$E2,$19,$E1
	.dc.b	$C3,$FE,$B3,$8C,$47,$18,$C7,$C7,$C8,$FF,$30,$CA,$31,$1F,$E7,$68
	.dc.b	$C3,$F9,$C7,$F3,$35,$E0,$E6,$F4,$57,$A7,$E0,$7C,$04,$BE,$AD,$7F
	.dc.b	$DE,$3D,$1F,$C0,$30,$FD,$97,$C3,$FE,$CD,$B1,$9A,$FF,$76,$BB,$2D
	.dc.b	$7C,$56,$BF,$15,$B3,$FE,$8A,$AE,$91,$FC,$3B,$6F,$E1,$F5,$B3,$FC
	.dc.b	$B4,$E2,$BF,$82,$36,$78,$4F,$F1,$6F,$A9,$FE,$F3,$3C,$9B,$19,$F2
	.dc.b	$B6,$FF,$99,$63,$2D,$8F,$3C,$3E,$A9,$F1,$DF,$E6,$BC,$BF,$8A,$4E
	.dc.b	$63,$18,$F5,$C2,$9D,$6F,$8B,$F3,$0C,$4E,$95,$73,$74,$63,$0F,$1D
	.dc.b	$72,$C3,$C2,$DC,$3E,$03,$6B,$D4,$75,$BE,$95,$C3,$0C,$56,$EB,$B1
	.dc.b	$88,$8F,$8A,$B8,$DF,$B6,$31,$5C,$67,$31,$EC,$38,$61,$E8,$BE,$78
	.dc.b	$78,$67,$26,$30,$F4,$CD,$8F,$49,$D9,$C3,$F1,$B0,$F0,$5F,$58,$D1
	.dc.b	$2F,$EA,$C6,$5F,$01,$2F,$D4,$D9,$F4,$35,$F8,$67,$38,$F2,$B4,$F2
	.dc.b	$C7,$AD,$FC,$75,$7D,$E6,$DF,$FB,$63,$0F,$B9,$1F,$93,$F1,$BF,$86
	.dc.b	$BF,$EE,$E8,$FF,$1A,$F1,$EE,$73,$3B,$8C,$1F,$6E,$38,$E1,$FD,$97
	.dc.b	$8F,$FD,$33,$A6,$6A,$5F,$D9,$BD,$E7,$FD,$EA,$F7,$39,$AD,$E1,$F9
	.dc.b	$CB,$DC,$FF,$B2,$6B,$73,$9A,$DD,$FE,$BE,$F7,$6B,$DC,$E6,$B7,$3F
	.dc.b	$AF,$AD,$D6,$A3,$13,$A7,$F4,$73,$AF,$5F,$FD,$B7,$87,$86,$7D,$04
	.dc.b	$7E,$14,$6B,$0F,$04,$E5,$B3,$AC,$7B,$65,$F4,$2B,$47,$BE,$F9,$3E
	.dc.b	$43,$27,$93,$1F,$DC,$65,$E9,$8C,$1A,$AE,$98,$77,$8C,$65,$F1,$A7
	.dc.b	$39,$7D,$3C,$7B,$D7,$2B,$E0,$7B,$CB,$3E,$A3,$86,$5F,$C9,$26,$9F
	.dc.b	$E8,$93,$E9,$52,$35,$1E,$D6,$BF,$5F,$CA,$D3,$C7,$2F,$F3,$FC,$E9
	.dc.b	$B1,$99,$DC,$FF,$36,$9C,$E1,$F9,$39,$F5,$C3,$E1,$EC,$61,$B5,$F2
	.dc.b	$9B,$7F,$C8,$2B,$84,$FF,$42,$8F,$C9,$4F,$B5,$B1,$97,$F2,$11,$95
	.dc.b	$F5,$4F,$77,$DB,$CD,$FE,$AD,$87,$83,$8C,$3F,$CB,$B1,$B7,$FC,$8A
	.dc.b	$E9,$B5,$D6,$35,$7E,$56,$BF,$D6,$1A,$F4,$F0,$B8,$12,$BD,$13,$FE
	.dc.b	$58,$6C,$78,$E4,$7A,$9C,$3E,$CF,$5E,$D7,$18,$FF,$44,$BF,$EF,$BA
	.dc.b	$F4,$FD,$5E,$BD,$8D,$B6,$3C,$31,$EB,$23,$F3,$F3,$7D,$47,$C1,$C3
	.dc.b	$FD,$5F,$18,$D3,$CC,$EB,$B6,$D6,$E3,$E7,$B6,$FF,$D1,$ED,$BF,$F6
	.dc.b	$FA,$AE,$37,$D3,$E9,$1F,$55,$4E,$FF,$2F,$FD,$CB,$1F,$F4,$0E,$4A
	.dc.b	$DC,$61,$B9,$7A,$1B,$3F,$82,$AE,$53,$9B,$F9,$BF,$54,$FA,$5F,$6F
	.dc.b	$E4,$6F,$89,$CC,$61,$B9,$FE,$46,$B8,$36,$7F,$25,$18,$C3,$F3,$4D
	.dc.b	$CB,$FD,$FC,$DA,$F8,$BE,$CF,$4A,$F4,$78,$3F,$0C,$F8,$9F,$1D,$6B
	.dc.b	$DD,$16,$7D,$D3,$70,$FA,$3F,$57,$F5,$3A,$3E,$AB,$70,$F9,$3C,$68
	.dc.b	$C6,$1F,$7A,$31,$3A,$AF,$C8,$FD,$5B,$EA,$FF,$03,$F8,$6B,$EB,$33
	.dc.b	$18,$6D,$FE,$D6,$B8,$36,$7F,$6D,$18,$C7,$F7,$CD,$CF,$F8,$9C,$23
	.dc.b	$15,$FE,$D8,$E5,$7D,$2F,$6A,$F4,$79,$7F,$92,$CB,$F2,$6D,$BF,$7B
	.dc.b	$67,$71,$F0,$A7,$F2,$1E,$E4,$8F,$C5,$7B,$96,$BF,$15,$37,$3F,$E1
	.dc.b	$79,$2B,$17,$FE,$E6,$CF,$F9,$3D,$7F,$10,$E7,$FE,$DE,$FB,$23,$33
	.dc.b	$FE,$F2,$9C,$BC,$7C,$3F,$8A,$47,$E5,$37,$25,$F3,$8F,$D8,$FC,$2F
	.dc.b	$D6,$7F,$55,$F1,$9F,$AB,$F9,$3B,$87,$E9,$5B,$87,$E9,$5B,$87,$B1
	.dc.b	$FD,$DF,$89,$2B,$F0,$FE,$8D,$BF,$C3,$4D,$BF,$DF,$CD,$AF,$E3,$34
	.dc.b	$FA,$C6,$7F,$BD,$FA,$DF,$E3,$71,$8A,$FD,$C4,$62,$7F,$9D,$4E,$6B
	.dc.b	$FA,$13,$5F,$F4,$2B,$5E,$EE,$F0,$FC,$4B,$6F,$F1,$1D,$DD,$5F,$F8
	.dc.b	$03,$F5,$AF,$AD,$FE,$90,$D8,$FE,$98,$DC,$3F,$A9,$35,$FF,$4A,$FA
	.dc.b	$3F,$E2,$A6,$F9,$DF,$D4,$63,$11,$F0,$9A,$FF,$3C,$FA,$B7,$C8,$7F
	.dc.b	$57,$6E,$1F,$D6,$DB,$3F,$D8,$5A,$FF,$A9,$17,$87,$C4,$93,$C5,$E9
	.dc.b	$E5,$48,$E8,$DA,$FE,$D9,$E8,$39,$3F,$B4,$36,$BF,$B6,$36,$77,$3A
	.dc.b	$F9,$49,$5A,$6C,$FF,$77,$FA,$97,$D1,$7F,$73,$6D,$7F,$77,$6C,$EE
	.dc.b	$75,$E9,$CA,$D3,$67,$A1,$CC,$91,$CD,$FF,$00,$2F,$8F,$A4,$7F,$C0
	.dc.b	$C6,$4F,$F0,$47,$D4,$3C,$BF,$F0,$43,$63,$31,$83,$FC,$19,$F5,$0F
	.dc.b	$42,$92,$7F,$C2,$51,$96,$FC,$37,$02,$38,$3F,$C3,$1F,$4F,$FE,$FB
	.dc.b	$6F,$91,$FE,$19,$8C,$67,$FE,$21,$6B,$FE,$E4,$7F,$C0,$0D,$F0,$38
	.dc.b	$12,$BF,$C4,$9E,$E3,$D0,$FF,$12,$36,$33,$18,$23,$FE,$10,$EF,$FE
	.dc.b	$63,$3B,$BF,$E1,$F3,$FB,$CF,$A7,$78,$5F,$E1,$4A,$D5,$FF,$38,$FA
	.dc.b	$6F,$BA,$FF,$8C,$3E,$9D,$F1,$1F,$E1,$A9,$CE,$3D,$E7,$2F,$84,$FE
	.dc.b	$13,$1F,$86,$CB,$F8,$58,$70,$EB,$9B,$FC,$81,$18,$8F,$F1,$4B,$D9
	.dc.b	$B1,$FE,$18,$AF,$EF,$95,$DE,$72,$D7,$B7,$CF,$7F,$87,$BE,$9F,$FA
	.dc.b	$8C,$74,$8C,$7A,$5E,$B7,$C2,$4A,$FF,$2A,$62,$8F,$2F,$BF,$27,$2D
	.dc.b	$BF,$F2,$F5,$7F,$98,$1F,$9E,$7A,$A3,$E4,$3F,$25,$FF,$21,$E8,$F8
	.dc.b	$5F,$F3,$03,$6F,$FC,$C8,$DA,$FF,$1E,$4E,$8E,$27,$F9,$31,$FF,$27
	.dc.b	$F7,$FF,$94,$3E,$99,$F4,$BF,$CE,$AF,$91,$FE,$71,$9D,$46,$6F,$FD
	.dc.b	$13,$5C,$2F,$EB,$D6,$9F,$C6,$4E,$AB,$F8,$8C,$2B,$83,$FD,$72,$74
	.dc.b	$F2,$35,$3D,$F0,$E9,$1F,$D3,$AF,$FC,$CF,$F2,$5C,$B1,$F8,$0F,$49
	.dc.b	$F6,$87,$58,$7F,$A8,$1D,$42,$BA,$D7,$C9,$C7,$31,$8B,$F2,$D7,$84
	.dc.b	$F2,$9E,$EC,$3C,$AF,$82,$BB,$C6,$23,$B1,$FE,$99,$29,$FC,$C0,$C5
	.dc.b	$7F,$A7,$7E,$9B,$F8,$E7,$D3,$DF,$00,$D2,$57,$FA,$9E,$5F,$0B,$FE
	.dc.b	$06,$8C,$35,$FD,$5C,$6B,$3E,$D3,$99,$FE,$F3,$39,$7D,$A7,$E1,$36
	.dc.b	$FE,$FB,$9F,$D8,$DB,$FD,$7D,$3F,$52,$B8,$4F,$DC,$6C,$FF,$81,$0D
	.dc.b	$CF,$46,$CE,$EF,$D0,$DC,$3E,$FC,$E8,$E5,$87,$89,$F3,$D7,$06,$CF
	.dc.b	$09,$D5,$FF,$AB,$27,$93,$FE,$AE,$C3,$CA,$DB,$E4,$EE,$B8,$BF,$04
	.dc.b	$F7,$46,$9C,$1F,$B8,$2D,$FF,$96,$AF,$DE,$FF,$97,$2F,$BC,$EB,$4F
	.dc.b	$89,$E5,$7E,$BD,$B1,$FC,$4E,$F8,$C7,$F8,$B3,$E9,$FF,$80,$4F,$F1
	.dc.b	$76,$BE,$16,$E1,$FE,$6A,$7F,$07,$1B,$DF,$B3,$FE,$4A,$E5,$7C,$86
	.dc.b	$2D,$18,$9E,$07,$F2,$5F,$A5,$FE,$98,$FF,$27,$FA,$6F,$D8,$65,$FE
	.dc.b	$40,$6C,$7F,$B6,$A7,$6F,$C5,$6D,$FF,$9F,$A7,$72,$6C,$EA,$BF,$C9
	.dc.b	$6F,$F8,$5E,$7C,$5E,$57,$A9,$46,$5F,$A5,$4F,$A7,$D4,$E7,$0F,$07
	.dc.b	$47,$1A,$FE,$6D,$7D,$08,$72,$8D,$1C,$B0,$F0,$7D,$3F,$09,$FC,$03
	.dc.b	$6F,$FD,$32,$DC,$BE,$27,$C3,$70,$C3,$D1,$E6,$FF,$B8,$70,$FE,$A0
	.dc.b	$D7,$9C,$6B,$E2,$7C,$DB,$C9,$DC,$72,$F8,$4E,$7F,$4F,$27,$9C,$57
	.dc.b	$F2,$9F,$4F,$F2,$D0,$70,$B1,$57,$CD,$97,$BB,$F1,$FE,$62,$5F,$F6
	.dc.b	$79,$CB,$FD,$A7,$E9,$DE,$CF,$D5,$E7,$7C,$95,$3E,$C4,$EA,$BF,$B8
	.dc.b	$7D,$3F,$F9,$91,$FE,$E5,$5F,$E0,$E6,$F9,$7F,$7E,$75,$1F,$AF,$7F
	.dc.b	$3A,$FE,$C6,$5F,$1D,$E1,$38,$78,$1C,$7D,$9F,$C4,$7D,$83,$EF,$29
	.dc.b	$9C,$3F,$C0,$53,$F8,$7E,$8F,$F4,$DD,$3E,$13,$89,$F9,$19,$D5,$7C
	.dc.b	$C5,$F5,$3D,$B9,$E4,$F4,$8F,$CC,$7D,$4F,$F9,$9A,$3F,$C2,$1F,$62
	.dc.b	$F1,$BF,$38,$D8,$FF,$0A,$4E,$DF,$E1,$CB,$DE,$DF,$CE,$C9,$DC,$7E
	.dc.b	$97,$EA,$BF,$D6,$31,$DB,$A7,$A9,$9F,$89,$D7,$9F,$E0,$F9,$E9,$5F
	.dc.b	$E2,$3B,$FF,$36,$3F,$E2,$5F,$AA,$FF,$34,$3F,$E2,$7F,$B1,$7D,$47
	.dc.b	$EC,$DB,$1B,$F8,$4F,$DB,$53,$EA,$F5,$F2,$52,$BA,$47,$EE,$FC,$0F
	.dc.b	$F4,$EC,$7F,$8C,$DA,$FF,$2D,$6F,$96,$EA,$FF,$67,$FD,$4E,$1F,$F6
	.dc.b	$8C,$DA,$FC,$FC,$FE,$B6,$B8,$46,$3A,$3C,$1F,$15,$23,$FC,$85,$5B
	.dc.b	$C3,$AD,$66,$7F,$9F,$7D,$5B,$E4,$BF,$89,$36,$3F,$C9,$4D,$77,$EB
	.dc.b	$6A,$64,$69,$E3,$1C,$BE,$07,$9F,$9D,$2B,$9C,$65,$EB,$19,$7C,$71
	.dc.b	$F2,$7E,$82,$47,$C9,$7C,$BF,$57,$88,$FE,$AD,$F5,$7F,$09,$BA,$67
	.dc.b	$7C,$70,$30,$FE,$5D,$7B,$9E,$0D,$AF,$9F,$3A,$E6,$98,$7F,$9A,$EB
	.dc.b	$F0,$6F,$83,$83,$2F,$8A,$6E,$1D,$27,$BF,$CC,$F8,$39,$25,$F3,$8C
	.dc.b	$BD,$63,$2F,$B2,$3F,$A2,$7D,$53,$EA,$E9,$27,$FC,$FD,$F5,$7F,$E4
	.dc.b	$37,$FD,$01,$3B,$73,$A7,$F4,$D9,$F8,$51,$AA,$E4,$FF,$7F,$8C,$B6
	.dc.b	$73,$97,$CD,$97,$C6,$73,$F2,$25,$F8,$A3,$B1,$87,$A4,$61,$F3,$FD
	.dc.b	$1F,$8C,$8F,$AB,$A2,$BE,$38,$B8,$78,$67,$FB,$37,$D5,$BC,$3A,$4B
	.dc.b	$FF,$50,$78,$AF,$91,$C4,$7F,$86,$E3,$FB,$A9,$FD,$DA,$B9,$47,$FA
	.dc.b	$9D,$AE,$FA,$BF,$BA,$4E,$EE,$F4,$CA,$EE,$EB,$AB,$11,$DE,$2E,$7F
	.dc.b	$95,$BF,$F0,$94,$F8,$FC,$3E,$A4,$7F,$D7,$1F,$57,$F9,$37,$FA,$EC
	.dc.b	$A7,$51,$C4,$DF,$57,$F8,$27,$EA,$FF,$95,$47,$FB,$05,$FE,$60,$FF
	.dc.b	$8E,$DB,$87,$E9,$DD,$C6,$7C,$8E,$5E,$5C,$8E,$51,$FE,$15,$FA,$BF
	.dc.b	$E1,$51,$FD,$C1,$A9,$B5,$FE,$DE,$BF,$DA,$47,$0F,$89,$77,$BE,$53
	.dc.b	$FC,$73,$18,$7B,$72,$F7,$7C,$1C,$F3,$2F,$F6,$E6,$1C,$5B,$39,$7F
	.dc.b	$CB,$11,$FD,$75,$C9,$C3,$D5,$FF,$57,$75,$FF,$82,$23,$D0,$E1,$F6
	.dc.b	$35,$F6,$45,$9F,$F2,$AF,$C5,$FE,$99,$FD,$3B,$6B,$53,$B6,$57,$FB
	.dc.b	$51,$F0,$57,$06,$CE,$A7,$FD,$71,$1A,$F5,$3C,$06,$6F,$F6,$BF,$55
	.dc.b	$FD,$62,$B6,$7F,$88,$97,$D5,$BE,$4F,$C1,$39,$8C,$3C,$04,$F3,$AE
	.dc.b	$BE,$77,$C1,$8E,$77,$CA,$B7,$7C,$AF,$71,$9A,$DB,$D5,$B1,$E8,$9E
	.dc.b	$18,$FD,$19,$F0,$36,$33,$5E,$7B,$E9,$18,$8F,$3E,$1E,$58,$C5,$F6
	.dc.b	$BE,$57,$E3,$AD,$3C,$89,$7E,$58,$CB,$E7,$BE,$67,$48,$CE,$1E,$9A
	.dc.b	$E2,$FC,$0A,$E1,$18,$BF,$0C,$75,$30,$7A,$CD,$99,$2C,$F5,$6C,$7C
	.dc.b	$5A,$F0,$BD,$27,$47,$84,$AE,$49,$3F,$83,$9E,$2F,$86,$76,$E4,$F8
	.dc.b	$67,$E8,$CF,$BB,$55,$F4,$19,$C3,$EE,$B7,$C8,$FE,$1A,$FD,$B6,$BF
	.dc.b	$93,$6B,$F8,$E6,$C7,$F9,$AB,$97,$F1,$11,$FE,$45,$BF,$EA,$3C,$BF
	.dc.b	$E7,$0E,$6C,$D4,$C3,$F1,$97,$BC,$FF,$75,$7B,$9C,$D6,$F0,$FD,$3D
	.dc.b	$EE,$7F,$D0,$F5,$B9,$CD,$6E,$FD,$4F,$E7,$E7,$51,$8A,$EA,$4A,$F3
	.dc.b	$FB,$BF,$E3,$27,$9C,$66,$78,$46,$B1,$ED,$3F,$17,$0A,$FF,$98,$6F
	.dc.b	$95,$F0,$24,$6C,$FD,$C6,$1E,$47,$18,$78,$09,$3F,$12,$B9,$CE,$A3
	.dc.b	$9B,$F4,$EF,$DD,$3A,$8C,$1E,$93,$81,$F2,$1D,$65,$F6,$63,$47,$0B
	.dc.b	$F7,$38,$C3,$CC,$E7,$0F,$26,$3F,$02,$73,$7C,$EF,$89,$EA,$BF,$AF
	.dc.b	$E5,$F9,$69,$DF,$63,$0D,$C3,$EC,$46,$FE,$37,$17,$F2,$1B,$87,$E3
	.dc.b	$9B,$97,$E2,$9B,$3F,$86,$6A,$FA,$B7,$C2,$7F,$9C,$3A,$FF,$CA,$10
	.dc.b	$F1,$71,$9F,$F4,$D8,$DB,$FE,$D6,$9D,$CE,$AF,$FD,$B3,$5A,$53,$AA
	.dc.b	$E0,$73,$BE,$13,$E3,$BB,$7F,$E3,$4A,$DC,$BA,$DF,$4F,$FE,$E3,$9C
	.dc.b	$CF,$AF,$C7,$F1,$56,$BE,$17,$9C,$4A,$DE,$0F,$7B,$18,$79,$A7,$DD
	.dc.b	$EE,$F8,$BE,$AF,$C6,$D8,$E2,$D9,$E0,$7C,$07,$1B,$7C,$BF,$A9,$F9
	.dc.b	$55,$EA,$84,$62,$B8,$C6,$3C,$2F,$94,$EB,$7B,$5B,$8E,$91,$89,$CB
	.dc.b	$C3,$E9,$FF,$DD,$B8,$77,$6E,$1D,$BE,$77,$FB,$54,$E7,$E0,$F8,$11
	.dc.b	$F3,$46,$A7,$CB,$5C,$B1,$FF,$4A,$7D,$5B,$E9,$7C,$FF,$4E,$F8,$EF
	.dc.b	$B7,$39,$CB,$E0,$4E,$5F,$AF,$19,$C7,$E0,$CE,$5F,$96,$23,$58,$7E
	.dc.b	$56,$74,$F8,$B0,$EE,$7C,$6F,$76,$E1,$F0,$7C,$4F,$88,$7C,$5F,$9E
	.dc.b	$B1,$92,$65,$EE,$8C,$BF,$53,$85,$F5,$7F,$4B,$AF,$E9,$B6,$FE,$E4
	.dc.b	$E9,$B3,$A6,$9F,$96,$FA,$97,$CE,$FD,$EE,$D7,$D5,$7E,$8A,$73,$5F
	.dc.b	$87,$9C,$BC,$6B,$85,$6D,$B1,$92,$65,$F6,$3E,$AF,$51,$E6,$F3,$FF
	.dc.b	$17,$7F,$8C,$6C,$64,$F5,$C6,$5A,$7E,$2F,$DE,$FC,$71,$F8,$FF,$7E
	.dc.b	$C6,$49,$87,$E2,$23,$2F,$DC,$CF,$FD,$E9,$3F,$84,$F2,$BE,$07,$8F
	.dc.b	$F8,$A6,$C6,$4F,$91,$19,$69,$2B,$EB,$7C,$4F,$07,$F2,$FF,$25,$FA
	.dc.b	$3F,$96,$FD,$23,$6B,$F4,$8D,$9D,$C7,$E9,$E7,$F5,$FE,$2C,$8F,$BB
	.dc.b	$EA,$DA,$FC,$24,$DE,$AF,$DF,$4D,$AF,$F3,$97,$D5,$B2,$FD,$A7,$C9
	.dc.b	$7F,$1E,$8C,$4F,$ED,$63,$13,$A8,$FE,$37,$3A,$C7,$F6,$35,$A9,$CD
	.dc.b	$7E,$FF,$B3,$15,$FA,$8F,$AA,$7D,$5F,$F0,$AE,$E7,$C0,$7F,$2E,$9C
	.dc.b	$E3,$FC,$4E,$72,$FF,$2A,$F5,$3F,$C2,$1B,$7F,$43,$97,$9F,$FC,$DA
	.dc.b	$73,$5F,$D0,$67,$2F,$F3,$2F,$AA,$7D,$5F,$F1,$AF,$B1,$71,$7F,$3E
	.dc.b	$9C,$D7,$F2,$39,$CB,$C6,$FF,$90,$FE,$7A,$DF,$84,$61,$E5,$E7,$7F
	.dc.b	$26,$FB,$27,$CA,$63,$0F,$E5,$D1,$87,$FA,$2F,$AE,$E9,$FE,$5B,$F6
	.dc.b	$4F,$2B,$FA,$8C,$E6,$BF,$9B,$4E,$5E,$3E,$B7,$F3,$4F,$B2,$7A,$BC
	.dc.b	$23,$0F,$F5,$AF,$B0,$7D,$67,$F3,$AF,$14,$65,$C3,$BF,$77,$5F,$03
	.dc.b	$72,$FE,$8D,$3F,$EA,$07,$FA,$ED,$FC,$37,$FA,$57,$D9,$79,$F8,$FF
	.dc.b	$93,$CE,$5F,$E9,$71,$9E,$C5,$7F,$B1,$4F,$89,$3A,$F9,$CF,$EE,$31
	.dc.b	$89,$FF,$01,$52,$7F,$BA,$FD,$8F,$C3,$BF,$EA,$FE,$8F,$F3,$2C,$3F
	.dc.b	$DA,$13,$9C,$3F,$C1,$53,$9A,$FB,$5F,$62,$F6,$BF,$BF,$F3,$3E,$8F
	.dc.b	$FC,$03,$39,$CB,$FB,$DC,$E5,$FE,$FB,$0F,$92,$DF,$D2,$70,$6C,$FF
	.dc.b	$80,$BE,$C7,$FD,$E4,$F0,$8F,$F0,$6C,$63,$2F,$E9,$B7,$CE,$B8,$79
	.dc.b	$BF,$E1,$2F,$B1,$79,$FC,$23,$0F,$F8,$5B,$E2,$9E,$E6,$6F,$FC,$23
	.dc.b	$18,$79,$98,$8E,$07,$5F,$42,$C7,$F8,$63,$EC,$3F,$0E,$FF,$81,$7E
	.dc.b	$B1,$F2,$BF,$E2,$49,$CE,$BF,$E0,$69,$CD,$71,$ED,$FF,$03,$C6,$27
	.dc.b	$FB,$4D,$70,$E6,$FF,$1E,$4E,$5F,$C1,$CF,$EB,$67,$57,$FA,$F8,$46
	.dc.b	$1F,$F1,$CD,$7D,$B7,$37,$E5,$E2,$FF,$0B,$FD,$8F,$F5,$89,$F0,$C6
	.dc.b	$3D,$DD,$63,$FE,$21,$AD,$65,$E2,$BE,$F7,$BA,$F8,$9E,$87,$F8,$8B
	.dc.b	$A1,$F2,$1F,$0E,$30,$D9,$D9,$F2,$EF,$07,$F8,$2B,$0E,$31,$FE,$46
	.dc.b	$7F,$72,$D9,$FF,$28,$C6,$27,$84,$60,$FF,$2A,$BF,$C4,$FC,$6F,$E2
	.dc.b	$9F,$61,$FE,$78,$8F,$F3,$14,$E9,$4E,$BE,$3B,$E6,$5F,$8A,$57,$C5
	.dc.b	$FA,$15,$B7,$9E,$15,$C9,$C2,$F9,$E1,$F6,$AF,$94,$F9,$4D,$1C,$8A
	.dc.b	$72,$7C,$18,$78,$DE,$31,$E8,$37,$19,$AF,$F7,$46,$1E,$38,$C4,$FF
	.dc.b	$BB,$63,$17,$E3,$AD,$CF,$28,$C3,$B3,$18,$7B,$5F,$41,$FE,$F5,$3C
	.dc.b	$45,$AC,$3A,$3F,$2D,$F6,$1F,$D0,$3E,$3B,$83,$63,$FD,$FC,$D7,$F3
	.dc.b	$81,$C3,$CB,$E3,$7F,$D2,$AF,$06,$BC,$23,$FD,$0D,$97,$69,$CC,$F1
	.dc.b	$8F,$AD,$97,$F6,$68,$BA,$7E,$AF,$4F,$BD,$87,$29,$FA,$6D,$8F,$C1
	.dc.b	$57,$19,$FA,$ED,$8E,$38,$79,$9B,$87,$6A,$87,$49,$D5,$79,$63,$9E
	.dc.b	$1E,$38,$FD,$F1,$D4,$D5,$F4,$9D,$9D,$1C,$1F,$9F,$EC,$7D,$4F,$F4
	.dc.b	$59,$FF,$5A,$C6,$3E,$4B,$F3,$FB,$7F,$79,$8C,$4E,$AB,$83,$FE,$A8
	.dc.b	$8E,$7E,$27,$08,$47,$07,$52,$BD,$4F,$09,$F0,$F9,$6F,$82,$CB,$63
	.dc.b	$31,$8F,$27,$2C,$8C,$9C,$A3,$AF,$03,$E1,$6F,$A1,$2B,$DF,$39,$6C
	.dc.b	$E5,$F6,$47,$F8,$35,$C9,$F1,$BE,$8D,$F4,$1E,$A5,$F8,$05,$31,$12
	.dc.b	$78,$46,$5E,$51,$97,$E5,$4F,$27,$66,$8F,$AB,$F8,$93,$E8,$FE,$4B
	.dc.b	$E2,$8C,$D6,$27,$2F,$CE,$FC,$49,$F2,$74,$CD,$7E,$02,$32,$FE,$0B
	.dc.b	$F1,$37,$AB,$F5,$9B,$1B,$AF,$B3,$F8,$9F,$F9,$12,$3F,$12,$D7,$F0
	.dc.b	$8D,$EC,$FE,$AF,$C8,$8F,$F3,$64,$FE,$43,$ED,$2B,$F8,$B5,$F1,$FF
	.dc.b	$86,$6C,$6F,$E2,$BF,$13,$F8,$9F,$F7,$A8,$FC,$CB,$5F,$CA,$37,$1F
	.dc.b	$0D,$72,$9F,$CC,$7C,$8B,$E0,$7F,$26,$D8,$FD,$1B,$5D,$F8,$BF,$98
	.dc.b	$FA,$AF,$F2,$BC,$7E,$A1,$7A,$79,$35,$F0,$5F,$9E,$FA,$3D,$D7,$E9
	.dc.b	$DB,$9F,$C4,$6F,$87,$FA,$F7,$9C,$FE,$D2,$9F,$23,$FA,$86,$C6,$63
	.dc.b	$1F,$31,$FA,$CF,$C4,$DF,$49,$FA,$EE,$EF,$EE,$D9,$FD,$26,$E7,$FB
	.dc.b	$DA,$DE,$3F,$C0,$BF,$13,$63,$FB,$66,$C6,$63,$1E,$B7,$EE,$BF,$13
	.dc.b	$FF,$0F,$CF,$F0,$A6,$BF,$C1,$5B,$97,$E0,$9F,$F4,$DC,$7F,$10,$DD
	.dc.b	$C9,$FC,$11,$70,$53,$11,$FC,$23,$F1,$3F,$F7,$DD,$7F,$18,$6B,$FC
	.dc.b	$4D,$B8,$7E,$41,$7C,$17,$F1,$0F,$C4,$F3,$BF,$F1,$D6,$DF,$F1,$86
	.dc.b	$F6,$1E,$31,$FC,$9B,$8F,$FD,$12,$DF,$47,$F8,$DC,$1B,$F1,$1F,$C7
	.dc.b	$BF,$13,$FF,$5D,$D7,$F2,$E6,$BF,$CA,$5B,$A7,$E9,$9F,$E5,$AF,$F8
	.dc.b	$B7,$A3,$E4,$E7,$FE,$3D,$7E,$4D,$27,$5F,$2B,$FC,$C6,$13,$C1,$FF
	.dc.b	$59,$35,$CC,$6E,$39,$CF,$7E,$3C,$19,$99,$5F,$EB,$B9,$D3,$FD,$23
	.dc.b	$E4,$BF,$BA,$C7,$EF,$9B,$87,$F4,$96,$CF,$F4,$45,$F2,$9F,$D0,$9C
	.dc.b	$EF,$FC,$26,$7F,$A6,$46,$0D,$BF,$D5,$BF,$13,$7D,$F7,$F4,$A6,$C6
	.dc.b	$63,$1F,$51,$96,$46,$4E,$51,$F1,$FE,$C7,$FA,$14,$7C,$BC,$BF,$CD
	.dc.b	$51,$8C,$FF,$CB,$CD,$8F,$F3,$77,$D5,$BE,$C7,$97,$A5,$30,$F8,$D5
	.dc.b	$CA,$3E,$A5,$6E,$72,$7F,$B9,$6F,$A3,$FE,$E7,$87,$C3,$70,$3A,$47
	.dc.b	$94,$FF,$3E,$F9,$DF,$C2,$1B,$DB,$FD,$C7,$E1,$93,$FC,$2A,$78,$46
	.dc.b	$3C,$5D,$4F,$18,$DD,$60,$CC,$FF,$76,$F9,$DF,$EC,$57,$FD,$FD,$BB
	.dc.b	$7F,$3C,$6E,$BF,$E0,$77,$F9,$83,$7D,$6E,$44,$AF,$F0,$7C,$E9,$0F
	.dc.b	$B1,$F4,$5F,$FA,$35,$FE,$95,$7F,$E9,$D6,$CF,$07,$71,$9C,$7F,$45
	.dc.b	$1A,$9E,$9C,$98,$99,$1F,$48,$C9,$67,$81,$D1,$FF,$0D,$F9,$7F,$A6
	.dc.b	$BF,$F1,$23,$63,$3D,$1F,$E1,$C8,$7C,$3D,$89,$3F,$E2,$E9,$E1,$D1
	.dc.b	$FD,$46,$BE,$C7,$8F,$FD,$EE,$BE,$BB,$FE,$03,$6F,$93,$AF,$52,$47
	.dc.b	$E2,$B8,$95,$BD,$FF,$D8,$B1,$8B,$DD,$7E,$24,$FE,$5B,$E3,$FF,$07
	.dc.b	$1F,$0B,$4C,$32,$3E,$33,$A6,$C6,$5E,$91,$FC,$E1,$C9,$0C,$BF,$B7
	.dc.b	$FA,$BF,$DC,$7C,$8C,$9F,$DE,$46,$5F,$DF,$46,$5F,$F2,$77,$7A,$7F
	.dc.b	$C9,$3F,$89,$7F,$48,$9F,$F2,$CB,$E1,$7F,$CA,$8D,$CF,$FC,$22,$EE
	.dc.b	$33,$E5,$F2,$C6,$47,$28,$EB,$97,$6F,$22,$4F,$68,$CB,$E0,$8C,$BE
	.dc.b	$08,$F2,$F9,$B2,$3C,$AF,$A3,$C5,$C4,$79,$A2,$E5,$E6,$9F,$6F,$2F
	.dc.b	$BB,$8E,$5F,$BA,$32,$FC,$08,$CB,$F0,$23,$E2,$FC,$47,$FB,$53,$6F
	.dc.b	$F3,$D6,$C8,$C5,$7F,$9D,$A3,$0F,$FB,$96,$77,$F3,$3A,$E6,$56,$AB
	.dc.b	$FD,$CB,$5D,$2F,$53,$9F,$89,$FA,$D5,$D7,$0C,$D6,$BE,$C6,$4F,$73
	.dc.b	$C5,$1A,$71,$8F,$82,$77,$1E,$03,$67,$11,$F5,$DF,$C5,$B0,$F0,$C6
	.dc.b	$3E,$65,$F2,$3B,$A4,$62,$7A,$1C,$BC,$2F,$8A,$D3,$84,$6E,$FA,$D7
	.dc.b	$2B,$E9,$7C,$23,$17,$D1,$B5,$C2,$F9,$6B,$FC,$82,$BB,$1B,$8C,$BF
	.dc.b	$C9,$2B,$8C,$ED,$B1,$9B,$F0,$57,$08,$F8,$18,$7A,$0F,$2B,$4E,$CD
	.dc.b	$C3,$D7,$5C,$A3,$18,$77,$BE,$98,$7C,$D8,$C4,$EB,$0F,$35,$6A,$F9
	.dc.b	$47,$62,$61,$D9,$E0,$79,$8C,$DF,$B6,$78,$CF,$C7,$BD,$E1,$E1,$BE
	.dc.b	$F1,$8C,$3C,$11,$63,$77,$CB,$0F,$3C,$F6,$3E,$41,$F0,$CF,$A4,$78
	.dc.b	$CB,$7F,$69,$B7,$F8,$77,$3F,$0F,$FE,$14,$6F,$9B,$C9,$FB,$11,$97
	.dc.b	$EC,$35,$CC,$7E,$4E,$3F,$C3,$B9,$7E,$63,$E0,$3F,$CF,$FE,$0E,$6A
	.dc.b	$5F,$E2,$EF,$79,$7E,$26,$F7,$39,$AD,$E1,$F9,$6B,$DC,$FF,$6E,$AD
	.dc.b	$CE,$6B,$77,$E6,$7F,$59,$3A,$8C,$57,$52,$57,$9F,$A1,$8F,$29,$F6
	.dc.b	$9A,$AF,$34,$EA,$30,$6A,$BC,$4F,$EC,$AF,$C0,$E6,$BC,$33,$A3,$C1
	.dc.b	$7D,$E7,$F8,$0E,$3E,$E6,$E3,$EA,$C7,$C1,$7C,$70,$E6,$4A,$F8,$93
	.dc.b	$A8,$C4,$72,$3C,$E7,$FB,$92,$34,$F9,$48,$6E,$7C,$AF,$0C,$7D,$6E
	.dc.b	$6B,$E1,$E3,$E9,$9C,$E6,$CB,$D5,$18,$7E,$CE,$FF,$CC,$BE,$3F,$C2
	.dc.b	$FF,$34,$FB,$3B,$C8,$9B,$18,$6E,$3F,$CB,$27,$4D,$9D,$39,$9E,$7E
	.dc.b	$D6,$34,$FE,$61,$87,$86,$B4,$73,$9D,$6B,$FD,$02,$5F,$96,$FE,$12
	.dc.b	$BF,$CB,$F1,$8F,$A4,$FC,$17,$5F,$E0,$E3,$73,$83,$FC,$86,$ED,$E3
	.dc.b	$5C,$F1,$FF,$31,$4E,$5B,$19,$AE,$8A,$FA,$4A,$BC,$BF,$1B,$86,$E1
	.dc.b	$DB,$6C,$60,$AA,$49,$D1,$E4,$77,$C9,$C3,$E3,$2C,$FA,$1B,$8F,$99
	.dc.b	$B9,$F9,$9E,$FF,$13,$E0,$F8,$8B,$7F,$01,$B7,$EE,$6D,$FB,$9F,$57
	.dc.b	$E2,$6F,$91,$BC,$B8,$6E,$1F,$4A,$74,$D9,$D3,$4F,$F2,$67,$87,$FE
	.dc.b	$74,$6F,$57,$CA,$F2,$7F,$C8,$51,$83,$79,$E6,$BF,$D7,$B1,$83,$51
	.dc.b	$FE,$92,$8E,$8D,$AE,$0F,$17,$B7,$E6,$CF,$A1,$BF,$61,$86,$E1,$FE
	.dc.b	$61,$9D,$36,$74,$D3,$F1,$3E,$77,$DC,$3E,$EF,$9B,$63,$24,$C7,$F1
	.dc.b	$71,$97,$FD,$61,$C0,$FA,$C7,$D6,$FE,$32,$31,$F0,$96,$73,$87,$FA
	.dc.b	$FA,$72,$D8,$CD,$F4,$55,$C6,$F9,$36,$B8,$36,$B8,$3F,$99,$F0,$1F
	.dc.b	$3D,$4C,$7D,$5D,$BF,$F3,$2C,$5B,$FF,$32,$CD,$CB,$FD,$0F,$F8,$43
	.dc.b	$CB,$FB,$CB,$CA,$B1,$86,$E3,$FE,$91,$9D,$36,$74,$FF,$AB,$7E,$B5
	.dc.b	$E8,$7E,$27,$E4,$7C,$F3,$98,$C6,$9F,$B6,$D1,$7C,$2B,$47,$7F,$3B
	.dc.b	$51,$99,$FF,$4E,$37,$4F,$E9,$D1,$CB,$0F,$5D,$70,$7F,$4F,$84,$C3
	.dc.b	$F4,$9A,$DB,$FE,$1B,$D7,$E6,$6D,$7E,$BA,$75,$8F,$F1,$9F,$C5,$1C
	.dc.b	$BF,$C2,$DB,$8F,$F2,$05,$CB,$99,$93,$97,$84,$63,$1F,$C7,$FD,$93
	.dc.b	$E8,$7F,$1C,$BE,$92,$C6,$1B,$87,$E3,$E7,$4D,$9D,$34,$D4,$F0,$AE
	.dc.b	$EF,$F2,$3F,$C5,$1E,$D5,$31,$EE,$5B,$FE,$77,$F8,$A3,$DA,$FE,$54
	.dc.b	$BD,$BB,$18,$7F,$97,$7E,$28,$F2,$A9,$8F,$26,$E1,$FD,$1B,$E2,$7F
	.dc.b	$B1,$37,$CF,$FE,$6D,$1C,$63,$07,$F4,$CF,$C5,$3E,$7F,$07,$F4,$47
	.dc.b	$E7,$C6,$79,$3F,$9F,$46,$27,$FA,$4F,$E2,$8F,$8C,$FE,$84,$DB,$FE
	.dc.b	$AE,$FF,$B3,$DB,$1F,$D1,$FC,$87,$CD,$D3,$FA,$57,$99,$FE,$8E,$AF
	.dc.b	$EB,$F3,$A7,$FB,$0F,$E2,$9F,$31,$BC,$9F,$D4,$63,$19,$7F,$5C,$51
	.dc.b	$FD,$57,$F1,$4B,$5F,$DA,$9B,$5F,$D8,$97,$C9,$EE,$BF,$AE,$7E,$29
	.dc.b	$FD,$86,$FF,$B8,$B6,$FF,$B4,$BF,$39,$BE,$CF,$F6,$3F,$8F,$7C,$25
	.dc.b	$F8,$5C,$37,$0F,$EF,$7F,$8A,$7F,$B2,$DB,$E4,$FF,$6B,$8F,$E1,$51
	.dc.b	$83,$FC,$03,$F2,$DF,$D1,$32,$FE,$F2,$DF,$1B,$FC,$05,$F8,$A7,$EB
	.dc.b	$DF,$F0,$24,$66,$BF,$C0,$72,$F6,$7F,$BC,$3B,$8C,$C6,$3C,$AF,$EF
	.dc.b	$9F,$8A,$79,$71,$FF,$03,$B7,$1F,$F0,$83,$E6,$8C,$E7,$FE,$03,$F4
	.dc.b	$7E,$9D,$7F,$85,$7E,$8B,$E1,$E5,$FE,$1A,$9C,$BF,$E1,$68,$CD,$FF
	.dc.b	$85,$3F,$14,$FE,$77,$3B,$F1,$FF,$C1,$D1,$89,$FF,$0C,$DF,$27,$FC
	.dc.b	$29,$1F,$D6,$9E,$5E,$B7,$0F,$6B,$8B,$FE,$17,$FC,$51,$F4,$97,$EF
	.dc.b	$70,$DC,$FF,$C5,$F3,$A6,$CE,$9A,$6A,$78,$61,$FE,$2C,$AE,$0D,$AF
	.dc.b	$F1,$37,$E1,$EF,$BC,$FF,$14,$7D,$27,$F7,$AE,$0F,$F1,$D4,$E7,$2F
	.dc.b	$F1,$FC,$E5,$FF,$1D,$C6,$70,$FF,$1C,$CE,$5F,$F1,$A4,$66,$BF,$C8
	.dc.b	$93,$9F,$88,$7E,$21,$F7,$EF,$FC,$74,$D8,$CF,$B1,$FE,$3B,$8C,$13
	.dc.b	$9B,$11,$97,$FC,$81,$F8,$77,$EF,$6F,$FC,$96,$FD,$3B,$39,$C7,$FC
	.dc.b	$AC,$E4,$C6,$7F,$E5,$67,$1F,$21,$D3,$97,$FA,$D6,$01,$F1,$97,$EB
	.dc.b	$DF,$61,$FF,$00,$72,$6F,$FE,$A5,$9C,$DF,$FB,$52,$F8,$5E,$EB,$4F
	.dc.b	$FB,$3E,$33,$3D,$EF,$9E,$78,$8F,$11,$E2,$18,$76,$53,$C1,$CD,$F5
	.dc.b	$69,$D7,$0F,$0B,$E5,$32,$72,$72,$F8,$30,$F1,$57,$5B,$F0,$4F,$C5
	.dc.b	$9C,$D7,$08,$C1,$D2,$B8,$5F,$8E,$FD,$11,$89,$F1,$1F,$EF,$13,$DC
	.dc.b	$5B,$C3,$D0,$FF,$7F,$FD,$6B,$E7,$9F,$63,$C6,$30,$D8,$CC,$FF,$BF
	.dc.b	$67,$2D,$F0,$7E,$4B,$70,$E5,$18,$8E,$ED,$7E,$9C,$A3,$8E,$3F,$4F
	.dc.b	$2F,$F3,$D5,$7D,$96,$BF,$3A,$B8,$5F,$DD,$6C,$7D,$29,$DD,$73,$F8
	.dc.b	$DF,$C0,$BF,$7E,$B6,$DC,$3C,$35,$F8,$63,$89,$6B,$8B,$E8,$8D,$DF
	.dc.b	$E7,$6F,$9C,$EE,$3D,$15,$C1,$D5,$70,$55,$C1,$FD,$35,$78,$1B,$5C
	.dc.b	$5D,$D7,$83,$85,$F6,$3F,$EB,$59,$FF,$66,$BF,$EE,$96,$FA,$DF,$AD
	.dc.b	$E0,$FF,$65,$CE,$E7,$FD,$92,$DD,$3F,$DA,$4F,$18,$CD,$71,$7F,$D9
	.dc.b	$11,$BA,$FF,$6A,$FB,$0F,$B1,$D6,$3F,$E1,$D6,$C6,$6F,$FD,$C2,$BE
	.dc.b	$C7,$33,$C3,$5F,$F4,$F5,$7F,$B9,$9A,$FF,$B7,$AB,$93,$FE,$E0,$D9
	.dc.b	$F4,$DE,$05,$F3,$74,$C4,$49,$FF,$78,$CE,$9B,$39,$7F,$DD,$7F,$27
	.dc.b	$FE,$5A,$7F,$DE,$EF,$F9,$61,$BF,$0D,$EE,$8D,$47,$0F,$73,$13,$23
	.dc.b	$8C,$6C,$C3,$DA,$30,$FF,$17,$DB,$FD,$FF,$F6,$3D,$38,$91,$8B,$F9
	.dc.b	$B3,$98,$DF,$BF,$89,$FA,$3F,$4B,$FE,$34,$C3,$A1,$FE,$FF,$70,$FF
	.dc.b	$1E,$AE,$8E,$B8,$93,$AC,$7F,$C0,$1E,$77,$DC,$F3,$3F,$C2,$17,$BA
	.dc.b	$D6,$1F,$CF,$70,$7C,$6B,$E9,$3F,$06,$DB,$EC,$D7,$C4,$27,$8F,$D8
	.dc.b	$FE,$2B,$E1,$8D,$F3,$25,$79,$E7,$2F,$A7,$EC,$5F,$69,$F9,$27,$CB
	.dc.b	$1D,$5B,$81,$DC,$7C,$1F,$B2,$FF,$27,$B7,$DD,$FC,$DC,$62,$3C,$F4
	.dc.b	$8F,$EA,$F1,$F2,$7E,$B7,$F9,$53,$7E,$A7,$F4,$70,$6F,$8B,$FA,$6F
	.dc.b	$8B,$DD,$74,$7E,$27,$2F,$E0,$23,$E2,$46,$38,$BE,$73,$C2,$75,$97
	.dc.b	$F3,$9F,$A4,$FD,$8E,$0F,$A6,$A6,$23,$F6,$BF,$64,$FA,$3F,$DB,$F6
	.dc.b	$FF,$B1,$B1,$E6,$DC,$FE,$E5,$6F,$4F,$C3,$7D,$4F,$FB,$69,$BF,$31
	.dc.b	$FC,$02,$31,$1F,$19,$F1,$3F,$8D,$FB,$2A,$FC,$A6,$E3,$EA,$6E,$1F
	.dc.b	$8B,$7F,$CE,$33,$F9,$5E,$0F,$F5,$73,$78,$3F,$87,$C7,$FA,$BE,$35
	.dc.b	$CD,$FC,$4B,$EC,$9F,$3F,$7D,$E9,$5F,$9F,$9C,$BF,$A1,$FB,$2A,$FD
	.dc.b	$E7,$F8,$8E,$9F,$39,$BA,$7E,$7D,$DC,$7D,$2E,$2F,$F4,$8D,$7E,$AD
	.dc.b	$FE,$B8,$DF,$3F,$F9,$2C,$6A,$3F,$17,$F6,$4F,$8A,$FE,$52,$D8,$FD
	.dc.b	$A3,$5D,$FD,$3F,$F3,$35,$D5,$F2,$1F,$F1,$73,$F2,$1F,$D6,$9A,$F8
	.dc.b	$9E,$11,$B8,$E3,$D3,$89,$91,$FC,$20,$C9,$67,$81,$D7,$3F,$CB,$47
	.dc.b	$53,$55,$FD,$0B,$C3,$FF,$17,$C7,$EA,$1B,$3F,$98,$6E,$DF,$C2,$57
	.dc.b	$D2,$F1,$A8,$7F,$15,$8D,$BF,$C1,$DD,$BF,$C6,$7E,$B5,$F6,$5F,$D4
	.dc.b	$9B,$19,$8C,$7C,$B6,$59,$19,$39,$47,$CD,$FB,$2F,$E9,$B1,$F0,$DA
	.dc.b	$FC,$C6,$E5,$F0,$BE,$41,$F3,$5C,$7D,$4F,$AD,$0F,$B2,$A6,$70,$E9
	.dc.b	$38,$7E,$84,$ED,$C9,$FC,$EB,$EA,$BF,$9A,$F8,$BF,$DC,$DF,$E6,$FD
	.dc.b	$D7,$CE,$FE,$1D,$5B,$7F,$83,$71,$7F,$9C,$6B,$8C,$E6,$B8,$1C,$6F
	.dc.b	$FB,$CF,$D9,$7F,$65,$9F,$E3,$ED,$D7,$F8,$7B,$70,$FE,$80,$FF,$2A
	.dc.b	$9D,$FA,$DC,$88,$FF,$4D,$8D,$57,$E2,$3E,$CB,$CF,$3F,$96,$7F,$8F
	.dc.b	$3F,$E8,$76,$A7,$71,$C7,$1F,$D1,$46,$88,$E0,$CF,$24,$AF,$CE,$4F
	.dc.b	$2C,$7F,$AC,$79,$9F,$9B,$6F,$4F,$F8,$6A,$31,$5F,$D5,$57,$CA,$E6
	.dc.b	$64,$FF,$69,$9D,$F9,$87,$71,$F6,$7C,$8F,$EF,$0B,$81,$1F,$E0,$D8
	.dc.b	$FE,$C7,$18,$F2,$FE,$F4,$F0,$8D,$FA,$B8,$99,$5D,$63,$F8,$27,$93
	.dc.b	$FC,$8E,$3F,$80,$AF,$9B,$C9,$A6,$47,$EF,$DD,$36,$32,$F2,$8E,$2E
	.dc.b	$4E,$0F,$EB,$FE,$1F,$F6,$1F,$0F,$27,$F6,$91,$97,$F6,$D1,$97,$FB
	.dc.b	$5F,$7A,$7F,$C9,$3F,$55,$FC,$FA,$7F,$B7,$BE,$BE,$5F,$EB,$ED,$E2
	.dc.b	$FF,$08,$3B,$8C,$F9,$BC,$B2,$97,$FE,$1D,$AD,$5F,$0C,$FB,$F9,$12
	.dc.b	$B5,$1E,$23,$0F,$8A,$30,$F3,$F3,$FC,$E4,$7B,$78,$B5,$F4,$C5,$CF
	.dc.b	$D3,$3E,$EE,$5F,$7F,$34,$BF,$34,$65,$F4,$46,$5F,$87,$1F,$1B,$E5
	.dc.b	$3F,$D9,$9B,$FC,$26,$CF,$F8,$AF,$54,$63,$9D,$0F,$A9,$C3,$AA,$FF
	.dc.b	$69,$5F,$36,$9C,$EF,$75,$C1,$E3,$F5,$52,$77,$87,$69,$C4,$7F,$C0
	.dc.b	$0B,$FA,$F5,$8F,$8A,$FC,$04,$F5,$3B,$8F,$B1,$FE,$25,$1A,$8C,$7D
	.dc.b	$73,$EC,$B7,$48,$C4,$F7,$07,$3F,$0B,$E2,$B4,$E3,$1C,$2F,$B5,$73
	.dc.b	$BE,$B7,$C2,$31,$7D,$4F,$1C,$F2,$36,$D3,$51,$C2,$BA,$DF,$7C,$7C
	.dc.b	$B5,$E3,$37,$19,$7D,$35,$C6,$76,$D8,$CE,$1E,$5A,$E1,$1E,$43,$9E
	.dc.b	$1E,$46,$93,$0F,$1C,$65,$B1,$9C,$3B,$D7,$4C,$3E,$9C,$62,$FD,$D7
	.dc.b	$C0,$F9,$38,$78,$E3,$33,$BC,$3E,$1C,$E5,$EC,$7B,$B0,$ED,$19,$BE
	.dc.b	$95,$C5,$EB,$5D,$0E,$17,$5F,$95,$1C,$1C,$9D,$8C,$CF,$8D,$FF,$0A
	.dc.b	$E1,$F7,$1B,$87,$E2,$9B,$7F,$E1,$A7,$E9,$FC,$57,$F8,$41,$E2,$D7
	.dc.b	$EA,$B5,$DF,$9B,$C5,$FC,$E4,$7F,$86,$31,$FC,$7B,$71,$FC,$1F,$C0
	.dc.b	$B9,$BF,$D0,$55,$AC,$3F,$3F,$7B,$C3,$A9,$F9,$3B,$E1,$D1,$AA,$93
	.dc.b	$F9,$2A,$E0,$AB,$6E,$67,$77,$DE,$3A,$61,$E2,$C7,$DB,$5A,$6C,$EB
	.dc.b	$1C,$3E,$4B,$E4,$7E,$E7,$0E,$EE,$2B,$8B,$67,$F6,$F3,$A9,$C9,$23
	.dc.b	$A3,$F1,$B0,$F2,$46,$08,$7C,$EC,$3F,$85,$D7,$38,$CE,$8C,$BE,$1B
	.dc.b	$6B,$77,$D2,$B9,$1F,$14,$98,$FA,$E3,$55,$C6,$73,$8F,$9D,$F9,$D3
	.dc.b	$AC,$3E,$8E,$3E,$39,$EB,$87,$3C,$BE,$4C,$70,$6C,$EF,$1E,$D5,$FC
	.dc.b	$A3,$83,$F9,$DF,$D9,$7C,$26,$DF,$F3,$2E,$37,$D9,$65,$55,$8C,$36
	.dc.b	$78,$37,$2F,$E5,$57,$FD,$22,$B9,$7B,$99,$C7,$C5,$8F,$6C,$7C,$15
	.dc.b	$A6,$9F,$D1,$3E,$C8,$FF,$B7,$7A,$BF,$DB,$31,$FE,$A1,$8F,$E6,$F1
	.dc.b	$87,$47,$F3,$AC,$53,$BC,$BF,$DE,$6F,$49,$E2,$6D,$D5,$73,$8C,$CA
	.dc.b	$B9,$FA,$DD,$A3,$16,$F5,$71,$5E,$06,$C7,$E2,$6F,$9C,$E5,$B1,$C5
	.dc.b	$B3,$C3,$83,$8B,$CB,$97,$9F,$C4,$DA,$E2,$DC,$7D,$AD,$CF,$E3,$BE
	.dc.b	$2F,$88,$F1,$FC,$3D,$BF,$0B,$6F,$E2,$B6,$FE,$7B,$EC,$FA,$F7,$C7
	.dc.b	$DE,$6C,$37,$0F,$6C,$E9,$B3,$A6,$9F,$E9,$5F,$0F,$FB,$7B,$7A,$BE
	.dc.b	$67,$95,$FE,$4C,$8C,$1B,$CF,$35,$FE,$C0,$8C,$1A,$8F,$F3,$1C,$74
	.dc.b	$6D,$70,$78,$BD,$BE,$95,$F0,$17,$EC,$B0,$DC,$3F,$D5,$53,$A6,$CE
	.dc.b	$9A,$7D,$AF,$33,$EF,$9F,$80,$F3,$6C,$64,$98,$FE,$4A,$32,$B8,$3F
	.dc.b	$3D,$F6,$4F,$83,$FC,$74,$21,$F3,$D8,$75,$87,$E5,$A7,$53,$9A,$D3
	.dc.b	$67,$55,$C9,$B3,$A6,$9A,$9F,$CD,$F6,$BE,$62,$98,$F9,$9B,$5F,$E6
	.dc.b	$78,$B9,$7F,$9C,$A6,$E5,$FB,$A8,$FF,$4D,$7D,$07,$F9,$2E,$FF,$DA
	.dc.b	$CB,$C9,$FC,$0B,$99,$FF,$4E,$B9,$9D,$F9,$7A,$AD,$B6,$3F,$6F,$4F
	.dc.b	$B2,$7A,$1F,$8C,$F9,$0F,$55,$6A,$72,$FF,$B9,$E3,$35,$F4,$E3,$FC
	.dc.b	$5F,$3A,$8F,$F5,$7C,$EB,$2F,$64,$77,$BE,$DA,$7F,$05,$8C,$61,$DE
	.dc.b	$7D,$BE,$76,$71,$E9,$87,$C5,$F1,$1D,$9D,$70,$FE,$34,$DC,$7F,$89
	.dc.b	$2D,$FF,$69,$9C,$AF,$D8,$BF,$B7,$8C,$57,$AC,$FE,$35,$F6,$4E,$6F
	.dc.b	$E2,$ED,$BF,$E2,$AB,$9B,$33,$27,$2F,$08,$C5,$75,$CD,$F6,$4F,$B2
	.dc.b	$FC,$CA,$FB,$2B,$18,$6E,$3F,$99,$9D,$36,$74,$D3,$53,$C2,$FC,$15
	.dc.b	$C1,$F7,$6C,$FB,$27,$B5,$FC,$D1,$B8,$7F,$32,$7F,$DE,$0D,$8F,$E7
	.dc.b	$3F,$64,$F6,$AF,$B5,$86,$D7,$F3,$AF,$B2,$79,$5F,$D0,$17,$95,$63
	.dc.b	$0F,$F4,$4F,$88,$FE,$85,$AF,$99,$FF,$49,$5F,$F4,$47,$3C,$3F,$83
	.dc.b	$73,$3E,$CF,$B2,$FA,$CD,$E3,$FE,$9F,$18,$AF,$C5,$3F,$66,$FF,$A7
	.dc.b	$FD,$93,$E3,$3F,$AA,$B6,$33,$ED,$62,$7F,$AC,$F9,$0F,$B2,$A7,$F5
	.dc.b	$CF,$33,$FD,$55,$7F,$D6,$A7,$4F,$F5,$CF,$B2,$F9,$BA,$FD,$27,$E6
	.dc.b	$B7,$97,$FB,$2C,$63,$5F,$D0,$7D,$95,$BF,$EF,$0B,$E9,$BF,$B5,$B9
	.dc.b	$C3,$FB,$93,$99,$FE,$D7,$F6,$3F,$DE,$1F,$F0,$03,$FD,$AE,$7F,$06
	.dc.b	$E5,$FE,$F4,$E6,$FF,$61,$F1,$CF,$83,$FE,$F2,$BE,$12,$C6,$1F,$EF
	.dc.b	$BF,$63,$FE,$E6,$C7,$FB,$DB,$71,$FC,$8B,$FC,$67,$0F,$DE,$39,$7F
	.dc.b	$7D,$F2,$DF,$D6,$67,$F2,$EF,$F6,$76,$FA,$BF,$E0,$A8,$C4,$FF,$05
	.dc.b	$FB,$1F,$F3,$96,$1F,$E0,$A7,$F8,$27,$82,$73,$87,$E8,$9C,$FD,$8F
	.dc.b	$86,$BF,$C4,$8D,$9F,$F0,$E3,$E7,$8C,$F9,$9F,$E1,$AF,$47,$ED,$E3
	.dc.b	$FC,$57,$E8,$BE,$35,$7F,$86,$A7,$2F,$F8,$BE,$33,$87,$EC,$BE,$C7
	.dc.b	$FA,$73,$FB,$4A,$DB,$FE,$2B,$8C,$DF,$F2,$19,$CE,$69,$CE,$1F,$E3
	.dc.b	$B6,$DF,$1F,$67,$FC,$5F,$18,$A7,$D8,$BE,$8F,$FC,$66,$BE,$C6,$C6
	.dc.b	$1B,$8F,$F8,$D2,$74,$D9,$D3,$4D,$4F,$0C,$7F,$87,$D7,$07,$F8,$8F
	.dc.b	$D8,$3E,$C7,$FC,$87,$D5,$FE,$13,$DB,$F9,$F4,$E7,$1F,$F2,$1C,$E5
	.dc.b	$FF,$2C,$46,$6F,$F8,$DC,$E5,$FF,$29,$46,$6B,$FA,$44,$E7,$37,$D8
	.dc.b	$FE,$13,$F1,$47,$E2,$9F,$E5,$48,$C3,$7D,$69,$9F,$F9,$52,$7F,$A6
	.dc.b	$36,$3F,$CB,$7F,$66,$F1,$AF,$85,$86,$D7,$F9,$72,$74,$D9,$D3,$4F
	.dc.b	$EE,$BB,$07,$C6,$DF,$B2,$7D,$7E,$67,$FC,$49,$BF,$F8,$26,$73,$19
	.dc.b	$38,$87,$2F,$0B,$E9,$5F,$EF,$3B,$B3,$FE,$88,$BE,$93,$98,$E4,$F8
	.dc.b	$78,$3F,$87,$46,$0F,$05,$72,$7B,$19,$3D,$0E,$63,$8E,$1E,$49,$F8
	.dc.b	$13,$9B,$F0,$C7,$43,$95,$70,$BF,$25,$FB,$23,$13,$B8,$C5,$72,$8F
	.dc.b	$09,$F1,$4A,$53,$E3,$96,$F1,$F5,$EF,$9F,$7D,$9E,$1B,$5C,$E7,$53
	.dc.b	$93,$8F,$81,$FE,$59,$9D,$35,$FA,$D2,$8C,$47,$DD,$8C,$65,$C8,$F9
	.dc.b	$99,$FF,$A3,$A6,$FC,$5F,$E0,$F4,$FB,$B7,$C2,$7E,$83,$63,$F0,$35
	.dc.b	$C6,$7E,$A3,$63,$8E,$1E,$76,$E1,$DA,$A1,$D2,$75,$5E,$76,$C7,$E6
	.dc.b	$63,$A1,$F9,$72,$BF,$9D,$BF,$24,$71,$AE,$B7,$E6,$C3,$D9,$39,$BF
	.dc.b	$C7,$53,$F5,$7F,$F2,$CC,$7F,$9F,$57,$CD,$7E,$57,$7F,$F7,$0C,$62
	.dc.b	$75,$F0,$79,$7F,$D1,$91,$C9,$FF,$1B,$3B,$6D,$6E,$FC,$15,$7F,$32
	.dc.b	$7D,$05,$25,$FF,$A3,$E3,$F9,$73,$63,$2F,$F9,$0E,$38,$BB,$EF,$ED
	.dc.b	$F8,$FB,$C8,$A4,$9F,$F4,$D4,$7D,$9C,$7F,$E0,$03,$B8,$CC,$ED,$FF
	.dc.b	$28,$46,$F4,$FE,$19,$4F,$D5,$9F,$71,$ED,$6D,$FF,$A9,$DA,$FF,$96
	.dc.b	$5F,$CB,$39,$8E,$04,$78,$71,$D7,$84,$5A,$FF,$4F,$1E,$19,$F9,$9E
	.dc.b	$A3,$E5,$29,$2F,$FD,$71,$1F,$E4,$86,$FB,$9C,$08,$EA,$6B,$FE,$6F
	.dc.b	$FC,$83,$EE,$61,$E0,$6E,$7F,$0B,$0F,$F2,$AD,$7C,$48,$FE,$61,$F2
	.dc.b	$7F,$CC,$78,$BF,$0C,$FF,$33,$E3,$B5,$F8,$5C,$7F,$D7,$13,$B8,$E4
	.dc.b	$17,$CF,$BE,$9B,$9B,$67,$73,$A9,$EB,$7E,$0F,$BE,$7C,$05,$25,$F0
	.dc.b	$C3,$93,$FE,$99,$C3,$F1,$1C,$9E,$83,$F0,$11,$E3,$6C,$6A,$3C,$3F
	.dc.b	$90,$7F,$BA,$DB,$F2,$BF,$02,$31,$1D,$5F,$F7,$6B,$E8,$FC,$D7,$FC
	.dc.b	$D0,$DF,$80,$F8,$F1,$88,$F2,$3F,$E8,$E7,$DF,$F8,$4F,$E2,$9F,$43
	.dc.b	$77,$F9,$6D,$E1,$FA,$6B,$EA,$BE,$9F,$EA,$4F,$BD,$A4,$C3,$E5,$C3
	.dc.b	$EB,$BE,$CF,$E2,$1F,$00,$7C,$FE,$0C,$F8,$BF,$7B,$F3,$4F,$8A,$6B
	.dc.b	$0F,$AA,$D7,$E7,$3F,$2B,$F2,$3F,$87,$1F,$45,$BB,$7E,$11,$BB,$7E
	.dc.b	$31,$7B,$7F,$8C,$EE,$7D,$A5,$26,$1F,$84,$8F,$F0,$63,$F8,$6F,$C5
	.dc.b	$DF,$63,$F9,$46,$C6,$63,$1E,$27,$E6,$3F,$1B,$F8,$D3,$F8,$E6,$BF
	.dc.b	$87,$6E,$9F,$9A,$7E,$44,$7E,$53,$CC,$FE,$D9,$AF,$E9,$1F,$F1,$3C
	.dc.b	$67,$D8,$FD,$2C,$6A,$3F,$C3,$1F,$10,$FA,$6F,$E5,$D0,$8C,$3F,$A6
	.dc.b	$31,$1A,$8F,$F7,$84,$65,$DB,$FE,$F5,$36,$66,$38,$7C,$DF,$16,$46
	.dc.b	$4E,$51,$FB,$AF,$CF,$5F,$81,$DE,$A9,$5F,$B2,$9C,$B6,$72,$FE,$B6
	.dc.b	$3E,$3E,$7F,$C1,$9D,$C7,$EC,$7E,$DB,$FC,$0C,$FE,$F1,$FC,$C3,$7B
	.dc.b	$3F,$85,$C6,$A3,$87,$AF,$89,$91,$C6,$3E,$4F,$DC,$7F,$02,$8F,$90
	.dc.b	$D7,$E2,$37,$85,$8F,$F9,$17,$ED,$FF,$1B,$5C,$3C,$8E,$64,$7E,$94
	.dc.b	$69,$E7,$1F,$62,$3F,$23,$1F,$45,$B5,$DA,$3F,$CA,$D7,$E1,$BF,$E5
	.dc.b	$3F,$7E,$FB,$1F,$E5,$2D,$AF,$E3,$ED,$7F,$8C,$2F,$B3,$DC,$C9,$DB
	.dc.b	$CA,$3F,$99,$FC,$F3,$EA,$6F,$D5,$4B,$FE,$57,$39,$6C,$E5,$FE,$4D
	.dc.b	$09,$E7,$3E,$FF,$AE,$95,$F1,$1F,$85,$F8,$9E,$23,$E1,$45,$D7,$E1
	.dc.b	$43,$65,$67,$5F,$A5,$76,$BF,$CF,$95,$A7,$FA,$0F,$DC,$FE,$81,$FE
	.dc.b	$74,$BE,$83,$FA,$8B,$63,$7D,$18,$33,$32,$32,$7F,$4C,$F9,$FF,$EA
	.dc.b	$D8,$FF,$5C,$6C,$7F,$55,$6F,$A2,$7F,$DF,$D1,$C7,$D5,$FE,$C3,$1A
	.dc.b	$27,$CD,$7E,$F3,$87,$9D,$1A,$48,$D5,$49,$FE,$01,$39,$6C,$65,$FD
	.dc.b	$F4,$65,$FE,$B9,$F4,$8F,$B4,$FE,$E6,$D8,$FE,$D2,$D7,$7F,$65,$A9
	.dc.b	$91,$A7,$8C,$7F,$7C,$FA,$B7,$0D,$25,$7F,$74,$8F,$AC,$FF,$6D,$76
	.dc.b	$E6,$36,$E7,$8B,$FC,$0F,$F6,$9F,$AB,$E4,$FF,$02,$B7,$83,$FC,$1A
	.dc.b	$DA,$FF,$03,$3F,$D9,$E7,$7E,$5F,$22,$4F,$F8,$1A,$75,$5F,$E1,$6F
	.dc.b	$CD,$5F,$5D,$FE,$16,$7F,$99,$71,$9A,$6A,$FF,$C1,$ED,$7F,$C1,$0E
	.dc.b	$BC,$7E,$55,$FE,$20,$FC,$27,$F3,$B9,$FF,$87,$9B,$8F,$F8,$99,$B8
	.dc.b	$7F,$87,$DF,$A5,$3B,$9E,$44,$9F,$F0,$FC,$EA,$3F,$C5,$FF,$AA,$7F
	.dc.b	$01,$9F,$F6,$0F,$C9,$E3,$EE,$3F,$D8,$ED,$8C,$9F,$E2,$D9,$E0,$FF
	.dc.b	$B9,$A3,$9F,$7F,$08,$D5,$5D,$56,$8C,$3C,$BE,$47,$FB,$7C,$EA,$72
	.dc.b	$4A,$F0,$C7,$2B,$C4,$78,$4E,$E6,$AF,$B3,$63,$C8,$BE,$E5,$F7,$78
	.dc.b	$7F,$CC,$5F,$07,$EE,$9D,$7A,$BF,$E4,$A7,$B1,$C6,$B8,$1C,$63,$59
	.dc.b	$FF,$38,$BE,$75,$C1,$EE,$F5,$73,$1B,$9E,$3A,$2A,$56,$A3,$11,$E1
	.dc.b	$39,$4F,$07,$6D,$7D,$38,$77,$C3,$CC,$F0,$24,$79,$63,$2F,$9E,$79
	.dc.b	$5F,$86,$3C,$73,$AC,$3D,$17,$C9,$F8,$11,$9C,$3C,$F7,$C6,$FC,$67
	.dc.b	$73,$84,$6B,$0E,$E7,$82,$B9,$4F,$B3,$0D,$BE,$22,$9D,$30,$F1,$BD
	.dc.b	$4F,$54,$6B,$0F,$11,$F4,$A3,$24,$9E,$27,$A9,$E1,$7E,$4A,$F9,$51
	.dc.b	$F2,$0F,$A6,$7B,$0B,$5F,$80,$6C,$FE,$25,$CF,$DD,$7F,$B1,$DB,$F0
	.dc.b	$9C,$9F,$BF,$19,$7E,$4B,$5C,$C7,$E2,$63,$FD,$A1,$87,$F8,$1B,$DC
	.dc.b	$7D,$DF,$F8,$16,$3F,$25,$3A,$D3,$9D,$F1,$3F,$3D,$7C,$38,$71,$8F
	.dc.b	$E5,$EF,$83,$E0,$3F,$43,$3B,$75,$5C,$30,$EE,$F4,$C7,$C5,$8F,$82
	.dc.b	$B6,$4A,$F4,$4E,$A3,$B4,$72,$75,$3F,$A0,$BF,$04,$E4,$91,$D5,$E7
	.dc.b	$8F,$8E,$30,$4A,$FD,$65,$72,$8C,$47,$F0,$3C,$7D,$ED,$C7,$D5,$7C
	.dc.b	$E3,$CB,$5C,$2F,$94,$60,$87,$13,$E3,$E3,$ED,$71,$8F,$A1,$DC,$F2
	.dc.b	$73,$5E,$FA,$E1,$9B,$2F,$2D,$EE,$31,$97,$9B,$0F,$8E,$7B,$AF,$8D
	.dc.b	$FD,$AE,$07,$ED,$3F,$D6,$34,$FE,$56,$DB,$FE,$55,$3A,$F9,$AF,$A8
	.dc.b	$F0,$E6,$B5,$BD,$BF,$9B,$B7,$6F,$F7,$CA,$BD,$D7,$06,$E3,$E0,$6E
	.dc.b	$3E,$0A,$E1,$3E,$4A,$E0,$D3,$FA,$0F,$E8,$9F,$C9,$5F,$F7,$CB,$6B
	.dc.b	$FD,$E4,$B5,$FF,$33,$73,$FF,$99,$F4,$4E,$E3,$0D,$BF,$E6,$13,$AB
	.dc.b	$E3,$18,$9E,$8F,$82,$32,$25,$F7,$EF,$9B,$CC,$63,$DD,$B3,$9B,$FE
	.dc.b	$7D,$8F,$96,$3F,$D9,$B1,$83,$C6,$EF,$04,$EE,$B9,$E1,$D1,$EF,$FB
	.dc.b	$23,$96,$98,$F9,$1B,$97,$5C,$7D,$15,$B8,$C6,$68,$C4,$F8,$27,$E2
	.dc.b	$C2,$B9,$CE,$4F,$85,$F7,$5F,$E4,$7A,$E6,$D8,$D5,$A1,$E2,$FC,$67
	.dc.b	$5A,$7D,$76,$DF,$D4,$6D,$F6,$AE,$1E,$37,$CE,$FC,$0B,$E3,$61,$F7
	.dc.b	$DB,$87,$DC,$7E,$6D,$F8,$9C,$BF,$65,$B8,$9A,$70,$AE,$9B,$7E,$01
	.dc.b	$7D,$CB,$E4,$F0,$DC,$7E,$74,$BE,$AD,$DD,$AB,$95,$AB,$E1,$9F,$CB
	.dc.b	$C3,$85,$EA,$31,$3A,$9F,$CE,$B6,$BF,$17,$87,$67,$A6,$1F,$92,$F6
	.dc.b	$FF,$DE,$31,$F7,$9B,$87,$D4,$5E,$FE,$FD,$F9,$5B,$79,$46,$2B,$ED
	.dc.b	$FD,$9B,$E1,$FF,$3E,$FF,$9F,$2F,$F1,$6E,$5B,$E7,$6A,$76,$D8,$C9
	.dc.b	$F0,$FC,$5F,$96,$7C,$CF,$C2,$2C,$64,$98,$7C,$48,$CB,$F1,$A6,$EF
	.dc.b	$F1,$A6,$D7,$EF,$BE,$CB,$EF,$9F,$75,$C7,$17,$EE,$23,$18,$7E,$F5
	.dc.b	$DF,$5E,$67,$4F,$EF,$BD,$8D,$F3,$1A,$3E,$8F,$53,$31,$FD,$7C,$EA
	.dc.b	$71,$8F,$EA,$C7,$DE,$FF,$BE,$EB,$F1,$7F,$00,$FA,$CF,$E2,$35,$AF
	.dc.b	$81,$FC,$A4,$E6,$BF,$7D,$39,$6D,$F3,$8F,$C5,$FC,$8F,$28,$C7,$07
	.dc.b	$FB,$B1,$B5,$FE,$50,$6D,$7F,$12,$EA,$5F,$1C,$7F,$8F,$FD,$73,$E0
	.dc.b	$AF,$C1,$E1,$B9,$7F,$0E,$9D,$36,$74,$D3,$F8,$9F,$C5,$F8,$38,$8B
	.dc.b	$D3,$FE,$5A,$E4,$C4,$F0,$C7,$33,$A6,$C7,$EB,$FB,$3F,$6E,$7E,$E3
	.dc.b	$B2,$C6,$49,$B7,$EC,$23,$2F,$EC,$A6,$F0,$7E,$CA,$6D,$7F,$4A,$FA
	.dc.b	$57,$C7,$7F,$42,$5F,$3D,$63,$0F,$F4,$6F,$C8,$BF,$A9,$65,$FD,$2D
	.dc.b	$BB,$1F,$EA,$31,$FC,$A9,$CF,$4B,$E7,$7F,$A7,$B7,$2F,$E7,$6F,$F5
	.dc.b	$46,$C7,$F5,$4F,$CC,$BF,$80,$EB,$FD,$95,$B3,$FD,$89,$FE,$06,$FF
	.dc.b	$44,$73,$87,$F5,$FF,$CA,$9F,$39,$FD,$79,$B5,$FD,$29,$7A,$D9,$9F
	.dc.b	$EC,$7F,$99,$3E,$9E,$FE,$1B,$86,$E1,$FD,$4B,$F5,$2F,$D5,$9F,$A9
	.dc.b	$71,$B7,$F6,$C8,$C6,$7F,$D3,$3D,$FF,$8E,$4F,$63,$FB,$8B,$F2,$BF
	.dc.b	$1D,$F8,$53,$FD,$49,$BC,$1F,$E0,$55,$F8,$27,$F7,$87,$39,$7F,$81
	.dc.b	$3F,$33,$B7,$FD,$FD,$BC,$A7,$8C,$FF,$6A,$73,$F9,$6F,$F3,$5A,$FF
	.dc.b	$03,$36,$9E,$AF,$F3,$79,$FE,$CB,$39,$FC,$CB,$F0,$F8,$7F,$74,$5F
	.dc.b	$31,$FE,$0C,$73,$5F,$DB,$1C,$CF,$F7,$A1,$F4,$FF,$E1,$26,$9F,$DF
	.dc.b	$BE,$D7,$F8,$BB,$8A,$FE,$EB,$1F,$C6,$8F,$E3,$7B,$7D,$C8,$CF,$17
	.dc.b	$F8,$72,$1F,$1D,$79,$F0,$DC,$3F,$B0,$FE,$5A,$FB,$8F,$ED,$1F,$97
	.dc.b	$7F,$7A,$C1,$39,$AF,$EC,$F3,$98,$C6,$5F,$D8,$DB,$5F,$E3,$C7,$FA
	.dc.b	$74,$66,$FF,$BA,$46,$27,$5E,$B6,$3E,$4B,$2F,$F7,$9A,$7E,$0F,$4C
	.dc.b	$7C,$B5,$CF,$FB,$CC,$E9,$B3,$A6,$93,$2F,$EF,$13,$B6,$D7,$37,$FB
	.dc.b	$E7,$95,$F2,$F2,$FF,$2A,$B6,$FF,$C8,$6F,$EA,$1F,$F0,$4B,$99,$FF
	.dc.b	$20,$B6,$7F,$91,$6D,$FE,$13,$8E,$13,$DA,$3F,$C2,$DF,$97,$FF,$03
	.dc.b	$8F,$F0,$63,$6B,$FC,$CA,$FE,$71,$39,$BF,$F2,$AB,$67,$FC,$A2,$D9
	.dc.b	$DF,$C3,$F6,$F1,$7F,$C4,$9F,$AC,$B0,$BD,$38,$69,$FE,$24,$8D,$36
	.dc.b	$34,$D3,$FA,$CE,$21,$F1,$B7,$ED,$DF,$2B,$99,$FF,$32,$FB,$5E,$15
	.dc.b	$3A,$7F,$DE,$F5,$C6,$B8,$69,$FC,$2A,$B5,$5C,$EB,$4F,$F9,$DA,$FA
	.dc.b	$4E,$A3,$C9,$3A,$7C,$38,$57,$27,$8B,$0E,$F7,$C9,$FF,$80,$15,$C7
	.dc.b	$0F,$14,$62,$B8,$1E,$1B,$E8,$FC,$23,$27,$80,$A7,$83,$0F,$5D,$74
	.dc.b	$8F,$82,$70,$70,$FB,$CF,$94,$68,$A7,$A8,$A7,$84,$DF,$9D,$FC,$53
	.dc.b	$EE,$32,$D8,$E2,$D7,$7D,$C9,$DF,$9D,$FE,$9C,$9D,$B5,$EA,$D7,$EF
	.dc.b	$57,$F6,$AB,$E2,$7D,$1C,$9F,$57,$F8,$3C,$BE,$DC,$62,$FE,$B5,$F1
	.dc.b	$7E,$4B,$67,$75,$C6,$BF,$18,$71,$AD,$E1,$E7,$AF,$C3,$DF,$1A,$DD
	.dc.b	$F1,$BF,$CC,$47,$1A,$DD,$F1,$8F,$5B,$67,$75,$C5,$FC,$EB,$6F,$B4
	.dc.b	$7E,$8D,$AE,$8F,$85,$1B,$9E,$98,$3E,$E1,$F6,$7F,$A0,$D7,$FD,$A3
	.dc.b	$0F,$9E,$C6,$3F,$EC,$E8,$C3,$FE,$9A,$78,$B6,$B7,$7D,$6A,$D7,$0A
	.dc.b	$FF,$4D,$3C,$1F,$F4,$E7,$DC,$7F,$80,$5F,$F5,$2B,$5C,$F9,$BD,$21
	.dc.b	$F7,$16,$31,$E7,$CA,$E3,$1E,$2F,$B8,$FF,$80,$11,$FE,$B0,$6B,$FE
	.dc.b	$AB,$6E,$5F,$C1,$43,$F7,$5C,$3A,$D3,$FD,$EA,$D9,$FF,$78,$B6,$78
	.dc.b	$ED,$F0,$BE,$E3,$FC,$7F,$5F,$EC,$06,$BF,$EB,$B6,$CF,$85,$FF,$21
	.dc.b	$C6,$FD,$7E,$44,$79,$70,$57,$94,$5A,$E5,$3F,$4B,$E3,$3F,$E0,$05
	.dc.b	$7F,$B4,$1F,$F7,$B3,$7E,$4B,$5F,$29,$B6,$46,$4E,$11,$87,$FD,$A5
	.dc.b	$F7,$1F,$67,$0F,$90,$DD,$7F,$9A,$47,$FB,$86,$FF,$13,$E8,$7F,$37
	.dc.b	$E5,$B8,$7E,$0E,$BF,$9A,$4F,$F3,$BE,$5C,$D7,$0F,$91,$FF,$2D,$5F
	.dc.b	$4C,$7F,$A0,$5F,$FB,$A3,$C0,$D1,$D3,$0E,$FF,$70,$FA,$BB,$33,$0E
	.dc.b	$42,$D5,$B8,$CC,$3D,$8F,$BC,$DE,$4F,$5C,$7E,$17,$2F,$1B,$AF,$B8
	.dc.b	$FF,$16,$DF,$26,$AD,$BD,$F1,$86,$FE,$1D,$F1,$BE,$E7,$FC,$D3,$5E
	.dc.b	$27,$FC,$CE,$DF,$7B,$5E,$57,$CC,$FB,$9F,$C3,$CF,$5C,$FE,$2C,$7F
	.dc.b	$91,$23,$D6,$D7,$47,$C1,$78,$E3,$F4,$BE,$E7,$FD,$FF,$3F,$09,$AF
	.dc.b	$0D,$91,$89,$E1,$F1,$7F,$7B,$EE,$7E,$25,$FC,$D5,$F0,$19,$35,$C3
	.dc.b	$F8,$4F,$B9,$DE,$47,$77,$F4,$9A,$F8,$5D,$FD,$CB,$87,$F1,$AD,$9F
	.dc.b	$B2,$D7,$D0,$BE,$63,$F2,$7D,$BF,$E5,$69,$FB,$EF,$D0,$7D,$CD,$D7
	.dc.b	$F2,$AE,$E3,$E1,$7D,$CB,$DD,$FC,$EB,$FE,$FB,$8F,$C4,$B5,$DC,$7E
	.dc.b	$8B,$EE,$7E,$47,$1F,$E6,$9B,$3F,$93,$6E,$A7,$FD,$D7,$1F,$94,$F3
	.dc.b	$3F,$BC,$D7,$E8,$9F,$F7,$0C,$67,$E3,$BF,$5F,$1A,$8F,$F1,$37,$DC
	.dc.b	$BE,$E7,$F9,$DC,$23,$0F,$E9,$0C,$46,$A1,$F3,$16,$24,$EE,$78,$BF
	.dc.b	$8C,$3A,$4E,$9C,$46,$B6,$FF,$19,$47,$63,$94,$FF,$06,$FB,$97,$C8
	.dc.b	$DF,$91,$95,$FB,$49,$CB,$67,$2F,$E5,$63,$E4,$9A,$7F,$31,$F7,$3F
	.dc.b	$A1,$C3,$F8,$7A,$FA,$0F,$E2,$AD,$9F,$E0,$2E,$F6,$FE,$1E,$D8,$FE
	.dc.b	$10,$ED,$EB,$3C,$BD,$8F,$9B,$F7,$3F,$C2,$E7,$E2,$37,$0F,$82,$DC
	.dc.b	$3F,$CA,$DF,$73,$F8,$D9,$CF,$91,$F5,$A3,$F8,$0D,$FC,$D6,$CE,$A3
	.dc.b	$EC,$5F,$37,$FC,$C9,$3A,$BE,$BE,$1F,$93,$0F,$E6,$7F,$38,$F9,$7A
	.dc.b	$4C,$3F,$90,$C3,$E0,$F1,$1F,$C0,$63,$83,$A9,$E7,$3F,$CF,$BE,$11
	.dc.b	$F3,$57,$E6,$65,$FF,$2C,$9C,$B6,$72,$FF,$0D,$84,$F3,$9F,$85,$F1
	.dc.b	$32,$BF,$2D,$AA,$3E,$37,$C5,$66,$7E,$34,$5E,$0F,$C7,$D2,$B1,$BF
	.dc.b	$D0,$AB,$7F,$22,$7F,$AB,$FD,$CF,$E8,$23,$FA,$1A,$FB,$9C,$9A,$E9
	.dc.b	$93,$A8,$CB,$C2,$3F,$B0,$FD,$CF,$F5,$FC,$7F,$AC,$B6,$7F,$A9,$37
	.dc.b	$C0,$2F,$B9,$E3,$EA,$CA,$FE,$AB,$3B,$34,$FE,$E7,$87,$A2,$69,$2B
	.dc.b	$F8,$0B,$8A,$A6,$22,$CF,$38,$AF,$EF,$67,$BB,$FD,$0F,$EE,$5F,$15
	.dc.b	$FD,$F9,$B1,$FD,$A9,$AE,$FE,$AB,$53,$23,$4F,$18,$FF,$03,$FD,$CF
	.dc.b	$DD,$9F,$EE,$6D,$7F,$A9,$37,$8F,$FC,$12,$FD,$D8,$DF,$17,$22,$3C
	.dc.b	$B8,$BF,$C2,$7F,$73,$FA,$FE,$0F,$F0,$5B,$67,$FC,$08,$DE,$03,$FD
	.dc.b	$0E,$77,$F2,$BC,$89,$5F,$E0,$89,$D3,$FE,$0D,$FB,$97,$D8,$FF,$88
	.dc.b	$9F,$E6,$98,$1A,$6A,$7F,$C2,$2D,$7F,$B8,$3A,$F2,$39,$57,$F8,$B3
	.dc.b	$EE,$7F,$95,$ED,$FE,$2A,$6E,$7F,$E2,$66,$E1,$FE,$1F,$7E,$9C,$EE
	.dc.b	$79,$12,$7F,$C4,$33,$A8,$FF,$17,$FD,$C7,$F0,$39,$FF,$6E,$F8,$38
	.dc.b	$FB,$9F,$F7,$23,$63,$27,$F8,$B6,$78,$3F,$EF,$68,$E7,$DF,$C2,$35
	.dc.b	$57,$55,$A3,$0F,$2F,$92,$FF,$11,$CE,$A7,$24,$AF,$0C,$72,$BC,$47
	.dc.b	$84,$FF,$20,$7E,$64,$FA,$2F,$03,$63,$8E,$3E,$17,$D6,$3E,$E9,$F6
	.dc.b	$1E,$98,$C7,$AA,$F8,$6D,$D2,$31,$3E,$13,$95,$71,$9E,$71,$C2,$77
	.dc.b	$AF,$F4,$09,$E5,$7D,$2F,$85,$6B,$66,$12,$F6,$F7,$6E,$1C,$EF,$8E
	.dc.b	$1D,$C9,$5B,$8C,$BC,$70,$EC,$F5,$6B,$EA,$38,$D7,$9E,$33,$7E,$0C
	.dc.b	$3C,$B5,$B3,$A6,$1C,$23,$33,$B9,$F5,$61,$E2,$C3,$A5,$EE,$72,$53
	.dc.b	$E8,$B6,$BE,$5C,$76,$24,$F2,$7D,$C6,$F0,$F3,$9E,$8B,$F0,$61,$E4
	.dc.b	$8C,$E1,$F4,$6B,$83,$DE,$B7,$1F,$16,$30,$FC,$C3,$E0,$9E,$62,$94
	.dc.b	$CE,$1F,$61,$BF,$73,$FE,$D6,$7E,$2B,$5F,$BC,$D7,$F2,$8D,$83,$FE
	.dc.b	$E0,$C7,$F1,$4D,$C7,$E9,$FB,$AF,$BA,$71,$7E,$62,$B5,$8F,$6F,$86
	.dc.b	$C1,$FA,$2B,$E1,$58,$9F,$C7,$57,$07,$C7,$5B,$BE,$B1,$A8,$E1,$8F
	.dc.b	$91,$D5,$E3,$1F,$7C,$5A,$FC,$EE,$3E,$07,$91,$2F,$93,$FA,$D8,$D4
	.dc.b	$60,$95,$D2,$75,$87,$92,$73,$1E,$F3,$A4,$FC,$69,$D3,$C6,$77,$3E
	.dc.b	$E3,$9E,$1E,$77,$11,$AB,$F1,$D7,$12,$4F,$79,$FF,$1E,$CE,$A3,$13
	.dc.b	$A7,$9B,$E3,$AE,$D3,$B9,$C9,$F1,$AF,$CE,$F2,$31,$3C,$23,$31,$C3
	.dc.b	$89,$F7,$3F,$D5,$AB,$F9,$4F,$E7,$4F,$47,$9D,$6B,$A6,$CE,$6B,$F9
	.dc.b	$7D,$F1,$C3,$F9,$75,$EE,$73,$F7,$7C,$AB,$5A,$FF,$3A,$51,$9C,$74
	.dc.b	$DA,$E2,$76,$8C,$B4,$FE,$7F,$F7,$7F,$C8,$9F,$F7,$CB,$71,$FF,$6C
	.dc.b	$3F,$EF,$B8,$CC,$7E,$53,$AF,$F2,$B9,$2B,$84,$61,$B5,$FC,$BA,$75
	.dc.b	$C1,$C6,$31,$3A,$B5,$3E,$7D,$F6,$9B,$9C,$C7,$FB,$26,$FF,$9C,$E3
	.dc.b	$E2,$8D,$C6,$0F,$03,$B9,$FF,$79,$3A,$6F,$BF,$D7,$0F,$06,$16,$B8
	.dc.b	$61,$EE,$FB,$BF,$EA,$98,$F5,$C7,$CF,$18,$7F,$DF,$CB,$E4,$7A,$39
	.dc.b	$AE,$5D,$58,$33,$36,$FA,$64,$FB,$BF,$F2,$2C,$61,$B8,$B0,$FA,$30
	.dc.b	$F2,$7D,$4E,$A3,$3E,$56,$0C,$CD,$BE,$FF,$77,$F0,$EB,$ED,$B6,$3E
	.dc.b	$C3,$F0,$9B,$B7,$C7,$8D,$63,$F6,$1B,$91,$E1,$19,$9F,$B5,$6F,$B5
	.dc.b	$7C,$B6,$1B,$8F,$CD,$97,$DD,$FF,$81,$63,$13,$AC,$3E,$14,$7E,$0A
	.dc.b	$31,$5F,$95,$6C,$7E,$1A,$B9,$57,$03,$E5,$4E,$9B,$7E,$33,$CF,$3C
	.dc.b	$2F,$A4,$7E,$1F,$DE,$FF,$77,$E1,$F6,$1B,$7D,$73,$FF,$73,$7C,$16
	.dc.b	$FE,$0A,$5E,$DE,$51,$8A,$E1,$F7,$6F,$89,$FC,$A3,$FE,$B7,$BF,$C2
	.dc.b	$B9,$6F,$A5,$A9,$DB,$63,$27,$C8,$F1,$FE,$71,$F3,$FE,$EE,$C6,$49
	.dc.b	$87,$C5,$DB,$EB,$4E,$A3,$1C,$1F,$E3,$AA,$DD,$7C,$FA,$D3,$FC,$13
	.dc.b	$EE,$FF,$06,$7E,$0B,$63,$8E,$BF,$D8,$36,$FF,$42,$DC,$7E,$57,$B7
	.dc.b	$C4,$9F,$4F,$FB,$77,$17,$4C,$45,$BF,$DB,$53,$EE,$BF,$DF,$11,$F8
	.dc.b	$DF,$84,$7D,$DF,$F0,$DA,$D7,$C2,$FE,$5A,$73,$5C,$AF,$8D,$7C,$59
	.dc.b	$CB,$F8,$B3,$8C,$F5,$BE,$95,$C3,$5F,$EF,$75,$A3,$E5,$FC,$A6,$B1
	.dc.b	$FE,$25,$F7,$6F,$86,$BF,$0F,$86,$E7,$FC,$1A,$74,$D9,$D3,$4F,$E1
	.dc.b	$7F,$1D,$E1,$7F,$99,$2F,$63,$F9,$03,$9C,$BF,$98,$B9,$9E,$38,$EA
	.dc.b	$B6,$D8,$FD,$9F,$5F,$EE,$4F,$DD,$75,$D8,$C9,$30,$FC,$4C,$65,$FC
	.dc.b	$5C,$DD,$7F,$71,$36,$BF,$A4,$7D,$43,$E8,$7F,$9B,$2F,$A1,$B1,$87
	.dc.b	$F9,$D7,$DD,$FF,$4B,$E9,$34,$FE,$B5,$9F,$ED,$5D,$73,$BE,$83,$FA
	.dc.b	$13,$FE,$9D,$CB,$F9,$9B,$96,$FB,$BF,$D1,$FE,$EF,$F7,$FB,$7F,$62
	.dc.b	$5C,$D9,$E0,$D3,$FD,$37,$EE,$DF,$3F,$4C,$7B,$77,$2F,$E8,$3F,$76
	.dc.b	$FA,$8F,$EA,$8B,$EA,$6C,$61,$FE,$B1,$F7,$7F,$AD,$1E,$0E,$38,$32
	.dc.b	$FF,$5D,$F7,$FE,$21,$3D,$8C,$72,$FF,$61,$FB,$BF,$EE,$1A,$7F,$52
	.dc.b	$5F,$87,$67,$D9,$D3,$FD,$9F,$EF,$37,$7F,$C7,$DD,$DF,$F5,$A7,$3C
	.dc.b	$DF,$E0,$67,$33,$FC,$97,$EF,$3F,$98,$75,$9B,$EB,$FF,$2F,$C3,$FC
	.dc.b	$11,$1A,$FB,$CF,$C0,$47,$F6,$67,$FC,$55,$7F,$CB,$DC,$B7,$E0,$FF
	.dc.b	$BE,$4B,$EF,$2F,$D2,$61,$B1,$FD,$C3,$EF,$3F,$8B,$1F,$C5,$F5,$FA
	.dc.b	$F1,$9E,$4F,$F0,$34,$3E,$4A,$F5,$E1,$B5,$FD,$67,$EE,$DF,$75,$FD
	.dc.b	$83,$EF,$3F,$BA,$4F,$F8,$62,$72,$FF,$5F,$8C,$ED,$FD,$F6,$72,$FF
	.dc.b	$33,$6E,$3F,$E3,$17,$FA,$24,$66,$FF,$B6,$C6,$27,$51,$8F,$63,$2F
	.dc.b	$F7,$21,$F7,$94,$C7,$CD,$5C,$BF,$B9,$4E,$9B,$3A,$69,$31,$FE,$91
	.dc.b	$3B,$6D,$73,$7F,$A6,$79,$DF,$31,$BA,$FF,$81,$21,$E1,$58,$98,$7F
	.dc.b	$80,$E7,$86,$7F,$E5,$17,$71,$99,$FF,$06,$FD,$E7,$ED,$F0,$FE,$B4
	.dc.b	$FE,$ED,$BE,$4E,$BC,$9D,$B2,$FF,$C2,$15,$F9,$87,$FC,$9A,$F3,$C7
	.dc.b	$FC,$C1,$F7,$9F,$EC,$09,$FF,$0B,$3F,$62,$0E,$70,$FE,$D4,$E4,$C4
	.dc.b	$72,$EA,$FF,$11,$C6,$2B,$FC,$B6,$FE,$0A,$BF,$E0,$0F,$C6,$97,$DD
	.dc.b	$BE,$63,$11,$FE,$43,$F1,$7F,$D8,$F3,$98,$FF,$74,$D7,$18,$C4,$6F
	.dc.b	$0C,$5F,$FB,$B2,$AC,$FF,$9B,$2F,$9C,$E6,$3C,$2F,$83,$2F,$E1,$F7
	.dc.b	$C2,$7C,$E6,$8E,$4E,$5F,$55,$EE,$78,$C6,$CE,$38,$79,$A3,$2F,$A2
	.dc.b	$38,$9B,$8C,$DF,$96,$78,$61,$E8,$AF,$82,$6D,$C3,$E6,$3E,$49,$4F
	.dc.b	$71,$83,$B9,$AF,$47,$FD,$79,$F7,$99,$84,$61,$E0,$D7,$87,$72,$78
	.dc.b	$4F,$F5,$9C,$6F,$A3,$FE,$7B,$BE,$2D,$7C,$0D,$7E,$FC,$F1,$7E,$46
	.dc.b	$4F,$AF,$FC,$1E,$7F,$7A,$31,$3F,$56,$B8,$4F,$E0,$9B,$3F,$8A,$37
	.dc.b	$3D,$2B,$9C,$78,$2F,$CD,$7C,$5B,$1C,$67,$75,$F9,$78,$E7,$5B,$C3
	.dc.b	$C5,$7C,$2B,$57,$C1,$FC,$E3,$70,$F2,$C7,$E8,$5A,$E8,$FC,$86,$1D
	.dc.b	$8F,$35,$FE,$4F,$F3,$0F,$E6,$33,$FE,$42,$6B,$FE,$37,$6C,$7F,$8D
	.dc.b	$27,$F0,$6F,$F8,$DA,$33,$B7,$EB,$9B,$5F,$B3,$6B,$F9,$59,$E1,$CD
	.dc.b	$FA,$5A,$E0,$D9,$E0,$D9,$7D,$E7,$F7,$89,$F3,$4E,$D5,$FF,$24,$72
	.dc.b	$7F,$A9,$67,$83,$FE,$51,$FB,$FF,$EE,$EF,$F9,$81,$AE,$7C,$8E,$9F
	.dc.b	$80,$FF,$40,$8D,$47,$79,$DE,$3F,$E5,$66,$E5,$FC,$3A,$F8,$5F,$7F
	.dc.b	$AA,$E0,$47,$FD,$71,$F7,$9F,$E1,$E7,$C5,$19,$78,$1A,$EB,$C6,$2C
	.dc.b	$3F,$DF,$7F,$A9,$BB,$3E,$33,$6F,$99,$FE,$7C,$F9,$27,$93,$F1,$D7
	.dc.b	$C7,$D3,$11,$67,$84,$57,$FC,$F1,$3C,$9F,$F4,$3F,$D9,$3E,$27,$E9
	.dc.b	$36,$33,$1C,$73,$F4,$43,$CF,$FD,$03,$FE,$FA,$3F,$8D,$69,$FC,$E6
	.dc.b	$33,$1F,$EC,$51,$F0,$3D,$BE,$AE,$47,$89,$FC,$44,$F0,$BE,$EB,$EE
	.dc.b	$D9,$E9,$B7,$FD,$1E,$B7,$18,$8F,$F6,$A7,$D0,$3E,$2F,$F1,$DE,$87
	.dc.b	$FB,$CA,$7F,$D4,$ED,$7F,$D3,$8D,$BF,$F4,$CF,$E6,$28,$E3,$D3,$2B
	.dc.b	$8B,$F9,$99,$78,$7D,$B0,$95,$D9,$D7,$C5,$FB,$BE,$EB,$FC,$59,$3E
	.dc.b	$9E,$1F,$91,$18,$8F,$F7,$63,$73,$F8,$6F,$E9,$63,$CF,$3E,$C2,$3C
	.dc.b	$C4,$BF,$4E,$7C,$4F,$CE,$FC,$B3,$F0,$55,$FE,$F9,$6D,$7D,$06,$DF
	.dc.b	$DB,$5E,$9F,$ED,$FE,$E3,$C5,$5F,$EB,$E6,$D7,$F7,$66,$DF,$F7,$BF
	.dc.b	$CC,$1F,$37,$FB,$FF,$CC,$1F,$2D,$49,$5F,$ED,$AA,$7E,$27,$BF,$82
	.dc.b	$93,$FE,$DE,$B4,$7F,$09,$A7,$89,$F8,$A9,$FC,$D5,$7C,$2C,$FF,$29
	.dc.b	$18,$D3,$F1,$AD,$8F,$CC,$68,$E8,$F3,$7C,$FC,$C3,$C0,$FF,$1B,$E2
	.dc.b	$7B,$3F,$98,$F2,$FF,$93,$F7,$7F,$1A,$F6,$B0,$7E,$73,$CE,$7C,$2F
	.dc.b	$F1,$FF,$B7,$F3,$E7,$FE,$F8,$6B,$FE,$ED,$6F,$07,$FB,$AB,$ED,$FF
	.dc.b	$05,$C1,$FE,$22,$9C,$B7,$8B,$A1,$FC,$AE,$77,$7F,$E2,$BA,$DB,$67
	.dc.b	$6F,$E1,$C4,$F0,$6F,$AF,$CC,$8E,$E3,$FD,$AC,$D8,$CB,$F3,$CF,$F1
	.dc.b	$D4,$7F,$38,$C3,$D5,$F2,$5E,$5F,$9D,$F3,$7C,$6C,$BE,$91,$97,$B4
	.dc.b	$65,$E8,$79,$A3,$E5,$7D,$4C,$8F,$94,$FC,$1F,$C1,$71,$1F,$2E,$2E
	.dc.b	$1F,$2E,$6E,$BF,$2E,$6D,$7A,$27,$E5,$77,$48,$F8,$0F,$CD,$FB,$2C
	.dc.b	$47,$93,$4F,$D2,$BA,$3E,$34,$7F,$5D,$E5,$FC,$0C,$7C,$F7,$FC,$23
	.dc.b	$97,$F2,$C8,$FC,$5F,$B5,$BE,$29,$1B,$8E,$46,$1F,$87,$F9,$83,$E6
	.dc.b	$7F,$B5,$36,$3E,$E3,$5F,$93,$83,$EA,$77,$52,$76,$F9,$E3,$11,$DA
	.dc.b	$2E,$3E,$7A,$E5,$3E,$B7,$DD,$F8,$9C,$8C,$10,$F0,$B9,$7C,$71,$F1
	.dc.b	$7E,$DE,$47,$C5,$7E,$47,$DC,$62,$3D,$31,$6F,$C5,$3F,$03,$EB,$E5
	.dc.b	$7C,$07,$E8,$7D,$46,$23,$E3,$C5,$AF,$54,$DE,$4F,$F4,$6E,$AA,$C5
	.dc.b	$FF,$A4,$F2,$5F,$DB,$FB,$39,$3F,$6D,$FF,$0D,$7E,$61,$F6,$F2,$F8
	.dc.b	$D1,$A9,$FA,$2E,$B2,$DF,$34,$9F,$D9,$5A,$39,$F8,$59,$9F,$D0,$CE
	.dc.b	$5D,$CE,$9F,$DD,$79,$7F,$CE,$1F,$F5,$53,$53,$9F,$87,$7E,$50,$FC
	.dc.b	$03,$FC,$65,$3B,$27,$BD,$FE,$3B,$AD,$CF,$F7,$6F,$D9,$DD,$54,$98
	.dc.b	$FF,$51,$F1,$9F,$73,$FE,$4E,$DF,$FC,$09,$7C,$2B,$4D,$7F,$84,$52
	.dc.b	$31,$C3,$FE,$02,$BE,$55,$FE,$11,$FC,$6D,$F2,$3F,$E0,$86,$C7,$85
	.dc.b	$AF,$4E,$97,$D4,$6E,$A4,$ED,$FF,$08,$CE,$A3,$E5,$F9,$BF,$ED,$4F
	.dc.b	$8B,$FF,$68,$F1,$7F,$45,$9F,$F5,$EC,$62,$75,$1F,$43,$E5,$EB,$F4
	.dc.b	$22,$D7,$18,$7C,$87,$F7,$B8,$C3,$6B,$75,$CB,$ED,$FF,$C2,$D3,$FE
	.dc.b	$01,$6E,$1E,$1C,$3C,$67,$E0,$2B,$87,$D8,$48,$C9,$E3,$2D,$7F,$72
	.dc.b	$6F,$57,$B7,$0E,$CD,$C3,$FD,$E3,$5B,$AF,$4C,$63,$D8,$F3,$56,$E7
	.dc.b	$FD,$F5,$48,$C4,$23,$95,$F6,$C7,$C2,$49,$EA,$72,$BE,$B1,$99,$E1
	.dc.b	$3A,$3A,$4E,$9C,$57,$86,$B8,$61,$E7,$9A,$79,$B0,$F2,$D6,$E3,$8D
	.dc.b	$F3,$C3,$B9,$2F,$94,$F6,$BE,$24,$C3,$E6,$DF,$1A,$F9,$33,$9A,$EA
	.dc.b	$78,$8F,$88,$E6,$FE,$F5,$71,$9F,$35,$EF,$0F,$93,$87,$48,$C6,$1E
	.dc.b	$AB,$E1,$7D,$27,$84,$62,$FD,$D8,$7A,$EB,$E7,$1F,$44,$1D,$4A,$52
	.dc.b	$9D,$4A,$7D,$DE,$AF,$EB,$EE,$A7,$F8,$E7,$8D,$A8,$D3,$63,$31,$C9
	.dc.b	$B5,$F9,$FA,$FE,$D2,$FF,$98,$7E,$83,$F9,$B7,$3F,$E6,$A7,$3A,$7E
	.dc.b	$59,$B8,$7E,$49,$AA,$B8,$BF,$86,$8C,$E1,$E0,$AE,$58,$FE,$A2,$F7
	.dc.b	$8F,$8E,$3C,$04,$C3,$F2,$53,$AA,$E3,$1D,$67,$55,$E9,$F0,$7F,$D5
	.dc.b	$6F,$39,$D1,$1D,$65,$C2,$F9,$C6,$CC,$CE,$E3,$24,$7E,$CE,$1E,$37
	.dc.b	$E1,$61,$CD,$A7,$A3,$0F,$1D,$71,$8F,$AA,$75,$39,$C6,$9E,$37,$E3
	.dc.b	$9E,$17,$DE,$78,$CE,$5D,$CF,$E2,$A7,$58,$FC,$F9,$EB,$8F,$95,$FA
	.dc.b	$D8,$77,$79,$DF,$18,$C5,$71,$9D,$D7,$1B,$7C,$C7,$17,$FC,$EF,$1F
	.dc.b	$6E,$31,$E1,$65,$B1,$93,$F0,$11,$97,$15,$F7,$BE,$1F,$0F,$E0,$2A
	.dc.b	$E7,$F9,$59,$B9,$FF,$B8,$FE,$CD,$F5,$5F,$EE,$AE,$BF,$F6,$EC,$63
	.dc.b	$2F,$ED,$1F,$68,$F5,$B5,$1C,$6E,$E9,$FE,$F1,$BE,$0F,$F6,$86,$E1
	.dc.b	$FE,$E8,$C3,$95,$FF,$73,$9C,$BF,$EF,$5F,$75,$E9,$74,$F0,$7F,$D5
	.dc.b	$6D,$87,$A7,$62,$57,$F7,$D9,$E2,$D9,$CE,$5F,$E7,$6F,$CC,$5E,$A7
	.dc.b	$A2,$11,$8A,$FF,$6B,$78,$FF,$EA,$FE,$BD,$4E,$6B,$95,$E0,$F8,$35
	.dc.b	$BB,$FF,$4C,$56,$F6,$47,$F1,$8D,$1F,$6B,$FE,$1D,$8F,$F0,$A3,$71
	.dc.b	$FF,$07,$AF,$85,$F8,$4E,$6F,$9C,$AF,$9F,$93,$A3,$77,$70,$FF,$78
	.dc.b	$5B,$F1,$DF,$F7,$3C,$61,$B8,$7F,$A9,$67,$5E,$57,$DC,$78,$7A,$F6
	.dc.b	$B7,$1F,$87,$6D,$FF,$34,$8E,$2F,$59,$F9,$DF,$4D,$FE,$EC,$8F,$F5
	.dc.b	$E3,$70,$FF,$16,$FE,$41,$FE,$81,$9C,$F8,$7C,$08,$FF,$8E,$63,$53
	.dc.b	$FE,$B0,$AE,$58,$7F,$A5,$DB,$5F,$D1,$1B,$3F,$1E,$B8,$D6,$EF,$DD
	.dc.b	$F6,$8F,$1A,$9F,$07,$D3,$FF,$78,$65,$F2,$A7,$4D,$9D,$BF,$3A,$77
	.dc.b	$8F,$CB,$AD,$BF,$E4,$AF,$5B,$FA,$CD,$7F,$94,$DF,$F7,$8B,$63,$F4
	.dc.b	$11,$87,$81,$8F,$1B,$2D,$9F,$97,$E6,$7C,$E3,$EB,$7D,$AD,$8C,$92
	.dc.b	$7E,$F7,$02,$75,$1F,$2E,$AE,$BF,$76,$6D,$FE,$67,$ED,$7F,$5B,$7F
	.dc.b	$E6,$47,$F6,$71,$FE,$6D,$73,$5F,$96,$E5,$46,$AB,$87,$6F,$EF,$63
	.dc.b	$15,$B7,$8D,$F7,$EC,$FA,$BC,$32,$3E,$97,$C9,$DC,$7E,$74,$DC,$3F
	.dc.b	$8D,$70,$2B,$17,$F8,$3F,$33,$F2,$67,$F0,$DF,$C5,$9F,$15,$89,$FC
	.dc.b	$AC,$DC,$BF,$D0,$35,$A9,$C9,$23,$F2,$53,$CE,$B8,$63,$FA,$5F,$A8
	.dc.b	$FD,$47,$AF,$FA,$96,$E1,$FB,$16,$E1,$F5,$1B,$5F,$27,$F4,$67,$D2
	.dc.b	$7F,$23,$6D,$FF,$A6,$DF,$25,$FF,$A5,$9C,$B6,$78,$1C,$71,$B5,$FB
	.dc.b	$73,$FC,$53,$F6,$CF,$8F,$FE,$48,$D8,$FD,$07,$D9,$62,$64,$71,$C3
	.dc.b	$FC,$67,$5B,$7F,$41,$5A,$59,$7F,$18,$53,$B8,$C3,$6B,$F8,$57,$02
	.dc.b	$F9,$57,$E2,$FE,$8F,$F9,$65,$7F,$2B,$6E,$1F,$C8,$D7,$C9,$FF,$49
	.dc.b	$73,$0F,$CC,$7F,$DB,$55,$E9,$8C,$7B,$1F,$C7,$FF,$4F,$FF,$75,$CE
	.dc.b	$6B,$75,$FE,$C2,$F9,$9F,$E1,$7D,$66,$9F,$8F,$AF,$F6,$73,$FC,$42
	.dc.b	$5D,$FF,$EB,$E9,$F0,$36,$3F,$98,$76,$BD,$4F,$EA,$93,$9D,$BE,$43
	.dc.b	$FE,$44,$6F,$B5,$FC,$D6,$BF,$83,$46,$30,$EB,$F7,$9C,$8F,$E7,$5F
	.dc.b	$98,$BA,$B1,$3F,$EE,$3F,$B3,$FD,$8B,$FD,$55,$BA,$7F,$4D,$5F,$69
	.dc.b	$FD,$CD,$CC,$7F,$B1,$BC,$EF,$ED,$5E,$D7,$F5,$5B,$FE,$85,$C7,$FD
	.dc.b	$FA,$BF,$03,$1F,$E7,$69,$D4,$7F,$59,$F3,$3E,$43,$FE,$EE,$6E,$BF
	.dc.b	$C2,$5F,$E5,$2D,$F0,$BF,$BC,$7C,$33,$E2,$29,$FD,$5F,$F2,$FF,$E6
	.dc.b	$35,$F8,$F9,$D6,$5F,$7D,$7E,$3B,$C8,$FE,$BD,$EE,$3E,$E7,$15,$BF
	.dc.b	$81,$EE,$6B,$E8,$EC,$7E,$87,$B3,$FB,$EF,$CB,$FF,$85,$E0,$FF,$86
	.dc.b	$7E,$D1,$F5,$D9,$8F,$F0,$87,$E5,$CF,$B3,$A7,$F8,$7F,$77,$CC,$63
	.dc.b	$1E,$D1,$86,$C6,$F0,$E0,$F4,$F5,$E4,$FE,$77,$0F,$EB,$EF,$FA,$99
	.dc.b	$47,$8B,$E5,$A4,$FE,$8F,$3F,$D3,$D7,$0E,$7F,$F5,$7C,$62,$BF,$C5
	.dc.b	$0D,$AF,$D5,$6B,$FB,$9A,$FD,$FF,$D5,$7E,$86,$31,$A2,$31,$E6,$FF
	.dc.b	$8C,$67,$3B,$BE,$5B,$13,$31,$EA,$FF,$84,$E7,$51,$83,$C3,$CB,$CC
	.dc.b	$FF,$0C,$7E,$5C,$FC,$57,$13,$A2,$69,$FE,$1F,$9D,$46,$34,$FE,$2D
	.dc.b	$9B,$C4,$F3,$1F,$E2,$3F,$B5,$F6,$38,$AF,$82,$E7,$D8,$C6,$3F,$A0
	.dc.b	$3F,$C4,$9F,$4F,$FE,$4F,$C4,$DE,$6C,$33,$0F,$E2,$93,$A6,$CE,$6B
	.dc.b	$A1,$FA,$B1,$F6,$9F,$E7,$57,$FC,$65,$F3,$FF,$E4,$68,$C4,$6D,$B6
	.dc.b	$9D,$7C,$7F,$FB,$9B,$C4,$FF,$76,$D6,$A7,$37,$FE,$EC,$9C,$D7,$FB
	.dc.b	$C2,$F9,$63,$8A,$D1,$E4,$9D,$5F,$28,$F1,$46,$27,$84,$60,$F4,$B8
	.dc.b	$7F,$DF,$F7,$AB,$EF,$1D,$4E,$F8,$79,$EB,$84,$2B,$86,$1E,$18,$E6
	.dc.b	$7F,$BF,$CD,$16,$30,$E0,$10,$7A,$5F,$6B,$E2,$B1,$19,$8C,$38,$37
	.dc.b	$A2,$71,$18,$6C,$FF,$06,$CA,$E9,$FD,$A2,$B9,$4E,$5B,$19,$7C,$31
	.dc.b	$9C,$FE,$FE,$8F,$9E,$F4,$46,$1B,$1E,$7B,$ED,$3F,$86,$AE,$53,$EF
	.dc.b	$39,$5F,$D6,$6B,$F4,$4F,$87,$5C,$A3,$A6,$1F,$23,$E5,$7C,$15,$67
	.dc.b	$A6,$3F,$2E,$7C,$17,$EF,$6D,$F7,$9F,$3D,$FD,$2B,$F0,$E3,$A8,$F1
	.dc.b	$63,$E8,$75,$5C,$1F,$F6,$C3,$63,$50,$FA,$2F,$E6,$EF,$F8,$0D,$AF
	.dc.b	$EB,$9B,$87,$F8,$67,$0F,$BE,$FF,$86,$E3,$73,$F8,$26,$BF,$87,$6B
	.dc.b	$F8,$A6,$BE,$1F,$93,$FD,$57,$06,$DB,$1B,$6D,$3E,$D3,$F7,$13,$E7
	.dc.b	$9F,$CD,$59,$A4,$8F,$C7,$FD,$A3,$EB,$3F,$40,$D8,$CC,$60,$8F,$F8
	.dc.b	$97,$ED,$3F,$A3,$B7,$C3,$F2,$AA,$E4,$FF,$14,$5F,$4B,$A6,$C7,$F8
	.dc.b	$BD,$AF,$FB,$3F,$ED,$1F,$69,$BA,$F2,$92,$7D,$2D,$7F,$DA,$DE,$1B
	.dc.b	$ED,$3E,$2B,$63,$B4,$F9,$BE,$C9,$F6,$94,$93,$FE,$41,$8F,$F1,$23
	.dc.b	$63,$F8,$24,$70,$8F,$DB,$B5,$FD,$E3,$5F,$E0,$3F,$69,$FE,$8F,$7F
	.dc.b	$7A,$D7,$F8,$13,$5D,$FA,$9F,$C0,$BC,$2F,$CE,$47,$F0,$76,$BF,$C3
	.dc.b	$1A,$FF,$11,$6B,$FC,$4E,$5E,$77,$F1,$B6,$C7,$8A,$30,$77,$FB,$46
	.dc.b	$14,$93,$E1,$8C,$BF,$C5,$BE,$07,$FD,$C8,$DF,$94,$FE,$51,$18,$8F
	.dc.b	$E3,$CD,$7F,$91,$BF,$84,$6F,$27,$02,$3D,$7E,$63,$99,$1C,$1F,$0F
	.dc.b	$EC,$DE,$07,$C7,$6C,$66,$30,$79,$3B,$24,$7A,$5A,$FC,$BF,$A4,$FB
	.dc.b	$ED,$F8,$0F,$A5,$18,$8F,$9C,$D7,$FA,$07,$D8,$7E,$B5,$FE,$72,$D7
	.dc.b	$F9,$F3,$5F,$E8,$6D,$7F,$A4,$7D,$8B,$CA,$FE,$A0,$D8,$FE,$86,$D7
	.dc.b	$FA,$43,$5F,$E9,$AB,$E8,$78,$92,$38,$BF,$8B,$8F,$C7,$47,$F5,$8E
	.dc.b	$44,$62,$11,$8E,$5E,$07,$93,$E4,$E4,$F9,$1A,$FE,$4F,$C1,$FE,$37
	.dc.b	$5F,$CA,$7D,$2F,$CD,$46,$0F,$E3,$BF,$61,$FB,$EF,$17,$FD,$80,$DE
	.dc.b	$0F,$ED,$4D,$C3,$FB,$6F,$D8,$7D,$C7,$FC,$3B,$1F,$C1,$9F,$ED,$AD
	.dc.b	$78,$F6,$FF,$79,$8C,$47,$F7,$67,$A2,$9F,$AD,$7F,$ED,$16,$DF,$29
	.dc.b	$FF,$61,$47,$F2,$C7,$FC,$5A,$D9,$E1,$E7,$FF,$20,$8F,$F3,$AC,$F0
	.dc.b	$8C,$11,$E7,$ED,$6F,$E6,$78,$7C,$C4,$9E,$F1,$97,$C3,$19,$7C,$31
	.dc.b	$FC,$BB,$27,$C7,$D2,$4F,$A3,$E8,$AB,$FE,$0F,$9D,$BF,$E0,$E6,$B9
	.dc.b	$9F,$90,$7A,$FE,$62,$47,$FB,$8E,$77,$A7,$C7,$F9,$FF,$80,$49,$F8
	.dc.b	$FF,$41,$5F,$8F,$16,$7E,$74,$FF,$46,$F8,$3F,$F7,$0D,$7F,$BB,$DF
	.dc.b	$F0,$4C,$B8,$3F,$C5,$91,$86,$FC,$8F,$13,$97,$A9,$2F,$94,$7F,$52
	.dc.b	$FA,$F7,$D8,$52,$4F,$7F,$84,$AF,$F8,$CE,$76,$FF,$8C,$9A,$E6,$7F
	.dc.b	$C8,$91,$8C,$BC,$D3,$FE,$42,$CB,$FC,$D3,$97,$A3,$D6,$EB,$5A,$A5
	.dc.b	$47,$E0,$7A,$7F,$0C,$F8,$91,$23,$24,$3F,$CF,$AF,$16,$FD,$5F,$CB
	.dc.b	$23,$F2,$FE,$12,$BA,$3E,$74,$7D,$18,$F8,$DF,$29,$27,$FC,$B7,$87
	.dc.b	$9F,$34,$FD,$7F,$A2,$D5,$7C,$78,$F8,$64,$7E,$43,$F7,$75,$91,$D3
	.dc.b	$5F,$F3,$35,$F4,$8F,$C0,$7C,$77,$E0,$C8,$EC,$FF,$36,$FD,$87,$C9
	.dc.b	$CB,$E5,$6B,$FE,$FD,$9E,$24,$3F,$D4,$33,$BC,$3F,$CE,$8F,$E6,$1B
	.dc.b	$CF,$FE,$16,$8C,$4F,$6F,$0B,$13,$23,$C4,$27,$6D,$9F,$F0,$FF,$91
	.dc.b	$FD,$0E,$3F,$D0,$6D,$4D,$27,$93,$FE,$3A,$F8,$07,$E1,$FF,$C2,$2B
	.dc.b	$FD,$89,$18,$F0,$3A,$D7,$F9,$03,$C6,$7D,$8B,$B3,$0D,$C3,$E4,$4E
	.dc.b	$A7,$87,$D8,$FF,$95,$37,$FE,$81,$7B,$BF,$F0,$53,$5F,$F0,$53,$5D
	.dc.b	$D7,$F8,$4F,$EC,$5F,$4D,$49,$3F,$E1,$58,$CB,$63,$D1,$FB,$3B,$E7
	.dc.b	$69,$EB,$AD,$D7,$FB,$63,$81,$F5,$9F,$EA,$BC,$FF,$D5,$31,$89,$FF
	.dc.b	$57,$C6,$27,$51,$FE,$B1,$9D,$3F,$AF,$3E,$8C,$7F,$AD,$7E,$2B,$FB
	.dc.b	$A7,$D8,$FF,$85,$5B,$ED,$67,$E1,$B9,$55,$BF,$B6,$D5,$7C,$67,$FD
	.dc.b	$80,$D8,$D9,$FE,$D4,$8E,$6D,$3F,$D8,$F1,$B7,$9D,$75,$C3,$D3,$87
	.dc.b	$9B,$9B,$E0,$37,$1E,$57,$E2,$9C,$E3,$E2,$F5,$7C,$D9,$27,$35,$E0
	.dc.b	$6B,$C6,$F9,$CF,$F8,$37,$0E,$D3,$FE,$E5,$C3,$DB,$7C,$2B,$A5,$F0
	.dc.b	$C3,$BC,$E4,$91,$9B,$EB,$18,$7A,$1E,$38,$D3,$83,$C8,$5A,$EE,$D9
	.dc.b	$F1,$57,$82,$BA,$92,$78,$46,$5E,$11,$9C,$3F,$03,$39,$7D,$E4,$C3
	.dc.b	$E5,$56,$DF,$83,$56,$3E,$CD,$71,$9C,$C7,$63,$C2,$6C,$F5,$D7,$57
	.dc.b	$89,$9C,$3D,$B1,$98,$FC,$6D,$75,$7E,$31,$D6,$3D,$2F,$BC,$1E,$22
	.dc.b	$9C,$0A,$7C,$C2,$CF,$F8,$23,$C6,$CC,$7E,$53,$AB,$FB,$9B,$A6,$BF
	.dc.b	$9E,$6B,$B8,$FC,$4C,$7F,$7A,$AF,$F3,$07,$D8,$FF,$48,$DB,$35,$2F
	.dc.b	$EE,$5E,$F2,$F3,$DE,$E7,$35,$BC,$3F,$6F,$7B,$9F,$7D,$6E,$73,$5B
	.dc.b	$BF,$33,$FA,$69,$D4,$62,$BA,$92,$BC,$FE,$0F,$F7,$A7,$9C,$66,$78
	.dc.b	$46,$B1,$ED,$3F,$04,$D1,$E4,$AE,$55,$C0,$91,$BA,$E4,$F0,$3F,$85
	.dc.b	$E3,$ED,$6E,$3E,$9B,$E7,$1C,$6F,$9B,$A3,$F5,$18,$FB,$9D,$10,$D6
	.dc.b	$1E,$03,$DA,$49,$E7,$3C,$67,$71,$DB,$2F,$C3,$47,$6B,$F1,$BA,$3B
	.dc.b	$46,$AF,$B5,$74,$FB,$1D,$E1,$FE,$20,$CF,$EA,$46,$3B,$B2,$D8,$C9
	.dc.b	$F6,$63,$2D,$33,$8F,$C0,$8C,$BF,$CF,$E9,$5F,$E3,$4A,$FD,$0C,$61
	.dc.b	$FE,$93,$3F,$90,$9D,$3F,$EA,$3F,$B0,$7C,$CF,$F8,$F3,$C1,$FF,$1E
	.dc.b	$F6,$3A,$1E,$2E,$72,$E1,$77,$1F,$F5,$4D,$6A,$7F,$B9,$E3,$D1,$FF
	.dc.b	$7D,$36,$7F,$DF,$4D,$7F,$DF,$4D,$7F,$D6,$DE,$CB,$D1,$FE,$B9,$1F
	.dc.b	$EB,$68,$C5,$7E,$EB,$2F,$F6,$15,$69,$B1,$99,$DD,$33,$F6,$C6,$1B
	.dc.b	$E9,$7A,$EF,$8D,$6E,$FF,$86,$FD,$83,$D3,$F8,$1E,$2B,$F0,$2F,$F6
	.dc.b	$95,$7F,$8A,$2B,$4F,$F8,$EE,$75,$9F,$C7,$C1,$5A,$7F,$89,$B5,$FE
	.dc.b	$32,$D7,$C3,$87,$B7,$EC,$5F,$43,$E7,$6E,$3F,$E2,$25,$F0,$39,$99
	.dc.b	$39,$78,$46,$2F,$DC,$7C,$5F,$B0,$7A,$DF,$77,$81,$18,$C3,$FC,$AD
	.dc.b	$18,$F1,$F5,$39,$BE,$73,$C2,$F2,$6A,$AD,$FF,$93,$BE,$85,$E0,$FD
	.dc.b	$1F,$CF,$7D,$33,$98,$C6,$7F,$26,$B7,$3F,$CE,$B2,$46,$1F,$E7,$AD
	.dc.b	$7F,$A1,$B5,$FE,$8E,$EF,$2F,$F4,$B3,$67,$FA,$6F,$1B,$EC,$9E,$C7
	.dc.b	$87,$2F,$89,$F5,$8F,$26,$9F,$1F,$B7,$FC,$77,$8F,$C6,$9D,$36,$76
	.dc.b	$FC,$A9,$DE,$9F,$22,$B6,$FD,$4C,$9E,$C7,$EE,$B5,$FD,$AA,$F1,$BF
	.dc.b	$30,$E7,$1F,$4B,$97,$FC,$CA,$E6,$D7,$D2,$B8,$C7,$27,$E7,$79,$BF
	.dc.b	$81,$FC,$D5,$F3,$B8,$C7,$F0,$31,$87,$ED,$4F,$ED,$AF,$E9,$B7,$87
	.dc.b	$FC,$D3,$3F,$5D,$B3,$F5,$A7,$A3,$F8,$8E,$67,$D5,$63,$5F,$E2,$7F
	.dc.b	$67,$FB,$7C,$7E,$AA,$F0,$3F,$70,$E7,$0F,$6B,$99,$FF,$42,$B9,$D8
	.dc.b	$E6,$7E,$73,$93,$17,$FC,$8F,$83,$F0,$93,$FC,$25,$B0,$5F,$0D,$F8
	.dc.b	$4E,$49,$8F,$E0,$5F,$C3,$C6,$2F,$F5,$E7,$E6,$3C,$FF,$E2,$1F,$66
	.dc.b	$F3,$FF,$3D,$18,$6E,$1F,$26,$3E,$61,$1C,$1F,$AB,$79,$FD,$8F,$0B
	.dc.b	$F9,$9F,$07,$FB,$6F,$63,$F6,$ED,$BF,$DD,$F1,$23,$15,$C2,$73,$5F
	.dc.b	$C9,$7E,$CD,$F6,$74,$C6,$36,$BF,$D6,$73,$A6,$CE,$9A,$7F,$B3,$BE
	.dc.b	$CD,$F4,$DF,$C3,$FC,$B7,$D9,$7F,$27,$9C,$D7,$F1,$09,$CB,$C6,$BF
	.dc.b	$21,$3A,$CF,$FC,$7A,$D7,$FC,$86,$D7,$FC,$8E,$EF,$91,$58,$AF,$DE
	.dc.b	$FD,$9F,$85,$FE,$89,$F4,$3F,$CC,$F0,$FD,$4B,$7A,$0D,$35,$0F,$B3
	.dc.b	$FB,$98,$7F,$17,$6C,$7F,$94,$9A,$FF,$96,$1A,$FF,$12,$F9,$6F,$E3
	.dc.b	$0F,$EE,$1B,$66,$9A,$8F,$E0,$9F,$6B,$FB,$3F,$08,$D2,$61,$FC,$A6
	.dc.b	$3F,$CA,$53,$FE,$68,$6B,$FE,$6C,$6B,$FE,$70,$6B,$FC,$47,$E9,$FB
	.dc.b	$12,$7A,$3F,$DC,$FE,$CB,$F5,$D3,$FE,$76,$C5,$18,$8F,$E5,$ED,$7F
	.dc.b	$8A,$B5,$7B,$9F,$D7,$9F,$F3,$CB,$5F,$F3,$FB,$5F,$F7,$07,$9F,$FE
	.dc.b	$03,$F1,$3F,$6B,$8F,$F8,$1D,$B3,$FE,$86,$C3,$FA,$4F,$76,$CF,$E3
	.dc.b	$5F,$60,$F8,$9A,$4C,$3F,$A8,$46,$5B,$F5,$9C,$08,$E0,$DF,$BD,$D0
	.dc.b	$8E,$0F,$F0,$D0,$F9,$6F,$F0,$D3,$4C,$B8,$3F,$97,$7D,$73,$EB,$E9
	.dc.b	$27,$FB,$24,$65,$FF,$13,$74,$BE,$2F,$FC,$4C,$D8,$CC,$60,$EB,$E7
	.dc.b	$C8,$E6,$DD,$FF,$D5,$37,$FD,$92,$7B,$FA,$5E,$42,$38,$3F,$D6,$F8
	.dc.b	$7E,$BA,$BF,$D7,$35,$C5,$FE,$47,$F5,$FF,$E2,$39,$D4,$27,$51,$8E
	.dc.b	$1F,$DC,$5F,$4F,$A4,$FE,$6B,$7F,$D7,$3B,$7F,$C3,$11,$8A,$FE,$01
	.dc.b	$3F,$DF,$27,$FA,$B4,$62,$38,$57,$F8,$2E,$3A,$C6,$E3,$FC,$3F,$B3
	.dc.b	$E4,$BF,$C7,$55,$AA,$FE,$D7,$C3,$FE,$27,$FA,$F7,$CA,$6A,$30,$DC
	.dc.b	$7B,$47,$88,$8F,$87,$AF,$CA,$47,$04,$BF,$E9,$38,$27,$FC,$6D,$F5
	.dc.b	$FE,$F7,$FC,$9D,$B3,$EB,$FF,$BA,$37,$0F,$F1,$43,$70,$FF,$71,$B5
	.dc.b	$FF,$74,$B9,$3E,$99,$63,$FD,$DE,$BE,$BF,$BD,$7F,$97,$6B,$FD,$EB
	.dc.b	$DB,$FE,$F6,$FA,$F7,$CC,$F0,$F9,$CF,$C1,$46,$2B,$F7,$21,$C3,$FE
	.dc.b	$77,$8D,$CF,$F0,$08,$D3,$CF,$2E,$CE,$6B,$8E,$4B,$E3,$7D,$BE,$8B
	.dc.b	$31,$F9,$37,$BC,$7A,$03,$AC,$11,$8C,$3B,$1E,$17,$C6,$7A,$B0,$E0
	.dc.b	$E0,$E2,$7F,$1A,$31,$87,$96,$BD,$31,$C6,$BA,$57,$92,$38,$1F,$E5
	.dc.b	$FC,$7F,$CC,$0F,$9C,$C1,$D8,$F3,$F8,$FF,$81,$F8,$8F,$F3,$33,$63
	.dc.b	$47,$F5,$8D,$13,$A6,$C6,$A7,$EF,$65,$CB,$5F,$F5,$54,$E5,$AF,$86
	.dc.b	$76,$E6,$7D,$AF,$E0,$BE,$73,$E1,$E5,$F7,$B3,$46,$1F,$8F,$7C,$A7
	.dc.b	$F0,$6D,$8F,$96,$70,$AE,$AD,$7E,$84,$6D,$F8,$78,$7A,$2B,$C1,$7D
	.dc.b	$E7,$EE,$DF,$19,$EF,$5B,$8F,$1B,$67,$75,$C5,$F6,$1E,$EB,$F0,$3F
	.dc.b	$E0,$49,$E0,$E8,$FC,$25,$77,$3A,$E1,$F1,$BE,$83,$FA,$14,$FF,$28
	.dc.b	$D7,$F1,$31,$88,$FF,$24,$B6,$7F,$93,$D7,$EE,$2B,$7C,$FF,$B6,$78
	.dc.b	$FC,$2D,$C3,$F3,$78,$23,$10,$8E,$93,$E0,$F9,$1C,$C1,$A7,$3C,$5F
	.dc.b	$5F,$FA,$FB,$F4,$1D,$A7,$51,$FE,$17,$8F,$F6,$4B,$5F,$5C,$78,$BE
	.dc.b	$B9,$F4,$DE,$4F,$AF,$FF,$50,$D7,$40,$E1,$FF,$35,$CF,$07,$4D,$7F
	.dc.b	$CD,$F1,$BC,$3F,$A6,$B7,$2F,$DA,$D6,$E7,$FA,$8A,$FA,$4F,$21,$2F
	.dc.b	$FD,$C3,$F5,$EF,$AF,$C4,$7F,$B8,$3C,$27,$D7,$FC,$56,$C7,$92,$F7
	.dc.b	$8F,$F6,$31,$3A,$9E,$55,$AC,$BF,$84,$FD,$63,$EB,$FE,$BB,$6B,$E2
	.dc.b	$35,$FF,$46,$AF,$7B,$73,$F7,$BE,$BF,$FD,$2F,$5F,$2D,$AF,$08,$D7
	.dc.b	$A3,$BC,$64,$EE,$3F,$15,$E1,$FE,$A2,$39,$7D,$7F,$53,$2F,$F3,$2C
	.dc.b	$FF,$90,$E2,$9E,$3F,$40,$F9,$2A,$71,$BF,$C4,$E0,$FA,$F6,$DE,$05
	.dc.b	$F7,$14,$C4,$7E,$6B,$DC,$FF,$7A,$5F,$E1,$FE,$E7,$F2,$8F,$FA,$C1
	.dc.b	$BC,$7F,$E0,$E9,$DC,$FF,$AD,$1F,$C1,$6D,$FC,$E2,$BF,$D9,$D3,$CF
	.dc.b	$E3,$BB,$75,$77,$D7,$F5,$35,$FE,$A9,$AF,$07,$C3,$E5,$F5,$FD,$8F
	.dc.b	$F4,$C6,$F4,$FC,$A8,$FC,$54,$6B,$CA,$F5,$67,$FA,$DF,$C6,$5F,$45
	.dc.b	$4C,$D7,$C6,$FA,$6F,$BC,$DF,$93,$FA,$B1,$FE,$EA,$8D,$76,$7E,$FB
	.dc.b	$EC,$BF,$75,$1F,$AB,$FB,$AF,$F6,$83,$7E,$47,$5E,$2F,$F0,$4F,$B3
	.dc.b	$7C,$A7,$F0,$6B,$7C,$35,$35,$1B,$FA,$2D,$7C,$C4,$CB,$F7,$18,$27
	.dc.b	$F1,$2F,$E4,$3A,$FF,$C8,$9D,$E8,$FA,$71,$86,$F1,$FF,$72,$9D,$78
	.dc.b	$7E,$5F,$95,$F8,$DE,$9C,$CB,$BC,$7E,$5B,$CE,$FE,$37,$0B,$FE,$FF
	.dc.b	$97,$83,$D1,$FE,$51,$39,$AD,$AF,$B2,$FE,$04,$FF,$11,$6B,$FE,$FA
	.dc.b	$6D,$7F,$96,$3E,$CF,$EF,$4E,$7C,$DF,$E5,$B1,$F4,$23,$FE,$00,$B6
	.dc.b	$7F,$CB,$F5,$CB,$5F,$F0,$3B,$6B,$FC,$23,$5B,$9F,$F0,$5B,$63,$FC
	.dc.b	$2B,$3B,$51,$F9,$E9,$DF,$B5,$E2,$3C,$7F,$0D,$6B,$F9,$65,$6D,$EF
	.dc.b	$3A,$73,$F4,$BE,$AF,$1A,$46,$08,$7B,$1C,$BE,$E8,$FE,$9B,$8F,$E3
	.dc.b	$A3,$E4,$74,$E0,$F8,$BF,$49,$63,$C7,$E2,$7C,$83,$0F,$03,$24,$30
	.dc.b	$7C,$CF,$A3,$F9,$E4,$7D,$BE,$BD,$74,$7C,$08,$FA,$B1,$FD,$8B,$E2
	.dc.b	$DF,$21,$79,$E5,$FF,$51,$9C,$B6,$72,$B8,$BF,$D3,$33,$FE,$13,$6C
	.dc.b	$FF,$6E,$FB,$27,$D9,$D2,$5F,$5F,$8C,$AF,$3E,$5F,$F2,$4E,$3F,$CA
	.dc.b	$E3,$FC,$A6,$D5,$7D,$1F,$EC,$CD,$79,$1C,$3A,$3D,$C7,$6C,$BC,$F5
	.dc.b	$F0,$7E,$36,$57,$BE,$3E,$21,$87,$98,$7F,$B5,$3E,$47,$E5,$7C,$44
	.dc.b	$8F,$90,$FC,$DF,$8A,$C4,$72,$9F,$F5,$3B,$A3,$E7,$47,$C7,$FA,$D9
	.dc.b	$1E,$C7,$EB,$7D,$1E,$23,$96,$DF,$64,$C6,$5F,$EC,$AD,$17,$9A,$FB
	.dc.b	$55,$AD,$7E,$6F,$C9,$7E,$04,$93,$FE,$0F,$FB,$3F,$F5,$0C,$EA,$3F
	.dc.b	$05,$3A,$7E,$1C,$E9,$FF,$43,$BF,$9F,$C7,$FD,$B6,$DA,$FF,$41,$B6
	.dc.b	$77,$E0,$FF,$8C,$A3,$10,$69,$23,$FC,$6F,$E6,$7F,$40,$9F,$BB,$F1
	.dc.b	$2F,$B3,$FF,$7F,$47,$F8,$CA,$31,$E0,$F0,$F3,$A4,$E6,$B6,$F2,$6E
	.dc.b	$7F,$EF,$2C,$3B,$3F,$CD,$BE,$CB,$FC,$91,$5F,$DD,$9A,$FE,$3B,$C6
	.dc.b	$46,$27,$F9,$EF,$D9,$FF,$2E,$9C,$F6,$7F,$82,$AB,$F9,$AC,$70,$F8
	.dc.b	$8C,$4F,$F8,$3B,$EC,$9E,$6F,$F8,$43,$EB,$DF,$67,$4D,$44,$D3,$FC
	.dc.b	$13,$3A,$6C,$E5,$FC,$CE,$AF,$82,$FF,$5A,$FC,$87,$F3,$5F,$B3,$FF
	.dc.b	$01,$B7,$EB,$33,$EB,$FF,$AE,$AF,$86,$1F,$ED,$E8,$C6,$1A,$8F,$F6
	.dc.b	$3E,$1C,$A7,$FD,$82,$D8,$FF,$65,$D7,$25,$87,$F4,$58,$C3,$E1,$8E
	.dc.b	$C7,$4C,$3B,$29,$DD,$F6,$9C,$DF,$7A,$36,$B7,$1E,$6E,$7F,$5E,$1F
	.dc.b	$05,$B8,$7A,$1B,$3E,$CF,$57,$AC,$62,$3A,$B7,$03,$67,$79,$7A,$AB
	.dc.b	$6D,$AD,$B6,$77,$3F,$12,$FC,$53,$BB,$E9,$36,$23,$B3,$D8,$F3,$73
	.dc.b	$87,$B6,$2D,$7D,$4B,$EF,$39,$AE,$47,$98,$B8,$7F,$88,$9E,$27,$63
	.dc.b	$38,$79,$1B,$86,$9E,$55,$F8,$DB,$EF,$87,$B2,$76,$76,$8C,$BD,$67
	.dc.b	$C8,$71,$7A,$61,$F2,$1E,$31,$96,$D7,$58,$E2,$FB,$E3,$8E,$1F,$5C
	.dc.b	$F4,$4F,$C1,$81,$E5,$29,$EF,$29,$F1,$4B,$4F,$2F,$FB,$2C,$61,$AF
	.dc.b	$E8,$9A,$FE,$99,$AF,$E5,$9F,$ED,$B3,$FB,$36,$DF,$8B,$EC,$FF,$9D
	.dc.b	$F6,$7E,$26,$73,$7C,$FC,$8C,$1F,$8F,$AD,$C7,$DA,$9D,$BD,$E7,$6F
	.dc.b	$E0,$27,$78,$78,$67,$2D,$9D,$78,$0C,$33,$8F,$B6,$ED,$7C,$AC,$39
	.dc.b	$CE,$63,$C5,$3A,$53,$A7,$D4,$68,$98,$7F,$92,$63,$51,$E9,$33,$5D
	.dc.b	$67,$58,$79,$67,$27,$48,$F5,$1A,$C3,$C5,$97,$D9,$C7,$C9,$8F,$68
	.dc.b	$C5,$F2,$3C,$17,$C5,$C4,$6A,$3A,$E3,$F5,$F1,$F1,$46,$2B,$C5,$3E
	.dc.b	$73,$58,$F8,$B2,$FA,$B8,$73,$9C,$D7,$87,$1F,$F0,$85,$FD,$DF,$13
	.dc.b	$AD,$7E,$17,$DB,$B7,$F4,$66,$BF,$3A,$2D,$F2,$8F,$C0,$46,$27,$2D
	.dc.b	$BF,$C5,$36,$77,$1E,$13,$25,$7F,$89,$74,$3E,$C9,$F6,$7F,$EF,$FA
	.dc.b	$FF,$4D,$73,$3C,$6C,$61,$FE,$95,$BD,$D7,$F5,$A9,$CE,$1C,$F0,$B8
	.dc.b	$71,$AF,$F7,$3D,$F0,$C3,$FB,$94,$E5,$FE,$E9,$EC,$BD,$2E,$6B,$ED
	.dc.b	$EC,$61,$B8,$7F,$BC,$A7,$83,$7D,$4C,$DF,$F9,$0B,$ED,$1E,$A7,$AB
	.dc.b	$C8,$7D,$3F,$FA,$FF,$0F,$E1,$7E,$37,$E4,$F0,$61,$C2,$7F,$C2,$D5
	.dc.b	$B9,$E9,$87,$38,$C5,$F1,$BE,$15,$FE,$20,$9C,$DF,$FB,$17,$ED,$3F
	.dc.b	$DA,$AF,$80,$C7,$BD,$F2,$A1,$F0,$F6,$24,$EE,$FF,$D8,$D5,$C9,$FF
	.dc.b	$0E,$3C,$23,$5C,$68,$D6,$5F,$AF,$1F,$5A,$F6,$3E,$EC,$61,$B0,$F2
	.dc.b	$2C,$4C,$3C,$55,$B6,$D6,$DB,$3F,$88,$C5,$F4,$9F,$EF,$69,$F2,$FD
	.dc.b	$A7,$9E,$7F,$CE,$F3,$97,$FC,$6D,$19,$C3,$EA,$4E,$5F,$F7,$9C,$66
	.dc.b	$BF,$A0,$CE,$5F,$F7,$94,$66,$BF,$A3,$CE,$5B,$E2,$7E,$63,$ED,$1E
	.dc.b	$D7,$9E,$B5,$8B,$EC,$5E,$57,$C6,$6E,$7F,$0B,$BB,$FC,$6C,$FC,$48
	.dc.b	$CE,$7F,$E9,$CB,$E5,$7C,$4F,$9F,$7C,$27,$55,$C2,$BF,$92,$70,$BD
	.dc.b	$AF,$EB,$F8,$FD,$C7,$FD,$82,$DB,$FD,$8C,$61,$FF,$2F,$3C,$1B,$E4
	.dc.b	$69,$DB,$EF,$F3,$7F,$03,$F6,$6F,$B4,$C6,$1F,$7A,$30,$FE,$46,$FE
	.dc.b	$DD,$70,$AD,$C7,$BE,$B6,$FD,$C9,$CE,$5F,$91,$E4,$7C,$4E,$31,$FE
	.dc.b	$F7,$A3,$ED,$3F,$77,$8F,$F9,$DD,$77,$FF,$0E,$73,$8F,$C6,$73,$7F
	.dc.b	$A0,$72,$F4,$73,$5C,$FC,$2E,$57,$F6,$F9,$A5,$FD,$9F,$87,$B7,$F5
	.dc.b	$26,$E3,$FA,$9E,$25,$62,$FF,$23,$E7,$FF,$22,$FB,$4F,$F0,$74,$7C
	.dc.b	$7E,$7F,$F2,$7B,$FB,$48,$CE,$5F,$A0,$AE,$1E,$77,$E4,$5B,$5C,$5C
	.dc.b	$3C,$7E,$D2,$D7,$E3,$78,$9F,$09,$FB,$8F,$6F,$F7,$2D,$BF,$DA,$B6
	.dc.b	$B8,$37,$6F,$C7,$FD,$A3,$ED,$3F,$9D,$B6,$FF,$85,$AC,$B3,$32,$72
	.dc.b	$F0,$8C,$61,$FE,$F2,$FB,$37,$D5,$FF,$0E,$F2,$1F,$69,$FC,$16,$73
	.dc.b	$8F,$F0,$C9,$CB,$F8,$F8,$CD,$FE,$AE,$72,$F5,$BE,$97,$CD,$97,$FB
	.dc.b	$3F,$B4,$B5,$FE,$C1,$FA,$7F,$E6,$BA,$FE,$B9,$F9,$35,$FC,$A1,$CB
	.dc.b	$F9,$47,$32,$9F,$E8,$DF,$69,$F6,$9F,$E5,$9F,$7B,$FE,$C2,$BF,$D3
	.dc.b	$4E,$5F,$E9,$B1,$9A,$FF,$69,$7C,$E7,$F1,$3E,$E3,$FC,$06,$BF,$46
	.dc.b	$E4,$C4,$FF,$98,$3C,$37,$DA,$FF,$38,$F0,$5E,$C7,$F2,$F9,$CE,$1F
	.dc.b	$CD,$A7,$2F,$EE,$63,$FC,$9F,$A7,$F4,$39,$7C,$7F,$1F,$7F,$A3,$CF
	.dc.b	$EB,$7F,$A2,$FD,$CF,$EE,$31,$E5,$18,$BF,$F0,$0F,$DB,$7E,$BB,$1F
	.dc.b	$E3,$2F,$CA,$D5,$BA,$35,$87,$F0,$07,$53,$FE,$CE,$F5,$3F,$A9,$F0
	.dc.b	$BE,$87,$FA,$57,$C1,$FF,$5A,$C3,$E5,$C7,$F8,$43,$C4,$FD,$9C,$7F
	.dc.b	$60,$EC,$FE,$59,$A7,$F2,$A9,$CE,$5F,$F0,$06,$73,$5F,$E8,$8F,$B6
	.dc.b	$7C,$7F,$F6,$2F,$B8,$7C,$FF,$F5,$99,$CD,$7F,$60,$9C,$BF,$C8,$E1
	.dc.b	$F6,$7F,$DA,$BE,$13,$FC,$3D,$A7,$F8,$2E,$73,$5F,$DA,$27,$2F,$87
	.dc.b	$E1,$78,$7C,$DF,$76,$C7,$F7,$A1,$F1,$DF,$E2,$EB,$FE,$79,$87,$F4
	.dc.b	$69,$C9,$FE,$00,$8C,$BF,$EA,$4F,$B5,$7D,$BF,$F8,$DB,$ED,$DF,$67
	.dc.b	$FD,$D6,$73,$5F,$E0,$89,$CB,$FE,$0A,$C1,$F2,$DB,$C3,$FC,$7F,$18
	.dc.b	$9F,$E8,$97,$CB,$8F,$F4,$D1,$E1,$F7,$3B,$7B,$BD,$DB,$1C,$3D,$FC
	.dc.b	$34,$E7,$1E,$58,$91,$FD,$47,$3F,$C3,$E1,$FD,$E2,$3F,$AA,$76,$FF
	.dc.b	$06,$CB,$FC,$3D,$F5,$FF,$DF,$F3,$FC,$8F,$97,$FE,$1E,$8C,$56,$EF
	.dc.b	$FC,$83,$18,$9D,$7B,$3F,$E1,$9A,$D5,$FF,$56,$9F,$F1,$3F,$13,$E6
	.dc.b	$F5,$1B,$E7,$FF,$15,$FD,$BB,$E7,$71,$A7,$F8,$A2,$30,$FA,$27,$73
	.dc.b	$AC,$7F,$C1,$15,$A7,$E5,$9F,$D9,$BC,$5F,$F1,$BF,$DB,$FE,$73,$7F
	.dc.b	$F2,$C6,$A9,$CF,$B9,$E9,$66,$7F,$E3,$8B,$F8,$31,$FE,$35,$B7,$DB
	.dc.b	$DF,$80,$C3,$6F,$FC,$01,$3A,$6C,$71,$EA,$DC,$7F,$81,$73,$FF,$7A
	.dc.b	$F8,$BF,$EF,$6F,$B7,$7D,$0F,$FB,$DB,$17,$D1,$FF,$B0,$67,$37,$FE
	.dc.b	$F8,$BE,$17,$BA,$D3,$FE,$FC,$94,$E9,$FF,$3C,$CE,$9E,$42,$BA,$BF
	.dc.b	$CE,$63,$48,$F3,$9B,$AF,$06,$1D,$63,$47,$80,$D1,$E0,$74,$F2,$AE
	.dc.b	$F8,$7B,$A7,$33,$C6,$73,$7E,$28,$E1,$7E,$07,$A9,$E0,$C3,$C5,$87
	.dc.b	$38,$C9,$E8,$3E,$21,$E5,$2C,$61,$D1,$EA,$F5,$7E,$FF,$C8,$F0,$6C
	.dc.b	$7C,$D6,$BF,$00,$38,$79,$67,$FD,$8D,$E0,$D7,$84,$7F,$0E,$BE,$D3
	.dc.b	$99,$E3,$1F,$37,$0F,$EA,$31,$7D,$0F,$9F,$9F,$D8,$C3,$94,$FC,$46
	.dc.b	$C7,$DC,$AE,$33,$F2,$1B,$1C,$70,$F3,$37,$0E,$D5,$0E,$93,$AA,$F2
	.dc.b	$C7,$3C,$3C,$71,$F8,$23,$A9,$AB,$E9,$3B,$3A,$38,$3F,$3B,$C4,$FC
	.dc.b	$89,$F6,$3F,$A3,$DF,$F2,$5D,$CF,$CC,$1F,$07,$9C,$93,$98,$C6,$3C
	.dc.b	$F0,$B8,$F4,$BF,$F6,$35,$ED,$B8,$74,$7F,$D8,$B3,$B7,$F8,$6C,$FF
	.dc.b	$6F,$F6,$FC,$0D,$8E,$7E,$82,$72,$BC,$0B,$12,$76,$EA,$3D,$5C,$2F
	.dc.b	$B6,$FE,$97,$5F,$EC,$A6,$BF,$EA,$76,$C7,$FB,$5D,$FE,$AF,$1B,$F2
	.dc.b	$39,$C6,$B7,$FF,$7C,$B6,$7F,$C6,$CF,$07,$DB,$C9,$FE,$78,$7E,$2B
	.dc.b	$5C,$F6,$7C,$08,$FF,$5F,$C6,$A3,$04,$78,$FC,$6D,$78,$9B,$32,$7F
	.dc.b	$BC,$8F,$E8,$45,$EF,$FB,$2E,$0F,$E7,$BC,$5F,$C4,$FE,$37,$EE,$FC
	.dc.b	$EB,$E3,$69,$2B,$91,$FE,$FE,$8D,$3F,$F0,$03,$EC,$DE,$87,$E1,$9B
	.dc.b	$1B,$AF,$BB,$F9,$86,$23,$F0,$2F,$E0,$F8,$5C,$1F,$94,$8C,$37,$0E
	.dc.b	$37,$2F,$F1,$F5,$E2,$79,$F4,$7E,$2B,$F2,$3F,$F7,$6C,$7E,$51,$AF
	.dc.b	$E4,$1B,$8F,$82,$3F,$3F,$5C,$E3,$1C,$EF,$85,$FB,$CD,$EF,$FC,$BC
	.dc.b	$7D,$F9,$DB,$F8,$3E,$07,$E3,$6F,$57,$F5,$2D,$CB,$FA,$15,$FF,$BC
	.dc.b	$23,$11,$F5,$1F,$D0,$3F,$AC,$F5,$5F,$39,$FA,$66,$C6,$63,$1F,$37
	.dc.b	$FA,$AF,$23,$FC,$0D,$3F,$B4,$6B,$FB,$06,$E1,$EE,$7F,$C9,$71,$FB
	.dc.b	$AF,$CB,$3F,$DC,$2D,$F9,$3F,$DA,$C6,$23,$F7,$2F,$89,$FD,$DF,$E2
	.dc.b	$FD,$39,$BF,$86,$BD,$23,$3E,$87,$EF,$A3,$51,$FC,$3B,$F2,$ED,$65
	.dc.b	$F4,$17,$B3,$FC,$11,$B3,$FC,$39,$DE,$1F,$EF,$AF,$BA,$F9,$37,$F6
	.dc.b	$F8,$91,$89,$D7,$17,$F1,$A8,$F9,$B8,$7F,$12,$FC,$B9,$F3,$9F,$C4
	.dc.b	$9B,$1F,$C7,$1A,$EF,$E7,$7F,$8B,$FC,$E6,$33,$FC,$3B,$72,$FE,$5A
	.dc.b	$DA,$FE,$56,$AB,$F8,$FF,$A8,$F7,$A9,$2B,$F9,$5C,$3E,$2F,$F9,$2F
	.dc.b	$E2,$0F,$99,$BE,$94,$BF,$E6,$33,$97,$F1,$B9,$FF,$82,$BE,$CB,$C5
	.dc.b	$06,$93,$3F,$E8,$11,$9D,$EC,$E9,$B3,$86,$4F,$63,$FC,$39,$3C,$DF
	.dc.b	$E8,$7D,$1F,$D9,$B7,$FD,$5A,$FA,$6F,$E7,$ED,$9D,$FA,$F8,$33,$32
	.dc.b	$32,$77,$E5,$74,$3E,$8B,$FA,$AB,$FC,$22,$7F,$D8,$2D,$9D,$CE,$8D
	.dc.b	$F8,$D2,$BA,$57,$37,$D8,$3E,$C5,$F6,$3F,$19,$B1,$A6,$BF,$E6,$81
	.dc.b	$E1,$7F,$0E,$E1,$FF,$1D,$CE,$B6,$FE,$21,$5D,$6F,$A1,$F4,$AF,$84
	.dc.b	$7F,$91,$EB,$85,$FF,$91,$27,$FC,$7E,$DF,$6B,$8D,$7C,$5F,$C6,$7F
	.dc.b	$9F,$63,$FA,$36,$D7,$F9,$47,$54,$62,$BF,$DC,$EF,$ED,$23,$3E,$87
	.dc.b	$F8,$02,$35,$18,$9E,$DF,$07,$FD,$E3,$CD,$FE,$B3,$1F,$DF,$3F,$39
	.dc.b	$7A,$3F,$DF,$1B,$1F,$E0,$69,$DE,$7F,$CB,$9B,$9F,$F8,$52,$B7,$5C
	.dc.b	$EF,$B7,$A7,$F1,$74,$93,$F5,$B6,$F5,$3F,$43,$3C,$BF,$4E,$32,$48
	.dc.b	$CE,$7F,$4A,$33,$5F,$46,$7F,$C2,$7F,$97,$BE,$B2,$FE,$03,$31,$FF
	.dc.b	$0F,$CE,$5B,$39,$7F,$C3,$B0,$9E,$73,$FE,$2A,$FC,$6D,$F4,$14,$97
	.dc.b	$F9,$D8,$7C,$9E,$23,$FB,$BC,$70,$75,$3C,$E7,$F1,$1F,$27,$27,$E2
	.dc.b	$3F,$BC,$FB,$6F,$C7,$ED,$EF,$8D,$4F,$C5,$75,$C5,$89,$D4,$F2,$3F
	.dc.b	$47,$5C,$A7,$F1,$5F,$8D,$FF,$1D,$E5,$34,$9E,$17,$F9,$21,$C7,$85
	.dc.b	$4C,$47,$F9,$03,$F1,$BF,$F0,$26,$5F,$DC,$1B,$87,$F9,$59,$B9,$7F
	.dc.b	$95,$1F,$EE,$D1,$9F,$6F,$91,$27,$FC,$A7,$3A,$78,$7D,$7F,$72,$3A
	.dc.b	$FC,$12,$BF,$E5,$49,$4E,$30,$F2,$57,$9F,$F0,$3F,$47,$E1,$13,$0E
	.dc.b	$F1,$97,$C3,$19,$7A,$47,$BF,$D2,$91,$E6,$7E,$17,$C3,$E2,$3C,$51
	.dc.b	$70,$F8,$93,$FE,$72,$F7,$7F,$C0,$FF,$1B,$F9,$1E,$24,$63,$A3,$F7
	.dc.b	$CD,$C5,$3A,$BF,$F5,$F5,$6B,$1B,$5A,$BE,$17,$D2,$72,$F1,$9F,$F7
	.dc.b	$7D,$71,$69,$C6,$77,$5C,$8F,$ED,$F8,$79,$DC,$46,$8D,$E1,$E2,$A7
	.dc.b	$DA,$FF,$A5,$3C,$0E,$B1,$8F,$3F,$F8,$5E,$3F,$B1,$8C,$4E,$AB,$FD
	.dc.b	$0D,$16,$B7,$97,$FB,$B2,$7A,$4F,$FB,$F2,$31,$3A,$8F,$F7,$B4,$F2
	.dc.b	$74,$62,$34,$F8,$0C,$96,$FC,$93,$C5,$B3,$C2,$3C,$31,$8B,$F4,$7C
	.dc.b	$4F,$C1,$9C,$E1,$E4,$9F,$51,$AC,$29,$E0,$9D,$BE,$58,$D3,$9C,$3C
	.dc.b	$04,$8C,$E1,$E1,$9D,$47,$48,$F4,$E1,$E3,$AE,$15,$A3,$C1,$19,$AE
	.dc.b	$18,$79,$E7,$26,$0F,$21,$C6,$FA,$B9,$3E,$44,$71,$8C,$C6,$E3,$04
	.dc.b	$7E,$AD,$78,$5D,$1D,$A7,$89,$ED,$47,$94,$A7,$90,$A7,$DB,$2E,$1F
	.dc.b	$E2,$FF,$A2,$CC,$7E,$13,$D2,$46,$23,$F2,$F1,$87,$73,$C5,$B1,$9A
	.dc.b	$FC,$2B,$70,$F1,$F9,$DF,$E3,$FE,$7F,$CD,$4E,$6F,$C1,$8E,$0F,$D1
	.dc.b	$D6,$E3,$F0,$53,$B7,$C7,$3B,$7F,$2B,$3B,$C3,$C3,$39,$6C,$EB,$D3
	.dc.b	$FF,$52,$46,$09,$8F,$EA,$70,$ED,$87,$38,$C6,$1C,$DB,$87,$3C,$70
	.dc.b	$FA,$8F,$3C,$FA,$B1,$EF,$5C,$31,$EF,$5B,$7F,$7D,$87,$78,$C4,$EC
	.dc.b	$8F,$DE,$8D,$3F,$08,$EE,$49,$E7,$5E,$08,$E6,$43,$84,$6B,$0F,$4C
	.dc.b	$62,$FB,$BF,$E5,$BA,$E4,$E4,$E5,$97,$E0,$6F,$94,$78,$6B,$8C,$E5
	.dc.b	$FC,$0D,$72,$8C,$4F,$F3,$8F,$BB,$FF,$2F,$61,$FC,$BB,$89,$18,$9F
	.dc.b	$E2,$B1,$8F,$A0,$F6,$3A,$F4,$6D,$6F,$3F,$E9,$0D,$CF,$FC,$88,$AF
	.dc.b	$75,$31,$F0,$4E,$F1,$F2,$56,$5F,$0C,$67,$1F,$05,$70,$AD,$63,$FE
	.dc.b	$27,$F8,$47,$5B,$87,$FD,$E5,$CF,$FE,$01,$6E,$1F,$E0,$1F,$89,$BA
	.dc.b	$7F,$80,$66,$E1,$C9,$B8,$73,$BF,$F0,$0D,$E2,$73,$E5,$F0,$F8,$9F
	.dc.b	$F2,$AD,$7F,$81,$A3,$1C,$DA,$8F,$F4,$E6,$2A,$DE,$5F,$E0,$5B,$DD
	.dc.b	$6A,$5F,$3D,$D4,$ED,$F8,$1D,$9D,$45,$B6,$A9,$FE,$2B,$8F,$7B,$D4
	.dc.b	$E6,$93,$E3,$E3,$F5,$1E,$BF,$2A,$E1,$E3,$8B,$5D,$A3,$AE,$5F,$E5
	.dc.b	$37,$47,$FA,$AF,$ED,$3F,$CA,$6D,$F2,$BE,$7F,$22,$7F,$CA,$73,$92
	.dc.b	$5F,$53,$FC,$1B,$1C,$9C,$4F,$1F,$2B,$9C,$E7,$4F,$BF,$F5,$FF,$EF
	.dc.b	$0C,$3F,$D9,$50,$8C,$61,$D6,$31,$E3,$FE,$09,$D7,$91,$6B,$71,$D2
	.dc.b	$31,$39,$65,$F6,$AD,$BF,$D1,$2B,$AB,$99,$F8,$3E,$E7,$C7,$C3,$D6
	.dc.b	$DA,$FA,$99,$A1,$F5,$3C,$38,$26,$3F,$ED,$79,$E0,$FF,$AC,$5D,$B9
	.dc.b	$8F,$CD,$E8,$F6,$EF,$81,$86,$D7,$E5,$27,$4D,$9D,$34,$F9,$7F,$62
	.dc.b	$E7,$F9,$0D,$8F,$9F,$F0,$BF,$9C,$D7,$F6,$33,$9D,$BE,$83,$63,$37
	.dc.b	$F4,$6B,$85,$FC,$9A,$D3,$67,$B9,$BE,$9B,$1F,$AC,$FB,$AF,$4E,$3F
	.dc.b	$89,$E8,$43,$3F,$DE,$BA,$CF,$F4,$53,$A9,$CE,$BA,$6E,$1D,$5F,$CC
	.dc.b	$FC,$0B,$CC,$A6,$3C,$AB,$5F,$81,$3F,$DC,$93,$B7,$19,$FF,$9A,$23
	.dc.b	$0F,$E7,$14,$F2,$8F,$F8,$01,$F4,$AF,$6B,$F2,$DF,$69,$FC,$27,$0E
	.dc.b	$74,$AD,$47,$E8,$27,$53,$CA,$BF,$8F,$BB,$7F,$3D,$F5,$38,$BE,$33
	.dc.b	$BB,$EB,$18,$DB,$F5,$B5,$BB,$FD,$DF,$E2,$C8,$BD,$58,$6E,$1F,$03
	.dc.b	$A9,$5B,$8C,$3F,$A8,$CB,$FD,$16,$E5,$FD,$FB,$9E,$1F,$F6,$1F,$E0
	.dc.b	$DF,$CD,$BD,$6F,$E9,$97,$FC,$B1,$46,$7C,$9D,$36,$74,$DC,$3F,$27
	.dc.b	$E3,$7E,$68,$FC,$DF,$8F,$6B,$F3,$F1,$67,$F3,$33,$77,$FC,$EC,$DA
	.dc.b	$FC,$C4,$DA,$FE,$5B,$F9,$73,$C7,$FE,$9A,$FF,$C0,$1A,$FE,$86,$E5
	.dc.b	$BE,$3F,$F5,$2F,$CB,$7D,$71,$FD,$21,$B7,$F9,$D7,$E0,$B7,$3F,$EB
	.dc.b	$3F,$97,$7F,$2C,$8F,$D1,$37,$4F,$CF,$AF,$4F,$7E,$47,$F6,$0F,$67
	.dc.b	$FA,$46,$5F,$D5,$9B,$87,$EA,$DE,$51,$FD,$31,$CE,$1F,$C6,$3F,$18
	.dc.b	$FF,$31,$CF,$DF,$6D,$7F,$38,$7F,$CB,$B1,$98,$FE,$DF,$F0,$7F,$C0
	.dc.b	$1F,$EA,$6F,$F6,$08,$FE,$CE,$E7,$90,$E6,$7F,$7D,$CB,$FD,$D5,$BC
	.dc.b	$5F,$DE,$A3,$18,$FE,$F9,$47,$F7,$EF,$47,$F4,$B3,$FC,$09,$B8,$FF
	.dc.b	$68,$5F,$03,$BF,$3F,$FC,$0B,$F9,$6F,$D7,$CB,$F8,$73,$70,$FE,$16
	.dc.b	$FF,$34,$8F,$ED,$AE,$70,$FE,$E7,$EF,$3E,$B3,$FC,$1E,$D8,$FF,$02
	.dc.b	$FD,$DF,$CC,$77,$3F,$DA,$3C,$6F,$97,$8F,$F4,$66,$CF,$FA,$4B,$BB
	.dc.b	$F8,$F7,$B5,$BC,$FF,$C3,$7F,$42,$F8,$5B,$F2,$98,$6E,$1F,$E1,$4F
	.dc.b	$9C,$F7,$B7,$87,$FC,$47,$F9,$63,$ED,$FF,$C4,$B5,$AC,$BF,$97,$4E
	.dc.b	$6F,$FC,$3A,$D8,$CF,$CD,$72,$BE,$7F,$75,$D2,$2D,$F0,$7F,$C0,$BD
	.dc.b	$5F,$D6,$AF,$FB,$FC,$AB,$3E,$CF,$F8,$2A,$30,$7F,$0D,$AE,$1C,$5F
	.dc.b	$DA,$9B,$1F,$16,$3E,$95,$B8,$7F,$C4,$37,$FE,$14,$9C,$C7,$F8,$AB
	.dc.b	$C5,$C7,$BB,$97,$FC,$8B,$F8,$77,$F1,$32,$7B,$B8,$F3,$66,$7F,$E3
	.dc.b	$E9,$CB,$63,$39,$FF,$55,$9C,$B6,$3F,$C2,$5F,$5C,$FB,$7A,$63,$E3
	.dc.b	$EE,$1F,$E1,$C9,$D3,$67,$4D,$3F,$C9,$FF,$87,$7F,$5F,$7F,$C8,$6D
	.dc.b	$C3,$FC,$82,$FF,$65,$8D,$FB,$1F,$E6,$28,$D7,$09,$A6,$EB,$FB,$73
	.dc.b	$CF,$77,$53,$E2,$FF,$CD,$7F,$95,$7C,$C8,$C6,$5F,$E6,$28,$C4,$EB
	.dc.b	$87,$FC,$8B,$5A,$C7,$FB,$D5,$7D,$87,$FD,$DB,$F0,$DF,$EF,$0F,$AB
	.dc.b	$7C,$A7,$FB,$C3,$99,$F5,$BF,$E5,$D9,$CD,$FF,$BC,$AF,$85,$EE,$B4
	.dc.b	$FF,$BC,$A3,$33,$FE,$39,$9C,$BC,$45,$74,$7F,$CD,$F1,$A7,$CA,$6A
	.dc.b	$BB,$69,$D2,$33,$5C,$70,$F3,$CE,$64,$30,$E1,$3A,$8E,$87,$8E,$B9
	.dc.b	$61,$E1,$8E,$35,$C2,$31,$1C,$CE,$27,$A2,$FC,$0E,$4F,$09,$EF,$29
	.dc.b	$EC,$29,$D8,$BF,$50,$FA,$37,$D6,$EA,$53,$98,$C3,$63,$8C,$F0,$79
	.dc.b	$FC,$97,$FA,$69,$B9,$F0,$8C,$47,$92,$91,$88,$F4,$C6,$27,$E8,$69
	.dc.b	$FE,$61,$9B,$C9,$EE,$D7,$ED,$5E,$E7,$E2,$B6,$3E,$F5,$71,$9F,$92
	.dc.b	$D8,$E3,$87,$9D,$B8,$76,$A8,$74,$9D,$57,$99,$B1,$F8,$08,$E8,$7D
	.dc.b	$22,$BF,$84,$BF,$24,$71,$AE,$B7,$E6,$C3,$D9,$39,$BF,$C3,$F1,$7E
	.dc.b	$D3,$E7,$3F,$4D,$AF,$FB,$13,$EB,$7F,$C0,$55,$FE,$D4,$6B,$FE,$AE
	.dc.b	$E2,$57,$08,$FE,$51,$18,$78,$7A,$B9,$E7,$5F,$2D,$7F,$D3,$77,$BB
	.dc.b	$FF,$3A,$7D,$6B,$E5,$78,$36,$3F,$DC,$AD,$77,$EE,$EA,$7B,$6C,$EE
	.dc.b	$A6,$6B,$FD,$7F,$F5,$8F,$AD,$F4,$66,$8C,$4E,$A7,$31,$F8,$A8,$C1
	.dc.b	$0F,$F6,$33,$C1,$BC,$BE,$FE,$5F,$F0,$5C,$7F,$BF,$1A,$FF,$A4,$5B
	.dc.b	$7F,$E8,$E7,$FD,$51,$1B,$F5,$79,$11,$E5,$F1,$75,$E5,$16,$B9,$4F
	.dc.b	$D3,$E5,$FE,$0F,$E6,$5F,$8B,$FF,$43,$61,$FE,$6F,$6B,$FE,$E8,$3F
	.dc.b	$C3,$66,$7E,$17,$81,$1F,$F3,$B1,$A3,$FD,$DF,$1C,$B0,$FF,$6A,$FD
	.dc.b	$6B,$DD,$F1,$36,$BA,$1F,$E9,$EE,$77,$D6,$E6,$3F,$D6,$8F,$E0,$EF
	.dc.b	$FD,$62,$DE,$BF,$E4,$35,$C3,$0F,$C6,$4F,$86,$72,$DB,$57,$6C,$B7
	.dc.b	$D7,$F9,$3F,$45,$F6,$9F,$95,$FC,$F5,$B5,$35,$1F,$99,$FA,$FF,$BD
	.dc.b	$B7,$FB,$11,$FC,$26,$7F,$EC,$B6,$C6,$79,$FE,$97,$D7,$FF,$AE,$2B
	.dc.b	$FD,$3C,$D5,$3F,$EF,$C8,$C3,$7A,$FF,$51,$E6,$BE,$BE,$FA,$B3,$0F
	.dc.b	$95,$39,$7E,$67,$8D,$FE,$3B,$6F,$CA,$7E,$C6,$0D,$F6,$3F,$69,$F5
	.dc.b	$EF,$A9,$A4,$BF,$AD,$1D,$5F,$F7,$47,$D4,$F5,$6F,$B9,$FB,$D8,$36
	.dc.b	$FF,$80,$7D,$7E,$2B,$F0,$0F,$89,$B8,$EA,$BF,$31,$F5,$FF,$0B,$B3
	.dc.b	$FC,$B0,$DA,$FF,$1A,$37,$C0,$3F,$C0,$F8,$BF,$DE,$FF,$5F,$FC,$06
	.dc.b	$BF,$1C,$BD,$8F,$E2,$4D,$8C,$F9,$1F,$C5,$7E,$BF,$84,$7D,$4C,$BF
	.dc.b	$1F,$1B,$8C,$1F,$8C,$71,$82,$77,$5F,$C7,$FE,$61,$F5,$F4,$98,$7E
	.dc.b	$86,$1F,$4B,$FC,$97,$E8,$1F,$4D,$7D,$49,$7F,$A7,$9C,$BF,$EF,$CF
	.dc.b	$AF,$F3,$E9,$E1,$8F,$F0,$2C,$63,$4F,$EE,$ED,$8C,$FC,$0F,$F3,$75
	.dc.b	$19,$6E,$B9,$8C,$09,$C4,$49,$F1,$3B,$8F,$E7,$DD,$3F,$D8,$F0,$F2
	.dc.b	$79,$1F,$8D,$F8,$8C,$9A,$F6,$26,$3F,$C0,$ED,$3A,$8C,$4F,$F1,$BE
	.dc.b	$B7,$D3,$5F,$2A,$57,$F8,$93,$2F,$85,$D2,$BD,$CE,$77,$FF,$51,$46
	.dc.b	$23,$ED,$CE,$EF,$9C,$62,$75,$7F,$36,$1F,$5C,$FA,$FF,$94,$D8,$EC
	.dc.b	$7F,$8D,$9C,$3F,$88,$1E,$17,$F1,$07,$67,$49,$FA,$91,$FE,$41,$CB
	.dc.b	$FC,$5E,$D9,$E9,$3E,$03,$7E,$BC,$8E,$77,$E0,$C3,$E1,$FC,$9F,$F3
	.dc.b	$76,$FA,$7F,$D1,$63,$19,$7F,$B1,$5F,$DE,$4E,$AF,$D7,$AF,$FB,$B2
	.dc.b	$F8,$1C,$7C,$3F,$EF,$9E,$87,$F5,$8A,$FD,$2F,$23,$EB,$FF,$95,$C7
	.dc.b	$F3,$F6,$BA,$9E,$18,$FE,$19,$B9,$FF,$3E,$AD,$CF,$E2,$9B,$1F,$D0
	.dc.b	$A7,$75,$FE,$64,$DF,$D6,$DC,$BE,$59,$F3,$32,$B1,$F4,$27,$6F,$D4
	.dc.b	$9D,$39,$76,$FD,$98,$D9,$F1,$AF,$FA,$9F,$D7,$BE,$6F,$FC,$3C,$BE
	.dc.b	$9E,$98,$89,$7F,$D8,$67,$4D,$9C,$BF,$94,$FA,$F7,$D5,$7F,$7B,$6C
	.dc.b	$7D,$46,$BB,$F9,$BD,$4C,$8D,$3C,$63,$F1,$BF,$5F,$27,$E1,$3F,$C0
	.dc.b	$FE,$5F,$F3,$DC,$3F,$9F,$D5,$3B,$AF,$84,$EF,$5C,$56,$E2,$C7,$FA
	.dc.b	$1A,$F9,$BF,$E9,$8E,$37,$D1,$7F,$13,$6F,$81,$FE,$17,$8C,$1F,$C3
	.dc.b	$F1,$FC,$8B,$88,$D1,$BF,$03,$FC,$99,$F5,$9F,$E0,$9C,$FF,$87,$B7
	.dc.b	$1F,$F0,$3B,$70,$FF,$01,$BF,$E0,$5E,$73,$96,$FB,$BD,$0E,$BC,$FD
	.dc.b	$BE,$BE,$5F,$F8,$66,$D1,$E2,$FA,$FC,$CF,$E8,$A7,$2E,$E7,$4F,$1F
	.dc.b	$90,$F4,$FD,$54,$8C,$10,$E4,$E5,$E9,$1F,$03,$EA,$A4,$78,$DF,$87
	.dc.b	$ED,$62,$3C,$51,$76,$F8,$B3,$FE,$45,$F8,$7F,$F0,$97,$C8,$7E,$66
	.dc.b	$DC,$9F,$EA,$8C,$3F,$DD,$7D,$BF,$66,$F6,$DC,$33,$7C,$27,$2D,$AC
	.dc.b	$DF,$38,$FF,$68,$DF,$16,$9B,$2C,$6D,$FF,$6F,$47,$86,$79,$D7,$FA
	.dc.b	$07,$0F,$45,$ED,$F5,$70,$BE,$7D,$F3,$1E,$68,$C3,$CB,$0F,$4E,$4F
	.dc.b	$AF,$7A,$1F,$C6,$31,$F8,$93,$98,$C6,$9F,$EE,$2A,$DF,$C2,$7A,$AB
	.dc.b	$97,$C6,$F4,$9C,$C7,$A8,$FF,$7C,$B8,$C3,$AB,$6F,$A5,$71,$C3,$BD
	.dc.b	$F0,$6B,$E3,$C3,$C1,$6A,$E1,$18,$AD,$CF,$5A,$F0,$D6,$9F,$21,$CC
	.dc.b	$B1,$B7,$D6,$73,$A8,$EA,$BA,$D7,$3C,$3D,$71,$E0,$33,$3C,$23,$35
	.dc.b	$70,$F7,$CD,$3D,$D1,$5F,$09,$DD,$C6,$1E,$53,$E0,$46,$2F,$C1,$3F
	.dc.b	$13,$0F,$25,$FD,$79,$CD,$7B,$A7,$2F,$80,$EC,$64,$E9,$5E,$08,$F9
	.dc.b	$10,$3A,$45,$61,$F2,$CB,$7F,$82,$6F,$A6,$F9,$DD,$C7,$E1,$23,$0E
	.dc.b	$27,$FC,$87,$3A,$8F,$CB,$4E,$9E,$71,$F8,$28,$E3,$97,$E6,$3D,$1F
	.dc.b	$F2,$6F,$5E,$6A,$5F,$DE,$BD,$E9,$E9,$BD,$CE,$6B,$78,$7E,$5A,$F7
	.dc.b	$3F,$06,$B7,$39,$AD,$DF,$E9,$6F,$77,$FE,$3C,$F5,$7F,$CA,$D7,$B8
	.dc.b	$EC,$47,$D1,$1A,$AE,$57,$D2,$BE,$33,$FA,$FC,$3C,$33,$A9,$E4,$DB
	.dc.b	$E7,$8F,$19,$F9,$37,$6F,$8C,$74,$38,$96,$B8,$B8,$38,$96,$38,$B8
	.dc.b	$8F,$61,$2F,$E6,$61,$EB,$9E,$91,$E5,$23,$F2,$E3,$4F,$38,$CB,$4D
	.dc.b	$61,$EB,$CB,$F0,$B1,$E3,$AE,$31,$8B,$FE,$4B,$F4,$3F,$E6,$8B,$FE
	.dc.b	$5F,$C3,$FC,$97,$E5,$7C,$9C,$8A,$D6,$5F,$C5,$EB,$57,$FC,$D2,$73
	.dc.b	$87,$4F,$4F,$8C,$3E,$A2,$C4,$AE,$0E,$A3,$18,$7F,$23,$C7,$C1,$38
	.dc.b	$C3,$CF,$57,$5F,$EF,$3C,$5F,$E9,$DE,$9F,$EF,$AD,$CB,$FB,$EF,$C7
	.dc.b	$DC,$FF,$BE,$CD,$D7,$FB,$EB,$7B,$BF,$93,$5F,$F7,$9B,$C4,$E7,$E3
	.dc.b	$F8,$75,$39,$F1,$7F,$E7,$CC,$7F,$A6,$46,$32,$FF,$2F,$D6,$EB,$FC
	.dc.b	$09,$7C,$AB,$9F,$CE,$49,$E5,$4F,$99,$C6,$1C,$AA,$EB,$FE,$A8,$AD
	.dc.b	$5F,$09,$F2,$ED,$23,$9F,$9D,$70,$F1,$CD,$C3,$94,$DA,$FF,$19,$7D
	.dc.b	$77,$F9,$31,$FF,$3E,$B6,$FF,$D4,$6B,$CD,$F9,$9C,$A8,$D6,$1F,$EA
	.dc.b	$E6,$DF,$4F,$3B,$71,$C6,$3A,$C7,$D5,$97,$D7,$BC,$BC,$36,$BF,$D5
	.dc.b	$33,$AF,$2F,$35,$27,$35,$BC,$FF,$DE,$77,$B9,$CD,$6E,$73,$5B,$AF
	.dc.b	$C5,$3F,$ED,$BA,$D9,$FE,$9E,$F7,$BF,$CF,$0D,$D7,$E0,$C7,$F8,$2A
	.dc.b	$30,$4B,$FF,$64,$C7,$06,$F8,$9C,$2B,$F2,$5A,$3D,$CB,$E0,$E1,$B8
	.dc.b	$FE,$36,$74,$D9,$D3,$4F,$8B,$F5,$EE,$7F,$98,$D8,$F9,$3F,$0F,$F9
	.dc.b	$3C,$FE,$2C,$E5,$B5,$C0,$FA,$93,$BD,$BF,$5D,$5B,$6C,$EE,$FA,$F4
	.dc.b	$5C,$3F,$D7,$FF,$5F,$E8,$6E,$5F,$B1,$8C,$60,$BF,$ED,$1D,$49,$FF
	.dc.b	$7F,$C7,$EB,$5D,$BA,$73,$7D,$B5,$E9,$87,$E5,$3E,$09,$E6,$D3,$1E
	.dc.b	$55,$C7,$FD,$B3,$3A,$6C,$E9,$A7,$FB,$CB,$EB,$9E,$8F,$E5,$FE,$BF
	.dc.b	$FC,$95,$39,$8C,$77,$FF,$0F,$D1,$7C,$2B,$F3,$37,$C2,$7F,$E0,$04
	.dc.b	$E5,$FC,$BC,$67,$7F,$89,$1F,$E0,$B9,$D4,$62,$3A,$E1,$FA,$6F,$AF
	.dc.b	$F0,$56,$B0,$FF,$56,$7D,$52,$7F,$6E,$FE,$7E,$33,$18,$EB,$CB,$63
	.dc.b	$FD,$FB,$3C,$5B,$3F,$CA,$7D,$67,$D6,$DF,$23,$0D,$AF,$E4,$93,$A6
	.dc.b	$CE,$9A,$7E,$63,$C6,$FC,$B9,$F9,$8F,$1E,$C6,$49,$AF,$E6,$23,$2F
	.dc.b	$E6,$A6,$ED,$F9,$69,$B5,$F8,$CF,$AF,$FE,$4F,$6A,$F9,$BD,$09,$FF
	.dc.b	$7B,$57,$F3,$B7,$73,$A7,$35,$C6,$7F,$2F,$F5,$FE,$D8,$C0,$B4,$66
	.dc.b	$FB,$BF,$D8,$EB,$FA,$43,$97,$F4,$3F,$5F,$FC,$96,$F9,$DA,$1E,$9F
	.dc.b	$F5,$97,$58,$7A,$9C,$DF,$F5,$3F,$6B,$F5,$8D,$F1,$FF,$B0,$46,$31
	.dc.b	$EA,$FF,$36,$7F,$D7,$BF,$5F,$FE,$59,$AF,$F2,$D6,$1F,$92,$F2,$53
	.dc.b	$9A,$F7,$C6,$3D,$5F,$E8,$73,$9A,$F9,$BF,$09,$FC,$27,$0F,$DE,$AF
	.dc.b	$3B,$FB,$93,$9C,$7E,$43,$99,$FE,$D9,$CF,$FD,$D2,$7E,$1A,$E2,$FE
	.dc.b	$F4,$E7,$0F,$B6,$E6,$BF,$BA,$7A,$7F,$AB,$7F,$86,$B6,$FE,$B2,$F8
	.dc.b	$2F,$F0,$23,$9E,$27,$D7,$FE,$0A,$FF,$8A,$BF,$82,$C7,$F1,$2E,$70
	.dc.b	$FC,$03,$99,$FE,$37,$EF,$BE,$BB,$FC,$20,$BE,$06,$C6,$1F,$ED,$7E
	.dc.b	$37,$C9,$AF,$F4,$57,$1A,$31,$8F,$FB,$C9,$FF,$43,$4F,$D6,$36,$EB
	.dc.b	$37,$D2,$3E,$1B,$FC,$40,$BE,$52,$C6,$1F,$F1,$27,$CF,$FC,$3C,$3F
	.dc.b	$C3,$9F,$03,$CE,$31,$97,$E5,$23,$B7,$D7,$3E,$6B,$FC,$58,$DB,$FF
	.dc.b	$12,$2F,$AE,$CC,$EB,$EB,$B7,$16,$B6,$FF,$83,$35,$FE,$D9,$8F,$E3
	.dc.b	$A5,$59,$F6,$3F,$C2,$31,$83,$E1,$4B,$83,$F6,$53,$B9,$CC,$7F,$01
	.dc.b	$F7,$B1,$5F,$9A,$E6,$FB,$53,$C2,$B7,$18,$C7,$C6,$FF,$08,$FA,$EF
	.dc.b	$E5,$57,$FB,$F6,$C9,$FF,$05,$47,$EE,$9C,$93,$CA,$FF,$29,$3A,$AE
	.dc.b	$91,$82,$5F,$F9,$36,$74,$FF,$66,$7C,$32,$F9,$97,$D7,$7F,$8A,$DB
	.dc.b	$5F,$E2,$85,$F1,$B9,$99,$39,$78,$46,$31,$FE,$23,$F5,$DF,$DC,$1F
	.dc.b	$E3,$9F,$96,$3E,$53,$FC,$D5,$39,$D5,$39,$C7,$FC,$C9,$39,$27,$57
	.dc.b	$0F,$67,$FC,$93,$D6,$F8,$EF,$F3,$A7,$D7,$FD,$28,$C6,$1F,$E6,$D8
	.dc.b	$C4,$EB,$93,$8B,$67,$F0,$3A,$FF,$BB,$FE,$27,$FD,$E5,$F5,$DE,$78
	.dc.b	$FE,$37,$F1,$9F,$DB,$23,$11,$B0,$E3,$91,$3B,$E1,$FF,$79,$DF,$47
	.dc.b	$FD,$F9,$3A,$C3,$18,$7F,$BF,$AA,$BF,$D3,$EF,$93,$93,$BD,$F2,$8C
	.dc.b	$C7,$8F,$0F,$14,$62,$02,$BA,$E1,$E1,$8E,$67,$5A,$E3,$87,$8A,$38
	.dc.b	$D7,$0B,$E9,$1C,$CE,$27,$9E,$78,$39,$3C,$27,$A4,$A7,$C9,$29,$D8
	.dc.b	$BF,$56,$FA,$37,$D7,$EA,$53,$98,$C3,$63,$8C,$F0,$79,$F8,$1F,$DE
	.dc.b	$5B,$9F,$08,$C4,$79,$29,$18,$8F,$4C,$62,$7E,$AE,$9F,$E7,$19,$BC
	.dc.b	$3E,$ED,$7F,$0B,$7B,$9F,$8A,$D8,$FC,$4D,$71,$9F,$92,$D8,$E3,$87
	.dc.b	$9D,$B8,$76,$A8,$74,$9D,$57,$99,$B1,$F8,$08,$E8,$7D,$22,$BF,$84
	.dc.b	$BF,$24,$71,$AE,$B7,$E6,$C3,$D9,$39,$BF,$C3,$F5,$BE,$01,$E9,$7E
	.dc.b	$73,$6F,$F3,$9F,$D7,$BE,$BF,$18,$7F,$87,$63,$0F,$F9,$AA,$B3,$87
	.dc.b	$37,$7D,$4A,$DD,$7F,$87,$1B,$AF,$F8,$37,$2F,$0C,$FF,$BA,$F1,$FE
	.dc.b	$FC,$DF,$3F,$9C,$62,$3F,$C4,$AF,$FA,$7D,$D1,$0C,$C7,$0F,$86,$F1
	.dc.b	$F0,$71,$C3,$FC,$82,$DD,$7F,$D2,$2D,$BF,$F1,$D3,$D5,$FF,$19,$B9
	.dc.b	$CF,$F3,$38,$74,$3A,$E1,$F0,$3D,$D7,$AD,$49,$87,$F5,$6B,$FF,$34
	.dc.b	$FC,$06,$23,$FC,$7F,$5C,$1D,$35,$CC,$6F,$3F,$F2,$CC,$E5,$B7,$E1
	.dc.b	$9F,$A3,$F5,$AF,$9F,$A4,$C3,$C5,$1C,$75,$FF,$58,$BF,$E9,$E8,$CD
	.dc.b	$72,$38,$3F,$E5,$48,$D3,$FE,$56,$F3,$1F,$59,$F8,$06,$C7,$23,$E0
	.dc.b	$7E,$61,$FE,$03,$6F,$9B,$F8,$68,$FF,$05,$46,$BC,$DF,$C4,$F9,$18
	.dc.b	$8F,$F3,$1B,$76,$FF,$3C,$B7,$4F,$F6,$4D,$DA,$FF,$39,$7D,$77,$F9
	.dc.b	$3D,$FA,$0D,$7F,$CD,$AD,$BE,$78,$FE,$63,$EB,$BF,$8C,$57,$D7,$7F
	.dc.b	$4C,$DE,$7D,$77,$FE,$7B,$EA,$DF,$4B,$49,$5F,$7E,$1F,$59,$FA,$4F
	.dc.b	$37,$FC,$37,$3F,$83,$6B,$FE,$80,$6E,$7F,$E9,$65,$F8,$0F,$EB,$31
	.dc.b	$7D,$F5,$25,$FE,$36,$1F,$4F,$FB,$2E,$8F,$D6,$CF,$FA,$31,$B8,$FF
	.dc.b	$A8,$9B,$C5,$FE,$F2,$5F,$35,$FB,$AF,$A8,$7C,$B5,$CE,$63,$F9,$B9
	.dc.b	$CB,$F9,$DF,$AE,$F2,$B7,$3F,$E0,$71,$E4,$8D,$57,$F0,$6E,$27,$D5
	.dc.b	$D2,$4F,$E9,$E1,$F4,$5F,$C3,$3F,$52,$FE,$A3,$BF,$FA,$E9,$BB,$FF
	.dc.b	$66,$F0,$91,$8C,$3F,$D6,$8B,$E2,$BF,$8A,$FE,$09,$9A,$34,$91,$FC
	.dc.b	$6B,$1F,$C1,$DF,$FA,$D1,$B8,$FF,$B2,$9B,$8F,$85,$FC,$3C,$FF,$AE
	.dc.b	$BE,$7B,$F8,$FC,$F9,$5F,$F7,$4C,$7F,$03,$6B,$BF,$27,$F9,$7B,$FE
	.dc.b	$38,$F8,$8B,$D1,$FE,$97,$8F,$F0,$FB,$5F,$F6,$89,$FE,$3E,$F8,$3D
	.dc.b	$CC,$8D,$BC,$A3,$11,$C2,$79,$64,$C7,$FB,$FC,$61,$B5,$FC,$F3,$B3
	.dc.b	$F7,$B7,$FE,$D6,$5F,$6D,$93,$5E,$2C,$9F,$E4,$13,$A6,$CE,$5F,$F6
	.dc.b	$FF,$D5,$BE,$8F,$FA,$4A,$FA,$1A,$62,$24,$FF,$2B,$9D,$36,$72,$FC
	.dc.b	$4F,$AD,$FC,$4B,$FE,$46,$6B,$9E,$1F,$9B,$D9,$F4,$27,$3B,$FC,$3A
	.dc.b	$FF,$18,$C6,$36,$FF,$11,$B6,$3F,$C6,$BB,$FE,$89,$E8,$7D,$6E,$BF
	.dc.b	$B2,$75,$C7,$FB,$3F,$D6,$7E,$C2,$7F,$D3,$0D,$F5,$0D,$3F,$67,$1F
	.dc.b	$E9,$A7,$1E,$A7,$23,$9F,$81,$FD,$C3,$EB,$7F,$8E,$69,$FA,$35,$EF
	.dc.b	$64,$D7,$C9,$4C,$7F,$AA,$4E,$9B,$39,$7F,$0D,$F1,$9F,$92,$23,$FE
	.dc.b	$70,$F8,$EF,$8C,$E8,$FC,$AC,$69,$FC,$AC,$6B,$87,$7C,$1F,$EE,$39
	.dc.b	$D9,$0D,$E5,$FE,$0A,$FA,$1F,$F4,$EC,$7F,$69,$7F,$D3,$6D,$9D,$7A
	.dc.b	$12,$30,$41,$7F,$6F,$C1,$5F,$49,$CE,$1F,$EF,$0F,$9B,$7C,$5F,$F8
	.dc.b	$71,$B3,$FE,$CA,$6B,$C0,$FF,$68,$2F,$A6,$E1,$32,$78,$3C,$E3,$F7
	.dc.b	$BA,$3E,$62,$92,$BE,$67,$D5,$57,$E4,$C5,$D3,$E7,$D2,$B1,$97,$F9
	.dc.b	$DA,$3A,$5F,$D5,$BE,$F5,$FE,$34,$FA,$8F,$F0,$04,$B5,$46,$1B,$ED
	.dc.b	$EF,$DB,$E0,$CA,$FF,$0D,$56,$CF,$F0,$1E,$7F,$37,$E4,$3F,$44,$DE
	.dc.b	$EF,$F2,$1C,$63,$5F,$E6,$CF,$F4,$6B,$F9,$59,$23,$51,$FE,$FA,$AE
	.dc.b	$27,$2F,$3B,$FC,$9B,$F4,$4F,$B3,$A4,$AF,$F1,$74,$7F,$A0,$9B,$19
	.dc.b	$7B,$C7,$17,$7F,$0F,$D3,$D5,$EB,$E8,$C9,$E7,$19,$7A,$C6,$5E,$71
	.dc.b	$E4,$F2,$64,$78,$9F,$3F,$AF,$88,$F0,$C5,$DF,$F2,$98,$FF,$81,$A7
	.dc.b	$FD,$7F,$1F,$C4,$7C,$37,$A9,$FE,$53,$F8,$AF,$E4,$D4,$97,$EA,$AF
	.dc.b	$A7,$FF,$71,$79,$BF,$EC,$89,$CC,$62,$B8,$61,$CE,$B4,$D8,$E9,$7C
	.dc.b	$DF,$F6,$4C,$F3,$8C,$47,$FB,$4E,$38,$FC,$27,$5C,$3C,$ED,$AD,$E3
	.dc.b	$E8,$9F,$1D,$F4,$3C,$75,$C7,$0F,$14,$EA,$7B,$1E,$4C,$75,$5E,$B8
	.dc.b	$7D,$67,$E8,$E7,$A4,$63,$AF,$D5,$8F,$C8,$8C,$4E,$A9,$F4,$5C,$DE
	.dc.b	$53,$E0,$36,$57,$35,$CD,$E3,$1C,$EF,$BE,$9F,$EE,$89,$DB,$63,$33
	.dc.b	$FE,$FC,$9E,$2F,$58,$CB,$E0,$AE,$D1,$89,$DC,$63,$0F,$01,$2B,$CD
	.dc.b	$87,$AE,$F8,$12,$7E,$04,$65,$E1,$5C,$EB,$89,$8C,$3D,$13,$63,$D1
	.dc.b	$5C,$5F,$44,$65,$FA,$71,$9A,$E8,$4A,$F8,$91,$97,$AC,$F2,$3C,$25
	.dc.b	$9E,$0F,$C7,$BF,$85,$5C,$63,$17,$E1,$C3,$E2,$E1,$D3,$0E,$35,$F2
	.dc.b	$E3,$C2,$FC,$70,$78,$8A,$74,$29,$4B,$3F,$CE,$7D,$7C,$C7,$DB,$EE
	.dc.b	$FF,$27,$38,$9D,$36,$32,$F2,$77,$7F,$8B,$6D,$F8,$BE,$4B,$FC,$C3
	.dc.b	$D9,$F9,$79,$CD,$F7,$C7,$07,$E8,$2B,$71,$F9,$79,$DB,$FA,$66,$E3
	.dc.b	$E4,$C7,$F4,$95,$C1,$F1,$5F,$4B,$E4,$F5,$F3,$BF,$C4,$93,$A9,$C9
	.dc.b	$1E,$71,$AC,$BF,$45,$8F,$80,$97,$CF,$2E,$D7,$CC,$95,$F0,$6F,$96
	.dc.b	$1E,$F3,$B4,$74,$BE,$51,$83,$55,$C9,$FB,$B8,$FB,$6B,$84,$6C,$F8
	.dc.b	$95,$CA,$B8,$46,$0D,$57,$17,$AD,$72,$9D,$CE,$67,$84,$72,$CB,$E7
	.dc.b	$57,$AF,$1F,$7B,$C4,$95,$E6,$C7,$DE,$F2,$8F,$E7,$5F,$64,$F1,$FE
	.dc.b	$1F,$0A,$1E,$46,$1D,$63,$FC,$B6,$75,$39,$AD,$36,$75,$5C,$9B,$3A
	.dc.b	$69,$A9,$E1,$87,$F3,$4B,$EB,$3A,$8C,$5F,$37,$F9,$CF,$80,$E7,$7C
	.dc.b	$2F,$F9,$B7,$A3,$FB,$0B,$70,$FF,$5D,$4B,$E0,$F1,$87,$F8,$5E,$38
	.dc.b	$CF,$F7,$EA,$CE,$1D,$5B,$3C,$67,$47,$F8,$13,$E6,$33,$F0,$5A,$6C
	.dc.b	$64,$FF,$04,$C6,$5A,$7F,$83,$3E,$97,$A9,$DB,$CB,$B1,$92,$65,$FE
	.dc.b	$0E,$8C,$BF,$E1,$0F,$2F,$CA,$79,$BE,$62,$D7,$F8,$62,$2C,$FF,$85
	.dc.b	$A6,$ED,$FE,$06,$E3,$F6,$9E,$EF,$62,$E1,$FE,$20,$8B,$7F,$E1,$E9
	.dc.b	$B5,$FE,$0B,$FA,$B7,$D3,$7C,$CE,$44,$63,$1F,$F0,$89,$FE,$09,$5F
	.dc.b	$5B,$B7,$58,$FF,$98,$6B,$8D,$E4,$D5,$5B,$E7,$5F,$E1,$7B,$E3,$87
	.dc.b	$8E,$F7,$5A,$C3,$9B,$69,$EC,$78,$4F,$17,$B1,$67,$51,$71,$F2,$52
	.dc.b	$B1,$8F,$FA,$07,$CF,$7C,$77,$E2,$1F,$1E,$3F,$D0,$B9,$FF,$CC,$73
	.dc.b	$9D,$34,$D9,$FF,$30,$B7,$2F,$F4,$16,$0F,$AD,$FF,$44,$D7,$C0,$8C
	.dc.b	$71,$FE,$69,$D7,$1D,$AD,$DF,$26,$CF,$E9,$5B,$7F,$E4,$47,$FD,$1C
	.dc.b	$F5,$73,$3F,$E9,$E6,$CF,$82,$75,$5D,$EF,$F4,$FE,$B7,$F6,$09,$FF
	.dc.b	$56,$78,$58,$C7,$2D,$8C,$9F,$EB,$18,$CB,$F5,$FE,$13,$FD,$EC,$DE
	.dc.b	$4F,$C6,$47,$F9,$EE,$30,$49,$FC,$44,$70,$6F,$A7,$C2,$BF,$7F,$0F
	.dc.b	$A9,$FF,$1A,$DF,$E1,$7E,$1B,$F2,$6F,$E3,$27,$5E,$4F,$E6,$E7,$31
	.dc.b	$FE,$B5,$33,$3B,$AE,$71,$F9,$9F,$B2,$7C,$27,$82,$73,$C1,$FE,$DC
	.dc.b	$6D,$FF,$81,$1B,$7F,$E7,$1B,$E3,$3D,$71,$FE,$27,$CE,$F9,$6B,$E5
	.dc.b	$61,$B5,$FA,$4E,$7F,$D8,$C6,$31,$FF,$51,$DF,$1A,$D2,$BF,$E3,$FE
	.dc.b	$E7,$FB,$06,$79,$2F,$88,$CF,$9F,$A6,$CE,$9B,$87,$FA,$C2,$B8,$4F
	.dc.b	$E5,$FE,$B9,$E8,$FE,$5D,$B1,$F9,$BF,$6F,$FD,$1C,$EE,$74,$47,$5E
	.dc.b	$95,$C3,$F4,$B7,$C9,$B5,$FA,$3B,$E7,$18,$BF,$89,$F5,$78,$E3,$FF
	.dc.b	$7C,$2F,$83,$FE,$7E,$E6,$7F,$70,$E5,$F9,$9F,$08,$F9,$8F,$E8,$AD
	.dc.b	$8C,$FE,$15,$89,$FE,$93,$D2,$F8,$4A,$7F,$4C,$FA,$BF,$E2,$37,$F5
	.dc.b	$E7,$4F,$D9,$F8,$E7,$DD,$FF,$52,$6C,$66,$F1,$3F,$D5,$BE,$B7,$F8
	.dc.b	$8C,$FE,$05,$BB,$FF,$B8,$9F,$E2,$6F,$FB,$D5,$CE,$E9,$FE,$2D,$EF
	.dc.b	$3E,$6B,$FB,$0B,$F3,$67,$57,$FE,$FD,$8D,$57,$E3,$DD,$3F,$C7,$FE
	.dc.b	$AD,$F6,$1F,$DA,$97,$D8,$D8,$C3,$FD,$BB,$F3,$27,$C7,$53,$1E,$AD
	.dc.b	$C3,$F3,$9F,$AC,$76,$CD,$E1,$A3,$0B,$EB,$BE,$C4,$B6,$F4,$CE,$FB
	.dc.b	$3F,$BE,$FC,$83,$E3,$A9,$8E,$4B,$9F,$EA,$FE,$BF,$FC,$1C,$DF,$7F
	.dc.b	$FC,$0B,$0F,$98,$B1,$FE,$08,$FA,$9D,$EB,$E1,$BF,$E8,$51,$B7,$F6
	.dc.b	$CE,$B5,$F2,$B9,$9E,$07,$BB,$E7,$FF,$9A,$B8,$D9,$18,$AF,$6F,$E1
	.dc.b	$1F,$BD,$AD,$4E,$70,$FE,$A5,$39,$AD,$F9,$9F,$E1,$EF,$CC,$DF,$61
	.dc.b	$FE,$20,$AD,$5F,$F5,$99,$7C,$EE,$EB,$83,$67,$F8,$87,$D8,$7F,$89
	.dc.b	$73,$F3,$7D,$2F,$AF,$3F,$A3,$39,$7E,$9C,$66,$BF,$B3,$4E,$5B,$F1
	.dc.b	$5F,$E3,$28,$FF,$1A,$DF,$C1,$74,$FE,$09,$CD,$F6,$F8,$AE,$1F,$15
	.dc.b	$C5,$FF,$1D,$FE,$66,$F9,$4A,$63,$E5,$6E,$7F,$CA,$67,$4D,$9D,$34
	.dc.b	$99,$7D,$39,$DB,$6B,$9B,$F5,$FE,$C7,$FB,$5F,$21,$A7,$D5,$8F,$CD
	.dc.b	$B8,$F8,$2F,$F2,$8B,$AF,$83,$B1,$FE,$54,$FB,$0F,$ED,$D5,$F8,$36
	.dc.b	$ED,$F9,$E5,$F0,$7B,$99,$A7,$F4,$59,$DB,$FE,$15,$76,$E6,$3F,$C7
	.dc.b	$5F,$33,$FC,$1A,$3F,$C3,$5F,$61,$FA,$CC,$3F,$81,$FC,$6A,$B5,$7F
	.dc.b	$B4,$AD,$5F,$FB,$8A,$B5,$F1,$3F,$E6,$C9,$CD,$27,$35,$FD,$66,$73
	.dc.b	$3F,$EE,$8F,$97,$FF,$75,$FD,$6F,$F3,$18,$FF,$76,$E4,$FA,$FF,$E2
	.dc.b	$B3,$99,$E0,$1C,$4E,$F0,$E7,$EB,$FF,$BB,$B0,$B7,$D3,$0E,$73,$98
	.dc.b	$E4,$FF,$BE,$AF,$94,$60,$FF,$7D,$D7,$27,$F8,$85,$72,$72,$6C,$CD
	.dc.b	$ED,$09,$E5,$5D,$B0,$F6,$4E,$A7,$37,$D6,$3D,$F3,$EE,$8C,$46,$CF
	.dc.b	$05,$73,$C3,$D9,$1E,$03,$CA,$53,$E4,$94,$E4,$67,$EB,$5F,$3A,$FA
	.dc.b	$FD,$4A,$73,$18,$6C,$73,$6B,$CF,$E7,$FF,$9B,$39,$6B,$9C,$39,$46
	.dc.b	$23,$D1,$28,$E3,$97,$E0,$71,$FF,$13,$57,$D3,$6B,$F4,$6B,$85,$FD
	.dc.b	$76,$C7,$D5,$9D,$D7,$3F,$0F,$E3,$3F,$8D,$AD,$B7,$0F,$0D,$7C,$B3
	.dc.b	$89,$6B,$8B,$E8,$8D,$DF,$E0,$2F,$9C,$EE,$3D,$15,$C1,$D5,$70,$7E
	.dc.b	$73,$67,$BC,$7E,$29,$AE,$B0,$FC,$EE,$FF,$B7,$F9,$4F,$CD,$6F,$FE
	.dc.b	$95,$FA,$FF,$F1,$F5,$FF,$92,$5A,$FF,$AC,$F8,$FF,$D4,$EF,$FA,$AD
	.dc.b	$AF,$F5,$38,$DE,$1F,$DB,$9B,$5F,$C0,$6B,$73,$FD,$CB,$E1,$BF,$C1
	.dc.b	$D7,$FE,$57,$6B,$FD,$75,$B1,$BF,$3F,$CF,$3F,$E0,$09,$CD,$FF,$AB
	.dc.b	$73,$74,$EB,$AF,$FC,$6C,$DA,$DD,$FF,$82,$38,$2D,$7E,$5F,$D9,$E1
	.dc.b	$18,$3F,$CA,$F3,$FE,$6B,$71,$1A,$37,$97,$7E,$87,$D1,$52,$5F,$F9
	.dc.b	$CE,$1F,$47,$88,$FE,$F5,$1C,$1D,$4F,$39,$D6,$DF,$E1,$8A,$D5,$70
	.dc.b	$AD,$3F,$E7,$FF,$AF,$7D,$67,$C1,$7B,$C6,$70,$C4,$E5,$91,$93,$94
	.dc.b	$61,$FE,$F3,$F2,$3F,$2F,$C9,$FF,$4D,$C6,$CF,$F0,$97,$D7,$BC,$7F
	.dc.b	$A0,$BC,$AA,$62,$3E,$A7,$D0,$E6,$BF,$D2,$CD,$7F,$C3,$8D,$DB,$FC
	.dc.b	$6F,$77,$9B,$FD,$D5,$E8,$7F,$9D,$EF,$FD,$3C,$FF,$92,$1B,$ED,$FE
	.dc.b	$06,$35,$1F,$E2,$7F,$AF,$FE,$2E,$FF,$A9,$5F,$EB,$0D,$EA,$FC,$44
	.dc.b	$6A,$3F,$C5,$DF,$5F,$FE,$07,$7F,$D5,$8D,$73,$E6,$FE,$3E,$1E,$67
	.dc.b	$E4,$7E,$8D,$E7,$DF,$A1,$98,$7F,$AD,$67,$2F,$FA,$DF,$C8,$FF,$29
	.dc.b	$37,$DA,$FC,$D4,$7F,$A0,$A3,$5E,$DF,$E7,$7A,$3F,$57,$C3,$FD,$23
	.dc.b	$C0,$FC,$0F,$8B,$93,$5F,$3F,$FA,$3F,$AF,$7C,$85,$C6,$65,$FE,$C7
	.dc.b	$9C,$BF,$EC,$9F,$A8,$F1,$BF,$E4,$F5,$F5,$7F,$AD,$6C,$6E,$BF,$61
	.dc.b	$CC,$FA,$FA,$49,$FF,$6A,$43,$EA,$3F,$6D,$F1,$DF,$D3,$FB,$7F,$CE
	.dc.b	$AF,$F3,$66,$FA,$1A,$BF,$DE,$7C,$7E,$F8,$BE,$03,$B6,$CE,$A7,$F8
	.dc.b	$16,$1F,$6E,$FF,$D0,$2D,$AF,$90,$D9,$FF,$45,$BF,$84,$7F,$CD,$DF
	.dc.b	$41,$FC,$69,$FF,$47,$AF,$27,$F8,$73,$63,$FD,$E0,$ED,$74,$7F,$A1
	.dc.b	$5F,$F0,$BD,$FF,$6F,$8C,$47,$F9,$F8,$DF,$C2,$7F,$1E,$78,$FB,$F6
	.dc.b	$77,$97,$FA,$CA,$76,$74,$3F,$C1,$31,$AB,$ED,$5C,$6E,$56,$CE,$27
	.dc.b	$86,$BF,$CF,$AF,$3E,$7F,$F3,$37,$91,$FC,$AF,$C0,$7D,$7D,$31,$12
	.dc.b	$B8,$4E,$9B,$39,$7F,$D3,$1F,$5E,$F8,$BF,$E6,$CB,$E8,$69,$88,$93
	.dc.b	$E3,$9D,$36,$72,$FC,$9F,$AF,$FC,$63,$ED,$6B,$9E,$2F,$95,$D9,$F3
	.dc.b	$67,$3C,$3F,$2F,$6F,$CD,$DF,$C3,$EB,$D6,$12,$3E,$44,$EA,$B8,$56
	.dc.b	$67,$FD,$6F,$EC,$3E,$5F,$F6,$B1,$FE,$5D,$6C,$67,$C0,$FE,$B6,$EF
	.dc.b	$C0,$B3,$A2,$4E,$23,$FA,$E7,$D7,$FF,$1C,$C7,$FD,$B6,$FF,$21,$C3
	.dc.b	$FD,$88,$D8,$CF,$CD,$70,$38,$F9,$92,$BE,$F5,$73,$E9,$FC,$A7,$CB
	.dc.b	$FE,$3C,$93,$F8,$7F,$8F,$FC,$86,$3F,$1E,$73,$3F,$89,$7F,$31,$3A
	.dc.b	$AE,$73,$F2,$27,$75,$F9,$9A,$D4,$FF,$BE,$BE,$95,$E8,$FF,$79,$7F
	.dc.b	$53,$1F,$93,$6B,$FE,$E1,$76,$E6,$38,$70,$C8,$C1,$FA,$5E,$07,$CF
	.dc.b	$3E,$23,$E8,$B6,$74,$D7,$ED,$BF,$E0,$C7,$31,$C0,$8F,$0F,$7B,$E0
	.dc.b	$FD,$74,$8F,$90,$FC,$4F,$A0,$C4,$7E,$FF,$44,$E3,$5F,$89,$57,$7F
	.dc.b	$F3,$A5,$6A,$F9,$9F,$4E,$7F,$C3,$DF,$53,$FE,$04,$B5,$FF,$43,$8C
	.dc.b	$37,$DD,$DF,$B9,$C1,$98,$7E,$FE,$B6,$BC,$34,$7F,$82,$A1,$87,$67
	.dc.b	$5F,$4F,$FD,$21,$CE,$3F,$D4,$E3,$13,$BB,$46,$27,$5E,$77,$F8,$DE
	.dc.b	$31,$7D,$3B,$F0,$66,$7F,$C7,$9F,$28,$FA,$FA,$4A,$FE,$31,$1F,$E8
	.dc.b	$36,$C6,$5F,$29,$ED,$8E,$4F,$0F,$87,$EB,$EA,$F6,$F4,$64,$F4,$8C
	.dc.b	$BD,$A3,$2F,$4D,$17,$FD,$2D,$BE,$4F,$A0,$93,$E6,$F6,$6B,$C6,$BF
	.dc.b	$C1,$9B,$FC,$76,$EF,$FD,$F5,$FF,$06,$CF,$E6,$FD,$4F,$F2,$AB,$FE
	.dc.b	$C7,$F2,$BF,$DD,$DF,$3A,$FA,$3F,$F7,$44,$61,$B3,$FE,$EF,$AE,$3F
	.dc.b	$1B,$FC,$CE,$31,$F0,$BA,$EF,$63,$CE,$7F,$DA,$18,$F3,$69,$C0,$B2
	.dc.b	$C3,$AC,$EE,$7C,$C7,$FB,$FE,$BA,$38,$AE,$17,$E1,$9C,$E1,$E5,$9F
	.dc.b	$1E,$39,$3C,$D8,$F6,$C3,$FD,$D3,$F5,$EE,$BF,$46,$1E,$D9,$4E,$6F
	.dc.b	$D3,$18,$AD,$E7,$FE,$E3,$A7,$D3,$FC,$8A,$D5,$F0,$9E,$51,$EC,$32
	.dc.b	$57,$8B,$BB,$EA,$F9,$9A,$F0,$C3,$CD,$3C,$23,$94,$EA,$3B,$CE,$A3
	.dc.b	$17,$E5,$9C,$E1,$E9,$9E,$06,$AF,$C5,$84,$7A,$BE,$BB,$E5,$3C,$23
	.dc.b	$58,$79,$27,$31,$8B,$EC,$F1,$BC,$1E,$D6,$BF,$42,$78,$BE,$02,$57
	.dc.b	$57,$78,$7A,$9E,$47,$CB,$9E,$33,$F0,$6B,$71,$8B,$F1,$D7,$47,$ED
	.dc.b	$46,$63,$9D,$71,$7A,$1F,$0C,$1C,$8A,$78,$8A,$7D,$E2,$DF,$F3,$AF
	.dc.b	$92,$CC,$7E,$4B,$BD,$18,$8F,$C8,$C6,$1D,$CF,$16,$C6,$6B,$F0,$6D
	.dc.b	$C3,$C7,$F4,$DF,$E6,$9E,$1F,$C9,$4E,$6F,$C1,$8E,$0F,$CE,$D6,$E3
	.dc.b	$F2,$93,$B7,$C7,$3B,$7F,$35,$3B,$C3,$C3,$39,$6C,$EB,$CF,$FE,$85
	.dc.b	$18,$26,$3F,$A9,$C3,$B6,$1C,$E3,$18,$73,$6E,$1C,$F1,$C3,$EA,$3C
	.dc.b	$F3,$EA,$C7,$BD,$70,$C7,$BD,$6D,$FD,$F6,$1D,$E3,$13,$B2,$3F,$7A
	.dc.b	$34,$FC,$23,$B9,$27,$9D,$78,$23,$99,$0E,$11,$AC,$3D,$31,$8B,$EE
	.dc.b	$AF,$BE,$5F,$7E,$FD,$33,$B3,$C9,$87,$77,$04,$9E,$17,$F8,$29,$D4
	.dc.b	$62,$5B,$BE,$BF,$FD,$23,$97,$F3,$46,$CF,$F2,$F7,$FC,$5F,$19,$F1
	.dc.b	$BD,$D5,$C7,$C7,$B5,$C3,$3F,$E6,$4F,$0C,$39,$5E,$AB,$6D,$C3,$A1
	.dc.b	$E1,$8C,$B4,$CC,$FF,$44,$E9,$FE,$D3,$F0,$FF,$E0,$DE,$8F,$EC,$4D
	.dc.b	$AF,$EC,$5F,$11,$6F,$FB,$D4,$DB,$E3,$5F,$E0,$28,$C7,$4A,$EC,$63
	.dc.b	$E1,$33,$F5,$F6,$FF,$C1,$2D,$C7,$FB,$A3,$72,$FE,$E8,$F4,$F8,$59
	.dc.b	$5F,$E1,$0F,$2E,$DF,$F7,$B9,$B7,$FD,$EE,$6D,$7F,$7B,$F2,$FC,$A7
	.dc.b	$9B,$EB,$EC,$7F,$86,$A2,$E1,$FE,$05,$9B,$7F,$E0,$5E,$6F,$69,$EE
	.dc.b	$F6,$2E,$1F,$E2,$28,$B7,$FE,$0D,$9B,$5F,$E3,$9E,$57,$D5,$3E,$BB
	.dc.b	$E6,$46,$1B,$0F,$AF,$B1,$32,$FF,$36,$4F,$16,$CE,$71,$FF,$51,$CE
	.dc.b	$5B,$5D,$27,$58,$78,$E9,$EC,$78,$0F,$0F,$B1,$71,$EF,$17,$1E,$B3
	.dc.b	$6F,$FC,$9D,$E7,$FF,$BD,$6F,$FC,$B9,$97,$E1,$63,$17,$FE,$37,$7C
	.dc.b	$95,$C3,$E0,$B5,$5B,$6D,$FF,$BB,$65,$F5,$FF,$E8,$98,$C3,$70,$FF
	.dc.b	$26,$CE,$B8,$7F,$2C,$F0,$E2,$B5,$B8,$FC,$F3,$6F,$FA,$8C,$71,$7A
	.dc.b	$CF,$02,$3E,$0A,$B7,$FE,$A3,$3F,$45,$EB,$7F,$50,$9F,$F5,$44,$63
	.dc.b	$C2,$CB,$63,$27,$FA,$B6,$3F,$03,$F0,$7F,$EF,$38,$FF,$54,$36,$7F
	.dc.b	$DB,$AF,$F9,$EA,$33,$C7,$C0,$8F,$0F,$4E,$D7,$F9,$C0,$7D,$7F,$F8
	.dc.b	$B2,$7F,$D6,$9F,$0D,$F8,$D7,$E0,$9F,$DC,$23,$53,$FE,$EC,$9D,$3C
	.dc.b	$E3,$3A,$FE,$6E,$72,$FF,$A5,$67,$BF,$91,$F9,$C9,$CD,$F5,$8F,$E1
	.dc.b	$53,$A8,$FF,$63,$7D,$7B,$E1,$33,$8F,$69,$FF,$02,$F8,$5F,$A8,$F8
	.dc.b	$7F,$E2,$1D,$3F,$CF,$67,$D3,$8F,$FA,$A6,$13,$AF,$2B,$6D,$AD,$DF
	.dc.b	$36,$D7,$3B,$FF,$46,$7B,$9F,$E5,$B9,$DA,$F9,$5F,$E4,$0E,$67,$71
	.dc.b	$8F,$3F,$2D,$AE,$2F,$E5,$3E,$BD,$E9,$D3,$F2,$FE,$DF,$FA,$26,$BF
	.dc.b	$41,$3A,$6C,$ED,$FD,$24,$EF,$0F,$D0,$D6,$DB,$5B,$7F,$4B,$3B,$E3
	.dc.b	$FE,$7D,$5B,$7F,$D8,$DF,$5F,$9C,$3E,$7A,$F8,$3F,$E7,$CE,$70,$FF
	.dc.b	$6B,$39,$9F,$97,$F0,$8F,$98,$FE,$88,$D8,$CF,$D7,$E2,$7F,$A4,$74
	.dc.b	$3E,$13,$FA,$57,$D7,$FF,$BA,$AF,$EA,$CE,$5F,$AF,$19,$CB,$FD,$81
	.dc.b	$F1,$CF,$92,$FE,$A4,$BE,$CA,$C6,$1F,$EA,$DF,$61,$FC,$4B,$83,$FD
	.dc.b	$C8,$DC,$7F,$1A,$FF,$14,$6F,$8F,$FD,$77,$DF,$7C,$DD,$E6,$C3,$6B
	.dc.b	$F1,$7F,$60,$FB,$0F,$EC,$8B,$EC,$2C,$61,$FE,$D1,$F6,$0F,$90,$A6
	.dc.b	$3D,$5B,$87,$E5,$BE,$C3,$EB,$3F,$A2,$FB,$0F,$A5,$E1,$A7,$3C,$A8
	.dc.b	$4E,$BE,$45,$F1,$FF,$DD,$D7,$C9,$D8,$C3,$FD,$EF,$EC,$3F,$DD,$79
	.dc.b	$7E,$99,$B8,$FF,$BB,$D7,$CC,$6F,$E0,$FF,$C0,$5F,$61,$AE,$96,$1F
	.dc.b	$B8,$8D,$BF,$B0,$74,$E6,$BF,$C1,$5F,$40,$F5,$EF,$C0,$61,$B5,$FB
	.dc.b	$8F,$B0,$FD,$EB,$7C,$FF,$F0,$97,$D8,$3E,$C3,$FC,$29,$5A,$DD,$AF
	.dc.b	$91,$F8,$D4,$F9,$EE,$37,$C9,$B8,$FF,$06,$FB,$0F,$F0,$D3,$FC,$23
	.dc.b	$E9,$BD,$78,$74,$8F,$E2,$B3,$A8,$C1,$F4,$E7,$78,$F8,$79,$D5,$B8
	.dc.b	$C3,$6F,$9F,$C5,$7F,$8B,$63,$18,$7D,$D5,$19,$F8,$BD,$36,$BF,$8E
	.dc.b	$FD,$83,$E5,$FF,$C6,$CD,$8C,$FC,$BE,$26,$47,$07,$71,$8C,$7D,$79
	.dc.b	$AF,$94,$61,$B9,$FB,$55,$71,$8F,$81,$F6,$1F,$DB,$EB,$F2,$CD,$B3
	.dc.b	$F8,$1C,$7F,$9A,$B9,$27,$C2,$FF,$93,$DD,$57,$91,$E5,$F6,$1F,$EC
	.dc.b	$AA,$FE,$7B,$F6,$4F,$A4,$FF,$2B,$4E,$6F,$D7,$39,$AF,$E8,$73,$9C
	.dc.b	$BE,$B3,$63,$FA,$8B,$FD,$F6,$BE,$1B,$97,$FD,$03,$F3,$9F,$C2,$DF
	.dc.b	$D9,$7D,$8F,$F2,$57,$F0,$7F,$22,$9D,$6A,$AD,$63,$FE,$45,$9C,$C6
	.dc.b	$31,$FF,$72,$DF,$1F,$3F,$FC,$E1,$39,$C7,$FD,$CB,$F3,$1F,$EE,$AF
	.dc.b	$AF,$FE,$65,$1F,$E1,$CF,$B1,$FD,$04,$62,$3F,$DD,$B3,$BF,$6B,$FD
	.dc.b	$DF,$A2,$F7,$1F,$EE,$C9,$D5,$73,$CE,$BC,$86,$1D,$70,$E5,$3A,$9F
	.dc.b	$F7,$ED,$70,$9C,$C7,$0B,$EC,$D3,$B1,$4E,$D7,$D7,$0F,$2C,$7C,$03
	.dc.b	$55,$D2,$7D,$D1,$E0,$33,$3C,$1E,$75,$F0,$23,$11,$B0,$79,$0A,$77
	.dc.b	$29,$F0,$4B,$F5,$EF,$A1,$7D,$8E,$A7,$9C,$E6,$31,$08,$C3,$B6,$BD
	.dc.b	$3E,$8F,$F8,$CB,$96,$B9,$CB,$9C,$A7,$33,$E8,$8C,$4F,$D8,$C7,$FC
	.dc.b	$4B,$37,$C2,$F6,$E7,$F8,$5B,$DC,$FD,$16,$C7,$E2,$6B,$8C,$FD,$56
	.dc.b	$C7,$1C,$3C,$ED,$C3,$E5,$FC,$17,$1C,$7E,$5B,$70,$CE,$1E,$7C,$3D
	.dc.b	$F3,$C6,$7B,$57,$9A,$BE,$99,$E0,$38,$5F,$E5,$31,$F3,$C6,$A7,$85
	.dc.b	$7D,$C8,$E0,$FD,$4C,$3C,$F1,$FE,$4E,$FB,$17,$0F,$E9,$38,$30,$DB
	.dc.b	$FE,$DD,$3A,$7F,$C9,$6F,$FA,$ED,$B1,$F7,$B9,$13,$B9,$FC,$4C,$62
	.dc.b	$75,$39,$9F,$C6,$E1,$CD,$FF,$3D,$37,$3F,$F3,$73,$70,$FF,$32,$FD
	.dc.b	$A3,$E9,$3C,$D7,$F8,$3F,$97,$FC,$7B,$4F,$C8,$DF,$57,$BC,$6B,$F2
	.dc.b	$F7,$D6,$D2,$4F,$E6,$23,$2D,$8F,$89,$C6,$FB,$AC,$CF,$F6,$28,$C4
	.dc.b	$F9,$2B,$3A,$7F,$A1,$D7,$D1,$E4,$97,$FE,$7C,$B7,$43,$C5,$CC,$FE
	.dc.b	$B6,$F8,$BE,$5B,$F5,$11,$E5,$93,$E2,$3C,$ED,$8F,$F4,$48,$AE,$6F
	.dc.b	$F6,$0F,$B4,$7D,$CF,$E0,$79,$51,$89,$FF,$45,$B5,$E1,$F7,$5B,$8C
	.dc.b	$47,$F0,$A9,$E3,$5F,$85,$9C,$CF,$FA,$79,$AF,$F8,$EF,$F2,$6F,$B8
	.dc.b	$FF,$0B,$A4,$62,$3F,$16,$D7,$B3,$5F,$F5,$4F,$DB,$BE,$9F,$F2,$8D
	.dc.b	$8E,$8D,$7F,$2A,$D7,$D0,$BE,$9F,$89,$23,$8B,$F9,$FF,$59,$F0,$14
	.dc.b	$93,$E9,$8C,$BE,$BF,$B7,$FF,$0B,$37,$CF,$FD,$44,$62,$3F,$4E,$D7
	.dc.b	$E5,$7D,$BF,$F8,$89,$F8,$8D,$7F,$5A,$D7,$7E,$8F,$EC,$F9,$1E,$7D
	.dc.b	$24,$FC,$D8,$CB,$F3,$FC,$7F,$F4,$63,$7E,$33,$F7,$91,$88,$FD,$DB
	.dc.b	$5F,$BD,$F6,$EF,$AF,$FE,$04,$D8,$FB,$4D,$7F,$81,$B5,$FC,$37,$DB
	.dc.b	$FF,$9D,$9F,$C0,$35,$FE,$12,$D7,$7F,$05,$FC,$3F,$EC,$7F,$D4,$91
	.dc.b	$F8,$76,$BF,$C4,$1A,$FE,$4D,$AF,$FB,$BB,$D1,$7C,$17,$F1,$96,$C7
	.dc.b	$E4,$5A,$FF,$1A,$6B,$F9,$DF,$92,$7C,$F7,$F2,$16,$C7,$E6,$5A,$FF
	.dc.b	$22,$6B,$FA,$5F,$95,$AF,$FC,$00,$6B,$FE,$9D,$6F,$59,$BF,$63,$F8
	.dc.b	$59,$FE,$F3,$28,$C4,$7F,$A9,$5B,$19,$E6,$FE,$6D,$1F,$87,$9E,$2D
	.dc.b	$78,$72,$68,$FE,$4D,$E7,$FF,$1B,$24,$7F,$17,$F1,$7F,$69,$E3,$FF
	.dc.b	$1D,$9D,$CF,$F2,$96,$BF,$C6,$5F,$E5,$3F,$0F,$FE,$A1,$8F,$E4,$0D
	.dc.b	$7F,$D1,$67,$F1,$58,$FE,$25,$8F,$F8,$BF,$CE,$F7,$B6,$7F,$D8,$4D
	.dc.b	$CF,$FD,$05,$E3,$F0,$B9,$5F,$E1,$EC,$7F,$D2,$97,$B3,$F9,$24,$7F
	.dc.b	$A0,$3E,$DD,$F4,$BF,$C8,$5F,$F2,$34,$71,$CB,$E7,$C6,$23,$84,$EC
	.dc.b	$E1,$EB,$C8,$E0,$7F,$B5,$E7,$AB,$8F,$65,$F5,$7F,$E6,$78,$C3,$5F
	.dc.b	$F3,$5B,$72,$FF,$5E,$4F,$0F,$7F,$FB,$B3,$CB,$9B,$FC,$5E,$DB,$FF
	.dc.b	$1A,$35,$FF,$1D,$B5,$FF,$1F,$B9,$AE,$B7,$FC,$4D,$BA,$76,$8F,$16
	.dc.b	$9F,$90,$FA,$27,$CF,$FF,$47,$6E,$3F,$03,$85,$18,$8F,$F5,$6B,$FE
	.dc.b	$F6,$74,$E7,$1F,$F6,$43,$6F,$B5,$F5,$F8,$DF,$EA,$9F,$6D,$FC,$B3
	.dc.b	$0F,$F5,$BA,$FA,$8C,$9A,$F2,$64,$F8,$E7,$4D,$9C,$BF,$90,$F4,$7F
	.dc.b	$6F,$9F,$F7,$7E,$DF,$EF,$47,$F8,$03,$F0,$57,$DE,$62,$FF,$C0,$D3
	.dc.b	$FE,$0B,$23,$83,$FB,$7D,$BE,$B3,$FC,$08,$71,$8F,$F6,$7F,$D6,$FF
	.dc.b	$81,$B7,$FD,$DB,$FB,$B8,$F9,$4D,$77,$EC,$FF,$75,$8E,$10,$E2,$46
	.dc.b	$F4,$FF,$79,$36,$FA,$DF,$DB,$F8,$29,$87,$E4,$1F,$BD,$F2,$F8,$8F
	.dc.b	$C8,$C5,$C3,$F0,$B3,$FE,$30,$FC,$43,$FB,$35,$7F,$83,$1A,$FE,$E9
	.dc.b	$B2,$C3,$FD,$1F,$EE,$38,$3F,$C7,$97,$FE,$92,$BE,$2F,$FA,$41,$AF
	.dc.b	$FA,$51,$AE,$FD,$4F,$F1,$FC,$76,$BE,$BE,$13,$ED,$FF,$87,$9E,$7A
	.dc.b	$FF,$0E,$75,$7F,$86,$73,$F2,$63,$F2,$63,$18,$7F,$2C,$6C,$67,$DB
	.dc.b	$E3,$ED,$C8,$FC,$F4,$EB,$8B,$F7,$75,$FE,$29,$F2,$FC,$55,$FA,$36
	.dc.b	$AC,$7C,$35,$FD,$4B,$3E,$1E,$FF,$16,$57,$17,$2F,$EB,$7C,$5F,$EC
	.dc.b	$8F,$F4,$26,$BF,$83,$6D,$7F,$1C,$F3,$BF,$CE,$D1,$FD,$B7,$0F,$0B
	.dc.b	$FD,$A1,$BF,$45,$FE,$41,$8D,$6D,$FD,$3B,$F1,$0F,$F6,$1D,$7A,$1B
	.dc.b	$19,$C7,$FC,$91,$1F,$9F,$7F,$81,$3F,$E7,$4C,$4D,$24,$7F,$A2,$5F
	.dc.b	$F5,$7B,$FE,$49,$6B,$E1,$F8,$6D,$11,$EB,$D9,$E0,$3C,$31,$FE,$84
	.dc.b	$FB,$87,$D5,$FF,$93,$3C,$EF,$E1,$11,$8F,$85,$FF,$7B,$F8,$9F,$EE
	.dc.b	$08,$C6,$3F,$E7,$28,$C7,$D8,$7F,$B9,$2F,$8E,$1C,$E7,$83,$6B,$FD
	.dc.b	$D7,$7C,$AB,$BD,$F1,$BD,$E1,$E0,$BD,$E1,$C5,$61,$C6,$BC,$31,$B3
	.dc.b	$33,$B9,$D4,$72,$AD,$42,$FC,$13,$93,$D8,$5C,$3D,$75,$C2,$77,$3D
	.dc.b	$A3,$9D,$F7,$3C,$13,$AB,$F0,$D6,$CE,$A5,$BF,$82,$F2,$B8,$EA,$78
	.dc.b	$9E,$77,$33,$E4,$75,$7E,$07,$38,$7C,$A7,$37,$E9,$72,$7B,$63,$55
	.dc.b	$E2,$8D,$91,$EC,$64,$B1,$C0,$91,$F7,$30,$F8,$17,$F0,$EF,$BC,$FD
	.dc.b	$89,$CC,$63,$0F,$6C,$62,$7C,$17,$D1,$E0,$6E,$FA,$3F,$16,$78,$61
	.dc.b	$E4,$8F,$61,$E3,$3E,$58,$3E,$21,$4F,$C0,$14,$FA,$C5,$BF,$F3,$C7
	.dc.b	$83,$98,$FC,$CF,$B1,$FB,$97,$4D,$7F,$4C,$D7,$71,$F9,$08,$FE,$01
	.dc.b	$5F,$DD,$74,$FF,$2A,$6F,$9A,$97,$F8,$1B,$DE,$7E,$7B,$DC,$E6,$B7
	.dc.b	$87,$EE,$2F,$73,$EF,$AD,$CE,$6B,$77,$E6,$7F,$51,$3A,$8C,$57,$52
	.dc.b	$57,$9F,$D3,$FE,$86,$F3,$8C,$CF,$08,$D6,$3D,$A7,$E0,$9A,$3C,$95
	.dc.b	$CA,$B8,$12,$37,$5C,$9E,$07,$F0,$0C,$7D,$AD,$C7,$D3,$7C,$E3,$8D
	.dc.b	$F3,$74,$7E,$C3,$1F,$73,$A2,$1A,$C3,$C0,$7B,$49,$3C,$E7,$8C,$EE
	.dc.b	$3B,$65,$F8,$68,$ED,$7E,$37,$47,$68,$D5,$F6,$AE,$38,$FB,$27,$31
	.dc.b	$F7,$EF,$94,$62,$F9,$57,$C2,$C3,$C7,$39,$9E,$E4,$8C,$CF,$52,$63
	.dc.b	$F2,$32,$FA,$DC,$1F,$D4,$3E,$BF,$FB,$FF,$17,$F5,$9F,$90,$B5,$FD
	.dc.b	$76,$6E,$5F,$E2,$BE,$35,$63,$0E,$57,$FA,$76,$D7,$EB,$5A,$FF,$59
	.dc.b	$BF,$F0,$6B,$FD,$6F,$D3,$AF,$F6,$B6,$BD,$DB,$3F,$D9,$9B,$81,$F1
	.dc.b	$36,$FF,$B6,$FA,$98,$9C,$FD,$BE,$BD,$EB,$5F,$DC,$9B,$8F,$9F,$05
	.dc.b	$76,$25,$7B,$7B,$5F,$6F,$F0,$5B,$1F,$E1,$BF,$80,$C4,$C8,$FF,$0F
	.dc.b	$3B,$8C,$61,$FE,$1F,$9F,$F6,$BD,$3C,$9E,$87,$5F,$26,$CF,$F8,$9E
	.dc.b	$2D,$71,$3F,$CB,$D4,$AC,$F0,$FC,$EF,$C3,$DE,$D7,$DA,$F2,$DF,$06
	.dc.b	$F2,$3E,$EE,$4B,$DE,$9F,$EB,$6B,$DD,$6A,$73,$5F,$C9,$5A,$FF,$2B
	.dc.b	$6B,$FC,$BD,$AF,$F3,$37,$8E,$3E,$5C,$7C,$9E,$00,$EC,$FE,$43,$D7
	.dc.b	$FF,$12,$F8,$37,$8F,$FC,$33,$A7,$FB,$1D,$7D,$86,$66,$4E,$5E,$11
	.dc.b	$8C,$FE,$47,$47,$FB,$3A,$7F,$35,$E5,$72,$BF,$F2,$CC,$62,$F2,$D8
	.dc.b	$C9,$FE,$5E,$8F,$89,$E2,$FC,$5F,$22,$CF,$FB,$7B,$1F,$99,$18,$7E
	.dc.b	$44,$FE,$79,$FA,$0D,$DF,$FD,$DF,$5F,$D4,$67,$2F,$38,$CC,$7E,$07
	.dc.b	$85,$AF,$DB,$AD,$5F,$ED,$7D,$47,$CE,$DF,$9E,$C3,$6B,$FD,$84,$DC
	.dc.b	$7F,$D9,$2D,$E3,$FF,$39,$FD,$DB,$EE,$FF,$80,$36,$FF,$DB,$88,$CC
	.dc.b	$48,$CB,$C2,$31,$7F,$99,$CD,$D3,$FE,$E8,$BF,$EF,$18,$7F,$BF,$21
	.dc.b	$E6,$FE,$0D,$D5,$F1,$73,$5C,$0F,$F0,$05,$F4,$F6,$BF,$85,$B6,$3F
	.dc.b	$E0,$0C,$F2,$AF,$F7,$FB,$5E,$2F,$6C,$3C,$57,$E1,$F6,$BC,$4F,$F2
	.dc.b	$1F,$6F,$FC,$1D,$87,$FB,$E3,$C7,$C6,$39,$6C,$64,$FF,$7E,$C6,$5F
	.dc.b	$43,$FE,$D7,$FB,$37,$DC,$7F,$2D,$6C,$EB,$A9,$0F,$2B,$55,$2B,$53
	.dc.b	$C2,$FF,$3E,$EE,$33,$3F,$BB,$F0,$7F,$82,$7E,$94,$FB,$2C,$57,$F0
	.dc.b	$18,$C3,$F9,$19,$FC,$B1,$1F,$CB,$75,$7F,$11,$6C,$72,$C7,$F2,$DC
	.dc.b	$4F,$94,$C7,$17,$FA,$FB,$EC,$DE,$AF,$F3,$E6,$E3,$FD,$2D,$B8,$7F
	.dc.b	$4B,$6B,$FE,$3B,$E0,$43,$CE,$E4,$4A,$E4,$FF,$4A,$F8,$77,$AF,$49
	.dc.b	$3F,$D6,$63,$2F,$F5,$BF,$88,$74,$7F,$52,$6C,$66,$30,$7F,$56,$FB
	.dc.b	$2F,$E2,$B9,$FF,$61,$6B,$E7,$6B,$FE,$55,$CD,$18,$84,$3E,$C9,$F5
	.dc.b	$DF,$D8,$9B,$87,$F6,$F6,$D7,$F6,$F6,$BF,$17,$EC,$1F,$7D,$FD,$A9
	.dc.b	$B5,$FD,$DD,$B3,$FD,$DD,$AF,$CD,$FB,$0F,$E6,$93,$FD,$F5,$B1,$FD
	.dc.b	$F5,$AE,$FC,$EF,$EE,$FF,$61,$FE,$AF,$AF,$F0,$23,$67,$FC,$08,$D7
	.dc.b	$ED,$35,$F8,$9F,$5E,$FA,$BF,$F0,$0B,$6B,$FC,$1C,$D9,$FF,$07,$35
	.dc.b	$FC,$27,$C2,$BE,$23,$FC,$12,$DA,$FF,$0B,$36,$7F,$C2,$CD,$7F,$19
	.dc.b	$F5,$F5,$FA,$CD,$7E,$AB,$78,$0D,$F9,$9F,$F0,$1C,$FF,$88,$5A,$FE
	.dc.b	$5D,$AF,$E0,$5A,$FE,$03,$EB,$5F,$5F,$FE,$20,$6E,$3F,$E2,$D6,$CF
	.dc.b	$F8,$B5,$AF,$E8,$DF,$E5,$71,$BF,$27,$91,$1F,$F1,$B7,$CF,$3E,$1B
	.dc.b	$FC,$5A,$D8,$CC,$60,$FF,$18,$FD,$63,$E4,$E9,$27,$FC,$81,$19,$6C
	.dc.b	$7F,$91,$63,$18,$FF,$70,$5F,$2B,$99,$FF,$08,$FE,$04,$FB,$4F,$E2
	.dc.b	$7F,$61,$86,$E3,$FD,$AE,$74,$FF,$65,$6E,$3F,$EB,$76,$BF,$EB,$D6
	.dc.b	$BF,$EC,$17,$7E,$2F,$F9,$49,$B3,$FE,$C7,$F6,$53,$FD,$06,$38,$5F
	.dc.b	$9B,$ED,$3F,$93,$FC,$6D,$9F,$E4,$CD,$C7,$C5,$3F,$DF,$71,$34,$FF
	.dc.b	$2D,$FD,$3F,$F7,$4B,$FF,$34,$36,$BF,$CD,$0D,$7F,$8F,$35,$FE,$10
	.dc.b	$EF,$14,$E3,$D9,$FF,$3A,$4B,$DA,$B1,$FE,$77,$FA,$8F,$87,$87,$F0
	.dc.b	$E6,$F4,$9A,$4D,$78,$1F,$E7,$FF,$AC,$FE,$03,$9F,$82,$F9,$4F,$F9
	.dc.b	$29,$FF,$78,$7C,$8F,$FB,$C7,$EB,$3F,$CE,$D1,$FD,$F3,$E7,$F8,$46
	.dc.b	$23,$FD,$E9,$3B,$F8,$FF,$F7,$B6,$0B,$DC,$7F,$BC,$E7,$55,$CF,$7A
	.dc.b	$F2,$18,$75,$C3,$94,$EA,$7C,$A7,$F9,$BE,$FB,$E1,$EB,$F8,$BF,$24
	.dc.b	$E9,$E5,$19,$9F,$14,$F6,$78,$D7,$58,$63,$E4,$C3,$C3,$1B,$07,$9F
	.dc.b	$1E,$B5,$C6,$B8,$61,$F0,$2B,$67,$A0,$FF,$22,$47,$F9,$15,$C1,$E4
	.dc.b	$36,$3E,$B1,$F4,$8E,$6F,$F2,$64,$61,$B1,$99,$E2,$71,$70,$F4,$97
	.dc.b	$CB,$FE,$9A,$37,$18,$CF,$AC,$E6,$78,$35,$F5,$CE,$DF,$88,$7F,$AE
	.dc.b	$FD,$5F,$7E,$9F,$85,$D7,$FC,$0B,$3F,$7A,$B8,$4F,$E2,$5B,$3F,$18
	.dc.b	$DC,$F4,$6C,$EE,$FD,$0D,$C3,$F1,$33,$A3,$96,$1E,$27,$CF,$5C,$1B
	.dc.b	$3C,$27,$57,$F7,$27,$93,$F7,$B0,$F2,$B6,$F9,$3B,$AE,$2F,$C1,$3D
	.dc.b	$D1,$AE,$2F,$DA,$7D,$67,$E9,$37,$FF,$04,$71,$A3,$17,$FB,$B6,$D7
	.dc.b	$EE,$FD,$7B,$5B,$AF,$F2,$57,$23,$E1,$F6,$D8,$FF,$27,$4E,$DF,$F6
	.dc.b	$71,$F8,$0A,$EE,$D3,$FD,$3D,$97,$F8,$F5,$FF,$2B,$74,$7F,$AE,$DC
	.dc.b	$CF,$93,$6F,$E2,$91,$FE,$5B,$6B,$DA,$B9,$3E,$0F,$9E,$FE,$F4,$FF
	.dc.b	$B4,$FE,$AE,$BF,$EA,$16,$F0,$FF,$13,$84,$E3,$7F,$D4,$B6,$3F,$45
	.dc.b	$57,$90,$E6,$38,$F9,$5E,$EA,$7C,$AD,$24,$FF,$B9,$79,$DF,$59,$88
	.dc.b	$F2,$7C,$BF,$D4,$CB,$AB,$9D,$3C,$D8,$F4,$AF,$64,$EF,$33,$BD,$3F
	.dc.b	$C8,$DF,$59,$FE,$F7,$AF,$F4,$13,$5F,$F6,$DB,$70,$FE,$69,$CA,$87
	.dc.b	$CE,$FE,$11,$D5,$FF,$94,$7C,$9F,$AD,$87,$FB,$91,$B8,$72,$AD,$CF
	.dc.b	$F9,$52,$73,$5F,$E5,$9F,$B1,$7C,$DF,$E4,$1B,$5F,$E9,$7A,$DB,$FE
	.dc.b	$5C,$FA,$C7,$D6,$FE,$59,$7D,$0D,$31,$16,$7F,$D4,$1F,$3D,$C1,$B1
	.dc.b	$97,$FD,$F7,$E9,$FF,$86,$DF,$F5,$32,$FA,$6F,$D1,$B6,$33,$E8,$FE
	.dc.b	$9B,$E7,$7F,$C3,$91,$FE,$AD,$7F,$C2,$AD,$F4,$75,$E8,$FE,$B3,$EB
	.dc.b	$5F,$3B,$7A,$E5,$FF,$AD,$67,$2F,$FA,$DF,$B5,$F3,$FF,$B4,$7F,$DE
	.dc.b	$71,$9F,$3F,$13,$FB,$8F,$31,$F5,$B7,$E4,$25,$FF,$B0,$67,$2F,$FB
	.dc.b	$0F,$EB,$7F,$D8,$4D,$F9,$EF,$E0,$30,$6F,$CA,$FF,$04,$FA,$DF,$F7
	.dc.b	$5D,$7F,$B2,$DF,$F5,$1B,$7E,$3B,$5F,$35,$FC,$2B,$EB,$73,$3F,$E2
	.dc.b	$4F,$03,$FD,$31,$1E,$28,$C1,$FE,$98,$71,$E1,$7E,$49,$E2,$FE,$53
	.dc.b	$EB,$3F,$DF,$F3,$FE,$DB,$7F,$CB,$ED,$F8,$2F,$E3,$31,$A8,$EB,$E3
	.dc.b	$3E,$41,$F4,$DF,$C7,$9F,$F6,$04,$FF,$B9,$5A,$F0,$76,$FF,$BA,$7E
	.dc.b	$4A,$BD,$1D,$E3,$FA,$86,$E7,$FE,$B5,$77,$1F,$AC,$F8,$1F,$C3,$D7
	.dc.b	$91,$FC,$3B,$7B,$7F,$98,$C6,$A7,$87,$67,$F3,$4F,$BA,$FE,$3B,$1F
	.dc.b	$B6,$6E,$3F,$EC,$3A,$FC,$4E,$1F,$C8,$7D,$3F,$E2,$44,$7F,$88,$7C
	.dc.b	$65,$7E,$8F,$A2,$F8,$9F,$C6,$C6,$1B,$97,$F1,$0B,$E0,$FF,$8E,$6B
	.dc.b	$87,$77,$F1,$7A,$DE,$5F,$CA,$2F,$FA,$5F,$85,$FE,$2B,$8F,$F6,$C3
	.dc.b	$71,$FF,$6B,$B7,$8B,$FD,$0D,$E8,$70,$B9,$5F,$E2,$0C,$FF,$8A,$DE
	.dc.b	$CF,$E3,$F1,$9C,$3F,$DC,$9F,$5D,$FD,$4E,$3F,$02,$FF,$90,$DD,$1F
	.dc.b	$EE,$67,$0E,$BD,$BF,$EC,$31,$C6,$FF,$91,$BA,$6F,$C0,$6A,$7F,$B4
	.dc.b	$FC,$BB,$E9,$5D,$38,$6E,$3E,$09,$D4,$E7,$9B,$4E,$F4,$95,$BB,$E2
	.dc.b	$FE,$9A,$F8,$B9,$9D,$B9,$9D,$E5,$FC,$EB,$E7,$DF,$41,$FD,$F1,$B5
	.dc.b	$F0,$B8,$51,$88,$FD,$3A,$F9,$8E,$15,$27,$84,$EE,$FB,$57,$F8,$23
	.dc.b	$E8,$FF,$95,$57,$EC,$1F,$E5,$0D,$F4,$35,$E7,$C8,$F1,$4E,$AF,$FA
	.dc.b	$8B,$78,$FF,$CB,$F9,$A3,$AC,$FF,$62,$EF,$FE,$DC,$4B,$FB,$BD,$2F
	.dc.b	$AC,$7D,$76,$67,$FC,$09,$F4,$7F,$DD,$4C,$61,$CE,$3F,$0A,$7F,$80
	.dc.b	$A3,$4E,$3D,$EF,$DD,$E9,$E4,$FA,$EF,$F0,$26,$A5,$E0,$7F,$8A,$9B
	.dc.b	$7F,$3D,$D9,$23,$46,$FE,$07,$F1,$1F,$17,$2B,$F0,$4F,$E2,$FE,$83
	.dc.b	$11,$F7,$E2,$E9,$FE,$AB,$E9,$FC,$95,$7F,$8B,$BE,$BB,$FB,$2D,$FB
	.dc.b	$1A,$FF,$8D,$1B,$5F,$BB,$E0,$FE,$17,$3F,$E8,$E9,$CE,$BF,$ED,$DF
	.dc.b	$09,$F5,$1F,$BF,$3B,$7B,$B2,$FE,$8C,$E9,$B3,$97,$F9,$9F,$67,$F8
	.dc.b	$13,$87,$F5,$1F,$57,$FE,$50,$8C,$1F,$E1,$09,$FC,$B3,$88,$D1,$BF
	.dc.b	$84,$FF,$2B,$FC,$B7,$7A,$FE,$7E,$DD,$BF,$C3,$6D,$D3,$F1,$6F,$FA
	.dc.b	$4A,$7F,$44,$E5,$62,$8C,$37,$2E,$FF,$11,$FE,$66,$F1,$3F,$B1,$DF
	.dc.b	$F4,$E7,$FA,$BE,$1F,$98,$6C,$74,$77,$8F,$F8,$B9,$B3,$BF,$4B,$A1
	.dc.b	$D7,$D2,$FF,$39,$FD,$5F,$FA,$D2,$FF,$C6,$AD,$AF,$D2,$B7,$0F,$4A
	.dc.b	$EB,$E3,$1C,$BA,$23,$FC,$06,$78,$6A,$7F,$D4,$38,$76,$EE,$F0,$9E
	.dc.b	$2F,$8F,$97,$C6,$32,$F3,$8F,$F3,$EF,$D6,$FF,$82,$7C,$FF,$F7,$DF
	.dc.b	$87,$FE,$D2,$8C,$57,$FB,$7A,$31,$EC,$7F,$9F,$FE,$2F,$F8,$4D,$3E
	.dc.b	$BB,$FD,$FB,$5A,$C3,$9E,$1F,$EE,$2A,$E7,$7C,$30,$E5,$7D,$F0,$E3
	.dc.b	$7B,$9C,$B7,$0E,$6B,$0E,$77,$C6,$33,$18,$78,$1E,$92,$E1,$E8,$C3
	.dc.b	$BC,$73,$C3,$BB,$70,$E7,$18,$BF,$74,$E2,$63,$D4,$E0,$58,$DB,$67
	.dc.b	$91,$27,$CF,$5E,$2A,$F4,$57,$2B,$F8,$D5,$A8,$EF,$18,$7A,$46,$5A
	.dc.b	$48,$D1,$EF,$78,$E1,$F5,$5D,$3F,$25,$A7,$C6,$75,$1C,$63,$27,$96
	.dc.b	$32,$71,$8E,$EE,$23,$91,$F8,$1B,$F7,$61,$F0,$63,$9E,$1F,$1A,$F8
	.dc.b	$C7,$03,$B1,$C4,$F4,$99,$8F,$8E,$71,$BF,$04,$70,$24,$65,$FC,$18
	.dc.b	$3E,$11,$4C,$98,$3F,$45,$ED,$7E,$D5,$D1,$F9,$B7,$0E,$21,$18,$8E
	.dc.b	$3C,$1C,$1E,$0F,$E5,$A3,$A5,$FF,$72,$D3,$FA,$9F,$0E,$6A,$5F,$E2
	.dc.b	$EF,$78,$F9,$EF,$73,$9A,$DE,$1F,$B4,$BD,$CF,$BE,$B7,$39,$AD,$DF
	.dc.b	$9A,$BF,$35,$5A,$8C,$56,$E7,$B1,$AF,$09,$8F,$28,$F6,$9A,$9F,$64
	.dc.b	$EA,$31,$5D,$08,$F9,$63,$53,$DC,$93,$A9,$FF,$41,$CE,$A3,$19,$7D
	.dc.b	$0C,$7C,$55,$AC,$2C,$E6,$3C,$27,$78,$CE,$5F,$1A,$3F,$86,$37,$2F
	.dc.b	$85,$1E,$CC,$BC,$6D,$C7,$E9,$65,$E3,$8C,$10,$E7,$97,$F1,$8B,$E7
	.dc.b	$18,$3C,$65,$BF,$8F,$8F,$92,$BD,$75,$C3,$2F,$55,$78,$AF,$86,$3E
	.dc.b	$2C,$38,$47,$4B,$E1,$97,$B7,$0E,$37,$F2,$E7,$38,$FF,$36,$E1,$FE
	.dc.b	$99,$F4,$BF,$5F,$87,$FA,$CF,$C5,$DD,$3F,$AC,$CD,$D3,$FA,$CC,$DB
	.dc.b	$E3,$5F,$D5,$E3,$1C,$6B,$B1,$8F,$4F,$3C,$16,$FF,$AF,$37,$1F,$F6
	.dc.b	$1E,$0A,$E7,$EA,$67,$3F,$ED,$BF,$11,$DB,$E0,$FB,$B7,$2F,$ED,$0D
	.dc.b	$C7,$9B,$6B,$FC,$1F,$F5,$DE,$9E,$F3,$7E,$13,$0C,$C3,$FC,$22,$E5
	.dc.b	$A7,$F8,$53,$CA,$E4,$73,$F2,$6C,$64,$99,$7F,$88,$23,$2F,$FB,$B0
	.dc.b	$7D,$43,$DC,$F9,$DE,$83,$E2,$1E,$47,$D1,$BD,$D6,$A7,$38,$FF,$99
	.dc.b	$30,$E4,$FF,$9B,$A3,$35,$DE,$72,$FF,$B3,$23,$3B,$7E,$DE,$3A,$D7
	.dc.b	$3B,$F1,$F8,$03,$C5,$87,$2B,$FF,$2E,$72,$3B,$FF,$0D,$E7,$7F,$85
	.dc.b	$DB,$0F,$AE,$B1,$32,$FF,$58,$4F,$16,$CE,$73,$FF,$29,$D7,$E3,$FC
	.dc.b	$BE,$55,$FE,$50,$8C,$56,$5B,$19,$3F,$CA,$B1,$E8,$F2,$3D,$3E,$55
	.dc.b	$9F,$90,$DC,$3E,$33,$73,$F8,$9A,$7F,$9B,$2B,$E7,$5E,$5F,$E9,$9C
	.dc.b	$2D,$7E,$5C,$7F,$4D,$A5,$7E,$9F,$D3,$7B,$B7,$E6,$70,$DB,$FF,$57
	.dc.b	$B7,$4F,$F5,$B3,$73,$FE,$59,$E0,$BE,$B5,$F5,$DF,$B3,$6E,$1F,$EC
	.dc.b	$97,$4D,$8D,$11,$D4,$5C,$3F,$DB,$5D,$9F,$ED,$49,$FF,$72,$2F,$37
	.dc.b	$EE,$B9,$9D,$C6,$3D,$BC,$B6,$B8,$B6,$7A,$CF,$2F,$6B,$19,$FF,$B8
	.dc.b	$E7,$F8,$67,$B7,$FE,$65,$C3,$FD,$B7,$18,$F2,$32,$D8,$C9,$FE,$E0
	.dc.b	$8F,$E2,$9F,$5C,$FA,$EA,$63,$CE,$B5,$FE,$ED,$9D,$36,$74,$D3,$F5
	.dc.b	$9E,$0F,$ED,$3E,$B9,$F1,$9F,$A2,$8C,$36,$7F,$DB,$B3,$A7,$89,$88
	.dc.b	$D6,$5F,$B3,$AF,$DF,$F4,$BE,$5F,$3B,$7D,$EF,$AE,$7A,$BF,$CB,$FE
	.dc.b	$1D,$F3,$FF,$CC,$67,$3A,$7F,$87,$6F,$F0,$B5,$A8,$C1,$FE,$F6,$9F
	.dc.b	$E6,$8E,$33,$FF,$55,$C3,$D1,$E9,$E9,$75,$7F,$9F,$7C,$4B,$DA,$BF
	.dc.b	$21,$86,$E3,$CF,$B9,$F1,$6E,$AF,$E8,$CF,$E2,$67,$4E,$5B,$D9,$FD
	.dc.b	$2F,$EB,$BF,$81,$6C,$E0,$FC,$FC,$7F,$00,$8C,$57,$F8,$EF,$14,$F0
	.dc.b	$AD,$F9,$1F,$D5,$7E,$B5,$F5,$D7,$DE,$C3,$71,$F4,$7D,$6B,$EB,$BF
	.dc.b	$AE,$2F,$8A,$B1,$87,$FB,$07,$D7,$7F,$23,$AF,$80,$BD,$7C,$FA,$9A
	.dc.b	$7F,$B3,$7D,$77,$FA,$9E,$BE,$32,$FA,$7C,$E9,$A7,$FB,$67,$D6,$BE
	.dc.b	$BA,$F9,$78,$6E,$1F,$3F,$E2,$9F,$11,$FD,$CD,$FF,$22,$F2,$7F,$06
	.dc.b	$73,$3B,$F8,$4F,$EF,$1F,$5D,$8E,$9F,$96,$7F,$BB,$1E,$33,$F3,$5D
	.dc.b	$5F,$F1,$07,$33,$FE,$F2,$F9,$EF,$ED,$35,$F8,$15,$F0,$5F,$E0,$57
	.dc.b	$33,$DB,$AD,$1A,$AF,$C4,$7D,$73,$E6,$7F,$C1,$9F,$67,$FC,$07,$0E
	.dc.b	$13,$9A,$FC,$74,$E7,$3F,$E4,$72,$F5,$7F,$C2,$8E,$EF,$8C,$63,$0F
	.dc.b	$F0,$C7,$D1,$BE,$33,$FC,$35,$F4,$9C,$73,$FC,$D4,$E5,$FC,$E4,$67
	.dc.b	$0F,$57,$D7,$BE,$67,$9E,$1F,$E2,$38,$C5,$F2,$8C,$4C,$8F,$D3,$4B
	.dc.b	$E6,$F1,$7F,$E0,$3F,$AE,$FD,$C3,$F5,$3E,$11,$EC,$3E,$BF,$FC,$05
	.dc.b	$5A,$E1,$FF,$37,$56,$A7,$3E,$A7,$F1,$76,$E7,$FC,$51,$B8,$F7,$9F
	.dc.b	$F0,$7D,$FF,$3B,$3F,$C0,$58,$75,$BF,$F5,$33,$F9,$77,$F8,$EF,$47
	.dc.b	$8B,$BB,$F9,$23,$63,$27,$F0,$DC,$15,$C5,$F6,$67,$F8,$AE,$77,$D5
	.dc.b	$7F,$64,$BF,$E1,$5F,$63,$FC,$5A,$FF,$11,$39,$7F,$15,$19,$CF,$FB
	.dc.b	$1C,$E5,$BF,$07,$FE,$55,$87,$C2,$58,$FF,$2C,$7D,$67,$B7,$06,$4F
	.dc.b	$AD,$7D,$0E,$B8,$7F,$CB,$D1,$8D,$7F,$8F,$61,$C6,$CD,$9F,$E0,$98
	.dc.b	$7E,$7E,$3B,$3F,$EC,$FF,$99,$FF,$69,$7D,$87,$F9,$76,$3F,$9C,$7D
	.dc.b	$1F,$FA,$9E,$31,$1B,$0E,$37,$4E,$FE,$67,$FD,$9D,$7D,$1F,$F6,$E4
	.dc.b	$EB,$3C,$61,$FE,$DA,$AA,$FF,$2D,$BE,$4E,$4E,$F7,$CA,$33,$1E,$3C
	.dc.b	$3C,$51,$88,$FF,$76,$57,$4C,$3C,$11,$C8,$E9,$5C,$30,$F0,$C7,$03
	.dc.b	$47,$8D,$CA,$AE,$97,$DE,$30,$7F,$BB,$4E,$45,$8C,$38,$3F,$DF,$DF
	.dc.b	$60,$FA,$97,$4F,$18,$C3,$63,$33,$FE,$F9,$9C,$B7,$E7,$BF,$AA,$37
	.dc.b	$3E,$51,$88,$1E,$07,$0D,$8F,$A3,$A7,$F7,$C8,$BF,$07,$C1,$F8,$76
	.dc.b	$BE,$FA,$FE,$12,$31,$87,$0A,$87,$0A,$DE,$1D,$E3,$E6,$BD,$DB,$87
	.dc.b	$7C,$3A,$57,$09,$F4,$E1,$D9,$D9,$DA,$77,$87,$E2,$A7,$51,$F9,$8C
	.dc.b	$3B,$37,$0E,$D1,$E3,$8F,$D0,$70,$7E,$A3,$EF,$7F,$2F,$BF,$F7,$BE
	.dc.b	$04,$63,$3F,$DA,$C6,$1F,$D8,$FD,$56,$6F,$94,$7E,$D2,$B1,$5B,$8F
	.dc.b	$D9,$56,$F9,$FF,$BB,$47,$F8,$5D,$AF,$F0,$EF,$4B,$5F,$11,$B6,$46
	.dc.b	$DC,$BF,$E1,$FF,$B3,$FF,$52,$37,$D5,$F2,$46,$23,$FB,$03,$DB,$87
	.dc.b	$FC,$F2,$E5,$B3,$C0,$E3,$5E,$DF,$B2,$FF,$60,$DF,$F8,$B5,$AF,$F6
	.dc.b	$B6,$D7,$FA,$11,$7D,$1F,$1F,$9D,$95,$C5,$E9,$18,$8F,$EE,$9F,$A9
	.dc.b	$BD,$8F,$98,$D8,$E0,$7F,$8F,$7E,$CB,$FC,$4C,$DF,$77,$E9,$46,$23
	.dc.b	$FB,$F3,$FE,$29,$78,$7D,$BE,$88,$F1,$3F,$B6,$F3,$FD,$BF,$3B,$F8
	.dc.b	$F7,$3F,$F8,$F5,$B3,$FC,$DA,$D7,$F8,$1F,$B3,$FF,$7C,$CF,$F9,$45
	.dc.b	$AF,$F8,$3D,$BB,$FF,$AA,$1F,$F0,$AC,$7F,$95,$FE,$8D,$F5,$1F,$8B
	.dc.b	$6C,$66,$3B,$78,$DF,$8F,$F5,$FF,$D7,$73,$FE,$1F,$F9,$CC,$19,$F1
	.dc.b	$7F,$29,$F4,$7F,$EF,$B9,$FF,$32,$2F,$B2,$C9,$AF,$A4,$FC,$D7,$D9
	.dc.b	$3E,$CE,$92,$FF,$CD,$91,$E4,$7F,$C5,$5F,$51,$FD,$71,$FF,$38,$35
	.dc.b	$CF,$9F,$FA,$5F,$85,$FE,$19,$1F,$E3,$1F,$86,$FF,$7F,$71,$FF,$B2
	.dc.b	$97,$DF,$7E,$AD,$B3,$FE,$56,$E8,$FD,$77,$8B,$FE,$F9,$8F,$F3,$F3
	.dc.b	$5F,$F1,$E3,$6F,$FD,$AA,$BC,$0F,$DB,$7D,$A3,$E5,$2F,$CE,$4C,$3F
	.dc.b	$D1,$13,$97,$FD,$15,$F6,$9F,$EA,$E6,$CF,$EF,$E3,$11,$FE,$4B,$F9
	.dc.b	$0F,$5E,$7F,$E8,$3F,$5F,$FC,$4D,$3F,$E9,$36,$BF,$E5,$16,$E9,$FE
	.dc.b	$EA,$7F,$CD,$D1,$FE,$94,$F9,$8F,$E1,$9F,$2B,$FE,$A3,$8C,$1F,$E9
	.dc.b	$8F,$0F,$F8,$87,$CC,$66,$3F,$CB,$8D,$C3,$FD,$EA,$DC,$7F,$D2,$CF
	.dc.b	$E6,$BD,$17,$BD,$F8,$5B,$FF,$32,$36,$FF,$DF,$AD,$BF,$F4,$F3,$F8
	.dc.b	$99,$FF,$57,$7C,$DF,$F1,$DA,$FF,$5F,$FA,$7F,$C8,$E3,$17,$FE,$A5
	.dc.b	$5C,$9F,$CA,$ED,$F8,$FB,$EB,$70,$D8,$FF,$1C,$35,$FF,$42,$E2,$F9
	.dc.b	$27,$A9,$C2,$E5,$70,$7F,$9C,$56,$E7,$33,$FE,$8E,$3C,$33,$89,$65
	.dc.b	$FE,$D9,$BE,$B5,$FE,$CB,$F8,$77,$93,$FC,$FD,$77,$D3,$11,$27,$84
	.dc.b	$65,$E5,$1F,$D1,$FE,$D5,$F5,$54,$93,$FE,$D9,$8F,$EE,$0D,$8C,$BF
	.dc.b	$E9,$98,$E2,$EF,$DD,$FE,$AD,$F4,$2F,$C5,$7F,$D3,$53,$FE,$ED,$8C
	.dc.b	$7C,$76,$A7,$33,$FE,$B0,$9E,$1F,$15,$9A,$93,$C6,$B7,$BF,$C6,$85
	.dc.b	$F0,$75,$5E,$2F,$D0,$FF,$99,$56,$E9,$18,$E0,$FF,$0A,$C7,$92,$FF
	.dc.b	$DF,$1F,$50,$F9,$FF,$ED,$8F,$FB,$FE,$33,$F4,$38,$99,$18,$21,$FE
	.dc.b	$BD,$9E,$4F,$F8,$8B,$1F,$47,$5B,$EA,$FF,$94,$CF,$FB,$0D,$B3,$A6
	.dc.b	$E5,$FE,$FD,$7F,$CB,$F3,$CE,$31,$E9,$7F,$80,$67,$73,$FE,$7A,$6D
	.dc.b	$F7,$F4,$BF,$41,$F6,$6F,$56,$92,$FF,$D2,$DF,$6D,$FC,$9B,$3F,$9D
	.dc.b	$0A,$E0,$FF,$A6,$DA,$FE,$6D,$B1,$C2,$79,$67,$FE,$3B,$9C,$B6,$FC
	.dc.b	$17,$FA,$3A,$DB,$FE,$92,$FB,$6F,$F6,$D3,$EF,$31,$F2,$FF,$E1,$F8
	.dc.b	$CC,$7F,$BA,$DF,$C2,$DF,$F9,$43,$34,$6B,$2F,$A2,$D9,$D3,$6B,$C1
	.dc.b	$C7,$F9,$5F,$1F,$F2,$DE,$34,$9F,$C9,$46,$5F,$CA,$46,$5F,$CD,$73
	.dc.b	$7F,$1F,$7F,$2B,$F8,$B7,$C0,$6E,$DF,$A6,$8C,$73,$7E,$9E,$71,$31
	.dc.b	$FD,$44,$EA,$BE,$2D,$23,$89,$2B,$F5,$95,$FE,$43,$FB,$2F,$CF,$3F
	.dc.b	$EB,$C7,$FC,$24,$B5,$FC,$9C,$61,$BF,$01,$C4,$E5,$F4,$52,$BE,$97
	.dc.b	$02,$FA,$AE,$1F,$F4,$64,$E9,$FF,$64,$B7,$CC,$FF,$2B,$46,$A3,$C5
	.dc.b	$C5,$F5,$E3,$19,$7E,$99,$E3,$3A,$3F,$DA,$AE,$23,$5F,$1D,$FE,$61
	.dc.b	$FB,$3F,$C4,$E2,$69,$3E,$BF,$58,$43,$D7,$19,$7D,$BD,$5F,$DC,$5F
	.dc.b	$05,$23,$18,$7E,$E9,$B9,$7E,$3D,$FE,$E7,$19,$F5,$3F,$CE,$91,$AB
	.dc.b	$FF,$76,$B6,$FA,$7A,$7E,$9D,$25,$79,$1F,$67,$3E,$23,$C7,$9F,$E5
	.dc.b	$EF,$DD,$5A,$9F,$D3,$0F,$B6,$7D,$A7,$F9,$D3,$E5,$1F,$3B,$FE,$AD
	.dc.b	$9C,$FC,$77,$FB,$62,$31,$5B,$75,$F1,$7F,$EF,$77,$33,$BC,$3F,$D7
	.dc.b	$B7,$C9,$FF,$6B,$4E,$9A,$6A,$B9,$3C,$2F,$B3,$6F,$A3,$E9,$8C,$BE
	.dc.b	$B9,$E2,$E6,$77,$7E,$1C,$3D,$B3,$CC,$F0,$9C,$4D,$DF,$49,$E1,$87
	.dc.b	$9E,$B8,$D6,$CE,$33,$A7,$A4,$6A,$31,$1B,$8C,$57,$52,$5F,$9B,$0F
	.dc.b	$4D,$F0,$24,$F8,$E3,$2F,$0B,$F0,$B9,$7D,$35,$1C,$BD,$5C,$5F,$8A
	.dc.b	$2E,$1F,$6A,$69,$EA,$7E,$01,$97,$99,$27,$93,$F0,$AF,$CB,$5C,$A3
	.dc.b	$E9,$19,$31,$87,$B2,$3C,$91,$CE,$FB,$BE,$13,$F0,$87,$0C,$3E,$06
	.dc.b	$1F,$37,$0E,$F3,$C4,$E0,$7C,$FA,$FC,$73,$FE,$7C,$C3,$F2,$4D,$BF
	.dc.b	$F6,$83,$A3,$F4,$1F,$62,$E3,$FF,$41,$C7,$E6,$A3,$0E,$29,$18,$8E
	.dc.b	$CD,$79,$CF,$E4,$A3,$8E,$3F,$8A,$FB,$6F,$F2,$EF,$26,$6A,$5F,$E3
	.dc.b	$2F,$79,$22,$5F,$07,$55,$C3,$0F,$C8,$5F,$09,$DF,$AF,$FE,$23,$9C
	.dc.b	$D6,$F0,$F2,$D7,$E8,$B0,$E9,$5A,$8E,$A4,$7D,$17,$CE,$B8,$C6,$C8
	.dc.b	$F9,$E5,$5D,$23,$64,$8F,$99,$7F,$BC,$9D,$46,$0F,$04,$ED,$E7,$7E
	.dc.b	$17,$57,$E0,$6C,$F8,$6B,$9C,$7F,$1F,$C3,$E2,$37,$0F,$84,$FC,$6C
	.dc.b	$6B,$93,$C2,$64,$A6,$70,$F0,$3E,$48,$E3,$97,$E1,$2B,$9E,$1E,$28
	.dc.b	$C5,$72,$6D,$FC,$CB,$E3,$87,$9A,$73,$8F,$AF,$2F,$81,$87,$86,$3E
	.dc.b	$24,$EA,$3C,$D8,$79,$23,$11,$B3,$33,$C2,$FB,$1F,$13,$1F,$33,$F8
	.dc.b	$9D,$FF,$B3,$79,$5F,$E6,$BE,$6F,$F2,$AF,$12,$31,$AF,$F8,$AE,$1F
	.dc.b	$4F,$8C,$D8,$72,$AF,$ED,$77,$98,$C4,$FF,$58,$AC,$5F,$5B,$FE,$E3
	.dc.b	$E9,$60,$CF,$D6,$D8,$C9,$2F,$FB,$0C,$65,$FE,$C7,$E1,$F4,$3A,$FA
	.dc.b	$96,$BF,$BE,$C5,$9F,$ED,$33,$74,$FF,$07,$7C,$2F,$9F,$A1,$F1,$56
	.dc.b	$30,$DC,$7F,$C3,$8D,$C3,$FC,$0B,$5C,$27,$FC,$4B,$E0,$3E,$6F,$A7
	.dc.b	$5F,$C3,$8C,$4F,$F8,$96,$31,$3A,$8E,$2D,$BF,$F5,$27,$C9,$3C,$2F
	.dc.b	$A1,$E3,$A3,$17,$FE,$AD,$5E,$C6,$9C,$DF,$2B,$C1,$99,$B5,$C6,$2C
	.dc.b	$F8,$26,$ED,$FE,$A9,$B7,$DC,$75,$F0,$FB,$37,$1F,$13,$70,$F4,$B6
	.dc.b	$FF,$C9,$FF,$65,$F0,$CF,$89,$D5,$6F,$E4,$C5,$BF,$F1,$8E,$FF,$CC
	.dc.b	$A7,$27,$F8,$7B,$ED,$BD,$B8,$7F,$A5,$9B,$1F,$C5,$57,$AF,$F9,$7B
	.dc.b	$46,$9F,$E6,$0D,$73,$B6,$E3,$8D,$77,$8C,$11,$F1,$55,$AF,$F5,$D7
	.dc.b	$DB,$FE,$87,$D4,$37,$EA,$70,$CB,$FF,$35,$CE,$9B,$39,$AF,$F2,$8F
	.dc.b	$DB,$3E,$0F,$F5,$8B,$E0,$AC,$61,$B7,$FE,$76,$9D,$36,$74,$FF,$BC
	.dc.b	$FF,$44,$7D,$8F,$E3,$3D,$E7,$D3,$FE,$3A,$73,$5F,$EF,$99,$CB,$C6
	.dc.b	$B4,$6F,$CE,$B1,$92,$67,$FE,$FD,$B4,$EA,$3F,$CD,$BF,$01,$FE,$D3
	.dc.b	$AF,$F4,$6B,$63,$3F,$21,$FC,$2E,$31,$C9,$FE,$EA,$6D,$7F,$AF,$DB
	.dc.b	$5F,$C5,$15,$71,$F5,$FF,$8A,$7D,$BA,$EF,$C8,$61,$A7,$FA,$7A,$34
	.dc.b	$D8,$D3,$4F,$F3,$7F,$E8,$BF,$E6,$C6,$F9,$FF,$C9,$73,$79,$4F,$9E
	.dc.b	$FC,$95,$EF,$5F,$F4,$E0,$71,$5F,$EF,$06,$E1,$FE,$05,$8E,$75,$CB
	.dc.b	$09,$3E,$03,$FD,$CD,$3A,$71,$7F,$E9,$AF,$8B,$FA,$31,$FE,$BA,$6C
	.dc.b	$67,$8F,$F9,$C4,$62,$BE,$34,$FF,$3E,$FB,$77,$D2,$7F,$3D,$6E,$1F
	.dc.b	$EC,$47,$FC,$06,$D8,$FE,$85,$F1,$CF,$3E,$98,$F8,$8B,$5F,$EC,$CF
	.dc.b	$8F,$FE,$9A,$DF,$4F,$FA,$54,$7C,$A8,$C1,$FE,$D2,$FB,$67,$8D,$FD
	.dc.b	$3D,$B1,$9F,$1F,$13,$FD,$4F,$E4,$5F,$05,$4C,$7C,$ED,$BF,$F6,$EF
	.dc.b	$DA,$BE,$8B,$FA,$D3,$FC,$BE,$33,$18,$F3,$FF,$AF,$78,$DF,$A4,$E4
	.dc.b	$FF,$63,$B7,$97,$E4,$FB,$5F,$B7,$F4,$F7,$D3,$FD,$97,$E6,$1A,$DF
	.dc.b	$1B,$0D,$D7,$FD,$DF,$F3,$2F,$84,$FE,$D8,$D8,$CF,$BD,$89,$FE,$E1
	.dc.b	$F6,$6F,$94,$A6,$3E,$56,$DF,$FB,$EB,$EC,$F4,$D8,$FE,$F1,$18,$E2
	.dc.b	$FF,$70,$A8,$FE,$F9,$F4,$0F,$80,$A6,$3E,$02,$E1,$F9,$73,$FC,$B9
	.dc.b	$C6,$FB,$3F,$F2,$F4,$67,$2F,$F0,$4C,$61,$FF,$77,$FD,$9B,$ED,$3F
	.dc.b	$C1,$53,$9E,$7F,$97,$3E,$3B,$E5,$82,$31,$87,$2F,$8A,$FF,$09,$46
	.dc.b	$2F,$E9,$FD,$37,$F9,$B5,$B3,$FE,$17,$FB,$37,$C8,$FF,$86,$6B,$59
	.dc.b	$7A,$A7,$2F,$81,$B8,$B0,$FB,$F1,$FD,$22,$31,$3C,$6F,$B7,$BD,$23
	.dc.b	$97,$BF,$6B,$67,$F8,$B3,$EF,$7A,$E3,$F8,$AF,$89,$7C,$47,$F8,$B6
	.dc.b	$73,$87,$CE,$9C,$BF,$12,$33,$7C,$F2,$E1,$8F,$16,$56,$DC,$B4,$F9
	.dc.b	$BF,$55,$FE,$1D,$AF,$2F,$02,$31,$9F,$E7,$9F,$EB,$53,$A7,$35,$F8
	.dc.b	$26,$C7,$F9,$21,$B5,$CF,$E0,$FF,$C9,$11,$8F,$AB,$E7,$87,$EA,$9B
	.dc.b	$87,$E3,$17,$C1,$FF,$93,$DC,$E3,$FE,$54,$6E,$26,$9C,$2B,$F1,$2E
	.dc.b	$BE,$B3,$F9,$04,$7E,$61,$B5,$FE,$5E,$7F,$79,$19,$F8,$0F,$F2,$F4
	.dc.b	$6A,$CD,$37,$5C,$67,$FC,$24,$FF,$AF,$FE,$47,$FD,$81,$F5,$AF,$8F
	.dc.b	$FF,$56,$68,$F9,$EF,$E6,$33,$9B,$FF,$58,$0B,$E3,$7C,$2B,$6F,$FB
	.dc.b	$72,$33,$3F,$D2,$67,$2F,$28,$CB,$F9,$98,$CB,$FE,$BD,$8C,$DF,$8F
	.dc.b	$3D,$CF,$A4,$61,$E5,$79,$46,$9F,$31,$AC,$3C,$D1,$9B,$EB,$87,$82
	.dc.b	$BA,$5F,$3C,$3C,$11,$C4,$F4,$14,$EE,$53,$C8,$7F,$BF,$0F,$09,$6B
	.dc.b	$0E,$8F,$F6,$D7,$D7,$7F,$00,$FA,$FE,$0D,$8F,$9E,$1C,$3F,$49,$AF
	.dc.b	$2E,$7F,$E0,$4F,$08,$FF,$28,$DF,$49,$CC,$F1,$6B,$C6,$3E,$9E,$7F
	.dc.b	$DB,$A2,$FB,$7F,$EE,$FC,$BE,$76,$1D,$27,$E1,$36,$3E,$95,$71,$9F
	.dc.b	$8C,$D8,$E3,$87,$99,$B8,$2B,$FC,$2D,$5F,$EF,$D3,$97,$BF,$D3,$1F
	.dc.b	$F8,$00,$F0,$9D,$9C,$71,$F1,$57,$1A,$F4,$36,$7F,$1F,$3A,$78,$63
	.dc.b	$E3,$9D,$47,$A2,$B8,$BA,$C3,$F3,$7C,$1F,$AE,$F8,$9F,$D3,$EF,$FB
	.dc.b	$76,$D7,$ED,$DB,$3F,$B4,$F4,$2F,$17,$F9,$AE,$BF,$A4,$B5,$FE,$3A
	.dc.b	$DC,$3F,$BC,$BF,$E6,$F8,$FE,$0F,$AF,$4E,$3F,$F1,$E4,$60,$FE,$17
	.dc.b	$F4,$99,$9F,$C9,$63,$FC,$45,$B5,$C2,$BF,$77,$39,$7F,$1E,$FF,$2D
	.dc.b	$EC,$E3,$CF,$FD,$02,$72,$D7,$8B,$5F,$B4,$7F,$2D,$8F,$83,$FA,$0F
	.dc.b	$7C,$FF,$C2,$6D,$DB,$F9,$A4,$FC,$6F,$A9,$FF,$66,$5F,$F6,$C6,$BF
	.dc.b	$D0,$1B,$7F,$E1,$C7,$FC,$4D,$1B,$F5,$3D,$F1,$C2,$3C,$79,$23,$11
	.dc.b	$FE,$04,$6A,$8D,$4F,$D8,$FB,$0F,$F4,$3C,$FF,$7E,$6B,$E3,$8F,$BB
	.dc.b	$F6,$0F,$9F,$BE,$9C,$AF,$F0,$2C,$E5,$B1,$F5,$23,$FC,$63,$1F,$41
	.dc.b	$AF,$D4,$6B,$EA,$F1,$7F,$8A,$6B,$FE,$1C,$6C,$7A,$5B,$92,$B9,$E9
	.dc.b	$FE,$42,$FB,$0F,$F6,$1D,$FF,$86,$17,$C6,$7E,$59,$B1,$9F,$0B,$F3
	.dc.b	$3E,$8B,$CD,$FC,$5B,$67,$F0,$ED,$7F,$16,$D7,$E8,$F9,$EF,$96,$FD
	.dc.b	$1B,$63,$FC,$52,$D7,$FB,$EB,$FD,$45,$FF,$16,$FD,$83,$EB,$FF,$56
	.dc.b	$BE,$C2,$98,$8F,$D7,$FD,$87,$F9,$16,$BF,$36,$D7,$F3,$ED,$7F,$0A
	.dc.b	$D7,$FC,$8F,$EA,$FF,$AE,$1F,$F1,$EB,$5F,$F0,$8B,$77,$FF,$36,$FB
	.dc.b	$5F,$D9,$1F,$F3,$7F,$B9,$FE,$98,$AF,$F0,$BA,$F9,$0F,$E0,$6D,$8F
	.dc.b	$F2,$4B,$B9,$FF,$3B,$F8,$6F,$B8,$FD,$CB,$6B,$F6,$AD,$7F,$72,$D7
	.dc.b	$F3,$FF,$62,$FA,$5F,$E2,$2D,$8F,$F2,$BB,$5F,$F1,$5A,$FA,$6F,$E2
	.dc.b	$FF,$63,$99,$FD,$4B,$5F,$F3,$5B,$5F,$D4,$37,$A9,$EA,$BE,$07,$F8
	.dc.b	$8B,$6B,$F8,$6B,$5F,$E2,$2D,$7F,$6F,$F1,$8F,$AA,$FE,$32,$D8,$FE
	.dc.b	$2A,$D7,$F8,$CB,$5F,$DF,$FC,$8D,$7F,$D0,$2D,$7F,$7C,$DC,$0D,$F7
	.dc.b	$7F,$1D,$3F,$C8,$9A,$FF,$26,$6B,$FC,$31,$AF,$FA,$3B,$E4,$FF,$9D
	.dc.b	$3F,$C4,$1A,$E7,$D0,$FE,$67,$18,$8F,$EC,$1F,$62,$F7,$BF,$A4,$C2
	.dc.b	$30,$FF,$A5,$CC,$47,$F4,$EE,$4F,$F7,$EF,$A3,$C2,$B7,$5F,$E0,$86
	.dc.b	$BF,$E3,$86,$BF,$CD,$9A,$F8,$AB,$FA,$CF,$D1,$BE,$AD,$F2,$78,$6D
	.dc.b	$78,$EF,$BE,$3F,$EC,$EF,$A4,$FF,$60,$4F,$FA,$91,$FF,$5C,$37,$B7
	.dc.b	$FB,$1C,$6A,$3F,$CF,$6D,$9E,$53,$A3,$7E,$27,$F6,$AF,$9D,$7D,$8D
	.dc.b	$26,$1F,$EB,$28,$7B,$78,$8F,$F4,$3C,$70,$75,$3C,$E7,$D9,$F1,$92
	.dc.b	$7F,$D8,$9B,$7C,$18,$F8,$1D,$39,$9F,$F3,$0C,$E5,$FF,$1A,$B5,$CE
	.dc.b	$DF,$EE,$77,$2D,$7F,$D5,$B0,$FA,$0F,$E8,$91,$86,$EA,$F4,$3F,$C4
	.dc.b	$95,$D6,$3F,$DF,$7F,$60,$F9,$9F,$F0,$6A,$FB,$1A,$62,$26,$3F,$ED
	.dc.b	$69,$D3,$67,$2F,$FA,$B3,$E9,$3F,$99,$5F,$86,$3F,$B2,$46,$31,$FF
	.dc.b	$75,$B6,$33,$F0,$DF,$E1,$E8,$C7,$69,$A4,$8F,$CE,$FD,$8F,$F4,$EA
	.dc.b	$FF,$67,$87,$18,$7F,$29,$6E,$1F,$E0,$CF,$B1,$FE,$5C,$FF,$83,$A3
	.dc.b	$95,$7F,$A5,$1A,$E7,$D5,$FF,$37,$C6,$23,$FC,$9C,$F4,$5F,$0E,$F9
	.dc.b	$3F,$DF,$E3,$F0,$63,$1E,$87,$A3,$0F,$F1,$DF,$D8,$BE,$AE,$93,$1F
	.dc.b	$F2,$D6,$1F,$27,$75,$7E,$06,$30,$F1,$9D,$E9,$FE,$61,$AD,$DF,$39
	.dc.b	$FF,$6F,$D7,$1A,$FC,$1F,$2F,$F2,$1C,$3E,$B7,$27,$ED,$FE,$23,$F6
	.dc.b	$C7,$ED,$FC,$89,$8F,$E7,$2F,$FC,$0F,$5B,$8C,$4F,$F0,$0F,$B1,$F8
	.dc.b	$92,$69,$8A,$FA,$34,$9D,$6B,$FE,$9F,$6C,$7E,$92,$BE,$84,$62,$3F
	.dc.b	$53,$1D,$3E,$9F,$C0,$78,$6B,$FC,$CF,$F5,$7F,$A2,$D3,$E2,$2F,$9E
	.dc.b	$C9,$AF,$82,$98,$FC,$A3,$FC,$75,$3B,$71,$C2,$8C,$43,$7F,$AD,$1D
	.dc.b	$E3,$FC,$85,$C2,$F9,$2F,$F3,$CB,$FD,$72,$39,$B5,$E5,$97,$D9,$8E
	.dc.b	$11,$A9,$E3,$E7,$C8,$C1,$E4,$FA,$DF,$F6,$1E,$46,$92,$3F,$CF,$2F
	.dc.b	$D0,$8D,$7C,$C7,$FA,$32,$0D,$9F,$F4,$7F,$BF,$FD,$ED,$F8,$2D,$7C
	.dc.b	$E7,$F9,$7D,$C6,$9F,$97,$7F,$C0,$8D,$F5,$3F,$D1,$31,$C6,$7F,$CD
	.dc.b	$6D,$7F,$30,$D7,$FC,$D5,$7F,$9D,$7F,$D1,$AD,$7F,$CE,$6D,$7F,$40
	.dc.b	$D7,$8E,$7E,$DF,$06,$47,$C7,$8F,$EC,$BF,$5E,$FB,$1F,$F4,$A7,$CA
	.dc.b	$3E,$A3,$FD,$23,$39,$F8,$DF,$F6,$64,$E7,$0F,$F6,$74,$E5,$46,$54
	.dc.b	$7F,$9F,$A3,$0E,$DB,$7E,$0B,$EE,$DA,$EB,$F2,$DF,$EA,$2A,$DE,$1E
	.dc.b	$2B,$DB,$6B,$C7,$7C,$5F,$0D,$F1,$AD,$C7,$99,$B7,$D2,$30,$78,$0E
	.dc.b	$87,$AA,$BA,$61,$E9,$AE,$27,$C6,$BE,$D5,$E2,$C3,$8D,$6E,$30,$76
	.dc.b	$3C,$A6,$AF,$A9,$27,$8D,$74,$9F,$14,$6C,$8F,$3A,$F9,$D7,$E4,$8E
	.dc.b	$F8,$7B,$4F,$A3,$7C,$EB,$C9,$3D,$0E,$58,$7B,$A7,$4F,$8E,$33,$87
	.dc.b	$91,$F0,$DF,$5C,$3C,$11,$EC,$36,$48,$D3,$F3,$4C,$47,$0C,$3C,$67
	.dc.b	$AA,$3D,$31,$9B,$F2,$47,$82,$31,$7E,$38,$E6,$70,$21,$E3,$BF,$0B
	.dc.b	$EF,$6D,$F7,$8F,$84,$7F,$BC,$0A,$7E,$20,$C3,$F9,$F6,$FD,$2B,$EC
	.dc.b	$78,$47,$E9,$23,$0E,$25,$18,$8E,$33,$C1,$FD,$3C,$70,$AF,$D3,$B7
	.dc.b	$04,$CF,$B1,$7B,$5F,$E0,$DB,$DE,$1F,$90,$C3,$85,$F8,$E3,$F3,$77
	.dc.b	$C3,$9F,$51,$27,$F1,$55,$C1,$56,$DC,$CE,$EF,$BC,$7E,$6E,$75,$18
	.dc.b	$26,$3F,$C2,$AF,$A4,$EA,$31,$7C,$DB,$7C,$F1,$F6,$46,$30,$E7,$3E
	.dc.b	$B3,$58,$F1,$71,$1A,$7D,$35,$C9,$CD,$73,$23,$F2,$E3,$4F,$D2,$BE
	.dc.b	$8D,$3B,$9F,$08,$95,$DF,$0F,$64,$76,$21,$C6,$3F,$92,$46,$A3,$17
	.dc.b	$E2,$69,$E3,$C3,$C3,$97,$E2,$30,$F1,$DF,$18,$C5,$78,$AF,$94,$E7
	.dc.b	$2F,$B5,$5A,$C7,$CD,$97,$BA,$31,$7E,$16,$E1,$E1,$3A,$1E,$A7,$2F
	.dc.b	$CC,$CB,$E8,$72,$FF,$6F,$F3,$7F,$CB,$DC,$BF,$E3,$7E,$3F,$F3,$55
	.dc.b	$7F,$75,$9C,$BF,$D5,$BC,$BB,$87,$F5,$E9,$FF,$7F,$3F,$D6,$9A,$E6
	.dc.b	$7F,$5B,$87,$76,$DB,$EB,$37,$3F,$E0,$16,$C6,$67,$71,$83,$C3,$3B
	.dc.b	$8F,$F0,$57,$D5,$62,$B8,$5F,$F7,$EA,$FF,$06,$78,$B9,$DB,$4D,$C3
	.dc.b	$FC,$24,$F8,$08,$F9,$9F,$2F,$33,$E8,$7C,$2B,$E3,$2C,$61,$B5,$FE
	.dc.b	$26,$6E,$3E,$3A,$E1,$3F,$E3,$4E,$F7,$D2,$F4,$E8,$F6,$C6,$27,$FC
	.dc.b	$5F,$18,$9D,$47,$16,$DF,$FA,$B3,$E4,$DE,$07,$D6,$F1,$D0,$FB,$6C
	.dc.b	$3A,$C3,$FD,$7E,$E5,$B5,$F6,$AB,$8B,$67,$85,$FF,$B0,$9A,$FF,$1E
	.dc.b	$3F,$CC,$8E,$B9,$BF,$CE,$AD,$BF,$F5,$DA,$F6,$FC,$77,$E4,$FB,$2F
	.dc.b	$2F,$85,$6F,$CA,$DC,$7D,$CD,$C7,$AB,$F2,$3E,$AE,$57,$CD,$EE,$B7
	.dc.b	$F8,$3E,$05,$63,$0F,$F3,$0D,$5B,$F5,$7D,$97,$9F,$0F,$F5,$2F,$C8
	.dc.b	$BE,$33,$F4,$30,$9D,$7A,$6A,$D5,$FF,$9C,$EB,$5B,$7E,$91,$B1,$FD
	.dc.b	$26,$FE,$4C,$78,$27,$51,$E1,$8D,$9F,$EC,$67,$FD,$FB,$3D,$5F,$37
	.dc.b	$57,$FB,$3F,$E4,$FF,$4B,$EC,$9B,$F4,$58,$65,$FF,$A1,$27,$4D,$9C
	.dc.b	$E1,$FA,$9F,$B2,$7C,$4F,$EC,$17,$C7,$D8,$C3,$6F,$FD,$1F,$3A,$6C
	.dc.b	$E9,$FF,$4B,$74,$3E,$C9,$F5,$5F,$97,$F8,$17,$D5,$FE,$66,$73,$7C
	.dc.b	$27,$2F,$28,$CE,$7F,$E0,$1C,$3A,$61,$F9,$99,$CC,$FE,$56,$73,$5F
	.dc.b	$E0,$06,$C7,$FA,$11,$BC,$DF,$EC,$CA,$7A,$1E,$3C,$3F,$88,$FC,$33
	.dc.b	$D8,$A6,$3D,$9B,$9F,$FA,$C2,$74,$D9,$D3,$4F,$F5,$AF,$D9,$1F,$E3
	.dc.b	$AB,$E5,$6C,$61,$AF,$06,$EB,$FC,$7E,$78,$4F,$E4,$7E,$CB,$FD,$08
	.dc.b	$FC,$7F,$9E,$FC,$CC,$62,$BE,$06,$A9,$CF,$A7,$F9,$7A,$D5,$7F,$8B
	.dc.b	$5A,$FE,$FD,$AF,$F8,$79,$E9,$BF,$AE,$31,$3F,$D2,$23,$FA,$07,$D2
	.dc.b	$76,$C7,$FD,$AB,$F2,$5F,$39,$FE,$8E,$DB,$FF,$1E,$75,$A3,$11,$FC
	.dc.b	$3E,$31,$C5,$FD,$2B,$EC,$9F,$4F,$49,$87,$F5,$28,$CB,$FD,$53,$E5
	.dc.b	$5E,$A7,$F5,$26,$C6,$63,$07,$F5,$6F,$96,$7C,$A5,$24,$FF,$5E,$8C
	.dc.b	$BF,$D8,$3E,$CB,$F8,$7B,$7C,$9F,$EB,$F1,$8D,$7D,$AD,$7D,$9F,$65
	.dc.b	$FE,$FF,$AF,$ED,$2D,$8F,$86,$D7,$7F,$0D,$FD,$AB,$EC,$BF,$9F,$D7
	.dc.b	$FC,$00,$6C,$FF,$63,$5F,$65,$FD,$A9,$CF,$27,$D7,$E2,$FD,$EF,$83
	.dc.b	$FD,$D3,$C4,$FD,$7E,$1F,$49,$AF,$D1,$6B,$FE,$5E,$97,$CF,$7E,$2F
	.dc.b	$2F,$F3,$0E,$1D,$63,$F6,$ED,$EA,$FE,$F7,$18,$9E,$5F,$3E,$F9,$5F
	.dc.b	$EF,$ED,$8E,$66,$BE,$03,$33,$FE,$04,$F9,$17,$CC,$7F,$83,$1B,$5F
	.dc.b	$E1,$26,$E9,$F8,$C6,$BF,$8B,$FB,$2C,$E9,$F0,$D7,$CD,$7F,$84,$1C
	.dc.b	$DF,$BD,$CB,$E4,$A6,$CF,$A3,$FF,$04,$4E,$63,$1F,$03,$FE,$1A,$9F
	.dc.b	$F0,$1D,$7F,$85,$DD,$BF,$E1,$9F,$B2,$FF,$44,$DF,$F8,$A9,$B1,$FA
	.dc.b	$46,$BF,$A3,$6B,$F4,$FE,$CB,$CA,$FF,$88,$D7,$C8,$7F,$8A,$9C,$E9
	.dc.b	$F8,$17,$37,$E7,$FA,$8F,$F4,$54,$FE,$15,$B1,$F7,$FE,$C9,$F2,$7F
	.dc.b	$E3,$49,$CE,$3F,$0A,$73,$5F,$E4,$26,$C7,$EF,$1A,$FE,$ED,$AF,$5F
	.dc.b	$9E,$FF,$1F,$C7,$0C,$3F,$2C,$FF,$62,$9F,$F1,$F3,$9A,$FC,$97,$D9
	.dc.b	$59,$FA,$7F,$44,$F8,$8F,$F2,$4C,$E7,$85,$39,$BF,$AD,$3E,$2B,$FB
	.dc.b	$6D,$A3,$49,$A7,$43,$AF,$BB,$6F,$F0,$1F,$5D,$F7,$B4,$FD,$77,$D7
	.dc.b	$3E,$1B,$FC,$B7,$39,$AF,$D5,$4E,$5E,$3F,$19,$FE,$66,$9F,$C3,$E9
	.dc.b	$FC,$85,$AF,$EC,$5A,$FE,$4B,$EB,$FF,$BE,$BF,$CA,$9A,$FF,$29,$6B
	.dc.b	$FB,$96,$BC,$7E,$07,$FC,$DF,$1C,$B2,$FE,$0C,$BE,$03,$33,$FE,$74
	.dc.b	$FB,$0F,$E7,$15,$FC,$2D,$B9,$7F,$05,$7F,$85,$C6,$7D,$2E,$04,$9F
	.dc.b	$D3,$4E,$9E,$11,$FE,$30,$C7,$FD,$D7,$F2,$BF,$EE,$CF,$B2,$7C,$8F
	.dc.b	$F9,$47,$17,$D2,$7F,$51,$9C,$DF,$F9,$54,$5F,$1B,$E1,$5B,$75,$3A
	.dc.b	$15,$C9,$5F,$6C,$79,$47,$5C,$2B,$AC,$3A,$39,$1E,$C7,$3B,$E5,$18
	.dc.b	$C7,$C3,$68,$C1,$FC,$AA,$FF,$B6,$5F,$8E,$3E,$19,$E6,$AF,$5E,$1E
	.dc.b	$28,$CE,$3D,$EF,$C7,$1F,$E7,$7C,$3C,$F8,$F8,$AB,$A4,$FC,$17,$85
	.dc.b	$F3,$8F,$88,$7C,$62,$9E,$83,$06,$0F,$57,$D9,$7F,$06,$F8,$AE,$0D
	.dc.b	$8F,$9E,$D7,$EB,$35,$FF,$46,$2E,$CF,$A3,$1C,$1A,$F0,$6B,$C1,$AF
	.dc.b	$89,$E3,$7F,$6F,$3F,$F0,$3C,$FD,$86,$BF,$2C,$E1,$8F,$DC,$BF,$13
	.dc.b	$BF,$67,$FD,$41,$7C,$DB,$7E,$19,$E1,$87,$A3,$5D,$63,$F7,$5A,$FD
	.dc.b	$E1,$8F,$67,$0D,$C7,$64,$C3,$B4,$F9,$CE,$37,$EE,$8F,$F5,$A1,$EC
	.dc.b	$9E,$13,$EC,$3D,$DD,$4E,$07,$DD,$FE,$AF,$6F,$DC,$36,$BF,$AE,$CE
	.dc.b	$9F,$DB,$FC,$46,$6F,$95,$7E,$15,$B3,$FA,$C6,$BF,$E3,$CA,$E4,$FE
	.dc.b	$D7,$7F,$F0,$4F,$CF,$77,$6C,$7F,$6D,$6B,$FE,$FD,$E6,$7B,$DB,$F3
	.dc.b	$D7,$F9,$5B,$DD,$FF,$B5,$5A,$FE,$B5,$AF,$F7,$8E,$97,$D5,$7A,$5B
	.dc.b	$1E,$96,$BF,$E5,$46,$BF,$CC,$E3,$E1,$47,$92,$3F,$A0,$E1,$89,$F8
	.dc.b	$DF,$66,$FA,$1A,$4A,$FF,$03,$43,$D8,$C4,$7F,$98,$A3,$83,$F2,$DA
	.dc.b	$E6,$7E,$84,$63,$9F,$FC,$79,$F6,$6F,$92,$FA,$CD,$BE,$F7,$F6,$7E
	.dc.b	$CF,$FD,$67,$1F,$E1,$76,$BF,$C8,$39,$7F,$C7,$D1,$8C,$3F,$C9,$0F
	.dc.b	$FA,$9A,$78,$7C,$67,$E0,$63,$8C,$7F,$9D,$DA,$F2,$8F,$C3,$7D,$C7
	.dc.b	$F2,$48,$FC,$33,$5F,$F4,$0B,$5F,$F4,$03,$5F,$F2,$87,$DA,$3E,$3B
	.dc.b	$F2,$0D,$8F,$C8,$35,$FF,$45,$B5,$FF,$45,$79,$FF,$C6,$3A,$0D,$24
	.dc.b	$7E,$6F,$D2,$7D,$75,$24,$7E,$76,$32,$FE,$7F,$E5,$5E,$AF,$E8,$DB
	.dc.b	$19,$8C,$1F,$A7,$FB,$3F,$F7,$C4,$7E,$9D,$AF,$FA,$89,$AF,$FA,$85
	.dc.b	$AF,$F9,$A3,$E9,$7F,$61,$EF,$1A,$48,$FD,$97,$6B,$E1,$BF,$6C,$D8
	.dc.b	$CC,$60,$FD,$D7,$C0,$BE,$EE,$92,$7F,$79,$19,$7F,$7D,$F0,$9F,$EC
	.dc.b	$17,$FD,$72,$D7,$FD,$70,$D7,$FD,$10,$D7,$FC,$D3,$F0,$DF,$EE,$17
	.dc.b	$FD,$43,$F2,$3F,$C2,$E3,$11,$FE,$A2,$8F,$E9,$FF,$5F,$FD,$62,$3F
	.dc.b	$B4,$35,$34,$9F,$59,$C0,$91,$FD,$6E,$11,$D3,$EE,$E4,$E5,$AF,$F8
	.dc.b	$1A,$5F,$53,$E2,$24,$73,$7F,$B1,$FD,$83,$E5,$E9,$27,$FB,$54,$65
	.dc.b	$BF,$6D,$FC,$BD,$42,$30,$D8,$FE,$49,$F7,$9F,$BB,$7F,$6C,$7F,$82
	.dc.b	$63,$4F,$F8,$71,$AF,$F8,$8D,$AF,$F8,$3C,$EF,$3E,$83,$F9,$97,$C4
	.dc.b	$FF,$20,$F0,$7F,$DF,$9E,$5A,$31,$E1,$F1,$C3,$BD,$7F,$B2,$FE,$8F
	.dc.b	$FD,$1F,$8F,$FA,$85,$AF,$9B,$BD,$18,$9E,$1F,$51,$FD,$2E,$31,$87
	.dc.b	$FB,$35,$B7,$FE,$AE,$6D,$F2,$AF,$EA,$5F,$39,$FE,$94,$AF,$F5,$82
	.dc.b	$F8,$2C,$9A,$F8,$39,$3A,$8C,$BC,$23,$FB,$07,$CE,$BF,$31,$75,$61
	.dc.b	$B5,$FE,$C1,$9D,$4E,$7D,$7D,$3F,$E2,$78,$CC,$FF,$93,$9A,$FF,$95
	.dc.b	$5A,$FD,$37,$3B,$36,$FF,$22,$47,$8F,$D7,$F1,$1D,$AB,$FD,$9D,$F6
	.dc.b	$9F,$D8,$9B,$EB,$7F,$75,$8C,$47,$D1,$5F,$71,$99,$93,$C0,$E7,$1F
	.dc.b	$DF,$7E,$8F,$F8,$EC,$7D,$97,$FC,$68,$FF,$B7,$5A,$E7,$CA,$E0,$71
	.dc.b	$F7,$64,$6E,$3F,$37,$F6,$B2,$3F,$DD,$47,$E8,$1F,$F0,$17,$5F,$F0
	.dc.b	$BD,$BF,$86,$74,$27,$F5,$A4,$AF,$D5,$B5,$FD,$19,$FA,$AE,$8F,$D4
	.dc.b	$9D,$DF,$C9,$7D,$AB,$E5,$BF,$C3,$4B,$E7,$29,$88,$93,$FE,$FB,$9D
	.dc.b	$36,$72,$FE,$CB,$8B,$F1,$1C,$7F,$87,$24,$7E,$1F,$9B,$F8,$DE,$9F
	.dc.b	$76,$15,$C1,$D5,$70,$7A,$57,$08,$FC,$1E,$0E,$6F,$CB,$B6,$BF,$28
	.dc.b	$D7,$C9,$F2,$EF,$9D,$DD,$7E,$56,$B5,$5F,$91,$6B,$DA,$34,$7E,$93
	.dc.b	$F1,$AF,$F7,$4C,$7E,$85,$AF,$C1,$9F,$06,$7E,$DE,$E7,$D4,$BE,$6F
	.dc.b	$FC,$90,$DB,$FF,$50,$56,$E3,$11,$FC,$1D,$FF,$08,$69,$FE,$A5,$BE
	.dc.b	$77,$C8,$95,$CA,$75,$5C,$21,$4F,$AE,$FE,$23,$1B,$32,$E1,$FE,$2C
	.dc.b	$DC,$3D,$4B,$EB,$7F,$CC,$0E,$70,$F2,$36,$B9,$F1,$EA,$BF,$CC,$DF
	.dc.b	$5F,$FE,$8E,$9F,$E6,$18,$A3,$11,$ED,$6B,$E8,$6C,$6F,$E3,$39,$F9
	.dc.b	$32,$BD,$53,$96,$D7,$BD,$B8,$F8,$8F,$1F,$15,$8F,$F3,$B4,$ED,$FF
	.dc.b	$74,$57,$96,$BF,$CF,$93,$99,$F4,$FC,$47,$F7,$9A,$FF,$45,$69,$FE
	.dc.b	$81,$8C,$11,$FF,$43,$3F,$99,$9F,$F4,$5F,$9D,$89,$FE,$C1,$F6,$4F
	.dc.b	$B3,$7C,$CF,$FC,$00,$F8,$9F,$EA,$D3,$98,$C6,$1F,$EE,$78,$C3,$FE
	.dc.b	$FE,$9D,$FC,$97,$FB,$7D,$E3,$7D,$27,$47,$17,$1F,$23,$FE,$79,$85
	.dc.b	$F4,$8C,$3F,$E5,$F8,$E2,$6A,$78,$D7,$6C,$3C,$B3,$96,$D7,$1B,$E7
	.dc.b	$7C,$2B,$C5,$5A,$8C,$47,$8B,$0F,$53,$A3,$C3,$7E,$5C,$3C,$11,$89
	.dc.b	$DB,$E1,$37,$5D,$2B,$84,$7C,$99,$DC,$62,$38,$61,$E6,$C3,$D3,$5A
	.dc.b	$BE,$44,$8C,$D7,$A5,$8F,$81,$CB,$49,$5D,$5F,$41,$C1,$C1,$30,$FB
	.dc.b	$71,$97,$DF,$19,$8F,$51,$93,$E3,$39,$7E,$51,$93,$DB,$1C,$8E,$AF
	.dc.b	$91,$C3,$C9,$CC,$73,$C3,$E3,$13,$0F,$89,$7D,$63,$17,$70,$F8,$CF
	.dc.b	$B8,$F8,$47,$F8,$C4,$FA,$C5,$8F,$CD,$B6,$3F,$30,$E7,$E2,$3F,$CF
	.dc.b	$CF,$13,$F2,$EE,$1D,$7B,$79,$72,$DA,$FF,$62,$3F,$94,$9E,$78,$FE
	.dc.b	$87,$89,$E7,$FF,$55,$8F,$D0,$CE,$B4,$FD,$9B,$71,$FD,$93,$5F,$F3
	.dc.b	$2D,$71,$9D,$E1,$DC,$93,$FA,$AC,$3B,$EB,$8C,$7B,$D5,$BF,$F2,$B5
	.dc.b	$6A,$FF,$43,$8F,$7C,$BD,$71,$8B,$FD,$8D,$6F,$0E,$73,$92,$3E,$FC
	.dc.b	$BC,$D1,$EA,$AF,$03,$4F,$34,$69,$F9,$67,$2C,$7D,$AE,$B2,$E2,$EA
	.dc.b	$FC,$27,$78,$CE,$3E,$88,$CD,$7B,$AF,$B4,$E6,$3A,$CE,$AC,$7C,$EB
	.dc.b	$E7,$97,$C3,$3E,$96,$5F,$06,$F8,$92,$BD,$97,$E3,$7D,$46,$23,$4F
	.dc.b	$53,$27,$AD,$CC,$F5,$8E,$04,$34,$79,$9D,$38,$26,$1E,$D7,$58,$FC
	.dc.b	$0C,$3D,$59,$7C,$AB,$E3,$7B,$E0,$FE,$D1,$E4,$FF,$8A,$7A,$3F,$C1
	.dc.b	$DC,$7F,$E5,$CD,$BF,$B0,$CE,$5F,$ED,$9E,$4D,$D7,$FB,$64,$FF,$B4
	.dc.b	$F0,$E7,$5F,$D8,$A3,$16,$6F,$87,$27,$FB,$3B,$6F,$FB,$3D,$7F,$92
	.dc.b	$E7,$FB,$8C,$F7,$BF,$EF,$DE,$1D,$AD,$5F,$FA,$8D,$7A,$B9,$F5,$B4
	.dc.b	$DC,$7F,$BD,$BE,$42,$3F,$DE,$1F,$57,$4F,$F9,$96,$3F,$C0,$6D,$BF
	.dc.b	$F0,$DF,$A9,$E3,$8C,$6D,$FE,$1C,$F1,$F0,$66,$6D,$FF,$83,$3C,$07
	.dc.b	$57,$AD,$B1,$9F,$A1,$C4,$C8,$E0,$EE,$31,$97,$F9,$97,$5F,$F2,$15
	.dc.b	$7F,$92,$7E,$CB,$EA,$79,$3F,$E2,$3F,$12,$F8,$C9,$CD,$F0,$C3,$FD
	.dc.b	$50,$D7,$FD,$50,$D7,$FD,$4E,$D7,$FC,$DA,$F1,$DB,$FC,$E1,$87,$7C
	.dc.b	$7F,$D7,$5F,$6E,$F9,$2E,$F9,$F9,$23,$F5,$F3,$AC,$71,$D7,$96,$46
	.dc.b	$4E,$71,$88,$FF,$2D,$47,$C9,$F0,$BE,$53,$4F,$9C,$58,$DB,$F4,$63
	.dc.b	$F3,$10,$FA,$F7,$CC,$E1,$7C,$CD,$8C,$36,$FF,$A1,$4F,$06,$CF,$FA
	.dc.b	$46,$BF,$A9,$4E,$5F,$E9,$8D,$D7,$4D,$AF,$E2,$3E,$DF,$F8,$DB,$E6
	.dc.b	$3B,$5F,$FB,$17,$E1,$BF,$41,$EB,$6E,$3F,$CE,$2A,$33,$96,$9B,$87
	.dc.b	$FB,$2D,$B4,$CB,$FD,$15,$E8,$3E,$47,$F5,$8D,$BF,$F3,$0A,$F8,$3C
	.dc.b	$CC,$9C,$BC,$23,$17,$FE,$B7,$FA,$B7,$A5,$F9,$27,$F4,$B7,$FE,$F0
	.dc.b	$72,$FF,$BF,$DC,$CF,$FA,$E1,$CB,$FE,$FB,$6B,$FE,$FB,$6B,$CE,$7E
	.dc.b	$83,$67,$A4,$78,$7D,$2F,$13,$FC,$27,$DF,$7C,$D5,$E3,$C3,$6F,$FD
	.dc.b	$03,$3A,$6C,$E9,$A7,$FA,$1A,$BF,$8E,$37,$C8,$FE,$2D,$1F,$CA,$71
	.dc.b	$FA,$13,$FE,$AF,$8D,$C6,$0F,$F5,$5B,$8D,$BF,$D7,$CD,$9E,$1E,$4F
	.dc.b	$4B,$FC,$B7,$07,$F3,$1E,$47,$C5,$7E,$CE,$FF,$AD,$E5,$F9,$DA,$D4
	.dc.b	$E6,$3F,$AF,$F3,$7E,$26,$BF,$37,$5F,$2A,$72,$D7,$94,$68,$FF,$5F
	.dc.b	$38,$3D,$57,$FE,$8E,$FB,$1F,$2C,$EA,$DF,$05,$CD,$B1,$FA,$A8,$C1
	.dc.b	$99,$91,$FA,$E6,$DF,$EA,$BE,$77,$91,$27,$F6,$6F,$5F,$4A,$47,$ED
	.dc.b	$9B,$3F,$B2,$F9,$7F,$11,$27,$F7,$6F,$F5,$9F,$88,$7D,$3D,$24,$FF
	.dc.b	$91,$7D,$04,$EA,$31,$F1,$4F,$85,$FE,$C4,$D8,$CC,$60,$FE,$CD,$F0
	.dc.b	$CF,$94,$A4,$9E,$D1,$97,$C1,$F1,$AF,$96,$FE,$DC,$D8,$CC,$60,$FE
	.dc.b	$E5,$F1,$FF,$E3,$DD,$BF,$DC,$AD,$7F,$E0,$03,$5F,$F7,$DB,$74,$7C
	.dc.b	$67,$F3,$9B,$F8,$0D,$AF,$F6,$DB,$5F,$F7,$8B,$5F,$F1,$5C,$BE,$97
	.dc.b	$10,$8C,$79,$9F,$E0,$79,$CC,$62,$FE,$77,$C9,$7F,$6F,$BF,$F7,$EB
	.dc.b	$5E,$AD,$7A,$36,$5F,$26,$FA,$7F,$F0,$A3,$6F,$ED,$36,$FF,$DE,$AD
	.dc.b	$7F,$53,$5C,$BE,$8B,$89,$2B,$8B,$FE,$1F,$F9,$87,$35,$24,$FE,$1E
	.dc.b	$32,$FE,$27,$E8,$DE,$77,$F8,$9D,$B1,$98,$C1,$1F,$F2,$AF,$88,$7E
	.dc.b	$6C,$F0,$CB,$F8,$56,$48,$7C,$87,$F8,$9D,$D5,$F1,$AF,$A5,$F4,$7A
	.dc.b	$C3,$A9,$F6,$3E,$73,$FA,$14,$F4,$9D,$3D,$A3,$35,$CF,$15,$6A,$BE
	.dc.b	$4B,$5F,$C1,$B5,$F0,$E3,$FE,$47,$8E,$B5,$CD,$B8,$70,$F9,$3F,$F2
	.dc.b	$04,$62,$FF,$15,$F3,$CF,$7F,$FC,$88,$FF,$1B,$8F,$86,$E6,$FD,$B1
	.dc.b	$C3,$C6,$DD,$7F,$95,$EB,$53,$FE,$C9,$CD,$18,$8F,$C1,$B5,$FE,$EF
	.dc.b	$C4,$F9,$CF,$E5,$1E,$3F,$E4,$1B,$19,$CB,$FD,$99,$39,$6C,$66,$BF
	.dc.b	$94,$B7,$5F,$99,$F3,$FF,$DF,$EF,$85,$7D,$EF,$8F,$FF,$19,$66,$69
	.dc.b	$3D,$BF,$F3,$93,$FC,$FE,$35,$1F,$E6,$98,$FE,$19,$3C,$E3,$FC,$46
	.dc.b	$FF,$A8,$BE,$37,$FD,$47,$F4,$1F,$E3,$98,$FC,$8F,$37,$E2,$63,$11
	.dc.b	$FD,$3A,$77,$DB,$FD,$43,$4F,$CA,$56,$AB,$A0,$78,$C6,$A7,$FD,$E1
	.dc.b	$5C,$AB,$BE,$D5,$EC,$3E,$8B,$8E,$3F,$9B,$7B,$42,$BF,$D7,$71,$AA
	.dc.b	$B7,$DA,$BC,$4F,$9E,$FD,$8E,$71,$E1,$3A,$3F,$B5,$9D,$4C,$1F,$E2
	.dc.b	$B9,$E2,$EB,$0F,$46,$3E,$D8,$C1,$C8,$F5,$94,$E8,$58,$C7,$D1,$3E
	.dc.b	$6D,$BE,$1B,$5C,$E7,$53,$93,$8F,$3F,$F8,$FE,$74,$D7,$D7,$28,$C4
	.dc.b	$7F,$82,$23,$18,$72,$3E,$96,$6F,$95,$FF,$06,$E9,$F8,$28,$C5,$FD
	.dc.b	$7B,$E2,$FC,$66,$CE,$EB,$8D,$7F,$85,$8E,$35,$BC,$3D,$15,$F8,$EB
	.dc.b	$E3,$5B,$BE,$37,$FE,$F8,$8E,$35,$BB,$E3,$1E,$B6,$CE,$EB,$8B,$FE
	.dc.b	$95,$6D,$F6,$80,$70,$EC,$F8,$71,$C2,$7A,$F0,$FE,$CF,$C8,$FD,$16
	.dc.b	$FF,$D8,$B8,$BF,$C0,$98,$7F,$12,$AD,$36,$1E,$E7,$08,$FE,$CD,$1C
	.dc.b	$1F,$E7,$BF,$75,$FD,$21,$E4,$D7,$F8,$26,$1C,$77,$79,$1F,$EE,$F9
	.dc.b	$FE,$43,$18,$F5,$35,$39,$9F,$E9,$3C,$EF,$3F,$35,$E3,$D1,$F7,$7F
	.dc.b	$A2,$E4,$FE,$80,$DD,$FF,$D3,$39,$73,$9F,$E8,$9D,$9F,$D9,$33,$F8
	.dc.b	$0F,$56,$E5,$D6,$35,$1F,$CE,$3C,$4C,$4C,$8E,$31,$E3,$F7,$BF,$C4
	.dc.b	$51,$F2,$DA,$FF,$62,$6D,$FF,$55,$5F,$49,$C7,$CD,$F6,$78,$AF,$94
	.dc.b	$A6,$70,$E9,$38,$7E,$C4,$ED,$C9,$F6,$FE,$93,$FC,$74,$DF,$47,$E4
	.dc.b	$46,$23,$FB,$AB,$FF,$00,$1D,$1F,$83,$F2,$BF,$8E,$37,$C2,$E3,$5F
	.dc.b	$46,$73,$3F,$DF,$BE,$99,$E6,$7D,$56,$C7,$13,$F1,$DF,$4B,$FD,$55
	.dc.b	$FC,$87,$8F,$F5,$DF,$F1,$93,$79,$D3,$B9,$FF,$05,$CE,$7C,$EF,$F5
	.dc.b	$3C,$7E,$65,$AE,$8F,$C3,$79,$FF,$E1,$07,$F3,$CF,$F0,$D6,$F8,$3A
	.dc.b	$F3,$BF,$19,$ED,$FF,$B6,$67,$F4,$8F,$FB,$C9,$BF,$09,$AF,$3F,$F2
	.dc.b	$5F,$4B,$E0,$E9,$FF,$08,$3F,$EF,$6A,$FD,$63,$5F,$F1,$1B,$BE,$EF
	.dc.b	$F2,$FF,$C4,$7E,$6F,$D6,$FE,$1F,$7F,$E1,$96,$CE,$AF,$F6,$DE,$03
	.dc.b	$E9,$7F,$44,$FE,$06,$3F,$C5,$ED,$BF,$F1,$13,$B8,$FD,$F7,$B1,$FE
	.dc.b	$E9,$6F,$D9,$7E,$AE,$0D,$F8,$8F,$D7,$7B,$7F,$E3,$5A,$FF,$1D,$AF
	.dc.b	$8A,$C9,$AC,$FF,$69,$EE,$E6,$BF,$86,$3E,$17,$FC,$86,$DC,$FF,$C6
	.dc.b	$CE,$E3,$FD,$A0,$3E,$9B,$FB,$9D,$7F,$20,$F3,$3F,$D0,$3F,$4D,$FC
	.dc.b	$2F,$DF,$3F,$C5,$6F,$87,$AB,$70,$7C,$C6,$FE,$22,$57,$FB,$1A,$72
	.dc.b	$DF,$98,$E9,$F1,$72,$30,$6F,$E7,$3A,$74,$48,$C1,$FE,$D0,$FA,$4F
	.dc.b	$1B,$FC,$73,$57,$CC,$F9,$CF,$34,$EA,$7D,$EF,$F9,$1D,$E5,$F0,$7A
	.dc.b	$3F,$94,$7D,$2F,$DB,$E2,$FF,$74,$D6,$EB,$FD,$BB,$87,$F1,$6E,$4F
	.dc.b	$F3,$57,$1F,$FA,$9B,$D3,$E5,$5D,$0F,$23,$F2,$4C,$BF,$E7,$6F,$8E
	.dc.b	$7D,$47,$F2,$D6,$C7,$F4,$56,$BB,$FA,$0D,$4C,$8D,$3C,$63,$F9,$B7
	.dc.b	$C8,$FF,$81,$E7,$FA,$73,$FE,$F0,$6F,$C7,$EB,$E1,$64,$60,$87,$FA
	.dc.b	$2A,$79,$2C,$FF,$D1,$1F,$2C,$FA,$67,$85,$FD,$5E,$30,$D8,$CD,$7F
	.dc.b	$9C,$27,$87,$C2,$66,$B5,$F0,$52,$B5,$7C,$1F,$07,$1F,$F7,$2C,$3A
	.dc.b	$C6,$AB,$83,$A1,$1A,$7F,$D4,$7F,$4F,$FD,$4E,$BF,$DA,$6D,$8F,$F6
	.dc.b	$93,$7D,$32,$F9,$4F,$EC,$0E,$49,$F0,$78,$8F,$EC,$9F,$2F,$FC,$7E
	.dc.b	$FF,$D6,$2F,$F1,$F6,$FA,$5A,$F4,$A4,$7F,$80,$67,$5C,$1F,$EA,$96
	.dc.b	$E7,$FE,$E8,$7A,$47,$F0,$3E,$AF,$CE,$D7,$E8,$7B,$7F,$3A,$7E,$62
	.dc.b	$EC,$7E,$97,$A3,$F4,$26,$1F,$D0,$CE,$1D,$1C,$CF,$F0,$A7,$D3,$3E
	.dc.b	$8B,$FC,$00,$BE,$8A,$98,$89,$3C,$23,$2F,$28,$FC,$16,$12,$3E,$EB
	.dc.b	$F8,$5F,$2F,$11,$F6,$78,$FF,$19,$38,$D3,$F1,$FC,$BF,$C5,$DF,$C8
	.dc.b	$E5,$FE,$26,$BF,$C6,$35,$E4,$D8,$FC,$BF,$D3,$3E,$9F,$F2,$D3,$9C
	.dc.b	$3F,$19,$F4,$1F,$DA,$30,$F1,$37,$0E,$FD,$5C,$0F,$CF,$46,$6B,$8F
	.dc.b	$BD,$9B,$FF,$1E,$7D,$1B,$E5,$3F,$C5,$0D,$8C,$C6,$3E,$0B,$2C,$8C
	.dc.b	$9C,$A3,$FC,$61,$F4,$8E,$5A,$49,$FF,$25,$43,$E7,$B1,$1F,$A3,$EE
	.dc.b	$4F,$18,$DD,$74,$AF,$F1,$E7,$D0,$FF,$60,$95,$A1,$E6,$E8,$DF,$C9
	.dc.b	$CA,$FF,$2D,$4E,$9B,$39,$7C,$51,$E2,$77,$8F,$80,$F0,$F2,$58,$E7
	.dc.b	$F6,$FE,$43,$0F,$FB,$7F,$83,$FC,$A5,$F2,$9F,$83,$7F,$CD,$9D,$38
	.dc.b	$64,$FF,$9A,$E3,$66,$1F,$EB,$FF,$4C,$FA,$77,$D4,$7F,$B7,$FE,$4F
	.dc.b	$FC,$FF,$4A,$D4,$E6,$3F,$D0,$71,$87,$FD,$9D,$3C,$23,$FC,$79,$18
	.dc.b	$37,$5D,$9C,$1D,$5C,$7C,$D7,$F9,$8E,$31,$7D,$DF,$17,$D4,$7F,$BB
	.dc.b	$70,$ED,$7C,$1B,$5B,$C3,$C9,$87,$37,$C6,$DA,$E1,$87,$92,$B8,$46
	.dc.b	$23,$B9,$88,$8F,$A2,$BA,$DF,$BF,$0F,$24,$F3,$35,$7C,$EB,$71,$F2
	.dc.b	$27,$6F,$B8,$DD,$F6,$C3,$D1,$3A,$AE,$35,$B9,$F5,$57,$BE,$B5,$7D
	.dc.b	$C9,$19,$AE,$98,$7C,$67,$B9,$EB,$AE,$33,$A8,$F5,$5F,$89,$C1,$E1
	.dc.b	$AE,$51,$98,$E6,$F9,$1B,$1F,$15,$E3,$87,$A8,$97,$E0,$8D,$9E,$09
	.dc.b	$E0,$E0,$F0,$19,$3E,$51,$D3,$0F,$7B,$EB,$31,$87,$82,$BB,$47,$E7
	.dc.b	$0F,$A0,$5B,$FC,$D3,$6B,$F2,$CE,$7E,$4D,$EB,$F1,$8F,$D0,$C6,$1C
	.dc.b	$57,$FA,$A6,$75,$1F,$A9,$9D,$3C,$E3,$F2,$D1,$C7,$2F,$D0,$71,$3D
	.dc.b	$16,$DA,$B9,$87,$E4,$70,$E1,$AF,$ED,$B0,$E1,$7A,$C7,$F4,$15,$AB
	.dc.b	$F3,$E1,$FA,$BB,$DC,$73,$8E,$18,$7E,$E6,$B8,$4F,$F2,$0C,$7C,$6E
	.dc.b	$C9,$87,$6B,$E9,$3A,$8C,$4E,$AF,$A5,$6B,$1E,$F3,$E4,$34,$0F,$03
	.dc.b	$A7,$E5,$45,$C7,$D3,$7D,$89,$7D,$AB,$F8,$6C,$EA,$31,$3A,$9F,$03
	.dc.b	$F4,$30,$F4,$4E,$89,$3A,$7E,$74,$6A,$30,$43,$CA,$78,$1D,$38,$9F
	.dc.b	$03,$67,$9E,$1E,$DC,$BF,$11,$3F,$08,$F8,$25,$73,$5C,$B0,$F6,$4F
	.dc.b	$86,$73,$7E,$88,$E3,$5E,$09,$CC,$79,$EB,$8B,$E3,$39,$65,$F8,$08
	.dc.b	$CC,$F0,$E5,$FE,$C5,$E5,$FE,$0B,$BF,$F0,$71,$FD,$97,$CB,$C6,$9F
	.dc.b	$66,$36,$FF,$64,$8C,$E5,$FD,$B3,$9D,$86,$2F,$95,$FF,$6A,$8C,$4F
	.dc.b	$F9,$46,$3F,$B6,$F8,$B8,$FB,$7C,$B6,$32,$4C,$3F,$B7,$F8,$FA,$8F
	.dc.b	$EE,$BE,$D7,$4F,$77,$AB,$63,$27,$F7,$A8,$CB,$4F,$F0,$6F,$95,$E7
	.dc.b	$ED,$FF,$56,$C7,$F8,$43,$DA,$C4,$C8,$D9,$92,$E3,$FE,$13,$F1,$1D
	.dc.b	$5F,$05,$B3,$FE,$28,$FA,$9C,$4C,$8F,$F1,$5B,$B8,$C5,$7F,$91,$F7
	.dc.b	$FF,$17,$57,$FA,$37,$C7,$7D,$4F,$D0,$F3,$7F,$C3,$5E,$45,$9C,$D7
	.dc.b	$FA,$4E,$F8,$E1,$FE,$40,$C3,$84,$63,$9F,$FD,$3F,$7C,$2B,$C1,$86
	.dc.b	$8E,$DF,$52,$F8,$1C,$65,$FE,$9F,$8F,$DE,$36,$B7,$E0,$70,$64,$64
	.dc.b	$E1,$18,$8F,$91,$1F,$1F,$C3,$F9,$0D,$3F,$C9,$E5,$8D,$BF,$E5,$28
	.dc.b	$FC,$8C,$3E,$A1,$F1,$B8,$5F,$07,$63,$0D,$BF,$E2,$9A,$FF,$22,$BE
	.dc.b	$53,$97,$FD,$60,$62,$76,$7F,$8A,$AF,$BB,$88,$FE,$3D,$18,$7F,$D7
	.dc.b	$1E,$FF,$F8,$CB,$E8,$7A,$61,$FE,$A3,$F8,$EF,$D3,$7A,$FB,$8C,$63
	.dc.b	$FE,$A9,$CF,$35,$A6,$C6,$4F,$07,$82,$F4,$DF,$37,$FB,$16,$CE,$97
	.dc.b	$C5,$66,$64,$E5,$E1,$18,$D7,$FD,$31,$F5,$3F,$E7,$5A,$FF,$6F,$3F
	.dc.b	$A6,$8C,$FA,$9F,$93,$8C,$69,$FE,$EA,$78,$4F,$2E,$ED,$56,$DB,$7D
	.dc.b	$23,$97,$A9,$8A,$FF,$7B,$7C,$3F,$F0,$BF,$81,$FB,$77,$FF,$00,$23
	.dc.b	$1C,$B9,$6C,$64,$F8,$B3,$C2,$BF,$D1,$18,$7F,$16,$9F,$92,$D8,$CF
	.dc.b	$95,$FC,$66,$31,$A7,$F2,$46,$D7,$F2,$36,$D7,$FA,$31,$FE,$38,$F6
	.dc.b	$F2,$BF,$25,$C7,$FC,$9B,$99,$F2,$3F,$B1,$C3,$F1,$53,$98,$C5,$FE
	.dc.b	$2A,$31,$DB,$CA,$BF,$1D,$7C,$A3,$15,$F9,$6E,$95,$F7,$F4,$DF,$53
	.dc.b	$AD,$7F,$3D,$4F,$85,$F0,$7D,$AF,$E9,$DB,$A7,$E6,$F2,$FD,$66,$1F
	.dc.b	$AB,$F8,$1F,$D6,$36,$75,$83,$EA,$BD,$5E,$07,$EC,$5B,$3A,$F5,$BD
	.dc.b	$CF,$ED,$5A,$F2,$F8,$9F,$8D,$F7,$9F,$2B,$6F,$3C,$63,$9B,$FA,$67
	.dc.b	$C7,$BE,$AE,$FD,$06,$1B,$8F,$E9,$3E,$49,$F1,$7F,$D4,$DF,$D3,$46
	.dc.b	$63,$1E,$27,$F5,$7F,$8C,$7C,$2D,$F4,$F0,$DB,$FD,$7F,$CA,$33,$FE
	.dc.b	$BC,$BE,$86,$C6,$1F,$EC,$5F,$2F,$FE,$34,$E1,$FE,$DA,$FF,$B6,$2F
	.dc.b	$F7,$2E,$5F,$ED,$CE,$7A,$7F,$DB,$DF,$25,$FC,$C6,$7F,$7E,$BE,$2B
	.dc.b	$FB,$73,$99,$DC,$7F,$71,$FA,$BC,$61,$FD,$F1,$E1,$7F,$C2,$1C,$BF
	.dc.b	$DF,$5C,$D7,$FB,$B7,$E6,$BF,$77,$3F,$E0,$15,$EE,$7F,$7D,$73,$3F
	.dc.b	$EE,$C7,$33,$FC,$53,$E7,$1F,$39,$FE,$03,$1F,$53,$FE,$05,$8C,$CF
	.dc.b	$F1,$A9,$CB,$C7,$3D,$46,$DB,$1F,$C8,$3E,$81,$D7,$FE,$0F,$FA,$07
	.dc.b	$CC,$FF,$84,$67,$35,$FC,$9A,$72,$FF,$28,$FA,$B7,$B5,$BC,$3F,$C3
	.dc.b	$11,$89,$ED,$18,$99,$1C,$7C,$6B,$BB,$8B,$FC,$53,$F5,$69,$FF,$11
	.dc.b	$7D,$2F,$F3,$2A,$D5,$FF,$8A,$EB,$4F,$F3,$A9,$D5,$F2,$BE,$3E,$CF
	.dc.b	$F8,$AE,$F7,$87,$21,$ED,$30,$E3,$8F,$5C,$38,$1C,$7E,$6A,$C7,$F4
	.dc.b	$8F,$A6,$FA,$F8,$73,$C1,$1F,$C2,$A7,$5C,$FF,$E3,$B8,$C5,$6D,$E3
	.dc.b	$A7,$F9,$21,$CB,$67,$87,$A1,$FC,$87,$9B,$EE,$DF,$7F,$A7,$7C,$86
	.dc.b	$7A,$7F,$C9,$11,$8C,$FF,$8F,$DF,$0C,$BF,$D2,$FE,$A2,$31,$3F,$E5
	.dc.b	$5F,$A8,$FE,$D4,$FF,$A3,$E7,$58,$7F,$94,$FE,$5B,$F9,$6E,$EF,$A9
	.dc.b	$7D,$06,$7D,$8F,$F2,$C4,$63,$1F,$ED,$B7,$C0,$9E,$3E,$E3,$F9,$37
	.dc.b	$07,$FA,$77,$E5,$BF,$D3,$DF,$57,$FE,$27,$8F,$F0,$6F,$5F,$FA,$BA
	.dc.b	$31,$1F,$EA,$31,$3C,$3E,$27,$1C,$5F,$91,$95,$70,$7B,$18,$9D,$DF
	.dc.b	$57,$FD,$B5,$7C,$E7,$C3,$85,$7C,$11,$70,$ED,$87,$59,$F1,$9D,$4E
	.dc.b	$25,$38,$96,$38,$DF,$82,$31,$87,$8D,$FF,$7A,$5F,$78,$C4,$72,$BF
	.dc.b	$87,$7C,$A3,$67,$FB,$B8,$F1,$9F,$EF,$0A,$C3,$83,$E5,$7D,$5F,$DA
	.dc.b	$E1,$FF,$7A,$B6,$3F,$DF,$8D,$7E,$80,$70,$F2,$E9,$FD,$D3,$C1,$AF
	.dc.b	$08,$46,$23,$85,$F8,$23,$8E,$1F,$7B,$27,$CE,$FD,$F8,$C3,$63,$8D
	.dc.b	$FC,$DC,$3A,$E1,$F5,$E7,$8C,$DA,$E0,$FF,$BA,$EB,$6F,$1C,$3D,$3F
	.dc.b	$55,$DA,$3F,$DE,$0D,$4F,$E4,$A7,$8B,$67,$D3,$7F,$95,$3C,$35,$F8
	.dc.b	$16,$E3,$BC,$7F,$19,$87,$8A,$3B,$12,$7D,$AF,$E8,$2D,$F4,$FF,$9E
	.dc.b	$DB,$F4,$51,$8C,$7F,$C2,$0D,$7F,$CE,$3F,$95,$FE,$D9,$CD,$7E,$16
	.dc.b	$75,$1F,$71,$AF,$48,$FB,$EF,$F0,$5F,$6B,$C0,$FF,$28,$E8,$FF,$0C
	.dc.b	$3F,$E6,$A6,$BF,$C8,$DF,$F1,$4B,$98,$E6,$F1,$6F,$3A,$78,$CF,$EF
	.dc.b	$63,$13,$AE,$EC,$CF,$FB,$AD,$AF,$F3,$A6,$CF,$F3,$29,$E0,$7A,$1F
	.dc.b	$E6,$BE,$ED,$9F,$2D,$7F,$33,$AD,$78,$6E,$9A,$49,$FF,$43,$C7,$F8
	.dc.b	$79,$B1,$EA,$8E,$11,$F2,$30,$FF,$45,$CE,$49,$1C,$BE,$4F,$B9,$D5
	.dc.b	$FE,$95,$F5,$88,$F4,$36,$3F,$D2,$ED,$77,$F7,$3A,$9F,$97,$39,$9F
	.dc.b	$94,$D4,$7F,$60,$70,$FF,$62,$F2,$1F,$07,$F5,$1B,$1E,$B6,$BF,$C0
	.dc.b	$B1,$46,$23,$EC,$78,$2F,$90,$FB,$AD,$8F,$88,$D7,$93,$5E,$3D,$5C
	.dc.b	$08,$F0,$7F,$0F,$E7,$FF,$B5,$23,$E7,$35,$F2,$35,$FC,$53,$5F,$83
	.dc.b	$87,$FA,$2B,$E2,$7F,$23,$F0,$C6,$92,$7E,$D7,$D6,$3E,$AF,$F2,$ED
	.dc.b	$8C,$C6,$0F,$CD,$FA,$FF,$DD,$23,$F0,$4D,$7E,$43,$5F,$CF,$35,$F2
	.dc.b	$FB,$4F,$A2,$FD,$2B,$63,$F1,$8D,$7E,$93,$5F,$D4,$7B,$A3,$F5,$A5
	.dc.b	$7F,$28,$D7,$ED,$35,$FD,$87,$C0,$3E,$B3,$F6,$AD,$8F,$CE,$35,$FC
	.dc.b	$13,$5F,$DC,$7C,$1B,$E8,$BF,$7A,$D8,$FD,$23,$5F,$C5,$35,$FE,$01
	.dc.b	$F1,$4F,$99,$FE,$0A,$D8,$FD,$63,$5F,$C9,$35,$FE,$11,$F6,$5F,$E7
	.dc.b	$D7,$F6,$4D,$7F,$30,$D7,$7E,$F7,$F1,$2F,$B3,$C4,$7E,$19,$AF,$F9
	.dc.b	$19,$BC,$86,$FD,$2F,$FB,$22,$7F,$80,$35,$FD,$33,$5F,$E3,$8D,$7F
	.dc.b	$27,$F6,$BF,$EA,$E7,$F8,$43,$5F,$D7,$35,$FE,$48,$D7,$FC,$45,$F5
	.dc.b	$9F,$CF,$5F,$F2,$E0,$71,$E7,$3C,$BD,$FA,$DF,$CE,$AB,$5A,$DB,$E5
	.dc.b	$1F,$DF,$9A,$F0,$3F,$9C,$B8,$7F,$85,$BD,$AB,$C7,$E5,$F3,$3A,$5F
	.dc.b	$F3,$1F,$93,$7D,$0D,$26,$1F,$C5,$32,$F3,$3F,$E7,$49,$7D,$4E,$A3
	.dc.b	$FC,$DD,$1C,$9E,$18,$FE,$F2,$31,$1A,$9F,$E8,$1F,$2C,$9B,$F3,$12
	.dc.b	$B8,$9F,$E8,$29,$D3,$8E,$14,$62,$3F,$CF,$CA,$79,$46,$2B,$FB,$2F
	.dc.b	$D0,$BF,$46,$7C,$EE,$1B,$5E,$09,$D4,$E7,$B3,$4F,$F6,$68,$CC,$FF
	.dc.b	$72,$6B,$FE,$32,$6B,$FE,$63,$73,$AF,$CA,$7C,$5E,$9F,$84,$EB,$3E
	.dc.b	$DF,$91,$F9,$EF,$FA,$65,$B5,$F4,$B5,$FF,$7F,$CE,$FE,$5F,$FB,$55
	.dc.b	$6B,$83,$F9,$5B,$67,$FD,$56,$DA,$2A,$FE,$E5,$F3,$FF,$87,$AF,$F5
	.dc.b	$8A,$E6,$C9,$AE,$49,$3F,$4E,$74,$D9,$CB,$FA,$0E,$6F,$F6,$9B,$F7
	.dc.b	$0F,$D1,$7B,$DF,$9D,$3F,$3D,$F0,$F6,$3F,$37,$1F,$E6,$79,$E3,$1F
	.dc.b	$E2,$06,$BF,$A3,$8F,$F4,$0F,$CC,$7E,$C0,$B1,$FA,$26,$BE,$03,$FD
	.dc.b	$38,$FF,$50,$FA,$BF,$E3,$1F,$02,$8C,$78,$3C,$B7,$56,$A7,$87,$D6
	.dc.b	$66,$31,$87,$FB,$61,$B5,$C2,$F5,$5E,$4E,$47,$E6,$2F,$9E,$C2,$F9
	.dc.b	$BA,$62,$26,$5E,$81,$5C,$A7,$71,$98,$EB,$3F,$03,$6F,$F0,$2E,$BE
	.dc.b	$DA,$FE,$0B,$EB,$64,$D7,$C2,$58,$FF,$09,$F3,$FF,$0F,$9F,$F5,$E3
	.dc.b	$63,$27,$F1,$E8,$CB,$4F,$30,$FA,$C7,$B7,$F4,$9B,$1A,$C7,$CB,$C5
	.dc.b	$FE,$22,$D3,$FD,$27,$EE,$7C,$7F,$17,$FC,$73,$7F,$EC,$CA,$D4,$63
	.dc.b	$5F,$65,$FF,$A6,$BE,$1F,$86,$13,$0F,$89,$7B,$C3,$8D,$7D,$3F,$BE
	.dc.b	$F1,$D7,$53,$E6,$F9,$BE,$46,$ED,$BD,$A4,$7C,$F9,$D6,$66,$DF,$4D
	.dc.b	$3E,$85,$7C,$3F,$4B,$FC,$73,$18,$DF,$FC,$19,$97,$C8,$6D,$78,$CF
	.dc.b	$95,$F3,$7F,$26,$F8,$EB,$FE,$EF,$AD,$E1,$CE,$73,$7F,$E4,$C9,$5F
	.dc.b	$1C,$3B,$56,$AB,$C9,$1E,$19,$D4,$78,$6B,$AE,$1D,$EF,$84,$74,$9E
	.dc.b	$77,$1E,$13,$AB,$E7,$87,$96,$B4,$FC,$18,$CE,$1D,$67,$2F,$B6,$33
	.dc.b	$7E,$BB,$F0,$9D,$5D,$99,$AA,$FC,$82,$E1,$EA,$74,$7B,$6B,$AB,$F0
	.dc.b	$8F,$4C,$FC,$37,$8D,$78,$23,$11,$A2,$61,$ED,$7A,$1C,$23,$C3,$5D
	.dc.b	$AF,$C5,$3D,$EB,$E1,$3B,$39,$E1,$F0,$5C,$47,$53,$6E,$8F,$5B,$98
	.dc.b	$E1,$19,$3B,$3F,$35,$F2,$99,$9F,$19,$2B,$CD,$87,$BE,$77,$7D,$5E
	.dc.b	$F7,$CE,$3C,$66,$63,$B9,$A3,$F2,$55,$D3,$0F,$87,$1F,$87,$3F,$D4
	.dc.b	$27,$D4,$29,$F4,$CB,$87,$E9,$1B,$E8,$7F,$A5,$5F,$CF,$35,$FE,$20
	.dc.b	$D8,$FF,$30,$7B,$9F,$ED,$E6,$CF,$E7,$67,$FD,$45,$87,$ED,$BE,$39
	.dc.b	$E7,$FF,$88,$23,$F6,$D3,$AC,$FF,$60,$DB,$F0,$E3,$FC,$83,$0E,$3C
	.dc.b	$BF,$BB,$AD,$57,$92,$B0,$6F,$1E,$D8,$74,$8C,$CF,$95,$D5,$E3,$1F
	.dc.b	$1C,$5A,$F9,$38,$F8,$2F,$81,$2B,$6F,$F0,$B8,$D4,$60,$93,$E1,$9D
	.dc.b	$61,$E4,$7D,$A7,$38,$D4,$FC,$79,$D4,$FB,$4D,$1C,$23,$58,$79,$E3
	.dc.b	$15,$C8,$F4,$D7,$2C,$BE,$AD,$F9,$2B,$67,$FB,$9B,$0F,$4D,$F1,$8C
	.dc.b	$9E,$19,$EE,$F8,$8C,$46,$D4,$F2,$78,$D7,$82,$75,$39,$9E,$26,$B2
	.dc.b	$FB,$71,$9A,$E1,$39,$CB,$EB,$BE,$18,$CB,$C8,$F0,$E3,$EF,$BF,$C2
	.dc.b	$CE,$8D,$DF,$58,$C9,$2F,$AE,$3F,$17,$0E,$97,$C3,$0E,$95,$F3,$E7
	.dc.b	$C1,$87,$38,$C5,$6F,$2B,$8F,$5B,$FC,$0C,$62,$B9,$5F,$19,$FE,$CD
	.dc.b	$C0,$FC,$2D,$EE,$7E,$1E,$31,$F6,$F6,$73,$8F,$F8,$55,$46,$6B,$4D
	.dc.b	$9D,$3C,$FE,$6D,$E5,$53,$1E,$5D,$AF,$F1,$04,$E9,$B3,$A6,$9F,$B5
	.dc.b	$BF,$CB,$E7,$E2,$8C,$1F,$B9,$C3,$33,$FE,$A0,$8F,$E2,$73,$EA,$CF
	.dc.b	$FC,$EA,$7E,$2A,$74,$FE,$C9,$AF,$E5,$DE,$73,$FC,$17,$D5,$B3,$F2
	.dc.b	$BF,$30,$7D,$6F,$C9,$5F,$5B,$63,$0D,$BF,$F1,$F4,$E9,$B3,$A7,$FC
	.dc.b	$8D,$E3,$7F,$96,$3E,$C3,$F7,$4D,$8C,$F9,$1F,$E4,$F6,$CE,$A0,$D3
	.dc.b	$1E,$2D,$9F,$E3,$DF,$07,$F2,$B2,$F9,$6D,$8C,$9F,$CE,$B3,$FD,$C4
	.dc.b	$7C,$86,$C6,$5F,$D9,$C6,$5B,$1F,$E8,$7D,$93,$FE,$D2,$F4,$7F,$D4
	.dc.b	$B0,$E6,$C5,$7F,$22,$AB,$7F,$E9,$AF,$97,$BA,$BC,$FF,$F6,$35,$FF
	.dc.b	$A6,$67,$31,$8A,$65,$FE,$AF,$84,$EB,$D6,$DB,$6F,$FD,$71,$E0,$F6
	.dc.b	$F4,$CD,$FA,$6C,$32,$76,$7F,$AF,$E3,$19,$78,$23,$FC,$C5,$F4,$8F
	.dc.b	$99,$F5,$F7,$21,$F0,$D8,$75,$5F,$ED,$97,$2D,$AF,$A1,$5C,$5B,$3C
	.dc.b	$2F,$FD,$B0,$D7,$FD,$B0,$E7,$6F,$F6,$C3,$6F,$CB,$F5,$5E,$1B,$F9
	.dc.b	$BE,$E2,$2C,$61,$B3,$FE,$F9,$9D,$36,$74,$D3,$FD,$FD,$EE,$3E,$C7
	.dc.b	$EC,$B6,$33,$D3,$89,$91,$C1,$DC,$63,$3F,$98,$07,$D4,$3E,$6F,$F0
	.dc.b	$F1,$FD,$10,$B1,$9B,$FC,$94,$62,$7F,$2D,$39,$8F,$CF,$35,$FF,$0C
	.dc.b	$B5,$FF,$0E,$BB,$D7,$FD,$FA,$DA,$CF,$C4,$F5,$7C,$D3,$FE,$FF,$F8
	.dc.b	$37,$A5,$FA,$9E,$94,$62,$BF,$5D,$18,$9D,$4E,$71,$FF,$15,$CB,$EE
	.dc.b	$BF,$51,$7C,$1B,$3A,$6B,$FE,$34,$9E,$1D,$9E,$1F,$85,$7B,$94,$98
	.dc.b	$78,$A3,$2F,$93,$E1,$9F,$65,$E9,$6C,$66,$30,$7B,$7E,$E9,$F0,$94
	.dc.b	$93,$EE,$8C,$BF,$03,$D8,$7D,$F7,$C5,$6C,$66,$30,$7C,$9F,$8A,$7C
	.dc.b	$3D,$24,$FC,$A8,$CB,$F3,$21,$EF,$7D,$16,$C6,$63,$07,$D5,$F8,$CF
	.dc.b	$DB,$47,$D4,$6F,$0F,$D2,$6B,$F3,$5A,$FE,$57,$85,$F1,$AF,$A6,$FC
	.dc.b	$03,$6F,$EF,$B6,$BE,$EB,$5F,$B1,$EA,$FF,$45,$7F,$02,$D7,$3E,$6F
	.dc.b	$E2,$E3,$11,$E1,$FA,$8A,$B1,$46,$1B,$5F,$93,$8C,$7A,$8E,$CF,$CB
	.dc.b	$37,$2F,$CA,$B7,$0F,$C9,$B5,$FC,$77,$C9,$7F,$62,$9F,$CD,$B6,$3F
	.dc.b	$32,$D7,$7E,$9F,$E8,$E1,$F3,$B4,$95,$C9,$FD,$47,$CA,$3E,$96,$92
	.dc.b	$7F,$53,$19,$6F,$CC,$F0,$23,$82,$63,$FC,$33,$65,$2B,$FA,$7A,$F9
	.dc.b	$BB,$1F,$D4,$BE,$97,$18,$7E,$F8,$FC,$03,$C5,$C6,$1F,$46,$36,$F7
	.dc.b	$74,$DA,$FE,$09,$1A,$8F,$DC,$35,$EB,$1A,$3C,$47,$8F,$E3,$ED,$78
	.dc.b	$31,$7C,$E3,$EB,$BF,$B4,$BF,$D9,$7C,$75,$7E,$5E,$34,$F1,$74,$E6
	.dc.b	$7F,$8C,$46,$B1,$FE,$26,$D7,$F8,$73,$5F,$D1,$7C,$F7,$F7,$0A,$FE
	.dc.b	$3A,$D8,$FE,$36,$D7,$F8,$B3,$5E,$3E,$BF,$F8,$02,$38,$61,$E8,$9F
	.dc.b	$CE,$47,$F5,$B6,$CF,$F8,$22,$5F,$84,$FB,$AF,$FC,$35,$18,$7F,$C3
	.dc.b	$4D,$73,$1B,$AF,$F6,$CC,$E5,$FE,$A2,$D3,$C4,$BF,$50,$FE,$84,$FF
	.dc.b	$84,$27,$94,$FF,$89,$FE,$27,$FA,$C7,$D1,$7E,$7A,$BF,$A2,$ED,$EB
	.dc.b	$8C,$47,$FB,$AA,$77,$C9,$FE,$EC,$B5,$EE,$3F,$C6,$93,$AA,$E7,$95
	.dc.b	$79,$0C,$3A,$E1,$CA,$75,$3F,$DA,$EB,$FB,$67,$C9,$6B,$1F,$9C,$F9
	.dc.b	$27,$41,$D5,$70,$C7,$C7,$3D,$DF,$45,$79,$E3,$FB,$C1,$CB,$1E,$2E
	.dc.b	$CC,$8C,$3D,$58,$F9,$E3,$47,$B2,$78,$9E,$57,$81,$FE,$53,$29,$FE
	.dc.b	$55,$9C,$38,$3C,$7F,$46,$F9,$97,$D2,$7F,$96,$E3,$0D,$8C,$CF,$F9
	.dc.b	$8E,$72,$DF,$94,$FF,$27,$B7,$1E,$51,$88,$F6,$CA,$31,$1E,$27,$8E
	.dc.b	$3F,$82,$CF,$FC,$CF,$3F,$6D,$AF,$C2,$AE,$57,$F7,$DB,$1F,$1A,$77
	.dc.b	$5C,$FE,$0F,$E8,$BF,$8E,$AD,$B7,$0F,$0D,$7D,$63,$89,$6B,$8B,$E7
	.dc.b	$C3,$C1,$3C,$30,$FF,$1A,$4F,$06,$D7,$07,$FC,$71,$87,$82,$79,$47
	.dc.b	$9E,$74,$FF,$42,$6C,$EE,$3F,$C8,$15,$FE,$CA,$7F,$96,$9F,$2F,$E2
	.dc.b	$BF,$4D,$B7,$E3,$A7,$53,$FE,$48,$1F,$49,$F7,$DE,$31,$AC,$BF,$86
	.dc.b	$73,$BE,$97,$FC,$ED,$5F,$CA,$DA,$F3,$AE,$18,$7F,$1E,$EB,$F0,$CF
	.dc.b	$F1,$46,$DB,$9F,$FA,$64,$FF,$35,$7F,$A6,$36,$77,$0F,$9B,$A6,$23
	.dc.b	$9F,$4F,$2A,$FE,$B0,$F2,$6F,$5E,$A6,$47,$F3,$B8,$E2,$61,$FE,$C1
	.dc.b	$F4,$BF,$E2,$47,$FA,$43,$5C,$F9,$5E,$A8,$FF,$2C,$C6,$A3,$1F,$4B
	.dc.b	$23,$04,$3A,$B9,$79,$9F,$D5,$7E,$95,$1F,$19,$B1,$98,$C7,$CD,$65
	.dc.b	$F9,$7E,$13,$E6,$29,$9A,$FE,$C1,$19,$6F,$77,$03,$6F,$F6,$5F,$17
	.dc.b	$FD,$A0,$DE,$9F,$AF,$1F,$AE,$E6,$FC,$1B,$A9,$FF,$03,$39,$E4,$FF
	.dc.b	$1F,$3F,$E8,$F8,$CF,$C6,$F2,$BF,$C0,$F9,$0F,$B0,$FC,$1F,$D6,$7E
	.dc.b	$46,$FF,$C2,$0D,$CB,$F2,$35,$B9,$FE,$F5,$F4,$FF,$EC,$66,$DF,$E3
	.dc.b	$23,$51,$FE,$19,$74,$FF,$86,$FE,$9F,$FC,$74,$FF,$80,$DF,$F5,$33
	.dc.b	$7E,$0F,$F2,$B1,$A8,$FF,$11,$7A,$5E,$BC,$7F,$46,$BC,$EF,$CD,$B6
	.dc.b	$7F,$C1,$CE,$F7,$FF,$30,$7A,$9F,$EE,$89,$FF,$09,$2F,$8D,$FD,$1B
	.dc.b	$63,$3F,$1D,$FA,$6F,$5F,$FC,$F7,$1F,$E1,$95,$F5,$F9,$34,$FE,$B3
	.dc.b	$D9,$7D,$45,$F9,$B9,$7F,$E2,$19,$CB,$FE,$23,$F7,$53,$FB,$47,$CD
	.dc.b	$1F,$E4,$16,$EF,$FC,$09,$DC,$7F,$8A,$FE,$97,$FD,$4E,$DF,$A5,$FD
	.dc.b	$EC,$1B,$F0,$BF,$C0,$3E,$11,$F3,$57,$A6,$61,$FE,$35,$9C,$BF,$E3
	.dc.b	$7F,$86,$AF,$F9,$4D,$4F,$F0,$A6,$C6,$EB,$F8,$6F,$CF,$BE,$42,$FC
	.dc.b	$94,$AF,$F2,$14,$E5,$FF,$22,$7C,$67,$CC,$6F,$CB,$7F,$16,$83,$79
	.dc.b	$7F,$8D,$CB,$E8,$FF,$2D,$D5,$FD,$E9,$B1,$FE,$F3,$6B,$BE,$1E,$31
	.dc.b	$83,$FB,$04,$7F,$38,$6B,$FC,$D5,$D7,$B5,$C3,$D4,$93,$B8,$FE,$57
	.dc.b	$F2,$6F,$A8,$A4,$9F,$F3,$04,$3E,$A3,$11,$FE,$7D,$8E,$0E,$A7,$9C
	.dc.b	$FF,$36,$F9,$65,$5F,$9A,$97,$FE,$6B,$9C,$B6,$72,$FF,$A2,$61,$3C
	.dc.b	$E7,$FA,$1F,$CC,$BE,$A1,$F4,$38,$6E,$1F,$E8,$69,$D4,$E7,$C1,$D3
	.dc.b	$BF,$7E,$56,$EF,$8A,$EF,$FF,$32,$61,$CA,$35,$5C,$1D,$71,$14,$ED
	.dc.b	$FF,$51,$72,$7F,$97,$2F,$FB,$1B,$6B,$FD,$B8,$D9,$FF,$66,$3F,$EF
	.dc.b	$E7,$FD,$20,$E7,$DD,$FC,$5C,$FF,$62,$8C,$1B,$9F,$F5,$87,$CE,$FE
	.dc.b	$2A,$8B,$97,$FB,$4B,$67,$FD,$38,$EC,$91,$A3,$7C,$7F,$90,$F8,$69
	.dc.b	$5F,$8F,$7F,$27,$EF,$62,$3F,$C9,$DF,$49,$F9,$B3,$FC,$E0,$57,$FC
	.dc.b	$C8,$FE,$7E,$73,$1F,$9F,$6B,$FA,$58,$BD,$06,$CF,$0C,$9F,$41,$FC
	.dc.b	$2A,$FC,$27,$FB,$D5,$C7,$C4,$71,$8C,$78,$9C,$67,$F8,$74,$79,$0C
	.dc.b	$96,$BA,$CE,$CE,$1E,$2F,$A3,$89,$F5,$0F,$A8,$C2,$FA,$2A,$62,$26
	.dc.b	$3D,$EB,$8C,$EA,$31,$5C,$DF,$21,$FC,$1F,$DC,$F8,$64,$73,$BF,$FA
	.dc.b	$25,$B1,$F2,$BE,$99,$EF,$D3,$84,$FC,$63,$A1,$DA,$36,$D8,$DE,$FF
	.dc.b	$27,$2F,$F3,$77,$C0,$FC,$1F,$0F,$FC,$39,$39,$6F,$1F,$E8,$EB,$FC
	.dc.b	$E9,$F3,$DB,$B9,$8F,$C1,$BD,$BF,$A8,$7F,$C4,$6F,$DA,$A7,$D4,$79
	.dc.b	$23,$1E,$7F,$F8,$4D,$BA,$FD,$39,$D6,$7E,$76,$E7,$E3,$BE,$13,$EB
	.dc.b	$6C,$79,$E7,$77,$FB,$BD,$5E,$9F,$F8,$87,$0F,$C5,$4A,$73,$8F,$C9
	.dc.b	$AD,$61,$FE,$B4,$97,$D4,$7D,$39,$CD,$F5,$6B,$CD,$B3,$C2,$3A,$5C
	.dc.b	$7E,$C4,$78,$1B,$5D,$67,$58,$79,$AB,$4D,$9D,$57,$2B,$EE,$78,$23
	.dc.b	$33,$C6,$33,$7F,$85,$9C,$E1,$ED,$9C,$DF,$88,$F0,$C6,$6A,$C7,$CB
	.dc.b	$AE,$2F,$C1,$8C,$BF,$2E,$32,$FC,$38,$CE,$1F,$0E,$3D,$11,$8B,$F4
	.dc.b	$45,$8D,$D7,$2A,$F8,$37,$E3,$AF,$31,$C8,$98,$7B,$EF,$DD,$1C,$CC
	.dc.b	$CF,$DB,$BE,$98,$7C,$0B,$F9,$95,$B7,$93,$D9,$A4,$AF,$13,$C2,$3A
	.dc.b	$9F,$0C,$A7,$C3,$3A,$47,$03,$C4,$E0,$E1,$1C,$CE,$CE,$23,$83,$F4
	.dc.b	$4E,$07,$CD,$3C,$CE,$9C,$47,$27,$C4,$7C,$73,$FA,$F1,$E7,$29,$CC
	.dc.b	$B8,$7E,$55,$BE,$97,$F9,$FD,$FD,$23,$5F,$D7,$35,$FD,$63,$63,$FC
	.dc.b	$9B,$F3,$FF,$C5,$63,$FD,$1B,$5F,$B2,$F8,$07,$9C,$F0,$B5,$73,$1F
	.dc.b	$D3,$E1,$C3,$2F,$4E,$1C,$2F,$58,$FE,$9E,$B5,$7E,$4C,$3F,$98,$DE
	.dc.b	$E3,$9C,$70,$C3,$D6,$EB,$1F,$14,$E8,$98,$75,$9D,$63,$C6,$7B,$1A
	.dc.b	$BF,$2D,$F2,$BE,$13,$93,$6F,$C1,$8D,$4F,$8A,$74,$7C,$AB,$EE,$E3
	.dc.b	$2F,$96,$EB,$1F,$66,$1D,$89,$7C,$DE,$D8,$F9,$EB,$A1,$C7,$1F,$0B
	.dc.b	$E4,$AE,$04,$7F,$C4,$97,$DA,$31,$1C,$AF,$BB,$AA,$E6,$E6,$77,$87
	.dc.b	$A2,$BC,$73,$EB,$CB,$EB,$BB,$39,$63,$E7,$AE,$17,$E0,$C7,$C5,$5E
	.dc.b	$72,$1C,$43,$A7,$13,$C9,$DC,$66,$31,$1A,$BF,$1C,$E5,$F2,$CE,$8D
	.dc.b	$C6,$72,$F3,$4F,$0C,$BE,$0D,$6E,$38,$99,$C7,$75,$F4,$EF,$96,$1F
	.dc.b	$0E,$3E,$FE,$5F,$33,$81,$F5,$1F,$EA,$0C,$3F,$C2,$0B,$EA,$3E,$CC
	.dc.b	$EB,$D5,$B3,$9A,$DC,$62,$B2,$D8,$FF,$0D,$7C,$E3,$C8,$E6,$D8,$CF
	.dc.b	$95,$89,$91,$C1,$DC,$63,$4F,$F0,$A6,$BE,$76,$E3,$F8,$48,$C6,$9F
	.dc.b	$C0,$63,$19,$65,$B8,$FE,$E1,$B7,$D2,$79,$7A,$B8,$BF,$F0,$E7,$D4
	.dc.b	$7F,$BC,$F0,$FF,$1A,$2F,$9D,$F8,$AE,$70,$FF,$0A,$39,$9D,$FB,$FA
	.dc.b	$AD,$B6,$3F,$C9,$5E,$33,$E9,$7F,$4F,$E3,$FF,$96,$1B,$1F,$E4,$79
	.dc.b	$DC,$60,$8F,$2F,$12,$E5,$F1,$BE,$1B,$E2,$65,$F1,$5B,$5F,$C6,$1B
	.dc.b	$3B,$8F,$E3,$73,$B7,$FD,$31,$E7,$FF,$18,$9F,$D0,$D7,$F1,$A8,$C7
	.dc.b	$0B,$9F,$17,$FC,$AA,$AD,$B4,$FF,$47,$FC,$CE,$1F,$F4,$8F,$A9,$7D
	.dc.b	$04,$D8,$C3,$70,$FF,$49,$B6,$FF,$D5,$3D,$FC,$FD,$1F,$F1,$9D,$7F
	.dc.b	$A9,$3E,$07,$13,$23,$FD,$4E,$EE,$31,$5F,$EB,$63,$FD,$6B,$F4,$FF
	.dc.b	$EF,$DA,$FF,$63,$2F,$A8,$F5,$F7,$A3,$55,$C6,$31,$3A,$9C,$F0,$FF
	.dc.b	$B1,$9E,$57,$D3,$DE,$D5,$6D,$91,$E0,$F7,$AE,$3F,$ED,$F7,$E4,$7B
	.dc.b	$5D,$27,$FD,$99,$C2,$84,$EA,$F6,$DA,$FF,$76,$36,$FF,$DE,$DF,$51
	.dc.b	$F4,$FD,$B3,$7A,$70,$CB,$FF,$7F,$B9,$69,$F8,$63,$FC,$03,$F5,$0F
	.dc.b	$88,$FC,$0E,$1F,$CA,$0B,$19,$9E,$11,$FD,$08,$CC,$F0,$E9,$FF,$74
	.dc.b	$4E,$A5,$5C,$5F,$B0,$75,$8D,$46,$78,$3F,$30,$D8,$F2,$FD,$0F,$67
	.dc.b	$C3,$BB,$E0,$9E,$9F,$E5,$57,$A7,$63,$0F,$E6,$7E,$A3,$F4,$17,$C9
	.dc.b	$F3,$E7,$F8,$BC,$11,$A8,$C3,$AC,$FF,$C0,$72,$F8,$47,$B7,$FA,$55
	.dc.b	$F5,$16,$30,$FE,$A7,$E1,$B3,$97,$A9,$7C,$26,$67,$4F,$EB,$FE,$9D
	.dc.b	$F5,$16,$F0,$D7,$E1,$FB,$4F,$A8,$FD,$B2,$F8,$9B,$18,$7F,$75,$F1
	.dc.b	$4F,$89,$BE,$AE,$1B,$87,$CE,$7E,$5B,$7E,$BB,$F8,$14,$63,$93,$E4
	.dc.b	$28,$FE,$0D,$F1,$BF,$AF,$AF,$AE,$BC,$8C,$F8,$BA,$7F,$86,$7C,$73
	.dc.b	$EA,$6F,$B1,$86,$E1,$F8,$0F,$6B,$F9,$B6,$9F,$5D,$FE,$75,$D5,$F5
	.dc.b	$1C,$B7,$D5,$FE,$31,$F5,$38,$BF,$D8,$BF,$E0,$47,$F1,$6E,$6F,$F9
	.dc.b	$03,$97,$F0,$9E,$D3,$C2,$FE,$45,$F2,$8F,$4B,$F9,$24,$E7,$0F,$CA
	.dc.b	$CE,$5F,$CB,$FC,$AF,$99,$FE,$02,$F7,$F0,$4E,$71,$FE,$60,$E4,$7D
	.dc.b	$17,$F3,$17,$53,$C2,$31,$A7,$F3,$5F,$98,$7D,$4D,$F8,$5C,$37,$2F
	.dc.b	$D4,$CB,$E7,$B3,$18,$6E,$7F,$A3,$9D,$4E,$4E,$3F,$3D,$DE,$BF,$A2
	.dc.b	$7D,$4A,$2C,$61,$B3,$FB,$89,$D3,$67,$4D,$35,$38,$BF,$DA,$53,$E4
	.dc.b	$B1,$C9,$FB,$2F,$A0,$7D,$4F,$F5,$57,$E8,$5F,$EE,$DC,$E3,$FC,$21
	.dc.b	$CD,$71,$AE,$1C,$9C,$5B,$5F,$C3,$BE,$8B,$FA,$EB,$63,$FB,$14,$63
	.dc.b	$0F,$E1,$0F,$66,$FB,$DA,$37,$EC,$DB,$FE,$53,$8B,$EA,$3F,$D3,$D1
	.dc.b	$FC,$73,$E9,$3F,$54,$EA,$33,$87,$F1,$59,$CB,$FD,$86,$B9,$57,$F7
	.dc.b	$86,$BF,$CC,$7E,$A7,$F9,$54,$A7,$2F,$F3,$20,$3E,$26,$FD,$47,$E4
	.dc.b	$23,$FC,$17,$C1,$FE,$A2,$8C,$47,$FB,$22,$77,$D1,$FE,$CA,$BF,$AD
	.dc.b	$8E,$2B,$FD,$97,$55,$F1,$61,$CE,$FF,$D9,$F1,$E1,$18,$76,$C3,$9C
	.dc.b	$62,$77,$7C,$E7,$35,$E4,$AF,$5C,$E6,$FB,$47,$88,$E4,$7B,$30,$F2
	.dc.b	$3C,$0C,$9E,$67,$2F,$96,$BC,$93,$D2,$3E,$11,$C2,$FC,$78,$79,$A3
	.dc.b	$DA,$75,$2E,$18,$70,$7F,$BC,$BE,$A1,$F3,$8F,$A9,$E3,$18,$6C,$66
	.dc.b	$7E,$4C,$E5,$BF,$2F,$F7,$DB,$8F,$28,$C4,$7F,$BF,$A5,$18,$8F,$13
	.dc.b	$C7,$3F,$BF,$93,$E2,$7F,$0F,$5F,$1A,$30,$FD,$DB,$E7,$85,$C3,$95
	.dc.b	$70,$37,$5C,$6F,$9E,$1F,$8C,$C3,$8C,$FD,$26,$C7,$8A,$77,$5C,$DB
	.dc.b	$5F,$9A,$AD,$CF,$AB,$0F,$C8,$3E,$0B,$E5,$5E,$59,$8F,$8C,$DD,$F4
	.dc.b	$57,$D1,$FF,$32,$6F,$FD,$BD,$FD,$C7,$A1,$FA,$1D,$FF,$C7,$B3,$99
	.dc.b	$FD,$C7,$95,$97,$8D,$5F,$01,$5B,$CB,$FC,$1A,$DC,$39,$E1,$BE,$6F
	.dc.b	$E4,$58,$FF,$15,$E2,$C1,$9F,$9D,$B1,$92,$1F,$C5,$9F,$07,$57,$28
	.dc.b	$FE,$4D,$38,$33,$D9,$63,$F9,$44,$57,$F9,$64,$7A,$7E,$A1,$F2,$F4
	.dc.b	$93,$FE,$36,$87,$D3,$E2,$3F,$C5,$D1,$C1,$D4,$F3,$9D,$6D,$FE,$2E
	.dc.b	$FA,$CE,$35,$FE,$40,$6B,$CA,$BE,$57,$84,$F9,$AA,$49,$FF,$23,$FD
	.dc.b	$4F,$0C,$3F,$C7,$34,$FA,$9E,$24,$AF,$EC,$B7,$F5,$7E,$B5,$E6,$52
	.dc.b	$4F,$F6,$AA,$FB,$5E,$5F,$E7,$31,$FF,$22,$B7,$97,$FB,$FF,$56,$BE
	.dc.b	$4F,$F0,$18,$BE,$AA,$93,$2F,$F0,$3F,$C4,$FE,$17,$CF,$7D,$2D,$24
	.dc.b	$FF,$83,$BC,$3F,$C5,$7D,$57,$38,$FF,$32,$B5,$FF,$2E,$37,$6F,$F2
	.dc.b	$CB,$FE,$A5,$8F,$F1,$61,$F9,$3F,$53,$F7,$AF,$F8,$B4,$FC,$B7,$D5
	.dc.b	$7F,$A1,$5F,$F3,$8F,$B1,$FE,$EB,$6F,$CC,$7E,$6E,$31,$1F,$E6,$B5
	.dc.b	$EA,$FE,$7F,$DB,$7C,$B5,$25,$7F,$91,$BE,$5F,$F4,$9E,$FF,$F9,$1E
	.dc.b	$3F,$CA,$5E,$F6,$0C,$FC,$17,$EA,$7E,$09,$74,$95,$FE,$86,$87,$CF
	.dc.b	$FE,$BF,$E1,$37,$3F,$E6,$8A,$C1,$99,$FD,$A7,$27,$FA,$B2,$7F,$D1
	.dc.b	$ED,$7F,$D0,$CD,$E3,$FF,$41,$AF,$A7,$FD,$E7,$C5,$3E,$12,$92,$FF
	.dc.b	$D0,$1D,$7F,$C0,$7E,$33,$F1,$B1,$FE,$9C,$6B,$FE,$93,$6E,$3F,$E9
	.dc.b	$05,$F3,$DF,$C2,$7E,$8B,$ED,$57,$FA,$65,$B5,$FE,$96,$6F,$83,$FE
	.dc.b	$D0,$7F,$D9,$F3,$FE,$AA,$F9,$2F,$E1,$CF,$FB,$FD,$BE,$A7,$FA,$B7
	.dc.b	$D5,$FE,$2D,$1F,$4E,$7F,$D7,$3F,$65,$FE,$C5,$6F,$85,$FC,$75,$46
	.dc.b	$72,$65,$FD,$23,$E9,$BF,$A7,$D7,$FB,$16,$B5,$3F,$D3,$1B,$1F,$E1
	.dc.b	$46,$A9,$E2,$F8,$0E,$57,$E1,$9D,$CE,$AF,$FD,$6F,$F3,$3F,$EB,$4B
	.dc.b	$FF,$64,$AF,$B6,$FE,$64,$D8,$CF,$C9,$60,$CC,$C8,$C9,$FE,$D2,$F9
	.dc.b	$CF,$F2,$8B,$7E,$1B,$F9,$E4,$62,$3F,$D9,$8F,$FA,$1D,$D1,$0C,$C7
	.dc.b	$0F,$87,$FE,$8D,$F3,$CF,$AA,$7C,$96,$3F,$33,$FF,$6E,$E1,$FE,$F2
	.dc.b	$A4,$62,$3F,$C7,$DC,$3F,$E3,$C9,$E5,$5D,$2F,$9D,$71,$BE,$1B,$7D
	.dc.b	$68,$5F,$4F,$99,$E1,$87,$85,$93,$C5,$CB,$FE,$F4,$F0,$DF,$5B,$FD
	.dc.b	$75,$B1,$98,$C6,$79,$64,$64,$E5,$1F,$D9,$7E,$8B,$F9,$0C,$7F,$C0
	.dc.b	$05,$C5,$93,$5E,$04,$9D,$46,$5E,$11,$F9,$5E,$47,$C9,$7F,$96,$4F
	.dc.b	$CA,$16,$7F,$CC,$71,$A7,$FC,$97,$1F,$99,$F8,$0D,$47,$E4,$E2,$CE
	.dc.b	$E3,$FC,$D5,$1D,$1E,$27,$49,$DC,$48,$C9,$CC,$B3,$FD,$D7,$E2,$FC
	.dc.b	$5D,$8F,$A4,$7C,$DF,$9E,$31,$F4,$16,$7F,$8B,$57,$07,$78,$78,$AB
	.dc.b	$95,$EE,$E4,$E6,$B6,$FA,$8F,$C1,$FD,$5B,$E9,$7E,$37,$2A,$31,$3E
	.dc.b	$E8,$C4,$EA,$5F,$59,$62,$56,$E7,$8B,$DE,$BA,$57,$CC,$F8,$0F,$9F
	.dc.b	$F5,$4F,$AC,$A6,$6F,$E3,$CE,$1F,$97,$3B,$72,$43,$D8,$1D,$38,$6E
	.dc.b	$DF,$6F,$7F,$F0,$64,$FB,$7E,$4B,$F5,$D3,$F8,$5C,$BF,$CD,$3D,$3F
	.dc.b	$12,$73,$C5,$FB,$15,$F4,$79,$99,$87,$37,$84,$7C,$CF,$AC,$F0,$DF
	.dc.b	$EE,$7D,$2F,$0D,$FC,$56,$E5,$E1,$AD,$CF,$EE,$1B,$92,$B7,$7C,$F5
	.dc.b	$ED,$B7,$CC,$F4,$BD,$B9,$7E,$67,$7F,$ED,$75,$F2,$A7,$37,$FE,$80
	.dc.b	$9F,$F1,$C6,$7F,$1E,$72,$FF,$B2,$A3,$8B,$9A,$FA,$7F,$11,$DD,$E8
	.dc.b	$48,$FA,$11,$97,$81,$30,$F2,$46,$5F,$35,$F8,$27,$ED,$36,$FC,$B3
	.dc.b	$AB,$EC,$DA,$EE,$63,$0F,$65,$F1,$39,$61,$DE,$33,$7B,$AF,$74,$E5
	.dc.b	$F8,$15,$C6,$7D,$F1,$B3,$CA,$DC,$3C,$B7,$D2,$7D,$CD,$C3,$DF,$1B
	.dc.b	$BE,$8F,$A0,$D9,$E0,$BE,$98,$78,$E3,$B9,$A3,$F3,$CE,$70,$FA,$B7
	.dc.b	$F0,$CE,$F3,$B8,$CE,$1E,$33,$F0,$15,$C7,$0F,$1D,$F0,$39,$4E,$EF
	.dc.b	$C8,$79,$E3,$2E,$1A,$F9,$0F,$76,$1E,$A6,$9E,$3B,$F0,$46,$63,$CD
	.dc.b	$7E,$2B,$F2,$C7,$0B,$EB,$87,$CE,$8F,$19,$FE,$5C,$7F,$2A,$DA,$FD
	.dc.b	$6B,$F4,$CD,$7A,$6F,$82,$FF,$7C,$D7,$F9,$29,$B1,$99,$FF,$33,$9F
	.dc.b	$C2,$DC,$3C,$A3,$F8,$9C,$6E,$FF,$A3,$79,$6F,$3B,$FB,$84,$7E,$C6
	.dc.b	$75,$97,$EB,$DB,$87,$EB,$5A,$AF,$97,$81,$DB,$19,$3F,$C6,$EB,$75
	.dc.b	$A8,$C4,$EB,$0F,$03,$FD,$4A,$75,$18,$9D,$91,$F4,$E3,$E8,$C7,$CD
	.dc.b	$7C,$DB,$5F,$CB,$AF,$9C,$62,$FA,$11,$4C,$9D,$BE,$FA,$E8,$D3,$A5
	.dc.b	$F8,$5C,$5F,$36,$DF,$47,$E3,$4E,$DF,$AD,$5D,$1A,$74,$C7,$19,$7D
	.dc.b	$F8,$F8,$17,$CF,$1F,$2C,$70,$C7,$C0,$D9,$D6,$3D,$EF,$95,$69,$B3
	.dc.b	$AC,$3B,$E5,$F7,$27,$38,$7C,$3C,$39,$4E,$49,$97,$D7,$8C,$BC,$8C
	.dc.b	$5D,$AF,$49,$CD,$C6,$3E,$8C,$BD,$78,$F8,$2B,$D4,$7C,$A3,$81,$C8
	.dc.b	$C9,$23,$8E,$3E,$48,$D4,$63,$0E,$F9,$7B,$E7,$95,$F7,$3C,$15,$B8
	.dc.b	$CD,$FE,$63,$2F,$6C,$7F,$81,$7E,$B3,$FD,$E7,$9F,$F8,$3D,$FF,$57
	.dc.b	$37,$D2,$FC,$27,$12,$73,$5C,$67,$31,$8D,$7F,$BD,$3D,$27,$B7,$A7
	.dc.b	$CB,$DF,$FF,$38,$E3,$FE,$1D,$6C,$FF,$85,$D7,$D6,$6F,$C8,$95,$B7
	.dc.b	$94,$62,$BF,$49,$97,$E4,$F6,$F1,$70,$7E,$36,$31,$96,$5B,$8F,$ED
	.dc.b	$5B,$7C,$A3,$F7,$13,$B6,$9F,$E3,$3F,$AC,$7C,$07,$C4,$6C,$67,$DF
	.dc.b	$C4,$C8,$E0,$EE,$31,$7F,$E4,$3F,$17,$FC,$A5,$F5,$5F,$B1,$7F,$CA
	.dc.b	$9E,$0D,$F1,$13,$63,$0D,$C7,$FC,$9C,$DB,$FE,$0D,$8F,$C3,$3E,$27
	.dc.b	$C3,$DA,$FE,$29,$16,$7F,$86,$4D,$E4,$FE,$57,$37,$7F,$DB,$4D,$A6
	.dc.b	$6F,$2D,$E8,$FF,$A7,$FC,$FF,$F5,$04,$E7,$D2,$C6,$9F,$C5,$23,$85
	.dc.b	$63,$0E,$78,$2F,$15,$AB,$FF,$4B,$46,$27,$BC,$60,$FF,$53,$75,$F0
	.dc.b	$F4,$0D,$EF,$C3,$27,$67,$FA,$B6,$31,$8F,$FA,$F8,$FF,$5B,$FE,$66
	.dc.b	$F3,$BC,$DD,$C8,$C5,$FF,$B1,$FE,$55,$F5,$BA,$9C,$E1,$CE,$D5,$AB
	.dc.b	$FF,$67,$56,$B8,$3F,$D8,$D5,$AC,$39,$63,$E0,$C3,$86,$3C,$F0,$E1
	.dc.b	$8F,$FB,$4E,$73,$8F,$29,$CD,$7D,$6F,$59,$F3,$FF,$31,$B8,$7F,$BB
	.dc.b	$FE,$7F,$13,$23,$FD,$E4,$EE,$31,$5F,$17,$8B,$FD,$8F,$E9,$FD,$6F
	.dc.b	$55,$F0,$36,$30,$DC,$7F,$A9,$E9,$F8,$4A,$DC,$60,$FC,$23,$FE,$B2
	.dc.b	$71,$87,$E3,$A3,$6F,$D5,$1F,$6A,$F9,$8F,$C5,$46,$1B,$E2,$7F,$62
	.dc.b	$BE,$73,$96,$D7,$16,$D7,$09,$E7,$39,$7C,$0D,$C3,$FD,$F9,$1C,$CF
	.dc.b	$25,$F7,$77,$5F,$9A,$F7,$DF,$4F,$4C,$7A,$97,$2F,$7E,$DF,$E1,$9F
	.dc.b	$AD,$FD,$0F,$9A,$7F,$47,$F5,$9D,$75,$FD,$6C,$E6,$F8,$FC,$2B,$E0
	.dc.b	$BF,$52,$D8,$CF,$1E,$27,$F5,$BF,$0C,$FA,$FA,$63,$EB,$6D,$F9,$3E
	.dc.b	$07,$FD,$AC,$DF,$80,$FD,$AF,$D6,$A3,$F6,$F3,$AE,$5F,$C3,$7B,$8F
	.dc.b	$8B,$FD,$DB,$6B,$DE,$BE,$27,$33,$FC,$03,$E2,$FF,$A0,$D7,$C4,$6D
	.dc.b	$FE,$45,$FE,$8F,$C5,$FC,$39,$CB,$FC,$43,$E3,$1F,$69,$FC,$29,$B7
	.dc.b	$F3,$9F,$A6,$D8,$FE,$1F,$F1,$EF,$2E,$98,$F1,$AD,$7D,$6F,$90,$FE
	.dc.b	$B9,$8F,$F1,$E6,$C7,$F1,$D7,$FB,$2C,$67,$DA,$FE,$37,$EE,$FF,$93
	.dc.b	$6B,$EF,$B7,$5F,$D6,$2F,$67,$7E,$1F,$F2,$2F,$AC,$D6,$3F,$CB,$1B
	.dc.b	$87,$F2,$B7,$FA,$8E,$1F,$B5,$72,$FF,$36,$F9,$57,$8F,$FC,$B1,$B5
	.dc.b	$F9,$07,$F9,$5C,$FE,$F5,$CB,$FB,$FC,$DF,$59,$FC,$D1,$B5,$F9,$8F
	.dc.b	$98,$F0,$E5,$FD,$02,$73,$7F,$C1,$E7,$D9,$5B,$D3,$F9,$EC,$3E,$9A
	.dc.b	$C7,$F4,$0F,$9B,$F9,$35,$FD,$1D,$B7,$FD,$43,$E7,$1E,$DE,$FE,$0B
	.dc.b	$FA,$3C,$62,$FF,$56,$DA,$FE,$30,$BE,$B3,$77,$C1,$95,$C3,$EB,$2D
	.dc.b	$FF,$1F,$FA,$CE,$5B,$FF,$5C,$FA,$17,$C7,$7F,$56,$9C,$E1,$FD,$76
	.dc.b	$72,$FE,$F2,$33,$7F,$DA,$1B,$E7,$16,$3C,$F3,$97,$C5,$ED,$18,$BF
	.dc.b	$E6,$5F,$4B,$F5,$B3,$FE,$DE,$DB,$FE,$DC,$FD,$C8,$CF,$37,$F6,$C8
	.dc.b	$D5,$7F,$7A,$6E,$47,$85,$F2,$AF,$EE,$7F,$4D,$8B,$FF,$00,$B6,$CF
	.dc.b	$F7,$9A,$FE,$36,$E4,$9F,$05,$CB,$E0,$B9,$BF,$DF,$BE,$9F,$F6,$B8
	.dc.b	$7F,$4D,$6E,$7F,$E0,$A7,$F6,$F1,$9F,$0B,$81,$31,$FE,$53,$3A,$AE
	.dc.b	$13,$FD,$12,$BF,$D4,$7F,$0D,$FE,$A4,$FA,$9F,$C8,$47,$F5,$4F,$80
	.dc.b	$FE,$34,$A3,$2F,$FA,$7E,$78,$75,$7F,$A8,$70,$FE,$FB,$5A,$AE,$A7
	.dc.b	$19,$D5,$F4,$9D,$4A,$F9,$BE,$1C,$AB,$DC,$E4,$64,$A6,$67,$CB,$87
	.dc.b	$8B,$0F,$F7,$3D,$F2,$8C,$57,$FA,$DC,$F0,$1C,$9C,$BC,$E7,$CF,$7E
	.dc.b	$28,$F5,$DF,$6C,$3D,$55,$DA,$7F,$DE,$A7,$C0,$2E,$18,$70,$7F,$B1
	.dc.b	$FE,$A5,$F4,$6F,$AA,$E3,$18,$6C,$66,$7F,$E0,$0C,$E5,$BF,$30,$D3
	.dc.b	$8B,$63,$31,$82,$3F,$23,$17,$C5,$7F,$B6,$74,$FB,$B1,$89,$FA,$38
	.dc.b	$30,$E9,$5F,$82,$8C,$61,$F5,$AF,$85,$F5,$C3,$A4,$7B,$F0,$E2,$DA
	.dc.b	$FF,$79,$FD,$37,$82,$BF,$DD,$4D,$BC,$D6,$D4,$78,$6F,$9C,$7F,$BF
	.dc.b	$27,$C3,$8F,$18,$FC,$2E,$3E,$66,$D7,$07,$F9,$76,$1E,$67,$93,$54
	.dc.b	$F1,$AF,$D2,$F2,$38,$7F,$BB,$57,$EB,$BC,$FF,$D1,$70,$FF,$9E,$E7
	.dc.b	$33,$FB,$9F,$2B,$2F,$FA,$1A,$33,$B7,$EE,$EF,$38,$75,$6C,$FF,$99
	.dc.b	$AB,$F8,$4F,$16,$B9,$BF,$CE,$51,$83,$F8,$6F,$C0,$66,$64,$64,$FF
	.dc.b	$2B,$4F,$F9,$15,$FF,$16,$F4,$F2,$7F,$49,$8F,$F1,$FA,$FC,$DB,$63
	.dc.b	$F5,$07,$F2,$58,$CC,$F1,$CB,$FD,$3B,$F7,$DF,$D2,$EB,$FC,$72,$D7
	.dc.b	$F9,$AD,$7B,$BE,$A9,$F3,$34,$93,$FE,$40,$87,$D5,$62,$3C,$B1,$C1
	.dc.b	$F1,$E4,$8C,$46,$AB,$D3,$18,$C7,$FA,$B5,$FC,$EF,$07,$AD,$F7,$BF
	.dc.b	$A3,$E2,$3E,$6A,$F4,$CA,$FE,$D3,$15,$F8,$F3,$F9,$47,$C6,$D7,$33
	.dc.b	$F2,$BC,$77,$95,$F7,$9B,$3F,$E5,$E6,$BF,$DD,$27,$F0,$7E,$57,$EB
	.dc.b	$34,$FF,$64,$AF,$3B,$26,$BA,$FF,$13,$8B,$EA,$7E,$CB,$71,$F8,$ED
	.dc.b	$7E,$FB,$51,$FE,$D3,$F3,$9E,$C7,$E5,$1B,$1F,$E2,$58,$C7,$83,$F9
	.dc.b	$8F,$A9,$E9,$A7,$FB,$79,$FF,$12,$D7,$F9,$E9,$AF,$FA,$9D,$DF,$77
	.dc.b	$FB,$87,$D2,$FE,$4D,$CE,$69,$2B,$F2,$7E,$9B,$E4,$FF,$4E,$D8,$FF
	.dc.b	$29,$46,$3D,$2F,$D5,$FA,$CF,$A8,$BD,$92,$FF,$D1,$B3,$97,$FC,$C5
	.dc.b	$EB,$FE,$83,$D8,$FF,$30,$47,$E3,$9A,$FE,$99,$AF,$C8,$3F,$E0,$0F
	.dc.b	$B2,$F9,$2F,$DD,$B6,$3F,$CF,$11,$8F,$93,$FD,$FF,$BB,$99,$FF,$42
	.dc.b	$E7,$83,$35,$FC,$13,$95,$F4,$5F,$B4,$6D,$7E,$89,$AF,$EE,$9A,$F9
	.dc.b	$78,$9F,$06,$F7,$3F,$87,$B6,$7F,$D5,$ED,$7F,$DB,$8F,$E5,$39,$FF
	.dc.b	$C0,$74,$F8,$6C,$73,$FE,$22,$31,$87,$F1,$A9,$E5,$5C,$1D,$E0,$A7
	.dc.b	$C3,$BE,$4F,$F8,$7B,$67,$F7,$AD,$7F,$8A,$B5,$F6,$7C,$6F,$F2,$AC
	.dc.b	$BC,$4D,$8E,$CD,$B3,$F6,$B1,$FF,$6B,$FD,$3F,$EA,$EB,$FD,$C6,$D7
	.dc.b	$FD,$D8,$D8,$FF,$6E,$FD,$17,$F6,$39,$DF,$87,$FD,$F6,$B8,$47,$F8
	.dc.b	$35,$AF,$E1,$5A,$FE,$B9,$E9,$8F,$EC,$3E,$59,$F4,$AF,$A3,$FE,$8F
	.dc.b	$18,$7F,$A4,$DA,$F8,$BF,$2F,$E0,$3F,$DB,$CF,$FB,$A1,$7D,$3F,$F4
	.dc.b	$C6,$C6,$7E,$3B,$13,$23,$83,$B8,$FE,$A7,$F3,$0F,$83,$A4,$AF,$F7
	.dc.b	$AC,$3E,$0F,$11,$FE,$13,$D7,$F0,$9F,$77,$FE,$EB,$8F,$F7,$DB,$5D
	.dc.b	$1F,$83,$AE,$CF,$F1,$2E,$1F,$F2,$8E,$CF,$A8,$7D,$17,$C4,$E1,$7C
	.dc.b	$B5,$3E,$3C,$EB,$0E,$59,$FC,$DA,$DC,$60,$61,$F4,$2F,$94,$66,$3C
	.dc.b	$18,$62,$3C,$F1,$B7,$55,$D1,$CE,$1E,$17,$DB,$CD,$FE,$E9,$FA,$8F
	.dc.b	$EF,$BE,$29,$A6,$33,$B5,$E5,$AF,$F7,$86,$7C,$0C,$C7,$4F,$9E,$FE
	.dc.b	$66,$DE,$4F,$F0,$4C,$7F,$B7,$23,$93,$AD,$BF,$72,$E6,$38,$1A,$9E
	.dc.b	$3D,$FF,$8E,$E9,$FF,$65,$CF,$FA,$1B,$E2,$FF,$8F,$46,$23,$FC,$F1
	.dc.b	$18,$3F,$D0,$86,$BD,$CB,$1F,$D3,$27,$86,$9F,$E7,$16,$C7,$13,$FA
	.dc.b	$8C,$66,$BF,$96,$36,$3C,$47,$3F,$8E,$EA,$AB,$FC,$1D,$F1,$5E,$57
	.dc.b	$CC,$FA,$B6,$7D,$0B,$E5,$FC,$B1,$8F,$1A,$CF,$F2,$AA,$E0,$EF,$0F
	.dc.b	$0D,$6D,$B1,$BC,$FF,$91,$36,$3F,$8F,$3C,$5F,$8E,$7E,$43,$EA,$1F
	.dc.b	$2D,$F1,$BA,$11,$89,$F7,$C6,$27,$52,$FA,$2B,$1C,$7A,$7F,$C8,$D7
	.dc.b	$DA,$FF,$85,$FC,$56,$A3,$E4,$9F,$1F,$E0,$7E,$4F,$D3,$7F,$BE,$23
	.dc.b	$2D,$7C,$E6,$CE,$3C,$3F,$E2,$57,$7F,$29,$C8,$FA,$71,$98,$E3,$3C
	.dc.b	$F8,$3F,$10,$DD,$FF,$B4,$36,$7F,$0D,$87,$C3,$F8,$FF,$D6,$5F,$E4
	.dc.b	$B3,$FF,$49,$75,$E4,$D7,$8F,$70,$FA,$BF,$34,$FA,$CF,$BA,$D8,$F9
	.dc.b	$78,$F4,$D3,$F9,$87,$D4,$F8,$72,$7A,$9E,$9D,$3C,$0D,$AD,$F0,$7D
	.dc.b	$78,$C6,$3F,$C2,$9B,$7C,$B6,$FB,$98,$FB,$69,$BF,$F8,$16,$7E,$7F
	.dc.b	$A7,$F2,$F3,$F8,$AD,$CF,$E2,$FC,$47,$D6,$8C,$61,$FE,$36,$C3,$93
	.dc.b	$F3,$A3,$2F,$69,$E1,$87,$62,$47,$D4,$8C,$BC,$63,$38,$7D,$69,$CB
	.dc.b	$E4,$8C,$E1,$E4,$AA,$78,$EF,$C3,$87,$92,$F9,$4E,$8F,$83,$3A,$37
	.dc.b	$87,$6B,$FC,$1C,$67,$0F,$C9,$CE,$6B,$B4,$73,$C3,$FC,$FF,$87,$96
	.dc.b	$7D,$44,$7C,$2D,$C3,$AD,$71,$BE,$93,$E4,$8C,$3D,$70,$F1,$3F,$1C
	.dc.b	$C4,$69,$E8,$79,$0A,$7D,$13,$66,$30,$F3,$DF,$8A,$31,$5F,$09,$DE
	.dc.b	$1E,$B3,$D2,$59,$E0,$D8,$FC,$16,$1E,$AB,$FA,$15,$A3,$EB,$4F,$6C
	.dc.b	$3D,$17,$C6,$3E,$01,$E5,$BE,$58,$7C,$09,$F0,$DF,$3A,$F7,$47,$1C
	.dc.b	$3D,$71,$F0,$4F,$F3,$11,$D0,$AF,$E5,$9B,$5F,$C1,$5F,$E0,$3E,$B3
	.dc.b	$EA,$7F,$D9,$B1,$FA,$68,$C3,$FE,$D0,$A4,$62,$3A,$B5,$E7,$3F,$C6
	.dc.b	$A3,$8D,$FF,$57,$F4,$3F,$C9,$FE,$06,$6A,$5F,$F1,$6B,$DE,$5F,$E7
	.dc.b	$BB,$DC,$E6,$B7,$87,$F1,$AB,$DC,$FF,$AA,$2B,$73,$9A,$DD,$F9,$2B
	.dc.b	$F9,$5D,$6A,$31,$5B,$9E,$C6,$BD,$26,$3C,$A3,$F8,$06,$3E,$07,$E3
	.dc.b	$63,$FC,$F3,$1E,$91,$A8,$E0,$E8,$E9,$58,$CB,$E5,$CE,$AB,$8D,$6F
	.dc.b	$1F,$21,$32,$F8,$99,$79,$27,$CF,$39,$7C,$07,$63,$33,$EE,$23,$F1
	.dc.b	$E7,$94,$F1,$8F,$04,$E9,$FF,$6F,$4E,$A3,$18,$7A,$27,$A9,$CF,$1F
	.dc.b	$66,$5F,$16,$75,$87,$E0,$70,$E5,$7E,$68,$F0,$4E,$B0,$F6,$E1,$CB
	.dc.b	$2F,$89,$87,$0C,$3C,$A7,$1A,$F3,$65,$F2,$30,$E5,$39,$21,$EA,$C3
	.dc.b	$A3,$E8,$30,$58,$E2,$F6,$33,$8F,$9B,$2F,$61,$F8,$1B,$F0,$38,$CB
	.dc.b	$CF,$87,$6A,$F9,$95,$A9,$C6,$3E,$EC,$3E,$4E,$EF,$A9,$FF,$62,$63
	.dc.b	$FE,$10,$6C,$FF,$82,$D7,$C2,$7C,$AA,$E1,$EB,$D9,$CD,$F1,$8C,$56
	.dc.b	$5E,$1F,$03,$FE,$7A,$9F,$EE,$FE,$36,$3C,$EF,$AB,$E2,$3A,$93,$9A
	.dc.b	$FF,$11,$DF,$1D,$3F,$C3,$CB,$EA,$75,$1F,$AE,$CB,$EC,$70,$7A,$B8
	.dc.b	$7F,$81,$6B,$89,$CB,$71,$FD,$53,$6F,$F6,$ED,$AE,$33,$FA,$FA,$F8
	.dc.b	$5F,$52,$F8,$2F,$89,$F5,$0F,$86,$CC,$61,$B7,$FE,$3C,$AE,$0D,$F8
	.dc.b	$4C,$E7,$FE,$46,$E1,$FF,$2B,$F9,$1F,$13,$CA,$FF,$20,$5F,$F9,$43
	.dc.b	$EA,$71,$39,$6C,$64,$FF,$2A,$C7,$C2,$CF,$E1,$FC,$CD,$9F,$E1,$0D
	.dc.b	$C7,$F7,$87,$F1,$B9,$FD,$F4,$FF,$3C,$E1,$F3,$CE,$83,$88,$D4,$F3
	.dc.b	$E0,$FE,$61,$F1,$77,$A9,$87,$FC,$01,$F4,$FF,$D9,$98,$7F,$A5,$BD
	.dc.b	$BF,$E1,$9E,$AF,$FA,$5B,$DB,$B3,$C3,$85,$7C,$30,$FF,$51,$56,$9E
	.dc.b	$98,$5A,$DC,$DA,$E1,$87,$F1,$99,$7B,$1F,$D3,$31,$FF,$58,$3F,$EF
	.dc.b	$16,$F8,$3C,$E3,$19,$7F,$AF,$1E,$0D,$8D,$78,$9B,$7C,$9E,$0B,$A7
	.dc.b	$F3,$78,$7F,$B3,$5B,$7F,$EC,$CF,$3D,$F0,$1C,$27,$3A,$7F,$B8,$9B
	.dc.b	$7F,$2B,$0E,$9E,$07,$6C,$38,$CF,$58,$C6,$1C,$DB,$19,$3F,$DF,$3F
	.dc.b	$53,$F1,$FD,$D7,$D1,$FC,$9F,$79,$39,$8C,$32,$BF,$E0,$0F,$D4,$DD
	.dc.b	$BF,$DB,$FE,$EB,$EA,$7E,$93,$6B,$FD,$B8,$FF,$AC,$5B,$1A,$F7,$36
	.dc.b	$D8,$C9,$FE,$F7,$A7,$D4,$BE,$6B,$E3,$7A,$FF,$81,$9C,$C6,$2B,$85
	.dc.b	$6D,$FF,$06,$B6,$7F,$C2,$E7,$08,$CC,$FE,$76,$3F,$2F,$3B,$E4,$FD
	.dc.b	$2B,$6B,$FC,$0D,$3E,$18,$C2,$E8,$7C,$33,$EA,$7F,$28,$DB,$FF,$7F
	.dc.b	$2F,$6F,$33,$F9,$9F,$51,$F5,$3E,$3C,$11,$8C,$3D,$31,$89,$D4,$E7
	.dc.b	$1F,$F1,$9C,$BE,$25,$F0,$FF,$A5,$5F,$11,$63,$0F,$EA,$7E,$2B,$F9
	.dc.b	$96,$5F,$B8,$DF,$F5,$11,$8C,$51,$8E,$7E,$1E,$AF,$EC,$7E,$1F,$FD
	.dc.b	$BB,$87,$99,$B7,$E4,$6B,$FC,$10,$F9,$0E,$1F,$E0,$1E,$F3,$E9,$FF
	.dc.b	$78,$FE,$E2,$BC,$86,$A3,$3E,$2F,$F0,$1F,$8D,$FE,$89,$3A,$BF,$A9
	.dc.b	$18,$7D,$31,$FC,$36,$3F,$A6,$67,$FC,$15,$D3,$FC,$1F,$E3,$9F,$47
	.dc.b	$F4,$DB,$5F,$49,$B7,$F4,$1A,$FB,$78,$9F,$21,$E9,$8F,$09,$F8,$75
	.dc.b	$F4,$1F,$C6,$5D,$63,$FC,$5D,$CE,$1E,$BB,$7C,$9F,$F7,$0C,$39,$C6
	.dc.b	$3D,$CF,$E4,$53,$FD,$EB,$0F,$E4,$0E,$DF,$E4,$5E,$F7,$F3,$AC,$3F
	.dc.b	$16,$DC,$3F,$12,$D7,$D4,$D7,$F9,$A7,$D3,$EE,$BF,$96,$2F,$6F,$F9
	.dc.b	$93,$9B,$DC,$7F,$35,$F9,$67,$8B,$4C,$7A,$56,$BD,$DA,$3E,$9F,$F4
	.dc.b	$4D,$8F,$D0,$B7,$3F,$CF,$B5,$FA,$1F,$33,$E3,$9F,$90,$FF,$5A,$D3
	.dc.b	$FA,$13,$99,$F9,$2E,$67,$FA,$87,$CD,$FC,$96,$FC,$17,$F4,$DF,$9C
	.dc.b	$7B,$3F,$D3,$EB,$58,$7F,$01,$9C,$DF,$ED,$9B,$1F,$B4,$6B,$F8,$56
	.dc.b	$BF,$D7,$D7,$CA,$77,$BF,$03,$6B,$F0,$5F,$4F,$D3,$3F,$EB,$9F,$3C
	.dc.b	$F8,$FF,$EC,$53,$9A,$FE,$D5,$39,$78,$E1,$FC,$36,$7C,$39,$7F,$19
	.dc.b	$6F,$39,$A7,$38,$D4,$F6,$FA,$7E,$EF,$F7,$1F,$A2,$FB,$B8,$7E,$5F
	.dc.b	$E8,$B8,$62,$9C,$E1,$F9,$89,$CC,$62,$3F,$50,$D7,$FC,$04,$D7,$F9
	.dc.b	$73,$65,$F4,$6E,$5F,$F0,$13,$70,$FD,$43,$FE,$02,$7F,$C1,$8E,$6F
	.dc.b	$F5,$2D,$DB,$FC,$0C,$D9,$DF,$C0,$F4,$BF,$F0,$77,$D2,$7F,$00,$BF
	.dc.b	$F0,$63,$6B,$FC,$30,$FF,$3F,$C3,$F6,$AE,$5F,$E8,$CD,$C4,$D3,$85
	.dc.b	$74,$F0,$7F,$AF,$57,$F8,$AB,$E1,$3F,$C5,$7F,$4A,$F8,$8F,$E5,$57
	.dc.b	$FE,$0C,$8C,$57,$FB,$D2,$B7,$5A,$F7,$BF,$D7,$55,$AA,$FE,$65,$5A
	.dc.b	$7F,$A0,$4E,$AB,$FB,$3C,$E6,$7F,$A0,$CE,$5E,$C3,$E9,$3A,$61,$FE
	.dc.b	$1E,$7B,$47,$A4,$D5,$F8,$F4,$FF,$22,$CE,$A5,$87,$7B,$F6,$5F,$8A
	.dc.b	$58,$F6,$C7,$C5,$1E,$A3,$33,$D5,$A7,$9A,$78,$63,$EB,$8F,$21,$FE
	.dc.b	$04,$3C,$25,$29,$9C,$35,$F4,$DF,$E7,$FF,$A7,$BF,$20,$69,$2B,$FC
	.dc.b	$1F,$2F,$99,$FE,$D3,$18,$6B,$FD,$8C,$46,$F4,$EF,$3A,$9F,$ED,$13
	.dc.b	$97,$DA,$7E,$0F,$E2,$FF,$B6,$67,$F7,$74,$46,$1E,$37,$E2,$9F,$C1
	.dc.b	$5F,$37,$85,$6E,$FF,$CF,$37,$C9,$FC,$15,$6E,$39,$E1,$F6,$9F,$EF
	.dc.b	$86,$E7,$EF,$9C,$AB,$FC,$01,$B7,$83,$1F,$86,$DB,$FC,$75,$F1,$C7
	.dc.b	$C7,$87,$82,$FD,$13,$E0,$8F,$82,$EC,$F0,$E1,$E9,$8F,$97,$3C,$1F
	.dc.b	$AD,$E4,$3A,$7F,$53,$BF,$FA,$5B,$5F,$ED,$B1,$8A,$DB,$5F,$E7,$39
	.dc.b	$AB,$85,$F3,$9F,$F2,$6D,$F2,$6C,$65,$FB,$D1,$97,$FC,$6F,$CD,$F8
	.dc.b	$57,$F8,$BE,$3E,$1E,$4F,$E3,$AD,$9F,$F6,$74,$E8,$FF,$21,$BC,$5E
	.dc.b	$71,$F9,$B9,$E3,$1F,$E4,$8E,$7F,$F3,$53,$7D,$DD,$46,$23,$FC,$92
	.dc.b	$D7,$F9,$9C,$78,$E3,$B6,$7F,$EA,$B6,$E7,$FC,$DA,$7D,$1F,$49,$FD
	.dc.b	$4E,$FF,$C3,$0D,$7F,$C2,$ED,$BF,$F5,$A7,$63,$E6,$78,$F7,$FA,$6B
	.dc.b	$85,$FF,$8A,$DA,$FF,$B1,$5A,$FF,$9C,$5E,$98,$FF,$60,$F8,$CF,$95
	.dc.b	$E0,$1E,$B5,$FE,$8D,$F9,$CF,$96,$FF,$8D,$03,$8C,$FF,$DC,$AD,$BF
	.dc.b	$F2,$C2,$F5,$7F,$08,$F2,$AF,$DD,$35,$DF,$57,$F0,$B8,$C4,$7F,$B9
	.dc.b	$1F,$B0,$F0,$E1,$D1,$1F,$E2,$91,$F4,$DF,$0E,$FC,$C8,$EB,$B3,$B1
	.dc.b	$1C,$1C,$BC,$2F,$F3,$86,$5E,$66,$C7,$B1,$AF,$95,$BC,$6F,$A5,$FF
	.dc.b	$4A,$CF,$FB,$F2,$BF,$CB,$D1,$88,$FD,$21,$FE,$F9,$71,$1A,$F7,$B9
	.dc.b	$F9,$FF,$C7,$E3,$FC,$CC,$DA,$FF,$26,$9A,$3C,$3E,$83,$E6,$69,$23
	.dc.b	$C4,$0F,$F5,$94,$6D,$C9,$FE,$6F,$F9,$7F,$5F,$A6,$69,$23,$79,$FF
	.dc.b	$9E,$3E,$63,$64,$8C,$D2,$3E,$2F,$AD,$FE,$7A,$9E,$2D,$7E,$19,$FE
	.dc.b	$C5,$71,$9F,$FA,$15,$B1,$FE,$A0,$E8,$F9,$FE,$D1,$A4,$8F,$2C,$7A
	.dc.b	$5B,$97,$02,$3E,$AE,$67,$CB,$7D,$A6,$C6,$63,$07,$DE,$F8,$17,$D7
	.dc.b	$D2,$4F,$C4,$8C,$B7,$E0,$F8,$11,$C1,$F2,$97,$D3,$FE,$29,$B1,$98
	.dc.b	$C1,$F9,$05,$F3,$8F,$A8,$FF,$0A,$F4,$7F,$B1,$E3,$13,$FE,$23,$8C
	.dc.b	$4E,$A3,$FB,$97,$E2,$8F,$90,$C4,$74,$9E,$D7,$CA,$3E,$A9,$93,$F0
	.dc.b	$E5,$32,$EF,$D0,$FF,$40,$C7,$E3,$E1,$F3,$FD,$9B,$1E,$53,$55,$93
	.dc.b	$53,$23,$47,$7A,$EA,$DC,$7E,$04,$EE,$3F,$C4,$B1,$F0,$1E,$37,$E5
	.dc.b	$8D,$C6,$27,$C0,$EA,$F9,$5F,$6B,$F3,$CF,$43,$CE,$6C,$E4,$43,$47
	.dc.b	$94,$D9,$87,$9C,$75,$AF,$0C,$E5,$F0,$DF,$9A,$38,$5E,$1F,$99,$35
	.dc.b	$EC,$7E,$11,$C5,$F6,$AF,$85,$3D,$4F,$35,$F8,$5F,$51,$92,$39,$7C
	.dc.b	$27,$08,$F2,$46,$4F,$8D,$7F,$47,$93,$FC,$6D,$7F,$21,$B8,$7E,$30
	.dc.b	$98,$7D,$4F,$0F,$FA,$E3,$F1,$5B,$87,$5C,$7B,$9F,$9D,$8C,$B7,$C6
	.dc.b	$F2,$57,$C4,$3A,$31,$F5,$C7,$27,$F2,$93,$C9,$B3,$C9,$F9,$47,$1A
	.dc.b	$FC,$14,$EE,$32,$74,$38,$9E,$99,$E5,$8F,$F6,$B7,$95,$F9,$E7,$81
	.dc.b	$FB,$AA,$E9,$5B,$8E,$46,$9C,$5F,$8A,$F9,$CF,$BC,$DC,$65,$B7,$EF
	.dc.b	$7B,$4F,$06,$9D,$6F,$C2,$48,$F4,$B6,$39,$9D,$EB,$B1,$2B,$D8,$FD
	.dc.b	$78,$F8,$6E,$5B,$1E,$3B,$F5,$BC,$09,$1D,$5A,$F9,$A7,$93,$F4,$1B
	.dc.b	$3C,$A3,$A9,$C4,$A7,$19,$E7,$1F,$A3,$8C,$BC,$5D,$1C,$4A,$47,$47
	.dc.b	$17,$15,$F1,$6B,$E0,$46,$2F,$E7,$56,$CE,$35,$BB,$F8,$4D,$AF,$1E
	.dc.b	$1F,$C9,$63,$11,$E1,$8C,$5F,$49,$FC,$3C,$62,$7E,$A9,$B3,$53,$E2
	.dc.b	$8D,$61,$F2,$2B,$8D,$68,$E3,$87,$9E,$3C,$2F,$C0,$AE,$38,$7C,$A9
	.dc.b	$E3,$5C,$1D,$CE,$A3,$E6,$C6,$AF,$C1,$87,$8E,$7F,$68,$7F,$06,$9F
	.dc.b	$E0,$2D,$8F,$F6,$76,$1F,$A3,$8F,$E1,$7F,$31,$FE,$17,$6F,$66,$A3
	.dc.b	$11,$FD,$85,$AF,$F5,$17,$FC,$3D,$87,$F3,$B6,$E5,$FE,$23,$97,$CD
	.dc.b	$B8,$7F,$C6,$B7,$BC,$3F,$A4,$E1,$C2,$FA,$47,$F5,$4B,$E1,$AE,$A2
	.dc.b	$4F,$F3,$FA,$E0,$F6,$C3,$AE,$3E,$7A,$DE,$3D,$EB,$FA,$4E,$3D,$1E
	.dc.b	$53,$AA,$E1,$1A,$BE,$6F,$90,$D9,$2F,$FB,$3D,$F3,$7B,$12,$B9,$BA
	.dc.b	$AE,$75,$FE,$0E,$C3,$D3,$91,$A7,$B3,$0E,$98,$72,$25,$78,$F0,$E9
	.dc.b	$97,$A6,$31,$3A,$7E,$24,$65,$F0,$65,$6F,$AC,$70,$21,$A3,$E1,$4F
	.dc.b	$1C,$7C,$D5,$C7,$2F,$5C,$62,$BE,$26,$1E,$1C,$3A,$C7,$3C,$7E,$23
	.dc.b	$DA,$33,$5C,$67,$33,$BA,$FC,$55,$6B,$2F,$8B,$8F,$87,$2F,$24,$62
	.dc.b	$FC,$E7,$88,$F5,$E3,$D7,$0E,$79,$7A,$2B,$86,$5E,$0C,$BB,$65,$D0
	.dc.b	$FC,$95,$6D,$C6,$3D,$AB,$CF,$8F,$58,$F6,$56,$DF,$11,$EB,$3F,$D0
	.dc.b	$D7,$D9,$CC,$F1,$7E,$B9,$8C,$3C,$B8,$FA,$32,$F0,$E3,$E7,$C7,$B3
	.dc.b	$E2,$AF,$C4,$46,$2B,$F0,$58,$75,$C7,$C1,$7B,$CB,$DB,$8F,$CB,$CB
	.dc.b	$CF,$5A,$C3,$C1,$18,$C7,$A3,$EA,$6B,$E8,$C3,$C9,$7C,$DB,$1C,$63
	.dc.b	$15,$F8,$5A,$E3,$7F,$B3,$9F,$E1,$15,$F0,$9B,$19,$8F,$1C,$62,$7F
	.dc.b	$57,$3F,$8E,$3F,$D1,$31,$EC,$9F,$AD,$5E,$AB,$FD,$BF,$1F,$0C,$25
	.dc.b	$EE,$72,$DA,$E0,$4B,$FB,$37,$C1,$EA,$7C,$39,$FE,$31,$19,$AF,$DA
	.dc.b	$CE,$8F,$E4,$71,$9A,$FC,$EB,$63,$AD,$78,$5A,$71,$C3,$DA,$D9,$E0
	.dc.b	$F4,$9E,$0D,$34,$E2,$78,$DF,$90,$F7,$CE,$88,$E6,$FC,$B5,$C6,$FB
	.dc.b	$DF,$8A,$3E,$AC,$7F,$BC,$1A,$7F,$A6,$A3,$57,$C8,$FE,$63,$3F,$CB
	.dc.b	$3C,$C7,$6B,$E2,$BE,$DC,$FF,$30,$8F,$E9,$1F,$01,$D6,$FF,$DD,$D1
	.dc.b	$8C,$7F,$D7,$D3,$99,$E7,$A2,$31,$4E,$77,$C4,$E1,$B8,$7F,$B3,$63
	.dc.b	$06,$7E,$26,$47,$FB,$45,$B8,$FF,$BB,$2D,$D6,$F9,$CE,$ED,$BF,$F7
	.dc.b	$14,$60,$F1,$F7,$3C,$0A,$49,$FF,$76,$BA,$F0,$64,$7F,$BC,$1B,$8F
	.dc.b	$86,$5F,$31,$EE,$6D,$27,$53,$93,$E1,$F8,$6F,$8B,$A4,$AE,$07,$F7
	.dc.b	$9D,$13,$B8,$CD,$FD,$28,$F5,$46,$23,$ED,$C6,$1F,$8C,$D8,$E8,$D8
	.dc.b	$FB,$7E,$2B,$D6,$FA,$ED,$AE,$96,$8C,$57,$D1,$6C,$F9,$FC,$7B,$3F
	.dc.b	$86,$6B,$F8,$E6,$D1,$A3,$AF,$F0,$ED,$AF,$0B,$6B,$F0,$0D,$8F,$E2
	.dc.b	$BF,$39,$F9,$19,$FF,$0F,$B7,$A8,$D3,$19,$7F,$C0,$09,$FE,$86,$D7
	.dc.b	$FC,$4E,$D9,$F7,$C7,$F2,$2C,$30,$4B,$FE,$4F,$1F,$D3,$23,$83,$23
	.dc.b	$2F,$F8,$C8,$C1,$C3,$EC,$9E,$57,$85,$B3,$C1,$B3,$B9,$D0,$9F,$C7
	.dc.b	$7A,$36,$7F,$21,$5C,$23,$10,$8F,$DA,$4F,$03,$FC,$8C,$ED,$FD,$EF
	.dc.b	$A8,$F6,$FA,$5F,$5D,$9E,$A3,$EA,$B3,$87,$0B,$B9,$7F,$0C,$AD,$4F
	.dc.b	$78,$FE,$1F,$8F,$58,$F9,$38,$23,$13,$FD,$4A,$31,$7F,$5D,$B5,$C2
	.dc.b	$B7,$8F,$95,$B1,$EA,$F8,$3F,$1D,$70,$C1,$7E,$BF,$61,$EB,$E9,$FE
	.dc.b	$17,$87,$9E,$D1,$A8,$C3,$AB,$E7,$1F,$C9,$23,$51,$E9,$6B,$FB,$B6
	.dc.b	$BB,$AF,$67,$B3,$CE,$3F,$80,$35,$FE,$5E,$D7,$F9,$BB,$6C,$F8,$F0
	.dc.b	$FA,$CD,$C3,$8F,$C4,$FC,$28,$C5,$BC,$67,$C5,$3E,$4B,$F4,$91,$86
	.dc.b	$C6,$70,$FC,$FC,$E5,$B1,$9A,$F8,$67,$57,$18,$FE,$4B,$F0,$77,$AD
	.dc.b	$F8,$6C,$39,$5F,$03,$FD,$31,$E6,$7F,$29,$F8,$17,$C0,$FF,$A6,$F1
	.dc.b	$7B,$7F,$9F,$9C,$DF,$F8,$62,$F8,$5E,$EB,$4F,$ED,$A3,$33,$FA,$89
	.dc.b	$CB,$C4,$57,$47,$F0,$F1,$A7,$CA,$7E,$96,$BB,$69,$FC,$FF,$97,$84
	.dc.b	$FF,$41,$C3,$15,$A8,$E0,$77,$C3,$B6,$3E,$CC,$3C,$B3,$DA,$73,$8F
	.dc.b	$86,$3A,$1E,$6B,$E9,$3E,$08,$61,$CB,$0E,$91,$B3,$DC,$7F,$14,$8F
	.dc.b	$E2,$AF,$43,$07,$F1,$CF,$82,$FD,$2E,$7F,$C6,$9B,$19,$E3,$FC,$53
	.dc.b	$67,$5E,$FF,$EB,$A3,$93,$5F,$F2,$6D,$23,$11,$FE,$D7,$8C,$61,$C8
	.dc.b	$F8,$79,$7F,$55,$9B,$DD,$FE,$DD,$CF,$EB,$5F,$09,$F7,$07,$11,$F6
	.dc.b	$EB,$94,$E9,$B1,$CB,$0F,$43,$70,$EF,$50,$EB,$7D,$2F,$D1,$7C,$1F
	.dc.b	$F7,$81,$D0,$B8,$79,$DD,$DF,$DE,$AE,$57,$FE,$07,$9E,$0D,$AE,$0F
	.dc.b	$F8,$2E,$77,$8B,$E0,$5F,$05,$FA,$5D,$FF,$C2,$13,$98,$C5,$FE,$FE
	.dc.b	$B6,$FF,$33,$1E,$9F,$F4,$6C,$FF,$A5,$5F,$19,$FE,$45,$F6,$CF,$81
	.dc.b	$E7,$D4,$8C,$4F,$F3,$DA,$E0,$FF,$02,$9F,$E2,$BB,$F0,$E9,$FC,$04
	.dc.b	$FF,$2D,$6B,$B7,$F9,$9B,$98,$EF,$F6,$FF,$E0,$58,$FE,$68,$D7,$84
	.dc.b	$79,$3B,$3C,$71,$E1,$7A,$E7,$FC,$FE,$75,$9C,$AF,$E7,$71,$C4,$C3
	.dc.b	$E9,$F8,$17,$C2,$7C,$06,$C7,$F4,$56,$BE,$FF,$01,$ED,$EE,$BE,$3F
	.dc.b	$84,$F7,$29,$9B,$E5,$38,$7F,$A8,$4E,$DC,$9F,$5F,$E1,$3E,$2B,$FC
	.dc.b	$9D,$BA,$7F,$7E,$6F,$37,$F9,$46,$51,$FE,$66,$79,$75,$FE,$11,$CE
	.dc.b	$B6,$7E,$3D,$71,$79,$CF,$F9,$6D,$CB,$FE,$5D,$FA,$DF,$F0,$E3,$FC
	.dc.b	$E5,$B8,$78,$EF,$A6,$FD,$7B,$A5,$FA,$E2,$BF,$C6,$39,$FC,$07,$F4
	.dc.b	$22,$BF,$CA,$DF,$24,$E6,$3C,$91,$5F,$7F,$8B,$72,$36,$78,$73,$3E
	.dc.b	$17,$FC,$63,$7D,$4F,$F2,$B1,$AF,$37,$FA,$65,$7F,$A8,$27,$73,$E7
	.dc.b	$36,$71,$AF,$F3,$37,$97,$DB,$CE,$FC,$B7,$87,$FA,$EA,$D9,$C6,$B6
	.dc.b	$F1,$EE,$FE,$A1,$3B,$9F,$DF,$78,$5C,$F0,$F4,$F9,$FF,$DA,$B0,$F0
	.dc.b	$2F,$4F,$26,$BC,$7B,$3E,$5F,$4D,$F2,$3F,$09,$B1,$CA,$B1,$3F,$1B
	.dc.b	$E4,$BF,$C7,$D3,$ED,$6B,$E9,$AE,$2F,$AF,$D6,$78,$3F,$39,$F3,$46
	.dc.b	$71,$C4,$E5,$FA,$7E,$C3,$E1,$E9,$9A,$F9,$91,$97,$E7,$7C,$A3,$E0
	.dc.b	$3E,$DB,$63,$75,$F7,$BD,$EC,$47,$CF,$5F,$01,$93,$57,$67,$ED,$43
	.dc.b	$E2,$3F,$0E,$D8,$FB,$D5,$C9,$DC,$FE,$33,$E5,$FF,$D8,$71,$F8,$36
	.dc.b	$BF,$7E,$B9,$3F,$89,$1F,$08,$F8,$7F,$CD,$36,$3F,$28,$6A,$3F,$41
	.dc.b	$E7,$BE,$25,$F1,$3C,$27,$8F,$C4,$CB,$E3,$87,$37,$C0,$DE,$CF,$CA
	.dc.b	$47,$8A,$78,$3A,$AE,$8F,$28,$CC,$F0,$AE,$5E,$07,$F0,$F6,$D7,$FB
	.dc.b	$85,$57,$09,$D5,$F4,$AF,$1C,$E5,$E1,$19,$BE,$D3,$99,$F8,$47,$B2
	.dc.b	$71,$7D,$A3,$07,$9C,$87,$63,$25,$25,$76,$72,$D7,$37,$ED,$BF,$23
	.dc.b	$63,$AC,$EC,$F8,$97,$E4,$AF,$2B,$67,$75,$F0,$9A,$7C,$79,$DB,$83
	.dc.b	$F0,$BE,$6F,$F4,$C7,$EC,$B6,$FC,$15,$E3,$9C,$F9,$DF,$80,$AD,$4E
	.dc.b	$6B,$75,$F5,$67,$2D,$AF,$06,$3F,$2E,$39,$92,$B8,$BF,$87,$3B,$4C
	.dc.b	$7E,$39,$EB,$31,$1F,$87,$AE,$F1,$C4,$D9,$EE,$9D,$DF,$9C,$FC,$64
	.dc.b	$F0,$BF,$ED,$53,$AA,$EC,$79,$6B,$71,$98,$ED,$7E,$18,$C5,$72,$9F
	.dc.b	$0B,$67,$51,$CC,$EA,$E0,$95,$DA,$38,$DF,$94,$97,$F2,$9B,$1C,$CF
	.dc.b	$1C,$F1,$AE,$71,$F4,$4E,$75,$E6,$AF,$81,$87,$F0,$47,$9D,$70,$AE
	.dc.b	$67,$17,$18,$7A,$F0,$F9,$53,$97,$C0,$47,$DA,$D7,$E9,$1E,$78,$FC
	.dc.b	$14,$F4,$AF,$0D,$74,$AD,$CE,$70,$F8,$07,$A4,$B1,$CE,$BD,$6E,$4E
	.dc.b	$D8,$7B,$67,$15,$F8,$EC,$3D,$55,$E9,$21,$F0,$0F,$0E,$1E,$98,$DC
	.dc.b	$65,$A4,$C3,$C3,$1E,$3B,$FC,$1E,$1E,$18,$C1,$F4,$8B,$87,$8E,$BE
	.dc.b	$0C,$78,$AF,$E2,$E1,$E1,$9C,$CF,$85,$B3,$E2,$BE,$D7,$F1,$EB,$B3
	.dc.b	$4E,$D8,$7F,$8E,$63,$F1,$C5,$AF,$CC,$36,$FF,$82,$39,$F8,$9F,$F3
	.dc.b	$23,$FC,$CD,$AF,$F3,$36,$BF,$D4,$1B,$27,$FC,$DD,$7F,$CF,$5B,$8F
	.dc.b	$F9,$2F,$E0,$7F,$B4,$F9,$BF,$D3,$27,$38,$75,$F0,$30,$7F,$58,$AD
	.dc.b	$C7,$F5,$09,$DB,$E1,$9D,$BF,$D8,$E7,$78,$78,$67,$2D,$9D,$70,$B0
	.dc.b	$CE,$3E,$FB,$B4,$18,$74,$9D,$47,$8E,$76,$EA,$FA,$39,$26,$1E,$13
	.dc.b	$FB,$3D,$F3,$8F,$80,$6A,$FA,$CE,$AB,$9B,$C6,$32,$49,$F3,$57,$37
	.dc.b	$E2,$1C,$CE,$B3,$CF,$0F,$3E,$5F,$5F,$1F,$4D,$F1,$8C,$5F,$13,$9D
	.dc.b	$F1,$71,$1A,$8F,$24,$78,$B1,$F3,$E3,$E5,$8E,$2D,$BE,$D9,$7D,$AC
	.dc.b	$39,$4E,$6F,$8D,$FA,$23,$94,$EA,$3B,$65,$EA,$BE,$47,$E1,$30,$EB
	.dc.b	$96,$4F,$C5,$63,$D2,$FC,$67,$52,$1F,$66,$F8,$3F,$06,$B9,$BF,$02
	.dc.b	$FB,$57,$3C,$3C,$93,$D0,$95,$C3,$2F,$8B,$18,$AE,$17,$E8,$C7,$E8
	.dc.b	$C6,$1F,$0D,$7C,$F3,$8E,$1D,$8E,$F8,$75,$CB,$E2,$E1,$E4,$CB,$8C
	.dc.b	$7C,$B6,$E5,$EE,$AE,$0F,$C0,$BE,$79,$78,$63,$F4,$B5,$E4,$23,$F0
	.dc.b	$E3,$58,$F9,$70,$ED,$39,$AD,$CF,$C0,$C3,$BC,$FC,$D3,$44,$8D,$5F
	.dc.b	$4B,$FE,$0C,$D9,$FB,$E7,$90,$FD,$F7,$2F,$DB,$BF,$D0,$12,$71,$19
	.dc.b	$66,$1F,$82,$3F,$86,$46,$2B,$A6,$1E,$B7,$F5,$71,$98,$C4,$7A,$5F
	.dc.b	$DC,$C6,$70,$F3,$D7,$5A,$F2,$B6,$B8,$E1,$E6,$BE,$15,$DA,$B5,$7F
	.dc.b	$C9,$67,$33,$FC,$41,$B5,$FE,$97,$AD,$CF,$13,$D3,$5C,$63,$31,$C5
	.dc.b	$B3,$C6,$36,$71,$BF,$2D,$F6,$25,$70,$6C,$F0,$3A,$1E,$E3,$FD,$B5
	.dc.b	$C4,$F0,$9F,$13,$F4,$AF,$FD,$47,$F1,$3F,$BC,$97,$33,$E2,$B1,$B2
	.dc.b	$B5,$39,$BF,$F5,$3F,$C5,$6A,$79,$75,$23,$13,$FE,$AC,$F8,$AC,$C6
	.dc.b	$9B,$1F,$EC,$2F,$8A,$F1,$77,$7F,$31,$D1,$DE,$9C,$E9,$E7,$F0,$9E
	.dc.b	$1F,$A6,$B5,$8F,$FB,$DE,$72,$F5,$31,$E3,$71,$AE,$4D,$8F,$F7,$57
	.dc.b	$C5,$7C,$3F,$1B,$FD,$BF,$18,$3F,$DC,$DE,$37,$C7,$F2,$1F,$15,$F2
	.dc.b	$71,$61,$98,$C3,$B8,$C5,$FF,$BF,$1B,$D0,$EA,$76,$7F,$5D,$E1,$79
	.dc.b	$8F,$80,$FA,$DE,$6B,$EC,$3E,$C4,$E7,$3F,$EC,$98,$F2,$CB,$F0,$17
	.dc.b	$B9,$CE,$5D,$7D,$0E,$15,$E1,$8F,$06,$18,$C3,$96,$6B,$C5,$6B,$3F
	.dc.b	$C4,$43,$C4,$B1,$86,$E5,$FE,$07,$F8,$BF,$E0,$9E,$87,$F0,$56,$C7
	.dc.b	$F8,$2E,$B0,$D3,$81,$FE,$0F,$AF,$EF,$3E,$BF,$F7,$AD,$7F,$C0,$35
	.dc.b	$8A,$FF,$0D,$55,$9F,$EE,$33,$5F,$F0,$39,$7E,$65,$EB,$E7,$C7,$FF
	.dc.b	$06,$61,$FE,$0D,$C7,$B5,$6A,$30,$7F,$84,$6D,$F1,$6F,$8C,$E5,$1D
	.dc.b	$E3,$13,$C2,$3F,$C0,$66,$66,$46,$63,$9F,$89,$85,$87,$F8,$63,$85
	.dc.b	$EF,$BE,$1F,$DD,$0F,$7D,$F1,$98,$AD,$65,$E1,$AD,$4E,$6F,$8D,$F0
	.dc.b	$AD,$5F,$0C,$7C,$31,$F9,$39,$D4,$60,$FE,$19,$3B,$C7,$CC,$DA,$FC
	.dc.b	$D2,$F6,$FC,$11,$E3,$F8,$27,$C0,$FA,$9F,$DF,$C6,$63,$1E,$2E,$5B
	.dc.b	$19,$39,$47,$C2,$F8,$4E,$39,$79,$76,$7C,$27,$E7,$B4,$F9,$55,$A7
	.dc.b	$E6,$46,$70,$FD,$84,$E5,$B8,$75,$F8,$CE,$CF,$E3,$7D,$27,$C5,$BE
	.dc.b	$33,$E9,$FC,$43,$E1,$B5,$39,$C7,$E9,$4E,$5F,$C2,$5F,$2B,$D9,$F9
	.dc.b	$7F,$8C,$7C,$1E,$2B,$F0,$DC,$1F,$BE,$1F,$18,$F0,$BF,$29,$C2,$F8
	.dc.b	$8F,$F2,$FC,$E6,$BF,$D2,$1E,$37,$F9,$A6,$31,$1F,$8E,$9D,$DF,$FA
	.dc.b	$56,$3F,$8A,$1D,$9E,$B1,$A8,$E6,$64,$FD,$43,$97,$F4,$D5,$DE,$78
	.dc.b	$E1,$67,$8D,$F8,$B0,$F4,$4E,$67,$89,$D4,$F3,$39,$7E,$14,$F1,$C3
	.dc.b	$C1,$18,$BE,$EF,$13,$07,$9A,$B9,$3E,$E3,$E0,$14,$F8,$58,$61,$FA
	.dc.b	$07,$FA,$E7,$E3,$3F,$33,$AF,$06,$C7,$C9,$6B,$BE,$64,$EF,$87,$F7
	.dc.b	$B3,$B6,$BD,$5A,$FF,$0F,$AF,$C6,$69,$E6,$FB,$7F,$D4,$CF,$F0,$96
	.dc.b	$BF,$B5,$0E,$23,$F9,$45,$3D,$0F,$C3,$E7,$FC,$6E,$B4,$FF,$1D,$97
	.dc.b	$A5,$AC,$3F,$D6,$11,$87,$C1,$9E,$5F,$F6,$77,$D0,$DF,$80,$34,$D5
	.dc.b	$E2,$7F,$90,$4D,$7D,$0D,$CF,$F9,$F3,$71,$FE,$61,$56,$BF,$D8,$53
	.dc.b	$5F,$3F,$85,$D8,$F0,$3F,$CB,$7E,$01,$F0,$3F,$EF,$81,$F0,$5F,$D9
	.dc.b	$27,$FD,$99,$3F,$F0,$01,$B3,$B9,$D4,$7F,$C0,$16,$CF,$46,$CF,$49
	.dc.b	$FF,$80,$3E,$7F,$F4,$7A,$D4,$7F,$6F,$3A,$D7,$2F,$83,$FD,$3D,$6E
	.dc.b	$FC,$15,$C8,$D9,$C4,$5F,$4E,$4F,$F0,$04,$79,$4D,$46,$6B,$A6,$1D
	.dc.b	$E7,$CE,$C7,$77,$E9,$BE,$77,$F0,$E7,$BC,$F9,$27,$38,$7A,$27,$CB
	.dc.b	$1C,$27,$96,$1E,$09,$F5,$1C,$B0,$F0,$57,$28,$B3,$B8,$E5,$3E,$79
	.dc.b	$F2,$BA,$7B,$CF,$CD,$BF,$14,$75,$36,$4A,$EC,$D9,$EC,$47,$F0,$33
	.dc.b	$CB,$0F,$91,$1D,$88,$7C,$DA,$F2,$BC,$CE,$27,$94,$F1,$3A,$3C,$25
	.dc.b	$AF,$04,$69,$F5,$3C,$67,$C2,$E4,$EA,$77,$30,$47,$EB,$1C,$09,$87
	.dc.b	$D2,$8E,$35,$E1,$7D,$73,$BB,$F8,$15,$F7,$6F,$C7,$5D,$EB,$9E,$1F
	.dc.b	$4F,$0E,$53,$E1,$BE,$CF,$C5,$BF,$6D,$F5,$8F,$64,$EA,$31,$1D,$63
	.dc.b	$06,$70,$F1,$E1,$C0,$F0,$5F,$49,$DC,$62,$BD,$71,$87,$F1,$F7,$E8
	.dc.b	$6C,$67,$0F,$A7,$7C,$EF,$E8,$61,$E0,$BF,$24,$75,$7D,$CE,$9F,$21
	.dc.b	$F1,$63,$85,$7C,$4C,$3D,$AD,$C3,$C1,$87,$CC,$8C,$61,$CF,$0E,$E7
	.dc.b	$33,$AC,$73,$C3,$AD,$79,$E7,$D7,$5C,$6B,$9C,$71,$39,$BC,$5E,$73
	.dc.b	$F5,$2B,$9C,$7C,$8A,$F7,$61,$F1,$EF,$84,$F2,$BE,$0F,$23,$F1,$37
	.dc.b	$C9,$F8,$A7,$90,$F1,$4F,$90,$E6,$59,$E7,$1D,$4E,$67,$53,$F3,$93
	.dc.b	$D6,$BC,$93,$F2,$4F,$85,$5D,$1A,$74,$7B,$57,$43,$B1,$D0,$ED,$3F
	.dc.b	$5A,$78,$C6,$E7,$8B,$B3,$8C,$ED,$F9,$07,$BA,$78,$BB,$9E,$33,$B8
	.dc.b	$F8,$B1,$E7,$78,$B5,$CC,$FA,$4E,$73,$C4,$F8,$31,$E5,$8C,$E1,$FA
	.dc.b	$78,$F5,$B5,$F3,$B6,$BC,$6D,$9F,$E6,$93,$FE,$9B,$3F,$94,$16,$BF
	.dc.b	$AB,$B6,$7F,$AE,$39,$CB,$FD,$80,$FF,$91,$E5,$18,$8D,$79,$DF,$E2
	.dc.b	$C9,$CB,$67,$8E,$1F,$D3,$9B,$87,$FA,$B7,$CC,$7B,$4D,$FF,$B7,$D6
	.dc.b	$B2,$FE,$EE,$D7,$FB,$1B,$6B,$8C,$F0,$C6,$57,$F6,$4A,$D5,$F5,$8C
	.dc.b	$4E,$B1,$F2,$3F,$E0,$0C,$39,$CE,$89,$3A,$7F,$BA,$46,$A3,$04,$C7
	.dc.b	$FC,$05,$87,$83,$1E,$D1,$8C,$2E,$01,$87,$39,$D4,$78,$1B,$87,$F8
	.dc.b	$02,$75,$3F,$12,$B9,$BF,$11,$F5,$63,$EF,$AF,$24,$EC,$E8,$78,$6B
	.dc.b	$9B,$C4,$C9,$63,$C1,$3C,$EF,$C6,$49,$E9,$3A,$C3,$C7,$97,$E0,$A7
	.dc.b	$9E,$1D,$A3,$15,$E4,$9E,$D7,$F8,$3C,$3B,$47,$48,$EC,$E8,$F4,$46
	.dc.b	$9C,$12,$BA,$CF,$3C,$3C,$79,$7E,$26,$73,$5C,$A7,$39,$7C,$AC,$BC
	.dc.b	$D7,$B2,$47,$CC,$9D,$D7,$B3,$1F,$16,$1D,$23,$C5,$7F,$81,$7E,$5C
	.dc.b	$64,$D6,$1D,$B0,$E7,$97,$C9,$AD,$DF,$81,$FC,$7E,$1D,$A3,$13,$F2
	.dc.b	$6F,$95,$70,$BF,$33,$4F,$36,$3D,$B2,$F9,$75,$C2,$FE,$96,$5E,$4C
	.dc.b	$BB,$63,$BC,$7E,$31,$CE,$BC,$13,$A8,$C6,$1F,$3D,$A7,$D9,$38,$CF
	.dc.b	$6C,$BD,$33,$B8,$C5,$FD,$1B,$E9,$18,$BF,$83,$97,$8D,$FB,$F9,$7A
	.dc.b	$70,$FC,$2E,$5E,$9A,$D3,$6B,$C6,$70,$C3,$AD,$70,$CB,$EB,$63,$DC
	.dc.b	$F6,$E1,$CA,$33,$3B,$CB,$CD,$39,$7E,$7D,$71,$9F,$9B,$7D,$B0,$F8
	.dc.b	$47,$86,$F8,$3E,$C3,$CB,$8F,$9E,$F9,$47,$0C,$BD,$CF,$29,$D6,$1E
	.dc.b	$38,$FB,$A7,$2C,$BC,$EE,$DF,$9E,$FD,$28,$E1,$7E,$E8,$E1,$97,$AA
	.dc.b	$72,$7A,$B1,$F3,$D7,$E5,$4F,$2D,$7A,$CE,$27,$AD,$CE,$1E,$68,$F1
	.dc.b	$5F,$C8,$77,$3C,$EF,$BC,$7B,$F2,$F9,$0F,$EA,$AB,$C8,$F2,$3D,$F5
	.dc.b	$DB,$1F,$7C,$6C,$CD,$F4,$7C,$D3,$BA,$E7,$1B,$8F,$84,$65,$F5,$4F
	.dc.b	$1C,$3D,$51,$B8,$F8,$8F,$A9,$FC,$D1,$BB,$F1,$CF,$EF,$DF,$A3,$3F
	.dc.b	$CF,$B0,$F6,$CE,$70,$F1,$B6,$7D,$27,$E2,$3F,$28,$7C,$8F,$F5,$9E
	.dc.b	$29,$F6,$CF,$8C,$C6,$DF,$D2,$F0,$61,$C2,$3F,$A7,$5F,$0A,$DD,$7F
	.dc.b	$51,$AD,$29,$FE,$CC,$F2,$C3,$70,$C1,$F3,$7C,$AB,$58,$F6,$3C,$58
	.dc.b	$74,$BE,$6D,$8F,$1D,$F2,$78,$57,$8F,$0E,$9F,$6C,$F4,$7C,$37,$FE
	.dc.b	$05,$BD,$DA,$B5,$1C,$B1,$F3,$3C,$AE,$39,$FC,$F9,$F5,$F8,$C3,$E0
	.dc.b	$4F,$F8,$95,$71,$D7,$E6,$35,$FE,$F0,$D7,$67,$1E,$F7,$63,$C0,$E5
	.dc.b	$18,$6C,$66,$78,$35,$FF,$1B,$4B,$BB,$EA,$47,$56,$BA,$6B,$F8,$18
	.dc.b	$E1,$A7,$A7,$BF,$C7,$3F,$22,$9D,$DF,$87,$AF,$BF,$18,$9F,$9B,$5C
	.dc.b	$1F,$8A,$7F,$85,$DC,$4F,$0A,$E5,$9F,$E4,$6F,$95,$F0,$17,$C2,$7F
	.dc.b	$2D,$D0,$EF,$F9,$8D,$AF,$A0,$D7,$EA,$35,$FF,$2D,$39,$C7,$F2,$F8
	.dc.b	$BB,$FC,$75,$FE,$2D,$8C,$3F,$88,$AE,$18,$78,$B0,$DC,$FC,$4B,$B7
	.dc.b	$C2,$3D,$57,$C1,$FF,$3D,$F5,$FF,$9E,$5A,$7F,$9D,$CA,$7F,$A1,$5F
	.dc.b	$F3,$CF,$7F,$FA,$C3,$BB,$FD,$E1,$3F,$EF,$69,$CB,$FE,$AB,$E1,$FF
	.dc.b	$43,$75,$FF,$A0,$A3,$91,$D1,$FF,$43,$57,$28,$FF,$79,$19,$EF,$FF
	.dc.b	$55,$47,$FB,$4A,$3B,$1C,$5F,$F4,$23,$70,$F2,$1F,$EA,$69,$E2,$FF
	.dc.b	$B3,$F0,$F2,$E1,$FE,$EA,$AE,$07,$9E,$3E,$13,$FE,$9C,$3E,$11,$4F
	.dc.b	$49,$70,$F4,$D7,$78,$FF,$4D,$BE,$27,$FD,$63,$7E,$27,$FD,$60,$78
	.dc.b	$2F,$C4,$F3,$C3,$D6,$DA,$E5,$3F,$EB,$4A,$E4,$FC,$CA,$F9,$16,$9C
	.dc.b	$E1,$EA,$8F,$A0,$79,$B0,$F1,$9E,$63,$E0,$1E,$E7,$75,$E2,$6E,$1E
	.dc.b	$2A,$FF,$5E,$DF,$6B,$E1,$7E,$28,$C4,$7C,$6B,$E2,$F6,$31,$3B,$7C
	.dc.b	$58,$7C,$29,$DB,$67,$53,$FE,$C4,$C3,$C4,$FE,$14,$E3,$87,$FB,$B5
	.dc.b	$F9,$8D,$7E,$51,$E1,$BF,$85,$7F,$2E,$38,$9F,$8C,$36,$F8,$6B,$EE
	.dc.b	$47,$84,$E5,$7F,$14,$F2,$CF,$FB,$5A,$7C,$B3,$F8,$5C,$3E,$1D,$F8
	.dc.b	$30,$F9,$33,$E0,$3F,$3E,$6C,$C4,$7D,$B6,$CF,$2C,$3C,$17,$E4,$C3
	.dc.b	$C1,$5E,$48,$F0,$DF,$E0,$2F,$84,$62,$FF,$55,$5E,$0A,$D3,$D8,$FE
	.dc.b	$87,$87,$FB,$B9,$87,$FA,$F6,$3F,$DD,$AE,$0F,$F5,$71,$FE,$E8,$3F
	.dc.b	$DD,$98,$FF,$BA,$DF,$F6,$09,$9C,$7F,$C2,$B5,$FE,$0D,$ED,$FF,$03
	.dc.b	$D7,$F8,$32,$FF,$C2,$A7,$F8,$26,$B7,$7F,$E2,$76,$BF,$E1,$93,$FC
	.dc.b	$5D,$B7,$F8,$F3,$0F,$F0,$CC,$6D,$FF,$0E,$CF,$F8,$8C,$FF,$1B,$46
	.dc.b	$4E,$27,$F8,$84,$FF,$26,$19,$3F,$C5,$F1,$E7,$7F,$C5,$66,$6B,$CF
	.dc.b	$7E,$93,$FC,$4E,$5C,$3D,$A7,$F9,$FF,$BF,$FC,$C3,$5E,$93,$FC,$55
	.dc.b	$8F,$A5,$F0,$99,$BE,$B5,$C7,$BF,$FC,$5F,$87,$F8,$CB,$0E,$75,$F1
	.dc.b	$F1,$F8,$38,$71,$AF,$1E,$1C,$71,$ED,$3E,$03,$E0,$17,$2F,$8D,$7C
	.dc.b	$B0,$E9,$3F,$1A,$72,$F9,$A3,$2F,$C6,$3E,$26,$1E,$07,$FD,$2F,$8F
	.dc.b	$9A,$3D,$19,$7A,$B2,$FF,$35,$61,$D2,$FC,$58,$FC,$F8,$FA,$F1,$EB
	.dc.b	$9F,$96,$F4,$AE,$4F,$F9,$38,$D1,$F5,$8B,$3C,$B1,$F3,$9F,$32,$39
	.dc.b	$65,$F1,$32,$F3,$61,$E4,$C7,$C4,$7C,$3C,$BE,$E5,$F5,$C3,$9C,$7B
	.dc.b	$F1,$FF,$45,$E1,$CA,$BC,$95,$AC,$3D,$35,$AC,$BD,$2F,$E7,$27,$55
	.dc.b	$D7,$0E,$F3,$E9,$6E,$3E,$3C,$3E,$E5,$7D,$AC,$7C,$17,$CE,$F7,$3F
	.dc.b	$5D,$AF,$CE,$3E,$BD,$79,$2F,$FD,$1D,$7C,$9F,$D4,$9F,$0A,$79,$4F
	.dc.b	$82,$FE,$C5,$6E,$7F,$0D,$5B,$7B,$1F,$49,$F4,$3F,$40,$F4,$63,$F0
	.dc.b	$A3,$BC,$79,$0F,$85,$7D,$E3,$A5,$77,$9F,$C8,$1D,$4F,$29,$FE,$60
	.dc.b	$8F,$F3,$CF,$67,$F9,$C3,$BF,$FC,$FD,$DD,$FE,$79,$AF,$F4,$37,$77
	.dc.b	$FA,$47,$6F,$F3,$9C,$E7,$97,$FD,$0D,$7C,$30,$E5,$5A,$7F,$D1,$B8
	.dc.b	$76,$8F,$F3,$75,$FF,$A4,$A3,$FD,$34,$7F,$A3,$0F,$F3,$A7,$77,$F9
	.dc.b	$F9,$FF,$29,$E1,$FE,$7E,$AF,$F4,$1C,$E6,$3F,$D4,$B1,$FE,$7D,$6B
	.dc.b	$AE,$EF,$F4,$5B,$FE,$99,$AF,$F3,$D9,$FE,$7D,$DF,$FC,$F8,$47,$FC
	.dc.b	$EB,$D9,$FE,$78,$23,$83,$FC,$E9,$7F,$E7,$EE,$AF,$F3,$9C,$07,$5F
	.dc.b	$F9,$D1,$D9,$FE,$8F,$EB,$FF,$32,$3B,$9F,$F3,$4D,$7F,$9B,$63,$FC
	.dc.b	$E7,$D9,$FE,$56,$AD,$C7,$FA,$17,$AB,$FC,$B3,$7C,$33,$FF,$61,$56
	.dc.b	$B9,$BF,$CB,$17,$FE,$5D,$3F,$D5,$37,$FE,$5D,$C7,$FD,$7F,$3A,$37
	.dc.b	$19,$BF,$F6,$63,$6B,$D9,$3B,$EC,$FF,$33,$F6,$49,$FF,$37,$47,$F9
	.dc.b	$B3,$A7,$FC,$2B,$CD,$FE,$0D,$6C,$FF,$82,$5A,$FF,$6D,$EA,$FE,$DF
	.dc.b	$1F,$DE,$E3,$FC,$15,$C3,$FE,$0D,$7F,$C6,$66,$2B,$A5,$7F,$83,$1F
	.dc.b	$F2,$B9,$8E,$BF,$F0,$03,$89,$F0,$4E,$DF,$F1,$09,$89,$DB,$83,$C8
	.dc.b	$FF,$98,$DC,$5F,$96,$FF,$C1,$67,$94,$FF,$13,$47,$94,$E8,$7A,$1D
	.dc.b	$DF,$96,$30,$F9,$5E,$E7,$F9,$80,$FF,$16,$BB,$8E,$A7,$28,$EA,$E2
	.dc.b	$39,$4F,$95,$BD,$9F,$E1,$CA,$FF,$10,$12,$BD,$13,$FE,$2E,$8F,$41
	.dc.b	$E2,$8F,$F3,$45,$FC,$48,$F1,$9E,$09,$E1,$3E,$38,$E1,$1F,$E6,$BC
	.dc.b	$3E,$74,$70,$32,$7F,$98,$AB,$A6,$1F,$82,$3D,$11,$F3,$B0,$F4,$61
	.dc.b	$E6,$8F,$F2,$D1,$E6,$9D,$BD,$5A,$73,$9E,$38,$7C,$93,$CE,$7C,$49
	.dc.b	$FF,$1D,$1C,$E3,$6F,$B4,$EF,$5E,$DA,$EB,$87,$B1,$ED,$3C,$A7,$55
	.dc.b	$CA,$3D,$E7,$F9,$EB,$0F,$7D,$F1,$23,$82,$47,$E1,$0F,$84,$75,$C3
	.dc.b	$D0,$F9,$CC,$1A,$3F,$9D,$1F,$E5,$19,$4F,$F2,$C8,$C3,$56,$1F,$E5
	.dc.b	$67,$27,$F9,$53,$AF,$50,$D3,$FC,$AD,$3F,$E5,$D6,$BF,$E5,$C3,$FB
	.dc.b	$1F,$67,$F5,$57,$FA,$E9,$FD,$57,$3F,$EE,$97,$FD,$56,$31,$97,$F5
	.dc.b	$E9,$FE,$FD,$BF,$F8,$36,$F9,$57,$F5,$8C,$7F,$BF,$D7,$28,$FE,$C8
	.dc.b	$FF,$6C,$9D,$CF,$F8,$5E,$39,$CF,$F6,$8A,$E8,$47,$04,$8E,$51,$96
	.dc.b	$C6,$6F,$FB,$5D,$F6,$C7,$FC,$35,$1F,$02,$F8,$3A,$6D,$7B,$8F,$ED
	.dc.b	$F7,$E4,$3A,$92,$78,$47,$73,$81,$FE,$04,$C3,$E0,$C7,$C7,$3C,$11
	.dc.b	$B3,$FB,$E1,$BB,$FF,$0C,$F6,$7F,$80,$6B,$E2,$57,$0B,$F0,$5F,$6A
	.dc.b	$D5,$FB,$23,$FC,$F9,$18,$CB,$FC,$41,$18,$CB,$E1,$46,$23,$FC,$ED
	.dc.b	$3B,$7F,$C2,$67,$D6,$C3,$DB,$87,$82,$3C,$F3,$B8,$FB,$87,$C5,$9D
	.dc.b	$92,$7C,$D7,$E0,$7A,$CE,$DC,$E5,$F4,$1E,$B3,$B9,$EB,$5F,$04,$E9
	.dc.b	$3A,$AE,$19,$7C,$E8,$FA,$F3,$B7,$A9,$F3,$E3,$4F,$CF,$C7,$F3,$8F
	.dc.b	$33,$59,$7D,$2C,$58,$7F,$B2,$27,$F5,$55,$5F,$F1,$07,$0E,$9F,$F1
	.dc.b	$7F,$5F,$F8,$AE,$BF,$C1,$93,$93,$FC,$5E,$FF,$8B,$4C,$9A,$EC,$FF
	.dc.b	$1B,$52,$7A,$3D,$A7,$55,$D6,$3F,$C5,$B5,$C2,$B7,$D7,$FE,$3D,$8C
	.dc.b	$5F,$F8,$76,$9D,$7F,$E3,$18,$F0,$1F,$E3,$12,$9F,$E3,$13,$CB,$1F
	.dc.b	$E4,$A7,$4D,$72,$F5,$DB,$FC,$64,$FF,$17,$6B,$E1,$E9,$74,$BA,$BF
	.dc.b	$C6,$B3,$FE,$4A,$9C,$C6,$23,$87,$57,$F8,$D9,$FF,$1A,$B6,$7F,$C6
	.dc.b	$D1,$C6,$35,$C1,$FE,$1F,$8F,$F0,$E9,$FE,$2A,$E7,$F8,$5B,$7F,$8A
	.dc.b	$9F,$F1,$14,$7C,$FC,$FF,$C5,$3D,$1A,$7E,$45,$FF,$8D,$3F,$24,$7C
	.dc.b	$06,$6F,$FC,$91,$18,$7E,$1B,$5E,$87,$F9,$37,$F0,$C7,$AB,$C6,$13
	.dc.b	$BC,$FE,$FB,$71,$FB,$0D,$9E,$55,$C6,$31,$0B,$74,$75,$6C,$79,$31
	.dc.b	$FF,$1F,$75,$FF,$5A,$8F,$EB,$07,$F7,$3E,$CF,$EC,$5D,$5F,$DA,$3B
	.dc.b	$3F,$AE,$F2,$FF,$85,$EB,$FA,$C3,$FD,$A1,$FE,$B6,$71,$2C,$FF,$6E
	.dc.b	$7F,$B0,$12,$38,$37,$2F,$EC,$B1,$82,$38,$24,$78,$1A,$F7,$39,$19
	.dc.b	$21,$98,$E6,$E2,$7C,$91,$88,$D4,$F4,$9C,$B7,$0F,$8D,$5F,$06,$B6
	.dc.b	$73,$8C,$CF,$F8,$5A,$31,$7F,$22,$31,$1C,$09,$7D,$9D,$9F,$49,$C3
	.dc.b	$AA,$EA,$FF,$87,$27,$C9,$5D,$63,$07,$2B,$F3,$5F,$22,$5F,$29,$FF
	.dc.b	$04,$BF,$E7,$8C,$3E,$F9,$23,$EC,$C6,$E4,$EE,$33,$7E,$17,$9B,$FE
	.dc.b	$6F,$7D,$0E,$67,$8B,$97,$C6,$E6,$7A,$4F,$DA,$38,$9F,$E2,$19,$FF
	.dc.b	$0F,$3D,$8C,$47,$F8,$99,$FF,$10,$10,$FF,$69,$C7,$F8,$87,$AB,$FC
	.dc.b	$47,$C1,$FE,$20,$8C,$4F,$F8,$C0,$FF,$0B,$47,$F3,$FD,$FF,$A5,$B6
	.dc.b	$3F,$A1,$D7,$F4,$6A,$FE,$93,$D9,$FD,$87,$4F,$EF,$95,$FD,$B7,$0E
	.dc.b	$37,$BC,$39,$69,$FD,$A7,$1E,$3D,$9F,$D7,$71,$E3,$8F,$F7,$28,$C3
	.dc.b	$FD,$68,$E8,$78,$8F,$EE,$87,$F5,$CA,$E9,$7D,$AB,$FB,$2D,$F6,$C7
	.dc.b	$C7,$7C,$0F,$EB,$D7,$C3,$1F,$31,$23,$55,$C5,$F0,$4F,$F7,$77,$8D
	.dc.b	$F6,$C3,$C1,$1E,$38,$C5,$F3,$AE,$17,$CD,$F0,$57,$6B,$E2,$F4,$39
	.dc.b	$10,$C1,$27,$95,$F6,$8E,$57,$DE,$3F,$C0,$11,$BA,$EB,$5C,$23,$A3
	.dc.b	$5D,$57,$17,$E7,$DF,$98,$F4,$D7,$22,$4E,$E7,$EC,$61,$F3,$AB,$73
	.dc.b	$9B,$FC,$7C,$7F,$32,$AF,$C7,$CE,$6C,$FF,$82,$EB,$FC,$7F,$3E,$2C
	.dc.b	$3D,$D7,$D2,$3F,$31,$F9,$1B,$E4,$FF,$C1,$99,$7F,$83,$BE,$51,$F2
	.dc.b	$58,$B0,$BE,$15,$FD,$F3,$99,$7C,$23,$FB,$DC,$EA,$BF,$AC,$9C,$70
	.dc.b	$E9,$6E,$E7,$D8,$F7,$AD,$61,$CA,$F1,$87,$3C,$3F,$C4,$58,$70,$AD
	.dc.b	$1F,$E2,$3B,$FF,$0D,$F8,$3F,$E1,$DA,$DB,$67,$FC,$35,$E0,$FF,$87
	.dc.b	$7C,$8C,$5F,$F8,$74,$6D,$FE,$10,$70,$D5,$DF,$FE,$0E,$7F,$D5,$CD
	.dc.b	$7F,$C2,$5E,$0F,$F8,$55,$CB,$5F,$F0,$76,$5F,$EB,$39,$FF,$08,$C7
	.dc.b	$F8,$5A,$31,$E0,$7F,$84,$BC,$0F,$F0,$6C,$7F,$8E,$BC,$37,$83,$FE
	.dc.b	$01,$E8,$FF,$0B,$61,$F3,$A7,$2F,$F8,$B1,$AF,$D2,$8F,$F0,$97,$77
	.dc.b	$FB,$0E,$3F,$C1,$2D,$F0,$3F,$C2,$11,$F0,$DA,$FF,$82,$A3,$FC,$09
	.dc.b	$D1,$23,$FC,$08,$43,$EF,$F8,$1F,$97,$CB,$FC,$A2,$DE,$2E,$53,$FE
	.dc.b	$40,$EA,$FE,$76,$FF,$59,$EE,$FE,$90,$7F,$3C,$E2,$FE,$DD,$3A,$7F
	.dc.b	$AA,$C6,$BB,$FF,$A2,$36,$BF,$9A,$57,$F4,$86,$F8,$1F,$D5,$5F,$E8
	.dc.b	$D1,$CF,$3F,$ED,$AD,$8F,$EC,$67,$F4,$AC,$3B,$E1,$FD,$3C,$EE,$58
	.dc.b	$EF,$87,$F4,$B9,$FD,$B5,$7F,$88,$67,$D9,$5B,$9C,$D6,$EF,$FC,$D9
	.dc.b	$7B,$9E,$F7,$C0,$8E,$09,$3E,$8A,$F8,$11,$D2,$FF,$81,$61,$EF,$8F
	.dc.b	$93,$3A,$AE,$EF,$86,$FF,$7B,$87,$86,$7A,$61,$E7,$76,$F4,$9F,$EF
	.dc.b	$27,$92,$36,$D3,$E4,$CF,$CB,$7F,$B8,$5F,$95,$FA,$67,$27,$D6,$F0
	.dc.b	$BF,$33,$E4,$7C,$CA,$7C,$CE,$EF,$CC,$ED,$F3,$3F,$7A,$79,$BE,$2B
	.dc.b	$F9,$F8,$7C,$57,$FC,$47,$1E,$93,$47,$33,$CE,$73,$AE,$EE,$8F,$24
	.dc.b	$F3,$7F,$C5,$53,$F1,$30,$F6,$C6,$F0,$FC,$A4,$7E,$1A,$B8,$4F,$E3
	.dc.b	$EB,$81,$1F,$71,$E5,$AE,$31,$F5,$23,$CB,$5F,$7A,$FF,$8D,$47,$A4
	.dc.b	$FF,$0D,$E1,$FE,$1F,$6C,$7F,$87,$3B,$A4,$7F,$88,$4F,$C9,$EF,$FE
	.dc.b	$1F,$3F,$A6,$BF,$AC,$6D,$FF,$BB,$23,$FA,$71,$FD,$92,$3F,$AE,$BF
	.dc.b	$DB,$09,$1F,$D4,$49,$87,$F6,$EC,$7F,$B5,$12,$B8,$CE,$AB,$A9,$FD
	.dc.b	$B8,$E7,$5C,$67,$6D,$BF,$EC,$87,$F6,$38,$F3,$C7,$F8,$9A,$FF,$B5
	.dc.b	$BE,$E9,$FF,$79,$78,$1F,$E3,$3B,$FF,$14,$1F,$DD,$0F,$F0,$CC,$F5
	.dc.b	$7C,$27,$92,$BD,$15,$E0,$8F,$8C,$F4,$F0,$3F,$C6,$70,$BE,$CD,$8C
	.dc.b	$C6,$08,$E0,$9A,$FF,$81,$70,$F1,$65,$F1,$6F,$CC,$FF,$83,$E3,$24
	.dc.b	$8F,$36,$5F,$19,$B8,$F9,$EB,$F8,$BC,$FC,$69,$F8,$39,$7C,$67,$F8
	.dc.b	$2C,$65,$F4,$61,$E3,$79,$9F,$10,$F1,$E1,$FE,$10,$3C,$4E,$0F,$C3
	.dc.b	$61,$E1,$8F,$A5,$1F,$E6,$BC,$BE,$0D,$75,$8E,$07,$E0,$71,$F8,$18
	.dc.b	$F9,$63,$FC,$46,$78,$5C,$1F,$E2,$27,$8B,$F5,$9D,$63,$EB,$8D,$1C
	.dc.b	$1F,$75,$F8,$A7,$F0,$27,$BA,$BC,$33,$C4,$DE,$4C,$3F,$21,$97,$F8
	.dc.b	$A2,$BF,$C7,$1D,$9F,$E1,$99,$FF,$17,$78,$1F,$E1,$CE,$EF,$F1,$1F
	.dc.b	$81,$FE,$29,$CF,$FC,$4F,$7B,$5C,$7F,$E2,$EB,$E1,$87,$F8,$C6,$F8
	.dc.b	$4F,$86,$BF,$C6,$F7,$FE,$31,$7F,$C3,$8D,$7F,$C7,$67,$F8,$B8,$FE
	.dc.b	$0E,$E5,$FF,$1F,$F6,$BC,$6F,$F1,$BF,$80,$F0,$3F,$C4,$F8,$7F,$8E
	.dc.b	$5B,$19,$CB,$FC,$65,$3F,$E3,$07,$FC,$92,$F9,$70,$F0,$EB,$FE,$2B
	.dc.b	$7E,$31,$FE,$13,$7F,$C4,$9D,$FF,$E2,$78,$F3,$F7,$FF,$8B,$5B,$19
	.dc.b	$7F,$CA,$B1,$96,$F7,$FC,$87,$FC,$6A,$D7,$FC,$62,$62,$FF,$93,$AE
	.dc.b	$37,$7F,$F8,$C3,$2F,$8F,$DD,$FE,$2A,$BF,$55,$7F,$8C,$1F,$F1,$A7
	.dc.b	$7E,$F0,$FF,$33,$CF,$F8,$B9,$F8,$37,$F4,$3B,$7F,$C4,$51,$FE,$6C
	.dc.b	$6B,$FE,$32,$3F,$C5,$57,$FE,$28,$EA,$FF,$17,$47,$F9,$AB,$BD,$DF
	.dc.b	$F6,$AB,$E6,$46,$27,$B9,$D6,$B8,$CE,$AB,$86,$1F,$D8,$38,$3F,$AA
	.dc.b	$F6,$7F,$4C,$C3,$FB,$04,$7F,$55,$7F,$BB,$61,$FD,$4D,$FE,$B5,$DD
	.dc.b	$FD,$D2,$BF,$B9,$C6,$3B,$BF,$B8,$DF,$F5,$38,$FE,$B1,$5F,$E3,$BB
	.dc.b	$F0,$1C,$EF,$C1,$3F,$D9,$7B,$FF,$B1,$C7,$F8,$24,$C9,$1C,$D7,$F7
	.dc.b	$C8,$C4,$F9,$8F,$EC,$F7,$EC,$9F,$04,$7F,$74,$33,$7E,$18,$CC,$7F
	.dc.b	$71,$32,$43,$D8,$7F,$6E,$8F,$4B,$9A,$F2,$CF,$C3,$BF,$2E,$1F,$E0
	.dc.b	$69,$F2,$D7,$F8,$24,$F8,$71,$FE,$4A,$69,$FE,$4A,$7E,$4D,$F6,$8F
	.dc.b	$0E,$1F,$4D,$E4,$FB,$67,$93,$63,$91,$E7,$3E,$01,$CA,$7F,$C4,$2F
	.dc.b	$C7,$30,$7F,$90,$AF,$D2,$F9,$27,$FC,$70,$78,$4F,$B8,$77,$33,$3D
	.dc.b	$23,$CA,$FC,$DA,$F2,$B6,$3C,$B3,$CE,$3D,$D7,$F2,$AB,$E0,$D7,$A2
	.dc.b	$30,$FA,$CF,$AC,$7A,$0F,$F1,$89,$EB,$77,$5D,$63,$96,$1F,$7A,$7F
	.dc.b	$18,$D3,$F1,$47,$D9,$8F,$F2,$E9,$FE,$26,$3F,$C5,$71,$FE,$8B,$70
	.dc.b	$7F,$B1,$DF,$E1,$1D,$BF,$E2,$A8,$C6,$FF,$E2,$98,$FF,$16,$DF,$F6
	.dc.b	$E7,$FC,$09,$9F,$F5,$DB,$FE,$A1,$18,$BF,$E9,$E7,$F5,$8B,$FE,$F0
	.dc.b	$D8,$FE,$D0,$FF,$74,$78,$5F,$F8,$06,$3F,$AB,$65,$FD,$C2,$FA,$C7
	.dc.b	$F6,$DE,$EF,$F0,$5B,$DA,$33,$5F,$E3,$B8,$EC,$7F,$66,$BF,$EB,$6F
	.dc.b	$1C,$7D,$18,$FF,$61,$9F,$0E,$3F,$DF,$EB,$C2,$43,$9D,$FF,$68,$9E
	.dc.b	$6F,$7C,$3C,$35,$FE,$35,$BE,$97,$DC,$F3,$38,$AE,$0D,$9E,$17,$D1
	.dc.b	$B5,$FE,$32,$39,$57,$C5,$7C,$71,$F1,$6B,$94,$7F,$88,$1A,$FF,$87
	.dc.b	$8F,$0D,$75,$74,$D7,$55,$D6,$FD,$71,$A3,$C5,$1E,$89,$E1,$DF,$FE
	.dc.b	$0D,$8E,$17,$E0,$8C,$61,$F2,$EF,$CE,$7C,$BC,$3A,$BE,$08,$D4,$62
	.dc.b	$78,$CF,$F8,$B8,$FB,$47,$8A,$7E,$2D,$FE,$4E,$3E,$2C,$72,$BF,$16
	.dc.b	$5F,$86,$AF,$2E,$3E,$88,$C5,$6E,$31,$97,$E1,$1B,$87,$C8,$3C,$07
	.dc.b	$C3,$BF,$01,$F0,$DC,$9E,$B3,$D0,$FE,$1D,$CE,$3E,$C7,$FD,$85,$8F
	.dc.b	$F9,$4F,$B3,$FC,$4D,$C1,$FE,$25,$C7,$FB,$0D,$7F,$87,$63,$FC,$48
	.dc.b	$65,$FF,$16,$F7,$FF,$8C,$3B,$FF,$C5,$71,$FE,$2E,$AF,$F1,$A9,$CA
	.dc.b	$FA,$F7,$FF,$8B,$A3,$FC,$53,$18,$AF,$F1,$0F,$6C,$9F,$F1,$09,$FE
	.dc.b	$20,$ED,$FF,$12,$61,$D4,$FF,$16,$F7,$7F,$8B,$2B,$CB,$5F,$EB,$8D
	.dc.b	$BF,$C4,$94,$7F,$C2,$AF,$F8,$98,$FF,$1C,$8E,$DF,$F1,$03,$7B,$BF
	.dc.b	$C4,$8D,$F0,$FF,$C5,$67,$F8,$7A,$5D,$FF,$E2,$39,$FF,$16,$B5,$FF
	.dc.b	$14,$F0,$6E,$31,$1F,$E2,$5E,$FF,$F0,$F2,$EF,$FF,$12,$67,$FE,$4C
	.dc.b	$E3,$FF,$12,$3F,$E2,$8E,$FD,$C7,$F8,$8E,$BE,$10,$EF,$FF,$10,$9F
	.dc.b	$E2,$2E,$FF,$F0,$D4,$7F,$87,$FB,$FF,$C3,$38,$7F,$8A,$7A,$BF,$C3
	.dc.b	$32,$3F,$C5,$06,$CF,$C0,$DF,$48,$DC,$6B,$5F,$F2,$97,$67,$F4,$2E
	.dc.b	$EF,$EA,$B3,$FD,$EF,$BB,$FA,$24,$7F,$79,$7F,$A2,$47,$F5,$C3,$51
	.dc.b	$C1,$FE,$A5,$3F,$E2,$23,$FA,$8D,$FF,$63,$E4,$FE,$CE,$FF,$61,$7F
	.dc.b	$B3,$CF,$F6,$F3,$FA,$C5,$F8,$1F,$EE,$13,$FE,$FD,$9F,$EB,$51,$FD
	.dc.b	$82,$BA,$BC,$4F,$21,$E2,$EF,$FE,$F3,$3F,$DD,$6B,$91,$E3,$3F,$B7
	.dc.b	$DF,$78,$D5,$7F,$90,$27,$4F,$F6,$8C,$3D,$D8,$78,$23,$FB,$79,$E0
	.dc.b	$70,$7C,$1B,$E9,$5E,$F3,$C3,$3E,$7B,$F8,$6F,$CA,$3F,$C0,$71,$E0
	.dc.b	$7D,$D3,$C9,$D4,$FC,$87,$C2,$F2,$7E,$01,$F2,$AF,$CF,$3C,$63,$27
	.dc.b	$9D,$C5,$FA,$30,$F8,$F8,$79,$EB,$E1,$1E,$77,$E3,$DF,$42,$47,$A0
	.dc.b	$D1,$F1,$CC,$1E,$68,$E2,$E2,$BE,$34,$F2,$C3,$D9,$18,$9D,$5F,$91
	.dc.b	$F8,$37,$F2,$9E,$87,$7A,$E3,$3B,$8F,$7C,$7C,$5A,$F7,$47,$C5,$3A
	.dc.b	$9F,$E1,$E3,$FD,$86,$FF,$87,$5B,$83,$0F,$F5,$2C,$E6,$7F,$C2,$07
	.dc.b	$F8,$93,$6F,$F0,$F3,$7B,$BF,$C3,$F9,$7F,$BF,$31,$FE,$7C,$7F,$40
	.dc.b	$7F,$A8,$99,$2C,$F4,$8D,$C6,$63,$FA,$43,$5F,$E8,$C7,$F4,$29,$FF
	.dc.b	$11,$B4,$FE,$B6,$7F,$4A,$8F,$EF,$D1,$FD,$52,$7F,$AD,$CF,$F4,$CE
	.dc.b	$FF,$EB,$6E,$CF,$ED,$D7,$FD,$8B,$6F,$EB,$91,$B6,$D7,$AE,$3C,$A7
	.dc.b	$A7,$0F,$5B,$E3,$8F,$5B,$93,$66,$63,$78,$FF,$92,$A3,$D2,$64,$F1
	.dc.b	$E3,$E9,$9F,$EE,$11,$E3,$3F,$B7,$57,$8D,$F8,$B5,$FE,$06,$3A,$47
	.dc.b	$8A,$3A,$1A,$C3,$FC,$9B,$3F,$E0,$9C,$7E,$1C,$EA,$31,$8F,$C0,$C3
	.dc.b	$E9,$77,$FF,$73,$9E,$D8,$32,$FF,$21,$E1,$F2,$CF,$F0,$C9,$27,$9B
	.dc.b	$E8,$BE,$78,$F8,$27,$9D,$72,$AF,$15,$FD,$37,$81,$F2,$CF,$54,$7E
	.dc.b	$0A,$7D,$F7,$F3,$2B,$C9,$5F,$32,$7A,$B6,$3E,$09,$D7,$2F,$C2,$9F
	.dc.b	$2F,$1F,$74,$6B,$0F,$A5,$5E,$DB,$F5,$DF,$0C,$3D,$79,$7C,$7A,$F3
	.dc.b	$63,$CB,$0F,$36,$3E,$4C,$3E,$A6,$3E,$8B,$FC,$2B,$5F,$69,$FE,$3A
	.dc.b	$8F,$26,$5F,$1E,$BF,$D0,$BD,$BF,$E1,$3C,$BF,$C3,$1D,$FF,$E1,$D6
	.dc.b	$9F,$D9,$F8,$3F,$C4,$7D,$FF,$E1,$BC,$BF,$B5,$5F,$17,$FC,$4B,$5B
	.dc.b	$3F,$C4,$6E,$0F,$F0,$D5,$7F,$8A,$C4,$7F,$73,$87,$80,$EE,$FF,$0F
	.dc.b	$4F,$F8,$92,$54,$EF,$FF,$12,$BC,$A3,$FC,$4B,$3E,$17,$73,$A8,$ED
	.dc.b	$7D,$5D,$57,$3E,$FF,$F1,$5D,$FF,$89,$B6,$FF,$15,$9F,$E2,$A7,$FC
	.dc.b	$58,$7F,$8A,$8F,$F1,$31,$7B,$7F,$C5,$64,$EF,$FF,$15,$BF,$E2,$8B
	.dc.b	$FF,$12,$12,$78,$77,$7F,$89,$BA,$FF,$C3,$33,$EB,$9F,$F0,$86,$5F
	.dc.b	$16,$72,$FC,$7E,$A7,$87,$FE,$16,$9C,$D7,$F8,$75,$AF,$23,$FC,$41
	.dc.b	$3C,$1F,$F2,$FF,$E2,$AF,$B5,$FF,$0F,$B6,$33,$1F,$E2,$2F,$A1,$7D
	.dc.b	$26,$A2,$4E,$8F,$F3,$4C,$E9,$C1,$FE,$38,$AE,$8D,$9E,$95,$FE,$2E
	.dc.b	$E9,$FF,$16,$4F,$F4,$E3,$FB,$87,$6F,$F5,$FE,$DF,$EB,$53,$FD,$7F
	.dc.b	$8B,$FA,$CD,$7F,$55,$6B,$C0,$E4,$FF,$5B,$EE,$FE,$F2,$4B,$E9,$3F
	.dc.b	$DE,$0F,$EC,$A7,$F8,$29,$CA,$A4,$FF,$8F,$2B,$A5,$6A,$72,$FF,$93
	.dc.b	$63,$FC,$63,$1E,$C7,$CA,$70,$AF,$2D,$FF,$7E,$8C,$B4,$F7,$F7,$FF
	.dc.b	$75,$AF,$EE,$75,$ED,$AD,$BF,$DE,$EB,$6F,$C3,$6B,$B6,$C7,$84,$E3
	.dc.b	$1E,$F3,$C6,$73,$3E,$24,$FC,$F9,$E2,$79,$63,$35,$E5,$C3,$E6,$47
	.dc.b	$23,$A5,$F6,$BF,$85,$7C,$A7,$B1,$F3,$6B,$E6,$47,$58,$C6,$1F,$2E
	.dc.b	$3F,$C7,$D5,$CA,$7C,$27,$BC,$A7,$BC,$B7,$EF,$9F,$04,$79,$1B,$3C
	.dc.b	$E3,$B9,$FE,$A9,$8E,$F7,$FE,$1B,$8F,$F3,$69,$D2,$3C,$D3,$F8,$08
	.dc.b	$E0,$7F,$8F,$0E,$C6,$CF,$7B,$BA,$FC,$2D,$75,$9F,$F5,$AC,$FF,$89
	.dc.b	$DF,$19,$88,$D3,$FE,$30,$EC,$FF,$14,$BC,$1A,$FF,$86,$0F,$EF,$D5
	.dc.b	$FD,$35,$FE,$97,$DB,$FD,$07,$BB,$FA,$59,$FD,$9F,$0F,$EB,$B5,$FD
	.dc.b	$32,$FF,$A7,$63,$FD,$8E,$B5,$1F,$D6,$F0,$EB,$3F,$E0,$CA,$FE,$9B
	.dc.b	$C7,$FD,$66,$7F,$C0,$FD,$FF,$D6,$5F,$EA,$55,$FD,$79,$FE,$E6,$6B
	.dc.b	$2F,$23,$A3,$CF,$5C,$F2,$FE,$B9,$7C,$F0,$FF,$7F,$1E,$0A,$E7,$3A
	.dc.b	$BE,$73,$AC,$3C,$0F,$C3,$AF,$06,$1D,$71,$F7,$C7,$F6,$A6,$BF,$DA
	.dc.b	$A7,$93,$E4,$3B,$C6,$B1,$F2,$5F,$D1,$AF,$21,$DF,$1F,$26,$1D,$F2
	.dc.b	$F9,$67,$D2,$C7,$FC,$18,$DA,$FA,$4E,$A3,$E2,$1E,$83,$C2,$E6,$3C
	.dc.b	$B1,$97,$C1,$8F,$97,$2F,$87,$8F,$94,$F8,$67,$D9,$2D,$7F,$85,$67
	.dc.b	$FC,$39,$19,$8F,$81,$5D,$2F,$C1,$3C,$E7,$2E,$DF,$13,$FE,$71,$C7
	.dc.b	$FC,$3E,$FF,$9C,$FF,$13,$78,$FF,$DC,$29,$18,$CF,$FB,$AF,$E2,$AF
	.dc.b	$1B,$53,$9A,$FF,$01,$D7,$F8,$26,$31,$D3,$FD,$A5,$B1,$FE,$19,$BE
	.dc.b	$4F,$F8,$16,$BB,$E1,$D0,$91,$E0,$C2,$CF,$F8,$67,$0F,$EE,$9E,$0B
	.dc.b	$C3,$FF,$03,$C7,$F8,$4F,$C9,$F5,$4E,$70,$FF,$14,$DF,$17,$FC,$55
	.dc.b	$E1,$2B,$78,$7C,$1A,$FE,$90,$7F,$83,$4A,$5E,$0F,$F0,$91,$9E,$FF
	.dc.b	$F0,$77,$91,$7C,$1F,$F0,$79,$FE,$10,$F0,$7F,$C4,$F7,$FE,$0E,$D7
	.dc.b	$FC,$20,$FF,$8B,$C5,$FF,$85,$5F,$F0,$E7,$81,$FE,$08,$F0,$1C,$9F
	.dc.b	$E3,$5B,$FF,$0A,$46,$2B,$E5,$57,$F8,$38,$E1,$5B,$DF,$FC,$2C,$FF
	.dc.b	$85,$7C,$6F,$F0,$3D,$7F,$8A,$7C,$0F,$F0,$1B,$E3,$6B,$FE,$43,$6A
	.dc.b	$0F,$F8,$2B,$1F,$F1,$37,$82,$F4,$3F,$C2,$C7,$68,$FF,$15,$4E,$F1
	.dc.b	$FF,$27,$56,$EF,$FC,$2F,$7B,$9F,$E7,$7D,$DF,$D4,$A7,$F9,$BB,$5F
	.dc.b	$E7,$67,$F4,$6E,$1F,$E7,$3D,$FF,$D5,$63,$FD,$DE,$7F,$3F,$37,$D7
	.dc.b	$FC,$FF,$5F,$E9,$73,$B9,$FF,$77,$C7,$82,$31,$3F,$DB,$E7,$93,$FD
	.dc.b	$54,$E4,$FF,$4D,$71,$7E,$56,$CF,$2F,$03,$FB,$A3,$FE,$15,$9D,$1F
	.dc.b	$D6,$F0,$FB,$57,$E2,$3D,$F3,$C5,$FE,$D8,$7B,$F0,$F4,$BF,$D8,$8C
	.dc.b	$1E,$3A,$E9,$1F,$E3,$67,$BC,$FB,$1F,$F0,$2D,$7F,$80,$DD,$D7,$C8
	.dc.b	$39,$9F,$E3,$2C,$3D,$6F,$F6,$C3,$11,$C1,$F4,$1F,$39,$EE,$F9,$63
	.dc.b	$E3,$BC,$2B,$C4,$7C,$33,$DC,$FF,$81,$CF,$0C,$7D,$16,$E1,$F2,$9F
	.dc.b	$F0,$E4,$79,$2F,$F0,$31,$DE,$FC,$31,$F3,$27,$FC,$6E,$FF,$84,$67
	.dc.b	$ED,$CF,$A6,$73,$87,$E0,$67,$DD,$1E,$13,$D2,$59,$EB,$3F,$37,$0F
	.dc.b	$4D,$F5,$BF,$39,$F6,$2F,$8E,$1E,$99,$FA,$15,$F0,$AB,$D7,$87,$8C
	.dc.b	$FA,$B8,$7F,$89,$1C,$4F,$C7,$9F,$2D,$F3,$AD,$4F,$3B,$EB,$5F,$0E
	.dc.b	$3C,$B3,$E7,$8E,$F1,$FC,$77,$0F,$F0,$F1,$30,$FF,$0E,$CF,$EF,$E1
	.dc.b	$DD,$FE,$1C,$8C,$35,$FF,$7E,$F7,$7F,$52,$7F,$DF,$93,$FD,$B9,$FE
	.dc.b	$CA,$7F,$6C,$3F,$B4,$B9,$C7,$FA,$D4,$65,$FE,$D3,$19,$C7,$FA,$D3
	.dc.b	$67,$FB,$25,$7F,$BE,$9F,$09,$FE,$22,$32,$53,$35,$FE,$02,$AE,$77
	.dc.b	$FD,$CF,$3F,$F0,$15,$78,$67,$FB,$53,$FD,$A0,$E7,$5C,$5B,$1F,$DD
	.dc.b	$9E,$AD,$C3,$FC,$72,$FF,$8D,$E3,$A1,$F1,$2F,$C0,$FC,$4A,$FF,$06
	.dc.b	$BA,$39,$9F,$E2,$1C,$3E,$63,$83,$FC,$45,$5E,$6B,$F8,$91,$96,$9E
	.dc.b	$27,$91,$D5,$FF,$00,$BE,$A3,$FC,$0E,$53,$E0,$1C,$EF,$CF,$7F,$58
	.dc.b	$F3,$92,$BD,$CD,$9E,$ED,$7C,$E7,$C1,$BF,$5D,$F8,$A3,$B4,$62,$FF
	.dc.b	$C3,$F1,$E3,$9F,$5C,$7F,$97,$EB,$E9,$4F,$03,$D9,$1C,$1C,$47,$87
	.dc.b	$C0,$FF,$21,$C7,$84,$DD,$FE,$26,$FE,$34,$79,$8F,$9D,$3C,$1B,$E0
	.dc.b	$7F,$8E,$1E,$57,$D6,$19,$7F,$48,$C7,$FC,$AF,$E0,$7F,$88,$2D,$C4
	.dc.b	$FC,$5D,$F2,$7F,$E1,$C8,$E3,$3A,$9C,$E1,$FE,$25,$BE,$55,$C6,$F8
	.dc.b	$4F,$F7,$38,$C3,$BF,$07,$9E,$1F,$E0,$DB,$FF,$1B,$36,$32,$FF,$8D
	.dc.b	$DA,$FF,$7D,$F0,$3F,$C5,$0F,$F8,$F0,$F4,$F9,$2F,$33,$FC,$4F,$87
	.dc.b	$84,$F1,$46,$5F,$F1,$7E,$3E,$C1,$19,$AF,$F1,$91,$D2,$BF,$C6,$0B
	.dc.b	$6F,$F1,$6B,$FE,$31,$30,$57,$FC,$5D,$DF,$23,$FC,$54,$7F,$8A,$BB
	.dc.b	$FF,$C6,$31,$FE,$2E,$8F,$F0,$C1,$FE,$2E,$EF,$FF,$19,$BF,$E3,$1E
	.dc.b	$FF,$F1,$83,$63,$2F,$0E,$EF,$F1,$5B,$FE,$57,$EF,$FE,$EF,$A7,$CE
	.dc.b	$9F,$F1,$87,$7F,$F8,$BE,$BF,$C4,$D5,$AD,$BF,$C6,$31,$FE,$36,$9F
	.dc.b	$F0,$F3,$FE,$2B,$EE,$FF,$11,$3F,$3C,$38,$C3,$FC,$53,$E3,$70,$9C
	.dc.b	$F7,$FF,$8B,$9F,$F3,$57,$81,$FE,$2D,$7F,$C6,$CD,$76,$7F,$8B,$8D
	.dc.b	$CF,$9D,$FE,$F1,$DF,$FD,$19,$FF,$08,$9F,$D3,$8F,$EB,$5D,$FF,$D8
	.dc.b	$1F,$F0,$37,$6F,$F5,$C7,$33,$C5,$E6,$71,$3F,$BE,$46,$DB,$3C,$DF
	.dc.b	$F0,$4C,$7F,$7B,$EF,$FF,$0E,$F2,$7F,$79,$AF,$EC,$75,$FD,$E2,$FF
	.dc.b	$AF,$57,$F6,$08,$FE,$BC,$77,$AF,$EE,$57,$DD,$B7,$DE,$3F,$BF,$CF
	.dc.b	$BA,$7F,$C1,$27,$7B,$FE,$DC,$E2,$FF,$B4,$36,$7E,$0D,$F3,$3C,$AF
	.dc.b	$F8,$00,$F6,$D7,$C3,$BF,$64,$7F,$8D,$63,$D2,$FA,$CF,$4D,$75,$7E
	.dc.b	$1D,$7F,$80,$1F,$EE,$E6,$AF,$E2,$9E,$42,$BE,$4A,$F0,$4F,$1A,$FF
	.dc.b	$01,$9E,$D9,$E6,$D3,$35,$CE,$3C,$C7,$72,$C6,$E3,$E8,$36,$3D,$4D
	.dc.b	$7E,$13,$F4,$2B,$C6,$7C,$08,$CC,$FF,$8B,$E7,$E0,$5F,$D4,$9F,$C4
	.dc.b	$C6,$2F,$E6,$CF,$DB,$AE,$07,$C0,$AF,$8F,$7D,$67,$51,$F7,$67,$51
	.dc.b	$D4,$FF,$1D,$D7,$C5,$8F,$F3,$5C,$FF,$AC,$E7,$FC,$45,$5F,$E2,$5A
	.dc.b	$FF,$11,$CA,$7F,$C4,$7D,$FF,$E2,$9E,$DF,$F1,$14,$62,$7F,$89,$BF
	.dc.b	$D7,$4F,$E9,$BD,$FF,$D2,$DC,$4F,$F5,$BC,$7F,$C1,$0D,$AF,$EE,$67
	.dc.b	$F5,$A8,$FE,$DA,$47,$FB,$6B,$5F,$EB,$13,$D2,$FF,$A9,$1F,$D7,$6B
	.dc.b	$A4,$FF,$7C,$9F,$1F,$7F,$F5,$B8,$FE,$BB,$C5,$FE,$22,$D3,$FA,$E3
	.dc.b	$71,$ED,$97,$A7,$1F,$ED,$D8,$F2,$AD,$61,$E4,$21,$E4,$9D,$9E,$29
	.dc.b	$FE,$F7,$97,$9A,$7F,$C6,$D1,$CD,$FF,$12,$61,$E0,$CB,$CD,$3E,$D9
	.dc.b	$FF,$24,$DF,$F8,$52,$BF,$C7,$77,$FE,$09,$3E,$01,$D7,$2F,$F0,$3D
	.dc.b	$F5,$7F,$C3,$53,$D7,$1F,$4E,$1E,$7B,$E2,$F6,$7A,$3E,$6C,$3D,$47
	.dc.b	$85,$FF,$0B,$35,$FB,$07,$C2,$CB,$E4,$4F,$4C,$3F,$CB,$B5,$BC,$7C
	.dc.b	$98,$FA,$B0,$FF,$0A,$9D,$EF,$D5,$1F,$86,$6C,$7E,$1A,$78,$BE,$28
	.dc.b	$F3,$DF,$C0,$BF,$F0,$9F,$67,$F7,$3C,$BF,$C1,$1D,$FF,$E0,$EC,$3F
	.dc.b	$A8,$77,$7F,$83,$A3,$1C,$7F,$E0,$A8,$C7,$7F,$F8,$42,$BF,$C2,$CD
	.dc.b	$9F,$F0,$8C,$72,$C3,$B3,$AC,$3F,$C0,$67,$F8,$1B,$B7,$FC,$13,$1F
	.dc.b	$E0,$DA,$FF,$02,$C6,$1F,$F0,$47,$77,$F8,$1D,$FF,$0D,$F7,$FF,$85
	.dc.b	$1C,$CF,$F8,$23,$5F,$F0,$3D,$FF,$80,$8F,$F0,$39,$5F,$F0,$33,$FD
	.dc.b	$FF,$BF,$FC,$0A,$FF,$82,$3B,$90,$EF,$FF,$04,$48,$FF,$04,$61,$86
	.dc.b	$9F,$E1,$0F,$C2,$1C,$DF,$E0,$1A,$77,$3E,$57,$0D,$BF,$F1,$41,$D6
	.dc.b	$F8,$B8,$0E,$1F,$F1,$95,$74,$AE,$67,$F7,$99,$D6,$1F,$E4,$05,$F3
	.dc.b	$5F,$E2,$96,$DF,$C5,$6B,$E9,$8D,$1E,$62,$3E,$38,$DC,$F5,$9C,$CF
	.dc.b	$93,$0C,$F0,$7F,$8B,$AE,$CF,$F3,$13,$FA,$4F,$47,$F4,$D7,$F9,$57
	.dc.b	$87,$FE,$D9,$7F,$A4,$C6,$7A,$7F,$DD,$DD,$9F,$D2,$27,$FA,$91,$C0
	.dc.b	$FF,$07,$9F,$DF,$E7,$FA,$1B,$D8,$FE,$C4,$64,$F1,$1C,$4F,$EE,$A6
	.dc.b	$4F,$E9,$EF,$F4,$6E,$CF,$E9,$E7,$F7,$D9,$FF,$01,$C6,$5B,$3D,$CF
	.dc.b	$EA,$27,$9A,$BD,$2F,$57,$FC,$02,$F8,$78,$BF,$B2,$39,$BF,$EA,$4F
	.dc.b	$C2,$3D,$31,$FD,$86,$BB,$C7,$1B,$F7,$B4,$F6,$DF,$82,$B8,$47,$F7
	.dc.b	$E9,$F9,$98,$7C,$F8,$FE,$E2,$70,$71,$5F,$DA,$23,$E1,$9E,$D7,$FB
	.dc.b	$46,$1E,$68,$E2,$68,$F1,$16,$3C,$55,$F1,$E3,$C4,$7B,$1C,$5F,$F8
	.dc.b	$42,$30,$77,$AF,$A3,$87,$E0,$67,$FC,$19,$7E,$53,$E5,$9D,$AB,$84
	.dc.b	$FC,$68,$C4,$79,$1D,$E1,$FE,$82,$AF,$F1,$14,$FF,$7F,$BF,$EF,$B5
	.dc.b	$FE,$00,$8F,$EF,$3D,$5F,$DF,$09,$D7,$FC,$E7,$2F,$F7,$84,$7F,$39
	.dc.b	$3F,$A1,$C7,$F3,$08,$FE,$6D,$3F,$D2,$4F,$E7,$78,$7F,$33,$9F,$E7
	.dc.b	$31,$FD,$CE,$B8,$BA,$AE,$35,$FD,$1A,$BA,$4F,$F3,$FA,$DC,$FF,$76
	.dc.b	$8F,$EC,$E7,$F4,$FE,$2F,$EC,$F3,$FD,$C2,$BB,$E1,$FD,$C2,$BC,$97
	.dc.b	$C4,$FE,$C6,$F1,$D7,$FB,$1F,$67,$F5,$88,$FE,$A4,$70,$CB,$FB,$43
	.dc.b	$E0,$3F,$B4,$1A,$AF,$43,$EB,$3D,$18,$78,$B1,$F9,$0D,$3D,$B1,$F0
	.dc.b	$DC,$3C,$4C,$1E,$13,$CC,$78,$6F,$CD,$5F,$36,$34,$79,$1C,$1E,$43
	.dc.b	$E0,$5F,$F6,$F6,$C7,$F6,$93,$C1,$8F,$C7,$8E,$13,$F0,$1F,$8F,$3D
	.dc.b	$63,$EB,$9F,$84,$BF,$5D,$71,$8C,$4F,$FB,$75,$FF,$69,$DF,$F4,$8C
	.dc.b	$7F,$07,$EC,$BE,$6F,$FB,$4F,$2B,$F0,$3D,$63,$FE,$C5,$AD,$5F,$F4
	.dc.b	$AA,$D6,$DF,$DC,$29,$5A,$7F,$A4,$DF,$2E,$56,$1F,$EE,$4A,$D4,$62
	.dc.b	$BF,$B1,$56,$EB,$58,$E6,$7F,$DD,$67,$67,$0E,$A3,$FB,$95,$F8,$63
	.dc.b	$6B,$87,$DB,$6F,$A9,$F4,$4E,$6F,$D9,$39,$7F,$B2,$46,$6F,$D7,$5C
	.dc.b	$2F,$9B,$FE,$03,$8F,$F0,$11,$FE,$02,$3C,$E5,$3D,$3C,$1F,$E0,$47
	.dc.b	$FB,$F0,$E9,$FF,$01,$EB,$98,$FF,$01,$46,$1F,$F0,$47,$8E,$D7,$FC
	.dc.b	$09,$28,$C4,$EA,$7F,$B8,$74,$7F,$68,$7E,$75,$FF,$81,$A3,$FB,$D1
	.dc.b	$27,$EE,$74,$7F,$6A,$9D,$47,$F7,$39,$FF,$00,$78,$BF,$DF,$49,$1F
	.dc.b	$DF,$5A,$FF,$75,$F9,$2F,$F0,$89,$C1,$FE,$F9,$18,$7B,$C6,$5F,$66
	.dc.b	$FF,$D3,$B9,$7F,$95,$93,$7F,$E8,$FD,$1F,$D9,$6B,$F9,$E9,$FD,$8B
	.dc.b	$7F,$E6,$F1,$97,$F9,$F4,$67,$9B,$FA,$9B,$FD,$EA,$3F,$AB,$9A,$BF
	.dc.b	$E9,$FD,$3F,$EB,$BB,$E5,$7F,$ED,$03,$8D,$FF,$7A,$BE,$31,$98,$FE
	.dc.b	$96,$7F,$41,$9F,$EB,$17,$FD,$25,$A7,$48,$FE,$E0,$79,$4E,$31,$EF
	.dc.b	$7C,$D3,$FD,$60,$EC,$EC,$91,$E8,$7F,$C0,$58,$7C,$6C,$3F,$AD,$57
	.dc.b	$A3,$0F,$F0,$09,$E8,$AF,$0D,$F1,$8F,$93,$5C,$27,$C7,$7C,$DF,$F0
	.dc.b	$44,$7C,$53,$4F,$8B,$0F,$ED,$2E,$8F,$19,$EA,$3F,$B3,$9F,$62,$FC
	.dc.b	$18,$79,$4F,$C0,$5F,$7C,$3C,$D5,$FE,$00,$AF,$ED,$6F,$33,$E7,$61
	.dc.b	$E3,$7C,$13,$EA,$6D,$F5,$7F,$C2,$07,$1A,$F9,$CF,$F2,$53,$FC,$E0
	.dc.b	$FF,$73,$7F,$BB,$57,$FB,$CC,$FF,$78,$74,$48,$FE,$EE,$D7,$FD,$F5
	.dc.b	$D3,$FD,$85,$FE,$87,$D3,$FD,$1B,$0F,$E6,$F1,$DA,$75,$8F,$F3,$E8
	.dc.b	$FE,$9A,$FF,$3E,$AE,$11,$FD,$68,$E7,$D3,$FD,$BB,$0F,$F0,$24,$73
	.dc.b	$6B,$FE,$EC,$3F,$AB,$46,$DF,$EC,$D8,$7F,$44,$BF,$EA,$31,$FE,$C8
	.dc.b	$3F,$A6,$1D,$DF,$23,$83,$FB,$1B,$8B,$F2,$C6,$27,$A1,$FD,$5C,$FE
	.dc.b	$BC,$57,$FA,$F4,$F8,$32,$FF,$64,$C7,$86,$B9,$C7,$F5,$98,$F6,$1E
	.dc.b	$82,$CF,$F6,$CA,$EF,$1F,$21,$EE,$D3,$BB,$E1,$7B,$5F,$C3,$9E,$51
	.dc.b	$D2,$7E,$AB,$F2,$63,$E9,$4E,$E7,$34,$C7,$EC,$C7,$8A,$FE,$34,$FA
	.dc.b	$5E,$51,$E4,$73,$3F,$E1,$27,$ED,$EB,$FE,$10,$C3,$C7,$1F,$DA,$BE
	.dc.b	$8B,$FA,$BF,$D5,$7F,$56,$6D,$7F,$54,$6C,$FF,$75,$6F,$2F,$F6,$7E
	.dc.b	$94,$7F,$66,$9F,$F6,$64,$62,$BF,$B3,$CE,$67,$C1,$1F,$DA,$BA,$FF
	.dc.b	$B3,$EB,$FD,$99,$B1,$FD,$91,$B3,$FE,$05,$6E,$1D,$F2,$FF,$72,$CF
	.dc.b	$F6,$26,$BF,$D7,$3B,$BF,$DC,$6F,$FB,$97,$9F,$FB,$03,$FD,$8E,$3F
	.dc.b	$DC,$8B,$EE,$DE,$57,$F6,$8E,$CF,$EC,$91,$89,$E9,$7F,$EE,$98,$C4
	.dc.b	$73,$8C,$57,$F7,$0E,$2F,$DF,$4D,$E7,$FF,$39,$93,$9B,$F8,$04,$7F
	.dc.b	$7A,$9D,$BF,$C1,$A7,$4E,$5A,$FF,$08,$31,$3C,$9C,$10,$F0,$39,$69
	.dc.b	$C0,$86,$CF,$03,$B6,$9A,$38,$9B,$13,$E8,$6C,$F9,$9D,$90,$E6,$71
	.dc.b	$8E,$04,$8E,$11,$A3,$81,$D2,$7C,$8D,$3C,$86,$C8,$64,$ED,$1D,$1C
	.dc.b	$1D,$27,$C1,$3D,$1B,$1D,$0D,$9D,$1C,$BD,$23,$84,$F5,$9C,$CF,$58
	.dc.b	$C1,$27,$64,$8E,$D3,$C1,$ED,$3B,$72,$D7,$C0,$62,$79,$38,$21,$E1
	.dc.b	$72,$D3,$81,$0D,$9E,$07,$6D,$34,$71,$36,$7C,$42,$4F,$A5,$B3,$E7
	.dc.b	$78,$10,$E8,$72,$8E,$04,$8E,$11,$A3,$81,$D2,$7C,$AD,$3C,$A6,$C8
	.dc.b	$64,$ED,$1D,$1C,$1D,$27,$C1,$3D,$1B,$1D,$0D,$9D,$1C,$BD,$23,$84
	.dc.b	$F5,$9C,$CF,$58,$C1,$27,$64,$8E,$D3,$C1,$ED,$3B,$72,$D7,$C0,$62
	.dc.b	$79,$38,$21,$E1,$72,$D3,$81,$0D,$9E,$07,$6D,$34,$71,$36,$7C,$42
	.dc.b	$4F,$A5,$B3,$E7,$78,$10,$E8,$72,$8E,$04,$8F,$95,$C6,$E6,$FF,$10
	.dc.b	$57,$F2,$B8,$C3,$FE,$24,$21,$FE,$24,$8D,$BF,$DF,$0C,$1A,$39,$3A
	.dc.b	$7F,$8C,$C7,$47,$64,$36,$71,$3A,$18,$34,$60,$87,$48,$ED,$5F,$DE
	.dc.b	$DB,$3D,$E7,$91,$0E,$33,$B7,$C0,$73,$2B,$97,$C2,$64,$AE,$5E,$86
	.dc.b	$4A,$E5,$EF,$1E,$36,$CF,$98,$8F,$82,$34,$F3,$36,$66,$36,$F8,$A3
	.dc.b	$73,$DE,$38,$11,$C1,$23,$64,$9E,$73,$E8,$8C,$3D,$C8,$77,$8D,$BE
	.dc.b	$43,$06,$8E,$4E,$9F,$14,$74,$76,$43,$67,$13,$A1,$83,$46,$08,$74
	.dc.b	$8E,$D3,$ED,$6C,$F7,$9E,$44,$38,$CE,$DF,$01,$CC,$AE,$5F,$09,$92
	.dc.b	$B9,$7A,$19,$2B,$97,$BC,$78,$DB,$3E,$62,$3E,$08,$D3,$CC,$D9,$98
	.dc.b	$DB,$E2,$8D,$CF,$78,$E0,$47,$04,$8D,$92,$79,$CF,$A2,$30,$F7,$21
	.dc.b	$DE,$36,$F9,$0C,$1A,$39,$3A,$7C,$51,$D1,$D9,$0D,$9C,$4E,$86,$0D
	.dc.b	$18,$21,$D2,$3B,$4F,$B5,$B3,$DE,$79,$10,$E3,$3B,$7C,$07,$32,$B9
	.dc.b	$7C,$26,$4A,$E5,$E8,$64,$AE,$5E,$F1,$E3,$7F,$8C,$E7,$FE,$60,$C7
	.dc.b	$FC,$C5,$A7,$F9,$9B,$4F,$F2,$FC,$E6,$BF,$CC,$67,$F9,$A6,$B7,$A7
	.dc.b	$F9,$BA,$57,$FE,$72,$AC,$9B,$38,$DF,$48,$ED,$7A,$8F,$F3,$6E,$3E
	.dc.b	$7A,$F5,$1E,$8C,$BE,$59,$1F,$8B,$97,$AD,$D9,$27,$91,$23,$4D,$9E
	.dc.b	$53,$F2,$A7,$24,$8C,$C7,$CA,$8C,$3E,$08,$CB,$E5,$30,$43,$A3,$A7
	.dc.b	$C9,$1C,$9C,$9B,$3B,$46,$E7,$CD,$19,$3B,$1C,$08,$70,$34,$76,$8E
	.dc.b	$53,$DE,$34,$70,$39,$98,$21,$82,$1C,$CE,$B3,$E4,$79,$19,$2B,$97
	.dc.b	$99,$92,$B9,$7A,$B5,$D1,$88,$F4,$36,$7D,$64,$7C,$71,$A7,$89,$27
	.dc.b	$A1,$23,$4D,$9E,$53,$E8,$9C,$92,$33,$1E,$88,$C3,$E0,$8F,$AD,$9F
	.dc.b	$F8,$86,$3F,$C3,$E7,$F8,$84,$FF,$12,$E9,$FE,$2E,$69,$FE,$2E,$36
	.dc.b	$43,$27,$F8,$9E,$3A,$38,$3A,$5F,$F8,$A6,$7A,$36,$3A,$1B,$3A,$39
	.dc.b	$7A,$47,$09,$EB,$39,$9E,$B1,$82,$4E,$C9,$1D,$A7,$83,$DA,$76,$E5
	.dc.b	$AF,$80,$C4,$F2,$70,$43,$C2,$E5,$A7,$02,$1B,$3C,$0E,$DA,$68,$E2
	.dc.b	$6C,$F8,$84,$9F,$4B,$67,$CE,$F0,$21,$D0,$E5,$1C,$09,$1C,$23,$47
	.dc.b	$03,$A4,$F9,$5A,$79,$4D,$90,$C9,$DA,$3A,$38,$3A,$4F,$82,$7A,$36
	.dc.b	$3A,$1B,$3A,$39,$7E,$16,$7F,$CE,$D3,$49,$9F,$F3,$48,$FE,$7E,$0F
	.dc.b	$E7,$65,$32,$60,$C1,$BD,$6B,$86,$BC,$5A,$EC,$DF,$1B,$81,$C3,$E1
	.dc.b	$E1,$B4,$E1,$77,$38,$F3,$5A,$A6,$3C,$F0,$E4,$7A,$F0,$E3,$7C,$23
	.dc.b	$D9,$3B,$A5,$F8,$2F,$75,$CA,$7E,$25,$6E,$B5,$87,$8E,$B5,$5E,$2E
	.dc.b	$4F,$23,$D9,$AF,$26,$BB,$8F,$8F,$69,$F3,$4F,$C1,$A7,$2E,$23,$81
	.dc.b	$E8,$BF,$74,$E9,$F1,$DF,$37,$D5,$8F,$CA,$BE,$53,$9B,$E4,$77,$8E
	.dc.b	$DB,$FF,$A5,$F5,$73,$7F,$AA,$1B,$5F,$87,$E7,$F8,$4A,$32,$FF,$A7
	.dc.b	$5B,$5C,$EB,$91,$F8,$DA,$E3,$3A,$E5,$FF,$61,$B7,$97,$F1,$D5,$C3
	.dc.b	$0F,$07,$07,$E1,$AB,$51,$E1,$9D,$3E,$17,$B4,$ED,$CE,$3E,$39,$FD
	.dc.b	$03,$56,$3E,$49,$DE,$3E,$47,$AC,$66,$78,$CE,$5B,$8F,$A1,$A7,$94
	.dc.b	$F4,$EB,$97,$CC,$6B,$2F,$59,$F2,$AF,$B4,$68,$98,$F9,$E3,$2F,$A4
	.dc.b	$98,$78,$E7,$D7,$97,$C1,$7B,$65,$EC,$CB,$C5,$87,$2B,$E1,$87,$29
	.dc.b	$CE,$1C,$70,$F0,$D6,$8E,$D7,$C6,$BD,$97,$C6,$BE,$01,$E2,$CB,$37
	.dc.b	$C6,$A9,$DA,$2C,$F0,$AE,$59,$7C,$3C,$7C,$51,$8C,$BE,$9C,$63,$1E
	.dc.b	$97,$C5,$F2,$C6,$71,$EB,$87,$CE,$BF,$C2,$E3,$87,$E2,$61,$E8,$CB
	.dc.b	$D8,$F3,$BD,$3F,$68,$C1,$2F,$A3,$96,$BF,$22,$3D,$CE,$EB,$BC,$FB
	.dc.b	$6F,$AC,$EC,$F1,$61,$E9,$78,$1F,$68,$ED,$5C,$8F,$87,$3C,$B2,$F8
	.dc.b	$D3,$A3,$81,$D6,$FC,$0F,$43,$15,$F1,$1C,$4F,$27,$ED,$18,$CB,$E7
	.dc.b	$D7,$23,$DF,$3B,$CB,$E4,$C6,$27,$E9,$E1,$E5,$78,$E1,$DE,$75,$97
	.dc.b	$C2,$BE,$33,$ED,$C7,$D4,$79,$8F,$5C,$69,$F6,$E3,$FA,$5C,$BE,$46
	.dc.b	$1D,$27,$E5,$E5,$E1,$C7,$B5,$6B,$1E,$D5,$EE,$BE,$07,$DA,$C7,$B4
	.dc.b	$78,$4D,$4F,$0C,$7C,$18,$F5,$9F,$99,$5C,$27,$E2,$61,$DE,$7C,$F9
	.dc.b	$7C,$47,$E6,$65,$E4,$C3,$96,$5E,$1B,$E1,$5E,$1C,$7A,$61,$F2,$A3
	.dc.b	$E2,$E3,$F5,$CF,$1B,$D0,$F1,$D7,$28,$F1,$E1,$DB,$2F,$C4,$D6,$F1
	.dc.b	$F0,$4E,$63,$E7,$DF,$1C,$3C,$11,$F0,$8F,$5E,$5E,$97,$96,$1D,$30
	.dc.b	$F9,$6D,$3E,$45,$7C,$18,$CC,$F6,$C3,$C0,$FF,$21,$CB,$DF,$87,$5A
	.dc.b	$DC,$E7,$1F,$24,$E7,$1E,$AF,$EC,$2B,$79,$7C,$46,$E5,$E1,$AD,$E1
	.dc.b	$F8,$63,$D2,$78,$CA,$77,$CB,$CB,$7C,$E7,$8D,$FE,$8D,$A7,$E7,$4B
	.dc.b	$5F,$5A,$FE,$4D,$71,$8F,$92,$7F,$8B,$1F,$F6,$EF,$2F,$F9,$C2,$FF
	.dc.b	$B5,$AE,$5F,$E5,$ED,$AF,$E5,$1D,$1F,$E3,$1C,$7F,$C7,$58,$2B,$58
	.dc.b	$72,$C3,$89,$FC,$E2,$F8,$61,$FD,$92,$B7,$87,$26,$C6,$6B,$83,$C8
	.dc.b	$FE,$71,$1A,$58,$78,$67,$71,$A7,$86,$1E,$38,$F4,$98,$BF,$04,$D8
	.dc.b	$FE,$EF,$87,$99,$EB,$87,$7C,$3B,$47,$53,$99,$70,$EC,$4C,$3E,$2D
	.dc.b	$F3,$AE,$33,$9A,$F7,$1C,$9C,$3E,$0C,$3C,$8F,$BA,$B9,$61,$E3,$24
	.dc.b	$FC,$0A,$E3,$7D,$B0,$F6,$E1,$D5,$B3,$AB,$F6,$61,$D3,$0E,$F8,$71
	.dc.b	$C3,$E1,$98,$AE,$EF,$33,$D3,$7C,$EB,$C5,$87,$62,$4E,$9F,$86,$F3
	.dc.b	$8E,$F8,$78,$63,$E5,$4F,$07,$E0,$1B,$32,$79,$0A,$7C,$72,$E0,$9E
	.dc.b	$97,$E5,$9F,$41,$B7,$07,$F9,$2A,$37,$5E,$1A,$F8,$55,$E2,$3A,$E1
	.dc.b	$F0,$AF,$84,$79,$5F,$A3,$7C,$A3,$E0,$10,$E5,$1E,$0B,$ED,$3B,$9F
	.dc.b	$24,$61,$E8,$D7,$C3,$87,$C8,$72,$75,$C3,$E0,$38,$39,$9F,$1E,$FC
	.dc.b	$0E,$EF,$C1,$3E,$E2,$38,$C3,$DD,$87,$D8,$AF,$09,$F3,$A7,$58,$78
	.dc.b	$E3,$38,$7C,$27,$C5,$87,$46,$E1,$F9,$06,$D7,$E2,$1F,$54,$78,$1F
	.dc.b	$46,$1F,$4E,$31,$1F,$2A,$31,$3E,$1B,$F6,$CF,$C0,$34,$74,$8D,$4F
	.dc.b	$08,$F5,$1E,$23,$F0,$B7,$F0,$EF,$BB,$F3,$23,$DB,$3F,$4E,$FB,$CF
	.dc.b	$B6,$72,$7C,$D3,$A4,$7E,$19,$C9,$E5,$8E,$91,$CA,$BE,$24,$EA,$7A
	.dc.b	$4E,$6F,$E3,$4F,$1A,$F2,$E1,$F4,$A3,$D6,$71,$8E,$75,$C6,$3D,$B1
	.dc.b	$EC,$33,$87,$C3,$9F,$B1,$3C,$6F,$BE,$1F,$0D,$F9,$31,$F2,$CE,$F3
	.dc.b	$D4,$F7,$16,$BF,$10,$D8,$FB,$67,$17,$B9,$97,$EB,$9E,$D9,$EE,$D3
	.dc.b	$BC,$F9,$CF,$A4,$57,$E8,$1E,$93,$DF,$1F,$92,$C3,$E3,$57,$82,$95
	.dc.b	$3B,$DD,$78,$8F,$F6,$CF,$13,$B5,$D9,$9E,$7F,$F7,$E4,$62,$BF,$DE
	.dc.b	$F8,$72,$6E,$1B,$E0,$EF,$8F,$3C,$71,$48,$C3,$C0,$87,$8C,$F3,$55
	.dc.b	$7C,$D7,$DD,$C3,$5E,$0B,$B7,$BD,$7A,$DB,$8E,$6B,$D3,$1E,$45,$59
	.dc.b	$7E,$1C,$DC,$BE,$0D,$6E,$B1,$97,$C9,$EB,$77,$2F,$D1,$8A,$73,$18
	.dc.b	$6C,$7B,$5A,$FC,$15,$D9,$F1,$E3,$4D,$74,$D7,$C4,$D7,$E5,$6F,$E2
	.dc.b	$7F,$02,$D4,$D3,$CF,$D3,$FD,$2F,$1F,$C3,$E0,$7E,$DF,$53,$BF,$18
	.dc.b	$7B,$E7,$F1,$57,$C6,$3F,$17,$5B,$9E,$66,$B8,$F8,$E1,$CB,$81,$5B
	.dc.b	$BE,$87,$E6,$30,$E9,$5B,$9D,$3E,$29,$4B,$B7,$11,$E4,$6B,$CD,$AF
	.dc.b	$06,$C7,$1C,$7C,$F3,$F9,$7B,$75,$7A,$27,$A1,$AC,$7D,$31,$9A,$ED
	.dc.b	$39,$AE,$2F,$48,$F0,$39,$9E,$11,$FE,$93,$7F,$CE,$2F,$FA,$3A,$3F
	.dc.b	$CC,$BD,$3F,$E6,$D8,$FF,$2E,$EB,$FE,$8C,$7F,$CD,$7D,$1F,$E6,$17
	.dc.b	$FD,$58,$D3,$FD,$35,$3C,$63,$2D,$8C,$F5,$7F,$9F,$7A,$3F,$CD,$B5
	.dc.b	$FE,$93,$37,$19,$7C,$31,$FE,$91,$69,$FE,$7D,$BF,$F4,$1C,$EE,$38
	.dc.b	$9F,$E8,$38,$CB,$67,$BF,$5F,$FA,$52,$3E,$01,$EC,$C3,$CC,$7F,$9F
	.dc.b	$8B,$7F,$E9,$C9,$E1,$7F,$0C,$EA,$7F,$A2,$6B,$CA,$7F,$A3,$CE,$93
	.dc.b	$E5,$C3,$E0,$C7,$8C,$DB,$F2,$6F,$C1,$3B,$6C,$EC,$F6,$D7,$76,$E1
	.dc.b	$FE,$97,$79,$CF,$A1,$FF,$50,$E1,$E8,$9F,$95,$1C,$0F,$F4,$E5,$FB
	.dc.b	$72,$FC,$1E,$1E,$DA,$F8,$F1,$87,$8E,$3E,$A9,$F1,$36,$3C,$57,$D4
	.dc.b	$FB,$91,$97,$B9,$E4,$AE,$8F,$DB,$9E,$F9,$7E,$4B,$0F,$84,$79,$5C
	.dc.b	$4F,$C3,$BE,$8F,$6C,$BE,$3E,$3E,$46,$E1,$E4,$AD,$E5,$F0,$EF,$C0
	.dc.b	$7C,$8C,$7B,$BE,$5C,$7B,$E3,$CE,$7C,$53,$F7,$0F,$16,$3D,$9F,$6C
	.dc.b	$FF,$B7,$AF,$B4,$7E,$0E,$3F,$0F,$8F,$8E,$B8,$4F,$A6,$FA,$D6,$A3
	.dc.b	$9D,$7F,$B1,$2B,$FD,$F3,$7D,$31,$FC,$46,$1C,$2B,$E4,$DF,$97,$0F
	.dc.b	$43,$EB,$31,$1E,$19,$E1,$1F,$04,$F5,$B8,$7E,$D5,$F7,$C3,$F1,$97
	.dc.b	$ED,$CB,$F8,$4D,$7F,$6B,$C7,$FD,$C3,$D5,$FD,$51,$FF,$5E,$C6,$7A
	.dc.b	$FF,$DC,$0D,$9D,$47,$F8,$A5,$FF,$12,$B6,$3F,$C1,$E7,$F8,$E7,$8B
	.dc.b	$FC,$1F,$A7,$F8,$5A,$3F,$C4,$0D,$7F,$C3,$98,$FF,$84,$DC,$9F,$E1
	.dc.b	$A7,$FC,$49,$87,$F9,$1A,$7F,$C3,$D3,$FE,$49,$AE,$AE,$E3,$33,$FE
	.dc.b	$48,$6D,$FF,$8F,$A7,$A9,$C4,$FF,$2A,$C7,$F8,$D8,$DE,$1E,$C3,$FC
	.dc.b	$51,$1A,$3F,$C7,$27,$A2,$3D,$55,$FE,$45,$8F,$F1,$A9,$D9,$F1,$36
	.dc.b	$BC,$4F,$13,$D0,$66,$BC,$0F,$C9,$3C,$07,$F9,$12,$7B,$61,$F1,$E3
	.dc.b	$91,$EB,$3F,$C8,$0F,$F9,$4A,$FD,$71,$F5,$23,$FC,$EF,$87,$C8,$8E
	.dc.b	$11,$F3,$A3,$15,$E8,$9E,$AD,$3C,$35,$F0,$30,$FC,$05,$F7,$AF,$51
	.dc.b	$DE,$7E,$9D,$F4,$AF,$83,$7C,$AB,$E0,$9D,$A3,$27,$4A,$FA,$91,$D2
	.dc.b	$FB,$3B,$C3,$E3,$C7,$B2,$3E,$3D,$FE,$1D,$E1,$5C,$9E,$07,$94,$FA
	.dc.b	$D8,$79,$6F,$9C,$FA,$27,$9C,$E5,$E0,$7C,$47,$17,$E4,$BF,$64,$FC
	.dc.b	$CA,$FA,$33,$9C,$3D,$93,$F8,$AA,$D9,$F3,$63,$31,$F7,$DB,$1E,$B8
	.dc.b	$F4,$CF,$3C,$3F,$2B,$1F,$80,$BF,$75,$74,$BF,$97,$5A,$BF,$BD,$7E
	.dc.b	$C2,$1C,$4F,$24,$F8,$30,$F6,$E1,$D2,$31,$3C,$2B,$9B,$C4,$F4,$5F
	.dc.b	$7A,$F8,$27,$17,$11,$FE,$8F,$E8,$FF,$38,$91,$FF,$38,$90,$FF,$37
	.dc.b	$B9,$7F,$C8,$27,$F9,$D1,$FF,$38,$1F,$E5,$BE,$8F,$F3,$B4,$7F,$A0
	.dc.b	$23,$06,$73,$FF,$3F,$F8,$12,$7F,$D2,$F1,$FE,$79,$30,$FF,$9F,$A3
	.dc.b	$0F,$F8,$D7,$A3,$FC,$E6,$FF,$9F,$7A,$7F,$CF,$8F,$F9,$DD,$AF,$F9
	.dc.b	$DD,$AF,$F9,$CB,$93,$FC,$FD,$D3,$FE,$73,$8F,$F4,$13,$6B,$FC,$FB
	.dc.b	$3F,$E7,$19,$FF,$3B,$6B,$5F,$F4,$4B,$FE,$8A,$F0,$9E,$67,$F9,$37
	.dc.b	$A9,$E2,$3C,$4F,$F2,$4D,$6B,$A7,$FC,$B3,$5A,$9C,$DF,$1B,$E1,$5F
	.dc.b	$E5,$3B,$E1,$8F,$F9,$C3,$0E,$13,$FE,$57,$F1,$DE,$37,$87,$C8,$79
	.dc.b	$1F,$E5,$89,$CD,$78,$27,$2F,$1A,$E1,$5B,$6C,$64,$E1,$19,$69,$FE
	.dc.b	$67,$EA,$75,$FF,$6E,$E5,$FF,$0A,$E7,$FD,$E7,$A7,$FB,$6F,$0F,$F7
	.dc.b	$58,$FE,$F0,$7F,$6D,$71,$D9,$FD,$9F,$B7,$FC,$17,$1F,$D9,$F2,$FF
	.dc.b	$05,$DF,$F7,$1B,$E0,$F4,$23,$FD,$CE,$7F,$BC,$9C,$5F,$EF,$27,$F8
	.dc.b	$10,$FF,$27,$DF,$77,$FB,$C9,$D0,$E5,$3D,$E7,$91,$E6,$C3,$DB,$97
	.dc.b	$F8,$2A,$B8,$1F,$E0,$08,$CE,$5F,$E0,$0B,$FF,$09,$65,$FE,$BE,$7F
	.dc.b	$C8,$71,$FE,$1A,$C7,$C5,$8F,$A2,$77,$87,$F9,$BA,$3C,$19,$7C,$28
	.dc.b	$F5,$1F,$E9,$37,$0F,$CC,$8F,$3C,$FC,$48,$F2,$4E,$AB,$BC,$FB,$4E
	.dc.b	$F5,$FE,$59,$AF,$82,$76,$CB,$EE,$E3,$E9,$8F,$F2,$09,$D0,$F9,$A7
	.dc.b	$86,$B9,$9E,$E8,$FF,$12,$4F,$F9,$2B,$2F,$8B,$1F,$7A,$77,$5D,$2B
	.dc.b	$58,$F9,$EF,$73,$F5,$B0,$F6,$BE,$63,$C0,$7D,$C8,$DC,$7C,$B8,$C4
	.dc.b	$F3,$23,$E8,$C7,$E1,$BE,$FC,$BF,$35,$3A,$AF,$A3,$3B,$BE,$B8,$FC
	.dc.b	$98,$EA,$66,$B8,$47,$D1,$C7,$D7,$39,$25,$F4,$24,$EF,$0F,$C2,$61
	.dc.b	$DA,$BC,$99,$7D,$17,$F5,$87,$C1,$CB,$FD,$68,$FF,$92,$E9,$3F,$E4
	.dc.b	$B6,$C0,$D4,$7F,$92,$BA,$E4,$64,$99,$7F,$94,$9B,$D9,$FE,$52,$7F
	.dc.b	$A5,$9F,$D5,$A7,$FA,$5B,$FD,$93,$2F,$E9,$CF,$F8,$A4,$FE,$A9,$5F
	.dc.b	$DA,$4F,$EC,$93,$FD,$FA,$3F,$C0,$27,$F5,$B9,$EA,$FF,$5A,$39,$67
	.dc.b	$FD,$78,$91,$AA,$E9,$7F,$D9,$2B,$BF,$6F,$F6,$87,$CC,$72,$AE,$11
	.dc.b	$98,$ED,$7F,$E1,$97,$89,$E3,$3F,$C3,$27,$27,$47,$89,$C9,$CA,$7F
	.dc.b	$C3,$0F,$22,$46,$DF,$8C,$73,$8D,$C6,$6F,$E2,$CF,$F8,$79,$FF,$0A
	.dc.b	$D7,$9D,$F7,$E1,$F4,$1D,$D7,$F8,$4E,$BD,$C7,$92,$36,$E0,$F0,$9E
	.dc.b	$73,$E8,$9E,$6B,$F8,$AF,$C1,$37,$1C,$A7,$E0,$CE,$A7,$81,$DE,$FC
	.dc.b	$C4,$79,$57,$9A,$7F,$0B,$5C,$A7,$81,$AC,$3E,$D4,$7C,$36,$CF,$78
	.dc.b	$C6,$1F,$06,$73,$5C,$8F,$19,$E8,$C3,$C4,$E3,$0F,$13,$F4,$4F,$83
	.dc.b	$7D,$1F,$8D,$1A,$7E,$31,$FE,$4F,$3C,$D7,$E1,$3F,$43,$E5,$3C,$AF
	.dc.b	$D5,$72,$38,$3F,$59,$E2,$BC,$BC,$C6,$1B,$8F,$ED,$70,$E4,$FE,$DE
	.dc.b	$F9,$39,$C3,$6B,$0E,$8E,$F0,$E8,$D3,$81,$45,$F6,$7F,$C7,$73,$C9
	.dc.b	$B8,$76,$9D,$61,$E2,$97,$98,$F3,$3C,$7E,$43,$86,$9E,$69,$D5,$78
	.dc.b	$5A,$FF,$92,$A7,$83,$E5,$6B,$FE,$4E,$8D,$DF,$87,$BF,$FC,$83,$3F
	.dc.b	$C6,$DA,$9A,$4F,$03,$FC,$78,$4F,$03,$FC,$82,$7F,$8E,$E7,$F9,$01
	.dc.b	$FC,$97,$8B,$0E,$1F,$F1,$F7,$7F,$F8,$D7,$C0,$FF,$1B,$F0,$FF,$8E
	.dc.b	$BC,$0F,$F1,$BD,$FF,$99,$3C,$0F,$F1,$BE,$9F,$60,$FF,$1B,$C6,$70
	.dc.b	$E5,$7F,$E2,$D8,$F0,$8F,$03,$FC,$6E,$E1,$AF,$99,$AF,$F3,$77,$75
	.dc.b	$DF,$C0,$FF,$1E,$CF,$F9,$DA,$73,$5F,$8B,$9C,$D7,$DB,$9C,$D2,$13
	.dc.b	$AA,$ED,$E0,$7F,$91,$2B,$FD,$03,$D3,$FE,$04,$7F,$C2,$DC,$7F,$D6
	.dc.b	$7B,$3F,$BF,$F6,$FF,$5A,$C3,$FB,$1D,$7F,$57,$EF,$FE,$B8,$FF,$78
	.dc.b	$3F,$C3,$47,$F6,$4B,$E4,$FF,$8B,$30,$FE,$D3,$C3,$FD,$9A,$39,$5F
	.dc.b	$47,$FC,$25,$19,$6D,$72,$8F,$F0,$3C,$60,$E2,$77,$2B,$C6,$BB,$CF
	.dc.b	$13,$FC,$2B,$3F,$DC,$0E,$44,$8D,$BC,$A3,$83,$FE,$04,$9E,$98,$7C
	.dc.b	$38,$F0,$1C,$5C,$46,$B1,$F9,$24,$9F,$85,$3A,$AE,$B1,$F2,$2B,$83
	.dc.b	$AA,$E1,$1F,$02,$FC,$B3,$DE,$73,$5E,$C9,$FF,$2D,$56,$EF,$B5,$FA
	.dc.b	$C8,$71,$3B,$12,$38,$5F,$48,$F8,$17,$E6,$BF,$C3,$57,$F9,$36,$79
	.dc.b	$3C,$1E,$53,$E3,$BF,$F3,$44,$72,$3C,$35,$EA,$24,$79,$A3,$9C,$F1
	.dc.b	$BF,$F2,$64,$71,$32,$7F,$88,$6B,$F3,$0F,$F8,$88,$FD,$8B,$87,$FC
	.dc.b	$E2,$7F,$04,$7F,$C4,$4F,$1E,$EF,$F1,$13,$7B,$7F,$C5,$2E,$E3,$FA
	.dc.b	$9B,$FD,$44,$FE,$8E,$7F,$4C,$73,$3F,$D3,$4F,$ED,$97,$FE,$02,$3F
	.dc.b	$AF,$12,$7F,$AE,$63,$FD,$54,$FE,$D0,$F5,$3F,$AE,$57,$42,$5F,$F5
	.dc.b	$58,$F0,$36,$78,$DF,$87,$BB,$FB,$2C,$75,$C3,$C5,$7E,$07,$C4,$F7
	.dc.b	$8D,$3F,$DA,$23,$35,$CE,$3D,$F3,$DB,$0F,$1C,$78,$89,$19,$C3,$C9
	.dc.b	$7E,$09,$EC,$F4,$AE,$35,$B9,$E3,$3B,$BF,$04,$7C,$2B,$E6,$F9,$4F
	.dc.b	$51,$0E,$A5,$8E,$35,$EC,$8F,$EF,$C7,$33,$B9,$CC,$E2,$48,$E3,$5E
	.dc.b	$7A,$F0,$CF,$2A,$E9,$5F,$E3,$A8,$F8,$B5,$E9,$BF,$6B,$6F,$8D,$6E
	.dc.b	$BC,$57,$F0,$49,$5B,$8F,$29,$DD,$FF,$3D,$75,$FF,$78,$D3,$FC,$07
	.dc.b	$3F,$DF,$0F,$F0,$14,$65,$FF,$02,$99,$27,$07,$F7,$F8,$D4,$7F,$82
	.dc.b	$3B,$BF,$AB,$BC,$0F,$ED,$31,$FE,$0B,$6C,$69,$FF,$03,$F6,$7F,$81
	.dc.b	$1F,$EF,$D8,$7F,$81,$62,$46,$08,$7F,$80,$5F,$F0,$FF,$57,$F7,$CF
	.dc.b	$07,$FA,$B4,$FF,$68,$8C,$63,$FD,$F2,$31,$D7,$FD,$FD,$BD,$7F,$DF
	.dc.b	$1F,$F4,$F6,$B5,$FF,$00,$1F,$DC,$BA,$7F,$BC,$ED,$FE,$A3,$6C,$7F
	.dc.b	$65,$F3,$9E,$77,$F8,$0A,$3F,$BE,$57,$F8,$B6,$B7,$92,$F7,$39,$EB
	.dc.b	$FF,$02,$3F,$DB,$0F,$F0,$14,$FF,$81,$FD,$07,$A1,$AC,$7F,$D7,$1D
	.dc.b	$BF,$DD,$9F,$F0,$37,$66,$5B,$87,$F8,$FF,$06,$1C,$A3,$A9,$DE,$7F
	.dc.b	$C0,$95,$D6,$B7,$87,$3C,$3B,$73,$7F,$38,$E4,$FF,$7D,$61,$FC,$D8
	.dc.b	$FE,$59,$1F,$CC,$BA,$BF,$A0,$47,$F5,$39,$FE,$81,$D7,$FE,$EA,$8F
	.dc.b	$EB,$F5,$B7,$91,$B3,$F9,$EC,$7F,$3B,$8D,$70,$FF,$45,$25,$7F,$3D
	.dc.b	$A4,$7F,$BF,$5E,$47,$F3,$E2,$CF,$29,$E1,$3B,$8F,$64,$FF,$43,$AF
	.dc.b	$E9,$EF,$07,$C4,$4B,$F6,$CF,$09,$FF,$01,$3D,$5F,$87,$3B,$7D,$B1
	.dc.b	$B3,$FB,$6D,$F8,$8F,$F7,$77,$88,$ED,$FF,$11,$B7,$6F,$F0,$E5,$75
	.dc.b	$BF,$26,$3E,$2C,$7E,$66,$5F,$E0,$03,$E2,$CF,$8A,$74,$7F,$96,$F2
	.dc.b	$FB,$33,$A9,$E6,$79,$AB,$9C,$FD,$8A,$E1,$3F,$15,$FE,$D6,$F0,$AE
	.dc.b	$57,$ED,$BF,$92,$DA,$DB,$6B,$EB,$DF,$C0,$CB,$DE,$4C,$7D,$F8,$7C
	.dc.b	$9C,$7A,$BE,$4B,$FC,$2E,$3E,$C7,$FC,$9E,$60,$F5,$3E,$87,$EA,$C7
	.dc.b	$3C,$3E,$84,$0F,$F2,$86,$3E,$5B,$FC,$25,$77,$9F,$C5,$4E,$6B,$D1
	.dc.b	$87,$9B,$0F,$16,$1F,$E1,$BA,$FA,$C7,$F8,$26,$FE,$41,$FE,$0C,$C1
	.dc.b	$5F,$CA,$27,$52,$7F,$C1,$AD,$EC,$FF,$06,$B7,$B3,$FC,$1A,$FF,$4D
	.dc.b	$3F,$9F,$CF,$F4,$5C,$3F,$A6,$D2,$BE,$14,$A3,$15,$F4,$8F,$F7,$01
	.dc.b	$FD,$1C,$FE,$C4,$FF,$52,$7F,$AF,$47,$F5,$C7,$33,$FD,$52,$79,$D7
	.dc.b	$FA,$DA,$BB,$4F,$83,$B7,$FC,$17,$9F,$9E,$BF,$AB,$E1,$D3,$0E,$CF
	.dc.b	$84,$FE,$D4,$E3,$0F,$65,$71,$3F,$AF,$57,$2A,$E1,$3E,$57,$8E,$1E
	.dc.b	$53,$E0,$DF,$9A,$FF,$01,$3F,$EF,$AA,$EA,$FB,$0F,$97,$3B,$7F,$C7
	.dc.b	$87,$C7,$38,$39,$7E,$A1,$98,$FF,$1C,$BE,$DA,$F1,$39,$9E,$13,$A8
	.dc.b	$F4,$C7,$BC,$C9,$AC,$3E,$4C,$79,$70,$F8,$15,$E8,$8F,$79,$E8,$BF
	.dc.b	$79,$F4,$5F,$F0,$69,$E3,$34,$78,$DD,$3F,$E0,$CC,$3E,$69,$1E,$15
	.dc.b	$E0,$C3,$CB,$5C,$23,$F1,$EF,$82,$7E,$16,$1E,$87,$CB,$3C,$67,$4F
	.dc.b	$E1,$E7,$58,$7D,$92,$5F,$C3,$AD,$E1,$F6,$27,$E4,$1F,$04,$DC,$7C
	.dc.b	$EB,$FC,$95,$7C,$9C,$3E,$09,$F3,$A7,$57,$CE,$75,$3F,$04,$FF,$1E
	.dc.b	$76,$FF,$87,$89,$DB,$F8,$8E,$26,$DF,$8A,$F1,$1E,$8E,$63,$FC,$5F
	.dc.b	$57,$0F,$F1,$5D,$FF,$8A,$63,$10,$BE,$2F,$F8,$CA,$33,$87,$82,$3F
	.dc.b	$8B,$57,$81,$FE,$31,$5C,$1F,$F1,$21,$AB,$FC,$EF,$86,$F0,$3F,$C5
	.dc.b	$73,$FE,$FB,$9F,$D0,$F8,$8F,$4B,$38,$F5,$AF,$F1,$A9,$CC,$FE,$4D
	.dc.b	$87,$37,$02,$B9,$38,$F0,$7F,$C6,$2F,$F8,$CF,$C2,$FF,$1A,$BF,$B0
	.dc.b	$10,$8C,$C7,$ED,$21,$E1,$7F,$8D,$E3,$84,$63,$C2,$FF,$1E,$47,$F9
	.dc.b	$43,$93,$FC,$70,$FF,$80,$8F,$F2,$73,$FE,$54,$F5,$1E,$AF,$F8,$77
	.dc.b	$C2,$7A,$BF,$E2,$79,$CF,$07,$F8,$9E,$72,$FB,$A3,$38,$7C,$93,$FC
	.dc.b	$81,$5C,$6F,$A3,$FE,$41,$DF,$EB,$0F,$1D,$E4,$67,$1F,$B7,$18,$7F
	.dc.b	$C8,$47,$F8,$DD,$C6,$1F,$E4,$A3,$C1,$5C,$EB,$94,$23,$15,$CA,$37
	.dc.b	$5C,$A7,$EE,$75,$FF,$59,$9F,$EC,$1E,$07,$F8,$35,$FF,$19,$F7,$FF
	.dc.b	$65,$7F,$C1,$07,$F6,$A3,$FB,$69,$FD,$89,$FE,$C9,$3F,$D7,$9A,$71
	.dc.b	$8F,$F0,$27,$67,$F6,$E7,$FB,$6C,$6B,$BF,$FB,$3C,$FF,$84,$8D,$5F
	.dc.b	$8F,$83,$FC,$35,$18,$F0,$3F,$C0,$98,$79,$AF,$8B,$FD,$B2,$BF,$BB
	.dc.b	$C7,$F5,$EC,$3F,$C2,$27,$79,$F5,$4F,$F7,$58,$F3,$63,$F0,$E7,$CB
	.dc.b	$5C,$4F,$F0,$BC,$F1,$AF,$F0,$8C,$66,$7F,$A1,$63,$E7,$C7,$FC,$4D
	.dc.b	$1F,$1B,$0F,$89,$5B,$BF,$2C,$E7,$0F,$09,$D6,$FD,$DA,$7F,$82,$E7
	.dc.b	$CF,$1F,$15,$FA,$75,$E2,$79,$5F,$C2,$7F,$C3,$AD,$C3,$FC,$9B,$1C
	.dc.b	$2B,$AB,$F2,$4E,$A7,$C9,$7F,$C3,$6E,$6B,$B1,$F1,$EB,$A6,$1F,$31
	.dc.b	$F6,$57,$09,$FA,$F5,$DC,$87,$E0,$8B,$7F,$3F,$1F,$91,$1F,$E3,$79
	.dc.b	$EF,$5E,$F8,$E0,$4A,$E1,$97,$E0,$AB,$BC,$E7,$1F,$94,$F6,$7F,$C5
	.dc.b	$AF,$F8,$BF,$1F,$F1,$74,$62,$3F,$89,$5F,$F8,$C0,$FF,$17,$93,$BB
	.dc.b	$FC,$5E,$DE,$FF,$E8,$F3,$FE,$02,$C3,$FA,$C3,$FD,$32,$7F,$A9,$57
	.dc.b	$F5,$78,$FE,$D8,$7F,$62,$34,$FF,$68,$70,$7F,$4F,$3F,$AA,$1F,$DD
	.dc.b	$27,$FA,$CD,$7F,$6D,$CB,$FC,$41,$3F,$DC,$9F,$EE,$8F,$09,$FE,$D8
	.dc.b	$F8,$1F,$2B,$D6,$FC,$B5,$EA,$8F,$F0,$BC,$E8,$FE,$F2,$F1,$AF,$ED
	.dc.b	$DE,$07,$F7,$4C,$3D,$95,$E4,$AF,$4B,$C0,$C1,$CE,$BC,$F1,$E5,$7C
	.dc.b	$F8,$78,$A3,$DD,$1F,$DC,$70,$F7,$47,$F8,$76,$37,$7F,$1E,$FA,$4E
	.dc.b	$4F,$84,$43,$B1,$6F,$B4,$7C,$43,$CB,$5E,$C9,$FA,$77,$FE,$0C,$7F
	.dc.b	$CC,$35,$C2,$7C,$33,$99,$FF,$1F,$B6,$7F,$C2,$37,$E5,$AE,$95,$F3
	.dc.b	$E3,$0F,$46,$BC,$8F,$9C,$66,$BA,$61,$F3,$DF,$AD,$7F,$E6,$A9,$F9
	.dc.b	$F7,$F3,$23,$CD,$1E,$FB,$EB,$1E,$4A,$E7,$3E,$E3,$B9,$E1,$C3,$E1
	.dc.b	$BE,$13,$33,$C5,$B8,$78,$DF,$AE,$7E,$87,$0F,$7D,$7E,$05,$FB,$51
	.dc.b	$E6,$3E,$B1,$F5,$0F,$F3,$E7,$57,$F8,$25,$FE,$D4,$7F,$82,$63,$FC
	.dc.b	$47,$D4,$7F,$C2,$7B,$FF,$86,$BB,$33,$12,$BF,$C2,$37,$CF,$87,$FC
	.dc.b	$30,$DB,$FF,$0E,$D6,$EB,$FC,$4A,$78,$36,$78,$5F,$E3,$2F,$5D,$EB
	.dc.b	$D3,$FC,$6B,$3A,$AF,$F0,$FB,$5F,$4C,$F0,$7F,$C4,$6D,$7B,$57,$F8
	.dc.b	$FB,$B3,$FC,$47,$1F,$E1,$AE,$EF,$F0,$FB,$FE,$D1,$8F,$F1,$16,$9F
	.dc.b	$E1,$C8,$FF,$11,$77,$7F,$88,$5B,$DD,$FE,$1E,$E1,$91,$FE,$2B,$1D
	.dc.b	$BF,$E1,$6B,$FF,$08,$F7,$7F,$86,$A7,$E7,$CE,$78,$3F,$C9,$73,$9A
	.dc.b	$FF,$20,$4E,$6B,$FC,$3B,$5F,$E2,$1E,$EF,$F0,$EB,$FE,$0D,$AF,$F0
	.dc.b	$EF,$67,$F8,$36,$3F,$D5,$7E,$43,$C8,$FF,$10,$C7,$F8,$7B,$BF,$FC
	.dc.b	$2F,$7F,$E1,$6C,$7F,$C3,$8F,$28,$CE,$1C,$E7,$FD,$E9,$87,$3B,$FF
	.dc.b	$12,$57,$49,$FE,$7D,$DB,$FD,$49,$FF,$0D,$F7,$7F,$45,$E7,$FE,$E7
	.dc.b	$BF,$F5,$99,$FE,$BE,$7F,$66,$2F,$67,$F6,$67,$FA,$55,$F3,$7F,$C0
	.dc.b	$67,$F6,$E3,$7C,$FF,$D8,$5C,$57,$F5,$87,$FB,$51,$E2,$CB,$FB,$D4
	.dc.b	$66,$3C,$0D,$BF,$01,$FD,$72,$77,$87,$F7,$69,$FE,$BB,$39,$AF,$EE
	.dc.b	$93,$93,$DA,$7F,$67,$8D,$4F,$0B,$FE,$C7,$5E,$C3,$CF,$5C,$5F,$EE
	.dc.b	$0D,$7F,$C7,$98,$7C,$1E,$FF,$EE,$B1,$83,$89,$F4,$0A,$75,$BF,$9C
	.dc.b	$7B,$6F,$E3,$38,$BE,$D8,$7F,$89,$2B,$8B,$E8,$8F,$5C,$7C,$A3,$E8
	.dc.b	$CF,$17,$C0,$D7,$C1,$1F,$06,$BD,$F5,$CE,$7D,$A7,$09,$D6,$1E,$DC
	.dc.b	$BF,$0D,$3A,$39,$9E,$1C,$7E,$0D,$F9,$F1,$F0,$3D,$67,$E2,$46,$31
	.dc.b	$F0,$93,$0E,$D9,$7E,$2E,$FF,$C1,$D3,$FE,$11,$BF,$F6,$D3,$50,$8F
	.dc.b	$F0,$A7,$7F,$F8,$48,$9D,$FF,$BF,$9D,$46,$1F,$E6,$A7,$F7,$CD,$3F
	.dc.b	$9A,$9F,$D1,$E3,$FA,$2F,$7F,$F3,$BC,$FF,$B2,$4F,$F3,$87,$FB,$6D
	.dc.b	$7F,$BD,$A3,$FB,$0B,$FD,$08,$FE,$95,$1F,$D2,$1F,$E9,$51,$C4,$EE
	.dc.b	$FF,$40,$C3,$FA,$AB,$FD,$05,$A4,$BF,$1C,$74,$9E,$4F,$F5,$3A,$ED
	.dc.b	$5F,$D4,$DB,$7E,$9A,$F2,$C6,$2B,$91,$FD,$54,$E5,$1D,$1F,$39,$92
	.dc.b	$9F,$1C,$D9,$9A,$F0,$47,$16,$C6,$6B,$C4,$FF,$71,$C3,$FB,$7B,$FE
	.dc.b	$01,$6D,$FF,$65,$8E,$07,$98,$F0,$1E,$2A,$FE,$FD,$1F,$49,$B7,$EE
	.dc.b	$3C,$57,$E0,$AF,$99,$87,$CA,$7C,$07,$D7,$2D,$FC,$56,$D7,$27,$C5
	.dc.b	$7D,$9F,$82,$79,$1F,$7B,$6F,$C1,$1F,$67,$0F,$6D,$7E,$35,$B5,$F8
	.dc.b	$1C,$3C,$98,$7C,$18,$E4,$78,$21,$3F,$E4,$08,$FC,$81,$F2,$CF,$54
	.dc.b	$F1,$BF,$34,$7C,$7A,$E2,$FF,$83,$CE,$27,$C4,$AF,$1A,$ED,$FF,$02
	.dc.b	$3F,$AE,$EF,$FF,$07,$3F,$B0,$E3,$6B,$FE,$08,$8F,$F0,$6F,$7F,$F8
	.dc.b	$49,$E1,$A7,$F8,$2E,$BF,$C1,$F3,$99,$FF,$0D,$1C,$EB,$93,$FE,$10
	.dc.b	$9E,$AF,$EF,$B6,$FF,$05,$46,$21,$DD,$FE,$09,$8C,$36,$BF,$82,$CA
	.dc.b	$B8,$C6,$1B,$1F,$C2,$A7,$83,$7B,$FF,$C3,$7D,$FF,$E1,$3E,$FF,$F0
	.dc.b	$8C,$BB,$FF,$C2,$AF,$67,$FC,$26,$FF,$83,$57,$7F,$F8,$4E,$3F,$C3
	.dc.b	$0D,$7C,$BC,$7F,$E1,$47,$F8,$B3,$FC,$5F,$B7,$FC,$19,$AB,$BB,$FC
	.dc.b	$21,$B7,$B2,$73,$18,$BF,$E4,$71,$8F,$03,$FC,$01,$1F,$E1,$EE,$FF
	.dc.b	$F0,$9B,$FE,$4A,$ED,$FF,$03,$3F,$E2,$36,$BF,$20,$38,$7F,$C2,$FD
	.dc.b	$FF,$45,$B8,$FF,$83,$8F,$F0,$F1,$FE,$19,$2A,$CB,$FC,$34,$7D,$9A
	.dc.b	$F0,$BF,$E1,$B9,$FC,$14,$FF,$51,$3F,$A6,$76,$FF,$50,$7F,$A4,$F2
	.dc.b	$FF,$54,$7F,$B1,$F2,$7F,$49,$6F,$67,$F4,$A7,$FC,$02,$73,$3F,$C1
	.dc.b	$73,$FD,$B2,$7F,$DB,$FB,$FF,$71,$AF,$E9,$6F,$E7,$AB,$FB,$5D,$77
	.dc.b	$8F,$EB,$13,$D5,$FF,$04,$36,$3F,$B9,$CF,$63,$C1,$D7,$FE,$18,$BF
	.dc.b	$35,$F2,$9F,$39,$1C,$12,$BC,$04,$8F,$87,$8F,$F8,$81,$EB,$5F,$E0
	.dc.b	$89,$D4,$78,$4F,$EE,$A5,$BF,$EE,$47,$F7,$58,$DB,$E5,$C3,$D5,$3F
	.dc.b	$DF,$63,$75,$F1,$AB,$E4,$57,$F7,$58,$F0,$57,$90,$F9,$E7,$82,$BA
	.dc.b	$3E,$DA,$EF,$3C,$23,$C8,$FC,$48,$CD,$7D,$27,$F7,$3D,$BF,$E0,$C9
	.dc.b	$F1,$57,$C0,$AD,$3F,$E4,$93,$C3,$7E,$88,$F3,$1A,$8F,$B6,$79,$F1
	.dc.b	$FB,$2F,$94,$C6,$1F,$6A,$BD,$33,$F4,$E7,$27,$C5,$BF,$43,$83,$F2
	.dc.b	$A5,$9F,$F1,$C3,$FA,$B7,$FC,$37,$7F,$E1,$37,$FC,$2F,$D7,$7B,$3F
	.dc.b	$C2,$95,$B7,$FC,$2D,$87,$F8,$42,$3F,$B1,$9F,$D5,$0A,$FF,$3F,$8F
	.dc.b	$E7,$AF,$F6,$37,$FD,$BA,$63,$1F,$F0,$7C,$FF,$40,$BF,$EA,$8F,$F4
	.dc.b	$88,$FE,$A8,$D3,$FA,$A5,$FF,$44,$9F,$E9,$67,$0B,$FE,$CC,$78,$63
	.dc.b	$FA,$A4,$6A,$FF,$AD,$DF,$4A,$DB,$4F,$ED,$97,$EB,$3F,$AF,$96,$BB
	.dc.b	$F6,$FF,$6A,$BE,$73,$E4,$3C,$C7,$AD,$CB,$70,$F4,$47,$C2,$7C,$AD
	.dc.b	$8F,$EC,$55,$E7,$9D,$9F,$87,$7D,$D7,$E6,$C3,$DD,$5C,$2F,$94,$7C
	.dc.b	$FB,$E1,$1E,$2B,$E9,$5E,$66,$D7,$C2,$8F,$53,$5F,$41,$B9,$EB,$7F
	.dc.b	$62,$B8,$BE,$33,$E5,$4F,$E0,$9A,$FC,$AA,$F3,$9E,$57,$FC,$BF,$3F
	.dc.b	$31,$FF,$06,$5F,$C4,$3D,$27,$C2,$77,$7F,$89,$7F,$C3,$F3,$C5,$FA
	.dc.b	$D1,$97,$FC,$1D,$3C,$DE,$2D,$AF,$CD,$47,$96,$FC,$F7,$CE,$7C,$F1
	.dc.b	$EF,$BF,$83,$5E,$07,$B5,$30,$FE,$D3,$5F,$DF,$BB,$7F,$C2,$F7,$FE
	.dc.b	$0C,$3F,$A8,$6B,$FE,$0C,$DF,$FC,$23,$1F,$E1,$63,$FA,$79,$B7,$FC
	.dc.b	$19,$DB,$FD,$05,$DC,$72,$3F,$D1,$F4,$ED,$FF,$46,$47,$F8,$9D,$B1
	.dc.b	$98,$FF,$48,$35,$FE,$D8,$7F,$84,$FA,$FF,$C2,$B5,$FE,$18,$EC,$B8
	.dc.b	$FF,$86,$3B,$7F,$C2,$77,$B3,$CF,$C5,$FE,$18,$7F,$C4,$35,$C1,$FF
	.dc.b	$0C,$1A,$C3,$FC,$31,$87,$FA,$87,$B3,$FC,$2F,$3F,$EA,$87,$FC,$2E
	.dc.b	$FF,$7F,$7F,$C2,$FE,$06,$FB,$7F,$C2,$D1,$89,$F7,$F1,$FF,$85,$DF
	.dc.b	$5B,$FE,$17,$EA,$FF,$07,$F6,$7F,$85,$3C,$7B,$13,$7F,$EF,$B9,$FF
	.dc.b	$86,$FB,$71,$8F,$CD,$8C,$4F,$F8,$22,$3F,$C2,$3D,$BF,$E4,$29,$9D
	.dc.b	$CF,$27,$FC,$3B,$DF,$FE,$0A,$C3,$FC,$27,$87,$F8,$77,$0E,$11,$FE
	.dc.b	$57,$BE,$0F,$5A,$E1,$5F,$E1,$5F,$09,$E1,$74,$C3,$8C,$E6,$3F,$C5
	.dc.b	$7D,$9F,$D3,$61,$C1,$FE,$09,$6F,$5F,$F5,$8F,$07,$FC,$31,$5F,$DB
	.dc.b	$E7,$FA,$27,$47,$F5,$A7,$FB,$33,$FD,$8E,$3F,$A7,$D7,$F6,$29,$E6
	.dc.b	$7F,$5D,$BF,$EA,$E7,$F5,$37,$FB,$4F,$81,$FD,$E1,$FE,$A2,$60,$FE
	.dc.b	$B8,$58,$E1,$5D,$A3,$C7,$3F,$D9,$9E,$E7,$F6,$63,$85,$78,$EB,$FC
	.dc.b	$49,$E0,$FF,$73,$AE,$11,$8C,$7D,$F8,$FB,$71,$F8,$D3,$F1,$4D,$6D
	.dc.b	$FE,$19,$CB,$E0,$E1,$FD,$D7,$1E,$91,$E2,$BE,$39,$7F,$84,$23,$18
	.dc.b	$FF,$70,$7C,$A7,$A8,$CD,$FC,$3A,$FA,$11,$A2,$3F,$05,$B3,$F4,$49
	.dc.b	$3E,$3C,$BD,$6F,$BA,$FE,$E6,$3E,$9C,$3C,$B8,$7F,$84,$CF,$93,$8F
	.dc.b	$AB,$2F,$81,$7C,$27,$FC,$2B,$97,$A6,$FC,$44,$7C,$31,$CB,$2F,$0B
	.dc.b	$EC,$3E,$DC,$EF,$2F,$36,$5E,$3B,$E8,$72,$C7,$E7,$DF,$82,$BE,$24
	.dc.b	$FD,$53,$6F,$28,$D1,$DA,$39,$63,$E9,$AF,$F1,$3C,$FF,$86,$E5,$5F
	.dc.b	$E1,$B6,$C7,$F1,$EC,$3F,$C3,$AD,$F0,$3F,$C3,$AF,$FA,$D2,$11,$FE
	.dc.b	$1E,$C3,$FD,$F1,$3F,$D6,$CF,$EC,$AF,$F5,$F6,$99,$3F,$B3,$E1,$FD
	.dc.b	$A6,$BF,$B2,$C6,$5A,$FF,$4F,$38,$94,$E3,$3D,$5A,$71,$AF,$EA,$2F
	.dc.b	$F5,$C3,$FC,$01,$E0,$FF,$66,$7F,$BB,$36,$FC,$53,$C2,$FF,$AB,$57
	.dc.b	$F6,$E3,$89,$FD,$90,$B3,$FD,$92,$3F,$B1,$5F,$F6,$4A,$FE,$BF,$E0
	.dc.b	$FF,$88,$70,$FE,$E9,$7B,$9C,$E1,$D6,$3F,$BF,$19,$9F,$EE,$17,$F3
	.dc.b	$27,$E0,$61,$F1,$0E,$EF,$C4,$6B,$F0,$8D,$DF,$AF,$0F,$14,$7C,$03
	.dc.b	$D5,$87,$82,$FB,$5F,$23,$B5,$F9,$AB,$FC,$0F,$87,$91,$09,$FF,$17
	.dc.b	$BF,$46,$3E,$65,$7D,$59,$E7,$1C,$67,$85,$7F,$8B,$CF,$33,$F0,$EB
	.dc.b	$FC,$15,$3C,$A3,$A1,$E8,$8C,$B4,$FF,$07,$47,$5A,$F1,$C7,$C2,$8E
	.dc.b	$4D,$38,$1F,$85,$BF,$84,$FB,$27,$89,$F0,$EF,$CB,$1F,$08,$FE,$3D
	.dc.b	$87,$F8,$69,$FF,$06,$F8,$1F,$DE,$DF,$F0,$A6,$9F,$E1,$93,$FC,$3D
	.dc.b	$19,$F0,$3F,$C3,$E4,$78,$78,$2F,$0F,$51,$2F,$FC,$40,$7F,$86,$C4
	.dc.b	$ED,$FF,$04,$1F,$E2,$4F,$0B,$FC,$22,$7F,$84,$B7,$FF,$08,$13,$C1
	.dc.b	$F3,$3F,$E1,$38,$FF,$0A,$77,$FF,$83,$A8,$D3,$FC,$14,$FF,$83,$FC
	.dc.b	$0F,$F0,$68,$F0,$7F,$C1,$91,$F8,$43,$FD,$E6,$7F,$84,$0F,$F1,$4F
	.dc.b	$B0,$F6,$BF,$BA,$F7,$BB,$9C,$3F,$DE,$2D,$7B,$8F,$F0,$8C,$E9,$FF
	.dc.b	$20,$C6,$67,$FC,$25,$39,$7E,$34,$66,$BF,$C2,$35,$FE,$18,$7F,$C2
	.dc.b	$7E,$05,$8F,$F0,$73,$69,$3F,$E1,$C7,$E8,$78,$6F,$0B,$D3,$8A,$F8
	.dc.b	$5E,$CF,$5D,$FE,$9A,$F8,$61,$EA,$AF,$F0,$9E,$FF,$CE,$AB,$FA,$44
	.dc.b	$FF,$43,$EB,$FE,$7E,$7F,$59,$8F,$F0,$57,$5F,$FB,$76,$BF,$9E,$C6
	.dc.b	$0F,$F6,$E4,$F0,$7F,$A8,$75,$7F,$82,$27,$71,$FE,$05,$EA,$FE,$A4
	.dc.b	$F8,$0F,$E9,$33,$BA,$FE,$98,$FF,$86,$2F,$FC,$14,$7F,$4B,$3C,$51
	.dc.b	$A6,$CF,$39,$E0,$FF,$7D,$7F,$B5,$BE,$73,$FC,$3F,$5E,$28,$FE,$BE
	.dc.b	$70,$8F,$EF,$B5,$D5,$F6,$9F,$D7,$5C,$3C,$AF,$DB,$DB,$FD,$E7,$0F
	.dc.b	$F0,$85,$F8,$2B,$E0,$56,$BB,$7F,$BA,$CE,$6F,$85,$7F,$80,$E3,$C7
	.dc.b	$7E,$C6,$C7,$89,$F8,$CF,$BE,$34,$7C,$57,$4F,$C8,$7E,$06,$3F,$E1
	.dc.b	$C3,$E0,$46,$E3,$D5,$8F,$B2,$75,$8F,$9F,$2F,$C2,$57,$C1,$3C,$07
	.dc.b	$63,$E7,$5F,$C0,$23,$EB,$7B,$52,$BF,$C1,$57,$F8,$7C,$15,$FD,$F6
	.dc.b	$7F,$C0,$77,$FE,$07,$AF,$F0,$31,$1F,$F0,$27,$6F,$FA,$91,$FE,$76
	.dc.b	$DC,$3F,$C0,$AD,$ED,$FF,$03,$4F,$F2,$D7,$FA,$04,$7F,$34,$9F,$E5
	.dc.b	$CD,$8F,$EA,$18,$FF,$B3,$2B,$F9,$E9,$FD,$C3,$1F,$EA,$F3,$A7,$FB
	.dc.b	$85,$7F,$41,$8E,$B3,$FC,$FA,$FF,$BD,$BF,$CE,$6F,$F9,$FB,$E8,$3F
	.dc.b	$B2,$3F,$DC,$5E,$C7,$9E,$BF,$BC,$DF,$81,$EC,$62,$34,$72,$8D,$3C
	.dc.b	$8E,$D3,$FE,$0E,$9F,$71,$CA,$38,$1C,$A3,$4F,$A7,$B7,$FB,$CC,$60
	.dc.b	$F3,$1E,$AB,$FE,$AF,$87,$C3,$AF,$F0,$23,$5F,$F0,$24,$7C,$E7,$A1
	.dc.b	$FD,$BC,$F9,$B3,$CE,$7E,$4B,$DB,$0F,$ED,$ED,$9D,$C6,$0F,$7C,$71
	.dc.b	$C3,$E2,$4F,$C0,$9F,$F0,$15,$7F,$6A,$AF,$B9,$3F,$DC,$67,$93,$F0
	.dc.b	$5D,$D7,$C6,$8F,$95,$7E,$E8,$F9,$15,$F0,$EB,$C2,$72,$AF,$01,$FE
	.dc.b	$08,$8F,$43,$DC,$F4,$90,$C1,$2B,$F0,$F1,$E6,$6C,$FE,$3A,$7F,$C0
	.dc.b	$27,$F8,$8A,$7F,$52,$FF,$89,$32,$FE,$D9,$E2,$7F,$80,$FA,$7F,$B9
	.dc.b	$75,$7F,$70,$31,$12,$BF,$BC,$9F,$DF,$6B,$8B,$FD,$EE,$38,$3F,$E3
	.dc.b	$0E,$3F,$E9,$AB,$DA,$75,$7F,$84,$7C,$66,$54,$FF,$0A,$4E,$AB,$97
	.dc.b	$17,$F7,$78,$C4,$7F,$83,$6B,$8B,$A3,$FC,$7E,$E1,$FF,$09,$47,$0C
	.dc.b	$3C,$7D,$9F,$E0,$7A,$FE,$D3,$5F,$E0,$33,$FC,$8F,$08,$CC,$7F,$82
	.dc.b	$FB,$3F,$85,$4F,$F8,$3A,$3F,$C1,$4D,$78,$E5,$FE,$0F,$7F,$C1,$87
	.dc.b	$F1,$1D,$BF,$C2,$2F,$F8,$23,$B3,$FB,$AE,$7F,$E0,$BE,$CF,$F0,$13
	.dc.b	$F2,$7A,$FF,$C0,$DA,$7C,$A9,$CD,$FF,$80,$27,$35,$FE,$1B,$9C,$D7
	.dc.b	$F8,$5E,$3F,$C2,$DD,$7F,$DF,$2F,$DC,$1C,$3F,$05,$AF,$F8,$95,$AF
	.dc.b	$F8,$17,$B3,$FC,$17,$D9,$FE,$0A,$C7,$FC,$19,$7F,$46,$BA,$56,$A3
	.dc.b	$17,$E1,$C3,$FC,$37,$1B,$E6,$FF,$6E,$DF,$FB,$D8,$FE,$8E,$7F,$4E
	.dc.b	$2B,$FD,$3B,$AB,$FA,$87,$2F,$F4,$49,$FE,$9D,$B7,$F6,$07,$83,$6B
	.dc.b	$84,$72,$9E,$13,$FD,$90,$D3,$8E,$7F,$EC,$53,$FD,$CA,$FB,$9F,$E1
	.dc.b	$D7,$07,$37,$1D,$7F,$DD,$23,$D5,$7D,$27,$3D,$9F,$DE,$AF,$FA,$FD
	.dc.b	$69,$FE,$BF,$87,$86,$3F,$B2,$9F,$D8,$CF,$EE,$6E,$5F,$F0,$79,$E1
	.dc.b	$3F,$C5,$31,$E8,$C3,$C5,$8F,$CC,$25,$7F,$8E,$71,$F2,$5F,$03,$B9
	.dc.b	$E3,$3B,$E3,$E8,$CB,$FB,$DC,$FC,$7C,$7C,$F3,$9A,$FF,$19,$C7,$87
	.dc.b	$2F,$74,$7A,$4F,$8F,$5E,$03,$FC,$35,$1A,$8F,$01,$A9,$EA,$F9,$4F
	.dc.b	$7C,$F1,$8F,$93,$8F,$B2,$FF,$C0,$91,$8C,$7D,$CF,$2A,$EF,$8F,$AA
	.dc.b	$3C,$CD,$C3,$E0,$E1,$CD,$FC,$5C,$F1,$3A,$61,$FE,$0B,$8F,$F0,$71
	.dc.b	$FA,$A9,$FF,$0B,$47,$F8,$27,$AA,$4F,$F8,$40,$87,$F0,$D9,$FF,$66
	.dc.b	$BF,$D1,$BA,$FF,$AC,$DF,$F4,$48,$FE,$7D,$3F,$CE,$9F,$EC,$B5,$FD
	.dc.b	$2A,$7F,$A7,$C7,$F6,$43,$FC,$0C,$FF,$4B,$9F,$EA,$11,$D6,$7F,$AE
	.dc.b	$C6,$31,$FE,$C1,$3F,$DE,$8F,$EA,$27,$FB,$3D,$FE,$A3,$1E,$18,$FE
	.dc.b	$E7,$7F,$DD,$63,$D3,$3C,$8F,$EB,$93,$FD,$8C,$FE,$B5,$7E,$B7,$FA
	.dc.b	$B7,$67,$F8,$12,$3D,$33,$DA,$31,$3C,$0E,$47,$8E,$79,$57,$8C,$F0
	.dc.b	$BE,$FB,$F9,$33,$C2,$7F,$C1,$17,$D6,$3F,$C1,$07,$88,$B1,$CE,$7F
	.dc.b	$BF,$57,$F6,$7A,$F6,$3E,$63,$C3,$7E,$6B,$F2,$1D,$EB,$E2,$1C,$9F
	.dc.b	$AE,$7D,$13,$E7,$5F,$D1,$AF,$46,$1F,$52,$BD,$35,$F4,$AB,$D3,$5E
	.dc.b	$08,$F3,$DF,$86,$73,$1E,$1B,$FC,$23,$F2,$4E,$57,$E4,$8F,$F0,$A6
	.dc.b	$1E,$C9,$D5,$78,$D0,$7F,$01,$9F,$F7,$C8,$FF,$7D,$75,$7F,$81,$1F
	.dc.b	$F0,$0E,$FF,$DF,$23,$FD,$F3,$D7,$FE,$00,$9F,$F0,$27,$67,$F9,$BD
	.dc.b	$FF,$3D,$36,$CD,$35,$D9,$FD,$D7,$83,$FD,$E9,$D9,$FD,$E7,$2F,$F0
	.dc.b	$4B,$63,$FC,$05,$D5,$FE,$06,$3F,$C1,$18,$FF,$51,$9D,$3E,$CB,$E8
	.dc.b	$FF,$7C,$8F,$EF,$6F,$F7,$86,$B9,$BF,$EF,$B1,$88,$75,$FF,$7D,$6D
	.dc.b	$FF,$7D,$7F,$C0,$87,$F6,$C8,$D3,$FE,$2F,$3F,$C0,$27,$F7,$0D,$BE
	.dc.b	$22,$7F,$D2,$9D,$5F,$DF,$23,$FC,$09,$97,$F8,$6F,$17,$B8,$F7,$3F
	.dc.b	$C0,$DD,$B2,$B5,$D9,$FE,$02,$C3,$ED,$E3,$FE,$02,$8C,$47,$F7,$49
	.dc.b	$FE,$E7,$D9,$FF,$00,$32,$94,$F2,$3F,$C0,$F3,$2B,$67,$68,$C5,$AF
	.dc.b	$FC,$7B,$DB,$FD,$F6,$BF,$C3,$1D,$DF,$E0,$53,$B3,$FC,$AA,$BF,$96
	.dc.b	$9F,$D3,$8F,$E7,$DC,$FF,$EC,$E7,$1B,$FF,$38,$7F,$A6,$9F,$CD,$B9
	.dc.b	$7F,$9E,$BA,$EB,$E9,$3F,$D5,$63,$07,$F5,$1E,$EF,$EC,$75,$C2,$3F
	.dc.b	$A7,$1A,$3F,$A8,$C6,$BB,$7F,$A5,$C7,$1E,$EF,$EA,$D8,$FF,$45,$C3
	.dc.b	$B4,$F2,$25,$F1,$7D,$A6,$2B,$8E,$1E,$19,$D5,$72,$6E,$3F,$D5,$2F
	.dc.b	$DE,$7F,$60,$7F,$A9,$3E,$3C,$BF,$AC,$39,$3D,$58,$F9,$67,$95,$FF
	.dc.b	$9F,$A7,$FC,$47,$1C,$9F,$7E,$3E,$4C,$BD,$F1,$8C,$3A,$E1,$F4,$E3
	.dc.b	$D5,$18,$CB,$FD,$FD,$1E,$69,$D6,$5E,$A8,$FA,$D7,$D2,$7C,$91,$83
	.dc.b	$E2,$CE,$DF,$51,$D8,$FC,$1D,$7F,$87,$E3,$E6,$47,$F7,$02,$1F,$F0
	.dc.b	$06,$BF,$B7,$BF,$DB,$7B,$5D,$B8,$8F,$E3,$A7,$FB,$A4,$D3,$FC,$93
	.dc.b	$3F,$E6,$71,$FC,$DA,$7F,$99,$E3,$FC,$96,$3F,$A4,$1F,$DB,$8F,$EE
	.dc.b	$BD,$DF,$CC,$E7,$F9,$74,$7F,$73,$D7,$FD,$65,$39,$C3,$F9,$AC,$FF
	.dc.b	$48,$3F,$9A,$B8,$BF,$E7,$71,$83,$FA,$45,$7F,$6C,$70,$FF,$39,$79
	.dc.b	$47,$94,$FE,$B5,$3E,$B8,$F2,$9F,$D2,$08,$7A,$4F,$5C,$F9,$9C,$B5
	.dc.b	$D4,$F4,$7D,$98,$7F,$65,$3E,$3C,$F9,$67,$C5,$7E,$FA,$F4,$D7,$F6
	.dc.b	$97,$FA,$B1,$83,$FA,$8E,$1F,$38,$98,$7F,$63,$24,$7D,$36,$E1,$DE
	.dc.b	$77,$39,$24,$66,$77,$5E,$19,$F9,$2F,$A0,$C6,$1E,$88,$F8,$31,$ED
	.dc.b	$7E,$49,$F0,$B0,$FC,$4D,$FE,$AF,$3F,$EC,$9C,$1F,$D8,$7B,$7F,$B0
	.dc.b	$3F,$ED,$E8,$FE,$BB,$7F,$D9,$B6,$FC,$8C,$62,$3F,$B5,$F6,$F0,$9C
	.dc.b	$B6,$3F,$B9,$F7,$6E,$FF,$DD,$3D,$DF,$D6,$A7,$FD,$A9,$3F,$D8,$1F
	.dc.b	$F6,$EF,$6F,$F6,$37,$FD,$65,$3B,$38,$77,$3D,$DF,$D1,$E1,$FE,$E2
	.dc.b	$8F,$EB,$DD,$FF,$EE,$38,$FF,$6E,$E1,$FD,$77,$0F,$F6,$EB,$5F,$F6
	.dc.b	$EF,$81,$A3,$FD,$B6,$FF,$5F,$3F,$DB,$C7,$F8,$07,$7F,$F0,$6F,$BE
	.dc.b	$F4,$BC,$3E,$FB,$83,$31,$FD,$96,$D5,$B7,$FC,$2B,$92,$F8,$46,$AB
	.dc.b	$85,$6E,$3F,$AA,$4E,$A3,$11,$FD,$CD,$D5,$73,$BF,$0C,$FF,$C0,$1C
	.dc.b	$2B,$CA,$2B,$C7,$0F,$4C,$62,$3F,$BD,$1E,$33,$CF,$87,$86,$7C,$A7
	.dc.b	$08,$D1,$F8,$C2,$DF,$8D,$C9,$DC,$E7,$5E,$6C,$5F,$06,$F8,$2F,$ED
	.dc.b	$B4,$9C,$C6,$21,$ED,$BE,$0B,$75,$AC,$3E,$69,$FE,$42,$71,$5F,$DF
	.dc.b	$63,$FC,$8B,$F1,$0F,$84,$71,$72,$CF,$C5,$5C,$AD,$7B,$AF,$2A,$F8
	.dc.b	$77,$C3,$61,$B7,$C6,$75,$F0,$F9,$A9,$39,$F8,$87,$C8,$E3,$2F,$31
	.dc.b	$E0,$8C,$CF,$3F,$8A,$73,$F8,$9E,$F5,$C3,$24,$6A,$B8,$63,$BB,$E0
	.dc.b	$D8,$C9,$E7,$F8,$C7,$A7,$EE,$6C,$67,$E3,$31,$32,$3D,$DF,$1A,$F8
	.dc.b	$EC,$69,$E3,$8C,$4F,$5A,$F9,$31,$89,$F4,$61,$D9,$E1,$7F,$7F,$E3
	.dc.b	$FF,$C9,$44,$F1,$31,$E9,$CC,$BE,$B5,$BC,$CF,$F4,$0C,$62,$FF,$CA
	.dc.b	$51,$89,$F9,$A6,$CF,$C6,$7C,$83,$E4,$7F,$1D,$3B,$EE,$FD,$AD,$2B
	.dc.b	$58,$78,$29,$E7,$70,$E6,$E2,$D9,$D4,$7E,$FB,$C2,$FF,$0D,$53,$E4
	.dc.b	$B1,$E6,$7F,$7C,$8F,$BF,$97,$F9,$AE,$31,$3A,$8F,$F4,$EC,$F1,$C3
	.dc.b	$FD,$3F,$7D,$3C,$3D,$57,$C3,$BB,$7F,$C5,$EF,$78,$7E,$8B,$E4,$FD
	.dc.b	$31,$8B,$FC,$FC,$78,$B1,$56,$F3,$F1,$75,$2B,$53,$CF,$4C,$FC,$A6
	.dc.b	$31,$E7,$18,$26,$5D,$F0,$7C,$AB,$E4,$B3,$7F,$04,$FD,$8C,$66,$B8
	.dc.b	$7C,$C3,$E5,$FC,$6B,$E0,$2C,$61,$93,$DE,$1F,$31,$8C,$3B,$65,$CE
	.dc.b	$3D,$B3,$AA,$F3,$7C,$CF,$F4,$F7,$94,$FC,$28,$C5,$F2,$F9,$A7,$A7
	.dc.b	$F1,$9F,$35,$FB,$DC,$B7,$0D,$1F,$33,$C2,$7C,$DE,$27,$E3,$46,$1B
	.dc.b	$1F,$4B,$E7,$1E,$E7,$18,$C3,$F5,$BE,$75,$F2,$97,$E5,$70,$DC,$3E
	.dc.b	$94,$A7,$EE,$D3,$C3,$D6,$DF,$1A,$31,$7F,$65,$B1,$F2,$BE,$79,$F3
	.dc.b	$FC,$AF,$F0,$F1,$8B,$FC,$75,$7E,$DF,$E6,$3F,$2A,$7E,$77,$D4,$FE
	.dc.b	$5F,$F4,$18,$AD,$F8,$2A,$FC,$DE,$79,$C3,$F0,$C7,$F0,$17,$1E,$0F
	.dc.b	$EB,$9B,$DF,$FA,$D6,$CF,$EA,$9A,$F1,$C3,$BD,$F6,$69,$E2,$9D,$E1
	.dc.b	$F4,$23,$35,$E2,$8C,$36,$B9,$36,$79,$1C,$30,$F3,$37,$0F,$33,$A9
	.dc.b	$E9,$87,$80,$E7,$5C,$23,$A1,$A9,$E1,$19,$3F,$87,$DF,$0B,$EB,$5B
	.dc.b	$38,$CE,$E3,$2D,$8F,$E1,$F3,$C5,$F9,$15,$DA,$BF,$8C,$4E,$67,$E5
	.dc.b	$1D,$6B,$C3,$5B,$8F,$F6,$B6,$1F,$1A,$7F,$D7,$9F,$02,$F8,$0F,$E5
	.dc.b	$75,$C7,$1F,$F5,$23,$40,$9E,$1E,$1F,$58,$C5,$72,$BF,$54,$62,$75
	.dc.b	$5C,$23,$A1,$92,$B9,$9E,$2D,$BF,$F6,$86,$1E,$97,$CB,$3A,$C3,$FD
	.dc.b	$96,$FB,$1F,$14,$7F,$BF,$0F,$F6,$D9,$C0,$C5,$FF,$B2,$E7,$FD,$D5
	.dc.b	$9F,$FB,$6E,$3F,$D9,$AD,$8C,$E5,$FE,$D9,$CB,$D7,$87,$6B,$FF,$76
	.dc.b	$1F,$ED,$B7,$C7,$87,$78,$CD,$F9,$6B,$F6,$79,$7F,$B7,$27,$FD,$B3
	.dc.b	$97,$FB,$6E,$3F,$DC,$79,$78,$70,$E5,$5F,$7B,$0E,$4F,$EC,$AB,$E6
	.dc.b	$3F,$ED,$C3,$EB,$B8,$7E,$D9,$FE,$D8,$3F,$DB,$66,$BE,$0F,$FD,$BE
	.dc.b	$E0,$FF,$6F,$1F,$EE,$08,$46,$A3,$94,$FE,$2A,$3F,$D5,$6F,$FA,$A6
	.dc.b	$3F,$D5,$75,$FE,$A6,$9C,$B7,$E0,$FF,$D5,$B3,$FE,$BA,$6B,$FE,$D4
	.dc.b	$9D,$B9,$39,$96,$BB,$12,$BA,$46,$8E,$55,$D2,$FF,$DC,$A7,$FA,$DE
	.dc.b	$3F,$D6,$87,$FA,$E0,$FF,$66,$D7,$37,$B5,$7C,$1B,$F8,$B3,$FA,$8F
	.dc.b	$83,$FF,$5D,$CF,$FB,$0E,$BE,$7C,$61,$A7,$FB,$AC,$FD,$43,$96,$99
	.dc.b	$B4,$65,$FF,$74,$57,$CD,$C3,$F0,$F3,$93,$58,$79,$30,$F3,$4E,$B0
	.dc.b	$87,$E1,$67,$4F,$DF,$8C,$DF,$AE,$3C,$57,$D5,$CD,$FE,$06,$3C,$9F
	.dc.b	$39,$99,$FC,$84,$5A,$FE,$17,$36,$B8,$57,$23,$F2,$71,$97,$E2,$9C
	.dc.b	$8F,$D1,$B9,$7E,$29,$93,$F3,$CE,$71,$F8,$F1,$D4,$FE,$29,$19,$BD
	.dc.b	$3E,$88,$D4,$F0,$79,$9E,$03,$33,$B7,$27,$86,$FE,$33,$87,$DE,$F0
	.dc.b	$39,$19,$BF,$82,$F1,$7F,$55,$7D,$67,$99,$23,$33,$CA,$30,$FC,$67
	.dc.b	$BB,$D4,$F9,$66,$E3,$FD,$61,$F4,$2F,$90,$FC,$97,$85,$FE,$16,$FA
	.dc.b	$2B,$CD,$FE,$32,$C3,$FC,$84,$D7,$FC,$1D,$1C,$3A,$BF,$C3,$B1,$FE
	.dc.b	$0F,$8E,$7C,$07,$4E,$F9,$6C,$FF,$87,$23,$0F,$F8,$AB,$E4,$BC,$7D
	.dc.b	$DF,$E0,$B8,$FF,$29,$35,$FF,$21,$F7,$A7,$D3,$E0,$FF,$B3,$E7,$91
	.dc.b	$0F,$F2,$07,$C0,$6B,$E1,$B6,$C8,$FF,$20,$47,$C2,$F0,$DF,$13,$F1
	.dc.b	$3C,$6F,$F5,$45,$7F,$93,$BD,$6D,$4F,$C8,$F2,$BF,$C9,$73,$FE,$59
	.dc.b	$8C,$BF,$E6,$2F,$23,$59,$6D,$91,$FE,$66,$8F,$A3,$E7,$7F,$6C,$8F
	.dc.b	$F2,$FF,$98,$F4,$7E,$AB,$63,$FC,$E9,$3B,$EB,$FF,$40,$C7,$FA,$10
	.dc.b	$E7,$E1,$47,$FC,$F5,$E8,$7F,$3A,$9D,$F5,$3E,$4D,$EB,$FE,$0A,$B5
	.dc.b	$DF,$FE,$91,$C3,$F0,$B5,$A7,$FC,$8B,$F4,$7E,$07,$FC,$7A,$BD,$CF
	.dc.b	$E3,$0D,$8D,$D6,$3E,$1F,$2C,$8E,$71,$C3,$D9,$FF,$04,$C7,$FB,$0A
	.dc.b	$93,$8F,$67,$FD,$7F,$18,$9F,$F6,$6B,$FE,$C5,$6F,$C1,$EF,$CD,$E7
	.dc.b	$F4,$9F,$EC,$B9,$FF,$66,$B6,$F9,$D6,$E7,$FC,$33,$7F,$EF,$A7,$99
	.dc.b	$0F,$27,$C0,$6D,$46,$DC,$9E,$48,$D4,$3C,$0E,$DC,$2E,$E7,$D1,$E3
	.dc.b	$E0,$3F,$B4,$E3,$FE,$F8,$F3,$F7,$5A,$8F,$F1,$6E,$8D,$BF,$C5,$F8
	.dc.b	$F3,$C7,$96,$12,$F7,$5A,$C3,$94,$30,$E4,$B0,$E4,$F0,$31,$3C,$27
	.dc.b	$73,$A9,$E5,$5E,$2F,$80,$FF,$00,$56,$BE,$0B,$EA,$56,$8E,$CB,$7F
	.dc.b	$86,$D8,$F0,$60,$8E,$35,$CF,$8B,$77,$2F,$75,$AC,$7E,$1F,$C2,$F3
	.dc.b	$9F,$06,$2F,$84,$FD,$76,$5E,$A6,$DF,$0C,$7E,$AE,$1C,$6B,$DE,$D9
	.dc.b	$D6,$1E,$68,$FC,$04,$FC,$AF,$5F,$FD,$CF,$F4,$BF,$7B,$E2,$3C,$97
	.dc.b	$F3,$5B,$3A,$8F,$9C,$D7,$E3,$36,$BA,$78,$0B,$FC,$87,$4F,$D8,$BF
	.dc.b	$98,$3E,$33,$E6,$5F,$F9,$CB,$E3,$3F,$84,$4F,$F0,$BF,$84,$FD,$6C
	.dc.b	$63,$C4,$46,$1F,$D8,$DF,$2A,$57,$28,$C3,$C6,$B8,$4F,$FA,$02,$B9
	.dc.b	$BE,$17,$9E,$FF,$8F,$8E,$71,$8C,$35,$1C,$6B,$9B,$E7,$32,$7B,$6B
	.dc.b	$95,$F8,$DB,$7F,$C2,$0F,$0C,$F6,$3A,$1C,$B0,$F4,$C7,$86,$34,$77
	.dc.b	$2B,$CA,$BB,$61,$C6,$FD,$55,$FE,$64,$BF,$8C,$D8,$CB,$D4,$D7,$5B
	.dc.b	$5F,$F3,$44,$7D,$78,$C3,$FE,$6B,$A4,$62,$38,$CF,$07,$FD,$5B,$E1
	.dc.b	$54,$D2,$3B,$7E,$4E,$1F,$CA,$B8,$7E,$5C,$F2,$3E,$0C,$65,$F8,$66
	.dc.b	$72,$FC,$4C,$65,$FC,$5C,$67,$3F,$E6,$9D,$DF,$C2,$A7,$33,$E4,$1F
	.dc.b	$11,$CA,$7F,$87,$BA,$C7,$FD,$8E,$E7,$1B,$87,$FA,$36,$7D,$35,$D6
	.dc.b	$31,$7E,$88,$C3,$FD,$74,$C5,$7B,$9F,$16,$1D,$5F,$C2,$64,$F8,$A7
	.dc.b	$C6,$7F,$22,$D3,$F7,$51,$8C,$3F,$91,$5B,$E3,$1E,$4E,$E7,$FD,$BF
	.dc.b	$8F,$3C,$71,$7F,$7F,$0E,$4F,$FA,$AA,$75,$5F,$EA,$96,$C6,$67,$6F
	.dc.b	$F6,$2E,$97,$47,$F4,$9B,$FE,$21,$E0,$F6,$9C,$FC,$65,$AD,$5F,$F2
	.dc.b	$06,$BC,$DA,$FF,$26,$AF,$E1,$4D,$BF,$F5,$E5,$FF,$31,$E3,$CC,$FF
	.dc.b	$38,$6C,$66,$77,$D9,$F9,$86,$CF,$96,$3C,$7B,$FF,$43,$8C,$61,$E9
	.dc.b	$EE,$FC,$2D,$FF,$40,$6E,$3F,$CE,$1F,$F0,$64,$67,$97,$81,$E2,$8D
	.dc.b	$CF,$C7,$6B,$FD,$4E,$34,$7C,$7F,$09,$E7,$72,$BF,$E9,$F1,$87,$EF
	.dc.b	$FC,$67,$F8,$91,$BE,$57,$DC,$C1,$2F,$83,$C3,$B8,$FE,$66,$D7,$FD
	.dc.b	$D2,$D7,$FC,$3A,$D4,$74,$9F,$9E,$BE,$07,$37,$FC,$21,$B1,$FE,$F1
	.dc.b	$6B,$FE,$F6,$6B,$FE,$2A,$5E,$6F,$46,$CF,$38,$C1,$9E,$F9,$1D,$DB
	.dc.b	$5D,$FE,$2F,$FC,$0F,$3D,$DB,$1D,$DA,$FB,$2D,$18,$E3,$79,$DE,$2A
	.dc.b	$FE,$B5,$18,$C7,$53,$9C,$3F,$44,$DA,$FF,$0F,$BF,$D6,$23,$E0,$D7
	.dc.b	$4C,$78,$92,$33,$E7,$EE,$FB,$B6,$3F,$5C,$D8,$FF,$15,$B5,$FF,$12
	.dc.b	$7A,$4F,$53,$D4,$DA,$F0,$B6,$7D,$6D,$8F,$DD,$2F,$8A,$E2,$48,$E2
	.dc.b	$FC,$5A,$7C,$55,$24,$FB,$A3,$2F,$C0,$F5,$DE,$AF,$CD,$6C,$66,$30
	.dc.b	$66,$64,$71,$6E,$3F,$4B,$A1,$F1,$5F,$5D,$B5,$F3,$9B,$3B,$9F,$B9
	.dc.b	$EF,$3E,$2A,$92,$FE,$AC,$65,$BF,$15,$C0,$8E,$0F,$B7,$EF,$7E,$B5
	.dc.b	$BD,$3F,$89,$8C,$63,$F8,$86,$CF,$E8,$7D,$FF,$F2,$DF,$99,$FE,$95
	.dc.b	$E9,$46,$3C,$FF,$E6,$FC,$EF,$89,$C6,$5F,$97,$AA,$FE,$C5,$FC,$DB
	.dc.b	$71,$FD,$6E,$1D,$1B,$3A,$8C,$4F,$F7,$4C,$3F,$65,$97,$49,$CE,$5E
	.dc.b	$58,$C5,$E5,$B8,$7F,$7E,$BE,$2D,$C3,$CC,$D7,$F8,$34,$F0,$7C,$07
	.dc.b	$EC,$5C,$5F,$79,$E1,$87,$8A,$FA,$DF,$03,$DC,$58,$5F,$29,$DD,$FC
	.dc.b	$DC,$3D,$18,$74,$6E,$1C,$5B,$3B,$9F,$0D,$E2,$F9,$9F,$C5,$EB,$17
	.dc.b	$E9,$AD,$9D,$4A,$F9,$6B,$6F,$A2,$FD,$CE,$67,$71,$E1,$8C,$11,$D9
	.dc.b	$92,$DF,$92,$3C,$47,$AE,$79,$C7,$C8,$7F,$98,$57,$E0,$48,$FE,$19
	.dc.b	$FC,$5F,$C3,$BE,$27,$F9,$9D,$F8,$6F,$FD,$BF,$39,$EC,$EB,$3F,$D1
	.dc.b	$23,$0E,$CC,$57,$2C,$7C,$F1,$8A,$E2,$AB,$8C,$EA,$31,$5C,$6F,$DD
	.dc.b	$18,$BE,$37,$C2,$72,$D9,$F8,$77,$FF,$00,$63,$11,$F9,$2C,$7F,$29
	.dc.b	$18,$26,$1D,$E7,$8D,$FC,$08,$C4,$78,$4F,$C9,$4F,$4C,$FF,$23,$7D
	.dc.b	$2B,$73,$D3,$1F,$3C,$60,$8E,$0E,$04,$BF,$94,$77,$C7,$D5,$3C,$23
	.dc.b	$35,$D5,$F1,$C7,$9B,$0F,$1C,$62,$FA,$5F,$27,$C1,$19,$24,$7D,$1B
	.dc.b	$E8,$EE,$3B,$DF,$86,$30,$6D,$C6,$3F,$1E,$FA,$CE,$A3,$15,$CA,$7C
	.dc.b	$71,$89,$E5,$18,$9F,$C3,$1B,$3C,$B3,$B3,$EF,$D7,$82,$37,$5C,$9D
	.dc.b	$98,$9D,$D7,$58,$F9,$91,$87,$66,$0F,$C0,$9C,$48,$64,$FC,$51,$C2
	.dc.b	$FE,$3C,$73,$AF,$81,$3C,$30,$FA,$71,$88,$E5,$5F,$89,$CB,$F6,$38
	.dc.b	$FA,$E7,$B5,$F0,$C3,$E0,$47,$8A,$F8,$65,$EB,$7C,$31,$E8,$3A,$DF
	.dc.b	$3B,$E5,$1C,$63,$18,$78,$32,$E5,$1F,$89,$8C,$4F,$E1,$4F,$F2,$09
	.dc.b	$F5,$0A,$7E,$8C,$B7,$FC,$09,$BE,$CB,$E2,$7F,$D9,$72,$8C,$4F,$23
	.dc.b	$F8,$13,$87,$8C,$F0,$8D,$E9,$FB,$B6,$E9,$FC,$51,$B5,$FC,$4F,$E2
	.dc.b	$7C,$F5,$BB,$FD,$FD,$6A,$BB,$C7,$F0,$2A,$DD,$F3,$8D,$92,$7F,$7F
	.dc.b	$3A,$7F,$90,$57,$3C,$3D,$95,$BC,$3F,$88,$5E,$EF,$B9,$FC,$22,$F8
	.dc.b	$1C,$63,$57,$CA,$75,$7C,$F0,$FE,$63,$5A,$BE,$33,$9A,$F8,$05,$3B
	.dc.b	$E1,$F0,$5F,$49,$9C,$3B,$46,$4F,$3C,$66,$BA,$BC,$0F,$19,$23,$86
	.dc.b	$1E,$48,$E1,$7D,$70,$F3,$38,$8E,$55,$C1,$F9,$31,$B7,$51,$97,$FA
	.dc.b	$24,$F2,$7C,$15,$F5,$5E,$6D,$9E,$18,$7B,$23,$C9,$3C,$0E,$E5,$3A
	.dc.b	$9E,$07,$91,$C2,$79,$E1,$F0,$9F,$8B,$3C,$6B,$B3,$D2,$32,$70,$C3
	.dc.b	$D9,$1B,$AE,$57,$D4,$EF,$1C,$5A,$F9,$0E,$13,$CA,$BA,$47,$87,$0F
	.dc.b	$89,$18,$3C,$35,$CB,$0F,$2D,$7E,$0A,$3C,$0F,$72,$19,$3F,$12,$78
	.dc.b	$0E,$C7,$E9,$4F,$8C,$6A,$7B,$BC,$A7,$B8,$AF,$CD,$9E,$6A,$F2,$38
	.dc.b	$9E,$67,$BE,$B9,$BE,$32,$57,$37,$C4,$79,$63,$71,$98,$6B,$FC,$4A
	.dc.b	$BF,$8B,$B6,$3F,$D9,$F3,$B8,$C4,$7E,$28,$F6,$5F,$F8,$7B,$E2,$7C
	.dc.b	$D1,$93,$CD,$19,$7C,$78,$78,$B0,$FD,$CD,$70,$8C,$4F,$F8,$3A,$31
	.dc.b	$5F,$4A,$39,$9F,$E2,$98,$CB,$E2,$21,$FE,$25,$7A,$BE,$08,$E5,$82
	.dc.b	$73,$3E,$78,$F8,$95,$FE,$21,$9C,$BE,$49,$FF,$0C,$C6,$2B,$9D,$D9
	.dc.b	$F2,$1D,$CF,$1D,$F7,$7D,$86,$4F,$59,$E7,$32,$78,$E3,$A3,$93,$EB
	.dc.b	$9F,$2C,$A7,$80,$A6,$F0,$FE,$9B,$7C,$AB,$C5,$3F,$E8,$AF,$39,$F1
	.dc.b	$2F,$89,$FE,$6D,$C9,$FE,$84,$F8,$8F,$C1,$D8,$46,$5F,$F5,$DC,$69
	.dc.b	$CF,$02,$31,$2E,$8F,$F4,$7D,$7F,$B2,$2B,$31,$88,$F5,$CE,$E7,$3C
	.dc.b	$B2,$7F,$D9,$ED,$7F,$DB,$9F,$0F,$DB,$B0,$D2,$47,$FB,$AF,$D4,$F1
	.dc.b	$F8,$06,$92,$3F,$DE,$1F,$01,$A2,$38,$3F,$DE,$DF,$0D,$EC,$F0,$4D
	.dc.b	$24,$7C,$7E,$4F,$F0,$0E,$AF,$17,$FD,$F7,$39,$3E,$64,$67,$C3,$F8
	.dc.b	$F3,$A9,$F8,$6D,$AE,$6F,$A6,$39,$1F,$51,$C3,$C3,$2E,$67,$D2,$F1
	.dc.b	$BF,$D2,$B1,$F3,$9B,$5E,$66,$C7,$91,$AF,$E0,$3C,$8F,$F0,$CC,$FB
	.dc.b	$1B,$1E,$A6,$BB,$3F,$03,$F0,$CF,$2E,$92,$3D,$39,$A7,$51,$EC,$1F
	.dc.b	$08,$ED,$FC,$74,$61,$B1,$98,$E0,$D7,$F4,$6B,$DB,$FF,$14,$47,$E6
	.dc.b	$E3,$0E,$1A,$FF,$8B,$5A,$FF,$8E,$17,$C1,$F2,$8F,$2B,$63,$31,$CE
	.dc.b	$31,$EF,$C8,$EC,$D7,$B7,$AB,$C4,$8F,$8D,$AB,$D6,$FD,$FD,$A7,$FC
	.dc.b	$9B,$18,$6E,$DF,$E5,$09,$74,$62,$FF,$73,$CC,$F8,$4C,$6E,$F5,$3F
	.dc.b	$C9,$D5,$FB,$FC,$FF,$87,$CE,$7C,$7D,$E1,$C6,$B5,$9F,$E6,$A3,$11
	.dc.b	$FE,$5E,$A4,$62,$3F,$46,$D7,$86,$3F,$52,$76,$F6,$24,$76,$3F,$88
	.dc.b	$FA,$FF,$E5,$FC,$7C,$BE,$C7,$D0,$9F,$8D,$8F,$EB,$21,$BF,$B9,$B3
	.dc.b	$CF,$14,$71,$BE,$7C,$1E,$38,$E9,$5F,$BA,$6B,$D6,$3D,$1E,$CF,$48
	.dc.b	$F8,$0D,$7F,$80,$B5,$FD,$FB,$5F,$E5,$6F,$1A,$F9,$ED,$C7,$78,$FD
	.dc.b	$68,$FC,$05,$7C,$3F,$15,$F0,$4F,$82,$FC,$EC,$61,$B1,$99,$FF,$44
	.dc.b	$4E,$5B,$19,$7E,$48,$8D,$38,$6E,$3F,$D4,$39,$BF,$A8,$BF,$D4,$BE
	.dc.b	$0B,$E7,$BF,$D1,$FE,$05,$F0,$5F,$EA,$21,$ED,$7F,$A9,$23,$38,$FF
	.dc.b	$4D,$9C,$BC,$5F,$F5,$4D,$F4,$84,$F0,$51,$98,$C4,$F2,$9D,$BE,$1A
	.dc.b	$F0,$3D,$35,$FD,$64,$E8,$FE,$BD,$7D,$FD,$EE,$77,$FE,$BA,$6C,$FF
	.dc.b	$63,$7C,$53,$BB,$E9,$8F,$A6,$BD,$97,$E0,$69,$93,$D3,$87,$8B,$1E
	.dc.b	$6D,$BF,$44,$ED,$F6,$4E,$AF,$96,$1F,$16,$BF,$D9,$AD,$7C,$67,$FB
	.dc.b	$1D,$C3,$EE,$EA,$FB,$CE,$5A,$F2,$6B,$F4,$9A,$FF,$B6,$BD,$0F,$E4
	.dc.b	$49,$A4,$7F,$93,$3E,$FA,$FE,$65,$C3,$F0,$67,$FC,$19,$39,$EC,$FF
	.dc.b	$0E,$E3,$F7,$67,$27,$D5,$3F,$91,$46,$B2,$FF,$71,$57,$1B,$E9,$5E
	.dc.b	$31,$AF,$E1,$F1,$F0,$C6,$2F,$8B,$4F,$E4,$18,$7F,$8B,$AB,$C5,$1C
	.dc.b	$F1,$E5,$3E,$18,$F8,$47,$96,$FA,$CF,$0C,$7C,$93,$E9,$3D,$71,$CB
	.dc.b	$1F,$C3,$72,$BE,$99,$F1,$7F,$E3,$EC,$7F,$98,$C6,$2B,$F8,$C2,$FA
	.dc.b	$CD,$47,$F8,$2A,$35,$8F,$EE,$B0,$E9,$86,$A3,$FC,$73,$3A,$C3,$AE
	.dc.b	$19,$BD,$E1,$FC,$97,$99,$F2,$79,$6C,$67,$EA,$31,$32,$38,$3B,$8C
	.dc.b	$65,$E3,$EA,$7C,$27,$F9,$5A,$3F,$C9,$71,$8B,$FF,$27,$47,$F2,$F9
	.dc.b	$D6,$3F,$CD,$9B,$5B,$9F,$77,$D4,$FF,$81,$6F,$F9,$E4,$E7,$4F,$EB
	.dc.b	$33,$9A,$F8,$9D,$B6,$7F,$AE,$B6,$FF,$C6,$6F,$F2,$B7,$FA,$33,$99
	.dc.b	$FE,$A3,$D3,$E4,$8E,$91,$FE,$3E,$6B,$CA,$34,$7C,$DF,$07,$FB,$85
	.dc.b	$74,$8C,$77,$FD,$1F,$AA,$7C,$6D,$9C,$36,$BF,$AC,$4F,$6C,$3F,$CE
	.dc.b	$6D,$D8,$D3,$1F,$56,$FA,$0F,$BC,$DB,$E7,$8E,$27,$F0,$5F,$58,$FA
	.dc.b	$4B,$F1,$38,$6D,$FF,$69,$FA,$A7,$C2,$FD,$66,$C7,$DA,$D7,$FC,$BB
	.dc.b	$18,$7F,$C0,$8D,$7F,$CC,$BF,$1B,$F5,$E7,$FC,$C6,$D8,$FF,$47,$BF
	.dc.b	$90,$7F,$38,$E7,$67,$A0,$F8,$5F,$CD,$3E,$9B,$FE,$FC,$E5,$B9,$7E
	.dc.b	$7F,$D2,$C6,$3F,$E1,$76,$BF,$E6,$3B,$FC,$8C,$62,$3F,$CE,$FE,$53
	.dc.b	$3F,$D4,$37,$0F,$F0,$63,$F3,$67,$F6,$4E,$5F,$DA,$7A,$8F,$9B,$FD
	.dc.b	$83,$FE,$8A,$AF,$F0,$A3,$96,$FA,$FF,$B5,$F6,$3F,$7B,$7F,$E8,$C6
	.dc.b	$BF,$E8,$A6,$BF,$EB,$66,$EC,$F9,$4F,$F5,$05,$FF,$88,$5B,$5F,$C1
	.dc.b	$D7,$B3,$FC,$0D,$CF,$6F,$F9,$0B,$DA,$7C,$6F,$F0,$77,$CD,$7F,$C3
	.dc.b	$DC,$E1,$FE,$2D,$73,$5F,$E4,$AF,$5D,$F3,$DF,$BC,$6D,$7E,$1E,$FF
	.dc.b	$D9,$B1,$87,$FC,$94,$D7,$FD,$57,$F2,$CF,$8D,$FE,$30,$DA,$FF,$1E
	.dc.b	$2F,$8D,$CC,$FF,$1D,$F8,$37,$D5,$DF,$8E,$C3,$70,$FF,$22,$7C,$2F
	.dc.b	$FA,$5D,$F2,$71,$23,$19,$FF,$97,$5A,$FF,$AF,$9A,$FF,$AF,$3E,$63
	.dc.b	$F9,$B5,$7F,$93,$DB,$1F,$CD,$57,$CA,$7F,$31,$73,$C1,$FE,$FC,$F9
	.dc.b	$AF,$F4,$7D,$FF,$98,$6D,$18,$7F,$C6,$E7,$E8,$4F,$E7,$C6,$BE,$02
	.dc.b	$C6,$E3,$F6,$8D,$7F,$97,$B5,$FF,$7F,$35,$E5,$3F,$E7,$1F,$B3,$7D
	.dc.b	$A3,$EB,$FA,$E1,$FD,$36,$73,$1F,$D2,$FD,$27,$D8,$DF,$B1,$C3,$71
	.dc.b	$FF,$39,$4A,$71,$97,$F5,$98,$E3,$3A,$9F,$EA,$DF,$66,$F9,$0A,$63
	.dc.b	$EC,$ED,$7F,$9F,$67,$4F,$F9,$FF,$ED,$3F,$08,$E6,$BF,$B0,$46,$31
	.dc.b	$FF,$30,$E8,$F9,$B7,$D2,$7B,$78,$3F,$C1,$3C,$4C,$98,$30,$E1,$7B
	.dc.b	$F8,$39,$5F,$27,$0E,$2F,$0A,$DB,$99,$DD,$F8,$AB,$6F,$5C,$7C,$75
	.dc.b	$B3,$FD,$35,$E3,$BE,$6B,$FB,$D3,$63,$3F,$1D,$89,$91,$C3,$EB,$2E
	.dc.b	$1F,$E0,$7B,$47,$F8,$0F,$ED,$7F,$98,$6D,$FE,$C2,$5F,$01,$9F,$81
	.dc.b	$D3,$31,$FF,$56,$4E,$9B,$39,$C7,$F2,$FE,$8B,$D6,$F9,$51,$FE,$49
	.dc.b	$F6,$7F,$C0,$33,$9B,$FF,$22,$C6,$3E,$37,$7F,$11,$FE,$FD,$AD,$5F
	.dc.b	$13,$E9,$B8,$BF,$EA,$4D,$8E,$B8,$78,$27,$2D,$8C,$C7,$81,$E7,$F1
	.dc.b	$5E,$0C,$73,$1D,$CF,$EA,$11,$95,$E2,$3E,$D9,$F2,$D9,$8C,$7C,$FD
	.dc.b	$9C,$E3,$DE,$72,$D8,$4E,$CE,$15,$67,$CD,$D2,$FB,$7F,$87,$B3,$4F
	.dc.b	$C7,$46,$1B,$ED,$F9,$EB,$86,$D6,$B5,$AF,$E6,$FA,$9F,$1F,$9A,$E1
	.dc.b	$18,$CF,$E3,$57,$F7,$1F,$6F,$E1,$3F,$DC,$FE,$06,$F1,$FF,$61,$DB
	.dc.b	$EC,$46,$2B,$2D,$BE,$75,$FD,$A6,$B7,$18,$9D,$5F,$D1,$AD,$6D,$FA
	.dc.b	$BF,$77,$FA,$CD,$FC,$46,$D7,$E1,$76,$47,$F6,$0C,$7E,$34,$FE,$E6
	.dc.b	$37,$18,$9D,$FB,$3F,$67,$D2,$FF,$44,$CE,$6B,$78,$FF,$02,$E5,$FE
	.dc.b	$A9,$E0,$FC,$DF,$B8,$F8,$13,$FB,$F7,$CF,$9F,$F0,$77,$2D,$DB,$81
	.dc.b	$35,$FB,$B3,$AD,$4F,$2A,$7C,$EF,$F0,$1F,$A1,$FA,$6D,$8F,$8D,$E9
	.dc.b	$7D,$48,$C6,$DF,$02,$78,$F8,$6B,$E2,$D8,$E3,$96,$26,$47,$E8,$6F
	.dc.b	$9E,$9F,$B8,$C7,$B6,$1C,$71,$ED,$5A,$BF,$E4,$ED,$7F,$C7,$15,$C3
	.dc.b	$0E,$B5,$C2,$F9,$61,$D7,$1C,$BF,$E8,$6C,$79,$1C,$F8,$BF,$BB,$4E
	.dc.b	$B2,$EF,$7C,$9B,$8F,$E3,$B2,$F0,$47,$AE,$74,$FC,$49,$D5,$73,$8C
	.dc.b	$1B,$8C,$B4,$99,$7A,$63,$2D,$8C,$E5,$E9,$BE,$F8,$FA,$27,$27,$98
	.dc.b	$A7,$C6,$C3,$C5,$97,$C0,$AE,$39,$7A,$B1,$EC,$4C,$7E,$46,$1C,$AF
	.dc.b	$86,$5E,$49,$F1,$3D,$30,$E7,$8D,$8F,$81,$97,$97,$1E,$98,$F9,$5A
	.dc.b	$F9,$4E,$D7,$F2,$B1,$FA,$99,$79,$72,$EF,$5E,$CC,$3A,$46,$30,$E7
	.dc.b	$7C,$32,$F1,$5E,$F0,$E4,$F8,$6F,$81,$EE,$BE,$38,$F8,$4F,$5C,$F9
	.dc.b	$32,$FF,$27,$3F,$68,$FF,$4B,$63,$FA,$76,$CF,$E8,$5C,$F8,$8F,$A1
	.dc.b	$E3,$48,$C4,$FF,$96,$63,$13,$A9,$CC,$F2,$8F,$DA,$C7,$0C,$3F,$4A
	.dc.b	$DC,$7F,$39,$F4,$3F,$EE,$29,$CD,$70,$25,$FE,$C6,$75,$5F,$A3,$BE
	.dc.b	$53,$FA,$E6,$BF,$C1,$9B,$82,$B8,$CE,$EF,$AE,$1E,$08,$EC,$63,$0E
	.dc.b	$D3,$63,$F7,$75,$C9,$ED,$19,$69,$C8,$FE,$11,$87,$88,$91,$FC,$62
	.dc.b	$79,$61,$F2,$AB,$73,$CE,$3D,$11,$6B,$8B,$D7,$0F,$41,$30,$F3,$C6
	.dc.b	$5B,$19,$8F,$80,$64,$AE,$5F,$4B,$FC,$6B,$0F,$2D,$78,$23,$57,$E4
	.dc.b	$AF,$39,$23,$35,$C3,$0F,$41,$EF,$9A,$6F,$0F,$C0,$B8,$AE,$2D,$8F
	.dc.b	$49,$A7,$81,$E8,$72,$FD,$43,$81,$93,$07,$03,$DA,$F0,$7F,$A5,$D7
	.dc.b	$9D,$D5,$FD,$3A,$F8,$51,$D0,$E2,$74,$7E,$1E,$1F,$16,$FC,$D1,$FD
	.dc.b	$AF,$DC,$FD,$1B,$71,$FC,$8F,$02,$73,$5F,$E6,$5F,$B9,$7C,$5E,$9B
	.dc.b	$97,$F8,$82,$B6,$FF,$9C,$5B,$69,$DC,$FF,$80,$7E,$5F,$FD,$5E,$FF
	.dc.b	$8B,$FA,$3F,$D3,$4D,$F5,$3C,$73,$A8,$E1,$EC,$60,$8F,$F8,$C2,$3F
	.dc.b	$CC,$EF,$F7,$0C,$BF,$32,$47,$FC,$4B,$F3,$7E,$64,$DE,$0F,$2E,$FF
	.dc.b	$D9,$AB,$F5,$53,$B3,$4F,$3A,$FF,$28,$47,$BB,$EE,$FF,$DD,$71,$FE
	.dc.b	$42,$FB,$BF,$F4,$C3,$7D,$0F,$9B,$3A,$8F,$F1,$13,$70,$FF,$0F,$37
	.dc.b	$87,$FD,$28,$D9,$FF,$4E,$3F,$E5,$27,$FC,$A9,$F6,$BF,$C2,$3B,$BF
	.dc.b	$CB,$CD,$F2,$4D,$F6,$65,$70,$21,$FE,$77,$AF,$F3,$34,$7C,$0C,$3E
	.dc.b	$0F,$C8,$4A,$F9,$31,$97,$FD,$5F,$DB,$FE,$B0,$7F,$D6,$3B,$7F,$86
	.dc.b	$E3,$11,$F3,$A7,$51,$89,$D4,$F1,$7F,$C4,$5A,$7F,$88,$DB,$59,$3E
	.dc.b	$9F,$83,$EF,$F3,$DE,$D7,$87,$D2,$FF,$52,$37,$2F,$F2,$14,$EB,$2F
	.dc.b	$F5,$1F,$C1,$E5,$FF,$42,$C2,$31,$B2,$38,$4B,$0E,$AF,$23,$FD,$7F
	.dc.b	$F7,$7C,$3D,$6F,$CB,$4F,$FA,$AA,$D1,$8C,$7F,$CA,$8A,$B8,$12,$F9
	.dc.b	$9F,$EC,$A9,$F3,$76,$B9,$B9,$4F,$FB,$75,$B1,$9E,$AD,$FD,$E3,$CA
	.dc.b	$E3,$5A,$24,$EA,$25,$FF,$BA,$34,$FF,$60,$E1,$CE,$31,$8F,$FB,$97
	.dc.b	$0E,$F8,$F5,$9F,$F7,$E7,$AD,$F2,$3D,$97,$81,$4F,$97,$ED,$FF,$B5
	.dc.b	$AB,$64,$3E,$66,$3F,$E7,$78,$D4,$60,$FF,$5C,$AF,$63,$EA,$F2,$3E
	.dc.b	$0A,$92,$BC,$E7,$FB,$02,$1F,$05,$C0,$D4,$75,$E4,$FC,$4D,$66,$38
	.dc.b	$8F,$AD,$79,$DF,$2B,$DC,$93,$A9,$CB,$67,$FC,$36,$F1,$6C,$6A,$3F
	.dc.b	$46,$64,$FF,$10,$B9,$E3,$FF,$78,$C6,$4F,$84,$57,$3D,$8F,$1B,$F2
	.dc.b	$FF,$06,$F8,$BA,$4B,$F8,$7A,$27,$F2,$71,$8F,$45,$EF,$FF,$8C,$A9
	.dc.b	$18,$9F,$3C,$62,$75,$39,$9F,$DE,$CB,$E1,$5E,$BF,$EA,$97,$AD,$4C
	.dc.b	$47,$EB,$BE,$1F,$37,$E6,$6B,$F5,$5B,$B7,$EA,$DE,$B1,$EC,$F0,$FF
	.dc.b	$8E,$BE,$66,$BE,$46,$BF,$C5,$71,$FD,$F7,$93,$EF,$F4,$7F,$C9,$13
	.dc.b	$F1,$97,$D5,$7F,$02,$6C,$67,$D4,$FE,$0B,$F1,$7C,$23,$F1,$4D,$CF
	.dc.b	$F8,$03,$7B,$7F,$DA,$4B,$D7,$FE,$1B,$F1,$9F,$4A,$FE,$4B,$5F,$33
	.dc.b	$5F,$F6,$D3,$56,$8F,$8D,$F2,$C7,$E6,$D7,$D1,$7F,$1A,$6C,$7E,$09
	.dc.b	$DE,$5F,$C5,$7E,$3D,$F0,$1F,$C8,$1B,$5F,$88,$6B,$FA,$75,$F0,$7F
	.dc.b	$C9,$BE,$03,$F3,$F9,$7F,$BE,$AB,$F1,$51,$83,$3F,$46,$FA,$7F,$E5
	.dc.b	$AB,$E5,$29,$88,$FE,$63,$F2,$9F,$CC,$EF,$F3,$2D,$7F,$70,$DD,$3F
	.dc.b	$97,$2F,$86,$FE,$73,$D5,$DE,$BF,$32,$D7,$F2,$ED,$7F,$08,$D4,$7C
	.dc.b	$EF,$95,$7C,$67,$F4,$47,$F6,$DC,$DA,$D9,$18,$AE,$15,$FD,$2B,$E5
	.dc.b	$BF,$2F,$1F,$C3,$DB,$9F,$F4,$56,$E1,$CF,$E5,$DF,$43,$FD,$49,$B7
	.dc.b	$FB,$26,$BF,$B0,$6B,$F9,$D6,$BC,$FE,$1F,$94,$C8,$FE,$07,$3F,$EA
	.dc.b	$2C,$3F,$AB,$73,$FF,$99,$DF,$EB,$1F,$32,$FA,$9F,$EC,$6D,$BF,$E1
	.dc.b	$75,$BC,$BE,$13,$FE,$6F,$8F,$D7,$35,$CF,$C5,$7F,$0D,$8C,$47,$E6
	.dc.b	$9E,$4D,$8F,$EE,$31,$8D,$FF,$37,$87,$F7,$4F,$9D,$7C,$07,$F7,$6F
	.dc.b	$9D,$7C,$25,$35,$13,$2F,$E3,$35,$C1,$B1,$FC,$7E,$31,$1F,$C6,$BE
	.dc.b	$7A,$ED,$FD,$E5,$BA,$FD,$F6,$CF,$EC,$97,$CC,$7F,$82,$1C,$E8,$69
	.dc.b	$23,$FC,$1B,$F4,$1F,$CD,$EB,$F9,$AB,$5F,$EA,$8D,$C3,$FC,$14,$FF
	.dc.b	$92,$A3,$3E,$87,$22,$3C,$BE,$2F,$FC,$38,$BE,$8F,$FA,$35,$FE,$73
	.dc.b	$E8,$DF,$31,$FE,$37,$3F,$3B,$59,$71,$D5,$FE,$3B,$9D,$CF,$E9,$2F
	.dc.b	$9A,$8C,$BD,$FD,$5F,$F1,$33,$9B,$E6,$F5,$3F,$5A,$F8,$1D,$1D,$63
	.dc.b	$8C,$F4,$AE,$FD,$3D,$AF,$83,$70,$FF,$28,$3B,$BF,$0B,$71,$F0,$63
	.dc.b	$F0,$2B,$D9,$7E,$48,$F0,$9E,$99,$DE,$1E,$EC,$7B,$4F,$2A,$F4,$47
	.dc.b	$58,$E3,$3F,$11,$FF,$2F,$B5,$FF,$32,$35,$F7,$57,$F8,$09,$FA,$DD
	.dc.b	$CF,$A4,$C5,$7F,$1B,$6C,$66,$7F,$C1,$33,$97,$FC,$1F,$F1,$9F,$E1
	.dc.b	$04,$D2,$3A,$3D,$F7,$FA,$AE,$3F,$8B,$1F,$C8,$27,$E3,$CE,$63,$E2
	.dc.b	$99,$38,$F2,$AB,$8C,$78,$0C,$B8,$AE,$73,$B9,$E6,$6F,$2F,$C4,$DF
	.dc.b	$77,$F1,$71,$DD,$CC,$F1,$73,$1C,$73,$FF,$40,$4E,$5F,$F1,$6D,$F8
	.dc.b	$30,$F1,$57,$90,$FC,$7F,$23,$89,$F4,$9E,$5C,$FF,$5E,$DB,$FE,$7D
	.dc.b	$2F,$A4,$C6,$1F,$DC,$A3,$8C,$FE,$EA,$B3,$87,$56,$CF,$19,$D1,$FE
	.dc.b	$34,$E6,$FF,$2E,$37,$E0,$78,$C3,$E1,$6C,$49,$D1,$C6,$31,$8F,$F9
	.dc.b	$6F,$A7,$FD,$47,$3F,$E4,$06,$C6,$6B,$CB,$18,$D3,$FD,$2C,$DA,$FE
	.dc.b	$B8,$DA,$FF,$07,$BE,$67,$B7,$C1,$7B,$BE,$91,$F1,$F4,$C7,$D2,$DC
	.dc.b	$3F,$CA,$33,$A6,$CE,$9A,$4B,$FF,$3A,$78,$6F,$94,$E5,$18,$7E,$6F
	.dc.b	$8B,$FD,$02,$BF,$AD,$F8,$38,$CB,$2E,$B1,$93,$A3,$FB,$0B,$F5,$BC
	.dc.b	$7F,$F2,$AC,$7F,$99,$9B,$8F,$FA,$17,$0F,$43,$FD,$E7,$DC,$FB,$BE
	.dc.b	$4F,$F8,$8A,$FF,$CD,$AD,$9F,$F4,$63,$FC,$BA,$33,$E8,$7E,$13,$CB
	.dc.b	$FF,$10,$47,$F9,$D1,$B5,$C3,$D2,$D7,$9F,$F8,$AF,$3B,$E9,$75,$9A
	.dc.b	$7D,$3E,$6F,$F1,$0F,$2F,$E4,$3E,$97,$FD,$2F,$9F,$FA,$01,$B9,$7F
	.dc.b	$A7,$DF,$F1,$FC,$67,$E1,$3F,$2F,$E9,$EB,$0F,$F6,$C3,$67,$FC,$3C
	.dc.b	$FB,$63,$33,$F9,$EF,$53,$FC,$FF,$5F,$E4,$7D,$B1,$C3,$97,$F4,$7E
	.dc.b	$AF,$FB,$8A,$7F,$CA,$5F,$23,$8F,$37,$2F,$EA,$3D,$7F,$D5,$ED,$FE
	.dc.b	$B5,$6E,$9F,$EE,$B7,$F5,$51,$9F,$0F,$F5,$FE,$CB,$E9,$E9,$8F,$90
	.dc.b	$B9,$FF,$9F,$7E,$9B,$F6,$BE,$F1,$BE,$A6,$1F,$DC,$FC,$0B,$E9,$A9
	.dc.b	$FB,$CF,$BD,$FE,$41,$87,$FA,$46,$3F,$7F,$F0,$9F,$EE,$08,$FF,$4D
	.dc.b	$7C,$7E,$3E,$9F,$2F,$F0,$4F,$87,$FF,$69,$CF,$FA,$D1,$BC,$3F,$ED
	.dc.b	$35,$F3,$1B,$F9,$0F,$E1,$7F,$14,$FA,$8A,$7F,$0E,$FB,$EF,$F2,$BE
	.dc.b	$3F,$EB,$B8,$FE,$23,$F1,$DF,$1B,$A7,$F9,$87,$8B,$FE,$C1,$F5,$F3
	.dc.b	$F3,$1A,$7F,$8D,$EC,$FA,$9F,$F0,$DE,$5F,$EE,$26,$C7,$FB,$4B,$C4
	.dc.b	$EE,$FF,$B5,$23,$38,$FF,$B6,$1B,$5F,$E6,$A9,$E9,$F0,$9D,$AF,$8D
	.dc.b	$EE,$3C,$37,$C2,$BF,$CE,$2D,$7F,$DD,$B7,$E3,$9F,$E5,$BE,$47,$C6
	.dc.b	$8F,$F6,$EA,$FA,$9C,$F1,$69,$B3,$A6,$DF,$FC,$01,$AE,$13,$FE,$E6
	.dc.b	$F9,$67,$D4,$7F,$3A,$6C,$67,$EA,$71,$32,$38,$3B,$8C,$5B,$4F,$EC
	.dc.b	$1F,$38,$F1,$FF,$DD,$F8,$7F,$A3,$A7,$3F,$4D,$FD,$BD,$DC,$EA,$F9
	.dc.b	$57,$FA,$52,$D7,$B8,$C5,$6E,$F9,$FC,$1F,$F5,$6B,$F0,$7C,$B5,$C3
	.dc.b	$86,$29,$E7,$59,$38,$61,$CF,$1F,$05,$F0,$E5,$79,$5F,$ED,$B9,$F5
	.dc.b	$1F,$EA,$37,$18,$FA,$0F,$F6,$BA,$F9,$4F,$EC,$AE,$B4,$FC,$AB,$6B
	.dc.b	$97,$CA,$6A,$B6,$FF,$6C,$F9,$FF,$E4,$13,$FE,$B1,$7F,$B3,$3E,$2C
	.dc.b	$3F,$DC,$B1,$A6,$FC,$1F,$1F,$83,$E4,$DA,$E1,$9A,$F9,$DF,$E7,$FD
	.dc.b	$8F,$F0,$C7,$C3,$7F,$5E,$8F,$F6,$2F,$5F,$F2,$A7,$FC,$5F,$EF,$7E
	.dc.b	$7E,$3F,$D8,$F8,$FF,$8A,$E3,$17,$C9,$B1,$9A,$E7,$39,$7A,$FC,$65
	.dc.b	$57,$DF,$0F,$14,$62,$B5,$39,$C3,$C5,$78,$9E,$99,$3E,$81,$F5,$18
	.dc.b	$5F,$53,$63,$0D,$D3,$CD,$5C,$63,$19,$FC,$01,$19,$9F,$67,$3B,$EA
	.dc.b	$7C,$95,$F8,$F6,$C6,$7E,$03,$E0,$3B,$CE,$CE,$89,$38,$8F,$7F,$B5
	.dc.b	$F0,$3D,$BB,$8F,$CB,$C7,$E3,$C6,$1F,$31,$88,$D7,$07,$F5,$88,$CB
	.dc.b	$FE,$00,$E1,$F9,$AE,$5B,$5C,$DB,$5E,$EF,$75,$EB,$FA,$9B,$19,$F6
	.dc.b	$31,$32,$38,$3B,$8C,$63,$F2,$F6,$7D,$4F,$87,$97,$F8,$03,$E5,$6E
	.dc.b	$BF,$22,$31,$5F,$BB,$78,$37,$4D,$4E,$DF,$C6,$7D,$33,$A9,$F5,$3F
	.dc.b	$26,$30,$FC,$6F,$43,$73,$AC,$3E,$5D,$69,$B1,$9E,$0F,$93,$5C,$2F
	.dc.b	$A4,$E7,$3F,$E8,$98,$71,$BD,$EF,$AB,$98,$F4,$7C,$51,$89,$FD,$0D
	.dc.b	$71,$BE,$78,$75,$C3,$F6,$B8,$F8,$5E,$44,$BE,$ED,$AE,$38,$F5,$7F
	.dc.b	$75,$5C,$6B,$BD,$F5,$B6,$1E,$17,$BE,$3C,$51,$A9,$E1,$3B,$8F,$DD
	.dc.b	$DF,$5C,$3C,$31,$E3,$39,$94,$E5,$8F,$57,$8E,$16,$FA,$1D,$78,$58
	.dc.b	$75,$AF,$1C,$EB,$1F,$55,$6A,$F9,$47,$98,$CD,$F1,$BF,$09,$E3,$AE
	.dc.b	$0F,$96,$FA,$61,$F5,$A3,$17,$CE,$31,$97,$CA,$7B,$98,$CB,$E0,$C6
	.dc.b	$6B,$B6,$3E,$1C,$3A,$C7,$ED,$A3,$18,$75,$8C,$65,$D1,$E2,$DB,$E8
	.dc.b	$7E,$00,$B8,$78,$09,$3F,$42,$32,$FC,$8A,$FD,$03,$FE,$1B,$3E,$19
	.dc.b	$70,$FD,$13,$63,$6F,$E7,$BB,$5F,$35,$FE,$FF,$AF,$EC,$4D,$8C,$CF
	.dc.b	$53,$F1,$AE,$1E,$CE,$AF,$F4,$2D,$C3,$B5,$7E,$E7,$E6,$BF,$C6,$15
	.dc.b	$A8,$FD,$E5,$6E,$FF,$06,$D7,$F3,$75,$C2,$F9,$C7,$EF,$A7,$58,$7A
	.dc.b	$23,$F8,$0D,$6D,$E7,$7D,$88,$E8,$B8,$78,$DE,$91,$E7,$73,$87,$82
	.dc.b	$74,$7E,$FA,$BB,$4F,$C4,$AD,$BE,$03,$B1,$D0,$93,$CA,$BF,$87,$D7
	.dc.b	$2B,$F5,$4E,$6B,$93,$DD,$EA,$77,$39,$D7,$27,$33,$CA,$3D,$06,$63
	.dc.b	$67,$47,$2F,$EC,$AB,$C5,$87,$CC,$26,$1F,$0E,$78,$9E,$1C,$3C,$CF
	.dc.b	$D6,$3C,$33,$C1,$F8,$31,$AC,$3C,$91,$E1,$3C,$07,$0A,$F4,$DF,$89
	.dc.b	$C4,$72,$C3,$CE,$F8,$5B,$4A,$E1,$3E,$68,$EE,$F9,$4C,$61,$E5,$C3
	.dc.b	$E6,$46,$30,$F0,$BE,$08,$F4,$5F,$47,$D5,$5F,$90,$8F,$C7,$70,$BE
	.dc.b	$A5,$F1,$98,$F7,$FF,$4D,$87,$F8,$D6,$91,$8D,$3F,$CD,$17,$C3,$2F
	.dc.b	$F0,$AB,$5F,$EE,$13,$B7,$FC,$C8,$DA,$F0,$61,$E2,$F9,$87,$A7,$49
	.dc.b	$7F,$E2,$08,$7D,$16,$23,$FB,$E4,$70,$75,$3C,$E7,$F3,$BF,$43,$E1
	.dc.b	$9F,$E9,$B7,$F9,$73,$3B,$C8,$FE,$ED,$19,$7F,$C4,$13,$C4,$E5,$C1
	.dc.b	$ED,$DB,$FB,$3D,$71,$AF,$F0,$77,$D4,$7F,$B7,$5F,$F1,$BB,$5C,$FB
	.dc.b	$FF,$42,$3F,$C8,$B1,$A8,$C1,$1E,$3E,$17,$F0,$3F,$0B,$FC,$9D,$EA
	.dc.b	$5F,$15,$F4,$54,$C4,$4B,$FF,$26,$CE,$9B,$39,$7F,$C7,$3A,$FF,$98
	.dc.b	$34,$FF,$30,$B7,$97,$FD,$9C,$D9,$FE,$E2,$BC,$FF,$9B,$32,$7E,$3B
	.dc.b	$F4,$23,$11,$DA,$2D,$76,$9F,$67,$92,$F8,$2F,$F6,$E7,$9D,$FE,$EC
	.dc.b	$87,$AB,$8E,$EF,$F4,$A3,$72,$FF,$49,$DD,$DF,$FD,$1C,$DB,$E3,$3F
	.dc.b	$E9,$26,$BF,$EA,$99,$DB,$FE,$96,$F3,$DF,$51,$CD,$B1,$98,$C6,$39
	.dc.b	$64,$64,$E5,$1E,$5E,$B7,$D4,$7E,$56,$7F,$D7,$11,$8E,$BD,$4E,$75
	.dc.b	$FF,$6D,$FD,$47,$F9,$82,$FF,$D6,$16,$8C,$47,$2A,$E1,$87,$FB,$65
	.dc.b	$B1,$FE,$D1,$9D,$DF,$4F,$2B,$BE,$32,$BA,$D6,$B0,$FF,$16,$E8,$F6
	.dc.b	$1E,$97,$CD,$6D,$FF,$BE,$E3,$1E,$9E,$59,$19,$3F,$E0,$04,$7D,$5E
	.dc.b	$27,$C0,$5F,$A7,$95,$FE,$C6,$9C,$B6,$72,$FD,$78,$4F,$39,$FC,$14
	.dc.b	$BE,$9F,$FD,$76,$FF,$B5,$CC,$79,$9A,$8C,$C7,$FB,$A4,$75,$3C,$5D
	.dc.b	$54,$9E,$55,$C3,$89,$BF,$E9,$2F,$55,$CE,$B5,$3F,$E5,$68,$F0,$9E
	.dc.b	$3C,$3B,$3F,$EE,$4F,$80,$7C,$4F,$E6,$5B,$19,$8C,$74,$7E,$77,$CD
	.dc.b	$7B,$74,$95,$FE,$EE,$87,$E0,$5F,$A3,$F8,$57,$A7,$49,$5F,$EF,$48
	.dc.b	$7D,$8F,$EA,$7E,$25,$EA,$D2,$57,$FB,$EA,$39,$BF,$EF,$CF,$0B,$F8
	.dc.b	$63,$FA,$B6,$F1,$FC,$68,$FF,$6C,$C6,$38,$7F,$1A,$BE,$0B,$F7,$1E
	.dc.b	$8B,$D4,$A4,$BF,$F3,$FE,$1C,$FB,$BF,$A8,$FA,$BF,$BF,$F8,$CE,$15
	.dc.b	$AC,$FF,$33,$18,$8F,$DF,$36,$F9,$3F,$EB,$58,$E0,$7F,$BB,$BE,$39
	.dc.b	$CB,$FC,$2D,$B1,$AC,$51,$88,$FE,$14,$BE,$37,$F8,$97,$C7,$BE,$A6
	.dc.b	$92,$B9,$47,$8F,$3F,$D7,$FC,$8B,$E0,$3F,$8D,$36,$BC,$0D,$7F,$8C
	.dc.b	$2F,$81,$FE,$41,$F0,$3F,$E6,$9D,$FE,$33,$70,$FD,$DB,$6D,$AB,$F0
	.dc.b	$07,$C6,$FF,$29,$FA,$A7,$D5,$52,$5F,$C1,$8F,$C8,$BF,$0F,$E5,$BF
	.dc.b	$9C,$37,$A7,$F9,$94,$62,$3F,$97,$2F,$A5,$FE,$6D,$DF,$DA,$7E,$63
	.dc.b	$5F,$E6,$8D,$DB,$F0,$1F,$33,$FA,$C9,$F4,$AF,$8B,$FE,$80,$D8,$FA
	.dc.b	$CD,$7F,$9F,$3F,$E8,$E9,$FC,$57,$CD,$FE,$6A,$7F,$A1,$36,$33,$7F
	.dc.b	$D2,$FE,$6D,$F4,$F4,$97,$F8,$38,$CB,$7E,$1B,$81,$1C,$1F,$89,$FC
	.dc.b	$05,$F5,$5F,$D5,$FE,$AF,$FC,$81,$3F,$8D,$FA,$2C,$B6,$75,$F1,$B2
	.dc.b	$3F,$25,$5C,$1F,$EB,$4D,$D3,$F4,$CF,$28,$CC,$F1,$69,$CA,$BF,$B4
	.dc.b	$7D,$27,$F7,$B8,$FC,$F3,$5F,$EC,$CD,$BF,$D9,$2F,$AD,$E3,$F0,$72
	.dc.b	$B8,$BD,$23,$FB,$A7,$D3,$71,$ED,$FE,$9A,$DA,$FC,$9B,$7B,$0B,$E6
	.dc.b	$B8,$E1,$2B,$F6,$53,$B3,$4F,$ED,$7E,$9F,$F9,$4B,$FD,$F5,$B9,$FF
	.dc.b	$09,$6E,$3F,$D7,$97,$CC,$FF,$82,$5C,$D7,$E9,$9B,$91,$A7,$08,$FE
	.dc.b	$9D,$F5,$5F,$D2,$70,$43,$EA,$9F,$55,$FD,$0C,$FC,$A3,$88,$DC,$60
	.dc.b	$FE,$91,$5C,$E3,$94,$63,$DC,$5F,$6B,$E5,$5D,$BB,$3F,$A8,$46,$2F
	.dc.b	$A4,$F9,$30,$E9,$3C,$1B,$3E,$5A,$E2,$DA,$EC,$F8,$6E,$C7,$3C,$3C
	.dc.b	$6F,$98,$F5,$D7,$26,$9C,$05,$F8,$1E,$55,$9B,$F3,$C7,$AC,$E1,$3B
	.dc.b	$78,$DF,$81,$FE,$DC,$62,$BF,$DF,$CD,$3F,$E0,$04,$FC,$B7,$6F,$F1
	.dc.b	$3F,$AB,$DB,$63,$7D,$C9,$CB,$63,$A8,$F8,$CA,$D3,$2E,$08,$FC,$53
	.dc.b	$EC,$F0,$FF,$65,$D3,$EC,$C7,$E2,$27,$E3,$63,$FE,$6D,$E0,$F9,$13
	.dc.b	$93,$94,$65,$E8,$66,$BF,$07,$19,$7A,$C7,$95,$CD,$F8,$74,$F9,$33
	.dc.b	$97,$FC,$E3,$19,$8F,$9B,$87,$92,$31,$1F,$E1,$D8,$C6,$1E,$68,$E4
	.dc.b	$7F,$78,$9D,$BE,$5A,$EB,$3F,$8A,$C1,$F6,$DF,$E2,$C8,$FE,$E7,$18
	.dc.b	$7F,$C1,$4F,$F8,$2E,$9F,$6F,$FD,$62,$77,$5C,$67,$F4,$B1,$FE,$08
	.dc.b	$8C,$57,$F0,$1F,$A3,$E7,$39,$2D,$7E,$FE,$2F,$17,$F1,$2F,$A2,$BC
	.dc.b	$A6,$FD,$16,$19,$87,$F1,$4F,$89,$B8,$FF,$A3,$6F,$FC,$D3,$3F,$CB
	.dc.b	$27,$AF,$53,$E2,$6C,$61,$B3,$FC,$C9,$B7,$FC,$C9,$F2,$FD,$6F,$F3
	.dc.b	$19,$FF,$0A,$36,$7F,$D3,$8F,$E1,$75,$FF,$14,$39,$6F,$67,$2F,$AD
	.dc.b	$E6,$DB,$E6,$DB,$E6,$DA,$FF,$19,$78,$8B,$F8,$ED,$BF,$EC,$95,$89
	.dc.b	$91,$FD,$9F,$EB,$6E,$1F,$EB,$1B,$47,$23,$9F,$CA,$DA,$FE,$DC,$FD
	.dc.b	$4F,$2B,$F9,$5C,$7F,$7E,$F4,$B1,$CF,$96,$47,$F8,$0B,$D4,$B9,$7F
	.dc.b	$80,$9F,$BB,$E5,$BE,$B7,$EF,$F9,$DF,$4E,$FF,$C2,$71,$8E,$2F,$C1
	.dc.b	$FD,$6F,$F8,$D6,$7F,$C8,$4B,$E3,$33,$8E,$9F,$C4,$FA,$DB,$CF,$FD
	.dc.b	$9C,$FF,$8D,$EF,$FC,$94,$E7,$93,$FC,$6F,$AF,$E4,$3D,$77,$D8,$5F
	.dc.b	$84,$C3,$70,$FF,$24,$F9,$7F,$96,$F6,$BF,$CA,$F1,$83,$FC,$9B,$EF
	.dc.b	$7E,$6B,$DE,$7C,$9D,$F1,$30,$DA,$FF,$2E,$7C,$07,$F9,$71,$BF,$29
	.dc.b	$FA,$28,$C7,$0F,$F9,$AA,$3F,$4B,$F0,$9F,$E9,$8A,$FF,$32,$3F,$E6
	.dc.b	$16,$FC,$66,$BE,$0F,$F5,$7F,$5A,$FC,$0A,$D6,$1A,$FF,$A3,$7E,$4B
	.dc.b	$F6,$3F,$14,$FA,$EB,$18,$7F,$6B,$F1,$AF,$9E,$B3,$86,$DF,$F9,$DF
	.dc.b	$EB,$5F,$4F,$FB,$C7,$FB,$AC,$66,$31,$C1,$FC,$03,$E5,$1F,$4D,$79
	.dc.b	$30,$DB,$FF,$66,$7D,$9F,$F1,$1B,$FF,$65,$37,$BF,$F8,$B3,$F8,$D9
	.dc.b	$DF,$C6,$FF,$0E,$A4,$EE,$7F,$D8,$DE,$17,$92,$79,$D6,$E3,$17,$D2
	.dc.b	$31,$3C,$EF,$93,$FE,$10,$AD,$DF,$FB,$EB,$E6,$9F,$43,$FC,$71,$B1
	.dc.b	$FE,$EF,$FA,$CC,$4C,$8F,$F7,$93,$B8,$C5,$FF,$2C,$3F,$D1,$3F,$3B
	.dc.b	$FE,$CE,$AF,$F5,$22,$F6,$7F,$94,$B9,$9E,$1E,$D2,$35,$5C,$3E,$63
	.dc.b	$77,$C1,$B1,$FE,$B4,$FA,$0F,$F1,$AC,$FF,$36,$6C,$67,$C5,$ED,$C1
	.dc.b	$FE,$CF,$9C,$E5,$E7,$6C,$66,$7F,$DD,$35,$C6,$FA,$46,$2F,$83,$E6
	.dc.b	$6F,$CD,$EA,$76,$FF,$46,$F1,$7F,$DB,$17,$FE,$C5,$6D,$7C,$07,$FD
	.dc.b	$85,$19,$F9,$8E,$04,$78,$7C,$BD,$AF,$8F,$F4,$5F,$D3,$E7,$FD,$A0
	.dc.b	$FF,$15,$8C,$FC,$07,$F5,$68,$C6,$9F,$EE,$57,$84,$F2,$F4,$B5,$5B
	.dc.b	$7F,$19,$F1,$53,$0F,$F7,$0D,$7F,$1B,$EC,$C3,$F7,$CF,$E3,$D3,$FE
	.dc.b	$FB,$F8,$57,$B5,$F9,$8F,$53,$0F,$E6,$A6,$D7,$E4,$E6,$BC,$B0,$FF
	.dc.b	$0F,$57,$16,$C7,$E6,$6B,$7D,$5F,$CA,$70,$EF,$3F,$ED,$8F,$A3,$74
	.dc.b	$F6,$E9,$46,$31,$E7,$18,$FA,$5D,$4E,$6F,$FC,$71,$9F,$F9,$DA,$B8
	.dc.b	$E1,$A3,$77,$70,$F2,$F4,$3E,$B3,$F9,$D6,$3C,$84,$67,$6F,$82,$EF
	.dc.b	$D6,$B5,$EB,$CF,$FD,$0B,$39,$C3,$AC,$E7,$0E,$6F,$6D,$7E,$0D,$75
	.dc.b	$AF,$8F,$5F,$D4,$7D,$BF,$8E,$7C,$8F,$9E,$B3,$AE,$04,$FC,$CF,$5F
	.dc.b	$37,$E9,$3F,$AD,$38,$C7,$E7,$C6,$2F,$95,$71,$7E,$1F,$B2,$F4,$FD
	.dc.b	$ED,$8C,$FD,$66,$26,$47,$07,$71,$8B,$F9,$5A,$3E,$B3,$CF,$7F,$41
	.dc.b	$67,$F3,$1C,$D7,$07,$33,$BC,$B5,$5B,$6C,$7C,$8F,$AB,$7D,$07,$F4
	.dc.b	$4C,$3E,$14,$63,$A7,$E4,$7D,$17,$E6,$23,$13,$AA,$E1,$BF,$C6,$9C
	.dc.b	$BF,$23,$81,$7C,$6B,$7C,$7F,$EC,$4C,$18,$70,$8C,$36,$3F,$9D,$57
	.dc.b	$1C,$7D,$17,$CA,$7F,$D9,$B5,$B7,$FC,$25,$5B,$7C,$D5,$C0,$F5,$E5
	.dc.b	$63,$24,$CB,$D1,$5C,$5F,$66,$1D,$E7,$CE,$DC,$BD,$B7,$CE,$B7,$85
	.dc.b	$CB,$FA,$DE,$3C,$DF,$6E,$1C,$F2,$F4,$63,$CB,$2F,$15,$6B,$0E,$51
	.dc.b	$F0,$B2,$AF,$8A,$2D,$71,$CB,$C3,$3C,$8F,$8B,$8F,$9F,$2F,$14,$62
	.dc.b	$F8,$C6,$23,$A1,$C4,$F9,$78,$78,$32,$F9,$C7,$A2,$78,$38,$BE,$D3
	.dc.b	$B7,$DC,$62,$BA,$4E,$EF,$E9,$C6,$27,$71,$8C,$3D,$64,$CB,$E2,$57
	.dc.b	$ED,$F2,$F8,$11,$89,$D4,$FB,$4D,$65,$E6,$C3,$AC,$63,$2F,$36,$3F
	.dc.b	$09,$B1,$F0,$71,$EA,$F3,$C3,$E6,$BF,$EE,$EC,$BF,$46,$DC,$3F,$54
	.dc.b	$DC,$3F,$DE,$32,$F0,$1F,$37,$86,$D7,$FB,$42,$75,$39,$24,$7E,$D2
	.dc.b	$77,$87,$EB,$DB,$A7,$EA,$3E,$63,$FD,$A4,$FE,$AA,$33,$5F,$A6,$6B
	.dc.b	$FB,$06,$E0,$AE,$33,$BC,$3F,$83,$56,$8D,$CE,$A7,$94,$E8,$8F,$F0
	.dc.b	$A8,$D5,$7B,$23,$D5,$3A,$C3,$F8,$45,$6A,$FB,$4E,$70,$F6,$5D,$3C
	.dc.b	$87,$23,$B1,$93,$C6,$66,$FA,$B9,$C3,$C5,$1F,$26,$BF,$91,$61,$E1
	.dc.b	$75,$87,$BA,$EC,$71,$9D,$47,$5C,$3E,$74,$F2,$8D,$99,$3F,$8A,$CF
	.dc.b	$0B,$F0,$46,$4F,$E4,$31,$9A,$E7,$19,$C3,$EA,$3D,$2B,$89,$EC,$75
	.dc.b	$87,$C8,$7E,$19,$8C,$3D,$0F,$1B,$EB,$5C,$4E,$6E,$09,$1E,$83,$81
	.dc.b	$F4,$23,$84,$6A,$3E,$7C,$72,$78,$1E,$13,$D5,$87,$B5,$E5,$3C,$70
	.dc.b	$F6,$C7,$F6,$0F,$AC,$7D,$67,$E4,$B8,$DE,$E7,$E9,$AB,$FB,$2B,$72
	.dc.b	$FF,$01,$DF,$09,$FF,$0A,$B5,$DE,$7C,$EB,$86,$5F,$E1,$0A,$D5,$FF
	.dc.b	$83,$3E,$5F,$FD,$8B,$87,$F8,$79,$AF,$F7,$46,$C6,$FD,$1F,$34,$62
	.dc.b	$FF,$C3,$6D,$AF,$F1,$1B,$72,$3F,$EA,$C8,$FF,$0C,$FD,$2F,$91,$FE
	.dc.b	$B7,$A7,$EA,$0F,$D7,$46,$60,$DD,$7D,$38,$7F,$68,$AD,$EB,$2F,$AB
	.dc.b	$5F,$F0,$F5,$BE,$7F,$E7,$7D,$63,$EB,$2F,$85,$2F,$FC,$7F,$39,$6C
	.dc.b	$E5,$FF,$0B,$C2,$79,$CF,$F0,$AF,$13,$FC,$A7,$EC,$DF,$1B,$FD,$AD
	.dc.b	$87,$F9,$51,$AF,$F9,$26,$7F,$CB,$3F,$4F,$BA,$93,$B7,$94,$7C,$4F
	.dc.b	$98,$F8,$BB,$CA,$F9,$71,$97,$E6,$C6,$5F,$83,$19,$7F,$CF,$9A,$A3
	.dc.b	$FC,$29,$E6,$BB,$7F,$C2,$D3,$9A,$FF,$42,$AF,$9E,$DC,$FF,$9A,$7C
	.dc.b	$E7,$CD,$E2,$DA,$A5,$F4,$99,$8C,$36,$38,$CF,$FA,$7A,$B1,$32,$30
	.dc.b	$7F,$AB,$6F,$97,$A5,$E7,$8C,$1F,$EB,$1F,$83,$CC,$C8,$C9,$FE,$B3
	.dc.b	$8F,$37,$53,$EA,$9E,$4E,$1F,$F2,$E5,$7F,$B4,$5A,$FF,$B2,$DD,$B9
	.dc.b	$8E,$1E,$54,$8C,$10,$E1,$1D,$4E,$DF,$15,$F0,$FD,$A7,$B9,$7D,$59
	.dc.b	$7F,$EE,$C8,$D1,$87,$FD,$D7,$29,$CC,$0F,$F7,$7C,$BE,$BF,$E8,$36
	.dc.b	$35,$CB,$F2,$78,$FC,$35,$AF,$85,$DB,$2B,$7C,$1F,$EA,$7C,$39,$47
	.dc.b	$6A,$FC,$03,$FB,$6E,$3B,$CD,$FC,$02,$3F,$13,$E1,$E2,$64,$60,$87
	.dc.b	$E5,$1C,$B4,$FC,$8B,$D3,$97,$F1,$5F,$04,$F8,$8B,$CF,$2F,$CE,$7E
	.dc.b	$4A,$74,$E3,$6F,$F3,$AF,$BB,$F9,$3F,$41,$EF,$D3,$11,$F9,$7F,$86
	.dc.b	$7A,$F7,$D4,$98,$F3,$9C,$BD,$7E,$1E,$AE,$6F,$9D,$1F,$EB,$28,$D5
	.dc.b	$7E,$8E,$37,$3F,$4F,$7F,$E0,$AD,$F0,$FF,$51,$06,$F8,$9F,$AB,$F4
	.dc.b	$3F,$D3,$78,$FB,$5F,$F1,$83,$7D,$2D,$7A,$9F,$B1,$F8,$A7,$D6,$D8
	.dc.b	$95,$F1,$67,$2F,$C7,$F0,$07,$EE,$4F,$96,$FB,$7B,$7F,$DD,$51,$8D
	.dc.b	$7F,$18,$F0,$9F,$A1,$F1,$BE,$47,$81,$FC,$19,$7C,$17,$F0,$56,$C7
	.dc.b	$0C,$FF,$84,$7C,$7B,$DA,$BE,$FC,$AF,$B9,$39,$7E,$F7,$D5,$3E,$4B
	.dc.b	$F8,$82,$F8,$AA,$62,$3F,$8A,$7D,$4B,$E2,$2F,$D5,$4C,$3F,$11,$39
	.dc.b	$7F,$15,$D4,$9F,$E3,$9F,$2B,$FC,$6A,$7D,$BE,$0A,$31,$C7,$FA,$FB
	.dc.b	$E1,$5F,$93,$F9,$77,$C5,$7F,$25,$7D,$51,$C9,$B5,$FB,$77,$71,$C7
	.dc.b	$E1,$38,$54,$8C,$1F,$9F,$F9,$9F,$F0,$55,$FE,$FD,$FE,$C5,$19,$F7
	.dc.b	$BF,$9A,$46,$BA,$7F,$85,$68,$FA,$97,$D3,$EB,$E0,$66,$3F,$AB,$9C
	.dc.b	$B6,$7F,$9F,$46,$23,$E1,$B7,$2F,$E2,$0F,$F8,$5D,$FE,$25,$F5,$1F
	.dc.b	$E1,$06,$D7,$F4,$78,$E1,$8F,$91,$D3,$7E,$4F,$53,$30,$FD,$F5,$69
	.dc.b	$FE,$A6,$DC,$8D,$27,$C3,$7F,$55,$FA,$17,$CF,$5F,$9F,$98,$7F,$09
	.dc.b	$9C,$B6,$72,$FE,$02,$13,$CE,$7F,$B0,$7D,$17,$0A,$FC,$2B,$FD,$AD
	.dc.b	$FE,$28,$D7,$31,$C0,$8E,$9E,$73,$FC,$C3,$D5,$FE,$AB,$B7,$F3,$BF
	.dc.b	$A8,$7C,$B7,$F6,$F6,$E1,$FC,$90,$FF,$19,$FB,$FC,$48,$E6,$F9,$F0
	.dc.b	$FF,$4B,$87,$DB,$FE,$6A,$31,$F5,$3F,$E3,$CC,$BF,$A8,$7D,$43,$E8
	.dc.b	$F8,$1F,$96,$C1,$19,$7F,$AB,$46,$63,$10,$AE,$8F,$09,$EF,$8F,$F2
	.dc.b	$B8,$C1,$CB,$AB,$CD,$7C,$AB,$A1,$C3,$0F,$2B,$6B,$89,$E7,$AE,$31
	.dc.b	$E8,$6C,$7F,$6B,$C3,$C7,$7E,$66,$D6,$E7,$A9,$E3,$29,$EB,$C3,$D1
	.dc.b	$5C,$5B,$3C,$67,$C6,$FA,$1F,$04,$69,$B8,$7D,$37,$D4,$5B,$F8,$ED
	.dc.b	$9F,$8C,$E7,$DC,$7D,$4F,$18,$F6,$46,$1C,$57,$F8,$EA,$75,$1F,$4E
	.dc.b	$74,$E3,$E4,$B2,$FF,$83,$DA,$9A,$47,$E7,$E1,$FE,$4F,$D3,$E2,$C7
	.dc.b	$13,$3F,$17,$FE,$3B,$D3,$F0,$71,$97,$F0,$B5,$C6,$3E,$C9,$9D,$3F
	.dc.b	$CB,$35,$C5,$F7,$61,$FE,$23,$EE,$E3,$3F,$E4,$77,$AD,$79,$CF,$F4
	.dc.b	$25,$72,$C7,$83,$CA,$3C,$B1,$E8,$78,$63,$D2,$31,$87,$E0,$31,$F3
	.dc.b	$4F,$68,$C6,$3E,$5B,$FF,$02,$7D,$47,$F9,$BF,$0F,$F0,$BE,$08,$C4
	.dc.b	$FF,$66,$1F,$53,$FD,$66,$B7,$3A,$AF,$F3,$11,$C6,$3F,$CC,$95,$FE
	.dc.b	$7C,$9E,$6D,$8C,$BF,$E4,$28,$CB,$FD,$03,$9D,$F3,$BD,$2B,$0D,$9F
	.dc.b	$F3,$EC,$BD,$DC,$76,$A3,$8D,$FF,$1C,$BE,$F7,$C1,$B5,$B6,$D7,$09
	.dc.b	$FC,$F4,$BA,$DF,$53,$C1,$B8,$7E,$6E,$75,$39,$24,$7F,$37,$F8,$8F
	.dc.b	$2D,$70,$BF,$E6,$D1,$E8,$FA,$9F,$E6,$F5,$FE,$45,$6C,$FF,$7A,$7F
	.dc.b	$21,$AF,$F9,$9D,$CB,$7B,$7D,$71,$C2,$7F,$94,$64,$8C,$47,$F6,$C6
	.dc.b	$BF,$EA,$67,$9D,$E7,$BB,$C9,$7E,$57,$E6,$F8,$BF,$D5,$27,$FB,$E4
	.dc.b	$62,$72,$C8,$F8,$93,$BB,$F9,$6D,$7C,$AD,$7F,$BF,$AF,$A9,$FA,$AD
	.dc.b	$9E,$71,$83,$EC,$F9,$1F,$CF,$E3,$E9,$36,$BE,$CB,$5F,$86,$D7,$FC
	.dc.b	$1A,$BD,$2E,$64,$9E,$6F,$E1,$BC,$AF,$F1,$D4,$7D,$F6,$D7,$E1,$9A
	.dc.b	$FD,$06,$BF,$E1,$AF,$A9,$7D,$07,$E4,$1B,$5F,$89,$6C,$FE,$41,$AF
	.dc.b	$DA,$F9,$FA,$FF,$89,$1A,$FF,$B7,$9B,$B9,$BE,$AB,$EA,$7F,$3E,$DC
	.dc.b	$3F,$30,$DB,$FC,$FB,$5F,$C4,$FB,$3F,$DE,$A7,$F3,$CD,$8F,$D2,$B5
	.dc.b	$DF,$93,$FA,$CF,$6D,$E7,$52,$5F,$E9,$E3,$2F,$EA,$7D,$CF,$F3,$EB
	.dc.b	$63,$F6,$B1,$8B,$FD,$9B,$5F,$CE,$FB,$EF,$A8,$FD,$DB,$6F,$F6,$8D
	.dc.b	$9F,$DD,$B5,$FD,$2F,$D4,$3E,$57,$F8,$1B,$6B,$F7,$8D,$9F,$E0,$6D
	.dc.b	$7F,$5B,$F0,$CF,$A8,$FE,$16,$DA,$FE,$08,$D9,$FE,$16,$D7,$F6,$BF
	.dc.b	$10,$F8,$6F,$E2,$6D,$AF,$E1,$8D,$9F,$E2,$6D,$7F,$7B,$F5,$0F,$6F
	.dc.b	$F8,$DB,$6B,$F8,$A3,$67,$F8,$DB,$5F,$E0,$BF,$1A,$F6,$3F,$91,$B6
	.dc.b	$BF,$8E,$36,$7F,$91,$B5,$FF,$01,$7D,$47,$F4,$37,$FA,$33,$7C,$4F
	.dc.b	$E8,$2F,$E9,$9B,$F0,$7F,$CE,$65,$3C,$5E,$71,$FC,$D9,$CB,$63,$51
	.dc.b	$FE,$43,$7B,$5F,$F9,$17,$E1,$53,$D6,$72,$E2,$33,$3F,$EF,$1F,$93
	.dc.b	$7D,$47,$F4,$86,$C6,$7E,$A3,$13,$23,$83,$B8,$C5,$FF,$AB,$BE,$5F
	.dc.b	$FD,$F3,$5F,$EF,$D7,$FA,$A3,$7E,$47,$FA,$AC,$62,$78,$46,$3D,$1C
	.dc.b	$B6,$BD,$7A,$9F,$6F,$CD,$7F,$9B,$2B,$FB,$13,$63,$66,$BB,$BC,$78
	.dc.b	$7F,$B7,$A7,$4D,$8F,$13,$FE,$6D,$8C,$B6,$32,$F3,$AF,$ED,$33,$CA
	.dc.b	$FC,$92,$8F,$F3,$FF,$C4,$F9,$2F,$DD,$5A,$9C,$E1,$E6,$B7,$88,$F9
	.dc.b	$47,$C6,$FF,$7A,$6D,$6E,$75,$F1,$D9,$A9,$39,$AD,$DF,$FB,$81,$FF
	.dc.b	$7D,$C6,$67,$FC,$0B,$F3,$7F,$DF,$72,$F4,$E2,$87,$A7,$AF,$5F,$6D
	.dc.b	$C3,$C5,$3A,$6C,$E9,$FF,$4E,$72,$FF,$BA,$9B,$F0,$7F,$9F,$8F,$D9
	.dc.b	$63,$FE,$24,$75,$A7,$F8,$B5,$AF,$F9,$B5,$AF,$E7,$67,$F4,$91,$88
	.dc.b	$FC,$E4,$7D,$2F,$71,$F3,$BF,$A3,$F3,$F0,$F1,$C7,$FC,$CC,$E5,$B7
	.dc.b	$F4,$BE,$79,$D1,$C2,$30,$DD,$D2,$FA,$0C,$3C,$2F,$9C,$EA,$B2,$6A
	.dc.b	$AC,$F8,$39,$1F,$51,$FD,$C6,$30,$FF,$62,$D3,$D7,$97,$E3,$63,$4F
	.dc.b	$F9,$F7,$3F,$D1,$C6,$DB,$EA,$7A,$70,$EB,$AD,$AE,$38,$74,$9C,$E3
	.dc.b	$F0,$B1,$6E,$F5,$3E,$14,$61,$B0,$F9,$8B,$12,$BF,$86,$D7,$F1,$38
	.dc.b	$FD,$2C,$7D,$B6,$C6,$63,$6F,$F1,$58,$EB,$F4,$FC,$E3,$DC,$7F,$19
	.dc.b	$7D,$9E,$AB,$E2,$E9,$8F,$9F,$B8,$7C,$D9,$D3,$67,$4D,$3E,$97,$D4
	.dc.b	$7B,$DB,$7F,$26,$31,$87,$8B,$44,$27,$59,$6D,$B7,$C8,$F9,$9F,$4C
	.dc.b	$F9,$9F,$EA,$D8,$FB,$04,$67,$CC,$FA,$5B,$BE,$77,$E8,$CE,$6B,$83
	.dc.b	$78,$7F,$0B,$1C,$B0,$EB,$18,$BE,$58,$F8,$A7,$3C,$BF,$D1,$AF,$86
	.dc.b	$59,$3C,$39,$78,$2B,$73,$FD,$33,$2F,$05,$F1,$6C,$7F,$4D,$AE,$17
	.dc.b	$CE,$3C,$F7,$C9,$FC,$6C,$E9,$CC,$EF,$2F,$44,$E5,$F6,$DF,$6A,$F3
	.dc.b	$9F,$27,$2F,$6B,$C8,$93,$C3,$0F,$04,$EA,$31,$87,$7C,$7D,$55,$AC
	.dc.b	$7C,$11,$89,$D6,$1E,$DC,$AB,$E2,$C3,$9B,$E3,$BE,$2F,$C7,$C3,$E0
	.dc.b	$65,$E9,$CB,$C5,$97,$5A,$F3,$B7,$0F,$83,$8F,$6B,$F8,$07,$1C,$7A
	.dc.b	$DF,$86,$F8,$1F,$E0,$18,$CC,$F0,$C7,$D9,$7C,$DF,$19,$8B,$F0,$9E
	.dc.b	$7C,$BD,$B9,$79,$A3,$06,$8F,$01,$F0,$CC,$5F,$79,$FC,$4D,$F7,$79
	.dc.b	$E1,$E4,$C3,$C3,$1A,$BE,$31,$87,$E5,$57,$17,$C1,$87,$8D,$F0,$9F
	.dc.b	$82,$3F,$34,$53,$E6,$18,$C3,$F5,$0D,$EF,$7C,$AF,$08,$FC,$E4,$61
	.dc.b	$C4,$FF,$72,$9D,$47,$ED,$A7,$4F,$38,$FD,$EC,$71,$D3,$F2,$ED,$D3
	.dc.b	$F4,$DF,$51,$FA,$F9,$CD,$70,$25,$FE,$F2,$75,$5F,$AD,$BE,$53,$E6
	.dc.b	$AD,$DF,$E9,$2F,$73,$93,$F8,$2D,$70,$7D,$15,$FC,$4A,$77,$18,$37
	.dc.b	$1D,$2F,$CF,$14,$E9,$7E,$87,$CC,$64,$E4,$E4,$F2,$39,$C3,$E5,$3D
	.dc.b	$8E,$25,$8E,$13,$EF,$AF,$34,$F1,$C3,$D5,$7C,$AF,$AC,$F1,$BE,$0D
	.dc.b	$3F,$8B,$61,$EE,$AE,$13,$AA,$EF,$5A,$3B,$61,$E7,$BE,$18,$78,$09
	.dc.b	$1E,$49,$EC,$7C,$0B,$E8,$7C,$87,$38,$78,$1F,$93,$19,$C3,$C3,$1C
	.dc.b	$4C,$9E,$33,$8D,$F8,$1F,$8F,$1C,$2B,$A4,$71,$3F,$A4,$E1,$E8,$C3
	.dc.b	$D7,$1C,$30,$F2,$BD,$8E,$14,$AE,$11,$E0,$AF,$75,$F7,$C3,$C4,$FD
	.dc.b	$03,$11,$F1,$E7,$FC,$35,$F5,$0F,$A8,$FC,$9F,$1B,$D8,$FD,$3E,$1F
	.dc.b	$E0,$F6,$E9,$FE,$7A,$BE,$18,$72,$C3,$8D,$6A,$3F,$CD,$8D,$C1,$5C
	.dc.b	$27,$FB,$A7,$C9,$FF,$BB,$9F,$F1,$63,$5C,$FB,$BE,$48,$FF,$5F,$46
	.dc.b	$A3,$04,$78,$FA,$7F,$A6,$FA,$2F,$14,$7F,$76,$CB,$F1,$86,$75,$91
	.dc.b	$FE,$1A,$13,$FC,$1B,$8B,$F8,$AB,$63,$FB,$83,$5F,$F5,$BB,$D8,$93
	.dc.b	$E1,$F9,$1F,$97,$F5,$0F,$A8,$BE,$1C,$C3,$FC,$95,$39,$6C,$E5,$FF
	.dc.b	$0C,$47,$F9,$5D,$C9,$FE,$56,$F1,$BF,$DE,$0D,$F6,$BF,$85,$F8,$FF
	.dc.b	$E5,$E6,$CB,$E8,$AC,$4A,$E0,$73,$8F,$91,$F2,$5F,$27,$9A,$4F,$F9
	.dc.b	$37,$64,$EA,$3F,$96,$C6,$1B,$1F,$CC,$1F,$96,$D8,$D4,$3C,$BF,$F5
	.dc.b	$8B,$A3,$FD,$58,$53,$5C,$1F,$43,$EA,$3F,$C4,$13,$FE,$A0,$F3,$9F
	.dc.b	$2D,$8A,$FF,$74,$ED,$FE,$A3,$FA,$4C,$63,$FE,$B6,$B4,$62,$7F,$D5
	.dc.b	$17,$67,$9D,$7F,$A9,$AB,$06,$67,$87,$A5,$E8,$C7,$FD,$62,$DC,$7C
	.dc.b	$2D,$BF,$F6,$2A,$F8,$3E,$33,$27,$FD,$80,$F4,$8F,$3F,$63,$EA,$1E
	.dc.b	$66,$1F,$F3,$45,$FF,$B4,$67,$5E,$5E,$6A,$4F,$F6,$59,$E5,$1F,$D8
	.dc.b	$9B,$19,$AE,$AD,$9E,$2F,$FB,$BE,$32,$DF,$87,$F9,$1E,$DB,$DD,$BE
	.dc.b	$BC,$AF,$F6,$F4,$E5,$B3,$97,$ED,$9F,$EE,$88,$FF,$6B,$BA,$3F,$DC
	.dc.b	$D0,$FA,$8F,$AC,$D8,$CC,$63,$EA,$32,$C8,$C9,$CA,$3F,$02,$FE,$F3
	.dc.b	$86,$F3,$3C,$4A,$62,$24,$FE,$71,$FC,$C1,$1C,$1F,$96,$9F,$C3,$BE
	.dc.b	$C3,$B1,$9E,$4F,$C7,$7C,$13,$EA,$3F,$C2,$7D,$08,$C5,$7F,$86,$E3
	.dc.b	$13,$A9,$CC,$FF,$89,$A7,$FD,$19,$EF,$F2,$F4,$4D,$24,$F3,$F8,$5F
	.dc.b	$F5,$F3,$7D,$7F,$0C,$62,$3A,$35,$FF,$46,$5B,$E1,$B1,$1F,$AC,$6A
	.dc.b	$8D,$57,$B2,$73,$1F,$AE,$F0,$BD,$FA,$7F,$16,$BF,$53,$5E,$ED,$7F
	.dc.b	$73,$E6,$FF,$AB,$DF,$88,$D7,$E1,$35,$F3,$B5,$FE,$03,$F1,$4F,$A8
	.dc.b	$F9,$ED,$8F,$9C,$D7,$E6,$35,$F8,$3D,$03,$EB,$95,$FA,$CD,$7E,$A3
	.dc.b	$5F,$93,$F1,$8F,$4B,$EF,$B6,$3E,$F3,$5F,$B8,$D7,$E8,$FC,$6F,$FB
	.dc.b	$E9,$FC,$23,$5F,$C1,$35,$DF,$AD,$F8,$CF,$A8,$7D,$45,$24,$FE,$36
	.dc.b	$32,$FE,$3F,$EA,$1F,$11,$F9,$46,$C6,$63,$07,$E6,$3D,$E7,$D4,$52
	.dc.b	$4F,$E6,$63,$2F,$E6,$FE,$51,$F0,$FF,$A0,$6C,$66,$30,$67,$E0,$24
	.dc.b	$71,$6B,$FA,$1F,$96,$FF,$3C,$BF,$96,$6B,$FD,$01,$AF,$E5,$5B,$CE
	.dc.b	$FA,$77,$D3,$FE,$C9,$B5,$FB,$16,$BF,$AF,$6B,$FA,$47,$3E,$83,$CF
	.dc.b	$FE,$9F,$18,$6D,$7F,$4D,$8F,$F0,$F4,$EE,$BF,$51,$F4,$FF,$E1,$D8
	.dc.b	$C6,$08,$E5,$3A,$BF,$EA,$D1,$BC,$3F,$AE,$BF,$E2,$59,$FD,$BB,$5C
	.dc.b	$FB,$FF,$C3,$A3,$11,$E0,$8E,$64,$47,$F3,$4B,$F1,$7B,$FF,$D9,$7E
	.dc.b	$75,$F3,$77,$E7,$25,$F0,$D3,$D3,$5A,$8C,$67,$FD,$CA,$31,$1F,$C6
	.dc.b	$5A,$FF,$0E,$6B,$BB,$FE,$4B,$F3,$F8,$8F,$E1,$CD,$C8,$D2,$4F,$F7
	.dc.b	$B7,$BE,$1F,$31,$B8,$7F,$81,$16,$3F,$D1,$3C,$CF,$ED,$78,$FF,$B2
	.dc.b	$DA,$FF,$B5,$1A,$FC,$63,$FA,$47,$AB,$FE,$1D,$8E,$38,$7F,$86,$9B
	.dc.b	$86,$DB,$5F,$3B,$E9,$BF,$A3,$D7,$F8,$83,$E9,$BF,$A9,$3F,$4B,$E9
	.dc.b	$1F,$4B,$DC,$FD,$3D,$A3,$2F,$D8,$8C,$C6,$21,$5D,$1E,$13,$E9,$8C
	.dc.b	$4E,$E3,$89,$DF,$9F,$DD,$1C,$E7,$73,$C7,$1F,$F1,$C5,$BE,$17,$96
	.dc.b	$1F,$15,$B1,$F8,$C7,$94,$F0,$6E,$3E,$FB,$F0,$55,$3D,$97,$E2,$8F
	.dc.b	$04,$E6,$7D,$B7,$E2,$C7,$BC,$7B,$8E,$26,$F1,$F4,$BF,$9E,$7E,$23
	.dc.b	$F6,$CE,$75,$5D,$57,$B9,$D4,$F0,$7E,$57,$A4,$FA,$6E,$11,$F4,$23
	.dc.b	$0F,$39,$FF,$2A,$CE,$9B,$1F,$E6,$5F,$8A,$FF,$33,$26,$91,$D1,$ED
	.dc.b	$3F,$CD,$1F,$03,$FC,$2F,$2F,$C0,$47,$EE,$A7,$E3,$65,$FE,$76,$DF
	.dc.b	$E4,$4E,$4E,$51,$97,$A1,$9A,$FC,$54,$65,$EB,$1E,$53,$8D,$F8,$74
	.dc.b	$F9,$18,$78,$23,$13,$F3,$63,$18,$78,$A3,$95,$F6,$AD,$C7,$EB,$6B
	.dc.b	$6F,$90,$DD,$FF,$A0,$27,$AF,$23,$E9,$DD,$1F,$CB,$73,$FE,$57,$18
	.dc.b	$CF,$F8,$9E,$3F,$E1,$78,$74,$6E,$7F,$D0,$F1,$B9,$FE,$44,$DA,$FE
	.dc.b	$6B,$F6,$DF,$E8,$EA,$E7,$8F,$87,$1E,$58,$7F,$88,$EB,$4F,$F9,$7B
	.dc.b	$77,$7B,$E7,$77,$18,$78,$E1,$FC,$73,$E6,$F3,$5B,$CB,$D3,$7B,$7C
	.dc.b	$3F,$6E,$F2,$77,$8F,$28,$C4,$F9,$BB,$9F,$35,$D9,$FF,$23,$C6,$63
	.dc.b	$1E,$BE,$5B,$19,$39,$47,$9B,$E9,$9F,$11,$4C,$75,$DA,$F8,$B3,$A7
	.dc.b	$E3,$FD,$67,$F5,$D7,$35,$F0,$23,$1C,$3F,$D4,$6B,$83,$FD,$4E,$33
	.dc.b	$C9,$FD,$9A,$72,$FF,$68,$8C,$DF,$F8,$07,$9F,$E5,$F8,$FF,$9D,$AF
	.dc.b	$ED,$AF,$09,$D1,$F6,$9D,$3F,$EC,$8F,$C0,$DF,$51,$CA,$FE,$AC,$62
	.dc.b	$7F,$02,$DC,$BF,$B4,$7D,$73,$EA,$37,$87,$DC,$8C,$5F,$E2,$39,$9E
	.dc.b	$EF,$E0,$17,$9D,$63,$0C,$9F,$C7,$BF,$83,$C7,$FB,$B5,$70,$9C,$F3
	.dc.b	$7E,$2B,$EB,$5F,$57,$79,$B0,$DA,$FC,$CF,$D5,$FF,$4E,$6E,$7F,$92
	.dc.b	$8C,$65,$FE,$03,$7C,$F3,$FA,$0F,$9F,$CF,$1F,$F8,$61,$CD,$7F,$82
	.dc.b	$5C,$CF,$FB,$C9,$CC,$FF,$87,$BD,$87,$C1,$FE,$81,$FF,$5F,$57,$EB
	.dc.b	$1C,$B7,$E2,$3F,$4B,$ED,$BE,$D2,$FD,$7E,$1B,$5F,$B4,$F7,$FF,$7A
	.dc.b	$DF,$9D,$FD,$6C,$63,$5F,$F0,$F2,$8F,$D8,$FC,$0F,$FA,$D6,$BF,$7A
	.dc.b	$BE,$4F,$3F,$15,$A7,$F6,$FF,$0A,$FA,$7B,$9E,$1B,$87,$F1,$0C,$DF
	.dc.b	$53,$F8,$7E,$2F,$F2,$17,$94,$FC,$0D,$F3,$BF,$C2,$AF,$7E,$07,$F0
	.dc.b	$69,$CE,$3C,$F1,$7C,$CF,$F0,$AF,$86,$FF,$29,$4E,$63,$0F,$39,$FF
	.dc.b	$1C,$BF,$C2,$A5,$F3,$19,$F8,$1E,$4E,$FE,$03,$87,$99,$6F,$C0,$F3
	.dc.b	$F8,$CF,$07,$D4,$D9,$E8,$F7,$F4,$24,$7F,$18,$F9,$FF,$F1,$7B,$E9
	.dc.b	$8C,$F0,$7F,$25,$D3,$F9,$74,$FF,$8B,$75,$FE,$2B,$39,$F6,$37,$C9
	.dc.b	$F2,$2F,$73,$FE,$58,$AD,$CE,$6B,$F9,$5F,$CA,$BE,$7E,$98,$E5,$B9
	.dc.b	$7F,$3A,$9D,$36,$74,$D3,$F9,$FF,$CB,$FF,$A9,$9B,$F1,$9F,$CE,$23
	.dc.b	$19,$FF,$9A,$5F,$F5,$2C,$F0,$F8,$DD,$56,$DB,$1F,$D5,$3E,$6D,$F6
	.dc.b	$AC,$BF,$A5,$C6,$1B,$0F,$9A,$B1,$2B,$94,$FF,$00,$6C,$70,$F3,$75
	.dc.b	$9B,$7F,$F4,$96,$3E,$09,$CC,$63,$0E,$55,$FE,$83,$F0,$7F,$96,$BF
	.dc.b	$EF,$F6,$FB,$FF,$A5,$F9,$CF,$EB,$31,$82,$7C,$76,$23,$2F,$F5,$EF
	.dc.b	$9D,$FA,$FA,$FF,$A4,$2D,$0F,$82,$D7,$46,$DB,$5F,$DC,$27,$4D,$9D
	.dc.b	$3F,$A5,$9F,$D0,$1F,$DB,$BA,$7F,$9E,$E7,$F8,$78,$CC,$FE,$AF,$8E
	.dc.b	$CF,$C5,$8F,$C9,$46,$36,$FF,$5C,$4F,$EE,$A3,$13,$A3,$FC,$07,$F8
	.dc.b	$37,$EC,$A7,$FD,$54,$FE,$BA,$32,$7F,$7E,$71,$1A,$8C,$77,$E5,$A7
	.dc.b	$E1,$7E,$CF,$FD,$6A,$D9,$FC,$2C,$3E,$52,$C4,$9D,$1C,$63,$11,$ED
	.dc.b	$1C,$8F,$C1,$BF,$56,$DE,$CE,$BF,$15,$F9,$29,$C9,$AF,$C1,$9F,$37
	.dc.b	$9A,$FC,$BC,$63,$2F,$CE,$56,$EB,$51,$F9,$F9,$D6,$3F,$1B,$D0,$F8
	.dc.b	$87,$C5,$F9,$6B,$5F,$32,$2C,$FC,$C9,$BC,$3F,$EE,$8F,$57,$F2,$ED
	.dc.b	$EC,$FF,$15,$C7,$F8,$83,$0F,$F1,$BB,$A7,$F2,$0D,$B7,$42,$30,$DF
	.dc.b	$6F,$9D,$FF,$8D,$B5,$7D,$6D,$FA,$DC,$37,$2F,$F2,$2C,$E9,$B3,$A6
	.dc.b	$9A,$FB,$1E,$87,$5F,$B6,$B1,$93,$9F,$D8,$63,$5F,$F7,$B6,$08,$F2
	.dc.b	$78,$FE,$5F,$22,$DF,$89,$B7,$C1,$BC,$09,$FD,$96,$1F,$D3,$AB,$F0
	.dc.b	$C3,$EB,$1F,$5B,$F8,$F8,$FF,$4B,$7A,$5F,$EE,$39,$CF,$5F,$FA,$46
	.dc.b	$3F,$DC,$DF,$23,$FD,$EA,$B6,$0E,$97,$CF,$E5,$7F,$AD,$C6,$70,$E6
	.dc.b	$F0,$37,$87,$57,$95,$F5,$AE,$31,$A3,$85,$72,$8C,$B7,$0F,$2F,$C0
	.dc.b	$FC,$1A,$DE,$3E,$39,$F6,$36,$3C,$8D,$7F,$CE,$13,$C6,$FA,$CE,$6F
	.dc.b	$B5,$70,$C7,$D0,$F0,$CE,$CE,$A3,$D1,$7C,$AB,$79,$7B,$72,$F1,$C7
	.dc.b	$0C,$7B,$1E,$39,$D5,$F7,$3D,$73,$B3,$9E,$5E,$7B,$E7,$3A,$26,$1F
	.dc.b	$2A,$F9,$C7,$92,$77,$3E,$58,$FA,$93,$AC,$7D,$44,$C7,$C9,$87,$7A
	.dc.b	$F2,$5F,$12,$61,$F4,$B2,$F3,$4F,$8A,$3B,$99,$70,$72,$72,$FE,$0A
	.dc.b	$BB,$3E,$73,$44,$3E,$81,$C8,$D1,$31,$F5,$61,$F3,$32,$F1,$E3,$F0
	.dc.b	$EF,$AE,$5E,$D9,$CE,$3E,$5A,$F5,$E3,$F4,$4F,$CF,$14,$F9,$E6,$0F
	.dc.b	$CB,$16,$9F,$42,$F1,$75,$3F,$A3,$8C,$38,$9D,$CE,$A3,$F6,$D3,$A7
	.dc.b	$A3,$C2,$BF,$42,$DB,$FD,$C3,$67,$F6,$1D,$7F,$EF,$19,$CD,$FE,$5A
	.dc.b	$B5,$5D,$A3,$F3,$35,$B8,$F3,$46,$C8,$FE,$CA,$34,$F1,$3F,$6B,$3D
	.dc.b	$30,$F5,$C6,$CE,$38,$7A,$2E,$E1,$FC,$3F,$0E,$77,$C2,$B5,$7D,$2F
	.dc.b	$CA,$F0,$3C,$A6,$4E,$C5,$8E,$24,$BE,$D1,$C4,$F4,$46,$5F,$8E,$66
	.dc.b	$BC,$53,$F2,$23,$AC,$ED,$E3,$19,$8F,$E2,$77,$EA,$C3,$E1,$13,$0F
	.dc.b	$4D,$F9,$6F,$81,$1F,$09,$93,$DB,$7D,$F0,$FF,$4E,$47,$96,$FA,$3E
	.dc.b	$4A,$E4,$F9,$0E,$2F,$C0,$7A,$19,$24,$F4,$72,$FC,$07,$89,$E4,$AF
	.dc.b	$E7,$7D,$2F,$95,$FF,$14,$F0,$FF,$2F,$8C,$78,$7F,$A6,$8C,$4F,$FA
	.dc.b	$52,$FF,$A3,$46,$38,$74,$7F,$91,$E7,$9B,$6B,$35,$FD,$EA,$79,$3F
	.dc.b	$E1,$FB,$F1,$56,$9F,$F1,$07,$65,$E3,$7D,$8D,$31,$12,$FF,$BD,$B5
	.dc.b	$FE,$E3,$F5,$1C,$7E,$73,$97,$34,$9F,$F0,$0C,$65,$FF,$01,$C6,$5F
	.dc.b	$EF,$31,$E3,$AF,$F4,$04,$7F,$85,$BD,$0C,$19,$F4,$6C,$64,$87,$B6
	.dc.b	$7F,$D0,$8F,$F9,$4B,$C5,$F7,$79,$26,$98,$F9,$9B,$3B,$21,$FE,$58
	.dc.b	$7E,$5F,$92,$FA,$BF,$F7,$14,$FF,$8E,$7E,$2F,$3E,$49,$DC,$EA,$7F
	.dc.b	$C9,$D5,$C7,$D3,$DD,$49,$FF,$50,$56,$F9,$FF,$D2,$6D,$E5,$FF,$03
	.dc.b	$61,$C6,$B7,$87,$93,$89,$F5,$7E,$6F,$C1,$FF,$07,$5F,$E5,$C6,$D9
	.dc.b	$A7,$E1,$32,$FF,$50,$B8,$C7,$91,$CF,$A7,$F0,$D9,$BC,$7B,$E3,$CC
	.dc.b	$BF,$2B,$7F,$E9,$9A,$D4,$60,$FF,$25,$B8,$7F,$24,$A7,$94,$62,$BF
	.dc.b	$32,$BE,$AD,$E5,$E1,$FC,$2C,$7F,$8D,$1B,$9F,$FA,$51,$DC,$78,$4E
	.dc.b	$53,$C7,$9E,$46,$67,$6F,$39,$EC,$47,$D5,$F0,$6F,$C2,$3E,$6F,$77
	.dc.b	$FA,$DB,$E0,$3F,$55,$EB,$BE,$6E,$98,$89,$87,$FA,$DE,$32,$FF,$AE
	.dc.b	$63,$F6,$5F,$79,$FE,$F5,$8F,$91,$68,$C4,$7E,$E5,$BA,$FF,$09,$5E
	.dc.b	$6F,$2F,$3A,$57,$29,$DF,$1F,$FB,$35,$DC,$7F,$BB,$F9,$5F,$19,$F8
	.dc.b	$9F,$8B,$93,$98,$FC,$61,$87,$F3,$72,$9C,$C6,$C8,$6A,$7F,$29,$D7
	.dc.b	$FC,$3F,$E0,$1F,$45,$7E,$16,$5F,$C0,$9C,$B6,$72,$FF,$19,$84,$F3
	.dc.b	$9F,$E3,$5C,$EF,$98,$A4,$BF,$97,$0F,$4F,$11,$FC,$8E,$38,$3A,$9E
	.dc.b	$73,$F8,$FF,$63,$F9,$5F,$43,$E4,$9F,$17,$86,$E1,$F8,$BA,$E0,$FE
	.dc.b	$22,$32,$FF,$B3,$EB,$95,$6A,$3F,$35,$57,$97,$F3,$4F,$83,$AB,$FD
	.dc.b	$1F,$3F,$97,$FA,$9F,$07,$64,$8C,$1F,$9F,$FC,$21,$E7,$7E,$A9,$FF
	.dc.b	$4B,$47,$FA,$69,$B8,$FF,$BE,$9D,$C7,$1F,$73,$D9,$F0,$F2,$30,$75
	.dc.b	$F7,$7E,$07,$C1,$48,$C1,$E1,$F8,$1F,$89,$A4,$8C,$1F,$9C,$F9,$0F
	.dc.b	$E2,$1D,$DF,$EA,$1F,$05,$0F,$AB,$FE,$A8,$DA,$E1,$E5,$7F,$57,$F9
	.dc.b	$2F,$D8,$9C,$A3,$FD,$AE,$DB,$3F,$B6,$EF,$7C,$4F,$F5,$F5,$F5,$14
	.dc.b	$C4,$7F,$63,$F9,$5F,$06,$1F,$D9,$DF,$A2,$DF,$84,$D6,$9F,$DA,$7E
	.dc.b	$5B,$E5,$ED,$EB,$5E,$DE,$4D,$7C,$2F,$F6,$FF,$AB,$73,$5E,$59,$9F
	.dc.b	$F0,$09,$CB,$FC,$0B,$E4,$1F,$1B,$FD,$DD,$7C,$6D,$31,$1F,$DE,$FE
	.dc.b	$65,$E8,$DF,$4A,$61,$FC,$32,$72,$FF,$0E,$F9,$BF,$F3,$97,$81,$FE
	.dc.b	$13,$F9,$CF,$0C,$FF,$81,$DB,$1F,$E1,$79,$DF,$17,$D9,$FC,$25,$F3
	.dc.b	$BF,$E0,$B9,$CE,$0E,$C4,$61,$57,$09,$FE,$3F,$F3,$EF,$A9,$FF,$0A
	.dc.b	$7D,$07,$F6,$B9,$FE,$46,$D7,$FC,$35,$3B,$7F,$93,$FD,$63,$E1,$7F
	.dc.b	$C3,$93,$99,$DD,$FF,$87,$EB,$51,$C2,$B0,$47,$9F,$C9,$7F,$89,$7E
	.dc.b	$5D,$F3,$FF,$E2,$7F,$A4,$FC,$1F,$0F,$F8,$E5,$BC,$3F,$9B,$AD,$E1
	.dc.b	$FC,$E9,$AF,$F8,$C6,$76,$FC,$56,$EE,$9D,$DF,$F9,$15,$B7,$D7,$0F
	.dc.b	$21,$E5,$FA,$1F,$F1,$D7,$D3,$7F,$0E,$BF,$F2,$62,$F9,$5C,$9A,$F5
	.dc.b	$26,$9F,$D3,$A7,$4D,$9C,$BF,$C5,$FC,$5F,$E9,$2F,$DB,$FA,$7F,$B4
	.dc.b	$7D,$EF,$59,$3A,$6F,$C1,$7F,$28,$8C,$1C,$B7,$7D,$67,$83,$0F,$F2
	.dc.b	$A7,$CA,$FF,$4B,$7F,$B0,$FD,$4B,$E9,$FF,$CA,$CD,$8F,$EC,$B3,$B7
	.dc.b	$8F,$C2,$7F,$96,$E3,$1C,$1F,$E6,$A8,$C7,$C2,$7F,$83,$3E,$AF,$FB
	.dc.b	$B4,$AD,$F5,$6F,$63,$FD,$AE,$7E,$EA,$35,$19,$78,$46,$5E,$2F,$FB
	.dc.b	$7A,$BA,$C6,$0E,$CE,$27,$94,$77,$F3,$BF,$DC,$B1,$8A,$E8,$EA,$B8
	.dc.b	$47,$5F,$19,$87,$92,$3C,$F7,$C9,$B8,$64,$F0,$D7,$07,$C8,$2F,$AB
	.dc.b	$EC,$AE,$4D,$39,$1B,$9E,$58,$79,$5B,$7E,$09,$EA,$FA,$8F,$46,$1C
	.dc.b	$B0,$F8,$B3,$FF,$00,$5A,$E8,$FF,$7F,$B8,$7E,$4F,$C2,$38,$FF,$DF
	.dc.b	$D3,$FE,$0A,$6C,$66,$78,$B5,$FB,$C4,$1F,$35,$5C,$35,$DB,$5F,$C3
	.dc.b	$35,$E2,$7C,$AB,$FC,$B7,$5F,$C6,$8D,$99,$E1,$46,$63,$8E,$9F,$82
	.dc.b	$AE,$31,$F5,$0E,$31,$9A,$E3,$19,$3C,$3E,$0E,$6F,$F3,$4F,$E3,$23
	.dc.b	$C2,$E5,$FE,$59,$A7,$BF,$1F,$B7,$8F,$83,$1E,$15,$E4,$AF,$F6,$96
	.dc.b	$3E,$09,$CD,$F0,$8C,$4F,$3A,$F7,$63,$F8,$FE,$4F,$F6,$F7,$CA,$FA
	.dc.b	$32,$FF,$6D,$B6,$7F,$DB,$1E,$0D,$BF,$F6,$B4,$DB,$E3,$5F,$55,$B3
	.dc.b	$F6,$5A,$FD,$C6,$BF,$BB,$DB,$FA,$7F,$8F,$D5,$B5,$FE,$6C,$E7,$FE
	.dc.b	$D2,$FF,$9F,$3C,$7F,$EB,$2D,$AF,$9F,$7F,$88,$BE,$2F,$E2,$5A,$E6
	.dc.b	$B8,$BE,$5E,$D7,$C1,$72,$8F,$F0,$F4,$63,$0F,$63,$63,$84,$ED,$FE
	.dc.b	$D1,$19,$D7,$F9,$DF,$99,$F1,$63,$17,$D2,$3E,$47,$D5,$FF,$80,$23
	.dc.b	$FD,$12,$DB,$FF,$80,$0F,$FA,$56,$33,$E6,$F0,$3C,$1B,$27,$85,$7C
	.dc.b	$D6,$BF,$E3,$76,$BF,$DF,$5E,$D8,$FF,$63,$F1,$7A,$61,$FE,$9A,$F1
	.dc.b	$BF,$AE,$3F,$E9,$CF,$AB,$FC,$FD,$7E,$73,$14,$62,$7F,$C8,$8F,$46
	.dc.b	$F2,$FE,$16,$3F,$4F,$97,$E0,$5A,$EA,$7F,$11,$18,$8F,$F0,$97,$95
	.dc.b	$FE,$78,$9F,$C6,$36,$3F,$18,$D7,$F1,$8D,$7F,$CA,$DC,$1F,$EF,$D9
	.dc.b	$FC,$A3,$63,$F2,$8D,$7F,$28,$D7,$8F,$9B,$F3,$A0,$D2,$4F,$E7,$FC
	.dc.b	$EF,$F6,$FB,$7E,$0F,$EA,$C6,$2F,$F4,$4D,$7F,$45,$F5,$7F,$D6,$6B
	.dc.b	$F5,$0D,$8F,$D4,$35,$DF,$57,$E0,$3E,$1B,$11,$FE,$73,$6B,$FE,$39
	.dc.b	$6E,$DF,$EA,$19,$FC,$3F,$B1,$FE,$AB,$C3,$F6,$CD,$BF,$DB,$35,$FD
	.dc.b	$B3,$5F,$F4,$07,$91,$FE,$2C,$9F,$DF,$36,$3F,$7C,$D7,$F7,$CD,$7F
	.dc.b	$D1,$1E,$EF,$F8,$FE,$7F,$83,$36,$3F,$83,$35,$FE,$0C,$D7,$FD,$21
	.dc.b	$EF,$7F,$B6,$67,$F8,$73,$63,$F8,$73,$5F,$E1,$CD,$7F,$D3,$1F,$00
	.dc.b	$FA,$BF,$D4,$B6,$BF,$8C,$36,$7F,$8C,$35,$FE,$A5,$F4,$FF,$8D,$7F
	.dc.b	$C2,$4D,$4D,$27,$A1,$FC,$16,$32,$D8,$FF,$1F,$57,$F5,$96,$CF,$F5
	.dc.b	$D5,$F5,$5B,$24,$ED,$E5,$F5,$52,$39,$36,$B9,$7C,$77,$72,$4F,$77
	.dc.b	$C5,$DF,$FC,$EE,$7D,$2F,$F3,$1F,$A4,$5E,$71,$FF,$4D,$E1,$FE,$31
	.dc.b	$EC,$FE,$5F,$18,$C3,$D1,$87,$4A,$F6,$36,$3F,$DD,$5E,$EB,$C7,$E9
	.dc.b	$3F,$CF,$E3,$15,$FF,$00,$3D,$4F,$E9,$9F,$1E,$6F,$C9,$E1,$93,$B3
	.dc.b	$C8,$7F,$B5,$65,$3A,$C7,$FC,$63,$3F,$ED,$9F,$91,$FF,$5D,$BC,$23
	.dc.b	$FD,$41,$0F,$82,$FE,$B8,$EA,$38,$9A,$8C,$FC,$1E,$9B,$5C,$FA,$D7
	.dc.b	$CE,$3B,$7C,$CB,$EA,$59,$FF,$6B,$8C,$36,$1F,$43,$62,$57,$F8,$86
	.dc.b	$B6,$DA,$E6,$D9,$FE,$A9,$3A,$CF,$D2,$F8,$7C,$2E,$71,$8B,$FE,$2D
	.dc.b	$C7,$FC,$F7,$D9,$FE,$83,$3F,$03,$44,$62,$75,$F1,$5F,$E0,$18,$C6
	.dc.b	$7F,$E7,$5A,$DF,$97,$93,$55,$67,$E3,$7C,$BF,$E2,$70,$FE,$98,$FF
	.dc.b	$7E,$C7,$C0,$E6,$7A,$39,$9D,$F2,$FF,$84,$63,$13,$A8,$C7,$C3,$FE
	.dc.b	$64,$91,$F7,$FD,$27,$5F,$F5,$9C,$7F,$77,$5A,$7F,$64,$FE,$DC,$95
	.dc.b	$FB,$77,$F4,$9C,$9B,$CB,$F4,$6F,$F8,$8F,$E8,$3F,$C9,$35,$F8,$35
	.dc.b	$F1,$D9,$8D,$35,$D3,$70,$FA,$13,$C2,$7F,$13,$F5,$3F,$EA,$16,$CF
	.dc.b	$E2,$61,$F1,$96,$24,$E8,$E3,$18,$8F,$C1,$FD,$27,$EC,$1B,$DD,$CF
	.dc.b	$CA,$FC,$A4,$E7,$4F,$F2,$DF,$42,$38,$5F,$F9,$DA,$31,$7F,$90,$9C
	.dc.b	$CF,$FA,$DE,$73,$8F,$E6,$67,$A3,$9F,$A2,$7D,$4F,$83,$0F,$D2,$46
	.dc.b	$27,$E5,$7C,$5F,$CD,$3E,$77,$D4,$D8,$C9,$32,$F9,$31,$97,$F3,$5F
	.dc.b	$4B,$F9,$E9,$FC,$B3,$61,$D4,$8F,$F1,$BC,$61,$D4,$67,$B7,$FC,$C1
	.dc.b	$18,$37,$19,$AF,$E0,$5A,$3E,$5F,$FC,$6E,$D8,$FF,$33,$37,$0F,$F3
	.dc.b	$33,$5F,$F3,$33,$9D,$FF,$C6,$0B,$E5,$FF,$CE,$0E,$6B,$F6,$0E,$5F
	.dc.b	$0F,$D0,$68,$DF,$D0,$D8,$C9,$E2,$8F,$29,$1F,$2F,$8F,$E8,$23,$82
	.dc.b	$63,$FC,$16,$FF,$97,$C7,$C0,$9F,$2F,$8F,$FD,$37,$E9,$DF,$31,$F9
	.dc.b	$6C,$3F,$D7,$7E,$7F,$FA,$FA,$73,$D3,$FE,$B8,$8F,$F6,$0F,$CE,$7F
	.dc.b	$85,$7E,$1D,$1B,$C3,$94,$63,$00,$C3,$9B,$D8,$CB,$88,$E1,$0B,$ED
	.dc.b	$7D,$70,$F1,$4F,$0C,$3B,$CE,$63,$A1,$C7,$E7,$BF,$D7,$F8,$74,$C7
	.dc.b	$B5,$6E,$31,$7C,$AF,$CF,$8F,$81,$ED,$8F,$81,$F5,$4E,$DB,$19,$9E
	.dc.b	$11,$83,$C7,$87,$81,$E6,$79,$A3,$71,$98,$E1,$7E,$4A,$E9,$8F,$AB
	.dc.b	$1F,$41,$2B,$9D,$7C,$8C,$39,$E3,$E5,$C7,$A1,$D6,$33,$8F,$AE,$33
	.dc.b	$3E,$B9,$CD,$FA,$23,$91,$9C,$7C,$51,$E0,$C7,$C4,$FA,$1F,$C2,$C6
	.dc.b	$A3,$06,$C8,$7D,$5D,$3C,$55,$DB,$1A,$7C,$08,$B3,$E3,$8B,$97,$BE
	.dc.b	$FB,$65,$EA,$9C,$E3,$E2,$C2,$E1,$E0,$AF,$85,$7C,$31,$F0,$4F,$53
	.dc.b	$EA,$46,$70,$F0,$C7,$E6,$8F,$B2,$53,$EF,$16,$BF,$3A,$DF,$A8,$FE
	.dc.b	$B6,$EC,$FD,$13,$87,$5E,$1E,$5C,$B6,$53,$CB,$3F,$CD,$37,$3F,$DB
	.dc.b	$36,$FF,$57,$D3,$E7,$AD,$E1,$F7,$EB,$55,$D8,$FC,$AD,$6E,$F9,$D6
	.dc.b	$C9,$3A,$2D,$F4,$7F,$3B,$7D,$30,$61,$CE,$7F,$7B,$39,$25,$FE,$46
	.dc.b	$75,$3E,$3B,$E8,$4C,$3F,$89,$4E,$B0,$F8,$77,$E2,$9F,$3D,$F9,$4F
	.dc.b	$1C,$65,$F0,$BD,$DC,$1D,$E3,$8D,$FF,$05,$BE,$D8,$79,$AB,$99,$D6
	.dc.b	$7C,$8F,$43,$2F,$17,$AE,$1E,$67,$37,$EA,$73,$5D,$6B,$84,$68,$E8
	.dc.b	$F6,$7F,$94,$5F,$93,$0E,$27,$AE,$38,$1E,$B3,$8D,$F7,$73,$87,$8E
	.dc.b	$FA,$3F,$5A,$39,$9A,$BF,$B1,$CC,$F9,$16,$DF,$A1,$E6,$B7,$FD,$D2
	.dc.b	$7F,$C5,$2F,$F9,$B5,$AF,$F9,$55,$DB,$67,$F3,$D5,$D5,$B3,$FE,$B3
	.dc.b	$8C,$4F,$F7,$2E,$8D,$F2,$7F,$80,$63,$04,$3F,$BA,$C7,$96,$FF,$B7
	.dc.b	$B6,$3F,$C1,$33,$C5,$FE,$C9,$F4,$D9,$BF,$F1,$47,$C8,$DA,$FE,$A1
	.dc.b	$5F,$E0,$6E,$4C,$4C,$8F,$F0,$2B,$E4,$23,$FB,$DA,$FF,$0B,$79,$DD
	.dc.b	$71,$FF,$45,$CF,$FA,$45,$AF,$F8,$83,$CF,$D4,$C8,$D3,$C6,$3C,$DE
	.dc.b	$2F,$FA,$D6,$7F,$CB,$1E,$0E,$0C,$FC,$65,$8C,$90,$FF,$2B,$BF,$0F
	.dc.b	$C4,$7D,$2F,$FA,$22,$7F,$D4,$3E,$23,$E9,$F4,$D8,$FF,$52,$CE,$DF
	.dc.b	$F4,$BB,$5F,$F2,$73,$5F,$F2,$63,$55,$78,$3D,$1F,$E3,$CD,$8E,$B1
	.dc.b	$E5,$9E,$DC,$BF,$E5,$CE,$37,$D3,$FA,$7E,$97,$F1,$73,$FE,$60,$6F
	.dc.b	$37,$F9,$6D,$B8,$F8,$1F,$C9,$CE,$E7,$91,$27,$FD,$83,$3A,$8F,$F5
	.dc.b	$36,$0F,$0F,$F0,$CB,$C4,$A6,$22,$5F,$FB,$36,$74,$D9,$CB,$FE,$D6
	.dc.b	$1F,$4E,$F2,$3F,$2D,$18,$E5,$B3,$F8,$6C,$FF,$52,$EE,$7F,$DC,$87
	.dc.b	$2A,$FE,$90,$D3,$FA,$54,$6D,$FE,$A0,$D7,$FA,$C3,$99,$EB,$3E,$03
	.dc.b	$F5,$1E,$83,$E9,$7E,$5F,$4B,$D3,$7C,$77,$EA,$57,$CB,$D3,$11,$33
	.dc.b	$FF,$65,$C6,$5F,$F6,$6C,$7E,$C7,$E9,$DE,$65,$24,$FF,$BD,$E1,$E6
	.dc.b	$62,$3F,$DC,$11,$C1,$D4,$F3,$9F,$C3,$7C,$27,$E1,$FE,$66,$57,$DE
	.dc.b	$8C,$BF,$EF,$7E,$64,$7E,$51,$B1,$98,$C1,$C1,$FF,$7B,$C6,$9B,$D3
	.dc.b	$FC,$2F,$DC,$7C,$C5,$F8,$29,$5F,$E6,$2E,$4F,$E5,$3C,$2B,$DC,$E7
	.dc.b	$AB,$FD,$ED,$18,$8F,$F7,$F2,$BE,$71,$8C,$3F,$8D,$73,$BE,$9E,$92
	.dc.b	$7A,$1F,$EA,$A8,$7A,$39,$86,$7E,$49,$E5,$1C,$33,$46,$23,$79,$FF
	.dc.b	$9F,$27,$37,$F9,$8F,$95,$FC,$4F,$C5,$BE,$76,$99,$C3,$F3,$3D,$8A
	.dc.b	$FC,$CB,$63,$F3,$31,$82,$3C,$1E,$4F,$88,$75,$FE,$5F,$D5,$FD,$4F
	.dc.b	$D3,$74,$24,$FE,$91,$FD,$3F,$D3,$FF,$3E,$8F,$F1,$A3,$5F,$F1,$E3
	.dc.b	$5F,$F2,$03,$5F,$F4,$DB,$FE,$C9,$7F,$C7,$0D,$79,$FB,$3D,$8E,$FE
	.dc.b	$CC,$8D,$B5,$F0,$FB,$9E,$42,$3D,$1F,$EC,$FF,$15,$FD,$B2,$3F,$CA
	.dc.b	$5C,$08,$C4,$7D,$36,$BF,$EE,$86,$C3,$E3,$3F,$B7,$13,$8B,$FB,$9B
	.dc.b	$FD,$C7,$AD,$F0,$9F,$DD,$9B,$87,$F5,$D6,$BF,$E6,$1B,$FD,$4C,$FF
	.dc.b	$7C,$F8,$D7,$D3,$52,$57,$F6,$58,$CB,$FD,$9F,$E3,$FF,$DA,$0F,$36
	.dc.b	$BE,$E6,$BF,$8F,$6A,$B7,$D3,$3E,$2F,$FC,$1A,$D9,$FE,$E8,$D7,$D0
	.dc.b	$D7,$E5,$FC,$63,$E2,$BF,$C2,$AD,$8F,$EF,$8D,$7E,$03,$5F,$A7,$F2
	.dc.b	$5F,$C9,$1F,$F0,$13,$5F,$8C,$D7,$7E,$4F,$F8,$8B,$E5,$3F,$9C,$E4
	.dc.b	$69,$27,$FC,$4F,$F4,$CF,$9C,$A4,$8F,$F0,$84,$65,$FF,$09,$7C,$BB
	.dc.b	$E7,$FF,$C5,$ED,$8C,$C6,$0F,$F1,$9F,$D2,$BD,$CA,$49,$FF,$0E,$46
	.dc.b	$5B,$1F,$E2,$98,$C4,$7E,$CF,$24,$63,$C6,$FE,$3E,$FF,$93,$E7,$5F
	.dc.b	$24,$FA,$5F,$F1,$97,$CE,$BC,$2A,$7F,$8D,$67,$55,$FC,$4D,$AF,$E4
	.dc.b	$5A,$FE,$91,$AF,$F1,$56,$BF,$87,$78,$E1,$FC,$C9,$BD,$9B,$C3,$C1
	.dc.b	$F4,$BF,$E4,$5F,$A0,$FE,$35,$87,$F2,$35,$F1,$B9,$35,$E4,$C9,$FF
	.dc.b	$21,$4F,$07,$F4,$CD,$7F,$72,$D4,$7F,$1A,$EC,$FE,$B8,$FF,$B3,$7C
	.dc.b	$74,$62,$3F,$DA,$51,$8F,$57,$FD,$90,$D8,$FF,$67,$35,$FE,$4C,$DD
	.dc.b	$7F,$CF,$2D,$D3,$F9,$D7,$D0,$FB,$2B,$FC,$F5,$EA,$FF,$9D,$08,$FF
	.dc.b	$97,$7E,$2D,$18,$8F,$F3,$F3,$FE,$6B,$8C,$C7,$FA,$0F,$D5,$FF,$0E
	.dc.b	$FD,$1F,$F8,$1B,$2F,$7F,$D1,$BC,$AF,$F7,$60,$F8,$BF,$F7,$74,$67
	.dc.b	$0F,$8B,$39,$78,$BF,$EF,$3B,$E9,$09,$E0,$A3,$31,$89,$E5,$3B,$7C
	.dc.b	$35,$E0,$7A,$7B,$9F,$D9,$27,$47,$D6,$BE,$FD,$9C,$EF,$EC,$CF,$EB
	.dc.b	$EF,$9B,$F7,$27,$C7,$5C,$27,$AE,$3E,$AA,$F6,$DF,$85,$A6,$4F,$56
	.dc.b	$1E,$3C,$79,$B6,$FB,$CF,$98,$F9,$95,$CD,$E3,$7C,$F0,$F8,$F1,$E8
	.dc.b	$32,$52,$99,$76,$FC,$FF,$5F,$FB,$7B,$B8,$FF,$0F,$F0,$71,$8F,$A1
	.dc.b	$3A,$79,$35,$ED,$F2,$1F,$E5,$84,$D2,$3F,$E5,$C7,$E1,$5B,$E8,$BF
	.dc.b	$DA,$3A,$7D,$EA,$FD,$4D,$6A,$73,$9F,$69,$F8,$97,$C7,$8B,$E6,$47
	.dc.b	$03,$35,$C1,$F8,$97,$F6,$6F,$FD,$C1,$3F,$EE,$0D,$BC,$15,$E8,$3F
	.dc.b	$00,$7F,$8F,$63,$53,$E1,$C7,$F2,$78,$F8,$6B,$EB,$1E,$18,$CD,$78
	.dc.b	$63,$F1,$78,$7A,$F1,$F3,$C7,$7A,$DC,$7E,$27,$BB,$FD,$ED,$F2,$1F
	.dc.b	$EF,$4C,$7F,$C9,$BC,$48,$C5,$7F,$8D,$E1,$F2,$B8,$C7,$FD,$F5,$57
	.dc.b	$0E,$77,$FE,$FB,$8C,$5F,$08,$C1,$FE,$45,$E5,$FE,$E0,$DF,$2F,$94
	.dc.b	$3E,$66,$C4,$9F,$F7,$34,$F0,$C7,$FC,$14,$D7,$FC,$3A,$D7,$FC,$9D
	.dc.b	$D4,$F9,$A5,$7F,$2E,$6C,$FF,$2E,$6B,$FC,$B9,$AF,$27,$DD,$F7,$CF
	.dc.b	$9F,$F8,$17,$D6,$FF,$CC,$13,$97,$FC,$C5,$F4,$9F,$E7,$C6,$F3,$7C
	.dc.b	$78,$FF,$3A,$46,$09,$8F,$F8,$96,$38,$36,$7C,$6F,$F8,$66,$32,$FF
	.dc.b	$8F,$1D,$F0,$FF,$A1,$BE,$41,$F3,$1D,$EF,$EB,$46,$30,$FF,$3B,$7D
	.dc.b	$27,$FA,$C9,$B3,$F6,$E3,$18,$FF,$75,$7A,$CF,$F9,$E5,$F0,$C5,$9F
	.dc.b	$2F,$1A,$73,$3F,$E2,$F6,$C3,$E9,$FF,$D1,$77,$FE,$87,$6D,$77,$BF
	.dc.b	$C6,$46,$32,$FF,$50,$F9,$EF,$B5,$FC,$83,$FE,$99,$C3,$FD,$20,$E5
	.dc.b	$BF,$0B,$F9,$5F,$C2,$BF,$B3,$C7,$F1,$7C,$D1,$8C,$FF,$CA,$8D,$7F
	.dc.b	$CC,$ED,$7F,$D3,$1E,$A7,$F8,$FE,$BF,$D3,$CD,$8F,$F0,$D3,$F8,$36
	.dc.b	$FA,$DF,$A5,$FA,$67,$D3,$DD,$30,$DC,$3F,$D5,$1F,$0D,$A7,$CC,$DC
	.dc.b	$3F,$CD,$0D,$7F,$CF,$6D,$7F,$CB,$DF,$2B,$FB,$3F,$70,$DF,$94,$C3
	.dc.b	$FB,$7F,$39,$F3,$17,$D8,$C3,$72,$FF,$5D,$7C,$0F,$EE,$72,$FF,$3C
	.dc.b	$35,$FF,$45,$B5,$FF,$5E,$35,$FF,$37,$7C,$13,$EE,$FF,$83,$36,$BF
	.dc.b	$D8,$CB,$EA,$33,$3F,$C2,$FE,$15,$EC,$7E,$31,$B7,$F6,$5B,$87,$FA
	.dc.b	$2D,$AF,$FA,$7B,$EE,$FF,$D1,$73,$F7,$DB,$1F,$E9,$16,$BB,$F8,$5F
	.dc.b	$CC,$FD,$93,$E6,$29,$2F,$F1,$11,$97,$F1,$5F,$66,$F9,$BF,$D0,$B6
	.dc.b	$33,$18,$3F,$4B,$F7,$7F,$E1,$78,$FC,$93,$71,$FF,$52,$35,$FF,$5B
	.dc.b	$B5,$FF,$74,$7D,$AF,$F3,$59,$FC,$D3,$63,$FD,$58,$D7,$FD,$7E,$D7
	.dc.b	$FC,$1F,$F5,$7F,$E5,$29,$FF,$7E,$C2,$30,$FF,$B9,$7C,$37,$D4,$FF
	.dc.b	$A0,$EB,$75,$A9,$FE,$4E,$F0,$AF,$F0,$23,$5E,$FF,$05,$FC,$61,$EC
	.dc.b	$D8,$E5,$5F,$E5,$AF,$A0,$7E,$17,$F9,$48,$EF,$5F,$D3,$E7,$3D,$1F
	.dc.b	$D3,$FF,$0C,$FE,$77,$5F,$E6,$2C,$BF,$9D,$F8,$C9,$7D,$6E,$EB,$83
	.dc.b	$5E,$07,$F9,$AA,$78,$46,$74,$FF,$4A,$57,$2A,$DC,$7F,$B2,$7E,$71
	.dc.b	$F2,$FF,$D8,$5B,$3A,$C1,$1F,$EB,$18,$C3,$AA,$FF,$50,$37,$1F,$4B
	.dc.b	$67,$C8,$F1,$9D,$DF,$C3,$D7,$FD,$61,$5E,$4D,$3E,$4F,$76,$7D,$AD
	.dc.b	$32,$7F,$C6,$D3,$A6,$D7,$2C,$3F,$C7,$93,$96,$C6,$72,$FE,$CF,$2F
	.dc.b	$C3,$7F,$82,$46,$27,$57,$F6,$E7,$2F,$CD,$F8,$67,$D9,$FF,$81,$D7
	.dc.b	$85,$63,$0D,$9F,$95,$3A,$6C,$E9,$F9,$DF,$45,$FE,$03,$C7,$F0,$2D
	.dc.b	$8F,$B8,$FF,$82,$DD,$FB,$9F,$E1,$78,$D6,$8E,$14,$61,$BE,$EE,$EB
	.dc.b	$FC,$41,$4F,$B0,$FE,$F2,$B1,$FF,$04,$C6,$1B,$F1,$3C,$23,$19,$7F
	.dc.b	$A0,$1E,$35,$FE,$6B,$6B,$9F,$A0,$FD,$44,$62,$3F,$D7,$EF,$25,$F1
	.dc.b	$0F,$97,$7D,$B7,$7B,$FF,$1B,$4E,$63,$FC,$6D,$F5,$3F,$C8,$B3,$FD
	.dc.b	$3B,$72,$FD,$13,$FE,$EB,$7F,$34,$E7,$1F,$8C,$DD,$8D,$38,$47,$FB
	.dc.b	$ED,$E1,$CB,$FD,$07,$E0,$BF,$75,$E5,$BE,$CA,$C6,$1B,$8F,$E6,$E7
	.dc.b	$4D,$9D,$39,$D7,$99,$D3,$E8,$EC,$64,$FC,$FD,$7C,$9C,$3F,$A9,$57
	.dc.b	$F8,$A9,$B3,$A8,$F2,$FD,$5F,$2C,$FE,$E4,$FF,$15,$B4,$FC,$AF,$BB
	.dc.b	$FF,$30,$7D,$6F,$F8,$AE,$FE,$04,$63,$E1,$35,$87,$F0,$3A,$D7,$07
	.dc.b	$D0,$AD,$4E,$7D,$CE,$7F,$11,$D1,$B1,$93,$94,$65,$A7,$F0,$FF,$AF
	.dc.b	$FF,$72,$3F,$E1,$D6,$E5,$FE,$B7,$E3,$46,$23,$FD,$A3,$18,$F9,$0F
	.dc.b	$F1,$CC,$F0,$CB,$F9,$5B,$FE,$31,$BF,$E4,$CE,$5B,$F4,$FF,$E8,$18
	.dc.b	$FE,$8F,$1E,$3F,$33,$1E,$7E,$59,$1E,$57,$B9,$1F,$47,$B7,$E2,$23
	.dc.b	$82,$67,$F5,$0F,$DC,$5B,$ED,$FC,$55,$E3,$7F,$6F,$F6,$2F,$9C,$7D
	.dc.b	$9F,$FB,$03,$D1,$FF,$78,$CE,$63,$18,$87,$D9,$7F,$95,$3E,$55,$7C
	.dc.b	$30,$E7,$5F,$ED,$28,$C5,$F2,$86,$1D,$5D,$1D,$70,$E9,$5C,$23,$27
	.dc.b	$0B,$EB,$87,$82,$3C,$97,$CE,$31,$1E,$6D,$FF,$D3,$71,$89,$D5,$F5
	.dc.b	$8C,$5F,$48,$F7,$5F,$1D,$7D,$17,$C2,$3B,$D6,$DB,$97,$B7,$2F,$4C
	.dc.b	$60,$F4,$D7,$17,$D5,$3B,$CB,$CB,$8F,$8C,$93,$C7,$2F,$5E,$1D,$23
	.dc.b	$89,$D3,$1F,$13,$89,$DD,$F4,$8E,$57,$D1,$D3,$EE,$BE,$98,$F9,$49
	.dc.b	$87,$48,$CB,$DA,$33,$8F,$A7,$2F,$8B,$39,$24,$67,$0F,$33,$A2,$3F
	.dc.b	$72,$FA,$C6,$32,$FA,$6F,$16,$CF,$22,$3E,$EB,$E1,$8F,$B7,$1F,$9D
	.dc.b	$87,$4B,$DC,$FC,$18,$C5,$F4,$24,$7C,$3C,$BE,$45,$7C,$63,$ED,$1F
	.dc.b	$84,$3E,$89,$6F,$F3,$4D,$CB,$F4,$2E,$7E,$CF,$FD,$9E,$F1,$3F,$42
	.dc.b	$E1,$D7,$46,$5C,$B6,$93,$CB,$1F,$CD,$37,$1F,$DB,$B6,$FF,$47,$E7
	.dc.b	$FE,$6A,$B7,$87,$E6,$AB,$55,$D8,$FC,$AD,$6E,$F9,$D6,$C9,$3A,$2D
	.dc.b	$F4,$7F,$3B,$7D,$30,$61,$CE,$7F,$6B,$39,$25,$FE,$B2,$75,$3E,$3B
	.dc.b	$E8,$4C,$3F,$89,$CE,$B0,$F8,$77,$E2,$9F,$3D,$F9,$4F,$1C,$65,$F0
	.dc.b	$BD,$DC,$1D,$E3,$8D,$FF,$05,$BE,$D8,$79,$AB,$99,$D6,$7C,$8F,$43
	.dc.b	$2F,$17,$AE,$1E,$67,$37,$EA,$73,$5D,$6B,$84,$68,$E8,$F6,$7F,$93
	.dc.b	$DF,$93,$0E,$27,$AE,$38,$1E,$B3,$8D,$F7,$73,$87,$8E,$FA,$3F,$5A
	.dc.b	$39,$9A,$BF,$B1,$CB,$FD,$93,$A7,$F9,$1F,$2F,$F6,$0F,$B7,$BC,$1F
	.dc.b	$D8,$1B,$87,$F5,$0A,$B7,$C6,$3F,$B8,$47,$E7,$63,$11,$FD,$C7,$91
	.dc.b	$39,$F0,$5F,$3B,$49,$5F,$A9,$BF,$F1,$34,$62,$3F,$B0,$46,$1F,$F4
	.dc.b	$74,$F2,$6F,$D7,$F9,$30,$FD,$D6,$7F,$DE,$BE,$37,$06,$6A,$1F,$E0
	.dc.b	$73,$D1,$3F,$B5,$7F,$B6,$9E,$BC,$9F,$21,$F0,$17,$85,$4C,$44,$9F
	.dc.b	$F1,$04,$65,$FF,$11,$47,$C8,$F1,$9F,$41,$7E,$12,$57,$F9,$42,$34
	.dc.b	$61,$FF,$26,$46,$1F,$F5,$07,$E1,$DF,$EB,$26,$FD,$3F,$D6,$F2,$12
	.dc.b	$8C,$3B,$9F,$F0,$54,$EB,$D9,$FE,$28,$D8,$FF,$53,$CE,$DF,$EC,$F1
	.dc.b	$97,$FD,$4F,$1C,$F7,$FF,$52,$B7,$97,$FD,$B8,$D9,$53,$F0,$FF,$AB
	.dc.b	$5F,$EE,$15,$F5,$9E,$95,$E3,$53,$11,$2F,$FD,$81,$3A,$6C,$E5,$FC
	.dc.b	$3F,$DA,$FE,$41,$FF,$6A,$35,$CF,$17,$E6,$A1,$F1,$B6,$24,$F0,$75
	.dc.b	$1F,$A2,$87,$D8,$FB,$7D,$27,$C8,$3E,$D7,$F4,$F1,$87,$84,$FC,$3A
	.dc.b	$E3,$C3,$F8,$F8,$F9,$D1,$FE,$14,$6B,$FE,$33,$6B,$C0,$FF,$79,$94
	.dc.b	$FF,$22,$17,$85,$BF,$F8,$7B,$EC,$7E,$7C,$FF,$AF,$FD,$BF,$F4,$93
	.dc.b	$FE,$CA,$ED,$FD,$DC,$62,$3F,$5E,$BE,$8F,$33,$27,$81,$FE,$D2,$8F
	.dc.b	$E0,$7F,$68,$F0,$AF,$9B,$2B,$FC,$99,$9F,$91,$FF,$68,$56,$DC,$E7
	.dc.b	$FE,$51,$F0,$93,$99,$FE,$02,$FF,$97,$DD,$9C,$AF,$C3,$87,$E2,$78
	.dc.b	$BF,$45,$F0,$3F,$AE,$AD,$F4,$FF,$BE,$7B,$75,$1F,$8A,$AF,$D0,$F4
	.dc.b	$FE,$0A,$3F,$BD,$57,$E5,$E7,$24,$8E,$04,$3F,$57,$2F,$3F,$F9,$2F
	.dc.b	$C2,$3E,$BE,$92,$BD,$3B,$25,$EF,$66,$7F,$8F,$C7,$17,$6D,$73,$1F
	.dc.b	$CC,$BD,$A7,$CB,$5F,$97,$95,$DA,$72,$D9,$CB,$FC,$B2,$13,$CE,$7F
	.dc.b	$3D,$F6,$EF,$9F,$FD,$A7,$C9,$4C,$3F,$23,$39,$6C,$FF,$29,$8C,$47
	.dc.b	$F8,$8D,$AF,$F8,$B5,$AF,$F9,$C9,$DF,$3F,$E1,$61,$18,$E7,$D5,$7F
	.dc.b	$52,$A7,$B2,$FA,$BF,$EA,$7F,$20,$F8,$8F,$EA,$4D,$CB,$FA,$63,$5F
	.dc.b	$F6,$23,$5F,$F6,$1F,$B5,$FB,$97,$FA,$93,$5F,$F6,$53,$5D,$F8,$7F
	.dc.b	$D7,$FE,$4B,$F8,$CC,$7F,$59,$6B,$FE,$D1,$6B,$FE,$D0,$6B,$F9,$FF
	.dc.b	$93,$AB,$95,$18,$6F,$97,$FD,$B6,$31,$E8,$FE,$3E,$BF,$B4,$35,$FF
	.dc.b	$70,$35,$FF,$6F,$B5,$FD,$77,$DF,$3D,$5F,$EF,$0D,$8F,$EE,$4D,$7F
	.dc.b	$DD,$4D,$7F,$DD,$3F,$88,$67,$87,$F7,$99,$23,$1C,$7F,$BB,$6D,$7B
	.dc.b	$5F,$EF,$0F,$F7,$CF,$8E,$7C,$4F,$F8,$25,$B1,$FE,$03,$6B,$FE,$FB
	.dc.b	$6B,$FE,$FA,$F9,$BF,$C4,$63,$E2,$5F,$27,$FE,$16,$6C,$FE,$C1,$DE
	.dc.b	$3F,$C3,$FA,$FF,$B3,$4F,$ED,$1A,$FF,$84,$1B,$7E,$EF,$9C,$FE,$57
	.dc.b	$A3,$EC,$DF,$13,$FE,$21,$6E,$1F,$E1,$B6,$BF,$E8,$D9,$E1,$3F,$E2
	.dc.b	$7F,$8E,$FC,$76,$1F,$29,$7D,$B7,$F8,$B9,$B7,$FC,$21,$DE,$7F,$C9
	.dc.b	$FE,$83,$F9,$C5,$FD,$26,$C6,$7D,$0F,$F1,$BC,$7F,$3B,$BF,$E2,$3E
	.dc.b	$13,$E3,$3F,$C7,$0D,$8F,$F1,$93,$5F,$43,$5F,$F7,$4E,$8F,$B4,$7C
	.dc.b	$9F,$F9,$11,$FF,$00,$46,$7D,$7C,$4F,$F9,$23,$E9,$3F,$BC,$5F,$F8
	.dc.b	$F1,$BA,$FE,$29,$BA,$FF,$43,$FB,$37,$CD,$FF,$94,$1B,$7F,$E4,$76
	.dc.b	$BF,$35,$AF,$99,$AF,$3F,$4B,$94,$C8,$FE,$67,$3F,$D1,$DF,$F2,$7F
	.dc.b	$D9,$3E,$57,$FC,$BF,$F6,$7F,$E6,$B9,$FF,$29,$B7,$3F,$D0,$D6,$E7
	.dc.b	$F9,$EB,$5F,$C0,$35,$F9,$2D,$7F,$B0,$35,$FE,$04,$F1,$8E,$B5,$FE
	.dc.b	$70,$8C,$65,$C3,$E8,$FF,$CE,$7F,$57,$FE,$CF,$BF,$D9,$FD,$5B,$EC
	.dc.b	$3F,$CE,$ED,$CB,$FA,$8D,$6F,$0F,$EE,$0F,$FB,$5E,$3E,$EB,$5C,$FC
	.dc.b	$9F,$F9,$F6,$31,$1F,$C5,$FE,$B7,$A3,$F9,$D6,$BF,$86,$6B,$FD,$F5
	.dc.b	$AF,$1C,$3F,$D1,$71,$CB,$1F,$E1,$6D,$C3,$87,$9F,$FE,$90,$FA,$EF
	.dc.b	$E7,$55,$FC,$41,$B3,$FE,$09,$6E,$1F,$CA,$D7,$89,$C7,$D0,$93,$FD
	.dc.b	$CA,$77,$D0,$77,$1F,$E2,$EF,$B1,$FE,$EB,$8A,$1F,$63,$F9,$AF,$8B
	.dc.b	$7D,$96,$67,$31,$4C,$BC,$67,$F4,$37,$D2,$36,$75,$71,$1D,$6B,$A3
	.dc.b	$67,$95,$77,$AD,$B7,$DA,$61,$E7,$E0,$E1,$7F,$83,$AF,$26,$3E,$79
	.dc.b	$F1,$DF,$0A,$D1,$E3,$71,$8F,$22,$5F,$7C,$3C,$75,$B6,$BF,$E7,$AC
	.dc.b	$3D,$33,$E0,$8C,$E3,$E8,$78,$1F,$24,$A5,$7E,$33,$70,$F3,$39,$F9
	.dc.b	$17,$C5,$71,$8F,$95,$18,$71,$5F,$EB,$A9,$D4,$23,$0F,$FB,$07,$E5
	.dc.b	$FF,$D8,$49,$A4,$74,$7D,$83,$E7,$FC,$47,$F9,$77,$4F,$E3,$B1,$F8
	.dc.b	$79,$F8,$7A,$7F,$9A,$F8,$BE,$2C,$E4,$E5,$19,$7A,$19,$AF,$E4,$B7
	.dc.b	$D7,$4F,$76,$1D,$23,$15,$F2,$63,$18,$76,$8F,$51,$DA,$76,$FE,$5E
	.dc.b	$76,$F8,$0D,$DF,$82,$B9,$4E,$A3,$FD,$61,$C1,$FA,$7F,$96,$FF,$3E
	.dc.b	$65,$FB,$36,$DF,$EC,$FE,$CA,$E3,$FB,$39,$B7,$C6,$BF,$65,$18,$E2
	.dc.b	$5D,$8D,$F3,$7F,$4E,$AF,$F3,$4B,$70,$FC,$FC,$7F,$0D,$FA,$3D,$F8
	.dc.b	$72,$36,$F2,$8C,$4F,$03,$FC,$61,$D3,$FD,$8E,$BF,$92,$F8,$FE,$68
	.dc.b	$ED,$AF,$F8,$79,$D5,$71,$D5,$18,$9D,$57,$1C,$79,$3F,$03,$EC,$9C
	.dc.b	$B4,$C7,$AF,$73,$FF,$42,$CE,$9B,$3A,$69,$33,$F1,$7B,$EE,$EE,$51
	.dc.b	$87,$E7,$FD,$97,$FA,$3A,$BF,$D2,$2D,$9F,$52,$C7,$7D,$5C,$31,$97
	.dc.b	$C2,$2D,$FF,$79,$F7,$BE,$DF,$92,$73,$BF,$F7,$B7,$EF,$FD,$97,$FA
	.dc.b	$3B,$0F,$F5,$0B,$6F,$E5,$3C,$63,$3E,$AF,$E1,$BE,$CB,$F9,$84,$7F
	.dc.b	$AA,$1B,$5C,$3E,$6F,$4F,$E3,$3D,$1F,$F1,$9D,$7F,$AB,$9B,$1F,$5D
	.dc.b	$FA,$D1,$FE,$29,$F5,$7F,$27,$F6,$5F,$EA,$A8,$FF,$5A,$BE,$5D,$3F
	.dc.b	$CB,$AE,$6F,$F0,$0E,$67,$FD,$B1,$F2,$B9,$AF,$F3,$1B,$63,$39,$FE
	.dc.b	$82,$31,$7F,$E7,$FF,$65,$F1,$BF,$A4,$6D,$FF,$B1,$9F,$F6,$63,$63
	.dc.b	$F5,$3E,$DF,$F8,$92,$3F,$CB,$9F,$57,$8F,$53,$2F,$EB,$BC,$DF,$D5
	.dc.b	$CF,$FB,$41,$B8,$FF,$9C,$7E,$9B,$5E,$27,$ED,$3E,$C9,$E9,$5F,$53
	.dc.b	$0D,$C7,$FD,$B3,$F6,$4F,$9E,$FD,$DB,$63,$3A,$E2,$7F,$7F,$F6,$4F
	.dc.b	$B2,$A6,$3E,$9E,$DF,$FB,$93,$E2,$5F,$21,$FC,$19,$B1,$9F,$8F,$C4
	.dc.b	$FF,$0A,$F8,$D7,$D5,$53,$1F,$33,$6F,$FD,$DF,$F6,$5F,$DD,$9B,$ED
	.dc.b	$FF,$12,$87,$D5,$58,$FE,$2B,$F2,$6F,$5A,$98,$E7,$B8,$7F,$80,$FC
	.dc.b	$1F,$F6,$C7,$D9,$3E,$AB,$F8,$FF,$62,$31,$5B,$38,$70,$FF,$8D,$5C
	.dc.b	$4F,$F8,$36,$76,$FF,$15,$39,$57,$48,$CD,$FF,$30,$BE,$55,$B9,$CD
	.dc.b	$FF,$84,$DB,$1D,$6F,$F9,$9D,$FF,$BA,$BE,$03,$A4,$7F,$34,$FB,$26
	.dc.b	$17,$E4,$30,$D3,$CF,$EC,$A7,$71,$87,$FD,$62,$D3,$86,$3F,$C9,$FE
	.dc.b	$65,$F1,$7F,$D0,$9B,$3D,$17,$D3,$E6,$64,$E5,$E1,$18,$AF,$E7,$5F
	.dc.b	$3C,$FB,$27,$8F,$FD,$42,$30,$D8,$FB,$98,$7D,$09,$CB,$63,$8B,$67
	.dc.b	$87,$1F,$16,$47,$1F,$A0,$FD,$5F,$0F,$C8,$C5,$8F,$47,$CB,$3A,$AF
	.dc.b	$F2,$56,$CF,$B2,$7C,$57,$F6,$47,$FC,$AF,$3F,$39,$CB,$7C,$AD,$4E
	.dc.b	$DB,$19,$3E,$AF,$D2,$7F,$58,$7B,$B7,$16,$08,$7C,$1F,$F7,$47,$51
	.dc.b	$9F,$82,$C1,$99,$B7,$F9,$8B,$7D,$97,$F4,$98,$C3,$63,$F3,$4F,$F4
	.dc.b	$FD,$BE,$34,$70,$F7,$BF,$4F,$3B,$9C,$B6,$BF,$CD,$8F,$16,$CF,$02
	.dc.b	$3E,$0F,$76,$DF,$F7,$5F,$AB,$FE,$21,$3F,$98,$6C,$67,$C8,$FF,$09
	.dc.b	$C6,$31,$FB,$5A,$23,$17,$F6,$1B,$7F,$29,$5F,$2F,$27,$F2,$1E,$7B
	.dc.b	$EC,$AF,$CB,$E1,$B8,$FE,$46,$74,$D9,$D3,$4D,$6B,$D0,$EB,$F4,$56
	.dc.b	$32,$4C,$FF,$47,$69,$D4,$62,$AE,$BF,$97,$E7,$FF,$6F,$56,$30,$F7
	.dc.b	$7A,$9F,$E3,$6F,$AA,$FF,$06,$47,$C9,$8C,$7C,$2F,$C8,$6C,$64,$FE
	.dc.b	$0B,$3C,$32,$FF,$13,$3F,$DC,$B1,$F9,$AF,$F7,$28,$FC,$9B,$99,$FD
	.dc.b	$2B,$9E,$13,$99,$DF,$C4,$79,$7D,$EF,$33,$FE,$4C,$FA,$D7,$C8,$53
	.dc.b	$1F,$13,$72,$FE,$33,$3A,$6C,$E9,$A6,$BC,$BE,$87,$5F,$4A,$C6,$49
	.dc.b	$97,$F9,$0E,$FF,$A9,$C6,$35,$FC,$5C,$7F,$7A,$9D,$7D,$37,$9F,$D5
	.dc.b	$F4,$32,$FC,$AE,$5A,$7E,$BF,$EB,$FF,$1F,$F6,$3F,$E8,$DF,$3B,$FD
	.dc.b	$A1,$18,$C3,$FD,$31,$1F,$ED,$3F,$53,$FC,$59,$18,$3F,$DB,$53,$BF
	.dc.b	$92,$FE,$B1,$18,$EE,$61,$D6,$30,$D7,$FB,$3F,$C5,$7F,$1F,$AE,$30
	.dc.b	$BE,$98,$F9,$30,$E7,$3E,$13,$4E,$2B,$8E,$3E,$67,$9D,$F2,$AF,$09
	.dc.b	$C5,$C3,$E2,$CB,$C9,$18,$25,$F5,$C3,$E1,$63,$E7,$7C,$99,$70,$9D
	.dc.b	$61,$DB,$1F,$1D,$EE,$31,$7C,$1F,$15,$70,$7C,$46,$2F,$A1,$1F,$81
	.dc.b	$97,$B6,$B9,$61,$EE,$BE,$58,$FB,$2B,$57,$E0,$21,$F2,$30,$FA,$51
	.dc.b	$A8,$C1,$E1,$25,$FA,$B2,$F1,$3D,$EA,$E3,$F1,$26,$D7,$6C,$3C,$59
	.dc.b	$7C,$18,$EB,$87,$17,$AD,$F1,$69,$C7,$2F,$89,$97,$97,$0F,$A7,$97
	.dc.b	$6A,$F8,$F8,$73,$26,$1F,$10,$E8,$E2,$3A,$4F,$C0,$C7,$C9,$7D,$27
	.dc.b	$FE,$00,$3F,$A1,$47,$D7,$29,$F8,$A2,$9F,$97,$FB,$1F,$F5,$43,$C9
	.dc.b	$AE,$A1,$18,$8F,$D7,$B5,$FD,$5B,$CA,$FF,$32,$DC,$37,$D3,$F9,$6F
	.dc.b	$33,$FD,$E1,$18,$BF,$CE,$56,$DE,$75,$B7,$F3,$B1,$97,$D1,$18,$34
	.dc.b	$F2,$70,$D3,$F5,$15,$E0,$7A,$99,$BF,$3C,$F2,$6C,$66,$FC,$D1,$FB
	.dc.b	$4A,$E5,$3E,$23,$47,$48,$D3,$E2,$79,$92,$7F,$8B,$CF,$2C,$3E,$6D
	.dc.b	$EE,$7B,$9A,$BC,$1D,$5C,$BF,$C6,$AB,$96,$1F,$16,$F9,$CF,$90,$FE
	.dc.b	$29,$3A,$BE,$71,$E9,$C3,$CE,$EE,$B8,$C7,$9E,$3B,$1C,$23,$38,$78
	.dc.b	$A3,$A9,$9C,$3C,$47,$8F,$0F,$04,$75,$8C,$3B,$C3,$A4,$EB,$0F,$04
	.dc.b	$F9,$6F,$E1,$3F,$CE,$7A,$5C,$3F,$D9,$23,$FA,$DF,$4F,$E7,$B8,$FF
	.dc.b	$A7,$B7,$87,$FC,$9F,$5F,$EC,$EA,$D9,$0F,$EB,$13,$B7,$FC,$8B,$1B
	.dc.b	$55,$FE,$24,$EF,$7D,$8E,$F1,$46,$2B,$FD,$0A,$D7,$8E,$3F,$E6,$08
	.dc.b	$7D,$37,$17,$58,$75,$C3,$06,$67,$B6,$3F,$B8,$9F,$ED,$78,$60,$CF
	.dc.b	$15,$7F,$BE,$94,$FE,$E8,$F9,$F3,$7C,$7D,$F0,$65,$7F,$86,$23,$46
	.dc.b	$1F,$F0,$CC,$61,$FF,$25,$7C,$F7,$C1,$F1,$BF,$DC,$51,$FE,$4D,$F8
	.dc.b	$1C,$4C,$8C,$10,$FF,$2A,$3F,$3F,$C4,$7D,$8F,$FA,$92,$7F,$D2,$8D
	.dc.b	$7F,$D3,$AD,$D3,$FC,$71,$F5,$9F,$0F,$8B,$F8,$FD,$F2,$F5,$BA,$5E
	.dc.b	$E7,$2D,$C3,$FD,$2B,$7C,$5F,$F4,$BC,$66,$F9,$47,$8A,$FC,$D5,$B9
	.dc.b	$2F,$A9,$CF,$87,$2F,$FD,$59,$39,$6C,$E5,$FF,$5B,$C7,$F8,$D9,$C9
	.dc.b	$FE,$CC,$FB,$1F,$CE,$3F,$88,$3F,$D2,$EE,$33,$FF,$3A,$B7,$3F,$F5
	.dc.b	$5B,$F9,$78,$CF,$0F,$E5,$63,$55,$CA,$31,$F1,$7F,$9F,$87,$D8,$FC
	.dc.b	$56,$EB,$98,$C7,$D8,$FC,$79,$C9,$FE,$A1,$8E,$06,$9E,$13,$1D,$3D
	.dc.b	$E7,$E8,$75,$7F,$AC,$7A,$BF,$55,$EA,$7E,$AA,$FF,$D7,$31,$8F,$9D
	.dc.b	$CB,$23,$27,$FA,$FA,$3F,$65,$F6,$2F,$87,$BF,$63,$2B,$FD,$EF,$39
	.dc.b	$6C,$E5,$F9,$D8,$A5,$5D,$2B,$F0,$3A,$FE,$5F,$DC,$74,$53,$38,$FE
	.dc.b	$02,$6B,$F7,$9F,$C4,$F9,$58,$E3,$FC,$CC,$5E,$4F,$CC,$F2,$AB,$F8
	.dc.b	$67,$C0,$3E,$C6,$93,$1F,$EF,$B8,$FF,$AD,$A5,$EC,$66,$7D,$71,$C5
	.dc.b	$DE,$3F,$EF,$08,$C4,$6A,$7F,$8D,$7A,$AF,$A9,$BF,$2B,$2B,$89,$FE
	.dc.b	$E8,$9D,$38,$E3,$FF,$57,$56,$E3,$E5,$CA,$BA,$CF,$E3,$FE,$C5,$EA
	.dc.b	$D2,$5F,$6D,$1E,$AB,$E9,$73,$3F,$A3,$17,$C9,$FF,$79,$35,$FD,$45
	.dc.b	$72,$9E,$2F,$FB,$C2,$74,$DB,$FE,$80,$EE,$BD,$9F,$16,$F8,$2F,$E7
	.dc.b	$CD,$8C,$C6,$3E,$0B,$FA,$17,$AA,$FB,$1A,$4A,$F8,$D0,$FA,$DF,$E9
	.dc.b	$1F,$21,$FC,$16,$7F,$22,$DC,$BF,$DE,$2D,$E5,$FF,$1D,$F1,$A3,$F8
	.dc.b	$45,$7E,$63,$E4,$74,$A5,$18,$6F,$91,$FD,$5E,$38,$E1,$FE,$FE,$F3
	.dc.b	$BF,$21,$7F,$6D,$AF,$E9,$1B,$1B,$F7,$3F,$B0,$7D,$8F,$F5,$AA,$FC
	.dc.b	$13,$5F,$D5,$B6,$59,$FA,$BE,$56,$D7,$E2,$7E,$C7,$FB,$0C,$FB,$9E
	.dc.b	$0D,$F5,$BF,$B7,$46,$A7,$DD,$C4,$FB,$1F,$EE,$4D,$8F,$CA,$35,$FD
	.dc.b	$EA,$FA,$CF,$EF,$1F,$37,$E2,$9F,$E0,$2B,$E4,$B2,$6B,$2F,$EF,$DD
	.dc.b	$CF,$91,$A4,$BF,$D1,$7C,$E7,$F1,$CF,$B1,$7C,$4F,$F8,$15,$B3,$FA
	.dc.b	$76,$BF,$C4,$17,$C4,$FF,$82,$FE,$C7,$F1,$53,$FA,$D7,$F1,$0D,$EB
	.dc.b	$D7,$6F,$F8,$47,$E8,$7F,$94,$4F,$F1,$A6,$F3,$7C,$EE,$C4,$63,$7F
	.dc.b	$C5,$AF,$4B,$FC,$35,$E3,$78,$B0,$FD,$E3,$5F,$E4,$AD,$C3,$A7,$D8
	.dc.b	$FF,$6B,$C4,$BE,$4F,$FC,$48,$DB,$FE,$0C,$D7,$F9,$7A,$F9,$3F,$F1
	.dc.b	$57,$D3,$7E,$53,$4F,$CF,$B6,$BF,$38,$DC,$1F,$62,$FA,$0F,$F1,$93
	.dc.b	$6B,$F8,$A3,$5F,$E7,$AD,$CF,$E2,$3F,$CE,$A3,$7E,$A7,$22,$3C,$BE
	.dc.b	$0F,$FC,$7D,$D8,$F9,$3F,$F2,$07,$D8,$BE,$36,$9A,$89,$7F,$C9,$6B
	.dc.b	$83,$63,$9F,$D1,$6A,$3F,$A8,$CE,$9B,$3A,$7F,$97,$7D,$5F,$FA,$D3
	.dc.b	$4F,$B8,$DE,$0F,$E0,$AD,$9F,$E0,$4B,$E7,$BF,$CA,$CE,$78,$0D,$24
	.dc.b	$7F,$96,$FE,$B3,$85,$7F,$3D,$6B,$FD,$99,$B8,$7E,$35,$FE,$6D,$1B
	.dc.b	$9E,$44,$79,$7A,$9F,$E6,$7F,$AD,$FE,$55,$1F,$DC,$1B,$7F,$97,$6E
	.dc.b	$5F,$C8,$17,$8D,$C7,$D2,$97,$FD,$46,$76,$FF,$21,$76,$FF,$58,$1F
	.dc.b	$5F,$FD,$9B,$0F,$F7,$2F,$D7,$FF,$2A,$F8,$B7,$D2,$FF,$B9,$A3,$0D
	.dc.b	$8F,$F7,$4D,$71,$B5,$7F,$32,$8C,$3C,$23,$AC,$FF,$BC,$0E,$AE,$0F
	.dc.b	$F7,$AC,$F0,$8E,$BE,$E7,$FA,$3A,$7A,$C7,$1B,$F1,$47,$FB,$F2,$39
	.dc.b	$BA,$39,$CF,$1B,$F2,$36,$79,$4F,$84,$3C,$A7,$78,$67,$0F,$25,$77
	.dc.b	$C3,$C2,$F0,$6C,$6A,$3E,$6E,$1E,$94,$7B,$2A,$B0,$DB,$5F,$4B,$5F
	.dc.b	$AF,$F2,$0F,$B0,$C5,$FF,$A6,$1B,$19,$9E,$53,$97,$90,$F9,$6A,$D3
	.dc.b	$2E,$08,$FC,$B3,$FD,$7F,$E1,$FF,$56,$CF,$FB,$7C,$7B,$27,$E5,$63
	.dc.b	$FC,$A7,$83,$E6,$4E,$4E,$51,$97,$A1,$9A,$FE,$E3,$7D,$75,$F8,$58
	.dc.b	$74,$8C,$57,$C7,$8C,$61,$DA,$3D,$57,$CA,$31,$3A,$9E,$E6,$AF,$FD
	.dc.b	$C9,$7D,$23,$E9,$CE,$EF,$FD,$25,$CA,$DD,$F3,$BF,$DD,$B3,$FD,$A3
	.dc.b	$73,$FF,$09,$FD,$D3,$E6,$71,$9F,$EC,$6A,$E3,$CD,$B4,$F6,$73,$97
	.dc.b	$6C,$7F,$C9,$53,$97,$F8,$87,$C7,$F2,$E9,$37,$D1,$C3,$27,$FC,$A5
	.dc.b	$F7,$6F,$B8,$C6,$BF,$E7,$C8,$C4,$FF,$03,$FC,$1D,$F5,$5E,$8E,$87
	.dc.b	$DE,$BE,$0F,$D7,$8A,$F7,$87,$F9,$8A,$F7,$5A,$9C,$D7,$FA,$DE,$73
	.dc.b	$5D,$F0,$ED,$87,$58,$F1,$70,$F1,$3C,$5C,$9F,$E6,$08,$C9,$30,$E9
	.dc.b	$5F,$E8,$58,$C4,$FF,$8A,$B0,$F4,$FC,$5F,$39,$F3,$56,$EF,$FC,$D1
	.dc.b	$97,$CF,$79,$D7,$D6,$FB,$47,$C5,$53,$1F,$6F,$71,$FF,$3F,$4E,$9F
	.dc.b	$F4,$07,$DC,$BE,$87,$58,$FF,$AA,$A3,$13,$A9,$FC,$1F,$A2,$F9,$AF
	.dc.b	$C2,$D6,$AF,$FD,$63,$39,$BF,$F4,$67,$DB,$59,$FF,$5A,$AF,$93,$FC
	.dc.b	$6B,$9E,$1F,$F0,$F3,$9A,$E3,$3F,$92,$FB,$67,$D0,$F0,$8C,$3F,$96
	.dc.b	$F6,$FF,$C3,$15,$FE,$9B,$7F,$C2,$8D,$F6,$B5,$EC,$FE,$73,$ED,$5F
	.dc.b	$6D,$7A,$F0,$DB,$FF,$52,$7D,$B6,$9B,$3F,$A4,$8C,$6F,$FE,$D0,$7F
	.dc.b	$08,$FF,$B4,$BE,$15,$F1,$DF,$AA,$7F,$09,$19,$8C,$7C,$AF,$EB,$BE
	.dc.b	$2B,$FC,$25,$8F,$FA,$CD,$FC,$9F,$1F,$F9,$05,$CD,$FF,$97,$BC,$0F
	.dc.b	$DA,$FE,$23,$FB,$9B,$FF,$5D,$3F,$E0,$76,$F9,$7A,$F5,$FF,$77,$F6
	.dc.b	$FF,$EE,$09,$FF,$60,$2F,$7F,$3F,$01,$A7,$F8,$0F,$C8,$3E,$CA,$FC
	.dc.b	$0E,$1B,$87,$FB,$27,$ED,$DF,$57,$FC,$21,$FF,$3E,$46,$63,$1F,$23
	.dc.b	$FC,$33,$E5,$5F,$5F,$7E,$83,$0D,$BF,$F6,$A7,$CC,$FF,$BE,$9B,$F5
	.dc.b	$9F,$C5,$23,$1C,$BF,$EB,$08,$FE,$2F,$F6,$FF,$C1,$A1,$D1,$FE,$69
	.dc.b	$87,$CD,$EB,$D5,$DB,$C3,$E1,$3F,$8F,$FC,$E7,$EB,$6F,$84,$63,$D6
	.dc.b	$FE,$17,$F6,$7F,$ED,$BD,$7F,$C2,$8D,$CF,$F8,$47,$D5,$BE,$EF,$7F
	.dc.b	$17,$2F,$F7,$7F,$3D,$6F,$F8,$1B,$73,$FE,$05,$ED,$23,$15,$FD,$CB
	.dc.b	$DB,$FE,$3B,$1F,$CD,$3E,$DD,$F6,$B4,$C7,$DA,$5B,$FF,$7E,$4E,$9B
	.dc.b	$3A,$69,$FF,$00,$7E,$AB,$FB,$FB,$7D,$3F,$E8,$31,$8D,$3E,$5C,$FF
	.dc.b	$B6,$A3,$F6,$6F,$18,$DC,$F0,$6C,$7F,$82,$B4,$FF,$63,$EF,$FC,$27
	.dc.b	$CA,$FE,$9E,$D9,$D1,$FE,$10,$7F,$87,$75,$6B,$D9,$FD,$4D,$F2,$6D
	.dc.b	$6D,$76,$6F,$0E,$0D,$9E,$B8,$B1,$FE,$35,$39,$8F,$69,$9B,$E7,$5C
	.dc.b	$3E,$1D,$F6,$CF,$BB,$FE,$CA,$D9,$D3,$A6,$C6,$88,$EA,$2D,$F8,$29
	.dc.b	$F5,$DF,$D9,$AB,$C0,$FF,$B5,$DB,$F0,$BF,$DD,$23,$15,$C5,$E0,$DF
	.dc.b	$86,$D4,$ED,$FC,$CF,$D7,$4B,$FE,$C5,$AF,$FA,$8D,$FC,$DC,$66,$FE
	.dc.b	$18,$EB,$7D,$5E,$78,$7F,$DE,$11,$8B,$E2,$7E,$92,$74,$FF,$92,$23
	.dc.b	$55,$EF,$FB,$3F,$E0,$F3,$F3,$5F,$E0,$F1,$9F,$23,$FC,$23,$18,$C3
	.dc.b	$FD,$C4,$F0,$9E,$5E,$46,$AB,$6F,$E3,$3E,$D3,$57,$F9,$06,$E1,$F8
	.dc.b	$D7,$A3,$6B,$81,$1F,$C3,$E8,$9D,$E1,$F8,$97,$9B,$FE,$66,$C9,$F4
	.dc.b	$3F,$C1,$8F,$F0,$A7,$A7,$FC,$D6,$7E,$5C,$EB,$0F,$E1,$35,$FE,$19
	.dc.b	$FA,$3F,$A5,$5A,$8C,$7C,$DF,$3A,$FF,$0C,$56,$B4,$FA,$35,$A9,$FC
	.dc.b	$B7,$3F,$F4,$1C,$3F,$18,$DA,$4F,$F9,$F3,$97,$FB,$D7,$87,$FE,$3E
	.dc.b	$AD,$4E,$5B,$6A,$FF,$00,$7A,$BB,$BE,$0F,$1F,$80,$98,$78,$BE,$02
	.dc.b	$E1,$E2,$9B,$97,$8A,$3A,$65,$FD,$D9,$D5,$FE,$BF,$81,$F4,$5F,$E5
	.dc.b	$25,$F4,$F6,$30,$D9,$E0,$DA,$FD,$65,$70,$9E,$9F,$2F,$CC,$E9,$F5
	.dc.b	$76,$32,$78,$7E,$0B,$17,$C9,$F1,$7A,$3E,$3F,$46,$D7,$9D,$B5,$E0
	.dc.b	$6E,$BE,$FE,$2F,$C5,$47,$F9,$53,$ED,$5E,$5F,$F8,$72,$FF,$D3,$DF
	.dc.b	$19,$FE,$AE,$9C,$F9,$3F,$E9,$78,$FF,$58,$FD,$B7,$F8,$3B,$E6,$7F
	.dc.b	$C7,$17,$FE,$B6,$BD,$BC,$8C,$61,$D6,$77,$09,$DC,$62,$BA,$5F,$86
	.dc.b	$F8,$37,$0F,$16,$1D,$1A,$AB,$8B,$93,$C4,$77,$AE,$6E,$6F,$C1,$B7
	.dc.b	$E0,$DF,$2E,$3E,$4C,$7C,$15,$CA,$30,$FF,$B5,$30,$EE,$F5,$C7,$C7
	.dc.b	$3A,$8C,$1E,$29,$DC,$64,$98,$F9,$27,$D5,$8F,$8F,$0F,$0E,$3E,$4A
	.dc.b	$D5,$F5,$C3,$9C,$60,$98,$FC,$48,$CB,$E4,$8C,$E3,$E3,$C7,$B3,$F0
	.dc.b	$F1,$EE,$47,$DD,$1A,$7D,$78,$F8,$EF,$A1,$DE,$75,$8F,$B2,$33,$5D
	.dc.b	$F5,$B7,$C6,$3E,$66,$3F,$07,$2F,$3C,$62,$F7,$97,$E4,$32,$EB,$7E
	.dc.b	$AC,$39,$1F,$6B,$1D,$E1,$EB,$21,$F3,$1C,$BE,$AB,$F7,$39,$3C,$04
	.dc.b	$07,$E0,$8A,$7E,$34,$A7,$DE,$2E,$5F,$E5,$2E,$DC,$C7,$E9,$7B,$FF
	.dc.b	$DB,$CE,$27,$4D,$8C,$BC,$9D,$E1,$F9,$46,$DF,$ED,$5B,$3F,$B1,$F7
	.dc.b	$3F,$DE,$73,$9C,$3F,$27,$5A,$AF,$D8,$37,$05,$F0,$9F,$35,$70,$23
	.dc.b	$FA,$18,$D3,$CA,$33,$87,$95,$FD,$0C,$EA,$BD,$B3,$9A,$FD,$55,$F8
	.dc.b	$6B,$4F,$8C,$97,$CA,$7A,$E1,$EB,$78,$C6,$5A,$4B,$EC,$7E,$EA,$FA
	.dc.b	$CF,$29,$F6,$9E,$B3,$07,$AC,$E5,$5F,$C8,$67,$8C,$60,$E3,$5D,$5E
	.dc.b	$38,$7B,$AB,$CD,$39,$9E,$07,$57,$CB,$1D,$63,$8E,$1F,$01,$F2,$1C
	.dc.b	$2B,$A3,$E0,$9E,$0E,$67,$85,$77,$3A,$C6,$C8,$7F,$2F,$BF,$5C,$6A
	.dc.b	$3C,$95,$D2,$3F,$9D,$EF,$FE,$0F,$F5,$7F,$3F,$C3,$FD,$27,$97,$F9
	.dc.b	$76,$1F,$D9,$78,$73,$B7,$F7,$B9,$5E,$2B,$FB,$C4,$61,$E0,$43,$FB
	.dc.b	$2C,$6F,$0F,$F0,$F4,$59,$FF,$72,$BF,$DC,$FE,$1F,$0C,$3F,$BB,$14
	.dc.b	$FE,$DF,$C5,$CF,$C6,$FF,$39,$C7,$F9,$A5,$7D,$96,$4D,$79,$B2,$75
	.dc.b	$19,$78,$47,$97,$B3,$FC,$A3,$1F,$E1,$EF,$23,$06,$7E,$7A,$C6,$48
	.dc.b	$7F,$88,$9F,$77,$92,$F1,$A9,$27,$FC,$A5,$F4,$75,$FF,$2B,$45,$E3
	.dc.b	$FF,$28,$CF,$CD,$97,$DA,$3E,$AB,$E8,$7B,$0F,$8C,$F3,$D7,$FA,$32
	.dc.b	$31,$A7,$1B,$E1,$3F,$E9,$43,$9D,$F2,$BE,$38,$FF,$89,$F0,$E3,$5B
	.dc.b	$BE,$2F,$F9,$D9,$B3,$FE,$2D,$8E,$66,$EF,$CB,$F7,$BF,$ED,$58,$FF
	.dc.b	$2E,$E8,$F2,$7F,$11,$E7,$23,$13,$A9,$CC,$3C,$BB,$12,$74,$74,$8F
	.dc.b	$C6,$7D,$B7,$E2,$A3,$FD,$3C,$BC,$4C,$9A,$E8,$93,$FE,$AF,$9D,$36
	.dc.b	$72,$FF,$A5,$3E,$DB,$F4,$6D,$E9,$FA,$91,$FE,$2B,$8D,$77,$60,$CC
	.dc.b	$C8,$FA,$13,$CB,$7F,$90,$DE,$EF,$E9,$ED,$9F,$EA,$6D,$E2,$70,$3E
	.dc.b	$23,$F5,$5E,$AF,$F8,$BB,$1F,$F2,$02,$FB,$8C,$9A,$F0,$E6,$5F,$EB
	.dc.b	$E8,$CB,$FE,$C0,$8F,$DA,$7B,$6F,$6A,$92,$7F,$DF,$1E,$DD,$7F,$D9
	.dc.b	$53,$B7,$FD,$F9,$3C,$A7,$EC,$FD,$0F,$DA,$F9,$B9,$5F,$82,$8C,$BF
	.dc.b	$84,$8C,$BF,$EE,$4E,$9F,$C2,$4E,$A3,$1D,$28,$C6,$BF,$EF,$E8,$7C
	.dc.b	$DF,$F0,$DF,$B5,$78,$D7,$E9,$26,$3F,$EE,$D9,$CB,$67,$2F,$FB,$96
	.dc.b	$13,$CE,$7F,$8C,$66,$F5,$29,$2F,$FD,$F1,$0F,$53,$11,$FE,$EF,$8E
	.dc.b	$0E,$A7,$9C,$FE,$29,$FD,$97,$BD,$27,$F0,$F1,$97,$FC,$2B,$A7,$ED
	.dc.b	$7C,$35,$6E,$7F,$2D,$5B,$69,$C8,$E3,$D5,$FE,$1E,$E9,$79,$DF,$E1
	.dc.b	$2C,$7F,$D1,$1D,$DF,$EC,$CB,$D9,$FE,$B3,$F8,$7F,$DD,$29,$FF,$47
	.dc.b	$46,$1E,$3E,$DF,$F3,$DA,$6D,$78,$A6,$1D,$3C,$FF,$DA,$4F,$F4,$3F
	.dc.b	$8A,$FD,$04,$17,$B1,$93,$5E,$1F,$F4,$9F,$8B,$D9,$AC,$3A,$35,$F3
	.dc.b	$70,$7F,$B4,$63,$AF,$4B,$CF,$FE,$A4,$BE,$62,$98,$8F,$EA,$DF,$69
	.dc.b	$FD,$2B,$0F,$42,$F8,$DC,$9A,$F5,$3F,$AE,$7D,$A7,$F5,$2A,$F0,$9F
	.dc.b	$ED,$B7,$93,$8D,$D1,$8E,$4F,$F5,$C3,$C5,$FF,$5D,$7D,$57,$FA,$E2
	.dc.b	$BE,$42,$F6,$BF,$B5,$36,$33,$EE,$FF,$6D,$FA,$9F,$34,$7C,$F5,$F1
	.dc.b	$99,$35,$A7,$F7,$3F,$81,$7C,$25,$F0,$26,$1F,$5E,$72,$FD,$9F,$B2
	.dc.b	$FC,$33,$EB,$5F,$65,$FD,$F5,$B1,$BE,$EF,$F0,$07,$C3,$FE,$12,$3F
	.dc.b	$04,$FE,$E1,$BD,$DA,$EE,$FF,$03,$7C,$F7,$F1,$DE,$CF,$F3,$AF,$32
	.dc.b	$3F,$8F,$5F,$E2,$DA,$FC,$D7,$82,$87,$D9,$7F,$58,$AE,$7F,$10,$FB
	.dc.b	$3F,$F0,$A3,$6F,$F2,$D5,$B7,$93,$97,$ED,$39,$BE,$BE,$EF,$F8,$76
	.dc.b	$35,$D3,$FE,$F3,$FA,$4F,$D7,$65,$EA,$FA,$47,$C5,$FF,$88,$DB,$97
	.dc.b	$E9,$6B,$6F,$A7,$EC,$BF,$C1,$4E,$7E,$33,$FC,$55,$3A,$D1,$5A,$AF
	.dc.b	$C7,$37,$26,$5F,$2A,$30,$DF,$86,$FF,$1A,$43,$E1,$BF,$AE,$F4,$7F
	.dc.b	$83,$BF,$0B,$FE,$39,$8F,$F0,$4F,$3D,$18,$8F,$EC,$D3,$C1,$D4,$67
	.dc.b	$2F,$F0,$63,$72,$FE,$BF,$5B,$AE,$27,$F6,$67,$FC,$3B,$E7,$FE,$2A
	.dc.b	$3F,$B9,$7C,$1F,$8C,$8F,$F6,$C7,$FB,$99,$87,$FC,$45,$7F,$E0,$3C
	.dc.b	$3F,$B7,$CF,$F8,$AF,$E2,$3F,$8C,$63,$F8,$CF,$49,$F2,$3F,$E5,$56
	.dc.b	$E3,$FE,$27,$AD,$BF,$EC,$66,$A6,$9A,$E0,$7D,$5F,$D5,$FA,$D7,$D9
	.dc.b	$7F,$98,$7E,$BB,$FA,$8E,$1F,$C0,$DB,$AF,$D6,$8F,$CC,$57,$07,$F9
	.dc.b	$43,$5F,$E1,$4D,$8E,$18,$23,$1A,$7E,$A1,$F3,$FD,$87,$9E,$BF,$DB
	.dc.b	$FF,$61,$FD,$53,$0F,$EB,$9F,$25,$FE,$E0,$8C,$6D,$FE,$7B,$94,$6E
	.dc.b	$F9,$46,$27,$FD,$CF,$18,$9E,$35,$CE,$7F,$DD,$67,$38,$E5,$B7,$F4
	.dc.b	$BB,$E4,$7F,$BC,$AF,$95,$70,$BE,$53,$FE,$F6,$8E,$47,$82,$B9,$5F
	.dc.b	$4C,$3D,$B3,$C6,$A9,$C6,$F9,$AC,$3B,$BE,$1C,$39,$D7,$18,$F2,$9F
	.dc.b	$18,$E6,$71,$8D,$BC,$6C,$C3,$CC,$D6,$94,$E4,$D7,$C9,$E5,$B9,$71
	.dc.b	$82,$31,$3C,$63,$13,$A9,$CC,$F3,$E4,$E8,$F9,$1A,$FF,$80,$8C,$B8
	.dc.b	$3D,$B7,$FE,$22,$D7,$DF,$3C,$4F,$87,$19,$7E,$29,$9D,$3F,$19,$19
	.dc.b	$7F,$1D,$19,$CF,$FD,$9F,$EF,$7E,$7E,$73,$3E,$3F,$8E,$E3,$3F,$A3
	.dc.b	$73,$8F,$CC,$73,$8D,$C3,$E5,$4F,$A2,$BA,$46,$2F,$CF,$18,$7E,$94
	.dc.b	$EE,$BD,$B1,$F0,$F0,$EA,$FE,$53,$83,$F5,$9F,$3D,$FB,$5C,$FF,$5F
	.dc.b	$C2,$8C,$61,$FD,$5B,$89,$DB,$8C,$7F,$77,$57,$1E,$77,$F8,$16,$CF
	.dc.b	$E1,$5A,$FE,$25,$AF,$F6,$2E,$5F,$F2,$0C,$FF,$8A,$DB,$1F,$E0,$D7
	.dc.b	$F8,$DC,$67,$CB,$F0,$46,$A7,$A3,$5E,$91,$FE,$BB,$6C,$FE,$4F,$EB
	.dc.b	$DF,$4F,$99,$FC,$EC,$62,$75,$39,$94,$E5,$FC,$F3,$75,$FF,$67,$7C
	.dc.b	$FF,$FA,$D6,$FF,$DA,$93,$97,$FC,$AD,$19,$AF,$F1,$59,$D3,$3F,$98
	.dc.b	$D9,$FE,$A9,$7F,$E9,$97,$9C,$EA,$FE,$B4,$7A,$E3,$CD,$18,$AF,$2C
	.dc.b	$60,$FB,$1C,$AC,$1F,$5F,$CE,$73,$18,$C3,$FC,$B9,$F5,$FF,$E6,$06
	.dc.b	$F9,$7F,$7E,$3F,$D2,$B1,$82,$63,$F7,$E3,$83,$67,$E7,$6E,$9C,$CF
	.dc.b	$E0,$FC,$47,$C6,$69,$B8,$7E,$2D,$B3,$F8,$B6,$BF,$8B,$FA,$FF,$F3
	.dc.b	$14,$FE,$45,$B1,$F9,$16,$BB,$F7,$3C,$71,$FE,$67,$9E,$1E,$EE,$08
	.dc.b	$FE,$6F,$CF,$7C,$E7,$A9,$B1,$98,$C1,$EE,$FA,$E7,$8D,$49,$3F,$A5
	.dc.b	$8C,$BF,$A7,$F8,$7B,$9F,$E9,$DA,$FE,$9D,$AF,$F9,$21,$AF,$F9,$0F
	.dc.b	$EB,$9F,$5D,$F3,$9B,$5F,$B3,$6C,$FE,$CD,$AF,$EC,$FE,$B9,$F3,$FF
	.dc.b	$59,$B5,$FB,$B6,$CF,$EE,$DA,$FE,$EF,$EB,$BF,$82,$CF,$F0,$16,$C7
	.dc.b	$F0,$16,$BB,$EF,$FC,$1F,$D7,$3E,$BA,$92,$FF,$85,$46,$5F,$E1,$9F
	.dc.b	$5C,$FA,$EF,$C5,$B6,$33,$18,$3F,$1F,$F5,$DF,$E8,$D8,$FE,$28,$DC
	.dc.b	$7F,$8A,$35,$FE,$28,$D7,$FD,$5D,$F2,$FF,$98,$F8,$D3,$49,$5F,$C7
	.dc.b	$FE,$41,$F1,$9F,$9E,$6C,$66,$30,$7E,$8B,$EB,$BF,$D8,$71,$FC,$9D
	.dc.b	$B8,$7F,$27,$6B,$FC,$9D,$AF,$FA,$21,$7A,$9C,$C9,$3C,$DF,$E3,$1F
	.dc.b	$50,$FA,$3F,$E1,$AD,$9F,$F3,$07,$CD,$3E,$BB,$35,$27,$FB,$C5,$6F
	.dc.b	$1F,$EF,$8D,$7F,$C0,$0D,$7F,$C0,$8E,$FC,$BF,$E4,$7F,$1B,$CE,$7F
	.dc.b	$D9,$3E,$5F,$F1,$18,$FE,$91,$F5,$CF,$AE,$A6,$3E,$42,$E5,$FE,$ED
	.dc.b	$9D,$36,$74,$D3,$FD,$E5,$F3,$FF,$E1,$66,$FD,$2F,$F5,$58,$C6,$5F
	.dc.b	$ED,$75,$19,$8D,$36,$78,$9F,$E2,$0C,$FF,$CB,$5B,$FF,$BB,$FE,$83
	.dc.b	$FB,$13,$67,$47,$F8,$99,$FF,$7A,$72,$EB,$C9,$FE,$75,$8F,$FB,$AA
	.dc.b	$78,$3F,$E2,$96,$BF,$E2,$D6,$BC,$E3,$8B,$97,$9E,$1E,$45,$F5,$3D
	.dc.b	$F1,$F0,$3F,$DD,$7E,$B5,$F5,$D4,$C3,$73,$F4,$E3,$FE,$CD,$8D,$C6
	.dc.b	$3D,$04,$62,$7F,$E0,$03,$89,$D3,$9C,$BF,$CF,$3F,$4B,$FD,$E9,$BF
	.dc.b	$0B,$FE,$07,$8C,$57,$FB,$89,$FE,$FB,$6A,$E3,$EC,$EE,$F8,$36,$FC
	.dc.b	$F4,$E7,$7D,$17,$E7,$8F,$EB,$FC,$0F,$8B,$C5,$F0,$AD,$BF,$9E,$AF
	.dc.b	$D3,$91,$C1,$BF,$AB,$E3,$87,$F7,$8A,$47,$F8,$87,$E9,$7F,$86,$69
	.dc.b	$FE,$98,$5F,$5D,$9F,$9D,$D3,$71,$F9,$93,$A6,$CE,$9F,$A7,$A3,$EB
	.dc.b	$7A,$C7,$E3,$5B,$9A,$75,$87,$E5,$27,$31,$C6,$F0,$66,$6D,$FC,$3F
	.dc.b	$99,$FE,$0A,$7F,$81,$BC,$CF,$E7,$39,$FF,$AE,$72,$F9,$D3,$AA,$FE
	.dc.b	$0A,$FF,$0E,$23,$83,$FD,$D9,$F3,$3F,$45,$EC,$63,$83,$FC,$91,$D2
	.dc.b	$F8,$AA,$63,$C3,$B5,$F9,$19,$D3,$67,$4D,$35,$EB,$F4,$3A,$FC,$E5
	.dc.b	$8C,$93,$3F,$C7,$3F,$E9,$CE,$A6,$0A,$E1,$5A,$13,$F4,$FE,$63,$FC
	.dc.b	$37,$C6,$69,$30,$FF,$31,$46,$5F,$D5,$2F,$92,$FF,$34,$39,$BE,$D7
	.dc.b	$F7,$E7,$3F,$1B,$2F,$8F,$D3,$5B,$DE,$0A,$F1,$92,$BC,$2F,$9B,$E9
	.dc.b	$24,$79,$1B,$5F,$E7,$5A,$FE,$A3,$C7,$FE,$F0,$5F,$31,$E6,$AF,$3B
	.dc.b	$FE,$56,$FA,$E7,$95,$FC,$AB,$07,$C8,$7F,$8E,$9F,$F6,$3C,$EB,$C8
	.dc.b	$FF,$5B,$7C,$2F,$FB,$27,$E9,$7F,$CC,$57,$BB,$5E,$DE,$75,$FE,$CF
	.dc.b	$8C,$4E,$DB,$1C,$FE,$99,$B7,$6B,$F0,$E5,$E5,$C3,$83,$6F,$A3,$67
	.dc.b	$FD,$BB,$8F,$92,$73,$97,$39,$F2,$D6,$DB,$19,$AE,$11,$E0,$8E,$19
	.dc.b	$79,$A3,$07,$B0,$EF,$97,$A1,$C1,$D0,$B1,$C5,$C6,$3F,$1E,$FB,$C6
	.dc.b	$31,$F3,$BC,$31,$F3,$4E,$AF,$91,$23,$38,$78,$70,$F8,$B8,$FA,$27
	.dc.b	$DD,$5E,$1B,$E5,$18,$24,$FC,$1B,$E5,$8F,$8B,$0E,$78,$F8,$EB,$59
	.dc.b	$75,$64,$64,$F8,$38,$F6,$8C,$D7,$23,$E0,$61,$DA,$BE,$5C,$67,$0F
	.dc.b	$01,$F8,$3C,$38,$B8,$9D,$B9,$7C,$13,$C5,$C9,$F5,$27,$8B,$83,$E7
	.dc.b	$19,$38,$CF,$57,$23,$0F,$53,$ED,$07,$D6,$29,$F8,$42,$9F,$81,$2E
	.dc.b	$4E,$EF,$F7,$8C,$68,$FD,$3B,$87,$6D,$7F,$50,$D7,$F6,$6F,$2C,$3F
	.dc.b	$54,$DB,$DF,$6F,$E6,$3D,$EF,$2D,$6A,$3F,$37,$5B,$BE,$71,$B7,$2C
	.dc.b	$7F,$33,$1B,$7F,$7B,$5D,$2F,$BC,$6A,$B8,$5F,$EF,$6B,$53,$D8,$D1
	.dc.b	$23,$4F,$0A,$E7,$1E,$78,$FD,$4C,$F2,$7E,$3D,$6D,$E9,$3C,$9E,$95
	.dc.b	$C8,$8F,$F1,$AA,$E5,$87,$D3,$BE,$53,$D0,$EE,$62,$34,$F4,$39,$19
	.dc.b	$C3,$E2,$46,$B0,$F1,$1E,$92,$61,$E2,$AF,$92,$71,$BE,$D8,$79,$30
	.dc.b	$E6,$48,$FA,$93,$E1,$3C,$D7,$D1,$F1,$99,$C3,$C5,$87,$77,$FD,$87
	.dc.b	$1D,$CD,$61,$DE,$F1,$7E,$8B,$FE,$D5,$C1,$FE,$94,$BF,$EB,$F9,$7E
	.dc.b	$7B,$87,$FA,$63,$74,$DF,$A7,$6B,$FB,$6D,$F1,$7F,$AA,$47,$E7,$DC
	.dc.b	$B7,$9B,$FC,$48,$D9,$D4,$F2,$30,$43,$FB,$99,$DE,$7F,$C5,$8D,$8F
	.dc.b	$D8,$E1,$FD,$87,$83,$5E,$2F,$F9,$B2,$BF,$CC,$4F,$FA,$F9,$BE,$AE
	.dc.b	$BC,$C9,$1F,$B5,$C7,$FC,$A7,$1E,$7A,$E5,$3F,$BE,$9C,$BB,$EC,$FF
	.dc.b	$25,$BF,$E1,$FF,$1B,$CD,$18,$8F,$F0,$2A,$F9,$6C,$CC,$9E,$07,$F8
	.dc.b	$9E,$3D,$DE,$3F,$F9,$86,$3F,$C9,$FE,$1E,$0C,$FC,$CD,$8C,$90,$FF
	.dc.b	$09,$CA,$3E,$B4,$3E,$B9,$F3,$DF,$67,$D7,$7C,$2F,$9F,$8B,$FA,$E4
	.dc.b	$63,$4E,$37,$C3,$0F,$F3,$CB,$5F,$E3,$8D,$7F,$91,$B5,$FE,$4E,$D7
	.dc.b	$F9,$5B,$C6,$BF,$CA,$CD,$E5,$DE,$3E,$5F,$AC,$FF,$6F,$5F,$F9,$B3
	.dc.b	$4F,$F2,$83,$7C,$8F,$C9,$79,$A9,$CC,$FF,$AD,$DF,$F7,$53,$B2,$19
	.dc.b	$9E,$31,$A8,$FF,$1D,$FD,$73,$E2,$7F,$20,$BC,$2A,$62,$24,$FF,$AB
	.dc.b	$E7,$4D,$9C,$BF,$E5,$31,$F5,$DF,$D3,$DF,$F5,$1B,$FE,$33,$74,$DD
	.dc.b	$BF,$CF,$4E,$E3,$3D,$BF,$A8,$8C,$47,$F5,$26,$BF,$D5,$9A,$FF,$5A
	.dc.b	$79,$4F,$49,$FC,$2C,$63,$6F,$EB,$FC,$2D,$9F,$5D,$FB,$1F,$57,$FC
	.dc.b	$81,$8F,$FB,$1B,$EB,$B0,$67,$C1,$9A,$7F,$B4,$69,$3A,$8C,$4F,$EE
	.dc.b	$FD,$B7,$B5,$49,$3F,$ED,$98,$7B,$78,$8F,$F3,$7C,$70,$75,$3C,$E7
	.dc.b	$ED,$FD,$67,$DC,$F9,$39,$5F,$89,$8C,$BF,$DC,$F9,$FF,$11,$1F,$C5
	.dc.b	$DB,$19,$E6,$4E,$5B,$1C,$C8,$7E,$3F,$E9,$BF,$8B,$7D,$77,$FB,$72
	.dc.b	$3F,$DE,$4B,$E7,$F2,$6B,$C4,$93,$A8,$CB,$C2,$3F,$90,$65,$FE,$87
	.dc.b	$8F,$F7,$E3,$5F,$F4,$8B,$78,$7F,$CA,$9E,$22,$1E,$9F,$2F,$4E,$5F
	.dc.b	$27,$AC,$7E,$09,$FE,$09,$E1,$3D,$DA,$66,$BF,$2F,$3F,$91,$23,$FC
	.dc.b	$FF,$AF,$C8,$FE,$4E,$38,$3F,$9C,$7F,$88,$79,$BF,$E2,$1C,$FC,$FD
	.dc.b	$CF,$6F,$FA,$33,$6F,$F9,$A3,$5F,$E6,$4D,$7F,$C6,$7A,$FE,$D7,$5E
	.dc.b	$24,$8E,$2F,$F5,$1F,$86,$FD,$1C,$9A,$4F,$03,$FA,$B7,$C3,$E8,$8E
	.dc.b	$0F,$EB,$5C,$DF,$ED,$0C,$3F,$A4,$35,$FE,$8C,$D7,$CA,$D7,$FD,$B1
	.dc.b	$F5,$BF,$D4,$DF,$E9,$8D,$7D,$6D,$77,$E7,$7F,$69,$FA,$DF,$EA,$71
	.dc.b	$F0,$1A,$FF,$B8,$DA,$FF,$B8,$1B,$93,$EB,$7F,$D8,$33,$FD,$7D,$AF
	.dc.b	$F5,$D6,$BF,$31,$AF,$FB,$BF,$E8,$DF,$5B,$FD,$ED,$B1,$FD,$A5,$AF
	.dc.b	$F6,$76,$BF,$5B,$DD,$FF,$7B,$3F,$DB,$DA,$FF,$6D,$6B,$BF,$7F,$FC
	.dc.b	$11,$F5,$9F,$7E,$3F,$E0,$03,$5F,$F7,$EB,$71,$37,$E0,$FF,$45,$3F
	.dc.b	$DF,$5A,$FF,$7B,$6B,$F8,$A6,$BF,$E8,$09,$7C,$C7,$FB,$7E,$3F,$1E
	.dc.b	$D7,$4D,$51,$C3,$C6,$FF,$10,$7D,$5F,$F6,$18,$34,$9E,$8E,$BE,$B2
	.dc.b	$5F,$F8,$4E,$51,$88,$FF,$09,$7C,$EF,$FA,$15,$EE,$DB,$34,$C7,$D5
	.dc.b	$FF,$82,$E7,$FC,$34,$D7,$F5,$2D,$7E,$03,$5F,$5B,$C7,$B3,$F8,$6B
	.dc.b	$7D,$1D,$FB,$1F,$E3,$58,$FF,$08,$57,$F5,$CE,$1F,$F0,$77,$D5,$FF
	.dc.b	$12,$8F,$F5,$5E,$1F,$E0,$E8,$C4,$7F,$60,$9F,$EA,$67,$F5,$D2,$3F
	.dc.b	$C3,$5A,$EC,$EC,$74,$C3,$FA,$CF,$80,$FA,$DF,$EB,$BE,$B7,$88,$95
	.dc.b	$FD,$C1,$CB,$FC,$65,$FE,$D8,$47,$FC,$B5,$ED,$FF,$84,$9F,$F5,$D3
	.dc.b	$5F,$F6,$0B,$5F,$F6,$2B,$5F,$F6,$4B,$C7,$26,$1A,$F9,$3F,$F1,$AC
	.dc.b	$E6,$3F,$21,$F4,$FF,$C9,$5F,$C9,$7D,$4B,$EB,$3F,$CD,$4D,$C7,$FC
	.dc.b	$B0,$D7,$FC,$AC,$D7,$F9,$2B,$9C,$7F,$A2,$B6,$BA,$D7,$F9,$B6,$3C
	.dc.b	$55,$FE,$E0,$FA,$CF,$F2,$C5,$FF,$B8,$BE,$B3,$FA,$87,$95,$FE,$7F
	.dc.b	$6F,$C3,$7F,$B9,$1F,$EA,$95,$C2,$D1,$8D,$18,$73,$C3,$FD,$D6,$73
	.dc.b	$2E,$1C,$DF,$F7,$7E,$1C,$EF,$95,$7F,$BD,$3D,$FF,$0E,$36,$F8,$B4
	.dc.b	$F5,$63,$E0,$53,$C9,$F2,$CF,$06,$9F,$F0,$01,$F2,$3C,$AB,$D0,$D3
	.dc.b	$62,$36,$DC,$7D,$4E,$F0,$F0,$23,$DB,$87,$95,$C1,$D4,$A3,$1F,$98
	.dc.b	$E1,$78,$0F,$7B,$FD,$61,$18,$6C,$66,$79,$9C,$DC,$3F,$EB,$7E,$3F
	.dc.b	$F5,$9A,$69,$1F,$F5,$BB,$F0,$9F,$B7,$E9,$FE,$71,$F8,$67,$B6,$35
	.dc.b	$A7,$F8,$87,$83,$84,$E4,$F9,$31,$97,$89,$E1,$33,$5D,$73,$FF,$06
	.dc.b	$C6,$5F,$F1,$04,$67,$5F,$6E,$1E,$08,$C5,$7C,$A8,$C6,$1E,$28,$E5
	.dc.b	$5F,$ED,$FA,$DC,$7D,$2A,$DB,$E4,$6D,$FF,$8C,$FD,$7F,$F2,$2F,$0F
	.dc.b	$C1,$CB,$FC,$63,$F5,$9F,$8E,$CF,$F8,$A6,$2B,$DC,$63,$B1,$5B,$AF
	.dc.b	$D2,$CF,$E1,$6B,$71,$83,$FA,$6C,$EE,$FE,$84,$E9,$F0,$F1,$78,$EF
	.dc.b	$FA,$D7,$E2,$5F,$8F,$7C,$0D,$C3,$A6,$28,$7C,$3F,$19,$FE,$B7,$1B
	.dc.b	$C3,$FC,$9A,$DB,$FE,$CD,$F6,$2F,$8F,$E1,$8F,$3C,$2C,$E6,$BE,$07
	.dc.b	$D9,$BC,$9D,$B6,$33,$EB,$E2,$64,$70,$77,$18,$BF,$4F,$DE,$FF,$97
	.dc.b	$EB,$E6,$36,$33,$EE,$F9,$23,$1A,$FF,$99,$9B,$5F,$DF,$DB,$5F,$DE
	.dc.b	$95,$71,$F5,$A5,$76,$7A,$46,$2B,$E5,$FD,$5F,$F9,$6E,$BE,$F3,$FD
	.dc.b	$4D,$BE,$CF,$C6,$8C,$4F,$0F,$0F,$06,$66,$D7,$28,$F9,$BF,$62,$FA
	.dc.b	$BC,$DF,$DA,$8C,$57,$E3,$7E,$B1,$F5,$FF,$51,$7D,$65,$8C,$3F,$63
	.dc.b	$EC,$DE,$2D,$33,$7F,$97,$8C,$BF,$99,$F4,$FF,$D0,$CD,$F9,$FF,$C0
	.dc.b	$C6,$33,$FC,$6A,$8F,$C2,$FD,$73,$E4,$EE,$D8,$6E,$1F,$A5,$F8,$4D
	.dc.b	$36,$3F,$1B,$1E,$EC,$7F,$32,$E9,$FC,$DF,$D7,$7F,$B2,$2B,$F5,$CB
	.dc.b	$E5,$7F,$2A,$E6,$77,$1F,$98,$FA,$E7,$D8,$7E,$6B,$F1,$37,$CD,$66
	.dc.b	$30,$FE,$7B,$EC,$1F,$47,$4C,$7B,$F7,$4F,$DF,$7D,$7B,$ED,$FF,$4A
	.dc.b	$BE,$BE,$C6,$1F,$D4,$FD,$83,$ED,$2E,$98,$6E,$1F,$C2,$7E,$C7,$FC
	.dc.b	$08,$DF,$47,$F6,$30,$F7,$2C,$7E,$D3,$E3,$DF,$19,$7E,$2B,$0D,$C3
	.dc.b	$F8,$A7,$C8,$3E,$C3,$F7,$7F,$8A,$7F,$86,$27,$31,$87,$F7,$FF,$65
	.dc.b	$FD,$85,$FE,$38,$DE,$AF,$E1,$AF,$FB,$4A,$33,$EE,$FF,$06,$FC,$57
	.dc.b	$E6,$C7,$F2,$37,$F8,$E3,$FE,$71,$73,$5F,$C5,$9C,$BF,$E0,$AF,$C2
	.dc.b	$7E,$BC,$FF,$2E,$F4,$51,$89,$FE,$51,$0F,$9D,$E6,$FF,$37,$9D,$37
	.dc.b	$93,$F9,$03,$C9,$B3,$C2,$FF,$CB,$35,$C1,$B3,$C2,$31,$7F,$E1,$EA
	.dc.b	$E0,$F9,$E7,$F9,$2F,$DA,$3E,$AE,$FD,$86,$1B,$8F,$F4,$59,$D3,$67
	.dc.b	$4D,$3F,$A5,$7D,$0F,$ED,$30,$FE,$76,$BE,$6B,$F9,$9B,$9D,$3F,$BB
	.dc.b	$7B,$88,$D5,$F2,$F3,$B7,$7C,$1B,$1F,$D6,$B0,$7E,$26,$FA,$CF,$E8
	.dc.b	$D1,$87,$FA,$47,$C1,$BF,$16,$7D,$86,$AF,$86,$1C,$DB,$87,$F5,$DB
	.dc.b	$FF,$7A,$7D,$DE,$F1,$E7,$D4,$8C,$61,$FD,$B2,$B6,$F3,$AD,$CF,$8B
	.dc.b	$EE,$DF,$51,$4C,$7D,$2D,$BE,$53,$A6,$CE,$9A,$7F,$77,$FA,$CF,$EC
	.dc.b	$19,$FF,$AF,$9F,$EC,$F1,$98,$FE,$C3,$18,$7F,$B6,$BC,$38,$8E,$78
	.dc.b	$FE,$9B,$93,$15,$FE,$A6,$FC,$57,$FB,$25,$FE,$69,$B5,$FE,$0F,$7F
	.dc.b	$97,$37,$E1,$7F,$33,$1A,$AE,$51,$8F,$85,$CB,$23,$F5,$13,$BB,$FF
	.dc.b	$3A,$BD,$23,$89,$FE,$17,$FA,$9F,$E1,$F3,$FE,$08,$7F,$87,$46,$7C
	.dc.b	$6F,$F0,$0C,$62,$75,$18,$F1,$F2,$DA,$FB,$5D,$2F,$BE,$FC,$4B,$63
	.dc.b	$3F,$75,$89,$91,$C1,$DC,$63,$0F,$C5,$57,$87,$5F,$F1,$2B,$6F,$FC
	.dc.b	$58,$FE,$2D,$F0,$E3,$E0,$23,$E4,$E3,$B7,$D2,$6D,$F4,$9B,$5F,$E7
	.dc.b	$BE,$0F,$EB,$9E,$87,$F8,$8B,$D9,$37,$8F,$0C,$AF,$F2,$14,$E9,$B3
	.dc.b	$9C,$BF,$C6,$B3,$F4,$D5,$FF,$69,$E2,$ED,$EE,$77,$66,$1D,$1C,$B4
	.dc.b	$EF,$F6,$FE,$53,$CD,$C9,$63,$24,$C3,$C3,$19,$7C,$7F,$41,$ED,$FA
	.dc.b	$2F,$73,$6B,$D6,$D9,$DC,$7B,$A7,$E2,$7C,$3C,$8F,$83,$F0,$F6,$BD
	.dc.b	$F3,$78,$7F,$8B,$ED,$FE,$26,$6C,$7F,$0E,$FC,$2D,$F7,$1F,$E9,$1F
	.dc.b	$96,$F9,$F8,$7F,$9B,$E5,$F3,$F9,$E4,$FF,$36,$47,$0D,$3F,$D2,$D1
	.dc.b	$89,$D5,$71,$C2,$F8,$2C,$39,$E1,$C2,$B4,$D9,$D5,$F0,$AE,$D5,$FE
	.dc.b	$AB,$C3,$B5,$EE,$73,$8F,$A3,$0E,$33,$9C,$39,$1C,$F1,$B1,$E1,$C5
	.dc.b	$F6,$EF,$B8,$FF,$3C,$C7,$73,$FD,$9D,$19,$F9,$6E,$1E,$53,$EB,$7D
	.dc.b	$F5,$CA,$F8,$E9,$F2,$A3,$17,$DF,$0E,$51,$D4,$FE,$9D,$7C,$A3,$35
	.dc.b	$C5,$BB,$7F,$B7,$1B,$5F,$EE,$1C,$7C,$59,$75,$CB,$C3,$7B,$6D,$7A
	.dc.b	$9A,$F8,$27,$86,$52,$3C,$98,$76,$74,$62,$B8,$C6,$4D,$61,$E2,$9D
	.dc.b	$57,$3C,$7D,$57,$E0,$AD,$46,$27,$58,$F9,$DF,$9B,$97,$C3,$BE,$E4
	.dc.b	$8C,$E5,$C3,$1F,$64,$57,$CF,$87,$57,$C6,$47,$EA,$63,$E2,$C7,$C1
	.dc.b	$39,$6C,$EB,$1F,$15,$F5,$CB,$DF,$18,$BE,$07,$88,$FB,$98,$75,$71
	.dc.b	$7E,$34,$78,$0A,$71,$29,$F8,$82,$9F,$96,$F3,$BF,$CD,$6E,$2F,$FC
	.dc.b	$45,$F0,$5A,$8D,$36,$33,$1B,$9E,$35,$F8,$D6,$E5,$F8,$D6,$CF,$E8
	.dc.b	$3C,$8F,$F7,$D4,$E7,$0F,$CD,$56,$AB,$B9,$F8,$8A,$DB,$FA,$88,$D6
	.dc.b	$1E,$2C,$3C,$17,$FA,$3B,$DD,$F2,$9C,$BF,$92,$BE,$33,$AB,$E5,$87
	.dc.b	$8E,$B5,$87,$37,$CA,$C7,$27,$94,$C6,$1D,$89,$3F,$AD,$BE,$D8,$7A
	.dc.b	$09,$7D,$6F,$9D,$F0,$7D,$D1,$98,$C4,$7A,$EF,$CD,$1F,$BB,$BE,$B5
	.dc.b	$E2,$9C,$9E,$1A,$C1,$D0,$EE,$E5,$A6,$6B,$BE,$1F,$E4,$B8,$E2,$78
	.dc.b	$4D,$4F,$19,$FE,$4D,$EE,$FE,$2A,$7F,$8F,$70,$A3,$13,$AC,$3E,$DC
	.dc.b	$BE,$B3,$FD,$7B,$7F,$CD,$DA,$F2,$BE,$38,$7E,$7A,$BA,$5F,$2B,$DE
	.dc.b	$1F,$CD,$DB,$5B,$BE,$38,$30,$E2,$F7,$F4,$B9,$1E,$49,$DC,$EA,$7D
	.dc.b	$77,$D1,$FE,$FF,$CE,$F9,$3F,$D7,$F4,$7F,$64,$6F,$1F,$F8,$46,$75
	.dc.b	$5F,$E0,$CF,$93,$CB,$CA,$AF,$83,$FE,$F1,$6D,$70,$9F,$F0,$7D,$63
	.dc.b	$E7,$73,$F5,$B2,$7F,$C2,$F1,$97,$FC,$33,$19,$7F,$C2,$F1,$D3,$ED
	.dc.b	$A4,$7F,$87,$9E,$FF,$45,$88,$FF,$0E,$C5,$D3,$FB,$FC,$F9,$BD,$1F
	.dc.b	$F5,$C5,$7F,$10,$9F,$F4,$CC,$62,$11,$8F,$67,$87,$D9,$71,$64,$F1
	.dc.b	$DF,$FC,$6D,$7C,$6F,$9D,$FC,$4F,$41,$F7,$74,$98,$7F,$98,$7E,$EE
	.dc.b	$BF,$E6,$18,$B7,$FE,$AC,$A5,$7D,$0A,$7C,$87,$F7,$57,$FD,$54,$D7
	.dc.b	$D3,$97,$F2,$A8,$C5,$A3,$1D,$BC,$E7,$31,$BF,$07,$06,$66,$DF,$68
	.dc.b	$AF,$63,$F9,$CD,$7F,$B4,$5F,$F5,$57,$BA,$F2,$BF,$0F,$D2,$8C,$4E
	.dc.b	$A7,$30,$F3,$2C,$49,$D1,$D2,$3F,$1F,$EB,$7F,$8D,$A3,$FD,$55,$E3
	.dc.b	$E0,$CF,$8F,$23,$24,$3F,$D5,$2F,$CB,$87,$C1,$52,$4F,$CC,$DA,$BE
	.dc.b	$F3,$FC,$83,$18,$DF,$E6,$47,$C2,$9C,$1F,$D6,$AB,$AC,$7F,$B7,$DA
	.dc.b	$F3,$D7,$8B,$F2,$FB,$5F,$6A,$FA,$1F,$D6,$AF,$8E,$A6,$22,$57,$E4
	.dc.b	$CF,$F3,$DC,$ED,$C5,$74,$AD,$C7,$FB,$97,$E1,$7F,$D4,$CF,$FB,$85
	.dc.b	$AE,$7D,$9F,$B3,$0F,$66,$C4,$9E,$0E,$A3,$F2,$DF,$8A,$FF,$90,$E7
	.dc.b	$E0,$7C,$9F,$CF,$33,$E7,$C8,$C9,$0F,$9D,$3C,$5B,$1B,$69,$F2,$FB
	.dc.b	$FF,$88,$FD,$B7,$F9,$37,$8B,$E9,$EA,$8F,$F7,$BD,$7C,$26,$BF,$3D
	.dc.b	$E1,$C5,$FE,$AF,$6C,$7F,$A9,$1D,$D7,$46,$CE,$FC,$BF,$E4,$3E,$43
	.dc.b	$E0,$AF,$D1,$4C,$7E,$8C,$E5,$B3,$97,$F0,$10,$9E,$73,$F9,$5F,$97
	.dc.b	$95,$F8,$97,$F3,$1F,$39,$88,$FC,$A7,$42,$71,$A7,$E3,$21,$58,$CF
	.dc.b	$F2,$95,$6F,$FC,$63,$5A,$7F,$0F,$F0,$6F,$88,$FE,$7A,$BE,$0E,$98
	.dc.b	$8F,$E8,$3C,$4F,$32,$92,$FF,$23,$0F,$37,$FA,$3F,$C2,$FF,$8F,$36
	.dc.b	$FA,$EB,$E2,$32,$6A,$FF,$A6,$FC,$7E,$F7,$FF,$72,$3B,$8F,$CE,$B5
	.dc.b	$FD,$53,$B7,$FD,$BB,$83,$EA,$FF,$AB,$B6,$3F,$4A,$D7,$F6,$4F,$E4
	.dc.b	$71,$FC,$5F,$A8,$F3,$FF,$AF,$AE,$CA,$62,$3F,$B1,$FD,$A7,$F4,$3C
	.dc.b	$FF,$DF,$4F,$1C,$7F,$66,$D7,$F8,$03,$B7,$FD,$E7,$F6,$9F,$B2,$C7
	.dc.b	$F3,$6D,$AF,$F3,$3F,$62,$31,$5C,$3B,$FF,$B9,$F1,$75,$C3,$F4,$6B
	.dc.b	$E4,$72,$6B,$3F,$EF,$3C,$EF,$84,$BF,$0B,$32,$FE,$13,$39,$7F,$85
	.dc.b	$FD,$A7,$F7,$DC,$BF,$5E,$DD,$BC,$AD,$96,$C8,$7B,$3F,$E0,$8F,$B4
	.dc.b	$FC,$45,$7F,$16,$7F,$0C,$DE,$4D,$6F,$FE,$0E,$F9,$FF,$E4,$93,$FC
	.dc.b	$71,$FE,$3C,$FF,$29,$6E,$7F,$C0,$9D,$C7,$F2,$DF,$39,$F0,$FF,$E1
	.dc.b	$86,$C7,$F2,$66,$BF,$CC,$FE,$D2,$DF,$F0,$C5,$F1,$1F,$E2,$07,$3B
	.dc.b	$FD,$46,$D7,$C0,$76,$FF,$32,$FB,$2F,$EF,$6F,$F3,$4F,$A5,$FE,$08
	.dc.b	$FF,$41,$6E,$1F,$C6,$A7,$73,$FD,$15,$B1,$9F,$6F,$FC,$63,$18,$C3
	.dc.b	$F0,$AF,$E1,$6B,$C1,$D3,$AA,$91,$83,$FA,$2F,$87,$FE,$16,$6F,$BB
	.dc.b	$FE,$3C,$8C,$47,$F5,$57,$B3,$F7,$1C,$E3,$FC,$C1,$B8,$FE,$59,$B3
	.dc.b	$BB,$E9,$72,$3B,$9E,$28,$FF,$25,$FD,$7B,$E6,$E9,$27,$FA,$DC,$7F
	.dc.b	$9B,$B1,$FE,$7E,$EE,$33,$3B,$7F,$B6,$46,$F1,$FD,$4F,$8D,$B9,$FE
	.dc.b	$E1,$F3,$1F,$E1,$17,$FB,$3B,$5F,$EE,$91,$BC,$3F,$A6,$B6,$B9,$E1
	.dc.b	$FE,$62,$8C,$57,$E8,$1F,$03,$FE,$79,$E5,$FD,$BF,$4B,$EC,$3F,$97
	.dc.b	$4F,$F7,$66,$BF,$E2,$DF,$7B,$F9,$47,$A9,$FE,$2A,$BE,$35,$AC,$FF
	.dc.b	$AF,$B7,$3F,$F4,$74,$FF,$04,$C1,$AB,$E0,$FF,$CB,$B0,$FA,$E6,$FF
	.dc.b	$EF,$03,$F9,$53,$88,$FF,$7A,$C6,$0F,$F7,$AD,$73,$CB,$FD,$F3,$5B
	.dc.b	$AE,$F7,$DA,$31,$7C,$67,$9F,$95,$FE,$FE,$9E,$18,$73,$9F,$25,$76
	.dc.b	$BE,$4F,$96,$F8,$36,$7C,$2E,$F0,$B0,$3C,$6E,$1F,$7C,$F2,$7C,$66
	.dc.b	$6F,$C5,$87,$93,$0F,$0C,$E8,$F2,$96,$35,$7C,$67,$E6,$CF,$84,$F1
	.dc.b	$57,$B5,$C1,$E4,$29,$4C,$BF,$DA,$FE,$C3,$31,$F6,$BB,$51,$89,$D4
	.dc.b	$F2,$6B,$C3,$E2,$B8,$BF,$0D,$A9,$A4,$7D,$A7,$FA,$8B,$EC,$3F,$CB
	.dc.b	$B9,$7D,$D8,$E4,$6B,$24,$EB,$4F,$5D,$FC,$F9,$CE,$DF,$46,$F8,$E1
	.dc.b	$F8,$79,$CC,$78,$0C,$DF,$FB,$53,$0E,$91,$F2,$30,$E9,$18,$8F,$9D
	.dc.b	$18,$C3,$B4,$7B,$4F,$34,$ED,$EA,$7F,$B0,$5C,$3B,$C3,$C9,$1E,$0E
	.dc.b	$1F,$F0,$C7,$3F,$F9,$FB,$2F,$F2,$6F,$D8,$7F,$9D,$71,$FF,$38,$56
	.dc.b	$9F,$F3,$7C,$67,$1F,$DD,$70,$FF,$8A,$67,$3E,$F7,$37,$F6,$F8,$DA
	.dc.b	$FE,$CF,$7C,$71,$FF,$2D,$FD,$87,$E6,$EF,$FC,$F6,$DA,$FF,$44,$2F
	.dc.b	$92,$E5,$CE,$8D,$5F,$2A,$CC,$62,$FF,$C1,$6F,$17,$AC,$72,$FA,$FF
	.dc.b	$F5,$D4,$7F,$A2,$9B,$1F,$E9,$77,$FD,$71,$19,$F8,$6E,$04,$78,$7C
	.dc.b	$35,$AF,$F4,$F7,$D7,$BD,$DF,$3B,$6B,$FD,$2E,$FF,$81,$9B,$1A,$23
	.dc.b	$AF,$2A,$D7,$49,$B9,$FF,$A0,$AF,$94,$E6,$FF,$D4,$7F,$5F,$FD,$E9
	.dc.b	$BE,$3F,$C8,$8F,$F1,$14,$60,$DF,$93,$93,$53,$63,$27,$F7,$BF,$6F
	.dc.b	$17,$FD,$ED,$FA,$7F,$5F,$FE,$72,$9F,$F5,$9A,$FA,$FC,$FB,$DA,$7E
	.dc.b	$CF,$D7,$FF,$98,$EB,$FD,$72,$DC,$3F,$D8,$D7,$C6,$73,$E1,$FE,$03
	.dc.b	$CF,$FA,$75,$FE,$23,$E2,$E7,$39,$7F,$0B,$F5,$EF,$93,$BA,$E1,$B8
	.dc.b	$FF,$B2,$3E,$BF,$4D,$8F,$C6,$C7,$FA,$3E,$30,$7F,$B3,$BE,$33,$FC
	.dc.b	$CC,$DF,$67,$F2,$71,$FE,$51,$8C,$1F,$ED,$4F,$AF,$7C,$BF,$E6,$1B
	.dc.b	$1F,$ED,$8F,$AF,$FD,$13,$FE,$60,$8C,$ED,$FE,$E7,$FA,$FF,$F7,$24
	.dc.b	$FF,$B7,$9B,$19,$F8,$AF,$D1,$C6,$3A,$BF,$D8,$BF,$5E,$FA,$FF,$D3
	.dc.b	$AF,$AF,$B1,$87,$F5,$7F,$39,$FE,$13,$CB,$FD,$D6,$F8,$AF,$FD,$9E
	.dc.b	$E7,$2F,$F7,$D3,$99,$FF,$3B,$FD,$7F,$F0,$19,$FF,$7F,$36,$BF,$DA
	.dc.b	$EB,$E9,$FF,$72,$E6,$BF,$D3,$FF,$0B,$FB,$CF,$8B,$F9,$58,$7F,$A7
	.dc.b	$BE,$5F,$13,$FC,$03,$E3,$51,$4F,$E0,$7F,$5C,$F6,$F1,$8F,$F6,$5D
	.dc.b	$3F,$DC,$FF,$2D,$FE,$02,$8C,$1F,$E8,$E7,$19,$27,$7F,$6D,$FC,$3A
	.dc.b	$73,$1F,$EF,$0F,$AF,$7D,$17,$F1,$27,$EC,$DF,$FA,$09,$CE,$1D,$A3
	.dc.b	$1E,$A7,$F2,$1B,$7D,$7F,$C9,$9F,$03,$61,$F3,$3C,$A7,$C9,$7E,$BA
	.dc.b	$D1,$3B,$3A,$56,$EF,$F9,$D4,$E6,$BC,$51,$E0,$AD,$E5,$E3,$AF,$17
	.dc.b	$A6,$F9,$E7,$D0,$7F,$2E,$7F,$DC,$D8,$FA,$E7,$53,$9F,$84,$D3,$67
	.dc.b	$55,$C9,$B3,$FC,$E3,$EB,$3F,$7D,$9A,$7F,$D7,$F0,$FA,$7D,$79,$5B
	.dc.b	$6C,$F8,$67,$4D,$9D,$3F,$3F,$77,$D7,$3E,$A3,$FA,$84,$61,$FE,$A5
	.dc.b	$EF,$BE,$BD,$F3,$DA,$BE,$18,$73,$6E,$1F,$48,$FF,$18,$38,$D3,$D5
	.dc.b	$0C,$3A,$5F,$3C,$66,$1F,$DA,$BB,$1F,$5D,$8C,$3F,$CC,$51,$89,$FE
	.dc.b	$DF,$F3,$EF,$AE,$F3,$B6,$33,$F5,$38,$99,$1C,$1D,$C6,$2F,$D1,$CA
	.dc.b	$FA,$7F,$EC,$91,$FE,$D8,$7F,$B2,$3A,$73,$1B,$F8,$3F,$ED,$F1,$83
	.dc.b	$7B,$FC,$09,$E9,$5A,$AF,$CF,$7D,$7F,$F6,$AA,$FF,$06,$37,$2F,$DA
	.dc.b	$3F,$CE,$A3,$3F,$05,$C0,$93,$F9,$89,$D3,$C2,$33,$7F,$4E,$95,$D6
	.dc.b	$3F,$74,$F0,$6C,$FF,$82,$3E,$AB,$F8,$84,$7F,$BE,$1F,$E2,$97,$F9
	.dc.b	$D7,$38,$9C,$CF,$CA,$72,$63,$0F,$84,$E3,$C9,$FC,$6F,$3F,$F1,$3C
	.dc.b	$3F,$C5,$2D,$BF,$F5,$02,$F9,$FD,$F8,$B2,$B6,$F2,$8C,$57,$0B,$EA
	.dc.b	$76,$E3,$B7,$DA,$2C,$F6,$9B,$97,$E5,$B9,$99,$7A,$70,$FF,$45,$46
	.dc.b	$30,$E9,$1C,$CF,$E0,$1E,$B3,$D1,$FF,$18,$B6,$33,$C9,$89,$91,$C1
	.dc.b	$DC,$63,$3F,$C9,$D7,$D5,$DF,$FB,$B7,$B3,$D3,$EB,$FA,$B3,$1F,$0B
	.dc.b	$96,$9E,$4F,$9F,$F1,$9E,$4E,$3B,$19,$26,$3E,$08,$CB,$E2,$FA,$0F
	.dc.b	$5F,$D7,$FB,$1B,$5F,$01,$B3,$B8,$F8,$73,$F0,$BE,$0E,$47,$C6,$F8
	.dc.b	$4B,$5E,$E9,$BC,$7F,$A7,$D9,$58,$BF,$A3,$F5,$EF,$9F,$FE,$17,$F2
	.dc.b	$5F,$3F,$2F,$E1,$92,$FA,$2C,$F2,$70,$BB,$5A,$9B,$CC,$BE,$98,$FF
	.dc.b	$A9,$70,$EF,$5A,$9C,$DF,$FA,$9E,$73,$5B,$8F,$0E,$1B,$BE,$0D,$9D
	.dc.b	$61,$DE,$31,$5C,$CE,$33,$AC,$3D,$4F,$58,$7D,$7B,$EB,$BF,$CF,$D1
	.dc.b	$C4,$CB,$8C,$FE,$24,$7F,$47,$FA,$9F,$82,$EB,$C0,$E9,$7C,$DB,$7C
	.dc.b	$75,$FF,$6C,$79,$4C,$38,$DE,$F0,$F0,$37,$1E,$ED,$BE,$35,$C3,$0E
	.dc.b	$97,$FE,$F3,$AD,$36,$32,$F9,$F2,$F1,$E5,$23,$BC,$EB,$2F,$25,$67
	.dc.b	$2F,$0C,$F2,$C3,$AD,$F3,$38,$E1,$BC,$BF,$DE,$93,$AC,$79,$D6,$89
	.dc.b	$87,$4B,$FA,$53,$A8,$C4,$EA,$F9,$E1,$E5,$CB,$E0,$61,$DE,$BD,$CF
	.dc.b	$28,$CC,$71,$32,$7D,$BC,$3B,$5F,$4A,$E5,$3F,$64,$D5,$72,$C3,$C6
	.dc.b	$4A,$FA,$39,$70,$CB,$E1,$9E,$FA,$E1,$3D,$6C,$43,$C6,$53,$C0,$59
	.dc.b	$FC,$13,$7C,$DF,$F1,$FB,$B3,$F2,$EE,$1D,$79,$F9,$72,$D9,$FF,$15
	.dc.b	$BF,$94,$9E,$78,$FE,$31,$B7,$F6,$3C,$8F,$F7,$FB,$F6,$A3,$38,$7E
	.dc.b	$35,$AF,$63,$F2,$F3,$C1,$D5,$F7,$23,$A2,$DF,$86,$2D,$F8,$66,$9C
	.dc.b	$8F,$D1,$DF,$83,$0F,$F1,$64,$68,$F1,$5F,$37,$B1,$C2,$F5,$3F,$0A
	.dc.b	$33,$3B,$25,$FB,$EB,$F7,$38,$7A,$EB,$AC,$E5,$E8,$72,$71,$5C,$A3
	.dc.b	$D0,$FE,$E2,$79,$5F,$86,$35,$3D,$0F,$DC,$4E,$B0,$F7,$13,$0F,$1C
	.dc.b	$FA,$A3,$0F,$84,$EF,$87,$88,$8F,$43,$F9,$2F,$D4,$7F,$B0,$9F,$E4
	.dc.b	$FE,$E7,$E2,$71,$FE,$7D,$C4,$8C,$5F,$F7,$4B,$E1,$7F,$CA,$E9,$18
	.dc.b	$BE,$37,$C2,$BF,$CA,$4D,$AD,$FA,$5F,$85,$6C,$F8,$B2,$C5,$6F,$0E
	.dc.b	$98,$F8,$70,$CD,$4B,$F5,$C7,$E3,$FB,$FF,$DD,$DF,$1F,$FA,$DE,$9F
	.dc.b	$EB,$CD,$E1,$FD,$8C,$EA,$FF,$C1,$7F,$21,$95,$19,$7F,$C0,$B5,$98
	.dc.b	$C5,$FF,$69,$AC,$75,$3E,$6E,$FC,$B4,$C7,$FC,$2F,$1A,$30,$FF,$84
	.dc.b	$23,$0F,$F8,$73,$EB,$F9,$91,$FF,$0F,$FC,$ED,$7F,$C2,$F1,$71,$FE
	.dc.b	$F5,$3E,$5F,$3F,$FD,$6F,$5F,$EA,$56,$BF,$E7,$06,$CF,$F9,$9D,$7C
	.dc.b	$57,$1F,$6A,$57,$17,$A4,$7C,$1F,$3D,$F4,$34,$95,$FE,$57,$FA,$1A
	.dc.b	$FF,$95,$62,$E3,$FE,$54,$9F,$9D,$2F,$8C,$FE,$DC,$FF,$A9,$DA,$FF
	.dc.b	$A7,$4D,$78,$3C,$23,$FD,$87,$1B,$8C,$11,$E5,$D9,$5E,$53,$D2,$FF
	.dc.b	$D4,$B1,$D0,$D9,$FE,$E5,$EB,$FF,$68,$79,$3F,$83,$6C,$67,$DC,$4E
	.dc.b	$57,$99,$62,$4E,$DD,$47,$E2,$7D,$5F,$F1,$64,$FF,$B6,$5F,$F1,$53
	.dc.b	$7C,$8D,$79,$12,$30,$43,$FD,$48,$FC,$A8,$7C,$1D,$24,$FC,$BE,$DA
	.dc.b	$FC,$B9,$DB,$FE,$D5,$6B,$F1,$A3,$7A,$7C,$69,$CB,$6B,$E9,$4F,$E8
	.dc.b	$FE,$BD,$F0,$17,$EB,$E6,$3F,$EF,$59,$CB,$67,$2F,$FB,$0A,$13,$CE
	.dc.b	$7E,$87,$C1,$BD,$9A,$4B,$FA,$D1,$FE,$97,$6C,$65,$FF,$72,$70,$A7
	.dc.b	$94,$70,$F6,$7F,$23,$F5,$FF,$E2,$F8,$F8,$5F,$59,$F2,$CC,$F9,$D2
	.dc.b	$3F,$DF,$53,$B7,$F2,$8D,$DF,$E3,$76,$EE,$73,$56,$73,$1F,$E0,$07
	.dc.b	$FD,$AD,$F5,$FF,$E4,$86,$FB,$3F,$C3,$63,$FD,$DD,$3B,$E1,$FE,$2F
	.dc.b	$1B,$6F,$97,$BA,$95,$B3,$F8,$CD,$6D,$FE,$37,$E4,$BE,$03,$F8,$D2
	.dc.b	$FA,$FA,$62,$24,$F3,$F2,$D5,$B8,$C3,$5F,$F7,$DB,$D5,$FC,$67,$CA
	.dc.b	$7E,$54,$8F,$E3,$FE,$6A,$BF,$93,$8B,$AF,$E3,$E6,$EB,$FE,$25,$EA
	.dc.b	$56,$2F,$FC,$53,$0A,$FE,$69,$F0,$4F,$87,$BF,$41,$2F,$D5,$39,$7D
	.dc.b	$9C,$5F,$CD,$DB,$E7,$7F,$3D,$8C,$43,$46,$9F,$D2,$25,$E7,$7F,$43
	.dc.b	$F8,$5F,$F1,$C5,$6A,$FF,$A4,$C6,$2F,$FA,$7B,$63,$3E,$CF,$F4,$BF
	.dc.b	$8E,$C7,$11,$A4,$8F,$EA,$18,$BE,$BE,$92,$FE,$B4,$7E,$4F,$1F,$EB
	.dc.b	$BE,$BF,$F3,$C9,$FB,$4B,$B3,$FA,$D3,$63,$3E,$87,$F5,$DF,$AF,$FE
	.dc.b	$8B,$7F,$D9,$9E,$53,$F0,$E3,$13,$BE,$DF,$EC,$9F,$55,$FA,$EA,$D1
	.dc.b	$FD,$A9,$C7,$07,$C5,$DD,$18,$C3,$FB,$9B,$FB,$BA,$E3,$3F,$DA,$F0
	.dc.b	$FE,$E1,$C5,$E2,$C3,$93,$70,$FE,$F6,$FA,$EF,$8F,$3B,$E0,$FF,$BC
	.dc.b	$36,$3F,$38,$D7,$7F,$07,$FD,$F7,$EB,$FF,$3F,$1F,$A1,$6B,$E4,$6E
	.dc.b	$3E,$B5,$ED,$7F,$81,$7E,$BF,$F1,$33,$EF,$7F,$0F,$5F,$AB,$6B,$EE
	.dc.b	$77,$9F,$F8,$53,$E7,$7F,$93,$CF,$C2,$6D,$7C,$96,$CF,$F8,$6B,$CE
	.dc.b	$FF,$7C,$D7,$ED,$DA,$FC,$A6,$D7,$D1,$6C,$FF,$88,$BE,$BE,$D7,$F8
	.dc.b	$81,$7C,$3F,$F8,$7D,$CF,$09,$A6,$A7,$F8,$37,$D7,$3D,$9F,$F1,$37
	.dc.b	$D7,$FF,$03,$9F,$B6,$DC,$FF,$05,$5B,$9E,$35,$FE,$2D,$8C,$47,$E0
	.dc.b	$9B,$3C,$39,$74,$49,$F0,$BC,$7D,$BF,$F1,$B7,$85,$FE,$16,$9F,$E3
	.dc.b	$4D,$7F,$1A,$DA,$FC,$A3,$DB,$5F,$F2,$2B,$96,$D7,$43,$AE,$36,$FA
	.dc.b	$C5,$7A,$C7,$F9,$1F,$EB,$1F,$5F,$49,$3F,$CB,$63,$FC,$D7,$8F,$E9
	.dc.b	$5D,$C6,$67,$6F,$E9,$23,$78,$7F,$96,$BC,$5D,$CF,$EA,$1B,$7F,$AF
	.dc.b	$6C,$EF,$E0,$FF,$CB,$51,$8A,$FF,$2F,$B7,$83,$F9,$56,$3A,$AF,$F1
	.dc.b	$17,$81,$FB,$7B,$FF,$32,$FD,$7F,$F2,$B7,$F6,$ED,$CB,$F9,$17,$A7
	.dc.b	$FC,$9B,$D4,$E2,$7F,$9A,$67,$5A,$7F,$9C,$9B,$6A,$E0,$FF,$5B,$1A
	.dc.b	$BE,$0B,$F9,$A4,$3E,$B5,$F5,$DF,$EE,$B1,$F2,$FF,$EE,$D8,$CC,$FF
	.dc.b	$BC,$27,$2F,$FB,$CA,$32,$F6,$AE,$B1,$8B,$F0,$69,$BB,$ED,$87,$56
	.dc.b	$E1,$93,$FD,$FB,$5C,$1F,$34,$7F,$BF,$E3,$17,$E0,$8F,$1C,$E8,$F3
	.dc.b	$57,$57,$9E,$1E,$49,$E4,$D8,$D9,$E1,$3E,$1C,$F1,$7D,$98,$67,$0F
	.dc.b	$0D,$FA,$27,$C0,$70,$35,$7D,$5E,$20,$87,$92,$AB,$E4,$6B,$92,$D7
	.dc.b	$FA,$97,$EB,$F3,$1F,$73,$C3,$46,$27,$53,$C2,$72,$F0,$1F,$17,$5A
	.dc.b	$65,$C1,$1F,$6D,$BE,$BF,$FB,$B6,$9F,$80,$9F,$65,$7C,$9A,$D6,$FF
	.dc.b	$2E,$72,$72,$BE,$3A,$7B,$AF,$84,$62,$7E,$1C,$62,$B9,$47,$9B,$0F
	.dc.b	$C7,$E1,$DA,$3E,$89,$DA,$33,$7C,$E3,$E7,$DF,$37,$67,$3C,$3F,$DB
	.dc.b	$D3,$B8,$E5,$18,$C3,$F2,$9B,$3E,$9D,$F7,$BF,$E1,$7C,$FF,$BA,$7D
	.dc.b	$87,$F9,$CA,$FF,$D4,$55,$A7,$FC,$85,$19,$C7,$FC,$DB,$39,$7F,$BB
	.dc.b	$C6,$6B,$F8,$0C,$E5,$EB,$F0,$1D,$EF,$FD,$E3,$F6,$1F,$B6,$7F,$C9
	.dc.b	$CD,$C3,$FC,$02,$BE,$4F,$84,$7F,$BD,$23,$57,$FE,$0B,$6D,$7F,$BF
	.dc.b	$3E,$CB,$F8,$1B,$D3,$97,$8F,$5A,$73,$87,$F8,$5B,$07,$CC,$3C,$6E
	.dc.b	$F1,$86,$C3,$CC,$B1,$32,$FF,$38,$56,$DB,$5B,$6C,$F8,$BE,$C7,$FD
	.dc.b	$93,$1F,$E6,$D6,$E1,$FE,$20,$7F,$D9,$71,$9F,$8A,$E0,$47,$87,$C5
	.dc.b	$5A,$E1,$37,$1F,$F1,$75,$F2,$9C,$DF,$FB,$76,$7E,$B3,$FB,$BE,$46
	.dc.b	$AF,$90,$F9,$F1,$FE,$61,$E2,$B3,$9A,$FF,$73,$37,$0F,$F5,$7B,$70
	.dc.b	$FE,$82,$76,$8F,$E7,$91,$99,$F0,$36,$3F,$A0,$47,$F8,$DC,$EB,$87
	.dc.b	$89,$C4,$7F,$91,$3E,$C2,$DF,$F9,$2D,$FF,$0B,$CF,$FA,$5D,$CB,$FC
	.dc.b	$5E,$D1,$AA,$E5,$F3,$3F,$84,$83,$7C,$DC,$3F,$88,$F3,$3F,$D2,$38
	.dc.b	$7F,$A8,$FC,$1F,$E4,$4F,$FA,$96,$33,$87,$F9,$62,$72,$FF,$96,$BA
	.dc.b	$3F,$D2,$D5,$FE,$AB,$5F,$61,$F9,$47,$33,$B8,$FA,$3D,$3C,$B5,$FE
	.dc.b	$D6,$DB,$FB,$0B,$DB,$DD,$FE,$7B,$E4,$1F,$65,$7E,$3F,$0D,$C3,$FD
	.dc.b	$79,$F5,$DF,$57,$88,$BE,$2F,$F4,$EE,$4C,$4F,$FB,$0F,$EB,$DF,$31
	.dc.b	$FA,$C6,$C6,$7E,$23,$13,$FB,$0F,$AD,$7C,$E5,$F4,$B0,$DB,$FF,$67
	.dc.b	$7E,$2E,$F9,$CF,$DB,$AF,$95,$B1,$87,$F2,$BF,$69,$F0,$F8,$FF,$C1
	.dc.b	$0D,$F0,$0F,$C5,$C3,$FD,$C9,$18,$9F,$E0,$7E,$CF,$FA,$F2,$7F,$DB
	.dc.b	$CB,$D5,$CC,$69,$FE,$13,$F5,$6F,$96,$BF,$39,$86,$D7,$FB,$9F,$EB
	.dc.b	$5E,$F7,$F1,$06,$C6,$7E,$93,$13,$FC,$53,$EA,$FE,$91,$F3,$B4,$FF
	.dc.b	$50,$B8,$E9,$FE,$27,$F5,$7F,$2B,$0F,$E1,$ED,$AF,$E3,$4F,$F9,$3A
	.dc.b	$33,$C9,$C0,$8F,$0E,$5B,$5C,$26,$F8,$3F,$D0,$EF,$94,$E6,$FF,$C0
	.dc.b	$13,$FE,$B9,$F9,$FF,$5C,$FF,$AD,$96,$9F,$CB,$9C,$CE,$E3,$1A,$65
	.dc.b	$B7,$D0,$FD,$B5,$72,$7E,$BF,$43,$E9,$BF,$89,$BF,$C9,$B0,$FF,$26
	.dc.b	$B9,$9D,$F9,$DA,$AD,$B6,$32,$7F,$92,$FE,$AB,$FD,$2C,$DF,$84,$FE
	.dc.b	$7F,$F5,$5C,$F5,$FF,$6C,$8A,$DB,$86,$E5,$FD,$A7,$F1,$7F,$D0,$CE
	.dc.b	$6F,$FA,$4C,$62,$58,$7F,$8A,$2F,$8B,$F3,$27,$4E,$6B,$8F,$7F,$F3
	.dc.b	$2F,$84,$7D,$D5,$31,$F6,$36,$FF,$3B,$3A,$6C,$E9,$A7,$CF,$CF,$F9
	.dc.b	$FF,$D2,$FF,$68,$6F,$99,$FD,$9A,$31,$E4,$1F,$EE,$55,$C6,$31,$EE
	.dc.b	$6E,$B8,$36,$7F,$2B,$F1,$1F,$98,$3F,$B5,$FC,$A7,$EF,$63,$24,$C7
	.dc.b	$F3,$33,$AF,$1F,$84,$E7,$8B,$F2,$93,$98,$C1,$FC,$F7,$EB,$3C,$E7
	.dc.b	$9A,$D5,$F4,$1F,$EB,$B1,$AC,$BF,$C0,$6D,$AF,$63,$67,$8C,$7F,$65
	.dc.b	$F9,$D7,$D1,$FF,$64,$6C,$7F,$0C,$7F,$AD,$79,$07,$35,$C8,$FF,$4E
	.dc.b	$38,$C7,$FD,$4A,$DA,$ED,$ED,$75,$C3,$F2,$7D,$32,$7F,$C1,$BD,$3F
	.dc.b	$DE,$F0,$FF,$1C,$3F,$A3,$8F,$DA,$B7,$1F,$82,$D9,$3C,$67,$75,$D6
	.dc.b	$BF,$C4,$53,$B3,$4F,$28,$FF,$01,$E7,$FC,$96,$BF,$95,$B6,$7D,$D9
	.dc.b	$A3,$F9,$34,$EB,$C8,$E2,$47,$8F,$8D,$6F,$E3,$71,$FF,$4C,$F1,$3F
	.dc.b	$C6,$FD,$3F,$C8,$F5,$FB,$53,$FD,$43,$6C,$D4,$C7,$F2,$97,$B7,$FC
	.dc.b	$00,$FC,$DC,$3F,$C3,$AE,$5B,$E5,$F6,$3B,$FB,$76,$B9,$C5,$9F,$2D
	.dc.b	$7F,$67,$3F,$AA,$4F,$E8,$FE,$A1,$CB,$E8,$E6,$FF,$27,$C6,$27,$CB
	.dc.b	$18,$9D,$47,$16,$DF,$C2,$AF,$E7,$9A,$FC,$06,$D7,$F8,$6D,$FE,$6B
	.dc.b	$19,$F1,$BB,$92,$7C,$9E,$3D,$BF,$AB,$E8,$3E,$B7,$F4,$9F,$13,$FD
	.dc.b	$FE,$7F,$DE,$13,$9B,$FF,$51,$F0,$7F,$91,$63,$F9,$A4,$FF,$AC,$23
	.dc.b	$17,$C9,$B5,$C2,$79,$CE,$5E,$BE,$45,$E7,$61,$DE,$7A,$1C,$2F,$94
	.dc.b	$EA,$F8,$57,$8B,$1F,$44,$E6,$B7,$3C,$E7,$D9,$C9,$FE,$16,$F8,$9E
	.dc.b	$38,$78,$E3,$17,$E7,$C5,$F5,$8F,$3F,$FC,$59,$1E,$43,$2E,$31,$F9
	.dc.b	$11,$FE,$95,$FA,$1F,$73,$AF,$6B,$A5,$F3,$6D,$F1,$DF,$FD,$D3,$E0
	.dc.b	$B0,$E3,$7B,$C3,$C0,$DC,$7B,$B6,$F8,$D7,$0C,$3A,$5F,$FB,$D6,$B4
	.dc.b	$D8,$CB,$E7,$CB,$C7,$94,$8E,$F3,$AC,$BC,$95,$9C,$BC,$33,$CB,$0E
	.dc.b	$B7,$CC,$E3,$86,$F2,$47,$86,$3E,$76,$3D,$23,$64,$C3,$AC,$EB,$2F
	.dc.b	$63,$CE,$79,$1F,$4E,$B9,$57,$08,$E5,$7E,$07,$E9,$1B,$BE,$6E,$8E
	.dc.b	$73,$F6,$AB,$9D,$71,$C3,$DA,$4C,$7E,$26,$1C,$AF,$C0,$4A,$FA,$79
	.dc.b	$79,$F1,$F0,$CE,$5F,$87,$8F,$67,$C3,$5D,$F0,$FC,$7C,$7A,$4F,$49
	.dc.b	$71,$FC,$23,$63,$6F,$9F,$EA,$BF,$CD,$0F,$13,$F3,$6E,$1E,$4D,$78
	.dc.b	$73,$ED,$DB,$F9,$88,$E7,$8F,$E4,$1B,$87,$C5,$E4,$F9,$D3,$9A,$E0
	.dc.b	$4B,$FC,$CC,$EA,$BB,$C7,$E7,$2F,$9B,$F9,$76,$BF,$B2,$9E,$0E,$6B
	.dc.b	$A5,$F7,$8E,$A6,$2F,$AC,$64,$FD,$3D,$72,$9F,$69,$E1,$AE,$0F,$E9
	.dc.b	$E3,$58,$B0,$F1,$B6,$B7,$5E,$52,$C6,$5E,$87,$90,$B7,$E3,$C3,$CA
	.dc.b	$7C,$98,$E3,$7D,$8E,$F8,$78,$9C,$61,$DA,$30,$F0,$3F,$86,$B8,$9E
	.dc.b	$F5,$D9,$F6,$1B,$32,$60,$FE,$5F,$C4,$F7,$9E,$4F,$D0,$E1,$FF,$5E
	.dc.b	$F8,$77,$AC,$FD,$4B,$E1,$56,$F8,$E1,$FC,$E1,$AF,$66,$EB,$FC,$E1
	.dc.b	$7D,$8F,$1A,$95,$C5,$E9,$18,$8E,$51,$6B,$94,$F9,$FA,$FB,$F5,$FF
	.dc.b	$67,$AF,$E0,$4D,$BF,$EE,$D5,$8E,$4F,$ED,$55,$65,$AB,$0E,$75,$A9
	.dc.b	$E2,$7F,$7F,$F8,$CE,$15,$BF,$2A,$DF,$F9,$DA,$B4,$43,$FB,$E4,$68
	.dc.b	$DD,$70,$FB,$89,$3F,$E0,$B7,$9F,$DE,$E2,$3F,$C0,$91,$74,$FF,$02
	.dc.b	$4F,$83,$D9,$95,$FE,$13,$7C,$7E,$7E,$23,$FC,$19,$16,$BF,$C1,$93
	.dc.b	$F0,$78,$5E,$A3,$CA,$C3,$70,$FF,$74,$B5,$FF,$1F,$AF,$88,$DC,$C9
	.dc.b	$DB,$CA,$3E,$74,$3E,$2D,$F0,$38,$6C,$7F,$A5,$4D,$47,$0F,$82,$FF
	.dc.b	$3D,$47,$FA,$D2,$76,$D9,$E4,$6F,$8E,$C7,$39,$DE,$9C,$9B,$3F,$D1
	.dc.b	$EF,$AC,$FF,$A1,$F3,$7D,$63,$CE,$F8,$CD,$C7,$FD,$55,$3A,$9C,$F9
	.dc.b	$DA,$64,$EA,$B8,$3F,$E6,$65,$3F,$8A,$FA,$BF,$F0,$5E,$1F,$E5,$E7
	.dc.b	$FD,$1F,$5F,$EA,$F6,$BF,$E6,$67,$79,$7F,$6E,$E7,$46,$27,$F3,$6F
	.dc.b	$0A,$E9,$E7,$F7,$EF,$FA,$1F,$05,$2F,$E3,$3F,$4F,$EA,$31,$1F,$3E
	.dc.b	$2D,$FC,$99,$B9,$7F,$6F,$E0,$56,$2F,$FC,$03,$0A,$FD,$3F,$2F,$F9
	.dc.b	$9A,$7F,$D9,$5F,$5B,$FD,$57,$1F,$F3,$EB,$71,$F6,$56,$EB,$FD,$0A
	.dc.b	$DA,$DF,$89,$FB,$2B,$E1,$5F,$BA,$AD,$5F,$F8,$4B,$89,$F0,$9E,$6A
	.dc.b	$FB,$2D,$7F,$D3,$2D,$C7,$F0,$4F,$A6,$37,$A7,$E9,$E3,$57,$D5,$B2
	.dc.b	$BE,$F9,$FE,$83,$EB,$5F,$4F,$F9,$FF,$95,$97,$F9,$F9,$CB,$67,$2F
	.dc.b	$E7,$E7,$94,$48,$D1,$0F,$D2,$7D,$67,$EE,$FE,$9F,$FC,$FD,$1F,$96
	.dc.b	$5F,$3D,$93,$5E,$94,$9D,$46,$5E,$11,$FC,$7B,$EB,$1E,$FD,$F4,$A5
	.dc.b	$7F,$8F,$B8,$BF,$D6,$B5,$A8,$C7,$0F,$E5,$63,$11,$FE,$BE,$55,$CA
	.dc.b	$31,$7F,$C1,$BE,$E7,$F8,$76,$1F,$C0,$1B,$87,$F8,$F5,$B5,$FE,$4E
	.dc.b	$7F,$89,$3F,$C2,$5C,$A6,$93,$C1,$ED,$E1,$49,$E0,$F8,$63,$F9,$EF
	.dc.b	$BF,$FE,$83,$9F,$59,$FD,$19,$FF,$03,$38,$EE,$FF,$80,$1E,$97,$0F
	.dc.b	$53,$FA,$2F,$C1,$3E,$AE,$92,$F8,$47,$FA,$05,$E5,$F6,$5E,$CC,$0B
	.dc.b	$E6,$7F,$A7,$B6,$77,$A7,$F5,$2F,$B2,$E0,$66,$3F,$DD,$6D,$E0,$FF
	.dc.b	$74,$BF,$23,$B1,$F8,$C7,$F5,$EA,$FC,$C3,$63,$3E,$9F,$F4,$58,$FE
	.dc.b	$EF,$3E,$FF,$AE,$7C,$AF,$F6,$25,$F1,$14,$C4,$7F,$66,$EC,$FD,$55
	.dc.b	$7F,$BF,$9E,$2F,$C9,$6B,$9F,$07,$FB,$6F,$D5,$BB,$29,$27,$E8,$43
	.dc.b	$EA,$FF,$B9,$FD,$1F,$6E,$07,$67,$F2,$58,$F2,$DF,$D7,$6A,$9D,$BF
	.dc.b	$E0,$28,$E5,$3F,$CB,$7C,$AE,$D3,$F7,$D7,$CD,$FF,$80,$9B,$19,$C3
	.dc.b	$FB,$67,$E0,$5F,$2C,$9C,$58,$F7,$3F,$C1,$3F,$58,$FB,$3B,$F0,$13
	.dc.b	$1F,$C6,$4E,$5F,$C7,$7C,$F3,$C9,$FF,$08,$FD,$6F,$F2,$0E,$F3,$67
	.dc.b	$55,$F9,$5F,$AC,$7A,$3F,$E1,$86,$C6,$63,$1E,$66,$7D,$DB,$3C,$63
	.dc.b	$0F,$F6,$0F,$53,$8B,$F2,$79,$7F,$97,$7E,$28,$BF,$F0,$63,$67,$86
	.dc.b	$1F,$E0,$FB,$E3,$D1,$FD,$BD,$B8,$7F,$37,$6D,$7F,$2B,$3F,$96,$F8
	.dc.b	$3F,$EC,$CF,$53,$96,$12,$7F,$B9,$9F,$CC,$B7,$7D,$53,$E3,$FF,$C6
	.dc.b	$ED,$9F,$DC,$35,$F5,$1F,$E4,$27,$FA,$4B,$63,$89,$1E,$3F,$21,$5E
	.dc.b	$35,$E0,$9C,$69,$F8,$C8,$F1,$D7,$F9,$3E,$1F,$8C,$FD,$DB,$F0,$C6
	.dc.b	$39,$F5,$39,$8F,$CA,$9C,$39,$F3,$32,$BF,$8A,$CF,$16,$D7,$F9,$F5
	.dc.b	$FE,$2D,$F5,$DF,$BA,$7C,$FE,$BF,$F0,$F3,$F8,$84,$66,$BD,$0D,$E1
	.dc.b	$FF,$07,$36,$BF,$CC,$ED,$CC,$BE,$33,$FC,$99,$F4,$9F,$D4,$2B,$F9
	.dc.b	$37,$D7,$7F,$28,$7E,$3B,$71,$FF,$18,$4E,$E7,$E9,$F1,$A3,$13,$F2
	.dc.b	$EB,$86,$5F,$E2,$16,$DF,$F7,$1A,$DC,$A2,$70,$FF,$A7,$EF,$1F,$5A
	.dc.b	$D7,$55,$7D,$8B,$39,$8E,$53,$FD,$6A,$31,$1F,$8E,$E1,$FE,$14,$A3
	.dc.b	$38,$77,$8C,$57,$63,$91,$6F,$BB,$E2,$9E,$15,$CA,$FA,$BE,$09,$E3
	.dc.b	$7E,$4B,$E2,$FB,$AB,$73,$CD,$B1,$B3,$E0,$C7,$85,$A7,$1C,$2B,$ED
	.dc.b	$6C,$78,$23,$E3,$1A,$C3,$CF,$1D,$E7,$E5,$CD,$29,$CC,$B3,$F1,$5D
	.dc.b	$AF,$15,$F5,$DC,$27,$F7,$71,$87,$8B,$5E,$2D,$7B,$FC,$77,$17,$F0
	.dc.b	$4D,$4D,$23,$F5,$8F,$F0,$E7,$CA,$7F,$44,$D3,$EF,$C7,$23,$47,$F4
	.dc.b	$98,$D6,$5E,$08,$CE,$49,$CC,$7C,$4B,$F9,$B3,$9D,$7E,$7C,$F5,$C3
	.dc.b	$F1,$B3,$98,$EE,$67,$0F,$F6,$EE,$1D,$23,$F3,$38,$74,$8C,$46,$CF
	.dc.b	$5E,$1D,$9F,$82,$79,$A3,$6E,$30,$F0,$C7,$A3,$83,$F6,$5F,$1B,$FC
	.dc.b	$4A,$FF,$D2,$1F,$5C,$F9,$9C,$63,$FB,$1A,$BC,$CA,$D5,$FF,$AA,$38
	.dc.b	$D5,$A9,$CD,$FF,$00,$9C,$D7,$F8,$B6,$73,$B7,$F2,$BF,$AE,$FF,$29
	.dc.b	$CF,$F8,$C5,$B7,$FC,$C9,$FF,$26,$C6,$79,$3C,$71,$AA,$E5,$18,$F8
	.dc.b	$0C,$BE,$A8,$FF,$18,$7D,$6D,$F6,$8D,$F3,$B0,$CC,$7F,$98,$CE,$9B
	.dc.b	$39,$CB,$FA,$2F,$D6,$FF,$BB,$67,$FC,$90,$D8,$CF,$C0,$FC,$88,$C6
	.dc.b	$3F,$D2,$5B,$5F,$EB,$F6,$D3,$81,$1F,$EF,$39,$EB,$ED,$F3,$23,$E0
	.dc.b	$C2,$DF,$F2,$EC,$3E,$6C,$FF,$A9,$DB,$1F,$E9,$B5,$F3,$9F,$7B,$93
	.dc.b	$FC,$51,$C9,$73,$7C,$B5,$35,$7C,$AB,$CD,$19,$C7,$FD,$D2,$3E,$C7
	.dc.b	$19,$7F,$40,$8C,$1F,$E6,$DF,$AE,$B1,$FD,$B9,$7D,$77,$E3,$9C,$E3
	.dc.b	$FD,$BD,$CD,$71,$F5,$F4,$6E,$AC,$7F,$8B,$3C,$AF,$CB,$FA,$1F,$ED
	.dc.b	$A8,$E0,$64,$FF,$3F,$75,$7F,$B6,$DB,$EE,$7E,$7A,$31,$9F,$F8,$09
	.dc.b	$FF,$32,$CF,$FA,$23,$6F,$F4,$FC,$FC,$9E,$9F,$EA,$93,$95,$F3,$B9
	.dc.b	$9F,$D4,$FC,$A3,$EB,$A9,$FA,$CF,$C6,$BF,$CB,$B8,$7F,$A4,$67,$4F
	.dc.b	$FA,$4F,$EB,$FE,$CF,$29,$A3,$E4,$FF,$6A,$EB,$E8,$5F,$3F,$FB,$86
	.dc.b	$DF,$FA,$75,$F8,$99,$FF,$87,$9C,$BF,$E2,$0F,$A9,$FF,$00,$4F,$FA
	.dc.b	$8D,$FF,$18,$37,$CF,$FE,$07,$18,$9F,$F1,$3F,$CA,$3D,$7F,$E0,$EF
	.dc.b	$E3,$AF,$FD,$58,$E5,$BD,$DF,$91,$FA,$FF,$93,$7F,$E0,$66,$F6,$7F
	.dc.b	$9E,$B9,$51,$F3,$2F,$E9,$3A,$7F,$C0,$9C,$1F,$E5,$77,$F8,$BF,$B5
	.dc.b	$F3,$30,$FF,$1D,$AF,$99,$CC,$FF,$1C,$FA,$FC,$3F,$EC,$16,$D7,$F9
	.dc.b	$09,$7C,$96,$EB,$F9,$1F,$D7,$BE,$B6,$98,$FA,$7B,$8F,$FB,$2F,$EB
	.dc.b	$FE,$C8,$F9,$DF,$E5,$6E,$1F,$E2,$DF,$5F,$F4,$34,$FF,$24,$B7,$87
	.dc.b	$FC,$8E,$FC,$F6,$F3,$70,$26,$1F,$E4,$89,$D3,$FE,$FB,$79,$36,$37
	.dc.b	$5D,$BE,$06,$DF,$F9,$77,$E8,$FD,$97,$FE,$E3,$59,$FF,$42,$73,$7F
	.dc.b	$E6,$67,$33,$BD,$35,$5B,$6C,$7E,$A7,$A1,$F5,$FF,$C4,$5F,$E4,$3C
	.dc.b	$C7,$33,$BF,$3B,$55,$B6,$C6,$4F,$F7,$DE,$4F,$AF,$FF,$52,$C6,$1B
	.dc.b	$1F,$E7,$4F,$AD,$EA,$FF,$A1,$8E,$1F,$0D,$FD,$81,$B5,$C6,$3F,$A5
	.dc.b	$4F,$0D,$BF,$D9,$B5,$A3,$95,$F6,$8F,$F3,$EF,$D7,$FE,$B1,$DC,$67
	.dc.b	$0F,$F2,$D4,$27,$51,$FE,$5D,$D1,$F3,$6F,$AA,$FE,$6C,$D9,$FE,$24
	.dc.b	$BE,$47,$33,$27,$2F,$08,$C6,$5F,$E9,$CF,$AF,$FE,$D1,$5F,$A8,$7F
	.dc.b	$B6,$C6,$7D,$BF,$EF,$F1,$8C,$FF,$D4,$0F,$09,$FE,$AC,$E6,$B9,$57
	.dc.b	$0F,$33,$F3,$7F,$17,$27,$FD,$37,$F3,$3F,$C0,$70,$FC,$EC,$E5,$B3
	.dc.b	$AF,$1B,$19,$7E,$82,$AC,$FC,$8B,$E5,$39,$8F,$96,$7F,$02,$FA,$FF
	.dc.b	$37,$88,$7C,$D8,$FF,$6D,$72,$62,$35,$9F,$F8,$7A,$30,$6B,$3F,$EC
	.dc.b	$DF,$43,$FC,$BF,$0F,$E8,$4D,$C3,$FD,$3A,$FF,$68,$8C,$F9,$9C,$08
	.dc.b	$F0,$F6,$6D,$7E,$4F,$A3,$F1,$A7,$E3,$BD,$BB,$5F,$EC,$2E,$8F,$F0
	.dc.b	$2E,$C6,$9F,$E0,$5B,$E5,$3B,$BE,$31,$F9,$FB,$B7,$FC,$AF,$47,$D4
	.dc.b	$FF,$80,$5B,$87,$F5,$D7,$F9,$0C,$EE,$74,$49,$D7,$8D,$70,$FF,$73
	.dc.b	$F9,$DF,$C7,$67,$E9,$C7,$F8,$6E,$30,$EB,$5F,$F2,$5C,$63,$0F,$F7
	.dc.b	$63,$6B,$F9,$DB,$6B,$95,$70,$F1,$78,$F9,$92,$FB,$F1,$FF,$60,$F7
	.dc.b	$31,$87,$92,$AE,$5D,$A6,$DF,$F4,$8F,$AD,$73,$F9,$3A,$3F,$CB,$F1
	.dc.b	$89,$EB,$AF,$E4,$1E,$15,$B7,$C3,$3B,$CB,$E1,$DF,$2F,$1E,$56,$BC
	.dc.b	$7B,$4C,$D3,$E3,$AC,$C6,$27,$FC,$89,$87,$F3,$C6,$CF,$F8,$E7,$D0
	.dc.b	$7D,$2E,$FE,$23,$FA,$75,$3C,$25,$6B,$2F,$F5,$4F,$17,$F0,$F8,$FE
	.dc.b	$73,$5F,$EA,$A8,$C6,$1C,$DB,$5C,$27,$A5,$73,$78,$18,$F2,$78,$3A
	.dc.b	$BB,$87,$89,$B5,$C7,$1F,$3C,$F9,$AF,$35,$30,$F0,$4F,$16,$DF,$A2
	.dc.b	$F8,$BC,$F1,$FF,$80,$3F,$5C,$FA,$5F,$F1,$9E,$7E,$61,$19,$F6,$7D
	.dc.b	$90,$F0,$1F,$5F,$EA,$8D,$DF,$2C,$BE,$6C,$62,$F8,$E1,$CE,$3B,$61
	.dc.b	$CB,$3C,$E1,$C7,$2F,$2B,$68,$73,$38,$E5,$E1,$C3,$A6,$5E,$06,$DF
	.dc.b	$2C,$BD,$39,$77,$C3,$DD,$5A,$7A,$B5,$ED,$7E,$A9,$E0,$F5,$C7,$C9
	.dc.b	$3A,$C3,$B3,$E4,$CB,$D7,$8C,$8E,$F5,$C3,$0E,$77,$C3,$0E,$97,$B7
	.dc.b	$C1,$97,$B6,$FB,$E5,$E5,$3C,$D8,$72,$C7,$D3,$3A,$26,$1F,$02,$F9
	.dc.b	$3D,$71,$F0,$1F,$32,$BE,$E4,$E9,$F3,$CE,$DF,$A3,$7D,$DF,$80,$78
	.dc.b	$23,$4F,$2C,$AE,$3E,$7C,$7E,$06,$1D,$72,$F7,$DF,$84,$E4,$53,$EF
	.dc.b	$19,$AF,$C4,$37,$D2,$FF,$35,$3B,$8F,$F2,$8F,$27,$18,$FC,$CC,$E9
	.dc.b	$E1,$19,$9D,$C7,$E0,$A7,$9D,$7E,$25,$BA,$7E,$9B,$87,$E3,$4E,$70
	.dc.b	$FC,$DD,$6A,$FF,$12,$D7,$F5,$35,$C2,$FF,$5F,$39,$C3,$C3,$87,$3B
	.dc.b	$FC,$9D,$ED,$E7,$87,$62,$3A,$2C,$F1,$26,$1E,$78,$CB,$E0,$9B,$1F
	.dc.b	$9E,$AE,$CF,$C6,$8C,$B4,$EC,$7C,$57,$37,$E8,$AE,$D3,$92,$5F,$89
	.dc.b	$EC,$70,$C3,$E0,$1F,$07,$0F,$0D,$F3,$9F,$79,$CA,$FA,$DF,$38,$F2
	.dc.b	$36,$55,$CA,$FA,$E1,$E6,$8C,$3E,$13,$F8,$8B,$F3,$77,$7C,$87,$F9
	.dc.b	$17,$1F,$E5,$2D,$7F,$9B,$C6,$BC,$7E,$12,$F9,$BB,$12,$B8,$BA,$8C
	.dc.b	$47,$29,$E9,$9F,$F3,$A6,$E9,$FE,$5B,$7A,$4F,$22,$1C,$6F,$A3,$FD
	.dc.b	$37,$B1,$CD,$E7,$E7,$FE,$C5,$18,$AF,$DE,$46,$27,$5C,$5F,$DE,$F3
	.dc.b	$CB,$FD,$F6,$AE,$1F,$DA,$9B,$87,$38,$FE,$F3,$58,$B7,$C5,$DF,$2A
	.dc.b	$57,$F7,$E8,$D1,$87,$FB,$D6,$3F,$E0,$B9,$CC,$7F,$81,$3E,$BF,$A1
	.dc.b	$1F,$F0,$3F,$D7,$D7,$89,$FD,$FA,$31,$A7,$B6,$7C,$9E,$CC,$AF,$F0
	.dc.b	$7B,$E7,$F4,$71,$1C,$B5,$F9,$33,$8B,$F8,$D5,$F0,$B8,$9E,$AB,$C9
	.dc.b	$C3,$71,$FF,$1C,$CE,$A7,$3E,$4E,$99,$3A,$AE,$0F,$AA,$BA,$D7,$D0
	.dc.b	$87,$C4,$BE,$BF,$0D,$8F,$F5,$CB,$5F,$F5,$A7,$BF,$FE,$AE,$CF,$FD
	.dc.b	$5F,$39,$6D,$72,$3B,$4E,$F8,$7F,$A0,$56,$DB,$3B,$BF,$F6,$3D,$6D
	.dc.b	$FF,$13,$7D,$7F,$F9,$91,$F8,$2D,$7D,$8D,$C3,$FC,$69,$92,$1E,$5F
	.dc.b	$CD,$75,$8F,$FA,$ED,$B7,$DB,$CC,$D5,$7E,$4B,$EB,$FF,$C1,$78,$FE
	.dc.b	$79,$79,$B9,$35,$D9,$30,$FF,$2F,$CE,$9B,$39,$7F,$DB,$9B,$3E,$1F
	.dc.b	$E7,$92,$3E,$7F,$D2,$D7,$71,$71,$F9,$F3,$75,$F9,$73,$6B,$FC,$2F
	.dc.b	$6A,$FD,$5F,$27,$F9,$AA,$7F,$CE,$7F,$5F,$FD,$57,$1F,$DC,$B7,$1F
	.dc.b	$F0,$FF,$3A,$F8,$5F,$E6,$5B,$7C,$3C,$2F,$DB,$E1,$C6,$97,$C7,$CD
	.dc.b	$FC,$E7,$C1,$FB,$E7,$F3,$2D,$7F,$28,$DC,$3E,$CB,$E9,$8D,$EB,$C8
	.dc.b	$8F,$2D,$3F,$35,$F5,$EF,$AF,$FC,$CF,$CB,$CA,$FF,$4F,$89,$D3,$86
	.dc.b	$BF,$BE,$35,$3D,$0F,$D1,$FD,$1D,$8E,$11,$5F,$B7,$1F,$B9,$FA,$9F
	.dc.b	$F3,$FC,$7E,$D9,$7D,$16,$4D,$7A,$52,$75,$19,$78,$47,$F2,$2F,$A9
	.dc.b	$7B,$B7,$D3,$95,$FE,$AC,$9C,$B6,$72,$FF,$0F,$8F,$F3,$43,$93,$E1
	.dc.b	$FD,$7F,$F0,$4E,$1F,$F4,$D4,$23,$1C,$FF,$E3,$56,$C7,$F9,$89,$78
	.dc.b	$5F,$C2,$5C,$C7,$1F,$13,$06,$66,$C7,$82,$2B,$F3,$E7,$9B,$FE,$C6
	.dc.b	$F7,$BF,$D0,$8D,$F4,$7F,$A1,$C1,$BE,$A7,$F4,$6F,$81,$7D,$7D,$25
	.dc.b	$FF,$B4,$23,$FD,$0E,$FF,$B4,$BE,$BF,$D7,$D8,$69,$34,$FE,$A3,$F5
	.dc.b	$FA,$26,$B8,$8F,$E6,$DF,$5F,$FD,$73,$0F,$F5,$D3,$7C,$93,$4F,$ED
	.dc.b	$B1,$FD,$07,$EB,$DF,$2B,$FD,$79,$7C,$35,$31,$1F,$D8,$BA,$BF,$45
	.dc.b	$5F,$D2,$5E,$2F,$FB,$B1,$AE,$7B,$FF,$B4,$FD,$7B,$A2,$92,$7F,$DE
	.dc.b	$70,$FA,$FF,$ED,$FF,$49,$DB,$AB,$FB,$DB,$7A,$BF,$81,$B6,$FF,$7A
	.dc.b	$F8,$27,$FC,$03,$E4,$F3,$9F,$F7,$FB,$5F,$EC,$0D,$8D,$FC,$D7,$F5
	.dc.b	$FF,$AF,$F9,$04,$F7,$31,$EE,$7F,$80,$BE,$BD,$F5,$97,$DF,$99,$FF
	.dc.b	$AD,$32,$FF,$08,$D6,$A3,$1A,$FF,$1A,$F9,$D7,$91,$FE,$0C,$6C,$E8
	.dc.b	$FE,$2E,$E1,$FE,$ED,$F5,$DF,$C8,$5D,$77,$A3,$11,$AF,$4B,$FC,$2F
	.dc.b	$18,$CB,$FC,$46,$FF,$00,$9E,$2B,$DC,$DF,$97,$27,$31,$FE,$01,$F4
	.dc.b	$F9,$69,$FE,$D0,$EF,$7D,$77,$1B,$FF,$00,$B5,$FE,$F9,$5C,$14,$A3
	.dc.b	$0D,$CF,$B1,$FE,$0F,$9F,$E6,$17,$DD,$A7,$22,$CF,$FB,$8B,$D1,$E1
	.dc.b	$F1,$7F,$E3,$6F,$AB,$7C,$65,$25,$FC,$88,$FE,$8F,$87,$F4,$C7,$71
	.dc.b	$99,$DB,$FE,$25,$8D,$BF,$D2,$9D,$B9,$8D,$EF,$FE,$50,$9D,$AF,$AE
	.dc.b	$FC,$15,$7F,$8E,$23,$1C,$DF,$E4,$E6,$C7,$E0,$5E,$1C,$D9,$AE,$30
	.dc.b	$C7,$F1,$D1,$86,$DE,$EB,$FC,$AF,$F5,$DF,$BE,$BF,$F3,$1B,$6B,$FB
	.dc.b	$5B,$70,$FE,$CA,$FF,$A2,$9F,$C6,$B9,$E4,$34,$9F,$13,$FE,$45,$FA
	.dc.b	$2F,$C6,$57,$F9,$2F,$EB,$BF,$92,$3F,$96,$6B,$FE,$3F,$8D,$E3,$FE
	.dc.b	$24,$6E,$3F,$DF,$AB,$73,$FE,$7A,$6C,$7E,$46,$77,$41,$AF,$FA,$6A
	.dc.b	$F1,$F5,$AF,$6B,$FD,$F8,$7F,$54,$A4,$65,$E5,$19,$8C,$47,$4D,$55
	.dc.b	$D2,$BC,$13,$98,$C3,$CB,$D6,$CD,$F2,$AE,$87,$0C,$3C,$D5,$E2,$8C
	.dc.b	$61,$DA,$30,$73,$3A,$9E,$BB,$EB,$3C,$30,$CD,$F4,$C3,$C1,$5A,$BE
	.dc.b	$47,$43,$D0,$59,$E1,$7F,$1A,$78,$4F,$73,$AE,$1E,$5B,$F8,$B7,$EA
	.dc.b	$9F,$86,$74,$2B,$97,$CC,$68,$FA,$FE,$1B,$EB,$B1,$18,$6C,$A7,$53
	.dc.b	$C6,$31,$3C,$23,$F0,$5F,$1B,$53,$48,$ED,$F9,$79,$BE,$BB,$F5,$93
	.dc.b	$F1,$4E,$67,$C7,$8D,$6D,$F4,$A3,$39,$7E,$1A,$73,$1C,$4F,$09,$C6
	.dc.b	$BA,$CF,$E3,$23,$8E,$BF,$ED,$49,$CB,$F1,$70,$F2,$C6,$27,$E6,$C6
	.dc.b	$30,$F1,$47,$6B,$F1,$56,$E3,$F2,$B5,$B7,$C8,$D9,$F4,$67,$F0,$31
	.dc.b	$FF,$54,$37,$0F,$F7,$3E,$97,$0F,$EB,$93,$5F,$D1,$B6,$9E,$A5,$8C
	.dc.b	$3B,$E9,$7D,$B5,$3A,$53,$F1,$B7,$26,$73,$FF,$32,$DF,$07,$F9,$24
	.dc.b	$FF,$3C,$73,$7C,$A3,$19,$7F,$99,$EB,$FC,$37,$5F,$CA,$FF,$1C,$FF
	.dc.b	$4E,$3D,$3D,$0F,$0C,$62,$B8,$46,$3B,$7D,$0D,$AF,$1F,$DD,$FA,$CF
	.dc.b	$4F,$D7,$FF,$A6,$E3,$24,$BF,$E7,$DF,$8E,$FB,$23,$1B,$7C,$F8,$C4
	.dc.b	$F9,$EF,$E1,$C6,$27,$FC,$C3,$9A,$73,$B7,$D1,$FC,$7B,$FC,$13,$8F
	.dc.b	$FA,$29,$B8,$FF,$A1,$FB,$5E,$76,$FC,$A9,$7B,$F7,$2D,$FF,$A3,$70
	.dc.b	$FF,$2B,$CE,$4E,$9E,$B7,$7A,$F0,$3F,$83,$D3,$F9,$1F,$85,$E3,$FA
	.dc.b	$DC,$6F,$FE,$24,$C3,$C9,$19,$7F,$C3,$95,$FE,$B2,$E9,$C6,$7F,$F0
	.dc.b	$07,$64,$75,$A7,$17,$7F,$C1,$7F,$9F,$47,$23,$27,$47,$37,$FE,$A7
	.dc.b	$5E,$9F,$E3,$BD,$2F,$F0,$F5,$7F,$AC,$27,$2F,$F6,$F8,$CD,$71,$E6
	.dc.b	$E1,$1F,$9F,$AF,$C5,$CF,$F7,$98,$19,$F9,$1F,$CD,$FE,$3E,$F9,$FC
	.dc.b	$63,$FD,$EE,$30,$FC,$2F,$A5,$FC,$BF,$0F,$F1,$0E,$07,$C8,$FF,$89
	.dc.b	$6B,$5E,$6F,$CF,$8C,$69,$FE,$C3,$F1,$1F,$3F,$FA,$AF,$C8,$3F,$BA
	.dc.b	$E1,$FE,$CC,$9C,$BF,$E0,$B8,$CD,$7F,$5A,$8F,$F1,$D7,$E4,$3F,$E7
	.dc.b	$37,$EA,$BF,$88,$6F,$BD,$F7,$E3,$13,$FE,$D3,$FA,$0F,$DA,$3F,$E1
	.dc.b	$BF,$C8,$9F,$53,$FC,$16,$73,$BF,$F0,$D9,$CE,$3F,$EE,$09,$7D,$17
	.dc.b	$F0,$2B,$FF,$65,$CE,$63,$1E,$8F,$3F,$BE,$D7,$DC,$5A,$FF,$13,$5A
	.dc.b	$3F,$29,$F5,$3F,$A9,$9F,$F2,$B2,$F9,$7C,$F9,$1A,$7F,$8D,$7E,$46
	.dc.b	$F9,$0F,$95,$EE,$E1,$B8,$7E,$63,$D8,$F8,$3C,$3F,$5D,$7D,$C7,$E8
	.dc.b	$DC,$CF,$F9,$95,$CB,$FE,$FE,$FC,$8F,$F7,$4D,$F3,$FF,$90,$FD,$8B
	.dc.b	$E2,$3F,$96,$56,$B9,$FF,$E0,$03,$72,$FE,$6D,$F1,$0F,$B8,$DE,$3F
	.dc.b	$CA,$A3,$19,$75,$3F,$C9,$4E,$30,$E7,$3B,$AF,$DD,$7E,$48,$FA,$2A
	.dc.b	$63,$C1,$B8,$FE,$E2,$74,$D9,$D3,$4F,$F2,$C7,$D7,$FC,$0C,$7F,$CB
	.dc.b	$CB,$F0,$1F,$E7,$EE,$74,$FE,$A0,$E6,$B8,$F0,$EA,$B6,$D8,$FF,$52
	.dc.b	$5B,$C6,$FF,$78,$46,$1B,$1F,$9D,$3F,$D1,$FF,$84,$FE,$B2,$75,$F3
	.dc.b	$5F,$D6,$6B,$79,$7E,$C6,$BF,$05,$5A,$3F,$8B,$B8,$BE,$87,$86,$FF
	.dc.b	$91,$46,$6B,$FC,$13,$39,$AE,$3C,$BF,$7B,$A6,$47,$F9,$F7,$97,$F8
	.dc.b	$0B,$FE,$C7,$7F,$31,$5F,$E4,$E6,$D9,$A7,$08,$D5,$7F,$33,$FB,$BF
	.dc.b	$EE,$37,$FC,$B9,$FD,$E4,$FC,$3F,$25,$1A,$8C,$3A,$C3,$F8,$93,$96
	.dc.b	$F8,$BD,$7E,$03,$B3,$F2,$7A,$E6,$3F,$80,$F0,$2E,$3F,$1B,$81,$5F
	.dc.b	$DF,$BE,$D1,$F1,$7A,$CF,$FC,$21,$18,$C7,$ED,$56,$EB,$51,$FE,$EB
	.dc.b	$9D,$61,$EB,$F6,$BF,$9F,$E1,$FC,$A5,$78,$FF,$D5,$5C,$DF,$FB,$5D
	.dc.b	$CB,$FD,$49,$CD,$7F,$B5,$1C,$BD,$4E,$53,$FB,$6F,$0A,$57,$FA,$D7
	.dc.b	$DA,$FE,$67,$87,$F8,$A7,$8F,$F8,$74,$FD,$DA,$DE,$1F,$EB,$FB,$DC
	.dc.b	$FD,$F9,$FC,$97,$1C,$8F,$EC,$BF,$39,$FD,$EB,$1F,$C5,$EF,$F8,$4C
	.dc.b	$7F,$41,$7F,$DC,$7F,$04,$7D,$0F,$19,$FB,$D1,$8C,$7F,$49,$5B,$AD
	.dc.b	$61,$FD,$3E,$B5,$7F,$B6,$F9,$DF,$DA,$9F,$B6,$F9,$DB,$8F,$EF,$22
	.dc.b	$CF,$EF,$26,$DF,$F5,$9F,$B4,$FE,$9F,$3F,$8F,$6C,$67,$1F,$F0,$44
	.dc.b	$63,$0F,$F5,$C7,$4A,$31,$68,$C7,$A7,$C2,$F8,$BF,$E5,$1F,$C3,$FF
	.dc.b	$97,$D7,$FA,$27,$C7,$7D,$7E,$7E,$17,$FC,$B1,$F1,$E9,$CE,$9F,$C2
	.dc.b	$B0,$E3,$39,$BE,$13,$FE,$CB,$D7,$FD,$07,$39,$8F,$F0,$CC,$FE,$95
	.dc.b	$F4,$B7,$1F,$43,$FA,$B8,$CF,$77,$F8,$9E,$35,$AF,$9B,$E6,$F1,$39
	.dc.b	$7F,$C5,$9F,$87,$FF,$83,$2F,$FC,$4F,$F8,$D3,$A7,$3E,$FF,$F8,$C6
	.dc.b	$31,$3F,$E2,$CA,$E1,$9F,$EF,$E7,$2F,$F0,$18,$CE,$BF,$E0,$F9,$CB
	.dc.b	$FC,$63,$E9,$BF,$B5,$5F,$FA,$AB,$EB,$7F,$DB,$B1,$FE,$26,$C3,$FD
	.dc.b	$57,$18,$F5,$BF,$D4,$FE,$EA,$37,$F0,$98,$75,$87,$3B,$FF,$7B,$61
	.dc.b	$CE,$F8,$46,$30,$E0,$EB,$0A,$AB,$CA,$F8,$8D,$5F,$57,$55,$C2,$7B
	.dc.b	$61,$E4,$AD,$46,$2B,$C1,$18,$BE,$FF,$6B,$FE,$03,$F0,$BA,$36,$BC
	.dc.b	$07,$8F,$DC,$DF,$97,$FC,$42,$33,$3B,$D3,$DB,$08,$C6,$7F,$32,$F8
	.dc.b	$E1,$DB,$0E,$58,$70,$BD,$E1,$F2,$2F,$6F,$6A,$DE,$08,$95,$C1,$F3
	.dc.b	$65,$D7,$19,$1C,$5F,$87,$87,$4C,$7C,$18,$F4,$AE,$6D,$79,$9C,$49
	.dc.b	$1C,$2B,$9D,$F6,$8C,$4E,$B1,$F8,$13,$97,$D1,$1F,$16,$FB,$63,$E9
	.dc.b	$CF,$FC,$8B,$7B,$AF,$1E,$35,$F1,$E3,$D8,$98,$F4,$C3,$96,$5E,$D8
	.dc.b	$C6,$5D,$8F,$66,$5D,$B0,$F3,$1D,$8A,$76,$9E,$38,$F8,$70,$EB,$95
	.dc.b	$C3,$F0,$57,$EC,$3D,$95,$5E,$86,$6B,$F1,$8D,$F3,$5F,$35,$B8,$FC
	.dc.b	$44,$61,$E5,$28,$C4,$71,$9E,$13,$97,$F3,$33,$C7,$4F,$C6,$37,$2F
	.dc.b	$8D,$E1,$7D,$29,$CD,$70,$26,$1F,$9C,$9D,$57,$CD,$BE,$53,$F9,$06
	.dc.b	$BE,$8A,$FD,$4D,$70,$8D,$47,$4B,$EF,$1D,$4C,$5F,$59,$B1,$F9,$CA
	.dc.b	$E4,$F5,$3F,$5F,$7E,$0C,$3E,$04,$73,$33,$5C,$9E,$18,$7E,$B6,$F9
	.dc.b	$D7,$08,$F1,$93,$0F,$8B,$5C,$67,$51,$E5,$8E,$D3,$B7,$DA,$78,$0C
	.dc.b	$C6,$E3,$51,$E2,$8E,$AE,$E3,$33,$DA,$3E,$01,$E1,$C3,$DC,$7B,$EF
	.dc.b	$D2,$E3,$0F,$34,$7B,$3C,$59,$5F,$C6,$9F,$93,$F8,$87,$E2,$31,$FE
	.dc.b	$2A,$DC,$BF,$51,$3C,$2B,$F8,$FC,$E5,$FE,$1C,$D7,$F8,$E4,$6E,$B9
	.dc.b	$E1,$D8,$C5,$F4,$AF,$E6,$D7,$CF,$A5,$F7,$3F,$E4,$69,$FE,$BF,$F8
	.dc.b	$57,$F0,$4E,$5F,$EA,$CD,$E1,$FF,$03,$D6,$EB,$FC,$B8,$DA,$DC,$72
	.dc.b	$3F,$BD,$4F,$0F,$37,$94,$7F,$83,$7F,$16,$78,$D8,$C9,$0F,$8B,$C3
	.dc.b	$6B,$99,$C6,$B2,$47,$35,$CA,$3A,$7E,$2B,$FC,$0F,$2F,$F4,$2E,$BF
	.dc.b	$EA,$3B,$E7,$E1,$3E,$ED,$F2,$3E,$68,$C3,$E8,$95,$71,$7F,$D2,$6D
	.dc.b	$BF,$F0,$BD,$F2,$9E,$AD,$77,$EC,$73,$FB,$6F,$F7,$04,$7F,$AB,$1A
	.dc.b	$FF,$A7,$9B,$5F,$E9,$C7,$F8,$84,$6F,$AF,$91,$1E,$5D,$3E,$7E,$E7
	.dc.b	$C6,$D2,$4F,$7E,$4A,$FF,$AB,$27,$6F,$FA,$F9,$AF,$8E,$3E,$17,$D8
	.dc.b	$BC,$0E,$DF,$63,$2B,$FD,$8B,$39,$6C,$E5,$ED,$3C,$A2,$46,$88,$78
	.dc.b	$FC,$6F,$57,$CC,$48,$F2,$3E,$DF,$AB,$C4,$7A,$62,$EF,$FE,$C2,$CD
	.dc.b	$5F,$0D,$F8,$9E,$BC,$BF,$74,$65,$F8,$11,$97,$E2,$47,$C0,$CB,$E6
	.dc.b	$7D,$04,$9F,$7C,$65,$F9,$11,$97,$FD,$AF,$A2,$75,$18,$6E,$3E,$7F
	.dc.b	$99,$FB,$1E,$37,$D9,$F8,$D9,$7F,$4E,$32,$FD,$58,$CB,$F6,$63,$F2
	.dc.b	$DF,$62,$F5,$E9,$27,$FD,$F3,$1F,$B0,$6C,$65,$FF,$76,$C7,$17,$7D
	.dc.b	$BF,$A1,$FC,$93,$F2,$7C,$9F,$E0,$8A,$7C,$1F,$80,$D7,$44,$C3,$BC
	.dc.b	$65,$FD,$C8,$BE,$6F,$F8,$A5,$AE,$67,$93,$E4,$F5,$BD,$BE,$9F,$BB
	.dc.b	$E0,$64,$FB,$23,$2F,$BA,$32,$FB,$23,$E2,$FD,$14,$8F,$86,$FC,$8F
	.dc.b	$07,$11,$F0,$A2,$F1,$35,$F8,$15,$F0,$FE,$A3,$E7,$F9,$73,$0F,$6E
	.dc.b	$A9,$D4,$7D,$09,$D2,$C3,$A3,$F0,$67,$6E,$70,$FF,$1A,$DF,$38,$EB
	.dc.b	$3F,$7B,$E5,$FF,$94,$7E,$4B,$FC,$96,$78,$1F,$EB,$9E,$9F,$E4,$18
	.dc.b	$7F,$AF,$9B,$1F,$EB,$D9,$E1,$8F,$8A,$1F,$93,$7F,$0E,$79,$C3,$F2
	.dc.b	$97,$47,$F3,$A8,$C3,$FC,$96,$B8,$47,$B3,$CA,$E7,$F3,$3F,$CF,$65
	.dc.b	$E9,$D3,$11,$C7,$DC,$FE,$5D,$F9,$47,$F4,$BD,$BF,$CA,$AD,$D3,$FC
	.dc.b	$DD,$87,$FB,$39,$FC,$5A,$F5,$B8,$FC,$04,$9C,$C7,$F3,$9F,$CA,$7F
	.dc.b	$D0,$7A,$FB,$F0,$FC,$79,$F9,$08,$CE,$3F,$EE,$4F,$3F,$F9,$E3,$7E
	.dc.b	$2B,$FA,$B4,$1B,$E7,$FF,$45,$F9,$7F,$95,$8F,$DC,$3E,$47,$BB,$F1
	.dc.b	$F9,$0E,$F1,$FE,$C1,$F9,$43,$A2,$FE,$07,$38,$3F,$AD,$CE,$5F,$D4
	.dc.b	$E4,$F8,$6F,$EA,$9F,$95,$7E,$CA,$7F,$5A,$D7,$F0,$62,$78,$3C,$BF
	.dc.b	$03,$E3,$E1,$C8,$C1,$DB,$A7,$AF,$85,$23,$07,$F6,$EF,$B1,$7D,$87
	.dc.b	$F7,$95,$F8,$35,$31,$1F,$DA,$7E,$E9,$F6,$9F,$DF,$FE,$13,$C7,$87
	.dc.b	$E6,$3C,$1F,$F3,$8F,$06,$EB,$FC,$09,$2F,$13,$0F,$F0,$AC,$39,$BF
	.dc.b	$E0,$03,$46,$7C,$2F,$07,$E0,$72,$30,$7F,$12,$CD,$F8,$07,$96,$A4
	.dc.b	$E6,$3F,$C0,$13,$A8,$C1,$0F,$9A,$FE,$19,$FF,$61,$7E,$1D,$FB,$77
	.dc.b	$F5,$3F,$5E,$C3,$F5,$0D,$8F,$E4,$73,$BE,$DF,$D9,$37,$AF,$FD,$99
	.dc.b	$0A,$E1,$5D,$2F,$FC,$29,$7B,$8F,$0F,$93,$82,$3F,$E2,$8B,$7D,$BB
	.dc.b	$B3,$8F,$4F,$F8,$77,$F0,$E7,$B7,$4C,$E1,$E6,$9C,$3F,$CE,$A7,$6E
	.dc.b	$4F,$E7,$BE,$67,$EA,$9F,$E3,$DF,$95,$FE,$45,$F1,$35,$FC,$AC,$6D
	.dc.b	$34,$9E,$7F,$F0,$F9,$76,$7F,$8C,$E7,$F8,$CF,$0F,$FB,$69,$FE,$B9
	.dc.b	$7D,$3D,$6D,$7C,$44,$8F,$2B,$5F,$D2,$F4,$7F,$24,$5F,$4C,$FA,$FF
	.dc.b	$E6,$D0,$FB,$7F,$C6,$BF,$93,$F4,$FF,$DD,$31,$8B,$FE,$4B,$23,$83
	.dc.b	$FE,$EE,$BE,$9C,$99,$AE,$71,$8A,$D9,$C6,$B9,$B8,$3C,$07,$17,$A6
	.dc.b	$1E,$6A,$E6,$D8,$FF,$7E,$9C,$A3,$CD,$87,$A2,$38,$E1,$CE,$FD,$18
	.dc.b	$78,$EF,$71,$93,$C2,$78,$E3,$55,$CA,$3C,$CF,$73,$C3,$7D,$E3,$AE
	.dc.b	$1E,$56,$E1,$FE,$EF,$BF,$85,$1F,$EF,$96,$BD,$4C,$7D,$BB,$E2,$B8
	.dc.b	$36,$3F,$DF,$D2,$8C,$47,$09,$FF,$7E,$3F,$57,$E4,$2A,$69,$1D,$BF
	.dc.b	$06,$FF,$23,$BF,$C3,$9E,$47,$D5,$8C,$BF,$5C,$CE,$9F,$88,$8C,$BF
	.dc.b	$8A,$8C,$E9,$FE,$AF,$E0,$FF,$01,$CE,$67,$C8,$7F,$45,$9D,$5F,$E3
	.dc.b	$7C,$BD,$63,$67,$35,$B3,$F2,$47,$63,$F8,$5E,$1D,$A3,$31,$93,$FC
	.dc.b	$29,$19,$73,$3F,$06,$31,$1D,$23,$17,$F1,$73,$FE,$C0,$FF,$B2,$5B
	.dc.b	$3F,$C6,$B5,$AF,$F8,$86,$2E,$7F,$5E,$BC,$73,$FC,$82,$7F,$C2,$8F
	.dc.b	$38,$CB,$C5,$BE,$1B,$ED,$FA,$B6,$3F,$90,$7D,$BF,$F8,$15,$FF,$28
	.dc.b	$C6,$6B,$FB,$2B,$5F,$86,$D7,$E4,$B5,$FD,$E3,$52,$FB,$3F,$3C,$69
	.dc.b	$AE,$9B,$1B,$9F,$4C,$7C,$6C,$5F,$6E,$F8,$9F,$5C,$E6,$31,$5E,$AF
	.dc.b	$AB,$F7,$9C,$79,$BF,$D3,$F1,$92,$65,$F3,$32,$57,$88,$93,$F2,$9F
	.dc.b	$17,$87,$F1,$EB,$E2,$C6,$3D,$9F,$B3,$F6,$EF,$87,$A7,$DC,$FB,$67
	.dc.b	$C1,$E3,$2F,$F3,$FD,$6D,$FB,$2D,$C7,$EF,$B5,$FC,$9B,$5F,$CC,$BF
	.dc.b	$D5,$23,$8F,$87,$F3,$23,$13,$CA,$3E,$8E,$9F,$E6,$DF,$82,$BC,$5F
	.dc.b	$EF,$CC,$3E,$13,$6F,$EF,$B5,$F0,$FD,$BF,$52,$4F,$A1,$EA,$B8,$29
	.dc.b	$1F,$9C,$64,$FC,$BF,$DB,$BE,$37,$F0,$6D,$8C,$C6,$0C,$F2,$48,$E2
	.dc.b	$DC,$3F,$4B,$5F,$EF,$C9,$EB,$F2,$3D,$08,$FD,$FF,$A4,$7A,$3F,$11
	.dc.b	$B1,$98,$C1,$F9,$6E,$D7,$C2,$52,$4F,$EC,$23,$2F,$E2,$FE,$D5,$ED
	.dc.b	$FD,$06,$C6,$63,$07,$E8,$BE,$3B,$F5,$04,$EF,$C1,$9F,$93,$95,$FB
	.dc.b	$E9,$46,$36,$FE,$0B,$0F,$B4,$E8,$4A,$E0,$FE,$0B,$DF,$FC,$D5,$7E
	.dc.b	$E9,$AF,$EE,$9A,$FE,$E9,$AB,$ED,$1F,$1D,$FB,$A6,$DF,$F1,$26,$CF
	.dc.b	$F1,$86,$BF,$C2,$3C,$CF,$D1,$CF,$EC,$5B,$1F,$C3,$9A,$EF,$B7,$F8
	.dc.b	$3F,$D9,$FE,$DA,$BF,$DC,$ED,$9F,$F5,$9F,$D5,$3E,$D3,$F9,$5C,$E7
	.dc.b	$B3,$FD,$74,$EE,$BF,$9E,$B5,$FE,$86,$D7,$FC,$2C,$D4,$F3,$BF,$F2
	.dc.b	$8F,$D4,$3E,$D3,$BF,$8D,$FC,$5E,$31,$3F,$E5,$63,$F9,$2F,$D9,$BE
	.dc.b	$CF,$FA,$0F,$86,$8C,$4E,$F2,$FE,$53,$0F,$B4,$DB,$A8,$FF,$42,$37
	.dc.b	$4F,$F6,$63,$6B,$BD,$70,$C3,$F9,$57,$DA,$3E,$62,$FC,$86,$1B,$7F
	.dc.b	$E6,$39,$D3,$67,$4D,$3C,$D9,$3C,$07,$CF,$FF,$5A,$6C,$FA,$F5,$7D
	.dc.b	$A7,$AE,$38,$CE,$AF,$FD,$66,$D7,$FB,$8B,$5F,$EE,$ED,$7F,$C9,$4D
	.dc.b	$7C,$57,$F3,$2B,$F2,$3B,$7F,$BE,$6B,$B3,$FE,$F1,$8C,$DF,$FA,$2B
	.dc.b	$93,$F9,$E4,$FF,$B5,$5B,$19,$F0,$BF,$A9,$46,$31,$52,$30,$DF,$3B
	.dc.b	$75,$C1,$FB,$FF,$68,$F9,$BB,$E1,$61,$B9,$7C,$79,$D3,$67,$4D,$3E
	.dc.b	$4F,$D3,$7C,$F3,$EE,$FC,$B5,$8C,$9F,$E0,$4D,$93,$FD,$C1,$B1,$98
	.dc.b	$C1,$FD,$BF,$EC,$DF,$69,$CF,$7F,$7D,$7F,$AC,$27,$2F,$DD,$F4,$7F
	.dc.b	$CF,$8D,$F1,$7F,$3B,$18,$AF,$F6,$D2,$8C,$C6,$9B,$3C,$4F,$F5,$A7
	.dc.b	$0B,$E5,$3E,$89,$FE,$03,$FA,$47,$C9,$E2,$B7,$37,$0F,$39,$F5,$AB
	.dc.b	$73,$A3,$F2,$3C,$5F,$97,$3F,$67,$F6,$9F,$DC,$79,$7F,$11,$19,$9F
	.dc.b	$94,$FC,$92,$38,$3F,$3D,$F6,$8F,$95,$E1,$8F,$FB,$E6,$31,$3F,$AD
	.dc.b	$FA,$3F,$D9,$9F,$B4,$F8,$8B,$19,$25,$7A,$B1,$77,$2B,$73,$EE,$ED
	.dc.b	$7D,$A7,$F5,$28,$C3,$6D,$87,$F8,$DA,$3C,$73,$AF,$4F,$F8,$9C,$6E
	.dc.b	$75,$18,$D3,$2F,$F9,$5B,$ED,$3F,$9A,$DF,$F9,$A9,$AF,$FA,$F5,$A9
	.dc.b	$A7,$F3,$EC,$3C,$71,$FE,$97,$F1,$9F,$47,$D7,$CC,$FF,$32,$FC,$52
	.dc.b	$73,$9F,$F0,$BA,$FD,$2D,$FC,$16,$E1,$D3,$5E,$7D,$FF,$CB,$A7,$37
	.dc.b	$FE,$56,$78,$DF,$F9,$69,$AE,$7C,$1F,$F1,$3C,$62,$3F,$CB,$7F,$64
	.dc.b	$FB,$3F,$F1,$5B,$6F,$FC,$94,$D9,$FF,$33,$B5,$FF,$33,$B9,$CB,$CD
	.dc.b	$F6,$6E,$9F,$E8,$73,$9B,$FF,$35,$CE,$5F,$E5,$FF,$66,$FA,$2F,$93
	.dc.b	$1F,$E9,$68,$C3,$FA,$FC,$7F,$DD,$D1,$87,$FD,$0E,$D8,$E1,$3F,$4B
	.dc.b	$DB,$7C,$9F,$FA,$92,$BF,$C6,$57,$FE,$C1,$8C,$56,$DD,$7C,$05,$9E
	.dc.b	$F3,$55,$72,$BE,$B1,$AC,$3F,$D9,$D8,$72,$6C,$64,$F0,$1E,$AC,$3A
	.dc.b	$5F,$AE,$F9,$CF,$69,$F0,$7D,$9F,$F8,$53,$C3,$E7,$28,$C3,$D0,$F2
	.dc.b	$FB,$9C,$3D,$AF,$E2,$91,$99,$DE,$1D,$9B,$9F,$C5,$BE,$18,$76,$9C
	.dc.b	$D8,$F0,$61,$D3,$0F,$3E,$5F,$E3,$8C,$F8,$4E,$72,$FF,$69,$E1,$CB
	.dc.b	$0E,$38,$FA,$B0,$E1,$1E,$5B,$E0,$F7,$AE,$07,$19,$DE,$3E,$9A,$DD
	.dc.b	$74,$39,$5F,$5C,$BD,$71,$82,$57,$A2,$32,$F3,$9F,$49,$C7,$2B,$5C
	.dc.b	$32,$F5,$E1,$E0,$C7,$C7,$18,$BD,$E1,$CE,$FE,$A5,$69,$F6,$5F,$A3
	.dc.b	$2F,$6B,$CF,$2F,$1E,$3C,$EB,$C0,$7E,$82,$32,$F2,$C7,$BC,$FD,$9C
	.dc.b	$AC,$E4,$E2,$7E,$11,$CB,$F1,$CF,$B8,$6C,$A6,$4F,$11,$8C,$3F,$C4
	.dc.b	$7E,$27,$FB,$D1,$B1,$FE,$FA,$6B,$BF,$3F,$FA,$DB,$96,$E9,$F9,$16
	.dc.b	$E1,$F8,$66,$CF,$E0,$7C,$3F,$2D,$6F,$0F,$CA,$56,$AB,$F4,$4D,$C1
	.dc.b	$7C,$30,$E9,$7C,$09,$7F,$A3,$9D,$3F,$8F,$BE,$44,$9F,$D1,$4E,$A7
	.dc.b	$BC,$72,$C3,$C9,$1C,$CC,$5F,$81,$FD,$85,$F9,$6F,$96,$1F,$22,$F8
	.dc.b	$92,$FF,$4F,$3A,$AF,$15,$F2,$C3,$C6,$F6,$3D,$0D,$7F,$82,$4F,$18
	.dc.b	$C6,$1D,$EF,$AD,$62,$FA,$E1,$FC,$0F,$0E,$13,$DC,$91,$97,$86,$1D
	.dc.b	$27,$73,$E9,$9F,$19,$F0,$CF,$E2,$9F,$4B,$F8,$27,$F8,$97,$8D,$F0
	.dc.b	$8F,$99,$19,$9F,$E4,$2D,$C7,$F8,$4C,$EE,$BF,$90,$CE,$5F,$E2,$8D
	.dc.b	$7F,$92,$46,$DF,$0E,$1D,$8F,$CF,$E7,$F9,$EA,$F2,$36,$33,$3F,$A4
	.dc.b	$9C,$BF,$C8,$7A,$9F,$65,$FE,$50,$8F,$F4,$17,$23,$EC,$FF,$57,$D1
	.dc.b	$FD,$89,$B9,$71,$BE,$17,$FE,$1D,$6B,$FB,$2C,$3A,$F9,$BF,$B2,$8E
	.dc.b	$0D,$51,$A7,$4B,$EC,$FF,$C5,$51,$86,$BF,$BC,$8D,$78,$DC,$DB,$1B
	.dc.b	$8F,$3B,$DA,$78,$E5,$FE,$11,$6E,$3F,$0F,$EC,$FF,$83,$CF,$F8,$82
	.dc.b	$9A,$BD,$5E,$7E,$1B,$EC,$DF,$15,$86,$DF,$FA,$C2,$F8,$36,$FE,$03
	.dc.b	$D1,$B1,$A8,$F8,$CD,$7F,$8C,$FD,$9F,$FB,$E9,$FF,$5D,$B5,$FF,$34
	.dc.b	$87,$18,$FA,$5F,$F5,$FC,$70,$EC,$F8,$6E,$DB,$1C,$34,$FE,$52,$F0
	.dc.b	$9E,$BD,$CF,$87,$E2,$D8,$FF,$66,$B5,$F1,$CF,$C9,$C3,$E5,$4E,$5B
	.dc.b	$5C,$8F,$0D,$F9,$3E,$C9,$DF,$DB,$EC,$A6,$5F,$ED,$A9,$CB,$67,$E5
	.dc.b	$C6,$23,$F1,$CD,$7F,$A0,$B5,$FE,$8A,$EF,$0F,$BC,$DB,$CD,$7B,$BC
	.dc.b	$4F,$7F,$C9,$CC,$3D,$EF,$C2,$FB,$3C,$47,$C6,$8B,$8F,$C0,$9F,$8F
	.dc.b	$A7,$C8,$F5,$A5,$FC,$88,$CB,$F2,$A3,$2F,$D0,$8F,$89,$A7,$D0,$F9
	.dc.b	$99,$3F,$0E,$32,$FE,$AF,$54,$7C,$42,$46,$5D,$BF,$88,$8D,$BF,$6B
	.dc.b	$E4,$FE,$EF,$8D,$F7,$BE,$72,$4F,$DE,$8C,$BF,$80,$8C,$BF,$86,$8F
	.dc.b	$C6,$FD,$9B,$D7,$A4,$9F,$F1,$86,$DF,$AD,$9F,$D8,$C7,$F8,$CD,$E0
	.dc.b	$EB,$8B,$FB,$DC,$62,$3F,$3E,$D9,$E5,$DB,$F9,$CF,$B3,$FC,$D5,$73
	.dc.b	$F8,$3F,$19,$9E,$89,$1F,$C4,$67,$F5,$4F,$F8,$19,$AF,$F8,$29,$AF
	.dc.b	$F9,$1D,$EA,$9A,$63,$07,$A9,$F0,$BD,$2F,$87,$E9,$CB,$F6,$1E,$F8
	.dc.b	$D3,$81,$49,$E1,$1F,$02,$7E,$6F,$CE,$48,$E1,$48,$FA,$3E,$06,$67
	.dc.b	$E3,$C5,$D7,$E8,$4F,$C7,$FA,$5F,$AF,$E6,$4B,$F1,$68,$9D,$47,$C7
	.dc.b	$7F,$8C,$C6,$63,$04,$3F,$89,$B9,$6C,$7E,$13,$E7,$BF,$99,$F8,$EF
	.dc.b	$A0,$7C,$BF,$F3,$96,$DF,$F4,$26,$BF,$CF,$DA,$FF,$8F,$3E,$CF,$F8
	.dc.b	$B3,$7D,$3F,$E5,$B1,$88,$FF,$21,$B5,$FE,$9E,$BE,$23,$FA,$43,$63
	.dc.b	$FA,$8B,$5F,$E9,$CD,$7C,$1E,$D7,$62,$3F,$D6,$E3,$FD,$CA,$DF,$57
	.dc.b	$C6,$47,$07,$F6,$1F,$B2,$FF,$52,$BF,$E5,$66,$BF,$E5,$A6,$BF,$E9
	.dc.b	$76,$E7,$FD,$26,$3F,$B3,$F9,$5F,$D1,$DF,$EF,$8D,$7F,$B6,$B5,$7C
	.dc.b	$A7,$F5,$D7,$FC,$CA,$D7,$FC,$D2,$D7,$FD,$4C,$D9,$7D,$8B,$7F,$EE
	.dc.b	$0D,$AF,$EF,$2D,$7F,$BB,$35,$EF,$F0,$3F,$D8,$AC,$D2,$47,$F8,$16
	.dc.b	$1F,$4D,$B2,$46,$5E,$5E,$CC,$8E,$0D,$7F,$C0,$DF,$60,$FA,$CF,$F0
	.dc.b	$33,$63,$FC,$1E,$D7,$71,$FD,$DB,$EB,$9F,$5F,$49,$3F,$E1,$98,$CB
	.dc.b	$FE,$18,$1E,$5F,$F8,$4D,$A6,$5C,$1D,$79,$24,$73,$6B,$FE,$27,$B7
	.dc.b	$D7,$FF,$82,$1B,$1F,$E2,$96,$BB,$8D,$1F,$DF,$A3,$3D,$BF,$E1,$B6
	.dc.b	$F8,$DF,$EA,$67,$FC,$21,$1F,$E1,$FF,$AE,$FD,$C2,$9F,$E6,$FF,$4A
	.dc.b	$FA,$FF,$F1,$5C,$A3,$18,$FF,$19,$BE,$37,$FE,$B2,$7F,$81,$47,$FA
	.dc.b	$C9,$AE,$7B,$BF,$C4,$71,$88,$FF,$7A,$3C,$9B,$1F,$E1,$88,$C5,$BE
	.dc.b	$BD,$F5,$F9,$E3,$FF,$0E,$FD,$7B,$D4,$A6,$72,$F3,$4F,$80,$8F,$F9
	.dc.b	$4F,$1F,$C7,$BF,$E5,$5F,$07,$F6,$2F,$F1,$7F,$13,$FC,$73,$18,$23
	.dc.b	$FE,$2C,$4D,$3E,$3F,$D7,$FF,$35,$9F,$E7,$D3,$FC,$69,$FF,$6B,$B5
	.dc.b	$1F,$C9,$7C,$4E,$E4,$78,$BD,$CF,$F1,$DF,$2F,$F9,$63,$83,$FD,$C3
	.dc.b	$F4,$2F,$AE,$70,$F4,$FA,$FF,$C9,$E3,$F9,$6F,$53,$FD,$CF,$18,$C5
	.dc.b	$18,$79,$DF,$2A,$57,$2C,$3B,$61,$D6,$FF,$DE,$27,$2A,$C1,$EC,$70
	.dc.b	$F2,$7A,$C6,$A1,$3D,$1F,$E1,$D3,$D3,$0F,$F7,$EF,$83,$E7,$6C,$72
	.dc.b	$9D,$9F,$2A,$B9,$61,$B8,$ED,$3E,$28,$F5,$63,$ED,$6E,$3A,$C3,$E1
	.dc.b	$5F,$A5,$AF,$9A,$B8,$5F,$86,$3E,$C9,$E1,$8F,$90,$E0,$E1,$5F,$EA
	.dc.b	$1C,$3D,$D1,$95,$F5,$EF,$82,$D3,$67,$FD,$55,$3A,$9C,$9F,$E8,$C8
	.dc.b	$7C,$55,$23,$FE,$90,$6E,$7F,$82,$C3,$FC,$11,$AF,$D1,$AF,$C8,$D7
	.dc.b	$D3,$9C,$C7,$D7,$3D,$51,$9F,$3B,$F8,$34,$64,$F8,$31,$9A,$E9,$1E
	.dc.b	$43,$55,$D2,$FF,$27,$1E,$03,$37,$D2,$33,$1C,$5F,$44,$67,$4F,$F6
	.dc.b	$66,$1D,$EB,$C4,$73,$AD,$E1,$E2,$9D,$47,$C1,$CF,$D3,$8F,$FA,$5D
	.dc.b	$B5,$FE,$E0,$D2,$E1,$FE,$58,$9A,$FE,$B1,$B5,$F6,$2B,$FC,$74,$FE
	.dc.b	$C5,$B1,$C3,$C6,$7D,$3D,$3A,$D3,$EB,$DE,$5E,$70,$FF,$22,$5F,$06
	.dc.b	$DF,$0A,$DC,$60,$DF,$D5,$F7,$3A,$FD,$7B,$E7,$71,$3A,$26,$9F,$C1
	.dc.b	$9F,$F3,$AC,$ED,$CE,$7F,$EE,$3C,$9F,$4B,$EB,$3D,$3C,$EF,$82,$C5
	.dc.b	$FA,$26,$E3,$D4,$FE,$27,$5B,$9D,$12,$35,$87,$F4,$DD,$B8,$BF,$5B
	.dc.b	$C3,$F7,$CF,$FB,$83,$EB,$FF,$CD,$F7,$FE,$69,$FA,$F7,$C0,$FD,$69
	.dc.b	$CD,$7F,$B9,$A7,$35,$BA,$FB,$71,$89,$E9,$EE,$62,$73,$E1,$CB,$E0
	.dc.b	$77,$7F,$0B,$BF,$F2,$EF,$66,$E8,$6F,$3E,$1E,$FF,$5F,$30,$FF,$41
	.dc.b	$DF,$17,$D5,$1D,$1F,$F3,$29,$A2,$E7,$F2,$E5,$AF,$E3,$E3,$13,$AC
	.dc.b	$7F,$6B,$0F,$AF,$7C,$37,$E6,$9B,$5F,$E9,$4A,$7C,$CF,$F8,$82,$3F
	.dc.b	$D2,$93,$AB,$FB,$47,$3E,$8E,$47,$E6,$A9,$F8,$3E,$2F,$8F,$C6,$77
	.dc.b	$3F,$3F,$F7,$0F,$97,$C4,$FC,$2F,$A5,$FF,$3A,$37,$D5,$F9,$B0,$F6
	.dc.b	$6C,$7E,$CB,$C4,$7C,$FD,$3F,$57,$F9,$67,$F8,$CB,$0F,$F5,$64,$E9
	.dc.b	$F9,$FF,$65,$FE,$AD,$6F,$C4,$7D,$A8,$7D,$BD,$8F,$DF,$FC,$77,$EC
	.dc.b	$78,$FE,$0B,$79,$7F,$84,$FD,$C3,$E5,$B7,$E5,$70,$F9,$59,$A7,$CD
	.dc.b	$FA,$A6,$38,$CB,$9F,$DA,$76,$3F,$81,$CB,$E7,$B1,$7F,$86,$F8,$6F
	.dc.b	$D0,$EB,$FC,$C3,$C0,$FF,$3D,$F9,$3F,$92,$9F,$93,$97,$E1,$5D,$DF
	.dc.b	$F5,$4F,$B4,$7C,$97,$F1,$A6,$C7,$E0,$79,$9F,$33,$9A,$FC,$EF,$B1
	.dc.b	$FA,$AD,$3C,$0D,$BF,$F0,$BE,$9F,$DE,$3E,$CB,$7E,$67,$F2,$6F,$B3
	.dc.b	$FE,$0F,$5F,$ED,$D6,$ED,$FB,$EF,$AD,$7D,$A6,$FD,$39,$7B,$7F,$92
	.dc.b	$CE,$A3,$04,$AF,$C9,$DF,$4D,$3F,$9D,$D6,$E7,$37,$FB,$6F,$B4,$7D
	.dc.b	$9D,$31,$F6,$97,$2F,$D9,$CE,$9B,$3A,$69,$FA,$7F,$AF,$7C,$E7,$F3
	.dc.b	$A7,$F0,$51,$98,$C7,$76,$5B,$19,$39,$47,$F4,$CF,$1D,$F3,$34,$FE
	.dc.b	$1B,$F6,$8C,$F1,$7F,$E4,$6D,$FF,$7B,$7C,$23,$07,$F8,$49,$C6,$DF
	.dc.b	$C1,$EF,$94,$EA,$3F,$C9,$F3,$AA,$E7,$87,$69,$E1,$85,$9F,$9B,$D7
	.dc.b	$F5,$0F,$AB,$E4,$5C,$7F,$85,$70,$FE,$43,$0F,$F2,$8B,$6C,$FE,$59
	.dc.b	$E7,$3F,$C8,$BE,$D1,$F3,$DF,$CB,$1F,$DB,$5F,$F1,$F7,$33,$BF,$1B
	.dc.b	$55,$B6,$C6,$4F,$75,$35,$7D,$47,$CF,$3E,$C7,$85,$6B,$73,$A6,$CE
	.dc.b	$9A,$4A,$FE,$47,$3B,$7F,$AC,$F1,$AF,$A4,$7F,$37,$F6,$7F,$CF,$B7
	.dc.b	$FD,$01,$FE,$06,$D8,$FE,$97,$18,$78,$18,$F1,$72,$D9,$F9,$FE,$6F
	.dc.b	$E9,$8F,$F0,$5F,$83,$FC,$B6,$32,$49,$FD,$84,$EB,$CE,$E1,$3F,$03
	.dc.b	$8F,$F2,$47,$F5,$CF,$B3,$FE,$C9,$C5,$F8,$E8,$CC,$EC,$FE,$B3,$3A
	.dc.b	$7F,$21,$1A,$C3,$E2,$BC,$FD,$BC,$5F,$F8,$C3,$67,$D1,$7E,$A4,$FD
	.dc.b	$57,$D1,$5C,$7F,$5D,$17,$1F,$9B,$36,$FF,$91,$7D,$97,$F3,$D9,$FE
	.dc.b	$C8,$D8,$CF,$CE,$7F,$76,$8C,$61,$FC,$61,$B4,$F1,$91,$86,$D7,$1F
	.dc.b	$53,$FC,$83,$F6,$5F,$C8,$6F,$FD,$6E,$DF,$8C,$FE,$B9,$C3,$AF,$22
	.dc.b	$67,$F9,$2A,$D4,$FD,$5B,$FE,$D7,$1C,$F2,$7D,$17,$E7,$A7,$FB,$F3
	.dc.b	$75,$FE,$80,$FE,$91,$BD,$9F,$E1,$28,$D6,$3F,$CD,$1B,$5E,$D7,$84
	.dc.b	$66,$7F,$C3,$3F,$63,$FE,$00,$8F,$F0,$9B,$6B,$FC,$1D,$F8,$FF,$F6
	.dc.b	$B7,$FC,$21,$19,$C3,$FA,$93,$6B,$FA,$1B,$67,$E6,$57,$1A,$DF,$BB
	.dc.b	$FE,$6A,$FB,$17,$D9,$7F,$A2,$BD,$CF,$F7,$4E,$1F,$EE,$29,$CE,$5F
	.dc.b	$E9,$4F,$96,$FF,$56,$C7,$F6,$58,$C4,$EE,$F9,$56,$9B,$19,$9D,$AA
	.dc.b	$E4,$FF,$AA,$2B,$84,$EA,$7C,$33,$93,$C9,$F0,$7D,$1E,$57,$67,$FC
	.dc.b	$4D,$87,$AA,$73,$7C,$EF,$57,$B7,$87,$B5,$CF,$EC,$BF,$BA,$F8,$3E
	.dc.b	$18,$C7,$AD,$FE,$08,$C7,$C4,$7A,$63,$33,$BC,$7D,$B7,$CA,$B6,$DC
	.dc.b	$BC,$8D,$D3,$E5,$E1,$CA,$B7,$8F,$7B,$DB,$6B,$78,$76,$8C,$50,$61
	.dc.b	$D5,$C3,$5F,$44,$F1,$C6,$47,$81,$F5,$34,$98,$78,$A3,$2D,$8C,$E1
	.dc.b	$FE,$F0,$BE,$DA,$32,$F6,$5E,$C9,$5E,$AB,$E2,$79,$B2,$F5,$BE,$A6
	.dc.b	$DF,$3C,$BC,$F7,$CA,$B8,$65,$E4,$C3,$C5,$87,$2C,$BE,$01,$E4,$CB
	.dc.b	$C7,$3A,$AE,$6F,$8F,$0E,$98,$78,$67,$2F,$E8,$EB,$8B,$E0,$3E,$81
	.dc.b	$CC,$A6,$4B,$7F,$83,$6F,$99,$FE,$71,$76,$7D,$87,$0E,$B9,$B2,$E5
	.dc.b	$B3,$FE,$16,$7F,$1F,$3C,$F0,$FC,$6B,$71,$FA,$9E,$07,$9A,$73,$5C
	.dc.b	$09,$87,$E5,$A7,$55,$F6,$EF,$94,$FE,$29,$AF,$8C,$E0,$72,$BE,$AF
	.dc.b	$E8,$23,$78,$7F,$BB,$23,$55,$D7,$0B,$1F,$96,$C3,$C2,$F7,$8C,$BD
	.dc.b	$F0,$FD,$4C,$EA,$31,$5C,$DF,$05,$7E,$CA,$75,$7F,$1E,$72,$4B,$E6
	.dc.b	$F6,$8D,$47,$A8,$CC,$6E,$FB,$46,$5F,$9B,$1C,$70,$F7,$9F,$02,$FC
	.dc.b	$2E,$30,$ED,$1E,$43,$2E,$27,$D0,$73,$6B,$FC,$43,$E9,$7F,$0B,$87
	.dc.b	$F0,$AF,$0F,$DA,$7C,$38,$CE,$3F,$94,$6D,$7F,$02,$9D,$BF,$CC,$A3
	.dc.b	$29,$A4,$BE,$9F,$2D,$D5,$93,$C5,$CB,$FD,$0B,$A5,$F6,$3F,$E3,$E8
	.dc.b	$FF,$76,$B5,$FF,$74,$37,$9B,$FD,$55,$F5,$AF,$7F,$F5,$71,$FE,$AD
	.dc.b	$9D,$72,$7F,$A7,$9B,$7F,$E0,$76,$E0,$C3,$94,$FF,$BF,$3E,$C5,$E2
	.dc.b	$74,$C3,$FD,$4F,$18,$9D,$4E,$5F,$F0,$97,$D8,$FE,$EB,$2F,$F5,$F5
	.dc.b	$6E,$7F,$E0,$03,$63,$35,$C7,$15,$F1,$C3,$E9,$7A,$BD,$71,$F9,$CD
	.dc.b	$DF,$FD,$B1,$1F,$E8,$3F,$0D,$F1,$5F,$EE,$C8,$FF,$34,$CE,$9F,$F8
	.dc.b	$02,$DC,$D8,$76,$9E,$55,$C5,$4E,$9F,$E8,$DF,$61,$FE,$D2,$7F,$CE
	.dc.b	$AD,$77,$5F,$C7,$23,$13,$C6,$51,$FE,$A6,$78,$F5,$FB,$63,$75,$C6
	.dc.b	$31,$D1,$F0,$7B,$5F,$13,$49,$87,$8A,$EB,$E2,$3F,$AE,$4F,$07,$FD
	.dc.b	$16,$D7,$47,$92,$3E,$77,$D7,$FF,$51,$8F,$F4,$8B,$5D,$B7,$2F,$65
	.dc.b	$7F,$BC,$27,$7D,$BC,$88,$F8,$3C,$FA,$F2,$8B,$5C,$A7,$DD,$CB,$2B
	.dc.b	$CE,$FC,$1F,$94,$C4,$7A,$30,$FC,$C3,$A3,$DB,$1F,$1E,$3E,$47,$A3
	.dc.b	$27,$E0,$C6,$5F,$87,$19,$7A,$47,$C2,$FB,$09,$1F,$09,$FA,$5E,$8E
	.dc.b	$23,$E1,$C5,$CB,$E1,$CD,$D7,$FA,$EE,$AA,$C5,$FD,$4A,$FB,$9E,$1F
	.dc.b	$DD,$F9,$D9,$3F,$5A,$32,$FD,$88,$CB,$F0,$63,$F4,$9F,$61,$FE,$92
	.dc.b	$8F,$F6,$3B,$5F,$BA,$DC,$BF,$24,$FE,$C7,$2F,$EE,$2E,$5B,$E9,$74
	.dc.b	$3A,$F6,$7E,$C7,$EC,$3F,$2B,$87,$F8,$9D,$A9,$B6,$F4,$38,$73,$CA
	.dc.b	$FF,$6D,$D6,$A7,$11,$F9,$28,$E6,$65,$F5,$F9,$DE,$EF,$66,$46,$08
	.dc.b	$7B,$9C,$BF,$02,$3E,$2F,$CF,$C8,$F8,$AF,$C8,$F9,$5C,$47,$C6,$8B
	.dc.b	$B7,$C6,$9F,$81,$F4,$F2,$BD,$2F,$D0,$F2,$31,$1E,$98,$B5,$F3,$A6
	.dc.b	$F0,$7F,$87,$B3,$56,$2F,$FC,$A1,$B2,$FF,$91,$7D,$0B,$E6,$E9,$2B
	.dc.b	$FC,$AD,$97,$F1,$3F,$19,$F4,$3F,$E6,$D9,$FD,$DF,$D8,$7F,$0D,$6F
	.dc.b	$85,$FC,$A6,$31,$7F,$C6,$C7,$E5,$AE,$DF,$E6,$52,$F8,$4A,$4C,$3C
	.dc.b	$3C,$3F,$E6,$7C,$DF,$63,$FE,$30,$CB,$F8,$F6,$1F,$EC,$2B,$D9,$FE
	.dc.b	$6D,$70,$F9,$BD,$AF,$E7,$9F,$75,$F1,$23,$62,$7F,$C4,$0F,$B3,$D5
	.dc.b	$F5,$1C,$3E,$E3,$FA,$26,$3F,$E9,$7C,$7F,$97,$37,$3F,$F2,$6B,$6F
	.dc.b	$FC,$FD,$8B,$E2,$3F,$A8,$79,$BF,$D0,$EF,$D8,$DA,$FF,$76,$37,$5F
	.dc.b	$EF,$DE,$07,$F5,$3F,$94,$FE,$79,$9F,$F3,$D6,$E1,$FE,$5C,$6E,$1F
	.dc.b	$D2,$B8,$9F,$1D,$FD,$7F,$EC,$FF,$01,$3F,$1B,$24,$62,$3F,$9D,$46
	.dc.b	$3A,$78,$7B,$3E,$1F,$83,$95,$F2,$32,$47,$F6,$7F,$B9,$7C,$9F,$9F
	.dc.b	$EA,$E5,$ED,$D7,$D4,$7A,$FD,$E9,$3C,$DE,$58,$FF,$75,$FB,$47,$D8
	.dc.b	$D2,$57,$A7,$E0,$3C,$2F,$DA,$FB,$37,$DA,$7F,$79,$6C,$66,$3F,$C0
	.dc.b	$0F,$E6,$8F,$84,$53,$E0,$97,$9B,$F4,$0D,$E6,$FC,$92,$F9,$DE,$DE
	.dc.b	$0C,$8F,$13,$FE,$09,$C9,$F6,$94,$93,$F9,$08,$CB,$7E,$03,$81,$A7
	.dc.b	$FB,$3C,$69,$B1,$FE,$15,$FB,$0F,$CD,$47,$F7,$1C,$9F,$56,$FB,$4F
	.dc.b	$F1,$34,$61,$E1,$E6,$CA,$FD,$0D,$F2,$AF,$F0,$43,$75,$FE,$03,$F6
	.dc.b	$8F,$B4,$E1,$7F,$DE,$AB,$4F,$FB,$6D,$B6,$AE,$13,$DB,$9F,$D0,$47
	.dc.b	$F6,$3F,$68,$F8,$EC,$47,$59,$B7,$FE,$1B,$AD,$4F,$EE,$3D,$0F,$D7
	.dc.b	$26,$9F,$B1,$97,$DA,$3D,$5F,$F1,$E4,$EB,$C0,$A7,$C6,$9D,$5F,$5F
	.dc.b	$4B,$64,$7F,$83,$FA,$5F,$E4,$77,$FD,$F5,$7F,$CA,$5E,$04,$3F,$C3
	.dc.b	$F1,$B5,$F4,$6F,$AF,$FF,$1F,$C3,$ED,$3F,$25,$4F,$7F,$FD,$AF,$18
	.dc.b	$6C,$7F,$B6,$AF,$96,$AB,$E5,$39,$C3,$AE,$1C,$67,$4D,$76,$71,$BE
	.dc.b	$6E,$0F,$3C,$F2,$AF,$F7,$49,$E5,$8E,$6E,$30,$F2,$C7,$39,$E0,$DC
	.dc.b	$36,$F9,$EB,$93,$DD,$B8,$7E,$CF,$0F,$1B,$6F,$9D,$7F,$52,$C3,$BB
	.dc.b	$4E,$A7,$A3,$0B,$7E,$C8,$FF,$7F,$B5,$F8,$D1,$FE,$F8,$8C,$47,$C3
	.dc.b	$F0,$DF,$17,$96,$C7,$CF,$3E,$93,$87,$AB,$5D,$1F,$57,$E4,$2B,$86
	.dc.b	$BF,$EF,$F2,$86,$B4,$FA,$97,$F9,$BE,$7F,$7C,$75,$33,$C3,$FE,$89
	.dc.b	$D3,$F8,$E4,$65,$FE,$39,$19,$D7,$C7,$87,$C8,$C3,$A5,$FE,$04,$E1
	.dc.b	$3F,$EB,$EA,$E1,$1E,$93,$BE,$1E,$27,$CD,$87,$68,$C6,$1D,$6B,$8D
	.dc.b	$7B,$EB,$4F,$73,$8D,$F7,$BF,$C3,$E0,$FC,$BB,$FD,$53,$1F,$DF,$A3
	.dc.b	$1F,$47,$FE,$0D,$8F,$F0,$BC,$EA,$FF,$92,$29,$D4,$7F,$AB,$A3,$55
	.dc.b	$FE,$15,$75,$5D,$5C,$BD,$2B,$8C,$BE,$D5,$F4,$18,$6D,$79,$57,$B5
	.dc.b	$99,$93,$9E,$EB,$97,$F3,$9B,$FE,$71,$3E,$DF,$B2,$FE,$60,$DE,$EF
	.dc.b	$0C,$3A,$6C,$49,$D1,$C6,$31,$7F,$D1,$7A,$3F,$97,$4F,$C9,$6C,$67
	.dc.b	$DC,$F6,$C6,$35,$FF,$1C,$36,$BF,$C8,$6D,$AF,$F6,$BB,$FE,$3B,$8E
	.dc.b	$9D,$BC,$88,$F7,$BB,$5F,$D5,$FF,$2F,$7A,$DF,$31,$FF,$29,$57,$DD
	.dc.b	$72,$DC,$34,$47,$53,$6B,$A4,$FD,$6F,$BD,$FB,$71,$8A,$FC,$44,$63
	.dc.b	$2F,$F2,$E1,$FE,$63,$8F,$F1,$23,$F4,$27,$FC,$46,$E5,$BA,$7E,$2B
	.dc.b	$D4,$7D,$A3,$D0,$C3,$70,$E7,$58,$9E,$10,$FA,$8B,$E4,$E1,$B1,$FE
	.dc.b	$35,$FC,$C3,$FA,$33,$7B,$BC,$11,$8C,$FF,$C6,$EA,$3C,$9F,$98,$BD
	.dc.b	$4A,$79,$FA,$DE,$9E,$31,$FF,$21,$7E,$65,$F9,$66,$F9,$5E,$D8,$C6
	.dc.b	$7F,$E4,$45,$1F,$07,$EA,$33,$C7,$FE,$85,$73,$7F,$E4,$D7,$2F,$F9
	.dc.b	$E5,$CD,$FF,$93,$BE,$7F,$FA,$C4,$FF,$94,$DB,$19,$F8,$FF,$9D,$18
	.dc.b	$BF,$F4,$97,$DA,$7C,$DA,$FF,$2B,$BF,$C6,$23,$3C,$7F,$56,$31,$87
	.dc.b	$FA,$27,$F2,$B7,$C7,$7D,$95,$F6,$B6,$30,$FD,$DF,$CC,$FF,$05,$8F
	.dc.b	$FA,$85,$B7,$FE,$93,$7F,$64,$FF,$9A,$DC,$E4,$EA,$7E,$68,$FB,$9F
	.dc.b	$C4,$2F,$B4,$B1,$87,$F1,$9B,$7F,$0A,$BF,$F3,$9B,$FC,$11,$BE,$46
	.dc.b	$BC,$8F,$C9,$7E,$31,$F1,$74,$FF,$53,$36,$BC,$4B,$8F,$71,$F9,$9F
	.dc.b	$45,$F2,$77,$D4,$C3,$70,$FE,$5D,$F6,$5E,$96,$ED,$FC,$8A,$1F,$53
	.dc.b	$62,$4E,$8E,$31,$8C,$7F,$CA,$D5,$CD,$B5,$CE,$75,$C7,$FE,$CF,$F8
	.dc.b	$1F,$F4,$CE,$1F,$E9,$75,$E7,$7E,$C5,$CE,$1F,$E9,$87,$33,$BF,$8A
	.dc.b	$D5,$6D,$B1,$FE,$A1,$FC,$95,$F5,$3F,$BB,$5F,$3B,$63,$0D,$9E,$0D
	.dc.b	$C7,$FD,$4F,$5C,$27,$FD,$97,$F6,$DE,$4C,$D1,$FE,$07,$8C,$3F,$EE
	.dc.b	$CD,$BF,$04,$F0,$CB,$FA,$AE,$0A,$DA,$B7,$E0,$5F,$EA,$EB,$E0,$7F
	.dc.b	$63,$71,$3B,$8F,$05,$FF,$58,$9C,$D7,$92,$BF,$C9,$17,$DA,$3A,$FC
	.dc.b	$5F,$F1,$DF,$8D,$FD,$74,$7F,$C0,$16,$E3,$FE,$F3,$7F,$5B,$7F,$9F
	.dc.b	$72,$DE,$9E,$47,$3F,$72,$D7,$FB,$47,$ED,$7F,$D4,$AD,$9F,$E5,$91
	.dc.b	$8A,$FF,$69,$BD,$3B,$8E,$71,$F5,$46,$2B,$8F,$C4,$F2,$7F,$2D,$F2
	.dc.b	$F3,$1F,$CD,$FC,$5D,$CB,$F2,$73,$71,$F8,$83,$85,$59,$BF,$B1,$56
	.dc.b	$FE,$A5,$6A,$7F,$4F,$3F,$D1,$7F,$16,$7D,$B5,$F6,$70,$DA,$FF,$79
	.dc.b	$4E,$9B,$3A,$69,$FE,$F7,$F9,$EF,$D4,$67,$F0,$1B,$1F,$E8,$1D,$91
	.dc.b	$FB,$48,$E1,$D9,$FD,$5E,$37,$A7,$F1,$56,$D7,$0E,$CD,$57,$EF,$7D
	.dc.b	$19,$87,$F8,$BF,$E9,$FF,$A4,$BF,$C1,$23,$38,$FE,$F1,$FF,$12,$CE
	.dc.b	$A3,$13,$B2,$3F,$ED,$78,$D6,$5F,$43,$7F,$EB,$15,$FE,$9B,$8F,$BB
	.dc.b	$F8,$87,$DF,$8F,$73,$6F,$F9,$73,$F8,$53,$19,$9A,$4E,$2F,$EF,$0E
	.dc.b	$BE,$D7,$FC,$05,$7F,$25,$B7,$FC,$E5,$7C,$D7,$F7,$F7,$38,$1A,$48
	.dc.b	$D7,$AB,$FE,$05,$FB,$37,$D0,$53,$1F,$29,$72,$F7,$1F,$66,$76,$E3
	.dc.b	$2F,$E9,$51,$87,$F7,$86,$27,$93,$8C,$BF,$C9,$FD,$3F,$E1,$5F,$87
	.dc.b	$FD,$41,$F1,$0A,$E7,$6D,$37,$1F,$05,$3E,$CF,$FA,$53,$70,$FF,$10
	.dc.b	$47,$F4,$B9,$DB,$A2,$46,$BC,$BD,$BF,$E2,$7F,$CC,$1F,$51,$FC,$47
	.dc.b	$E3,$BF,$CB,$53,$FD,$9E,$5F,$7B,$9F,$8B,$FC,$E4,$70,$CB,$FD,$55
	.dc.b	$18,$9D,$57,$1C,$2E,$6C,$39,$E1,$C2,$B4,$D9,$D5,$F0,$AD,$D7,$FA
	.dc.b	$C7,$0E,$D7,$B9,$F1,$4E,$4F,$45,$71,$3C,$3F,$99,$FF,$8C,$EB,$84
	.dc.b	$63,$E0,$BC,$F4,$F7,$5F,$5F,$8F,$B6,$7D,$5F,$F7,$7A,$EC,$6E,$75
	.dc.b	$19,$DF,$E4,$CE,$72,$F7,$2B,$E7,$C1,$8C,$7B,$D6,$A7,$35,$D9,$B7
	.dc.b	$D1,$BB,$7F,$B7,$EF,$8D,$6F,$1F,$14,$FF,$B8,$EB,$72,$3C,$53,$A7
	.dc.b	$15,$CA,$FB,$4E,$67,$78,$F9,$63,$15,$BC,$3B,$46,$F0,$EF,$7C,$0F
	.dc.b	$8F,$5B,$7C,$51,$A2,$4F,$4C,$BE,$5E,$1D,$C9,$5C,$A3,$2F,$BE,$33
	.dc.b	$8F,$D0,$C3,$B4,$F8,$63,$D7,$3A,$C7,$E4,$5F,$29,$CD,$73,$3B,$65
	.dc.b	$B7,$CF,$8F,$B4,$F3,$46,$63,$B5,$FC,$2C,$3E,$E6,$1D,$63,$13,$D7
	.dc.b	$0F,$34,$E6,$3A,$D7,$D5,$61,$F2,$4B,$97,$E1,$1B,$8F,$E4,$DC,$F3
	.dc.b	$FF,$A2,$9E,$27,$E3,$DC,$3A,$F8,$4C,$B9,$6D,$7F,$24,$7F,$25,$3C
	.dc.b	$F1,$FC,$83,$72,$FC,$67,$8F,$E7,$9C,$D7,$02,$61,$F9,$99,$D5,$7E
	.dc.b	$4E,$F9,$4F,$E0,$9A,$F8,$CE,$07,$2B,$EA,$FE,$8E,$37,$82,$FA,$CF
	.dc.b	$3C,$38,$61,$DD,$FD,$64,$EA,$7B,$61,$D0,$95,$EC,$8E,$27,$33,$C1
	.dc.b	$5C,$9C,$9C,$A3,$B3,$E8,$36,$49,$EA,$FE,$82,$BC,$57,$F3,$E3,$55
	.dc.b	$CA,$BF,$7D,$5A,$C3,$DE,$4C,$3C,$57,$C0,$F3,$57,$67,$F8,$75,$76
	.dc.b	$39,$4F,$A3,$0F,$91,$3C,$AB,$84,$78,$DF,$0C,$7F,$33,$FC,$C5,$F6
	.dc.b	$5F,$3B,$91,$F9,$9B,$DC,$C3,$6F,$85,$6E,$F8,$C6,$08,$7F,$37,$B4
	.dc.b	$7F,$A2,$A3,$C1,$EB,$F1,$A9,$18,$21,$C6,$7A,$B6,$3C,$13,$E7,$FC
	.dc.b	$C7,$FC,$D9,$3F,$DC,$DB,$E2,$7F,$95,$9B,$C9,$FE,$37,$5E,$0F,$EC
	.dc.b	$E7,$53,$4D,$16,$BF,$D9,$95,$C9,$B5,$9C,$9F,$6F,$FD,$AA,$7F,$BD
	.dc.b	$AF,$97,$D3,$63,$3D,$58,$33,$32,$32,$7F,$81,$3F,$1B,$7D,$3F,$76
	.dc.b	$C6,$63,$1F,$23,$96,$46,$4E,$51,$E7,$FC,$6D,$E8,$D2,$4F,$F8,$4A
	.dc.b	$1F,$67,$88,$FF,$0F,$47,$07,$53,$CE,$7E,$17,$E3,$7F,$32,$7F,$C3
	.dc.b	$6D,$7F,$C4,$CD,$E1,$FF,$44,$BF,$E3,$E8,$DE,$FC,$88,$F2,$F2,$FE
	.dc.b	$92,$F9,$97,$DC,$61,$B1,$FE,$39,$35,$19,$FB,$8D,$32,$35,$3C,$1F
	.dc.b	$F1,$E9,$D6,$74,$E2,$35,$8F,$E3,$9B,$CB,$FC,$ED,$57,$E0,$FF,$19
	.dc.b	$CD,$B9,$51,$86,$E3,$BA,$E0,$CA,$FF,$21,$D6,$CE,$AF,$F9,$1B,$F1
	.dc.b	$77,$BD,$F9,$15,$F5,$74,$C4,$49,$E1,$19,$79,$47,$CC,$FB,$6F,$B9
	.dc.b	$1F,$09,$BA,$FE,$B1,$BA,$FF,$59,$7F,$0B,$3F,$3D,$CB,$FE,$11,$6A
	.dc.b	$69,$C1,$ED,$C5,$E1,$23,$CF,$8B,$F5,$9B,$BE,$1E,$FC,$3C,$AF,$F3
	.dc.b	$3C,$E5,$B3,$97,$FC,$E9,$09,$E7,$3F,$B7,$F8,$57,$C3,$52,$5F,$F9
	.dc.b	$CA,$3F,$8B,$3F,$E7,$D7,$6E,$63,$6E,$7B,$FF,$2B,$F6,$EF,$AC,$F9
	.dc.b	$9E,$5C,$AF,$CA,$4E,$5B,$39,$7E,$84,$F2,$89,$1A,$21,$F5,$7B,$7F
	.dc.b	$88,$7E,$6A,$F4,$6F,$C1,$4A,$FF,$45,$CE,$5B,$39,$7F,$D3,$11,$FE
	.dc.b	$D8,$72,$7F,$A4,$BF,$36,$7C,$17,$F1,$D5,$F0,$74,$C4,$49,$E1,$19
	.dc.b	$79,$47,$EE,$3F,$0F,$7E,$07,$49,$3F,$B7,$8F,$CD,$B6,$32,$FE,$D2
	.dc.b	$38,$BB,$EA,$CF,$57,$F4,$09,$CC,$EE,$BC,$15,$FC,$D7,$C7,$FE,$61
	.dc.b	$3F,$EB,$75,$F1,$D9,$34,$FF,$3B,$FB,$2F,$7C,$FF,$AD,$1A,$FF,$AF
	.dc.b	$DB,$8B,$B7,$FC,$E1,$E1,$A5,$F2,$BF,$D1,$7E,$E5,$F0,$37,$E3,$A6
	.dc.b	$5F,$EC,$09,$CB,$FE,$C2,$F6,$7F,$9E,$BF,$EC,$D5,$E6,$FF,$4F,$6C
	.dc.b	$6F,$A3,$FA,$97,$E6,$EF,$A4,$A4,$9F,$F6,$6C,$3E,$3B,$FA,$C7,$E7
	.dc.b	$0F,$AF,$BF,$05,$2F,$FD,$A7,$39,$7F,$DA,$BF,$97,$D7,$FD,$C0,$FF
	.dc.b	$AA,$DC,$CF,$F6,$38,$D7,$27,$C5,$9F,$E4,$5F,$8D,$FC,$8E,$5F,$8E
	.dc.b	$D8,$CF,$67,$F6,$B8,$F9,$31,$FE,$E4,$FB,$A7,$CD,$7F,$6F,$7F,$39
	.dc.b	$1F,$EF,$16,$E1,$DF,$A1,$1C,$27,$FD,$DB,$F9,$CF,$C8,$7F,$DF,$0F
	.dc.b	$FA,$EA,$B8,$37,$2F,$B6,$EE,$3A,$7E,$74,$FA,$BF,$F0,$03,$6F,$FD
	.dc.b	$F2,$D5,$3C,$38,$FF,$C0,$DF,$11,$FE,$7B,$8F,$33,$F1,$B1,$FC,$4B
	.dc.b	$6B,$FA,$43,$B8,$FC,$6F,$DD,$FF,$09,$9E,$B1,$8F,$A1,$FF,$09,$49
	.dc.b	$BE,$4F,$F8,$57,$E2,$31,$7F,$EC,$5C,$7E,$34,$62,$39,$9F,$D5,$DC
	.dc.b	$47,$CB,$8F,$EB,$03,$F1,$5F,$F1,$03,$8F,$6A,$C7,$F8,$9E,$5F,$32
	.dc.b	$F9,$4C,$7E,$33,$C7,$0F,$16,$BE,$38,$C4,$78,$AF,$8D,$73,$BE,$53
	.dc.b	$C0,$F9,$AE,$1E,$0D,$9F,$C8,$1E,$2E,$FE,$47,$3A,$95,$CE,$74,$DF
	.dc.b	$BB,$7D,$17,$AD,$B1,$98,$C7,$46,$59,$19,$39,$46,$1F,$64,$F2,$6C
	.dc.b	$F2,$7C,$5F,$9D,$FD,$F7,$F1,$2D,$B3,$4F,$C3,$38,$E3,$FA,$11,$96
	.dc.b	$F7,$FD,$3F,$BB,$7D,$55,$FA,$A9,$8F,$C1,$9C,$B6,$78,$11,$D7,$E0
	.dc.b	$BF,$77,$F3,$CF,$AF,$9F,$CE,$7F,$C3,$E9,$A6,$B8,$FF,$20,$DA,$FD
	.dc.b	$2B,$AE,$5E,$9C,$BF,$E3,$EF,$A9,$FD,$FE,$0B,$7D,$DF,$F3,$AF,$43
	.dc.b	$F8,$93,$77,$FF,$76,$3F,$CF,$AF,$8D,$70,$AF,$F7,$84,$62,$78,$4F
	.dc.b	$56,$BB,$3C,$11,$B9,$E3,$C9,$FB,$C9,$D1,$FE,$7D,$BF,$0C,$FF,$BF
	.dc.b	$4F,$29,$C3,$0E,$CD,$9E,$2E,$9F,$73,$55,$F8,$5F,$1D,$F8,$70,$F3
	.dc.b	$92,$BA,$3F,$06,$3C,$D3,$F2,$4E,$37,$5F,$11,$6F,$E3,$B7,$D8,$7D
	.dc.b	$E6,$E3,$E1,$C6,$1C,$4F,$F8,$46,$75,$1E,$89,$D3,$8F,$96,$CB,$F6
	.dc.b	$DA,$9A,$47,$E6,$DB,$D0,$FE,$95,$97,$F7,$A9,$E4,$6A,$33,$9F,$9E
	.dc.b	$FE,$7C,$FF,$1F,$9C,$BF,$06,$32,$FF,$36,$8C,$BF,$1E,$33,$7E,$1E
	.dc.b	$0D,$CF,$83,$0F,$F0,$5C,$E6,$3F,$06,$67,$0F,$F6,$66,$1D,$23,$07
	.dc.b	$8E,$F9,$BF,$04,$F5,$38,$C3,$C3,$3E,$8B,$7D,$E7,$F8,$3E,$BF,$8C
	.dc.b	$3F,$ED,$A6,$9F,$C8,$0F,$E7,$CE,$6F,$F9,$63,$C5,$BF,$1B,$B7,$FA
	.dc.b	$BC,$66,$BF,$94,$BC,$5F,$E5,$7F,$79,$FD,$36,$FF,$94,$36,$BF,$9C
	.dc.b	$2F,$A2,$E9,$48,$D3,$CB,$C3,$CC,$62,$FF,$AA,$3C,$5E,$B1,$CB,$E0
	.dc.b	$BF,$C4,$31,$FC,$DD,$B8,$7F,$47,$5F,$03,$BF,$56,$56,$DE,$51,$8A
	.dc.b	$FE,$9D,$F7,$4F,$BC,$F3,$B6,$FF,$A2,$3F,$EC,$F6,$C6,$88,$EB,$E2
	.dc.b	$6D,$74,$9B,$A7,$F5,$2B,$E5,$39,$BF,$EA,$1F,$79,$FE,$F2,$6F,$C6
	.dc.b	$7C,$88,$FF,$78,$C6,$0D,$D6,$4D,$4D,$8C,$9F,$D6,$FE,$F3,$EC,$3F
	.dc.b	$DF,$E9,$5F,$52,$73,$7F,$ED,$5C,$1E,$DF,$E0,$FD,$4F,$F0,$CE,$D6
	.dc.b	$73,$97,$F7,$19,$CB,$7E,$F3,$FC,$BA,$DF,$7B,$85,$F3,$9C,$E3,$FE
	.dc.b	$0E,$BF,$F3,$47,$A9,$DF,$0F,$F1,$61,$C4,$EC,$62,$78,$7D,$E7,$F9
	.dc.b	$0A,$3C,$4F,$FA,$85,$BE,$07,$A6,$31,$3D,$FE,$F3,$FC,$27,$5E,$96
	.dc.b	$C7,$8F,$A9,$E9,$7C,$19,$CE,$DF,$E4,$BF,$BC,$FC,$A5,$FA,$5B,$19
	.dc.b	$FA,$1F,$91,$18,$C7,$FD,$07,$F7,$76,$DB,$23,$0D,$BF,$A1,$18,$F7
	.dc.b	$FF,$B0,$5F,$CF,$6D,$FC,$95,$E7,$FD,$77,$39,$7F,$98,$7E,$EF,$E7
	.dc.b	$B7,$C2,$FB,$91,$8C,$3F,$D2,$EF,$F1,$99,$FF,$34,$7D,$DF,$F8,$1A
	.dc.b	$7E,$F2,$F4,$7F,$08,$E7,$2F,$B4,$E6,$7F,$09,$F7,$7F,$B4,$BF,$F3
	.dc.b	$93,$63,$FD,$4A,$FE,$D9,$DF,$3F,$E3,$FE,$EF,$FD,$9F,$1F,$8D,$6E
	.dc.b	$1F,$87,$7F,$DA,$31,$9F,$8E,$FC,$B7,$57,$F0,$9A,$F9,$AD,$E2,$FF
	.dc.b	$57,$F8,$DA,$F1,$7F,$39,$F7,$7F,$1B,$1F,$F5,$BB,$70,$FA,$AF,$CC
	.dc.b	$8F,$CF,$39,$BF,$EE,$7E,$BF,$E9,$7C,$FF,$C9,$5F,$E6,$DF,$F2,$3D
	.dc.b	$FE,$95,$CE,$FF,$E5,$2F,$B4,$F5,$B7,$5F,$E4,$B1,$EC,$BF,$F3,$03
	.dc.b	$AB,$FF,$3B,$36,$3F,$D1,$6D,$A3,$C5,$E9,$3F,$E5,$B8,$D1,$3E,$17
	.dc.b	$BD,$FE,$F7,$E3,$5F,$2B,$7C,$CC,$37,$0F,$DF,$4E,$9B,$3A,$69,$FC
	.dc.b	$0F,$EE,$DF,$31,$FC,$25,$7C,$CD,$8C,$36,$78,$37,$5F,$E0,$95,$C2
	.dc.b	$7F,$CF,$D0,$FB,$BF,$13,$FE,$96,$31,$97,$CA,$F9,$1F,$EB,$51,$8A
	.dc.b	$E3,$5C,$37,$FF,$69,$4E,$4F,$9F,$5C,$27,$55,$FE,$27,$6C,$FF,$AE
	.dc.b	$BE,$69,$F7,$7B,$C7,$FA,$EC,$62,$FF,$92,$7C,$AF,$EB,$6F,$F7,$ED
	.dc.b	$E5,$3F,$AF,$9F,$F2,$E4,$63,$A7,$F9,$84,$EC,$9E,$06,$23,$F9,$A7
	.dc.b	$DD,$BE,$D6,$C6,$1B,$87,$F3,$59,$D3,$67,$4D,$3F,$0D,$F5,$FF,$92
	.dc.b	$3F,$27,$F7,$76,$32,$4C,$FF,$2B,$19,$7F,$2F,$37,$87,$F3,$D3,$6B
	.dc.b	$DB,$F7,$6F,$97,$FE,$98,$FF,$5F,$C3,$FA,$13,$99,$FE,$98,$E6,$FF
	.dc.b	$DF,$4E,$67,$7E,$BF,$2B,$E6,$FF,$57,$F9,$FF,$D3,$70,$1A,$7E,$97
	.dc.b	$1F,$9E,$E3,$AF,$81,$C7,$B2,$CF,$EA,$7E,$3B,$F6,$A7,$F8,$6B,$EE
	.dc.b	$FF,$A2,$C6,$49,$9F,$EE,$27,$5E,$8F,$09,$CE,$DF,$C0,$67,$31,$83
	.dc.b	$F0,$3F,$77,$F5,$AF,$FB,$23,$69,$3A,$E2,$FE,$E9,$39,$8E,$3C,$98
	.dc.b	$33,$36,$3F,$BB,$7D,$DF,$F7,$ED,$CD,$1F,$55,$FD,$FD,$C7,$AB,$4C
	.dc.b	$45,$AF,$F0,$07,$DD,$BE,$4F,$FC,$0C,$D8,$CF,$C9,$E2,$64,$70,$77
	.dc.b	$18,$DB,$FC,$05,$E6,$7A,$DF,$EC,$AF,$E9,$CC,$67,$FE,$4B,$D3,$FC
	.dc.b	$29,$18,$35,$AE,$23,$FC,$31,$F7,$7F,$CD,$C9,$EA,$E3,$09,$9F,$F8
	.dc.b	$66,$72,$D8,$FF,$10,$7D,$D3,$EE,$FE,$8C,$7F,$A5,$7E,$1B,$F6,$F2
	.dc.b	$FB,$2C,$FC,$57,$F9,$42,$BF,$BE,$CF,$1C,$3F,$D3,$D7,$C2,$F6,$F2
	.dc.b	$C2,$DF,$09,$BB,$30,$E5,$3E,$13,$58,$74,$69,$99,$E1,$5C,$AB,$8E
	.dc.b	$3E,$59,$CD,$F1,$6D,$F1,$AF,$64,$E5,$E3,$8F,$A6,$3C,$12,$F6,$9F
	.dc.b	$75,$EA,$87,$DC,$BE,$EB,$FC,$2B,$1E,$93,$84,$EE,$75,$BF,$4B,$E7
	.dc.b	$87,$69,$CD,$F1,$9C,$BC,$E3,$3A,$7D,$18,$F0,$E3,$E0,$8C,$4E,$B2
	.dc.b	$FF,$71,$5F,$29,$CD,$6E,$7F,$DC,$D5,$BB,$1D,$31,$F3,$65,$E9,$BE
	.dc.b	$13,$AA,$E0,$EC,$F4,$4E,$EB,$D9,$5C,$27,$58,$7A,$71,$F2,$63,$E2
	.dc.b	$AD,$63,$D0,$EB,$3A,$BF,$04,$E8,$98,$79,$B2,$EE,$C9,$F9,$13,$E1
	.dc.b	$AE,$51,$E2,$C3,$C1,$5B,$C7,$C9,$97,$A6,$B8,$CE,$6B,$E5,$CE,$6B
	.dc.b	$E5,$CE,$6B,$E4,$DF,$6C,$BD,$58,$F9,$A7,$E2,$E5,$B7,$C4,$7B,$2F
	.dc.b	$D4,$F8,$08,$7D,$43,$45,$29,$6B,$F1,$2D,$E4,$FF,$3E,$BB,$3F,$2E
	.dc.b	$E1,$D7,$66,$5C,$B6,$7F,$90,$3F,$95,$9E,$79,$7E,$39,$BA,$7E,$2B
	.dc.b	$C5,$F3,$4E,$6B,$81,$30,$FC,$D4,$EA,$BF,$23,$7C,$A7,$F1,$8D,$7C
	.dc.b	$67,$03,$95,$F5,$7F,$49,$1B,$C1,$7D,$67,$9E,$1C,$30,$EE,$FE,$BE
	.dc.b	$75,$3D,$B0,$E8,$4A,$F6,$47,$13,$99,$E0,$AE,$4E,$4E,$51,$D9,$F4
	.dc.b	$1B,$24,$F5,$7F,$55,$5E,$2B,$F9,$F1,$AA,$E5,$5F,$BC,$AD,$61,$EF
	.dc.b	$26,$1E,$2B,$E0,$79,$AB,$B3,$FC,$5A,$BB,$1C,$A7,$D1,$87,$C8,$9E
	.dc.b	$55,$C2,$3C,$6F,$86,$3F,$95,$7D,$CB,$EE,$7E,$77,$0A,$31,$3A,$9C
	.dc.b	$C3,$D3,$B1,$27,$47,$48,$D9,$97,$F9,$F4,$F4,$6C,$69,$E2,$6A,$70
	.dc.b	$FF,$45,$FB,$9F,$F3,$0F,$81,$FD,$73,$04,$63,$83,$FC,$A6,$DA,$FF
	.dc.b	$6B,$AE,$EF,$D8,$5F,$18,$D5,$F2,$BC,$9C,$8B,$7C,$AE,$CB,$EE,$7F
	.dc.b	$B5,$4F,$F2,$EC,$3C,$90,$F9,$2D,$B6,$75,$87,$F9,$A2,$38,$C2,$B8
	.dc.b	$F4,$70,$B9,$3C,$0F,$F7,$8D,$70,$AF,$85,$F7,$2F,$8C,$F2,$B6,$B8
	.dc.b	$64,$8C,$47,$7C,$BF,$CF,$30,$F8,$0E,$37,$2B,$8C,$F0,$C3,$BD,$7C
	.dc.b	$2F,$B9,$79,$94,$97,$E0,$87,$D9,$E2,$3C,$03,$34,$F2,$8E,$15,$D6
	.dc.b	$BE,$7F,$DC,$FD,$99,$F5,$B5,$E7,$68,$C6,$3F,$85,$7F,$C8,$B3,$C3
	.dc.b	$83,$99,$1E,$7E,$3F,$DE,$5F,$28,$F9,$CC,$36,$3F,$D7,$6D,$7F,$D7
	.dc.b	$4B,$EE,$77,$32,$76,$F2,$8C,$47,$08,$B8,$FF,$AF,$2B,$94,$FD,$8F
	.dc.b	$B9,$ED,$3F,$9F,$6F,$17,$FA,$C1,$B3,$F2,$DE,$2F,$C4,$73,$A1,$A4
	.dc.b	$C7,$F3,$7F,$73,$FE,$BC,$AF,$C3,$AF,$B9,$C9,$AF,$6A,$4E,$A3,$2F
	.dc.b	$08,$F9,$5F,$73,$F6,$E3,$FD,$BC,$DC,$FF,$DB,$2D,$CF,$FB,$13,$F8
	.dc.b	$69,$FF,$73,$B9,$7F,$C3,$2D,$4D,$38,$3D,$B8,$BC,$24,$79,$F1,$7E
	.dc.b	$E3,$6F,$F7,$EC,$7E,$91,$FF,$7E,$B7,$E0,$B5,$F0,$72,$30,$43,$F1
	.dc.b	$F3,$C9,$FC,$8F,$BC,$F7,$FF,$82,$B6,$3F,$62,$D7,$7E,$06,$A6,$46
	.dc.b	$9E,$31,$F9,$3F,$B9,$7D,$57,$CC,$F2,$65,$FF,$03,$13,$A7,$0D,$7F
	.dc.b	$5E,$6A,$7A,$1F,$5B,$BA,$C7,$08,$AF,$D8,$8F,$E3,$FF,$72,$FA,$1B
	.dc.b	$EF,$4A,$FE,$1B,$39,$6C,$E5,$FD,$AC,$7F,$B6,$DC,$9F,$C5,$3E,$E5
	.dc.b	$EF,$FF,$2C,$5E,$FD,$31,$12,$78,$46,$5E,$51,$FB,$9F,$B9,$7D,$CD
	.dc.b	$24,$FC,$78,$FC,$F3,$F1,$1D,$B9,$8D,$B9,$EA,$BD,$5F,$EA,$A8,$57
	.dc.b	$07,$95,$78,$6B,$FA,$2F,$91,$FC,$FE,$7F,$8A,$AF,$8B,$C9,$A7,$FA
	.dc.b	$5F,$DC,$FC,$09,$FE,$6E,$D7,$F8,$EB,$71,$FE,$C0,$BE,$2F,$FA,$9F
	.dc.b	$DC,$BD,$DA,$4B,$FE,$83,$0F,$90,$FE,$B3,$EB,$FF,$3B,$F0,$BF,$D0
	.dc.b	$BE,$3F,$FB,$8F,$9F,$26,$BA,$BF,$AF,$FD,$CB,$EC,$E9,$33,$FE,$9B
	.dc.b	$0F,$89,$FE,$CB,$F7,$2F,$AB,$BE,$FC,$BF,$EA,$73,$97,$FA,$AF,$DC
	.dc.b	$D7,$F9,$DB,$FE,$B5,$73,$3F,$DC,$23,$5C,$9F,$E0,$1F,$B9,$F8,$9B
	.dc.b	$FA,$9B,$7F,$D3,$5B,$B9,$F9,$11,$FD,$93,$EE,$5F,$49,$FD,$F1,$B1
	.dc.b	$98,$C7,$85,$FE,$00,$FB,$9F,$EC,$D3,$FD,$B1,$AF,$F5,$26,$E5,$FE
	.dc.b	$15,$5F,$13,$FE,$08,$FB,$97,$D3,$D2,$5F,$F7,$68,$F9,$CF,$F7,$8F
	.dc.b	$86,$FF,$68,$BF,$D7,$9F,$83,$B7,$D8,$6C,$7F,$6B,$77,$8F,$F8,$97
	.dc.b	$EE,$7F,$81,$37,$C9,$FF,$0C,$C6,$2F,$F0,$3F,$0C,$F9,$EF,$F0,$EB
	.dc.b	$71,$FF,$04,$B5,$FE,$DE,$DC,$3F,$C6,$51,$FD,$38,$7D,$97,$F8,$99
	.dc.b	$C7,$B7,$63,$FC,$57,$2F,$94,$7C,$8E,$3E,$E7,$8E,$1F,$E1,$FA,$46
	.dc.b	$23,$FC,$13,$1C,$67,$9D,$72,$9E,$0D,$76,$D9,$67,$FE,$37,$87,$83
	.dc.b	$C8,$95,$C9,$EB,$F8,$2F,$A3,$EE,$BF,$55,$3E,$56,$BE,$46,$E1,$DA
	.dc.b	$3F,$C7,$91,$F8,$C7,$5E,$27,$32,$3C,$FA,$2B,$A3,$B4,$78,$63,$E4
	.dc.b	$FD,$D7,$DE,$CF,$FC,$A0,$DA,$FF,$25,$37,$03,$F8,$66,$F2,$72,$24
	.dc.b	$FC,$F9,$D7,$DD,$3E,$EB,$EB,$B6,$7E,$A3,$5F,$A6,$BE,$EB,$73,$27
	.dc.b	$6F,$28,$FC,$0F,$DD,$7D,$BC,$7E,$30,$7F,$C3,$6D,$E4,$DF,$24,$C3
	.dc.b	$F0,$95,$A2,$47,$F8,$E7,$E9,$BF,$77,$6A,$7D,$CF,$F3,$3F,$43,$F8
	.dc.b	$73,$7C,$7F,$F7,$93,$FC,$DE,$F8,$D7,$0A,$FF,$7B,$46,$27,$84,$F5
	.dc.b	$6B,$B3,$C1,$1B,$9E,$3C,$3F,$E7,$D9,$D1,$FE,$85,$BF,$0C,$F9,$E3
	.dc.b	$07,$57,$8E,$16,$3A,$D7,$3C,$3C,$24,$AD,$CF,$6C,$3C,$4F,$C3,$8F
	.dc.b	$82,$77,$2B,$D2,$FB,$29,$F8,$6F,$94,$C5,$FA,$DC,$46,$9F,$09,$8F
	.dc.b	$65,$F7,$5C,$63,$E2,$C6,$1C,$4F,$F8,$36,$75,$1E,$19,$D3,$8F,$94
	.dc.b	$CB,$F7,$5A,$9A,$47,$E7,$E0,$FA,$4F,$DA,$67,$FD,$D6,$79,$1A,$8C
	.dc.b	$E7,$E7,$BF,$87,$7C,$79,$7E,$2D,$F0,$C3,$FB,$F4,$E6,$3E,$E9,$9B
	.dc.b	$FF,$67,$E1,$D2,$3E,$56,$1D,$23,$11,$F8,$18,$C6,$1D,$A3,$D8,$79
	.dc.b	$5C,$46,$A3,$B9,$AC,$3C,$75,$DF,$ED,$1F,$95,$3D,$1F,$D7,$E3,$FA
	.dc.b	$59,$CF,$8D,$8C,$51,$C2,$B0,$EA,$BF,$DA,$4E,$5B,$7F,$D7,$9B,$3C
	.dc.b	$23,$15,$FE,$C0,$A7,$E0,$2F,$9E,$EA,$DC,$3A,$AF,$96,$CC,$C9,$F0
	.dc.b	$FE,$59,$FE,$2B,$77,$1E,$68,$C6,$1D,$FF,$02,$7D,$D7,$AB,$6F,$E0
	.dc.b	$98,$7A,$A7,$2F,$B2,$33,$5C,$6B,$83,$F0,$AB,$6F,$AE,$77,$C5,$FC
	.dc.b	$FF,$47,$E6,$7F,$C8,$F1,$F8,$EB,$EB,$3E,$5B,$9C,$3E,$23,$99,$DF
	.dc.b	$6E,$AB,$6D,$8C,$9D,$23,$2D,$3E,$AF,$E5,$9F,$E3,$76,$FB,$5F,$6A
	.dc.b	$3F,$C6,$D1,$82,$61,$F5,$7E,$BF,$EB,$E7,$EB,$AE,$53,$9E,$5F,$C1
	.dc.b	$D6,$AF,$F1,$5F,$91,$FF,$B8,$5F,$C2,$C6,$6B,$F1,$73,$97,$F0,$5F
	.dc.b	$99,$BE,$B7,$F1,$AD,$BF,$C6,$AF,$91,$CC,$FE,$4B,$F1,$67,$BB,$79
	.dc.b	$F0,$DC,$3F,$2B,$E1,$3C,$AF,$CC,$BF,$E5,$38,$CC,$63,$C4,$FC,$EF
	.dc.b	$E4,$5E,$1D,$7F,$2C,$DE,$2F,$B8,$F8,$EF,$F4,$2E,$5F,$D5,$7C,$57
	.dc.b	$B6,$7F,$04,$BF,$1D,$FD,$4B,$99,$DC,$7E,$B3,$EC,$DF,$3F,$7E,$7F
	.dc.b	$0D,$AF,$D7,$FE,$5B,$FE,$7D,$6F,$77,$ED,$23,$1A,$7E,$B5,$FE,$25
	.dc.b	$9F,$E4,$FF,$32,$FA,$5D,$E7,$71,$9E,$AF,$DE,$C6,$31,$FE,$09,$F6
	.dc.b	$9F,$E3,$57,$F7,$ED,$C7,$F7,$5F,$89,$BD,$CF,$E0,$B3,$9C,$7F,$45
	.dc.b	$F4,$6F,$A6,$FE,$12,$DC,$7F,$84,$BF,$C4,$DB,$1F,$C3,$BF,$2D,$7C
	.dc.b	$A5,$FA,$3C,$36,$BF,$88,$FC,$23,$EB,$FF,$8A,$BF,$E3,$A8,$CC,$63
	.dc.b	$E0,$3F,$8C,$FE,$5D,$FE,$4E,$9F,$E3,$2B,$E0,$73,$F0,$7A,$7F,$8F
	.dc.b	$FD,$43,$F0,$8B,$F4,$78,$6E,$1F,$C8,$BF,$2E,$F9,$A4,$F9,$8F,$E4
	.dc.b	$EF,$E8,$78,$DF,$78,$FC,$13,$F9,$74,$63,$F0,$EB,$39,$C3,$FC,$C0
	.dc.b	$A3,$35,$A6,$CE,$9B,$D5,$FE,$64,$AE,$13,$CA,$B8,$3F,$CE,$3E,$BB
	.dc.b	$E6,$B7,$93,$F9,$EC,$39,$6C,$49,$D1,$C6,$31,$87,$F3,$BF,$BB,$FE
	.dc.b	$70,$DF,$33,$FA,$3C,$7F,$39,$AF,$E8,$EE,$B4,$FE,$4C,$D8,$FE,$A4
	.dc.b	$DA,$FE,$A2,$F1,$7A,$47,$F8,$EE,$7F,$DE,$5F,$7D,$FE,$8E,$8C,$1F
	.dc.b	$E2,$97,$12,$9F,$F3,$97,$CC,$3D,$4F,$DC,$4F,$F9,$16,$B7,$5D,$3E
	.dc.b	$1F,$FB,$15,$BF,$30,$F2,$5A,$F7,$8F,$FB,$A2,$73,$18,$95,$71,$CF
	.dc.b	$FB,$64,$71,$C3,$AD,$FB,$EB,$C9,$EA,$78,$33,$7D,$65,$8C,$37,$3F
	.dc.b	$7E,$3D,$AB,$71,$8B,$E6,$D8,$CC,$FF,$71,$FC,$6F,$E7,$79,$06,$9F
	.dc.b	$3F,$6F,$EB,$4E,$35,$FF,$00,$BA,$F8,$1B,$1F,$9A,$FC,$77,$FD,$B7
	.dc.b	$97,$FA,$DD,$FE,$8F,$7F,$E7,$C7,$38,$FF,$AE,$1C,$CF,$DB,$73,$3B
	.dc.b	$F7,$3A,$7C,$C7,$56,$D7,$56,$DF,$37,$FC,$2B,$F6,$9F,$9C,$8F,$F0
	.dc.b	$73,$6F,$FB,$A3,$F9,$DA,$FF,$0C,$B9,$7F,$C4,$4D,$9F,$F1,$0B,$74
	.dc.b	$3C,$63,$AF,$37,$E1,$BC,$9F,$ED,$71,$F8,$FB,$FF,$48,$70,$25,$F6
	.dc.b	$3B,$FA,$1E,$0D,$AF,$B1,$3A,$6C,$E9,$FB,$9E,$6F,$F8,$39,$EA,$FF
	.dc.b	$5D,$73,$87,$E6,$23,$0F,$8C,$FC,$9B,$C5,$BE,$CE,$DE,$0D,$9E,$ED
	.dc.b	$8E,$58,$AB,$AE,$BF,$BA,$F6,$7F,$7F,$ED,$FF,$00,$6E,$3F,$C0,$1B
	.dc.b	$87,$F0,$76,$D7,$D4,$F1,$FE,$CF,$D4,$7D,$A6,$DF,$D6,$6D,$FD,$C6
	.dc.b	$D7,$DC,$F2,$FF,$03,$E5,$7E,$09,$B7,$F7,$9B,$7F,$84,$6D,$7E,$13
	.dc.b	$D7,$FC,$49,$F8,$AF,$36,$DF,$E1,$A2,$DF,$E3,$26,$D7,$E3,$2B,$F2
	.dc.b	$3E,$BF,$E4,$9A,$7E,$38,$B5,$F9,$46,$CF,$EB,$BE,$BD,$1F,$E9,$BF
	.dc.b	$8B,$FF,$3A,$77,$27,$37,$FE,$9D,$9C,$FD,$BE,$3E,$3B,$FC,$F9,$3F
	.dc.b	$E7,$8B,$E1,$87,$38,$C6,$1B,$C3,$AB,$63,$86,$17,$AF,$F8,$24,$DC
	.dc.b	$39,$D7,$84,$ED,$F6,$19,$9D,$33,$1F,$05,$FF,$0D,$8D,$46,$2D,$1E
	.dc.b	$1B,$FF,$1D,$FD,$DD,$FB,$AF,$EF,$9D,$3E,$99,$CE,$16,$73,$1E,$33
	.dc.b	$25,$73,$AF,$C4,$8C,$B6,$33,$AF,$FB,$BD,$B2,$0E,$1E,$31,$BE,$0F
	.dc.b	$F2,$64,$77,$6D,$75,$6C,$FF,$BE,$F9,$D9,$7B,$67,$37,$DF,$0E,$C7
	.dc.b	$7A,$DE,$3E,$FA,$DD,$79,$1B,$5B,$BE,$99,$6C,$95,$F1,$70,$E6,$F9
	.dc.b	$E3,$39,$79,$DF,$9D,$3A,$8C,$4E,$C8,$FC,$B8,$D6,$5F,$0E,$B9,$B6
	.dc.b	$BC,$17,$E4,$78,$C6,$6F,$94,$E5,$EE,$48,$F9,$11,$97,$81,$27,$EB
	.dc.b	$5F,$89,$F9,$58,$FC,$42,$61,$EC,$BE,$53,$A8,$FB,$CC,$3E,$63,$5D
	.dc.b	$17,$1F,$C7,$BB,$5C,$7F,$E6,$18,$D3,$5D,$35,$FC,$73,$5F,$C0,$3C
	.dc.b	$72,$FC,$A3,$72,$DF,$B5,$F9,$FF,$BB,$F3,$D6,$A3,$F4,$55,$BB,$E7
	.dc.b	$1B,$72,$C7,$F4,$91,$B7,$F3,$B5,$D2,$FB,$C6,$AB,$85,$FE,$9E,$B5
	.dc.b	$3D,$8D,$12,$34,$F0,$AE,$71,$E7,$8F,$DA,$4F,$27,$F5,$B7,$E3,$C3
	.dc.b	$E5,$E1,$C6,$7F,$49,$5B,$9C,$C7,$7C,$3A,$9F,$B9,$9D,$5F,$32,$5F
	.dc.b	$C3,$AE,$07,$A7,$0F,$21,$E8,$72,$79,$23,$C4,$FC,$62,$E1,$E2,$9E
	.dc.b	$38,$78,$67,$B4,$61,$ED,$19,$7E,$09,$CA,$FE,$06,$1F,$33,$83,$FC
	.dc.b	$65,$E3,$DE,$1F,$F2,$0D,$FF,$27,$8C,$7C,$34,$8C,$10,$FE,$69,$5F
	.dc.b	$E3,$B8,$D4,$62,$BF,$DD,$CF,$F3,$18,$ED,$E4,$EB,$3F,$E9,$37,$C6
	.dc.b	$B4,$7F,$37,$9F,$27,$DC,$7E,$A3,$9F,$FC,$8B,$F0,$77,$97,$FA,$BB
	.dc.b	$6F,$FA,$95,$5C,$3F,$C9,$CD,$CF,$FD,$C9,$3F,$E5,$A9,$FF,$29,$56
	.dc.b	$6B,$5E,$6C,$AD,$5A,$39,$7C,$CE,$67,$FC,$B9,$17,$2F,$F2,$CC,$F7
	.dc.b	$F0,$7C,$1C,$D2,$F9,$C6,$5E,$B1,$97,$FC,$CD,$1E,$7F,$63,$D1,$EC
	.dc.b	$C9,$F1,$C6,$5F,$2C,$65,$FF,$37,$47,$C1,$FC,$BF,$FA,$96,$3F,$DE
	.dc.b	$B6,$43,$E0,$B4,$6F,$9E,$61,$F1,$67,$4D,$9C,$BF,$E3,$D1,$8C,$7F
	.dc.b	$D0,$A6,$32,$FA,$8D,$33,$1B,$79,$37,$1F,$F0,$7C,$6C,$CC,$F2,$9D
	.dc.b	$70,$FF,$2A,$6C,$FF,$69,$7F,$C8,$CF,$69,$E1,$1F,$85,$F1,$9F,$73
	.dc.b	$FE,$FE,$AF,$F4,$C3,$FE,$D1,$6F,$AD,$B8,$D4,$7F,$94,$9B,$8F,$FA
	.dc.b	$F1,$E1,$19,$9E,$1E,$67,$E4,$7D,$C7,$2D,$25,$7E,$46,$3F,$26,$FF
	.dc.b	$99,$F8,$11,$C2,$35,$3C,$1D,$47,$0C,$7E,$3F,$DC,$BE,$A3,$E7,$AF
	.dc.b	$9E,$A6,$22,$5F,$D6,$9D,$36,$72,$FD,$D8,$E8,$E4,$86,$67,$F5,$DF
	.dc.b	$4A,$EF,$A4,$9F,$D7,$43,$F0,$2C,$47,$BE,$38,$3A,$9E,$73,$F5,$FE
	.dc.b	$DD,$F3,$D7,$09,$7F,$EC,$F9,$CB,$67,$2F,$FA,$22,$3D,$2D,$8F,$C9
	.dc.b	$FD,$C3,$EC,$FF,$2B,$F2,$0F,$C0,$29,$A8,$97,$F9,$C9,$D3,$67,$2F
	.dc.b	$E8,$EB,$A3,$FA,$26,$B9,$AE,$FF,$15,$FB,$5F,$B3,$93,$FA,$F7,$F7
	.dc.b	$1F,$1B,$88,$FC,$27,$72,$71,$C7,$FB,$B8,$57,$F1,$FE,$3F,$F0,$64
	.dc.b	$FE,$FD,$BD,$BF,$E4,$DE,$7F,$BF,$8F,$F2,$19,$FF,$25,$BF,$91,$76
	.dc.b	$A3,$6E,$4E,$DE,$6F,$EF,$3F,$2C,$78,$97,$AA,$57,$F0,$29,$CB,$67
	.dc.b	$2F,$F0,$98,$FC,$F3,$63,$A1,$AC,$BF,$81,$CE,$9B,$1F,$CF,$7F,$03
	.dc.b	$7A,$B4,$97,$FC,$F6,$1E,$B7,$F4,$3F,$1D,$F5,$57,$E5,$E5,$FF,$43
	.dc.b	$9C,$BF,$D1,$7F,$20,$7C,$1F,$F4,$B5,$EF,$D3,$11,$FD,$3F,$F2,$ED
	.dc.b	$D7,$F4,$D7,$6D,$9D,$4F,$F5,$5E,$F7,$9B,$7C,$A9,$7F,$D5,$67,$2F
	.dc.b	$F5,$7F,$CB,$BE,$5B,$FC,$1D,$BC,$5F,$CF,$5B,$D1,$F7,$DF,$90,$FE
	.dc.b	$07,$CA,$FE,$B5,$5F,$D8,$5A,$FF,$11,$6D,$7F,$47,$79,$C7,$E0,$FF
	.dc.b	$3D,$7B,$DF,$DB,$7C,$CF,$E6,$F7,$FD,$35,$B8,$7E,$3E,$B7,$7F,$DB
	.dc.b	$FC,$5F,$F4,$9B,$7A,$3F,$BB,$47,$6E,$CF,$DF,$F2,$7F,$79,$DF,$F7
	.dc.b	$B7,$FC,$95,$FC,$CE,$7F,$D6,$DB,$97,$E7,$5D,$C7,$F5,$FE,$AB,$3F
	.dc.b	$A1,$7F,$81,$C6,$78,$3F,$C0,$D1,$AB,$FD,$07,$E0,$1F,$CC,$9F,$F0
	.dc.b	$33,$5F,$E6,$CD,$AF,$ED,$8B,$E6,$BF,$C2,$1F,$9F,$7F,$1F,$AF,$F0
	.dc.b	$72,$FB,$2C,$9A,$F2,$FF,$C2,$DC,$CF,$C3,$6F,$D2,$CC,$3F,$C2,$D3
	.dc.b	$97,$FC,$31,$F2,$3F,$01,$FD,$B3,$78,$8D,$38,$47,$F4,$DF,$3B,$C4
	.dc.b	$FF,$88,$7E,$8D,$F0,$9F,$E2,$76,$C7,$F8,$96,$76,$FF,$55,$FC,$C3
	.dc.b	$FB,$0F,$27,$F0,$B9,$CF,$17,$F8,$35,$B1,$FC,$3E,$77,$5F,$C2,$DB
	.dc.b	$91,$A4,$F8,$AF,$F1,$C7,$C5,$68,$FE,$DB,$8F,$F8,$65,$B8,$7F,$1B
	.dc.b	$75,$EC,$70,$ED,$99,$7F,$8F,$67,$2D,$8C,$BF,$DB,$E3,$C7,$F6,$1F
	.dc.b	$CD,$B0,$FF,$13,$3F,$D5,$27,$FB,$A3,$63,$3E,$97,$F9,$36,$38,$63
	.dc.b	$FC,$B5,$D3,$7D,$6D,$4F,$F9,$4F,$E2,$DF,$80,$FF,$95,$7E,$C3,$F8
	.dc.b	$2E,$5F,$E0,$16,$E5,$FE,$37,$AD,$CF,$F9,$5D,$AF,$F8,$1D,$B1,$BA
	.dc.b	$ED,$F8,$17,$F9,$87,$F3,$0F,$E4,$37,$FE,$5F,$5F,$1B,$93,$5E,$84
	.dc.b	$9D,$46,$5E,$11,$FE,$3A,$D9,$E7,$B2,$53,$EC,$9A,$FF,$B7,$87,$C8
	.dc.b	$FF,$B8,$23,$33,$C2,$72,$F2,$8C,$AC,$3B,$DE,$6B,$FD,$D3,$18,$9F
	.dc.b	$04,$77,$BE,$CD,$8E,$07,$9A,$78,$5F,$66,$CE,$E7,$99,$E5,$AE,$6F
	.dc.b	$1C,$3E,$03,$A3,$CC,$E2,$FD,$35,$EC,$9D,$36,$3C,$13,$B8,$F5,$B6
	.dc.b	$7C,$07,$A0,$F9,$75,$E2,$7B,$E1,$9C,$3E,$14,$7D,$23,$CD,$15,$EA
	.dc.b	$D7,$25,$8F,$F1,$6F,$D9,$E6,$3E,$87,$92,$8C,$4E,$A7,$84,$E5,$E0
	.dc.b	$3E,$2E,$B4,$CB,$82,$3F,$26,$FF,$94,$72,$7B,$E3,$F1,$73,$F0,$67
	.dc.b	$31,$F3,$CC,$9C,$7C,$55,$71,$8F,$06,$BF,$8A,$AE,$2F,$E3,$23,$8B
	.dc.b	$9A,$E3,$1D,$4E,$3A,$FF,$B3,$E3,$33,$C6,$72,$7E,$82,$BA,$D7,$CD
	.dc.b	$C3,$D1,$7C,$23,$15,$E7,$C3,$CD,$5A,$7F,$2B,$D6,$FB,$37,$C4,$79
	.dc.b	$32,$FF,$5C,$47,$FA,$DE,$BF,$DB,$0E,$9B,$F1,$1B,$7F,$C7,$D1,$9C
	.dc.b	$7F,$D1,$8F,$16,$DE,$DF,$F1,$F4,$66,$FF,$D8,$DF,$66,$F9,$AD,$64
	.dc.b	$8C,$63,$FE,$E9,$5F,$37,$A7,$38,$FF,$A6,$DB,$87,$4C,$35,$5B,$7A
	.dc.b	$FD,$8B,$EC,$BB,$F0,$7F,$0C,$CB,$FD,$ED,$39,$7F,$DF,$11,$9C,$BF
	.dc.b	$D5,$13,$97,$FD,$55,$19,$C7,$FD,$C1,$39,$7F,$DC,$51,$9A,$FF,$32
	.dc.b	$7D,$93,$CA,$F7,$AF,$B2,$B1,$86,$CF,$E0,$4E,$2E,$31,$FF,$5D,$57
	.dc.b	$19,$D4,$7F,$9A,$E3,$B4,$63,$64,$6E,$78,$DF,$7A,$D7,$D9,$7F,$94
	.dc.b	$1B,$F3,$FF,$52,$31,$5F,$EC,$95,$19,$3E,$CF,$D0,$FD,$DA,$FF,$41
	.dc.b	$4E,$9F,$14,$66,$BF,$D0,$0D,$C5,$E5,$FF,$B6,$AB,$7C,$BF,$84,$8C
	.dc.b	$5F,$FB,$67,$EC,$9F,$31,$F8,$86,$DF,$BD,$7C,$C6,$67,$F1,$BF,$65
	.dc.b	$FE,$12,$AF,$88,$FE,$21,$FF,$4C,$39,$BF,$F7,$33,$97,$FD,$2D,$E0
	.dc.b	$BE,$27,$F2,$ED,$AE,$7C,$F8,$9F,$CD,$FD,$97,$92,$FF,$DE,$0D,$E6
	.dc.b	$FE,$07,$BA,$3C,$F8,$7D,$77,$4F,$FA,$AF,$E4,$BE,$0D,$70,$5F,$35
	.dc.b	$FA,$87,$33,$B8,$FD,$5F,$D9,$7F,$A6,$23,$F0,$4D,$C7,$FB,$64,$7F
	.dc.b	$AD,$E3,$FB,$BB,$AF,$8C,$FD,$9F,$D9,$3E,$62,$98,$F8,$DB,$5F,$8D
	.dc.b	$FB,$2F,$B4,$6D,$C7,$3F,$EF,$23,$19,$78,$0F,$F6,$07,$D2,$7F,$95
	.dc.b	$EB,$F2,$CD,$8D,$70,$3E,$71,$F4,$5F,$C1,$AB,$5C,$3F,$2B,$EC,$BF
	.dc.b	$DA,$B8,$FE,$7D,$B8,$F2,$7F,$8B,$4E,$FE,$5B,$F8,$7F,$D9,$3D,$CB
	.dc.b	$EE,$61,$B8,$FE,$A7,$E6,$BE,$2E,$5E,$46,$E3,$F6,$97,$B5,$FC,$69
	.dc.b	$CD,$7F,$B8,$FE,$CB,$2F,$EC,$DB,$7E,$E5,$EF,$FF,$22,$73,$87,$E1
	.dc.b	$BE,$CB,$FD,$73,$7F,$BB,$5F,$1B,$FC,$A1,$CE,$1F,$21,$CC,$FF,$01
	.dc.b	$FB,$2F,$D7,$13,$A3,$F9,$7B,$FB,$3C,$1F,$62,$FB,$1F,$E7,$11,$8F
	.dc.b	$B2,$B3,$9C,$3F,$87,$CE,$5B,$19,$CB,$EF,$4E,$5B,$19,$8F,$EA,$0A
	.dc.b	$BA,$DF,$66,$DF,$47,$FA,$2F,$D9,$7D,$78,$FE,$38,$DC,$3F,$0E,$B8
	.dc.b	$B7,$C5,$2B,$6F,$28,$C5,$7E,$3B,$EC,$BF,$7D,$D4,$69,$FD,$42,$FF
	.dc.b	$95,$38,$F1,$7F,$AB,$BA,$F5,$6C,$7F,$5F,$B7,$D9,$3E,$A7,$0D,$C3
	.dc.b	$F9,$BF,$D4,$BD,$0F,$DF,$DF,$F9,$7A,$B4,$FE,$9C,$E7,$7C,$8F,$ED
	.dc.b	$95,$C7,$CD,$C6,$1F,$D0,$AB,$6F,$6F,$8E,$FE,$35,$18,$3F,$65,$5F
	.dc.b	$EB,$AC,$B9,$E5,$FC,$4A,$72,$BE,$AF,$3E,$86,$9B,$8F,$AA,$74,$D9
	.dc.b	$D3,$FD,$47,$EA,$FE,$69,$36,$FE,$FE,$E3,$1A,$62,$2C,$FE,$3B,$EC
	.dc.b	$9F,$41,$F9,$47,$FA,$66,$7F,$8E,$73,$3B,$F5,$F5,$5B,$6C,$64,$E9
	.dc.b	$19,$69,$9A,$FE,$23,$F6,$5F,$8D,$EB,$FE,$6E,$D9,$FF,$49,$F2,$A3
	.dc.b	$F2,$11,$FD,$BD,$D5,$27,$5C,$FF,$E1,$E9,$CC,$71,$E7,$C4,$FD,$CF
	.dc.b	$1F,$FB,$7E,$3F,$E4,$8E,$34,$3E,$CB,$5F,$47,$B6,$E9,$F8,$C9,$D3
	.dc.b	$67,$4F,$E4,$3C,$BF,$F0,$A6,$AC,$3F,$97,$7A,$BF,$8C,$9C,$C6,$2B
	.dc.b	$85,$66,$B4,$D8,$C9,$D6,$32,$D3,$F5,$3F,$15,$FA,$8F,$95,$FD,$4B
	.dc.b	$63,$27,$EB,$A3,$2D,$3E,$67,$C9,$FC,$9F,$27,$E5,$36,$32,$7D,$18
	.dc.b	$CB,$4F,$A9,$F5,$FF,$47,$CC,$FA,$4D,$8C,$9F,$66,$32,$D3,$EE,$7C
	.dc.b	$77,$D9,$3E,$D7,$C7,$D8,$C9,$34,$FC,$4C,$65,$FC,$5C,$7E,$07,$D7
	.dc.b	$FC,$13,$4F,$C4,$16,$36,$FE,$2E,$3F,$4F,$F6,$5F,$E8,$AF,$8B,$FF
	.dc.b	$3C,$D7,$E3,$7B,$D3,$9F,$8B,$FC,$AC,$3E,$8F,$15,$BA,$BD,$6A,$E3
	.dc.b	$97,$FA,$93,$0E,$35,$A9,$E5,$39,$BE,$4D,$BE,$55,$BF,$B2,$E0,$D9
	.dc.b	$D5,$F6,$BF,$F5,$D6,$1E,$6B,$DD,$69,$F3,$CB,$89,$F4,$58,$FB,$26
	.dc.b	$FF,$E3,$3A,$E6,$0E,$13,$B8,$D6,$DF,$22,$73,$8F,$AD,$5F,$4E,$6C
	.dc.b	$63,$E0,$AD,$4E,$6B,$B9,$C3,$0E,$B9,$FF,$B9,$67,$58,$F7,$C3,$95
	.dc.b	$F0,$9C,$D6,$EF,$8C,$FA,$67,$2F,$AE,$B8,$DF,$A6,$73,$3E,$1C,$7C
	.dc.b	$B7,$CA,$3D,$A7,$77,$13,$C0,$E3,$7D,$30,$F1,$5F,$22,$5F,$BA,$74
	.dc.b	$F9,$6F,$99,$27,$E3,$E5,$F3,$DE,$38,$F9,$E7,$C9,$7C,$AF,$D0,$79
	.dc.b	$6B,$78,$7D,$98,$F4,$DF,$19,$CC,$78,$5E,$F9,$62,$39,$CE,$DF,$46
	.dc.b	$3E,$37,$B1,$23,$EA,$46,$5E,$04,$BF,$16,$1E,$7C,$BE,$46,$5F,$8F
	.dc.b	$C3,$CC,$7A,$8B,$1F,$85,$6C,$6D,$FC,$27,$1F,$F9,$69,$E2,$7E,$41
	.dc.b	$C3,$C9,$AF,$0F,$6B,$6E,$DF,$C9,$C7,$3C,$7F,$08,$DC,$7F,$25,$F3
	.dc.b	$3F,$2E,$73,$5C,$09,$7F,$93,$9D,$57,$78,$FC,$B5,$F3,$7F,$28,$D7
	.dc.b	$F4,$F3,$C1,$CD,$74,$BE,$F1,$D4,$C5,$F5,$8C,$9F,$9B,$AE,$53,$ED
	.dc.b	$3C,$35,$C1,$FD,$24,$6B,$16,$1E,$36,$D6,$EB,$CA,$58,$CB,$D0,$F2
	.dc.b	$16,$FC,$78,$79,$4F,$93,$1C,$6F,$B1,$DF,$0F,$13,$8C,$3B,$46,$1E
	.dc.b	$07,$F0,$87,$D0,$66,$7C,$15,$D9,$F1,$DF,$89,$F9,$66,$23,$F9,$DF
	.dc.b	$97,$F3,$F8,$BF,$9C,$79,$57,$8B,$FC,$6D,$7F,$CB,$AB,$06,$66,$46
	.dc.b	$48,$7F,$28,$AF,$E8,$53,$C2,$3F,$9B,$F9,$35,$D5,$FF,$BC,$1E,$38
	.dc.b	$FF,$52,$BF,$E9,$55,$E3,$FA,$BF,$DD,$78,$5F,$ED,$EC,$3F,$DA,$F8
	.dc.b	$7F,$90,$FE,$0B,$3C,$9F,$D5,$AE,$CF,$F5,$B6,$E3,$FD,$56,$AD,$F2
	.dc.b	$AF,$F6,$2A,$AC,$F9,$D2,$7F,$CA,$0F,$0F,$A5,$C4,$70,$8B,$97,$96
	.dc.b	$7A,$F8,$5D,$B9,$E5,$FF,$96,$63,$2F,$F9,$6E,$32,$F8,$A3,$CB,$F3
	.dc.b	$DE,$6F,$6A,$4F,$F9,$86,$32,$FF,$98,$E3,$2F,$AA,$3D,$FF,$5B,$FD
	.dc.b	$57,$16,$FF,$E4,$E8,$7C,$16,$8D,$F2,$CC,$3F,$B0,$5F,$07,$FC,$A1
	.dc.b	$5C,$1C,$DF,$3B,$E1,$1F,$E8,$DC,$7F,$D9,$6F,$FA,$25,$FF,$78,$37
	.dc.b	$2F,$AA,$FF,$B4,$22,$9C,$1E,$51,$83,$31,$23,$91,$0F,$03,$F8,$0F
	.dc.b	$19,$F5,$1F,$EF,$39,$FF,$52,$BF,$EB,$B6,$F9,$9B,$8D,$47,$FA,$69
	.dc.b	$B9,$FF,$99,$DE,$11,$99,$E1,$E6,$7E,$33,$DC,$FC,$44,$FC,$43,$FC
	.dc.b	$EC,$E1,$5F,$CB,$E6,$4E,$63,$0F,$E4,$DB,$C9,$D4,$95,$D6,$B8,$F3
	.dc.b	$7C,$AF,$B2,$7D,$45,$FA,$E9,$7F,$EB,$99,$CB,$67,$2F,$FB,$0E,$13
	.dc.b	$CE,$64,$F1,$8D,$E3,$F8,$38,$57,$47,$DB,$C5,$FE,$9A,$FB,$27,$7F
	.dc.b	$EC,$1B,$3E,$96,$BC,$0F,$F4,$E2,$FA,$6E,$13,$27,$83,$CE,$3E,$F7
	.dc.b	$D2,$BE,$7A,$E9,$2F,$FD,$C7,$39,$6C,$E5,$FF,$6B,$C7,$A7,$3F,$F5
	.dc.b	$23,$B8,$FF,$73,$7D,$8B,$E9,$7F,$2F,$F2,$0F,$B2,$A7,$E6,$A7,$55
	.dc.b	$FE,$E7,$6B,$CA,$B4,$72,$F8,$AB,$1B,$8A,$FF,$BD,$6D,$3F,$BA,$FB
	.dc.b	$29,$3F,$81,$EE,$47,$EF,$BE,$37,$33,$F2,$F9,$95,$A8,$E1,$5B,$7F
	.dc.b	$15,$CB,$FE,$10,$DB,$F2,$8D,$E4,$FF,$69,$B6,$7F,$DA,$0F,$F8,$5E
	.dc.b	$77,$1F,$CB,$23,$58,$7E,$2A,$31,$E7,$FE,$FB,$EB,$5E,$2D,$EA,$95
	.dc.b	$FA,$71,$3A,$70,$D7,$F0,$C6,$A3,$8B,$88,$D6,$3F,$95,$6E,$5F,$E4
	.dc.b	$17,$F5,$CD,$AF,$25,$FF,$44,$FB,$27,$AB,$49,$8F,$EB,$E1,$EB,$7F
	.dc.b	$4A,$F2,$1F,$25,$7E,$5A,$5F,$ED,$67,$2F,$ED,$FE,$C9,$F0,$5F,$D4
	.dc.b	$9B,$19,$8C,$7C,$1F,$F5,$6F,$B2,$CC,$FD,$B6,$F0,$FF,$09,$6F,$07
	.dc.b	$FA,$57,$8D,$0B,$FE,$BF,$DE,$F3,$AF,$97,$32,$FE,$15,$39,$7F,$86
	.dc.b	$7D,$97,$C8,$58,$A3,$0D,$E4,$FE,$D7,$1F,$32,$77,$E5,$BD,$0F,$ED
	.dc.b	$ED,$8F,$E3,$0D,$7F,$2C,$F4,$FB,$2F,$F6,$64,$7F,$1D,$F3,$7F,$87
	.dc.b	$70,$7F,$27,$6F,$07,$BE,$B7,$5F,$9F,$F1,$BF,$D7,$6F,$F2,$56,$B9
	.dc.b	$E4,$FF,$00,$C7,$87,$0F,$E6,$5B,$FE,$FD,$B1,$FE,$06,$8C,$61,$F3
	.dc.b	$5F,$E0,$CF,$CF,$EB,$CE,$1F,$19,$CB,$78,$3F,$C1,$F1,$AE,$07,$D9
	.dc.b	$7F,$40,$AF,$E7,$4D,$7F,$70,$DC,$BF,$A3,$2F,$A3,$FF,$0C,$FD,$97
	.dc.b	$F2,$5A,$FE,$84,$D7,$F8,$03,$6B,$FA,$77,$23,$E8,$BF,$C4,$2B,$EC
	.dc.b	$6C,$4B,$FE,$99,$39,$7F,$A7,$7C,$87,$EE,$DF,$E1,$8D,$BF,$EB,$4D
	.dc.b	$DB,$F2,$4F,$19,$FE,$25,$E6,$3E,$CB,$FC,$63,$F4,$7F,$EF,$19,$FE
	.dc.b	$B4,$D7,$71,$FE,$36,$FB,$2F,$EE,$38,$7E,$76,$F9,$57,$F6,$16,$BF
	.dc.b	$C7,$E7,$6F,$E6,$17,$C4,$F3,$F5,$E4,$FF,$69,$9C,$BF,$DA,$FE,$2B
	.dc.b	$8F,$19,$FE,$FF,$AF,$F7,$76,$E5,$FA,$F7,$64,$EA,$FF,$26,$BA,$BF
	.dc.b	$D4,$BA,$C7,$33,$27,$C1,$1F,$E5,$4F,$AE,$FE,$8F,$87,$F8,$15,$FE
	.dc.b	$CA,$DF,$57,$5E,$94,$BF,$F0,$15,$6B,$2F,$E0,$4D,$8F,$DC,$BD,$23
	.dc.b	$F8,$37,$C5,$BE,$CB,$FC,$C6,$DA,$FF,$06,$35,$FE,$94,$BE,$CB,$73
	.dc.b	$27,$6F,$28,$FF,$36,$FD,$97,$F2,$69,$FF,$0B,$2F,$8A,$C9,$AF,$36
	.dc.b	$4E,$A3,$2F,$08,$FF,$22,$E4,$F3,$3F,$C5,$32,$FB,$17,$D8,$FF,$BF
	.dc.b	$47,$C8,$FF,$BF,$E3,$32,$37,$19,$9C,$42,$FB,$70,$EE,$F8,$CE,$70
	.dc.b	$F0,$C6,$2F,$BC,$ED,$F1,$61,$D6,$3C,$18,$74,$7D,$13,$E0,$8C,$57
	.dc.b	$3A,$E1,$5E,$63,$A1,$CA,$FC,$37,$C5,$B1,$ED,$8F,$1B,$D8,$C4,$F4
	.dc.b	$BE,$F3,$C7,$0E,$B5,$DC,$E0,$5C,$3E,$0D,$70,$8F,$14,$FC,$87,$91
	.dc.b	$4F,$8C,$D6,$99,$8F,$F1,$B7,$D9,$65,$B1,$9F,$19,$39,$6C,$73,$1F
	.dc.b	$13,$5A,$65,$C1,$1F,$5D,$FF,$29,$E2,$F7,$C7,$E0,$E7,$E0,$CE,$63
	.dc.b	$E0,$99,$38,$F8,$7F,$EA,$1C,$BF,$07,$19,$7F,$0B,$3C,$5C,$C7,$91
	.dc.b	$CC,$F1,$CF,$E3,$5F,$82,$31,$5F,$2E,$31,$87,$9A,$39,$57,$FB,$82
	.dc.b	$B7,$1F,$98,$AD,$BE,$83,$67,$FA,$E2,$37,$87,$B4,$FE,$D6,$FF,$AE
	.dc.b	$DF,$F3,$D4,$7A,$EB,$B5,$FF,$8C,$E7,$83,$FE,$F0,$F4,$B8,$3A,$AF
	.dc.b	$F1,$9C,$FF,$88,$A3,$FC,$75,$F0,$9F,$B2,$6E,$DC,$2B,$F6,$53,$FB
	.dc.b	$39,$C9,$23,$FD,$05,$4F,$8A,$F0,$C7,$0B,$E5,$39,$CB,$A4,$B0,$7C
	.dc.b	$36,$BF,$3E,$7C,$96,$A7,$39,$FF,$AD,$71,$56,$A3,$D3,$7C,$6F,$83
	.dc.b	$63,$DB,$F0,$FC,$CF,$1F,$E5,$DF,$E1,$F8,$C9,$32,$F5,$C7,$F9,$82
	.dc.b	$31,$8F,$D1,$8C,$4F,$F9,$73,$25,$F4,$F5,$7B,$61,$E7,$BD,$E1,$FE
	.dc.b	$B1,$F8,$FF,$F5,$34,$FF,$9A,$5B,$8F,$F9,$9B,$F0,$9F,$F4,$B4,$E7
	.dc.b	$E2,$3E,$3D,$6A,$FF,$D7,$5E,$2F,$FB,$22,$B4,$7F,$9C,$DC,$61,$FE
	.dc.b	$70,$6C,$FF,$B0,$70,$7D,$17,$D2,$9C,$DF,$F2,$C9,$CB,$CF,$CE,$FA
	.dc.b	$FF,$97,$3E,$7B,$84,$61,$93,$FE,$83,$9F,$AF,$97,$FA,$01,$E1,$19
	.dc.b	$E1,$E7,$CD,$F8,$2F,$A3,$7C,$86,$2B,$FD,$19,$18,$7F,$D1,$DF,$7B
	.dc.b	$FE,$D4,$7F,$05,$EF,$FD,$9C,$BF,$D1,$B3,$97,$F2,$71,$9E,$0F,$F7
	.dc.b	$05,$7F,$B3,$CF,$C5,$7E,$7F,$FE,$84,$7F,$D3,$46,$6B,$F3,$71,$97
	.dc.b	$FD,$33,$F9,$47,$E8,$5C,$E3,$F9,$A8,$C5,$FE,$D6,$93,$9F,$9A,$FD
	.dc.b	$0F,$92,$F9,$9C,$65,$F9,$B8,$C3,$FE,$AB,$C7,$F1,$6E,$63,$B7,$57
	.dc.b	$27,$F5,$3E,$13,$CC,$FD,$6F,$E5,$FF,$BB,$DB,$FD,$5F,$39,$C3,$F5
	.dc.b	$13,$9D,$91,$FC,$4F,$5F,$DA,$FE,$81,$F2,$67,$38,$7F,$2F,$9C,$FA
	.dc.b	$1F,$B8,$FD,$09,$F3,$1F,$BC,$F9,$BF,$91,$87,$0C,$15,$AC,$BF,$71
	.dc.b	$5A,$7F,$BC,$C3,$E8,$3F,$82,$FE,$85,$F2,$A7,$31,$8B,$F0,$7D,$DF
	.dc.b	$F0,$AF,$CC,$9F,$29,$8B,$FE,$09,$18,$7F,$DA,$9F,$28,$F8,$AF,$E2
	.dc.b	$38,$FF,$1D,$C7,$FD,$A7,$39,$AF,$E1,$93,$9D,$7D,$5F,$99,$7F,$B3
	.dc.b	$27,$3F,$2B,$FC,$62,$31,$87,$F2,$0B,$4F,$F1,$CF,$C2,$3F,$CE,$D7
	.dc.b	$FE,$E5,$6D,$7F,$26,$FA,$F7,$CF,$EF,$E2,$A5,$FF,$1D,$9F,$E0,$35
	.dc.b	$FE,$E8,$BE,$39,$FF,$41,$9C,$BF,$3E,$33,$5F,$E2,$78,$4E,$A3,$FC
	.dc.b	$4F,$3A,$A5,$6B,$E7,$FF,$9A,$FD,$57,$F0,$29,$CB,$64,$FC,$D8,$FE
	.dc.b	$7E,$E4,$9F,$21,$CB,$8B,$9B,$FD,$03,$2F,$E1,$75,$F0,$7A,$71,$E7
	.dc.b	$65,$BA,$7C,$16,$CE,$E3,$D8,$7F,$90,$3E,$B5,$F5,$9F,$D2,$9B,$1F
	.dc.b	$D1,$73,$F7,$52,$72,$FC,$09,$E5,$5F,$D5,$3F,$35,$FC,$14,$E4,$FF
	.dc.b	$28,$57,$0D,$7F,$B8,$4E,$9F,$CF,$46,$6F,$E8,$4E,$5E,$A7,$F5,$77
	.dc.b	$15,$F4,$33,$FF,$3D,$4F,$DF,$F8,$5F,$8D,$E1,$FC,$76,$46,$4F,$C2
	.dc.b	$46,$2B,$E8,$FA,$1F,$63,$EF,$BE,$CB,$6B,$E1,$36,$77,$1F,$16,$7E
	.dc.b	$47,$B3,$23,$F2,$1F,$0F,$6B,$E1,$1F,$26,$7F,$BC,$FC,$9A,$F5,$B7
	.dc.b	$F8,$4E,$31,$87,$D6,$BF,$93,$7C,$2F,$73,$F9,$5A,$DD,$7E,$2B,$E6
	.dc.b	$FF,$23,$E7,$FE,$49,$B5,$F6,$DB,$3B,$8F,$BD,$3F,$8D,$F1,$A4,$7E
	.dc.b	$A7,$D9,$B5,$FA,$F9,$B9,$FE,$A6,$6D,$7F,$A2,$BF,$1D,$FD,$9B,$E9
	.dc.b	$B4,$61,$FC,$FF,$7F,$F8,$62,$31,$7F,$C7,$DB,$5F,$82,$F9,$AD,$77
	.dc.b	$FE,$35,$B1,$F8,$36,$E1,$FE,$28,$E5,$FE,$AD,$3F,$B9,$E5,$FF,$78
	.dc.b	$3F,$C0,$FE,$4B,$FC,$53,$EB,$3D,$CF,$D6,$44,$9D,$1F,$E2,$59,$D3
	.dc.b	$8C,$3F,$6F,$F5,$DF,$BE,$3F,$7F,$F8,$9D,$9F,$D4,$45,$9F,$DE,$CD
	.dc.b	$DB,$F9,$E7,$E6,$3F,$CE,$E7,$F0,$AF,$F4,$BB,$FF,$59,$B9,$9D,$FC
	.dc.b	$37,$F8,$FA,$31,$3A,$8C,$7A,$19,$75,$E1,$C9,$FA,$55,$FE,$C9,$6C
	.dc.b	$67,$0F,$DB,$7C,$53,$A7,$FA,$F4,$79,$27,$FC,$B1,$39,$F5,$7C,$5D
	.dc.b	$3E,$3D,$AE,$3D,$9B,$87,$B1,$BA,$FC,$07,$FC,$AF,$EC,$FF,$49,$F5
	.dc.b	$9F,$55,$FE,$D2,$8F,$E1,$B1,$8B,$FF,$6C,$46,$E7,$5E,$2F,$F6,$CA
	.dc.b	$D5,$2B,$4F,$09,$D5,$75,$F8,$8B,$39,$CB,$FD,$45,$3D,$1B,$19,$3F
	.dc.b	$D5,$F1,$C6,$33,$1D,$EB,$9B,$63,$37,$FE,$F4,$6C,$EA,$31,$3D,$0F
	.dc.b	$6F,$A3,$CB,$EB,$5F,$20,$F4,$F9,$CE,$63,$15,$C4,$F8,$11,$99,$DE
	.dc.b	$3E,$6B,$E7,$16,$F8,$E9,$F2,$27,$38,$73,$C7,$F9,$CC,$E6,$B7,$9A
	.dc.b	$A5,$F0,$AD,$61,$E0,$9F,$24,$E6,$B8,$42,$30,$F2,$BE,$AF,$C2,$C7
	.dc.b	$CD,$72,$39,$CE,$9F,$7C,$E7,$0F,$0C,$6A,$BC,$51,$82,$61,$E2,$C3
	.dc.b	$C3,$5C,$8E,$04,$9D,$BE,$33,$13,$C7,$0F,$15,$72,$C7,$E3,$CE,$70
	.dc.b	$F2,$36,$FE,$66,$5F,$55,$E6,$7C,$19,$F3,$CE,$F1,$F0,$CF,$C7,$C7
	.dc.b	$E0,$61,$DA,$38,$1C,$8D,$5F,$27,$35,$F1,$1F,$A2,$68,$A5,$33,$8F
	.dc.b	$E0,$9B,$EE,$FF,$25,$77,$1F,$E6,$FE,$BE,$31,$F6,$E7,$4F,$08,$CC
	.dc.b	$EE,$3F,$07,$3C,$EB,$F2,$0D,$CB,$EA,$78,$7F,$EF,$B7,$EB,$46,$6B
	.dc.b	$F1,$ED,$7C,$18,$7E,$76,$B8,$46,$A3,$F4,$13,$BC,$3C,$58,$7E,$7E
	.dc.b	$F7,$7C,$8F,$2D,$F0,$3A,$34,$C9,$27,$93,$97,$D7,$17,$0F,$5C,$D3
	.dc.b	$C8,$71,$72,$D3,$33,$C9,$FD,$4D,$79,$AF,$D3,$5F,$12,$FA,$57,$EA
	.dc.b	$EB,$58,$7B,$C9,$3E,$A9,$EC,$7B,$F0,$EC,$7A,$9C,$E1,$D9,$F1,$E1
	.dc.b	$E6,$9D,$1C,$1F,$14,$79,$8E,$D1,$97,$89,$FC,$1B,$0F,$9F,$F9,$67
	.dc.b	$F3,$69,$FE,$6E,$BE,$A3,$26,$AE,$5F,$F2,$E9,$E1,$1F,$C2,$9B,$1F
	.dc.b	$C4,$D7,$C6,$76,$38,$C7,$F2,$58,$DA,$76,$F5,$8D,$DF,$92,$B9,$4F
	.dc.b	$87,$CB,$FE,$33,$CB,$FE,$B3,$C3,$FD,$4F,$3F,$E1,$D6,$BF,$E3,$E5
	.dc.b	$3C,$38,$BD,$0D,$9F,$47,$A5,$5F,$23,$6B,$8C,$75,$EF,$E1,$D4,$FA
	.dc.b	$AC,$CF,$F7,$86,$E5,$E5,$AE,$44,$9E,$47,$F8,$CF,$F0,$FF,$E0,$73
	.dc.b	$FE,$CD,$7F,$B4,$46,$79,$3C,$91,$FD,$3A,$3F,$CB,$2E,$9B,$CB,$A9
	.dc.b	$91,$D6,$3D,$9F,$42,$FB,$1A,$49,$FF,$22,$7E,$61,$FE,$DD,$C7,$FD
	.dc.b	$BD,$5B,$9F,$F2,$43,$5F,$F3,$4B,$63,$FD,$C9,$3C,$78,$3F,$A6,$CB
	.dc.b	$E3,$FC,$24,$BE,$7E,$E5,$79,$C5,$C3,$C3,$3F,$4F,$F3,$07,$D9,$7A
	.dc.b	$BE,$52,$61,$FE,$5B,$9C,$B6,$72,$F9,$A7,$94,$48,$D1,$0F,$5F,$C7
	.dc.b	$FE,$07,$EC,$5E,$25,$24,$FF,$9A,$61,$C7,$88,$FF,$45,$C7,$07,$53
	.dc.b	$CE,$7A,$78,$FD,$7A,$25,$73,$8C,$BD,$63,$2F,$48,$FC,$A7,$D8,$BE
	.dc.b	$C7,$CD,$D9,$2B,$FC,$F9,$39,$6C,$E5,$F2,$CF,$28,$91,$A2,$1E,$AF
	.dc.b	$B0,$FD,$17,$53,$DD,$BE,$EC,$AF,$F4,$54,$E5,$B3,$97,$FD,$5B,$09
	.dc.b	$E7,$3D,$3E,$8E,$57,$27,$BF,$D0,$62,$38,$C5,$E6,$74,$FF,$86,$31
	.dc.b	$5F,$ED,$FC,$9F,$F2,$7D,$6B,$0F,$F1,$25,$7E,$4D,$B7,$BD,$64,$7A
	.dc.b	$A3,$DE,$61,$FF,$62,$DF,$48,$FF,$13,$B6,$3F,$C6,$0D,$5F,$52,$FB
	.dc.b	$3F,$E1,$4F,$E8,$63,$FD,$60,$BA,$B2,$6B,$AE,$4E,$A3,$2F,$08,$E7
	.dc.b	$E6,$C8,$E6,$F6,$EB,$C4,$7F,$B7,$2D,$38,$D7,$E2,$4F,$9A,$BE,$47
	.dc.b	$B9,$FC,$7F,$7F,$CD,$CF,$FB,$15,$AF,$C2,$8D,$B9,$E8,$E0,$61,$D1
	.dc.b	$FE,$CD,$D5,$39,$C3,$C3,$78,$AF,$F6,$97,$E6,$0F,$A5,$FE,$61,$F9
	.dc.b	$87,$E1,$E7,$FD,$AC,$D7,$FD,$EF,$3B,$78,$F9,$BC,$39,$64,$60,$FF
	.dc.b	$6F,$7E,$2C,$FA,$2F,$E7,$9F,$A0,$7F,$17,$E6,$FF,$30,$B6,$BF,$CD
	.dc.b	$55,$B7,$8F,$C7,$70,$F1,$24,$60,$FF,$75,$7E,$87,$FC,$C7,$83,$E6
	.dc.b	$E0,$F1,$FB,$9F,$D2,$E7,$59,$7F,$9C,$17,$CC,$79,$3C,$C9,$8F,$48
	.dc.b	$FE,$A5,$EF,$3E,$9F,$FA,$A7,$E6,$3F,$53,$4F,$B6,$DC,$BF,$CF,$75
	.dc.b	$BC,$3F,$DF,$CB,$E3,$F8,$F9,$32,$73,$1F,$D7,$FE,$A5,$5D,$BD,$89
	.dc.b	$5F,$E8,$DC,$BF,$13,$5A,$8C,$63,$FE,$8A,$87,$D4,$FF,$69,$F6,$5F
	.dc.b	$0B,$4E,$D1,$DF,$E5,$F5,$F2,$93,$1E,$67,$FA,$6F,$04,$FF,$73,$FC
	.dc.b	$39,$F0,$3D,$7C,$39,$7C,$04,$E9,$C7,$E3,$4F,$96,$FE,$F9,$F8,$77
	.dc.b	$F4,$69,$70,$7E,$92,$30,$FE,$9A,$B8,$4F,$5F,$87,$7D,$1F,$F8,$16
	.dc.b	$5F,$25,$4F,$F0,$44,$EA,$B8,$7C,$4E,$2E,$46,$0E,$9E,$5F,$93,$D0
	.dc.b	$91,$83,$D9,$F9,$13,$E9,$7A,$D4,$9C,$C7,$F8,$5E,$75,$18,$21,$EE
	.dc.b	$E8,$FF,$5F,$EC,$F9,$2F,$F0,$F7,$E2,$1F,$85,$AF,$95,$3C,$9E,$07
	.dc.b	$FB,$35,$C7,$07,$F0,$A9,$E1,$3A,$0E,$23,$E6,$63,$F4,$0E,$FD,$B2
	.dc.b	$BE,$C5,$6F,$1F,$F6,$9B,$63,$C1,$F6,$9E,$6E,$1F,$B3,$E9,$49,$E3
	.dc.b	$19,$7D,$31,$97,$C9,$5F,$59,$FC,$37,$E6,$1F,$E4,$E7,$DB,$9B,$EB
	.dc.b	$3F,$9A,$47,$E4,$A3,$1C,$3F,$E4,$3A,$E1,$1F,$CA,$9B,$9F,$FB,$A9
	.dc.b	$B5,$FE,$F3,$6A,$7F,$4A,$FE,$5F,$17,$C8,$7F,$93,$DF,$F1,$E5,$FF
	.dc.b	$36,$6D,$FF,$BD,$9D,$D7,$1A,$E0,$47,$F4,$5F,$61,$FB,$2B,$43,$EC
	.dc.b	$17,$FF,$00,$2B,$53,$98,$ED,$7C,$23,$07,$28,$F0,$B8,$13,$CA,$BA
	.dc.b	$C7,$82,$31,$3D,$23,$C7,$3A,$7C,$75,$DE,$31,$5C,$27,$97,$81,$FB
	.dc.b	$67,$C1,$7E,$2A,$E4,$D3,$D9,$3B,$8E,$6D,$8F,$14,$F2,$BF,$04,$70
	.dc.b	$C3,$CA,$D9,$D1,$83,$E1,$4F,$07,$CB,$39,$7E,$1D,$79,$DF,$F7,$EE
	.dc.b	$1F,$39,$B1,$97,$B1,$AF,$41,$F1,$9F,$EF,$AA,$FE,$B0,$D8,$CC,$FD
	.dc.b	$53,$E7,$38,$7F,$E0,$0F,$C7,$E5,$D3,$5F,$B8,$D7,$F0,$AD,$4F,$AF
	.dc.b	$0F,$F5,$CE,$9E,$D9,$D9,$F8,$18,$CB,$F8,$33,$3A,$7D,$D8,$CB,$F7
	.dc.b	$E3,$3A,$7F,$AE,$3C,$5F,$F1,$B4,$E6,$7C,$23,$E0,$B9,$4F,$F8,$ED
	.dc.b	$D6,$3F,$9C,$73,$8D,$C3,$E0,$4F,$A6,$BA,$C6,$2F,$D1,$18,$7E,$29
	.dc.b	$8A,$F7,$3E,$2C,$3A,$B9,$3F,$8F,$63,$F0,$0F,$F1,$5B,$EC,$3F,$B3
	.dc.b	$57,$F9,$46,$3F,$D5,$4E,$A3,$35,$F4,$E7,$E4,$C7,$FB,$02,$74,$BE
	.dc.b	$BF,$F8,$2D,$7F,$5D,$6C,$7F,$2F,$7F,$7B,$19,$DF,$C0,$FE,$E6,$37
	.dc.b	$3F,$87,$6B,$99,$EF,$18,$8F,$E0,$EF,$F4,$F9,$FC,$43,$63,$F1,$8D
	.dc.b	$7A,$73,$F2,$3C,$2F,$F6,$CF,$B0,$7C,$FE,$23,$D9,$3A,$C9,$5A,$C3
	.dc.b	$DD,$E7,$EF,$1F,$F1,$2C,$7F,$33,$E6,$FE,$B0,$E8,$98,$F9,$2B,$D6
	.dc.b	$48,$F4,$BD,$E7,$FD,$A5,$1C,$30,$FA,$D1,$8A,$FE,$B5,$F2,$9F,$DF
	.dc.b	$2B,$FC,$18,$D8,$FE,$BD,$F5,$EF,$2F,$D9,$39,$C3,$FC,$DA,$EE,$BF
	.dc.b	$6C,$D7,$F9,$63,$5F,$E6,$AD,$5F,$60,$F8,$BF,$C1,$B7,$1F,$DD,$36
	.dc.b	$7F,$7C,$D7,$F8,$22,$F9,$5E,$24,$9E,$2F,$E3,$BE,$C1,$F2,$14,$93
	.dc.b	$FC,$1E,$32,$DF,$9C,$E0,$47,$07,$F0,$DF,$B0,$7C,$0F,$E6,$9B,$19
	.dc.b	$8C,$1F,$9E,$7F,$CE,$87,$F1,$52,$E5,$FC,$65,$AF,$F1,$F6,$BF,$D5
	.dc.b	$7E,$BF,$F4,$2F,$F1,$F6,$BF,$C9,$5A,$EC,$FD,$6A,$F9,$7A,$49,$FE
	.dc.b	$4D,$19,$6F,$1F,$02,$38,$3F,$97,$F9,$DF,$B8,$ED,$34,$91,$FC,$D3
	.dc.b	$C8,$D1,$1C,$1F,$CD,$7E,$B3,$FC,$E2,$DF,$67,$F8,$1C,$63,$3F,$E7
	.dc.b	$8D,$7F,$A1,$3F,$EA,$E9,$E5,$F1,$3C,$08,$F5,$F5,$F9,$91,$C1,$FD
	.dc.b	$17,$DE,$FF,$36,$B7,$D8,$FE,$27,$18,$C3,$FA,$63,$5F,$EA,$5F,$54
	.dc.b	$FB,$DF,$E3,$2D,$BF,$EA,$0D,$9F,$EA,$8D,$7F,$AD,$7D,$5B,$E1,$BF
	.dc.b	$83,$36,$BF,$C6,$5F,$46,$FA,$5C,$D4,$9F,$EB,$95,$BB,$FE,$C0,$D7
	.dc.b	$AC,$FF,$2C,$8D,$77,$FF,$35,$C3,$F7,$57,$FE,$89,$9C,$BF,$E9,$F8
	.dc.b	$CD,$FF,$91,$BE,$AF,$F6,$17,$FC,$F5,$B5,$FE,$A5,$5F,$07,$FC,$75
	.dc.b	$CD,$7F,$A3,$9B,$46,$9C,$2B,$FC,$A0,$EA,$FF,$D3,$11,$F0,$B8,$FF
	.dc.b	$94,$46,$2F,$FC,$B2,$FF,$AF,$67,$E3,$FB,$DA,$AD,$B6,$3F,$D4,$DF
	.dc.b	$56,$F9,$1F,$E6,$8D,$8F,$0E,$B8,$99,$1F,$E0,$99,$D6,$3F,$E0,$D6
	.dc.b	$BF,$E1,$36,$BF,$E4,$97,$79,$FF,$63,$FA,$7F,$C5,$2A,$D5,$FF,$43
	.dc.b	$9C,$C7,$D1,$F9,$8F,$79,$F0,$3E,$66,$C7,$E1,$22,$E3,$F0,$A6,$E1
	.dc.b	$F5,$7C,$4F,$A5,$F5,$5F,$4D,$B8,$7C,$E6,$DF,$DD,$6D,$79,$63,$FC
	.dc.b	$D3,$9F,$F5,$57,$FD,$3C,$D8,$FC,$57,$D5,$FC,$AF,$42,$61,$F1,$DF
	.dc.b	$EF,$75,$C6,$FE,$34,$60,$FE,$C5,$F4,$EC,$B8,$CF,$FA,$52,$31,$7F
	.dc.b	$84,$FA,$BF,$C6,$7C,$37,$E3,$5B,$19,$3F,$13,$19,$69,$FA,$AF,$1B
	.dc.b	$F1,$27,$E2,$BE,$02,$C6,$49,$87,$E3,$23,$2F,$FA,$F7,$EA,$FF,$65
	.dc.b	$3F,$EE,$F6,$C7,$E9,$7C,$1F,$EF,$D1,$8D,$FE,$54,$A3,$6F,$F9,$65
	.dc.b	$AF,$FA,$3D,$AF,$FA,$61,$AF,$1E,$CF,$C3,$7C,$27,$66,$E1,$FC,$1B
	.dc.b	$E3,$3F,$4A,$7F,$83,$7C,$C7,$A3,$89,$D1,$30,$FF,$60,$CE,$A3,$19
	.dc.b	$7E,$F7,$E8,$7F,$74,$7E,$EF,$E2,$2C,$FE,$DA,$2D,$FF,$05,$9B,$5F
	.dc.b	$03,$EA,$BF,$9B,$4F,$FA,$AD,$AF,$FA,$D1,$A9,$A7,$F3,$F8,$E5,$F5
	.dc.b	$7F,$E2,$98,$C6,$1C,$E3,$1E,$AC,$8F,$F5,$FD,$78,$7E,$13,$18,$7F
	.dc.b	$BD,$70,$FF,$03,$3E,$3B,$FF,$27,$C6,$23,$C9,$18,$E4,$98,$7F,$A2
	.dc.b	$69,$18,$AF,$F4,$74,$62,$7C,$27,$BF,$3F,$45,$7A,$5F,$F2,$17,$A5
	.dc.b	$FE,$4F,$BF,$EE,$5E,$EF,$F9,$82,$30,$47,$F6,$5E,$8B,$E8,$3F,$DA
	.dc.b	$93,$FC,$4A,$31,$5F,$ED,$E8,$FF,$32,$C6,$23,$FB,$F7,$87,$CA,$B8
	.dc.b	$B6,$38,$35,$F1,$E5,$FE,$A4,$9D,$B6,$36,$FF,$83,$2F,$F9,$1B,$FE
	.dc.b	$B3,$6C,$70,$3F,$D0,$AF,$F6,$98,$D7,$9D,$FC,$35,$A7,$37,$13,$EE
	.dc.b	$7C,$9F,$54,$F9,$07,$A1,$E3,$AF,$F0,$D5,$6A,$73,$1F,$EB,$AC,$7F
	.dc.b	$DF,$57,$C6,$F8,$36,$74,$7C,$9B,$E5,$3A,$BE,$79,$FC,$BB,$E0,$FC
	.dc.b	$DC,$BF,$C0,$8E,$67,$78,$F8,$1B,$88,$5F,$29,$F7,$65,$DF,$19,$5E
	.dc.b	$43,$77,$C9,$F7,$CE,$B0,$F0,$55,$BE,$75,$C7,$0E,$CF,$85,$AF,$49
	.dc.b	$F9,$58,$FA,$AB,$D1,$1C,$E7,$86,$3E,$1A,$D4,$62,$74,$F3,$3D,$11
	.dc.b	$A7,$18,$79,$9F,$25,$73,$CB,$E7,$C6,$2F,$9D,$F1,$C7,$CE,$4B,$E1
	.dc.b	$97,$82,$31,$5E,$AC,$39,$E3,$DF,$1F,$A0,$79,$CA,$52,$D7,$E0,$DB
	.dc.b	$F5,$7F,$CA,$1D,$9F,$61,$C3,$AE,$7C,$B9,$6C,$A7,$95,$7E,$25,$BA
	.dc.b	$7E,$51,$B7,$F5,$7B,$BC,$F5,$BC,$3F,$2F,$5A,$AE,$C7,$DD,$AD,$DF
	.dc.b	$3A,$D9,$27,$45,$BE,$8F,$E1,$6F,$A6,$1F,$E9,$88,$D5,$F3,$75,$7F
	.dc.b	$A8,$C3,$BC,$7E,$2E,$F8,$92,$75,$1C,$8D,$35,$FD,$0C,$F2,$7D,$78
	.dc.b	$79,$09,$7F,$B8,$9D,$61,$F2,$EF,$93,$6F,$DA,$79,$89,$7C,$9F,$98
	.dc.b	$70,$C3,$CC,$F0,$8D,$3E,$2C,$3E,$04,$7E,$FF,$0F,$09,$E2,$AE,$31
	.dc.b	$9A,$F7,$9F,$DF,$8F,$69,$FE,$B0,$BF,$98,$E5,$B9,$FE,$4B,$EA,$FF
	.dc.b	$9B,$37,$1F,$CB,$B6,$BF,$D7,$33,$A7,$F3,$6F,$F3,$DB,$FE,$7C,$DA
	.dc.b	$C9,$F3,$57,$C1,$D3,$11,$2B,$F9,$54,$E9,$B3,$F9,$C5,$19,$7F,$46
	.dc.b	$D7,$D5,$1B,$8F,$F7,$84,$7F,$62,$3F,$90,$3F,$E8,$47,$07,$4F,$43
	.dc.b	$F6,$BD,$1F,$E2,$69,$E3,$28,$C7,$91,$D4,$91,$FB,$09,$DB,$F2,$FA
	.dc.b	$FC,$44,$74,$D7,$A6,$7F,$2A,$3C,$2D,$8F,$2F,$D5,$FF,$06,$C3,$FC
	.dc.b	$72,$FF,$B9,$E3,$3C,$5C,$1F,$E7,$91,$FD,$41,$EA,$63,$93,$4A,$39
	.dc.b	$3F,$C3,$1A,$FF,$14,$6B,$FC,$61,$AF,$F4,$9F,$99,$7D,$5F,$46,$C7
	.dc.b	$CC,$6B,$E2,$1F,$57,$FE,$52,$C7,$FC,$A7,$3A,$6D,$73,$3C,$F3,$BC
	.dc.b	$BF,$94,$D3,$E9,$7C,$04,$C3,$C1,$F2,$15,$F0,$4E,$DF,$24,$F9,$E7
	.dc.b	$E1,$FD,$5B,$E7,$FD,$5F,$1B,$2F,$F0,$B3,$96,$CF,$E4,$E3,$11,$FC
	.dc.b	$D1,$AF,$F3,$C6,$BF,$D0,$1D,$E9,$FE,$72,$6E,$59,$AF,$95,$F0,$BF
	.dc.b	$5B,$EA,$DE,$0D,$26,$3F,$98,$87,$1E,$23,$E7,$47,$07,$53,$CE,$7A
	.dc.b	$78,$9D,$79,$E5,$73,$8C,$BD,$63,$2F,$48,$FC,$4F,$D5,$BE,$AF,$CD
	.dc.b	$D7,$2B,$F5,$73,$96,$CF,$EE,$63,$11,$FD,$69,$AF,$F6,$26,$BF,$D9
	.dc.b	$9D,$EF,$FE,$98,$6E,$D9,$AF,$8D,$F2,$3F,$9C,$FA,$B7,$B3,$7D,$99
	.dc.b	$97,$EF,$A7,$2D,$9C,$BF,$95,$84,$F3,$9E,$9F,$4B,$2B,$93,$DF,$EA
	.dc.b	$71,$1C,$62,$E9,$FE,$AE,$C1,$5F,$B0,$F2,$3F,$CC,$53,$9D,$3D,$06
	.dc.b	$76,$98,$FF,$7A,$AF,$F3,$43,$FE,$06,$6B,$FE,$0A,$6B,$FE,$3B,$7A
	.dc.b	$CE,$67,$9C,$FE,$C7,$EA,$9F,$57,$FC,$09,$FD,$24,$67,$AB,$06,$7A
	.dc.b	$E4,$7F,$26,$94,$EA,$31,$3C,$FC,$B9,$1B,$7B,$76,$62,$3C,$51,$74
	.dc.b	$FF,$69,$E0,$AF,$8D,$EB,$7F,$16,$DB,$F3,$B8,$7F,$36,$6B,$FC,$2F
	.dc.b	$64,$F0,$8F,$90,$FF,$47,$8C,$C6,$08,$71,$8E,$0F,$F1,$59,$DB,$FC
	.dc.b	$5F,$EA,$5F,$57,$FD,$45,$B1,$FE,$2E,$6B,$FE,$34,$6B,$FE,$3C,$7F
	.dc.b	$B7,$B7,$CC,$E4,$47,$FC,$6F,$F5,$4F,$9A,$FE,$BC,$D8,$CC,$60,$CF
	.dc.b	$C1,$C8,$E2,$D7,$FC,$8B,$F5,$3F,$CD,$93,$49,$E4,$FF,$6B,$5F,$21
	.dc.b	$49,$3F,$E4,$A8,$CB,$FE,$4C,$F8,$F7,$D4,$FF,$73,$6C,$66,$30,$67
	.dc.b	$E2,$64,$71,$6B,$FE,$5A,$FA,$75,$7F,$7D,$6C,$7F,$96,$9A,$EE,$35
	.dc.b	$ED,$C8,$D3,$5E,$9F,$21,$CC,$8F,$37,$FC,$15,$F4,$EF,$6A,$92,$7F
	.dc.b	$CD,$31,$97,$FC,$D7,$F4,$EF,$8D,$FF,$09,$36,$33,$18,$3F,$C2,$DF
	.dc.b	$28,$FA,$7A,$49,$FF,$3A,$C6,$5B,$F0,$1C,$08,$E0,$DF,$9D,$D0,$8E
	.dc.b	$0F,$F3,$FF,$D3,$BE,$83,$FC,$4C,$D8,$CC,$60,$8F,$F8,$82,$34,$FF
	.dc.b	$75,$6F,$82,$F1,$FF,$D2,$FF,$4E,$F8,$9F,$F0,$4F,$D3,$39,$BF,$C1
	.dc.b	$73,$C2,$B7,$96,$A6,$65,$FE,$94,$BE,$0D,$8F,$F1,$E4,$62,$3F,$D3
	.dc.b	$AD,$C7,$C5,$1F,$ED,$78,$C5,$FF,$AA,$9B,$1F,$ED,$BC,$1F,$4F,$E6
	.dc.b	$F8,$DF,$2F,$8D,$2B,$FC,$7F,$69,$D4,$79,$5F,$19,$1F,$F5,$5F,$D3
	.dc.b	$FF,$93,$5F,$F5,$CB,$5F,$F7,$53,$53,$4F,$F2,$B6,$5E,$A8,$CF,$C1
	.dc.b	$FF,$8C,$A7,$39,$F4,$E9,$FF,$1A,$E4,$F8,$8F,$F3,$03,$70,$FF,$5E
	.dc.b	$B5,$FF,$61,$B5,$FF,$65,$B9,$C7,$C0,$DB,$FF,$68,$B6,$43,$E9,$D8
	.dc.b	$7F,$97,$E1,$F4,$DF,$CD,$5D,$CE,$B0,$E3,$39,$9F,$E7,$58,$FF,$B7
	.dc.b	$A3,$18,$79,$DD,$5F,$1A,$EB,$3F,$EE,$48,$C4,$F0,$AE,$53,$E8,$6D
	.dc.b	$78,$B0,$F3,$57,$47,$D1,$EB,$71,$9F,$6B,$53,$BC,$73,$87,$86,$B9
	.dc.b	$BD,$EB,$DA,$D7,$CE,$78,$67,$2D,$32,$7C,$23,$25,$33,$8F,$BA,$7C
	.dc.b	$13,$FE,$F8,$8F,$89,$3F,$50,$E8,$53,$EB,$57,$D5,$71,$1F,$D2,$7E
	.dc.b	$9F,$97,$8C,$69,$27,$FD,$55,$F1,$1F,$EA,$B4,$D2,$3A,$3D,$78,$7F
	.dc.b	$AE,$35,$F5,$C7,$E2,$27,$DB,$39,$8F,$82,$64,$E3,$BF,$EE,$B5,$FC
	.dc.b	$34,$65,$FC,$44,$F1,$73,$1E,$43,$8D,$F8,$1C,$47,$0A,$E9,$A7,$C5
	.dc.b	$9C,$BF,$EB,$A8,$CC,$7C,$9C,$3C,$F1,$88,$F9,$D1,$8B,$EC,$D8,$F0
	.dc.b	$9F,$E9,$9C,$3D,$C7,$FB,$B5,$EE,$7E,$A9,$CC,$F8,$4F,$F6,$6B,$88
	.dc.b	$F1,$1D,$1F,$E0,$9F,$4D,$FB,$F9,$FF,$69,$B6,$7F,$D5,$0F,$ED,$A3
	.dc.b	$3B,$FF,$24,$AE,$2D,$8F,$C4,$4F,$19,$EB,$59,$3F,$7F,$19,$9D,$F3
	.dc.b	$66,$53,$FD,$2E,$7F,$DD,$14,$FA,$7F,$F5,$34,$79,$AF,$85,$6E,$FC
	.dc.b	$F1,$DE,$F8,$7C,$16,$A3,$7E,$44,$C3,$AF,$D3,$FF,$4D,$7E,$0C,$67
	.dc.b	$0F,$82,$F3,$9D,$46,$27,$67,$2F,$57,$86,$3F,$16,$BF,$D8,$D3,$FE
	.dc.b	$FE,$FA,$77,$D3,$FB,$FE,$43,$FA,$14,$E6,$31,$D5,$23,$EE,$1F,$B3
	.dc.b	$9D,$E5,$FE,$DB,$BE,$73,$A8,$FE,$57,$3A,$C7,$FA,$07,$D3,$BE,$0F
	.dc.b	$E7,$36,$F8,$5A,$F8,$47,$F0,$19,$D4,$FF,$BA,$57,$C7,$F2,$3E,$C7
	.dc.b	$E8,$9F,$EA,$0A,$E7,$3A,$7A,$FD,$43,$E2,$BE,$F3,$63,$77,$89,$CF
	.dc.b	$CB,$49,$E3,$C7,$72,$FF,$7D,$7D,$43,$D9,$FC,$43,$6B,$F0,$FE,$5F
	.dc.b	$F8,$B5,$F6,$C6,$69,$C1,$FD,$8E,$7F,$CC,$07,$C2,$2C,$69,$FD,$67
	.dc.b	$D9,$3E,$9F,$F2,$D1,$AA,$FE,$D9,$F5,$3F,$A5,$BE,$07,$F5,$C7,$11
	.dc.b	$FA,$7B,$FF,$1D,$57,$D0,$9C,$BF,$CB,$D7,$B5,$FA,$36,$C7,$E8,$BE
	.dc.b	$29,$F0,$59,$AD,$76,$4A,$D7,$C5,$5C,$FC,$67,$F2,$EF,$4F,$F5,$DE
	.dc.b	$19,$FD,$37,$D9,$7F,$05,$9D,$5F,$E1,$2B,$FC,$5C,$DF,$47,$F5,$9F
	.dc.b	$64,$FA,$5E,$71,$FB,$9F,$AB,$FF,$35,$D7,$E3,$1B,$87,$C4,$E2,$7D
	.dc.b	$97,$F7,$EA,$D7,$BF,$C7,$E2,$24,$7F,$00,$97,$C0,$63,$0F,$5E,$38
	.dc.b	$AE,$D3,$FB,$FF,$B1,$FC,$BE,$5F,$E1,$FA,$D5,$FC,$2F,$9B,$FC,$23
	.dc.b	$7E,$0B,$F8,$36,$69,$FE,$24,$6A,$FF,$C5,$1F,$5B,$F8,$1C,$FF,$4E
	.dc.b	$DC,$3D,$97,$FB,$E7,$E7,$C6,$6B,$FC,$61,$F5,$AF,$88,$FE,$3C,$DC
	.dc.b	$3F,$8E,$FD,$63,$E9,$73,$52,$7F,$6B,$5B,$C7,$E5,$5E,$E7,$35,$B9
	.dc.b	$FB,$F5,$BC,$3F,$9B,$5E,$E7,$35,$FC,$97,$EB,$7E,$3E,$1F,$D2,$5B
	.dc.b	$3F,$A6,$5F,$33,$BF,$36,$57,$F3,$49,$DE,$27,$71,$99,$FE,$65,$8F
	.dc.b	$E8,$30,$F4,$F2,$E3,$E1,$F2,$DB,$F8,$6D,$9D,$C7,$F1,$6F,$AC,$7C
	.dc.b	$9F,$F4,$76,$C7,$8F,$5C,$4C,$8E,$0F,$BE,$31,$5F,$A4,$AE,$2D,$9D
	.dc.b	$57,$03,$EA,$F7,$7A,$8F,$5F,$85,$63,$24,$AF,$AB,$19,$7F,$41,$AF
	.dc.b	$F2,$36,$FD,$5F,$D1,$F3,$7E,$93,$6F,$E6,$36,$77,$1F,$07,$EB,$3E
	.dc.b	$01,$F3,$3B,$AC,$64,$93,$F8,$68,$CB,$EF,$9B,$C5,$F8,$49,$B5,$F8
	.dc.b	$1F,$AB,$FC,$27,$97,$F8,$56,$C7,$DF,$6E,$1E,$C6,$DF,$E3,$BC,$CF
	.dc.b	$BE,$7E,$03,$D5,$B7,$F9,$E8,$B7,$F9,$A9,$B5,$FE,$95,$FA,$BF,$CB
	.dc.b	$37,$AB,$F2,$DD,$9F,$E0,$59,$CE,$3F,$E5,$56,$E1,$FE,$8A,$6D,$FE
	.dc.b	$4E,$B8,$5C,$BF,$EB,$F5,$AE,$05,$67,$0F,$D6,$FD,$5F,$E6,$0F,$F0
	.dc.b	$57,$A3,$FD,$97,$0F,$DB,$C6,$5A,$7F,$85,$A3,$53,$CA,$3F,$4D,$F4
	.dc.b	$1F,$AF,$3F,$61,$F1,$D6,$32,$49,$FC,$3C,$65,$FE,$73,$F5,$5F,$CA
	.dc.b	$B5,$FF,$5D,$36,$BF,$D5,$3F,$0B,$FE,$28,$F7,$3F,$C5,$53,$9B,$FF
	.dc.b	$56,$B5,$39,$9E,$15,$FE,$02,$8D,$1F,$C0,$A7,$7E,$07,$4F,$88,$EA
	.dc.b	$CA,$E0,$E5,$A7,$3E,$5F,$19,$FE,$41,$C3,$FC,$1D,$F0,$98,$9D,$13
	.dc.b	$0F,$ED,$33,$A8,$C6,$7F,$ED,$9F,$4B,$FC,$91,$5F,$EC,$86,$BF,$DA
	.dc.b	$3C,$7F,$F2,$54,$60,$9F,$09,$88,$CB,$AF,$03,$FD,$FB,$E8,$3E,$77
	.dc.b	$FD,$41,$7F,$C1,$23,$18,$77,$BD,$D6,$AE,$CF,$8D,$AF,$F7,$A6,$DF
	.dc.b	$F9,$AE,$F9,$CE,$72,$FF,$4D,$5F,$83,$CB,$D9,$C3,$DA,$B3,$AA,$F8
	.dc.b	$13,$87,$FD,$53,$5E,$8A,$C5,$F6,$FA,$9F,$F0,$3E,$9C,$6D,$18,$78
	.dc.b	$9D,$31,$D4,$66,$B9,$7A,$FF,$D9,$6B,$67,$27,$11,$EC,$CF,$FB,$6B
	.dc.b	$C3,$0E,$73,$AC,$BF,$D9,$98,$75,$7E,$06,$3A,$A9,$8F,$6A,$D3,$E5
	.dc.b	$31,$7C,$D1,$E3,$9E,$38,$78,$F3,$FA,$77,$CA,$B4,$72,$AF,$89,$97
	.dc.b	$A3,$2F,$2E,$3D,$6B,$78,$FA,$DF,$83,$5B,$7C,$38,$75,$25,$71,$9D
	.dc.b	$4F,$2B,$EB,$87,$99,$F8,$31,$E3,$26,$1D,$CF,$56,$5B,$8E,$F8,$79
	.dc.b	$71,$F6,$63,$D9,$F0,$E3,$BA,$FF,$25,$E5,$F0,$AF,$81,$F6,$E3,$FD
	.dc.b	$F9,$5F,$83,$6C,$65,$F3,$9F,$ED,$D8,$7D,$57,$FA,$1E,$30,$D7,$9B
	.dc.b	$5F,$BE,$D7,$E3,$BC,$72,$FC,$B3,$73,$DF,$C0,$FE,$4B,$C1,$FF,$72
	.dc.b	$C6,$2F,$F0,$95,$B7,$9D,$6D,$FC,$CC,$65,$FC,$FB,$70,$F4,$CE,$E7
	.dc.b	$9B,$C0,$8F,$E4,$6F,$C2,$F6,$36,$4C,$3F,$5D,$1A,$8E,$F3,$CA,$7D
	.dc.b	$AF,$76,$C7,$1C,$3D,$91,$CA,$38,$46,$08,$FE,$3A,$BB,$5F,$AA,$39
	.dc.b	$1E,$5A,$E4,$E0,$F1,$96,$3F,$D5,$0F,$63,$85,$76,$7F,$82,$D7,$6C
	.dc.b	$3E,$41,$30,$F5,$57,$EE,$A7,$C6,$4A,$FE,$13,$E0,$D7,$E5,$1C,$0F
	.dc.b	$D3,$BF,$1F,$EA,$BF,$91,$E5,$FC,$91,$B7,$FB,$CA,$46,$1F,$DF,$AE
	.dc.b	$5E,$53,$27,$F8,$F4,$ED,$E3,$3D,$CE,$9C,$8E,$1E,$D5,$CE,$BF,$DC
	.dc.b	$6D,$7F,$D5,$2B,$EA,$F8,$CF,$0B,$B5,$E4,$6B,$E0,$39,$CE,$2B,$FD
	.dc.b	$D3,$1D,$BC,$8C,$1C,$B9,$DE,$E7,$96,$74,$43,$C7,$3B,$8D,$47,$9B
	.dc.b	$BB,$B1,$1F,$F7,$EF,$D5,$FC,$B7,$47,$BA,$34,$FF,$B1,$A3,$5A,$74
	.dc.b	$7F,$DA,$71,$CB,$9F,$DB,$18,$8F,$EE,$F6,$F3,$33,$52,$78,$1F,$CD
	.dc.b	$AF,$D9,$97,$BA,$BE,$3F,$D4,$BE,$AE,$92,$F8,$70,$7E,$0F,$EA,$FD
	.dc.b	$F8,$FC,$0B,$E1,$5A,$3F,$A1,$B8,$7F,$DC,$2D,$8F,$8F,$3C,$AB,$F1
	.dc.b	$D2,$F8,$AF,$19,$2F,$A0,$87,$D2,$66,$3A,$CD,$C3,$C9,$3F,$82,$FA
	.dc.b	$B7,$C6,$D2,$5F,$4C,$1F,$11,$88,$FF,$7B,$D7,$07,$C2,$D7,$D7,$1B
	.dc.b	$C3,$F0,$33,$96,$DF,$CC,$9F,$C9,$FD,$5B,$CA,$A4,$C3,$DB,$0E,$4C
	.dc.b	$43,$0F,$C6,$CF,$18,$DD,$74,$AE,$BE,$3F,$6F,$6E,$57,$48,$CB,$DA
	.dc.b	$32,$F5,$DF,$F7,$D8,$2B,$F5,$3F,$56,$FA,$BA,$4B,$F3,$FC,$BD,$79
	.dc.b	$1F,$E0,$29,$E0,$FD,$86,$BA,$16,$9E,$5A,$7E,$AA,$B6,$DB,$F9,$33
	.dc.b	$FB,$CF,$AA,$FC,$F6,$1F,$84,$5E,$D6,$4D,$73,$C9,$D4,$65,$E1,$1C
	.dc.b	$FE,$AE,$47,$37,$B7,$D4,$E2,$3D,$16,$9C,$65,$FC,$02,$15,$FC,$43
	.dc.b	$C9,$FF,$26,$CF,$F0,$46,$EF,$FC,$AD,$B8,$7F,$8A,$3E,$A9,$F5,$7C
	.dc.b	$7C,$59,$3F,$A0,$AD,$AA,$DB,$99,$DB,$F3,$67,$73,$A8,$FA,$0F,$EB
	.dc.b	$5F,$D5,$BF,$AA,$6F,$57,$F2,$88,$D4,$70,$EB,$C4,$C8,$E3,$1B,$F3
	.dc.b	$E4,$74,$79,$7C,$CE,$23,$C3,$17,$4F,$E4,$F8,$2B,$E4,$7B,$12,$FE
	.dc.b	$03,$FC,$EB,$6F,$F6,$2C,$60,$87,$BE,$34,$71,$70,$6A,$FF,$9A,$CF
	.dc.b	$28,$FC,$97,$D5,$3E,$AF,$FA,$3B,$63,$F8,$7B,$5F,$CC,$7D,$8B,$E1
	.dc.b	$77,$E7,$C9,$DB,$FD,$42,$78,$63,$FD,$13,$EB,$1F,$3F,$FD,$51,$B7
	.dc.b	$FC,$71,$AF,$E9,$7E,$C7,$EC,$57,$1F,$88,$E0,$49,$FE,$47,$2F,$7F
	.dc.b	$FA,$FF,$D6,$FF,$43,$BF,$EA,$6D,$E9,$FE,$F4,$DB,$FF,$3B,$DB,$E9
	.dc.b	$78,$FC,$BC,$9F,$E5,$F4,$8F,$ED,$75,$FD,$52,$FF,$C0,$9F,$28,$FA
	.dc.b	$DF,$ED,$ED,$C3,$F9,$BB,$5F,$E0,$1F,$63,$FC,$65,$FE,$05,$1F,$2D
	.dc.b	$DF,$17,$23,$FB,$CC,$FC,$4D,$3F,$B6,$FD,$72,$FF,$BE,$B6,$FF,$A2
	.dc.b	$B5,$EF,$E2,$6B,$E1,$24,$69,$FF,$02,$FB,$4F,$73,$9F,$CB,$4A,$E6
	.dc.b	$EB,$C9,$FF,$05,$FD,$73,$E0,$A9,$2B,$FA,$AE,$1E,$FD,$7F,$C0,$7F
	.dc.b	$5C,$F9,$3F,$F0,$9B,$6B,$FA,$DB,$5F,$E4,$8F,$8D,$FE,$BF,$F3,$8F
	.dc.b	$AE,$FF,$0D,$B6,$33,$18,$F8,$A9,$1C,$1A,$FF,$2F,$FA,$CF,$C3,$3D
	.dc.b	$3D,$6E,$64,$7F,$B6,$FD,$73,$E2,$FF,$C5,$2D,$8C,$C6,$0F,$F1,$6D
	.dc.b	$3E,$BA,$99,$9D,$9F,$DE,$23,$4F,$F7,$AF,$AE,$FC,$03,$E5,$AE,$67
	.dc.b	$F8,$E2,$33,$B7,$F8,$B1,$BD,$9F,$E5,$36,$D7,$FB,$41,$AA,$FF,$38
	.dc.b	$F7,$EB,$E8,$47,$C9,$E0,$77,$E7,$91,$E5,$7D,$BE,$E6,$23,$FA,$C6
	.dc.b	$1F,$55,$B3,$A7,$FC,$2B,$F5,$DF,$E5,$7C,$FF,$C8,$6D,$CF,$FC,$D4
	.dc.b	$D9,$FF,$73,$AE,$AF,$F2,$BB,$98,$34,$91,$FE,$5C,$F3,$5F,$1F,$49
	.dc.b	$5F,$E2,$6F,$AE,$E9,$8F,$FB,$D3,$EB,$BF,$97,$4F,$FB,$DA,$F8,$BF
	.dc.b	$E2,$C6,$BF,$DD,$5B,$3F,$E5,$CA,$E7,$2A,$E6,$A3,$31,$8A,$E6,$DB
	.dc.b	$FF,$45,$3C,$27,$73,$FE,$8D,$6D,$79,$6B,$73,$FE,$91,$39,$9B,$C3
	.dc.b	$FC,$49,$3E,$38,$C6,$1E,$F7,$85,$72,$9E,$86,$41,$E9,$AE,$78,$79
	.dc.b	$5B,$5C,$A7,$81,$FD,$5E,$BA,$3C,$6F,$38,$7A,$A3,$47,$0A,$F4,$BE
	.dc.b	$08,$D1,$F3,$1F,$F4,$23,$5F,$59,$FE,$87,$3B,$15,$FE,$97,$F0,$D9
	.dc.b	$8F,$A5,$E5,$A3,$13,$A9,$E1,$39,$78,$0F,$8D,$AD,$32,$E0,$8F,$CF
	.dc.b	$3F,$D1,$5D,$BF,$EA,$5D,$3F,$07,$1F,$32,$7E,$3E,$5F,$EC,$5D,$BE
	.dc.b	$4C,$E4,$E4,$7F,$97,$6F,$96,$7E,$FB,$E3,$18,$9F,$F5,$44,$62,$B9
	.dc.b	$C7,$9F,$0F,$C9,$E1,$DE,$3A,$9D,$E3,$37,$CE,$3F,$DB,$97,$CD,$D9
	.dc.b	$CF,$0F,$F6,$AC,$EE,$39,$46,$30,$F0,$D7,$1B,$F8,$35,$C4,$F8,$B3
	.dc.b	$A7,$DF,$87,$96,$77,$5C,$6F,$F2,$7F,$52,$E1,$FF,$12,$63,$FE,$AB
	.dc.b	$8C,$4F,$FB,$4F,$0F,$F6,$A4,$6F,$F1,$EC,$57,$2B,$E3,$97,$F9,$D2
	.dc.b	$B8,$CF,$3F,$C5,$3F,$C0,$0F,$5A,$DE,$1D,$27,$37,$C7,$47,$CA,$76
	.dc.b	$86,$3F,$F0,$07,$87,$0A,$B7,$97,$A5,$FE,$2E,$DB,$DC,$62,$FB,$46
	.dc.b	$2F,$9F,$E2,$8F,$9C,$BD,$F8,$6E,$3F,$16,$74,$D9,$D3,$4D,$4F,$0C
	.dc.b	$3E,$25,$70,$6D,$7A,$BF,$14,$FF,$5B,$5F,$D1,$6C,$FC,$B7,$FD,$6D
	.dc.b	$19,$F8,$9E,$07,$C2,$FC,$1F,$FC,$2D,$19,$3E,$D3,$8A,$FA,$8D,$DF
	.dc.b	$ED,$B6,$7F,$07,$7E,$09,$CF,$9B,$F3,$BF,$14,$FF,$6D,$D7,$E1,$57
	.dc.b	$D0,$E7,$E3,$74,$C9,$FC,$57,$E2,$8F,$81,$C6,$9F,$85,$8C,$4F,$E3
	.dc.b	$FB,$9F,$27,$F7,$1B,$1C,$AF,$13,$F8,$0F,$C5,$3C,$78,$7E,$3D,$BA
	.dc.b	$FE,$59,$F2,$DF,$E6,$DC,$BF,$A0,$F8,$9F,$F2,$D3,$F9,$F5,$F5,$7F
	.dc.b	$8C,$73,$7F,$9B,$73,$3F,$A5,$FC,$51,$F3,$BF,$92,$5E,$85,$8C,$3F
	.dc.b	$96,$FC,$53,$FB,$85,$7E,$B5,$FF,$3A,$B7,$6D,$78,$9F,$9C,$FC,$53
	.dc.b	$F8,$66,$DF,$B0,$5F,$3D,$9E,$FD,$3F,$A2,$F3,$FF,$DB,$98,$FE,$E5
	.dc.b	$BB,$FE,$CD,$7D,$E6,$FD,$BF,$D4,$FE,$29,$FE,$3F,$BF,$E0,$0D,$BF
	.dc.b	$DD,$3F,$85,$7F,$62,$E7,$3F,$E0,$7F,$8A,$7F,$BD,$EB,$F8,$42,$FA
	.dc.b	$9F,$DA,$B9,$BF,$E0,$8E,$67,$F8,$77,$BB,$FE,$A3,$6F,$CA,$7E,$F2
	.dc.b	$31,$5F,$C2,$DF,$F7,$14,$FF,$15,$FC,$53,$FD,$C2,$DF,$C0,$7F,$81
	.dc.b	$C6,$2B,$F8,$92,$8F,$E0,$FF,$8A,$7F,$93,$EB,$F8,$F3,$FE,$6D,$6E
	.dc.b	$FA,$F8,$0F,$E1,$BF,$8A,$7D,$29,$FE,$4E,$DC,$7F,$89,$BF,$52,$3F
	.dc.b	$94,$39,$D9,$F6,$3E,$19,$FF,$00,$C7,$F2,$B8,$F8,$8D,$CB,$F8,$F3
	.dc.b	$C3,$4B,$3C,$70,$4F,$19,$CD,$69,$B7,$D8,$EF,$7D,$9B,$3B,$7F,$9F
	.dc.b	$7E,$29,$9C,$3F,$96,$B7,$CF,$3F,$5E,$3F,$9E,$39,$26,$1F,$CC,$9D
	.dc.b	$63,$FD,$19,$E5,$F8,$A7,$F3,$EC,$3F,$A7,$B6,$FF,$9D,$AF,$4F,$F9
	.dc.b	$DB,$9C,$4D,$24,$6B,$D4,$FE,$89,$4F,$B4,$FE,$4A,$FC,$6C,$3F,$C5
	.dc.b	$EE,$7E,$0B,$FA,$57,$E0,$2F,$A6,$7D,$87,$2C,$38,$EC,$C7,$8C,$61
	.dc.b	$FF,$1C,$DF,$FB,$B3,$E2,$38,$63,$D3,$C1,$FF,$58,$57,$D1,$CF,$FC
	.dc.b	$7F,$87,$15,$87,$19,$D4,$79,$5E,$11,$98,$FE,$D5,$D6,$FB,$0F,$EB
	.dc.b	$EF,$68,$CC,$63,$2C,$B6,$FF,$BD,$57,$16,$CF,$F6,$8F,$C4,$BE,$71
	.dc.b	$F4,$63,$FC,$00,$E3,$83,$5C,$1B,$6C,$E9,$FC,$BF,$D9,$3E,$BA,$FD
	.dc.b	$AE,$1B,$5F,$0F,$85,$5B,$8C,$2A,$E1,$18,$6E,$BF,$6C,$FD,$0C,$EE
	.dc.b	$78,$DF,$69,$FF,$08,$64,$F8,$6F,$71,$EF,$F4,$6D,$FB,$A2,$D7,$09
	.dc.b	$B9,$7E,$06,$3E,$BF,$9B,$FC,$7E,$FF,$02,$FF,$23,$E1,$39,$9D,$FA
	.dc.b	$DF,$8E,$8C,$4E,$A3,$1F,$19,$97,$5F,$53,$BF,$AA,$B8,$FE,$0B,$04
	.dc.b	$63,$1F,$85,$8F,$82,$31,$97,$9E,$31,$7C,$70,$E8,$D3,$A5,$7E,$07
	.dc.b	$E0,$3F,$05,$EB,$59,$FC,$0B,$71,$F9,$6D,$C7,$FC,$83,$BF,$F7,$49
	.dc.b	$F9,$E7,$E2,$FE,$07,$F1,$8D,$BF,$C5,$B4,$DB,$F8,$E8,$FF,$1C,$FE
	.dc.b	$73,$FA,$5F,$00,$BE,$4B,$3D,$DA,$6D,$FF,$95,$67,$4D,$9D,$39,$FA
	.dc.b	$1E,$A7,$6F,$81,$B1,$92,$65,$FE,$57,$BF,$F1,$04,$7F,$99,$15,$79
	.dc.b	$BE,$13,$CF,$F0,$76,$BC,$4D,$BE,$4D,$DB,$FC,$CF,$82,$3D,$FD,$7F
	.dc.b	$03,$E6,$EE,$1E,$B6,$E1,$E3,$6E,$1C,$9E,$13,$F0,$7A,$ED,$7A,$5F
	.dc.b	$96,$48,$E8,$6C,$FE,$1F,$C5,$FD,$7E,$7F,$D6,$5F,$4D,$FC,$9A,$31
	.dc.b	$F1,$28,$C7,$9F,$FA,$38,$C5,$7F,$B6,$AB,$75,$C7,$CF,$D3,$70,$FF
	.dc.b	$5A,$3C,$9B,$5C,$67,$6F,$FB,$93,$EC,$DF,$4F,$FE,$80,$8F,$29,$97
	.dc.b	$18,$F8,$A3,$FC,$8F,$5C,$95,$72,$E7,$EF,$18,$BE,$38,$FF,$44,$AD
	.dc.b	$4F,$63,$9E,$3C,$23,$38,$F7,$6D,$71,$BE,$77,$EB,$BE,$11,$88,$F0
	.dc.b	$61,$83,$C7,$97,$C8,$8F,$64,$65,$F1,$61,$D9,$E8,$62,$BA,$46,$63
	.dc.b	$A9,$E9,$AE,$8E,$31,$F4,$57,$03,$C9,$3A,$AE,$44,$8C,$E5,$F2,$E3
	.dc.b	$18,$79,$6F,$E8,$5F,$1B,$E9,$39,$CB,$F0,$18,$F9,$E3,$0F,$B8,$F3
	.dc.b	$63,$D6,$B8,$C6,$CC,$E5,$C1,$F6,$9F,$DE,$8A,$7D,$1A,$C3,$D4,$D6
	.dc.b	$5F,$E7,$4F,$B3,$CC,$7E,$13,$E0,$11,$89,$D4,$F2,$6B,$CA,$76,$FD
	.dc.b	$46,$E9,$F8,$36,$E7,$F8,$BF,$A2,$FF,$7E,$BF,$8D,$8C,$D7,$E7,$5A
	.dc.b	$F8,$F0,$FC,$75,$70,$8D,$47,$E4,$67,$78,$79,$B0,$FC,$95,$EE,$F9
	.dc.b	$1E,$5B,$E0,$74,$69,$92,$4F,$27,$2F,$AE,$2E,$1E,$B9,$A7,$90,$E2
	.dc.b	$E5,$A6,$67,$93,$F9,$AA,$F3,$5F,$A6,$BE,$25,$F4,$AF,$DB,$D6,$B0
	.dc.b	$F7,$92,$7D,$53,$D8,$F7,$E1,$D8,$F5,$39,$C3,$B3,$E3,$C3,$CD,$3A
	.dc.b	$23,$83,$F8,$17,$C9,$79,$B1,$95,$E0,$AF,$E0,$13,$F5,$7A,$FF,$AA
	.dc.b	$F7,$7F,$93,$39,$3F,$9E,$B7,$6F,$E9,$0D,$C1,$5F,$D7,$3B,$F0,$D8
	.dc.b	$FE,$B5,$38,$E8,$FE,$FF,$1F,$D7,$6B,$1E,$6E,$59,$1F,$D8,$63,$2E
	.dc.b	$C8,$7F,$5B,$7A,$7B,$BD,$7E,$02,$4F,$F6,$98,$CB,$FD,$AE,$32,$FF
	.dc.b	$63,$8F,$27,$8B,$23,$FB,$83,$E7,$F7,$F1,$1F,$DA,$62,$EB,$FD,$82
	.dc.b	$7D,$9C,$3E,$DF,$1E,$5F,$F7,$98,$CB,$FD,$EE,$32,$FF,$73,$8F,$89
	.dc.b	$EF,$7C,$5F,$7E,$4F,$F8,$06,$32,$FF,$80,$E3,$2F,$F7,$B8,$F9,$FE
	.dc.b	$87,$F9,$FA,$3F,$C1,$1C,$FF,$3C,$CF,$A7,$23,$24,$3F,$CE,$2F,$DB
	.dc.b	$5F,$5F,$FD,$55,$FF,$2A,$18,$F2,$B5,$06,$F6,$49,$FF,$3F,$CE,$5B
	.dc.b	$39,$79,$C7,$5D,$BF,$CA,$51,$D4,$D9,$CA,$BF,$1D,$E5,$F9,$3E,$16
	.dc.b	$DB,$E2,$69,$88,$97,$FE,$8E,$9D,$36,$72,$FF,$A4,$3B,$7E,$49,$1F
	.dc.b	$F4,$5F,$65,$74,$7F,$A3,$A3,$E7,$C7,$C7,$F8,$67,$D5,$DF,$AC,$97
	.dc.b	$FE,$A0,$9C,$B6,$72,$FC,$C8,$4F,$39,$93,$C6,$37,$AF,$C6,$AE,$73
	.dc.b	$FE,$87,$EE,$74,$3E,$67,$F6,$2B,$EA,$A9,$88,$95,$FE,$B9,$AE,$33
	.dc.b	$A8,$C5,$73,$75,$EE,$BE,$AF,$F7,$AB,$E8,$E9,$88,$93,$C2,$32,$F2
	.dc.b	$8F,$C7,$79,$0F,$0B,$F2,$1F,$56,$F3,$29,$A8,$97,$FE,$CA,$9D,$36
	.dc.b	$72,$FE,$5A,$BA,$3F,$ED,$06,$B9,$AE,$FF,$37,$FC,$5B,$E0,$BF,$51
	.dc.b	$3F,$F0,$03,$B3,$07,$F1,$AF,$C4,$FF,$B2,$C6,$4F,$91,$3C,$38,$FF
	.dc.b	$D5,$5C,$28,$C7,$94,$AF,$E4,$B7,$CA,$F7,$F0,$EF,$63,$F9,$3B,$6B
	.dc.b	$EA,$35,$FB,$CF,$FA,$1D,$CC,$70,$23,$C3,$D9,$FD,$D7,$D6,$3C,$6B
	.dc.b	$E3,$CA,$FF,$79,$CE,$5B,$39,$7F,$7F,$1F,$E5,$1C,$7F,$6A,$EE,$33
	.dc.b	$3B,$9C,$D7,$82,$3F,$A0,$FC,$4B,$EC,$69,$2F,$F2,$1F,$89,$FF,$00
	.dc.b	$C3,$E4,$FA,$BF,$9A,$9F,$CA,$2F,$37,$FA,$5B,$63,$3D,$9F,$D3,$BF
	.dc.b	$13,$7B,$37,$DA,$95,$F9,$D9,$CB,$F9,$FF,$C4,$EB,$FA,$97,$FD,$E8
	.dc.b	$E6,$7F,$AC,$C6,$B2,$FB,$BD,$4F,$86,$FE,$BA,$BE,$1A,$98,$8F,$EC
	.dc.b	$3F,$89,$B5,$A4,$CB,$F6,$11,$FA,$D7,$F7,$5F,$89,$FE,$71,$BE,$07
	.dc.b	$F6,$A8,$C6,$5F,$8E,$7A,$4F,$EE,$7F,$13,$7C,$D7,$F7,$05,$F6,$94
	.dc.b	$C4,$7F,$74,$F6,$DE,$1D,$CA,$61,$FC,$0E,$72,$FF,$05,$F4,$7E,$83
	.dc.b	$FC,$3D,$7B,$7F,$DF,$9B,$1B,$E7,$FF,$00,$F8,$EF,$6E,$DC,$AF,$E2
	.dc.b	$13,$97,$F8,$97,$E1,$9F,$8F,$F1,$7F,$51,$E3,$A3,$13,$FE,$0E,$6C
	.dc.b	$67,$DC,$FF,$07,$FE,$26,$F5,$69,$2F,$F8,$F7,$07,$F8,$41,$FE,$51
	.dc.b	$F8,$9F,$F0,$4E,$6F,$DC,$CE,$5B,$F0,$1F,$E1,$A9,$D6,$9E,$45,$C5
	.dc.b	$9E,$C9,$8F,$F2,$B9,$CB,$FC,$B7,$F1,$27,$B1,$FE,$24,$6C,$66,$3F
	.dc.b	$C4,$FF,$89,$3E,$7A,$9B,$89,$3F,$CD,$E7,$4D,$9D,$B4,$D7,$DC,$FF
	.dc.b	$8C,$81,$7E,$6E,$47,$F4,$18,$CB,$5C,$BF,$D3,$54,$73,$64,$71,$8D
	.dc.b	$ED,$FC,$8A,$79,$CF,$F5,$5F,$A9,$FE,$AF,$3F,$C9,$9F,$EB,$F1,$FD
	.dc.b	$45,$AE,$FD,$DF,$F2,$4C,$70,$DF,$F0,$6E,$AB,$97,$BB,$AA,$FF,$27
	.dc.b	$E4,$F9,$5F,$F2,$8F,$8A,$F8,$DA,$6A,$26,$7F,$D8,$6B,$83,$63,$9F
	.dc.b	$D6,$7F,$88,$F3,$FC,$FE,$7F,$CE,$FE,$AB,$F8,$15,$FF,$3D,$CB,$FC
	.dc.b	$3C,$A3,$2F,$F0,$7F,$86,$FE,$84,$E6,$6D,$F4,$8C,$1F,$D1,$DC,$57
	.dc.b	$29,$EA,$DA,$E6,$F6,$9D,$CF,$4E,$F6,$1E,$78,$F0,$9C,$8B,$7C,$9E
	.dc.b	$11,$BC,$3C,$55,$C9,$F2,$9E,$B3,$45,$33,$7E,$27,$8C,$EA,$BC,$11
	.dc.b	$F1,$EF,$C0,$E8,$E3,$7D,$F0,$F7,$47,$19,$DA,$C3,$B3,$6B,$85,$FB
	.dc.b	$23,$81,$E6,$9D,$9F,$3D,$FF,$77,$35,$F8,$27,$FB,$B9,$C3,$F4,$F9
	.dc.b	$DF,$57,$FE,$EE,$9F,$E0,$2D,$8C,$CF,$16,$BF,$EF,$4F,$8C,$FF,$7A
	.dc.b	$BF,$E0,$36,$A6,$91,$F9,$F9,$FF,$61,$E2,$F9,$F1,$C4,$CF,$86,$8C
	.dc.b	$C7,$33,$2E,$2B,$EB,$D7,$39,$FB,$67,$3D,$7F,$2F,$19,$7F,$33,$19
	.dc.b	$CF,$FD,$85,$87,$78,$E2,$77,$9E,$31,$97,$8C,$66,$78,$E1,$E3,$8F
	.dc.b	$29,$CB,$0F,$2D,$F1,$AF,$D3,$CF,$84,$FF,$15,$61,$E5,$8C,$C7,$53
	.dc.b	$33,$C5,$E3,$5D,$2F,$C1,$87,$99,$FC,$F7,$80,$FA,$BF,$EF,$91,$FE
	.dc.b	$63,$6D,$FF,$9B,$5F,$EF,$D1,$9C,$7E,$16,$5F,$E4,$09,$DB,$63,$33
	.dc.b	$FD,$6A,$78,$D7,$39,$CB,$FE,$E8,$C9,$F1,$38,$C7,$FC,$41,$C1,$67
	.dc.b	$38,$FF,$9D,$DF,$F5,$83,$6B,$53,$B6,$C7,$F9,$F7,$EA,$9F,$27,$D5
	.dc.b	$FE,$49,$19,$8C,$73,$65,$B1,$93,$94,$65,$A6,$72,$FF,$49,$57,$47
	.dc.b	$FD,$29,$F5,$2F,$96,$F6,$36,$BF,$D2,$0F,$F9,$85,$B1,$A3,$E1,$7D
	.dc.b	$57,$FA,$72,$32,$7F,$A5,$DC,$57,$FA,$85,$BC,$1F,$E7,$26,$CF,$F8
	.dc.b	$2A,$FC,$15,$BF,$63,$E7,$7D,$57,$FB,$26,$BF,$D4,$4F,$F8,$A9,$BF
	.dc.b	$1B,$AF,$42,$47,$FA,$9F,$EA,$9F,$39,$8C,$7F,$D6,$51,$89,$D4,$FD
	.dc.b	$BE,$C7,$CC,$7D,$DA,$D5,$FF,$AE,$67,$37,$FE,$B3,$FA,$AF,$24,$FF
	.dc.b	$AF,$5B,$BF,$FA,$39,$F3,$46,$73,$FC,$47,$C8,$BE,$DA,$FC,$9E,$1B
	.dc.b	$8F,$FA,$FF,$E7,$9F,$37,$F8,$F6,$C6,$7C,$9C,$4F,$E4,$FE,$AF,$FC
	.dc.b	$7F,$3F,$EC,$76,$E7,$FE,$D1,$7F,$D7,$51,$9D,$BF,$35,$F5,$7F,$C3
	.dc.b	$F7,$FF,$52,$BF,$E4,$39,$FF,$68,$39,$7F,$C7,$EE,$6F,$FD,$B9,$ED
	.dc.b	$BE,$6F,$F4,$8D,$AF,$F6,$C3,$FE,$9D,$6C,$7E,$A7,$EA,$DE,$CD,$31
	.dc.b	$D3,$6B,$FD,$C1,$F5,$6F,$A8,$FD,$82,$F9,$EB,$18,$7F,$68,$FF,$8A
	.dc.b	$A7,$FD,$D0,$DC,$FF,$DE,$4B,$E4,$37,$C5,$FB,$BF,$AB,$C5,$FF,$BB
	.dc.b	$DB,$7F,$EF,$97,$C5,$19,$AF,$E0,$5F,$1F,$FE,$E3,$8F,$F7,$B3,$FE
	.dc.b	$8B,$6E,$FA,$F8,$8F,$E1,$1F,$29,$F2,$F1,$79,$68,$C3,$7E,$13,$7E
	.dc.b	$67,$F1,$2E,$67,$D5,$BE,$AF,$0F,$93,$1F,$F0,$5E,$9F,$CC,$A3,$4F
	.dc.b	$F3,$43,$1C,$3C,$2B,$8B,$63,$55,$FE,$14,$AE,$97,$FC,$A9,$E7,$1D
	.dc.b	$8D,$57,$17,$13,$E6,$8F,$E4,$DF,$31,$CA,$FF,$D0,$ED,$AF,$82,$FD
	.dc.b	$B8,$FF,$7B,$CF,$02,$3F,$EF,$AE,$3B,$5E,$1D,$1F,$3B,$FC,$F2,$BF
	.dc.b	$D2,$6D,$9F,$9A,$FF,$3C,$C7,$FE,$00,$B9,$AE,$3E,$8F,$F3,$B8,$C1
	.dc.b	$BB,$E7,$1F,$D0,$AD,$F5,$5F,$C8,$DF,$84,$63,$E6,$7F,$A3,$FD,$0B
	.dc.b	$EA,$DF,$3B,$C6,$F8,$69,$FD,$D7,$0E,$19,$7D,$06,$9F,$37,$0E,$57
	.dc.b	$C7,$19,$5F,$5A,$7F,$D3,$BE,$F6,$2F,$87,$B7,$FE,$6D,$AF,$B7,$B7
	.dc.b	$E9,$FE,$AF,$FA,$C7,$06,$6F,$E5,$28,$CE,$1A,$6D,$F2,$9D,$36,$7F
	.dc.b	$B1,$7D,$4F,$CB,$3E,$45,$FE,$71,$C7,$CA,$EB,$87,$6D,$BF,$2C,$F0
	.dc.b	$DD,$F5,$6F,$A9,$FD,$02,$F9,$9B,$18,$6D,$FE,$62,$B8,$C6,$32,$FC
	.dc.b	$9D,$F1,$AD,$3D,$5B,$3B,$AF,$F0,$3C,$EE,$BF,$1D,$83,$EA,$FD,$E7
	.dc.b	$C0,$F5,$6E,$3E,$A8,$B5,$C2,$6E,$1F,$E3,$3E,$27,$9D,$FC,$83,$0F
	.dc.b	$B4,$FF,$62,$B4,$EB,$C8,$FC,$5C,$E6,$31,$5C,$3C,$8C,$D6,$9D,$FC
	.dc.b	$77,$0F,$8F,$B5,$D9,$B8,$FE,$37,$14,$63,$2F,$34,$62,$BC,$B7,$CD
	.dc.b	$A7,$33,$EC,$FC,$0F,$E0,$BD,$8F,$C1,$B6,$32,$7D,$E8,$CB,$4F,$C0
	.dc.b	$E3,$F8,$AF,$81,$FC,$5B,$63,$27,$E1,$A3,$2D,$3F,$3B,$F5,$3F,$AC
	.dc.b	$53,$FC,$B2,$1F,$45,$FE,$37,$75,$19,$8C,$76,$E5,$B5,$D1,$E1,$F5
	.dc.b	$72,$3B,$FC,$0D,$BF,$E2,$58,$2B,$1B,$7F,$9C,$7B,$93,$FE,$1C,$8C
	.dc.b	$3B,$C3,$D3,$F0,$9E,$5F,$84,$F3,$36,$32,$7B,$A3,$2D,$3E,$0F,$57
	.dc.b	$B7,$E7,$FD,$CD,$8C,$9F,$1A,$32,$D2,$67,$EA,$EC,$E0,$D8,$CE,$5E
	.dc.b	$89,$CB,$63,$F8,$C7,$0F,$F5,$FA,$FF,$53,$7D,$1F,$FB,$AA,$31,$F1
	.dc.b	$28,$C7,$A3,$FE,$95,$8C,$57,$D8,$AD,$D7,$1F,$47,$4D,$C3,$FD,$68
	.dc.b	$F2,$6D,$71,$9D,$BF,$82,$FA,$B7,$D3,$7F,$9E,$23,$CA,$65,$C6,$3E
	.dc.b	$28,$FF,$20,$D7,$25,$5C,$B9,$FB,$C6,$2F,$8E,$3F,$90,$8C,$4E,$B2
	.dc.b	$F5,$C7,$78,$E3,$7C,$E3,$C6,$70,$2C,$F0,$71,$8F,$96,$78,$37,$0F
	.dc.b	$0C,$FC,$1C,$BA,$C6,$23,$8E,$3E,$5B,$E2,$FC,$B8,$F0,$C6,$1F,$31
	.dc.b	$89,$E5,$87,$8E,$75,$3D,$0D,$E1,$F2,$67,$55,$E1,$AD,$9C,$23,$38
	.dc.b	$F9,$89,$7F,$03,$2F,$A1,$5C,$2F,$91,$E2,$8C,$B8,$AE,$38,$78,$67
	.dc.b	$31,$8C,$7E,$05,$71,$CB,$18,$FB,$8F,$F6,$CE,$3F,$4D,$B3,$F8,$76
	.dc.b	$DF,$FB,$69,$FC,$1F,$BC,$FA,$BC,$5F,$F9,$D9,$B1,$99,$FF,$70,$CE
	.dc.b	$5B,$19,$CF,$EE,$37,$2F,$AC,$D9,$FC,$D7,$CC,$FE,$36,$B7,$87,$E2
	.dc.b	$AB,$55,$E4,$8F,$C6,$D6,$EF,$9C,$6C,$93,$F8,$09,$D3,$FA,$7A,$E7
	.dc.b	$87,$9F,$0E,$F3,$B3,$8B,$82,$61,$F8,$DB,$E7,$3E,$33,$F1,$B7,$D0
	.dc.b	$EB,$1A,$BE,$D5,$E6,$AD,$E1,$EA,$7B,$C6,$70,$EF,$1F,$10,$C9,$E9
	.dc.b	$9A,$FE,$E2,$FB,$61,$E1,$9C,$E0,$9C,$C6,$27,$CB,$56,$37,$3E,$72
	.dc.b	$47,$94,$FD,$4F,$CB,$61,$F8,$C4,$C3,$F7,$11,$F3,$B9,$FF,$A1,$F6
	.dc.b	$FF,$07,$E2,$FE,$87,$DF,$78,$3F,$AC,$D7,$F4,$88,$C6,$3F,$D6,$EE
	.dc.b	$CF,$28,$FE,$C3,$38,$E9,$FE,$F7,$1F,$D9,$2B,$1E,$66,$59,$1F,$DA
	.dc.b	$23,$2E,$C8,$7F,$5C,$7A,$7A,$DD,$7E,$02,$4F,$F7,$08,$CB,$FD,$CA
	.dc.b	$32,$FF,$64,$8F,$27,$87,$23,$FB,$BB,$E7,$F5,$F1,$1F,$DA,$A2,$EB
	.dc.b	$FD,$9A,$7D,$9C,$9E,$DF,$16,$5F,$F8,$02,$32,$FF,$80,$A3,$2F,$F7
	.dc.b	$48,$F8,$9E,$FF,$C5,$FA,$69,$3F,$E0,$88,$CB,$FE,$0A,$8C,$BF,$DF
	.dc.b	$23,$E7,$FA,$1F,$E7,$A8,$FF,$07,$F3,$7C,$F3,$3E,$94,$8C,$90,$FF
	.dc.b	$36,$BF,$69,$7C,$F7,$F0,$DD,$3F,$D0,$4B,$C9,$D3,$67,$FD,$08,$EF
	.dc.b	$9B,$FA,$73,$63,$FC,$3E,$ED,$FF,$49,$B6,$BA,$F6,$74,$3A,$DC,$AE
	.dc.b	$B1,$F8,$EF,$2F,$CB,$F0,$B7,$07,$C3,$D3,$11,$2F,$FD,$25,$3A,$6C
	.dc.b	$E5,$FF,$46,$F6,$FC,$A2,$3F,$E8,$BE,$CA,$E8,$FF,$47,$47,$D0,$8F
	.dc.b	$91,$F0,$EF,$95,$BF,$2D,$2F,$FD,$51,$39,$6C,$E5,$FF,$5A,$C2,$79
	.dc.b	$CC,$9E,$31,$BD,$BF,$D6,$95,$CE,$7F,$D9,$DD,$0F,$AB,$FD,$8A,$FA
	.dc.b	$BA,$62,$24,$FF,$AC,$27,$4D,$9C,$BF,$ED,$9F,$75,$F5,$5F,$BC,$5F
	.dc.b	$57,$4C,$44,$9F,$F5,$DC,$E9,$B3,$97,$FD,$A7,$E4,$3C,$2F,$C8,$7D
	.dc.b	$13,$CC,$A7,$E4,$E7,$55,$FE,$D1,$6B,$CA,$B4,$72,$FA,$0B,$1B,$8A
	.dc.b	$FF,$B8,$AD,$3F,$C5,$BE,$0F,$F5,$13,$FE,$F7,$6F,$07,$F9,$95,$BD
	.dc.b	$7F,$C8,$3E,$AF,$FB,$45,$6F,$B7,$91,$1F,$F6,$DC,$6A,$FF,$CD,$DF
	.dc.b	$0E,$F6,$7F,$92,$B6,$BF,$DC,$2D,$53,$C3,$D9,$DD,$49,$DB,$CA,$3F
	.dc.b	$71,$F1,$2F,$16,$F8,$D2,$FF,$DF,$13,$96,$CE,$5E,$FA,$7F,$91,$67
	.dc.b	$FC,$9D,$8F,$FC,$01,$78,$4E,$8F,$F4,$CB,$88,$D5,$F3,$78,$5F,$F4
	.dc.b	$2F,$89,$7D,$5D,$26,$1F,$F0,$03,$EA,$D9,$7F,$48,$F5,$5D,$77,$CC
	.dc.b	$97,$FE,$A8,$CF,$FD,$31,$5A,$8C,$5F,$F5,$4F,$AB,$7B,$3F,$D4,$9B
	.dc.b	$3A,$1E,$A2,$33,$1E,$D8,$C7,$B3,$FD,$6B,$EA,$F3,$97,$FA,$9D,$B8
	.dc.b	$FF,$62,$6D,$F4,$55,$FD,$8B,$AD,$F0,$D7,$E1,$A5,$F8,$F2,$FD,$14
	.dc.b	$FF,$68,$FA,$B6,$D4,$95,$B8,$FD,$F3,$C7,$EA,$FF,$50,$DF,$07,$FB
	.dc.b	$8C,$62,$3D,$8F,$47,$C1,$F5,$6F,$9A,$FE,$EE,$BE,$AE,$98,$8F,$EF
	.dc.b	$7E,$DF,$EA,$2B,$D0,$D7,$E4,$37,$8B,$FD,$98,$FF,$4A,$8F,$77,$A2
	.dc.b	$E6,$FF,$02,$B6,$33,$18,$F6,$BF,$C1,$3E,$31,$BE,$DC,$AF,$91,$39
	.dc.b	$7E,$57,$D5,$FE,$63,$A3,$FC,$2A,$D8,$FB,$ED,$E4,$2F,$73,$FC,$2F
	.dc.b	$E4,$FF,$85,$2B,$F0,$5F,$57,$FE,$31,$7E,$B3,$5F,$C3,$46,$F7,$FF
	.dc.b	$75,$7D,$5F,$F1,$47,$FD,$D9,$19,$CB,$FC,$4E,$DA,$FC,$8D,$6E,$7F
	.dc.b	$C5,$6B,$83,$FC,$52,$E6,$77,$D5,$89,$FF,$16,$FD,$5B,$D6,$A4,$AF
	.dc.b	$C6,$7D,$5B,$EA,$BF,$C6,$6D,$8F,$C8,$57,$16,$D6,$AA,$47,$26,$BF
	.dc.b	$A5,$A7,$CD,$7F,$8F,$8B,$1F,$9B,$5F,$57,$93,$4C,$8D,$46,$5E,$11
	.dc.b	$97,$F6,$4F,$27,$66,$8F,$D4,$7D,$57,$F6,$1D,$FF,$84,$C7,$EF,$23
	.dc.b	$FB,$53,$AF,$6B,$FC,$A7,$1B,$DF,$FC,$B2,$D9,$E1,$3A,$37,$ED,$FF
	.dc.b	$96,$B3,$7C,$AD,$25,$7E,$E6,$1F,$57,$88,$FE,$15,$1C,$1D,$4F,$39
	.dc.b	$FF,$17,$60,$E9,$FE,$7F,$4F,$AA,$FE,$0E,$8F,$E1,$3F,$09,$FE,$F9
	.dc.b	$8C,$C6,$0D,$38,$AE,$9A,$E3,$C6,$FF,$80,$11,$8A,$ED,$1D,$5E,$46
	.dc.b	$0F,$03,$FE,$37,$AE,$98,$79,$9B,$7D,$6B,$78,$79,$6F,$A1,$C4,$F0
	.dc.b	$DF,$16,$D7,$1A,$F0,$DF,$B2,$78,$47,$C1,$C3,$D1,$3E,$08,$F3,$1C
	.dc.b	$27,$BE,$1E,$18,$E1,$87,$76,$E1,$DE,$BA,$C7,$13,$27,$4C,$3B,$C7
	.dc.b	$48,$E1,$7D,$54,$E9,$E2,$7C,$B3,$37,$F7,$1C,$47,$D1,$6C,$65,$FA
	.dc.b	$BC,$8F,$AB,$FF,$7D,$57,$F0,$36,$C6,$67,$B1,$E2,$70,$FF,$C0,$1F
	.dc.b	$8B,$CB,$FE,$F3,$28,$6B,$4C,$BF,$4F,$2F,$E6,$DC,$3F,$56,$7F,$25
	.dc.b	$39,$F0,$3F,$5F,$AF,$E4,$A7,$27,$D4,$BE,$2F,$D6,$33,$1F,$98,$8C
	.dc.b	$BC,$63,$C4,$E6,$7A,$EB,$EE,$BF,$06,$1E,$3B,$DC,$FF,$B5,$6B,$71
	.dc.b	$F4,$6B,$6F,$84,$D9,$FE,$09,$8D,$BF,$16,$FC,$17,$D6,$B8,$C7,$95
	.dc.b	$B7,$E3,$AF,$75,$F2,$7D,$B1,$E3,$BF,$D2,$77,$3E,$AF,$FC,$9F,$1F
	.dc.b	$DF,$DB,$5F,$E0,$F7,$FD,$5F,$19,$EC,$FE,$37,$9F,$F1,$CB,$E5,$18
	.dc.b	$9F,$F7,$8F,$D6,$70,$C3,$94,$E5,$FF,$0E,$C6,$6F,$96,$4F,$90,$D6
	.dc.b	$3F,$DC,$DB,$0F,$98,$B1,$32,$FF,$0B,$4F,$16,$CE,$71,$FF,$22,$61
	.dc.b	$FE,$09,$F5,$BC,$4D,$AF,$F1,$83,$FE,$0D,$F5,$F3,$E4,$FF,$95,$E7
	.dc.b	$2D,$BE,$6D,$AF,$86,$AB,$8D,$CB,$E2,$F4,$8C,$5F,$F8,$EF,$EA,$9E
	.dc.b	$D7,$C0,$F2,$7F,$96,$61,$FE,$31,$9C,$BF,$E3,$38,$CD,$7F,$90,$DB
	.dc.b	$97,$D3,$6C,$FF,$35,$D9,$7C,$AF,$81,$DC,$F9,$FF,$89,$FF,$AA,$2B
	.dc.b	$B4,$63,$E0,$BE,$3F,$E5,$DF,$E8,$59,$F8,$EB,$E6,$73,$EF,$E9,$DE
	.dc.b	$B2,$B7,$17,$0F,$A7,$F9,$73,$F0,$DF,$BF,$DF,$F7,$6F,$E9,$CE,$5F
	.dc.b	$F2,$94,$66,$BF,$CC,$5F,$59,$FE,$43,$9F,$F2,$B3,$FE,$7A,$6F,$27
	.dc.b	$E2,$A3,$13,$FE,$69,$F8,$CF,$F7,$9D,$7F,$97,$5B,$19,$F5,$3F,$23
	.dc.b	$18,$E4,$FC,$F7,$CE,$FF,$82,$2F,$FC,$C8,$BE,$13,$F2,$EE,$70,$FF
	.dc.b	$3B,$39,$9F,$F3,$57,$D5,$BE,$63,$F3,$AF,$FA,$0A,$33,$18,$F8,$2F
	.dc.b	$D3,$7C,$9B,$F2,$D7,$C5,$E1,$B7,$FE,$87,$9F,$95,$F2,$D9,$AF,$D4
	.dc.b	$FC,$07,$E4,$EB,$FC,$F4,$FF,$92,$DB,$DF,$AF,$6F,$F5,$FF,$57,$FC
	.dc.b	$93,$87,$F7,$8F,$F9,$76,$BF,$D0,$4E,$5F,$C6,$39,$E0,$FF,$4A,$7D
	.dc.b	$5B,$E9,$7F,$72,$DA,$FF,$45,$3F,$E1,$66,$C7,$EF,$BE,$A1,$F5,$55
	.dc.b	$C3,$67,$FD,$25,$F5,$6F,$9D,$FE,$0A,$F5,$C7,$FD,$48,$E6,$77,$87
	.dc.b	$CA,$FC,$AB,$F6,$D1,$FE,$17,$E2,$46,$2D,$18,$DF,$87,$93,$FC,$4B
	.dc.b	$E4,$FF,$C4,$58,$7F,$B8,$BD,$4C,$78,$19,$7F,$85,$FD,$67,$E0,$78
	.dc.b	$FF,$92,$B6,$BF,$D8,$8F,$E0,$F3,$FE,$26,$E5,$FF,$75,$B6,$4D,$38
	.dc.b	$3D,$3D,$FE,$E4,$AF,$E3,$33,$A7,$BC,$7F,$2A,$F9,$6F,$8D,$5F,$CD
	.dc.b	$DF,$A5,$7F,$EC,$B7,$35,$FC,$1D,$CC,$EF,$7E,$5E,$E7,$36,$D7,$FB
	.dc.b	$17,$E6,$3F,$A3,$D7,$F4,$37,$F9,$E4,$7F,$11,$72,$DF,$8A,$FE,$7B
	.dc.b	$18,$C7,$FD,$BE,$F0,$6F,$AD,$A9,$FE,$91,$0F,$AB,$FD,$43,$FE,$FF
	.dc.b	$31,$E0,$7D,$2E,$7D,$CE,$B2,$FF,$38,$D6,$B2,$F8,$F9,$EB,$0F,$B2
	.dc.b	$D8,$C9,$FE,$F3,$9E,$07,$D8,$8F,$EA,$B8,$7F,$9B,$27,$FD,$CC,$D9
	.dc.b	$FF,$7B,$2F,$81,$DF,$C0,$4A,$DB,$CA,$31,$5C,$2F,$AE,$DD,$9B,$7D
	.dc.b	$9B,$3B,$8F,$0C,$FE,$4A,$A4,$7E,$13,$E0,$AD,$7E,$42,$6E,$7F,$96
	.dc.b	$9B,$5F,$96,$9B,$5F,$94,$9B,$D0,$FB,$17,$D0,$7F,$16,$7F,$93,$61
	.dc.b	$FA,$D7,$2F,$F1,$57,$37,$FA,$E7,$2B,$1F,$F5,$2D,$F5,$FA,$0E,$98
	.dc.b	$75,$7F,$91,$F9,$4F,$97,$A6,$3D,$CB,$9F,$F2,$39,$D3,$67,$4D,$3F
	.dc.b	$4F,$F4,$1F,$B8,$3F,$73,$F1,$56,$32,$4D,$3F,$6F,$19,$7F,$73,$37
	.dc.b	$4F,$DF,$CD,$AF,$E1,$DF,$19,$F8,$17,$F9,$B3,$6F,$E0,$46,$B7,$FE
	.dc.b	$7F,$39,$BF,$86,$D9,$E7,$C1,$AA,$DB,$FD,$1B,$E6,$7F,$9E,$61,$D0
	.dc.b	$FE,$36,$E3,$0F,$9A,$FC,$48,$D7,$02,$35,$87,$F2,$56,$CF,$2F,$27
	.dc.b	$B6,$BD,$DF,$D0,$F8,$F2,$FF,$43,$CF,$6F,$F4,$33,$72,$FF,$23,$4F
	.dc.b	$F3,$59,$CD,$7E,$A3,$83,$F5,$7E,$3F,$EB,$1B,$19,$3F,$5B,$19,$69
	.dc.b	$FB,$0F,$8B,$FD,$99,$FB,$4F,$02,$C6,$49,$7F,$B4,$8C,$BF,$B6,$FA
	.dc.b	$4F,$DD,$9F,$BC,$F8,$CB,$5F,$BC,$8B,$3F,$BC,$9B,$9F,$B3,$E7,$BF
	.dc.b	$81,$98,$BF,$F1,$86,$88,$CC,$7F,$06,$FC,$22,$E1,$FC,$19,$FE,$3B
	.dc.b	$F0,$CF,$98,$57,$FA,$73,$CC,$FF,$71,$4E,$63,$17,$FE,$F1,$87,$A3
	.dc.b	$C2,$3F,$D5,$55,$C2,$F1,$F1,$7F,$E5,$59,$E1,$18,$9D,$61,$D2,$7F
	.dc.b	$CE,$57,$97,$71,$9F,$A6,$F6,$7D,$25,$D0,$F1,$CF,$CD,$7E,$20,$2B
	.dc.b	$8B,$93,$15,$C9,$C4,$6B,$3F,$44,$7F,$A1,$73,$F2,$3C,$30,$F1,$11
	.dc.b	$E9,$87,$8A,$B8,$DF,$3C,$FC,$75,$C4,$EF,$8F,$86,$BB,$DF,$1A,$DD
	.dc.b	$F1,$BF,$5B,$5E,$55,$C3,$1F,$2C,$E4,$98,$75,$BF,$9F,$7E,$0C,$7D
	.dc.b	$18,$79,$AB,$53,$F4,$0D,$12,$35,$3F,$3F,$0E,$F5,$C2,$BC,$77,$E8
	.dc.b	$8C,$63,$EB,$9F,$4C,$6F,$1F,$2E,$1C,$F2,$F1,$E1,$C6,$2D,$F8,$F2
	.dc.b	$F0,$BE,$39,$D3,$F4,$32,$F1,$D7,$D3,$8F,$9E,$7C,$F2,$E1,$F8,$26
	.dc.b	$E3,$F8,$17,$3C,$0F,$5B,$8C,$7D,$08,$C3,$8A,$46,$23,$8C,$F0,$7F
	.dc.b	$21,$1C,$32,$FC,$1B,$72,$FB,$AD,$9F,$BB,$DF,$F8,$6A,$DD,$FE,$86
	.dc.b	$B5,$5D,$23,$F4,$95,$BB,$E7,$1B,$24,$FE,$72,$74,$FA,$E3,$37,$DA
	.dc.b	$34,$62,$B9,$BF,$A4,$BF,$0E,$0C,$39,$CF,$8C,$D6,$1C,$5F,$59,$4D
	.dc.b	$61,$DC,$FD,$2C,$F2,$AF,$97,$5B,$6D,$79,$0E,$A4,$AE,$58,$7E,$9A
	.dc.b	$F9,$61,$E6,$26,$1F,$1A,$FA,$46,$1E,$97,$C9,$F8,$87,$5B,$E8,$F1
	.dc.b	$36,$E1,$FE,$0F,$DF,$E9,$C2,$46,$EB,$F8,$0C,$FD,$0F,$88,$FF,$45
	.dc.b	$7D,$DF,$F0,$2E,$4F,$EA,$BE,$05,$E1,$FE,$72,$DC,$57,$C1,$FF,$1F
	.dc.b	$DE,$63,$11,$FE,$40,$8C,$72,$FF,$94,$30,$FE,$C3,$86,$09,$5F,$E4
	.dc.b	$8F,$82,$AF,$F9,$22,$2C,$FF,$68,$9E,$DE,$73,$E0,$69,$2F,$FC,$E5
	.dc.b	$F4,$15,$FF,$01,$66,$AE,$11,$FE,$77,$AE,$75,$D3,$C7,$EB,$F0,$32
	.dc.b	$BF,$CF,$91,$97,$FC,$FD,$19,$7A,$DA,$7C,$BF,$63,$27,$FD,$08,$FA
	.dc.b	$3C,$AC,$47,$08,$BA,$7F,$A1,$E7,$E7,$78,$EF,$A2,$6D,$86,$E1,$FE
	.dc.b	$DF,$6B,$FE,$1A,$5C,$3B,$99,$3B,$79,$47,$6F,$75,$F4,$B4,$95,$FE
	.dc.b	$EB,$87,$D2,$E2,$3F,$C4,$F1,$C1,$D4,$F3,$9F,$C3,$43,$F0,$27,$C4
	.dc.b	$E1,$B1,$FE,$8B,$6B,$FE,$99,$5F,$17,$B9,$93,$B7,$94,$62,$38,$45
	.dc.b	$DB,$FD,$3B,$5C,$A7,$FB,$7C,$7F,$9B,$3E,$75,$C7,$F3,$9B,$3A,$E7
	.dc.b	$FF,$72,$46,$23,$FC,$8A,$FE,$21,$B3,$C4,$8F,$13,$FD,$D3,$1B,$AF
	.dc.b	$F3,$DF,$CD,$BD,$EF,$BA,$BD,$3A,$62,$25,$F3,$E3,$56,$E3,$0D,$7F
	.dc.b	$CB,$2F,$57,$FD,$6D,$F1,$4F,$33,$E9,$B6,$33,$18,$F3,$F2,$C8,$C9
	.dc.b	$CA,$30,$FF,$B3,$67,$93,$67,$93,$FE,$E1,$E1,$FD,$AF,$C7,$C8,$C1
	.dc.b	$0F,$D3,$B9,$7F,$53,$1F,$C2,$7D,$3F,$F1,$CC,$7C,$BE,$FF,$F4,$F7
	.dc.b	$9B,$93,$5E,$65,$9D,$57,$07,$F7,$ED,$E0,$FF,$21,$5F,$FA,$9E,$BF
	.dc.b	$2B,$E5,$BE,$B7,$F2,$FB,$7E,$1F,$1F,$F7,$53,$70,$FF,$73,$D6,$E7
	.dc.b	$FD,$C4,$D7,$7D,$1C,$8E,$B7,$23,$6D,$7A,$BF,$C9,$FE,$03,$FA,$C4
	.dc.b	$7B,$D7,$8F,$93,$5E,$4C,$9D,$46,$5E,$11,$FC,$C7,$E0,$FC,$99,$7C
	.dc.b	$07,$F5,$AD,$EC,$D6,$32,$FE,$75,$6B,$AF,$FC,$AB,$BA,$31,$3F,$D0
	.dc.b	$1E,$B3,$FE,$48,$F8,$DF,$C1,$E3,$FE,$6F,$A4,$62,$75,$CB,$FC,$12
	.dc.b	$1C,$16,$25,$7E,$A4,$E2,$EE,$33,$1F,$EB,$B3,$26,$E7,$B3,$FE,$6C
	.dc.b	$7A,$BF,$89,$F0,$BF,$75,$87,$E0,$9F,$DA,$61,$EE,$6C,$FF,$5C,$77
	.dc.b	$1F,$E7,$6E,$97,$E4,$6F,$AD,$FE,$6A,$DB,$FD,$B3,$5F,$F6,$5B,$F1
	.dc.b	$9F,$EC,$FF,$51,$FB,$57,$FA,$9A,$F8,$EF,$ED,$0D,$8F,$D0,$3B,$CB
	.dc.b	$F3,$3F,$51,$72,$FA,$AE,$EB,$FA,$EB,$67,$7D,$FF,$DC,$FC,$07,$CC
	.dc.b	$52,$5F,$EB,$61,$F5,$FF,$DE,$7E,$9D,$F1,$37,$C0,$97,$FB,$39,$CB
	.dc.b	$FB,$5F,$A7,$7C,$5F,$F8,$05,$CC,$7F,$70,$6E,$9F,$B2,$77,$1F,$C7
	.dc.b	$7F,$20,$7D,$87,$F6,$06,$C6,$63,$1C,$9F,$E0,$DD,$FC,$F3,$FC,$15
	.dc.b	$7D,$56,$4D,$6D,$FE,$13,$DB,$FD,$4F,$9F,$F0,$25,$E2,$E4,$D7,$B5
	.dc.b	$FE,$19,$FA,$9C,$E5,$FC,$45,$AF,$F8,$29,$B7,$FC,$31,$7C,$4F,$F8
	.dc.b	$87,$CD,$7D,$75,$F9,$39,$87,$F1,$B9,$CB,$FD,$1F,$07,$E4,$0F,$AF
	.dc.b	$FF,$15,$7D,$63,$E1,$E9,$FE,$31,$AD,$CD,$F9,$A7,$C4,$D2,$63,$D6
	.dc.b	$3F,$D0,$6D,$F8,$AE,$1F,$0D,$23,$07,$F2,$FF,$AB,$DB,$FE,$28,$FA
	.dc.b	$27,$D5,$FF,$8F,$5B,$1B,$AF,$F2,$05,$70,$CF,$F9,$7B,$7C,$4F,$E1
	.dc.b	$6D,$AF,$F2,$4B,$FE,$16,$7C,$1E,$FF,$2B,$91,$83,$E3,$7E,$3F,$FE
	.dc.b	$A3,$C3,$FD,$F8,$1C,$61,$FD,$F9,$B2,$9D,$65,$F2,$67,$31,$C7,$E3
	.dc.b	$70,$66,$6D,$78,$22,$BC,$EF,$B4,$FF,$97,$FE,$B1,$F2,$D4,$95,$FD
	.dc.b	$66,$1F,$31,$88,$FF,$29,$C7,$07,$53,$CE,$7F,$CD,$5C,$2F,$8C,$FF
	.dc.b	$36,$7D,$57,$E9,$2F,$FC,$B6,$DD,$3F,$AE,$D6,$E7,$FB,$43,$5F,$F3
	.dc.b	$0B,$63,$75,$DB,$E6,$7F,$C6,$F9,$3E,$77,$E4,$60,$FA,$B6,$48,$8F
	.dc.b	$E0,$9D,$39,$77,$AD,$9E,$31,$C6,$FB,$65,$CE,$73,$29,$CB,$D0,$EF
	.dc.b	$5C,$EF,$B1,$E1,$AE,$35,$C1,$B8,$7A,$9E,$B3,$B3,$8D,$73,$BF,$25
	.dc.b	$6E,$7D,$57,$CD,$DC,$F9,$DA,$F9,$4F,$44,$F1,$79,$E1,$B7,$D2,$53
	.dc.b	$E2,$96,$BB,$3C,$0E,$95,$E7,$78,$9D,$E3,$6F,$82,$7A,$4E,$E3,$A9
	.dc.b	$BA,$F2,$3B,$3E,$B1,$DC,$B5,$F4,$1B,$1F,$82,$73,$E2,$BE,$B3,$8C
	.dc.b	$7D,$B8,$C3,$8A,$FF,$2F,$4E,$A3,$EE,$4E,$9C,$7C,$9E,$5F,$C6,$35
	.dc.b	$34,$8F,$CF,$C7,$F9,$7F,$0F,$CB,$8E,$26,$7C,$14,$66,$39,$99,$71
	.dc.b	$5F,$4E,$B9,$CF,$DB,$39,$EB,$F9,$48,$CB,$F9,$68,$CE,$9F,$EE,$7C
	.dc.b	$3B,$C7,$13,$BC,$F1,$8C,$BC,$63,$33,$C7,$0F,$1C,$79,$4E,$58,$79
	.dc.b	$6F,$8D,$7D,$08,$C5,$70,$BE,$CD,$9E,$97,$C8,$E9,$3F,$B2,$F6,$DF
	.dc.b	$59,$FD,$E2,$FF,$CD,$CD,$AF,$F7,$3B,$FE,$02,$8C,$F9,$DF,$CC,$1F
	.dc.b	$F2,$2C,$6F,$3F,$F3,$6B,$63,$3E,$46,$ED,$87,$83,$0C,$3F,$DA,$3E
	.dc.b	$5E,$EA,$6F,$A7,$86,$63,$FE,$83,$9D,$36,$73,$87,$F3,$53,$FA,$07
	.dc.b	$CA,$3C,$7F,$16,$A8,$C5,$FF,$71,$5E,$56,$9C,$DF,$3A,$E1,$79,$35
	.dc.b	$56,$F9,$45,$AE,$17,$D2,$72,$7F,$A6,$FE,$A9,$F5,$3F,$09,$B1,$FE
	.dc.b	$9F,$FA,$B7,$7E,$6B,$E3,$C7,$F3,$19,$D6,$19,$69,$1F,$EF,$1F,$17
	.dc.b	$C1,$F9,$FE,$2A,$3E,$1F,$D5,$FF,$BB,$AB,$FC,$24,$FF,$BD,$9B,$F1
	.dc.b	$3A,$F8,$6D,$E5,$27,$71,$71,$F8,$DF,$56,$F9,$4F,$9E,$DA,$EB,$78
	.dc.b	$9F,$C0,$FC,$27,$D3,$9F,$F5,$F2,$F9,$9C,$F0,$E9,$FC,$3F,$C8,$3E
	.dc.b	$62,$98,$F9,$3B,$87,$FB,$1F,$EA,$FF,$DC,$0D,$F8,$BF,$C7,$C3,$E7
	.dc.b	$AC,$7E,$4B,$EA,$FF,$CE,$55,$FE,$CF,$6F,$07,$E2,$FB,$D0,$FA,$FE
	.dc.b	$1F,$15,$F9,$DF,$8C,$7D,$63,$D9,$C3,$72,$FF,$7D,$72,$3E,$83,$35
	.dc.b	$FA,$2F,$8A,$FC,$36,$BD,$D7,$D1,$E7,$C3,$D3,$FA,$8F,$9D,$FE,$2F
	.dc.b	$E2,$FF,$B7,$17,$85,$9F,$1F,$4F,$EB,$BE,$B1,$EA,$D3,$1E,$CD,$E0
	.dc.b	$FF,$77,$7D,$27,$C5,$6F,$4F,$ED,$A3,$FC,$25,$18,$3F,$DE,$9F,$59
	.dc.b	$8E,$0F,$8C,$DD,$7F,$DD,$EF,$F8,$92,$77,$EC,$FC,$7F,$AC,$FD,$6D
	.dc.b	$7F,$28,$E7,$46,$2B,$FD,$60,$FC,$49,$D3,$9C,$3F,$DA,$5E,$CF,$F0
	.dc.b	$AF,$97,$7B,$16,$30,$FF,$00,$FA,$CF,$C1,$D7,$EE,$DB,$87,$F9,$29
	.dc.b	$FF,$69,$E3,$FE,$73,$72,$FF,$87,$E1,$18,$9F,$F2,$63,$63,$3F,$31
	.dc.b	$DF,$C1,$97,$CD,$F1,$46,$2F,$F1,$9F,$39,$F5,$2B,$FE,$00,$71,$7F
	.dc.b	$24,$8C,$4E,$A3,$1C,$79,$6D,$70,$3C,$C7,$FB,$E3,$E9,$7F,$99,$57
	.dc.b	$F3,$07,$F9,$D3,$F9,$B7,$3E,$19,$CE,$7F,$D6,$A3,$1E,$BF,$F3,$69
	.dc.b	$CD,$F5,$8C,$7B,$1F,$CF,$70,$7D,$63,$EB,$30,$DB,$F9,$9C,$D8,$99
	.dc.b	$1F,$EF,$F9,$D5,$7D,$5A,$CB,$FE,$FF,$8C,$E7,$F5,$F5,$CB,$6F,$E0
	.dc.b	$D7,$F4,$EB,$FF,$37,$C6,$34,$FF,$35,$7C,$1F,$F5,$09,$7D,$5D,$89
	.dc.b	$5B,$9E,$37,$FC,$05,$DC,$66,$79,$63,$CF,$7B,$3C,$5B,$7C,$5B,$97
	.dc.b	$CD,$C3,$F4,$93,$F9,$3C,$7F,$2C,$7E,$5F,$EA,$6E,$1F,$93,$8B,$3F
	.dc.b	$99,$9B,$7F,$99,$9B,$5F,$97,$E4,$56,$2D,$D2,$FA,$C7,$D4,$FF,$19
	.dc.b	$7F,$96,$61,$F2,$A7,$53,$9F,$3F,$4D,$9D,$57,$26,$CF,$F2,$3E,$7F
	.dc.b	$F7,$6C,$73,$6E,$1F,$DD,$70,$47,$F1,$A9,$FF,$46,$46,$3E,$53,$91
	.dc.b	$23,$97,$87,$6F,$F7,$5F,$45,$FB,$D3,$F7,$DF,$47,$6B,$F7,$51,$67
	.dc.b	$F8,$04,$DE,$0F,$E0,$13,$6B,$E4,$4D,$AF,$8F,$F3,$DF,$80,$7D,$0D
	.dc.b	$BF,$E3,$0F,$CE,$AF,$90,$E5,$F5,$B6,$79,$F0,$FF,$47,$8E,$11,$D2
	.dc.b	$3F,$A5,$7D,$67,$F4,$0A,$FA,$4D,$9F,$90,$FC,$AB,$FE,$50,$E5,$BE
	.dc.b	$4F,$23,$9E,$96,$FF,$39,$E2,$FE,$84,$FD,$17,$2D,$8C,$93,$1F,$D0
	.dc.b	$46,$5F,$D1,$70,$7E,$A7,$C6,$FD,$53,$6B,$F5,$0D,$9D,$C7,$EA,$E7
	.dc.b	$F6,$1F,$23,$23,$F5,$DE,$15,$AF,$D7,$4D,$D3,$F4,$D3,$6B,$FD,$A1
	.dc.b	$D9,$FE,$E3,$F8,$FF,$DD,$1F,$BB,$F8,$EB,$3F,$B9,$8B,$8F,$EE,$66
	.dc.b	$E1,$FB,$1F,$A5,$FE,$04,$62,$FA,$EA,$8C,$C7,$F0,$37,$57,$FC,$35
	.dc.b	$FE,$41,$F1,$0F,$A6,$FF,$40,$FC,$6F,$F9,$2A,$BF,$DC,$D3,$9B,$FF
	.dc.b	$4E,$FA,$3F,$EE,$58,$7A,$5C,$23,$18,$5B,$E2,$D7,$FC,$C8,$DA,$E3
	.dc.b	$7D,$27,$6D,$5F,$59,$EB,$F8,$8C,$36,$3F,$D5,$B8,$78,$2B,$39,$FF
	.dc.b	$94,$27,$27,$FB,$FE,$B8,$67,$ED,$8E,$B3,$B8,$C6,$3E,$F9,$E2,$6B
	.dc.b	$1E,$A4,$73,$1B,$9E,$97,$DB,$15,$78,$70,$ED,$18,$8E,$87,$2C,$BE
	.dc.b	$35,$71,$8E,$18,$F7,$C3,$CB,$5A,$C3,$9D,$68,$F3,$DF,$17,$E7,$E5
	.dc.b	$C7,$1F,$04,$E4,$D6,$5E,$98,$CD,$F4,$BF,$2E,$3E,$C8,$C5,$6E,$7C
	.dc.b	$13,$9C,$BB,$47,$DA,$3C,$75,$CF,$0F,$03,$CB,$2F,$55,$70,$AE,$F8
	.dc.b	$FD,$B7,$8C,$66,$FC,$F5,$C2,$72,$7F,$BE,$A3,$F0,$8F,$CD,$3E,$69
	.dc.b	$5C,$BF,$0B,$7F,$F2,$D3,$C5,$AE,$DA,$EF,$D6,$E8,$FE,$2E,$38,$E5
	.dc.b	$F8,$06,$E3,$B9,$FA,$7D,$BF,$7E,$B5,$7F,$9A,$BD,$D7,$38,$FC,$ED
	.dc.b	$6E,$F9,$5F,$0A,$FB,$B7,$C1,$FD,$2D,$6D,$E4,$6A,$FB,$1D,$1A,$64
	.dc.b	$95,$F9,$89,$E5,$1F,$B0,$AE,$D8,$1A,$62,$77,$19,$BE,$B7,$E7,$3A
	.dc.b	$5E,$30,$F5,$91,$FD,$A5,$F2,$C3,$AD,$EE,$72,$53,$C8,$DC,$15,$C6
	.dc.b	$31,$5E,$5A,$B1,$B9,$F3,$92,$3E,$21,$FA,$8E,$DC,$3F,$60,$CE,$3F
	.dc.b	$BC,$8F,$9D,$C6,$E1,$7C,$9F,$EE,$F8,$FF,$C4,$8D,$E2,$FE,$66,$DC
	.dc.b	$7F,$C7,$91,$6F,$94,$DB,$E7,$5F,$E3,$98,$C5,$48,$FF,$1E,$35,$FF
	.dc.b	$1E,$7B,$17,$0F,$F1,$A4,$D3,$FA,$1C,$77,$F3,$9F,$03,$49,$7F,$E4
	.dc.b	$D8,$FF,$6C,$36,$32,$FF,$9B,$DE,$5F,$03,$CF,$C5,$E9,$F0,$12,$B9
	.dc.b	$46,$5E,$91,$97,$FC,$F1,$1E,$3F,$AC,$91,$E1,$7C,$DE,$4E,$23,$FC
	.dc.b	$FD,$17,$4F,$F3,$E6,$5F,$E4,$09,$FF,$1F,$7D,$3F,$FB,$DD,$BA,$7C
	.dc.b	$FF,$25,$39,$9F,$F1,$C3,$F3,$9B,$3B,$8D,$5F,$FA,$ED,$BB,$7F,$8F
	.dc.b	$1E,$11,$FE,$BF,$F7,$1F,$59,$DD,$B8,$7F,$94,$5A,$FF,$91,$D7,$D5
	.dc.b	$6E,$64,$ED,$E5,$1F,$87,$87,$D6,$3E,$2B,$0D,$8F,$F4,$39,$A8,$CF
	.dc.b	$C5,$E9,$91,$A9,$E0,$B1,$67,$FD,$4E,$57,$C2,$B5,$39,$BC,$5F,$09
	.dc.b	$FE,$BB,$7C,$1F,$F3,$87,$CF,$38,$FE,$93,$63,$31,$8E,$5C,$B2,$32
	.dc.b	$72,$8F,$B5,$F5,$9F,$EE,$48,$FF,$3E,$2F,$43,$26,$BD,$D9,$3A,$8C
	.dc.b	$BC,$23,$E6,$7C,$53,$CA,$A4,$9F,$97,$1F,$E7,$C6,$C6,$5F,$86,$7F
	.dc.b	$7C,$8E,$4F,$0F,$37,$5A,$FC,$58,$4E,$9E,$26,$2B,$C7,$5F,$B5,$E1
	.dc.b	$FD,$B7,$C7,$CA,$FD,$A4,$65,$FD,$B4,$65,$FF,$71,$47,$F0,$BF,$43
	.dc.b	$FC,$67,$1F,$EC,$76,$BE,$8F,$01,$18,$CB,$FD,$CC,$FF,$90,$67,$87
	.dc.b	$99,$CC,$8F,$3F,$27,$F2,$9E,$5B,$EB,$3F,$2D,$A7,$E0,$6B,$DA,$DC
	.dc.b	$7E,$15,$6E,$7F,$2E,$D7,$E1,$B6,$37,$5D,$0E,$E4,$9E,$2E,$5F,$F7
	.dc.b	$27,$C0,$7F,$42,$6F,$8D,$FC,$B6,$0D,$F2,$24,$E8,$87,$C5,$9E,$4F
	.dc.b	$C7,$F8,$2F,$27,$5F,$F9,$E1,$FD,$4E,$3F,$A1,$DD,$18,$9D,$65,$FC
	.dc.b	$F6,$38,$DA,$B7,$D7,$93,$55,$FB,$FF,$8D,$FC,$36,$3F,$89,$6E,$3F
	.dc.b	$E4,$76,$E1,$FE,$4C,$7F,$2C,$FE,$31,$CA,$69,$36,$ED,$C9,$27,$F8
	.dc.b	$4C,$E5,$FF,$2B,$DF,$B7,$C0,$FD,$C4,$7E,$25,$FE,$92,$EB,$0F,$F4
	.dc.b	$DC,$62,$35,$D7,$FC,$C7,$EB,$1F,$4D,$49,$3F,$EF,$58,$F8,$BB,$3E
	.dc.b	$B3,$F7,$D3,$F9,$76,$E5,$F3,$DB,$97,$FA,$A9,$FE,$05,$3E,$33,$E7
	.dc.b	$7D,$66,$9B,$D9,$FD,$BE,$31,$81,$E3,$3F,$A5,$ED,$7C,$C7,$F7,$56
	.dc.b	$C7,$5A,$FE,$21,$39,$FA,$7F,$EF,$5F,$58,$F8,$8B,$DB,$30,$E1,$39
	.dc.b	$79,$7D,$63,$E2,$BF,$C0,$4E,$63,$F6,$CD,$E0,$FC,$73,$B8,$F7,$79
	.dc.b	$8F,$AC,$7D,$67,$F5,$F6,$CE,$9A,$FF,$18,$9E,$1C,$BF,$E1,$1E,$26
	.dc.b	$D4,$95,$EF,$87,$D0,$7F,$85,$F7,$FF,$54,$69,$FA,$16,$F2,$FF,$B7
	.dc.b	$1B,$7F,$9F,$5F,$05,$FE,$1F,$FA,$BF,$E2,$33,$F2,$DA,$FF,$15,$6D
	.dc.b	$FE,$B5,$7C,$37,$F8,$9F,$CD,$7D,$5D,$FA,$09,$87,$D5,$9C,$BF,$83
	.dc.b	$C1,$F5,$6F,$A9,$FF,$19,$46,$1B,$19,$9E,$1F,$30,$F8,$6D,$B6,$33
	.dc.b	$18,$33,$F0,$5A,$F8,$7B,$3C,$A3,$0F,$E2,$7E,$A7,$FC,$D4,$FE,$2F
	.dc.b	$E8,$9F,$51,$FE,$43,$6C,$6E,$BF,$C8,$D5,$C2,$3F,$9D,$37,$4F,$E2
	.dc.b	$4D,$E4,$FF,$67,$F8,$28,$FF,$0B,$C6,$1B,$5F,$C4,$DB,$3B,$F8,$3E
	.dc.b	$FE,$97,$C6,$FA,$9F,$F7,$55,$FF,$60,$C5,$18,$7F,$C0,$46,$3E,$2B
	.dc.b	$86,$BC,$59,$5F,$4E,$B8,$67,$FC,$39,$F7,$C7,$57,$2F,$F8,$55,$B7
	.dc.b	$D6,$39,$5F,$F9,$A7,$EA,$5F,$29,$49,$3F,$B7,$87,$CB,$62,$3F,$B1
	.dc.b	$47,$07,$53,$CE,$7F,$CE,$9C,$5F,$E7,$D9,$FE,$00,$D7,$FB,$53,$70
	.dc.b	$FE,$84,$BE,$5B,$8F,$C5,$4A,$E2,$F4,$8F,$F1,$E6,$2F,$99,$FF,$19
	.dc.b	$53,$EA,$3F,$5E,$8F,$D8,$E9,$5D,$78,$34,$F9,$33,$C2,$17,$D7,$3E
	.dc.b	$53,$98,$C4,$F2,$3B,$57,$23,$A5,$F7,$AE,$18,$79,$9B,$87,$4A,$DD
	.dc.b	$D3,$9D,$F1,$8F,$46,$1E,$0B,$E3,$5E,$13,$DA,$71,$29,$98,$E1,$3C
	.dc.b	$B0,$F1,$3E,$09,$E4,$FA,$5F,$91,$3C,$1B,$3C,$1F,$29,$DE,$78,$3F
	.dc.b	$22,$34,$DC,$35,$87,$AE,$38,$9E,$59,$DD,$FC,$C8,$FA,$C7,$12,$94
	.dc.b	$CB,$B7,$E2,$F8,$4F,$A9,$DC,$7D,$A8,$C3,$CA,$51,$88,$E6,$D7,$BF
	.dc.b	$C7,$72,$7F,$14,$D4,$D2,$3F,$4F,$1F,$E5,$FC,$5F,$4E,$38,$99,$F0
	.dc.b	$7F,$D0,$1A,$FE,$3E,$32,$FE,$46,$B8,$C7,$DD,$33,$AF,$FB,$96,$B8
	.dc.b	$BE,$68,$E9,$7D,$DF,$29,$83,$91,$8B,$ED,$5C,$A3,$E7,$E1,$F8,$9C
	.dc.b	$3C,$35,$E5,$9E,$18,$78,$70,$ED,$7B,$C3,$95,$7E,$B6,$31,$3A,$8F
	.dc.b	$15,$F3,$9C,$C7,$91,$B1,$FB,$8C,$1F,$97,$B9,$FF,$92,$E1,$FE,$2C
	.dc.b	$F7,$9F,$2B,$8C,$BF,$D6,$57,$BA,$FE,$B1,$5A,$7F,$AD,$46,$67,$F9
	.dc.b	$94,$E5,$FE,$69,$19,$C3,$F9,$9C,$E5,$FF,$49,$7D,$1B,$C0,$F0,$43
	.dc.b	$E3,$BF,$89,$63,$FE,$CA,$AD,$3F,$DA,$63,$38,$FF,$A6,$27,$2F,$F4
	.dc.b	$38,$CD,$76,$BE,$B1,$8B,$FF,$25,$FE,$8B,$C2,$BF,$A3,$B6,$3F,$C9
	.dc.b	$FF,$12,$F9,$3F,$2C,$E7,$1F,$F6,$B3,$6D,$5F,$E0,$2B,$FE,$FD,$39
	.dc.b	$8C,$57,$F5,$78,$C7,$93,$D3,$2F,$83,$5A,$C3,$FD,$6D,$F9,$93,$E6
	.dc.b	$3E,$A6,$28,$76,$61,$D6,$7F,$EF,$07,$2D,$8E,$3F,$A3,$7F,$92,$A3
	.dc.b	$07,$FB,$D9,$C5,$7F,$B0,$9B,$87,$FB,$1D,$B3,$FC,$FE,$FA,$D6,$FD
	.dc.b	$9F,$B3,$F5,$7F,$E6,$CA,$FF,$0F,$36,$7F,$C7,$DD,$8F,$9E,$DF,$A9
	.dc.b	$C3,$E1,$7F,$03,$F3,$1F,$B6,$D7,$FB,$CE,$3F,$EB,$BC,$BF,$C5,$13
	.dc.b	$97,$FA,$97,$AE,$7F,$DE,$07,$F9,$48,$B9,$FF,$81,$67,$82,$9E,$1E
	.dc.b	$F7,$E7,$D7,$E2,$5F,$EB,$4B,$FF,$2D,$FE,$65,$F8,$17,$FC,$BB,$19
	.dc.b	$BF,$F2,$C4,$E5,$FF,$74,$FE,$8E,$F3,$BA,$7D,$4F,$F7,$1C,$3F,$D0
	.dc.b	$54,$AD,$47,$F9,$A2,$75,$3F,$E6,$3F,$CC,$DF,$35,$E5,$77,$19,$8C
	.dc.b	$61,$E9,$FC,$53,$F5,$D8,$7C,$45,$EA,$E7,$C5,$D3,$E4,$FB,$3F,$2F
	.dc.b	$89,$6F,$C4,$FC,$3F,$A1,$FA,$D1,$FE,$7C,$5F,$27,$9E,$9D,$3F,$27
	.dc.b	$F3,$2F,$CC,$57,$D4,$55,$9F,$0F,$4F,$E6,$BE,$95,$F3,$BF,$47,$E0
	.dc.b	$D1,$98,$C3,$F5,$7F,$32,$F8,$3D,$BF,$7F,$93,$ED,$47,$F9,$8F,$3F
	.dc.b	$F4,$8B,$A7,$F0,$AE,$63,$FD,$1D,$F3,$3F,$32,$7F,$D2,$EE,$E3,$F0
	.dc.b	$8E,$5D,$F2,$79,$7C,$8F,$C3,$7D,$5F,$BA,$7F,$15,$A7,$F0,$3E,$D4
	.dc.b	$E5,$7D,$C7,$C8,$CB,$FD,$4A,$EC,$C5,$FF,$A8,$5F,$E1,$8B,$E1,$BD
	.dc.b	$1A,$FB,$E3,$1E,$DF,$F2,$6F,$AA,$CE,$1F,$E3,$1B,$B5,$FE,$35,$6D
	.dc.b	$A9,$F0,$9F,$C2,$A7,$31,$8B,$FC,$AC,$62,$75,$87,$77,$87,$BB,$FA
	.dc.b	$7F,$B8,$7C,$EE,$30,$FF,$60,$C6,$1E,$73,$B9,$D6,$7F,$EB,$FA,$D2
	.dc.b	$CF,$DD,$97,$EB,$5F,$F3,$CF,$E4,$6F,$93,$F4,$76,$FF,$46,$8C,$67
	.dc.b	$E8,$8C,$4E,$A3,$FD,$5D,$3A,$C3,$B4,$7F,$95,$FE,$B5,$CD,$FC,$07
	.dc.b	$F3,$DF,$F7,$34,$FE,$DA,$72,$FF,$B7,$23,$35,$C8,$FF,$31,$4E,$9F
	.dc.b	$F6,$D4,$6B,$2F,$B1,$39,$7F,$31,$C6,$9D,$5F,$8F,$0F,$17,$C7,$FF
	.dc.b	$3E,$C3,$84,$FF,$1E,$AD,$E3,$E0,$FA,$FF,$C1,$70,$ED,$B5,$E9,$6C
	.dc.b	$F9,$C5,$A3,$39,$F9,$BB,$3F,$91,$7C,$17,$FA,$63,$0F,$85,$18,$F8
	.dc.b	$7F,$23,$63,$27,$FC,$00,$9E,$18,$F7,$FC,$67,$FD,$F3,$3F,$C8,$1B
	.dc.b	$1F,$EB,$7F,$C0,$BC,$53,$9F,$93,$FE,$BB,$5A,$CB,$F0,$3C,$F8,$9F
	.dc.b	$C0,$36,$BE,$EB,$70,$FF,$61,$71,$FE,$32,$FF,$51,$87,$FB,$B6,$FF
	.dc.b	$D3,$11,$F4,$BE,$F7,$FA,$DD,$FF,$A6,$EB,$4F,$DD,$E4,$4E,$A1,$3A
	.dc.b	$AE,$5E,$C7,$A3,$E4,$BF,$A5,$5E,$E9,$F3,$BF,$E1,$CF,$94,$7E,$03
	.dc.b	$EE,$8C,$37,$4F,$29,$E6,$9D,$BF,$26,$76,$FF,$4A,$F4,$3F,$9E,$B7
	.dc.b	$CF,$FE,$B5,$F8,$2F,$F4,$2D,$BF,$D6,$D5,$A7,$F0,$4D,$DB,$F2,$73
	.dc.b	$FC,$86,$53,$9B,$FE,$DB,$39,$8C,$4F,$F6,$98,$C5,$F8,$3C,$8F,$F0
	.dc.b	$AD,$7F,$81,$30,$FD,$CE,$D8,$F8,$0F,$DC,$37,$2F,$DA,$B7,$2F,$B2
	.dc.b	$F2,$F3,$FF,$BA,$7E,$7A,$FA,$4C,$5F,$C6,$BF,$EF,$53,$98,$F8,$95
	.dc.b	$C2,$B7,$5F,$6E,$B4,$FD,$DB,$FF,$27,$5F,$3E,$44,$3E,$C7,$FC,$08
	.dc.b	$EA,$B8,$46,$3E,$23,$2D,$BE,$55,$C5,$E9,$FA,$23,$D6,$FF,$07,$7E
	.dc.b	$89,$FC,$76,$7F,$C3,$33,$97,$C5,$19,$C7,$FC,$2D,$39,$7F,$29,$19
	.dc.b	$C9,$39,$DF,$F8,$6C,$E7,$F4,$57,$DE,$7F,$87,$57,$C6,$D8,$C3,$72
	.dc.b	$F8,$53,$A6,$CE,$9F,$8D,$E1,$BE,$83,$FC,$50,$FF,$8D,$74,$FD,$43
	.dc.b	$99,$DF,$C3,$6A,$B6,$D8,$C9,$F4,$7F,$3A,$FF,$1A,$EC,$6D,$F5,$D7
	.dc.b	$F9,$4C,$68,$9F,$29,$D2,$F9,$3F,$E6,$6F,$B8,$FE,$9B,$F2,$BF,$E1
	.dc.b	$2D,$02,$33,$EE,$E2,$31,$E8,$FF,$83,$6B,$84,$6F,$C7,$C5,$6E,$F8
	.dc.b	$CF,$46,$BC,$B1,$FF,$6E,$CE,$EB,$8C,$F8,$BE,$E3,$FC,$8D,$F3,$BA
	.dc.b	$8C,$78,$08,$C4,$78,$70,$B5,$C2,$3F,$DC,$F8,$7A,$63,$19,$7A,$1B
	.dc.b	$1E,$1A,$E3,$87,$5B,$E2,$78,$30,$C9,$E0,$CB,$D7,$87,$29,$D1,$C0
	.dc.b	$39,$7C,$C6,$5C,$12,$BC,$18,$FC,$CA,$EB,$18,$C7,$D4,$DB,$EF,$5B
	.dc.b	$CB,$E4,$C7,$82,$74,$DC,$7D,$2F,$97,$0E,$B1,$8B,$F8,$95,$C2,$73
	.dc.b	$5E,$78,$E7,$8F,$AE,$73,$5C,$2F,$96,$3F,$2F,$0E,$11,$87,$CF,$19
	.dc.b	$9E,$11,$A2,$3E,$BC,$B8,$47,$A4,$FC,$91,$F3,$8B,$87,$DF,$6E,$3F
	.dc.b	$65,$CF,$83,$FE,$D5,$78,$9F,$51,$C3,$AF,$94,$CB,$96,$D7,$FA,$19
	.dc.b	$FC,$6C,$F3,$C3,$F1,$4D,$C7,$F1,$DE,$07,$E9,$E7,$35,$C0,$98,$7E
	.dc.b	$06,$75,$5F,$95,$BE,$53,$FA,$16,$BE,$33,$81,$CA,$FA,$BF,$8B,$8D
	.dc.b	$E1,$FE,$92,$8D,$57,$5C,$2C,$7E,$B7,$0F,$0B,$DE,$32,$F7,$C3,$F4
	.dc.b	$D3,$A8,$C5,$73,$7C,$15,$FB,$09,$D5,$FC,$79,$C9,$2F,$9B,$DA,$35
	.dc.b	$1E,$A3,$31,$BB,$ED,$19,$7E,$6C,$71,$C3,$DE,$7C,$0B,$F0,$B8,$C3
	.dc.b	$B4,$79,$09,$3E,$73,$F8,$3F,$AB,$87,$E6,$93,$0F,$E0,$D1,$F1,$B9
	.dc.b	$1F,$21,$FE,$F5,$AF,$EC,$8D,$7F,$C4,$8D,$E4,$FF,$2D,$FE,$8D,$FD
	.dc.b	$D6,$77,$F0,$BF,$CA,$E7,$FC,$BF,$5B,$E0,$FF,$7B,$B6,$BF,$DE,$57
	.dc.b	$C5,$FF,$18,$2F,$AD,$E5,$5F,$EF,$E8,$D5,$F8,$3C,$5C,$DF,$2B,$B9
	.dc.b	$2F,$94,$FF,$7D,$78,$E1,$CE,$F7,$1A,$9F,$F2,$07,$E5,$0F,$95,$E6
	.dc.b	$BE,$66,$98,$89,$3F,$E0,$99,$D3,$67,$2F,$F9,$2F,$F3,$57,$DC,$79
	.dc.b	$BC,$47,$D0,$D3,$D3,$3A,$AF,$F0,$A3,$5F,$F0,$0C,$6D,$CD,$7C,$0F
	.dc.b	$CB,$9F,$2B,$7C,$59,$5F,$E1,$D9,$CB,$67,$2F,$F9,$7E,$3F,$C0,$EE
	.dc.b	$4F,$F1,$70,$FC,$7B,$FD,$7E,$E3,$E0,$7F,$C4,$CD,$F3,$7E,$9B,$BF
	.dc.b	$0A,$D7,$08,$FF,$49,$57,$06,$D6,$AB,$6F,$29,$7D,$05,$33,$5E,$4B
	.dc.b	$E5,$CF,$FE,$E5,$AD,$CF,$96,$B6,$FF,$A8,$7F,$1E,$78,$5E,$6F,$BB
	.dc.b	$FF,$0D,$4F,$B9,$AE,$E3,$D7,$3C,$23,$5E,$5F,$13,$33,$23,$E3,$CE
	.dc.b	$DE,$47,$5F,$37,$9F,$2F,$87,$C4,$93,$C2,$32,$F8,$E3,$2F,$A6,$3C
	.dc.b	$7E,$7F,$CE,$FB,$A7,$C2,$53,$35,$E2,$9C,$3F,$6A,$76,$E4,$FB,$B9
	.dc.b	$7F,$87,$B7,$77,$7F,$88,$EA,$D7,$18,$FB,$B5,$B7,$FC,$50,$D8,$FF
	.dc.b	$16,$CF,$07,$FD,$A0,$DA,$C4,$EE,$7E,$9E,$5D,$08,$FE,$43,$EB,$1F
	.dc.b	$43,$E0,$E7,$FC,$6C,$3C,$DA,$66,$BD,$73,$87,$94,$65,$FF,$71,$60
	.dc.b	$F4,$7F,$2F,$E1,$3E,$12,$9E,$E8,$95,$FA,$29,$D3,$67,$2F,$FB,$B7
	.dc.b	$F3,$4F,$F2,$63,$BF,$C1,$3A,$9F,$A4,$8C,$C6,$FD,$1C,$BB,$3F,$5F
	.dc.b	$1A,$DF,$E5,$37,$87,$FC,$AD,$8B,$0E,$37,$CC,$E3,$A7,$EC,$BE,$DF
	.dc.b	$F9,$74,$1B,$87,$C9,$8D,$9F,$BE,$8D,$9A,$9F,$47,$E0,$1C,$FD,$19
	.dc.b	$5E,$68,$CB,$C6,$32,$FC,$38,$FE,$07,$FA,$43,$E6,$FF,$82,$FC,$13
	.dc.b	$EF,$69,$A8,$97,$FC,$42,$74,$D9,$CA,$DF,$F1,$D7,$D6,$3F,$8B,$B5
	.dc.b	$D6,$FF,$E7,$D8,$C4,$FE,$56,$FF,$D7,$F5,$FE,$C1,$F0,$1F,$7D,$FD
	.dc.b	$29,$7B,$F4,$C4,$7F,$4E,$FB,$A7,$DE,$5E,$69,$87,$FB,$2E,$72,$FF
	.dc.b	$B3,$7F,$1E,$FE,$1B,$9F,$EB,$3D,$34,$3E,$A7,$FA,$CB,$6B,$87,$8D
	.dc.b	$FD,$73,$93,$FD,$73,$1F,$ED,$75,$F1,$39,$35,$E3,$7F,$32,$FC,$C9
	.dc.b	$F7,$5F,$D9,$3F,$49,$7C,$7D,$35,$1F,$DA,$3B,$7F,$9B,$6F,$E5,$7E
	.dc.b	$2E,$7F,$28,$6E,$8C,$CE,$FD,$0F,$EE,$3F,$92,$72,$EA,$FF,$57,$DA
	.dc.b	$37,$18,$6E,$1D,$1E,$0F,$6F,$BB,$FD,$B3,$FE,$F5,$7E,$5B,$E3,$A4
	.dc.b	$63,$0F,$98,$F0,$9D,$7C,$BE,$70,$F1,$BF,$BA,$8C,$F3,$7F,$82,$23
	.dc.b	$51,$EE,$FC,$D7,$E7,$BD,$DF,$9E,$FF,$B2,$B0,$FB,$51,$88,$D7,$0F
	.dc.b	$F6,$2F,$D2,$9F,$3B,$FE,$14,$FB,$5F,$DD,$57,$CA,$6A,$9D,$BD,$73
	.dc.b	$F9,$3F,$13,$FB,$78,$F8,$0F,$C6,$DB,$FD,$8D,$C1,$FE,$20,$8D,$69
	.dc.b	$F8,$DF,$CB,$BF,$C5,$D4,$69,$FE,$31,$87,$E9,$7F,$E6,$BC,$3E,$F4
	.dc.b	$3E,$9F,$FC,$4F,$1F,$AF,$9D,$4F,$F4,$43,$8F,$C5,$7F,$8C,$9E,$1E
	.dc.b	$65,$BF,$F1,$64,$6D,$BA,$7C,$96,$E1,$F8,$C1,$F2,$FF,$E0,$AF,$8A
	.dc.b	$7D,$05,$24,$FF,$85,$23,$F9,$13,$7C,$FC,$BF,$97,$11,$C9,$D1,$C7
	.dc.b	$C1,$ED,$32,$30,$7D,$13,$F8,$77,$DD,$FD,$8C,$CF,$EA,$A3,$51,$8E
	.dc.b	$EF,$85,$3B,$23,$83,$3E,$8C,$AE,$33,$B8,$E3,$F3,$BF,$E2,$AF,$D3
	.dc.b	$3F,$73,$5D,$B8,$7F,$CA,$8D,$F8,$AD,$78,$52,$30,$43,$E1,$57,$27
	.dc.b	$EE,$F7,$FF,$B3,$B4,$FB,$5B,$FF,$85,$DB,$EB,$7D,$B9,$D7,$27,$E7
	.dc.b	$DB,$7F,$CC,$6F,$84,$FE,$8D,$B1,$FC,$D6,$76,$BE,$E9,$F7,$7F,$CE
	.dc.b	$61,$DD,$FE,$38,$7F,$C7,$FF,$78,$64,$67,$D1,$4E,$67,$8C,$5B,$EB
	.dc.b	$A7,$08,$C6,$1D,$F0,$E5,$87,$16,$E1,$B3,$C1,$87,$27,$A9,$8A,$E5
	.dc.b	$7E,$13,$C7,$1C,$4F,$49,$67,$95,$7B,$5B,$3C,$6F,$C4,$DA,$E1,$7D
	.dc.b	$9F,$06,$1D,$67,$BC,$79,$0E,$E5,$BE,$AF,$03,$51,$C3,$0F,$7C,$6E
	.dc.b	$FB,$3F,$55,$B3,$B3,$C9,$1B,$BF,$29,$F4,$5F,$9A,$D7,$C0,$66,$FE
	.dc.b	$53,$7E,$8D,$F7,$3B,$8F,$C2,$46,$1E,$52,$8C,$47,$19,$E0,$FE,$17
	.dc.b	$E5,$2A,$69,$1D,$BF,$56,$DF,$2F,$FE,$5C,$CF,$F2,$15,$CC,$F9,$93
	.dc.b	$AD,$FE,$E4,$66,$7F,$2D,$39,$8F,$A4,$67,$3F,$F7,$05,$F5,$7E,$26
	.dc.b	$1F,$9B,$FB,$0E,$33,$FE,$6C,$7C,$35,$E9,$0E,$B1,$E1,$7C,$A7,$CB
	.dc.b	$5D,$5F,$A5,$7E,$D8,$C4,$7E,$96,$31,$5E,$A8,$ED,$8F,$57,$D2,$62
	.dc.b	$FD,$33,$BA,$E3,$3A,$6D,$7A,$1B,$5F,$D8,$FE,$F3,$FC,$7D,$7F,$C9
	.dc.b	$1B,$7F,$D3,$1F,$F5,$8C,$67,$DB,$FE,$94,$F1,$7F,$D1,$13,$63,$F1
	.dc.b	$99,$AB,$9C,$FF,$01,$8C,$57,$08,$FE,$11,$E4,$FF,$59,$BF,$C7,$36
	.dc.b	$3F,$2C,$D7,$F3,$4D,$7F,$86,$3B,$AF,$EB,$DE,$C7,$F1,$8A,$FF,$71
	.dc.b	$CE,$5F,$F0,$14,$66,$BF,$B4,$FD,$E6,$5F,$F2,$B3,$6F,$FD,$37,$F3
	.dc.b	$3F,$02,$73,$E2,$59,$57,$8E,$73,$87,$C2,$8C,$4E,$AB,$E0,$36,$7F
	.dc.b	$A9,$FC,$27,$E2,$9F,$EA,$B5,$E1,$8C,$5F,$F6,$5F,$98,$76,$F8,$E1
	.dc.b	$0F,$A1,$C3,$AC,$3F,$B2,$4E,$A7,$27,$D1,$71,$1F,$BD,$9E,$4D,$9D
	.dc.b	$7D,$E7,$F9,$4A,$39,$B6,$BC,$93,$E3,$8C,$65,$FE,$67,$E6,$7C,$DF
	.dc.b	$C7,$FB,$CF,$F3,$46,$3F,$DD,$FC,$EF,$E1,$B3,$9F,$77,$5F,$0F,$F1
	.dc.b	$AF,$FD,$81,$7F,$8C,$6E,$9F,$D8,$DA,$FF,$4D,$C3,$FD,$C7,$F1,$1F
	.dc.b	$5D,$B5,$FE,$00,$FB,$CF,$E0,$5A,$7F,$9E,$E1,$5B,$FB,$CF,$F0,$8C
	.dc.b	$7F,$25,$B4,$63,$0F,$E5,$51,$8F,$13,$F3,$4D,$AF,$E4,$DF,$77,$FA
	.dc.b	$8A,$FE,$5C,$D7,$F9,$93,$5F,$EC,$8D,$7F,$CE,$FF,$74,$FB,$AE,$AD
	.dc.b	$AF,$23,$67,$C6,$D7,$E1,$7D,$4F,$EC,$1F,$F3,$DB,$5F,$F4,$1B,$70
	.dc.b	$37,$EE,$DF,$27,$EC,$6E,$1F,$0D,$B7,$F0,$9A,$FD,$1F,$92,$FB,$F3
	.dc.b	$F2,$5B,$1F,$21,$AE,$F9,$3E,$4F,$DD,$FE,$7A,$3E,$D3,$5F,$F4,$9B
	.dc.b	$5F,$F4,$DB,$57,$DD,$BC,$DF,$CE,$37,$0F,$EB,$0D,$9F,$EC,$0D,$7F
	.dc.b	$11,$F7,$7F,$47,$53,$49,$E7,$FD,$EF,$67,$E7,$57,$FA,$99,$AF,$FA
	.dc.b	$B9,$B5,$FE,$62,$E1,$FC,$2F,$DD,$FF,$05,$83,$49,$E1,$FE,$2B,$EE
	.dc.b	$9E,$AD,$26,$BF,$93,$8C,$BF,$DC,$FE,$DD,$F7,$3F,$C0,$DB,$1F,$E5
	.dc.b	$EF,$BA,$BA,$FF,$9B,$66,$F6,$AB,$FC,$69,$1A,$9F,$F0,$33,$5C,$DF
	.dc.b	$F0,$78,$C4,$7F,$8F,$DE,$73,$FE,$AD,$B7,$DD,$78,$EF,$F9,$34,$62
	.dc.b	$7F,$89,$D3,$E7,$7D,$07,$ED,$FE,$21,$F3,$58,$AD,$CE,$7E,$23,$19
	.dc.b	$FF,$B2,$ED,$1F,$07,$C0,$97,$FE,$7E,$FB,$87,$C1,$E3,$0F,$85,$5F
	.dc.b	$CE,$7E,$E3,$BD,$F9,$23,$19,$FF,$2D,$F8,$DF,$F4,$DC,$FF,$2A,$6D
	.dc.b	$FF,$B6,$FE,$05,$F5,$1B,$EA,$97,$FE,$27,$BD,$DF,$F8,$C5,$AF,$F9
	.dc.b	$8D,$AF,$F9,$A5,$EB,$C9,$F5,$63,$FD,$F3,$F7,$4F,$A5,$EB,$1F,$EE
	.dc.b	$B8,$C5,$F9,$7E,$EB,$D0,$7D,$0F,$4A,$C6,$4F,$C2,$72,$62,$FF,$D9
	.dc.b	$3C,$DF,$4E,$7D,$DE,$17,$C1,$3F,$95,$7E,$35,$FD,$96,$9A,$FE,$AF
	.dc.b	$5F,$34,$62,$BF,$98,$7C,$13,$ED,$39,$63,$FC,$FE,$30,$FF,$54,$FB
	.dc.b	$A7,$C4,$FF,$5B,$6C,$7D,$9D,$F1,$32,$3F,$CA,$53,$AC,$7F,$CB,$4D
	.dc.b	$7F,$CB,$CD,$7F,$D2,$0E,$F2,$FF,$71,$73,$FE,$22,$BC,$57,$F8,$3E
	.dc.b	$AF,$17,$DC,$49,$FE,$AD,$F8,$6B,$DF,$C6,$7F,$ED,$7C,$D7,$FD,$3F
	.dc.b	$EF,$3F,$B0,$E1,$E0,$8C,$7C,$7F,$B7,$E4,$64,$FE,$D7,$D2,$B8,$F3
	.dc.b	$9F,$F0,$67,$D1,$BE,$7B,$3C,$3F,$EF,$98,$C6,$1F,$DF,$3D,$57,$CF
	.dc.b	$FF,$80,$DB,$1F,$DB,$7E,$F3,$FC,$77,$7F,$5A,$73,$97,$FB,$BD,$AF
	.dc.b	$FA,$15,$AF,$FA,$DD,$AF,$FA,$F5,$AF,$F4,$8B,$FE,$DD,$5F,$D7,$EB
	.dc.b	$B4,$63,$D1,$F2,$77,$4C,$7F,$69,$EF,$5C,$7C,$53,$FB,$3E,$0C,$63
	.dc.b	$F5,$9F,$67,$9D,$30,$FF,$04,$7E,$0D,$F3,$B0,$F2,$4E,$55,$FE,$A7
	.dc.b	$D2,$7D,$EB,$E6,$35,$39,$8C,$63,$FE,$13,$CB,$EF,$3F,$E0,$16,$E1
	.dc.b	$FE,$00,$6B,$FE,$0F,$6B,$FE,$E1,$C3,$F0,$7D,$1F,$E1,$19,$7D,$8D
	.dc.b	$8C,$3F,$E1,$5F,$BD,$FE,$D7,$97,$9D,$B8,$F3,$7F,$9F,$E3,$EC,$8C
	.dc.b	$73,$F1,$3F,$C2,$33,$FD,$CA,$73,$3F,$E1,$98,$C4,$32,$7D,$EB,$ED
	.dc.b	$7F,$C3,$2D,$BF,$F1,$03,$6B,$FC,$3E,$D7,$FC,$5C,$E7,$1F,$9C,$BE
	.dc.b	$3B,$FC,$60,$E6,$BE,$23,$97,$FC,$81,$E2,$3E,$73,$FC,$52,$D8,$FF
	.dc.b	$18,$37,$0F,$F1,$7B,$5F,$F1,$F3,$9C,$3E,$F3,$FE,$44,$6D,$7F,$90
	.dc.b	$63,$15,$F5,$5B,$5F,$0F,$EF,$7F,$C8,$73,$9E,$6F,$F2,$44,$62,$75
	.dc.b	$F2,$D8,$3A,$4D,$AF,$F3,$6F,$CE,$7F,$4A,$AF,$F6,$E7,$C1,$99,$1F
	.dc.b	$C3,$DC,$98,$F5,$3F,$9F,$7C,$B2,$B8,$4A,$B8,$4E,$63,$9F,$8F,$C2
	.dc.b	$B7,$5A,$6D,$E6,$B7,$87,$FB,$C6,$7A,$FD,$EB,$B9,$F4,$39,$BF,$44
	.dc.b	$E6,$31,$1E,$3C,$BA,$29,$D3,$EC,$31,$5D,$1C,$46,$B3,$F5,$47,$3C
	.dc.b	$7A,$CE,$71,$F2,$DF,$1B,$DC,$E6,$FD,$67,$49,$D5,$70,$C7,$9C,$62
	.dc.b	$77,$7C,$A3,$11,$E6,$9D,$9E,$EC,$FE,$54,$7C,$29,$DE,$1E,$18,$E0
	.dc.b	$7B,$E7,$6E,$2F,$B5,$F4,$8D,$65,$F2,$2F,$91,$30,$F1,$E1,$CA,$31
	.dc.b	$7C,$70,$F8,$F3,$9A,$F1,$92,$7D,$17,$D6,$31,$87,$C5,$9C,$E5,$CF
	.dc.b	$2F,$8D,$39,$3C,$B1,$97,$E0,$4F,$E4,$9D,$97,$2F,$BE,$DB,$FB,$CE
	.dc.b	$7C,$07,$CA,$F1,$8F,$C5,$C6,$1C,$57,$F5,$89,$D4,$7E,$2E,$74,$F3
	.dc.b	$8F,$C2,$C7,$1D,$3F,$06,$DC,$BF,$37,$DF,$FA,$89,$CD,$70,$25,$FE
	.dc.b	$4A,$75,$5F,$A2,$BE,$53,$E6,$AD,$DF,$E4,$2F,$73,$93,$F3,$75,$C1
	.dc.b	$F4,$57,$E6,$A7,$71,$83,$71,$D2,$FC,$F1,$4E,$97,$E8,$7C,$C6,$4E
	.dc.b	$4E,$4F,$23,$9C,$3E,$53,$D8,$E2,$58,$E1,$3E,$FA,$F3,$4F,$1C,$3D
	.dc.b	$57,$CA,$FA,$CF,$1B,$E0,$D3,$F8,$2E,$1E,$EA,$E1,$3A,$AE,$F5,$A3
	.dc.b	$B9,$E4,$7A,$9F,$BC,$F6,$79,$1C,$EE,$4F,$F0,$48,$FA,$3D,$2F,$92
	.dc.b	$7C,$47,$F2,$6E,$8F,$EF,$D1,$8B,$FE,$61,$18,$9D,$70,$7F,$75,$FB
	.dc.b	$CF,$F0,$15,$6A,$31,$E1,$EA,$3F,$B3,$61,$CD,$FE,$E0,$62,$F5,$8B
	.dc.b	$0E,$73,$A8,$C7,$DE,$EA,$73,$97,$F8,$4D,$B9,$78,$E7,$47,$E8,$AB
	.dc.b	$8D,$FE,$A5,$AF,$EC,$DA,$FF,$46,$5F,$7A,$F9,$6F,$2A,$FB,$6A,$62
	.dc.b	$25,$75,$9E,$0D,$8F,$14,$62,$3C,$1F,$79,$FE,$9E,$7C,$8D,$73,$F0
	.dc.b	$1E,$98,$C4,$79,$5D,$69,$FE,$33,$6E,$7F,$E2,$76,$CF,$F8,$55,$7D
	.dc.b	$9F,$CE,$FB,$B7,$CA,$DF,$1A,$5F,$D9,$3F,$C3,$93,$A7,$1B,$23,$11
	.dc.b	$F2,$5A,$FC,$96,$BF,$E3,$96,$BF,$DE,$7E,$EF,$FD,$9A,$F7,$3F,$D6
	.dc.b	$5E,$57,$E1,$DA,$66,$36,$FE,$20,$38,$D3,$FD,$4E,$DD,$3F,$BF,$B6
	.dc.b	$7F,$C1,$1F,$07,$FE,$8F,$7A,$5F,$78,$7D,$05,$3C,$55,$FE,$4E,$AD
	.dc.b	$90,$F1,$39,$7C,$92,$F5,$FD,$5F,$75,$FE,$03,$9F,$77,$DD,$FF,$8D
	.dc.b	$5B,$E5,$EB,$C3,$91,$FC,$E2,$B8,$3F,$CF,$DA,$FF,$49,$6B,$FE,$1C
	.dc.b	$77,$AF,$FB,$21,$BA,$E6,$BE,$66,$1F,$E4,$BB,$FA,$15,$F9,$B7,$CB
	.dc.b	$E0,$68,$F4,$F4,$48,$DB,$5F,$A5,$4F,$A4,$F7,$FA,$1F,$67,$EE,$DF
	.dc.b	$0F,$4C,$DF,$82,$6B,$F6,$F2,$47,$1C,$7F,$05,$9F,$F8,$73,$5F,$F7
	.dc.b	$1E,$3F,$EE,$46,$EB,$AD,$66,$5F,$D8,$2F,$FC,$46,$FF,$6F,$6B,$FE
	.dc.b	$42,$6B,$FE,$4B,$7B,$4E,$67,$9C,$FE,$3F,$EE,$DF,$45,$E4,$E9,$FC
	.dc.b	$B4,$3C,$DA,$66,$BF,$39,$8F,$E9,$B2,$FF,$7D,$B7,$2F,$16,$1C,$BE
	.dc.b	$33,$DF,$E7,$FE,$87,$C6,$7C,$3D,$33,$97,$49,$AF,$E8,$F1,$47,$13
	.dc.b	$F5,$5F,$65,$FE,$5F,$9F,$D7,$5B,$E3,$BB,$1A,$F4,$A4,$FF,$82,$2B
	.dc.b	$83,$FE,$0F,$6B,$FE,$19,$6B,$FE,$6C,$77,$AB,$5C,$57,$4A,$FD,$C6
	.dc.b	$0F,$8E,$F2,$EF,$FB,$FF,$BB,$FE,$89,$8F,$1B,$FF,$3F,$46,$31,$ED
	.dc.b	$5C,$27,$F8,$46,$5F,$45,$B1,$CF,$EF,$3E,$0F,$DE,$75,$F4,$E5,$70
	.dc.b	$8C,$B4,$87,$F0,$DC,$5F,$51,$FC,$4F,$EF,$2C,$FF,$17,$F8,$6F,$E5
	.dc.b	$6D,$F2,$B5,$52,$3F,$C6,$35,$C1,$FF,$1C,$35,$FF,$21,$35,$FF,$4A
	.dc.b	$BB,$DB,$E1,$42,$31,$B6,$AB,$F9,$36,$0F,$81,$FE,$D3,$E3,$BE,$F3
	.dc.b	$FB,$03,$71,$FE,$C0,$D7,$FB,$03,$5F,$EC,$1F,$76,$FB,$CF,$ED,$0D
	.dc.b	$8F,$ED,$0D,$7F,$B4,$35,$FE,$D1,$F7,$9F,$C7,$5F,$ED,$8D,$73,$E4
	.dc.b	$7F,$73,$8C,$47,$FB,$F7,$8B,$F8,$A3,$FD,$CD,$AF,$F7,$36,$BB,$F8
	.dc.b	$3F,$F2,$EA,$FB,$C7,$7E,$1B,$1F,$E6,$26,$BF,$E6,$96,$BF,$DF,$FA
	.dc.b	$3F,$A6,$B7,$E0,$3F,$C1,$11,$88,$FF,$68,$7A,$28,$C4,$7F,$C0,$1F
	.dc.b	$BB,$B2,$69,$2B,$FC,$25,$F7,$7F,$B6,$AF,$F0,$93,$5F,$F0,$93,$5F
	.dc.b	$F0,$93,$5F,$0F,$CB,$FC,$F6,$F0,$7F,$87,$63,$10,$C1,$18,$5F,$77
	.dc.b	$F2,$A7,$4D,$8C,$EB,$FE,$27,$8D,$4B,$EE,$DF,$3D,$F4,$1B,$1F,$E8
	.dc.b	$E6,$BF,$E9,$56,$BF,$E2,$BF,$86,$FD,$B3,$FE,$2F,$6B,$9F,$07,$FC
	.dc.b	$6D,$18,$8F,$9F,$F7,$7F,$C0,$5F,$63,$72,$39,$EE,$7D,$DB,$E1,$7F
	.dc.b	$CA,$10,$FA,$5F,$F2,$5E,$1F,$E5,$D6,$BF,$E5,$C6,$A8,$E1,$E3,$FF
	.dc.b	$92,$61,$F2,$DD,$29,$F0,$74,$FF,$2C,$CE,$8F,$F0,$E7,$C3,$FF,$1F
	.dc.b	$6F,$95,$FE,$1D,$8C,$47,$F8,$6D,$FE,$68,$F0,$F3,$3F,$3B,$18,$8F
	.dc.b	$F6,$03,$5F,$F6,$53,$51,$1F,$D9,$65,$FC,$29,$E4,$D9,$F1,$77,$FC
	.dc.b	$6F,$BB,$FE,$73,$E3,$FF,$9F,$9F,$F0,$D6,$1F,$E4,$3B,$46,$23,$FC
	.dc.b	$F3,$18,$F3,$79,$1F,$ED,$07,$11,$FE,$D7,$6B,$FE,$E1,$6B,$FE,$8F
	.dc.b	$9F,$0B,$F4,$5B,$1C,$0F,$E8,$87,$8B,$E6,$FF,$C8,$DF,$77,$FB,$C8
	.dc.b	$CF,$0F,$F9,$B5,$BF,$0F,$AF,$0E,$63,$FE,$E4,$BF,$E0,$2F,$C8,$0E
	.dc.b	$1F,$96,$DA,$FE,$E7,$1E,$DF,$4F,$F8,$8C,$62,$3D,$D1,$83,$DD,$A7
	.dc.b	$FA,$A1,$B1,$F8,$D6,$CF,$F7,$EE,$8F,$F3,$07,$DC,$BE,$EB,$F2,$56
	.dc.b	$FB,$AF,$F2,$AB,$F7,$BC,$B7,$DD,$FA,$67,$33,$FE,$5A,$8C,$4F,$B6
	.dc.b	$FA,$63,$F8,$6B,$DC,$62,$FB,$57,$C3,$3A,$BB,$71,$7D,$23,$F1,$F3
	.dc.b	$C6,$B9,$E1,$E8,$7C,$77,$CA,$BA,$9E,$9B,$E4,$AF,$94,$F9,$EF,$93
	.dc.b	$E0,$9F,$11,$C4,$A7,$18,$DB,$E4,$7E,$4F,$C0,$F8,$AF,$EF,$36,$7F
	.dc.b	$4E,$FA,$31,$B7,$E1,$7A,$BE,$9C,$78,$57,$A8,$F0,$4F,$8B,$1F,$4C
	.dc.b	$79,$4D,$57,$13,$A3,$C6,$3E,$83,$F9,$13,$99,$4A,$65,$DB,$F6,$FE
	.dc.b	$8D,$F7,$1B,$8F,$95,$18,$79,$4A,$31,$1C,$DA,$F7,$F9,$1F,$ED,$69
	.dc.b	$A4,$7F,$B7,$BF,$43,$07,$CA,$7F,$8D,$F2,$FC,$8D,$7F,$12,$AD,$4E
	.dc.b	$72,$ED,$3F,$26,$F8,$F0,$7D,$A8,$E0,$66,$B8,$3F,$16,$F9,$CE,$E3
	.dc.b	$CC,$6D,$C4,$6A,$3E,$B1,$D3,$0F,$D2,$DF,$82,$32,$72,$8C,$DF,$82
	.dc.b	$3F,$82,$57,$39,$F6,$C6,$F0,$FE,$11,$7C,$E7,$35,$B9,$F7,$C7,$4C
	.dc.b	$3C,$B7,$E7,$3F,$47,$6F,$BB,$FF,$27,$CF,$F2,$F6,$CF,$F2,$E5,$EE
	.dc.b	$FF,$22,$9F,$E7,$F5,$C7,$D7,$C5,$FF,$59,$BE,$33,$FC,$A1,$B5,$C1
	.dc.b	$E6,$F7,$AE,$34,$F2,$BF,$B4,$4F,$FB,$D5,$B1,$FD,$03,$E6,$BF,$D1
	.dc.b	$53,$9F,$0B,$8D,$6B,$1F,$F2,$CB,$67,$CF,$5C,$2F,$FD,$F5,$39,$7F
	.dc.b	$B8,$7D,$DD,$B3,$7E,$09,$F2,$FE,$E9,$CE,$56,$B8,$1B,$9C,$44,$CB
	.dc.b	$FA,$D5,$FF,$B5,$DF,$EB,$95,$D7,$97,$E0,$FC,$B3,$B6,$FC,$F6,$1B
	.dc.b	$7F,$40,$FE,$91,$29,$CE,$9F,$E0,$99,$CD,$63,$4F,$F3,$DF,$DD,$BE
	.dc.b	$3B,$E7,$B6,$BA,$5F,$F7,$AC,$7A,$46,$21,$3B,$7B,$1F,$E3,$FE,$A7
	.dc.b	$CC,$7C,$5F,$BA,$7D,$DF,$E1,$3D,$1F,$E0,$F3,$98,$D7,$C4,$4C,$BF
	.dc.b	$BF,$5F,$06,$C7,$CF,$FB,$BF,$D4,$6D,$FE,$B5,$C7,$FD,$7D,$F1,$59
	.dc.b	$EF,$ED,$82,$3C,$1F,$76,$F3,$E9,$8E,$8B,$5E,$DF,$BB,$FC,$9F,$0F
	.dc.b	$B0,$38,$CF,$C1,$F7,$5F,$12,$75,$39,$AF,$F3,$AF,$DD,$BE,$E3,$E0
	.dc.b	$36,$3E,$6E,$C8,$FD,$1D,$FF,$99,$DD,$3F,$32,$1F,$53,$A6,$F8,$3F
	.dc.b	$2E,$3C,$13,$FE,$6F,$75,$93,$E5,$7E,$83,$71,$F0,$BF,$E5,$56,$C7
	.dc.b	$D5,$F9,$3F,$B3,$1E,$66,$DF,$8D,$E4,$DE,$7F,$B9,$F7,$7F,$9B,$8F
	.dc.b	$43,$77,$FA,$CB,$ED,$77,$E0,$7E,$FB,$EE,$DE,$8D,$3F,$0D,$F7,$4F
	.dc.b	$96,$C6,$BF,$23,$EE,$FD,$FC,$66,$8F,$8E,$FC,$73,$AF,$7B,$E5,$ED
	.dc.b	$F8,$76,$DF,$FA,$A9,$F9,$86,$37,$FF,$4C,$65,$FC,$0F,$13,$7E,$FB
	.dc.b	$F8,$0D,$FD,$A8,$D7,$DE,$7C,$BB,$FB,$0D,$C7,$F2,$CF,$E4,$6B,$F0
	.dc.b	$2E,$5F,$F2,$07,$03,$EE,$FF,$90,$B6,$3F,$CC,$1F,$73,$70,$FC,$A4
	.dc.b	$DC,$3F,$DE,$AD,$AE,$77,$FC,$56,$F1,$7F,$E7,$4A,$AF,$D3,$9E,$E6
	.dc.b	$CB,$5F,$BF,$97,$D0,$79,$0F,$DD,$7C,$5B,$E2,$B1,$5B,$9B,$87,$E6
	.dc.b	$6B,$53,$93,$A7,$83,$EE,$3F,$95,$7D,$DF,$F3,$3D,$3D,$71,$99,$D9
	.dc.b	$FE,$81,$9D,$3F,$22,$3F,$A3,$FC,$73,$8A,$9F,$C3,$FE,$05,$F5,$38
	.dc.b	$BF,$E3,$55,$B6,$D6,$E7,$59,$FF,$01,$AD,$3F,$B3,$FB,$B7,$D2,$EB
	.dc.b	$4F,$E2,$51,$8C,$7A,$FD,$D7,$90,$FA,$5E,$AD,$8C,$93,$0F,$F4,$CF
	.dc.b	$27,$F0,$BC,$13,$FE,$83,$C9,$87,$BB,$B2,$4F,$F0,$AF,$BB,$FE,$B1
	.dc.b	$8F,$FB,$BA,$72,$FA,$5F,$E6,$11,$AB,$FD,$7D,$7F,$3B,$FB,$97,$C4
	.dc.b	$FF,$70,$6C,$7C,$FC,$71,$32,$38,$3F,$62,$31,$8F,$F2,$7A,$E2,$D9
	.dc.b	$D5,$70,$3C,$1F,$3D,$DC,$FE,$89,$F7,$4F,$7F,$13,$A2,$5F,$F2,$39
	.dc.b	$D4,$63,$5F,$8F,$3F,$ED,$0C,$7F,$DC,$B3,$ED,$F9,$1F,$D1,$79,$3E
	.dc.b	$D6,$46,$4F,$D6,$C6,$2B,$FA,$07,$97,$FC,$2E,$7F,$B0,$36,$3F,$9A
	.dc.b	$7D,$C7,$F3,$87,$F9,$AF,$8B,$FE,$18,$9D,$65,$E4,$9C,$DF,$F4,$0B
	.dc.b	$E1,$5A,$8F,$F6,$03,$67,$FD,$D2,$DF,$11,$5C,$30,$ED,$CB,$30,$FD
	.dc.b	$27,$AD,$71,$F4,$CD,$CB,$FD,$AF,$DA,$AC,$5F,$83,$E3,$7D,$07,$F7
	.dc.b	$5F,$B8,$FE,$F9,$7E,$9D,$7F,$8B,$65,$E4,$AD,$DF,$F5,$AB,$DB,$CC
	.dc.b	$FE,$BB,$C7,$F5,$AB,$FA,$D3,$70,$FE,$F0,$BE,$9B,$FC,$08,$E6,$94
	.dc.b	$DD,$39,$8C,$36,$33,$5C,$B9,$3F,$C1,$DF,$6F,$FD,$5B,$1F,$F0,$33
	.dc.b	$71,$EB,$68,$F6,$B7,$8B,$89,$27,$FB,$BC,$E8,$E5,$97,$97,$ED,$DF
	.dc.b	$63,$FE,$1D,$6D,$7F,$7E,$7F,$72,$D8,$D1,$1D,$74,$DA,$FF,$00,$7D
	.dc.b	$BB,$E5,$BF,$C5,$0D,$AF,$F0,$4B,$FE,$25,$9D,$CE,$89,$3A,$F8,$0B
	.dc.b	$3D,$39,$3F,$C6,$5F,$6F,$FE,$2A,$AF,$87,$87,$E1,$5F,$F0,$AC,$66
	.dc.b	$BE,$94,$E5,$FF,$0A,$46,$76,$4E,$77,$FF,$1A,$4F,$1A,$7C,$BF,$F4
	.dc.b	$27,$FD,$79,$EB,$FD,$E7,$3F,$43,$FB,$AF,$91,$4E,$6B,$84,$62,$BF
	.dc.b	$0B,$0A,$E1,$5B,$F3,$AD,$6A,$6B,$E2,$8B,$5C,$74,$FF,$63,$5F,$8F
	.dc.b	$ED,$FF,$D3,$BF,$2F,$8C,$3F,$27,$18,$7D,$27,$5A,$D2,$8D,$65,$E7
	.dc.b	$8C,$7C,$8A,$F7,$5A,$C3,$DD,$9F,$9A,$31,$7E,$0B,$E0,$F8,$0E,$77
	.dc.b	$E0,$CB,$D5,$19,$AE,$2E,$8C,$57,$2C,$7C,$95,$C1,$F3,$57,$02,$4E
	.dc.b	$F1,$F2,$E5,$E8,$8C,$61,$E5,$C7,$07,$8F,$2F,$54,$EA,$7C,$58,$F6
	.dc.b	$8C,$65,$E1,$C7,$E8,$46,$F2,$F0,$E3,$CF,$2E,$B8,$FA,$A2,$E3,$EA
	.dc.b	$CB,$BC,$63,$1E,$6F,$1A,$E0,$4A,$F1,$46,$5B,$19,$9E,$2F,$CF,$CB
	.dc.b	$CC,$FB,$CE,$67,$E6,$4F,$9C,$5C,$7F,$0C,$DC,$BE,$CB,$9F,$03,$FD
	.dc.b	$92,$F1,$3F,$10,$E1,$D7,$C7,$E5,$CB,$6B,$FC,$BC,$FE,$2A,$79,$E3
	.dc.b	$F6,$9B,$8F,$E6,$FB,$BF,$5D,$39,$AE,$04,$C3,$F1,$13,$AA,$FD,$15
	.dc.b	$F2,$9F,$D0,$B5,$F1,$9C,$0E,$57,$D5,$FC,$94,$6F,$05,$F5,$9E,$78
	.dc.b	$70,$C3,$BB,$FA,$49,$D4,$F6,$C3,$A1,$2B,$D9,$1C,$4E,$67,$82,$B9
	.dc.b	$39,$39,$47,$67,$D0,$6C,$93,$D5,$FD,$FD,$78,$AF,$E7,$C6,$AB,$95
	.dc.b	$7E,$A2,$B5,$87,$BC,$98,$78,$AF,$81,$E6,$AE,$CF,$F0,$7A,$ED,$E8
	.dc.b	$77,$64,$7F,$06,$8C,$BC,$E7,$C0,$E4,$FE,$2D,$C6,$FB,$77,$CE,$FF
	.dc.b	$2D,$E4,$FF,$73,$7C,$6F,$FB,$D2,$FF,$DA,$2D,$7F,$D9,$D5,$C1,$FF
	.dc.b	$6A,$B5,$FF,$69,$46,$E7,$9F,$C0,$75,$E4,$FE,$D7,$7B,$9F,$F7,$A2
	.dc.b	$F9,$DD,$61,$FD,$EE,$30,$F6,$8C,$E5,$E9,$F9,$0B,$5A,$AF,$2D,$60
	.dc.b	$CC,$F0,$FB,$73,$49,$3F,$EE,$E8,$79,$38,$8F,$F7,$64,$70,$75,$3C
	.dc.b	$E7,$C7,$F6,$FF,$E8,$89,$FF,$7B,$B5,$FF,$7A,$37,$2F,$F2,$8E,$48
	.dc.b	$7D,$77,$2F,$5E,$5F,$27,$AC,$7C,$0F,$B7,$7C,$75,$25,$7F,$88,$B1
	.dc.b	$EB,$3F,$E8,$28,$C6,$FF,$E1,$A8,$E3,$18,$3F,$C5,$0E,$14,$F2,$AF
	.dc.b	$07,$89,$F5,$25,$F6,$DF,$DA,$DE,$46,$3B,$F5,$19,$8E,$78,$FF,$79
	.dc.b	$F6,$BF,$CF,$11,$C2,$75,$8F,$F9,$85,$B5,$FD,$FD,$B3,$FE,$3E,$AE
	.dc.b	$2F,$43,$FC,$83,$0F,$B7,$F0,$92,$3C,$B2,$F9,$6C,$47,$13,$D7,$38
	.dc.b	$C3,$FC,$E5,$5B,$9F,$6F,$DB,$3C,$1F,$84,$D8,$CC,$7C,$5F,$B7,$7C
	.dc.b	$DE,$A2,$4E,$8F,$8F,$3A,$70,$78,$EB,$A3,$67,$A5,$78,$3E,$DE,$4F
	.dc.b	$91,$F0,$7C,$6E,$23,$94,$5C,$BF,$D1,$98,$AB,$D3,$E8,$CB,$F6,$BF
	.dc.b	$77,$ED,$D8,$66,$7C,$93,$70,$FF,$4B,$56,$A7,$F0,$39,$7F,$86,$67
	.dc.b	$FD,$34,$DE,$0F,$F1,$C3,$67,$FC,$EF,$F6,$EF,$9E,$FC,$64,$E6,$7E
	.dc.b	$77,$91,$83,$F2,$15,$B8,$EF,$19,$69,$F9,$28,$D3,$F7,$67,$F3,$98
	.dc.b	$79,$39,$BB,$9D,$3E,$1A,$D7,$AA,$2B,$F8,$5B,$4F,$9B,$D0,$93,$F0
	.dc.b	$1F,$D2,$78,$BF,$EA,$78,$C1,$0F,$14,$68,$FD,$57,$D8,$7F,$97,$A3
	.dc.b	$F5,$32,$FA,$9E,$66,$BC,$E9,$3F,$AE,$AE,$0F,$E6,$5B,$9F,$FB,$26
	.dc.b	$DF,$11,$C2,$B5,$76,$B5,$3E,$2F,$46,$57,$57,$F7,$DF,$6E,$F9,$2C
	.dc.b	$CF,$79,$B7,$FE,$D7,$AD,$4F,$97,$DC,$F6,$91,$FE,$0B,$0F,$96,$C4
	.dc.b	$7A,$66,$D7,$39,$FE,$1D,$F6,$FB,$AF,$E1,$AD,$7F,$70,$D9,$FF,$74
	.dc.b	$7D,$BF,$FB,$02,$77,$5C,$88,$FF,$16,$8D,$3C,$A3,$2D,$35,$F0,$1F
	.dc.b	$D3,$BE,$99,$F6,$F4,$93,$FC,$A7,$CC,$4F,$FB,$56,$31,$F6,$EF,$B3
	.dc.b	$FE,$AE,$D8,$CC,$63,$F1,$2F,$EB,$7F,$71,$FE,$B9,$9E,$8D,$7F,$8F
	.dc.b	$60,$8C,$6B,$FD,$29,$FC,$84,$EB,$DC,$FE,$18,$DF,$1B,$FB,$3C,$62
	.dc.b	$BF,$A8,$BC,$BC,$1F,$F5,$67,$DC,$7E,$7A,$FF,$9D,$B5,$FE,$4C,$DA
	.dc.b	$FE,$B3,$F7,$CF,$90,$FE,$E7,$D9,$FC,$DB,$43,$49,$E9,$FF,$78,$FB
	.dc.b	$9D,$13,$D1,$C4,$7F,$7C,$FB,$9F,$5E,$5F,$D9,$9B,$AF,$FB,$C9,$BA
	.dc.b	$FF,$BF,$D7,$CA,$FF,$81,$3E,$53,$E7,$4F,$C8,$6D,$FF,$70,$6D,$FF
	.dc.b	$BF,$17,$1F,$F8,$33,$EE,$7E,$34,$FF,$C0,$06,$DB,$64,$61,$BF,$15
	.dc.b	$FD,$97,$EE,$5F,$35,$49,$7F,$87,$FB,$1F,$E9,$8F,$DC,$F8,$A7,$89
	.dc.b	$FE,$1C,$6C,$7E,$35,$AE,$FB,$FF,$C4,$BB,$BE,$E3,$FC,$CF,$3F,$93
	.dc.b	$FB,$9F,$F1,$86,$E8,$F4,$46,$1F,$EF,$B3,$C1,$FE,$D5,$4F,$AB,$FF
	.dc.b	$1A,$5F,$F8,$2D,$B3,$FE,$04,$6C,$EF,$E1,$7F,$C6,$72,$F9,$9F,$2D
	.dc.b	$7F,$8C,$E9,$7C,$EF,$75,$FE,$0B,$F8,$BF,$E2,$F8,$7F,$82,$9A,$FF
	.dc.b	$75,$6D,$F2,$8F,$64,$7F,$27,$78,$79,$FC,$C8,$F3,$E2,$EC,$61,$E9
	.dc.b	$18,$7F,$C4,$54,$FB,$9F,$E5,$79,$7F,$8A,$DF,$AF,$1E,$13,$5F,$01
	.dc.b	$F4,$A3,$75,$FE,$28,$6C,$F3,$F4,$34,$6F,$C0,$B5,$E0,$AD,$92,$3F
	.dc.b	$C5,$DF,$73,$FA,$FC,$7E,$D5,$23,$0F,$DD,$6E,$1F,$EA,$8F,$73,$8F
	.dc.b	$AB,$2B,$FC,$6B,$5A,$F8,$EF,$EF,$11,$F8,$0E,$FF,$F6,$A6,$1E,$D6
	.dc.b	$E3,$FE,$4B,$6C,$FF,$91,$1F,$F6,$AC,$EE,$3F,$27,$1A,$8F,$D9,$C6
	.dc.b	$39,$FF,$C9,$1F,$6E,$FB,$9F,$F2,$6C,$BA,$FF,$94,$BF,$97,$F3,$9F
	.dc.b	$73,$F0,$27,$33,$FC,$BA,$31,$3F,$12,$FA,$63,$F9,$FB,$DC,$62,$FB
	.dc.b	$57,$CA,$3A,$BB,$71,$7D,$23,$F5,$13,$C6,$B9,$E1,$E8,$7C,$77,$CA
	.dc.b	$BA,$9E,$9B,$E4,$AF,$94,$F9,$EF,$93,$E0,$9F,$11,$C4,$A7,$18,$DB
	.dc.b	$E4,$6B,$F0,$CD,$C7,$8E,$FC,$F8,$72,$BF,$0D,$75,$3C,$E7,$12,$BA
	.dc.b	$7C,$11,$D3,$0F,$5D,$F5,$3D,$51,$D7,$0F,$B7,$5F,$96,$38,$94,$A6
	.dc.b	$6F,$EA,$37,$EE,$5F,$6D,$B8,$F9,$11,$87,$94,$A3,$11,$C6,$78,$3F
	.dc.b	$C7,$FE,$4E,$A6,$91,$DB,$F3,$B0,$7D,$A7,$F0,$DD,$3F,$1F,$5C,$CF
	.dc.b	$8F,$3A,$DF,$EA,$C6,$67,$F2,$B3,$98,$F9,$A6,$74,$FF,$76,$5F,$57
	.dc.b	$E2,$61,$E1,$8C,$4F,$C9,$8C,$61,$DA,$3D,$47,$8A,$76,$FF,$04,$9D
	.dc.b	$BE,$03,$78,$79,$27,$71,$E3,$3F,$D8,$CE,$1D,$DF,$9A,$FC,$17,$D2
	.dc.b	$FE,$1D,$68,$E3,$3A,$7C,$8F,$E6,$BF,$01,$7E,$2B,$FD,$72,$3F,$D9
	.dc.b	$D1,$8F,$0F,$F9,$54,$7F,$55,$8D,$F9,$98,$B4,$6E,$78,$DF,$39,$FF
	.dc.b	$60,$4B,$EB,$31,$AF,$F4,$58,$E1,$3A,$9C,$DF,$7D,$7F,$C3,$0F,$91
	.dc.b	$FF,$1C,$B7,$A7,$C5,$7F,$84,$DA,$D7,$0C,$3C,$2F,$06,$D6,$A7,$8F
	.dc.b	$E3,$FF,$E0,$CC,$3B,$56,$3C,$BE,$6D,$8C,$9F,$1A,$78,$47,$63,$45
	.dc.b	$8D,$4F,$27,$E6,$F7,$FF,$28,$6F,$37,$AA,$1D,$D6,$24,$FD,$4F,$A7
	.dc.b	$FC,$C6,$FF,$26,$F9,$63,$F3,$A7,$35,$F5,$DB,$1F,$5B,$F0,$57,$D1
	.dc.b	$FC,$89,$CF,$02,$73,$7F,$82,$FC,$59,$F0,$BF,$39,$B1,$F8,$6F,$C7
	.dc.b	$7F,$D8,$8E,$E7,$5F,$45,$D7,$EC,$6C,$FA,$5F,$AF,$83,$E6,$3E,$CF
	.dc.b	$81,$F5,$F8,$3F,$09,$39,$BF,$C9,$CE,$5F,$C9,$7C,$4F,$F8,$59,$FC
	.dc.b	$BB,$78,$3F,$1E,$BE,$E3,$F0,$AE,$70,$FC,$87,$C8,$7A,$2B,$F2,$AB
	.dc.b	$EE,$FF,$16,$E6,$77,$1F,$8E,$FC,$5B,$FD,$DF,$8F,$E9,$97,$CE,$E7
	.dc.b	$CC,$D3,$F9,$4F,$C5,$9F,$81,$DF,$47,$0D,$C3,$F5,$F9,$FB,$F4,$F1
	.dc.b	$FC,$4F,$8B,$8B,$C6,$FE,$7B,$C6,$FF,$63,$DF,$ED,$9B,$5F,$AE,$7E
	.dc.b	$D9,$8D,$FF,$5D,$F8,$97,$F7,$BA,$FD,$EB,$6B,$F6,$AF,$F3,$78,$CF
	.dc.b	$1F,$EB,$3F,$16,$FF,$66,$47,$F0,$45,$F7,$79,$F9,$2D,$3F,$B2,$E2
	.dc.b	$7D,$BD,$FA,$CC,$37,$0F,$E1,$BF,$8B,$74,$D3,$F8,$2B,$D2,$33,$8F
	.dc.b	$EF,$23,$13,$FC,$2B,$27,$47,$97,$E9,$BC,$CD,$C3,$F8,$CF,$E2,$4F
	.dc.b	$9A,$FE,$0A,$D8,$CF,$D7,$E2,$7F,$8D,$71,$3E,$E1,$F3,$B8,$5F,$25
	.dc.b	$63,$0D,$C3,$FC,$A7,$3A,$6C,$E9,$A6,$A7,$86,$BF,$E7,$AA,$E0,$FF
	.dc.b	$26,$FC,$4F,$E2,$61,$FC,$95,$B7,$FC,$C5,$F8,$71,$9E,$0F,$E4,$31
	.dc.b	$AA,$FE,$66,$DF,$20,$F0,$BE,$55,$FC,$A3,$F1,$5F,$E5,$F7,$FC,$E9
	.dc.b	$B6,$5E,$B6,$FA,$A6,$1F,$D1,$A7,$66,$9C,$47,$F4,$5B,$7D,$E3,$E1
	.dc.b	$30,$DA,$FF,$3D,$FD,$C3,$D0,$CD,$49,$FF,$3F,$FE,$17,$E6,$C3,$9C
	.dc.b	$FD,$88,$C6,$1C,$E7,$FD,$4B,$18,$9D,$63,$EF,$F8,$4F,$81,$F5,$37
	.dc.b	$1F,$53,$70,$FE,$AF,$A2,$31,$C3,$FD,$5E,$31,$5F,$07,$D8,$F9,$47
	.dc.b	$CB,$F8,$EB,$19,$26,$1D,$23,$2F,$E0,$7B,$FF,$1D,$C7,$F8,$F6,$D7
	.dc.b	$E3,$1B,$3B,$8F,$C7,$CE,$DA,$6C,$B1,$EA,$F8,$8F,$BC,$7D,$FF,$6E
	.dc.b	$CF,$D9,$8B,$3F,$02,$6E,$7F,$EB,$9F,$90,$7A,$EF,$A0,$FF,$00,$46
	.dc.b	$3E,$AA,$CE,$72,$FF,$6D,$5F,$18,$4E,$AF,$6D,$AD,$9F,$EE,$1B,$E8
	.dc.b	$FE,$13,$CE,$FD,$AF,$43,$D6,$B1,$86,$CF,$E2,$9B,$A7,$FB,$9A,$B8
	.dc.b	$4F,$33,$45,$8D,$4F,$B3,$F3,$6F,$EB,$B0,$D3,$E1,$43,$E7,$B5,$EC
	.dc.b	$ED,$B8,$7A,$AD,$9A,$F8,$4E,$63,$17,$C6,$3F,$55,$EE,$FE,$CF,$DE
	.dc.b	$FD,$A3,$63,$27,$EB,$E3,$2D,$3F,$65,$87,$EE,$FE,$27,$F7,$8D,$8C
	.dc.b	$9F,$B7,$8C,$B4,$FF,$1F,$FE,$36,$F8,$BF,$F1,$5B,$63,$3F,$05,$89
	.dc.b	$91,$C1,$DC,$62,$FF,$C7,$F5,$F2,$B6,$FF,$1F,$3F,$0A,$33,$9F,$6D
	.dc.b	$BB,$B7,$1F,$1B,$67,$9E,$FE,$4E,$1F,$2B,$6B,$D2,$D9,$DC,$7B,$27
	.dc.b	$D7,$EC,$48,$C1,$F0,$1C,$4F,$F0,$0F,$99,$B9,$FE,$FD,$FE,$35,$BF
	.dc.b	$F9,$6F,$E2,$5F,$4F,$FE,$FE,$8F,$E4,$17,$FE,$99,$8C,$78,$4C,$BF
	.dc.b	$D3,$D1,$BF,$1F,$15,$B0,$7F,$90,$2B,$38,$76,$8C,$E1,$DA,$31,$1B
	.dc.b	$C3,$AB,$AB,$E3,$87,$48,$7D,$B7,$F5,$CE,$4D,$C6,$3C,$D4,$63,$2E
	.dc.b	$D7,$C6,$F8,$5F,$9C,$E1,$5C,$A3,$39,$79,$2F,$8D,$7B,$27,$37,$CF
	.dc.b	$0E,$F3,$CE,$72,$FA,$72,$E9,$3A,$C3,$DF,$5A,$AF,$09,$E3,$C3,$BB
	.dc.b	$89,$DD,$71,$C3,$C1,$8F,$B7,$2F,$8B,$7E,$5A,$D3,$67,$C2,$6F,$2B
	.dc.b	$1E,$79,$DB,$DE,$FE,$E4,$EB,$0F,$0D,$D9,$CD,$71,$8C,$3D,$CF,$56
	.dc.b	$1E,$47,$A1,$9A,$E6,$EA,$F8,$BA,$26,$3F,$0B,$2F,$8C,$4C,$BC,$F1
	.dc.b	$97,$D3,$1F,$72,$7F,$0C,$7C,$C2,$B9,$7E,$49,$A3,$F1,$7E,$6B,$ED
	.dc.b	$B1,$18,$6C,$FF,$8A,$1B,$3B,$9E,$53,$97,$F0,$51,$9C,$7E,$FB,$74
	.dc.b	$FC,$97,$C8,$7F,$BA,$5F,$CA,$46,$6B,$F1,$CD,$7C,$47,$E3,$67,$83
	.dc.b	$AC,$3F,$31,$3B,$C3,$C5,$87,$E3,$AF,$77,$CA,$3C,$77,$C0,$FD,$04
	.dc.b	$EA,$B8,$E1,$E4,$9D,$5F,$84,$B7,$C9,$F7,$C7,$27,$33,$C5,$C9,$C4
	.dc.b	$B3,$C4,$91,$F1,$23,$89,$ED,$36,$E5,$A6,$63,$8B,$FB,$5B,$F5,$61
	.dc.b	$F0,$C9,$7E,$59,$E2,$7C,$3B,$F3,$5F,$C7,$8F,$E0,$5E,$4F,$C3,$C2
	.dc.b	$46,$08,$7A,$E7,$83,$FC,$06,$7E,$AF,$9A,$E8,$7D,$0F,$F1,$9E,$4F
	.dc.b	$F1,$87,$CC,$DE,$1F,$E4,$F7,$FE,$55,$8C,$62,$8E,$57,$86,$C7,$F5
	.dc.b	$E8,$C6,$1D,$A3,$FB,$0F,$D8,$63,$D4,$CB,$23,$FB,$2C,$65,$D9,$0F
	.dc.b	$EB,$EF,$4F,$92,$EB,$EA,$C9,$FE,$DB,$19,$7F,$B7,$C6,$5F,$EC,$F1
	.dc.b	$E4,$F4,$E4,$7F,$74,$7C,$FE,$F6,$23,$FB,$6C,$5D,$BF,$B6,$CF,$B3
	.dc.b	$EA,$7D,$BF,$1F,$2F,$FB,$EC,$65,$FE,$FF,$19,$7F,$BB,$C7,$C4,$E4
	.dc.b	$F8,$BF,$17,$27,$FC,$0B,$19,$7F,$C0,$F1,$97,$FB,$FC,$7D,$2F,$1D
	.dc.b	$F8,$9F,$FA,$DA,$7F,$C4,$0B,$DD,$D3,$63,$3E,$EE,$0C,$CC,$8C,$9F
	.dc.b	$E2,$D1,$F5,$FF,$EB,$D7,$0B,$87,$F0,$71,$A6,$C6,$E3,$FC,$79,$CF
	.dc.b	$FD,$96,$31,$BF,$F7,$49,$E0,$74,$AE,$3C,$52,$30,$43,$99,$FD,$EA
	.dc.b	$B9,$DF,$F9,$6F,$F1,$77,$1F,$C5,$73,$5F,$E4,$66,$EA,$77,$1F,$E4
	.dc.b	$76,$BD,$A3,$58,$73,$3A,$57,$E7,$3E,$C2,$57,$F8,$F1,$FC,$FF,$BB
	.dc.b	$88,$FF,$1A,$C5,$D3,$FC,$6B,$3F,$23,$EB,$9E,$3D,$F8,$99,$87,$D1
	.dc.b	$9C,$B6,$72,$FC,$21,$09,$E9,$32,$79,$47,$0B,$FA,$15,$D2,$7F,$DC
	.dc.b	$5F,$3D,$FB,$9F,$81,$34,$FD,$CF,$87,$67,$FC,$E1,$DF,$FE,$DE,$9C
	.dc.b	$47,$F9,$4A,$79,$78,$7C,$FC,$29,$3C,$30,$4E,$A3,$13,$A7,$FD,$E9
	.dc.b	$1F,$8C,$FA,$67,$97,$F8,$EF,$AE,$7D,$BD,$35,$12,$FF,$31,$3A,$6C
	.dc.b	$E5,$F9,$15,$D1,$FC,$F3,$5C,$D7,$7F,$77,$F8,$E7,$80,$F8,$3A,$4A
	.dc.b	$F8,$71,$FE,$D9,$6C,$65,$FC,$BE,$E9,$E5,$1C,$3E,$0F,$F9,$47,$C3
	.dc.b	$3D,$3A,$49,$F9,$F1,$FE,$B3,$6C,$65,$EB,$1C,$5D,$FA,$9F,$B4,$FD
	.dc.b	$2B,$FD,$55,$1F,$BB,$7F,$D2,$0D,$F6,$75,$EC,$48,$C1,$0F,$C7,$4F
	.dc.b	$26,$C7,$59,$91,$D6,$3F,$A1,$7E,$31,$FE,$D9,$8F,$C3,$B5,$F8,$6D
	.dc.b	$E9,$FF,$3A,$EA,$8F,$F6,$CC,$FE,$3B,$BF,$E3,$BF,$29,$F9,$CD,$D7
	.dc.b	$FA,$84,$6A,$F9,$FD,$F3,$CB,$FE,$A8,$BC,$CA,$62,$3F,$AC,$7E,$31
	.dc.b	$BC,$7F,$3A,$ED,$B3,$A9,$EF,$EE,$F8,$3D,$89,$5E,$78,$FE,$C5,$F8
	.dc.b	$C3,$E1,$E9,$27,$F5,$31,$E7,$F2,$BF,$B6,$F9,$EF,$88,$FE,$D4,$DA
	.dc.b	$FD,$7B,$5F,$C5,$BD,$5F,$D9,$F8,$BF,$C6,$37,$F9,$6F,$F2,$0E,$3F
	.dc.b	$E0,$BC,$3F,$DD,$63,$58,$FF,$7C,$F2,$9A,$FF,$79,$5F,$0F,$4C,$47
	.dc.b	$F7,$DE,$4F,$F2,$66,$7F,$9C,$5E,$56,$4D,$78,$FD,$FD,$2F,$07,$A7
	.dc.b	$2F,$CF,$1F,$E0,$7E,$CF,$E1,$F1,$FC,$31,$AF,$EA,$1B,$C1,$F8,$6F
	.dc.b	$B3,$F0,$BF,$88,$F9,$A7,$3F,$F8,$47,$F1,$BF,$E0,$18,$7E,$C1,$B5
	.dc.b	$B9,$FF,$0B,$7E,$31,$FD,$9E,$BF,$6B,$E9,$F6,$E3,$FF,$0F,$B6,$3F
	.dc.b	$C3,$D3,$BB,$FC,$BA,$FA,$9E,$64,$9F,$E4,$BF,$8A,$FE,$25,$EB,$E7
	.dc.b	$7F,$89,$74,$FF,$59,$C7,$F0,$46,$E1,$FA,$2C,$7F,$92,$CF,$F2,$EC
	.dc.b	$5F,$3D,$FE,$2D,$9C,$CF,$43,$51,$EC,$23,$83,$96,$7F,$E3,$5F,$A7
	.dc.b	$7D,$CD,$CE,$57,$F3,$C9,$CB,$67,$2F,$F1,$A8,$F0,$E5,$FB,$87,$71
	.dc.b	$99,$DC,$E6,$BC,$11,$E1,$F9,$EF,$17,$CC,$4B,$F0,$46,$5F,$14,$65
	.dc.b	$F0,$47,$F9,$37,$F0,$9F,$F0,$D4,$7F,$2C,$F0,$5F,$17,$FE,$51,$6C
	.dc.b	$7F,$56,$9D,$E5,$FC,$39,$BA,$7F,$97,$1B,$3F,$E5,$B6,$F8,$25,$F3
	.dc.b	$5F,$E1,$ED,$BF,$85,$60,$B7,$D4,$7E,$D7,$2F,$D5,$C6,$63,$1D,$FB
	.dc.b	$8D,$CD,$BE,$71,$83,$A3,$8A,$E3,$1C,$E7,$91,$D9,$E0,$27,$84,$75
	.dc.b	$F2,$18,$71,$AF,$44,$F1,$BE,$B1,$CA,$F8,$BE,$2B,$E2,$F6,$AE,$07
	.dc.b	$8B,$0F,$65,$FF,$15,$BE,$35,$E3,$BE,$33,$D8,$97,$E5,$C3,$C9,$5B
	.dc.b	$6C,$76,$8F,$3E,$1D,$A3,$84,$79,$EF,$AC,$78,$E3,$8C,$ED,$F5,$57
	.dc.b	$6C,$3E,$7C,$7F,$BC,$9A,$F4,$31,$7F,$01,$D7,$84,$FA,$9F,$F7,$8D
	.dc.b	$7F,$93,$9B,$19,$9F,$F7,$C4,$E5,$FF,$7B,$8F,$89,$AD,$32,$E0,$8F
	.dc.b	$DC,$C3,$FA,$D6,$FF,$1A,$3F,$21,$3F,$22,$73,$1F,$82,$32,$71,$F2
	.dc.b	$95,$C6,$3C,$1A,$FF,$4B,$AE,$2F,$F4,$D8,$E2,$E6,$B8,$C7,$53,$8E
	.dc.b	$BF,$EC,$78,$CC,$F1,$9C,$9F,$A5,$AE,$B5,$FA,$18,$EA,$F1,$BF,$26
	.dc.b	$1E,$BC,$39,$CE,$EB,$99,$F9,$A8,$CB,$CE,$78,$57,$28,$F8,$0D,$C3
	.dc.b	$D4,$DB,$F0,$4F,$EB,$F8,$1F,$53,$FE,$4E,$AF,$F2,$EB,$67,$FC,$72
	.dc.b	$BE,$3B,$E1,$E7,$FE,$18,$8D,$37,$D8,$CD,$FF,$BA,$27,$2D,$AC,$DF
	.dc.b	$FB,$A6,$72,$FF,$BA,$F4,$7C,$BE,$B2,$FF,$0D,$67,$67,$38,$FF,$9C
	.dc.b	$9F,$F0,$43,$6B,$53,$B6,$C7,$F9,$E3,$EA,$1F,$27,$CF,$1E,$D1,$89
	.dc.b	$E9,$18,$9D,$47,$16,$C7,$F9,$72,$78,$3F,$E5,$EC,$3F,$92,$56,$CE
	.dc.b	$67,$FA,$23,$E2,$BF,$BB,$B7,$97,$DD,$18,$AF,$F3,$3B,$FE,$14,$9D
	.dc.b	$1F,$E9,$0F,$98,$FC,$FF,$07,$F9,$22,$75,$5F,$E6,$E9,$CD,$7F,$35
	.dc.b	$D5,$18,$94,$3C,$17,$D4,$7C,$FA,$D4,$E6,$3E,$97,$D4,$3E,$5A,$9F
	.dc.b	$57,$EA,$1F,$75,$8C,$FF,$D4,$B4,$F7,$7A,$79,$FC,$9B,$3F,$EA,$AD
	.dc.b	$3F,$D7,$0D,$F8,$9C,$ED,$FE,$05,$7F,$05,$F2,$1F,$BA,$9F,$F5,$8B
	.dc.b	$67,$FD,$1A,$FF,$81,$A3,$27,$E2,$7E,$9B,$CA,$FF,$AE,$17,$D4,$E7
	.dc.b	$4D,$3F,$8F,$FA,$9F,$EB,$F5,$FE,$BE,$6E,$1F,$E9,$A5,$E4,$6F,$C5
	.dc.b	$FC,$B7,$D4,$FF,$72,$BF,$F6,$2B,$6F,$FD,$42,$FC,$38,$CF,$91,$E5
	.dc.b	$D7,$CD,$F5,$36,$7C,$CF,$E8,$7C,$2F,$F5,$44,$7F,$B4,$1B,$87,$FA
	.dc.b	$B1,$FA,$E6,$3A,$BF,$D1,$7F,$53,$FE,$7F,$AF,$F6,$BB,$6B,$FD,$66
	.dc.b	$BC,$7F,$D7,$39,$AF,$F4,$9F,$D3,$BE,$2F,$F6,$4B,$EA,$6C,$61,$FD
	.dc.b	$B6,$FF,$13,$2F,$F7,$2B,$72,$FF,$60,$BC,$E3,$3C,$1F,$BD,$FA,$9C
	.dc.b	$57,$FB,$19,$B9,$7F,$A9,$1F,$F6,$FD,$7F,$BC,$1C,$BF,$9D,$C9,$CD
	.dc.b	$E6,$F9,$9F,$3B,$67,$FD,$93,$F0,$5F,$C2,$FE,$2C,$DF,$99,$C3,$FC
	.dc.b	$53,$27,$D4,$BE,$87,$0B,$E8,$2C,$61,$B8,$FF,$3F,$3F,$A0,$4E,$DC
	.dc.b	$6D,$FE,$12,$8E,$91,$8B,$EF,$5C,$AB,$FC,$2F,$3C,$A3,$13,$FC,$C2
	.dc.b	$BA,$57,$FB,$B3,$E5,$FE,$1B,$FB,$C6,$F8,$C7,$EC,$61,$E2,$8C,$72
	.dc.b	$7F,$2D,$9D,$93,$93,$11,$FC,$C7,$EA,$3F,$9A,$D1,$A7,$F4,$3B,$E8
	.dc.b	$E3,$A3,$81,$C7,$D1,$B3,$F0,$B2,$FF,$53,$FC,$C7,$F4,$26,$C7,$C3
	.dc.b	$FA,$57,$99,$9F,$88,$5D,$25,$6E,$6D,$FC,$83,$FC,$A9,$5C,$27,$E7
	.dc.b	$7D,$4F,$F4,$E9,$EE,$6F,$85,$1E,$CF,$97,$F8,$1F,$17,$F0,$5B,$7E
	.dc.b	$E6,$DF,$7D,$FE,$24,$63,$0F,$F7,$DE,$FF,$D5,$2B,$4D,$8C,$FC,$87
	.dc.b	$CD,$F5,$E6,$3F,$29,$FA,$11,$8B,$F3,$72,$BE,$A7,$F2,$7C,$9F,$94
	.dc.b	$6E,$1F,$8B,$9D,$36,$74,$D3,$53,$C2,$BA,$B7,$1F,$4F,$C7,$7E,$04
	.dc.b	$FC,$17,$B3,$6F,$EF,$C5,$BF,$6C,$DA,$FF,$13,$FC,$FF,$F5,$0F,$A9
	.dc.b	$FE,$FE,$FF,$52,$FA,$9B,$EC,$26,$C6,$1B,$97,$EA,$2B,$93,$69,$5F
	.dc.b	$B7,$E8,$7C,$D5,$F5,$30,$DB,$FC,$63,$74,$FC,$05,$70,$9F,$C8,$57
	.dc.b	$06,$9C,$0B,$3F,$8D,$C5,$F5,$3F,$D8,$E7,$CC,$BE,$6B,$DE,$E6,$B8
	.dc.b	$39,$9D,$FB,$5A,$AD,$B6,$3F,$57,$EE,$7E,$C7,$E3,$FF,$64,$D8,$C9
	.dc.b	$FB,$08,$CB,$4F,$D9,$E1,$FB,$9F,$A9,$FD,$D3,$63,$27,$EE,$23,$2D
	.dc.b	$3F,$83,$7D,$33,$E7,$7F,$C6,$4D,$8C,$FC,$1E,$26,$47,$07,$71,$8B
	.dc.b	$FD,$35,$7C,$3E,$4F,$F3,$5F,$72,$3E,$24,$EB,$4E,$FA,$F8,$1B,$97
	.dc.b	$91,$B3,$D3,$83,$CB,$C5,$E6,$6D,$7A,$9B,$3B,$8F,$6C,$FB,$3D,$79
	.dc.b	$18,$3E,$0B,$89,$F4,$BC,$B3,$F4,$E5,$FD,$F5,$FF,$42,$CB,$7F,$F3
	.dc.b	$47,$C3,$3E,$5B,$FD,$C1,$1F,$C9,$E7,$FD,$45,$18,$F2,$3F,$D1,$B5
	.dc.b	$B9,$FF,$76,$56,$E3,$11,$D6,$77,$5D,$3C,$8B,$3C,$26,$BE,$38,$BF
	.dc.b	$53,$FD,$6F,$93,$FD,$53,$7F,$EF,$38,$C5,$79,$0F,$0D,$70,$F3,$57
	.dc.b	$C2,$7F,$E0,$05,$70,$C7,$CA,$DC,$3C,$9B,$6E,$93,$97,$8E,$1D,$1F
	.dc.b	$01,$A7,$17,$F0,$33,$F8,$58,$79,$E7,$2D,$C7,$B6,$1D,$70,$F8,$18
	.dc.b	$E2,$F7,$1E,$0A,$C1,$E5,$C3,$AC,$64,$F1,$DF,$33,$A4,$6A,$3C,$D3
	.dc.b	$C9,$DD,$76,$7C,$26,$23,$58,$FC,$69,$D5,$73,$BE,$D1,$C5,$EF,$97
	.dc.b	$D1,$BE,$39,$7B,$F1,$F8,$F8,$71,$C3,$81,$2B,$D9,$7C,$6B,$C4,$4C
	.dc.b	$7E,$64,$65,$F2,$4F,$FB,$3E,$7D,$A7,$FB,$47,$1F,$C2,$BF,$ED,$96
	.dc.b	$BF,$89,$F3,$5F,$31,$C6,$7F,$C4,$CD,$8C,$CF,$36,$BC,$DC,$DF,$E2
	.dc.b	$9B,$9F,$5A,$FB,$BF,$20,$BF,$BF,$5A,$7F,$26,$D7,$F1,$ED,$C1,$3C
	.dc.b	$67,$EE,$CE,$49,$5D,$A3,$AB,$1D,$B9,$69,$30,$F4,$46,$5B,$19,$C3
	.dc.b	$D4,$FE,$32,$FC,$53,$A2,$4E,$9F,$D5,$C6,$A3,$04,$9F,$7D,$77,$8C
	.dc.b	$47,$C0,$8F,$45,$72,$8B,$3A,$C3,$EA,$C7,$8C,$C9,$D0,$F4,$1F,$BE
	.dc.b	$BF,$13,$9A,$F3,$B6,$BF,$85,$4E,$A3,$F7,$FE,$6D,$BF,$A3,$1E,$73
	.dc.b	$F8,$4C,$64,$D4,$C8,$FD,$DC,$78,$39,$FF,$A3,$7D,$47,$D9,$E2,$FE
	.dc.b	$62,$DE,$1F,$F6,$F7,$1A,$31,$87,$FB,$7A,$1F,$51,$8C,$39,$CF,$F5
	.dc.b	$EA,$C5,$71,$8F,$EC,$35,$8F,$4F,$2C,$8F,$EC,$B1,$97,$64,$3F,$AF
	.dc.b	$BD,$3D,$CE,$BE,$A4,$9F,$ED,$B1,$97,$FB,$7C,$65,$FE,$CF,$1E,$4F
	.dc.b	$4A,$47,$F7,$47,$CF,$F1,$F8,$8F,$ED,$B1,$75,$FE,$AB,$3E,$CF,$A1
	.dc.b	$F6,$FC,$5C,$BF,$EF,$B1,$97,$FB,$FC,$65,$FE,$EF,$1F,$13,$8F,$E2
	.dc.b	$FC,$44,$9F,$F0,$2C,$65,$FF,$03,$C6,$5F,$EF,$F1,$F4,$BC,$67,$CE
	.dc.b	$FF,$AD,$A7,$FC,$78,$BD,$CD,$36,$33,$EE,$60,$CC,$C8,$C9,$FE,$8B
	.dc.b	$FA,$8F,$F5,$DB,$78,$7F,$02,$A1,$18,$77,$1F,$E8,$73,$5C,$39,$99
	.dc.b	$1D,$63,$81,$87,$84,$F4,$74,$7F,$95,$FE,$5F,$FD,$D8,$D9,$F8,$51
	.dc.b	$E0,$E0,$FF,$32,$BA,$9F,$F2,$63,$63,$33,$C0,$E3,$C5,$F9,$6F,$A8
	.dc.b	$97,$FE,$36,$7F,$33,$EE,$62,$3F,$C5,$F1,$6F,$FC,$2F,$3F,$13,$E7
	.dc.b	$DF,$51,$7E,$22,$61,$FE,$B5,$9C,$B6,$72,$FC,$78,$4F,$39,$93,$C6
	.dc.b	$37,$7F,$E2,$4A,$E7,$3F,$ED,$7F,$A8,$FD,$97,$C1,$1A,$7E,$CB,$C3
	.dc.b	$B3,$FE,$D7,$9D,$36,$74,$E7,$C1,$E7,$E0,$C8,$C1,$0F,$F7,$33,$97
	.dc.b	$FD,$D1,$1F,$86,$FA,$67,$95,$F8,$8F,$A8,$7D,$45,$35,$12,$FF,$DB
	.dc.b	$55,$C1,$B1,$C7,$DD,$D4,$7E,$36,$74,$D9,$D3,$FE,$F3,$EE,$7C,$17
	.dc.b	$F1,$56,$C6,$63,$1F,$07,$96,$46,$4E,$51,$FC,$77,$E1,$BF,$D4,$71
	.dc.b	$FE,$FF,$6B,$FE,$D7,$6F,$17,$F9,$C9,$7A,$7C,$7D,$39,$5C,$5E,$91
	.dc.b	$FA,$CF,$A4,$FF,$4F,$4F,$FB,$ED,$FF,$43,$37,$D8,$D7,$AF,$23,$04
	.dc.b	$3F,$61,$3C,$9B,$1D,$66,$47,$58,$FE,$71,$F5,$1F,$EC,$78,$FF,$01
	.dc.b	$65,$FE,$8A,$8C,$47,$FB,$CD,$B1,$9F,$77,$FA,$07,$7F,$CC,$AF,$F7
	.dc.b	$C2,$D7,$26,$B6,$FE,$8D,$F5,$0F,$2A,$92,$FC,$3E,$8F,$F8,$4F,$D1
	.dc.b	$53,$CE,$FE,$9D,$F5,$1C,$6F,$62,$91,$98,$6B,$FE,$A4,$8C,$36,$BD
	.dc.b	$3E,$EF,$AB,$DA,$97,$D3,$06,$CA,$FE,$BD,$F5,$0F,$87,$A4,$BE,$11
	.dc.b	$F0,$F2,$FF,$59,$F9,$EF,$88,$FE,$CC,$DA,$EC,$D7,$CF,$BF,$C1,$8E
	.dc.b	$D1,$E3,$F1,$BF,$8D,$4F,$C5,$7F,$90,$37,$3F,$EE,$51,$A9,$F8,$9E
	.dc.b	$5F,$A5,$F6,$35,$EE,$6B,$4F,$EF,$51,$FE,$AF,$8F,$85,$C9,$FE,$4C
	.dc.b	$CF,$FD,$9A,$DC,$7E,$B3,$67,$E9,$2F,$2F,$C7,$E9,$79,$3D,$39,$5E
	.dc.b	$B8,$FF,$04,$76,$7F,$12,$8F,$1F,$12,$31,$97,$FB,$71,$B9,$7E,$17
	.dc.b	$EA,$3C,$14,$F9,$8F,$46,$1C,$9B,$87,$FB,$91,$B7,$F8,$DF,$43,$9D
	.dc.b	$7E,$27,$E9,$7F,$B5,$3E,$28,$CE,$06,$CE,$B0,$FF,$77,$AF,$A7,$FF
	.dc.b	$11,$36,$FF,$14,$D7,$DB,$F4,$EF,$67,$FC,$51,$87,$F8,$EA,$7D,$1F
	.dc.b	$01,$FE,$2B,$D3,$FD,$75,$1E,$3F,$3B,$AF,$D3,$F0,$64,$76,$8F,$0E
	.dc.b	$DA,$8C,$1E,$27,$FC,$6D,$F4,$6F,$A7,$B7,$2B,$F4,$13,$96,$CE,$5F
	.dc.b	$AB,$1E,$7D,$D7,$FB,$78,$E1,$3A,$AE,$15,$A8,$FD,$B4,$F1,$AE,$15
	.dc.b	$E4,$FA,$7F,$2F,$CB,$CB,$F1,$C6,$5F,$2C,$65,$F1,$C7,$F9,$47,$E9
	.dc.b	$FF,$C3,$D1,$F8,$A6,$E3,$E1,$E3,$46,$31,$FE,$1C,$BE,$67,$97,$BF
	.dc.b	$30,$FE,$05,$3B,$7F,$87,$3B,$7F,$88,$8D,$3E,$E6,$3F,$EA,$BF,$A6
	.dc.b	$FD,$E7,$85,$FA,$F6,$F7,$FF,$A9,$DF,$E0,$15,$C2,$D1,$8E,$5F,$E3
	.dc.b	$31,$8B,$FF,$57,$1C,$CB,$7C,$DF,$F5,$95,$F3,$BE,$58,$78,$E7,$77
	.dc.b	$4F,$F6,$03,$CD,$FF,$61,$57,$18,$E7,$87,$8A,$FD,$35,$C4,$FE,$39
	.dc.b	$87,$85,$E3,$3B,$AE,$38,$6D,$ED,$3C,$0E,$58,$77,$78,$CE,$EF,$A4
	.dc.b	$6C,$F8,$46,$8A,$7D,$1B,$F0,$3E,$C8,$E3,$7E,$0B,$ED,$3B,$BF,$F7
	.dc.b	$A4,$7D,$18,$FF,$6D,$19,$29,$4C,$AE,$F7,$D3,$F2,$6C,$F0,$9D,$4E
	.dc.b	$4E,$71,$FE,$D6,$28,$F8,$8A,$D3,$2E,$0F,$A1,$7F,$DA,$77,$FA,$53
	.dc.b	$FE,$F9,$9C,$F9,$7F,$AF,$D7,$FA,$64,$E4,$F8,$F7,$C5,$F9,$26,$63
	.dc.b	$FA,$94,$65,$E3,$1E,$37,$33,$D7,$3F,$8B,$7E,$0C,$3C,$77,$B9,$FF
	.dc.b	$66,$56,$E3,$E6,$D6,$DF,$09,$B3,$FC,$23,$1B,$7E,$35,$F8,$2F,$AD
	.dc.b	$71,$8F,$2B,$6F,$C7,$5E,$EB,$E4,$FB,$63,$C7,$7F,$A1,$E1,$7D,$3B
	.dc.b	$EA,$7F,$8B,$67,$FC,$1E,$31,$5F,$E0,$37,$FC,$FA,$D9,$DB,$FE,$17
	.dc.b	$8D,$E1,$FD,$6B,$EA,$9F,$1B,$A8,$FF,$0B,$4E,$B0,$FE,$29,$5A,$BF
	.dc.b	$F0,$86,$7F,$E0,$D6,$F9,$7C,$73,$FF,$14,$7A,$16,$B5,$97,$F6,$86
	.dc.b	$E3,$FC,$7D,$B8,$7F,$8A,$DF,$F3,$1B,$D0,$FF,$0F,$7E,$40,$F9,$1E
	.dc.b	$CD,$8C,$FA,$B8,$99,$1C,$1D,$C6,$30,$FE,$F3,$5C,$5B,$3A,$AE,$07
	.dc.b	$F8,$B7,$D1,$7C,$C7,$B5,$B1,$9F,$98,$C4,$FC,$1F,$97,$FF,$21,$4F
	.dc.b	$F8,$D5,$B8,$7F,$82,$2F,$8B,$FE,$0A,$8C,$E9,$FE,$01,$F9,$8F,$95
	.dc.b	$E2,$9E,$97,$87,$E7,$7D,$43,$E9,$69,$F4,$7F,$4D,$FE,$BF,$9F,$F9
	.dc.b	$16,$74,$FF,$91,$FF,$08,$FF,$1F,$37,$AB,$EC,$47,$D2,$D7,$FC,$AC
	.dc.b	$E9,$FF,$0F,$F8,$37,$5F,$EA,$CD,$BF,$F3,$13,$C7,$9B,$FC,$11,$87
	.dc.b	$E0,$FE,$45,$F7,$97,$E6,$70,$DC,$3F,$CB,$5F,$25,$E8,$CF,$FC,$5A
	.dc.b	$DC,$FF,$B2,$2F,$37,$F1,$CE,$6B,$FC,$E5,$F5,$0F,$9F,$FC,$92,$F8
	.dc.b	$BB,$18,$7F,$2D,$F5,$1F,$81,$C7,$FB,$82,$ED,$CF,$66,$9F,$CE,$7E
	.dc.b	$43,$FE,$23,$CB,$FC,$E0,$DD,$7F,$C8,$6F,$F2,$F8,$CF,$1F,$E8,$FE
	.dc.b	$33,$EE,$C7,$FA,$1D,$E1,$97,$F7,$F7,$37,$FE,$85,$73,$3F,$E8,$DF
	.dc.b	$A8,$FF,$3E,$BF,$E7,$F5,$F3,$1F,$AF,$73,$5F,$E5,$47,$33,$FE,$85
	.dc.b	$FA,$7F,$F6,$DB,$7E,$47,$F6,$D1,$FE,$2D,$8C,$1F,$E8,$BF,$4D,$F4
	.dc.b	$FF,$BB,$78,$E1,$FE,$60,$73,$3B,$BF,$E0,$1F,$51,$CE,$BF,$CC,$AB
	.dc.b	$D9,$CF,$B3,$A7,$F8,$37,$C6,$FF,$0D,$C3,$FD,$54,$DB,$FF,$53,$3F
	.dc.b	$C6,$9F,$F5,$E7,$A1,$FC,$37,$E4,$5F,$3B,$7E,$7F,$0D,$C3,$FD,$5F
	.dc.b	$C6,$FA,$6F,$C4,$46,$1B,$8F,$FA,$32,$75,$DF,$FC,$85,$E1,$DF,$6B
	.dc.b	$71,$FC,$C1,$B7,$FE,$D2,$7F,$3A,$F5,$9E,$04,$7C,$15,$6F,$FD,$13
	.dc.b	$F3,$5F,$33,$0F,$F2,$03,$F2,$5B,$C3,$FC,$AE,$31,$7F,$EC,$F7,$83
	.dc.b	$78,$35,$3B,$78,$76,$C9,$FF,$63,$57,$D8,$C3,$C3,$39,$6C,$EB,$93
	.dc.b	$18,$FF,$98,$E5,$F3,$CF,$80,$FE,$8D,$18,$7F,$A4,$7C,$FF,$FB,$3F
	.dc.b	$AF,$35,$BC,$7F,$CA,$2D,$BF,$F5,$DD,$7F,$B5,$7D,$8D,$E1,$CB,$D6
	.dc.b	$FE,$9D,$5F,$EC,$03,$B6,$1F,$EE,$AB,$E0,$F4,$AE,$17,$F1,$DF,$91
	.dc.b	$E1,$D9,$F8,$2D,$BF,$A2,$DC,$FE,$8B,$F3,$FA,$E5,$7C,$8F,$81,$B7
	.dc.b	$F6,$26,$DF,$D8,$9B,$4A,$FF,$7D,$FD,$37,$ED,$2B,$FB,$BB,$67,$50
	.dc.b	$8F,$DB,$CF,$E4,$DD,$3C,$BE,$3F,$F2,$91,$89,$E1,$E0,$E2,$65,$FE
	.dc.b	$82,$B4,$74,$AF,$0C,$FE,$2F,$CA,$95,$F8,$4F,$83,$B8,$FE,$52,$6D
	.dc.b	$FE,$52,$6D,$7B,$77,$70,$3E,$CB,$EC,$7D,$5F,$D9,$6D,$7D,$79,$D3
	.dc.b	$67,$4D,$3E,$E7,$E8,$DF,$AC,$CD,$82,$30,$DE,$BF,$C7,$C6,$09,$D9
	.dc.b	$88,$CB,$2F,$AB,$96,$9F,$A4,$FC,$6B,$F5,$DF,$19,$FA,$F6,$C6,$4F
	.dc.b	$D4,$C6,$5A,$7E,$B3,$E9,$FF,$6D,$F1,$3F,$B7,$6C,$64,$FD,$8C,$65
	.dc.b	$A7,$ED,$30,$FD,$F1,$FB,$FF,$4A,$C6,$49,$A7,$F0,$88,$CB,$FC,$2A
	.dc.b	$3F,$83,$65,$FC,$1D,$A7,$F0,$02,$C6,$DF,$E0,$71,$FE,$37,$E5,$6D
	.dc.b	$4C,$7B,$16,$BF,$2E,$7E,$77,$D5,$FE,$A3,$39,$C1,$C9,$FE,$D9,$C3
	.dc.b	$FC,$1F,$18,$BE,$11,$D9,$E7,$0B,$FE,$B7,$F3,$5E,$33,$C9,$C9,$6B
	.dc.b	$73,$A6,$DA,$B7,$18,$7D,$9F,$61,$ED,$F6,$EC,$F0,$70,$D3,$9C,$63
	.dc.b	$1F,$E2,$1F,$32,$F0,$95,$FE,$99,$F9,$3F,$DF,$52,$BF,$8B,$FC,$42
	.dc.b	$73,$1C,$6B,$85,$7F,$05,$8D,$CE,$AF,$AC,$FF,$AB,$63,$8F,$87,$8C
	.dc.b	$3A,$4F,$1A,$C5,$6B,$0E,$B2,$FA,$87,$23,$E3,$F3,$18,$77,$9F,$8A
	.dc.b	$BF,$8C,$4E,$5F,$11,$8A,$E4,$E0,$EC,$E3,$2F,$2C,$7C,$3C,$FC,$AF
	.dc.b	$1C,$3C,$35,$BC,$BE,$4E,$1D,$F0,$E3,$8F,$8F,$0E,$B8,$70,$23,$E1
	.dc.b	$8D,$57,$1B,$E7,$87,$5A,$F0,$1B,$C7,$C6,$F4,$3D,$55,$C5,$DE,$3E
	.dc.b	$29,$DC,$E6,$78,$4F,$BA,$31,$87,$8A,$FC,$15,$AC,$3C,$78,$FC,$EC
	.dc.b	$79,$D6,$B0,$E8,$72,$8C,$CF,$BB,$1E,$D5,$D3,$1E,$D3,$9C,$3A,$E3
	.dc.b	$ED,$CB,$0F,$AA,$33,$87,$AE,$73,$3F,$04,$FC,$C9,$4E,$66,$32,$FC
	.dc.b	$13,$7A,$1F,$4B,$C2,$3F,$0F,$18,$71,$28,$C4,$71,$9E,$0F,$E5,$63
	.dc.b	$85,$FC,$B6,$E1,$F8,$46,$CF,$E5,$3D,$3F,$D7,$D6,$EF,$F3,$75,$AA
	.dc.b	$E9,$1F,$9E,$AD,$DF,$38,$D9,$27,$F0,$33,$A7,$D7,$19,$BE,$D1,$A3
	.dc.b	$15,$CD,$FC,$4D,$F8,$70,$FF,$2F,$46,$8E,$66,$AF,$C8,$E4,$F3,$99
	.dc.b	$3B,$16,$FC,$24,$BF,$D3,$E1,$E7,$9E,$33,$97,$BB,$F9,$58,$D5,$FD
	.dc.b	$3B,$F0,$36,$BB,$D7,$17,$9D,$FB,$63,$94,$70,$C3,$DC,$7B,$F0,$EB
	.dc.b	$F4,$9D,$99,$3F,$C3,$A3,$2F,$C0,$3B,$DF,$F0,$08,$FA,$1C,$9F,$E9
	.dc.b	$AF,$AE,$FE,$57,$CF,$FD,$25,$BA,$FF,$48,$6E,$1F,$D1,$F9,$9F,$4F
	.dc.b	$8C,$3F,$D4,$6D,$85,$53,$11,$9F,$56,$5F,$F4,$A7,$87,$D4,$62,$3F
	.dc.b	$AC,$1D,$E3,$19,$7F,$5D,$9E,$DC,$D2,$BF,$A9,$BE,$1E,$6C,$47,$2C
	.dc.b	$FF,$D6,$F3,$8B,$FE,$CD,$5E,$7F,$8D,$F4,$7C,$E4,$C3,$FA,$EC,$65
	.dc.b	$FE,$BF,$19,$7A,$6A,$9F,$81,$F0,$92,$7F,$B2,$BF,$0F,$C7,$C4,$70
	.dc.b	$8B,$8F,$F7,$79,$F9,$1F,$17,$F2,$7E,$32,$5F,$F6,$E8,$CB,$FD,$C2
	.dc.b	$32,$F8,$A3,$E8,$7C,$AF,$D1,$EC,$93,$FD,$DA,$32,$FF,$78,$8C,$BE
	.dc.b	$A8,$FB,$AB,$E1,$DF,$1F,$86,$C7,$F8,$20,$D4,$67,$E3,$34,$C8,$D4
	.dc.b	$F0,$7F,$D1,$4A,$75,$18,$9B,$9F,$FA,$7E,$B5,$3F,$E8,$FF,$CF,$7F
	.dc.b	$DC,$6D,$F7,$3F,$27,$E7,$A5,$18,$77,$3F,$E9,$E6,$DF,$F7,$EA,$E1
	.dc.b	$5B,$38,$7B,$3F,$99,$B8,$7F,$87,$8F,$CE,$78,$78,$7F,$02,$5B,$FF
	.dc.b	$18,$47,$CA,$F9,$E7,$05,$25,$7C,$98,$71,$62,$3F,$D5,$D1,$C1,$D4
	.dc.b	$F3,$9D,$5F,$FB,$0E,$B5,$5C,$2B,$4F,$F8,$DB,$E0,$5F,$47,$FB,$56
	.dc.b	$C6,$63,$1F,$59,$96,$46,$4E,$51,$C7,$E2,$F9,$7C,$5C,$9E,$D1,$97
	.dc.b	$C1,$19,$7F,$61,$69,$FC,$6F,$D4,$3C,$DF,$C7,$FD,$4F,$F5,$7B,$FC
	.dc.b	$63,$5F,$F6,$84,$ED,$FC,$83,$5F,$F6,$A4,$6F,$5F,$F7,$33,$7C,$5F
	.dc.b	$F2,$2D,$6E,$7F,$DD,$4D,$8C,$DF,$90,$F1,$FA,$9F,$1F,$EA,$26,$3F
	.dc.b	$2D,$F9,$5F,$51,$88,$F6,$C5,$C7,$FC,$93,$E2,$35,$5F,$F2,$5F,$A9
	.dc.b	$FB,$F8,$7F,$9A,$9F,$F5,$CB,$7C,$0D,$72,$C8,$C1,$0F,$E3,$53,$C9
	.dc.b	$FF,$7A,$7D,$1B,$E1,$7F,$72,$D8,$FD,$B3,$5D,$F9,$BA,$99,$1A,$78
	.dc.b	$C6,$23,$AC,$5D,$3F,$CC,$B0,$AF,$F2,$7C,$7C,$0F,$8C,$FE,$7F,$3F
	.dc.b	$E8,$65,$EA,$FF,$45,$6C,$67,$CE,$FE,$91,$E5,$72,$E4,$FF,$4E,$36
	.dc.b	$7F,$D2,$8D,$E1,$2F,$A4,$FE,$A1,$F5,$39,$1B,$E0,$CE,$FF,$D5,$7E
	.dc.b	$A7,$91,$AC,$7F,$D3,$6D,$7F,$A3,$3D,$FE,$13,$FC,$6B,$BF,$DC,$6E
	.dc.b	$9F,$EB,$16,$D7,$FA,$A5,$FF,$3A,$BF,$D3,$FE,$A7,$F6,$B5,$F8,$47
	.dc.b	$EC,$37,$8F,$FB,$44,$6A,$FF,$D7,$9F,$52,$F3,$7F,$B6,$39,$C3,$FD
	.dc.b	$6C,$DE,$03,$B8,$FF,$5A,$FE,$43,$F9,$D7,$FD,$78,$D9,$34,$FF,$5C
	.dc.b	$47,$FA,$F3,$F4,$E7,$85,$FD,$EF,$E9,$BD,$7A,$FE,$71,$B9,$FF,$B4
	.dc.b	$6B,$77,$FE,$C4,$F3,$3F,$C6,$2F,$F6,$F5,$F5,$9F,$E0,$56,$C6,$FD
	.dc.b	$AF,$F0,$47,$92,$FA,$CA,$49,$FF,$67,$C7,$E6,$73,$FD,$6F,$6F,$66
	.dc.b	$FB,$7F,$E1,$18,$37,$0F,$F0,$A7,$DA,$FF,$40,$D7,$FD,$D0,$DC,$7F
	.dc.b	$DC,$0D,$E1,$2F,$92,$FF,$0E,$7A,$EE,$1A,$4B,$FF,$71,$C3,$A3,$FC
	.dc.b	$45,$F5,$7F,$D7,$F6,$FE,$06,$DC,$FF,$DF,$0D,$BF,$F7,$83,$FD,$79
	.dc.b	$FF,$0B,$7D,$1F,$E3,$34,$34,$C7,$D5,$BE,$5F,$FC,$60,$DA,$FF,$7A
	.dc.b	$B5,$FF,$10,$37,$1F,$E2,$BF,$1E,$F8,$AF,$F1,$C3,$6B,$FD,$FA,$D7
	.dc.b	$FC,$50,$D9,$E3,$F1,$7F,$E3,$FF,$8F,$FE,$5D,$4C,$D1,$86,$FC,$36
	.dc.b	$FE,$3B,$83,$28,$7F,$8D,$6B,$80,$9D,$46,$66,$E1,$FC,$BA,$B5,$3C
	.dc.b	$A3,$F9,$8F,$D5,$BE,$47,$FC,$A8,$D9,$D7,$0F,$83,$24,$E6,$7F,$C8
	.dc.b	$CB,$E5,$78,$DC,$9E,$27,$BE,$78,$61,$E0,$AF,$F3,$16,$6F,$92,$FF
	.dc.b	$32,$78,$CF,$9B,$A6,$A2,$63,$D3,$3F,$81,$18,$8F,$F2,$C5,$BE,$53
	.dc.b	$FC,$6B,$9B,$E7,$DA,$7C,$AF,$AA,$FE,$89,$A3,$E2,$3E,$6E,$49,$CC
	.dc.b	$61,$FE,$95,$87,$3B,$E5,$7C,$23,$17,$C6,$39,$9D,$5D,$9F,$69,$C3
	.dc.b	$B3,$EE,$C7,$2A,$EF,$87,$A2,$74,$7F,$88,$B0,$F2,$5F,$66,$C7,$29
	.dc.b	$E1,$5C,$B0,$F0,$CF,$12,$DF,$8A,$75,$1D,$B0,$F1,$4F,$58,$DE,$1D
	.dc.b	$9F,$C1,$E1,$DA,$7E,$13,$63,$DD,$3B,$9E,$C6,$B0,$F1,$DF,$6A,$D9
	.dc.b	$D8,$CD,$F3,$7B,$CF,$FB,$D2,$7E,$7C,$7D,$83,$27,$FB,$FD,$CB,$F5
	.dc.b	$7B,$1F,$57,$C5,$B1,$B6,$BC,$E7,$F8,$D4,$F5,$8E,$DF,$23,$5F,$F7
	.dc.b	$E9,$43,$5A,$65,$FA,$37,$FE,$86,$E3,$FA,$73,$FB,$C9,$CF,$C4,$A7
	.dc.b	$33,$CB,$5F,$EC,$73,$98,$FC,$19,$C4,$FC,$2D,$F5,$72,$71,$D3,$FD
	.dc.b	$AF,$3C,$6B,$FC,$49,$3C,$5E,$B7,$E0,$C3,$C6,$D9,$EB,$5C,$6F,$E2
	.dc.b	$56,$A7,$BB,$6F,$AC,$78,$4F,$E2,$15,$C5,$F9,$67,$4B,$ED,$3E,$28
	.dc.b	$E6,$64,$FD,$37,$0B,$EA,$DF,$23,$F0,$F3,$FF,$25,$C6,$30,$FE,$0C
	.dc.b	$BE,$47,$51,$FC,$6E,$35,$87,$FB,$7D,$B5,$CA,$F7,$3F,$CD,$E7,$95
	.dc.b	$74,$9C,$E1,$FE,$C0,$F7,$2E,$66,$F8,$78,$65,$FF,$19,$9D,$36,$73
	.dc.b	$8F,$FB,$C7,$EA,$BF,$D0,$53,$FC,$7D,$B1,$9F,$86,$F0,$C6,$34,$FF
	.dc.b	$35,$B6,$BF,$9C,$36,$BF,$A8,$2A,$E3,$F0,$D2,$BB,$3D,$23,$15,$FC
	.dc.b	$A2,$7F,$A1,$7C,$77,$F8,$BD,$FE,$64,$D8,$D1,$FD,$75,$FF,$5D,$C6
	.dc.b	$BD,$5F,$8F,$F5,$5F,$ED,$B8,$FE,$74,$DA,$E3,$B7,$FB,$7E,$B7,$7F
	.dc.b	$3B,$C7,$FF,$51,$61,$FD,$05,$B7,$C3,$0E,$5A,$7F,$A3,$27,$2F,$FA
	.dc.b	$3B,$EA,$5E,$27,$D8,$6E,$1F,$D2,$FE,$AB,$F6,$0F,$F4,$D8,$CE,$3E
	.dc.b	$4F,$AA,$FF,$68,$CF,$F5,$16,$C6,$78,$FF,$07,$18,$C7,$FD,$3B,$DB
	.dc.b	$F7,$AB,$DA,$E6,$FF,$B6,$39,$9D,$F5,$7E,$33,$BD,$F1,$17,$E6,$B0
	.dc.b	$DC,$3F,$C9,$47,$E4,$BD,$6F,$4E,$5F,$E4,$BF,$9D,$C4,$FE,$5B,$EA
	.dc.b	$9F,$43,$4C,$7C,$9D,$9F,$ED,$9F,$55,$F8,$7E,$93,$4F,$F7,$2C,$6B
	.dc.b	$AF,$F4,$3F,$55,$FD,$B6,$3F,$BA,$37,$3F,$B6,$FF,$90,$E3,$3C,$5F
	.dc.b	$A8,$F9,$CF,$AF,$1F,$E0,$D7,$87,$1F,$FB,$11,$CD,$FF,$89,$DC,$CF
	.dc.b	$F8,$4F,$EA,$BF,$C9,$4F,$F8,$0D,$7C,$E7,$ED,$1C,$D7,$E3,$1C,$CF
	.dc.b	$F8,$27,$EA,$9F,$35,$FB,$A6,$C6,$7D,$8C,$4F,$EF,$BD,$27,$0D,$31
	.dc.b	$36,$BF,$C2,$9F,$55,$FE,$4F,$6C,$7F,$06,$8F,$F3,$7C,$60,$FF,$60
	.dc.b	$79,$1F,$C2,$BE,$43,$F8,$8C,$7F,$8B,$DF,$E1,$FA,$7E,$91,$CB,$FE
	.dc.b	$49,$F9,$47,$D5,$7F,$12,$7F,$C5,$15,$FE,$25,$72,$DF,$5B,$F8,$DF
	.dc.b	$13,$EA,$9D,$78,$6D,$7F,$8E,$27,$5D,$B9,$A9,$39,$AD,$E5,$FC,$02
	.dc.b	$F7,$39,$AD,$CE,$6B,$75,$FC,$DD,$FC,$ED,$6C,$FF,$1E,$7D,$57,$D2
	.dc.b	$AF,$E0,$6F,$CB,$6F,$17,$F3,$08,$C5,$FF,$63,$EB,$47,$18,$C3,$AB
	.dc.b	$FF,$2D,$39,$6F,$0F,$3F,$81,$91,$E2,$E3,$B9,$78,$A6,$E1,$D2,$6D
	.dc.b	$7F,$92,$3E,$59,$F5,$5F,$D2,$30,$46,$27,$FC,$C7,$F4,$1F,$D1,$7A
	.dc.b	$F5,$39,$AF,$E5,$27,$0B,$C9,$FD,$4A,$B8,$5F,$F9,$B2,$74,$DA,$E5
	.dc.b	$8F,$F8,$16,$BF,$A9,$7C,$B7,$2A,$F0,$F9,$08,$F9,$58,$FC,$BF,$0A
	.dc.b	$D7,$D2,$6E,$5F,$49,$B8,$FC,$C7,$E9,$75,$CA,$FB,$3F,$55,$6F,$EC
	.dc.b	$CD,$BF,$A7,$36,$BF,$D2,$5F,$55,$FB,$69,$FE,$F4,$D8,$CF,$81,$F9
	.dc.b	$88,$C6,$1F,$E7,$A6,$D1,$A7,$EF,$A3,$F2,$8E,$3E,$1F,$91,$CE,$AD
	.dc.b	$FE,$43,$C9,$FC,$61,$F8,$DF,$81,$B1,$92,$61,$F9,$48,$CB,$FD,$83
	.dc.b	$83,$FC,$01,$F5,$5F,$5F,$87,$FD,$91,$7F,$42,$31,$F0,$D9,$6C,$64
	.dc.b	$FA,$B1,$F7,$FE,$1B,$F4,$F1,$FE,$C0,$B4,$63,$04,$63,$BB,$87,$74
	.dc.b	$BF,$C7,$DE,$EB,$51,$89,$D5,$FA,$6B,$47,$ED,$1C,$47,$EB,$FE,$17
	.dc.b	$F6,$1F,$0B,$6F,$F6,$ED,$AF,$DB,$B7,$8B,$F6,$2F,$ED,$FE,$1B,$F7
	.dc.b	$1F,$0D,$6F,$F7,$ED,$BF,$DF,$B6,$FF,$72,$FE,$FF,$49,$5F,$C1,$BD
	.dc.b	$1B,$7F,$C1,$A6,$DF,$F0,$09,$B5,$FC,$00,$FE,$11,$9F,$F0,$96,$9F
	.dc.b	$C4,$0B,$7F,$C4,$1B,$5F,$C3,$78,$FE,$DC,$FF,$85,$9B,$19,$F5,$BF
	.dc.b	$C8,$71,$8A,$FF,$74,$36,$BF,$E0,$03,$7C,$2F,$EA,$F5,$C3,$87,$8F
	.dc.b	$D5,$4C,$3F,$DB,$D6,$E3,$C6,$3E,$1A,$B9,$79,$66,$DF,$83,$E7,$3C
	.dc.b	$7F,$3F,$E4,$6C,$74,$6E,$1E,$B6,$DF,$FA,$36,$1F,$4C,$F0,$BF,$D2
	.dc.b	$1F,$0C,$F9,$AF,$F7,$3C,$FF,$0B,$C7,$FD,$35,$18,$AD,$BA,$F0,$EE
	.dc.b	$03,$AC,$E1,$EF,$16,$FC,$17,$D2,$78,$47,$77,$FD,$5F,$7F,$EF,$8F
	.dc.b	$A9,$72,$3E,$9B,$85,$27,$31,$8B,$E5,$9F,$AE,$BC,$95,$A7,$A9,$8B
	.dc.b	$E8,$E0,$E8,$E2,$79,$47,$AB,$1F,$0D,$73,$9C,$B7,$1F,$45,$F1,$7E
	.dc.b	$35,$70,$8C,$5F,$87,$1C,$1E,$9C,$BE,$46,$1C,$E7,$47,$95,$C3,$E3
	.dc.b	$31,$5D,$1F,$4D,$74,$7D,$99,$7C,$4C,$7C,$B3,$9C,$BD,$37,$C6,$B5
	.dc.b	$7E,$09,$CD,$6E,$FD,$73,$E0,$8C,$3B,$8C,$CF,$08,$D1,$1F,$97,$5D
	.dc.b	$A3,$CE,$66,$7C,$AF,$A3,$1F,$87,$87,$B7,$1F,$82,$7C,$2B,$E4,$E2
	.dc.b	$3E,$DD,$7E,$4C,$F9,$25,$3B,$96,$3F,$08,$DE,$87,$D5,$6E,$3F,$13
	.dc.b	$18,$71,$28,$C4,$71,$9E,$0F,$E5,$63,$86,$5F,$86,$6E,$5F,$9D,$6C
	.dc.b	$FE,$5B,$E4,$BD,$15,$BB,$FC,$7D,$6A,$BA,$47,$E4,$AB,$77,$CE,$36
	.dc.b	$49,$FC,$44,$E9,$F5,$C6,$6F,$B4,$68,$C5,$73,$7F,$21,$7E,$1C,$18
	.dc.b	$73,$9F,$19,$AC,$38,$BE,$B2,$9A,$C3,$B9,$FA,$89,$E5,$5F,$2E,$B6
	.dc.b	$DA,$F2,$1D,$49,$5C,$B0,$FE,$0D,$7C,$B0,$F3,$13,$0F,$8D,$7D,$23
	.dc.b	$0F,$4B,$E4,$FC,$43,$AD,$F4,$7F,$7F,$F2,$7E,$6C,$24,$60,$87,$39
	.dc.b	$E0,$FE,$DE,$7E,$87,$47,$FA,$6B,$EA,$BF,$29,$C5,$FD,$47,$E8,$AF
	.dc.b	$17,$F4,$C6,$E1,$FD,$32,$AD,$F1,$8F,$E9,$33,$8E,$83,$6B,$1E,$B4
	.dc.b	$AF,$E9,$EE,$FE,$37,$11,$FD,$7E,$2E,$5F,$D8,$E7,$A7,$34,$AF,$EA
	.dc.b	$EF,$7E,$5C,$47,$F6,$78,$B5,$FD,$AE,$7C,$9F,$17,$E5,$FA,$A9,$7F
	.dc.b	$D8,$23,$2F,$F6,$28,$CB,$FD,$C6,$3D,$9F,$07,$23,$FB,$3B,$EF,$F1
	.dc.b	$F1,$1F,$DD,$E2,$E1,$FD,$EE,$7E,$27,$C5,$7C,$5F,$A9,$97,$FD,$C2
	.dc.b	$32,$FF,$72,$8C,$BF,$E0,$18,$F9,$9F,$29,$F3,$7B,$24,$FF,$78,$8C
	.dc.b	$BF,$DE,$A3,$2F,$F8,$26,$3E,$CA,$F8,$67,$C7,$E1,$B1,$FE,$7F,$6B
	.dc.b	$FE,$86,$5F,$1B,$B9,$93,$B7,$94,$62,$38,$45,$CF,$FB,$F5,$72,$9F
	.dc.b	$F4,$AF,$D5,$7F,$B7,$9B,$EE,$7E,$3B,$CF,$4F,$FB,$A2,$77,$EC,$E6
	.dc.b	$38,$47,$F9,$8D,$D3,$6B,$53,$F9,$38,$87,$F8,$78,$FC,$B7,$89,$87
	.dc.b	$FC,$5C,$5C,$7F,$C6,$51,$F2,$3E,$69,$C1,$49,$5F,$EA,$58,$71,$62
	.dc.b	$3F,$D5,$B1,$C1,$D4,$F3,$9D,$5F,$F8,$86,$B5,$5C,$2B,$4F,$FB,$B7
	.dc.b	$EA,$BF,$5F,$F0,$26,$98,$F9,$FB,$3B,$20,$C3,$FD,$B1,$49,$E7,$F1
	.dc.b	$5D,$3E,$4A,$5F,$FB,$EE,$32,$FF,$BF,$63,$2F,$3F,$0B,$FC,$59,$E1
	.dc.b	$30,$5F,$E3,$BE,$A5,$E6,$FE,$43,$EA,$BF,$AA,$63,$FE,$CB,$6B,$FE
	.dc.b	$D2,$9D,$BF,$EC,$F6,$B9,$9E,$87,$3F,$52,$C6,$E2,$BF,$ED,$CA,$4F
	.dc.b	$C8,$FA,$A9,$3F,$E4,$3D,$BE,$1C,$7C,$CF,$AA,$CC,$FB,$46,$9E,$D9
	.dc.b	$CD,$7C,$AA,$FE,$43,$F5,$1F,$80,$BE,$9B,$7C,$48,$78,$1A,$37,$CB
	.dc.b	$2B,$67,$69,$DB,$8C,$91,$88,$F2,$46,$27,$F7,$7F,$43,$FF,$00,$27
	.dc.b	$FD,$EC,$D7,$FD,$FA,$DC,$BF,$CD,$CF,$F9,$21,$BF,$0B,$C8,$93,$CB
	.dc.b	$CD,$AF,$28,$B3,$CA,$7F,$A0,$FC,$67,$F3,$EA,$F4,$2F,$57,$26,$BC
	.dc.b	$DF,$E8,$FE,$4F,$87,$7F,$C4,$B6,$FE,$7B,$75,$2F,$AA,$FE,$9F,$F5
	.dc.b	$59,$26,$78,$8F,$EA,$7F,$55,$C4,$91,$88,$FE,$AF,$F0,$9F,$E2,$FD
	.dc.b	$99,$FE,$6A,$30,$DC,$F7,$E0,$FF,$5E,$FA,$AF,$DA,$DF,$09,$FC,$2C
	.dc.b	$7E,$05,$FA,$AD,$8D,$6A,$8E,$37,$FA,$8F,$AA,$79,$BF,$DA,$9C,$E1
	.dc.b	$F9,$06,$D9,$DC,$7E,$17,$EA,$BF,$4B,$8F,$66,$E1,$FB,$46,$CF,$E6
	.dc.b	$97,$BD,$FD,$E3,$EA,$BF,$81,$4F,$E3,$DA,$FE,$39,$B7,$EA,$FA,$5F
	.dc.b	$8B,$3F,$A4,$F3,$7F,$C5,$EF,$E5,$DA,$FE,$59,$B3,$F0,$97,$B3,$FE
	.dc.b	$08,$F2,$9F,$53,$49,$7F,$A1,$8F,$CC,$69,$FC,$3B,$B3,$9B,$6F,$FC
	.dc.b	$23,$1D,$2B,$F5,$0E,$9F,$DC,$FD,$4B,$E4,$7F,$C2,$EF,$F4,$8C,$FE
	.dc.b	$AB,$70,$FE,$36,$EE,$BF,$65,$EB,$35,$FF,$0F,$B6,$33,$18,$E6,$FF
	.dc.b	$11,$FD,$4F,$F6,$8C,$7F,$02,$DC,$7F,$93,$B6,$FF,$86,$BF,$D8,$DF
	.dc.b	$DE,$FD,$17,$CE,$E5,$34,$C7,$D4,$BE,$57,$FC,$62,$DA,$FE,$10,$D7
	.dc.b	$F8,$3B,$71,$FC,$9F,$C7,$3E,$27,$FC,$72,$DA,$FE,$20,$D7,$F8,$7B
	.dc.b	$67,$8F,$C5,$7F,$90,$3E,$3B,$F9,$CC,$7E,$79,$BA,$FF,$44,$6D,$FF
	.dc.b	$2A,$5F,$1B,$C7,$E2,$24,$FF,$1F,$9D,$EC,$77,$19,$9D,$BF,$C8,$A3
	.dc.b	$66,$9F,$E4,$BF,$52,$F9,$0F,$F2,$AB,$63,$F9,$63,$5D,$FC,$A6,$A6
	.dc.b	$46,$9E,$31,$FE,$5E,$C5,$F2,$34,$95,$FC,$DE,$1F,$25,$88,$FE,$6F
	.dc.b	$1C,$1D,$4F,$39,$FF,$19,$E0,$F9,$EF,$EE,$D4,$FA,$8F,$E1,$9E,$8B
	.dc.b	$E1,$BF,$D9,$11,$86,$C7,$FB,$2E,$F9,$70,$AF,$94,$E7,$0E,$B8,$71
	.dc.b	$9D,$35,$D9,$C6,$F9,$B8,$0E,$23,$81,$2B,$B5,$F4,$6D,$71,$7D,$4E
	.dc.b	$4F,$33,$87,$B5,$75,$8E,$35,$D6,$77,$87,$C0,$7C,$D8,$7B,$2B,$67
	.dc.b	$FB,$B4,$E9,$5C,$67,$D3,$3B,$78,$E1,$E3,$7F,$C7,$37,$D5,$A6,$70
	.dc.b	$F2,$3C,$0F,$F7,$93,$F1,$9F,$F7,$D3,$5F,$82,$63,$A9,$F5,$3F,$EF
	.dc.b	$29,$FE,$18,$D8,$CC,$F4,$6B,$FE,$F8,$F9,$0F,$F7,$CB,$F7,$5A,$9A
	.dc.b	$47,$EC,$61,$FE,$7E,$E4,$F7,$47,$13,$3F,$0F,$FD,$6F,$4F,$EB,$51
	.dc.b	$97,$FA,$D4,$67,$5F,$C6,$4C,$C3,$E4,$61,$D6,$3F,$0C,$71,$9F,$F6
	.dc.b	$5D,$71,$8E,$27,$9C,$F0,$E1,$E4,$71,$87,$82,$31,$87,$6A,$E5,$5E
	.dc.b	$13,$89,$E0,$38,$DF,$82,$B8,$C6,$AF,$DA,$E8,$FF,$21,$57,$47,$D0
	.dc.b	$62,$78,$CF,$8E,$3F,$45,$F8,$47,$F0,$29,$FF,$1C,$C2,$30,$FF,$92
	.dc.b	$71,$7E,$75,$FE,$23,$AD,$D6,$A7,$FC,$15,$5C,$78,$7F,$C1,$91,$3F
	.dc.b	$1D,$7C,$C6,$30,$FA,$38,$71,$7F,$CB,$53,$A7,$AF,$1E,$F1,$FF,$40
	.dc.b	$65,$FE,$72,$9C,$9F,$E5,$0C,$7C,$BF,$01,$BA,$F5,$5F,$19,$CD,$70
	.dc.b	$9F,$65,$F3,$7F,$CC,$C3,$E2,$DF,$1F,$CD,$B5,$DE,$BF,$C2,$2F,$C8
	.dc.b	$8C,$D7,$26,$DF,$C2,$A7,$CC,$73,$C3,$C9,$18,$C3,$94,$62,$75,$1D
	.dc.b	$DF,$67,$CB,$7F,$50,$8F,$AE,$DC,$7D,$7E,$0E,$BA,$A5,$7D,$6F,$9A
	.dc.b	$B5,$EE,$E0,$47,$C7,$6C,$FC,$BF,$C8,$7F,$B6,$3C,$4C,$61,$BF,$9D
	.dc.b	$F1,$F9,$D6,$7C,$6E,$1F,$C5,$F8,$6F,$2B,$19,$FC,$E8,$C4,$FE,$2B
	.dc.b	$E7,$7E,$C7,$E3,$BE,$0A,$F3,$46,$32,$F3,$79,$FF,$83,$E5,$7E,$38
	.dc.b	$F9,$CC,$E5,$FE,$92,$9C,$C7,$E0,$FE,$FD,$F2,$BF,$83,$FD,$3D,$F4
	.dc.b	$19,$8C,$3F,$0F,$E5,$FB,$7E,$39,$FE,$6C,$CB,$F3,$F1,$87,$E4,$57
	.dc.b	$E3,$7E,$3B,$FA,$B4,$7E,$92,$31,$C1,$F9,$38,$FF,$18,$1F,$97,$FC
	.dc.b	$E3,$FD,$76,$FE,$C0,$CC,$7E,$BE,$32,$FD,$7F,$57,$F3,$5F,$A8,$3E
	.dc.b	$73,$EC,$46,$1F,$B5,$BF,$C5,$FC,$75,$DF,$8C,$BE,$24,$61,$FD,$FF
	.dc.b	$98,$EC,$F9,$9F,$10,$E9,$FD,$54,$E6,$BE,$64,$E5,$FE,$03,$DB,$F4
	.dc.b	$FF,$1D,$F3,$F8,$DF,$C0,$BC,$1F,$F3,$2D,$7D,$39,$CB,$FC,$03,$ED
	.dc.b	$FF,$69,$F9,$C7,$F8,$96,$3F,$85,$C6,$3D,$8F,$DD,$7E,$6F,$FE,$C8
	.dc.b	$9F,$DF,$79,$9F,$1B,$2F,$E1,$F3,$9E,$6F,$E2,$D3,$9A,$FE,$7B,$F8
	.dc.b	$CF,$BE,$7F,$8D,$7E,$7A,$FA,$8F,$E0,$D3,$9C,$7F,$8B,$CE,$6F,$F9
	.dc.b	$1F,$EA,$0F,$A2,$FE,$18,$D8,$FE,$05,$F0,$18,$99,$1F,$C4,$FE,$BB
	.dc.b	$F1,$19,$FF,$0C,$BE,$0F,$F2,$B8,$CE,$3F,$CA,$27,$2F,$F2,$88,$CD
	.dc.b	$7F,$56,$9C,$BF,$EF,$F8,$CD,$29,$1F,$CA,$63,$B7,$AB,$FC,$86,$3E
	.dc.b	$8C,$6D,$D4,$F7,$9D,$1B,$F9,$EB,$7E,$BF,$17,$97,$C1,$73,$6C,$64
	.dc.b	$F8,$11,$96,$9F,$D3,$3E,$6F,$F8,$B6,$1F,$D1,$9B,$1E,$BF,$4F,$F9
	.dc.b	$C4,$63,$3F,$F2,$03,$6B,$DB,$C9,$AA,$F6,$36,$FE,$83,$6F,$E7,$3F
	.dc.b	$07,$D9,$93,$F5,$7D,$8B,$5F,$4E,$6E,$1F,$46,$6D,$7C,$FF,$83,$F9
	.dc.b	$5F,$07,$F2,$DB,$1F,$7D,$B8,$7D,$D6,$DF,$F6,$4F,$C3,$5F,$63,$FD
	.dc.b	$55,$B7,$F8,$7F,$67,$13,$23,$F1,$5E,$25,$C7,$F1,$0D,$C3,$F0,$CD
	.dc.b	$9F,$C2,$3F,$7F,$C4,$FC,$07,$A9,$67,$F2,$8D,$BF,$C9,$36,$FF,$20
	.dc.b	$FE,$1F,$CF,$95,$F9,$BF,$3A,$DF,$E6,$66,$DF,$E5,$E6,$D7,$F8,$B3
	.dc.b	$F1,$EF,$EB,$97,$FD,$B1,$B5,$F9,$DF,$37,$FB,$B4,$62,$B8,$7B,$7F
	.dc.b	$DD,$FF,$01,$7E,$07,$A8,$C3,$67,$FC,$01,$3A,$7F,$C3,$35,$CF,$1F
	.dc.b	$D2,$4E,$5F,$D8,$79,$FF,$A3,$3F,$49,$E5,$DB,$FD,$C4,$59,$FD,$B4
	.dc.b	$DC,$FF,$69,$E4,$7E,$AC,$FF,$06,$7C,$FF,$F0,$9C,$3F,$69,$19,$7F
	.dc.b	$81,$3A,$F4,$31,$87,$F8,$3B,$F3,$97,$77,$F8,$77,$E1,$DF,$22,$F2
	.dc.b	$3F,$C3,$55,$AC,$BF,$D4,$5A,$1F,$F5,$25,$F0,$AD,$E1,$CF,$0E,$55
	.dc.b	$FE,$3C,$C3,$96,$3F,$C1,$E3,$8C,$EA,$3F,$C4,$3E,$53,$E7,$3F,$92
	.dc.b	$9E,$EF,$94,$B7,$FE,$46,$9D,$3F,$DB,$FF,$1E,$FE,$DA,$E6,$BF,$BB
	.dc.b	$46,$2F,$E0,$7D,$2F,$BF,$93,$1B,$FF,$92,$34,$4E,$7A,$7E,$0B,$72
	.dc.b	$F5,$37,$0F,$95,$E6,$FF,$8F,$FE,$BB,$FB,$8C,$62,$BF,$CB,$71,$8C
	.dc.b	$7B,$56,$EB,$E6,$B6,$BD,$9E,$CF,$F4,$43,$13,$F1,$1C,$47,$F8,$37
	.dc.b	$F1,$EF,$EA,$17,$FE,$60,$DF,$FB,$A4,$E7,$E9,$BC,$9F,$13,$F5,$F1
	.dc.b	$FF,$4E,$70,$FF,$A4,$34,$7D,$26,$7D,$CF,$A1,$5F,$ED,$F9,$FF,$25
	.dc.b	$C6,$21,$1C,$31,$FF,$54,$5F,$3C,$39,$3D,$70,$B7,$CE,$6B,$CA,$2F
	.dc.b	$DA,$7F,$88,$BE,$8F,$9C,$FF,$BB,$E3,$0F,$A0,$F2,$4F,$0F,$49,$5C
	.dc.b	$25,$5C,$33,$F4,$37,$1F,$26,$5B,$8C,$57,$7C,$39,$BD,$CE,$6E,$2F
	.dc.b	$E1,$67,$F0,$67,$55,$C9,$DE,$3D,$6F,$86,$3E,$09,$DC,$61,$F0,$4E
	.dc.b	$CF,$55,$F5,$C3,$D3,$5B,$AE,$A7,$2A,$DE,$5F,$0D,$E5,$19,$3D,$57
	.dc.b	$CF,$0F,$F6,$7E,$3E,$72,$61,$D6,$75,$5E,$43,$96,$3E,$2C,$3B,$E5
	.dc.b	$DA,$31,$3A,$C7,$DD,$18,$23,$E0,$C3,$B6,$5E,$7A,$E0,$77,$AF,$9B
	.dc.b	$19,$7A,$C7,$FB,$7B,$2F,$C1,$B6,$BE,$BB,$71,$FF,$71,$BF,$89,$F4
	.dc.b	$BF,$CA,$2E,$29,$18,$8E,$8D,$7E,$2B,$5E,$5D,$BF,$87,$8E,$78,$7E
	.dc.b	$31,$B9,$7C,$CF,$07,$FD,$F4,$F2,$AE,$04,$BF,$C3,$CE,$9E,$A7,$E2
	.dc.b	$AB,$9B,$8B,$FC,$CC,$EE,$3F,$3B,$3A,$6D,$75,$AD,$C5,$BF,$13,$D6
	.dc.b	$7F,$41,$7C,$F0,$F8,$71,$B9,$F0,$57,$AD,$C1,$D8,$E4,$7A,$8C,$CF
	.dc.b	$27,$37,$E2,$6C,$F8,$A3,$E4,$57,$8A,$78,$E1,$EB,$BE,$55,$C4,$DC
	.dc.b	$F2,$3C,$A7,$F0,$5C,$3D,$55,$C4,$97,$D6,$7F,$6B,$E0,$FB,$EA,$47
	.dc.b	$F0,$18,$F8,$46,$1F,$61,$C0,$FE,$19,$FA,$27,$F2,$07,$FD,$79,$F8
	.dc.b	$FB,$C7,$46,$1A,$FF,$BE,$A7,$83,$78,$73,$CB,$FC,$CE,$13,$A8,$C3
	.dc.b	$78,$B9,$FE,$74,$F9,$8E,$31,$8F,$96,$B8,$78,$AB,$4F,$F7,$69,$E3
	.dc.b	$5A,$84,$AB,$0D,$8E,$07,$39,$C9,$A9,$DF,$93,$E0,$F1,$9F,$86,$FF
	.dc.b	$82,$AF,$8C,$63,$C6,$DC,$74,$3C,$35,$C6,$B8,$47,$F8,$0C,$57,$17
	.dc.b	$86,$3F,$E3,$27,$96,$6F,$AC,$F8,$3F,$1A,$FC,$7D,$F5,$98,$6D,$F0
	.dc.b	$AF,$85,$7C,$63,$04,$3C,$5C,$0F,$37,$CA,$FD,$7F,$C5,$DF,$37,$E9
	.dc.b	$FB,$49,$7E,$D9,$CB,$67,$2F,$B7,$6F,$8F,$5F,$4E,$B7,$3F,$04,$7D
	.dc.b	$BB,$EE,$3E,$C3,$63,$E5,$35,$E0,$FF,$B2,$DC,$C7,$02,$3C,$3E,$1F
	.dc.b	$F0,$5F,$50,$FB,$6A,$49,$FA,$D0,$C7,$11,$E5,$8E,$0E,$A7,$9C,$FE
	.dc.b	$37,$F1,$D7,$C0,$F6,$F1,$65,$FE,$06,$72,$D9,$CB,$D6,$79,$44,$8D
	.dc.b	$10,$F1,$7A,$9F,$9B,$E8,$7C,$1D,$24,$FE,$3E,$1F,$07,$88,$F9,$D1
	.dc.b	$C1,$D4,$F3,$9F,$D3,$7D,$57,$CD,$E4,$FD,$B3,$F3,$6B,$EB,$36,$33
	.dc.b	$C1,$C0,$E3,$C1,$2B,$F4,$15,$CF,$C0,$FE,$0B,$E5,$BE,$DB,$F6,$99
	.dc.b	$7C,$FB,$FD,$33,$5F,$C2,$4E,$DF,$D5,$35,$FC,$44,$6F,$2F,$E1,$8D
	.dc.b	$BE,$F7,$FC,$06,$F8,$57,$F1,$16,$CE,$AF,$C3,$5E,$4F,$03,$CB,$C1
	.dc.b	$27,$CB,$19,$7C,$F1,$97,$C7,$1F,$C3,$FE,$D7,$E7,$47,$E7,$5B,$8F
	.dc.b	$F2,$16,$E1,$FC,$A1,$FA,$F3,$FC,$1D,$CB,$FC,$B9,$A9,$A7,$07,$F5
	.dc.b	$5F,$8E,$BE,$2B,$F9,$07,$B5,$F5,$EF,$F9,$AB,$6B,$F9,$CD,$6D,$FE
	.dc.b	$2C,$D7,$F6,$91,$B4,$D8,$D6,$1F,$CE,$DB,$5F,$D0,$9B,$5D,$AF,$C7
	.dc.b	$F7,$7F,$CC,$FD,$DF,$D5,$CF,$F0,$17,$EC,$3F,$D3,$1A,$FF,$AD,$3A
	.dc.b	$78,$1C,$7A,$64,$7F,$2D,$9E,$79,$7F,$4D,$7F,$8C,$E5,$FD,$51,$B3
	.dc.b	$D7,$5F,$16,$DE,$36,$47,$37,$2F,$F3,$9F,$96,$FE,$8C,$FF,$17,$F2
	.dc.b	$7E,$86,$3F,$D6,$DB,$8F,$F6,$2A,$DC,$FF,$5F,$6C,$7F,$66,$9D,$D7
	.dc.b	$F6,$A6,$A9,$DB,$FD,$B9,$AE,$7C,$9F,$EB,$91,$89,$7C,$5F,$F5,$FF
	.dc.b	$CD,$9F,$25,$FD,$9B,$27,$E2,$4F,$AA,$C3,$73,$FE,$AD,$7C,$1F,$EB
	.dc.b	$1F,$9B,$FF,$73,$7F,$BF,$C7,$F3,$67,$7E,$27,$F6,$D9,$D4,$3F,$37
	.dc.b	$7C,$0F,$87,$97,$F2,$19,$FF,$31,$EF,$FE,$E9,$5A,$BF,$EB,$BE,$EF
	.dc.b	$F5,$CF,$65,$F3,$D4,$FE,$B3,$1D,$F8,$3F,$B0,$FE,$3B,$FD,$2B,$2F
	.dc.b	$F5,$F7,$A6,$8C,$73,$FF,$85,$37,$F6,$67,$FE,$15,$F7,$FF,$B4,$FD
	.dc.b	$EF,$E8,$F0,$3F,$9F,$F4,$BF,$D8,$B5,$FD,$C3,$BF,$E3,$FA,$32,$30
	.dc.b	$7F,$7D,$FC,$DF,$D4,$FF,$5B,$FA,$E7,$C9,$7F,$84,$5B,$1B,$E7,$FE
	.dc.b	$E7,$E6,$7C,$5F,$56,$4F,$F7,$E8,$FE,$EF,$D1,$FD,$FB,$F1,$DF,$87
	.dc.b	$3F,$E1,$6D,$D1,$88,$FF,$18,$4F,$F8,$7F,$F3,$7F,$E6,$10,$E5,$7E
	.dc.b	$A2,$E0,$FC,$23,$67,$FC,$2D,$5C,$1F,$F1,$AF,$B7,$FE,$0E,$F8,$6E
	.dc.b	$79,$FF,$B4,$34,$FF,$05,$CF,$F8,$B7,$F1,$BF,$E0,$DF,$1B,$FB,$83
	.dc.b	$F7,$2F,$F0,$9F,$CD,$DD,$BF,$E3,$3F,$9C,$FD,$E9,$FB,$E8,$CE,$7F
	.dc.b	$E2,$7A,$F2,$3F,$E0,$58,$CB,$6B,$C1,$B7,$F8,$73,$8B,$F3,$DF,$61
	.dc.b	$FE,$20,$AF,$F0,$DB,$78,$7F,$C9,$B3,$FE,$43,$FC,$5B,$F8,$AD,$7F
	.dc.b	$94,$A3,$93,$FE,$37,$8C,$93,$D1,$E0,$7E,$AE,$5E,$7F,$3F,$88,$F8
	.dc.b	$5D,$93,$0E,$51,$97,$D9,$19,$7F,$C4,$18,$27,$51,$86,$E3,$FE,$67
	.dc.b	$D9,$F8,$07,$F8,$EB,$E6,$3C,$5F,$51,$30,$FF,$22,$C6,$5F,$F2,$DD
	.dc.b	$FF,$8E,$E7,$51,$FE,$75,$9F,$F2,$1F,$D7,$FF,$05,$F5,$BF,$81,$4F
	.dc.b	$F9,$09,$B7,$FE,$74,$9D,$CF,$F9,$49,$AF,$F9,$23,$67,$C8,$FF,$9B
	.dc.b	$FE,$DB,$FC,$F1,$7F,$E8,$6F,$B6,$FF,$1F,$F0,$76,$6F,$A7,$FE,$88
	.dc.b	$7F,$C8,$75,$C3,$14,$63,$D5,$FC,$D4,$62,$FF,$D1,$E7,$32,$DF,$37
	.dc.b	$FD,$27,$7C,$EF,$96,$1E,$39,$DD,$D3,$FD,$36,$F3,$7F,$D3,$B5,$C6
	.dc.b	$39,$E1,$E2,$BF,$4D,$71,$3F,$CC,$B8,$78,$5E,$33,$BA,$E3,$86,$DE
	.dc.b	$D3,$C0,$E5,$87,$77,$8C,$EE,$FA,$46,$CF,$84,$68,$A2,$79,$3B,$C3
	.dc.b	$CE,$EA,$32,$7C,$F7,$FD,$72,$71,$C3,$E7,$B8,$8D,$3E,$A3,$67,$7F
	.dc.b	$5F,$FD,$32,$E2,$3F,$D5,$DF,$11,$A8,$D3,$63,$31,$91,$F2,$55,$A6
	.dc.b	$5C,$11,$F7,$E4,$F1,$DE,$77,$D3,$9F,$F6,$AD,$7D,$5A,$D4,$E6,$7A
	.dc.b	$E7,$FD,$26,$F8,$C7,$CA,$38,$9D,$9C,$C7,$FA,$62,$33,$F0,$BC,$6B
	.dc.b	$FD,$C6,$E6,$3D,$2E,$53,$9C,$3C,$97,$D3,$6F,$97,$18,$BF,$33,$6F
	.dc.b	$C8,$7A,$27,$FD,$D7,$5C,$63,$EC,$CE,$F0,$ED,$1E,$9B,$E6,$F8,$AF
	.dc.b	$8C,$7A,$1F,$0C,$64,$FF,$7E,$E3,$F2,$1C,$0F,$AB,$FF,$48,$D7,$FB
	.dc.b	$75,$B1,$FE,$CA,$FB,$5F,$F6,$54,$E7,$E1,$BE,$6D,$FF,$C0,$1B,$DD
	.dc.b	$FE,$41,$A8,$8E,$1B,$3F,$59,$E1,$F0,$B5,$E0,$79,$5C,$3C,$9A,$E6
	.dc.b	$FF,$29,$5F,$EA,$67,$F9,$3E,$1F,$E2,$87,$D9,$5E,$87,$E1,$C6,$6B
	.dc.b	$FD,$B1,$E3,$FF,$5C,$9F,$F7,$F3,$63,$E1,$7D,$F7,$17,$D9,$19,$CB
	.dc.b	$D4,$EE,$3F,$62,$D7,$D6,$D7,$8F,$87,$EA,$8E,$57,$E4,$EA,$ED,$49
	.dc.b	$CB,$FE,$01,$C5,$3A,$F9,$57,$8D,$4F,$36,$4A,$75,$E7,$2F,$C1,$9F
	.dc.b	$3A,$F8,$BF,$BD,$6E,$3F,$C1,$1A,$FF,$0C,$6B,$F5,$B6,$7D,$27,$D4
	.dc.b	$6D,$7D,$48,$C1,$9F,$46,$47,$D8,$6D,$77,$F0,$BE,$EE,$1F,$DA,$EB
	.dc.b	$C9,$18,$3B,$7D,$1C,$8F,$C0,$B6,$BF,$8D,$E2,$FA,$3F,$21,$2B,$F0
	.dc.b	$EF,$E2,$BE,$C9,$E6,$52,$4F,$F1,$F8,$CB,$F8,$EF,$A4,$F7,$11,$C1
	.dc.b	$F9,$29,$F8,$44,$70,$7F,$29,$1D,$FF,$9A,$69,$97,$07,$CB,$F9,$89
	.dc.b	$1F,$9E,$6E,$5E,$EF,$9A,$FA,$04,$9F,$D1,$BF,$A6,$ED,$FF,$2C,$C7
	.dc.b	$F3,$96,$CF,$F1,$6E,$0D,$1F,$67,$79,$5F,$AC,$6C,$FC,$CF,$A3,$FB
	.dc.b	$C4,$9F,$D8,$BF,$B4,$FA,$CF,$F0,$4C,$7F,$47,$6C,$FF,$4E,$6B,$FD
	.dc.b	$59,$AF,$F0,$0F,$9E,$71,$FE,$F9,$B5,$FD,$45,$B3,$FD,$61,$AF,$F6
	.dc.b	$0F,$AB,$FF,$4B,$CF,$E7,$A3,$F8,$94,$61,$FE,$39,$F4,$BF,$AE,$8C
	.dc.b	$74,$FF,$13,$75,$3C,$FD,$CF,$E1,$F1,$88,$FE,$2A,$D7,$AC,$7E,$C3
	.dc.b	$EA,$FF,$AC,$E1,$E1,$95,$76,$8F,$E6,$73,$A9,$FF,$3E,$46,$3E,$AF
	.dc.b	$91,$CF,$D4,$B1,$FE,$42,$A4,$E2,$B8,$9F,$E8,$58,$56,$71,$FE,$6F
	.dc.b	$0F,$95,$F3,$7C,$47,$9D,$B8,$F5,$6E,$1D,$2D,$7C,$6B,$C5,$7F,$C5
	.dc.b	$BE,$43,$F8,$3D,$7F,$3B,$6C,$FB,$BC,$1D,$11,$FF,$05,$47,$F0,$AC
	.dc.b	$3F,$A3,$35,$34,$E6,$FB,$7E,$3F,$93,$E5,$C7,$FD,$CF,$5F,$17,$C8
	.dc.b	$FA,$47,$D3,$F2,$2C,$64,$95,$C6,$32,$F3,$FA,$BF,$B5,$F3,$3F,$6D
	.dc.b	$B5,$F4,$5B,$3B,$8F,$A9,$3F,$CE,$3E,$AD,$F5,$74,$FC,$27,$8D,$6B
	.dc.b	$F0,$93,$A7,$FC,$58,$DB,$FF,$19,$B5,$FF,$1D,$B5,$FE,$CA,$EF,$7F
	.dc.b	$BB,$D9,$D6,$7E,$23,$F9,$3F,$1B,$F2,$9E,$2D,$C3,$F1,$4D,$AF,$94
	.dc.b	$DB,$FC,$3B,$F9,$BF,$12,$57,$E4,$7C,$4B,$7F,$46,$6D,$FE,$3A,$6D
	.dc.b	$7F,$7A,$FA,$BF,$EA,$EF,$E8,$DB,$7F,$67,$C4,$FE,$BF,$18,$BF,$F0
	.dc.b	$23,$A9,$FF,$2D,$B5,$E9,$E5,$7F,$80,$A3,$11,$FE,$0F,$79,$5F,$F8
	.dc.b	$83,$E6,$1F,$47,$DA,$FF,$B2,$C6,$27,$F5,$9E,$3F,$EE,$4F,$DD,$78
	.dc.b	$D6,$32,$4C,$7F,$21,$19,$7F,$25,$E3,$7F,$01,$3F,$B8,$FB,$3F,$C5
	.dc.b	$B4,$FD,$8C,$66,$FF,$74,$EB,$C6,$C5,$FE,$E8,$7D,$57,$E8,$AB,$FC
	.dc.b	$E0,$D8,$FF,$3B,$B5,$FF,$3F,$B5,$FF,$14,$BB,$BF,$F3,$1F,$AE,$F9
	.dc.b	$97,$91,$FE,$0E,$AD,$61,$FE,$37,$AD,$4F,$F8,$4B,$C9,$FF,$72,$CF
	.dc.b	$F1,$B6,$E3,$FE,$C7,$F5,$F5,$F0,$32,$BF,$9B,$3F,$CE,$48,$FA,$BE
	.dc.b	$AB,$FC,$73,$91,$A4,$8D,$1F,$CE,$72,$71,$7F,$72,$C7,$E3,$5F,$1A
	.dc.b	$D4,$E7,$5F,$F2,$64,$FF,$35,$AF,$97,$E6,$78,$09,$7F,$31,$FF,$17
	.dc.b	$FD,$37,$F9,$5E,$3F,$9C,$57,$F3,$7A,$FF,$2D,$4E,$5F,$AD,$AB,$EA
	.dc.b	$BE,$31,$CA,$2D,$FC,$78,$FE,$B4,$47,$FA,$D3,$FE,$5A,$AF,$F2,$74
	.dc.b	$7F,$AE,$1B,$3F,$E6,$26,$BE,$2F,$5F,$FD,$3B,$3B,$7F,$D3,$FE,$8B
	.dc.b	$7E,$1F,$25,$87,$94,$DB,$F5,$CD,$5F,$53,$FD,$17,$E3,$3F,$D5,$F7
	.dc.b	$FE,$72,$8C,$4F,$FA,$F2,$3A,$E3,$A8,$FF,$3F,$4E,$B1,$F1,$46,$3D
	.dc.b	$06,$1B,$8E,$F1,$87,$A1,$E4,$BD,$5F,$3C,$3C,$55,$BC,$3A,$DF,$2C
	.dc.b	$3A,$66,$CB,$D9,$8F,$1A,$F1,$57,$4C,$39,$63,$EA,$9F,$09,$CC,$B5
	.dc.b	$C5,$C6,$3E,$0C,$7B,$65,$CF,$19,$5C,$31,$EA,$FA,$CC,$46,$AF,$D3
	.dc.b	$87,$5C,$39,$D6,$89,$3A,$BE,$70,$9D,$CE,$6F,$A6,$3F,$19,$F8,$79
	.dc.b	$7C,$2C,$38,$E3,$EB,$3C,$11,$99,$E3,$8F,$C6,$3E,$1E,$3D,$B0,$E9
	.dc.b	$1E,$46,$E3,$F0,$EF,$94,$FA,$4F,$F5,$C9,$F2,$CB,$87,$E2,$9B,$8F
	.dc.b	$E0,$5C,$FA,$1F,$E4,$07,$89,$F9,$47,$0E,$BA,$32,$E5,$B4,$9E,$58
	.dc.b	$7C,$E6,$E3,$F9,$B6,$DF,$E6,$3B,$3F,$D9,$F5,$BC,$3F,$23,$5A,$AE
	.dc.b	$C7,$E1,$2B,$77,$CE,$B6,$49,$D1,$6F,$A3,$F8,$CB,$E9,$83,$0E,$73
	.dc.b	$FA,$A9,$C9,$2F,$F4,$73,$A9,$F1,$DF,$42,$61,$FB,$79,$D6,$1F,$0E
	.dc.b	$FC,$53,$E7,$BF,$29,$E3,$8C,$BE,$17,$BB,$83,$BC,$71,$BF,$CC,$DF
	.dc.b	$6C,$3C,$D5,$CC,$EB,$3E,$47,$A1,$97,$8B,$D7,$0F,$33,$9B,$F5,$3F
	.dc.b	$C0,$BB,$3D,$78,$48,$C1,$06,$DF,$8E,$94,$7F,$37,$3F,$8A,$B8,$76
	.dc.b	$7F,$1D,$71,$1F,$98,$F9,$2F,$D7,$4F,$E3,$9A,$FE,$51,$AE,$F6,$FC
	.dc.b	$9A,$8D,$F2,$F6,$8C,$4F,$86,$31,$3A,$AF,$27,$D4,$3E,$0B,$51,$8F
	.dc.b	$62,$E3,$EC,$AD,$3F,$E8,$D6,$BF,$E0,$97,$FD,$9A,$E6,$3C,$B1,$88
	.dc.b	$FD,$6B,$5F,$DB,$B5,$F1,$F8,$5E,$0E,$B7,$D1,$FF,$9D,$27,$FB,$A3
	.dc.b	$5F,$04,$F9,$95,$7F,$8C,$1F,$DB,$B5,$F0,$C7,$90,$F9,$39,$3E,$9E
	.dc.b	$92,$7E,$93,$5F,$F1,$34,$6D,$FF,$57,$35,$FF,$4C,$78,$7E,$5D,$7E
	.dc.b	$97,$D3,$BE,$0B,$D9,$9B,$E2,$29,$E9,$9D,$61,$FE,$B9,$6B,$FC,$51
	.dc.b	$AF,$F1,$E6,$BF,$8D,$6B,$F9,$2F,$9E,$7C,$97,$C5,$6C,$7F,$B1,$5A
	.dc.b	$FF,$90,$57,$CA,$EE,$64,$ED,$E5,$1F,$3F,$E9,$DF,$3F,$49,$5F,$ED
	.dc.b	$18,$F7,$36,$32,$FF,$93,$E3,$8B,$BC,$7E,$D7,$D3,$BD,$8E,$DD,$72
	.dc.b	$BF,$DB,$93,$96,$CF,$E8,$A3,$11,$FC,$E9,$AF,$F4,$46,$BF,$B3,$77
	.dc.b	$BF,$E9,$DB,$CF,$9A,$F7,$79,$5F,$8D,$E0,$7B,$74,$98,$FF,$BB,$E1
	.dc.b	$ED,$E2,$3F,$CD,$91,$C1,$D4,$F3,$9F,$CC,$7D,$3F,$D2,$CF,$F7,$0F
	.dc.b	$D2,$AF,$F7,$D3,$5F,$F3,$B3,$BC,$BF,$86,$36,$3F,$88,$BB,$9E,$7C
	.dc.b	$1D,$B8,$7F,$4D,$E2,$3E,$9F,$F5,$19,$7D,$6C,$7F,$D0,$4D,$C3,$F8
	.dc.b	$55,$6E,$7F,$8E,$C6,$2B,$F8,$8C,$E6,$7F,$B4,$B5,$FE,$E6,$D7,$F9
	.dc.b	$13,$B9,$E5,$1D,$32,$CD,$79,$0F,$2F,$C0,$79,$B8,$65,$6C,$FE,$5F
	.dc.b	$8F,$EB,$EF,$F9,$73,$63,$33,$C1,$F6,$D7,$F0,$7F,$A7,$FA,$91,$FE
	.dc.b	$A1,$6E,$3F,$CA,$5B,$87,$F3,$97,$EE,$CF,$84,$3A,$7F,$A0,$B5,$CE
	.dc.b	$DD,$F1,$FE,$2D,$F4,$EF,$84,$FE,$31,$E8,$FD,$7C,$7F,$A3,$37,$0F
	.dc.b	$E9,$75,$B7,$C0,$D7,$FC,$2E,$D7,$FC,$42,$D7,$FA,$83,$5E,$75,$E2
	.dc.b	$23,$EB,$97,$CC,$7F,$2B,$F5,$7F,$6B,$3F,$EC,$57,$F0,$98,$7F,$5B
	.dc.b	$6C,$7F,$60,$77,$A7,$F5,$16,$CE,$FA,$3A,$1D,$7A,$BB,$6F,$30,$F9
	.dc.b	$97,$F6,$5F,$F6,$47,$06,$5F,$14,$DC,$7B,$4F,$F4,$1F,$A7,$FE,$9D
	.dc.b	$1F,$D8,$7C,$6F,$AF,$7F,$ED,$D6,$C7,$F6,$59,$DE,$7F,$DD,$5F,$EA
	.dc.b	$55,$FD,$E9,$A9,$A4,$F1,$BC,$04,$8F,$C2,$5B,$E1,$7F,$AA,$7D,$33
	.dc.b	$E9,$3F,$BB,$B6,$BF,$C9,$6D,$7F,$CA,$2D,$7F,$CB,$7F,$47,$FE,$00
	.dc.b	$7F,$C1,$4D,$7F,$C1,$ED,$4D,$FA,$67,$B3,$FE,$06,$6C,$FF,$96,$9A
	.dc.b	$FF,$98,$1A,$FF,$84,$FD,$6F,$EC,$A4,$7F,$C2,$8F,$F8,$42,$BC,$D1
	.dc.b	$E4,$C3,$04,$7F,$C3,$5E,$AF,$F6,$E2,$38,$3F,$C3,$DC,$BF,$DC,$88
	.dc.b	$E0,$FF,$37,$FD,$1F,$91,$FF,$15,$35,$FF,$17,$B5,$37,$CB,$FE,$BB
	.dc.b	$1F,$E2,$DF,$03,$FB,$F9,$1F,$F1,$77,$83,$FE,$02,$23,$83,$FC,$FF
	.dc.b	$F4,$3C,$9F,$F1,$D3,$5F,$F1,$FB,$53,$69,$F4,$3F,$E0,$E2,$4F,$F8
	.dc.b	$F5,$FF,$1C,$FC,$F7,$F3,$D8,$FF,$4C,$B5,$FF,$24,$35,$FF,$26,$B5
	.dc.b	$7D,$0B,$7F,$F2,$1B,$63,$FD,$2E,$D7,$FD,$3A,$D7,$BE,$9E,$22,$3F
	.dc.b	$E9,$EA,$FE,$06,$FF,$94,$FC,$FF,$F1,$39,$1C,$1F,$EA,$4F,$A0,$FE
	.dc.b	$1C,$9A,$4F,$07,$FC,$AE,$FF,$19,$EB,$FF,$2B,$37,$AF,$FC,$CC,$DA
	.dc.b	$4A,$FF,$DE,$2F,$0F,$5B,$FC,$81,$8F,$F8,$9A,$3F,$CC,$0D,$C3,$9C
	.dc.b	$FF,$91,$70,$3E,$22,$61,$E9,$6B,$E0,$9E,$2F,$F3,$F9,$E2,$FF,$41
	.dc.b	$A7,$D0,$7F,$94,$77,$7C,$27,$F9,$A3,$2F,$47,$D0,$5D,$0E,$E6,$63
	.dc.b	$FE,$7F,$7F,$CA,$64,$7F,$D0,$5F,$31,$FE,$56,$23,$82,$1F,$E5,$47
	.dc.b	$C4,$77,$FA,$0F,$F3,$27,$CF,$FF,$9F,$71,$FF,$4A,$FA,$3F,$E4,$CF
	.dc.b	$5D,$F4,$1F,$E9,$58,$C3,$63,$FD,$31,$5C,$71,$57,$F9,$52,$31,$F3
	.dc.b	$FC,$27,$FD,$42,$D4,$EF,$C4,$FB,$71,$8B,$EB,$18,$9E,$04,$8C,$E3
	.dc.b	$D1,$B5,$C4,$FF,$57,$BE,$A7,$46,$31,$FF,$58,$57,$86,$B3,$2C,$3D
	.dc.b	$98,$F9,$C9,$87,$67,$84,$F7,$3F,$CF,$58,$74,$9F,$F5,$E1,$E1,$8D
	.dc.b	$57,$C4,$8F,$3E,$1E,$B9,$F2,$1E,$29,$D3,$FE,$C8,$7E,$73,$FE,$C7
	.dc.b	$3C,$D8,$FC,$87,$11,$A7,$E1,$98,$F1,$5E,$EF,$18,$E9,$18,$71,$28
	.dc.b	$C4,$71,$9F,$F6,$5A,$F8,$1A,$47,$FD,$9E,$DC,$3F,$61,$F4,$1F,$ED
	.dc.b	$C7,$FD,$27,$1F,$3B,$54,$E7,$3F,$24,$E6,$31,$7E,$F8,$C7,$3F,$D6
	.dc.b	$8F,$F7,$5C,$61,$FA,$46,$0F,$A9,$7D,$A7,$84,$7A,$4C,$D7,$67,$FD
	.dc.b	$E5,$5C,$F0,$F3,$C6,$AB,$8C,$E7,$0F,$3C,$E4,$F0,$9D,$6B,$E0,$D7
	.dc.b	$08,$D4,$74,$32,$FE,$27,$E7,$7F,$D1,$33,$FE,$49,$6C,$7F,$0F,$3F
	.dc.b	$8F,$FC,$F7,$F1,$F9,$D7,$A8,$69,$27,$86,$6A,$E1,$1C,$C5,$F6,$9E
	.dc.b	$50,$33,$EA,$62,$33,$18,$BF,$F3,$74,$3F,$2F,$7C,$96,$E7,$95,$6C
	.dc.b	$F0,$61,$FE,$A2,$8F,$76,$1B,$9F,$E7,$77,$CA,$72,$7F,$9A,$FE,$83
	.dc.b	$FA,$D3,$7C,$3E,$19,$7B,$7E,$87,$96,$0B,$DB,$EF,$6C,$FA,$9B,$7F
	.dc.b	$E9,$8D,$17,$CA,$73,$7F,$0D,$B7,$F2,$FA,$B5,$8F,$8B,$0F,$E7,$B8
	.dc.b	$FC,$45,$D5,$E4,$F8,$A7,$CE,$EE,$30,$CA,$FF,$44,$4E,$9E,$53,$9C
	.dc.b	$FE,$4F,$CD,$7B,$B8,$0E,$74,$FA,$B1,$8B,$FC,$57,$CC,$EC,$F8,$FF
	.dc.b	$A9,$3E,$5F,$13,$E0,$F1,$FC,$C6,$32,$FC,$36,$9F,$D9,$27,$F2,$9F
	.dc.b	$35,$E1,$3E,$7F,$DB,$BE,$E3,$13,$F9,$9F,$9C,$F3,$1F,$77,$95,$F6
	.dc.b	$FE,$19,$D1,$FE,$A7,$FA,$57,$CF,$7D,$A6,$C7,$FA,$EE,$9F,$97,$7F
	.dc.b	$C0,$18,$E1,$5F,$0F,$B6,$46,$FC,$77,$CE,$E3,$83,$F2,$37,$F2,$4F
	.dc.b	$C4,$FE,$A7,$FE,$62,$BF,$D8,$47,$E2,$9F,$E4,$47,$FA,$EC,$B3,$FA
	.dc.b	$CF,$B7,$7D,$36,$FC,$9F,$A9,$CD,$2F,$F4,$BF,$6F,$FC,$47,$0F,$D0
	.dc.b	$CF,$D9,$F9,$A9,$5E,$DF,$8C,$7D,$16,$2F,$F6,$DE,$5B,$B7,$F3,$AD
	.dc.b	$BF,$89,$E9,$5A,$DD,$7E,$17,$B6,$57,$C9,$FB,$7F,$3B,$FC,$36,$3F
	.dc.b	$15,$8C,$8F,$DD,$FD,$BF,$E4,$70,$FD,$BC,$FE,$B3,$EA,$DE,$AD,$3F
	.dc.b	$63,$F0,$8F,$57,$18,$FF,$B8,$7E,$8D,$E2,$7E,$D1,$B1,$FC,$23,$ED
	.dc.b	$DF,$09,$9A,$FD,$CF,$D6,$7D,$3D,$7F,$8B,$37,$CA,$FE,$25,$E7,$7D
	.dc.b	$CC,$7F,$DD,$D3,$97,$F8,$7B,$6B,$F8,$D3,$6B,$F9,$17,$7F,$6F,$8A
	.dc.b	$97,$C9,$FD,$D4,$62,$FF,$94,$FD,$67,$C8,$6F,$77,$F1,$08,$FF,$7B
	.dc.b	$5F,$F2,$27,$58,$7F,$27,$6C,$7F,$2B,$84,$62,$BF,$CB,$2F,$27,$AC
	.dc.b	$F3,$F9,$7E,$DF,$49,$DD,$B1,$93,$C1,$19,$69,$DC,$FE,$63,$F2,$5F
	.dc.b	$6F,$0F,$E7,$2D,$8F,$26,$17,$0F,$E8,$13,$53,$4D,$57,$AB,$D3,$FE
	.dc.b	$5F,$19,$69,$F1,$4B,$7F,$17,$D9,$F9,$07,$C9,$F6,$6C,$64,$98,$7D
	.dc.b	$08,$CB,$F4,$BE,$B3,$E8,$7C,$E7,$D1,$6D,$7D,$16,$CE,$E3,$EA,$4F
	.dc.b	$F4,$4F,$AB,$7D,$65,$3E,$D7,$B1,$6B,$E4,$5F,$F6,$1A,$DC,$62,$11
	.dc.b	$87,$F0,$B3,$C5,$FE,$D9,$19,$D7,$FB,$3B,$6B,$E7,$D6,$E7,$F0,$DE
	.dc.b	$B7,$E2,$FD,$4F,$C6,$36,$BF,$18,$DC,$3F,$26,$DA,$FC,$77,$A1,$F9
	.dc.b	$33,$F2,$9E,$75,$BF,$CA,$45,$BF,$CD,$CD,$AF,$EF,$7F,$59,$FD,$62
	.dc.b	$7F,$C0,$8D,$8F,$CD,$79,$9F,$DB,$23,$18,$7F,$7C,$6D,$7F,$81,$DB
	.dc.b	$5F,$9E,$AE,$1E,$E4,$BF,$95,$5A,$E5,$FE,$85,$87,$82,$BF,$59,$E6
	.dc.b	$CB,$FD,$67,$91,$6B,$F6,$73,$72,$FD,$74,$DA,$FC,$DF,$8F,$FB,$53
	.dc.b	$FC,$0B,$F0,$BF,$C4,$30,$FD,$DC,$66,$7F,$70,$EB,$C1,$C4,$FF,$89
	.dc.b	$7E,$AF,$F4,$35,$F8,$16,$C7,$F8,$CF,$E1,$3F,$C1,$DD,$5F,$E1,$09
	.dc.b	$CE,$7F,$E2,$66,$DF,$F8,$89,$B7,$FE,$72,$F5,$FF,$C6,$13,$9E,$DF
	.dc.b	$E4,$DF,$31,$23,$FC,$5B,$AF,$FA,$B1,$FE,$59,$19,$BF,$F1,$D7,$90
	.dc.b	$8F,$7F,$C3,$E7,$3F,$17,$D6,$79,$4F,$F1,$76,$DF,$CD,$3C,$DC,$65
	.dc.b	$ED,$9B,$97,$B7,$83,$F9,$FC,$E4,$F8,$5E,$3F,$BC,$F8,$1F,$39,$63
	.dc.b	$24,$AF,$95,$19,$7C,$FF,$59,$FD,$0C,$FF,$1B,$FD,$57,$F9,$3E,$BF
	.dc.b	$A4,$46,$5A,$7E,$C6,$35,$3C,$A3,$1F,$0B,$FE,$58,$F5,$FE,$14,$FF
	.dc.b	$96,$6B,$FC,$83,$18,$D5,$F0,$4F,$93,$E1,$C5,$FE,$56,$BD,$B7,$1C
	.dc.b	$E3,$CD,$B8,$72,$BC,$CA,$F8,$BF,$E6,$9F,$AB,$79,$AF,$AC,$FF,$4B
	.dc.b	$E5,$FE,$64,$9C,$C6,$2B,$D1,$F3,$3D,$27,$31,$FE,$AF,$AE,$17,$CE
	.dc.b	$30,$77,$71,$AF,$C2,$8F,$25,$F3,$9C,$D7,$2C,$3C,$14,$C3,$BE,$1C
	.dc.b	$71,$63,$C6,$F7,$5A,$C3,$96,$1E,$23,$BE,$DA,$CB,$BC,$79,$6F,$07
	.dc.b	$83,$1E,$B5,$89,$E5,$87,$57,$81,$D6,$76,$EA,$FA,$BE,$96,$CF,$18
	.dc.b	$F1,$9B,$BE,$D9,$7C,$59,$D5,$F4,$8C,$5F,$22,$57,$9F,$2F,$85,$87
	.dc.b	$32,$57,$C7,$8C,$BC,$A7,$BE,$5F,$8D,$9C,$BF,$22,$75,$5E,$D9,$CB
	.dc.b	$F0,$0F,$F7,$F6,$3F,$82,$6E,$1F,$81,$6E,$1F,$EF,$B7,$ED,$79,$FF
	.dc.b	$E3,$07,$FD,$FD,$48,$C4,$74,$6B,$F9,$36,$BF,$99,$79,$61,$F1,$DB
	.dc.b	$8E,$F8,$FF,$07,$E0,$7F,$BA,$23,$17,$F6,$AB,$6F,$3A,$DB,$F8,$A8
	.dc.b	$CB,$E8,$8C,$1A,$79,$38,$69,$F8,$6A,$F0,$3D,$4C,$DF,$9E,$79,$36
	.dc.b	$33,$7E,$68,$FC,$15,$72,$9F,$11,$A3,$A4,$69,$F1,$3C,$C9,$3F,$BB
	.dc.b	$9E,$58,$7C,$FB,$DC,$F7,$35,$78,$3A,$B9,$7F,$84,$57,$2C,$3E,$2D
	.dc.b	$F3,$9F,$21,$FB,$79,$D5,$F3,$8F,$4E,$1E,$77,$F5,$7E,$07,$90,$8F
	.dc.b	$F0,$FE,$07,$C2,$3E,$AF,$33,$C0,$FD,$FD,$7C,$EA,$DC,$FC,$7F,$91
	.dc.b	$FE,$41,$5F,$E2,$56,$AE,$B7,$1A,$24,$EA,$73,$19,$FA,$CF,$F4,$53
	.dc.b	$CF,$93,$83,$CE,$ED,$F1,$6C,$EA,$F9,$D2,$F9,$BC,$4E,$DF,$07,$BB
	.dc.b	$EB,$28,$C4,$E8,$FF,$20,$47,$FB,$75,$B1,$A7,$B3,$63,$94,$70,$F8
	.dc.b	$9C,$65,$23,$9B,$DB,$37,$E7,$BF,$D1,$A7,$24,$3D,$07,$AE,$78,$46
	.dc.b	$E3,$3A,$3E,$B7,$C8,$D8,$FA,$91,$EC,$D1,$E1,$FF,$9F,$27,$24,$39
	.dc.b	$CE,$E3,$51,$E2,$FA,$D7,$C9,$FC,$16,$C7,$F9,$95,$AF,$F8,$97,$8F
	.dc.b	$F8,$7E,$1F,$52,$72,$DA,$E4,$7C,$39,$DF,$07,$E0,$3E,$B7,$FD,$8B
	.dc.b	$87,$F9,$C5,$AF,$C4,$6D,$FE,$01,$7C,$F7,$D5,$73,$5C,$FE,$03,$06
	.dc.b	$67,$ED,$7D,$6B,$E9,$E9,$31,$FF,$40,$43,$C5,$C4,$7D,$48,$E0,$EA
	.dc.b	$79,$CF,$E1,$7E,$B5,$EE,$52,$57,$4F,$2E,$BF,$7A,$76,$FF,$A3,$DA
	.dc.b	$F7,$8D,$EB,$F9,$E9,$CB,$6B,$D7,$3F,$95,$FA,$D7,$A3,$49,$87,$FA
	.dc.b	$6E,$1E,$FE,$23,$F1,$F1,$C1,$D4,$F3,$9F,$D0,$7D,$6F,$CC,$9F,$DE
	.dc.b	$3F,$41,$FF,$53,$35,$FC,$CB,$B4,$D3,$58,$FE,$E1,$BC,$1F,$C0,$1E
	.dc.b	$1E,$3B,$EB,$7F,$60,$DB,$FF,$5A,$35,$FD,$3E,$17,$0F,$E0,$73,$70
	.dc.b	$FE,$16,$DA,$FE,$28,$D7,$E1,$DF,$62,$3D,$9F,$E0,$17,$B3,$F4,$B7
	.dc.b	$FC,$41,$FF,$18,$37,$C7,$DF,$A5,$2F,$FD,$95,$5A,$C7,$F8,$EB,$63
	.dc.b	$F9,$33,$D2,$3F,$90,$FD,$6F,$D3,$9F,$E0,$6D,$C3,$F8,$FB,$67,$75
	.dc.b	$FC,$56,$31,$5F,$CC,$5A,$9A,$7E,$09,$FF,$6F,$7C,$83,$E8,$BF,$8F
	.dc.b	$36,$33,$1F,$C8,$7E,$B5,$F4,$9A,$89,$3A,$3F,$DD,$13,$A7,$07,$F1
	.dc.b	$CA,$E8,$D9,$E9,$5F,$CB,$7E,$B7,$F5,$B3,$FC,$89,$FD,$93,$FE,$F3
	.dc.b	$6B,$9E,$7E,$07,$1E,$79,$1B,$8E,$9B,$48,$E4,$F7,$DF,$11,$DE,$2E
	.dc.b	$FF,$D3,$30,$FA,$D2,$FA,$DF,$E7,$EF,$F3,$66,$CF,$F5,$26,$F0,$7F
	.dc.b	$5A,$F8,$F7,$D6,$FF,$49,$9C,$DF,$F6,$56,$BD,$7C,$AF,$E9,$D5,$C2
	.dc.b	$3F,$B5,$61,$D6,$79,$12,$33,$87,$7C,$3F,$AB,$7D,$6B,$E4,$A9,$2B
	.dc.b	$67,$F7,$7F,$B4,$FE,$8A,$E0,$7D,$6F,$F6,$F7,$FB,$D3,$5C,$F9,$FD
	.dc.b	$BF,$01,$FE,$C9,$F5,$EF,$7A,$92,$BB,$ED,$FC,$BF,$E9,$7F,$AB,$FB
	.dc.b	$B2,$7F,$C1,$4F,$F7,$2F,$83,$FF,$2C,$77,$FF,$6F,$D3,$D5,$5E,$56
	.dc.b	$BF,$D9,$9B,$87,$F8,$32,$9F,$69,$FD,$8E,$BF,$B3,$FB,$DF,$D9,$88
	.dc.b	$FC,$29,$7D,$A7,$FB,$16,$3F,$C3,$9D,$BF,$DB,$08,$FA,$B8,$5F,$4F
	.dc.b	$FE,$03,$FB,$1F,$81,$3F,$25,$AF,$F7,$A6,$E7,$FE,$16,$FA,$BF,$E7
	.dc.b	$3D,$1F,$E1,$CF,$3F,$FB,$C9,$2F,$E3,$F8,$1D,$9F,$F1,$77,$9F,$FD
	.dc.b	$F8,$8F,$A7,$ED,$3F,$8D,$7D,$97,$F2,$F2,$5F,$CB,$FA,$D7,$E1,$7F
	.dc.b	$E0,$6F,$87,$98,$7F,$8E,$5F,$A1,$F0,$FF,$CC,$A3,$FC,$71,$F6,$35
	.dc.b	$34,$C7,$CE,$7E,$CB,$83,$FC,$63,$F6,$6E,$CF,$F1,$33,$6F,$F1,$8D
	.dc.b	$7F,$C3,$B7,$FC,$15,$BC,$DF,$E2,$D8,$C5,$FF,$8F,$1B,$C5,$FA,$2E
	.dc.b	$27,$CF,$FF,$88,$3D,$E9,$8F,$F9,$4D,$FF,$1B,$7D,$9F,$F0,$EC,$7F
	.dc.b	$C8,$AD,$C3,$FC,$9A,$DA,$FF,$2B,$36,$7F,$C4,$FE,$6F,$F8,$8C,$E7
	.dc.b	$E7,$CA,$FC,$DE,$1F,$C1,$2D,$F0,$FE,$9F,$7E,$57,$EA,$E7,$2D,$9F
	.dc.b	$29,$1F,$D8,$6A,$FB,$3C,$47,$F9,$0A,$B8,$B6,$BE,$34,$FF,$90,$3E
	.dc.b	$56,$4F,$F9,$CD,$FF,$20,$7E,$47,$F3,$F5,$FF,$32,$FC,$2A,$77,$5F
	.dc.b	$E4,$A7,$7C,$9F,$E4,$AF,$02,$4F,$F9,$F5,$FF,$22,$7C,$9E,$23,$FC
	.dc.b	$A3,$17,$0F,$F3,$D9,$FE,$4E,$9F,$F3,$37,$DA,$3D,$3F,$F4,$25,$BB
	.dc.b	$7F,$C8,$FC,$2F,$AF,$FF,$3F,$C6,$1B,$1F,$E8,$2B,$E5,$8F,$FA,$2E
	.dc.b	$BF,$C9,$F1,$8B,$E9,$5F,$E8,$83,$A3,$B7,$17,$D2,$3F,$D2,$93,$C6
	.dc.b	$B9,$E1,$E6,$7C,$53,$C4,$FF,$4D,$5F,$47,$B9,$CA,$78,$47,$9D,$B1
	.dc.b	$C2,$FC,$0F,$F9,$A9,$B7,$E1,$AE,$18,$77,$AD,$16,$3C,$18,$79,$A7
	.dc.b	$55,$C6,$BB,$1B,$BE,$B8,$78,$A1,$3B,$7A,$9E,$A2,$DF,$57,$11,$F3
	.dc.b	$5F,$F5,$99,$CA,$7E,$DB,$88,$D3,$EA,$36,$78,$3D,$47,$C5,$E2,$30
	.dc.b	$D9,$FF,$59,$36,$77,$3B,$1F,$0F,$5A,$65,$C1,$1F,$66,$4F,$1B,$FD
	.dc.b	$9F,$97,$FA,$5A,$7F,$DA,$95,$F3,$EB,$5D,$5F,$46,$72,$72,$BE,$3A
	.dc.b	$FB,$2F,$84,$62,$7E,$2C,$7F,$B7,$FE,$0F,$FD,$3F,$3F,$ED,$E7,$C1
	.dc.b	$3E,$87,$37,$BB,$E9,$3D,$E3,$C4,$FB,$B1,$DB,$F5,$CF,$2B,$83,$AE
	.dc.b	$1E,$49,$E2,$D9,$E0,$73,$AF,$F7,$46,$1E,$98,$FF,$77,$CE,$EF,$9C
	.dc.b	$78,$CF,$CF,$FC,$9B,$A3,$FD,$5F,$8F,$F8,$2A,$31,$E4,$7F,$02,$8F
	.dc.b	$F8,$01,$1B,$F1,$31,$68,$DC,$F1,$BE,$73,$F8,$E9,$CB,$63,$28,$EA
	.dc.b	$E1,$EF,$93,$E2,$17,$F9,$2B,$FC,$14,$62,$BB,$56,$EB,$4D,$8C,$CF
	.dc.b	$A2,$9E,$C3,$97,$17,$BE,$0E,$95,$A9,$E1,$F8,$E7,$FB,$27,$2F,$8C
	.dc.b	$DC,$97,$C2,$3F,$C5,$13,$AF,$7B,$89,$E7,$AF,$80,$7C,$99,$E4,$E3
	.dc.b	$04,$62,$3E,$8B,$5E,$13,$EB,$F8,$B7,$1F,$6C,$3E,$D4,$61,$FB,$9F
	.dc.b	$2B,$FF,$00,$1B,$1F,$0E,$31,$7E,$A9,$FB,$F0,$9D,$1E,$99,$CB,$5F
	.dc.b	$C1,$46,$E7,$EB,$36,$3F,$05,$F8,$1B,$D8,$FA,$4D,$AF,$19,$F9,$17
	.dc.b	$13,$E3,$3F,$1D,$F8,$DB,$E9,$FE,$D3,$FE,$9D,$AF,$CC,$39,$6F,$8F
	.dc.b	$F7,$FF,$1A,$7B,$DE,$7E,$04,$62,$BF,$18,$DC,$7F,$30,$D7,$EA,$77
	.dc.b	$E7,$8B,$F0,$AD,$AF,$D2,$BF,$89,$6E,$1F,$8F,$F8,$87,$CC,$DF,$47
	.dc.b	$0D,$AF,$D8,$7C,$87,$AF,$2F,$D4,$35,$FC,$23,$5F,$D9,$B5,$FD,$97
	.dc.b	$E3,$4F,$A2,$FC,$F3,$6B,$F7,$8B,$E8,$B3,$3F,$A3,$FC,$6B,$F2,$F7
	.dc.b	$F3,$F8,$1F,$85,$FE,$D7,$1F,$E0,$95,$AE,$0F,$D2,$CE,$6F,$87,$E3
	.dc.b	$4F,$9A,$F9,$CD,$C3,$F7,$0D,$9F,$E0,$8D,$7F,$41,$E6,$BC,$9F,$DB
	.dc.b	$36,$BF,$8A,$3C,$71,$FD,$AB,$9A,$F4,$7E,$34,$FA,$CF,$BC,$DB,$FE
	.dc.b	$10,$D9,$FE,$28,$D7,$F6,$1F,$8D,$3E,$87,$F0,$CD,$AF,$E2,$0D,$9F
	.dc.b	$E3,$8D,$7F,$71,$F3,$55,$FE,$4E,$D7,$F9,$33,$78,$0D,$FC,$69,$F4
	.dc.b	$FF,$95,$6E,$1F,$C7,$DB,$7F,$CA,$DA,$FF,$07,$FC,$53,$F9,$43,$FC
	.dc.b	$CD,$B8,$1A,$63,$F1,$57,$D1,$7E,$8D,$B8,$7F,$2D,$6D,$FF,$39,$6B
	.dc.b	$FE,$08,$FA,$87,$D3,$7F,$42,$84,$62,$BF,$D8,$BF,$68,$F8,$9D,$4E
	.dc.b	$6B,$F7,$87,$0B,$FE,$14,$DB,$FF,$06,$4E,$E3,$C1,$3A,$8E,$5C,$9F
	.dc.b	$E5,$6C,$3F,$A7,$BF,$CB,$63,$37,$FE,$EE,$9C,$BF,$D6,$B1,$FC,$15
	.dc.b	$7F,$3E,$5F,$8D,$7F,$45,$73,$3B,$8C,$72,$65,$B1,$FD,$9E,$78,$B7
	.dc.b	$A3,$91,$CF,$F1,$8B,$19,$26,$9F,$AF,$A4,$EA,$3F,$90,$7C,$8B,$F1
	.dc.b	$DF,$C8,$67,$FE,$E3,$6E,$1F,$DA,$67,$5F,$09,$FD,$D1,$E1,$D1,$6B
	.dc.b	$F2,$51,$8A,$FF,$1D,$B7,$0F,$F2,$1B,$5F,$EA,$AE,$70,$F0,$73,$73
	.dc.b	$C3,$A3,$F1,$FC,$7F,$91,$D9,$71,$F8,$8D,$C7,$DC,$DC,$3E,$83,$F3
	.dc.b	$FD,$49,$5F,$2F,$CF,$B7,$F1,$66,$DF,$D7,$9B,$5F,$DA,$BF,$01,$FE
	.dc.b	$81,$7F,$8C,$6D,$7F,$A6,$BE,$5F,$F2,$55,$FE,$0F,$8F,$EE,$53,$A8
	.dc.b	$E5,$E7,$FE,$4A,$39,$DF,$F9,$99,$AE,$A7,$FC,$C1,$1C,$63,$FC,$03
	.dc.b	$7F,$8E,$7F,$C3,$91,$9B,$4E,$7B,$1C,$6F,$BD,$FC,$81,$F9,$1F,$5E
	.dc.b	$E5,$F8,$D8,$B9,$7E,$1E,$6E,$3F,$E2,$1E,$2F,$F4,$1E,$BF,$81,$F8
	.dc.b	$1F,$F5,$B6,$3E,$F3,$F0,$CA,$35,$C3,$B6,$CE,$F1,$5F,$48,$C3,$FE
	.dc.b	$87,$6D,$FE,$82,$7F,$2E,$DC,$3C,$8F,$F0,$A6,$F4,$79,$4F,$F0,$45
	.dc.b	$FF,$A6,$9B,$3C,$EB,$F2,$2D,$7D,$31,$ED,$FA,$EE,$58,$78,$9F,$DC
	.dc.b	$7D,$0F,$EE,$7E,$32,$CF,$ED,$5B,$7F,$AF,$6E,$5F,$C0,$5F,$E0,$1D
	.dc.b	$FF,$C0,$7D,$9B,$7F,$BD,$6D,$FE,$DD,$B7,$FC,$25,$FF,$26,$78,$7F
	.dc.b	$EE,$1A,$FF,$2B,$36,$BF,$CC,$4D,$7F,$9C,$7C,$2F,$F8,$4E,$35,$7F
	.dc.b	$E5,$0F,$CD,$BF,$9F,$BF,$E5,$28,$CD,$FF,$9C,$A7,$2D,$EC,$FF,$2E
	.dc.b	$47,$F0,$4C,$3F,$CE,$8B,$BB,$33,$2B,$FC,$D5,$3F,$D4,$C9,$1F,$CC
	.dc.b	$BE,$C3,$FB,$A5,$FF,$95,$9A,$FF,$9D,$DA,$9A,$7F,$75,$AF,$F3,$E7
	.dc.b	$C3,$BE,$95,$EB,$7F,$47,$AD,$5F,$FB,$2A,$73,$7F,$E9,$1C,$7F,$C2
	.dc.b	$CF,$9B,$D3,$FF,$1C,$46,$2B,$FD,$18,$D7,$FC,$D0,$E7,$5F,$95,$E9
	.dc.b	$FC,$4A,$F8,$AF,$C6,$FB,$39,$87,$78,$B8,$7F,$88,$70,$56,$30,$E7
	.dc.b	$7E,$58,$FF,$45,$7C,$E7,$FA,$06,$3F,$CF,$0D,$BF,$F5,$3B,$5F,$F5
	.dc.b	$33,$7B,$CF,$FA,$27,$E2,$5C,$DF,$77,$ED,$7F,$D3,$91,$8B,$E3,$F0
	.dc.b	$78,$AD,$63,$FE,$E0,$7F,$D1,$7F,$64,$FA,$A7,$CC,$E7,$0F,$F5,$74
	.dc.b	$E6,$31,$0F,$A4,$F1,$E1,$8A,$D4,$FF,$BD,$B0,$EB,$87,$48,$C1,$D1
	.dc.b	$C6,$5E,$78,$E9,$A7,$9D,$D6,$5D,$6B,$77,$D2,$3C,$79,$78,$AB,$A8
	.dc.b	$C7,$C7,$8F,$56,$DF,$19,$DD,$71,$AD,$61,$DA,$7C,$27,$4B,$E2,$FF
	.dc.b	$B4,$6B,$85,$F3,$BE,$38,$79,$CF,$2E,$3E,$BB,$F2,$63,$D5,$B3,$B7
	.dc.b	$C7,$3B,$C7,$BC,$68,$F1,$57,$1C,$7C,$31,$E6,$C3,$C2,$EB,$0F,$0E
	.dc.b	$5D,$67,$24,$CB,$FD,$97,$7C,$AF,$B4,$E5,$E9,$3A,$CB,$C9,$8F,$A7
	.dc.b	$2F,$29,$E6,$C3,$93,$88,$FC,$05,$FF,$B6,$0F,$51,$63,$F0,$ED,$8D
	.dc.b	$BF,$94,$F7,$FF,$C8,$EF,$13,$EE,$38,$79,$35,$E1,$DB,$B7,$6F,$E5
	.dc.b	$63,$9E,$5F,$83,$6E,$5F,$A2,$E9,$FF,$76,$CE,$6B,$81,$2F,$F0,$53
	.dc.b	$AA,$EF,$1F,$85,$BE,$6F,$EA,$1A,$FE,$4E,$78,$39,$AE,$97,$DE,$3A
	.dc.b	$98,$BE,$B1,$93,$F3,$55,$CA,$7D,$A7,$86,$B8,$3F,$B8,$8D,$63,$FE
	.dc.b	$CA,$9C,$CF,$0A,$E6,$E2,$74,$F3,$3C,$65,$8F,$15,$F9,$0F,$91,$1C
	.dc.b	$70,$F7,$D7,$2A,$EB,$3C,$6B,$83,$F0,$4E,$B1,$D1,$C3,$5F,$90,$7F
	.dc.b	$13,$E0,$FE,$43,$F3,$1F,$29,$72,$3E,$D7,$0D,$AE,$1B,$AF,$83,$63
	.dc.b	$FD,$A4,$7F,$2E,$83,$8B,$EE,$D9,$E0,$B8,$3F,$A4,$FD,$F7,$08,$FE
	.dc.b	$84,$DE,$6F,$EE,$8D,$9F,$EE,$5F,$6F,$6B,$FB,$7C,$DC,$3D,$6D,$C8
	.dc.b	$DA,$C7,$B7,$6F,$FB,$04,$61,$FE,$C5,$1F,$A1,$BF,$8A,$DB,$FF,$64
	.dc.b	$CF,$F7,$A8,$FD,$32,$FB,$47,$3F,$97,$BD,$F6,$3F,$E9,$EA,$FF,$1F
	.dc.b	$46,$3C,$7F,$54,$71,$AB,$7C,$A3,$F6,$4D,$7F,$74,$D7,$F8,$23,$5F
	.dc.b	$F0,$B7,$C3,$79,$C8,$FF,$84,$1A,$FF,$6F,$7E,$31,$F1,$FE,$E7,$E5
	.dc.b	$70,$F9,$E7,$FB,$A7,$D9,$7C,$B3,$38,$C8,$FE,$19,$B7,$F6,$C8,$F9
	.dc.b	$15,$C6,$7F,$BA,$C7,$C9,$F1,$3E,$A7,$D6,$C9,$FF,$08,$47,$D7,$30
	.dc.b	$FF,$80,$E3,$0F,$F8,$57,$B7,$EE,$75,$C8,$C1,$0F,$F0,$53,$97,$FC
	.dc.b	$19,$1F,$87,$F3,$5F,$9D,$3E,$B3,$1F,$35,$FC,$B7,$AB,$FC,$F8,$DE
	.dc.b	$AF,$F5,$F5,$6F,$1F,$F3,$9B,$5F,$F5,$CB,$5F,$F5,$CB,$5E,$37,$FC
	.dc.b	$AF,$0F,$0B,$E3,$9E,$58,$7F,$A2,$DB,$7F,$EC,$DC,$3F,$A0,$CF,$91
	.dc.b	$F3,$FC,$9C,$C7,$FD,$0B,$03,$D6,$65,$FF,$6D,$A8,$CC,$7F,$8F,$FD
	.dc.b	$9F,$D3,$11,$FF,$21,$7C,$85,$7F,$C6,$91,$6F,$FC,$CD,$3F,$98,$FC
	.dc.b	$F5,$BF,$E6,$BF,$3D,$79,$74,$D4,$4C,$7F,$A6,$57,$06,$C7,$EA,$A3
	.dc.b	$11,$FD,$5D,$B8,$7F,$07,$8F,$E1,$B1,$8D,$3F,$84,$36,$3F,$88,$E0
	.dc.b	$F2,$BF,$82,$7A,$0F,$A4,$A4,$AF,$F3,$06,$15,$63,$FE,$81,$9C,$3F
	.dc.b	$E8,$59,$E7,$F3,$FD,$3E,$7E,$5F,$FA,$E7,$A9,$8A,$B7,$39,$85,$70
	.dc.b	$8E,$F7,$FC,$A2,$F8,$DF,$E5,$3E,$77,$F2,$BF,$49,$2F,$EF,$C6,$5F
	.dc.b	$EF,$9C,$FF,$81,$8F,$C7,$72,$9A,$49,$E2,$43,$F0,$BF,$61,$F1,$88
	.dc.b	$F9,$DF,$93,$F3,$78,$8F,$38,$CB,$DF,$39,$DB,$FD,$51,$5F,$42,$FE
	.dc.b	$8E,$53,$0F,$85,$19,$7E,$24,$65,$EB,$87,$CA,$9F,$E4,$BF,$3A,$C7
	.dc.b	$FA,$06,$2F,$AC,$7B,$58,$6E,$3F,$EE,$B1,$7C,$5C,$35,$FF,$12,$35
	.dc.b	$FF,$17,$35,$FE,$A8,$D7,$FA,$EB,$C6,$BF,$29,$1D,$AF,$CE,$78,$F2
	.dc.b	$FE,$A9,$F0,$2F,$A9,$FE,$98,$DA,$FE,$52,$D7,$FC,$7D,$A2,$31,$1F
	.dc.b	$E9,$EF,$7B,$F9,$69,$1F,$E6,$3E,$22,$3F,$AC,$7E,$39,$67,$F9,$AB
	.dc.b	$5E,$AD,$7F,$B8,$5A,$31,$1F,$93,$FC,$72,$BF,$7D,$B2,$69,$8F,$96
	.dc.b	$FE,$41,$3F,$85,$6C,$9A,$63,$F1,$C7,$D4,$FF,$6F,$6D,$7F,$4C,$6B
	.dc.b	$F1,$1A,$FA,$3F,$1C,$FB,$EF,$E2,$1B,$C1,$FC,$1F,$04,$63,$A7,$F5
	.dc.b	$CF,$FA,$C6,$BF,$71,$F3,$8E,$1F,$F0,$03,$63,$FA,$EB,$5F,$AA,$D7
	.dc.b	$E4,$FE,$39,$FE,$1C,$C7,$C0,$FF,$77,$6F,$A3,$FE,$15,$8D,$63,$FB
	.dc.b	$FF,$A0,$FE,$4F,$7F,$C2,$5A,$FE,$89,$B7,$EB,$5F,$07,$FE,$1F,$FB
	.dc.b	$AF,$F3,$C5,$7F,$09,$6B,$FB,$56,$EA,$6F,$DC,$74,$C7,$F8,$83,$D2
	.dc.b	$BF,$8D,$35,$FD,$9B,$B7,$F8,$07,$E3,$9F,$E2,$C7,$F9,$03,$5F,$DC
	.dc.b	$B7,$2F,$A4,$FF,$92,$E3,$F7,$BF,$48,$F8,$FF,$F1,$73,$63,$FC,$16
	.dc.b	$D7,$F4,$6D,$7F,$2F,$F8,$E3,$E4,$7F,$C8,$0F,$E9,$E3,$F9,$93,$5F
	.dc.b	$E1,$6E,$DF,$E6,$DF,$90,$FF,$6E,$7F,$8D,$37,$53,$7F,$1C,$FE,$75
	.dc.b	$3F,$CF,$5A,$FE,$D9,$AF,$EA,$9A,$FF,$3B,$6B,$FC,$9D,$78,$9F,$E5
	.dc.b	$77,$9E,$9F,$97,$6E,$DC,$67,$FC,$B9,$E7,$3D,$0F,$F2,$F7,$EA,$7F
	.dc.b	$57,$0F,$CF,$36,$FF,$A5,$56,$F1,$FE,$A8,$D7,$F8,$6B,$5F,$E0,$2D
	.dc.b	$78,$DF,$F9,$86,$38,$5F,$EA,$DB,$5A,$AE,$BE,$B7,$F9,$C7,$E4,$DE
	.dc.b	$EF,$F9,$CF,$C7,$FF,$67,$63,$FB,$36,$E1,$FD,$7E,$B7,$87,$F6,$96
	.dc.b	$BF,$C8,$9A,$FF,$16,$6B,$C6,$FF,$CE,$DE,$D0,$A4,$8F,$F2,$AB,$96
	.dc.b	$BF,$E8,$73,$0F,$F5,$06,$E1,$FC,$25,$FF,$06,$42,$BF,$31,$3F,$E9
	.dc.b	$EF,$B0,$FF,$18,$4F,$FA,$83,$CD,$FE,$27,$18,$8F,$F1,$A7,$99,$FE
	.dc.b	$9D,$8C,$5F,$29,$46,$39,$FF,$C2,$31,$8B,$FF,$52,$1D,$0B,$7C,$DF
	.dc.b	$F5,$45,$F3,$BE,$58,$79,$67,$77,$4F,$35,$F1,$7F,$D6,$F7,$C6,$79
	.dc.b	$E1,$E2,$C3,$D5,$1E,$1A,$E2,$73,$3E,$24,$EE,$38,$E1,$FE,$C2,$F9
	.dc.b	$3E,$38,$FF,$AF,$5B,$79,$AD,$A8,$F4,$BF,$04,$EE,$53,$E5,$9E,$1C
	.dc.b	$7D,$0F,$AB,$0F,$13,$C3,$1E,$EF,$19,$DB,$E8,$C7,$FD,$A0,$F8,$8B
	.dc.b	$1F,$55,$B5,$F5,$1C,$FC,$D3,$E0,$B8,$C7,$CA,$8C,$38,$A4,$62,$38
	.dc.b	$CF,$FB,$21,$7C,$55,$23,$FE,$CC,$6E,$3F,$8B,$C5,$F3,$5F,$E6,$38
	.dc.b	$FC,$75,$7F,$BC,$2B,$EA,$56,$B8,$3E,$B4,$E4,$F9,$51,$97,$E6,$19
	.dc.b	$AE,$D7,$D7,$5F,$6E,$1F,$5E,$BF,$DF,$B7,$C7,$5F,$0C,$7A,$8F,$B5
	.dc.b	$5F,$EF,$9B,$FF,$3C,$4E,$AF,$CB,$1D,$6B,$E8,$C6,$31,$F3,$C7,$3C
	.dc.b	$7C,$71,$89,$D4,$FA,$0D,$61,$DE,$75,$1E,$EB,$EC,$8F,$85,$87,$47
	.dc.b	$CE,$78,$8A,$7F,$A0,$A3,$D3,$5F,$E7,$5F,$AF,$FF,$03,$5F,$F9,$FD
	.dc.b	$B5,$FE,$58,$7F,$CA,$D1,$9F,$1B,$FD,$31,$9F,$F8,$4E,$77,$5F,$E3
	.dc.b	$06,$BD,$67,$FC,$96,$DA,$FF,$76,$E0,$FA,$7C,$CF,$F9,$C6,$31,$3A
	.dc.b	$9C,$F6,$7E,$19,$B3,$AA,$FF,$36,$F2,$FE,$07,$2F,$F1,$34,$E5,$FE
	.dc.b	$6C,$FA,$30,$FF,$4A,$7E,$3F,$F2,$5F,$F4,$F3,$6F,$FC,$EE,$BE,$53
	.dc.b	$84,$7F,$A1,$63,$57,$FE,$33,$78,$47,$F9,$41,$AF,$63,$D2,$E1,$E4
	.dc.b	$74,$CB,$FD,$B5,$E8,$59,$EB,$87,$FA,$A7,$E0,$7F,$CC,$4F,$FA,$CF
	.dc.b	$E6,$9E,$27,$16,$C6,$7E,$C3,$13,$23,$F7,$78,$AB,$79,$7F,$9D,$1A
	.dc.b	$FF,$9A,$1A,$FF,$AE,$7E,$BF,$FD,$81,$5C,$9B,$1F,$E7,$C6,$BF,$E7
	.dc.b	$06,$BF,$EC,$1F,$AF,$77,$FD,$A6,$D7,$91,$B3,$FE,$86,$6B,$FE,$7A
	.dc.b	$FA,$F6,$7F,$82,$6D,$7B,$1B,$3F,$E8,$E6,$BF,$E8,$2E,$DF,$C3,$37
	.dc.b	$93,$F1,$71,$88,$FF,$6B,$35,$FF,$61,$78,$0F,$07,$F2,$2D,$C3,$E5
	.dc.b	$B6,$7F,$D3,$AD,$7F,$98,$FC,$BB,$E2,$BF,$3B,$08,$C5,$7F,$2E,$8C
	.dc.b	$4E,$A7,$31,$FE,$A8,$FA,$F7,$CC,$FE,$89,$B7,$F7,$1B,$3B,$9F,$D3
	.dc.b	$FA,$9F,$99,$AF,$F3,$FF,$4E,$08,$F8,$70,$FF,$07,$37,$C6,$F1,$11
	.dc.b	$FC,$4E,$CF,$A5,$FD,$9B,$63,$31,$83,$F6,$FE,$CB,$E3,$A9,$27,$F2
	.dc.b	$71,$97,$E1,$7E,$09,$FE,$43,$6F,$97,$F2,$23,$1B,$7F,$54,$6B,$FE
	.dc.b	$BE,$F9,$47,$D0,$7F,$06,$6D,$FE,$81,$B3,$FE,$D2,$6B,$FE,$C6,$F8
	.dc.b	$0F,$A9,$3F,$ED,$0D,$BF,$87,$C6,$27,$FD,$B2,$D7,$FB,$6E,$1F,$96
	.dc.b	$7F,$99,$07,$0F,$F3,$56,$B9,$EA,$FE,$37,$F1,$2F,$B2,$A4,$C7,$F6
	.dc.b	$F1,$97,$F0,$DF,$85,$7F,$BA,$5F,$EA,$E7,$81,$C5,$BC,$5F,$E6,$BF
	.dc.b	$0D,$FE,$93,$AD,$CF,$F4,$7A,$DD,$7F,$BF,$9E,$31,$FE,$08,$6B,$D7
	.dc.b	$E1,$7F,$33,$18,$8E,$51,$FC,$92,$57,$FE,$04,$BE,$F4,$BE,$F1,$FD
	.dc.b	$BA,$75,$3E,$78,$FE,$B5,$E1,$B1,$7D,$BF,$F2,$56,$CF,$AA,$75,$F5
	.dc.b	$D9,$A9,$39,$AD,$DF,$51,$7C,$27,$55,$FC,$C7,$EC,$7F,$CF,$B1,$D5
	.dc.b	$7C,$C6,$7C,$0D,$36,$74,$DC,$3B,$60,$BE,$35,$F4,$3E,$37,$FD,$79
	.dc.b	$F5,$5F,$D2,$9B,$19,$FC,$D7,$F1,$DC,$08,$FE,$F9,$3A,$C7,$D4,$D7
	.dc.b	$F1,$ED,$77,$F0,$0C,$3F,$B5,$CE,$64,$F3,$BE,$D1,$8C,$3F,$CA,$D8
	.dc.b	$79,$F0,$E4,$F9,$7D,$5F,$44,$FF,$52,$AF,$1B,$6F,$9E,$08,$7D,$17
	.dc.b	$0F,$46,$63,$EF,$79,$C6,$30,$F9,$3C,$3F,$DD,$EB,$F0,$6B,$A7,$F0
	.dc.b	$5E,$0F,$E0,$DB,$3F,$7D,$B5,$C9,$FE,$EE,$FF,$79,$6F,$87,$CE,$7F
	.dc.b	$47,$A3,$EA,$BF,$93,$5F,$E1,$CE,$13,$F9,$AA,$E1,$1F,$E6,$46,$BF
	.dc.b	$E6,$E6,$BF,$C0,$9A,$FF,$8D,$5E,$35,$FE,$28,$A7,$BB,$E4,$C3,$F5
	.dc.b	$51,$8C,$3F,$35,$E8,$FE,$87,$CE,$FD,$13,$63,$27,$E7,$E3,$2D,$3F
	.dc.b	$21,$F7,$3F,$E0,$AC,$FF,$02,$DE,$13,$F0,$33,$FB,$2E,$49,$3F,$97
	.dc.b	$75,$5F,$94,$7A,$43,$EB,$3F,$BC,$D7,$E3,$23,$1E,$07,$F0,$5F,$D3
	.dc.b	$FF,$C9,$73,$9A,$DE,$1F,$CA,$1F,$E4,$A4,$AF,$EC,$ED,$8F,$AF,$F9
	.dc.b	$E3,$E9,$3E,$EC,$F2,$C7,$F9,$04,$E6,$BF,$75,$E8,$FE,$F4,$FD,$F7
	.dc.b	$D0,$D8,$C9,$2B,$F7,$3D,$89,$D4,$7E,$53,$DE,$FE,$24,$7F,$13,$F6
	.dc.b	$6D,$FF,$0F,$8B,$3F,$9A,$9B,$AF,$EB,$BB,$FF,$8E,$7C,$97,$F4,$02
	.dc.b	$61,$FC,$F9,$B7,$FA,$2F,$95,$FE,$41,$1F,$C8,$5F,$E4,$5D,$12,$BF
	.dc.b	$8F,$7D,$4D,$BF,$D4,$CF,$F4,$C2,$4F,$F4,$A7,$FC,$23,$F5,$9F,$CE
	.dc.b	$64,$D2,$79,$5A,$3F,$52,$FF,$91,$EB,$F9,$DB,$FC,$EA,$7F,$D2,$0E
	.dc.b	$5F,$DC,$F6,$7F,$5C,$24,$FF,$5A,$6C,$7E,$F8,$FE,$78,$FF,$3D,$7F
	.dc.b	$CC,$5D,$8F,$A3,$A6,$6B,$FA,$04,$65,$B1,$FD,$D8,$8F,$F9,$7A,$35
	.dc.b	$3F,$EE,$2D,$5F,$9D,$3C,$AF,$E4,$37,$FD,$E2,$73,$18,$AF,$EA,$75
	.dc.b	$BA,$FF,$3E,$D6,$9B,$19,$9F,$F6,$FF,$85,$FD,$6B,$07,$CA,$FF,$2E
	.dc.b	$F8,$6F,$E4,$75,$FD,$4A,$F8,$46,$31,$FE,$3D,$5B,$C3,$FC,$17,$F5
	.dc.b	$7F,$E6,$8D,$7F,$DC,$51,$8E,$DF,$F4,$A6,$5E,$03,$FD,$85,$C5,$FC
	.dc.b	$5B,$2E,$F0,$C3,$93,$B3,$15,$CA,$77,$3A,$78,$D7,$3B,$F0,$61,$E2
	.dc.b	$C3,$FD,$A1,$7B,$9C,$C7,$5A,$E0,$EB,$0E,$D3,$AC,$3B,$C6,$27,$D3
	.dc.b	$A7,$C1,$78,$63,$DD,$F3,$B7,$1F,$04,$70,$D7,$17,$CE,$36,$7B,$E7
	.dc.b	$6B,$2F,$2B,$AC,$7C,$78,$72,$3C,$13,$AA,$E2,$4B,$ED,$19,$7C,$11
	.dc.b	$99,$E3,$8F,$AF,$0E,$D9,$7C,$12,$3E,$93,$CD,$7C,$89,$5C,$B2,$F2
	.dc.b	$4E,$A2,$E5,$F5,$B0,$ED,$8F,$86,$BF,$CA,$67,$CA,$29,$F1,$0C,$BF
	.dc.b	$84,$6F,$BB,$FE,$95,$77,$1F,$E9,$9F,$4F,$FD,$2C,$D8,$CC,$ED,$E3
	.dc.b	$3F,$56,$39,$5F,$E3,$1B,$9F,$E7,$3D,$CF,$F3,$3C,$E6,$B8,$12,$FE
	.dc.b	$C4,$EA,$BF,$47,$7C,$A7,$F2,$2D,$7F,$3C,$DC,$3F,$D4,$F1,$CF,$0E
	.dc.b	$CF,$E0,$E7,$57,$D6,$39,$E1,$BC,$3B,$BF,$8F,$9D,$4F,$68,$D3,$C0
	.dc.b	$F3,$98,$34,$77,$79,$39,$C3,$E5,$61,$E4,$3F,$63,$5D,$A3,$37,$E2
	.dc.b	$AE,$47,$62,$D7,$18,$D3,$CA,$3B,$61,$F0,$CF,$7B,$8C,$3C,$58,$7C
	.dc.b	$1D,$FF,$92,$E5,$47,$D5,$53,$5C,$3F,$E3,$3C,$3F,$81,$57,$07,$F9
	.dc.b	$1D,$7F,$A8,$B6,$FE,$59,$5B,$9F,$F0,$25,$6D,$B5,$FE,$31,$7F,$9C
	.dc.b	$78,$FF,$EE,$39,$FF,$20,$36,$75,$3F,$EB,$28,$FF,$57,$F5,$7F,$6D
	.dc.b	$9C,$B6,$33,$87,$F3,$3F,$C6,$AD,$FF,$A0,$AD,$F5,$F8,$C7,$FC,$9D
	.dc.b	$56,$4D,$99,$5F,$D3,$5F,$21,$1F,$F7,$D7,$E1,$3F,$C3,$5F,$F5,$0B
	.dc.b	$FC,$D9,$FF,$26,$37,$2F,$F1,$33,$B8,$CF,$2F,$ED,$90,$8D,$3F,$BC
	.dc.b	$6B,$FE,$D0,$6B,$ED,$F7,$7E,$41,$1F,$EC,$8D,$7F,$D2,$F8,$FF,$BD
	.dc.b	$5F,$ED,$B1,$F0,$7E,$DF,$E1,$7B,$FF,$0F,$ED,$65,$7F,$72,$8C,$BF
	.dc.b	$DB,$DF,$AC,$47,$FB,$97,$C8,$7C,$B2,$3F,$DE,$BE,$D6,$BE,$28,$BA
	.dc.b	$7D,$B9,$FA,$83,$D1,$A4,$AF,$F6,$14,$7E,$36,$7F,$C5,$CE,$E3,$33
	.dc.b	$B7,$FD,$6B,$1B,$E1,$FF,$7B,$FD,$4B,$A3,$F0,$B8,$3F,$19,$FF,$3A
	.dc.b	$E1,$FE,$86,$6B,$FE,$85,$AE,$0F,$FB,$79,$AF,$F2,$E6,$BF,$CC,$9A
	.dc.b	$F8,$2B,$EF,$28,$E3,$5F,$E6,$96,$DE,$E7,$FD,$2F,$FA,$73,$D3,$FC
	.dc.b	$DF,$A6,$FD,$01,$F5,$B8,$6D,$FF,$B4,$5A,$FF,$B2,$97,$D0,$6E,$64
	.dc.b	$ED,$E5,$1F,$A3,$F9,$0F,$D2,$7C,$6C,$AF,$F1,$9C,$65,$FF,$1A,$C6
	.dc.b	$5F,$C6,$C7,$EC,$1F,$77,$81,$7E,$B7,$FC,$DD,$1F,$9E,$DF,$FD,$11
	.dc.b	$18,$8F,$9D,$3C,$77,$FE,$77,$1F,$36,$3E,$4C,$62,$3E,$74,$60,$E0
	.dc.b	$47,$B6,$FF,$AC,$54,$F5,$3F,$80,$79,$BF,$C0,$7E,$4E,$57,$F9,$42
	.dc.b	$32,$FF,$94,$A3,$2F,$EA,$A3,$E9,$FE,$69,$FC,$22,$3F,$E0,$03,$5F
	.dc.b	$F7,$CB,$73,$FF,$73,$BF,$C6,$31,$FF,$7F,$B9,$6F,$83,$D0,$EB,$E1
	.dc.b	$F6,$F1,$7B,$F9,$93,$1E,$F1,$97,$FB,$FF,$2B,$1F,$E7,$71,$86,$BF
	.dc.b	$E0,$68,$D4,$FF,$11,$6C,$74,$21,$EA,$F3,$BF,$6B,$EE,$7E,$DB,$A2
	.dc.b	$4F,$ED,$23,$2F,$ED,$A3,$2F,$EC,$A3,$F9,$77,$97,$23,$FC,$FA,$FF
	.dc.b	$31,$F8,$1C,$47,$F1,$68,$BB,$FF,$3F,$9F,$C6,$7E,$11,$FC,$A6,$3F
	.dc.b	$DA,$1F,$2D,$FA,$5B,$FB,$E7,$F2,$A7,$0F,$FB,$F6,$78,$70,$FE,$0D
	.dc.b	$FE,$53,$06,$93,$C9,$FD,$BB,$C2,$3F,$9A,$B5,$E1,$1F,$9F,$A6,$5F
	.dc.b	$D2,$FE,$07,$FA,$D1,$2B,$FD,$2C,$FF,$02,$EE,$FE,$0B,$1F,$D0,$1A
	.dc.b	$FE,$F7,$C0,$D1,$FD,$97,$EA,$5F,$09,$49,$3F,$E3,$EF,$39,$3A,$8C
	.dc.b	$7D,$4D,$7F,$D4,$8D,$7F,$D4,$4D,$7F,$A8,$CA,$31,$1E,$5F,$92,$F4
	.dc.b	$11,$FE,$B1,$8A,$3F,$BD,$7D,$4F,$F5,$4B,$34,$9B,$7F,$7F,$FA,$9F
	.dc.b	$EC,$D8,$7F,$AC,$DA,$F8,$9A,$A9,$18,$87,$CE,$3E,$53,$F9,$6B,$63
	.dc.b	$FB,$63,$5F,$F3,$1B,$5F,$F6,$17,$D4,$FF,$83,$9B,$DD,$FE,$13,$8C
	.dc.b	$47,$C4,$6B,$E3,$FA,$97,$DB,$FF,$86,$5B,$1F,$4D,$AF,$FB,$4D,$AF
	.dc.b	$F9,$DA,$DE,$EF,$A0,$91,$E8,$7F,$C4,$9F,$53,$C2,$3E,$AB,$5F,$88
	.dc.b	$DD,$0D,$FA,$97,$C5,$FF,$8B,$5B,$5F,$87,$6B,$FE,$E7,$6B,$FE,$E5
	.dc.b	$E8,$79,$9F,$E3,$56,$C7,$E3,$DA,$FF,$BB,$DA,$FF,$BB,$7E,$A7,$FD
	.dc.b	$A6,$DF,$03,$FC,$7D,$18,$8F,$C7,$B5,$FB,$DF,$53,$FE,$EA,$7F,$38
	.dc.b	$D7,$FD,$F8,$D7,$FD,$F4,$D7,$F3,$3F,$52,$FA,$4F,$F2,$63,$63,$F4
	.dc.b	$CD,$7F,$D4,$FA,$A3,$11,$FD,$1E,$31,$3F,$E4,$79,$E1,$E1,$3C,$2E
	.dc.b	$F3,$9F,$98,$FF,$26,$79,$D9,$C7,$F8,$9E,$2F,$A8,$FE,$E7,$8E,$63
	.dc.b	$1E,$AF,$0C,$A5,$FE,$EF,$0E,$4F,$99,$AF,$95,$AF,$EE,$1D,$DF,$56
	.dc.b	$CE,$4F,$2F,$D8,$FF,$98,$7E,$A7,$FB,$85,$78,$FF,$09,$7C,$CF,$F9
	.dc.b	$95,$B8,$FA,$AB,$78,$7F,$06,$7F,$BC,$47,$C7,$6B,$9F,$92,$FF,$3E
	.dc.b	$46,$23,$F7,$5E,$1F,$E8,$1F,$E3,$0D,$7E,$9B,$5F,$A4,$D7,$F8,$AB
	.dc.b	$BC,$7E,$53,$6F,$3D,$5F,$E7,$D8,$FF,$15,$D7,$F9,$B2,$DE,$CB,$0F
	.dc.b	$F3,$7F,$90,$F9,$2F,$F0,$31,$FE,$40,$71,$7F,$E7,$48,$C1,$FE,$0B
	.dc.b	$BE,$75,$CA,$31,$7D,$69,$F5,$3C,$9E,$DF,$21,$FE,$13,$8C,$57,$F9
	.dc.b	$39,$D6,$1D,$2B,$8C,$F4,$C3,$FD,$0B,$5C,$2F,$94,$FA,$1B,$5E,$FC
	.dc.b	$3C,$55,$D1,$B5,$B9,$E6,$D7,$B9,$C6,$F2,$D3,$FA,$8C,$6F,$1F,$14
	.dc.b	$7C,$13,$4F,$79,$E3,$1F,$00,$B3,$EF,$3D,$8F,$80,$63,$F0,$70,$EB
	.dc.b	$19,$9E,$0F,$8E,$3B,$3F,$E3,$D3,$CC,$59,$F9,$4D,$C3,$EA,$39,$F8
	.dc.b	$27,$D5,$7F,$AC,$69,$18,$9E,$87,$91,$C3,$C6,$7F,$CF,$6B,$E3,$69
	.dc.b	$1F,$F3,$FB,$70,$FC,$16,$4F,$73,$F8,$34,$7E,$16,$BF,$D8,$35,$F2
	.dc.b	$2B,$5C,$1F,$2A,$72,$7C,$78,$CB,$F2,$4C,$D7,$6B,$EB,$A7,$AB,$0F
	.dc.b	$AD,$5F,$A1,$9F,$E2,$DA,$F8,$63,$D4,$7D,$E3,$FC,$ED,$1A,$AE,$51
	.dc.b	$E1,$8E,$AF,$D1,$AF,$44,$62,$3F,$CE,$11,$8C,$7D,$51,$DB,$0F,$34
	.dc.b	$62,$75,$3E,$B3,$57,$FF,$00,$2F,$A4,$7B,$A3,$15,$C6,$3C,$8F,$E6
	.dc.b	$39,$FF,$C0,$BF,$57,$FC,$CF,$3F,$E6,$1E,$2A,$31,$5F,$E2,$08,$7D
	.dc.b	$7E,$30,$FF,$08,$46,$E7,$F9,$DB,$63,$FA,$4B,$5F,$E9,$0D,$7A,$57
	.dc.b	$F3,$6C,$3F,$01,$F1,$77,$33,$49,$87,$F8,$EE,$32,$FF,$4A,$F5,$7F
	.dc.b	$B9,$BF,$D3,$23,$38,$7F,$1B,$8F,$F1,$CC,$EA,$30,$7F,$8A,$7E,$B1
	.dc.b	$CB,$ED,$C5,$18,$BF,$EA,$EB,$D0,$D3,$9B,$E7,$5C,$2F,$D9,$39,$9F
	.dc.b	$ED,$0D,$7F,$C9,$6D,$7C,$51,$E1,$F9,$77,$8D,$DA,$30,$F9,$7C,$87
	.dc.b	$C4,$DD,$B0,$DD,$3F,$B6,$CF,$C2,$C7,$FB,$C3,$5F,$F2,$DB,$5F,$F2
	.dc.b	$53,$73,$7D,$63,$EC,$3E,$23,$6F,$AE,$38,$9F,$91,$F5,$AF,$6A,$98
	.dc.b	$EF,$B7,$FE,$00,$FA,$DF,$4B,$FE,$75,$6E,$9F,$E1,$06,$BF,$E0,$F6
	.dc.b	$BF,$E7,$2F,$03,$EE,$F4,$FF,$81,$D7,$D2,$FD,$97,$37,$B8,$FB,$9F
	.dc.b	$4A,$F1,$29,$86,$CF,$F8,$53,$D4,$FF,$28,$BF,$E5,$7F,$17,$F2,$D1
	.dc.b	$8C,$BF,$CB,$11,$F9,$9F,$AD,$7D,$5F,$E2,$5F,$F0,$3D,$FF,$87,$9C
	.dc.b	$B7,$C9,$FC,$77,$D6,$FE,$27,$6F,$F1,$03,$67,$FD,$1E,$FE,$3D,$FF
	.dc.b	$13,$B9,$C7,$FD,$01,$F7,$EF,$94,$FD,$53,$6B,$FD,$40,$D9,$FF,$1D
	.dc.b	$B5,$FF,$1D,$75,$3E,$57,$F4,$0B,$E5,$EC,$61,$FD,$27,$D7,$3E,$63
	.dc.b	$F7,$0D,$C7,$FD,$5C,$DC,$3F,$C9,$2D,$7F,$C9,$1F,$47,$F0,$9F,$F2
	.dc.b	$6B,$5C,$F0,$7F,$03,$8C,$47,$FA,$C3,$EB,$BE,$5B,$FE,$54,$6B,$FE
	.dc.b	$53,$6B,$BE,$0F,$E1,$9D,$BF,$EC,$4C,$7F,$D8,$4D,$9F,$F2,$EB,$5F
	.dc.b	$F2,$E3,$5F,$F5,$FA,$F9,$EE,$64,$9E,$6F,$F1,$9F,$BF,$F0,$D7,$FA
	.dc.b	$CD,$BE,$09,$A6,$A1,$F8,$DF,$B3,$97,$FC,$B6,$DC,$13,$FE,$33,$FC
	.dc.b	$57,$FD,$95,$1F,$EF,$A8,$E3,$9F,$F3,$36,$C7,$F7,$46,$9F,$DD,$AB
	.dc.b	$C0,$49,$FE,$75,$6B,$7B,$1C,$EB,$C5,$7E,$18,$FE,$49,$8B,$E6,$E9
	.dc.b	$8E,$0B,$9F,$F9,$F2,$74,$DF,$9A,$CE,$1F,$C9,$A5,$1D,$3E,$DA,$57
	.dc.b	$1F,$72,$DE,$E7,$C3,$F0,$F8,$DB,$FD,$0D,$8F,$FB,$3B,$0F,$8F,$1F
	.dc.b	$E8,$DF,$95,$7C,$D3,$DF,$FE,$B1,$18,$6C,$7F,$B9,$6B,$89,$F4,$27
	.dc.b	$4E,$32,$46,$1F,$F1,$63,$5F,$AE,$6A,$7C,$13,$FE,$26,$9D,$4F,$59
	.dc.b	$CB,$FD,$4A,$3C,$FE,$0F,$92,$FC,$AF,$D0,$F8,$89,$87,$CD,$FB,$CB
	.dc.b	$87,$CB,$9F,$AB,$F7,$98,$CF,$E2,$BF,$63,$C6,$98,$7D,$5E,$BB,$87
	.dc.b	$D3,$9F,$BB,$F6,$38,$C3,$E6,$BF,$8D,$FB,$27,$CD,$53,$F1,$3F,$A0
	.dc.b	$BE,$93,$19,$FE,$56,$B6,$FE,$4D,$B9,$7F,$92,$1A,$FF,$95,$5A,$FC
	.dc.b	$A7,$78,$FF,$85,$6F,$F8,$D3,$FE,$C0,$8C,$D7,$E0,$27,$2F,$EA,$3B
	.dc.b	$BF,$61,$E8,$7E,$60,$FC,$CF,$9B,$67,$F2,$D1,$F9,$CE,$DC,$65,$F9
	.dc.b	$57,$F0,$BC,$32,$BF,$05,$CB,$F8,$CE,$2C,$5F,$D9,$E6,$5F,$E3,$39
	.dc.b	$33,$BF,$E4,$E5,$1F,$90,$EB,$99,$74,$9F,$E4,$95,$B9,$CE,$1F,$8E
	.dc.b	$9F,$D3,$D7,$E2,$DB,$19,$9F,$F6,$FE,$5F,$A0,$7F,$3B,$E0,$72,$C3
	.dc.b	$C4,$FE,$EB,$E4,$E6,$3F,$B7,$F4,$AE,$3F,$B5,$9F,$DF,$F4,$E3,$0F
	.dc.b	$D7,$BF,$E2,$5F,$A7,$7D,$CD,$3F,$82,$F6,$5C,$BF,$DF,$B3,$A7,$F8
	.dc.b	$21,$9C,$7F,$81,$3C,$FE,$0B,$FC,$85,$3F,$E4,$E8,$FF,$46,$F3,$23
	.dc.b	$15,$FE,$9B,$8C,$57,$03,$C9,$EB,$F3,$BF,$0B,$E7,$F8,$C9,$1E,$4F
	.dc.b	$82,$B8,$7F,$8E,$31,$57,$88,$95,$E2,$8C,$BF,$CB,$7E,$B1,$F3,$C8
	.dc.b	$C3,$6F,$FD,$4B,$3A,$9C,$92,$3F,$CA,$3F,$14,$FB,$EF,$E0,$7C,$3C
	.dc.b	$65,$7C,$5D,$CE,$AF,$F3,$D8,$BE,$43,$E4,$92,$3D,$0D,$AF,$77,$DB
	.dc.b	$7D,$98,$FB,$4C,$9F,$AE,$E5,$A4,$9F,$8D,$F7,$5B,$6C,$7D,$EF,$77
	.dc.b	$E0,$3F,$C8,$B8,$BE,$C4,$62,$7C,$ED,$51,$A9,$FE,$BE,$71,$C7,$FC
	.dc.b	$76,$FF,$B0,$3E,$0F,$DF,$F5,$FF,$37,$E3,$7F,$CC,$B8,$72,$BD,$BF
	.dc.b	$EC,$FA,$DD,$F1,$8F,$ED,$7F,$5C,$E9,$7D,$7E,$65,$39,$8C,$7C,$5F
	.dc.b	$FA,$2B,$2E,$F5,$AB,$E2,$F7,$3A,$B8,$8D,$DF,$39,$D1,$FE,$89,$F6
	.dc.b	$3D,$59,$7F,$48,$9D,$57,$0B,$E5,$39,$6D,$FF,$81,$30,$F2,$4E,$71
	.dc.b	$F5,$61,$CA,$31,$5E,$0A,$DB,$E6,$AD,$DF,$48,$F6,$47,$C0,$C7,$CB
	.dc.b	$1B,$3D,$38,$75,$71,$8F,$C4,$C3,$AC,$72,$37,$87,$5A,$E1,$1F,$07
	.dc.b	$2F,$0E,$32,$36,$F8,$63,$18,$7C,$38,$B9,$7F,$6E,$C7,$BE,$1D,$09
	.dc.b	$8F,$AE,$32,$F8,$49,$5D,$71,$EE,$4C,$3B,$3F,$1F,$2F,$82,$FA,$E3
	.dc.b	$2F,$F7,$E3,$DC,$5C,$3E,$F3,$6F,$F0,$AE,$7B,$BF,$C9,$4F,$13,$F1
	.dc.b	$EE,$1D,$7B,$1F,$E3,$9A,$FF,$55,$36,$78,$DF,$E3,$1B,$9F,$E2,$5B
	.dc.b	$7F,$9E,$F0,$7F,$C2,$75,$BB,$FC,$2D,$6A,$BC,$31,$F8,$8A,$DD,$F3
	.dc.b	$8D,$92,$7F,$03,$3A,$7F,$57,$5C,$F0,$F3,$C7,$E1,$A3,$13,$B8,$EE
	.dc.b	$66,$BA,$61,$FA,$69,$D5,$F2,$8F,$45,$6C,$F0,$9C,$30,$F3,$39,$8E
	.dc.b	$4E,$4F,$2D,$D7,$F7,$B7,$DB,$0F,$41,$27,$E1,$4E,$F0,$F1,$61,$DB
	.dc.b	$0E,$B3,$DE,$30,$47,$E8,$1C,$6F,$A3,$E8,$3F,$83,$70,$7F,$5B,$F7
	.dc.b	$BE,$53,$FD,$77,$4B,$C0,$6C,$CA,$FE,$4A,$7F,$40,$7F,$C9,$FC,$1F
	.dc.b	$C9,$DB,$19,$36,$7F,$25,$F9,$9F,$F2,$AF,$4F,$F5,$5C,$7F,$B6,$7D
	.dc.b	$2E,$30,$FE,$96,$DB,$E0,$7F,$7A,$BC,$57,$F7,$36,$E4,$6F,$C6,$4A
	.dc.b	$E0,$43,$FB,$73,$97,$FB,$D4,$A3,$FD,$31,$5F,$DF,$DB,$19,$9F,$F4
	.dc.b	$E7,$D7,$3E,$C7,$CD,$DC,$FB,$D7,$D0,$E1,$B7,$FE,$A7,$BE,$0D,$BF
	.dc.b	$F5,$32,$8C,$BF,$B5,$6B,$FB,$F6,$BF,$DC,$3E,$CB,$CA,$47,$FC,$0C
	.dc.b	$D7,$FC,$18,$FC,$63,$E3,$FC,$B7,$C8,$EB,$F9,$37,$2B,$FC,$4F,$19
	.dc.b	$7F,$C2,$0F,$C2,$23,$FE,$13,$3E,$8F,$C9,$47,$FC,$5E,$7D,$42,$18
	.dc.b	$3F,$C6,$7D,$9F,$6B,$E9,$DF,$55,$7D,$A9,$5F,$E9,$FE,$25,$6A,$30
	.dc.b	$D5,$CD,$F8,$B8,$C3,$6F,$B7,$D6,$7E,$2F,$D1,$7D,$7B,$E1,$30,$BD
	.dc.b	$3A,$62,$26,$5F,$EA,$99,$E0,$D8,$FF,$6F,$57,$46,$CE,$A7,$FD,$58
	.dc.b	$DC,$7F,$D8,$8D,$EB,$FF,$5F,$7E,$6B,$FC,$C6,$7F,$D9,$0F,$FB,$22
	.dc.b	$BF,$CF,$EB,$E6,$BD,$4D,$8C,$FA,$D8,$23,$83,$8F,$CC,$7C,$FF,$B3
	.dc.b	$FA,$1D,$72,$7E,$5C,$65,$A4,$3E,$6E,$5F,$E9,$BE,$17,$D8,$7E,$D6
	.dc.b	$59,$3E,$83,$1F,$3B,$FE,$74,$9F,$F6,$C0,$70,$F4,$07,$28,$E9,$3C
	.dc.b	$E3,$FA,$B1,$92,$B9,$7F,$AD,$35,$FE,$C8,$E7,$6F,$F2,$D3,$71,$FF
	.dc.b	$72,$B6,$7C,$36,$F5,$FF,$29,$F2,$7F,$95,$F9,$29,$5F,$90,$8C,$B4
	.dc.b	$87,$AF,$A3,$F2,$FF,$27,$F9,$DD,$3F,$3D,$F2,$B2,$BF,$33,$19,$69
	.dc.b	$0E,$5B,$FF,$B6,$DF,$C0,$F0,$7E,$03,$82,$46,$08,$7F,$80,$F9,$FE
	.dc.b	$9C,$FE,$53,$73,$49,$3C,$38,$7F,$1F,$1F,$8C,$E2,$FD,$A7,$2F,$ED
	.dc.b	$7E,$66,$5F,$EC,$23,$2D,$21,$E8,$E5,$FD,$E7,$9B,$FB,$DE,$79,$3F
	.dc.b	$B8,$8C,$B4,$87,$C0,$F6,$BF,$19,$F4,$5F,$EF,$E8,$FC,$2E,$9F,$E9
	.dc.b	$B3,$F8,$3F,$B1,$FD,$15,$B1,$F8,$E9,$E0,$FE,$34,$38,$7F,$C5,$0D
	.dc.b	$7F,$22,$6B,$C0,$FE,$67,$19,$8F,$CA,$BD,$27,$FD,$48,$DE,$BE,$2D
	.dc.b	$9F,$E2,$7E,$03,$DD,$FE,$3D,$F5,$DF,$D3,$39,$8D,$25,$7F,$B3,$A1
	.dc.b	$F4,$5B,$24,$65,$FE,$AB,$BF,$F4,$28,$FF,$4C,$B7,$AC,$D3,$1F,$5D
	.dc.b	$FE,$3D,$9F,$F5,$23,$5F,$D8,$CA,$31,$1F,$EA,$46,$C3,$E6,$5E,$2F
	.dc.b	$F6,$66,$D7,$FB,$99,$AF,$FB,$95,$AF,$FA,$CF,$EB,$5E,$47,$F6,$E6
	.dc.b	$C7,$FB,$B9,$AF,$FB,$B5,$AF,$FA,$EF,$E6,$FE,$76,$0F,$AD,$FF,$31
	.dc.b	$71,$FF,$82,$DB,$C7,$EC,$AD,$FC,$EB,$E0,$7F,$BE,$B7,$1F,$F7,$EB
	.dc.b	$5F,$F7,$E3,$5F,$F6,$57,$D6,$FE,$A9,$FE,$A8,$D7,$F8,$03,$71,$FF
	.dc.b	$0E,$3F,$C9,$A3,$FA,$D7,$D0,$3E,$33,$FC,$32,$BE,$6A,$98,$8F,$F0
	.dc.b	$EF,$CB,$FF,$53,$AF,$F3,$B6,$C8,$C4,$7F,$A0,$63,$1E,$87,$F8,$62
	.dc.b	$73,$3F,$EE,$1F,$AD,$F0,$3F,$DA,$DA,$FF,$17,$6E,$7F,$E3,$37,$C9
	.dc.b	$1F,$DC,$7E,$B5,$F3,$3F,$E3,$05,$F3,$54,$C4,$7F,$8D,$3D,$FF,$EA
	.dc.b	$95,$E5,$6B,$FE,$F3,$6B,$F9,$06,$BF,$EF,$2F,$AD,$FD,$0B,$FE,$01
	.dc.b	$6B,$FC,$B5,$B8,$FF,$92,$5F,$D5,$C7,$F8,$1F,$EB,$7F,$C0,$3C,$BF
	.dc.b	$9C,$6C,$7F,$BF,$9B,$B3,$EB,$7F,$01,$87,$F9,$4D,$FC,$05,$7E,$91
	.dc.b	$AA,$78,$72,$FF,$91,$A7,$31,$CF,$C5,$FF,$2A,$C7,$4C,$79,$56,$27
	.dc.b	$FC,$B3,$F6,$3F,$D3,$F0,$FF,$0F,$7D,$73,$E8,$7F,$CB,$AD,$8F,$F1
	.dc.b	$14,$ED,$FE,$9C,$D7,$F8,$CE,$48,$C4,$7E,$F1,$AF,$A5,$B1,$BF,$63
	.dc.b	$C6,$49,$EF,$F2,$FF,$E6,$EF,$AE,$7C,$E7,$F9,$C3,$CB,$FF,$0C,$57
	.dc.b	$C3,$6E,$0A,$DC,$FF,$8E,$5A,$FE,$5D,$AF,$A1,$AF,$F1,$26,$BF,$3F
	.dc.b	$E7,$5F,$35,$FE,$74,$6C,$7E,$91,$AF,$E8,$DA,$FC,$47,$3C,$5F,$E8
	.dc.b	$F6,$F0,$FF,$20,$6C,$F7,$F9,$BF,$F0,$E6,$0F,$53,$FC,$9B,$2F,$AE
	.dc.b	$7C,$37,$F4,$83,$F8,$BB,$88,$FF,$0D,$C6,$0F,$E9,$B5,$CF,$2F,$F0
	.dc.b	$FD,$6E,$BE,$45,$F6,$8C,$5F,$19,$E7,$E4,$7F,$56,$9E,$18,$73,$9F
	.dc.b	$25,$76,$BE,$4F,$96,$F8,$36,$7C,$2E,$F0,$B1,$FE,$32,$BE,$98,$78
	.dc.b	$AF,$8C,$F8,$8D,$9E,$8A,$B5,$E3,$F3,$78,$4F,$F8,$F1,$AF,$F5,$CC
	.dc.b	$3C,$0D,$3C,$18,$FA,$1E,$C7,$B0,$B8,$FC,$08,$D4,$78,$4D,$1F,$53
	.dc.b	$1F,$3B,$4F,$2C,$70,$C3,$E4,$47,$F8,$00,$E4,$57,$2F,$98,$C7,$B8
	.dc.b	$F8,$AF,$F2,$9C,$FF,$97,$DB,$19,$9E,$87,$E8,$DC,$3F,$E5,$AF,$91
	.dc.b	$FF,$5F,$A6,$91,$FF,$61,$BF,$3B,$4F,$F4,$06,$FF,$42,$7F,$CC,$F3
	.dc.b	$9F,$65,$39,$9E,$87,$C4,$AE,$2F,$C6,$38,$99,$9E,$2E,$4E,$3B,$7E
	.dc.b	$56,$78,$BF,$97,$8E,$2E,$67,$8B,$98,$F0,$EB,$9B,$F2,$61,$E7,$6C
	.dc.b	$F8,$5B,$3B,$8F,$14,$ED,$A6,$CB,$1B,$3C,$11,$B7,$F6,$5C,$4F,$AE
	.dc.b	$FF,$3E,$5F,$FB,$2D,$B7,$FE,$77,$5C,$DF,$E4,$D9,$FF,$11,$C6,$B3
	.dc.b	$F8,$B7,$CF,$9F,$E3,$E1,$FC,$C2,$B9,$B6,$32,$F2,$6B,$FD,$27,$47
	.dc.b	$CC,$62,$FF,$D7,$8D,$8C,$E3,$FE,$00,$6D,$7F,$7F,$6E,$3E,$4F,$85
	.dc.b	$FE,$2D,$5F,$EE,$79,$CB,$FE,$3C,$8C,$D7,$FA,$57,$EB,$9F,$17,$EF
	.dc.b	$8F,$EB,$91,$8A,$FF,$78,$2F,$8C,$D6,$68,$D5,$73,$BF,$54,$62,$3F
	.dc.b	$A9,$B5,$FF,$26,$B5,$EF,$1D,$3E,$41,$EF,$76,$AE,$BA,$7D,$09,$CE
	.dc.b	$5F,$E8,$23,$FC,$AF,$F0,$1E,$EC,$7F,$BA,$2F,$99,$F3,$B9,$CC,$E6
	.dc.b	$B9,$57,$0A,$FA,$53,$98,$FF,$16,$35,$FE,$DA,$D7,$FC,$CF,$F5,$DF
	.dc.b	$88,$AE,$76,$8C,$4F,$FB,$0E,$BA,$63,$F8,$A9,$CB,$FE,$9B,$FD,$53
	.dc.b	$FC,$C7,$87,$92,$11,$F4,$63,$0E,$AF,$FD,$D1,$3A,$DB,$ED,$3C,$36
	.dc.b	$FC,$7B,$6F,$F8,$9B,$6B,$FC,$CF,$7F,$4E,$3B,$F6,$F2,$8F,$EB,$B5
	.dc.b	$E0,$D7,$13,$F8,$2F,$43,$FC,$87,$3F,$01,$FF,$21,$B7,$CD,$D7,$9D
	.dc.b	$F8,$8F,$51,$F2,$FF,$9F,$6D,$7F,$97,$1B,$A7,$F9,$6D,$AF,$F9,$C3
	.dc.b	$E6,$7F,$23,$EB,$9B,$DF,$87,$F2,$BE,$C7,$E2,$30,$FF,$4D,$BF,$26
	.dc.b	$FF,$DE,$8E,$5F,$F1,$D3,$9E,$0F,$DE,$FB,$2F,$8D,$FD,$A3,$6F,$FC
	.dc.b	$E2,$DB,$FF,$38,$35,$FF,$18,$FC,$BB,$E3,$FF,$4E,$DA,$FC,$2B,$FE
	.dc.b	$F2,$6C,$7E,$B3,$DF,$C5,$7F,$A9,$FE,$2B,$F8,$24,$62,$3F,$C7,$AD
	.dc.b	$7F,$D6,$FF,$58,$F7,$7F,$85,$37,$0F,$F4,$3B,$67,$FD,$0C,$D7,$FC
	.dc.b	$8F,$F0,$BF,$2E,$7F,$DA,$1B,$FF,$13,$8C,$4F,$FA,$35,$AF,$F9,$3B
	.dc.b	$93,$93,$FE,$C8,$6B,$FE,$65,$6F,$77,$FA,$1B,$1F,$E3,$D0,$F9,$2A
	.dc.b	$4C,$7F,$D3,$B1,$96,$FC,$DF,$02,$38,$3F,$B6,$E0,$FA,$BE,$0F,$FB
	.dc.b	$FD,$BC,$7F,$D7,$3E,$41,$F8,$CF,$F9,$2A,$FF,$80,$DF,$1C,$BF,$91
	.dc.b	$36,$3A,$E1,$D2,$3F,$C0,$CD,$44,$70,$D9,$FE,$59,$7E,$0F,$53,$AD
	.dc.b	$79,$2F,$8C,$E5,$FE,$13,$9F,$FA,$05,$BE,$DF,$F3,$08,$7D,$65,$89
	.dc.b	$3C,$3E,$B2,$E6,$D7,$FA,$BE,$09,$E5,$F4,$F2,$FB,$FC,$45,$C3,$84
	.dc.b	$7D,$B9,$F0,$7C,$46,$71,$E6,$FB,$38,$1F,$43,$4C,$78,$57,$0F,$E5
	.dc.b	$BB,$FD,$09,$FE,$C2,$D8,$FF,$12,$B7,$2F,$F1,$93,$5F,$9A,$1D,$63
	.dc.b	$F6,$F0,$7B,$9D,$EB,$D2,$6E,$33,$7F,$33,$C4,$F9,$47,$CB,$FA,$7B
	.dc.b	$19,$3E,$8F,$06,$2F,$9F,$47,$FB,$7A,$3E,$97,$C0,$4A,$FA,$FE,$3D
	.dc.b	$BD,$D7,$FA,$9A,$7E,$DF,$D4,$67,$5F,$A6,$FE,$3B,$EA,$9F,$2F,$4F
	.dc.b	$E7,$3F,$50,$F9,$7C,$61,$F9,$6A,$DB,$FA,$06,$E5,$FE,$4E,$6B,$FE
	.dc.b	$5B,$6B,$F2,$9D,$E3,$FE,$3E,$1F,$82,$7E,$6A,$BB,$63,$FD,$2A,$31
	.dc.b	$7E,$CE,$4F,$C9,$1F,$93,$E3,$B7,$F9,$88,$FC,$B7,$2E,$2B,$E1,$BF
	.dc.b	$87,$EA,$95,$F8,$4E,$5F,$CA,$36,$33,$5F,$6B,$9B,$F6,$FD,$78,$D7
	.dc.b	$F0,$0F,$ED,$BE,$7D,$C3,$4F,$EC,$3B,$BD,$0C,$67,$FB,$EA,$DB,$FC
	.dc.b	$21,$B9,$7F,$9D,$9A,$FF,$A0,$DA,$FE,$49,$DE,$3F,$26,$DF,$1F,$DA
	.dc.b	$BF,$B9,$C6,$2F,$F7,$67,$EE,$4F,$DD,$78,$15,$C9,$FC,$03,$E1,$F1
	.dc.b	$5F,$8A,$7F,$C5,$FF,$A9,$3E,$AE,$98,$F8,$2B,$7F,$E1,$09,$D3,$7E
	.dc.b	$07,$38,$7E,$6A,$7F,$CB,$53,$FC,$63,$B9,$1E,$E9,$46,$38,$3F,$CA
	.dc.b	$93,$98,$C5,$7F,$A7,$1B,$5D,$7E,$6B,$CF,$7E,$87,$D3,$F1,$F2,$FC
	.dc.b	$DF,$53,$6F,$C9,$3D,$89,$3D,$71,$47,$F9,$A2,$5F,$56,$F9,$EC,$36
	.dc.b	$FF,$A1,$B6,$7F,$D6,$2D,$7F,$D7,$CE,$78,$FF,$D3,$FF,$16,$FA,$BF
	.dc.b	$F5,$4D,$FF,$44,$AD,$3F,$CF,$A7,$4F,$D4,$F9,$3F,$8A,$47,$D0,$DC
	.dc.b	$3E,$87,$C2,$6E,$7E,$B3,$F7,$3E,$26,$46,$0F,$B4,$E2,$7E,$C3,$F8
	.dc.b	$29,$C6,$5F,$E9,$2F,$AD,$FF,$4C,$DF,$F2,$6E,$24,$63,$0F,$BC,$D7
	.dc.b	$EE,$B6,$37,$C7,$FE,$C0,$F8,$DF,$99,$5F,$82,$E1,$7D,$5E,$78,$79
	.dc.b	$DD,$AD,$4D,$5F,$59,$FD,$13,$B7,$FD,$AB,$7F,$CE,$A3,$13,$FE,$DF
	.dc.b	$8F,$E7,$BF,$51,$DB,$5D,$CF,$F6,$CA,$DE,$7E,$48,$EB,$E2,$B1,$DC
	.dc.b	$61,$B8,$73,$8E,$C7,$39,$E0,$E0,$E0,$5B,$EF,$8F,$93,$2F,$64,$62
	.dc.b	$79,$61,$E0,$8C,$43,$0F,$06,$3E,$79,$C9,$31,$F0,$C7,$AE,$77,$8F
	.dc.b	$C5,$9C,$B7,$0E,$D1,$8C,$3F,$C1,$F5,$C1,$F1,$9E,$98,$D3,$EC,$BE
	.dc.b	$D8,$FA,$A7,$38,$F8,$F0,$F8,$F3,$97,$B9,$2F,$E3,$61,$DD,$E9,$8F
	.dc.b	$95,$F0,$4F,$2C,$B8,$65,$F0,$72,$EE,$79,$30,$E5,$97,$87,$0E,$27
	.dc.b	$6C,$BC,$4E,$0F,$F0,$09,$F5,$8B,$1F,$8A,$6D,$7D,$C7,$3E,$2F,$FA
	.dc.b	$15,$E2,$7E,$51,$C3,$AF,$67,$FC,$93,$5F,$E9,$46,$CF,$1C,$7E,$D3
	.dc.b	$74,$FC,$D3,$6F,$F2,$DE,$BF,$F8,$5A,$B7,$7F,$85,$AD,$57,$86,$3F
	.dc.b	$11,$5B,$BE,$71,$B2,$4F,$E4,$E7,$4F,$E8,$EB,$9E,$1E,$78,$FC,$C4
	.dc.b	$62,$77,$1D,$CC,$D7,$4C,$3F,$29,$3A,$BE,$51,$E8,$AD,$9E,$13,$86
	.dc.b	$1E,$67,$31,$C9,$C9,$E5,$BA,$FF,$02,$BE,$D8,$7A,$09,$3F,$0A,$77
	.dc.b	$87,$8B,$0E,$D8,$75,$9E,$F1,$82,$3F,$40,$E3,$7D,$1F,$41,$FC,$03
	.dc.b	$83,$F9,$1F,$77,$CA,$7F,$AF,$FB,$D7,$88,$E7,$8F,$FB,$75,$FF,$12
	.dc.b	$AE,$2F,$F8,$88,$FE,$53,$7F,$E5,$4E,$1F,$E6,$15,$C2,$51,$89,$D4
	.dc.b	$7F,$6F,$C1,$F5,$9F,$E5,$CE,$5D,$CF,$F2,$F8,$C5,$ED,$B1,$FC,$CF
	.dc.b	$C4,$B9,$7F,$78,$9B,$87,$4A,$FF,$01,$46,$2B,$3E,$2D,$9F,$F0,$34
	.dc.b	$61,$FF,$04,$47,$FA,$AA,$FF,$C0,$6D,$BF,$F1,$54,$FF,$73,$8F,$F5
	.dc.b	$7A,$FA,$A7,$C5,$79,$BB,$5F,$56,$FA,$6C,$36,$FF,$D5,$F7,$C1,$B7
	.dc.b	$D2,$31,$1F,$B8,$6B,$FC,$09,$AF,$CB,$8F,$31,$1F,$F0,$43,$5F,$F1
	.dc.b	$03,$F1,$8F,$8F,$F1,$5F,$23,$AF,$E4,$FC,$64,$AF,$F1,$54,$65,$FF
	.dc.b	$09,$3F,$0C,$8F,$F8,$54,$FA,$3F,$31,$1F,$F1,$91,$F5,$08,$60,$FF
	.dc.b	$1B,$77,$7D,$AF,$A4,$F8,$91,$FE,$8E,$7F,$D0,$0F,$FB,$33,$93,$F0
	.dc.b	$BC,$29,$CC,$EF,$6E,$44,$78,$72,$7F,$8F,$67,$AD,$F2,$5F,$49,$F8
	.dc.b	$CA,$7D,$5B,$DF,$C7,$CE,$BD,$EA,$6E,$75,$5F,$EE,$86,$BF,$CB,$9A
	.dc.b	$FF,$38,$6B,$FE,$BA,$6B,$FE,$E8,$78,$E3,$FA,$28,$F1,$DF,$73,$C7
	.dc.b	$EE,$7E,$8B,$D5,$77,$BE,$3B,$0B,$E9,$E9,$88,$98,$7F,$B4,$A7,$83
	.dc.b	$63,$8F,$85,$F4,$FE,$2F,$EA,$7A,$F2,$BE,$59,$FE,$64,$E1,$4E,$E3
	.dc.b	$3C,$9F,$56,$31,$1F,$6F,$EA,$7E,$77,$D4,$BE,$43,$E8,$7C,$CB,$D7
	.dc.b	$A6,$A2,$5F,$F5,$81,$5C,$5C,$35,$FE,$BA,$D7,$FB,$3B,$9E,$3F,$98
	.dc.b	$DE,$8F,$F7,$5B,$64,$30,$7A,$FF,$95,$C3,$F2,$DF,$21,$2B,$DF,$B7
	.dc.b	$1D,$BF,$29,$5B,$9C,$BF,$9D,$FA,$BF,$CF,$7C,$F7,$E7,$FE,$AE,$57
	.dc.b	$28,$CB,$48,$7E,$8F,$CD,$FB,$B5,$F7,$B2,$93,$F6,$E3,$2F,$F7,$EE
	.dc.b	$8F,$C2,$C7,$C9,$6C,$67,$9D,$39,$6C,$73,$21,$FE,$D1,$7E,$FB,$F0
	.dc.b	$3E,$77,$F7,$1F,$43,$23,$04,$3F,$75,$1F,$BD,$3F,$7D,$F3,$9F,$BF
	.dc.b	$E9,$FE,$01,$F5,$72,$BE,$5C,$65,$A4,$3F,$82,$7C,$FF,$ED,$FF,$09
	.dc.b	$FE,$AF,$1F,$9B,$EA,$FD,$09,$9F,$52,$47,$E2,$67,$6F,$F8,$89,$AF
	.dc.b	$F8,$B1,$AF,$E0,$83,$C3,$AB,$F6,$8D,$EA,$D3,$65,$C1,$FB,$4E,$DF
	.dc.b	$E3,$9D,$FF,$D2,$FA,$7F,$69,$5F,$EC,$76,$BF,$E3,$CE,$9F,$E3,$93
	.dc.b	$BE,$F9,$3B,$6B,$BA,$FF,$1C,$BF,$E9,$F6,$BF,$C7,$FD,$7F,$01,$FD
	.dc.b	$6F,$87,$FC,$E9,$1F,$ED,$76,$BE,$16,$BF,$EA,$96,$BF,$EE,$5F,$99
	.dc.b	$FE,$6A,$FF,$AB,$5A,$E7,$C8,$FE,$D1,$18,$8F,$F7,$4F,$D6,$7F,$94
	.dc.b	$DF,$F5,$A3,$5F,$F7,$83,$5D,$F9,$5F,$DC,$7E,$6F,$F9,$BF,$0F,$FB
	.dc.b	$D5,$B3,$FE,$F4,$6E,$C7,$F9,$DC,$7F,$12,$F9,$C7,$C1,$7F,$7B,$6C
	.dc.b	$7F,$BD,$DA,$FD,$16,$BF,$EC,$9F,$9D,$FD,$1A,$E5,$FE,$A3,$0E,$FF
	.dc.b	$F0,$8B,$67,$F9,$0B,$C2,$7F,$AB,$FC,$F7,$F4,$99,$FE,$4A,$D7,$A6
	.dc.b	$7F,$C4,$33,$4E,$6B,$8F,$97,$FE,$1E,$F9,$47,$CD,$FF,$85,$5B,$8F
	.dc.b	$F9,$F3,$04,$62,$3F,$18,$D7,$FD,$C9,$F5,$7D,$72,$F0,$3F,$D4,$1B
	.dc.b	$87,$F8,$B2,$35,$3F,$07,$F5,$5F,$E6,$CF,$F4,$2C,$91,$88,$F5,$9B
	.dc.b	$F3,$7F,$C6,$F3,$AF,$80,$B4,$F8,$6F,$E7,$93,$C9,$B1,$9B,$FF,$4B
	.dc.b	$CE,$5B,$FA,$B7,$F6,$58,$FF,$23,$9E,$2F,$8E,$EA,$78,$23,$5A,$7C
	.dc.b	$DF,$AB,$78,$1F,$E4,$86,$E1,$FD,$61,$AF,$AB,$81,$F5,$6F,$8B,$FF
	.dc.b	$27,$B6,$7E,$83,$5F,$DD,$35,$FE,$15,$3C,$A7,$FC,$99,$2F,$8C,$E7
	.dc.b	$12,$BF,$B4,$CE,$5F,$ED,$7E,$D7,$CD,$D7,$FC,$43,$F5,$6F,$9F,$FF
	.dc.b	$30,$B6,$B7,$C7,$FE,$64,$BF,$A9,$1E,$26,$BF,$8F,$6A,$CD,$18,$6C
	.dc.b	$F3,$24,$F9,$38,$FF,$CE,$3F,$54,$F9,$9F,$F3,$9F,$D5,$BE,$3E,$9A
	.dc.b	$89,$7F,$97,$AE,$0D,$8F,$F3,$84,$62,$3F,$92,$7D,$B3,$9B,$FC,$E8
	.dc.b	$D8,$FD,$03,$5D,$C6,$8F,$F4,$2C,$67,$7F,$F2,$4B,$79,$7F,$61,$87
	.dc.b	$E1,$AF,$FC,$03,$83,$D4,$FE,$01,$0F,$AC,$7C,$3F,$F4,$51,$F0,$BF
	.dc.b	$D1,$E3,$33,$FE,$0A,$9C,$BF,$E0,$C8,$CA,$BE,$D9,$E6,$BF,$A8,$46
	.dc.b	$27,$C1,$1D,$EF,$B3,$63,$81,$E8,$C3,$CF,$87,$6B,$E4,$F8,$2B,$C7
	.dc.b	$5D,$0F,$35,$F4,$79,$61,$D1,$F3,$61,$D2,$BC,$11,$C4,$F8,$75,$C5
	.dc.b	$DD,$71,$C3,$D5,$E7,$78,$6B,$FB,$63,$6B,$FC,$5E,$F9,$F0,$F9,$ED
	.dc.b	$BF,$16,$1E,$8C,$78,$5F,$87,$1F,$6C,$FC,$8C,$3C,$D3,$EC,$38,$E3
	.dc.b	$E5,$AD,$C7,$B8,$FE,$FC,$52,$BF,$45,$B3,$F8,$27,$3E,$DB,$E2,$FF
	.dc.b	$CA,$94,$8C,$4F,$43,$EF,$38,$78,$CF,$FB,$21,$7C,$85,$23,$FE,$CC
	.dc.b	$6E,$5F,$88,$CD,$ED,$7F,$B9,$A3,$F1,$75,$FE,$58,$AF,$B1,$5A,$DF
	.dc.b	$ED,$4E,$4F,$B1,$19,$7E,$D1,$9A,$ED,$7D,$75,$F5,$61,$F5,$9F,$F7
	.dc.b	$B4,$7F,$BD,$B6,$F0,$DF,$A8,$FF,$49,$4E,$B1,$E1,$1F,$3A,$FC,$B7
	.dc.b	$D6,$30,$79,$1C,$3F,$E3,$18,$DC,$FA,$A3,$B6,$1E,$68,$C4,$EA,$7D
	.dc.b	$66,$AF,$C4,$E2,$3C,$6F,$F7,$BB,$F5,$4F,$BA,$3D,$27,$FB,$3E,$32
	.dc.b	$FF,$25,$FC,$11,$7F,$D4,$0F,$EF,$0E,$1F,$F5,$0A,$F9,$9D,$4F,$19
	.dc.b	$BD,$0A,$E1,$3F,$E4,$16,$BF,$D2,$1A,$F6,$9F,$F6,$23,$FE,$07,$BF
	.dc.b	$EA,$91,$FE,$01,$6C,$7F,$92,$DA,$FF,$52,$6B,$C7,$B7,$C7,$F0,$7F
	.dc.b	$99,$AF,$F5,$CD,$72,$DB,$FD,$DF,$39,$BF,$EC,$92,$F6,$7D,$6D,$BF
	.dc.b	$F6,$0F,$D6,$7F,$26,$79,$CE,$8F,$4D,$7F,$4F,$AE,$0F,$9E,$33,$9F
	.dc.b	$E7,$9B,$3A,$C7,$C6,$F3,$BF,$F6,$7E,$3F,$9B,$9F,$AB,$87,$F5,$BF
	.dc.b	$7B,$FC,$83,$5F,$ED,$46,$C7,$F7,$D5,$F3,$3F,$51,$CD,$72,$C9,$1C
	.dc.b	$2B,$FB,$DB,$5F,$F1,$C3,$5F,$07,$A3,$EA,$FA,$2F,$B3,$1F,$DB,$5B
	.dc.b	$46,$98,$A7,$CA,$FC,$26,$E5,$FE,$3D,$6D,$FF,$A1,$DA,$F9,$3E,$57
	.dc.b	$81,$27,$83,$F3,$3A,$BF,$03,$3F,$DF,$DB,$8F,$F5,$FE,$1D,$1F,$4B
	.dc.b	$CF,$7C,$6D,$26,$3F,$E4,$F8,$CB,$FE,$F5,$F7,$1F,$2F,$F7,$B2,$46
	.dc.b	$27,$53,$98,$C4,$FF,$A8,$7D,$57,$BF,$F8,$26,$DF,$F9,$75,$B3,$B9
	.dc.b	$FC,$3F,$B1,$F8,$F8,$FF,$3D,$75,$E0,$97,$E5,$F8,$5F,$01,$1C,$1F
	.dc.b	$E6,$9E,$97,$C1,$FE,$49,$B1,$98,$C1,$F9,$6F,$7B,$38,$7F,$AE,$5A
	.dc.b	$FF,$A9,$63,$47,$C3,$FC,$0B,$F8,$25,$7F,$45,$94,$61,$FE,$90,$D7
	.dc.b	$FA,$AB,$5F,$F4,$37,$D5,$BE,$AB,$F4,$ED,$AF,$F4,$13,$67,$FD,$98
	.dc.b	$D7,$FC,$A6,$FF,$A2,$A3,$7E,$E7,$22,$3E,$5B,$7C,$FF,$A4,$91,$97
	.dc.b	$C3,$E7,$48,$F0,$B7,$0F,$E9,$3E,$EF,$F2,$27,$FA,$FB,$FC,$93,$E1
	.dc.b	$9F,$10,$8D,$79,$17,$05,$FF,$A3,$0E,$2F,$C9,$8D,$47,$68,$C4,$FE
	.dc.b	$46,$73,$E1,$7F,$04,$C3,$BF,$D4,$76,$AF,$16,$28,$FE,$6B,$B3,$EA
	.dc.b	$69,$8F,$A2,$B9,$F9,$32,$FE,$0D,$5B,$8C,$1F,$5A,$B8,$38,$D3,$FC
	.dc.b	$DD,$3B,$7F,$D1,$CF,$F9,$EE,$35,$DE,$87,$D1,$F9,$3D,$BF,$2B,$6B
	.dc.b	$73,$A6,$CE,$9F,$05,$F2,$F9,$7F,$A5,$67,$E0,$D6,$19,$5F,$95,$9D
	.dc.b	$36,$3F,$79,$18,$AF,$F0,$F3,$72,$F9,$DA,$F8,$E7,$EC,$4E,$6F,$ED
	.dc.b	$EF,$F8,$08,$F9,$FB,$C9,$F5,$EF,$6F,$E2,$CD,$C7,$E7,$CD,$AF,$D1
	.dc.b	$FC,$B7,$D8,$F0,$BE,$CB,$63,$E2,$B7,$0F,$9E,$DB,$FE,$8F,$0F,$A6
	.dc.b	$7C,$6F,$E5,$DB,$7F,$ED,$FF,$A4,$FF,$42,$3F,$32,$33,$5F,$95,$6B
	.dc.b	$FE,$4C,$6B,$F9,$A3,$B5,$7F,$25,$8C,$C7,$E7,$9E,$77,$FE,$DC,$BF
	.dc.b	$EE,$0F,$9A,$33,$7F,$91,$9C,$BF,$C1,$3C,$5F,$CC,$FD,$4F,$E6,$9B
	.dc.b	$1F,$86,$6E,$1F,$90,$6E,$1F,$A5,$FC,$0B,$F9,$6B,$7C,$4F,$DB,$43
	.dc.b	$E4,$EC,$4C,$3F,$39,$C4,$AE,$31,$F7,$9E,$0E,$9C,$E3,$FB,$CF,$8F
	.dc.b	$7D,$27,$F0,$56,$D7,$E9,$F2,$FF,$7B,$C6,$66,$4F,$F2,$98,$DD,$7F
	.dc.b	$9B,$DA,$FF,$9F,$1A,$FD,$17,$75,$FE,$06,$1F,$13,$FE,$BA,$AF,$D3
	.dc.b	$CE,$9F,$DF,$46,$69,$5F,$C2,$FF,$01,$FE,$2B,$5F,$09,$B3,$A8,$FC
	.dc.b	$FC,$2F,$F9,$03,$AA,$E5,$58,$33,$36,$BF,$7B,$F5,$7F,$C5,$0F,$E2
	.dc.b	$BE,$9D,$8C,$92,$BF,$85,$C6,$5F,$E1,$BE,$4F,$F4,$62,$4F,$F4,$66
	.dc.b	$C7,$F1,$2F,$81,$FE,$43,$1F,$C8,$9F,$E4,$7E,$24,$AF,$E1,$FE,$7D
	.dc.b	$BF,$E3,$33,$FD,$48,$93,$FD,$49,$FF,$01,$7D,$3F,$F7,$89,$FF,$57
	.dc.b	$B5,$FD,$23,$5F,$E7,$ED,$4F,$E2,$3D,$F7,$C8,$3E,$17,$FA,$64,$E6
	.dc.b	$31,$87,$F2,$4F,$7F,$FB,$29,$1F,$EC,$AD,$C3,$F9,$8C,$7F,$40,$8F
	.dc.b	$E8,$2F,$F9,$DB,$E6,$79,$CF,$DB,$1E,$3A,$33,$C5,$FD,$0A,$B8,$65
	.dc.b	$1F,$F6,$15,$6E,$3F,$CB,$B5,$D2,$FF,$82,$4E,$6B,$7F,$AA,$BB,$3F
	.dc.b	$CB,$71,$A8,$C3,$F8,$EA,$F1,$C6,$1F,$E3,$74,$8C,$4F,$1A,$FF,$32
	.dc.b	$F9,$7F,$BB,$BF,$5E,$0F,$92,$CF,$6F,$F7,$0C,$B9,$61,$8B,$DD,$55
	.dc.b	$F5,$1F,$E4,$0E,$8F,$F6,$96,$1F,$C7,$E3,$13,$FE,$DE,$8F,$E4,$5F
	.dc.b	$45,$FB,$8C,$FC,$53,$CB,$97,$F4,$CE,$CF,$F0,$0D,$F2,$9D,$B6,$F9
	.dc.b	$F6,$78,$E3,$15,$C2,$F9,$DF,$06,$DF,$0D,$BC,$F5,$BC,$AB,$E3,$6E
	.dc.b	$1D,$21,$97,$A6,$FD,$D8,$74,$9E,$47,$A4,$E4,$59,$E0,$EB,$1F,$26
	.dc.b	$1D,$AB,$94,$F6,$CB,$D9,$18,$C7,$C4,$78,$F0,$E7,$5D,$B0,$E6,$F8
	.dc.b	$67,$57,$CB,$0E,$C4,$9F,$24,$65,$E1,$87,$73,$C7,$87,$6B,$F0,$C6
	.dc.b	$6F,$E3,$CE,$63,$15,$CA,$73,$7D,$70,$F2,$CF,$F6,$33,$E2,$97,$0F
	.dc.b	$C3,$B6,$FF,$06,$E7,$A3,$FD,$16,$F1,$3F,$22,$E1,$D7,$B3,$FE,$6D
	.dc.b	$A4,$E6,$7F,$28,$F2,$BF,$C5,$B7,$3C,$DF,$E0,$3D,$0F,$EF,$15,$A8
	.dc.b	$FC,$1D,$6E,$FF,$2A,$D7,$F4,$75,$C2,$F9,$C7,$D3,$9D,$61,$E4,$BE
	.dc.b	$D3,$D1,$8E,$DC,$B4,$98,$78,$DF,$BD,$87,$77,$C7,$3C,$A7,$47,$EA
	.dc.b	$27,$53,$F2,$27,$58,$79,$AB,$53,$83,$99,$E0,$8F,$DE,$57,$2B,$F5
	.dc.b	$4E,$4D,$D7,$27,$07,$72,$C2,$79,$61,$F1,$09,$87,$C0,$BF,$4E,$1E
	.dc.b	$52,$61,$FC,$12,$FE,$07,$07,$F1,$FE,$BB,$C4,$F8,$CA,$6B,$93,$FD
	.dc.b	$E1,$87,$F0,$0B,$E0,$FF,$2B,$6B,$FC,$AA,$36,$DB,$FF,$1D,$71,$7F
	.dc.b	$38,$C1,$F1,$BF,$EB,$BC,$3F,$A2,$36,$33,$7F,$D0,$9A,$FE,$86,$76
	.dc.b	$FF,$A0,$23,$3C,$EB,$FE,$65,$18,$6F,$8B,$FE,$4B,$9D,$57,$F3,$AF
	.dc.b	$9F,$C5,$6A,$3A,$56,$1D,$1D,$E7,$CC,$47,$CC,$FF,$B5,$57,$D2,$BD
	.dc.b	$BE,$5D,$EF,$A6,$FF,$45,$57,$FA,$76,$31,$E9,$77,$8E,$35,$6F,$94
	.dc.b	$7F,$B8,$5A,$FE,$ED,$AF,$F0,$56,$BF,$0B,$E1,$3E,$61,$1F,$98,$D7
	.dc.b	$FB,$B6,$DF,$DD,$63,$E0,$36,$3E,$77,$C2,$E3,$77,$CA,$5F,$82,$95
	.dc.b	$F4,$E3,$47,$DB,$23,$F6,$DF,$BD,$F0,$1C,$C9,$1F,$6F,$BA,$BE,$4B
	.dc.b	$4E,$38,$FE,$CE,$5F,$E0,$BB,$FA,$DE,$6B,$E4,$3F,$DA,$58,$FF,$9B
	.dc.b	$9A,$FF,$84,$E3,$5E,$DF,$09,$7B,$96,$25,$71,$75,$1F,$57,$E9,$3F
	.dc.b	$65,$3F,$EC,$6F,$93,$FC,$FE,$7F,$EA,$46,$F4,$FF,$A6,$6B,$75,$FE
	.dc.b	$F3,$6B,$FC,$ED,$AE,$FB,$BF,$15,$18,$8F,$F7,$DB,$CA,$FF,$D6,$2D
	.dc.b	$AE,$13,$E3,$F3,$FE,$1F,$DE,$BE,$0E,$92,$7F,$D2,$31,$FE,$7A,$6C
	.dc.b	$65,$F8,$D1,$C5,$DF,$A7,$F8,$1F,$88,$FC,$17,$C5,$49,$FD,$94,$65
	.dc.b	$FD,$A4,$65,$FB,$51,$F9,$DF,$BD,$FF,$01,$C7,$FB,$BB,$E3,$15,$FA
	.dc.b	$16,$C7,$FB,$CA,$76,$FF,$92,$C3,$87,$FB,$23,$5F,$F6,$43,$51,$C3
	.dc.b	$CE,$FD,$B4,$74,$DB,$FD,$B6,$DD,$78,$CF,$8F,$1F,$D0,$77,$7E,$87
	.dc.b	$EA,$66,$3F,$C4,$63,$2F,$F1,$38,$CB,$F9,$98,$FD,$47,$D4,$48,$FE
	.dc.b	$30,$FE,$AF,$EA,$71,$1F,$9F,$8B,$9F,$E7,$BB,$15,$FC,$17,$F3,$A6
	.dc.b	$5F,$A0,$FC,$51,$85,$35,$13,$1F,$F0,$00,$AE,$2E,$1A,$FF,$81,$5A
	.dc.b	$FF,$83,$DC,$E3,$FF,$00,$1B,$AF,$FB,$D5,$B2,$EB,$63,$E8,$8F,$E6
	.dc.b	$DF,$33,$FC,$27,$79,$18,$21,$FB,$F7,$2F,$F0,$18,$FE,$20,$C3,$FA
	.dc.b	$01,$FC,$4F,$D4,$C3,$FC,$1C,$B9,$7C,$98,$FC,$87,$95,$F1,$F8,$25
	.dc.b	$7E,$36,$32,$FC,$6E,$24,$7E,$18,$91,$97,$6F,$E4,$E3,$6F,$FB,$F3
	.dc.b	$27,$17,$F2,$AF,$8E,$7D,$87,$F5,$F6,$D7,$FA,$79,$AF,$F5,$D6,$BF
	.dc.b	$E3,$FF,$01,$EF,$7F,$4D,$6C,$7D,$C6,$BF,$E3,$D6,$BF,$EA,$69,$FE
	.dc.b	$9E,$F5,$F2,$BC,$44,$7C,$5F,$4B,$E5,$23,$83,$B7,$77,$84,$8E,$0F
	.dc.b	$F5,$BF,$D2,$FF,$61,$F1,$8D,$24,$7F,$7F,$FA,$5F,$E2,$0F,$F7,$B6
	.dc.b	$BF,$EB,$C6,$BF,$EA,$B6,$BF,$BA,$FB,$57,$B7,$FD,$BD,$B1,$F9,$E6
	.dc.b	$BF,$E5,$F6,$BF,$EC,$8F,$95,$7D,$5F,$F8,$45,$B1,$FE,$D0,$6B,$FE
	.dc.b	$0F,$6B,$FE,$D0,$FA,$67,$C6,$7F,$86,$5B,$1F,$ED,$86,$BF,$E1,$76
	.dc.b	$BF,$E7,$9B,$FF,$36,$BE,$7F,$7B,$D4,$47,$FD,$C3,$F0,$9C,$1F,$F6
	.dc.b	$63,$5F,$F7,$5B,$75,$FF,$7E,$E1,$FE,$2B,$FB,$77,$C5,$D2,$5F,$FB
	.dc.b	$B2,$32,$FF,$BB,$BE,$91,$ED,$FF,$8C,$DB,$19,$8C,$19,$EC,$91,$C5
	.dc.b	$AF,$F8,$D7,$E9,$7F,$A4,$37,$CE,$FF,$1F,$46,$23,$FD,$F2,$D7,$FD
	.dc.b	$D1,$7F,$DF,$9F,$27,$AB,$E3,$23,$83,$F9,$DF,$85,$FE,$02,$9E,$11
	.dc.b	$FE,$1A,$6C,$67,$D5,$FF,$0D,$FD,$1F,$F7,$4A,$6E,$FA,$47,$CD,$FF
	.dc.b	$88,$1B,$3D,$BC,$BF,$F7,$E5,$F1,$B3,$FE,$5E,$9F,$E9,$A7,$47,$0F
	.dc.b	$0F,$81,$FF,$2F,$46,$23,$51,$E4,$23,$DC,$F2,$FA,$FF,$C3,$1F,$F1
	.dc.b	$AC,$FE,$21,$FF,$1B,$E1,$FE,$18,$78,$5F,$AF,$D8,$F7,$F3,$C9,$FF
	.dc.b	$60,$4F,$A2,$3C,$64,$8C,$F8,$7F,$E8,$16,$FC,$1F,$F9,$D6,$31,$1F
	.dc.b	$E7,$16,$BF,$02,$BF,$C6,$13,$FE,$48,$9D,$EF,$FE,$36,$6E,$BF,$DC
	.dc.b	$AB,$FC,$6D,$F1,$1F,$E5,$FD,$1F,$3A,$C3,$FB,$BF,$D1,$FF,$97,$BC
	.dc.b	$D7,$9F,$F3,$A3,$0D,$8F,$A3,$7C,$AB,$8C,$62,$FA,$6A,$FA,$3E,$2F
	.dc.b	$F8,$11,$AE,$67,$85,$65,$B1,$93,$FC,$16,$F9,$1D,$18,$C3,$F0,$75
	.dc.b	$FE,$76,$C3,$C9,$3C,$4F,$C4,$E1,$E4,$9F,$09,$2F,$BB,$C6,$3D,$38
	.dc.b	$F9,$89,$3E,$47,$58,$F8,$AF,$FD,$21,$8F,$8A,$BF,$2A,$79,$E3,$C5
	.dc.b	$8F,$93,$1E,$31,$E5,$9D,$47,$8E,$3F,$C5,$33,$B9,$F8,$71,$FA,$03
	.dc.b	$D7,$7F,$E0,$E9,$F7,$C6,$5A,$FD,$6F,$A4,$7C,$2E,$DB,$1C,$CF,$D6
	.dc.b	$B8,$79,$35,$D7,$C5,$7F,$AB,$D3,$48,$FF,$AC,$DF,$89,$6F,$A2,$FF
	.dc.b	$60,$E5,$FE,$A6,$9C,$C6,$32,$EB,$7F,$5E,$F8,$70,$7C,$D8,$FF,$25
	.dc.b	$DF,$08,$C4,$78,$86,$BF,$EB,$6A,$E3,$18,$BE,$B3,$9C,$3F,$DA,$6F
	.dc.b	$8E,$7C,$11,$E3,$7E,$76,$1D,$DF,$C5,$1E,$23,$78,$78,$A3,$58,$F3
	.dc.b	$3B,$46,$67,$67,$23,$D1,$1A,$71,$3C,$B0,$FD,$C7,$D2,$3E,$4F,$FD
	.dc.b	$D3,$9F,$FA,$CA,$31,$8F,$F1,$27,$FB,$AB,$67,$6F,$27,$FD,$E3,$87
	.dc.b	$F0,$99,$E2,$FF,$BF,$5A,$E6,$36,$FF,$09,$C9,$F1,$18,$8F,$E7,$71
	.dc.b	$8B,$FE,$3E,$D8,$FC,$B7,$1F,$E0,$6B,$83,$6B,$F9,$4F,$BE,$F7,$FC
	.dc.b	$F7,$D2,$ED,$AB,$57,$FC,$C3,$E9,$3F,$C7,$ED,$F5,$F7,$1F,$E7,$68
	.dc.b	$C1,$30,$F1,$C7,$06,$CF,$92,$31,$1F,$CB,$5D,$6B,$FE,$E0,$97,$CC
	.dc.b	$F6,$7C,$A6,$2F,$FA,$1F,$D2,$7F,$7C,$6F,$8F,$EA,$8C,$63,$FF,$00
	.dc.b	$1F,$F2,$FC,$E8,$FE,$8E,$E2,$3E,$4B,$5F,$E8,$0D,$7F,$9E,$78,$2F
	.dc.b	$95,$F8,$ED,$AE,$93,$FC,$F5,$DC,$FC,$CF,$BF,$7A,$D4,$C5,$DA,$FE
	.dc.b	$BB,$F7,$EF,$1F,$EC,$B6,$3F,$AE,$B7,$1F,$BE,$D7,$FA,$B7,$7F,$E0
	.dc.b	$F7,$57,$F5,$38,$7C,$77,$DE,$75,$19,$8F,$C0,$F9,$EF,$9F,$A6,$3E
	.dc.b	$7E,$D7,$F7,$2F,$57,$FC,$EE,$FF,$90,$FD,$EF,$C7,$C6,$33,$FF,$22
	.dc.b	$47,$E4,$BD,$77,$CA,$FE,$39,$7D,$0D,$8C,$3F,$92,$FA,$6F,$C1,$E3
	.dc.b	$FD,$AD,$B8,$7F,$87,$1F,$C8,$3F,$3C,$C7,$47,$E6,$BD,$D7,$D6,$FE
	.dc.b	$89,$B7,$FE,$06,$6D,$7F,$9D,$67,$F4,$FE,$AB,$E3,$BF,$46,$BE,$06
	.dc.b	$C6,$1F,$D3,$FC,$1E,$F1,$FF,$4F,$4F,$EC,$23,$18,$7F,$A7,$E3,$1E
	.dc.b	$D4,$AF,$F5,$23,$67,$FD,$49,$F6,$1C,$49,$3F,$EA,$97,$A7,$D6,$48
	.dc.b	$FF,$56,$36,$7F,$D5,$99,$F8,$49,$3F,$EB,$17,$C9,$F0,$52,$3F,$D6
	.dc.b	$8D,$9F,$EA,$9F,$6A,$F9,$4F,$E0,$B0,$8C,$57,$F8,$E7,$E3,$5F,$8B
	.dc.b	$7F,$93,$23,$FC,$79,$5B,$BE,$37,$C2,$3A,$C7,$F1,$B7,$53,$FE,$F8
	.dc.b	$9D,$47,$F6,$F7,$87,$77,$E6,$A3,$FD,$E1,$F8,$B3,$E7,$BA,$E1,$FC
	.dc.b	$6B,$24,$E7,$1F,$F2,$3F,$0F,$F1,$AC,$3F,$0D,$E6,$21,$F6,$5F,$6B
	.dc.b	$CF,$FB,$6D,$BE,$2D,$BE,$35,$C2,$7A,$F8,$7F,$82,$3F,$07,$EB,$D8
	.dc.b	$C9,$2B,$8C,$65,$FC,$C6,$7E,$8F,$5F,$D5,$A7,$F6,$CB,$F9,$CB,$E9
	.dc.b	$B3,$32,$7F,$C1,$55,$C5,$B5,$C6,$FF,$C3,$13,$96,$C6,$78,$3E,$64
	.dc.b	$FC,$8E,$FE,$95,$E4,$AF,$DB,$DF,$CA,$F4,$FF,$3F,$E7,$7E,$81,$B1
	.dc.b	$93,$E7,$C6,$5A,$7D,$2F,$07,$F4,$E7,$EA,$3D,$0B,$19,$26,$1F,$3A
	.dc.b	$32,$FE,$F4,$FF,$10,$FD,$57,$E5,$3D,$83,$FD,$6B,$57,$E0,$0F,$82
	.dc.b	$D4,$CC,$7F,$C6,$18,$72,$6E,$1C,$9B,$3F,$95,$8C,$4F,$F0,$DC,$3A
	.dc.b	$F4,$73,$8C,$63,$F9,$6C,$3F,$D0,$1C,$BF,$5D,$FC,$8F,$1F,$F0,$6E
	.dc.b	$1F,$E0,$EC,$8C,$9F,$96,$8C,$5F,$D1,$E6,$FC,$1F,$67,$E1,$1B,$5F
	.dc.b	$84,$6C,$EE,$3F,$0F,$3F,$B8,$F9,$37,$CC,$53,$E3,$7D,$FB,$C3,$C5
	.dc.b	$FE,$E2,$B6,$FC,$66,$E1,$FE,$57,$6B,$FE,$63,$6B,$FE,$C4,$77,$BF
	.dc.b	$F8,$A2,$FF,$90,$BF,$E2,$28,$CD,$7F,$B7,$27,$2F,$C8,$EE,$FE,$57
	.dc.b	$D5,$FC,$B1,$B1,$FB,$A6,$E1,$FB,$26,$E1,$FC,$17,$B9,$F4,$6F,$AF
	.dc.b	$FF,$0D,$BF,$C5,$6F,$F8,$0B,$9C,$3F,$D9,$6E,$6F,$93,$99,$DF,$C5
	.dc.b	$72,$BE,$6F,$4E,$8F,$F1,$6D,$7E,$3F,$0F,$F6,$8B,$6F,$FC,$66,$D7
	.dc.b	$FD,$94,$E7,$0E,$FC,$DC,$EB,$C6,$FA,$3E,$B6,$47,$9B,$E8,$2E,$1E
	.dc.b	$A9,$F0,$92,$7C,$2F,$93,$D2,$91,$E4,$6C,$F9,$1F,$8B,$D7,$F1,$9B
	.dc.b	$3F,$13,$D3,$F6,$91,$F6,$B6,$BE,$5F,$C8,$FC,$E8,$F9,$EC,$9F,$9A
	.dc.b	$E5,$A4,$9F,$A7,$F5,$BB,$6C,$7D,$4F,$5B,$FA,$03,$F2,$1B,$1F,$EE
	.dc.b	$DD,$91,$88,$71,$7E,$0A,$30,$DA,$FF,$4D,$F5,$F1,$FA,$8F,$E9,$9E
	.dc.b	$07,$F8,$53,$0E,$38,$F4,$BD,$C6,$2B,$76,$89,$F5,$0F,$97,$75,$F0
	.dc.b	$94,$E6,$31,$F3,$BF,$E3,$9C,$FC,$15,$AB,$E2,$FF,$AF,$B0,$EC,$E6
	.dc.b	$FA,$61,$CE,$34,$78,$F0,$E9,$5C,$27,$58,$76,$6C,$67,$0E,$F3,$E2
	.dc.b	$AE,$37,$EC,$9C,$D6,$F0,$F0,$C6,$27,$C1,$EE,$FC,$47,$84,$E8,$8F
	.dc.b	$98,$5F,$67,$D8,$67,$1F,$2B,$C8,$E1,$1B,$8C,$C7,$0C,$7D,$98,$78
	.dc.b	$E3,$AE,$1E,$37,$58,$FA,$5E,$D3,$BC,$3A,$3C,$70,$EA,$76,$8C,$E3
	.dc.b	$EA,$C7,$BB,$F6,$B1,$ED,$5C,$2F,$91,$D6,$32,$E2,$79,$63,$F7,$30
	.dc.b	$EB,$8F,$92,$7F,$AA,$E1,$F7,$63,$F0,$27,$C6,$2B,$97,$EF,$7C,$87
	.dc.b	$F8,$8D,$D9,$F9,$07,$0F,$06,$BC,$3B,$3B,$C7,$1C,$FF,$0C,$DC,$75
	.dc.b	$5F,$57,$A3,$F2,$75,$AB,$FC,$9D,$EE,$BF,$46,$D7,$F4,$35,$C2,$F9
	.dc.b	$DF,$19,$FC,$05,$6F,$0F,$F2,$84,$74,$C2,$3A,$7F,$43,$87,$38,$EF
	.dc.b	$1A,$BE,$87,$E7,$E7,$53,$F0,$E7,$58,$78,$8F,$D3,$4E,$AD,$87,$8C
	.dc.b	$95,$E0,$BE,$27,$80,$F2,$1F,$14,$C9,$DC,$CE,$1E,$37,$35,$DA,$6D
	.dc.b	$F8,$6F,$BD,$70,$C3,$CA,$F0,$8C,$E1,$CE,$3F,$80,$F0,$7F,$60,$F8
	.dc.b	$0F,$88,$FF,$62,$F8,$0B,$C4,$6C,$CA,$FE,$35,$C5,$FD,$52,$3F,$87
	.dc.b	$B6,$32,$7F,$2F,$8C,$BF,$CC,$71,$FE,$2B,$D7,$FD,$A3,$DD,$B7,$FC
	.dc.b	$D5,$B5,$FC,$B2,$AE,$5F,$DD,$9B,$97,$D0,$9F,$EE,$71,$8E,$29,$18
	.dc.b	$21,$FD,$E9,$CB,$FD,$B9,$FF,$1E,$D7,$F8,$09,$B5,$FC,$D6,$7F,$BA
	.dc.b	$FC,$17,$83,$E1,$E4,$7F,$80,$A3,$C6,$70,$23,$C0,$FD,$6D,$BE,$65
	.dc.b	$F5,$3E,$06,$CE,$9A,$FE,$EA,$78,$4E,$CF,$87,$3A,$D7,$FC,$34,$DD
	.dc.b	$7F,$AF,$1A,$F9,$2D,$FC,$8C,$C7,$FA,$95,$F0,$9D,$47,$85,$F2,$11
	.dc.b	$F2,$7C,$47,$83,$EA,$64,$72,$C3,$FA,$E6,$7E,$DA,$F2,$9A,$9F,$F5
	.dc.b	$4E,$4B,$CC,$71,$3F,$B8,$C3,$E1,$FD,$BF,$31,$27,$58,$7C,$99,$F8
	.dc.b	$26,$63,$AC,$E1,$FF,$65,$FD,$AF,$F6,$6D,$3F,$D8,$50,$F9,$7F,$B8
	.dc.b	$FF,$A1,$5B,$1A,$9E,$1A,$7F,$6A,$7F,$B7,$D7,$F2,$C6,$B9,$E9,$FF
	.dc.b	$42,$A8,$CB,$FD,$FD,$E6,$D8,$FC,$EB,$B6,$DB,$8B,$F4,$3E,$D3,$F1
	.dc.b	$67,$CB,$61,$FE,$C7,$87,$FB,$AD,$AF,$FB,$11,$DB,$98,$E1,$E0,$C8
	.dc.b	$C1,$F3,$38,$1F,$53,$F8,$C2,$46,$7E,$4A,$BF,$6A,$2E,$FF,$90,$9F
	.dc.b	$5F,$D1,$7D,$9A,$FF,$7D,$FC,$9F,$F8,$51,$FF,$59,$36,$BE,$9C,$EE
	.dc.b	$7F,$D6,$CD,$7F,$AC,$B5,$FF,$0D,$B5,$FF,$15,$35,$3C,$E3,$AF,$1F
	.dc.b	$90,$F9,$F5,$FD,$92,$3F,$4F,$E3,$7E,$A8,$97,$F0,$FE,$A6,$BF,$98
	.dc.b	$8B,$3F,$AE,$9F,$D8,$FD,$3F,$EC,$BE,$36,$5F,$CB,$8C,$BF,$36,$32
	.dc.b	$FE,$8A,$3E,$D5,$FD,$BA,$93,$F6,$E3,$2F,$FB,$DB,$81,$1F,$ED,$2E
	.dc.b	$1F,$C2,$36,$33,$3B,$8E,$2F,$E0,$DF,$C4,$6D,$FC,$23,$E3,$E4,$FE
	.dc.b	$05,$FE,$17,$F4,$D8,$8F,$DB,$C5,$D3,$F8,$6C,$FF,$10,$F6,$FF,$88
	.dc.b	$FC,$C4,$BF,$C5,$46,$5F,$C6,$46,$5F,$E0,$31,$FA,$0F,$B6,$F3,$C7
	.dc.b	$E5,$F8,$FF,$38,$67,$39,$1F,$EF,$09,$DB,$FE,$1C,$6B,$FE,$5E,$6B
	.dc.b	$FE,$6D,$7A,$E3,$FC,$01,$BD,$7B,$AF,$D5,$71,$FF,$2A,$FA,$8F,$E6
	.dc.b	$44,$C3,$F9,$93,$FC,$D7,$9E,$47,$F3,$56,$BF,$CD,$79,$FF,$5E,$DE
	.dc.b	$8F,$E7,$84,$7F,$9E,$79,$FF,$CF,$C8,$E0,$FE,$80,$7F,$43,$21,$83
	.dc.b	$FA,$4F,$C2,$76,$7F,$CF,$BE,$02,$31,$1F,$CC,$9A,$FF,$A7,$1B,$0F
	.dc.b	$93,$7C,$8F,$F6,$C7,$FD,$F9,$5F,$CB,$9A,$B8,$3F,$DC,$D1,$C2,$3F
	.dc.b	$99,$FC,$A7,$F0,$87,$FA,$EB,$5F,$EB,$0D,$7C,$ED,$7F,$A3,$FC,$AB
	.dc.b	$E3,$FF,$BE,$B6,$3F,$9F,$B5,$F1,$71,$70,$F8,$6F,$F0,$1F,$CB,$7F
	.dc.b	$AE,$A7,$FA,$2A,$F6,$32,$6B,$D9,$FF,$04,$9F,$E4,$69,$FE,$D8,$D7
	.dc.b	$E5,$B5,$FE,$B6,$D7,$FD,$8B,$F4,$7D,$DF,$19,$88,$FF,$0A,$C6,$75
	.dc.b	$63,$FD,$92,$3C,$95,$FD,$6B,$E8,$DE,$B7,$F8,$71,$FF,$17,$46,$7D
	.dc.b	$7C,$4F,$F8,$83,$BF,$EF,$4F,$F8,$29,$AF,$F8,$15,$AF,$E1,$1A,$FF
	.dc.b	$74,$97,$C2,$7F,$8A,$9B,$1F,$DB,$1A,$FB,$39,$1F,$7D,$FE,$82,$D3
	.dc.b	$D7,$3F,$EE,$6F,$A4,$FC,$13,$B6,$C6,$A9,$18,$D7,$FC,$0B,$F4,$8F
	.dc.b	$27,$FC,$76,$DC,$3F,$C4,$AD,$7F,$C4,$0D,$7F,$3E,$D7,$9E,$9C,$A6
	.dc.b	$47,$F8,$1A,$53,$FE,$48,$F3,$BF,$A2,$5F,$F8,$5B,$ED,$1F,$09,$FE
	.dc.b	$4C,$6E,$1F,$E0,$EA,$DA,$CF,$FC,$11,$E1,$7F,$91,$A7,$FA,$94,$62
	.dc.b	$7F,$C7,$B1,$83,$8C,$66,$3E,$64,$60,$7A,$7F,$E5,$CF,$A3,$FF,$08
	.dc.b	$E5,$FE,$2A,$FA,$57,$C4,$FF,$98,$1B,$87,$F8,$8A,$B6,$F9,$5F,$F0
	.dc.b	$DB,$FC,$11,$AF,$F8,$D1,$AE,$FD,$EF,$F2,$FF,$D2,$BD,$AA,$49,$FF
	.dc.b	$2B,$C6,$5B,$1F,$E7,$28,$C4,$7E,$1D,$BC,$9F,$81,$7F,$C7,$D1,$FE
	.dc.b	$3D,$16,$F9,$9F,$E1,$90,$FA,$47,$A0,$88,$FF,$2C,$8C,$C6,$3C,$9D
	.dc.b	$3C,$66,$DF,$5C,$79,$5F,$0A,$F0,$E1,$E0,$8C,$5F,$53,$99,$67,$C5
	.dc.b	$7C,$DE,$C6,$2B,$94,$F0,$C3,$C7,$7D,$1B,$1E,$49,$DD,$71,$C3,$CB
	.dc.b	$3C,$62,$9E,$D2,$9E,$88,$B8,$78,$FE,$3F,$B4,$FC,$76,$A7,$E0,$46
	.dc.b	$EB,$C5,$5D,$1F,$75,$7D,$0C,$3C,$EF,$94,$F7,$9E,$12,$C7,$03,$8D
	.dc.b	$7C,$5C,$7B,$E3,$F7,$2B,$EA,$9C,$4B,$3F,$2D,$B1,$B7,$EC,$FC,$D3
	.dc.b	$E9,$78,$C7,$D2,$8C,$38,$A4,$62,$3A,$B5,$E3,$F0,$FF,$E7,$44,$D2
	.dc.b	$3F,$E7,$87,$E7,$E6,$F9,$AF,$E3,$F9,$7E,$26,$BF,$29,$5A,$9C,$E7
	.dc.b	$DA,$7E,$3D,$F1,$E3,$FB,$31,$C0,$CD,$70,$7E,$2D,$FD,$9B,$FF,$71
	.dc.b	$CF,$F2,$ED,$BC,$15,$E8,$3E,$E4,$EA,$F8,$47,$23,$AB,$E0,$78,$63
	.dc.b	$FA,$1C,$7C,$91,$F7,$0F,$24,$67,$0F,$24,$7E,$B3,$1F,$75,$F3,$6C
	.dc.b	$71,$3A,$E1,$E3,$71,$1A,$BF,$4B,$67,$FD,$8B,$F4,$AF,$88,$FE,$97
	.dc.b	$97,$F4,$78,$C6,$3F,$EB,$E5,$F1,$3A,$8E,$35,$72,$FE,$69,$5C,$1F
	.dc.b	$EB,$0D,$BE,$97,$BE,$94,$E7,$0F,$F6,$BE,$9F,$EC,$F7,$FD,$EA,$DC
	.dc.b	$3F,$DE,$0D,$7F,$C5,$ED,$78,$F5,$75,$F7,$DE,$5F,$7A,$E5,$9F,$F6
	.dc.b	$39,$CE,$1F,$ED,$AF,$A5,$FE,$C4,$DF,$0B,$CD,$0F,$80,$B1,$27,$F4
	.dc.b	$7A,$AB,$8E,$9F,$AE,$8C,$57,$FB,$7A,$3F,$DE,$73,$A7,$FC,$8E,$F0
	.dc.b	$C7,$FB,$B7,$B9,$F8,$47,$FB,$C4,$67,$0F,$F7,$DC,$E5,$FF,$7E,$FC
	.dc.b	$E7,$FB,$E5,$FF,$79,$AF,$87,$F9,$CE,$70,$FF,$05,$39,$9D,$CF,$96
	.dc.b	$7F,$B6,$4E,$5F,$53,$5F,$F2,$EB,$5F,$F0,$D7,$C2,$7D,$BF,$2B,$FD
	.dc.b	$CD,$7D,$0F,$F0,$D4,$7D,$FF,$A5,$7A,$17,$D1,$C3,$6B,$FA,$76,$BF
	.dc.b	$E1,$7F,$A5,$7C,$2F,$97,$85,$18,$AF,$A2,$DD,$3E,$83,$5F,$F3,$9F
	.dc.b	$83,$FD,$F2,$7A,$9F,$D6,$EB,$FB,$07,$91,$F9,$39,$FE,$E1,$3A,$AF
	.dc.b	$F0,$74,$71,$7F,$C9,$A7,$F8,$5B,$E8,$BF,$41,$93,$F2,$3F,$F0,$CD
	.dc.b	$FF,$6D,$F3,$55,$A9,$CD,$F2,$97,$CD,$FD,$18,$FF,$10,$B6,$33,$7F
	.dc.b	$CC,$27,$2F,$E4,$3D,$8F,$F4,$73,$FE,$74,$F7,$7F,$59,$1F,$85,$9F
	.dc.b	$F1,$33,$A7,$D3,$F1,$1F,$77,$3F,$DF,$FE,$A6,$F9,$2F,$DD,$52,$B5
	.dc.b	$8F,$D4,$AD,$3F,$E9,$3F,$A5,$FA,$9F,$00,$69,$23,$F4,$BD,$EF,$A9
	.dc.b	$FE,$00,$D8,$F9,$DC,$9F,$AD,$FA,$AC,$D6,$BE,$A3,$7F,$53,$6F,$78
	.dc.b	$23,$9F,$1F,$E1,$63,$E3,$ED,$FB,$16,$BF,$EB,$46,$BF,$81,$7D,$15
	.dc.b	$D3,$3F,$2F,$C2,$F9,$9F,$3F,$D1,$7A,$39,$ED,$78,$9F,$5F,$91,$E1
	.dc.b	$24,$7C,$07,$C9,$E8,$FC,$03,$E0,$FC,$2D,$9F,$F1,$97,$D4,$3E,$87
	.dc.b	$E8,$B6,$3F,$25,$D0,$FC,$87,$EE,$6E,$AD,$3F,$94,$6B,$FD,$F9,$AF
	.dc.b	$F8,$D1,$AF,$F8,$ED,$AF,$3B,$FC,$DC,$BC,$7F,$1C,$61,$FB,$27,$DA
	.dc.b	$EF,$B8,$FD,$46,$DF,$FA,$5F,$87,$FC,$99,$3F,$62,$B8,$3F,$6B,$B7
	.dc.b	$F0,$67,$E1,$3C,$DB,$19,$25,$F3,$8C,$BF,$EB,$EE,$C7,$CA,$FF,$9D
	.dc.b	$2B,$F7,$11,$8F,$7B,$FA,$A3,$AD,$2D,$6E,$F9,$35,$FF,$2C,$CF,$1B
	.dc.b	$FF,$0F,$B6,$3F,$C5,$2D,$77,$3C,$F7,$FF,$59,$CF,$93,$BF,$BE,$3E
	.dc.b	$07,$F4,$35,$FA,$2F,$A8,$B8,$FE,$71,$B8,$FC,$96,$E3,$FE,$B9,$D9
	.dc.b	$1F,$AA,$F6,$25,$FE,$93,$E9,$AE,$1F,$46,$6E,$1C,$66,$D7,$F8,$2F
	.dc.b	$D5,$FD,$FF,$B1,$FD,$F2,$3F,$1F,$FA,$47,$E3,$A3,$18,$FF,$20,$6B
	.dc.b	$FE,$26,$AF,$E3,$11,$88,$E9,$5F,$C5,$E3,$8E,$1F,$E0,$CB,$FC,$64
	.dc.b	$FF,$A4,$67,$2F,$F4,$08,$CD,$7F,$B7,$FF,$36,$FF,$72,$CF,$F0,$66
	.dc.b	$C7,$E3,$7E,$83,$FC,$15,$18,$C3,$FD,$BC,$DA,$FE,$7C,$DA,$FF,$64
	.dc.b	$6A,$BE,$3F,$15,$F8,$6F,$D4,$BF,$93,$63,$F6,$9F,$E5,$5C,$67,$39
	.dc.b	$7F,$A8,$E3,$1C,$BC,$79,$B9,$36,$7F,$27,$5C,$8F,$F0,$7E,$7F,$99
	.dc.b	$6E,$1D,$FB,$FC,$5D,$D2,$FF,$23,$5C,$2F,$F2,$AD,$7C,$4D,$7F,$9C
	.dc.b	$3D,$23,$3D,$FC,$E7,$C0,$FC,$8E,$7F,$93,$F5,$16,$BE,$73,$6B,$E9
	.dc.b	$37,$2F,$9A,$FD,$0E,$9F,$A3,$D5,$6F,$EB,$36,$FE,$D3,$6F,$EA,$BF
	.dc.b	$63,$E9,$FF,$BE,$72,$1A,$4B,$FF,$16,$C6,$5F,$C1,$7D,$47,$16,$9F
	.dc.b	$85,$F3,$BF,$07,$F5,$1F,$84,$6C,$7E,$26,$2E,$5F,$8D,$9B,$97,$E2
	.dc.b	$3E,$A3,$C6,$4B,$F0,$B6,$BF,$29,$F5,$1F,$90,$8F,$C8,$BF,$92,$F5
	.dc.b	$FD,$64,$AF,$4B,$6B,$F3,$9E,$BE,$E7,$F3,$8F,$E6,$BD,$A9,$18,$3F
	.dc.b	$40,$E2,$7F,$48,$FE,$82,$71,$8A,$BF,$C3,$97,$FE,$0E,$7F,$DE,$7E
	.dc.b	$8A,$31,$86,$B3,$E0,$C1,$3B,$8C,$D7,$2A,$FF,$0C,$F7,$FF,$30,$BD
	.dc.b	$EE,$F9,$2E,$73,$9C,$79,$61,$FE,$B6,$F1,$7F,$67,$5A,$58,$4F,$A8
	.dc.b	$7C,$63,$E6,$38,$4E,$63,$0F,$FB,$22,$B8,$AF,$9A,$7C,$E7,$58,$E1
	.dc.b	$7C,$96,$1D,$A3,$0D,$8E,$87,$92,$B9,$E5,$AC,$7C,$3A,$7C,$1A,$E7
	.dc.b	$18,$CB,$C7,$87,$3A,$D6,$1C,$AF,$A9,$C3,$1E,$84,$73,$1B,$9E,$D5
	.dc.b	$CE,$FA,$E1,$DE,$FD,$A7,$0C,$BE,$03,$D4,$EF,$5C,$5D,$DF,$92,$77
	.dc.b	$39,$9E,$13,$E0,$AE,$D1,$F3,$6B,$71,$8A,$E0,$71,$C3,$D1,$87,$9B
	.dc.b	$1F,$66,$3C,$DF,$A5,$3A,$C7,$C7,$5B,$7C,$F8,$78,$27,$51,$E4,$9D
	.dc.b	$5F,$38,$F8,$19,$76,$9C,$92,$3E,$D6,$5E,$8C,$7E,$74,$7E,$20,$F8
	.dc.b	$C5,$CB,$F0,$4D,$CB,$F1,$EE,$7E,$23,$FA,$43,$C4,$FC,$5B,$87,$5E
	.dc.b	$2F,$F9,$DE,$BF,$DD,$8D,$9E,$39,$FD,$C6,$E5,$F8,$86,$DF,$E8,$FB
	.dc.b	$7F,$E0,$04,$E6,$FF,$2F,$5A,$AF,$0C,$7E,$6A,$B7,$1F,$AB,$9D,$61
	.dc.b	$E2,$8F,$C1,$C6,$27,$71,$C8,$CD,$76,$3A,$34,$C9,$30,$FC,$AC,$F2
	.dc.b	$BE,$D1,$ED,$AE,$0E,$AA,$E1,$EB,$79,$3F,$A8,$BE,$58,$7A,$09,$7E
	.dc.b	$2A,$D9,$ED,$79,$BF,$1C,$A7,$AE,$FC,$4F,$E8,$AB,$AC,$EB,$0F,$3E
	.dc.b	$1E,$1A,$C6,$1E,$07,$81,$FC,$4B,$7F,$E5,$5D,$1F,$39,$70,$BC,$8C
	.dc.b	$36,$B8,$71,$AF,$83,$63,$F9,$0F,$0F,$F2,$D9,$FE,$36,$D8,$C9,$FC
	.dc.b	$F2,$32,$FF,$87,$31,$FE,$43,$CB,$FE,$20,$6D,$66,$7F,$C5,$1C,$18
	.dc.b	$7F,$9B,$CD,$CF,$D2,$DC,$8D,$AC,$7C,$0D,$BE,$67,$F8,$CD,$CB,$FD
	.dc.b	$D9,$FF,$6B,$5F,$C4,$6D,$9B,$87,$7F,$A7,$97,$FE,$3D,$8F,$11,$A7
	.dc.b	$FC,$02,$D7,$F7,$16,$FC,$15,$F5,$1E,$B6,$CF,$FB,$31,$AB,$4F,$EB
	.dc.b	$F5,$C7,$EA,$38,$1C,$6E,$57,$18,$E9,$F5,$1D,$7E,$1E,$4F,$58,$CB
	.dc.b	$FC,$0E,$FC,$91,$F0,$B1,$34,$C7,$C2,$79,$FE,$52,$5F,$9E,$3D,$46
	.dc.b	$1F,$46,$9F,$D9,$B0,$F8,$15,$A9,$F7,$7D,$1F,$C1,$F8,$E9,$18,$21
	.dc.b	$C8,$FE,$FB,$92,$37,$19,$BF,$99,$5F,$87,$F4,$1F,$88,$BE,$03,$0B
	.dc.b	$E0,$A9,$88,$98,$7F,$AA,$27,$83,$63,$F9,$42,$AE,$B1,$86,$C7,$FA
	.dc.b	$EA,$78,$BF,$EB,$F6,$F1,$7F,$AB,$97,$C1,$79,$FE,$D5,$F2,$D7,$C5
	.dc.b	$98,$7F,$B1,$27,$2D,$9C,$BF,$EC,$B8,$FF,$05,$39,$3F,$15,$F1,$9F
	.dc.b	$8D,$23,$F8,$DF,$A7,$AF,$C9,$E1,$4E,$33,$FC,$7C,$2B,$F6,$54,$F6
	.dc.b	$3F,$48,$FF,$88,$AB,$FC,$65,$F3,$2F,$88,$DB,$67,$5C,$DF,$E2,$0B
	.dc.b	$E3,$1F,$74,$E9,$DB,$F3,$9A,$7D,$08,$DB,$FD,$85,$AF,$38,$F2,$C7
	.dc.b	$4C,$3F,$C8,$8D,$AF,$F1,$7D,$FD,$0B,$FD,$77,$B9,$FB,$12,$3F,$B1
	.dc.b	$F2,$EB,$F1,$62,$E1,$F1,$27,$F7,$1F,$39,$FB,$9F,$93,$97,$FB,$98
	.dc.b	$CB,$FB,$B8,$CB,$F3,$E3,$F0,$13,$F8,$1D,$E4,$FD,$F8,$CB,$FE,$FA
	.dc.b	$E0,$FF,$22,$B6,$3F,$21,$18,$8F,$F7,$EB,$70,$FF,$7D,$B7,$3F,$F6
	.dc.b	$03,$67,$F3,$4F,$DD,$7F,$88,$7C,$B7,$F1,$32,$3F,$C4,$FE,$4E,$BF
	.dc.b	$87,$8B,$7F,$86,$9F,$E3,$5F,$4B,$FC,$6F,$E9,$E5,$FF,$1B,$8C,$BF
	.dc.b	$C7,$63,$2F,$E4,$63,$F1,$BF,$4F,$F9,$28,$FC,$39,$FC,$87,$F1,$16
	.dc.b	$3F,$D0,$F1,$46,$27,$62,$B9,$3F,$8D,$6B,$FE,$29,$6B,$FE,$72,$6B
	.dc.b	$FE,$7C,$6B,$CF,$5F,$29,$1F,$2D,$78,$BC,$5F,$E7,$5D,$12,$7F,$9D
	.dc.b	$3F,$CF,$7A,$7F,$9F,$7A,$7F,$CB,$49,$3F,$CC,$7A,$BF,$C6,$EF,$FA
	.dc.b	$83,$E4,$9F,$25,$FD,$16,$21,$F1,$3C,$34,$EA,$30,$FE,$75,$FE,$48
	.dc.b	$DC,$FF,$A4,$B7,$93,$F9,$A3,$FB,$59,$DF,$85,$FD,$8A,$3F,$93,$47
	.dc.b	$F4,$56,$BF,$D6,$9A,$F9,$FE,$4B,$FB,$A4,$7F,$1E,$AE,$BE,$1E,$27
	.dc.b	$FB,$C7,$6F,$F9,$67,$1F,$EB,$6D,$AF,$E8,$CD,$C3,$FD,$76,$FF,$96
	.dc.b	$A7,$FB,$0F,$D3,$FD,$6C,$CD,$27,$17,$F0,$9F,$B2,$7B,$DF,$87,$F9
	.dc.b	$57,$B9,$4F,$E3,$73,$AC,$FE,$67,$D4,$7F,$B0,$DB,$EE,$FF,$83,$E3
	.dc.b	$11,$FD,$3D,$7B,$FF,$C9,$3F,$57,$FF,$08,$CF,$F4,$66,$BF,$E6,$F6
	.dc.b	$BF,$E7,$96,$BF,$EC,$2F,$D6,$3F,$2F,$BF,$FB,$70,$7C,$73,$E4,$FF
	.dc.b	$40,$DA,$FC,$FD,$70,$E3,$FE,$DD,$F5,$3E,$0C,$3F,$C0,$ED,$8C,$FB
	.dc.b	$7F,$E2,$98,$FF,$1F,$5F,$F6,$9F,$21,$F2,$7F,$E2,$26,$C7,$F8,$15
	.dc.b	$AF,$F8,$69,$AF,$FB,$DB,$57,$C1,$FF,$8D,$57,$C5,$D3,$11,$FE,$39
	.dc.b	$FA,$87,$CD,$52,$5F,$E7,$7E,$A1,$E6,$7F,$8F,$DB,$1F,$E1,$76,$BF
	.dc.b	$E2,$C6,$BF,$EE,$CE,$C4,$62,$3F,$C2,$4F,$F0,$77,$FC,$10,$F2,$E0
	.dc.b	$FF,$16,$B7,$83,$7E,$57,$F9,$3F,$E9,$FF,$93,$5F,$F8,$3F,$EA,$3F
	.dc.b	$CC,$4D,$F8,$4D,$7A,$F2,$3F,$C5,$95,$C1,$F2,$35,$FF,$12,$35,$66
	.dc.b	$8E,$F9,$FF,$90,$1B,$97,$0B,$F2,$7B,$7F,$E6,$3F,$A6,$FE,$BD,$8F
	.dc.b	$2F,$5D,$F1,$DF,$E6,$76,$C7,$F0,$99,$DE,$67,$FB,$5C,$78,$9B,$19
	.dc.b	$F6,$7F,$CC,$B1,$8F,$A8,$FF,$19,$57,$F1,$36,$BF,$E3,$36,$EB,$FE
	.dc.b	$58,$5F,$03,$FE,$79,$73,$AF,$F9,$15,$B5,$F0,$9B,$89,$E3,$1F,$CF
	.dc.b	$B0,$FD,$26,$08,$7D,$3F,$DD,$E2,$FF,$13,$C6,$1D,$3F,$D2,$BD,$3B
	.dc.b	$3F,$83,$BE,$73,$B3,$A3,$88,$E9,$3C,$2B,$99,$DA,$39,$78,$3F,$E7
	.dc.b	$7B,$E4,$7C,$D7,$11,$CA,$3D,$0D,$77,$7D,$EB,$CE,$D8,$F7,$47,$73
	.dc.b	$C3,$7E,$38,$F3,$DF,$69,$E2,$F8,$FE,$6F,$B1,$23,$38,$FC,$98,$F4
	.dc.b	$BE,$48,$EB,$54,$F0,$DF,$77,$D9,$39,$69,$F0,$27,$CC,$F0,$35,$87
	.dc.b	$AD,$D9,$F5,$0A,$79,$0C,$18,$3E,$9F,$D0,$3E,$A3,$FC,$E7,$28,$C4
	.dc.b	$F0,$8C,$4E,$A7,$33,$FE,$7A,$F9,$1F,$F2,$72,$69,$1F,$F2,$93,$ED
	.dc.b	$C1,$F3,$6E,$0F,$81,$5F,$91,$AD,$4E,$63,$F0,$26,$4F,$C1,$D7,$2A
	.dc.b	$E3,$1E,$13,$33,$C7,$6F,$E2,$D1,$C7,$5F,$E3,$17,$C6,$11,$A7,$93
	.dc.b	$AA,$F1,$39,$C3,$CB,$87,$81,$B5,$E6,$3C,$58,$71,$9D,$47,$91,$B1
	.dc.b	$98,$E1,$3C,$A3,$B4,$79,$34,$FF,$7E,$E1,$E2,$7E,$DD,$7E,$3D,$CD
	.dc.b	$DA,$FE,$43,$F5,$0F,$2B,$FB,$14,$FF,$B2,$63,$0F,$F8,$85,$FE,$E2
	.dc.b	$D8,$DC,$F2,$7E,$93,$FD,$0E,$78,$BF,$EF,$26,$B9,$8D,$BF,$ED,$7C
	.dc.b	$3F,$8B,$D7,$FB,$95,$B1,$FE,$FA,$6B,$FD,$3D,$AF,$1E,$3F,$1F,$1B
	.dc.b	$EA,$1F,$3B,$8B,$E7,$BF,$FB,$F2,$73,$8F,$F8,$F7,$EA,$1F,$07,$C9
	.dc.b	$B1,$9F,$4F,$13,$23,$FD,$F5,$3A,$D3,$F5,$99,$A3,$11,$FD,$95,$AF
	.dc.b	$FB,$D5,$E1,$8F,$E6,$67,$FB,$8F,$D0,$F5,$AF,$F7,$77,$86,$8F,$68
	.dc.b	$F7,$E9,$8F,$8F,$B9,$7F,$95,$E7,$4F,$EF,$30,$E5,$5F,$43,$0F,$EC
	.dc.b	$35,$A7,$D2,$D7,$FC,$1D,$F2,$5B,$AF,$F1,$8B,$7C,$B3,$4D,$7D,$47
	.dc.b	$F9,$0B,$1F,$03,$6F,$E2,$B5,$FF,$04,$35,$FE,$99,$6F,$A8,$FF,$71
	.dc.b	$57,$A5,$B3,$F3,$9A,$FF,$84,$5A,$F2,$EF,$7C,$8F,$A5,$B5,$F0,$DB
	.dc.b	$3F,$61,$AF,$F2,$5F,$D0,$9F,$3D,$F1,$E1,$18,$AF,$E4,$F1,$89,$D4
	.dc.b	$E6,$3F,$95,$FC,$8B,$E4,$B9,$36,$F8,$B6,$77,$3D,$FD,$C7,$D0,$52
	.dc.b	$5F,$82,$32,$FD,$DF,$80,$7C,$7F,$D7,$6C,$66,$30,$78,$BE,$56,$47
	.dc.b	$89,$B8,$FF,$A3,$FD,$4F,$C0,$7C,$37,$F9,$22,$BF,$D2,$91,$83,$39
	.dc.b	$48,$FF,$4C,$36,$BF,$A5,$FC,$E3,$DD,$F9,$4D,$AF,$92,$D9,$DC,$F8
	.dc.b	$3C,$29,$5F,$EA,$36,$CF,$FA,$8F,$E8,$BC,$A4,$9F,$F5,$43,$DB,$E6
	.dc.b	$E4,$76,$6C,$FF,$AA,$FE,$9F,$DA,$49,$FF,$58,$3F,$7D,$FE,$BC,$7F
	.dc.b	$B7,$8B,$0F,$03,$FD,$11,$F9,$43,$E6,$FF,$91,$CF,$F5,$7F,$3E,$DF
	.dc.b	$08,$FE,$E0,$D7,$FC,$4A,$D7,$FC,$58,$D7,$93,$FE,$E0,$A7,$A9,$D2
	.dc.b	$BF,$98,$E0,$9C,$D7,$FB,$53,$AB,$F9,$1F,$17,$FA,$6E,$3F,$DA,$F7
	.dc.b	$89,$91,$B3,$25,$C7,$DC,$FE,$D3,$EE,$DE,$C7,$F2,$E7,$F3,$19,$7F
	.dc.b	$19,$73,$3F,$EE,$77,$2F,$19,$E1,$D3,$C5,$B1,$FE,$4B,$AF,$EF,$5D
	.dc.b	$4F,$CA,$BF,$55,$18,$6C,$FF,$94,$67,$5F,$01,$FD,$49,$E1,$DD,$6B
	.dc.b	$F6,$59,$FF,$94,$A7,$37,$FB,$49,$CC,$F3,$6A,$9D,$BF,$08,$C4,$7F
	.dc.b	$AB,$A7,$D0,$71,$AF,$07,$C0,$F8,$7E,$9E,$4E,$AF,$A7,$1B,$A7,$31
	.dc.b	$D7,$2F,$81,$78,$D3,$E1,$DE,$F0,$E5,$5A,$7A,$78,$6E,$3F,$F0,$B5
	.dc.b	$FF,$4B,$7F,$A4,$4E,$9C,$BF,$CB,$9C,$E3,$ED,$AE,$5E,$7F,$E2,$3C
	.dc.b	$8F,$C4,$B6,$FD,$54,$FC,$93,$EE,$BF,$E8,$36,$E2,$FA,$7F,$F5,$1D
	.dc.b	$70,$9D,$D7,$1E,$4F,$E5,$91,$FC,$86,$3F,$CA,$8D,$7F,$D1,$2D,$7B
	.dc.b	$47,$E8,$2B,$B5,$75,$8C,$47,$E8,$3E,$1B,$F4,$3F,$1F,$6B,$E1,$36
	.dc.b	$FE,$06,$A8,$C6,$5F,$25,$FC,$67,$CB,$7E,$D2,$BE,$B3,$67,$F2,$EB
	.dc.b	$E5,$B7,$33,$0F,$AF,$3B,$7F,$30,$ED,$CC,$7F,$08,$F6,$BF,$C5,$AF
	.dc.b	$E3,$CF,$F3,$37,$9F,$FB,$EF,$C5,$5E,$DE,$E2,$5F,$F9,$B5,$FE,$AD
	.dc.b	$7C,$AF,$FC,$FC,$D7,$FD,$66,$F6,$CB,$FC,$D0,$3F,$14,$78,$9D,$2B
	.dc.b	$F4,$13,$98,$FE,$01,$EC,$4C,$7F,$7D,$F0,$B6,$FF,$15,$E0,$AB,$18
	.dc.b	$7F,$05,$AB,$7F,$91,$FA,$BF,$E2,$A7,$F1,$6F,$A6,$B1,$F9,$78,$B8
	.dc.b	$FD,$69,$B7,$F9,$DE,$EF,$EA,$A4,$BF,$F0,$F7,$27,$E9,$EF,$FC,$66
	.dc.b	$DA,$FD,$27,$C0,$FF,$24,$BF,$E4,$AF,$F9,$23,$E9,$BF,$79,$7F,$A3
	.dc.b	$6D,$7F,$8D,$57,$81,$BF,$06,$65,$FE,$2D,$9E,$37,$FE,$E0,$EF,$46
	.dc.b	$23,$FC,$3A,$D5,$1A,$F0,$BF,$C5,$D3,$FC,$4A,$31,$5D,$67,$27,$9F
	.dc.b	$C4,$F4,$36,$3F,$81,$D7,$E9,$2F,$FC,$A6,$D7,$C8,$D7,$C7,$DB,$E0
	.dc.b	$3E,$17,$C4,$7C,$36,$D7,$4F,$AA,$F8,$C6,$2B,$E3,$38,$8F,$84,$FA
	.dc.b	$88,$FF,$94,$7F,$58,$FF,$21,$3F,$15,$B3,$FE,$40,$6B,$F1,$1B,$99
	.dc.b	$FF,$25,$74,$B8,$FE,$5F,$AD,$FE,$C1,$8C,$5F,$1F,$98,$C5,$6B,$1F
	.dc.b	$F6,$42,$99,$F5,$3F,$E5,$AF,$07,$31,$8F,$77,$F7,$8F,$FB,$5E,$B8
	.dc.b	$2F,$0B,$FD,$01,$97,$82,$18,$73,$78,$1C,$F0,$EA,$EB,$0E,$B5,$C6
	.dc.b	$F9,$E1,$DB,$0E,$53,$D5,$B1,$9A,$E2,$F9,$2B,$8D,$F7,$3C,$27,$7C
	.dc.b	$3C,$37,$CA,$3E,$14,$F0,$74,$70,$8D,$38,$C3,$CD,$F0,$5F,$32,$3B
	.dc.b	$47,$97,$1F,$1C,$63,$1F,$25,$6E,$39,$47,$A7,$1F,$1D,$70,$9F,$99
	.dc.b	$5B,$C3,$C2,$49,$D6,$3E,$78,$C5,$F2,$3E,$6C,$66,$7A,$4E,$AF,$E1
	.dc.b	$CE,$6B,$84,$7B,$EB,$AD,$F8,$5E,$24,$9E,$B1,$97,$8C,$FE,$10,$F9
	.dc.b	$25,$BF,$C1,$36,$BF,$16,$E7,$D5,$FF,$01,$3C,$4F,$C4,$B8,$75,$D5
	.dc.b	$FE,$8E,$AF,$F3,$53,$67,$8E,$7F,$85,$6E,$7F,$89,$6D,$FE,$7B,$9B
	.dc.b	$FA,$0C,$E6,$FF,$07,$5A,$AF,$0C,$7E,$1A,$B7,$1F,$A6,$9D,$61,$E2
	.dc.b	$8F,$C3,$46,$27,$71,$C8,$CD,$76,$3A,$34,$C9,$30,$FC,$F4,$F2,$BE
	.dc.b	$D1,$ED,$AE,$0E,$AA,$E1,$EB,$79,$3F,$A9,$BE,$58,$7A,$09,$7E,$2A
	.dc.b	$D9,$ED,$79,$BF,$1C,$A7,$AE,$FC,$4F,$E9,$6B,$AC,$EB,$0F,$3E,$1E
	.dc.b	$1A,$C6,$1E,$07,$81,$FC,$3B,$87,$F8,$E7,$D5,$7C,$E7,$FA,$3F,$C7
	.dc.b	$DE,$73,$9E,$45,$53,$FD,$13,$7F,$CA,$5A,$FF,$29,$6D,$F6,$8F,$F0
	.dc.b	$D1,$FD,$42,$DF,$57,$FE,$DA,$9F,$EA,$ED,$8C,$CF,$F5,$D8,$C1,$FC
	.dc.b	$8B,$B3,$FB,$23,$79,$3F,$AE,$B6,$7B,$72,$DC,$3F,$B1,$E1,$44,$53
	.dc.b	$0C,$9F,$ED,$8F,$8C,$8F,$F6,$FF,$17,$96,$52,$3F,$B0,$C7,$53,$D4
	.dc.b	$47,$81,$FD,$DF,$C4,$F7,$11,$C1,$0F,$F0,$13,$E7,$3D,$1F,$11,$E9
	.dc.b	$E5,$F5,$70,$4A,$FE,$E3,$19,$7F,$C0,$8F,$CB,$23,$FE,$07,$F4,$3E
	.dc.b	$11,$1F,$EF,$3E,$B5,$7F,$C2,$51,$78,$3F,$C2,$53,$F2,$FF,$18,$7B
	.dc.b	$34,$97,$FE,$84,$87,$B5,$88,$FF,$3F,$47,$07,$53,$CE,$7E,$C4,$7F
	.dc.b	$98,$F8,$FE,$D6,$0F,$C5,$9F,$2D,$86,$D7,$FA,$13,$D0,$5F,$26,$C7
	.dc.b	$F1,$D3,$A1,$5C,$BF,$C8,$5A,$FF,$2B,$73,$7A,$6C,$FF,$72,$6C,$FF
	.dc.b	$BC,$FF,$16,$7C,$07,$E4,$7D,$87,$EA,$5F,$E9,$3A,$FF,$54,$B5,$FF
	.dc.b	$58,$37,$1F,$F6,$8A,$F2,$79,$39,$CB,$FD,$A0,$DA,$FF,$63,$36,$FB
	.dc.b	$D7,$E7,$FE,$0D,$F1,$3F,$9F,$C2,$61,$FE,$B8,$9C,$B6,$72,$FF,$94
	.dc.b	$67,$94,$7C,$BF,$27,$EA,$FC,$A7,$D9,$AF,$F4,$33,$73,$F8,$D3,$B9
	.dc.b	$F9,$CF,$D1,$7F,$D3,$CD,$D8,$D3,$84,$7F,$A4,$63,$2F,$FB,$02,$9F
	.dc.b	$01,$ED,$DB,$DD,$F1,$52,$BD,$91,$97,$DD,$19,$7F,$D9,$DA,$7D,$67
	.dc.b	$FD,$B7,$FA,$13,$E8,$7E,$C3,$63,$31,$8F,$A2,$CB,$23,$27,$28,$E3
	.dc.b	$F4,$7C,$BE,$B6,$4F,$68,$CB,$FD,$BF,$8D,$82,$7F,$2D,$18,$6C,$66
	.dc.b	$78,$35,$FF,$7C,$37,$83,$F3,$F5,$C6,$7D,$BF,$5B,$FB,$7D,$BF,$71
	.dc.b	$F1,$32,$7F,$6D,$19,$7F,$71,$19,$7F,$6F,$1F,$8B,$FA,$F9,$1F,$89
	.dc.b	$7F,$1D,$F1,$F8,$8F,$9B,$17,$1F,$D8,$F9,$2E,$05,$FE,$FB,$F2,$57
	.dc.b	$D4,$7F,$00,$F8,$8F,$E9,$F9,$7F,$BF,$9A,$FF,$B4,$E7,$6F,$F9,$A7
	.dc.b	$1F,$DC,$C6,$23,$FC,$34,$D7,$FC,$48,$D7,$FC,$D2,$D7,$FC,$E4,$F1
	.dc.b	$8E,$75,$E5,$3C,$9E,$67,$EA,$FE,$5F,$FA,$B6,$46,$93,$0F,$E1,$DF
	.dc.b	$4E,$F1,$BF,$40,$D8,$CC,$60,$E7,$F1,$32,$3F,$C6,$5B,$7C,$78,$C4
	.dc.b	$7F,$8F,$63,$E5,$7C,$6F,$88,$91,$C3,$14,$78,$3C,$A9,$3E,$46,$BC
	.dc.b	$E5,$F0,$9E,$C2,$47,$37,$DF,$F0,$B2,$3B,$35,$FF,$24,$67,$FD,$A9
	.dc.b	$F8,$3A,$FC,$02,$3E,$CF,$B5,$FD,$CB,$79,$7F,$C0,$11,$88,$F7,$35
	.dc.b	$FF,$2E,$7D,$2F,$FB,$ED,$FE,$74,$D7,$FC,$B0,$D7,$78,$7D,$0F,$0E
	.dc.b	$47,$D4,$6B,$FE,$5F,$FD,$0B,$ED,$3F,$D1,$5A,$E7,$8B,$F8,$C4,$62
	.dc.b	$3F,$DD,$9F,$4C,$DF,$FC,$30,$D8,$FB,$2D,$7E,$C3,$5F,$F7,$A7,$99
	.dc.b	$FE,$B7,$6F,$B5,$FE,$21,$8C,$47,$FB,$17,$8F,$FA,$FC,$FF,$89,$7D
	.dc.b	$5F,$E8,$93,$F8,$66,$BF,$85,$6B,$FF,$00,$5A,$F3,$FA,$67,$CA,$FF
	.dc.b	$8B,$DB,$1F,$90,$6B,$F8,$F6,$BF,$ED,$0B,$7D,$33,$EB,$BF,$C7,$0D
	.dc.b	$9F,$CC,$B5,$FC,$C3,$5E,$4E,$7B,$3F,$82,$B7,$CE,$F8,$0D,$95,$F8
	.dc.b	$3E,$BF,$FC,$75,$F4,$EF,$86,$FF,$1E,$7D,$47,$F9,$1A,$D9,$7F,$10
	.dc.b	$8C,$3F,$C4,$AB,$84,$FF,$09,$CB,$E7,$46,$23,$F5,$CD,$7E,$03,$5D
	.dc.b	$DF,$F9,$5E,$38,$57,$58,$F2,$9E,$6E,$6F,$F2,$87,$D3,$3D,$BF,$F2
	.dc.b	$97,$A5,$FD,$73,$1F,$3E,$5F,$7E,$31,$8F,$F2,$8B,$E1,$5E,$73,$E1
	.dc.b	$B8,$7F,$82,$35,$FA,$CD,$7F,$14,$D7,$E5,$FD,$47,$F0,$07,$F8,$7B
	.dc.b	$5F,$E1,$CD,$7F,$00,$D7,$F2,$8E,$E7,$F9,$EB,$6F,$3E,$07,$F9,$BE
	.dc.b	$3F,$89,$D7,$A4,$5B,$E6,$FE,$BC,$3E,$45,$F2,$C8,$8B,$EA,$31,$1B
	.dc.b	$8A,$65,$E1,$3C,$9E,$57,$DF,$3C,$D7,$28,$C4,$BE,$9F,$83,$E2,$C3
	.dc.b	$BC,$62,$BB,$4E,$DE,$58,$FA,$A7,$8D,$F4,$78,$5F,$2A,$F3,$B6,$FC
	.dc.b	$B7,$C6,$79,$E1,$F0,$6F,$8D,$59,$E2,$D3,$8C,$D7,$E6,$D7,$CC,$C7
	.dc.b	$8C,$7C,$73,$E5,$E1,$F2,$A3,$CF,$3B,$BF,$0D,$7B,$4D,$9D,$5F,$83
	.dc.b	$87,$0C,$3C,$18,$FB,$E7,$CB,$5C,$4F,$94,$6A,$FD,$CF,$A0,$FB,$65
	.dc.b	$2C,$FD,$46,$C7,$E1,$9C,$F8,$6F,$A8,$FF,$3E,$57,$F9,$11,$B1,$99
	.dc.b	$EA,$7E,$41,$C3,$FE,$83,$F8,$EF,$F2,$A2,$69,$1F,$F2,$C3,$F3,$71
	.dc.b	$79,$DF,$E8,$CC,$FF,$0F,$5F,$97,$AD,$4E,$73,$F0,$C6,$38,$3E,$B5
	.dc.b	$6C,$F8,$D8,$74,$8C,$C7,$8C,$FB,$07,$F1,$CB,$FF,$4D,$EB,$E0,$8D
	.dc.b	$3F,$98,$AE,$51,$C0,$ED,$8E,$6B,$C9,$1F,$43,$0F,$23,$F5,$CD,$E3
	.dc.b	$FE,$ED,$9D,$C7,$EC,$27,$75,$E1,$6D,$7B,$A7,$D7,$5C,$1E,$67,$F8
	.dc.b	$DD,$C3,$ED,$AF,$26,$3F,$9B,$E8,$7D,$43,$E2,$3F,$D7,$39,$FF,$92
	.dc.b	$A1,$F1,$38,$BF,$EC,$D1,$B7,$F9,$35,$F2,$9D,$63,$FD,$2A,$B4,$FF
	.dc.b	$30,$31,$1A,$9F,$0E,$16,$BB,$DF,$4C,$9E,$9F,$FB,$1B,$1F,$EA,$B1
	.dc.b	$89,$FF,$1C,$B6,$3F,$C7,$0D,$7F,$B5,$39,$CB,$FC,$CB,$F5,$1F,$DD
	.dc.b	$2B,$B5,$F8,$E3,$17,$E7,$C9,$F5,$1F,$E3,$8B,$F7,$B6,$74,$FF,$90
	.dc.b	$A3,$3C,$FE,$D8,$D4,$FF,$7B,$6B,$FE,$0F,$6B,$FD,$C1,$B5,$E0,$9F
	.dc.b	$8D,$F5,$1F,$CE,$6F,$A9,$F4,$1C,$5F,$9F,$F3,$9F,$F7,$9C,$FD,$47
	.dc.b	$FB,$13,$4F,$A0,$E2,$38,$75,$FC,$DC,$13,$99,$FF,$09,$35,$FF,$13
	.dc.b	$35,$FF,$03,$F8,$AF,$8B,$FB,$8D,$C3,$C5,$86,$27,$F0,$1F,$52,$F8
	.dc.b	$FA,$63,$E4,$6D,$FE,$33,$EA,$7E,$66,$5F,$E7,$36,$BF,$E2,$56,$BF
	.dc.b	$E3,$96,$BF,$E1,$AE,$E7,$C3,$7E,$3D,$7C,$35,$8C,$3F,$93,$E8,$FF
	.dc.b	$72,$E1,$F9,$C6,$DF,$E0,$D7,$CD,$6F,$D4,$FC,$FC,$BE,$A7,$6F,$F2
	.dc.b	$E6,$BF,$DA,$9A,$FF,$6F,$6B,$C7,$E2,$FF,$53,$D4,$FB,$8F,$D4,$AC
	.dc.b	$B9,$4E,$A7,$31,$FE,$59,$F7,$1F,$67,$FA,$D6,$C7,$ED,$9F,$F1,$D3
	.dc.b	$63,$5F,$57,$F8,$19,$34,$93,$FE,$E6,$DF,$CB,$F5,$EF,$9B,$F8,$3F
	.dc.b	$13,$F3,$AB,$FC,$51,$A2,$B5,$1E,$1E,$DF,$21,$E5,$F6,$6C,$79,$7C
	.dc.b	$7F,$11,$1F,$F5,$73,$E5,$F5,$BD,$9E,$AF,$B5,$B3,$ED,$F3,$FE,$01
	.dc.b	$F0,$7E,$76,$C7,$C1,$F9,$DF,$71,$1F,$F5,$D3,$FA,$1F,$C8,$BF,$DC
	.dc.b	$B1,$F1,$F9,$BF,$8B,$46,$1F,$E0,$9C,$7F,$E2,$79,$FF,$17,$7D,$CB
	.dc.b	$E8,$7F,$06,$D8,$F8,$7F,$97,$F1,$E2,$FB,$58,$7F,$8B,$EB,$4F,$F2
	.dc.b	$50,$E1,$FE,$EE,$D7,$FC,$66,$D7,$95,$FF,$25,$78,$BF,$8A,$F9,$3F
	.dc.b	$34,$75,$9E,$F5,$FC,$FA,$72,$F9,$3C,$96,$6F,$92,$F8,$27,$C2,$F9
	.dc.b	$1B,$A7,$F4,$09,$D3,$7E,$4F,$39,$FF,$29,$94,$7C,$AF,$AF,$95,$C7
	.dc.b	$3B,$7B,$9F,$9F,$F2,$B8,$D3,$C9,$8A,$3F,$A6,$F5,$3F,$45,$FF,$80
	.dc.b	$19,$7F,$9D,$E3,$1F,$23,$A9,$FF,$64,$4E,$5D,$E5,$FD,$4E,$B7,$3F
	.dc.b	$E1,$86,$C7,$F8,$99,$AF,$F9,$89,$AF,$F9,$9D,$DE,$3F,$71,$F2,$79
	.dc.b	$BE,$1C,$3C,$4F,$E2,$BE,$76,$61,$F2,$7E,$1E,$E3,$F1,$E7,$F1,$FF
	.dc.b	$3B,$8C,$BE,$0B,$F9,$2F,$32,$61,$F4,$7E,$5E,$E1,$F3,$E7,$F2,$FE
	.dc.b	$FE,$30,$F9,$2F,$F1,$6F,$AD,$76,$D3,$F8,$7F,$57,$C6,$C3,$F2,$53
	.dc.b	$AC,$BF,$D1,$4D,$8F,$F1,$EB,$5F,$F2,$5B,$5F,$F4,$3B,$5F,$F4,$63
	.dc.b	$C7,$0F,$F0,$6D,$7F,$45,$9E,$F7,$E2,$DB,$F5,$FF,$35,$31,$FC,$77
	.dc.b	$D7,$5B,$FC,$64,$FE,$D3,$DD,$C6,$5F,$86,$7F,$01,$EA,$4C,$3F,$01
	.dc.b	$F4,$D7,$0F,$C0,$73,$2B,$F0,$DE,$8E,$71,$FB,$53,$FD,$01,$BF,$4F
	.dc.b	$F8,$B3,$F1,$9F,$51,$72,$FC,$64,$7F,$85,$7A,$7F,$28,$DB,$34,$93
	.dc.b	$F9,$7E,$97,$D4,$73,$D3,$F2,$91,$89,$FD,$77,$D5,$7F,$12,$3F,$89
	.dc.b	$FD,$75,$8C,$9F,$B3,$F9,$3C,$5F,$EC,$5F,$F1,$77,$D4,$B5,$A7,$F1
	.dc.b	$CF,$8A,$B7,$FE,$3A,$9D,$3F,$BD,$CF,$38,$7E,$ED,$E7,$E9,$FF,$81
	.dc.b	$EB,$F9,$EC,$FF,$A3,$3C,$84,$62,$3F,$D9,$31,$89,$E0,$79,$36,$E7
	.dc.b	$87,$85,$F3,$F5,$7A,$3D,$8B,$3E,$56,$E1,$FE,$35,$C5,$18,$E3,$FE
	.dc.b	$35,$18,$AF,$67,$27,$9C,$8F,$FA,$5B,$3F,$3C,$63,$0E,$9E,$D7,$C8
	.dc.b	$8F,$92,$FC,$AE,$7F,$80,$4B,$E5,$82,$31,$7E,$5F,$7B,$85,$7C,$D7
	.dc.b	$E9,$91,$FA,$13,$67,$84,$7D,$3E,$8C,$67,$EC,$7E,$DF,$C4,$7F,$4A
	.dc.b	$9F,$9B,$8F,$F9,$A2,$FF,$A6,$8D,$3E,$9D,$6E,$3F,$DF,$D1,$8A,$FE
	.dc.b	$8B,$1B,$AE,$5F,$14,$E5,$DC,$3E,$6F,$FA,$CC,$E6,$B8,$63,$8B,$DD
	.dc.b	$59,$0F,$AA,$7C,$C3,$D4,$FF,$5D,$5E,$AB,$53,$97,$FD,$8D,$19,$F0
	.dc.b	$FF,$C3,$31,$83,$AD,$F4,$BE,$71,$8B,$E5,$97,$9E,$3A,$CE,$9F,$E4
	.dc.b	$7F,$27,$E7,$72,$75,$2C,$F1,$BE,$D1,$9C,$3C,$73,$9C,$3B,$E3,$EB
	.dc.b	$BD,$CE,$71,$F0,$5F,$0A,$D5,$F0,$C3,$D1,$5E,$5C,$38,$D7,$C1,$C7
	.dc.b	$CD,$18,$AD,$9E,$77,$0F,$B3,$5E,$58,$75,$8F,$5D,$F8,$30,$F1,$E1
	.dc.b	$D2,$B5,$87,$23,$9E,$3E,$7C,$3B,$E5,$EB,$7C,$15,$D6,$75,$1D,$8D
	.dc.b	$E3,$DD,$F9,$18,$77,$C7,$CA,$48,$CE,$57,$1E,$C4,$BF,$85,$19,$7A
	.dc.b	$4F,$E2,$4F,$B2,$5C,$BE,$BB,$72,$FC,$63,$9F,$13,$F9,$8B,$C4,$FC
	.dc.b	$B3,$87,$5D,$59,$72,$DA,$FF,$50,$3F,$5E,$79,$E3,$F8,$76,$E1,$F8
	.dc.b	$3E,$AF,$F7,$E3,$F8,$58,$CE,$1F,$9A,$6B,$D8,$FB,$93,$C1,$D5,$F7
	.dc.b	$23,$A2,$DF,$86,$2D,$F8,$66,$9C,$8F,$C4,$5F,$83,$0F,$F3,$E4,$68
	.dc.b	$F1,$5F,$37,$B1,$C2,$F5,$3F,$0A,$33,$3B,$25,$FB,$EB,$F3,$98,$7A
	.dc.b	$EB,$AC,$E5,$E8,$72,$71,$5C,$A3,$D0,$FE,$CE,$79,$5F,$86,$35,$3D
	.dc.b	$0F,$E0,$B3,$AC,$3D,$C4,$C3,$C7,$3F,$C2,$38,$7F,$AC,$FC,$3F,$CF
	.dc.b	$7F,$AD,$FC,$1D,$E3,$36,$65,$7F,$2A,$AE,$2D,$A4,$FF,$1D,$DB,$F9
	.dc.b	$BB,$67,$F9,$6F,$D5,$7F,$21,$E6,$FE,$B3,$F5,$56,$FF,$98,$B6,$BF
	.dc.b	$9B,$55,$C7,$FB,$7B,$71,$34,$C6,$92,$BF,$BF,$46,$5F,$F0,$04,$65
	.dc.b	$FE,$F5,$3F,$EF,$B7,$FA,$EC,$7E,$86,$BF,$BE,$B6,$5E,$1F,$87,$8A
	.dc.b	$5F,$F8,$2E,$3C,$86,$C8,$FF,$68,$79,$65,$E9,$F8,$99,$3F,$E1,$38
	.dc.b	$F6,$18,$7F,$C0,$11,$87,$FA,$6E,$9F,$E9,$9F,$2F,$FD,$FB,$18,$AF
	.dc.b	$EB,$F0,$FA,$6F,$90,$D9,$D1,$FD,$9D,$E2,$AB,$A7,$C4,$7C,$19,$CC
	.dc.b	$FF,$87,$1A,$F6,$8D,$7C,$7F,$6F,$85,$93,$DB,$75,$AB,$C6,$49,$D4
	.dc.b	$65,$BF,$4B,$E8,$FA,$2F,$4F,$CF,$C9,$DC,$65,$A4,$3D,$5A,$FC,$2F
	.dc.b	$A3,$FC,$57,$9E,$FC,$B9,$E6,$61,$79,$F4,$C4,$4C,$3F,$D4,$D3,$C1
	.dc.b	$B1,$FC,$96,$BA,$36,$75,$3F,$EA,$F6,$F0,$FF,$A9,$DB,$8F,$FA,$AD
	.dc.b	$FE,$ED,$19,$F5,$7C,$9B,$4B,$FF,$5F,$52,$7D,$04,$8C,$BE,$BF,$5B
	.dc.b	$EF,$F8,$9F,$80,$F9,$29,$5F,$3A,$32,$D2,$1F,$20,$FC,$27,$CD,$FE
	.dc.b	$C5,$7D,$33,$D7,$C7,$CE,$7C,$89,$FC,$58,$70,$FF,$BF,$67,$83,$A6
	.dc.b	$BF,$D4,$5A,$FF,$5B,$6B,$FE,$25,$6B,$FE,$2C,$78,$EB,$FE,$40,$6E
	.dc.b	$DB,$BF,$2E,$FF,$A0,$F8,$CF,$D0,$FC,$84,$C7,$F1,$F1,$96,$90,$F8
	.dc.b	$5D,$3F,$98,$F9,$0F,$D4,$FC,$6F,$EA,$BE,$3E,$57,$E6,$23,$2D,$21
	.dc.b	$CB,$95,$F1,$FF,$8F,$F8,$8F,$C8,$7C,$4C,$AF,$BB,$19,$7F,$1E,$FE
	.dc.b	$2C,$8E,$0D,$90,$FC,$77,$D3,$7E,$FB,$E3,$FF,$7F,$F2,$12,$7F,$5B
	.dc.b	$19,$69,$0F,$37,$AB,$FC,$B7,$E9,$57,$FC,$23,$D6,$95,$FB,$B9,$CB
	.dc.b	$67,$81,$1F,$DD,$9F,$89,$FB,$67,$67,$F3,$BE,$96,$8F,$13,$0D,$BF
	.dc.b	$C5,$DF,$06,$DF,$F8,$4D,$46,$5F,$F0,$D3,$5F,$F1,$23,$5F,$F3,$73
	.dc.b	$BB,$FF,$43,$B7,$A7,$35,$E2,$F2,$3E,$16,$1F,$0F,$82,$63,$F0,$63
	.dc.b	$E3,$7C,$7F,$F4,$5D,$4D,$24,$FF,$AB,$E1,$F1,$9F,$34,$91,$F2,$FE
	.dc.b	$3F,$E8,$7C,$6F,$D1,$F8,$F9,$3F,$3E,$3E,$A7,$C6,$F7,$24,$75,$7C
	.dc.b	$5E,$2F,$D9,$24,$7D,$7F,$8E,$FD,$D7,$D3,$FF,$A6,$E3,$FA,$0F,$1B
	.dc.b	$E5,$3F,$D7,$51,$86,$BF,$C0,$C4,$F1,$7F,$4B,$AF,$F3,$C7,$DD,$D9
	.dc.b	$F0,$C8,$FF,$B3,$FF,$01,$71,$7F,$79,$6C,$7F,$45,$6B,$FB,$07,$F1
	.dc.b	$AF,$87,$E9,$FF,$08,$FF,$15,$C9,$FA,$CB,$D0,$FC,$63,$67,$F8,$9D
	.dc.b	$70,$AF,$F6,$2F,$E3,$8F,$8A,$FC,$53,$6B,$FC,$BC,$D7,$FC,$C8,$D7
	.dc.b	$FD,$F7,$F8,$0B,$DF,$FF,$09,$36,$3F,$AF,$35,$FE,$12,$FF,$88,$9F
	.dc.b	$EC,$5F,$54,$F6,$FF,$C3,$4F,$F8,$2A,$33,$EE,$62,$7F,$C3,$DE,$67
	.dc.b	$F2,$A9,$FF,$63,$79,$88,$C4,$7F,$B7,$1A,$FF,$B8,$5B,$1F,$E3,$2C
	.dc.b	$FF,$C5,$1F,$03,$ED,$D5,$E1,$7F,$B9,$23,$0D,$F8,$FF,$F1,$77,$D5
	.dc.b	$3B,$29,$33,$FE,$F9,$F5,$4F,$2B,$FC,$68,$D8,$FF,$46,$68,$8C,$47
	.dc.b	$FA,$5E,$31,$3F,$E2,$D9,$CC,$FB,$97,$97,$D2,$64,$F8,$23,$FC,$85
	.dc.b	$F5,$2F,$47,$FC,$89,$F5,$5F,$95,$DF,$DB,$6C,$93,$9D,$BF,$DE,$B0
	.dc.b	$C3,$95,$7F,$85,$DA,$F9,$DA,$FD,$26,$BC,$9A,$F7,$78,$DF,$8A,$7B
	.dc.b	$61,$E8,$3D,$3E,$9F,$F9,$6F,$EA,$7C,$F8,$7F,$50,$F6,$3F,$63,$87
	.dc.b	$16,$E5,$E8,$AD,$CF,$57,$8D,$7E,$19,$AE,$E7,$FC,$B5,$18,$8F,$89
	.dc.b	$F5,$4F,$6F,$FC,$D8,$BD,$BA,$62,$24,$FF,$8E,$E7,$4D,$9C,$BF,$C1
	.dc.b	$EB,$FC,$2D,$82,$1F,$53,$FD,$89,$C7,$AF,$F7,$9D,$46,$5F,$EC,$FC
	.dc.b	$3C,$5E,$33,$6F,$AC,$60,$EC,$E2,$B9,$4F,$59,$D1,$D0,$EE,$F3,$F7
	.dc.b	$7F,$CC,$11,$AA,$E5,$3E,$13,$95,$74,$BF,$23,$E7,$9D,$D7,$8A,$F9
	.dc.b	$CE,$E3,$B9,$F0,$2B,$8D,$F8,$70,$CC,$B0,$F6,$FB,$7C,$63,$FC,$AF
	.dc.b	$5E,$3B,$F8,$93,$E1,$C7,$D2,$DC,$39,$D7,$A0,$E1,$7D,$B0,$F3,$BD
	.dc.b	$EF,$CF,$3A,$8F,$8E,$F1,$7D,$95,$C6,$7E,$3C,$7D,$53,$94,$57,$65
	.dc.b	$AF,$AE,$ED,$FF,$41,$FD,$56,$E3,$E0,$79,$08,$C4,$EA,$79,$4E,$5F
	.dc.b	$F0,$17,$C7,$7F,$80,$D3,$48,$E8,$FC,$0D,$BE,$73,$F8,$2E,$BF,$8F
	.dc.b	$9F,$CB,$57,$D2,$AD,$70,$FD,$49,$C9,$CA,$F8,$E9,$EF,$BE,$11,$89
	.dc.b	$F8,$91,$FA,$1D,$7F,$39,$87,$49,$FB,$66,$8A,$7F,$AC,$6B,$F8,$CC
	.dc.b	$F9,$63,$DA,$78,$70,$F2,$38,$C7,$AD,$78,$63,$F5,$18,$79,$2B,$6D
	.dc.b	$8F,$D1,$9E,$CA,$E3,$3A,$8E,$C6,$8E,$D8,$FF,$BD,$EF,$B4,$71,$8C
	.dc.b	$4F,$E7,$B9,$5F,$54,$F9,$3F,$E6,$7A,$7F,$23,$8C,$57,$FB,$15,$7C
	.dc.b	$BE,$A3,$F9,$AC,$6B,$1F,$F3,$9B,$6B,$95,$EE,$7F,$9F,$4F,$2A,$E9
	.dc.b	$39,$C3,$F0,$1F,$11,$FC,$91,$B3,$FD,$5B,$34,$E5,$B1,$98,$E5,$F5
	.dc.b	$4F,$5B,$A5,$6B,$2F,$F4,$3C,$E6,$FF,$DC,$1F,$55,$FE,$15,$6F,$9F
	.dc.b	$E3,$8F,$F3,$04,$60,$95,$FD,$96,$38,$36,$7D,$11,$88,$FF,$01,$BA
	.dc.b	$D7,$FA,$77,$C2,$7F,$BB,$AF,$9C,$63,$93,$E1,$7D,$53,$CB,$BF,$1F
	.dc.b	$86,$DF,$FB,$DE,$53,$F1,$DB,$1F,$DF,$1B,$97,$F7,$B6,$BF,$E1,$6F
	.dc.b	$25,$F2,$5F,$45,$B5,$D6,$F1,$3F,$57,$EA,$9F,$37,$7E,$73,$0D,$BF
	.dc.b	$E9,$5A,$FF,$6B,$FA,$AF,$85,$9F,$F7,$A6,$BF,$DF,$DB,$CC,$6F,$80
	.dc.b	$F9,$4F,$C2,$37,$0F,$F1,$C9,$88,$FC,$65,$BE,$41,$E4,$7E,$3F,$27
	.dc.b	$E6,$2F,$95,$DD,$6B,$1F,$F3,$6C,$E5,$FF,$72,$FE,$A9,$FD,$BD,$FE
	.dc.b	$40,$DC,$7F,$94,$35,$FE,$58,$D7,$FC,$2F,$EF,$BE,$97,$F3,$8D,$AF
	.dc.b	$F5,$A6,$B8,$9F,$4F,$D6,$3E,$8B,$D1,$F5,$8F,$9D,$CC,$61,$F7,$F2
	.dc.b	$3E,$27,$FC,$0C,$DC,$BF,$9B,$37,$4F,$E7,$CD,$7F,$CA,$BA,$3E,$5B
	.dc.b	$CA,$79,$BC,$3B,$7F,$EC,$9F,$A1,$F4,$FD,$37,$A9,$B1,$FE,$CD,$F9
	.dc.b	$BF,$DB,$60,$69,$23,$FD,$59,$F4,$1F,$08,$F8,$7E,$85,$8F,$C3,$9F
	.dc.b	$E3,$EF,$C8,$9F,$4D,$FC,$0D,$7D,$3D,$8C,$39,$FA,$4F,$E0,$D2,$69
	.dc.b	$2B,$FC,$BB,$0F,$83,$E8,$75,$E4,$B3,$E2,$E1,$FF,$11,$7D,$5F,$FA
	.dc.b	$46,$31,$08,$C2,$EE,$FF,$2C,$F6,$FF,$98,$BC,$0F,$C9,$D6,$A7,$FC
	.dc.b	$8E,$F0,$AF,$EE,$4D,$7F,$C5,$0D,$7C,$3F,$0B,$FE,$41,$8E,$58,$7F
	.dc.b	$8D,$70,$7E,$73,$FE,$84,$8F,$2C,$EA,$FF,$45,$39,$7C,$D3,$F3,$7A
	.dc.b	$3E,$61,$F3,$70,$B1,$F0,$7C,$46,$DF,$BE,$8C,$2F,$17,$78,$74,$3F
	.dc.b	$CF,$11,$99,$DF,$A7,$F6,$4F,$B5,$F5,$36,$32,$74,$F5,$B1,$7C,$F1
	.dc.b	$F0,$C7,$F4,$7D,$BD,$FF,$13,$76,$FF,$5A,$61,$F5,$E7,$E1,$F9,$F9
	.dc.b	$A9,$3F,$E1,$D9,$E2,$DA,$E3,$8F,$F8,$8E,$72,$D8,$CF,$0F,$DC,$7F
	.dc.b	$94,$B6,$FC,$5E,$6F,$E5,$7C,$E9,$8F,$C6,$FA,$9B,$8F,$AF,$81,$5F
	.dc.b	$9B,$F1,$33,$97,$FA,$3E,$7F,$7D,$87,$CE,$FA,$8F,$D0,$1F,$A1,$F8
	.dc.b	$EB,$19,$3E,$9F,$A3,$8C,$BB,$3F,$C8,$BE,$AF,$FC,$F5,$3F,$98,$6C
	.dc.b	$FF,$BC,$FE,$EB,$EE,$3F,$64,$46,$B2,$FF,$4E,$BC,$23,$FC,$90,$D7
	.dc.b	$FC,$FC,$D7,$FD,$10,$D7,$99,$FC,$C6,$7F,$08,$F7,$8E,$D5,$FD,$1A
	.dc.b	$73,$5F,$0F,$7F,$E0,$07,$F0,$1D,$ED,$7E,$4A,$3F,$82,$71,$63,$1F
	.dc.b	$9C,$FE,$17,$96,$57,$E1,$78,$BD,$F8,$FD,$B9,$CD,$FE,$67,$8F,$F5
	.dc.b	$7C,$F8,$AF,$C0,$3F,$2F,$AE,$63,$FC,$16,$5F,$76,$F0,$F3,$8F,$B6
	.dc.b	$FF,$D7,$1C,$1F,$D5,$E3,$FC,$DC,$D7,$E5,$C6,$8F,$EA,$D4,$F0,$F9
	.dc.b	$E1,$FE,$32,$8C,$63,$FB,$5F,$D2,$1F,$41,$FE,$27,$6C,$7E,$D8,$C4
	.dc.b	$47,$86,$17,$1F,$D7,$57,$97,$0F,$F7,$02,$FA,$BE,$87,$5F,$52,$E1
	.dc.b	$E1,$8F,$07,$BF,$8C,$39,$BE,$3F,$77,$FC,$71,$87,$F7,$C2,$38,$3D
	.dc.b	$1F,$49,$BC,$3D,$8F,$AF,$E8,$E4,$7B,$FE,$AE,$E3,$E8,$9B,$A7,$A2
	.dc.b	$6D,$7F,$B6,$7C,$36,$5F,$DC,$F0,$46,$3E,$33,$CA,$4A,$FF,$51,$46
	.dc.b	$5F,$9D,$F1,$7F,$36,$7E,$73,$F3,$FD,$CF,$69,$27,$8E,$1F,$E0,$BC
	.dc.b	$3F,$D5,$D8,$78,$AB,$EB,$7C,$77,$1B,$FB,$4F,$D9,$F9,$59,$18,$3E
	.dc.b	$F3,$8B,$FA,$EF,$28,$C9,$FC,$DF,$EE,$5F,$1C,$8F,$DC,$62,$9C,$C6
	.dc.b	$1B,$7B,$3E,$43,$88,$FF,$7D,$4F,$1A,$D7,$C6,$3D,$7F,$9D,$A3,$E1
	.dc.b	$B3,$1C,$AA,$E5,$F4,$2B,$58,$FF,$AE,$29,$33,$EB,$BF,$CB,$3F,$53
	.dc.b	$98,$C7,$8C,$8C,$43,$E5,$7F,$C6,$5A,$78,$E3,$73,$AB,$E7,$1F,$EC
	.dc.b	$EC,$3B,$A8,$CD,$71,$8E,$9C,$7E,$87,$58,$F8,$5E,$47,$62,$9A,$C3
	.dc.b	$CB,$8F,$93,$1E,$F8,$F6,$6D,$F3,$C7,$C1,$39,$C3,$E1,$DF,$17,$B4
	.dc.b	$EA,$B8,$5F,$2C,$3B,$E3,$E7,$C7,$96,$1E,$33,$CA,$E3,$1E,$91,$9C
	.dc.b	$7A,$5F,$C7,$C7,$AC,$F8,$B0,$E9,$18,$8E,$18,$F8,$5F,$83,$87,$4C
	.dc.b	$7C,$24,$AE,$11,$97,$94,$67,$0F,$4E,$5E,$16,$5F,$A2,$B8,$E3,$E2
	.dc.b	$9C,$E5,$E8,$AF,$0C,$E5,$E1,$3F,$79,$F3,$17,$0F,$C1,$B6,$FE,$E3
	.dc.b	$9F,$2F,$F9,$D3,$C4,$FC,$33,$87,$5E,$16,$5C,$B6,$93,$CB,$1F,$BA
	.dc.b	$DD,$3F,$06,$DB,$FD,$07,$AF,$FF,$00,$27,$38,$7E,$66,$B5,$5D,$8F
	.dc.b	$B7,$5B,$7F,$51,$1A,$C3,$C3,$5C,$4F,$D1,$56,$EB,$96,$1E,$49,$D4
	.dc.b	$74,$2C,$65,$FC,$4D,$F5,$AE,$57,$DA,$73,$7C,$30,$FC,$FD,$6A,$FC
	.dc.b	$84,$9F,$1D,$6C,$F4,$B5,$FD,$A4,$F1,$8C,$5F,$7C,$3F,$4F,$39,$AF
	.dc.b	$09,$23,$2F,$0B,$F0,$3E,$D3,$58,$7C,$E6,$9F,$48,$FE,$21,$C3,$FC
	.dc.b	$9F,$3F,$9C,$FF,$4A,$F8,$EB,$C0,$73,$C7,$F8,$AC,$BF,$89,$57,$1C
	.dc.b	$3F,$97,$61,$FC,$43,$83,$F9,$5D,$FE,$41,$B1,$99,$FE,$57,$F5,$7F
	.dc.b	$C5,$BA,$BF,$B9,$B6,$B3,$3F,$E0,$1F,$AB,$C3,$FC,$B6,$6E,$3F,$D8
	.dc.b	$1B,$A1,$B5,$8E,$4B,$7C,$CF,$F0,$63,$97,$FA,$DB,$F9,$FB,$FE,$D6
	.dc.b	$DB,$FE,$65,$3F,$D8,$32,$EF,$CB,$23,$FC,$2D,$1E,$23,$81,$1E,$07
	.dc.b	$F7,$0E,$87,$D5,$FA,$9B,$1F,$E9,$36,$BF,$E1,$83,$51,$1D,$3F,$01
	.dc.b	$E5,$F5,$5C,$FE,$8E,$47,$38,$CB,$E1,$7D,$D4,$69,$8F,$A9,$F2,$7D
	.dc.b	$54,$AF,$24,$79,$C8,$E0,$F4,$7D,$57,$AF,$E7,$FD,$9F,$2D,$27,$D9
	.dc.b	$19,$69,$0F,$F5,$16,$3F,$EF,$D7,$FD,$A3,$F1,$2F,$6B,$F0,$30,$F9
	.dc.b	$C6,$58,$6C,$70,$9D,$D7,$18,$FE,$34,$D7,$F9,$03,$5F,$E5,$4D,$7F
	.dc.b	$B4,$BB,$E7,$FF,$04,$42,$31,$CF,$AA,$F1,$67,$F9,$6F,$4D,$F9,$7B
	.dc.b	$E2,$30,$FF,$71,$CB,$FD,$D2,$DC,$3F,$DB,$CE,$F2,$FF,$5F,$35,$FF
	.dc.b	$3C,$36,$78,$D7,$4F,$8A,$FC,$57,$CB,$7E,$2F,$EA,$A6,$1F,$8B,$8C
	.dc.b	$B4,$87,$D0,$FA,$BF,$9B,$F5,$AF,$94,$F9,$FF,$20,$F8,$EA,$6A,$25
	.dc.b	$FF,$45,$AE,$0D,$8F,$E9,$0A,$32,$FF,$4F,$6E,$9F,$31,$BB,$7F,$86
	.dc.b	$1B,$3F,$E2,$8C,$1F,$1F,$FA,$77,$F5,$18,$49,$FD,$44,$65,$A4,$3F
	.dc.b	$1B,$F2,$9F,$AF,$F2,$3F,$61,$CB,$27,$F6,$11,$96,$90,$FC,$AF,$07
	.dc.b	$DA,$F9,$2F,$B7,$CB,$27,$EB,$C6,$5F,$F7,$97,$27,$E6,$DF,$F7,$83
	.dc.b	$5F,$F1,$AF,$2F,$E1,$A7,$FB,$73,$63,$99,$0F,$C0,$F4,$7F,$07,$F3
	.dc.b	$7F,$84,$73,$C9,$FE,$11,$19,$69,$0F,$D6,$FB,$DF,$CA,$3E,$FD,$E7
	.dc.b	$5E,$79,$5F,$E5,$6E,$7F,$F3,$37,$90,$BD,$CE,$7A,$3F,$D2,$11,$88
	.dc.b	$57,$08,$C3,$70,$F0,$73,$FE,$EB,$EB,$9E,$BF,$E9,$3C,$29,$8F,$53
	.dc.b	$FC,$CF,$3A,$70,$1C,$3F,$E1,$96,$BF,$E2,$36,$BF,$E6,$26,$BD,$27
	.dc.b	$C4,$47,$F5,$FE,$1F,$C5,$F9,$39,$1E,$3D,$7E,$14,$7C,$9F,$93,$F9
	.dc.b	$5F,$23,$FD,$1F,$63,$49,$7F,$EA,$E8,$7C,$97,$D1,$F8,$69,$3D,$67
	.dc.b	$2D,$F9,$1F,$AC,$47,$5F,$11,$F6,$7E,$4B,$C0,$48,$EC,$F8,$FC,$8F
	.dc.b	$BD,$F2,$52,$7B,$47,$E0,$7E,$42,$47,$85,$FC,$27,$C9,$7F,$2E,$FB
	.dc.b	$57,$C1,$FE,$E9,$B5,$FE,$43,$6B,$FE,$49,$6B,$FE,$53,$FA,$B7,$C3
	.dc.b	$7F,$78,$EB,$7D,$63,$E2,$30,$DB,$FB,$B7,$C1,$FA,$1F,$5A,$E9,$FE
	.dc.b	$FE,$D8,$CC,$63,$D6,$FE,$8B,$FA,$D7,$F0,$E9,$FF,$32,$35,$FF,$51
	.dc.b	$B5,$FF,$59,$B5,$7D,$7F,$F0,$E7,$F1,$FF,$97,$FF,$88,$5B,$F0,$9F
	.dc.b	$D4,$A7,$51,$FA,$0F,$13,$FD,$E5,$F4,$0F,$91,$FF,$0A,$B6,$7F,$14
	.dc.b	$D7,$81,$FE,$F6,$FA,$EF,$E9,$BC,$3F,$B0,$F8,$D7,$95,$FE,$13,$6D
	.dc.b	$7F,$BE,$1A,$FF,$BC,$1A,$FF,$B6,$B9,$FF,$16,$F6,$F1,$10,$F0,$3F
	.dc.b	$C5,$2D,$9F,$CF,$3C,$30,$FE,$03,$F5,$CF,$73,$FC,$5E,$DA,$FD,$23
	.dc.b	$5E,$DF,$5D,$FE,$8E,$AF,$E1,$2F,$FA,$42,$3F,$DD,$AD,$7F,$DC,$78
	.dc.b	$A3,$10,$8C,$7C,$1F,$F8,$CE,$7F,$D3,$D5,$D2,$3F,$C8,$31,$DF,$83
	.dc.b	$9B,$A7,$AF,$D6,$FF,$BD,$F0,$FD,$D7,$D7,$3C,$2F,$F2,$4B,$63,$F7
	.dc.b	$D3,$B7,$E4,$35,$FF,$75,$E4,$8C,$47,$6F,$91,$FF,$24,$46,$23,$D4
	.dc.b	$F3,$CF,$F9,$5B,$71,$E1,$3E,$0E,$0F,$F2,$DF,$E5,$9F,$85,$BF,$E6
	.dc.b	$3F,$06,$F1,$FF,$CB,$ED,$BF,$E2,$95,$BC,$3E,$5B,$F8,$D8,$F3,$B5
	.dc.b	$CF,$3F,$F9,$76,$31,$1F,$3F,$EB,$DF,$0B,$FE,$6C,$6C,$7F,$21,$6B
	.dc.b	$F8,$B5,$F2,$9B,$99,$3B,$79,$47,$F8,$E3,$EB,$BF,$C1,$9A,$29,$F5
	.dc.b	$DF,$AD,$7F,$D1,$DF,$19,$97,$3D,$F4,$46,$67,$85,$F4,$AD,$9D,$5C
	.dc.b	$47,$5A,$E3,$3E,$18,$C5,$73,$9F,$05,$70,$6F,$C3,$7F,$5C,$BE,$A7
	.dc.b	$98,$E5,$1E,$5A,$E5,$1E,$19,$E1,$7E,$66,$D6,$E7,$89,$F0,$DF,$2B
	.dc.b	$DA,$FC,$4F,$3C,$3E,$47,$6F,$A2,$2D,$73,$7F,$83,$BE,$89,$CC,$71
	.dc.b	$3D,$31,$DB,$0E,$51,$DF,$0F,$2B,$B8,$C9,$27,$A4,$6E,$BE,$B3,$82
	.dc.b	$E1,$F4,$DB,$7F,$25,$CF,$C4,$BE,$BF,$8C,$7D,$78,$C3,$8A,$46,$23
	.dc.b	$8C,$FF,$93,$17,$C9,$D2,$3F,$E5,$06,$E5,$F8,$DC,$5F,$12,$E3,$FA
	.dc.b	$37,$F8,$FB,$B5,$A9,$CC,$FE,$5E,$73,$5C,$63,$F0,$66,$4F,$C2,$D7
	.dc.b	$17,$27,$98,$FA,$A7,$F8,$B6,$39,$6B,$FA,$18,$E4,$F9,$E3,$3A,$FC
	.dc.b	$6B,$F0,$E1,$E4,$BE,$98,$EE,$31,$3A,$9F,$0B,$67,$B4,$70,$AF,$F7
	.dc.b	$85,$FA,$E7,$35,$B9,$F0,$57,$A0,$FF,$1E,$D7,$91,$FC,$0C,$FE,$39
	.dc.b	$CD,$FE,$43,$A5,$F5,$FF,$C6,$63,$FC,$54,$D9,$FE,$A8,$FF,$89,$23
	.dc.b	$3C,$FF,$37,$4F,$F2,$8C,$EE,$BF,$98,$B5,$FE,$92,$D7,$FA,$43,$6F
	.dc.b	$EF,$FC,$E7,$FA,$E2,$7F,$A7,$E4,$8C,$47,$F4,$F8,$C6,$3F,$94,$AE
	.dc.b	$2D,$8C,$CF,$F6,$BF,$AF,$FF,$68,$CF,$F9,$16,$73,$8F,$F9,$A2,$73
	.dc.b	$5F,$DC,$67,$FC,$47,$5B,$F4,$3D,$CF,$F6,$2F,$AF,$7C,$F6,$2F,$C9
	.dc.b	$57,$2F,$CC,$BE,$86,$DF,$96,$31,$87,$6A,$FD,$AD,$69,$F4,$7C,$8F
	.dc.b	$39,$D6,$7E,$78,$ED,$F5,$EF,$9A,$A6,$3E,$9E,$E7,$FE,$63,$9D,$3F
	.dc.b	$1D,$B8,$F8,$1A,$FF,$68,$6B,$FE,$00,$F2,$9E,$CF,$AD,$B5,$F4,$5B
	.dc.b	$3E,$A6,$BF,$DC,$BE,$4B,$FC,$5D,$3F,$5D,$B1,$F0,$1A,$EF,$C2,$F9
	.dc.b	$10,$FA,$1A,$4A,$E4,$FC,$DF,$B0,$7D,$3D,$24,$FE,$16,$32,$FF,$1A
	.dc.b	$C1,$FA,$DF,$FB,$A6,$30,$D8,$FE,$3E,$D7,$F9,$4B,$5F,$F6,$1F,$87
	.dc.b	$F9,$45,$F2,$1F,$DB,$AB,$FD,$8B,$39,$8C,$4F,$F2,$AF,$AF,$7D,$1F
	.dc.b	$16,$DF,$95,$B3,$B9,$ED,$F5,$FF,$E4,$3A,$F5,$36,$7F,$9A,$B5,$FE
	.dc.b	$7E,$D7,$FD,$B3,$F0,$1E,$32,$4F,$FB,$69,$F5,$7C,$3B,$E3,$69,$27
	.dc.b	$E4,$46,$5F,$F5,$37,$2F,$E7,$7E,$20,$D2,$47,$CF,$F9,$17,$D2,$7C
	.dc.b	$76,$C6,$63,$07,$3E,$99,$1F,$EB,$36,$E7,$FA,$6C,$38,$B6,$BC,$44
	.dc.b	$AF,$05,$77,$23,$83,$FD,$7D,$F5,$5E,$92,$38,$3F,$0D,$F8,$F3,$E3
	.dc.b	$FF,$7D,$82,$31,$3F,$EE,$1E,$27,$E0,$6F,$8D,$DD,$6A,$FF,$D1,$27
	.dc.b	$1C,$3F,$11,$19,$8F,$EE,$AD,$7F,$C5,$8D,$7F,$C6,$4F,$3F,$03,$F9
	.dc.b	$87,$E7,$BF,$09,$FE,$BD,$3D,$72,$FF,$7A,$42,$75,$1F,$E7,$AD,$BF
	.dc.b	$B6,$3F,$ED,$B3,$F8,$C3,$FA,$6A,$FF,$49,$3A,$6C,$68,$DF,$55,$7F
	.dc.b	$DE,$1F,$4F,$FB,$76,$F7,$7F,$23,$8C,$57,$FA,$75,$FE,$03,$8F,$F7
	.dc.b	$5F,$11,$1A,$8C,$3A,$BF,$F4,$8B,$96,$DF,$D4,$FD,$33,$DB,$87,$F9
	.dc.b	$5D,$BC,$1F,$53,$E4,$75,$C9,$30,$FE,$05,$5F,$8A,$BF,$F1,$2B,$5F
	.dc.b	$F2,$E3,$5E,$93,$FC,$FE,$FE,$FF,$9B,$CE,$7F,$D8,$B1,$8B,$F8,$7F
	.dc.b	$57,$F0,$4F,$85,$E6,$D8,$C9,$2B,$E2,$46,$5F,$8D,$E5,$7C,$9F,$2F
	.dc.b	$E5,$36,$BE,$7B,$67,$71,$F4,$A7,$F2,$EF,$F8,$D0,$FF,$3D,$6F,$F8
	.dc.b	$67,$13,$FE,$DC,$87,$D7,$FE,$05,$D7,$AB,$C8,$FE,$3D,$5C,$2B,$FC
	.dc.b	$96,$D7,$FD,$04,$D7,$FD,$10,$F6,$9F,$F0,$35,$3A,$BA,$4F,$E5,$A3
	.dc.b	$0F,$E2,$BE,$5B,$F1,$1E,$0F,$E2,$5B,$19,$3F,$1F,$19,$69,$FB,$3E
	.dc.b	$C7,$CD,$7D,$76,$C7,$ED,$FA,$1D,$F9,$A9,$3C,$5F,$BD,$3A,$8C,$1F
	.dc.b	$28,$FF,$50,$7D,$73,$0F,$E0,$AD,$8F,$D8,$F7,$3E,$5F,$35,$27,$FA
	.dc.b	$95,$6F,$7F,$F3,$6B,$5F,$F3,$BB,$5E,$F3,$FE,$0F,$DF,$F8,$A6,$1C
	.dc.b	$E7,$ED,$3F,$C1,$6F,$FD,$71,$CA,$FA,$9F,$A7,$5F,$96,$6D,$FF,$99
	.dc.b	$67,$5F,$11,$F9,$D9,$CD,$F2,$6D,$FF,$5E,$6D,$70,$9F,$E4,$51,$AE
	.dc.b	$EF,$DD,$FC,$BF,$F0,$13,$F8,$17,$D6,$DA,$FE,$0F,$17,$0F,$E0,$93
	.dc.b	$72,$FE,$03,$E6,$7F,$8A,$7C,$1F,$ED,$35,$FE,$FC,$6B,$FC,$12,$34
	.dc.b	$7F,$0F,$F2,$FF,$88,$37,$1F,$E2,$BF,$0B,$FC,$4C,$FE,$29,$F0,$56
	.dc.b	$32,$7F,$92,$F0,$4F,$84,$91,$9F,$73,$C8,$47,$04,$C7,$F8,$F7,$C3
	.dc.b	$FF,$21,$9F,$E4,$4F,$F2,$3F,$AC,$F6,$12,$BC,$0D,$AF,$E4,$DF,$59
	.dc.b	$B9,$FE,$5A,$FF,$2B,$E1,$91,$83,$F9,$8B,$89,$FF,$25,$E5,$F8,$19
	.dc.b	$F6,$4F,$CB,$3F,$CA,$9F,$9A,$7F,$BD,$A3,$F0,$B4,$E2,$4E,$63,$FA
	.dc.b	$6C,$63,$E1,$BF,$D5,$95,$FF,$00,$63,$0F,$26,$CF,$FB,$03,$E9,$FF
	.dc.b	$9C,$D7,$FA,$CF,$E0,$90,$9F,$06,$1D,$E7,$31,$E3,$CB,$15,$B8,$7D
	.dc.b	$77,$F8,$33,$A3,$FD,$81,$39,$8C,$4F,$84,$ED,$A7,$F8,$B3,$0E,$B1
	.dc.b	$CB,$CA,$61,$CA,$3C,$97,$CA,$B6,$DC,$3C,$58,$73,$7C,$18,$72,$BF
	.dc.b	$01,$D7,$0E,$8D,$33,$5C,$5E,$F3,$C7,$0E,$ED,$BE,$B7,$CA,$F8,$5F
	.dc.b	$3B,$DC,$F6,$C3,$CD,$3F,$03,$97,$E8,$61,$CA,$3C,$B3,$A7,$94,$7B
	.dc.b	$27,$71,$E4,$AE,$98,$F8,$E3,$07,$AA,$76,$F2,$33,$87,$91,$D6,$DC
	.dc.b	$AE,$47,$8A,$FA,$3F,$C5,$EF,$9E,$3E,$52,$4E,$B2,$F8,$31,$8B,$E7
	.dc.b	$8F,$DC,$9C,$D7,$4A,$D6,$1E,$23,$C1,$7E,$27,$F1,$87,$D5,$2C,$7E
	.dc.b	$29,$B3,$F7,$5C,$FA,$5F,$D2,$9E,$27,$E5,$1C,$3A,$F0,$7F,$D2,$15
	.dc.b	$FE,$B7,$6C,$F1,$C7,$F1,$AD,$D3,$F3,$4D,$BF,$C9,$FD,$17,$E8,$6B
	.dc.b	$77,$F7,$EB,$55,$E1,$8F,$C1,$56,$EF,$9C,$6C,$93,$F8,$29,$D3,$FA
	.dc.b	$0A,$E7,$87,$9E,$3F,$0F,$18,$9D,$C7,$73,$35,$D3,$0F,$D3,$CE,$AF
	.dc.b	$94,$7A,$2B,$67,$84,$E1,$87,$99,$CC,$72,$72,$79,$6E,$BF,$C0,$AF
	.dc.b	$B6,$1E,$82,$4F,$C2,$9D,$E1,$E2,$C3,$B6,$1D,$67,$BC,$60,$8F,$D0
	.dc.b	$38,$DF,$47,$D0,$7F,$0D,$E1,$FE,$4E,$FC,$A7,$FC,$B7,$9D,$DF,$F1
	.dc.b	$18,$7F,$1C,$9D,$D7,$F0,$F6,$BF,$CB,$B8,$3F,$C5,$AF,$F2,$58,$CF
	.dc.b	$07,$F2,$F9,$CC,$FF,$25,$3F,$DB,$93,$FE,$6A,$6D,$FF,$B7,$DD,$1F
	.dc.b	$D7,$4F,$E5,$5D,$3F,$E6,$E6,$DF,$F9,$B9,$A7,$F4,$3F,$C0,$DF,$5D
	.dc.b	$8C,$10,$FA,$EC,$3A,$AF,$F1,$31,$C6,$B3,$32,$3F,$4F,$5C,$A3,$D3
	.dc.b	$5C,$E7,$2D,$7F,$6B,$3E,$0F,$9D,$EF,$F3,$B2,$7B,$CE,$5B,$4A,$EA
	.dc.b	$48,$EA,$F8,$3E,$9E,$47,$81,$AF,$82,$51,$EE,$24,$76,$F7,$3E,$0F
	.dc.b	$D4,$7C,$2F,$AA,$93,$E2,$8C,$BE,$C7,$DE,$47,$DF,$F3,$5F,$28,$8F
	.dc.b	$AB,$E2,$6A,$E0,$FF,$54,$64,$AC,$C6,$2B,$EC,$F8,$EF,$C0,$7F,$DB
	.dc.b	$58,$7F,$B5,$DA,$FF,$A6,$1B,$5F,$E4,$35,$F4,$FC,$9C,$F1,$FF,$93
	.dc.b	$1B,$4B,$ED,$E5,$6E,$FF,$13,$2F,$AD,$7C,$F6,$3E,$67,$D1,$3F,$86
	.dc.b	$0E,$1F,$F7,$3C,$F0,$74,$D7,$F9,$7B,$5F,$E7,$2D,$7F,$C0,$CD,$7F
	.dc.b	$C1,$CF,$1C,$FF,$DC,$ED,$CF,$77,$E5,$F5,$FE,$4F,$E0,$3F,$80,$C3
	.dc.b	$FD,$FC,$D7,$FD,$7E,$DC,$3F,$CD,$4B,$8F,$8F,$24,$AE,$2F,$48,$FD
	.dc.b	$3D,$FE,$A3,$EC,$25,$7E,$62,$32,$FE,$6A,$32,$FE,$07,$2F,$E0,$9C
	.dc.b	$6A,$F9,$FF,$61,$FB,$1A,$D1,$FE,$2D,$F9,$27,$C0,$7D,$46,$C7,$D1
	.dc.b	$9E,$1A,$B1,$FF,$17,$47,$39,$FE,$CA,$D7,$3D,$BF,$C3,$63,$11,$FE
	.dc.b	$3B,$79,$CE,$A3,$B4,$F8,$7E,$1F,$F8,$3F,$C7,$FF,$08,$F9,$A9,$5F
	.dc.b	$B8,$8C,$BF,$BA,$8C,$BE,$CE,$EF,$E2,$33,$FC,$43,$92,$4F,$F0,$07
	.dc.b	$F8,$9E,$98,$8E,$1B,$FE,$4D,$D1,$FC,$1A,$3F,$2F,$F5,$0F,$47,$F5
	.dc.b	$3F,$31,$FE,$A9,$AF,$E3,$7D,$68,$C4,$7C,$E1,$5C,$5F,$CE,$35,$FF
	.dc.b	$02,$B5,$FF,$07,$B5,$FF,$2E,$35,$FF,$32,$3C,$78,$7E,$A4,$62,$FC
	.dc.b	$E7,$9B,$D1,$FE,$63,$E5,$7F,$32,$FB,$09,$5F,$C8,$A3,$2F,$F2,$48
	.dc.b	$CB,$F1,$23,$F9,$CF,$CD,$48,$FE,$52,$FF,$3C,$FA,$FC,$47,$CA,$8B
	.dc.b	$8F,$F1,$29,$FD,$FF,$E6,$8F,$6F,$FA,$4C,$BF,$3A,$7B,$B8,$6E,$3F
	.dc.b	$C0,$6F,$83,$6F,$FA,$C4,$62,$3F,$C5,$AD,$7F,$C6,$ED,$7F,$D0,$8E
	.dc.b	$EF,$E3,$C2,$31,$D3,$AA,$FB,$BE,$CF,$F5,$8F,$53,$FA,$89,$2F,$FA
	.dc.b	$8B,$FD,$A7,$EA,$5E,$BD,$24,$FF,$90,$04,$69,$FE,$BB,$F5,$EF,$84
	.dc.b	$FE,$EA,$D8,$CC,$60,$CF,$D8,$C8,$E2,$D7,$FC,$9B,$4F,$63,$C8,$48
	.dc.b	$F1,$46,$5F,$0F,$B5,$E0,$23,$83,$D1,$F8,$4E,$DF,$F2,$CB,$5F,$F3
	.dc.b	$1B,$5F,$F5,$83,$5F,$F5,$C4,$FF,$2D,$7E,$0F,$97,$F1,$C8,$FC,$5F
	.dc.b	$BD,$7C,$37,$F8,$4D,$B1,$98,$C1,$FE,$17,$FA,$F7,$8B,$49,$3F,$3E
	.dc.b	$32,$FC,$0D,$3E,$B1,$1C,$1E,$1F,$43,$C0,$47,$07,$F8,$4B,$EB,$9F
	.dc.b	$3D,$FE,$2A,$6C,$66,$30,$7F,$8B,$BE,$B1,$E0,$52,$4F,$F8,$72,$32
	.dc.b	$DE,$6E,$04,$70,$7F,$88,$BE,$B5,$E8,$7F,$8E,$9B,$19,$8C,$1F,$E3
	.dc.b	$EF,$0F,$FC,$49,$1F,$EF,$16,$BF,$E4,$76,$BF,$EF,$B6,$F9,$7F,$C7
	.dc.b	$7D,$DF,$F1,$D4,$7F,$8B,$7B,$FF,$86,$CE,$6B,$FC,$57,$3F,$CE,$94
	.dc.b	$3E,$A1,$F1,$BF,$E4,$3F,$5F,$FB,$55,$69,$B5,$B9,$FF,$24,$57,$0A
	.dc.b	$FF,$20,$B5,$FF,$1E,$B5,$FF,$7F,$4F,$17,$73,$C8,$8F,$F9,$03,$EA
	.dc.b	$DF,$13,$FE,$55,$FA,$BF,$B3,$3F,$D1,$F1,$7C,$CF,$9C,$D7,$1C,$9F
	.dc.b	$F2,$9D,$70,$7C,$2D,$7F,$CE,$4D,$7E,$1F,$D5,$FF,$37,$CC,$D2,$6B
	.dc.b	$FE,$74,$7F,$9D,$78,$DF,$E7,$B6,$D7,$29,$FF,$36,$43,$E3,$7F,$C5
	.dc.b	$3B,$FE,$8B,$81,$6F,$33,$DE,$E3,$77,$C6,$EA,$35,$14,$CB,$CA,$78
	.dc.b	$BF,$6F,$B7,$FC,$98,$F0,$3B,$E1,$E1,$7F,$B0,$57,$3B,$ED,$E8,$F2
	.dc.b	$8E,$B3,$A3,$8D,$F4,$9E,$F7,$CD,$E1,$7C,$AF,$A5,$FB,$A7,$77,$C6
	.dc.b	$7D,$97,$C5,$DD,$77,$3F,$03,$7C,$DF,$0E,$3C,$E7,$8D,$FB,$B1,$F0
	.dc.b	$D7,$C3,$9E,$71,$E0,$AC,$4F,$23,$B4,$FB,$B1,$E3,$1E,$1C,$7C,$8E
	.dc.b	$E3,$24,$9F,$0D,$78,$E7,$A1,$F5,$CF,$A4,$FF,$4B,$AF,$A1,$19,$6B
	.dc.b	$F0,$BD,$B7,$D5,$ED,$B1,$CC,$FD,$B3,$87,$93,$5D,$7C,$57,$FB,$15
	.dc.b	$34,$8F,$FB,$25,$F4,$3F,$E9,$1F,$5B,$FC,$27,$97,$E2,$A3,$F2,$B3
	.dc.b	$AC,$55,$AC,$BB,$D7,$D7,$9C,$F0,$F4,$8E,$26,$6B,$83,$F1,$6F,$E9
	.dc.b	$BE,$19,$AF,$0D,$3F,$37,$7D,$23,$F3,$47,$E8,$EB,$C9,$7D,$23,$07
	.dc.b	$7B,$F1,$3F,$08,$C1,$F8,$D3,$CB,$8F,$69,$E5,$18,$C7,$FD,$D7,$18
	.dc.b	$8E,$53,$BA,$EB,$7E,$26,$CF,$69,$FD,$DF,$13,$EA,$FF,$D4,$F5,$FE
	.dc.b	$DC,$7F,$B4,$B7,$C6,$FF,$21,$E7,$FC,$E2,$73,$87,$F9,$95,$E2,$DF
	.dc.b	$17,$77,$FC,$2B,$0E,$D8,$62,$BF,$9E,$DF,$2C,$3F,$A1,$E7,$FD,$93
	.dc.b	$0F,$F7,$BB,$6B,$FD,$E6,$D7,$F9,$BB,$5E,$CE,$E7,$FA,$6F,$D5,$FF
	.dc.b	$A9,$6B,$FD,$EF,$39,$7C,$91,$9A,$FF,$3F,$7D,$5B,$E7,$7C,$D9,$A3
	.dc.b	$17,$FF,$00,$17,$CE,$E9,$CD,$F2,$BF,$F2,$1B,$F0,$63,$2F,$F4,$E6
	.dc.b	$B9,$8D,$BF,$D8,$23,$9E,$5F,$D9,$BA,$7F,$BB,$E1,$E1,$BF,$2C,$62
	.dc.b	$FF,$D2,$DF,$56,$FA,$BF,$4A,$FA,$BB,$18,$66,$1F,$B9,$D1,$5B,$8C
	.dc.b	$36,$FF,$84,$4E,$AB,$CE,$F5,$7E,$39,$BC,$BF,$BE,$D5,$BF,$C1,$4D
	.dc.b	$BF,$EF,$BF,$7B,$F8,$19,$FC,$33,$63,$3F,$15,$F4,$A3,$18,$BC,$B7
	.dc.b	$D0,$F8,$9B,$87,$F8,$B6,$30,$79,$BE,$B1,$F4,$3F,$71,$FF,$36,$CF
	.dc.b	$F5,$ED,$7F,$C2,$71,$A8,$C1,$F9,$97,$79,$FF,$B0,$FD,$86,$9F,$86
	.dc.b	$6D,$70,$9F,$F0,$CC,$3D,$0D,$57,$E8,$65,$F5,$9F,$E5,$0B,$FF,$3E
	.dc.b	$36,$3F,$96,$35,$FE,$70,$D7,$FB,$87,$C2,$3E,$77,$F5,$99,$BE,$F7
	.dc.b	$F2,$B8,$78,$ED,$7B,$8F,$35,$6F,$A3,$D5,$D9,$27,$D5,$C4,$DB,$E8
	.dc.b	$9F,$BE,$BF,$F4,$FF,$D6,$7B,$6B,$FD,$28,$DC,$BF,$A2,$35,$FE,$A0
	.dc.b	$D7,$FC,$11,$F5,$8F,$96,$F4,$FD,$EA,$3E,$31,$9B,$F3,$C6,$5F,$2E
	.dc.b	$2F,$91,$FA,$47,$D3,$AB,$3A,$F4,$3F,$2F,$5F,$E3,$48,$FF,$53,$C6
	.dc.b	$0C,$FD,$87,$DB,$F9,$0F,$B8,$D9,$EF,$D9,$F7,$CF,$C0,$5D,$8F,$1F
	.dc.b	$77,$62,$3F,$EB,$47,$C3,$F0,$5F,$87,$FA,$EF,$C4,$36,$7E,$B7,$DE
	.dc.b	$7C,$D7,$FC,$58,$DE,$0F,$B3,$F9,$7F,$CD,$E3,$FF,$67,$CE,$64,$FF
	.dc.b	$32,$7F,$BF,$35,$E7,$CB,$FC,$39,$DB,$63,$B7,$3A,$DF,$0B,$F9,$79
	.dc.b	$CC,$63,$1F,$A3,$DF,$FA,$1F,$87,$FD,$13,$63,$27,$D6,$8C,$B4,$FB
	.dc.b	$27,$EA,$4F,$D5,$7B,$75,$C9,$33,$FD,$84,$65,$68,$9F,$F2,$BE,$7E
	.dc.b	$9F,$82,$FE,$9C,$D8,$FD,$A3,$ED,$F8,$7C,$E6,$AA,$7F,$32,$9E,$18
	.dc.b	$7F,$89,$DA,$FE,$C2,$78,$5F,$F8,$86,$77,$87,$8E,$BF,$5F,$FA,$7B
	.dc.b	$E2,$7A,$DF,$8A,$30,$FE,$BB,$E8,$7F,$AA,$57,$FA,$81,$7D,$77,$F5
	.dc.b	$27,$3E,$01,$CD,$71,$F7,$F5,$5B,$6C,$75,$F3,$B9,$1C,$FC,$3B,$19
	.dc.b	$26,$3F,$31,$FE,$F1,$87,$F0,$A8,$C1,$E5,$AF,$E3,$5F,$63,$F8,$38
	.dc.b	$C8,$71,$AF,$F8,$23,$E6,$9F,$4F,$C3,$A2,$65,$FE,$49,$C3,$84,$7E
	.dc.b	$57,$0E,$91,$FE,$57,$6B,$FE,$8F,$7A,$CF,$F3,$9B,$7C,$97,$4A,$F3
	.dc.b	$E2,$9C,$D7,$CF,$F3,$BE,$61,$F3,$7E,$BA,$C6,$49,$7F,$03,$87,$E8
	.dc.b	$4E,$A3,$E1,$F2,$38,$FF,$9C,$65,$F8,$18,$C7,$BB,$F9,$29,$CE,$3F
	.dc.b	$96,$51,$9F,$80,$D3,$67,$F0,$CF,$F4,$7F,$AB,$7D,$55,$3F,$9F,$FA
	.dc.b	$2F,$1F,$17,$FC,$FE,$B6,$FC,$D6,$E9,$FE,$77,$6B,$FE,$83,$6B,$FE
	.dc.b	$BB,$77,$A7,$F9,$02,$DE,$47,$6A,$F9,$B1,$8B,$FC,$D7,$D6,$FE,$5F
	.dc.b	$EB,$3F,$30,$D8,$C9,$F9,$F8,$CB,$4F,$D1,$72,$7E,$7F,$E5,$7F,$40
	.dc.b	$D8,$C9,$FA,$78,$CB,$4F,$D5,$7A,$7F,$E1,$4E,$BF,$EB,$D9,$7F,$86
	.dc.b	$1A,$FF,$BF,$9A,$FE,$7A,$3F,$5D,$D5,$FA,$F6,$D7,$ED,$7E,$1F,$F6
	.dc.b	$5C,$3F,$B3,$6C,$64,$FD,$D4,$65,$A7,$EF,$7E,$0F,$C8,$47,$FC,$61
	.dc.b	$8A,$31,$8F,$E3,$E1,$EE,$7F,$01,$9F,$E0,$4F,$F0,$3F,$8F,$F7,$12
	.dc.b	$F9,$B6,$F9,$7D,$66,$E7,$F8,$73,$FC,$2C,$8F,$F1,$19,$B3,$C2,$3F
	.dc.b	$8A,$7C,$96,$39,$3F,$8E,$76,$7F,$11,$8F,$E2,$9F,$57,$FC,$7E,$BE
	.dc.b	$63,$6F,$F8,$4B,$5F,$E3,$8D,$AF,$E3,$4F,$F2,$7F,$87,$7C,$DF,$F1
	.dc.b	$0F,$8C,$C3,$CB,$0F,$F5,$FB,$96,$E0,$1F,$58,$F3,$D8,$EA,$73,$18
	.dc.b	$7F,$DA,$37,$C5,$6D,$FC,$5B,$0E,$18,$73,$C2,$E1,$C7,$2F,$3E,$1E
	.dc.b	$1A,$D3,$D0,$F2,$61,$CD,$C1,$E2,$BE,$64,$73,$1B,$AE,$B1,$B3,$8D
	.dc.b	$76,$8C,$B7,$1F,$55,$71,$7D,$91,$9A,$F3,$61,$E9,$9C,$B6,$3C,$FC
	.dc.b	$BF,$41,$DE,$3E,$5C,$7B,$D6,$E7,$E3,$D6,$DF,$34,$EA,$B8,$5F,$92
	.dc.b	$B5,$87,$8E,$F7,$3E,$2A,$FA,$D8,$FA,$67,$C2,$7C,$0C,$7C,$17,$C6
	.dc.b	$75,$3C,$8D,$E3,$E1,$7E,$5E,$3E,$0B,$E2,$4C,$3B,$C6,$5F,$3C,$F2
	.dc.b	$C7,$D1,$96,$23,$E9,$1F,$87,$29,$E9,$30,$60,$FC,$27,$A7,$FE,$01
	.dc.b	$73,$9A,$31,$1C,$E3,$1E,$67,$F6,$36,$C7,$09,$E5,$9F,$E0,$1B,$87
	.dc.b	$DE,$6C,$FE,$4F,$C8,$FC,$E5,$6F,$0F,$CD,$D6,$AB,$F4,$ED,$C3,$FC
	.dc.b	$7F,$1A,$98,$FE,$42,$75,$7D,$30,$F0,$56,$B0,$ED,$7F,$96,$AD,$3F
	.dc.b	$0A,$75,$5C,$23,$46,$0D,$1D,$E3,$F6,$33,$CA,$3C,$F1,$B6,$9C,$8E
	.dc.b	$04,$AF,$19,$FA,$69,$E5,$87,$CC,$8C,$DF,$8D,$D5,$75,$9D,$4E,$4F
	.dc.b	$D5,$D7,$2C,$3E,$41,$30,$F3,$DF,$18,$C3,$E5,$C3,$F8,$3E,$EF,$CA
	.dc.b	$9F,$47,$F1,$A3,$FC,$79,$57,$9F,$F1,$D7,$C2,$F7,$3F,$C8,$2B,$95
	.dc.b	$FF,$19,$D7,$F8,$D4,$62,$7F,$92,$46,$2B,$8C,$7F,$95,$44,$ED,$C3
	.dc.b	$E5,$FA,$CF,$E4,$9C,$BF,$D0,$29,$F2,$B8,$C3,$F9,$85,$FA,$E3,$18
	.dc.b	$74,$A6,$18,$AE,$93,$FE,$CC,$9C,$B6,$75,$0F,$AC,$A6,$22,$4F,$18
	.dc.b	$CB,$FB,$0A,$F9,$91,$FE,$4B,$6C,$66,$BA,$37,$E9,$FB,$F9,$B2,$3C
	.dc.b	$91,$E2,$39,$49,$A4,$8E,$9F,$05,$E8,$FA,$C9,$1E,$D8,$F5,$98,$7F
	.dc.b	$C9,$54,$9C,$C7,$C3,$FA,$CF,$83,$F5,$92,$30,$43,$F8,$86,$5F,$89
	.dc.b	$9F,$29,$23,$DD,$E6,$7C,$AF,$9B,$91,$F3,$E3,$E6,$98,$7C,$71,$87
	.dc.b	$EA,$7D,$4F,$D2,$F3,$E4,$60,$87,$A9,$CB,$EC,$8F,$C0,$AF,$A9,$7B
	.dc.b	$98,$F9,$C7,$CA,$53,$73,$AA,$FE,$F5,$CF,$F9,$58,$C4,$7F,$2C,$6B
	.dc.b	$FC,$C5,$AF,$F7,$96,$BF,$E0,$37,$8C,$73,$BE,$E7,$8E,$3F,$2F,$EA
	.dc.b	$FF,$81,$BC,$1A,$3C,$5A,$61,$98,$7F,$B2,$9D,$36,$32,$FD,$EA,$FC
	.dc.b	$D9,$1F,$D0,$63,$5D,$1F,$A3,$8F,$D0,$C7,$CA,$FA,$4F,$F5,$3C,$FC
	.dc.b	$4E,$BF,$F0,$B4,$FF,$87,$63,$07,$F8,$85,$C3,$FE,$B2,$9E,$2D,$F1
	.dc.b	$3F,$D5,$CA,$32,$FF,$5B,$6B,$FE,$23,$6B,$FE,$2D,$79,$E9,$FC,$25
	.dc.b	$B7,$BA,$FC,$0D,$BE,$07,$F7,$7E,$24,$C7,$F8,$03,$FB,$EF,$A7,$C4
	.dc.b	$7E,$92,$2E,$5F,$C3,$E7,$F8,$1F,$CD,$7F,$04,$FA,$79,$7F,$C2,$63
	.dc.b	$2F,$F0,$B8,$CB,$FA,$E8,$FB,$7E,$8F,$DC,$CE,$4F,$D9,$8C,$BF,$EF
	.dc.b	$9E,$04,$7F,$B4,$F8,$7F,$C9,$2D,$8C,$CE,$E3,$8B,$FE,$C2,$8C,$B7
	.dc.b	$D3,$FE,$41,$F2,$7F,$92,$E0,$FE,$4E,$D9,$FE,$62,$DE,$BF,$E8,$9E
	.dc.b	$4F,$0E,$69,$5F,$7A,$55,$BE,$FF,$F0,$4F,$0F,$F9,$D2,$3F,$BC,$D7
	.dc.b	$FB,$5D,$7C,$27,$88,$F1,$F9,$72,$7A,$4E,$9B,$39,$7F,$DF,$7F,$74
	.dc.b	$F5,$FF,$4B,$EB,$C9,$FC,$88,$9D,$38,$6B,$FE,$24,$6B,$FE,$29,$6B
	.dc.b	$FE,$71,$6B,$FE,$7B,$78,$E3,$FD,$5D,$BB,$6E,$FF,$6F,$EC,$7C,$EF
	.dc.b	$67,$E7,$FC,$A4,$C7,$DB,$1F,$4B,$E6,$FF,$A2,$6A,$69,$27,$FA,$54
	.dc.b	$3E,$43,$EC,$92,$3E,$3F,$C5,$FD,$CF,$90,$FB,$BF,$23,$27,$E5,$C7
	.dc.b	$E0,$3E,$03,$B9,$23,$AB,$E2,$F9,$0F,$C2,$92,$3E,$9F,$C1,$FF,$2A
	.dc.b	$FB,$A7,$5F,$F6,$CE,$47,$DE,$BC,$CC,$37,$0F,$A7,$7C,$1F,$AB,$F5
	.dc.b	$6F,$89,$FE,$02,$D8,$FF,$2B,$87,$0F,$F9,$69,$AF,$F8,$33,$85,$F2
	.dc.b	$7F,$9C,$24,$7E,$2B,$E5,$3F,$A0,$FD,$4F,$E2,$E1,$A7,$F8,$3F,$EE
	.dc.b	$7F,$DB,$D1,$F8,$86,$BA,$EB,$FF,$59,$57,$1A,$FF,$0B,$7D,$53,$AB
	.dc.b	$F8,$CB,$6F,$D2,$D7,$FC,$E2,$D7,$FC,$47,$F7,$5F,$DE,$5F,$D5,$35
	.dc.b	$F0,$1F,$E2,$27,$15,$FE,$DE,$5E,$E7,$F8,$67,$E7,$5F,$25,$7E,$4E
	.dc.b	$61,$FB,$59,$CB,$FB,$7F,$8F,$FF,$3E,$BF,$E3,$46,$BF,$E2,$86,$BF
	.dc.b	$E3,$26,$F9,$0E,$07,$67,$F8,$A5,$B5,$FC,$0D,$AF,$AA,$78,$76,$FF
	.dc.b	$8B,$FE,$A5,$EE,$52,$57,$F0,$BF,$8E,$FF,$18,$BC,$5A,$FF,$92,$9A
	.dc.b	$FF,$8F,$9A,$FF,$92,$1B,$67,$FC,$5B,$1C,$FE,$13,$FC,$81,$1D,$B8
	.dc.b	$5C,$BC,$FE,$13,$FC,$89,$F5,$2F,$8E,$FF,$23,$7D,$CB,$C4,$A6,$A2
	.dc.b	$67,$FE,$3F,$AE,$0D,$8F,$F2,$04,$62,$3F,$C4,$ED,$C7,$C3,$1F,$E6
	.dc.b	$08,$C6,$3F,$E5,$A6,$C6,$7E,$43,$FC,$B1,$F5,$4E,$6F,$F2,$D7,$C1
	.dc.b	$3C,$8A,$6A,$25,$FF,$94,$EB,$83,$63,$FC,$A9,$18,$8F,$F1,$DF,$D5
	.dc.b	$30,$FF,$33,$3F,$A4,$CF,$CF,$C4,$8C,$69,$FE,$70,$78,$5F,$F4,$56
	.dc.b	$BF,$83,$76,$FF,$9F,$9B,$3C,$EF,$FC,$6F,$F5,$5F,$C2,$F0,$5B,$EA
	.dc.b	$7F,$6B,$F2,$BF,$E1,$E8,$C3,$A4,$E6,$3F,$77,$5C,$E7,$67,$47,$11
	.dc.b	$D3,$E1,$B9,$36,$7C,$2F,$2F,$2F,$F8,$4C,$72,$15,$CA,$FC,$47,$84
	.dc.b	$B7,$D1,$F3,$46,$E7,$C5,$5D,$2F,$C1,$1D,$6B,$93,$AC,$3C,$71,$CA
	.dc.b	$BC,$7E,$FF,$02,$46,$67,$E4,$C7,$A5,$E3,$5F,$32,$7A,$BF,$23,$1F
	.dc.b	$33,$DC,$F7,$63,$E7,$74,$72,$29,$AC,$3D,$55,$E1,$8F,$A8,$73,$2C
	.dc.b	$FD,$36,$CF,$CC,$73,$D2,$FA,$AF,$E8,$75,$FC,$21,$B1,$99,$EA,$7D
	.dc.b	$27,$0F,$F4,$AF,$91,$FF,$26,$A6,$91,$FF,$28,$BE,$BC,$BF,$D0,$7C
	.dc.b	$1E,$D9,$FC,$8C,$E7,$E1,$BF,$6F,$B7,$F4,$69,$C9,$F8,$3B,$E2,$FE
	.dc.b	$14,$CC,$7E,$A6,$78,$E9,$E8,$BF,$AA,$FF,$B7,$23,$F8,$1E,$DD,$6F
	.dc.b	$D0,$1E,$58,$F0,$8F,$F7,$99,$E1,$9E,$AE,$30,$DB,$E3,$39,$B8,$3B
	.dc.b	$61,$ED,$AF,$81,$3A,$C3,$C9,$18,$9D,$4F,$36,$D7,$28,$F1,$4F,$E3
	.dc.b	$B8,$1F,$54,$F5,$BF,$8F,$E9,$FC,$2A,$31,$5F,$E4,$B7,$FB,$63,$67
	.dc.b	$6F,$FB,$F2,$37,$7F,$D3,$5B,$5C,$FC,$9D,$CB,$0F,$0E,$18,$7F,$D7
	.dc.b	$FF,$3D,$FC,$C2,$7F,$84,$46,$31,$FF,$7C,$6A,$AD,$57,$F1,$2A,$D3
	.dc.b	$FE,$E0,$9D,$57,$F9,$8F,$EA,$9D,$FD,$E7,$35,$FD,$8E,$73,$5F,$E6
	.dc.b	$9F,$AA,$7D,$07,$95,$B1,$9C,$B1,$32,$3F,$3F,$92,$B7,$87,$FB,$C9
	.dc.b	$AF,$B1,$AF,$A5,$DE,$BF,$CE,$BC,$27,$C8,$75,$AF,$EE,$0F,$C7,$FA
	.dc.b	$AF,$EC,$18,$FF,$9F,$D7,$B3,$9F,$13,$4C,$9F,$2C,$E9,$B1,$F2,$AF
	.dc.b	$EA,$4E,$5F,$E0,$B9,$FF,$83,$6F,$EB,$79,$3E,$09,$FF,$05,$56,$F3
	.dc.b	$FE,$A3,$39,$7F,$17,$EA,$FF,$BF,$72,$FF,$4A,$7D,$57,$E2,$B0,$FF
	.dc.b	$15,$D7,$F9,$2F,$E4,$3F,$05,$39,$8F,$EB,$7E,$7B,$E8,$7C,$DA,$A3
	.dc.b	$18,$FD,$06,$E3,$E8,$6B,$FE,$53,$F9,$BF,$C7,$7B,$26,$FD,$36,$1C
	.dc.b	$FC,$96,$BE,$7A,$E3,$FE,$A0,$78,$FD,$2F,$8B,$03,$49,$3F,$82,$F1
	.dc.b	$FF,$1B,$B7,$F7,$46,$C7,$FA,$F0,$FD,$03,$8C,$FF,$48,$FF,$82,$63
	.dc.b	$FD,$74,$FE,$11,$FC,$F3,$98,$FF,$1D,$39,$78,$FA,$FF,$1A,$3F,$CA
	.dc.b	$33,$E3,$8C,$19,$F7,$BC,$27,$8B,$82,$CF,$83,$E1,$7C,$BF,$0D,$E6
	.dc.b	$6C,$79,$3E,$23,$B1,$1F,$80,$FE,$2F,$37,$BD,$FE,$00,$9F,$F0,$3C
	.dc.b	$7F,$0E,$8C,$69,$F8,$27,$FC,$89,$3A,$36,$EA,$7E,$57,$C8,$FC,$8F
	.dc.b	$85,$F9,$2D,$AF,$8F,$F2,$1F,$04,$8F,$D3,$7E,$27,$BF,$F4,$0F,$A3
	.dc.b	$F4,$36,$7F,$DA,$BF,$4D,$FE,$BB,$6F,$C5,$7F,$18,$D3,$F8,$C7,$CC
	.dc.b	$DA,$D4,$7F,$A1,$03,$87,$FB,$6B,$5F,$F1,$13,$5F,$F1,$4B,$5E,$33
	.dc.b	$DF,$F4,$0F,$B9,$18,$BF,$F7,$4C,$7D,$FE,$DF,$F5,$E7,$8D,$F8,$AF
	.dc.b	$4B,$F1,$6D,$AF,$C4,$C5,$BF,$C3,$CD,$E0,$FF,$23,$FE,$5E,$E4,$FE
	.dc.b	$64,$DC,$3E,$95,$FF,$37,$F8,$37,$CC,$6A,$E5,$7E,$5B,$DB,$B7,$F9
	.dc.b	$47,$F9,$FE,$3F,$CD,$FD,$0F,$E8,$3F,$2F,$FE,$7C,$E2,$CF,$A1,$FD
	.dc.b	$D2,$BF,$CD,$F3,$9C,$7C,$2D,$C7,$FD,$62,$D9,$FE,$AF,$5C,$6B,$FC
	.dc.b	$A7,$39,$68,$BF,$0B,$4F,$6D,$79,$5E,$E6,$6F,$F8,$AF,$E1,$BF,$BE
	.dc.b	$8D,$B6,$3F,$8C,$FE,$9C,$FA,$3F,$7C,$E6,$FF,$B5,$B6,$FF,$73,$F0
	.dc.b	$DA,$F1,$3F,$80,$BF,$D8,$7C,$AF,$F3,$E4,$FF,$00,$FA,$6C,$70,$7F
	.dc.b	$06,$64,$64,$FF,$65,$46,$34,$FC,$97,$EA,$5F,$EF,$99,$FB,$30,$8C
	.dc.b	$3A,$EE,$FC,$E4,$63,$4F,$EE,$0E,$A7,$FC,$8A,$D7,$FC,$F4,$D7,$FD
	.dc.b	$02,$D7,$B7,$8B,$FA,$7F,$94,$FC,$3E,$1E,$2A,$EC,$FF,$A0,$E3,$35
	.dc.b	$FC,$6B,$E7,$3F,$91,$9F,$0F,$5B,$5F,$C8,$A3,$F9,$3F,$C2,$62,$BF
	.dc.b	$90,$BF,$73,$E9,$25,$7E,$23,$BB,$FA,$85,$7E,$22,$13,$A8,$C7,$C8
	.dc.b	$5D,$7F,$23,$5F,$EF,$A3,$F0,$7C,$9F,$88,$3F,$33,$83,$CC,$C4,$E8
	.dc.b	$FF,$2F,$F4,$7F,$B6,$2B,$FB,$63,$63,$27,$E1,$CB,$1F,$A4,$FA,$8F
	.dc.b	$F5,$5C,$73,$CB,$F3,$F1,$87,$FD,$F1,$DE,$F2,$BF,$6E,$7F,$8A,$7F
	.dc.b	$4A,$7D,$2E,$30,$FD,$CE,$8B,$E1,$1F,$D3,$30,$D1,$FD,$28,$FF,$19
	.dc.b	$7E,$9E,$FA,$8A,$7E,$57,$8A,$D7,$49,$D3,$75,$CD,$B8,$3F,$DA,$75
	.dc.b	$EF,$8C,$57,$F8,$DE,$3F,$C5,$F1,$FE,$0E,$6B,$FE,$40,$6B,$BF,$82
	.dc.b	$FF,$00,$C7,$0A,$EB,$F2,$7E,$89,$F4,$BE,$AE,$49,$87,$9F,$92,$E3
	.dc.b	$E2,$3F,$CA,$58,$FE,$12,$7D,$95,$F0,$AB,$78,$FA,$BE,$A3,$C2,$4A
	.dc.b	$FF,$1F,$DA,$31,$87,$5D,$D3,$F5,$2B,$E4,$57,$CE,$E0,$FF,$30,$79
	.dc.b	$DF,$2B,$0F,$F1,$7B,$5F,$F3,$0B,$5F,$F2,$F3,$5F,$F3,$53,$C7,$1F
	.dc.b	$9F,$F5,$1D,$38,$7E,$DB,$F7,$3E,$0A,$65,$F6,$62,$E1,$E7,$CF,$FC
	.dc.b	$9F,$3F,$76,$B8,$B4,$FE,$17,$E0,$7F,$16,$AF,$F3,$8B,$63,$FD,$96
	.dc.b	$D8,$FF,$80,$0D,$41,$E5,$F1,$4F,$A6,$DF,$9C,$BF,$F5,$EC,$72,$F0
	.dc.b	$F1,$87,$FB,$1E,$37,$3A,$8E,$BF,$54,$F6,$DE,$06,$65,$39,$8C,$66
	.dc.b	$F5,$7A,$E9,$E3,$BE,$78,$72,$7C,$06,$2F,$A6,$1C,$E3,$27,$B6,$F9
	.dc.b	$E5,$F0,$23,$C3,$7C,$F0,$ED,$18,$C3,$9C,$E6,$BE,$4D,$F2,$AF,$01
	.dc.b	$DB,$0F,$06,$1C,$9B,$5C,$32,$E7,$8C,$BE,$18,$F5,$7E,$9E,$5F,$06
	.dc.b	$B5,$8F,$3A,$D5,$79,$4F,$6D,$6D,$E7,$7C,$30,$F7,$63,$E2,$9E,$B9
	.dc.b	$7A,$5E,$58,$F5,$BE,$11,$AC,$AB,$AC,$39,$E5,$E2,$CB,$C1,$87,$1A
	.dc.b	$F0,$91,$F9,$71,$AA,$E3,$87,$4C,$BC,$97,$F5,$71,$C6,$3E,$E9,$CE
	.dc.b	$1C,$6B,$E7,$E3,$DD,$FB,$47,$C3,$2C,$FE,$05,$B5,$F8,$77,$3E,$EB
	.dc.b	$E9,$B8,$C7,$E5,$A3,$0E,$2B,$FD,$29,$3A,$8F,$CB,$CE,$9E,$71,$F9
	.dc.b	$08,$E3,$7F,$81,$6E,$7F,$4B,$E9,$3F,$21,$39,$AE,$04,$BF,$C5,$4E
	.dc.b	$AB,$EC,$DF,$29,$F3,$56,$EF,$F1,$D7,$B9,$C9,$F9,$3A,$E0,$FA,$2B
	.dc.b	$F4,$B3,$B8,$C1,$B8,$E9,$7E,$78,$A7,$4B,$F4,$3E,$63,$27,$27,$27
	.dc.b	$91,$CE,$1F,$29,$EC,$71,$2C,$70,$9F,$7D,$79,$A7,$8E,$1E,$AB,$E5
	.dc.b	$7D,$67,$8D,$F0,$69,$FC,$43,$0F,$75,$70,$9D,$57,$7A,$D1,$FC,$4F
	.dc.b	$F1,$6F,$E0,$31,$FC,$09,$7D,$47,$C7,$6C,$EE,$B0,$7F,$0B,$D7,$F2
	.dc.b	$73,$FC,$29,$B1,$99,$FE,$21,$39,$7F,$B8,$7A,$1F,$C4,$F9,$FF,$BA
	.dc.b	$B6,$F3,$3F,$DF,$7E,$4B,$0F,$F6,$69,$BD,$9F,$D7,$1B,$CF,$FE,$6D
	.dc.b	$9E,$07,$F8,$43,$A9,$DF,$AC,$24,$E6,$39,$CE,$A3,$F4,$B8,$23,$11
	.dc.b	$FD,$92,$7F,$49,$39,$6D,$BE,$AB,$C7,$E4,$CB,$FF,$0B,$47,$98,$D9
	.dc.b	$1F,$EF,$2F,$2F,$86,$F5,$92,$3F,$C3,$DE,$AD,$7F,$B3,$45,$DB,$FC
	.dc.b	$3B,$3F,$0B,$83,$E1,$FA,$F2,$FF,$C4,$D1,$97,$FC,$0E,$FA,$08,$FF
	.dc.b	$82,$BE,$4F,$E5,$F1,$C8,$FF,$17,$47,$CE,$23,$83,$E7,$78,$5F,$53
	.dc.b	$C9,$79,$DF,$53,$AE,$57,$F9,$0A,$72,$D9,$E0,$47,$FC,$8F,$FA,$4B
	.dc.b	$E7,$FF,$03,$E8,$3E,$B9,$F1,$38,$5F,$49,$4C,$44,$C3,$FC,$C7,$3C
	.dc.b	$1B,$1F,$C7,$6B,$A3,$67,$53,$FE,$BE,$6F,$17,$FB,$11,$BD,$9F,$EB
	.dc.b	$ED,$5F,$1F,$E3,$F9,$DF,$26,$93,$1F,$F6,$4E,$09,$D4,$61,$AF,$FB
	.dc.b	$32,$FE,$03,$FE,$68,$FC,$0D,$F5,$5F,$A0,$5F,$59,$4C,$44,$9E,$0D
	.dc.b	$7F,$C8,$5D,$FF,$E4,$57,$FD,$67,$F3,$CF,$87,$FD,$6C,$3E,$9D,$F1
	.dc.b	$F8,$6C,$70,$9D,$D7,$18,$FE,$8C,$D7,$FA,$6B,$5F,$EA,$ED,$7F,$C2
	.dc.b	$4E,$F8,$7F,$C5,$70,$8C,$70,$EA,$BB,$E5,$F0,$7C,$99,$7F,$15,$FD
	.dc.b	$CF,$A3,$23,$04,$3E,$0B,$FC,$2B,$17,$E2,$4F,$8D,$F9,$9F,$39,$2F
	.dc.b	$FD,$E7,$39,$6C,$F1,$8C,$47,$F9,$E6,$B9,$CF,$5E,$0F,$C2,$CB,$A3
	.dc.b	$F3,$1F,$39,$F9,$72,$47,$FC,$00,$FD,$35,$A6,$23,$F2,$13,$71,$FF
	.dc.b	$27,$8A,$DC,$F1,$AD,$6F,$FE,$5C,$C3,$BC,$E4,$FD,$87,$A5,$F8,$9C
	.dc.b	$64,$60,$87,$FA,$F2,$3F,$74,$7F,$2E,$F9,$9F,$C7,$C7,$E4,$FE,$B3
	.dc.b	$F2,$44,$7F,$2D,$F0,$9F,$95,$30,$FF,$A4,$7A,$13,$98,$FF,$38,$EB
	.dc.b	$F8,$0E,$67,$C6,$FE,$4B,$3F,$F6,$05,$6B,$5F,$F2,$C4,$62,$3F,$D5
	.dc.b	$07,$E5,$E7,$F2,$C7,$F8,$5D,$C1,$FE,$19,$AE,$AD,$8D,$46,$23,$9B
	.dc.b	$6F,$C9,$3F,$A4,$EF,$F0,$47,$FA,$03,$CF,$66,$FA,$8F,$DA,$13,$1F
	.dc.b	$DC,$74,$FE,$DF,$C0,$FE,$8F,$5F,$BC,$9F,$F1,$CD,$F2,$9C,$9A,$F9
	.dc.b	$CF,$DF,$78,$12,$BF,$81,$47,$E6,$7F,$30,$FD,$9C,$7E,$45,$B9,$FE
	.dc.b	$19,$BC,$1F,$39,$7D,$37,$F0,$9F,$13,$C6,$4B,$FE,$BF,$A7,$F1,$A6
	.dc.b	$F8,$BF,$C3,$FE,$7A,$47,$F1,$38,$FE,$25,$ED,$48,$FE,$2C,$FF,$15
	.dc.b	$E2,$FE,$2D,$ED,$FC,$02,$4F,$67,$E1,$FC,$7F,$F1,$BF,$9C,$93,$FC
	.dc.b	$7E,$3F,$BE,$FC,$83,$E4,$E9,$27,$F9,$27,$06,$BE,$83,$FB,$9D,$1A
	.dc.b	$49,$FE,$F7,$1F,$11,$73,$23,$E2,$C7,$E3,$CC,$B8,$7F,$DD,$4F,$18
	.dc.b	$7D,$57,$F2,$B9,$FF,$77,$37,$93,$F0,$CD,$C3,$FD,$9F,$5C,$3D,$2F
	.dc.b	$F0,$CF,$CF,$FE,$B3,$0F,$C1,$B5,$FF,$5B,$4A,$31,$1C,$9A,$B0,$7E
	.dc.b	$61,$FC,$12,$3F,$E0,$07,$1F,$F4,$F7,$D9,$68,$C4,$7F,$19,$9E,$15
	.dc.b	$F9,$DF,$98,$7B,$BF,$E2,$E6,$D7,$26,$BE,$1C,$FD,$BF,$32,$F8,$DF
	.dc.b	$F1,$AB,$67,$FC,$56,$D7,$F4,$2D,$7D,$ED,$79,$FC,$77,$29,$91,$EA
	.dc.b	$94,$FF,$90,$BE,$A9,$F2,$5F,$E4,$4F,$AB,$F7,$5F,$C0,$6B,$E7,$3F
	.dc.b	$03,$3C,$1F,$8A,$D7,$F6,$AD,$7E,$9B,$5F,$B6,$D7,$9C,$F8,$C8,$F4
	.dc.b	$F2,$FF,$CA,$FF,$55,$F8,$9F,$AB,$78,$7F,$E5,$A6,$D7,$D6,$9F,$C7
	.dc.b	$1F,$BE,$7F,$06,$D7,$F1,$6D,$7F,$36,$DE,$3F,$ED,$57,$FE,$66,$F5
	.dc.b	$CD,$31,$EE,$DB,$FC,$2C,$E9,$B3,$A7,$F4,$83,$EA,$FF,$0B,$C8,$87
	.dc.b	$D5,$3E,$4B,$11,$76,$B3,$98,$E1,$3F,$3E,$31,$1D,$79,$BF,$38,$F4
	.dc.b	$9E,$78,$78,$23,$07,$2B,$EB,$3E,$03,$A1,$C2,$B9,$DF,$6B,$E4,$F8
	.dc.b	$27,$77,$CC,$F6,$46,$EB,$8C,$F3,$C3,$2F,$7C,$3C,$37,$C4,$F2,$9E
	.dc.b	$B9,$DB,$F1,$F0,$F5,$7C,$17,$86,$BE,$3B,$67,$E6,$3E,$1A,$B7,$E1
	.dc.b	$9E,$AF,$A1,$F9,$C7,$63,$C0,$78,$AF,$DF,$1E,$78,$D3,$C6,$33,$1E
	.dc.b	$5C,$3E,$8B,$F6,$8E,$65,$8F,$A4,$D8,$DB,$F5,$BE,$2D,$F5,$7C,$63
	.dc.b	$E0,$46,$1C,$52,$31,$1D,$5A,$F1,$F8,$FF,$EF,$89,$A4,$7F,$C0,$0F
	.dc.b	$C1,$C5,$F1,$2E,$AF,$89,$5F,$97,$AD,$4E,$63,$EB,$99,$3E,$CD,$72
	.dc.b	$AE,$31,$E1,$33,$3C,$76,$FC,$CC,$71,$D7,$F3,$77,$C6,$11,$A7,$93
	.dc.b	$AA,$F1,$39,$C3,$CB,$87,$81,$B5,$E6,$3C,$58,$71,$9D,$47,$91,$B1
	.dc.b	$98,$E1,$3C,$A3,$B4,$79,$35,$FF,$79,$61,$E2,$7E,$84,$64,$FF,$1B
	.dc.b	$C6,$5F,$ED,$3F,$57,$FE,$58,$C3,$FA,$6B,$63,$F9,$77,$1A,$3F,$D3
	.dc.b	$33,$AF,$6B,$FC,$B6,$FF,$64,$8E,$15,$FD,$81,$AE,$67,$FD,$0A,$D9
	.dc.b	$FF,$7A,$F3,$7D,$D6,$9A,$8F,$EF,$B1,$97,$FC,$5E,$D7,$FB,$6B,$9D
	.dc.b	$7F,$C7,$9F,$56,$F8,$FF,$2D,$72,$BB,$92,$B5,$1F,$D3,$4C,$F5,$F3
	.dc.b	$7F,$BF,$AF,$23,$31,$0F,$D2,$52,$7C,$97,$C2,$7D,$53,$96,$CF,$59
	.dc.b	$FF,$0C,$FD,$5F,$F8,$96,$7C,$47,$F7,$E7,$11,$F4,$BD,$CF,$F6,$84
	.dc.b	$FF,$80,$9B,$19,$F5,$7D,$B1,$8D,$3F,$CB,$CE,$A7,$FC,$26,$D7,$FC
	.dc.b	$36,$D7,$FC,$56,$DC,$4F,$FB,$92,$FD,$ED,$9F,$F3,$0B,$5E,$BE,$97
	.dc.b	$D9,$B7,$D5,$FE,$A6,$7F,$A5,$C6,$38,$B5,$39,$9F,$F4,$5B,$5F,$F1
	.dc.b	$22,$F8,$FD,$93,$1F,$E2,$97,$DA,$3E,$7F,$98,$FA,$AA,$4A,$FF,$24
	.dc.b	$BA,$FA,$29,$1F,$E4,$C6,$E3,$FE,$7C,$F9,$2F,$F2,$94,$FD,$E6,$C6
	.dc.b	$7D,$0F,$BB,$18,$8F,$F1,$FE,$0F,$AA,$F1,$92,$FF,$CB,$2F,$9B,$DE
	.dc.b	$91,$FE,$5C,$6D,$7F,$A3,$29,$F4,$3E,$C2,$57,$F9,$89,$F7,$FC,$3C
	.dc.b	$8F,$F3,$2B,$6B,$97,$CD,$7C,$42,$4F,$F9,$A9,$F4,$FC,$74,$8F,$73
	.dc.b	$67,$FA,$47,$03,$EA,$9E,$5F,$DD,$6D,$FD,$26,$D7,$06,$BF,$13,$E9
	.dc.b	$7E,$A9,$27,$FC,$F6,$FD,$8F,$AA,$91,$FE,$7E,$6C,$FC,$3C,$7E,$E9
	.dc.b	$27,$FD,$06,$FF,$07,$87,$C9,$7F,$16,$7F,$D9,$AD,$CB,$FA,$34,$3E
	.dc.b	$AB,$FD,$81,$7F,$E2,$4A,$D3,$FE,$20,$79,$4F,$F6,$96,$B9,$F9,$2F
	.dc.b	$04,$62,$3F,$C5,$4F,$27,$FD,$0F,$F3,$7F,$C4,$9F,$2C,$DE,$D4,$FF
	.dc.b	$14,$E2,$FF,$18,$61,$FE,$7E,$6D,$FF,$A6,$3D,$6D,$79,$12,$B4,$F1
	.dc.b	$8C,$4F,$F3,$8F,$A5,$FF,$3D,$57,$F9,$FF,$E6,$FF,$90,$C6,$27,$F9
	.dc.b	$EF,$B4,$9C,$FB,$5F,$C9,$A3,$17,$FE,$90,$BE,$1C,$1F,$E8,$DB,$FE
	.dc.b	$5A,$BE,$A5,$F4,$B8,$6D,$FF,$AC,$E7,$58,$E6,$A4,$FE,$EA,$B7,$97
	.dc.b	$E4,$1A,$FF,$08,$6B,$FE,$57,$BF,$05,$78,$BE,$0B,$FB,$B4,$7F,$80
	.dc.b	$CC,$DF,$86,$32,$F1,$B7,$D3,$79,$CF,$47,$9B,$67,$D1,$3A,$6C,$E9
	.dc.b	$A7,$FA,$D3,$F4,$AF,$F7,$AB,$F6,$7E,$93,$FA,$7C,$63,$3F,$F6,$8C
	.dc.b	$63,$E3,$3F,$00,$DA,$E0,$7F,$9D,$BE,$A7,$F5,$7A,$7E,$61,$B1,$FE
	.dc.b	$CC,$FA,$8F,$C6,$46,$30,$FF,$3D,$3F,$B0,$7F,$CF,$2D,$73,$DF,$FA
	.dc.b	$78,$C4,$1E,$4D,$EE,$EB,$3D,$9F,$CB,$7B,$DF,$97,$8B,$87,$E5,$DB
	.dc.b	$5F,$EE,$E6,$E3,$FE,$35,$EA,$FF,$7B,$47,$F7,$4F,$97,$FF,$7C,$F9
	.dc.b	$76,$7F,$9F,$61,$FE,$F9,$9F,$F7,$27,$D0,$E6,$A4,$E6,$B7,$8F,$FB
	.dc.b	$0E,$F7,$39,$AF,$8B,$E5,$48,$E9,$93,$CA,$C6,$1F,$EC,$B3,$E1,$57
	.dc.b	$C5,$25,$7F,$98,$63,$2D,$DB,$FA,$1D,$FF,$94,$71,$65,$FE,$F0,$AF
	.dc.b	$B5,$87,$B7,$DF,$FD,$8F,$C4,$C9,$FD,$67,$0D,$CF,$F7,$13,$72,$F9
	.dc.b	$1D,$8A,$C5,$FF,$AD,$F2,$78,$1F,$EE,$88,$FE,$11,$F0,$3F,$90,$F3
	.dc.b	$7F,$86,$32,$32,$7E,$5E,$31,$9F,$F8,$FB,$A3,$FA,$7C,$FF,$87,$1B
	.dc.b	$1F,$E8,$9C,$11,$88,$FF,$62,$C6,$3C,$CF,$F0,$FC,$F0,$D3,$F0,$19
	.dc.b	$3E,$9B,$F4,$75,$FC,$7D,$FE,$41,$84,$C7,$F5,$5D,$16,$3F,$CA,$37
	.dc.b	$FE,$D5,$9F,$F1,$7B,$63,$FB,$31,$FE,$00,$3C,$3E,$4C,$9F,$53,$63
	.dc.b	$D5,$92,$3F,$C8,$67,$EF,$63,$E1,$38,$EC,$FD,$CF,$4F,$B0,$97,$F1
	.dc.b	$5B,$5F,$27,$AF,$73,$FC,$0D,$EB,$F5,$12,$30,$7F,$19,$3E,$D6,$08
	.dc.b	$D5,$7C,$F8,$C4,$FF,$93,$9B,$19,$9F,$EC,$1D,$9F,$81,$8F,$F5,$B3
	.dc.b	$5F,$F6,$03,$5F,$F7,$DB,$5F,$F8,$00,$F1,$F7,$DF,$49,$AC,$7F,$D5
	.dc.b	$91,$8A,$DF,$26,$2B,$59,$7F,$AC,$D4,$CF,$A7,$7C,$87,$E0,$63,$FD
	.dc.b	$81,$3A,$53,$AE,$2E,$51,$E4,$F4,$7B,$3C,$EF,$96,$5E,$38,$E3,$18
	.dc.b	$C3,$A6,$1C,$E7,$B9,$CF,$1E,$AF,$0C,$6C,$EF,$0E,$8D,$BF,$8F,$8F
	.dc.b	$8A,$F8,$1D,$0B,$5C,$31,$F1,$47,$73,$57,$CD,$F5,$61,$E0,$9D,$C3
	.dc.b	$46,$1D,$EF,$84,$F6,$C7,$DB,$7C,$2F,$79,$73,$C6,$5F,$82,$B5,$8F
	.dc.b	$37,$BE,$3D,$0E,$B8,$72,$AE,$17,$CB,$2F,$63,$F2,$27,$51,$89,$D9
	.dc.b	$1F,$A9,$1A,$CB,$CF,$5C,$DB,$5F,$26,$74,$FC,$43,$94,$6A,$B9,$E3
	.dc.b	$E9,$BF,$6D,$7E,$24,$FA,$65,$C3,$F0,$0D,$C7,$F2,$0E,$7E,$4B,$FC
	.dc.b	$8A,$F1,$3F,$22,$E1,$D7,$9F,$FC,$FE,$93,$99,$FC,$6B,$CB,$1F,$C3
	.dc.b	$B7,$2C,$DF,$E3,$3E,$07,$F3,$D5,$A8,$FC,$7D,$6E,$FF,$4C,$D7,$F4
	.dc.b	$75,$C2,$F9,$C7,$E2,$A7,$58,$79,$2F,$B4,$F4,$63,$B7,$2D,$26,$1E
	.dc.b	$37,$F2,$B8,$77,$7C,$73,$CA,$74,$7E,$92,$75,$3F,$22,$75,$87,$9A
	.dc.b	$B5,$38,$39,$9E,$08,$FD,$AD,$72,$BF,$54,$E4,$DD,$72,$70,$77,$2C
	.dc.b	$27,$96,$1F,$10,$98,$7C,$0B,$F4,$E1,$E5,$26,$1F,$C5,$6F,$D7,$F4
	.dc.b	$EF,$6A,$D4,$AF,$E0,$F3,$96,$CF,$F8,$0D,$46,$5F,$C6,$35,$FC,$B3
	.dc.b	$5F,$E7,$FE,$4F,$F8,$2F,$B3,$FA,$93,$79,$4D,$24,$FF,$47,$3F,$C1
	.dc.b	$B1,$A7,$13,$C9,$FF,$29,$70,$FF,$BB,$B8,$BF,$C0,$DD,$0E,$4C,$57
	.dc.b	$F8,$5A,$3F,$88,$3C,$0F,$F2,$D4,$6A,$FF,$BF,$BE,$79,$E5,$3F,$E0
	.dc.b	$58,$C1,$97,$86,$1F,$E8,$27,$CA,$D7,$FC,$1C,$75,$9F,$95,$5F,$E0
	.dc.b	$C8,$E4,$7F,$84,$4F,$89,$E6,$47,$FB,$13,$5F,$EC,$CF,$80,$8F,$C4
	.dc.b	$FC,$7F,$FB,$63,$63,$CB,$F1,$D2,$30,$43,$FC,$3D,$29,$FA,$CD,$8F
	.dc.b	$E1,$14,$8C,$47,$F7,$7F,$0F,$E0,$91,$F1,$35,$78,$1F,$E6,$3C,$93
	.dc.b	$98,$FF,$33,$56,$2B,$EE,$FD,$33,$E0,$3E,$A3,$63,$31,$8F,$8E,$CB
	.dc.b	$23,$27,$28,$FC,$0A,$F8,$D4,$7E,$13,$CA,$FF,$74,$FC,$0D,$EA,$3C
	.dc.b	$2B,$77,$F9,$33,$95,$FE,$8E,$33,$1F,$CB,$5A,$FF,$37,$6B,$FE,$01
	.dc.b	$77,$9F,$FB,$D1,$BC,$79,$AF,$25,$3D,$7F,$26,$72,$3F,$D0,$D9,$7C
	.dc.b	$37,$FD,$04,$3C,$7D,$3F,$E8,$58,$D6,$5F,$E7,$76,$E3,$FE,$18,$D3
	.dc.b	$FC,$33,$7F,$9A,$FA,$9F,$CD,$F9,$53,$1F,$CD,$C6,$5F,$CE,$C6,$5F
	.dc.b	$BB,$7F,$E3,$C9,$F9,$FF,$2D,$FE,$ED,$9F,$91,$B3,$DB,$F9,$86,$BD
	.dc.b	$89,$3F,$E8,$FA,$E0,$FF,$57,$6B,$FD,$85,$AF,$F8,$AD,$DE,$AC,$71
	.dc.b	$5D,$2B,$FD,$3D,$5F,$98,$AF,$C7,$FB,$7F,$B8,$FA,$59,$5F,$A9,$8F
	.dc.b	$E0,$5C,$A7,$31,$EE,$3F,$89,$CE,$A3,$F8,$6D,$3E,$8D,$86,$1B,$7F
	.dc.b	$C6,$E7,$53,$9F,$17,$4C,$9D,$57,$07,$C1,$5D,$6B,$A7,$37,$83,$E8
	.dc.b	$FF,$A2,$57,$56,$BC,$2F,$F3,$6F,$49,$FF,$5E,$5F,$9D,$B3,$A9,$FF
	.dc.b	$07,$4E,$5F,$F0,$43,$5E,$93,$E2,$7C,$FF,$2D,$FB,$68,$FE,$55,$F1
	.dc.b	$1F,$0E,$7F,$9B,$B5,$F9,$5D,$AE,$BF,$06,$1F,$EE,$76,$F0,$FF,$AF
	.dc.b	$EF,$84,$FF,$BA,$DB,$1F,$EC,$39,$DD,$FF,$46,$F9,$1F,$E6,$64,$7F
	.dc.b	$99,$EF,$5F,$0F,$72,$71,$8F,$F1,$5A,$FE,$0F,$F9,$73,$D4,$FC,$DF
	.dc.b	$A3,$31,$FE,$05,$39,$6C,$FF,$65,$8C,$47,$F8,$BD,$AF,$F8,$E5,$AF
	.dc.b	$FA,$0D,$DD,$FE,$66,$3F,$D1,$71,$8C,$F5,$5F,$C6,$BE,$85,$F0,$FF
	.dc.b	$D4,$3D,$9F,$ED,$44,$C3,$FD,$50,$FF,$6D,$F7,$24,$7F,$AA,$DA,$FF
	.dc.b	$AB,$FC,$DF,$EE,$44,$7F,$D6,$0F,$E1,$BE,$79,$EE,$52,$4F,$E1,$23
	.dc.b	$2F,$F4,$6F,$BC,$7C,$5F,$F7,$F6,$C6,$63,$07,$F8,$07,$E0,$24,$7F
	.dc.b	$AF,$DA,$FF,$B0,$FC,$0F,$F0,$31,$1F,$F6,$23,$FE,$09,$F4,$E4,$7F
	.dc.b	$B1,$DA,$FF,$B2,$FE,$07,$FC,$1C,$47,$FD,$98,$FF,$84,$7E,$56,$47
	.dc.b	$FB,$3D,$AF,$FC,$01,$C5,$EE,$FA,$C9,$1E,$F7,$FC,$35,$E5,$C8,$FF
	.dc.b	$6C,$35,$FF,$3C,$76,$BE,$6F,$F4,$51,$FE,$C3,$8C,$78,$3F,$E2,$59
	.dc.b	$CC,$FF,$B3,$27,$2F,$EC,$BE,$6F,$F7,$AF,$FB,$2F,$A9,$18,$8F,$F6
	.dc.b	$9C,$63,$D9,$FD,$FC,$E6,$7F,$B9,$BF,$CC,$DE,$9E,$4E,$88,$FF,$80
	.dc.b	$3E,$29,$EA,$FF,$8E,$5B,$19,$8C,$19,$F5,$24,$71,$6B,$F0,$A7,$FB
	.dc.b	$F3,$FC,$09,$B8,$FE,$87,$24,$63,$BB,$E1,$2F,$3F,$FC,$6D,$F3,$7F
	.dc.b	$E9,$EB,$E0,$FF,$8F,$30,$FE,$1A,$D8,$CF,$BF,$C0,$FF,$29,$47,$FA
	.dc.b	$9A,$7E,$93,$5F,$A0,$D5,$8A,$3C,$18,$EA,$BA,$57,$F9,$53,$E5,$5F
	.dc.b	$07,$FE,$4B,$F9,$B7,$C9,$7F,$93,$7E,$5F,$F8,$BE,$66,$CE,$B0,$FF
	.dc.b	$2E,$B5,$FC,$4B,$5F,$C4,$35,$FC,$33,$5F,$3F,$AB,$E9,$AF,$F1,$06
	.dc.b	$5F,$E7,$18,$C1,$1F,$F3,$03,$EC,$9E,$31,$8F,$7B,$FC,$53,$F3,$4F
	.dc.b	$1B,$FC,$77,$4E,$DF,$F0,$B2,$3F,$C3,$15,$5D,$7B,$94,$FF,$35,$B8
	.dc.b	$38,$5F,$5E,$2E,$57,$D2,$53,$95,$F3,$59,$BE,$6D,$C3,$C0,$F1,$AE
	.dc.b	$0D,$C7,$D2,$EE,$FD,$0D,$7A,$DF,$A6,$B8,$B6,$B8,$BE,$BB,$EE,$DA
	.dc.b	$E6,$EF,$1B,$1E,$EA,$F9,$38,$F7,$25,$6E,$7E,$21,$F0,$27,$BB,$F3
	.dc.b	$63,$DE,$77,$7C,$E7,$5A,$EE,$F1,$6C,$70,$8E,$57,$E8,$78,$1C,$8E
	.dc.b	$07,$D9,$32,$5C,$3E,$93,$6F,$F0,$6E,$7C,$F7,$CD,$F1,$8F,$BB,$18
	.dc.b	$71,$5F,$DC,$67,$51,$F7,$A7,$4E,$3E,$17,$FD,$A0,$9A,$47,$FD,$A8
	.dc.b	$FC,$0C,$BF,$DB,$9C,$BF,$0A,$7F,$29,$39,$F2,$BF,$B7,$6D,$F8,$D9
	.dc.b	$C9,$F0,$6F,$8B,$F0,$CC,$C7,$E4,$E3,$2F,$18,$F2,$1F,$62,$38,$ED
	.dc.b	$9B,$FC,$DE,$DF,$CD,$F0,$FE,$FD,$39,$BF,$04,$79,$27,$E4,$57,$9F
	.dc.b	$1F,$55,$76,$C7,$C5,$7B,$8F,$AD,$5B,$7D,$06,$EF,$C1,$7C,$A7,$57
	.dc.b	$E1,$39,$C6,$6B,$C6,$FE,$63,$85,$F3,$7F,$E5,$D8,$FF,$78,$B6,$7F
	.dc.b	$C7,$78,$3C,$2F,$F0,$D6,$9F,$CE,$EB,$4D,$F0,$B3,$87,$F5,$D9,$FF
	.dc.b	$42,$3D,$23,$38,$7F,$30,$9C,$BF,$E4,$8E,$67,$CD,$FF,$89,$E3,$FB
	.dc.b	$54,$62,$FF,$DA,$ED,$8F,$EE,$2D,$77,$3D,$B3,$FE,$65,$56,$B8,$63
	.dc.b	$FD,$4D,$FF,$46,$47,$F6,$C3,$FB,$A7,$CD,$E9,$B7,$CE,$1F,$3B,$62
	.dc.b	$4F,$FB,$AA,$7C,$58,$F0,$8C,$4F,$AE,$30,$7F,$99,$A7,$C2,$D8,$ED
	.dc.b	$3E,$9F,$61,$F4,$57,$E1,$30,$D9,$FF,$38,$C7,$5B,$FE,$B7,$5F,$EE
	.dc.b	$D9,$FF,$7E,$B6,$3F,$C2,$ED,$7F,$C2,$CB,$E8,$7C,$07,$58,$FF,$3D
	.dc.b	$C6,$67,$E3,$47,$F6,$8F,$25,$E4,$FC,$F6,$CE,$B7,$FF,$12,$79,$19
	.dc.b	$AF,$C0,$F0,$3E,$77,$F6,$78,$FF,$87,$DA,$FF,$87,$9A,$A3,$87,$47
	.dc.b	$BD,$E5,$E8,$78,$7E,$9F,$AD,$A3,$19,$F8,$F4,$7D,$2F,$9C,$F4,$7D
	.dc.b	$2D,$9F,$47,$AE,$F9,$AF,$99,$C0,$8C,$4F,$F2,$18,$C4,$EA,$73,$B7
	.dc.b	$F9,$13,$17,$D2,$FC,$5F,$A6,$F8,$CD,$C7,$C9,$CB,$F2,$4F,$95,$CD
	.dc.b	$63,$D1,$CD,$E2,$C3,$E9,$63,$D6,$30,$67,$E0,$FE,$97,$C6,$FD,$36
	.dc.b	$DF,$C1,$FA,$5F,$AC,$7D,$7F,$86,$B1,$F0,$3E,$87,$D4,$47,$DD,$BA
	.dc.b	$3E,$87,$D1,$3C,$CF,$B6,$BE,$66,$C6,$1F,$C2,$7C,$5C,$C3,$E9,$7C
	.dc.b	$65,$DF,$FD,$8F,$C0,$F9,$8F,$71,$2B,$E7,$46,$5B,$E1,$7E,$3F,$C4
	.dc.b	$FC,$83,$63,$F2,$B6,$FA,$2F,$D4,$C6,$1B,$1F,$EF,$0F,$B9,$78,$1F
	.dc.b	$C5,$23,$87,$3D,$AD,$C7,$F6,$76,$BF,$E2,$16,$BF,$E2,$56,$BC,$AF
	.dc.b	$C7,$E5,$7F,$1E,$F0,$BF,$8E,$57,$89,$FE,$3D,$19,$9F,$F4,$27,$CB
	.dc.b	$7E,$2A,$7F,$CD,$0D,$8F,$F4,$46,$BF,$C2,$27,$3D,$DF,$C7,$EB,$58
	.dc.b	$FF,$97,$1B,$DF,$FE,$95,$C3,$55,$FC,$97,$E7,$9F,$45,$4E,$5F,$13
	.dc.b	$6B,$F7,$53,$A6,$CE,$9A,$7F,$04,$F9,$7F,$F0,$2F,$89,$FB,$EF,$A2
	.dc.b	$7C,$CF,$F4,$09,$CD,$7F,$0A,$D9,$5A,$8E,$5E,$BB,$0E,$58,$71,$8C
	.dc.b	$4F,$1A,$E1,$BF,$F0,$58,$FF,$52,$C6,$30,$FE,$AE,$D8,$FF,$2D,$63
	.dc.b	$E2,$AE,$07,$F0,$D9,$D3,$7E,$47,$D5,$F3,$7E,$B6,$DF,$85,$B3,$CB
	.dc.b	$0F,$5C,$61,$FB,$3F,$53,$F5,$7E,$8F,$EB,$36,$FF,$93,$71,$A3,$18
	.dc.b	$79,$38,$D3,$9A,$DD,$FE,$5E,$7F,$D5,$11,$FA,$06,$D7,$B0,$7E,$A7
	.dc.b	$FC,$82,$75,$F0,$DC,$4F,$E9,$13,$FE,$A2,$6D,$72,$8F,$F6,$EC,$60
	.dc.b	$FD,$37,$D2,$7F,$BA,$2B,$99,$FB,$07,$13,$E9,$C5,$F2,$3F,$95,$3F
	.dc.b	$2D,$F0,$B6,$FE,$0C,$5A,$E1,$37,$55,$7D,$87,$FD,$28,$FE,$3D,$7C
	.dc.b	$7F,$E3,$5C,$D7,$DC,$E2,$FE,$91,$39,$F8,$0D,$36,$F9,$3C,$7C,$A9
	.dc.b	$1D,$7E,$0E,$E1,$CF,$1F,$E8,$53,$FE,$61,$C7,$FA,$A5,$FF,$71,$6C
	.dc.b	$66,$77,$5F,$E6,$F9,$CB,$F5,$75,$75,$F6,$AF,$F1,$04,$62,$FF,$63
	.dc.b	$F1,$9F,$B7,$E0,$FD,$C3,$63,$27,$ED,$A3,$2D,$3F,$1B,$F4,$DF,$C1
	.dc.b	$7E,$97,$F8,$33,$63,$27,$E4,$A3,$2D,$3F,$2B,$F2,$BF,$D8,$B8,$CD
	.dc.b	$24,$7F,$63,$8C,$BF,$9C,$FA,$0F,$E2,$93,$FC,$55,$FE,$2D,$E8,$49
	.dc.b	$FD,$0F,$D2,$5D,$FF,$11,$3F,$C6,$FE,$67,$1B,$FF,$22,$7F,$8F,$7C
	.dc.b	$8F,$8C,$98,$78,$5B,$7F,$91,$F9,$CD,$CF,$EB,$9F,$E4,$DF,$11,$EB
	.dc.b	$24,$7A,$5B,$3F,$99,$F9,$5D,$CF,$ED,$9F,$E5,$C4,$7F,$73,$36,$78
	.dc.b	$47,$EF,$3D,$4C,$65,$FC,$E9,$FE,$9D,$F4,$4F,$8F,$FD,$8B,$6B,$FD
	.dc.b	$4E,$DB,$FF,$57,$35,$FF,$5C,$39,$BF,$F3,$33,$FD,$CE,$7F,$B0,$B9
	.dc.b	$7F,$DE,$CE,$67,$F8,$27,$D2,$7F,$6F,$BF,$F5,$4F,$CC,$FF,$BE,$E3
	.dc.b	$13,$FE,$B3,$8E,$3C,$1F,$B2,$8C,$01,$3C,$3E,$3D,$5C,$FD,$7F,$75
	.dc.b	$70,$36,$E2,$77,$1C,$F1,$61,$DE,$73,$3C,$31,$F3,$BE,$1C,$7C,$18
	.dc.b	$F7,$9C,$CB,$0E,$79,$62,$F7,$3C,$72,$F1,$61,$C6,$3C,$D7,$C1,$F4
	.dc.b	$57,$07,$C4,$62,$B8,$BE,$9C,$BD,$17,$CF,$2F,$4D,$F1,$39,$61,$D6
	.dc.b	$33,$87,$2C,$3A,$C6,$32,$F5,$D6,$F0,$E8,$4A,$F5,$4E,$B1,$F0,$C6
	.dc.b	$0F,$34,$7C,$1C,$3B,$5F,$39,$D6,$1D,$DD,$9F,$0A,$F9,$BE,$4C,$3C
	.dc.b	$0D,$3E,$06,$5D,$71,$91,$CA,$FE,$CE,$1D,$E3,$81,$F2,$8F,$C5,$1E
	.dc.b	$72,$99,$3E,$A1,$8C,$9E,$CF,$FB,$4E,$36,$7E,$49,$C3,$B6,$BF,$9D
	.dc.b	$6B,$F8,$97,$95,$7D,$66,$E3,$BF,$4B,$F1,$1F,$11,$E6,$AD,$47,$E2
	.dc.b	$EB,$77,$CE,$36,$E5,$8F,$E2,$A3,$6F,$EA,$2B,$A5,$F7,$8D,$57,$0B
	.dc.b	$FC,$4D,$6A,$7B,$1A,$24,$69,$E1,$5C,$E3,$CF,$1F,$9A,$9E,$4F,$C7
	.dc.b	$AD,$BD,$27,$93,$D2,$B9,$11,$FD,$85,$72,$C3,$E9,$DF,$29,$E8,$77
	.dc.b	$31,$1A,$7A,$1C,$8C,$E1,$F1,$23,$58,$78,$8F,$49,$30,$F1,$57,$C9
	.dc.b	$3F,$7F,$F4,$9F,$EB,$57,$FD,$02,$BE,$27,$F1,$BC,$1F,$C6,$65,$39
	.dc.b	$77,$E1,$6B,$7F,$C9,$CF,$39,$FC,$6B,$5F,$CA,$B5,$DC,$FE,$49,$FF
	.dc.b	$38,$7B,$BC,$DF,$F3,$96,$F7,$C9,$FF,$1D,$B7,$B7,$FD,$55,$3F,$E4
	.dc.b	$2A,$C1,$1F,$F2,$2F,$16,$CC,$3F,$E2,$D8,$C3,$A3,$FB,$94,$78,$EF
	.dc.b	$FC,$92,$DA,$FE,$67,$3F,$AE,$97,$D1,$BE,$5B,$C9,$30,$FF,$7D,$46
	.dc.b	$5B,$1D,$63,$11,$CD,$AF,$EF,$27,$AC,$EF,$D5,$F8,$54,$F4,$29,$88
	.dc.b	$97,$E0,$6B,$FE,$63,$97,$A3,$CB,$C6,$E7,$EA,$4A,$F3,$C6,$5F,$F7
	.dc.b	$B5,$23,$D0,$48,$CF,$8F,$E4,$F9,$99,$1F,$02,$3C,$E6,$1F,$14,$61
	.dc.b	$FE,$C5,$C9,$FD,$97,$57,$D0,$BE,$1F,$F0,$0B,$E1,$E9,$88,$97,$CB
	.dc.b	$9F,$FB,$65,$F2,$AD,$CE,$44,$76,$AF,$C7,$79,$DF,$DC,$FC,$3F,$C8
	.dc.b	$53,$E5,$FC,$0F,$FB,$91,$AF,$F9,$94,$47,$26,$9F,$E5,$D2,$9F,$E5
	.dc.b	$92,$9F,$CE,$27,$C0,$EE,$39,$11,$FF,$62,$0F,$1F,$B7,$97,$DF,$CB
	.dc.b	$93,$A8,$CB,$C2,$32,$F9,$E3,$F1,$3F,$47,$23,$F2,$AF,$E3,$3E,$93
	.dc.b	$11,$F8,$78,$BC,$1F,$57,$91,$5F,$1B,$E1,$BF,$DA,$B7,$F3,$42,$F9
	.dc.b	$FD,$1B,$E7,$93,$FB,$F9,$E2,$FF,$52,$6B,$FD,$69,$AF,$F5,$E7,$78
	.dc.b	$7F,$13,$6F,$1E,$6B,$EC,$FD,$0F,$FA,$A2,$FF,$63,$F3,$5F,$A8,$FA
	.dc.b	$39,$18,$21,$FA,$57,$2F,$E9,$E3,$F8,$96,$2F,$9A,$7C,$76,$1F,$ED
	.dc.b	$75,$FE,$F3,$6B,$FE,$DB,$76,$E6,$38,$79,$72,$30,$71,$F8,$EE,$84
	.dc.b	$7F,$03,$F0,$1F,$8D,$30,$F5,$8C,$3A,$39,$10,$FC,$B4,$7E,$AA,$24
	.dc.b	$7E,$8F,$AB,$F5,$F1,$FC,$33,$DB,$CC,$FF,$BA,$7B,$53,$8C,$FF,$05
	.dc.b	$5F,$C4,$BC,$B9,$7F,$C7,$9F,$E2,$BE,$56,$23,$F8,$84,$5A,$FC,$54
	.dc.b	$FD,$6E,$8F,$D7,$FD,$1B,$D7,$A6,$70,$FC,$67,$1F,$FA,$A1,$FE,$8C
	.dc.b	$D7,$FC,$32,$D7,$FC,$42,$D7,$FC,$4A,$EF,$0F,$EB,$4D,$E5,$CE,$1F
	.dc.b	$7B,$C5,$7C,$3F,$F2,$EE,$9F,$E5,$FF,$2B,$32,$FE,$77,$1F,$CC,$BD
	.dc.b	$0F,$E7,$44,$8F,$E4,$EF,$F3,$DF,$80,$FE,$6E,$48,$FE,$87,$F0,$1F
	.dc.b	$8C,$FA,$1F,$ED,$9B,$7F,$A1,$47,$D1,$F8,$AB,$F8,$C4,$63,$3E,$1B
	.dc.b	$7F,$63,$FA,$37,$C2,$7F,$57,$6D,$FF,$8F,$3A,$51,$88,$FE,$B0,$D7
	.dc.b	$FA,$77,$47,$E7,$5F,$C2,$35,$63,$FC,$D6,$30,$FF,$03,$F6,$7F,$BA
	.dc.b	$43,$DA,$BF,$35,$2F,$9C,$6B,$E4,$3F,$B6,$51,$A4,$9F,$44,$3E,$67
	.dc.b	$A9,$23,$CB,$F1,$BE,$0F,$98,$F0,$FC,$CC,$9F,$4C,$7C,$5F,$83,$F8
	.dc.b	$84,$8F,$89,$B2,$3C,$7F,$33,$EA,$24,$FC,$2F,$94,$FA,$DA,$FB,$BD
	.dc.b	$59,$3F,$53,$9F,$E0,$37,$83,$FC,$35,$C4,$94,$61,$B3,$FE,$70,$9D
	.dc.b	$4F,$F8,$3F,$EA,$1E,$8F,$F8,$7D,$B5,$FB,$76,$BF,$53,$9F,$FA,$FC
	.dc.b	$7F,$34,$8F,$DE,$7A,$6F,$63,$FC,$50,$D8,$CC,$63,$D9,$91,$F8,$79
	.dc.b	$47,$F8,$85,$B1,$98,$C7,$A7,$C2,$3C,$47,$F6,$C9,$DB,$8E,$AF,$EB
	.dc.b	$13,$FE,$39,$FA,$3F,$F0,$FE,$BF,$DF,$FE,$93,$FC,$56,$DF,$81,$D7
	.dc.b	$C0,$4C,$7F,$35,$7F,$E2,$47,$FB,$EB,$5F,$EF,$6D,$C4,$F6,$DB,$75
	.dc.b	$CE,$BC,$98,$FF,$A9,$B1,$FF,$08,$7D,$27,$F4,$79,$F4,$3F,$A7,$8C
	.dc.b	$F8,$1F,$E5,$18,$D4,$70,$F9,$1F,$F2,$34,$71,$AF,$F1,$C3,$5F,$F0
	.dc.b	$9B,$5F,$F0,$8F,$DB,$F9,$5F,$F0,$9F,$BF,$FE,$5A,$7E,$43,$6F,$FC
	.dc.b	$49,$3B,$D4,$D2,$67,$FE,$65,$AF,$66,$9F,$EC,$4F,$A5,$7A,$7F,$BA
	.dc.b	$C9,$F4,$BF,$E1,$4F,$80,$F0,$37,$E1,$FF,$D9,$6F,$F8,$62,$F8,$EA
	.dc.b	$BE,$37,$D7,$0E,$97,$C6,$74,$D7,$67,$1B,$E6,$E0,$38,$8E,$04,$AE
	.dc.b	$D7,$D1,$B5,$C5,$F4,$9F,$EE,$22,$BD,$EF,$CB,$1C,$3C,$CF,$21,$1C
	.dc.b	$1D,$2F,$C1,$5F,$E4,$3C,$7E,$1C,$E4,$91,$93,$C8,$53,$D9,$8F,$AE
	.dc.b	$3C,$F1,$FE,$F4,$8F,$82,$DC,$3E,$A3,$71,$FF,$51,$4B,$DC,$7C,$C6
	.dc.b	$1B,$5F,$EF,$89,$D4,$E4,$FF,$55,$47,$D7,$F8,$BA,$9A,$47,$6F,$A9
	.dc.b	$FE,$7F,$EC,$B6,$C4,$73,$3E,$C6,$3F,$EC,$58,$CC,$75,$33,$3C,$5F
	.dc.b	$BC,$71,$D7,$FC,$C7,$5C,$63,$27,$1D,$B3,$5C,$5F,$F3,$3C,$67,$5F
	.dc.b	$F6,$9C,$E5,$FC,$A4,$4C,$7D,$57,$D7,$0E,$31,$8B,$F2,$46,$2B,$CF
	.dc.b	$8F,$AA,$B4,$D8,$C9,$E6,$8C,$B4,$CC,$F3,$3D,$33,$F8,$4E,$2F,$F4
	.dc.b	$2E,$5F,$DF,$63,$FD,$2B,$92,$31,$7F,$E9,$48,$7A,$18,$BE,$37,$FE
	.dc.b	$37,$9E,$0D,$BD,$47,$08,$C1,$FE,$3D,$FA,$47,$CF,$FF,$BE,$71,$FF
	.dc.b	$52,$C6,$27,$FC,$0C,$D8,$FF,$6E,$35,$FF,$69,$39,$D7,$FD,$6D,$D5
	.dc.b	$FE,$D1,$AE,$D7,$E7,$8C,$5F,$F4,$17,$FD,$9F,$3F,$E4,$F6,$C7,$F4
	.dc.b	$5B,$FC,$74,$E7,$D6,$E4,$79,$E7,$F9,$EB,$73,$FE,$7A,$DB,$E7,$3F
	.dc.b	$EE,$77,$AE,$3F,$EC,$BD,$FF,$C7,$78,$75,$3F,$CC,$0E,$2B,$FA,$B7
	.dc.b	$D1,$BC,$CF,$A0,$BC,$DB,$18,$65,$FF,$8B,$67,$4D,$8F,$34,$62,$BF
	.dc.b	$2A,$7F,$B4,$3D,$0F,$F6,$D5,$7F,$8B,$BE,$23,$9C,$FF,$67,$9C,$BE
	.dc.b	$4F,$2B,$C6,$79,$3C,$DB,$7F,$E7,$9C,$1F,$45,$E0,$24,$79,$5F,$91
	.dc.b	$86,$A3,$FC,$7B,$1E,$EC,$7F,$1D,$39,$C7,$C4,$E0,$F8,$65,$AF,$F2
	.dc.b	$2F,$C1,$74,$23,$FE,$38,$7B,$FC,$37,$43,$E5,$FC,$95,$9F,$F2,$6F
	.dc.b	$C2,$78,$0F,$9F,$F4,$36,$3F,$CA,$33,$E8,$23,$FE,$44,$7D,$7C,$FE
	.dc.b	$83,$F4,$DF,$06,$F8,$5C,$57,$F9,$6A,$FD,$A6,$BC,$9B,$1F,$E5,$E3
	.dc.b	$E2,$90,$FF,$28,$9F,$8F,$FA,$DF,$C3,$73,$FF,$1B,$EA,$46,$27,$FB
	.dc.b	$24,$31,$7C,$5F,$EE,$6B,$5C,$2A,$E3,$F3,$79,$3D,$3F,$2B,$71,$FF
	.dc.b	$5A,$5F,$AD,$FF,$3D,$36,$FF,$17,$7E,$2D,$7C,$C7,$9F,$E7,$7C,$07
	.dc.b	$C4,$F4,$AE,$1F,$EF,$5F,$A1,$75,$7F,$10,$C5,$18,$9F,$F7,$D7,$C0
	.dc.b	$FF,$BE,$9E,$15,$B9,$D5,$7E,$70,$E3,$1F,$B7,$6C,$E5,$A6,$E7,$8F
	.dc.b	$07,$FA,$7F,$EB,$1B,$F5,$C3,$F9,$1C,$A7,$38,$7C,$4E,$2F,$98,$7F
	.dc.b	$25,$FA,$AF,$A1,$19,$26,$5F,$E5,$1E,$0F,$F5,$2C,$7F,$B7,$E7,$73
	.dc.b	$9A,$E5,$E0,$FD,$7F,$4E,$DF,$D2,$6D,$73,$6E,$5F,$83,$C5,$1F,$2F
	.dc.b	$EC,$5F,$37,$4F,$9D,$FA,$37,$C0,$C7,$FA,$BC,$E9,$FF,$16,$61,$FD
	.dc.b	$9F,$14,$E5,$FF,$36,$98,$BE,$7B,$75,$BE,$27,$48,$C4,$F0,$C3,$C0
	.dc.b	$FF,$8E,$6F,$A6,$1F,$9B,$E8,$FF,$1E,$61,$F9,$0F,$9F,$FD,$5D,$FE
	.dc.b	$65,$B1,$93,$BC,$7E,$6E,$78,$BF,$3E,$7E,$A7,$C6,$4A,$DF,$43,$E9
	.dc.b	$3F,$9A,$C7,$EA,$38,$95,$BC,$3F,$D1,$11,$89,$E3,$5C,$27,$F5,$53
	.dc.b	$97,$E2,$DB,$E5,$3F,$38,$DB,$FC,$DF,$C1,$62,$64,$7E,$56,$75,$5F
	.dc.b	$EE,$79,$FC,$D4,$62,$3F,$5F,$1F,$A1,$F2,$B8,$C7,$D4,$E8,$FE,$91
	.dc.b	$7F,$8D,$F8,$EF,$E0,$1F,$43,$F9,$56,$C6,$4F,$DE,$C6,$5A,$7F,$BF
	.dc.b	$BE,$91,$EE,$7E,$8B,$EC,$9E,$27,$E9,$27,$35,$F4,$B9,$15,$A8,$E5
	.dc.b	$7C,$6F,$83,$63,$EC,$E6,$FA,$47,$CA,$FF,$03,$6C,$E9,$FC,$3B,$FE
	.dc.b	$2E,$73,$3F,$94,$6B,$FD,$79,$AE,$E7,$F8,$24,$62,$3F,$D1,$4F,$2C
	.dc.b	$7F,$40,$FE,$AA,$BC,$15,$E3,$EB,$FF,$12,$FC,$DF,$F2,$9E,$6B,$4F
	.dc.b	$9C,$CF,$27,$E0,$F2,$FF,$28,$5F,$F9,$66,$73,$7C,$1B,$7F,$9A,$9D
	.dc.b	$3F,$8E,$F8,$8F,$E7,$1E,$4F,$F8,$C7,$A7,$FD,$D5,$3F,$CD,$23,$13
	.dc.b	$A8,$E0,$64,$FE,$1D,$F2,$BF,$E1,$7E,$A3,$49,$1F,$CA,$74,$4E,$A3
	.dc.b	$FA,$0F,$CB,$7F,$1D,$AF,$E9,$4F,$F4,$BF,$2E,$4F,$E9,$3E,$06,$F1
	.dc.b	$FF,$1F,$9B,$C1,$FD,$5F,$5F,$F4,$55,$FF,$4C,$9F,$E5,$1E,$EF,$94
	.dc.b	$91,$E3,$CF,$FC,$95,$87,$F4,$E6,$E1,$FC,$7F,$8F,$8D,$7F,$56,$7E
	.dc.b	$35,$FF,$8D,$71,$E3,$87,$F9,$D3,$5F,$E0,$13,$92,$5F,$EF,$35,$E2
	.dc.b	$D3,$FB,$0F,$D2,$7F,$69,$3E,$94,$58,$C9,$2D,$97,$BE,$75,$1F,$CA
	.dc.b	$3E,$87,$F8,$AB,$FE,$32,$6B,$FE,$88,$6B,$FE,$DD,$6A,$7F,$8E,$CF
	.dc.b	$F3,$5F,$CF,$BF,$A4,$BF,$CD,$E3,$37,$FE,$92,$9C,$BF,$E9,$4F,$A4
	.dc.b	$7A,$6F,$8C,$FF,$5A,$65,$98,$C6,$5F,$EB,$58,$F2,$70,$FF,$AA,$7E
	.dc.b	$13,$F1,$33,$FE,$F1,$C3,$9E,$1C,$A3,$18,$72,$AD,$C2,$B6,$F5,$3D
	.dc.b	$D7,$D7,$0F,$16,$1C,$E7,$86,$1C,$E7,$31,$C4,$F8,$95,$C5,$F0,$D7
	.dc.b	$1B,$EF,$87,$89,$B7,$EC,$C7,$58,$F5,$C3,$89,$D6,$B6,$F9,$A7,$78
	.dc.b	$48,$D5,$71,$7C,$73,$9C,$3C,$13,$9B,$E7,$5A,$38,$DE,$F0,$F6,$DF
	.dc.b	$C8,$CD,$97,$B6,$FE,$1D,$ED,$F3,$D6,$E7,$58,$76,$26,$3E,$5C,$39
	.dc.b	$E5,$C7,$0E,$67,$88,$F9,$91,$A7,$18,$7A,$B1,$E1,$39,$8C,$4F,$13
	.dc.b	$FD,$1B,$87,$DE,$6E,$1F,$8C,$6D,$FF,$8C,$E5,$EB,$3D,$2C,$36,$BF
	.dc.b	$C7,$93,$A9,$C9,$23,$EB,$4E,$F0,$FB,$CD,$D3,$F4,$1F,$25,$FE,$F3
	.dc.b	$7F,$45,$19,$AF,$CD,$35,$FC,$B3,$70,$57,$19,$DE,$1F,$8D,$AD,$1B
	.dc.b	$9D,$4F,$29,$D1,$1F,$C6,$46,$AB,$D9,$1E,$A9,$D6,$1F,$97,$AD,$5F
	.dc.b	$69,$CE,$1E,$CB,$A7,$90,$E4,$76,$32,$78,$CC,$DF,$57,$38,$78,$A3
	.dc.b	$E4,$D7,$EA,$70,$F0,$BA,$C3,$DD,$76,$38,$CE,$A3,$AE,$1F,$3A,$79
	.dc.b	$47,$EE,$FE,$91,$F1,$28,$FF,$00,$FD,$C7,$F1,$D6,$ED,$97,$FC,$9F
	.dc.b	$5C,$1F,$F0,$3B,$5F,$F4,$03,$5F,$F6,$EB,$5E,$4D,$7F,$D4,$1E,$93
	.dc.b	$E9,$BB,$72,$7F,$9A,$5B,$C5,$FE,$AA,$9D,$57,$F9,$03,$E2,$32,$F2
	.dc.b	$AB,$D1,$FE,$B2,$6D,$70,$9F,$E8,$77,$FE,$B4,$E9,$61,$C1,$FC,$F4
	.dc.b	$E8,$FF,$5A,$3A,$7C,$07,$F9,$43,$6E,$07,$F9,$1E,$38,$C7,$F5,$D6
	.dc.b	$BD,$E3,$FC,$99,$D7,$CE,$3F,$D2,$5E,$CF,$98,$F4,$C6,$74,$FF,$66
	.dc.b	$3B,$9F,$ED,$0D,$7F,$CC,$D8,$74,$3F,$CB,$19,$3E,$23,$E0,$92,$7F
	.dc.b	$CD,$FF,$40,$F9,$AC,$47,$FA,$8E,$6E,$3F,$ED,$BA,$D4,$FF,$A9,$73
	.dc.b	$F9,$7F,$33,$23,$04,$3F,$C0,$91,$EE,$23,$FE,$1B,$9F,$F6,$6B,$FE
	.dc.b	$DA,$F6,$3E,$A9,$F6,$23,$31,$F4,$78,$51,$8C,$3F,$D5,$95,$C2,$BD
	.dc.b	$7E,$CF,$E0,$BC,$7F,$E3,$0D,$FB,$27,$C4,$53,$59,$F3,$E4,$FD,$05
	.dc.b	$F0,$C3,$FC,$81,$7C,$1B,$7C,$30,$FC,$7C,$3E,$45,$EC,$61,$7A,$F4
	.dc.b	$C4,$49,$FF,$70,$4F,$06,$C7,$FB,$8D,$E8,$D8,$D4,$7F,$AF,$9B,$AF
	.dc.b	$FB,$0D,$B9,$FE,$D7,$E1,$1E,$BF,$E9,$29,$F4,$49,$C7,$C0,$BC,$EA
	.dc.b	$6E,$75,$97,$FA,$25,$AF,$E8,$EF,$FD,$E1,$3C,$23,$57,$FB,$0F,$8A
	.dc.b	$EF,$1F,$EB,$CE,$9F,$E1,$4F,$FB,$BF,$C0,$D6,$32,$35,$1F,$B6,$30
	.dc.b	$FF,$B7,$2D,$F4,$DF,$43,$E8,$9E,$E5,$3E,$9C,$EA,$BF,$D3,$2D,$7F
	.dc.b	$8C,$1F,$C7,$5C,$38,$6B,$F2,$6B,$E9,$47,$1D,$3F,$DE,$0D,$DF,$55
	.dc.b	$FE,$A6,$FA,$3F,$F5,$2B,$FE,$DA,$D7,$E9,$AE,$B4,$61,$BC,$7F,$9F
	.dc.b	$C3,$E4,$32,$B9,$39,$7F,$6F,$E4,$FF,$1A,$F9,$59,$18,$21,$FA,$7E
	.dc.b	$14,$6A,$3F,$55,$DB,$FD,$F1,$FF,$5E,$DF,$FB,$B1,$BE,$C7,$E7,$23
	.dc.b	$11,$FC,$84,$FC,$EF,$C1,$66,$64,$70,$39,$C6,$1E,$13,$CB,$76,$FF
	.dc.b	$D2,$63,$0D,$AF,$C4,$79,$9F,$CD,$BC,$F9,$87,$CB,$8C,$BF,$8B,$8C
	.dc.b	$BF,$C1,$78,$3F,$C0,$6F,$E3,$CF,$E7,$E4,$3E,$A7,$D0,$D7,$E3,$17
	.dc.b	$2F,$2B,$F5,$E5,$3F,$BE,$F3,$1E,$DF,$EA,$3E,$0E,$5F,$F6,$68,$D1
	.dc.b	$FB,$99,$DC,$7E,$40,$FE,$D8,$E1,$C3,$5F,$CA,$B9,$C3,$C7,$08,$C6
	.dc.b	$9F,$DC,$5B,$1F,$DD,$FE,$D9,$F1,$9F,$D2,$3E,$02,$4F,$F7,$BC,$3F
	.dc.b	$A4,$C7,$F6,$4F,$A1,$7C,$6F,$F4,$EE,$BF,$EA,$64,$BF,$F5,$F3,$FD
	.dc.b	$5B,$E5,$3F,$A9,$92,$3A,$69,$FE,$07,$EB,$60,$DF,$FA,$CF,$D3,$CB
	.dc.b	$D9,$FA,$3F,$45,$1E,$3F,$82,$FE,$BF,$F4,$DF,$D9,$09,$5F,$EC,$F7
	.dc.b	$FB,$3F,$C2,$7F,$64,$24,$74,$CB,$C3,$F0,$1F,$DA,$3E,$12,$4E,$72
	.dc.b	$47,$C1,$F5,$FF,$B6,$FC,$FF,$F7,$22,$57,$FB,$75,$FE,$EB,$F3,$FF
	.dc.b	$DC,$89,$1D,$F7,$FE,$37,$CB,$FD,$D3,$E8,$7F,$BB,$92,$75,$DB,$FD
	.dc.b	$EB,$E3,$7F,$BF,$92,$3F,$DD,$4F,$F8,$0B,$E2,$3F,$BF,$92,$3D,$99
	.dc.b	$B9,$BF,$C0,$7F,$01,$27,$D8,$FF,$87,$A5,$C7,$78,$FF,$21,$5F,$C4
	.dc.b	$65,$18,$E3,$46,$27,$F2,$D1,$89,$D4,$FC,$89,$F1,$F0,$F7,$3C,$1C
	.dc.b	$3F,$E2,$EF,$A1,$69,$49,$5F,$56,$1A,$EC,$F1,$38,$8F,$0B,$5F,$E6
	.dc.b	$F9,$23,$11,$FE,$51,$8C,$69,$CE,$3B,$57,$8E,$3F,$C8,$3F,$36,$F9
	.dc.b	$9F,$F2,$1F,$D0,$3E,$5E,$9A,$89,$87,$C2,$AE,$0D,$8F,$91,$18,$8F
	.dc.b	$2B,$79,$3F,$84,$E4,$8C,$76,$7E,$95,$B5,$FD,$0F,$17,$CB,$79,$3B
	.dc.b	$7C,$FF,$33,$2B,$F3,$95,$FA,$B6,$CF,$D9,$8C,$47,$C6,$6B,$F1,$5A
	.dc.b	$FE,$D7,$E8,$3F,$AC,$3F,$D4,$DB,$91,$A7,$E3,$E3,$F5,$8E,$3C,$0F
	.dc.b	$F3,$5B,$AE,$8F,$F1,$0F,$D0,$7F,$A2,$36,$FF,$44,$7C,$FF,$F9,$3F
	.dc.b	$0F,$F4,$57,$C2,$FF,$52,$72,$FF,$A2,$4C,$28,$E0,$AB,$A3,$99,$FF
	.dc.b	$48,$46,$27,$71,$C9,$BE,$8F,$FA,$55,$E7,$3B,$9E,$3C,$8C,$3C,$78
	.dc.b	$79,$B0,$E1,$5E,$56,$C7,$FA,$8E,$F9,$36,$F9,$3E,$53,$91,$D0,$FF
	.dc.b	$38,$4F,$27,$D9,$1C,$30,$F2,$BC,$BE,$1F,$D3,$3F,$EB,$16,$BF,$E9
	.dc.b	$4C,$3D,$38,$E5,$B1,$E1,$3B,$9E,$17,$E2,$1A,$BE,$F3,$9C,$3C,$93
	.dc.b	$AA,$F6,$4F,$A7,$1F,$A3,$1F,$E2,$86,$BE,$03,$18,$FC,$27,$5D,$0F
	.dc.b	$A0,$FE,$23,$5F,$C1,$1B,$19,$9F,$EC,$93,$97,$FC,$C1,$F1,$BF,$E6
	.dc.b	$14,$D2,$3A,$3E,$06,$0F,$9B,$FE,$F3,$A7,$E2,$A7,$FD,$B5,$5F,$4E
	.dc.b	$B5,$CF,$F5,$67,$27,$2B,$E3,$A7,$F7,$7B,$E1,$18,$9F,$F2,$4C,$7D
	.dc.b	$17,$FB,$64,$7F,$82,$36,$F0,$47,$A0,$FB,$06,$AF,$83,$FE,$E9,$AE
	.dc.b	$B1,$E3,$8E,$18,$FE,$8B,$1F,$3C,$7D,$D3,$CF,$19,$C3,$C9,$1C,$CF
	.dc.b	$55,$73,$BE,$B1,$C4,$EB,$8F,$FB,$B2,$B8,$C7,$C2,$C7,$A5,$F9,$A3
	.dc.b	$15,$B8,$F8,$1C,$5F,$EC,$BF,$93,$FF,$13,$E7,$FC,$82,$91,$8A,$FF
	.dc.b	$02,$43,$E3,$F1,$87,$F9,$8E,$37,$3F,$DE,$5B,$1F,$D0,$1A,$FF,$3F
	.dc.b	$6B,$D2,$BF,$C4,$AF,$F2,$DF,$91,$AF,$F9,$BF,$A9,$18,$CB,$F9,$A3
	.dc.b	$63,$38,$7F,$3F,$8F,$2C,$FF,$39,$F3,$AD,$6A,$FF,$DA,$4F,$F8,$BA
	.dc.b	$FF,$C2,$0F,$F9,$71,$A7,$0F,$96,$E2,$C3,$FA,$61,$E1,$21,$83,$FC
	.dc.b	$35,$F4,$1F,$DF,$DE,$66,$39,$FC,$BE,$AF,$FB,$32,$3F,$D1,$DF,$0F
	.dc.b	$8F,$A4,$CB,$23,$FB,$DC,$EF,$2F,$ED,$6D,$7F,$B5,$35,$FF,$1D,$2F
	.dc.b	$A1,$F9,$0D,$9E,$71,$83,$3F,$15,$23,$BB,$6B,$BF,$C5,$F2,$24,$F9
	.dc.b	$1E,$BF,$43,$23,$CE,$D9,$F3,$F8,$7E,$22,$4F,$B1,$F2,$FC,$14,$8F
	.dc.b	$7B,$67,$DF,$C7,$EA,$24,$FC,$47,$DB,$C1,$23,$E3,$B6,$7F,$C2,$3E
	.dc.b	$F3,$E8,$3F,$14,$DA,$FF,$13,$B6,$77,$3A,$F6,$65,$69,$B3,$D3,$C8
	.dc.b	$E6,$49,$E6,$F6,$F8,$69,$1D,$9B,$3E,$4F,$67,$C6,$49,$F1,$BE,$6F
	.dc.b	$98,$91,$E6,$6C,$FB,$3E,$5B,$D6,$49,$F5,$BE,$EF,$4E,$47,$B9,$B3
	.dc.b	$FE,$9B,$F9,$27,$C4,$7E,$F2,$11,$8A,$FF,$3F,$F8,$4F,$BE,$F8,$CF
	.dc.b	$FA,$72,$75,$5F,$EC,$D6,$BF,$EF,$66,$A2,$38,$6D,$FF,$00,$7C,$2F
	.dc.b	$8F,$E7,$DD,$1E,$5A,$EF,$7F,$EA,$C9,$CE,$3F,$E8,$5E,$2F,$E0,$3E
	.dc.b	$F1,$FE,$06,$3E,$57,$2C,$9E,$07,$FA,$45,$D3,$8C,$FF,$C1,$CF,$FA
	.dc.b	$96,$3F,$DE,$6F,$FA,$4F,$C8,$EB,$E7,$F8,$1B,$19,$3A,$46,$5F,$F5
	.dc.b	$4D,$3E,$87,$D2,$FF,$BD,$5B,$9F,$FB,$21,$76,$7D,$57,$35,$F2,$EB
	.dc.b	$96,$DF,$26,$79,$C6,$4C,$CF,$FA,$2A,$33,$1C,$5E,$6F,$D5,$FA,$1F
	.dc.b	$82,$F8,$EB,$F0,$91,$8C,$3F,$CC,$B9,$FD,$BC,$3E,$45,$FF,$30,$F9
	.dc.b	$A7,$C2,$FD,$08,$93,$A3,$FD,$7D,$3A,$71,$1F,$38,$79,$29,$FE,$55
	.dc.b	$19,$87,$03,$E3,$3C,$1E,$3F,$E0,$1B,$87,$FB,$1D,$B7,$C6,$B8,$4F
	.dc.b	$D2,$E4,$7C,$67,$E3,$4F,$C8,$7D,$2D,$9D,$47,$E5,$67,$71,$FE,$B4
	.dc.b	$6D,$FE,$AE,$74,$7E,$AF,$2F,$F5,$3D,$72,$3F,$02,$E2,$7D,$99,$FF
	.dc.b	$B0,$27,$FD,$BD,$F2,$DF,$2F,$E0,$FF,$36,$D8,$C9,$F2,$A3,$2F,$ED
	.dc.b	$BE,$8B,$E6,$5F,$CB,$E1,$46,$1F,$EF,$EF,$CC,$6F,$07,$F7,$C7,$6D
	.dc.b	$8E,$15,$FB,$99,$E7,$C1,$9A,$FE,$11,$F4,$3F,$85,$AF,$CA,$37,$0F
	.dc.b	$E0,$1A,$3E,$0F,$7D,$52,$FF,$DB,$B7,$F9,$0B,$FC,$C3,$5E,$73,$FC
	.dc.b	$16,$3A,$41,$7D,$0F,$38,$FC,$51,$97,$FC,$2F,$F3,$9F,$8C,$E5,$B9
	.dc.b	$7E,$0A,$BF,$21,$3F,$E8,$BF,$53,$35,$32,$F8,$07,$F1,$99,$E4,$E2
	.dc.b	$91,$C6,$30,$EA,$3E,$EF,$C3,$7E,$94,$FD,$36,$96,$BF,$47,$16,$7F
	.dc.b	$07,$37,$7F,$F6,$9F,$1B,$E2,$7F,$C4,$4D,$C3,$FD,$FE,$DB,$FF,$77
	.dc.b	$B5,$FF,$77,$39,$AF,$F6,$EF,$47,$F8,$EE,$BF,$67,$7F,$CA,$3D,$8F
	.dc.b	$DD,$7B,$B2,$7F,$6F,$EF,$DC,$BF,$2F,$3E,$22,$4F,$89,$F2,$F8,$52
	.dc.b	$3C,$AD,$9F,$2B,$FC,$DB,$83,$F8,$43,$67,$F8,$2F,$4F,$B8,$8F,$B9
	.dc.b	$B5,$FA,$8A,$FE,$21,$1F,$C4,$5F,$E2,$58,$CA,$FE,$1F,$16,$BF,$5F
	.dc.b	$3F,$28,$93,$F2,$9F,$EA,$5F,$98,$FE,$AA,$7F,$95,$35,$FE,$60,$D7
	.dc.b	$FA,$E3,$53,$FA,$87,$FB,$07,$CC,$7F,$27,$7F,$2B,$19,$51,$9D,$97
	.dc.b	$FC,$F7,$E8,$7F,$CC,$3F,$19,$FE,$A2,$C7,$1E,$9A,$31,$8F,$6F,$37
	.dc.b	$8B,$FE,$B0,$AD,$CE,$A1,$7D,$1D,$1F,$DD,$5C,$7C,$07,$F9,$FE,$3A
	.dc.b	$63,$C6,$73,$7D,$27,$89,$D3,$1F,$23,$8C,$D9,$79,$31,$E9,$3B,$CB
	.dc.b	$A5,$F0,$9C,$E3,$DA,$BC,$27,$67,$0F,$13,$13,$C7,$1F,$EC,$98,$F8
	.dc.b	$32,$EB,$8C,$BD,$BF,$0E,$75,$18,$26,$3E,$2C,$3A,$47,$8A,$31,$5C
	.dc.b	$DB,$1E,$FC,$BD,$D8,$FC,$59,$C9,$1F,$2C,$6A,$7E,$85,$F6,$7D,$95
	.dc.b	$B7,$C1,$5C,$C8,$F0,$BF,$55,$F3,$8E,$33,$A6,$9C,$8E,$07,$BC,$FF
	.dc.b	$03,$65,$F8,$36,$E1,$F6,$9B,$7F,$E0,$B7,$F0,$DE,$AF,$F9,$19,$C5
	.dc.b	$23,$11,$D1,$AF,$E6,$DA,$F2,$F0,$FF,$19,$1C,$F2,$FC,$0B,$71,$FD
	.dc.b	$17,$CA,$FF,$BB,$DE,$55,$C0,$97,$F8,$59,$D3,$D4,$FC,$45,$73,$71
	.dc.b	$7F,$8E,$9D,$C7,$E9,$27,$4D,$AE,$B5,$B8,$B7,$E2,$7A,$CF,$E9,$6F
	.dc.b	$9E,$1F,$0E,$37,$3E,$0A,$F5,$B8,$3B,$1C,$8F,$51,$99,$E4,$E6,$FC
	.dc.b	$4D,$9F,$14,$7C,$8A,$F1,$4F,$1C,$3D,$77,$CA,$B8,$9B,$9E,$47,$94
	.dc.b	$FE,$05,$87,$AA,$B8,$92,$FA,$CF,$F0,$0F,$A1,$72,$7F,$94,$F8,$3F
	.dc.b	$88,$CA,$5C,$B9,$E1,$FF,$2C,$46,$27,$FC,$C6,$F2,$6D,$FE,$3D,$FF
	.dc.b	$44,$46,$5F,$C9,$35,$FF,$2F,$35,$FF,$07,$F4,$3E,$0B,$97,$77,$F9
	.dc.b	$2D,$B1,$99,$FE,$D3,$18,$3C,$CF,$F9,$4A,$34,$FF,$8D,$A3,$33,$FE
	.dc.b	$55,$CE,$FC,$BF,$F9,$2A,$BF,$CB,$B1,$AB,$FD,$1E,$9F,$CD,$2D,$7B
	.dc.b	$BF,$F2,$E3,$5F,$F3,$E4,$6B,$C7,$F4,$CE,$67,$FC,$56,$D7,$FC,$CE
	.dc.b	$BE,$5F,$CB,$DB,$E8,$C3,$FB,$FE,$7E,$69,$97,$FB,$BC,$7C,$71,$D8
	.dc.b	$91,$9C,$1F,$43,$E2,$24,$65,$AF,$A9,$F8,$67,$C4,$F7,$BE,$2F,$97
	.dc.b	$E4,$DA,$57,$F8,$A7,$1F,$87,$3A,$8F,$03,$F0,$08,$F8,$FC,$EF,$9E
	.dc.b	$47,$96,$3F,$EA,$66,$C6,$5F,$9A,$FD,$6F,$43,$EE,$F8,$BF,$ED,$E8
	.dc.b	$B8,$FF,$B3,$AB,$FC,$72,$D7,$FC,$8A,$3D,$7D,$CC,$8D,$BD,$1D,$1F
	.dc.b	$EB,$EF,$51,$E1,$FE,$32,$1F,$3F,$F8,$1E,$2F,$D3,$B7,$A3,$FD,$79
	.dc.b	$7C,$2B,$95,$FE,$8D,$FF,$7B,$B6,$35,$1F,$CD,$5A,$FF,$B0,$5C,$DF
	.dc.b	$ED,$1B,$5D,$AB,$C9,$0F,$43,$F4,$34,$F9,$D4,$E1,$B8,$7F,$98,$67
	.dc.b	$53,$9E,$CD,$32,$75,$5C,$1F,$F4,$A3,$FD,$1D,$FB,$3F,$1F,$FA,$CF
	.dc.b	$82,$F0,$47,$E1,$5A,$FE,$9A,$3F,$5E,$6E,$64,$72,$3F,$DD,$BF,$39
	.dc.b	$FE,$84,$7E,$97,$C7,$7B,$8F,$A5,$19,$8F,$F4,$3B,$5F,$F7,$43,$5F
	.dc.b	$F7,$73,$50,$70,$DF,$43,$A9,$1F,$F7,$FF,$CE,$BE,$7B,$F8,$3E,$7E
	.dc.b	$FE,$F9,$5F,$9A,$8C,$BF,$06,$32,$FF,$06,$8F,$E1,$FA,$C8,$FD,$0B
	.dc.b	$FC,$4B,$E2,$F1,$1F,$C3,$22,$EF,$FC,$47,$A5,$5F,$93,$C1,$EC,$DF
	.dc.b	$52,$63,$FE,$9D,$3F,$DC,$B1,$B3,$F2,$53,$C2,$3F,$9B,$35,$FF,$03
	.dc.b	$35,$F8,$E7,$C8,$75,$87,$F4,$66,$F2,$E9,$B3,$BF,$96,$FE,$5B,$F0
	.dc.b	$DF,$85,$F2,$64,$FF,$00,$8C,$BF,$87,$8C,$BF,$CB,$23,$F1,$9E,$1C
	.dc.b	$8F,$C6,$3F,$CE,$BE,$5D,$F0,$F9,$96,$7F,$CD,$FA,$7F,$96,$47,$03
	.dc.b	$F8,$C5,$70,$7E,$37,$CD,$FC,$C7,$E4,$7C,$E3,$E7,$7F,$80,$36,$37
	.dc.b	$C5,$FC,$0E,$B8,$47,$F8,$B5,$AF,$C8,$0E,$1F,$96,$D7,$F2,$6F,$2C
	.dc.b	$7F,$B8,$B7,$83,$75,$FB,$CC,$3F,$CE,$35,$FD,$4F,$17,$C5,$7F,$66
	.dc.b	$6E,$3F,$D1,$5A,$F8,$FC,$74,$62,$3F,$AB,$BE,$D7,$4D,$73,$7D,$4E
	.dc.b	$DF,$2F,$23,$B3,$5E,$3E,$27,$8C,$8F,$8D,$F3,$76,$48,$F3,$35,$F0
	.dc.b	$FC,$E7,$AC,$8F,$AD,$F7,$7C,$4C,$8F,$73,$5F,$4F,$C0,$FC,$32,$3F
	.dc.b	$0D,$F8,$DE,$D4,$8F,$8C,$D7,$E0,$FB,$1F,$2C,$8F,$CB,$7F,$C2,$74
	.dc.b	$F0,$E9,$27,$FB,$A4,$65,$BF,$07,$C0,$8E,$0D,$F8,$FD,$08,$E0,$E7
	.dc.b	$F0,$3E,$02,$38,$3B,$FC,$0F,$90,$8E,$0F,$F0,$37,$CD,$BE,$6F,$FC
	.dc.b	$56,$D8,$CC,$60,$CF,$B5,$23,$8B,$5F,$C5,$4F,$F9,$05,$FD,$7F,$4F
	.dc.b	$F2,$CE,$64,$E7,$A3,$FC,$50,$D9,$D2,$F8,$2F,$F1,$9F,$CD,$3E,$3F
	.dc.b	$FC,$6B,$F2,$AF,$42,$9A,$89,$9F,$F8,$86,$B8,$36,$3F,$C8,$F1,$88
	.dc.b	$FD,$03,$72,$FF,$1D,$37,$2F,$43,$65,$9A,$3F,$49,$5C,$3A,$7C,$FD
	.dc.b	$52,$FF,$C4,$55,$FE,$08,$7E,$19,$C9,$C3,$FE,$44,$6B,$E0,$9F,$39
	.dc.b	$FE,$5E,$F9,$A7,$25,$24,$FF,$8F,$E3,$2D,$8F,$F2,$DC,$62,$3C,$38
	.dc.b	$A3,$18,$FF,$96,$5F,$F0,$34,$FF,$45,$F9,$9F,$F3,$EC,$FF,$64,$F7
	.dc.b	$1F,$31,$FD,$00,$7C,$7F,$F4,$28,$CD,$FF,$6A,$9C,$BC,$5F,$E8,$F7
	.dc.b	$D2,$3F,$CF,$51,$8E,$4F,$E6,$B0,$F9,$9E,$6F,$83,$0F,$E9,$F1,$8A
	.dc.b	$EB,$3B,$79,$63,$FD,$EA,$78,$E1,$DD,$E1,$7C,$AB,$D4,$DB,$FF,$00
	.dc.b	$D6,$E7,$9C,$75,$3D,$93,$B7,$C8,$77,$8D,$D7,$83,$0F,$5E,$3E,$0A
	.dc.b	$F8,$47,$22,$8C,$3E,$24,$71,$C7,$A4,$EC,$F1,$1E,$02,$CF,$9D,$F4
	.dc.b	$D7,$CA,$BF,$04,$70,$C3,$E3,$C7,$58,$E3,$3F,$41,$FE,$A9,$5F,$55
	.dc.b	$B1,$97,$E1,$9A,$F8,$37,$CD,$7F,$AE,$EB,$FD,$8E,$D8,$CC,$FF,$5F
	.dc.b	$3D,$CE,$1F,$F6,$17,$C2,$FF,$98,$53,$48,$FF,$99,$5F,$86,$FF,$7A
	.dc.b	$E7,$72,$62,$3F,$C7,$33,$AD,$3F,$85,$4E,$67,$A9,$F2,$A3,$2F,$CC
	.dc.b	$38,$99,$AE,$AE,$4E,$39,$FE,$52,$78,$BF,$96,$8E,$2E,$67,$8B,$98
	.dc.b	$E3,$A7,$FB,$AE,$72,$FC,$5C,$3D,$11,$89,$FF,$03,$46,$30,$F5,$36
	.dc.b	$3C,$8D,$BE,$CD,$BE,$93,$EB,$AE,$2D,$3F,$39,$C1,$FB,$CF,$87,$F7
	.dc.b	$E5,$FE,$DB,$B4,$62,$FF,$9D,$47,$FA,$81,$FF,$5F,$3A,$6C,$6E,$FF
	.dc.b	$93,$CB,$E2,$F1,$87,$7B,$E3,$18,$E9,$7C,$4D,$31,$EB,$DA,$FF,$66
	.dc.b	$4E,$9B,$3A,$69,$AF,$47,$A3,$D5,$AF,$22,$E5,$FC,$92,$93,$8D,$3F
	.dc.b	$DF,$91,$F8,$EA,$E0,$75,$2D,$7F,$2C,$F3,$DF,$33,$EE,$6C,$FF,$A7
	.dc.b	$FE,$1B,$13,$9F,$99,$93,$FE,$A3,$8B,$7C,$E7,$97,$C3,$E3,$2F,$F2
	.dc.b	$C3,$D7,$E1,$FB,$7C,$25,$C7,$FD,$58,$F8,$7E,$66,$46,$0F,$F5,$83
	.dc.b	$E5,$F1,$3C,$DF,$11,$67,$FD,$68,$FA,$7E,$93,$D5,$F3,$56,$7F,$D7
	.dc.b	$0F,$B7,$92,$47,$FA,$EF,$E0,$AE,$FF,$0F,$D8,$F8,$5C,$DF,$0D,$B5
	.dc.b	$FE,$C1,$3E,$37,$37,$C7,$69,$FE,$B2,$E1,$FD,$8F,$47,$FA,$6B,$DB
	.dc.b	$FC,$4F,$C0,$7F,$9F,$2B,$47,$F8,$35,$7D,$3E,$A7,$7F,$3B,$C3,$DC
	.dc.b	$B7,$C3,$5F,$EF,$78,$7F,$99,$A3,$18,$FF,$AA,$3D,$BE,$FE,$7F,$81
	.dc.b	$B3,$AA,$7C,$E7,$90,$F2,$FA,$16,$75,$F0,$DE,$8F,$43,$D2,$D8,$E5
	.dc.b	$EE,$FB,$3E,$4B,$DA,$D8,$ED,$F3,$9F,$00,$F8,$3E,$FD,$8F,$17,$91
	.dc.b	$F1,$3D,$BF,$8A,$D8,$FF,$7D,$7C,$E7,$C7,$E2,$FF,$08,$8F,$8F,$FF
	.dc.b	$80,$33,$AE,$6F,$DC,$4E,$7B,$2D,$6A,$38,$B7,$A7,$F8,$3E,$1D,$5F
	.dc.b	$E4,$F3,$C7,$0F,$F6,$75,$6D,$FF,$63,$B6,$3F,$D6,$75,$C6,$73,$7F
	.dc.b	$7B,$E0,$64,$7C,$FF,$A3,$B7,$FE,$1A,$E5,$56,$32,$FC,$3C,$2B,$17
	.dc.b	$F8,$5F,$82,$FC,$4F,$B1,$F8,$A6,$C7,$DB,$8B,$8F,$59,$B8,$7B,$3E
	.dc.b	$71,$E2,$FC,$65,$E4,$58,$C3,$71,$FF,$77,$4E,$9B,$3A,$69,$A9,$E1
	.dc.b	$87,$D6,$AE,$0F,$F8,$E3,$E5,$1F,$35,$FC,$A9,$B7,$F9,$3F,$85,$C4
	.dc.b	$C8,$FC,$B5,$53,$E4,$1C,$7D,$E9,$1F,$A0,$BF,$F4,$C3,$63,$33,$E1
	.dc.b	$7C,$1F,$0B,$8D,$3E,$FF,$CE,$BC,$FF,$C3,$FC,$6F,$E2,$A3,$18,$FE
	.dc.b	$0A,$32,$D3,$F6,$58,$A7,$6F,$E3,$A7,$6D,$9D,$CE,$AF,$F1,$FC,$EB
	.dc.b	$F8,$9F,$41,$F1,$7E,$22,$E1,$ED,$70,$D3,$E0,$46,$31,$FF,$3A,$E0
	.dc.b	$F9,$DF,$F3,$7D,$7F,$07,$8C,$6B,$FD,$75,$D6,$B6,$B7,$7C,$9B,$3F
	.dc.b	$DB,$1F,$AF,$19,$AE,$1B,$FE,$C3,$C8,$7C,$47,$EC,$BE,$77,$F3,$B9
	.dc.b	$FC,$69,$CB,$F2,$23,$38,$7E,$F1,$B9,$7E,$C9,$B2,$9F,$F5,$95,$F5
	.dc.b	$EB,$93,$CD,$F1,$46,$27,$FD,$2F,$F3,$0C,$3F,$BD,$5F,$F2,$98,$C7
	.dc.b	$BB,$FD,$FD,$D7,$C3,$5A,$DD,$F2,$69,$FE,$2C,$9F,$F5,$5F,$C9,$6E
	.dc.b	$F9,$68,$8F,$CF,$65,$32,$FC,$B7,$C9,$5C,$BF,$49,$3F,$A3,$F8,$DC
	.dc.b	$67,$FD,$23,$A5,$1F,$A7,$F9,$D9,$8F,$E7,$FE,$82,$E3,$FA,$D9,$B8
	.dc.b	$71,$9B,$5F,$ED,$9C,$1F,$17,$FB,$13,$F6,$5F,$1F,$67,$F5,$51,$FB
	.dc.b	$5F,$8A,$C6,$5F,$B9,$7F,$71,$F1,$52,$BF,$61,$E3,$DA,$FD,$F4,$DC
	.dc.b	$FD,$53,$6B,$C5,$F3,$3F,$C0,$4F,$E0,$5C,$36,$3F,$75,$17,$0F,$E1
	.dc.b	$13,$6F,$E2,$71,$FF,$09,$31,$7E,$D7,$0F,$F0,$3F,$89,$B7,$FC,$45
	.dc.b	$FE,$8B,$F3,$DF,$9A,$BF,$E7,$ED,$AF,$C7,$52,$3F,$85,$BF,$D4,$1D
	.dc.b	$3F,$93,$6B,$BE,$AE,$BD,$BD,$9F,$EA,$7F,$3D,$FE,$44,$F1,$7F,$CF
	.dc.b	$B9,$76,$8C,$5F,$FA,$22,$3C,$3E,$4B,$D4,$FF,$3C,$FC,$67,$F9,$6A
	.dc.b	$BF,$45,$7C,$23,$10,$8C,$5F,$5F,$81,$63,$E0,$AF,$D5,$63,$D7,$5C
	.dc.b	$47,$1A,$DB,$E3,$3B,$B8,$C7,$BE,$3D,$A7,$9E,$5E,$6C,$BC,$19,$74
	.dc.b	$AD,$57,$58,$EF,$97,$86,$3B,$19,$2C,$F0,$38,$C6,$F0,$F0,$E3,$E3
	.dc.b	$C3,$BD,$F0,$8C,$3E,$4C,$BC,$51,$88,$ED,$5C,$B1,$F1,$E5,$E9,$9F
	.dc.b	$7C,$70,$23,$E6,$C7,$C5,$8F,$82,$F9,$4F,$86,$B9,$E3,$C2,$72,$4A
	.dc.b	$FB,$B8,$77,$BF,$1C,$73,$23,$E9,$3A,$12,$79,$39,$6B,$F2,$4F,$CC
	.dc.b	$94,$F8,$A6,$32,$FC,$7B,$7E,$0B,$FC,$54,$FF,$B7,$61,$18,$8D,$79
	.dc.b	$1F,$EE,$17,$F1,$D1,$97,$F3,$4F,$2A,$FC,$2B,$6F,$35,$F8,$3F,$97
	.dc.b	$FD,$6D,$6A,$3F,$0D,$5B,$BF,$CE,$B5,$FC,$E5,$70,$BE,$71,$F8,$69
	.dc.b	$D6,$1E,$8B,$ED,$3D,$18,$ED,$CB,$49,$87,$8D,$FC,$8E,$1D,$DF,$1C
	.dc.b	$F2,$9D,$1F,$A2,$9D,$4F,$C4,$9D,$61,$E6,$AD,$4E,$0E,$67,$82,$3F
	.dc.b	$81,$57,$2B,$F5,$4E,$4D,$D7,$27,$07,$72,$C2,$79,$61,$F1,$09,$87
	.dc.b	$C0,$BF,$4E,$1E,$52,$61,$FC,$32,$FD,$7F,$3C,$E8,$FE,$47,$C7,$FE
	.dc.b	$64,$8C,$57,$F8,$81,$AF,$F9,$D1,$75,$70,$DB,$F9,$5C,$6A,$B9,$5E
	.dc.b	$6F,$FB,$E4,$E6,$B7,$79,$F0,$5D,$9F,$93,$C7,$FB,$27,$87,$98,$37
	.dc.b	$69,$97,$F9,$0D,$AF,$F8,$07,$1F,$F0,$56,$0F,$99,$FF,$18,$4F,$FA
	.dc.b	$A1,$AE,$C4,$6F,$CB,$F2,$4E,$B2,$FF,$5E,$AF,$9C,$CC,$CB,$E8,$F0
	.dc.b	$8E,$3F,$35,$27,$C0,$F4,$E0,$C4,$78,$71,$F3,$4E,$31,$EF,$5E,$2F
	.dc.b	$8C,$F1,$FC,$7C,$C3,$D5,$19,$69,$0E,$99,$FF,$BF,$9F,$81,$E8,$FB
	.dc.b	$08,$FC,$3F,$8E,$AF,$08,$B8,$FC,$09,$F8,$7E,$AF,$C4,$F5,$65,$FC
	.dc.b	$B8,$CB,$48,$78,$BE,$3B,$F0,$5E,$77,$F8,$F7,$E0,$AF,$27,$FA,$52
	.dc.b	$BF,$C6,$CD,$7F,$C7,$62,$38,$B4,$FF,$1E,$14,$DB,$AF,$82,$FC,$84
	.dc.b	$3D,$B7,$93,$86,$C7,$F9,$F1,$AF,$FB,$41,$7A,$FB,$99,$3B,$79,$44
	.dc.b	$8C,$10,$E4,$FE,$87,$6F,$C4,$78,$17,$9F,$C9,$5F,$E7,$76,$BF,$E7
	.dc.b	$C7,$F9,$AB,$98,$E0,$47,$87,$C6,$FE,$5B,$E3,$A4,$7E,$75,$FC,$CF
	.dc.b	$C7,$48,$C1,$0F,$C2,$BF,$1B,$E7,$1F,$0D,$49,$3F,$EA,$0F,$9C,$6F
	.dc.b	$88,$FF,$7A,$D7,$07,$FD,$4A,$D7,$FD,$F3,$3C,$1E,$53,$C7,$4F,$F7
	.dc.b	$8D,$6D,$FF,$79,$D6,$EB,$F3,$3B,$FE,$C8,$8F,$EE,$73,$AE,$8F,$DE
	.dc.b	$47,$EE,$23,$F7,$3F,$21,$FB,$AD,$65,$7F,$01,$8C,$B4,$87,$EA,$3C
	.dc.b	$CF,$C4,$F8,$0F,$66,$FA,$72,$BF,$D8,$73,$96,$CE,$5F,$E1,$87,$FA
	.dc.b	$FE,$79,$C4,$8E,$31,$BD,$1E,$1F,$F9,$52,$FA,$57,$EE,$BE,$43,$F8
	.dc.b	$99,$1F,$E3,$53,$FC,$58,$C3,$B3,$27,$FA,$AE,$BA,$BF,$C7,$BE,$69
	.dc.b	$F0,$DF,$CC,$5B,$1B,$AF,$E5,$31,$88,$FE,$2C,$BE,$47,$85,$49,$E0
	.dc.b	$75,$7A,$47,$EC,$BE,$6D,$B5,$25,$7F,$B5,$FE,$3F,$FC,$BB,$97,$9A
	.dc.b	$15,$C1,$FF,$78,$B5,$F9,$AD,$8F,$68,$9E,$79,$7C,$16,$D7,$FB,$36
	.dc.b	$B8,$4F,$FA,$09,$B1,$FE,$C9,$9D,$F9,$2C,$DF,$25,$F1,$FE,$4A,$5F
	.dc.b	$C0,$AD,$4F,$CB,$B9,$39,$BF,$23,$7E,$4B,$E8,$7C,$97,$D1,$F9,$29
	.dc.b	$3B,$8F,$A9,$D7,$F5,$7C,$29,$3C,$E3,$EC,$76,$C8,$EC,$FD,$BF,$5F
	.dc.b	$EE,$7C,$97,$DD,$F9,$29,$5E,$38,$FC,$07,$C8,$FE,$07,$E4,$A4,$F9
	.dc.b	$E3,$F0,$9F,$13,$23,$D4,$FE,$1F,$E4,$7F,$11,$F2,$3F,$89,$F8,$09
	.dc.b	$5E,$F8,$FC,$63,$F8,$DF,$80,$8F,$C3,$3F,$C0,$18,$BE,$0A,$92,$3E
	.dc.b	$64,$3E,$0B,$7F,$1F,$C3,$E3,$A5,$6A,$39,$F8,$BD,$3C,$F9,$3C,$A3
	.dc.b	$BF,$C6,$48,$EA,$F8,$BE,$37,$C7,$F1,$96,$7C,$31,$87,$EF,$7C,$DF
	.dc.b	$FB,$B1,$BF,$07,$FE,$19,$8F,$F1,$1C,$6B,$E3,$70,$47,$07,$2F,$8D
	.dc.b	$E0,$47,$07,$6F,$37,$FC,$49,$F3,$6F,$06,$92,$7F,$21,$1F,$8C,$7F
	.dc.b	$C1,$4E,$DC,$C6,$DC,$F8,$7D,$0C,$79,$3F,$C1,$3C,$65,$F1,$AD,$9D
	.dc.b	$BC,$2F,$19,$1F,$0F,$87,$5F,$0C,$5B,$E9,$4A,$FF,$1F,$7C,$DF,$F6
	.dc.b	$CB,$FD,$42,$E6,$C9,$AE,$59,$3A,$6B,$FE,$27,$97,$37,$F7,$EF,$9B
	.dc.b	$70,$B3,$5B,$E6,$9E,$0F,$FC,$00,$1D,$DA,$9D,$46,$63,$13,$C2,$79
	.dc.b	$E1,$DA,$FA,$CE,$65,$39,$79,$9E,$2B,$EE,$F0,$7B,$34,$EB,$87,$A5
	.dc.b	$E2,$74,$71,$5D,$E7,$9C,$E9,$B1,$E3,$AF,$13,$6B,$8C,$F8,$2B,$86
	.dc.b	$1E,$E8,$F8,$57,$C1,$B5,$C2,$BD,$55,$B6,$9E,$93,$CD,$85,$BE,$51
	.dc.b	$A3,$B9,$9A,$E8,$F0,$32,$4C,$3D,$15,$B3,$FD,$A9,$5F,$36,$7F,$DC
	.dc.b	$2D,$7C,$F1,$AF,$8B,$7C,$DD,$24,$FF,$BA,$0F,$AF,$1A,$7F,$DC,$03
	.dc.b	$E3,$AB,$4C,$B8,$23,$F5,$A9,$CF,$FD,$BB,$1F,$C2,$4F,$E4,$AB,$E3
	.dc.b	$56,$BC,$FF,$91,$39,$39,$5F,$1D,$7C,$D7,$C2,$31,$3F,$F0,$02,$31
	.dc.b	$5C,$A3,$CD,$87,$E4,$F0,$ED,$1F,$8A,$3B,$46,$6F,$9C,$7F,$BD,$8F
	.dc.b	$0D,$73,$BE,$B1,$C4,$EB,$87,$FB,$1A,$B8,$C7,$48,$C5,$71,$8E,$67
	.dc.b	$1A,$E7,$7F,$12,$39,$9F,$E1,$5B,$F6,$F2,$FE,$DB,$E4,$BF,$BD,$E1
	.dc.b	$FE,$BA,$C7,$D7,$97,$F1,$75,$3A,$C3,$FD,$73,$1A,$AE,$13,$9B,$FE
	.dc.b	$03,$85,$AE,$35,$FD,$E5,$E1,$7F,$C2,$3E,$05,$E3,$F5,$5E,$CB,$B3
	.dc.b	$C1,$49,$DE,$1F,$E2,$27,$73,$A7,$33,$FE,$CF,$73,$3F,$E2,$88,$E5
	.dc.b	$7D,$EF,$B4,$F3,$FC,$67,$FC,$E2,$F5,$F8,$47,$CA,$79,$67,$38,$75
	.dc.b	$9C,$D7,$1F,$43,$D5,$18,$9F,$15,$70,$E0,$FF,$22,$E6,$AD,$46,$27
	.dc.b	$58,$7C,$59,$CB,$F1,$E3,$37,$F0,$E7,$2F,$FB,$B3,$E7,$3F,$CB,$B5
	.dc.b	$FE,$27,$6C,$7F,$B8,$57,$B5,$F5,$DC,$E1,$FD,$55,$B5,$FE,$FC,$FA
	.dc.b	$1F,$E9,$0F,$33,$EF,$38,$BF,$F7,$63,$6B,$FA,$F7,$0F,$F9,$66,$BF
	.dc.b	$C7,$13,$97,$FA,$3F,$17,$EE,$BD,$AF,$C4,$7D,$8F,$92,$7F,$C7,$CF
	.dc.b	$8B,$15,$23,$51,$FE,$FB,$75,$19,$9F,$C9,$7D,$AF,$86,$B5,$18,$BC
	.dc.b	$FB,$BF,$97,$F3,$FE,$4E,$0C,$3F,$CE,$F7,$FE,$24,$E3,$E7,$ED,$6F
	.dc.b	$CB,$B3,$FE,$28,$7F,$41,$B7,$F9,$8E,$3D,$FE,$6E,$3C,$AF,$8E,$F9
	.dc.b	$3D,$9F,$07,$81,$6B,$FC,$62,$FA,$3C,$BF,$27,$89,$67,$FC,$6A,$FB
	.dc.b	$3D,$8F,$47,$7D,$9F,$F1,$CB,$F0,$3D,$8F,$66,$96,$7F,$C7,$AF,$C4
	.dc.b	$E2,$F8,$1E,$15,$9F,$F2,$0B,$EA,$F6,$7E,$ED,$59,$FB,$2F,$CC,$F0
	.dc.b	$BE,$47,$C3,$D9,$FF,$24,$BF,$43,$C9,$F9,$9F,$2B,$67,$FC,$9A,$FD
	.dc.b	$4E,$BF,$A1,$F3,$16,$7F,$CA,$2F,$F0,$9F,$CF,$FF,$CA,$D1,$FE,$54
	.dc.b	$F9,$5C,$7B,$5F,$5D,$FD,$F6,$BF,$C4,$FC,$27,$CF,$3E,$07,$38,$7F
	.dc.b	$A9,$A7,$31,$8C,$FF,$C3,$13,$FE,$B7,$AE,$0F,$F0,$5F,$27,$FD,$EA
	.dc.b	$F1,$8C,$CF,$FB,$3A,$72,$F2,$3F,$C4,$51,$AB,$FF,$56,$4E,$9F,$F5
	.dc.b	$3C,$66,$BF,$43,$9F,$F8,$B1,$FF,$4E,$36,$FA,$78,$28,$FD,$44,$7E
	.dc.b	$89,$D5,$FF,$92,$DB,$5D,$3D,$1F,$95,$EA,$7C,$B7,$F9,$A7,$C4,$7F
	.dc.b	$9F,$67,$FD,$55,$CB,$FE,$38,$7F,$D5,$91,$9C,$7C,$33,$97,$C7,$19
	.dc.b	$C7,$F7,$B1,$D3,$DD,$FD,$0B,$F4,$FF,$45,$7C,$6F,$D5,$F2,$1C,$79
	.dc.b	$8C,$37,$3F,$AD,$3A,$7E,$C5,$72,$73,$3B,$BF,$A1,$8F,$FB,$B2,$30
	.dc.b	$FF,$B5,$2F,$94,$E9,$72,$F8,$CF,$27,$35,$C3,$C3,$16,$B8,$4D,$BF
	.dc.b	$9D,$EF,$FE,$57,$D9,$FC,$B3,$70,$FB,$2D,$BF,$B3,$C4,$8C,$5F,$E1
	.dc.b	$3C,$4F,$CC,$7E,$A8,$F1,$3F,$AC,$46,$1B,$8F,$E6,$67,$4F,$E5,$67
	.dc.b	$4D,$35,$3C,$31,$F8,$B5,$C1,$F1,$FB,$FF,$AD,$F8,$0F,$D7,$36,$FF
	.dc.b	$22,$D9,$DC,$7E,$52,$7F,$5D,$C6,$F8,$4A,$63,$C4,$B5,$FA,$F9,$D3
	.dc.b	$67,$4D,$3F,$5D,$DB,$FB,$9F,$C6,$5F,$31,$FD,$E6,$30,$D9,$FD,$D4
	.dc.b	$E9,$E2,$62,$35,$97,$E1,$27,$4F,$ED,$EB,$9D,$7E,$7B,$B3,$F4,$67
	.dc.b	$E9,$3C,$1B,$19,$25,$7E,$82,$32,$FD,$CF,$3F,$F8,$B1,$FC,$5F,$CC
	.dc.b	$B5,$FB,$A8,$B3,$F8,$29,$BC,$6A,$FF,$51,$7E,$19,$F9,$DA,$FE,$15
	.dc.b	$EF,$3E,$FB,$FD,$75,$5F,$E1,$7A,$D5,$F1,$BF,$F3,$7D,$69,$FE,$24
	.dc.b	$27,$75,$FE,$B4,$BE,$9F,$05,$FE,$2A,$9C,$D7,$28,$FE,$61,$E8,$7F
	.dc.b	$63,$F9,$3B,$87,$FA,$F3,$2F,$F6,$C7,$9C,$9C,$E1,$FC,$5A,$31,$5F
	.dc.b	$C6,$AF,$93,$FC,$73,$CF,$FE,$FB,$EF,$7F,$3D,$6C,$64,$FE,$FF,$19
	.dc.b	$69,$FC,$93,$D0,$FE,$89,$E7,$FF,$45,$6C,$64,$FE,$53,$19,$69,$FC
	.dc.b	$B3,$C2,$FF,$06,$F8,$FF,$D3,$5B,$19,$3F,$C1,$F1,$96,$9F,$CD,$3C
	.dc.b	$6F,$F0,$AF,$C8,$7F,$55,$6C,$64,$FF,$0B,$C6,$5A,$7F,$93,$FE,$79
	.dc.b	$F2,$0F,$8A,$FF,$4B,$63,$FE,$E3,$9C,$C6,$38,$3F,$DD,$7F,$16,$F5
	.dc.b	$BF,$01,$F0,$CC,$38,$63,$CE,$FF,$D5,$51,$8C,$39,$43,$1E,$AE,$8E
	.dc.b	$B8,$74,$AE,$11,$93,$85,$F5,$C3,$C1,$1F,$EF,$EB,$E7,$18,$86,$3E
	.dc.b	$7C,$BC,$35,$AC,$3A,$DE,$E3,$17,$E2,$AD,$E1,$DE,$F6,$49,$D1,$E5
	.dc.b	$AE,$2F,$9B,$2E,$27,$C0,$C7,$C5,$1B,$BE,$8F,$A5,$AF,$13,$E0,$47
	.dc.b	$0C,$7C,$D3,$AB,$EE,$78,$27,$58,$79,$6F,$95,$F0,$C3,$C5,$97,$CD
	.dc.b	$9F,$04,$79,$9F,$23,$AC,$B1,$5C,$67,$27,$92,$B9,$3D,$AF,$C1,$8F
	.dc.b	$8E,$72,$F9,$B2,$FC,$25,$0F,$BA,$53,$C6,$7F,$BD,$63,$F0,$AE,$4F
	.dc.b	$C4,$7C,$8F,$EF,$DD,$35,$D4,$23,$11,$F9,$B6,$BF,$87,$79,$61,$F8
	.dc.b	$16,$E1,$BE,$5F,$C3,$F8,$3E,$5A,$D4,$7E,$2A,$B7,$7C,$E3,$6E,$58
	.dc.b	$FE,$16,$36,$FE,$86,$BA,$5F,$78,$D5,$70,$BF,$CA,$56,$A7,$B1,$A2
	.dc.b	$46,$9E,$15,$CE,$3C,$F1,$F9,$59,$E4,$FC,$AA,$DB,$D2,$79,$3D,$2B
	.dc.b	$91,$1F,$D7,$57,$2C,$3E,$75,$F2,$9E,$87,$73,$11,$A7,$A1,$C8,$CE
	.dc.b	$1F,$12,$35,$87,$88,$F4,$93,$0F,$15,$7C,$93,$F8,$A7,$A7,$FE,$F4
	.dc.b	$E1,$FC,$57,$47,$F5,$38,$FF,$5A,$63,$8D,$BF,$D7,$BE,$FF,$E1,$70
	.dc.b	$FE,$95,$1F,$EC,$0E,$6F,$F2,$2C,$60,$FF,$5E,$B8,$7F,$D7,$95,$CD
	.dc.b	$FF,$60,$D7,$37,$FC,$CB,$3B,$C3,$C4,$FF,$82,$A3,$FD,$93,$F2,$3C
	.dc.b	$71,$FF,$64,$36,$75,$3C,$B7,$D6,$0A,$A6,$E3,$87,$CD,$F1,$89,$5E
	.dc.b	$9C,$18,$2B,$73,$98,$E8,$DA,$CB,$EA,$F9,$BF,$0F,$14,$8E,$11,$E7
	.dc.b	$30,$FF,$B7,$E3,$0F,$5F,$9A,$F4,$E5,$23,$04,$3F,$DC,$CE,$5F,$93
	.dc.b	$1F,$35,$7E,$48,$F2,$FE,$87,$84,$FC,$49,$E0,$E1,$B7,$F6,$4E,$35
	.dc.b	$CE,$F9,$47,$DD,$BE,$57,$C4,$D5,$7D,$3F,$9A,$95,$F1,$DF,$AD,$F3
	.dc.b	$78,$8F,$BF,$17,$8F,$EF,$CF,$DD,$F9,$0F,$9B,$5F,$EA,$86,$CF,$E5
	.dc.b	$5B,$CD,$FC,$8F,$91,$1F,$89,$8C,$3A,$A4,$EB,$8B,$B1,$DF,$9F,$E4
	.dc.b	$7C,$BB,$3F,$CB,$64,$F3,$9F,$39,$86,$E5,$F1,$EF,$83,$6F,$55,$23
	.dc.b	$04,$3F,$CB,$8E,$5F,$F1,$FC,$F2,$7F,$D4,$7F,$AB,$3C,$7F,$1B,$63
	.dc.b	$FD,$C4,$D7,$7F,$0D,$A9,$91,$A7,$8C,$7E,$9F,$E7,$7F,$51,$F2,$12
	.dc.b	$BF,$39,$19,$7F,$3D,$19,$7F,$6B,$1F,$1B,$AF,$F1,$DF,$02,$F8,$6A
	.dc.b	$66,$BE,$24,$E1,$FC,$64,$ED,$C9,$0F,$F2,$FB,$96,$CF,$37,$F8,$2F
	.dc.b	$77,$EF,$FE,$76,$47,$EC,$E3,$F8,$11,$87,$F8,$5C,$61,$FE,$29,$C2
	.dc.b	$FD,$0F,$FB,$5B,$FE,$71,$31,$E4,$FE,$02,$33,$D1,$FE,$67,$5F,$13
	.dc.b	$99,$97,$D2,$76,$AF,$B5,$73,$F2,$FA,$7C,$54,$AE,$11,$97,$94,$65
	.dc.b	$FF,$71,$DF,$F8,$C9,$D9,$A2,$1A,$9F,$D4,$7D,$07,$EA,$7C,$49,$3F
	.dc.b	$C6,$FB,$53,$A8,$C4,$E9,$FD,$74,$7F,$2B,$F2,$24,$7F,$1F,$7F,$97
	.dc.b	$79,$D8,$8F,$E6,$F1,$74,$FE,$6F,$3F,$AA,$EE,$FC,$2F,$CB,$4B,$FD
	.dc.b	$34,$65,$FC,$1C,$65,$FD,$0C,$65,$FC,$1C,$66,$BF,$D4,$D0,$EB,$FD
	.dc.b	$E7,$81,$FC,$DB,$E7,$BF,$9F,$E7,$2B,$F9,$B4,$7F,$45,$E3,$FE,$87
	.dc.b	$E5,$49,$FE,$75,$1F,$D2,$FE,$CD,$F1,$3F,$D2,$3E,$A6,$57,$F4,$39
	.dc.b	$CB,$7E,$4B,$FA,$6F,$6C,$8C,$1F,$3F,$AB,$FA,$B7,$CB,$7F,$13,$8F
	.dc.b	$E2,$2D,$DB,$F8,$BC,$F8,$7E,$23,$FA,$BF,$8B,$2F,$C1,$1E,$5F,$88
	.dc.b	$FE,$B7,$E8,$49,$F2,$47,$A7,$D7,$FE,$BF,$E6,$49,$F4,$47,$B7,$CD
	.dc.b	$FE,$C7,$E5,$C9,$F6,$47,$C1,$F3,$BF,$B3,$F9,$32,$7E,$04,$7F,$33
	.dc.b	$F9,$7F,$5F,$0C,$9F,$C6,$47,$C9,$F1,$3F,$B7,$F9,$52,$7E,$44,$7C
	.dc.b	$DF,$1B,$FB,$9F,$14,$9F,$99,$1F,$47,$C7,$FA,$5E,$34,$9F,$A1,$1F
	.dc.b	$DF,$FE,$5D,$F0,$FB,$9C,$9F,$DD,$E3,$2F,$F7,$9E,$86,$7F,$E0,$7F
	.dc.b	$9C,$75,$53,$EA,$C6,$7E,$76,$DF,$18,$C3,$FE,$03,$A4,$E2,$3F,$7D
	.dc.b	$DF,$FD,$46,$73,$1B,$F9,$6F,$F0,$F7,$CB,$3E,$6B,$FC,$E5,$E3,$FF
	.dc.b	$BF,$E1,$CE,$F4,$B6,$DC,$3F,$C5,$53,$C5,$E7,$7C,$B2,$FF,$41,$4F
	.dc.b	$2B,$E2,$6A,$A1,$CA,$FA,$3E,$81,$3C,$5E,$F7,$D5,$FF,$47,$7D,$EB
	.dc.b	$E0,$BD,$7E,$95,$9F,$2B,$5E,$82,$71,$9F,$EE,$1B,$9F,$A2,$AD,$78
	.dc.b	$9A,$F1,$BF,$F1,$F7,$CF,$3E,$42,$FC,$94,$AF,$F1,$B4,$E5,$B3,$97
	.dc.b	$DD,$E3,$A5,$5D,$2B,$FC,$4D,$EA,$7D,$1D,$3F,$D5,$9F,$3C,$F4,$7F
	.dc.b	$D5,$87,$F4,$0C,$B2,$E3,$B7,$FD,$61,$3B,$9F,$F5,$A5,$F3,$51,$97
	.dc.b	$FC,$2B,$C7,$FD,$15,$CD,$F3,$7A,$9F,$EB,$D7,$C0,$E8,$EB,$1C,$67
	.dc.b	$A5,$77,$7B,$57,$06,$DF,$9A,$3F,$D8,$D3,$B6,$D7,$37,$C5,$7D,$CF
	.dc.b	$85,$5C,$5F,$0E,$1E,$67,$D6,$D9,$F0,$1C,$B0,$F2,$DF,$5A,$E2,$75
	.dc.b	$9D,$BE,$FB,$EC,$F4,$3C,$58,$79,$DD,$1D,$E3,$71,$98,$E0,$7F,$B7
	.dc.b	$DF,$A4,$DA,$FA,$0D,$C3,$FD,$C8,$FD,$8F,$88,$7B,$D8,$8F,$F2,$5C
	.dc.b	$61,$DD,$A3,$11,$DD,$AF,$4F,$84,$E4,$FF,$BD,$DA,$9A,$47,$E8,$5B
	.dc.b	$E6,$1C,$7F,$3E,$79,$1A,$8C,$BF,$5C,$CE,$BF,$8F,$AE,$55,$C6,$7C
	.dc.b	$07,$17,$16,$87,$C4,$71,$C3,$AB,$F9,$98,$F0,$B9,$AE,$0E,$63,$AE
	.dc.b	$19,$7C,$86,$36,$FF,$66,$DF,$A2,$78,$61,$F8,$BC,$7C,$F7,$C2,$31
	.dc.b	$5B,$BF,$99,$3E,$93,$45,$BE,$6F,$AA,$3B,$E3,$E6,$78,$E3,$E5,$3F
	.dc.b	$71,$C5,$FB,$FF,$89,$FF,$21,$57,$FA,$2E,$91,$8B,$FF,$44,$C7,$F9
	.dc.b	$CE,$7F,$D8,$CE,$9B,$3F,$E1,$69,$E4,$D7,$8B,$6B,$C1,$3F,$8F,$E8
	.dc.b	$78,$BB,$BF,$E6,$AD,$8C,$CE,$E3,$07,$96,$3F,$DA,$FE,$FF,$9E,$B8
	.dc.b	$D7,$F3,$68,$D7,$53,$E7,$A9,$8F,$3E,$D7,$F4,$28,$F4,$61,$FD,$65
	.dc.b	$B5,$C6,$7F,$53,$39,$6C,$67,$0F,$E9,$D3,$FD,$7B,$E3,$FA,$C6,$3B
	.dc.b	$91,$F3,$74,$78,$9F,$A0,$C7,$FB,$D1,$95,$E5,$58,$87,$EE,$A7,$8B
	.dc.b	$F3,$67,$96,$1F,$9D,$3F,$D8,$CE,$1C,$4B,$C6,$7C,$8F,$D5,$6D,$F5
	.dc.b	$6D,$73,$3F,$D9,$6E,$1E,$7F,$13,$E6,$9F,$23,$63,$85,$A3,$11,$C3
	.dc.b	$1F,$C2,$47,$96,$3F,$DB,$8D,$7F,$DB,$2D,$9E,$55,$F8,$BF,$9D,$FF
	.dc.b	$2C,$BF,$0D,$B5,$E0,$6B,$F0,$1A,$FF,$BA,$5F,$F1,$2C,$71,$F2,$7A
	.dc.b	$11,$F9,$DD,$0F,$1B,$F3,$4D,$8C,$C6,$0F,$27,$AB,$23,$C8,$DC,$3C
	.dc.b	$9C,$9E,$92,$4F,$A5,$F6,$7A,$72,$3D,$8D,$9F,$67,$A5,$F0,$49,$3F
	.dc.b	$05,$F8,$9E,$94,$8F,$88,$D9,$FA,$79,$7F,$B6,$67,$DB,$F0,$7E,$C2
	.dc.b	$3F,$2F,$AB,$E7,$11,$C1,$F3,$FC,$5F,$A4,$47,$07,$D3,$BF,$AC,$47
	.dc.b	$07,$E6,$7E,$5D,$F2,$7F,$C1,$5B,$19,$8C,$1F,$C0,$F0,$FD,$F7,$57
	.dc.b	$A6,$FF,$C7,$9E,$1B,$E7,$3E,$DC,$7F,$8D,$27,$53,$FE,$AA,$78,$47
	.dc.b	$E6,$DA,$FE,$9D,$AF,$F9,$3A,$BC,$18,$7F,$BA,$AD,$F2,$BE,$29,$FE
	.dc.b	$4D,$87,$86,$3F,$79,$93,$E7,$29,$8E,$BB,$8F,$F5,$1B,$FE,$B7,$D0
	.dc.b	$BE,$13,$93,$FA,$B4,$66,$F9,$46,$73,$4E,$63,$0B,$E2,$7F,$D1,$31
	.dc.b	$8E,$1F,$EC,$0E,$A3,$35,$FE,$48,$9C,$BF,$E4,$A8,$CE,$3E,$69,$CB
	.dc.b	$E8,$8C,$D7,$9A,$72,$FB,$7E,$71,$F1,$BF,$19,$B1,$E9,$BF,$F6,$8F
	.dc.b	$C9,$E6,$A4,$FF,$85,$A5,$5C,$31,$FF,$45,$B5,$FE,$3D,$1A,$AF,$EA
	.dc.b	$0D,$AE,$DF,$07,$FD,$F9,$F0,$5F,$39,$FF,$05,$9F,$EE,$2F,$27,$ED
	.dc.b	$D3,$E7,$29,$8F,$93,$B8,$74,$8B,$97,$E4,$A9,$58,$C3,$EE,$F0,$7F
	.dc.b	$9C,$B0,$FA,$AB,$E7,$3F,$0F,$EA,$7E,$21,$B3,$A6,$E3,$FE,$7C,$AE
	.dc.b	$13,$F8,$18,$FF,$4D,$F7,$FE,$90,$FC,$9F,$CE,$7F,$53,$C7,$F4,$11
	.dc.b	$9A,$FE,$E0,$FF,$6D,$23,$83,$EA,$4B,$C0,$E1,$7F,$AE,$8C,$4F,$AB
	.dc.b	$D5,$FC,$FF,$CE,$7E,$81,$B1,$93,$DF,$19,$69,$F8,$1F,$79,$F3,$9F
	.dc.b	$AA,$6C,$67,$C3,$C4,$C8,$E0,$EE,$31,$8F,$EF,$FB,$3F,$82,$9C,$FE
	.dc.b	$66,$DF,$C7,$F0,$3F,$72,$FF,$84,$67,$FC,$1C,$47,$07,$E5,$7E,$55
	.dc.b	$F2,$7C,$B0,$FE,$1F,$18,$9F,$D5,$78,$1F,$DE,$FB,$BF,$B3,$F1,$FE
	.dc.b	$97,$C3,$C4,$CA,$D9,$FB,$18,$C6,$1F,$77,$CF,$FE,$12,$7F,$0A,$F9
	.dc.b	$CB,$5F,$88,$8B,$3F,$83,$9B,$87,$F9,$92,$1F,$37,$FA,$57,$FC,$04
	.dc.b	$D7,$FC,$C0,$D7,$7E,$07,$F8,$C6,$3F,$49,$87,$F9,$97,$DC,$7C,$E3
	.dc.b	$8B,$FC,$5D,$5A,$C3,$FD,$13,$39,$BF,$CE,$FB,$9F,$C8,$8F,$E4,$7F
	.dc.b	$3B,$63,$27,$F4,$18,$FE,$80,$47,$FA,$27,$B3,$FD,$0C,$8E,$09,$7F
	.dc.b	$ED,$EF,$4D,$F0,$5F,$CC,$2B,$FC,$59,$18,$AF,$D7,$FB,$7F,$D3,$48
	.dc.b	$FF,$51,$6D,$FE,$B7,$E0,$BF,$9C,$C7,$F3,$A7,$F9,$DF,$8B,$2B,$F7
	.dc.b	$31,$6B,$F6,$93,$FD,$64,$93,$FD,$75,$FE,$B9,$E4,$48,$FE,$C0,$D9
	.dc.b	$FF,$12,$F8,$9F,$C5,$5F,$67,$CB,$7F,$4A,$9F,$5B,$FE,$5D,$F9,$D7
	.dc.b	$CE,$BE,$2B,$FD,$5D,$97,$F8,$FE,$73,$18,$C3,$FD,$3D,$F1,$6F,$63
	.dc.b	$FD,$53,$F0,$9F,$C5,$A7,$FD,$45,$7C,$1F,$F5,$B6,$1C,$F1,$ED,$18
	.dc.b	$36,$E2,$B9,$43,$0E,$DA,$F8,$61,$87,$5A,$DB,$70,$ED,$96,$F0,$E3
	.dc.b	$7D,$CE,$38,$F7,$AD,$C6,$6F,$93,$EE,$CB,$DB,$3A,$C7,$BD,$F1,$AF
	.dc.b	$4E,$3D,$EF,$71,$83,$A4,$ED,$F0,$CE,$F2,$F6,$63,$E0,$9C,$9C,$27
	.dc.b	$4D,$26,$1E,$28,$CB,$63,$38,$F9,$CF,$F3,$C6,$5F,$0A,$74,$48,$D4
	.dc.b	$FB,$F0,$F1,$E3,$E7,$C7,$C8,$79,$70,$E5,$7C,$2F,$CF,$87,$37,$91
	.dc.b	$E2,$3E,$11,$FE,$7F,$AF,$C2,$36,$FF,$06,$DC,$7F,$D0,$8F,$E2,$BE
	.dc.b	$65,$E9,$62,$3F,$03,$18,$77,$68,$C4,$77,$6B,$F9,$B7,$86,$1F,$88
	.dc.b	$6E,$39,$BF,$A7,$E1,$FF,$A7,$EB,$51,$F5,$AB,$77,$E5,$8D,$B9,$63
	.dc.b	$F8,$B8,$DB,$F9,$8A,$E9,$7D,$E3,$55,$C2,$FF,$37,$5A,$9E,$C6,$89
	.dc.b	$1A,$78,$57,$38,$F3,$C7,$E6,$E7,$93,$F2,$2B,$6F,$49,$E4,$F4,$AE
	.dc.b	$44,$7F,$77,$5C,$B0,$F9,$D7,$CA,$7A,$1D,$CC,$46,$9E,$87,$23,$38
	.dc.b	$7C,$48,$D6,$1E,$23,$D2,$4C,$3C,$55,$F2,$4F,$E2,$5E,$A7,$F3,$6E
	.dc.b	$DF,$F1,$BE,$DF,$EB,$EC,$6D,$7F,$AF,$9B,$5B,$8F,$F2,$DF,$3F,$F9
	.dc.b	$29,$FF,$70,$35,$FF,$66,$B5,$FF,$6E,$B5,$FF,$34,$BF,$DE,$9E,$3A
	.dc.b	$7E,$56,$7F,$A4,$B6,$75,$3D,$23,$C7,$18,$EE,$FF,$3B,$B7,$0E,$BF
	.dc.b	$33,$AF,$96,$DB,$30,$F4,$C6,$5F,$4B,$DC,$8F,$B3,$E4,$3E,$0F,$57
	.dc.b	$F3,$E8,$FF,$3D,$B5,$63,$E2,$9F,$29,$E9,$24,$73,$3E,$3F,$C9,$FB
	.dc.b	$48,$E0,$87,$4C,$7E,$3C,$7C,$03,$E0,$FC,$A7,$D6,$97,$CE,$3C,$DF
	.dc.b	$B3,$E2,$BE,$75,$F0,$38,$6D,$FF,$95,$EF,$83,$6F,$F8,$33,$D1,$B1
	.dc.b	$A8,$FE,$1A,$D7,$EE,$FC,$AF,$D4,$23,$F8,$6F,$96,$AF,$AC,$FF,$52
	.dc.b	$F1,$27,$3B,$FD,$AA,$FC,$7F,$C7,$B8,$E9,$30,$FF,$55,$C7,$E3,$EB
	.dc.b	$FD,$6E,$EE,$33,$3B,$7C,$1C,$A9,$E1,$C2,$99,$F3,$7E,$37,$FC,$9D
	.dc.b	$C3,$FA,$2C,$5E,$8A,$3E,$85,$6E,$BD,$81,$C3,$D4,$EB,$A7,$CC,$8C
	.dc.b	$C6,$5A,$E8,$FF,$45,$3D,$6F,$FC,$94,$DC,$3F,$DA,$DF,$3B,$FE,$A3
	.dc.b	$AF,$F5,$C2,$F2,$3D,$0D,$8C,$C6,$3E,$4B,$2C,$8C,$9C,$A3,$F5,$BF
	.dc.b	$2F,$FA,$EF,$8F,$93,$FC,$02,$32,$FF,$02,$8C,$BF,$3B,$89,$3F,$1B
	.dc.b	$A7,$F2,$DE,$AB,$E1,$A9,$9B,$F7,$CF,$C5,$23,$FD,$5F,$54,$62,$3E
	.dc.b	$7C,$61,$E2,$7F,$AC,$6B,$FD,$32,$FF,$16,$C3,$F8,$3F,$CC,$48,$C1
	.dc.b	$0F,$43,$97,$D5,$1F,$81,$F9,$DF,$ED,$BC,$1F,$ED,$46,$F0,$7E,$26
	.dc.b	$11,$8A,$FE,$06,$BE,$2B,$97,$91,$27,$FD,$DD,$3B,$AE,$35,$DF,$CA
	.dc.b	$F0,$7C,$5C,$BD,$C6,$5F,$F7,$95,$FF,$8F,$5B,$1F,$DF,$6F,$C6,$D9
	.dc.b	$D4,$FE,$C6,$72,$F4,$3C,$7F,$41,$FB,$0F,$9E,$FD,$8F,$8D,$27,$F5
	.dc.b	$F1,$97,$F6,$31,$97,$F5,$BD,$69,$FE,$6B,$E3,$C9,$FE,$8E,$FF,$38
	.dc.b	$F3,$F1,$1F,$87,$8B,$8F,$F3,$A9,$FD,$6F,$7F,$E2,$3E,$5A,$5F,$E9
	.dc.b	$23,$2F,$E1,$5F,$D6,$11,$C1,$B2,$1F,$B6,$E4,$6F,$FD,$2B,$E7,$BF
	.dc.b	$A7,$12,$7F,$AD,$BF,$D4,$78,$E4,$7F,$5E,$6B,$FB,$D1,$E0,$BE,$75
	.dc.b	$F1,$7F,$D6,$DB,$3F,$E3,$49,$D4,$E4,$D7,$C9,$49,$D5,$23,$11,$DE
	.dc.b	$BF,$8D,$74,$FF,$67,$E2,$34,$93,$CB,$E2,$BC,$44,$70,$76,$F8,$AF
	.dc.b	$31,$1C,$1E,$2F,$5B,$D4,$47,$07,$9B,$CA,$F7,$11,$C1,$EA,$F3,$3E
	.dc.b	$11,$1C,$1F,$0F,$17,$CA,$7A,$C9,$19,$7E,$57,$0C,$8F,$8A,$D7,$E2
	.dc.b	$6D,$FA,$F7,$E4,$F6,$FD,$12,$3F,$33,$BB,$EA,$11,$C1,$F5,$7E,$55
	.dc.b	$F1,$3F,$E1,$36,$C6,$63,$07,$F8,$5F,$E5,$BF,$8F,$C7,$DD,$6B,$F5
	.dc.b	$DA,$FF,$43,$E2,$46,$23,$FA,$94,$3E,$71,$F3,$BF,$E2,$36,$CF,$E2
	.dc.b	$5A,$FE,$0D,$AF,$57,$3D,$AE,$8E,$B1,$86,$D7,$F8,$F2,$1F,$2F,$FE
	.dc.b	$40,$F8,$C7,$CC,$FE,$77,$0E,$58,$A5,$E6,$3C,$DE,$0D,$9F,$E4,$99
	.dc.b	$FF,$6D,$AE,$71,$D7,$0F,$F0,$6C,$62,$3F,$D2,$ED,$7F,$58,$D7,$FB
	.dc.b	$53,$9C,$3C,$6D,$AE,$D5,$E7,$6F,$07,$F8,$97,$E7,$5F,$07,$ED,$F2
	.dc.b	$EE,$5E,$63,$D8,$E1,$FE,$F7,$37,$2F,$83,$08,$D3,$FC,$05,$AF,$F7
	.dc.b	$F6,$BF,$E0,$D6,$F0,$3E,$75,$F2,$5F,$E5,$F6,$D7,$F0,$56,$BF,$BF
	.dc.b	$6B,$F8,$97,$37,$F5,$2B,$A5,$7F,$9D,$A7,$39,$FE,$31,$7C,$97,$F9
	.dc.b	$1F,$AF,$EC,$61,$FE,$21,$F8,$57,$91,$FE,$17,$3F,$55,$9A,$32,$FF
	.dc.b	$89,$E3,$31,$88,$FF,$16,$71,$FF,$98,$9E,$13,$FE,$21,$BE,$F0,$F9
	.dc.b	$DE,$13,$D9,$BD,$7F,$E2,$77,$A5,$70,$6D,$F5,$77,$7F,$E3,$96,$BD
	.dc.b	$6F,$C1,$3B,$6E,$3B,$7D,$D8,$77,$6D,$64,$FF,$22,$61,$E1,$8E,$31
	.dc.b	$EC,$3E,$14,$FF,$8F,$71,$F0,$9E,$6B,$F9,$95,$C3,$1F,$29,$1C,$12
	.dc.b	$F9,$9E,$13,$E9,$DF,$95,$F1,$E3,$EB,$6D,$FF,$BE,$F1,$F9,$B1,$FD
	.dc.b	$4D,$AF,$A0,$C7,$C4,$BD,$EE,$0D,$8F,$EB,$47,$F9,$95,$C3,$C2,$3F
	.dc.b	$D0,$6B,$E1,$29,$1F,$F4,$3B,$73,$FC,$26,$0F,$99,$FE,$13,$1F,$87
	.dc.b	$AF,$F3,$85,$7B,$EB,$5C,$1F,$06,$72,$7D,$B3,$FC,$5D,$7C,$B5,$F3
	.dc.b	$DF,$18,$C4,$FF,$73,$8C,$57,$38,$F4,$1C,$CF,$F5,$BC,$F0,$7F,$CE
	.dc.b	$D3,$C1,$D1,$E2,$C3,$D7,$3B,$8C,$5F,$F9,$F7,$0F,$4C,$70,$3B,$D7
	.dc.b	$0C,$3C,$58,$75,$8F,$85,$3A,$BE,$58,$75,$8C,$1F,$02,$F8,$BE,$9C
	.dc.b	$3B,$DF,$ED,$38,$BF,$73,$F1,$5F,$CD,$A7,$F8,$26,$28,$C6,$1F,$D3
	.dc.b	$A3,$FD,$0C,$FF,$21,$AE,$2D,$8E,$0D,$9E,$38,$7F,$87,$A7,$8B,$FE
	.dc.b	$AC,$76,$FF,$A8,$BA,$7F,$C5,$B5,$FC,$B6,$3F,$A4,$78,$5C,$A5,$E0
	.dc.b	$FF,$99,$6B,$51,$C2,$B9,$57,$81,$E3,$3F,$CD,$E7,$8F,$16,$5E,$19
	.dc.b	$FF,$44,$EB,$7C,$EF,$76,$DF,$15,$E4,$66,$64,$E5,$E1,$18,$C7,$FD
	.dc.b	$B1,$87,$F6,$49,$F5,$77,$3E,$47,$59,$7F,$54,$8C,$63,$F3,$7C,$3F
	.dc.b	$F0,$EB,$4F,$97,$C3,$FE,$02,$F1,$AC,$FF,$77,$9D,$13,$3F,$F6,$CC
	.dc.b	$72,$6C,$72,$E0,$7C,$9D,$39,$C5,$3E,$F7,$C3,$71,$CF,$FA,$43,$63
	.dc.b	$FC,$22,$BD,$0F,$2B,$9C,$3F,$CB,$AF,$95,$FF,$75,$B6,$B9,$E5,$C2
	.dc.b	$FD,$BF,$3B,$FE,$62,$7F,$1C,$DB,$FF,$79,$AF,$9F,$DF,$A5,$C3,$D2
	.dc.b	$97,$C3,$E5,$AE,$1F,$EF,$AE,$CF,$F0,$BC,$FE,$65,$B1,$D3,$CC,$F9
	.dc.b	$B1,$8D,$3F,$B3,$66,$F4,$BC,$A7,$9B,$CC,$B8,$F9,$BA,$7D,$27,$AB
	.dc.b	$E5,$2C,$7A,$BD,$3F,$69,$EE,$EA,$B1,$EE,$F4,$FE,$09,$F0,$BE,$02
	.dc.b	$C7,$C2,$F5,$3E,$29,$F1,$BE,$5E,$C7,$AE,$7D,$27,$CA,$F0,$EC,$7C
	.dc.b	$AF,$0F,$E6,$9F,$3B,$DD,$B1,$F3,$BC,$4F,$A2,$7D,$2F,$7E,$C7,$D2
	.dc.b	$E6,$FA,$A7,$D6,$F5,$2C,$7F,$03,$F9,$27,$CB,$FE,$59,$B1,$9F,$63
	.dc.b	$13,$FB,$1D,$64,$FF,$8A,$FA,$1F,$3D,$FE,$99,$8F,$DE,$CE,$B2,$FE
	.dc.b	$0C,$FF,$0C,$AD,$F6,$E5,$FF,$17,$F9,$29,$C5,$FF,$BB,$F0,$E9,$87
	.dc.b	$28,$F4,$7C,$F7,$9C,$F4,$78,$B6,$32,$49,$63,$FE,$29,$D7,$F5,$33
	.dc.b	$98,$FF,$16,$7B,$BF,$E8,$28,$FC,$75,$7E,$E2,$30,$FE,$4C,$C6,$DF
	.dc.b	$90,$6C,$F0,$3E,$27,$CF,$3E,$4B,$E0,$36,$3E,$47,$88,$F9,$5C,$D4
	.dc.b	$9E,$2F,$C7,$9D,$46,$09,$7F,$EF,$DB,$E8,$FF,$BF,$E7,$6E,$6B,$97
	.dc.b	$CF,$4B,$F0,$FC,$AD,$AE,$13,$74,$FE,$99,$82,$B1,$7F,$E9,$8D,$5F
	.dc.b	$3D,$F8,$7F,$9E,$FC,$43,$67,$F1,$8D,$CB,$E8,$C7,$18,$C6,$3F,$80
	.dc.b	$F1,$7F,$9E,$1F,$85,$F9,$EF,$E7,$37,$FD,$36,$33,$7F,$D2,$1D,$78
	.dc.b	$58,$9D,$4E,$4F,$CA,$57,$2B,$ED,$1F,$9F,$F7,$FF,$3D,$EF,$7E,$7D
	.dc.b	$B1,$93,$F4,$91,$96,$9F,$9F,$E3,$78,$5F,$9B,$6C,$67,$C4,$C4,$C8
	.dc.b	$E0,$EE,$31,$97,$F6,$9F,$33,$FA,$E1,$CF,$C7,$B7,$FD,$9A,$2C,$FC
	.dc.b	$A3,$F3,$78,$2B,$39,$FE,$AA,$AD,$AF,$75,$85,$E9,$BA,$7F,$47,$1D
	.dc.b	$63,$1D,$5F,$A1,$F0,$3F,$44,$CC,$3F,$C1,$57,$C3,$2F,$D4,$D6,$70
	.dc.b	$E5,$CB,$FD,$91,$BE,$AF,$F0,$DF,$9E,$FE,$1C,$CC,$7F,$89,$39,$69
	.dc.b	$FE,$0B,$F9,$EF,$CE,$E3,$FE,$B4,$6C,$7B,$7D,$CF,$DE,$F9,$DF,$E0
	.dc.b	$99,$CE,$3F,$05,$A9,$CC,$F0,$AE,$DD,$DF,$C8,$7D,$D9,$1F,$C9,$7E
	.dc.b	$76,$E1,$EC,$9B,$97,$F0,$B9,$B5,$FA,$5F,$92,$FE,$54,$7F,$2B,$F8
	.dc.b	$1B,$1F,$CB,$E2,$E1,$F1,$A6,$DF,$F1,$5F,$7B,$F9,$91,$FC,$CF,$D5
	.dc.b	$B8,$7F,$37,$8B,$7F,$3A,$6D,$7F,$1D,$F0,$BF,$9D,$1F,$CE,$FE,$1A
	.dc.b	$E1,$FC,$FE,$2D,$FD,$69,$B5,$FC,$97,$E0,$FF,$A1,$1F,$D0,$FC,$6B
	.dc.b	$87,$F4,$78,$B7,$F7,$A6,$D7,$FB,$A3,$E7,$1F,$3A,$F4,$3F,$D0,$39
	.dc.b	$FE,$4F,$05,$6A,$73,$1F,$97,$F4,$5E,$BB,$4C,$CF,$0C,$7A,$3F,$E9
	.dc.b	$5C,$7A,$63,$CE,$30,$7E,$85,$C7,$C2,$7F,$1B,$8E,$D8,$75,$76,$75
	.dc.b	$CB,$BD,$70,$BE,$B8,$78,$30,$E5,$6C,$BC,$F9,$70,$AD,$5F,$0C,$70
	.dc.b	$7A,$AB,$77,$E0,$C7,$B3,$E7,$BE,$2D,$AE,$19,$79,$F2,$F0,$DF,$0A
	.dc.b	$F4,$1C,$EF,$8B,$E5,$AE,$18,$F8,$E3,$58,$79,$1F,$83,$8F,$93,$1E
	.dc.b	$B8,$72,$C7,$B1,$DE,$75,$87,$3A,$DE,$1C,$DF,$8B,$3A,$C3,$D4,$78
	.dc.b	$09,$1A,$C7,$CB,$1A,$8E,$6E,$AF,$F1,$15,$CE,$3D,$F3,$A2,$4E,$AB
	.dc.b	$9C,$EA,$3E,$03,$F9,$13,$E3,$97,$0F,$C4,$37,$1F,$C6,$B9,$F8,$EF
	.dc.b	$EE,$CF,$FB,$A2,$51,$88,$D7,$9F,$FE,$EC,$7F,$3B,$19,$7F,$2A,$F2
	.dc.b	$C7,$EC,$B7,$0C,$D7,$E4,$3E,$43,$FD,$AB,$18,$BF,$C3,$56,$DF,$CE
	.dc.b	$35,$FD,$5C,$F0,$73,$5F,$9A,$9D,$61,$E8,$8E,$D3,$96,$3D,$9C,$B4
	.dc.b	$98,$7A,$23,$2D,$8C,$E1,$E7,$8F,$CA,$E1,$E3,$9F,$87,$5D,$A7,$81
	.dc.b	$A8,$C5,$71,$23,$FB,$9A,$E5,$7E,$98,$EE,$7B,$8F,$D5,$46,$B0,$F7
	.dc.b	$93,$0F,$3D,$F4,$AF,$19,$C8,$F9,$27,$88,$E9,$87,$9A,$3F,$86,$DF
	.dc.b	$CD,$F0,$9F,$15,$FE,$3D,$BF,$F7,$63,$5F,$ED,$4D,$E0,$FF,$28,$63
	.dc.b	$FC,$85,$FE,$21,$5C,$BA,$B1,$C3,$FD,$DE,$31,$87,$49,$CB,$FD,$E3
	.dc.b	$C0,$FC,$85,$F2,$C5,$96,$B0,$F2,$B6,$33,$7E,$6F,$8E,$C4,$C8,$F4
	.dc.b	$BB,$8E,$1F,$3B,$27,$DA,$F3,$F9,$1C,$47,$BA,$2E,$3E,$39,$F0,$7C
	.dc.b	$A4,$AF,$8A,$F8,$FE,$7F,$11,$F1,$A2,$D7,$AE,$7D,$1F,$3D,$2B,$E6
	.dc.b	$BE,$BF,$9A,$C4,$7C,$E8,$B5,$F0,$E7,$E3,$CB,$E7,$5E,$7E,$1B,$1F
	.dc.b	$E2,$5A,$FE,$84,$EE,$7E,$7F,$8A,$E6,$7A,$3C,$66,$57,$1E,$5F,$EB
	.dc.b	$98,$F4,$8F,$F1,$B6,$1C,$6F,$E9,$7C,$DC,$9F,$C3,$BF,$57,$E7,$71
	.dc.b	$1F,$88,$8B,$9F,$DD,$9F,$B9,$F2,$4E,$3A,$4B,$F1,$1F,$E3,$C8,$FC
	.dc.b	$86,$7F,$8B,$78,$46,$B9,$91,$88,$FF,$20,$37,$0F,$CA,$BC,$A3,$FC
	.dc.b	$95,$F3,$4C,$BE,$6B,$63,$C2,$1C,$3F,$E5,$7A,$7C,$E3,$C3,$E3,$72
	.dc.b	$B8,$DF,$0B,$FE,$E8,$70,$AD,$4F,$F9,$84,$ED,$58,$9D,$DF,$59,$FF
	.dc.b	$6A,$2F,$27,$F5,$33,$E2,$6C,$66,$77,$E4,$E2,$64,$72,$77,$1F,$AB
	.dc.b	$F9,$CF,$D6,$7C,$9C,$AF,$E0,$31,$97,$F8,$1C,$65,$FE,$07,$1F,$23
	.dc.b	$A6,$47,$C0,$7F,$2F,$E9,$FF,$A3,$23,$04,$3E,$06,$A9,$DF,$B1,$CA
	.dc.b	$73,$17,$4F,$CD,$DF,$39,$FE,$27,$DF,$FC,$0F,$E6,$A4,$7F,$18,$8F
	.dc.b	$E0,$C6,$1F,$E3,$31,$87,$FD,$FD,$F3,$9F,$D9,$2F,$EA,$B7,$93,$F5
	.dc.b	$4D,$9F,$F7,$7A,$F8,$8F,$A2,$E6,$8D,$24,$6B,$CB,$95,$E0,$78,$FC
	.dc.b	$4E,$23,$F1,$95,$FE,$31,$74,$79,$A3,$57,$E6,$AB,$87,$9A,$7F,$57
	.dc.b	$E1,$7E,$B3,$C4,$98,$7E,$97,$A1,$3A,$8C,$4E,$9F,$D3,$C7,$F2,$DF
	.dc.b	$26,$47,$F4,$27,$F9,$87,$CE,$62,$3F,$A1,$C5,$CF,$F9,$C4,$FE,$AB
	.dc.b	$B2,$57,$E8,$9F,$C1,$FC,$5E,$23,$F0,$31,$6B,$EE,$CD,$D3,$FC,$D9
	.dc.b	$CB,$FB,$96,$D7,$2B,$FE,$AB,$F2,$9F,$D0,$C8,$FF,$58,$F3,$BF,$A3
	.dc.b	$54,$8F,$EB,$6F,$F4,$9F,$93,$FE,$99,$F3,$8F,$88,$A4,$AF,$E1,$90
	.dc.b	$F9,$4D,$FC,$7C,$9C,$BC,$BE,$0F,$F8,$DF,$4C,$9F,$A0,$F6,$F9,$4E
	.dc.b	$FF,$13,$27,$AB,$E2,$F8,$0F,$1F,$C9,$C9,$F0,$BE,$6F,$86,$F3,$FB
	.dc.b	$52,$7C,$AF,$AB,$E2,$BD,$7E,$8C,$9F,$4B,$EE,$F9,$0F,$7F,$C9,$C9
	.dc.b	$F6,$BF,$0B,$C7,$FE,$67,$C9,$27,$F1,$0F,$C6,$F9,$3F,$8F,$D7,$27
	.dc.b	$E2,$BF,$2B,$C2,$F9,$7E,$2C,$9F,$92,$FC,$EF,$13,$E7,$FB,$F2,$7E
	.dc.b	$6B,$FD,$F7,$E5,$9F,$19,$49,$3F,$CF,$63,$2F,$D4,$3F,$C0,$9F,$2E
	.dc.b	$F3,$69,$23,$FA,$27,$CE,$3A,$7F,$C1,$4D,$8C,$C6,$0C,$F4,$48,$C9
	.dc.b	$07,$8B,$FE,$1E,$A4,$FF,$86,$BE,$59,$F3,$3F,$E7,$3A,$FF,$59,$43
	.dc.b	$D1,$7A,$5B,$6E,$3F,$D7,$A7,$8B,$CE,$F9,$78,$A8,$4E,$9E,$13,$A7
	.dc.b	$2F,$F1,$78,$D4,$7A,$27,$94,$63,$0F,$2D,$71,$8F,$4F,$07,$F8,$3B
	.dc.b	$E7,$1F,$11,$49,$8F,$9F,$83,$FC,$12,$FF,$66,$9D,$B6,$76,$F8,$4F
	.dc.b	$4C,$F0,$71,$EC,$BE,$71,$F2,$7F,$E3,$F6,$CF,$F7,$D6,$BC,$1E,$5A
	.dc.b	$7F,$94,$1C,$B6,$79,$1C,$FD,$DF,$F1,$4F,$A9,$F4,$A7,$E6,$7C,$E3
	.dc.b	$B7,$FD,$F4,$7F,$3D,$C7,$2E,$3A,$FF,$DF,$D3,$B9,$FA,$97,$CD,$46
	.dc.b	$5F,$F0,$9F,$27,$E2,$1C,$DF,$37,$A9,$F7,$5F,$03,$A3,$AC,$71,$9E
	.dc.b	$95,$DD,$ED,$5C,$1B,$7E,$68,$FC,$45,$F8,$B0,$EE,$D3,$B1,$F0,$67
	.dc.b	$83,$E3,$6B,$BA,$F0,$E1,$E3,$BF,$54,$F2,$7C,$47,$83,$0F,$33,$67
	.dc.b	$84,$F5,$3C,$71,$C1,$F7,$E1,$E6,$AE,$31,$B3,$B6,$1E,$48,$CC,$7A
	.dc.b	$CF,$F7,$13,$F4,$DB,$7F,$45,$B5,$FE,$DB,$97,$C7,$BE,$03,$0D,$AF
	.dc.b	$F7,$61,$FE,$4C,$70,$F6,$6B,$AF,$82,$FF,$73,$3F,$B7,$6B,$FE,$6F
	.dc.b	$32,$E0,$FA,$76,$F8,$D7,$3F,$D6,$AE,$47,$D6,$9D,$46,$63,$A6,$BF
	.dc.b	$8F,$9C,$C7,$DC,$38,$C6,$6B,$AC,$64,$E3,$AF,$FA,$E6,$33,$3F,$99
	.dc.b	$9E,$33,$9B,$F0,$61,$E8,$BD,$CF,$C1,$AF,$34,$61,$FC,$89,$D8,$FF
	.dc.b	$4A,$39,$7F,$2B,$3C,$EB,$AC,$72,$6D,$FA,$5B,$7E,$39,$F5,$57,$17
	.dc.b	$F5,$1C,$3F,$BB,$D7,$FC,$DF,$7F,$ED,$1B,$FE,$62,$FF,$A1,$DF,$C0
	.dc.b	$B6,$37,$4A,$E5,$3F,$D2,$BC,$7C,$C6,$27,$F8,$CF,$5A,$73,$C8,$F5
	.dc.b	$AF,$B7,$86,$E3,$FE,$97,$9D,$36,$74,$D3,$FD,$23,$D4,$F1,$3C,$B9
	.dc.b	$7F,$42,$7C,$F0,$F0,$7F,$CA,$95,$C2,$77,$3C,$63,$FA,$75,$FF,$A2
	.dc.b	$5D,$18,$F7,$B7,$5D,$27,$58,$7F,$A5,$6F,$93,$FC,$ED,$46,$B8,$FF
	.dc.b	$E0,$0E,$4F,$57,$E1,$9F,$13,$D9,$B9,$FC,$18,$F8,$FE,$BE,$32,$FE
	.dc.b	$87,$68,$CD,$7F,$80,$7E,$4E,$D7,$F4,$8D,$91,$F4,$3C,$AF,$A3,$54
	.dc.b	$F9,$C6,$0F,$F0,$57,$CA,$5B,$E4,$FD,$CE,$9F,$F1,$F4,$FF,$83,$FE
	.dc.b	$67,$1E,$C7,$E0,$32,$46,$FD,$AC,$7C,$C5,$BF,$F0,$B3,$CF,$D9,$E1
	.dc.b	$F3,$16,$7F,$C3,$4F,$7F,$6F,$A7,$8B,$67,$FC,$3C,$F8,$FC,$EF,$07
	.dc.b	$8F,$67,$FC,$44,$F9,$FC,$EF,$27,$CB,$D9,$FF,$13,$3E,$BF,$6F,$D1
	.dc.b	$F2,$F6,$7F,$C5,$4F,$BF,$DB,$F6,$7C,$B5,$9F,$F1,$73,$F0,$FD,$DF
	.dc.b	$81,$E5,$D9,$FF,$19,$3F,$1F,$C5,$F8,$9E,$45,$9F,$F1,$B3,$F2,$FD
	.dc.b	$DF,$91,$E3,$59,$FF,$1D,$3F,$3F,$DC,$F9,$9D,$76,$7F,$C7,$CF,$D3
	.dc.b	$F7,$7E,$87,$CB,$59,$FF,$21,$3F,$A2,$FA,$9F,$F2,$04,$7F,$BD,$9B
	.dc.b	$E0,$7F,$BD,$7E,$69,$F2,$3B,$F7,$25,$FF,$8A,$FC,$AF,$F7,$9F,$CD
	.dc.b	$FF,$91,$63,$94,$FF,$02,$94,$E7,$0F,$F7,$B6,$3D,$A7,$8F,$5F,$49
	.dc.b	$FF,$0F,$47,$08,$FC,$67,$C1,$4C,$7C,$07,$F8,$EF,$D2,$FE,$81,$F0
	.dc.b	$59,$C3,$FC,$47,$C0,$BC,$63,$FC,$E2,$F7,$87,$2B,$FE,$2D,$7B,$8C
	.dc.b	$4F,$F4,$BF,$82,$FC,$E7,$AB,$F9,$D6,$CE,$A9,$18,$BF,$0B,$70,$FE
	.dc.b	$8F,$1F,$E3,$4F,$95,$FE,$23,$E0,$9F,$8D,$E0,$FF,$01,$C7,$FC,$1F
	.dc.b	$39,$7C,$D5,$C5,$F4,$5F,$F9,$A3,$A7,$9E,$1C,$9B,$3A,$C3,$F9,$BD
	.dc.b	$FF,$92,$4E,$47,$CE,$73,$7F,$17,$27,$C1,$FE,$EC,$FD,$E7,$AB,$70
	.dc.b	$F8,$11,$70,$F9,$F3,$6F,$97,$C0,$7F,$03,$F7,$7F,$82,$37,$0F,$90
	.dc.b	$DB,$FA,$ED,$AF,$8F,$EB,$BB,$3F,$98,$9F,$8B,$E7,$79,$18,$C7,$F9
	.dc.b	$EC,$E7,$CA,$C6,$5F,$8A,$B4,$63,$1F,$CB,$B6,$FF,$35,$3A,$7F,$03
	.dc.b	$F1,$7F,$EA,$A7,$EF,$5F,$F5,$88,$C3,$F8,$E3,$1F,$09,$FC,$85,$B3
	.dc.b	$C0,$FA,$FE,$73,$E3,$FF,$48,$D8,$CF,$CB,$E2,$64,$70,$77,$18,$AF
	.dc.b	$E9,$5E,$67,$F5,$43,$F5,$DF,$37,$FD,$A6,$FF,$AF,$C6,$5F,$C3,$C3
	.dc.b	$CA,$C5,$7E,$BA,$31,$3C,$0D,$CE,$B6,$FF,$0F,$89,$DD,$7E,$DB,$37
	.dc.b	$7F,$EF,$DB,$87,$E5,$97,$8D,$99,$93,$97,$84,$62,$BF,$D3,$7B,$3D
	.dc.b	$5F,$A0,$7D,$1F,$5E,$E1,$F5,$A2,$D7,$09,$B9,$FE,$C3,$E6,$DF,$17
	.dc.b	$FE,$14,$F8,$37,$D0,$3C,$EF,$F0,$1D,$6B,$3F,$E0,$95,$A9,$CD,$F1
	.dc.b	$BE,$15,$FB,$4B,$E1,$97,$FA,$D7,$0E,$13,$F8,$AF,$5F,$F1,$07,$E2
	.dc.b	$7D,$8B,$19,$3F,$DE,$9F,$0F,$8B,$FE,$E8,$FE,$3F,$D8,$FC,$87,$B3
	.dc.b	$6B,$F2,$AD,$AF,$8E,$DD,$BF,$8F,$70,$A3,$F3,$1E,$47,$E6,$7D,$9B
	.dc.b	$87,$E7,$9B,$87,$D0,$6E,$1C,$9F,$D0,$7B,$32,$B8,$1F,$A6,$71,$5F
	.dc.b	$E0,$BE,$BB,$7F,$CB,$FD,$A4,$7E,$A7,$D8,$FD,$57,$B3,$71,$FD,$7B
	.dc.b	$70,$FC,$0B,$70,$F5,$BF,$E3,$8F,$9B,$FF,$12,$78,$FF,$E8,$5C,$7F
	.dc.b	$D6,$B1,$8B,$7C,$2F,$81,$FF,$80,$1E,$5F,$D8,$73,$5C,$3E,$05,$87
	.dc.b	$39,$0C,$3A,$61,$CE,$31,$5C,$1E,$67,$6E,$BF,$F4,$2C,$EE,$B8,$E1
	.dc.b	$DF,$0E,$D7,$C9,$B8,$F9,$6F,$86,$3D,$5B,$3C,$30,$E7,$7D,$F4,$C5
	.dc.b	$EF,$1E,$D3,$E4,$3B,$17,$1E,$D1,$AC,$7A,$CE,$F0,$EA,$FF,$42,$CB
	.dc.b	$D7,$5B,$24,$7C,$69,$DD,$F4,$8C,$61,$E3,$6E,$5E,$63,$DF,$87,$48
	.dc.b	$C9,$E1,$9E,$0F,$A3,$1F,$A1,$7D,$63,$18,$F8,$9A,$7C,$33,$96,$1D
	.dc.b	$DA,$7C,$1C,$3C,$0F,$02,$63,$EA,$8C,$BF,$08,$FC,$64,$7F,$AC,$9A
	.dc.b	$FC,$E3,$18,$FE,$0D,$D7,$C1,$3E,$6F,$FD,$C3,$1F,$8B,$8C,$3F,$EE
	.dc.b	$4A,$46,$23,$8C,$F0,$9C,$BF,$81,$9E,$39,$FE,$35,$B9,$7E,$8B,$E4
	.dc.b	$7F,$D8,$6F,$E9,$23,$38,$7D,$76,$BE,$3B,$FC,$65,$70,$8D,$47,$72
	.dc.b	$3F,$85,$8D,$3C,$4F,$C4,$4F,$4C,$3C,$F1,$B3,$8E,$1E,$7B,$B8,$79
	.dc.b	$E6,$9E,$03,$9C,$79,$1C,$9D,$E7,$93,$D4,$D9,$1C,$9F,$B1,$C3,$CF
	.dc.b	$7E,$48,$CD,$FC,$5B,$ED,$7C,$63,$0D,$3F,$3D,$7C,$B0,$F7,$12,$B8
	.dc.b	$DF,$4A,$F5,$35,$FE,$15,$5F,$37,$D6,$73,$BD,$EF,$E1,$F5,$FD,$83
	.dc.b	$D4,$BC,$7F,$CE,$9B,$C5,$FE,$5F,$BE,$15,$FE,$16,$BC,$BF,$EE,$F8
	.dc.b	$56,$30,$EE,$D8,$E3,$3F,$DB,$BD,$3D,$73,$7F,$85,$63,$FC,$C2,$D7
	.dc.b	$47,$F8,$92,$39,$11,$FF,$13,$FB,$DD,$7B,$7B,$32,$3F,$C5,$71,$97
	.dc.b	$64,$3F,$BF,$BE,$5F,$33,$CD,$C5,$27,$FC,$67,$19,$7F,$C6,$B1,$97
	.dc.b	$FC,$0F,$1E,$DE,$1F,$5F,$CE,$49,$FF,$1D,$C6,$5A,$43,$FC,$1F,$1F
	.dc.b	$29,$7C,$C7,$F9,$45,$FF,$01,$FD,$05,$7F,$BF,$37,$8B,$FC,$FB,$37
	.dc.b	$8B,$ED,$B6,$9E,$22,$F0,$DF,$4F,$C1,$5B,$E3,$7C,$67,$F8,$63,$0F
	.dc.b	$F4,$9B,$56,$5F,$E9,$B8,$C3,$7D,$AF,$C0,$46,$33,$FF,$53,$36,$97
	.dc.b	$DB,$CA,$DD,$FE,$2B,$07,$CB,$FE,$19,$FF,$20,$B7,$1F,$F6,$03,$70
	.dc.b	$F1,$BF,$90,$7F,$D2,$AE,$67,$9F,$47,$58,$C1,$B9,$FF,$4A,$7C,$A3
	.dc.b	$E5,$FE,$97,$4F,$C9,$9F,$D3,$EF,$F5,$6B,$51,$FE,$A4,$6B,$FE,$9D
	.dc.b	$9E,$07,$18,$CE,$3F,$EB,$26,$E5,$FB,$06,$D7,$F8,$01,$AF,$ED,$BD
	.dc.b	$1F,$89,$F2,$FF,$CF,$6F,$FC,$C6,$DB,$FF,$71,$36,$BE,$AA,$F7,$38
	.dc.b	$F8,$92,$7F,$D7,$33,$BB,$3B,$8F,$BF,$F2,$FF,$6B,$C0,$91,$82,$1F
	.dc.b	$82,$C7,$EE,$C6,$A3,$07,$DE,$E4,$57,$C9,$E3,$F9,$BF,$03,$2B,$E5
	.dc.b	$46,$5F,$F6,$7E,$E9,$D4,$7D,$69,$D3,$4E,$27,$0F,$7F,$F2,$3F,$0D
	.dc.b	$F9,$2F,$9C,$93,$F9,$A8,$CB,$F9,$C8,$CB,$E8,$E8,$FD,$0C,$FE,$6B
	.dc.b	$E6,$E4,$FE,$89,$FC,$EF,$36,$23,$86,$A9,$C7,$0F,$E9,$6B,$F7,$9F
	.dc.b	$A8,$7F,$A5,$AF,$FD,$D3,$C0,$F9,$8F,$D0,$36,$3F,$DD,$F3,$B7,$FD
	.dc.b	$D0,$DB,$FF,$51,$CE,$E7,$E7,$C2,$30,$FF,$C0,$09,$E1,$C5,$FC,$F9
	.dc.b	$B5,$FE,$4E,$6A,$73,$3C,$FE,$6F,$F7,$7F,$39,$FB,$CF,$9D,$97,$FC
	.dc.b	$12,$32,$FF,$06,$8C,$BF,$12,$3F,$81,$E5,$23,$F8,$5B,$FC,$1A,$B1
	.dc.b	$1F,$2A,$2D,$FF,$0D,$9F,$E4,$9F,$5B,$FE,$EA,$AF,$F3,$2C,$FF,$B1
	.dc.b	$3E,$25,$E4,$7F,$09,$6C,$EB,$4F,$EB,$B7,$C6,$3F,$2A,$1C,$5F,$3B
	.dc.b	$E4,$78,$3E,$1E,$57,$33,$FC,$EB,$5C,$2F,$AC,$FE,$DB,$BF,$F9,$0F
	.dc.b	$C7,$7F,$5C,$F2,$26,$1F,$CA,$A3,$F9,$2F,$CE,$FF,$65,$F9,$67,$8D
	.dc.b	$4C,$D7,$F3,$0F,$97,$D7,$6C,$8C,$1F,$CD,$3E,$63,$97,$3C,$8C,$1F
	.dc.b	$CE,$3D,$4E,$DE,$14,$8C,$1F,$CF,$3D,$6F,$17,$81,$23,$07,$F4,$0F
	.dc.b	$98,$F3,$78,$12,$30,$7F,$44,$F9,$8F,$57,$7C,$8C,$1F,$D2,$39,$BD
	.dc.b	$DE,$04,$8C,$1F,$2F,$C2,$7C,$CB,$E6,$7E,$27,$85,$2B,$53,$96,$FC
	.dc.b	$7F,$C9,$F0,$E4,$60,$DF,$6F,$CD,$F0,$A4,$60,$E7,$F2,$3F,$47,$C4
	.dc.b	$91,$83,$BF,$CA,$FD,$5E,$F9,$18,$3C,$7E,$2F,$F8,$77,$45,$53,$11
	.dc.b	$9F,$91,$FB,$DE,$0C,$BF,$5C,$71,$F9,$1D,$46,$0F,$7B,$96,$90,$FE
	.dc.b	$2D,$83,$E6,$7F,$C1,$1F,$39,$FA,$09,$FE,$63,$F2,$AF,$0F,$1E,$53
	.dc.b	$A7,$FD,$9B,$39,$F2,$B8,$92,$FF,$3D,$19,$6D,$F2,$AF,$F2,$3E,$16
	.dc.b	$7F,$DA,$ED,$4D,$35,$4F,$1D,$F1,$7F,$E4,$5F,$A3,$E5,$D8,$5E,$FE
	.dc.b	$4D,$69,$32,$FC,$44,$E9,$B3,$97,$F9,$D7,$CE,$BE,$2B,$FC,$A5,$E9
	.dc.b	$BE,$2E,$9F,$E5,$59,$D5,$7E,$4D,$AF,$2A,$D1,$B9,$FF,$1B,$7A,$7F
	.dc.b	$E5,$EE,$2F,$F7,$BF,$CE,$FF,$57,$BF,$F7,$CE,$3F,$DA,$D4,$65,$FE
	.dc.b	$B9,$EC,$FF,$9C,$23,$17,$CE,$77,$5F,$E7,$48,$C4,$F1,$8E,$8D,$AE
	.dc.b	$51,$BF,$86,$FF,$1A,$DF,$20,$7F,$8F,$1C,$D7,$18,$F4,$CF,$FA,$08
	.dc.b	$EA,$73,$C3,$31,$CB,$0F,$1C,$EF,$0F,$0C,$FC,$06,$BD,$EB,$8B,$4F
	.dc.b	$21,$F0,$8C,$CF,$81,$ED,$87,$9A,$3C,$51,$C1,$D3,$C4,$E7,$7E,$47
	.dc.b	$8D,$74,$BF,$1D,$70,$AD,$57,$8C,$FF,$50,$94,$EE,$61,$FA,$8D,$F8
	.dc.b	$CF,$F3,$5B,$C0,$FF,$33,$B8,$75,$F0,$3F,$EF,$09,$4E,$67,$87,$C4
	.dc.b	$7F,$BD,$1F,$AA,$D4,$D2,$3F,$23,$17,$C3,$FE,$EE,$3E,$59,$C8,$D3
	.dc.b	$9D,$3F,$21,$19,$7F,$25,$3C,$BA,$3E,$AC,$E7,$3F,$F6,$0D,$F1,$7C
	.dc.b	$F0,$F8,$5E,$33,$E1,$7A,$C7,$A0,$D4,$F2,$AF,$C8,$CE,$B1,$FC,$1E
	.dc.b	$1E,$1B,$F4,$9C,$71,$EB,$7E,$48,$C6,$1C,$DD,$DF,$2C,$7C,$D1,$F6
	.dc.b	$A3,$15,$C5,$B1,$C0,$F2,$9F,$E0,$A9,$F5,$57,$BB,$1F,$0C,$EB,$97
	.dc.b	$FC,$A5,$BF,$CF,$CF,$E8,$46,$32,$FE,$B5,$87,$DE,$6D,$6D,$FD,$8C
	.dc.b	$EE,$FF,$1E,$D7,$FD,$E2,$D7,$75,$FE,$EA,$9F,$C7,$4F,$F2,$9F,$57
	.dc.b	$F0,$15,$C3,$A1,$18,$8E,$F1,$8A,$E2,$D8,$F0,$B7,$1F,$5B,$FD,$1E
	.dc.b	$7F,$BA,$39,$79,$F5,$FA,$EB,$FA,$4F,$B9,$57,$5A,$34,$BE,$23,$FD
	.dc.b	$25,$7F,$E2,$9A,$DD,$71,$AE,$13,$FA,$B9,$CB,$63,$31,$FA,$F6,$BF
	.dc.b	$B8,$73,$8F,$E8,$DF,$EC,$1F,$09,$D6,$FC,$78,$A3,$D5,$A7,$F5,$3F
	.dc.b	$56,$E8,$6F,$AF,$86,$69,$F2,$9F,$9D,$A7,$FB,$6A,$B8,$4E,$4F,$9F
	.dc.b	$EA,$EB,$0F,$EC,$0D,$93,$4C,$7A,$92,$BE,$BB,$6F,$EB,$C7,$12,$1F
	.dc.b	$70,$ED,$C9,$FE,$29,$7F,$0E,$D8,$FC,$7B,$5F,$C7,$B5,$FC,$AA,$F4
	.dc.b	$39,$92,$79,$BD,$BC,$F9,$1D,$9B,$5D,$BC,$EF,$19,$27,$C6,$F9,$BC
	.dc.b	$C9,$1E,$66,$CF,$9B,$CC,$F5,$92,$7D,$6F,$BB,$CB,$91,$EE,$6C,$FB
	.dc.b	$BC,$9F,$86,$49,$F8,$6F,$C6,$F2,$E4,$7C,$66,$CF,$C6,$F9,$DF,$96
	.dc.b	$49,$F9,$6F,$CE,$F8,$B9,$1F,$39,$B3,$F3,$BD,$EF,$A6,$49,$FA,$6F
	.dc.b	$D6,$F9,$69,$1F,$59,$B3,$FD,$2F,$E7,$1E,$97,$EB,$5B,$5F,$E9,$1E
	.dc.b	$D7,$CE,$BD,$BD,$5C,$AF,$D5,$5F,$0C,$3A,$4F,$E6,$14,$65,$FE,$C2
	.dc.b	$EF,$0F,$E3,$DF,$37,$FE,$5C,$BF,$DC,$4E,$5F,$F5,$8C,$66,$BF,$DD
	.dc.b	$1C,$CF,$9D,$FD,$6B,$6B,$FC,$FC,$BE,$27,$33,$27,$2F,$08,$C5,$74
	.dc.b	$F9,$DE,$7E,$37,$46,$DF,$46,$CE,$E3,$BC,$FF,$15,$F9,$4F,$F5,$6C
	.dc.b	$7F,$11,$6C,$FF,$BA,$39,$FF,$82,$4E,$7E,$17,$81,$F1,$EB,$75,$FE
	.dc.b	$06,$6B,$FE,$2C,$0E,$1F,$F1,$7B,$DB,$6F,$F6,$B7,$93,$D2,$FE,$D6
	.dc.b	$1F,$CF,$6F,$E4,$FC,$9F,$C8,$3E,$4F,$8F,$6B,$E4,$C5,$9F,$9F,$37
	.dc.b	$0F,$EA,$7B,$BE,$33,$E8,$FA,$5F,$49,$B8,$FD,$26,$E1,$F6,$1B,$7F
	.dc.b	$D3,$3C,$A7,$4F,$F3,$53,$FA,$04,$39,$1E,$06,$71,$FE,$79,$7C,$1F
	.dc.b	$E7,$17,$F8,$F8,$C6,$3F,$CE,$63,$07,$4B,$78,$3C,$E5,$8F,$E4,$A3
	.dc.b	$F9,$BF,$C5,$BE,$73,$F1,$B5,$86,$E5,$FE,$CB,$9D,$3F,$90,$9D,$34
	.dc.b	$FD,$57,$88,$F9,$2F,$6B,$63,$3F,$27,$89,$91,$C1,$DC,$62,$BC,$F3
	.dc.b	$F8,$3D,$5E,$2F,$F5,$D3,$E4,$FC,$EB,$C4,$C5,$7C,$89,$F8,$C4,$7F
	.dc.b	$B0,$36,$FA,$9F,$12,$9E,$2F,$28,$CE,$E9,$CD,$FF,$07,$DB,$F6,$0D
	.dc.b	$EE,$FB,$70,$F1,$2C,$49,$D1,$C6,$31,$87,$8B,$CB,$70,$BE,$77,$EA
	.dc.b	$9F,$5B,$E7,$6E,$1F,$4E,$2D,$F1,$A5,$63,$4F,$F0,$DF,$CE,$3E,$3F
	.dc.b	$F8,$B3,$71,$FF,$02,$37,$0F,$F3,$83,$5F,$F3,$CB,$9A,$FF,$32,$7B
	.dc.b	$6F,$9D,$79,$9F,$E1,$7A,$D6,$1F,$EA,$4A,$D4,$FE,$3B,$E7,$64,$FE
	.dc.b	$2F,$E7,$6E,$5E,$99,$FE,$F9,$A9,$A4,$AF,$F1,$BC,$7F,$B7,$9B,$F0
	.dc.b	$BC,$08,$E0,$FC,$C7,$CE,$FE,$7A,$3F,$3E,$FE,$83,$C8,$EE,$49,$EA
	.dc.b	$DC,$3E,$67,$C0,$EE,$7F,$44,$FE,$A3,$E0,$A4,$60,$FD,$33,$89,$FF
	.dc.b	$29,$E3,$FE,$E7,$9F,$4C,$FB,$8F,$7F,$C0,$C8,$F6,$B6,$B8,$BF,$B6
	.dc.b	$F9,$DF,$DB,$B6,$7F,$C8,$1F,$3B,$FE,$2D,$F1,$FF,$D3,$FE,$0F,$F2
	.dc.b	$98,$C6,$1F,$EA,$08,$FE,$59,$E5,$FD,$8B,$DC,$FF,$B6,$AB,$7F,$04
	.dc.b	$BE,$5E,$E7,$F2,$1A,$E4,$6D,$C4,$EE,$3F,$DC,$78,$75,$76,$2F,$AD
	.dc.b	$F4,$C7,$C3,$39,$C3,$9C,$F9,$B9,$3F,$D6,$58,$F5,$AD,$5F,$0C,$39
	.dc.b	$C7,$8E,$39,$E3,$E1,$AE,$15,$AA,$E2,$79,$1C,$4E,$F0,$F2,$4E,$71
	.dc.b	$F1,$BD,$4F,$5C,$ED,$C5,$F6,$9D,$C7,$68,$F4,$4E,$E3,$15,$DA,$78
	.dc.b	$61,$E8,$C7,$DF,$39,$25,$76,$7E,$26,$5F,$07,$1E,$F9,$7B,$4F,$5D
	.dc.b	$F2,$BE,$04,$3C,$27,$D6,$74,$E0,$91,$CA,$BB,$61,$27,$FD,$E9,$3F
	.dc.b	$8C,$7F,$DF,$AD,$7E,$49,$8F,$75,$F2,$DF,$E9,$19,$46,$27,$9B,$5E
	.dc.b	$13,$B8,$D6,$7F,$69,$B9,$FE,$75,$B5,$F9,$9F,$96,$FF,$04,$56,$EF
	.dc.b	$F0,$D5,$AA,$FC,$93,$70,$5F,$0C,$3A,$5F,$02,$5F,$E1,$E7,$4F,$E9
	.dc.b	$EF,$91,$27,$F2,$93,$A9,$EF,$1C,$B0,$F2,$47,$33,$17,$E0,$7F,$29
	.dc.b	$7E,$5B,$E5,$87,$C8,$C3,$75,$E1,$7B,$B8,$3B,$9C,$9C,$B4,$98,$78
	.dc.b	$8F,$D2,$D7,$6B,$F5,$4E,$B0,$FE,$15,$39,$BE,$D8,$78,$EB,$D4,$6C
	.dc.b	$F2,$14,$FE,$01,$3F,$23,$BF,$FA,$EF,$C0,$7F,$8F,$7B,$7F,$DF,$FC
	.dc.b	$8F,$4F,$1B,$FF,$BF,$23,$18,$FF,$84,$6F,$2B,$0E,$6F,$F4,$48,$FF
	.dc.b	$6F,$46,$23,$F9,$8C,$62,$5E,$8D,$24,$FF,$4C,$6B,$FD,$AD,$EF,$5F
	.dc.b	$DD,$1B,$2F,$88,$BF,$15,$86,$5F,$F8,$9E,$32,$FF,$58,$7C,$44,$7F
	.dc.b	$AD,$F9,$BD,$38,$24,$7F,$8B,$E3,$B9,$E9,$23,$C0,$FE,$C7,$C1,$ED
	.dc.b	$23,$82,$1F,$E0,$57,$D2,$7A,$BE,$1B,$E5,$C3,$E7,$3F,$CA,$A8,$FF
	.dc.b	$12,$F8,$4F,$9D,$CC,$61,$BC,$4E,$1F,$F3,$C5,$E1,$BE,$A7,$F9,$EE
	.dc.b	$E4,$F2,$6C,$65,$D9,$78,$5F,$15,$FE,$19,$8F,$26,$1D,$DF,$C1,$46
	.dc.b	$5F,$F1,$53,$63,$53,$CB,$DA,$FB,$73,$9C,$7D,$D3,$DB,$C9,$C9,$AA
	.dc.b	$FC,$7D,$BE,$75,$CD,$86,$E5,$FE,$C9,$6B,$FE,$48,$7F,$32,$D8,$E0
	.dc.b	$47,$87,$3F,$D2,$F9,$D7,$CC,$7D,$4E,$8F,$95,$5F,$4D,$AA,$76,$FC
	.dc.b	$4D,$7F,$D5,$11,$88,$F9,$8D,$7E,$93,$5D,$D7,$D1,$7F,$01,$1C,$72
	.dc.b	$F9,$D1,$DE,$7C,$C7,$97,$D0,$F8,$BF,$3B,$FE,$05,$9F,$F3,$3B,$6B
	.dc.b	$EB,$07,$18,$7D,$A5,$EE,$72,$F0,$65,$FF,$BC,$E7,$6F,$FA,$61,$DB
	.dc.b	$EF,$3F,$03,$F3,$7F,$73,$BE,$46,$78,$FF,$13,$3F,$80,$33,$1F,$86
	.dc.b	$9C,$3C,$CF,$99,$C7,$F4,$7E,$02,$46,$77,$FC,$43,$F1,$A7,$E4,$3C
	.dc.b	$0F,$95,$1F,$93,$7E,$7C,$6A,$38,$C6,$CF,$F4,$89,$75,$F9,$FC,$0F
	.dc.b	$7F,$F2,$FF,$0B,$F9,$8F,$9B,$97,$ED,$E8,$4E,$A3,$13,$A7,$F4,$71
	.dc.b	$FA,$0F,$9B,$91,$C5,$FD,$1F,$2E,$23,$F5,$31,$72,$FD,$4E,$1F,$83
	.dc.b	$E4,$FF,$7A,$7C,$EB,$9B,$F4,$BC,$2F,$98,$A7,$F0,$C9,$D6,$5F,$35
	.dc.b	$AF,$EF,$DA,$FF,$05,$6B,$DA,$BE,$6C,$70,$C7,$F2,$6D,$C3,$55,$F8
	.dc.b	$39,$FE,$F7,$F3,$3F,$BE,$F9,$AF,$DF,$FC,$DC,$AF,$91,$19,$7E,$54
	.dc.b	$65,$FE,$13,$1F,$C1,$B2,$91,$F3,$DF,$E1,$35,$88,$FE,$1F,$16,$FE
	.dc.b	$3C,$FF,$2C,$F9,$DF,$F7,$8D,$79,$7E,$25,$E4,$FF,$0C,$6C,$7E,$46
	.dc.b	$76,$FE,$11,$AF,$F2,$96,$BF,$CD,$43,$87,$F9,$C3,$5C,$FC,$3F,$8C
	.dc.b	$8F,$EB,$7C,$3F,$E4,$7F,$1B,$FD,$A3,$88,$D2,$4F,$F3,$38,$7C,$E6
	.dc.b	$C9,$19,$79,$7C,$AC,$8E,$0D,$7F,$A1,$77,$FF,$3C,$F9,$7F,$01,$23
	.dc.b	$B3,$E3,$F4,$A4,$78,$5A,$F2,$C1,$EA,$FA,$49,$1E,$77,$D9,$F2,$B2
	.dc.b	$3D,$4D,$79,$7C,$AF,$C1,$23,$EF,$7E,$27,$3C,$8F,$84,$D7,$D7,$F3
	.dc.b	$8F,$9D,$FF,$08,$B6,$3F,$AD,$35,$DC,$6B,$E3,$E4,$69,$AF,$4E,$2E
	.dc.b	$64,$79,$BD,$BE,$46,$47,$66,$BE,$4F,$97,$F1,$91,$F1,$BE,$6E,$69
	.dc.b	$1E,$66,$BE,$CF,$8F,$F5,$91,$F5,$BE,$EF,$8F,$91,$EE,$6B,$F1,$1F
	.dc.b	$F1,$4E,$66,$98,$F9,$7F,$F0,$8F,$CE,$FE,$8A,$9D,$BF,$DF,$7E,$4D
	.dc.b	$D1,$98,$C3,$78,$7F,$45,$76,$FF,$C3,$0F,$37,$FC,$20,$D7,$7D,$BF
	.dc.b	$CE,$23,$11,$C2,$3C,$64,$7F,$A7,$60,$F9,$97,$C5,$7F,$91,$3E,$77
	.dc.b	$C1,$D8,$7F,$C8,$97,$FD,$45,$AF,$B3,$2F,$F0,$7C,$70,$3C,$33,$C7
	.dc.b	$1F,$F2,$CC,$60,$CF,$CE,$BE,$27,$FC,$BA,$DC,$3F,$C4,$CD,$7F,$5A
	.dc.b	$D7,$F8,$35,$A3,$5B,$7F,$89,$1B,$8F,$F8,$89,$B5,$FE,$6D,$5F,$15
	.dc.b	$FE,$46,$ED,$FF,$3B,$5F,$F9,$17,$DF,$7A,$9F,$C4,$CF,$EE,$B8,$23
	.dc.b	$2F,$F9,$2E,$33,$18,$8F,$F1,$EF,$B3,$F8,$07,$84,$FF,$1D,$BE,$F0
	.dc.b	$F9,$DE,$13,$D9,$BE,$D7,$F2,$37,$A5,$70,$6D,$F5,$77,$7F,$E5,$D6
	.dc.b	$BD,$6F,$C1,$3B,$6E,$3B,$7D,$D8,$77,$6D,$64,$FF,$34,$61,$E1,$8E
	.dc.b	$31,$EC,$3E,$14,$FF,$39,$C7,$C2,$79,$AF,$E6,$57,$0C,$7C,$A4,$70
	.dc.b	$4B,$E6,$78,$4F,$A7,$7E,$57,$C7,$8F,$AD,$B7,$FE,$3F,$C7,$E6,$C7
	.dc.b	$F9,$D5,$AF,$A0,$C7,$C5,$3E,$07,$83,$63,$FC,$F1,$28,$C4,$70,$9F
	.dc.b	$F1,$2A,$F8,$8A,$47,$FC,$52,$DC,$FF,$09,$87,$FA,$DB,$7F,$65,$7F
	.dc.b	$A4,$EB,$DD,$39,$8F,$B0,$7F,$A3,$63,$3F,$0F,$FE,$85,$BF,$C7,$4F
	.dc.b	$13,$67,$F0,$E8,$D6,$BF,$E9,$1A,$FA,$33,$FE,$8E,$9F,$F4,$76,$DC
	.dc.b	$A3,$D4,$7D,$B3,$96,$3C,$1F,$EE,$95,$E2,$8F,$4C,$70,$C3,$DB,$7E
	.dc.b	$08,$F4,$1E,$4C,$7D,$93,$DE,$F9,$4F,$A0,$F5,$47,$2B,$EB,$3A,$8E
	.dc.b	$13,$CE,$FB,$3E,$43,$15,$F9,$7E,$1F,$DC,$FC,$5F,$F9,$DA,$3F,$D4
	.dc.b	$79,$A3,$15,$FE,$EF,$8F,$E2,$53,$FE,$6D,$74,$D9,$FF,$43,$CF,$26
	.dc.b	$BC,$5B,$5E,$09,$FF,$5F,$57,$FC,$01,$6E,$FA,$8C,$5F,$F8,$21,$AF
	.dc.b	$F1,$C7,$39,$FF,$AF,$BB,$5F,$2F,$E8,$C7,$F8,$AE,$16,$73,$8F,$A7
	.dc.b	$E2,$75,$8F,$FB,$EE,$3D,$9F,$05,$89,$91,$E6,$7C,$84,$70,$7C,$3D
	.dc.b	$77,$A7,$F3,$F3,$E4,$3D,$BE,$27,$BF,$C4,$DA,$F9,$4E,$1F,$86,$FC
	.dc.b	$12,$3F,$1B,$8B,$CE,$6E,$FB,$6B,$FC,$CA,$3E,$07,$7B,$E1,$AF,$C1
	.dc.b	$61,$B8,$7F,$A9,$65,$ED,$FD,$46,$C7,$FB,$C5,$B8,$FF,$8E,$1A,$FF
	.dc.b	$4C,$5E,$FF,$80,$F0,$FC,$1D,$AE,$DE,$FF,$93,$E0,$3C,$AD,$8F,$17
	.dc.b	$BB,$E1,$23,$E1,$7C,$9E,$DF,$B0,$F6,$FB,$B6,$7D,$5E,$D7,$C0,$F7
	.dc.b	$7E,$0B,$63,$DD,$E6,$7B,$48,$FB,$5F,$81,$EC,$FC,$83,$E4,$FB,$56
	.dc.b	$7E,$37,$B3,$F1,$48,$FC,$57,$E4,$30,$F9,$05,$E1,$FF,$2F,$FC,$1B
	.dc.b	$D8,$FD,$0B,$67,$FC,$F5,$21,$D1,$FE,$FD,$AD,$9C,$3D,$79,$5C,$2B
	.dc.b	$76,$9C,$C7,$83,$D6,$EE,$4A,$E1,$82,$D3,$E4,$F5,$25,$6A,$51,$8B
	.dc.b	$F0,$7C,$DF,$FA,$0A,$BF,$DD,$0D,$8F,$D0,$78,$CF,$99,$EB,$5A,$F7
	.dc.b	$BF,$85,$D2,$F8,$3F,$09,$AF,$FA,$75,$AF,$C5,$6B,$C6,$3F,$8A,$DB
	.dc.b	$F4,$C7,$BD,$DA,$FC,$92,$C3,$E7,$3B,$9F,$E0,$D9,$FF,$04,$C7,$F5
	.dc.b	$6D,$9E,$11,$F3,$BD,$AD,$F9,$73,$0E,$67,$D9,$DD,$E5,$3E,$3B,$58
	.dc.b	$7E,$DA,$31,$5C,$23,$EC,$FC,$0F,$88,$F1,$D5,$9D,$5F,$D7,$9F,$E3
	.dc.b	$46,$6B,$A3,$F0,$B6,$72,$D3,$1F,$03,$70,$FF,$73,$4E,$9F,$C5,$6B
	.dc.b	$F2,$6B,$F9,$BB,$63,$33,$BC,$3F,$2F,$39,$7E,$FF,$8C,$EE,$ED,$18
	.dc.b	$7E,$6F,$C2,$4C,$3D,$7C,$2F,$98,$C5,$FD,$0A,$FE,$6D,$18,$C7,$F9
	.dc.b	$E3,$F5,$BC,$F9,$87,$2F,$9C,$B8,$7D,$89,$FE,$7D,$86,$30,$FE,$88
	.dc.b	$FE,$43,$E0,$1E,$CD,$31,$77,$1F,$A8,$A7,$71,$FA,$F8,$C4,$FF,$5C
	.dc.b	$6C,$66,$77,$1F,$B5,$6D,$FE,$28,$FD,$7F,$A1,$DE,$7F,$4F,$DB,$FE
	.dc.b	$1B,$9F,$C7,$F2,$7E,$D7,$E2,$3F,$2A,$7E,$5B,$E7,$2E,$5F,$2A,$3F
	.dc.b	$AE,$6D,$8C,$7F,$1C,$FE,$0F,$C7,$95,$F8,$0F,$22,$D7,$ED,$B9,$55
	.dc.b	$F8,$9F,$89,$CE,$9E,$79,$47,$C9,$CE,$63,$F8,$AF,$3E,$E3,$FD,$7C
	.dc.b	$F9,$13,$F1,$78,$0D,$25,$7C,$77,$F3,$1C,$FC,$B2,$F0,$3F,$B4,$E2
	.dc.b	$95,$F8,$8F,$92,$B8,$FE,$DA,$7F,$BE,$72,$E3,$0F,$F0,$13,$FB,$C9
	.dc.b	$97,$F9,$0F,$8D,$B1,$FB,$E3,$FC,$09,$96,$2F,$FC,$14,$FF,$82,$7E
	.dc.b	$99,$F3,$5F,$E1,$86,$E1,$FC,$A9,$B7,$FC,$EB,$95,$18,$8F,$F5,$8B
	.dc.b	$AC,$3F,$C2,$DC,$BD,$70,$FF,$0A,$3D,$FE,$62,$63,$FC,$02,$DF,$37
	.dc.b	$8C,$7F,$BF,$C2,$BC,$84,$AF,$23,$FE,$30,$F8,$07,$DD,$3E,$77,$0D
	.dc.b	$AF,$E9,$53,$A9,$C9,$23,$F8,$8F,$8A,$E3,$FF,$21,$D6,$E7,$31,$F1
	.dc.b	$FE,$2B,$E3,$12,$3E,$33,$73,$F4,$FC,$CE,$E7,$DC,$FC,$FF,$87,$91
	.dc.b	$83,$E1,$38,$9F,$82,$FD,$59,$C7,$0F,$EA,$36,$FF,$29,$D7,$D3,$6B
	.dc.b	$F4,$BE,$0F,$ED,$13,$0F,$6F,$3F,$DF,$8F,$F2,$33,$FD,$5B,$E6,$5E
	.dc.b	$43,$A3,$FD,$25,$F0,$3F,$DD,$F1,$56,$A7,$31,$FD,$EF,$7F,$E7,$EF
	.dc.b	$FA,$86,$B8,$56,$E3,$07,$F8,$05,$C7,$C3,$27,$78,$78,$6B,$67,$D3
	.dc.b	$F3,$3F,$AA,$D6,$E3,$94,$EE,$31,$1E,$68,$C5,$74,$C7,$CF,$3C,$F4
	.dc.b	$B8,$F8,$AB,$53,$E0,$8C,$3E,$DC,$7C,$33,$AC,$7B,$D6,$AB,$C1,$87
	.dc.b	$6C,$BE,$1D,$6B,$2F,$1D,$7A,$2B,$C3,$18,$8F,$21,$E1,$71,$7D,$4E
	.dc.b	$D7,$D3,$0F,$11,$2B,$A1,$F2,$23,$4E,$30,$F1,$61,$E0,$3C,$D8,$FB
	.dc.b	$27,$57,$CD,$F5,$9C,$E3,$4E,$2B,$9B,$E7,$AE,$79,$75,$8C,$12,$7E
	.dc.b	$AE,$1E,$48,$E5,$87,$D0,$7F,$C8,$B8,$FE,$1D,$B9,$7E,$0D,$B8,$FF
	.dc.b	$55,$97,$AE,$F9,$9C,$36,$BF,$C9,$13,$A9,$C9,$23,$EF,$4E,$F2,$FB
	.dc.b	$6D,$C7,$F3,$3E,$0F,$FB,$79,$FC,$DC,$66,$BF,$34,$D7,$F4,$AD,$C3
	.dc.b	$FD,$1F,$1F,$8E,$8C,$13,$0F,$C8,$DF,$59,$E7,$1B,$23,$F9,$38,$D3
	.dc.b	$DA,$33,$87,$A1,$E7,$18,$C3,$CF,$16,$3F,$61,$5C,$B0,$F8,$91,$F0
	.dc.b	$A7,$4F,$EF,$A7,$58,$79,$E3,$47,$8C,$DD,$72,$C3,$C5,$19,$36,$E0
	.dc.b	$98,$7C,$C8,$E3,$5D,$67,$F5,$77,$E0,$C3,$27,$B2,$BF,$84,$F6,$FF
	.dc.b	$50,$F8,$6F,$E2,$9C,$7F,$EE,$3D,$2E,$FF,$E5,$5E,$8F,$C4,$4F,$F2
	.dc.b	$7A,$CE,$1F,$E9,$0E,$44,$7F,$87,$F1,$E9,$3F,$D7,$E3,$13,$A8,$FE
	.dc.b	$C5,$E4,$F2,$AF,$EC,$2D,$93,$7E,$1B,$0C,$AF,$EB,$CE,$5F,$EC,$1F
	.dc.b	$1B,$5F,$EB,$F3,$CE,$24,$7F,$71,$7C,$64,$7F,$AD,$FE,$41,$FE,$AF
	.dc.b	$7F,$AC,$9E,$5C,$FF,$CB,$3A,$7F,$AB,$FB,$15,$BA,$FF,$7C,$46,$3D
	.dc.b	$BF,$F2,$03,$63,$AC,$62,$78,$47,$F7,$EF,$86,$F8,$A4,$7F,$C0,$CD
	.dc.b	$7F,$BC,$BF,$14,$F8,$DF,$0F,$F4,$A3,$E8,$FB,$5F,$2F,$C8,$FE,$F1
	.dc.b	$1F,$E8,$26,$BF,$E2,$FE,$14,$F1,$6C,$7F,$95,$DA,$FF,$B1,$DA,$FF
	.dc.b	$B4,$1A,$FF,$A6,$7D,$E7,$95,$E1,$F2,$64,$FF,$A4,$A7,$2D,$9F,$C0
	.dc.b	$46,$21,$CB,$FE,$5A,$8C,$37,$E1,$F7,$F0,$BE,$DF,$2A,$5F,$58,$CB
	.dc.b	$48,$74,$E2,$F3,$75,$FC,$7F,$87,$FE,$BB,$8F,$F9,$B7,$F3,$67,$C4
	.dc.b	$FE,$72,$11,$8C,$3F,$CF,$B5,$C3,$0F,$EF,$4F,$F6,$68,$FF,$79,$B5
	.dc.b	$CF,$83,$FE,$81,$51,$97,$FD,$F0,$F3,$6C,$79,$4F,$15,$7E,$8A,$3F
	.dc.b	$D8,$17,$E4,$F8,$EC,$F9,$52,$73,$49,$E2,$48,$CB,$D3,$CA,$FA,$3F
	.dc.b	$13,$F8,$7F,$3E,$57,$D7,$8C,$B4,$87,$CB,$1E,$13,$8D,$F1,$5F,$37
	.dc.b	$C2,$F9,$5F,$40,$F8,$6A,$67,$0F,$97,$3F,$17,$87,$DD,$1F,$E2,$53
	.dc.b	$E2,$C6,$48,$F0,$21,$FE,$89,$8F,$E1,$8F,$97,$1F,$27,$AD,$FA,$3F
	.dc.b	$9E,$93,$F9,$38,$FC,$89,$1C,$1F,$97,$F8,$CF,$CA,$FB,$5F,$AB,$F9
	.dc.b	$E9,$3F,$9B,$8C,$B4,$87,$87,$67,$B9,$F8,$BE,$FF,$E1,$33,$F9,$EE
	.dc.b	$FF,$F2,$2B,$F7,$1A,$FC,$DE,$8F,$F4,$4B,$C4,$FF,$65,$38,$7F,$DE
	.dc.b	$33,$F8,$D8,$C4,$7F,$BB,$A3,$BC,$60,$FF,$65,$38,$3F,$9D,$F5,$7E
	.dc.b	$E7,$E3,$7F,$84,$7A,$F2,$FF,$71,$1F,$B6,$23,$83,$F7,$9F,$07,$FB
	.dc.b	$AF,$13,$F8,$87,$54,$9F,$E0,$11,$96,$90,$F8,$DF,$1B,$FC,$67,$E6
	.dc.b	$3F,$95,$69,$E6,$F6,$7E,$41,$9F,$32,$5F,$FB,$AE,$BF,$94,$BF,$E6
	.dc.b	$B6,$A0,$E1,$B1,$C9,$FF,$68,$47,$83,$5E,$16,$F8,$2F,$E0,$15,$F6
	.dc.b	$F8,$BF,$AE,$1F,$9F,$7F,$2C,$7E,$89,$C3,$FE,$BB,$6B,$FE,$7F,$E8
	.dc.b	$FD,$37,$B1,$C8,$91,$CA,$32,$D9,$7C,$57,$80,$91,$97,$FB,$77,$D9
	.dc.b	$3E,$66,$92,$7A,$46,$5B,$F3,$3C,$08,$E0,$DF,$CC,$F4,$23,$83,$9F
	.dc.b	$CC,$FF,$77,$A3,$4C,$69,$9F,$8F,$F2,$92,$BC,$6F,$A3,$E0,$79,$7C
	.dc.b	$7C,$9F,$A3,$C3,$F6,$77,$64,$BF,$07,$C6,$4B,$D9,$FC,$4A,$51,$E5
	.dc.b	$F8,$FF,$37,$C6,$7C,$62,$57,$C4,$7E,$4F,$C5,$7A,$FD,$89,$3D,$32
	.dc.b	$F0,$CF,$BF,$E1,$A4,$E7,$2F,$F6,$E3,$7E,$2B,$E2,$7C,$1F,$D4,$24
	.dc.b	$7D,$27,$FC,$49,$F2,$FF,$DC,$BA,$9C,$1E,$FC,$1D,$BF,$34,$FF,$12
	.dc.b	$56,$E7,$E2,$3F,$C0,$6B,$8E,$1F,$CC,$A3,$13,$AF,$13,$8F,$B5,$2B
	.dc.b	$F1,$D4,$8F,$19,$23,$3E,$66,$DF,$E5,$53,$B7,$E2,$28,$F2,$F8,$7E
	.dc.b	$6F,$2A,$56,$A3,$2F,$81,$F8,$24,$70,$6C,$87,$83,$C7,$F8,$3E,$A7
	.dc.b	$F8,$6F,$D6,$93,$E2,$8C,$BE,$C7,$E6,$91,$FD,$67,$CC,$3D,$EF,$F2
	.dc.b	$A3,$63,$31,$82,$3F,$81,$8D,$3F,$D2,$DB,$BF,$D0,$E6,$47,$F7,$99
	.dc.b	$7C,$C7,$F9,$8A,$BF,$89,$7C,$BB,$A7,$FE,$00,$1F,$7A,$B2,$E3,$93
	.dc.b	$F8,$C4,$27,$83,$FC,$6E,$FA,$39,$AD,$C6,$27,$73,$DE,$B9,$C6,$2B
	.dc.b	$C1,$3C,$4E,$9F,$01,$FE,$30,$70,$7A,$9C,$4F,$1A,$E7,$5E,$7A,$F4
	.dc.b	$9C,$B0,$AF,$16,$C7,$82,$B9,$4E,$B0,$F4,$4F,$AB,$0F,$17,$C0,$EA
	.dc.b	$8D,$3D,$B5,$F0,$9B,$3E,$6A,$F2,$57,$A9,$F0,$1D,$70,$F0,$4E,$E7
	.dc.b	$FA,$34,$E9,$B8,$FB,$9F,$61,$F1,$30,$FA,$2F,$F8,$55,$AF,$53,$FA
	.dc.b	$1B,$87,$EC,$7C,$C3,$E1,$3F,$A3,$C7,$E1,$A3,$0F,$06,$BD,$5A,$FF
	.dc.b	$4E,$F8,$AF,$F2,$52,$69,$1F,$F2,$73,$F4,$B2,$FF,$54,$F4,$FD,$49
	.dc.b	$FE,$C9,$39,$F6,$7F,$84,$ED,$F8,$89,$C9,$F8,$2B,$E2,$FE,$10,$CC
	.dc.b	$7E,$3A,$32,$F1,$8F,$21,$C6,$7A,$E9,$E9,$8C,$61,$E3,$77,$5C,$EB
	.dc.b	$71,$FE,$41,$C3,$CD,$5B,$6C,$78,$CE,$37,$FE,$D8,$AE,$31,$FA,$A9
	.dc.b	$DD,$F9,$5B,$5E,$38,$F0,$5F,$5A,$F2,$9F,$A5,$E0,$FD,$C7,$C6,$FF
	.dc.b	$2E,$CB,$F8,$06,$08,$C5,$FF,$57,$8F,$F3,$83,$FE,$AF,$8C,$DF,$F9
	.dc.b	$9A,$78,$4E,$4F,$F5,$A7,$1E,$B0,$EF,$1F,$EB,$7E,$0C,$57,$FB,$B2
	.dc.b	$3F,$D7,$5A,$E2,$64,$6C,$C9,$70,$FF,$31,$57,$FA,$72,$3F,$D8,$5F
	.dc.b	$05,$D3,$E2,$7A,$B2,$32,$7F,$B1,$E3,$15,$CA,$7C,$7B,$E3,$5F,$86
	.dc.b	$F9,$BE,$0E,$61,$FE,$CF,$F8,$9B,$87,$8B,$9F,$D6,$63,$0F,$96,$FB
	.dc.b	$B9,$A5,$FF,$B5,$E2,$D7,$F9,$EF,$DE,$7C,$3F,$C7,$6D,$7F,$B7,$7E
	.dc.b	$2F,$13,$9F,$87,$93,$FD,$EE,$BF,$DA,$AD,$8E,$5F,$09,$23,$5F,$11
	.dc.b	$71,$F0,$7C,$17,$7F,$86,$F0,$36,$BA,$7C,$17,$8C,$F2,$7C,$25,$8F
	.dc.b	$07,$C0,$F9,$CF,$47,$C2,$58,$F2,$7C,$0F,$AF,$E0,$FD,$8D,$8F,$47
	.dc.b	$CC,$7B,$CF,$81,$F0,$96,$3D,$9F,$1B,$F0,$CF,$89,$F0,$56,$3E,$07
	.dc.b	$CA,$FC,$73,$E4,$7C,$25,$8F,$88,$7C,$B3,$E6,$7C,$25,$7F,$C4,$5F
	.dc.b	$2A,$F8,$7F,$CB,$36,$33,$F0,$D8,$9C,$FC,$C4,$9E,$3F,$0F,$79,$FF
	.dc.b	$D6,$98,$7F,$9B,$67,$AF,$C2,$F4,$F8,$7E,$AF,$6F,$8A,$93,$E2,$F8
	.dc.b	$8B,$87,$FB,$44,$7C,$BB,$8F,$F5,$F1,$87,$F6,$3E,$2B,$E6,$3C,$13
	.dc.b	$C6,$B8,$61,$FE,$4A,$8E,$0D,$8C,$C7,$FA,$CE,$78,$D7,$F8,$C2,$36
	.dc.b	$73,$BF,$0D,$70,$9F,$E1,$15,$C2,$FF,$BE,$CE,$5F,$F2,$F6,$8F,$89
	.dc.b	$F9,$A7,$CE,$F8,$AB,$3F,$4A,$2E,$5F,$2A,$6E,$3F,$E1,$3B,$7C,$47
	.dc.b	$D6,$3E,$BF,$C5,$5C,$7E,$DC,$5C,$3E,$9C,$DB,$FE,$3D,$F2,$EE,$8F
	.dc.b	$92,$DC,$3F,$CD,$F2,$87,$46,$AA,$56,$A7,$86,$1F,$ED,$B7,$71,$99
	.dc.b	$DC,$FF,$B7,$FE,$35,$E0,$63,$1F,$E6,$11,$8B,$FC,$9F,$CC,$7E,$48
	.dc.b	$FC,$9F,$C6,$D8,$C9,$30,$FC,$7C,$65,$FC,$8F,$A9,$F9,$A3,$F3,$7F
	.dc.b	$1B,$6B,$F3,$F1,$67,$F2,$F3,$71,$FC,$07,$CC,$7F,$9D,$AF,$F1,$0D
	.dc.b	$AF,$CA,$3F,$ED,$18,$CF,$C0,$7E,$52,$35,$5C,$A3,$15,$96,$57,$03
	.dc.b	$AC,$60,$FD,$4F,$C6,$7E,$C8,$FD,$9F,$C7,$59,$FD,$BC,$59,$FD,$7C
	.dc.b	$DB,$FB,$BE,$47,$DC,$3E,$EF,$95,$70,$FB,$11,$6F,$EC,$4D,$AF,$83
	.dc.b	$7E,$B3,$F0,$BE,$7D,$C3,$E1,$C5,$BF,$C0,$4D,$AE,$35,$A3,$F0,$D1
	.dc.b	$97,$F4,$79,$7F,$0E,$3F,$87,$FC,$75,$8F,$E2,$71,$6F,$F8,$5C,$DB
	.dc.b	$FD,$14,$7F,$16,$3F,$8B,$FC,$75,$BF,$E3,$71,$4F,$E2,$71,$67,$FC
	.dc.b	$F9,$F2,$AF,$8E,$FE,$E8,$DB,$FF,$62,$2F,$97,$CC,$C9,$CB,$C2,$31
	.dc.b	$82,$FA,$FC,$9F,$43,$AF,$C8,$5C,$3A,$C5,$AE,$11,$FE,$E3,$C7,$FD
	.dc.b	$70,$E8,$FF,$70,$FC,$43,$E6,$3C,$91,$86,$D7,$21,$F0,$AB,$FF,$5D
	.dc.b	$56,$F3,$FF,$0B,$E1,$CD,$B8,$72,$CF,$E2,$FC,$86,$27,$E2,$FC,$74
	.dc.b	$C3,$E2,$FC,$85,$D3,$C1,$A7,$7A,$C6,$3F,$3A,$AD,$FB,$7E,$3F,$E7
	.dc.b	$9F,$43,$E4,$2C,$7D,$08,$B8,$F4,$D9,$58,$C3,$EC,$FC,$67,$D8,$3E
	.dc.b	$CF,$C6,$5C,$7E,$CC,$5C,$39,$4D,$AF,$E7,$9F,$31,$FC,$73,$97,$FC
	.dc.b	$CF,$F1,$1F,$EB,$48,$C5,$FF,$9C,$63,$FD,$6F,$CE,$F2,$3F,$DE,$7F
	.dc.b	$1E,$9E,$1E,$4B,$1E,$78,$06,$3D,$2F,$84,$66,$31,$7D,$B0,$EA,$F4
	.dc.b	$BE,$75,$CB,$0E,$CF,$94,$E4,$E3,$0E,$B0,$C7,$D1,$AF,$2C,$BB,$CF
	.dc.b	$0C,$31,$87,$1C,$7C,$57,$89,$95,$DC,$F1,$3E,$4C,$3A,$C6,$70,$EA
	.dc.b	$E8,$C6,$1D,$5F,$89,$3C,$31,$F3,$D7,$02,$56,$EB,$D5,$97,$C4,$7D
	.dc.b	$07,$4C,$3C,$B7,$F0,$30,$E9,$97,$C9,$AD,$12,$75,$7C,$E3,$CD,$1E
	.dc.b	$2C,$3C,$18,$FA,$B0,$E7,$5E,$0A,$D6,$3E,$0C,$7E,$3C,$F5,$7C,$13
	.dc.b	$A8,$F9,$15,$C9,$EF,$3E,$57,$F0,$C7,$B0,$A7,$C9,$2D,$FE,$51,$BE
	.dc.b	$DB,$E1,$BF,$D9,$F2,$8C,$4F,$23,$F2,$2E,$1E,$33,$C2,$37,$87,$D2
	.dc.b	$6D,$FE,$8D,$B5,$F9,$7E,$EF,$F5,$84,$E6,$FF,$19,$5A,$AE,$F1,$F8
	.dc.b	$FA,$DC,$7E,$7E,$75,$87,$9E,$B8,$9B,$9A,$7E,$2E,$76,$70,$AE,$71
	.dc.b	$D8,$D4,$F0,$C3,$F1,$73,$AB,$ED,$5A,$24,$E9,$E1,$87,$AD,$CB,$FC
	.dc.b	$0E,$B9,$61,$E2,$C3,$75,$F0,$CA,$78,$EF,$E5,$47,$AE,$39,$9F,$05
	.dc.b	$DE,$0B,$EB,$18,$BF,$75,$58,$DC,$FC,$AE,$6F,$E7,$9F,$1B,$F8,$8E
	.dc.b	$0F,$EB,$6D,$E0,$FE,$94,$DC,$3F,$A1,$F2,$3E,$3B,$18,$72,$9F,$EB
	.dc.b	$F5,$8A,$36,$B1,$F1,$92,$BF,$B2,$BB,$F8,$EC,$47,$F3,$78,$B9,$7F
	.dc.b	$6E,$9E,$9F,$1B,$2B,$FB,$6B,$DF,$E3,$B1,$1F,$CF,$E2,$D7,$F7,$69
	.dc.b	$F5,$78,$0F,$97,$FF,$4B,$5F,$F6,$46,$BF,$D8,$63,$5F,$01,$EB,$E5
	.dc.b	$E3,$5A,$C1,$C1,$FC,$D6,$F7,$87,$3C,$38,$3F,$E0,$2F,$8F,$F8,$84
	.dc.b	$7F,$C0,$BF,$1F,$5F,$EA,$D1,$70,$FF,$06,$CF,$D1,$9F,$97,$EE,$5E
	.dc.b	$1F,$F4,$D4,$7F,$A4,$DA,$FF,$5F,$E5,$D4,$C8,$D3,$F5,$63,$EC,$FC
	.dc.b	$2B,$D4,$A4,$AF,$EA,$18,$FB,$9F,$F0,$1C,$FF,$BB,$1C,$EB,$FD,$57
	.dc.b	$99,$3C,$E7,$27,$0B,$E8,$FF,$7E,$8E,$4D,$F8,$CF,$19,$1E,$35,$D3
	.dc.b	$A3,$11,$B3,$34,$AF,$5C,$FC,$7F,$8E,$FE,$B3,$97,$FA,$C5,$B9,$7F
	.dc.b	$26,$6E,$1F,$E2,$FF,$97,$74,$71,$E5,$93,$FE,$BF,$AD,$E0,$BD,$CE
	.dc.b	$6B,$6F,$FB,$3E,$B7,$3A,$8F,$F1,$65,$71,$F0,$BF,$58,$D8,$FF,$53
	.dc.b	$CA,$72,$FF,$86,$23,$31,$82,$3C,$3C,$4F,$B7,$F1,$D2,$3E,$2B,$F7
	.dc.b	$BE,$3F,$11,$F4,$62,$E1,$F7,$78,$3E,$CD,$7C,$9D,$25,$FC,$17,$E6
	.dc.b	$72,$62,$3D,$D1,$6F,$E7,$4D,$DB,$E7,$4D,$AF,$F0,$D6,$BF,$C1,$1F
	.dc.b	$B3,$F1,$DF,$92,$23,$F7,$3E,$3E,$BF,$87,$8B,$8F,$9F,$8D,$5F,$9A
	.dc.b	$F5,$65,$FE,$11,$FC,$EF,$AB,$88,$FC,$84,$5B,$E5,$3F,$4B,$D3,$FE
	.dc.b	$CF,$1F,$EC,$AF,$27,$EA,$F4,$7F,$B6,$30,$FF,$6D,$D6,$B6,$FD,$2B
	.dc.b	$63,$FC,$F0,$DB,$FF,$42,$47,$5F,$27,$B3,$0E,$66,$4F,$CD,$FC,$7F
	.dc.b	$ED,$08,$FE,$7B,$DD,$AF,$EA,$62,$D7,$C3,$9F,$DD,$F6,$CA,$FD,$2B
	.dc.b	$FB,$EF,$8F,$C4,$7E,$C6,$2D,$7C,$B9,$FE,$1D,$F2,$DF,$C7,$73,$FF
	.dc.b	$05,$B6,$FF,$CA,$4D,$A7,$C5,$3C,$6E,$3E,$24,$BF,$E0,$22,$F8,$56
	.dc.b	$E3,$31,$D5,$D4,$F1,$7F,$4F,$3B,$7F,$84,$72,$7F,$18,$23,$FC,$0B
	.dc.b	$E4,$3F,$8D,$F1,$48,$FE,$0A,$FF,$1E,$F3,$3F,$8F,$FC,$84,$9F,$E1
	.dc.b	$2F,$F2,$2F,$83,$FE,$B7,$F2,$CF,$91,$A4,$AF,$E1,$FC,$1A,$F9,$09
	.dc.b	$1F,$C5,$A9,$1C,$B6,$E7,$F2,$12,$B4,$F6,$F9,$1E,$FF,$21,$27,$93
	.dc.b	$E2,$E1,$F1,$FC,$6C,$9E,$CF,$9B,$E4,$7C,$FE,$D7,$A3,$E3,$E5,$78
	.dc.b	$E3,$D7,$F1,$72,$3C,$CF,$BB,$E3,$BD,$FF,$17,$27,$D4,$FC,$2F,$8D
	.dc.b	$F8,$7F,$15,$F1,$3E,$36,$57,$BE,$3E,$3F,$C0,$C8,$F8,$4F,$CA,$F8
	.dc.b	$CF,$97,$E2,$C9,$F8,$CF,$CE,$F8,$CF,$F0,$4F,$CB,$7F,$B9,$67,$F9
	.dc.b	$FB,$5F,$F6,$1B,$7C,$3F,$E2,$DD,$3F,$AF,$F4,$93,$FA,$EA,$DF,$93
	.dc.b	$D0,$EB,$F1,$B2,$3A,$B5,$EB,$F1,$7E,$22,$3E,$27,$CB,$F1,$BE,$6F
	.dc.b	$5F,$CF,$F1,$52,$BC,$F1,$96,$90,$F4,$FC,$5E,$5F,$5C,$5D,$7F,$98
	.dc.b	$66,$AF,$87,$ED,$4B,$F8,$6F,$C6,$F8,$BC,$47,$C4,$8B,$7C,$A7,$FC
	.dc.b	$67,$F2,$CF,$82,$A4,$BF,$ED,$71,$FD,$83,$80,$EE,$33,$3B,$7F,$DE
	.dc.b	$71,$FD,$99,$7C,$B7,$F8,$F3,$0F,$F6,$A7,$CB,$3A,$7F,$D2,$23,$97
	.dc.b	$FD,$27,$19,$BF,$F6,$D4,$E5,$E2,$FF,$A6,$2F,$A4,$27,$83,$E0,$F7
	.dc.b	$58,$70,$8E,$78,$76,$C3,$C2,$D8,$FF,$74,$61,$D3,$0E,$6D,$AD,$CF
	.dc.b	$33,$D5,$5C,$5F,$05,$E6,$B8,$47,$80,$E1,$5E,$B7,$D0,$FA,$F0,$F3
	.dc.b	$C6,$0F,$1B,$8C,$3D,$B3,$C5,$B8,$78,$6F,$BD,$F8,$8E,$47,$98,$F0
	.dc.b	$C6,$EF,$A5,$F8,$70,$CD,$43,$8D,$F4,$3E,$7B,$FE,$C6,$6B,$C4,$FF
	.dc.b	$60,$38,$7E,$5F,$C5,$3E,$1F,$FD,$89,$1F,$84,$8C,$3C,$1A,$F5,$6B
	.dc.b	$FE,$CB,$F8,$DE,$28,$38,$69,$A7,$07,$D7,$BF,$F1,$BF,$99,$F6,$67
	.dc.b	$94,$E7,$E0,$3F,$CB,$1A,$FE,$2A,$72,$7C,$AB,$E2,$FC,$C3,$31,$F9
	.dc.b	$08,$CB,$C6,$3C,$7F,$11,$9A,$FF,$32,$39,$BF,$53,$9C,$2E,$9E,$FB
	.dc.b	$F0,$5F,$92,$31,$86,$E7,$C4,$7C,$07,$07,$57,$8D,$7B,$DD,$4F,$FA
	.dc.b	$E6,$BA,$47,$FC,$00,$C3,$E1,$56,$DB,$1D,$9B,$7D,$4E,$F7,$E4,$3A
	.dc.b	$57,$1A,$F0,$9C,$4B,$3E,$5C,$3D,$45,$AE,$32,$8E,$D1,$89,$B3,$FE
	.dc.b	$8E,$C3,$D2,$FE,$0C,$F6,$99,$BF,$14,$F9,$A3,$DC,$75,$AE,$F7,$E1
	.dc.b	$8E,$06,$4D,$CF,$17,$B6,$1F,$0A,$BE,$04,$70,$9F,$2B,$FC,$43,$0F
	.dc.b	$2B,$F8,$AB,$F7,$D7,$DD,$6C,$76,$8F,$71,$D1,$E5,$87,$AA,$33,$1B
	.dc.b	$3C,$0F,$33,$D6,$57,$D7,$3E,$4C,$7F,$2B,$3D,$5B,$3C,$CE,$E7,$D7
	.dc.b	$C3,$E0,$3E,$99,$EA,$FC,$AA,$F4,$BF,$10,$FB,$B1,$D7,$0F,$83,$1F
	.dc.b	$50,$EB,$5E,$38,$CE,$3F,$A0,$9F,$3D,$7A,$5F,$1D,$EF,$1F,$85,$1F
	.dc.b	$00,$F4,$63,$EA,$7C,$58,$FA,$1F,$D8,$DF,$A1,$F3,$5F,$A1,$FD,$CC
	.dc.b	$7C,$C7,$8B,$5C,$9D,$CC,$94,$F8,$87,$A0,$F8,$8F,$77,$89,$F0,$1F
	.dc.b	$89,$7E,$A8,$C1,$EA,$C7,$E1,$BF,$91,$3E,$0C,$FC,$3C,$3F,$88,$DF
	.dc.b	$C3,$9F,$AD,$3E,$77,$E7,$4F,$C3,$7C,$2D,$73,$5D,$B0,$F8,$D8,$FE
	.dc.b	$6E,$7C,$58,$F9,$6F,$84,$E6,$7E,$09,$E8,$2C,$69,$FC,$09,$C8,$F2
	.dc.b	$9F,$1C,$F0,$CF,$5B,$F1,$61,$E9,$BE,$D5,$E8,$9C,$BC,$63,$2F,$23
	.dc.b	$A9,$E3,$BF,$E6,$A2,$FF,$85,$B6,$3C,$8D,$7F,$3E,$7F,$AE,$AB,$0F
	.dc.b	$F1,$23,$13,$FD,$62,$7A,$BF,$C3,$D7,$C5,$F5,$7A,$1A,$9F,$EA,$CE
	.dc.b	$A7,$FC,$AF,$F0,$DB,$7F,$AD,$71,$FF,$95,$3E,$2F,$16,$8F,$F2,$BD
	.dc.b	$61,$D5,$72,$3F,$CC,$35,$9E,$85,$58,$E7,$83,$F5,$9F,$89,$7E,$43
	.dc.b	$CA,$BD,$5D,$BA,$CB,$FC,$53,$3A,$9C,$BA,$6B,$CE,$3F,$C5,$F3,$CE
	.dc.b	$3F,$D7,$74,$E1,$F6,$D7,$F8,$D1,$B0,$E4,$B1,$2B,$E8,$3F,$52,$BE
	.dc.b	$5B,$5E,$51,$A3,$BF,$C7,$F1,$8F,$49,$F4,$9F,$9B,$F2,$4F,$C2,$DF
	.dc.b	$0D,$F2,$7A,$F3,$18,$C3,$FD,$1B,$18,$9D,$47,$BE,$B9,$46,$30,$EF
	.dc.b	$0C,$1E,$E7,$D5,$E8,$F0,$C6,$30,$FF,$4C,$C6,$27,$51,$D6,$75,$7F
	.dc.b	$E9,$CA,$FC,$17,$E2,$8F,$99,$F8,$9C,$A8,$C6,$1F,$D8,$EF,$F1,$B1
	.dc.b	$87,$13,$BA,$E3,$39,$E0,$FF,$46,$E3,$DA,$7F,$18,$F8,$3E,$3A,$63
	.dc.b	$F7,$F0,$F0,$7C,$3E,$32,$F9,$F1,$E3,$AC,$63,$FB,$18,$DD,$F1,$C9
	.dc.b	$5E,$CF,$8C,$F1,$92,$7D,$58,$7E,$8E,$93,$98,$FF,$5F,$F1,$71,$AF
	.dc.b	$F6,$1B,$FA,$0F,$D2,$3F,$35,$3F,$B7,$6B,$FB,$46,$BF,$E8,$66,$E4
	.dc.b	$7F,$44,$FA,$70,$7C,$77,$5C,$3E,$9B,$F5,$38,$A4,$7B,$62,$E1,$F1
	.dc.b	$2B,$F1,$8D,$AE,$35,$F0,$DF,$E0,$3E,$03,$F0,$9F,$7C,$FC,$28,$C6
	.dc.b	$9F,$A3,$6C,$66,$77,$7F,$81,$9C,$BF,$EE,$4E,$7E,$B3,$AC,$FF,$D9
	.dc.b	$71,$FB,$4F,$95,$7C,$2B,$D1,$FF,$32,$F7,$7F,$8B,$67,$31,$8C,$7F
	.dc.b	$CE,$5E,$9B,$D9,$FE,$5D,$E2,$7F,$12,$9F,$F3,$B5,$F0,$7F,$CE,$F8
	.dc.b	$73,$C7,$B4,$60,$DB,$8A,$E5,$0C,$3B,$6D,$E1,$86,$1D,$6B,$6D,$C3
	.dc.b	$B6,$5B,$C3,$8D,$F7,$38,$E3,$DE,$B7,$19,$BE,$4F,$BB,$2F,$6C,$EB
	.dc.b	$1E,$F7,$C6,$BD,$38,$F7,$BD,$C6,$0E,$93,$B7,$C3,$3B,$CB,$D9,$8F
	.dc.b	$82,$72,$70,$9D,$34,$98,$78,$A3,$2D,$8C,$E3,$E7,$3F,$CC,$D9,$7C
	.dc.b	$29,$D1,$23,$53,$EF,$C3,$C7,$8F,$9F,$1F,$21,$E5,$C3,$95,$F0,$BF
	.dc.b	$3E,$1C,$DE,$47,$88,$F8,$47,$F1,$5A,$FC,$23,$6F,$F0,$6D,$C7,$FD
	.dc.b	$8F,$2F,$85,$7C,$9E,$1B,$5F,$C1,$A7,$53,$92,$47,$DB,$9D,$E1,$F8
	.dc.b	$66,$E3,$F9,$DE,$7F,$F3,$EB,$F9,$F8,$CD,$7E,$69,$AF,$E4,$1B,$87
	.dc.b	$F1,$A8,$FB,$31,$82,$61,$F7,$2F,$AC,$F3,$8D,$91,$FC,$84,$69,$ED
	.dc.b	$19,$C3,$D0,$F3,$8C,$61,$E7,$8B,$1F,$B0,$AE,$58,$7C,$48,$F8,$53
	.dc.b	$A7,$F5,$33,$AC,$3C,$F1,$A3,$C6,$6E,$B9,$61,$E2,$8C,$9B,$70,$4C
	.dc.b	$3E,$64,$71,$AE,$B3,$FB,$6B,$F0,$61,$93,$D9,$5F,$C0,$F8,$5F,$8F
	.dc.b	$BC,$4F,$F1,$86,$FF,$E2,$8F,$17,$FC,$63,$7F,$D4,$DA,$FF,$80,$2F
	.dc.b	$8D,$70,$84,$61,$FE,$59,$7D,$5B,$3F,$D4,$1D,$C6,$5F,$ED,$3C,$EF
	.dc.b	$8F,$E9,$D9,$FD,$C5,$B1,$99,$FE,$BD,$F2,$38,$99,$1F,$D8,$5D,$C7
	.dc.b	$96,$3F,$C4,$5F,$1B,$87,$89,$FE,$2D,$32,$5E,$74,$79,$E3,$27,$F6
	.dc.b	$88,$CB,$EB,$6B,$EA,$F8,$FF,$4E,$AF,$2E,$9E,$A8,$93,$FD,$CB,$3F
	.dc.b	$F0,$74,$7C,$56,$C6,$73,$4E,$5E,$FD,$5F,$DA,$72,$F8,$A7,$C6,$F9
	.dc.b	$0B,$7F,$E0,$38,$AF,$F6,$78,$F9,$7E,$37,$9B,$E3,$E4,$FF,$82,$63
	.dc.b	$2F,$85,$F0,$11,$FF,$04,$F9,$1F,$E3,$3F,$33,$EA,$FC,$74,$9F,$F1
	.dc.b	$74,$7E,$16,$75,$1F,$E1,$59,$D3,$FE,$3C,$F2,$3F,$CC,$CF,$F9,$07
	.dc.b	$CC,$65,$F8,$0E,$C9,$39,$8E,$93,$A8,$FF,$10,$4E,$9F,$F2,$3C,$65
	.dc.b	$FF,$2B,$8F,$92,$7C,$BF,$E1,$57,$BD,$4C,$44,$9F,$CC,$4F,$06,$C7
	.dc.b	$E1,$63,$11,$FE,$2C,$6F,$07,$FA,$8D,$BD,$89,$91,$87,$1E,$97,$F9
	.dc.b	$36,$BF,$D4,$1F,$AB,$BE,$1F,$F5,$D0,$87,$C4,$E1,$B7,$FE,$87,$9D
	.dc.b	$56,$4D,$54,$9D,$57,$57,$FC,$91,$F2,$3F,$AE,$F2,$24,$60,$87,$F8
	.dc.b	$BD,$CB,$FE,$31,$8F,$D7,$62,$F6,$6F,$93,$2B,$F5,$D3,$96,$CF,$EB
	.dc.b	$63,$11,$FE,$54,$6B,$FE,$6F,$6B,$F2,$43,$C3,$3F,$F6,$83,$75,$D3
	.dc.b	$67,$E8,$7B,$3F,$EA,$3A,$FA,$5E,$CF,$29,$CF,$82,$9C,$DF,$2C,$7C
	.dc.b	$17,$BB,$F0,$C6,$27,$C9,$87,$59,$56,$E3,$0F,$19,$F7,$B6,$B9,$57
	.dc.b	$8A,$FC,$2D,$39,$D7,$8D,$C9,$E2,$2D,$F6,$8F,$1E,$1E,$B9,$E2,$7E
	.dc.b	$3A,$FB,$BC,$4C,$8A,$E4,$F8,$70,$F3,$3E,$0B,$F4,$9D,$B0,$F0,$E1
	.dc.b	$DE,$B7,$5D,$2B,$6F,$37,$83,$9C,$3C,$71,$A8,$F1,$57,$38,$DD,$74
	.dc.b	$23,$82,$4F,$B4,$F3,$57,$BB,$0F,$5C,$74,$C3,$2C,$3E,$0C,$68,$DD
	.dc.b	$78,$CF,$23,$9C,$3C,$93,$DC,$F5,$1F,$0A,$A2,$F5,$BE,$19,$2F,$E0
	.dc.b	$BF,$85,$AE,$73,$F4,$F1,$F6,$BF,$1C,$86,$6B,$F0,$4F,$A2,$78,$E1
	.dc.b	$F9,$1B,$ED,$3E,$57,$B6,$1F,$C4,$5B,$1E,$82,$61,$82,$61,$D9,$F1
	.dc.b	$C7,$C9,$3C,$15,$CE,$7C,$75,$E0,$9D,$3C,$9F,$8D,$5C,$AF,$E5,$5F
	.dc.b	$D1,$8C,$61,$EA,$3F,$31,$7C,$F0,$FE,$0A,$FA,$0E,$98,$7C,$33,$AC
	.dc.b	$7E,$09,$F3,$57,$A2,$3E,$CD,$75,$BF,$45,$7C,$59,$F4,$9F,$50,$E4
	.dc.b	$7E,$07,$1F,$C5,$57,$19,$F8,$F3,$F1,$4F,$06,$3F,$45,$E7,$1D,$4F
	.dc.b	$DD,$63,$E8,$8E,$91,$F8,$6C,$71,$87,$E8,$8E,$F8,$FE,$2A,$FE,$14
	.dc.b	$7F,$21,$7E,$15,$F9,$A3,$85,$F3,$8C,$3F,$35,$F2,$E3,$F3,$1F,$2C
	.dc.b	$7F,$01,$AE,$6F,$E3,$A3,$CA,$FD,$E2,$61,$F2,$EF,$CB,$19,$AF,$2B
	.dc.b	$F3,$0F,$C0,$D7,$89,$F8,$E7,$E6,$B1,$FD,$6D,$7A,$23,$AC,$EE,$FF
	.dc.b	$07,$8F,$AE,$7E,$E9,$0F,$BB,$3B,$3E,$93,$D7,$1F,$D5,$C7,$AA,$BF
	.dc.b	$0F,$87,$D5,$AF,$A8,$7C,$19,$FC,$03,$F0,$70,$F2,$57,$0B,$E7,$18
	.dc.b	$AF,$97,$87,$F9,$6F,$0F,$E8,$18,$7F,$55,$3F,$4C,$57,$2F,$F4,$6F
	.dc.b	$53,$FC,$CE,$EC,$FE,$60,$E1,$E0,$D7,$87,$91,$FE,$53,$5E,$5D,$23
	.dc.b	$FE,$57,$6E,$7F,$D5,$F0,$FF,$29,$6B,$FD,$22,$BF,$C2,$55,$FD,$2A
	.dc.b	$73,$1F,$D0,$8C,$9F,$D1,$1C,$D7,$F6,$18,$CB,$D7,$C4,$5F,$59,$F2
	.dc.b	$9C,$63,$2F,$53,$33,$C6,$78,$5F,$8E,$B7,$3D,$EB,$79,$FF,$80,$70
	.dc.b	$F1,$BF,$26,$F8,$61,$CA,$BC,$0D,$BE,$D1,$E1,$9D,$B4,$D9,$63,$67
	.dc.b	$86,$36,$D8,$DB,$4D,$D7,$8E,$32,$D8,$CC,$7E,$16,$79,$CE,$4F,$66
	.dc.b	$1F,$01,$C4,$FF,$91,$F0,$F7,$BC,$4E,$A5,$BE,$AF,$F8,$96,$FA,$B9
	.dc.b	$AE,$B8,$7C,$36,$C7,$BD,$B5,$EB,$69,$E2,$8F,$F1,$26,$1E,$57,$A1
	.dc.b	$E5,$70,$F8,$4F,$29,$67,$CA,$FE,$22,$7A,$BC,$8E,$B5,$E5,$7A,$35
	.dc.b	$C9,$E7,$32,$53,$33,$E9,$72,$D7,$33,$EB,$C3,$E5,$C7,$96,$3C,$6D
	.dc.b	$38,$BB,$BF,$6C,$7A,$4C,$9E,$99,$E6,$E4,$EE,$66,$38,$BF,$9A,$38
	.dc.b	$9D,$CD,$96,$39,$BB,$32,$75,$73,$87,$DE,$8E,$13,$DA,$31,$1B,$8C
	.dc.b	$3F,$50,$C6,$1F,$36,$B9,$C6,$4F,$03,$8C,$3E,$CC,$62,$B9,$C6,$2F
	.dc.b	$BC,$E6,$FA,$56,$B0,$F9,$B3,$E0,$3E,$C6,$1D,$63,$31,$C4,$CC,$F1
	.dc.b	$7E,$39,$C7,$0F,$B9,$5C,$63,$27,$18,$CC,$F1,$8F,$92,$66,$BE,$E5
	.dc.b	$F9,$89,$1E,$06,$D7,$26,$CF,$09,$D1,$FC,$22,$FF,$8E,$BF,$96,$6B
	.dc.b	$EF,$72,$7F,$2E,$2D,$FF,$24,$77,$1F,$05,$CD,$7F,$01,$3F,$6E,$EB
	.dc.b	$0F,$E3,$2F,$F2,$CF,$11,$E2,$BC,$CE,$97,$FE,$CF,$9C,$F9,$58,$C5
	.dc.b	$1C,$2B,$0E,$AB,$C1,$3A,$BE,$91,$89,$D4,$FF,$A9,$E3,$B4,$EA,$30
	.dc.b	$71,$BF,$06,$0F,$63,$FA,$3D,$FF,$B5,$D7,$87,$A7,$37,$CA,$B8,$4E
	.dc.b	$6B,$4D,$79,$3E,$D9,$FF,$5D,$70,$5A,$37,$8F,$0C,$BF,$7D,$70,$6D
	.dc.b	$2B,$53,$EC,$34,$58,$EB,$3C,$27,$71,$FD,$83,$C3,$77,$BD,$3F,$9F
	.dc.b	$AB,$D8,$FA,$95,$A9,$CC,$62,$B8,$63,$F4,$6F,$83,$F5,$AF,$83,$9A
	.dc.b	$E7,$E2,$3C,$6E,$FD,$58,$6D,$71,$9D,$3D,$67,$4D,$3F,$01,$E4,$BD
	.dc.b	$8F,$C3,$B6,$3C,$9E,$7E,$26,$47,$07,$DF,$18,$E0,$FC,$2F,$2A,$DE
	.dc.b	$87,$43,$AF,$B1,$71,$F0,$C5,$BE,$34,$AC,$71,$7E,$8E,$1E,$B7,$98
	.dc.b	$F3,$FC,$B5,$C7,$D7,$17,$0E,$73,$6F,$F6,$9E,$CF,$E4,$9B,$7F,$8D
	.dc.b	$FD,$2D,$E5,$FE,$72,$B5,$3F,$DD,$DB,$82,$D1,$88,$5F,$2A,$FD,$7D
	.dc.b	$69,$F9,$BC,$F1,$FA,$3E,$D5,$BF,$83,$36,$BE,$04,$DA,$53,$C3,$7B
	.dc.b	$5F,$12,$7F,$7B,$18,$D7,$F5,$1F,$50,$F9,$6F,$E0,$D3,$9A,$DD,$F1
	.dc.b	$C3,$F7,$B3,$97,$F4,$97,$C6,$B7,$87,$EC,$FD,$A7,$98,$F2,$FF,$CA
	.dc.b	$3E,$2F,$9B,$15,$6A,$73,$1E,$AF,$6D,$EE,$7F,$92,$BC,$8F,$E3,$D3
	.dc.b	$EE,$BE,$0F,$F9,$93,$0E,$91,$89,$DC,$F6,$C7,$C4,$EC,$ED,$7D,$6F
	.dc.b	$8C,$EA,$B8,$DF,$39,$CE,$3E,$6C,$18,$F9,$B3,$C5,$EE,$7C,$D8,$F3
	.dc.b	$C3,$8C,$78,$4E,$39,$78,$5D,$E3,$CA,$B8,$46,$0E,$18,$F7,$79,$61
	.dc.b	$DD,$F7,$57,$13,$CD,$3B,$CB,$D7,$7D,$2B,$67,$27,$13,$BC,$7C,$F5
	.dc.b	$C2,$FA,$5F,$8B,$1F,$43,$CA,$74,$F3,$9F,$0C,$EB,$1F,$31,$31,$EF
	.dc.b	$19,$7C,$91,$99,$E3,$97,$83,$0E,$A4,$C7,$F1,$59,$79,$F2,$F8,$D3
	.dc.b	$E2,$7F,$D6,$98,$7E,$19,$B8,$FE,$09,$B8,$FF,$AC,$DF,$C9,$FA,$8E
	.dc.b	$CC,$5A,$31,$3F,$EB,$98,$C4,$EA,$73,$3E,$07,$78,$7E,$29,$B9,$7E
	.dc.b	$59,$B3,$F9,$BE,$BF,$F3,$D4,$E6,$FF,$17,$5A,$AF,$2C,$7E,$3A,$B7
	.dc.b	$1F,$A3,$9D,$61,$E4,$8F,$C2,$C6,$27,$71,$C8,$CD,$76,$3A,$34,$C9
	.dc.b	$30,$FD,$1C,$F2,$BE,$D1,$ED,$AE,$0E,$AA,$E1,$EB,$79,$3F,$B6,$BE
	.dc.b	$58,$7A,$09,$7E,$2A,$D9,$ED,$79,$BF,$1C,$A7,$AE,$FC,$4F,$E9,$2B
	.dc.b	$AC,$EB,$0F,$3E,$1E,$1A,$C6,$1E,$07,$81,$FC,$4B,$81,$EE,$3C,$AF
	.dc.b	$F6,$96,$FF,$CB,$A7,$31,$8A,$FE,$61,$5B,$7F,$C1,$B8,$7F,$90,$AF
	.dc.b	$84,$E6,$7F,$95,$B7,$9B,$FD,$CF,$87,$49,$E3,$5D,$7B,$BF,$B4,$7B
	.dc.b	$97,$B8,$D3,$1E,$E5,$BF,$EF,$31,$5F,$ED,$31,$E7,$7F,$C3,$DE,$DD
	.dc.b	$F2,$8D,$35,$F9,$C7,$F8,$AA,$7F,$C4,$F3,$A6,$CE,$DF,$F1,$5C,$ED
	.dc.b	$FE,$EF,$EE,$F9,$C8,$FF,$83,$3E,$62,$BA,$3F,$C1,$F1,$ED,$8F,$75
	.dc.b	$49,$FF,$09,$BF,$07,$DC,$C4,$7F,$81,$22,$EF,$FE,$03,$9F,$8D,$E5
	.dc.b	$4A,$FF,$0D,$BF,$27,$C0,$C4,$7F,$83,$22,$D7,$F8,$2E,$7E,$E7,$98
	.dc.b	$F2,$FF,$CA,$F7,$FE,$24,$F7,$7E,$91,$A8,$CC,$7F,$98,$A7,$87,$A3
	.dc.b	$A8,$E0,$7F,$8C,$5E,$BE,$83,$3F,$B9,$D5,$2B,$FC,$7D,$39,$6C,$E5
	.dc.b	$FF,$11,$CF,$28,$FB,$9E,$E7,$F8,$5A,$7F,$CB,$AD,$7F,$CB,$4D,$D3
	.dc.b	$FD,$33,$EA,$FF,$73,$9C,$F9,$9C,$88,$FF,$99,$A3,$53,$FE,$A0,$AE
	.dc.b	$5D,$3F,$DE,$DB,$5F,$E1,$26,$AA,$E3,$3F,$E1,$6A,$7B,$4F,$73,$F5
	.dc.b	$51,$8F,$72,$C9,$B3,$30,$FF,$14,$CE,$5B,$39,$7F,$C8,$FE,$E7,$E9
	.dc.b	$FC,$79,$1F,$E5,$C8,$FD,$51,$87,$FC,$9B,$18,$7E,$07,$B9,$FA,$7C
	.dc.b	$CD,$3F,$55,$EE,$7F,$A1,$F3,$FF,$18,$FB,$9E,$FB,$FF,$19,$DF,$07
	.dc.b	$99,$0F,$F4,$65,$71,$53,$C5,$F9,$17,$DD,$E1,$7F,$ED,$0A,$E1,$5B
	.dc.b	$6D,$FF,$8B,$4F,$05,$74,$71,$7F,$E3,$1A,$E1,$DB,$FE,$57,$AD,$1F
	.dc.b	$E4,$AB,$E6,$FA,$AB,$85,$F2,$8F,$7B,$FE,$3B,$9D,$C1,$B1,$E2,$BF
	.dc.b	$63,$4E,$17,$E1,$8F,$1E,$1D,$2F,$C0,$53,$E1,$96,$FB,$9D,$B0,$F2
	.dc.b	$C7,$03,$D6,$78,$EF,$B6,$1E,$0B,$E8,$F8,$E7,$86,$1E,$27,$8E,$1D
	.dc.b	$5D,$E1,$D9,$E9,$7E,$3C,$3A,$C7,$2C,$3E,$33,$CA,$32,$F0,$3C,$B1
	.dc.b	$E1,$7D,$A6,$E7,$A1,$E0,$61,$92,$61,$F3,$2F,$C9,$5C,$F0,$F8,$B8
	.dc.b	$78,$23,$B1,$97,$C7,$1E,$DB,$F1,$D7,$B6,$38,$E1,$E0,$C3,$B5,$79
	.dc.b	$0F,$0E,$1E,$E8,$CC,$60,$FC,$56,$1E,$1C,$3C,$A7,$57,$C2,$66,$3B
	.dc.b	$57,$D5,$79,$1F,$12,$BA,$9F,$54,$91,$E1,$24,$78,$2F,$CD,$3C,$27
	.dc.b	$33,$F4,$63,$8B,$F5,$C8,$66,$3C,$51,$C6,$33,$87,$C4,$7E,$39,$E8
	.dc.b	$3C,$07,$A2,$7E,$7E,$1E,$FB,$F1,$C7,$88,$FA,$4E,$27,$C3,$18,$79
	.dc.b	$4F,$C8,$74,$79,$4D,$BD,$5D,$9E,$33,$44,$8F,$1B,$CA,$BF,$01,$7E
	.dc.b	$D8,$DC,$62,$7E,$3B,$EF,$9D,$47,$95,$F9,$33,$F5,$63,$78,$7D,$CC
	.dc.b	$3C,$51,$B3,$E5,$9F,$07,$0F,$45,$F2,$8F,$2D,$FE,$03,$0F,$13,$F4
	.dc.b	$30,$E8,$FD,$02,$57,$8A,$77,$1E,$33,$E9,$C7,$5B,$F9,$51,$D6,$BF
	.dc.b	$20,$F2,$BF,$2B,$C1,$F3,$3F,$1E,$37,$5F,$1E,$FE,$25,$6A,$BA,$61
	.dc.b	$F0,$30,$F6,$CF,$E6,$E7,$FB,$75,$7F,$3F,$6E,$1F,$CC,$5B,$7F,$E9
	.dc.b	$39,$7B,$5F,$EF,$B9,$FE,$EC,$7F,$2D,$70,$F4,$6B,$FC,$CD,$AF,$F7
	.dc.b	$DF,$2B,$FD,$3C,$83,$86,$9A,$70,$7F,$31,$A7,$B8,$E1,$FE,$69,$5C
	.dc.b	$EB,$53,$98,$FE,$62,$64,$FE,$67,$5C,$AB,$8C,$78,$34,$FE,$C1,$39
	.dc.b	$8E,$26,$73,$F4,$28,$D4,$FF,$67,$8E,$CF,$2B,$F1,$39,$8E,$47,$F8
	.dc.b	$46,$33,$5F,$33,$0F,$4E,$1E,$46,$D7,$89,$B7,$CD,$B5,$C6,$7D,$B7
	.dc.b	$D1,$FA,$06,$2F,$A3,$4E,$85,$BE,$8D,$AE,$91,$EB,$9E,$2D,$9E,$2D
	.dc.b	$38,$96,$78,$9F,$23,$0F,$8E,$E2,$7E,$15,$74,$8D,$5F,$A6,$7E,$7E
	.dc.b	$1E,$E7,$D7,$87,$56,$D7,$0B,$E5,$1D,$0F,$2C,$65,$A6,$67,$CF,$19
	.dc.b	$6C,$66,$3C,$CF,$DF,$BF,$23,$F7,$A3,$81,$AC,$3D,$B1,$E4,$39,$14
	.dc.b	$E5,$7E,$66,$9E,$42,$DF,$89,$B5,$D6,$FC,$31,$F8,$A3,$27,$5B,$F0
	.dc.b	$BF,$20,$FB,$46,$63,$8E,$1F,$2E,$35,$7F,$2D,$B7,$EC,$78,$61,$E7
	.dc.b	$C3,$AC,$7C,$13,$DA,$70,$29,$93,$CA,$64,$A6,$5E,$6F,$AD,$B7,$E5
	.dc.b	$8D,$61,$E8,$7D,$67,$39,$DB,$83,$75,$E7,$9E,$18,$7B,$67,$27,$CF
	.dc.b	$AE,$31,$98,$F1,$4F,$0C,$3E,$14,$77,$8C,$3E,$E3,$77,$DC,$8F,$58
	.dc.b	$DB,$F0,$0D,$99,$8D,$D7,$DF,$BF,$1D,$F8,$2B,$CF,$39,$BE,$75,$A9
	.dc.b	$F1,$61,$F1,$DE,$B5,$BC,$3A,$E1,$F8,$96,$CF,$DC,$3C,$44,$9E,$0D
	.dc.b	$9E,$0D,$38,$46,$89,$1A,$6C,$6A,$30,$47,$04,$3E,$64,$F9,$1A,$77
	.dc.b	$2C,$F4,$6C,$F3,$69,$CC,$F8,$EE,$5A,$64,$F8,$AE,$5A,$66,$BE,$51
	.dc.b	$1F,$C6,$1A,$3B,$90,$EA,$53,$46,$08,$60,$8F,$59,$FC,$83,$67,$F0
	.dc.b	$13,$A3,$F3,$FF,$93,$3E,$0B,$FA,$E6,$33,$D7,$78,$B8,$CF,$FA,$AD
	.dc.b	$ED,$FF,$6F,$1F,$ED,$F9,$DB,$8C,$3F,$B9,$4E,$DF,$EB,$91,$BA,$E9
	.dc.b	$1A,$C2,$E0,$C3,$9C,$F0,$35,$7D,$67,$57,$CA,$58,$77,$9F,$54,$E7
	.dc.b	$F4,$B7,$CC,$76,$FC,$4D,$F0,$F9,$8C,$37,$2E,$D3,$A7,$D9,$92,$F9
	.dc.b	$C6,$AB,$84,$F9,$F1,$5F,$18,$F5,$1C,$70,$EF,$7D,$30,$E7,$3C,$6B
	.dc.b	$85,$FF,$83,$E3,$85,$77,$3C,$7F,$94,$3C,$7F,$84,$D8,$F8,$DF,$A7
	.dc.b	$BE,$33,$35,$27,$E5,$7E,$60,$EE,$C6,$3E,$98,$C4,$F6,$AF,$AB,$18
	.dc.b	$9F,$5E,$3D,$DE,$17,$F4,$FD,$B7,$C0,$FD,$EF,$49,$F6,$0F,$73,$53
	.dc.b	$98,$C6,$9F,$3A,$DF,$07,$9F,$98,$E8,$D9,$FA,$7E,$D3,$C3,$7B,$BF
	.dc.b	$E4,$AD,$93,$98,$C3,$63,$FC,$9E,$D7,$FC,$62,$70,$F2,$DE,$E7,$06
	.dc.b	$C7,$06,$BC,$25,$18,$8F,$F1,$DF,$D7,$BE,$67,$FC,$2D,$97,$F9,$57
	.dc.b	$85,$F6,$0F,$99,$CE,$09,$CC,$62,$18,$73,$7F,$26,$BE,$6B,$8D,$74
	.dc.b	$6D,$E6,$B7,$1C,$FF,$53,$7C,$F7,$5C,$FF,$CA,$B9,$FF,$94,$9B,$7F
	.dc.b	$E6,$BA,$DB,$FA,$26,$BF,$9C,$8D,$BE,$FF,$1B,$B5,$C8,$E0,$D7,$87
	.dc.b	$CD,$7B,$7F,$4D,$7B,$DE,$FD,$1F,$09,$4D,$44,$BF,$D9,$CE,$9B,$39
	.dc.b	$7F,$5F,$5D,$1C,$90,$CD,$6F,$C2,$FD,$75,$F3,$9F,$DB,$B6,$3F,$65
	.dc.b	$5C,$E7,$F7,$7C,$DF,$AD,$7F,$82,$FB,$EF,$92,$FB,$0D,$8F,$E1,$33
	.dc.b	$B7,$8F,$87,$AA,$91,$A8,$FC,$04,$EF,$0F,$E0,$57,$CE,$BF,$85,$B6
	.dc.b	$3F,$83,$4E,$F8,$7F,$D7,$F5,$FE,$C7,$27,$A3,$FE,$BC,$7F,$DD,$A7
	.dc.b	$FB,$06,$BF,$D7,$8D,$3F,$D9,$07,$FB,$55,$FF,$7A,$F0,$7F,$B6,$BC
	.dc.b	$FF,$F6,$A7,$A5,$FE,$C8,$8F,$F7,$2C,$EB,$D3,$FF,$69,$D6,$A3,$AD
	.dc.b	$FF,$B2,$8E,$A7,$92,$FF,$DC,$33,$FE,$F2,$AF,$F6,$6C,$FF,$B5,$CF
	.dc.b	$39,$C6,$34,$F9,$DA,$E6,$FF,$DB,$11,$E1,$8E,$F1,$FE,$F7,$7A,$1F
	.dc.b	$15,$C2,$C3,$FD,$DF,$18,$6D,$79,$27,$57,$F0,$5F,$8C,$2F,$FD,$E8
	.dc.b	$FF,$B4,$DF,$F7,$64,$7B,$AF,$CA,$7D,$6B,$EB,$5E,$B8,$CB,$E6,$8C
	.dc.b	$BE,$D3,$E1,$D7,$B4,$F8,$25,$9E,$31,$E6,$C3,$EA,$3E,$38,$FA,$EF
	.dc.b	$9D,$FC,$39,$F7,$4F,$25,$FB,$63,$EB,$4F,$27,$EC,$DF,$E2,$DF,$95
	.dc.b	$87,$C5,$8F,$3E,$1E,$39,$F3,$E1,$D2,$FF,$07,$39,$9E,$27,$C1,$36
	.dc.b	$7C,$1C,$3C,$55,$F4,$DB,$87,$76,$CF,$13,$99,$F1,$CB,$5D,$EB,$DD
	.dc.b	$1D,$6F,$C7,$1E,$58,$FC,$3D,$F5,$9C,$C6,$CC,$9D,$E7,$9B,$F3,$CF
	.dc.b	$98,$7E,$39,$CE,$1F,$1D,$B8,$7E,$0E,$7D,$E7,$1C,$3F,$12,$FC,$E8
	.dc.b	$D3,$EA,$9F,$34,$FE,$01,$A7,$E0,$2B,$B6,$1F,$B2,$8C,$D7,$69,$CB
	.dc.b	$F8,$33,$F3,$35,$E0,$3F,$2F,$87,$DF,$6C,$7B,$0F,$35,$77,$C3,$CA
	.dc.b	$FB,$2B,$83,$F8,$EA,$E8,$D3,$F1,$F8,$79,$6F,$E6,$4F,$AD,$B3,$D6
	.dc.b	$B9,$C7,$9A,$38,$BE,$93,$F2,$A7,$97,$0F,$C8,$C7,$C0,$8F,$90,$74
	.dc.b	$3E,$16,$0A,$E5,$48,$C4,$FF,$81,$0F,$B8,$7F,$81,$A3,$53,$FE,$0F
	.dc.b	$6C,$6D,$7A,$7F,$DE,$E3,$0D,$79,$35,$12,$38,$F4,$7F,$86,$E3,$3E
	.dc.b	$95,$4D,$23,$B7,$FD,$00,$FF,$AE,$8F,$F4,$7B,$87,$FD,$90,$DB,$FF
	.dc.b	$60,$46,$0D,$93,$2F,$F4,$5B,$72,$FF,$45,$D7,$1F,$4F,$FD,$92,$4D
	.dc.b	$7F,$D6,$15,$FE,$96,$9D,$46,$23,$FD,$25,$7D,$09,$7D,$23,$B1,$E9
	.dc.b	$8C,$B6,$35,$18,$8F,$43,$70,$F5,$93,$0F,$24,$FF,$B0,$4F,$21,$D8
	.dc.b	$E3,$5D,$08,$F7,$21,$A2,$CF,$5B,$F6,$61,$EE,$9E,$C7,$BA,$B9,$BC
	.dc.b	$49,$5C,$5D,$CF,$1C,$3F,$DA,$91,$E6,$7D,$98,$7A,$9C,$C7,$53,$C7
	.dc.b	$19,$BF,$F6,$95,$73,$6C,$FB,$EF,$C1,$7C,$AB,$53,$FE,$D7,$35,$7C
	.dc.b	$DC,$4F,$8B,$0F,$7C,$69,$B1,$98,$F5,$99,$2B,$93,$C0,$E5,$A6,$70
	.dc.b	$F3,$B6,$3C,$31,$F2,$B0,$F4,$BA,$31,$7E,$79,$FF,$74,$61,$EC,$69
	.dc.b	$DE,$78,$3F,$23,$0F,$4C,$72,$6C,$64,$F4,$C6,$5A,$67,$0F,$4C,$65
	.dc.b	$3F,$C2,$6B,$F8,$B6,$0C,$18,$7F,$19,$F4,$7F,$DB,$0F,$FB,$BB,$D3
	.dc.b	$FF,$71,$FA,$7F,$EA,$DB,$FF,$61,$56,$B3,$FF,$71,$9F,$EC,$F9,$FF
	.dc.b	$73,$CE,$AB,$8F,$A7,$FE,$E5,$BE,$13,$FE,$E8,$C3,$FD,$B7,$18,$BE
	.dc.b	$31,$FC,$4B,$0F,$F6,$79,$FE,$D8,$F4,$7F,$DA,$B7,$E8,$8F,$E5,$14
	.dc.b	$F8,$EF,$F7,$0C,$FF,$B5,$EB,$77,$FC,$72,$3F,$DC,$7E,$A6,$A3,$1E
	.dc.b	$9F,$FB,$79,$F0,$57,$0C,$7F,$9A,$4E,$5F,$E4,$B1,$9B,$E7,$5F,$EE
	.dc.b	$9F,$53,$FD,$D7,$5E,$76,$CF,$FB,$A7,$D1,$F9,$55,$ED,$6D,$7F,$41
	.dc.b	$F5,$3F,$DC,$D3,$FE,$DF,$24,$7F,$B7,$1B,$5F,$04,$7A,$7F,$EE,$37
	.dc.b	$B7,$A9,$FE,$E6,$F5,$3F,$DD,$71,$B9,$D5,$7C,$DF,$47,$FD,$CA,$FC
	.dc.b	$53,$FD,$89,$E9,$7F,$B8,$B1,$FE,$A9,$1B,$F4,$FF,$DC,$DE,$A7,$FB
	.dc.b	$8A,$38,$4F,$FB,$1E,$9E,$A6,$63,$89,$97,$D1,$3F,$77,$D1,$FF,$70
	.dc.b	$4D,$F0,$BF,$DD,$2F,$FB,$91,$AF,$FB,$75,$AF,$1F,$47,$FD,$BC,$FF
	.dc.b	$58,$9F,$F7,$0F,$A5,$FE,$E7,$24,$7F,$69,$F4,$7F,$DB,$0F,$93,$C1
	.dc.b	$FF,$6E,$BE,$A6,$BE,$83,$FD,$BF,$E7,$FF,$B8,$67,$FD,$B5,$1C,$7C
	.dc.b	$BB,$C5,$F9,$3B,$FF,$70,$4F,$FB,$78,$FF,$67,$9F,$18,$FF,$6A,$16
	.dc.b	$FF,$DA,$FE,$87,$FB,$6A,$BF,$39,$28,$C6,$BF,$EE,$3F,$0F,$FD,$CC
	.dc.b	$7F,$B6,$6B,$73,$AA,$FF,$73,$7A,$19,$BF,$F7,$57,$A1,$FE,$CD,$7F
	.dc.b	$59,$9F,$FB,$2D,$BE,$86,$6F,$FC,$03,$E7,$FF,$B3,$72,$FF,$69,$CF
	.dc.b	$FB,$86,$7D,$53,$9A,$FD,$B3,$63,$FC,$13,$7D,$5B,$5D,$54,$ED,$BE
	.dc.b	$86,$69,$F1,$FF,$ED,$0F,$43,$FD,$A7,$5F,$EC,$EF,$47,$0F,$FB,$54
	.dc.b	$FF,$6F,$BA,$CF,$FD,$A3,$3A,$9F,$F6,$E9,$B3,$FD,$AE,$FF,$0D,$8F
	.dc.b	$9F,$E6,$FF,$8D,$A7,$FC,$68,$7F,$8D,$8F,$F2,$69,$FE,$69,$F4,$3F
	.dc.b	$C8,$CE,$09,$C3,$FE,$4F,$7F,$C8,$1E,$87,$FA,$09,$FF,$40,$46,$7C
	.dc.b	$FF,$F2,$8C,$7F,$92,$4E,$87,$F9,$26,$3F,$D1,$98,$7F,$9F,$A3,$B6
	.dc.b	$1E,$56,$DF,$5B,$FF,$45,$61,$E4,$AF,$F3,$5D,$7F,$9D,$EF,$FC,$AE
	.dc.b	$FF,$95,$2B,$C2,$DC,$3F,$CB,$2E,$CF,$4C,$F8,$6B,$9D,$7F,$9C,$0E
	.dc.b	$35,$E6,$AF,$F4,$89,$EE,$BF,$16,$1F,$E7,$26,$E1,$E2,$AE,$17,$FE
	.dc.b	$6D,$BF,$5C,$F7,$BE,$11,$F0,$EF,$FD,$40,$F6,$AF,$7B,$93,$FD,$41
	.dc.b	$1D,$9F,$49,$BB,$FF,$38,$1F,$44,$F7,$CF,$AD,$A7,$FA,$0E,$3E,$99
	.dc.b	$F0,$E3,$DF,$3E,$27,$D1,$87,$C4,$AE,$15,$F8,$2A,$F1,$57,$08,$F2
	.dc.b	$9E,$73,$E3,$15,$F5,$BD,$E7,$C6,$FE,$14,$E8,$6D,$F5,$CF,$67,$71
	.dc.b	$F6,$CF,$4C,$FB,$5B,$1E,$6A,$FC,$51,$E9,$8E,$CE,$2F,$E8,$C7,$8C
	.dc.b	$E2,$69,$FC,$84,$F8,$A3,$C8,$7E,$A1,$F9,$CF,$03,$C8,$7A,$8E,$85
	.dc.b	$AF,$33,$63,$CD,$3D,$A3,$A5,$FB,$DE,$93,$F8,$A3,$FD,$68,$FE,$02
	.dc.b	$BE,$AD,$7A,$9E,$B7,$ED,$C3,$F2,$58,$7D,$BC,$38,$D7,$D7,$C3,$EA
	.dc.b	$9E,$6B,$E0,$FE,$30,$FF,$58,$61,$FE,$B4,$6E,$1F,$EB,$18,$C4,$7F
	.dc.b	$86,$67,$FD,$52,$FF,$AC,$39,$7F,$D6,$DE,$87,$FA,$BE,$3F,$D5,$CD
	.dc.b	$7F,$C4,$1E,$87,$FA,$A9,$FF,$60,$1F,$EA,$E3,$FD,$58,$4C,$7F,$CA
	.dc.b	$8F,$F9,$31,$FF,$52,$ED,$FE,$40,$8F,$F1,$07,$A1,$FE,$27,$8C,$47
	.dc.b	$63,$F9,$7E,$9F,$C8,$E7,$4F,$F9,$52,$3F,$D0,$B1,$FE,$5B,$7F,$D0
	.dc.b	$51,$FE,$A5,$9E,$7A,$7F,$92,$F9,$3F,$D9,$31,$DA,$BF,$D8,$E7,$63
	.dc.b	$C7,$1A,$AE,$D5,$C6,$7F,$DF,$87,$0A,$EB,$7E,$48,$FF,$22,$1F,$0E
	.dc.b	$33,$7C,$E3,$38,$D7,$FD,$D9,$5C,$5F,$81,$5C,$1F,$21,$F2,$70,$FF
	.dc.b	$34,$1F,$E9,$68,$D1,$EB,$8F,$04,$7C,$7B,$ED,$1F,$10,$FF,$39,$47
	.dc.b	$13,$4F,$63,$D8,$E2,$7A,$5F,$9F,$0F,$85,$18,$AF,$2C,$62,$BF,$CF
	.dc.b	$51,$8C,$3D,$8F,$43,$E6,$9D,$CF,$39,$8A,$E8,$7B,$27,$9B,$63,$E4
	.dc.b	$9D,$70,$F6,$BC,$8F,$96,$E3,$0F,$45,$F8,$A7,$58,$79,$DE,$77,$D5
	.dc.b	$FA,$38,$7C,$13,$C2,$57,$6D,$76,$7A,$5D,$B4,$D9,$4D,$9E,$63,$E1
	.dc.b	$5F,$9B,$0F,$44,$7B,$4C,$96,$BE,$2D,$78,$63,$E6,$1C,$E3,$2F,$73
	.dc.b	$37,$EF,$7C,$2D,$73,$87,$C5,$C3,$C5,$1E,$0A,$F8,$8E,$B0,$F2,$C6
	.dc.b	$23,$FD,$D8,$7A,$F1,$F0,$BE,$3C,$7E,$1B,$4F,$F6,$A9,$F9,$22,$9F
	.dc.b	$BA,$AF,$F6,$5B,$4F,$F6,$49,$67,$AE,$3E,$98,$FB,$78,$FA,$5F,$F6
	.dc.b	$C1,$D2,$3F,$DC,$2D,$8F,$E0,$CF,$A2,$32,$FF,$B8,$30,$F6,$63,$F0
	.dc.b	$9F,$F6,$BC,$79,$CF,$D9,$14,$FD,$61,$E5,$C7,$F8,$3D,$F6,$AE,$51
	.dc.b	$F0,$AF,$D1,$1F,$EE,$FA,$E3,$1E,$26,$D7,$17,$D6,$7D,$87,$DD,$7E
	.dc.b	$68,$C9,$C3,$13,$FC,$58,$F8,$AF,$D8,$C3,$F1,$11,$83,$EF,$38,$C3
	.dc.b	$C9,$19,$AF,$9F,$7F,$37,$1F,$E7,$73,$DC,$ED,$5E,$E7,$F3,$E7,$DF
	.dc.b	$3D,$18,$78,$5F,$CA,$1F,$46,$BF,$DC,$D5,$FD,$03,$CE,$FF,$7C,$49
	.dc.b	$FF,$78,$78,$DF,$EF,$5F,$42,$61,$FD,$42,$51,$FE,$94,$AF,$F7,$A4
	.dc.b	$62,$75,$39,$9F,$8B,$3F,$EF,$BF,$43,$9C,$7F,$B0,$B5,$FF,$80,$0F
	.dc.b	$FB,$96,$FF,$DD,$07,$FB,$2A,$1E,$87,$FB,$EE,$33,$3B,$3F,$A0,$8F
	.dc.b	$35,$E6,$F8,$63,$0D,$8C,$CF,$F6,$A7,$FD,$F1,$E7,$7F,$BE,$EB,$73
	.dc.b	$E1,$8C,$4E,$BC,$FF,$F6,$BD,$7F,$C0,$0B,$FF,$7D,$DF,$FB,$E7,$5F
	.dc.b	$F7,$6B,$F8,$4C,$9E,$77,$FB,$EA,$FF,$D9,$5E,$7F,$FB,$F9,$2B,$F4
	.dc.b	$36,$FF,$BF,$B6,$7D,$ED,$7F,$B0,$B9,$BF,$F7,$E7,$9F,$E2,$94,$3C
	.dc.b	$FF,$F7,$FC,$7E,$63,$CF,$FF,$7F,$BF,$29,$AF,$F6,$C6,$BF,$EF,$FC
	.dc.b	$3F,$C2,$30,$F5,$7E,$73,$63,$31,$83,$E9,$79,$5F,$EF,$CA,$FF,$7A
	.dc.b	$79,$9F,$EF,$86,$FA,$BF,$E1,$C9,$D4,$1F,$F7,$CB,$F7,$9A,$FF,$80
	.dc.b	$1A,$FF,$C0,$1F,$33,$FC,$4B,$E6,$7F,$BF,$23,$FC,$0A,$D7,$F6,$E7
	.dc.b	$FB,$E8,$DF,$99,$FE,$2C,$51,$FE,$FC,$78,$FA,$99,$8C,$47,$26,$BF
	.dc.b	$EE,$E3,$FD,$F5,$DF,$FE,$F4,$7F,$DF,$E7,$D2,$D3,$C3,$E6,$BC,$DF
	.dc.b	$03,$67,$FC,$7A,$D7,$F3,$AD,$7F,$C3,$EF,$FB,$F5,$34,$C6,$2F,$37
	.dc.b	$FD,$D5,$E6,$FF,$BD,$EB,$C4,$7F,$BF,$3C,$CC,$BF,$97,$C7,$FE,$00
	.dc.b	$BF,$EF,$C8,$FF,$72,$47,$FB,$F1,$F8,$DE,$69,$A4,$9F,$F2,$AC,$7F
	.dc.b	$BE,$C9,$1C,$A3,$F9,$3F,$9B,$FE,$FC,$7F,$DE,$5C,$CB,$FA,$BE,$A3
	.dc.b	$CE,$FF,$7D,$47,$D8,$E5,$79,$DF,$EF,$BC,$3F,$C8,$31,$87,$FC,$CA
	.dc.b	$D7,$F8,$2B,$5D,$C7,$8D,$46,$F0,$F1,$47,$F3,$C8,$C5,$79,$2B,$87
	.dc.b	$99,$FE,$82,$F1,$FF,$CD,$2B,$CD,$FD,$7E,$1F,$E4,$AE,$47,$9D,$F9
	.dc.b	$0F,$33,$FC,$DF,$18,$9F,$F2,$A5,$F1,$D2,$E1,$CA,$F7,$87,$1C,$3F
	.dc.b	$32,$71,$C3,$FC,$AF,$87,$16,$DF,$FA,$9D,$FF,$75,$9F,$EE,$13,$FA
	.dc.b	$BC,$6F,$1F,$4C,$F7,$AE,$B7,$DE,$FF,$AE,$C7,$7C,$37,$8F,$98,$FF
	.dc.b	$7B,$1F,$D8,$B0,$CD,$7C,$1B,$E4,$AF,$91,$FE,$A3,$C3,$C1,$1E,$46
	.dc.b	$BB,$9E,$33,$C2,$35,$7E,$17,$D8,$7B,$AF,$FD,$7F,$1E,$39,$EC,$FF
	.dc.b	$A0,$5E,$53,$D4,$E5,$87,$FA,$FE,$BF,$DE,$27,$FB,$1B,$0F,$95,$8F
	.dc.b	$FB,$96,$BF,$DB,$2F,$E0,$6F,$AC,$7B,$CE,$D8,$FC,$09,$FA,$98,$75
	.dc.b	$AF,$F7,$35,$7F,$B2,$EF,$A9,$FE,$D5,$8F,$F6,$FE,$3F,$32,$7F,$DB
	.dc.b	$51,$FE,$D4,$3B,$D7,$E1,$23,$B9,$EA,$8F,$F8,$02,$7F,$B8,$DF,$B3
	.dc.b	$87,$FB,$9D,$FF,$76,$1F,$EF,$9A,$ED,$8F,$96,$3F,$DD,$A7,$FB,$E8
	.dc.b	$FC,$4D,$F7,$7D,$C7,$E3,$A7,$E9,$4F,$D0,$8F,$39,$E0,$39,$63,$F2
	.dc.b	$70,$F2,$47,$D1,$3E,$15,$78,$A3,$D3,$87,$F3,$6B,$FC,$59,$F8,$CC
	.dc.b	$3C,$44,$9F,$86,$70,$BF,$C0,$1F,$86,$13,$C6,$3F,$3B,$1F,$20,$F5
	.dc.b	$12,$7A,$3E,$93,$11,$D9,$C4,$69,$61,$F7,$A7,$51,$ED,$9D,$4E,$63
	.dc.b	$C9,$1C,$67,$F2,$73,$99,$E7,$8F,$E8,$A7,$F2,$E6,$9F,$D2,$CE,$E3
	.dc.b	$38,$7E,$3E,$72,$71,$7F,$52,$67,$1F,$C3,$5F,$AA,$BE,$1B,$EE,$3C
	.dc.b	$AE,$30,$F4,$DF,$4A,$F9,$11,$87,$9D,$7B,$EB,$81,$F0,$B0,$F3,$61
	.dc.b	$E2,$9D,$C6,$1E,$73,$BC,$3D,$11,$83,$C6,$43,$B1,$E9,$31,$3C,$DF
	.dc.b	$CA,$E1,$F7,$DF,$C1,$63,$F9,$53,$F3,$CF,$E7,$2F,$CE,$F3,$3E,$8E
	.dc.b	$3F,$9C,$8F,$D8,$47,$26,$BA,$3B,$BC,$9C,$57,$85,$F3,$92,$FE,$F1
	.dc.b	$FC,$25,$F6,$4F,$DA,$7E,$2E,$1F,$C1,$2F,$F3,$71,$FB,$83,$F8,$8C
	.dc.b	$65,$FC,$EC,$FE,$40,$E4,$FD,$B3,$83,$FB,$03,$FA,$46,$1F,$C7,$6B
	.dc.b	$FC,$D3,$3E,$B3,$FC,$4A,$53,$F8,$E9,$8B,$79,$AF,$3B,$8C,$61,$B1
	.dc.b	$99,$FF,$3A,$CE,$5F,$F1,$77,$9D,$FE,$2F,$4D,$23,$A3,$FD,$5E,$7F
	.dc.b	$A6,$76,$FF,$6B,$E1,$FE,$B9,$8C,$1F,$ED,$87,$13,$FE,$AC,$24,$72
	.dc.b	$6E,$3F,$EA,$C8,$E2,$FF,$B4,$1F,$F6,$61,$FE,$D9,$AE,$FE,$77,$FB
	.dc.b	$16,$78,$57,$2C,$FF,$C9,$78,$78,$30,$EC,$4A,$DB,$67,$77,$CD,$B8
	.dc.b	$76,$24,$FA,$E7,$4F,$FB,$84,$F3,$CF,$FB,$82,$77,$5E,$9B,$F1,$3E
	.dc.b	$02,$4F,$6A,$E9,$5C,$23,$94,$60,$93,$CE,$FC,$31,$F1,$0F,$F7,$99
	.dc.b	$4F,$F6,$A4,$6E,$BA,$47,$C9,$39,$D7,$69,$D4,$60,$D3,$82,$46,$C9
	.dc.b	$1A,$6B,$E2,$BF,$7C,$7B,$67,$8C,$EC,$86,$EF,$D7,$5C,$23,$E6,$61
	.dc.b	$F6,$DA,$7D,$73,$FD,$75,$3F,$EE,$8B,$FF,$44,$FC,$7B,$DB,$FF,$61
	.dc.b	$E1,$FE,$D6,$CB,$FD,$8B,$82,$B5,$1F,$ED,$BF,$37,$FC,$BC,$FF,$9B
	.dc.b	$E4,$6E,$F9,$CE,$A3,$FC,$EE,$66,$79,$1D,$AB,$AC,$7F,$BA,$E7,$85
	.dc.b	$FF,$96,$A3,$81,$9A,$FF,$78,$C7,$F9,$78,$FF,$78,$9B,$F3,$FF,$DE
	.dc.b	$D1,$E5,$94,$61,$F1,$B6,$78,$7A,$1E,$8F,$43,$AD,$FF,$B8,$70,$46
	.dc.b	$30,$FF,$69,$C6,$27,$B6,$1C,$DF,$F5,$AF,$9F,$FE,$F6,$AF,$F7,$F6
	.dc.b	$3E,$7B,$7A,$DF,$3A,$B5,$3E,$F7,$FD,$ED,$7E,$96,$DF,$28,$FF,$7A
	.dc.b	$7A,$3F,$EF,$CF,$43,$FD,$F3,$3F,$EE,$3F,$8A,$FF,$7E,$F3,$3D,$57
	.dc.b	$CE,$FD,$CF,$39,$5A,$9C,$E5,$F1,$E7,$FD,$D7,$A7,$FA,$2B,$0E,$5E
	.dc.b	$BF,$1F,$63,$83,$F8,$6F,$3B,$FD,$BE,$4F,$07,$FD,$C9,$AF,$FB,$85
	.dc.b	$B1,$FE,$DE,$8F,$C3,$B5,$FF,$73,$7A,$EF,$3F,$FD,$D4,$D8,$FE,$76
	.dc.b	$7F,$B7,$3C,$FF,$F5,$96,$4F,$1D,$18,$6C,$FE,$07,$CF,$AF,$E4,$96
	.dc.b	$7B,$F6,$E4,$EE,$3F,$D9,$6F,$E6,$29,$E6,$7F,$AB,$FC,$CF,$F6,$1C
	.dc.b	$FE,$36,$72,$D9,$D7,$99,$23,$FD,$A4,$43,$A7,$91,$FE,$B9,$E0,$FF
	.dc.b	$5A,$C7,$EA,$BC,$CF,$F6,$43,$F9,$96,$BF,$AA,$8D,$BF,$EC,$0F,$2F
	.dc.b	$FD,$9A,$F1,$F3,$3A,$44,$8E,$51,$A7,$81,$FE,$D8,$E7,$FF,$6B,$C6
	.dc.b	$6B,$AF,$C3,$7F,$66,$AF,$D5,$79,$8C,$7F,$DA,$0D,$8F,$D7,$4E,$DF
	.dc.b	$DF,$B7,$A3,$F7,$53,$B9,$FE,$06,$D8,$FD,$F4,$EF,$1F,$E1,$74,$8C
	.dc.b	$3F,$C3,$AB,$84,$E9,$B1,$FC,$4A,$76,$FF,$82,$3C,$DF,$F3,$97,$91
	.dc.b	$FE,$98,$F2,$FF,$C1,$13,$FE,$B0,$F3,$3F,$CF,$95,$A8,$FF,$44,$9F
	.dc.b	$E7,$C3,$58,$7F,$A7,$5F,$F4,$81,$8C,$BF,$C3,$B5,$FE,$23,$8F,$01
	.dc.b	$89,$E0,$FF,$88,$27,$FD,$0C,$FF,$8C,$CF,$F3,$B5,$F9,$0F,$F1,$F6
	.dc.b	$1F,$E8,$38,$FF,$49,$9D,$0B,$7D,$C9,$7F,$E2,$BA,$FF,$22,$C7,$B2
	.dc.b	$7C,$57,$1F,$41,$D2,$FF,$C5,$67,$FA,$E1,$CD,$72,$BF,$4C,$FF,$8C
	.dc.b	$70,$FF,$59,$61,$E0,$C3,$D7,$87,$97,$0F,$F6,$8D,$F3,$7D,$71,$EF
	.dc.b	$3F,$D5,$0F,$BE,$3F,$CA,$75,$CE,$7C,$31,$F0,$67,$9C,$7D,$07,$9B
	.dc.b	$DE,$3E,$E1,$DC,$E7,$87,$C3,$79,$1A,$8E,$4E,$AF,$D3,$7E,$1B,$EA
	.dc.b	$D9,$D5,$79,$0F,$53,$F2,$1E,$67,$DA,$7C,$4D,$32,$7C,$07,$2D,$32
	.dc.b	$F3,$7F,$0C,$77,$3D,$C7,$E5,$27,$E4,$3D,$4D,$BE,$8B,$F8,$11,$E6
	.dc.b	$C3,$EB,$C6,$23,$99,$FE,$81,$9F,$BD,$87,$E1,$5E,$4F,$C4,$9E,$18
	.dc.b	$7D,$5A,$FF,$35,$42,$7F,$CD,$2D,$8F,$F3,$3B,$5F,$F3,$87,$9B,$FE
	.dc.b	$6B,$6F,$4F,$F9,$B8,$8F,$F9,$9C,$FF,$3D,$F9,$BF,$E6,$A6,$9F,$E6
	.dc.b	$BC,$7F,$C8,$6F,$F8,$79,$AF,$F8,$EC,$FF,$26,$9F,$E2,$83,$FC,$7B
	.dc.b	$9F,$F8,$C1,$A6,$4D,$9F,$E2,$DD,$7F,$C8,$2D,$3F,$C7,$C5,$9F,$F1
	.dc.b	$AB,$67,$A3,$C8,$EA,$64,$ED,$5F,$E3,$57,$CC,$DC,$BF,$C7,$51,$83
	.dc.b	$FC,$70,$5A,$F6,$C6,$89,$E6,$FF,$90,$89,$87,$A2,$BC,$24,$3C,$84
	.dc.b	$BF,$04,$F0,$8F,$69,$B8,$CC,$F3,$8E,$07,$92,$BC,$31,$99,$DC,$F2
	.dc.b	$37,$7E,$8B,$EC,$FC,$88,$EC,$73,$8C,$E1,$F2,$AB,$9D,$6E,$B9,$CF
	.dc.b	$F9,$94,$D3,$C6,$FC,$4F,$FB,$28,$FF,$2F,$BF,$ED,$07,$FC,$AB,$5F
	.dc.b	$ED,$2C,$18,$7F,$98,$E7,$FC,$B3,$E6,$7F,$94,$A7,$FC,$A0,$DB,$FF
	.dc.b	$2E,$C3,$EA,$DF,$47,$FE,$59,$F3,$E5,$7F,$97,$0F,$F2,$85,$F1,$C3
	.dc.b	$FC,$8F,$1A,$7F,$CB,$4E,$A7,$FC,$CE,$FF,$93,$63,$FC,$AF,$E7,$FF
	.dc.b	$9A,$BD,$9C,$79,$FF,$E4,$D9,$FF,$37,$B5,$DD,$7F,$9D,$7D,$0F,$F2
	.dc.b	$B6,$9F,$E6,$48,$FE,$D7,$5B,$38,$79,$DF,$E5,$08,$FF,$2B,$F9,$BF
	.dc.b	$E7,$8F,$3F,$FC,$A8,$C8,$FF,$3F,$5B,$CF,$FF,$28,$46,$1B,$97,$F0
	.dc.b	$38,$FF,$2A,$FB,$7C,$63,$07,$F0,$79,$47,$F9,$4F,$CF,$FF,$29,$7A
	.dc.b	$16,$FF,$D1,$9D,$1F,$E5,$98,$FF,$28,$46,$3D,$0F,$F1,$CB,$FE,$57
	.dc.b	$F3,$FF,$CA,$6F,$F9,$FF,$CF,$FF,$29,$7A,$DF,$E4,$8D,$3F,$CA,$4D
	.dc.b	$8F,$F2,$BF,$9F,$FE,$4D,$9F,$F0,$79,$FE,$52,$F4,$3F,$CA,$24,$7F
	.dc.b	$CA,$3E,$7F,$F9,$2B,$83,$FC,$C9,$E6,$5C,$BF,$D4,$3B,$7F,$94,$FD
	.dc.b	$0F,$F7,$5C,$7F,$86,$AF,$58,$7F,$A9,$9F,$F2,$8B,$EE,$B7,$A1,$FE
	.dc.b	$52,$8F,$F2,$B7,$A1,$FE,$52,$CB,$FD,$45,$5F,$E4,$E9,$FF,$55,$FA
	.dc.b	$1F,$E5,$27,$FC,$AD,$E8,$78,$1F,$F5,$9C,$EE,$BF,$D5,$55,$B6,$FA
	.dc.b	$1F,$E5,$5A,$FF,$29,$9F,$E4,$33,$FC,$94,$7F,$92,$0C,$1F,$E4,$83
	.dc.b	$FD,$77,$E6,$7F,$95,$27,$FD,$6C,$DD,$3F,$C7,$3E,$6F,$F9,$58,$FF
	.dc.b	$2B,$C6,$6B,$A9,$FE,$5B,$F3,$FF,$CA,$CD,$3B,$C7,$F9,$4B,$CA,$FF
	.dc.b	$27,$B2,$3F,$C9,$F1,$EF,$7F,$C8,$CE,$5F,$F2,$6C,$67,$CE,$FF,$21
	.dc.b	$92,$70,$7F,$96,$BE,$23,$97,$9D,$FD,$FE,$FF,$B6,$79,$9F,$E1,$F7
	.dc.b	$FD,$BB,$E6,$FF,$70,$7F,$BC,$1F,$DC,$23,$FC,$0B,$1F,$DA,$23,$FB
	.dc.b	$C4,$F3,$77,$3C,$DF,$F2,$64,$EC,$91,$C8,$91,$E2,$F3,$FF,$CA,$8F
	.dc.b	$F7,$78,$FF,$0D,$BF,$E0,$37,$FC,$16,$7F,$82,$4E,$D3,$E7,$71,$3C
	.dc.b	$AB,$FC,$89,$5F,$E4,$AF,$3F,$FB,$E9,$37,$FF,$00,$12,$3D,$71,$96
	.dc.b	$C6,$70,$F4,$C7,$9C,$FF,$05,$61,$E6,$7D,$07,$F8,$4C,$F9,$6F,$A3
	.dc.b	$0F,$F0,$AD,$73,$8F,$83,$3F,$22,$7D,$17,$FE,$52,$6D,$76,$9F,$01
	.dc.b	$EB,$79,$38,$BF,$F2,$4B,$F5,$CC,$5F,$F9,$0D,$C1,$DC,$E8,$7F,$89
	.dc.b	$B0,$F8,$F3,$CA,$7E,$69,$E9,$9F,$59,$FE,$3A,$BF,$5B,$F1,$89,$3F
	.dc.b	$19,$EC,$79,$CC,$CF,$C0,$3E,$BB,$E2,$AF,$C5,$46,$23,$84,$F3,$8C
	.dc.b	$3E,$5A,$F8,$B3,$FE,$7D,$7C,$71,$B6,$CF,$76,$CF,$C4,$3F,$18,$7A
	.dc.b	$CF,$69,$F1,$A3,$B9,$FE,$B6,$A6,$1F,$E4,$1B,$4E,$63,$FC,$85,$57
	.dc.b	$0F,$F2,$39,$1F,$F7,$ED,$B2,$FF,$23,$47,$56,$A1,$1B,$E8,$FF,$24
	.dc.b	$C7,$F9,$18,$87,$F3,$C1,$E7,$FF,$92,$DC,$35,$FE,$82,$D7,$FD,$A2
	.dc.b	$D7,$FC,$12,$7F,$95,$4E,$25,$38,$94,$E2,$53,$89,$E4,$36,$53,$65
	.dc.b	$36,$76,$30,$7A,$8E,$85,$32,$F9,$DC,$1C,$48,$78,$0A,$78,$0A,$78
	.dc.b	$0D,$10,$D1,$4D,$18,$21,$82,$1E,$23,$81,$4D,$94,$DB,$E9,$23,$EE
	.dc.b	$6B,$EE,$30,$43,$04,$3D,$67,$52,$3C,$9C,$B5,$F2,$B5,$F2,$B5,$F2
	.dc.b	$9D,$CE,$25,$38,$94,$E2,$53,$89,$E5,$36,$53,$65,$36,$76,$30,$7A
	.dc.b	$8E,$85,$32,$F9,$DC,$1C,$48,$78,$0A,$78,$0A,$78,$0D,$10,$D1,$4D
	.dc.b	$18,$21,$82,$1E,$23,$81,$4D,$94,$DB,$E9,$23,$EE,$6B,$EE,$30,$43
	.dc.b	$04,$3D,$67,$52,$3C,$9C,$B5,$F2,$B5,$F2,$B5,$F2,$9D,$CE,$25,$38
	.dc.b	$94,$E2,$53,$89,$E5,$36,$53,$65,$36,$76,$30,$7A,$8E,$85,$32,$F9
	.dc.b	$DC,$1C,$48,$78,$0A,$78,$0A,$78,$0D,$10,$D1,$FC,$C6,$7F,$98,$2F
	.dc.b	$7D,$C7,$FC,$DE,$FF,$9B,$46,$27,$86,$09,$CD,$6E,$79,$36,$B8,$47
	.dc.b	$F3,$EA,$E0,$E8,$E7,$5C,$9A,$64,$FE,$8B,$97,$F5,$07,$D0,$D7,$39
	.dc.b	$7F,$51,$7F,$D0,$38,$FF,$5A,$8C,$47,$F5,$73,$FA,$AC,$F1,$9D,$46
	.dc.b	$27,$6F,$66,$BD,$8D,$10,$EF,$3C,$C9,$1C,$27,$AC,$EA,$31,$3A,$8E
	.dc.b	$47,$94,$D1,$83,$C4,$53,$C4,$68,$87,$12,$1C,$0C,$94,$C9,$C8,$B3
	.dc.b	$E6,$6C,$79,$5E,$26,$CE,$44,$30,$48,$E4,$E5,$AE,$63,$A3,$DA,$7E
	.dc.b	$19,$1F,$74,$F0,$24,$F0,$8E,$8D,$7A,$4F,$17,$CF,$3C,$23,$11,$DC
	.dc.b	$EB,$3C,$67,$51,$89,$DB,$D9,$AF,$63,$44,$3B,$CF,$32,$47,$09,$EB
	.dc.b	$3A,$8C,$4E,$A3,$91,$E5,$34,$60,$F1,$14,$F1,$1A,$21,$C4,$87,$03
	.dc.b	$25,$32,$72,$2C,$F9,$9B,$1E,$57,$89,$B3,$91,$0C,$12,$39,$39,$6B
	.dc.b	$98,$E8,$F6,$9F,$86,$47,$DD,$3C,$09,$3C,$23,$A3,$5E,$93,$C5,$F3
	.dc.b	$CF,$08,$C4,$77,$3A,$CF,$19,$D4,$62,$76,$F6,$6B,$D8,$D1,$0E,$F3
	.dc.b	$CC,$91,$C2,$7A,$CE,$A3,$13,$A8,$E4,$79,$4D,$18,$3C,$45,$3C,$46
	.dc.b	$88,$71,$21,$C0,$C9,$4C,$9C,$8B,$3E,$66,$C7,$95,$F9,$77,$FE,$78
	.dc.b	$7F,$9D,$5F,$F9,$D9,$FE,$80,$7F,$30,$8D,$57,$F4,$C8,$C4,$ED,$FE
	.dc.b	$7A,$60,$FF,$43,$CF,$02,$46,$E3,$B9,$A7,$13,$D6,$3C,$24,$3A,$94
	.dc.b	$E2,$77,$21,$97,$93,$C0,$87,$03,$53,$E4,$7C,$44,$39,$9C,$23,$9B
	.dc.b	$C4,$C4,$71,$76,$64,$91,$C0,$91,$B8,$F0,$CF,$C0,$8C,$47,$48,$C1
	.dc.b	$1F,$21,$C0,$A6,$CA,$78,$23,$9B,$63,$93,$E6,$9F,$1C,$62,$3A,$19
	.dc.b	$9E,$33,$A8,$EA,$D7,$53,$C5,$F1,$18,$37,$3C,$89,$1B,$8F,$21,$A7
	.dc.b	$13,$D6,$3C,$64,$3A,$94,$E2,$77,$21,$97,$93,$C0,$87,$03,$53,$E4
	.dc.b	$7C,$44,$39,$9C,$23,$9B,$C4,$C4,$71,$76,$64,$91,$C0,$91,$B8,$F0
	.dc.b	$CF,$C0,$8C,$47,$48,$C1,$1F,$21,$C0,$A6,$CA,$78,$23,$9B,$63,$93
	.dc.b	$E6,$9F,$1C,$62,$3A,$19,$9E,$33,$A8,$EA,$D7,$53,$C5,$F1,$18,$37
	.dc.b	$3C,$89,$1B,$8F,$21,$A7,$13,$D6,$3C,$64,$3A,$94,$E2,$77,$21,$97
	.dc.b	$93,$C0,$87,$03,$53,$E4,$7C,$44,$39,$9C,$23,$9B,$C4,$C4,$71,$76
	.dc.b	$64,$91,$C0,$91,$B8,$F0,$CF,$C0,$8C,$47,$48,$C1,$1F,$21,$C0,$A6
	.dc.b	$CA,$78,$23,$9B,$63,$93,$E6,$9F,$1C,$62,$3A,$19,$9E,$33,$A8,$EA
	.dc.b	$D7,$53,$C5,$F1,$18,$37,$3C,$89,$1B,$8F,$21,$A7,$13,$D6,$3C,$64
	.dc.b	$3A,$94,$E2,$77,$21,$97,$93,$C0,$87,$03,$53,$E4,$7C,$44,$39,$9C
	.dc.b	$23,$9B,$C4,$C4,$71,$76,$64,$91,$C0,$91,$B8,$F0,$CF,$C0,$8C,$47
	.dc.b	$48,$C1,$1F,$21,$C0,$A6,$CA,$78,$23,$9B,$63,$93,$E6,$9F,$1C,$62
	.dc.b	$3F,$97,$FE,$B8,$20

;--------------------------------------------------------------------------------
;	FDNEG.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fdneg_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FDNEG.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FDNEG.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fdneg_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdneg.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDNEG.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdneg.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDNEG.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fdneg_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$E6,$3A,$29,$F7,$3F,$80,$9C,$C6,$1C,$4E,$E7,$51
	.dc.b	$E6,$9D,$3F,$88,$9D,$3C,$23,$2F,$96,$78,$CE,$A3,$84,$E9,$F0,$CE
	.dc.b	$A3,$11,$DE,$30,$7F,$4A,$FA,$EF,$F3,$4F,$A3,$FC,$76,$51,$8F,$57
	.dc.b	$F9,$DB,$FE,$6E,$C3,$CA,$DC,$3D,$CE,$EB,$A5,$FE,$81,$A7,$90,$EA
	.dc.b	$FA,$DA,$7F,$5C,$7C,$51,$CE,$32,$7A,$4B,$5E,$29,$ED,$1A,$78,$9D
	.dc.b	$67,$CD,$87,$8A,$53,$A7,$C1,$87,$BA,$38,$12,$32,$71,$78,$3E,$3A
	.dc.b	$EA,$FF,$5D,$8F,$1F,$C3,$7A,$4C,$91,$C9,$23,$E2,$46,$5E,$E7,$32
	.dc.b	$B9,$7C,$11,$F0,$27,$E3,$1E,$22,$39,$24,$7D,$18,$CB,$D9,$E3,$1E
	.dc.b	$F8,$EA,$65,$E2,$E0,$8F,$C2,$3E,$1E,$1F,$4E,$3B,$47,$03,$38,$FE
	.dc.b	$19,$EC,$71,$C3,$E8,$4F,$02,$1F,$59,$CB,$4F,$B5,$5D,$DF,$1C,$6A
	.dc.b	$7B,$E3,$AB,$F0,$57,$AC,$87,$AE,$F9,$57,$8E,$38,$54,$7F,$19,$7D
	.dc.b	$4F,$1D,$78,$4E,$2E,$63,$86,$1F,$39,$F4,$1C,$8C,$DF,$C2,$9E,$11
	.dc.b	$F9,$CC,$3C,$EF,$B0,$FC,$84,$78,$23,$37,$E0,$9C,$9D,$0F,$93,$5E
	.dc.b	$B7,$A6,$19,$7C,$78,$7E,$01,$F9,$F5,$CD,$F3,$1C,$09,$1B,$78,$D7
	.dc.b	$E1,$E7,$84,$E5,$F8,$97,$EA,$9F,$14,$FC,$39,$DC,$7E,$31,$DD,$FA
	.dc.b	$AF,$C3,$5B,$BE,$58,$7C,$68,$C5,$6D,$F0,$4F,$C4,$37,$59,$C3,$CD
	.dc.b	$18,$8E,$B1,$8A,$D4,$7E,$13,$0F,$1B,$A3,$15,$D6,$77,$1F,$08,$F9
	.dc.b	$31,$D6,$FC,$73,$F3,$EF,$E0,$9D,$EB,$A1,$C4,$FC,$95,$71,$8F,$88
	.dc.b	$F7,$8D,$C7,$E7,$1F,$A6,$FE,$B7,$0F,$C2,$57,$77,$E3,$47,$7B,$FB
	.dc.b	$0F,$17,$F2,$8F,$92,$BD,$35,$EC,$8D,$1F,$95,$8F,$7E,$1F,$12,$B9
	.dc.b	$DF,$82,$7C,$67,$EB,$2B,$8B,$F9,$28,$F1,$BD,$8F,$5B,$C6,$7C,$6E
	.dc.b	$A3,$E1,$DF,$8E,$35,$5E,$39,$F9,$37,$E3,$8F,$85,$1F,$9D,$8F,$C6
	.dc.b	$BF,$62,$BE,$23,$EA,$C3,$E6,$61,$EA,$8E,$11,$F4,$5E,$07,$B0,$FA
	.dc.b	$D7,$EC,$7F,$68,$78,$A3,$9C,$78,$A3,$85,$F8,$9F,$BC,$F9,$DF,$8D
	.dc.b	$87,$B1,$F8,$C7,$97,$0F,$3B,$D5,$B8,$7E,$32,$3C,$97,$F6,$5F,$8E
	.dc.b	$7C,$C3,$D7,$3F,$25,$A6,$4F,$2E,$1E,$F7,$CA,$FA,$9F,$29,$C2,$BE
	.dc.b	$33,$C2,$7A,$34,$E9,$1F,$B4,$6D,$7E,$44,$F8,$91,$93,$99,$5C,$9B
	.dc.b	$8F,$09,$D6,$37,$19,$8F,$A7,$7F,$25,$FA,$73,$F1,$4F,$C5,$BF,$2C
	.dc.b	$F8,$F1,$F8,$DC,$3E,$94,$72,$3D,$D5,$E2,$7F,$2D,$5E,$06,$C6,$5F
	.dc.b	$9F,$19,$6C,$66,$7A,$BF,$55,$B5,$E1,$C3,$E7,$57,$84,$F7,$57,$18
	.dc.b	$CB,$C5,$B3,$C7,$0F,$C1,$CF,$F0,$66,$D7,$E6,$30,$F8,$2F,$E2,$4F
	.dc.b	$82,$7C,$37,$27,$C7,$9F,$D6,$36,$3F,$26,$71,$77,$1D,$E3,$99,$97
	.dc.b	$89,$98,$DC,$78,$E3,$2D,$3A,$C7,$77,$EC,$C7,$13,$6F,$E5,$27,$DA
	.dc.b	$D3,$C3,$3C,$70,$FD,$75,$7E,$D5,$B7,$F8,$56,$D7,$DE,$3F,$1B,$5F
	.dc.b	$91,$6D,$7E,$34,$FA,$17,$E0,$7A,$61,$FA,$76,$C7,$E8,$AB,$95,$7D
	.dc.b	$B6,$CF,$28,$F3,$BE,$D7,$E0,$B5,$F8,$27,$C5,$29,$EF,$8F,$96,$D8
	.dc.b	$F8,$A7,$89,$FD,$79,$F9,$63,$D6,$53,$C8,$FB,$70,$FC,$54,$7F,$7E
	.dc.b	$A7,$C7,$7F,$9A,$9B,$1F,$E5,$26,$BF,$C8,$1B,$5F,$E8,$76,$D2,$64
	.dc.b	$70,$C7,$FC,$18,$EA,$33,$7D,$CF,$E7,$8F,$FB,$A2,$FF,$D2,$4D,$BF
	.dc.b	$F5,$3B,$FE,$2A,$30,$7F,$AA,$A3,$6F,$FA,$08,$C9,$FE,$B3,$72,$FF
	.dc.b	$9B,$0C,$90,$FF,$7C,$7B,$1D,$A0,$D2,$47,$76,$BD,$5A,$FA,$5A,$F4
	.dc.b	$38,$1E,$73,$B1,$0C,$10,$F4,$19,$29,$93,$E3,$19,$29,$FE,$CC,$F8
	.dc.b	$47,$B5,$F2,$1B,$19,$8C,$11,$C1,$0F,$F6,$D4,$70,$6C,$6D,$A6,$CF
	.dc.b	$C0,$1F,$81,$87,$B7,$E1,$24,$79,$5A,$F6,$6B,$C8,$47,$27,$81,$0E
	.dc.b	$05,$3F,$DD,$1F,$09,$E2,$CF,$F2,$1F,$09,$74,$F2,$EA,$8C,$5F,$FB
	.dc.b	$EA,$EA,$94,$61,$E1,$1F,$98,$3F,$E0,$0C,$E6,$64,$64,$E5,$1C,$4C
	.dc.b	$AB,$FD,$24,$E6,$38,$57,$67,$57,$E8,$9E,$35,$EC,$9D,$DE,$2B,$51
	.dc.b	$82,$1C,$8F,$71,$DB,$E1,$3D,$84,$7C,$3F,$09,$5E,$07,$C3,$8C,$5F
	.dc.b	$EE,$E6,$D7,$CC,$9B,$5F,$C0,$71,$56,$23,$91,$E1,$3B,$94,$EF,$09
	.dc.b	$E0,$F2,$8D,$5F,$F0,$B9,$CD,$7D,$5F,$84,$FA,$DF,$09,$2F,$E5,$C6
	.dc.b	$5F,$9B,$19,$7B,$C6,$5F,$0C,$66,$FF,$8C,$4E,$5F,$E3,$51,$9C,$7F
	.dc.b	$17,$39,$7F,$1B,$19,$AF,$85,$39,$7E,$24,$65,$FC,$74,$7E,$37,$E1
	.dc.b	$3F,$1D,$F0,$92,$7F,$07,$19,$7F,$0B,$19,$7E,$3C,$65,$F9,$31,$9C
	.dc.b	$3F,$98,$4E,$5F,$E6,$51,$9C,$3F,$47,$39,$7F,$4B,$19,$AF,$B5,$39
	.dc.b	$7E,$E4,$65,$FD,$34,$7E,$97,$E1,$24,$7E,$69,$FD,$47,$06,$23,$EE
	.dc.b	$C5,$C3,$FA,$24,$DC,$3F,$6B,$36,$BF,$15,$35,$FD,$A4,$5D,$7F,$82
	.dc.b	$47,$43,$C0,$43,$F1,$73,$B9,$E0,$FE,$FB,$81,$CD,$FC,$94,$95,$FC
	.dc.b	$A5,$AF,$F2,$26,$BF,$C5,$63,$8B,$B2,$1B,$29,$BE,$BF,$01,$1F,$E0
	.dc.b	$D1,$C9,$FD,$7A,$8D,$46,$1F,$44,$E3,$2F,$EE,$75,$BA,$E7,$5A,$69
	.dc.b	$FD,$0D,$FE,$84,$47,$04,$3F,$9F,$B9,$69,$93,$D4,$53,$F9,$FC,$3B
	.dc.b	$3F,$A7,$12,$3F,$A9,$35,$FE,$96,$D7,$E5,$35,$E6,$D7,$FA,$53,$FD
	.dc.b	$67,$B6,$47,$F5,$D6,$BF,$D5,$9A,$FD,$56,$BE,$66,$BF,$D5,$5A,$FF
	.dc.b	$4E,$EE,$FE,$CC,$47,$FB,$53,$5F,$EC,$6D,$7F,$04,$D7,$E1,$B5,$FE
	.dc.b	$C4,$FF,$71,$EF,$91,$FD,$D5,$AF,$F6,$E6,$BF,$8D,$6B,$F3,$9A,$FF
	.dc.b	$6D,$6A,$FB,$BF,$F0,$BB,$FD,$45,$FF,$1A,$6F,$FD,$D1,$B1,$FE,$35
	.dc.b	$76,$FF,$3E,$6E,$7F,$E3,$E7,$71,$D7,$AF,$FB,$2C,$6A,$BA,$75,$62
	.dc.b	$65,$78,$63,$66,$1E,$D1,$87,$47,$9D,$D5,$F8,$67,$53,$C0,$D1,$C0
	.dc.b	$FE,$AB,$F0,$EF,$0F,$FB,$CB,$63,$FB,$93,$5F,$77,$C2,$EA,$64,$69
	.dc.b	$F6,$C6,$23,$AC,$5B,$F8,$93,$99,$FF,$0B,$57,$56,$D7,$F9,$5A,$31
	.dc.b	$1F,$E2,$46,$BF,$E6,$56,$A6,$92,$39,$12,$3F,$C4,$87,$17,$0F,$D9
	.dc.b	$EE,$FA,$A7,$9F,$2F,$04,$E2,$7F,$09,$39,$79,$18,$21,$F0,$0F,$88
	.dc.b	$63,$0F,$C2,$47,$18,$F4,$4F,$57,$C4,$70,$39,$18,$24,$FF,$A2,$AF
	.dc.b	$EA,$37,$C9,$FB,$24,$9F,$CD,$7C,$1D,$7E,$FC,$59,$FC,$B4,$DC,$7E
	.dc.b	$74,$DA,$FC,$F4,$D7,$E7,$45,$D5,$39,$C1,$39,$8C,$36,$BF,$16,$FD
	.dc.b	$D2,$3F,$50,$EA,$59,$E6,$D8,$DB,$F6,$A3,$2D,$8C,$BF,$8C,$6B,$EC
	.dc.b	$3F,$79,$F0,$5F,$B2,$3C,$AE,$08,$7C,$63,$CE,$63,$1F,$DE,$47,$08
	.dc.b	$FE,$00,$F3,$76,$71,$32,$71,$C7,$C4,$E6,$79,$B9,$3F,$39,$5F,$36
	.dc.b	$BA,$46,$EB,$8D,$70,$8E,$A6,$4E,$27,$33,$F8,$E4,$FD,$5F,$3F,$F5
	.dc.b	$84,$8F,$9C,$D7,$E6,$35,$FD,$AB,$5F,$D9,$35,$FD,$93,$FB,$7F,$42
	.dc.b	$47,$D9,$6B,$F5,$DA,$FF,$00,$6B,$FB,$D6,$BF,$BD,$6B,$FB,$DF,$47
	.dc.b	$F8,$21,$1F,$C3,$35,$FC,$23,$5F,$E1,$AD,$7F,$85,$35,$FE,$14,$FF
	.dc.b	$0F,$F4,$A4,$7E,$45,$AF,$E3,$DA,$FF,$18,$6B,$FC,$55,$AF,$F1,$56
	.dc.b	$BF,$C5,$7D,$3F,$E3,$84,$7F,$3C,$D7,$F3,$8D,$7F,$92,$B5,$FE,$44
	.dc.b	$D7,$F9,$13,$FC,$9F,$D4,$91,$FA,$96,$BF,$A7,$6B,$FC,$C1,$AF,$F2
	.dc.b	$D6,$BF,$CB,$5A,$FF,$2D,$F5,$7F,$9A,$11,$FD,$B3,$5F,$DA,$35,$FE
	.dc.b	$7A,$D7,$F9,$D3,$5F,$E7,$4F,$F3,$FF,$5A,$47,$F0,$16,$BF,$BF,$6B
	.dc.b	$FD,$21,$AF,$F4,$56,$BF,$D1,$5A,$FF,$45,$F8,$7F,$E9,$84,$7F,$87
	.dc.b	$35,$FE,$18,$D7,$FA,$AB,$5F,$EA,$4D,$7F,$A9,$3F,$C6,$5F,$F0,$57
	.dc.b	$C1,$7F,$B8,$BE,$21,$B5,$3F,$C0,$31,$27,$FD,$BD,$3A,$6C,$E5,$FF
	.dc.b	$06,$4F,$17,$27,$F8,$55,$C3,$FE,$CE,$9E,$4D,$9E,$5B,$B8,$13,$2B
	.dc.b	$73,$98,$C5,$74,$8F,$F0,$D5,$DB,$E3,$4B,$E3,$18,$6C,$7F,$89,$67
	.dc.b	$83,$D0,$F4,$9E,$C2,$D7,$85,$E9,$7C,$1F,$96,$6A,$FB,$46,$79,$17
	.dc.b	$BC,$3C,$37,$C8,$F8,$C7,$13,$C4,$7C,$53,$25,$26,$1E,$38,$E7,$5D
	.dc.b	$A7,$91,$D4,$DE,$1F,$E4,$5B,$F1,$61,$E5,$6C,$F0,$AE,$4F,$03,$C7
	.dc.b	$87,$7A,$FA,$F7,$EE,$9E,$F8,$78,$B0,$E7,$1C,$4F,$8A,$58,$F8,$26
	.dc.b	$DE,$35,$E3,$69,$E3,$9E,$77,$E1,$8F,$86,$D7,$A4,$F9,$1F,$3D,$F9
	.dc.b	$5F,$16,$1E,$CA,$E0,$49,$D3,$F8,$A9,$F0,$9C,$2F,$C3,$3A,$3E,$54
	.dc.b	$6A,$79,$3F,$8B,$31,$1B,$70,$79,$8F,$61,$A3,$CE,$6A,$39,$BF,$65
	.dc.b	$B5,$E9,$9E,$93,$96,$C6,$63,$99,$92,$B9,$70,$4C,$3E,$93,$E5,$3E
	.dc.b	$01,$E5,$37,$1C,$63,$EC,$E1,$F5,$E7,$85,$F9,$B0,$FD,$FC,$7E,$C8
	.dc.b	$FC,$91,$6B,$F8,$0B,$6B,$F8,$A4,$62,$3E,$C1,$93,$FC,$60,$E5,$FD
	.dc.b	$29,$93,$FC,$CE,$E5,$EE,$7C,$6D,$1E,$0F,$F0,$4C,$3F,$75,$18,$9F
	.dc.b	$73,$54,$6A,$7F,$03,$1F,$C5,$27,$53,$98,$E3,$3D,$72,$FE,$23,$5A
	.dc.b	$7E,$BD,$FF,$00,$8C,$57,$3A,$EA,$A3,$2D,$8C,$C6,$0C,$CF,$D2,$9F
	.dc.b	$E4,$F7,$E3,$7D,$37,$E2,$BE,$27,$96,$3C,$07,$A0,$E3,$1B,$70,$72
	.dc.b	$36,$7D,$E7,$4E,$4F,$31,$C2,$7C,$47,$E6,$AB,$E0,$57,$86,$76,$73
	.dc.b	$3A,$CF,$0A,$F8,$F8,$6F,$0F,$81,$7E,$AB,$F3,$9C,$2B,$6E,$0E,$B7
	.dc.b	$CD,$FD,$1D,$72,$74,$FC,$AB,$ED,$72,$72,$FF,$3D,$9E,$11,$EF,$78
	.dc.b	$3A,$21,$98,$E0,$73,$70,$7F,$34,$8E,$75,$E5,$8F,$D9,$C6,$2F,$E6
	.dc.b	$CF,$F0,$28,$C4,$FB,$CF,$5D,$75,$9E,$58,$7C,$E9,$F1,$C6,$30,$D6
	.dc.b	$1D,$09,$3C,$8D,$B8,$7C,$77,$C1,$F0,$18,$36,$62,$FC,$0F,$B8,$97
	.dc.b	$E9,$9F,$F0,$86,$1F,$AD,$9F,$E0,$8D,$7F,$16,$62,$36,$47,$FD,$61
	.dc.b	$1B,$7F,$26,$6C,$CC,$6D,$FF,$74,$C7,$EA,$B2,$7C,$17,$E4,$5B,$3E
	.dc.b	$A3,$F0,$F8,$FF,$82,$23,$31,$F9,$D6,$E1,$F5,$EF,$A5,$F3,$3E,$35
	.dc.b	$7E,$E7,$C5,$FC,$C1,$1F,$CA,$5B,$E0,$71,$1F,$8B,$9B,$87,$F8,$36
	.dc.b	$B5,$3F,$96,$9C,$E3,$FC,$31,$F5,$5F,$29,$F9,$71,$F3,$88,$FC,$86
	.dc.b	$BF,$11,$AF,$03,$44,$34,$5B,$FE,$4F,$E0,$7E,$CC,$F1,$C7,$F2,$39
	.dc.b	$C3,$FA,$C8,$CB,$F0,$CC,$10,$F8,$07,$B4,$C4,$F6,$70,$68,$FD,$84
	.dc.b	$76,$71,$1F,$CE,$A7,$0F,$18,$CD,$FC,$77,$AB,$FC,$D2,$BC,$B9,$2F
	.dc.b	$8C,$E6,$BC,$37,$D5,$E2,$7F,$41,$BE,$35,$C2,$30,$70,$AF,$03,$E8
	.dc.b	$38,$9F,$0C,$F2,$9C,$0D,$18,$35,$5E,$B2,$47,$47,$C9,$1C,$1C,$BF
	.dc.b	$4C,$F1,$1D,$5C,$BD,$23,$DC,$F1,$6C,$74,$37,$6B,$F7,$57,$13,$A5
	.dc.b	$70,$23,$E1,$3A,$1A,$21,$C4,$FE,$BD,$1D,$27,$B4,$7A,$0C,$9F,$21
	.dc.b	$CB,$F1,$5B,$5E,$69,$FB,$67,$32,$CF,$16,$BE,$43,$A9,$CD,$E0,$F8
	.dc.b	$67,$AD,$7B,$23,$75,$F1,$63,$13,$A9,$F6,$BF,$8B,$7C,$07,$A6,$FA
	.dc.b	$BE,$18,$D3,$6F,$E1,$4E,$AF,$E8,$B6,$FB,$BF,$A1,$AF,$E5,$35,$FC
	.dc.b	$B0,$8F,$F2,$66,$BF,$C8,$9A,$F0,$34,$43,$45,$AF,$F4,$F7,$C5,$3A
	.dc.b	$FF,$BC,$36,$7D,$A7,$F8,$05,$C3,$63,$D0,$FF,$AE,$A3,$33,$D2,$73
	.dc.b	$29,$CE,$1F,$EB,$86,$C7,$CC,$BF,$0D,$48,$F0,$DF,$37,$FC,$01,$5C
	.dc.b	$1F,$AF,$7F,$66,$30,$67,$1F,$B1,$3A,$3C,$D7,$D5,$FA,$87,$47,$17
	.dc.b	$FA,$AA,$E7,$18,$BE,$CD,$AE,$67,$F0,$09,$DB,$CC,$C4,$6D,$F8,$4F
	.dc.b	$EB,$A3,$71,$F2,$CD,$9C,$4C,$1F,$9B,$7E,$8B,$FE,$6F,$6B,$F2,$E3
	.dc.b	$FE,$00,$46,$23,$CA,$7F,$9D,$4F,$09,$4A,$65,$E0,$E0,$ED,$1D,$08
	.dc.b	$F1,$8F,$9F,$ED,$7E,$20,$F7,$38,$8F,$AF,$18,$9D,$46,$0F,$75,$FD
	.dc.b	$D7,$F1,$D3,$A7,$E0,$C6,$4F,$CB,$C6,$5C,$5F,$C1,$9E,$87,$23,$AB
	.dc.b	$F8,$F9,$E0,$D8,$C9,$F9,$98,$CB,$4C,$9F,$99,$72,$D3,$36,$8C,$B6
	.dc.b	$33,$18,$3E,$29,$5C,$9E,$8A,$F6,$B8,$3D,$44,$7D,$EF,$94,$C5,$7A
	.dc.b	$DF,$8E,$D7,$F0,$D3,$F8,$42,$3F,$65,$AF,$E0,$5A,$F0,$34,$43,$45
	.dc.b	$9F,$F0,$D7,$C0,$FF,$8B,$CF,$64,$7F,$8A,$DC,$3E,$DA,$F3,$4F,$08
	.dc.b	$FF,$13,$CF,$CC,$D5,$EF,$FF,$8E,$49,$3F,$E7,$FF,$1D,$86,$23,$FC
	.dc.b	$71,$37,$0C,$C0,$56,$1D,$9F,$E3,$69,$E1,$5B,$8D,$B8,$79,$5F,$91
	.dc.b	$E0,$78,$E3,$4A,$34,$F6,$3F,$D5,$15,$E1,$BF,$F2,$05,$F0,$AD,$C7
	.dc.b	$96,$75,$7F,$C9,$7A,$FF,$76,$4C,$3F,$60,$D7,$F7,$0D,$7F,$82,$35
	.dc.b	$E4,$70,$21,$FB,$FE,$BF,$E0,$E4,$7F,$78,$D7,$46,$08,$60,$87,$F1
	.dc.b	$37,$83,$FC,$57,$AF,$F8,$99,$1F,$E1,$0D,$7F,$88,$35,$E0,$68,$86
	.dc.b	$8A,$7F,$21,$EB,$FE,$3E,$47,$F8,$A3,$5F,$E3,$8D,$78,$1A,$21,$A2
	.dc.b	$9F,$CB,$78,$3F,$95,$91,$FE,$40,$D7,$F9,$43,$5E,$06,$88,$68,$A7
	.dc.b	$F3,$9E,$BF,$E6,$E4,$7F,$96,$35,$FE,$68,$D7,$81,$A2,$1A,$29,$FD
	.dc.b	$17,$AF,$FA,$19,$1F,$E7,$0D,$7F,$A0,$35,$E0,$68,$86,$8A,$7F,$51
	.dc.b	$EB,$FE,$9E,$47,$FA,$23,$5F,$E9,$8D,$78,$1A,$21,$A2,$9F,$D7,$7E
	.dc.b	$37,$FA,$D9,$1F,$EA,$0D,$7F,$AC,$35,$E0,$68,$86,$8A,$7F,$69,$D3
	.dc.b	$FB,$39,$1F,$EB,$8D,$7F,$B2,$35,$E0,$68,$86,$8A,$7F,$0B,$1F,$19
	.dc.b	$FC,$11,$C3,$5F,$E2,$0D,$7F,$8C,$35,$E3,$C3,$FD,$DD,$A7,$F7,$98
	.dc.b	$D1,$23,$4D,$8D,$46,$3E,$0A,$47,$16,$BF,$E0,$59,$46,$23,$CE,$D7
	.dc.b	$CE,$D7,$CE,$D7,$DF,$F0,$7E,$22,$3E,$76,$BD,$DA,$FC,$46,$BF,$11
	.dc.b	$AF,$C4,$7D,$BC,$92,$3E,$13,$5F,$53,$5F,$9E,$D7,$E7,$B5,$F9,$ED
	.dc.b	$7E,$BF,$C2,$7C,$A2,$3F,$3D,$AF,$C7,$6B,$F7,$1A,$FD,$C6,$BF,$71
	.dc.b	$FA,$BD,$72,$3E,$D3,$5F,$A4,$D7,$F0,$ED,$7F,$0E,$D7,$F0,$ED,$7F
	.dc.b	$1F,$F0,$BF,$82,$23,$F8,$76,$BF,$7D,$AF,$E5,$1A,$FE,$51,$AF,$E5
	.dc.b	$1F,$C6,$F6,$C8,$FC,$93,$5F,$C5,$35,$FC,$FB,$5F,$CF,$B5,$FC,$FB
	.dc.b	$5F,$D3,$FC,$37,$E6,$88,$FE,$7D,$AF,$E5,$DA,$FE,$B1,$AF,$EB,$1A
	.dc.b	$FE,$B1,$FD,$2F,$C3,$FF,$7F,$E9,$FF,$02,$6A,$8C,$6F,$FB,$46,$DF
	.dc.b	$F8,$2F,$A5,$C5,$88,$FD,$EC,$DB,$E5,$1F,$E1,$1C,$3A,$CE,$EF,$8C
	.dc.b	$E7,$0E,$CD,$3F,$80,$CF,$33,$86,$16,$3F,$80,$47,$A9,$A4,$3F,$74
	.dc.b	$E5,$A6,$4F,$E2,$2E,$5A,$71,$BF,$F0,$97,$1F,$9C,$F0,$42,$3D,$6F
	.dc.b	$F8,$CE,$34,$FA,$4F,$86,$5C,$7E,$1B,$FE,$22,$8D,$BC,$F0,$F8,$2D
	.dc.b	$3F,$90,$F8,$FF,$2C,$8F,$F9,$07,$96,$BB,$3F,$96,$C6,$30,$FF,$15
	.dc.b	$4D,$AF,$E5,$B3,$6B,$ED,$4F,$27,$F9,$9C,$7F,$28,$23,$FC,$B5,$AF
	.dc.b	$F2,$56,$BC,$0D,$10,$D1,$6B,$FC,$79,$CD,$F8,$13,$C5,$1E,$83,$FC
	.dc.b	$B8,$E5,$F9,$13,$AD,$13,$99,$F8,$19,$2B,$F2,$1F,$15,$27,$E8,$3F
	.dc.b	$8F,$E8,$C4,$70,$8B,$8F,$F9,$3E,$6E,$7F,$D4,$A6,$D7,$E7,$E7,$8B
	.dc.b	$FD,$62,$3F,$A6,$91,$FE,$A4,$D7,$FA,$53,$5E,$06,$88,$68,$B5,$FE
	.dc.b	$61,$E9,$FD,$31,$E2,$8F,$ED,$33,$87,$FD,$05,$19,$C5,$39,$9F,$F3
	.dc.b	$7C,$E7,$CC,$FD,$A1,$23,$FD,$15,$D7,$5F,$97,$17,$0F,$F3,$94,$DC
	.dc.b	$BF,$BA,$CD,$AF,$E0,$53,$C5,$FE,$F7,$1F,$DC,$08,$FF,$75,$6B,$FD
	.dc.b	$B5,$AF,$03,$44,$34,$5A,$FF,$42,$76,$7F,$09,$3C,$51,$FE,$0C,$9C
	.dc.b	$3F,$EA,$58,$CE,$49,$CC,$FF,$A4,$A7,$8F,$C6,$BD,$CF,$E3,$4D,$9F
	.dc.b	$DD,$35,$FF,$4C,$F9,$3A,$99,$1A,$7F,$8E,$46,$23,$F6,$93,$C7,$1F
	.dc.b	$F5,$83,$73,$FF,$57,$70,$AF,$85,$FF,$89,$EB,$C0,$DC,$3F,$C4,$11
	.dc.b	$88,$FF,$5B,$35,$FF,$5F,$35,$34,$91,$C8,$91,$FE,$35,$38,$B8,$7F
	.dc.b	$D6,$FF,$0F,$FC,$DC,$F6,$63,$EA,$9C,$4F,$FB,$4E,$73,$8F,$92,$73
	.dc.b	$3C,$8F,$8C,$7C,$CF,$5F,$E7,$91,$FF,$6E,$78,$75,$F9,$31,$70,$FF
	.dc.b	$62,$4D,$BF,$83,$36,$BF,$AA,$4F,$17,$E5,$47,$C8,$23,$EF,$3E,$2B
	.dc.b	$87,$E1,$62,$8C,$47,$23,$84,$65,$B1,$A7,$C6,$7F,$BD,$7C,$CF,$EB
	.dc.b	$C7,$95,$EE,$64,$87,$B8,$F8,$86,$30,$FF,$71,$47,$1A,$7B,$1F,$96
	.dc.b	$25,$7F,$4F,$D5,$F0,$F8,$8F,$C8,$CD,$C3,$FD,$D9,$37,$3F,$C4,$CD
	.dc.b	$AF,$EF,$93,$C5,$FC,$94,$7E,$40,$8F,$C2,$6B,$F2,$1A,$F0,$34,$43
	.dc.b	$45,$AF,$F7,$F7,$C3,$FF,$80,$CF,$14,$7E,$62,$70,$FB,$63,$38,$FC
	.dc.b	$28,$C1,$0F,$79,$FB,$8F,$7B,$F7,$64,$7E,$37,$C3,$D7,$F6,$B1,$70
	.dc.b	$65,$FB,$4A,$C3,$FA,$E9,$B5,$FE,$21,$9E,$4F,$ED,$E3,$F6,$A4,$7F
	.dc.b	$0E,$D7,$F1,$AD,$78,$1A,$21,$A2,$D7,$91,$FF,$18,$7C,$3F,$F8,$A4
	.dc.b	$F1,$BF,$BE,$9C,$3F,$83,$8C,$E0,$9C,$CF,$F8,$1F,$45,$6B,$DB,$FE
	.dc.b	$42,$49,$FC,$67,$9B,$5F,$E3,$71,$71,$F2,$42,$D5,$9C,$FF,$8A,$DD
	.dc.b	$AF,$F2,$44,$F3,$7F,$8F,$47,$F1,$C2,$3F,$AE,$6B,$FB,$46,$BC,$0D
	.dc.b	$10,$D1,$6F,$CB,$6F,$87,$FF,$28,$1E,$39,$FE,$4B,$58,$7F,$55,$19
	.dc.b	$CD,$39,$9F,$91,$39,$F6,$FF,$A3,$92,$3F,$C8,$1F,$78,$F8,$5C,$47
	.dc.b	$F4,$29,$B9,$7C,$8A,$D4,$FF,$4E,$9C,$E9,$FC,$F1,$FF,$36,$4E,$A7
	.dc.b	$91,$FC,$4A,$3F,$A1,$91,$C1,$0F,$E3,$6E,$5A,$64,$FE,$9C,$E5,$A7
	.dc.b	$AB,$DC,$F0,$9E,$47,$FA,$6C,$61,$D1,$E9,$21,$FE,$53,$94,$EE,$3D
	.dc.b	$27,$F6,$47,$FB,$79,$1F,$93,$EE,$D7,$47,$F7,$28,$F8,$71,$9C,$3F
	.dc.b	$B5,$C5,$C1,$52,$D7,$CE,$3F,$B6,$57,$F6,$A2,$3F,$CF,$1A,$FF,$42
	.dc.b	$6B,$C0,$D1,$0D,$16,$FE,$E7,$A9,$F6,$0F,$24,$7F,$74,$9C,$3F,$86
	.dc.b	$8C,$BF,$DD,$A3,$2F,$C9,$8C,$E3,$F9,$98,$FF,$46,$35,$FF,$40,$36
	.dc.b	$BF,$18,$FF,$AC,$CC,$63,$F1,$25,$18,$7E,$3B,$56,$3A,$7B,$E7,$CC
	.dc.b	$8F,$E7,$A6,$BF,$E0,$97,$C8,$F7,$3B,$C7,$F6,$F6,$9F,$DC,$63,$6F
	.dc.b	$95,$AF,$92,$38,$46,$23,$89,$DE,$32,$47,$A3,$FA,$8D,$A4,$7A,$88
	.dc.b	$7E,$55,$CB,$F2,$A3,$D7,$87,$ED,$67,$2F,$F1,$EA,$F8,$C3,$8B,$E7
	.dc.b	$12,$FF,$80,$FC,$45,$7E,$AC,$5A,$FE,$01,$37,$2F,$81,$36,$BF,$31
	.dc.b	$3C,$5F,$95,$7D,$DF,$91,$3F,$10,$8E,$08,$7D,$37,$2D,$32,$74,$73
	.dc.b	$7F,$C4,$F7,$FE,$1E,$78,$E7,$F0,$F3,$87,$F8,$EC,$66,$BC,$91,$FB
	.dc.b	$F8,$D3,$F0,$0F,$CB,$F3,$FE,$4C,$8F,$F2,$6F,$88,$AF,$E6,$A2,$E1
	.dc.b	$FC,$96,$6E,$5F,$85,$9B,$5F,$BD,$9E,$2F,$C6,$8F,$C3,$91,$FC,$73
	.dc.b	$5F,$CB,$35,$E0,$68,$86,$8B,$5F,$CD,$F9,$7F,$99,$9E,$28,$FD,$5C
	.dc.b	$E1,$FE,$83,$19,$C5,$39,$9F,$E4,$93,$9F,$89,$FD,$B1,$23,$FA,$4F
	.dc.b	$C3,$57,$F7,$B1,$70,$FE,$91,$37,$2F,$D4,$4D,$AF,$E4,$13,$C5,$FC
	.dc.b	$4C,$7E,$A8,$8F,$EC,$5A,$FE,$E5,$AF,$03,$44,$34,$5A,$FE,$B5,$F0
	.dc.b	$DF,$D5,$8F,$14,$7F,$0E,$9C,$3F,$D9,$23,$39,$27,$33,$FD,$1E,$73
	.dc.b	$E2,$7F,$19,$24,$7F,$6C,$EB,$AF,$F2,$08,$B8,$7F,$6B,$9B,$97,$F0
	.dc.b	$B9,$B5,$FD,$0E,$78,$BF,$AC,$8F,$E1,$A4,$7F,$8A,$35,$FE,$38,$D7
	.dc.b	$81,$A2,$1A,$2D,$7F,$7D,$F8,$6F,$EF,$27,$8A,$3F,$98,$CE,$1F,$F0
	.dc.b	$3C,$67,$24,$E6,$7F,$B5,$4E,$7C,$AF,$E7,$84,$8F,$F0,$7F,$B7,$5F
	.dc.b	$E8,$71,$70,$FF,$07,$4D,$CB,$F9,$74,$DA,$FE,$CD,$3C,$5F,$E1,$F1
	.dc.b	$FC,$C0,$8F,$F3,$76,$BF,$CF,$DA,$F0,$34,$43,$45,$AF,$F1,$07,$81
	.dc.b	$FE,$1C,$3C,$51,$FD,$52,$70,$FF,$8A,$A3,$39,$27,$33,$FE,$0D,$9C
	.dc.b	$F9,$FF,$D7,$C9,$1F,$E3,$2F,$85,$AF,$F6,$68,$B8,$7F,$8C,$66,$E5
	.dc.b	$FD,$46,$6D,$7F,$82,$67,$8B,$FC,$CA,$3F,$A9,$91,$FE,$B4,$D7,$FB
	.dc.b	$13,$5E,$06,$88,$68,$B5,$FE,$43,$F7,$7F,$C7,$E7,$8A,$3F,$BB,$CE
	.dc.b	$1F,$F2,$6C,$67,$24,$E6,$7F,$C6,$13,$9C,$3F,$B4,$92,$3F,$B8,$35
	.dc.b	$FE,$F0,$D7,$FA,$7B,$5F,$EE,$8D,$7F,$C0,$EF,$F7,$BF,$8F,$7B,$74
	.dc.b	$93,$FE,$ED,$8C,$B6,$32,$FF,$81,$44,$69,$C3,$5F,$F0,$53,$5F,$F0
	.dc.b	$D3,$5F,$F1,$FB,$5E,$71,$E5,$F6,$E4,$78,$DA,$F6,$6B,$D9,$AF,$66
	.dc.b	$BD,$9A,$FA,$3D,$BF,$84,$47,$E0,$35,$F5,$B5,$F5,$B5,$F5,$B5,$F5
	.dc.b	$BF,$37,$DB,$91,$F2,$DA,$FC,$66,$BF,$19,$AF,$C6,$6B,$F1,$9A,$FD
	.dc.b	$0F,$6F,$ED,$11,$FB,$0D,$7E,$9B,$5F,$A6,$D7,$E9,$B5,$FA,$6F,$E1
	.dc.b	$7D,$AF,$F2,$86,$DF,$EC,$6C,$91,$8C,$FE,$DB,$6F,$FC,$F3,$F7,$BF
	.dc.b	$E4,$9B,$E1,$39,$9F,$C3,$B6,$F9,$C7,$FA,$02,$7F,$DA,$75,$C6,$FA
	.dc.b	$CE,$5A,$A7,$AB,$93,$C2,$75,$BF,$53,$F9,$B6,$BF,$96,$6B,$F9,$66
	.dc.b	$BD,$8C,$1C,$48,$71,$3C,$02,$9F,$0D,$EF,$6C,$6B,$0F,$7B,$E9,$C3
	.dc.b	$D7,$87,$C0,$BE,$07,$A6,$78,$61,$87,$BE,$1D,$1E,$47,$EA,$A3,$E6
	.dc.b	$7C,$1C,$8F,$03,$E3,$F4,$F1,$1C,$0F,$DC,$C6,$30,$FA,$33,$71,$FD
	.dc.b	$BC,$DA,$FB,$B3,$C9,$FD,$EC,$7F,$0B,$23,$FC,$25,$AF,$F0,$46,$BC
	.dc.b	$0D,$10,$D1,$6B,$F0,$7F,$1F,$F2,$CF,$14,$7A,$0F,$86,$E5,$F7,$90
	.dc.b	$F7,$62,$8D,$47,$C1,$CD,$7F,$93,$F8,$E9,$3F,$5D,$FB,$5F,$11,$88
	.dc.b	$E1,$17,$1F,$CB,$CD,$CF,$F8,$EC,$DA,$FD,$1C,$F1,$7F,$91,$C7,$F3
	.dc.b	$12,$3F,$CB,$DA,$FF,$2A,$6B,$C0,$D1,$0D,$16,$BF,$53,$EC,$FE,$44
	.dc.b	$F1,$47,$F3,$69,$C3,$F8,$A8,$CE,$29,$CC,$FC,$89,$CF,$C4,$7E,$DC
	.dc.b	$91,$F9,$26,$BF,$2D,$B7,$FB,$A6,$E3,$FD,$01,$B3,$FC,$19,$E2,$FF
	.dc.b	$46,$7F,$AA,$11,$FE,$A4,$D7,$FA,$6B,$5E,$06,$88,$68,$B3,$FC,$33
	.dc.b	$E3,$FF,$68,$78,$A3,$FA,$CC,$E1,$FD,$6C,$67,$14,$E6,$7F,$17,$39
	.dc.b	$D3,$F8,$D1,$23,$F6,$BE,$ED,$7F,$1D,$17,$0F,$E3,$B3,$71,$FE,$C7
	.dc.b	$36,$BF,$94,$CF,$17,$FB,$4C,$7F,$77,$23,$FD,$D5,$AF,$F7,$06,$BC
	.dc.b	$0D,$10,$D1,$6B,$F9,$7E,$7F,$C5,$CF,$14,$7F,$7C,$9C,$3F,$C4,$23
	.dc.b	$39,$7E,$AE,$30,$43,$DE,$7F,$1F,$3F,$9F,$10,$FE,$33,$F1,$15,$FD
	.dc.b	$91,$6F,$FA,$0C,$5C,$7F,$C0,$F3,$4F,$E9,$B1,$C5,$FF,$06,$9F,$E1
	.dc.b	$B2,$1F,$E1,$92,$9F,$E1,$42,$9C,$0D,$10,$D1,$67,$FA,$9F,$C0,$FF
	.dc.b	$3B,$3C,$51,$FE,$20,$30,$7F,$34,$73,$7F,$D6,$63,$FB,$1C,$69,$F7
	.dc.b	$9F,$D2,$3E,$45,$DB,$FD,$95,$B1,$98,$FE,$95,$C3,$96,$46,$4F,$E8
	.dc.b	$71,$87,$F9,$E4,$F2,$6C,$F2,$78,$1F,$E3,$98,$D2,$20,$72,$E1,$AE
	.dc.b	$DA,$EE,$3F,$C8,$8E,$5A,$E4,$CC,$70,$24,$69,$B1,$FD,$7B,$E0,$FF
	.dc.b	$B5,$1E,$DC,$59,$F8,$EF,$35,$B3,$DB,$3A,$7C,$46,$4F,$E8,$75,$CE
	.dc.b	$BD,$9E,$17,$D1,$24,$FF,$6E,$F8,$4A,$FC,$C8,$B7,$FE,$16,$9B,$5C
	.dc.b	$AB,$EE,$57,$F8,$06,$78,$C7,$CC,$3E,$64,$7B,$C8,$FC,$46,$BF,$55
	.dc.b	$AF,$1C,$51,$88,$F0,$1B,$39,$1F,$E0,$6E,$7F,$F0,$91,$E6,$7E,$C4
	.dc.b	$61,$D1,$E8,$21,$FD,$A2,$34,$FC,$13,$F3,$1E,$47,$E5,$C8,$FF,$85
	.dc.b	$3E,$16,$BF,$92,$8B,$97,$F8,$F2,$6E,$5F,$8B,$9B,$5F,$E4,$89,$E2
	.dc.b	$FB,$E3,$F0,$84,$7F,$12,$D7,$F3,$4D,$78,$1A,$21,$A2,$D7,$F9,$3B
	.dc.b	$B3,$FC,$6C,$78,$A3,$F4,$13,$87,$FC,$5F,$19,$C9,$39,$9F,$F0,$6C
	.dc.b	$E7,$CC,$FD,$D1,$23,$FC,$6F,$DF,$5F,$DA,$C5,$C3,$FC,$BF,$37,$2F
	.dc.b	$D7,$4D,$AF,$F3,$5C,$F1,$7F,$05,$1F,$A7,$23,$FA,$C6,$BF,$BD,$6B
	.dc.b	$C0,$D1,$0D,$16,$BF,$CE,$3E,$07,$F9,$64,$F1,$47,$F0,$39,$C3,$FE
	.dc.b	$52,$8C,$E4,$9C,$CF,$F8,$C2,$73,$E8,$7F,$1D,$24,$7F,$9A,$3E,$FB
	.dc.b	$F6,$CF,$F1,$88,$CB,$4D,$65,$8C,$77,$8A,$B1,$3F,$C6,$2F,$9D,$6A
	.dc.b	$7F,$86,$12,$FF,$88,$B5,$FE,$44,$D7,$F5,$CD,$7F,$90,$35,$FE,$40
	.dc.b	$F1,$8F,$4C,$BE,$1F,$FD,$3C,$DA,$FE,$7D,$5E,$67,$F9,$74,$65,$FF
	.dc.b	$39,$36,$BD,$11,$C0,$F8,$D5,$D0,$E4,$FF,$35,$F5,$FF,$A4,$11,$F9
	.dc.b	$9E,$35,$7F,$A1,$C5,$AF,$3C,$DC,$7F,$9E,$4D,$AF,$2C,$F1,$7F,$88
	.dc.b	$47,$F3,$52,$3F,$CE,$9A,$FF,$4B,$6B,$C0,$D1,$0D,$16,$BE,$4F,$C3
	.dc.b	$FC,$83,$C5,$1E,$D3,$F0,$CE,$5F,$ED,$13,$AC,$53,$99,$F7,$9F,$DB
	.dc.b	$7E,$4B,$FB,$61,$1F,$C8,$7C,$45,$7F,$B3,$C5,$C3,$ED,$CD,$CB,$FB
	.dc.b	$04,$DA,$FB,$33,$C5,$FE,$69,$1F,$D6,$48,$FF,$5E,$6B,$FD,$BD,$AF
	.dc.b	$03,$44,$34,$5A,$FC,$6F,$99,$F8,$C3,$C5,$1F,$DD,$67,$0F,$EA,$E3
	.dc.b	$3B,$27,$33,$F8,$E9,$E3,$F2,$6F,$7F,$FD,$E8,$D9,$FF,$6D,$B5,$FC
	.dc.b	$AF,$C4,$EA,$64,$69,$FC,$FC,$62,$3F,$D9,$93,$C7,$1F,$DE,$B7,$3F
	.dc.b	$DB,$56,$EB,$FB,$24,$F7,$6D,$FF,$82,$23,$11,$FE,$14,$6B,$FE,$17
	.dc.b	$6A,$69,$23,$91,$23,$FC,$2B,$19,$7F,$89,$7C,$4F,$EE,$0F,$56,$1E
	.dc.b	$8A,$C4,$E8,$F8,$04,$3F,$73,$3A,$7E,$0C,$7C,$CF,$83,$91,$F3,$1F
	.dc.b	$E0,$BE,$9E,$23,$E3,$C5,$BF,$E1,$73,$72,$F8,$13,$6B,$F8,$4C,$F1
	.dc.b	$7E,$4C,$7C,$72,$3E,$E6,$BF,$E1,$13,$E3,$38,$78,$9B,$21,$BC,$51
	.dc.b	$CA,$FF,$8F,$FC,$47,$F1,$E3,$C9,$5D,$0F,$E7,$8E,$5F,$80,$43,$F8
	.dc.b	$FB,$97,$E0,$9F,$90,$F8,$7F,$C9,$91,$FE,$8D,$F1,$15,$FC,$6C,$5A
	.dc.b	$FE,$63,$37,$2F,$C3,$4D,$AF,$E5,$F3,$C5,$FC,$7C,$7E,$34,$8F,$E1
	.dc.b	$1A,$FC,$16,$BC,$0D,$10,$D1,$6B,$FA,$1F,$B3,$FD,$08,$F1,$47,$E5
	.dc.b	$67,$0F,$F6,$18,$CE,$49,$CC,$FF,$44,$9C,$E3,$FB,$62,$47,$F6,$99
	.dc.b	$AF,$EC,$22,$DF,$F5,$49,$B8,$FE,$A6,$6D,$7F,$52,$9E,$2F,$EC,$A3
	.dc.b	$F6,$04,$7F,$4E,$D7,$F0,$8D,$78,$1A,$21,$A2,$D7,$F6,$6F,$87,$FE
	.dc.b	$CA,$78,$A3,$F7,$13,$87,$FC,$09,$19,$C5,$39,$9F,$EC,$F3,$9D,$FF
	.dc.b	$8C,$92,$3F,$C1,$9E,$E5,$7F,$89,$C5,$C3,$FB,$BC,$DC,$BF,$86,$4D
	.dc.b	$AF,$EE,$B3,$C5,$FE,$2B,$1F,$C4,$C8,$FF,$0A,$6B,$FA,$76,$BC,$0D
	.dc.b	$10,$D1,$6B,$FC,$13,$F0,$FF,$E0,$83,$C5,$1F,$C6,$E7,$0F,$F8,$9E
	.dc.b	$33,$92,$73,$3F,$E0,$A9,$CF,$2F,$F2,$92,$47,$F2,$06,$BF,$BD,$6B
	.dc.b	$FC,$C9,$AF,$F2,$F6,$BF,$CB,$DF,$E6,$9C,$D2,$3F,$95,$B5,$FE,$10
	.dc.b	$D7,$F9,$EB,$5F,$E7,$4D,$7F,$9D,$35,$FE,$75,$D3,$FD,$04,$8F,$F3
	.dc.b	$86,$BF,$C5,$5A,$FF,$4A,$6B,$FD,$1D,$AF,$F4,$77,$FA,$67,$54,$8F
	.dc.b	$E8,$6D,$7F,$90,$35,$FE,$AA,$D7,$FA,$93,$5F,$EA,$4D,$5D,$2F,$87
	.dc.b	$FF,$21,$92,$3F,$CC,$1F,$05,$5F,$F2,$0C,$5D,$7F,$C9,$73,$77,$FF
	.dc.b	$2E,$4D,$AF,$F2,$4C,$D7,$BC,$5B,$4D,$8C,$36,$7F,$CE,$11,$FE,$56
	.dc.b	$8F,$F7,$FD,$FF,$9C,$BD,$27,$83,$FE,$F8,$6D,$FF,$BF,$1A,$FF,$4B
	.dc.b	$6B,$FE,$DD,$6A,$23,$86,$C7,$52,$3D,$5A,$F5,$7C,$5E,$C4,$8F,$13
	.dc.b	$5F,$13,$5F,$13,$5F,$10,$94,$66,$3C,$AD,$75,$EC,$7C,$4C,$F8,$D2
	.dc.b	$31,$3E,$C6,$BF,$B3,$F8,$6B,$3B,$8F,$85,$8F,$CD,$AC,$9C,$38,$97
	.dc.b	$8A,$DE,$1E,$0A,$C4,$2F,$FD,$4F,$7C,$23,$67,$8C,$FA,$D3,$F6,$48
	.dc.b	$E0,$87,$D9,$72,$D3,$27,$DB,$72,$D3,$B9,$FC,$47,$C1,$F5,$1E,$6A
	.dc.b	$F7,$46,$A9,$3A,$8E,$38,$FC,$57,$D9,$87,$BD,$E1,$3B,$72,$7B,$4A
	.dc.b	$7D,$BF,$2B,$E7,$91,$F0,$7B,$75,$DD,$A7,$18,$FC,$AA,$B8,$7D,$E9
	.dc.b	$B7,$F9,$9C,$55,$CE,$3F,$3D,$3F,$A5,$23,$FA,$56,$BF,$A5,$6B,$C0
	.dc.b	$D1,$0D,$16,$FE,$EF,$C1,$FE,$14,$F1,$C7,$BC,$F9,$8E,$5F,$97,$3A
	.dc.b	$B4,$E6,$7E,$0C,$FE,$53,$D0,$91,$F4,$9F,$C9,$79,$18,$8F,$8F,$17
	.dc.b	$0F,$C7,$CD,$CB,$F4,$73,$6B,$E3,$4F,$17,$F8,$14,$7F,$09,$23,$FC
	.dc.b	$25,$AF,$F0,$96,$BC,$0D,$10,$D1,$6B,$F4,$3F,$03,$FA,$93,$C5,$1F
	.dc.b	$BD,$9C,$3F,$91,$8C,$E8,$9C,$CF,$E2,$B3,$56,$BD,$6F,$DC,$92,$7F
	.dc.b	$35,$F0,$35,$FC,$5C,$5C,$7F,$67,$37,$3F,$E0,$F3,$6B,$F1,$53,$C5
	.dc.b	$FE,$4D,$1F,$CB,$48,$FF,$2D,$6B,$FC,$B5,$AF,$03,$44,$34,$5A,$FE
	.dc.b	$0B,$E8,$7F,$0D,$3C,$51,$FC,$86,$70,$FE,$D6,$33,$9A,$73,$3F,$16
	.dc.b	$73,$3F,$C7,$09,$1F,$BD,$F7,$EB,$FA,$E8,$B7,$FC,$5A,$6E,$3F,$CA
	.dc.b	$A6,$D7,$EB,$67,$8B,$FD,$26,$3F,$A8,$11,$FE,$A0,$D7,$FA,$83,$5E
	.dc.b	$06,$88,$68,$B5,$FC,$A3,$DF,$FE,$60,$78,$A3,$FA,$24,$E1,$FE,$31
	.dc.b	$19,$8F,$E9,$11,$82,$1E,$F3,$F8,$DF,$BD,$FD,$00,$8F,$F2,$6F,$BF
	.dc.b	$FE,$A2,$FF,$12,$8C,$B4,$D1,$8B,$FE,$E5,$17,$1F,$EB,$35,$CA,$73
	.dc.b	$5F,$DC,$EB,$FB,$B1,$1C,$10,$FE,$EC,$E5,$A6,$4F,$EF,$0E,$5A,$7A
	.dc.b	$5F,$EB,$B2,$7C,$91,$FD,$F1,$C9,$1F,$86,$43,$F9,$BC,$71,$7A,$19
	.dc.b	$95,$7F,$76,$F9,$57,$C5,$7F,$6C,$6D,$7F,$50,$6B,$D3,$E0,$35,$32
	.dc.b	$34,$FA,$23,$11,$FD,$22,$78,$DF,$82,$3F,$C2,$D1,$8B,$F9,$F5,$C2
	.dc.b	$7E,$CC,$F8,$1B,$7F,$EE,$B7,$8B,$54,$6D,$F6,$46,$63,$11,$C0,$E4
	.dc.b	$E1,$E3,$1B,$23,$A7,$E9,$FC,$04,$8F,$79,$0F,$9E,$E5,$F8,$11,$F0
	.dc.b	$B0,$F8,$10,$C5,$5B,$9C,$DF,$C3,$AF,$8B,$AF,$D3,$23,$F9,$DD,$EB
	.dc.b	$F3,$62,$D7,$6C,$3F,$86,$4F,$D9,$9C,$E7,$F7,$DF,$C6,$DF,$29,$F9
	.dc.b	$F1,$F1,$C8,$FC,$66,$BF,$AC,$3E,$7B,$87,$89,$B2,$1B,$C1,$1C,$B0
	.dc.b	$F3,$7B,$BF,$9E,$3C,$B5,$F0,$0F,$DD,$39,$7D,$44,$3F,$3A,$E5,$F8
	.dc.b	$47,$E5,$FE,$13,$F3,$44,$7F,$82,$7B,$B5,$FC,$A4,$5A,$F9,$13,$72
	.dc.b	$FD,$24,$DA,$FF,$36,$66,$AE,$51,$F9,$69,$FC,$71,$1F,$C6,$B5,$F8
	.dc.b	$4D,$78,$1A,$21,$A2,$DF,$DB,$E8,$FE,$06,$78,$E3,$F3,$F3,$87,$F9
	.dc.b	$04,$67,$34,$E6,$7F,$81,$CE,$79,$7F,$7C,$48,$FE,$51,$ED,$57,$F7
	.dc.b	$11,$70,$FC,$64,$DC,$FF,$84,$4D,$AF,$8B,$3C,$5F,$DC,$C7,$EC,$48
	.dc.b	$FE,$C1,$AF,$E1,$5A,$F0,$34,$43,$45,$AF,$CF,$7B,$3F,$C9,$8F,$14
	.dc.b	$7F,$02,$9C,$3F,$D0,$E3,$39,$7F,$84,$23,$04,$3D,$E7,$F1,$FF,$85
	.dc.b	$FE,$44,$47,$FA,$67,$85,$5F,$E3,$91,$70,$FD,$84,$DC,$BF,$96,$4D
	.dc.b	$AF,$C4,$CF,$17,$F8,$EC,$7F,$15,$23,$FC,$51,$AF,$EA,$5A,$F0,$34
	.dc.b	$43,$45,$AF,$E1,$FF,$80,$7F,$10,$6D,$F8,$E7,$F9,$3D,$7F,$AA,$AB
	.dc.b	$53,$FE,$B8,$6C,$FF,$36,$9E,$11,$88,$37,$04,$FF,$49,$EF,$95,$FD
	.dc.b	$25,$ED,$E4,$62,$3F,$A2,$45,$C7,$CF,$37,$1F,$EA,$93,$6B,$FD,$B9
	.dc.b	$8A,$B9,$47,$F4,$69,$FE,$7C,$47,$F9,$EB,$5F,$E2,$4D,$78,$1A,$21
	.dc.b	$A2,$DF,$CA,$F2,$BE,$31,$E3,$8F,$E9,$F3,$87,$FD,$6B,$69,$D6,$5F
	.dc.b	$E9,$A9,$CC,$6D,$F8,$27,$F6,$DF,$8A,$FE,$E0,$47,$9F,$7D,$7F,$B5
	.dc.b	$45,$C7,$EE,$CD,$CF,$FB,$D4,$DA,$F8,$F3,$C5,$FE,$D7,$1F,$D8,$C8
	.dc.b	$FF,$62,$6B,$FC,$DD,$AF,$03,$44,$34,$5A,$FC,$7F,$AB,$F8,$B3,$C5
	.dc.b	$1F,$DD,$27,$0F,$C6,$8C,$E2,$9C,$CF,$FB,$3B,$25,$72,$F9,$77,$93
	.dc.b	$FF,$00,$1B,$5F,$EE,$46,$BC,$7D,$5D,$4C,$8D,$3F,$9F,$8C,$47,$FB
	.dc.b	$72,$78,$E5,$FD,$D9,$BA,$7E,$8E,$B7,$5F,$E4,$C9,$EE,$DB,$FF,$09
	.dc.b	$46,$20,$47,$0D,$8D,$91,$FF,$0E,$46,$93,$49,$1B,$8F,$8D,$EA,$7E
	.dc.b	$E4,$F5,$E1,$E9,$9C,$4F,$E5,$27,$38,$79,$23,$AC,$E9,$E6,$7C,$63
	.dc.b	$E6,$7C,$4F,$CF,$23,$F9,$EF,$4E,$BF,$26,$2E,$1F,$C4,$26,$DF,$7C
	.dc.b	$55,$8B,$FD,$DD,$72,$7E,$64,$7C,$A2,$3F,$11,$AF,$F8,$46,$79,$B9
	.dc.b	$21,$98,$DB,$D0,$CD,$FF,$24,$F4,$BF,$90,$1E,$38,$F9,$F3,$87,$F7
	.dc.b	$71,$9A,$EF,$1F,$95,$8D,$3F,$04,$FC,$95,$FE,$54,$8F,$F0,$4F,$62
	.dc.b	$BF,$8F,$8B,$87,$F3,$69,$B8,$FC,$79,$B5,$FC,$82,$78,$BF,$91,$8F
	.dc.b	$C7,$91,$FC,$4B,$5F,$8C,$D7,$81,$A2,$1A,$2D,$7F,$48,$CB,$FA,$21
	.dc.b	$E2,$8F,$C9,$CE,$1F,$E3,$F1,$9C,$53,$99,$FD,$D4,$E7,$D7,$FD,$C1
	.dc.b	$23,$F4,$B7,$FE,$64,$FC,$05,$E8,$66,$7F,$69,$56,$B5,$37,$3F,$C6
	.dc.b	$4D,$AF,$D7,$DF,$27,$F8,$24,$7E,$D0,$8F,$EB,$9A,$FE,$2D,$AF,$EF
	.dc.b	$5A,$F6,$38,$10,$E0,$7A,$A7,$F5,$92,$F3,$FF,$5D,$87,$B5,$F3,$57
	.dc.b	$EF,$2F,$37,$C2,$6C,$FA,$9D,$E6,$AC,$CB,$0E,$CF,$84,$E0,$7C,$03
	.dc.b	$F9,$27,$0F,$F2,$72,$3B,$39,$73,$E1,$FE,$45,$17,$2F,$A9,$37,$2F
	.dc.b	$DD,$4D,$AF,$9F,$3C,$5F,$E4,$91,$FC,$88,$8F,$F1,$C6,$BF,$B9,$6B
	.dc.b	$C0,$D1,$0D,$16,$BF,$0D,$E7,$FC,$2C,$0F,$8E,$7F,$93,$D6,$62,$47
	.dc.b	$C2,$20,$AE,$6F,$C3,$AE,$6E,$A3,$FA,$17,$3F,$F3,$C2,$47,$F3,$66
	.dc.b	$BF,$C4,$DA,$FF,$45,$6B,$FC,$85,$AF,$F4,$37,$FA,$3F,$B7,$23,$FA
	.dc.b	$0B,$5F,$E3,$CD,$7F,$A8,$35,$FE,$58,$D7,$FA,$73,$5F,$E9,$3D,$3F
	.dc.b	$D5,$08,$FF,$4E,$6B,$FC,$AD,$AF,$F5,$D6,$BF,$CE,$5A,$FF,$5B,$7F
	.dc.b	$AF,$FA,$D2,$3F,$AC,$B5,$FE,$6C,$D7,$FB,$43,$5F,$E8,$8D,$7F,$B3
	.dc.b	$35,$FE,$C3,$D7,$FD,$B0,$8F,$F6,$66,$BF,$D0,$DA,$FF,$75,$6B,$FD
	.dc.b	$45,$AF,$F7,$37,$FB,$BF,$AD,$23,$FB,$8B,$5F,$E9,$CD,$7F,$C0,$0D
	.dc.b	$7F,$AE,$35,$FE,$FC,$D5,$F3,$2F,$80,$FD,$CB,$63,$F6,$4D,$7F,$29
	.dc.b	$C1,$FC,$2B,$D4,$DD,$49,$DB,$FA,$18,$C4,$70,$3F,$C1,$F5,$CF,$55
	.dc.b	$FF,$7F,$8C,$3D,$67,$99,$FE,$FB,$3C,$F7,$5A,$66,$7C,$CF,$F8,$24
	.dc.b	$C9,$4A,$65,$DB,$E1,$8E,$91,$88,$F3,$C6,$08,$F7,$7F,$7D,$E1,$C8
	.dc.b	$F3,$9E,$F3,$C3,$7F,$D2,$63,$51,$9A,$F7,$BD,$8F,$DF,$46,$A7,$E2
	.dc.b	$C6,$4E,$1E,$1F,$D6,$23,$FC,$17,$D3,$AE,$8F,$B3,$1F,$C4,$E3,$33
	.dc.b	$D2,$7F,$03,$38,$C7,$FA,$2D,$72,$9F,$8D,$3F,$10,$8F,$CF,$6B,$F3
	.dc.b	$70,$46,$23,$89,$B2,$1B,$78,$DF,$F6,$DF,$4F,$F9,$41,$E3,$9F,$C3
	.dc.b	$CE,$1F,$E4,$51,$97,$D2,$60,$87,$C1,$3E,$77,$C1,$7E,$70,$8F,$F2
	.dc.b	$BF,$36,$BA,$3F,$3F,$1F,$CE,$63,$37,$F1,$F0,$4E,$32,$FE,$D7,$5C
	.dc.b	$AF,$F1,$93,$F8,$A2,$3F,$95,$6B,$F1,$DA,$F0,$34,$43,$45,$BF,$F0
	.dc.b	$97,$A9,$FD,$34,$F1,$47,$EB,$27,$0F,$F4,$78,$CB,$FB,$E8,$CB,$F9
	.dc.b	$5E,$54,$EB,$17,$99,$FC,$08,$95,$FD,$CF,$E2,$3F,$59,$8F,$C3,$9C
	.dc.b	$CF,$E6,$DF,$F0,$0C,$EB,$3F,$DE,$46,$AB,$85,$73,$38,$9F,$B3,$8F
	.dc.b	$D8,$91,$FD,$E3,$5F,$C7,$35,$E0,$68,$86,$8B,$1F,$9F,$A7,$A9,$E6
	.dc.b	$3C,$97,$E6,$AF,$74,$65,$FE,$43,$7A,$7F,$D4,$F5,$C9,$D9,$A3,$F9
	.dc.b	$3F,$C3,$FF,$2A,$23,$F2,$7D,$5A,$FF,$01,$8B,$8F,$96,$6E,$3F,$C8
	.dc.b	$E6,$D7,$82,$78,$BF,$C6,$63,$F8,$B9,$1F,$E4,$4D,$7F,$68,$D7,$81
	.dc.b	$A2,$1A,$2D,$7C,$7F,$53,$ED,$9E,$28,$FE,$8F,$38,$7F,$09,$19,$CD
	.dc.b	$39,$9F,$E4,$58,$AB,$47,$F4,$E2,$47,$E3,$7D,$4A,$FF,$25,$8B,$87
	.dc.b	$D9,$9B,$97,$F4,$79,$B5,$F5,$27,$8B,$FC,$F6,$3F,$9D,$91,$FE,$8C
	.dc.b	$D7,$F8,$C3,$5E,$06,$88,$68,$B5,$F8,$BF,$53,$F3,$E7,$8A,$3F,$B5
	.dc.b	$CE,$1F,$D4,$46,$72,$F8,$51,$82,$1E,$F3,$FB,$7F,$C4,$7F,$72,$23
	.dc.b	$FB,$1D,$6B,$FD,$26,$2E,$1F,$9D,$9B,$97,$F6,$B9,$B5,$F9,$89,$E2
	.dc.b	$FF,$61,$8F,$EB,$E4,$7F,$B5,$35,$FE,$78,$D7,$81,$A2,$1A,$2D,$7E
	.dc.b	$BF,$D3,$FE,$06,$78,$A3,$FC,$1F,$38,$7F,$86,$46,$70,$4E,$67,$F1
	.dc.b	$33,$97,$FC,$2A,$43,$F8,$A7,$A7,$5F,$ED,$85,$BF,$E0,$11,$71,$FF
	.dc.b	$07,$4D,$3F,$77,$1C,$5F,$F0,$21,$FE,$02,$21,$FE,$0D,$29,$FD,$7C
	.dc.b	$A7,$03,$44,$34,$59,$FE,$25,$CD,$FC,$98,$F1,$47,$F8,$C8,$C1,$FC
	.dc.b	$BD,$CE,$3F,$CA,$E3,$F8,$8C,$69,$F7,$9F,$EA,$7E,$E7,$C3,$FF,$AC
	.dc.b	$9B,$19,$8F,$E6,$BE,$A6,$59,$19,$3F,$9F,$46,$1F,$F0,$D4,$F2,$6C
	.dc.b	$F2,$78,$1F,$E2,$D8,$D3,$FE,$31,$6A,$23,$86,$C7,$F8,$EE,$31,$06
	.dc.b	$92,$34,$48,$FF,$22,$46,$5F,$F6,$47,$E0,$5F,$E1,$DC,$7F,$C3,$CD
	.dc.b	$CF,$C5,$3D,$6F,$DD,$5A,$9F,$F6,$CD,$6E,$BC,$A7,$BA,$74,$F6,$9E
	.dc.b	$27,$4B,$75,$7D,$12,$4F,$93,$C0,$AF,$CC,$8B,$7E,$69,$B7,$CE,$6D
	.dc.b	$7C,$9C,$15,$CA,$3E,$74,$FC,$92,$3F,$11,$AF,$D5,$6B,$C0,$D1,$0D
	.dc.b	$16,$FE,$4E,$1F,$10,$F1,$C7,$D5,$9C,$3F,$42,$32,$7C,$27,$04,$3E
	.dc.b	$01,$F9,$7F,$87,$FC,$B1,$1F,$B3,$E2,$57,$F2,$31,$6F,$ED,$CD,$CB
	.dc.b	$E8,$4D,$AF,$8D,$3C,$5F,$C9,$47,$E3,$48,$FE,$21,$AF,$E6,$5A,$F0
	.dc.b	$34,$43,$45,$AF,$C6,$F8,$DF,$88,$3C,$51,$F9,$99,$C3,$F9,$48,$CE
	.dc.b	$09,$CC,$FF,$A1,$F3,$56,$BC,$2F,$DD,$12,$7F,$3B,$E9,$57,$F6,$B1
	.dc.b	$71,$FC,$FC,$DC,$FF,$29,$36,$BF,$15,$3C,$5F,$DB,$47,$EC,$48,$FE
	.dc.b	$B1,$AF,$EF,$5A,$F0,$34,$43,$45,$AF,$D8,$FA,$5F,$AC,$3C,$51,$FB
	.dc.b	$D9,$C3,$FB,$88,$CE,$69,$CC,$FC,$59,$CF,$C3,$FF,$1D,$24,$7F,$00
	.dc.b	$F4,$2B,$FC,$62,$2E,$1F,$C0,$A6,$E5,$FB,$79,$B5,$FA,$D9,$E2,$FF
	.dc.b	$19,$8F,$E2,$84,$7F,$87,$B5,$FE,$40,$D7,$81,$A2,$1A,$2D,$7F,$14
	.dc.b	$F4,$BF,$87,$9E,$28,$FE,$41,$38,$7F,$8D,$C6,$72,$4E,$67,$F1,$13
	.dc.b	$9F,$3F,$FA,$01,$23,$F9,$1F,$A5,$5F,$E7,$71,$70,$FE,$4B,$37,$2F
	.dc.b	$E3,$53,$6B,$F8,$84,$F1,$7F,$9E,$47,$F3,$72,$3F,$CC,$9A,$FF,$43
	.dc.b	$6B,$C0,$D1,$0D,$16,$BF,$9B,$FA,$5F,$CC,$8F,$14,$7F,$43,$9C,$3F
	.dc.b	$CF,$A3,$39,$27,$33,$FA,$B9,$CF,$C3,$7F,$63,$24,$7F,$53,$BF,$F7
	.dc.b	$17,$E0,$6F,$53,$33,$FD,$82,$AD,$6A,$6E,$7F,$CF,$E6,$D7,$F5,$DB
	.dc.b	$E4,$FF,$73,$8F,$EB,$84,$7F,$AB,$B5,$FE,$D0,$D7,$FB,$43,$5E,$C7
	.dc.b	$02,$1C,$0F,$54,$FF,$5C,$9F,$85,$18,$3C,$98,$F9,$9F,$ED,$95,$FD
	.dc.b	$8B,$0F,$83,$36,$F9,$3B,$C9,$59,$BE,$67,$51,$87,$81,$F9,$06,$7E
	.dc.b	$71,$A7,$18,$46,$1D,$46,$65,$E9,$EE,$B5,$4C,$D9,$FF,$7A,$9C,$C6
	.dc.b	$27,$FC,$1D,$18,$9F,$89,$5C,$F0,$95,$E0,$3D,$58,$78,$1C,$1F,$E0
	.dc.b	$63,$9C,$57,$C0,$D7,$25,$BF,$3B,$C5,$AF,$03,$51,$C5,$F8,$6D,$8D
	.dc.b	$C7,$43,$C0,$E1,$D9,$E8,$F4,$FE,$41,$EC,$7E,$36,$1D,$DA,$7E,$56
	.dc.b	$36,$FB,$B1,$F9,$0F,$43,$67,$DC,$8D,$B8,$AF,$98,$FC,$EF,$87,$FA
	.dc.b	$C4,$7E,$5F,$25,$78,$45,$BF,$D7,$CD,$C3,$ED,$4D,$AF,$D5,$4F,$17
	.dc.b	$E5,$47,$C5,$23,$F3,$DA,$FC,$66,$BC,$0D,$10,$D1,$6B,$F8,$0F,$A1
	.dc.b	$F9,$13,$C5,$1F,$96,$9C,$3F,$86,$8C,$E0,$9C,$CF,$C8,$CD,$5A,$E5
	.dc.b	$FC,$D1,$27,$F1,$FE,$85,$7E,$64,$5C,$7F,$89,$CD,$CF,$F3,$D3,$6B
	.dc.b	$F8,$74,$F1,$7F,$1D,$1F,$89,$23,$F9,$36,$BF,$8A,$6B,$C0,$D1,$0D
	.dc.b	$16,$BF,$92,$F6,$7E,$D4,$F1,$47,$EE,$A7,$0F,$EA,$A3,$39,$A7,$33
	.dc.b	$F1,$67,$3D,$1F,$9F,$24,$7E,$A5,$AF,$E8,$1A,$FE,$AD,$AF,$F0,$16
	.dc.b	$BF,$C0,$1F,$D5,$7C,$3C,$8F,$DA,$35,$FD,$5B,$5F,$DB,$35,$FE,$18
	.dc.b	$D7,$F8,$5B,$5F,$E3,$3D,$5F,$B7,$23,$FC,$05,$AF,$EE,$1A,$FF,$03
	.dc.b	$6B,$FC,$65,$AF,$F1,$87,$F8,$17,$6F,$F2,$12,$47,$F1,$6F,$46,$BF
	.dc.b	$B3,$8B,$9F,$F5,$A9,$BA,$FF,$25,$9B,$5F,$D5,$67,$8B,$FC,$56,$3B
	.dc.b	$BF,$C6,$5A,$FF,$10,$6B,$B3,$44,$34,$53,$47,$E8,$FC,$FF,$E7,$87
	.dc.b	$89,$FE,$83,$18,$7F,$98,$C6,$75,$4E,$67,$91,$FD,$2D,$FE,$95,$F8
	.dc.b	$23,$E3,$A9,$27,$FA,$A4,$3D,$1C,$47,$F1,$C8,$E0,$EA,$79,$CE,$B3
	.dc.b	$FE,$47,$5A,$AE,$15,$A7,$81,$0C,$9E,$0A,$E5,$9F,$FB,$3A,$B9,$5F
	.dc.b	$4B,$E5,$7C,$09,$19,$79,$1E,$AF,$82,$F7,$F9,$D2,$30,$43,$D8,$E5
	.dc.b	$F7,$47,$C6,$31,$3E,$C9,$C4,$F1,$8C,$3F,$22,$32,$F2,$30,$43,$04
	.dc.b	$8E,$2F,$88,$E9,$87,$FB,$BB,$E7,$9E,$B7,$F8,$05,$B3,$FE,$09,$6B
	.dc.b	$F5,$7D,$1D,$4C,$8D,$3F,$80,$8C,$47,$F5,$69,$E3,$7F,$D6,$9B,$7F
	.dc.b	$46,$B7,$5F,$DE,$E7,$BB,$6F,$FB,$94,$62,$3D,$11,$82,$3F,$DF,$E3
	.dc.b	$48,$38,$6C,$70,$7F,$C1,$11,$93,$CA,$F1,$CD,$E8,$FE,$70,$F6,$E3
	.dc.b	$EB,$BC,$57,$E7,$27,$38,$F9,$A3,$04,$3E,$24,$FE,$9B,$E3,$64,$7E
	.dc.b	$91,$FD,$4F,$A1,$88,$FD,$24,$5C,$3F,$43,$36,$FF,$35,$36,$BD,$13
	.dc.b	$56,$29,$C2,$9B,$18,$6C,$F8,$CF,$DB,$BF,$BC,$F8,$6F,$DE,$F9,$F2
	.dc.b	$7F,$77,$19,$7F,$7B,$19,$7F,$79,$19,$7F,$7D,$19,$C7,$F7,$73,$97
	.dc.b	$F7,$B1,$9C,$7F,$6D,$39,$7F,$71,$19,$AF,$AD,$39,$7E,$C4,$65,$F8
	.dc.b	$31,$F9,$DB,$91,$F9,$D7,$F8,$BF,$9F,$88,$FC,$C4,$5C,$3F,$87,$CD
	.dc.b	$C3,$F5,$13,$6B,$F0,$D3,$C5,$FC,$D4,$7E,$60,$8F,$E5,$5A,$FE,$41
	.dc.b	$AF,$03,$44,$34,$5A,$FF,$21,$F8,$3F,$E4,$46,$DF,$8E,$7F,$3F,$5F
	.dc.b	$E7,$BA,$D4,$FE,$8E,$51,$87,$F4,$DB,$7E,$0D,$F8,$87,$FA,$26,$34
	.dc.b	$F0,$9E,$D0,$F6,$BB,$12,$78,$78,$D5,$FF,$2B,$CE,$DE,$93,$E5,$9D
	.dc.b	$63,$EC,$8B,$7E,$7B,$56,$2F,$F7,$91,$CD,$EB,$5C,$DE,$A4,$32,$79
	.dc.b	$0C,$5F,$76,$DF,$7B,$F9,$1F,$19,$F2,$7A,$E5,$7C,$38,$CB,$F1,$63
	.dc.b	$2F,$BA,$32,$FC,$08,$CD,$FD,$89,$CB,$F6,$A3,$37,$EF,$9C,$BF,$06
	.dc.b	$33,$5F,$02,$72,$FC,$28,$CB,$F7,$23,$F9,$D7,$AD,$23,$F9,$D3,$F8
	.dc.b	$3E,$EC,$47,$F3,$58,$B8,$7E,$32,$6E,$1F,$D1,$66,$D7,$D0,$9E,$2F
	.dc.b	$F3,$78,$FE,$6A,$47,$F9,$93,$5F,$E5,$AD,$78,$1A,$21,$A2,$D7,$E7
	.dc.b	$BC,$0F,$CD,$1E,$28,$FE,$77,$38,$7F,$2D,$19,$D5,$39,$9F,$C8,$CF
	.dc.b	$1D,$7F,$D0,$91,$F9,$DF,$13,$F4,$C7,$E9,$FC,$CB,$19,$21,$F9,$A7
	.dc.b	$2F,$E7,$22,$E3,$FB,$49,$B9,$FE,$66,$6D,$7E,$62,$6B,$FB,$28,$B6
	.dc.b	$9B,$18,$7F,$75,$F1,$3F,$BB,$F3,$65,$FE,$D6,$32,$FE,$DE,$32,$FE
	.dc.b	$C2,$32,$FE,$CA,$33,$8F,$F0,$A9,$CB,$FC,$32,$33,$8F,$EC,$67,$2F
	.dc.b	$EC,$E3,$35,$FB,$29,$CB,$FB,$48,$CB,$FC,$3A,$3F,$C1,$8C,$3F,$C1
	.dc.b	$87,$F1,$3A,$FF,$6E,$FE,$0A,$F3,$B5,$3F,$E0,$92,$CE,$A2,$E1,$FE
	.dc.b	$19,$9A,$7F,$7F,$AE,$4F,$63,$FC,$08,$7F,$82,$88,$60,$87,$F8,$08
	.dc.b	$C9,$4C,$9F,$E1,$53,$25,$3F,$C1,$D5,$EA,$F3,$BD,$95,$FC,$BE,$91
	.dc.b	$E6,$AF,$F0,$99,$A3,$D0,$D5,$1C,$1F,$51,$F0,$CE,$F3,$D6,$78,$4E
	.dc.b	$E7,$42,$BE,$6F,$9F,$A2,$47,$23,$E4,$6B,$87,$86,$1E,$99,$C5,$7D
	.dc.b	$6A,$B3,$9A,$E4,$7A,$2B,$C7,$3A,$8F,$A9,$15,$F6,$C5,$B4,$D8,$C3
	.dc.b	$67,$AC,$BE,$73,$DF,$E5,$C9,$CC,$7C,$23,$0F,$A8,$FC,$24,$65,$D9
	.dc.b	$D1,$DE,$3F,$92,$9D,$4F,$13,$E9,$CE,$B1,$FB,$93,$AA,$FC,$CC,$E6
	.dc.b	$BF,$25,$39,$AF,$C1,$4E,$67,$EE,$CF,$DF,$FA,$07,$C4,$52,$4F,$C0
	.dc.b	$F3,$6B,$E3,$9D,$BF,$41,$AF,$9F,$24,$F0,$C9,$97,$F8,$06,$F7,$18
	.dc.b	$BF,$F3,$34,$62,$7F,$31,$5E,$09,$FF,$04,$09,$EE,$E1,$AF,$5C,$36
	.dc.b	$D2,$9C,$5A,$D3,$31,$C9,$F1,$C6,$A3,$11,$B3,$AB,$83,$D0,$FF,$01
	.dc.b	$F3,$7F,$46,$7C,$1B,$F1,$47,$C5,$8C,$BF,$26,$B4,$F1,$3F,$1F,$18
	.dc.b	$7F,$67,$EB,$FE,$B8,$8F,$E5,$3C,$CA,$FD,$18,$B9,$7F,$1C,$9B,$87
	.dc.b	$F0,$39,$B5,$FC,$5E,$6B,$D8,$FC,$FC,$7E,$F2,$33,$9A,$2C,$61,$B3
	.dc.b	$FB,$2F,$C1,$BF,$DB,$77,$FE,$3F,$7F,$DD,$D1,$9F,$2F,$F8,$74,$6A
	.dc.b	$3F,$D4,$CD,$4F,$07,$F4,$4D,$C7,$F1,$11,$AF,$2F,$99,$D2,$A4,$78
	.dc.b	$E3,$2F,$84,$87,$47,$2F,$68,$E2,$E4,$E4,$70,$9F,$0D,$78,$EF,$DD
	.dc.b	$7F,$AB,$E2,$97,$FA,$57,$DF,$EF,$62,$3F,$55,$17,$0E,$73,$6F,$F4
	.dc.b	$D3,$6B,$E2,$CF,$17,$F4,$11,$F9,$52,$3F,$9A,$6B,$F9,$66,$BC,$0D
	.dc.b	$10,$D1,$6B,$DB,$E5,$FD,$63,$C5,$1F,$B1,$9C,$3F,$7E,$33,$C1,$FE
	.dc.b	$7C,$8C,$10,$F7,$9F,$85,$F7,$BF,$02,$47,$F1,$1E,$5D,$7E,$FC,$5C
	.dc.b	$3E,$7C,$DC,$BF,$19,$36,$BF,$13,$35,$FC,$4C,$5D,$3F,$DD,$31,$E1
	.dc.b	$21,$D8,$F0,$98,$BF,$BB,$3C,$6B,$F2,$FE,$4F,$E6,$3C,$69,$7F,$9F
	.dc.b	$8C,$BF,$A1,$8C,$BF,$98,$8C,$BF,$9A,$8C,$DF,$E2,$A7,$2F,$E3,$23
	.dc.b	$38,$FE,$A6,$72,$FE,$AE,$33,$5F,$A8,$9C,$BF,$AA,$8C,$BF,$AA,$8F
	.dc.b	$E7,$DE,$B4,$8F,$E7,$0F,$EC,$BD,$CC,$47,$F3,$D8,$B8,$7E,$7A,$6E
	.dc.b	$1F,$CE,$66,$D7,$EE,$27,$8B,$FC,$CE,$3F,$94,$11,$FE,$5A,$D7,$F9
	.dc.b	$4B,$5E,$06,$88,$68,$B5,$FB,$0F,$2F,$F8,$49,$E2,$8F,$E8,$93,$87
	.dc.b	$F8,$94,$67,$54,$E6,$7F,$85,$CF,$1D,$5E,$6F,$F1,$42,$4F,$F8,$33
	.dc.b	$F0,$7F,$F6,$8C,$EA,$3F,$8F,$4E,$9F,$EE,$D3,$A8,$C6,$3F,$C1,$9F
	.dc.b	$F1,$64,$66,$78,$90,$E3,$9F,$03,$F9,$44,$D8,$09,$E4,$FF,$24,$8D
	.dc.b	$39,$9E,$FE,$5F,$83,$CB,$93,$EB,$8C,$BE,$D8,$CB,$EA,$8C,$BE,$C8
	.dc.b	$CE,$1F,$CC,$30,$FF,$70,$36,$B3,$7F,$EF,$D9,$FF,$73,$DF,$72,$1C
	.dc.b	$CF,$01,$8A,$E4,$E0,$D1,$E3,$3F,$C7,$BF,$2E,$F6,$BF,$CA,$CD,$8C
	.dc.b	$C7,$CE,$F4,$F2,$C8,$C9,$F6,$E3,$0F,$F8,$B6,$79,$36,$79,$3C,$0F
	.dc.b	$F0,$5C,$69,$FF,$09,$B5,$FA,$A1,$C3,$F6,$1A,$E6,$3F,$C2,$31,$82
	.dc.b	$38,$24,$7F,$86,$23,$2F,$E4,$FC,$EF,$C5,$9E,$AB,$FE,$C7,$3E,$A7
	.dc.b	$EB,$46,$A7,$D5,$1A,$AC,$47,$E6,$27,$96,$3F,$67,$E1,$7F,$2F,$E5
	.dc.b	$CA,$FD,$14,$7E,$68,$C3,$FA,$08,$C3,$A3,$F5,$0E,$B0,$FC,$44,$EA
	.dc.b	$78,$1A,$38,$1F,$AA,$78,$57,$E2,$EB,$85,$72,$AE,$07,$23,$81,$FB
	.dc.b	$7F,$2F,$F5,$E4,$7F,$75,$E5,$D7,$F6,$F1,$67,$F3,$93,$6F,$F3,$53
	.dc.b	$6B,$F6,$3A,$FF,$B9,$5E,$C4,$39,$9D,$8C,$5F,$E9,$E7,$94,$7E,$E6
	.dc.b	$ED,$26,$C7,$E7,$3C,$F9,$5F,$A0,$7F,$84,$FB,$58,$8F,$CC,$45,$C3
	.dc.b	$F6,$93,$71,$FC,$9E,$CA,$C5,$FA,$AB,$93,$F9,$A8,$FC,$99,$1F,$C8
	.dc.b	$35,$FD,$0B,$5E,$06,$88,$68,$B7,$FC,$1F,$CD,$FE,$42,$78,$A3,$F4
	.dc.b	$D3,$87,$F9,$74,$66,$FD,$11,$82,$1F,$00,$FF,$5F,$CB,$E2,$7F,$DB
	.dc.b	$0F,$FB,$46,$FF,$C8,$AD,$C7,$F6,$AE,$EB,$8D,$70,$F3,$24,$60,$87
	.dc.b	$FA,$C6,$79,$3F,$CC,$9A,$FC,$D9,$F0,$9C,$9E,$13,$87,$9D,$F9,$23
	.dc.b	$4F,$9A,$75,$3C,$0F,$6C,$E2,$7D,$7E,$67,$B8,$93,$F0,$BC,$1A,$F9
	.dc.b	$E2,$CF,$C2,$9B,$8F,$C2,$9B,$5E,$B9,$AF,$E1,$E2,$E3,$FE,$3B,$8E
	.dc.b	$87,$D0,$8C,$CE,$9B,$1F,$CD,$AB,$F9,$A9,$1F,$AF,$ED,$D7,$F9,$84
	.dc.b	$5C,$3E,$BC,$D8,$FC,$C9,$67,$E8,$47,$17,$F9,$8C,$7F,$2A,$23,$FC
	.dc.b	$9D,$AF,$F3,$86,$BC,$0D,$10,$D1,$67,$F1,$3E,$6F,$E3,$0F,$14,$7F
	.dc.b	$3D,$9C,$3F,$92,$8C,$E9,$FE,$94,$AE,$33,$93,$E2,$1F,$98,$F2,$FF
	.dc.b	$2F,$E5,$C8,$C1,$0F,$C8,$39,$7F,$25,$1F,$9E,$31,$3F,$9E,$9C,$4F
	.dc.b	$18,$C3,$FA,$68,$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$1F,$AA,$F2
	.dc.b	$7F,$5E,$48,$FD,$9F,$95,$5F,$D4,$45,$AF,$D9,$CD,$AE,$55,$FB,$AA
	.dc.b	$FD,$CC,$E6,$3F,$5D,$15,$FE,$35,$17,$6F,$F6,$94,$71,$3C,$25,$3C
	.dc.b	$25,$BF,$0B,$FE,$2A,$7F,$D9,$3F,$46,$F8,$5F,$EF,$31,$86,$E3,$FD
	.dc.b	$F2,$75,$5C,$63,$FB,$FF,$E1,$1F,$E0,$08,$5F,$17,$47,$F8,$E2,$B8
	.dc.b	$BA,$AE,$97,$DA,$3F,$C7,$B7,$CE,$F9,$46,$CE,$53,$B5,$7D,$27,$FD
	.dc.b	$C7,$5C,$6F,$AD,$79,$4E,$25,$38,$DD,$69,$9B,$F3,$BE,$D6,$BD,$0E
	.dc.b	$8E,$1E,$A6,$23,$A4,$66,$3A,$99,$3E,$28,$AF,$F7,$53,$F0,$BC,$DF
	.dc.b	$74,$FB,$6F,$C1,$87,$BA,$7E,$15,$F0,$BE,$B3,$F2,$0D,$3D,$A3,$57
	.dc.b	$CE,$3E,$7F,$B5,$CB,$CC,$91,$B8,$EA,$61,$FA,$47,$9E,$32,$EC,$EC
	.dc.b	$EE,$FE,$9C,$EA,$78,$DA,$B5,$8F,$E2,$AB,$57,$F6,$A7,$37,$E6,$3B
	.dc.b	$1C,$9C,$BF,$26,$7E,$1F,$99,$23,$E0,$3F,$1B,$CC,$C4,$78,$A2,$DF
	.dc.b	$E2,$26,$E1,$F9,$C9,$B5,$F9,$49,$AF,$D7,$8B,$8F,$FB,$B2,$3A,$1E
	.dc.b	$02,$1F,$02,$77,$3C,$1F,$BB,$ED,$49,$FC,$1E,$68,$FB,$BF,$0B,$99
	.dc.b	$FA,$33,$97,$73,$A6,$90,$DB,$96,$9F,$9D,$7F,$2A,$47,$04,$3F,$3A
	.dc.b	$E5,$A6,$4F,$D2,$39,$69,$F8,$FF,$3B,$F2,$A7,$A7,$3F,$E1,$F3,$89
	.dc.b	$D1,$E7,$8C,$B8,$3F,$0F,$3C,$73,$FE,$1F,$ED,$7E,$83,$CB,$95,$F9
	.dc.b	$D8,$FD,$19,$87,$F1,$F1,$87,$47,$E8,$DD,$5F,$F0,$89,$D4,$F0,$34
	.dc.b	$70,$3F,$32,$F0,$AF,$E4,$35,$C2,$B9,$57,$03,$91,$C0,$FD,$97,$95
	.dc.b	$FB,$82,$3F,$C0,$BF,$09,$FE,$D1,$7F,$CB,$27,$33,$FA,$87,$F9,$04
	.dc.b	$EA,$3F,$93,$C6,$AB,$F9,$0B,$AA,$C1,$CE,$7C,$27,$77,$2F,$23,$27
	.dc.b	$EF,$AB,$B3,$FC,$56,$76,$D3,$91,$F9,$8F,$4B,$F8,$B9,$1F,$67,$C2
	.dc.b	$57,$F3,$31,$70,$61,$FE,$C8,$C1,$79,$8F,$EA,$15,$6F,$FE,$00,$D6
	.dc.b	$9F,$01,$FC,$DE,$3F,$97,$91,$C1,$0F,$E6,$EE,$5A,$64,$FE,$7C,$E5
	.dc.b	$A7,$D5,$F3,$7E,$89,$E3,$7F,$A8,$C6,$1D,$1E,$43,$DE,$60,$F9,$91
	.dc.b	$C6,$7F,$CF,$B1,$F3,$FC,$AF,$BA,$7D,$EE,$89,$1F,$85,$8C,$BF,$87
	.dc.b	$8C,$BC,$4D,$19,$21,$92,$65,$F4,$1F,$97,$1C,$A3,$10,$A5,$78,$23
	.dc.b	$F0,$95,$BC,$7E,$C4,$E6,$F8,$DF,$B1,$BE,$57,$E6,$09,$87,$E7,$FC
	.dc.b	$AA,$FE,$3E,$2C,$F2,$9B,$87,$E1,$66,$D7,$E0,$A6,$BE,$78,$B6,$9B
	.dc.b	$18,$6C,$FE,$BE,$3F,$57,$1F,$E0,$BF,$A6,$77,$52,$57,$ED,$7C,$CA
	.dc.b	$FF,$62,$9D,$BF,$DE,$9A,$FE,$A6,$37,$97,$CC,$9C,$B6,$BF,$77,$3D
	.dc.b	$2B,$FC,$55,$5C,$9B,$3F,$E1,$58,$C4,$08,$E1,$B1,$B2,$3F,$E1,$F8
	.dc.b	$D2,$69,$23,$71,$FC,$07,$CA,$FE,$1E,$7A,$B3,$F0,$47,$19,$CB,$FC
	.dc.b	$66,$74,$F8,$8C,$10,$F9,$11,$FC,$6B,$D9,$FE,$37,$E5,$49,$FE,$45
	.dc.b	$19,$7F,$92,$46,$5F,$E2,$91,$97,$F8,$B4,$67,$2F,$C8,$CE,$5F,$C9
	.dc.b	$C6,$70,$FE,$1D,$39,$7F,$88,$46,$6B,$F8,$6C,$E5,$FE,$1F,$19,$7F
	.dc.b	$49,$1F,$E5,$1F,$C2,$BF,$DD,$51,$FE,$24,$6D,$FE,$A9,$B7,$F2,$1F
	.dc.b	$F7,$39,$89,$FB,$4D,$4D,$37,$E5,$F7,$24,$7F,$92,$23,$2D,$8C,$BE
	.dc.b	$3F,$2E,$DF,$8E,$6D,$F8,$E6,$D7,$8A,$6B,$E2,$8F,$D2,$7C,$84,$8F
	.dc.b	$CF,$3F,$07,$E1,$71,$1F,$A3,$8B,$87,$C1,$9B,$87,$E2,$36,$56,$2F
	.dc.b	$E0,$D7,$27,$F4,$31,$F9,$F2,$3F,$99,$6B,$FE,$7B,$9E,$6E,$48,$66
	.dc.b	$36,$F4,$33,$7F,$57,$CD,$FB,$47,$8E,$3F,$49,$38,$7E,$74,$66,$BB
	.dc.b	$C7,$DF,$8D,$3F,$04,$FA,$9E,$DF,$E1,$7C,$C9,$18,$21,$F8,$77,$2F
	.dc.b	$E2,$63,$F1,$86,$30,$FC,$54,$E2,$78,$C6,$1F,$C1,$46,$5E,$46,$08
	.dc.b	$60,$91,$C5,$F1,$1D,$32,$FC,$B7,$97,$F9,$D2,$47,$E3,$F8,$6B,$F9
	.dc.b	$F8,$B5,$F9,$F9,$B5,$CA,$BF,$2D,$5F,$A7,$9C,$C7,$E8,$22,$BF,$A0
	.dc.b	$8B,$BA,$6C,$61,$B3,$FC,$BF,$DC,$FE,$70,$47,$F4,$9F,$07,$5F,$E6
	.dc.b	$D1,$71,$FD,$A4,$DC,$7F,$29,$36,$BF,$67,$3C,$5F,$E6,$71,$FC,$C4
	.dc.b	$8F,$F2,$B6,$BF,$93,$6B,$C0,$D1,$0D,$16,$BF,$83,$79,$9F,$C2,$CF
	.dc.b	$14,$7F,$35,$9C,$3F,$BF,$8C,$EC,$9C,$CF,$F1,$19,$E3,$E1,$BB,$BF
	.dc.b	$DC,$CF,$FB,$14,$8E,$0C,$F9,$72,$BF,$DA,$73,$C5,$B3,$97,$FD,$83
	.dc.b	$1C,$9C,$9C,$27,$58,$FF,$82,$A7,$94,$F0,$AC,$E7,$FD,$2A,$79,$4E
	.dc.b	$4E,$44,$30,$78,$3C,$AF,$59,$1F,$3F,$97,$5D,$1E,$A8,$F8,$31,$9C
	.dc.b	$3C,$91,$6F,$C5,$36,$BF,$B6,$5A,$B1,$1F,$EB,$39,$B3,$98,$EC,$78
	.dc.b	$8A,$78,$89,$7F,$0E,$7F,$B4,$BF,$E2,$EF,$A8,$7A,$78,$C7,$FB,$6C
	.dc.b	$62,$FF,$BF,$56,$EB,$ED,$E1,$B9,$FF,$01,$56,$DF,$F7,$AB,$6B,$84
	.dc.b	$7D,$C9,$E0,$A7,$57,$E0,$BC,$C6,$27,$E1,$D6,$27,$FC,$03,$12,$7C
	.dc.b	$0D,$8C,$BC,$CA,$77,$6B,$4C,$C7,$27,$81,$C9,$C3,$EE,$8E,$31,$88
	.dc.b	$E4,$6E,$3F,$3F,$E5,$C8,$F9,$37,$D3,$0F,$87,$1F,$9A,$9D,$3F,$22
	.dc.b	$BD,$CF,$E8,$A3,$33,$C0,$FC,$E7,$B7,$FA,$CF,$2A,$57,$EA,$23,$F5
	.dc.b	$E6,$1F,$D3,$C6,$1D,$1F,$AE,$75,$3D,$23,$27,$03,$46,$1D,$39,$23
	.dc.b	$82,$61,$FA,$D9,$EA,$FE,$BE,$BF,$81,$F9,$52,$FF,$7E,$FF,$06,$F2
	.dc.b	$B1,$1F,$BE,$8B,$87,$EE,$E6,$DF,$EE,$66,$D7,$EA,$26,$AD,$13,$85
	.dc.b	$36,$30,$D9,$F1,$9F,$C4,$9F,$C9,$F8,$D2,$3E,$D9,$FA,$07,$F8,$EF
	.dc.b	$C1,$66,$3F,$43,$17,$1F,$E2,$53,$71,$FD,$6C,$DA,$FD,$FC,$F1,$7F
	.dc.b	$39,$1F,$8E,$23,$F9,$66,$BF,$95,$6B,$C0,$D1,$0D,$16,$BF,$D5,$BF
	.dc.b	$85,$FF,$BB,$0F,$F7,$6E,$FF,$EA,$E9,$F6,$9F,$B5,$AF,$DB,$CE,$F0
	.dc.b	$F4,$4E,$4D,$C6,$5E,$35,$CD,$5F,$83,$D9,$E7,$E5,$C9,$E3,$1D,$8C
	.dc.b	$3C,$30,$FF,$3F,$E0,$AD,$4F,$03,$C2,$F0,$BE,$D1,$93,$91,$FE,$97
	.dc.b	$9E,$4F,$03,$51,$C1,$D8,$8D,$B8,$6C,$6C,$95,$E6,$AF,$13,$E8,$BF
	.dc.b	$93,$E5,$CC,$3E,$3B,$F3,$32,$C4,$75,$8B,$87,$D6,$9B,$7F,$0A,$6D
	.dc.b	$7A,$E6,$BE,$78,$B8,$FD,$1B,$FB,$F1,$83,$C1,$3E,$22,$1E,$13,$ED
	.dc.b	$D7,$F3,$2F,$8D,$91,$F7,$1F,$C1,$FC,$26,$23,$F9,$FC,$5B,$FC,$6C
	.dc.b	$DC,$7F,$A5,$CD,$AF,$A5,$3C,$5F,$E7,$51,$FC,$B0,$8F,$F3,$46,$BF
	.dc.b	$CC,$DA,$F0,$34,$43,$45,$AF,$CF,$F9,$7F,$9A,$3C,$51,$FD,$46,$70
	.dc.b	$FE,$6A,$33,$AA,$73,$3F,$81,$9E,$39,$FF,$55,$8F,$C3,$79,$5F,$A6
	.dc.b	$3F,$4F,$CD,$23,$F4,$B1,$97,$F4,$F1,$97,$F1,$91,$97,$F1,$D1,$9C
	.dc.b	$7F,$6F,$39,$7F,$73,$19,$CF,$F4,$B3,$97,$F4,$F1,$9A,$FD,$04,$E5
	.dc.b	$FD,$14,$7E,$F3,$C9,$91,$FB,$A7,$F7,$FE,$56,$23,$F4,$11,$70,$FE
	.dc.b	$13,$37,$0F,$DA,$CD,$AF,$D7,$4D,$7F,$55,$17,$14,$D8,$C3,$67,$C4
	.dc.b	$7F,$0C,$7F,$C0,$FF,$00,$F8,$0A,$49,$FE,$2F,$F8,$67,$F9,$4D,$FE
	.dc.b	$33,$3B,$6C,$ED,$FD,$EC,$ED,$FD,$FC,$ED,$B3,$BC,$99,$26,$5F,$19
	.dc.b	$CC,$66,$F9,$3F,$E1,$66,$BC,$43,$87,$85,$FF,$85,$23,$11,$A8,$C1
	.dc.b	$1F,$F0,$DC,$68,$91,$98,$FF,$10,$C7,$AE,$7C,$B0,$A7,$99,$FE,$32
	.dc.b	$6C,$7C,$62,$1F,$1A,$34,$7C,$3C,$3D,$57,$D5,$F6,$57,$2A,$E0,$72
	.dc.b	$BF,$0C,$3D,$BF,$07,$97,$2B,$EC,$C7,$8C,$C3,$F5,$A3,$0E,$8F,$BE
	.dc.b	$EA,$BA,$CE,$A7,$81,$A3,$81,$F8,$07,$85,$FA,$2B,$85,$72,$AE,$07
	.dc.b	$23,$81,$F8,$EF,$2F,$E4,$11,$FC,$8F,$97,$5F,$C6,$C5,$9F,$83,$36
	.dc.b	$FE,$3C,$DA,$FC,$5C,$D7,$EB,$45,$C5,$36,$30,$D9,$F2,$1E,$23,$F4
	.dc.b	$3E,$B7,$EA,$08,$FE,$9B,$E0,$EB,$FA,$08,$B8,$FD,$69,$B8,$FD,$E9
	.dc.b	$B5,$FA,$09,$E2,$FE,$76,$3F,$30,$47,$F2,$AD,$7F,$26,$D7,$81,$A2
	.dc.b	$1A,$2D,$7E,$23,$CC,$FC,$B1,$E2,$8F,$D4,$CE,$1F,$E0,$31,$9D,$53
	.dc.b	$99,$FD,$FC,$F1,$CD,$ED,$7E,$87,$C9,$95,$FC,$2A,$3F,$48,$61,$FE
	.dc.b	$0F,$18,$74,$7F,$10,$75,$8F,$E7,$A7,$53,$C0,$D1,$C0,$FE,$22,$F0
	.dc.b	$CF,$F5,$35,$C2,$B9,$57,$03,$91,$C0,$FF,$1A,$FE,$1B,$FE,$DC,$8F
	.dc.b	$DE,$B6,$BF,$87,$36,$7F,$8C,$2F,$2B,$FD,$62,$E6,$0D,$24,$6B,$CA
	.dc.b	$93,$E0,$78,$C6,$23,$FC,$4F,$3C,$67,$31,$B2,$1A,$72,$D3,$37,$EA
	.dc.b	$8C,$BF,$C6,$35,$F4,$4F,$F3,$CB,$FE,$86,$7A,$A2,$C6,$48,$79,$1C
	.dc.b	$BF,$CF,$A2,$CF,$C8,$9C,$CF,$F3,$79,$FE,$64,$47,$F9,$7B,$5F,$E5
	.dc.b	$AD,$78,$1A,$21,$A2,$CF,$D6,$F2,$7E,$89,$E2,$C7,$F9,$5D,$62,$BE
	.dc.b	$EC,$E5,$FE,$99,$19,$7E,$7C,$65,$FB,$3B,$7D,$CF,$26,$47,$E1,$A3
	.dc.b	$EF,$98,$7E,$BC,$61,$D1,$F8,$C7,$58,$7E,$32,$75,$3C,$0D,$1C,$0F
	.dc.b	$C0,$BC,$30,$FC,$8D,$70,$AE,$55,$C0,$E4,$70,$3F,$37,$E4,$7E,$50
	.dc.b	$8F,$E7,$BC,$9A,$FE,$3A,$2C,$FE,$82,$6D,$FE,$82,$6D,$7E,$82,$6B
	.dc.b	$F9,$E8,$BA,$A6,$C6,$1B,$3E,$43,$C4,$7F,$A9,$3E,$B1,$F4,$1F,$DE
	.dc.b	$F1,$FE,$DB,$18,$9D,$4F,$EB,$BA,$B5,$97,$F8,$06,$B5,$5F,$E8,$E6
	.dc.b	$DF,$2B,$E1,$3C,$88,$6D,$CB,$4F,$F0,$61,$D5,$AD,$28,$9F,$0B,$C1
	.dc.b	$AE,$CE,$2E,$1F,$F1,$03,$5E,$51,$D1,$A7,$68,$D1,$23,$F8,$7F,$93
	.dc.b	$27,$D7,$5E,$F7,$F8,$DC,$65,$F6,$12,$3F,$8E,$C6,$5E,$11,$FC,$43
	.dc.b	$D9,$F9,$AF,$F8,$D7,$D3,$FF,$0F,$CF,$F8,$81,$B8,$FF,$24,$AF,$E4
	.dc.b	$27,$0A,$4E,$9E,$53,$A7,$F9,$7C,$EE,$75,$1C,$27,$4F,$8C,$FF,$74
	.dc.b	$C6,$8D,$EA,$AD,$DF,$96,$FA,$18,$3C,$D7,$C9,$CB,$E2,$3A,$F9,$3E
	.dc.b	$32,$3E,$0F,$2A,$BF,$ED,$AC,$53,$8B,$F8,$95,$6F,$FC,$A9,$5A,$BE
	.dc.b	$A7,$32,$9C,$EF,$83,$F3,$EA,$BF,$2A,$2C,$FF,$B9,$AB,$AC,$60,$F9
	.dc.b	$55,$F4,$74,$7B,$F4,$98,$7C,$AF,$82,$AF,$E9,$E2,$E5,$F7,$66,$E3
	.dc.b	$FA,$D9,$B5,$F8,$39,$E2,$FE,$7E,$3F,$31,$7E,$08,$FC,$C4,$60,$8E
	.dc.b	$08,$7E,$91,$CB,$4C,$9D,$4F,$C0,$F9,$7F,$87,$3C,$6F,$EC,$63,$0E
	.dc.b	$8F,$E1,$8E,$B3,$FF,$7C,$61,$F5,$2B,$1E,$CF,$E4,$49,$7F,$8E,$F2
	.dc.b	$6B,$F7,$E2,$E7,$FA,$19,$B7,$F7,$A6,$D7,$E9,$E6,$BF,$A2,$8B,$82
	.dc.b	$6C,$61,$B3,$F9,$E8,$FD,$24,$7E,$97,$C8,$FD,$37,$93,$2B,$F4,$71
	.dc.b	$97,$F4,$B1,$97,$F2,$D1,$97,$F3,$11,$9C,$7F,$75,$39,$7F,$79,$19
	.dc.b	$C7,$F3,$33,$97,$F3,$71,$9A,$FE,$09,$39,$7F,$83,$46,$5F,$E0,$31
	.dc.b	$F9,$3F,$62,$47,$F3,$F7,$F8,$07,$C0,$E2,$3F,$9C,$45,$C3,$F8,$74
	.dc.b	$DC,$3F,$9F,$4D,$AF,$E2,$93,$C5,$FE,$61,$1F,$92,$23,$FC,$A5,$AF
	.dc.b	$F3,$26,$BC,$0D,$10,$D1,$6B,$F9,$07,$93,$FC,$5C,$F1,$47,$F4,$19
	.dc.b	$C3,$FC,$5A,$33,$BA,$73,$3F,$C2,$67,$8F,$E1,$D9,$8F,$F7,$15,$FF
	.dc.b	$B3,$23,$10,$8C,$57,$F2,$D9,$FF,$62,$46,$27,$8E,$3F,$EE,$39,$E3
	.dc.b	$39,$9E,$53,$9F,$2B,$95,$C9,$E2,$7F,$BC,$EF,$8D,$6E,$32,$23,$2E
	.dc.b	$E7,$93,$D8,$E5,$58,$BF,$11,$2F,$A4,$EC,$E9,$5B,$8F,$53,$EF,$AF
	.dc.b	$81,$E5,$7C,$1F,$2A,$5F,$93,$1F,$F7,$13,$67,$2F,$58,$4F,$28,$D1
	.dc.b	$C5,$DE,$3F,$46,$B5,$3C,$67,$38,$FC,$D9,$CD,$7D,$79,$CD,$7C,$99
	.dc.b	$CD,$7D,$39,$CC,$FC,$99,$FE,$CD,$F5,$CD,$E9,$27,$D9,$C9,$5F,$8D
	.dc.b	$3B,$7E,$9B,$5F,$74,$6F,$16,$29,$95,$B9,$CC,$62,$FF,$19,$7D,$E3
	.dc.b	$FC,$43,$5D,$5B,$3F,$E1,$58,$C4,$75,$0E,$1E,$87,$F8,$99,$C3,$A8
	.dc.b	$E4,$47,$4D,$7F,$C5,$F1,$E7,$BF,$CE,$79,$5F,$A1,$3D,$CF,$86,$35
	.dc.b	$3F,$99,$9D,$3F,$1A,$B6,$E4,$FC,$44,$61,$FC,$E7,$B3,$FA,$D2,$3F
	.dc.b	$8D,$F2,$AB,$F8,$B8,$B9,$7E,$DE,$6E,$3F,$B1,$9B,$5F,$AB,$9A,$F6
	.dc.b	$3F,$31,$1F,$BB,$8C,$EA,$8B,$18,$6C,$FE,$7F,$C9,$FE,$05,$E4,$C8
	.dc.b	$C1,$0F,$D0,$39,$7F,$45,$1F,$C2,$4C,$63,$FC,$3E,$71,$3C,$63,$0F
	.dc.b	$EB,$63,$2F,$23,$04,$30,$48,$E2,$F8,$8E,$98,$FD,$7F,$8D,$FD,$09
	.dc.b	$23,$F8,$6F,$E1,$FF,$E7,$0A,$FC,$AC,$7E,$32,$74,$FE,$89,$FF,$27
	.dc.b	$4E,$EB,$A4,$FE,$92,$77,$5C,$AB,$53,$FA,$29,$FC,$D1,$1F,$CB,$35
	.dc.b	$FC,$8B,$5F,$D1,$B5,$EE,$68,$E2,$43,$F8,$E4,$7F,$BB,$9B,$E4,$F9
	.dc.b	$2F,$8D,$F9,$63,$D6,$66,$7C,$C7,$C4,$72,$FB,$4F,$8F,$EC,$7C,$22
	.dc.b	$3F,$2B,$CA,$AF,$8A,$2D,$7C,$39,$B7,$FB,$9C,$15,$8B,$FE,$11,$0A
	.dc.b	$C4,$23,$6F,$88,$F0,$14,$F0,$16,$FC,$0D,$BF,$05,$F4,$AE,$D3,$F8
	.dc.b	$2F,$2B,$EE,$F9,$52,$30,$43,$E6,$B9,$7E,$7C,$7E,$10,$C6,$1F,$81
	.dc.b	$9C,$4F,$18,$C3,$F9,$08,$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$1F
	.dc.b	$C4,$FE,$03,$F9,$E9,$23,$F3,$BF,$05,$5F,$E7,$71,$6B,$F2,$D3,$6B
	.dc.b	$95,$7F,$3E,$AF,$CF,$CF,$18,$FE,$6B,$1F,$CC,$C8,$FF,$2F,$6B,$FC
	.dc.b	$A9,$AF,$03,$44,$34,$5A,$FD,$47,$97,$FA,$E3,$C9,$1F,$D2,$A7,$0F
	.dc.b	$EE,$23,$2F,$F4,$28,$CB,$FA,$28,$CE,$4F,$67,$F7,$24,$9F,$E0,$3E
	.dc.b	$55,$7F,$53,$17,$0F,$DD,$4D,$CB,$F3,$13,$6B,$F2,$F3,$5F,$CB,$45
	.dc.b	$B4,$D8,$C3,$67,$F8,$8C,$7F,$0C,$8F,$E1,$BE,$4F,$FB,$6F,$F1,$0F
	.dc.b	$F6,$A9,$FE,$D6,$BF,$F6,$85,$FF,$18,$9D,$61,$23,$04,$3F,$C3,$D5
	.dc.b	$CA,$3A,$C6,$71,$FE,$3F,$39,$8C,$4F,$ED,$6B,$A4,$72,$8E,$2F,$F0
	.dc.b	$E8,$CD,$72,$9C,$BF,$CA,$BE,$C7,$E8,$57,$F8,$E9,$AF,$F8,$C5,$B1
	.dc.b	$E2,$D3,$FD,$F5,$18,$AF,$EF,$CD,$AF,$F0,$63,$56,$0A,$E3,$9C,$9F
	.dc.b	$03,$F2,$E3,$11,$FE,$09,$39,$C5,$7B,$B5,$C9,$6B,$D2,$F1,$6B,$C0
	.dc.b	$D4,$71,$51,$A7,$99,$DD,$C3,$A8,$E1,$1E,$DA,$FF,$78,$4F,$F8,$AD
	.dc.b	$AE,$CF,$69,$FE,$66,$72,$6A,$3E,$31,$CF,$CA,$FC,$61,$F0,$B1,$FC
	.dc.b	$0F,$B5,$F8,$F2,$5F,$3F,$2A,$BF,$8A,$8B,$3F,$88,$9B,$9F,$F8,$CF
	.dc.b	$15,$62,$FD,$15,$5F,$AF,$16,$7F,$DC,$B1,$ED,$8C,$1F,$A1,$9B,$1E
	.dc.b	$D8,$C3,$EF,$F2,$7F,$4D,$E4,$C8,$C1,$0F,$D0,$B9,$7F,$47,$1F,$AD
	.dc.b	$31,$8F,$E8,$A7,$13,$C6,$30,$FC,$D8,$CB,$C8,$C1,$0C,$12,$38,$BE
	.dc.b	$23,$A6,$3F,$9C,$E3,$FC,$F9,$23,$ED,$7C,$15,$7F,$4D,$16,$BF,$6F
	.dc.b	$36,$B9,$57,$EB,$6B,$F8,$44,$F1,$8F,$CB,$47,$E3,$08,$FE,$6D,$AF
	.dc.b	$E4,$DA,$F0,$34,$43,$45,$AF,$E1,$3E,$07,$F1,$73,$C9,$1F,$B6,$9C
	.dc.b	$3F,$95,$8C,$BF,$A4,$8C,$BF,$C5,$A3,$3A,$3D,$7F,$E4,$44,$9F,$CF
	.dc.b	$79,$15,$FE,$3B,$17,$0F,$E3,$73,$72,$FC,$A4,$DA,$FD,$54,$D7,$F8
	.dc.b	$4C,$5B,$4D,$8C,$36,$7F,$5D,$1F,$CD,$63,$FD,$25,$7F,$E4,$8D,$BF
	.dc.b	$DA,$B4,$7F,$DC,$F8,$7F,$A5,$DB,$1F,$EA,$57,$64,$F1,$F9,$CE,$DB
	.dc.b	$3B,$F2,$2B,$D5,$B5,$D2,$3C,$13,$89,$91,$FE,$EA,$9D,$12,$33,$1F
	.dc.b	$EE,$38,$CC,$E2,$FD,$95,$DA,$25,$73,$8F,$E6,$FF,$2B,$23,$F9,$F3
	.dc.b	$E4,$F8,$3C,$47,$F3,$F8,$B7,$F1,$E6,$E3,$FC,$FA,$6D,$7C,$59,$E2
	.dc.b	$FF,$2F,$8F,$E4,$A4,$7F,$9A,$35,$FE,$58,$D7,$81,$A2,$1A,$2D,$7D
	.dc.b	$9F,$2B,$E5,$9E,$28,$FE,$73,$38,$7E,$54,$67,$95,$39,$9F,$95,$3C
	.dc.b	$73,$7B,$3F,$58,$93,$F4,$FC,$8A,$FD,$18,$B8,$FE,$2E,$6E,$7F,$89
	.dc.b	$9B,$5F,$89,$9A,$FE,$3E,$2D,$A6,$C6,$1B,$3F,$83,$8F,$C4,$C7,$E2
	.dc.b	$BC,$6F,$C5,$F8,$F2,$BF,$11,$19,$7F,$15,$19,$7F,$11,$19,$7F,$15
	.dc.b	$19,$C7,$F4,$93,$97,$F4,$D1,$9C,$7F,$49,$39,$7F,$4D,$19,$AF,$D3
	.dc.b	$4E,$5F,$D4,$46,$5F,$D7,$47,$E9,$3D,$17,$C1,$52,$4F,$E8,$5A,$B0
	.dc.b	$FF,$5F,$CF,$08,$19,$C2,$A8,$D3,$FA,$7A,$E7,$5C,$9B,$5F,$BF,$7B
	.dc.b	$CF,$F8,$3E,$B8,$B6,$3F,$C0,$D1,$88,$EE,$D7,$C6,$D7,$C5,$1C,$08
	.dc.b	$FF,$85,$E3,$44,$8C,$C7,$EF,$7C,$8F,$E0,$E7,$99,$EB,$18,$7F,$83
	.dc.b	$46,$5E,$E6,$08,$7C,$53,$89,$F2,$2F,$FD,$D3,$F8,$8B,$D9,$FE,$2E
	.dc.b	$FF,$B4,$E3,$3E,$4E,$26,$57,$1C,$BE,$7D,$ED,$B5,$A9,$F0,$09,$EC
	.dc.b	$F1,$3E,$CB,$C5,$B1,$C6,$FC,$B7,$E4,$89,$3C,$4E,$91,$BB,$F0,$C6
	.dc.b	$6F,$A3,$A3,$13,$C9,$F2,$C6,$8F,$2B,$8F,$23,$E1,$91,$E9,$E4,$D7
	.dc.b	$47,$78,$F9,$31,$9C,$7E,$6C,$5C,$3E,$3C,$DA,$F8,$33,$5F,$8D,$16
	.dc.b	$FF,$DD,$31,$D8,$F3,$C6,$67,$4F,$EA,$F8,$24,$FC,$5D,$7F,$8D,$BF
	.dc.b	$0F,$E0,$72,$FD,$48,$CB,$67,$F0,$33,$AC,$BF,$5F,$17,$2F,$BF,$3C
	.dc.b	$AB,$F4,$D3,$FA,$42,$3F,$9D,$6B,$FE,$77,$BE,$EF,$03,$44,$35,$1C
	.dc.b	$1D,$1F,$5B,$B7,$F2,$A7,$8D,$FD,$7C,$61,$FC,$04,$65,$FE,$05,$19
	.dc.b	$79,$1F,$15,$FC,$FF,$B3,$FA,$0F,$26,$4F,$E2,$63,$2F,$E2,$E3,$2F
	.dc.b	$E2,$23,$2F,$E2,$A3,$39,$7E,$B2,$72,$FE,$BA,$33,$97,$EA,$27,$2F
	.dc.b	$EA,$A3,$35,$FA,$49,$CB,$FA,$68,$CB,$FA,$B8,$FD,$BF,$8B,$23,$F4
	.dc.b	$0F,$EE,$BC,$9C,$47,$E7,$A2,$E1,$FC,$02,$6E,$1F,$BA,$9B,$5F,$B4
	.dc.b	$9A,$FE,$E6,$2E,$FF,$EE,$68,$EE,$43,$B1,$FB,$19,$DC,$F0,$7F,$A0
	.dc.b	$76,$C9,$FC,$B3,$FC,$43,$E0,$71,$1F,$95,$8B,$8F,$F1,$69,$B8,$FF
	.dc.b	$30,$9B,$5F,$C2,$A7,$8B,$FC,$E6,$3F,$9B,$91,$FE,$60,$D7,$F2,$8D
	.dc.b	$78,$1A,$21,$A2,$D2,$BF,$A7,$FE,$25,$67,$F8,$A3,$5F,$51,$EE,$33
	.dc.b	$1F,$E4,$48,$D4,$FF,$5C,$AF,$5D,$27,$35,$D8,$CB,$FD,$76,$3B,$F9
	.dc.b	$1C,$CE,$9C,$12,$32,$43,$F9,$9C,$F1,$6C,$71,$75,$87,$A6,$74,$D9
	.dc.b	$D4,$F2,$AE,$CF,$23,$8D,$FC,$48,$CD,$71,$9C,$9F,$DE,$2F,$FD,$CA
	.dc.b	$76,$AF,$97,$E4,$7C,$52,$4F,$CC,$EC,$AF,$AA,$2D,$F9,$E6,$D7,$2A
	.dc.b	$FA,$75,$F3,$E7,$31,$E1,$8A,$FC,$F8,$B7,$FE,$E4,$8E,$27,$84,$A7
	.dc.b	$84,$B7,$E1,$7F,$AA,$7D,$A3,$E7,$3F,$D0,$2D,$8C,$C7,$E0,$7E,$13
	.dc.b	$2C,$8C,$9F,$8A,$8C,$3F,$DA,$A7,$93,$67,$93,$C0,$FF,$10,$9F,$E2
	.dc.b	$48,$DB,$86,$BF,$E2,$66,$BB,$8E,$2E,$5A,$E4,$B1,$C1,$34,$91,$C2
	.dc.b	$32,$D8,$F4,$D7,$E3,$7C,$AF,$D1,$1E,$E7,$F5,$11,$AC,$FF,$5B,$3A
	.dc.b	$BF,$75,$6A,$F1,$1F,$9C,$9E,$55,$FB,$4F,$67,$F5,$DE,$54,$AF,$DA
	.dc.b	$C7,$EC,$4C,$3F,$A4,$8C,$3A,$3F,$76,$EB,$0F,$D2,$4E,$A7,$81,$A3
	.dc.b	$81,$FA,$E7,$85,$7F,$06,$AE,$15,$CA,$B8,$1C,$8E,$07,$F0,$AF,$27
	.dc.b	$F8,$31,$1F,$E1,$BE,$4D,$7F,$77,$16,$7F,$71,$36,$FF,$8A,$CD,$AF
	.dc.b	$D8,$4D,$7F,$87,$45,$D7,$FD,$D7,$1E,$12,$1D,$8F,$09,$8B,$FE,$0D
	.dc.b	$3C,$6B,$F3,$FA,$CA,$FD,$1B,$FE,$E2,$FC,$4D,$F0,$D9,$9F,$CE,$4D
	.dc.b	$C3,$F9,$45,$6A,$78,$90,$D9,$FE,$41,$74,$E0,$FC,$CE,$EA,$FC,$81
	.dc.b	$23,$F3,$2D,$76,$D7,$F4,$CD,$7F,$4C,$D7,$F5,$8F,$AF,$92,$47,$96
	.dc.b	$FF,$33,$38,$8F,$F3,$0D,$FF,$BF,$EF,$C9,$1B,$D3,$FD,$2F,$2C,$38
	.dc.b	$61,$CA,$BD,$DE,$CF,$C9,$23,$F0,$7C,$8A,$F6,$8B,$3F,$2A,$6E,$1D
	.dc.b	$A6,$D7,$FA,$A3,$05,$62,$10,$9C,$3A,$8F,$19,$0F,$01,$E3,$31,$7F
	.dc.b	$6A,$78,$D7,$E0,$7B,$7F,$05,$E4,$4B,$FB,$51,$97,$EE,$46,$5F,$93
	.dc.b	$19,$7E,$5C,$66,$FF,$0F,$39,$7F,$13,$19,$C7,$E6,$CE,$5F,$9F,$19
	.dc.b	$AF,$85,$39,$7E,$24,$65,$F8,$71,$FC,$EB,$A2,$47,$F3,$E7,$F3,$9F
	.dc.b	$07,$88,$FE,$6D,$17,$0F,$CD,$CD,$C3,$F9,$84,$DA,$FA,$53,$C5,$FE
	.dc.b	$6D,$1F,$C9,$C8,$FF,$30,$6B,$F9,$E6,$BC,$0D,$10,$D1,$6B,$F5,$BE
	.dc.b	$4F,$EC,$CF,$14,$7F,$3A,$9C,$3F,$AD,$8C,$ED,$FD,$D6,$3F,$11,$1A
	.dc.b	$7D,$E7,$EC,$FD,$7F,$DF,$11,$FD,$B7,$91,$5F,$D1,$C5,$C3,$F7,$F3
	.dc.b	$72,$FD,$1C,$DA,$FC,$B4,$D7,$F2,$B1,$77,$4D,$8C,$36,$7C,$87,$88
	.dc.b	$FF,$64,$7A,$DF,$EB,$EC,$3F,$C6,$CF,$FB,$18,$9E,$42,$76,$D7,$FD
	.dc.b	$B7,$1B,$F2,$71,$5C,$AA,$E5,$BA,$EC,$DA,$CD,$F5,$9E,$EF,$63,$CB
	.dc.b	$1A,$51,$A7,$C3,$59,$8F,$14,$71,$32,$72,$21,$83,$FD,$E7,$F6,$DF
	.dc.b	$E7,$96,$F8,$5D,$E3,$18,$FC,$4F,$83,$C4,$CA,$FF,$78,$56,$B2,$F7
	.dc.b	$56,$67,$99,$AA,$F3,$46,$63,$FC,$2B,$5E,$16,$CF,$F8,$1E,$31,$1F
	.dc.b	$3D,$AF,$D3,$6A,$69,$23,$91,$23,$FC,$33,$19,$7F,$03,$E5,$7C,$93
	.dc.b	$D1,$87,$F9,$DE,$7D,$93,$F5,$67,$4F,$AC,$87,$E0,$9C,$BF,$00,$FC
	.dc.b	$2E,$3F,$57,$C9,$91,$F9,$08,$FB,$06,$1F,$C5,$C6,$1D,$1F,$96,$75
	.dc.b	$7F,$96,$9D,$4F,$03,$47,$03,$C2,$60,$D1,$82,$47,$E8,$27,$AE,$5F
	.dc.b	$97,$AF,$C6,$79,$12,$FF,$4E,$FE,$3F,$C9,$C4,$7E,$8A,$2E,$1F,$AA
	.dc.b	$9B,$7F,$A1,$9B,$5F,$9F,$9A,$FE,$72,$2D,$A6,$C6,$1B,$3E,$23,$F7
	.dc.b	$2F,$E5,$FE,$8E,$47,$E5,$1F,$D2,$7C,$26,$23,$F3,$F1,$71,$FD,$FC
	.dc.b	$DC,$7F,$1D,$36,$BF,$69,$3C,$5F,$CB,$C7,$E5,$48,$FE,$31,$AF,$F9
	.dc.b	$A3,$44,$62,$38,$9B,$21,$B7,$8D,$FF,$12,$D7,$F7,$87,$8E,$7F,$3D
	.dc.b	$38,$7F,$8C,$C6,$6B,$BC,$60,$87,$C0,$3F,$8F,$FB,$1F,$C1,$C8,$FF
	.dc.b	$1D,$BF,$F6,$B7,$E2,$9F,$EC,$E8,$FE,$29,$3A,$7F,$8F,$BF,$C4,$E3
	.dc.b	$51,$FC,$52,$35,$3F,$CA,$9D,$65,$83,$F9,$13,$67,$F9,$65,$72,$D5
	.dc.b	$39,$9E,$87,$96,$BF,$98,$56,$E7,$55,$C2,$B4,$74,$AF,$E3,$5F,$2D
	.dc.b	$FC,$2C,$91,$FC,$09,$AF,$D7,$6B,$FC,$51,$AF,$F1,$16,$BF,$C4,$5F
	.dc.b	$F4,$E5,$FF,$63,$83,$49,$E5,$FC,$7E,$A9,$3E,$63,$6F,$CB,$8C,$C3
	.dc.b	$2F,$A7,$38,$7E,$AC,$DC,$3E,$1C,$DA,$F9,$73,$5E,$A7,$7B,$F1,$B6
	.dc.b	$75,$18,$8F,$E8,$7C,$7F,$CC,$48,$FF,$28,$6B,$F9,$D6,$BF,$CE,$9A
	.dc.b	$FF,$37,$6B,$FC,$DD,$FE,$79,$E9,$C8,$FE,$66,$D7,$F5,$0D,$7F,$A2
	.dc.b	$B5,$FE,$84,$D7,$FA,$13,$5F,$E8,$5C,$DF,$D2,$48,$FF,$40,$6B,$FB
	.dc.b	$56,$BF,$D4,$9A,$FF,$4F,$6B,$FD,$3D,$FE,$A9,$CD,$23,$FA,$5B,$5F
	.dc.b	$E0,$0D,$7F,$AE,$B5,$FE,$B4,$D7,$FA,$D3,$5F,$EB,$47,$F6,$12,$1F
	.dc.b	$D6,$0A,$7F,$0D,$29,$FD,$A8,$A7,$F6,$72,$9F,$D9,$CF,$F7,$97,$DC
	.dc.b	$3E,$96,$92,$3F,$DF,$B1,$96,$C6,$5F,$F1,$01,$FE,$23,$8D,$38,$0E
	.dc.b	$1D,$B5,$FF,$14,$35,$E5,$1D,$5D,$F4,$F9,$88,$F9,$1A,$F7,$9F,$F1
	.dc.b	$FB,$92,$19,$8D,$BD,$0C,$9E,$4E,$AF,$86,$47,$E0,$B5,$EC,$D7,$81
	.dc.b	$A2,$1A,$29,$F0,$3D,$4F,$9E,$47,$E6,$B5,$F5,$35,$E0,$68,$86,$8A
	.dc.b	$7C,$CF,$88,$FB,$64,$7E,$CB,$5F,$8C,$D7,$81,$A2,$1A,$29,$F6,$3B
	.dc.b	$BF,$0E,$47,$F0,$AD,$7E,$93,$5E,$06,$88,$68,$A7,$E1,$3D,$6F,$C9
	.dc.b	$91,$FC,$8B,$5F,$BC,$D7,$81,$A2,$1A,$29,$F9,$0F,$07,$F3,$E4,$7F
	.dc.b	$3A,$D7,$F1,$4D,$78,$1A,$21,$A2,$9F,$9C,$F9,$EF,$D5,$91,$FD,$4B
	.dc.b	$5F,$CB,$35,$E0,$68,$86,$8A,$7E,$A3,$C4,$FD,$B9,$1F,$DA,$B5,$FD
	.dc.b	$13,$5E,$06,$88,$68,$A7,$ED,$3C,$4F,$E0,$64,$7F,$80,$B5,$FD,$73
	.dc.b	$5E,$06,$88,$68,$A7,$F0,$0F,$84,$FE,$1E,$47,$F8,$6B,$5F,$DD,$35
	.dc.b	$E0,$68,$86,$8A,$7F,$0C,$F8,$7F,$E3,$64,$7F,$8C,$B5,$FE,$0C,$D7
	.dc.b	$81,$A2,$1A,$29,$FC,$63,$C7,$FE,$4E,$47,$F9,$2B,$5F,$E2,$4D,$78
	.dc.b	$1A,$21,$A2,$9F,$C9,$3C,$8F,$E6,$64,$7F,$98,$B5,$FE,$3C,$D7,$81
	.dc.b	$A2,$1A,$29,$FC,$C3,$DB,$FE,$7E,$47,$F9,$EB,$5F,$E5,$4D,$78,$1A
	.dc.b	$21,$A2,$9F,$CF,$3E,$53,$FA,$59,$1F,$E9,$2D,$7F,$9B,$35,$E0,$68
	.dc.b	$86,$8A,$7F,$48,$F2,$FF,$AB,$91,$FE,$AA,$D7,$FA,$13,$5E,$06,$88
	.dc.b	$68,$A7,$F5,$4F,$77,$FB,$19,$1F,$EC,$2D,$7F,$A7,$35,$E0,$68,$86
	.dc.b	$8A,$7F,$60,$F3,$BF,$B7,$91,$FE,$DA,$D7,$FA,$D3,$5E,$06,$88,$68
	.dc.b	$A7,$F6,$CF,$97,$FE,$F6,$47,$FB,$CB,$5F,$EC,$CD,$78,$1A,$21,$A2
	.dc.b	$9F,$DF,$A9,$E8,$7F,$7D,$80,$8C,$B9,$6B,$FE,$03,$6B,$CE,$39,$3C
	.dc.b	$5B,$1A,$8E,$A6,$4E,$91,$A2,$38,$3C,$1E,$8F,$22,$38,$21,$E3,$72
	.dc.b	$D3,$27,$BD,$CB,$4F,$67,$C4,$F9,$88,$E0,$87,$BD,$CB,$4C,$9F,$2D
	.dc.b	$CB,$4F,$91,$EB,$FC,$22,$38,$21,$F2,$DC,$B4,$C9,$F5,$DC,$B4,$FA
	.dc.b	$9E,$AF,$CE,$23,$82,$1F,$5D,$CB,$4C,$9F,$83,$72,$D3,$F0,$1E,$B7
	.dc.b	$DA,$23,$82,$1F,$83,$72,$D3,$27,$E3,$DC,$B4,$FC,$67,$C0,$FE,$18
	.dc.b	$8E,$08,$7E,$3D,$CB,$4C,$9F,$9B,$72,$D3,$F3,$1F,$53,$F9,$22,$38
	.dc.b	$21,$F9,$B7,$2D,$32,$7E,$9D,$CB,$4F,$D2,$7B,$3F,$9E,$23,$82,$1F
	.dc.b	$A7,$72,$D3,$27,$EC,$DC,$B4,$FD,$87,$C1,$FE,$A8,$8E,$08,$7E,$CD
	.dc.b	$CB,$4C,$9F,$BF,$72,$D3,$F7,$9E,$CF,$ED,$88,$E0,$87,$EF,$DC,$B4
	.dc.b	$C9,$FC,$2D,$CB,$4F,$E1,$1F,$0F,$FC,$08,$8E,$08,$7F,$0B,$72,$D3
	.dc.b	$27,$F1,$77,$2D,$3F,$8A,$7B,$9F,$C3,$88,$E0,$87,$F1,$77,$2D,$32
	.dc.b	$7F,$23,$72,$D3,$F9,$07,$BB,$FC,$68,$8E,$08,$7F,$23,$72,$D3,$27
	.dc.b	$F2,$F7,$2D,$3F,$96,$7C,$5F,$F2,$62,$38,$21,$FC,$BD,$CB,$4C,$9F
	.dc.b	$CE,$DC,$B4,$FE,$71,$F4,$5F,$CC,$88,$E0,$87,$F3,$B7,$2D,$32,$7F
	.dc.b	$47,$72,$D3,$FA,$27,$9F,$FC,$F8,$8E,$08,$7F,$47,$72,$D3,$27,$F5
	.dc.b	$37,$2D,$3F,$A8,$79,$FF,$D2,$88,$E0,$87,$F5,$37,$2D,$32,$7F,$5F
	.dc.b	$72,$D3,$FA,$E7,$C3,$FF,$56,$23,$82,$1F,$D7,$DC,$B4,$C9,$FD,$AD
	.dc.b	$CB,$4F,$ED,$1F,$21,$FD,$88,$8E,$08,$7F,$6B,$72,$D3,$27,$F7,$77
	.dc.b	$2D,$3F,$BA,$7A,$5F,$DB,$88,$E0,$87,$F7,$77,$2D,$32,$7F,$80,$DC
	.dc.b	$B4,$FF,$02,$7D,$CB,$D7,$FE,$FA,$D8,$CC,$60,$8E,$08,$7F,$80,$0F
	.dc.b	$F0,$6C,$71,$70,$1C,$3B,$6B,$FE,$08,$6B,$CA,$3C,$7F,$07,$23,$CE
	.dc.b	$D7,$AC,$A3,$11,$D5,$AF,$56,$BE,$66,$BE,$5F,$A8,$F8,$24,$7E,$2B
	.dc.b	$5E,$ED,$7D,$4D,$7D,$4D,$7E,$1B,$F3,$3D,$B9,$1F,$41,$AF,$A9,$AF
	.dc.b	$C5,$6B,$F1,$5A,$FC,$E6,$BF,$37,$E1,$7E,$C9,$1F,$BA,$D7,$E3,$B5
	.dc.b	$FA,$4D,$7E,$93,$5F,$B6,$FE,$13,$DE,$91,$F8,$86,$BF,$49,$AF,$DD
	.dc.b	$6B,$F7,$5A,$FE,$19,$AF,$E1,$7E,$67,$F2,$24,$7F,$2A,$D7,$EF,$B5
	.dc.b	$FC,$53,$5F,$C5,$35,$FC,$9B,$F9,$CF,$80,$91,$FA,$06,$BF,$8A,$6B
	.dc.b	$F9,$56,$BF,$95,$6B,$F9,$E6,$BF,$9D,$F8,$1F,$D4,$91,$FD,$6B,$5F
	.dc.b	$CB,$B5,$FD,$13,$5F,$D1,$35,$FD,$5B,$FB,$4F,$8D,$91,$FB,$86,$BF
	.dc.b	$A2,$6B,$FA,$D6,$BF,$AD,$6B,$FB,$66,$BF,$B5,$F8,$2F,$E0,$24,$7F
	.dc.b	$82,$B5,$FD,$7B,$5F,$DD,$35,$FD,$D3,$5F,$E0,$6F,$F0,$CF,$83,$91
	.dc.b	$FC,$41,$AF,$EE,$9A,$FF,$05,$6B,$FC,$15,$AF,$F0,$E6,$BF,$C3,$7E
	.dc.b	$0B,$F8,$C9,$1F,$E3,$AD,$7F,$83,$B5,$FE,$24,$D7,$F8,$93,$5F,$E3
	.dc.b	$6F,$F2,$4F,$89,$91,$FC,$A1,$AF,$F1,$26,$BF,$C7,$5A,$FF,$1D,$6B
	.dc.b	$FC,$99,$AF,$F2,$5F,$83,$FE,$62,$47,$F9,$AB,$5F,$E3,$ED,$7F,$95
	.dc.b	$35,$FE,$54,$D7,$F9,$9B,$FC,$F3,$E1,$64,$7F,$40,$6B,$FC,$A9,$AF
	.dc.b	$F3,$56,$BF,$CD,$5A,$FF,$3E,$6B,$FC,$F7,$E2,$BF,$A4,$91,$FE,$9A
	.dc.b	$D7,$F9,$BB,$5F,$E8,$4D,$7F,$A1,$35,$FE,$96,$FF,$54,$F8,$39,$1F
	.dc.b	$D6,$1A,$FF,$42,$6B,$FD,$35,$AF,$F4,$D6,$BF,$D5,$9A,$FF,$55,$F8
	.dc.b	$3F,$EC,$24,$7F,$B2,$B5,$FE,$9E,$D7,$FA,$D3,$5F,$EB,$4D,$7F,$B1
	.dc.b	$BC,$7E,$15,$D1,$FD,$7F,$E0,$FF,$DC,$F1,$8A,$FE,$D3,$18,$9E,$71
	.dc.b	$FE,$F8,$9D,$3E,$73,$11,$CD,$A7,$22,$9B,$11,$CB,$A1,$7D,$E7,$81
	.dc.b	$AA,$E8,$D3,$D6,$74,$95,$FF,$82,$09,$1F,$E0,$C6,$BF,$DD,$9A,$FF
	.dc.b	$80,$5A,$F6,$30,$71,$3F,$C2,$A7,$F8,$50,$86,$08,$7F,$80,$8C,$94
	.dc.b	$C9,$FE,$0F,$32,$53,$FC,$2A,$3E,$E9,$F0,$5F,$0E,$30,$D8,$CC,$7C
	.dc.b	$13,$11,$1F,$9F,$1A,$7E,$41,$FE,$21,$34,$61,$D6,$3F,$71,$B8,$7F
	.dc.b	$88,$E7,$67,$F8,$B6,$3A,$36,$BD,$2F,$98,$F5,$C6,$91,$1C,$36,$38
	.dc.b	$C6,$23,$91,$FE,$38,$70,$48,$DF,$C1,$49,$EE,$D7,$83,$5F,$43,$5F
	.dc.b	$19,$F3,$DC,$3E,$56,$BF,$4F,$E0,$BC,$A4,$7D,$AD,$7C,$CD,$7E,$33
	.dc.b	$5E,$F4,$8C,$47,$C3,$7E,$0F,$C1,$48,$F9,$2D,$7E,$13,$5F,$A4,$D7
	.dc.b	$AB,$5F,$9C,$D7,$F1,$9F,$05,$F3,$88,$FD,$66,$BF,$3D,$AF,$DF,$6B
	.dc.b	$EA,$6B,$F6,$DF,$B3,$F0,$52,$3F,$02,$D7,$ED,$35,$FC,$53,$5F,$8A
	.dc.b	$D7,$F0,$CD,$7F,$49,$F0,$5F,$86,$23,$F8,$E6,$BF,$87,$6B,$F9,$76
	.dc.b	$BF,$49,$AF,$E4,$DE,$3A,$7E,$87,$E0,$7F,$5C,$7E,$BE,$24,$FE,$B2
	.dc.b	$32,$FE,$BA,$32,$FE,$AE,$32,$FE,$B6,$2B,$FA,$98,$FC,$F9,$1F,$D3
	.dc.b	$B5,$FD,$03,$5E,$06,$88,$68,$FD,$7F,$C0,$FE,$A8,$8E,$08,$7E,$AD
	.dc.b	$CB,$4C,$9F,$BD,$72,$D3,$F7,$7F,$03,$FB,$62,$38,$21,$FB,$77,$2D
	.dc.b	$32,$7F,$09,$72,$D3,$F8,$3F,$C0,$FF,$02,$23,$82,$1F,$C0,$DC,$B4
	.dc.b	$C9,$FC,$55,$CB,$4F,$E2,$7F,$03,$FC,$38,$8E,$08,$7F,$0F,$72,$D3
	.dc.b	$27,$F2,$17,$2D,$3F,$8F,$FC,$0F,$F1,$A2,$38,$21,$FC,$6D,$CB,$4C
	.dc.b	$9F,$CB,$5C,$B4,$E3,$F8,$73,$DD,$FE,$16,$7F,$1A,$71,$1F,$CB,$63
	.dc.b	$07,$F1,$09,$B1,$FC,$C5,$AF,$F4,$F8,$BC,$3C,$A1,$56,$15,$A3,$75
	.dc.b	$2B,$F9,$ED,$6E,$71,$1C,$A7,$CE,$E9,$AE,$4F,$21,$93,$D4,$53,$CC
	.dc.b	$5A,$F3,$57,$78,$C4,$F8,$23,$D0,$F9,$E3,$27,$F5,$2E,$4F,$EB,$51
	.dc.b	$FD,$09,$B1,$93,$CA,$E1,$FF,$0F,$9E,$22,$D7,$C7,$8D,$12,$38,$90
	.dc.b	$F5,$9C,$3E,$03,$C0,$FF,$44,$6B,$FD,$95,$AF,$F8,$0D,$B2,$69,$23
	.dc.b	$D0,$71,$29,$C4,$B1,$C5,$B1,$C5,$F5,$FC,$07,$62,$3C,$1A,$E8,$C1
	.dc.b	$0C,$12,$BF,$C0,$D1,$C1,$FF,$04,$7C,$07,$A8,$93,$E4,$6B,$F3,$1B
	.dc.b	$1C,$23,$44,$34,$57,$FB,$DA,$FB,$C7,$C0,$FC,$A6,$C7,$C4,$6B,$F7
	.dc.b	$1B,$3C,$9B,$1F,$E0,$66,$BF,$E1,$A6,$BD,$E7,$B1,$1E,$CD,$7B,$7C
	.dc.b	$0F,$8C,$8F,$8D,$AE,$8C,$10,$C1,$2B,$CD,$1C,$1F,$47,$C0,$FB,$C9
	.dc.b	$3E,$F6,$BE,$F6,$C7,$08,$D1,$0D,$15,$D3,$FB,$73,$E7,$7A,$8C,$7F
	.dc.b	$6D,$8A,$31,$3B,$85,$FF,$10,$6C,$70,$9D,$ED,$FC,$46,$79,$5F,$4B
	.dc.b	$E0,$71,$AD,$ED,$F5,$5B,$87,$D5,$9D,$57,$52,$47,$8E,$32,$F9,$67
	.dc.b	$8B,$EC,$9F,$03,$C3,$0F,$24,$EB,$0F,$2C,$F3,$35,$87,$8E,$EC,$78
	.dc.b	$CA,$75,$3C,$0F,$66,$92,$FB,$C6,$5B,$19,$9E,$CF,$84,$C1,$F9,$5B
	.dc.b	$F3,$1E,$7C,$3D,$F5,$C8,$95,$F9,$D8,$CB,$D2,$33,$5E,$93,$A9,$E9
	.dc.b	$72,$FE,$BE,$78,$BE,$1B,$F7,$E1,$F7,$A3,$8B,$F2,$2B,$8B,$96,$E1
	.dc.b	$E8,$AF,$81,$87,$8C,$97,$D4,$F8,$15,$C2,$7B,$E1,$E2,$9F,$35,$70
	.dc.b	$7C,$C7,$92,$FB,$3C,$A3,$58,$79,$5B,$5C,$E3,$13,$E1,$C3,$F2,$D3
	.dc.b	$9B,$E7,$18,$9E,$D8,$79,$1A,$7C,$2A,$E1,$19,$8F,$8B,$5C,$B0,$FA
	.dc.b	$91,$C2,$FE,$3D,$F0,$8E,$46,$67,$CE,$D3,$CC,$F2,$7A,$DF,$7A,$F9
	.dc.b	$11,$C8,$C9,$5C,$9E,$07,$A1,$F3,$0F,$C2,$3A,$9F,$0C,$72,$8E,$0F
	.dc.b	$E7,$0F,$E5,$97,$FC,$11,$B5,$FE,$E0,$C3,$F5,$71,$83,$F9,$D9,$FA
	.dc.b	$73,$04,$3F,$42,$53,$F0,$85,$3F,$A1,$7A,$8E,$EE,$CD,$8E,$B2,$8C
	.dc.b	$47,$F4,$16,$BF,$D0,$5A,$FF,$41,$6B,$DA,$7A,$91,$EB,$1A,$9E,$D3
	.dc.b	$99,$EC,$FF,$28,$9F,$26,$1F,$E3,$1F,$01,$E4,$35,$D5,$CC,$3F,$96
	.dc.b	$61,$C3,$4F,$F6,$DE,$1C,$2F,$58,$FF,$2B,$AD,$5F,$A7,$0F,$F1,$65
	.dc.b	$EE,$39,$C7,$0C,$3E,$03,$AC,$7C,$53,$A2,$61,$D6,$75,$8F,$19,$EC
	.dc.b	$6A,$FC,$B7,$CA,$F8,$4E,$4D,$BF,$06,$35,$3E,$29,$D1,$F3,$CF,$EA
	.dc.b	$58,$7A,$6B,$99,$27,$83,$CE,$FC,$95,$C6,$34,$85,$73,$AE,$24,$8F
	.dc.b	$2B,$EA,$AE,$79,$7D,$C9,$F2,$D7,$17,$64,$3C,$15,$C5,$F5,$CF,$1B
	.dc.b	$F1,$C7,$8B,$0F,$82,$F1,$9D,$57,$8E,$35,$87,$8B,$2F,$BB,$1D,$8F
	.dc.b	$66,$3E,$37,$04,$BE,$6E,$B0,$EE,$E6,$B9,$CE,$A3,$B9,$E8,$8C,$9E
	.dc.b	$0A,$ED,$3B,$24,$66,$BA,$C6,$32,$FA,$75,$C2,$B9,$C6,$2B,$71,$8B
	.dc.b	$F9,$B1,$89,$F8,$07,$E4,$23,$C8,$F1,$C3,$31,$C1,$B9,$7C,$EC,$3C
	.dc.b	$15,$8C,$BE,$A4,$FD,$2C,$BC,$D8,$7E,$3B,$1F,$03,$D8,$C6,$3C,$5F
	.dc.b	$BD,$8F,$47,$57,$F9,$16,$E1,$F6,$67,$53,$DF,$2F,$8D,$7C,$EF,$EB
	.dc.b	$D6,$AF,$F0,$18,$FD,$38,$C4,$FD,$9A,$F8,$0F,$BC,$98,$7B,$F1,$F4
	.dc.b	$BE,$33,$13,$FE,$86,$AF,$E0,$4F,$F9,$A9,$B5,$F9,$78,$C5,$7F,$02
	.dc.b	$69,$C5,$C4,$6C,$87,$94,$FF,$54,$CF,$47,$13,$DF,$C8,$E3,$1F,$98
	.dc.b	$6B,$FB,$46,$C9,$A4,$9F,$53,$67,$BB,$63,$B1,$A3,$E0,$79,$1D,$08
	.dc.b	$E0,$93,$E8,$8C,$B6,$32,$7C,$78,$CB,$4F,$99,$E4,$7A,$08,$E0,$93
	.dc.b	$F1,$23,$2D,$8C,$9F,$4E,$32,$D3,$EC,$79,$1F,$10,$8E,$09,$3F,$42
	.dc.b	$32,$D8,$C9,$F7,$E3,$2D,$3F,$09,$E4,$7D,$02,$38,$24,$FD,$C8,$CB
	.dc.b	$63,$27,$E2,$E3,$2D,$3F,$21,$E4,$7D,$C2,$38,$24,$FE,$22,$32,$D8
	.dc.b	$C9,$F9,$78,$CB,$4E,$3A,$7F,$6B,$8F,$E3,$B9,$BE,$1E,$C4,$AF,$EC
	.dc.b	$6E,$A3,$17,$FD,$3A,$2D,$FF,$75,$9B,$BF,$F7,$06,$D7,$F3,$06,$BF
	.dc.b	$D3,$1B,$87,$87,$0E,$E7,$F7,$28,$C9,$1F,$F1,$0C,$F1,$6C,$F1,$8F
	.dc.b	$F1,$2B,$5F,$DC,$4F,$F7,$66,$CF,$08,$C4,$79,$E3,$04,$79,$C6,$9F
	.dc.b	$54,$78,$38,$71,$1F,$D9,$9B,$5F,$E2,$D9,$D1,$27,$CB,$3A,$24,$66
	.dc.b	$3F,$87,$46,$1D,$C7,$B9,$EA,$47,$04,$7E,$09,$92,$B9,$7F,$8D,$19
	.dc.b	$2B,$F3,$3E,$F5,$F1,$1E,$C6,$C6,$63,$04,$70,$49,$FF,$2A,$CF,$06
	.dc.b	$CE,$DB,$3B,$7F,$C1,$C7,$F8,$3E,$5E,$4F,$84,$93,$E5,$6C,$F6,$6B
	.dc.b	$C8,$4F,$29,$E0,$48,$E0,$D7,$84,$3C,$9F,$80,$49,$F8,$8D,$9F,$6B
	.dc.b	$5E,$2D,$7D,$73,$CE,$78,$92,$3E,$2F,$91,$FE,$48,$F1,$BF,$29,$5B
	.dc.b	$94,$62,$75,$5D,$2E,$D7,$C7,$9A,$FC,$08,$B6,$9B,$18,$6D,$7F,$8F
	.dc.b	$6F,$B3,$4E,$C5,$BE,$CF,$E1,$67,$F0,$84,$70,$48,$F6,$B5,$F9,$B1
	.dc.b	$B7,$24,$3E,$8F,$91,$F8,$F2,$4F,$E4,$DB,$1F,$8C,$6B,$C2,$34,$47
	.dc.b	$4D,$7E,$EF,$91,$F9,$B2,$47,$E7,$DB,$1F,$98,$6B,$C2,$34,$47,$4D
	.dc.b	$7F,$13,$E4,$7E,$9C,$91,$FA,$B6,$C7,$E9,$1A,$F0,$8D,$11,$D3,$5F
	.dc.b	$CA,$FC,$3F,$EC,$C9,$1F,$B7,$6C,$7E,$C1,$AF,$08,$D1,$1D,$35,$FD
	.dc.b	$0F,$93,$FB,$F2,$47,$F0,$36,$C7,$EF,$1A,$F0,$8D,$11,$D3,$5F,$D6
	.dc.b	$F9,$3F,$C2,$C9,$1F,$C3,$DB,$1F,$C2,$1A,$F0,$8D,$11,$D3,$5D,$1F
	.dc.b	$A8,$7F,$63,$1F,$A2,$F2,$3F,$46,$C8,$FC,$E4,$FF,$06,$8D,$46,$36
	.dc.b	$46,$E7,$A1,$FC,$55,$DB,$83,$94,$F1,$AF,$E5,$04,$7F,$96,$36,$3F
	.dc.b	$92,$B5,$FE,$22,$D7,$F8,$4B,$5F,$E4,$8F,$F3,$3F,$22,$4F,$F3,$76
	.dc.b	$CF,$F3,$06,$BF,$C7,$1A,$FF,$14,$6B,$FC,$BD,$AF,$F2,$8F,$87,$FE
	.dc.b	$13,$1F,$C2,$9B,$3F,$15,$B3,$F0,$70,$46,$37,$FE,$36,$DA,$FE,$46
	.dc.b	$DA,$FE,$7C,$FF,$4D,$24,$7F,$51,$6C,$7F,$4A,$6B,$FC,$D9,$AF,$48
	.dc.b	$E0,$47,$FA,$8F,$C3,$FF,$59,$24,$7F,$5D,$6C,$6A,$30,$47,$04,$3F
	.dc.b	$A2,$C7,$07,$FA,$3F,$3F,$F6,$52,$4F,$F6,$96,$C7,$F6,$26,$BC,$23
	.dc.b	$44,$74,$D7,$FA,$9F,$91,$FD,$C4,$91,$FD,$D5,$B1,$FD,$B9,$AF,$08
	.dc.b	$D1,$1D,$35,$FE,$BF,$E4,$7F,$7D,$24,$7F,$80,$5B,$1F,$DE,$9A,$F0
	.dc.b	$8D,$11,$D3,$5F,$ED,$7E,$27,$F8,$24,$91,$FE,$0D,$6C,$7F,$81,$9A
	.dc.b	$F0,$8D,$11,$D3,$5F,$EF,$DF,$7C,$F2,$7F,$C6,$2D,$9F,$F2,$2B,$67
	.dc.b	$FC,$08,$1C,$3C,$A7,$89,$27,$8B,$63,$8C,$F0,$7F,$C2,$ED,$47,$58
	.dc.b	$F1,$F9,$32,$BC,$6D,$9E,$6D,$7D,$6D,$7A,$B5,$EB,$48,$C4,$7A,$FC
	.dc.b	$9F,$82,$4A,$F8,$2D,$8F,$4B,$5F,$92,$D7,$D6,$D7,$BB,$F3,$3E,$22
	.dc.b	$4F,$CC,$6C,$FC,$46,$BF,$51,$AF,$C6,$6B,$EA,$6B,$F4,$FE,$23,$EC
	.dc.b	$92,$7E,$CB,$63,$E8,$B5,$FC,$0B,$5F,$A6,$D7,$E3,$BF,$84,$F8,$89
	.dc.b	$3F,$84,$6C,$FD,$C6,$BF,$8C,$6B,$F7,$9A,$FD,$26,$BF,$8B,$F8,$8F
	.dc.b	$C8,$92,$7F,$22,$D8,$FC,$4B,$5F,$CC,$B5,$FC,$5B,$5F,$BE,$FE,$73
	.dc.b	$E2,$24,$FE,$71,$B3,$F9,$46,$BF,$A4,$6B,$F9,$66,$BF,$8A,$6B,$FA
	.dc.b	$3F,$88,$FD,$49,$27,$F5,$2D,$8F,$D0,$B5,$FD,$8B,$5F,$D1,$B5,$FC
	.dc.b	$BB,$FB,$4F,$88,$93,$FB,$46,$CF,$EB,$1A,$FE,$F1,$AF,$EB,$9A,$FE
	.dc.b	$89,$AF,$EE,$FE,$23,$F8,$09,$27,$F8,$0B,$63,$F7,$2D,$7F,$84,$B5
	.dc.b	$FD,$DB,$5F,$D7,$BF,$C3,$3E,$22,$4F,$F0,$C6,$CF,$F0,$46,$BF,$C5
	.dc.b	$1A,$FF,$06,$6B,$FB,$A6,$BF,$C4,$FE,$23,$F8,$C9,$27,$F8,$CB,$63
	.dc.b	$F8,$8B,$5F,$E4,$2D,$7F,$89,$B5,$FE,$0E,$FF,$24,$F8,$89,$3F,$C9
	.dc.b	$1B,$3F,$C7,$1A,$FF,$2C,$6B,$FC,$79,$AF,$F1,$26,$BF,$CA,$FE,$23
	.dc.b	$F9,$89,$27,$F9,$8B,$63,$F9,$4B,$5F,$E7,$2D,$7F,$95,$B5,$FE,$3E
	.dc.b	$FF,$3C,$F8,$89,$3F,$CF,$1B,$3F,$CD,$1A,$FF,$44,$6B,$FC,$D9,$AF
	.dc.b	$F2,$A6,$BF,$D0,$FD,$6F,$E9,$24,$9F,$E9,$2D,$8F,$E8,$2D,$7F,$A8
	.dc.b	$B5,$FE,$86,$D7,$F9,$BB,$FD,$53,$D6,$93,$FD,$51,$B3,$FD,$31,$AF
	.dc.b	$F5,$C6,$BF,$D3,$9A,$FF,$42,$6B,$FD,$6F,$E2,$3F,$B0,$92,$7F,$B0
	.dc.b	$B6,$3F,$AC,$B5,$FE,$D2,$D7,$FA,$DB,$5F,$E9,$EF,$F6,$CF,$88,$93
	.dc.b	$FD,$B1,$B3,$FD,$91,$AF,$F7,$46,$BF,$D9,$9A,$FF,$5A,$6B,$FD,$CE
	.dc.b	$3F,$BC,$90,$FE,$F2,$57,$FB,$8B,$5F,$F0,$0B,$5F,$EE,$6D,$7F,$B3
	.dc.b	$A3,$FC,$02,$BE,$FD,$F1,$39,$8F,$F0,$2C,$62,$76,$D8,$CC,$ED,$FE
	.dc.b	$FE,$7F,$80,$23,$4E,$0E,$D1,$B7,$71,$A6,$C6,$5E,$4D,$78,$CF,$8A
	.dc.b	$31,$3C,$67,$C1,$F1,$38,$25,$7A,$1B,$1E,$C6,$BE,$59,$46,$23,$CA
	.dc.b	$D7,$AC,$FC,$2F,$89,$F0,$91,$C1,$23,$E4,$C6,$5B,$19,$7C,$11,$96
	.dc.b	$C7,$CE,$F8,$9F,$69,$1C,$12,$3E,$AC,$65,$B1,$97,$D9,$19,$6C,$7D
	.dc.b	$AF,$89,$F9,$24,$70,$48,$FC,$0C,$65,$B1,$97,$E4,$46,$5B,$1F,$86
	.dc.b	$F8,$9F,$AA,$47,$04,$8F,$C6,$C6,$5B,$19,$7E,$A4,$65,$B1,$F9,$2F
	.dc.b	$89,$FC,$09,$1C,$12,$3F,$33,$19,$6C,$65,$FC,$04,$65,$B1,$F9,$EF
	.dc.b	$89,$FC,$69,$1C,$12,$3F,$4B,$19,$6C,$65,$FC,$64,$65,$B1,$FA,$AF
	.dc.b	$89,$FC,$C9,$1C,$12,$3F,$63,$19,$6C,$65,$FC,$C4,$65,$B1,$FB,$6F
	.dc.b	$89,$FD,$29,$1C,$12,$3F,$7B,$19,$6C,$65,$FD,$24,$65,$B1,$FC,$0B
	.dc.b	$E0,$BF,$62,$47,$04,$8F,$E1,$31,$96,$C6,$5F,$D8,$46,$5B,$1F,$C3
	.dc.b	$BE,$0B,$F7,$A4,$70,$48,$FE,$2B,$19,$6C,$65,$FD,$E4,$65,$B1,$FC
	.dc.b	$6B,$E0,$BF,$84,$91,$C1,$23,$F9,$0C,$65,$B1,$97,$F8,$44,$65,$B1
	.dc.b	$FC,$9B,$E0,$BF,$8A,$91,$C1,$23,$F9,$6C,$65,$B1,$97,$F8,$A4,$65
	.dc.b	$B1,$FC,$CB,$E0,$BF,$90,$91,$C1,$23,$F9,$CC,$65,$B1,$97,$F9,$04
	.dc.b	$65,$B1,$FC,$FB,$E0,$BF,$96,$91,$C1,$23,$FA,$2C,$65,$B1,$97,$F9
	.dc.b	$64,$65,$B1,$FD,$2B,$E0,$BF,$9C,$91,$C1,$23,$FA,$8C,$65,$B1,$97
	.dc.b	$F9,$C4,$65,$B1,$FD,$5B,$E2,$7F,$A2,$91,$C1,$23,$FA,$EC,$65,$B1
	.dc.b	$97,$FA,$24,$65,$B1,$FD,$8B,$E0,$7F,$A8,$91,$C1,$23,$FB,$4C,$65
	.dc.b	$B1,$97,$FA,$84,$65,$B1,$FD,$BB,$E0,$BF,$AE,$91,$C1,$23,$FB,$AC
	.dc.b	$65,$B1,$97,$FA,$E4,$65,$B1,$FD,$EB,$E0,$BF,$B4,$91,$C1,$23,$FC
	.dc.b	$03,$19,$6C,$65,$FE,$D1,$19,$6C,$7F,$81,$9F,$EE,$A4,$70,$47,$FC
	.dc.b	$1A,$64,$AE,$5F,$EE,$86,$4A,$FF,$86,$3A,$DF,$05,$FE,$05,$6C,$66
	.dc.b	$30,$47,$04,$9F,$F0,$A0,$9E,$2E,$1A,$FF,$86,$5A,$FF,$87,$9A,$8E
	.dc.b	$53,$E3,$F8,$29,$5E,$76,$CF,$56,$BD,$5A,$F5,$6B,$D6,$91,$88,$F4
	.dc.b	$7C,$17,$C1,$25,$7C,$56,$C7,$AD,$AF,$AD,$AF,$AD,$AF,$77,$E6,$7C
	.dc.b	$44,$9F,$A0,$D9,$F8,$CD,$7E,$33,$5F,$8C,$D7,$D4,$D7,$E8,$7C,$1F
	.dc.b	$D9,$24,$FD,$D6,$C7,$D3,$6B,$F4,$DA,$FD,$36,$BF,$1D,$FC,$27,$C1
	.dc.b	$C9,$FC,$43,$67,$EF,$35,$FB,$CD,$7E,$F3,$5F,$A4,$D7,$F1,$1F,$07
	.dc.b	$F9,$12,$4F,$E5,$5B,$1F,$8B,$6B,$F8,$B6,$BF,$8B,$6B,$F7,$DF,$CE
	.dc.b	$7C,$1C,$9F,$D0,$36,$7F,$2C,$D7,$F2,$CD,$7F,$2C,$D7,$F1,$4D,$7F
	.dc.b	$41,$F0,$7F,$A9,$24,$FE,$B5,$B1,$FA,$36,$BF,$A3,$6B,$FA,$36,$BF
	.dc.b	$97,$7F,$69,$F0,$72,$7F,$70,$D9,$FD,$73,$5F,$D7,$35,$FD,$73,$5F
	.dc.b	$D1,$35,$FD,$C7,$C1,$FF,$01,$24,$FF,$05,$6C,$7E,$ED,$AF,$EE,$DA
	.dc.b	$FE,$ED,$AF,$EB,$DF,$E1,$9F,$11,$27,$F8,$83,$67,$F8,$33,$5F,$E0
	.dc.b	$CD,$7F,$83,$35,$FD,$D3,$5F,$E2,$1F,$09,$FC,$64,$93,$FC,$75,$B1
	.dc.b	$FC,$4D,$AF,$F1,$36,$BF,$C4,$DA,$FF,$07,$7F,$92,$7C,$24,$9F,$E5
	.dc.b	$0D,$9F,$E3,$CD,$7F,$8F,$35,$FE,$3C,$D7,$F8,$93,$5F,$E5,$1F,$11
	.dc.b	$FC,$C4,$93,$FC,$D5,$B1,$FC,$AD,$AF,$F2,$B6,$BF,$CA,$DA,$FF,$1F
	.dc.b	$7F,$9E,$7C,$1C,$9F,$E8,$0D,$9F,$E6,$CD,$7F,$9B,$35,$FE,$6C,$D7
	.dc.b	$F9,$53,$5F,$E8,$1F,$07,$FD,$24,$93,$FD,$35,$B1,$FD,$0D,$AF,$F4
	.dc.b	$36,$BF,$D0,$DA,$FF,$37,$7F,$DB,$5F,$89,$BF,$2B,$7A,$7F,$EE,$DF
	.dc.b	$86,$FF,$77,$4E,$63,$18,$7F,$4B,$8E,$1E,$AE,$31,$46,$E7,$FA,$85
	.dc.b	$F2,$C3,$AC,$60,$E4,$E2,$93,$BB,$E9,$7C,$E7,$37,$C3,$0C,$E1,$C7
	.dc.b	$0E,$13,$D6,$BC,$73,$98,$E0,$66,$FA,$57,$0A,$EA,$70,$BE,$A4,$AE
	.dc.b	$16,$C3,$15,$C3,$0E,$75,$DA,$31,$5B,$8E,$37,$F1,$E7,$FB,$E7,$C2
	.dc.b	$7F,$81,$49,$3F,$E0,$96,$C7,$F7,$F6,$BF,$DF,$DA,$FF,$7F,$6B,$FD
	.dc.b	$CD,$47,$F8,$2B,$17,$EC,$0E,$9F,$B1,$18,$6C,$7F,$7A,$9F,$95,$39
	.dc.b	$6C,$71,$6C,$F0,$EA,$E2,$C8,$C9,$C2,$31,$87,$CE,$BE,$8D,$BE,$8D
	.dc.b	$9D,$CF,$E4,$E7,$2D,$BF,$05,$F6,$BE,$51,$C2,$3F,$23,$5E,$B9,$CC
	.dc.b	$7E,$15,$7E,$06,$F8,$2F,$C7,$FE,$4D,$F5,$E7,$F0,$66,$A3,$3B,$FD
	.dc.b	$9A,$E1,$7F,$8B,$BE,$19,$7D,$0C,$3F,$C6,$15,$A7,$F0,$D3,$AA,$FF
	.dc.b	$1A,$47,$57,$95,$F5,$C3,$BC,$79,$EF,$95,$70,$38,$1F,$E4,$67,$2F
	.dc.b	$46,$BE,$29,$F2,$3C,$63,$2F,$23,$11,$A7,$93,$5E,$53,$FE,$54,$78
	.dc.b	$1A,$71,$3E,$FC,$3B,$B6,$3F,$84,$7E,$4C,$F1,$3D,$D1,$ED,$C2,$E1
	.dc.b	$FA,$BA,$D4,$E4,$91,$F5,$2F,$A3,$FC,$42,$76,$E6,$BF,$3B,$FA,$9B
	.dc.b	$E2,$29,$FA,$0F,$49,$F1,$38,$CB,$F3,$F5,$B6,$D6,$E7,$58,$7E,$66
	.dc.b	$B4,$D9,$D5,$70,$24,$65,$FD,$5D,$F8,$1C,$E9,$FC,$BE,$F9,$B6,$F9
	.dc.b	$DF,$12,$46,$6B,$9E,$1E,$48,$FE,$6F,$E7,$7F,$9F,$E7,$F8,$F2,$F1
	.dc.b	$FF,$95,$B9,$BF,$E3,$0E,$67,$71,$A9,$DB,$63,$27,$48,$CB,$4C,$C7
	.dc.b	$48,$CB,$63,$35,$E6,$9C,$B6,$32,$7F,$4D,$AF,$25,$79,$8F,$11,$F8
	.dc.b	$1F,$81,$FE,$33,$FB,$0B,$D2,$FD,$A4,$61,$B7,$FC,$6A,$74,$FE,$AE
	.dc.b	$74,$D3,$53,$C2,$BF,$8B,$57,$07,$C1,$3F,$9E,$AF,$D2,$47,$F0,$A6
	.dc.b	$F0,$1A,$49,$E0,$49,$FD,$C4,$65,$B1,$97,$F6,$D1,$E3,$C7,$F9,$17
	.dc.b	$91,$6B,$F7,$D3,$6B,$FB,$4F,$2A,$B1,$1E,$5A,$F7,$3F,$CB,$E3,$87
	.dc.b	$5B,$F1,$B7,$87,$FD,$C1,$78,$96,$30,$DC,$3F,$B5,$4E,$9B,$3A,$69
	.dc.b	$A9,$E1,$87,$F6,$7A,$E0,$DA,$E0,$DA,$E0,$72,$78,$57,$2C,$FF,$C2
	.dc.b	$B7,$C9,$B7,$C9,$B7,$CA,$FF,$71,$CB,$FC,$10,$FE,$0B,$F1,$56,$7F
	.dc.b	$80,$C5,$9F,$DF,$4D,$AF,$DB,$4D,$AF,$F0,$EF,$2A,$B1,$7F,$E1,$A8
	.dc.b	$FE,$2B,$66,$9B,$AE,$E4,$9F,$2B,$FC,$4A,$B9,$D7,$1B,$B7,$C6,$F6
	.dc.b	$7F,$23,$F8,$0F,$F1,$3F,$D1,$3E,$07,$FA,$B4,$61,$B9,$7F,$88,$67
	.dc.b	$4F,$F4,$D9,$D3,$4D,$4F,$0A,$FF,$0F,$57,$07,$C1,$3F,$D0,$2B,$F4
	.dc.b	$24,$70,$48,$FE,$AD,$19,$6C,$65,$FE,$A9,$19,$6C,$78,$DF,$0F,$9D
	.dc.b	$E3,$CB,$FA,$6D,$5A,$FF,$2A,$5A,$B1,$7F,$DC,$AA,$EE,$AD,$5F,$5C
	.dc.b	$3F,$C1,$DE,$6F,$F8,$8B,$CE,$FF,$11,$B6,$7F,$C4,$2D,$C7,$FC,$30
	.dc.b	$DA,$E3,$3F,$E2,$6A,$E2,$D3,$89,$6B,$8B,$67,$8C,$7F,$88,$E7,$8B
	.dc.b	$67,$8B,$4E,$25,$9F,$F3,$BF,$D8,$BE,$0B,$FC,$D2,$DA,$FF,$30,$3F
	.dc.b	$02,$BF,$CB,$0E,$5B,$3C,$0E,$3A,$5A,$E3,$16,$78,$CD,$BF,$0C,$DA
	.dc.b	$FF,$43,$5F,$79,$CC,$72,$25,$F2,$6C,$6A,$3C,$71,$E8,$BE,$90,$1F
	.dc.b	$84,$BE,$03,$FC,$B7,$F3,$EF,$4F,$FC,$DD,$39,$C3,$FC,$B7,$39,$7F
	.dc.b	$C8,$51,$9B,$FF,$2A,$4E,$5F,$F2,$0C,$67,$1F,$F5,$24,$E5,$FF,$15
	.dc.b	$C6,$6B,$FD,$3F,$39,$7F,$C7,$B1,$9A,$FF,$4E,$CE,$5F,$54,$78,$5F
	.dc.b	$F1,$E3,$5E,$38,$7F,$4E,$8C,$47,$08,$C1,$2B,$FC,$69,$3A,$9F,$8B
	.dc.b	$39,$7F,$C9,$06,$27,$71,$9B,$F5,$BF,$E2,$8F,$8E,$F9,$2C,$9F,$8E
	.dc.b	$E5,$A4,$AF,$8F,$19,$6C,$7F,$AC,$FF,$1A,$7D,$EF,$F8,$86,$7F,$D5
	.dc.b	$F1,$8F,$0F,$FD,$94,$EB,$CC,$B5,$BC,$BF,$CD,$93,$BC,$3A,$4E,$5F
	.dc.b	$C2,$47,$17,$37,$C8,$8F,$0B,$B8,$78,$AF,$92,$BE,$CE,$67,$78,$78
	.dc.b	$AB,$75,$A8,$C4,$EA,$FC,$75,$A2,$46,$6F,$FB,$BF,$AD,$FE,$B0,$DB
	.dc.b	$FD,$62,$DB,$FF,$57,$B6,$7F,$D4,$CD,$C3,$FD,$48,$D9,$FF,$4F,$B6
	.dc.b	$7A,$C7,$FA,$E2,$BA,$B4,$EA,$58,$EA,$DA,$EA,$D3,$A9,$63,$F3,$3E
	.dc.b	$1F,$FB,$DB,$EB,$3F,$CE,$35,$FA,$48,$C7,$95,$FB,$26,$DF,$EC,$1B
	.dc.b	$7F,$AE,$6D,$71,$9F,$F7,$B5,$F4,$7F,$57,$3F,$A3,$23,$FA,$06,$D7
	.dc.b	$E7,$B8,$3F,$C7,$0F,$EB,$23,$2D,$9E,$04,$78,$47,$2B,$F1,$47,$FB
	.dc.b	$F7,$EB,$9E,$57,$AE,$FD,$44,$9D,$46,$5F,$64,$66,$BF,$D8,$9D,$4B
	.dc.b	$E7,$39,$3F,$C1,$7F,$23,$FE,$98,$3F,$D3,$3F,$25,$63,$FD,$2B,$17
	.dc.b	$0F,$F4,$94,$DC,$3F,$D1,$53,$6B,$FC,$2B,$36,$BF,$C2,$93,$6B,$FC
	.dc.b	$29,$37,$BD,$36,$30,$DB,$F0,$9F,$BB,$FB,$DF,$F0,$95,$FE,$D1,$FF
	.dc.b	$4C,$C7,$EC,$9C,$B7,$CD,$FD,$EC,$63,$1F,$D7,$3C,$1B,$E1,$EA,$76
	.dc.b	$DA,$F0,$36,$7F,$84,$D7,$3E,$63,$CD,$B3,$BC,$3C,$CD,$BF,$33,$6F
	.dc.b	$C3,$1E,$DC,$3C,$2D,$3F,$A3,$FB,$FF,$E9,$0F,$84,$FE,$9A,$D8,$C9
	.dc.b	$FE,$8F,$8C,$B4,$CD,$7F,$A3,$23,$2D,$8C,$BF,$D0,$A7,$FA,$01,$23
	.dc.b	$F9,$EB,$63,$F3,$CD,$78,$46,$88,$E9,$B8,$7F,$58,$94,$7F,$6B,$8B
	.dc.b	$3F,$E8,$AF,$77,$CB,$1B,$9F,$4B,$A1,$E5,$A7,$73,$FD,$82,$ED,$F3
	.dc.b	$8E,$27,$E1,$7B,$7F,$06,$7E,$13,$C7,$B1,$F8,$D8,$B3,$F7,$A6,$E3
	.dc.b	$F8,$F9,$B5,$F8,$F9,$B5,$F8,$C9,$BA,$FE,$12,$6E,$C9,$B1,$86,$DF
	.dc.b	$84,$F8,$87,$F9,$FB,$E4,$BE,$03,$FE,$80,$F9,$3B,$8F,$F9,$FE,$6E
	.dc.b	$5F,$E7,$F9,$BC,$69,$B1,$86,$DF,$6B,$E4,$DB,$E4,$DB,$E4,$DB,$E4
	.dc.b	$D3,$57,$D2,$D7,$F5,$EF,$8C,$62,$BD,$47,$03,$93,$93,$C6,$59,$F1
	.dc.b	$B6,$79,$C7,$5B,$F8,$72,$FC,$2D,$EC,$7C,$6E,$77,$C0,$70,$CB,$FC
	.dc.b	$75,$7B,$AD,$4E,$63,$FC,$7D,$7C,$AB,$E3,$5F,$2C,$FB,$CE,$6B,$D1
	.dc.b	$39,$AF,$99,$1D,$5E,$37,$D6,$30,$70,$71,$5C,$E3,$CA,$68,$AE,$5F
	.dc.b	$4B,$5F,$3E,$1F,$AA,$76,$D7,$6D,$77,$5F,$E3,$47,$B1,$A7,$13,$D6
	.dc.b	$32,$70,$3E,$B1,$EA,$CB,$FA,$6F,$CF,$7F,$2B,$8C,$57,$37,$75,$F1
	.dc.b	$EB,$7C,$BF,$2B,$09,$7F,$22,$B4,$FF,$9B,$F2,$63,$D6,$73,$8F,$E3
	.dc.b	$B9,$1E,$77,$E8,$9B,$87,$E6,$DF,$F7,$B3,$FC,$3D,$CC,$FE,$35,$B9
	.dc.b	$FF,$0D,$6C,$EF,$D0,$E9,$72,$3A,$3D,$E3,$17,$CE,$2D,$F3,$9B,$5E
	.dc.b	$19,$B5,$E9,$9B,$5F,$C7,$26,$DF,$A2,$BD,$D1,$FD,$6B,$A7,$FA,$DF
	.dc.b	$55,$C7,$FB,$23,$6B,$FA,$AB,$70,$FE,$CC,$D9,$FE,$CC,$D9,$FE,$C6
	.dc.b	$DE,$6F,$EB,$6D,$E6,$34,$95,$FD,$CF,$0F,$1B,$4F,$19,$F6,$FA,$3F
	.dc.b	$89,$FC,$AF,$EA,$1B,$19,$3F,$87,$46,$5A,$67,$1F,$E0,$11,$96,$C6
	.dc.b	$5F,$D2,$CF,$E7,$5F,$E0,$4D,$DB,$FA,$34,$FE,$72,$30,$47,$04,$9F
	.dc.b	$F4,$BC,$F0,$6C,$ED,$B3,$E5,$CF,$F8,$F7,$A7,$E5,$7F,$65,$7C,$5B
	.dc.b	$7F,$CD,$27,$55,$FB,$A8,$B7,$FC,$A3,$9D,$58,$D7,$FC,$1B,$CD,$FE
	.dc.b	$1A,$E7,$FF,$0D,$B7,$1F,$F1,$23,$70,$FF,$0A,$B7,$2F,$F1,$33,$67
	.dc.b	$FC,$4C,$D9,$FF,$11,$B6,$FB,$CF,$F8,$BE,$FB,$B4,$EE,$5A,$EE,$DB
	.dc.b	$EE,$D3,$F6,$1F,$05,$FB,$53,$F6,$DD,$16,$3F,$6D,$16,$BF,$69,$36
	.dc.b	$BF,$69,$36,$BF,$75,$36,$BF,$81,$4D,$AF,$F0,$2C,$DE,$34,$D8,$C3
	.dc.b	$6F,$C2,$7F,$11,$F8,$0F,$E9,$1F,$52,$F8,$AF,$E7,$F1,$86,$CF,$F4
	.dc.b	$49,$D3,$C4,$C4,$6B,$2F,$E8,$B3,$A7,$F9,$1D,$73,$AE,$11,$F9,$08
	.dc.b	$FC,$C1,$23,$F9,$9B,$63,$51,$82,$38,$21,$FD,$22,$38,$65,$F9,$47
	.dc.b	$FB,$27,$C5,$3E,$2F,$CD,$39,$BF,$1B,$FE,$88,$9D,$46,$0F,$EA,$35
	.dc.b	$C1,$FE,$6D,$1A,$BF,$6C,$7F,$82,$7E,$57,$FC,$27,$F2,$DF,$E1,$46
	.dc.b	$C6,$4F,$F0,$74,$65,$A6,$67,$FC,$01,$19,$6C,$67,$0F,$F0,$74,$E5
	.dc.b	$B1,$93,$FC,$9D,$C6,$9D,$46,$1B,$5F,$B1,$6D,$2B,$85,$7F,$A4,$78
	.dc.b	$FF,$D1,$67,$FA,$33,$9A,$C6,$49,$87,$FA,$1A,$32,$FF,$A2,$26,$DF
	.dc.b	$FA,$66,$6D,$7F,$A6,$66,$D7,$FA,$5A,$6F,$1F,$FA,$3A,$6F,$4A,$6C
	.dc.b	$61,$E3,$4F,$98,$FF,$50,$62,$F7,$9E,$FF,$F8,$A2,$73,$18,$CF,$FC
	.dc.b	$33,$18,$9D,$78,$FF,$E4,$EA,$D6,$1F,$E0,$2A,$D5,$F0,$AD,$5F,$08
	.dc.b	$FF,$07,$4A,$7F,$9B,$36,$33,$18,$23,$82,$57,$F8,$90,$F6,$46,$9C
	.dc.b	$4F,$36,$C6,$70,$FF,$12,$4F,$C2,$E7,$7C,$4F,$96,$FD,$57,$6F,$FC
	.dc.b	$05,$37,$0F,$F2,$3E,$1F,$EC,$9B,$F0,$C7,$F8,$36,$75,$1F,$EC,$CE
	.dc.b	$1F,$F6,$7F,$0F,$FB,$41,$B5,$FE,$D9,$6C,$FF,$B2,$9B,$87,$FB,$6D
	.dc.b	$B3,$FE,$DB,$6C,$F5,$8F,$F7,$15,$75,$69,$D4,$B1,$D5,$B5,$D5,$A7
	.dc.b	$52,$C7,$F5,$1F,$86,$FF,$44,$9F,$E8,$AF,$84,$B3,$FE,$83,$8B,$3F
	.dc.b	$E7,$39,$B5,$FE,$7F,$9B,$5F,$B0,$9B,$5D,$63,$91,$FE,$8D,$8C,$D7
	.dc.b	$EB,$EA,$F1,$A6,$C6,$1E,$77,$F9,$DE,$0F,$F3,$F7,$C7,$3C,$BF,$CE
	.dc.b	$46,$1B,$97,$F9,$EA,$74,$FE,$7A,$74,$D3,$53,$C2,$BF,$CF,$B5,$C1
	.dc.b	$FD,$05,$7E,$68,$95,$F9,$E6,$C7,$E5,$1A,$FF,$9F,$36,$FD,$C4,$EC
	.dc.b	$91,$B6,$BB,$9F,$16,$7F,$EE,$6F,$19,$87,$EC,$A3,$F3,$B3,$AB,$FC
	.dc.b	$BB,$F1,$38,$B1,$80,$F9,$13,$EF,$7D,$B1,$AA,$FF,$76,$F4,$7F,$9A
	.dc.b	$FE,$2B,$FC,$D8,$DB,$FF,$36,$36,$FF,$CD,$4D,$AF,$F3,$53,$67,$FC
	.dc.b	$E0,$D9,$FF,$3C,$36,$7B,$CF,$F9,$DE,$BB,$B4,$EE,$5A,$EE,$D9,$EE
	.dc.b	$D3,$DD,$D0,$E8,$70,$7F,$BA,$23,$0D,$8F,$C7,$4F,$FB,$8E,$72,$D8
	.dc.b	$E2,$D9,$E1,$C1,$C5,$91,$93,$84,$63,$0F,$F6,$E5,$F4,$6D,$F4,$6C
	.dc.b	$EF,$05,$FA,$AB,$51,$8A,$5F,$38,$C3,$EF,$BE,$57,$C7,$0B,$7C,$30
	.dc.b	$F7,$BE,$28,$FE,$85,$F2,$DD,$71,$FF,$7C,$7C,$BE,$30,$FC,$44,$61
	.dc.b	$FE,$8B,$3B,$7F,$B6,$CE,$EF,$FD,$EC,$DA,$E1,$1F,$D1,$EB,$83,$FE
	.dc.b	$5F,$9E,$0F,$E9,$63,$FA,$29,$2B,$FA,$4B,$63,$F9,$FB,$5E,$11,$A2
	.dc.b	$3A,$6D,$7F,$42,$9F,$0B,$FD,$F5,$F3,$47,$F3,$E6,$CF,$F9,$CD,$AF
	.dc.b	$F6,$B7,$33,$C6,$7E,$19,$F8,$78,$FC,$65,$7E,$35,$B1,$93,$F1,$91
	.dc.b	$96,$99,$AF,$C2,$C6,$5B,$19,$AF,$BF,$82,$B5,$18,$6C,$FE,$25,$B5
	.dc.b	$F8,$16,$E5,$FA,$56,$DF,$3A,$FF,$74,$7C,$CF,$FB,$AC,$FF,$76,$7C
	.dc.b	$6D,$8C,$93,$0F,$F7,$54,$65,$FF,$76,$4D,$C7,$FD,$D9,$36,$BF,$DE
	.dc.b	$53,$6B,$FD,$F5,$36,$9D,$4A,$C2,$9B,$18,$79,$53,$F0,$C7,$CB,$FE
	.dc.b	$F7,$E1,$BF,$1F,$39,$8C,$65,$F9,$C3,$F9,$24,$65,$FE,$01,$3C,$2B
	.dc.b	$9C,$60,$E1,$3F,$CA,$E3,$2F,$EF,$27,$85,$76,$86,$5F,$B9,$7B,$1F
	.dc.b	$E1,$BA,$E2,$F8,$8C,$95,$CB,$E6,$6B,$E1,$9F,$F7,$04,$62,$38,$9B
	.dc.b	$8C,$92,$FF,$C5,$51,$A3,$93,$EC,$7C,$B5,$F8,$EF,$87,$F5,$CE,$6F
	.dc.b	$DF,$3E,$B6,$C7,$F1,$79,$DD,$7C,$08,$F1,$1C,$6F,$F9,$84,$EA,$FF
	.dc.b	$88,$75,$23,$F6,$AB,$1B,$18,$6D,$FD,$2A,$FF,$3B,$46,$30,$F0,$1F
	.dc.b	$EE,$B7,$11,$D9,$B3,$FE,$84,$6C,$EE,$3A,$5F,$8D,$B3,$AA,$EF,$1A
	.dc.b	$BF,$23,$4D,$96,$FC,$73,$C1,$FE,$83,$F0,$9F,$D4,$4F,$EA,$5F,$07
	.dc.b	$6B,$FA,$84,$59,$FE,$8F,$36,$BF,$80,$CD,$AF,$E8,$D3,$6B,$F9,$F4
	.dc.b	$DE,$1F,$EC,$53,$78,$13,$63,$0D,$BF,$09,$F7,$FC,$4F,$F4,$07,$8D
	.dc.b	$FA,$36,$C6,$4F,$A5,$E1,$FF,$B7,$73,$FC,$24,$EB,$3F,$A6,$7E,$D5
	.dc.b	$C4,$6A,$B8,$DF,$57,$F2,$BC,$1F,$EC,$BB,$FD,$15,$FE,$78,$93,$A6
	.dc.b	$C7,$EA,$DA,$F0,$8D,$11,$D3,$E9,$7E,$AF,$1F,$FA,$9E,$31,$87,$96
	.dc.b	$7D,$37,$FE,$DD,$8F,$EC,$D5,$B9,$CD,$FA,$5B,$7F,$95,$9D,$57,$33
	.dc.b	$FB,$BF,$DF,$BE,$6B,$FC,$25,$5F,$E1,$58,$C3,$FE,$11,$31,$1A,$7F
	.dc.b	$C2,$D1,$AB,$FF,$05,$C6,$8E,$06,$A7,$84,$7F,$84,$A7,$85,$7F,$4A
	.dc.b	$AE,$13,$C9,$E0,$72,$78,$57,$29,$F4,$B6,$7F,$C5,$F5,$D1,$FF,$0C
	.dc.b	$1F,$BA,$D2,$3F,$AC,$D6,$E1,$FA,$4E,$2F,$C9,$CF,$EC,$2B,$F8,$0D
	.dc.b	$6B,$1F,$E0,$B1,$67,$FD,$F7,$36,$FF,$96,$DF,$27,$A5,$72,$AE,$84
	.dc.b	$8C,$BE,$33,$17,$D9,$A7,$62,$C7,$F3,$FF,$95,$FA,$DC,$9F,$D1,$5B
	.dc.b	$3F,$C9,$5B,$3F,$88,$6D,$7F,$27,$6C,$FF,$3B,$6C,$F5,$8F,$D0,$C7
	.dc.b	$F3,$C2,$38,$24,$7F,$49,$8C,$B6,$32,$FF,$49,$8C,$B6,$3C,$C7,$E2
	.dc.b	$BE,$17,$D4,$E2,$3D,$0E,$4C,$57,$A5,$A7,$E0,$27,$8C,$FE,$AB,$E1
	.dc.b	$7F,$C5,$BB,$3E,$12,$C6,$1B,$7F,$A5,$6D,$7E,$C9,$B8,$7E,$B9,$B3
	.dc.b	$FE,$52,$BE,$8F,$EB,$6F,$A3,$4E,$85,$BE,$8D,$BE,$8D,$3A,$5F,$69
	.dc.b	$FE,$EF,$7D,$9F,$D4,$7C,$C7,$ED,$4F,$DB,$7C,$CD,$BF,$D9,$C5,$9F
	.dc.b	$D7,$4D,$C3,$F7,$F3,$6B,$F4,$F3,$6B,$FC,$F7,$C0,$AC,$73,$A1,$58
	.dc.b	$75,$7E,$03,$F8,$1C,$66,$B4,$F5,$B7,$E1,$AF,$96,$FE,$DB,$7F,$E2
	.dc.b	$48,$C3,$FD,$80,$C4,$69,$FF,$02,$46,$A7,$F7,$11,$A3,$81,$A9,$E1
	.dc.b	$1F,$D0,$27,$83,$FE,$1D,$6B,$FE,$30,$6B,$FE,$D5,$6B,$E8,$8F,$F1
	.dc.b	$43,$B6,$BB,$6B,$B9,$FF,$0E,$CF,$22,$4E,$9B,$1E,$73,$FC,$15,$FA
	.dc.b	$A7,$F1,$9A,$E9,$19,$7D,$55,$FE,$88,$9D,$47,$73,$33,$BC,$FF,$D5
	.dc.b	$53,$AB,$F9,$11,$FD,$03,$E6,$FF,$9F,$1F,$CF,$FE,$12,$DF,$F3,$C8
	.dc.b	$B8,$7F,$37,$9B,$5C,$6B,$47,$F4,$28,$CB,$FC,$CA,$AD,$7E,$32,$6E
	.dc.b	$BF,$8B,$9B,$AA,$6C,$61,$CF,$47,$F5,$3F,$9B,$FE,$A8,$DC,$7F,$A8
	.dc.b	$B7,$1F,$E9,$8D,$CB,$FA,$EB,$67,$FA,$2B,$67,$F3,$ED,$CB,$F3,$CD
	.dc.b	$CB,$FB,$75,$7F,$4C,$C3,$C2,$FF,$60,$6D,$FF,$5B,$6E,$1F,$DB,$1F
	.dc.b	$CE,$7C,$5F,$F6,$BF,$87,$B7,$FD,$A1,$B7,$FD,$8D,$B7,$FD,$D1,$B3
	.dc.b	$FD,$71,$B3,$FB,$67,$8B,$F9,$BA,$FC,$61,$27,$F2,$2D,$8F,$F3,$E1
	.dc.b	$FB,$27,$0F,$18,$D9,$1D,$EA,$9E,$57,$FE,$89,$8C,$57,$F8,$0B,$CF
	.dc.b	$F2,$DF,$58,$FD,$BD,$FF,$82,$E7,$35,$C2,$BF,$69,$39,$AD,$9F,$E0
	.dc.b	$CE,$EF,$F0,$91,$FE,$13,$F2,$EC,$64,$95,$FE,$0B,$8C,$BF,$E0,$D9
	.dc.b	$B8,$FF,$87,$A6,$D7,$F8,$1A,$6D,$7F,$20,$9B,$B7,$F1,$F9,$BA,$A6
	.dc.b	$C6,$1C,$F9,$BF,$E2,$7F,$57,$FC,$50,$DC,$7F,$C4,$AD,$C7,$FC,$40
	.dc.b	$DC,$BF,$C6,$2D,$9F,$F0,$CB,$67,$F9,$9B,$72,$FE,$64,$DC,$8D,$25
	.dc.b	$7F,$8D,$30,$F1,$B4,$FE,$53,$E6,$1F,$F6,$4F,$9E,$F5,$95,$FC,$CE
	.dc.b	$B5,$97,$FA,$6A,$B5,$39,$BE,$37,$C1,$EA,$62,$35,$5C,$67,$4F,$F3
	.dc.b	$3B,$EB,$9F,$6C,$3A,$E1,$CA,$BF,$23,$5F,$C9,$C9,$1F,$CB,$1B,$1A
	.dc.b	$8C,$11,$C1,$0F,$E8,$11,$C2,$BF,$34,$F8,$A5,$2A,$F8,$7C,$3E,$AB
	.dc.b	$DD,$79,$5F,$46,$29,$D4,$FF,$4B,$C2,$E1,$CE,$B5,$3C,$4F,$F2,$BF
	.dc.b	$C3,$7F,$9B,$CF,$F3,$87,$7D,$8F,$F3,$6C,$5A,$FF,$34,$4D,$C7,$FC
	.dc.b	$F3,$36,$BF,$CC,$33,$6B,$FB,$DC,$DD,$BF,$BD,$4D,$D9,$36,$30,$DB
	.dc.b	$F0,$9F,$E7,$EF,$9B,$FF,$42,$7C,$5F,$FA,$15,$B1,$93,$FD,$07,$19
	.dc.b	$69,$9C,$7F,$D0,$11,$96,$C6,$72,$FF,$49,$CE,$5B,$19,$3F,$CF,$B1
	.dc.b	$96,$99,$9F,$F0,$CC,$65,$B1,$F7,$27,$F3,$70,$FC,$3D,$F0,$FF,$E9
	.dc.b	$FF,$9C,$FC,$14,$FE,$0A,$73,$7F,$E9,$E9,$CC,$F1,$AE,$13,$D4,$FC
	.dc.b	$4C,$65,$E0,$65,$E9,$1C,$1F,$F6,$9D,$76,$7F,$D4,$95,$DA,$31,$1D
	.dc.b	$0E,$87,$F3,$A7,$2F,$81,$AF,$6A,$FE,$7E,$F0,$39,$46,$9C,$1C,$6B
	.dc.b	$FC,$33,$19,$23,$83,$C8,$7F,$B1,$7E,$62,$D7,$B6,$BC,$B8,$7C,$EB
	.dc.b	$F6,$3F,$EC,$EA,$DC,$FC,$0B,$B3,$9B,$E5,$7C,$1F,$D9,$ED,$F9,$6E
	.dc.b	$FF,$AA,$D8,$FF,$68,$E1,$8D,$7F,$C3,$B1,$C1,$B0,$9E,$0F,$FB,$44
	.dc.b	$E1,$52,$BE,$B4,$EB,$0F,$F7,$2D,$64,$3F,$C9,$23,$53,$DE,$BF,$DD
	.dc.b	$15,$BA,$D5,$F8,$AF,$73,$FE,$6E,$35,$7C,$4F,$0C,$EF,$7F,$E2,$71
	.dc.b	$8C,$3D,$DF,$37,$EF,$F8,$AB,$3C,$1B,$87,$9A,$3E,$14,$63,$2F,$1B
	.dc.b	$63,$27,$82,$BA,$34,$E8,$5A,$E8,$7C,$BC,$55,$CE,$30,$DC,$BF,$86
	.dc.b	$B6,$77,$5C,$F0,$F1,$BF,$80,$F8,$3F,$D8,$F3,$3E,$3D,$C7,$FA,$39
	.dc.b	$CC,$62,$B8,$1F,$D6,$9C,$67,$FA,$C9,$DD,$71,$AD,$36,$7F,$AE,$57
	.dc.b	$0C,$3F,$31,$7F,$A2,$25,$FE,$AB,$74,$62,$FF,$64,$D7,$F5,$AD,$7F
	.dc.b	$68,$D7,$F8,$63,$C6,$BB,$D7,$F7,$2F,$8E,$FF,$27,$D6,$A3,$DA,$FA
	.dc.b	$DD,$78,$3E,$9A,$E3,$39,$3F,$BB,$57,$27,$F8,$F5,$FE,$97,$E5,$26
	.dc.b	$1F,$8E,$F9,$8B,$5F,$73,$83,$FA,$1D,$63,$1F,$CC,$D5,$BF,$D6,$4D
	.dc.b	$BF,$C1,$4D,$BE,$35,$F9,$AA,$FD,$A4,$E7,$74,$58,$C3,$67,$81,$FA
	.dc.b	$2F,$92,$FD,$E7,$CB,$7E,$F5,$B1,$93,$F4,$F1,$96,$99,$C7,$D9,$82
	.dc.b	$75,$18,$6E,$5F,$B6,$6D,$7F,$09,$6C,$FE,$7D,$B9,$7F,$68,$6D,$F5
	.dc.b	$AF,$05,$F3,$7F,$8D,$FB,$7F,$C5,$7C,$3F,$E7,$8D,$8C,$9F,$C0,$23
	.dc.b	$2D,$33,$8F,$AA,$32,$D8,$CB,$FC,$F2,$7F,$9C,$92,$3F,$30,$D8,$FE
	.dc.b	$90,$D7,$84,$68,$8E,$9B,$5F,$CE,$E7,$F9,$77,$C5,$78,$A3,$13,$E6
	.dc.b	$7F,$8F,$C6,$63,$FA,$A4,$DA,$FE,$11,$37,$2F,$F0,$8F,$C7,$7F,$39
	.dc.b	$3F,$9D,$7C,$7D,$C3,$F9,$1C,$5B,$FB,$73,$71,$FE,$5B,$36,$BF,$A0
	.dc.b	$4D,$AF,$E2,$D3,$6F,$FC,$3F,$37,$34,$D8,$C3,$6F,$C2,$7F,$34,$F8
	.dc.b	$AF,$E9,$DF,$17,$FD,$3D,$B1,$93,$F9,$CC,$65,$A6,$71,$FC,$9C,$65
	.dc.b	$B1,$9C,$BF,$A3,$CE,$5B,$19,$3F,$AE,$46,$5A,$66,$7F,$9A,$46,$5B
	.dc.b	$1F,$DD,$7B,$DE,$CF,$F2,$EF,$1D,$E4,$FF,$87,$27,$35,$FE,$84,$9C
	.dc.b	$BC,$6B,$85,$6D,$B1,$93,$84,$65,$A6,$6F,$FD,$A3,$7D,$9B,$5F,$E3
	.dc.b	$47,$FC,$53,$19,$7F,$C4,$0D,$73,$1B,$7F,$CE,$F1,$96,$C6,$54,$65
	.dc.b	$B1,$98,$C1,$1F,$44,$7F,$A7,$EF,$FD,$2F,$8F,$F9,$22,$BD,$0D,$AD
	.dc.b	$BF,$E9,$99,$DD,$F9,$8F,$81,$1B,$7E,$67,$8B,$C1,$FF,$0B,$78,$DF
	.dc.b	$E1,$93,$FC,$35,$F1,$36,$3F,$C0,$71,$72,$FE,$37,$36,$FF,$C1,$B3
	.dc.b	$6B,$FC,$41,$36,$BF,$BB,$4D,$E1,$FF,$3F,$CD,$E1,$4D,$8C,$36,$FC
	.dc.b	$27,$F8,$53,$E0,$3F,$C5,$DF,$0B,$FE,$2F,$6C,$64,$FF,$0C,$C6,$5A
	.dc.b	$67,$1F,$E6,$71,$96,$C6,$72,$FF,$13,$CE,$5B,$19,$3F,$C7,$91,$96
	.dc.b	$99,$9F,$F0,$A4,$65,$B1,$FA,$1F,$3F,$FC,$8D,$F0,$5F,$9F,$6C,$64
	.dc.b	$FF,$18,$C6,$5A,$66,$7F,$A5,$C6,$5B,$19,$7F,$2F,$3F,$8D,$24,$7F
	.dc.b	$9C,$B8,$11,$89,$FD,$1B,$5E,$33,$B8,$C4,$7E,$91,$E1,$87,$E4,$DF
	.dc.b	$F3,$07,$C3,$F8,$E3,$15,$E7,$7F,$CA,$31,$98,$E4,$64,$FF,$33,$39
	.dc.b	$AF,$F1,$E7,$B3,$FE,$6B,$3F,$CD,$9E,$CD,$AF,$F2,$AC,$59,$FE,$DB
	.dc.b	$36,$FF,$CB,$F3,$6B,$FC,$E5,$36,$BA,$C7,$23,$FC,$BF,$19,$E0,$FF
	.dc.b	$69,$D5,$E0,$4D,$8C,$3B,$FC,$01,$F1,$7F,$E8,$47,$FD,$11,$8F,$F9
	.dc.b	$FD,$CE,$3F,$E7,$C7,$38,$9C,$CF,$F8,$85,$C9,$8C,$BF,$CF,$CE,$2F
	.dc.b	$B7,$9B,$DD,$B8,$75,$6D,$F3,$6D,$F2,$6C,$FF,$A7,$2F,$B3,$6F,$A3
	.dc.b	$E7,$6D,$F9,$5F,$E6,$DC,$BF,$E6,$6F,$C0,$9F,$23,$8B,$FF,$67,$79
	.dc.b	$69,$CC,$7F,$41,$AE,$15,$BC,$3F,$CF,$77,$B8,$C6,$6B,$52,$B5,$1C
	.dc.b	$CE,$F8,$76,$9F,$F4,$29,$FC,$C3,$0F,$E5,$C4,$9F,$D1,$36,$3F,$A1
	.dc.b	$B5,$FE,$7E,$D7,$FA,$1B,$5F,$E7,$EF,$1B,$FE,$68,$FF,$A2,$3F,$03
	.dc.b	$70,$7A,$AF,$89,$E7,$BF,$7C,$EA,$31,$7C,$9A,$7C,$0C,$53,$BA,$F6
	.dc.b	$F1,$B1,$FC,$87,$CA,$4C,$3F,$1D,$F2,$B6,$FF,$19,$37,$2F,$C5,$CD
	.dc.b	$AF,$C1,$4D,$AF,$CC,$4D,$E1,$FC,$B4,$DE,$04,$D8,$C3,$6F,$F3,$D1
	.dc.b	$FA,$38,$E1,$F2,$9F,$A5,$E0,$FD,$33,$6B,$F4,$8D,$C7,$F4,$4D,$CB
	.dc.b	$F4,$2D,$9F,$CC,$36,$7F,$5C,$DC,$BF,$58,$DC,$8D,$25,$7E,$CF,$0F
	.dc.b	$1B,$4F,$F0,$B5,$FE,$EF,$E4,$FF,$C2,$8D,$8C,$9F,$BB,$8C,$B4,$CC
	.dc.b	$7E,$EE,$32,$D8,$CB,$FE,$0C,$3F,$C0,$44,$7F,$9B,$34,$FF,$0B,$94
	.dc.b	$E0,$E8,$8E,$9B,$8F,$F8,$1E,$3F,$8A,$79,$7E,$28,$C4,$F9,$9F,$E1
	.dc.b	$B1,$98,$FF,$10,$CD,$3F,$8D,$45,$CB,$F8,$C7,$C9,$FF,$1F,$3F,$90
	.dc.b	$7C,$8D,$BF,$E3,$D1,$6F,$F8,$E4,$DC,$3F,$8D,$CD,$AF,$E2,$53,$6B
	.dc.b	$F9,$44,$DB,$FE,$4D,$37,$44,$D8,$C3,$6F,$C2,$7F,$2F,$F9,$3F,$E6
	.dc.b	$9F,$27,$FC,$D5,$B1,$93,$F9,$A4,$65,$A6,$71,$FE,$69,$19,$6C,$67
	.dc.b	$2F,$E6,$B3,$96,$C6,$4F,$E5,$D1,$96,$99,$9F,$E8,$D1,$96,$C7,$F2
	.dc.b	$09,$FE,$45,$1F,$C8,$BF,$10,$79,$DF,$D3,$FE,$53,$35,$AA,$56,$AF
	.dc.b	$FA,$85,$69,$FE,$57,$3A,$BE,$51,$DC,$CD,$71,$9F,$EA,$95,$C5,$F1
	.dc.b	$57,$07,$F9,$7C,$EE,$FF,$AB,$D6,$E7,$B5,$F8,$1E,$B7,$FE,$65,$BE
	.dc.b	$D1,$96,$BA,$9E,$0F,$81,$AF,$43,$31,$C1,$D1,$89,$EC,$49,$E3,$3E
	.dc.b	$77,$B4,$FF,$71,$F9,$4F,$44,$E6,$39,$E1,$F3,$63,$1B,$7F,$79,$AD
	.dc.b	$E1,$30,$F9,$58,$70,$7F,$B7,$4E,$63,$FB,$F7,$CA,$7F,$81,$CF,$F0
	.dc.b	$47,$9F,$6B,$FC,$0D,$16,$7F,$C0,$93,$72,$FF,$01,$CD,$AF,$EF,$53
	.dc.b	$6B,$FC,$21,$37,$1F,$F0,$74,$DE,$04,$D8,$C3,$6F,$C2,$7F,$85,$FE
	.dc.b	$4F,$FC,$39,$F2,$9F,$E1,$D6,$C6,$4F,$F0,$E4,$65,$A6,$71,$FF,$0E
	.dc.b	$46,$5B,$19,$CB,$FC,$3B,$39,$6C,$64,$FF,$0B,$46,$5A,$66,$7F,$C5
	.dc.b	$D1,$96,$C7,$D2,$7F,$47,$E4,$FF,$8D,$7E,$47,$EE,$B6,$75,$BA,$31
	.dc.b	$5F,$9C,$6E,$1F,$81,$6D,$7E,$99,$B3,$FB,$57,$F3,$84,$8F,$C7,$B6
	.dc.b	$3F,$2E,$D7,$F4,$0D,$7A,$47,$02,$3C,$1F,$3C,$7F,$93,$FE,$43,$D7
	.dc.b	$18,$24,$FF,$92,$E3,$2F,$AE,$33,$5F,$E4,$C9,$CB,$FC,$3B,$67,$E0
	.dc.b	$5F,$EC,$EA,$FF,$18,$C6,$3C,$EF,$F2,$F4,$7F,$03,$8C,$3F,$A0,$6E
	.dc.b	$0B,$E9,$7C,$E3,$FC,$C9,$5C,$63,$15,$C2,$B7,$7D,$AB,$A5,$68,$91
	.dc.b	$9B,$ED,$3E,$69,$CB,$C2,$75,$5D,$23,$04,$AF,$F3,$83,$E3,$9F,$27
	.dc.b	$C8,$79,$7E,$46,$CF,$89,$B5,$E0,$18,$FC,$88,$CE,$1F,$01,$B5,$E5
	.dc.b	$6D,$70,$C5,$18,$BE,$55,$F0,$30,$F1,$34,$F1,$16,$FC,$53,$C1,$B6
	.dc.b	$AE,$31,$86,$E1,$FC,$6F,$C9,$FE,$79,$E2,$3D,$27,$67,$F2,$29,$CC
	.dc.b	$63,$3F,$D3,$C6,$27,$51,$FE,$80,$9D,$63,$FD,$12,$B5,$3C,$0D,$57
	.dc.b	$09,$FE,$73,$3F,$D0,$C9,$3F,$CD,$9B,$1A,$8C,$11,$C1,$0F,$E9,$51
	.dc.b	$C3,$0F,$74,$FF,$A5,$3E,$5F,$FD,$3F,$5F,$DA,$EB,$C5,$5A,$8F,$F4
	.dc.b	$9C,$E9,$FE,$D7,$3A,$AF,$EC,$AF,$C8,$9D,$70,$A3,$F3,$9F,$23,$2B
	.dc.b	$F3,$3F,$23,$70,$FA,$B3,$71,$FD,$2C,$DA,$FC,$D4,$DA,$FC,$34,$DD
	.dc.b	$FF,$6B,$37,$74,$D8,$C3,$6F,$F5,$31,$FA,$F8,$E1,$F2,$1F,$B1,$F9
	.dc.b	$0F,$D9,$36,$BF,$5E,$DC,$7F,$16,$DC,$BF,$76,$D9,$FD,$83,$67,$F3
	.dc.b	$8D,$CB,$F8,$5B,$72,$34,$95,$FB,$EC,$3C,$6D,$3F,$9D,$7E,$22,$F1
	.dc.b	$3F,$84,$7A,$5F,$E0,$D8,$C4,$FF,$06,$8C,$4E,$A3,$8B,$72,$FF,$62
	.dc.b	$D7,$07,$F4,$F5,$C2,$B9,$1C,$0F,$F0,$7C,$EE,$7F,$C3,$4D,$7F,$C5
	.dc.b	$6D,$7F,$DA,$ED,$4D,$25,$7F,$8D,$E3,$8B,$5E,$2D,$78,$B6,$38,$CF
	.dc.b	$96,$7F,$7F,$F0,$5E,$97,$DB,$57,$4F,$E6,$33,$6B,$B9,$C6,$32,$D3
	.dc.b	$F9,$57,$C8,$FF,$35,$3F,$9B,$7C,$75,$8F,$E6,$71,$70,$FE,$1F,$36
	.dc.b	$BF,$9F,$CD,$AF,$E6,$93,$6B,$F9,$04,$DD,$BF,$A7,$CD,$D9,$36,$30
	.dc.b	$DB,$F0,$9F,$D1,$3D,$BF,$E9,$5F,$21,$FD,$2D,$B1,$93,$FA,$4C,$65
	.dc.b	$A6,$71,$FE,$4F,$19,$6C,$67,$2F,$EB,$33,$96,$C6,$4F,$EA,$11,$96
	.dc.b	$99,$9F,$E7,$D1,$96,$C7,$E7,$3C,$4F,$EC,$5E,$E7,$E6,$DB,$19,$3F
	.dc.b	$B0,$C6,$5A,$66,$7F,$9F,$C6,$5B,$19,$7F,$2F,$3F,$89,$24,$7E,$71
	.dc.b	$B1,$FE,$6E,$DD,$18,$8E,$33,$B2,$3B,$8E,$38,$7F,$A4,$6B,$8D,$7F
	.dc.b	$78,$F4,$3C,$93,$9A,$B3,$FD,$DE,$6D,$FE,$A2,$6E,$DF,$E1,$29,$BB
	.dc.b	$3D,$3F,$F0,$41,$FE,$09,$F8,$FB,$8F,$F8,$1A,$2E,$3F,$D8,$A6,$E5
	.dc.b	$FE,$12,$9B,$5F,$E0,$79,$B5,$FD,$CE,$6E,$3F,$E2,$29,$B8,$78,$6F
	.dc.b	$B1,$FE,$14,$8C,$D6,$9D,$FC,$4F,$F8,$77,$E3,$FF,$C3,$CD,$AF,$F0
	.dc.b	$DB,$71,$FE,$EE,$DC,$BF,$C4,$ED,$9F,$F0,$E3,$67,$FC,$10,$DC,$BF
	.dc.b	$C6,$ED,$C8,$D2,$57,$F8,$AB,$0F,$1B,$4F,$E5,$A7,$F8,$CF,$D5,$FE
	.dc.b	$58,$D3,$27,$F8,$C9,$CB,$4C,$E1,$FE,$10,$8C,$B6,$32,$FF,$28,$7F
	.dc.b	$8D,$10,$FE,$5A,$57,$F2,$4D,$78,$3A,$23,$A6,$E3,$F9,$17,$FC,$99
	.dc.b	$F1,$FE,$28,$C4,$79,$9F,$F2,$9C,$66,$3F,$9C,$96,$7F,$C6,$91,$72
	.dc.b	$7E,$04,$F8,$FF,$F3,$03,$6F,$FC,$AB,$78,$99,$1F,$E6,$39,$D4,$FF
	.dc.b	$8D,$0C,$9F,$E3,$87,$38,$7F,$9C,$5B,$3F,$E5,$B9,$DC,$E6,$AC,$E4
	.dc.b	$EC,$E2,$7F,$CA,$75,$C9,$FF,$27,$19,$BF,$5C,$DD,$13,$9A,$7C,$7F
	.dc.b	$BB,$E3,$FD,$ED,$C7,$D8,$DA,$F5,$37,$2F,$53,$67,$FD,$13,$7D,$1F
	.dc.b	$89,$7D,$1A,$74,$2D,$F4,$6D,$F4,$69,$D2,$FB,$4F,$C0,$6A,$FC,$4B
	.dc.b	$FC,$1F,$3F,$EF,$4F,$19,$F6,$1F,$E5,$6C,$13,$F0,$AF,$87,$81,$FE
	.dc.b	$64,$AD,$4E,$70,$E7,$5F,$E8,$1C,$39,$F8,$DC,$A3,$FD,$08,$72,$AF
	.dc.b	$05,$F2,$8F,$9F,$7C,$B1,$E1,$87,$2B,$FF,$6C,$4E,$5F,$AB,$19,$AE
	.dc.b	$B3,$E2,$7D,$07,$72,$94,$CB,$B7,$D5,$8F,$F8,$36,$31,$1E,$48,$C1
	.dc.b	$2B,$9C,$73,$38,$B8,$9E,$31,$E1,$BF,$84,$D9,$F4,$FA,$9F,$E6,$57
	.dc.b	$DB,$7F,$1E,$B7,$7E,$FB,$F1,$57,$D7,$BE,$11,$F0,$2F,$84,$F8,$A3
	.dc.b	$0E,$FC,$5C,$C7,$23,$F5,$DF,$1B,$FB,$8F,$8E,$FD,$CB,$6B,$F6,$CD
	.dc.b	$9D,$C7,$EE,$67,$6D,$36,$58,$D9,$FB,$B8,$DB,$63,$6D,$37,$97,$EF
	.dc.b	$E7,$6D,$9D,$B6,$7F,$83,$7A,$DF,$C4,$0F,$E2,$3F,$1D,$6B,$F8,$74
	.dc.b	$59,$FE,$19,$36,$BF,$86,$4D,$AF,$D9,$CD,$AF,$E0,$F3,$79,$7F,$8F
	.dc.b	$CD,$E3,$4D,$8C,$36,$FC,$27,$E6,$3B,$FF,$92,$FC,$77,$D2,$C7,$F6
	.dc.b	$51,$89,$D1,$FC,$96,$75,$C0,$9D,$57,$E7,$9B,$9F,$E1,$9B,$7F,$B3
	.dc.b	$7F,$34,$47,$F2,$8D,$9F,$D1,$D7,$59,$C9,$23,$33,$B7,$F4,$ED,$7B
	.dc.b	$47,$A1,$FE,$71,$F1,$DE,$89,$93,$9B,$FE,$75,$38,$AF,$54,$E4,$FE
	.dc.b	$47,$19,$9F,$E8,$1F,$1B,$FD,$33,$E3,$BF,$A6,$B6,$BF,$A5,$36,$FF
	.dc.b	$A4,$36,$B8,$CF,$F4,$FA,$E2,$D3,$89,$6B,$8B,$67,$8C,$7F,$53,$9E
	.dc.b	$2D,$9E,$2D,$38,$96,$7F,$AC,$7C,$77,$F6,$43,$FB,$2F,$A3,$6B,$FB
	.dc.b	$14,$5A,$FE,$C1,$36,$B8,$D6,$8F,$ED,$11,$97,$FA,$4D,$5A,$FE,$BB
	.dc.b	$37,$83,$FB,$AC,$DE,$04,$D8,$C3,$C6,$BF,$A2,$76,$FF,$7A,$F5,$BF
	.dc.b	$22,$DC,$7F,$98,$B6,$7F,$97,$B7,$2F,$E2,$ED,$9F,$E8,$8D,$9F,$E7
	.dc.b	$AF,$F2,$D2,$47,$F2,$76,$C7,$E5,$1A,$FF,$36,$6B,$D2,$38,$11,$E0
	.dc.b	$F9,$8F,$F0,$77,$C7,$7A,$9C,$47,$A1,$C9,$8A,$F4,$B4,$FE,$EB,$3C
	.dc.b	$67,$FC,$29,$F1,$DF,$E1,$EF,$8E,$FF,$0F,$B6,$BF,$C3,$8D,$9D,$C7
	.dc.b	$F8,$7E,$76,$D3,$65,$8D,$9F,$E2,$18,$DB,$63,$6D,$37,$7F,$E2,$59
	.dc.b	$DB,$67,$6D,$9F,$F1,$67,$C6,$FF,$8D,$8F,$F1,$BF,$C7,$5A,$FF,$1A
	.dc.b	$45,$9F,$F1,$94,$DA,$FF,$19,$4D,$AF,$F0,$CC,$DA,$FF,$16,$CD,$DF
	.dc.b	$FC,$8B,$37,$74,$D8,$C3,$6F,$C2,$7F,$8C,$30,$FE,$CF,$ED,$BA,$3F
	.dc.b	$C9,$BF,$1D,$FC,$A6,$31,$5F,$E4,$B8,$C4,$EA,$39,$1F,$E1,$99,$DE
	.dc.b	$7F,$E4,$DA,$DC,$F1,$37,$1F,$E4,$09,$DD,$FF,$94,$6B,$73,$FE,$0A
	.dc.b	$9E,$6F,$F8,$39,$AF,$7B,$FE,$FE,$DB,$FF,$03,$CE,$A3,$B4,$E9,$B3
	.dc.b	$A8,$C4,$27,$83,$63,$31,$83,$CF,$3D,$E7,$FC,$D5,$D9,$E8,$C3,$E5
	.dc.b	$DD,$AF,$F2,$8C,$DA,$F0,$9F,$E3,$6B,$E6,$FF,$9A,$23,$5A,$3D,$E7
	.dc.b	$C7,$7F,$83,$23,$1F,$1D,$67,$38,$FC,$C9,$CB,$63,$38,$33,$FF,$23
	.dc.b	$56,$A3,$15,$FE,$72,$8F,$06,$1D,$70,$EC,$D9,$E8,$DA,$E2,$2B,$8C
	.dc.b	$F8,$6B,$8B,$4E,$25,$9E,$33,$C1,$B8,$2B,$8C,$61,$B8,$7A,$A7,$57
	.dc.b	$EF,$F8,$CF,$88,$7C,$5F,$8C,$B1,$92,$61,$DA,$32,$F8,$26,$DF,$59
	.dc.b	$B5,$F1,$6D,$58,$BF,$81,$0A,$C6,$3F,$66,$AC,$E6,$FC,$38,$7A,$67
	.dc.b	$37,$CF,$0F,$C9,$74,$7F,$92,$BC,$47,$C9,$BD,$1F,$DA,$4E,$63,$19
	.dc.b	$7E,$B2,$F8,$BF,$E8,$F8,$CC,$7E,$3E,$78,$57,$2A,$D3,$67,$FD,$19
	.dc.b	$5C,$32,$FE,$0B,$7F,$A2,$25,$FE,$B9,$B5,$FB,$76,$BF,$B7,$6B,$FC
	.dc.b	$15,$AF,$F0,$17,$8C,$F6,$9F,$F5,$37,$7B,$D6,$F2,$4E,$6F,$D5,$5A
	.dc.b	$9C,$BD,$4F,$85,$C4,$C3,$E5,$61,$CE,$73,$1F,$A9,$F8,$DF,$D5,$7C
	.dc.b	$6D,$9F,$D4,$37,$1F,$C9,$B7,$2F,$C8,$36,$7F,$08,$D9,$FC,$1B,$78
	.dc.b	$7F,$74,$DE,$13,$49,$5F,$B7,$C3,$C6,$D3,$C6,$7E,$EF,$E3,$7F,$80
	.dc.b	$1F,$C0,$7D,$CB,$19,$26,$3F,$AA,$8C,$BF,$AC,$9B,$97,$EA,$66,$D7
	.dc.b	$E7,$66,$D7,$E7,$26,$E5,$FC,$4E,$6E,$49,$B1,$87,$8D,$7F,$2B,$E8
	.dc.b	$FE,$2D,$E6,$FF,$4B,$6E,$3F,$59,$B3,$FC,$8D,$B9,$7F,$54,$6C,$FF
	.dc.b	$4D,$6C,$FF,$49,$7F,$9A,$92,$3F,$A0,$B6,$3F,$A5,$35,$FE,$94,$D7
	.dc.b	$A4,$70,$23,$C1,$F3,$1F,$CA,$FE,$37,$D4,$E2,$3D,$0E,$4C,$57,$A5
	.dc.b	$A7,$F0,$99,$E3,$3F,$C7,$BE,$37,$F9,$D7,$C6,$FF,$3B,$6D,$7F,$39
	.dc.b	$6C,$EE,$3F,$9E,$4E,$DA,$6C,$B1,$B3,$F9,$5C,$6D,$B1,$B6,$9B,$BF
	.dc.b	$E5,$93,$B6,$CE,$DB,$3F,$CD,$BD,$AF,$EA,$47,$F5,$3F,$42,$D7,$F5
	.dc.b	$18,$B3,$FC,$F2,$6D,$7F,$39,$9B,$5F,$CB,$66,$D7,$F2,$C9,$BB,$FF
	.dc.b	$63,$9B,$BA,$6C,$61,$B7,$E1,$3F,$92,$7E,$26,$E1,$FE,$D1,$F1,$BF
	.dc.b	$EC,$58,$C4,$FF,$67,$8C,$4E,$A3,$8B,$72,$FE,$BE,$2B,$8B,$FD,$62
	.dc.b	$B8,$D7,$33,$88,$88,$49,$E2,$E0,$E9,$5F,$E1,$48,$CC,$6E,$30,$49
	.dc.b	$E1,$1D,$67,$6D,$9D,$3A,$8C,$BD,$A7,$8B,$4F,$6C,$7B,$33,$FE,$DD
	.dc.b	$F1,$BE,$C9,$CD,$D9,$CD,$FB,$A7,$2D,$8C,$C6,$0F,$EF,$BF,$1B,$FE
	.dc.b	$19,$D3,$FC,$34,$DA,$FF,$0C,$36,$77,$1F,$E1,$B9,$DB,$4D,$96,$36
	.dc.b	$7F,$82,$A3,$6D,$8D,$B4,$DD,$7F,$82,$E7,$6D,$9D,$B6,$7F,$C2,$BD
	.dc.b	$1F,$E3,$13,$FC,$65,$F1,$B6,$BF,$C6,$11,$67,$FC,$37,$36,$BF,$C3
	.dc.b	$13,$6B,$FC,$19,$36,$BF,$C1,$73,$77,$FF,$21,$4D,$C3,$FC,$AD,$7D
	.dc.b	$8F,$F1,$F4,$66,$B4,$F3,$BF,$CB,$6F,$FE,$47,$AF,$D0,$35,$FD,$2B
	.dc.b	$63,$F0,$2D,$C5,$3F,$97,$8C,$34,$FF,$27,$1D,$0F,$C5,$EA,$8F,$CD
	.dc.b	$12,$32,$DA,$FC,$A3,$5F,$D1,$4F,$19,$D9,$23,$6D,$AF,$2C,$7E,$86
	.dc.b	$BF,$CC,$DF,$81,$BE,$3B,$C9,$3E,$CB,$F7,$9F,$E6,$79,$D3,$88,$D1
	.dc.b	$2F,$FC,$CF,$3A,$8C,$4F,$F9,$9E,$1F,$1B,$E0,$F8,$EF,$0B,$67,$B3
	.dc.b	$6B,$FC,$BF,$5C,$27,$C5,$5C,$1A,$70,$2C,$F0,$AE,$57,$E4,$6C,$F8
	.dc.b	$DB,$3F,$E7,$3B,$EC,$F8,$2F,$B3,$6B,$B3,$4F,$2F,$C6,$7C,$A3,$E5
	.dc.b	$FC,$65,$9F,$91,$16,$7D,$73,$71,$F8,$F3,$6B,$E3,$4D,$AF,$24,$DE
	.dc.b	$06,$2A,$C2,$9B,$18,$6D,$F8,$8F,$E7,$BB,$FF,$99,$FE,$51,$F0,$5F
	.dc.b	$CF,$63,$0D,$9F,$F6,$8F,$8A,$AD,$C7,$23,$2E,$33,$FE,$47,$7C,$A7
	.dc.b	$51,$FE,$A0,$AE,$57,$D6,$3F,$97,$57,$F3,$92,$57,$E9,$1B,$1A,$8C
	.dc.b	$11,$C1,$0F,$D1,$C7,$0D,$3F,$3F,$1F,$D4,$63,$FD,$0D,$F2,$6F,$8B
	.dc.b	$F1,$56,$B8,$FF,$83,$DF,$AB,$0F,$83,$7B,$7D,$75,$BC,$38,$C7,$F5
	.dc.b	$7C,$95,$BB,$E4,$7F,$62,$FC,$47,$F4,$8F,$F5,$96,$BF,$D7,$9A,$FF
	.dc.b	$61,$6A,$69,$38,$FF,$DE,$6D,$BF,$EA,$AD,$9F,$EB,$EF,$0E,$2D,$92
	.dc.b	$76,$D9,$DB,$5F,$33,$5F,$33,$5F,$33,$E2,$E3,$93,$E2,$6C,$F8,$9A
	.dc.b	$FC,$16,$BF,$05,$AF,$C1,$6B,$F0,$78,$3F,$8B,$FC,$7F,$F1,$AD,$7E
	.dc.b	$EE,$3F,$A1,$C3,$F8,$B4,$E5,$F9,$95,$C1,$FF,$60,$4E,$EF,$F8,$6D
	.dc.b	$6D,$FA,$15,$B7,$EC,$36,$FB,$BF,$40,$97,$F4,$1B,$3F,$41,$AF,$DF
	.dc.b	$6B,$E4,$8E,$04,$78,$3E,$68,$FE,$6B,$F0,$3E,$59,$F8,$53,$92,$3F
	.dc.b	$C9,$A3,$58,$7B,$6F,$CF,$5A,$7F,$9A,$CE,$AF,$7F,$1F,$FD,$08,$B5
	.dc.b	$FC,$FD,$B8,$7F,$32,$6D,$FF,$3E,$6C,$FF,$3D,$6C,$FF,$2B,$78,$BF
	.dc.b	$9F,$7F,$2E,$48,$FC,$BB,$63,$F2,$ED,$78,$46,$88,$E9,$B3,$F8,$29
	.dc.b	$FE,$B7,$F1,$FE,$2A,$FD,$44,$79,$39,$FF,$94,$CE,$72,$4E,$79,$D3
	.dc.b	$9F,$2B,$FB,$2F,$2F,$CC,$6E,$7F,$8A,$6E,$5F,$BD,$6E,$7F,$94,$6C
	.dc.b	$FE,$F5,$B3,$FB,$D7,$F7,$24,$8F,$DC,$B6,$3F,$72,$D7,$F8,$43,$5E
	.dc.b	$91,$C0,$8F,$07,$CC,$7F,$79,$8F,$53,$88,$F4,$39,$31,$3E,$96,$9F
	.dc.b	$DE,$67,$8C,$FF,$1B,$F2,$BF,$C1,$3F,$1D,$F9,$06,$D7,$EB,$9B,$3B
	.dc.b	$8F,$F0,$54,$ED,$A6,$CB,$1B,$3F,$8C,$CF,$F1,$E2,$47,$F1,$E6,$C6
	.dc.b	$A3,$04,$70,$43,$F9,$84,$70,$BF,$E1,$2F,$F8,$8B,$E3,$FC,$93,$E1
	.dc.b	$9C,$D7,$F8,$86,$73,$18,$9D,$4F,$F8,$76,$72,$FF,$3B,$F8,$EF,$F1
	.dc.b	$6D,$FE,$D1,$B1,$FC,$49,$B5,$FD,$15,$B7,$FC,$71,$B3,$CC,$FF,$19
	.dc.b	$57,$37,$F9,$FC,$7F,$42,$23,$82,$57,$F4,$28,$CB,$63,$2F,$F5,$38
	.dc.b	$CB,$63,$CC,$7F,$90,$7C,$FF,$04,$79,$E3,$2F,$F8,$DA,$33,$7E,$99
	.dc.b	$CB,$FE,$44,$8C,$C7,$F5,$FE,$BF,$F2,$8F,$C7,$7F,$18,$6C,$7F,$36
	.dc.b	$6D,$FF,$69,$6D,$FF,$40,$6C,$F3,$3F,$CA,$D5,$CD,$FE,$C7,$1F,$D9
	.dc.b	$88,$E0,$95,$FD,$9A,$32,$D8,$CB,$FD,$DE,$32,$D8,$F3,$1F,$E6,$5C
	.dc.b	$3C,$11,$E7,$8C,$BF,$E5,$A8,$CD,$FA,$67,$2F,$F9,$A2,$33,$1F,$EE
	.dc.b	$0F,$C5,$1F,$1B,$FE,$76,$F8,$EF,$E8,$51,$89,$FF,$3A,$46,$27,$51
	.dc.b	$FE,$E9,$BF,$F1,$0D,$6F,$0F,$F3,$5D,$EE,$B8,$9B,$3C,$17,$CA,$7F
	.dc.b	$DF,$35,$C5,$FF,$78,$35,$DC,$FF,$B9,$23,$11,$FE,$F9,$6A,$69,$27
	.dc.b	$FC,$6B,$1C,$8F,$F7,$AC,$66,$7A,$4E,$A7,$A3,$E9,$7F,$D0,$BF,$1F
	.dc.b	$70,$F7,$CD,$AF,$F4,$84,$DA,$E3,$1E,$F6,$9C,$9C,$3E,$79,$F9,$FF
	.dc.b	$15,$FE,$A6,$CE,$D7,$FA,$8D,$B5,$FE,$96,$6E,$1F,$EA,$26,$CF,$FA
	.dc.b	$85,$B3,$FE,$8F,$78,$BF,$32,$BD,$CF,$C7,$6E,$26,$92,$BE,$83,$67
	.dc.b	$E4,$B5,$E9,$1A,$23,$F7,$23,$FD,$83,$F0,$5E,$48,$F0,$4F,$C2,$AF
	.dc.b	$03,$FE,$C1,$9D,$46,$27,$94,$E4,$E1,$5F,$96,$F8,$AF,$F6,$6C,$D9
	.dc.b	$FF,$65,$34,$FF,$5E,$97,$1F,$F6,$4B,$63,$FD,$90,$D3,$FD,$6C,$71
	.dc.b	$3F,$27,$3F,$18,$91,$F9,$36,$C7,$E3,$1A,$F0,$8D,$11,$D3,$67,$F5
	.dc.b	$33,$FE,$45,$E9,$7E,$09,$FA,$B9,$CD,$7C,$7A,$F3,$56,$B4,$F4,$61
	.dc.b	$FA,$19,$CE,$1C,$6F,$F7,$D1,$FB,$CF,$87,$ED,$2F,$8F,$C5,$7E,$5F
	.dc.b	$1F,$0C,$63,$06,$5F,$EF,$C9,$CD,$7D,$9C,$39,$DE,$EB,$58,$75,$AD
	.dc.b	$1D,$5C,$61,$FC,$02,$FF,$22,$49,$FD,$F3,$63,$F6,$ED,$7F,$82,$B5
	.dc.b	$FE,$34,$D7,$F8,$23,$C6,$BF,$8A,$BF,$03,$E3,$FC,$D5,$C0,$F2,$D7
	.dc.b	$B9,$2C,$7E,$9C,$F0,$8E,$B3,$C2,$FD,$F3,$99,$FE,$25,$ED,$7C,$ED
	.dc.b	$7F,$95,$B6,$FF,$40,$D9,$FE,$42,$DB,$FE,$56,$D9,$FE,$6C,$D9,$FE
	.dc.b	$40,$FE,$E0,$91,$FC,$95,$B1,$FC,$79,$AF,$F2,$C6,$BD,$23,$81,$1E
	.dc.b	$0F,$98,$FB,$1F,$1D,$EA,$71,$1E,$87,$26,$2B,$D2,$D3,$E5,$E3,$F9
	.dc.b	$CA,$E5,$5F,$D0,$3E,$1B,$F2,$7F,$07,$FD,$41,$B7,$FC,$15,$B3,$B8
	.dc.b	$FC,$54,$ED,$A6,$CB,$1B,$3F,$9F,$CF,$F1,$E2,$47,$F4,$B6,$C6,$A3
	.dc.b	$04,$70,$43,$FA,$AC,$70,$C3,$FB,$0B,$FA,$CD,$FC,$93,$E1,$9C,$D7
	.dc.b	$E8,$67,$31,$89,$D4,$FC,$6D,$95,$A8,$FE,$ED,$F0,$9F,$B8,$F8,$EF
	.dc.b	$EE,$AD,$8F,$E5,$4D,$BF,$ED,$6D,$C3,$FB,$AB,$67,$99,$FB,$3A,$E6
	.dc.b	$FF,$6D,$8F,$E8,$64,$70,$4A,$FE,$D3,$19,$6C,$65,$FE,$F3,$19,$6C
	.dc.b	$79,$8F,$E0,$5F,$1B,$E0,$8F,$3C,$65,$FD,$91,$FB,$1C,$53,$BC,$3D
	.dc.b	$75,$B7,$F8,$7C,$66,$3F,$73,$4F,$71,$F0,$9F,$C8,$38,$FF,$DB,$D1
	.dc.b	$8A,$FE,$29,$18,$9D,$47,$F7,$69,$D5,$79,$38,$3F,$9A,$DE,$EB,$89
	.dc.b	$BB,$E3,$5E,$A3,$C4,$FF,$70,$BE,$71,$E6,$33,$35,$FE,$23,$1A,$7D
	.dc.b	$6D,$73,$3C,$C9,$3C,$63,$99,$CE,$78,$38,$3C,$0E,$0F,$6E,$29,$EF
	.dc.b	$5F,$D1,$BE,$37,$DD,$7E,$BB,$DD,$68,$FE,$79,$3A,$BE,$55,$E6,$9F
	.dc.b	$88,$68,$AF,$3F,$85,$FE,$A7,$F1,$BF,$51,$B3,$F4,$5B,$3B,$8F,$E9
	.dc.b	$53,$B6,$9B,$2C,$6C,$F8,$53,$F4,$49,$1F,$1F,$04,$62,$77,$39,$8C
	.dc.b	$47,$C8,$6B,$F5,$1E,$17,$F0,$23,$F2,$11,$FD,$BB,$93,$CD,$1E,$39
	.dc.b	$D5,$F8,$2F,$FB,$25,$6A,$73,$76,$7F,$82,$4D,$BF,$C2,$FC,$17,$F7
	.dc.b	$AF,$8D,$FC,$DB,$6B,$F3,$0D,$9F,$B0,$DC,$3C,$2D,$9F,$C6,$B6,$7A
	.dc.b	$C7,$C5,$8F,$CC,$11,$C1,$23,$F1,$D1,$96,$C6,$5F,$C7,$C6,$5B,$1E
	.dc.b	$63,$FC,$09,$F1,$BE,$A7,$11,$E8,$72,$62,$BD,$2D,$3F,$94,$4F,$19
	.dc.b	$FC,$B7,$C1,$FF,$87,$B9,$BF,$7E,$DA,$FD,$E3,$67,$71,$FE,$15,$9D
	.dc.b	$B4,$D9,$63,$67,$EB,$67,$F7,$84,$8F,$C6,$36,$35,$18,$23,$82,$1F
	.dc.b	$B4,$8E,$17,$F8,$67,$FC,$6F,$F1,$BE,$49,$F0,$CE,$6B,$FC,$59,$39
	.dc.b	$8C,$4E,$A7,$FA,$64,$E5,$FE,$17,$ED,$FF,$90,$BE,$33,$F9,$1B,$63
	.dc.b	$F9,$03,$6B,$F8,$03,$6F,$F3,$2D,$9E,$67,$F8,$F6,$B9,$BF,$AD,$8F
	.dc.b	$E4,$04,$70,$4A,$FE,$2D,$19,$6C,$65,$FE,$2F,$19,$6C,$79,$8F,$F2
	.dc.b	$67,$3F,$82,$3C,$F1,$97,$F9,$C7,$3A,$75,$87,$AA,$B4,$FF,$96,$23
	.dc.b	$31,$FC,$C3,$E0,$BF,$9C,$11,$FE,$10,$DB,$FE,$5C,$D7,$F9,$6B,$5E
	.dc.b	$51,$C0,$8F,$F3,$5F,$82,$FE,$88,$48,$FE,$28,$D8,$D4,$60,$8E,$08
	.dc.b	$7F,$3C,$8E,$0F,$F3,$EF,$80,$FE,$A0,$49,$FE,$40,$D8,$FE,$8C,$D7
	.dc.b	$84,$68,$8E,$9A,$FF,$4A,$F8,$1F,$EB,$84,$8F,$E5,$8D,$8F,$EA,$4D
	.dc.b	$78,$46,$88,$E9,$AF,$F5,$6F,$81,$FE,$D0,$48,$FE,$70,$D8,$FE,$BC
	.dc.b	$D7,$84,$68,$8E,$9A,$FF,$62,$F7,$BF,$BA,$12,$3F,$A2,$36,$3F,$B5
	.dc.b	$35,$E1,$1A,$23,$A6,$BF,$DF,$7F,$15,$7B,$BF,$EB,$0E,$9F,$EC,$BB
	.dc.b	$7F,$47,$9C,$DF,$FA,$96,$72,$FF,$86,$A3,$37,$F9,$A9,$CB,$FE,$96
	.dc.b	$8C,$D7,$FA,$C2,$73,$29,$CD,$7F,$93,$67,$31,$89,$FF,$17,$46,$2F
	.dc.b	$FA,$D4,$78,$5A,$F7,$6B,$D5,$B3,$E6,$8F,$81,$18,$25,$7F,$85,$67
	.dc.b	$53,$D6,$31,$3A,$8F,$F1,$14,$F9,$5E,$D3,$FE,$C9,$F8,$DF,$33,$FE
	.dc.b	$B6,$C3,$E5,$5E,$2B,$F7,$F5,$67,$35,$B3,$C1,$3A,$70,$FB,$5A,$FC
	.dc.b	$9F,$83,$FF,$7A,$7E,$09,$FE,$20,$8F,$A1,$18,$EF,$F9,$6D,$AE,$39
	.dc.b	$23,$15,$FE,$43,$8C,$4F,$FB,$BE,$FA,$4F,$78,$C5,$F4,$7E,$45,$FC
	.dc.b	$E2,$38,$24,$FC,$88,$CB,$63,$2F,$E4,$E3,$2D,$8F,$3C,$7B,$BE,$3B
	.dc.b	$C1,$3E,$78,$B8,$FF,$C0,$1C,$15,$8B,$FC,$B5,$5B,$FF,$74,$56,$A7
	.dc.b	$A9,$F1,$3D,$EF,$97,$F1,$DF,$92,$6C,$64,$F8,$91,$96,$99,$AF,$F3
	.dc.b	$95,$27,$51,$86,$BF,$B1,$AF,$CA,$92,$BF,$24,$D8,$FC,$7B,$5E,$11
	.dc.b	$A2,$3A,$6E,$3F,$B9,$9F,$C2,$F8,$1E,$28,$C4,$F9,$DF,$91,$19,$8F
	.dc.b	$DE,$4D,$AF,$8F,$35,$FD,$C7,$B5,$F9,$0F,$8D,$FD,$BB,$6F,$F1,$AD
	.dc.b	$9F,$C6,$37,$0F,$E1,$8D,$9F,$E1,$4D,$9F,$E1,$0F,$EE,$09,$1F,$B5
	.dc.b	$6C,$7E,$C9,$AF,$F1,$B6,$BD,$23,$81,$1E,$0F,$98,$FC,$F7,$C4,$7A
	.dc.b	$9C,$47,$A1,$C9,$8A,$F4,$B4,$47,$E3,$27,$93,$FA,$BF,$6B,$F6,$BE
	.dc.b	$17,$F1,$C6,$D7,$EC,$9B,$3B,$8F,$D8,$4E,$DA,$6C,$B1,$B3,$F8,$9C
	.dc.b	$FF,$1E,$24,$7F,$1B,$6C,$6A,$30,$47,$04,$3F,$A2,$47,$0B,$FE,$84
	.dc.b	$FF,$0E,$F8,$DF,$24,$F8,$67,$35,$FC,$1E,$73,$18,$9D,$4F,$C6,$F2
	.dc.b	$5B,$AF,$73,$FD,$2F,$D9,$FE,$37,$F1,$BF,$D0,$9B,$1F,$C6,$1B,$87
	.dc.b	$F1,$76,$E3,$FD,$5D,$B3,$CC,$FE,$35,$5C,$DF,$E7,$D1,$FD,$10,$8E
	.dc.b	$09,$5F,$CF,$A3,$2D,$8C,$BF,$DA,$63,$2D,$8F,$31,$FC,$A7,$C4,$F0
	.dc.b	$47,$9E,$32,$FF,$06,$8C,$DF,$A6,$72,$FF,$1B,$8C,$C7,$F6,$FF,$63
	.dc.b	$F9,$FF,$C6,$7F,$66,$6C,$7F,$3C,$6D,$FF,$3B,$6D,$FF,$7B,$6C,$F3
	.dc.b	$3F,$9F,$57,$37,$FB,$14,$7F,$68,$23,$82,$57,$F6,$28,$CB,$63,$2F
	.dc.b	$F8,$36,$32,$D8,$F3,$1F,$D3,$7E,$37,$C1,$1E,$78,$CB,$FC,$AA,$33
	.dc.b	$7E,$99,$CB,$FC,$FE,$33,$1F,$E1,$39,$FE,$C7,$E2,$FF,$82,$9B,$1F
	.dc.b	$D8,$1A,$7F,$5F,$2D,$7F,$87,$DB,$1C,$CF,$EC,$53,$CD,$FF,$03,$47
	.dc.b	$F8,$30,$8E,$08,$FF,$81,$8C,$95,$CB,$FE,$31,$32,$57,$CC,$7F,$71
	.dc.b	$F8,$DF,$04,$79,$E3,$2F,$F4,$E8,$CD,$7A,$67,$2F,$F6,$38,$CC,$7F
	.dc.b	$16,$97,$E2,$CF,$5F,$FC,$13,$E3,$7F,$8D,$23,$15,$FE,$01,$8C,$4E
	.dc.b	$A3,$FC,$23,$3A,$AF,$EB,$F5,$A9,$E0,$6A,$B8,$4B,$04,$F8,$EF,$94
	.dc.b	$F9,$4F,$2D,$7F,$8E,$A3,$31,$B8,$C1,$27,$84,$7F,$32,$9D,$B6,$74
	.dc.b	$EA,$32,$F6,$9E,$2D,$35,$3E,$B7,$FB,$DF,$8D,$6F,$D3,$1D,$6B,$DC
	.dc.b	$7F,$89,$E7,$4F,$9C,$CC,$72,$6C,$F2,$8F,$A7,$EA,$FF,$8E,$BC,$7F
	.dc.b	$A4,$DA,$F9,$ED,$9F,$95,$1F,$26,$31,$5C,$F1,$46,$27,$47,$F8,$EE
	.dc.b	$FA,$BF,$02,$76,$D9,$FA,$04,$70,$4B,$F9,$51,$96,$C6,$5F,$B5,$3D
	.dc.b	$1C,$9E,$83,$EB,$4F,$F9,$63,$C7,$F1,$C6,$4F,$87,$5F,$00,$FF,$29
	.dc.b	$46,$B0,$F7,$55,$9F,$E9,$9C,$AA,$C5,$FE,$0F,$D3,$FF,$32,$F9,$1F
	.dc.b	$9D,$6D,$FE,$69,$B3,$E1,$C9,$18,$CB,$E6,$36,$BF,$30,$DA,$ED,$1F
	.dc.b	$2E,$3F,$30,$47,$04,$8F,$C9,$46,$5B,$19,$7E,$4C,$65,$B1,$E7,$3F
	.dc.b	$CE,$9E,$47,$AD,$C4,$7A,$67,$FD,$07,$39,$8C,$4E,$AB,$FC,$CD,$39
	.dc.b	$7F,$11,$E8,$FF,$A2,$A7,$F8,$13,$5F,$DF,$B6,$BE,$33,$67,$8C,$FF
	.dc.b	$A1,$CE,$25,$38,$9F,$8F,$8F,$DE,$11,$C1,$27,$F6,$D1,$96,$C6,$5F
	.dc.b	$C6,$C6,$5B,$1D,$9F,$C7,$75,$3C,$DF,$F5,$B6,$4B,$F2,$D7,$AA,$FF
	.dc.b	$D5,$B7,$BF,$27,$38,$FB,$2B,$47,$4C,$79,$E1,$FC,$9F,$CF,$FF,$6F
	.dc.b	$E1,$DF,$E2,$B1,$97,$4B,$46,$30,$E4,$DC,$7F,$85,$DF,$1B,$FE,$5B
	.dc.b	$87,$4C,$38,$C6,$0E,$97,$CF,$0F,$D7,$D7,$F2,$72,$4F,$EF,$5B,$1F
	.dc.b	$C8,$5A,$F0,$8D,$11,$D3,$70,$FE,$19,$3E,$3F,$8A,$F4,$4F,$F3,$98
	.dc.b	$B4,$C7,$E8,$56,$23,$FA,$7C,$EA,$73,$5E,$B9,$FD,$87,$A1,$EF,$F2
	.dc.b	$7F,$A9,$36,$7F,$A7,$B6,$77,$1F,$36,$76,$D3,$65,$8D,$9F,$D0,$67
	.dc.b	$FA,$61,$23,$F9,$13,$63,$51,$82,$38,$21,$FC,$7A,$38,$63,$FC,$C5
	.dc.b	$FB,$1F,$13,$E4,$9F,$0C,$E6,$BF,$09,$39,$8C,$4E,$A7,$FC,$CD,$A2
	.dc.b	$B5,$1F,$D9,$FC,$EF,$C2,$7C,$57,$F7,$76,$C7,$F7,$46,$DF,$F1,$B6
	.dc.b	$E1,$FD,$15,$B3,$CC,$FC,$B5,$73,$7F,$A2,$47,$F7,$12,$38,$25,$7F
	.dc.b	$6B,$8C,$B6,$32,$FF,$42,$8C,$B6,$3C,$C7,$E9,$7C,$BF,$04,$79,$E3
	.dc.b	$2F,$EB,$E3,$37,$E9,$9C,$BF,$A8,$C5,$3A,$9F,$F0,$F7,$E2,$FF,$EF
	.dc.b	$17,$F6,$5F,$15,$FD,$E2,$FF,$B8,$4E,$70,$FD,$E4,$E6,$B8,$F9,$DF
	.dc.b	$B1,$8C,$4F,$FA,$FA,$B8,$5F,$86,$73,$29,$CD,$FD,$A9,$CC,$62,$7F
	.dc.b	$4D,$8A,$73,$87,$F8,$0E,$7C,$37,$FE,$14,$BF,$34,$7F,$89,$5A,$E8
	.dc.b	$FF,$10,$B8,$7C,$D3,$CD,$B3,$CE,$31,$1E,$76,$CF,$18,$C4,$EC,$FF
	.dc.b	$6E,$E1,$F2,$2B,$F8,$A7,$C4,$FA,$27,$27,$8E,$74,$FF,$21,$8D,$57
	.dc.b	$F7,$DB,$F8,$4F,$63,$F9,$64,$6A,$BE,$AF,$99,$FC,$87,$CD,$F9,$2D
	.dc.b	$C3,$BD,$FD,$38,$C3,$D5,$B5,$CA,$7F,$99,$57,$26,$9C,$AB,$EC,$57
	.dc.b	$C2,$24,$7D,$16,$C7,$CE,$6B,$F4,$DA,$FE,$41,$AF,$96,$39,$1D,$F0
	.dc.b	$4F,$F3,$DF,$89,$F2,$57,$A6,$B5,$39,$69,$23,$47,$F5,$39,$E3,$87
	.dc.b	$E7,$3C,$BF,$E9,$1F,$13,$F9,$06,$DF,$85,$B7,$F4,$DB,$5C,$67,$FA
	.dc.b	$C5,$71,$69,$C4,$FC,$A4,$FE,$20,$8E,$09,$5F,$94,$8C,$B6,$32,$FE
	.dc.b	$6E,$32,$D8,$EC,$FF,$6F,$F3,$BC,$4F,$9E,$33,$7F,$C0,$67,$2F,$A6
	.dc.b	$74,$71,$9D,$9F,$C1,$3C,$BF,$CF,$91,$FD,$73,$63,$F5,$6D,$7F,$60
	.dc.b	$D7,$94,$70,$23,$FB,$2F,$2B,$F5,$64,$8F,$DD,$36,$35,$18,$23,$82
	.dc.b	$1F,$BF,$8E,$0F,$FA,$E7,$A3,$FC,$39,$F1,$1F,$E0,$53,$FC,$0D,$F1
	.dc.b	$36,$FF,$C2,$11,$6B,$FC,$31,$37,$1F,$E5,$13,$6B,$FA,$7C,$DA,$FE
	.dc.b	$2B,$36,$FF,$C5,$53,$79,$13,$63,$0D,$BF,$09,$F9,$FF,$2B,$F8,$69
	.dc.b	$1F,$E3,$8D,$C3,$F8,$CB,$5F,$E3,$CD,$7F,$9D,$B5,$FE,$64,$FF,$18
	.dc.b	$F2,$64,$FF,$27,$6C,$FF,$24,$6B,$FC,$A5,$AF,$F4,$66,$BF,$CF,$5A
	.dc.b	$FF,$3B,$F2,$7F,$92,$92,$7F,$9A,$36,$3F,$98,$B5,$FE,$6C,$D7,$FA
	.dc.b	$9B,$5F,$E9,$4F,$F3,$0F,$1E,$4F,$F3,$F6,$CF,$F3,$C6,$BF,$D0,$5A
	.dc.b	$FF,$5E,$6B,$FD,$55,$AF,$F5,$3F,$17,$FA,$CE,$AF,$82,$FE,$A1,$7F
	.dc.b	$C4,$E7,$31,$8A,$FE,$6D,$18,$EA,$FE,$B2,$DA,$E0,$7F,$32,$BE,$93
	.dc.b	$A8,$C5,$F3,$9F,$ED,$F5,$FD,$38,$93,$FD,$75,$B7,$FD,$69,$AF,$F5
	.dc.b	$F6,$BF,$DF,$1A,$FF,$73,$78,$CF,$69,$F0,$F4,$F8,$EB,$D7,$1E,$87
	.dc.b	$FD,$A5,$3F,$0E,$74,$F1,$8D,$9A,$B4,$EA,$93,$3E,$1B,$FC,$EA,$FF
	.dc.b	$A1,$7D,$1F,$F0,$1C,$63,$0F,$F0,$B4,$63,$C5,$FF,$3E,$36,$FF,$D0
	.dc.b	$CD,$BE,$51,$FD,$56,$F9,$3F,$E4,$49,$D4,$F8,$8F,$15,$7F,$7C,$AF
	.dc.b	$1C,$64,$38,$78,$35,$E3,$5B,$69,$63,$4F,$29,$EC,$49,$E1,$19,$6C
	.dc.b	$FA,$23,$9D,$FF,$8D,$FE,$27,$DC,$AB,$C7,$78,$7F,$CC,$33,$6B,$8C
	.dc.b	$68,$FF,$56,$46,$63,$DB,$E1,$7E,$6B,$8F,$E8,$B6,$BE,$53,$65,$87
	.dc.b	$85,$FB,$51,$98,$C4,$74,$9F,$CD,$D7,$46,$9D,$0F,$7D,$7D,$12,$38
	.dc.b	$24,$FC,$B8,$CB,$63,$2F,$D7,$C7,$FD,$01,$18,$8F,$AF,$3E,$17,$F2
	.dc.b	$9C,$77,$0F,$3D,$F0,$9F,$6D,$F0,$3F,$67,$5B,$71,$1A,$3E,$0B,$A3
	.dc.b	$D8,$E4,$FC,$4E,$FF,$ED,$3F,$89,$FC,$EB,$67,$F2,$CD,$9E,$17,$FC
	.dc.b	$06,$7F,$D3,$75,$C1,$FF,$72,$36,$7F,$DD,$8D,$7E,$15,$7E,$6C,$93
	.dc.b	$F8,$C6,$C7,$E5,$5A,$F0,$8D,$11,$D3,$70,$F8,$F3,$FE,$ED,$F4,$BC
	.dc.b	$51,$89,$F5,$9F,$EF,$98,$CB,$FA,$89,$B5,$FE,$7A,$9A,$FE,$BB,$B7
	.dc.b	$F9,$5F,$C4,$FF,$02,$6D,$FE,$ED,$B3,$F1,$9B,$87,$DA,$6C,$FE,$D9
	.dc.b	$B3,$FC,$3D,$FD,$F1,$23,$F5,$ED,$8F,$DC,$35,$FC,$63,$5E,$91,$C0
	.dc.b	$8F,$07,$CC,$7F,$3A,$CB,$D4,$E2,$3D,$0E,$4C,$57,$A5,$A7,$FA,$23
	.dc.b	$8B,$F8,$8B,$89,$FE,$51,$B7,$FA,$B3,$A5,$D2,$CF,$F9,$4C,$E6,$31
	.dc.b	$7F,$C8,$E3,$13,$A8,$FF,$41,$D7,$28,$C5,$FE,$76,$31,$3C,$4D,$DF
	.dc.b	$49,$FC,$BD,$FF,$29,$24,$FF,$1A,$6C,$7F,$23,$6B,$FB,$76,$BF,$CB
	.dc.b	$1A,$FB,$A3,$91,$E1,$7C,$93,$F1,$1F,$0C,$67,$1F,$DD,$D7,$C1,$CB
	.dc.b	$D7,$18,$9D,$1F,$D4,$B0,$F9,$33,$A8,$FE,$7B,$C3,$FE,$00,$FD,$83
	.dc.b	$FC,$01,$1F,$D2,$E3,$1A,$7F,$56,$6C,$64,$FE,$F9,$3C,$32,$FD,$DC
	.dc.b	$E9,$B3,$AA,$E4,$FF,$40,$9F,$EA,$44,$70,$49,$FE,$99,$19,$6C,$65
	.dc.b	$FE,$45,$19,$6C,$79,$5F,$EF,$34,$F8,$7F,$0C,$79,$09,$97,$FB,$FE
	.dc.b	$75,$5C,$63,$F9,$3D,$F0,$75,$76,$7F,$B2,$EB,$F9,$1C,$3F,$BD,$B6
	.dc.b	$3F,$BA,$36,$3F,$8F,$B6,$BF,$99,$B6,$79,$9F,$90,$9E,$6F,$F6,$68
	.dc.b	$FE,$F0,$47,$04,$8F,$EE,$51,$96,$C6,$5F,$E9,$11,$96,$C7,$98,$FD
	.dc.b	$2F,$E0,$CF,$8A,$F0,$CF,$A2,$74,$FE,$9E,$75,$18,$BF,$5B,$4F,$C8
	.dc.b	$46,$A7,$91,$FD,$1F,$F1,$9C,$57,$D1,$F8,$8F,$ED,$CF,$F6,$C8,$CD
	.dc.b	$FF,$8A,$A1,$A2,$B7,$3F,$E1,$DA,$DD,$FF,$78,$AD,$4F,$FB,$E6,$73
	.dc.b	$7C,$AF,$8D,$FF,$8D,$A7,$37,$C2,$3C,$D8,$FE,$96,$B7,$0A,$F6,$3F
	.dc.b	$E2,$96,$BE,$39,$FF,$02,$B7,$0F,$F1,$0C,$EA,$3C,$13,$A6,$CE,$A3
	.dc.b	$11,$DD,$B3,$C6,$31,$3B,$23,$E8,$7F,$6B,$F1,$1E,$89,$F8,$B3,$C0
	.dc.b	$D1,$FA,$68,$D5,$4A,$F9,$0F,$F0,$58,$C4,$FD,$9D,$BF,$D0,$FF,$AF
	.dc.b	$FF,$97,$AF,$E7,$46,$27,$C0,$D8,$C9,$FE,$83,$9E,$13,$CE,$30,$F0
	.dc.b	$69,$C1,$FF,$28,$61,$FC,$3D,$FA,$15,$F1,$09,$3F,$39,$B1,$C1,$AF
	.dc.b	$08,$D1,$1D,$3D,$F3,$FC,$9F,$4F,$C8,$BF,$E3,$D1,$F9,$5B,$F3,$5F
	.dc.b	$FA,$1A,$B5,$18,$9D,$5F,$FA,$86,$72,$FE,$4F,$E2,$FF,$AE,$6C,$F8
	.dc.b	$CF,$AD,$18,$6C,$FE,$D1,$B8,$FE,$C5,$B8,$7F,$95,$79,$18,$73,$BE
	.dc.b	$C7,$12,$DF,$CA,$C3,$F1,$64,$8F,$CA,$B6,$35,$18,$23,$82,$1F,$A1
	.dc.b	$8E,$17,$FB,$97,$DB,$ED,$F9,$AB,$F6,$D1,$69,$87,$A2,$B1,$1F,$AD
	.dc.b	$9D,$4E,$6B,$D3,$7F,$C1,$78,$3F,$DA,$F9,$BD,$BF,$CF,$46,$1B,$5F
	.dc.b	$45,$B8,$FD,$76,$DF,$DE,$6C,$FF,$0F,$BE,$8F,$F0,$9B,$FD,$89,$1F
	.dc.b	$DD,$B6,$FF,$2A,$D7,$F8,$23,$5F,$E2,$CD,$7F,$88,$3C,$6F,$D9,$3F
	.dc.b	$87,$C7,$C7,$5A,$8F,$11,$92,$BF,$CA,$EB,$74,$B5,$F2,$AF,$E3,$1F
	.dc.b	$17,$F9,$0D,$BF,$82,$37,$1F,$E4,$CD,$9F,$E6,$CD,$C3,$F9,$A3,$67
	.dc.b	$F9,$83,$67,$F9,$53,$FC,$58,$91,$FC,$81,$B1,$FB,$A6,$BF,$CA,$5A
	.dc.b	$F4,$8E,$04,$78,$3E,$63,$FA,$BF,$2F,$A9,$C4,$7A,$1C,$98,$AF,$4B
	.dc.b	$4F,$EA,$B3,$C6,$7F,$AA,$7C,$57,$E3,$3D,$8F,$E5,$0D,$AF,$E9,$4D
	.dc.b	$9D,$C7,$EF,$67,$6D,$36,$58,$D9,$FC,$6A,$7F,$9D,$92,$3F,$A2,$B6
	.dc.b	$35,$18,$23,$82,$1F,$D4,$63,$85,$FF,$69,$7F,$4B,$87,$92,$7C,$33
	.dc.b	$9A,$FE,$29,$39,$8C,$4E,$A7,$F4,$53,$97,$FB,$77,$C4,$FF,$18,$DB
	.dc.b	$FA,$63,$63,$FB,$73,$6B,$FB,$F3,$6F,$FB,$E3,$67,$99,$FE,$10,$AE
	.dc.b	$6F,$F6,$68,$FE,$BE,$47,$04,$AF,$E8,$71,$96,$C6,$5F,$EE,$91,$96
	.dc.b	$C7,$98,$FF,$0F,$E9,$E0,$8F,$3C,$65,$FF,$0F,$46,$6F,$D3,$39,$7F
	.dc.b	$98,$70,$FE,$BA,$30,$2A,$7D,$B3,$E2,$3F,$8B,$FB,$1F,$E0,$A8,$C5
	.dc.b	$7F,$45,$8C,$4E,$A3,$FC,$40,$73,$A5,$70,$7F,$A8,$57,$09,$E4,$70
	.dc.b	$8F,$0D,$73,$7F,$DE,$D5,$C5,$E8,$74,$8F,$F0,$7C,$66,$3C,$0D,$79
	.dc.b	$CF,$06,$96,$34,$F3,$8D,$34,$D4,$F6,$24,$6A,$7D,$98,$23,$C3,$3F
	.dc.b	$CF,$B2,$F7,$5F,$AA,$2D,$7F,$76,$9B,$5C,$0F,$85,$5E,$19,$D4,$62
	.dc.b	$FB,$BE,$EF,$87,$FE,$E7,$F1,$BF,$55,$B1,$B6,$DF,$CA,$6D,$F9,$E3
	.dc.b	$EF,$C6,$27,$A1,$FE,$6C,$AE,$8F,$BE,$3E,$79,$1C,$12,$7E,$64,$65
	.dc.b	$B1,$97,$EC,$DA,$75,$18,$7B,$BF,$E7,$7F,$80,$B8,$79,$E3,$F1,$73
	.dc.b	$8B,$FF,$3B,$D5,$9F,$5C,$7C,$13,$FD,$0B,$19,$8F,$C9,$FC,$3F,$F9
	.dc.b	$AA,$DF,$AF,$BD,$AF,$CF,$4E,$63,$18,$7D,$18,$C4,$EA,$3F,$31,$5C
	.dc.b	$A3,$16,$8F,$F4,$2D,$F2,$8F,$CD,$DF,$2B,$E8,$7E,$2D,$FC,$D9,$1C
	.dc.b	$12,$3F,$31,$19,$6C,$65,$F9,$B1,$96,$C7,$67,$C8,$FF,$A5,$67,$E9
	.dc.b	$39,$AF,$9D,$92,$FD,$71,$EA,$CA,$D7,$EB,$A6,$E1,$EF,$C3,$9B,$EE
	.dc.b	$9F,$CC,$7A,$DF,$B6,$23,$FA,$96,$C7,$EC,$DA,$FE,$19,$AF,$EF,$1A
	.dc.b	$FF,$16,$7F,$80,$FC,$3C,$9F,$DA,$36,$7F,$7C,$D7,$F2,$2D,$7F,$83
	.dc.b	$B5,$FE,$42,$D7,$F8,$CF,$AD,$FC,$38,$93,$FC,$05,$B1,$FC,$2D,$AF
	.dc.b	$E7,$9A,$FF,$14,$6B,$FC,$B9,$FE,$33,$EB,$49,$FE,$18,$D9,$FE,$2C
	.dc.b	$D7,$F5,$2D,$7F,$8F,$B5,$FE,$72,$D7,$F9,$8F,$C3,$FF,$26,$24,$FF
	.dc.b	$19,$6C,$7F,$23,$6B,$FB,$66,$BF,$CB,$1A,$FF,$46,$7F,$98,$FC,$3C
	.dc.b	$9F,$E4,$8D,$9F,$E5,$CD,$7F,$80,$B5,$FE,$6E,$D7,$FA,$8B,$5F,$E9
	.dc.b	$3E,$BF,$ED,$FF,$5E,$7C,$7F,$F5,$38,$C3,$6F,$F6,$73,$A7,$F9,$44
	.dc.b	$E9,$A6,$A7,$85,$7E,$90,$FD,$C5,$71,$7F,$A8,$D7,$F4,$E2,$57,$F4
	.dc.b	$16,$C7,$F4,$B6,$BF,$C7,$9A,$FF,$54,$6B,$FD,$B9,$E3,$C1,$FB,$07
	.dc.b	$DD,$F0,$9E,$48,$C5,$7A,$2B,$CD,$0B,$56,$EB,$B1,$B9,$FE,$7C,$6E
	.dc.b	$35,$5F,$DF,$BE,$23,$FD,$77,$E1,$FF,$7D,$6D,$FF,$4F,$6C,$FF,$72
	.dc.b	$6D,$FF,$06,$6C,$FF,$70,$6C,$F5,$8F,$EC,$D1,$FD,$D8,$8E,$09,$1F
	.dc.b	$DC,$E3,$2D,$8C,$BF,$D2,$23,$2D,$8F,$31,$FE,$09,$F1,$3F,$D9,$7E
	.dc.b	$4F,$AE,$33,$3E,$97,$5D,$A9,$D4,$E4,$98,$7F,$89,$A3,$2F,$F5,$1F
	.dc.b	$C7,$5F,$11,$F9,$3F,$8F,$F9,$B8,$FF,$87,$E7,$2F,$E4,$A3,$38,$7F
	.dc.b	$8B,$67,$2F,$DA,$D9,$3A,$84,$EB,$0F,$B3,$5A,$BF,$F0,$2C,$E5,$FF
	.dc.b	$1C,$4F,$E0,$EF,$94,$E6,$FF,$37,$18,$C3,$FB,$54,$79,$DA,$F9,$5A
	.dc.b	$F8,$9B,$3E,$88,$FF,$18,$B5,$E5,$7F,$E2,$88,$C4,$73,$8C,$4F,$22
	.dc.b	$38,$3D,$6F,$EF,$7E,$1B,$CD,$18,$24,$FC,$78,$CB,$FB,$73,$CD,$1A
	.dc.b	$C3,$E3,$CE,$9F,$B1,$F1,$1F,$4B,$56,$AF,$1F,$E3,$CE,$63,$18,$7C
	.dc.b	$E8,$C4,$EA,$3F,$12,$DC,$7F,$84,$5F,$48,$C6,$1F,$E6,$E7,$53,$F2
	.dc.b	$2F,$E5,$92,$FE,$43,$63,$FC,$AD,$9A,$31,$1C,$67,$64,$77,$1C,$6F
	.dc.b	$DF,$3F,$9E,$F1,$BC,$6F,$E5,$66,$D3,$1F,$87,$58,$8F,$C8,$CE,$A7
	.dc.b	$35,$F0,$CF,$C8,$FC,$3F,$F8,$A7,$E3,$47,$E4,$1C,$36,$3F,$C4,$AE
	.dc.b	$9E,$26,$1D,$5F,$F1,$C7,$4F,$DD,$C9,$5D,$2B,$8C,$FC,$68,$FC,$A9
	.dc.b	$23,$F2,$4D,$8D,$46,$08,$E0,$87,$E7,$23,$86,$5F,$BF,$7D,$DF,$0D
	.dc.b	$E6,$8C,$57,$89,$FE,$5B,$82,$75,$3F,$C0,$EA,$DF,$F9,$0E,$B5,$3C
	.dc.b	$4F,$E1,$1F,$11,$F9,$1E,$8F,$DC,$B6,$3F,$7E,$DA,$F8,$8D,$C7,$F8
	.dc.b	$A3,$67,$99,$FC,$7A,$B9,$BF,$B5,$8F,$DD,$91,$C1,$2B,$F2,$31,$96
	.dc.b	$C6,$5F,$E0,$31,$96,$C7,$98,$FE,$5B,$CF,$E0,$8F,$3C,$65,$FD,$3F
	.dc.b	$07,$F5,$88,$C6,$1E,$A6,$9F,$ED,$3D,$7F,$6F,$7C,$E7,$F9,$97,$C3
	.dc.b	$7E,$47,$B7,$F8,$FB,$6F,$F9,$2B,$71,$FC,$53,$6B,$8C,$FE,$52,$B8
	.dc.b	$B4,$E2,$7F,$1B,$9F,$E4,$24,$70,$4A,$FD,$BC,$65,$B1,$97,$F9,$3C
	.dc.b	$65,$B1,$D9,$FD,$07,$6F,$89,$F3,$C6,$6F,$F1,$B3,$97,$D3,$3A,$38
	.dc.b	$CE,$CF,$EA,$5F,$03,$FC,$6F,$CD,$FE,$86,$D8,$FE,$92,$D9,$FD,$73
	.dc.b	$6B,$FA,$F3,$67,$99,$FD,$A2,$B9,$BF,$CF,$E3,$FA,$29,$1C,$12,$BF
	.dc.b	$8D,$C6,$5B,$19,$7F,$A5,$C6,$5B,$1E,$63,$FB,$BF,$B7,$E0,$8F,$3C
	.dc.b	$65,$FC,$5C,$66,$FD,$33,$97,$FD,$BD,$A7,$F4,$09,$D4,$FF,$77,$F7
	.dc.b	$7F,$8D,$7C,$2F,$F6,$86,$C7,$F6,$E6,$E1,$FC,$4D,$B8,$7F,$80,$DB
	.dc.b	$3C,$CF,$E3,$F5,$CD,$FE,$C9,$1F,$DA,$88,$E0,$95,$FD,$02,$32,$D8
	.dc.b	$CB,$FD,$C2,$32,$D8,$F3,$1F,$CB,$BE,$17,$C1,$1E,$78,$CB,$FC,$B6
	.dc.b	$33,$7E,$99,$CB,$FC,$6A,$33,$1F,$E1,$B9,$FF,$73,$7E,$B9,$FE,$C3
	.dc.b	$8F,$F0,$8C,$63,$C8,$FF,$07,$B6,$32,$7F,$B9,$27,$84,$7F,$3B,$34
	.dc.b	$58,$D4,$F2,$7F,$C1,$51,$FE,$10,$23,$82,$3F,$D9,$8C,$95,$CB,$FE
	.dc.b	$16,$32,$57,$CA,$FF,$9A,$A5,$F1,$DE,$18,$F2,$13,$0F,$F3,$4C,$EA
	.dc.b	$B8,$C7,$F9,$0E,$78,$3A,$AB,$29,$8B,$D7,$FF,$07,$7A,$5F,$E3,$F8
	.dc.b	$C1,$FE,$54,$71,$1A,$8F,$F1,$E9,$CE,$4F,$07,$F9,$5C,$F0,$8E,$47
	.dc.b	$08,$F0,$CF,$37,$FC,$51,$1C,$5E,$87,$48,$FF,$20,$C6,$63,$C0,$D7
	.dc.b	$9B,$C1,$A5,$74,$F3,$74,$D3,$51,$D8,$91,$A3,$D9,$7E,$E9,$EA,$FE
	.dc.b	$EB,$95,$E8,$FB,$E7,$D7,$76,$B5,$36,$FD,$33,$6B,$FC,$41,$7C,$A7
	.dc.b	$E3,$47,$96,$FC,$0F,$CE,$F5,$3F,$43,$E5,$7D,$76,$D7,$00,$E2,$7E
	.dc.b	$6B,$6B,$9C,$E8,$FD,$25,$73,$69,$CE,$7E,$89,$1F,$88,$DA,$F9,$CD
	.dc.b	$7E,$CC,$75,$72,$F0,$23,$82,$57,$92,$7F,$7B,$E5,$78,$E7,$D3,$5E
	.dc.b	$33,$F7,$51,$9A,$E1,$19,$9F,$D5,$4E,$63,$F0,$FE,$8F,$F9,$AF,$C6
	.dc.b	$FC,$F3,$63,$E4,$36,$BF,$26,$DB,$F8,$ED,$9E,$67,$FA,$E2,$B9,$BF
	.dc.b	$87,$8F,$CB,$91,$C1,$2B,$F2,$91,$96,$C6,$5F,$93,$19,$6C,$79,$8F
	.dc.b	$F6,$3F,$8B,$E0,$8F,$3C,$65,$FF,$0D,$C6,$6F,$D3,$39,$7F,$31,$19
	.dc.b	$8F,$D5,$F8,$BF,$CB,$3C,$9F,$E0,$4D,$8F,$C6,$36,$FF,$6E,$DB,$FC
	.dc.b	$5B,$67,$99,$FC,$BE,$B9,$BF,$AB,$8F,$DD,$91,$C1,$2B,$F7,$11,$96
	.dc.b	$C6,$5F,$C6,$C6,$5B,$1E,$63,$F9,$F7,$C3,$F8,$23,$CF,$19,$7F,$97
	.dc.b	$46,$6F,$D3,$39,$7F,$96,$C6,$63,$F8,$7F,$7F,$FA,$DF,$F0,$A7,$AB
	.dc.b	$FC,$9E,$30,$D9,$FF,$80,$13,$A7,$F6,$93,$A6,$9A,$9E,$18,$7F,$AE
	.dc.b	$2B,$83,$FC,$92,$BF,$90,$92,$BF,$8A,$36,$3F,$8F,$35,$FD,$8B,$5F
	.dc.b	$4C,$70,$23,$C1,$F1,$BE,$1F,$90,$F1,$E3,$FC,$1A,$AD,$7F,$98,$6D
	.dc.b	$58,$BF,$E9,$75,$7B,$55,$AB,$EB,$87,$F1,$0E,$6F,$E5,$4C,$F6,$3F
	.dc.b	$D9,$93,$FC,$66,$72,$FF,$57,$94,$EB,$1F,$E8,$F5,$A5,$3A,$F2,$BF
	.dc.b	$AD,$4E,$63,$17,$D4,$FE,$A4,$E2,$B9,$C6,$2F,$B4,$F9,$E3,$FA,$15
	.dc.b	$7F,$52,$23,$82,$57,$F4,$F8,$CB,$63,$2F,$F2,$48,$CB,$63,$C8,$FA
	.dc.b	$9F,$81,$4F,$6B,$D7,$1A,$BF,$25,$6C,$F8,$35,$B9,$D1,$FD,$9A,$F8
	.dc.b	$CE,$EB,$E5,$DE,$1F,$E8,$7C,$3F,$9D,$F5,$FF,$C0,$4D,$C3,$FA,$63
	.dc.b	$6F,$FB,$BB,$6B,$8C,$FE,$82,$B8,$B4,$E2,$7F,$6B,$9F,$EF,$64,$70
	.dc.b	$4A,$FE,$F1,$19,$6C,$65,$FE,$9B,$19,$6C,$76,$7F,$55,$EB,$78,$9F
	.dc.b	$3C,$66,$FF,$CE,$3A,$2B,$51,$EA,$AD,$9C,$AB,$81,$FE,$1F,$FC,$7D
	.dc.b	$8F,$E3,$74,$79,$3F,$93,$B5,$6A,$FF,$19,$39,$8C,$4F,$D3,$AE,$35
	.dc.b	$FE,$00,$9C,$E1,$FA,$5B,$E1,$87,$76,$D7,$43,$84,$7F,$81,$B0,$F3
	.dc.b	$35,$FF,$20,$5F,$81,$FF,$16,$B5,$DC,$FF,$89,$23,$11,$C2,$79,$3E
	.dc.b	$C6,$C6,$A3,$13,$C6,$72,$47,$CA,$7E,$B7,$E1,$7D,$33,$F9,$93,$89
	.dc.b	$6B,$E0,$CF,$E8,$A7,$78,$7F,$88,$A7,$35,$E5,$AE,$87,$67,$97,$99
	.dc.b	$E6,$24,$7B,$5B,$1F,$E4,$79,$46,$23,$E2,$B5,$FC,$03,$5F,$96,$FC
	.dc.b	$2F,$3A,$57,$C9,$6C,$F5,$6B,$F4,$1A,$FE,$2D,$AF,$D6,$6B,$F8,$AF
	.dc.b	$0F,$E7,$92,$7E,$B3,$63,$D4,$D7,$EE,$B5,$FC,$C3,$5F,$C1,$BF,$6B
	.dc.b	$C2,$93,$F8,$16,$CF,$C5,$6B,$F8,$86,$BF,$A3,$6B,$F8,$E6,$BF,$A2
	.dc.b	$F5,$FF,$0E,$49,$FC,$73,$63,$E9,$35,$FC,$AB,$5F,$D8,$35,$FC,$DB
	.dc.b	$F9,$2F,$5E,$4F,$E6,$5B,$3F,$75,$AF,$E8,$1A,$FE,$ED,$AF,$E9,$9A
	.dc.b	$FE,$EB,$CE,$FC,$F9,$27,$F4,$CD,$8F,$C5,$35,$FD,$6B,$5F,$E1,$0D
	.dc.b	$7F,$66,$FE,$AB,$CD,$93,$FB,$16,$CF,$E5,$5A,$FE,$E1,$AF,$F1,$36
	.dc.b	$BF,$BE,$6B,$FC,$4B,$C3,$FD,$B9,$27,$F7,$CD,$8F,$D1,$35,$FE,$0A
	.dc.b	$D7,$F9,$03,$5F,$E1,$6F,$F0,$2F,$0A,$4F,$F0,$96,$CF,$EB,$5A,$FF
	.dc.b	$10,$6B,$FC,$AD,$AF,$F1,$66,$BF,$CA,$BE,$2F,$F8,$79,$27,$F8,$B3
	.dc.b	$63,$F7,$4D,$7F,$8E,$B5,$FE,$70,$D7,$F9,$1B,$FC,$6B,$D8,$93,$FC
	.dc.b	$85,$B3,$FC,$15,$AF,$F2,$86,$BF,$D0,$DA,$FF,$2E,$6B,$FD,$0B,$CC
	.dc.b	$FE,$4E,$49,$FE,$5C,$D8,$FE,$24,$D7,$F9,$AB,$5F,$EA,$0D,$7F,$9D
	.dc.b	$BF,$CC,$BC,$B9,$3F,$CE,$5B,$3F,$C7,$5A,$FF,$40,$6B,$FD,$6D,$AF
	.dc.b	$F4,$66,$BF,$D6,$BC,$4F,$E7,$E4,$9F,$E8,$CD,$8F,$E5,$4D,$7F,$A6
	.dc.b	$B5,$FE,$D0,$D7,$FA,$9B,$FD,$2B,$C3,$93,$FD,$45,$B3,$FC,$D5,$AF
	.dc.b	$F5,$86,$BF,$DC,$DA,$FF,$5E,$6B,$FD,$CA,$3F,$AB,$90,$FE,$BC,$57
	.dc.b	$FA,$13,$5F,$EC,$AD,$7F,$C0,$0D,$7F,$B5,$BF,$D9,$3F,$20,$7A,$D4
	.dc.b	$90,$98,$49,$DC,$62,$31,$3C,$1C,$09,$DB,$86,$BC,$5A,$F2,$9E,$2F
	.dc.b	$19,$E8,$EF,$E2,$FD,$04,$9D,$D2,$31,$3D,$DA,$F1,$9D,$C6,$23,$D8
	.dc.b	$D5,$E1,$FF,$A0,$B9,$BF,$D5,$07,$FA,$A7,$9E,$DF,$FA,$9A,$2D,$7F
	.dc.b	$A8,$66,$F4,$FF,$8D,$A6,$D7,$FA,$46,$6D,$76,$EF,$56,$30,$FF,$48
	.dc.b	$D5,$9C,$DF,$73,$FD,$6B,$19,$AD,$3D,$2F,$E8,$7C,$7F,$F8,$EF,$A9
	.dc.b	$E8,$FF,$8F,$75,$4E,$63,$0D,$C7,$FD,$30,$7D,$F7,$17,$FE,$1F,$AE
	.dc.b	$33,$A8,$FC,$35,$F6,$8C,$47,$89,$B1,$98,$E5,$5F,$91,$24,$FD,$A6
	.dc.b	$C7,$E1,$DA,$FE,$1D,$AF,$96,$38,$11,$E0,$F8,$AB,$CF,$1E,$9F,$8E
	.dc.b	$F7,$CE,$6A,$CE,$63,$04,$7C,$EF,$39,$DE,$A8,$F1,$4E,$8E,$45,$8D
	.dc.b	$3F,$AB,$BF,$1C,$62,$3E,$31,$C1,$C4,$6A,$7A,$C6,$2F,$ED,$5F,$8A
	.dc.b	$72,$4B,$ED,$19,$7C,$13,$5F,$DA,$45,$BE,$D1,$AA,$F7,$1F,$06,$FC
	.dc.b	$75,$B2,$46,$4F,$E0,$8E,$27,$E0,$46,$0D,$3E,$46,$BE,$26,$D2,$79
	.dc.b	$57,$BE,$7C,$75,$D0,$E4,$53,$91,$FC,$36,$7F,$8E,$12,$3F,$82,$B6
	.dc.b	$35,$18,$23,$82,$1F,$C5,$63,$84,$FC,$29,$FC,$67,$C6,$7D,$27,$CB
	.dc.b	$19,$AF,$D3,$4E,$5F,$B9,$3F,$0D,$CB,$C0,$C4,$F1,$3F,$03,$3C,$5C
	.dc.b	$46,$8F,$E6,$51,$97,$13,$F1,$27,$8C,$75,$8F,$C8,$4E,$DB,$19,$7F
	.dc.b	$9B,$46,$5B,$19,$B4,$E5,$B1,$98,$C1,$1C,$12,$7E,$09,$E0,$29,$D8
	.dc.b	$B3,$D1,$B3,$CD,$B3,$CE,$BE,$1C,$73,$3F,$A5,$4F,$F5,$B2,$4F,$F4
	.dc.b	$46,$C7,$F5,$16,$BC,$23,$44,$74,$D9,$F9,$71,$FE,$BF,$D9,$F0,$1F
	.dc.b	$CA,$EE,$5F,$E1,$0F,$E4,$33,$A7,$15,$F8,$02,$5F,$F0,$19,$D4,$62
	.dc.b	$7E,$E9,$C2,$75,$7D,$A7,$47,$CB,$8C,$CE,$CE,$31,$A3,$FA,$DC,$F1
	.dc.b	$9F,$59,$27,$FB,$ED,$78,$A3,$15,$D6,$79,$C7,$68,$E9,$87,$D7,$8D
	.dc.b	$C7,$69,$F1,$3C,$48,$64,$E3,$87,$8E,$3C,$38,$59,$DD,$F5,$9C,$9D
	.dc.b	$23,$F1,$17,$E0,$C3,$D1,$3E,$62,$4F,$82,$32,$F4,$C3,$B3,$E3,$AF
	.dc.b	$4D,$F7,$8E,$18,$6A,$31,$1E,$0C,$3C,$6F,$A9,$F2,$61,$DA,$34,$7F
	.dc.b	$8A,$5C,$4F,$26,$E1,$FD,$13,$0F,$98,$7A,$F0,$F2,$CE,$B0,$FC,$5C
	.dc.b	$60,$8F,$81,$B8,$7D,$F9,$E0,$FE,$5E,$33,$5D,$6B,$81,$27,$84,$EA
	.dc.b	$3E,$4C,$7C,$56,$D7,$4A,$F3,$C6,$2B,$85,$7C,$88,$E9,$3D,$E3,$AC
	.dc.b	$60,$E0,$57,$81,$FE,$48,$AF,$D1,$BF,$E4,$D6,$BF,$40,$C5,$7F,$69
	.dc.b	$70,$43,$E8,$39,$6C,$FF,$93,$27,$9C,$60,$F0,$4E,$DF,$80,$66,$BF
	.dc.b	$BB,$7E,$95,$FE,$32,$8F,$99,$7F,$50,$8A,$77,$BF,$DD,$9F,$C3,$C7
	.dc.b	$C8,$9E,$91,$FE,$6C,$AE,$93,$99,$FF,$28,$FB,$DC,$5F,$F3,$97,$E8
	.dc.b	$4F,$07,$11,$AC,$3F,$C5,$F3,$97,$F8,$8D,$6A,$31,$3F,$C6,$63,$15
	.dc.b	$CB,$34,$62,$F8,$9D,$27,$51,$F1,$67,$73,$F2,$E3,$04,$70,$4A,$FB
	.dc.b	$91,$96,$C6,$5F,$BD,$19,$C3,$F9,$23,$EE,$F2,$1E,$6F,$8A,$FC,$B7
	.dc.b	$EB,$7F,$C9,$F3,$A7,$37,$C9,$A7,$FA,$6A,$35,$3C,$49,$3B,$7E,$D4
	.dc.b	$EE,$BC,$11,$A8,$FE,$4F,$7E,$59,$CD,$71,$24,$F1,$3C,$35,$FC,$0E
	.dc.b	$B7,$19,$8E,$86,$5D,$AB,$F3,$D7,$48,$D1,$C0,$D5,$F4,$73,$1B,$73
	.dc.b	$5D,$30,$FB,$31,$C6,$FA,$CF,$B0,$D6,$1E,$D6,$9E,$48,$DC,$F6,$8E
	.dc.b	$0F,$3C,$3C,$4F,$E5,$4C,$47,$18,$CC,$F2,$9D,$61,$FE,$63,$8C,$61
	.dc.b	$E6,$8C,$57,$02,$5F,$32,$5F,$62,$4F,$C1,$BE,$78,$7A,$E7,$35,$E8
	.dc.b	$AE,$AF,$27,$85,$F6,$9D,$9E,$F3,$FD,$C9,$7D,$2B,$F2,$33,$BA,$E3
	.dc.b	$1C,$CE,$C7,$DF,$73,$5D,$0F,$83,$19,$36,$F6,$9E,$91,$88,$FB,$67
	.dc.b	$1C,$3E,$34,$71,$8F,$31,$AA,$EE,$48,$E5,$7E,$99,$FC,$F9,$DE,$3F
	.dc.b	$DE,$31,$FA,$68,$FF,$7F,$35,$FC,$C1,$8B,$FE,$51,$19,$8E,$40,$D4
	.dc.b	$F4,$9D,$91,$C1,$23,$C6,$7A,$5E,$4B,$27,$87,$F6,$F0,$FB,$33,$C3
	.dc.b	$0F,$C6,$56,$AF,$EA,$92,$75,$87,$F2,$5C,$39,$4F,$C9,$9F,$01,$FC
	.dc.b	$C2,$32,$FB,$30,$EE,$7F,$AE,$6F,$B4,$74,$AE,$D3,$A7,$D3,$3A,$8C
	.dc.b	$4F,$39,$F0,$4E,$5F,$16,$1F,$A5,$3D,$77,$E2,$AF,$29,$2B,$83,$CC
	.dc.b	$86,$63,$84,$F3,$C3,$E0,$5F,$98,$EE,$E2,$BD,$44,$C3,$CE,$FD,$1B
	.dc.b	$E7,$3E,$3C,$3C,$33,$E4,$C3,$B6,$1F,$D5,$2F,$6F,$79,$F9,$8D,$78
	.dc.b	$5F,$87,$0E,$F7,$CA,$7F,$07,$87,$68,$C4,$F9,$23,$15,$DA,$B7,$85
	.dc.b	$AD,$37,$0F,$9E,$D9,$F8,$F7,$CA,$7C,$98,$7D,$06,$E1,$EF,$BE,$17
	.dc.b	$F3,$67,$27,$AE,$F9,$47,$C9,$8E,$E6,$9C,$1F,$08,$A7,$C2,$7E,$41
	.dc.b	$BC,$3D,$2D,$39,$C7,$0C,$3E,$1C,$7E,$3A,$FC,$B1,$E9,$9D,$61,$E9
	.dc.b	$AE,$51,$F0,$2B,$0F,$8B,$0F,$4E,$1F,$9D,$9C,$BE,$29,$D4,$7E,$33
	.dc.b	$B7,$F8,$09,$2F,$F8,$4B,$67,$F7,$27,$F2,$87,$0F,$F0,$C6,$BF,$DD
	.dc.b	$5A,$EE,$3F,$89,$F6,$FF,$0E,$23,$82,$47,$63,$FB,$44,$69,$C6,$48
	.dc.b	$C4,$7F,$1C,$6B,$FE,$05,$7F,$8D,$F6,$CA,$FE,$46,$D9,$E6,$D7,$F9
	.dc.b	$2B,$5F,$F0,$73,$5F,$E8,$ED,$7F,$BE,$76,$FF,$28,$24,$FF,$30,$6C
	.dc.b	$7A,$1A,$FF,$32,$6B,$FE,$1B,$6B,$FD,$51,$FE,$67,$DB,$27,$F9,$DB
	.dc.b	$67,$E1,$B5,$FE,$7A,$D7,$FC,$4C,$D7,$FA,$FB,$5F,$F0,$A4,$7F,$40
	.dc.b	$21,$FD,$20,$AF,$D0,$6B,$FD,$29,$AF,$F8,$CD,$AF,$F6,$C7,$FC,$E5
	.dc.b	$F3,$6E,$EA,$48,$FB,$D2,$9D,$46,$1B,$1F,$83,$6B,$F9,$26,$BF,$D3
	.dc.b	$1A,$FF,$57,$6B,$D6,$7C,$0F,$37,$CB,$DD,$2B,$C0,$D9,$F0,$35,$F0
	.dc.b	$35,$F0,$35,$F0,$35,$F6,$7C,$9F,$C2,$24,$FB,$5B,$1E,$D6,$BE,$D6
	.dc.b	$BE,$D6,$BE,$D7,$E6,$FB,$DF,$E8,$DE,$07,$EB,$4F,$93,$FF,$46,$E6
	.dc.b	$AD,$4E,$71,$FE,$F9,$67,$E5,$4E,$AF,$8B,$63,$66,$8B,$8F,$FA,$9E
	.dc.b	$FA,$3F,$E0,$0C,$3C,$13,$AC,$3C,$B8,$74,$8C,$5F,$09,$F1,$5F,$D9
	.dc.b	$C3,$C0,$F3,$6C,$6A,$3F,$11,$18,$AF,$24,$7F,$88,$A7,$BC,$79,$A3
	.dc.b	$6F,$82,$FC,$51,$BB,$F1,$4F,$2A,$FF,$7E,$46,$27,$C9,$3C,$4E,$98
	.dc.b	$7B,$A3,$95,$F4,$BE,$2D,$AF,$9F,$7C,$EB,$C6,$7E,$77,$0F,$B7,$3F
	.dc.b	$A1,$6C,$66,$30,$47,$04,$3F,$49,$1C,$1B,$1E,$68,$FC,$0F,$99,$E3
	.dc.b	$9F,$59,$F3,$67,$E9,$C6,$63,$CB,$87,$CF,$C7,$DC,$FC,$49,$D4,$62
	.dc.b	$BE,$41,$EE,$8C,$DF,$7A,$FC,$AC,$62,$7C,$87,$80,$B7,$F8,$9E,$EF
	.dc.b	$CE,$7D,$6B,$4F,$CD,$C6,$1B,$5F,$8D,$9D,$3F,$52,$74,$D3,$53,$C2
	.dc.b	$BF,$0D,$5C,$1F,$04,$E8,$E0,$EA,$78,$61,$F9,$61,$5C,$5F,$8D,$7E
	.dc.b	$0A,$E6,$71,$BE,$73,$C6,$B8,$47,$F1,$9A,$FE,$4E,$47,$F9,$13,$6B
	.dc.b	$F9,$13,$5F,$E4,$4D,$79,$47,$02,$3C,$1F,$09,$FC,$02,$5E,$97,$B2
	.dc.b	$38,$CF,$A2,$FC,$D3,$98,$FE,$59,$76,$7E,$0C,$E7,$A5,$E2,$BB,$3E
	.dc.b	$36,$3F,$E8,$99,$CF,$87,$8C,$91,$C2,$B0,$EA,$BF,$86,$B9,$6D,$AF
	.dc.b	$6F,$78,$CC,$62,$B2,$D8,$C9,$CA,$32,$D3,$37,$FE,$AA,$BF,$03,$6F
	.dc.b	$C1,$3B,$8C,$5F,$56,$C6,$71,$FE,$37,$7E,$B9,$FE,$49,$EF,$7F,$64
	.dc.b	$25,$7F,$5F,$6C,$7F,$5F,$6B,$FD,$7D,$AF,$F5,$F6,$BF,$D7,$DF,$ED
	.dc.b	$FF,$21,$27,$FB,$53,$67,$FB,$53,$5F,$ED,$4D,$7F,$B5,$35,$FE,$D4
	.dc.b	$D7,$FB,$B7,$87,$FD,$F0,$93,$FD,$DD,$B1,$FD,$DD,$AF,$F7,$76,$BF
	.dc.b	$DD,$DA,$FF,$77,$7F,$C0,$FE,$FC,$9F,$F0,$13,$67,$FC,$04,$D7,$FC
	.dc.b	$04,$D7,$FC,$04,$D7,$FC,$04,$D7,$FC,$1D,$F2,$1F,$E1,$42,$4F,$F8
	.dc.b	$3D,$B1,$FE,$0F,$6B,$FE,$0F,$6B,$FE,$0F,$6B,$FE,$0F,$78,$E5,$FC
	.dc.b	$9D,$FE,$E3,$1F,$DC,$99,$3F,$D2,$1C,$B4,$91,$FD,$A6,$32,$D8,$FE
	.dc.b	$9E,$FF,$8B,$08,$FF,$89,$DA,$7F,$89,$CA,$7F,$89,$CA,$7F,$89,$CA
	.dc.b	$7F,$89,$CF,$ED,$87,$ED,$F4,$79,$FF,$AE,$3F,$5F,$E0,$5A,$FD,$9C
	.dc.b	$5A,$FD,$34,$DC,$3F,$43,$7C,$AF,$67,$EE,$A7,$4F,$7B,$B7,$FC,$12
	.dc.b	$BF,$DC,$7A,$FF,$0C,$9C,$B6,$B9,$1D,$2B,$F6,$D3,$AA,$E2,$78,$63
	.dc.b	$F7,$5E,$77,$F0,$8F,$3F,$F8,$4B,$6F,$F8,$6B,$67,$71,$FC,$42,$76
	.dc.b	$D3,$65,$8D,$9F,$C1,$E3,$6D,$8D,$B4,$DE,$5F,$1A,$76,$D9,$DB,$67
	.dc.b	$F8,$97,$9D,$FC,$83,$CF,$FE,$42,$DA,$FE,$4A,$D9,$DC,$7F,$28,$9D
	.dc.b	$B4,$D9,$63,$67,$F1,$F8,$DB,$63,$6D,$37,$5F,$7A,$76,$D9,$DB,$67
	.dc.b	$F9,$57,$89,$FC,$E0,$FE,$73,$E7,$DA,$FE,$7B,$16,$7F,$98,$4D,$AF
	.dc.b	$C5,$CD,$AF,$CB,$4D,$AF,$C4,$4D,$D7,$FA,$9C,$DD,$13,$63,$0D,$BF
	.dc.b	$09,$FD,$3B,$C3,$FE,$9B,$E2,$7F,$4E,$6C,$64,$FE,$A7,$19,$69,$9C
	.dc.b	$7F,$A4,$46,$5B,$19,$CB,$F4,$B3,$96,$C6,$4F,$D9,$46,$5A,$66,$7F
	.dc.b	$4D,$19,$6C,$7F,$97,$E5,$E4,$7E,$F2,$7F,$CF,$ED,$51,$A9,$FF,$59
	.dc.b	$B6,$3F,$DE,$4D,$7F,$E0,$0B,$51,$D2,$38,$CF,$59,$D4,$E4,$E7,$19
	.dc.b	$35,$EC,$CA,$F0,$B6,$7C,$2D,$7C,$2D,$7C,$27,$A9,$C3,$FE,$9C,$6B
	.dc.b	$EE,$F8,$5F,$21,$27,$DE,$D8,$F7,$B5,$F7,$B5,$EF,$68,$C4,$7F,$AA
	.dc.b	$9F,$81,$F0,$92,$BE,$5B,$67,$E5,$B5,$F9,$6D,$7A,$B5,$FF,$5A,$B5
	.dc.b	$FA,$FE,$D7,$CD,$24,$FD,$86,$C7,$D8,$6B,$F6,$1A,$FA,$9A,$FF,$B0
	.dc.b	$9E,$37,$FE,$AD,$0F,$F8,$DE,$3E,$7F,$C4,$FE,$03,$E7,$1D,$9C,$F0
	.dc.b	$FC,$85,$EF,$0E,$31,$FE,$B1,$C3,$F0,$F8,$71,$BE,$11,$F9,$29,$B5
	.dc.b	$E1,$BF,$04,$F0,$18,$78,$1F,$45,$75,$58,$75,$72,$70,$C3,$F0,$EF
	.dc.b	$3A,$E4,$FC,$CA,$E0,$FB,$8E,$07,$28,$DB,$C6,$7D,$31,$88,$F0,$9C
	.dc.b	$0E,$85,$38,$1C,$E3,$C6,$F0,$8F,$1E,$1F,$22,$38,$19,$9E,$E7,$0B
	.dc.b	$F3,$CE,$EB,$C5,$18,$C3,$FD,$D7,$18,$BF,$25,$6E,$F9,$46,$2B,$71
	.dc.b	$8A,$F1,$C6,$27,$E4,$DF,$5C,$3E,$3C,$7B,$0F,$7D,$77,$38,$C6,$A3
	.dc.b	$D4,$7C,$13,$99,$E1,$75,$3D,$63,$E0,$B6,$7A,$5F,$C2,$9E,$31,$D9
	.dc.b	$D9,$CC,$AE,$4E,$2F,$84,$FB,$CE,$63,$E0,$1D,$63,$2D,$7D,$D8,$7C
	.dc.b	$97,$C3,$19,$7D,$A7,$C2,$79,$BD,$CE,$38,$7C,$B7,$A9,$93,$8C,$F5
	.dc.b	$72,$73,$3B,$9B,$BF,$89,$5D,$A3,$47,$09,$E4,$F8,$30,$F9,$AF,$C5
	.dc.b	$C3,$E1,$BC,$0D,$1F,$22,$BD,$97,$DA,$3C,$26,$CF,$3E,$1E,$B7,$8D
	.dc.b	$74,$C3,$C3,$1A,$C3,$C3,$3F,$3E,$39,$CE,$AF,$B4,$7A,$6F,$A3,$C4
	.dc.b	$F2,$1F,$04,$F5,$61,$F0,$6B,$94,$60,$EB,$3C,$AB,$AC,$79,$AB,$84
	.dc.b	$E6,$38,$1E,$1C,$3C,$F3,$F0,$A3,$27,$13,$AC,$69,$F2,$61,$ED,$79
	.dc.b	$1E,$83,$E4,$5F,$89,$F6,$47,$C7,$BF,$13,$C6,$3B,$1C,$9D,$57,$78
	.dc.b	$E6,$79,$0E,$77,$EC,$8D,$61,$F1,$2B,$A5,$68,$F4,$9E,$58,$D5,$7B
	.dc.b	$E7,$78,$79,$C9,$5D,$DB,$1A,$8C,$11,$C1,$27,$A1,$23,$E3,$B5,$F5
	.dc.b	$C6,$DC,$90,$CC,$6E,$BC,$F1,$96,$BA,$31,$87,$CC,$9E,$0D,$8C,$E1
	.dc.b	$F1,$A3,$11,$C4,$95,$D0,$91,$A6,$CF,$12,$4F,$CB,$9D,$12,$33,$7F
	.dc.b	$6A,$75,$18,$BF,$B1,$18,$24,$EC,$91,$A6,$C7,$DA,$6B,$F8,$78,$D1
	.dc.b	$1D,$35,$D1,$82,$18,$21,$F9,$22,$1F,$89,$32,$53,$38,$7E,$12,$3E
	.dc.b	$7D,$7B,$67,$A9,$A3,$F0,$D1,$A7,$DE,$F9,$5D,$57,$E0,$DF,$A7,$7F
	.dc.b	$2D,$D9,$D0,$F1,$12,$FC,$4F,$DE,$36,$7A,$1C,$BD,$2F,$CC,$E4,$F2
	.dc.b	$9C,$6B,$E6,$BE,$4B,$F4,$46,$23,$C0,$70,$3A,$1E,$03,$81,$CE,$3D
	.dc.b	$CF,$08,$F7,$61,$F8,$C8,$E0,$66,$7B,$9C,$2F,$D3,$3B,$AF,$14,$63
	.dc.b	$0F,$8D,$18,$BF,$25,$6E,$F9,$46,$2B,$71,$8A,$F1,$C6,$27,$E7,$5F
	.dc.b	$5C,$3E,$3C,$7B,$0F,$7D,$77,$38,$C6,$A3,$D4,$7C,$13,$99,$E1,$75
	.dc.b	$3D,$63,$E0,$B6,$7A,$5F,$C2,$9F,$F0,$4F,$BF,$FE,$01,$3F,$C3,$C5
	.dc.b	$8F,$F0,$EB,$63,$71,$FE,$50,$8D,$B4,$D9,$63,$CB,$3E,$A7,$C6,$79
	.dc.b	$CA,$E4,$F2,$BE,$D3,$FC,$94,$7F,$8F,$8F,$F2,$29,$6B,$0F,$F9,$74
	.dc.b	$C1,$87,$2F,$03,$FC,$6E,$68,$86,$08,$7F,$AD,$3F,$21,$6A,$FC,$93
	.dc.b	$C5,$18,$6C,$66,$7F,$CA,$73,$96,$C6,$5F,$F3,$4C,$65,$B1,$97,$FD
	.dc.b	$67,$19,$6C,$65,$F2,$35,$46,$DC,$CF,$F9,$DB,$F2,$31,$00

;--------------------------------------------------------------------------------
;	FDSQRT.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fdsqrt_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FDSQRT.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FDSQRT.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fdsqrt_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdsqrt.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDSQRT.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdsqrt.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDSQRT.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fdsqrt_data_compressed::
	.dc.b	$00,$02,$A1,$14,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$FD
	.dc.b	$4F,$E5,$1F,$97,$BE,$4B,$CF,$E9,$7A,$3D,$2D,$4B,$E1,$2F,$C9,$E3
	.dc.b	$E4,$73,$F1,$7A,$F8,$8D,$FB,$9C,$3D,$3E,$3E,$9F,$2F,$4F,$9F,$A7
	.dc.b	$D3,$D3,$EB,$E9,$F6,$F4,$FB,$FA,$7E,$0F,$4F,$C3,$E9,$F8,$BD,$3F
	.dc.b	$97,$EC,$7C,$CF,$63,$E6,$FB,$1F,$D8,$FF,$14,$FE,$40,$7E,$63,$C5
	.dc.b	$3F,$98,$7F,$B9,$3F,$D9,$9B,$BF,$1D,$F9,$6F,$CF,$7E,$9B,$F5,$DF
	.dc.b	$B6,$FD,$F7,$F0,$6F,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$FE,$ED,$F8,$AB
	.dc.b	$E6,$2F,$C7,$FE,$13,$DC,$FE,$F5,$F3,$9F,$47,$8F,$FC,$0D,$7F,$E7
	.dc.b	$B8,$FA,$11,$FE,$09,$7F,$DB,$87,$FA,$50,$FF,$03,$1F,$CF,$CF,$80
	.dc.b	$7B,$D3,$EF,$6F,$1F,$83,$8F,$C3,$C7,$E2,$E3,$F1,$F1,$F9,$38,$FC
	.dc.b	$BC,$7E,$6E,$3F,$3F,$1F,$A3,$8F,$D3,$C7,$EA,$E3,$FF,$0A,$77,$7B
	.dc.b	$F7,$FF,$10,$3F,$ED,$CE,$1F,$F2,$6B,$FE,$97,$2F,$C6,$F0,$F8,$AE
	.dc.b	$3F,$05,$CB,$DE,$E7,$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83
	.dc.b	$DE,$F0,$FB,$DE,$2F,$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$F5,$38,$3C
	.dc.b	$FE,$F7,$A3,$DE,$FF,$16,$7C,$EF,$F8,$14,$7C,$5E,$3E,$2F,$3F,$17
	.dc.b	$AF,$8B,$DF,$C5,$F0,$F8,$BE,$3F,$11,$CB,$E1,$79,$FC,$2F,$4F,$85
	.dc.b	$EB,$F0,$BD,$BE,$17,$BF,$C2,$F8,$3E,$17,$C3,$F0,$BE,$2F,$83,$F1
	.dc.b	$FC,$1F,$93,$DB,$F2,$FB,$7E,$6F,$6F,$CF,$EC,$7F,$90,$FE,$7B,$F1
	.dc.b	$5E,$61,$FC,$4B,$7C,$BD,$7C,$16,$FE,$0B,$87,$C1,$71,$F8,$2E,$5F
	.dc.b	$05,$CF,$E0,$BF,$33,$E8,$F5,$F7,$BB,$7B,$DD,$FD,$EF,$07,$BD,$E1
	.dc.b	$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD,$EF,$79,$FD,$EF
	.dc.b	$47,$BD,$FE,$5B,$F9,$FF,$C4,$73,$BD,$FF,$E7,$2A,$1F,$03,$BF,$8A
	.dc.b	$E1,$F2,$BC,$7D,$5E,$5C,$BC,$FE,$0B,$A7,$C1,$75,$F8,$2E,$DF,$05
	.dc.b	$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE,$0B,$C9,$F0,$5E
	.dc.b	$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$7B,$AF,$EF,$B5,$FD,$FE,$BF,$C0
	.dc.b	$35,$FE,$03,$AF,$B7,$E0,$BD,$DF,$05,$FE,$79,$F7,$FF,$CA,$47,$E3
	.dc.b	$BA,$3F,$D0,$CB,$E0,$34,$EF,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1C
	.dc.b	$FE,$03,$A7,$C0,$7F,$10,$E8,$ED,$F0,$1D,$FE,$03,$C1,$F0,$1F,$C5
	.dc.b	$3A,$FF,$8A,$F5,$FF,$16,$EB,$FE,$2F,$D7,$FC,$63,$AF,$F8,$CF,$5F
	.dc.b	$F1,$AE,$6F,$E3,$7C,$DF,$E9,$1F,$C5,$BF,$CD,$47,$A7,$9C,$FA,$5F
	.dc.b	$F5,$03,$FE,$FC,$3F,$D6,$C7,$F8,$98,$FE,$BC,$7F,$5D,$78,$F0,$F5
	.dc.b	$E1,$ED,$C1,$DF,$83,$E4,$7C,$0F,$C9,$F8,$1F,$95,$F0,$3F,$2F,$E0
	.dc.b	$7E,$67,$C0,$FC,$DF,$81,$F9,$DF,$03,$F3,$FE,$07,$E8,$7C,$0F,$FA
	.dc.b	$A7,$E8,$7F,$4E,$7E,$98,$7C,$0E,$7E,$07,$5F,$03,$BF,$81,$E1,$F0
	.dc.b	$3C,$7E,$07,$97,$C0,$F3,$F8,$1E,$9F,$03,$D7,$E0,$7B,$7C,$0F,$7F
	.dc.b	$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03,$F2,$FC,$9F,$99,$D7,$F3,$7A
	.dc.b	$FE,$77,$5F,$CF,$EB,$FA,$1D,$7F,$47,$AF,$E9,$75,$FD,$3E,$BF,$A9
	.dc.b	$D7,$FE,$C5,$F4,$FF,$81,$9F,$C0,$8F,$6F,$83,$FE,$D3,$5D,$7B,$EB
	.dc.b	$E1,$D7,$C7,$AF,$97,$5F,$3E,$BE,$9D,$7F,$27,$E0,$7E,$57,$C0,$FC
	.dc.b	$BF,$81,$F0,$77,$7C,$DF,$81,$FF,$6B,$7D,$17,$EF,$0F,$DC,$F0,$7F
	.dc.b	$B8,$D7,$BF,$AF,$7F,$7E,$FF,$0F,$7F,$8F,$BF,$CB,$DF,$E7,$EF,$F4
	.dc.b	$E7,$EB,$EF,$FB,$FC,$8F,$81,$E4,$7C,$1F,$23,$E1,$79,$1F,$0F,$C8
	.dc.b	$F1,$FB,$FE,$4F,$67,$CB,$EC,$F9,$BD,$9F,$3F,$B3,$E8,$F6,$7D,$3E
	.dc.b	$CF,$AB,$D9,$F5,$FB,$3E,$CF,$67,$FD,$E5,$F4,$7F,$EC,$B3,$F4,$47
	.dc.b	$E8,$7B,$0B,$DF,$DF,$BF,$C3,$DF,$E3,$EF,$F2,$F7,$FE,$C7,$B1,$D3
	.dc.b	$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F
	.dc.b	$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$F0,$DE,$A7
	.dc.b	$AB,$DC,$74,$4F,$3F,$FA,$59,$FD,$28,$FC,$4A,$7F,$12,$EF,$DB,$E2
	.dc.b	$E7,$DB,$B3,$FE,$33,$3F,$C2,$87,$F8,$4C,$7B,$1A,$F5,$B7,$E6,$70
	.dc.b	$F3,$38,$F9,$9C,$BC,$CE,$7E,$67,$4E,$CE,$BB,$F6,$F6,$3B,$FB,$1E
	.dc.b	$0F,$63,$C7,$F4,$9F,$D3,$0F,$C8,$F0,$9F,$C8,$BF,$01,$F6,$37,$D7
	.dc.b	$E3,$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$F6,$1A,$F8
	.dc.b	$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F,$5F
	.dc.b	$CF,$EB,$FA,$3D,$7F,$4F,$AF,$F3,$3E,$97,$E6,$F2,$9F,$E6,$07,$E3
	.dc.b	$CF,$C7,$3F,$69,$7A,$DC,$7D,$6E,$5E,$B7,$F0,$5E,$BE,$9E,$B7,$5F
	.dc.b	$5B,$B7,$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$78,$FD,$6F,$27
	.dc.b	$AD,$E5,$F5,$BC,$DE,$B7,$9F,$D6,$F4,$7A,$DE,$9F,$5B,$D5,$EB,$7E
	.dc.b	$1B,$E9,$BF,$C2,$E7,$F8,$58,$FC,$69,$F8,$CE,$72,$F5,$F8,$7A,$FC
	.dc.b	$7D,$EF,$E3,$DE,$77,$3F,$7B,$A7,$BD,$F9,$2F,$C5,$FF,$C1,$67,$E0
	.dc.b	$0F,$BF,$91,$FB,$ED,$F4,$38,$7A,$1C,$7D,$0E,$5E,$87,$3F,$43,$A7
	.dc.b	$A1,$D7,$D0,$ED,$E7,$F7,$F3,$FC,$1E,$7F,$87,$CF,$F1,$79,$FE,$3F
	.dc.b	$3F,$C9,$E7,$FC,$1F,$6B,$E1,$7B,$5F,$0F,$DA,$FF,$57,$FA,$7F,$EB
	.dc.b	$0F,$33,$FD,$63,$D3,$FE,$B7,$FC,$C3,$FC,$D9,$F2,$47,$F9,$5B,$7E
	.dc.b	$03,$5E,$F6,$FA,$78,$74,$F1,$E9,$E5,$D3,$CF,$A7,$A7,$4F,$5E,$9E
	.dc.b	$DD,$3D,$FA,$7C,$1D,$3F,$5F,$DE,$FB,$1E,$F7,$8F,$AF,$ED,$7B,$DF
	.dc.b	$6F,$DE,$FB,$9E,$F7,$DD,$F7,$BE,$F7,$4F,$DF,$E9,$FC,$07,$4F,$F0
	.dc.b	$6F,$82,$FF,$1F,$9F,$B4,$3E,$A9,$F5,$07,$36,$F9,$B8,$73,$71,$E6
	.dc.b	$E5,$CD,$CF,$9B,$A7,$37,$5E,$6E,$DC,$DD,$F9,$BC,$1C,$9E,$1F,$03
	.dc.b	$C5,$E0,$78,$FC,$0F,$27,$BD,$E5,$F7,$BC,$DE,$F7,$F1,$CF,$A9,$FD
	.dc.b	$89,$FE,$1F,$3F,$9F,$1F,$B0,$F1,$4F,$EC,$1F,$E5,$6B,$DD,$E5,$EE
	.dc.b	$F3,$F7,$7A,$7B,$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF
	.dc.b	$17,$BB,$E3,$F7,$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE
	.dc.b	$FA,$7D,$DF,$E6,$BF,$55,$FE,$86,$3F,$D0,$A7,$F3,$C3,$F6,$07,$EB
	.dc.b	$F9,$7F,$A1,$AF,$73,$8F,$B9,$CB,$DC,$E7,$EE,$74,$F6,$FA,$FB,$7D
	.dc.b	$BD,$BE,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F
	.dc.b	$2F,$AF,$E6,$F5,$FC,$FE,$BF,$F4,$CF,$AB,$FF,$42,$9F,$CE,$8F,$DD
	.dc.b	$9F,$BA,$3F,$0F,$C9,$FD,$69,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7
	.dc.b	$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F
	.dc.b	$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5,$BF,$B1,$FC,$07,$F3,$83
	.dc.b	$FC,$CC,$7F,$81,$4F,$C4,$1F,$87,$E4,$FE,$6F,$EB,$70,$F5,$B8,$FA
	.dc.b	$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD,$DB,$D6,$EF,$EB,$78,$3D,$6F
	.dc.b	$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6,$F2,$FA,$DE,$6F,$5B,$CF,$EB
	.dc.b	$7F,$7A,$FA,$CF,$F6,$59,$FE,$CA,$3F,$C0,$27,$F3,$53,$F0,$DC,$87
	.dc.b	$F0,$CF,$F8,$29,$7A,$BC,$FD,$5E,$9E,$AF,$5F,$57,$B7,$AB,$DF,$D5
	.dc.b	$F0,$7A,$BE,$1F,$57,$C5,$EA,$F8,$FD,$5F,$27,$AB,$E5,$F5,$7C,$DE
	.dc.b	$AF,$9F,$D5,$F4,$7A,$BF,$E1,$2F,$AD,$FF,$65,$9F,$E0,$03,$FA,$09
	.dc.b	$FD,$00,$FC,$3F,$19,$FC,$3B,$FE,$22,$7F,$DA,$23,$D3,$E9,$E9,$F5
	.dc.b	$F4,$FB,$7A,$7D,$FD,$3F,$07,$A7,$E1,$F4,$FC,$5E,$9F,$8F,$D3,$F2
	.dc.b	$7A,$7E,$5F,$4F,$CD,$E9,$F9,$FD,$3F,$47,$A7,$FE,$2A,$FA,$EF,$EF
	.dc.b	$E7,$FA,$B4,$FF,$26,$1F,$C3,$CF,$C4,$1F,$C3,$78,$BF,$C7,$2B,$D2
	.dc.b	$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9,$78
	.dc.b	$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5E,$7F,$4B
	.dc.b	$D1,$E9,$7F,$90,$BD,$B7,$CC,$5F,$97,$C7,$C5,$E7,$DF,$D7,$C2,$EF
	.dc.b	$DE,$E1,$EC,$71,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1
	.dc.b	$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6
	.dc.b	$3C,$DE,$C7,$F9,$63,$EB,$DF,$2F,$7E,$47,$1F,$0B,$9F,$85,$D7,$66
	.dc.b	$FE,$0B,$87,$C1,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7
	.dc.b	$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$E5
	.dc.b	$F7,$BC,$DE,$F7,$F9,$BB,$EC,$3F,$C8,$67,$FB,$C4,$FF,$48,$1F,$D4
	.dc.b	$CF,$E1,$87,$E1,$7D,$F3,$F8,$57,$FC,$FE,$FF,$BC,$47,$B9,$D7,$DC
	.dc.b	$ED,$EE,$77,$F7,$3C,$1E,$E7,$87,$DC,$F1,$7B,$9E,$3F,$73,$C9,$EE
	.dc.b	$79,$7D,$CF,$37,$B9,$E7,$F7,$3D,$1E,$E7,$A7,$DC,$FF,$43,$FD,$8F
	.dc.b	$E1,$B8,$BF,$D2,$2B,$DC,$CF,$B9,$AF,$73,$7E,$E7,$0F,$73,$8F,$B9
	.dc.b	$CB,$DC,$E7,$EE,$74,$F7,$3A,$FB,$9D,$BD,$CE,$FE,$E7,$83,$DC,$F0
	.dc.b	$FB,$9E,$2F,$73,$C7,$EE,$79,$3D,$CF,$2F,$B9,$E6,$F7,$3C,$FE,$E7
	.dc.b	$FA,$7B,$8B,$F0,$FC,$FF,$EA,$87,$FD,$1A,$7F,$93,$8F,$F2,$69,$FC
	.dc.b	$38,$FE,$1A,$FF,$96,$FD,$5E,$5E,$AF,$3F,$57,$A7,$AB,$D7,$D5,$ED
	.dc.b	$EA,$F7,$F5,$7C,$1E,$AF,$87,$D5,$F1,$7A,$BE,$3F,$57,$C9,$EA,$F9
	.dc.b	$7D,$5F,$37,$AB,$E7,$F5,$7F,$D6,$BF,$65,$FE,$8A,$1F,$31,$8F,$95
	.dc.b	$CF,$C8,$EB,$E2,$37,$F1,$1C,$3E,$17,$8F,$C2,$F2,$F7,$39,$FB,$9D
	.dc.b	$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC,$F1
	.dc.b	$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$FE,$CC,$FB,$3F,$C3
	.dc.b	$79,$07,$F0,$CF,$FB,$61,$FF,$6A,$9F,$E8,$93,$FA,$A9,$FD,$51,$E1
	.dc.b	$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D
	.dc.b	$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1
	.dc.b	$EC,$7F,$B9,$3E,$D3,$F0,$FC,$47,$FD,$A6,$7F,$A5,$0F,$F4,$99,$FD
	.dc.b	$50,$FC,$38,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1,$DF
	.dc.b	$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6,$3C
	.dc.b	$DE,$C7,$9F,$D8,$FF,$7D,$FA,$7F,$ED,$03,$F0,$FC,$63,$FC,$B9,$EB
	.dc.b	$6B,$D6,$DF,$AD,$C2,$B8,$FA,$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD
	.dc.b	$DB,$D6,$EF,$EB,$78,$3D,$6F,$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6
	.dc.b	$F2,$FA,$DE,$6F,$5B,$CF,$EB,$7A,$3D,$6F,$37,$33,$ED,$7F,$12,$7E
	.dc.b	$21,$C7,$BB,$A8,$FF,$68,$1F,$E1,$C3,$F1,$07,$F8,$6A,$3E,$02,$F5
	.dc.b	$B8,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77
	.dc.b	$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B,$C9,$EB,$7A,$BE,$DB
	.dc.b	$F1,$23,$E1,$B1,$F0,$D9,$F8,$6D,$7C,$36,$FE,$1B,$87,$B7,$C7,$DB
	.dc.b	$E5,$ED,$F3,$F6,$FA,$7B,$7D,$7D,$BE,$DE,$DF,$7F,$6F,$C1,$ED,$F8
	.dc.b	$7D,$BF,$17,$B7,$E3,$F6,$FC,$9E,$DF,$97,$DB,$F3,$7B,$7F,$43,$ED
	.dc.b	$FE,$8F,$8E,$7F,$12,$DF,$73,$5E,$E6,$FD,$CE,$1E,$E7,$1F,$73,$97
	.dc.b	$B9,$CF,$DC,$E9,$EE,$75,$F7,$3B,$7B,$9D,$FD,$CF,$07,$B9,$E1,$F7
	.dc.b	$3C,$5E,$E7,$8F,$DC,$F2,$7B,$9E,$5F,$73,$CD,$EE,$79,$FD,$CF,$47
	.dc.b	$B9,$F8,$8F,$1F,$F8,$79,$F8,$71,$EE,$67,$E1,$35,$F0,$7B,$F8,$4E
	.dc.b	$1F,$03,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D,$FE,$DE,$FF,$7F
	.dc.b	$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E,$FF,$97,$DF,$F3
	.dc.b	$7B,$FF,$9E,$FB,$8F,$E1,$A7,$E1,$7C,$BF,$DB,$2F,$7B,$5E,$F6,$FD
	.dc.b	$EE,$1D,$5C,$7D,$EE,$5E,$F7,$3F,$7B,$A7,$BD,$D7,$DE,$ED,$EF,$77
	.dc.b	$F7,$BC,$1E,$F7,$87,$DE,$F1,$7B,$DE,$3F,$7B,$C9,$EF,$79,$7D,$EF
	.dc.b	$37,$BD,$E7,$F7,$BF,$6B,$F7,$3F,$86,$3F,$0A,$3D,$EC,$FB,$DA,$F7
	.dc.b	$B7,$EF,$70,$F7,$B8,$FB,$DC,$BD,$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB
	.dc.b	$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD,$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2
	.dc.b	$FB,$DE,$6F,$7B,$CF,$EF,$7F,$0C,$FB,$AF,$EA,$67,$F0,$B3,$F0,$BE
	.dc.b	$67,$F2,$55,$EF,$6F,$DE,$E1,$EF,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E
	.dc.b	$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD
	.dc.b	$EF,$27,$BD,$E5,$F7,$BC,$DE,$F7,$9F,$DE,$FE,$47,$C3,$FD,$48,$FC
	.dc.b	$37,$2F,$FB,$F7,$DD,$CF,$BB,$AF,$77,$7E,$EF,$0F,$77,$8F,$BB,$CB
	.dc.b	$DD,$E7,$EE,$F4,$F7,$7A,$FB,$BD,$BD,$DE,$FE,$EF,$83,$DD,$F0,$FB
	.dc.b	$BE,$2F,$77,$C7,$EE,$F9,$3D,$DF,$2F,$BB,$E6,$F7,$7F,$9C,$FD,$DF
	.dc.b	$F0,$A3,$F8,$49,$F8,$5E,$6F,$EA,$4B,$DC,$DF,$B9,$C3,$DC,$E3,$EE
	.dc.b	$72,$F7,$39,$FB,$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C
	.dc.b	$3E,$E7,$8B,$DC,$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9
	.dc.b	$E8,$F7,$3F,$A8,$FD,$E7,$F8,$60,$FE,$A0,$7E,$1B,$9C,$FE,$19,$FE
	.dc.b	$D8,$FF,$BB,$0F,$F0,$C1,$FD,$45,$DF,$B3,$D3,$D9,$EB,$EC,$F6,$F6
	.dc.b	$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE
	.dc.b	$CF,$9B,$D9,$F3,$FB,$3E,$8F,$67,$FB,$4F,$DE,$FF,$86,$0F,$C3,$9F
	.dc.b	$C2,$F8,$7F,$C0,$4B,$D8,$DF,$B1,$C3,$D8,$E3,$EC,$72,$F6,$39,$FB
	.dc.b	$1D,$3D,$8E,$BE,$C7,$6F,$63,$BF,$B1,$E0,$F6,$3C,$3E,$C7,$8B,$D8
	.dc.b	$F1,$FB,$1E,$4F,$63,$CB,$EC,$79,$BD,$8F,$3F,$B1,$FE,$00,$F5,$3F
	.dc.b	$A8,$9F,$D4,$0F,$C4,$73,$7F,$B9,$7D,$7D,$7A,$FB,$F5,$F8,$57,$1F
	.dc.b	$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F,$07,$AF
	.dc.b	$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB,$F9,$FD
	.dc.b	$7F,$F0,$BF,$DF,$7F,$97,$8F,$F0,$B1,$F8,$73,$F0,$DC,$C5,$EE,$F0
	.dc.b	$F7,$38,$FB,$BC,$BD,$AE,$7E,$D7,$4F,$6B,$AF,$B5,$DB,$DA,$EF,$ED
	.dc.b	$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7,$93,$DA,$F2,$FB,$5E,$6F
	.dc.b	$6B,$CF,$ED,$7F,$8B,$BE,$FF,$FC,$BA,$7F,$0C,$3F,$0D,$CA,$7F,$0C
	.dc.b	$FF,$92,$57,$B3,$C7,$D9,$E5,$EC,$F3,$F6,$7A,$7B,$3D,$7D,$97,$F0
	.dc.b	$0C,$2E,$5C,$E1,$F8,$07,$F9,$57,$E1,$9C,$FC,$63,$0D,$AF,$EF,$6D
	.dc.b	$9D,$C7,$39,$DB,$4D,$96,$36,$FB,$6B,$C2,$E8,$C4,$F2,$69,$9D,$53
	.dc.b	$C2,$B9,$C6,$2B,$83,$6B,$85,$6E,$FD,$04,$8F,$23,$5E,$F7,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF
	.dc.b	$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63
	.dc.b	$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF
	.dc.b	$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88
	.dc.b	$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A
	.dc.b	$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB
	.dc.b	$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B
	.dc.b	$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D
	.dc.b	$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B
	.dc.b	$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60
	.dc.b	$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC
	.dc.b	$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE
	.dc.b	$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78
	.dc.b	$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12
	.dc.b	$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71
	.dc.b	$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C
	.dc.b	$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C
	.dc.b	$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6
	.dc.b	$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09
	.dc.b	$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC
	.dc.b	$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE
	.dc.b	$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A
	.dc.b	$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12
	.dc.b	$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D
	.dc.b	$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12
	.dc.b	$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2
	.dc.b	$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C
	.dc.b	$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09
	.dc.b	$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA
	.dc.b	$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB
	.dc.b	$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E
	.dc.b	$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD
	.dc.b	$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9
	.dc.b	$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A
	.dc.b	$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46
	.dc.b	$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31
	.dc.b	$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9
	.dc.b	$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26
	.dc.b	$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12
	.dc.b	$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3
	.dc.b	$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C
	.dc.b	$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F
	.dc.b	$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18
	.dc.b	$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A
	.dc.b	$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C
	.dc.b	$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67
	.dc.b	$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB
	.dc.b	$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D
	.dc.b	$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38
	.dc.b	$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E
	.dc.b	$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97
	.dc.b	$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D
	.dc.b	$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57
	.dc.b	$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA
	.dc.b	$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3
	.dc.b	$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF
	.dc.b	$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17
	.dc.b	$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF
	.dc.b	$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E
	.dc.b	$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D
	.dc.b	$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35
	.dc.b	$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1
	.dc.b	$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D
	.dc.b	$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4
	.dc.b	$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F
	.dc.b	$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D2,$12,$95,$2F,$10,$BE,$15,$BA,$D4,$F1,$9C,$BC,$AB
	.dc.b	$8C,$62,$77,$3C,$EF,$C2,$EC,$E4,$E2,$37,$F4,$6F,$83,$7C,$77,$92
	.dc.b	$FC,$B3,$98,$C4,$7A,$E3,$0E,$1A,$F1,$6B,$CA,$7A,$BB,$6B,$B6,$BB
	.dc.b	$9E,$EA,$79,$BE,$19,$DB,$67,$72,$F0,$1F,$27,$86,$D7,$9D,$AF,$9D
	.dc.b	$AF,$C0,$6B,$EF,$6B,$E3,$6C,$F7,$9E,$C4,$9E,$CD,$7B,$3F,$83,$FB
	.dc.b	$B7,$C1,$77,$8C,$36,$53,$A7,$E8,$73,$FE,$22,$31,$1E,$3A,$E2,$F9
	.dc.b	$DA,$E4,$E7,$5F,$72,$B9,$3D,$EB,$9D,$FA,$EB,$67,$32,$9F,$4E,$B8
	.dc.b	$B9,$9E,$38,$7C,$E6,$E1,$E2,$B5,$F0,$BF,$25,$69,$EB,$3E,$18,$F1
	.dc.b	$CE,$E3,$18,$79,$C9,$5F,$10,$ED,$87,$86,$33,$1D,$9F,$80,$EB,$08
	.dc.b	$F8,$27,$83,$CC,$C4,$6B,$0F,$31,$E4,$8C,$57,$6B,$F2,$C7,$19,$DC
	.dc.b	$E6,$FA,$CF,$C9,$38,$9F,$26,$39,$61,$EF,$9D,$DF,$4A,$F5,$1D,$6B
	.dc.b	$A6,$1C,$A3,$A1,$E8,$26,$1E,$27,$81,$DC,$87,$43,$8B,$B7,$15,$E1
	.dc.b	$3E,$CB,$DD,$C5,$F9,$B0,$F4,$5E,$63,$15,$63,$84,$EA,$3C,$B8,$7C
	.dc.b	$07,$46,$0F,$39,$E2,$AF,$86,$74,$74,$66,$BF,$1D,$5D,$70,$F5,$61
	.dc.b	$E9,$8C,$61,$F2,$2F,$F1,$F1,$87,$E1,$CE,$DF,$1D,$72,$9D,$47,$73
	.dc.b	$89,$AA,$F5,$3E,$4A,$F8,$B7,$E4,$8D,$1D,$AB,$9C,$EA,$39,$9C,$70
	.dc.b	$F7,$D7,$18,$C3,$F2,$9E,$67,$D6,$AE,$73,$C2,$39,$E1,$F0,$A3,$0F
	.dc.b	$E6,$63,$99,$E7,$9E,$35,$D6,$33,$3F,$05,$EF,$7F,$15,$CE,$1F,$02
	.dc.b	$7B,$57,$16,$CF,$53,$8E,$1F,$66,$32,$75,$8E,$67,$C5,$37,$3F,$06
	.dc.b	$37,$5D,$CF,$B4,$F1,$33,$3C,$4E,$67,$C5,$39,$9F,$12,$38,$9C,$DF
	.dc.b	$C0,$D7,$C6,$8F,$E0,$F1,$CA,$FE,$05,$76,$8E,$86,$6B,$D9,$5D,$23
	.dc.b	$C5,$3B,$AE,$33,$99,$E3,$39,$AE,$31,$E2,$BF,$95,$5B,$BE,$33,$93
	.dc.b	$E6,$1E,$19,$F8,$D3,$BA,$E7,$3F,$97,$BF,$8D,$39,$AD,$CF,$39,$CB
	.dc.b	$67,$E2,$3C,$5B,$7F,$97,$9D,$CE,$4F,$7C,$F5,$7E,$71,$CC,$C9,$4F
	.dc.b	$82,$EE,$3F,$28,$7C,$3A,$F2,$CF,$53,$67,$1A,$F8,$6E,$5A,$79,$6B
	.dc.b	$CF,$7F,$6E,$7C,$26,$DF,$8B,$3B,$8F,$2F,$CE,$BE,$79,$EF,$FF,$8D
	.dc.b	$BF,$00,$7C,$CF,$F8,$EE,$B5,$39,$8C,$57,$0B,$FF,$20,$5F,$05,$7D
	.dc.b	$7E,$07,$84,$62,$B8,$2A,$E1,$5D,$A3,$9D,$F6,$C3,$8C,$F3,$9C,$9C
	.dc.b	$E3,$37,$D2,$F9,$B6,$FC,$F8,$6B,$E0,$66,$1C,$B1,$F8,$51,$CB,$E0
	.dc.b	$73,$3E,$5B,$F7,$CE,$31,$EB,$56,$73,$5C,$F1,$5F,$38,$C7,$C0,$F9
	.dc.b	$BE,$06,$46,$08,$77,$3E,$33,$A7,$19,$7D,$79,$F7,$99,$CB,$DF,$58
	.dc.b	$AF,$67,$C2,$7F,$97,$67,$E6,$BF,$E5,$26,$F2,$7C,$C8,$D6,$7F,$7D
	.dc.b	$B5,$F6,$9E,$11,$99,$E1,$30,$EE,$F3,$30,$FA,$27,$FC,$BA,$F3,$3F
	.dc.b	$CB,$F8,$A3,$2D,$8C,$C6,$08,$E0,$87,$F9,$DA,$38,$36,$36,$DF,$8C
	.dc.b	$F7,$7C,$1C,$8F,$8B,$1F,$04,$C3,$E7,$8C,$3A,$3E,$5B,$AC,$7E,$5C
	.dc.b	$FC,$9F,$83,$95,$F4,$1F,$99,$F0,$78,$8F,$85,$16,$BE,$9C,$DC,$FE
	.dc.b	$84,$DA,$FA,$13,$F5,$7E,$0E,$47,$DC,$7E,$C7,$C1,$E2,$3E,$74,$5C
	.dc.b	$3E,$FC,$DC,$3E,$E4,$DA,$FB,$93,$F8,$1F,$83,$E7,$1F,$88,$8C,$3F
	.dc.b	$8A,$8C,$BF,$6E,$32,$FD,$D8,$CD,$F8,$23,$81,$F9,$2F,$83,$DC,$7E
	.dc.b	$4A,$7F,$21,$12,$76,$73,$38,$10,$C1,$27,$F3,$71,$F9,$9F,$83,$FC
	.dc.b	$D7,$C1,$CA,$FD,$0C,$65,$FD,$1C,$65,$FC,$A4,$65,$FC,$B4,$66,$BF
	.dc.b	$4F,$3F,$A5,$F8,$3D,$46,$2B,$F5,$B1,$97,$B9,$0F,$D0,$39,$7B,$99
	.dc.b	$8F,$D9,$C6,$5B,$1D,$0F,$DC,$FC,$1E,$74,$FD,$A5,$59,$FD,$A4,$DE
	.dc.b	$2F,$D5,$4D,$E2,$4F,$EF,$4C,$45,$AE,$A7,$73,$CE,$60,$FD,$C4,$70
	.dc.b	$71,$87,$F0,$A8,$FD,$A9,$23,$F6,$0D,$7F,$DD,$97,$FB,$D8,$C4,$0F
	.dc.b	$E1,$CE,$1A,$72,$21,$CA,$FD,$6F,$8B,$0D,$BF,$C3,$7E,$0F,$D8,$FA
	.dc.b	$0D,$1F,$C6,$5C,$E1,$E9,$8C,$E1,$FC,$46,$7F,$91,$7C,$1C,$9F,$E5
	.dc.b	$2F,$F2,$5F,$83,$C4,$7F,$1B,$8B,$7F,$CB,$26,$E1,$FC,$A6,$6D,$7F
	.dc.b	$29,$9E,$4F,$42,$1F,$C9,$5C,$BD,$4C,$D7,$57,$F9,$DC,$74,$C3,$F9
	.dc.b	$45,$70,$AD,$57,$33,$67,$F4,$2F,$83,$D6,$1F,$CF,$70,$B3,$FC,$F6
	.dc.b	$6E,$1F,$CC,$E6,$E4,$9F,$E8,$DF,$07,$FD,$1F,$E0,$E6,$1F,$D3,$A3
	.dc.b	$2F,$F5,$08,$CB,$FD,$06,$32,$FF,$43,$8C,$E5,$FD,$52,$7F,$A9,$7C
	.dc.b	$1F,$F5,$3F,$83,$97,$FD,$6A,$32,$FF,$5C,$8C,$BF,$D3,$63,$2F,$F4
	.dc.b	$F8,$CD,$7F,$60,$9F,$F0,$4F,$E6,$5F,$E1,$5A,$FE,$D6,$BE,$5F,$26
	.dc.b	$B6,$93,$FE,$05,$9D,$36,$72,$F2,$9E,$2E,$4E,$0F,$83,$CC,$F0,$F9
	.dc.b	$92,$7A,$46,$5E,$D1,$97,$CB,$19,$7C,$F1,$9C,$7F,$C4,$F8,$2B,$DB
	.dc.b	$E4,$4B,$F4,$3F,$03,$E1,$31,$1F,$02,$2D,$F2,$9B,$8F,$F8,$FF,$05
	.dc.b	$62,$FF,$C7,$70,$AF,$97,$F0,$92,$7E,$33,$F3,$BE,$13,$11,$F3,$A2
	.dc.b	$E3,$E8,$9B,$8F,$79,$B5,$DA,$79,$3D,$08,$7D,$87,$2F,$53,$35,$D5
	.dc.b	$FB,$11,$D3,$91,$8A,$F8,$DE,$EF,$91,$1C,$13,$0F,$34,$76,$7E,$36
	.dc.b	$1C,$9C,$C7,$17,$2F,$F9,$37,$F0,$F7,$95,$FE,$20,$6C,$7F,$86,$5A
	.dc.b	$FF,$28,$6B,$FC,$9C,$FF,$35,$B8,$7F,$CE,$8D,$73,$1D,$48,$F5,$6B
	.dc.b	$D5,$FF,$18,$E2,$F8,$5F,$D2,$92,$7F,$41,$F0,$B5,$FD,$3C,$5C,$BF
	.dc.b	$23,$37,$1F,$C4,$4D,$AF,$C3,$CF,$EC,$3D,$5F,$D8,$FC,$2C,$9F,$D6
	.dc.b	$46,$5F,$D7,$46,$5F,$DA,$C6,$5F,$DB,$C6,$70,$FD,$0C,$FE,$F3,$E1
	.dc.b	$65,$7E,$DD,$FD,$FF,$C2,$E2,$3F,$7F,$16,$BF,$53,$37,$2F,$D0,$4D
	.dc.b	$AF,$CF,$CF,$F0,$8F,$85,$FE,$13,$F0,$B2,$7F,$82,$46,$5F,$E0,$D1
	.dc.b	$97,$F8,$6C,$65,$FE,$1F,$19,$C3,$F7,$33,$FC,$53,$E1,$7F,$8A,$FC
	.dc.b	$2C,$BF,$E2,$11,$97,$F8,$94,$65,$FE,$33,$19,$7F,$8D,$C6,$6B,$F8
	.dc.b	$2C,$E5,$B3,$D0,$FE,$39,$ED,$67,$3F,$E0,$15,$67,$F7,$F3,$79,$7F
	.dc.b	$53,$37,$85,$36,$30,$DA,$EC,$7F,$2E,$70,$FA,$0F,$09,$FE,$BA,$F7
	.dc.b	$5F,$8D,$3C,$CF,$F6,$3C,$61,$B1,$99,$FF,$60,$8F,$A7,$79,$9A,$A9
	.dc.b	$3D,$6B,$83,$FE,$AD,$38,$4E,$9F,$F5,$85,$F7,$78,$57,$FB,$61,$D9
	.dc.b	$E0,$BE,$EE,$88,$75,$72,$DA,$ED,$AF,$FB,$96,$FB,$4E,$4E,$93,$AD
	.dc.b	$7F,$C4,$D3,$AA,$FE,$B1,$7D,$1B,$5F,$EA,$C8,$EF,$7F,$C1,$CF,$39
	.dc.b	$4A,$65,$DB,$FC,$35,$AF,$F8,$D5,$AF,$03,$51,$DC,$8F,$53,$47,$F4
	.dc.b	$6F,$87,$FE,$DE,$47,$FB,$57,$C3,$D7,$FB,$8C,$5B,$FE,$B7,$37,$1F
	.dc.b	$EA,$53,$6B,$FA,$8C,$FF,$23,$A5,$7F,$92,$F7,$7E,$66,$F7,$F5,$39
	.dc.b	$8C,$61,$FE,$FF,$6B,$FE,$EC,$5F,$2D,$C2,$17,$D2,$F0,$D8,$E5,$18
	.dc.b	$AE,$31,$82,$1D,$63,$55,$BF,$92,$E1,$F0,$32,$BC,$91,$97,$CD,$19
	.dc.b	$7C,$D1,$97,$D1,$19,$CB,$E1,$C7,$A3,$05,$79,$7E,$2B,$CD,$F1,$52
	.dc.b	$BE,$14,$65,$F8,$91,$97,$E2,$46,$5F,$8D,$19,$BE,$F9,$FD,$0A,$F8
	.dc.b	$5F,$15,$2F,$E7,$3F,$17,$E2,$B1,$1F,$3E,$2D,$F2,$C9,$58,$DB,$EC
	.dc.b	$55,$BF,$64,$FE,$CF,$4F,$E2,$31,$F9,$16,$BF,$9F,$6B,$FA,$36,$BF
	.dc.b	$DA,$DA,$FF,$A4,$DA,$9A,$4E,$6F,$CF,$46,$5B,$19,$8F,$D1,$7E,$71
	.dc.b	$FE,$33,$7F,$8A,$AF,$2B,$F8,$53,$63,$3E,$56,$0C,$CC,$8C,$90,$FE
	.dc.b	$31,$3C,$5B,$1C,$5F,$E4,$74,$E4,$F2,$92,$3E,$33,$5F,$8C,$D7,$E3
	.dc.b	$35,$F8,$CD,$7E,$33,$FB,$FD,$FF,$5F,$F1,$7F,$A1,$F8,$99,$5F,$AD
	.dc.b	$8F,$D2,$18,$7F,$5D,$18,$74,$7E,$CD,$D7,$07,$E0,$E7,$EF,$73,$7C
	.dc.b	$E2,$57,$DF,$6B,$F7,$DA,$FD,$F6,$BF,$7D,$AF,$DF,$7D,$5C,$2F,$90
	.dc.b	$FD,$D7,$C8,$4A,$FE,$0D,$1F,$BD,$30,$FF,$07,$8C,$3A,$3F,$86,$3A
	.dc.b	$C3,$F3,$93,$FD,$27,$EA,$7F,$87,$D7,$F3,$56,$BF,$E3,$D6,$BF,$EE
	.dc.b	$56,$A6,$93,$A7,$F2,$6D,$8F,$E6,$4D,$7F,$A0,$3D,$34,$FE,$4B,$EB
	.dc.b	$7F,$17,$F9,$19,$5F,$C8,$E3,$F8,$D1,$87,$F9,$24,$61,$D1,$FC,$A5
	.dc.b	$D6,$3F,$B9,$9F,$77,$5F,$90,$95,$E4,$6B,$E4,$6B,$F0,$DA,$FC,$36
	.dc.b	$BF,$0D,$FA,$FF,$23,$FC,$BF,$E3,$24,$FF,$3B,$8F,$E6,$46,$1F,$E7
	.dc.b	$91,$87,$47,$F4,$17,$57,$FC,$46,$7E,$B7,$6F,$CC,$25,$7C,$C6,$BF
	.dc.b	$31,$AF,$DB,$6B,$F6,$DA,$FD,$B7,$F1,$FE,$77,$F4,$7F,$8A,$93,$FD
	.dc.b	$4E,$3F,$A5,$18,$7F,$AA,$46,$1D,$1F,$D6,$5D,$5F,$F2,$99,$FC,$77
	.dc.b	$37,$E1,$09,$5F,$84,$6B,$F8,$46,$BF,$93,$6B,$F9,$36,$BF,$93,$7F
	.dc.b	$4F,$F1,$9F,$D7,$C9,$1F,$DA,$FE,$2E,$BF,$DA,$E2,$DF,$F3,$D9,$BD
	.dc.b	$FF,$DC,$E6,$D7,$F5,$79,$FE,$D5,$F1,$7F,$DB,$BF,$35,$FF,$0D,$61
	.dc.b	$FD,$39,$B8,$7F,$7B,$9D,$D8,$56,$DE,$46,$8F,$EF,$D1,$87,$84,$6D
	.dc.b	$FF,$02,$47,$ED,$FC,$2F,$D6,$92,$3F,$5A,$D7,$F5,$AD,$7F,$78,$D7
	.dc.b	$F7,$8D,$7F,$78,$FF,$29,$B7,$C6,$7B,$FE,$2E,$57,$93,$1F,$F0,$E5
	.dc.b	$6A,$30,$7F,$85,$1C,$3F,$02,$37,$3C,$A3,$06,$EB,$E6,$19,$C7,$E6
	.dc.b	$4E,$27,$FC,$D7,$CC,$F1,$7F,$86,$92,$3F,$86,$B5,$FE,$1A,$D7,$F8
	.dc.b	$E3,$5F,$E3,$8D,$7F,$8E,$3F,$07,$E3,$3E,$F7,$C6,$49,$F6,$F4,$B4
	.dc.b	$57,$E0,$CD,$4F,$B2,$75,$59,$8E,$35,$C1,$B1,$C3,$2F,$C5,$5F,$F3
	.dc.b	$3F,$1F,$F9,$49,$2B,$F9,$4B,$5F,$E5,$2D,$7F,$9C,$35,$FE,$70,$D7
	.dc.b	$F9,$C3,$F7,$7E,$33,$F3,$5F,$19,$27,$CB,$1F,$9D,$30,$FA,$F1,$4E
	.dc.b	$63,$67,$B1,$DE,$1F,$A3,$AF,$E9,$9E,$4F,$F4,$22,$57,$F4,$26,$BF
	.dc.b	$D0,$9A,$FF,$51,$6B,$FD,$45,$AF,$F5,$17,$F2,$DE,$6F,$EC,$FE,$32
	.dc.b	$4F,$CE,$8F,$DB,$18,$7D,$51,$87,$47,$D4,$75,$7F,$BB,$9F,$EC,$9E
	.dc.b	$2F,$F5,$A2,$57,$F5,$A6,$BF,$D6,$9A,$FF,$69,$6B,$FD,$A5,$AF,$F6
	.dc.b	$97,$F5,$DE,$67,$F0,$BF,$8B,$93,$F8,$68,$FE,$1C,$61,$FA,$51,$87
	.dc.b	$47,$E3,$1D,$5F,$F1,$39,$FE,$F9,$E4,$7F,$72,$25,$7F,$72,$6B,$FD
	.dc.b	$C9,$AF,$F8,$05,$AF,$F8,$05,$AF,$F8,$05,$FE,$0D,$F1,$9F,$C8,$FE
	.dc.b	$32,$4F,$E7,$A3,$F9,$31,$87,$F1,$51,$87,$47,$E9,$1D,$5F,$F2,$B9
	.dc.b	$FF,$0A,$78,$FF,$E0,$A2,$57,$F8,$29,$AF,$F8,$29,$AF,$F8,$65,$AF
	.dc.b	$F8,$65,$AF,$F8,$65,$FE,$3D,$F1,$9F,$CE,$FE,$32,$4F,$ED,$A3,$F9
	.dc.b	$F1,$87,$F4,$51,$87,$47,$EF,$1D,$5F,$F4,$39,$FF,$16,$79,$5F,$E2
	.dc.b	$22,$57,$F8,$89,$AF,$F8,$89,$AF,$F8,$C5,$AF,$F8,$C5,$AF,$F8,$C5
	.dc.b	$FE,$6D,$F1,$9F,$D4,$FE,$32,$4F,$F0,$E8,$FE,$AC,$61,$FD,$D4,$61
	.dc.b	$D1,$FC,$51,$D5,$FF,$5B,$9F,$F2,$27,$95,$FE,$3A,$25,$7F,$8E,$9A
	.dc.b	$FF,$8E,$9A,$FF,$92,$5A,$FF,$92,$5A,$FF,$92,$5F,$E9,$DF,$19,$FD
	.dc.b	$AF,$E2,$E4,$FF,$26,$8F,$ED,$C6,$1F,$E2,$51,$87,$47,$F2,$C7,$57
	.dc.b	$FD,$CE,$7F,$CB,$9E,$6F,$F9,$48,$95,$FE,$52,$6B,$FE,$52,$6B,$FE
	.dc.b	$61,$6B,$FE,$61,$6B,$FE,$61,$7F,$B3,$7C,$67,$F8,$0F,$E2,$E4,$FF
	.dc.b	$3E,$8F,$F0,$31,$87,$F9,$54,$61,$D1,$FD,$11,$D5,$FF,$82,$E7,$FC
	.dc.b	$E9,$E6,$FF,$9A,$89,$5F,$E6,$A6,$BF,$E6,$A6,$BF,$E7,$96,$BF,$E7
	.dc.b	$96,$BF,$E7,$97,$FB,$F7,$C6,$7F,$85,$63,$FA,$C7,$C1,$7F,$BD,$9B
	.dc.b	$E3,$FF,$87,$E6,$47,$F6,$29,$D6,$1F,$EE,$E6,$D7,$F8,$12,$B7,$5C
	.dc.b	$8E,$11,$8A,$ED,$E2,$BD,$0F,$F4,$39,$2B,$FD,$0E,$D7,$FD,$0E,$D7
	.dc.b	$FD,$2C,$D7,$FD,$2C,$D7,$FD,$2C,$FF,$B0,$F0,$7C,$67,$BE,$FD,$1E
	.dc.b	$CD,$C7,$FC,$6D,$E3,$AF,$FD,$CD,$8F,$C4,$C3,$4A,$75,$18,$78,$5D
	.dc.b	$AE,$38,$7C,$5A,$C1,$99,$97,$F2,$E3,$FD,$65,$E8,$7F,$AA,$09,$1F
	.dc.b	$EA,$86,$BF,$EA,$86,$BF,$EB,$76,$BF,$EB,$76,$BF,$EB,$77,$E1,$7B
	.dc.b	$1E,$5F,$8C,$93,$F6,$E3,$D0,$61,$FC,$1C,$61,$D1,$F8,$17,$58,$7E
	.dc.b	$06,$7F,$D9,$5E,$9F,$FB,$00,$95,$FE,$C0,$6B,$FE,$C0,$6B,$FE,$CF
	.dc.b	$6B,$FE,$CF,$6B,$FE,$CF,$7E,$D7,$C6,$7C,$DF,$8C,$93,$F9,$38,$FA
	.dc.b	$06,$1F,$CD,$C6,$1D,$1F,$99,$75,$7F,$99,$9F,$F7,$17,$A9,$FE,$D8
	.dc.b	$25,$7F,$B6,$1A,$FF,$B6,$1A,$FF,$B9,$DA,$FF,$B9,$DA,$FF,$B9,$DF
	.dc.b	$C9,$7C,$67,$E1,$7E,$2E,$4F,$EA,$E3,$F1,$06,$1F,$D9,$C6,$1D,$1F
	.dc.b	$B1,$75,$7F,$B1,$9F,$F7,$D7,$AD,$FE,$F0,$25,$7F,$BC,$1A,$FF,$BC
	.dc.b	$1A,$FF,$BF,$DA,$FF,$BF,$DA,$FF,$BF,$DF,$D5,$7C,$5F,$E7,$7E,$2A
	.dc.b	$4F,$F0,$38,$FD,$01,$87,$F8,$5C,$61,$D1,$FC,$25,$D5,$FF,$09,$9F
	.dc.b	$D4,$7C,$5F,$EA,$78,$E5,$FF,$0F,$8C,$BF,$C4,$63,$2F,$F1,$88,$CB
	.dc.b	$FC,$6A,$33,$5F,$C5,$67,$F7,$32,$E7,$7B,$3F,$C8,$62,$46,$1C,$1F
	.dc.b	$EB,$97,$0E,$1A,$F0,$8D,$BB,$6B,$DD,$B1,$C2,$3F,$98,$35,$E0,$72
	.dc.b	$70,$F1,$3C,$7C,$DF,$C5,$7E,$2F,$F8,$77,$C5,$49,$FE,$69,$1F,$C4
	.dc.b	$4C,$3F,$CF,$23,$0E,$8F,$E7,$4E,$B1,$FE,$75,$3F,$1B,$EA,$BD,$34
	.dc.b	$69,$1F,$6B,$5F,$82,$D7,$E4,$B5,$F9,$CD,$7F,$97,$E6,$F8,$CF,$E5
	.dc.b	$1F,$15,$27,$FA,$74,$7F,$2B,$30,$FF,$56,$8C,$3A,$3F,$AA,$3A,$C3
	.dc.b	$FA,$A4,$FA,$FD,$DF,$39,$2B,$CE,$D7,$CE,$D7,$CE,$D7,$FB,$36,$1F
	.dc.b	$ED,$C7,$27,$23,$EA,$E0,$F8,$CF,$E8,$5F,$11,$27,$FB,$44,$7F,$45
	.dc.b	$30,$FF,$70,$8C,$3A,$3F,$B7,$3A,$C7,$FB,$74,$FD,$5F,$7F,$E8,$92
	.dc.b	$BE,$8B,$5F,$A2,$D7,$E8,$B5,$F5,$B5,$1F,$7D,$FE,$D1,$F0,$FF,$D6
	.dc.b	$FE,$1E,$4F,$F8,$06,$3F,$AF,$18,$7F,$C1,$31,$87,$47,$F8,$1D,$D5
	.dc.b	$FF,$81,$E7,$F1,$DE,$FF,$E2,$89,$5F,$8A,$6B,$F8,$A6,$BF,$8A,$6B
	.dc.b	$F5,$1A,$FA,$B3,$FF,$77,$3F,$BD,$E1,$7E,$6C,$F8,$7F,$EF,$0B,$E1
	.dc.b	$E9,$88,$95,$FF,$00,$27,$4D,$9C,$BF,$EF,$F8,$E8,$E4,$86,$67,$F5
	.dc.b	$9F,$07,$FA,$82,$47,$EA,$1A,$FE,$A1,$AF,$EA,$1A,$FE,$49,$AF,$BD
	.dc.b	$FE,$23,$7E,$EF,$87,$F6,$FC,$3C,$AF,$54,$7C,$03,$0F,$A6,$30,$E8
	.dc.b	$F8,$0E,$B5,$79,$3F,$E2,$6C,$97,$FC,$1F,$DF,$FE,$06,$4C,$3F,$81
	.dc.b	$B5,$FE,$06,$D7,$F8,$1B,$5F,$D6,$B5,$FB,$4F,$B2,$5F,$0F,$F6,$3E
	.dc.b	$1E,$57,$D5,$8F,$B6,$61,$FA,$91,$87,$47,$DB,$75,$8F,$B2,$3F,$C7
	.dc.b	$95,$D9,$FD,$1C,$7F,$98,$F7,$7B,$9F,$C8,$09,$3F,$C8,$1A,$FF,$20
	.dc.b	$6B,$FC,$81,$AF,$F0,$A6,$BF,$97,$7D,$DF,$0F,$F9,$42,$47,$E4,$7E
	.dc.b	$12,$BF,$8F,$8B,$7E,$9D,$FF,$3B,$58,$DB,$EB,$55,$BF,$6C,$DA,$FD
	.dc.b	$1C,$DC,$93,$63,$0D,$AF,$D1,$47,$EA,$23,$F5,$3F,$07,$FA,$AF,$81
	.dc.b	$95,$FA,$78,$CB,$FA,$98,$CB,$FA,$78,$CB,$FA,$98,$CE,$5E,$BC,$15
	.dc.b	$FB,$6F,$80,$FD,$BF,$BD,$30,$FD,$A4,$65,$FD,$B4,$65,$FD,$A4,$65
	.dc.b	$FD,$B4,$66,$FB,$4F,$F0,$2F,$73,$F8,$1F,$B5,$2F,$F8,$04,$65,$FE
	.dc.b	$05,$19,$7F,$80,$46,$5F,$E0,$51,$9A,$F5,$4E,$5F,$64,$67,$6F,$D5
	.dc.b	$4E,$5F,$D6,$46,$6B,$F2,$73,$97,$F2,$B1,$9D,$BF,$8D,$CF,$F7,$7F
	.dc.b	$73,$FC,$3B,$F0,$4F,$86,$A6,$70,$FC,$DE,$3F,$D6,$AB,$31,$83,$FA
	.dc.b	$0D,$71,$70,$7F,$80,$63,$92,$CB,$FD,$8D,$3B,$8C,$57,$FB,$2E,$31
	.dc.b	$3F,$A9,$AE,$35,$FE,$02,$7F,$C1,$24,$7F,$C1,$2D,$7F,$C1,$2D,$78
	.dc.b	$1A,$21,$A2,$C7,$77,$D3,$C6,$9F,$8B,$D3,$EA,$AD,$57,$A0,$FE,$91
	.dc.b	$3A,$78,$BB,$C3,$E4,$47,$C4,$C1,$7F,$D2,$BD,$B9,$3F,$D1,$DF,$E9
	.dc.b	$BE,$D6,$23,$FA,$34,$5C,$7F,$3D,$37,$0F,$E5,$93,$6B,$F8,$9C,$DD
	.dc.b	$7F,$AC,$CD,$E2,$FF,$74,$57,$72,$1D,$8F,$EB,$15,$B9,$E0,$FF,$60
	.dc.b	$F6,$24,$FF,$5D,$7F,$B1,$FA,$D8,$8F,$EB,$91,$72,$FD,$A4,$DC,$BF
	.dc.b	$A1,$4D,$AF,$E5,$33,$72,$FE,$DF,$37,$24,$D8,$C3,$6B,$C4,$7F,$6F
	.dc.b	$7F,$BA,$FA,$9F,$DD,$BD,$29,$3F,$DC,$E3,$2F,$F7,$58,$CB,$FD,$CE
	.dc.b	$32,$FF,$75,$8C,$E5,$FC,$36,$72,$FF,$0F,$8C,$E5,$FD,$76,$72,$FF
	.dc.b	$5F,$8C,$D7,$F4,$69,$CB,$FD,$22,$33,$97,$F8,$3A,$72,$FF,$84,$23
	.dc.b	$39,$27,$2A,$33,$18,$6D,$7F,$84,$7C,$FF,$F1,$AB,$F9,$7B,$7E,$6F
	.dc.b	$FE,$99,$CB,$FA,$13,$72,$FE,$ED,$7F,$E1,$E3,$8C,$EA,$EC,$FF,$90
	.dc.b	$2F,$9C,$6C,$F0,$1F,$C1,$5C,$B6,$FB,$3F,$DA,$2F,$83,$E5,$8C,$CF
	.dc.b	$08,$FF,$13,$4E,$AF,$C5,$1C,$CC,$5F,$A1,$C1,$0E,$27,$87,$CB,$F4
	.dc.b	$AB,$F8,$7E,$46,$1B,$1B,$21,$E5,$8D,$62,$6D,$49,$F7,$46,$8C,$3E
	.dc.b	$D8,$C3,$A3,$AE,$1F,$4A,$33,$1D,$CC,$E1,$F3,$ED,$3A,$AF,$B3,$7D
	.dc.b	$9F,$B7,$E3,$79,$89,$87,$D1,$E2,$AF,$CF,$8B,$3F,$82,$9B,$8F,$79
	.dc.b	$B5,$F1,$F1,$FE,$32,$FD,$B8,$CE,$5F,$8C,$7E,$2D,$E2,$BF,$CD,$78
	.dc.b	$79,$2F,$C4,$71,$21,$E3,$32,$52,$3C,$5F,$C3,$4F,$5A,$EC,$D8,$E7
	.dc.b	$3E,$B7,$F2,$1F,$0F,$F6,$39,$E4,$60,$87,$E4,$5C,$BF,$93,$8F,$BE
	.dc.b	$62,$7F,$4F,$38,$9E,$31,$87,$F3,$D1,$97,$91,$82,$18,$24,$71,$7C
	.dc.b	$47,$4C,$BF,$01,$E1,$FE,$40,$91,$FA,$BF,$87,$AF,$EA,$62,$D7,$EF
	.dc.b	$26,$D7,$2A,$FD,$95,$7E,$6A,$7F,$37,$CF,$27,$F6,$CF,$E7,$BE,$1F
	.dc.b	$11,$FB,$58,$B8,$7F,$08,$9B,$87,$E5,$66,$D7,$E1,$E6,$ED,$FC,$42
	.dc.b	$6E,$C9,$B1,$86,$D7,$88,$FE,$10,$FE,$C3,$C3,$91,$FC,$2D,$FD,$9F
	.dc.b	$C3,$E2,$3F,$85,$45,$CB,$F8,$EC,$DC,$BF,$55,$36,$BF,$39,$3F,$BC
	.dc.b	$F8,$69,$1F,$C5,$DF,$DF,$FC,$3E,$23,$F8,$B4,$5C,$3F,$94,$CD,$C3
	.dc.b	$F6,$D3,$6B,$F5,$13,$FC,$23,$E1,$BF,$84,$FC,$3C,$9F,$E4,$91,$97
	.dc.b	$F9,$34,$65,$FE,$49,$19,$7F,$93,$46,$70,$FE,$73,$3F,$C5,$3E,$1E
	.dc.b	$57,$F2,$F7,$F8,$BF,$81,$88,$FE,$5D,$16,$BF,$A0,$CD,$CB,$F8,$74
	.dc.b	$DA,$FE,$01,$3F,$C8,$3E,$1A,$47,$F3,$B7,$F9,$1F,$C3,$E2,$3F,$9D
	.dc.b	$45,$C3,$FA,$6C,$DC,$3F,$8D,$4D,$AF,$E1,$93,$FC,$B3,$E1,$E4,$7F
	.dc.b	$47,$7F,$9A,$7E,$71,$FE,$AF,$8F,$E8,$D1,$A6,$9F,$D3,$E3,$58,$7F
	.dc.b	$5C,$9D,$36,$75,$87,$F4,$28,$7C,$A3,$E0,$70,$DA,$FF,$2F,$4E,$A7
	.dc.b	$3F,$01,$A6,$4E,$AB,$83,$FE,$45,$AE,$B5,$A8,$C4,$EA,$7F,$D9,$47
	.dc.b	$58,$E5,$3F,$03,$E2,$24,$7B,$5F,$87,$F1,$38,$8F,$EF,$99,$A7,$17
	.dc.b	$FE,$BD,$85,$63,$2F,$EF,$11,$BB,$E8,$7C,$99,$F8,$3F,$03,$F0,$BD
	.dc.b	$F9,$3E,$D8,$CB,$EF,$8C,$BE,$C8,$CB,$EE,$8C,$E1,$EE,$CD,$5F,$73
	.dc.b	$E2,$25,$FD,$97,$EF,$FC,$4E,$23,$DB,$16,$FD,$93,$73,$F5,$CD,$AF
	.dc.b	$EF,$99,$AB,$ED,$FC,$07,$DC,$F7,$A5,$7D,$78,$CB,$F6,$63,$2F,$D6
	.dc.b	$8C,$BF,$62,$33,$8F,$AE,$7F,$2B,$F1,$1F,$96,$F8,$99,$7F,$93,$8C
	.dc.b	$BF,$95,$8C,$BF,$62,$32,$FD,$A8,$CD,$7D,$99,$FC,$B7,$E1,$4F,$81
	.dc.b	$BE,$FC,$C3,$F2,$B3,$96,$CE,$5F,$C9,$42,$79,$CC,$9E,$31,$BB,$FC
	.dc.b	$9D,$73,$9F,$C6,$FC,$4F,$EC,$BE,$2A,$47,$EB,$E3,$F6,$A6,$1F,$C9
	.dc.b	$C6,$1D,$1F,$B5,$75,$5F,$95,$9F,$F1,$9C,$BE,$41,$F0,$78,$6C,$7F
	.dc.b	$B3,$1A,$FF,$AC,$47,$D7,$7E,$AF,$8F,$FC,$BF,$3A,$C3,$F8,$43,$63
	.dc.b	$37,$E0,$AE,$16,$3C,$51,$E1,$23,$FE,$DC,$8D,$BE,$0E,$AC,$BE,$3B
	.dc.b	$EC,$F3,$9D,$9C,$23,$31,$D8,$E9,$3E,$33,$9E,$1E,$8B,$C4,$E8,$F8
	.dc.b	$4E,$B0,$FF,$3E,$E1,$DA,$F1,$5E,$BC,$3B,$CE,$63,$89,$0F,$94,$E5
	.dc.b	$F5,$5F,$27,$DC,$78,$4C,$9C,$70,$FE,$69,$3C,$5F,$44,$67,$0F,$2E
	.dc.b	$1E,$A8,$E2,$49,$E3,$19,$78,$C6,$67,$8B,$E1,$8C,$92,$33,$1E,$A9
	.dc.b	$FF,$4F,$B2,$BF,$99,$B6,$32,$F0,$30,$60,$F7,$9A,$8E,$2E,$C8,$7D
	.dc.b	$D8,$E0,$F0,$39,$9C,$5F,$64,$F9,$AB,$F2,$7A,$E2,$B9,$1E,$D9,$FC
	.dc.b	$CC,$66,$3C,$73,$F1,$9F,$D1,$46,$6F,$E8,$57,$09,$E4,$74,$3E,$21
	.dc.b	$C0,$C9,$D1,$F2,$1C,$5D,$92,$BC,$0F,$E7,$6B,$C1,$7E,$5C,$7F,$83
	.dc.b	$56,$89,$7D,$2F,$F2,$57,$C1,$F0,$D7,$09,$D4,$78,$CF,$45,$7C,$67
	.dc.b	$CF,$87,$B9,$E6,$7F,$74,$FD,$45,$E7,$7F,$7B,$7F,$D1,$F7,$FB,$BB
	.dc.b	$46,$29,$18,$AE,$07,$1F,$42,$4F,$58,$CB,$DE,$32,$FF,$7D,$AF,$0B
	.dc.b	$FE,$11,$F8,$DF,$F0,$BE,$92,$30,$43,$FB,$BB,$97,$FB,$CC,$7F,$87
	.dc.b	$8C,$61,$FD,$E6,$71,$3F,$E1,$DE,$EF,$F1,$1F,$C5,$48,$C1,$0F,$F0
	.dc.b	$1B,$97,$FC,$0B,$1F,$E2,$A3,$15,$FE,$05,$9C,$4F,$F8,$A7,$E3,$3F
	.dc.b	$C5,$E4,$7F,$C5,$9D,$95,$FF,$07,$C5,$AF,$F0,$74,$DD,$7F,$C1,$B3
	.dc.b	$6B,$FB,$9C,$FF,$8E,$BB,$A4,$7F,$8D,$DF,$F1,$EF,$C6,$62,$3F,$C3
	.dc.b	$71,$70,$FF,$0D,$4D,$C3,$FC,$33,$36,$BF,$BF,$CF,$F9,$1B,$E3,$24
	.dc.b	$7F,$90,$DF,$F2,$4F,$5E,$23,$FC,$4F,$17,$0F,$F1,$34,$DC,$3F,$C4
	.dc.b	$B3,$6B,$FC,$17,$3F,$E5,$2E,$C9,$1F,$E4,$F7,$FC,$AB,$F1,$98,$8F
	.dc.b	$F1,$9C,$5C,$3F,$C6,$53,$70,$FF,$18,$CD,$AF,$F0,$BC,$FF,$51,$7F
	.dc.b	$2B,$3F,$D4,$FD,$27,$9A,$F8,$CD,$62,$AD,$4F,$EA,$ED,$7C,$3A,$77
	.dc.b	$5A,$C2,$D6,$AF,$A3,$FA,$79,$E7,$5A,$BE,$53,$93,$85,$E6,$F8,$C7
	.dc.b	$A0,$E9,$59,$35,$3B,$EE,$E1,$D7,$2B,$A6,$1E,$E9,$D4,$63,$0F,$ED
	.dc.b	$73,$F0,$AB,$33,$DE,$38,$3B,$8E,$8E,$67,$FC,$40,$F3,$F8,$CF,$41
	.dc.b	$1E,$DF,$19,$5E,$11,$74,$F9,$D1,$F1,$E7,$1A,$7D,$0A,$B3,$93,$D1
	.dc.b	$F1,$9F,$17,$E3,$24,$60,$87,$95,$CB,$E7,$8F,$98,$63,$0F,$0E,$7F
	.dc.b	$82,$AC,$D7,$C5,$F8,$BF,$A5,$F1,$92,$30,$43,$E1,$39,$7E,$24,$7D
	.dc.b	$93,$17,$DB,$45,$66,$BE,$97,$C5,$FD,$FF,$8C,$91,$82,$1F,$3D,$CB
	.dc.b	$F4,$63,$F0,$C6,$2F,$B4,$E2,$7E,$F7,$C5,$FE,$2F,$E3,$24,$60,$87
	.dc.b	$DB,$72,$FD,$D8,$FC,$91,$8A,$F5,$4E,$27,$F1,$5F,$17,$F9,$72,$3F
	.dc.b	$8D,$F8,$BA,$FE,$1E,$2D,$7C,$39,$BC,$3F,$AE,$9B,$5F,$A9,$9B,$BF
	.dc.b	$ED,$A6,$EA,$9B,$18,$6D,$79,$0F,$11,$F9,$FF,$8B,$FD,$61,$1F,$D1
	.dc.b	$7C,$5D,$7F,$39,$17,$2F,$B1,$37,$2F,$E0,$73,$6B,$F7,$D3,$FB,$7F
	.dc.b	$8A,$FD,$C7,$C5,$49,$FD,$74,$65,$FD,$84,$65,$FD,$54,$65,$FD,$64
	.dc.b	$67,$0F,$C4,$4E,$5F,$C5,$46,$70,$FE,$2B,$39,$7F,$8B,$C6,$6B,$F8
	.dc.b	$94,$E5,$FE,$29,$19,$D3,$F9,$0C,$FF,$0F,$F8,$AF,$E2,$9F,$21,$FE
	.dc.b	$3E,$C7,$F4,$AD,$C7,$F9,$1C,$EF,$1F,$E1,$B3,$97,$F9,$2B,$70,$FE
	.dc.b	$5B,$3B,$9F,$E1,$93,$93,$5B,$27,$58,$78,$2B,$C6,$63,$0F,$0B,$E0
	.dc.b	$32,$7F,$1B,$9D,$3E,$23,$27,$F1,$77,$38,$30,$F1,$4E,$30,$F1,$3D
	.dc.b	$48,$71,$3B,$18,$C3,$CC,$E0,$F6,$FC,$5F,$CA,$23,$F1,$BE,$2E,$BA
	.dc.b	$3E,$4C,$7D,$08,$CC,$FD,$08,$B9,$7F,$57,$C1,$58,$BF,$9F,$57,$1F
	.dc.b	$64,$DA,$4D,$8C,$36,$BE,$CF,$99,$F8,$02,$3F,$73,$E2,$EB,$F6,$22
	.dc.b	$E5,$F8,$49,$B9,$79,$66,$D7,$E0,$A6,$E5,$F3,$E6,$E4,$9B,$18,$6D
	.dc.b	$79,$0F,$11,$F8,$EF,$3B,$F2,$A4,$7F,$23,$F1,$75,$FC,$6C,$5C,$BF
	.dc.b	$33,$37,$2F,$95,$36,$BF,$2F,$37,$2F,$C2,$4D,$C9,$36,$30,$DA,$F2
	.dc.b	$1E,$22,$9E,$22,$E1,$E2,$6E,$1E,$26,$9E,$22,$E1,$E2,$6E,$1E,$27
	.dc.b	$F5,$FF,$17,$FB,$62,$3F,$B2,$F8,$BA,$FE,$BA,$2E,$1F,$BA,$9B,$87
	.dc.b	$E2,$E6,$D7,$EE,$26,$E1,$FA,$59,$B8,$26,$C6,$1B,$5E,$43,$C4,$7F
	.dc.b	$04,$F8,$AF,$E1,$64,$7F,$83,$FC,$55,$7F,$81,$C5,$CB,$F8,$7C,$DC
	.dc.b	$BF,$41,$36,$BF,$86,$CD,$CB,$F7,$53,$72,$4D,$8C,$36,$BC,$87,$88
	.dc.b	$FE,$33,$F1,$5F,$C8,$3E,$2A,$46,$08,$7F,$18,$72,$FF,$1A,$8F,$E4
	.dc.b	$E6,$32,$FE,$4F,$38,$9E,$31,$87,$F9,$3C,$65,$E4,$60,$86,$09,$1C
	.dc.b	$5F,$11,$D3,$2F,$E0,$95,$E1,$7F,$83,$57,$5A,$FE,$73,$39,$AF,$E6
	.dc.b	$B3,$9A,$E3,$1E,$0A,$EC,$70,$7F,$9E,$F1,$FF,$44,$23,$FD,$03,$E2
	.dc.b	$AB,$FC,$F2,$2C,$FF,$48,$9B,$5F,$C6,$66,$D7,$F4,$69,$BA,$7F,$33
	.dc.b	$9B,$A2,$6C,$61,$B5,$E4,$3C,$47,$F5,$2F,$89,$FE,$B3,$F1,$52,$30
	.dc.b	$43,$FA,$8B,$97,$FA,$9C,$7F,$60,$31,$97,$F6,$09,$C4,$F1,$8C,$3F
	.dc.b	$D8,$23,$2F,$23,$04,$30,$48,$E2,$F8,$8E,$99,$7F,$3D,$F8,$AF,$EE
	.dc.b	$3F,$15,$27,$FB,$6C,$7F,$74,$30,$FF,$6A,$8C,$3A,$3F,$B9,$BA,$AF
	.dc.b	$EF,$33,$A9,$E0,$68,$E0,$7F,$74,$78,$57,$F4,$FA,$E1,$5C,$AB,$81
	.dc.b	$C8,$E0,$72,$7F,$C0,$F1,$CA,$3F,$C0,$75,$CA,$BA,$47,$2A,$E8,$F2
	.dc.b	$3A,$3C,$9E,$91,$FD,$87,$E2,$BF,$C2,$A4,$8F,$F0,$8F,$C5,$57,$FC
	.dc.b	$1B,$16,$BF,$C3,$33,$6B,$95,$7F,$86,$2B,$FC,$3F,$39,$8F,$F0,$CC
	.dc.b	$5D,$BF,$C0,$D3,$76,$4D,$8C,$36,$BF,$C4,$7F,$15,$FE,$2A,$23,$FE
	.dc.b	$B2,$FC,$EB,$FD,$3E,$FF,$88,$A3,$2D,$35,$8E,$32,$FF,$18,$CD,$8F
	.dc.b	$F6,$C5,$72,$9C,$D7,$80,$87,$FA,$E6,$35,$3F,$E0,$9A,$E5,$3F,$E3
	.dc.b	$67,$FD,$A5,$3A,$8C,$1C,$AB,$AB,$CE,$BF,$DB,$AB,$E2,$1E,$7E,$3F
	.dc.b	$22,$E5,$26,$CE,$A3,$FD,$BC,$71,$9E,$15,$23,$33,$B7,$FD,$C0,$6C
	.dc.b	$E3,$5D,$8E,$B5,$E4,$9E,$13,$FE,$FF,$38,$14,$A6,$5D,$A2,$38,$6C
	.dc.b	$71,$69,$63,$4F,$43,$91,$4D,$9E,$17,$67,$B6,$FF,$06,$DB,$E2,$D9
	.dc.b	$F2,$7E,$44,$D3,$DD,$84,$9F,$8E,$FD,$49,$D4,$60,$87,$CA,$AE,$8A
	.dc.b	$36,$FE,$1F,$E3,$7F,$07,$F1,$B2,$3F,$31,$1F,$86,$30,$FE,$36,$30
	.dc.b	$E8,$FC,$F3,$AC,$BF,$DD,$BA,$2B,$75,$C4,$D9,$C4,$F1,$18,$34,$60
	.dc.b	$93,$D2,$BB,$67,$FA,$4A,$FC,$DE,$3F,$9C,$F8,$B9,$87,$EC,$63,$2F
	.dc.b	$EC,$E3,$2F,$E9,$E3,$2F,$EA,$63,$38,$7A,$A7,$2F,$B2,$33,$7F,$B4
	.dc.b	$9C,$BF,$B6,$8C,$D3,$85,$97,$EA,$63,$C8,$78,$C8,$78,$CF,$E1,$1F
	.dc.b	$15,$FB,$6F,$8B,$91,$82,$1F,$BF,$72,$FF,$01,$8F,$DE,$98,$CB,$E9
	.dc.b	$CE,$27,$8C,$61,$FE,$2D,$19,$79,$18,$21,$82,$47,$17,$C4,$74,$C7
	.dc.b	$F8,$BD,$78,$5F,$E3,$35,$D6,$BF,$0B,$39,$AF,$8D,$6A,$D5,$F2,$9F
	.dc.b	$0D,$F7,$38,$BF,$CB,$7E,$27,$F8,$DF,$C5,$48,$C1,$0F,$E4,$8E,$5F
	.dc.b	$E4,$D1,$FC,$88,$C4,$FE,$62,$71,$3C,$63,$0F,$F3,$B8,$CB,$C8,$C1
	.dc.b	$0C,$12,$38,$BE,$23,$A5,$7F,$3C,$F8,$7F,$E6,$7F,$0F,$27,$FA,$44
	.dc.b	$7F,$36,$30,$FF,$3D,$8C,$3A,$3F,$A7,$3A,$AF,$D8,$CE,$A7,$81,$A3
	.dc.b	$81,$FD,$1D,$E1,$5F,$D3,$AB,$85,$72,$AE,$07,$23,$81,$FD,$77,$E1
	.dc.b	$FF,$A5,$91,$FE,$C1,$F0,$F5,$FE,$A9,$16,$7F,$82,$CD,$BF,$EB,$B3
	.dc.b	$6B,$F0,$73,$76,$FD,$4C,$DD,$53,$63,$0D,$AF,$21,$E2,$29,$E2,$2E
	.dc.b	$1E,$26,$E1,$E2,$69,$E2,$2E,$1E,$26,$E1,$E2,$7F,$BF,$7C,$3F,$F6
	.dc.b	$C2,$3F,$E0,$1F,$87,$AF,$F7,$58,$B8,$7F,$26,$9B,$87,$F7,$E9,$B5
	.dc.b	$FA,$49,$B8,$7F,$0A,$9B,$82,$6C,$61,$B5,$E4,$3C,$47,$F8,$4F,$E1
	.dc.b	$FF,$C0,$24,$7F,$C2,$DF,$0F,$5F,$F0,$54,$5C,$BF,$9D,$CD,$CB,$FC
	.dc.b	$27,$36,$BF,$73,$37,$2F,$E3,$F3,$72,$4D,$8C,$36,$BC,$87,$88,$FF
	.dc.b	$14,$7C,$3F,$F8,$58,$8F,$F8,$AF,$E1,$EB,$FE,$1F,$8B,$97,$F5,$09
	.dc.b	$B9,$7F,$8A,$26,$D7,$F0,$E9,$B9,$7F,$34,$9B,$92,$6C,$61,$B5,$E4
	.dc.b	$3C,$45,$3C,$45,$C3,$C4,$DC,$3C,$4D,$3C,$45,$C3,$C4,$DC,$3C,$4F
	.dc.b	$F9,$27,$E1,$FF,$C6,$E4,$7F,$C9,$9F,$0F,$5F,$F2,$04,$5C,$3F,$B8
	.dc.b	$CD,$C3,$FC,$93,$36,$BF,$97,$CD,$C3,$FA,$AC,$DC,$13,$63,$0D,$AF
	.dc.b	$21,$E2,$3F,$C5,$58,$30,$FF,$16,$60,$FC,$ED,$F0,$FF,$E2,$6A,$D5
	.dc.b	$FF,$65,$97,$C4,$67,$0E,$31,$89,$EB,$18,$BD,$47,$16,$E1,$DE,$ED
	.dc.b	$F1,$A6,$1E,$0A,$F2,$5E,$8D,$DC,$AE,$E7,$52,$CF,$06,$DF,$5B,$E8
	.dc.b	$DC,$3C,$D3,$AC,$3A,$3E,$28,$F6,$E1,$E1,$8D,$1D,$1D,$DF,$4A,$F5
	.dc.b	$FC,$44,$AF,$0B,$EE,$F0,$31,$1D,$4F,$99,$18,$AF,$2C,$DC,$3E,$6C
	.dc.b	$DA,$FA,$53,$72,$FB,$13,$72,$FF,$2A,$46,$2B,$C2,$43,$C0,$FD,$FA
	.dc.b	$E1,$5C,$5B,$3B,$C3,$C6,$DC,$3C,$6D,$3C,$65,$BF,$03,$DB,$06,$1F
	.dc.b	$06,$3E,$9F,$C3,$FE,$03,$E2,$24,$60,$87,$17,$2F,$38,$FC,$39,$8C
	.dc.b	$7E,$DC,$E2,$78,$C6,$1F,$C0,$C6,$5E,$46,$08,$60,$91,$C5,$F1,$1D
	.dc.b	$31,$FC,$FF,$C4,$7E,$63,$E2,$24,$FE,$3A,$3F,$36,$61,$F8,$51,$87
	.dc.b	$47,$E5,$1D,$57,$E5,$27,$53,$C0,$D1,$C0,$F9,$8F,$0A,$FD,$8D,$70
	.dc.b	$AE,$55,$C0,$E4,$70,$3F,$47,$F1,$1F,$B0,$23,$FA,$6F,$88,$AF,$D9
	.dc.b	$8B,$3F,$A8,$9B,$7F,$C0,$26,$D7,$F0,$59,$BA,$FF,$0B,$9B,$AA,$6C
	.dc.b	$61,$B5,$E4,$3C,$45,$3C,$45,$C3,$C4,$DC,$3C,$4D,$3C,$45,$C3,$C4
	.dc.b	$DC,$3C,$4F,$F0,$4F,$88,$FE,$22,$47,$F8,$3F,$C4,$57,$F3,$51,$70
	.dc.b	$FE,$13,$37,$0F,$E4,$33,$6B,$F9,$34,$DC,$3F,$96,$4D,$C1,$36,$30
	.dc.b	$DA,$F2,$1E,$23,$F8,$CF,$C4,$7F,$26,$23,$FC,$73,$E2,$2B,$FA,$F8
	.dc.b	$B9,$7F,$1E,$9B,$97,$F3,$69,$B5,$FC,$EE,$6E,$5F,$D0,$66,$E4,$9B
	.dc.b	$18,$6D,$79,$0F,$11,$FC,$BB,$E2,$3F,$9D,$FC,$44,$8C,$10,$FE,$08
	.dc.b	$E5,$FE,$0D,$1F,$D0,$8C,$65,$FC,$DA,$71,$3C,$63,$0F,$F3,$C8,$CB
	.dc.b	$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$5F,$D6,$AB,$C2,$FF,$5C,$AE,$B5
	.dc.b	$FD,$22,$73,$5F,$D8,$E7,$35,$C6,$3C,$15,$D8,$E0,$FF,$52,$F8,$8F
	.dc.b	$EB,$E4,$7F,$AA,$FC,$45,$7F,$94,$45,$9F,$EA,$F3,$6B,$FB,$7C,$DA
	.dc.b	$FE,$E9,$37,$4F,$EF,$53,$74,$4D,$8C,$36,$BC,$87,$88,$FE,$CF,$F0
	.dc.b	$FF,$DD,$3E,$22,$46,$08,$7F,$3D,$72,$FF,$3F,$8F,$EF,$66,$32,$FE
	.dc.b	$DF,$38,$9E,$31,$87,$FB,$AC,$65,$E4,$60,$86,$09,$1C,$5F,$11,$D3
	.dc.b	$23,$3C,$DF,$82,$FF,$81,$87,$92,$BF,$F0,$AC,$61,$B3,$C2,$3F,$BB
	.dc.b	$07,$17,$FE,$24,$8E,$07,$5A,$E8,$F7,$9D,$DF,$92,$39,$63,$E0,$AD
	.dc.b	$C7,$66,$DF,$5B,$E3,$3E,$BA,$E4,$C8,$F3,$36,$32,$F7,$8C,$B6,$32
	.dc.b	$F3,$2C,$69,$C1,$C9,$C3,$E8,$37,$1C,$9C,$1B,$7E,$49,$1F,$88,$D7
	.dc.b	$E5,$C7,$B1,$E3,$3F,$2E,$B1,$32,$B8,$4E,$5F,$F2,$3F,$C4,$7F,$92
	.dc.b	$B8,$3E,$1C,$FF,$91,$DB,$8F,$F9,$52,$77,$8F,$F9,$12,$72,$B6,$FE
	.dc.b	$3C,$E8,$FF,$0C,$4F,$56,$1F,$E4,$A7,$57,$DE,$32,$70,$34,$77,$32
	.dc.b	$FF,$94,$23,$8D,$FF,$9C,$2F,$C7,$7E,$2B,$E2,$7F,$8D,$A1,$F0,$FE
	.dc.b	$FF,$88,$93,$EB,$8F,$84,$61,$E1,$18,$74,$7C,$17,$58,$FC,$79,$D4
	.dc.b	$F0,$34,$70,$3C,$4F,$0C,$3D,$15,$C2,$B9,$57,$03,$91,$C0,$FA,$1F
	.dc.b	$11,$F5,$FE,$22,$46,$08,$7B,$DC,$BF,$06,$3E,$F1,$89,$FB,$D3,$89
	.dc.b	$E3,$18,$7E,$F4,$65,$E4,$60,$86,$09,$1C,$5F,$11,$D2,$FE,$B5,$78
	.dc.b	$5F,$B1,$5D,$6B,$F2,$13,$9A,$FC,$66,$2A,$D5,$F2,$9F,$0D,$F7,$38
	.dc.b	$BF,$95,$F8,$8F,$CE,$11,$FC,$C7,$C4,$57,$F0,$91,$67,$F4,$13,$6B
	.dc.b	$F1,$93,$6B,$C7,$37,$4F,$EF,$93,$71,$4D,$8C,$36,$BC,$87,$88,$FD
	.dc.b	$57,$C4,$7E,$C7,$E2,$24,$60,$87,$E6,$5C,$BF,$9B,$8F,$DC,$18,$CB
	.dc.b	$F7,$13,$89,$E3,$18,$7F,$71,$19,$79,$18,$21,$82,$47,$17,$C4,$74
	.dc.b	$CB,$F6,$1F,$11,$FC,$24,$91,$FC,$17,$E2,$2B,$FB,$18,$B5,$FC,$36
	.dc.b	$6D,$72,$AF,$E1,$95,$FC,$4E,$73,$1F,$82,$8B,$A7,$F8,$AE,$6E,$89
	.dc.b	$B1,$86,$D7,$42,$1E,$33,$CC,$63,$0F,$E4,$11,$C2,$34,$48,$CB,$C0
	.dc.b	$FE,$19,$F1,$1F,$CA,$48,$FF,$25,$F8,$8A,$E8,$FE,$51,$1F,$CB,$E3
	.dc.b	$33,$FC,$BE,$2E,$5F,$C6,$E6,$D7,$E8,$A6,$E5,$FE,$4B,$9B,$92,$6C
	.dc.b	$61,$B5,$FC,$DF,$E2,$3F,$9F,$7C,$44,$8C,$10,$FE,$4C,$E5,$FE,$51
	.dc.b	$1F,$D1,$4C,$65,$FD,$16,$71,$3C,$63,$0F,$F4,$58,$CB,$C8,$C1,$0C
	.dc.b	$12,$38,$BE,$23,$A6,$5F,$CF,$7E,$23,$FA,$B7,$C4,$49,$FE,$A5,$1F
	.dc.b	$D6,$4C,$3F,$CF,$E3,$0E,$8F,$EB,$0E,$AB,$FA,$F4,$EA,$78,$1A,$38
	.dc.b	$1F,$D2,$5E,$15,$FD,$3E,$B8,$57,$2A,$E0,$72,$38,$1C,$9F,$ED,$91
	.dc.b	$CA,$3F,$8E,$D7,$2A,$E9,$1C,$AB,$A3,$C8,$E8,$F2,$7A,$47,$F6,$1F
	.dc.b	$87,$FE,$F5,$F1,$12,$7F,$BB,$47,$F7,$D3,$0F,$F6,$38,$C3,$A3,$FB
	.dc.b	$E3,$AA,$FF,$01,$4E,$A7,$81,$A3,$81,$FD,$B5,$E1,$5F,$DC,$EB,$85
	.dc.b	$72,$AE,$07,$23,$81,$FE,$0B,$F8,$7F,$F0,$97,$C2,$C8,$C1,$0F,$EF
	.dc.b	$4E,$5F,$EF,$91,$FE,$19,$31,$3F,$E1,$99,$C4,$F1,$8C,$3F,$E1,$98
	.dc.b	$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A5,$FF,$84,$7E,$1F,$FC,$A9,$28
	.dc.b	$FF,$34,$F0,$BF,$3C,$FF,$2F,$4E,$6F,$FC,$51,$18,$AE,$51,$FE,$2D
	.dc.b	$BE,$71,$CA,$75,$3C,$23,$FC,$71,$32,$3F,$CE,$F5,$D6,$FC,$15,$CA
	.dc.b	$78,$1D,$48,$27,$85,$F6,$AE,$33,$A9,$FF,$22,$5F,$39,$CC,$F1,$BF
	.dc.b	$31,$E3,$7F,$C6,$13,$BB,$E8,$75,$E2,$46,$F0,$F0,$DE,$CC,$1F,$12
	.dc.b	$FB,$1E,$A2,$E1,$E2,$9F,$2C,$79,$E7,$C6,$FC,$46,$BE,$43,$07,$D2
	.dc.b	$F8,$5F,$37,$C3,$48,$CD,$FD,$99,$F5,$19,$8F,$0E,$4A,$CC,$70,$84
	.dc.b	$EF,$0F,$9B,$5A,$AE,$33,$9C,$3E,$F4,$E6,$BE,$9C,$E6,$BF,$0F,$39
	.dc.b	$AF,$C2,$4E,$70,$FC,$74,$E6,$BF,$1B,$E0,$FD,$12,$3E,$5F,$86,$AF
	.dc.b	$AA,$2E,$3F,$82,$9B,$8F,$E4,$66,$D7,$E6,$26,$E1,$F9,$D9,$B9,$26
	.dc.b	$C6,$1B,$5F,$16,$3F,$09,$1F,$85,$F8,$69,$3F,$25,$FC,$47,$C3,$62
	.dc.b	$3E,$7C,$5C,$BF,$2F,$37,$2F,$D3,$4D,$AF,$D6,$CD,$CB,$F6,$53,$72
	.dc.b	$4D,$8C,$36,$BC,$47,$DB,$70,$F8,$8C,$1A,$30,$4C,$3C,$AE,$5F,$D6
	.dc.b	$CF,$68,$FC,$37,$C3,$7E,$94,$8F,$E2,$7C,$6A,$FE,$3A,$2D,$7E,$DA
	.dc.b	$6E,$5F,$C1,$66,$D7,$F0,$C9,$B9,$7F,$11,$9B,$97,$FB,$E2,$BC,$24
	.dc.b	$3B,$1E,$13,$18,$7F,$04,$9E,$37,$FB,$BF,$86,$FD,$E7,$C3,$4B,$FD
	.dc.b	$04,$65,$FD,$14,$65,$FD,$3C,$65,$FD,$4C,$66,$FF,$88,$CE,$5F,$E2
	.dc.b	$71,$9C,$BF,$91,$4E,$5F,$E4,$91,$9A,$FE,$57,$39,$7F,$96,$C6,$72
	.dc.b	$FE,$65,$3F,$C4,$FE,$1B,$F8,$A7,$C3,$4B,$FD,$C4,$65,$FD,$D4,$65
	.dc.b	$FD,$FC,$65,$FE,$03,$19,$C7,$F9,$4C,$E5,$FE,$57,$19,$C7,$F9,$D4
	.dc.b	$E5,$FE,$79,$19,$AF,$E8,$73,$97,$FA,$2C,$67,$0F,$E9,$53,$97,$FA
	.dc.b	$64,$67,$44,$E5,$46,$63,$0D,$AF,$E2,$DF,$0D,$FC,$D7,$E1,$A4,$60
	.dc.b	$87,$F1,$E7,$2F,$F2,$08,$FE,$78,$63,$2F,$E9,$33,$89,$E3,$18,$7F
	.dc.b	$92,$46,$5E,$46,$08,$60,$91,$C5,$F1,$1D,$32,$FE,$C9,$F0,$DF,$D3
	.dc.b	$49,$1F,$CC,$7E,$1A,$BF,$CD,$A2,$D7,$F5,$E9,$B5,$CA,$BF,$9C,$57
	.dc.b	$F5,$79,$CC,$7F,$76,$8B,$A7,$F7,$B9,$BA,$26,$C6,$1B,$5F,$D0,$FE
	.dc.b	$1B,$FB,$11,$1F,$E8,$DF,$0D,$5F,$E9,$71,$72,$FE,$E7,$37,$2F,$F0
	.dc.b	$14,$DA,$FF,$04,$CD,$CB,$FC,$1D,$37,$24,$D8,$C3,$6B,$C8,$78,$8A
	.dc.b	$78,$8B,$87,$89,$B8,$78,$9A,$78,$8B,$87,$89,$B8,$78,$9F,$EC,$9F
	.dc.b	$0B,$FE,$03,$F8,$59,$18,$21,$FD,$B1,$CB,$FD,$BA,$3F,$C1,$46,$30
	.dc.b	$FF,$0B,$CE,$27,$8C,$61,$FE,$E5,$19,$79,$18,$21,$82,$47,$17,$C4
	.dc.b	$74,$C3,$FC,$77,$F9,$EB,$E1,$7F,$C4,$4B,$E1,$A9,$88,$95,$FE,$0A
	.dc.b	$9D,$36,$72,$FF,$86,$A3,$A3,$92,$19,$9D,$DF,$F8,$F6,$79,$CF,$1A
	.dc.b	$CD,$F9,$AF,$C7,$3C,$AF,$84,$6A,$19,$B0,$E7,$5D,$70,$D4,$E7,$E1
	.dc.b	$7D,$E4,$BF,$67,$C3,$57,$D5,$16,$FC,$D3,$72,$F9,$33,$6B,$C3,$8A
	.dc.b	$B1,$87,$79,$F9,$15,$F3,$A7,$34,$8B,$18,$6C,$F0,$21,$E4,$3C,$E6
	.dc.b	$30,$F9,$51,$C2,$34,$48,$CB,$C0,$FA,$DF,$0D,$F8,$02,$3F,$77,$E1
	.dc.b	$AB,$A3,$F0,$11,$F8,$68,$CC,$FD,$98,$B9,$7E,$32,$6D,$7C,$49,$B9
	.dc.b	$7D,$59,$B9,$26,$C6,$1B,$5F,$8F,$F8,$6F,$CA,$91,$FC,$97,$C3,$57
	.dc.b	$F2,$11,$72,$FC,$64,$DC,$BF,$3F,$36,$BE,$BC,$DC,$BF,$11,$37,$24
	.dc.b	$D8,$C3,$6B,$C8,$78,$8F,$F3,$88,$FB,$47,$A5,$FE,$97,$87,$E4,$AE
	.dc.b	$8C,$36,$38,$D7,$06,$D6,$AA,$46,$08,$7F,$A2,$8E,$AE,$9C,$4F,$36
	.dc.b	$C9,$84,$3D,$0D,$73,$1C,$5D,$35,$CC,$6D,$E9,$19,$6C,$65,$EA,$6E
	.dc.b	$34,$E0,$F0,$38,$24,$7C,$97,$D5,$A7,$FB,$3E,$78,$36,$DD,$3B,$B9
	.dc.b	$5F,$1A,$B5,$38,$8F,$86,$67,$5F,$87,$7C,$A3,$F8,$F7,$C4,$7F,$24
	.dc.b	$23,$FC,$87,$E2,$2B,$FC,$7E,$2E,$1F,$C6,$E6,$DF,$F2,$E9,$B5,$FC
	.dc.b	$1A,$6E,$FF,$C5,$E6,$EE,$9B,$18,$6D,$79,$0F,$11,$FC,$CF,$E2,$3F
	.dc.b	$9C,$FC,$44,$8C,$10,$FE,$68,$E5,$FE,$6D,$1F,$D0,$0C,$65,$FC,$CE
	.dc.b	$71,$3C,$63,$0F,$F4,$18,$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$5F
	.dc.b	$D4,$BE,$23,$FA,$89,$23,$FA,$77,$9B,$5F,$E9,$91,$6B,$FA,$44,$DA
	.dc.b	$E5,$5F,$D5,$6B,$FA,$DC,$E6,$3F,$9A,$C5,$D3,$FA,$34,$DD,$13,$63
	.dc.b	$0D,$AE,$84,$3C,$67,$98,$C6,$1F,$D6,$E3,$84,$68,$91,$97,$81,$FD
	.dc.b	$9F,$E2,$3F,$B8,$91,$FE,$DD,$F1,$15,$D1,$FD,$C6,$3F,$BB,$C6,$67
	.dc.b	$FB,$54,$5C,$BF,$BE,$CD,$AF,$EA,$B3,$72,$FE,$CD,$37,$24,$D8,$C3
	.dc.b	$6B,$FC,$01,$F0,$FF,$E0,$62,$3F,$E0,$3F,$85,$AF,$F8,$06,$2E,$5F
	.dc.b	$DF,$66,$E5,$FE,$10,$9B,$5F,$DA,$26,$E5,$FD,$E6,$6E,$49,$B1,$86
	.dc.b	$D7,$90,$F1,$1F,$E1,$5F,$87,$FF,$0D,$FC,$3C,$8C,$10,$FF,$0B,$39
	.dc.b	$7F,$C3,$11,$FE,$22,$31,$97,$F8,$56,$71,$3C,$63,$0F,$F8,$8E,$32
	.dc.b	$F2,$30,$43,$04,$8E,$2F,$88,$E9,$97,$F8,$D2,$BC,$2F,$F8,$DA,$BA
	.dc.b	$D7,$F8,$9E,$73,$5F,$E1,$09,$CD,$71,$8F,$05,$76,$38,$3F,$E3,$9F
	.dc.b	$87,$FE,$83,$09,$FE,$89,$9B,$F3,$EF,$E8,$93,$9B,$FF,$16,$C6,$2B
	.dc.b	$FC,$93,$7F,$E4,$83,$9B,$CA,$35,$3C,$0D,$1E,$13,$FA,$95,$75,$BF
	.dc.b	$04,$72,$38,$1B,$3F,$C9,$97,$98,$DC,$72,$BF,$03,$D8,$D9,$93,$98
	.dc.b	$AF,$05,$F9,$AB,$FC,$91,$7C,$67,$35,$DE,$B5,$A2,$B1,$F1,$19,$F8
	.dc.b	$89,$7D,$F1,$F9,$53,$A8,$C6,$0A,$DC,$78,$F2,$F9,$15,$B8,$C1,$F4
	.dc.b	$27,$79,$7A,$27,$4F,$AA,$33,$5C,$AB,$BD,$61,$E4,$60,$87,$6C,$3F
	.dc.b	$01,$5B,$7D,$31,$E8,$F8,$8F,$85,$F1,$12,$30,$43,$96,$48,$D4,$62
	.dc.b	$7E,$51,$8C,$BE,$8C,$E2,$79,$46,$1F,$8F,$19,$79,$18,$21,$82,$47
	.dc.b	$17,$C7,$1C,$63,$EF,$CF,$87,$2F,$CB,$4E,$5F,$CC,$46,$6B,$F3,$F3
	.dc.b	$97,$F4,$31,$9C,$3F,$4D,$39,$7F,$51,$F1,$1F,$87,$F8,$89,$7F,$80
	.dc.b	$8F,$C5,$18,$7E,$3C,$61,$D1,$F8,$67,$58,$FE,$5A,$75,$3C,$0D,$1C
	.dc.b	$0F,$A2,$F0,$C7,$F6,$55,$C2,$B9,$57,$03,$91,$C0,$FC,$AF,$C4,$7E
	.dc.b	$7C,$8F,$E6,$3E,$22,$BF,$7A,$2C,$FE,$AA,$6D,$FF,$01,$9B,$5F,$C2
	.dc.b	$26,$E3,$FC,$36,$6E,$89,$B1,$86,$D7,$90,$F1,$1F,$AA,$F8,$8F,$DA
	.dc.b	$11,$FD,$6F,$81,$5F,$C9,$C5,$CB,$F7,$F3,$72,$FE,$2D,$36,$BF,$8E
	.dc.b	$CD,$CB,$F9,$14,$DC,$BF,$CA,$37,$C2,$FC,$44,$3F,$43,$7B,$AE,$0D
	.dc.b	$8D,$61,$E2,$6D,$73,$7C,$E7,$A8,$C9,$FC,$C2,$B9,$BF,$C2,$BE,$1F
	.dc.b	$F8,$A1,$1F,$E1,$BF,$11,$5F,$D9,$C5,$BF,$E3,$F3,$6F,$F9,$B4,$DA
	.dc.b	$FE,$7B,$37,$2F,$E8,$53,$72,$4D,$8C,$36,$BC,$87,$88,$FE,$3F,$F1
	.dc.b	$1F,$CA,$7E,$22,$46,$08,$7F,$08,$72,$FF,$0A,$8F,$E6,$06,$32,$FE
	.dc.b	$6F,$38,$9E,$31,$87,$F9,$6C,$65,$E4,$60,$86,$09,$1C,$5F,$11,$D3
	.dc.b	$2F,$EB,$5F,$0F,$FD,$07,$E2,$24,$FF,$3A,$8F,$E8,$86,$1F,$E4,$31
	.dc.b	$87,$47,$F4,$07,$55,$FD,$42,$75,$3C,$0D,$1C,$0F,$E5,$6F,$0A,$FE
	.dc.b	$D9,$5C,$2B,$95,$70,$39,$1C,$0E,$4F,$F5,$E8,$E5,$1F,$DE,$AB,$95
	.dc.b	$74,$8E,$55,$D1,$E4,$74,$79,$3D,$23,$FB,$F7,$C4,$7F,$65,$F8,$89
	.dc.b	$3F,$D7,$A3,$FB,$41,$87,$FA,$2C,$61,$D1,$FD,$91,$D5,$7F,$74,$9D
	.dc.b	$4F,$03,$47,$03,$FA,$7B,$C2,$BF,$C2,$15,$C2,$B9,$57,$03,$91,$C0
	.dc.b	$FE,$E7,$F1,$1F,$DF,$7E,$22,$46,$08,$7F,$5C,$72,$FF,$5E,$8F,$F0
	.dc.b	$21,$89,$FF,$05,$CE,$27,$8C,$61,$FF,$00,$C6,$5E,$46,$08,$60,$91
	.dc.b	$C5,$F1,$1D,$2F,$FC,$5D,$F1,$1F,$E1,$52,$47,$F8,$47,$DC,$FF,$73
	.dc.b	$57,$F8,$7A,$73,$3F,$DD,$9F,$F1,$54,$EA,$B9,$CF,$F8,$5A,$77,$5F
	.dc.b	$E3,$5A,$D4,$FF,$90,$A7,$35,$C2,$3F,$C1,$53,$AE,$2F,$F2,$55,$6A
	.dc.b	$78,$1E,$5A,$C4,$EA,$3C,$06,$4A,$79,$CB,$7E,$39,$E3,$C4,$BE,$78
	.dc.b	$78,$63,$17,$C2,$BC,$D7,$C6,$3E,$37,$C4,$7C,$22,$3F,$27,$E2,$2B
	.dc.b	$EC,$8B,$3F,$0E,$6D,$7F,$9A,$30,$56,$2D,$92,$57,$98,$C5,$59,$CD
	.dc.b	$79,$08,$78,$4F,$21,$8C,$3E,$A4,$F1,$BF,$B7,$F1,$12,$BF,$04,$FD
	.dc.b	$EF,$88,$C4,$7D,$68,$B7,$F7,$66,$E5,$EB,$9B,$5E,$AC,$55,$8C,$BD
	.dc.b	$95,$67,$35,$E0,$21,$DC,$FC,$A5,$6E,$78,$3F,$92,$F8,$7F,$C9,$FC
	.dc.b	$44,$AF,$CD,$46,$5F,$CE,$46,$5F,$C7,$C6,$5F,$C8,$C6,$72,$FC,$C4
	.dc.b	$E5,$FC,$D4,$67,$2F,$B1,$39,$7E,$D4,$66,$BE,$04,$E5,$F8,$51,$9C
	.dc.b	$BE,$E4,$E5,$FB,$D1,$9C,$93,$95,$19,$8C,$36,$BF,$71,$F1,$1F,$B4
	.dc.b	$23,$FB,$BF,$88,$AF,$EB,$A2,$E5,$FB,$59,$B9,$7E,$4E,$6D,$7D,$99
	.dc.b	$B9,$7E,$4E,$6E,$49,$B1,$86,$D7,$90,$F1,$1F,$C3,$7E,$1F,$F8,$49
	.dc.b	$1F,$E2,$1F,$11,$5F,$E0,$71,$72,$FE,$15,$37,$2F,$D4,$4D,$AF,$C7
	.dc.b	$4D,$CB,$F5,$13,$72,$4D,$8C,$36,$BC,$87,$88,$FE,$45,$F1,$1F,$C7
	.dc.b	$88,$FF,$25,$F8,$8A,$FF,$18,$8B,$97,$F1,$F9,$B9,$7E,$F6,$6D,$7E
	.dc.b	$8E,$6E,$5F,$BD,$9B,$92,$6C,$61,$B5,$E4,$3C,$45,$3C,$45,$C3,$C4
	.dc.b	$DC,$3C,$4D,$3C,$45,$C3,$C4,$DC,$3C,$4F,$F4,$37,$F9,$F9,$1F,$E8
	.dc.b	$DE,$C5,$7F,$9C,$C5,$BF,$E8,$13,$6F,$F8,$F4,$DA,$FE,$09,$36,$FF
	.dc.b	$8F,$4D,$11,$5C,$36,$7C,$87,$88,$FE,$B1,$F1,$1F,$D5,$08,$FF,$5B
	.dc.b	$F6,$EB,$FD,$3A,$2E,$3F,$D5,$66,$E3,$FC,$CE,$6D,$7F,$19,$9B,$8F
	.dc.b	$F3,$39,$B8,$A6,$C6,$1B,$5E,$43,$C4,$7F,$6D,$F8,$8F,$ED,$3F,$11
	.dc.b	$23,$04,$3F,$B1,$B9,$7F,$B2,$C7,$F7,$03,$19,$7F,$6D,$9C,$4F,$18
	.dc.b	$C3,$FD,$EE,$32,$F2,$30,$43,$04,$8E,$2F,$88,$E9,$97,$F5,$7A,$F0
	.dc.b	$BF,$D6,$6B,$AD,$7F,$7F,$9C,$D7,$F4,$19,$CD,$71,$8F,$05,$76,$38
	.dc.b	$3F,$E1,$1F,$87,$FF,$06,$FC,$3C,$8C,$10,$FF,$03,$B9,$7F,$C1,$31
	.dc.b	$FE,$14,$31,$3F,$E1,$19,$C4,$F1,$8C,$3F,$E1,$F8,$CB,$C8,$C1,$0C
	.dc.b	$12,$38,$BE,$23,$A5,$1F,$EF,$3F,$9F,$BE,$23,$10,$A5,$6F,$65,$EE
	.dc.b	$B9,$56,$9E,$77,$CA,$31,$5C,$3E,$23,$18,$73,$C3,$96,$1D,$E3,$18
	.dc.b	$6A,$39,$9D,$A3,$85,$66,$FC,$75,$6F,$9E,$1E,$06,$DF,$0C,$39,$1D
	.dc.b	$0F,$01,$67,$85,$73,$C3,$A3,$27,$8E,$1E,$06,$E0,$AD,$47,$49,$D3
	.dc.b	$67,$C6,$FD,$09,$F4,$7C,$3F,$A7,$E2,$24,$F4,$8C,$BD,$A3,$2F,$38
	.dc.b	$CB,$D6,$33,$8F,$C1,$9C,$BF,$0E,$33,$97,$D1,$9C,$BF,$4E,$33,$5F
	.dc.b	$84,$9C,$BF,$86,$8C,$E3,$F8,$29,$CB,$F8,$48,$CE,$29,$CA,$8C,$C6
	.dc.b	$1B,$5F,$2B,$E1,$FE,$D7,$C4,$48,$C1,$0F,$94,$E5,$F9,$91,$F8,$13
	.dc.b	$19,$7E,$06,$71,$3C,$63,$0F,$D8,$8C,$BC,$8C,$10,$C1,$23,$8B,$E2
	.dc.b	$3A,$65,$F9,$BF,$88,$FC,$91,$23,$F0,$BF,$11,$5F,$C1,$45,$AF,$CB
	.dc.b	$4D,$AE,$55,$F8,$CA,$FC,$E4,$E6,$3F,$6F,$17,$4F,$D9,$CD,$D1,$36
	.dc.b	$30,$DA,$FD,$08,$FB,$7F,$F3,$9B,$FA,$21,$F9,$3B,$CD,$FD,$94,$61
	.dc.b	$B3,$C3,$3F,$E0,$13,$CF,$C4,$DD,$F1,$8E,$ED,$7B,$57,$1D,$3F,$87
	.dc.b	$B6,$3F,$8A,$4E,$EF,$AB,$6B,$77,$E2,$3F,$8B,$D4,$BF,$5B,$63,$2F
	.dc.b	$02,$9E,$72,$C7,$07,$47,$07,$0F,$F2,$76,$BC,$4D,$47,$47,$07,$84
	.dc.b	$F9,$F1,$ED,$78,$5F,$B3,$CF,$F5,$46,$AA,$BF,$4F,$04,$E2,$BE,$7D
	.dc.b	$F3,$3A,$4E,$AB,$F8,$9F,$C4,$7F,$23,$F8,$89,$18,$21,$FC,$4D,$CB
	.dc.b	$FC,$56,$3F,$95,$18,$BF,$E5,$53,$89,$E3,$18,$7F,$90,$C6,$5E,$46
	.dc.b	$08,$60,$91,$C5,$F1,$1D,$32,$FE,$89,$F1,$1F,$CE,$FE,$22,$4F,$F2
	.dc.b	$E8,$FE,$7C,$61,$FE,$59,$18,$74,$7F,$34,$75,$5F,$D0,$E7,$53,$C0
	.dc.b	$D1,$C0,$FE,$6C,$F0,$AF,$EA,$D5,$C2,$B9,$57,$03,$91,$C0,$FE,$87
	.dc.b	$F0,$FF,$D4,$C8,$FF,$46,$F8,$8A,$FF,$43,$8B,$3F,$D5,$E6,$DF,$F6
	.dc.b	$59,$B5,$FD,$DA,$6F,$1F,$F7,$29,$BC,$69,$B1,$86,$D7,$90,$F1,$14
	.dc.b	$F1,$17,$0F,$13,$70,$F1,$34,$F1,$17,$0F,$13,$70,$F1,$3F,$D9,$3E
	.dc.b	$1F,$FB,$A7,$C4,$48,$C1,$0F,$EC,$8E,$5F,$EC,$D1,$FD,$EC,$C6,$1F
	.dc.b	$DE,$E7,$13,$C6,$30,$FF,$72,$8C,$BC,$8C,$10,$C1,$23,$8B,$E2,$3A
	.dc.b	$61,$FE,$15,$F8,$8F,$F0,$67,$C4,$49,$FE,$FF,$1F,$E0,$F3,$0F,$F7
	.dc.b	$D8,$C3,$A3,$FC,$0A,$EA,$BF,$C2,$93,$A9,$E0,$68,$E0,$7F,$81,$DE
	.dc.b	$15,$FE,$27,$AE,$15,$CA,$B8,$1C,$8E,$07,$F8,$53,$E2,$3F,$C4,$84
	.dc.b	$7F,$C2,$FF,$11,$5F,$F0,$A4,$59,$FF,$14,$4D,$BF,$F1,$B4,$DA,$FF
	.dc.b	$23,$CD,$D3,$FC,$87,$37,$44,$D8,$C3,$6B,$C8,$78,$8A,$78,$8B,$87
	.dc.b	$89,$B8,$78,$9A,$78,$8B,$87,$89,$B8,$78,$9F,$F1,$AF,$C3,$FF,$91
	.dc.b	$FF,$40,$FE,$8F,$87,$F9,$51,$B8,$7F,$94,$E7,$FC,$94,$6E,$A1,$C8
	.dc.b	$87,$F8,$F6,$78,$AC,$11,$87,$FC,$CF,$3D,$27,$FC,$93,$39,$35,$82
	.dc.b	$75,$3D,$4F,$09,$FE,$49,$73,$87,$A6,$32,$FF,$97,$E3,$DD,$F1,$1E
	.dc.b	$FF,$88,$95,$E0,$8C,$BE,$28,$CB,$EB,$8C,$BE,$D8,$CD,$7C,$79,$CB
	.dc.b	$F2,$63,$39,$7F,$A1,$71,$56,$A3,$13,$AB,$F8,$D5,$A7,$C8,$78,$C8
	.dc.b	$78,$CF,$8F,$F1,$1F,$67,$E2,$24,$60,$87,$D3,$72,$FD,$58,$FC,$01
	.dc.b	$8C,$7F,$01,$38,$9E,$31,$87,$EA,$C6,$5E,$46,$08,$60,$91,$C5,$F1
	.dc.b	$1D,$30,$F8,$95,$E1,$7E,$35,$75,$AF,$C9,$4E,$6B,$E2,$CE,$6B,$8C
	.dc.b	$78,$2B,$B1,$C1,$FC,$5F,$C4,$7E,$77,$E2,$24,$60,$87,$E5,$DC,$BF
	.dc.b	$99,$8F,$D2,$18,$9F,$D2,$4E,$27,$8C,$61,$FC,$CC,$65,$E4,$60,$86
	.dc.b	$09,$1C,$5F,$11,$D2,$BF,$11,$F1,$1F,$B5,$F8,$89,$3F,$A6,$8F,$DC
	.dc.b	$18,$7F,$61,$18,$74,$7E,$B1,$D5,$7E,$F6,$75,$3C,$0D,$1C,$0F,$DD
	.dc.b	$3C,$2B,$F3,$55,$C2,$B9,$57,$03,$91,$C0,$FD,$DF,$C3,$FF,$0D,$F8
	.dc.b	$89,$18,$21,$FC,$1D,$CB,$FC,$26,$3F,$8A,$18,$9F,$E2,$93,$89,$E3
	.dc.b	$18,$7F,$84,$C6,$5E,$46,$08,$60,$91,$C5,$F1,$1D,$2F,$F7,$15,$E1
	.dc.b	$7F,$75,$5D,$6B,$F9,$34,$E6,$BF,$73,$39,$AE,$31,$E0,$AE,$C7,$07
	.dc.b	$FD,$1F,$F7,$2E,$EF,$F3,$33,$63,$FC,$E6,$D7,$FC,$EE,$D7,$FC,$FA
	.dc.b	$D4,$47,$0D,$8E,$A4,$7A,$B5,$EA,$FF,$A0,$B9,$3F,$9B,$FC,$4F,$F4
	.dc.b	$5F,$0E,$57,$F3,$68,$FE,$90,$61,$FE,$81,$18,$74,$7F,$3C,$75,$87
	.dc.b	$F4,$D9,$FF,$54,$F9,$2F,$57,$FD,$0C,$DB,$FF,$49,$B5,$FF,$4B,$B5
	.dc.b	$FF,$4E,$B5,$F8,$21,$C3,$F1,$1A,$E6,$3B,$91,$EE,$D7,$FD,$53,$87
	.dc.b	$F5,$8F,$89,$FE,$CC,$48,$FE,$AF,$E4,$57,$FB,$04,$5C,$7F,$B5,$4D
	.dc.b	$D7,$FA,$14,$DA,$FE,$7F,$36,$FB,$47,$F6,$29,$FE,$E9,$39,$E6,$45
	.dc.b	$8C,$36,$78,$10,$F1,$9E,$63,$18,$7F,$7D,$8E,$11,$A2,$46,$5E,$07
	.dc.b	$F7,$BF,$88,$FF,$03,$7C,$4C,$8F,$EE,$F1,$FE,$09,$30,$FF,$80,$63
	.dc.b	$0E,$8F,$EF,$AE,$A7,$FC,$1D,$3A,$9E,$06,$8E,$07,$84,$C1,$A3,$04
	.dc.b	$8F,$F0,$C4,$F5,$CB,$FB,$85,$75,$9C,$D7,$27,$C2,$F6,$72,$F5,$72
	.dc.b	$F5,$9F,$EF,$55,$D6,$3F,$BB,$D7,$F8,$A7,$E2,$64,$7F,$86,$9F,$F1
	.dc.b	$67,$C4,$E2,$3F,$C4,$91,$70,$FF,$17,$4D,$C3,$FC,$0D,$36,$BF,$C0
	.dc.b	$73,$74,$FF,$14,$CD,$D1,$36,$30,$DA,$F1,$1F,$E2,$C7,$0F,$88,$C1
	.dc.b	$A3,$04,$C3,$CA,$E5,$FF,$22,$4F,$68,$FF,$1B,$FC,$4F,$F9,$24,$8F
	.dc.b	$F8,$EB,$E2,$6B,$FE,$43,$8B,$5F,$E4,$D9,$B9,$7F,$89,$66,$D7,$F8
	.dc.b	$8A,$6E,$5F,$E4,$89,$B9,$26,$C6,$1B,$5E,$43,$C4,$53,$C4,$5C,$3C
	.dc.b	$4D,$C3,$C4,$D3,$C4,$5C,$3C,$4D,$C3,$C4,$FF,$96,$3E,$27,$F8,$8D
	.dc.b	$AF,$F9,$9D,$BF,$41,$C5,$FF,$36,$AB,$39,$AF,$F3,$95,$CA,$D4,$72
	.dc.b	$BE,$30,$BE,$31,$87,$FC,$F1,$5C,$67,$55,$C4,$F2,$95,$F4,$10,$D9
	.dc.b	$FE,$77,$AE,$91,$B8,$FF,$3E,$4F,$5A,$F0,$CF,$49,$E0,$E4,$FF,$30
	.dc.b	$4E,$6F,$DD,$3E,$C3,$67,$86,$BD,$2E,$63,$6F,$91,$B1,$A8,$F1,$9B
	.dc.b	$BF,$3B,$D0,$C9,$23,$CE,$EE,$FC,$CF,$BF,$E2,$7C,$D4,$F4,$69,$88
	.dc.b	$93,$DE,$32,$F8,$63,$2F,$9E,$79,$3B,$34,$43,$53,$C9,$E1,$3C,$AF
	.dc.b	$EE,$CE,$5F,$BF,$19,$BF,$C2,$4E,$5F,$C3,$46,$6B,$ED,$4E,$5F,$71
	.dc.b	$E1,$AE,$AF,$94,$F0,$99,$AE,$AE,$63,$C2,$FC,$FF,$89,$F8,$FF,$13
	.dc.b	$23,$04,$3E,$13,$97,$E2,$47,$CE,$31,$3F,$9A,$9C,$4F,$18,$C3,$F7
	.dc.b	$63,$2F,$23,$04,$30,$48,$E2,$F8,$8E,$95,$FA,$BA,$F0,$BF,$AD,$AE
	.dc.b	$B5,$FA,$C9,$CD,$7E,$8E,$73,$5C,$63,$C1,$5D,$8E,$0F,$E4,$FE,$27
	.dc.b	$F1,$64,$7F,$2D,$F1,$35,$FC,$34,$59,$FD,$DC,$DA,$FE,$01,$36,$BF
	.dc.b	$69,$37,$4F,$E1,$93,$74,$4D,$8C,$36,$BC,$87,$88,$A7,$88,$B8,$78
	.dc.b	$9B,$87,$89,$A7,$88,$B8,$78,$9B,$87,$89,$FD,$A7,$C4,$FE,$B0,$8F
	.dc.b	$ED,$FE,$26,$BF,$A7,$8B,$87,$F1,$C9,$B8,$7F,$21,$9B,$5F,$C5,$66
	.dc.b	$E1,$FC,$B6,$6E,$09,$B1,$86,$D7,$90,$F1,$14,$F1,$17,$0F,$13,$70
	.dc.b	$F1,$34,$F1,$17,$0F,$13,$70,$F1,$3F,$C5,$7E,$27,$F8,$69,$1F,$E3
	.dc.b	$1F,$13,$5F,$E1,$11,$70,$FE,$7B,$37,$0F,$E8,$53,$6B,$F9,$B4,$DC
	.dc.b	$3F,$A7,$4D,$C1,$36,$30,$DA,$F2,$1E,$22,$9E,$22,$E1,$E2,$6E,$1E
	.dc.b	$26,$9E,$22,$E1,$E2,$6E,$1E,$27,$F9,$B7,$C4,$FF,$2E,$F8,$99,$18
	.dc.b	$21,$FC,$A5,$CB,$FC,$AE,$3F,$9A,$98,$C3,$FB,$0C,$E2,$78,$C6,$1F
	.dc.b	$E8,$91,$97,$91,$82,$18,$24,$71,$7C,$47,$4C,$3F,$B9,$57,$85,$FE
	.dc.b	$E9,$5D,$6B,$FB,$9C,$E6,$BF,$B5,$4E,$6B,$8C,$78,$2B,$B1,$C1,$FE
	.dc.b	$B5,$F1,$3F,$D4,$BB,$A4,$60,$87,$F4,$D7,$2F,$F4,$F8,$FE,$B2,$62
	.dc.b	$7F,$C0,$B3,$89,$E3,$18,$7F,$B4,$46,$5E,$46,$08,$60,$91,$C5,$F1
	.dc.b	$1D,$2B,$FC,$2D,$5E,$17,$FC,$31,$5D,$6B,$FC,$2F,$39,$AF,$F0,$74
	.dc.b	$E6,$B8,$C7,$82,$BB,$1C,$1F,$EF,$DF,$13,$FD,$DB,$E2,$64,$60,$87
	.dc.b	$F7,$17,$2F,$F7,$38,$FE,$FA,$62,$7F,$C5,$33,$89,$E3,$18,$7F,$C1
	.dc.b	$91,$97,$91,$82,$18,$24,$71,$7C,$47,$4A,$FF,$1D,$57,$85,$FF,$1E
	.dc.b	$57,$5A,$FF,$1D,$CE,$6B,$FC,$65,$39,$AE,$31,$E0,$AE,$C7,$07,$FC
	.dc.b	$58,$3E,$D7,$F8,$8B,$FE,$13,$F8,$AF,$F0,$E9,$FE,$37,$75,$C3,$67
	.dc.b	$75,$FE,$4C,$AD,$D7,$49,$CD,$FF,$BB,$9B,$1F,$E4,$79,$DB,$FE,$4A
	.dc.b	$3A,$9C,$6F,$FD,$EA,$C9,$F2,$B6,$32,$F5,$6A,$39,$9A,$8E,$2E,$CD
	.dc.b	$B8,$78,$35,$E2,$77,$37,$1D,$9E,$B7,$FE,$6E,$9F,$6C,$62,$FF,$CE
	.dc.b	$5F,$15,$F1,$EE,$4F,$B4,$FF,$29,$46,$1F,$85,$5C,$8F,$8F,$C1,$FE
	.dc.b	$83,$9F,$F2,$7F,$C5,$4A,$FF,$2D,$3F,$E5,$2F,$8A,$C4,$7F,$92,$E2
	.dc.b	$E5,$FE,$80,$9B,$97,$FA,$16,$6D,$7F,$9E,$66,$E1,$FE,$91,$9B,$C4
	.dc.b	$9B,$18,$6D,$78,$8F,$F3,$63,$87,$C4,$60,$D1,$82,$61,$E5,$72,$FF
	.dc.b	$A5,$A7,$B4,$7F,$9F,$7C,$57,$C5,$7F,$A1,$5B,$19,$8C,$7C,$56,$59
	.dc.b	$19,$39,$46,$1F,$F4,$34,$F2,$6C,$F2,$78,$18,$2D,$FF,$AC,$A5,$5C
	.dc.b	$E3,$15,$73,$FF,$5D,$D6,$AF,$8C,$62,$75,$0C,$D5,$B8,$F0,$12,$FA
	.dc.b	$BE,$A3,$99,$C6,$3D,$51,$A8,$F6,$E1,$DA,$BA,$DD,$BE,$D7,$EB,$F8
	.dc.b	$AF,$99,$F1,$53,$1F,$A1,$1F,$3C,$C3,$F2,$23,$0E,$8F,$AC,$EA,$7A
	.dc.b	$46,$4E,$06,$8C,$3A,$72,$47,$04,$CB,$E5,$CF,$57,$E6,$D7,$8A,$B8
	.dc.b	$46,$9E,$AE,$9F,$01,$A3,$E4,$47,$1A,$FB,$95,$D6,$72,$7E,$47,$CA
	.dc.b	$FC,$87,$C5,$48,$C1,$0F,$C5,$B9,$7F,$1B,$1F,$97,$31,$5F,$89,$9C
	.dc.b	$4F,$18,$C3,$F9,$F8,$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$1F,$92
	.dc.b	$AF,$0B,$F9,$4A,$EB,$5F,$9B,$9C,$D7,$E2,$E7,$35,$C6,$3C,$15,$D8
	.dc.b	$E0,$FE,$D7,$E2,$7F,$69,$F1,$52,$30,$43,$F5,$EE,$5F,$D8,$C7,$EE
	.dc.b	$CC,$4F,$EB,$67,$13,$FD,$2F,$B7,$F9,$87,$21,$A4,$8F,$E7,$6D,$7F
	.dc.b	$A0,$35,$FE,$94,$D7,$FA,$A3,$5F,$D5,$7C,$47,$F0,$DF,$46,$47,$F1
	.dc.b	$18,$FE,$20,$61,$FE,$13,$18,$74,$7F,$17,$75,$87,$F0,$79,$FE,$D1
	.dc.b	$DF,$EA,$25,$7F,$52,$6B,$FD,$5D,$AF,$F6,$16,$BF,$DA,$DA,$FF,$58
	.dc.b	$7F,$B0,$F1,$7F,$0A,$F8,$8F,$E4,$DF,$11,$2B,$F9,$54,$7F,$29,$30
	.dc.b	$FF,$22,$8C,$3A,$3F,$98,$3A,$C3,$F9,$04,$FF,$80,$FC,$57,$17,$F6
	.dc.b	$36,$DF,$F7,$26,$BF,$DD,$9A,$FF,$7C,$6B,$FD,$DC,$38,$7F,$BD,$35
	.dc.b	$CC,$77,$23,$DD,$AF,$F8,$23,$0F,$E5,$7F,$11,$FD,$1F,$E2,$24,$FF
	.dc.b	$4B,$8F,$E9,$46,$1F,$E8,$71,$87,$47,$F5,$17,$58,$FF,$41,$9D,$4F
	.dc.b	$03,$47,$03,$FA,$6B,$C3,$8B,$FA,$1D,$70,$AE,$55,$C0,$E4,$70,$39
	.dc.b	$3F,$D3,$E3,$94,$7F,$3F,$AE,$55,$D2,$39,$57,$47,$91,$D1,$E4,$F4
	.dc.b	$8F,$EA,$BF,$11,$FD,$AF,$E2,$24,$FF,$6F,$8F,$ED,$C6,$1F,$EC,$F1
	.dc.b	$87,$47,$F7,$57,$55,$FD,$96,$75,$3C,$0D,$1C,$0F,$EE,$2F,$0A,$FE
	.dc.b	$CF,$5C,$2B,$95,$70,$39,$1C,$0E,$4F,$F7,$38,$E5,$1F,$D8,$EB,$95
	.dc.b	$74,$8E,$55,$D1,$E4,$74,$79,$3D,$23,$FB,$CF,$C4,$7F,$83,$FE,$22
	.dc.b	$4F,$F8,$4E,$3F,$C2,$46,$1F,$F0,$5C,$61,$D1,$FE,$19,$75,$5F,$E0
	.dc.b	$99,$D4,$F0,$34,$70,$3F,$C2,$AF,$0A,$FF,$05,$D7,$0A,$E5,$5C,$0E
	.dc.b	$47,$03,$93,$FE,$17,$8E,$51,$FE,$07,$AE,$55,$D2,$39,$57,$47,$91
	.dc.b	$D1,$E4,$F4,$8F,$F0,$EF,$C4,$7F,$8C,$FE,$22,$4F,$F8,$DE,$3F,$C6
	.dc.b	$C6,$1F,$F1,$7C,$61,$D1,$FE,$3D,$75,$5F,$E2,$D9,$D4,$F0,$34,$70
	.dc.b	$3F,$C7,$2F,$0A,$FF,$17,$D7,$0A,$E5,$5C,$0E,$47,$03,$93,$FE,$3B
	.dc.b	$8E,$51,$FE,$2B,$AE,$55,$D2,$39,$57,$47,$91,$D1,$E4,$F4,$8F,$F2
	.dc.b	$0F,$C3,$FF,$1D,$94,$7F,$A5,$76,$7E,$85,$FE,$9E,$9C,$DF,$F9,$0E
	.dc.b	$31,$5C,$A3,$FC,$C5,$7C,$E3,$94,$EA,$78,$1F,$E6,$59,$8F,$FA,$8E
	.dc.b	$BA,$DF,$82,$39,$1C,$0E,$A4,$11,$C2,$BB,$4F,$19,$D4,$FF,$98,$EB
	.dc.b	$9C,$E6,$78,$DF,$99,$F1,$9F,$E7,$09,$DD,$F4,$3A,$F7,$23,$78,$78
	.dc.b	$6B,$66,$2F,$E2,$3D,$8F,$51,$70,$F1,$4F,$96,$3C,$F3,$E3,$7E,$23
	.dc.b	$5F,$21,$83,$E9,$7C,$47,$9B,$E2,$24,$66,$FE,$CC,$FA,$8C,$C7,$87
	.dc.b	$25,$66,$38,$42,$77,$87,$CD,$AD,$57,$19,$CE,$1F,$7A,$73,$5F,$4E
	.dc.b	$73,$5F,$87,$9C,$D7,$E1,$27,$38,$7E,$3A,$73,$5E,$C3,$C6,$EB,$0F
	.dc.b	$24,$79,$AB,$13,$C2,$BD,$F3,$C0,$87,$9C,$F6,$FC,$47,$DC,$F8,$89
	.dc.b	$18,$21,$F1,$5C,$BF,$1E,$3F,$06,$63,$2F,$C8,$4E,$27,$8C,$61,$F9
	.dc.b	$71,$97,$91,$82,$18,$24,$71,$7C,$47,$4C,$BF,$57,$5E,$17,$F5,$B5
	.dc.b	$D6,$BF,$45,$39,$AF,$DB,$4E,$6B,$8C,$78,$2B,$B1,$C1,$FC,$0F,$C4
	.dc.b	$7E,$80,$8F,$E1,$3E,$22,$BF,$89,$8B,$3F,$B1,$9B,$5F,$C0,$26,$D7
	.dc.b	$F0,$79,$B8,$FF,$0C,$9B,$A2,$6C,$61,$B5,$E4,$3C,$45,$3C,$45,$C3
	.dc.b	$C4,$DC,$3C,$4D,$3C,$45,$C3,$C4,$DC,$3C,$4F,$E8,$BE,$23,$F8,$0F
	.dc.b	$C4,$48,$C1,$0F,$D5,$39,$7F,$59,$1F,$C2,$0C,$61,$FC,$56,$71,$3C
	.dc.b	$63,$0F,$EC,$23,$2F,$23,$04,$30,$48,$E2,$F8,$8E,$98,$7F,$30,$AF
	.dc.b	$0B,$FC,$CA,$BA,$D7,$F2,$79,$CD,$7F,$3B,$9C,$D7,$18,$F0,$57,$63
	.dc.b	$83,$FC,$1B,$E2,$3F,$92,$FC,$44,$8C,$10,$FE,$1C,$E5,$FE,$21,$1F
	.dc.b	$CB,$0C,$4F,$F3,$99,$C4,$F1,$8C,$3F,$C5,$23,$2F,$23,$04,$30,$48
	.dc.b	$E2,$F8,$8E,$95,$FD,$52,$B9,$CF,$F4,$97,$F9,$FF,$C3,$62,$75,$5C
	.dc.b	$0F,$EC,$11,$97,$FB,$14,$DC,$FF,$B3,$CF,$F4,$8F,$87,$95,$FC,$B1
	.dc.b	$FE,$97,$F0,$F8,$8F,$E6,$51,$71,$FE,$B5,$37,$1F,$ED,$33,$6B,$FB
	.dc.b	$84,$DC,$3F,$BA,$CD,$D5,$36,$30,$DA,$F1,$1F,$CF,$9C,$3E,$23,$06
	.dc.b	$8C,$13,$0F,$2B,$97,$FB,$84,$F6,$8F,$E9,$3C,$BF,$DA,$3E,$1A,$46
	.dc.b	$08,$7F,$51,$72,$FF,$53,$8F,$ED,$E6,$2B,$FB,$E4,$E2,$78,$C6,$1F
	.dc.b	$EA,$F1,$97,$91,$82,$18,$24,$71,$7C,$47,$4C,$BF,$C2,$F5,$E1,$7F
	.dc.b	$C3,$35,$D6,$BF,$C1,$D3,$9A,$FF,$11,$4E,$6B,$8C,$78,$2B,$B1,$C1
	.dc.b	$FE,$DB,$F0,$FF,$E0,$CF,$86,$91,$82,$1F,$DD,$5C,$BF,$DD,$E3,$FC
	.dc.b	$26,$62,$7F,$C4,$13,$89,$E3,$18,$7F,$BD,$C6,$5E,$46,$08,$60,$91
	.dc.b	$C5,$F1,$1D,$2B,$FC,$77,$5E,$17,$FC,$7B,$5D,$6B,$FC,$65,$39,$AF
	.dc.b	$F2,$34,$E6,$B8,$C7,$82,$BB,$1C,$1F,$F0,$E0,$FB,$66,$FF,$E4,$28
	.dc.b	$71,$7F,$36,$7F,$C5,$86,$38,$F5,$5C,$23,$9B,$5D,$CF,$1B,$FF,$2D
	.dc.b	$C7,$13,$77,$C8,$F0,$4E,$F0,$4C,$8C,$38,$BE,$D7,$D1,$F0,$47,$03
	.dc.b	$57,$59,$1A,$74,$D7,$66,$CA,$57,$4F,$C7,$8D,$3A,$3A,$B8,$74,$6A
	.dc.b	$38,$9E,$8B,$F0,$CF,$C1,$9F,$A7,$6E,$5C,$47,$D0,$BE,$0F,$93,$0F
	.dc.b	$97,$5B,$8F,$96,$43,$85,$FC,$CB,$FF,$36,$FC,$44,$9F,$F2,$8B,$FE
	.dc.b	$70,$F8,$8C,$47,$F9,$5E,$2E,$1F,$E7,$F9,$B9,$7F,$A3,$A6,$D7,$FA
	.dc.b	$56,$6F,$1F,$FA,$6A,$6F,$1A,$6C,$61,$B5,$E2,$3F,$CC,$6E,$1F,$11
	.dc.b	$83,$46,$09,$87,$95,$CB,$FE,$95,$9E,$D1,$FE,$6E,$F8,$8F,$F4,$69
	.dc.b	$1F,$F3,$8F,$C3,$D7,$FC,$ED,$16,$BF,$D3,$D3,$72,$FF,$56,$CD,$AF
	.dc.b	$F5,$A4,$DC,$BF,$D7,$33,$72,$4D,$8C,$36,$BC,$87,$88,$A7,$88,$B8
	.dc.b	$78,$9B,$87,$89,$A7,$88,$B8,$78,$9B,$87,$89,$FF,$4B,$7E,$87,$FB
	.dc.b	$2B,$FF,$60,$B6,$FF,$DA,$4D,$95,$A1,$F1,$1F,$EB,$C7,$51,$9A,$C1
	.dc.b	$99,$B5,$E0,$8A,$FF,$AC,$EB,$9B,$92,$19,$8E,$4E,$2F,$CF,$1C,$9F
	.dc.b	$4D,$58,$F4,$96,$7D,$35,$D9,$E8,$73,$8D,$BF,$12,$78,$5F,$AE,$34
	.dc.b	$72,$8C,$D7,$C0,$23,$F0,$FE,$23,$E2,$7C,$44,$AF,$9F,$19,$7E,$8C
	.dc.b	$65,$F8,$51,$97,$E2,$46,$71,$FA,$B3,$97,$EB,$C6,$71,$FA,$F3,$97
	.dc.b	$EC,$C6,$6B,$E0,$CE,$5F,$87,$19,$C7,$EE,$4E,$5B,$3E,$83,$F0,$FE
	.dc.b	$5E,$72,$55,$8C,$36,$7F,$1B,$F1,$1F,$85,$F8,$89,$18,$21,$F8,$17
	.dc.b	$2F,$E0,$E3,$F1,$86,$33,$FC,$AC,$E2,$7F,$AB,$FD,$BF,$DB,$7F,$9E
	.dc.b	$35,$FE,$80,$D7,$FA,$4B,$53,$49,$AF,$F3,$76,$C7,$F3,$C6,$BF,$D0
	.dc.b	$DE,$99,$FE,$8F,$E2,$7F,$45,$F0,$52,$BF,$5B,$1F,$A5,$30,$FE,$7E
	.dc.b	$30,$E8,$FD,$9B,$AC,$7F,$67,$3A,$9E,$06,$8E,$07,$EA,$5E,$1B,$FE
	.dc.b	$E2,$B8,$57,$2A,$E0,$72,$38,$1F,$C0,$BD,$6F,$DD,$7A,$92,$30,$43
	.dc.b	$F6,$CE,$5F,$DC,$47,$F0,$13,$13,$FC,$2A,$71,$3F,$5B,$6F,$98,$47
	.dc.b	$E6,$35,$F9,$8D,$7E,$DB,$5F,$B6,$D7,$ED,$BF,$DC,$32,$FE,$29,$F1
	.dc.b	$3F,$C4,$FE,$26,$57,$F1,$E8,$FE,$2C,$61,$FE,$21,$18,$74,$7F,$24
	.dc.b	$75,$87,$F2,$49,$FF,$11,$6A,$E3,$FC,$31,$2F,$F0,$CD,$7F,$0C,$D7
	.dc.b	$F2,$AD,$7F,$2A,$D7,$F2,$AF,$AF,$8D,$F1,$3F,$CB,$7E,$26,$57,$F3
	.dc.b	$88,$FE,$60,$61,$FE,$55,$18,$74,$7F,$3E,$75,$8F,$F3,$E9,$D4,$F0
	.dc.b	$34,$70,$3F,$9A,$3C,$38,$3F,$A1,$D7,$0A,$E5,$5C,$0E,$47,$03,$93
	.dc.b	$FD,$32,$39,$47,$F3,$2A,$E5,$5D,$23,$95,$74,$79,$1D,$1E,$4F,$48
	.dc.b	$FE,$AB,$F1,$3F,$D4,$49,$1F,$D7,$3E,$26,$BF,$D3,$62,$D7,$F5,$E9
	.dc.b	$B5,$CA,$BF,$B1,$D7,$F5,$B9,$CC,$7F,$48,$8B,$B7,$F6,$19,$B9,$26
	.dc.b	$C6,$1B,$5D,$08,$78,$CF,$31,$8C,$3F,$B8,$C7,$08,$D1,$23,$2F,$03
	.dc.b	$FB,$47,$C4,$FF,$71,$23,$FD,$F3,$E2,$6B,$A3,$FB,$FC,$7F,$77,$8C
	.dc.b	$CF,$F8,$02,$2E,$5F,$E0,$09,$B5,$FD,$A2,$6E,$5F,$E0,$19,$B9,$26
	.dc.b	$C6,$1B,$5E,$02,$1E,$33,$FC,$01,$1B,$73,$87,$F8,$4A,$31,$1B,$8C
	.dc.b	$3F,$E1,$8F,$87,$FF,$08,$11,$FF,$0D,$FC,$45,$7F,$C1,$91,$67,$FC
	.dc.b	$3B,$37,$2F,$F0,$EC,$DA,$FF,$02,$CD,$CB,$FC,$3D,$37,$24,$D8,$C3
	.dc.b	$6B,$C8,$78,$8A,$78,$8B,$87,$89,$B8,$78,$9A,$78,$8B,$87,$89,$B8
	.dc.b	$78,$9F,$F5,$1D,$2B,$FD,$4B,$4F,$D1,$1F,$11,$FE,$A1,$AD,$5F,$F8
	.dc.b	$EA,$5F,$13,$9C,$38,$C6,$27,$AC,$62,$F5,$1C,$5B,$87,$7B,$B7,$C6
	.dc.b	$98,$78,$2B,$C9,$7A,$37,$72,$BB,$9D,$4B,$3C,$1B,$7D,$6F,$A3,$70
	.dc.b	$F3,$4E,$B0,$E8,$F8,$A3,$DB,$87,$86,$34,$74,$77,$7D,$2B,$53,$C4
	.dc.b	$F9,$13,$AA,$EF,$19,$C3,$CC,$DC,$3C,$8F,$03,$B9,$9B,$E4,$E4,$C1
	.dc.b	$EE,$F8,$9F,$96,$47,$E0,$FC,$05,$7B,$45,$CB,$E1,$CD,$C7,$ED,$CD
	.dc.b	$AF,$C0,$4D,$C7,$F0,$D3,$71,$4D,$8C,$36,$BC,$87,$88,$A7,$88,$B8
	.dc.b	$78,$9B,$87,$89,$A7,$88,$B8,$78,$9B,$87,$89,$FB,$FF,$13,$F8,$CF
	.dc.b	$89,$91,$82,$1F,$3D,$CB,$F4,$63,$F2,$66,$30,$FC,$3C,$E2,$78,$C6
	.dc.b	$1F,$C6,$C6,$5E,$46,$08,$60,$91,$C5,$F1,$1D,$30,$FD,$5D,$78,$5F
	.dc.b	$D6,$D7,$5A,$FC,$CC,$E6,$BF,$69,$39,$AE,$31,$E0,$AE,$C7,$07,$F4
	.dc.b	$7F,$13,$FB,$02,$3F,$A6,$F8,$9A,$FE,$4E,$2C,$FE,$A2,$6D,$7F,$00
	.dc.b	$9B,$5F,$C1,$66,$E5,$FC,$2E,$6E,$49,$B1,$86,$D7,$90,$F1,$14,$F1
	.dc.b	$17,$0F,$13,$70,$F1,$34,$F1,$17,$0F,$13,$70,$F1,$3F,$C1,$3E,$1F
	.dc.b	$F8,$8F,$C4,$48,$C1,$0F,$DA,$39,$7F,$6D,$1F,$C6,$0C,$61,$FC,$32
	.dc.b	$71,$3C,$63,$0F,$F1,$28,$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$1F
	.dc.b	$CC,$2B,$C2,$FF,$32,$AE,$B5,$FC,$7A,$73,$5F,$CE,$67,$35,$C6,$3C
	.dc.b	$15,$D8,$E0,$FF,$28,$F8,$7F,$E6,$BF,$11,$23,$04,$3F,$8C,$39,$7F
	.dc.b	$8D,$47,$F3,$C3,$13,$FC,$C2,$71,$3C,$63,$0F,$F3,$68,$CB,$C8,$C1
	.dc.b	$0C,$12,$38,$BE,$23,$A5,$7F,$54,$AE,$73,$FD,$01,$FE,$9F,$F0,$F8
	.dc.b	$9D,$57,$03,$FA,$F4,$65,$FE,$C1,$37,$3F,$EC,$D3,$FD,$63,$E2,$3F
	.dc.b	$AC,$FC,$3C,$BF,$E9,$F1,$97,$FA,$8C,$65,$FE,$7D,$19,$7F,$A0,$46
	.dc.b	$71,$FE,$AB,$39,$7F,$AB,$C6,$71,$FE,$DF,$39,$7F,$B8,$C6,$6B,$FB
	.dc.b	$B4,$E5,$FE,$F1,$19,$C3,$FB,$F4,$E5,$FF,$00,$46,$75,$4E,$54,$66
	.dc.b	$30,$DA,$FE,$DB,$F1,$1F,$DE,$BE,$1E,$46,$08,$7F,$5D,$72,$FF,$5F
	.dc.b	$8F,$F0,$09,$8C,$BF,$BA,$CE,$27,$8C,$61,$FE,$F7,$19,$79,$18,$21
	.dc.b	$82,$47,$17,$C4,$74,$CB,$FC,$33,$5C,$E7,$FC,$0A,$FF,$85,$3E,$23
	.dc.b	$13,$AA,$E0,$7F,$88,$E3,$2F,$F8,$96,$6E,$7F,$E2,$B9,$FF,$0E,$FC
	.dc.b	$3F,$F8,$7B,$E2,$25,$FF,$85,$23,$2F,$F8,$5A,$32,$FF,$80,$E3,$2F
	.dc.b	$F8,$16,$33,$8F,$F8,$6A,$72,$FF,$87,$23,$38,$FF,$8D,$27,$2F,$F8
	.dc.b	$DA,$33,$5F,$E3,$B9,$CB,$FE,$3D,$8C,$E1,$FE,$43,$9C,$BF,$E4,$58
	.dc.b	$CE,$A9,$CA,$8C,$C6,$1B,$5F,$E3,$61,$E2,$BD,$0F,$F2,$2E,$4F,$DD
	.dc.b	$3F,$BD,$57,$F8,$A6,$31,$D5,$FE,$EA,$A9,$3F,$E4,$09,$CB,$6B,$8C
	.dc.b	$7F,$96,$23,$BC,$F0,$7A,$4E,$B4,$FF,$31,$B6,$07,$03,$D1,$87,$9D
	.dc.b	$C9,$0D,$14,$A6,$5E,$0E,$5A,$E6,$B8,$3E,$08,$CB,$D9,$AE,$63,$6F
	.dc.b	$63,$4E,$23,$71,$E4,$BF,$85,$1E,$26,$CF,$D0,$A7,$57,$C5,$9C,$DF
	.dc.b	$C0,$AD,$4E,$70,$EF,$39,$3E,$84,$7F,$9E,$BE,$27,$FC,$FB,$C5,$F1
	.dc.b	$70,$FF,$3D,$36,$FF,$D0,$F3,$BB,$FF,$3B,$CE,$56,$FF,$C4,$1D,$1F
	.dc.b	$E6,$59,$EA,$C3,$FC,$FA,$EA,$FB,$C6,$4E,$06,$8E,$E6,$5F,$F4,$1C
	.dc.b	$71,$BF,$F5,$1D,$F8,$EF,$C5,$7C,$4E,$67,$F9,$E6,$34,$6F,$75,$F3
	.dc.b	$9F,$5C,$E9,$C5,$7F,$AB,$E7,$C5,$7E,$09,$CD,$3E,$27,$E5,$92,$3E
	.dc.b	$3F,$C4,$D7,$C5,$17,$2F,$9F,$36,$FD,$73,$6B,$C7,$3F,$58,$C6,$3F
	.dc.b	$CF,$A7,$13,$D8,$FA,$AE,$B0,$56,$A7,$C2,$68,$E0,$7C,$57,$84,$EF
	.dc.b	$0F,$2C,$EC,$E2,$6C,$E2,$FE,$22,$3C,$13,$F8,$CF,$89,$FC,$6F,$B7
	.dc.b	$2F,$F1,$31,$97,$F1,$71,$97,$EB,$46,$5F,$B1,$19,$BF,$CA,$CE,$5F
	.dc.b	$CB,$C6,$72,$FC,$1C,$E5,$FC,$2C,$66,$BE,$EC,$E5,$FB,$F1,$9C,$BF
	.dc.b	$B4,$CE,$5F,$ED,$71,$9C,$93,$95,$19,$8C,$36,$BF,$4F,$F1,$3F,$AE
	.dc.b	$F8,$09,$18,$21,$F9,$57,$2F,$E5,$E3,$F6,$A6,$32,$FD,$AC,$E2,$78
	.dc.b	$C7,$EE,$27,$11,$F9,$FF,$7E,$CE,$63,$79,$FE,$A2,$72,$FE,$AA,$AC
	.dc.b	$FE,$8A,$6E,$7F,$E0,$C9,$B9,$A6,$C6,$1B,$5D,$8E,$47,$F0,$8F,$89
	.dc.b	$FE,$1F,$F0,$12,$30,$43,$F6,$CE,$5F,$DC,$47,$F1,$63,$19,$7F,$16
	.dc.b	$9C,$4F,$F4,$0F,$6F,$F9,$61,$1F,$E6,$FC,$C8,$C4,$7F,$3D,$6B,$FD
	.dc.b	$05,$AF,$F3,$B6,$BF,$CF,$DF,$E8,$39,$FF,$8B,$BE,$27,$F9,$4F,$C4
	.dc.b	$CA,$FE,$4B,$1F,$CB,$0C,$3F,$C4,$63,$0E,$8F,$E5,$6E,$B1,$FE,$63
	.dc.b	$3F,$D6,$1F,$E8,$7E,$EF,$F4,$72,$57,$B1,$AF,$F5,$76,$BF,$D6,$DA
	.dc.b	$FF,$55,$6B,$FD,$65,$F5,$EC,$F8,$8F,$E8,$7F,$13,$2B,$F9,$FC,$7F
	.dc.b	$46,$30,$FF,$2B,$8C,$3A,$3F,$A2,$BA,$BF,$E9,$73,$EB,$E1,$FE,$CD
	.dc.b	$E7,$FF,$61,$25,$FD,$66,$BF,$DC,$5A,$FF,$75,$6B,$C8,$E0,$43,$FB
	.dc.b	$F6,$9F,$E6,$2F,$89,$FE,$BA,$48,$FE,$B3,$F1,$35,$FE,$89,$17,$1F
	.dc.b	$EC,$33,$70,$FE,$9B,$36,$BF,$A2,$CD,$BE,$D1,$FD,$92,$7F,$B5,$CE
	.dc.b	$73,$45,$8C,$36,$78,$10,$F1,$9E,$63,$18,$7F,$74,$8E,$11,$A2,$46
	.dc.b	$5E,$07,$F5,$FF,$89,$FE,$FB,$F1,$32,$3F,$BC,$C7,$F8,$00,$C3,$FD
	.dc.b	$96,$30,$E8,$FE,$FE,$EA,$7F,$C0,$B3,$A9,$E0,$68,$E0,$78,$4C,$1A
	.dc.b	$30,$48,$FF,$07,$CF,$5C,$BF,$BD,$D7,$59,$CD,$72,$7C,$2F,$67,$2F
	.dc.b	$57,$2F,$59,$FF,$02,$D7,$58,$FE,$FB,$5F,$E2,$0F,$87,$91,$FE,$1C
	.dc.b	$7F,$C4,$7F,$11,$88,$FF,$04,$45,$C3,$FC,$4B,$37,$0F,$F0,$8C,$DA
	.dc.b	$FF,$04,$CD,$D3,$FD,$67,$37,$44,$D8,$C3,$6B,$C4,$7F,$8B,$9C,$3E
	.dc.b	$23,$06,$8C,$13,$0F,$2B,$97,$FC,$77,$3D,$A3,$FC,$73,$F1,$1F,$E4
	.dc.b	$0F,$88,$91,$82,$1F,$E2,$77,$2F,$F8,$A6,$3F,$C8,$E6,$2B,$FC,$8F
	.dc.b	$38,$9E,$31,$87,$FC,$8F,$19,$79,$18,$21,$82,$47,$17,$C4,$74,$CB
	.dc.b	$FC,$7F,$5E,$17,$FC,$83,$5D,$6B,$FC,$B3,$39,$AF,$F1,$DC,$E6,$B8
	.dc.b	$C7,$82,$BB,$1C,$1F,$F4,$5A,$8F,$F4,$6E,$4F,$D1,$BF,$D2,$D3,$9A
	.dc.b	$FF,$30,$C3,$E2,$3F,$8F,$46,$1B,$5C,$6E,$D7,$27,$AB,$C6,$BC,$31
	.dc.b	$8A,$DB,$23,$27,$52,$D7,$96,$75,$0B,$F3,$4F,$86,$37,$7D,$DD,$5F
	.dc.b	$77,$2F,$2A,$E3,$3B,$6D,$E6,$7D,$27,$B1,$95,$C8,$87,$63,$D0,$5B
	.dc.b	$F4,$3A,$AE,$F5,$CA,$35,$7E,$18,$EF,$1B,$7D,$0E,$AF,$C8,$FD,$4F
	.dc.b	$84,$F9,$E4,$7E,$0F,$C1,$57,$AC,$5C,$3D,$F3,$6F,$EA,$CD,$AF,$BF
	.dc.b	$37,$1F,$C2,$4D,$C5,$36,$30,$DA,$FA,$31,$F8,$28,$C4,$78,$8F,$29
	.dc.b	$A2,$19,$3E,$1B,$C3,$0F,$AB,$3C,$E7,$EE,$F8,$FF,$8E,$EA,$91,$82
	.dc.b	$1F,$31,$CB,$F3,$A3,$F2,$A6,$2B,$F0,$33,$89,$E3,$18,$7F,$17,$19
	.dc.b	$79,$18,$21,$82,$47,$17,$C4,$74,$CB,$F4,$75,$E1,$7F,$4B,$5D,$6B
	.dc.b	$F2,$73,$9A,$FD,$8C,$E6,$B8,$C7,$82,$BB,$1C,$1F,$D0,$F4,$7E,$C8
	.dc.b	$8F,$E9,$3E,$22,$BF,$90,$8B,$3F,$A2,$9B,$5F,$B8,$9B,$5F,$C0,$A6
	.dc.b	$E9,$FC,$1E,$6E,$89,$B1,$86,$D7,$90,$F1,$14,$F1,$17,$0F,$13,$70
	.dc.b	$F1,$34,$F1,$17,$0F,$13,$70,$F1,$3F,$C0,$BE,$23,$F8,$9F,$C4,$48
	.dc.b	$C1,$0F,$D8,$B9,$7F,$67,$1F,$C6,$8C,$61,$FC,$1A,$71,$3C,$63,$0F
	.dc.b	$F1,$08,$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$1F,$CA,$2B,$C2,$FF
	.dc.b	$2A,$AE,$B5,$FC,$62,$73,$5F,$CD,$A7,$35,$C6,$3C,$15,$D8,$E0,$FF
	.dc.b	$26,$F8,$8F,$E6,$FF,$0F,$23,$04,$3F,$8A,$B9,$7F,$8B,$C7,$F3,$E3
	.dc.b	$13,$FC,$AA,$71,$3C,$63,$0F,$F3,$48,$CB,$C8,$C1,$0C,$12,$38,$BE
	.dc.b	$23,$A5,$7F,$4C,$AE,$73,$FC,$E9,$FE,$A3,$F1,$18,$9D,$57,$03,$FA
	.dc.b	$DC,$65,$FE,$BB,$37,$3F,$EC,$53,$FD,$6B,$E1,$FF,$AD,$FC,$44,$BF
	.dc.b	$E9,$B1,$97,$FA,$7C,$65,$FE,$77,$19,$7F,$9E,$C6,$71,$FE,$9F,$39
	.dc.b	$7F,$A8,$C6,$71,$FE,$CF,$39,$7F,$B4,$C6,$6B,$FB,$9C,$E5,$FE,$EB
	.dc.b	$19,$C3,$FB,$D4,$E5,$B3,$E8,$3F,$B2,$FC,$3E,$76,$55,$8C,$36,$7F
	.dc.b	$B5,$FC,$47,$F7,$CF,$88,$91,$82,$1F,$D6,$9C,$BF,$D7,$23,$FC,$06
	.dc.b	$63,$3F,$ED,$F3,$89,$E3,$18,$7F,$BC,$C6,$5E,$46,$08,$60,$91,$C5
	.dc.b	$F1,$1D,$33,$FF,$08,$D7,$39,$FE,$FE,$FF,$85,$BE,$23,$13,$AA,$E0
	.dc.b	$7F,$88,$23,$2F,$F8,$8A,$6E,$7F,$E2,$79,$FF,$0F,$FC,$47,$F8,$83
	.dc.b	$C3,$97,$FE,$12,$8C,$BF,$E1,$48,$CB,$FE,$00,$8C,$BF,$E0,$28,$CE
	.dc.b	$3F,$E1,$49,$CB,$FE,$16,$8C,$E3,$FE,$2C,$9C,$BF,$E2,$E8,$CD,$7F
	.dc.b	$8E,$27,$2F,$F8,$EA,$33,$87,$F8,$FE,$72,$D9,$F4,$1F,$E3,$0F,$D1
	.dc.b	$DF,$0F,$AD,$D7,$C2,$73,$5C,$0F,$61,$72,$FF,$1A,$C6,$A3,$FC,$4C
	.dc.b	$3C,$E8,$FF,$90,$07,$43,$D9,$FF,$27,$C6,$1B,$3C,$23,$9B,$6F,$FC
	.dc.b	$B1,$1B,$3A,$57,$37,$B4,$EF,$54,$48,$C3,$89,$E0,$FA,$1B,$5D,$A7
	.dc.b	$73,$EB,$79,$B0,$E8,$D7,$2F,$A5,$D3,$5C,$BB,$78,$95,$D3,$83,$A3
	.dc.b	$87,$C4,$6D,$E4,$E0,$DB,$F2,$88,$FC,$6F,$6E,$BF,$32,$2C,$FC,$CA
	.dc.b	$56,$2F,$CF,$7F,$32,$FE,$BD,$6A,$3F,$15,$F1,$1F,$E8,$6C,$9F,$11
	.dc.b	$4C,$44,$9F,$C8,$46,$5E,$24,$3F,$CD,$93,$C9,$D9,$9C,$BF,$1F,$5C
	.dc.b	$63,$75,$D5,$E3,$19,$9F,$CA,$CE,$5E,$33,$D0,$E2,$75,$C7,$FD,$4F
	.dc.b	$6B,$E1,$18,$7D,$DE,$E6,$29,$C0,$BE,$11,$CC,$CA,$9E,$13,$BA,$B3
	.dc.b	$9A,$FD,$A7,$B7,$F2,$FE,$22,$46,$08,$7C,$67,$2F,$C8,$8F,$A4,$63
	.dc.b	$2F,$DA,$CE,$27,$8C,$61,$FE,$05,$19,$79,$18,$21,$82,$47,$17,$C4
	.dc.b	$74,$CB,$E4,$D7,$39,$FE,$0A,$FE,$17,$E2,$31,$3A,$AE,$07,$C6,$8C
	.dc.b	$BF,$22,$6E,$5F,$36,$7F,$1D,$F1,$1F,$8F,$F8,$89,$7F,$C6,$E3,$2F
	.dc.b	$F1,$D8,$CB,$F8,$C8,$CB,$F8,$E8,$CE,$3F,$C6,$A7,$2F,$F1,$C8,$CE
	.dc.b	$3F,$84,$9C,$BF,$86,$8C,$D7,$DE,$9C,$BF,$80,$8C,$E1,$F8,$59,$CB
	.dc.b	$67,$D0,$7F,$33,$F8,$8C,$E8,$AB,$18,$6C,$FF,$39,$F8,$7F,$D8,$7C
	.dc.b	$44,$8C,$10,$FD,$5B,$97,$F5,$B1,$FB,$73,$19,$FF,$3A,$9C,$4F,$18
	.dc.b	$C3,$FD,$1E,$32,$F2,$30,$43,$04,$8E,$2F,$88,$E9,$9F,$EB,$AB,$9C
	.dc.b	$FF,$4A,$7F,$85,$7C,$46,$27,$55,$C0,$FD,$5C,$65,$FD,$6C,$DC,$FF
	.dc.b	$65,$3F,$C4,$FE,$23,$F8,$A7,$A3,$2F,$FB,$04,$65,$FE,$C5,$19,$7F
	.dc.b	$88,$C6,$5F,$E2,$71,$9C,$7F,$AF,$CF,$F5,$6F,$47,$F9,$8F,$39,$A4
	.dc.b	$BF,$E7,$8D,$7F,$9D,$B5,$FE,$8A,$D7,$FA,$4B,$5F,$E1,$3F,$0F,$FC
	.dc.b	$B7,$E2,$24,$7F,$79,$8F,$E6,$06,$1F,$E5,$31,$87,$47,$F7,$F7,$58
	.dc.b	$7F,$78,$9F,$ED,$FE,$97,$A8,$95,$FD,$4D,$AF,$F5,$26,$BF,$D7,$1A
	.dc.b	$FF,$60,$6B,$FD,$BD,$FE,$BD,$A7,$F2,$5E,$BF,$E8,$DF,$11,$2B,$FC
	.dc.b	$25,$1F,$D2,$4C,$3F,$D0,$A3,$0E,$8F,$F0,$C3,$AC,$3F,$C2,$33,$FE
	.dc.b	$0A,$FB,$A7,$A7,$FD,$91,$B7,$FD,$BD,$AF,$F7,$76,$BF,$DE,$DA,$FF
	.dc.b	$7E,$6B,$FE,$0A,$6A,$69,$23,$C0,$48,$FF,$02,$61,$FD,$17,$BF,$FB
	.dc.b	$37,$C3,$C9,$FF,$19,$47,$F6,$93,$0F,$F6,$28,$C3,$A3,$FC,$70,$EB
	.dc.b	$1F,$F1,$8C,$FC,$1F,$4F,$E0,$12,$BE,$03,$5F,$80,$D7,$E0,$35,$F8
	.dc.b	$0D,$7E,$03,$F6,$BF,$48,$FF,$10,$53,$9B,$FC,$5B,$8A,$5F,$11,$FE
	.dc.b	$26,$AD,$61,$FD,$9A,$72,$FF,$8D,$23,$35,$C6,$38,$5F,$4A,$EE,$DB
	.dc.b	$CC,$F3,$6D,$71,$8E,$93,$C5,$CD,$78,$15,$F8,$23,$0D,$8F,$C5,$7A
	.dc.b	$DF,$89,$23,$F8,$96,$BF,$89,$6B,$F8,$96,$BF,$89,$6B,$F8,$97,$E5
	.dc.b	$7C,$47,$93,$E2,$24,$FA,$0F,$8D,$1E,$83,$2F,$A3,$0F,$89,$39,$8E
	.dc.b	$18,$FD,$1A,$DE,$1E,$DA,$D5,$7F,$9D,$37,$7B,$1F,$A5,$24,$7E,$95
	.dc.b	$AF,$E9,$5A,$FE,$95,$AF,$E9,$5A,$FE,$95,$FC,$37,$C4,$7D,$0E,$89
	.dc.b	$3E,$6D,$53,$F5,$0C,$C7,$B4,$FC,$34,$E5,$E1,$1A,$C7,$EC,$4E,$67
	.dc.b	$D9,$E6,$FE,$FC,$8F,$EF,$DA,$FE,$FD,$AF,$EF,$DA,$FE,$FD,$AF,$EF
	.dc.b	$DF,$D0,$7C,$47,$E2,$BE,$22,$4F,$AE,$3F,$1A,$61,$F5,$6D,$FA,$39
	.dc.b	$CC,$6C,$F8,$4E,$F0,$FC,$95,$7F,$A8,$BE,$E9,$B7,$F9,$F1,$B7,$FE
	.dc.b	$87,$6B,$FE,$8F,$6B,$FE,$93,$6B,$FE,$96,$6B,$FE,$A2,$6A,$69,$23
	.dc.b	$C0,$48,$FE,$33,$A3,$E2,$7F,$57,$F1,$32,$7E,$E4,$7E,$B8,$C3,$F0
	.dc.b	$A5,$39,$8D,$9F,$83,$77,$8F,$EC,$EB,$E1,$7C,$17,$C1,$25,$7C,$16
	.dc.b	$BF,$05,$AF,$C1,$6B,$F0,$5A,$FC,$17,$F8,$D7,$C3,$FF,$04,$F8,$89
	.dc.b	$3F,$95,$8F,$E0,$E6,$1F,$54,$61,$D1,$F9,$B7,$57,$FC,$2E,$7F,$87
	.dc.b	$FC,$3F,$F1,$0F,$88,$97,$FA,$18,$CB,$FA,$38,$CB,$F1,$E3,$2F,$C9
	.dc.b	$8C,$D7,$F1,$79,$CB,$FC,$66,$33,$CF,$FC,$9E,$72,$FF,$29,$8C,$D7
	.dc.b	$F3,$19,$CB,$FC,$CE,$33,$C3,$FC,$E6,$7F,$93,$FC,$3F,$F2,$8F,$88
	.dc.b	$97,$FB,$98,$CB,$FB,$B8,$CB,$F7,$E3,$2F,$E0,$63,$38,$FF,$2F,$9C
	.dc.b	$BF,$CC,$63,$38,$FF,$3F,$9C,$BF,$D0,$63,$35,$FD,$26,$72,$FF,$4B
	.dc.b	$8C,$E1,$FD,$46,$72,$D9,$F4,$1F,$CB,$4F,$D0,$7C,$3E,$B8,$15,$D2
	.dc.b	$72,$70,$3D,$85,$C7,$FA,$DB,$A3,$24,$7D,$CF,$7E,$25,$F7,$AE,$D1
	.dc.b	$89,$E9,$97,$C1,$69,$FB,$7A,$E0,$E0,$F0,$5F,$67,$C1,$3C,$5F,$E5
	.dc.b	$46,$9C,$46,$C8,$FF,$05,$9E,$B3,$C8,$F0,$1E,$2C,$7E,$53,$93,$A1
	.dc.b	$F4,$CD,$63,$F2,$9F,$81,$3D,$2F,$BC,$74,$3B,$57,$48,$CC,$74,$8C
	.dc.b	$5F,$A0,$91,$C6,$79,$CE,$A3,$C4,$7C,$0C,$3E,$63,$DB,$0F,$23,$E9
	.dc.b	$8C,$D7,$29,$CB,$E2,$8C,$DF,$C5,$AF,$6E,$3F,$CD,$71,$91,$83,$DE
	.dc.b	$43,$C1,$1A,$7C,$91,$F0,$E7,$A4,$60,$E4,$52,$5F,$96,$B9,$3F,$90
	.dc.b	$BF,$2C,$FC,$08,$F4,$5E,$AF,$E3,$E3,$F1,$6B,$EB,$9F,$37,$1F,$7C
	.dc.b	$FC,$C8,$CE,$1D,$2B,$E6,$1C,$5E,$F1,$C7,$1F,$CC,$E1,$DE,$3C,$57
	.dc.b	$CE,$B6,$73,$C3,$BC,$E8,$FB,$53,$F0,$B1,$F7,$47,$C9,$7D,$D7,$F1
	.dc.b	$E3,$67,$E2,$4F,$EB,$CF,$F6,$52,$3F,$D9,$5A,$FF,$65,$6B,$C0,$D1
	.dc.b	$0D,$16,$3C,$2F,$8E,$7F,$CA,$5F,$11,$EF,$8F,$8F,$5F,$E4,$C8,$C6
	.dc.b	$1F,$02,$31,$3C,$4F,$49,$F4,$67,$71,$F2,$72,$FF,$34,$46,$27,$84
	.dc.b	$61,$E4,$78,$0F,$F1,$54,$6D,$FC,$3D,$F9,$9D,$DF,$8D,$D9,$8B,$F0
	.dc.b	$62,$BE,$58,$FC,$0A,$FE,$F1,$18,$C3,$EB,$1D,$63,$B9,$EC,$C3,$EF
	.dc.b	$4E,$E3,$EB,$E3,$E9,$76,$7A,$4E,$A7,$8C,$F7,$E1,$E3,$70,$76,$8E
	.dc.b	$58,$79,$49,$87,$92,$7F,$D1,$97,$E1,$7E,$71,$CA,$76,$E8,$C4,$F1
	.dc.b	$AE,$AF,$38,$F3,$4F,$FA,$16,$BA,$61,$EE,$9E,$27,$BB,$0F,$43,$EE
	.dc.b	$8F,$A0,$FB,$8C,$63,$F4,$2B,$9B,$F8,$68,$C9,$F0,$A3,$2F,$D8,$32
	.dc.b	$7C,$3A,$EF,$7E,$C8,$F3,$E1,$F5,$63,$18,$78,$9F,$54,$FA,$0F,$8A
	.dc.b	$74,$7E,$2E,$3F,$0A,$3E,$E6,$3E,$77,$85,$7C,$17,$D9,$5D,$A7,$D4
	.dc.b	$F6,$8E,$97,$E3,$23,$D0,$F6,$47,$FA,$C7,$83,$FD,$6A,$48,$FF,$43
	.dc.b	$55,$7F,$C9,$51,$67,$FD,$71,$36,$BF,$D8,$D3,$6B,$FD,$9D,$37,$0F
	.dc.b	$F6,$A4,$DC,$7F,$8A,$34,$FF,$62,$CD,$AF,$F4,$B4,$61,$A4,$3C,$67
	.dc.b	$98,$C6,$1E,$B6,$E1,$FE,$D9,$AE,$91,$C0,$FF,$32,$38,$9E,$11,$83
	.dc.b	$89,$FE,$E7,$AF,$44,$EE,$BD,$F3,$E1,$3A,$57,$22,$47,$86,$39,$E1
	.dc.b	$F0,$9B,$87,$C2,$6E,$1E,$F7,$9D,$76,$69,$F3,$0D,$9C,$B0,$F2,$3E
	.dc.b	$73,$C9,$87,$89,$C1,$CB,$06,$1E,$18,$C2,$C3,$C5,$5C,$6B,$D1,$5A
	.dc.b	$BE,$F1,$89,$FC,$1C,$F9,$0D,$15,$E5,$7D,$DB,$7D,$B0,$F7,$46,$2B
	.dc.b	$D3,$18,$23,$EA,$8E,$EF,$A2,$7E,$AB,$DE,$77,$7E,$EC,$3C,$F8,$7B
	.dc.b	$6B,$A9,$CE,$B7,$19,$8F,$C1,$1D,$4F,$9C,$75,$38,$1F,$4C,$D1,$F0
	.dc.b	$0B,$87,$C0,$AF,$0E,$1E,$F9,$E1,$3E,$F7,$BB,$EC,$30,$5B,$FC,$54
	.dc.b	$74,$6C,$66,$3C,$78,$7D,$39,$C9,$23,$38,$7B,$6B,$85,$7C,$59,$F6
	.dc.b	$35,$F1,$D7,$C4,$BF,$44,$E6,$BC,$B3,$D6,$30,$EE,$BE,$6C,$61,$F0
	.dc.b	$D7,$2B,$FC,$15,$7D,$0C,$31,$3F,$6A,$BC,$27,$03,$9D,$70,$7D,$73
	.dc.b	$C1,$FA,$A7,$0B,$F8,$4E,$8E,$35,$D7,$0F,$83,$7F,$0F,$0F,$15,$68
	.dc.b	$E7,$3A,$AE,$13,$EF,$AE,$15,$D4,$F0,$38,$BF,$09,$F0,$48,$E4,$98
	.dc.b	$7A,$67,$F1,$18,$77,$7C,$37,$DF,$0F,$0C,$7E,$22,$3E,$73,$E0,$AF
	.dc.b	$CE,$9C,$EB,$F2,$11,$EB,$3A,$CF,$37,$A9,$CC,$F9,$C7,$33,$F0,$77
	.dc.b	$FE,$49,$AE,$B1,$CE,$BA,$BF,$02,$FC,$B8,$7B,$9B,$7F,$5E,$3E,$1B
	.dc.b	$CE,$7A,$DF,$9F,$0F,$A3,$1C,$CF,$2D,$73,$6D,$73,$3E,$4B,$E3,$9E
	.dc.b	$38,$7C,$98,$F7,$81,$8B,$C7,$C4,$43,$51,$97,$06,$22,$B4,$B1,$B7
	.dc.b	$6D,$7A,$35,$EE,$D7,$91,$C2,$3A,$BE,$66,$C6,$F0,$FE,$6F,$1E,$8A
	.dc.b	$5F,$B6,$7F,$9D,$F8,$B9,$8C,$62,$8C,$62,$AE,$6F,$C5,$BE,$6E,$6B
	.dc.b	$D9,$18,$9E,$B3,$BB,$EB,$7D,$27,$9C,$AB,$85,$F2,$BD,$CF,$13,$B5
	.dc.b	$F1,$9D,$47,$D2,$8F,$82,$47,$E7,$B5,$F9,$ED,$7E,$7B,$5F,$9E,$D7
	.dc.b	$A1,$F7,$2F,$FB,$0F,$C4,$7A,$7E,$07,$E0,$CF,$A7,$0F,$99,$18,$C7
	.dc.b	$D7,$5E,$73,$84,$CA,$F6,$5F,$46,$E2,$AD,$47,$53,$E9,$61,$D6,$FE
	.dc.b	$BC,$FD,$69,$FC,$BB,$F8,$22,$3F,$90,$6B,$F9,$06,$BF,$90,$6B,$F9
	.dc.b	$06,$BF,$9E,$7F,$37,$6A,$78,$BF,$0F,$C3,$97,$EC,$9C,$B6,$72,$FA
	.dc.b	$EB,$95,$7C,$B9,$93,$C6,$37,$97,$E2,$EF,$9C,$FC,$31,$EA,$FD,$27
	.dc.b	$F6,$8D,$7F,$6E,$D7,$F7,$8D,$78,$E5,$FA,$A8,$C1,$1C,$12,$3F,$77
	.dc.b	$19,$7F,$09,$2F,$0F,$F0,$9E,$0C,$9F,$85,$1F,$87,$30,$FC,$19,$4E
	.dc.b	$63,$67,$C9,$77,$87,$C1,$AF,$BD,$7F,$46,$BC,$77,$E9,$23,$82,$1F
	.dc.b	$C5,$23,$83,$63,$6D,$36,$7C,$A7,$EF,$8F,$88,$FD,$07,$B5,$27,$EF
	.dc.b	$47,$E8,$CC,$3E,$D8,$C3,$A3,$F0,$8E,$B0,$FB,$93,$A9,$E0,$68,$E0
	.dc.b	$7C,$77,$86,$5F,$AE,$AE,$15,$CA,$B8,$1C,$8E,$07,$27,$F1,$51,$CA
	.dc.b	$3F,$57,$5C,$AB,$A4,$72,$AE,$8F,$23,$A3,$C9,$E9,$1F,$C0,$27,$A6
	.dc.b	$BF,$A5,$BE,$95,$D8,$E9,$5D,$9E,$97,$D8,$E8,$F6,$8E,$8F,$63,$F8
	.dc.b	$0C,$F6,$C3,$EA,$46,$6F,$C0,$F6,$3C,$07,$6A,$F0,$5F,$63,$C0,$F6
	.dc.b	$8F,$01,$D8,$E9,$5D,$9C,$D7,$67,$A1,$DA,$BA,$57,$63,$A3,$FC,$9B
	.dc.b	$EE,$9E,$4F,$F1,$7F,$5D,$F0,$94,$FE,$35,$3A,$C3,$FA,$00,$70,$FF
	.dc.b	$01,$9E,$2E,$9A,$FF,$05,$8D,$DF,$86,$FB,$4E,$EF,$BC,$48,$C1,$0E
	.dc.b	$4E,$5F,$04,$72,$72,$79,$AB,$BE,$1F,$E0,$6C,$3C,$78,$7F,$46,$6C
	.dc.b	$7F,$82,$A7,$75,$FE,$0F,$7D,$AF,$F8,$35,$AE,$F0,$FE,$FF,$18,$8E
	.dc.b	$11,$82,$3F,$E1,$08,$D3,$F2,$63,$2F,$F8,$5A,$32,$48,$F7,$5F,$C7
	.dc.b	$9F,$99,$F0,$B8,$BF,$14,$7B,$EB,$ED,$4E,$63,$C6,$64,$E2,$7F,$43
	.dc.b	$F8,$9F,$ED,$1F,$13,$23,$04,$3F,$92,$39,$7F,$93,$47,$F6,$F3,$18
	.dc.b	$7F,$43,$9C,$4F,$18,$FE,$8D,$38,$8F,$E5,$DF,$13,$67,$31,$BC,$7F
	.dc.b	$BB,$CE,$5F,$EF,$35,$67,$FB,$7C,$FF,$7F,$C3,$FC,$01,$F1,$32,$7F
	.dc.b	$AF,$46,$5F,$EC,$11,$97,$F9,$F4,$65,$FE,$81,$19,$C3,$FA,$E4,$FF
	.dc.b	$82,$F5,$95,$FD,$A5,$FF,$07,$7C,$46,$23,$FA,$44,$5A,$FE,$C9,$37
	.dc.b	$2F,$F0,$8C,$DA,$FF,$03,$CF,$F8,$5F,$83,$FC,$31,$F1,$32,$7F,$BB
	.dc.b	$46,$5F,$EF,$11,$97,$FA,$B4,$65,$FE,$B1,$19,$C3,$FB,$A4,$FF,$88
	.dc.b	$FE,$23,$FC,$49,$F1,$32,$FF,$C0,$51,$97,$FC,$09,$19,$7F,$B1,$46
	.dc.b	$5F,$EC,$91,$9A,$FF,$00,$4F,$F8,$BF,$CE,$FF,$18,$7C,$34,$BF,$F0
	.dc.b	$74,$65,$FF,$08,$46,$5F,$ED,$D1,$97,$FB,$84,$66,$BF,$C1,$93,$96
	.dc.b	$CF,$43,$FC,$35,$F1,$39,$D7,$FC,$7B,$56,$7F,$C6,$73,$79,$BF,$C4
	.dc.b	$93,$73,$4D,$8C,$36,$BB,$1F,$E0,$37,$0F,$A0,$F0,$9F,$E4,$6F,$D2
	.dc.b	$5F,$15,$FE,$69,$5F,$0F,$4C,$44,$9E,$11,$97,$94,$65,$FF,$19,$4F
	.dc.b	$27,$66,$88,$6A,$79,$3C,$27,$96,$7F,$E7,$FA,$E5,$3B,$38,$C6,$67
	.dc.b	$8B,$BA,$E3,$9F,$FA,$4E,$DF,$13,$ED,$26,$1F,$07,$E2,$6B,$E8,$8B
	.dc.b	$87,$86,$6D,$F3,$9B,$5F,$E9,$9B,$56,$35,$47,$82,$7E,$75,$7C,$DA
	.dc.b	$D4,$E6,$7C,$07,$8C,$D9,$0C,$9F,$32,$3E,$B7,$C4,$C8,$FB,$4F,$D9
	.dc.b	$EF,$C4,$7D,$28,$B9,$7C,$B9,$B9,$7C,$59,$B5,$EB,$9B,$97,$AA,$6E
	.dc.b	$3F,$E6,$4A,$EE,$43,$B1,$F8,$BA,$DC,$F0,$7F,$19,$DB,$27,$F2,$0F
	.dc.b	$E3,$FE,$1B,$11,$F8,$88,$B9,$7E,$06,$6E,$5F,$6E,$6D,$7D,$19,$B9
	.dc.b	$7D,$09,$B9,$26,$C6,$1B,$5E,$23,$F4,$4F,$E8,$7E,$27,$F4,$5F,$11
	.dc.b	$27,$F4,$D1,$97,$F5,$11,$97,$F3,$F1,$97,$F4,$31,$9C,$BF,$35,$39
	.dc.b	$7F,$39,$19,$CB,$F2,$D3,$97,$F3,$11,$9A,$FC,$74,$E5,$FC,$84,$67
	.dc.b	$2F,$C7,$CF,$EE,$7E,$1F,$F7,$5F,$11,$2F,$F7,$D1,$97,$F8,$04,$65
	.dc.b	$FD,$BC,$65,$FD,$CC,$67,$1F,$D9,$4E,$5B,$3D,$0F,$E1,$5F,$11,$9C
	.dc.b	$BF,$61,$56,$7F,$4D,$37,$1F,$D2,$CD,$D5,$3F,$C4,$4C,$45,$AE,$A7
	.dc.b	$73,$CE,$60,$FE,$1F,$1C,$1C,$61,$FC,$27,$E1,$FF,$8E,$7C,$44,$9F
	.dc.b	$E4,$11,$FC,$7C,$C3,$FC,$62,$30,$E8,$FE,$4C,$EA,$7A,$46,$4F,$E3
	.dc.b	$F3,$C1,$FE,$43,$F0,$F6,$77,$1A,$3F,$98,$46,$73,$FE,$2D,$56,$7F
	.dc.b	$86,$4F,$F3,$33,$19,$FF,$0C,$9C,$4F,$23,$C5,$7D,$8F,$E5,$15,$A8
	.dc.b	$F0,$12,$32,$FF,$41,$F8,$8F,$E8,$04,$7F,$A2,$7C,$3D,$7F,$9D,$C5
	.dc.b	$9F,$E6,$93,$71,$FE,$5D,$36,$BF,$92,$4D,$C7,$F9,$1C,$DD,$53,$63
	.dc.b	$0D,$AF,$21,$E2,$3F,$AB,$7C,$47,$F5,$5F,$88,$91,$82,$1F,$D4,$1C
	.dc.b	$BF,$D4,$A3,$FA,$E1,$8C,$BF,$A6,$4E,$27,$8C,$61,$FE,$C9,$19,$79
	.dc.b	$18,$21,$82,$47,$17,$C4,$74,$CB,$FA,$97,$92,$89,$29,$C6,$6F,$D2
	.dc.b	$B5,$5A,$9C,$C6,$23,$FB,$6D,$FF,$77,$37,$7C,$E1,$7C,$DC,$DF,$27
	.dc.b	$81,$FE,$01,$AC,$BC,$2F,$B5,$F3,$BE,$53,$B2,$08,$D4,$F0,$BE,$D3
	.dc.b	$A9,$FE,$F5,$7D,$AF,$9C,$F1,$AD,$BE,$C3,$83,$8A,$EB,$1E,$83,$81
	.dc.b	$E7,$F8,$9E,$98,$3E,$26,$98,$89,$3F,$2C,$F7,$46,$9C,$61,$F4,$67
	.dc.b	$71,$DA,$3E,$A4,$EE,$31,$4A,$DD,$FC,$2A,$D3,$F1,$23,$38,$7D,$A9
	.dc.b	$CB,$F7,$23,$37,$E4,$3C,$77,$EB,$6C,$7D,$65,$F5,$0E,$6F,$A5,$0F
	.dc.b	$C5,$DF,$0F,$F8,$A3,$94,$65,$C6,$5F,$8E,$9D,$CF,$E5,$6F,$A6,$7F
	.dc.b	$97,$9C,$CF,$83,$0E,$F1,$8C,$3A,$B7,$0E,$91,$F9,$AC,$39,$DF,$84
	.dc.b	$ED,$7C,$A7,$C7,$1D,$DA,$EC,$F0,$61,$97,$8B,$5F,$7D,$F8,$1F,$D7
	.dc.b	$4F,$90,$F1,$57,$78,$E2,$75,$8D,$3C,$27,$93,$E0,$3C,$57,$EA,$C3
	.dc.b	$DD,$1E,$1A,$E2,$EE,$3E,$3C,$EE,$7D,$71,$C2,$79,$DF,$8D,$E1,$3A
	.dc.b	$23,$E6,$9E,$73,$B9,$EE,$78,$2B,$B5,$6E,$38,$BF,$10,$F8,$D1,$A7
	.dc.b	$B9,$EF,$3D,$57,$F0,$23,$46,$E7,$BC,$66,$BD,$13,$98,$F8,$F3,$E3
	.dc.b	$9C,$9E,$78,$C9,$C7,$0F,$1C,$66,$FC,$26,$22,$3F,$56,$78,$B8,$C3
	.dc.b	$E1,$CF,$08,$D1,$9B,$FB,$F1,$99,$FA,$07,$09,$FA,$31,$99,$F8,$F8
	.dc.b	$7C,$0C,$3C,$95,$D0,$D9,$2F,$9D,$70,$9F,$7D,$F3,$AF,$98,$6E,$7C
	.dc.b	$6F,$3A,$E2,$FD,$72,$1F,$5C,$D4,$70,$8F,$41,$C2,$34,$E2,$34,$7E
	.dc.b	$00,$FA,$53,$F0,$E7,$C7,$3C,$6F,$DD,$3E,$79,$E7,$5C,$9F,$1B,$F8
	.dc.b	$07,$91,$F8,$83,$67,$47,$83,$97,$83,$F0,$4E,$78,$7D,$A7,$06,$DC
	.dc.b	$5F,$AE,$36,$79,$6F,$CE,$FC,$43,$C0,$7C,$87,$F2,$4E,$F0,$FA,$6E
	.dc.b	$CF,$39,$E9,$8D,$B9,$7A,$CF,$CD,$BF,$93,$3E,$67,$D7,$1E,$58,$F4
	.dc.b	$1D,$F0,$F8,$B3,$A3,$E1,$D7,$6A,$FA,$17,$E4,$7B,$61,$FA,$3C,$3B
	.dc.b	$3F,$0A,$F8,$D7,$E4,$AB,$8E,$1D,$AB,$57,$F0,$6B,$57,$F9,$3A,$D4
	.dc.b	$F9,$4E,$A5,$BF,$54,$F9,$E3,$94,$FE,$0E,$3E,$0C,$EA,$FB,$D7,$C9
	.dc.b	$BE,$EF,$D2,$31,$7C,$67,$D9,$5C,$67,$C1,$7F,$86,$3C,$0F,$03,$EB
	.dc.b	$3E,$03,$53,$E0,$C3,$EA,$C7,$E7,$63,$AB,$A9,$EA,$7E,$20,$FB,$98
	.dc.b	$7C,$F8,$E1,$1F,$9A,$AF,$A9,$1F,$1E,$3E,$29,$C3,$0F,$92,$D3,$38
	.dc.b	$7D,$26,$CF,$18,$F7,$46,$DA,$6C,$B1,$B3,$D1,$1B,$6C,$6D,$A6,$E3
	.dc.b	$D9,$19,$6C,$65,$FC,$99,$CC,$C9,$F5,$6B,$E4,$BF,$5E,$3E,$73,$A3
	.dc.b	$F4,$B1,$A7,$F4,$F1,$A7,$AC,$F8,$5B,$1E,$19,$F9,$71,$DE,$FE,$83
	.dc.b	$DF,$0F,$AE,$F7,$AE,$95,$C2,$3C,$35,$C1,$F0,$D7,$07,$B9,$D7,$0F
	.dc.b	$AB,$87,$8E,$F9,$61,$D3,$0F,$CA,$DF,$4B,$DC,$F9,$E7,$2E,$E3,$F3
	.dc.b	$13,$F8,$59,$CE,$1E,$1B,$EA,$D8,$CC,$EE,$B8,$BF,$C7,$2B,$6E,$4F
	.dc.b	$CE,$D7,$19,$DE,$1F,$C2,$EB,$75,$C6,$FA,$B6,$B7,$87,$F0,$46,$C7
	.dc.b	$EE,$0E,$A7,$7B,$FE,$1A,$D9,$FD,$0B,$CC,$D9,$63,$6D,$8D,$C7,$96
	.dc.b	$BC,$6D,$AF,$0D,$7E,$E6,$38,$99,$7A,$B4,$CD,$7A,$63,$FD,$34,$7F
	.dc.b	$8B,$4F,$F4,$E5,$FF,$91,$9B,$7F,$E5,$C6,$DF,$07,$81,$FD,$B0,$C9
	.dc.b	$D2,$79,$39,$21,$FE,$40,$3B,$F4,$BE,$C7,$A4,$A3,$13,$CD,$AF,$09
	.dc.b	$DC,$62,$3F,$CB,$CD,$7F,$CD,$4D,$7F,$CE,$ED,$9E,$D3,$C8,$93,$DD
	.dc.b	$FF,$5F,$7D,$B8,$80

;--------------------------------------------------------------------------------
;	FDSUB.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fdsub_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FDSUB.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FDSUB.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fdsub_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdsub.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDSUB.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fdsub.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FDSUB.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fdsub_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3D,$61,$F3,$7F,$DA,$BD,$E7,$99,$C6,$30,$DA,$FE
	.dc.b	$50,$D9,$DC,$7F,$72,$F8,$17,$BF,$FC,$8E,$3F,$B9,$88,$CF,$C6,$E2
	.dc.b	$31,$E6,$FF,$42,$7F,$BC,$CE,$9E,$2D,$73,$5D,$5D,$CF,$17,$F8,$A5
	.dc.b	$75,$AE,$11,$83,$9C,$ED,$F4,$57,$37,$47,$6A,$E5,$3B,$8F,$54,$ED
	.dc.b	$E3,$3D,$E3,$11,$EC,$F9,$3C,$4C,$8F,$5C,$EA,$D7,$D5,$F3,$CF,$43
	.dc.b	$75,$E2,$BE,$51,$E5,$32,$7B,$1C,$BE,$A2,$1F,$56,$BB,$5F,$84,$95
	.dc.b	$D6,$79,$9D,$48,$7F,$40,$AE,$AF,$94,$C9,$D5,$CB,$F1,$A2,$DF,$BA
	.dc.b	$6D,$78,$2B,$D0,$78,$0F,$39,$C4,$C9,$D8,$C1,$67,$A3,$C4,$F2,$1C
	.dc.b	$9C,$BB,$8F,$03,$F8,$89,$F1,$D7,$9A,$FE,$2B,$89,$F0,$9C,$EB,$8B
	.dc.b	$F0,$CF,$01,$E4,$3C,$13,$C9,$CD,$FD,$C7,$F0,$86,$27,$9B,$F2,$4C
	.dc.b	$57,$8E,$FC,$51,$EF,$BE,$8F,$AE,$7E,$C3,$4E,$F1,$A7,$D1,$1A,$69
	.dc.b	$A9,$EA,$FA,$1A,$FC,$A3,$C8,$75,$3C,$4E,$9F,$80,$68,$EA,$68,$F5
	.dc.b	$9D,$CA,$64,$E8,$78,$4C,$94,$C9,$E3,$3E,$49,$83,$C2,$60,$F0,$9C
	.dc.b	$4A,$71,$BF,$C1,$B8,$38,$B8,$75,$7F,$51,$E8,$62,$BE,$A3,$E6,$3A
	.dc.b	$5F,$CF,$74,$7E,$20,$A7,$BC,$F0,$57,$E0,$CF,$03,$E2,$AF,$04,$60
	.dc.b	$D3,$89,$ED,$3B,$AF,$B8,$F6,$3E,$01,$4F,$2D,$FD,$7B,$F0,$57,$27
	.dc.b	$AD,$72,$AE,$87,$D7,$3E,$74,$70,$AE,$57,$FC,$66,$F9,$3E,$58,$CD
	.dc.b	$F3,$AE,$B7,$CC,$F8,$35,$C6,$B7,$1F,$1C,$F0,$9F,$1C,$F5,$1F,$5A
	.dc.b	$FB,$DF,$D2,$8F,$2D,$73,$7E,$A1,$BA,$EA,$FE,$FC,$CD,$FA,$DB,$3C
	.dc.b	$5E,$75,$E9,$8F,$A6,$66,$3D,$31,$93,$E6,$5F,$92,$3D,$C7,$A2,$BB
	.dc.b	$3D,$0F,$73,$F1,$1F,$71,$F0,$0F,$B2,$53,$EC,$9E,$69,$F0,$3E,$A9
	.dc.b	$F0,$3C,$A7,$A3,$C9,$AF,$23,$C4,$7C,$32,$9F,$C6,$CA,$7E,$EC,$F5
	.dc.b	$57,$C1,$6D,$78,$9F,$C2,$D7,$89,$F0,$4F,$C8,$9D,$47,$E8,$27,$4F
	.dc.b	$8C,$F9,$67,$C4,$9E,$0D,$9F,$9B,$7F,$6A,$3C,$C7,$03,$D4,$53,$27
	.dc.b	$98,$E4,$7C,$77,$93,$ED,$3E,$71,$F8,$73,$E7,$1C,$0F,$6B,$C1,$FE
	.dc.b	$06,$70,$BF,$BA,$F8,$4E,$8F,$D0,$7A,$1F,$30,$E2,$E4,$E2,$F9,$CF
	.dc.b	$59,$4F,$59,$DE,$BE,$93,$6B,$E6,$BF,$8F,$BF,$9A,$F8,$0F,$21,$E0
	.dc.b	$9F,$9A,$F6,$39,$1E,$F3,$EB,$5F,$BE,$FF,$89,$4F,$09,$F1,$C7,$A6
	.dc.b	$3C,$6D,$BF,$1C,$FC,$5B,$F1,$CF,$BA,$3B,$9E,$17,$83,$E1,$3D,$57
	.dc.b	$E1,$9E,$37,$D8,$F1,$4F,$CE,$6B,$ED,$BF,$C3,$BF,$68,$ED,$1A,$BF
	.dc.b	$5C,$EA,$7E,$A4,$61,$FC,$31,$C4,$F7,$CF,$1A,$F1,$D7,$BA,$BE,$D4
	.dc.b	$62,$BD,$47,$C1,$BF,$51,$1E,$87,$A6,$BC,$4F,$D6,$AF,$13,$F9,$03
	.dc.b	$67,$B4,$F2,$9E,$8A,$F0,$3E,$E9,$E6,$FE,$24,$EC,$70,$3B,$47,$07
	.dc.b	$DB,$1C,$1F,$AA,$60,$E0,$73,$3C,$40,$FF,$41,$63,$1F,$F4,$66,$5F
	.dc.b	$E8,$86,$9F,$E8,$82,$9C,$4A,$7F,$A3,$4C,$1F,$E9,$CC,$FF,$D2,$AF
	.dc.b	$FA,$2B,$64,$63,$4F,$F4,$8C,$EE,$BC,$0D,$3C,$05,$8D,$BE,$9F,$95
	.dc.b	$FF,$49,$34,$AE,$13,$48,$E8,$87,$FA,$6D,$CB,$4C,$9F,$EA,$27,$2D
	.dc.b	$32,$7C,$07,$2D,$32,$7C,$77,$2D,$32,$7A,$4F,$01,$0E,$C5,$3A,$14
	.dc.b	$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08
	.dc.b	$7D,$03,$DA,$53,$CC,$53,$C2,$53,$B1,$4E,$A5,$3A,$BF,$29,$A7,$67
	.dc.b	$0E,$C8,$E0,$8E,$C8,$79,$1C,$B4,$C9,$E6,$72,$D3,$27,$A1,$CB,$4C
	.dc.b	$9E,$A7,$2D,$32,$7A,$8F,$01,$0E,$C5,$3A,$14,$E4,$53,$91,$C0,$87
	.dc.b	$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$7D,$03,$DA,$53,$CC
	.dc.b	$53,$C2,$53,$B1,$4E,$A5,$3A,$BF,$29,$A7,$67,$0E,$C8,$E0,$8E,$C8
	.dc.b	$79,$1C,$B4,$C9,$E6,$72,$D3,$27,$A1,$CB,$4C,$9E,$A7,$2D,$32,$7A
	.dc.b	$8F,$01,$0E,$C5,$3A,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34
	.dc.b	$43,$45,$34,$60,$86,$08,$7D,$03,$DA,$53,$CC,$53,$C2,$53,$B1,$4E
	.dc.b	$A5,$3A,$BF,$29,$A7,$67,$0E,$C8,$E0,$8E,$C8,$79,$1C,$B4,$C9,$E6
	.dc.b	$72,$D3,$27,$A1,$CB,$4C,$9E,$A7,$2D,$32,$7A,$8F,$01,$0E,$C5,$3A
	.dc.b	$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86
	.dc.b	$08,$7D,$03,$DA,$53,$CC,$53,$C2,$53,$B1,$4E,$A5,$3A,$BF,$29,$A7
	.dc.b	$67,$0E,$C8,$E0,$8E,$C8,$7C,$BB,$FF,$0A,$D7,$F8,$6A,$3F,$C3,$A4
	.dc.b	$3F,$C2,$87,$F8,$7A,$1B,$BE,$6F,$84,$A7,$31,$86,$C7,$F8,$8E,$78
	.dc.b	$3F,$E3,$61,$AB,$76,$BF,$E1,$8F,$9E,$71,$FF,$8C,$72,$7B,$EF,$86
	.dc.b	$FF,$1E,$56,$B2,$FF,$09,$35,$F3,$4E,$D4,$65,$FF,$15,$7D,$06,$EF
	.dc.b	$AD,$FF,$91,$E3,$53,$FD,$F5,$D5,$F3,$73,$BE,$F1,$F3,$4E,$5F,$F1
	.dc.b	$44,$F0,$BD,$5F,$9A,$77,$1E,$A7,$FC,$5D,$8E,$E7,$CF,$1C,$8D,$AA
	.dc.b	$F0,$B6,$33,$18,$39,$15,$C9,$CE,$3F,$CC,$98,$FC,$07,$91,$FE,$3B
	.dc.b	$9E,$33,$B8,$F9,$77,$EA,$7F,$C7,$93,$A7,$84,$F6,$69,$DB,$0F,$88
	.dc.b	$E0,$E6,$78,$0E,$71,$C5,$A7,$1A,$F0,$3F,$62,$34,$D3,$47,$33,$07
	.dc.b	$6A,$F8,$4E,$9A,$E6,$3D,$D7,$F0,$E7,$2D,$8C,$DF,$AA,$BC,$A7,$13
	.dc.b	$D9,$3B,$6C,$6D,$F1,$C6,$5E,$2D,$73,$1B,$7F,$21,$3E,$E6,$9E,$97
	.dc.b	$C7,$7F,$12,$32,$78,$CE,$B3,$CF,$1F,$B1,$7D,$E7,$6D,$8C,$CE,$E7
	.dc.b	$EA,$4E,$5B,$19,$C3,$F0,$11,$DA,$FC,$57,$C9,$B1,$C1,$B3,$C2,$3D
	.dc.b	$F1,$8C,$3D,$D5,$C6,$7E,$D1,$E6,$BF,$04,$EA,$3A,$1A,$9E,$17,$F6
	.dc.b	$2B,$73,$F1,$CF,$A7,$7D,$DC,$1E,$93,$91,$D2,$79,$46,$E3,$E6,$1E
	.dc.b	$83,$71,$EB,$7C,$D1,$BC,$3E,$44,$E6,$FC,$B8,$7D,$8B,$E2,$FA,$E7
	.dc.b	$51,$E1,$8D,$47,$8A,$36,$7A,$0F,$13,$92,$47,$02,$47,$C0,$38,$9E
	.dc.b	$A3,$CE,$62,$BC,$F1,$E3,$24,$77,$37,$3D,$9E,$8D,$7E,$53,$5F,$2C
	.dc.b	$78,$9C,$B5,$C9,$C9,$C3,$D9,$B5,$F2,$23,$44,$39,$94,$E6,$60,$86
	.dc.b	$08,$75,$38,$14,$D9,$4D,$BE,$17,$2D,$72,$7D,$73,$9B,$97,$91,$1C
	.dc.b	$11,$E0,$74,$29,$D2,$7F,$06,$49,$FA,$71,$F6,$67,$94,$E6,$3E,$B4
	.dc.b	$62,$77,$5F,$72,$72,$76,$3C,$85,$9E,$0E,$AB,$A3,$DC,$D1,$1C,$CF
	.dc.b	$58,$D4,$63,$0F,$D8,$C6,$2F,$C7,$18,$AE,$04,$70,$4C,$3F,$54,$DA
	.dc.b	$FC,$B3,$6F,$F1,$D1,$EA,$7C,$13,$C0,$91,$C1,$B1,$C3,$0F,$7C,$F9
	.dc.b	$4F,$81,$87,$9A,$B6,$FE,$4A,$3E,$03,$97,$A1,$2B,$A3,$A3,$99,$0E
	.dc.b	$C4,$C3,$E0,$57,$68,$FC,$49,$C4,$D6,$1E,$FB,$F1,$5F,$BA,$7C,$11
	.dc.b	$83,$C9,$1E,$23,$45,$72,$6E,$3C,$97,$E2,$79,$9E,$29,$E3,$3B,$8F
	.dc.b	$1D,$F9,$DF,$CC,$4E,$9E,$13,$D9,$A7,$6B,$F5,$38,$39,$9E,$03,$9C
	.dc.b	$71,$69,$C6,$BC,$0F,$E9,$63,$4D,$34,$73,$30,$76,$AF,$84,$E9,$AE
	.dc.b	$63,$DD,$7F,$0E,$72,$D8,$CD,$FA,$AB,$CA,$71,$3D,$93,$B6,$C6,$DF
	.dc.b	$1C,$65,$E2,$D7,$31,$B7,$F3,$93,$EE,$69,$E9,$7C,$77,$F1,$23,$27
	.dc.b	$8C,$EB,$3C,$F0,$FC,$FD,$F7,$9D,$B6,$33,$3B,$9F,$C9,$4E,$5B,$19
	.dc.b	$C3,$F0,$11,$DA,$FC,$57,$C9,$B1,$C1,$AF,$FB,$EF,$6F,$F4,$8B,$FE
	.dc.b	$91,$6B,$FE,$8F,$E1,$70,$7F,$A2,$DB,$1F,$E8,$BC,$BF,$D1,$DC,$1F
	.dc.b	$E8,$C8,$FF,$4B,$35,$12,$FF,$D2,$6E,$A7,$FC,$E0,$EB,$47,$C8,$70
	.dc.b	$8F,$F3,$89,$97,$1B,$7F,$A3,$6F,$FD,$3A,$FF,$A3,$F8,$7F,$D3,$5C
	.dc.b	$1F,$EA,$46,$DF,$FA,$B5,$72,$39,$75,$C7,$FE,$A0,$AF,$F5,$59,$F5
	.dc.b	$38,$FE,$76,$9F,$81,$E2,$F9,$93,$F9,$D3,$5C,$5F,$3B,$8F,$E9,$C7
	.dc.b	$E2,$A3,$15,$C6,$3E,$91,$C6,$B9,$D7,$E7,$AF,$BC,$FE,$34,$FC,$CD
	.dc.b	$F2,$70,$7D,$F3,$91,$D2,$79,$46,$E3,$F2,$47,$8C,$DC,$79,$9F,$49
	.dc.b	$8B,$F3,$46,$6F,$CB,$1F,$9E,$9D,$1E,$23,$A4,$F8,$1C,$9E,$03,$AB
	.dc.b	$92,$47,$36,$C7,$B8,$E3,$C1,$F9,$F8,$F1,$47,$B6,$78,$11,$E1,$1C
	.dc.b	$9F,$88,$D7,$CB,$1B,$72,$D7,$27,$27,$0F,$A1,$B5,$F0,$A3,$44,$39
	.dc.b	$94,$E6,$60,$86,$08,$75,$38,$14,$D9,$4D,$BE,$17,$2D,$72,$7C,$53
	.dc.b	$9B,$97,$91,$1C,$11,$E0,$74,$29,$D2,$7E,$89,$27,$F0,$A4,$8F,$C1
	.dc.b	$1D,$4E,$25,$8E,$0E,$A7,$C1,$18,$8F,$C3,$46,$27,$A4,$60,$E7,$19
	.dc.b	$79,$18,$24,$F3,$AF,$C4,$57,$26,$CF,$09,$D6,$1F,$AE,$8E,$D1,$8C
	.dc.b	$3F,$51,$5C,$27,$37,$CE,$F8,$4E,$48,$E0,$95,$C9,$B5,$F9,$C6,$DF
	.dc.b	$E5,$63,$E3,$5F,$5A,$E0,$48,$E0,$D9,$E1,$87,$C3,$9F,$29,$ED,$AE
	.dc.b	$4F,$02,$4F,$07,$E3,$C6,$A3,$07,$C1,$8D,$C7,$E8,$1D,$9D,$48,$78
	.dc.b	$09,$87,$C2,$BF,$34,$F1,$AE,$87,$67,$FD,$91,$5F,$D4,$78,$BF,$A5
	.dc.b	$BF,$D3,$F8,$FF,$A7,$11,$FE,$9C,$43,$FA,$41,$4F,$E8,$DC,$2E,$36
	.dc.b	$5F,$D4,$31,$63,$FD,$2C,$FE,$9B,$79,$71,$C7,$FD,$3E,$7F,$A6,$5F
	.dc.b	$2A,$46,$2F,$84,$7F,$53,$29,$DA,$B9,$5F,$92,$F9,$5F,$F5,$5D,$BF
	.dc.b	$A7,$5F,$F5,$D1,$C3,$FD,$59,$C3,$FD,$5B,$83,$FA,$AF,$17,$F5,$49
	.dc.b	$70,$7F,$68,$7D,$FC,$2E,$1C,$47,$F7,$09,$D6,$CA,$D4,$FF,$6C,$3F
	.dc.b	$B7,$4E,$A3,$3C,$7F,$22,$7C,$D3,$9B,$EB,$1F,$98,$3F,$07,$7F,$81
	.dc.b	$7F,$12,$7F,$12,$3F,$12,$7E,$1A,$FF,$43,$B7,$E1,$4F,$C9,$16,$39
	.dc.b	$C6,$9E,$5A,$FE,$2A,$75,$1C,$EF,$F1,$EE,$A3,$9B,$4E,$71,$A7,$F8
	.dc.b	$F4,$F3,$74,$73,$3D,$26,$A7,$9B,$4E,$66,$8F,$E4,$CE,$9A,$6A,$3F
	.dc.b	$40,$D8,$F4,$DF,$C7,$9C,$CF,$85,$B1,$99,$DD,$FC,$3A,$F1,$1B,$8C
	.dc.b	$DF,$B6,$3C,$26,$4F,$6D,$73,$6D,$73,$77,$19,$7A,$B5,$CC,$6D,$FC
	.dc.b	$01,$D4,$FC,$49,$F8,$B8,$FC,$74,$FE,$1B,$0F,$C4,$4F,$E3,$CF,$CA
	.dc.b	$1B,$AF,$C7,$DF,$E8,$E3,$F3,$A6,$6B,$93,$F8,$AB,$F2,$3E,$26,$DF
	.dc.b	$F2,$36,$C6,$EF,$D1,$39,$6C,$66,$BE,$43,$CE,$33,$87,$E1,$63,$E1
	.dc.b	$C6,$30,$FB,$D1,$BA,$E3,$7D,$5B,$1D,$4F,$C3,$1B,$2C,$6D,$A6,$EF
	.dc.b	$F0,$37,$DE,$3F,$10,$7D,$8A,$EB,$3E,$58,$FA,$B1,$8B,$F2,$DF,$5A
	.dc.b	$E7,$39,$AE,$67,$B1,$C3,$F6,$CE,$67,$92,$3B,$3E,$97,$9B,$9A,$EE
	.dc.b	$FA,$5F,$C9,$B9,$3B,$9D,$23,$89,$23,$6D,$8F,$E1,$E7,$E0,$4F,$C4
	.dc.b	$1F,$84,$C7,$F8,$A3,$FC,$82,$BF,$0B,$3E,$6B,$FE,$67,$8F,$E4,$AB
	.dc.b	$F9,$CE,$1F,$8D,$9F,$E5,$44,$8F,$F3,$FB,$FD,$3D,$FE,$A4,$7F,$A8
	.dc.b	$F8,$BF,$AA,$BF,$D3,$9A,$FF,$A6,$F7,$FF,$4F,$36,$FF,$D4,$64,$7F
	.dc.b	$AC,$E1,$E8,$C3,$FA,$8C,$79,$0F,$E9,$45,$1C,$3F,$D6,$EB,$FA,$6B
	.dc.b	$63,$53,$CA,$BA,$70,$FF,$50,$9F,$11,$CE,$BF,$D4,$B5,$FD,$7A,$3F
	.dc.b	$A4,$3E,$2D,$3F,$AE,$47,$F4,$3E,$0F,$EA,$4F,$FA,$C5,$AF,$FA,$E9
	.dc.b	$FE,$A8,$FF,$AA,$F8,$7F,$AD,$BF,$34,$FB,$13,$F5,$9F,$DF,$BF,$63
	.dc.b	$2F,$A8,$FD,$F8,$FC,$91,$F8,$73,$F2,$47,$E3,$EF,$F0,$A4,$7F,$44
	.dc.b	$6E,$7F,$7F,$97,$E0,$63,$F2,$46,$9E,$CF,$E4,$67,$93,$E0,$8C,$BD
	.dc.b	$CE,$11,$A7,$81,$82,$1C,$5D,$3C,$4C,$1A,$38,$98,$21,$82,$1C,$4B
	.dc.b	$AF,$EE,$63,$AB,$63,$A1,$B2,$39,$7A,$98,$3E,$19,$87,$4F,$63,$25
	.dc.b	$9F,$94,$49,$FD,$EB,$63,$F7,$67,$57,$0F,$D8,$9E,$AE,$67,$89,$1E
	.dc.b	$31,$A2,$3E,$23,$81,$A3,$0E,$A7,$BD,$7E,$02,$BA,$57,$84,$D9,$27
	.dc.b	$7B,$FF,$2B,$8D,$46,$37,$FE,$43,$7C,$E7,$C1,$39,$7F,$1D,$5C,$67
	.dc.b	$51,$CA,$FC,$31,$82,$4E,$C9,$7D,$9B,$5D,$23,$04,$70,$4B,$FC,$AE
	.dc.b	$1F,$07,$0E,$F7,$CA,$73,$5E,$F9,$CD,$F3,$9F,$F4,$04,$FF,$9B,$F8
	.dc.b	$BF,$8F,$3F,$C6,$DA,$FF,$08,$3F,$CD,$A4,$3F,$CB,$58,$FF,$1B,$CB
	.dc.b	$F8,$7C,$7F,$0A,$D8,$FF,$94,$F6,$FE,$1E,$FF,$12,$C3,$6F,$F1,$89
	.dc.b	$FE,$2C,$FF,$1E,$DF,$FC,$CB,$9F,$68,$FF,$2A,$3F,$C5,$4F,$E3,$0F
	.dc.b	$F9,$5E,$FA,$BC,$71,$F1,$30,$D5,$7F,$9E,$71,$FF,$38,$70,$FF,$9A
	.dc.b	$A7,$FC,$1D,$C1,$FC,$69,$FF,$37,$F0,$7F,$1A,$7F,$5B,$C1,$F3,$DF
	.dc.b	$C1,$E7,$F3,$1F,$A8,$FE,$D1,$FA,$06,$88,$7E,$14,$B9,$7E,$13,$7F
	.dc.b	$A9,$3C,$76,$FC,$71,$F6,$8A,$E4,$F0,$39,$69,$9A,$EB,$C1,$F8,$38
	.dc.b	$C4,$78,$49,$7E,$17,$D2,$7B,$8F,$49,$A9,$E4,$E6,$BC,$6D,$AF,$0B
	.dc.b	$4F,$0D,$FB,$1C,$1C,$DC,$3F,$C4,$4E,$F7,$E3,$AF,$64,$72,$3C,$A6
	.dc.b	$DF,$0B,$6B,$C2,$D3,$84,$72,$7C,$07,$23,$F8,$E4,$70,$69,$C1,$F8
	.dc.b	$F3,$E3,$6D,$FC,$29,$EC,$74,$AE,$D5,$B8,$F7,$B5,$DB,$5E,$E7,$1A
	.dc.b	$EB,$7F,$12,$33,$3B,$7F,$90,$9B,$3A,$DF,$8D,$DB,$5D,$C7,$1C,$3F
	.dc.b	$31,$1C,$4F,$1C,$EE,$33,$3B,$9F,$A7,$39,$6D,$7F,$B7,$1F,$F3,$B9
	.dc.b	$FC,$24,$FF,$12,$70,$7F,$08,$7F,$C2,$FC,$3F,$E2,$47,$F6,$87,$FB
	.dc.b	$D4,$FE,$18,$7E,$F4,$FF,$7C,$E5,$FC,$10,$99,$7F,$01,$7F,$C2,$66
	.dc.b	$23,$FC,$4A,$E2,$34,$7F,$04,$75,$1E,$07,$4F,$08,$D1,$C1,$FE,$28
	.dc.b	$E1,$D1,$83,$FC,$3D,$9F,$F0,$37,$FD,$EA,$DD,$3F,$84,$47,$F8,$96
	.dc.b	$35,$E9,$BC,$7F,$F7,$6C,$FF,$BF,$63,$19,$FF,$A7,$61,$18,$7F,$D4
	.dc.b	$0D,$7F,$95,$9F,$CB,$35,$FF,$2D,$1D,$0B,$3F,$E2,$FC,$BF,$C5,$9A
	.dc.b	$FF,$BF,$27,$4D,$9E,$71,$C2,$BF,$CD,$B5,$C2,$7B,$1A,$AE,$13,$E1
	.dc.b	$6B,$FE,$BE,$9E,$0A,$BB,$D7,$28,$E0,$D7,$86,$7D,$63,$13,$E3,$9F
	.dc.b	$F3,$D1,$CE,$3D,$93,$B8,$C1,$D0,$A7,$09,$F9,$EF,$FB,$32,$BC,$B1
	.dc.b	$E4,$8D,$BD,$DD,$4F,$58,$D1,$2B,$C3,$3D,$E3,$0D,$8C,$BE,$17,$91
	.dc.b	$1F,$9B,$35,$D1,$C4,$C9,$4C,$90,$F2,$99,$9F,$8C,$7A,$5F,$F7,$2D
	.dc.b	$7B,$AB,$B4,$E7,$6F,$F5,$0E,$1F,$62,$38,$BF,$6A,$FA,$9E,$08,$CD
	.dc.b	$7D,$78,$CC,$61,$F9,$33,$C5,$F0,$E1,$E0,$8E,$C6,$6F,$C3,$7F,$7A
	.dc.b	$32,$F5,$AE,$27,$B6,$32,$FD,$C3,$C2,$E2,$FE,$8C,$7E,$28,$E1,$3D
	.dc.b	$4E,$95,$E0,$8F,$3C,$65,$E8,$78,$0F,$81,$5D,$63,$ED,$D7,$BC,$EF
	.dc.b	$3E,$28,$E5,$18,$8E,$4E,$E3,$EF,$46,$CF,$09,$23,$F2,$AF,$09,$F0
	.dc.b	$3F,$81,$38,$47,$B8,$FB,$4F,$FB,$0D,$EC,$EC,$E5,$3E,$43,$C2,$E0
	.dc.b	$F0,$38,$AF,$89,$3D,$6B,$EB,$57,$57,$C6,$FE,$44,$E5,$1B,$AF,$9B
	.dc.b	$1B,$7C,$0F,$8E,$BA,$C7,$E0,$CD,$E1,$F9,$59,$FA,$E7,$D2,$9D,$B8
	.dc.b	$AF,$8B,$18,$8F,$AB,$5C,$B0,$E9,$1E,$C3,$F5,$26,$8C,$1F,$85,$31
	.dc.b	$5F,$5E,$BA,$3F,$2A,$34,$D3,$E7,$DF,$9E,$38,$DF,$E1,$30,$F8,$B5
	.dc.b	$AC,$3C,$6F,$5A,$F8,$87,$89,$D3,$ED,$AF,$A4,$EA,$3E,$63,$F0,$CC
	.dc.b	$47,$43,$C0,$7C,$E7,$EA,$61,$EA,$BF,$AF,$1F,$88,$9E,$0F,$C4,$AF
	.dc.b	$64,$FD,$EC,$3D,$91,$F0,$67,$81,$E5,$38,$4F,$B8,$E1,$87,$48,$F0
	.dc.b	$9E,$63,$81,$F2,$A3,$85,$FD,$78,$CD,$7C,$27,$E5,$9F,$87,$39,$18
	.dc.b	$39,$47,$BE,$38,$1E,$A7,$A4,$6E,$33,$3E,$08,$F9,$57,$EE,$8F,$8B
	.dc.b	$1F,$9A,$AF,$53,$F0,$0F,$01,$6B,$E2,$3B,$3D,$A6,$67,$9E,$1F,$46
	.dc.b	$BE,$A4,$E5,$AF,$13,$A9,$6B,$AC,$F1,$8F,$CB,$9C,$4A,$6E,$BB,$BF
	.dc.b	$1A,$78,$B6,$33,$7E,$FB,$F0,$DF,$5B,$FC,$A4,$E6,$FA,$CE,$6B,$F2
	.dc.b	$73,$9A,$DE,$1F,$7D,$F9,$B3,$F4,$F0,$FC,$1E,$1F,$71,$AF,$D3,$6C
	.dc.b	$FD,$78,$FA,$75,$CE,$72,$7D,$79,$FD,$FC,$FA,$EB,$D2,$D3,$CE,$58
	.dc.b	$F2,$BE,$18,$FC,$1C,$77,$39,$DF,$AE,$3D,$B7,$E1,$77,$7E,$17,$F5
	.dc.b	$B7,$FB,$78,$F2,$C7,$BC,$D9,$63,$D6,$78,$4A,$7C,$17,$6F,$36,$DF
	.dc.b	$EF,$8F,$2D,$78,$E7,$8C,$7A,$67,$C6,$ED,$A7,$E6,$23,$F0,$8F,$C6
	.dc.b	$BF,$74,$7E,$AD,$AF,$ED,$9B,$3F,$6A,$3C,$87,$82,$FE,$13,$FC,$02
	.dc.b	$BD,$D3,$E6,$8F,$73,$5F,$71,$C6,$36,$79,$9F,$D3,$9E,$63,$EC,$D7
	.dc.b	$8A,$BE,$6C,$78,$AB,$A1,$F0,$5F,$11,$9A,$E8,$FE,$AB,$0F,$CC,$3E
	.dc.b	$FC,$3E,$55,$72,$BF,$5C,$FF,$0A,$39,$1D,$23,$E6,$36,$FE,$65,$F4
	.dc.b	$8F,$91,$7D,$27,$C3,$5F,$B0,$C3,$D5,$1F,$70,$E3,$5F,$95,$7F,$1E
	.dc.b	$FC,$87,$8E,$1F,$B6,$BE,$93,$E1,$3E,$3E,$1E,$1C,$3E,$B1,$E6,$8F
	.dc.b	$53,$5F,$E0,$CF,$E8,$9F,$A6,$FC,$1B,$FA,$71,$D4,$F9,$07,$F2,$E3
	.dc.b	$FD,$8E,$FF,$B3,$DF,$F6,$3D,$FF,$AF,$DF,$F5,$E1,$FE,$90,$3F,$D7
	.dc.b	$7B,$7F,$AF,$78,$9B,$FF,$AF,$57,$17,$FA,$06,$BF,$D7,$C7,$FB,$00
	.dc.b	$E5,$1A,$C7,$FD,$40,$D8,$69,$FE,$B8,$D1,$C7,$FE,$A0,$8F,$F5,$CC
	.dc.b	$FF,$AE,$E7,$FD,$6C,$9D,$A3,$9B,$FE,$BD,$78,$1F,$EB,$C3,$84,$F5
	.dc.b	$8D,$57,$8E,$BC,$53,$9A,$FF,$61,$1C,$8E,$27,$78,$FF,$67,$CF,$1A
	.dc.b	$FF,$5B,$EC,$F8,$5F,$F5,$F5,$FF,$AF,$9B,$19,$8F,$F6,$13,$5F,$A0
	.dc.b	$7F,$B3,$B8,$FF,$D8,$DC,$6E,$3F,$F6,$87,$33,$9B,$31,$FE,$D5,$DB
	.dc.b	$FD,$0B,$D7,$FE,$16,$27,$27,$F6,$B6,$E5,$FE,$15,$6C,$FF,$93,$DB
	.dc.b	$3F,$E8,$A6,$CF,$FA,$45,$AF,$FA,$1D,$B3,$E0,$9E,$C4,$9E,$CD,$8E
	.dc.b	$CD,$3B,$16,$3B,$35,$EC,$D8,$ED,$1D,$08,$F4,$6C,$74,$69,$D0,$A7
	.dc.b	$42,$C7,$43,$91,$23,$93,$63,$93,$5E,$4D,$39,$47,$02,$47,$06,$BC
	.dc.b	$1B,$1C,$23,$44,$34,$57,$4E,$08,$E0,$9C,$9F,$EB,$FA,$FA,$AD,$AF
	.dc.b	$94,$DA,$F8,$2D,$AF,$53,$67,$CA,$D9,$F1,$35,$F0,$B6,$7C,$33,$FE
	.dc.b	$FC,$AF,$0B,$4F,$09,$6B,$C2,$D9,$F0,$B4,$F0,$96,$3C,$2D,$7C,$2D
	.dc.b	$8F,$0C,$7F,$BF,$A7,$C2,$D9,$F0,$B4,$F0,$96,$7C,$2D,$8F,$0B,$4F
	.dc.b	$09,$4F,$09,$63,$C2,$23,$F8,$A4,$21,$3E,$47,$51,$FC,$65,$CC,$F9
	.dc.b	$9C,$C7,$19,$E1,$3D,$27,$73,$EA,$76,$60,$F5,$98,$9E,$31,$DD,$AF
	.dc.b	$59,$F0,$1D,$63,$33,$CE,$3E,$1B,$CD,$DC,$F3,$6B,$CC,$F2,$3C,$CC
	.dc.b	$C7,$38,$F9,$6F,$37,$67,$32,$9C,$DF,$5C,$73,$8C,$C7,$37,$E9,$BC
	.dc.b	$CF,$87,$5F,$EF,$08,$FA,$4D,$83,$49,$3F,$CC,$A7,$6D,$9D,$B6,$77
	.dc.b	$3F,$E1,$ED,$BF,$BE,$BF,$DE,$27,$FC,$0B,$C1,$FE,$08,$8F,$F1,$41
	.dc.b	$FE,$14,$34,$7F,$80,$8F,$F0,$51,$FE,$1D,$3F,$C1,$47,$F7,$BB,$FF
	.dc.b	$09,$57,$68,$FF,$01,$D7,$F8,$41,$FF,$01,$BF,$E1,$03,$FC,$8D,$1B
	.dc.b	$8F,$F1,$A4,$FF,$82,$4F,$F1,$83,$E2,$3F,$C0,$F5,$E7,$8F,$F0,$64
	.dc.b	$79,$CE,$E7,$94,$EE,$79,$6F,$FC,$3C,$FA,$0E,$91,$E8,$70,$72,$3F
	.dc.b	$C5,$E6,$0F,$2D,$FA,$E3,$CB,$1B,$7F,$C4,$B1,$F3,$2F,$D9,$CB,$FE
	.dc.b	$29,$BF,$F2,$6D,$7F,$89,$EF,$99,$E7,$79,$B8,$3C,$27,$A8,$F3,$C7
	.dc.b	$A9,$C1,$E1,$7A,$DF,$86,$3D,$15,$E1,$8F,$F2,$24,$7D,$BC,$3E,$54
	.dc.b	$7D,$38,$FC,$31,$C0,$FB,$E7,$8A,$FC,$B5,$CE,$3F,$0F,$1E,$B8,$C3
	.dc.b	$F5,$0C,$47,$27,$E2,$BE,$1C,$3E,$2B,$9C,$3D,$2E,$4F,$16,$1E,$67
	.dc.b	$ED,$61,$F8,$87,$35,$CA,$75,$1F,$1A,$74,$FC,$63,$E6,$5F,$89,$FC
	.dc.b	$5D,$71,$23,$D0,$87,$1C,$3D,$B5,$F1,$EB,$F2,$D8,$7F,$9F,$2B,$FD
	.dc.b	$F0,$7F,$8F,$EB,$FC,$81,$CF,$FE,$43,$8F,$E5,$53,$FC,$B2,$31,$1F
	.dc.b	$CC,$A7,$78,$7F,$29,$AD,$3F,$EC,$2C,$77,$5F,$E6,$63,$FC,$96,$7F
	.dc.b	$A5,$1C,$BF,$E6,$8C,$7F,$C8,$53,$C4,$91,$FE,$69,$BF,$F2,$0A,$69
	.dc.b	$1F,$F1,$FF,$17,$F9,$B3,$9B,$FC,$DD,$CD,$FE,$6D,$3F,$CD,$DC,$DF
	.dc.b	$E6,$DA,$FF,$35,$B5,$FE,$9B,$5F,$E6,$B7,$3A,$FF,$9B,$A7,$FC,$E9
	.dc.b	$9F,$F9,$72,$7C,$E7,$F9,$72,$BF,$CB,$0F,$F9,$97,$2F,$F3,$5B,$F0
	.dc.b	$0F,$F3,$04,$69,$FF,$30,$99,$8D,$CF,$F9,$DA,$75,$18,$3F,$CE,$EF
	.dc.b	$13,$3C,$DF,$E6,$DA,$E7,$5F,$E7,$EA,$F0,$E1,$FD,$81,$BA,$7F,$6A
	.dc.b	$6D,$7F,$9D,$1A,$FF,$A0,$F8,$7F,$B2,$4E,$DB,$1F,$E8,$59,$DF,$1F
	.dc.b	$F9,$AF,$93,$FC,$D2,$7F,$9B,$B9,$3F,$CD,$95,$FE,$72,$9D,$3F,$E7
	.dc.b	$4F,$9D,$78,$9F,$D2,$DB,$97,$F4,$76,$CE,$E7,$5A,$CA,$D3,$67,$A3
	.dc.b	$FE,$02,$D3,$FB,$FD,$FF,$A8,$AB,$FC,$08,$7F,$82,$9F,$F4,$F3,$8A
	.dc.b	$FF,$50,$46,$DF,$F0,$74,$6E,$BA,$34,$E1,$09,$E6,$F0,$39,$CF,$06
	.dc.b	$CF,$09,$ED,$5E,$69,$D1,$CE,$AC,$EA,$BC,$13,$D4,$F0,$1A,$99,$1D
	.dc.b	$5E,$31,$83,$84,$F3,$7C,$66,$0D,$57,$C8,$9B,$5F,$1E,$69,$C2,$75
	.dc.b	$1E,$B8,$C9,$E9,$7C,$46,$4A,$F8,$A7,$75,$F2,$2B,$9C,$E6,$B7,$1F
	.dc.b	$15,$A7,$C2,$76,$66,$38,$11,$CC,$7A,$4E,$31,$E9,$79,$19,$78,$38
	.dc.b	$74,$7E,$10,$F3,$D7,$A1,$B1,$D1,$FA,$6D,$AF,$9D,$1F,$0C,$F9,$D1
	.dc.b	$99,$E4,$D3,$84,$74,$3C,$67,$85,$D3,$83,$B9,$67,$B9,$1E,$55,$F0
	.dc.b	$9E,$46,$EB,$C2,$DA,$F0,$9D,$A7,$84,$76,$6D,$76,$AF,$C6,$D6,$CF
	.dc.b	$91,$5B,$6C,$ED,$FC,$14,$77,$3D,$07,$99,$DB,$4F,$A2,$73,$2C,$73
	.dc.b	$7E,$59,$CE,$BC,$AF,$13,$E8,$3F,$85,$3D,$35,$F0,$1B,$5F,$02,$79
	.dc.b	$3E,$23,$D8,$F8,$0E,$95,$F3,$63,$31,$D9,$FC,$54,$F4,$8F,$6C,$6A
	.dc.b	$78,$47,$63,$A4,$66,$BE,$5C,$64,$E8,$7D,$8A,$E8,$7B,$A7,$81,$F3
	.dc.b	$1C,$BE,$98,$E5,$5E,$09,$CC,$78,$0C,$D7,$82,$33,$1E,$03,$E8,$16
	.dc.b	$7E,$84,$7C,$BA,$FB,$B5,$F2,$DA,$7C,$B8,$E7,$3E,$18,$CD,$7C,$43
	.dc.b	$C2,$FA,$4C,$CF,$B6,$32,$7D,$E7,$9D,$79,$E7,$67,$BE,$32,$F3,$9F
	.dc.b	$C3,$CF,$95,$B1,$E5,$38,$BF,$0E,$78,$BD,$E3,$2F,$C1,$3F,$03,$3E
	.dc.b	$17,$99,$EB,$9E,$27,$C5,$72,$FA,$CC,$CF,$E1,$8E,$B5,$F1,$67,$71
	.dc.b	$EB,$33,$5D,$E3,$E9,$3E,$B6,$C7,$AD,$E2,$7A,$DE,$27,$C3,$76,$79
	.dc.b	$DC,$BE,$F3,$89,$EF,$7F,$1A,$E4,$E3,$5E,$B8,$E2,$FB,$E3,$31,$E9
	.dc.b	$79,$D7,$B5,$C9,$E5,$33,$5E,$D8,$FC,$6C,$F1,$9D,$9E,$33,$F0,$75
	.dc.b	$DE,$77,$1C,$CD,$BD,$CC,$BC,$63,$E8,$BF,$AB,$9F,$49,$F4,$DF,$1B
	.dc.b	$F1,$E7,$89,$CE,$BF,$80,$CF,$E0,$A7,$51,$87,$E3,$D7,$96,$31,$3E
	.dc.b	$89,$EC,$F9,$A3,$E0,$1A,$9F,$2C,$7E,$C2,$BC,$33,$AA,$FE,$15,$3E
	.dc.b	$13,$EA,$4E,$A3,$C6,$68,$F1,$9E,$07,$51,$C2,$7C,$13,$F0,$A3,$E2
	.dc.b	$1F,$08,$D3,$CD,$F5,$1A,$3D,$4F,$81,$FE,$1E,$79,$6B,$E1,$47,$17
	.dc.b	$DB,$5F,$DF,$E3,$C1,$3E,$FA,$F1,$BE,$1A,$E2,$F4,$9F,$5B,$E1,$AF
	.dc.b	$5C,$E8,$FE,$1F,$1E,$19,$F8,$15,$D2,$3C,$E7,$B0,$E8,$7B,$0F,$14
	.dc.b	$79,$CF,$2D,$79,$E3,$C9,$1D,$0D,$38,$3C,$93,$DE,$7C,$B3,$B9,$EF
	.dc.b	$5F,$09,$F3,$1C,$23,$9B,$FC,$C7,$F3,$67,$C4,$FF,$91,$1B,$3F,$E5
	.dc.b	$7A,$7E,$A2,$F8,$ED,$5C,$AF,$F2,$AD,$F0,$C3,$FC,$A9,$87,$0B,$D4
	.dc.b	$7F,$92,$E1,$CF,$FE,$F7,$7F,$DD,$31,$FE,$99,$73,$7F,$E6,$08,$FF
	.dc.b	$5E,$71,$7F,$AE,$B9,$FF,$D4,$70,$AD,$BF,$EA,$13,$7C,$5F,$EB,$C7
	.dc.b	$FD,$94,$08,$7F,$AF,$CA,$52,$9F,$ED,$73,$FD,$7F,$CD,$FE,$BD,$E7
	.dc.b	$FF,$5F,$9F,$EB,$7A,$93,$FE,$C4,$E7,$FF,$5F,$3D,$F9,$FF,$D7,$CF
	.dc.b	$FB,$02,$3F,$D7,$B8,$7F,$A9,$78,$BF,$D5,$8F,$FA,$C6,$7F,$D7,$D3
	.dc.b	$A7,$FD,$7D,$3F,$EB,$1A,$FF,$5A,$4E,$67,$FD,$67,$5F,$EB,$88,$C4
	.dc.b	$FF,$AE,$27,$E3,$CF,$57,$FD,$8B,$3C,$63,$06,$67,$84,$FF,$A0,$63
	.dc.b	$FD,$6E,$7F,$AE,$23,$37,$FE,$BF,$E7,$FF,$5B,$D7,$FA,$EC,$FF,$4B
	.dc.b	$D7,$FA,$FB,$C8,$78,$9F,$32,$7E,$94,$61,$C3,$5F,$A2,$D7,$EA,$F3
	.dc.b	$FF,$AF,$1B,$D8,$EA,$FF,$61,$63,$FE,$C4,$9F,$F5,$AF,$17,$E6,$79
	.dc.b	$3F,$0B,$A7,$F9,$54,$FC,$D7,$27,$FA,$39,$B3,$FE,$78,$6D,$7F,$99
	.dc.b	$1A,$71,$AE,$73,$FE,$79,$8C,$B6,$33,$3F,$E7,$C9,$CB,$63,$2F,$F8
	.dc.b	$C2,$32,$D8,$CF,$47,$FA,$A6,$72,$D8,$CF,$0F,$FA,$FE,$72,$D8,$CD
	.dc.b	$FC,$1C,$3D,$D1,$8C,$3F,$6A,$DB,$FF,$6D,$36,$BF,$4D,$1D,$1E,$47
	.dc.b	$C5,$8D,$4F,$EC,$5B,$3B,$8E,$B5,$E0,$25,$71,$6C,$7C,$DA,$E2,$48
	.dc.b	$CD,$7C,$F6,$D7,$CD,$6D,$7C,$B6,$BF,$25,$B8,$7C,$76,$E1,$F1,$5B
	.dc.b	$5E,$CA,$F1,$46,$1F,$4C,$79,$C8,$E9,$B7,$E5,$69,$E5,$2C,$79,$5A
	.dc.b	$F9,$5B,$7E,$56,$DF,$95,$A7,$96,$BB,$57,$84,$8E,$09,$1E,$18,$CB
	.dc.b	$63,$31,$E3,$32,$57,$2F,$94,$C9,$5C,$D7,$9E,$32,$D8,$CD,$7A,$67
	.dc.b	$2D,$8C,$D7,$AC,$F2,$12,$7C,$4D,$7C,$0D,$AE,$CD,$AE,$8D,$3A,$4F
	.dc.b	$22,$1C,$8B,$3C,$9B,$3C,$9B,$3C,$9E,$04,$9E,$0D,$9E,$0D,$9E,$13
	.dc.b	$A2,$4E,$9B,$1A,$8C,$11,$C1,$27,$F2,$27,$CA,$71,$5F,$B6,$BF,$CB
	.dc.b	$36,$3F,$0F,$3C,$23,$E7,$B6,$3E,$33,$4F,$79,$4F,$51,$6F,$CE,$DB
	.dc.b	$F3,$36,$7C,$CF,$F6,$89,$4A,$7F,$9F,$46,$1B,$19,$9F,$E6,$A4,$39
	.dc.b	$39,$6C,$70,$3F,$D2,$0F,$F8,$B2,$BF,$C6,$51,$FE,$9C,$8F,$F0,$E9
	.dc.b	$FE,$2C,$3F,$C4,$2F,$F8,$B1,$CD,$7F,$8B,$5E,$51,$FE,$27,$9F,$F1
	.dc.b	$DB,$FE,$40,$8E,$51,$87,$FC,$71,$3C,$4F,$F1,$7B,$FE,$2B,$3B,$9F
	.dc.b	$E3,$53,$8C,$F9,$4E,$91,$C6,$BF,$C7,$04,$8F,$F1,$9C,$74,$9E,$27
	.dc.b	$FA,$70,$F4,$92,$39,$B4,$F2,$96,$BF,$C8,$11,$A2,$46,$9B,$1A,$8C
	.dc.b	$11,$C1,$23,$E5,$1C,$0B,$1B,$6C,$6E,$3E,$14,$FB,$E7,$DB,$5E,$F9
	.dc.b	$F8,$E7,$D2,$3E,$55,$7C,$F3,$93,$93,$E1,$9F,$EB,$03,$DD,$3E,$02
	.dc.b	$3E,$03,$25,$7D,$0D,$3D,$05,$8E,$F5,$F4,$4F,$44,$79,$9C,$3D,$0F
	.dc.b	$23,$A8,$F4,$4F,$8A,$72,$6E,$78,$C7,$8A,$3F,$CE,$1C,$FF,$9F,$7F
	.dc.b	$CA,$7C,$FF,$E6,$97,$FC,$A6,$D7,$31,$B7,$FC,$B5,$1F,$E6,$FB,$FD
	.dc.b	$43,$6B,$FB,$CF,$47,$F9,$DD,$FF,$3A,$DF,$EC,$6B,$F5,$C7,$F8,$48
	.dc.b	$FF,$09,$DF,$F9,$7F,$A1,$C7,$FE,$66,$9F,$F3,$AC,$62,$7F,$C0,$BA
	.dc.b	$73,$C3,$FC,$DD,$1F,$E7,$03,$FC,$D9,$3B,$7F,$CE,$D3,$FC,$06,$3F
	.dc.b	$CF,$51,$FE,$7F,$8C,$5F,$28,$74,$7F,$9F,$B0,$FF,$3E,$DF,$03,$A3
	.dc.b	$8A,$FF,$3B,$CF,$F9,$36,$BF,$CF,$07,$F9,$3D,$C3,$D0,$FF,$3D,$F3
	.dc.b	$FF,$9E,$9A,$7F,$9E,$6B,$FC,$F6,$FF,$A1,$39,$BF,$CF,$2F,$F9,$EF
	.dc.b	$9F,$FC,$FA,$A3,$FC,$CF,$8F,$BF,$1F,$F3,$55,$7F,$9C,$8F,$F3,$77
	.dc.b	$07,$F9,$F8,$FF,$3D,$39,$C3,$FC,$EB,$19,$C3,$FD,$3B,$87,$F9,$FA
	.dc.b	$73,$10,$61,$F3,$67,$51,$8C,$3F,$D1,$8E,$E7,$58,$78,$A3,$FC,$EB
	.dc.b	$87,$F9,$E2,$BF,$CF,$93,$99,$FF,$40,$71,$64,$D5,$7F,$9F,$BC,$4F
	.dc.b	$F5,$CC,$FF,$9E,$DA,$FF,$AA,$1A,$FF,$AB,$1A,$FF,$AC,$1D,$E5,$F9
	.dc.b	$3F,$D4,$DF,$09,$F8,$FE,$0F,$F2,$8C,$62,$FF,$CA,$71,$C7,$83,$FD
	.dc.b	$1F,$D1,$FE,$83,$9F,$F5,$DB,$5F,$F5,$F2,$ED,$69,$FE,$85,$9F,$F5
	.dc.b	$FF,$CC,$3E,$0F,$F5,$90,$8C,$63,$FE,$97,$6C,$7F,$B1,$1A,$FF,$61
	.dc.b	$F5,$BF,$CE,$59,$BE,$1B,$F4,$93,$FE,$CF,$EA,$56,$A7,$32,$9C,$E1
	.dc.b	$FD,$A2,$BC,$71,$89,$FF,$69,$B5,$F0,$4F,$C4,$7B,$4F,$8B,$E4,$7A
	.dc.b	$D7,$1E,$0F,$ED,$9F,$9E,$3E,$23,$CF,$F2,$7E,$28,$C6,$5F,$EF,$FD
	.dc.b	$13,$9A,$F8,$B5,$AC,$FF,$DC,$AD,$7F,$DD,$2D,$7F,$DD,$AE,$F0,$F0
	.dc.b	$CF,$C6,$F1,$7A,$63,$E9,$B5,$FE,$0A,$3C,$5F,$15,$E5,$E5,$F9,$AD
	.dc.b	$8F,$F6,$54,$EE,$BC,$75,$E6,$24,$F5,$9D,$63,$F8,$FF,$8C,$E5,$68
	.dc.b	$F8,$B9,$FD,$AF,$C0,$FE,$23,$70,$E4,$FC,$22,$3E,$26,$D7,$38,$F8
	.dc.b	$31,$F8,$4E,$BE,$71,$E3,$C3,$F0,$FF,$2B,$F4,$5B,$19,$3D,$AF,$D0
	.dc.b	$23,$83,$57,$DC,$F5,$6E,$F9,$6E,$93,$F0,$6F,$F2,$BF,$2F,$F7,$DB
	.dc.b	$19,$3E,$5B,$F7,$88,$E0,$D5,$F7,$D1,$87,$F8,$F6,$3C,$47,$DA,$AF
	.dc.b	$C6,$76,$FE,$93,$A6,$E1,$F8,$D6,$D7,$E2,$A7,$F1,$A4,$9F,$BA,$EB
	.dc.b	$0F,$E0,$D3,$64,$FE,$B7,$8B,$FD,$C5,$E2,$7E,$DA,$BF,$C9,$5C,$A9
	.dc.b	$CF,$57,$E9,$1B,$87,$E4,$67,$4F,$EA,$E9,$7C,$E3,$BE,$1F,$B7,$BE
	.dc.b	$2F,$C0,$BE,$2F,$C5,$AE,$0F,$96,$BA,$46,$23,$94,$63,$0F,$23,$4F
	.dc.b	$25,$F6,$C3,$D6,$D8,$F1,$E1,$E0,$9F,$EF,$98,$75,$8C,$4F,$F8,$0A
	.dc.b	$FA,$E1,$E5,$6D,$F0,$6C,$EC,$F0,$9E,$39,$F7,$61,$E1,$7D,$87,$38
	.dc.b	$DB,$4D,$BC,$6B,$D2,$D3,$D3,$3C,$E7,$E1,$9C,$67,$6D,$9D,$9C,$5F
	.dc.b	$0C,$7A,$63,$C2,$D8,$F0,$9C,$4D,$C7,$B5,$CC,$6C,$EE,$58,$EF,$1C
	.dc.b	$4F,$2C,$65,$E6,$7F,$09,$C3,$E8,$57,$AF,$0F,$A3,$39,$C3,$E6,$B7
	.dc.b	$0F,$39,$E3,$C3,$C2,$D3,$C2,$F7,$8F,$0C,$65,$F0,$C7,$E5,$A7,$AB
	.dc.b	$67,$AC,$F7,$3D,$EF,$17,$99,$98,$F0,$B6,$3C,$26,$EB,$D7,$3D,$67
	.dc.b	$99,$B3,$C7,$87,$C2,$AF,$AD,$1D,$A7,$81,$D8,$B1,$DA,$39,$1B,$9E
	.dc.b	$4F,$D1,$AF,$0C,$62,$BE,$24,$62,$FE,$0C,$E6,$FD,$18,$7D,$32,$3E
	.dc.b	$4A,$E0,$FC,$6B,$F1,$E1,$EA,$C3,$99,$27,$58,$75,$9F,$3D,$F2,$7E
	.dc.b	$1D,$F2,$7D,$57,$C5,$CD,$71,$9D,$57,$17,$E2,$46,$49,$5C,$63,$2D
	.dc.b	$BF,$91,$1C,$B0,$F6,$E1,$F1,$DB,$1F,$19,$B3,$F1,$0F,$84,$E2,$7D
	.dc.b	$6D,$9E,$93,$A2,$4F,$56,$C7,$57,$31,$FC,$98,$FD,$D9,$70,$FE,$4A
	.dc.b	$DA,$FF,$0E,$92,$78,$47,$F0,$C3,$4E,$1F,$CB,$9F,$85,$F8,$FF,$D5
	.dc.b	$B8,$AF,$C2,$C6,$1F,$D8,$C6,$E7,$F4,$6E,$EB,$FA,$3D,$6D,$FC,$4B
	.dc.b	$67,$EF,$B6,$7E,$CB,$6F,$EB,$B6,$7E,$9E,$28,$C5,$7E,$91,$AF,$8C
	.dc.b	$FD,$49,$E3,$9F,$E8,$B1,$AC,$3D,$71,$E8,$7D,$66,$27,$8B,$8C,$3D
	.dc.b	$73,$B7,$D6,$D3,$D6,$59,$F5,$B7,$0F,$5B,$4F,$5D,$F8,$EB,$AD,$78
	.dc.b	$1E,$AD,$9E,$AD,$3A,$96,$BA,$B6,$3A,$9D,$DF,$8C,$D7,$E3,$47,$F1
	.dc.b	$47,$E3,$1F,$DB,$6B,$C9,$3F,$20,$E1,$7F,$16,$B8,$3F,$16,$35,$5E
	.dc.b	$13,$B9,$4E,$EF,$F1,$37,$BC,$FF,$6C,$AE,$F1,$E1,$9E,$F7,$F1,$2B
	.dc.b	$BB,$F1,$1B,$1F,$11,$B3,$F1,$1A,$7C,$42,$C7,$C4,$8F,$25,$FA,$E3
	.dc.b	$C8,$FA,$E3,$C4,$6A,$BC,$53,$A7,$C4,$D3,$C4,$58,$F1,$36,$BC,$4D
	.dc.b	$3C,$4F,$92,$3E,$3B,$5F,$8E,$7F,$62,$1F,$9E,$3E,$2B,$FD,$43,$2F
	.dc.b	$25,$F0,$7F,$EA,$0A,$D6,$1F,$E2,$F9,$CB,$FE,$83,$8C,$E1,$FC,$11
	.dc.b	$AF,$F0,$D6,$BF,$E3,$46,$BF,$E4,$47,$8D,$A7,$58,$79,$2D,$EF,$78
	.dc.b	$E3,$C9,$FA,$26,$7E,$2E,$DF,$95,$B9,$78,$E7,$FD,$8F,$91,$A4,$AF
	.dc.b	$53,$DE,$F8,$E5,$E3,$BF,$37,$CE,$F8,$FE,$0B,$E2,$37,$0F,$84,$F6
	.dc.b	$23,$FC,$8F,$14,$62,$BF,$CE,$31,$89,$F9,$9F,$A7,$74,$7E,$67,$95
	.dc.b	$F0,$73,$FA,$4D,$8C,$9D,$5F,$69,$1C,$1A,$BF,$F4,$16,$CF,$63,$9C
	.dc.b	$FD,$9F,$7B,$E7,$7C,$57,$DF,$6C,$64,$F5,$3F,$28,$8E,$0D,$5F,$6D
	.dc.b	$39,$47,$E1,$7E,$3F,$ED,$7C,$E7,$E2,$DB,$19,$3E,$33,$F5,$88,$E0
	.dc.b	$D5,$7A,$BD,$AE,$51,$FE,$D3,$C9,$F2,$3F,$87,$D7,$FD,$EF,$18,$9F
	.dc.b	$CC,$46,$27,$EE,$5F,$F6,$1A,$D3,$FD,$71,$AE,$63,$6F,$FA,$32,$37
	.dc.b	$C5,$C9,$B7,$FE,$9E,$8F,$D5,$7B,$BD,$E3,$FD,$D1,$EC,$3E,$67,$FD
	.dc.b	$43,$E6,$7E,$36,$FC,$73,$9A,$FF,$48,$CE,$5F,$F4,$DC,$67,$1F,$EE
	.dc.b	$2D,$7F,$BD,$35,$DD,$FA,$1F,$F5,$8F,$67,$FB,$07,$0F,$F5,$7F,$81
	.dc.b	$D6,$72,$51,$3B,$C3,$FD,$33,$39,$8C,$4F,$09,$FF,$5A,$4F,$1A,$7A
	.dc.b	$74,$E9,$5F,$EC,$2F,$63,$FD,$87,$EB,$3C,$AF,$F5,$BD,$F0,$C7,$E5
	.dc.b	$18,$C3,$93,$8C,$3F,$C3,$18,$72,$BD,$4E,$67,$FD,$83,$39,$7E,$A5
	.dc.b	$7E,$EF,$0F,$04,$72,$0E,$16,$1E,$37,$38,$7F,$88,$2F,$A4,$6A,$FA
	.dc.b	$4E,$63,$67,$83,$0D,$BF,$04,$C9,$C5,$CA,$AE,$B7,$E1,$C3,$C9,$3C
	.dc.b	$5F,$34,$66,$FC,$11,$8C,$3C,$75,$ED,$8C,$61,$DB,$0E,$B5,$AA,$F8
	.dc.b	$E7,$F8,$E4,$F1,$16,$30,$E0,$FF,$24,$7A,$EE,$37,$B1,$FE,$41,$8C
	.dc.b	$36,$33,$3F,$E4,$89,$CB,$79,$3E,$7B,$74,$E5,$18,$8E,$ED,$7F,$DF
	.dc.b	$B2,$8E,$39,$FD,$9D,$1A,$AF,$EB,$C6,$1F,$B3,$7C,$70,$B8,$72,$AE
	.dc.b	$15,$F9,$5B,$B8,$75,$6C,$F8,$AB,$85,$77,$C3,$A3,$F7,$A7,$C0,$DA
	.dc.b	$E2,$EE,$B8,$AA,$E3,$7D,$E3,$D3,$5C,$1D,$57,$0A,$DD,$FA,$0E,$F7
	.dc.b	$C6,$BF,$1F,$3A,$7E,$24,$F9,$30,$FE,$46,$DD,$79,$63,$FB,$47,$DD
	.dc.b	$3E,$2A,$FD,$75,$3C,$7D,$47,$46,$BF,$24,$FD,$A7,$B1,$FC,$11,$BB
	.dc.b	$EA,$31,$1F,$B0,$C3,$E8,$C7,$DD,$8D,$9F,$C0,$70,$79,$4F,$47,$C3
	.dc.b	$18,$7A,$61,$E2,$F9,$57,$C1,$FE,$75,$B8,$7F,$A5,$DA,$FF,$A7,$1A
	.dc.b	$FF,$A9,$7D,$A7,$BA,$F7,$7F,$D3,$D1,$86,$FB,$9F,$0E,$BF,$D4,$B3
	.dc.b	$97,$FD,$4F,$D0,$F6,$BF,$C5,$FB,$3D,$BF,$E5,$91,$86,$F0,$FB,$E7
	.dc.b	$4F,$F1,$DA,$E4,$E6,$77,$7D,$AF,$FA,$03,$FC,$E9,$B5,$C2,$3F,$D7
	.dc.b	$3E,$CB,$BB,$C7,$EC,$7F,$41,$8C,$63,$E1,$8C,$4E,$A3,$F9,$74,$EA
	.dc.b	$BE,$C1,$FC,$A3,$25,$F0,$AF,$F6,$25,$6E,$BF,$DA,$0D,$7F,$5E,$D7
	.dc.b	$F7,$0F,$DD,$8F,$85,$ED,$79,$78,$FE,$4B,$63,$27,$92,$32,$D3,$C2
	.dc.b	$FC,$49,$E9,$5F,$0A,$31,$7F,$53,$DA,$F8,$4D,$BF,$7B,$6B,$DE,$2F
	.dc.b	$DF,$3F,$5E,$B8,$BF,$02,$BA,$B4,$F8,$7E,$D7,$E0,$BD,$8F,$A0,$D8
	.dc.b	$F9,$AD,$C3,$BE,$08,$C7,$3F,$F8,$6A,$31,$5D,$09,$5F,$7A,$BE,$05
	.dc.b	$FE,$43,$D9,$FB,$CD,$AF,$B6,$DA,$EE,$DC,$3B,$3C,$A7,$ED,$B6,$7F
	.dc.b	$17,$D9,$F9,$CE,$BF,$C5,$36,$BF,$0E,$D9,$F5,$B6,$FD,$4F,$D5,$24
	.dc.b	$FE,$21,$FC,$CF,$AD,$FE,$96,$F6,$1E,$CF,$F8,$4E,$30,$DA,$FE,$0D
	.dc.b	$3A,$7F,$51,$3A,$69,$A9,$E1,$87,$E6,$B0,$5F,$18,$FB,$FC,$8C,$39
	.dc.b	$46,$1B,$5F,$E4,$C9,$D6,$1D,$4E,$F5,$D6,$FA,$56,$EF,$8D,$7F,$50
	.dc.b	$BE,$37,$BC,$3C,$B7,$FE,$47,$C3,$8D,$EF,$0E,$37,$FE,$3E,$6B,$FE
	.dc.b	$3A,$6E,$1D,$2F,$96,$1D,$2F,$FC,$B1,$87,$4C,$39,$47,$A5,$F7,$36
	.dc.b	$BD,$0D,$3C,$45,$8E,$27,$A5,$F8,$31,$E0,$69,$E0,$AF,$14,$FF,$8C
	.dc.b	$2B,$C4,$FA,$67,$A3,$4E,$87,$90,$B5,$E4,$6C,$79,$1A,$79,$23,$C4
	.dc.b	$F1,$69,$C4,$B1,$C5,$B1,$C6,$3D,$07,$A9,$F4,$57,$AA,$7F,$12,$DA
	.dc.b	$F5,$DF,$D1,$BE,$37,$E1,$6D,$F8,$5A,$78,$4B,$1E,$13,$BB,$E1,$8E
	.dc.b	$ED,$AE,$F8,$7E,$00,$91,$F2,$AF,$C5,$7C,$8F,$11,$6F,$C4,$D9,$F1
	.dc.b	$34,$F1,$47,$81,$E2,$D3,$89,$63,$8B,$63,$8D,$F9,$A7,$C9,$7E,$69
	.dc.b	$FA,$E7,$DF,$AF,$A2,$D9,$F8,$6D,$3C,$85,$8E,$CD,$AE,$CF,$4B,$F8
	.dc.b	$33,$A6,$CE,$A3,$BC,$63,$0F,$B4,$4B,$F9,$CD,$BF,$83,$7C,$67,$CB
	.dc.b	$5C,$5D,$92,$77,$5C,$6F,$E3,$DF,$19,$F5,$DF,$17,$77,$C6,$B7,$7C
	.dc.b	$63,$E6,$DF,$19,$F6,$61,$F2,$DF,$5B,$67,$CA,$DB,$EE,$D9,$E3,$3F
	.dc.b	$50,$93,$F2,$67,$55,$DC,$97,$E3,$9F,$BA,$D7,$E9,$37,$0F,$66,$1E
	.dc.b	$AC,$3D,$87,$AA,$75,$87,$66,$C7,$8A,$7D,$6D,$3D,$65,$BF,$5B,$63
	.dc.b	$D7,$1E,$98,$C6,$1E,$87,$C3,$1E,$86,$BE,$86,$E1,$E8,$7C,$C6,$30
	.dc.b	$F3,$34,$EB,$87,$A2,$BC,$04,$C3,$D2,$D9,$F4,$B4,$F4,$96,$3D,$27
	.dc.b	$02,$47,$06,$C7,$06,$9C,$23,$44,$8D,$36,$35,$18,$23,$82,$61,$F1
	.dc.b	$E3,$F2,$64,$8F,$B4,$D9,$F8,$AD,$3E,$61,$4F,$E7,$47,$FA,$B8,$A7
	.dc.b	$F3,$73,$18,$7F,$62,$6E,$1F,$CA,$FD,$9F,$E8,$F1,$FC,$04,$FD,$F3
	.dc.b	$89,$FE,$3B,$18,$AF,$E6,$CE,$E3,$F7,$F3,$BA,$E9,$5A,$9E,$26,$9C
	.dc.b	$46,$AB,$F9,$39,$CD,$FF,$0A,$57,$39,$FE,$64,$D9,$FE,$0A,$DB,$FD
	.dc.b	$FB,$67,$F7,$2D,$9F,$E4,$AD,$7D,$47,$82,$FC,$B7,$E0,$AF,$6F,$A8
	.dc.b	$F5,$91,$FE,$A7,$F4,$DE,$CE,$31,$FF,$2C,$DE,$EB,$53,$9C,$F5,$4A
	.dc.b	$D3,$FE,$25,$BE,$77,$FE,$ED,$AD,$D6,$9E,$53,$AA,$EF,$87,$FA,$52
	.dc.b	$B9,$C7,$E0,$5A,$FF,$BC,$1A,$FF,$BE,$57,$AB,$E5,$C3,$C1,$6F,$63
	.dc.b	$17,$BA,$B8,$AB,$57,$E7,$3B,$4E,$A3,$35,$F1,$23,$17,$D2,$3E,$14
	.dc.b	$7A,$49,$5F,$99,$8F,$03,$5F,$91,$EA,$75,$F6,$39,$64,$8C,$57,$8E
	.dc.b	$BD,$D1,$87,$EA,$B7,$2F,$9A,$FC,$92,$4F,$7A,$E8,$DF,$2F,$D9,$8F
	.dc.b	$81,$B1,$DF,$0F,$B9,$18,$7F,$06,$DC,$3E,$DB,$F5,$C9,$3D,$4F,$07
	.dc.b	$AD,$F2,$7D,$7F,$6B,$63,$B6,$28,$C6,$1F,$90,$6E,$1F,$88,$7F,$08
	.dc.b	$4A,$EA,$7B,$7D,$6F,$AD,$EB,$FC,$A6,$C7,$66,$DF,$E7,$1B,$7F,$94
	.dc.b	$7F,$20,$49,$E8,$7D,$7F,$59,$EB,$FE,$0F,$4F,$C0,$46,$27,$F5,$51
	.dc.b	$89,$D4,$7C,$B9,$D6,$1F,$A5,$AD,$4F,$5A,$FF,$79,$D6,$AB,$F9,$03
	.dc.b	$51,$3D,$67,$AD,$FE,$9F,$F1,$BF,$25,$18,$C3,$FD,$3B,$18,$9D,$47
	.dc.b	$DD,$9D,$57,$FA,$3E,$BF,$35,$D0,$EC,$A6,$26,$DF,$E1,$67,$4D,$9D
	.dc.b	$34,$FC,$56,$AF,$3F,$F4,$CD,$8F,$C7,$5F,$B3,$D6,$F6,$FA,$8F,$4F
	.dc.b	$DF,$2B,$E0,$FF,$BB,$1A,$7F,$12,$2D,$FE,$BE,$F9,$D7,$29,$CD,$7F
	.dc.b	$4B,$9C,$BF,$C5,$70,$FE,$41,$87,$86,$34,$1C,$3E,$38,$DD,$F7,$8D
	.dc.b	$3D,$E3,$31,$CA,$78,$DF,$A9,$E7,$1F,$04,$E2,$78,$F0,$F2,$BF,$0B
	.dc.b	$0B,$5C,$70,$F0,$5F,$4C,$3B,$DE,$CE,$67,$9F,$0E,$6F,$79,$E1,$18
	.dc.b	$86,$1D,$AB,$9C,$79,$4F,$E3,$11,$FC,$65,$C1,$F0,$4F,$EE,$DE,$A3
	.dc.b	$D5,$7A,$DF,$C7,$E3,$0D,$8C,$CF,$F2,$59,$CB,$78,$FF,$00,$DD,$39
	.dc.b	$46,$23,$BB,$5F,$C3,$CA,$38,$E7,$F6,$34,$FC,$E5,$7C,$C6,$BF,$1A
	.dc.b	$B8,$5F,$D0,$6C,$7C,$A9,$DD,$73,$D7,$F2,$AF,$E1,$2B,$6D,$C3,$C3
	.dc.b	$5F,$9A,$38,$96,$B8,$BE,$88,$DD,$FE,$FA,$F9,$CE,$E3,$D1,$5C,$1D
	.dc.b	$57,$05,$5C,$1F,$E0,$F5,$E0,$6D,$71,$77,$5C,$67,$E1,$E1,$FD,$F1
	.dc.b	$BA,$F8,$6B,$F6,$4F,$9E,$7C,$71,$FD,$1E,$5E,$27,$EB,$EB,$E0,$36
	.dc.b	$7C,$27,$F4,$DF,$5B,$F8,$23,$77,$D2,$8C,$BE,$03,$F9,$0B,$F7,$5E
	.dc.b	$07,$F6,$5C,$1E,$73,$CD,$F0,$46,$1E,$B5,$D2,$3F,$C3,$3E,$F3,$E1
	.dc.b	$FC,$3F,$80,$3C,$7A,$79,$27,$59,$7F,$65,$6B,$E5,$13,$C5,$FE,$D4
	.dc.b	$D7,$51,$C2,$FF,$C5,$3E,$A3,$D7,$7A,$FF,$E0,$58,$C3,$7D,$7F,$9B
	.dc.b	$87,$F8,$2E,$72,$FF,$83,$7D,$F7,$9F,$FE,$34,$8F,$E5,$31,$8E,$3F
	.dc.b	$F0,$83,$63,$84,$ED,$FE,$9B,$F4,$0F,$1F,$10,$8C,$5F,$F8,$65,$B1
	.dc.b	$FE,$1D,$6B,$FE,$77,$FD,$03,$F0,$73,$FE,$1F,$6C,$7F,$88,$DA,$EF
	.dc.b	$9B,$BC,$6B,$6F,$F2,$6B,$6B,$FC,$95,$FA,$A7,$C7,$7F,$C2,$D1,$9B
	.dc.b	$FE,$CA,$D4,$E6,$95,$C6,$7E,$C4,$EF,$E1,$DF,$03,$F0,$1B,$87,$F8
	.dc.b	$DD,$B5,$FE,$3B,$6B,$FE,$95,$73,$87,$F9,$91,$B3,$E8,$DF,$FC,$8D
	.dc.b	$18,$AF,$EF,$6D,$4E,$6B,$CF,$3C,$2F,$8F,$9D,$E5,$26,$1E,$56,$E1
	.dc.b	$F9,$3C,$11,$FE,$5E,$C9,$F1,$1C,$AF,$FC,$7D,$5A,$6D,$6E,$75,$7C
	.dc.b	$FC,$8F,$88,$4A,$F8,$8D,$8D,$47,$EC,$F0,$7C,$F7,$C0,$A9,$3E,$07
	.dc.b	$C9,$18,$AF,$47,$99,$F4,$49,$5F,$45,$B1,$A8,$EC,$7C,$C8,$8F,$B1
	.dc.b	$F8,$0E,$23,$E2,$7C,$17,$DD,$24,$7D,$D6,$C6,$A3,$F8,$67,$73,$E7
	.dc.b	$FE,$CD,$49,$F9,$2F,$FA,$2A,$75,$18,$3F,$C4,$DF,$01,$9F,$81,$FF
	.dc.b	$16,$B6,$32,$76,$8C,$B4,$CE,$1F,$C8,$78,$BF,$8E,$CE,$A3,$16,$8C
	.dc.b	$4F,$FA,$D6,$31,$5C,$A3,$13,$A8,$FF,$5A,$4E,$AB,$F8,$D7,$96,$C3
	.dc.b	$CD,$18,$6E,$1F,$EC,$09,$D5,$F9,$EB,$4D,$9D,$57,$08,$FF,$1D,$D7
	.dc.b	$0A,$F4,$E1,$E3,$7F,$C7,$D8,$78,$F0,$E3,$7C,$CE,$38,$73,$AF,$F3
	.dc.b	$1E,$1C,$EB,$DB,$5E,$F3,$DB,$87,$58,$F9,$31,$9A,$EF,$39,$C3,$C1
	.dc.b	$3D,$8F,$04,$F6,$C3,$C0,$B0,$F0,$4F,$63,$A6,$1D,$A7,$E1,$9E,$E8
	.dc.b	$CD,$F6,$8E,$91,$D9,$E9,$7F,$CC,$2F,$E3,$47,$F9,$88,$CC,$FF,$B2
	.dc.b	$63,$31,$E7,$3B,$DF,$E0,$7E,$1B,$E6,$47,$03,$35,$F8,$08,$CD,$7A
	.dc.b	$A7,$C6,$7E,$32,$33,$7F,$15,$B5,$EA,$6C,$F4,$3E,$CB,$8B,$FB,$31
	.dc.b	$CE,$7E,$83,$E3,$8C,$9E,$78,$CD,$F8,$E7,$51,$E7,$33,$7E,$36,$9E
	.dc.b	$32,$C7,$8E,$BF,$C0,$77,$E3,$9F,$8D,$1D,$E7,$D0,$76,$3E,$0B,$93
	.dc.b	$CB,$1C,$09,$1E,$E7,$BC,$6A,$30,$74,$BF,$51,$F3,$63,$C7,$3C,$0E
	.dc.b	$6E,$27,$85,$FB,$5B,$3E,$D6,$9E,$DC,$18,$7B,$63,$D3,$3F,$1D,$DD
	.dc.b	$F9,$4E,$C7,$9D,$F8,$07,$C1,$73,$1E,$E8,$CE,$1E,$9B,$EF,$3E,$99
	.dc.b	$EE,$7C,$0A,$E9,$5F,$03,$0F,$E4,$98,$79,$6B,$B6,$1C,$9F,$2E,$1F
	.dc.b	$0A,$31,$87,$2B,$F8,$4D,$9F,$84,$DA,$F8,$57,$D7,$0F,$33,$6F,$C3
	.dc.b	$7E,$7A,$F0,$CE,$63,$F9,$EB,$63,$F2,$C7,$90,$F6,$DF,$C4,$7D,$E7
	.dc.b	$C4,$AE,$AE,$CE,$B5,$DE,$FC,$6D,$9D,$D7,$A4,$FA,$98,$7C,$09,$FA
	.dc.b	$15,$F2,$5F,$29,$C4,$B7,$DE,$3A,$B6,$FA,$CF,$27,$E4,$B4,$F9,$25
	.dc.b	$8F,$92,$F5,$7E,$0D,$F9,$63,$EA,$61,$F1,$63,$F4,$91,$C5,$F4,$C6
	.dc.b	$63,$F2,$11,$C6,$7E,$4E,$1E,$D7,$E4,$E1,$ED,$7B,$D7,$33,$B9,$E3
	.dc.b	$AE,$F1,$C3,$0F,$5B,$67,$D6,$D3,$D6,$7C,$37,$E3,$CF,$8E,$3E,$A9
	.dc.b	$93,$C2,$71,$7D,$71,$E1,$3D,$67,$96,$3D,$78,$7B,$DF,$D0,$46,$5E
	.dc.b	$B3,$DE,$FE,$AB,$6F,$EA,$CF,$7A,$F7,$BE,$39,$F7,$BF,$8D,$8F,$6B
	.dc.b	$4F,$6C,$73,$8F,$69,$E1,$AE,$ED,$AE,$F5,$C6,$BD,$AF,$79,$F6,$BE
	.dc.b	$33,$DA,$FC,$17,$DA,$64,$F6,$BD,$4F,$0B,$9A,$F0,$B6,$BC,$31,$F7
	.dc.b	$EB,$C2,$7A,$67,$E8,$9E,$5A,$F8,$71,$FC,$57,$89,$E8,$FF,$9C,$C9
	.dc.b	$3F,$E7,$36,$D7,$F9,$7D,$FF,$19,$7D,$07,$F9,$B9,$93,$FE,$5E,$72
	.dc.b	$D1,$3F,$EB,$9F,$D0,$5E,$9F,$FA,$81,$B3,$FD,$C6,$77,$18,$70,$D7
	.dc.b	$FB,$B3,$9C,$53,$DA,$53,$98,$FF,$18,$F6,$EA,$7C,$57,$6F,$11,$C2
	.dc.b	$31,$5C,$EF,$94,$FF,$8A,$2F,$2F,$9A,$AB,$E0,$FA,$1F,$61,$2F,$D9
	.dc.b	$19,$6D,$A7,$8B,$63,$FC,$C7,$EE,$E2,$64,$7C,$AC,$7C,$F5,$BB,$FF
	.dc.b	$2A,$DE,$9F,$9D,$E3,$FC,$D2,$5F,$81,$B1,$A8,$FF,$38,$FA,$7E,$29
	.dc.b	$91,$DC,$F8,$4F,$F9,$C6,$33,$3F,$6F,$E7,$7E,$D1,$2B,$DA,$D8,$D4
	.dc.b	$7F,$9F,$BD,$CF,$7C,$C8,$ED,$3F,$80,$8F,$F3,$F4,$E6,$3F,$11,$EB
	.dc.b	$FE,$1C,$95,$F2,$9B,$1A,$8F,$F4,$5F,$1F,$CC,$99,$1D,$A7,$F0,$B1
	.dc.b	$FE,$8B,$9C,$C7,$F9,$D3,$CB,$FF,$22,$12,$BF,$C8,$4D,$8D,$47,$A3
	.dc.b	$57,$BE,$D7,$F0,$7C,$AC,$29,$2F,$85,$6D,$F2,$4E,$DF,$37,$6F,$E7
	.dc.b	$F1,$FC,$E3,$5F,$D5,$B5,$34,$FC,$BB,$8E,$2F,$CB,$E4,$D5,$D5,$F9
	.dc.b	$8B,$DE,$2F,$4B,$D2,$7A,$B1,$B9,$79,$A2,$C7,$FA,$68,$B8,$FE,$A8
	.dc.b	$7A,$1F,$A4,$6D,$7D,$19,$FE,$07,$EB,$FF,$04,$F5,$5E,$9F,$EE,$6F
	.dc.b	$74,$7A,$61,$CC,$FD,$7C,$72,$15,$C9,$FD,$34,$69,$5F,$7B,$ED,$3F
	.dc.b	$C0,$EB,$85,$7F,$09,$AD,$3C,$B0,$AE,$4E,$F7,$CA,$33,$1E,$3A,$E7
	.dc.b	$87,$A9,$EB,$19,$AE,$0F,$23,$FD,$21,$7D,$AB,$84,$FF,$23,$8C,$61
	.dc.b	$E7,$8C,$57,$13,$E3,$46,$5E,$66,$6B,$9B,$8B,$F2,$CE,$E3,$D8,$7F
	.dc.b	$16,$BF,$E2,$EF,$CD,$30,$7F,$35,$F5,$5E,$B3,$D7,$FE,$3B,$18,$6C
	.dc.b	$66,$78,$35,$FF,$57,$4B,$D7,$FB,$D1,$D5,$AE,$9A,$FF,$3D,$8E,$19
	.dc.b	$78,$27,$8E,$5F,$6B,$3F,$CD,$5F,$D8,$6B,$F1,$EF,$8B,$6B,$8C,$FF
	.dc.b	$49,$37,$5C,$F5,$FE,$98,$FE,$1A,$F8,$B6,$B8,$DF,$F5,$EB,$E2,$73
	.dc.b	$BE,$2F,$A6,$77,$2B,$E7,$3B,$8F,$45,$70,$7F,$B4,$CE,$9B,$5C,$AB
	.dc.b	$BC,$78,$6B,$83,$D6,$75,$1F,$AD,$D7,$B4,$FE,$B5,$B8,$FF,$62,$C3
	.dc.b	$B4,$FE,$D2,$B8,$3E,$DF,$36,$BF,$B8,$6E,$3F,$3A,$FF,$B5,$FA,$DF
	.dc.b	$C0,$AB,$F9,$83,$5F,$01,$FE,$04,$71,$1C,$32,$EB,$18,$CB,$FB,$07
	.dc.b	$88,$F5,$FC,$76,$8C,$61,$DB,$87,$CB,$EB,$3B,$FF,$36,$D9,$FE,$E6
	.dc.b	$D7,$FC,$20,$D7,$FC,$14,$1D,$5F,$F8,$1B,$D7,$FE,$79,$3F,$D2,$DA
	.dc.b	$FB,$0E,$15,$F1,$BE,$1D,$F1,$1F,$AC,$6C,$FF,$81,$1A,$FF,$88,$5A
	.dc.b	$F8,$BD,$77,$B2,$F6,$7F,$C1,$71,$86,$FA,$AE,$4C,$D6,$EF,$FC,$59
	.dc.b	$EC,$BC,$AF,$F0,$FC,$7F,$89,$E3,$1C,$7F,$E1,$66,$C7,$09,$DB,$FE
	.dc.b	$3A,$F5,$FB,$57,$F6,$1F,$51,$F1,$DF,$E2,$29,$CD,$7D,$EE,$25,$6A
	.dc.b	$FF,$B9,$56,$AB,$E7,$B5,$FF,$1D,$35,$F0,$67,$E2,$51,$DA,$FF,$C7
	.dc.b	$2D,$AF,$F1,$35,$FC,$87,$BD,$FF,$8D,$BD,$07,$B1,$FD,$92,$9F,$23
	.dc.b	$98,$C3,$6F,$FC,$73,$5C,$1B,$3E,$38,$C6,$5F,$E5,$96,$E5,$FE,$4C
	.dc.b	$6D,$79,$37,$E9,$5F,$87,$97,$B3,$E3,$E5,$F3,$36,$FF,$C3,$13,$A7
	.dc.b	$C7,$1E,$12,$38,$36,$FB,$7D,$9E,$71,$CB,$C6,$F8,$15,$F0,$DB,$19
	.dc.b	$3D,$EF,$B4,$8E,$0D,$61,$F7,$75,$7B,$3C,$E7,$CF,$EC,$FC,$DF,$92
	.dc.b	$FA,$0D,$8C,$9F,$31,$F9,$44,$70,$6A,$FB,$3C,$8F,$87,$E5,$7D,$9F
	.dc.b	$4B,$EE,$36,$32,$7D,$87,$EB,$11,$C1,$A9,$FC,$B7,$80,$F6,$B9,$CF
	.dc.b	$F8,$D3,$D2,$FF,$14,$CB,$DA,$FF,$1E,$C6,$1B,$3F,$E4,$39,$D3,$C4
	.dc.b	$C4,$6A,$FB,$B7,$93,$FD,$25,$5C,$27,$F8,$57,$9C,$C3,$B6,$1C,$E7
	.dc.b	$8B,$6F,$8E,$1C,$E3,$F9,$3B,$67,$FC,$55,$87,$5A,$FF,$4F,$E1,$D6
	.dc.b	$FC,$A7,$33,$89,$6B,$8D,$F3,$AF,$E6,$0D,$BE,$95,$FE,$B1,$BE,$97
	.dc.b	$E6,$3D,$78,$7A,$9B,$3E,$06,$9B,$C3,$CB,$3B,$6C,$ED,$B3,$B8,$F0
	.dc.b	$4E,$DA,$6C,$B1,$B3,$FD,$A7,$5E,$6A,$F9,$33,$EF,$7D,$AD,$BF,$6D
	.dc.b	$70,$BE,$CD,$9E,$CD,$3B,$0B,$ED,$1F,$32,$3D,$F7,$E0,$6C,$66,$3D
	.dc.b	$A6,$4A,$E4,$F5,$B9,$69,$98,$F1,$46,$5B,$19,$7C,$95,$EA,$6C,$7A
	.dc.b	$9A,$7A,$AB,$D2,$FE,$0A,$3E,$26,$1F,$81,$23,$82,$47,$B0,$E0,$5B
	.dc.b	$F9,$AD,$C3,$E6,$BE,$08,$F4,$36,$7D,$17,$E2,$9E,$27,$C4,$9E,$AD
	.dc.b	$33,$3F,$12,$32,$D8,$CC,$7C,$43,$25,$72,$F8,$4C,$95,$CC,$F8,$EB
	.dc.b	$D2,$DB,$F8,$75,$ED,$6D,$79,$6B,$78,$7D,$1B,$E1,$87,$C8,$AD,$5F
	.dc.b	$08,$F4,$5F,$07,$EB,$57,$0C,$3B,$47,$13,$75,$C6,$3D,$95,$C5,$FB
	.dc.b	$98,$78,$49,$5D,$1B,$1A,$8C,$11,$C1,$30,$FA,$F8,$78,$DF,$4C,$F8
	.dc.b	$E7,$C3,$87,$8D,$F9,$58,$78,$EB,$8E,$1E,$06,$DF,$7A,$FF,$19,$E1
	.dc.b	$FC,$31,$B5,$FE,$03,$8C,$47,$EF,$8C,$CF,$F8,$E9,$FD,$A1,$83,$11
	.dc.b	$F9,$C7,$C7,$ED,$7F,$27,$7F,$00,$62,$7F,$91,$38,$9F,$C0,$4E,$DE
	.dc.b	$46,$27,$6E,$2B,$FA,$4C,$ED,$FE,$50,$D9,$FE,$1C,$DA,$FD,$23,$6F
	.dc.b	$F1,$0D,$9F,$C1,$B6,$7E,$EB,$5F,$41,$FB,$13,$D1,$5E,$29,$F2,$BE
	.dc.b	$26,$9E,$22,$C7,$89,$B5,$E2,$69,$E2,$2C,$78,$9A,$F8,$A3,$F6,$CF
	.dc.b	$8A,$3F,$B3,$57,$9E,$FE,$44,$F9,$DF,$91,$3A,$39,$BA,$AE,$6E,$A7
	.dc.b	$FB,$CD,$73,$BF,$74,$73,$7D,$CD,$BF,$73,$4F,$71,$6F,$DC,$D8,$F7
	.dc.b	$34,$F7,$14,$F7,$3F,$C3,$DF,$71,$E4,$7D,$33,$E4,$9F,$4B,$E4,$6C
	.dc.b	$79,$1B,$3E,$46,$9E,$42,$C7,$91,$B1,$E4,$6B,$E4,$3F,$8F,$1E,$47
	.dc.b	$FC,$1F,$5F,$1E,$7E,$74,$FA,$EF,$E7,$47,$AD,$F9,$D5,$C2,$3D,$67
	.dc.b	$07,$BC,$70,$9E,$EF,$09,$FF,$5C,$D7,$79,$F8,$87,$7B,$F8,$93,$DD
	.dc.b	$F8,$8F,$CB,$E2,$7B,$3F,$E7,$7F,$0F,$FC,$F4,$D9,$FF,$32,$BF,$E7
	.dc.b	$32,$3F,$E7,$86,$DF,$F9,$BD,$FF,$1D,$FB,$5C,$A5,$3F,$EB,$9F,$65
	.dc.b	$EA,$7F,$A1,$7E,$79,$5A,$DA,$DF,$0B,$DC,$FF,$A8,$EB,$35,$AA,$56
	.dc.b	$9F,$E2,$8D,$7F,$D9,$ED,$7C,$67,$0B,$FF,$50,$47,$19,$E8,$DB,$FF
	.dc.b	$72,$F8,$5E,$07,$8C,$66,$FC,$DE,$DF,$FA,$7F,$DA,$D4,$62,$BF,$D7
	.dc.b	$51,$88,$F3,$CF,$80,$93,$FC,$AE,$32,$DA,$4F,$16,$C7,$F9,$8F,$DA
	.dc.b	$C4,$E7,$BF,$FD,$61,$F1,$F6,$BF,$D9,$4D,$C3,$E2,$CF,$C0,$24,$F7
	.dc.b	$75,$9F,$F9,$A3,$DC,$B3,$F0,$3A,$BF,$D7,$9F,$39,$E3,$6C,$7D,$93
	.dc.b	$47,$A5,$F9,$A4,$70,$6F,$0F,$BD,$1F,$E8,$2F,$6F,$A4,$E6,$7E,$EE
	.dc.b	$DF,$EC,$BF,$97,$F8,$4D,$9D,$6E,$8F,$B6,$48,$F9,$4D,$AF,$C9,$3F
	.dc.b	$E9,$0E,$7E,$73,$DF,$A5,$EF,$7D,$AD,$9E,$F7,$FB,$E1,$FF,$3E,$C6
	.dc.b	$63,$15,$FE,$75,$8F,$ED,$04,$7E,$E5,$74,$6C,$7E,$0E,$31,$8F,$FB
	.dc.b	$EE,$9C,$79,$AF,$C4,$78,$9F,$8F,$C5,$B3,$D7,$CF,$3F,$9E,$31,$A7
	.dc.b	$C9,$BF,$D0,$67,$FB,$26,$A6,$98,$D1,$EC,$F9,$4F,$34,$DA,$F8,$D1
	.dc.b	$63,$FD,$56,$5C,$BF,$68,$3D,$6F,$D0,$36,$BE,$E4,$FF,$04,$F7,$BF
	.dc.b	$82,$FB,$8F,$6F,$F6,$B7,$BA,$3D,$30,$E6,$7E,$AE,$39,$0A,$E4,$FF
	.dc.b	$0B,$8D,$2B,$EF,$7D,$A7,$F8,$7D,$70,$AF,$E0,$B5,$A7,$96,$15,$C9
	.dc.b	$DE,$F9,$46,$63,$C7,$5C,$F0,$F5,$3D,$63,$35,$C1,$E4,$7F,$14,$3C
	.dc.b	$C6,$0F,$83,$87,$9A,$3F,$C6,$18,$78,$EF,$AC,$62,$B8,$46,$1E,$93
	.dc.b	$BA,$E5,$18,$36,$E1,$F6,$9F,$C4,$E3,$F8,$A3,$F3,$8C,$1F,$E4,$1F
	.dc.b	$71,$EE,$BD,$EF,$E3,$51,$86,$C6,$67,$83,$5F,$81,$3F,$9B,$C3,$8F
	.dc.b	$0F,$E0,$27,$B3,$5D,$B5,$FE,$77,$3C,$5F,$91,$97,$E6,$9F,$B0,$D7
	.dc.b	$E6,$46,$F3,$FC,$03,$67,$FA,$29,$B8,$FC,$23,$4E,$F5,$E5,$8E,$2D
	.dc.b	$8E,$3A,$FF,$4E,$9E,$2D,$8E,$38,$7E,$2E,$FC,$18,$7F,$66,$9E,$CD
	.dc.b	$AE,$0F,$F6,$A9,$DA,$6C,$F0,$8F,$D5,$4F,$07,$58,$79,$5E,$87,$F2
	.dc.b	$D9,$F8,$97,$FD,$AD,$BA,$F9,$A7,$E5,$DF,$ED,$23,$E2,$FA,$76,$7D
	.dc.b	$0D,$7C,$27,$F3,$9F,$7B,$F8,$13,$77,$D4,$62,$3F,$7B,$5F,$DD,$23
	.dc.b	$F8,$E4,$6C,$FE,$8D,$8B,$C8,$7B,$DE,$18,$C3,$D3,$0F,$17,$BC,$F0
	.dc.b	$69,$30,$FE,$99,$EF,$3E,$2F,$11,$FE,$9B,$AE,$0F,$F7,$36,$B9,$AF
	.dc.b	$D1,$FA,$CF,$88,$A4,$9F,$EF,$51,$97,$FB,$E7,$C0,$3C,$DF,$D5,$B6
	.dc.b	$33,$18,$3F,$5F,$F0,$0F,$7D,$F0,$1F,$E0,$58,$C3,$70,$72,$66,$B7
	.dc.b	$87,$FA,$D7,$DE,$CF,$2F,$CA,$F6,$9E,$FF,$F8,$4E,$73,$5F,$48,$FE
	.dc.b	$5F,$3A,$AF,$EE,$53,$AE,$5F,$F1,$8C,$E6,$B8,$09,$D5,$7F,$8E,$E7
	.dc.b	$55,$FE,$37,$9E,$F5,$FC,$7D,$AF,$F8,$D1,$AF,$F9,$21,$AB,$D4,$7B
	.dc.b	$DF,$DD,$6D,$EF,$FF,$8B,$67,$38,$FB,$E3,$15,$B6,$C6,$4F,$F0,$74
	.dc.b	$F0,$BF,$F6,$E3,$5F,$E6,$8D,$7F,$C9,$2D,$7F,$CB,$AB,$DF,$F2,$72
	.dc.b	$78,$DB,$5D,$9B,$5F,$22,$B8,$4F,$D5,$AE,$0F,$F9,$0E,$33,$3E,$4A
	.dc.b	$FF,$30,$3E,$26,$BE,$2F,$17,$E2,$4F,$C3,$6D,$7B,$9B,$3E,$16,$E9
	.dc.b	$FE,$6B,$AF,$F0,$25,$78,$67,$AD,$7D,$16,$CE,$A3,$E3,$FB,$9F,$4F
	.dc.b	$DD,$FA,$4D,$BF,$9A,$D9,$F8,$0D,$C3,$C0,$F4,$24,$FC,$17,$BF,$BB
	.dc.b	$F7,$FD,$2F,$BC,$D9,$FB,$2D,$BF,$98,$DB,$F6,$3C,$A7,$E6,$36,$7F
	.dc.b	$B8,$0F,$25,$E7,$FF,$90,$7C,$17,$B9,$FE,$37,$8C,$37,$0F,$E2,$B3
	.dc.b	$A7,$FC,$95,$3A,$69,$A9,$E1,$7F,$AF,$C7,$F8,$E4,$61,$FF,$4E,$DF
	.dc.b	$39,$FD,$6E,$1E,$4B,$E4,$74,$8E,$53,$BA,$FF,$52,$56,$E7,$37,$D6
	.dc.b	$FF,$D5,$D7,$B7,$FD,$47,$5B,$BF,$0D,$6E,$FF,$C6,$37,$B7,$FD,$77
	.dc.b	$5B,$BE,$75,$BB,$F5,$C7,$95,$E6,$D3,$65,$AE,$B3,$C6,$FB,$47,$D3
	.dc.b	$C3,$E5,$3F,$09,$B7,$EE,$6D,$FA,$9B,$3E,$66,$BC,$E3,$E4,$BC,$DB
	.dc.b	$3C,$DA,$73,$2C,$73,$6B,$CC,$F9,$AF,$AE,$3E,$0C,$75,$32,$57,$E1
	.dc.b	$B6,$3E,$1B,$4F,$86,$F3,$AF,$87,$87,$CF,$9C,$9B,$AE,$E7,$A5,$EE
	.dc.b	$D3,$B9,$6B,$BB,$63,$9B,$4E,$71,$F0,$63,$9B,$63,$9B,$4E,$65,$8E
	.dc.b	$6D,$8E,$71,$F2,$4E,$65,$79,$B6,$39,$B4,$E6,$58,$E6,$FD,$17,$8B
	.dc.b	$63,$8B,$4E,$25,$8E,$33,$F2,$DB,$7F,$86,$8C,$5F,$D0,$8C,$12,$FF
	.dc.b	$03,$3A,$C3,$F3,$04,$8F,$B0,$DC,$3E,$B5,$F3,$9E,$37,$CE,$7E,$19
	.dc.b	$E5,$9D,$38,$BE,$D5,$D0,$91,$D2,$FB,$61,$F6,$6F,$8B,$F1,$2F,$8D
	.dc.b	$6E,$F8,$CE,$EF,$8C,$7D,$7B,$E2,$FC,$AB,$E3,$5B,$8E,$26,$EB,$8D
	.dc.b	$FD,$5B,$E2,$FD,$0B,$E3,$5B,$38,$C6,$EB,$8B,$F0,$5B,$3C,$23,$53
	.dc.b	$C1,$D3,$63,$51,$83,$DC,$4B,$F4,$36,$7C,$4D,$3A,$94,$E5,$1C,$09
	.dc.b	$1C,$1B,$1C,$1A,$F0,$8D,$11,$D3,$5D,$18,$21,$82,$4F,$B0,$98,$7E
	.dc.b	$1E,$BE,$E3,$6B,$EB,$36,$BE,$83,$67,$F3,$B3,$F9,$02,$4F,$E3,$67
	.dc.b	$58,$7D,$8A,$DC,$E6,$B7,$1E,$06,$9E,$02,$C7,$81,$B1,$C8,$E1,$1C
	.dc.b	$C8,$E0,$91,$CE,$32,$D8,$CC,$75,$32,$57,$27,$78,$E0,$48,$D3,$63
	.dc.b	$51,$83,$A1,$FE,$F1,$BF,$ED,$0D,$BF,$F1,$0C,$62,$3F,$8F,$1F,$B7
	.dc.b	$F7,$3F,$B3,$3F,$C3,$CC,$5F,$F5,$D7,$13,$FC,$32,$76,$FF,$3B,$8D
	.dc.b	$9F,$D0,$4D,$D7,$F8,$46,$76,$FF,$65,$6D,$7F,$45,$6C,$FF,$0E,$6D
	.dc.b	$FF,$03,$6C,$FE,$F5,$B3,$FB,$76,$BE,$29,$FE,$88,$F8,$A7,$FC,$3B
	.dc.b	$3A,$BF,$54,$7B,$5F,$51,$89,$F5,$6C,$F7,$3F,$CE,$DD,$BF,$E7,$56
	.dc.b	$DF,$F9,$C1,$B5,$FE,$63,$6E,$3F,$E5,$37,$FC,$9E,$49,$FF,$32,$3F
	.dc.b	$E5,$4F,$51,$E8,$7F,$A0,$FD,$B5,$30,$FF,$4A,$D6,$A3,$FD,$45,$3A
	.dc.b	$9C,$D7,$19,$E1,$86,$65,$7C,$5F,$F1,$C7,$4B,$0E,$55,$BA,$D5,$78
	.dc.b	$4F,$13,$87,$0D,$73,$87,$AF,$DA,$FF,$4E,$7A,$DD,$32,$DC,$61,$F2
	.dc.b	$98,$3F,$D7,$91,$BA,$EF,$3B,$7E,$0B,$D6,$7F,$CB,$F1,$EC,$8C,$3F
	.dc.b	$E6,$2B,$FF,$74,$4F,$FB,$1F,$DA,$E9,$3F,$0B,$AB,$FD,$61,$18,$AD
	.dc.b	$CF,$C4,$8C,$4E,$AB,$CF,$3F,$0A,$3E,$C1,$23,$C7,$5C,$A7,$E4,$1F
	.dc.b	$43,$9B,$FD,$7D,$EC,$F7,$6D,$74,$9D,$3F,$ED,$D9,$D3,$4D,$52,$BF
	.dc.b	$10,$48,$F0,$BF,$8B,$CB,$FD,$99,$EC,$FB,$5B,$3C,$9B,$87,$E3,$1B
	.dc.b	$AF,$E0,$5E,$53,$EC,$6C,$FF,$A3,$B8,$DE,$BF,$D7,$B7,$B2,$F6,$FE
	.dc.b	$7D,$6B,$1F,$93,$5A,$9C,$DF,$1B,$E1,$5F,$0E,$F8,$61,$FE,$8D,$9C
	.dc.b	$BF,$E8,$AA,$FE,$36,$4A,$FC,$1D,$70,$9F,$A3,$DD,$F9,$FC,$7F,$3B
	.dc.b	$EC,$BD,$4F,$C1,$CE,$70,$FD,$3C,$E5,$FA,$91,$9B,$5A,$B5,$18,$AF
	.dc.b	$D3,$56,$EB,$57,$F0,$BD,$39,$5E,$DB,$A7,$FA,$92,$2D,$7F,$A8,$22
	.dc.b	$CF,$EC,$61,$E8,$FE,$69,$B1,$F8,$99,$FE,$0D,$EC,$FF,$07,$F5,$DE
	.dc.b	$B7,$1C,$38,$12,$F8,$5A,$F8,$56,$E7,$31,$89,$FE,$17,$5F,$C3,$EF
	.dc.b	$AC,$7E,$7A,$74,$FC,$69,$D5,$FF,$B5,$27,$38,$7A,$29,$7C,$9F,$E1
	.dc.b	$D3,$AB,$E3,$1D,$4D,$DF,$95,$A6,$6B,$95,$F5,$7A,$47,$C4,$36,$79
	.dc.b	$70,$F3,$B9,$8E,$37,$E0,$8F,$19,$C2,$BA,$5F,$AE,$31,$3B,$7B,$1F
	.dc.b	$C5,$30,$FE,$2A,$E0,$E0,$7F,$04,$F5,$DE,$C3,$D9,$FE,$37,$18,$6C
	.dc.b	$66,$7F,$DF,$B3,$96,$EB,$F9,$46,$E5,$CA,$31,$1D,$DA,$FF,$3B,$9F
	.dc.b	$D1,$4F,$D8,$C9,$AF,$F3,$FD,$3E,$A4,$62,$FE,$ED,$F0,$7E,$5B,$67
	.dc.b	$75,$C6,$BF,$A4,$1C,$6B,$78,$79,$EB,$F0,$17,$C6,$B7,$7C,$6F,$FB
	.dc.b	$04,$71,$AD,$DF,$18,$F5,$36,$77,$5C,$5F,$D0,$B6,$FB,$47,$F6,$B6
	.dc.b	$BA,$3E,$14,$6E,$7A,$46,$4F,$DF,$1D,$8F,$E7,$15,$ED,$C3,$FB,$75
	.dc.b	$F4,$8C,$6B,$F0,$BC,$FF,$DD,$36,$F8,$E3,$87,$FB,$F7,$B3,$FB,$E7
	.dc.b	$FA,$23,$5F,$08,$71,$8F,$F7,$D7,$F8,$E4,$70,$DF,$A4,$BC,$67,$B3
	.dc.b	$86,$D6,$CF,$EA,$5F,$A3,$3E,$33,$8F,$88,$F9,$3A,$79,$E9,$5B,$78
	.dc.b	$D6,$9F,$E7,$F5,$C5,$C9,$FD,$7F,$F4,$57,$CB,$FE,$99,$B1,$FD,$F1
	.dc.b	$AF,$F6,$A0,$E1,$FE,$DA,$EB,$3F,$F0,$CD,$5D,$FF,$C4,$2D,$AF,$F1
	.dc.b	$55,$57,$9D,$FB,$88,$F9,$5F,$77,$B2,$F5,$DE,$CF,$F8,$42,$30,$DF
	.dc.b	$65,$C9,$9A,$DE,$1E,$9F,$85,$FE,$B2,$FF,$81,$1A,$FF,$8B,$5A,$FF
	.dc.b	$8F,$DA,$BE,$19,$F0,$FF,$E3,$B8,$FF,$1A,$46,$36,$FF,$11,$B6,$3E
	.dc.b	$43,$5F,$F0,$9F,$D5,$3E,$5B,$94,$7F,$93,$63,$15,$FE,$2C,$6C,$7D
	.dc.b	$16,$BF,$E1,$CF,$AB,$78,$5C,$A1,$18,$AF,$F1,$AB,$63,$ED,$35,$FB
	.dc.b	$3E,$E7,$1E,$23,$49,$5F,$83,$F3,$B4,$47,$07,$E1,$FD,$5E,$44,$70
	.dc.b	$7E,$2F,$DB,$EC,$47,$07,$E3,$FC,$EF,$11,$1C,$1F,$93,$F5,$BC,$C4
	.dc.b	$70,$7E,$5F,$D5,$F5,$11,$C1,$F9,$BF,$3F,$DC,$47,$07,$E7,$FD,$3F
	.dc.b	$84,$47,$07,$E8,$FD,$3F,$8C,$47,$07,$E9,$FD,$2F,$94,$47,$07,$EA
	.dc.b	$F9,$FE,$71,$1C,$1F,$AF,$E9,$FA,$44,$70,$7D,$AF,$6F,$FC,$D3,$F1
	.dc.b	$6D,$FF,$CD,$7C,$89,$CC,$61,$FC,$24,$66,$B9,$13,$93,$F0,$73,$B8
	.dc.b	$FF,$31,$76,$7F,$0F,$7F,$CC,$95,$C6,$BF,$90,$B7,$C0,$FF,$4E,$36
	.dc.b	$7F,$D2,$D5,$C6,$B7,$7D,$5B,$3D,$BC,$3C,$61,$CA,$AC,$F2,$9B,$5F
	.dc.b	$EB,$1C,$3B,$36,$BF,$D1,$78,$77,$C3,$96,$1E,$89,$E3,$5F,$C2,$30
	.dc.b	$B7,$BA,$FF,$56,$61,$D2,$B5,$1C,$4F,$31,$5F,$E8,$33,$C5,$F1,$46
	.dc.b	$9F,$69,$AA,$EE,$48,$E5,$87,$FB,$0E,$B9,$57,$EE,$30,$F5,$DE,$E3
	.dc.b	$9D,$6E,$30,$FB,$A3,$B5,$F4,$3E,$41,$FE,$B4,$BF,$A5,$1F,$EA,$26
	.dc.b	$BA,$3F,$9E,$3E,$23,$C7,$93,$93,$C7,$5E,$DB,$F1,$3F,$EA,$79,$D3
	.dc.b	$93,$A4,$6A,$FF,$B3,$CE,$9F,$39,$23,$47,$FB,$96,$34,$E2,$79,$92
	.dc.b	$7F,$DE,$53,$A8,$C4,$7D,$83,$A3,$FE,$E7,$8D,$47,$FB,$E6,$FD,$D8
	.dc.b	$7B,$23,$CB,$85,$C3,$C1,$1B,$9F,$D7,$CE,$70,$EF,$7F,$1A,$31,$5B
	.dc.b	$9F,$84,$F9,$A3,$2F,$96,$FA,$AC,$3C,$AE,$6B,$AE,$1E,$07,$9C,$64
	.dc.b	$EC,$7A,$1C,$E1,$EC,$7E,$A6,$1E,$08,$C4,$EC,$F8,$A7,$2C,$3C,$78
	.dc.b	$78,$A7,$71,$8A,$E5,$87,$AA,$B4,$E3,$0E,$E7,$89,$D4,$E7,$0E,$E7
	.dc.b	$BD,$C5,$F5,$AE,$07,$BA,$FA,$B8,$AF,$44,$FC,$8C,$3C,$B3,$F7,$6F
	.dc.b	$9B,$F8,$2B,$E3,$7E,$0A,$FA,$D3,$CE,$FC,$13,$9C,$3C,$F3,$93,$96
	.dc.b	$1D,$27,$E3,$47,$73,$4E,$0F,$BC,$E1,$FA,$C7,$BD,$C5,$F8,$A3,$C1
	.dc.b	$1E,$F9,$ED,$1C,$30,$F4,$E1,$EF,$C3,$F0,$53,$C0,$F9,$73,$AB,$F2
	.dc.b	$46,$2F,$D6,$D8,$EC,$47,$04,$BF,$86,$48,$F8,$26,$DC,$47,$BC,$F6
	.dc.b	$1C,$63,$6D,$35,$5E,$57,$E2,$99,$3D,$D5,$E6,$AE,$B1,$8A,$F0,$1D
	.dc.b	$23,$38,$7C,$96,$E1,$F1,$1B,$3F,$92,$AE,$D5,$B8,$E4,$48,$F2,$DF
	.dc.b	$86,$31,$1E,$13,$8D,$F5,$8C,$D7,$39,$E3,$1E,$66,$9C,$1F,$15,$F9
	.dc.b	$23,$F3,$31,$CA,$BA,$47,$09,$D1,$FC,$86,$FC,$91,$9A,$E0,$74,$2B
	.dc.b	$D0,$D1,$0F,$54,$F4,$8F,$C2,$B6,$3E,$9B,$5F,$B6,$DC,$3F,$9B,$40
	.dc.b	$FC,$69,$4F,$B6,$53,$F0,$65,$7F,$8D,$9F,$CC,$21,$E0,$BD,$6F,$83
	.dc.b	$5E,$0B,$F2,$9F,$E8,$68,$CD,$F2,$C1,$3F,$E9,$A9,$CC,$7C,$AC,$3A
	.dc.b	$C6,$27,$E3,$5A,$73,$18,$8F,$01,$F0,$9F,$01,$9B,$F0,$3D,$CC,$5F
	.dc.b	$7C,$3D,$51,$83,$BD,$F6,$BF,$A8,$76,$3B,$B9,$3E,$F4,$71,$8F,$F6
	.dc.b	$DC,$66,$3E,$1C,$79,$49,$19,$70,$75,$BF,$9D,$5E,$48,$FC,$24,$ED
	.dc.b	$F0,$C6,$63,$C3,$1E,$08,$C5,$78,$27,$85,$7F,$54,$FD,$15,$EC,$FF
	.dc.b	$A3,$5B,$5F,$EC,$3C,$11,$87,$FD,$8C,$D7,$F0,$CE,$6F,$ED,$C7,$E3
	.dc.b	$63,$13,$F8,$D7,$FD,$7A,$FE,$46,$77,$3D,$AB,$8C,$62,$B7,$3E,$0D
	.dc.b	$3C,$C4,$BF,$37,$E8,$AF,$6B,$FD,$4C,$D8,$CC,$60,$8F,$84,$FD,$04
	.dc.b	$6E,$BF,$E0,$0B,$FE,$E8,$8D,$47,$E9,$E3,$58,$7A,$1D,$47,$FB,$A5
	.dc.b	$AF,$EA,$5A,$F6,$F1,$BC,$BB,$4B,$F4,$B6,$79,$3F,$0A,$3C,$AD,$8F
	.dc.b	$17,$6D,$57,$F7,$9C,$D7,$C4,$9F,$86,$47,$E3,$F6,$FC,$62,$38,$24
	.dc.b	$FC,$28,$ED,$5F,$E1,$A0,$E3,$83,$E7,$92,$BE,$9B,$F4,$BB,$64,$7D
	.dc.b	$66,$DF,$C7,$7A,$9D,$BB,$2C,$FD,$09,$F6,$C7,$DC,$24,$7E,$01,$FF
	.dc.b	$72,$6C,$E9,$A4,$9F,$F3,$33,$6B,$FC,$FA,$FC,$BD,$9E,$C3,$E2,$BF
	.dc.b	$D1,$F3,$2F,$FD,$4E,$EE,$75,$18,$3C,$DE,$DB,$BF,$FD,$EC,$D8,$FD
	.dc.b	$53,$71,$FE,$58,$1C,$3F,$CB,$DD,$63,$FA,$46,$CF,$E9,$FC,$C7,$BB
	.dc.b	$F8,$19,$CD,$FF,$3B,$6B,$D7,$0F,$75,$70,$BF,$3C,$7E,$27,$E3,$FF
	.dc.b	$75,$5F,$D0,$5A,$FF,$45,$6A,$69,$FB,$C7,$5D,$1F,$2F,$DA,$7B,$1D
	.dc.b	$63,$F2,$DE,$DB,$D5,$A6,$2E,$E5,$F7,$E7,$4D,$9D,$34,$FC,$26,$CF
	.dc.b	$53,$F4,$4D,$8F,$C4,$5F,$F5,$DF,$6F,$FA,$F7,$B4,$F6,$3F,$AF,$5E
	.dc.b	$EB,$FA,$55,$E3,$0E,$67,$EB,$63,$90,$BE,$4F,$F1,$98,$D2,$C3,$BE
	.dc.b	$1C,$EE,$CF,$F6,$B9,$AF,$F5,$9A,$FE,$4F,$7D,$23,$FB,$94,$EB,$0F
	.dc.b	$34,$E6,$7C,$6E,$E3,$38,$7E,$32,$B8,$5F,$48,$C5,$FA,$5F,$23,$4F
	.dc.b	$21,$4F,$26,$1E,$67,$BC,$70,$C3,$C5,$1E,$F3,$86,$1E,$27,$9C,$79
	.dc.b	$6F,$C5,$1D,$8F,$E9,$98,$7F,$4D,$7E,$69,$83,$F4,$BE,$CF,$E2,$BD
	.dc.b	$BF,$EA,$4D,$8C,$EC,$9C,$B6,$38,$FB,$7F,$E0,$56,$E5,$CA,$31,$1F
	.dc.b	$E1,$DB,$46,$23,$FC,$45,$18,$9F,$AF,$9F,$E6,$66,$EB,$FE,$27,$D3
	.dc.b	$EC,$5E,$E7,$E5,$36,$3E,$E5,$71,$9F,$9C,$D8,$E3,$87,$9D,$B8,$76
	.dc.b	$A8,$74,$9D,$57,$99,$B1,$FE,$3B,$8E,$87,$F8,$EC,$AF,$F8,$FE,$FC
	.dc.b	$91,$C6,$BA,$DF,$9B,$0F,$64,$E6,$FC,$F1,$FA,$8D,$7A,$36,$3C,$0D
	.dc.b	$58,$76,$72,$7F,$3B,$A7,$AD,$B6,$C6,$6B,$97,$B7,$FB,$D8,$FE,$84
	.dc.b	$D7,$F6,$E7,$F9,$65,$C6,$5F,$E0,$A7,$ED,$C7,$0D,$FC,$D2,$E5,$7B
	.dc.b	$78,$6D,$6C,$FE,$A1,$ED,$BD,$BF,$CB,$36,$3F,$C4,$8D,$7F,$9D,$B5
	.dc.b	$FF,$34,$39,$BF,$F0,$DF,$94,$F2,$FC,$6D,$AE,$F7,$F2,$7D,$47,$93
	.dc.b	$49,$3F,$E1,$2C,$3F,$CD,$FE,$56,$23,$FA,$5D,$70,$7F,$C7,$0D,$73
	.dc.b	$5F,$AF,$8C,$57,$92,$3C,$67,$37,$FC,$67,$F0,$AF,$57,$F6,$CD,$BF
	.dc.b	$F2,$23,$5F,$EB,$6D,$7F,$D1,$1E,$E3,$D4,$78,$FF,$E4,$B8,$C3,$7D
	.dc.b	$4F,$C0,$D7,$0A,$DB,$FD,$9B,$DC,$7B,$DF,$E4,$68,$FF,$4B,$46,$3A
	.dc.b	$FF,$CA,$CD,$8E,$13,$B7,$FD,$3D,$ED,$3C,$4C,$42,$31,$7F,$E5,$F6
	.dc.b	$C7,$F7,$66,$BF,$EA,$35,$ED,$F2,$6C,$F1,$8C,$1D,$BD,$77,$B7,$49
	.dc.b	$3F,$E6,$D8,$CB,$F8,$CE,$FF,$2F,$4F,$E9,$A3,$04,$C7,$EF,$FA,$5E
	.dc.b	$AE,$6F,$F1,$9C,$FE,$02,$30,$67,$D9,$91,$F8,$46,$D7,$E6,$0E,$E4
	.dc.b	$8F,$CE,$BD,$7D,$29,$1F,$8C,$6C,$7E,$43,$9B,$C4,$49,$FC,$8B,$F3
	.dc.b	$BF,$46,$7C,$A5,$24,$FF,$A3,$63,$2F,$E6,$3C,$EF,$61,$1C,$1F,$9C
	.dc.b	$F7,$7E,$01,$1C,$1C,$FE,$17,$C0,$47,$07,$E9,$3D,$4F,$90,$47,$07
	.dc.b	$EA,$3A,$3E,$61,$1C,$1F,$EA,$3F,$71,$B7,$F9,$C7,$87,$FC,$87,$B7
	.dc.b	$F1,$41,$3A,$AF,$E1,$7E,$B7,$F9,$DE,$31,$3F,$E4,$26,$BF,$EE,$DB
	.dc.b	$FC,$AC,$62,$3F,$DE,$8F,$0E,$FF,$F5,$3B,$6F,$FD,$F1,$CA,$F7,$7B
	.dc.b	$61,$27,$FC,$D5,$5B,$C3,$FE,$00,$5E,$E7,$35,$B9,$F5,$56,$D5,$6D
	.dc.b	$CC,$EF,$0F,$1B,$FE,$93,$9D,$46,$2F,$A9,$2B,$E3,$4E,$A3,$DB,$5C
	.dc.b	$DA,$7F,$A0,$63,$52,$3E,$6C,$ED,$C5,$F9,$E7,$71,$AA,$E5,$86,$E3
	.dc.b	$D6,$D9,$DD,$71,$8C,$12,$BD,$18,$7F,$B5,$AB,$78,$7B,$EB,$58,$7C
	.dc.b	$4B,$EE,$70,$3E,$AC,$6A,$FE,$24,$FC,$28,$FE,$0C,$49,$FE,$08,$D8
	.dc.b	$D4,$7F,$A8,$FB,$BF,$A5,$3D,$A3,$24,$AF,$F7,$5C,$EA,$31,$3E,$23
	.dc.b	$84,$EB,$C5,$F5,$46,$0D,$4F,$E6,$EB,$95,$7C,$1B,$DC,$F4,$3E,$A4
	.dc.b	$EA,$B8,$D6,$DE,$93,$AB,$E3,$3E,$19,$CC,$FF,$46,$9C,$C7,$DF,$8C
	.dc.b	$3A,$AE,$D1,$F9,$AB,$FF,$71,$46,$23,$EE,$CF,$C7,$AE,$71,$E1,$8E
	.dc.b	$51,$88,$EA,$7A,$CB,$7E,$97,$E1,$46,$EB,$B6,$3F,$8E,$6E,$1E,$29
	.dc.b	$D1,$30,$EE,$4B,$ED,$5C,$23,$CC,$67,$0F,$01,$F1,$67,$75,$CD,$B3
	.dc.b	$B3,$E7,$9F,$1E,$BA,$47,$A2,$34,$FC,$77,$BB,$C8,$95,$E0,$8E,$55
	.dc.b	$DE,$31,$3F,$24,$DC,$F8,$1C,$46,$9F,$93,$19,$69,$E1,$3C,$87,$23
	.dc.b	$11,$EF,$39,$57,$A9,$C4,$F3,$8C,$3E,$63,$DB,$3F,$80,$3A,$CF,$2A
	.dc.b	$F5,$4F,$16,$CF,$82,$39,$3E,$33,$94,$F8,$1D,$1E,$E8,$E4,$78,$30
	.dc.b	$FC,$15,$74,$9E,$33,$D1,$FC,$D1,$F8,$E2,$DF,$ED,$5B,$3F,$E1,$48
	.dc.b	$C4,$71,$21,$F4,$4D,$CF,$DA,$BF,$E8,$7E,$CF,$E4,$63,$07,$B2,$32
	.dc.b	$E2,$7D,$44,$3F,$AB,$CF,$17,$E4,$C7,$13,$FA,$8C,$66,$7E,$24,$FD
	.dc.b	$C3,$8C,$F5,$71,$5F,$B3,$CD,$7C,$6B,$A5,$78,$67,$31,$C6,$BF,$74
	.dc.b	$71,$AE,$15,$E7,$8C,$4E,$F0,$CC,$F8,$8F,$49,$B7,$FB,$85,$6A,$BA
	.dc.b	$C6,$2B,$91,$E9,$BF,$1D,$F2,$3D,$73,$B7,$15,$CA,$BF,$B6,$DF,$5A
	.dc.b	$F1,$DF,$2C,$3E,$3D,$EF,$0F,$ED,$B7,$BC,$3C,$0D,$8F,$05,$F1,$9F
	.dc.b	$8F,$1E,$27,$B9,$E8,$77,$3C,$6F,$EA,$D7,$19,$F8,$55,$D8,$E1,$3A
	.dc.b	$AE,$11,$DA,$B8,$57,$73,$AC,$70,$35,$3C,$23,$ED,$35,$FC,$24,$F2
	.dc.b	$BF,$81,$7D,$EF,$95,$70,$6B,$C2,$F9,$57,$DE,$8F,$9F,$87,$CC,$9C
	.dc.b	$BF,$4E,$FE,$6B,$9B,$E7,$7C,$67,$C1,$7E,$A2,$38,$24,$FB,$0F,$8D
	.dc.b	$1A,$70,$D3,$D0,$53,$D0,$F0,$8E,$71,$A6,$BA,$38,$4F,$96,$7A,$CE
	.dc.b	$A3,$13,$C6,$72,$47,$C2,$7F,$0C,$C3,$F2,$ED,$C3,$FC,$92,$49,$FC
	.dc.b	$3C,$7F,$BA,$2D,$E3,$3D,$CF,$95,$5F,$DB,$FD,$D7,$AD,$FE,$5D,$F8
	.dc.b	$A4,$FF,$8C,$70,$B8,$FF,$8C,$EF,$73,$FE,$E8,$AC,$D6,$A9,$5A,$7F
	.dc.b	$91,$98,$BF,$F5,$9C,$66,$3F,$93,$B5,$F2,$47,$4B,$A7,$8C,$F4,$69
	.dc.b	$99,$F2,$4D,$5E,$EB,$CA,$FF,$3C,$7C,$C6,$63,$18,$7F,$C0,$08,$C4
	.dc.b	$FF,$B1,$A7,$35,$E8,$6B,$FC,$F1,$AF,$F8,$5D,$DE,$5F,$05,$B7,$9F
	.dc.b	$2F,$AB,$E9,$AF,$F2,$E4,$FD,$38,$C4,$39,$9E,$97,$9F,$7F,$F4,$83
	.dc.b	$6B,$E5,$D7,$BA,$7D,$04,$8C,$DF,$59,$FF,$4B,$37,$1F,$37,$BF,$CE
	.dc.b	$73,$E1,$7F,$A7,$7E,$5B,$E3,$37,$0F,$2E,$1F,$ED,$0A,$F8,$93,$F1
	.dc.b	$C9,$3E,$07,$8B,$C3,$92,$CF,$8B,$C3,$FF,$55,$78,$FF,$4D,$B5,$D9
	.dc.b	$FA,$44,$7D,$6D,$9E,$51,$E4,$F8,$0E,$4F,$F9,$9F,$BF,$FD,$61,$5F
	.dc.b	$E4,$66,$C6,$4F,$93,$CA,$8F,$EE,$44,$8C,$BC,$5F,$B1,$B3,$E0,$3F
	.dc.b	$D8,$2F,$F9,$C2,$7F,$21,$EF,$BC,$4F,$CA,$4F,$EC,$27,$31,$8E,$BF
	.dc.b	$06,$1F,$80,$BE,$0F,$DC,$6E,$9F,$A8,$6B,$FE,$0C,$6B,$FE,$67,$77
	.dc.b	$87,$EC,$5B,$3F,$B9,$F4,$DF,$01,$E2,$8F,$C7,$FC,$07,$ED,$6F,$F7
	.dc.b	$3E,$7F,$FA,$AA,$31,$3F,$C0,$63,$1D,$BF,$BD,$CB,$FC,$ED,$5F,$B7
	.dc.b	$7F,$D6,$CD,$4D,$F8,$07,$B5,$FB,$56,$E1,$FA,$07,$FD,$7C,$D8,$D1
	.dc.b	$1D,$55,$3F,$DE,$18,$BD,$9F,$E0,$0D,$9F,$D5,$CF,$FB,$27,$E0,$3F
	.dc.b	$D9,$5E,$EF,$EB,$23,$F8,$E5,$F0,$25,$70,$85,$70,$EE,$FF,$4A,$61
	.dc.b	$D6,$7F,$E0,$01,$D4,$61,$D5,$46,$E3,$11,$B7,$51,$99,$FE,$4B,$5D
	.dc.b	$EB,$CB,$75,$FF,$52,$45,$AE,$11,$FE,$4E,$8C,$61,$FC,$DB,$0E,$F3
	.dc.b	$EA,$3C,$B8,$7A,$A7,$4F,$82,$32,$78,$0A,$78,$0B,$7E,$07,$15,$EA
	.dc.b	$AE,$71,$E3,$36,$E1,$E4,$7F,$18,$BF,$E3,$2F,$C9,$30,$7F,$46,$F7
	.dc.b	$5E,$FB,$E0,$3F,$8F,$C6,$1B,$19,$9E,$0D,$7F,$CE,$32,$F8,$0F,$A7
	.dc.b	$1D,$5A,$E9,$AF,$FB,$46,$38,$65,$E0,$9E,$39,$7D,$DC,$FF,$37,$7F
	.dc.b	$55,$AF,$C8,$BE,$2D,$AE,$33,$FE,$DD,$37,$5C,$F5,$FF,$70,$3F,$89
	.dc.b	$BE,$2D,$AE,$36,$3F,$A2,$D7,$0B,$E8,$D8,$D5,$F2,$9F,$65,$72,$9E
	.dc.b	$27,$A6,$B8,$2A,$E0,$F2,$9E,$15,$E8,$3B,$C7,$E5,$67,$6F,$43,$F4
	.dc.b	$3A,$E9,$A7,$A6,$7E,$3E,$3F,$E0,$7A,$E3,$18,$AF,$8D,$EB,$7E,$C1
	.dc.b	$B7,$FD,$9A,$74,$FC,$3F,$80,$FE,$04,$FF,$6A,$6B,$E0,$0E,$30,$FF
	.dc.b	$0B,$3F,$85,$8E,$1B,$F7,$7F,$73,$CA,$F8,$0C,$5A,$31,$3C,$0F,$EF
	.dc.b	$3E,$33,$E0,$39,$47,$F7,$B6,$C6,$63,$FA,$5F,$EA,$AF,$9C,$CD,$7E
	.dc.b	$92,$B5,$5F,$E2,$86,$BF,$EC,$C6,$BF,$83,$1E,$DB,$CB,$F0,$B6,$3F
	.dc.b	$C1,$AD,$7F,$C8,$9D,$DF,$E2,$57,$33,$C0,$8F,$F6,$08,$D3,$F2,$E7
	.dc.b	$93,$FE,$DE,$6B,$E3,$6B,$F9,$76,$AF,$80,$7C,$6F,$F0,$26,$C7,$F8
	.dc.b	$71,$AF,$F9,$21,$AF,$FB,$B3,$E0,$DE,$F7,$F0,$A6,$C7,$F8,$91,$AF
	.dc.b	$F9,$41,$AF,$FB,$D3,$E0,$DE,$DB,$C8,$FF,$15,$C6,$1B,$83,$A3,$35
	.dc.b	$BB,$FF,$2D,$FC,$B7,$F7,$A6,$F9,$1F,$D9,$5F,$DB,$46,$5F,$AC,$1C
	.dc.b	$3F,$67,$E0,$9E,$37,$18,$46,$3B,$3F,$C7,$ED,$8F,$F3,$5B,$5E,$AB
	.dc.b	$CD,$E4,$D9,$E3,$18,$3B,$5F,$C2,$8F,$F2,$5B,$6B,$DB,$8E,$8F,$27
	.dc.b	$27,$C3,$AF,$F2,$8B,$67,$F2,$D9,$E8,$F5,$71,$7F,$91,$E3,$F0,$DE
	.dc.b	$7E,$09,$7E,$0D,$BC,$64,$70,$76,$9F,$09,$1C,$1F,$E6,$2A,$79,$BE
	.dc.b	$02,$46,$5F,$7F,$CA,$C8,$F7,$B6,$BF,$D2,$BF,$02,$F8,$2F,$A4,$DA
	.dc.b	$FF,$37,$B6,$77,$3E,$EF,$3E,$57,$8D,$B3,$F8,$6F,$95,$F9,$E4,$9F
	.dc.b	$9E,$F6,$F8,$29,$1D,$9B,$3F,$EA,$8C,$FF,$0E,$FE,$AF,$97,$E6,$92
	.dc.b	$BE,$9F,$C1,$7F,$94,$3C,$0F,$CE,$47,$F1,$48,$C7,$47,$FA,$97,$8D
	.dc.b	$1F,$70,$93,$C8,$98,$7F,$29,$9F,$F2,$FC,$7F,$0F,$F7,$9F,$05,$CE
	.dc.b	$E4,$F7,$3F,$D3,$73,$A7,$17,$CE,$76,$FF,$35,$9D,$39,$9D,$F8,$9F
	.dc.b	$EC,$BC,$3C,$33,$FD,$86,$B7,$39,$BE,$6D,$BF,$0E,$18,$7F,$93,$61
	.dc.b	$CB,$0F,$F5,$75,$FF,$B1,$F0,$E7,$5A,$8C,$5F,$1A,$F0,$35,$57,$07
	.dc.b	$FB,$A5,$F5,$8C,$47,$6B,$F1,$9E,$C3,$E5,$57,$39,$DD,$71,$8C,$12
	.dc.b	$78,$57,$89,$AF,$8A,$34,$FC,$93,$E6,$38,$7F,$88,$D7,$CE,$6C,$7F
	.dc.b	$0F,$8C,$47,$92,$7C,$EF,$F6,$9F,$81,$7C,$17,$C4,$9C,$E5,$FE,$FF
	.dc.b	$BF,$7D,$6A,$72,$73,$BF,$1B,$FE,$DE,$8D,$46,$09,$5C,$F0,$FF,$7D
	.dc.b	$4E,$A3,$15,$D2,$BE,$11,$E5,$C3,$C8,$F2,$8F,$04,$9F,$EC,$D5,$CD
	.dc.b	$C9,$27,$9C,$ED,$FF,$17,$4E,$DC,$CF,$2A,$EF,$5B,$95,$6D,$FA,$95
	.dc.b	$EF,$AD,$BC,$4F,$54,$70,$8C,$BE,$06,$B9,$8D,$CF,$5C,$3E,$7D,$6A
	.dc.b	$31,$87,$C8,$8C,$12,$B8,$B6,$FE,$31,$DE,$32,$47,$27,$84,$B5,$DA
	.dc.b	$3D,$86,$F0,$FA,$B3,$F4,$E3,$07,$CD,$79,$46,$8F,$91,$3C,$E3,$81
	.dc.b	$DA,$BC,$D3,$BA,$F9,$13,$97,$AB,$F5,$0F,$51,$92,$47,$17,$2D,$7C
	.dc.b	$71,$E5,$AF,$14,$F9,$2B,$C3,$39,$6D,$77,$38,$96,$B8,$CF,$9A,$BB
	.dc.b	$46,$27,$C3,$3C,$CF,$4B,$F8,$63,$E0,$CF,$0A,$EF,$19,$8F,$24,$F0
	.dc.b	$C3,$F0,$2D,$C3,$F0,$0D,$3E,$AD,$7F,$84,$0F,$B4,$7F,$84,$B0,$FD
	.dc.b	$8B,$70,$FF,$08,$46,$23,$E7,$99,$9E,$4F,$23,$DC,$C7,$FD,$29,$9B
	.dc.b	$E0,$BC,$71,$EC,$9D,$3F,$E2,$F8,$CE,$5E,$C8,$D1,$CA,$33,$3C,$2F
	.dc.b	$F5,$97,$D6,$3D,$B7,$C4,$8F,$F9,$8E,$35,$7D,$0F,$AE,$47,$33,$C2
	.dc.b	$3E,$A9,$0F,$34,$F6,$3C,$58,$7F,$05,$9F,$E0,$15,$CA,$72,$6E,$BD
	.dc.b	$4D,$7C,$37,$EA,$C3,$DB,$1D,$E3,$13,$B7,$F8,$51,$88,$F5,$38,$C3
	.dc.b	$C5,$3E,$FA,$E7,$5D,$A7,$2F,$88,$E2,$EA,$FD,$11,$E1,$8C,$4F,$2B
	.dc.b	$ED,$7F,$EC,$FC,$3D,$F3,$9A,$DC,$F9,$DB,$7D,$A3,$E0,$61,$DA,$3C
	.dc.b	$91,$8A,$E4,$D3,$91,$63,$91,$F6,$67,$96,$1E,$48,$C5,$72,$6D,$72
	.dc.b	$69,$CB,$0F,$BC,$FD,$C3,$11,$F9,$4B,$EB,$87,$C8,$9D,$57,$09,$CD
	.dc.b	$7C,$33,$E8,$1E,$78,$D5,$FC,$26,$C6,$4E,$D1,$96,$99,$AF,$C8,$57
	.dc.b	$67,$FD,$B0,$DA,$FA,$33,$FB,$28,$FC,$EC,$FF,$A3,$1A,$F9,$0F,$E0
	.dc.b	$0F,$D1,$3E,$96,$4F,$6B,$DD,$19,$C2,$9F,$E2,$38,$B7,$EF,$BF,$C5
	.dc.b	$61,$E2,$7F,$C7,$13,$A7,$CF,$56,$7F,$C0,$93,$6B,$C9,$1C,$4C,$BD
	.dc.b	$6B,$E7,$BA,$23,$FE,$3F,$9E,$4F,$8A,$32,$FF,$8C,$A3,$2F,$23,$47
	.dc.b	$18,$ED,$7E,$F6,$DF,$BD,$A4,$9F,$5B,$FE,$5A,$AE,$B5,$C6,$7E,$0C
	.dc.b	$EE,$FD,$8F,$93,$0F,$9C,$E4,$F3,$D7,$D2,$BE,$78,$7A,$2B,$78,$78
	.dc.b	$0E,$A5,$C3,$E3,$B7,$0F,$F2,$DF,$C1,$7F,$AD,$08,$FF,$AD,$1F,$F6
	.dc.b	$57,$C1,$3D,$8F,$F4,$04,$7E,$76,$31,$3F,$47,$74,$E6,$B8,$92,$7F
	.dc.b	$A0,$D7,$0B,$FE,$99,$6F,$82,$FF,$62,$92,$BF,$D8,$AF,$FB,$6F,$DE
	.dc.b	$7C,$17,$FA,$46,$3F,$5D,$18,$9F,$14,$62,$78,$12,$3F,$A8,$CE,$EF
	.dc.b	$C3,$E9,$7F,$B5,$09,$3F,$ED,$47,$C1,$F0,$5F,$E9,$F7,$BB,$63,$F2
	.dc.b	$14,$8D,$C7,$66,$D7,$BB,$E0,$7F,$DC,$24,$9F,$F7,$0B,$ED,$F0,$BF
	.dc.b	$D5,$8F,$B1,$B1,$D9,$D3,$EA,$6C,$FC,$AF,$1B,$FD,$DA,$49,$FF,$76
	.dc.b	$BF,$27,$C2,$FF,$5C,$3F,$21,$B1,$EA,$74,$FC,$66,$CF,$F8,$07,$8B
	.dc.b	$FD,$CB,$3F,$DE,$9B,$19,$9F,$F4,$11,$FE,$5B,$7F,$B5,$75,$FF,$A9
	.dc.b	$63,$F8,$3B,$6B,$F5,$6E,$E7,$FC,$02,$D5,$CD,$FE,$D8,$6C,$67,$DF
	.dc.b	$78,$1F,$D5,$B9,$9E,$72,$BF,$99,$56,$B1,$FE,$25,$5A,$9C,$DF,$1B
	.dc.b	$E1,$5F,$C2,$EF,$86,$0C,$3F,$BC,$56,$A3,$14,$65,$70,$70,$62,$FD
	.dc.b	$04,$9F,$F1,$1F,$BD,$FA,$3C,$3F,$C0,$9E,$5F,$FA,$06,$31,$3F,$DA
	.dc.b	$63,$1D,$1E,$F6,$C7,$B1,$FF,$44,$7B,$CF,$6B,$FA,$93,$67,$F9,$33
	.dc.b	$E7,$6C,$68,$8E,$B2,$A7,$B8,$FE,$F5,$EB,$FF,$5D,$C8,$FF,$5F,$8F
	.dc.b	$85,$EF,$7C,$3F,$71,$D3,$F0,$AF,$78,$7F,$37,$6D,$F5,$3E,$76,$1C
	.dc.b	$D6,$1C,$EF,$94,$E6,$31,$3E,$4B,$AF,$D0,$8B,$7D,$B0,$F1,$CF,$D4
	.dc.b	$3A,$87,$2F,$8A,$B9,$BE,$C3,$58,$78,$A3,$31,$CA,$78,$D7,$7A,$FF
	.dc.b	$34,$5F,$8B,$0F,$F6,$25,$76,$BE,$8F,$08,$CD,$70,$9E,$E7,$B4,$A7
	.dc.b	$B4,$B7,$DD,$F6,$5F,$7C,$3C,$AF,$F9,$C2,$3D,$47,$F9,$C9,$C3,$F0
	.dc.b	$8F,$89,$ED,$BD,$D7,$BD,$FE,$78,$8C,$36,$33,$3F,$E8,$59,$CB,$7D
	.dc.b	$EF,$D4,$37,$1E,$51,$88,$EE,$D7,$F4,$F2,$8E,$3A,$7D,$4C,$FF,$39
	.dc.b	$5F,$79,$AF,$C4,$AE,$17,$F8,$26,$C7,$C8,$9D,$D7,$3D,$7F,$6A,$FE
	.dc.b	$02,$B6,$DC,$3C,$35,$FB,$A3,$89,$6B,$8B,$E8,$8D,$DF,$F0,$3B,$E7
	.dc.b	$3B,$8F,$45,$70,$75,$5C,$15,$70,$7F,$86,$57,$81,$B5,$C5,$DD,$71
	.dc.b	$9F,$87,$83,$4F,$05,$7E,$B6,$73,$1E,$79,$F1,$CF,$F8,$A6,$5E,$8F
	.dc.b	$ED,$EB,$E0,$36,$7C,$27,$F8,$BF,$DE,$FE,$02,$DD,$F4,$A3,$2F,$EF
	.dc.b	$5F,$B8,$EC,$FE,$67,$EE,$BD,$EE,$D9,$23,$13,$B3,$F8,$37,$73,$E6
	.dc.b	$37,$E7,$3E,$7E,$9E,$29,$D5,$7F,$91,$5A,$FF,$1B,$C1,$5C,$63,$FA
	.dc.b	$2B,$5E,$F7,$EC,$FD,$29,$7F,$AB,$6C,$FF,$B7,$1A,$FF,$B9,$DA,$FF
	.dc.b	$0F,$73,$7F,$EC,$1B,$7D,$3F,$A8,$ED,$72,$FC,$71,$97,$FC,$BD,$7C
	.dc.b	$B7,$FE,$9B,$DF,$E7,$9F,$E3,$8D,$73,$9F,$F0,$18,$C4,$7F,$57,$FA
	.dc.b	$27,$A3,$FC,$15,$B1,$FE,$F6,$6B,$FF,$00,$5A,$FA,$FC,$3F,$84,$47
	.dc.b	$E1,$3F,$C3,$FD,$D7,$B7,$FD,$EA,$7F,$98,$08,$CF,$8B,$88,$C7,$B7
	.dc.b	$F9,$27,$73,$A7,$8F,$C5,$BE,$2F,$FC,$0C,$DD,$B9,$36,$77,$3F,$E0
	.dc.b	$BF,$8D,$7C,$75,$25,$F8,$63,$2F,$DB,$F8,$3F,$F0,$97,$11,$A4,$8F
	.dc.b	$C0,$7C,$56,$88,$E0,$FC,$27,$95,$C8,$8E,$0F,$C4,$7C,$67,$62,$38
	.dc.b	$3F,$19,$F1,$DE,$22,$38,$3F,$21,$CB,$E6,$23,$83,$F2,$9C,$5E,$A2
	.dc.b	$38,$3F,$31,$C3,$EE,$23,$83,$F3,$9F,$07,$F0,$88,$E0,$FD,$07,$C0
	.dc.b	$FC,$62,$38,$3F,$48,$7C,$A2,$18,$3F,$01,$FA,$53,$E6,$FF,$C8,$0D
	.dc.b	$8C,$C6,$0F,$F2,$27,$C3,$BC,$FF,$F7,$77,$77,$F0,$1A,$FF,$61,$4E
	.dc.b	$5F,$E0,$1D,$BF,$EF,$38,$C7,$3F,$E4,$9A,$FF,$80,$DA,$FF,$8E,$1A
	.dc.b	$FF,$8F,$5E,$3E,$07,$F2,$D8,$C6,$1D,$7E,$47,$B3,$30,$F1,$B9,$69
	.dc.b	$27,$FC,$41,$7D,$23,$C1,$5C,$67,$2D,$BF,$04,$29,$39,$AE,$27,$5A
	.dc.b	$DC,$EA,$30,$6A,$BC,$37,$E5,$6C,$7C,$AB,$F9,$E4,$7F,$6D,$5C,$E7
	.dc.b	$E5,$4F,$1B,$EF,$EB,$5B,$ED,$1E,$29,$FA,$53,$9A,$F9,$91,$5F,$64
	.dc.b	$5B,$34,$97,$E5,$3E,$FB,$89,$F8,$F1,$83,$4F,$CE,$8F,$C4,$92,$3D
	.dc.b	$6D,$8D,$47,$F9,$EE,$FE,$7F,$C1,$3E,$47,$D3,$3D,$6B,$67,$F9,$66
	.dc.b	$75,$19,$AF,$86,$D3,$FD,$07,$1A,$9E,$33,$A6,$CE,$AB,$BC,$60,$FC
	.dc.b	$95,$FE,$46,$BC,$B3,$A7,$91,$2B,$91,$E1,$AE,$2D,$57,$E3,$9E,$11
	.dc.b	$FE,$9B,$8C,$4F,$E8,$EF,$C0,$F4,$9F,$21,$CA,$FC,$51,$9B,$E7,$39
	.dc.b	$8F,$05,$F4,$8F,$19,$F2,$6B,$8B,$E1,$6B,$C4,$EF,$1C,$5F,$C3,$CF
	.dc.b	$D7,$3C,$07,$9C,$C3,$CD,$F0,$CF,$49,$E5,$87,$E3,$A7,$75,$DE,$FD
	.dc.b	$35,$AC,$3E,$0C,$F3,$3C,$78,$73,$25,$F6,$24,$69,$B7,$D2,$77,$39
	.dc.b	$24,$66,$76,$7A,$E3,$E3,$38,$7E,$4C,$7E,$0A,$31,$1F,$10,$D9,$D1
	.dc.b	$E5,$1B,$8C,$1C,$A3,$E4,$B4,$F7,$47,$8D,$D9,$D5,$E4,$64,$91,$D8
	.dc.b	$E7,$1B,$6E,$1F,$3E,$3C,$B3,$C1,$F0,$1F,$46,$79,$3E,$26,$B9,$37
	.dc.b	$87,$D5,$8F,$B2,$D8,$EB,$1C,$23,$13,$E8,$8C,$5F,$D2,$AF,$F6,$25
	.dc.b	$7E,$D5,$B1,$FC,$C1,$FD,$01,$89,$FF,$0C,$4E,$E3,$24,$3D,$04,$1F
	.dc.b	$0A,$F9,$1F,$C5,$C7,$AC,$93,$F9,$D9,$D5,$7B,$2F,$73,$FA,$13,$57
	.dc.b	$88,$93,$B3,$F3,$75,$CA,$FE,$19,$E3,$BE,$33,$B8,$C9,$E3,$9E,$18
	.dc.b	$7C,$78,$E1,$7E,$89,$CD,$7D,$EB,$F2,$4E,$70,$F1,$DF,$49,$EE,$D9
	.dc.b	$CE,$1F,$D7,$30,$E7,$3D,$B0,$F2,$4E,$70,$DC,$77,$BE,$CE,$8C,$1F
	.dc.b	$02,$37,$5C,$C5,$73,$9E,$11,$98,$C3,$E5,$8C,$BF,$62,$79,$BF,$5A
	.dc.b	$BC,$2E,$CF,$93,$7E,$A7,$55,$CA,$31,$87,$C6,$8C,$5F,$48,$C5,$79
	.dc.b	$A4,$CA,$E3,$7B,$3C,$38,$7D,$59,$EE,$71,$38,$61,$BB,$EE,$7B,$30
	.dc.b	$EF,$5F,$00,$E4,$E2,$3C,$35,$F0,$5E,$17,$F0,$B0,$F1,$C6,$0F,$A0
	.dc.b	$59,$F8,$78,$6C,$91,$E2,$8C,$B7,$0F,$55,$74,$9F,$51,$E1,$36,$48
	.dc.b	$DD,$74,$8C,$57,$F1,$0A,$E5,$5E,$D8,$C5,$7E,$EE,$BE,$2C,$7F,$A4
	.dc.b	$6B,$F4,$D1,$F0,$0F,$01,$5F,$B8,$7D,$DF,$8A,$7B,$FF,$4A,$BE,$3C
	.dc.b	$EA,$FF,$7B,$5A,$9F,$8B,$7E,$CA,$D4,$7F,$50,$9D,$4E,$6B,$A3,$E6
	.dc.b	$3E,$1C,$6B,$0F,$64,$65,$AA,$35,$5E,$4A,$E7,$3F,$2E,$BF,$C9,$75
	.dc.b	$A8,$E5,$5B,$3C,$F3,$AC,$38,$D6,$EB,$95,$7E,$1A,$31,$87,$C5,$AD
	.dc.b	$92,$79,$9B,$71,$5E,$F8,$F8,$51,$8B,$F5,$DF,$4C,$3E,$5C,$E6,$BC
	.dc.b	$D7,$F1,$23,$EE,$3F,$05,$B5,$F3,$23,$EE,$35,$F1,$4F,$80,$FF,$80
	.dc.b	$3E,$C7,$67,$EA,$7C,$27,$22,$3F,$ED,$7A,$47,$6F,$8D,$93,$96,$DF
	.dc.b	$FB,$82,$9F,$01,$E4,$25,$73,$7C,$FF,$2F,$23,$B3,$6B,$96,$7E,$C2
	.dc.b	$4F,$8D,$F7,$F2,$C8,$F3,$36,$7C,$5E,$5F,$C4,$24,$FA,$DF,$8F,$F0
	.dc.b	$F2,$3D,$CD,$9F,$57,$97,$F3,$09,$3F,$0D,$F9,$FE,$24,$8F,$8C,$D9
	.dc.b	$F8,$5E,$27,$D4,$24,$FC,$B7,$EB,$F8,$92,$3E,$73,$67,$F6,$83,$CB
	.dc.b	$FA,$33,$88,$C7,$0F,$E0,$A7,$31,$C1,$AF,$EF,$5C,$F2,$7F,$AB,$1B
	.dc.b	$5F,$C8,$3E,$05,$A7,$F7,$69,$CD,$FF,$34,$9C,$D7,$85,$B1,$E6,$6B
	.dc.b	$FC,$2D,$AF,$F2,$CF,$83,$6B,$F9,$86,$D7,$AD,$B3,$F0,$1A,$FF,$15
	.dc.b	$73,$8F,$FB,$11,$B3,$D7,$4F,$F0,$6C,$62,$BF,$A5,$A8,$F0,$12,$7C
	.dc.b	$0F,$EA,$BE,$07,$BD,$7F,$42,$6B,$FD,$51,$A9,$A7,$77,$18,$7F,$9A
	.dc.b	$30,$7A,$7F,$ED,$F9,$E5,$3F,$E2,$2A,$E0,$79,$EB,$7D,$F9,$9F,$F1
	.dc.b	$3F,$C0,$BE,$73,$F7,$ED,$CF,$F0,$4D,$C7,$F0,$ED,$7F,$9E,$39,$C3
	.dc.b	$FD,$D6,$D9,$FF,$60,$60,$F9,$DE,$F8,$7F,$8D,$BE,$0D,$EF,$53,$19
	.dc.b	$5C,$3F,$D1,$33,$A6,$CE,$9A,$7F,$A3,$B2,$7B,$7F,$E4,$26,$C7,$FA
	.dc.b	$4A,$FF,$C0,$3F,$07,$FE,$02,$F8,$17,$BF,$FD,$D2,$F7,$5F,$D9,$9A
	.dc.b	$2F,$58,$7F,$93,$A7,$98,$BE,$6F,$FB,$7E,$34,$B0,$F0,$61,$CA,$EC
	.dc.b	$FF,$65,$9A,$FF,$8B,$2B,$C0,$F5,$AF,$DB,$CE,$B0,$F3,$E1,$E0,$AF
	.dc.b	$F0,$6C,$E6,$F8,$CE,$67,$CE,$EE,$B8,$DF,$49,$EC,$7A,$70,$F1,$BC
	.dc.b	$23,$D0,$E8,$EE,$59,$E6,$FA,$CC,$5F,$8A,$7B,$5F,$86,$31,$3B,$3F
	.dc.b	$88,$61,$FC,$45,$F9,$A6,$0F,$E2,$FF,$02,$F8,$27,$C1,$FF,$17,$8C
	.dc.b	$36,$33,$3C,$1A,$FF,$84,$E5,$F0,$7F,$E5,$A8,$EA,$D7,$4D,$7F,$C7
	.dc.b	$31,$C3,$3F,$04,$F1,$CB,$E9,$67,$F9,$CB,$FB,$6D,$7E,$45,$F1,$6D
	.dc.b	$71,$9F,$F2,$29,$BA,$E7,$AF,$F9,$21,$FB,$D7,$C5,$B5,$C6,$FF,$CB
	.dc.b	$B7,$C4,$E7,$7C,$5F,$4C,$EE,$57,$CE,$77,$1E,$8A,$E0,$FF,$9A,$27
	.dc.b	$4D,$AE,$55,$DE,$3C,$35,$C1,$EB,$3A,$8F,$D7,$6B,$DA,$7F,$CA,$75
	.dc.b	$F1,$71,$FC,$FC,$70,$7C,$4F,$F9,$E1,$CC,$3E,$0F,$9C,$FC,$06,$BF
	.dc.b	$BC,$BB,$F0,$7F,$C1,$23,$96,$1F,$E7,$28,$C4,$7F,$99,$1B,$87,$89
	.dc.b	$FC,$2C,$70,$3F,$D1,$2E,$A7,$C9,$CE,$F8,$3F,$1D,$FF,$34,$7C,$3C
	.dc.b	$3E,$4F,$09,$F0,$7F,$9D,$6E,$5F,$DB,$DA,$FF,$74,$6B,$FE,$69,$73
	.dc.b	$7F,$C1,$07,$C1,$BD,$2F,$0B,$6B,$D0,$2B,$8E,$7F,$E9,$2F,$73,$13
	.dc.b	$A6,$BF,$E8,$66,$E1,$DE,$FF,$A3,$4E,$EF,$F6,$D5,$C2,$3F,$CF,$CD
	.dc.b	$7B,$46,$8F,$2F,$C1,$C8,$E4,$79,$BD,$B7,$CC,$7F,$01,$6C,$7F,$84
	.dc.b	$DA,$FF,$86,$1A,$FF,$A3,$5F,$80,$E7,$3E,$44,$7F,$C3,$9F,$08,$F2
	.dc.b	$BF,$87,$36,$33,$18,$3F,$89,$7C,$83,$CB,$A4,$9F,$F1,$34,$65,$FF
	.dc.b	$14,$7C,$2B,$DF,$FF,$16,$47,$F8,$BA,$31,$E5,$6B,$DF,$FC,$D4,$E6
	.dc.b	$B6,$FF,$A9,$FE,$15,$F0,$BF,$E7,$26,$F2,$FF,$8D,$5B,$3F,$E3,$A6
	.dc.b	$BF,$87,$E0,$FF,$3C,$F1,$7C,$CA,$FC,$74,$60,$FF,$3E,$F1,$FB,$63
	.dc.b	$FC,$80,$DA,$FF,$22,$B5,$FB,$BB,$70,$F2,$E5,$70,$6C,$FE,$67,$E8
	.dc.b	$7C,$24,$9F,$CF,$3E,$0F,$84,$91,$E0,$6C,$F8,$3E,$13,$CA,$49,$F2
	.dc.b	$BE,$CF,$5E,$47,$EA,$5B,$3E,$8D,$FD,$A4,$9F,$6B,$F0,$38,$24,$7C
	.dc.b	$06,$CF,$EC,$7D,$CF,$92,$49,$FD,$B3,$F2,$3E,$0E,$47,$C8,$6C,$FC
	.dc.b	$88,$F9,$A4,$3E,$69,$FE,$AA,$E9,$FD,$D2,$89,$FA,$5F,$F8,$52,$3F
	.dc.b	$DB,$53,$A9,$CF,$8B,$FE,$E2,$8E,$53,$FE,$A9,$6C,$FF,$A9,$5D,$47
	.dc.b	$F0,$2E,$3F,$E0,$CF,$42,$4F,$FB,$8A,$7B,$3F,$C1,$FA,$F8,$B9,$9D
	.dc.b	$F7,$7F,$36,$BE,$73,$91,$19,$AE,$2D,$AE,$27,$BE,$7B,$38,$7E,$15
	.dc.b	$74,$9F,$F7,$5D,$F9,$23,$15,$FE,$E5,$8C,$5F,$68,$E9,$3E,$4B,$FE
	.dc.b	$51,$3F,$A2,$24,$FF,$2F,$78,$BF,$C9,$BE,$03,$FC,$D5,$3E,$43,$D7
	.dc.b	$1D,$27,$71,$83,$EC,$4F,$0A,$F2,$C7,$AA,$77,$87,$F5,$CA,$D4,$7C
	.dc.b	$E6,$DF,$87,$D0,$DB,$F1,$63,$37,$DC,$38,$BE,$95,$C2,$7C,$6D,$76
	.dc.b	$FF,$BE,$A4,$D2,$57,$FA,$FE,$3E,$09,$3E,$0B,$CD,$32,$7C,$C7,$C7
	.dc.b	$8C,$57,$96,$B8,$D7,$B6,$73,$7F,$05,$B7,$ED,$6D,$FC,$A6,$BE,$E6
	.dc.b	$D9,$A4,$9F,$43,$ED,$23,$83,$CC,$7D,$DF,$82,$F2,$38,$AF,$44,$FD
	.dc.b	$39,$CC,$62,$75,$7F,$66,$72,$D8,$C9,$F5,$E3,$2D,$33,$3F,$86,$8C
	.dc.b	$B6,$32,$FD,$C8,$CB,$63,$36,$9C,$B6,$33,$18,$23,$82,$4F,$AC,$FA
	.dc.b	$D1,$A7,$15,$EB,$6C,$FA,$5A,$79,$CA,$79,$4B,$5E,$36,$D7,$5A,$FC
	.dc.b	$05,$7D,$13,$C6,$E2,$7C,$71,$83,$F0,$86,$DE,$6E,$5F,$C3,$39,$8E
	.dc.b	$47,$A4,$AF,$A5,$B1,$E9,$6B,$E9,$6D,$7A,$5B,$5E,$9A,$F1,$D7,$98
	.dc.b	$F1,$92,$77,$1E,$58,$FC,$C3,$E5,$38,$C7,$95,$A7,$94,$B1,$E5,$6B
	.dc.b	$E5,$6D,$79,$5B,$5E,$59,$F8,$67,$96,$BC,$6F,$F2,$D3,$EC,$16,$BE
	.dc.b	$B3,$67,$F8,$21,$2B,$F8,$13,$63,$91,$E3,$3C,$C7,$F5,$7F,$82,$70
	.dc.b	$7E,$A6,$BE,$04,$E8,$4E,$A3,$31,$CC,$91,$97,$D5,$1F,$15,$E5,$19
	.dc.b	$C3,$E5,$47,$5C,$3E,$23,$E3,$BF,$13,$C8,$E2,$E2,$78,$C7,$82,$31
	.dc.b	$7D,$E3,$0F,$9A,$BA,$C6,$0E,$87,$A5,$CD,$F9,$67,$94,$70,$33,$5D
	.dc.b	$AB,$D1,$87,$F6,$EB,$E1,$18,$BF,$8B,$18,$8E,$D3,$C5,$F9,$75,$6F
	.dc.b	$B4,$F3,$8C,$3B,$AE,$53,$AB,$EB,$5A,$69,$EF,$AE,$11,$9B,$F7,$DF
	.dc.b	$29,$C6,$1F,$EB,$68,$C5,$F1,$3F,$D1,$EE,$2B,$E3,$E1,$F0,$A7,$51
	.dc.b	$D3,$0E,$71,$F4,$6B,$85,$FD,$6B,$E1,$7C,$B0,$F2,$DF,$2A,$FB,$77
	.dc.b	$C2,$3E,$8C,$FC,$37,$47,$5C,$38,$DF,$86,$3E,$0B,$49,$7D,$27,$6D
	.dc.b	$9D,$B6,$77,$1D,$4E,$6F,$EC,$6F,$F3,$8D,$AF,$F6,$EB,$FE,$21,$8C
	.dc.b	$B6,$3D,$0F,$F7,$FF,$82,$7B,$B8,$C3,$E8,$55,$C3,$FC,$55,$5A,$7E
	.dc.b	$8C,$66,$FF,$49,$1F,$EC,$29,$D4,$7F,$82,$F0,$FC,$71,$2B,$FC,$83
	.dc.b	$3A,$BE,$F7,$D0,$97,$E5,$C3,$B3,$F9,$48,$CC,$FF,$BE,$AB,$94,$7B
	.dc.b	$4F,$15,$70,$8C,$D7,$98,$E9,$3C,$8F,$6C,$EC,$8E,$63,$FC,$D1,$3C
	.dc.b	$9F,$11,$C8,$F7,$5F,$B2,$B8,$C6,$2B,$C6,$73,$55,$CF,$0F,$D9,$36
	.dc.b	$F9,$56,$E3,$E5,$D7,$F2,$F6,$BE,$88,$FE,$52,$F4,$BF,$C0,$BE,$87
	.dc.b	$9B,$FE,$3D,$E5,$7C,$16,$32,$5F,$FB,$DE,$3F,$D6,$E4,$7F,$D6,$EF
	.dc.b	$4A,$53,$DD,$C3,$6B,$C6,$D9,$FE,$7A,$D7,$FA,$42,$F8,$1E,$24,$9E
	.dc.b	$2F,$FB,$22,$A4,$7F,$B2,$1B,$5D,$3E,$03,$C2,$49,$F0,$BE,$4C,$A4
	.dc.b	$79,$1B,$3F,$ED,$0B,$FF,$6A,$92,$7F,$DA,$AF,$B3,$C6,$91,$EC,$6C
	.dc.b	$FB,$3D,$FF,$82,$49,$F8,$2F,$FB,$83,$E0,$64,$7F,$B8,$1B,$3F,$13
	.dc.b	$C1,$F9,$24,$9F,$92,$FC,$CF,$06,$47,$CC,$6C,$FF,$68,$E2,$7C,$77
	.dc.b	$D5,$6D,$7E,$11,$B3,$B9,$FB,$5E,$3F,$D3,$8F,$F5,$33,$5F,$F6,$33
	.dc.b	$53,$4F,$B9,$C5,$F8,$88,$DF,$07,$FA,$F3,$E0,$5A,$FF,$AF,$AF,$79
	.dc.b	$7F,$AC,$FE,$05,$F0,$1F,$EC,$2F,$8C,$FE,$C3,$18,$9F,$F6,$B6,$7D
	.dc.b	$73,$56,$AF,$F2,$57,$B9,$FF,$04,$B5,$FF,$0A,$35,$F0,$47,$FB,$42
	.dc.b	$38,$57,$55,$B6,$67,$FD,$AB,$F0,$5A,$C7,$C7,$F1,$4F,$1F,$35,$FE
	.dc.b	$E7,$C5,$39,$AF,$49,$FE,$D1,$BF,$16,$5F,$EF,$16,$A6,$92,$FA,$CF
	.dc.b	$29,$FF,$6F,$FA,$DF,$EC,$8A,$F2,$47,$FB,$EF,$E0,$DF,$05,$E3,$F0
	.dc.b	$9F,$15,$FE,$EE,$9C,$E3,$CF,$4F,$F5,$AC,$E6,$FF,$79,$5A,$C3,$FC
	.dc.b	$66,$D7,$FC,$7C,$D7,$FD,$D9,$F0,$6F,$73,$FD,$F6,$DC,$7D,$12,$83
	.dc.b	$A9,$93,$A9,$E1,$3F,$ED,$17,$71,$99,$FC,$76,$3F,$83,$1E,$D6,$27
	.dc.b	$FD,$FD,$18,$7C,$37,$E0,$F8,$3F,$0F,$C0,$BC,$CD,$E1,$C6,$BF,$1E
	.dc.b	$D5,$87,$37,$FD,$C8,$73,$15,$CD,$E3,$87,$81,$61,$E0,$C3,$BC,$D9
	.dc.b	$FF,$34,$4D,$7B,$E1,$FA,$BB,$E9,$1F,$20,$E6,$78,$2F,$9B,$4E,$65
	.dc.b	$39,$DF,$57,$D1,$3D,$30,$F6,$3D,$A3,$35,$D2,$31,$1E,$07,$85,$F9
	.dc.b	$67,$FC,$EF,$87,$A6,$F9,$C6,$2F,$C5,$18,$7C,$B7,$CB,$0F,$24,$7B
	.dc.b	$0F,$EB,$37,$FD,$69,$F9,$C6,$0F,$F4,$2F,$C0,$BE,$09,$F0,$7F,$D8
	.dc.b	$23,$0D,$8C,$CF,$06,$BF,$E9,$79,$7C,$1F,$DE,$8E,$AD,$74,$D7,$E9
	.dc.b	$47,$0C,$BC,$13,$C7,$2F,$AF,$9F,$E8,$EF,$EE,$35,$FA,$97,$C5,$B5
	.dc.b	$C6,$7E,$F1,$BA,$E7,$AF,$E0,$5F,$C1,$DF,$16,$D7,$1B,$FD,$6D,$F1
	.dc.b	$39,$DF,$17,$D3,$3B,$95,$F3,$9D,$C7,$A2,$B8,$3F,$B8,$9D,$36,$B9
	.dc.b	$57,$78,$F0,$D7,$07,$AC,$EA,$3F,$6B,$A7,$69,$FD,$6B,$71,$FE,$7D
	.dc.b	$87,$69,$FD,$B5,$70,$7D,$BF,$07,$5F,$DD,$37,$1F,$9D,$7F,$D1,$FE
	.dc.b	$0F,$F8,$2D,$7F,$12,$6B,$E0,$3F,$89,$B8,$8E,$19,$75,$8C,$65,$FD
	.dc.b	$4B,$99,$F0,$7E,$3B,$46,$30,$ED,$C3,$E5,$F4,$9F,$07,$49,$3F,$CC
	.dc.b	$7E,$0D,$F3,$58,$8F,$F7,$0D,$70,$7F,$C7,$6D,$73,$5F,$A5,$F7,$53
	.dc.b	$F0,$DB,$3F,$CC,$1A,$FB,$F2,$61,$FD,$1A,$35,$5C,$48,$F9,$7A,$6B
	.dc.b	$FE,$F4,$BE,$6F,$FB,$FD,$AF,$F1,$56,$BF,$CF,$BE,$75,$D9,$FB,$B6
	.dc.b	$C7,$F9,$59,$AF,$FC,$00,$6B,$FC,$7C,$7B,$0F,$7F,$F8,$23,$63,$FC
	.dc.b	$BE,$D7,$81,$C2,$37,$EC,$C8,$DB,$5D,$FB,$8F,$95,$FE,$20,$D8,$FF
	.dc.b	$35,$B5,$DC,$7F,$15,$F8,$47,$A1,$49,$3F,$E7,$28,$CB,$FE,$74,$F8
	.dc.b	$47,$C0,$7F,$8A,$A3,$CF,$18,$F3,$75,$8B,$B7,$85,$EE,$FE,$27,$BE
	.dc.b	$F8,$4F,$F3,$93,$7B,$3F,$D0,$2D,$9F,$86,$D7,$E6,$72,$FC,$B9,$FC
	.dc.b	$6F,$07,$F9,$EE,$31,$3F,$92,$8F,$F3,$F7,$27,$BA,$BF,$D1,$AD,$9F
	.dc.b	$C2,$6F,$A3,$C7,$F3,$D2,$BC,$6D,$9F,$CE,$FC,$27,$72,$4F,$E7,$9E
	.dc.b	$FF,$09,$23,$BB,$67,$D7,$F3,$FF,$00,$93,$F0,$1F,$57,$BD,$23,$F5
	.dc.b	$2D,$9F,$CD,$6F,$EC,$24,$FB,$1F,$97,$D3,$23,$E5,$B6,$7F,$6B,$F0
	.dc.b	$3F,$1C,$93,$FB,$67,$E3,$FC,$2C,$8F,$8E,$D9,$FA,$7F,$3B,$FE,$47
	.dc.b	$EE,$FF,$06,$77,$FF,$97,$79,$13,$9F,$0B,$FC,$25,$18,$C7,$EE,$CF
	.dc.b	$E0,$08,$E0,$FF,$17,$C6,$F0,$FE,$3C,$1F,$F1,$04,$6A,$7A,$91,$E3
	.dc.b	$E3,$DA,$E3,$5C,$F0,$F3,$5F,$92,$B5,$1E,$96,$DF,$F8,$5E,$BB,$DF
	.dc.b	$2B,$E1,$3E,$13,$B9,$63,$BD,$72,$7F,$C8,$F5,$CA,$FC,$35,$C2,$7F
	.dc.b	$3B,$5F,$A2,$24,$7E,$89,$EF,$1F,$ED,$FE,$FB,$5E,$7A,$E4,$FF,$A3
	.dc.b	$77,$FE,$1F,$18,$3D,$57,$C1,$FF,$19,$46,$AF,$CF,$1F,$E5,$E8,$C5
	.dc.b	$75,$9F,$04,$F3,$7F,$CC,$33,$A8,$EA,$71,$BE,$D3,$A6,$BF,$EE,$E6
	.dc.b	$CF,$86,$3F,$DD,$31,$FE,$F0,$24,$FF,$BC,$1F,$F2,$D4,$F7,$9F,$F3
	.dc.b	$8F,$3B,$BF,$35,$2F,$C4,$7F,$9E,$27,$4E,$2B,$D6,$4C,$3F,$D6,$93
	.dc.b	$A8,$C4,$FF,$95,$E3,$13,$F3,$6B,$FC,$D9,$18,$9F,$97,$18,$BF,$15
	.dc.b	$FE,$62,$B4,$FC,$3A,$F2,$4E,$AF,$C2,$D8,$F0,$3F,$C6,$C8,$FF,$6A
	.dc.b	$6C,$F5,$7F,$CF,$DF,$0B,$FD,$5A,$7A,$92,$74,$FF,$A6,$A3,$51,$82
	.dc.b	$57,$8A,$75,$C3,$EB,$69,$CE,$3F,$DB,$55,$C9,$E1,$5F,$B6,$8C,$4F
	.dc.b	$A8,$E3,$1D,$9F,$F4,$14,$6A,$3E,$DD,$FA,$AF,$C1,$1C,$DB,$7E,$06
	.dc.b	$DF,$26,$CF,$FA,$D2,$FA,$3F,$E0,$CB,$E9,$5E,$93,$81,$63,$85,$72
	.dc.b	$55,$CB,$0F,$B3,$18,$9D,$4E,$6F,$AC,$FB,$4E,$77,$D5,$E0,$D9,$E0
	.dc.b	$FC,$4C,$3E,$49,$ED,$75,$7D,$88,$E0,$95,$DC,$EA,$5A,$EA,$FC,$98
	.dc.b	$DD,$76,$7E,$0C,$6E,$BA,$B6,$36,$7D,$97,$91,$92,$4F,$37,$DE,$62
	.dc.b	$38,$47,$AE,$30,$7D,$E8,$F8,$06,$8F,$1C,$73,$8F,$5B,$6B,$D6,$F3
	.dc.b	$31,$5E,$86,$D7,$A2,$BB,$1E,$83,$FC,$65,$5F,$A0,$6D,$27,$FC,$5F
	.dc.b	$1F,$28,$CB,$89,$E1,$1F,$20,$C9,$FE,$36,$7F,$1E,$4F,$86,$FC,$1D
	.dc.b	$F9,$EB,$DA,$7E,$82,$32,$E2,$BE,$4C,$FE,$B2,$73,$18,$9D,$3F,$94
	.dc.b	$8C,$BF,$1A,$33,$49,$CB,$E4,$BF,$E0,$B1,$87,$2D,$9F,$A7,$5F,$A9
	.dc.b	$8C,$3F,$80,$BF,$24,$EA,$3C,$07,$94,$E0,$E6,$FC,$0F,$F0,$FA,$E7
	.dc.b	$5F,$0E,$B6,$75,$9F,$4D,$F9,$9E,$4F,$99,$A7,$5A,$E6,$FC,$06,$CF
	.dc.b	$8C,$F8,$D1,$C1,$EC,$78,$2B,$C4,$FA,$89,$7F,$26,$78,$4E,$A3,$13
	.dc.b	$C8,$DD,$F9,$DC,$47,$C0,$7F,$91,$5F,$91,$EA,$D8,$FC,$35,$F6,$3D
	.dc.b	$45,$9F,$54,$78,$2F,$C5,$5F,$0E,$7B,$B6,$33,$5F,$14,$93,$F9,$59
	.dc.b	$FC,$F3,$63,$F3,$91,$88,$FD,$2B,$EB,$9F,$E7,$BF,$0D,$6B,$F3,$07
	.dc.b	$DB,$3F,$A5,$39,$AF,$99,$1A,$7F,$A1,$46,$5F,$E5,$6D,$B3,$49,$3D
	.dc.b	$67,$8B,$6B,$C6,$D3,$FA,$4C,$FA,$66,$C7,$ED,$E3,$11,$FD,$0D,$AF
	.dc.b	$8A,$3D,$21,$FF,$26,$FB,$EE,$4F,$ED,$75,$2B,$75,$A9,$C3,$E6,$25
	.dc.b	$FC,$BB,$E5,$5C,$23,$D3,$3A,$8F,$02,$9D,$B9,$3D,$11,$C4,$DC,$6A
	.dc.b	$79,$92,$3E,$0C,$F9,$5E,$0F,$23,$A1,$DE,$BD,$33,$E7,$6C,$F6,$6C
	.dc.b	$FC,$5A,$F0,$4F,$6A,$E0,$F6,$9E,$8D,$9D,$C7,$86,$76,$F8,$A3,$2F
	.dc.b	$D3,$3D,$06,$6B,$C2,$F4,$31,$5D,$63,$ED,$CF,$C8,$9F,$F7,$73,$5F
	.dc.b	$54,$FF,$3C,$8E,$E7,$83,$E1,$9F,$09,$FA,$4C,$3C,$A7,$FA,$67,$C0
	.dc.b	$FF,$69,$92,$FF,$DA,$6F,$FB,$2B,$E1,$A4,$7F,$B2,$9B,$8F,$FB,$E3
	.dc.b	$89,$F0,$3F,$90,$6D,$7F,$4E,$6C,$EE,$7F,$DC,$2C,$9F,$F7,$0B,$4F
	.dc.b	$C3,$9F,$97,$F0,$38,$B8,$6B,$F8,$C6,$BB,$9F,$17,$C3,$4A,$F1,$36
	.dc.b	$7A,$8E,$47,$CB,$FF,$BA,$C9,$5F,$D9,$A3,$2F,$FB,$AB,$E1,$FF,$DD
	.dc.b	$A4,$70,$7B,$FE,$1F,$E1,$11,$C1,$CF,$E5,$FF,$DF,$04,$70,$7F,$BD
	.dc.b	$BE,$23,$FD,$F2,$47,$07,$CB,$F1,$BE,$71,$1C,$1F,$DF,$3E,$1D,$7F
	.dc.b	$C0,$9B,$19,$8C,$1F,$8C,$F8,$89,$1F,$89,$6E,$9F,$1B,$CC,$F9,$A9
	.dc.b	$A4,$DF,$F8,$7F,$C3,$FF,$1E,$BF,$F0,$6B,$6B,$E7,$B5,$FB,$AD,$7F
	.dc.b	$82,$FC,$3B,$E1,$3F,$2D,$5F,$C2,$2B,$97,$3F,$F3,$9A,$D4,$FF,$1B
	.dc.b	$8F,$E7,$93,$A8,$E7,$5F,$C8,$AB,$77,$F8,$06,$BF,$8E,$6B,$FC,$5B
	.dc.b	$E1,$37,$87,$E9,$FE,$15,$F0,$DE,$4A,$FE,$55,$5A,$C7,$C5,$5A,$7C
	.dc.b	$87,$3C,$78,$60,$9D,$5F,$FB,$FA,$F9,$29,$D6,$1D,$E1,$3E,$67,$C3
	.dc.b	$7F,$CF,$A3,$11,$A8,$FE,$83,$F0,$AF,$63,$C3,$E2,$3E,$1B,$A4,$E7
	.dc.b	$4F,$3E,$1E,$1A,$D4,$62,$13,$B7,$CD,$1F,$46,$77,$18,$3F,$A7,$4E
	.dc.b	$F3,$FE,$21,$3A,$7F,$B9,$7C,$33,$E0,$7D,$CD,$C3,$C0,$FF,$57,$6C
	.dc.b	$68,$8E,$B0,$A7,$84,$F9,$5E,$6F,$C9,$B3,$F2,$E3,$FB,$17,$C3,$7F
	.dc.b	$63,$F8,$47,$9B,$FD,$5A,$F7,$87,$C1,$6D,$F5,$69,$FD,$C2,$F9,$57
	.dc.b	$1C,$3B,$52,$72,$FE,$5E,$3F,$C2,$B8,$78,$23,$F8,$FC,$E9,$F4,$E1
	.dc.b	$5F,$38,$9E,$0E,$8C,$5F,$F1,$CA,$E5,$87,$A2,$B7,$5C,$9E,$31,$9A
	.dc.b	$E1,$3C,$8F,$35,$F8,$5F,$51,$93,$C1,$5D,$1C,$BD,$30,$F6,$D7,$48
	.dc.b	$C5,$73,$C3,$CB,$5A,$BF,$82,$7F,$A5,$0E,$E7,$FA,$5A,$30,$E0,$FF
	.dc.b	$05,$FC,$23,$E1,$5F,$0D,$FE,$9C,$8C,$36,$33,$3F,$D0,$27,$2D,$F8
	.dc.b	$6F,$D8,$37,$4E,$51,$88,$FE,$A7,$28,$C4,$78,$9E,$39,$FD,$4D,$3F
	.dc.b	$39,$3F,$6D,$AF,$C9,$AE,$57,$F7,$DB,$1F,$36,$77,$5C,$F5,$FF,$14
	.dc.b	$BF,$80,$AD,$B7,$0F,$0D,$7F,$8B,$8E,$25,$AE,$2F,$9F,$0F,$04,$F0
	.dc.b	$C3,$FC,$89,$3C,$1B,$5C,$1F,$F2,$46,$1E,$09,$E5,$1E,$79,$D3,$FA
	.dc.b	$A6,$CE,$E3,$99,$FC,$2E,$7E,$1D,$FF,$9C,$DB,$AF,$4C,$7E,$05,$6F
	.dc.b	$C8,$FD,$A4,$FB,$5B,$1D,$0F,$E0,$1F,$0D,$FB,$E7,$F8,$C3,$5C,$EF
	.dc.b	$C1,$47,$DB,$77,$1D,$31,$7C,$2B,$E1,$B0,$DA,$D9,$FC,$97,$F4,$FE
	.dc.b	$4F,$F2,$1A,$7E,$AC,$F2,$BA,$36,$78,$70,$F8,$EF,$8D,$FF,$8D,$6B
	.dc.b	$55,$CD,$B1,$FE,$99,$9D,$EB,$FE,$6C,$FD,$41,$EE,$7E,$A5,$B8,$7F
	.dc.b	$BF,$9A,$FF,$88,$1A,$FF,$8B,$9C,$DF,$87,$D0,$F3,$9F,$1E,$33,$7D
	.dc.b	$EF,$37,$FD,$22,$FD,$F7,$B9,$E1,$EE,$78,$C9,$5E,$37,$CD,$72,$3C
	.dc.b	$CD,$7F,$C6,$BF,$3C,$F9,$3F,$E0,$8D,$8F,$F1,$C8,$70,$F0,$8D,$FC
	.dc.b	$DC,$8D,$CA,$31,$1D,$BE,$17,$A9,$1D,$BE,$0F,$84,$91,$C9,$AF,$F9
	.dc.b	$2B,$E1,$FE,$8F,$97,$E4,$6C,$67,$E1,$DD,$39,$DB,$FC,$55,$3F,$E5
	.dc.b	$2F,$8F,$65,$FE,$B8,$6E,$DE,$E6,$CE,$E7,$FD,$7B,$F2,$FF,$DC,$23
	.dc.b	$FC,$CE,$D7,$FD,$5E,$D7,$FD,$80,$D5,$F6,$9F,$26,$FE,$5B,$63,$FC
	.dc.b	$CA,$D7,$FC,$E6,$D7,$FD,$76,$BE,$7B,$FD,$9E,$D9,$E7,$18,$3F,$DA
	.dc.b	$7C,$0F,$88,$A4,$9F,$AA,$EB,$C7,$91,$F6,$1B,$87,$E1,$2B,$99,$27
	.dc.b	$EE,$BD,$BB,$64,$7E,$01,$B3,$F8,$CF,$87,$F1,$92,$7F,$0A,$F9,$BE
	.dc.b	$86,$47,$E2,$1B,$3F,$94,$FA,$5F,$59,$27,$F1,$AF,$C2,$8F,$F1,$FF
	.dc.b	$6F,$FA,$36,$7F,$C5,$5A,$FF,$86,$E7,$F7,$9E,$E7,$F8,$F6,$31,$1F
	.dc.b	$38,$E5,$3F,$51,$A7,$F8,$76,$39,$76,$7F,$07,$F5,$F9,$4F,$83,$D0
	.dc.b	$97,$FB,$5C,$5F,$29,$8A,$FE,$19,$57,$0F,$E2,$13,$6F,$FC,$D9,$6F
	.dc.b	$94,$F5,$FB,$7E,$C6,$CF,$56,$E3,$FC,$5D,$B8,$27,$FE,$00,$8F,$8E
	.dc.b	$F8,$BF,$31,$FE,$5D,$8C,$1E,$77,$11,$C6,$FF,$D6,$F3,$A6,$C6,$5F
	.dc.b	$F6,$3C,$65,$B1,$CA,$3E,$17,$C7,$D9,$FE,$59,$F1,$BF,$58,$FA,$FE
	.dc.b	$CD,$AF,$91,$16,$7F,$99,$4D,$C7,$F9,$B7,$7F,$DE,$EE,$FB,$ED,$C3
	.dc.b	$E8,$36,$FF,$9D,$36,$BD,$13,$FE,$19,$3E,$B7,$CD,$FC,$3F,$4F,$F1
	.dc.b	$2D,$8F,$E3,$E6,$EF,$E1,$CF,$C1,$7E,$6B,$5F,$F3,$43,$C5,$FE,$95
	.dc.b	$E3,$DB,$F9,$83,$E4,$7F,$97,$5F,$E1,$CF,$F4,$BB,$8A,$FE,$A5,$18
	.dc.b	$F2,$3F,$3A,$D3,$81,$FD,$0E,$7F,$8B,$D7,$66,$C7,$FA,$DD,$FE,$64
	.dc.b	$EF,$C2,$EE,$47,$97,$91,$70,$FF,$3E,$43,$E6,$3C,$BE,$4F,$E2,$5B
	.dc.b	$1F,$E9,$73,$47,$C8,$7E,$24,$FE,$4D,$B1,$A8,$FC,$6C,$73,$8F,$57
	.dc.b	$87,$67,$FB,$E6,$3F,$C9,$72,$F6,$78,$7F,$1B,$C3,$F8,$ED,$C3,$E1
	.dc.b	$36,$F8,$BF,$33,$C2,$93,$F1,$FC,$3B,$5D,$A6,$E1,$F2,$66,$D2,$A7
	.dc.b	$DA,$BE,$1B,$EB,$F8,$BF,$EC,$78,$C5,$7D,$38,$C4,$EA,$3F,$D7,$93
	.dc.b	$AA,$F6,$D6,$A7,$81,$AA,$E1,$3F,$EA,$DA,$E1,$87,$DD,$BE,$13,$CA
	.dc.b	$F8,$1C,$9E,$17,$CA,$BF,$71,$7C,$AB,$F0,$97,$CA,$7A,$47,$2B,$E8
	.dc.b	$72,$7A,$47,$2B,$E9,$1F,$16,$57,$D5,$FF,$76,$68,$C3,$B5,$78,$0E
	.dc.b	$D8,$78,$2B,$B1,$E0,$7B,$47,$80,$ED,$7D,$63,$F5,$55,$D2,$BB,$B6
	.dc.b	$33,$5D,$8F,$04,$F6,$8F,$01,$D9,$F0,$47,$6C,$3D,$B1,$F8,$06,$BE
	.dc.b	$E6,$DF,$8F,$0F,$75,$77,$AF,$1B,$6F,$C7,$1E,$18,$F1,$C7,$84,$F1
	.dc.b	$BE,$13,$E0,$47,$85,$AF,$85,$B5,$E1,$C3,$E0,$5F,$8A,$BB,$1E,$22
	.dc.b	$DF,$88,$F0,$5F,$88,$F0,$47,$C2,$3C,$05,$3C,$05,$BF,$04,$7B,$6B
	.dc.b	$C1,$5D,$2B,$C0,$FC,$D7,$F0,$C7,$22,$D7,$E8,$DB,$87,$F0,$88,$C4
	.dc.b	$79,$0C,$DF,$F9,$4A,$33,$1E,$6A,$F0,$1F,$CB,$7C,$5F,$C2,$47,$A2
	.dc.b	$33,$82,$73,$5F,$97,$AD,$5F,$F2,$79,$CE,$1F,$31,$B1,$EF,$6C,$FC
	.dc.b	$CB,$E9,$3F,$81,$8E,$67,$95,$E8,$66,$FC,$07,$E2,$6B,$9E,$1E,$66
	.dc.b	$CF,$D4,$8F,$E7,$91,$8A,$E8,$7E,$B9,$C3,$E0,$8F,$85,$3D,$F0,$F5
	.dc.b	$34,$F5,$16,$7D,$55,$CA,$FC,$93,$9B,$F2,$34,$F2,$1E,$CA,$FC,$79
	.dc.b	$F0,$0F,$16,$1F,$11,$CD,$7C,$0C,$3F,$AF,$E1,$DD,$B7,$D6,$BD,$B8
	.dc.b	$7A,$9E,$33,$EA,$6C,$FA,$9F,$A5,$F4,$FF,$DA,$BE,$47,$FB,$5B,$67
	.dc.b	$FB,$33,$67,$FA,$6B,$6F,$FB,$47,$CF,$FF,$72,$3F,$B9,$FD,$0D,$BF
	.dc.b	$ED,$D1,$6F,$FA,$AC,$DA,$FF,$12,$7E,$A1,$FE,$07,$7F,$B9,$7B,$5F
	.dc.b	$E0,$0C,$3F,$93,$4E,$5F,$EE,$91,$9B,$FE,$9D,$3F,$E0,$F7,$F4,$ED
	.dc.b	$78,$F5,$FF,$6C,$8C,$47,$08,$E4,$A3,$B3,$FC,$CB,$E8,$73,$18,$C3
	.dc.b	$FC,$25,$E6,$FF,$84,$3E,$53,$FC,$22,$DC,$BF,$C1,$8D,$C3,$FB,$9B
	.dc.b	$6B,$FC,$1D,$E8,$FF,$86,$0F,$F0,$CF,$BD,$6F,$FC,$29,$16,$FF,$BD
	.dc.b	$CD,$AF,$F2,$57,$C5,$FF,$BD,$EF,$FC,$35,$5E,$A8,$C3,$FE,$02,$31
	.dc.b	$ED,$7F,$83,$23,$31,$FC,$31,$AF,$F5,$36,$BF,$DE,$5E,$75,$E0,$9F
	.dc.b	$A0,$BE,$61,$F1,$FD,$B0,$FF,$77,$43,$E7,$36,$BE,$42,$C6,$1B,$3C
	.dc.b	$1B,$93,$0F,$F2,$0D,$F1,$AE,$BD,$5D,$0E,$BF,$37,$63,$24,$BE,$4F
	.dc.b	$F8,$4E,$FD,$13,$B9,$C9,$FE,$17,$FD,$41,$F2,$5E,$8F,$43,$F0,$B1
	.dc.b	$8C,$3D,$B1,$89,$FF,$1B,$CE,$6B,$F9,$B8,$70,$FF,$3A,$6B,$FE,$01
	.dc.b	$78,$61,$FE,$85,$6D,$F9,$F2,$7C,$A7,$78,$F9,$BF,$19,$F3,$3E,$63
	.dc.b	$E6,$B6,$32,$7D,$28,$CB,$4F,$AB,$F2,$9F,$50,$FA,$BF,$41,$63,$24
	.dc.b	$CF,$E2,$E7,$F7,$E7,$51,$F2,$0F,$77,$B1,$F7,$BB,$3E,$3B,$6F,$E8
	.dc.b	$1E,$F7,$E3,$91,$F8,$F3,$C9,$BA,$A7,$9C,$7E,$33,$E1,$AD,$78,$6B
	.dc.b	$FD,$81,$3C,$3D,$9F,$A6,$47,$9F,$B3,$F5,$88,$E0,$EF,$EC,$FD,$A2
	.dc.b	$38,$3C,$7E,$C7,$DE,$23,$83,$CF,$EC,$FE,$08,$8E,$0F,$5F,$B3,$F8
	.dc.b	$62,$38,$3D,$FE,$CF,$E2,$88,$E0,$F8,$7E,$CF,$E3,$88,$E0,$F8,$FE
	.dc.b	$CF,$E4,$88,$E0,$FF,$00,$77,$FF,$81,$48,$E0,$F9,$E7,$E6,$88,$60
	.dc.b	$FA,$BE,$03,$DF,$FE,$14,$48,$CB,$FE,$98,$E9,$FB,$11,$F7,$1B,$C3
	.dc.b	$FA,$46,$BF,$E1,$C6,$BF,$E6,$37,$ED,$BB,$E0,$E8,$49,$E9,$C1,$DC
	.dc.b	$8E,$0F,$C5,$FC,$9F,$9D,$B8,$7F,$AB,$23,$17,$FB,$66,$BF,$E2,$E7
	.dc.b	$3C,$FF,$EF,$76,$D7,$F9,$DD,$7B,$BF,$D2,$1C,$D7,$FA,$7D,$CC,$F8
	.dc.b	$A7,$C0,$47,$F3,$FF,$25,$E0,$AF,$F4,$13,$5F,$F5,$43,$53,$4E,$EE
	.dc.b	$30,$FE,$67,$F1,$2F,$67,$FA,$DD,$68,$93,$A8,$CB,$FD,$83,$E4,$5D
	.dc.b	$3F,$ED,$06,$E3,$FB,$46,$E3,$FC,$71,$AF,$F9,$51,$CE,$89,$FF,$4E
	.dc.b	$46,$1F,$E6,$1E,$CB,$BF,$C1,$87,$F7,$0F,$91,$79,$54,$C6,$57,$0F
	.dc.b	$E1,$2F,$FB,$2E,$77,$18,$84,$61,$FE,$28,$FF,$B4,$63,$4E,$72,$F3
	.dc.b	$4F,$F4,$3F,$8C,$FF,$00,$C1,$FF,$01,$C7,$FB,$67,$E4,$7F,$DB,$5F
	.dc.b	$1E,$F4,$FF,$D7,$17,$BC,$7C,$4D,$BF,$E4,$8D,$3B,$BF,$EB,$FB,$E7
	.dc.b	$5C,$B0,$EF,$29,$CB,$E5,$BA,$F1,$18,$78,$DD,$18,$BF,$EA,$35,$C3
	.dc.b	$0F,$25,$6E,$B9,$36,$33,$5C,$27,$91,$D2,$78,$3F,$0C,$C9,$FD,$D7
	.dc.b	$0F,$97,$1A,$8F,$51,$AA,$EA,$FF,$A6,$AF,$C9,$5C,$2F,$C1,$7E,$38
	.dc.b	$C5,$6C,$E4,$E1,$F8,$47,$F4,$8A,$FE,$92,$F9,$4C,$1F,$DC,$BE,$41
	.dc.b	$E2,$3E,$47,$FA,$7C,$61,$B1,$99,$E0,$D7,$FD,$63,$3F,$A6,$CB,$8E
	.dc.b	$DF,$E1,$59,$EC,$D7,$6D,$7F,$DF,$D3,$C5,$F9,$39,$FE,$81,$FB,$AD
	.dc.b	$7E,$6C,$6F,$3F,$B0,$D9,$4C,$78,$4F,$DD,$69,$E0,$AF,$34,$72,$6C
	.dc.b	$72,$9F,$CE,$6D,$6B,$93,$67,$96,$1F,$92,$C3,$C5,$7E,$1A,$D3,$4D
	.dc.b	$61,$D9,$F3,$CF,$07,$B9,$F0,$27,$8A,$8D,$5F,$C3,$BF,$D4,$B7,$5F
	.dc.b	$43,$6B,$AE,$1D,$DE,$A7,$F8,$56,$1E,$8F,$EE,$2B,$8B,$67,$D0,$7F
	.dc.b	$86,$FE,$47,$F8,$13,$FC,$35,$AE,$77,$E0,$A3,$F9,$4B,$B8,$E9,$8B
	.dc.b	$9D,$EF,$E1,$B5,$B3,$F8,$D7,$C8,$F2,$7F,$C9,$2D,$7F,$E0,$0B,$5F
	.dc.b	$B0,$D7,$F0,$6F,$1A,$FE,$3A,$3E,$21,$E0,$F9,$1B,$5E,$0D,$BD,$7F
	.dc.b	$22,$F1,$A9,$27,$F9,$8E,$0F,$D5,$FB,$70,$7E,$26,$F8,$57,$F8,$E9
	.dc.b	$AF,$F9,$89,$AF,$E3,$C3,$87,$95,$F9,$7E,$4A,$47,$F3,$D9,$E7,$87
	.dc.b	$F2,$AD,$3E,$9C,$FC,$4D,$FF,$80,$C6,$23,$E2,$C7,$F0,$4A,$78,$D4
	.dc.b	$93,$E7,$75,$94,$8F,$53,$5F,$A7,$F2,$7D,$88,$FF,$96,$30,$47,$F1
	.dc.b	$2F,$87,$7C,$55,$25,$7F,$97,$A3,$2D,$F8,$BE,$04,$70,$7F,$99,$3E
	.dc.b	$7B,$FC,$C4,$DF,$4B,$F8,$FC,$62,$3F,$D1,$ED,$7E,$77,$C9,$FD,$BF
	.dc.b	$94,$FF,$36,$B6,$3F,$D2,$AB,$D0,$FD,$2C,$63,$8D,$5F,$E7,$4F,$94
	.dc.b	$65,$FD,$6A,$BF,$84,$46,$37,$D4,$E6,$7F,$8A,$CE,$5F,$EA,$7E,$96
	.dc.b	$5F,$EB,$CD,$F1,$0D,$35,$F2,$6D,$E9,$31,$FF,$43,$C6,$5B,$F0,$5C
	.dc.b	$08,$E0,$FA,$DF,$3B,$E1,$E4,$34,$91,$F5,$7C,$AF,$36,$06,$92,$3E
	.dc.b	$D4,$68,$8E,$0F,$C2,$7C,$87,$72,$38,$3F,$05,$F2,$1D,$88,$E0,$FC
	.dc.b	$37,$C8,$78,$88,$E0,$FF,$39,$FC,$07,$FB,$CB,$BD,$F4,$1F,$F0,$06
	.dc.b	$30,$D9,$D9,$38,$7F,$DA,$BE,$CF,$FB,$49,$FF,$7F,$56,$DF,$F5,$FD
	.dc.b	$FF,$86,$6B,$FA,$35,$7F,$B5,$27,$FD,$E9,$3F,$2A,$7F,$82,$B6,$38
	.dc.b	$F6,$79,$76,$95,$A7,$D1,$18,$9E,$FF,$45,$EB,$D7,$D8,$DB,$EA,$D9
	.dc.b	$DC,$78,$27,$F8,$FB,$FC,$4F,$C2,$F8,$9F,$4D,$8A,$FE,$CF,$E0,$A7
	.dc.b	$35,$FC,$35,$FC,$A4,$66,$3F,$8D,$46,$08,$EC,$F7,$F5,$75,$AF,$E3
	.dc.b	$73,$DF,$C3,$FE,$61,$F3,$77,$C6,$37,$E7,$70,$CC,$7F,$96,$CE,$9B
	.dc.b	$39,$C3,$FD,$E3,$3F,$46,$FF,$9E,$7C,$9F,$5E,$3E,$CD,$C3,$9B,$6F
	.dc.b	$8B,$E7,$EB,$FF,$32,$47,$4F,$5F,$0F,$9D,$B5,$FA,$E8,$FD,$99,$1C
	.dc.b	$1F,$E7,$19,$E7,$87,$F5,$0C,$5F,$33,$CA,$FD,$3F,$33,$F0,$CF,$89
	.dc.b	$EE,$5C,$7E,$0C,$59,$F7,$CD,$C3,$A7,$BF,$F2,$FC,$9F,$98,$DC,$3E
	.dc.b	$4B,$6F,$E3,$B6,$BF,$CE,$32,$F8,$9F,$A9,$F0,$1F,$66,$30,$7D,$27
	.dc.b	$11,$FC,$5E,$33,$7F,$D6,$4F,$F4,$F4,$64,$91,$FE,$93,$37,$5F,$0F
	.dc.b	$E1,$39,$C7,$DE,$F1,$7F,$08,$7E,$17,$E1,$2C,$64,$98,$7E,$02,$32
	.dc.b	$FE,$0B,$CB,$FC,$67,$97,$F8,$D6,$D7,$E2,$9B,$3B,$8F,$C6,$CF,$BF
	.dc.b	$DB,$FC,$A7,$C2,$59,$FC,$8B,$6B,$F1,$ED,$CF,$EE,$36,$7F,$CA,$31
	.dc.b	$FE,$0D,$8C,$4A,$31,$3F,$0E,$BB,$B4,$EE,$59,$EE,$D9,$EE,$D3,$BF
	.dc.b	$1A,$FB,$CF,$6A,$F0,$57,$F9,$6C,$FF,$3F,$9D,$9C,$9F,$EC,$3B,$F1
	.dc.b	$B8,$3F,$DB,$0E,$27,$F0,$ED,$9E,$8D,$9E,$51,$D6,$76,$D3,$65,$8D
	.dc.b	$9E,$E3,$C0,$F9,$9C,$47,$26,$C7,$26,$9C,$A3,$E2,$BC,$A7,$C8,$FA
	.dc.b	$A7,$C0,$D3,$89,$67,$89,$DC,$B1,$DD,$B1,$DE,$3E,$8D,$7D,$0B,$FA
	.dc.b	$51,$F5,$67,$F0,$11,$87,$C9,$1E,$23,$DD,$5D,$23,$C5,$1D,$0E,$51
	.dc.b	$F6,$CF,$23,$98,$E6,$7A,$9E,$66,$63,$71,$E2,$AF,$31,$27,$AD,$79
	.dc.b	$6B,$67,$78,$F8,$07,$32,$D7,$37,$F2,$CF,$F9,$65,$FB,$0D,$BF,$E0
	.dc.b	$C4,$8F,$E0,$D1,$F4,$4D,$38,$38,$BD,$8F,$86,$7C,$AF,$47,$F5,$6E
	.dc.b	$0C,$D7,$F9,$EE,$33,$3F,$0A,$7E,$B1,$C5,$FA,$8F,$17,$53,$C5,$C9
	.dc.b	$C6,$3A,$DF,$F9,$9A,$33,$7F,$7E,$FB,$B6,$32,$7C,$C8,$CB,$4C,$D7
	.dc.b	$D8,$BF,$E8,$93,$A8,$C4,$FE,$F2,$31,$3F,$52,$FC,$55,$A3,$DA,$79
	.dc.b	$A7,$75,$F5,$AB,$D9,$3B,$3E,$05,$FF,$71,$AD,$C7,$C6,$9D,$9F,$88
	.dc.b	$77,$5F,$DC,$A7,$71,$DE,$3D,$73,$E5,$9C,$C7,$CC,$BF,$3C,$F3,$8F
	.dc.b	$98,$66,$77,$7E,$38,$E6,$7A,$6B,$8B,$F8,$33,$8C,$6E,$FE,$15,$BE
	.dc.b	$2D,$F3,$3F,$E0,$DB,$FF,$08,$46,$1F,$F0,$59,$88,$D3,$FE,$BA,$8D
	.dc.b	$57,$F8,$32,$38,$FA,$11,$D7,$A7,$67,$F9,$4E,$1E,$1A,$C3,$FE,$CE
	.dc.b	$BF,$F4,$FD,$F5,$9F,$ED,$9E,$1F,$93,$E6,$6C,$F8,$5B,$87,$F6,$B8
	.dc.b	$4F,$F6,$F2,$3F,$A7,$77,$87,$7F,$84,$B3,$FD,$4B,$1F,$F4,$15,$F9
	.dc.b	$9F,$F3,$C4,$7B,$FC,$EF,$8B,$E7,$FC,$66,$C6,$4F,$89,$1F,$2B,$CF
	.dc.b	$91,$F1,$FD,$AB,$87,$59,$B9,$7C,$19,$B4,$A9,$F3,$2F,$0F,$EA,$FA
	.dc.b	$3F,$E4,$78,$C5,$7D,$38,$C4,$FF,$A9,$67,$35,$FE,$0F,$6B,$FE,$E9
	.dc.b	$6B,$DC,$3C,$30,$5E,$63,$0F,$C7,$F9,$9E,$EF,$E0,$FE,$63,$F1,$3F
	.dc.b	$33,$F8,$A6,$C6,$4F,$C4,$46,$5A,$7E,$2F,$E6,$7F,$22,$7E,$4B,$E6
	.dc.b	$6C,$64,$99,$7D,$58,$CB,$FD,$C7,$E6,$3F,$8A,$F1,$3E,$13,$FC,$D9
	.dc.b	$18,$6F,$57,$FB,$26,$3F,$94,$D7,$CF,$6A,$69,$FC,$74,$EB,$F3,$3C
	.dc.b	$B1,$FD,$17,$B7,$FA,$8F,$99,$FD,$4B,$63,$27,$E9,$E3,$2D,$3F,$55
	.dc.b	$F0,$DF,$B0,$3F,$63,$F1,$16,$32,$4C,$7F,$29,$19,$7F,$C2,$1E,$E3
	.dc.b	$E1,$7F,$73,$1F,$E8,$B8,$C5,$7E,$DE,$31,$3F,$EE,$99,$CC,$7F,$AE
	.dc.b	$A3,$04,$9F,$F1,$A9,$FE,$B8,$7F,$D2,$9F,$2A,$F9,$9E,$73,$AA,$FF
	.dc.b	$52,$F9,$FF,$EA,$22,$47,$F5,$AF,$69,$1F,$EF,$9F,$93,$7A,$34,$95
	.dc.b	$FD,$82,$32,$DF,$47,$81,$1C,$1D,$67,$FD,$62,$47,$07,$3F,$3B,$C0
	.dc.b	$47,$07,$7F,$3F,$C8,$47,$07,$A6,$BF,$D7,$A4,$70,$79,$FC,$FF,$61
	.dc.b	$1C,$1E,$BF,$3B,$E0,$11,$C1,$F1,$7D,$0F,$F6,$49,$1C,$1F,$E0,$2F
	.dc.b	$03,$EB,$ED,$FB,$9E,$1F,$DF,$12,$7E,$41,$F3,$48,$60,$FF,$05,$FC
	.dc.b	$EF,$FB,$DE,$4D,$27,$B1,$FC,$02,$3F,$D2,$DF,$29,$F6,$2B,$FC,$26
	.dc.b	$D8,$CF,$0E,$DF,$E1,$42,$34,$E1,$FB,$6F,$4E,$0E,$0F,$F1,$18,$CB
	.dc.b	$B8,$CB,$FC,$8F,$E5,$1E,$D7,$FB,$9F,$E7,$5D,$BF,$EE,$99,$CE,$5F
	.dc.b	$93,$B5,$6A,$31,$3A,$9F,$F5,$93,$5F,$4B,$5F,$E4,$8D,$7C,$57,$DC
	.dc.b	$F0,$7B,$16,$BC,$1F,$2B,$D7,$C9,$F9,$43,$E7,$5C,$7F,$EF,$F9,$D5
	.dc.b	$7F,$A8,$A7,$2F,$2C,$3E,$0E,$2A,$F1,$36,$33,$18,$39,$C7,$F4,$96
	.dc.b	$E0,$F8,$F7,$A5,$E1,$BF,$2E,$1F,$D0,$7E,$59,$F2,$18,$CB,$FD,$65
	.dc.b	$E9,$FD,$BA,$F9,$95,$A7,$F8,$06,$49,$D5,$7F,$A8,$EB,$4F,$FB,$7D
	.dc.b	$AF,$E1,$9A,$F8,$6F,$DF,$F2,$AF,$13,$C5,$28,$C6,$3F,$75,$FE,$CC
	.dc.b	$D9,$DC,$6A,$7F,$D6,$8D,$C7,$F2,$CF,$08,$CC,$F6,$97,$C8,$D3,$C3
	.dc.b	$3F,$DC,$FE,$3B,$FB,$A7,$CA,$7F,$4D,$8F,$F6,$CC,$EF,$1F,$C8,$5F
	.dc.b	$27,$F2,$53,$C4,$57,$17,$FD,$BD,$5D,$95,$F6,$F3,$7F,$BA,$61,$CE
	.dc.b	$78,$C7,$EB,$9D,$98,$3C,$78,$71,$7F,$67,$55,$EF,$15,$F2,$61,$E3
	.dc.b	$BE,$CF,$96,$33,$7E,$2C,$3F,$82,$4E,$67,$C2,$F1,$8C,$CF,$97,$0F
	.dc.b	$05,$F4,$8C,$5F,$9C,$F1,$94,$EE,$7A,$8C,$1D,$9F,$16,$1E,$98,$F8
	.dc.b	$27,$FA,$9A,$BF,$D4,$EF,$CF,$30,$7C,$7F,$94,$79,$CF,$95,$FF,$56
	.dc.b	$C6,$1B,$19,$9E,$0D,$7F,$8E,$CF,$E9,$B1,$E3,$B1,$A4,$9F,$04,$65
	.dc.b	$F8,$19,$7E,$75,$FA,$8D,$7E,$25,$75,$6C,$F1,$D3,$F1,$15,$B6,$EB
	.dc.b	$FE,$39,$7F,$19,$7C,$5B,$5C,$6B,$FC,$7C,$71,$2D,$71,$7F,$C8,$55
	.dc.b	$C5,$B3,$C6,$FF,$CA,$91,$C5,$B5,$C5,$FF,$2C,$57,$16,$D7,$17,$FC
	.dc.b	$69,$3B,$69,$F9,$59,$E0,$E0,$F9,$9A,$79,$9F,$E7,$97,$F1,$F1,$FE
	.dc.b	$A1,$5C,$23,$15,$EE,$F3,$3F,$52,$DB,$E3,$86,$1F,$F3,$77,$CA,$FF
	.dc.b	$00,$7F,$A4,$35,$F0,$9A,$DF,$82,$8F,$E4,$EF,$08,$E9,$97,$F2,$FF
	.dc.b	$66,$F3,$1E,$37,$D6,$7F,$B4,$7C,$97,$2C,$3B,$87,$11,$AD,$BA,$FE
	.dc.b	$B1,$FD,$BA,$31,$5F,$C0,$CE,$13,$FA,$2A,$E9,$1F,$DD,$1A,$FF,$9F
	.dc.b	$FC,$D7,$C2,$FB,$1B,$1F,$DA,$5A,$FF,$AE,$3E,$4B,$C7,$BF,$1A,$E0
	.dc.b	$49,$FE,$87,$3A,$7F,$D3,$55,$C9,$FF,$02,$B5,$E4,$7F,$91,$67,$F7
	.dc.b	$3F,$A9,$BD,$BA,$49,$F7,$46,$5F,$81,$8F,$D7,$7D,$DC,$1F,$04,$8F
	.dc.b	$F9,$47,$E1,$FF,$82,$E2,$69,$23,$AF,$B5,$E2,$23,$83,$BF,$C9,$7C
	.dc.b	$D2,$38,$3E,$97,$C1,$3E,$1B,$F8,$9B,$63,$31,$83,$F8,$BF,$A4,$F9
	.dc.b	$CA,$49,$FB,$91,$97,$EF,$7C,$C7,$F9,$A9,$BE,$7F,$F2,$08,$C4,$7F
	.dc.b	$45,$6B,$FE,$34,$F9,$8F,$BF,$F2,$7F,$87,$6C,$7F,$4B,$56,$ED,$D7
	.dc.b	$22,$FE,$8F,$C1,$7F,$85,$39,$0D,$27,$3F,$F3,$3F,$67,$FC,$33,$E2
	.dc.b	$9A,$48,$FC,$C7,$C8,$7F,$BC,$35,$FF,$24,$B5,$FF,$6A,$35,$FF,$72
	.dc.b	$35,$61,$FE,$BF,$9F,$C2,$F8,$5E,$52,$3F,$86,$F9,$6F,$F1,$5F,$51
	.dc.b	$A4,$8F,$B3,$E5,$7F,$8C,$39,$BF,$E0,$0E,$3F,$DB,$DA,$FF,$97,$9A
	.dc.b	$FF,$BA,$71,$FF,$27,$3F,$DD,$5A,$FF,$99,$1A,$EF,$D3,$F1,$FC,$54
	.dc.b	$C3,$F3,$6D,$9F,$07,$C5,$78,$C9,$3F,$8E,$7E,$97,$8F,$FE,$37,$F0
	.dc.b	$DE,$17,$FA,$D6,$BF,$CC,$59,$7F,$88,$56,$1F,$D5,$E7,$4F,$F1,$09
	.dc.b	$FF,$14,$3E,$7B,$EB,$5F,$B0,$6C,$6D,$F0,$53,$E7,$7F,$C6,$F7,$DE
	.dc.b	$BB,$76,$68,$DF,$CF,$5C,$7F,$83,$C5,$9F,$E1,$D3,$72,$FF,$34,$E2
	.dc.b	$F9,$1E,$FF,$13,$E0,$6E,$3F,$C4,$1B,$87,$F1,$76,$DF,$F2,$EA,$F3
	.dc.b	$FB,$7F,$E4,$CE,$44,$63,$0F,$F4,$44,$61,$FE,$1B,$1F,$88,$23,$83
	.dc.b	$6F,$C2,$F9,$BE,$8F,$C4,$F9,$C9,$3F,$C9,$3E,$7A,$E7,$FC,$AE,$6E
	.dc.b	$5E,$E9,$B5,$F5,$A3,$ED,$FA,$5F,$3B,$E7,$BE,$7B,$67,$F9,$7B,$70
	.dc.b	$FE,$6C,$DA,$FF,$5A,$7C,$53,$E0,$7E,$BF,$CE,$FF,$9B,$23,$18,$7F
	.dc.b	$38,$8C,$4F,$FA,$42,$7F,$D8,$C4,$70,$6A,$7F,$9F,$F9,$D7,$1F,$E8
	.dc.b	$DD,$BF,$84,$3F,$0B,$E5,$DA,$FE,$8B,$16,$7F,$A6,$4D,$CB,$EF,$FC
	.dc.b	$EF,$E3,$3B,$3F,$1A,$DC,$3F,$A6,$B6,$FF,$AA,$36,$BF,$DA,$5E,$7B
	.dc.b	$E7,$3F,$2B,$F3,$BF,$E8,$A8,$C6,$1F,$D5,$A3,$13,$FE,$AE,$9C,$E1
	.dc.b	$FE,$6D,$6B,$FD,$62,$34,$70,$AF,$EB,$FE,$45,$9F,$F0,$3E,$0F,$99
	.dc.b	$7C,$EF,$ED,$5B,$7F,$EA,$85,$F3,$B9,$99,$39,$78,$46,$2F,$AD,$7D
	.dc.b	$86,$FC,$BF,$5E,$2E,$CD,$C3,$B3,$67,$71,$FE,$E5,$F2,$7F,$D4,$59
	.dc.b	$7F,$B5,$B0,$FF,$64,$D2,$73,$5F,$93,$9C,$C6,$27,$FC,$C1,$18,$7D
	.dc.b	$DF,$03,$FE,$B8,$9D,$5F,$C3,$8C,$4E,$AF,$C4,$D3,$FC,$43,$5C,$2F
	.dc.b	$A4,$FC,$7A,$E2,$D3,$89,$6B,$89,$CE,$D1,$FD,$2A,$30,$D3,$84,$6A
	.dc.b	$3F,$A6,$9A,$2C,$68,$E0,$FD,$29,$F1,$B6,$3C,$6D,$3C,$6F,$17,$99
	.dc.b	$99,$F0,$3F,$3C,$F0,$16,$38,$34,$E0,$F8,$A3,$EC,$4F,$89,$B1,$E2
	.dc.b	$6C,$F8,$8E,$27,$89,$C9,$E2,$BF,$A1,$5C,$EB,$F0,$0D,$AF,$BE,$DA
	.dc.b	$FB,$93,$DC,$D9,$67,$6D,$8D,$C7,$8D,$F0,$C7,$AC,$E3,$19,$6C,$79
	.dc.b	$63,$E2,$D7,$9A,$7A,$5F,$E1,$A7,$DA,$7C,$7A,$EA,$FB,$E7,$99,$DD
	.dc.b	$F2,$9D,$CB,$1D,$E3,$E8,$D7,$F6,$87,$FC,$2F,$3F,$60,$FC,$19,$FE
	.dc.b	$18,$30,$7D,$E3,$0F,$02,$3F,$EF,$98,$FD,$53,$E7,$8F,$D7,$7C,$EF
	.dc.b	$CF,$8C,$1E,$93,$F5,$EE,$67,$89,$FA,$77,$11,$F5,$6D,$18,$AF,$8E
	.dc.b	$D9,$F8,$17,$FC,$5E,$30,$FC,$18,$FE,$1F,$18,$9F,$91,$3B,$7E,$F4
	.dc.b	$F9,$8F,$3C,$65,$F9,$86,$6F,$CD,$3C,$1F,$4C,$7F,$25,$AE,$CF,$BE
	.dc.b	$B9,$1E,$32,$57,$8A,$FD,$AF,$39,$F8,$15,$C9,$F1,$CE,$9F,$1D,$F8
	.dc.b	$2B,$F9,$95,$6E,$3C,$B3,$A3,$9C,$F5,$3E,$49,$5F,$93,$5D,$6F,$CB
	.dc.b	$1F,$5C,$F0,$DF,$57,$DD,$5C,$67,$67,$B9,$FC,$A4,$7C,$73,$67,$6F
	.dc.b	$9D,$FE,$94,$7F,$4B,$F8,$8B,$3F,$D2,$E2,$CF,$F4,$69,$B5,$FC,$F7
	.dc.b	$E2,$FF,$A9,$7A,$FF,$D4,$DB,$87,$F5,$36,$DF,$F4,$E6,$D7,$F1,$69
	.dc.b	$FF,$19,$1F,$D6,$7E,$43,$F9,$44,$67,$C5,$FE,$BD,$18,$9F,$F2,$7D
	.dc.b	$FF,$32,$9F,$D2,$B6,$3F,$DB,$07,$47,$67,$EB,$E3,$71,$9A,$EC,$FF
	.dc.b	$69,$F8,$AE,$D5,$B8,$FE,$D9,$E9,$7F,$6F,$3F,$B8,$7C,$8D,$8C,$92
	.dc.b	$7F,$B5,$C6,$5F,$ED,$BE,$C7,$F7,$7F,$5F,$FB,$C3,$6B,$FB,$C3,$67
	.dc.b	$71,$FD,$EE,$7C,$71,$FD,$43,$C3,$FF,$00,$7C,$8E,$29,$18,$7F,$C7
	.dc.b	$C7,$C6,$32,$7C,$47,$E0,$91,$C1,$C2,$7F,$8A,$47,$F8,$37,$DE,$EC
	.dc.b	$FF,$A0,$FC,$97,$CE,$FF,$84,$6F,$0D,$79,$6D,$FE,$62,$8C,$61,$FE
	.dc.b	$1A,$AE,$35,$FE,$17,$3A,$7C,$EC,$78,$C7,$FB,$7D,$B1,$99,$E4,$F8
	.dc.b	$FD,$FC,$61,$F8,$BF,$13,$CD,$F3,$BF,$15,$B8,$7B,$70,$FE,$F3,$1F
	.dc.b	$82,$24,$7E,$21,$B0,$AE,$93,$E4,$F8,$DE,$8F,$9B,$DA,$F8,$9E,$DF
	.dc.b	$C5,$6C,$64,$F5,$C6,$5A,$7B,$BD,$AF,$98,$7C,$DF,$87,$B1,$92,$63
	.dc.b	$F2,$63,$2F,$96,$3F,$47,$F2,$8F,$0F,$EB,$7C,$EF,$E2,$23,$13,$F3
	.dc.b	$23,$13,$FA,$89,$CD,$7E,$98,$38,$7F,$52,$D7,$BC,$72,$AF,$AF,$F3
	.dc.b	$B6,$7C,$87,$FB,$13,$D7,$7A,$DF,$EE,$62,$57,$79,$4E,$A3,$1E,$6B
	.dc.b	$C2,$FF,$35,$12,$32,$FF,$00,$F9,$77,$87,$49,$3D,$63,$2F,$7B,$7A
	.dc.b	$DF,$EF,$82,$46,$5F,$F3,$C5,$48,$F1,$B7,$1F,$8D,$6F,$03,$C4,$4A
	.dc.b	$F0,$3F,$C5,$30,$FD,$8D,$7C,$D6,$BF,$C5,$1A,$FF,$1C,$6B,$FE,$EC
	.dc.b	$C9,$E2,$7F,$A2,$C9,$7F,$19,$F8,$5E,$0C,$8F,$80,$DA,$FB,$9E,$AF
	.dc.b	$FA,$41,$FE,$52,$D7,$FD,$F2,$D4,$DB,$78,$9F,$E9,$B2,$5F,$D6,$7F
	.dc.b	$15,$F2,$F2,$3E,$E7,$1A,$31,$87,$E3,$7C,$CF,$F7,$E5,$7E,$05,$B1
	.dc.b	$9F,$7F,$F9,$DC,$62,$3F,$9C,$FC,$BF,$FA,$55,$BE,$7F,$F4,$08,$C4
	.dc.b	$7F,$44,$6A,$1E,$E7,$E2,$F0,$FC,$83,$63,$F4,$0D,$7F,$A3,$B5,$FE
	.dc.b	$9F,$F2,$EF,$83,$FE,$9E,$DA,$FC,$D3,$67,$F5,$4D,$7F,$A8,$FC,$BF
	.dc.b	$FA,$8A,$35,$E1,$7F,$9D,$A3,$17,$FC,$F9,$67,$FD,$71,$B8,$7E,$9D
	.dc.b	$B8,$7E,$DD,$AF,$F5,$EF,$96,$CE,$1F,$DA,$BE,$5F,$EC,$DF,$F2,$19
	.dc.b	$CB,$F1,$65,$3A,$AF,$DC,$8A,$DB,$FC,$8E,$35,$3F,$E8,$E9,$CB,$CB
	.dc.b	$87,$FA,$BC,$79,$2B,$FB,$BB,$5F,$A8,$D7,$B6,$3F,$DE,$BE,$5D,$F2
	.dc.b	$1F,$DC,$B0,$78,$9F,$6B,$3F,$E7,$55,$A7,$FA,$7C,$67,$0F,$4C,$E5
	.dc.b	$FF,$74,$C6,$6B,$FD,$4B,$39,$7F,$C1,$F1,$9A,$FF,$06,$FC,$BB,$E2
	.dc.b	$7F,$B8,$B6,$3F,$A5,$3F,$E6,$E3,$18,$7C,$86,$DF,$FA,$C9,$B3,$D2
	.dc.b	$79,$D7,$F7,$CA,$7C,$15,$3F,$BF,$CF,$E6,$7E,$2F,$F3,$5F,$2C,$F8
	.dc.b	$CF,$F1,$0D,$F0,$C7,$FA,$E3,$6F,$EC,$B4,$FF,$44,$DF,$2A,$E3,$87
	.dc.b	$6A,$4E,$5F,$C2,$47,$FB,$67,$0F,$04,$7F,$8A,$A7,$4F,$F6,$59,$D5
	.dc.b	$7F,$49,$9C,$DF,$8E,$F8,$DF,$F8,$E2,$73,$7C,$23,$A0,$9D,$D7,$1B
	.dc.b	$E9,$39,$7D,$38,$78,$A3,$E3,$19,$3D,$2E,$5F,$69,$93,$D0,$E7,$0F
	.dc.b	$1B,$63,$55,$E1,$7C,$A7,$F5,$79,$FE,$B0,$F8,$0C,$1F,$DA,$7E,$5B
	.dc.b	$FC,$67,$F2,$FF,$D7,$5B,$19,$F5,$53,$96,$C7,$1E,$0F,$F0,$83,$70
	.dc.b	$E5,$18,$8F,$F2,$DD,$FE,$9A,$72,$7D,$4C,$BF,$31,$37,$5F,$F3,$3E
	.dc.b	$9F,$4A,$F8,$4F,$C3,$6C,$7D,$6A,$E3,$3F,$1D,$B1,$C7,$0F,$2B,$70
	.dc.b	$ED,$50,$E9,$3A,$AF,$F3,$BD,$F6,$7F,$CF,$27,$42,$E1,$E6,$77,$3C
	.dc.b	$6B,$F9,$D5,$71,$7F,$9E,$4F,$16,$D7,$17,$75,$C5,$FD,$44,$7E,$A3
	.dc.b	$4E,$F1,$FE,$43,$AF,$7E,$09,$F0,$BE,$B3,$13,$E0,$F9,$0E,$4D,$39
	.dc.b	$5E,$17,$CB,$7E,$F1,$FF,$28,$35,$FE,$5A,$DA,$EB,$1C,$31,$E5,$39
	.dc.b	$C3,$F9,$B7,$CB,$BC,$FF,$0E,$08,$C6,$1D,$B8,$7C,$5F,$AB,$7F,$BD
	.dc.b	$A3,$FC,$E7,$FA,$CF,$FA,$19,$BC,$3A,$F8,$29,$18,$21,$FD,$0A,$B9
	.dc.b	$3F,$D1,$3F,$57,$7A,$7F,$A2,$6C,$7F,$8E,$DA,$FF,$92,$DA,$FF,$AC
	.dc.b	$9C,$E7,$FE,$BE,$C7,$F9,$FC,$F9,$B7,$F9,$55,$AD,$BF,$D7,$ED,$9F
	.dc.b	$F3,$A5,$70,$7F,$C9,$F6,$8C,$3F,$32,$78,$49,$7C,$CF,$A4,$91,$E9
	.dc.b	$7D,$9F,$33,$23,$D8,$D7,$D9,$F2,$FF,$04,$8F,$C1,$7E,$27,$2C,$8F
	.dc.b	$88,$D7,$E2,$78,$5F,$24,$8F,$C9,$7E,$67,$CE,$C8,$F9,$8D,$7F,$CE
	.dc.b	$DF,$21,$FE,$06,$7F,$CD,$4D,$73,$E0,$7F,$16,$8C,$47,$FB,$89,$7C
	.dc.b	$1E,$49,$1C,$9F,$B9,$F3,$D2,$3E,$E3,$5F,$F4,$37,$CD,$3E,$19,$F1
	.dc.b	$FF,$E7,$B8,$C3,$7D,$37,$7E,$6B,$77,$FE,$8F,$FA,$17,$C2,$FF,$BE
	.dc.b	$9B,$CF,$FE,$85,$6C,$EE,$7F,$DF,$FF,$44,$F9,$7A,$4B,$FF,$46,$46
	.dc.b	$5F,$CA,$78,$7F,$AE,$BF,$C5,$B5,$FF,$23,$74,$7F,$7A,$7F,$2D,$19
	.dc.b	$7F,$11,$9D,$9F,$5D,$7E,$7E,$30,$66,$64,$7E,$8D,$A7,$FA,$B7,$F5
	.dc.b	$77,$D0,$75,$6C,$FF,$A8,$DB,$3B,$9D,$7A,$B2,$B4,$D9,$E9,$F3,$9C
	.dc.b	$C9,$3C,$DF,$57,$96,$F8,$EF,$F1,$67,$80,$F9,$AF,$E3,$BE,$0F,$F1
	.dc.b	$2A,$D6,$1F,$C2,$27,$2F,$F7,$9B,$E6,$47,$FB,$23,$71,$FF,$24,$D7
	.dc.b	$27,$F6,$B8,$7F,$84,$EB,$91,$FC,$53,$E6,$BC,$7D,$7E,$46,$C6,$4F
	.dc.b	$E3,$31,$96,$9F,$C7,$3E,$6B,$D6,$7B,$3E,$32,$C6,$49,$87,$8A,$32
	.dc.b	$FF,$97,$BE,$59,$F2,$3F,$13,$E6,$BD,$F3,$FE,$89,$9C,$DF,$F2,$A9
	.dc.b	$CC,$A7,$2F,$FB,$A5,$AF,$FA,$44,$E5,$5F,$8B,$8C,$C6,$5E,$3B,$FF
	.dc.b	$42,$6D,$FF,$9D,$74,$FD,$6B,$FE,$77,$38,$5F,$F9,$97,$DF,$7C,$CD
	.dc.b	$F9,$9C,$36,$B8,$64,$CB,$FD,$69,$39,$8F,$E5,$37,$CA,$B7,$5F,$EB
	.dc.b	$9B,$ED,$F0,$32,$73,$4F,$8E,$C6,$1C,$AA,$E1,$F8,$CA,$D5,$F0,$9F
	.dc.b	$93,$A3,$E4,$3D,$55,$86,$D7,$A3,$0F,$9D,$1F,$24,$91,$F3,$63,$E5
	.dc.b	$46,$1F,$B8,$FC,$5F,$9A,$E9,$3E,$8F,$4B,$E4,$1F,$27,$C9,$B1,$92
	.dc.b	$5F,$C4,$8C,$BF,$1B,$E6,$7E,$87,$CC,$FD,$16,$D7,$C4,$6C,$EE,$3E
	.dc.b	$3C,$FC,$0F,$47,$EC,$7C,$85,$9F,$96,$DA,$FC,$54,$FE,$1C,$93,$EF
	.dc.b	$8F,$CA,$46,$F7,$F9,$DF,$1F,$6B,$FD,$79,$B7,$FA,$D9,$BF,$07,$F8
	.dc.b	$7F,$8C,$FC,$43,$23,$ED,$B9,$69,$F7,$BB,$BF,$1D,$F1,$DF,$8F,$6C
	.dc.b	$64,$FC,$1C,$65,$A7,$D9,$E3,$7A,$1F,$97,$F3,$7F,$02,$D9,$D3,$76
	.dc.b	$FD,$A3,$75,$FD,$23,$67,$F5,$0D,$9F,$DB,$35,$FD,$A3,$6F,$C3,$3F
	.dc.b	$98,$BF,$0B,$4F,$09,$6B,$C2,$DB,$F0,$B4,$F0,$96,$3C,$2D,$7C,$2D
	.dc.b	$AF,$0C,$7E,$BA,$BC,$2D,$9F,$0B,$4F,$09,$6B,$C2,$D8,$F0,$B4,$F0
	.dc.b	$94,$F0,$96,$BC,$33,$FB,$3A,$F0,$B6,$BC,$2D,$AF,$0B,$4F,$09,$63
	.dc.b	$C2,$D8,$F0,$B5,$F0,$B6,$BC,$33,$F9,$4A,$F0,$B4,$F0,$96,$BC,$2D
	.dc.b	$AF,$0B,$4F,$09,$63,$C2,$D7,$C2,$DA,$F0,$C7,$F2,$1D,$3F,$C3,$13
	.dc.b	$E0,$8C,$5F,$91,$B1,$9B,$57,$09,$F2,$9A,$2C,$6A,$B9,$36,$B9,$57
	.dc.b	$18,$FB,$75,$FC,$66,$FF,$0A,$DC,$3F,$A3,$12,$FF,$80,$47,$DC,$34
	.dc.b	$E2,$78,$C7,$C4,$3E,$19,$EF,$F3,$BD,$8F,$C4,$33,$7E,$C8,$CC,$FC
	.dc.b	$69,$D4,$E6,$BE,$34,$71,$33,$82,$F8,$E1,$F0,$23,$15,$C3,$0E,$87
	.dc.b	$D2,$C3,$A6,$1F,$28,$E9,$19,$AE,$31,$98,$E3,$7F,$D7,$F0,$EF,$87
	.dc.b	$CC,$6D,$7D,$2B,$E3,$87,$E2,$23,$E1,$9E,$89,$DC,$64,$F1,$9B,$8F
	.dc.b	$1D,$F7,$BF,$53,$ED,$AE,$31,$F3,$EB,$8B,$F4,$6B,$8B,$FD,$9A,$77
	.dc.b	$1F,$42,$FC,$B3,$BC,$3D,$AE,$CE,$E7,$B4,$B3,$ED,$AE,$6F,$AE,$33
	.dc.b	$7E,$BB,$EB,$3F,$02,$FA,$BF,$DF,$AF,$8C,$7C,$F7,$EC,$7C,$C7,$F8
	.dc.b	$2C,$FF,$06,$7A,$57,$1F,$EF,$F1,$6F,$FC,$0F,$37,$1F,$F2,$4F,$D0
	.dc.b	$7F,$86,$27,$F5,$D7,$FE,$2A,$94,$E5,$FD,$5C,$66,$31,$97,$FB,$C6
	.dc.b	$3F,$C2,$31,$87,$FC,$5F,$F3,$5C,$2F,$FD,$7F,$2F,$99,$CD,$F3,$6D
	.dc.b	$FE,$A2,$F8,$5F,$F8,$1D,$AF,$F8,$D5,$A8,$3D,$F0,$FF,$7A,$B6,$75
	.dc.b	$1E,$CF,$4B,$BD,$79,$3E,$63,$D8,$7B,$7E,$66,$DF,$B2,$2C,$FA,$A6
	.dc.b	$DF,$CA,$EE,$F8,$9F,$13,$F1,$5B,$87,$C4,$6D,$FC,$26,$D2,$7F,$CA
	.dc.b	$10,$C3,$E7,$7C,$27,$F9,$62,$31,$3F,$3A,$31,$3F,$E9,$F9,$CD,$FF
	.dc.b	$83,$DA,$FF,$8B,$DA,$F6,$8E,$35,$F6,$3E,$66,$CF,$D6,$F9,$9F,$BC
	.dc.b	$7D,$FF,$82,$B5,$F7,$A2,$CF,$DC,$9B,$97,$E3,$3B,$7F,$0D,$F4,$7F
	.dc.b	$87,$6E,$1F,$86,$6D,$FE,$11,$B5,$F6,$39,$BF,$1D,$F0,$5F,$8A,$6D
	.dc.b	$FF,$93,$78,$BF,$87,$E1,$F8,$59,$FF,$60,$46,$23,$F0,$D1,$83,$94
	.dc.b	$FE,$5B,$E0,$AC,$FE,$53,$E0,$7F,$38,$7E,$77,$E0,$AD,$7E,$72,$2C
	.dc.b	$FE,$6A,$6E,$5F,$AA,$EC,$FD,$27,$C1,$7E,$95,$B8,$7E,$91,$B7,$FA
	.dc.b	$26,$D7,$E0,$63,$F3,$5F,$40,$E7,$FD,$77,$BD,$FE,$98,$8C,$61,$FA
	.dc.b	$E8,$C4,$FF,$B7,$E7,$35,$FE,$4F,$6B,$FE,$6F,$6B,$FE,$9D,$77,$7E
	.dc.b	$18,$46,$2F,$F4,$FD,$9F,$EE,$89,$FF,$49,$47,$1B,$F2,$73,$78,$49
	.dc.b	$1F,$D8,$2F,$F1,$F5,$EC,$E9,$F3,$12,$74,$FA,$7A,$24,$71,$6C,$FA
	.dc.b	$FA,$7D,$C4,$9E,$CF,$C1,$E9,$91,$E1,$6C,$FC,$3E,$8F,$8C,$49,$F3
	.dc.b	$3F,$06,$3F,$D8,$10,$F8,$6C,$47,$F0,$D8,$C1,$FD,$F0,$AF,$F9,$F5
	.dc.b	$AF,$FA,$77,$E5,$7F,$D5,$2F,$FA,$05,$AE,$7C,$FE,$AA,$32,$FF,$B4
	.dc.b	$7E,$5F,$F8,$EA,$69,$3C,$0E,$3F,$33,$F9,$F9,$FE,$44,$D4,$D2,$78
	.dc.b	$5C,$09,$8F,$F8,$5A,$11,$D3,$C3,$93,$96,$D7,$FA,$6F,$E5,$9E,$87
	.dc.b	$A9,$B5,$FE,$20,$6C,$EE,$75,$E7,$4A,$D3,$67,$A7,$9F,$CC,$93,$CD
	.dc.b	$F8,$FF,$2A,$F4,$E9,$27,$FC,$5F,$19,$6C,$7F,$C0,$08,$C5,$7E,$68
	.dc.b	$FE,$90,$BD,$3D,$4E,$DE,$E4,$AE,$AF,$D8,$F9,$4E,$77,$FE,$DB,$6B
	.dc.b	$FD,$4C,$38,$7F,$AB,$35,$CD,$7F,$01,$ED,$FD,$94,$F9,$9C,$7C,$77
	.dc.b	$EC,$EB,$8D,$6F,$0E,$D7,$B6,$C7,$F7,$9A,$E1,$5D,$A7,$FD,$E9,$F2
	.dc.b	$8F,$9B,$FC,$8B,$63,$FC,$AE,$D9,$FF,$6C,$35,$FF,$7C,$39,$C7,$F8
	.dc.b	$43,$67,$FB,$6F,$67,$F7,$27,$BD,$F9,$EF,$FB,$B7,$CA,$BC,$2C,$42
	.dc.b	$31,$5F,$DD,$8F,$94,$FE,$C6,$3E,$C1,$B9,$E3,$1C,$23,$2D,$C7,$F8
	.dc.b	$E3,$C5,$B3,$E0,$5F,$23,$4F,$1D,$7F,$82,$FE,$0F,$FC,$19,$F2,$8F
	.dc.b	$23,$EE,$5F,$0C,$BC,$D7,$FE,$0C,$C3,$94,$62,$3F,$05,$87,$2B,$E3
	.dc.b	$5D,$A5,$39,$79,$55,$78,$8C,$3C,$4E,$8C,$5F,$E3,$DF,$41,$83,$51
	.dc.b	$C1,$E6,$DC,$3D,$51,$B3,$33,$E6,$AF,$14,$74,$AE,$18,$7F,$57,$7A
	.dc.b	$46,$6B,$83,$CC,$C1,$E2,$7C,$CF,$C0,$3C,$17,$E2,$9E,$71,$97,$BD
	.dc.b	$F8,$A3,$81,$FE,$AC,$C3,$FD,$5A,$FC,$03,$07,$6F,$94,$FE,$75,$F2
	.dc.b	$BF,$EB,$26,$C6,$7C,$04,$E5,$B1,$C7,$E4,$D9,$71,$6C,$66,$30,$47
	.dc.b	$E5,$67,$F9,$D7,$ED,$35,$F9,$D8,$2B,$9E,$7F,$56,$31,$7C,$35,$FE
	.dc.b	$15,$3F,$67,$0E,$4E,$EF,$8C,$7F,$0F,$BE,$38,$78,$AF,$8C,$FF,$13
	.dc.b	$38,$9B,$AE,$33,$FE,$69,$AE,$37,$DE,$F8,$CF,$F9,$B6,$F8,$BB,$BE
	.dc.b	$2A,$F8,$CF,$7A,$E3,$3F,$A9,$6C,$EA,$3A,$47,$EB,$34,$E1,$1F,$A9
	.dc.b	$AF,$91,$8F,$F8,$56,$F8,$4F,$33,$5F,$1B,$C5,$F3,$35,$D1,$FC,$23
	.dc.b	$E5,$7F,$7E,$FE,$D5,$AE,$77,$E1,$1F,$56,$35,$1D,$6D,$C8,$F1,$B0
	.dc.b	$DA,$D9,$FC,$07,$E5,$7F,$B2,$BF,$A0,$0E,$1F,$D1,$B5,$FD,$D3,$5F
	.dc.b	$E9,$4E,$EB,$FA,$37,$C2,$FF,$85,$67,$F8,$8C,$23,$11,$FC,$9A,$78
	.dc.b	$6B,$FD,$4F,$E4,$1E,$47,$C5,$6D,$7F,$17,$6B,$FD,$97,$07,$EB,$4D
	.dc.b	$B8,$5F,$EC,$2F,$77,$FC,$45,$AF,$78,$D1,$F2,$72,$F9,$91,$D5,$B8
	.dc.b	$7F,$94,$6B,$C2,$FF,$94,$F2,$77,$FE,$F9,$B7,$F2,$DA,$FC,$A8,$F2
	.dc.b	$7A,$F2,$3E,$85,$A3,$11,$FC,$4A,$1E,$BF,$AC,$91,$C1,$F0,$7C,$1C
	.dc.b	$8F,$03,$5E,$5C,$9F,$0C,$8F,$81,$F8,$DE,$AC,$8F,$1B,$5F,$E4,$7F
	.dc.b	$6A,$F8,$CF,$E3,$6D,$8F,$C4,$35,$DC,$7F,$20,$1F,$0D,$49,$1F,$8F
	.dc.b	$8C,$B7,$CB,$E0,$47,$07,$E5,$7E,$8D,$97,$F2,$C6,$C6,$63,$07,$F3
	.dc.b	$0F,$98,$7C,$1B,$E6,$3F,$3D,$18,$6F,$BD,$F8,$9B,$FE,$83,$39,$7F
	.dc.b	$A1,$FC,$C3,$E0,$FF,$CC,$6D,$EB,$FD,$43,$67,$73,$FE,$68,$E6,$7C
	.dc.b	$55,$25,$FE,$15,$FF,$36,$F2,$FF,$81,$6F,$F2,$DE,$16,$09,$5F,$86
	.dc.b	$EC,$D9,$1C,$1F,$8F,$F8,$AF,$01,$1C,$1F,$BD,$F9,$67,$CB,$7F,$9F
	.dc.b	$1B,$19,$8C,$19,$F9,$69,$1C,$5B,$8F,$FA,$E7,$E1,$7F,$C5,$1D,$FF
	.dc.b	$B2,$8F,$F6,$14,$63,$C2,$FF,$0E,$5F,$FA,$EE,$3F,$D8,$44,$7F,$D6
	.dc.b	$4D,$9F,$F1,$4F,$C0,$23,$AF,$85,$E6,$9C,$C6,$3E,$5B,$39,$5C,$3F
	.dc.b	$87,$DA,$31,$7F,$BB,$6E,$5F,$C5,$9E,$BF,$3F,$2F,$8F,$D1,$DB,$FE
	.dc.b	$05,$36,$FF,$8E,$CD,$AF,$E3,$1F,$27,$E6,$F8,$AF,$F2,$3F,$22,$31
	.dc.b	$3F,$C6,$5F,$BE,$47,$F8,$0B,$71,$FF,$24,$CF,$19,$F5,$FB,$BD,$2B
	.dc.b	$E1,$FC,$D7,$C4,$FA,$2B,$5F,$01,$B8,$7F,$19,$6E,$7F,$CD,$1F,$97
	.dc.b	$F4,$32,$BE,$3F,$B1,$6F,$F9,$04,$DB,$FE,$77,$36,$BF,$C8,$FF,$27
	.dc.b	$F5,$3E,$6F,$FC,$CB,$EC,$A3,$13,$FE,$4D,$8C,$3F,$E2,$38,$FF,$65
	.dc.b	$C6,$31,$FF,$17,$B5,$FF,$23,$39,$C3,$FE,$00,$36,$93,$FE,$20,$E3
	.dc.b	$FF,$22,$D6,$AF,$F0,$FF,$37,$F8,$8F,$26,$E1,$F8,$46,$E1,$FC,$F5
	.dc.b	$B7,$FD,$61,$FC,$7F,$CD,$CA,$FC,$5F,$CD,$DB,$FE,$89,$36,$FF,$AF
	.dc.b	$CD,$AF,$F5,$67,$CB,$7D,$9B,$FE,$95,$F3,$7F,$E8,$A8,$C5,$FF,$66
	.dc.b	$8C,$71,$FF,$BB,$A7,$85,$7F,$99,$DA,$F4,$8D,$1C,$2F,$F3,$FE,$4D
	.dc.b	$9F,$EC,$9F,$14,$F9,$BF,$DA,$AF,$9B,$B1,$86,$DF,$F7,$69,$D3,$67
	.dc.b	$4E,$7E,$17,$99,$D3,$E3,$EC,$64,$98,$78,$A3,$2F,$FA,$93,$37,$C8
	.dc.b	$F9,$31,$F8,$8D,$BF,$F4,$A4,$E9,$F2,$4E,$9A,$6A,$78,$57,$56,$EC
	.dc.b	$9F,$A7,$18,$69,$CA,$78,$36,$78,$47,$E3,$AF,$C4,$FC,$CB,$F1,$36
	.dc.b	$BC,$47,$13,$AB,$93,$A9,$6B,$AA,$AE,$B1,$FE,$ED,$A1,$D9,$C3,$6B
	.dc.b	$F4,$31,$A8,$C5,$7F,$A0,$9E,$33,$E4,$8F,$03,$E4,$9E,$6E,$5B,$1E
	.dc.b	$66,$9E,$6A,$F2,$CF,$82,$BA,$D6,$E3,$11,$E2,$AF,$24,$75,$32,$59
	.dc.b	$F3,$4E,$CE,$73,$DE,$FE,$B4,$FD,$16,$BF,$1D,$B7,$E4,$8E,$77,$D6
	.dc.b	$BC,$31,$E5,$32,$57,$27,$78,$FB,$11,$DD,$AF,$76,$CF,$79,$E3,$1F
	.dc.b	$2C,$F6,$CF,$5B,$F5,$C6,$63,$C3,$3D,$DF,$A6,$77,$29,$DC,$B5,$DE
	.dc.b	$B9,$C6,$E7,$EA,$D7,$57,$C7,$1C,$CF,$E1,$17,$FC,$2C,$93,$FE,$7D
	.dc.b	$BF,$CD,$CE,$A3,$13,$A8,$FF,$70,$CF,$CD,$7C,$95,$F9,$5F,$89,$FC
	.dc.b	$7C,$62,$FC,$58,$27,$37,$C4,$F5,$38,$C3,$F0,$0D,$AF,$B1,$5C,$5F
	.dc.b	$C3,$5F,$53,$E0,$3F,$91,$9E,$11,$F3,$DB,$87,$CD,$3C,$D3,$CB,$0F
	.dc.b	$16,$1F,$C8,$AB,$73,$E8,$6D,$F6,$3F,$54,$E1,$F1,$5F,$ED,$23,$17
	.dc.b	$EC,$AE,$D7,$EC,$69,$EC,$9D,$C7,$86,$33,$7E,$96,$DF,$A5,$F7,$18
	.dc.b	$3C,$C5,$8F,$34,$78,$E3,$11,$E6,$C3,$E5,$3F,$A1,$3E,$51,$9B,$F1
	.dc.b	$CF,$F3,$F6,$C7,$F4,$87,$F6,$9F,$19,$FD,$20,$FE,$93,$EB,$5A,$FE
	.dc.b	$8D,$16,$BF,$A7,$CD,$C3,$F8,$D7,$B9,$FD,$43,$E3,$3F,$A8,$B7,$0F
	.dc.b	$E9,$CD,$BF,$EA,$ED,$AF,$E2,$1E,$EF,$F5,$8F,$5F,$FB,$43,$6F,$FA
	.dc.b	$2B,$FD,$9C,$8F,$E7,$5B,$7F,$D3,$DF,$EC,$3E,$8F,$29,$FE,$BD,$F3
	.dc.b	$7F,$D9,$0F,$EC,$BE,$85,$8C,$92,$FF,$B5,$C6,$5F,$ED,$BD,$BF,$DB
	.dc.b	$3E,$6F,$FB,$6B,$6B,$FB,$53,$67,$71,$FD,$B6,$7F,$C0,$5F,$2F,$FD
	.dc.b	$D3,$DA,$B3,$FD,$C5,$B5,$FB,$29,$FF,$02,$12,$7F,$74,$EB,$5F,$E9
	.dc.b	$1E,$D5,$9F,$E6,$DF,$21,$FE,$00,$3F,$C0,$3E,$ED,$8F,$EF,$D1,$70
	.dc.b	$FF,$03,$CD,$BF,$EC,$23,$E5,$FF,$C2,$F7,$FD,$04,$FF,$3F,$B8,$AF
	.dc.b	$EA,$11,$8E,$FF,$F0,$7B,$4F,$F0,$61,$D0,$FF,$1C,$C3,$CE,$FE,$3B
	.dc.b	$7D,$FE,$6F,$53,$FE,$3E,$9C,$E0,$9C,$CF,$1E,$AF,$ED,$6F,$4A,$FE
	.dc.b	$AA,$F6,$35,$85,$7A,$4F,$06,$C7,$F6,$C3,$94,$7B,$BD,$CE,$F5,$E2
	.dc.b	$F9,$7F,$82,$7C,$2F,$98,$B7,$C7,$06,$7F,$26,$B3,$18,$9F,$EF,$39
	.dc.b	$AF,$37,$F4,$23,$FA,$70,$F9,$1F,$9F,$DF,$F4,$1B,$7D,$EB,$EB,$46
	.dc.b	$30,$EC,$78,$AF,$95,$79,$63,$FC,$63,$8F,$DA,$F8,$9F,$97,$8A,$31
	.dc.b	$5B,$8C,$3F,$26,$B8,$B6,$33,$1F,$35,$AF,$D1,$73,$8F,$F8,$81,$B3
	.dc.b	$F0,$BE,$63,$C3,$7F,$EA,$8C,$BF,$DC,$64,$9F,$07,$AB,$EA,$C7,$ED
	.dc.b	$7B,$7D,$9F,$C5,$B5,$FF,$42,$B5,$34,$E4,$FE,$15,$AF,$2A,$FC,$23
	.dc.b	$C1,$FF,$7A,$E3,$2F,$C6,$D9,$F1,$E7,$C2,$CD,$25,$7C,$98,$65,$B2
	.dc.b	$46,$5F,$F4,$15,$48,$FE,$F0,$D3,$F3,$4B,$E4,$3C,$ED,$3E,$B1,$63
	.dc.b	$71,$EC,$F9,$87,$8F,$49,$5F,$76,$32,$FD,$FF,$A0,$77,$7C,$46,$C6
	.dc.b	$63,$07,$C8,$F9,$87,$5D,$24,$FE,$26,$32,$DE,$AE,$04,$70,$7E,$3F
	.dc.b	$E8,$3C,$7A,$7E,$E5,$AF,$EF,$5A,$FF,$0C,$6B,$FE,$C6,$F6,$DE,$4F
	.dc.b	$DC,$6D,$7E,$6D,$B3,$FC,$15,$AF,$F0,$95,$E4,$71,$24,$F1,$7F,$0F
	.dc.b	$F3,$2F,$37,$F8,$AF,$D0,$3E,$3B,$F8,$BC,$E6,$FF,$65,$39,$7F,$69
	.dc.b	$19,$C7,$F8,$B3,$5F,$E3,$4D,$7F,$94,$B5,$FF,$71,$BC,$7A,$FF,$D1
	.dc.b	$AD,$9D,$F9,$DE,$2C,$7F,$94,$FC,$D6,$70,$FE,$2F,$F4,$1F,$71,$FE
	.dc.b	$12,$23,$58,$A7,$4F,$FA,$6A,$3C,$53,$BA,$FD,$4C,$60,$8F,$80,$FE
	.dc.b	$7B,$2F,$93,$FE,$45,$8F,$F1,$5A,$F2,$5B,$E6,$DE,$FF,$0E,$F7,$C2
	.dc.b	$7F,$42,$9C,$E1,$FC,$8E,$72,$FF,$1F,$8C,$DF,$F4,$16,$BF,$D1,$5A
	.dc.b	$FF,$54,$6A,$1F,$36,$F8,$DF,$EA,$2D,$AF,$E5,$8E,$DB,$1A,$23,$A9
	.dc.b	$A7,$CE,$D1,$F2,$DF,$D7,$1B,$3F,$CD,$E7,$C5,$F0,$5E,$3F,$9A,$FC
	.dc.b	$64,$7F,$B0,$AF,$81,$2B,$84,$2B,$87,$2F,$F7,$0C,$3A,$CF,$DE,$3A
	.dc.b	$8C,$3A,$BF,$D6,$B0,$FF,$71,$C6,$23,$6E,$A7,$9B,$98,$F0,$1C,$EB
	.dc.b	$C1,$5C,$9F,$EE,$57,$E4,$9F,$56,$15,$FF,$02,$C5,$C3,$AD,$70,$9E
	.dc.b	$87,$02,$9C,$0B,$1C,$1F,$F0,$14,$F3,$BF,$14,$60,$DB,$87,$96,$1E
	.dc.b	$D8,$C4,$7C,$63,$FB,$14,$FF,$63,$70,$79,$4F,$F0,$87,$CC,$FF,$35
	.dc.b	$F9,$BF,$ED,$0D,$8F,$F0,$93,$5D,$F8,$69,$DF,$CD,$FF,$82,$A7,$6D
	.dc.b	$7A,$B5,$FF,$15,$D2,$31,$1F,$99,$8C,$65,$F3,$73,$FC,$D4,$DD,$7F
	.dc.b	$C6,$7A,$7E,$02,$F7,$3F,$31,$B1,$F8,$4A,$E3,$3F,$41,$B1,$C7,$0F
	.dc.b	$43,$70,$ED,$50,$E9,$3A,$AF,$4B,$63,$F6,$11,$D0,$FF,$23,$95,$F6
	.dc.b	$46,$D5,$F9,$A7,$78,$7B,$1B,$3F,$8D,$9D,$3D,$A7,$F4,$9A,$EA,$3F
	.dc.b	$90,$D7,$BB,$04,$F9,$9F,$51,$89,$F3,$FC,$97,$26,$9F,$E2,$FA,$E5
	.dc.b	$5F,$E6,$0F,$9B,$FD,$FC,$FF,$8C,$DA,$FF,$30,$6D,$F5,$84,$7D,$D7
	.dc.b	$86,$FD,$E7,$F9,$57,$93,$79,$FF,$C2,$35,$A3,$FC,$81,$F2,$9F,$D8
	.dc.b	$9B,$E0,$79,$D4,$65,$FF,$24,$DB,$E4,$B9,$C7,$F4,$49,$D3,$FE,$6D
	.dc.b	$6B,$FE,$7C,$6B,$FC,$A5,$AB,$CD,$7C,$DF,$85,$B1,$FE,$55,$6B,$E2
	.dc.b	$A6,$FB,$A9,$3B,$7E,$14,$EA,$30,$7F,$92,$23,$F6,$D5,$D2,$3F,$D0
	.dc.b	$8D,$7F,$D2,$7D,$EF,$8E,$FD,$E3,$63,$FC,$C0,$D7,$FD,$14,$D7,$E1
	.dc.b	$7C,$0F,$F0,$3C,$8D,$24,$7F,$96,$BE,$03,$C6,$47,$07,$6F,$8F,$F0
	.dc.b	$91,$C1,$FE,$71,$A7,$C1,$77,$24,$65,$F7,$F2,$C8,$FF,$34,$B5,$F4
	.dc.b	$F8,$3E,$E2,$3E,$E7,$F8,$E7,$CC,$BE,$32,$92,$7F,$D0,$51,$97,$FD
	.dc.b	$09,$0F,$85,$FE,$48,$D8,$CC,$60,$FE,$51,$E7,$BE,$46,$92,$7F,$D1
	.dc.b	$D1,$97,$FD,$21,$F3,$3F,$01,$B9,$FF,$31,$8C,$47,$FA,$F5,$AF,$FB
	.dc.b	$07,$E6,$5F,$08,$F8,$CF,$F4,$D4,$61,$B8,$7E,$26,$B8,$56,$DF,$BF
	.dc.b	$F2,$7F,$E8,$4E,$43,$49,$E0,$7D,$5F,$8A,$FF,$44,$F1,$9A,$48,$FF
	.dc.b	$4D,$F8,$5F,$E8,$EE,$9F,$F0,$56,$3F,$ED,$56,$BF,$96,$8F,$07,$8F
	.dc.b	$2B,$F2,$6D,$9F,$F6,$E6,$1F,$E3,$97,$F6,$9E,$BF,$88,$95,$FE,$B9
	.dc.b	$F9,$87,$CC,$7F,$A7,$9B,$19,$8C,$1E,$FF,$17,$FC,$6F,$E0,$7F,$A6
	.dc.b	$27,$FD,$95,$BF,$F8,$2B,$C2,$FE,$44,$7F,$07,$F7,$7F,$DE,$B1,$FE
	.dc.b	$0E,$CF,$FD,$E4,$D7,$F5,$F8,$75,$BF,$D6,$BC,$0E,$FE,$15,$8F,$F6
	.dc.b	$F6,$2F,$9E,$F1,$FC,$FF,$91,$B3,$A6,$E1,$FC,$25,$B9,$F9,$AD,$F3
	.dc.b	$BE,$C3,$DB,$E1,$DC,$3B,$C5,$C3,$F8,$94,$DB,$FE,$69,$C2,$F9,$2F
	.dc.b	$8B,$F3,$7F,$0B,$1F,$E4,$51,$87,$F9,$8B,$F8,$42,$3F,$BB,$6E,$1F
	.dc.b	$CB,$5F,$9F,$E2,$73,$8F,$E5,$51,$F4,$BE,$7F,$E9,$FC,$F5,$BF,$8A
	.dc.b	$DA,$FE,$50,$DC,$BE,$03,$F6,$BA,$25,$7C,$CF,$90,$B7,$FC,$BE,$6D
	.dc.b	$FC,$79,$B5,$FE,$EE,$CF,$F0,$77,$E5,$3F,$DD,$AE,$23,$FC,$D1,$5F
	.dc.b	$D6,$7E,$19,$3F,$EE,$52,$4F,$FB,$8D,$B5,$FE,$90,$9D,$CF,$23,$9F
	.dc.b	$CD,$7E,$5A,$75,$5F,$E8,$A8,$7C,$ED,$F9,$DC,$36,$F9,$5B,$15,$F0
	.dc.b	$9C,$C6,$2F,$8C,$62,$51,$DF,$DC,$93,$9F,$22,$E1,$FD,$A2,$FF,$D2
	.dc.b	$2D,$AF,$34,$EB,$0F,$EF,$9E,$03,$E4,$BD,$5F,$1F,$F2,$E3,$19,$78
	.dc.b	$63,$13,$FC,$1C,$91,$FE,$95,$C7,$FA,$DC,$61,$FF,$5C,$D7,$27,$C9
	.dc.b	$F0,$BC,$AB,$D9,$E8,$7C,$9F,$4B,$E5,$36,$32,$7C,$18,$CB,$4F,$89
	.dc.b	$E3,$FD,$13,$E9,$7A,$36,$32,$4C,$3E,$24,$65,$FF,$66,$FB,$8F,$95
	.dc.b	$FB,$5F,$3B,$F8,$B8,$C5,$FD,$18,$C4,$EA,$3E,$DD,$7F,$A7,$63,$0F
	.dc.b	$46,$EF,$FE,$29,$BE,$73,$F5,$6F,$9D,$F5,$39,$CE,$AB,$F8,$DF,$12
	.dc.b	$B7,$7C,$6F,$78,$76,$7F,$D2,$ED,$8F,$F5,$BB,$5F,$34,$77,$7C,$CD
	.dc.b	$BF,$34,$EF,$0F,$14,$FE,$82,$FC,$35,$E4,$6C,$64,$F1,$9C,$CF,$19
	.dc.b	$6F,$C6,$D7,$C6,$7A,$CF,$19,$67,$C7,$1B,$AF,$1C,$FE,$A7,$0F,$89
	.dc.b	$3E,$08,$E0,$F8,$1A,$78,$0A,$78,$2B,$D5,$1C,$5B,$1E,$26,$9E,$27
	.dc.b	$8D,$78,$A3,$F7,$37,$F0,$A7,$C8,$D8,$CD,$75,$3E,$34,$F5,$6C,$75
	.dc.b	$3B,$BE,$96,$C7,$A5,$AF,$A6,$38,$9B,$2B,$B6,$BB,$3C,$AF,$8D,$A7
	.dc.b	$8C,$B1,$E3,$8E,$35,$EF,$BF,$9F,$5F,$58,$FF,$6D,$E1,$FE,$85,$7E
	.dc.b	$81,$8A,$FF,$6A,$C6,$63,$E9,$19,$3F,$DE,$2E,$5F,$31,$E7,$BF,$E4
	.dc.b	$BE,$27,$CD,$8C,$E1,$E3,$8C,$3F,$84,$8E,$18,$7F,$80,$F0,$F8,$91
	.dc.b	$F0,$C9,$7D,$23,$E1,$47,$4C,$3F,$9D,$61,$D1,$FB,$57,$CE,$79,$1B
	.dc.b	$AF,$E6,$95,$B7,$E6,$5F,$59,$F9,$86,$CF,$E6,$93,$B7,$E8,$46,$CF
	.dc.b	$A0,$F6,$C3,$DD,$5E,$67,$DC,$62,$3A,$B8,$C3,$D3,$3B,$7D,$37,$DE
	.dc.b	$FA,$9D,$E7,$E2,$DF,$96,$3F,$87,$CE,$E7,$E4,$C6,$E7,$C9,$5D,$27
	.dc.b	$B1,$F0,$30,$F8,$D3,$CE,$38,$C6,$30,$F3,$46,$F0,$ED,$5F,$52,$B5
	.dc.b	$78,$C3,$C1,$3C,$CD,$57,$D4,$F9,$DF,$EE,$1E,$A7,$F7,$16,$E3,$FD
	.dc.b	$9D,$B5,$FD,$8D,$B7,$FC,$B3,$E7,$7F,$BC,$1F,$DE,$7E,$76,$DF,$F6
	.dc.b	$F8,$B7,$FD,$AE,$6D,$7F,$B8,$FE,$6F,$FA,$E3,$FC,$DB,$E7,$7F,$C2
	.dc.b	$51,$8C,$3F,$C0,$F1,$8E,$6F,$CF,$4F,$F6,$3C,$3F,$B4,$07,$0F,$F6
	.dc.b	$B6,$A3,$51,$CE,$BF,$C0,$9E,$A5,$9F,$F7,$3F,$CC,$3E,$77,$FC,$71
	.dc.b	$92,$3F,$C8,$D7,$FE,$0E,$74,$DF,$17,$71,$8B,$E7,$18,$BC,$BA,$F8
	.dc.b	$A9,$7C,$FE,$52,$DF,$8A,$6E,$3F,$E7,$4C,$55,$8B,$F7,$7C,$AF,$82
	.dc.b	$7E,$4B,$63,$FB,$6E,$A8,$F9,$24,$8F,$93,$8F,$C5,$9D,$46,$23,$E6
	.dc.b	$4F,$48,$F8,$1E,$8D,$9F,$8F,$F0,$3F,$00,$F8,$3E,$F5,$AF,$8D,$16
	.dc.b	$7E,$6C,$DC,$FD,$9F,$01,$F2,$3E,$43,$E4,$B7,$0F,$9C,$DB,$FA,$AD
	.dc.b	$AF,$C5,$7A,$1F,$43,$E0,$3F,$12,$DB,$F8,$2F,$E2,$08,$FB,$75,$46
	.dc.b	$30,$FF,$26,$4F,$E2,$63,$ED,$FB,$9D,$23,$F0,$1F,$3B,$F7,$8F,$BF
	.dc.b	$EE,$58,$C9,$2F,$F1,$71,$97,$F1,$BD,$DF,$86,$F9,$DF,$C3,$B6,$BF
	.dc.b	$18,$D9,$DC,$7E,$3E,$7F,$4B,$E8,$7E,$3B,$DF,$B3,$F9,$36,$D7,$DC
	.dc.b	$D1,$5F,$A8,$25,$7C,$17,$7B,$7E,$37,$E0,$2C,$FF,$94,$FD,$77,$2F
	.dc.b	$FB,$88,$91,$F8,$58,$CB,$F8,$7F,$31,$F1,$DF,$EE,$32,$46,$5F,$EA
	.dc.b	$DE,$A3,$E6,$29,$27,$F2,$11,$97,$B7,$47,$FB,$C8,$8E,$0F,$07,$57
	.dc.b	$FB,$C4,$8E,$0E,$7F,$11,$E0,$23,$83,$F3,$DF,$3E,$EF,$FE,$D4,$D8
	.dc.b	$CC,$60,$FE,$23,$C5,$23,$F1,$4D,$3F,$56,$BE,$13,$FB,$A3,$4F,$D6
	.dc.b	$16,$37,$1F,$DE,$7E,$87,$11,$FC,$71,$AF,$FA,$15,$AF,$F1,$B6,$AF
	.dc.b	$99,$74,$FF,$80,$DB,$7F,$BB,$6C,$FF,$00,$6B,$FC,$AB,$E6,$5C,$9F
	.dc.b	$E0,$B6,$D7,$F0,$36,$CF,$F0,$86,$BF,$CC,$B6,$FF,$5E,$37,$E1,$BF
	.dc.b	$C2,$91,$88,$FF,$4F,$35,$FE,$69,$F3,$1F,$C7,$2F,$F8,$83,$63,$F8
	.dc.b	$AB,$5F,$E7,$ED,$7F,$D5,$1E,$BB,$C4,$FF,$11,$36,$BF,$8D,$36,$7F
	.dc.b	$8F,$B5,$FE,$93,$F3,$0F,$82,$FF,$15,$36,$BF,$91,$36,$7F,$93,$B5
	.dc.b	$FE,$A3,$F3,$1F,$87,$7F,$84,$9F,$D3,$E3,$87,$62,$BF,$D7,$D5,$B8
	.dc.b	$C5,$6E,$FF,$DE,$53,$CE,$3A,$F1,$FF,$8F,$63,$11,$CA,$3F,$C8,$1F
	.dc.b	$2F,$CF,$0F,$F7,$BF,$CC,$7D,$EC,$7A,$DA,$B4,$FF,$BA,$67,$55,$FE
	.dc.b	$D1,$9C,$D2,$73,$5F,$DA,$67,$31,$88,$E9,$7E,$22,$5F,$F4,$9A,$F3
	.dc.b	$CF,$F7,$1F,$98,$7B,$5F,$DB,$3B,$7F,$DD,$B9,$79,$A7,$38,$7F,$BE
	.dc.b	$A7,$2F,$FB,$96,$33,$7F,$3A,$BF,$87,$FC,$07,$CA,$79,$61,$D6,$31
	.dc.b	$7F,$EE,$DF,$98,$7C,$2E,$30,$FF,$02,$46,$27,$FD,$A6,$FF,$81,$DB
	.dc.b	$3B,$94,$ED,$FB,$E6,$23,$51,$C6,$32,$EF,$0F,$A6,$7F,$8F,$8F,$F1
	.dc.b	$FC,$7F,$85,$7D,$CF,$F0,$B7,$CB,$FE,$4E,$3F,$C3,$13,$BF,$73,$C4
	.dc.b	$F0,$85,$70,$E2,$FF,$0F,$42,$75,$39,$8E,$75,$F9,$0B,$ED,$1C,$27
	.dc.b	$4F,$19,$D5,$FF,$81,$A7,$36,$9C,$D7,$F8,$CA,$73,$18,$8E,$82,$FB
	.dc.b	$D7,$1C,$BB,$CF,$C1,$35,$87,$AA,$75,$18,$3B,$14,$F3,$4F,$27,$E1
	.dc.b	$99,$8E,$35,$EE,$8E,$CF,$8A,$FC,$D1,$8B,$F0,$46,$23,$99,$FE,$23
	.dc.b	$C3,$FC,$48,$F9,$CC,$1F,$E2,$AF,$96,$FE,$81,$F3,$1F,$E2,$96,$C6
	.dc.b	$7C,$04,$E5,$B1,$C7,$E6,$3E,$C3,$73,$E5,$18,$8F,$F2,$FD,$A3,$11
	.dc.b	$F8,$08,$C4,$FD,$AC,$BF,$3F,$37,$6F,$C2,$E9,$F8,$0B,$DC,$FC,$B6
	.dc.b	$C7,$E1,$2B,$8C,$FC,$F6,$C7,$1C,$3C,$ED,$C3,$B5,$43,$A4,$EA,$BC
	.dc.b	$CD,$8F,$F3,$A4,$74,$3F,$32,$57,$FC,$F1,$7E,$48,$E3,$5D,$6F,$CD
	.dc.b	$87,$B2,$73,$7E,$78,$FC,$56,$BD,$1B,$1E,$06,$AC,$3B,$39,$3F,$63
	.dc.b	$3E,$4F,$96,$DD,$9B,$7C,$BE,$63,$F4,$97,$C0,$FF,$48,$38,$7F,$83
	.dc.b	$B7,$2F,$E6,$AF,$DB,$8E,$01,$D6,$1F,$C0,$BE,$61,$E3,$7A,$AD,$18
	.dc.b	$BE,$37,$C1,$FF,$50,$73,$7F,$8D,$1F,$25,$3F,$5C,$FB,$AF,$F2,$96
	.dc.b	$E3,$FC,$2A,$B8,$57,$F1,$E6,$D6,$F2,$F1,$5F,$0A,$EF,$5F,$33,$F5
	.dc.b	$B7,$77,$EA,$1B,$3F,$E8,$D6,$BF,$E9,$26,$BF,$EB,$C7,$38,$79,$75
	.dc.b	$EE,$75,$A9,$5E,$CC,$3E,$A3,$67,$2F,$F3,$6B,$FE,$A2,$F8,$7A,$3C
	.dc.b	$84,$7C,$9F,$0F,$E7,$23,$83,$D1,$C9,$EB,$23,$83,$D9,$C9,$EF,$23
	.dc.b	$83,$E0,$7C,$3F,$C3,$23,$83,$E2,$74,$7C,$72,$38,$3E,$47,$C3,$7C
	.dc.b	$B2,$38,$3E,$67,$C3,$7C,$F2,$38,$3E,$87,$C2,$FD,$32,$38,$3E,$A7
	.dc.b	$C3,$7D,$72,$38,$3E,$C7,$C3,$7D,$B2,$38,$3E,$E7,$C2,$FD,$F2,$38
	.dc.b	$3F,$D9,$9F,$30,$F6,$93,$86,$CF,$FB,$52,$7F,$CA,$7C,$59,$AF,$F6
	.dc.b	$A7,$D1,$3E,$83,$FC,$3B,$7F,$D8,$23,$1D,$5A,$9C,$CF,$FC,$01,$9C
	.dc.b	$BF,$EE,$0F,$A3,$7C,$6E,$A1,$18,$BF,$F7,$13,$63,$84,$EF,$17,$CC
	.dc.b	$52,$5F,$FB,$A6,$32,$FF,$BA,$FF,$01,$FF,$30,$59,$A4,$F4,$BC,$5F
	.dc.b	$64,$F8,$CA,$4B,$FF,$7A,$46,$5E,$9F,$25,$FE,$1B,$F0,$3F,$79,$8F
	.dc.b	$F8,$5F,$14,$E7,$BF,$FC,$29,$5F,$E4,$49,$CD,$F2,$25,$FF,$82,$81
	.dc.b	$E7,$BF,$8D,$3C,$23,$FC,$42,$DA,$FF,$19,$7B,$CC,$FC,$57,$E5,$87
	.dc.b	$CA,$53,$1F,$29,$72,$FE,$2B,$3A,$6C,$E9,$A6,$BC,$FE,$97,$D5,$B1
	.dc.b	$93,$9C,$65,$A7,$F2,$EC,$BF,$C9,$79,$7A,$3E,$53,$E4,$78,$7E,$76
	.dc.b	$D7,$D9,$8F,$82,$47,$07,$37,$DD,$E2,$F0,$DB,$FC,$AD,$A7,$E2,$70
	.dc.b	$F8,$BC,$9F,$0F,$E5,$3E,$23,$31,$F8,$2E,$5A,$7C,$4F,$17,$E5,$65
	.dc.b	$F2,$DB,$19,$3E,$4C,$65,$A7,$BF,$C1,$FA,$5E,$2F,$E1,$5B,$19,$3F
	.dc.b	$2C,$FD,$E2,$38,$35,$9F,$B3,$C5,$E5,$1F,$63,$C4,$FB,$DE,$2F,$DF
	.dc.b	$6C,$64,$FB,$B1,$96,$9F,$80,$F1,$7F,$0C,$7E,$1F,$0B,$19,$26,$1F
	.dc.b	$32,$32,$FF,$9C,$BD,$0F,$D2,$3F,$43,$C6,$FF,$3C,$46,$38,$3F,$2B
	.dc.b	$18,$EC,$FE,$01,$3C,$2B,$FC,$B5,$18,$26,$1F,$DD,$27,$93,$FE,$74
	.dc.b	$E2,$78,$DC,$E3,$F3,$3E,$1F,$E7,$FC,$6F,$D0,$36,$32,$7E,$7A,$32
	.dc.b	$D3,$FB,$AF,$EB,$8F,$93,$FD,$E3,$63,$39,$62,$64,$70,$77,$18,$CF
	.dc.b	$E2,$7B,$9C,$BE,$4F,$E1,$B6,$FD,$8D,$9D,$C7,$8A,$76,$D3,$65,$8D
	.dc.b	$9F,$EB,$3C,$3F,$DF,$1B,$FF,$B4,$36,$61,$D6,$30,$DB,$FF,$48,$36
	.dc.b	$FF,$D4,$4D,$7F,$D3,$AD,$AE,$73,$E1,$AE,$6D,$39,$96,$B9,$B5,$E6
	.dc.b	$D9,$E7,$1E,$9C,$3D,$ED,$9F,$7B,$4F,$7D,$7F,$B3,$F0,$F7,$CF,$82
	.dc.b	$75,$3E,$46,$C6,$63,$85,$76,$6C,$76,$6C,$F6,$8F,$F6,$0D,$76,$8F
	.dc.b	$59,$D0,$B3,$C0,$F0,$16,$7C,$07,$2A,$F0,$34,$F0,$16,$7C,$07,$FB
	.dc.b	$26,$BC,$0F,$C8,$3C,$05,$AE,$CF,$8A,$79,$47,$89,$A7,$8A,$3A,$57
	.dc.b	$89,$B1,$E2,$69,$E2,$95,$78,$AF,$EB,$C7,$82,$7C,$55,$D9,$F1,$1C
	.dc.b	$A3,$C4,$D8,$F1,$47,$4B,$F1,$36,$3E,$FC,$FF,$B6,$F0,$FE,$0A,$48
	.dc.b	$FF,$61,$36,$3F,$DC,$8D,$78,$47,$AE,$7D,$3E,$C7,$BE,$7C,$E7,$9A
	.dc.b	$33,$3F,$06,$32,$6A,$32,$FD,$43,$38,$7C,$29,$FC,$2D,$75,$8C,$4F
	.dc.b	$5C,$53,$9C,$3D,$F1,$8B,$F0,$5F,$0B,$F0,$4E,$5F,$04,$64,$EB,$87
	.dc.b	$90,$F9,$B1,$E4,$BF,$E2,$F8,$77,$C3,$E1,$61,$CE,$F7,$87,$3B,$DD
	.dc.b	$78,$9F,$2D,$F1,$C3,$F0,$58,$71,$AF,$C5,$47,$77,$F9,$3C,$EF,$0F
	.dc.b	$BE,$79,$63,$78,$77,$76,$77,$3C,$B3,$C5,$FC,$2D,$F7,$BF,$C5,$9D
	.dc.b	$CF,$E3,$D3,$B7,$E7,$35,$F9,$2D,$AF,$8F,$3C,$E7,$95,$F9,$A3,$13
	.dc.b	$E5,$9F,$7D,$7D,$6F,$88,$FE,$83,$F2,$7F,$D0,$9B,$8F,$F4,$66,$D7
	.dc.b	$F3,$16,$E3,$FC,$57,$DE,$FE,$92,$7F,$4A,$CE,$DF,$F4,$E8,$B7,$FC
	.dc.b	$E6,$6D,$2B,$FC,$83,$2F,$D7,$3F,$CD,$33,$FC,$87,$E5,$35,$28,$AE
	.dc.b	$FC,$EF,$EB,$33,$A9,$CD,$7F,$9C,$AB,$FC,$9C,$D7,$BB,$5F,$F2,$8B
	.dc.b	$5E,$97,$FE,$E6,$8F,$1B,$FC,$9B,$E5,$3B,$3C,$B0,$FE,$D3,$F2,$5F
	.dc.b	$DA,$BE,$52,$CF,$F6,$E6,$E1,$FD,$65,$B8,$FF,$3D,$7F,$B8,$FC,$9C
	.dc.b	$AF,$EE,$BF,$13,$6F,$FB,$04,$DB,$FE,$89,$36,$BE,$4F,$C5,$FF,$82
	.dc.b	$2F,$F9,$67,$C9,$79,$9F,$F3,$6C,$66,$FF,$8F,$CE,$65,$3F,$E1,$57
	.dc.b	$9F,$3F,$FA,$5E,$31,$1C,$A3,$07,$18,$E9,$CC,$F9,$3C,$4F,$F8,$33
	.dc.b	$E2,$FF,$C1,$BE,$CD,$CF,$FC,$22,$DC,$7F,$BE,$37,$3F,$EC,$0F,$F8
	.dc.b	$53,$E9,$E5,$7F,$86,$3E,$8E,$DF,$F8,$0E,$6D,$FF,$67,$9B,$5F,$7B
	.dc.b	$E0,$BF,$C3,$AF,$BF,$95,$18,$8F,$BC,$FF,$AD,$63,$31,$FE,$95,$C3
	.dc.b	$F0,$B3,$92,$47,$03,$FD,$33,$7F,$91,$F9,$D7,$CC,$77,$AF,$F5,$44
	.dc.b	$7E,$2A,$FF,$34,$BE,$9F,$3C,$FA,$6E,$3D,$0E,$B3,$B7,$15,$CA,$B8
	.dc.b	$46,$31,$53,$C5,$F0,$9E,$2F,$A7,$B5,$B9,$D3,$67,$4D,$3B,$E0,$EE
	.dc.b	$FF,$58,$47,$1F,$82,$FC,$54,$62,$7F,$6D,$3E,$89,$CF,$B5,$FE,$E1
	.dc.b	$AF,$F6,$A9,$1C,$1A,$9F,$27,$CA,$5B,$E6,$7F,$B1,$A9,$ED,$F3,$25
	.dc.b	$7D,$79,$F3,$47,$7F,$6E,$4E,$6B,$FD,$81,$5F,$75,$AF,$FB,$4D,$FF
	.dc.b	$68,$37,$DA,$F4,$12,$7B,$78,$EF,$0F,$FC,$FA,$48,$CB,$EF,$F6,$E4
	.dc.b	$78,$5B,$7F,$8D,$EE,$79,$3F,$AF,$25,$7A,$E3,$2F,$B7,$E5,$5F,$11
	.dc.b	$FC,$8A,$FF,$94,$C6,$27,$53,$99,$FF,$76,$CE,$5F,$CE,$2F,$A2,$D4
	.dc.b	$23,$15,$DE,$73,$18,$9E,$3F,$45,$82,$57,$95,$E7,$F2,$AF,$16,$92
	.dc.b	$7E,$B4,$65,$BE,$27,$02,$38,$37,$E1,$74,$23,$83,$F0,$1E,$F3,$CE
	.dc.b	$F5,$B6,$33,$18,$33,$C9,$23,$8B,$73,$FE,$01,$5F,$E7,$87,$FA,$A8
	.dc.b	$70,$FF,$58,$6A,$35,$E6,$FC,$CF,$93,$FE,$53,$06,$93,$C8,$D1,$FC
	.dc.b	$CD,$FE,$75,$1B,$E3,$C4,$E5,$E8,$4C,$7F,$41,$5F,$86,$AF,$E3,$1F
	.dc.b	$25,$E0,$25,$FD,$E6,$D7,$EF,$1F,$E8,$9F,$1B,$FC,$56,$91,$E2,$25
	.dc.b	$7F,$4B,$F0,$95,$B9,$F3,$FC,$97,$C5,$C3,$F6,$4D,$C3,$F9,$53,$5F
	.dc.b	$96,$D7,$FC,$02,$EE,$FF,$BE,$B6,$7F,$96,$F8,$0E,$CF,$EB,$13,$9B
	.dc.b	$FA,$1F,$24,$F8,$EF,$EB,$8D,$C3,$F8,$CB,$F5,$5B,$1A,$23,$A9,$A7
	.dc.b	$F1,$03,$FC,$21,$EF,$7F,$67,$C8,$FF,$69,$8F,$C0,$77,$FE,$07,$E4
	.dc.b	$7F,$A6,$C7,$F4,$49,$E0,$4B,$FE,$57,$1C,$04,$F0,$7B,$C6,$95,$75
	.dc.b	$F2,$BF,$C4,$F8,$78,$AB,$15,$F9,$6A,$AF,$23,$AD,$7F,$5B,$8C,$D7
	.dc.b	$19,$C9,$C4,$A7,$1B,$F2,$3E,$77,$CD,$1C,$5C,$C7,$98,$E3,$19,$AF
	.dc.b	$04,$63,$0F,$F1,$DD,$70,$BF,$1C,$62,$B7,$3D,$A3,$15,$E5,$8C,$61
	.dc.b	$EE,$8F,$94,$7F,$34,$3F,$A1,$1E,$33,$07,$EE,$BE,$45,$F0,$4F,$92
	.dc.b	$FE,$91,$18,$6C,$66,$7F,$C0,$53,$96,$FC,$1F,$F9,$4D,$BA,$72,$8C
	.dc.b	$47,$76,$BF,$E5,$E9,$47,$1C,$7F,$01,$9F,$E6,$6B,$EC,$B5,$FA,$75
	.dc.b	$C2,$FE,$EB,$63,$EB,$CE,$EB,$9E,$BF,$E6,$C7,$F1,$95,$B6,$E1,$E1
	.dc.b	$AF,$F3,$89,$C4,$B5,$C5,$F4,$46,$EF,$F8,$F5,$F3,$9D,$C7,$A2,$B8
	.dc.b	$3A,$AE,$0A,$B8,$3F,$C9,$EB,$C0,$DA,$E2,$EE,$B8,$CF,$C3,$C3,$FD
	.dc.b	$02,$DC,$FC,$35,$FB,$7C,$3A,$BF,$E3,$EC,$3F,$4F,$3E,$EF,$8C,$B3
	.dc.b	$E1,$6B,$B3,$FC,$3F,$F2,$5F,$BE,$6E,$DA,$8C,$47,$F0,$57,$ED,$BA
	.dc.b	$3F,$C6,$5C,$BF,$DD,$9B,$E2,$71,$AE,$FC,$A3,$F2,$E5,$9F,$F0,$D3
	.dc.b	$5F,$F2,$93,$5F,$EA,$6E,$75,$FF,$50,$7C,$67,$C3,$9F,$F5,$33,$5E
	.dc.b	$D3,$EE,$F9,$2F,$EB,$51,$FE,$3C,$6B,$FC,$7D,$B5,$D3,$CC,$E1,$DD
	.dc.b	$FA,$C9,$DD,$7F,$64,$6B,$FE,$BA,$6B,$FD,$8D,$E7,$B7,$F9,$9B,$16
	.dc.b	$DF,$4B,$14,$62,$FC,$98,$7C,$6D,$3E,$29,$23,$E2,$BE,$3F,$88,$91
	.dc.b	$FE,$41,$C9,$18,$8F,$F3,$D4,$3E,$27,$D4,$48,$E0,$FD,$2F,$26,$47
	.dc.b	$D2,$6B,$CB,$97,$E1,$11,$F0,$3F,$17,$E2,$64,$78,$DA,$FD,$AF,$27
	.dc.b	$EF,$91,$FB,$EF,$CD,$F8,$89,$1E,$B6,$BE,$AF,$89,$FA,$44,$7E,$03
	.dc.b	$F8,$AF,$8F,$91,$F8,$A6,$BF,$0B,$E2,$7E,$D1,$1F,$90,$FD,$DF,$88
	.dc.b	$91,$F2,$DA,$FF,$A8,$FE,$91,$F0,$9F,$CB,$9B,$1F,$E7,$16,$BB,$8F
	.dc.b	$E6,$7F,$26,$F3,$9F,$17,$FE,$77,$8C,$37,$E0,$FF,$05,$7F,$E3,$09
	.dc.b	$CB,$FE,$B3,$F9,$3F,$F7,$B1,$3D,$FF,$F2,$CB,$FE,$F9,$FC,$0D,$F0
	.dc.b	$BF,$E5,$C6,$F7,$7F,$A1,$5B,$7F,$EB,$D6,$BF,$E3,$EE,$E7,$0F,$F9
	.dc.b	$91,$B5,$FE,$8D,$6C,$FF,$B1,$5A,$FF,$91,$BE,$4B,$FD,$44,$DF,$7F
	.dc.b	$FC,$DB,$18,$8F,$F2,$30,$70,$FF,$92,$7F,$5A,$F8,$B7,$FE,$BB,$F0
	.dc.b	$3F,$CF,$1E,$0A,$38,$CE,$AB,$FD,$4B,$3F,$AA,$F0,$3F,$DF,$93,$99
	.dc.b	$FF,$2C,$B5,$E3,$C5,$FE,$80,$8C,$47,$86,$38,$9F,$E3,$FC,$3F,$D0
	.dc.b	$17,$E1,$9D,$61,$E5,$D3,$46,$FA,$AC,$64,$93,$E6,$B4,$EA,$31,$F3
	.dc.b	$FD,$FE,$97,$C0,$DB,$F8,$4D,$9D,$C7,$C5,$9F,$26,$3E,$59,$FE,$59
	.dc.b	$E4,$FA,$7E,$97,$37,$F4,$3C,$7F,$F2,$25,$78,$E7,$2F,$F9,$E6,$32
	.dc.b	$FF,$2E,$AF,$F5,$2C,$62,$39,$46,$0E,$2F,$82,$BE,$C7,$D1,$E2,$BB
	.dc.b	$E1,$FC,$DB,$E9,$3E,$61,$F3,$7C,$EB,$7F,$7E,$2E,$1F,$1E,$6D,$7F
	.dc.b	$3E,$FA,$2F,$A9,$F4,$BF,$55,$B8,$7E,$1D,$B7,$F3,$DB,$5F,$17,$97
	.dc.b	$FC,$D5,$5F,$8D,$AF,$4C,$61,$FA,$C6,$3D,$7F,$CC,$92,$7F,$DD,$D3
	.dc.b	$FE,$E6,$9F,$CD,$BC,$6B,$B5,$7F,$A1,$FE,$21,$D1,$CF,$0F,$F5,$35
	.dc.b	$7D,$89,$FF,$79,$B7,$E1,$F8,$7C,$47,$16,$57,$03,$8D,$70,$6C,$EB
	.dc.b	$2F,$D1,$CB,$E1,$FC,$59,$78,$DB,$7C,$E9,$8A,$F8,$4E,$63,$17,$C6
	.dc.b	$31,$3E,$F7,$E0,$F9,$FE,$FF,$9A,$F9,$6D,$8E,$D8,$7E,$FE,$3F,$81
	.dc.b	$92,$3E,$5B,$63,$FD,$21,$AF,$F5,$A9,$EF,$E8,$F4,$7C,$3F,$33,$F3
	.dc.b	$4F,$9D,$F3,$56,$32,$4B,$F8,$71,$97,$E2,$FA,$5F,$57,$1F,$AC,$DA
	.dc.b	$F6,$B6,$77,$1F,$02,$7F,$0F,$E7,$FD,$DF,$6A,$CF,$C4,$6D,$7D,$26
	.dc.b	$EF,$F7,$DB,$3F,$69,$B3,$F4,$B6,$FF,$3A,$CF,$C8,$D5,$87,$78,$C3
	.dc.b	$6B,$E3,$B7,$0F,$CF,$36,$7F,$2A,$D9,$FF,$7B,$57,$49,$F4,$1D,$0B
	.dc.b	$1D,$1B,$5D,$1A,$74,$2C,$74,$AE,$CF,$F3,$D6,$D7,$78,$F8,$93,$DD
	.dc.b	$A7,$72,$D7,$76,$C7,$76,$9D,$DF,$0B,$DD,$B3,$DE,$7E,$84,$77,$6C
	.dc.b	$F7,$69,$DC,$B1,$DD,$B1,$DC,$F0,$9D,$CB,$3D,$E7,$EE,$1D,$CB,$1D
	.dc.b	$DB,$3D,$DA,$77,$2C,$77,$8F,$0B,$DD,$B3,$DE,$3F,$11,$87,$E2,$A7
	.dc.b	$F7,$4D,$C3,$F1,$73,$E1,$37,$7E,$56,$99,$AE,$31,$96,$C6,$67,$9C
	.dc.b	$7A,$63,$8B,$F5,$DF,$F0,$16,$1F,$E0,$02,$57,$F2,$86,$C7,$F8,$08
	.dc.b	$F9,$87,$08,$D1,$E1,$AF,$F6,$3B,$E7,$8F,$84,$FC,$77,$F5,$1E,$97
	.dc.b	$EA,$9C,$9F,$26,$3F,$93,$C6,$71,$F9,$13,$AC,$3F,$90,$4E,$70,$F9
	.dc.b	$0D,$BF,$13,$6B,$E1,$36,$BE,$11,$CA,$7C,$D7,$D8,$FA,$D1,$9A,$FA
	.dc.b	$47,$98,$F1,$61,$EE,$73,$3C,$70,$FE,$95,$87,$7C,$3E,$06,$1C,$CF
	.dc.b	$A5,$7C,$EF,$BE,$1C,$E7,$31,$CE,$BF,$9C,$DF,$3B,$F8,$C7,$CA,$8F
	.dc.b	$4E,$1F,$09,$A7,$C2,$2C,$7C,$26,$DF,$C2,$C3,$C1,$1F,$56,$7C,$95
	.dc.b	$EA,$8D,$9E,$A9,$DC,$F8,$8D,$C7,$98,$DB,$E6,$AF,$81,$1D,$70,$F6
	.dc.b	$4F,$F6,$0C,$3F,$41,$EB,$7F,$71,$F4,$FF,$B9,$36,$FF,$AC,$B6,$BF
	.dc.b	$B3,$B7,$0F,$EE,$9E,$9F,$F7,$93,$FB,$D6,$76,$FF,$B0,$C5,$BF,$ED
	.dc.b	$F3,$6B,$FC,$65,$EE,$3E,$03,$FC,$05,$F3,$5F,$E2,$68,$C6,$3F,$DA
	.dc.b	$A3,$13,$FE,$39,$C3,$FC,$CB,$5A,$7F,$4C,$1C,$3F,$A9,$6B,$C2,$39
	.dc.b	$D7,$F7,$6F,$5A,$CF,$F8,$13,$C9,$FF,$0A,$7A,$DF,$E1,$56,$D7,$F7
	.dc.b	$C6,$CE,$E3,$FB,$FC,$FF,$91,$B2,$7C,$9B,$E6,$B0,$DA,$FF,$04,$3E
	.dc.b	$99,$FF,$2A,$B9,$BF,$F3,$BB,$63,$35,$CB,$2E,$6F,$C7,$F8,$6E,$BF
	.dc.b	$0D,$71,$EA,$DC,$7E,$2E,$AA,$F9,$50,$3E,$43,$86,$D7,$CC,$69,$FE
	.dc.b	$22,$7E,$8C,$7B,$3E,$6B,$B6,$5E,$AF,$88,$F7,$FC,$4F,$C0,$6D,$FC
	.dc.b	$06,$E1,$EE,$6D,$7C,$2F,$9A,$F8,$E7,$C8,$C6,$DF,$C8,$8B,$7F,$1A
	.dc.b	$6D,$7F,$9F,$7E,$55,$ED,$7D,$0F,$91,$FF,$38,$C6,$31,$FA,$31,$89
	.dc.b	$F8,$13,$9C,$3F,$16,$D7,$F9,$93,$5F,$F5,$AB,$BB,$F0,$4F,$DE,$F9
	.dc.b	$AE,$98,$7D,$EF,$9A,$FC,$0F,$C0,$FE,$09,$B1,$93,$F0,$71,$96,$9F
	.dc.b	$86,$F9,$AF,$C4,$9F,$8A,$F9,$AB,$19,$26,$7F,$88,$8C,$BF,$5C,$FB
	.dc.b	$DF,$2B,$F8,$F9,$FF,$36,$F3,$BE,$53,$FD,$2B,$39,$AF,$F2,$FC,$E6
	.dc.b	$BF,$CA,$B3,$FA,$A9,$E7,$D1,$F8,$78,$C4,$1A,$73,$71,$1F,$A2,$F9
	.dc.b	$3E,$EF,$FA,$93,$E2,$5E,$87,$E4,$5B,$87,$E4,$5B,$7F,$BE,$6B,$FD
	.dc.b	$97,$99,$E8,$7E,$65,$B5,$F9,$96,$CF,$F0,$66,$BC,$6F,$81,$27,$83
	.dc.b	$E3,$F3,$E4,$78,$DB,$3F,$AC,$E6,$FD,$47,$59,$A4,$AF,$17,$A1,$E5
	.dc.b	$23,$83,$F5,$DF,$44,$F8,$BF,$D9,$36,$33,$18,$3F,$6D,$F2,$CF,$83
	.dc.b	$A4,$9F,$DC,$46,$5B,$F1,$9C,$08,$E0,$FD,$F7,$D0,$BA,$BF,$81,$36
	.dc.b	$33,$18,$3F,$83,$7C,$BF,$E1,$76,$FE,$62,$D7,$FC,$30,$D7,$F8,$70
	.dc.b	$70,$FF,$11,$5C,$BD,$09,$5F,$C4,$63,$2F,$F1,$3F,$A1,$FE,$4D,$5F
	.dc.b	$E2,$56,$BD,$9A,$FF,$1D,$6A,$F5,$DE,$27,$F1,$F6,$D7,$F1,$F6,$CF
	.dc.b	$F4,$C6,$BF,$E3,$35,$CB,$C4,$93,$C5,$FE,$55,$F3,$1F,$A1,$FA,$17
	.dc.b	$C0,$AB,$FD,$7F,$39,$8C,$61,$FB,$71,$1F,$AC,$27,$5F,$42,$57,$27
	.dc.b	$8F,$89,$CA,$B8,$F9,$38,$AF,$A3,$7F,$B3,$F9,$9F,$DF,$5B,$E8,$7E
	.dc.b	$F6,$1E,$19,$CC,$62,$7A,$47,$EF,$89,$8F,$F4,$76,$E1,$FE,$DA,$7A
	.dc.b	$FC,$A7,$F2,$58,$FC,$0C,$7F,$03,$F9,$AF,$A5,$1F,$EE,$9F,$85,$FF
	.dc.b	$7B,$65,$FD,$4E,$72,$FF,$19,$C5,$3A,$84,$E9,$FF,$33,$35,$FC,$7B
	.dc.b	$5E,$9C,$1F,$D7,$7E,$69,$F0,$FF,$16,$7F,$B1,$46,$31,$FF,$79,$BF
	.dc.b	$D9,$5B,$3B,$94,$ED,$FB,$0D,$C7,$E9,$3C,$63,$33,$C6,$5E,$C5,$3A
	.dc.b	$4F,$F7,$3F,$07,$FB,$A7,$CC,$BE,$1F,$F5,$37,$C3,$2F,$EF,$0F,$DF
	.dc.b	$31,$7F,$89,$69,$FD,$FA,$F9,$D7,$2C,$3B,$D7,$F8,$9C,$61,$DD,$EB
	.dc.b	$1F,$94,$8C,$47,$F8,$C0,$DC,$F3,$C3,$AC,$7F,$8B,$CC,$9C,$9C,$BC
	.dc.b	$67,$9E,$1E,$87,$AC,$7C,$13,$88,$8D,$57,$F8,$6A,$33,$18,$7C,$13
	.dc.b	$D2,$FD,$15,$B8,$C5,$F8,$AB,$B4,$61,$EE,$62,$7D,$8E,$0F,$8C,$7F
	.dc.b	$88,$F0,$FF,$12,$3E,$73,$07,$F9,$4B,$E6,$5F,$00,$F9,$AF,$F1,$54
	.dc.b	$61,$B1,$99,$E0,$D7,$FD,$81,$2F,$7F,$E1,$47,$56,$BA,$6B,$FC,$BA
	.dc.b	$38,$69,$E0,$9E,$39,$7D,$AC,$BF,$43,$7F,$51,$AF,$D0,$BE,$2D,$AE
	.dc.b	$33,$FC,$E8,$DD,$73,$D7,$F9,$EB,$F8,$6B,$E2,$DA,$E3,$7F,$E4,$FB
	.dc.b	$E2,$73,$BE,$2F,$A6,$77,$2B,$E7,$3B,$8F,$45,$70,$7F,$CB,$53,$A6
	.dc.b	$D7,$2A,$EF,$1E,$1A,$E0,$F5,$9D,$47,$EB,$B4,$ED,$3F,$CC,$EB,$E2
	.dc.b	$E3,$FC,$B7,$0E,$D3,$FB,$5B,$E0,$FB,$7E,$4A,$BF,$B9,$6D,$F2,$C3
	.dc.b	$FB,$1F,$CD,$7F,$02,$AF,$DD,$C6,$37,$E3,$39,$9F,$E8,$8F,$E0,$DD
	.dc.b	$9F,$E7,$97,$59,$7F,$42,$E5,$FF,$02,$4E,$A1,$18,$8E,$75,$E6,$E4
	.dc.b	$FE,$5B,$1D,$0F,$E5,$EE,$4F,$6D,$3F,$5D,$FF,$24,$4E,$E7,$8B,$FE
	.dc.b	$B2,$6B,$FE,$00,$39,$9F,$E8,$B7,$2F,$F8,$13,$CE,$FE,$E8,$F8,$5A
	.dc.b	$FF,$63,$6D,$7F,$5E,$F9,$2F,$F1,$34,$EF,$9B,$E7,$D6,$B4,$E9,$1E
	.dc.b	$8A,$DD,$73,$8F,$DC,$D7,$68,$F3,$35,$E2,$7F,$A8,$73,$FF,$54,$37
	.dc.b	$D9,$FE,$03,$18,$8F,$F6,$8B,$5F,$81,$97,$CF,$6F,$93,$FC,$22,$31
	.dc.b	$1F,$E2,$46,$BF,$EB,$3B,$7C,$37,$98,$91,$FE,$63,$7C,$3E,$DC,$8F
	.dc.b	$33,$5F,$A5,$E5,$77,$23,$F4,$5F,$83,$CD,$23,$FC,$DC,$D7,$D7,$E7
	.dc.b	$FB,$88,$FC,$37,$D5,$F2,$52,$3E,$E3,$5F,$F3,$B7,$C1,$FC,$E2,$3F
	.dc.b	$E7,$B7,$E4,$FB,$72,$3E,$73,$5F,$C5,$7C,$9F,$C7,$23,$F8,$97,$EC
	.dc.b	$FC,$1C,$8F,$F4,$33,$5F,$C8,$0F,$99,$7A,$DF,$CE,$1B,$1F,$ED,$D6
	.dc.b	$BC,$1A,$FE,$67,$E6,$DE,$FF,$F4,$06,$C7,$FB,$95,$AF,$81,$AF,$E8
	.dc.b	$7E,$69,$F3,$6F,$87,$FF,$76,$46,$1B,$83,$C1,$CD,$6E,$FF,$1B,$8F
	.dc.b	$FB,$D7,$A3,$C7,$D9,$F9,$58,$C1,$FE,$F9,$F9,$97,$CD,$52,$4F,$FB
	.dc.b	$EE,$32,$FF,$BF,$71,$70,$7F,$C0,$16,$C6,$63,$07,$87,$E1,$3F,$C5
	.dc.b	$1E,$07,$F0,$3F,$63,$FC,$49,$9A,$73,$E1,$7F,$9E,$35,$FF,$2E,$4E
	.dc.b	$71,$FD,$65,$73,$69,$27,$71,$CB,$6E,$AF,$4F,$A0,$95,$FC,$77,$0F
	.dc.b	$BC,$D8,$CD,$FF,$92,$65,$1E,$4E,$0C,$E7,$FE,$68,$9F,$F3,$A5,$F3
	.dc.b	$DF,$D2,$FA,$9A,$E4,$F0,$19,$3F,$94,$67,$CB,$1F,$F2,$C7,$86,$FA
	.dc.b	$0F,$8B,$5F,$22,$30,$F0,$3E,$E1,$0F,$E2,$05,$9F,$92,$F1,$FA,$0E
	.dc.b	$6F,$F3,$D9,$7D,$07,$53,$B7,$C4,$D9,$D4,$5B,$F2,$CD,$D3,$F9,$67
	.dc.b	$89,$E5,$F9,$2F,$33,$70,$EC,$DB,$F6,$B6,$BF,$D2,$1F,$3D,$F9,$DA
	.dc.b	$F2,$78,$FF,$E6,$78,$C5,$FC,$48,$C7,$57,$FA,$5E,$3F,$D8,$B3,$A7
	.dc.b	$FA,$43,$5C,$C6,$D0,$8E,$B7,$E2,$9F,$9D,$E2,$76,$C3,$E3,$74,$7D
	.dc.b	$13,$E9,$78,$56,$32,$4C,$7E,$CC,$65,$FB,$7F,$3D,$F6,$7E,$7B,$ED
	.dc.b	$36,$BE,$93,$67,$71,$F5,$67,$FD,$11,$AF,$FA,$16,$3F,$CF,$3C,$2F
	.dc.b	$9E,$CF,$A1,$F8,$56,$CF,$F6,$78,$F8,$18,$7F,$A8,$9B,$7F,$E9,$18
	.dc.b	$FE,$9F,$1C,$1F,$C8,$FB,$B7,$7F,$ED,$1F,$3D,$F9,$13,$F2,$5E,$95
	.dc.b	$C7,$F1,$51,$6F,$F3,$33,$6B,$FB,$87,$CE,$FE,$67,$E7,$BF,$34,$DC
	.dc.b	$3F,$24,$DB,$FD,$0B,$6B,$FD,$A9,$F3,$3F,$CC,$AB,$F3,$15,$FE,$AB
	.dc.b	$8C,$3F,$A8,$31,$E4,$FE,$9D,$B3,$C0,$FF,$0B,$4F,$0A,$FF,$02,$4E
	.dc.b	$8E,$53,$AA,$E4,$D3,$94,$7F,$BB,$EB,$94,$7D,$48,$E5,$5D,$23,$84
	.dc.b	$74,$38,$57,$46,$C7,$49,$F7,$CE,$5B,$19,$3F,$33,$2A,$F0,$BF,$E9
	.dc.b	$FD,$FF,$D9,$15,$C2,$3C,$0D,$3C,$8F,$80,$CD,$74,$AE,$72,$AE,$73
	.dc.b	$9B,$F2,$34,$F2,$5F,$EB,$2F,$8D,$70,$8C,$57,$06,$D7,$07,$D5,$19
	.dc.b	$6C,$66,$B9,$C7,$CC,$8F,$9C,$7C,$CB,$F5,$9E,$D7,$31,$ED,$8C,$C7
	.dc.b	$9E,$3D,$E7,$88,$A7,$80,$B3,$E0,$8E,$C7,$49,$ED,$3F,$3C,$F2,$16
	.dc.b	$36,$D3,$C4,$F1,$3C,$51,$E4,$8F,$13,$5F,$13,$67,$C5,$1C,$8E,$91
	.dc.b	$E2,$3E,$B3,$E2,$6C,$FE,$42,$3F,$CD,$53,$FA,$66,$D7,$F0,$48,$C4
	.dc.b	$7C,$53,$33,$FE,$22,$7E,$21,$83,$FC,$4B,$1F,$90,$7C,$F1,$FC,$B2
	.dc.b	$9D,$0F,$9E,$F8,$B5,$A3,$DF,$5B,$9D,$3F,$22,$34,$6E,$B9,$D7,$2B
	.dc.b	$F0,$1F,$96,$BE,$35,$F1,$4E,$31,$99,$E1,$87,$C5,$AD,$DF,$DA,$BD
	.dc.b	$CF,$F8,$52,$73,$87,$D6,$24,$FA,$05,$79,$23,$D1,$87,$92,$F9,$61
	.dc.b	$E0,$9F,$C1,$D6,$EF,$CD,$7B,$9F,$F5,$76,$1E,$09,$EC,$79,$23,$37
	.dc.b	$CE,$3F,$05,$7C,$CF,$55,$F3,$8F,$2D,$7C,$DA,$E1,$5F,$87,$BE,$D5
	.dc.b	$D4,$E6,$E2,$F9,$4E,$EB,$8D,$FC,$28,$F0,$3F,$12,$F9,$42,$3C,$07
	.dc.b	$C4,$C3,$E8,$57,$2A,$F9,$55,$C9,$F7,$DF,$82,$F8,$C7,$C6,$BE,$31
	.dc.b	$CC,$F9,$78,$79,$E3,$EC,$7C,$74,$9F,$97,$F1,$F6,$FE,$19,$F6,$2D
	.dc.b	$59,$C7,$D1,$39,$AC,$5F,$E5,$7E,$3F,$F0,$98,$FE,$15,$B1,$F6,$1B
	.dc.b	$5D,$AF,$F4,$31,$87,$FA,$CF,$A9,$F8,$DF,$9E,$FE,$B0,$DB,$FE,$A6
	.dc.b	$FE,$CC,$8F,$EC,$74,$FF,$5D,$B6,$38,$BF,$88,$F6,$2E,$3E,$1D,$5F
	.dc.b	$3D,$F9,$B3,$F3,$9F,$3D,$6F,$F1,$F1,$6B,$84,$DC,$FE,$19,$FB,$AF
	.dc.b	$9E,$FD,$26,$5F,$A5,$6E,$1F,$98,$6D,$F8,$5B,$5F,$DE,$3D,$6F,$D6
	.dc.b	$7C,$07,$F7,$76,$DF,$F7,$27,$F8,$91,$1F,$80,$DB,$FE,$DC,$FE,$D7
	.dc.b	$D9,$E5,$3F,$A8,$F7,$7F,$71,$F3,$DF,$B9,$6C,$64,$FD,$6C,$65,$A7
	.dc.b	$EC,$3E,$7B,$F8,$01,$FC,$06,$EC,$64,$98,$7C,$D8,$CB,$F8,$C1,$F0
	.dc.b	$8F,$88,$FE,$15,$F3,$DF,$2A,$FF,$BB,$CE,$5F,$DF,$C6,$6B,$FC,$A7
	.dc.b	$39,$7F,$DA,$2D,$7F,$BF,$9D,$AB,$FC,$77,$19,$8E,$B1,$CD,$D5,$7F
	.dc.b	$0C,$F7,$71,$5D,$76,$FB,$BC,$7F,$E6,$0A,$FE,$65,$F3,$DF,$C8,$7C
	.dc.b	$BF,$E4,$4D,$8F,$E2,$AD,$C7,$F1,$CF,$FA,$EB,$E6,$5F,$3D,$4C,$69
	.dc.b	$6F,$FA,$21,$FD,$16,$76,$E2,$FF,$95,$C6,$1F,$D5,$3E,$78,$E1,$5E
	.dc.b	$AF,$85,$F0,$7C,$7D,$9E,$5A,$2D,$39,$C3,$E5,$DF,$C8,$25,$FF,$5D
	.dc.b	$D5,$1B,$C3,$F4,$38,$3E,$7B,$17,$FE,$30,$F2,$BF,$AE,$BF,$D9,$5B
	.dc.b	$8F,$D6,$0E,$1F,$B2,$D7,$FC,$73,$F0,$FF,$C4,$A7,$FB,$73,$63,$93
	.dc.b	$5F,$C0,$B5,$FF,$20,$FC,$93,$C3,$FF,$3E,$B6,$BF,$BC,$36,$7C,$8D
	.dc.b	$7F,$17,$F3,$7F,$3D,$BB,$FF,$A1,$A3,$11,$FE,$4E,$6B,$FE,$A8,$C3
	.dc.b	$FC,$0E,$DF,$13,$FD,$1B,$18,$83,$77,$7C,$47,$E6,$09,$97,$D1,$7F
	.dc.b	$D2,$BF,$49,$F4,$A3,$FC,$20,$DB,$F9,$8D,$7F,$46,$D7,$FC,$C7,$F3
	.dc.b	$8F,$9C,$FF,$4F,$B6,$BF,$C3,$4D,$9F,$AC,$D7,$F5,$BF,$38,$F6,$BF
	.dc.b	$D4,$ED,$AF,$F1,$13,$67,$EF,$35,$FD,$AF,$D2,$3C,$4F,$F5,$7B,$6B
	.dc.b	$FC,$54,$D9,$FC,$33,$5F,$DE,$FC,$EB,$E7,$7F,$D6,$ED,$AF,$F1,$93
	.dc.b	$67,$F1,$CD,$7F,$82,$FC,$EB,$E7,$BF,$D7,$ED,$AF,$F1,$D3,$67,$F2
	.dc.b	$CD,$7F,$86,$FD,$97,$FB,$92,$7F,$C8,$0D,$8F,$CE,$35,$D9,$FE,$CC
	.dc.b	$F9,$FF,$C2,$A6,$93,$9B,$FD,$A3,$E7,$3E,$7E,$93,$0F,$F2,$74,$65
	.dc.b	$FF,$28,$7C,$FB,$DB,$FF,$6C,$B6,$33,$18,$3F,$DB,$BF,$3F,$FC,$1F
	.dc.b	$2F,$F7,$ED,$6D,$FE,$6D,$19,$51,$9F,$07,$FC,$03,$09,$FE,$58,$8C
	.dc.b	$78,$9D,$89,$87,$F9,$97,$E7,$BC,$16,$F9,$FF,$AC,$F4,$8C,$C6,$27
	.dc.b	$FD,$73,$18,$E3,$FE,$70,$D8,$FE,$49,$82,$3C,$6D,$C3,$FD,$F7,$18
	.dc.b	$33,$F3,$ED,$7F,$A6,$65,$FC,$4B,$D9,$FE,$89,$39,$C7,$F9,$3F,$89
	.dc.b	$FC,$3E,$BE,$0C,$E7,$2F,$F0,$D4,$E6,$BF,$D9,$F3,$9A,$FF,$02,$CE
	.dc.b	$6B,$E3,$7C,$FB,$E0,$7F,$A9,$37,$0F,$5B,$FE,$C3,$93,$97,$FB,$53
	.dc.b	$6F,$E9,$B6,$7F,$DB,$EF,$13,$13,$FE,$8F,$F4,$FF,$B0,$51,$FE,$C5
	.dc.b	$1F,$DB,$07,$CF,$3E,$07,$FD,$D3,$5B,$C3,$E7,$B6,$BF,$BE,$34,$FC
	.dc.b	$2B,$CE,$B9,$3E,$19,$FA,$EE,$6B,$AD,$FF,$93,$2B,$73,$FE,$02,$3C
	.dc.b	$73,$AA,$FC,$74,$EA,$79,$1D,$4D,$9F,$E0,$2F,$13,$75,$C6,$32,$75
	.dc.b	$9E,$F8,$7A,$DE,$91,$FE,$16,$37,$19,$9F,$F0,$7C,$63,$0F,$44,$62
	.dc.b	$B9,$C6,$27,$C4,$6F,$0F,$1B,$89,$DC,$62,$3E,$01,$FD,$BC,$FF,$66
	.dc.b	$1E,$33,$07,$F8,$7B,$E7,$9E,$C3,$E7,$FF,$DA,$51,$86,$C6,$67,$FB
	.dc.b	$84,$E5,$BE,$CF,$FA,$D9,$B9,$F2,$8C,$47,$76,$BF,$BF,$12,$8E,$58
	.dc.b	$FD,$AC,$FF,$3D,$5E,$D6,$BF,$4A,$B8,$5F,$C1,$6C,$7D,$69,$DD,$73
	.dc.b	$DB,$C0,$FE,$1A,$B6,$DC,$3C,$35,$E5,$38,$96,$B8,$BE,$9A,$E2,$D9
	.dc.b	$E3,$7F,$C6,$E3,$8B,$6B,$8B,$FC,$7E,$B8,$B6,$B8,$BF,$A9,$9D,$B4
	.dc.b	$F7,$46,$AB,$E1,$E1,$FC,$B1,$B9,$F9,$EB,$F8,$EE,$1D,$9E,$47,$F9
	.dc.b	$AE,$5F,$31,$FB,$6A,$F8,$0D,$9F,$41,$FC,$9B,$E7,$FF,$80,$37,$6D
	.dc.b	$28,$CB,$FE,$77,$7E,$E3,$B3,$FC,$93,$F2,$5F,$E0,$16,$F8,$FC,$A7
	.dc.b	$8E,$7F,$D1,$BE,$B1,$EF,$78,$E1,$E7,$FF,$90,$30,$FF,$29,$B5,$FF
	.dc.b	$43,$57,$07,$F9,$EB,$5F,$F6,$44,$6E,$7F,$D6,$2D,$D3,$F2,$95,$B9
	.dc.b	$FE,$AF,$F0,$6F,$9F,$FD,$5B,$70,$FF,$29,$07,$0F,$F9,$55,$AF,$FA
	.dc.b	$31,$CC,$F8,$3C,$EF,$66,$3F,$E2,$BA,$FE,$A7,$5A,$C7,$F4,$CD,$85
	.dc.b	$FC,$2B,$E2,$7D,$29,$D4,$F8,$FE,$87,$CE,$47,$CF,$08,$F5,$FC,$D4
	.dc.b	$9C,$B5,$F5,$79,$7F,$00,$8F,$27,$E1,$FC,$E4,$8E,$AD,$7E,$17,$CE
	.dc.b	$7C,$82,$3E,$27,$E5,$FC,$DC,$8F,$2B,$5F,$95,$F3,$9F,$40,$8F,$A9
	.dc.b	$FA,$7F,$37,$23,$DA,$D7,$E9,$6D,$F6,$08,$FC,$27,$F9,$2F,$D7,$3E
	.dc.b	$72,$92,$7E,$EC,$65,$BF,$39,$C0,$8E,$0D,$FC,$DF,$42,$38,$3F,$0F
	.dc.b	$F8,$07,$FA,$A1,$BF,$09,$FC,$DA,$31,$1F,$E9,$D6,$BF,$ED,$47,$FC
	.dc.b	$A0,$F2,$DF,$81,$1F,$3F,$CE,$FB,$08,$E0,$FC,$BF,$D0,$B8,$1F,$11
	.dc.b	$86,$CF,$FA,$B6,$5C,$3F,$C1,$A3,$13,$A7,$FD,$CF,$F4,$AD,$3F,$BC
	.dc.b	$C7,$F8,$EE,$31,$C1,$FA,$66,$C7,$FA,$E5,$AF,$FB,$C3,$EA,$1F,$25
	.dc.b	$CA,$11,$8A,$FD,$7B,$63,$FD,$84,$D7,$FD,$F3,$F4,$4E,$4F,$F1,$AF
	.dc.b	$67,$FA,$2F,$B3,$FC,$A3,$39,$C3,$F8,$4F,$83,$FE,$FB,$8C,$61,$FE
	.dc.b	$CD,$6B,$D3,$0F,$24,$62,$3F,$AD,$3C,$2F,$AC,$F8,$3E,$83,$A3,$E2
	.dc.b	$FA,$0F,$1F,$D0,$5C,$BF,$8C,$B7,$1F,$E2,$AD,$C3,$F8,$AB,$EA,$FA
	.dc.b	$09,$5F,$C8,$3E,$82,$DF,$F1,$C9,$B7,$FC,$72,$6D,$7F,$6F,$F2,$7F
	.dc.b	$DD,$33,$FC,$83,$E8,$13,$FE,$62,$9C,$BF,$E7,$39,$FE,$5F,$3A,$C3
	.dc.b	$F8,$8C,$F3,$F9,$3F,$AD,$7F,$EE,$F9,$FF,$3D,$35,$FE,$9B,$1A,$39
	.dc.b	$5F,$FA,$0A,$7B,$CF,$F1,$CE,$0F,$F4,$35,$F5,$3F,$9E,$FD,$07,$DA
	.dc.b	$F3,$7E,$DB,$63,$27,$F4,$28,$CB,$4F,$E8,$BF,$3F,$F8,$23,$F0,$7F
	.dc.b	$41,$63,$24,$C3,$FA,$2C,$65,$F7,$9E,$CE,$0F,$C5,$FA,$5E,$AA,$F9
	.dc.b	$91,$87,$E5,$3F,$10,$8F,$F4,$6C,$D1,$8A,$EA,$7F,$72,$7F,$2F,$1D
	.dc.b	$63,$C1,$3F,$4C,$FA,$9F,$AF,$3E,$83,$FD,$22,$D8,$FE,$C7,$38,$99
	.dc.b	$1C,$63,$FA,$4C,$ED,$FE,$C6,$68,$F7,$EC,$8F,$F4,$83,$7E,$53,$BF
	.dc.b	$BB,$E0,$6D,$F5,$9D,$36,$B8,$89,$F6,$5F,$F8,$43,$7F,$4F,$06,$33
	.dc.b	$46,$30,$FC,$A3,$F8,$E2,$3F,$2D,$B5,$F2,$A7,$8C,$F8,$7C,$DE,$95
	.dc.b	$F1,$BD,$E9,$3E,$5F,$A0,$B8,$7C,$0C,$FF,$D6,$13,$F3,$AB,$E6,$D6
	.dc.b	$B4,$FC,$CD,$7F,$11,$6F,$CF,$FD,$2E,$FF,$A6,$DC,$3E,$13,$71,$EC
	.dc.b	$FC,$2F,$A0,$FA,$57,$D3,$D5,$C7,$17,$DC,$6E,$1F,$2D,$B8,$79,$DB
	.dc.b	$5C,$67,$FD,$B7,$7D,$1F,$C3,$CE,$9A,$6A,$FB,$34,$EC,$5B,$EC,$DB
	.dc.b	$ED,$1F,$98,$BE,$CF,$E3,$2F,$B3,$6B,$81,$E0,$78,$5F,$81,$B1,$E0
	.dc.b	$69,$E0,$2D,$F8,$0F,$E2,$4A,$FC,$31,$F9,$7B,$F0,$B6,$BC,$35,$E3
	.dc.b	$39,$BE,$38,$E7,$5C,$B2,$FF,$27,$CF,$09,$E5,$5C,$0E,$4F,$0C,$3D
	.dc.b	$4D,$9F,$53,$67,$D4,$D3,$D5,$1F,$EA,$1B,$F0,$61,$E8,$C3,$C3,$7D
	.dc.b	$EF,$C2,$71,$BF,$41,$C7,$0F,$0B,$4F,$09,$67,$C2,$D9,$F0,$C0,$F9
	.dc.b	$53,$C9,$C6,$1E,$4A,$F8,$57,$E4,$BF,$15,$7C,$29,$E8,$79,$1E,$37
	.dc.b	$CD,$EB,$1E,$47,$2D,$72,$79,$8C,$94,$CC,$FA,$1D,$9F,$EC,$92,$9E
	.dc.b	$F3,$17,$FC,$75,$B5,$FE,$0B,$8C,$46,$C8,$78,$27,$DA,$7F,$21,$F9
	.dc.b	$0F,$CD,$C7,$8A,$FD,$EF,$F2,$A8,$CD,$7E,$8A,$34,$66,$7F,$22,$E0
	.dc.b	$CD,$75,$8C,$CF,$13,$F2,$55,$D2,$FF,$08,$D9,$FA,$6D,$BF,$C3,$60
	.dc.b	$8C,$61,$EA,$8C,$57,$6B,$E6,$D3,$99,$6F,$9D,$F2,$8F,$7D,$F9,$E7
	.dc.b	$F4,$47,$AF,$0F,$3C,$F1,$9F,$05,$71,$AF,$99,$3F,$D6,$23,$11,$F0
	.dc.b	$B0,$F2,$BD,$2F,$E1,$61,$FD,$72,$B7,$1D,$A7,$53,$EA,$6D,$FA,$9A
	.dc.b	$7A,$9F,$21,$EA,$BE,$F8,$7E,$6A,$7D,$55,$DE,$BF,$3F,$ED,$FF,$6C
	.dc.b	$3F,$B6,$FB,$B6,$3F,$AE,$45,$AF,$E9,$13,$6F,$FA,$2F,$B7,$FD,$D3
	.dc.b	$E8,$3F,$BA,$B7,$0F,$EC,$8D,$BF,$EA,$0D,$AF,$E3,$BF,$0F,$FE,$CB
	.dc.b	$AF,$ED,$5F,$3F,$FD,$A2,$31,$7F,$D5,$E3,$1C,$BF,$E6,$79,$CC,$FE
	.dc.b	$95,$AF,$EF,$C3,$87,$F8,$13,$D2,$FC,$33,$FE,$0C,$FA,$0E,$B8,$7F
	.dc.b	$7B,$F9,$FF,$F0,$79,$FE,$10,$FA,$0B,$19,$26,$3F,$DA,$E3,$2F,$F6
	.dc.b	$DF,$47,$FC,$2F,$F4,$1F,$E1,$86,$D7,$F8,$1D,$B3,$B8,$FF,$05,$4F
	.dc.b	$F8,$23,$EA,$5F,$19,$FE,$25,$D1,$F4,$98,$BF,$F0,$F6,$0B,$C3,$FE
	.dc.b	$1C,$9B,$5F,$D2,$A7,$33,$FC,$2D,$AF,$C7,$6B,$FC,$9D,$AF,$5B,$F2
	.dc.b	$1E,$3A,$7C,$9E,$30,$FF,$12,$43,$E8,$34,$6F,$DB,$B3,$A8,$B8,$79
	.dc.b	$E6,$E5,$FE,$68,$B7,$D0,$77,$F9,$5F,$03,$71,$EE,$DC,$3E,$03,$6F
	.dc.b	$F1,$BB,$F9,$FE,$8B,$F1,$8D,$BF,$93,$C8,$8F,$C6,$12,$3F,$98,$61
	.dc.b	$F8,$88,$C3,$FE,$5A,$9E,$13,$F0,$3E,$9F,$AB,$F0,$BE,$23,$E3,$1F
	.dc.b	$1F,$E8,$2C,$64,$97,$F5,$E3,$2F,$D9,$F9,$5F,$9D,$F4,$DF,$3D,$B5
	.dc.b	$F3,$9B,$3B,$8F,$A3,$3F,$E5,$0F,$8E,$FA,$51,$F5,$3E,$2F,$1C,$7F
	.dc.b	$65,$B1,$EE,$B5,$70,$7E,$0B,$5F,$EA,$4D,$7F,$5C,$EF,$7F,$F0,$BB
	.dc.b	$67,$FD,$59,$C4,$F8,$BE,$F7,$FE,$2B,$ED,$FE,$AC,$4A,$F1,$78,$88
	.dc.b	$FF,$72,$7C,$93,$92,$92,$BD,$11,$97,$D5,$F4,$7F,$98,$27,$57,$FB
	.dc.b	$B5,$FC,$DF,$3F,$F9,$D6,$7E,$03,$72,$FB,$ED,$7F,$B8,$35,$FE,$18
	.dc.b	$BD,$1F,$F7,$C3,$67,$9C,$60,$CF,$A1,$23,$BB,$6B,$F1,$BF,$41,$FE
	.dc.b	$C9,$27,$B1,$FE,$40,$8F,$E3,$51,$9F,$96,$C4,$7F,$86,$E1,$18,$84
	.dc.b	$62,$FE,$E3,$63,$BF,$C8,$6E,$0D,$24,$FE,$0F,$E7,$9D,$3D,$5B,$19
	.dc.b	$8C,$19,$E7,$91,$C5,$B8,$7E,$AA,$BF,$05,$3D,$79,$3A,$11,$FC,$97
	.dc.b	$A7,$FC,$31,$BE,$17,$B6,$31,$7F,$B4,$6B,$FE,$25,$5E,$16,$C9,$5D
	.dc.b	$1E,$5D,$72,$3B,$B6,$BF,$7F,$F7,$0F,$1F,$E6,$36,$BF,$4C,$D9,$DC
	.dc.b	$E8,$FF,$5F,$C6,$7B,$7F,$CF,$2B,$C7,$CC,$E9,$EA,$4B,$EA,$FD,$BF
	.dc.b	$0B,$B6,$1F,$E4,$16,$BF,$D7,$5A,$FF,$97,$5A,$9F,$14,$EE,$BF,$A5
	.dc.b	$77,$BE,$3B,$FD,$AB,$5A,$BE,$55,$83,$73,$6F,$FA,$7F,$CC,$BA,$3F
	.dc.b	$20,$D8,$FE,$14,$DB,$FE,$52,$D7,$FC,$B4,$E7,$4F,$F4,$EB,$67,$FC
	.dc.b	$ED,$1F,$9C,$F6,$7F,$DD,$33,$9A,$FF,$3D,$7C,$CB,$DC,$FF,$76,$B7
	.dc.b	$0F,$EC,$6F,$F9,$FD,$B1,$A2,$3A,$9A,$30,$FF,$57,$8F,$3F,$FD,$F2
	.dc.b	$D9,$FE,$DF,$3F,$E8,$BF,$07,$FD,$19,$F3,$0F,$80,$FF,$08,$5E,$EB
	.dc.b	$A5,$E3,$0E,$61,$E7,$0C,$39,$DE,$E7,$31,$89,$F8,$B5,$C5,$61,$E1
	.dc.b	$7F,$D0,$D3,$B7,$D9,$3B,$9D,$47,$F8,$96,$75,$87,$9E,$73,$3E,$D3
	.dc.b	$AE,$1E,$36,$9D,$8A,$76,$C3,$C0,$F1,$9E,$F5,$D6,$BF,$B5,$DF,$83
	.dc.b	$0F,$24,$62,$F9,$D6,$DE,$D7,$CA,$B7,$18,$AF,$24,$FA,$23,$E5,$1F
	.dc.b	$D6,$CA,$7F,$46,$31,$7F,$D7,$FE,$61,$EA,$3E,$67,$0D,$AF,$F0,$0C
	.dc.b	$EA,$72,$71,$F5,$BF,$11,$3A,$6B,$FE,$B7,$94,$62,$3F,$CA,$91,$8C
	.dc.b	$F9,$1F,$5F,$2F,$CE,$4D,$D7,$FC,$B5,$9F,$D8,$BE,$13,$F4,$1B,$1F
	.dc.b	$72,$B8,$CF,$D4,$6C,$71,$C3,$CE,$DC,$3B,$54,$3A,$4E,$AB,$CE,$D8
	.dc.b	$FF,$75,$47,$43,$FC,$DC,$57,$FD,$DD,$7E,$48,$E3,$5D,$6F,$CD,$87
	.dc.b	$B2,$73,$7E,$78,$FD,$4E,$BD,$1B,$1E,$06,$AC,$3B,$39,$3F,$CC,$74
	.dc.b	$F7,$F6,$D8,$CD,$72,$F9,$9F,$DF,$47,$F9,$A9,$AF,$EE,$00,$73,$97
	.dc.b	$EC,$63,$19,$F3,$DF,$CF,$3F,$E2,$AF,$8B,$BD,$07,$57,$FE,$78,$F9
	.dc.b	$97,$BD,$F9,$76,$C7,$F8,$CD,$AF,$F9,$49,$AF,$FA,$74,$3A,$BF,$E0
	.dc.b	$87,$F3,$5F,$1F,$C7,$D1,$FE,$06,$AF,$05,$FC,$AF,$69,$EA,$52,$4F
	.dc.b	$F3,$2E,$3F,$E6,$7F,$AF,$FF,$8E,$63,$FD,$1F,$5C,$1F,$F2,$43,$5F
	.dc.b	$F3,$3B,$5F,$33,$5F,$F5,$A3,$BC,$BC,$F8,$FF,$84,$A7,$C0,$7F,$33
	.dc.b	$70,$FD,$31,$3C,$6F,$FD,$7D,$9F,$EE,$A7,$E4,$73,$FC,$B2,$3E,$6F
	.dc.b	$1F,$F8,$25,$9A,$48,$F5,$F8,$DA,$23,$83,$EA,$71,$7D,$72,$38,$3E
	.dc.b	$1F,$CD,$76,$23,$83,$E3,$FC,$97,$88,$8E,$0F,$C0,$7B,$DF,$83,$23
	.dc.b	$83,$E7,$FC,$D7,$A8,$8E,$0F,$A7,$F1,$DE,$E2,$38,$3F,$19,$CF,$F8
	.dc.b	$F2,$38,$3F,$D0,$5F,$32,$F9,$AF,$E5,$4D,$8C,$C6,$0C,$F8,$B2,$38
	.dc.b	$B5,$FF,$54,$FE,$05,$F5,$5B,$C3,$FC,$D6,$31,$1F,$9B,$6B,$FE,$F9
	.dc.b	$F9,$BF,$F1,$DB,$FE,$95,$6B,$FE,$B4,$6B,$FA,$36,$BC,$7E,$6F,$FA
	.dc.b	$14,$1A,$48,$FF,$50,$7D,$53,$DF,$FE,$90,$D8,$CC,$60,$FE,$99,$F3
	.dc.b	$6E,$1F,$F1,$E4,$FF,$B1,$A3,$1E,$77,$FA,$AD,$7C,$DF,$F1,$19,$FD
	.dc.b	$C4,$69,$FF,$66,$FC,$D3,$E6,$FF,$CF,$4D,$EC,$FF,$5A,$36,$77,$3F
	.dc.b	$E7,$FF,$13,$D1,$5F,$A8,$EF,$78,$1E,$F8,$FD,$5D,$FF,$AF,$35,$FE
	.dc.b	$0B,$1F,$BC,$DF,$FD,$8D,$1C,$C9,$1F,$EC,$3C,$3A,$D7,$FB,$0A,$77
	.dc.b	$3D,$ED,$D5,$8A,$FE,$51,$8F,$8B,$5E,$F1,$FE,$2D,$AF,$91,$7C,$67
	.dc.b	$35,$E8,$DE,$D7,$C8,$AE,$D1,$8A,$CB,$AE,$89,$3E,$3E,$9B,$8F,$8B
	.dc.b	$05,$63,$1F,$E4,$B0,$AC,$5F,$F1,$4F,$53,$C2,$FA,$30,$FE,$1F,$18
	.dc.b	$8F,$E4,$2F,$FA,$BF,$BF,$FD,$1F,$1F,$C8,$A3,$04,$76,$7D,$5F,$07
	.dc.b	$F5,$8E,$2B,$E3,$C7,$DD,$C7,$F9,$1F,$4F,$C4,$8C,$57,$F9,$91,$FF
	.dc.b	$3C,$3C,$F8,$B5,$3B,$6C,$77,$F6,$F9,$1C,$FE,$72,$C6,$49,$8F,$8A
	.dc.b	$32,$FF,$2F,$F9,$BF,$93,$3E,$9F,$23,$E2,$46,$27,$D3,$18,$E0,$FF
	.dc.b	$69,$4E,$67,$FA,$B3,$5F,$ED,$61,$C3,$FD,$B9,$E9,$A7,$F6,$D6,$DF
	.dc.b	$C1,$F9,$BE,$CF,$CB,$F9,$5F,$8D,$F3,$7F,$1D,$B1,$93,$E8,$46,$5A
	.dc.b	$7D,$3F,$9B,$F9,$C7,$CF,$F9,$BB,$19,$26,$5F,$5A,$32,$FF,$53,$F4
	.dc.b	$5B,$FD,$7F,$9B,$FC,$44,$62,$FF,$01,$18,$9F,$F7,$B4,$E6,$BF,$BD
	.dc.b	$1F,$E1,$78,$C9,$2B,$F7,$33,$BD,$7F,$B3,$FD,$07,$39,$FC,$67,$CD
	.dc.b	$7E,$1F,$D4,$FC,$43,$63,$27,$E4,$63,$2D,$3F,$29,$F3,$9F,$8F,$3F
	.dc.b	$21,$F3,$96,$32,$4C,$FF,$2F,$19,$7E,$17,$07,$E5,$FD,$6F,$82,$DA
	.dc.b	$F4,$B6,$77,$1F,$A3,$9D,$B4,$D9,$63,$67,$E9,$A3,$6D,$8D,$B4,$DE
	.dc.b	$7F,$E0,$C9,$DB,$67,$6D,$9D,$C7,$EB,$67,$6D,$36,$58,$D9,$FE,$15
	.dc.b	$8D,$B6,$36,$D3,$73,$FE,$E7,$9D,$B6,$76,$D9,$DF,$4F,$FB,$6D,$B5
	.dc.b	$C2,$31,$3F,$E5,$1D,$58,$7C,$D8,$C3,$70,$FD,$7C,$E9,$F9,$93,$A6
	.dc.b	$9A,$9E,$18,$7F,$98,$2B,$83,$F3,$6B,$83,$6B,$81,$C9,$E1,$5C,$9F
	.dc.b	$F3,$93,$6B,$9D,$7D,$13,$99,$6B,$9D,$75,$39,$BD,$5E,$6D,$9E,$6F
	.dc.b	$D5,$9E,$6D,$39,$BD,$67,$99,$D4,$E6,$59,$E6,$FD,$98,$E6,$D9,$E6
	.dc.b	$75,$79,$C7,$57,$99,$FE,$75,$9F,$92,$DA,$FE,$02,$4A,$FE,$03,$1C
	.dc.b	$0F,$82,$6D,$CC,$F9,$23,$CF,$5F,$E6,$88,$FC,$8C,$62,$7C,$AF,$F9
	.dc.b	$A6,$32,$71,$3D,$B2,$BF,$F5,$35,$7E,$62,$B9,$1E,$C7,$FA,$85,$72
	.dc.b	$BF,$C9,$36,$3F,$33,$5C,$AB,$F4,$A7,$23,$9D,$F7,$AE,$11,$F9,$E6
	.dc.b	$C7,$CE,$35,$5E,$38,$F4,$5F,$F6,$37,$71,$EE,$9D,$D7,$82,$77,$3F
	.dc.b	$DB,$DD,$C7,$AC,$F8,$91,$D6,$BD,$6D,$3D,$73,$C2,$7D,$37,$ED,$BF
	.dc.b	$C1,$C7,$B6,$7E,$5C,$E6,$BE,$8C,$E6,$FD,$11,$D6,$3E,$03,$6F,$E0
	.dc.b	$57,$19,$F3,$1C,$67,$D4,$6C,$F8,$11,$B8,$D1,$D0,$F8,$55,$EC,$7F
	.dc.b	$88,$FC,$C3,$E6,$3F,$C5,$ED,$BF,$F0,$A4,$E2,$64,$7F,$92,$E7,$53
	.dc.b	$FE,$16,$32,$7F,$8F,$97,$C5,$74,$F8,$0E,$AD,$AE,$4D,$AE,$0D,$CB
	.dc.b	$F8,$D7,$07,$93,$3F,$ED,$2D,$BF,$EC,$2F,$F6,$82,$3F,$A4,$6D,$FF
	.dc.b	$73,$7D,$BD,$FC,$A7,$D5,$F0,$BF,$00,$F8,$3F,$39,$63,$24,$BF,$64
	.dc.b	$65,$F7,$7C,$E7,$C8,$F9,$CF,$92,$DA,$F8,$CD,$9D,$C7,$C9,$9F,$F2
	.dc.b	$57,$C9,$FC,$AD,$3F,$DE,$BB,$FF,$C0,$0A,$FF,$3A,$CE,$55,$A9,$5B
	.dc.b	$8F,$F7,$15,$7D,$47,$FC,$56,$1C,$3F,$E2,$E6,$BD,$1A,$EF,$83,$CA
	.dc.b	$7E,$13,$E7,$2E,$1F,$81,$F6,$7F,$0E,$7E,$23,$D5,$B1,$92,$5F,$E0
	.dc.b	$E3,$2F,$E1,$7D,$8F,$C7,$FB,$3F,$90,$6D,$7E,$35,$B3,$B8,$FC,$84
	.dc.b	$FF,$9C,$7E,$5F,$FD,$FF,$1F,$93,$F9,$5C,$71,$FE,$69,$B1,$FC,$AA
	.dc.b	$77,$7F,$E2,$C6,$BF,$19,$AF,$F9,$49,$DE,$E9,$C5,$FE,$9B,$E7,$3A
	.dc.b	$DF,$E7,$FE,$2B,$F5,$27,$EA,$BE,$66,$DF,$E9,$E2,$CF,$E9,$66,$E1
	.dc.b	$F8,$0F,$67,$F6,$3F,$39,$FB,$26,$E1,$FA,$F6,$DF,$EB,$5B,$5E,$58
	.dc.b	$FC,$2F,$CE,$3E,$63,$FD,$53,$F3,$57,$0F,$F5,$FC,$DA,$FF,$59,$CD
	.dc.b	$BE,$50,$9D,$3F,$8A,$6B,$FE,$76,$6B,$E0,$BE,$2E,$5F,$F5,$34,$75
	.dc.b	$BF,$CC,$D2,$DF,$27,$AB,$FD,$1F,$77,$2C,$7F,$28,$D7,$F3,$4D,$E8
	.dc.b	$FE,$AD,$8F,$F6,$65,$C5,$92,$46,$5E,$5F,$1C,$F9,$AA,$49,$FE,$DF
	.dc.b	$19,$7F,$B8,$FD,$3F,$F1,$A6,$F8,$5E,$38,$C6,$5F,$E7,$66,$BF,$B0
	.dc.b	$F9,$A7,$89,$E9,$6D,$FF,$7C,$6C,$FF,$9F,$9A,$F1,$BE,$04,$9E,$0F
	.dc.b	$C3,$FA,$67,$C7,$52,$4F,$F8,$22,$32,$FF,$82,$BE,$6F,$FD,$19,$7F
	.dc.b	$C1,$1A,$FF,$AC,$1A,$FF,$0E,$6A,$79,$BF,$E9,$46,$BC,$EB,$EB,$46
	.dc.b	$23,$FD,$77,$F4,$AF,$82,$FB,$4D,$BF,$F0,$EB,$67,$FD,$3E,$D7,$F8
	.dc.b	$CF,$CD,$3E,$6F,$F0,$4D,$AF,$F1,$2B,$67,$FD,$4E,$D7,$8F,$C7,$F0
	.dc.b	$24,$F0,$7F,$19,$F4,$9F,$9A,$93,$49,$CD,$F9,$1F,$33,$F2,$F7,$FE
	.dc.b	$34,$6D,$7F,$AD,$9A,$FF,$30,$6B,$FE,$DC,$5F,$03,$CC,$93,$CD,$FC
	.dc.b	$F7,$CD,$BC,$1F,$DE,$76,$FE,$D3,$E9,$1E,$CE,$30,$FF,$73,$DE,$DF
	.dc.b	$D6,$9F,$EE,$A9,$DB,$8C,$BF,$A4,$CE,$DF,$F6,$6B,$5F,$E8,$ED,$7A
	.dc.b	$E1,$E4,$3C,$0F,$26,$A8,$78,$38,$7B,$63,$FC,$33,$E7,$2E,$1F,$C0
	.dc.b	$BC,$7F,$C0,$4E,$71,$EA,$A3,$33,$FD,$83,$15,$F9,$5F,$EC,$6D,$73
	.dc.b	$7F,$C1,$E3,$07,$82,$37,$1D,$BE,$29,$F0,$BE,$AA,$D5,$3E,$73,$F8
	.dc.b	$6B,$FD,$BF,$0F,$E3,$3D,$3F,$DC,$E3,$FB,$9C,$EB,$3F,$AF,$39,$9E
	.dc.b	$5E,$1F,$F1,$9A,$E7,$1F,$DD,$1A,$FB,$5A,$FF,$7E,$F9,$CF,$F5,$DE
	.dc.b	$1F,$05,$B1,$93,$E8,$B8,$4D,$23,$AF,$67,$EB,$E2,$F3,$A9,$F6,$A7
	.dc.b	$FC,$1B,$E1,$FF,$83,$BE,$6D,$EE,$FD,$1B,$E1,$97,$E2,$EF,$FC,$25
	.dc.b	$87,$28,$C4,$72,$17,$C9,$FE,$83,$3D,$D6,$1D,$F0,$ED,$3F,$D6,$EB
	.dc.b	$85,$7E,$06,$B4,$F2,$BA,$E4,$EF,$7C,$A3,$31,$E3,$AE,$78,$7A,$A3
	.dc.b	$67,$F8,$9A,$F9,$C6,$67,$C3,$3C,$6F,$A3,$C0,$F0,$47,$4C,$3D,$13
	.dc.b	$FE,$30,$9F,$69,$BA,$E9,$19,$BE,$F5,$C2,$BC,$46,$DC,$3E,$D3,$FA
	.dc.b	$A9,$FE,$C4,$3A,$98,$3F,$C7,$FF,$36,$F4,$9F,$39,$FE,$CA,$8C,$36
	.dc.b	$33,$3E,$79,$CB,$7D,$4F,$F2,$73,$72,$E5,$18,$8E,$ED,$7F,$6D,$28
	.dc.b	$E3,$9F,$DA,$CD,$AF,$EE,$E3,$0D,$8E,$33,$C1,$B8,$7D,$E7,$8E,$16
	.dc.b	$B8,$3F,$E6,$4C,$3C,$15,$F8,$46,$C7,$D7,$AE,$2D,$AE,$2A,$77,$87
	.dc.b	$E4,$1B,$5F,$C4,$4E,$31,$E8,$9E,$D8,$79,$DF,$13,$5F,$D4,$DF,$6A
	.dc.b	$E2,$7B,$23,$C4,$7E,$AD,$F2,$E1,$FE,$32,$6E,$BB,$C3,$C1,$5F,$C4
	.dc.b	$CF,$F4,$1C,$BD,$DF,$DB,$57,$9D,$B3,$AC,$2A,$F9,$CF,$E0,$4F,$EE
	.dc.b	$DA,$F5,$35,$B7,$08,$FB,$91,$B8,$EB,$93,$99,$F1,$38,$6D,$6C,$FE
	.dc.b	$11,$F3,$4F,$9C,$F1,$B6,$33,$1D,$3E,$3F,$FC,$B5,$18,$3F,$58,$27
	.dc.b	$83,$FA,$F6,$BF,$EA,$06,$BF,$D2,$7D,$97,$CE,$7C,$26,$C7,$F1,$B6
	.dc.b	$BF,$37,$25,$BE,$73,$85,$CA,$E0,$F9,$E3,$11,$FC,$06,$F9,$BF,$EB
	.dc.b	$26,$BE,$39,$FD,$D7,$D9,$FF,$7C,$8F,$6B,$5F,$E1,$4D,$7F,$D7,$6D
	.dc.b	$7F,$B2,$63,$FB,$F7,$E2,$B5,$F8,$9C,$DF,$C1,$E3,$11,$E3,$F2,$7F
	.dc.b	$85,$60,$69,$23,$C7,$F1,$1E,$62,$38,$3B,$F3,$F5,$23,$83,$EB,$D3
	.dc.b	$E5,$7B,$92,$32,$FC,$2E,$79,$1F,$09,$AF,$DD,$F7,$FD,$44,$7D,$EF
	.dc.b	$AB,$E4,$E4,$78,$1A,$FC,$AF,$87,$F9,$E4,$7E,$7B,$F1,$79,$A4,$7C
	.dc.b	$A6,$BF,$CE,$FE,$75,$F3,$BF,$CC,$5B,$1F,$94,$6B,$B8,$FE,$6D,$F9
	.dc.b	$1B,$E5,$A9,$27,$F3,$71,$97,$F3,$BF,$3D,$F5,$9F,$F1,$03,$5F,$F8
	.dc.b	$02,$D4,$DF,$9E,$FF,$26,$C7,$E9,$9A,$FF,$53,$6B,$FE,$2B,$0E,$1F
	.dc.b	$F1,$77,$CD,$3C,$8F,$E9,$CD,$8F,$D8,$B5,$FE,$BC,$D7,$A7,$CF,$7F
	.dc.b	$9F,$DF,$DB,$B5,$FE,$C8,$D7,$7E,$D7,$F5,$BF,$9E,$7C,$E7,$F9,$2A
	.dc.b	$7F,$B5,$46,$3D,$3F,$E0,$2B,$0F,$E4,$13,$AA,$E0,$FE,$9B,$C5,$FF
	.dc.b	$25,$F8,$1F,$E7,$6F,$07,$FD,$BB,$DF,$FE,$2A,$7F,$7B,$F0,$DF,$E5
	.dc.b	$A8,$FE,$AB,$E4,$7E,$CA,$FF,$DA,$11,$8C,$3F,$67,$18,$38,$1D,$BD
	.dc.b	$7B,$5F,$E6,$FC,$3F,$84,$FD,$1F,$8B,$E8,$3C,$6D,$9D,$66,$8C,$63
	.dc.b	$FC,$46,$31,$3A,$AF,$5F,$C3,$48,$E3,$F0,$96,$F8,$4D,$C7,$FD,$13
	.dc.b	$82,$B1,$7F,$C2,$3C,$1F,$8B,$EA,$FF,$9E,$BC,$64,$62,$7F,$9B,$BF
	.dc.b	$EF,$18,$CC,$7F,$0E,$8C,$11,$D9,$EC,$FA,$3E,$91,$F4,$B9,$BE,$9F
	.dc.b	$93,$6B,$E0,$B7,$5F,$80,$DD,$3D,$2F,$DA,$F2,$25,$7C,$8F,$81,$B7
	.dc.b	$F1,$E6,$DF,$C0,$9B,$5F,$E8,$9F,$37,$FD,$17,$6A,$FE,$7B,$F4,$1F
	.dc.b	$E9,$A9,$CC,$62,$BF,$CB,$D1,$FE,$D3,$AE,$59,$27,$FA,$9C,$E6,$47
	.dc.b	$F6,$39,$D3,$FE,$43,$8F,$0B,$99,$FC,$0F,$93,$E1,$AF,$CB,$F5,$7E
	.dc.b	$63,$AA,$D7,$E1,$9B,$7F,$85,$6E,$3F,$79,$FC,$FF,$CF,$CA,$FC,$6F
	.dc.b	$91,$6F,$F1,$93,$6F,$F0,$B3,$6B,$FB,$37,$4F,$EA,$1F,$F5,$EF,$0F
	.dc.b	$FB,$61,$FC,$B9,$88,$F6,$E1,$FD,$75,$AF,$28,$FE,$83,$18,$8F,$8B
	.dc.b	$3C,$DF,$F0,$73,$6F,$FD,$BF,$E0,$7F,$2E,$BF,$EE,$97,$FE,$C4,$8F
	.dc.b	$E5,$51,$FA,$D7,$FD,$72,$DF,$13,$5E,$D6,$DB,$19,$38,$47,$3F,$9A
	.dc.b	$91,$CF,$E8,$2E,$7F,$E0,$1C,$15,$8C,$BC,$B5,$6F,$FD,$A9,$F4,$0D
	.dc.b	$BC,$DE,$A7,$FB,$92,$31,$3E,$78,$C4,$EA,$3F,$D5,$53,$AC,$BC,$15
	.dc.b	$A9,$E0,$6A,$B8,$4F,$F8,$D6,$B8,$61,$F2,$2F,$84,$F2,$BE,$07,$27
	.dc.b	$85,$F2,$AF,$B3,$7C,$AB,$F3,$31,$FD,$06,$F9,$BD,$63,$9D,$F5,$39
	.dc.b	$BD,$63,$9D,$F5,$8F,$8A,$D9,$FD,$2D,$F7,$7F,$65,$97,$F9,$22,$BA
	.dc.b	$DF,$79,$EA,$77,$7A,$C7,$73,$A9,$63,$AD,$F7,$57,$DE,$BD,$95,$D8
	.dc.b	$F7,$DF,$86,$BB,$CF,$84,$EE,$F8,$67,$BB,$4E,$EF,$85,$EF,$7E,$1B
	.dc.b	$F0,$5F,$5B,$E6,$75,$BE,$6D,$BE,$67,$53,$9C,$75,$AE,$78,$7D,$48
	.dc.b	$F4,$B6,$7E,$5B,$6B,$DF,$1E,$C9,$D9,$F0,$5D,$B6,$36,$79,$9D,$C7
	.dc.b	$13,$65,$8D,$C7,$A8,$DB,$C6,$36,$D3,$6F,$B9,$F8,$0F,$B8,$FF,$74
	.dc.b	$57,$E8,$DF,$F0,$B3,$5F,$A2,$63,$0F,$E1,$8F,$13,$15,$FE,$86,$9E
	.dc.b	$11,$DE,$3C,$CF,$E4,$3E,$83,$E4,$3E,$1B,$F3,$62,$9C,$DF,$E5,$EB
	.dc.b	$57,$F8,$89,$CE,$1F,$3E,$F8,$4F,$CD,$AD,$9E,$13,$C0,$E6,$FA,$1F
	.dc.b	$86,$BE,$71,$F4,$EB,$9D,$7D,$13,$99,$D3,$0F,$F1,$B5,$70,$8F,$A9
	.dc.b	$3C,$27,$E9,$47,$0B,$ED,$7E,$1C,$3D,$18,$71,$BF,$82,$D9,$F8,$F1
	.dc.b	$D9,$E0,$79,$63,$57,$ED,$9D,$3E,$7B,$F6,$DF,$F3,$9A,$DC,$FA,$70
	.dc.b	$F3,$61,$F5,$23,$CD,$5B,$8E,$F1,$EC,$AE,$37,$EA,$8E,$D5,$EA,$AD
	.dc.b	$E1,$EA,$69,$EA,$C3,$BD,$7D,$33,$BC,$7F,$4B,$C3,$ED,$7C,$DF,$F4
	.dc.b	$FF,$A0,$FE,$A0,$D9,$FE,$A0,$DA,$FE,$8C,$DC,$3F,$AC,$FD,$07,$F5
	.dc.b	$73,$FA,$C7,$D0,$5B,$FE,$B1,$16,$FF,$A7,$4D,$AF,$ED,$7E,$DF,$F5
	.dc.b	$0B,$FE,$BD,$F0,$FF,$E0,$A9,$CF,$0F,$F6,$38,$C4,$FF,$84,$30,$5F
	.dc.b	$17,$FC,$FC,$62,$BF,$4F,$19,$8F,$F4,$13,$5E,$4F,$1B,$EF,$5F,$DB
	.dc.b	$FD,$8E,$F5,$FE,$1E,$AD,$4F,$F2,$AF,$57,$FB,$B7,$AF,$FD,$DD,$B1
	.dc.b	$FD,$DD,$B5,$FD,$B5,$B8,$7F,$80,$3E,$83,$FB,$F1,$FD,$FF,$E8,$2D
	.dc.b	$FF,$7F,$8B,$7F,$DD,$66,$D7,$C8,$DF,$FC,$0D,$E6,$F9,$AB,$E1,$C6
	.dc.b	$30,$FE,$BA,$7B,$9F,$86,$47,$E3,$C7,$06,$E6,$9E,$8D,$8F,$F0,$93
	.dc.b	$88,$F8,$B6,$F9,$D7,$CE,$E1,$B7,$FE,$19,$C3,$13,$23,$FD,$11,$3A
	.dc.b	$9F,$F0,$91,$93,$FC,$B1,$1F,$DD,$A1,$F2,$7D,$7E,$53,$B3,$67,$9B
	.dc.b	$6B,$89,$FE,$61,$AE,$33,$E0,$AF,$6E,$DE,$7E,$7B,$3E,$46,$DF,$E3
	.dc.b	$34,$57,$E3,$09,$5F,$21,$DE,$7F,$99,$C7,$FD,$13,$7E,$23,$F2,$F3
	.dc.b	$F1,$7E,$3A,$5F,$C2,$FA,$0B,$5E,$8C,$3E,$65,$61,$E9,$36,$BE,$27
	.dc.b	$D0,$7C,$FF,$A0,$FA,$0D,$8F,$9A,$DC,$3A,$64,$8C,$61,$FE,$5C,$F9
	.dc.b	$AF,$A4,$FA,$7E,$1B,$F1,$91,$8C,$3E,$94,$63,$87,$FA,$9C,$FD,$9C
	.dc.b	$7F,$DD,$CD,$7F,$6C,$D7,$FD,$0E,$F3,$6C,$7E,$1F,$E8,$3A,$5F,$E6
	.dc.b	$27,$FD,$1B,$F5,$8F,$77,$FB,$0B,$67,$F3,$2D,$9F,$E0,$2D,$7C,$7C
	.dc.b	$5E,$62,$4F,$E8,$31,$FF,$80,$0D,$F9,$CF,$59,$1C,$1F,$A6,$F9,$1F
	.dc.b	$E1,$17,$FE,$A0,$6A,$69,$8F,$9C,$FC,$7C,$7F,$14,$6B,$FC,$64,$38
	.dc.b	$7F,$8E,$35,$FF,$56,$FB,$2E,$7F,$F0,$0B,$6F,$F7,$2D,$9F,$E4,$2D
	.dc.b	$7B,$7D,$57,$FB,$8E,$7F,$7E,$D8,$FE,$4E,$D7,$73,$FE,$0E,$F9,$DF
	.dc.b	$65,$7F,$06,$6C,$FF,$2E,$6B,$EB,$6B,$FC,$D7,$E7,$7E,$29,$3C,$2F
	.dc.b	$F0,$CB,$F2,$3E,$A9,$C9,$FE,$1C,$6E,$1F,$C5,$1B,$7F,$CF,$1A,$FC
	.dc.b	$9F,$9E,$FF,$69,$26,$93,$D5,$FF,$13,$FC,$F3,$DA,$A4,$C3,$F9,$0C
	.dc.b	$65,$FE,$47,$F5,$5F,$D3,$1B,$E0,$FF,$8C,$23,$18,$FF,$4E,$6B,$F6
	.dc.b	$FE,$7D,$F3,$FF,$E3,$66,$DF,$F2,$E6,$CF,$F5,$66,$BF,$83,$F0,$9F
	.dc.b	$11,$FE,$3E,$6D,$7F,$36,$6C,$FF,$5E,$6B,$F8,$BF,$9F,$79,$BF,$E4
	.dc.b	$66,$D7,$F3,$E6,$CF,$F6,$66,$BF,$93,$F9,$F7,$CF,$7F,$4B,$9D,$F8
	.dc.b	$7F,$EE,$09,$CD,$FE,$D6,$D5,$A8,$FE,$E9,$3A,$9F,$EE,$6D,$7B,$5F
	.dc.b	$F9,$6A,$31,$1F,$DE,$FE,$1B,$86,$1F,$E0,$1F,$9F,$71,$FF,$5D,$9C
	.dc.b	$E3,$FE,$B8,$9C,$BE,$38,$CD,$AD,$5A,$8F,$E6,$53,$A9,$CD,$F5,$C7
	.dc.b	$FC,$DB,$3E,$4A,$07,$0D,$8F,$F3,$9F,$CF,$FF,$0A,$8F,$E7,$1D,$FF
	.dc.b	$C7,$54,$67,$0F,$F0,$E4,$E6,$31,$97,$FB,$4A,$31,$E1,$7F,$6C,$79
	.dc.b	$7C,$FB,$CF,$FE,$FC,$DA,$FA,$4F,$99,$B1,$A2,$3A,$C6,$9F,$D4,$F1
	.dc.b	$7C,$B7,$F8,$25,$B3,$F7,$67,$E0,$FC,$FF,$C2,$F9,$D7,$AB,$FE,$31
	.dc.b	$BD,$D7,$3C,$39,$1F,$E1,$37,$88,$AE,$2F,$FB,$C2,$34,$AF,$B5,$F5
	.dc.b	$96,$18,$AF,$E5,$15,$5F,$BD,$3A,$7B,$57,$F9,$0A,$75,$87,$9B,$0F
	.dc.b	$0D,$7F,$92,$67,$37,$C6,$73,$3E,$87,$75,$C6,$FA,$4F,$63,$D5,$87
	.dc.b	$8D,$E1,$1E,$98,$C4,$7B,$4E,$07,$A1,$CE,$1E,$26,$C6,$AB,$C0,$FC
	.dc.b	$03,$FC,$6E,$73,$3F,$C7,$33,$87,$07,$F9,$97,$E7,$9F,$38,$F9,$FF
	.dc.b	$F1,$FC,$61,$B1,$99,$FF,$34,$CE,$5B,$F3,$BF,$AB,$6E,$1C,$A3,$11
	.dc.b	$FD,$6E,$51,$88,$F1,$3C,$73,$FA,$39,$FE,$6E,$7E,$D3,$5F,$8B,$5C
	.dc.b	$AF,$EF,$36,$3E,$4C,$EE,$B9,$EB,$FE,$83,$7E,$ED,$6D,$B8,$78,$6B
	.dc.b	$FD,$10,$71,$2D,$71,$7C,$F8,$78,$27,$86,$1F,$C0,$A7,$83,$6B,$83
	.dc.b	$FC,$1B,$0F,$04,$F2,$8F,$3C,$E9,$FD,$4B,$67,$71,$CC,$FE,$59,$3F
	.dc.b	$0E,$FF,$39,$1C,$0F,$E6,$13,$C1,$C1,$FE,$4D,$9F,$F5,$1B,$CB,$BB
	.dc.b	$F6,$DA,$FB,$A3,$F9,$C7,$CF,$FF,$00,$6E,$FA,$7F,$8F,$C6,$5F,$F2
	.dc.b	$C3,$F7,$1D,$9F,$E3,$5C,$5F,$18,$F8,$3E,$51,$87,$A4,$AB,$A4,$FB
	.dc.b	$3E,$FB,$FA,$FC,$FF,$4B,$FB,$AF,$D7,$E5,$FE,$BD,$6C,$7F,$32,$9D
	.dc.b	$BF,$E6,$B6,$D7,$C6,$9D,$C9,$B3,$AA,$FF,$64,$2F,$95,$F2,$CB,$E0
	.dc.b	$69,$E8,$9D,$5F,$F5,$F6,$BF,$E7,$B9,$E0,$FA,$67,$85,$48,$C1,$0F
	.dc.b	$43,$F1,$3E,$3F,$FC,$79,$1F,$07,$3F,$19,$9F,$46,$C7,$F6,$FF,$97
	.dc.b	$DB,$FE,$8B,$6E,$9F,$06,$BE,$7C,$63,$4F,$EB,$32,$FD,$87,$FC,$0B
	.dc.b	$7F,$65,$AF,$3C,$3E,$BF,$DE,$7F,$5F,$8F,$A9,$F1,$7F,$18,$CF,$83
	.dc.b	$F7,$3C,$97,$CA,$FD,$EF,$7D,$EA,$53,$51,$E4,$F9,$0F,$C2,$E8,$F8
	.dc.b	$8A,$67,$3F,$F0,$67,$DB,$7E,$3D,$FF,$54,$75,$7F,$8B,$1F,$F0,$83
	.dc.b	$5F,$F5,$64,$6F,$83,$F4,$FF,$AF,$BE,$33,$BF,$74,$BF,$F0,$C4,$E5
	.dc.b	$F9,$9F,$0B,$E4,$F8,$79,$18,$3F,$C3,$DF,$00,$F9,$3F,$CD,$FE,$4B
	.dc.b	$F3,$67,$FC,$0A,$D5,$3B,$7F,$C4,$DE,$FB,$E3,$7E,$0F,$C6,$C9,$CC
	.dc.b	$7D,$8F,$88,$F8,$AC,$8F,$C0,$47,$EA,$BE,$87,$FD,$71,$1F,$E3,$1F
	.dc.b	$A2,$76,$7E,$B9,$B1,$FE,$35,$9D,$BF,$ED,$18,$FA,$AF,$E0,$7C,$BF
	.dc.b	$A6,$71,$EA,$FC,$4F,$D2,$FD,$6F,$2E,$57,$E4,$63,$F7,$7F,$46,$F9
	.dc.b	$BF,$DE,$FE,$4C,$F1,$A9,$A8,$FE,$03,$F4,$DE,$4C,$3F,$DB,$9D,$FF
	.dc.b	$EA,$1E,$CF,$EF,$07,$E5,$63,$7C,$1F,$9A,$FA,$CF,$E1,$7F,$B0,$BE
	.dc.b	$4A,$99,$AF,$F2,$97,$D3,$3E,$7F,$FD,$8D,$1F,$EA,$D1,$19,$F9,$4C
	.dc.b	$46,$3E,$83,$F9,$3B,$FE,$B3,$9D,$3C,$7F,$19,$FF,$36,$BF,$E4,$20
	.dc.b	$E1,$FF,$22,$B5,$CF,$C4,$FF,$38,$78,$70,$BE,$BB,$11,$FE,$4C,$8C
	.dc.b	$7A,$3E,$46,$C7,$A5,$AF,$F8,$BF,$5F,$F6,$5F,$07,$C8,$6D,$7E,$F9
	.dc.b	$FF,$61,$46,$7E,$6B,$E6,$CE,$5B,$5F,$E6,$5A,$E2,$DF,$4B,$36,$9C
	.dc.b	$BF,$ED,$6F,$8C,$CF,$CB,$7F,$A6,$A3,$FC,$B7,$C1,$89,$99,$7F,$45
	.dc.b	$AD,$3F,$E5,$D9,$CE,$3F,$E7,$58,$F9,$9F,$87,$7F,$0A,$7F,$D1,$8D
	.dc.b	$8D,$1E,$7F,$86,$F4,$A8,$D6,$3F,$D9,$9B,$5D,$3D,$2F,$5E,$7F,$E6
	.dc.b	$EB,$FF,$3D,$D6,$9B,$3D,$23,$17,$F0,$70,$C4,$FC,$0F,$A3,$98,$FB
	.dc.b	$78,$6E,$3E,$6C,$3F,$DF,$75,$87,$F3,$BA,$7F,$AE,$1B,$1E,$4F,$9C
	.dc.b	$F2,$FA,$97,$1F,$ED,$9A,$A3,$18,$F5,$8C,$4E,$AB,$8B,$FE,$B3,$FB
	.dc.b	$07,$D2,$7C,$66,$C7,$D3,$FA,$9C,$4C,$8E,$0F,$DC,$8C,$61,$F4,$05
	.dc.b	$72,$70,$DC,$FD,$B3,$C1,$FC,$A7,$A3,$F4,$3E,$4F,$E8,$B7,$1F,$6B
	.dc.b	$70,$F6,$36,$BE,$3F,$CD,$FE,$38,$FC,$7F,$45,$BF,$C6,$45,$BF,$A1
	.dc.b	$C1,$FC,$46,$72,$7C,$3F,$4F,$E1,$1F,$82,$F9,$FB,$19,$25,$7C,$D8
	.dc.b	$CB,$EE,$9B,$8F,$EB,$E6,$D7,$F6,$CB,$FF,$62,$3F,$9F,$F8,$1F,$D2
	.dc.b	$FB,$DF,$A6,$6C,$64,$FD,$24,$65,$FB,$FE,$87,$E5,$8F,$CB,$FC,$DD
	.dc.b	$8C,$93,$3F,$C1,$46,$5F,$AD,$F6,$6C,$BF,$4A,$DA,$FD,$97,$4E,$26
	.dc.b	$47,$FB,$E6,$75,$5F,$2A,$B2,$F4,$AB,$A7,$E6,$A6,$DF,$F2,$3B,$E3
	.dc.b	$3F,$E5,$66,$C7,$C5,$9D,$DF,$FB,$AA,$72,$DB,$FE,$17,$3A,$84,$E9
	.dc.b	$B3,$AA,$F4,$3B,$9E,$76,$D3,$F7,$97,$C2,$F6,$47,$DC,$D9,$F4,$5F
	.dc.b	$99,$E5,$87,$58,$C5,$F4,$C3,$C8,$48,$F2,$61,$F1,$F0,$E8,$7C,$58
	.dc.b	$CE,$1D,$A3,$33,$F0,$5F,$AA,$61,$F4,$1C,$CF,$39,$63,$93,$63,$E7
	.dc.b	$1C,$6B,$C8,$F4,$8D,$1D,$CB,$3E,$E8,$E1,$5E,$EC,$3E,$04,$79,$A7
	.dc.b	$55,$C8,$F2,$BE,$23,$F0,$91,$C8,$E1,$87,$C7,$BE,$75,$EA,$AE,$15
	.dc.b	$F1,$4D,$DE,$A3,$32,$7C,$26,$1F,$0C,$F1,$7A,$36,$35,$87,$E1,$30
	.dc.b	$F1,$4F,$96,$B6,$7C,$3A,$DE,$1F,$5A,$79,$12,$74,$FC,$E7,$E1,$5F
	.dc.b	$CE,$79,$9E,$43,$64,$AF,$4C,$F3,$8D,$1B,$8C,$BC,$0C,$CF,$A7,$0F
	.dc.b	$89,$5E,$3A,$DD,$6A,$FE,$C4,$FB,$67,$E4,$9A,$BF,$33,$8C,$3E,$1D
	.dc.b	$FA,$AF,$E0,$3E,$1A,$DD,$FB,$EF,$E2,$BF,$58,$EC,$78,$EF,$9C,$6A
	.dc.b	$BE,$1D,$72,$7E,$14,$EA,$BE,$14,$FB,$5E,$73,$AC,$3D,$ED,$8C,$BF
	.dc.b	$1E,$B9,$3E,$7C,$3D,$2F,$C3,$9F,$33,$ED,$6E,$1F,$13,$99,$F1,$5F
	.dc.b	$E0,$43,$FC,$0B,$F1,$96,$BF,$B8,$C5,$BF,$EE,$13,$6F,$FB,$DF,$CD
	.dc.b	$7F,$88,$BE,$53,$FC,$46,$DC,$3F,$C4,$0D,$BF,$EA,$AD,$AF,$F7,$DF
	.dc.b	$CA,$7F,$80,$EB,$FC,$98,$DA,$FF,$1A,$7A,$5F,$E1,$E8,$C4,$1A,$6E
	.dc.b	$BF,$C6,$D3,$BE,$F9,$7F,$EC,$03,$AB,$FE,$22,$8C,$DF,$F6,$8F,$AE
	.dc.b	$FF,$1D,$9F,$E3,$CF,$A8,$B7,$FE,$39,$8B,$3F,$DB,$A6,$E1,$FE,$51
	.dc.b	$F7,$7F,$C5,$BF,$09,$FE,$2E,$6E,$1F,$E1,$A6,$DF,$F8,$65,$B5,$F0
	.dc.b	$BE,$2B,$DC,$7F,$93,$0B,$5E,$9A,$1F,$06,$32,$FF,$80,$1B,$3C,$E7
	.dc.b	$84,$ED,$FF,$29,$4E,$9F,$F3,$37,$97,$FE,$44,$3F,$C8,$BE,$8D,$8F
	.dc.b	$F1,$8C,$5A,$FF,$18,$4D,$BF,$F1,$DF,$9F,$FE,$62,$F4,$3F,$CC,$6D
	.dc.b	$C3,$FC,$C0,$DB,$FF,$0A,$B6,$93,$FE,$8B,$FC,$61,$F4,$DF,$E5,$C6
	.dc.b	$C7,$F9,$0B,$8B,$13,$23,$DB,$3A,$C3,$FC,$89,$59,$7F,$C7,$11,$9B
	.dc.b	$EF,$1F,$2A,$31,$5F,$E8,$7A,$E1,$5F,$E5,$6F,$A7,$FF,$3E,$9F,$E7
	.dc.b	$EF,$84,$B5,$FE,$7B,$8B,$87,$F8,$C2,$6D,$7F,$A4,$FC,$FF,$F3,$7F
	.dc.b	$D9,$7F,$9C,$1B,$87,$F9,$71,$B7,$FE,$5B,$6D,$79,$A1,$F8,$C7,$FA
	.dc.b	$FE,$BE,$DB,$63,$FC,$DD,$F6,$6F,$B2,$FF,$41,$CE,$6F,$93,$6F,$FD
	.dc.b	$70,$DB,$FE,$99,$7C,$67,$AF,$B3,$C8,$8F,$E4,$6D,$3B,$AF,$F3,$25
	.dc.b	$F3,$5F,$51,$4E,$D4,$F8,$F7,$8F,$FE,$98,$9E,$18,$7F,$A9,$2F,$F3
	.dc.b	$AD,$B7,$D9,$FF,$00,$8D,$4B,$E6,$1E,$8F,$FA,$89,$BA,$79,$ED,$DD
	.dc.b	$9A,$FF,$53,$FC,$E3,$CD,$A7,$FA,$FB,$EE,$38,$E5,$FE,$7F,$9D,$63
	.dc.b	$FA,$AF,$B6,$F0,$9E,$5A,$77,$FF,$AC,$A3,$CB,$E8,$FC,$33,$E0,$FD
	.dc.b	$DB,$ED,$FF,$D6,$F1,$F1,$FE,$F1,$F1,$FF,$EB,$A6,$E5,$F3,$FD,$67
	.dc.b	$85,$9A,$FF,$5F,$E0,$F8,$CA,$7F,$B5,$BF,$00,$FC,$DF,$07,$F0,$69
	.dc.b	$D6,$3F,$C0,$BE,$B9,$F1,$1F,$EC,$86,$C7,$DD,$B7,$C7,$66,$BF,$D9
	.dc.b	$BF,$80,$3E,$0E,$9F,$EE,$4F,$83,$7D,$96,$31,$F8,$DC,$1F,$0C,$FF
	.dc.b	$74,$FC,$6F,$E1,$79,$7F,$D2,$9B,$FF,$AA,$FE,$B9,$EB,$FF,$B6,$9B
	.dc.b	$87,$E3,$EF,$FA,$AB,$FC,$6A,$33,$5F,$EB,$3F,$C0,$DF,$07,$FE,$E2
	.dc.b	$6D,$7F,$20,$FC,$0D,$D3,$9A,$FF,$73,$C6,$27,$53,$FE,$EA,$F8,$5F
	.dc.b	$C6,$CE,$70,$FF,$60,$47,$E4,$0F,$CA,$65,$FD,$46,$FF,$96,$C7,$E1
	.dc.b	$3E,$E3,$FD,$E7,$E6,$FF,$56,$D9,$18,$F6,$FF,$0C,$CB,$FD,$67,$D6
	.dc.b	$FE,$43,$C9,$FE,$01,$7C,$27,$3E,$2E,$F4,$FF,$76,$5E,$E7,$FA,$1C
	.dc.b	$2B,$83,$FE,$D7,$FA,$BF,$C8,$4F,$F4,$26,$C7,$FB,$D7,$C9,$7C,$FF
	.dc.b	$F2,$4C,$BF,$88,$56,$9B,$DF,$9C,$3B,$CE,$5F,$EA,$73,$FB,$3A,$D6
	.dc.b	$1F,$D5,$AB,$4D,$BC,$A8,$CD,$71,$BF,$35,$6A,$5F,$4B,$4E,$B5,$FC
	.dc.b	$4B,$B7,$17,$ED,$9B,$87,$F2,$49,$B5,$EC,$F9,$9E,$87,$59,$B7,$F1
	.dc.b	$62,$C7,$C9,$2C,$FF,$33,$1E,$F7,$A9,$B5,$FD,$BA,$7F,$BA,$F9,$5F
	.dc.b	$DD,$BE,$A9,$F4,$3F,$5A,$F7,$5F,$E7,$BB,$C6,$1C,$CF,$8B,$1C,$85
	.dc.b	$F2,$7E,$7C,$69,$61,$DF,$0E,$77,$67,$F0,$A6,$8E,$71,$A6,$9A,$9E
	.dc.b	$0F,$FA,$3E,$BF,$B3,$DF,$86,$3F,$1F,$3A,$C3,$D7,$39,$9F,$43,$B8
	.dc.b	$CD,$F8,$A7,$C5,$39,$BE,$B3,$9B,$E9,$18,$AF,$21,$E6,$3C,$98,$79
	.dc.b	$9E,$F5,$EB,$BF,$14,$76,$8C,$1D,$5F,$11,$F3,$4B,$78,$70,$7E,$B3
	.dc.b	$EA,$9F,$0A,$FA,$BE,$31,$86,$C6,$67,$DD,$39,$6F,$C3,$FF,$AE,$5B
	.dc.b	$A7,$28,$C4,$7D,$69,$FF,$39,$CE,$4F,$A7,$8B,$5F,$DF,$E7,$F4,$E3
	.dc.b	$17,$F6,$2F,$8B,$F1,$DB,$3B,$AE,$35,$FC,$20,$E3,$5B,$C3,$CD,$5F
	.dc.b	$7E,$F8,$D6,$EF,$8D,$FF,$94,$A3,$8D,$6E,$F8,$C7,$C9,$6C,$EE,$B8
	.dc.b	$BF,$A4,$6D,$F6,$8F,$F2,$EB,$5D,$1F,$0A,$37,$3D,$23,$27,$EA,$CE
	.dc.b	$C7,$F9,$82,$BD,$B8,$7F,$8A,$EF,$A4,$63,$4F,$85,$E3,$FE,$D5,$B7
	.dc.b	$C7,$1C,$3F,$E7,$0F,$AB,$FE,$02,$FF,$9B,$1A,$F8,$43,$8C,$7F,$C7
	.dc.b	$4F,$FA,$BA,$38,$6F,$D2,$7F,$C3,$5F,$53,$79,$8F,$82,$B9,$D5,$FA
	.dc.b	$B7,$66,$7E,$AF,$F5,$DA,$7F,$42,$6D,$FF,$A1,$AB,$6F,$FA,$09,$AF
	.dc.b	$9E,$36,$9B,$1A,$CB,$FC,$A4,$FF,$89,$E7,$FD,$17,$3F,$CD,$1B,$CD
	.dc.b	$E5,$F5,$64,$60,$87,$C5,$AE,$4D,$8F,$54,$C8,$F3,$47,$C1,$F5,$1F
	.dc.b	$35,$FB,$56,$CF,$EE,$9A,$FF,$0D,$6B,$FC,$83,$0F,$96,$FF,$00,$6B
	.dc.b	$FC,$4D,$AE,$F5,$FE,$05,$E5,$72,$8F,$ED,$AD,$7F,$D6,$2D,$4D,$FA
	.dc.b	$B7,$CA,$7F,$0B,$6C,$FF,$10,$6B,$FC,$8D,$AF,$F3,$6F,$AB,$7C,$9F
	.dc.b	$F1,$36,$C7,$F1,$86,$BF,$CA,$DA,$FF,$3E,$B7,$A1,$EA,$24,$7F,$20
	.dc.b	$8C,$BF,$C8,$BE,$AB,$F2,$8F,$F3,$56,$BF,$D2,$1A,$FF,$83,$9A,$FF
	.dc.b	$B4,$BE,$AB,$FC,$12,$FF,$2C,$6B,$FC,$FD,$AF,$F5,$26,$BF,$33,$CE
	.dc.b	$F9,$E4,$7C,$CF,$F3,$3F,$8F,$7C,$45,$24,$FF,$3A,$8C,$BF,$CF,$3E
	.dc.b	$A7,$FC,$D6,$DF,$6F,$F9,$E4,$62,$3F,$A9,$35,$F9,$1F,$09,$F8,$02
	.dc.b	$3F,$1D,$FE,$89,$F5,$0F,$96,$A4,$9F,$E9,$71,$97,$FA,$6F,$D4,$7F
	.dc.b	$9D,$5B,$EF,$FF,$4D,$8C,$47,$F6,$36,$BF,$67,$E5,$7F,$22,$47,$EC
	.dc.b	$3F,$D5,$7E,$A1,$F4,$F4,$93,$FD,$72,$32,$FF,$5E,$F6,$3F,$CF,$ED
	.dc.b	$F3,$FF,$AF,$46,$23,$FB,$A3,$5F,$C5,$7C,$AF,$E8,$88,$FE,$25,$FE
	.dc.b	$CD,$F8,$77,$F0,$18,$FE,$F8,$D7,$FC,$1E,$D7,$FC,$AA,$D4,$3E,$9D
	.dc.b	$F2,$EF,$6F,$FB,$AC,$61,$BF,$0F,$FE,$94,$AF,$F0,$54,$E5,$FF,$06
	.dc.b	$7D,$3B,$E9,$BF,$DD,$11,$FE,$64,$8C,$79,$5F,$E0,$06,$C7,$09,$DA
	.dc.b	$F2,$9F,$3A,$FA,$7F,$F2,$7B,$70,$FF,$71,$3F,$C5,$1B,$1A,$23,$AF
	.dc.b	$26,$DF,$F5,$5B,$7C,$FF,$FB,$74,$FF,$6F,$7D,$6F,$C1,$BF,$F7,$74
	.dc.b	$67,$2F,$F2,$FB,$AD,$AC,$7F,$9A,$BE,$A3,$F8,$3D,$7F,$BD,$7E,$A7
	.dc.b	$FC,$F6,$FF,$9D,$0F,$F4,$2C,$6B,$2F,$14,$E9,$FE,$B1,$19,$7F,$C8
	.dc.b	$ED,$78,$F9,$1F,$19,$46,$5F,$B6,$F5,$CD,$7C,$0F,$8F,$F5,$3E,$7B
	.dc.b	$F5,$FC,$6C,$9F,$7F,$A7,$71,$F8,$DA,$FF,$B3,$E7,$E1,$D7,$C2,$AD
	.dc.b	$63,$FE,$97,$E1,$7D,$4F,$D6,$F2,$FE,$BB,$70,$FA,$ED,$CB,$E9,$57
	.dc.b	$09,$FB,$75,$FE,$94,$FA,$87,$D4,$FF,$A3,$BF,$38,$7C,$5E,$63,$0D
	.dc.b	$BF,$F2,$5D,$70,$6C,$FC,$78,$C6,$3F,$E6,$16,$E7,$F1,$A3,$F4,$71
	.dc.b	$8A,$FF,$55,$ED,$FE,$BD,$9C,$DF,$E4,$BC,$CF,$C9,$E3,$6F,$F2,$6D
	.dc.b	$AF,$6B,$71,$F8,$CF,$E2,$BE,$36,$57,$E3,$BD,$0B,$7F,$42,$6D,$FF
	.dc.b	$B0,$F2,$FF,$7B,$B6,$3E,$37,$D5,$7F,$AC,$B9,$BF,$E0,$0E,$3F,$8F
	.dc.b	$C3,$D7,$39,$F9,$2D,$CF,$C6,$8F,$88,$47,$07,$19,$6C,$9F,$D8,$F9
	.dc.b	$7C,$A7,$F0,$BE,$AF,$E9,$3D,$9F,$D2,$B7,$0F,$D4,$B6,$77,$1F,$AC
	.dc.b	$9F,$DD,$F9,$D2,$3F,$77,$F5,$56,$BF,$0B,$37,$6F,$C7,$4D,$AF,$F7
	.dc.b	$7F,$D5,$7F,$43,$7F,$6F,$F2,$1F,$ED,$C8,$C6,$1F,$9A,$8C,$67,$FB
	.dc.b	$A6,$D7,$03,$E0,$57,$0C,$3E,$84,$EA,$F9,$56,$AF,$93,$4E,$51,$FE
	.dc.b	$51,$BE,$51,$F8,$AA,$E5,$7D,$23,$85,$74,$38,$5F,$46,$C7,$49,$FC
	.dc.b	$B4,$2F,$AB,$FC,$9F,$85,$87,$6B,$F0,$1D,$AF,$C1,$5D,$0F,$03,$D2
	.dc.b	$F9,$BF,$C9,$2B,$94,$F5,$6C,$66,$BA,$1D,$A7,$A4,$76,$3A,$61,$EB
	.dc.b	$6C,$FA,$CF,$B3,$19,$69,$99,$FE,$AB,$7E,$4C,$3D,$B3,$E0,$9F,$23
	.dc.b	$6F,$C9,$1E,$28,$F2,$47,$82,$3D,$F1,$D9,$A7,$6B,$F1,$36,$3C,$4F
	.dc.b	$CF,$8C,$B6,$33,$7E,$5C,$3E,$2E,$1E,$6A,$F1,$1E,$62,$E1,$E6,$3C
	.dc.b	$87,$C7,$BF,$23,$63,$C4,$79,$8F,$01,$6F,$C1,$3F,$06,$BC,$15,$D2
	.dc.b	$3E,$09,$D0,$B5,$D0,$FF,$00,$94,$F8,$86,$0E,$E5,$AF,$D5,$B7,$0F
	.dc.b	$EA,$51,$88,$E2,$77,$3D,$35,$FC,$67,$B3,$F1,$71,$EA,$8F,$33,$F9
	.dc.b	$98,$CC,$FE,$72,$35,$7F,$CC,$A7,$37,$F3,$9B,$1F,$05,$B3,$F3,$B0
	.dc.b	$FE,$C5,$3E,$38,$E6,$79,$1E,$87,$E2,$DC,$9A,$C3,$FC,$0B,$5D,$1F
	.dc.b	$39,$89,$E1,$87,$9C,$FC,$B4,$EE,$3E,$14,$EE,$BF,$06,$76,$AD,$CF
	.dc.b	$2C,$3D,$8D,$3D,$85,$9F,$65,$F5,$7D,$47,$C9,$8D,$5F,$9B,$0F,$83
	.dc.b	$3A,$3C,$17,$FE,$1A,$BE,$51,$EF,$9D,$BA,$BF,$0B,$E8,$C3,$DD,$7C
	.dc.b	$67,$B9,$EE,$AE,$33,$A8,$E3,$87,$DE,$BE,$87,$F8,$33,$0F,$E1,$7D
	.dc.b	$2F,$86,$FF,$1E,$9F,$E3,$EF,$88,$B8,$7F,$8F,$A2,$DF,$F8,$22,$6E
	.dc.b	$1F,$D9,$7E,$AB,$FC,$6B,$F2,$BF,$E3,$66,$E1,$FE,$3A,$6D,$FF,$87
	.dc.b	$1B,$5F,$CB,$3E,$9B,$FD,$F7,$5F,$E3,$9F,$82,$FF,$02,$4E,$7C,$1F
	.dc.b	$ED,$D1,$89,$FD,$4D,$70,$C3,$F5,$01,$C3,$FA,$B6,$BC,$2B,$FC,$75
	.dc.b	$1C,$6F,$B5,$7F,$97,$3D,$4E,$D7,$FE,$76,$AD,$4F,$F8,$53,$E7,$7F
	.dc.b	$CA,$67,$F9,$53,$EA,$AC,$7F,$96,$22,$D7,$F8,$EA,$6E,$1F,$E3,$7F
	.dc.b	$73,$FC,$D7,$F0,$DF,$E6,$C6,$E1,$FE,$6C,$6D,$FF,$8D,$9B,$5F,$1F
	.dc.b	$E1,$7F,$CC,$7E,$CF,$9A,$BE,$04,$62,$FF,$C1,$59,$A3,$DE,$48,$F8
	.dc.b	$F3,$C1,$B3,$B7,$8C,$FF,$9C,$0E,$D1,$99,$FF,$24,$FD,$57,$FA,$10
	.dc.b	$FF,$42,$FA,$F6,$FF,$D0,$B1,$67,$FC,$9F,$37,$1F,$F1,$67,$AB,$FE
	.dc.b	$7C,$F5,$BF,$CF,$AD,$C3,$FD,$02,$DB,$FF,$31,$B6,$BF,$B9,$FD,$4B
	.dc.b	$EA,$BF,$C6,$5F,$71,$F0,$E3,$18,$7F,$A0,$E3,$13,$FE,$79,$9C,$E1
	.dc.b	$F8,$86,$BF,$CC,$5A,$FF,$36,$77,$7E,$09,$FF,$51,$E9,$D3,$0F,$F4
	.dc.b	$BF,$D5,$FF,$A5,$8F,$F4,$BF,$C2,$58,$C9,$31,$FF,$3E,$C6,$5F,$F3
	.dc.b	$9F,$AD,$FE,$AE,$FA,$BF,$F5,$7B,$6B,$FD,$5E,$D9,$DC,$7F,$AC,$A7
	.dc.b	$F0,$5F,$52,$FA,$AF,$F5,$3C,$E1,$AF,$F9,$36,$34,$FE,$09,$B5,$FA
	.dc.b	$A6,$BF,$D4,$DA,$FF,$58,$77,$A7,$F8,$91,$B3,$FE,$DC,$87,$D4,$3E
	.dc.b	$AF,$C5,$3F,$E5,$DF,$AA,$E7,$87,$FB,$89,$A9,$A6,$2B,$FD,$AF,$1F
	.dc.b	$D9,$5A,$FF,$6A,$6B,$FE,$EC,$6A,$F9,$27,$D0,$7F,$AE,$5B,$7F,$A2
	.dc.b	$6C,$FF,$0A,$6B,$FD,$CF,$E6,$DE,$77,$FB,$05,$B5,$FA,$A6,$CF,$F1
	.dc.b	$26,$BF,$DE,$FE,$7F,$C6,$4F,$80,$FF,$65,$3E,$6F,$5B,$E0,$93,$83
	.dc.b	$FD,$9C,$FC,$4F,$A7,$7C,$7F,$FB,$45,$B9,$7E,$E9,$B8,$FF,$22,$6B
	.dc.b	$FE,$0B,$F9,$F7,$C7,$7F,$B6,$5B,$5F,$C0,$9B,$3F,$CA,$9A,$FF,$84
	.dc.b	$FE,$9D,$F1,$BF,$EE,$16,$D7,$F0,$A6,$CF,$F3,$26,$BF,$E1,$BF,$A6
	.dc.b	$FF,$6D,$3F,$CD,$9A,$E7,$E1,$7F,$DD,$91,$88,$FF,$12,$EC,$F7,$FE
	.dc.b	$39,$30,$FE,$33,$19,$7F,$8D,$FD,$33,$D1,$FF,$7B,$36,$33,$18,$3F
	.dc.b	$DF,$5F,$4C,$F8,$3A,$49,$FE,$4B,$19,$7F,$93,$FD,$33,$8F,$FD,$07
	.dc.b	$9F,$F8,$EA,$31,$3A,$9C,$C8,$4E,$9F,$EA,$FF,$09,$F9,$62,$73,$70
	.dc.b	$8F,$F7,$7B,$7C,$1F,$EA,$31,$CD,$AB,$5F,$F2,$8E,$3F,$E4,$C9,$CC
	.dc.b	$7F,$98,$E7,$F2,$5E,$EF,$E4,$09,$5F,$DA,$3C,$07,$D3,$7F,$BB,$EB
	.dc.b	$A5,$7F,$62,$9C,$F7,$7F,$4F,$C3,$BD,$7B,$7E,$95,$D1,$FD,$4D,$B9
	.dc.b	$7F,$B8,$F0,$FD,$97,$D3,$66,$A4,$FF,$B7,$EB,$76,$BF,$F0,$0F,$80
	.dc.b	$C3,$73,$9B,$FF,$34,$DF,$0B,$FF,$5A,$B6,$3F,$89,$5F,$17,$B6,$1E
	.dc.b	$5F,$62,$47,$97,$C2,$FC,$CE,$1C,$A7,$38,$F7,$3D,$CF,$C3,$9D,$D7
	.dc.b	$8A,$3E,$07,$D0,$C8,$F6,$55,$39,$0C,$93,$9C,$3D,$33,$6B,$FD,$0B
	.dc.b	$0F,$57,$FB,$F3,$63,$F9,$71,$FC,$3E,$5F,$4A,$FA,$4E,$37,$C0,$95
	.dc.b	$C2,$BF,$8E,$1C,$27,$71,$98,$FF,$4A,$F8,$BF,$99,$BE,$91,$FC,$F6
	.dc.b	$D5,$9B,$F1,$56,$AB,$ED,$56,$9E,$C7,$82,$35,$87,$AE,$32,$78,$0A
	.dc.b	$E6,$3A,$47,$EA,$F0,$F0,$3E,$3C,$2B,$C6,$7A,$DF,$85,$E3,$0A,$DD
	.dc.b	$F5,$C3,$D2,$7C,$29,$F1,$9D,$CF,$1D,$F3,$8D,$C7,$84,$D9,$D4,$FF
	.dc.b	$16,$94,$FE,$8C,$62,$BF,$C6,$3F,$4A,$F7,$9F,$4D,$86,$D7,$F9,$72
	.dc.b	$75,$39,$38,$FC,$07,$E9,$E7,$4D,$7F,$D9,$53,$FE,$7D,$CB,$64,$8F
	.dc.b	$9F,$8F,$E4,$9F,$92,$D7,$E9,$CF,$16,$D7,$7C,$FE,$F5,$6D,$BA,$FF
	.dc.b	$B7,$1F,$C1,$5F,$16,$D7,$1A,$FF,$70,$9C,$4B,$5C,$5F,$F7,$2D,$71
	.dc.b	$6C,$F1,$BF,$E5,$91,$C5,$B5,$C5,$FE,$61,$5C,$5B,$5C,$5F,$D3,$4E
	.dc.b	$DA,$7E,$2A,$78,$38,$3F,$73,$5F,$17,$0F,$CE,$46,$5B,$3C,$27,$5A
	.dc.b	$FB,$BE,$9B,$F7,$6D,$BF,$D7,$4E,$9F,$E8,$DF,$4D,$FB,$F7,$F6,$4D
	.dc.b	$7F,$84,$B7,$0F,$0B,$FE,$B9,$8D,$EF,$D9,$FE,$55,$C8,$FA,$4C,$5F
	.dc.b	$F8,$8D,$D4,$AF,$A6,$3F,$D5,$E1,$F4,$CF,$83,$DF,$D3,$7E,$E2,$B4
	.dc.b	$DA,$FE,$BD,$5B,$7F,$85,$35,$FE,$43,$1B,$4D,$8D,$63,$F3,$1F,$F1
	.dc.b	$7C,$FF,$13,$97,$37,$88,$F4,$CE,$A7,$F9,$6B,$6B,$EA,$D7,$09,$E5
	.dc.b	$5C,$3D,$09,$18,$21,$C9,$F8,$BE,$FB,$E6,$69,$27,$F9,$37,$E1,$9D
	.dc.b	$5F,$6E,$DF,$18,$8F,$F2,$BA,$E4,$FF,$2D,$F2,$EB,$FD,$11,$B9,$FE
	.dc.b	$31,$BA,$FC,$EB,$FA,$FF,$4E,$F9,$0A,$4C,$3F,$9C,$47,$CA,$AF,$B5
	.dc.b	$F4,$EF,$8C,$FB,$ED,$AF,$E7,$CD,$7E,$AF,$A7,$F8,$4C,$9F,$31,$49
	.dc.b	$3F,$D1,$7E,$B3,$FC,$96,$FF,$47,$FA,$8F,$C9,$3F,$D7,$9B,$B7,$E9
	.dc.b	$9B,$9F,$E3,$7A,$7C,$D5,$FE,$25,$FA,$8F,$EF,$CF,$F5,$16,$BF,$DA
	.dc.b	$1B,$7F,$D2,$C7,$CC,$BE,$EF,$D1,$EA,$4B,$F9,$CF,$E8,$25,$F3,$7F
	.dc.b	$A0,$F8,$E7,$C2,$52,$5F,$F6,$1F,$B0,$7D,$17,$E9,$BE,$AB,$FC,$E3
	.dc.b	$3F,$D9,$5A,$FF,$7C,$6E,$9E,$5C,$BF,$CC,$F3,$F7,$FE,$6F,$E6,$11
	.dc.b	$FF,$12,$47,$CA,$7F,$C0,$BF,$54,$F9,$BF,$DA,$B6,$3F,$BA,$35,$FF
	.dc.b	$05,$FD,$53,$E9,$7F,$76,$F3,$C7,$F8,$AB,$79,$3F,$78,$DA,$FF,$17
	.dc.b	$63,$C5,$FC,$A7,$CC,$FE,$04,$8F,$9B,$D5,$FE,$0F,$F5,$4F,$98,$A4
	.dc.b	$9F,$F0,$47,$D7,$FE,$C7,$2F,$E1,$5F,$57,$F0,$A7,$FC,$42,$DD,$3F
	.dc.b	$94,$B6,$7F,$D9,$19,$FF,$17,$AF,$D3,$7C,$CF,$E4,$C8,$FC,$8F,$95
	.dc.b	$FE,$33,$F5,$AF,$9D,$FE,$DE,$D9,$FE,$34,$7F,$1B,$70,$E0,$38,$77
	.dc.b	$F5,$8F,$9A,$FE,$F2,$D8,$E0,$1C,$3C,$DA,$F8,$3E,$B1,$F3,$AF,$87
	.dc.b	$E3,$18,$6F,$CE,$BD,$AC,$D6,$EF,$D3,$C2,$FA,$1F,$C9,$F0,$FF,$9A
	.dc.b	$DB,$3F,$E5,$59,$D7,$A7,$FE,$77,$9C,$C7,$E9,$DB,$5F,$EF,$56,$D7
	.dc.b	$F9,$BD,$FE,$3D,$1D,$7B,$FF,$D0,$1F,$46,$F8,$BA,$7F,$A0,$FB,$7D
	.dc.b	$95,$FE,$7D,$9D,$36,$76,$FF,$A0,$67,$78,$79,$EB,$6F,$F9,$93,$EA
	.dc.b	$FF,$8B,$57,$E9,$3E,$23,$F7,$CF,$F5,$38,$CC,$FB,$FA,$3F,$CF,$8F
	.dc.b	$1F,$4B,$FD,$2D,$18,$D7,$FB,$AD,$70,$7F,$69,$87,$4B,$FF,$5A,$4F
	.dc.b	$47,$F0,$ED,$7C,$17,$FE,$B3,$8C,$47,$FA,$57,$1F,$F6,$DD,$FF,$13
	.dc.b	$6E,$3F,$EB,$F8,$47,$FA,$E1,$BF,$29,$F3,$09,$3F,$27,$DD,$B7,$F6
	.dc.b	$F7,$7D,$57,$DB,$F4,$BE,$E3,$6F,$F0,$4D,$AE,$0D,$C7,$FC,$2B,$F5
	.dc.b	$1F,$C8,$2B,$FD,$65,$F5,$4F,$91,$F8,$58,$2B,$4F,$E2,$E7,$55,$FE
	.dc.b	$83,$9C,$DF,$F3,$19,$CD,$7F,$A7,$27,$35,$F0,$67,$35,$E2,$F3,$FF
	.dc.b	$D9,$51,$89,$FF,$07,$57,$0B,$FF,$60,$4E,$5F,$CA,$7A,$1F,$9B,$C7
	.dc.b	$F3,$8D,$BF,$D1,$36,$7E,$43,$71,$FA,$9F,$27,$F9,$A3,$F3,$7E,$BD
	.dc.b	$BF,$CC,$45,$BF,$CA,$4D,$AF,$93,$F5,$3F,$8C,$FA,$AF,$8F,$A2,$31
	.dc.b	$8F,$C5,$6E,$1F,$0D,$B7,$CA,$7F,$9F,$DF,$29,$CC,$7E,$1F,$8F,$FB
	.dc.b	$E3,$FA,$BF,$81,$FD,$8F,$BD,$FB,$26,$C6,$4F,$D8,$46,$5F,$E0,$9E
	.dc.b	$97,$F0,$13,$F8,$17,$D5,$58,$C9,$33,$FC,$74,$65,$FF,$6A,$FD,$57
	.dc.b	$F5,$19,$FD,$E7,$CA,$FE,$1A,$31,$3F,$B9,$8C,$7A,$BF,$C1,$9B,$5C
	.dc.b	$0F,$F6,$55,$70,$DB,$F4,$D3,$AB,$E5,$5A,$BE,$4D,$39,$47,$92,$32
	.dc.b	$D8,$CD,$7F,$9C,$AF,$B4,$7C,$BB,$ED,$3E,$08,$E9,$5E,$03,$91,$63
	.dc.b	$95,$F6,$6C,$F6,$3F,$CC,$97,$DA,$BF,$81,$5F,$69,$F0,$1D,$AB,$C1
	.dc.b	$3D,$9A,$74,$7C,$11,$D2,$FC,$0D,$3C,$10,$17,$E1,$AF,$E1,$B7,$E1
	.dc.b	$9F,$1D,$78,$4F,$1B,$E1,$6C,$F8,$4F,$1B,$DE,$3C,$67,$7A,$ED,$A2
	.dc.b	$FB,$5F,$82,$FB,$5F,$80,$EC,$58,$ED,$1E,$03,$B3,$E0,$8E,$D8,$7C
	.dc.b	$06,$CF,$C0,$7D,$F1,$96,$C6,$70,$F2,$DF,$93,$0F,$34,$F9,$30,$F3
	.dc.b	$5F,$91,$A7,$91,$F3,$47,$90,$F3,$3E,$23,$E1,$9E,$02,$DF,$82,$3B
	.dc.b	$57,$82,$7B,$36,$BB,$1E,$09,$ED,$3E,$03,$B4,$7C,$37,$B3,$6B,$B4
	.dc.b	$74,$BF,$71,$2B,$C2,$D8,$D4,$60,$8E,$09,$5E,$4B,$F3,$4F,$92,$3C
	.dc.b	$D3,$E4,$8F,$86,$F9,$1B,$7E,$49,$E5,$7E,$16,$CF,$81,$B5,$E0,$8E
	.dc.b	$D7,$E0,$8E,$C7,$C3,$3B,$16,$FB,$4F,$0B,$EC,$D3,$B1,$67,$B1,$D2
	.dc.b	$FD,$6F,$46,$D7,$48,$D5,$F8,$DD,$36,$35,$18,$BF,$0B,$C1,$AE,$6F
	.dc.b	$C6,$E5,$B7,$FE,$39,$F8,$DF,$F1,$A1,$FE,$35,$F9,$0B,$3F,$E3,$B8
	.dc.b	$B5,$FE,$01,$9B,$5F,$E0,$DF,$AA,$FF,$19,$FC,$D7,$F8,$D1,$B8,$7F
	.dc.b	$8C,$5B,$7F,$E2,$D6,$D2,$7F,$DF,$9F,$4F,$FE,$13,$F0,$FF,$DD,$71
	.dc.b	$FE,$17,$33,$DF,$FE,$2A,$F8,$AF,$F0,$2D,$6D,$B1,$99,$E1,$8F,$F1
	.dc.b	$5A,$D4,$62,$11,$87,$FC,$81,$E8,$FF,$93,$0F,$F2,$6F,$D5,$5A,$FF
	.dc.b	$25,$45,$9F,$F2,$34,$DC,$BF,$CB,$5F,$01,$FE,$5D,$F8,$3F,$F2,$F3
	.dc.b	$70,$FF,$32,$36,$FF,$C5,$4D,$AF,$85,$1E,$7F,$8A,$FF,$2E,$FB,$5E
	.dc.b	$E6,$DF,$A7,$34,$62,$97,$C9,$F6,$C6,$63,$13,$C2,$BF,$D0,$ED,$AF
	.dc.b	$F4,$83,$5F,$F4,$AB,$5F,$F3,$A7,$D5,$7F,$9E,$4F,$F3,$D7,$B3,$6B
	.dc.b	$FD,$01,$16,$7F,$C8,$D3,$6F,$FC,$A5,$E9,$FF,$9E,$3D,$2F,$F3,$CB
	.dc.b	$70,$FF,$3B,$36,$FF,$CE,$4D,$AF,$67,$D5,$3D,$AF,$F3,$5F,$B3,$FE
	.dc.b	$46,$8C,$61,$FE,$D8,$8C,$4E,$A3,$F0,$13,$AC,$3F,$CE,$D5,$A9,$FF
	.dc.b	$54,$CE,$67,$FD,$6A,$D7,$FA,$F0,$70,$FF,$9B,$FE,$AB,$FD,$26,$7F
	.dc.b	$A5,$3E,$16,$D7,$FA,$46,$2C,$FF,$A3,$66,$E3,$FE,$A1,$F6,$7F,$D4
	.dc.b	$9F,$55,$FE,$A5,$6E,$1F,$EA,$B6,$DF,$F9,$C5,$B5,$F7,$FE,$A9,$F5
	.dc.b	$1F,$EA,$4B,$FF,$34,$C6,$1F,$F5,$11,$88,$D3,$FA,$38,$D5,$FF,$A8
	.dc.b	$63,$46,$67,$FD,$29,$4F,$9D,$7D,$56,$AF,$F2,$37,$F1,$BE,$AB,$9D
	.dc.b	$FE,$70,$7C,$7B,$C8,$FF,$3E,$CE,$AB,$FD,$29,$39,$7F,$6D,$8F,$FA
	.dc.b	$CA,$F9,$7D,$7D,$9F,$A7,$37,$1F,$D0,$CD,$5F,$32,$FA,$2F,$4C,$BB
	.dc.b	$FF,$D1,$B3,$9C,$3F,$83,$E2,$AD,$47,$F0,$7F,$A0,$79,$FF,$0B,$EB
	.dc.b	$FF,$D5,$79,$FF,$88,$67,$35,$CA,$73,$2F,$A3,$F3,$62,$97,$83,$FE
	.dc.b	$9E,$9C,$C6,$2F,$F8,$B7,$AD,$F1,$70,$FD,$EF,$C4,$BE,$0F,$FD,$4B
	.dc.b	$39,$C9,$39,$BF,$E3,$DF,$56,$FA,$CF,$B3,$ED,$7F,$C0,$09,$CC,$63
	.dc.b	$C3,$FF,$57,$FC,$7B,$EB,$3E,$FF,$D7,$BE,$77,$31,$87,$FD,$6B,$F5
	.dc.b	$AF,$9C,$DF,$D2,$E1,$B9,$FF,$2F,$FA,$DF,$8F,$C0,$F8,$8F,$F7,$14
	.dc.b	$FE,$B2,$73,$18,$E4,$FF,$26,$73,$38,$7F,$23,$3F,$E4,$FB,$FE,$75
	.dc.b	$39,$7E,$CC,$66,$BF,$26,$7F,$95,$7E,$B5,$EC,$FE,$6B,$E5,$5E,$DF
	.dc.b	$FB,$3A,$73,$87,$E0,$A7,$2F,$F4,$7F,$AD,$7C,$3F,$FB,$53,$F0,$77
	.dc.b	$D8,$FF,$B5,$A7,$35,$FD,$36,$72,$FF,$4F,$FA,$F7,$6E,$F0,$FF,$6E
	.dc.b	$C6,$27,$B4,$7F,$B8,$3E,$2F,$F3,$53,$9D,$BD,$67,$FB,$97,$E5,$38
	.dc.b	$E1,$FE,$E6,$F4,$1E,$FD,$3F,$19,$F0,$97,$3F,$8E,$DC,$39,$F2,$7F
	.dc.b	$83,$6F,$8D,$68,$93,$F2,$B9,$DF,$5F,$FA,$78,$E3,$87,$FB,$DB,$EB
	.dc.b	$DE,$05,$3F,$83,$F9,$7F,$6E,$93,$AC,$7F,$B9,$36,$3D,$AD,$C7,$F8
	.dc.b	$45,$71,$8C,$57,$F3,$26,$D7,$FA,$45,$B5,$FC,$03,$E5,$BC,$55,$DB
	.dc.b	$E1,$3C,$C7,$F1,$AF,$43,$F3,$11,$92,$61,$E2,$9D,$77,$F0,$9F,$3F
	.dc.b	$81,$EF,$3E,$05,$58,$C9,$1F,$59,$93,$EC,$E2,$F8,$3F,$E4,$93,$FD
	.dc.b	$32,$30,$FF,$85,$67,$F8,$60,$FA,$E7,$D6,$EE,$F8,$CF,$FA,$C6,$F8
	.dc.b	$BD,$A3,$2F,$F5,$8F,$3F,$FD,$6D,$7C,$A3,$84,$E9,$EF,$19,$7F,$C1
	.dc.b	$71,$FE,$14,$8C,$46,$AF,$C5,$87,$47,$FB,$2C,$64,$EB,$19,$73,$75
	.dc.b	$EB,$15,$E0,$60,$F8,$13,$D1,$4F,$1C,$3D,$F7,$E4,$9C,$9C,$0F,$09
	.dc.b	$4F,$84,$58,$E3,$5D,$6B,$A6,$1E,$F8,$C5,$6E,$30,$FA,$8F,$21,$F5
	.dc.b	$8B,$38,$70,$7F,$82,$3E,$BB,$F8,$7F,$D7,$F0,$6C,$7B,$5A,$FC,$A6
	.dc.b	$BC,$7E,$2F,$FC,$7A,$ED,$AE,$DA,$EF,$1E,$D1,$C3,$2F,$B9,$93,$4F
	.dc.b	$F0,$DC,$61,$B1,$C6,$FE,$46,$0B,$E5,$3F,$E2,$23,$94,$FD,$B6,$BC
	.dc.b	$23,$68,$4F,$3C,$38,$5F,$82,$75,$7C,$B0,$FC,$95,$EE,$7A,$5F,$06
	.dc.b	$18,$24,$FA,$DC,$BC,$B0,$F1,$C7,$EA,$B0,$FD,$86,$9E,$B6,$C7,$89
	.dc.b	$AF,$A2,$78,$E3,$EE,$C3,$FC,$DC,$BE,$77,$6D,$AF,$F4,$73,$BC,$3F
	.dc.b	$C9,$3F,$5F,$FC,$06,$7F,$D2,$4D,$7F,$9E,$B6,$FF,$CF,$4F,$D1,$8D
	.dc.b	$ED,$E6,$9F,$F0,$EF,$97,$79,$7F,$BB,$56,$8F,$F4,$F7,$D7,$FF,$44
	.dc.b	$7F,$D4,$1F,$5F,$FB,$DB,$FF,$43,$37,$2F,$F2,$ED,$6E,$BF,$A7,$B6
	.dc.b	$B7,$9F,$C4,$BE,$15,$F7,$6B,$57,$CB,$D7,$D7,$2C,$AF,$1C,$7C,$CA
	.dc.b	$FF,$15,$C7,$92,$74,$F8,$27,$46,$31,$F3,$CE,$27,$FD,$75,$F1,$6F
	.dc.b	$A0,$FA,$ED,$8C,$C7,$DA,$FA,$F5,$7A,$76,$95,$F1,$1E,$35,$F8,$0F
	.dc.b	$3F,$94,$FF,$70,$6E,$3F,$EA,$D6,$E9,$FE,$84,$D7,$E8,$D7,$F7,$6F
	.dc.b	$AF,$7C,$BF,$E2,$9B,$1F,$ED,$16,$BB,$F1,$3F,$1F,$F5,$EF,$A2,$A4
	.dc.b	$9F,$B3,$DD,$E5,$F4,$E4,$7F,$B6,$F4,$FF,$16,$3F,$ED,$DF,$82,$FF
	.dc.b	$1D,$3F,$E5,$CF,$D8,$9E,$47,$E2,$5B,$19,$8C,$7C,$4F,$E3,$7E,$13
	.dc.b	$ED,$CF,$FB,$31,$BC,$5F,$E6,$BD,$B5,$E6,$EE,$65,$7F,$9A,$9F,$F3
	.dc.b	$83,$7E,$8B,$A7,$9D,$D7,$59,$3B,$8F,$07,$1F,$87,$AE,$4F,$38,$F2
	.dc.b	$71,$79,$7C,$C9,$3D,$E3,$D1,$F2,$F2,$3C,$4F,$AF,$D9,$F6,$7C,$B4
	.dc.b	$9F,$33,$EF,$F8,$7F,$81,$E0,$49,$F5,$3F,$0F,$E1,$BE,$27,$C9,$7C
	.dc.b	$5F,$82,$95,$EF,$8F,$91,$F1,$FF,$27,$E3,$E4,$FC,$38,$FD,$C7,$CB
	.dc.b	$FF,$AA,$23,$E3,$78,$5F,$3C,$CF,$C7,$7F,$10,$F9,$2D,$4F,$F9,$01
	.dc.b	$BD,$7F,$EF,$D6,$D7,$FB,$53,$F3,$EF,$F8,$01,$39,$BE,$44,$BF,$F0
	.dc.b	$35,$6C,$C3,$D6,$2C,$A7,$53,$C0,$FE,$75,$F4,$CD,$7F,$13,$1F,$E0
	.dc.b	$08,$C7,$7F,$F6,$45,$B7,$F2,$39,$D5,$70,$7F,$DD,$7F,$66,$EE,$F0
	.dc.b	$47,$F7,$F8,$C6,$3F,$DB,$9B,$19,$9D,$BF,$E2,$9F,$A5,$7D,$3F,$FB
	.dc.b	$59,$B7,$FE,$BC,$7F,$83,$36,$34,$47,$5E,$55,$AF,$F3,$07,$CF,$7F
	.dc.b	$AE,$0F,$F5,$CF,$C8,$FE,$AA,$BF,$D8,$F1,$9C,$7F,$DC,$AE,$B6,$B1
	.dc.b	$27,$6F,$FB,$AA,$B9,$5F,$68,$FF,$1B,$FD,$3F,$F0,$1A,$FF,$06,$B6
	.dc.b	$7F,$C4,$4D,$7F,$C5,$ED,$7F,$B7,$07,$87,$07,$C6,$6C,$FF,$91,$F4
	.dc.b	$7D,$9F,$83,$BB,$D3,$E2,$4A,$F6,$F8,$D7,$2F,$87,$37,$2F,$F7,$E6
	.dc.b	$DF,$E8,$89,$CD,$FC,$FF,$A8,$FA,$5E,$5F,$D3,$6C,$64,$FA,$91,$96
	.dc.b	$99,$9F,$F4,$86,$9F,$EC,$78,$C6,$1F,$86,$11,$99,$FF,$52,$FD,$33
	.dc.b	$EA,$3F,$0B,$1F,$12,$31,$7F,$E3,$86,$C7,$F9,$35,$AF,$F9,$69,$CD
	.dc.b	$FE,$39,$B3,$FE,$6F,$CB,$FD,$45,$39,$BF,$C8,$7C,$4F,$E4,$7E,$26
	.dc.b	$DF,$E4,$5B,$8F,$B1,$B7,$EB,$C9,$1F,$8A,$F2,$A5,$FE,$3B,$CB,$B8
	.dc.b	$7E,$52,$6E,$1F,$4A,$6D,$79,$66,$D2,$99,$FF,$B9,$E3,$95,$7E,$F5
	.dc.b	$68,$FA,$6F,$F4,$4B,$67,$F5,$5F,$4B,$89,$91,$C6,$75,$5F,$62,$B2
	.dc.b	$FE,$AB,$CC,$FD,$2F,$9B,$FA,$66,$DF,$EA,$9B,$3B,$8F,$D6,$CF,$EE
	.dc.b	$FE,$9A,$47,$EE,$FE,$76,$D7,$E1,$A6,$E9,$F2,$E6,$D7,$E7,$E6,$D7
	.dc.b	$C3,$E0,$56,$2F,$F1,$75,$67,$37,$FC,$26,$30,$7F,$0D,$AB,$19,$3C
	.dc.b	$85,$BF,$23,$6F,$C8,$D3,$C8,$58,$F2,$4F,$47,$89,$FA,$27,$13,$F8
	.dc.b	$7A,$F1,$57,$18,$F4,$B6,$FD,$37,$CA,$78,$4E,$6F,$9B,$E2,$9E,$CF
	.dc.b	$1B,$F8,$53,$A9,$F9,$11,$E0,$7A,$4F,$1A,$EB,$1C,$8C,$9C,$23,$C0
	.dc.b	$75,$2B,$A8,$F4,$B6,$FD,$31,$D0,$93,$DC,$FE,$35,$7E,$58,$E2,$7C
	.dc.b	$EA,$E5,$7E,$AA,$F0,$47,$5A,$E1,$6A,$F6,$5F,$6A,$F4,$1B,$BE,$31
	.dc.b	$93,$F8,$8D,$F2,$AE,$33,$E5,$AE,$35,$BB,$F5,$CE,$63,$CE,$D8,$CC
	.dc.b	$F0,$8C,$57,$48,$C1,$2B,$8B,$F8,$89,$F2,$9C,$EF,$C8,$FE,$52,$3E
	.dc.b	$1B,$F4,$27,$C9,$5F,$16,$72,$7B,$CB,$1E,$9A,$E7,$1F,$76,$38,$D7
	.dc.b	$33,$CE,$58,$F3,$5F,$CD,$8C,$4F,$A0,$91,$99,$E9,$3C,$9F,$14,$F6
	.dc.b	$6C,$66,$3C,$8F,$C0,$8E,$F5,$E6,$9F,$C1,$1C,$4D,$C7,$B2,$7E,$31
	.dc.b	$C6,$3C,$73,$FC,$1A,$5F,$4C,$CF,$FA,$85,$FF,$6E,$8C,$3F,$E0,$2E
	.dc.b	$8C,$D7,$FB,$7E,$72,$FF,$31,$9E,$15,$A8,$EA,$FF,$82,$2B,$15,$C3
	.dc.b	$0F,$E8,$9F,$05,$FE,$1B,$3F,$C3,$9F,$29,$71,$FF,$0E,$45,$C7,$FB
	.dc.b	$3C,$DA,$FE,$EB,$E8,$FF,$85,$3D,$1F,$F0,$AB,$70,$FF,$0C,$B6,$FF
	.dc.b	$C3,$ED,$AE,$33,$FE,$1E,$AE,$2D,$3F,$A4,$7D,$2F,$F8,$06,$FF,$C2
	.dc.b	$21,$C3,$FE,$14,$6A,$35,$DB,$FD,$D9,$C4,$7F,$82,$EF,$FC,$7A,$DA
	.dc.b	$FF,$0C,$7D,$67,$C9,$8C,$57,$F5,$DE,$67,$A7,$FE,$32,$3F,$C6,$7E
	.dc.b	$9D,$C3,$FC,$6F,$16,$BF,$C7,$B3,$71,$FE,$E5,$F4,$FF,$E4,$9F,$84
	.dc.b	$FF,$25,$37,$0F,$F2,$53,$6F,$FC,$2A,$DA,$E3,$3F,$E4,$FA,$E2,$D3
	.dc.b	$E3,$FA,$BE,$FB,$F8,$CD,$CC,$D2,$4F,$03,$FC,$9E,$FC,$13,$FC,$BD
	.dc.b	$F4,$F2,$3F,$CB,$DF,$4F,$6F,$FC,$53,$37,$1F,$F0,$74,$DA,$FF,$0D
	.dc.b	$79,$FF,$E5,$8F,$3F,$FC,$B2,$D8,$FF,$2E,$B7,$0F,$F3,$1B,$6F,$FC
	.dc.b	$82,$D9,$E6,$7F,$99,$2B,$F3,$DF,$4C,$FA,$7F,$80,$D9,$FC,$3E,$1F
	.dc.b	$83,$8C,$38,$6B,$F8,$C7,$38,$7F,$A0,$1B,$3F,$E5,$BD,$3B,$DF,$F9
	.dc.b	$CB,$D7,$93,$FE,$76,$F6,$2E,$3F,$E7,$C9,$B8,$FF,$96,$A6,$D7,$F9
	.dc.b	$97,$D0,$FF,$45,$7B,$9F,$E8,$B6,$C7,$FA,$2D,$B8,$7F,$95,$9B,$7F
	.dc.b	$E3,$B6,$CF,$33,$FD,$25,$5F,$C1,$A5,$EF,$7E,$FE,$7F,$49,$18,$F2
	.dc.b	$B5,$39,$C3,$FC,$74,$A3,$38,$6B,$0B,$87,$FA,$DF,$0E,$B1,$DF,$E9
	.dc.b	$1D,$97,$B3,$0D,$BE,$1F,$49,$FA,$CC,$0D,$3F,$D6,$2F,$F9,$C5,$C7
	.dc.b	$57,$AD,$FD,$8C,$F2,$EA,$C4,$FB,$FE,$91,$F4,$34,$C7,$C1,$DB,$F4
	.dc.b	$79,$DF,$A2,$6F,$C0,$7C,$78,$C7,$1F,$F2,$B3,$FC,$B8,$A3,$E6,$7D
	.dc.b	$13,$E6,$2F,$B5,$86,$E1,$F1,$65,$F4,$7C,$3E,$7F,$8B,$67,$E5,$7A
	.dc.b	$FF,$8E,$CF,$FC,$D4,$DE,$22,$F8,$BF,$B8,$E4,$7C,$96,$0C,$FC,$9D
	.dc.b	$C3,$EB,$7D,$07,$F6,$1C,$3F,$D3,$8D,$8F,$C1,$BF,$E0,$48,$39,$C7
	.dc.b	$F9,$57,$D0,$7E,$0E,$BF,$CF,$5F,$40,$D7,$FA,$74,$E6,$BE,$40,$9D
	.dc.b	$57,$FA,$07,$E8,$1E,$E7,$0F,$03,$FA,$B4,$62,$7F,$D0,$D8,$73,$CF
	.dc.b	$FD,$E3,$EE,$FE,$12,$7F,$DA,$35,$BC,$BB,$CE,$5F,$E8,$11,$9A,$FF
	.dc.b	$7C,$F8,$3F,$CE,$A4,$DF,$A0,$7C,$6F,$F6,$78,$C3,$7D,$3F,$ED,$53
	.dc.b	$AC,$BF,$D3,$55,$A7,$FD,$39,$19,$BF,$C4,$60,$FA,$0F,$F5,$CD,$7F
	.dc.b	$BA,$2B,$52,$8F,$F7,$5F,$D0,$7E,$2D,$FF,$53,$36,$F9,$FC,$CB,$E8
	.dc.b	$3F,$BE,$4E,$6F,$95,$7F,$5F,$8C,$5F,$4E,$BF,$E0,$58,$71,$8E,$B1
	.dc.b	$C8,$8F,$EE,$E3,$55,$F6,$AB,$9B,$D7,$E5,$FC,$27,$F0,$DF,$A0,$78
	.dc.b	$38,$C3,$D3,$36,$BE,$1D,$6A,$72,$7B,$BA,$7D,$67,$B2,$6C,$64,$8F
	.dc.b	$8C,$C9,$F8,$EC,$DE,$B7,$F8,$7D,$B3,$FE,$CF,$9F,$F0,$17,$D0,$7F
	.dc.b	$80,$FE,$79,$F3,$BF,$E0,$59,$FF,$6C,$61,$C4,$97,$C6,$17,$C6,$B8
	.dc.b	$61,$DA,$7F,$3E,$76,$18,$76,$7F,$97,$E1,$F5,$A3,$11,$E1,$BE,$8E
	.dc.b	$B0,$AE,$4E,$78,$78,$2F,$93,$FD,$72,$B7,$87,$8F,$0E,$B7,$E3,$7B
	.dc.b	$C6,$6B,$84,$F4,$3C,$05,$3C,$13,$C2,$FB,$C7,$F8,$C2,$FB,$C6,$23
	.dc.b	$75,$E5,$C3,$D3,$1E,$18,$C3,$D4,$F8,$C7,$F7,$43,$E6,$94,$A6,$6B
	.dc.b	$5F,$3E,$F7,$9F,$41,$86,$D7,$F7,$E9,$D4,$E4,$E3,$F0,$0D,$29,$2B
	.dc.b	$84,$65,$B1,$F4,$B1,$6B,$F8,$7D,$7E,$CC,$62,$BE,$BE,$2C,$39,$57
	.dc.b	$DE,$8C,$61,$CF,$0E,$58,$7E,$44,$E5,$87,$D1,$AD,$B7,$0E,$73,$BA
	.dc.b	$E3,$3F,$E5,$D8,$E3,$87,$7A,$DC,$FF,$98,$6F,$9C,$FA,$2F,$8C,$FE
	.dc.b	$89,$B3,$AB,$FD,$23,$67,$B4,$F0,$9E,$E6,$AF,$BC,$F3,$3F,$C1,$BA
	.dc.b	$79,$DA,$72,$28,$BE,$CF,$13,$FA,$4D,$7F,$5C,$76,$D7,$37,$FB,$2A
	.dc.b	$FD,$D7,$D0,$7F,$02,$9E,$AD,$7F,$A9,$B7,$2F,$E4,$F0,$8F,$B0,$F0
	.dc.b	$DB,$A3,$F8,$7E,$3F,$EE,$0F,$F2,$5F,$17,$2F,$19,$BF,$40,$EB,$CF
	.dc.b	$CF,$FE,$C3,$3F,$E1,$CD,$BF,$F4,$A5,$6D,$F8,$6D,$7F,$B4,$C6,$D3
	.dc.b	$63,$59,$FA,$97,$C5,$78,$39,$A5,$F4,$8F,$29,$87,$AC,$61,$D1,$F5
	.dc.b	$2D,$1B,$BE,$D5,$F0,$3E,$05,$F2,$94,$98,$7D,$CF,$A0,$E2,$F3,$DB
	.dc.b	$E1,$91,$FC,$0D,$72,$7F,$07,$E8,$57,$FC,$12,$DC,$7E,$BB,$75,$F9
	.dc.b	$77,$F4,$BE,$81,$F1,$D4,$98,$7E,$3A,$3E,$3D,$7F,$3F,$FA,$07,$C5
	.dc.b	$FD,$A6,$D7,$E5,$1A,$FD,$2E,$FF,$BF,$97,$F8,$CA,$3C,$FE,$86,$0F
	.dc.b	$C1,$FB,$DF,$E5,$08,$FF,$32,$7D,$07,$F8,$39,$BE,$47,$E3,$37,$4E
	.dc.b	$67,$FC,$0E,$BD,$FD,$7D,$07,$A2,$7F,$8C,$36,$8D,$1F,$33,$C1,$93
	.dc.b	$FE,$75,$76,$DF,$97,$EB,$E9,$76,$D6,$4E,$E3,$C3,$C7,$E2,$E8,$93
	.dc.b	$CE,$3C,$BC,$5E,$6F,$3A,$4F,$78,$F4,$FC,$04,$8F,$13,$EC,$F8,$1F
	.dc.b	$6F,$BF,$27,$CC,$FC,$0F,$80,$F8,$3D,$72,$7D,$4F,$C4,$F7,$BE,$2F
	.dc.b	$C4,$FC,$6F,$89,$95,$EF,$8F,$93,$F0,$DF,$2B,$E1,$A4,$FC,$38,$FA
	.dc.b	$5F,$40,$F9,$EF,$9F,$E0,$4A,$FF,$4E,$4E,$5F,$E2,$FF,$41,$5F,$F2
	.dc.b	$EB,$7A,$BF,$85,$B6,$BF,$C1,$9F,$41,$FC,$3A,$73,$87,$F0,$8A,$D5
	.dc.b	$83,$85,$5C,$6F,$FC,$2D,$39,$25,$7F,$12,$9D,$57,$3F,$AA,$E6,$FF
	.dc.b	$6D,$6B,$FE,$E2,$6B,$FE,$21,$0E,$1F,$F1,$27,$D0,$B6,$FC,$B4,$61
	.dc.b	$AF,$A6,$1B,$7F,$35,$8D,$78,$1F,$EF,$0D,$1F,$30,$FA,$0F,$F6,$A3
	.dc.b	$76,$FF,$65,$3F,$C3,$1B,$1A,$23,$AF,$2A,$E1,$FE,$6E,$F9,$5F,$F4
	.dc.b	$06,$06,$FC,$5F,$F9,$FA,$FF,$DA,$D1,$A2,$78,$7F,$EE,$59,$D5,$F3
	.dc.b	$8F,$F7,$44,$EA,$BF,$DB,$D5,$AB,$FF,$37,$9D,$EA,$CF,$F9,$2F,$E8
	.dc.b	$3F,$86,$BF,$8A,$F8,$0C,$E3,$F7,$E7,$2F,$FA,$1B,$97,$FC,$DF,$87
	.dc.b	$F5,$49,$CC,$FD,$56,$BD,$7B,$BF,$AC,$B7,$0F,$9D,$9B,$C4,$F4,$9E
	.dc.b	$AF,$4A,$CF,$C2,$8B,$5F,$26,$6E,$3E,$CE,$07,$CF,$FD,$6F,$2F,$EB
	.dc.b	$B7,$1F,$AE,$DC,$3F,$D3,$F5,$C2,$7E,$DD,$70,$69,$C0,$B3,$FE,$C6
	.dc.b	$F9,$D7,$CF,$FC,$6C,$5F,$54,$F8,$8F,$93,$6B,$DE,$1F,$E0,$CB,$DD
	.dc.b	$6A,$73,$5F,$EB,$2B,$E5,$7F,$E1,$09,$CD,$FD,$A9,$CD,$7A,$BE,$2B
	.dc.b	$F2,$B8,$7E,$59,$B8,$7E,$59,$B3,$B8,$FC,$CC,$FE,$23,$CB,$91,$F9
	.dc.b	$0F,$32,$D7,$E5,$E6,$E3,$F8,$AE,$15,$62,$FF,$82,$4F,$E6,$EB,$17
	.dc.b	$FE,$59,$BD,$D7,$E7,$27,$E0,$7D,$07,$C9,$E5,$7D,$56,$2B,$F8,$DB
	.dc.b	$70,$FC,$83,$71,$FC,$DD,$70,$AF,$D3,$F9,$BF,$A2,$F3,$BF,$46,$D8
	.dc.b	$C9,$FA,$C8,$CB,$4F,$DD,$FD,$0F,$EF,$8F,$DF,$FD,$0D,$8C,$93,$2F
	.dc.b	$C4,$C6,$5F,$C5,$CD,$C7,$F2,$93,$6B,$E2,$F2,$AB,$17,$FE,$7D,$85
	.dc.b	$62,$FF,$88,$D5,$9C,$DF,$86,$7F,$41,$39,$AD,$36,$74,$78,$1C,$5F
	.dc.b	$9A,$32,$F8,$63,$DF,$18,$AF,$0B,$6F,$C3,$5C,$49,$3E,$29,$FE,$47
	.dc.b	$7E,$27,$83,$67,$71,$F0,$A3,$13,$C6,$7E,$5C,$F7,$BF,$3C,$E4,$91
	.dc.b	$EB,$3B,$9B,$3B,$96,$7B,$9F,$36,$BB,$C7,$18,$EE,$6A,$7D,$ED,$BF
	.dc.b	$7C,$78,$23,$E3,$4F,$67,$57,$E6,$8E,$93,$B3,$99,$F8,$27,$C8,$E6
	.dc.b	$BA,$C7,$02,$57,$63,$CB,$3B,$AE,$D1,$98,$C4,$FC,$28,$F1,$BE,$58
	.dc.b	$CC,$7A,$EF,$E0,$BC,$67,$E9,$1E,$43,$C3,$3F,$D8,$27,$6F,$AA,$37
	.dc.b	$3D,$8E,$8F,$17,$A1,$F3,$DC,$47,$18,$C1,$C5,$F8,$53,$DA,$FF,$49
	.dc.b	$5E,$B9,$F4,$9A,$79,$C6,$AB,$E2,$C7,$A2,$31,$3B,$9E,$D3,$9A,$ED
	.dc.b	$7F,$9E,$9C,$9E,$8A,$F8,$15,$DA,$75,$1E,$28,$E4,$64,$F3,$1E,$18
	.dc.b	$FA,$B1,$EF,$25,$F9,$27,$E4,$C7,$3A,$F0,$4F,$E0,$9B,$5C,$A3,$84
	.dc.b	$7F,$8B,$FE,$81,$F4,$3F,$DF,$BD,$2F,$E9,$B1,$8C,$7F,$C0,$F1,$89
	.dc.b	$D4,$7F,$B6,$27,$55,$FE,$11,$AF,$F0,$DF,$C2,$CA,$FF,$0D,$FC,$45
	.dc.b	$AF,$EC,$B3,$73,$FE,$DD,$36,$BF,$9F,$F1,$BD,$4F,$F0,$8F,$A9,$FE
	.dc.b	$12,$6C,$FF,$86,$9B,$8F,$F8,$85,$B8,$7F,$4B,$6C,$F4,$6C,$FF,$93
	.dc.b	$FE,$73,$FD,$AD,$3F,$E5,$EF,$7F,$FB,$6C,$62,$7F,$C2,$91,$8F,$03
	.dc.b	$FA,$63,$6B,$81,$FE,$2C,$F5,$BF,$C4,$E7,$F8,$A3,$D6,$B1,$92,$5F
	.dc.b	$F8,$DE,$32,$FF,$82,$3E,$83,$FC,$87,$F0,$7F,$E4,$46,$D7,$F9,$11
	.dc.b	$B3,$B8,$FF,$23,$CE,$DA,$6C,$B0,$D3,$FD,$FF,$9B,$D8,$FF,$1E,$FB
	.dc.b	$1F,$01,$B7,$FD,$C0,$FF,$3A,$5A,$B8,$47,$F9,$9E,$3E,$54,$FF,$8C
	.dc.b	$9B,$19,$78,$46,$6F,$FC,$C3,$F4,$1F,$E6,$03,$FC,$C3,$F4,$16,$FF
	.dc.b	$CC,$31,$6F,$FC,$57,$36,$BF,$C0,$DE,$97,$F9,$53,$D4,$FF,$2A,$B7
	.dc.b	$0F,$F2,$EB,$6F,$FC,$C8,$DA,$E3,$3F,$E6,$2A,$E2,$D3,$E0,$57,$FA
	.dc.b	$4F,$E8,$1F,$41,$FE,$0B,$F6,$BF,$C6,$31,$8A,$FF,$52,$C6,$27,$E1
	.dc.b	$52,$B4,$FC,$76,$BF,$E8,$F6,$BF,$A0,$F6,$FF,$CD,$E7,$F9,$C3,$DC
	.dc.b	$B8,$7F,$9E,$22,$CF,$F9,$FE,$6E,$5F,$E5,$7F,$63,$FD,$17,$EF,$7F
	.dc.b	$A3,$1B,$87,$FA,$31,$B7,$FE,$56,$6D,$71,$9F,$F4,$8D,$71,$69,$FE
	.dc.b	$4D,$F8,$0F,$DF,$CF,$FA,$01,$79,$BF,$E5,$77,$33,$FE,$38,$72,$F1
	.dc.b	$BD,$5E,$DB,$1F,$E5,$9F,$A0,$FE,$00,$DE,$DF,$F3,$14,$63,$2F,$CD
	.dc.b	$4F,$F9,$12,$13,$FE,$67,$FA,$0F,$DC,$67,$FE,$94,$6D,$7F,$92,$9F
	.dc.b	$F5,$29,$8B,$FD,$22,$EB,$E4,$73,$EB,$B5,$FE,$6D,$FA,$07,$D0,$7F
	.dc.b	$9D,$5B,$19,$F8,$5C,$4F,$F9,$E7,$D1,$FD,$3E,$5E,$83,$F6,$4B,$B7
	.dc.b	$5D,$9B,$7F,$D0,$1F,$40,$F9,$7B,$EF,$E1,$B9,$FF,$9F,$E5,$F4,$1C
	.dc.b	$3E,$7B,$8B,$67,$FD,$09,$ED,$FE,$4B,$8F,$FD,$6E,$D8,$FF,$36,$AF
	.dc.b	$8C,$FF,$48,$39,$DB,$F8,$42,$FA,$0C,$9A,$F9,$3B,$7F,$E9,$1F,$A0
	.dc.b	$FE,$9B,$89,$A7,$F5,$37,$1E,$A7,$FA,$6B,$E7,$FF,$0B,$97,$C5,$1F
	.dc.b	$40,$DB,$5E,$07,$FA,$82,$31,$97,$ED,$A9,$F4,$0F,$7B,$73,$98,$C3
	.dc.b	$BC,$7B,$E6,$8F,$E3,$BE,$FB,$E2,$3F,$D5,$B7,$C3,$1F,$24,$E6,$BA
	.dc.b	$63,$FE,$B2,$F0,$BF,$15,$46,$FC,$FB,$D8,$D7,$A9,$FE,$B8,$8C,$61
	.dc.b	$E9,$AD,$D6,$A3,$D9,$3A,$BF,$88,$DD,$3F,$DD,$FF,$40,$F9,$1E,$98
	.dc.b	$7F,$B1,$3E,$83,$F1,$18,$FC,$56,$E3,$F9,$1C,$1F,$33,$FE,$EE,$AF
	.dc.b	$E7,$B5,$A9,$F8,$94,$11,$99,$DF,$6F,$FB,$56,$F9,$56,$EB,$4D,$9F
	.dc.b	$21,$BF,$8D,$B7,$E5,$F6,$7D,$07,$FB,$7F,$E8,$1E,$36,$27,$44,$C3
	.dc.b	$D5,$3A,$8C,$67,$E8,$9F,$3F,$67,$C3,$3E,$23,$67,$E2,$45,$8F,$C8
	.dc.b	$67,$EF,$9C,$9F,$8D,$A7,$B3,$FE,$F0,$6C,$7E,$46,$BE,$37,$D0,$7C
	.dc.b	$7F,$9F,$7C,$F7,$F6,$8B,$DD,$7E,$75,$A7,$9C,$B7,$FD,$AE,$F9,$57
	.dc.b	$1C,$3F,$DF,$F7,$D8,$61,$D9,$FF,$72,$C7,$D8,$8C,$47,$F7,$8A,$E6
	.dc.b	$EA,$EB,$93,$9D,$F7,$8C,$C7,$7A,$E7,$87,$A9,$EB,$19,$AE,$0F,$23
	.dc.b	$FC,$0E,$2F,$BD,$71,$9F,$74,$EC,$FF,$0A,$61,$E9,$8C,$DF,$82,$31
	.dc.b	$3D,$2F,$9C,$62,$3D,$53,$BC,$3D,$31,$E1,$3F,$C2,$45,$3F,$C2,$95
	.dc.b	$87,$07,$F8,$7B,$E7,$DF,$02,$FA,$0F,$F0,$D4,$61,$B1,$99,$FF,$11
	.dc.b	$4E,$5B,$F0,$7F,$E2,$F6,$E9,$CA,$31,$1F,$1A,$51,$88,$F1,$3C,$72
	.dc.b	$FB,$7A,$36,$FF,$19,$C6,$1B,$1C,$30,$FB,$B8,$74,$7E,$FC,$71,$C2
	.dc.b	$D7,$07,$F7,$D5,$B7,$EA,$36,$FB,$47,$03,$AE,$1E,$5B,$F0,$47,$DB
	.dc.b	$7B,$36,$FB,$2B,$E9,$3C,$2F,$AC,$EC,$EB,$87,$77,$FA,$24,$F4,$8F
	.dc.b	$DF,$37,$4E,$4F,$F9,$58,$C4,$FF,$19,$7F,$AC,$3F,$D1,$DA,$F4,$C3
	.dc.b	$F6,$D3,$FE,$5E,$FA,$0F,$DF,$CF,$F4,$D6,$BF,$D5,$5B,$7F,$C3,$DF
	.dc.b	$B3,$1B,$DB,$9B,$FE,$6F,$E5,$FE,$D8,$FF,$93,$2B,$F9,$8F,$91,$A8
	.dc.b	$5F,$4C,$3F,$CD,$FF,$40,$ED,$D7,$D0,$7E,$CA,$7F,$CE,$6D,$BF,$E6
	.dc.b	$D5,$B7,$FB,$23,$5F,$ED,$91,$B4,$D8,$D6,$3F,$CA,$57,$C5,$F8,$3A
	.dc.b	$25,$F4,$8F,$44,$EA,$3F,$B9,$5A,$B7,$1F,$DE,$E7,$4D,$AC,$DF,$F3
	.dc.b	$7B,$E1,$3D,$4F,$EF,$9E,$83,$E5,$FE,$33,$63,$31,$F2,$7E,$83,$8C
	.dc.b	$7B,$B6,$F8,$E4,$76,$7F,$81,$FD,$3E,$4F,$F8,$3D,$B8,$FF,$49,$6E
	.dc.b	$BF,$3A,$BE,$B7,$D0,$7C,$68,$FF,$51,$AF,$8F,$C9,$AE,$0F,$AF,$DD
	.dc.b	$F7,$7E,$81,$F3,$34,$CE,$9F,$E1,$9F,$80,$FF,$25,$BE,$5F,$53,$F0
	.dc.b	$9A,$9F,$C3,$4E,$A7,$FB,$D7,$EC,$7F,$E4,$07,$F1,$8F,$F8,$9D,$BE
	.dc.b	$97,$E4,$B8,$13,$B9,$E3,$EA,$E6,$3F,$4F,$3A,$EC,$C4,$F1,$FA,$17
	.dc.b	$D5,$6F,$08,$7E,$6E,$32,$DF,$9E,$E0,$47,$5C,$FC,$F5,$91,$C5,$ED
	.dc.b	$F4,$9D,$FD,$29,$3D,$5F,$17,$CE,$F8,$FE,$57,$C9,$EC,$CA,$F1,$47
	.dc.b	$9F,$DA,$F4,$7C,$BC,$9F,$34,$7A,$FC,$AF,$67,$97,$27,$D5,$1E,$FF
	.dc.b	$22,$47,$B5,$F8,$5F,$49,$F0,$FE,$46,$4F,$C1,$7E,$37,$CA,$7C,$7F
	.dc.b	$94,$93,$F1,$5F,$95,$F2,$9F,$BF,$F9,$2F,$F6,$5C,$FF,$96,$5A,$FF
	.dc.b	$98,$5B,$D5,$FE,$1B,$7F,$84,$C6,$7E,$4B,$91,$1E,$5E,$3D,$74,$75
	.dc.b	$8F,$04,$78,$7E,$92,$4F,$85,$FE,$25,$F4,$9F,$E1,$3F,$27,$FD,$FF
	.dc.b	$82,$77,$5E,$57,$73,$FC,$67,$E8,$DF,$2D,$FE,$61,$AF,$F0,$08,$8C
	.dc.b	$FC,$0E,$23,$1F,$2F,$FC,$C9,$FF,$27,$1B,$AE,$4F,$F8,$2B,$E9,$7F
	.dc.b	$CE,$0F,$46,$B9,$F8,$6F,$F7,$7C,$62,$3F,$CB,$1B,$BE,$71,$E1,$7F
	.dc.b	$97,$22,$46,$34,$FE,$68,$FF,$0F,$6C,$ED,$E7,$1B,$AF,$F3,$53,$6B
	.dc.b	$9F,$85,$B9,$47,$86,$F0,$DB,$EA,$78,$2B,$37,$CE,$3D,$3F,$37,$F0
	.dc.b	$2B,$D0,$D8,$F6,$FC,$77,$FA,$3D,$FE,$99,$19,$C7,$C2,$DA,$36,$B8
	.dc.b	$4F,$FA,$0E,$38,$2F,$A4,$FE,$2F,$7F,$5D,$B1,$F7,$1A,$FF,$8F,$9A
	.dc.b	$FD,$A7,$77,$FE,$73,$6C,$FC,$AF,$13,$E3,$57,$7F,$3B,$E2,$E0,$FA
	.dc.b	$4A,$63,$D2,$B8,$7B,$A2,$DF,$A6,$6E,$5F,$1A,$6D,$7C,$A9,$B5,$F5
	.dc.b	$34,$78,$DF,$43,$C8,$FA,$2D,$BF,$C1,$B6,$BE,$FB,$71,$FF,$47,$7D
	.dc.b	$1B,$E9,$3D,$6D,$BF,$A7,$F1,$38,$99,$1F,$ED,$79,$D6,$1F,$46,$B2
	.dc.b	$FE,$33,$CA,$FC,$1F,$97,$F8,$46,$DF,$E5,$9B,$3B,$8F,$CC,$CE,$DA
	.dc.b	$6C,$B1,$F7,$FE,$33,$E7,$9F,$43,$D4,$B3,$F8,$A8,$B3,$F8,$59,$B9
	.dc.b	$FC,$0F,$A3,$F7,$9F,$98,$F3,$AE,$1E,$D8,$B7,$FA,$69,$B5,$F9,$AF
	.dc.b	$63,$EF,$FB,$BF,$80,$6E,$1F,$9A,$6D,$FE,$55,$B5,$C6,$7F,$3F,$5C
	.dc.b	$5A,$71,$2D,$71,$6C,$F1,$8F,$D0,$4F,$16,$CF,$16,$9C,$4B,$3C,$4F
	.dc.b	$D9,$CF,$16,$CF,$16,$CF,$16,$9C,$63,$F7,$B3,$C5,$A7,$12,$CF,$16
	.dc.b	$CF,$18,$FD,$EC,$F1,$6C,$F1,$69,$C4,$B3,$C4,$FE,$07,$3C,$5B,$3C
	.dc.b	$5B,$3C,$5A,$71,$E5,$FF,$22,$CE,$DB,$3B,$6C,$EE,$BB,$46,$2B,$F8
	.dc.b	$ED,$FD,$C6,$9F,$58,$B7,$F4,$1B,$7F,$29,$A7,$C4,$2C,$7B,$9B,$1E
	.dc.b	$86,$D7,$86,$7F,$92,$D7,$85,$B3,$E1,$69,$E1,$2D,$78,$5B,$1E,$16
	.dc.b	$9E,$12,$C7,$85,$B5,$E1,$8F,$E6,$75,$E1,$6D,$78,$5B,$5E,$16,$9E
	.dc.b	$12,$C7,$85,$B1,$E1,$69,$E1,$2D,$78,$4F,$F2,$C4,$F7,$6C,$F7,$6C
	.dc.b	$F7,$6C,$F7,$6C,$F7,$6C,$F7,$69,$DE,$BD,$CE,$2F,$F0,$64,$BF,$C5
	.dc.b	$36,$FF,$14,$DB,$FC,$54,$E8,$93,$A6,$C6,$A3,$04,$70,$48,$FC,$1C
	.dc.b	$F5,$69,$C8,$B3,$C5,$B3,$C6,$3E,$DC,$F1,$6C,$F1,$69,$C4,$B3,$C4
	.dc.b	$FA,$73,$C5,$FF,$0B,$79,$FF,$E0,$23,$FC,$07,$E8,$5B,$FF,$0A,$C5
	.dc.b	$BF,$F0,$8C,$DA,$57,$F8,$AB,$E8,$59,$FF,$6E,$6C,$7F,$81,$BA,$71
	.dc.b	$32,$3F,$C4,$E7,$17,$FB,$9C,$67,$0F,$F0,$FF,$A1,$FE,$16,$3F,$C2
	.dc.b	$FE,$8D,$BF,$F1,$4C,$59,$FF,$12,$CD,$C7,$FC,$5F,$36,$BF,$C5,$73
	.dc.b	$6B,$FC,$2F,$F4,$3F,$E0,$2F,$9C,$FF,$01,$B6,$BF,$C3,$AD,$9F,$F0
	.dc.b	$C3,$70,$F8,$30,$1E,$7F,$FB,$A6,$BD,$B1,$98,$C3,$FE,$3A,$F4,$BF
	.dc.b	$C2,$71,$8A,$E7,$3C,$BB,$B2,$7F,$8D,$AA,$DF,$F8,$C3,$E6,$FF,$C3
	.dc.b	$27,$F8,$6B,$E6,$EC,$64,$97,$FE,$2F,$8C,$BF,$E3,$19,$B8,$7F,$8F
	.dc.b	$A6,$D7,$F9,$43,$D3,$FF,$22,$FA,$BF,$E4,$66,$C7,$F9,$61,$B8,$7F
	.dc.b	$95,$1B,$7E,$AC,$DF,$03,$EC,$AF,$F3,$7C,$62,$FF,$CE,$8D,$8F,$F3
	.dc.b	$CB,$5F,$F3,$A9,$F3,$27,$E3,$9F,$E3,$82,$CF,$F9,$AB,$D5,$FF,$2D
	.dc.b	$9F,$E5,$CF,$5A,$C6,$49,$5F,$E6,$A8,$CB,$FE,$6C,$9B,$87,$F9,$DA
	.dc.b	$6D,$7F,$99,$3C,$FF,$F2,$3F,$9F,$FE,$48,$6C,$7F,$98,$9B,$87,$F9
	.dc.b	$75,$B7,$F7,$3E,$85,$EB,$FC,$3A,$FC,$44,$62,$FF,$D2,$CD,$8F,$F4
	.dc.b	$DB,$5F,$86,$E6,$BA,$57,$F9,$EB,$4F,$F2,$EC,$E6,$FF,$CF,$7F,$43
	.dc.b	$FE,$5E,$3F,$CB,$FF,$43,$63,$FC,$F3,$17,$1F,$F3,$A4,$DC,$7F,$D0
	.dc.b	$33,$6B,$FD,$0B,$36,$BF,$D1,$FF,$00,$F1,$7F,$D1,$ED,$AF,$D0,$E0
	.dc.b	$8F,$F3,$0C,$61,$D5,$7F,$A6,$23,$FD,$53,$1F,$AA,$75,$97,$4D,$FA
	.dc.b	$BF,$E9,$CF,$A0,$FF,$3E,$DF,$FA,$81,$B8,$17,$CA,$6F,$B7,$FD,$47
	.dc.b	$F3,$DF,$E9,$2A,$FF,$5A,$2F,$A1,$CF,$B3,$A7,$FD,$53,$F4,$2E,$EB
	.dc.b	$DF,$86,$EF,$EE,$FA,$1F,$F3,$6E,$5F,$EB,$F5,$F6,$5F,$EB,$27,$33
	.dc.b	$FE,$C6,$72,$FC,$6F,$31,$E0,$7F,$AE,$1B,$19,$F0,$31,$3F,$EB,$CF
	.dc.b	$A1,$6F,$4C,$7C,$55,$BF,$A3,$2F,$3B,$81,$C7,$E4,$AC,$FD,$6F,$59
	.dc.b	$F3,$5F,$EC,$85,$EE,$58,$C3,$9F,$A1,$D7,$C6,$5B,$E4,$FF,$B4,$3E
	.dc.b	$87,$FA,$0F,$67,$DD,$3F,$DA,$AB,$D7,$D7,$A1,$B7,$FD,$6B,$F4,$19
	.dc.b	$C3,$F9,$4D,$BE,$85,$8E,$AF,$FD,$77,$18,$AD,$FD,$0B,$DA,$D4,$61
	.dc.b	$B8,$78,$CF,$CC,$7C,$37,$FB,$12,$B8,$57,$83,$C7,$FF,$6A,$C6,$2B
	.dc.b	$87,$83,$FC,$DE,$BF,$A0,$42,$30,$FF,$43,$FA,$1F,$C8,$CF,$AA,$73
	.dc.b	$29,$CB,$D3,$D4,$FF,$6A,$C6,$3B,$FF,$A1,$D7,$08,$C5,$F8,$1B,$7F
	.dc.b	$F0,$03,$E8,$5E,$F7,$2F,$1F,$FD,$BF,$F3,$FF,$73,$0F,$96,$DB,$FA
	.dc.b	$99,$3E,$6D,$F4,$3C,$3A,$A6,$1C,$1F,$EB,$F5,$2B,$73,$92,$5F,$CD
	.dc.b	$C3,$B5,$6F,$0E,$73,$9F,$9B,$F2,$9D,$2F,$FB,$55,$3E,$85,$E0,$EA
	.dc.b	$7D,$35,$6F,$D3,$5A,$9C,$9E,$EF,$96,$F8,$27,$C2,$9B,$19,$23,$E7
	.dc.b	$32,$7E,$17,$07,$B7,$E5,$8F,$2B,$5F,$C0,$FD,$0F,$F8,$07,$E7,$DE
	.dc.b	$1F,$E1,$2F,$8E,$1F,$9A,$6D,$F5,$C3,$93,$FE,$09,$C3,$95,$F1,$9C
	.dc.b	$CA,$72,$F8,$AE,$BE,$21,$87,$89,$D1,$8B,$FE,$D6,$F3,$30,$6A,$38
	.dc.b	$3C,$DB,$87,$A1,$CC,$79,$CE,$22,$35,$3D,$AF,$CB,$18,$9F,$EF,$71
	.dc.b	$8C,$3F,$C5,$37,$D2,$7C,$8F,$82,$33,$7C,$E7,$BD,$F3,$7C,$D1,$F1
	.dc.b	$F0,$F8,$71,$E5,$3D,$E7,$BC,$B8,$61,$D1,$FB,$AF,$A0,$7C,$03,$E8
	.dc.b	$78,$C6,$1B,$19,$9F,$97,$39,$6F,$C1,$7C,$36,$E5,$CA,$31,$1F,$E1
	.dc.b	$C9,$46,$23,$C4,$F1,$D3,$E9,$E6,$D7,$F8,$7C,$61,$B1,$C3,$0F,$C1
	.dc.b	$E1,$D1,$FC,$2C,$71,$C2,$D7,$07,$FC,$B1,$5B,$7E,$DB,$6F,$B4,$70
	.dc.b	$3A,$E1,$E5,$BF,$04,$7E,$09,$EC,$DB,$EC,$AF,$A4,$F0,$BE,$B3,$B3
	.dc.b	$AE,$1D,$DF,$E6,$D3,$D2,$3F,$CC,$AD,$D3,$93,$FC,$C8,$C4,$FF,$9C
	.dc.b	$D7,$CD,$7E,$CE,$74,$D9,$D6,$35,$7D,$0F,$F0,$07,$FA,$8B,$5F,$E9
	.dc.b	$0D,$C3,$FC,$EE,$FD,$A8,$DE,$FD,$67,$F9,$67,$CC,$DE,$63,$DE,$F9
	.dc.b	$65,$FD,$2B,$E8,$5D,$DE,$3F,$A1,$FD,$8C,$FF,$4E,$6E,$7F,$E8,$FA
	.dc.b	$DB,$FD,$99,$AF,$F5,$D8,$DA,$6C,$6B,$2F,$F4,$7A,$F8,$BF,$07,$3C
	.dc.b	$BE,$91,$F1,$67,$51,$FD,$D2,$D5,$B8,$FE,$D9,$3A,$6D,$66,$FF,$D3
	.dc.b	$B7,$C2,$7A,$9F,$DF,$BE,$19,$F2,$BF,$4D,$B1,$98,$FA,$DF,$43,$C6
	.dc.b	$3D,$DB,$7C,$72,$3B,$3F,$C1,$3E,$B7,$27,$FC,$00,$DC,$7F,$D5,$ED
	.dc.b	$D7,$E7,$57,$E0,$FE,$87,$E3,$47,$F7,$E5,$F2,$19,$35,$C1,$F5,$FB
	.dc.b	$BF,$15,$F4,$2F,$80,$A6,$74,$FF,$0D,$FC,$27,$F8,$E1,$F2,$FA,$DF
	.dc.b	$91,$D4,$FE,$4E,$75,$3F,$DD,$BE,$85,$EB,$7E,$35,$FF,$11,$C7,$F9
	.dc.b	$B9,$AE,$E3,$5D,$9B,$EC,$93,$B8,$FC,$B7,$D0,$3E,$87,$A4,$C3,$FC
	.dc.b	$F3,$19,$6F,$CE,$70,$23,$AE,$7E,$7A,$48,$E2,$F6,$F9,$BE,$FE,$94
	.dc.b	$9E,$AF,$8B,$E6,$7C,$7E,$D7,$93,$D8,$95,$E2,$8F,$3F,$CB,$7A,$3E
	.dc.b	$5A,$4F,$9A,$3D,$7E,$57,$B3,$CB,$93,$EA,$8F,$7F,$91,$23,$DA,$FC
	.dc.b	$2F,$95,$F8,$7F,$29,$27,$E0,$BF,$1B,$E5,$7E,$3F,$C7,$49,$F8,$AF
	.dc.b	$CA,$F8,$FF,$E1,$5F,$41,$FE,$BE,$9F,$F2,$CB,$5F,$F2,$83,$7C,$6F
	.dc.b	$F7,$FB,$FC,$12,$37,$F2,$9C,$88,$F2,$F1,$EB,$CA,$2D,$72,$9F,$0F
	.dc.b	$D0,$4A,$F0,$BF,$C7,$7E,$85,$E4,$E6,$7C,$73,$6F,$30,$15,$87,$FC
	.dc.b	$D1,$F4,$CC,$BF,$B5,$36,$3F,$B7,$B5,$FF,$07,$B5,$FF,$7D,$FD,$13
	.dc.b	$E6,$BF,$CD,$11,$FE,$14,$8C,$7C,$17,$F7,$85,$F3,$5F,$CE,$A7,$55
	.dc.b	$C1,$FF,$17,$7C,$FB,$C1,$FF,$1B,$68,$8C,$6D,$FE,$D2,$7F,$86,$B6
	.dc.b	$77,$1A,$AF,$F3,$1B,$6B,$FC,$BE,$F0,$8C,$CF,$02,$57,$5F,$06,$CF
	.dc.b	$F9,$B3,$E7,$5F,$0F,$E4,$6D,$7F,$B9,$3E,$37,$FD,$A4,$FF,$98,$23
	.dc.b	$35,$C6,$B8,$1E,$CA,$DC,$7F,$8B,$E7,$79,$7D,$5F,$A1,$FF,$70,$FD
	.dc.b	$27,$F9,$D2,$BF,$D3,$D1,$8F,$83,$FF,$2C,$3F,$C5,$72,$FF,$12,$B6
	.dc.b	$7F,$C8,$AD,$7F,$D3,$A1,$C3,$E7,$FA,$2F,$61,$ED,$F5,$2D,$7F,$C0
	.dc.b	$0C,$93,$8A,$F8,$95,$72,$FF,$45,$D6,$AF,$85,$F3,$69,$F6,$7C,$6F
	.dc.b	$AF,$E4,$FD,$86,$C6,$4F,$BD,$19,$69,$FE,$CA,$FA,$1F,$F7,$8E,$5F
	.dc.b	$E5,$7F,$A6,$79,$5F,$E7,$B9,$CE,$69,$CD,$F1,$FA,$3F,$D1,$61,$FE
	.dc.b	$C0,$B4,$62,$3F,$51,$18,$9F,$F2,$8B,$63,$E5,$DF,$E5,$7C,$BF,$C8
	.dc.b	$E3,$F9,$26,$DF,$E5,$DB,$5F,$94,$6D,$7E,$51,$B3,$FE,$A8,$E6,$FF
	.dc.b	$74,$57,$E6,$AF,$F2,$1F,$1F,$2B,$EA,$7A,$B6,$BF,$2D,$37,$1F,$A7
	.dc.b	$36,$BD,$DF,$29,$FA,$CE,$57,$D3,$7C,$78,$C3,$67,$FD,$72,$7F,$AE
	.dc.b	$A7,$6F,$D1,$9D,$BF,$85,$F6,$FF,$49,$F0,$1F,$A5,$6C,$64,$FC,$54
	.dc.b	$65,$A6,$73,$FD,$74,$65,$B1,$9C,$BF,$1B,$39,$6C,$64,$FD,$84,$65
	.dc.b	$A6,$67,$F5,$D1,$96,$C6,$65,$CB,$FE,$29,$8C,$36,$7F,$81,$5F,$B1
	.dc.b	$B3,$EA,$69,$E8,$2D,$79,$9B,$1E,$46,$9E,$22,$C7,$8A,$FC,$CD,$9F
	.dc.b	$1C,$62,$FC,$93,$FB,$0B,$F2,$36,$BC,$8D,$BF,$23,$4F,$21,$63,$C8
	.dc.b	$D8,$F2,$34,$F1,$3E,$6A,$F0,$CF,$F1,$FA,$F0,$B4,$F0,$96,$7C,$2D
	.dc.b	$AF,$0B,$4F,$09,$63,$C2,$D8,$F0,$9E,$3A,$F0,$CF,$F2,$9A,$F0,$B6
	.dc.b	$BC,$2D,$3C,$25,$AF,$0B,$63,$C2,$D3,$C2,$58,$F0,$C7,$8E,$BC,$31
	.dc.b	$F8,$8A,$F0,$B6,$7C,$2D,$AF,$0B,$4F,$09,$63,$C2,$D8,$F0,$B4,$F0
	.dc.b	$BE,$3A,$F0,$9F,$E5,$A9,$EE,$D8,$EE,$D9,$EE,$D9,$EE,$D9,$EE,$D9
	.dc.b	$EE,$78,$5E,$F5,$EE,$74,$62,$BC,$31,$92,$38,$25,$79,$23,$2D,$8C
	.dc.b	$D7,$9A,$72,$D8,$C9,$E8,$8C,$B4,$CC,$FA,$A3,$2D,$8C,$CF,$B2,$BD
	.dc.b	$37,$FA,$B2,$4F,$EB,$89,$5F,$9F,$8C,$B6,$33,$5F,$93,$9C,$B6,$32
	.dc.b	$7F,$3B,$F3,$FF,$C2,$A7,$F8,$5B,$D0,$B1,$FE,$1B,$8B,$7F,$E1,$89
	.dc.b	$B5,$FE,$37,$FA,$47,$A5,$FE,$11,$F8,$EF,$F1,$EC,$E6,$31,$8F,$F3
	.dc.b	$8C,$3F,$C8,$17,$C2,$3F,$BD,$5F,$07,$FC,$89,$29,$D3,$FE,$1F,$9F
	.dc.b	$F1,$5F,$A5,$2B,$FC,$61,$E9,$DA,$FF,$16,$CD,$C7,$FC,$5B,$36,$BF
	.dc.b	$AD,$4D,$A5,$7F,$8E,$EB,$0F,$F3,$0F,$9D,$FF,$16,$7C,$F7,$F8,$B5
	.dc.b	$B8,$7F,$84,$5B,$7F,$E3,$36,$D7,$F9,$6F,$D1,$FF,$77,$D7,$C1,$87
	.dc.b	$1F,$FB,$72,$3E,$25,$AB,$6F,$92,$3F,$AF,$7C,$37,$C5,$8F,$F3,$44
	.dc.b	$E9,$FF,$14,$B5,$F9,$2D,$77,$E0,$FF,$82,$70,$FE,$97,$F3,$7F,$E4
	.dc.b	$73,$FC,$91,$F3,$97,$1F,$F1,$64,$5B,$FF,$27,$4D,$AF,$F1,$5C,$DA
	.dc.b	$FF,$26,$4D,$AF,$97,$D8,$F5,$7F,$CC,$3E,$C7,$F9,$89,$B7,$FE,$67
	.dc.b	$6D,$7F,$99,$1B,$8F,$C9,$F8,$6F,$F2,$E7,$D3,$7C,$1A,$F8,$D1,$8B
	.dc.b	$F9,$3A,$A3,$FC,$4D,$3A,$9F,$F2,$6B,$F2,$23,$33,$B9,$CD,$FF,$9E
	.dc.b	$FD,$8F,$F3,$B9,$FE,$78,$F6,$6C,$FF,$9F,$62,$E1,$FE,$7A,$9B,$5F
	.dc.b	$E7,$A9,$B5,$FE,$27,$9B,$5F,$E6,$BF,$47,$FC,$ED,$E8,$FF,$9D,$DB
	.dc.b	$5F,$E5,$B6,$CF,$F9,$F5,$B8,$7F,$A9,$7E,$95,$ED,$7F,$89,$FD,$7F
	.dc.b	$89,$18,$C3,$F1,$DC,$DF,$EA,$97,$F1,$B1,$9C,$7E,$1B,$5F,$F3,$8B
	.dc.b	$5F,$D8,$B5,$FF,$41,$7D,$2F,$FA,$34,$FF,$47,$7D,$2D,$AF,$F3,$94
	.dc.b	$59,$FF,$4A,$4D,$CB,$FC,$E3,$36,$BF,$D2,$53,$6B,$F6,$3F,$0B,$FE
	.dc.b	$62,$7F,$40,$63,$CA,$FF,$57,$C3,$D0,$B1,$8E,$29,$87,$EB,$69,$C5
	.dc.b	$8C,$39,$E6,$FA,$5F,$F4,$93,$FC,$59,$B7,$FE,$D3,$8D,$78,$3F,$EB
	.dc.b	$E9,$CD,$7F,$06,$FA,$1F,$F5,$2D,$17,$D2,$FF,$B1,$1C,$98,$C3,$F8
	.dc.b	$07,$D2,$BC,$1F,$F6,$4A,$F0,$AC,$61,$FF,$66,$FD,$27,$FA,$18,$7D
	.dc.b	$2E,$3D,$4C,$BF,$ED,$2F,$45,$E2,$53,$1E,$25,$CF,$F8,$87,$D2,$F9
	.dc.b	$DB,$AF,$FB,$6A,$1F,$1F,$63,$1E,$94,$AE,$3F,$2D,$78,$FE,$5F,$B2
	.dc.b	$F9,$DF,$F7,$1A,$F7,$AC,$61,$CF,$C7,$6B,$E8,$EE,$3F,$C8,$ED,$1F
	.dc.b	$EE,$CF,$A5,$FE,$7F,$D3,$FD,$29,$FF,$47,$B7,$D4,$D7,$B9,$FC,$0F
	.dc.b	$E9,$73,$8F,$F4,$8F,$A5,$E6,$FF,$2C,$8C,$EC,$13,$A7,$F1,$E3,$E9
	.dc.b	$5E,$FF,$5C,$3F,$88,$46,$30,$E9,$1F,$C4,$FC,$5F,$F7,$2C,$7F,$58
	.dc.b	$9D,$5F,$49,$CB,$E1,$8C,$E6,$AF,$EB,$FF,$13,$FD,$02,$31,$5E,$78
	.dc.b	$C4,$FF,$64,$FA,$57,$89,$FC,$8E,$73,$5E,$D9,$CD,$7F,$4C,$9C,$CA
	.dc.b	$72,$FC,$08,$CC,$62,$FC,$33,$BB,$FE,$A1,$0F,$A5,$7D,$2F,$F3,$3B
	.dc.b	$E1,$8F,$F7,$8F,$A5,$FC,$23,$FD,$71,$B8,$FE,$CB,$E8,$1F,$15,$FC
	.dc.b	$EE,$73,$97,$81,$B7,$CF,$A7,$C3,$7C,$23,$A4,$71,$23,$FD,$B6,$35
	.dc.b	$87,$DC,$AE,$6F,$E0,$5F,$EE,$0E,$7D,$DF,$6F,$D2,$FE,$8A,$FD,$11
	.dc.b	$E5,$23,$FD,$E6,$35,$3C,$63,$DD,$F3,$3E,$B3,$D9,$16,$32,$47,$C8
	.dc.b	$64,$FC,$8E,$4F,$77,$FA,$F3,$67,$FC,$0B,$3F,$99,$FA,$5F,$CD,$7D
	.dc.b	$1B,$C0,$FF,$13,$DE,$EB,$9E,$1C,$8F,$ED,$0F,$11,$5C,$5F,$E4,$11
	.dc.b	$A5,$7D,$AF,$AC,$FE,$52,$B8,$57,$F8,$DA,$B4,$F2,$C2,$B9,$3B,$DF
	.dc.b	$28,$CC,$78,$EB,$9E,$1E,$97,$AC,$66,$B8,$3C,$8F,$F2,$29,$E6,$30
	.dc.b	$7F,$8E,$B0,$F3,$47,$F8,$03,$0F,$1D,$F5,$8C,$57,$08,$C3,$D2,$77
	.dc.b	$5C,$A3,$06,$DC,$3E,$D3,$FC,$1A,$78,$CB,$18,$7E,$81,$E8,$FA,$47
	.dc.b	$C1,$BE,$97,$8C,$61,$B1,$99,$FF,$0C,$CE,$5B,$F0,$BF,$E5,$86,$E3
	.dc.b	$CA,$31,$1D,$DA,$FF,$9B,$25,$1C,$73,$FB,$79,$B5,$FF,$38,$C6,$1B
	.dc.b	$1C,$67,$83,$70,$FB,$EF,$1C,$2D,$70,$7F,$98,$61,$E0,$AF,$C2,$B6
	.dc.b	$3E,$CD,$71,$6D,$71,$53,$BC,$3F,$22,$DA,$FF,$43,$9C,$63,$D1,$3D
	.dc.b	$B0,$F3,$BE,$26,$BF,$A9,$BE,$D5,$C4,$F6,$47,$88,$FF,$3A,$BE,$5C
	.dc.b	$3F,$92,$B7,$5D,$E1,$E0,$AF,$E9,$72,$F9,$DF,$DA,$D7,$99,$B3,$FD
	.dc.b	$4E,$79,$3F,$D5,$7E,$97,$F8,$13,$FE,$84,$6B,$FD,$59,$B8,$7F,$36
	.dc.b	$7E,$DC,$6F,$7E,$B2,$E3,$7C,$D6,$1B,$5B,$3F,$D2,$3F,$4A,$F8,$9F
	.dc.b	$2C,$3E,$97,$F6,$BA,$FF,$AD,$1B,$1F,$DA,$EB,$83,$CA,$B8,$CE,$A3
	.dc.b	$96,$9D,$2B,$71,$FD,$C1,$7C,$77,$0E,$79,$3E,$48,$E0,$61,$FE,$EF
	.dc.b	$69,$CC,$6C,$FF,$55,$C6,$5B,$1A,$7F,$BF,$FC,$43,$E5,$FD,$ED,$8F
	.dc.b	$F5,$9B,$5F,$F0,$2F,$D2,$F3,$7C,$3B,$7B,$88,$F0,$AF,$93,$EA,$F1
	.dc.b	$E0,$FE,$D6,$DD,$FF,$DB,$AD,$AF,$F0,$A6,$7F,$51,$FF,$09,$FD,$2F
	.dc.b	$C6,$BF,$EE,$6B,$E4,$FE,$AB,$67,$FD,$92,$EF,$0F,$AD,$EA,$7D,$AF
	.dc.b	$A5,$7D,$2D,$3E,$AC,$7D,$EF,$83,$FF,$2D,$61,$FE,$D2,$D1,$E9,$FA
	.dc.b	$8D,$7A,$FF,$84,$FB,$47,$9D,$F8,$6E,$B5,$D3,$51,$D7,$C7,$FC,$9F
	.dc.b	$D3,$7F,$80,$32,$FF,$71,$35,$FF,$18,$E7,$FE,$4B,$7F,$DC,$6F,$FA
	.dc.b	$79,$DF,$CC,$72,$3D,$1F,$1D,$F9,$6F,$93,$7B,$DF,$98,$FB,$3F,$F4
	.dc.b	$35,$FF,$BB,$1A,$FF,$8F,$A7,$6F,$6B,$F2,$F6,$C8,$C1,$E2,$F9,$0F
	.dc.b	$49,$1C,$1F,$EF,$8F,$B2,$7D,$07,$EA,$3E,$CB,$D1,$3F,$E4,$C6,$F6
	.dc.b	$7F,$8B,$2B,$73,$EE,$FB,$5F,$8B,$F7,$32,$30,$7C,$2F,$84,$F9,$33
	.dc.b	$23,$07,$FA,$5F,$B7,$FC,$B8,$BE,$D3,$FC,$8A,$E4,$FF,$30,$7D,$A7
	.dc.b	$E0,$9C,$07,$0F,$F9,$8A,$38,$BF,$E6,$48,$FF,$31,$BF,$63,$C9,$FB
	.dc.b	$67,$F0,$68,$CC,$7D,$AF,$3F,$EF,$E3,$23,$ED,$C7,$F0,$CF,$C4,$BF
	.dc.b	$D1,$F1,$E2,$6B,$FE,$72,$6E,$BF,$E5,$2E,$0F,$F4,$2C,$EF,$C0,$FE
	.dc.b	$2B,$1F,$C3,$4F,$E1,$D5,$C3,$4F,$86,$DA,$FF,$40,$4E,$F3,$F1,$5F
	.dc.b	$2E,$79,$5F,$12,$BE,$BE,$1F,$31,$FD,$A6,$3E,$3A,$DD,$7F,$A2,$27
	.dc.b	$6F,$6F,$1F,$CB,$F5,$52,$3A,$3F,$CB,$3E,$AB,$FC,$DB,$87,$B6,$35
	.dc.b	$3C,$DD,$63,$FC,$CB,$EA,$5F,$3A,$F8,$1F,$F0,$7C,$61,$BF,$3B,$FD
	.dc.b	$3F,$0F,$F1,$28,$9D,$38,$FA,$F7,$D5,$7F,$9F,$A3,$F1,$31,$8C,$7F
	.dc.b	$C3,$6D,$8E,$13,$B7,$FD,$55,$1F,$D1,$E7,$7E,$A6,$65,$39,$7C,$11
	.dc.b	$C7,$47,$D1,$7E,$A2,$FF,$2D,$18,$ED,$FF,$47,$CF,$F3,$59,$CB,$BB
	.dc.b	$FF,$7D,$B6,$BF,$D4,$E7,$19,$DF,$A3,$CA,$FF,$D3,$3F,$4D,$FE,$B2
	.dc.b	$A3,$7E,$43,$FD,$F6,$E4,$9A,$7F,$A8,$23,$2F,$6C,$3F,$5F,$F5,$5F
	.dc.b	$CB,$1F,$F2,$0B,$73,$FA,$2D,$7F,$CA,$ED,$78,$FA,$1C,$DF,$F6,$A4
	.dc.b	$71,$C7,$D7,$A3,$EB,$75,$4A,$D6,$1C,$8F,$DD,$B8,$85,$FC,$EA,$E3
	.dc.b	$39,$8F,$83,$F5,$7F,$10,$F8,$BB,$DC,$3E,$9E,$A9,$C6,$3D,$27,$FD
	.dc.b	$D5,$7F,$2A,$F2,$FA,$3E,$83,$EC,$FA,$7F,$69,$B5,$F6,$9B,$7F,$69
	.dc.b	$B5,$FE,$D0,$FA,$A7,$CC,$FC,$06,$DF,$93,$E2,$F1,$32,$3F,$8D,$CE
	.dc.b	$B0,$F3,$1F,$0B,$4F,$F6,$BC,$EA,$B8,$4F,$FC,$00,$AE,$18,$7E,$2A
	.dc.b	$F7,$39,$AF,$C9,$FC,$7F,$E5,$3D,$3B,$8F,$E5,$1B,$5F,$94,$6E,$1F
	.dc.b	$69,$FC,$8F,$C8,$4A,$F9,$BE,$AD,$BF,$3D,$FE,$52,$B0,$F8,$E6,$D7
	.dc.b	$F3,$EC,$DE,$3F,$C5,$3F,$4F,$F2,$76,$7E,$1C,$5C,$7F,$4F,$37,$0F
	.dc.b	$9D,$36,$BE,$DF,$2A,$B1,$61,$87,$82,$BA,$9F,$B4,$9D,$CE,$AF,$8E
	.dc.b	$1F,$BE,$9C,$DF,$1A,$FA,$15,$E8,$BE,$71,$86,$DF,$E8,$AF,$8D,$74
	.dc.b	$C3,$CD,$5A,$3E,$24,$66,$77,$19,$BF,$D1,$D7,$07,$B1,$2F,$E1,$CF
	.dc.b	$AC,$E1,$1E,$B7,$94,$7A,$DC,$E1,$EB,$8E,$33,$F3,$A7,$38,$78,$DA
	.dc.b	$72,$BF,$15,$70,$24,$7A,$8F,$15,$79,$5C,$34,$F5,$E1,$EE,$71,$3E
	.dc.b	$57,$47,$A4,$CD,$78,$2B,$9C,$F8,$27,$EF,$61,$EB,$9F,$C1,$DF,$5A
	.dc.b	$D1,$D4,$E1,$5D,$5C,$DF,$58,$E7,$87,$1B,$E6,$7C,$28,$F1,$DE,$5D
	.dc.b	$C7,$C0,$BE,$B8,$78,$E7,$27,$CD,$3D,$AE,$5F,$14,$F6,$73,$7F,$45
	.dc.b	$B1,$F4,$6F,$C1,$1C,$0F,$97,$1E,$5B,$F5,$CE,$AF,$C5,$1F,$06,$7E
	.dc.b	$49,$2B,$CB,$1F,$97,$8E,$46,$5E,$11,$A8,$E0,$F8,$63,$2F,$8C,$93
	.dc.b	$E0,$C3,$E6,$46,$30,$F6,$57,$C6,$6D,$F4,$BF,$88,$71,$C3,$D5,$3F
	.dc.b	$AF,$BF,$86,$7D,$2A,$F3,$BE,$53,$E1,$9A,$BF,$87,$1A,$8F,$46,$1E
	.dc.b	$4C,$3B,$DF,$8A,$79,$1E,$7A,$FB,$87,$96,$35,$19,$3C,$47,$78,$F9
	.dc.b	$15,$D6,$7F,$0D,$7F,$1A,$FE,$9E,$DF,$EA,$3B,$FE,$AF,$F0,$1F,$DF
	.dc.b	$BD,$DF,$EF,$ED,$8C,$9F,$D7,$23,$2F,$F8,$33,$EA,$BF,$C1,$E7,$F8
	.dc.b	$43,$EA,$AC,$64,$9A,$7F,$84,$23,$2F,$F4,$E1,$F5,$5F,$58,$FD,$16
	.dc.b	$7F,$E0,$EE,$BF,$EF,$31,$9D,$E9,$FE,$14,$8B,$87,$F9,$D2,$74,$FF
	.dc.b	$72,$8C,$DF,$F8,$83,$E2,$7F,$C4,$A7,$F8,$9B,$E2,$ED,$7F,$89,$A2
	.dc.b	$CF,$F8,$9A,$6E,$3F,$D7,$7E,$7F,$FC,$47,$F3,$FF,$E2,$46,$E1,$FE
	.dc.b	$00,$6D,$FF,$54,$E0,$46,$2D,$5F,$E9,$0F,$A4,$FF,$1F,$46,$2F,$FA
	.dc.b	$44,$62,$7F,$A6,$FD,$6F,$F8,$6A,$31,$3A,$EE,$C1,$FE,$40,$9B,$5F
	.dc.b	$E9,$6A,$DD,$1E,$37,$FE,$4A,$AD,$CF,$9F,$E7,$FF,$C7,$E7,$F9,$03
	.dc.b	$DF,$B8,$7F,$87,$22,$CF,$B6,$6D,$FF,$8C,$3D,$EF,$F2,$8F,$C1,$FF
	.dc.b	$94,$9B,$87,$F9,$49,$B7,$FE,$52,$6D,$7C,$18,$F3,$7C,$57,$F9,$8F
	.dc.b	$33,$49,$7F,$E7,$69,$DB,$EF,$8F,$F2,$B9,$F1,$C8,$E4,$FF,$1A,$B8
	.dc.b	$8F,$F3,$1F,$D2,$FF,$9A,$0F,$F3,$4F,$D5,$5A,$FF,$34,$C5,$9F,$F3
	.dc.b	$4C,$DC,$3F,$C4,$1E,$7F,$F9,$9B,$D1,$FF,$33,$B7,$0F,$F2,$1B,$6F
	.dc.b	$F0,$8D,$AF,$81,$F5,$2F,$5F,$FD,$04,$DB,$F7,$36,$FF,$D2,$AD,$7F
	.dc.b	$DB,$21,$D5,$74,$AF,$F4,$1E,$9F,$E6,$29,$CD,$FF,$B8,$1B,$7F,$E1
	.dc.b	$76,$DF,$FA,$3A,$B8,$5F,$F9,$AB,$E4,$7F,$A4,$1F,$CF,$3E,$19,$E5
	.dc.b	$E7,$C3,$FF,$37,$FC,$D6,$34,$FF,$73,$B7,$4F,$F1,$BD,$B7,$DE,$1C
	.dc.b	$38,$26,$1F,$E9,$CF,$A7,$FE,$2D,$87,$F0,$7C,$55,$FE,$7C,$F9,$7F
	.dc.b	$F5,$76,$FF,$89,$7F,$D2,$DF,$1B,$F9,$FC,$7F,$DE,$F1,$A7,$FD,$FF
	.dc.b	$F8,$47,$F9,$61,$BE,$77,$FA,$2E,$DF,$4F,$FE,$83,$CD,$7B,$7F,$C7
	.dc.b	$8B,$E0,$7F,$D1,$AD,$C3,$FD,$67,$F4,$EF,$7B,$35,$FE,$92,$FB,$47
	.dc.b	$C1,$53,$C1,$E8,$FF,$06,$C3,$FD,$0A,$7F,$92,$72,$57,$08,$CC,$7F
	.dc.b	$93,$7D,$65,$7F,$A7,$9B,$3F,$EC,$4F,$A7,$F4,$52,$72,$FF,$31,$5D
	.dc.b	$BC,$4F,$85,$2F,$89,$C5,$F8,$2F,$F8,$CF,$27,$FA,$B2,$3E,$99,$FE
	.dc.b	$D2,$E7,$72,$E6,$7F,$D6,$3F,$85,$BC,$BA,$7C,$FA,$FF,$09,$46,$6F
	.dc.b	$B7,$D9,$FF,$A2,$B9,$BF,$82,$36,$3F,$D5,$3F,$4E,$F6,$7F,$D7,$B3
	.dc.b	$9D,$3F,$CE,$1F,$4E,$F6,$FF,$D8,$3F,$4E,$FA,$8F,$F6,$1C,$E7,$2F
	.dc.b	$64,$E5,$F7,$7D,$43,$DE,$DE,$1F,$EC,$A8,$C4,$FC,$3B,$E5,$8F,$F5
	.dc.b	$BC,$FF,$1C,$FF,$A0,$6B,$6E,$7E,$0F,$96,$3F,$ED,$4F,$89,$FE,$6A
	.dc.b	$7F,$36,$FA,$77,$D4,$67,$D5,$FF,$6C,$C6,$34,$FA,$55,$BA,$D4,$7D
	.dc.b	$59,$D6,$3F,$DB,$5B,$C5,$FE,$FE,$F1,$BE,$34,$A7,$35,$F8,$FF,$A7
	.dc.b	$7C,$DF,$FB,$A1,$B1,$F9,$0C,$1F,$44,$FA,$8D,$5C,$AF,$C3,$5F,$0C
	.dc.b	$7F,$D6,$B7,$B9,$CD,$6E,$7F,$C0,$55,$BC,$3F,$B1,$1D,$B0,$D3,$FD
	.dc.b	$16,$1F,$35,$E5,$3F,$71,$F5,$0F,$94,$C6,$5E,$29,$B8,$FF,$B1,$AB
	.dc.b	$53,$93,$CF,$F2,$DE,$F3,$E0,$4D,$8C,$91,$F3,$99,$3F,$C3,$58,$BD
	.dc.b	$AF,$4C,$FE,$E6,$30,$FE,$A6,$7F,$D8,$9E,$7F,$FB,$17,$E9,$BF,$85
	.dc.b	$47,$FB,$1E,$B6,$F4,$BE,$6E,$5A,$FF,$05,$AE,$5E,$5F,$79,$CC,$F6
	.dc.b	$17,$D9,$5F,$68,$C4,$6D,$D4,$66,$BF,$A6,$CA,$FB,$BE,$3C,$2B,$E3
	.dc.b	$3B,$CF,$8E,$33,$1D,$63,$0E,$8F,$EB,$33,$CA,$BC,$11,$83,$47,$A3
	.dc.b	$0F,$F1,$36,$1E,$88,$F2,$1A,$9E,$2F,$8E,$3D,$73,$C2,$FD,$77,$DE
	.dc.b	$31,$5D,$CF,$F0,$64,$7F,$83,$5F,$94,$60,$F8,$7F,$4F,$FC,$6F,$EA
	.dc.b	$3F,$C2,$4D,$8C,$F8,$89,$CB,$63,$8F,$C2,$7D,$26,$E7,$CA,$31,$1F
	.dc.b	$DF,$2D,$18,$8F,$E8,$91,$89,$FA,$39,$35,$FE,$91,$A7,$DF,$8C,$5F
	.dc.b	$D6,$BE,$0F,$CC,$6C,$EE,$B8,$D7,$F5,$13,$8D,$6F,$0F,$45,$7E,$2E
	.dc.b	$F8,$D6,$EF,$8D,$8F,$EC,$CE,$14,$F0,$74,$D7,$67,$B2,$38,$46,$4D
	.dc.b	$4F,$92,$BD,$15,$ED,$BF,$2C,$F8,$E3,$FA,$C3,$EA,$C3,$FC,$9C,$DD
	.dc.b	$7E,$5D,$70,$AF,$44,$E5,$BE,$C7,$EE,$EB,$E2,$36,$37,$95,$5F,$51
	.dc.b	$FC,$0D,$FE,$D0,$D7,$FD,$40,$DC,$7F,$CC,$0F,$DE,$8D,$EF,$D6,$7F
	.dc.b	$97,$FC,$CD,$E7,$3D,$EF,$96,$5F,$E0,$7F,$A8,$76,$F8,$FE,$9F,$F7
	.dc.b	$73,$FE,$0B,$6E,$7F,$56,$B6,$FF,$7D,$6B,$FE,$B5,$8D,$A6,$C6,$B2
	.dc.b	$FF,$3C,$2F,$8F,$F0,$73,$CB,$E9,$1F,$16,$75,$1F,$E0,$BB,$FE,$76
	.dc.b	$DA,$D5,$48,$C1,$0F,$F6,$45,$72,$7F,$D9,$5F,$51,$C9,$FF,$66,$7C
	.dc.b	$1B,$E6,$7E,$A3,$63,$FC,$35,$3B,$CB,$FD,$21,$97,$CC,$9F,$26,$DF
	.dc.b	$24,$E3,$85,$8F,$F1,$17,$AB,$CD,$FF,$6D,$36,$F9,$D7,$E0,$FE,$A3
	.dc.b	$E5,$D7,$F9,$D9,$AF,$F8,$F5,$AF,$E9,$DA,$FC,$FE,$DF,$E2,$58,$FF
	.dc.b	$5D,$8F,$F5,$03,$5F,$E6,$AD,$7F,$D0,$7F,$01,$FE,$32,$7F,$D4,$CD
	.dc.b	$7F,$9D,$B5,$DF,$A7,$FC,$7F,$E9,$FF,$C5,$91,$FE,$AD,$6B,$FD,$09
	.dc.b	$AF,$FA,$41,$AF,$F9,$5E,$7F,$C1,$2F,$83,$C3,$FE,$55,$18,$8F,$0C
	.dc.b	$63,$E0,$E4,$79,$9A,$FF,$50,$F8,$77,$C4,$7F,$33,$6C,$7F,$B0,$5A
	.dc.b	$EE,$3B,$FD,$2C,$8F,$82,$D7,$E0,$FB,$DE,$42,$3F,$19,$FE,$81,$F4
	.dc.b	$8F,$9D,$A4,$9F,$F6,$84,$65,$F9,$BE,$E7,$B8,$8E,$0F,$A3,$F1,$1F
	.dc.b	$08,$8E,$0F,$F6,$D7,$D1,$FF,$90,$1B,$E8,$7F,$51,$8C,$47,$F7,$26
	.dc.b	$BF,$EC,$0F,$A3,$FF,$2B,$3F,$EE,$36,$BF,$DE,$1A,$EF,$D3,$FA,$7E
	.dc.b	$2C,$8F,$C2,$B5,$FE,$FF,$F4,$7F,$8D,$EA,$FF,$7A,$37,$C2,$FF,$4A
	.dc.b	$36,$7F,$9F,$6F,$FE,$6D,$7F,$D0,$D1,$F9,$37,$7D,$3F,$DA,$E3,$11
	.dc.b	$FD,$15,$A9,$E8,$FF,$A9,$BC,$8F,$E5,$95,$FC,$4A,$72,$FF,$B9,$E7
	.dc.b	$FD,$28,$D7,$FC,$3B,$7F,$EA,$98,$C6,$08,$C7,$67,$4E,$69,$7C,$9E
	.dc.b	$F1,$E0,$F8,$59,$3D,$9F,$E7,$DF,$47,$CE,$30,$43,$C2,$7C,$48,$DD
	.dc.b	$FF,$45,$FB,$1F,$F6,$0C,$FF,$8B,$03,$87,$FC,$5E,$D7,$FD,$FA,$D7
	.dc.b	$FD,$89,$F4,$8F,$91,$7B,$3C,$E3,$0D,$F8,$77,$9F,$9A,$DD,$FF,$8F
	.dc.b	$3E,$91,$EE,$7A,$E3,$FD,$A0,$23,$39,$F9,$DB,$1C,$67,$83,$FE,$59
	.dc.b	$F9,$FF,$D4,$57,$FA,$1D,$B1,$F1,$38,$51,$FC,$D6,$75,$DD,$FE,$F1
	.dc.b	$8D,$CF,$FA,$09,$B3,$D7,$C9,$D5,$7F,$BD,$FE,$5F,$FD,$5F,$66,$FD
	.dc.b	$27,$C4,$BF,$F4,$C4,$68,$99,$7F,$BF,$E7,$59,$79,$67,$FD,$72,$BE
	.dc.b	$8D,$F3,$7F,$E7,$2C,$5F,$5F,$FE,$9B,$AF,$BE,$6E,$75,$19,$9F,$99
	.dc.b	$5C,$B2,$FC,$03,$6F,$FC,$DA,$D7,$E8,$B5,$FF,$15,$1D,$63,$FD,$E5
	.dc.b	$1C,$67,$E3,$7D,$2F,$FA,$12,$31,$3F,$E8,$BC,$5F,$4B,$F1,$CF,$91
	.dc.b	$EB,$59,$F8,$B1,$71,$F8,$53,$71,$FA,$5F,$23,$F5,$7E,$3F,$EB,$37
	.dc.b	$0F,$BC,$DB,$FB,$DC,$48,$C5,$FE,$37,$E9,$1F,$2F,$F2,$FE,$C7,$FD
	.dc.b	$F5,$8F,$FA,$84,$F7,$CE,$9F,$F6,$84,$6B,$1F,$9F,$39,$7F,$D2,$AD
	.dc.b	$7F,$32,$D7,$9D,$7E,$59,$FF,$2A,$47,$1B,$F2,$4F,$E2,$FE,$9B,$FD
	.dc.b	$5B,$39,$C3,$F3,$9F,$1D,$F9,$5F,$8D,$FC,$B3,$63,$F3,$CD,$C7,$E0
	.dc.b	$B7,$1F,$F6,$DF,$03,$C9,$FC,$C1,$F9,$9F,$66,$E1,$F9,$68,$B8,$7E
	.dc.b	$4E,$6D,$FC,$9F,$3B,$F1,$BE,$8F,$99,$B8,$7C,$9B,$F6,$C6,$1F,$8E
	.dc.b	$DA,$E5,$3F,$EF,$BE,$75,$F3,$8C,$36,$FF,$71,$3A,$9E,$D0,$D5,$87
	.dc.b	$59,$FB,$2D,$9F,$1C,$74,$38,$CF,$07,$11,$E6,$6E,$1E,$68,$F0,$E1
	.dc.b	$E3,$8F,$E6,$07,$EC,$9C,$4F,$FB,$A7,$0F,$35,$73,$38,$C7,$83,$0F
	.dc.b	$24,$7A,$5B,$87,$8E,$36,$71,$C3,$BD,$F4,$7C,$86,$30,$F0,$CF,$FB
	.dc.b	$B2,$75,$87,$C2,$3D,$65,$8E,$B5,$CA,$3D,$15,$CA,$3D,$91,$E7,$38
	.dc.b	$CF,$CA,$38,$BC,$E3,$C4,$EE,$32,$6E,$32,$FA,$E3,$B9,$B8,$C9,$E3
	.dc.b	$8F,$54,$F8,$EB,$D7,$1C,$EB,$8B,$CC,$DB,$CC,$CD,$73,$3D,$71,$B6
	.dc.b	$99,$9E,$B1,$F4,$E7,$DE,$D3,$D6,$59,$F3,$C7,$C1,$3B,$96,$3C,$67
	.dc.b	$79,$EB,$7F,$76,$B8,$CF,$CB,$3F,$83,$E1,$FC,$15,$F0,$18,$BF,$EE
	.dc.b	$D5,$F8,$58,$C1,$FC,$42,$78,$46,$63,$C4,$79,$E7,$FB,$27,$D1,$FE
	.dc.b	$1A,$30,$7C,$B1,$7F,$D9,$23,$55,$E7,$AF,$64,$E5,$F9,$D5,$C4,$E7
	.dc.b	$19,$7E,$A1,$D8,$CF,$1A,$FA,$E1,$F8,$F8,$C5,$F4,$AE,$18,$78,$5F
	.dc.b	$2C,$E9,$EB,$19,$7A,$DF,$4A,$F2,$E1,$E3,$3F,$19,$3D,$30,$FE,$FD
	.dc.b	$5B,$8F,$C2,$61,$D6,$B7,$87,$53,$C4,$F3,$C3,$E6,$C6,$70,$E7,$1E
	.dc.b	$98,$C3,$FE,$11,$C3,$B5,$7E,$4E,$FB,$4F,$2B,$F8,$55,$A8,$F9,$F3
	.dc.b	$AC,$3E,$76,$1D,$1F,$EB,$B5,$B7,$ED,$61,$E5,$77,$87,$4C,$3C,$8D
	.dc.b	$8F,$92,$D9,$F7,$B6,$7C,$87,$CA,$C3,$C5,$5F,$7B,$57,$D3,$7F,$88
	.dc.b	$7D,$FF,$F1,$13,$6F,$FC,$3E,$DA,$FF,$0E,$37,$1F,$F1,$67,$D3,$7F
	.dc.b	$8B,$CF,$F1,$87,$D3,$5B,$FF,$1C,$45,$BF,$EF,$73,$6B,$FC,$8D,$F4
	.dc.b	$DF,$15,$FF,$18,$7A,$BF,$CD,$30,$FF,$0B,$CE,$73,$4E,$70,$09,$FC
	.dc.b	$63,$FE,$50,$6B,$CB,$6F,$DA,$46,$23,$8B,$F1,$0E,$DD,$FD,$63,$A4
	.dc.b	$FF,$92,$FE,$1E,$63,$FE,$53,$F8,$AB,$5F,$E1,$69,$B9,$FF,$93,$E6
	.dc.b	$D7,$F8,$A7,$E9,$7F,$C9,$DF,$4D,$FE,$4F,$6C,$7F,$93,$1B,$87,$F9
	.dc.b	$25,$B7,$FE,$6B,$FA,$37,$83,$FE,$44,$F8,$8F,$34,$63,$0F,$EE,$51
	.dc.b	$89,$FE,$25,$39,$C3,$FC,$E6,$1C,$3F,$E7,$66,$BF,$E4,$67,$85,$F8
	.dc.b	$67,$FC,$D9,$F4,$DD,$70,$FF,$31,$FD,$37,$F9,$A4,$FF,$35,$7C,$05
	.dc.b	$8C,$93,$1F,$F3,$24,$65,$FF,$1E,$FC,$0F,$F9,$DF,$E3,$7F,$CF,$0D
	.dc.b	$AF,$F4,$03,$67,$71,$FE,$83,$9F,$83,$F1,$5F,$E7,$9F,$B1,$B3,$FE
	.dc.b	$76,$6D,$7C,$0D,$55,$F1,$09,$5F,$EA,$98,$CB,$FE,$14,$6C,$FF,$9E
	.dc.b	$E7,$9E,$BF,$E8,$BF,$9F,$FF,$47,$1F,$E8,$FF,$A6,$B8,$7F,$A5,$E2
	.dc.b	$DF,$F9,$6A,$6D,$7F,$8E,$BD,$6F,$F4,$87,$AB,$FE,$91,$6E,$1F,$E8
	.dc.b	$E6,$DF,$FA,$2D,$B5,$FE,$AD,$FA,$77,$BD,$FE,$86,$F7,$FF,$31,$18
	.dc.b	$C3,$FC,$69,$18,$9F,$EA,$93,$9C,$3F,$06,$D7,$FD,$6E,$D7,$FD,$0E
	.dc.b	$EE,$FC,$13,$FE,$AD,$D7,$A6,$1F,$E3,$AF,$67,$FA,$C6,$1F,$B4,$6B
	.dc.b	$FE,$C8,$F1,$7F,$B2,$46,$3E,$69,$E9,$7F,$67,$6D,$FF,$B6,$DB,$7F
	.dc.b	$98,$6B,$FE,$CC,$5C,$1C,$49,$3C,$5F,$EE,$5F,$2D,$FE,$CE,$AF,$F6
	.dc.b	$9B,$5F,$F5,$1B,$5F,$E1,$6D,$5F,$4B,$FE,$81,$7F,$87,$B5,$CF,$A5
	.dc.b	$FD,$FE,$30,$3E,$13,$04,$C3,$FD,$ED,$19,$7F,$DF,$1F,$48,$E2,$FF
	.dc.b	$04,$36,$33,$18,$3F,$C1,$9E,$B3,$D1,$A4,$9F,$DE,$E5,$FE,$A7,$8D
	.dc.b	$F2,$C9,$CD,$23,$15,$FC,$12,$7E,$59,$C8,$AE,$75,$FF,$0E,$46,$23
	.dc.b	$FD,$FB,$F4,$6F,$13,$FC,$40,$D9,$F1,$B6,$7F,$88,$B5,$FF,$69,$8F
	.dc.b	$A2,$7A,$FF,$E2,$96,$D7,$B1,$B3,$FC,$69,$AF,$2F,$A1,$7D,$17,$F3
	.dc.b	$96,$D7,$FA,$3B,$E8,$9E,$CE,$6B,$F9,$F7,$D1,$7F,$AA,$EB,$53,$8F
	.dc.b	$6B,$FA,$15,$F0,$9F,$F4,$BD,$6F,$1F,$F7,$BB,$7B,$93,$FE,$EE,$F6
	.dc.b	$FF,$E0,$0C,$E6,$3F,$DE,$3F,$0B,$FC,$B1,$BE,$5F,$F4,$FF,$A2,$FE
	.dc.b	$BB,$1F,$EA,$49,$D6,$3F,$D7,$A7,$2F,$F8,$FA,$78,$9F,$EA,$A8,$CE
	.dc.b	$3E,$29,$FE,$CB,$5C,$94,$EA,$BF,$B4,$C6,$3D,$4F,$ED,$96,$EE,$F9
	.dc.b	$EF,$39,$D5,$7F,$AE,$3E,$85,$F2,$9F,$D9,$DB,$1F,$EC,$9A,$7C,$EB
	.dc.b	$E8,$B5,$72,$B9,$3E,$09,$D4,$60,$98,$7B,$CF,$EF,$F8,$76,$71,$87
	.dc.b	$6C,$3A,$46,$C4,$BD,$2F,$31,$EE,$FA,$27,$C9,$63,$1F,$64,$DB,$EB
	.dc.b	$5A,$9C,$9F,$07,$E4,$FE,$01,$F0,$66,$C6,$48,$F8,$F0,$FF,$0B,$C6
	.dc.b	$32,$F9,$83,$D1,$FF,$08,$36,$3F,$DC,$B5,$FC,$50,$7D,$07,$F0,$D7
	.dc.b	$FC,$43,$3B,$7E,$03,$4E,$B5,$FE,$1B,$8E,$53,$C5,$FF,$43,$C6,$57
	.dc.b	$91,$FE,$2C,$9D,$57,$57,$84,$F2,$F1,$FF,$13,$39,$7F,$C1,$B5,$C5
	.dc.b	$CD,$FE,$4E,$F8,$DF,$63,$D7,$3A,$C3,$84,$7F,$36,$32,$7B,$AB,$94
	.dc.b	$F6,$8C,$9E,$A9,$E2,$FF,$3C,$33,$87,$95,$E9,$7E,$CB,$E7,$18,$BF
	.dc.b	$13,$D0,$F7,$1E,$1B,$EA,$FD,$33,$F9,$7E,$1F,$CC,$1C,$1E,$53,$F8
	.dc.b	$EF,$D0,$FF,$25,$FA,$2F,$E6,$AD,$8F,$F1,$9B,$5D,$F8,$49,$DF,$B7
	.dc.b	$FE,$BC,$9D,$B5,$EA,$D7,$F7,$74,$8C,$47,$F9,$9A,$31,$A7,$C6,$C9
	.dc.b	$B7,$F0,$2D,$7F,$09,$18,$BF,$97,$7C,$1F,$9E,$D9,$DD,$71,$AF,$E1
	.dc.b	$67,$1A,$DE,$1E,$9A,$FC,$85,$F1,$AD,$DF,$1B,$FF,$3D,$C7,$1A,$DD
	.dc.b	$F1,$8F,$6B,$67,$75,$C5,$FE,$E0,$DB,$ED,$1F,$06,$78,$3A,$3E,$14
	.dc.b	$EE,$33,$5E,$9C,$3F,$90,$37,$4F,$3D,$FE,$C1,$F1,$CF,$28,$C1,$EA
	.dc.b	$F3,$7F,$4B,$3F,$11,$B1,$E8,$36,$3E,$8B,$F8,$13,$86,$BF,$C6,$DB
	.dc.b	$5F,$CC,$DF,$BD,$1B,$DF,$AC,$FF,$30,$F9,$6B,$CC,$7C,$35,$DE,$AF
	.dc.b	$D1,$7F,$A3,$23,$C7,$F4,$1F,$C0,$33,$FE,$8A,$D8,$FF,$57,$CE,$DF
	.dc.b	$E5,$CD,$C7,$FA,$34,$EE,$4D,$9D,$57,$F5,$05,$F0,$FE,$49,$FE,$6D
	.dc.b	$5E,$2E,$7F,$2C,$C8,$C1,$0F,$0B,$97,$FA,$04,$F2,$6C,$E7,$0F,$EB
	.dc.b	$5F,$45,$97,$FA,$E7,$B4,$F9,$4F,$8A,$DB,$FA,$B5,$BC,$FF,$B3,$CE
	.dc.b	$67,$BE,$BE,$93,$8D,$D8,$FB,$BE,$86,$F1,$FE,$D4,$DC,$3E,$6D,$FD
	.dc.b	$4F,$A2,$F8,$B7,$FE,$DC,$6A,$6E,$2F,$73,$E4,$77,$4A,$FB,$2F,$3F
	.dc.b	$8A,$FB,$FF,$40,$FA,$2A,$4A,$FC,$8F,$D6,$3E,$27,$F0,$BF,$46,$F3
	.dc.b	$69,$2B,$F2,$FD,$0F,$3F,$AF,$91,$27,$F3,$79,$FF,$81,$1F,$F0,$17
	.dc.b	$D6,$3E,$BB,$C7,$F0,$52,$76,$FE,$62,$5E,$FF,$E5,$BE,$19,$F0,$F4
	.dc.b	$97,$FA,$8F,$AD,$F3,$3E,$0F,$95,$F7,$91,$F0,$F2,$FC,$2F,$81,$91
	.dc.b	$E6,$7C,$1C,$BF,$A4,$FA,$67,$CE,$D2,$57,$ED,$BE,$B7,$FC,$34,$FB
	.dc.b	$BE,$9F,$E6,$91,$F7,$FC,$F7,$D0,$F8,$A9,$1F,$11,$F7,$72,$FE,$CB
	.dc.b	$E9,$DE,$D5,$25,$7F,$04,$FA,$77,$CA,$7E,$E1,$FA,$11,$9E,$AC,$1F
	.dc.b	$27,$C4,$FB,$FE,$9C,$9F,$A6,$FC,$DE,$AF,$E0,$3F,$4F,$FE,$20,$9F
	.dc.b	$E2,$2D,$7F,$C5,$0D,$F2,$7F,$C8,$1C,$BF,$E6,$C9,$DF,$37,$22,$3F
	.dc.b	$C6,$A3,$4F,$28,$FE,$C9,$5D,$63,$FC,$FE,$D7,$FC,$A2,$D4,$7F,$A8
	.dc.b	$EF,$F8,$BD,$7F,$B7,$A3,$B4,$63,$1C,$B2,$32,$78,$E3,$8F,$D3,$48
	.dc.b	$F3,$3F,$C8,$BE,$9F,$FC,$63,$E5,$7F,$98,$B0,$4E,$EB,$FC,$95,$18
	.dc.b	$7F,$9B,$D7,$07,$CF,$F4,$FF,$C3,$DF,$F4,$DB,$5F,$F3,$6B,$5D,$F9
	.dc.b	$FF,$E0,$78,$FF,$60,$47,$FA,$E1,$AF,$FA,$90,$38,$7A,$D7,$F8,$3F
	.dc.b	$E9,$DF,$2A,$F8,$4F,$F6,$0C,$61,$BF,$1E,$F6,$73,$5B,$C3,$FD,$63
	.dc.b	$C4,$F9,$FF,$D0,$EF,$FE,$79,$6C,$FF,$9B,$9F,$E4,$D1,$9E,$FF,$F4
	.dc.b	$3C,$6A,$BF,$D2,$07,$38,$CF,$A9,$A7,$FD,$1B,$F3,$AF,$A6,$A7,$FA
	.dc.b	$43,$C0,$F5,$63,$FE,$89,$9D,$36,$76,$FF,$A3,$67,$79,$7F,$A0,$AB
	.dc.b	$6F,$FA,$07,$E9,$7F,$8E,$57,$FA,$4F,$E8,$FF,$D1,$B3,$9F,$4B,$FD
	.dc.b	$3F,$18,$D3,$FD,$3D,$7C,$2B,$FD,$1B,$39,$BF,$F5,$3C,$E6,$BD,$F3
	.dc.b	$9A,$FC,$CC,$E6,$BF,$CE,$B3,$99,$4E,$55,$27,$92,$C1,$F4,$BF,$20
	.dc.b	$F9,$3F,$19,$71,$F8,$D1,$70,$F8,$53,$6F,$E9,$EE,$F9,$6F,$AF,$F3
	.dc.b	$3F,$61,$B8,$FD,$F6,$E1,$F7,$DB,$7F,$EC,$1F,$A4,$FE,$6B,$5F,$EB
	.dc.b	$DF,$A5,$FE,$8C,$FF,$86,$A3,$37,$FE,$B4,$9C,$BF,$EB,$38,$CD,$7E
	.dc.b	$1A,$72,$F5,$F4,$BF,$DA,$31,$8B,$FF,$6A,$D7,$07,$FD,$3F,$19,$BF
	.dc.b	$E0,$F3,$97,$E5,$C6,$6F,$F3,$9F,$27,$F9,$6F,$63,$F2,$ED,$C3,$F3
	.dc.b	$CD,$BF,$CF,$36,$BF,$37,$BF,$FB,$E6,$7E,$BB,$63,$EB,$7B,$FF,$9C
	.dc.b	$F1,$E6,$1F,$99,$7F,$3F,$18,$BF,$85,$E6,$FE,$43,$D2,$FD,$3D,$FB
	.dc.b	$A3,$0F,$CB,$6C,$F0,$8F,$CD,$4F,$08,$F8,$33,$AC,$3E,$B3,$6B,$A1
	.dc.b	$C0,$E8,$59,$E8,$7E,$7A,$BA,$2D,$55,$CE,$3B,$D7,$1B,$EE,$D3,$BC
	.dc.b	$73,$7B,$B4,$EF,$1F,$A8,$BE,$F7,$DA,$B9,$57,$63,$85,$76,$6C,$76
	.dc.b	$8E,$4F,$66,$C7,$68,$FD,$8D,$76,$BF,$54,$FB,$5B,$7E,$18,$EA,$78
	.dc.b	$6F,$99,$E1,$2D,$F8,$4E,$A7,$84,$B1,$E1,$3F,$7B,$C8,$C3,$C5,$3E
	.dc.b	$E9,$F1,$36,$FC,$57,$DA,$7C,$53,$D2,$3C,$4D,$3C,$51,$DA,$BA,$BE
	.dc.b	$A3,$A9,$6B,$AC,$F1,$8E,$A7,$17,$D4,$D3,$D5,$5F,$C5,$6B,$AB,$E5
	.dc.b	$8E,$AD,$AE,$B1,$C6,$3C,$ED,$AF,$3C,$73,$7B,$B6,$3B,$DF,$C7,$9C
	.dc.b	$B6,$33,$7E,$29,$F5,$47,$89,$B7,$E2,$BE,$87,$88,$B1,$E2,$BE,$91
	.dc.b	$96,$C6,$6F,$B5,$FA,$4F,$8F,$5C,$9B,$5C,$1B,$3C,$27,$51,$E5,$39
	.dc.b	$15,$CD,$F9,$EB,$94,$E5,$B1,$9A,$E9,$18,$C3,$F3,$24,$C3,$F1,$4D
	.dc.b	$BF,$C0,$37,$0F,$AC,$D9,$F8,$ED,$9F,$2E,$1D,$E7,$E8,$E1,$E0,$23
	.dc.b	$82,$4F,$D4,$6D,$FC,$B9,$DC,$E4,$91,$99,$D9,$F3,$CB,$1F,$29,$AF
	.dc.b	$C3,$9E,$73,$C6,$3F,$C0,$1E,$E7,$F8,$33,$E0,$BF,$C1,$AD,$BF,$F0
	.dc.b	$53,$67,$71,$FE,$0D,$9F,$F0,$CF,$D2,$C8,$FF,$10,$7D,$2D,$AF,$F1
	.dc.b	$04,$DC,$3F,$B5,$CD,$AF,$F2,$BD,$3E,$95,$F4,$BF,$E1,$EF,$4B,$F9
	.dc.b	$5C,$62,$BF,$C3,$91,$89,$D4,$7F,$98,$30,$3D,$2B,$85,$FF,$87,$6F
	.dc.b	$84,$F2,$38,$1B,$BE,$73,$FE,$37,$F8,$BF,$F1,$A9,$FE,$36,$F8,$FB
	.dc.b	$5F,$E3,$D8,$B3,$FE,$3D,$9B,$8F,$F6,$0F,$A5,$FF,$1B,$FA,$BF,$E3
	.dc.b	$86,$E1,$FE,$35,$6D,$FF,$8C,$5B,$5E,$A8,$FF,$3D,$FD,$0B,$BF,$FC
	.dc.b	$5F,$E7,$7F,$4E,$8C,$61,$FE,$35,$8C,$4E,$A3,$DB,$5C,$A3,$0F,$F8
	.dc.b	$A2,$3E,$8D,$72,$8F,$F3,$43,$5F,$AA,$1C,$3C,$27,$CC,$FD,$AB,$FF
	.dc.b	$18,$7D,$27,$F9,$58,$FF,$2B,$FB,$F7,$0F,$F2,$A4,$5A,$FF,$28,$4D
	.dc.b	$AF,$F1,$7F,$C0,$7F,$99,$3E,$13,$FC,$CA,$DC,$3F,$CD,$8D,$BF,$F3
	.dc.b	$63,$6B,$C7,$93,$E2,$3F,$CC,$FF,$49,$F1,$DB,$87,$C4,$C3,$D9,$18
	.dc.b	$7E,$21,$F4,$1C,$5F,$47,$DF,$1F,$21,$AE,$E7,$FD,$20,$F4,$7F,$D2
	.dc.b	$53,$FE,$85,$FA,$1F,$F3,$F9,$FE,$80,$F8,$AB,$3F,$E8,$78,$B5,$FE
	.dc.b	$87,$9B,$5F,$E2,$AF,$4B,$FD,$03,$E8,$FF,$A0,$9B,$87,$F9,$FD,$B7
	.dc.b	$FE,$7B,$6D,$7D,$6F,$A4,$7B,$7F,$E7,$8F,$5F,$FC,$2F,$18,$C3,$FC
	.dc.b	$FD,$18,$9D,$47,$C7,$9D,$61,$FE,$7C,$AD,$4E,$6B,$FD,$37,$B6,$2B
	.dc.b	$E5,$9F,$EA,$9F,$0B,$FC,$61,$8B,$E2,$5F,$4D,$FE,$28,$F5,$FF,$A8
	.dc.b	$46,$32,$FA,$DF,$36,$F4,$38,$CB,$97,$FC,$5F,$39,$AF,$CF,$E3,$D6
	.dc.b	$73,$C7,$E0,$FA,$C7,$8B,$8B,$DC,$61,$FC,$27,$CC,$FF,$A8,$B0,$FF
	.dc.b	$59,$FC,$53,$0F,$F1,$EC,$E6,$BF,$88,$4E,$5F,$C6,$7D,$37,$F9,$AE
	.dc.b	$DF,$58,$F1,$BF,$C8,$B3,$98,$C5,$70,$F9,$8F,$8B,$1F,$C6,$A3,$07
	.dc.b	$B1,$C7,$27,$CC,$FA,$77,$47,$CC,$F4,$1E,$37,$F9,$4A,$73,$87,$C4
	.dc.b	$9C,$BF,$A1,$F5,$D1,$F5,$3E,$AD,$F2,$5F,$E5,$A9,$CD,$7C,$C9,$CB
	.dc.b	$C7,$BF,$87,$D6,$ED,$B1,$FA,$D3,$E2,$69,$FA,$CE,$1F,$B5,$F1,$5F
	.dc.b	$E6,$69,$CC,$7F,$B8,$3A,$D3,$FE,$6A,$FA,$97,$9B,$F8,$89,$71,$E6
	.dc.b	$30,$FF,$9C,$3E,$A1,$F5,$3F,$8E,$FA,$CF,$F5,$C6,$7F,$BF,$9C,$BF
	.dc.b	$86,$8C,$E9,$F9,$6F,$AA,$7D,$57,$F9,$E9,$B5,$FC,$1F,$EA,$9E,$FE
	.dc.b	$6B,$FC,$FF,$18,$9D,$4F,$FA,$0B,$EA,$BF,$D4,$18,$2B,$4F,$E9,$23
	.dc.b	$26,$BE,$07,$F1,$90,$11,$CB,$E3,$3F,$9B,$71,$FF,$30,$6E,$3F,$DA
	.dc.b	$BE,$AB,$F1,$77,$FB,$09,$CB,$CE,$33,$5C,$7D,$6F,$F4,$BC,$62,$FF
	.dc.b	$B9,$DF,$0B,$4E,$5E,$5E,$0F,$EE,$BC,$8F,$97,$1F,$BE,$9D,$2F,$A8
	.dc.b	$7C,$C7,$FA,$91,$B8,$FE,$DB,$E9,$3F,$DF,$91,$8A,$FE,$0E,$DD,$3F
	.dc.b	$A6,$B6,$BF,$BF,$57,$0B,$E8,$48,$E9,$F5,$57,$0E,$7E,$DF,$84,$FE
	.dc.b	$23,$F5,$5F,$8E,$AF,$E8,$58,$A7,$51,$87,$7D,$B8,$BF,$2F,$CC,$FA
	.dc.b	$CF,$63,$6B,$9C,$58,$F5,$97,$4F,$EC,$E3,$DB,$FF,$63,$B6,$BF,$AA
	.dc.b	$4F,$E0,$7C,$CF,$C1,$7D,$4F,$E8,$23,$FC,$3D,$7C,$09,$5C,$21,$5C
	.dc.b	$3B,$BF,$BA,$57,$FB,$52,$7A,$0B,$E8,$B0,$F1,$39,$AF,$F0,$0B,$D0
	.dc.b	$C1,$E2,$C3,$B3,$FE,$06,$C2,$BC,$F0,$FF,$0D,$61,$DA,$3C,$2F,$83
	.dc.b	$0E,$EE,$63,$B3,$99,$E3,$87,$8A,$FA,$DF,$08,$C5,$F9,$0F,$06,$1E
	.dc.b	$27,$BC,$7C,$28,$C4,$76,$8C,$3C,$4D,$9E,$D3,$07,$C1,$3F,$BE,$D7
	.dc.b	$F7,$E7,$E5,$98,$3E,$3F,$D4,$FF,$18,$FA,$AF,$F0,$1B,$63,$3E,$0A
	.dc.b	$72,$D8,$E3,$F0,$5F,$4D,$B9,$72,$8C,$47,$F8,$D2,$D1,$88,$FF,$35
	.dc.b	$C6,$27,$E9,$E2,$D7,$FC,$DF,$AF,$E0,$23,$17,$F6,$2F,$83,$F2,$DB
	.dc.b	$3B,$AE,$35,$FE,$78,$38,$D6,$F0,$F4,$57,$E3,$2F,$8D,$6E,$F8,$DF
	.dc.b	$F1,$D8,$E3,$5B,$BE,$31,$EA,$6C,$EE,$B8,$BF,$D0,$9B,$7D,$A3,$F9
	.dc.b	$43,$5D,$1F,$0A,$37,$3D,$23,$27,$F8,$5A,$BD,$78,$7F,$97,$1B,$A7
	.dc.b	$8E,$39,$4F,$F9,$26,$5E,$B7,$F9,$86,$BE,$23,$67,$D0,$7F,$93,$FE
	.dc.b	$AB,$F8,$03,$77,$D2,$8C,$BF,$E2,$87,$EE,$BB,$3F,$94,$7C,$E7,$5C
	.dc.b	$5C,$E8,$E7,$5C,$A7,$FC,$EB,$F5,$5F,$E8,$9B,$FF,$30,$7D,$57,$EC
	.dc.b	$73,$FF,$58,$B6,$3F,$A8,$4E,$DF,$F1,$CB,$6B,$FC,$FF,$3B,$93,$67
	.dc.b	$55,$FE,$BC,$7F,$C5,$F3,$E0,$E5,$F7,$D7,$F2,$F8,$FF,$39,$4E,$9E
	.dc.b	$FE,$C6,$AA,$46,$A3,$E3,$CE,$F0,$F1,$D6,$AB,$89,$FE,$7D,$FA,$AE
	.dc.b	$CF,$F9,$4B,$E0,$DF,$37,$F4,$1B,$1B,$AC,$4F,$9B,$59,$3F,$E8,$7A
	.dc.b	$DB,$FE,$5C,$CF,$E8,$3F,$E8,$BF,$4F,$39,$7F,$A9,$1B,$AF,$CE,$C3
	.dc.b	$EF,$FD,$57,$C6,$BF,$EF,$CD,$4D,$C5,$F0,$BF,$23,$B2,$57,$DA,$79
	.dc.b	$FC,$77,$E2,$BE,$01,$EE,$D2,$57,$FA,$7F,$EA,$9F,$35,$F8,$9C,$1E
	.dc.b	$4D,$25,$7F,$9A,$A1,$F5,$FF,$8F,$FD,$93,$D1,$C4,$69,$3E,$7B,$5E
	.dc.b	$B4,$C3,$FC,$E7,$3F,$E3,$46,$FC,$E7,$3F,$98,$E9,$CD,$27,$71,$DF
	.dc.b	$87,$C1,$D9,$27,$9C,$78,$F7,$F2,$79,$F2,$7B,$C7,$9F,$E4,$64,$78
	.dc.b	$9F,$57,$C1,$7A,$FE,$46,$4F,$99,$F7,$7C,$8F,$BF,$A6,$4F,$A9,$F8
	.dc.b	$5F,$23,$F0,$FE,$03,$E2,$7C,$8C,$AF,$7C,$7C,$7F,$B2,$F9,$1F,$65
	.dc.b	$27,$E1,$C7,$EE,$7E,$B5,$F4,$DF,$37,$C3,$95,$FE,$9A,$9C,$BF,$ED
	.dc.b	$7F,$D0,$1F,$49,$FC,$21,$B1,$98,$C7,$8F,$96,$46,$4E,$51,$87,$FD
	.dc.b	$6B,$3C,$9B,$3C,$9F,$F7,$37,$D4,$7F,$9D,$1F,$F6,$EB,$5F,$F0,$2B
	.dc.b	$5F,$F2,$C1,$FE,$59,$70,$E1,$E1,$DC,$C7,$7F,$1D,$FC,$2C,$FE,$19
	.dc.b	$3A,$8C,$FC,$BF,$F2,$69,$FF,$6C,$DF,$FB,$E5,$AF,$FB,$39,$FD,$9B
	.dc.b	$63,$81,$1E,$1E,$4F,$3D,$64,$73,$7F,$99,$FD,$47,$F8,$DB,$0F,$2C
	.dc.b	$6A,$7B,$3A,$8C,$65,$EB,$9F,$F3,$C6,$3F,$61,$D9,$FE,$7B,$8D,$2B
	.dc.b	$EC,$FF,$9F,$FE,$9D,$C7,$FB,$48,$FF,$18,$C6,$3C,$6F,$F1,$9C,$21
	.dc.b	$C9,$FD,$4A,$37,$5C,$5F,$A6,$3E,$C5,$E2,$F8,$A3,$EB,$C6,$34,$EA
	.dc.b	$D8,$FF,$20,$B5,$FB,$B9,$BE,$9B,$F2,$75,$FE,$F0,$6C,$FF,$AF,$DF
	.dc.b	$E5,$F1,$9E,$2F,$F8,$03,$1A,$AF,$90,$73,$8C,$F2,$69,$CD,$7C,$9F
	.dc.b	$A4,$FF,$52,$D1,$BF,$51,$FC,$6A,$7F,$E0,$0C,$68,$99,$F2,$CB,$FC
	.dc.b	$BD,$5B,$7F,$DA,$93,$AA,$E3,$5B,$B5,$ED,$F3,$D7,$FA,$DA,$DF,$51
	.dc.b	$FC,$B3,$0F,$AA,$D8,$FF,$34,$B5,$FA,$6D,$7F,$4E,$EE,$BF,$0A,$D9
	.dc.b	$FC,$46,$6F,$9C,$F5,$5F,$83,$C2,$F3,$62,$F2,$29,$8F,$88,$B8,$FC
	.dc.b	$48,$B7,$F0,$26,$E7,$F4,$B2,$FF,$71,$B7,$E4,$7E,$BF,$CA,$FD,$86
	.dc.b	$E3,$F7,$DB,$87,$DF,$6D,$71,$9F,$C2,$57,$17,$F4,$47,$F0,$6F,$A5
	.dc.b	$7D,$47,$B1,$B1,$F6,$BC,$FC,$4C,$8E,$38,$FF,$B9,$6B,$6F,$DA,$AD
	.dc.b	$3F,$91,$F9,$0F,$C7,$FC,$7F,$E4,$1B,$7F,$97,$6C,$EE,$3F,$35,$3F
	.dc.b	$33,$CF,$91,$F8,$CF,$42,$D7,$E2,$26,$E9,$F8,$49,$B5,$F9,$79,$B5
	.dc.b	$FE,$AD,$E1,$56,$2F,$F3,$55,$67,$37,$F9,$A8,$C1,$F7,$AA,$C6,$4F
	.dc.b	$21,$6F,$C8,$DB,$F2,$34,$F2,$16,$3C,$93,$D1,$E2,$7E,$91,$C4,$FE
	.dc.b	$82,$BC,$55,$C6,$3D,$2D,$BF,$4D,$F2,$9E,$13,$9B,$E6,$F8,$A7,$B3
	.dc.b	$C6,$FE,$14,$EA,$7E,$44,$78,$1E,$93,$C6,$BA,$C7,$23,$27,$08,$F0
	.dc.b	$1D,$4A,$EA,$3D,$2D,$BF,$4C,$74,$24,$F7,$3F,$89,$5F,$96,$38,$9F
	.dc.b	$3A,$B9,$5F,$AA,$BC,$11,$D6,$B8,$5A,$BD,$97,$DA,$BD,$06,$EF,$8C
	.dc.b	$64,$FE,$41,$7C,$AB,$8C,$F9,$6B,$8D,$6E,$FD,$73,$98,$F3,$B6,$33
	.dc.b	$3C,$23,$15,$D2,$30,$4A,$E2,$FE,$22,$7C,$A7,$3B,$F2,$3F,$94,$8F
	.dc.b	$86,$FD,$09,$F2,$57,$C5,$9C,$9E,$F2,$C7,$A6,$B9,$C7,$DD,$8E,$35
	.dc.b	$CC,$F3,$96,$3C,$D7,$F3,$63,$13,$E8,$24,$66,$7A,$4F,$27,$C5,$3D
	.dc.b	$9B,$19,$8F,$23,$F0,$23,$BD,$79,$A7,$F0,$47,$13,$71,$EC,$9F,$8C
	.dc.b	$71,$8F,$1C,$FF,$06,$96,$6F,$9F,$FE,$48,$7F,$68,$FA,$2B,$87,$F2
	.dc.b	$2A,$DC,$E1,$FE,$E1,$37,$0F,$EA,$FE,$8F,$F4,$6F,$4B,$FA,$3B,$70
	.dc.b	$FE,$CE,$DB,$FE,$C6,$DA,$FE,$89,$F4,$BF,$DE,$0F,$EF,$3F,$4B,$6F
	.dc.b	$FC,$01,$16,$FF,$C0,$13,$6B,$8D,$68,$FF,$03,$46,$5F,$F2,$3F,$D2
	.dc.b	$B2,$FF,$0A,$58,$46,$67,$FC,$42,$D8,$FF,$0E,$9A,$9E,$2D,$9F,$8D
	.dc.b	$3B,$BF,$EF,$3D,$3F,$DA,$A3,$13,$FD,$BB,$E1,$FF,$C2,$A7,$F8,$5B
	.dc.b	$E2,$EC,$7F,$87,$62,$E1,$FE,$1D,$9B,$87,$F8,$13,$D6,$FE,$E7,$EB
	.dc.b	$7F,$74,$6E,$1F,$E1,$46,$DF,$F8,$41,$B5,$C6,$7F,$C3,$35,$C5,$A7
	.dc.b	$DF,$FA,$47,$6F,$B9,$B1,$FE,$44,$6E,$1F,$11,$AF,$F9,$09,$CD,$FF
	.dc.b	$94,$9B,$3F,$E0,$EF,$B1,$FF,$06,$C6,$2B,$FA,$C7,$B1,$FE,$0E,$3F
	.dc.b	$C1,$FE,$C5,$BF,$F1,$7C,$59,$FF,$15,$CD,$C3,$FC,$19,$F4,$BF,$E4
	.dc.b	$0F,$89,$FF,$20,$B7,$0F,$F2,$43,$6F,$FC,$90,$DA,$E3,$3F,$E4,$DA
	.dc.b	$E2,$D3,$E2,$7B,$5F,$0C,$8F,$C3,$6D,$AB,$E0,$47,$F8,$A0,$F7,$9F
	.dc.b	$E5,$4F,$A4,$95,$FE,$5B,$F8,$3B,$7F,$E5,$B9,$B7,$FE,$27,$9B,$5F
	.dc.b	$E5,$5F,$3F,$FC,$67,$E7,$FF,$8D,$1B,$1F,$E5,$46,$E1,$FE,$50,$6D
	.dc.b	$FF,$93,$1B,$3C,$CF,$F2,$DD,$7F,$94,$FE,$0F,$FC,$D1,$1F,$E8,$B7
	.dc.b	$FC,$DB,$7E,$D8,$C7,$9D,$BF,$3F,$86,$BC,$78,$AD,$7F,$8A,$B3,$62
	.dc.b	$9F,$F3,$07,$D2,$FF,$25,$9F,$F3,$AB,$FD,$19,$39,$AF,$F4,$E3,$9F
	.dc.b	$84,$6D,$FE,$69,$6C,$F7,$5C,$39,$9F,$F3,$77,$D2,$3C,$0A,$63,$C1
	.dc.b	$B8,$79,$E5,$EF,$70,$38,$FB,$F6,$7D,$9E,$37,$F9,$A7,$8F,$F0,$0B
	.dc.b	$E0,$7F,$CF,$8E,$77,$FF,$19,$C7,$F9,$FF,$E8,$BF,$DE,$77,$C8,$C7
	.dc.b	$B9,$C1,$D7,$B9,$BF,$89,$B1,$F2,$E9,$1F,$E8,$9F,$4F,$FC,$AB,$8F
	.dc.b	$FA,$79,$B5,$FE,$C3,$5F,$47,$BF,$4F,$87,$D6,$CA,$FA,$D4,$FA,$4C
	.dc.b	$6F,$F7,$CF,$CE,$7D,$12,$7F,$D3,$0D,$AF,$BC,$FF,$A2,$5B,$1F,$E8
	.dc.b	$AF,$A2,$7D,$15,$3F,$D1,$BF,$44,$F6,$31,$7F,$D3,$E3,$0D,$8F,$F4
	.dc.b	$9F,$D1,$7E,$0F,$2F,$E7,$15,$A7,$CC,$23,$4D,$F6,$3F,$D3,$7E,$CF
	.dc.b	$FA,$AE,$73,$F0,$9F,$C4,$68,$DF,$A1,$FB,$59,$7F,$61,$8C,$74,$7F
	.dc.b	$A9,$1B,$19,$F3,$FF,$D4,$F5,$C2,$FC,$6D,$E5,$FF,$68,$FD,$13,$DD
	.dc.b	$DE,$5F,$EA,$EF,$9F,$7D,$15,$3F,$D8,$9F,$32,$FA,$2C,$6B,$FD,$CE
	.dc.b	$B6,$DA,$DC,$EB,$4F,$87,$5A,$6C,$EA,$B8,$1C,$FB,$3C,$47,$FB,$4F
	.dc.b	$E8,$9F,$2B,$89,$D1,$2F,$F0,$9B,$2B,$73,$98,$3F,$DB,$65,$8B,$E7
	.dc.b	$BE,$01,$F0,$62,$CF,$B6,$2C,$7A,$8B,$5D,$1F,$EF,$0B,$CF,$FF,$6B
	.dc.b	$B6,$BF,$C3,$13,$FD,$F7,$E8,$BF,$BF,$7D,$0B,$E8,$3C,$97,$BA,$FC
	.dc.b	$4B,$4E,$CF,$F3,$BC,$3F,$DC,$6F,$31,$5C,$DF,$1C,$69,$5F,$82,$FB
	.dc.b	$CF,$A6,$B8,$57,$C5,$AD,$3C,$B0,$AE,$4E,$F7,$CA,$33,$1E,$3A,$E7
	.dc.b	$87,$AD,$EB,$19,$AE,$0F,$23,$EA,$1E,$63,$07,$BE,$BA,$B9,$9E,$B8
	.dc.b	$7A,$6B,$FC,$47,$87,$9B,$0F,$04,$62,$F9,$C6,$1E,$F3,$BA,$E5,$18
	.dc.b	$3F,$C5,$A7,$CD,$2C,$61,$C1,$F8,$BF,$A1,$7B,$EF,$A2,$E3,$18,$6C
	.dc.b	$66,$7F,$C7,$33,$96,$FC,$0F,$F7,$76,$E3,$CA,$31,$1F,$1A,$51,$88
	.dc.b	$F1,$3C,$73,$FB,$79,$FE,$76,$7E,$CB,$5F,$9B,$5C,$AF,$EE,$B6,$3E
	.dc.b	$8C,$EE,$B9,$EB,$FA,$97,$F0,$F5,$B6,$E1,$E1,$AF,$D7,$1C,$4B,$5C
	.dc.b	$5F,$3E,$1E,$09,$E1,$87,$F9,$86,$78,$36,$B8,$3F,$E6,$5C,$3C,$13
	.dc.b	$CA,$3C,$F3,$A7,$F5,$8D,$9D,$C7,$33,$FD,$15,$3F,$0E,$FF,$5A,$DD
	.dc.b	$7A,$63,$F0,$2B,$7F,$45,$FB,$69,$F6,$B6,$3A,$1F,$D0,$3E,$8B,$F8
	.dc.b	$13,$FE,$97,$6B,$9D,$F8,$28,$FB,$6E,$E3,$A6,$2E,$67,$CB,$61,$B5
	.dc.b	$B3,$FD,$49,$F4,$5F,$E6,$67,$FD,$4D,$F4,$3F,$B2,$CF,$F9,$03,$63
	.dc.b	$FD,$1D,$3B,$7F,$AB,$B7,$0F,$E1,$F3,$B9,$36,$75,$5F,$CA,$D7,$A9
	.dc.b	$EC,$9F,$F0,$1D,$78,$BA,$3D,$F5,$A8,$F3,$D6,$E6,$46,$08,$7F,$6A
	.dc.b	$AE,$6D,$8E,$73,$F2,$BE,$8B,$38,$FF,$26,$F8,$27,$BD,$F3,$9B,$87
	.dc.b	$FB,$26,$B7,$9F,$F4,$39,$CC,$F7,$D7,$D4,$71,$BB,$1F,$ED,$1F,$53
	.dc.b	$78,$FF,$34,$6E,$1F,$3A,$FE,$EF,$D1,$7C,$6B,$FF,$5E,$35,$37,$1F
	.dc.b	$F2,$4C,$7D,$8E,$DF,$94,$47,$FD,$C1,$3F,$E4,$77,$FD,$C5,$EF,$BD
	.dc.b	$DF,$C5,$B6,$33,$1F,$8E,$FA,$27,$C5,$FE,$33,$0F,$F0,$6D,$7F,$98
	.dc.b	$DA,$FF,$6A,$6E,$BF,$D7,$4F,$EE,$3F,$45,$F9,$4F,$A2,$F6,$41,$A4
	.dc.b	$F9,$AD,$79,$13,$0F,$F3,$84,$FF,$83,$DB,$F2,$FC,$FD,$0E,$9C,$72
	.dc.b	$77,$1D,$FE,$07,$C1,$CF,$27,$9C,$78,$F6,$F2,$79,$72,$7B,$C7,$9F
	.dc.b	$E3,$A4,$78,$9F,$57,$C7,$7A,$FE,$36,$4F,$99,$F7,$7C,$6F,$BF,$A2
	.dc.b	$4F,$A9,$F8,$5F,$45,$F0,$FE,$37,$E2,$7C,$6C,$AF,$7C,$7C,$7F,$87
	.dc.b	$F9,$1F,$0F,$27,$E1,$C7,$EF,$BE,$89,$F1,$1F,$37,$C1,$95,$FE,$97
	.dc.b	$9C,$BF,$E0,$E8,$FF,$0D,$7D,$13,$E3,$FF,$85,$B6,$74,$03,$97,$FC
	.dc.b	$AA,$FE,$61,$D3,$9E,$CF,$F0,$7B,$6B,$BD,$76,$F1,$3B,$B2,$FB,$9D
	.dc.b	$9D,$38,$FA,$2F,$F1,$CB,$7C,$0F,$E4,$31,$88,$FF,$33,$FC,$5F,$F9
	.dc.b	$D1,$D1,$FD,$CA,$34,$FF,$81,$9A,$FF,$8E,$DA,$FB,$E1,$F4,$5F,$CB
	.dc.b	$9B,$1E,$66,$BE,$0F,$1F,$53,$23,$4F,$18,$E5,$B4,$9E,$4F,$F3,$AF
	.dc.b	$A2,$FF,$1A,$ED,$FE,$36,$C3,$EB,$D7,$B2,$B8,$3F,$E8,$4A,$DE,$BF
	.dc.b	$E5,$3A,$DD,$F1,$BD,$BF,$E4,$0A,$DD,$FD,$BF,$AB,$5F,$F8,$4D,$B1
	.dc.b	$FE,$18,$6B,$FE,$2C,$0E,$1F,$F1,$7F,$D1,$B9,$7F,$80,$C7,$F9,$96
	.dc.b	$31,$F0,$5F,$E2,$25,$CC,$F4,$F1,$7C,$2D,$9B,$E7,$FF,$DE,$F1,$86
	.dc.b	$C6,$79,$3F,$86,$6F,$FE,$68,$9C,$93,$93,$37,$C6,$AD,$F5,$9B,$F3
	.dc.b	$CF,$26,$9A,$FA,$2F,$F5,$4E,$5F,$EF,$F9,$D3,$67,$6F,$E9,$AD,$5C
	.dc.b	$31,$FF,$45,$DF,$06,$DF,$07,$FD,$1F,$3B,$AF,$E4,$D5,$B7,$FC,$EF
	.dc.b	$F4,$5F,$EE,$FF,$AA,$FF,$54,$63,$FA,$68,$C7,$C2,$7B,$B1,$46,$27
	.dc.b	$5E,$2F,$C0,$BE,$35,$E4,$F2,$FE,$09,$F0,$BE,$1E,$DF,$BE,$2C,$FA
	.dc.b	$A6,$E5,$FC,$EF,$81,$F4,$5F,$31,$FF,$74,$57,$FA,$A9,$CE,$1F,$15
	.dc.b	$CD,$71,$AE,$1F,$1F,$C5,$B5,$B6,$DF,$36,$D7,$FA,$6B,$E8,$9F,$29
	.dc.b	$F7,$3E,$AD,$F1,$DF,$0E,$73,$8F,$D5,$9C,$BF,$C3,$A3,$38,$FF,$B0
	.dc.b	$27,$2F,$FB,$C2,$33,$87,$F5,$99,$CB,$F0,$7E,$43,$E0,$61,$F0,$5B
	.dc.b	$87,$BD,$B8,$7C,$36,$D7,$DE,$F4,$BF,$22,$7E,$4B,$D3,$B7,$F8,$F8
	.dc.b	$B7,$F8,$A9,$B5,$C6,$B4,$7E,$56,$32,$FF,$B2,$B9,$57,$8B,$FD,$6C
	.dc.b	$2B,$16,$8D,$DF,$4A,$F0,$36,$7C,$11,$F9,$CB,$F4,$4E,$4F,$03,$D2
	.dc.b	$BB,$46,$67,$C3,$39,$BD,$36,$75,$7E,$38,$C5,$77,$9C,$DF,$9A,$3C
	.dc.b	$A6,$49,$5E,$63,$F5,$77,$DA,$B8,$B6,$77,$1F,$1E,$31,$3C,$63,$07
	.dc.b	$82,$FC,$71,$9A,$F5,$CF,$43,$45,$7A,$9A,$2C,$7C,$FA,$EF,$3F,$20
	.dc.b	$FB,$47,$29,$DD,$FC,$A6,$DF,$8A,$78,$46,$1E,$2D,$38,$D7,$67,$C5
	.dc.b	$7F,$12,$7B,$3E,$B3,$C8,$E0,$EA,$EE,$3A,$3B,$BF,$23,$6F,$C9,$3D
	.dc.b	$AB,$89,$1F,$C8,$DF,$8E,$BE,$F4,$FA,$1E,$4F,$C9,$6C,$FA,$A3,$C8
	.dc.b	$F5,$39,$B8,$8E,$71,$F8,$47,$89,$E7,$9F,$05,$FE,$26,$3D,$86,$AB
	.dc.b	$CC,$FC,$2A,$F7,$C6,$27,$DE,$DB,$F7,$D7,$80,$E5,$5E,$43,$89,$E7
	.dc.b	$7A,$1A,$38,$4F,$E3,$A3,$71,$87,$93,$BA,$F4,$4E,$E3,$15,$FB,$2A
	.dc.b	$E3,$5A,$78,$D7,$9C,$F9,$85,$7E,$61,$C2,$79,$57,$E7,$61,$9B,$E8
	.dc.b	$FF,$A5,$7D,$5F,$F2,$D6,$E1,$FC,$AD,$B5,$FC,$A5,$B7,$FD,$73,$D4
	.dc.b	$FE,$DF,$EA,$FF,$70,$6D,$FF,$6D,$6D,$FF,$68,$6D,$7F,$4F,$FA,$4F
	.dc.b	$E9,$C7,$F4,$FF,$A4,$B7,$FD,$36,$2D,$FF,$51,$9B,$5C,$6B,$47,$F5
	.dc.b	$38,$CB,$FE,$5D,$97,$D1,$BE,$93,$FC,$0B,$EC,$FF,$6D,$8C,$57,$F8
	.dc.b	$0E,$31,$3A,$8F,$F2,$CC,$EA,$BF,$C0,$35,$FD,$97,$E3,$25,$7F,$63
	.dc.b	$F8,$DB,$5F,$D9,$E6,$E5,$FD,$B6,$6D,$7F,$3B,$F6,$FF,$C2,$FE,$DF
	.dc.b	$F8,$61,$B1,$FE,$15,$6E,$1F,$E1,$06,$DF,$F8,$31,$B3,$CC,$FF,$0E
	.dc.b	$57,$F8,$27,$E9,$1D,$DF,$E0,$7F,$85,$C3,$6B,$FC,$EB,$3A,$7F,$C8
	.dc.b	$98,$A3,$0F,$F9,$1D,$AF,$CD,$35,$29,$FF,$26,$63,$FD,$47,$DD,$FF
	.dc.b	$19,$9F,$E3,$4F,$76,$CF,$F8,$C6,$2D,$FF,$8B,$26,$D7,$F8,$4F,$E8
	.dc.b	$7F,$C2,$5F,$15,$FE,$13,$6E,$1F,$E1,$16,$DF,$F8,$55,$B5,$C6,$7F
	.dc.b	$C2,$D5,$C5,$A7,$C0,$F7,$FF,$C8,$FE,$FF,$C7,$6C,$64,$F8,$7A,$23
	.dc.b	$DD,$8F,$C3,$6B,$F0,$E3,$53,$F0,$BC,$57,$D2,$7F,$8A,$CF,$F1,$67
	.dc.b	$D1,$59,$FF,$15,$45,$BF,$F1,$74,$DC,$FF,$C9,$3E,$9F,$F9,$73,$D2
	.dc.b	$FF,$2E,$B7,$0F,$F2,$D3,$6F,$FC,$AA,$DA,$E3,$3F,$E5,$FA,$E2,$D3
	.dc.b	$FC,$BD,$F0,$FF,$E7,$21,$E9,$FF,$9B,$1C,$37,$7C,$7D,$25,$AE,$2F
	.dc.b	$F9,$CB,$E9,$3F,$8B,$62,$7F,$96,$E7,$FE,$82,$72,$62,$1F,$12,$D7
	.dc.b	$FC,$F6,$DC,$3F,$CE,$4B,$8F,$33,$FE,$80,$FA,$47,$85,$4C,$78,$77
	.dc.b	$0F,$F3,$D4,$BE,$17,$81,$C7,$E1,$AC,$FF,$9F,$FC,$B7,$A9,$FE,$8C
	.dc.b	$5E,$1D,$8C,$3F,$E9,$0F,$A4,$FF,$7A,$D7,$13,$1F,$03,$C5,$D7,$C0
	.dc.b	$6F,$C2,$B1,$FE,$8D,$A4,$7F,$A6,$BD,$5F,$F2,$AC,$FF,$BD,$78,$D1
	.dc.b	$8E,$5F,$F5,$5A,$FA,$4E,$1E,$AF,$1F,$A4,$98,$7F,$A6,$69,$F4,$98
	.dc.b	$C3,$FD,$67,$3D,$7E,$93,$FC,$F2,$DA,$FF,$55,$C3,$E9,$6C,$7F,$A4
	.dc.b	$FE,$95,$F4,$B4,$FF,$4B,$7D,$2B,$DA,$C6,$3F,$F0,$02,$30,$D8,$FF
	.dc.b	$4E,$7D,$2F,$F5,$BC,$D5,$A7,$CA,$04,$6C,$E1,$EE,$7F,$AA,$EF,$F9
	.dc.b	$F4,$73,$F8,$8F,$E2,$FC,$5F,$B4,$6C,$FF,$43,$FA,$5F,$B6,$FC,$D3
	.dc.b	$FA,$34,$6B,$0F,$04,$E9,$F1,$88,$D2,$8D,$57,$7F,$47,$FD,$71,$18
	.dc.b	$BF,$07,$4F,$FB,$5B,$E9,$5F,$29,$A8,$FF,$5F,$7D,$1B,$E9,$69,$FE
	.dc.b	$DC,$F9,$DF,$F5,$EE,$7F,$C3,$A7,$58,$FA,$74,$F8,$D1,$87,$CE,$D9
	.dc.b	$FE,$2E,$2B,$9D,$7F,$65,$9D,$36,$FC,$2F,$FB,$86,$77,$7D,$A7,$CF
	.dc.b	$D7,$2B,$FD,$CD,$F4,$AF,$95,$C5,$F8,$AA,$E5,$E2,$AD,$5F,$A6,$99
	.dc.b	$7E,$8A,$30,$BE,$8F,$E2,$9F,$1A,$6C,$EA,$2C,$7C,$02,$DF,$F3,$41
	.dc.b	$E7,$7F,$BA,$DB,$5E,$19,$FC,$77,$D2,$FE,$3F,$E8,$9F,$47,$FE,$09
	.dc.b	$BD,$D7,$E0,$9A,$75,$C3,$FD,$ED,$1C,$85,$72,$7F,$75,$1A,$57,$DE
	.dc.b	$FB,$4F,$F4,$CA,$E1,$5F,$E1,$9A,$D3,$CB,$0A,$E4,$EF,$7C,$A3,$31
	.dc.b	$E3,$AE,$78,$7A,$9E,$B1,$9A,$E0,$F2,$3F,$BE,$5F,$6A,$E1,$3F,$DF
	.dc.b	$E7,$D2,$6C,$46,$9E,$26,$5C,$3E,$7A,$F1,$E1,$EF,$AE,$97,$DE,$31
	.dc.b	$5C,$8F,$F0,$9C,$7F,$85,$1F,$9C,$60,$F2,$7D,$23,$E0,$1F,$4B,$FE
	.dc.b	$1A,$8C,$36,$33,$3C,$1A,$FF,$86,$E5,$F0,$5E,$C8,$EA,$D7,$4D,$7F
	.dc.b	$C9,$D1,$C3,$3F,$04,$F1,$CF,$ED,$E7,$F9,$BB,$FB,$2D,$7E,$55,$F1
	.dc.b	$6D,$71,$9F,$F2,$D1,$BA,$E7,$AF,$F9,$75,$FC,$3D,$F1,$6D,$71,$BF
	.dc.b	$E5,$57,$C4,$E7,$7C,$5F,$4C,$EE,$57,$CE,$77,$1E,$8A,$E0,$FF,$35
	.dc.b	$9D,$36,$B9,$57,$78,$F0,$D7,$07,$AC,$EA,$3F,$5D,$AF,$69,$FE,$5F
	.dc.b	$5F,$17,$1F,$F1,$6E,$1D,$A7,$F6,$B7,$C1,$F6,$FC,$AD,$7F,$72,$DB
	.dc.b	$E5,$87,$FA,$0F,$E9,$7F,$81,$D7,$F1,$16,$BF,$CF,$9B,$8F,$F8,$F1
	.dc.b	$FC,$1C,$6F,$7F,$04,$FF,$2B,$F9,$AB,$CC,$7C,$57,$CB,$0F,$F4,$97
	.dc.b	$D2,$FF,$9C,$B1,$FE,$41,$F4,$BF,$B4,$9F,$F4,$B3,$63,$FB,$04,$ED
	.dc.b	$FE,$A2,$DC,$7F,$C9,$F3,$B9,$36,$75,$5E,$0F,$53,$E2,$4F,$F8,$3A
	.dc.b	$BC,$3C,$FF,$1E,$D5,$4D,$C4,$9F,$E6,$D3,$A6,$CE,$5F,$EC,$B5,$D5
	.dc.b	$F3,$C7,$D3,$FA,$5C,$63,$FE,$66,$F8,$3F,$F2,$6D,$7F,$41,$6B,$FD
	.dc.b	$BA,$77,$97,$FA,$E6,$73,$5F,$03,$4F,$59,$F7,$E3,$33,$F2,$A3,$F0
	.dc.b	$5E,$9E,$30,$FF,$05,$35,$F8,$F1,$C7,$D8,$FC,$47,$D2,$FC,$89,$FF
	.dc.b	$08,$35,$FF,$40,$46,$BD,$9F,$97,$DD,$27,$FA,$D5,$7F,$94,$9F,$EB
	.dc.b	$9F,$4B,$FE,$4A,$7F,$AF,$7C,$07,$F9,$79,$BE,$EF,$E5,$E7,$51,$FE
	.dc.b	$20,$D9,$FB,$2B,$ED,$3F,$1E,$FF,$8C,$27,$FC,$01,$3A,$F3,$F3,$5A
	.dc.b	$EB,$DF,$64,$BD,$E0,$9F,$CC,$EA,$FB,$6E,$B5,$0E,$B1,$96,$FC,$FF
	.dc.b	$02,$3A,$F8,$4E,$7E,$A4,$8E,$2F,$6F,$57,$BF,$A5,$27,$AB,$E2,$F9
	.dc.b	$9F,$1F,$C8,$F9,$3D,$69,$5E,$28,$F3,$FC,$9F,$A3,$E4,$E4,$F9,$A3
	.dc.b	$D7,$E4,$FB,$3C,$B9,$3E,$A8,$F7,$F9,$12,$3D,$AF,$C2,$F8,$8F,$87
	.dc.b	$DD,$27,$E0,$BF,$1B,$E1,$FE,$3F,$C6,$C9,$F8,$AF,$CA,$F8,$CF,$E2
	.dc.b	$1F,$4A,$D2,$EB,$2F,$FC,$3D,$39,$6C,$E5,$FF,$26,$C2,$79,$CC,$9E
	.dc.b	$31,$BE,$DF,$F6,$7D,$73,$9E,$3C,$DE,$22,$3E,$0E,$6A,$E8,$F1,$C7
	.dc.b	$A2,$3F,$94,$7D,$2F,$F8,$E6,$7F,$E0,$0B,$5F,$F0,$03,$5F,$F8,$00
	.dc.b	$D7,$FC,$EC,$1E,$17,$FE,$E4,$F8,$6F,$E2,$C7,$F1,$78,$CF,$8F,$C2
	.dc.b	$1F,$4B,$FC,$E5,$B7,$FE,$45,$6B,$FE,$72,$7F,$97,$39,$8E,$04,$78
	.dc.b	$78,$F5,$E1,$17,$0F,$24,$F7,$F8,$89,$5D,$DF,$E9,$5F,$4B,$FE,$49
	.dc.b	$C7,$CD,$1A,$9F,$13,$A8,$FE,$A1,$F4,$8F,$90,$FF,$43,$57,$F8,$9E
	.dc.b	$31,$F0,$5F,$E2,$98,$43,$E4,$7F,$69,$1B,$AE,$2F,$CC,$1F,$52,$FA
	.dc.b	$56,$3C,$E3,$0D,$8C,$CF,$F8,$DE,$72,$F3,$D5,$F2,$8E,$3F,$F4,$2E
	.dc.b	$1F,$EC,$F8,$C3,$FC,$39,$71,$EA,$FF,$C7,$F1,$A7,$F8,$C3,$6F,$FD
	.dc.b	$1E,$DA,$E9,$5C,$67,$C3,$76,$B5,$7D,$23,$15,$97,$87,$CF,$FF,$AF
	.dc.b	$E7,$FC,$9E,$DB,$7D,$2F,$FA,$D7,$1F,$E4,$B3,$92,$67,$E1,$9D,$5F
	.dc.b	$F2,$F9,$E8,$BE,$93,$FE,$00,$DF,$FA,$55,$B1,$EE,$F1,$BE,$13,$BF
	.dc.b	$93,$B3,$AC,$BE,$8C,$E6,$FE,$55,$EA,$FD,$7F,$47,$27,$CF,$F1,$76
	.dc.b	$FC,$93,$70,$FF,$5C,$E5,$FE,$D5,$4A,$35,$7F,$EE,$79,$CC,$7C,$8B
	.dc.b	$CD,$FD,$8F,$23,$E8,$79,$5F,$45,$B8,$7D,$F6,$CF,$DC,$6E,$1E,$CF
	.dc.b	$A1,$7C,$9F,$C2,$6D,$FD,$0F,$8B,$C4,$C8,$FF,$5F,$4E,$B0,$FA,$15
	.dc.b	$97,$F1,$3E,$67,$E0,$FC,$DF,$C2,$36,$FF,$24,$D9,$DC,$7E,$56,$76
	.dc.b	$D3,$65,$8D,$9F,$96,$8D,$B6,$36,$D3,$78,$FE,$6A,$76,$D9,$DB,$67
	.dc.b	$71,$FA,$99,$DB,$4D,$96,$36,$7E,$AA,$36,$D8,$DB,$4D,$CF,$EE,$27
	.dc.b	$6D,$9D,$B6,$77,$CC,$AD,$B6,$B6,$D9,$DC,$E8,$93,$A6,$C6,$A3,$04
	.dc.b	$70,$4A,$FE,$03,$7F,$5D,$B3,$F4,$DB,$3F,$3D,$A7,$CB,$2C,$7C,$76
	.dc.b	$C7,$C3,$69,$EF,$2D,$FA,$DB,$7E,$86,$D7,$9D,$B5,$E7,$8F,$E4,$17
	.dc.b	$E7,$69,$E7,$2C,$79,$DB,$3E,$76,$9E,$72,$C7,$9D,$B1,$E7,$6D,$F9
	.dc.b	$DB,$7E,$76,$9E,$72,$D7,$9C,$FE,$63,$7E,$76,$C7,$9D,$A7,$9C,$B7
	.dc.b	$E7,$6C,$79,$DA,$79,$CB,$1E,$76,$DF,$9D,$B7,$E7,$6D,$79,$DA,$79
	.dc.b	$EB,$FA,$7D,$F9,$DB,$3E,$76,$CF,$9D,$A7,$9C,$B1,$E7,$6C,$79,$DA
	.dc.b	$79,$CB,$7E,$76,$DF,$9D,$B5,$E7,$6D,$79,$E9,$7E,$76,$D7,$9D,$B3
	.dc.b	$E7,$6D,$79,$DB,$3E,$76,$CF,$9D,$B3,$E7,$69,$E7,$2D,$F9,$F7,$7D
	.dc.b	$17,$F6,$23,$FB,$1F,$B7,$71,$FE,$BF,$17,$1F,$EB,$93,$71,$FE,$71
	.dc.b	$F4,$3F,$CD,$CF,$EE,$1E,$85,$C3,$F9,$AC,$5B,$FE,$F7,$36,$BF,$A8
	.dc.b	$FB,$9F,$DC,$BD,$FF,$EE,$6D,$C3,$FB,$7B,$6F,$FB,$63,$6B,$8C,$FF
	.dc.b	$74,$AE,$2D,$3F,$C1,$DE,$57,$F8,$0C,$FF,$02,$79,$56,$32,$4C,$3F
	.dc.b	$C1,$F1,$95,$7F,$DC,$3E,$85,$97,$F7,$66,$DF,$F8,$13,$A7,$13,$23
	.dc.b	$FB,$41,$FE,$5A,$9D,$DF,$F8,$16,$B4,$FF,$87,$BD,$2F,$F0,$C1,$FE
	.dc.b	$19,$F4,$ED,$FF,$89,$E2,$CF,$F8,$92,$6E,$3F,$E2,$49,$B5,$FE,$2F
	.dc.b	$9B,$5F,$E1,$CF,$A1,$FF,$0F,$FD,$0F,$F8,$81,$B5,$FE,$1C,$6C,$FF
	.dc.b	$86,$5B,$87,$F8,$43,$D1,$76,$FF,$84,$AB,$FC,$EB,$18,$BF,$F2,$93
	.dc.b	$63,$FC,$9C,$1C,$3F,$E5,$17,$55,$D6,$BF,$C7,$1E,$A7,$F8,$8A,$7F
	.dc.b	$C2,$9C,$CF,$A1,$FF,$1B,$9F,$E3,$8F,$A1,$B9,$7F,$8D,$22,$E5,$FE
	.dc.b	$31,$9B,$97,$F8,$7A,$6D,$7F,$86,$ED,$58,$BF,$F2,$8F,$AB,$FE,$49
	.dc.b	$F5,$FF,$C9,$4D,$BF,$F2,$C3,$67,$FC,$AA,$DC,$7E,$37,$C1,$FC,$59
	.dc.b	$1F,$05,$C3,$6B,$E3,$C6,$33,$3F,$04,$FF,$2A,$96,$3F,$CD,$1E,$BF
	.dc.b	$F9,$74,$FF,$2F,$7B,$16,$32,$4B,$FF,$34,$C6,$5F,$F3,$5C,$DC,$3F
	.dc.b	$CD,$73,$6B,$FC,$7D,$F0,$5F,$EF,$97,$F3,$37,$FA,$28,$C3,$FA,$48
	.dc.b	$FF,$22,$79,$1E,$E7,$76,$AD,$BF,$E8,$83,$1C,$BF,$E7,$78,$37,$97
	.dc.b	$0F,$F9,$EF,$E8,$1E,$25,$31,$E4,$DC,$7B,$FC,$1B,$8B,$FD,$04,$B9
	.dc.b	$AC,$61,$FF,$43,$7D,$07,$F2,$0A,$F3,$AF,$0F,$3E,$26,$9D,$FA,$7C
	.dc.b	$3D,$3B,$7C,$1F,$F4,$84,$BC,$5A,$63,$C5,$B5,$F1,$25,$F4,$1C,$0E
	.dc.b	$3F,$41,$67,$E4,$FD,$03,$E8,$BF,$D3,$8D,$8C,$FD,$2E,$27,$FD,$41
	.dc.b	$E9,$3D,$BF,$F5,$17,$E0,$DF,$EB,$AB,$FA,$73,$97,$EA,$C6,$7B,$7E
	.dc.b	$D1,$FC,$7A,$5F,$01,$D3,$E3,$FA,$B6,$FE,$EF,$D0,$7F,$AB,$3B,$CF
	.dc.b	$FA,$EA,$33,$ED,$FF,$AC,$E3,$18,$F6,$1F,$40,$F8,$CF,$F4,$FB,$6F
	.dc.b	$F1,$F2,$FA,$0F,$EB,$AF,$0A,$FF,$52,$C6,$0C,$FD,$07,$DB,$9E,$D1
	.dc.b	$8E,$AF,$F5,$5C,$E7,$3F,$F5,$4F,$AF,$FE,$B4,$8C,$63,$FB,$6C,$1F
	.dc.b	$07,$FC,$52,$30,$D8,$FD,$DF,$D0,$3E,$7F,$FD,$6F,$39,$94,$E6,$BD
	.dc.b	$73,$98,$C4,$FC,$38,$C5,$7F,$B0,$27,$38,$FE,$9F,$E8,$3F,$43,$87
	.dc.b	$C4,$9C,$BE,$7F,$9E,$7D,$07,$FB,$21,$B1,$FB,$2F,$9A,$75,$66,$A4
	.dc.b	$FC,$EA,$DE,$9F,$82,$C9,$87,$0A,$D5,$F0,$AF,$B5,$7C,$2F,$F7,$97
	.dc.b	$B9,$CD,$78,$F9,$A6,$1F,$C1,$3E,$81,$F2,$18,$BE,$75,$71,$FE,$0D
	.dc.b	$5A,$BF,$34,$FB,$7E,$52,$47,$A6,$E9,$E4,$8B,$7F,$EE,$6C,$FF,$8A
	.dc.b	$4E,$6B,$F1,$B3,$FC,$1D,$BE,$47,$FB,$C2,$7F,$9B,$FD,$07,$F3,$8F
	.dc.b	$9F,$7C,$F7,$E1,$70,$E1,$5F,$9B,$69,$FE,$F4,$2E,$1F,$D8,$2F,$95
	.dc.b	$71,$C3,$B5,$A7,$2F,$FB,$B6,$15,$88,$FC,$A4,$DB,$F0,$E1,$E4,$9F
	.dc.b	$CC,$1C,$C3,$97,$FC,$33,$7D,$1C,$E1,$E5,$69,$E1,$C3,$C0,$FC,$23
	.dc.b	$27,$27,$2F,$7A,$ED,$7E,$4A,$E9,$18,$BF,$03,$E3,$AE,$8E,$6F,$9E
	.dc.b	$1E,$DA,$DD,$F2,$3E,$44,$EA,$FF,$C1,$67,$94,$B5,$87,$07,$EE,$FE
	.dc.b	$7D,$E9,$3E,$83,$8C,$61,$B1,$99,$FF,$0B,$CE,$5B,$E9,$FF,$90,$9B
	.dc.b	$A7,$28,$C4,$7D,$29,$FE,$B7,$39,$3E,$D6,$4D,$BF,$87,$6B,$F5,$63
	.dc.b	$17,$F7,$EF,$8B,$F4,$1B,$3B,$AE,$35,$FC,$5C,$E3,$5B,$C3,$CD,$5F
	.dc.b	$81,$BE,$35,$BB,$E3,$7F,$E1,$48,$E3,$5B,$BE,$31,$F2,$5B,$3B,$AE
	.dc.b	$2F,$E9,$1B,$7D,$A3,$FC,$3E,$D7,$47,$C2,$8D,$CF,$48,$C9,$FD,$16
	.dc.b	$BD,$78,$7E,$5E,$32,$DB,$E9,$3A,$D3,$E1,$7C,$C7,$ED,$5B,$7F,$D2
	.dc.b	$A7,$4F,$F8,$CB,$E8,$3F,$C0,$CF,$F4,$E6,$BF,$D5,$5B,$87,$85,$FF
	.dc.b	$0A,$C6,$F7,$EC,$FF,$A1,$39,$5F,$2B,$8B,$FE,$52,$EA,$57,$D3,$1F
	.dc.b	$E9,$DF,$41,$FE,$6A,$9F,$EB,$FF,$41,$FA,$A9,$FE,$A4,$D8,$FF,$24
	.dc.b	$CE,$DF,$ED,$4D,$C3,$E3,$4E,$E4,$D9,$D5,$78,$3C,$FF,$24,$B9,$A9
	.dc.b	$E7,$A4,$D3,$37,$E6,$9C,$3F,$DD,$A7,$6E,$48,$79,$A7,$E1,$7D,$07
	.dc.b	$78,$FE,$FB,$EF,$3E,$57,$E3,$36,$3F,$C0,$33,$B7,$FC,$11,$19,$7F
	.dc.b	$C1,$58,$FD,$07,$E4,$EB,$CA,$FD,$AF,$D1,$F4,$B8,$6B,$F8,$B6,$EB
	.dc.b	$F3,$A7,$C3,$EF,$7D,$7F,$A0,$F8,$F8,$FF,$9E,$1A,$FF,$9D,$23,$5E
	.dc.b	$AF,$CA,$EF,$93,$FE,$18,$AF,$F2,$13,$FE,$1A,$FA,$07,$C5,$7E,$0F
	.dc.b	$E0,$1E,$ED,$3F,$0D,$3A,$AF,$F3,$4F,$CF,$BE,$83,$F1,$8F,$F8,$9E
	.dc.b	$33,$E8,$62,$73,$DB,$AF,$02,$57,$28,$FC,$A6,$AF,$A0,$E7,$30,$FF
	.dc.b	$3B,$C6,$5B,$EA,$70,$23,$AE,$6E,$7E,$B4,$8E,$2F,$6F,$84,$EF,$E9
	.dc.b	$49,$EA,$F8,$BE,$5F,$C7,$F2,$1E,$4F,$95,$95,$E2,$8F,$3F,$C9,$FA
	.dc.b	$3E,$4E,$4F,$9A,$3D,$7E,$57,$B3,$CC,$93,$EA,$8F,$7F,$93,$23,$DA
	.dc.b	$FC,$2F,$86,$F8,$7D,$D2,$7E,$0B,$F1,$BE,$33,$E3,$FC,$64,$9F,$8A
	.dc.b	$FC,$AF,$8C,$FD,$E7,$D0,$34,$BA,$CB,$FF,$2C,$CE,$5B,$39,$7F,$CC
	.dc.b	$50,$9E,$73,$27,$8C,$6F,$C6,$FE,$45,$5C,$E7,$8F,$37,$88,$8F,$83
	.dc.b	$9A,$BA,$3C,$71,$E8,$8F,$E2,$9F,$41,$FE,$32,$9F,$F3,$7B,$5F,$F3
	.dc.b	$B3,$6F,$F9,$97,$C1,$7E,$46,$73,$D3,$FD,$BE,$B8,$47,$F8,$49,$AF
	.dc.b	$F9,$AD,$AF,$FB,$8A,$5F,$41,$FC,$95,$B1,$FE,$80,$6B,$FE,$87,$7F
	.dc.b	$7C,$D8,$E0,$47,$87,$8F,$5D,$1C,$E3,$B4,$77,$F8,$59,$3D,$DF,$E6
	.dc.b	$DF,$41,$FE,$39,$F2,$BF,$D0,$58,$7F,$AA,$EB,$44,$3F,$D3,$53,$B7
	.dc.b	$FD,$FB,$F4,$FF,$EA,$87,$FC,$2A,$D7,$FC,$6A,$1C,$3F,$E3,$86,$BF
	.dc.b	$E9,$1F,$A1,$7C,$6F,$FA,$4A,$38,$C6,$3E,$0F,$FC,$42,$BE,$39,$E2
	.dc.b	$E2,$F8,$5F,$F9,$5F,$E6,$FF,$D6,$C4,$E3,$FF,$79,$6A,$97,$1A,$B7
	.dc.b	$C9,$A9,$CC,$71,$AF,$F7,$EB,$A6,$CE,$EB,$95,$DA,$FF,$63,$7C,$AB
	.dc.b	$E8,$34,$DC,$3C,$57,$FE,$55,$1F,$41,$FE,$B2,$8C,$9A,$9E,$19,$7F
	.dc.b	$A5,$DB,$5C,$5B,$3D,$2B,$95,$6E,$FD,$8B,$E8,$3F,$9A,$D7,$68,$F8
	.dc.b	$7F,$4F,$FE,$9B,$C7,$FC,$25,$5A,$9F,$F5,$34,$E6,$BF,$D9,$13,$99
	.dc.b	$FF,$39,$B5,$CF,$A7,$FE,$1B,$6C,$FE,$12,$D9,$3E,$5B,$E2,$9F,$1B
	.dc.b	$E6,$ED,$79,$6B,$73,$8B,$F8,$55,$67,$E1,$4D,$AF,$91,$36,$BE,$DF
	.dc.b	$95,$F5,$7C,$DF,$AC,$DA,$FB,$AD,$9F,$F6,$26,$E8,$C6,$3F,$E3,$5F
	.dc.b	$A0,$7D,$0F,$E0,$7E,$9D,$E7,$FC,$E9,$CE,$5E,$99,$CB,$FE,$FE,$8C
	.dc.b	$E5,$FE,$D1,$9C,$BF,$94,$9F,$F7,$A4,$EB,$0F,$C3,$56,$9F,$C4,$64
	.dc.b	$F3,$FF,$21,$8F,$E4,$5B,$87,$E6,$1B,$8F,$B1,$B7,$C9,$B3,$F9,$4B
	.dc.b	$E8,$FE,$82,$FA,$34,$E8,$5B,$E8,$DB,$E8,$D3,$A5,$F6,$9F,$D6,$B6
	.dc.b	$7F,$64,$DB,$F0,$CF,$D2,$BF,$0B,$6F,$C2,$D3,$C2,$5B,$F0,$B6,$3C
	.dc.b	$27,$8D,$F0,$B6,$3C,$2D,$AF,$0C,$7C,$EA,$F0,$B6,$7C,$2D,$AF,$0B
	.dc.b	$4F,$09,$63,$C3,$1E,$33,$C2,$58,$F0,$B6,$BC,$33,$F4,$EB,$C2,$D3
	.dc.b	$C2,$59,$F0,$B6,$BC,$2D,$3C,$2F,$8E,$3C,$2D,$3C,$25,$AF,$0C,$7F
	.dc.b	$22,$AF,$0B,$6B,$C2,$D3,$C2,$5A,$F0,$B6,$3C,$27,$8D,$F0,$B6,$3C
	.dc.b	$2D,$AF,$0C,$7F,$35,$AF,$0B,$67,$C2,$DA,$F0,$B4,$F0,$96,$3C,$31
	.dc.b	$E3,$3C,$25,$8F,$0B,$6B,$C2,$7E,$BE,$78,$CF,$39,$E2,$D3,$89,$67
	.dc.b	$8D,$7A,$1E,$53,$C1,$B3,$C1,$A7,$0A,$F6,$B8,$BF,$D2,$12,$3F,$2C
	.dc.b	$DA,$FC,$43,$6B,$ED,$36,$7E,$63,$67,$DC,$D9,$F4,$36,$BC,$15,$D8
	.dc.b	$95,$D9,$B5,$D9,$A7,$62,$C7,$66,$C7,$66,$CF,$69,$E8,$49,$E8,$D9
	.dc.b	$E8,$D3,$A1,$63,$E9,$6E,$F9,$1F,$ED,$E7,$F7,$0F,$72,$DF,$F4,$B9
	.dc.b	$D3,$67,$4D,$3F,$9B,$7D,$17,$F7,$DE,$57,$CB,$FF,$3D,$8C,$36,$7F
	.dc.b	$AB,$4E,$9E,$26,$23,$FC,$03,$EF,$7F,$80,$BE,$0A,$CF,$8D,$BA,$7F
	.dc.b	$7D,$6E,$9F,$DF,$5B,$3F,$E0,$36,$CF,$F8,$6B,$3F,$F6,$CC,$FF,$85
	.dc.b	$BC,$CF,$F0,$91,$FE,$13,$F3,$2C,$64,$95,$FD,$7A,$32,$FF,$51,$3F
	.dc.b	$C4,$DF,$44,$F5,$BF,$C3,$3F,$27,$FE,$0A,$8C,$5F,$CB,$8C,$4F,$F7
	.dc.b	$FA,$E2,$D5,$5C,$1F,$F2,$2B,$5F,$F1,$C7,$AD,$FE,$29,$3F,$C5,$5E
	.dc.b	$BD,$C3,$FC,$63,$16,$7F,$BA,$CD,$D7,$FB,$1C,$DA,$FE,$B5,$36,$BE
	.dc.b	$37,$D0,$FF,$8C,$BE,$8B,$FC,$66,$DA,$FC,$23,$67,$FC,$5C,$DC,$13
	.dc.b	$FE,$49,$97,$AA,$EF,$FC,$37,$C4,$7F,$8B,$A3,$15,$F8,$48,$C4,$FC
	.dc.b	$39,$CD,$7F,$99,$9A,$F5,$3A,$C7,$CB,$BF,$F1,$3F,$D1,$7F,$92,$8F
	.dc.b	$F2,$5F,$D1,$5C,$7F,$39,$16,$FF,$C8,$D3,$6B,$FC,$8D,$36,$BF,$C9
	.dc.b	$93,$6B,$FC,$BF,$EC,$FF,$96,$7D,$CF,$F2,$D3,$6B,$FC,$C2,$D9,$FF
	.dc.b	$12,$B7,$0F,$8B,$1F,$E6,$DC,$9F,$11,$FE,$60,$FA,$7F,$88,$DC,$3F
	.dc.b	$CA,$15,$F2,$A3,$E2,$12,$3F,$D1,$31,$97,$FC,$91,$DE,$F7,$3F,$CD
	.dc.b	$C7,$F9,$BF,$DD,$B3,$FE,$76,$8B,$5F,$E3,$A9,$B9,$7F,$8A,$26,$D7
	.dc.b	$F8,$7E,$6D,$7F,$A1,$3E,$1D,$F0,$3F,$E7,$D6,$D7,$F9,$4F,$E0,$7F
	.dc.b	$D1,$4F,$F9,$56,$33,$5F,$A9,$C5,$0E,$7E,$47,$3E,$7B,$7F,$E5,$DF
	.dc.b	$A3,$7D,$1F,$FA,$2D,$B1,$9F,$3B,$13,$FE,$8F,$F8,$97,$2D,$F8,$4C
	.dc.b	$36,$FF,$CD,$1F,$47,$FC,$7B,$6F,$22,$F2,$BF,$D3,$0E,$77,$FF,$49
	.dc.b	$39,$AE,$3E,$D6,$BD,$9D,$B6,$3F,$CE,$96,$F2,$3F,$D4,$4D,$8C,$F9
	.dc.b	$18,$9C,$FD,$1C,$9E,$3F,$47,$74,$F9,$1F,$47,$FC,$A5,$BE,$67,$FA
	.dc.b	$B6,$3F,$D2,$11,$FE,$B8,$75,$B7,$FA,$A7,$D6,$7C,$1F,$FA,$CF,$E8
	.dc.b	$FF,$6F,$97,$FA,$26,$72,$FF,$A2,$E3,$38,$7D,$79,$7C,$27,$FA,$ED
	.dc.b	$79,$16,$30,$FF,$AF,$FE,$8D,$F0,$17,$E0,$30,$DC,$BF,$3B,$F4,$6F
	.dc.b	$8E,$FF,$58,$B6,$3F,$43,$F4,$7F,$E2,$ED,$13,$97,$F9,$90,$8D,$1B
	.dc.b	$FA,$3F,$B1,$5F,$C4,$E3,$1E,$D7,$FA,$FA,$73,$5F,$ED,$FF,$73,$FD
	.dc.b	$8A,$FF,$B8,$63,$2B,$E1,$FF,$89,$F1,$FE,$EA,$31,$E9,$7F,$B2,$3E
	.dc.b	$89,$EC,$7F,$B2,$AF,$7A,$7F,$23,$9C,$BF,$BC,$8C,$DF,$FB,$BE,$72
	.dc.b	$FB,$23,$2A,$33,$87,$85,$B5,$FE,$E6,$FA,$3F,$D1,$D7,$C4,$9C,$BF
	.dc.b	$EF,$BF,$A0,$7D,$17,$FB,$79,$B7,$FE,$C9,$F9,$CF,$C4,$D2,$72,$FC
	.dc.b	$E6,$E3,$F0,$5B,$5F,$EF,$6A,$E3,$18,$AF,$05,$F7,$3F,$DF,$55,$BF
	.dc.b	$A3,$C3,$FC,$7A,$BF,$16,$DE,$7D,$DB,$17,$D1,$BE,$53,$D5,$52,$F8
	.dc.b	$1F,$EE,$39,$D3,$8B,$F5,$7C,$B7,$C1,$3E,$13,$63,$C5,$16,$3E,$01
	.dc.b	$70,$FC,$01,$FC,$32,$7F,$9A,$B6,$3F,$26,$07,$D1,$DF,$A2,$7D,$0F
	.dc.b	$F6,$5A,$E0,$74,$F3,$3D,$F7,$EF,$6B,$CE,$F9,$57,$19,$CC,$F6,$17
	.dc.b	$D9,$F1,$5F,$96,$31,$1F,$DD,$6F,$9B,$AA,$FE,$AB,$3A,$8F,$01,$D2
	.dc.b	$FC,$18,$70,$8E,$46,$4E,$4E,$5E,$33,$E0,$C3,$C5,$5F,$DE,$67,$D1
	.dc.b	$39,$BF,$0C,$E7,$0E,$D7,$C2,$FF,$C0,$11,$8A,$DC,$60,$F1,$1F,$27
	.dc.b	$0F,$13,$DA,$FD,$2E,$8F,$69,$FD,$AA,$BF,$B5,$B8,$3C,$47,$9F,$E8
	.dc.b	$9E,$9B,$E8,$FF,$B8,$C6,$1B,$19,$9E,$0D,$7F,$C3,$13,$FD,$7F,$2E
	.dc.b	$3E,$AF,$E7,$67,$B3,$5D,$B5,$FF,$16,$4F,$17,$DF,$8B,$5F,$F1,$86
	.dc.b	$BF,$7A,$31,$3F,$5E,$B8,$4F,$E0,$9B,$3F,$E3,$73,$73,$D2,$B9,$C7
	.dc.b	$82,$FC,$D7,$C5,$B1,$C6,$77,$5F,$A1,$8E,$75,$BC,$3C,$57,$C2,$B5
	.dc.b	$7C,$1F,$D2,$B7,$0F,$2C,$7E,$AD,$AE,$8F,$C9,$61,$D8,$F3,$5E,$2B
	.dc.b	$F5,$DA,$79,$6B,$FC,$57,$5E,$EA,$FE,$75,$7D,$23,$15,$F0,$FE,$53
	.dc.b	$8D,$FA,$DA,$F8,$4F,$EA,$3F,$47,$FE,$0E,$7F,$C6,$ED,$73,$BF,$08
	.dc.b	$FF,$0F,$46,$A3,$AE,$2E,$67,$CC,$E1,$B5,$B3,$FC,$87,$F4,$7F,$E6
	.dc.b	$B7,$FC,$8B,$F4,$5F,$AD,$CF,$FC,$E0,$D8,$FE,$25,$3B,$7F,$B3,$B7
	.dc.b	$0F,$EB,$53,$B9,$36,$75,$5F,$E7,$B5,$E8,$FC,$29,$FF,$00,$D7,$8B
	.dc.b	$9B,$E2,$D6,$A3,$CF,$5B,$99,$18,$21,$FD,$EA,$B9,$B6,$39,$CF,$D0
	.dc.b	$FA,$3C,$E3,$FD,$C7,$E0,$9E,$4F,$D4,$6E,$1F,$E6,$8A,$DE,$7F,$E9
	.dc.b	$39,$CC,$F7,$D7,$D4,$71,$BB,$1F,$E7,$0F,$53,$78,$FF,$7F,$6E,$1F
	.dc.b	$3A,$FC,$1F,$01,$F8,$4F,$A3,$F8,$F8,$7F,$41,$6B,$FE,$7E,$8D,$7A
	.dc.b	$DF,$27,$BA,$4F,$F9,$FA,$BF,$C8,$CF,$FA,$03,$E8,$DF,$31,$F9,$0F
	.dc.b	$81,$7B,$D4,$FC,$9C,$EA,$BF,$D0,$FF,$6B,$FE,$74,$7F,$D1,$3D,$1F
	.dc.b	$E1,$46,$F9,$5F,$9A,$9D,$47,$F8,$A9,$FC,$AB,$FE,$2C,$97,$CD,$FE
	.dc.b	$81,$B1,$DB,$C0,$FD,$1F,$DE,$BC,$8F,$CD,$FC,$AF,$E4,$79,$8F,$E3
	.dc.b	$67,$57,$FE,$9C,$F8,$CF,$C3,$3F,$E3,$AF,$B6,$7C,$A7,$EB,$9B,$1B
	.dc.b	$ED,$FD,$8D,$3E,$03,$5D,$92,$BA,$47,$ED,$BE,$F5,$F4,$DE,$4F,$6A
	.dc.b	$57,$FA,$B2,$72,$FF,$AB,$BF,$15,$AF,$F9,$33,$ED,$FF,$CB,$EF,$FA
	.dc.b	$C9,$AF,$F9,$42,$37,$C1,$FE,$43,$FB,$87,$CC,$FF,$06,$BF,$F0,$B5
	.dc.b	$F1,$EE,$FE,$13,$5A,$8E,$9D,$7F,$23,$CE,$91,$83,$FD,$83,$F7,$0F
	.dc.b	$8E,$FE,$23,$F7,$0F,$98,$A7,$F1,$49,$D5,$70,$F9,$6D,$F9,$32,$30
	.dc.b	$7F,$B3,$3E,$17,$CA,$FF,$AC,$FB,$3F,$DD,$9C,$3F,$ED,$63,$EC,$C7
	.dc.b	$F2,$09,$DC,$7F,$98,$7F,$0A,$7C,$DF,$F2,$46,$D7,$FB,$61,$AF,$F9
	.dc.b	$DB,$EB,$1F,$69,$BA,$FE,$57,$1F,$ED,$79,$FF,$6F,$B5,$FF,$3E,$D7
	.dc.b	$2B,$F0,$57,$0F,$99,$95,$DE,$39,$FC,$FF,$4F,$06,$4F,$48,$CB,$DA
	.dc.b	$32,$F9,$63,$C7,$F1,$72,$3C,$6F,$9B,$BF,$11,$EB,$8B,$8F,$FA,$0F
	.dc.b	$3F,$E0,$B3,$FE,$41,$FA,$DF,$9A,$FF,$90,$CC,$74,$FF,$83,$A3,$31
	.dc.b	$96,$BA,$3F,$E0,$0F,$C6,$FF,$35,$3F,$9B,$47,$D2,$F4,$F8,$57,$F5
	.dc.b	$49,$D7,$B9,$FD,$5A,$B5,$7F,$E5,$39,$7D,$6E,$C9,$1D,$BD,$5A,$FF
	.dc.b	$AB,$27,$6F,$8A,$7B,$4F,$7F,$91,$93,$E6,$7F,$B2,$FD,$6F,$F9,$9B
	.dc.b	$3F,$81,$87,$FB,$42,$B4,$43,$FD,$73,$3B,$7F,$CD,$1F,$58,$F9,$8F
	.dc.b	$F5,$84,$7F,$91,$E3,$1F,$19,$FE,$49,$84,$3E,$61,$F0,$79,$8D,$CF
	.dc.b	$F9,$3F,$EB,$FF,$8D,$E7,$C5,$B1,$9F,$57,$FD,$78,$FD,$78,$CB,$F6
	.dc.b	$47,$CE,$7F,$B7,$2B,$C4,$D8,$FF,$2E,$35,$E0,$D7,$FC,$E7,$F4,$1B
	.dc.b	$E6,$FD,$73,$6B,$3D,$1F,$EA,$09,$4E,$6B,$FD,$44,$D8,$CF,$46,$E3
	.dc.b	$FB,$1D,$FF,$A0,$0E,$F1,$CF,$E9,$7F,$DB,$B4,$6F,$D6,$FF,$B7,$6B
	.dc.b	$86,$1F,$EA,$F9,$D9,$34,$F2,$56,$E7,$FD,$69,$5F,$E8,$EA,$FF,$74
	.dc.b	$C3,$EB,$7F,$AC,$CF,$48,$C7,$BF,$E7,$FA,$2F,$F4,$FD,$6A,$7F,$DF
	.dc.b	$6D,$CB,$FD,$DC,$DB,$F9,$35,$C2,$E4,$75,$9D,$5F,$C6,$AC,$BE,$63
	.dc.b	$D9,$E5,$FC,$33,$E2,$7C,$AD,$BF,$83,$8F,$C5,$9C,$57,$18,$F9,$15
	.dc.b	$87,$EC,$FD,$6F,$D1,$F9,$CF,$A4,$DC,$39,$F0,$23,$18,$73,$8C,$4E
	.dc.b	$AB,$FD,$83,$F5,$8F,$AD,$A7,$D9,$E4,$B5,$F2,$E7,$4D,$9D,$34,$98
	.dc.b	$7C,$11,$3C,$1F,$67,$07,$E7,$27,$79,$9F,$E3,$15,$B8,$C5,$7F,$B7
	.dc.b	$63,$13,$F8,$1C,$3C,$B3,$DE,$72,$EF,$0F,$2B,$97,$77,$D2,$3C,$B1
	.dc.b	$A8,$C4,$F2,$9C,$DF,$98,$F0,$15,$EC,$D9,$FC,$F5,$F6,$8F,$5D,$DA
	.dc.b	$FC,$F4,$D7,$A9,$E5,$9F,$C1,$E1,$F1,$63,$04,$BF,$2D,$F9,$EB,$A9
	.dc.b	$E3,$9E,$73,$FD,$AE,$B9,$3C,$27,$58,$7C,$68,$F3,$C6,$2F,$CD,$18
	.dc.b	$BF,$04,$F6,$69,$EA,$78,$1F,$06,$FB,$BB,$9F,$04,$F0,$7A,$18,$24
	.dc.b	$6C,$F9,$B1,$D8,$F5,$B9,$7C,$06,$6F,$EB,$1E,$BA,$F0,$C7,$D4,$35
	.dc.b	$7F,$1D,$C5,$F8,$1B,$7E,$08,$C4,$78,$0C,$9E,$07,$37,$E0,$AE,$93
	.dc.b	$B2,$47,$8F,$0F,$BC,$EE,$FE,$EC,$62,$BE,$7E,$1E,$8B,$E3,$18,$9D
	.dc.b	$57,$CD,$6C,$FC,$BA,$E1,$7D,$63,$13,$AC,$3D,$C7,$D0,$BF,$95,$3D
	.dc.b	$9F,$19,$F6,$9E,$2D,$8E,$37,$DB,$0E,$93,$F1,$CE,$78,$78,$6B,$F2
	.dc.b	$18,$6B,$0F,$93,$1E,$17,$E5,$9E,$13,$F8,$1C,$6D,$FA,$11,$B7,$4F
	.dc.b	$8A,$35,$3F,$36,$BC,$C7,$E1,$67,$84,$78,$E3,$18,$7A,$5D,$DF,$DB
	.dc.b	$7D,$38,$7C,$F7,$31,$E2,$9E,$0F,$E2,$A3,$E3,$B4,$F8,$E5,$BF,$34
	.dc.b	$7F,$03,$C9,$F5,$9F,$D9,$FE,$73,$FB,$43,$67,$F9,$BB,$6F,$F9,$B3
	.dc.b	$6F,$F7,$5F,$3B,$FD,$AC,$FE,$D9,$E7,$DB,$FE,$0F,$16,$FF,$96,$CD
	.dc.b	$AF,$E3,$7F,$59,$FC,$F8,$FE,$F1,$EC,$DC,$3F,$9E,$45,$BF,$E8,$D3
	.dc.b	$6B,$8D,$68,$FE,$9B,$19,$7F,$B9,$7C,$5F,$F7,$FF,$81,$FF,$00,$36
	.dc.b	$3F,$8D,$B6,$FF,$A0,$B6,$FF,$C2,$7E,$67,$F8,$2C,$FF,$06,$79,$B6
	.dc.b	$FF,$AD,$C5,$BF,$EB,$53,$6B,$F9,$D7,$D6,$7F,$A8,$AF,$FC,$00,$DA
	.dc.b	$FE,$9F,$D7,$FE,$8D,$9F,$F3,$24,$E6,$B8,$E7,$9B,$09,$D5,$7F,$83
	.dc.b	$EF,$53,$9A,$B5,$A8,$E2,$DB,$FF,$10,$56,$AB,$FC,$51,$F0,$D3,$0F
	.dc.b	$EE,$FF,$2F,$71,$FE,$ED,$37,$0F,$F1,$BC,$DA,$FF,$10,$7D,$67,$F8
	.dc.b	$B7,$D2,$FF,$17,$36,$3F,$A4,$B7,$0F,$ED,$ED,$BF,$EC,$1F,$56,$ED
	.dc.b	$FF,$28,$D7,$C0,$8C,$5F,$F9,$7D,$B1,$FE,$53,$0E,$1F,$F2,$B3,$AA
	.dc.b	$EB,$5F,$E4,$5F,$A9,$F8,$F3,$9B,$FF,$4A,$36,$FC,$15,$B9,$FF,$1C
	.dc.b	$4E,$6F,$FC,$91,$EE,$7F,$92,$CF,$F2,$67,$D6,$5A,$FE,$CF,$17,$0F
	.dc.b	$F0,$6C,$DA,$FF,$2F,$FC,$47,$F9,$5F,$E3,$3F,$CB,$0D,$C3,$FC,$4E
	.dc.b	$DB,$FF,$13,$36,$BE,$3F,$C9,$FC,$69,$F8,$2D,$C8,$D2,$5F,$A9,$FF
	.dc.b	$31,$1F,$08,$8E,$4F,$EF,$EE,$27,$FC,$51,$E6,$BE,$A3,$F1,$ED,$AF
	.dc.b	$F2,$3F,$8F,$FE,$90,$7F,$CB,$D1,$9A,$FF,$18,$7D,$F3,$EB,$7F,$2E
	.dc.b	$DA,$FF,$27,$FC,$A3,$AF,$35,$F9,$CF,$A1,$FF,$35,$9F,$E7,$8F,$AD
	.dc.b	$FA,$D8,$A3,$FC,$DF,$8B,$83,$A9,$FE,$89,$F8,$97,$56,$34,$FF,$39
	.dc.b	$7D,$13,$C1,$FD,$4B,$63,$FC,$C3,$F5,$BF,$D8,$63,$19,$FF,$92,$7E
	.dc.b	$9B,$FC,$FA,$7F,$A2,$A7,$E7,$D7,$F9,$F6,$3F,$D0,$3F,$70,$FA,$7F
	.dc.b	$DA,$B6,$7F,$CD,$BE,$3B,$C8,$CD,$6B,$E5,$25,$6B,$DB,$B9,$FF,$96
	.dc.b	$69,$BF,$F0,$06,$D7,$F9,$E3,$EB,$7F,$DB,$AF,$FA,$3E,$33,$5F,$E6
	.dc.b	$18,$78,$BF,$C2,$1A,$7F,$A0,$3E,$B5,$F5,$D9,$9F,$E1,$BF,$7A,$F6
	.dc.b	$29,$FE,$B5,$F5,$FE,$F7,$87,$FE,$59,$9D,$63,$F0,$F8,$7F,$8A,$F1
	.dc.b	$9F,$F5,$E7,$C7,$7E,$07,$5F,$99,$1F,$C6,$BE,$BD,$F3,$34,$FE,$3B
	.dc.b	$F5,$FF,$D6,$31,$FF,$55,$4E,$9F,$F5,$64,$66,$FF,$CF,$CD,$CF,$FC
	.dc.b	$D9,$F5,$FF,$3E,$B7,$EB,$7F,$28,$8C,$5F,$FA,$15,$B5,$FE,$73,$F5
	.dc.b	$FF,$D7,$71,$8B,$FE,$67,$4F,$85,$FC,$99,$F9,$4D,$9F,$43,$FC,$D2
	.dc.b	$30,$DF,$43,$E1,$B6,$FF,$10,$E1,$FF,$58,$FD,$7F,$F8,$C6,$F9,$C7
	.dc.b	$FB,$27,$EB,$FF,$15,$8F,$FA,$6D,$BA,$7F,$A5,$07,$D5,$BE,$8F,$FA
	.dc.b	$2C,$EA,$FD,$2D,$BF,$E9,$AD,$BE,$35,$C3,$AE,$5F,$85,$FF,$7A,$61
	.dc.b	$D6,$30,$79,$7E,$A7,$B9,$FE,$F9,$F2,$5F,$29,$89,$D1,$33,$FF,$66
	.dc.b	$4E,$A3,$19,$7F,$A8,$F3,$7C,$EF,$A8,$F5,$DD,$AF,$89,$16,$3E,$21
	.dc.b	$6B,$FD,$58,$3D,$5F,$EC,$ED,$AF,$F7,$A4,$FF,$6E,$F2,$3F,$B7,$FD
	.dc.b	$73,$E5,$BE,$FD,$EE,$BF,$D9,$4D,$3A,$E1,$FD,$CA,$39,$0A,$E4,$FF
	.dc.b	$7A,$9E,$18,$7F,$59,$C3,$BE,$1D,$B0,$FF,$66,$E1,$C3,$0F,$C7,$D6
	.dc.b	$9E,$53,$AA,$E5,$1D,$DE,$57,$B8,$E2,$64,$E2,$E5,$F3,$DF,$7C,$3D
	.dc.b	$CF,$86,$33,$5D,$1E,$47,$48,$E1,$7E,$38,$FD,$28,$C3,$C5,$7D,$2F
	.dc.b	$CF,$7C,$30,$F1,$5E,$CE,$B5,$B7,$9D,$FF,$7F,$7D,$65,$70,$FD,$03
	.dc.b	$C3,$F5,$DF,$C7,$BE,$BF,$83,$63,$FC,$10,$D7,$7E,$1A,$77,$E7,$30
	.dc.b	$A4,$AE,$31,$96,$C7,$CF,$CF,$F2,$F1,$75,$FE,$0F,$A7,$DB,$CB,$FA
	.dc.b	$9D,$FD,$56,$D7,$CC,$BE,$2D,$BE,$31,$FC,$42,$77,$7C,$F0,$F1,$3F
	.dc.b	$87,$9E,$2D,$AE,$37,$FE,$4A,$BE,$35,$CE,$F8,$BF,$C7,$0E,$25,$AE
	.dc.b	$34,$AE,$2F,$3A,$E2,$FF,$95,$AB,$8B,$6B,$8D,$70,$8F,$D6,$47,$EB
	.dc.b	$34,$ED,$1F,$A5,$6B,$F4,$5F,$36,$3F,$2E,$34,$77,$F5,$79,$34,$C9
	.dc.b	$D8,$BF,$5F,$FC,$05,$BB,$F0,$51,$97,$C0,$7F,$9C,$1F,$F0,$4B,$C0
	.dc.b	$F0,$D7,$F2,$EF,$97,$F0,$B6,$34,$72,$1F,$5F,$FE,$5D,$7C,$BF,$5F
	.dc.b	$FA,$FC,$BF,$A5,$B6,$3F,$CF,$F3,$B7,$93,$69,$5C,$35,$F5,$5F,$1B
	.dc.b	$EF,$DB,$ED,$C3,$FC,$3D,$1F,$A9,$6B,$FE,$69,$6B,$FB,$86,$BF,$D7
	.dc.b	$9D,$E1,$E0,$8F,$F4,$9D,$39,$FC,$27,$7F,$4E,$4F,$BA,$32,$FD,$4B
	.dc.b	$E7,$97,$F6,$AF,$AF,$E5,$3F,$BD,$6B,$FE,$7D,$6B,$FC,$31,$AF,$F7
	.dc.b	$2F,$75,$F2,$9F,$C0,$9B,$1F,$C1,$DA,$FF,$A1,$DA,$FF,$16,$C9,$B7
	.dc.b	$F0,$A6,$C7,$F0,$F6,$BF,$E8,$F6,$BF,$C7,$A5,$F5,$FE,$12,$47,$85
	.dc.b	$FE,$2B,$F5,$EF,$8B,$A4,$9F,$E3,$91,$97,$F8,$F7,$D7,$BE,$2B,$F8
	.dc.b	$EB,$63,$31,$83,$F9,$0F,$D7,$FF,$9B,$63,$F9,$33,$5F,$F5,$33,$5F
	.dc.b	$E6,$AD,$7F,$C3,$6F,$E7,$9D,$F5,$74,$23,$FC,$C7,$EB,$FF,$C1,$8D
	.dc.b	$F1,$7F,$98,$C6,$23,$FD,$68,$D7,$FA,$1F,$BD,$F8,$C7,$FD,$B6,$D7
	.dc.b	$FC,$F1,$CB,$FC,$EA,$30,$3E,$BB,$04,$9F,$E8,$51,$97,$FA,$27,$D7
	.dc.b	$3D,$DF,$E8,$6D,$8C,$C6,$0F,$E8,$FF,$5D,$98,$FF,$65,$35,$FE,$B2
	.dc.b	$D7,$FC,$7E,$D7,$FD,$E7,$F5,$AF,$90,$FE,$A2,$D8,$FE,$AC,$D7,$FD
	.dc.b	$AC,$D7,$FB,$2F,$D7,$7F,$9E,$9F,$EB,$8D,$7F,$DB,$8D,$77,$F0,$1F
	.dc.b	$D7,$BE,$B7,$FC,$E5,$1F,$D8,$DA,$FF,$B8,$DA,$FF,$72,$6B,$FE,$58
	.dc.b	$5F,$5B,$CC,$91,$CD,$FE,$D7,$EE,$7F,$8E,$23,$FB,$83,$5F,$F7,$83
	.dc.b	$5F,$EF,$ED,$7F,$CC,$BF,$56,$F9,$4F,$E5,$CD,$8F,$E6,$AD,$7F,$93
	.dc.b	$7D,$53,$EB,$37,$5F,$DF,$2B,$55,$FE,$0C,$6B,$DA,$34,$7F,$3B,$A7
	.dc.b	$D6,$53,$9C,$FF,$44,$8F,$F6,$73,$7E,$67,$81,$C7,$E4,$24,$64,$87
	.dc.b	$99,$ED,$F4,$92,$3A,$3E,$1F,$13,$11,$EC,$8B,$E4,$FF,$29,$86,$1F
	.dc.b	$C5,$23,$1F,$59,$F3,$5B,$EC,$7F,$55,$8C,$79,$3F,$E9,$0F,$AA,$FF
	.dc.b	$52,$E1,$FD,$7A,$7E,$82,$0E,$1B,$5F,$E2,$68,$C1,$D2,$38,$4F,$F4
	.dc.b	$93,$FA,$54,$66,$10,$FA,$CF,$ED,$0D,$8F,$ED,$ED,$7F,$B0,$3F,$CA
	.dc.b	$30,$F8,$0E,$5B,$3C,$8E,$7E,$57,$4F,$AC,$97,$D1,$FE,$ED,$F5,$9F
	.dc.b	$E7,$AC,$FE,$14,$6A,$7B,$BA,$8F,$EF,$9F,$63,$E8,$9F,$F2,$43,$5F
	.dc.b	$F2,$A8,$70,$FF,$96,$1A,$FF,$AF,$BE,$B1,$F5,$BF,$EB,$F8,$E3,$18
	.dc.b	$F8,$5F,$F2,$A2,$F9,$37,$B3,$8B,$E1,$7F,$E6,$4F,$C5,$FF,$D9,$19
	.dc.b	$7F,$96,$DB,$1E,$46,$BB,$C7,$F1,$0F,$FA,$27,$E9,$5C,$BF,$EB,$FB
	.dc.b	$C3,$67,$FC,$B1,$3F,$D9,$27,$FC,$2E,$F0,$6C,$EF,$85,$39,$AF,$F1
	.dc.b	$0C,$63,$9B,$53,$98,$FF,$67,$7D,$33,$EB,$79,$37,$1F,$F1,$2F,$D6
	.dc.b	$FB,$2B,$75,$A2,$4E,$B3,$B9,$7F,$89,$FE,$B7,$FB,$04,$FF,$8B,$BE
	.dc.b	$BF,$FD,$9D,$39,$F4,$FD,$11,$8C,$7F,$59,$7C,$2B,$EE,$B5,$FF,$49
	.dc.b	$B5,$FF,$7C,$B5,$FD,$43,$C6,$39,$D7,$C0,$FA,$2F,$99,$4A,$D6,$1F
	.dc.b	$EB,$BD,$1F,$31,$F0,$8F,$87,$EF,$D9,$FF,$23,$CE,$9B,$3A,$69,$F5
	.dc.b	$BE,$BB,$E7,$FC,$CF,$D0,$6C,$64,$FB,$71,$96,$9F,$E6,$0F,$AD,$7D
	.dc.b	$6F,$DD,$FB,$27,$CB,$7A,$A7,$35,$FE,$5A,$9C,$BC,$6B,$85,$6D,$B1
	.dc.b	$93,$84,$65,$A6,$75,$FC,$4D,$74,$7F,$CC,$57,$E0,$73,$5C,$5B,$5C
	.dc.b	$78,$3F,$C9,$C2,$F9,$3F,$CA,$6F,$93,$A8,$CC,$FF,$21,$A5,$6A,$3F
	.dc.b	$DE,$13,$A9,$FE,$4D,$39,$7F,$D7,$D1,$97,$8E,$1F,$01,$F4,$B6,$FC
	.dc.b	$4D,$BF,$05,$F3,$9F,$46,$1E,$D7,$CC,$D9,$F1,$B6,$FC,$2D,$9E,$33
	.dc.b	$F0,$CD,$96,$76,$D8,$DC,$79,$A3,$E9,$B4,$F9,$A5,$8F,$1C,$71,$C3
	.dc.b	$E0,$5F,$A5,$B8,$7B,$6F,$E3,$B5,$F8,$B5,$DA,$FE,$1B,$4F,$87,$3B
	.dc.b	$BE,$98,$7B,$70,$EB,$5B,$C3,$AB,$5E,$A7,$A3,$0E,$AD,$8E,$A6,$EF
	.dc.b	$AB,$EB,$BE,$C4,$70,$49,$F8,$4F,$CC,$7E,$31,$F8,$93,$C2,$59,$FC
	.dc.b	$C3,$70,$FE,$66,$FA,$0F,$E9,$71,$A7,$CA,$62,$36,$FB,$2F,$D6,$7F
	.dc.b	$2A,$FA,$FF,$C1,$46,$6F,$8C,$E5,$FC,$04,$64,$F0,$46,$5F,$59,$88
	.dc.b	$E2,$7F,$15,$9E,$B7,$F9,$F9,$E5,$4D,$55,$AC,$3F,$25,$18,$C3,$C3
	.dc.b	$87,$13,$F3,$D1,$9C,$3C,$27,$94,$EA,$E7,$0F,$7D,$79,$27,$CC,$23
	.dc.b	$C5,$87,$86,$7F,$39,$5C,$5E,$15,$D6,$FC,$35,$B8,$F8,$37,$CA,$BC
	.dc.b	$F8,$7A,$AB,$53,$E2,$AF,$06,$1D,$5F,$19,$D6,$BF,$AD,$D7,$0B,$F4
	.dc.b	$5F,$5A,$F7,$E1,$F0,$EF,$75,$E0,$3D,$37,$C6,$BD,$F7,$C4,$F1,$3D
	.dc.b	$E7,$FA,$2C,$EE,$3E,$4B,$6F,$DF,$3D,$67,$DF,$87,$A5,$E4,$62,$BF
	.dc.b	$85,$73,$3E,$63,$FB,$CE,$7F,$DE,$9B,$87,$F8,$11,$B7,$FE,$00,$AE
	.dc.b	$13,$FE,$08,$AF,$EF,$BF,$43,$23,$FA,$DF,$CE,$DB,$FF,$03,$E4,$AC
	.dc.b	$69,$FE,$06,$AB,$7F,$C9,$BE,$BF,$FB,$CF,$D4,$FF,$53,$6C,$7F,$59
	.dc.b	$7F,$4F,$B1,$A4,$8F,$F1,$CE,$5F,$CF,$A3,$94,$7F,$87,$BD,$8B,$3E
	.dc.b	$4F,$85,$FF,$0A,$FB,$FF,$E1,$66,$D7,$F7,$46,$CE,$E3,$FB,$BC,$FF
	.dc.b	$88,$3C,$F9,$1F,$E2,$AF,$42,$D7,$C7,$9B,$AF,$F5,$8E,$15,$62,$FF
	.dc.b	$AA,$7D,$7F,$FB,$1D,$FF,$01,$FC,$E7,$F8,$32,$31,$8F,$CA,$8C,$69
	.dc.b	$FC,$22,$73,$3F,$C5,$5A,$F4,$8D,$1C,$2B,$FC,$1F,$DD,$67,$EE,$7C
	.dc.b	$27,$F8,$E8,$FF,$1D,$FC,$AD,$AF,$F2,$2C,$59,$FC,$14,$DD,$3F,$C7
	.dc.b	$FF,$5F,$FE,$3E,$F5,$FF,$C7,$ED,$C3,$FC,$3A,$DB,$FC,$1B,$6B,$FD
	.dc.b	$41,$F5,$8F,$07,$FC,$73,$F2,$1F,$E2,$C8,$C6,$1F,$E3,$F8,$C4,$FF
	.dc.b	$28,$9C,$E1,$FE,$78,$6B,$FC,$DD,$AF,$F8,$05,$DD,$F8,$27,$FC,$B3
	.dc.b	$F5,$FD,$30,$FF,$19,$FD,$7F,$F9,$60,$FF,$2C,$FD,$7D,$8C,$93,$1F
	.dc.b	$CE,$C6,$5F,$D3,$7C,$47,$F9,$93,$E5,$BF,$CC,$AD,$AF,$F3,$7B,$67
	.dc.b	$71,$FE,$72,$9F,$81,$F2,$3F,$E6,$8F,$83,$B3,$FE,$46,$6D,$7B,$E7
	.dc.b	$E1,$12,$7F,$43,$AA,$37,$AF,$F9,$D1,$B3,$FE,$23,$F5,$1D,$FF,$D4
	.dc.b	$9B,$1F,$AA,$6E,$1F,$EA,$46,$BF,$D8,$FE,$B1,$F5,$DF,$D6,$9B,$5F
	.dc.b	$B2,$6C,$FF,$AB,$1A,$FF,$6B,$F9,$FF,$F4,$23,$FE,$B0,$6B,$9F,$4B
	.dc.b	$FB,$2C,$62,$3F,$C7,$52,$DF,$C0,$4B,$F0,$3F,$DB,$3E,$79,$E0,$D2
	.dc.b	$4F,$F0,$68,$CB,$FC,$23,$E8,$BB,$61,$FE,$E7,$6A,$69,$8F,$AC,$7D
	.dc.b	$1F,$F7,$A6,$D7,$F1,$26,$D7,$FB,$31,$AF,$F8,$2D,$7C,$87,$12,$4F
	.dc.b	$17,$FC,$0B,$4F,$8D,$A4,$9F,$E3,$F1,$97,$F9,$08,$EE,$FF,$06,$B4
	.dc.b	$CB,$83,$FC,$23,$F5,$3F,$E9,$1C,$7F,$DC,$2D,$7F,$C4,$0D,$7F,$CD
	.dc.b	$4D,$7E,$37,$97,$FE,$1A,$CC,$D2,$5F,$F3,$4F,$A8,$7D,$4F,$F8,$81
	.dc.b	$B1,$98,$C1,$FE,$24,$FA,$9F,$E9,$F8,$7F,$BD,$1A,$FF,$8C,$DA,$FF
	.dc.b	$9F,$5A,$87,$D4,$FF,$51,$AF,$E8,$AD,$8F,$F7,$FB,$5F,$F1,$F3,$5F
	.dc.b	$F4,$47,$BB,$FC,$29,$E8,$D4,$D2,$78,$5F,$EC,$47,$F8,$7D,$7F,$A4
	.dc.b	$73,$79,$59,$AF,$F6,$4C,$62,$7F,$25,$3F,$2C,$8F,$F5,$CF,$A7,$FC
	.dc.b	$4E,$1F,$E9,$B6,$BE,$F6,$A6,$9F,$92,$EE,$4F,$1E,$1F,$F7,$F7,$D4
	.dc.b	$7F,$B9,$E3,$07,$C9,$9D,$DF,$5A,$FF,$6F,$DF,$3A,$D5,$F1,$AD,$47
	.dc.b	$13,$CD,$E3,$7F,$97,$A8,$D2,$47,$B5,$EA,$7F,$AF,$9F,$17,$5B,$E3
	.dc.b	$B3,$1D,$6F,$E0,$7D,$2F,$90,$F2,$B6,$34,$FC,$38,$C3,$FE,$4D,$8F
	.dc.b	$94,$E3,$3F,$7B,$FE,$51,$F1,$7F,$DF,$B0,$7F,$DF,$F1,$F8,$DE,$BF
	.dc.b	$C7,$7D,$3B,$E9,$7F,$D9,$D7,$BC,$3C,$2D,$BF,$01,$F3,$70,$E6,$B0
	.dc.b	$E7,$7C,$A7,$31,$89,$E1,$75,$FF,$6E,$45,$BE,$D8,$78,$E7,$FD,$31
	.dc.b	$39,$BF,$B7,$39,$9E,$21,$C5,$72,$C3,$FD,$41,$3E,$53,$4F,$97,$0F
	.dc.b	$14,$7C,$33,$27,$98,$A7,$6A,$EB,$7E,$2A,$EB,$87,$96,$F9,$DF,$7B
	.dc.b	$DC,$62,$75,$3D,$0F,$01,$FC,$EC,$FF,$20,$9E,$73,$07,$FA,$E7,$E9
	.dc.b	$FF,$92,$7D,$47,$F9,$19,$B1,$EE,$6B,$BE,$C4,$EF,$CF,$FF,$4A,$4E
	.dc.b	$DA,$F5,$6B,$FC,$44,$52,$33,$1F,$C5,$63,$1A,$7D,$BC,$DA,$F1,$D3
	.dc.b	$EB,$46,$2F,$F0,$17,$C1,$F9,$AD,$9D,$D7,$1A,$F1,$1C,$6B,$78,$7A
	.dc.b	$AB,$F0,$57,$C6,$B7,$7C,$6F,$F9,$54,$71,$AD,$DF,$18,$F7,$5F,$1A
	.dc.b	$DD,$F1,$7F,$48,$D8,$D4,$7C,$4A,$E4,$F0,$3E,$14,$EE,$33,$1C,$8F
	.dc.b	$DD,$D7,$AF,$0F,$DB,$B7,$4F,$1D,$70,$8C,$57,$AB,$C6,$FD,$AD,$7C
	.dc.b	$46,$C7,$84,$FD,$A8,$FA,$8F,$E0,$0F,$F0,$A6,$BA,$6E,$1F,$C2,$1F
	.dc.b	$BD,$1B,$DF,$AC,$B9,$1F,$23,$86,$D6,$CF,$E2,$FF,$51,$FE,$63,$7F
	.dc.b	$8C,$FD,$3F,$ED,$33,$FC,$63,$63,$FA,$F4,$ED,$F4,$B7,$0F,$E3,$73
	.dc.b	$B9,$36,$75,$5F,$97,$7F,$C4,$F3,$E2,$F1,$BE,$2A,$8C,$BF,$1A,$9F
	.dc.b	$69,$A9,$CE,$1E,$0E,$8E,$35,$FD,$02,$FA,$10,$E5,$1A,$39,$61,$E0
	.dc.b	$FA,$9E,$13,$F5,$1B,$7F,$CD,$9B,$97,$C4,$FA,$5F,$3F,$9E,$F8,$EA
	.dc.b	$4C,$3F,$A1,$DF,$07,$FC,$0D,$9F,$D2,$7D,$9B,$FA,$0E,$0F,$E1,$2D
	.dc.b	$F5,$3D,$89,$1F,$D3,$A1,$F1,$3F,$23,$EA,$5F,$03,$49,$5F,$D5,$23
	.dc.b	$E8,$BF,$E1,$7F,$A9,$7B,$BF,$41,$B1,$FD,$69,$AF,$D1,$F0,$7E,$AF
	.dc.b	$D4,$FF,$9C,$A3,$FA,$FB,$5F,$CD,$B7,$8B,$FA,$EF,$43,$DE,$E3,$DB
	.dc.b	$2F,$8B,$F8,$09,$F2,$C3,$EA,$5C,$BF,$84,$6C,$7F,$6E,$6B,$FD,$7B
	.dc.b	$07,$C9,$3E,$DB,$F1,$7E,$BF,$E1,$BA,$7F,$93,$36,$FF,$C7,$0D,$5F
	.dc.b	$69,$FE,$5B,$7F,$BD,$C3,$EA,$F9,$92,$3F,$BF,$CB,$E5,$FF,$33,$F5
	.dc.b	$7F,$E6,$E9,$FF,$01,$B5,$FD,$FB,$73,$FF,$0F,$9E,$8E,$2F,$D0,$FD
	.dc.b	$5E,$73,$FE,$78,$DB,$FF,$27,$35,$7D,$57,$E9,$1F,$F0,$8F,$D5,$BE
	.dc.b	$27,$F5,$6D,$8C,$C7,$EB,$A5,$F3,$77,$DE,$95,$FE,$1A,$9F,$F4,$13
	.dc.b	$FE,$21,$CF,$E6,$D7,$ED,$BE,$AD,$F3,$54,$95,$FE,$22,$FB,$47,$C4
	.dc.b	$EC,$91,$FE,$26,$AC,$BB,$F2,$3F,$80,$FD,$67,$F8,$C6,$3F,$C5,$6D
	.dc.b	$7F,$93,$B7,$2F,$93,$D1,$FA,$B7,$FC,$57,$F5,$7F,$8F,$AF,$F1,$9B
	.dc.b	$5F,$E5,$ED,$8F,$F1,$8F,$D5,$3E,$B3,$F8,$8C,$E7,$93,$FB,$63,$6F
	.dc.b	$FC,$FC,$D5,$5C,$5E,$FE,$07,$F1,$A8,$7D,$65,$3A,$C6,$7E,$5B,$5F
	.dc.b	$2B,$2F,$FC,$93,$69,$D4,$62,$74,$FF,$48,$9D,$3C,$3A,$FC,$04,$79
	.dc.b	$F9,$F5,$E1,$17,$5F,$83,$3F,$CC,$3E,$B3,$E6,$6B,$FE,$0B,$6D,$FF
	.dc.b	$A7,$5A,$BE,$B3,$FD,$5A,$FF,$98,$23,$24,$ED,$FE,$77,$3A,$9F,$EB
	.dc.b	$F3,$CA,$3F,$A3,$1F,$D1,$DC,$7D,$67,$02,$4F,$F9,$B2,$7E,$F5,$FF
	.dc.b	$87,$D4,$EA,$7F,$B5,$C6,$1E,$1E,$87,$8B,$EA,$E4,$73,$7F,$A7,$FD
	.dc.b	$67,$F9,$CB,$4F,$F3,$9D,$A7,$75,$D1,$DE,$3F,$D5,$BE,$F9,$ED,$FF
	.dc.b	$90,$9B,$5F,$E5,$A0,$E1,$FF,$2E,$B5,$FF,$1C,$0F,$AD,$FD,$D3,$FE
	.dc.b	$62,$6B,$A6,$BF,$E3,$F6,$BF,$ED,$1F,$AD,$7D,$67,$FB,$1A,$3F,$CD
	.dc.b	$71,$8F,$8C,$F2,$2F,$AD,$F1,$4F,$1A,$E0,$FF,$A3,$3E,$83,$F8,$E6
	.dc.b	$7F,$C0,$5B,$1F,$CB,$DF,$EC,$11,$9E,$5F,$F5,$54,$B7,$57,$07,$F8
	.dc.b	$FB,$6B,$FD,$80,$F2,$9E,$95,$CA,$9F,$48,$F9,$9C,$37,$0F,$E2,$9E
	.dc.b	$37,$FB,$91,$FE,$2D,$19,$C3,$FA,$23,$6F,$F9,$7B,$67,$FD,$A1,$5C
	.dc.b	$6B,$77,$EA,$5F,$57,$FD,$9A,$BF,$91,$78,$3F,$ED,$19,$CF,$B1,$E7
	.dc.b	$8C,$56,$E7,$FA,$A4,$62,$75,$EA,$75,$AD,$67,$FC,$6A,$31,$3A,$C3
	.dc.b	$D7,$E4,$FC,$03,$E0,$FC,$8D,$8C,$93,$2F,$7C,$65,$FA,$DF,$57,$F2
	.dc.b	$F8,$3F,$C0,$F5,$F6,$E3,$1E,$EE,$5B,$19,$3F,$01,$1F,$0F,$E9,$5F
	.dc.b	$55,$F5,$7E,$AD,$F5,$39,$8C,$36,$FF,$A4,$4E,$9F,$F7,$2D,$72,$73
	.dc.b	$3B,$CB,$FB,$D5,$6D,$FF,$32,$D6,$DC,$CE,$F7,$FE,$D3,$5B,$7F,$DD
	.dc.b	$75,$B7,$33,$BB,$5F,$E5,$6F,$84,$7F,$AA,$6F,$83,$AB,$4E,$AB,$8D
	.dc.b	$69,$E5,$7D,$5B,$7E,$EC,$3E,$1B,$4F,$79,$63,$D6,$DC,$3C,$ED,$3C
	.dc.b	$58,$78,$65,$87,$86,$BC,$B8,$78,$5B,$7C,$6B,$C6,$7C,$AC,$3C,$6D
	.dc.b	$8F,$1B,$4F,$19,$67,$C6,$D8,$F1,$D7,$7C,$3C,$98,$78,$DB,$87,$6C
	.dc.b	$3E,$9B,$6F,$E7,$37,$0F,$96,$D7,$DA,$DB,$E9,$87,$72,$57,$B6,$FD
	.dc.b	$F7,$E5,$3D,$EF,$82,$3D,$AD,$8F,$6B,$4F,$69,$63,$DA,$75,$C3,$D6
	.dc.b	$F2,$3D,$25,$9F,$4B,$5F,$4B,$6F,$D2,$F3,$BF,$4C,$F9,$EF,$D2,$78
	.dc.b	$5F,$4C,$F6,$7C,$ED,$8F,$3B,$4F,$3C,$73,$BF,$3B,$C0,$F2,$94,$F2
	.dc.b	$96,$FC,$AF,$1B,$F2,$DF,$8C,$F2,$DF,$79,$F2,$9D,$23,$C6,$D8,$F1
	.dc.b	$C7,$1B,$F1,$BA,$9F,$0B,$6F,$C2,$EE,$FC,$37,$DE,$3C,$35,$D4,$F0
	.dc.b	$9C,$A3,$B9,$BB,$EE,$E2,$38,$4F,$27,$81,$CA,$FC,$33,$C6,$FC,$33
	.dc.b	$C2,$F9,$BB,$C3,$F2,$84,$AF,$BC,$D8,$F9,$8D,$C3,$DC,$D9,$F2,$4F
	.dc.b	$8A,$7C,$98,$78,$A3,$C1,$87,$89,$D6,$1E,$2A,$DE,$1C,$E7,$E9,$C7
	.dc.b	$F6,$EF,$AA,$FE,$7D,$7F,$CF,$DB,$8F,$F7,$96,$D7,$F7,$96,$D7,$F6
	.dc.b	$4F,$AA,$FE,$DA,$7F,$6E,$F9,$DB,$7F,$DB,$22,$DF,$F6,$C9,$B5,$FC
	.dc.b	$C3,$EA,$BF,$B1,$7D,$27,$F4,$76,$E1,$FD,$01,$B7,$FC,$F1,$B5,$C6
	.dc.b	$7F,$C5,$1A,$A5,$18,$6D,$7F,$8A,$0F,$E4,$6E,$1F,$E7,$A7,$F8,$47
	.dc.b	$A7,$FD,$C7,$3F,$E0,$4F,$4B,$FC,$13,$F0,$DF,$E0,$A6,$C6,$4F,$F0
	.dc.b	$4C,$65,$FF,$10,$79,$DF,$DA,$4F,$ED,$5E,$75,$8C,$93,$0F,$F1,$24
	.dc.b	$65,$FF,$11,$47,$93,$EA,$BF,$D6,$8F,$F8,$5B,$E6,$FF,$C1,$D1,$89
	.dc.b	$FF,$0B,$C6,$3D,$BF,$F0,$FB,$6B,$81,$FD,$54,$E9,$19,$9F,$F0,$FF
	.dc.b	$5D,$DF,$FC,$5F,$F0,$BF,$E0,$13,$FC,$05,$F2,$F7,$0F,$F1,$E4,$5B
	.dc.b	$FF,$1E,$4D,$AF,$F1,$3F,$D5,$7F,$8C,$3D,$5F,$F1,$8B,$70,$FF,$17
	.dc.b	$B6,$FF,$C5,$ED,$AF,$A9,$F4,$EE,$FF,$F1,$37,$C8,$7F,$8B,$63,$18
	.dc.b	$7F,$8B,$23,$13,$A8,$F7,$61,$FD,$E2,$B6,$FF,$8B,$2B,$73,$A8,$4E
	.dc.b	$AF,$FC,$9F,$F5,$5A,$9F,$F2,$1F,$D5,$7F,$92,$8F,$F2,$5F,$D5,$5C
	.dc.b	$3F,$C9,$31,$67,$FC,$93,$36,$FF,$CB,$3F,$13,$FE,$2B,$F8,$CF,$F1
	.dc.b	$63,$70,$FF,$33,$36,$FF,$CC,$CD,$AF,$87,$F2,$1F,$E5,$AF,$83,$F8
	.dc.b	$4D,$BF,$56,$88,$C6,$1F,$8A,$3E,$13,$8B,$E7,$5C,$23,$FC,$CA,$74
	.dc.b	$7E,$27,$AA,$EF,$FF,$36,$7D,$63,$E6,$BF,$CB,$D3,$9A,$FF,$7F,$4E
	.dc.b	$5F,$99,$F5,$4F,$AB,$FF,$3A,$7D,$A3,$BB,$FC,$CD,$39,$AF,$AF,$A2
	.dc.b	$B5,$1F,$57,$E8,$BF,$CE,$1B,$7F,$A1,$7E,$D2,$A8,$CD,$A7,$38,$38
	.dc.b	$BF,$D0,$9F,$1A,$EC,$F0,$C6,$1F,$F3,$C7,$D1,$BC,$5F,$F4,$57,$DA
	.dc.b	$3D,$3C,$C6,$1F,$F3,$F7,$D3,$F8,$25,$D3,$8F,$63,$8B,$FE,$83,$FA
	.dc.b	$E7,$D4,$7F,$A5,$7C,$F1,$97,$0E,$FE,$BB,$5E,$F5,$E6,$F8,$B3,$FE
	.dc.b	$B6,$7F,$31,$4E,$1F,$F5,$07,$DA,$3E,$23,$FD,$25,$39,$AF,$9B,$39
	.dc.b	$7F,$40,$7C,$33,$E8,$FD,$A3,$E6,$BF,$D3,$51,$9E,$AF,$C5,$74,$AB
	.dc.b	$53,$FA,$9F,$B0,$F8,$CF,$DA,$F7,$3F,$D3,$D1,$8B,$E3,$1F,$EA,$1E
	.dc.b	$4F,$83,$5B,$9F,$97,$B7,$FA,$E6,$BF,$CD,$F3,$FB,$5F,$B0,$7D,$87
	.dc.b	$FA,$B5,$79,$96,$30,$FF,$AC,$7E,$C1,$EE,$66,$FF,$7D,$18,$7A,$57
	.dc.b	$FA,$DF,$EC,$1E,$EF,$1B,$C3,$6B,$F8,$37,$C5,$7F,$05,$7F,$A3,$E0
	.dc.b	$F6,$BF,$D8,$13,$9D,$BF,$D0,$F3,$9A,$E3,$E9,$E8,$FC,$4F,$B3,$63
	.dc.b	$F8,$97,$0A,$31,$7F,$B9,$11,$99,$E6,$7F,$19,$FB,$0F,$C6,$BF,$AA
	.dc.b	$FA,$B7,$C9,$7F,$B4,$E7,$37,$F6,$A7,$35,$C7,$AB,$47,$FB,$A6,$3F
	.dc.b	$DA,$D5,$C3,$A3,$FB,$33,$6B,$FB,$3B,$6F,$FD,$F1,$5C,$2F,$C1,$C7
	.dc.b	$FE,$F3,$F2,$BF,$D4,$14,$BF,$B3,$87,$F0,$09,$CC,$62,$7E,$03,$71
	.dc.b	$FE,$F2,$7E,$FF,$E5,$3A,$DF,$C1,$F4,$7D,$C7,$BE,$EC,$64,$8F,$43
	.dc.b	$27,$CD,$A7,$AB,$FE,$F9,$6C,$FF,$45,$9F,$F6,$0F,$93,$FE,$C2,$FA
	.dc.b	$F7,$D7,$7F,$82,$2F,$75,$CF,$0E,$47,$A3,$0F,$F0,$86,$1C,$6F,$84
	.dc.b	$E6,$BF,$C2,$22,$74,$FF,$B2,$E3,$FC,$33,$87,$82,$37,$3A,$58,$78
	.dc.b	$A7,$51,$FE,$1C,$9D,$4E,$6F,$C7,$2A,$DC,$E6,$3A,$99,$9F,$3E,$1D
	.dc.b	$E3,$E1,$99,$3C,$CE,$5F,$51,$93,$CB,$87,$AA,$7C,$37,$E0,$8C,$4F
	.dc.b	$82,$31,$5C,$F0,$F4,$D6,$AB,$D8,$7F,$8E,$CF,$19,$63,$0E,$0F,$F1
	.dc.b	$EF,$D7,$FF,$41,$FB,$0E,$0D,$8F,$F2,$23,$5F,$A6,$D7,$8F,$BF,$F3
	.dc.b	$9D,$B5,$DB,$5D,$E7,$DA,$38,$65,$F5,$33,$FE,$D5,$5F,$65,$AF,$C7
	.dc.b	$B5,$70,$9E,$73,$F2,$DD,$D7,$3D,$3F,$CB,$2F,$E0,$2F,$8E,$1F,$85
	.dc.b	$6D,$F1,$AD,$DF,$F9,$80,$E6,$76,$BE,$58,$78,$68,$7E,$82,$B8,$B8
	.dc.b	$AE,$2E,$AF,$BD,$F2,$C3,$DD,$7C,$2B,$91,$E5,$C3,$B3,$83,$F9,$06
	.dc.b	$1F,$12,$BF,$CE,$CD,$D3,$CC,$FA,$6F,$F6,$51,$EB,$F5,$ED,$F9,$5A
	.dc.b	$F5,$BF,$F4,$07,$D8,$7F,$00,$8F,$F2,$13,$5F,$E6,$0D,$8D,$E3,$CA
	.dc.b	$31,$7F,$12,$5C,$9F,$DC,$6F,$F9,$97,$8F,$BE,$0F,$0F,$D7,$FF,$96
	.dc.b	$E7,$FC,$A1,$F5,$DF,$B0,$C7,$FD,$22,$D8,$F9,$F3,$B7,$FA,$13,$71
	.dc.b	$FF,$4C,$4E,$E4,$D9,$D5,$7F,$A7,$1F,$F0,$E4,$FF,$97,$DA,$F4,$F4
	.dc.b	$BE,$2D,$3E,$C3,$CF,$18,$E8,$95,$C6,$32,$FA,$A3,$2F,$33,$FC,$D9
	.dc.b	$F6,$1F,$DD,$9B,$1F,$4A,$31,$1F,$D7,$6D,$F2,$3F,$5B,$DF,$7C,$F5
	.dc.b	$25,$72,$3E,$4E,$FE,$82,$3F,$E7,$B9,$E2,$DF,$B0,$E8,$47,$5E,$5F
	.dc.b	$E1,$3E,$C1,$F1,$94,$93,$FE,$85,$8F,$96,$FE,$8B,$EC,$1F,$17,$F8
	.dc.b	$C6,$C7,$FA,$2D,$AF,$B7,$C0,$FC,$9F,$2B,$ED,$BF,$CD,$93,$FE,$60
	.dc.b	$7E,$0B,$77,$DC,$6A,$3F,$C6,$3F,$3C,$F2,$F9,$B6,$3F,$CD,$2D,$77
	.dc.b	$F4,$DE,$0F,$B2,$F4,$E2,$69,$3D,$EE,$FF,$45,$2F,$C8,$F8,$BA,$39
	.dc.b	$7A,$3C,$FD,$69,$5F,$E7,$8C,$3F,$C5,$2D,$E3,$F0,$75,$C8,$C1,$EF
	.dc.b	$FB,$7F,$6F,$6C,$8C,$1C,$FE,$43,$D0,$47,$B7,$C8,$FA,$FE,$42,$47
	.dc.b	$85,$F7,$7C,$87,$C8,$F9,$59,$3F,$35,$F9,$7E,$5F,$C3,$FA,$0F,$89
	.dc.b	$F0,$72,$BD,$51,$F1,$FE,$DB,$E4,$7C,$14,$9F,$74,$7F,$0A,$FD,$07
	.dc.b	$DF,$1F,$E3,$57,$FC,$F6,$FF,$B2,$DA,$E7,$93,$81,$C7,$92,$46,$E3
	.dc.b	$2F,$18,$CB,$C4,$FF,$6B,$7D,$53,$7F,$EB,$AD,$8F,$EC,$4D,$7F,$C1
	.dc.b	$61,$C3,$FE,$0E,$75,$D2,$F2,$FF,$9E,$CB,$EA,$5F,$55,$FC,$A1,$B1
	.dc.b	$E2,$C3,$1F,$01,$9F,$93,$93,$FE,$EB,$AF,$F3,$23,$67,$2F,$F9,$6E
	.dc.b	$3F,$D8,$4E,$4E,$1F,$4B,$DC,$8F,$3F,$36,$BC,$62,$E7,$E9,$8F,$EA
	.dc.b	$B5,$FC,$01,$D1,$D0,$87,$86,$BF,$A1,$FD,$53,$C7,$FF,$03,$B6,$7F
	.dc.b	$C1,$8D,$7E,$5B,$5F,$F1,$BB,$9B,$F1,$43,$C1,$FE,$68,$7F,$35,$9D
	.dc.b	$6D,$F3,$A3,$FC,$EB,$3F,$D2,$63,$1E,$FF,$F5,$69,$CC,$FF,$A0,$5F
	.dc.b	$E9,$8D,$8D,$11,$E4,$7F,$76,$8D,$DF,$91,$7D,$4F,$52,$4F,$01,$F5
	.dc.b	$5E,$AA,$E1,$3A,$9F,$13,$FD,$AA,$75,$87,$AE,$35,$5C,$2B,$A5,$69
	.dc.b	$FF,$04,$FD,$4B,$AB,$F8,$B4,$7E,$42,$31,$E4,$FF,$8E,$D7,$5B,$DA
	.dc.b	$C5,$F0,$BF,$F0,$9F,$DB,$7F,$23,$D3,$FC,$84,$D8,$FD,$0B,$5F,$F3
	.dc.b	$6B,$5F,$EF,$1F,$4E,$E4,$FF,$39,$6A,$8C,$5F,$CC,$8C,$71,$EA,$73
	.dc.b	$7F,$EE,$46,$CF,$FA,$59,$B5,$DA,$B8,$5F,$6F,$A2,$79,$74,$F0,$FD
	.dc.b	$57,$A7,$1F,$F4,$FC,$EB,$0F,$13,$72,$FF,$79,$B6,$7F,$D4,$B5,$C6
	.dc.b	$BF,$C2,$93,$96,$DF,$77,$D9,$3B,$BF,$F4,$4D,$6D,$FF,$1F,$FD,$4F
	.dc.b	$F5,$EA,$FC,$33,$63,$E0,$7C,$FF,$FA,$65,$46,$71,$FD,$1B,$6B,$F8
	.dc.b	$6B,$67,$FD,$2F,$A2,$F9,$4E,$6D,$83,$E6,$7E,$41,$F2,$7E,$32,$E5
	.dc.b	$F1,$22,$E5,$F1,$26,$DF,$D3,$FA,$9F,$B1,$EC,$FD,$96,$E1,$F7,$74
	.dc.b	$46,$30,$FB,$8D,$BE,$55,$BB,$E4,$D3,$94,$ED,$FC,$6B,$6B,$DE,$DA
	.dc.b	$FC,$73,$6F,$C3,$5C,$EB,$C2,$DA,$F0,$C7,$E3,$EF,$C2,$D9,$F0,$9B
	.dc.b	$3C,$25,$7C,$2D,$9F,$0B,$6B,$C3,$3C,$CF,$09,$67,$C2,$7E,$76,$BC
	.dc.b	$2D,$AF,$0C,$ED,$F0,$B4,$F0,$96,$3C,$2D,$AF,$0C,$73,$8F,$0B,$4F
	.dc.b	$0C,$FE,$E2,$BC,$2D,$3C,$31,$B7,$C2,$D9,$F0,$B4,$F0,$96,$BC,$33
	.dc.b	$CE,$7C,$2D,$9F,$0C,$7E,$62,$BC,$2D,$9F,$09,$B3,$C2,$57,$C2,$D9
	.dc.b	$F0,$B6,$BC,$33,$CC,$F0,$96,$7C,$27,$F1,$5A,$F0,$B6,$BC,$33,$B7
	.dc.b	$C2,$D3,$C2,$58,$F0,$B6,$BC,$31,$CE,$3C,$2D,$3C,$30,$AF,$0B,$6B
	.dc.b	$C3,$3B,$7C,$2D,$AF,$0B,$4F,$09,$67,$C2,$73,$8F,$0B,$63,$C3,$5E
	.dc.b	$02,$57,$82,$74,$F8,$1B,$5E,$06,$D7,$81,$A7,$82,$B9,$CF,$81,$B3
	.dc.b	$E0,$AE,$D1,$87,$B3,$67,$B3,$67,$B3,$6B,$B5,$72,$25,$77,$6D,$77
	.dc.b	$8C,$11,$C1,$23,$C2,$70,$2C,$6D,$B1,$B8,$F2,$C6,$DA,$6C,$FF,$7E
	.dc.b	$FD,$33,$EA,$3F,$9E,$36,$7F,$A8,$FA,$78,$99,$1F,$E0,$99,$D5,$7F
	.dc.b	$54,$AC,$BF,$D9,$7E,$47,$FB,$47,$D4,$7F,$69,$6D,$FF,$31,$6C,$EE
	.dc.b	$3F,$9A,$4F,$F6,$7F,$26,$47,$F6,$2F,$26,$D7,$F6,$29,$B9,$FF,$68
	.dc.b	$9B,$5F,$D8,$A6,$D7,$F5,$E9,$AB,$27,$9D,$FE,$3A,$8F,$F3,$1C,$62
	.dc.b	$FF,$CD,$71,$B9,$D3,$F2,$1A,$E6,$3E,$24,$7F,$81,$FE,$9F,$E5,$BF
	.dc.b	$E0,$1F,$5B,$FC,$05,$EB,$5A,$FE,$FA,$DB,$FE,$FA,$DC,$FF,$C0,$6F
	.dc.b	$F8,$47,$C9,$95,$FD,$67,$CA,$B7,$FE,$13,$9B,$7F,$E1,$39,$B5,$FE
	.dc.b	$20,$9B,$5F,$DC,$A6,$AF,$A8,$67,$EC,$6D,$FC,$26,$CF,$FA,$39,$AF
	.dc.b	$FA,$28,$E5,$3F,$E0,$58,$DE,$1F,$E1,$CF,$7B,$13,$FE,$2B,$F8,$99
	.dc.b	$87,$F7,$3F,$A0,$B8,$FF,$8B,$66,$E1,$FE,$2D,$9B,$5F,$E1,$EF,$63
	.dc.b	$FC,$59,$E8,$7F,$8B,$5B,$1F,$E2,$86,$E1,$FE,$28,$6D,$FF,$8B,$9B
	.dc.b	$3C,$CF,$F1,$7D,$7F,$87,$7E,$9D,$D9,$F7,$9B,$3F,$84,$6E,$1F,$EB
	.dc.b	$06,$BF,$0D,$CE,$1F,$E8,$66,$CF,$F9,$4B,$EA,$3B,$5F,$F9,$17,$DA
	.dc.b	$93,$FE,$40,$F6,$AE,$3F,$E4,$09,$B8,$FF,$91,$A6,$D7,$F9,$73,$E8
	.dc.b	$FF,$CA,$DF,$17,$FE,$57,$6C,$7F,$87,$DB,$87,$F9,$65,$B7,$FE,$59
	.dc.b	$6C,$F3,$3F,$C4,$D5,$FE,$66,$F1,$5F,$41,$4C,$7C,$DD,$AF,$F3,$47
	.dc.b	$D3,$7E,$01,$64,$C9,$2F,$03,$FC,$E2,$EE,$74,$E6,$B9,$FA,$5A,$F8
	.dc.b	$3D,$B6,$3F,$CE,$DF,$42,$EC,$FF,$3D,$AE,$CB,$18,$7F,$CF,$FF,$49
	.dc.b	$F4,$B2,$F2,$37,$43,$FE,$63,$6F,$27,$0E,$49,$97,$FA,$16,$9C,$98
	.dc.b	$C3,$DD,$3F,$03,$07,$84,$F1,$7F,$D1,$ED,$9F,$F4,$73,$FE,$5F,$6C
	.dc.b	$6B,$E3,$37,$F4,$96,$76,$FF,$A6,$3E,$8D,$F3,$97,$D1,$C3,$6B,$FD
	.dc.b	$35,$F4,$7F,$EA,$C6,$FA,$FF,$EA,$08,$C6,$BE,$83,$EB,$3F,$EB,$79
	.dc.b	$FF,$51,$79,$EF,$7B,$FD,$4E,$D8,$CF,$C8,$E2,$73,$ED,$49,$E3,$F2
	.dc.b	$F7,$1E,$C3,$E8,$9D,$FF,$E8,$B5,$F1,$B6,$30,$FF,$A3,$FE,$87,$F0
	.dc.b	$95,$CA,$B8,$E1,$FE,$DB,$9C,$F3,$7F,$BD,$27,$35,$E0,$FA,$27,$A9
	.dc.b	$CF,$C3,$FF,$4D,$46,$31,$E5,$1F,$E9,$DF,$07,$F0,$D6,$CB,$F1,$D5
	.dc.b	$FE,$E9,$9C,$C6,$27,$DB,$1C,$0E,$3F,$42,$FA,$2E,$9E,$07,$FA,$A6
	.dc.b	$31,$87,$38,$C4,$EA,$3F,$DE,$DF,$43,$F7,$9F,$F6,$A7,$CE,$7D,$1C
	.dc.b	$7E,$5C,$E5,$FB,$5B,$A7,$4A,$BF,$67,$2A,$E1,$39,$8F,$DB,$C6,$27
	.dc.b	$F3,$54,$F9,$6F,$F6,$04,$E6,$B7,$D5,$83,$FD,$D7,$5B,$BF,$4C,$70
	.dc.b	$E2,$FF,$77,$FD,$17,$F8,$23,$2F,$06,$1E,$39,$CF,$B9,$BC,$3E,$0C
	.dc.b	$70,$F8,$CC,$57,$C0,$E7,$F1,$9E,$46,$D7,$C8,$8B,$1F,$18,$B8,$78
	.dc.b	$07,$9D,$FE,$DB,$6D,$7E,$E4,$FE,$B9,$2F,$A2,$FE,$37,$F4,$0F,$95
	.dc.b	$F9,$77,$C2,$BF,$3A,$D3,$B6,$1F,$EE,$78,$FE,$47,$7C,$AB,$8C,$E6
	.dc.b	$7B,$0B,$EC,$FF,$0B,$8F,$AD,$18,$8F,$B7,$5C,$DD,$61,$5C,$9C,$EF
	.dc.b	$BC,$66,$3C,$66,$4E,$25,$38,$D7,$7C,$3D,$CF,$86,$38,$B9,$C3,$CB
	.dc.b	$7F,$23,$0F,$15,$F0,$8C,$5F,$58,$C3,$FE,$09,$9D,$D7,$28,$C1,$B7
	.dc.b	$0F,$B0,$FF,$1A,$C7,$F8,$D9,$F5,$98,$3E,$CF,$D0,$BD,$87,$D1,$7F
	.dc.b	$8F,$23,$0D,$8C,$CF,$06,$BF,$E1,$D9,$7B,$5F,$D6,$E3,$AB,$5D,$35
	.dc.b	$FD,$7C,$70,$C7,$C1,$3C,$73,$FB,$59,$B5,$FD,$AC,$61,$B1,$CA,$78
	.dc.b	$36,$BE,$F3,$C7,$0B,$5C,$1F,$EE,$38,$78,$5F,$B8,$DC,$3B,$47,$03
	.dc.b	$F1,$11,$9A,$ED,$1F,$83,$7B,$36,$BB,$29,$E0,$F8,$30,$F3,$BF,$E0
	.dc.b	$2A,$E3,$7E,$08,$F1,$CE,$9F,$D6,$5F,$EB,$35,$FE,$2F,$7E,$0C,$7F
	.dc.b	$83,$B6,$3C,$B8,$73,$9C,$F8,$1C,$70,$F0,$35,$EA,$7E,$F4,$FF,$0A
	.dc.b	$FD,$17,$F0,$27,$01,$C3,$FD,$61,$B3,$FC,$61,$FC,$1C,$70,$73,$7F
	.dc.b	$CA,$3C,$3F,$15,$7F,$33,$7C,$53,$AA,$FF,$12,$FD,$17,$F9,$39,$F0
	.dc.b	$FD,$0F,$EB,$F0,$FE,$6E,$D8,$DF,$67,$AE,$B8,$47,$F6,$C9,$D2,$6C
	.dc.b	$6B,$1F,$E5,$EB,$E0,$79,$12,$BC,$3F,$37,$5E,$9C,$FF,$27,$0F,$EE
	.dc.b	$B1,$F1,$AF,$93,$6F,$55,$23,$07,$CC,$FA,$2E,$2F,$F8,$09,$B8,$FF
	.dc.b	$42,$6E,$7F,$03,$BB,$EC,$79,$EF,$8F,$A4,$C3,$A6,$3E,$5E,$09,$1F
	.dc.b	$29,$D3,$FE,$10,$A7,$D1,$74,$24,$7E,$02,$1E,$67,$E1,$FE,$89,$F0
	.dc.b	$F4,$95,$F8,$68,$F9,$9A,$7F,$6C,$FA,$27,$C3,$FE,$3D,$B5,$F8,$C6
	.dc.b	$BF,$5F,$C1,$FC,$2F,$D1,$3E,$A2,$F0,$CA,$FC,$B7,$1A,$B5,$18,$FA
	.dc.b	$37,$99,$F8,$E6,$C6,$63,$1F,$03,$F9,$2F,$6B,$CF,$3F,$E7,$36,$E1
	.dc.b	$FE,$55,$EC,$D7,$CA,$79,$3E,$6E,$5F,$8B,$34,$7A,$3A,$3A,$7A,$9D
	.dc.b	$7D,$D9,$7E,$8C,$FF,$C5,$ED,$E6,$F1,$76,$C8,$C1,$D7,$E8,$FE,$1F
	.dc.b	$83,$23,$07,$3F,$8C,$F5,$11,$ED,$F1,$9E,$DF,$83,$91,$E1,$7E,$07
	.dc.b	$C1,$7C,$EF,$8F,$93,$EB,$7E,$8F,$95,$F1,$7E,$67,$E3,$7C,$0C,$AF
	.dc.b	$54,$7C,$9F,$A2,$F9,$5F,$47,$27,$DD,$1F,$C3,$7E,$8F,$DF,$1F,$E5
	.dc.b	$E7,$FD,$02,$FF,$0C,$6B,$9E,$4E,$07,$1E,$49,$1B,$8C,$BC,$63,$2F
	.dc.b	$13,$F8,$BF,$D1,$B7,$FE,$3C,$D8,$CC,$7F,$21,$FA,$27,$6E,$A3,$1F
	.dc.b	$47,$2B,$67,$F2,$2A,$FF,$56,$37,$E3,$39,$1C,$FE,$36,$46,$48,$7F
	.dc.b	$9F,$A7,$8B,$7E,$7B,$C4,$47,$B7,$9F,$5D,$1E,$18,$F4,$46,$7C,$2F
	.dc.b	$F2,$3D,$27,$1E,$57,$F9,$72,$3F,$83,$5F,$F3,$AF,$A3,$79,$3F,$D0
	.dc.b	$5B,$19,$8F,$E8,$9E,$17,$FA,$4A,$30,$7F,$42,$7F,$95,$9F,$CB,$27
	.dc.b	$8C,$7F,$A6,$A3,$FD,$48,$EF,$E2,$B9,$9D,$3D,$1B,$19,$21,$C5,$F0
	.dc.b	$FD,$1C,$8F,$03,$FD,$5F,$E8,$FF,$CA,$D9,$FF,$9E,$30,$4E,$EB,$C4
	.dc.b	$EE,$71,$AF,$F9,$B2,$B7,$5D,$0F,$EB,$B7,$D1,$FF,$7A,$FD,$5B,$C2
	.dc.b	$FF,$15,$36,$3F,$CC,$21,$C3,$FE,$64,$6B,$FE,$8C,$1F,$48,$EE,$FE
	.dc.b	$27,$19,$8C,$79,$9D,$17,$7B,$DA,$C5,$F0,$BF,$F7,$57,$CF,$FD,$FD
	.dc.b	$BF,$91,$B6,$3F,$D3,$EF,$F4,$58,$CF,$2F,$FA,$6F,$79,$3B,$6C,$66
	.dc.b	$7A,$F3,$5B,$52,$F8,$57,$FB,$EF,$E8,$9E,$76,$E9,$18,$9F,$E6,$FF
	.dc.b	$25,$FC,$D1,$FE,$73,$19,$C3,$FD,$6A,$DA,$F3,$36,$53,$E2,$AE,$4F
	.dc.b	$0B,$F0,$56,$E3,$CE,$DB,$57,$07,$AD,$7B,$69,$F4,$7F,$D5,$B0,$FE
	.dc.b	$9B,$F5,$7F,$EB,$79,$FF,$7A,$CE,$5F,$F6,$5C,$66,$FF,$D6,$F3,$99
	.dc.b	$FC,$03,$5F,$F6,$6B,$5F,$5B,$63,$E3,$FD,$27,$A8,$F5,$FA,$77,$0F
	.dc.b	$8F,$16,$FE,$3C,$DA,$FB,$7F,$49,$F4,$75,$FE,$EF,$87,$E1,$23,$1E
	.dc.b	$DE,$5B,$19,$3F,$13,$19,$69,$9C,$3F,$17,$19,$6C,$66,$FF,$17,$39
	.dc.b	$6C,$64,$FC,$BC,$65,$A6,$67,$F1,$B7,$E8,$7F,$33,$7E,$66,$CF,$91
	.dc.b	$A7,$88,$B7,$E0,$6C,$76,$69,$DA,$FC,$4D,$8F,$14,$FE,$76,$72,$D8
	.dc.b	$CD,$F9,$63,$E7,$C7,$8D,$B5,$E3,$69,$E3,$2C,$78,$DB,$1E,$13,$CA
	.dc.b	$F7,$6C,$F7,$9F,$DF,$DF,$C3,$AE,$26,$CB,$3B,$AF,$97,$3B,$AE,$27
	.dc.b	$9C,$B1,$E7,$6D,$79,$DA,$79,$CB,$1E,$78,$EF,$7F,$1A,$FC,$11,$FC
	.dc.b	$16,$FC,$15,$E2,$BF,$03,$4F,$01,$67,$C0,$D8,$F0,$34,$F0,$3E,$28
	.dc.b	$F8,$0D,$AF,$08,$AF,$0C,$F8,$E7,$C2,$DA,$F0,$B4,$F0,$96,$3C,$2D
	.dc.b	$8F,$09,$E3,$7C,$2D,$AF,$0D,$78,$1F,$15,$78,$1B,$3E,$06,$D7,$81
	.dc.b	$B3,$E0,$6C,$F8,$27,$C4,$78,$0B,$1E,$09,$ED,$5E,$12,$38,$25,$78
	.dc.b	$63,$2D,$8C,$D7,$8E,$72,$D8,$C9,$E5,$8C,$B4,$CC,$F9,$EB,$D3,$5E
	.dc.b	$76,$9E,$78,$F0,$12,$3B,$36,$7A,$34,$E9,$5E,$06,$C7,$82,$7B,$4E
	.dc.b	$5B,$19,$AF,$04,$7E,$3A,$9F,$46,$F8,$BF,$EB,$1F,$56,$FA,$5F,$E8
	.dc.b	$93,$9C,$7F,$D1,$F3,$97,$FD,$C7,$19,$BF,$EC,$B1,$FD,$5B,$05,$6E
	.dc.b	$72,$7F,$5A,$AE,$13,$FD,$CE,$74,$FF,$74,$F9,$5F,$E7,$58,$7F,$3B
	.dc.b	$6E,$1F,$DF,$9B,$87,$F7,$E6,$D7,$F6,$8F,$33,$FB,$19,$FD,$93,$CC
	.dc.b	$B7,$FD,$C2,$2D,$FF,$70,$9B,$5C,$6B,$47,$F7,$68,$CB,$FD,$CA,$7E
	.dc.b	$4F,$A3,$FD,$FF,$8F,$FD,$6D,$3F,$12,$31,$EE,$FC,$A5,$1F,$E2,$8D
	.dc.b	$51,$8B,$FF,$13,$1D,$27,$51,$98,$F0,$63,$FE,$0C,$F6,$3F,$BF,$FB
	.dc.b	$1F,$E0,$06,$DF,$F8,$31,$B5,$FE,$0C,$6D,$7F,$88,$3C,$BF,$EC,$C7
	.dc.b	$F6,$7F,$32,$DF,$F8,$AA,$2D,$FF,$8A,$A6,$D7,$1A,$D1,$FE,$30,$8C
	.dc.b	$BF,$17,$E9,$5F,$4B,$FE,$14,$F9,$5F,$F0,$74,$62,$7F,$C4,$31,$89
	.dc.b	$91,$FE,$24,$F8,$2B,$8F,$F8,$EB,$E3,$3F,$BE,$1F,$DF,$7E,$3A,$D7
	.dc.b	$F9,$02,$2C,$FF,$90,$26,$E7,$FE,$2A,$F7,$7F,$C4,$BE,$AF,$F8,$99
	.dc.b	$B8,$7F,$8C,$9B,$7F,$E3,$26,$D7,$19,$FF,$1B,$D7,$16,$9F,$9F,$3F
	.dc.b	$09,$F4,$AE,$FF,$C2,$7C,$0F,$F8,$B6,$31,$3F,$E2,$C8,$C4,$C8,$FF
	.dc.b	$2C,$FD,$2D,$CB,$FC,$87,$F0,$1F,$E3,$C3,$FC,$7B,$F0,$16,$BF,$C9
	.dc.b	$31,$67,$FC,$93,$37,$3F,$F2,$EF,$D2,$FF,$88,$FE,$53,$FC,$48,$DC
	.dc.b	$3F,$CC,$ED,$BF,$F3,$3B,$6B,$8C,$FF,$9B,$2B,$8B,$4F,$81,$5F,$E6
	.dc.b	$4F,$A5,$7D,$2F,$F9,$85,$B3,$AB,$43,$E7,$F5,$5F,$E6,$7F,$A5,$FC
	.dc.b	$26,$1F,$C1,$AD,$0F,$0F,$5C,$FB,$78,$7A,$BC,$7E,$06,$E9,$DE,$FF
	.dc.b	$DB,$6F,$87,$E9,$5D,$DF,$E7,$55,$DD,$63,$0F,$F9,$E7,$E9,$7E,$BE
	.dc.b	$BE,$36,$EB,$FE,$89,$7F,$CD,$86,$30,$FE,$40,$B9,$39,$1C,$F9,$2C
	.dc.b	$FC,$3F,$19,$F1,$5F,$E8,$76,$C6,$7C,$FC,$4E,$7E,$43,$5F,$4B,$6B
	.dc.b	$93,$FE,$8F,$FA,$57,$CF,$5F,$5B,$0D,$AF,$A5,$F4,$BF,$EA,$C6,$FB
	.dc.b	$7F,$E9,$78,$C7,$0F,$D1,$7F,$D6,$D3,$F6,$7D,$67,$C2,$FF,$A7,$5B
	.dc.b	$19,$F8,$5C,$4E,$7D,$E9,$3C,$7E,$6E,$E3,$FD,$08,$7D,$2F,$E0,$E7
	.dc.b	$F1,$71,$8F,$1F,$FD,$29,$39,$8E,$0A,$3F,$D3,$1F,$49,$F8,$5A,$E3
	.dc.b	$3B,$E9,$FE,$51,$39,$BE,$3E,$1F,$FA,$7F,$E9,$5E,$C7,$2F,$0F,$0D
	.dc.b	$CB,$C7,$E1,$FD,$DC,$3F,$D6,$38,$FE,$2A,$73,$E5,$78,$BC,$1F,$F5
	.dc.b	$55,$EE,$CD,$FA,$57,$A5,$AB,$FF,$58,$46,$30,$E6,$7C,$27,$17,$F1
	.dc.b	$EF,$A7,$D2,$FD,$EA,$FD,$67,$D2,$FD,$2A,$F9,$93,$97,$F5,$51,$9E
	.dc.b	$0F,$E9,$53,$96,$F5,$7F,$B5,$E7,$FD,$75,$7F,$D7,$6B,$4D,$9F,$F6
	.dc.b	$34,$62,$FF,$DC,$75,$C1,$FF,$6A,$C6,$53,$F5,$EB,$F8,$17,$D2,$FF
	.dc.b	$80,$1E,$DC,$9F,$EE,$78,$C5,$72,$8E,$E6,$7E,$3E,$D7,$E1,$AD,$CD
	.dc.b	$E4,$3C,$B7,$67,$51,$63,$E2,$97,$3F,$ED,$83,$CE,$FF,$71,$36,$BF
	.dc.b	$27,$3F,$EF,$0F,$A5,$FF,$78,$FD,$1B,$E6,$3F,$B9,$E1,$C0,$97,$C2
	.dc.b	$D7,$C2,$B7,$39,$8C,$4F,$F8,$9E,$BF,$C0,$57,$D6,$3F,$C1,$C7,$20
	.dc.b	$E5,$61,$E7,$73,$87,$F8,$2A,$FA,$C6,$AF,$AC,$E6,$36,$78,$70,$DB
	.dc.b	$EF,$32,$71,$72,$F8,$F0,$F1,$61,$E1,$7C,$91,$C5,$CD,$F8,$23,$D7
	.dc.b	$87,$92,$B8,$61,$E3,$BD,$E1,$DA,$B5,$87,$1C,$3F,$C6,$11,$FE,$26
	.dc.b	$69,$4F,$8C,$65,$D7,$D2,$3D,$87,$D2,$E1,$B5,$FE,$30,$9D,$4E,$4E
	.dc.b	$3E,$D7,$CE,$9D,$35,$FA,$B2,$8C,$47,$F9,$16,$31,$A7,$23,$E9,$E9
	.dc.b	$F9,$89,$BA,$FF,$92,$F5,$FB,$97,$C2,$7E,$8B,$63,$F0,$15,$C6,$7E
	.dc.b	$AB,$63,$8E,$1E,$76,$E1,$DA,$A1,$D2,$75,$5E,$76,$C0,$FC,$F6,$1E
	.dc.b	$59,$FF,$30,$1D,$CE,$15,$F8,$EC,$3C,$D1,$A9,$E1,$5C,$E3,$83,$D6
	.dc.b	$76,$E6,$7F,$CB,$BA,$F1,$8F,$E8,$75,$EF,$BF,$F3,$15,$F0,$9F,$29
	.dc.b	$AE,$FE,$2F,$76,$BA,$3F,$A8,$7D,$2F,$EF,$DF,$E9,$ED,$73,$BF,$08
	.dc.b	$FB,$91,$A8,$EB,$6E,$47,$85,$86,$D6,$CF,$EB,$7F,$49,$FE,$43,$7F
	.dc.b	$AE,$FD,$2F,$EA,$35,$FF,$45,$36,$3E,$84,$ED,$FE,$CC,$DC,$3F,$D0
	.dc.b	$53,$B9,$36,$75,$5F,$E9,$55,$F0,$3F,$09,$B5,$E4,$8F,$8B,$4F,$93
	.dc.b	$D4,$63,$A2,$56,$C8,$7F,$79,$AF,$EE,$4F,$B2,$38,$BF,$DE,$BE,$97
	.dc.b	$2F,$F7,$F6,$E5,$FE,$9B,$6E,$5F,$0B,$E8,$BE,$BF,$9E,$F9,$1A,$4C
	.dc.b	$3A,$63,$E8,$DE,$47,$CC,$74,$FF,$83,$69,$F4,$BD,$09,$1F,$E0,$E8
	.dc.b	$79,$9F,$86,$FA,$5F,$93,$3F,$A1,$5F,$0F,$93,$5E,$2F,$C8,$F0,$7F
	.dc.b	$1B,$F4,$AF,$87,$A6,$70,$FF,$30,$7D,$27,$FB,$B5,$BF,$21,$F8,$A8
	.dc.b	$C4,$7F,$97,$D7,$DA,$7E,$3B,$EE,$7F,$CD,$73,$FE,$69,$7F,$CD,$0D
	.dc.b	$F4,$B5,$F0,$BB,$F0,$38,$78,$72,$B8,$47,$7F,$82,$F0,$6D,$27,$CB
	.dc.b	$1F,$9C,$FA,$77,$CC,$D2,$4F,$F9,$E6,$3E,$A3,$79,$38,$11,$D7,$C4
	.dc.b	$7B,$7C,$F9,$1F,$09,$F8,$1E,$E7,$7F,$88,$F0,$7B,$12,$BB,$47,$8F
	.dc.b	$D6,$F2,$7C,$34,$9F,$14,$7C,$9F,$73,$E5,$77,$C9,$F9,$F1,$EB,$F2
	.dc.b	$E4,$7A,$5F,$77,$C5,$7B,$FC,$19,$3E,$D7,$E1,$7C,$57,$F0,$8F,$A4
	.dc.b	$FF,$3D,$4F,$F9,$51,$AF,$F9,$65,$BE,$4F,$FB,$D5,$7C,$3F,$1F,$88
	.dc.b	$95,$C5,$E9,$18,$8E,$51,$6B,$94,$F8,$7D,$FF,$17,$C4,$4B,$F1,$46
	.dc.b	$5F,$24,$65,$F2,$47,$F2,$0F,$A4,$7C,$75,$24,$FF,$9B,$BE,$8D,$F1
	.dc.b	$D8,$8F,$F3,$A4,$3E,$93,$F9,$4B,$63,$FC,$EB,$7D,$1E,$5C,$9A,$E5
	.dc.b	$91,$82,$1D,$1C,$BD,$A3,$A9,$8D,$5E,$8F,$F5,$19,$61,$FE,$B6,$79
	.dc.b	$CE,$9F,$EA,$D1,$F9,$87,$D9,$DD,$E8,$3D,$3E,$85,$8C,$90,$F7,$BF
	.dc.b	$D1,$BE,$91,$E1,$52,$4F,$FA,$53,$E8,$FF,$62,$FF,$A6,$67,$6E,$7D
	.dc.b	$CF,$E9,$F3,$FC,$D6,$73,$3F,$E9,$F8,$FF,$54,$BC,$FA,$F6,$70,$EC
	.dc.b	$B1,$92,$1C,$5C,$BC,$E2,$E9,$F0,$E1,$5F,$D4,$9F,$07,$D2,$79,$08
	.dc.b	$FF,$AB,$BE,$93,$FC,$C3,$BF,$CF,$A5,$6A,$B8,$9B,$9F,$ED,$7F,$46
	.dc.b	$F8,$EF,$F4,$EC,$FF,$95,$84,$67,$E2,$3F,$C6,$CB,$E8,$DE,$5E,$2F
	.dc.b	$8D,$FF,$97,$BE,$A3,$FD,$4F,$8F,$F8,$F5,$B1,$D5,$AE,$F1,$F8,$0F
	.dc.b	$F8,$5F,$57,$D0,$7F,$1A,$BC,$19,$71,$3F,$E6,$88,$C7,$27,$E2,$EA
	.dc.b	$56,$9B,$CB,$9A,$E4,$DB,$FF,$38,$D7,$F5,$79,$EF,$18,$9D,$57,$91
	.dc.b	$B5,$D6,$BC,$27,$09,$D3,$7E,$91,$F4,$9C,$9B,$1B,$F2,$BE,$2D,$7A
	.dc.b	$67,$38,$FB,$4F,$F4,$2B,$88,$D5,$F9,$2B,$8D,$FC,$9A,$D3,$FB,$1F
	.dc.b	$A4,$FE,$B8,$FF,$BD,$1B,$87,$5C,$3E,$7F,$9B,$F0,$E1,$3A,$9F,$CA
	.dc.b	$36,$BA,$7B,$1F,$0A,$F8,$4E,$6F,$CF,$F3,$1E,$F3,$E0,$7B,$77,$0F
	.dc.b	$89,$16,$7C,$59,$7F,$BD,$72,$29,$D4,$7C,$57,$9D,$72,$25,$72,$8C
	.dc.b	$B6,$33,$3C,$5F,$A9,$7D,$1F,$91,$56,$BE,$7C,$DA,$F1,$CD,$BF,$31
	.dc.b	$E5,$BE,$A7,$23,$A9,$8B,$F8,$6D,$AF,$6D,$F3,$BE,$8D,$9D,$5F,$69
	.dc.b	$C5,$FC,$6A,$B7,$EB,$8C,$5F,$0A,$E6,$D3,$9D,$70,$7C,$D3,$C0,$F8
	.dc.b	$C4,$AE,$87,$E4,$E7,$84,$7A,$8E,$11,$A9,$E6,$D8,$F9,$31,$E1,$7A
	.dc.b	$3F,$95,$9E,$73,$F3,$23,$E3,$CE,$8E,$D1,$CC,$F8,$F1,$CD,$C5,$FC
	.dc.b	$58,$C1,$D4,$B3,$C9,$C3,$D6,$7C,$0D,$9E,$4D,$BF,$C4,$CF,$67,$E5
	.dc.b	$C6,$6B,$E5,$9D,$A7,$99,$F1,$DC,$57,$E1,$67,$E3,$C7,$C6,$23,$E5
	.dc.b	$3E,$39,$D4,$F2,$5F,$C6,$9E,$55,$D6,$74,$FC,$29,$D5,$FA,$2F,$AC
	.dc.b	$F9,$1B,$7E,$E8,$F6,$47,$D0,$6D,$F8,$23,$D3,$5C,$E3,$CD,$3C,$DB
	.dc.b	$5C,$EB,$AD,$6E,$B4,$49,$F4,$CF,$D2,$8E,$E7,$A5,$EE,$78,$CE,$67
	.dc.b	$CE,$AF,$7B,$EA,$BF,$7B,$B8,$F1,$DF,$D6,$AE,$B3,$B3,$F6,$1A,$BE
	.dc.b	$93,$F9,$CF,$A5,$FC,$E9,$B3,$FC,$E6,$74,$D9,$D3,$4F,$E8,$B8,$3E
	.dc.b	$81,$EE,$7F,$5A,$8F,$F7,$84,$62,$7F,$B6,$46,$E7,$4D,$8C,$CE,$E3
	.dc.b	$FA,$01,$93,$F9,$2F,$CD,$7F,$35,$8D,$65,$FD,$7B,$D4,$FE,$A7,$F4
	.dc.b	$7F,$D5,$1B,$87,$AD,$B8,$7F,$3F,$CD,$18,$BE,$55,$F0,$6F,$93,$4F
	.dc.b	$EB,$5E,$67,$F5,$93,$FA,$D7,$9B,$63,$24,$C7,$F9,$DC,$65,$F5,$E4
	.dc.b	$FA,$1F,$82,$DA,$F8,$31,$82,$3F,$3A,$35,$3E,$9F,$5B,$E1,$57,$F6
	.dc.b	$BF,$81,$FE,$D9,$F1,$17,$0F,$EE,$0D,$BF,$E9,$AD,$D3,$FB,$73,$67
	.dc.b	$FB,$43,$67,$FB,$7B,$FD,$FB,$CC,$93,$F8,$5F,$36,$D7,$CB,$9B,$7F
	.dc.b	$E1,$09,$B5,$FE,$61,$FA,$16,$7F,$E0,$1F,$97,$FE,$DD,$18,$9F,$F0
	.dc.b	$34,$62,$64,$7F,$82,$7B,$2E,$5F,$67,$E4,$3F,$C2,$27,$F8,$4B,$E4
	.dc.b	$ED,$7E,$66,$2C,$FD,$F9,$B9,$FF,$89,$26,$D7,$F8,$AE,$6D,$7F,$83
	.dc.b	$BE,$87,$FC,$2D,$EB,$7F,$85,$DB,$5F,$E1,$B6,$CF,$F7,$86,$E1,$FE
	.dc.b	$59,$FA,$1F,$EF,$5F,$46,$F8,$CF,$F0,$64,$61,$BE,$04,$C7,$F6,$1E
	.dc.b	$FD,$CB,$FC,$07,$EE,$7F,$89,$CF,$F1,$47,$B9,$6B,$FC,$59,$16,$7F
	.dc.b	$C1,$33,$73,$FF,$15,$4D,$AF,$F1,$24,$DA,$53,$C6,$FC,$6B,$FE,$49
	.dc.b	$5E,$BF,$F9,$01,$CD,$FF,$95,$5C,$BF,$E3,$FF,$A2,$77,$FF,$91,$9B
	.dc.b	$19,$F1,$71,$39,$DE,$4F,$1F,$72,$E7,$FE,$33,$F8,$4F,$C2,$37,$A3
	.dc.b	$FC,$A3,$1F,$B8,$AF,$F2,$73,$AD,$7F,$CC,$0F,$F9,$0A,$BD,$AF,$FB
	.dc.b	$12,$77,$E7,$F2,$F7,$BF,$CB,$7F,$44,$E7,$BD,$18,$6D,$FF,$96,$FC
	.dc.b	$4F,$F1,$A1,$FE,$4C,$CB,$E4,$B8,$F3,$FF,$CC,$AB,$9A,$98,$E6,$B7
	.dc.b	$FE,$65,$FA,$16,$DF,$E6,$C6,$C6,$7D,$EC,$4F,$F9,$C3,$E8,$5F,$21
	.dc.b	$4C,$7D,$15,$BF,$F3,$87,$A0,$F8,$8F,$F3,$BA,$F5,$6C,$61,$CF,$C5
	.dc.b	$EB,$E3,$2D,$F2,$7F,$CC,$5F,$45,$F9,$98,$FE,$7E,$DE,$3F,$E6,$AA
	.dc.b	$77,$E2,$7F,$9A,$7E,$87,$EB,$56,$A7,$1E,$9F,$47,$FC,$DD,$F4,$4F
	.dc.b	$53,$39,$7F,$4C,$8C,$3F,$D3,$B0,$E7,$DD,$FE,$C5,$EE,$79,$1D,$FD
	.dc.b	$0F,$F3,$C4,$62,$F8,$46,$26,$47,$F5,$8F,$A2,$79,$F8,$DB,$FA,$7C
	.dc.b	$62,$7F,$AE,$E1,$CE,$73,$5B,$C7,$FD,$AD,$F4,$3F,$3E,$FF,$D4,$1F
	.dc.b	$38,$F8,$8F,$F4,$54,$E7,$0F,$F4,$74,$E6,$BF,$B4,$CE,$6F,$9F,$37
	.dc.b	$FA,$BA,$9F,$31,$CA,$25,$F0,$75,$18,$C3,$FB,$3E,$1F,$57,$0F,$F6
	.dc.b	$17,$D1,$7F,$66,$C1,$52,$B5,$1F,$EC,$F9,$D4,$E7,$97,$D2,$7F,$B0
	.dc.b	$21,$DD,$8A,$F4,$FC,$A7,$B8,$F7,$DD,$8C,$91,$F2,$99,$3F,$BE,$E2
	.dc.b	$F9,$3F,$F5,$83,$67,$FC,$1B,$3E,$8F,$A2,$F4,$FD,$0F,$F1,$48,$FF
	.dc.b	$73,$5F,$02,$57,$08,$57,$0F,$4F,$E0,$C7,$F2,$0A,$FF,$5F,$3D,$45
	.dc.b	$F5,$75,$7C,$5E,$27,$43,$75,$D3,$0F,$44,$60,$DE,$1C,$5E,$66,$4F
	.dc.b	$1B,$98,$E3,$7E,$19,$F2,$3E,$38,$CD,$78,$A7,$C5,$39,$C1,$87,$79
	.dc.b	$E3,$5E,$59,$CC,$78,$4C,$9F,$2A,$78,$3D,$27,$84,$63,$0F,$35,$72
	.dc.b	$BF,$35,$FC,$33,$FC,$7B,$3F,$E3,$E7,$E6,$18,$3D,$1F,$42,$F5,$DF
	.dc.b	$45,$FE,$44,$8C,$36,$33,$3C,$1A,$FD,$29,$7B,$3F,$D6,$A3,$AB,$5D
	.dc.b	$35,$FC,$0C,$70,$D3,$C1,$3C,$71,$FC,$06,$8D,$7F,$0F,$18,$6C,$72
	.dc.b	$9E,$0D,$AF,$BC,$F1,$C2,$D7,$07,$F6,$38,$78,$5F,$C1,$37,$0E,$D1
	.dc.b	$C0,$FC,$44,$66,$BB,$47,$E2,$5E,$CD,$AE,$CA,$78,$3E,$0C,$3C,$EF
	.dc.b	$F0,$2A,$E3,$7E,$08,$F1,$CE,$9F,$F0,$0D,$FE,$77,$5F,$D4,$DF,$83
	.dc.b	$1F,$C5,$D6,$8F,$F4,$1E,$3E,$07,$07,$F9,$EA,$5F,$17,$CE,$3C,$4D
	.dc.b	$77,$5D,$3E,$8B,$F8,$24,$7F,$A3,$9A,$FF,$A1,$1B,$1B,$CB,$94,$63
	.dc.b	$1F,$E3,$F2,$E6,$7C,$97,$86,$30,$F0,$E1,$F1,$7D,$17,$F4,$CC,$BF
	.dc.b	$D3,$BF,$43,$FB,$1C,$7F,$84,$B6,$3F,$95,$CE,$DF,$F4,$BB,$70,$F8
	.dc.b	$B3,$B9,$36,$75,$5F,$C5,$17,$C1,$78,$89,$5E,$2F,$A2,$AF,$FA,$92
	.dc.b	$FF,$C9,$8F,$9B,$A3,$38,$FD,$4C,$39,$CE,$48,$79,$B0,$FA,$1F,$45
	.dc.b	$CA,$3F,$D5,$ED,$AF,$BA,$DC,$BE,$0F,$5F,$D7,$F3,$5F,$1F,$49,$87
	.dc.b	$FB,$12,$F8,$3F,$D4,$34,$FA,$2F,$CC,$DF,$F0,$11,$88,$FF,$5F,$DF
	.dc.b	$47,$FD,$9A,$BE,$73,$C0,$48,$CF,$9D,$F8,$9F,$A2,$F9,$71,$FD,$79
	.dc.b	$7C,$46,$4D,$78,$FE,$0F,$03,$F2,$1F,$44,$F8,$7A,$67,$0F,$C4,$F0
	.dc.b	$FE,$2B,$E8,$BF,$DB,$B1,$83,$FC,$D2,$E1,$FF,$35,$C6,$FE,$3F,$F1
	.dc.b	$D2,$FA,$2F,$C8,$7D,$13,$D5,$BE,$9C,$C7,$A7,$42,$B5,$18,$7F,$52
	.dc.b	$DF,$13,$99,$D3,$C2,$EF,$EC,$F8,$38,$25,$74,$D3,$FC,$9E,$F6,$FA
	.dc.b	$27,$CB,$7E,$8D,$B1,$98,$C7,$66,$79,$E4,$F1,$78,$7C,$5F,$BB,$CF
	.dc.b	$93,$DD,$F8,$3E,$D7,$5F,$86,$ED,$ED,$4A,$F1,$47,$87,$D7,$F1,$7C
	.dc.b	$7C,$9F,$34,$7C,$AF,$57,$E5,$F8,$72,$7D,$91,$E9,$F2,$A4,$7B,$5F
	.dc.b	$67,$C6,$FB,$7C,$39,$3F,$05,$F8,$1F,$45,$FC,$47,$E8,$BF,$CF,$13
	.dc.b	$F7,$43,$87,$EE,$9F,$E2,$A7,$1D,$7F,$C8,$97,$C1,$F3,$F8,$49,$5C
	.dc.b	$E7,$7E,$17,$F8,$49,$DC,$66,$76,$F6,$3C,$2F,$96,$BC,$DF,$01,$E7
	.dc.b	$F8,$59,$5D,$63,$2F,$78,$CB,$CE,$FF,$CD,$CF,$B7,$E8,$9F,$13,$FC
	.dc.b	$C5,$B1,$98,$FE,$69,$F4,$4F,$8C,$D4,$7F,$7D,$AD,$D7,$F8,$4D,$AF
	.dc.b	$F8,$99,$AF,$F8,$C2,$9C,$9D,$39,$64,$FC,$F9,$CB,$67,$2F,$C3,$84
	.dc.b	$F9,$26,$4F,$18,$DE,$FF,$D5,$2B,$9C,$F1,$ED,$F5,$91,$F0,$76,$D7
	.dc.b	$47,$8E,$3E,$0C,$7F,$55,$FA,$2F,$F3,$0C,$FE,$15,$AF,$D7,$6D,$FF
	.dc.b	$64,$FA,$1F,$F5,$24,$EF,$C2,$FF,$12,$57,$08,$FF,$25,$B5,$FF,$2A
	.dc.b	$B5,$FF,$3B,$CF,$FA,$2D,$FC,$A3,$5F,$C3,$36,$BC,$1D,$5F,$DA,$E3
	.dc.b	$1C,$BF,$E6,$06,$D7,$4E,$AD,$56,$D9,$1D,$9C,$BE,$0F,$A1,$E8,$47
	.dc.b	$F4,$DF,$45,$FE,$02,$C7,$E0,$4E,$67,$CA,$FF,$7D,$9D,$63,$FE,$72
	.dc.b	$FA,$57,$B5,$FE,$47,$6D,$FF,$97,$43,$87,$FC,$C0,$D7,$FC,$E7,$F4
	.dc.b	$6F,$A2,$7C,$4F,$18,$C3,$7E,$8F,$F9,$4D,$70,$AD,$BF,$E7,$CD,$1F
	.dc.b	$3D,$FC,$93,$3F,$E6,$6D,$9F,$E0,$33,$AE,$3F,$F3,$5B,$27,$4D,$E4
	.dc.b	$CD,$FF,$84,$A7,$2D,$AC,$DA,$FC,$37,$86,$D7,$8A,$31,$7F,$DC,$23
	.dc.b	$07,$F4,$8F,$A1,$7C,$A7,$16,$C7,$F4,$CF,$23,$E2,$BB,$9D,$11,$D6
	.dc.b	$B7,$2F,$E2,$5F,$45,$FD,$96,$7F,$AA,$FC,$EF,$FB,$51,$FE,$4D,$19
	.dc.b	$9F,$8B,$39,$78,$FB,$1E,$B8,$C6,$BF,$E4,$8A,$E0,$FC,$A8,$CD,$FF
	.dc.b	$8E,$33,$7D,$17,$AC,$F6,$7B,$76,$7E,$04,$5C,$BE,$AD,$7F,$BD,$F0
	.dc.b	$FA,$AF,$2A,$D9,$27,$77,$C6,$B7,$76,$BC,$53,$6B,$EB,$5F,$37,$BD
	.dc.b	$79,$2B,$A5,$F2,$69,$C8,$B3,$CA,$7A,$4A,$BD,$15,$F0,$67,$31,$DE
	.dc.b	$7D,$77,$89,$CB,$6B,$AC,$F3,$8C,$5F,$9A,$7F,$0D,$18,$9F,$34,$FC
	.dc.b	$29,$E3,$7D,$DF,$11,$C0,$CD,$F8,$1A,$70,$BE,$8F,$0A,$E8,$78,$49
	.dc.b	$7E,$03,$D7,$1C,$A3,$2F,$C7,$69,$F1,$DE,$C6,$E3,$93,$B6,$C6,$CE
	.dc.b	$2F,$C3,$8E,$07,$95,$F0,$C6,$AF,$E0,$3C,$63,$27,$43,$EB,$5F,$C1
	.dc.b	$8E,$0D,$8D,$4F,$27,$75,$F2,$6B,$6F,$A9,$A7,$86,$FE,$4B,$E1,$8F
	.dc.b	$61,$E7,$23,$93,$F2,$71,$C2,$7B,$C7,$D0,$7C,$77,$E5,$AE,$31,$A3
	.dc.b	$CF,$5C,$6F,$D0,$7E,$4A,$FB,$4E,$EB,$8C,$E6,$31,$3A,$8E,$35,$CE
	.dc.b	$FF,$49,$3E,$EA,$EF,$7F,$4A,$38,$56,$E3,$E3,$4F,$A1,$F8,$CD,$9F
	.dc.b	$8D,$1C,$63,$C2,$7D,$D8,$ED,$1E,$69,$E9,$1F,$0E,$3D,$35,$F0,$E3
	.dc.b	$D0,$DA,$E7,$7F,$0A,$3D,$31,$E1,$3D,$33,$D8,$F4,$D7,$EC,$36,$7C
	.dc.b	$EF,$F3,$6F,$47,$F9,$BB,$67,$FA,$5B,$6B,$FA,$5D,$70,$9F,$E9,$F5
	.dc.b	$FD,$12,$5F,$42,$F7,$3F,$A2,$7D,$2B,$E7,$F5,$39,$C3,$FD,$13,$39
	.dc.b	$7F,$CB,$51,$9B,$FE,$B3,$39,$7F,$C9,$D1,$9C,$FF,$AD,$CE,$5F,$EB
	.dc.b	$9E,$9F,$F5,$0C,$3F,$A8,$B7,$0F,$EC,$AD,$C3,$D6,$DA,$E3,$3F,$DA
	.dc.b	$AB,$8B,$4F,$EB,$1E,$67,$F5,$73,$FA,$C7,$99,$63,$24,$C3,$F9,$EC
	.dc.b	$65,$F4,$FD,$17,$F4,$DF,$A5,$F8,$6D,$AF,$82,$9F,$F0,$66,$88,$C5
	.dc.b	$08,$58,$D4,$F0,$9D,$CE,$B2,$FE,$71,$F0,$5F,$DB,$FE,$23,$FB,$83
	.dc.b	$6F,$FB,$A3,$6B,$FA,$AB,$6B,$FA,$AB,$67,$FB,$03,$67,$FC,$09,$C3
	.dc.b	$FE,$68,$9F,$F0,$9F,$99,$FE,$02,$F3,$3F,$C0,$6C,$8C,$9F,$E1,$88
	.dc.b	$C5,$7F,$A2,$FE,$8D,$EC,$7F,$81,$7E,$5F,$F1,$D1,$8B,$FF,$05,$46
	.dc.b	$27,$DF,$5F,$12,$B4,$FC,$63,$E4,$38,$70,$D7,$FC,$41,$F1,$FF,$E1
	.dc.b	$83,$FC,$33,$F2,$77,$0F,$F1,$6C,$59,$FD,$1C,$DD,$7F,$C5,$93,$6B
	.dc.b	$FC,$55,$36,$BF,$B9,$FC,$FF,$F8,$7F,$D4,$FF,$10,$36,$BF,$C4,$8D
	.dc.b	$9F,$F0,$2B,$70,$FB,$7F,$45,$FC,$6F,$C0,$7C,$87,$EC,$63,$0D,$F0
	.dc.b	$7E,$1E,$1F,$87,$94,$62,$11,$8C,$7F,$CB,$4D,$8F,$F2,$77,$82,$F6
	.dc.b	$BF,$C6,$C7,$F8,$DF,$DA,$B7,$FE,$3B,$8B,$5F,$E1,$C9,$B9,$FF,$87
	.dc.b	$26,$D7,$F8,$9E,$6D,$7F,$92,$FE,$03,$F3,$2F,$F9,$61,$7C,$9F,$F9
	.dc.b	$35,$CE,$1F,$E5,$A7,$37,$FA,$38,$FF,$2E,$FD,$1F,$F0,$38,$C3,$6F
	.dc.b	$FC,$B0,$FF,$19,$7F,$CC,$AE,$64,$FC,$49,$E8,$72,$CF,$FC,$BD,$0F
	.dc.b	$A3,$B1,$FE,$61,$F8,$6F,$C7,$DF,$F9,$B9,$B7,$E5,$5E,$1E,$F9,$F8
	.dc.b	$79,$7C,$7D,$EB,$7E,$0D,$3F,$CD,$D8,$7F,$B6,$E3,$18,$BE,$8D,$CD
	.dc.b	$FE,$71,$7F,$DC,$B3,$A7,$2D,$E7,$FF,$3B,$78,$9F,$D8,$CF,$F3,$7E
	.dc.b	$3F,$E8,$67,$1E,$76,$B9,$65,$F8,$25,$CB,$8B,$FF,$46,$FD,$1B,$5F
	.dc.b	$F4,$13,$6F,$CE,$BD,$EC,$CF,$FA,$1F,$E8,$DF,$25,$4C,$7D,$1D,$C3
	.dc.b	$DF,$E7,$BE,$27,$FD,$1C,$BD,$5B,$18,$73,$F2,$1A,$F9,$0B,$7C,$9F
	.dc.b	$EA,$DF,$47,$FA,$88,$F9,$DB,$A3,$0F,$FA,$76,$3F,$96,$C2,$78,$F8
	.dc.b	$7F,$D7,$FE,$8B,$ED,$D7,$2A,$D3,$D2,$33,$7C,$CF,$E7,$13,$A7,$03
	.dc.b	$E8,$DE,$9F,$5E,$FF,$ED,$91,$8C,$3A,$47,$F6,$FE,$EF,$F1,$05,$B1
	.dc.b	$FE,$60,$F4,$8C,$C6,$3B,$7F,$A1,$C7,$03,$8F,$D1,$BC,$CE,$9E,$47
	.dc.b	$F7,$B8,$C6,$5F,$1A,$B7,$5A,$8F,$93,$F4,$7F,$55,$FC,$AF,$CE,$BE
	.dc.b	$1F,$FC,$09,$39,$D3,$FD,$A7,$39,$AE,$3C,$BA,$3F,$D9,$51,$FE,$AC
	.dc.b	$9F,$EB,$D5,$C7,$4F,$EB,$6D,$BF,$EC,$2D,$9F,$F6,$A5,$71,$51,$96
	.dc.b	$F1,$7F,$B5,$FE,$8F,$FC,$15,$7C,$A3,$1E,$DF,$A1,$91,$FB,$09,$FC
	.dc.b	$D3,$F0,$23,$DB,$F2,$D2,$3E,$1D,$D3,$EE,$E0,$9C,$65,$EA,$9B,$5F
	.dc.b	$DD,$7E,$4F,$FC,$55,$27,$FC,$59,$3F,$DF,$FE,$8F,$FC,$01,$F4,$5F
	.dc.b	$C9,$23,$F9,$04,$F0,$25,$FE,$72,$38,$09,$E0,$FF,$BE,$67,$AA,$AE
	.dc.b	$BE,$87,$8E,$EC,$FF,$69,$9A,$F1,$8A,$F6,$BE,$B8,$7A,$67,$33,$D4
	.dc.b	$F1,$5F,$56,$9D,$4A,$75,$29,$D5,$E1,$5E,$A8,$C1,$F0,$A7,$BD,$F9
	.dc.b	$AF,$C5,$3C,$8F,$15,$79,$B0,$FF,$15,$BC,$E3,$38,$76,$3F,$C1,$0F
	.dc.b	$A4,$A5,$33,$78,$7E,$79,$FE,$0F,$FA,$27,$A8,$FA,$3E,$31,$86,$C6
	.dc.b	$67,$BC,$E5,$BE,$B3,$3E,$2D,$8C,$C6,$08,$FC,$AC,$7F,$20,$FD,$86
	.dc.b	$BF,$3A,$95,$CF,$2F,$C2,$C6,$2F,$86,$BF,$E4,$E9,$FC,$4E,$1C,$9D
	.dc.b	$DF,$18,$FF,$29,$DF,$1C,$3C,$57,$C6,$7F,$CA,$E7,$13,$75,$C6,$7F
	.dc.b	$CB,$95,$C6,$FB,$DF,$19,$FF,$30,$5F,$17,$77,$C5,$5F,$19,$EF,$5C
	.dc.b	$67,$FC,$16,$D9,$D4,$74,$8F,$F3,$36,$BC,$23,$FA,$2D,$7C,$8C,$7F
	.dc.b	$CD,$77,$C2,$79,$9A,$F9,$AE,$2F,$99,$AE,$8F,$E7,$FF,$47,$FB,$D7
	.dc.b	$FA,$8B,$5C,$EF,$C2,$3E,$AC,$6A,$3A,$DB,$91,$F2,$38,$6D,$6C,$FE
	.dc.b	$BB,$F4,$7F,$E4,$07,$FA,$FF,$D0,$FE,$DB,$5F,$F4,$4B,$63,$FD,$19
	.dc.b	$3B,$7F,$AA,$B7,$0F,$F4,$5C,$EE,$4D,$9D,$57,$FA,$51,$7C,$0F,$C2
	.dc.b	$6D,$79,$23,$E2,$D3,$E4,$75,$18,$E7,$95,$B2,$1F,$DA,$2B,$FC,$C2
	.dc.b	$FB,$23,$8B,$FD,$F3,$E8,$B2,$FF,$6E,$6E,$5F,$EA,$46,$E5,$F0,$BE
	.dc.b	$8F,$EB,$F9,$8F,$8F,$A4,$C3,$A6,$3E,$8D,$E4,$7C,$C7,$4F,$F7,$CA
	.dc.b	$7C,$E7,$42,$47,$F8,$42,$1E,$6F,$E1,$BE,$8F,$E4,$CF,$FA,$F5,$7C
	.dc.b	$3E,$4D,$78,$FF,$23,$D3,$FC,$6F,$D1,$BE,$46,$99,$C3,$FC,$3B,$F8
	.dc.b	$BF,$FB,$25,$FF,$0F,$F4,$FF,$9E,$DB,$ED,$7E,$5A,$75,$1F,$E1,$4F
	.dc.b	$B8,$FF,$7C,$3C,$3D,$BF,$CD,$D9,$FC,$EC,$EA,$7F,$C3,$7F,$09,$FE
	.dc.b	$37,$7F,$CA,$FF,$48,$E6,$FC,$CB,$63,$FC,$63,$3B,$7F,$C4,$7F,$42
	.dc.b	$F7,$FF,$53,$F6,$FF,$A5,$EC,$FF,$6D,$B7,$AF,$FD,$C3,$5B,$BF,$F1
	.dc.b	$DB,$FE,$56,$7F,$C7,$B2,$F9,$4E,$67,$ED,$27,$53,$E8,$F9,$2F,$2F
	.dc.b	$64,$8F,$2C,$7E,$EB,$ED,$9F,$3F,$FB,$CF,$B5,$F8,$FA,$7F,$BB,$9B
	.dc.b	$A7,$FB,$D2,$B7,$87,$F9,$3B,$ED,$9F,$37,$C7,$D7,$93,$98,$FE,$0F
	.dc.b	$C2,$FA,$EF,$8F,$E4,$4A,$F8,$F1,$FC,$33,$0F,$F4,$94,$7F,$96,$BE
	.dc.b	$D9,$F3,$FF,$C4,$1B,$1E,$1F,$91,$FE,$27,$F6,$DF,$E3,$F8,$E1,$DF
	.dc.b	$B3,$3E,$EF,$1F,$83,$FA,$FE,$A4,$AE,$D1,$FC,$73,$2F,$0C,$7F,$96
	.dc.b	$9B,$C8,$E3,$4C,$9C,$C6,$21,$FA,$0B,$EB,$7A,$6B,$FC,$A3,$EA,$B9
	.dc.b	$57,$F9,$D9,$AF,$78,$DB,$9B,$E0,$61,$FF,$36,$CF,$E7,$DB,$E2,$F5
	.dc.b	$3B,$7C,$7D,$8E,$D1,$5E,$31,$E3,$D3,$C9,$E2,$49,$F2,$46,$5F,$34
	.dc.b	$65,$F0,$C7,$F4,$2F,$AA,$7C,$8D,$24,$FF,$A4,$BE,$A9,$F2,$B8,$8F
	.dc.b	$F4,$2C,$7C,$87,$E5,$B5,$57,$47,$FD,$36,$D7,$8E,$BF,$D4,$A7,$84
	.dc.b	$E7,$87,$E5,$C3,$E7,$75,$F3,$B3,$0F,$24,$71,$30,$F9,$23,$0F,$A7
	.dc.b	$B7,$B1,$1F,$67,$AD,$5F,$5C,$5C,$3C,$33,$FD,$9B,$EA,$BF,$D8,$D5
	.dc.b	$FE,$B4,$6B,$FE,$A6,$6C,$FE,$0B,$C3,$FF,$5A,$4E,$FD,$0F,$ED,$F1
	.dc.b	$FD,$54,$FE,$AD,$5C,$39,$7F,$20,$DA,$FC,$AC,$EF,$A7,$C5,$7C,$BC
	.dc.b	$29,$5F,$EC,$6A,$FF,$4E,$37,$D7,$CB,$FE,$BA,$8C,$B6,$7C,$73,$E4
	.dc.b	$FA,$A9,$3C,$DF,$F0,$2F,$D5,$7E,$73,$4F,$64,$6A,$7A,$3A,$D3,$FC
	.dc.b	$17,$F5,$2F,$A9,$7C,$77,$F9,$52,$30,$DF,$01,$F0,$79,$AD,$E3,$FE
	.dc.b	$5C,$FA,$FE,$38,$FF,$97,$1B,$19,$C3,$F5,$EF,$ED,$E3,$2F,$EE,$47
	.dc.b	$DE,$3A,$B7,$1F,$C0,$23,$17,$FE,$68,$6C,$7F,$9B,$1A,$FF,$03,$D5
	.dc.b	$F4,$BF,$C9,$AB,$F8,$4C,$7F,$02,$33,$CF,$63,$33,$FD,$DD,$D9,$88
	.dc.b	$D5,$7F,$9A,$67,$A3,$FE,$F3,$AE,$75,$C2,$35,$CF,$C2,$73,$1B,$FA
	.dc.b	$3F,$EA,$95,$B9,$46,$27,$A7,$CD,$7C,$A9,$F4,$4E,$5B,$EB,$75,$8C
	.dc.b	$5F,$BA,$B8,$3F,$02,$33,$7F,$CA,$BE,$AF,$FB,$F4,$F9,$E1,$18,$7D
	.dc.b	$99,$3E,$93,$E0,$C6,$BC,$BB,$5B,$CF,$F9,$5B,$70,$F9,$77,$AA,$91
	.dc.b	$DE,$75,$7F,$2E,$B2,$FF,$A5,$67,$F9,$E4,$EA,$3C,$11,$F1,$F0,$F2
	.dc.b	$3C,$EF,$8E,$1E,$27,$E8,$4E,$AB,$A1,$F0,$63,$75,$D9,$DE,$18,$9D
	.dc.b	$4F,$53,$C1,$19,$8F,$19,$D0,$B7,$F8,$DB,$F0,$3E,$3A,$E5,$3B,$C3
	.dc.b	$E0,$CF,$E5,$AF,$A4,$7B,$CC,$DF,$2B,$F3,$1C,$AB,$A5,$FD,$6B,$E3
	.dc.b	$87,$A6,$7A,$56,$CE,$35,$E1,$9F,$7C,$66,$3D,$E6,$5F,$79,$97,$DA
	.dc.b	$F2,$9F,$75,$78,$23,$94,$78,$27,$91,$1F,$39,$C2,$3A,$61,$EF,$BF
	.dc.b	$1D,$6E,$7B,$11,$FC,$33,$C2,$3D,$D8,$7C,$59,$F4,$DF,$6A,$E0,$F8
	.dc.b	$CC,$46,$A3,$33,$C6,$3D,$A6,$49,$3C,$67,$C5,$1A,$37,$19,$78,$19
	.dc.b	$BF,$B3,$87,$C5,$8F,$61,$EA,$29,$BC,$3E,$0D,$79,$A3,$17,$F4,$9D
	.dc.b	$DF,$C7,$8C,$4E,$E3,$13,$F8,$39,$F8,$C7,$D4,$C3,$DD,$3A,$AF,$86
	.dc.b	$7A,$6B,$B9,$E6,$71,$1B,$AF,$8F,$87,$AE,$78,$DF,$4C,$31,$5F,$C0
	.dc.b	$0F,$0E,$1C,$E7,$51,$8A,$FC,$14,$FB,$2B,$ED,$DF,$28,$FA,$37,$C1
	.dc.b	$D5,$FC,$03,$E9,$E1,$E5,$8F,$99,$87,$C3,$C3,$9D,$77,$3D,$25,$9E
	.dc.b	$B8,$77,$7E,$DE,$1D,$EF,$95,$77,$9F,$BB,$87,$79,$F8,$B8,$79,$DB
	.dc.b	$5D,$5B,$5F,$CF,$FE,$AF,$F9,$B1,$FC,$DF,$E0,$2D,$7F,$41,$CD,$38
	.dc.b	$AF,$E8,$D5,$6F,$F8,$67,$B3,$FC,$73,$E0,$FF,$8E,$B7,$1F,$E4,$8D
	.dc.b	$BF,$E7,$2D,$AF,$F2,$FF,$D4,$BE,$5B,$F9,$DB,$6F,$C5,$F6,$18,$99
	.dc.b	$1F,$D1,$67,$58,$7F,$59,$AC,$BF,$C8,$63,$37,$DA,$FF,$AA,$56,$A3
	.dc.b	$F9,$27,$CD,$FF,$34,$FA,$BF,$E6,$AD,$C7,$F9,$E3,$71,$FE,$A2,$DA
	.dc.b	$FE,$C3,$E6,$FF,$5F,$3F,$B0,$79,$D6,$FE,$54,$5B,$FE,$DD,$36,$BE
	.dc.b	$0F,$D5,$FF,$83,$25,$A2,$30,$DC,$3F,$C2,$93,$FE,$90,$8C,$38,$6D
	.dc.b	$74,$9F,$EA,$9F,$51,$F3,$2E,$47,$C9,$7F,$A9,$C6,$2F,$FC,$05,$F1
	.dc.b	$3F,$DE,$7E,$3F,$FB,$D3,$6F,$F0,$8D,$9D,$C7,$E1,$E7,$FA,$EF,$9D
	.dc.b	$23,$FB,$27,$9D,$6B,$FB,$CC,$DC,$BF,$AA,$4D,$AF,$F5,$47,$D5,$B2
	.dc.b	$FF,$05,$7B,$3F,$DF,$A3,$13,$F9,$58,$C4,$C8,$FC,$C7,$4D,$CB,$FB
	.dc.b	$15,$F0,$6D,$7F,$85,$67,$58,$FF,$65,$F9,$6F,$EE,$A7,$F7,$6F,$98
	.dc.b	$B8,$FF,$7D,$8B,$87,$F8,$4A,$6D,$7F,$88,$BE,$A7,$FC,$41,$EB,$FF
	.dc.b	$88,$5B,$87,$EA,$DB,$7F,$E2,$F6,$D7,$FB,$03,$EA,$9D,$9F,$E3,$5F
	.dc.b	$AB,$7C,$47,$F8,$7A,$73,$8F,$F8,$D6,$72,$FF,$89,$E3,$38,$FF,$80
	.dc.b	$5B,$87,$F8,$07,$EB,$B9,$63,$2F,$83,$FE,$14,$8C,$5F,$F8,$27,$CE
	.dc.b	$7D,$57,$F8,$B9,$B7,$FE,$45,$F1,$BF,$DF,$4F,$F8,$DE,$33,$5C,$7A
	.dc.b	$78,$1F,$E4,$FF,$87,$78,$78,$BF,$F1,$F7,$DD,$E1,$B1,$BF,$96,$FF
	.dc.b	$1E,$CE,$70,$FF,$0E,$7D,$B3,$AF,$FC,$82,$DC,$3F,$CB,$1A,$BC,$3C
	.dc.b	$D7,$F9,$1B,$E8,$7F,$C9,$F8,$1F,$F2,$8F,$17,$BB,$8F,$5E,$6F,$F9
	.dc.b	$33,$EF,$FF,$84,$E7,$FE,$FB,$6E,$0B,$7D,$7F,$F9,$0E,$73,$2F,$B7
	.dc.b	$7B,$5F,$E5,$66,$E1,$FE,$6D,$FC,$4B,$FC,$C7,$43,$85,$6F,$DF,$95
	.dc.b	$FE,$73,$F2,$5D,$D8,$BF,$F1,$F7,$9C,$9F,$F3,$1B,$6B,$FC,$F5,$F6
	.dc.b	$3E,$48,$C5,$7F,$91,$65,$1F,$E6,$A6,$9F,$E8,$1F,$B1,$7C,$4E,$67
	.dc.b	$5E,$D4,$AF,$F4,$37,$5F,$E0,$38,$91,$EB,$C6,$D7,$F9,$DB,$F0,$07
	.dc.b	$8D,$4F,$F4,$96,$1F,$C6,$2F,$84,$E9,$FF,$41,$7D,$7F,$E7,$B0,$FF
	.dc.b	$2E,$37,$1E,$F1,$F2,$61,$F6,$1F,$E8,$47,$55,$E3,$FA,$F7,$A9,$9A
	.dc.b	$FF,$44,$C6,$27,$BC,$7F,$A3,$3E,$C1,$EA,$EA,$F0,$DA,$FF,$4B,$7C
	.dc.b	$0F,$EB,$9E,$DE,$9F,$FA,$52,$31,$9F,$DD,$C9,$39,$F2,$E5,$FF,$A7
	.dc.b	$FE,$C1,$EA,$63,$0F,$F3,$D4,$62,$7B,$D7,$FA,$82,$31,$3F,$E7,$FC
	.dc.b	$7B,$AF,$AE,$7D,$87,$FA,$FF,$EA,$9F,$17,$FE,$A8,$9C,$E5,$FE,$BF
	.dc.b	$9C,$BF,$EB,$18,$CD,$FF,$A2,$DB,$C7,$FE,$D7,$A7,$D0,$F3,$C6,$61
	.dc.b	$C5,$F0,$46,$30,$E5,$EA,$7F,$B4,$7E,$C3,$F4,$F9,$7D,$69,$FD,$64
	.dc.b	$62,$7C,$C6,$FE,$3A,$E3,$E5,$FA,$EF,$51,$EB,$BB,$19,$23,$E9,$32
	.dc.b	$7F,$A9,$74,$7D,$2F,$FB,$29,$B3,$FE,$D1,$9F,$DA,$79,$9F,$B5,$FA
	.dc.b	$F7,$C9,$7F,$BE,$F0,$E0,$4B,$E1,$7F,$C3,$45,$70,$7F,$DF,$F3,$FE
	.dc.b	$D8,$7A,$9A,$AE,$95,$DD,$FE,$FF,$76,$78,$CD,$7F,$5D,$17,$0F,$4C
	.dc.b	$73,$3F,$64,$78,$F0,$F1,$3F,$C6,$70,$FD,$4C,$62,$3C,$6E,$A7,$9B
	.dc.b	$98,$E5,$87,$99,$EC,$78,$8B,$3C,$23,$D3,$18,$AE,$F1,$8A,$ED,$18
	.dc.b	$9F,$3D,$FA,$E3,$11,$D4,$FF,$10,$9F,$34,$B1,$87,$E5,$9F,$03,$EB
	.dc.b	$FF,$87,$FD,$87,$06,$C7,$F8,$A9,$AE,$FC,$14,$EF,$DE,$FE,$7D,$3B
	.dc.b	$6B,$D5,$AF,$F1,$FA,$46,$23,$F9,$E4,$63,$3F,$B1,$9B,$5F,$E4,$DA
	.dc.b	$7D,$48,$C5,$FD,$DB,$E0,$FC,$E6,$CE,$EB,$8D,$7F,$2F,$38,$D6,$F0
	.dc.b	$F4,$D7,$E0,$2F,$8D,$6E,$F8,$DF,$F5,$18,$E3,$5B,$BE,$31,$ED,$6C
	.dc.b	$EE,$B8,$BF,$A4,$6D,$F6,$8F,$83,$3C,$1D,$1F,$0A,$77,$19,$AF,$4E
	.dc.b	$1F,$CF,$2B,$67,$F6,$FD,$7D,$0F,$F6,$78,$D6,$1D,$E3,$71,$C6,$9E
	.dc.b	$D7,$F5,$38,$C3,$6F,$9D,$75,$3F,$B9,$7D,$87,$F0,$47,$B4,$7F,$5B
	.dc.b	$8C,$46,$C3,$8A,$F4,$3F,$82,$8E,$27,$4D,$FC,$0F,$F8,$27,$99,$F2
	.dc.b	$B8,$B4,$62,$78,$9D,$AB,$8A,$FB,$0F,$F2,$3B,$AF,$B0,$FD,$C6,$5F
	.dc.b	$DC,$5B,$1F,$E0,$D9,$DB,$E9,$6E,$1F,$3E,$77,$26,$CE,$AB,$FB,$E2
	.dc.b	$F8,$6F,$1B,$6B,$CE,$D7,$E3,$53,$E9,$37,$5E,$A8,$FF,$2E,$CF,$A7
	.dc.b	$9F,$2F,$29,$D4,$48,$C1,$F1,$3E,$C3,$8B,$F4,$DB,$97,$E2,$5B,$97
	.dc.b	$C3,$EB,$F9,$31,$8A,$FF,$18,$7D,$83,$E4,$FE,$6B,$63,$B6,$3E,$7E
	.dc.b	$09,$1F,$3D,$DD,$70,$F8,$19,$3F,$62,$D1,$D3,$CD,$FB,$3F,$61,$F4
	.dc.b	$AB,$F9,$AB,$5F,$F1,$43,$5F,$F2,$23,$5F,$A9,$DD,$FC,$7F,$80,$D2
	.dc.b	$47,$F2,$6F,$AF,$FF,$65,$37,$E1,$7F,$93,$46,$23,$FD,$96,$D7,$FA
	.dc.b	$3F,$D7,$FF,$9A,$5F,$E5,$ED,$7F,$DA,$2D,$77,$F0,$5F,$CC,$7E,$1B
	.dc.b	$FC,$1F,$1F,$CD,$9A,$FF,$B6,$1A,$FF,$54,$6B,$FE,$47,$FA,$D7,$C1
	.dc.b	$7F,$3B,$6C,$7F,$40,$6B,$FE,$E2,$6B,$FD,$79,$7D,$1F,$12,$47,$17
	.dc.b	$CD,$F5,$B2,$3C,$CD,$7F,$DD,$BF,$58,$F8,$0F,$E9,$8D,$8F,$EA,$2D
	.dc.b	$77,$1F,$D4,$7E,$AD,$F3,$94,$93,$FD,$62,$32,$FC,$5F,$AC,$F9,$04
	.dc.b	$70,$6F,$DD,$E8,$47,$07,$F6,$2F,$AB,$FF,$23,$B7,$D6,$FE,$C5,$18
	.dc.b	$8F,$F0,$08,$70,$FF,$81,$3E,$B3,$FC,$BA,$FF,$6C,$6B,$C1,$AE,$FD
	.dc.b	$AF,$ED,$DF,$46,$C3,$F8,$9B,$67,$F8,$BB,$5F,$E4,$DF,$58,$FA,$CD
	.dc.b	$D7,$22,$4F,$F7,$CA,$E0,$FF,$86,$1E,$5E,$2F,$F4,$3F,$AC,$FE,$88
	.dc.b	$DF,$1B,$9D,$7F,$26,$9F,$F4,$BB,$7E,$17,$F9,$3C,$62,$3F,$9C,$3F
	.dc.b	$86,$74,$43,$31,$C3,$97,$8E,$9C,$BE,$22,$57,$78,$CB,$E1,$8C,$BD
	.dc.b	$A3,$F9,$BF,$D5,$BE,$62,$92,$7F,$9E,$FD,$5B,$EB,$31,$1F,$62,$B8
	.dc.b	$3F,$E2,$06,$B9,$AF,$F1,$04,$63,$5F,$ED,$17,$F2,$35,$FF,$59,$36
	.dc.b	$FF,$D9,$D5,$B9,$CF,$C1,$FF,$4C,$AF,$F7,$65,$79,$3E,$53,$13,$23
	.dc.b	$04,$3F,$B1,$4F,$FA,$71,$F5,$F4,$F3,$23,$EE,$F2,$AB,$C2,$2E,$5E
	.dc.b	$89,$FE,$BB,$F5,$9F,$C7,$2B,$FB,$0B,$5F,$EE,$0D,$9F,$F0,$37,$67
	.dc.b	$FB,$02,$77,$E3,$7F,$90,$EB,$84,$7F,$98,$9A,$FF,$B5,$5A,$FF,$BC
	.dc.b	$5E,$9A,$7F,$B6,$6B,$FB,$19,$FD,$92,$33,$3F,$EF,$78,$FF,$55,$BF
	.dc.b	$DD,$9A,$FF,$81,$5B,$7E,$6F,$33,$FB,$CC,$62,$FB,$7C,$16,$0C,$CE
	.dc.b	$BE,$42,$61,$D5,$FF,$00,$FD,$67,$E8,$AF,$CD,$1A,$9F,$03,$A8,$FF
	.dc.b	$03,$FD,$A7,$F7,$19,$FF,$34,$B5,$FE,$0A,$D7,$FD,$0E,$D7,$FD,$F5
	.dc.b	$F5,$AF,$AD,$7C,$47,$F9,$CA,$30,$DE,$DF,$F1,$7D,$7F,$10,$9C,$BF
	.dc.b	$C4,$BE,$B1,$F3,$FF,$E3,$D8,$12,$33,$97,$F9,$F9,$B1,$C6,$78,$2E
	.dc.b	$07,$D5,$FF,$30,$BF,$F1,$7C,$63,$95,$18,$9F,$EF,$CD,$8C,$D7,$F9
	.dc.b	$06,$B8,$4E,$4E,$1C,$BA,$89,$7D,$EB,$87,$D5,$7F,$58,$BF,$0B,$63
	.dc.b	$A7,$D6,$7C,$89,$CF,$9D,$CD,$FF,$2D,$56,$F0,$EC,$D8,$FF,$2F,$57
	.dc.b	$1B,$EF,$7C,$5F,$93,$F5,$8F,$A5,$C6,$2F,$A5,$FF,$7A,$E3,$FE,$4A
	.dc.b	$AD,$3F,$1A,$33,$7F,$E5,$B9,$CB,$F0,$E3,$35,$FE,$73,$9C,$BF,$0E
	.dc.b	$33,$5F,$6A,$95,$A8,$FF,$3B,$4E,$A7,$CF,$7E,$68,$C4,$70,$8C,$7B
	.dc.b	$DE,$D8,$C5,$79,$E3,$17,$D6,$7F,$05,$7C,$A3,$18,$7C,$1A,$DE,$1D
	.dc.b	$30,$F8,$4D,$C3,$E0,$36,$BD,$CD,$9F,$65,$F4,$C3,$B4,$E6,$FD,$75
	.dc.b	$E5,$3A,$95,$EA,$DB,$EA,$D3,$AC,$F3,$9F,$79,$F4,$70,$F4,$CE,$E3
	.dc.b	$B9,$97,$BB,$4E,$E5,$AE,$ED,$8E,$E7,$37,$BC,$7D,$B9,$EF,$1B,$3B
	.dc.b	$C6,$67,$BB,$63,$BB,$4E,$E5,$8E,$F1,$CC,$EE,$FE,$12,$7B,$CE,$E7
	.dc.b	$BC,$E4,$EE,$57,$BB,$63,$BB,$4E,$EF,$38,$EE,$7E,$1A,$7B,$CE,$E3
	.dc.b	$B9,$97,$BB,$4E,$E5,$9E,$ED,$8F,$6B,$FA,$13,$EC,$97,$0F,$D5,$B6
	.dc.b	$FF,$8C,$BA,$38,$90,$FB,$A6,$E7,$C7,$5F,$C7,$BE,$97,$D3,$3E,$83
	.dc.b	$DB,$19,$9F,$4C,$7C,$03,$3A,$AB,$8C,$67,$0F,$D3,$CE,$6B,$8C,$7E
	.dc.b	$B6,$78,$D7,$08,$C1,$C1,$C6,$1E,$4C,$3C,$55,$F2,$0E,$D1,$FC,$53
	.dc.b	$0E,$D1,$9B,$E7,$39,$BE,$78,$7C,$0B,$DE,$1F,$43,$0E,$73,$FC,$EE
	.dc.b	$B7,$1F,$5A,$7C,$26,$F0,$EB,$5B,$3D,$B8,$79,$F0,$EB,$1F,$88,$25
	.dc.b	$FE,$12,$F9,$C6,$23,$C2,$7F,$47,$9E,$2F,$D3,$BE,$EE,$70,$F2,$DF
	.dc.b	$E2,$0E,$71,$BC,$3C,$53,$AC,$3C,$35,$AA,$EE,$F1,$C3,$BC,$62,$77
	.dc.b	$18,$C3,$C3,$5B,$BE,$97,$EB,$AD,$61,$FD,$1B,$57,$D6,$53,$1F,$29
	.dc.b	$73,$FE,$B7,$17,$0F,$EB,$F3,$72,$FE,$C7,$F3,$7F,$77,$DE,$FB,$CD
	.dc.b	$C3,$FB,$83,$6F,$E4,$36,$BF,$B8,$C3,$EA,$9F,$59,$FD,$6B,$ED,$1E
	.dc.b	$E7,$F4,$39,$CE,$3F,$DF,$A7,$2F,$F8,$2B,$0E,$73,$AC,$3F,$D4,$2D
	.dc.b	$7F,$6C,$D7,$FA,$C3,$5F,$F0,$9B,$C6,$39,$E1,$F1,$FE,$B7,$C4,$FE
	.dc.b	$6F,$E6,$FF,$39,$95,$C7,$FC,$1C,$DC,$7F,$0C,$DC,$3F,$C1,$2F,$D1
	.dc.b	$F3,$A5,$7F,$82,$3C,$FB,$7F,$E0,$C9,$B7,$FE,$04,$9B,$5E,$8F,$81
	.dc.b	$FE,$D5,$EE,$FC,$56,$C7,$C3,$34,$7F,$B0,$B0,$F9,$D1,$FC,$4B,$53
	.dc.b	$49,$3C,$67,$E1,$C7,$EE,$7E,$AF,$B4,$FF,$87,$BE,$B7,$F1,$5F,$15
	.dc.b	$F8,$B6,$C6,$4F,$F1,$1C,$65,$A7,$F8,$CF,$D0,$FE,$08,$7F,$05,$F4
	.dc.b	$6C,$64,$98,$FE,$BE,$32,$FF,$90,$63,$FC,$21,$F5,$AC,$7F,$39,$EF
	.dc.b	$7F,$82,$63,$13,$FE,$34,$8C,$4F,$F3,$29,$CD,$7F,$B9,$9A,$FF,$3B
	.dc.b	$6B,$DA,$38,$D7,$F8,$FB,$AE,$CF,$F0,$3F,$94,$FE,$3C,$7F,$1F,$F9
	.dc.b	$8B,$5F,$E5,$18,$B3,$FC,$26,$6F,$17,$F9,$03,$E9,$7F,$69,$EC,$7E
	.dc.b	$D5,$B8,$7F,$93,$1B,$7F,$E5,$06,$D7,$EA,$BE,$A9,$DD,$FE,$2D,$F8
	.dc.b	$EF,$F1,$54,$63,$0F,$F2,$04,$62,$7F,$A9,$CE,$70,$FE,$B4,$1C,$3F
	.dc.b	$D7,$5A,$FF,$8F,$DE,$17,$E1,$9F,$E7,$DE,$5F,$5C,$3F,$15,$6F,$95
	.dc.b	$7D,$5F,$F6,$26,$CF,$E5,$5B,$3E,$26,$BF,$DB,$5F,$CB,$C6,$F9,$F9
	.dc.b	$11,$FC,$F7,$D6,$BC,$3F,$ED,$ED,$8C,$C6,$0F,$EE,$7F,$5A,$E7,$A4
	.dc.b	$9F,$D4,$46,$5F,$D5,$7C,$F7,$F9,$F7,$2F,$F0,$03,$5F,$F3,$13,$5F
	.dc.b	$F4,$EB,$57,$D6,$7F,$0F,$6F,$87,$FE,$04,$8C,$19,$FA,$C7,$B3,$FE
	.dc.b	$08,$6E,$1F,$BB,$6D,$FD,$86,$BF,$E1,$35,$EE,$F1,$24,$F1,$7F,$C2
	.dc.b	$7E,$4B,$C4,$A4,$9F,$E1,$11,$97,$F8,$50,$7F,$C3,$6D,$32,$E0,$CF
	.dc.b	$C3,$48,$E2,$DC,$3F,$1F,$F5,$2E,$CF,$F1,$2B,$6B,$F8,$C3,$67,$73
	.dc.b	$FE,$2A,$FA,$97,$5D,$25,$FF,$1F,$8C,$B7,$EA,$7F,$D1,$90,$7F,$D1
	.dc.b	$DF,$50,$F3,$33,$1F,$E9,$1F,$A9,$79,$9A,$9C,$37,$0F,$CE,$FB,$DF
	.dc.b	$A2,$7F,$D4,$58,$3C,$9F,$F4,$D4,$E7,$5F,$F6,$FC,$E6,$B8,$F8,$1A
	.dc.b	$3F,$D4,$1F,$52,$F2,$B1,$3F,$AA,$8C,$3C,$0C,$46,$AF,$F9,$21,$FE
	.dc.b	$EC,$FA,$87,$D4,$FF,$B1,$3E,$89,$F0,$FF,$EA,$D9,$CE,$1F,$EE,$89
	.dc.b	$CB,$FB,$88,$CD,$F4,$0E,$3C,$7F,$F7,$25,$7F,$AE,$A3,$15,$FE,$C3
	.dc.b	$6D,$78,$F0,$F1,$7B,$7E,$17,$C5,$CB,$FE,$D8,$FA,$9F,$EE,$18,$71
	.dc.b	$8C,$78,$1A,$8F,$24,$6F,$E2,$B1,$8F,$97,$27,$D3,$FB,$4F,$74,$5B
	.dc.b	$F6,$45,$8E,$05,$D7,$FC,$26,$3E,$93,$FD,$A4,$DA,$FE,$37,$3F,$BC
	.dc.b	$1F,$50,$FA,$7F,$F7,$15,$6E,$7E,$43,$4E,$B7,$FE,$DD,$8E,$55,$C5
	.dc.b	$FF,$0D,$46,$57,$8C,$7B,$DF,$69,$61,$9B,$E5,$5F,$C4,$8D,$9D,$23
	.dc.b	$55,$FC,$FA,$EC,$E4,$D9,$C6,$B9,$E1,$EB,$7A,$C7,$A2,$BC,$57,$CA
	.dc.b	$38,$46,$1E,$66,$CE,$D3,$C9,$F6,$9E,$0A,$ED,$3E,$58,$C6,$1E,$E7
	.dc.b	$99,$C2,$BC,$B1,$97,$A6,$1F,$E1,$F9,$FA,$C7,$F3,$83,$0F,$F8,$8F
	.dc.b	$EA,$3F,$88,$7D,$4D,$F0,$0D,$25,$7F,$64,$97,$A8,$CF,$93,$67,$4D
	.dc.b	$7F,$D8,$93,$C2,$76,$7C,$7C,$FF,$2F,$37,$5F,$BC,$FD,$2A,$E0,$DB
	.dc.b	$ED,$9F,$DE,$BE,$33,$F6,$1B,$1F,$82,$AE,$2D,$AE,$33,$D8,$D9,$6F
	.dc.b	$C3,$3B,$C3,$EE,$5F,$2A,$D6,$1E,$73,$38,$74,$7F,$2D,$3B,$7F,$31
	.dc.b	$3D,$9B,$87,$68,$F7,$CE,$E7,$F5,$53,$FA,$AD,$7B,$C7,$F1,$1A,$F7
	.dc.b	$E1,$FE,$88,$BE,$11,$8A,$F6,$79,$FC,$6F,$C2,$D7,$A9,$FE,$14,$FA
	.dc.b	$9F,$DF,$37,$7D,$46,$23,$F9,$03,$F8,$07,$47,$F2,$7F,$90,$ED,$93
	.dc.b	$A1,$1C,$AF,$C1,$F5,$3F,$E3,$4B,$FF,$11,$FD,$27,$EB,$33,$FB,$ED
	.dc.b	$8F,$DF,$4E,$DF,$E6,$0D,$C7,$FD,$4F,$3B,$93,$67,$55,$F8,$A5,$EE
	.dc.b	$78,$89,$5E,$2F,$A9,$AF,$F3,$DB,$FE,$D6,$F9,$BA,$33,$8F,$FA,$DB
	.dc.b	$0E,$73,$92,$1E,$6C,$3E,$77,$D4,$F2,$8F,$E9,$0D,$AF,$F5,$FB,$72
	.dc.b	$F8,$3D,$BC,$31,$FA,$DF,$52,$F8,$EA,$67,$1F,$F2,$5C,$65,$F9,$7B
	.dc.b	$FD,$DD,$7E,$8C,$6A,$B9,$FA,$DD,$3E,$A6,$4F,$28,$FC,$27,$D4,$FC
	.dc.b	$88,$FE,$5C,$D4,$DC,$55,$DF,$BE,$57,$DC,$79,$FC,$1F,$E3,$FE,$A7
	.dc.b	$FD,$5B,$3F,$E6,$26,$BF,$DA,$DB,$C3,$FE,$E2,$E5,$7D,$A7,$E5,$FE
	.dc.b	$AB,$FD,$C3,$7F,$E6,$A6,$BF,$DC,$DB,$5C,$7C,$EF,$CF,$7D,$57,$F8
	.dc.b	$02,$BF,$CE,$2D,$7F,$BD,$36,$7F,$96,$0F,$B2,$7D,$27,$E9,$FE,$AD
	.dc.b	$F2,$D4,$98,$7F,$9F,$3E,$C9,$3B,$FA,$49,$3F,$E8,$08,$47,$EC,$67
	.dc.b	$CF,$D7,$FC,$7B,$37,$C7,$FA,$09,$87,$FB,$47,$4F,$27,$1F,$EE,$7E
	.dc.b	$B3,$FC,$B7,$3F,$E8,$C6,$BF,$E1,$56,$CF,$F4,$78,$7D,$97,$F8,$FD
	.dc.b	$FF,$49,$FD,$6B,$E7,$7F,$82,$36,$33,$1F,$C1,$E5,$E0,$7C,$CF,$AD
	.dc.b	$95,$F3,$1F,$E0,$3C,$6F,$9E,$E7,$F0,$52,$BB,$BF,$C4,$7E,$B1,$F7
	.dc.b	$1F,$C4,$FE,$BB,$FC,$7D,$5F,$EA,$66,$BF,$E2,$F6,$ED,$EB,$A7,$CA
	.dc.b	$73,$EA,$97,$FE,$AE,$9F,$F2,$BB,$F2,$73,$FD,$1B,$FE,$39,$FA,$46
	.dc.b	$7F,$C8,$DB,$1F,$EB,$66,$BF,$E3,$FF,$AE,$FC,$1B,$99,$E0,$47,$FD
	.dc.b	$7B,$1A,$7B,$47,$F2,$FF,$AE,$FF,$15,$F4,$7F,$5C,$6C,$FF,$86,$EF
	.dc.b	$BE,$9F,$5E,$5F,$5D,$C4,$95,$FE,$CA,$8F,$C5,$B6,$32,$FF,$95,$63
	.dc.b	$8B,$BE,$BF,$16,$BE,$3F,$8C,$93,$CE,$32,$F5,$8C,$BC,$E3,$FA,$47
	.dc.b	$D6,$FD,$08,$FF,$6D,$B5,$FF,$32,$37,$1F,$C8,$7D,$6B,$EB,$B8,$F1
	.dc.b	$4B,$E2,$FF,$53,$9D,$46,$0F,$F7,$3B,$F1,$36,$FF,$01,$B6,$3F,$C8
	.dc.b	$58,$77,$9C,$FC,$4F,$3F,$95,$95,$CA,$3B,$18,$7F,$CF,$57,$FE,$DF
	.dc.b	$74,$78,$3A,$FC,$84,$7C,$7E,$75,$78,$C5,$D3,$DF,$3F,$DB,$7E,$B9
	.dc.b	$E0,$52,$5F,$F9,$66,$BF,$69,$F2,$5F,$B3,$8F,$F4,$65,$70,$75,$A7
	.dc.b	$F8,$8A,$31,$1A,$BF,$EF,$55,$C6,$FF,$7D,$5A,$7A,$FB,$7A,$3C,$37
	.dc.b	$FE,$99,$AF,$F5,$0B,$D8,$FF,$36,$38,$75,$F0,$FD,$BD,$4E,$EC,$8E
	.dc.b	$07,$FB,$B1,$D3,$89,$FF,$7B,$B7,$E5,$3C,$E7,$F8,$47,$EB,$BF,$43
	.dc.b	$1F,$EF,$48,$CB,$B3,$AC,$6B,$EA,$FF,$DC,$91,$FE,$44,$3F,$C8,$AE
	.dc.b	$1C,$07,$0E,$DA,$F0,$F8,$2F,$F3,$24,$7F,$86,$E3,$8B,$5E,$11,$FE
	.dc.b	$69,$FA,$C7,$D6,$BE,$47,$BD,$27,$31,$85,$DE,$F8,$3C,$5F,$1B,$F5
	.dc.b	$70,$3E,$AB,$F9,$97,$27,$F3,$F6,$CF,$FB,$6A,$75,$CB,$FD,$5A,$BF
	.dc.b	$15,$5A,$9F,$EB,$2D,$8C,$F3,$6E,$FF,$CE,$D3,$FE,$0B,$9E,$D3,$9A
	.dc.b	$7D,$57,$F5,$E9,$FE,$A0,$DB,$F1,$79,$3F,$06,$73,$E9,$F1,$AD,$5F
	.dc.b	$F6,$A6,$CF,$E7,$EB,$85,$FA,$E7,$2F,$FA,$EF,$EB,$1F,$49,$E3,$FA
	.dc.b	$DF,$E9,$F8,$7E,$3A,$72,$FF,$63,$8C,$D7,$FB,$CE,$72,$FC,$78,$CE
	.dc.b	$3F,$D8,$67,$2F,$F7,$E8,$CD,$7F,$A3,$A7,$2F,$ED,$63,$35,$E5,$BF
	.dc.b	$27,$9F,$F2,$E1,$7C,$1F,$F5,$15,$70,$73,$5C,$2F,$D4,$D8,$F4,$37
	.dc.b	$0F,$33,$67,$C8,$D9,$F1,$47,$C1,$C3,$C5,$87,$86,$D8,$78,$63,$0D
	.dc.b	$C3,$96,$1E,$4C,$39,$1E,$42,$BE,$46,$DF,$91,$A7,$90,$B1,$E4,$3E
	.dc.b	$86,$1E,$4A,$EF,$87,$8A,$31,$5C,$2B,$C4,$70,$7C,$4D,$3C,$45,$BF
	.dc.b	$13,$63,$C4,$D3,$C5,$1F,$76,$7B,$C6,$4E,$F1,$99,$EE,$D8,$EE,$D3
	.dc.b	$B9,$63,$BB,$63,$BC,$7E,$1E,$7B,$CE,$67,$BC,$E4,$EE,$57,$BB,$63
	.dc.b	$BB,$4E,$E5,$8E,$E7,$E3,$A7,$BC,$66,$3B,$99,$7B,$B4,$EE,$59,$EE
	.dc.b	$D8,$EE,$D3,$BC,$7E,$32,$7B,$C6,$4E,$F1,$99,$EE,$D8,$EE,$D3,$B9
	.dc.b	$63,$BB,$63,$BE,$1F,$97,$9F,$C3,$5F,$E2,$5E,$B1,$99,$FB,$93,$97
	.dc.b	$BC,$66,$BE,$94,$E5,$F0,$C6,$6B,$E4,$4E,$5F,$1C,$66,$BD,$D3,$97
	.dc.b	$CB,$7F,$0A,$FD,$11,$F0,$AB,$84,$62,$B8,$4E,$6B,$85,$68,$8E,$09
	.dc.b	$7E,$E6,$CF,$B1,$B7,$EA,$6C,$FA,$1B,$3E,$68,$E4,$47,$93,$67,$93
	.dc.b	$4E,$47,$F4,$0F,$AC,$FE,$78,$7F,$3D,$F9,$6B,$19,$25,$7F,$3B,$8C
	.dc.b	$BF,$D5,$F7,$7C,$EF,$F4,$D5,$F0,$76,$30,$DB,$FE,$C0,$DA,$E4,$FF
	.dc.b	$2F,$FA,$97,$CA,$7F,$57,$FA,$C7,$D5,$E6,$30,$DC,$3F,$D5,$13,$A7
	.dc.b	$FC,$2F,$5C,$9C,$CE,$F0,$6B,$FE,$A7,$AD,$47,$F5,$7F,$AC,$D4,$FF
	.dc.b	$6E,$F9,$DF,$65,$FB,$5B,$87,$F7,$96,$DF,$C3,$6D,$7F,$65,$F3,$BF
	.dc.b	$B4,$1F,$DA,$7C,$FB,$7F,$DC,$22,$DF,$F6,$79,$B5,$E5,$FA,$6F,$EC
	.dc.b	$DE,$F7,$C5,$6E,$1E,$E6,$DF,$B2,$0C,$D1,$3A,$BE,$95,$C4,$FF,$09
	.dc.b	$7D,$5F,$57,$FC,$07,$F1,$1F,$DF,$FE,$3F,$FC,$00,$D8,$C9,$FE,$08
	.dc.b	$8C,$B4,$FF,$0E,$7A,$1F,$83,$3F,$09,$E8,$D8,$C9,$33,$FC,$54,$65
	.dc.b	$F5,$1E,$AF,$AC,$63,$FE,$11,$F9,$EF,$EE,$91,$8B,$FF,$0C,$C6,$27
	.dc.b	$51,$FE,$58,$9D,$57,$F8,$4E,$B5,$39,$AF,$F0,$E7,$76,$2B,$F3,$9E
	.dc.b	$4F,$E6,$CF,$CE,$7C,$D5,$BF,$F1,$C4,$59,$FD,$14,$DD,$BF,$C4,$9F
	.dc.b	$59,$FE,$27,$F5,$FF,$C5,$0D,$C3,$FC,$5E,$DB,$FF,$13,$36,$BF,$DF
	.dc.b	$5F,$58,$EF,$FF,$13,$7C,$9F,$F8,$4E,$31,$87,$F8,$96,$31,$3A,$8F
	.dc.b	$F3,$BC,$EB,$0F,$F1,$35,$6A,$73,$5F,$B7,$FA,$CC,$57,$D0,$3F,$CB
	.dc.b	$5F,$59,$FE,$47,$FA,$D7,$CC,$FF,$88,$FC,$F4,$E6,$31,$D3,$FE,$4D
	.dc.b	$F8,$C5,$E3,$3F,$F4,$86,$C9,$CC,$7F,$A3,$3E,$B7,$F8,$3B,$A8,$7C
	.dc.b	$7F,$F8,$C6,$73,$A7,$F9,$96,$73,$5F,$E9,$4F,$AD,$75,$7F,$97,$75
	.dc.b	$FE,$27,$39,$8C,$79,$1F,$E3,$BF,$A2,$FF,$29,$E9,$EA,$F8,$7C,$7A
	.dc.b	$5C,$5F,$F2,$07,$D6,$FF,$04,$96,$3F,$EB,$BF,$B2,$77,$EB,$CC,$FF
	.dc.b	$36,$46,$27,$FD,$4F,$F5,$CF,$7B,$FC,$E1,$F6,$5F,$EE,$59,$CC,$63
	.dc.b	$E2,$3F,$C9,$D4,$F8,$1F,$F3,$B7,$98,$F0,$B3,$18,$7F,$CA,$9C,$49
	.dc.b	$FF,$3E,$7D,$93,$E5,$F3,$18,$7F,$CB,$59,$FC,$4E,$9F,$F4,$77,$D9
	.dc.b	$3E,$33,$38,$FF,$A1,$63,$15,$FE,$C0,$1C,$BF,$E8,$9F,$5B,$FD,$BF
	.dc.b	$19,$8C,$7C,$57,$F9,$9F,$EC,$5E,$56,$FE,$2B,$0D,$BF,$F6,$5F,$D8
	.dc.b	$BC,$AF,$F3,$72,$F8,$1B,$18,$73,$F6,$32,$BF,$DA,$7F,$60,$F6,$B1
	.dc.b	$C5,$F9,$98,$C4,$EA,$7F,$CF,$5F,$62,$F6,$FF,$CF,$95,$AB,$E1,$EB
	.dc.b	$E0,$FF,$40,$7C,$57,$EA,$7C,$3F,$C0,$CF,$F2,$CA,$56,$BD,$29,$9F
	.dc.b	$FB,$9B,$EC,$7F,$05,$7F,$AB,$C7,$94,$FF,$34,$84,$EB,$1E,$B1,$FE
	.dc.b	$CF,$BF,$F4,$84,$E6,$B8,$5F,$2A,$FE,$7B,$F5,$EF,$B1,$FF,$5E,$7D
	.dc.b	$5B,$E4,$7F,$D3,$13,$9C,$3F,$63,$39,$7F,$DF,$71,$9B,$FE,$00,$DC
	.dc.b	$FD,$E2,$BF,$D3,$D1,$9A,$F5,$36,$BC,$78,$78,$BE,$1F,$C2,$F8,$B9
	.dc.b	$3F,$DA,$1F,$63,$FA,$BC,$38,$C6,$3C,$AD,$47,$92,$37,$D7,$8C,$7C
	.dc.b	$BF,$5F,$EC,$3D,$AD,$AF,$5C,$58,$F5,$97,$5F,$9A,$3E,$9B,$FD,$78
	.dc.b	$DA,$FE,$CE,$FD,$29,$FE,$D3,$E8,$7F,$6A,$FB,$07,$77,$D9,$AE,$15
	.dc.b	$FC,$9D,$A7,$6C,$3F,$D9,$11,$C8,$57,$27,$BC,$69,$5F,$7B,$ED,$3F
	.dc.b	$90,$AE,$15,$F8,$9A,$D3,$CB,$0A,$E4,$EF,$7C,$A3,$31,$E3,$AE,$78
	.dc.b	$7A,$9E,$B1,$9A,$E0,$F2,$3F,$30,$79,$8C,$1F,$E2,$AC,$3C,$D1,$A8
	.dc.b	$D4,$F6,$9C,$C7,$2C,$3C,$F7,$E1,$BE,$11,$8A,$DD,$7A,$EF,$96,$1E
	.dc.b	$A8,$F1,$1D,$CA,$72,$30,$60,$FF,$11,$FD,$87,$F1,$4F,$B1,$E0,$D8
	.dc.b	$F4,$35,$F7,$B5,$FD,$C2,$F8,$1F,$F2,$4C,$70,$6B,$C1,$AF,$06,$BE
	.dc.b	$27,$8D,$FD,$CC,$9A,$7F,$05,$8C,$36,$39,$69,$F7,$B0,$F1,$3F,$80
	.dc.b	$8E,$38,$5A,$E0,$F7,$9E,$18,$78,$2B,$F8,$91,$C2,$7E,$DB,$70,$FC
	.dc.b	$7E,$1D,$A7,$EF,$B6,$3C,$B1,$DC,$EB,$71,$E7,$87,$89,$FD,$4B,$6B
	.dc.b	$95,$F8,$A3,$F5,$99,$5A,$FD,$9B,$5F,$F0,$A5,$70,$7F,$CB,$B5,$E9
	.dc.b	$BF,$E5,$2B,$DE,$E7,$3A,$6B,$E2,$3A,$9F,$E6,$6F,$B1,$FE,$08,$E0
	.dc.b	$38,$7F,$C4,$8D,$AF,$E6,$EF,$E0,$E3,$83,$9C,$3F,$85,$F3,$3E,$67
	.dc.b	$C7,$7F,$CA,$DF,$1D,$6E,$FF,$CE,$FF,$63,$FE,$3F,$7C,$5F,$61,$FB
	.dc.b	$6C,$7F,$A4,$36,$37,$DB,$EC,$AE,$11,$FE,$39,$9D,$26,$C6,$B1,$FE
	.dc.b	$9E,$BE,$27,$91,$2B,$C3,$F0,$75,$E9,$CF,$F2,$B0,$FE,$F1,$1F,$1E
	.dc.b	$F9,$36,$F5,$52,$30,$7C,$DF,$B1,$E2,$FF,$93,$DB,$8F,$F6,$26,$E7
	.dc.b	$F0,$3B,$3E,$CC,$62,$BF,$D3,$BF,$60,$F9,$8F,$BA,$D8,$ED,$8E,$27
	.dc.b	$D1,$C1,$27,$EF,$56,$DF,$F3,$03,$FE,$D8,$73,$F6,$1D,$0E,$BE,$7F
	.dc.b	$E2,$BE,$C7,$E5,$69,$FD,$ED,$B5,$FE,$B0,$6A,$B5,$FC,$9E,$E9,$3F
	.dc.b	$79,$E9,$F1,$FF,$86,$FD,$9B,$84,$FF,$9A,$17,$D8,$E4,$D6,$1F,$8C
	.dc.b	$B7,$8F,$49,$7F,$E7,$08,$7B,$BF,$92,$FA,$0F,$3F,$21,$A4,$F8,$4F
	.dc.b	$17,$CF,$CC,$3C,$4F,$97,$E7,$F9,$7A,$1C,$FD,$D9,$5F,$E7,$EC,$3F
	.dc.b	$C5,$CD,$E0,$F0,$75,$48,$C1,$ED,$FB,$9F,$81,$F5,$92,$30,$73,$F9
	.dc.b	$3F,$41,$1E,$DF,$05,$EB,$FA,$89,$1E,$17,$DD,$F0,$3F,$B0,$FB,$9F
	.dc.b	$F3,$94,$FF,$A4,$9A,$FF,$B3,$B8,$91,$87,$FC,$43,$D1,$FE,$08,$EB
	.dc.b	$FF,$02,$EE,$E3,$F9,$5C,$73,$1F,$81,$1F,$37,$7F,$9D,$BC,$9F,$89
	.dc.b	$1F,$C4,$BE,$AB,$FC,$6D,$1F,$C3,$DF,$E0,$CD,$EE,$D7,$AD,$23,$04
	.dc.b	$3F,$D6,$D3,$C9,$B1,$D6,$64,$75,$8F,$E4,$38,$7F,$9E,$A3,$F8,$F3
	.dc.b	$5F,$F6,$03,$75,$FF,$29,$DF,$FA,$2E,$77,$F0,$3F,$CA,$BE,$AB,$E2
	.dc.b	$40,$38,$7E,$34,$6F,$8A,$47,$F2,$D9,$47,$3E,$2C,$CF,$FB,$46,$72
	.dc.b	$EE,$74,$F1,$E8,$F1,$74,$48,$C1,$0E,$4E,$5E,$91,$E9,$31,$97,$F9
	.dc.b	$C2,$FF,$D5,$53,$C6,$3F,$CE,$7F,$55,$FE,$B1,$8F,$E8,$8D,$7F,$DC
	.dc.b	$CD,$8F,$F3,$CF,$D5,$7E,$BA,$77,$CF,$FD,$42,$3E,$67,$2F,$FA,$05
	.dc.b	$B3,$F2,$6F,$94,$E7,$DC,$E1,$F2,$12,$BF,$AA,$5F,$FA,$41,$B3,$97
	.dc.b	$FD,$F1,$1F,$EE,$27,$27,$0F,$A7,$F1,$11,$E7,$E3,$D7,$8C,$5C,$FD
	.dc.b	$73,$71,$F4,$4D,$AF,$C0,$5F,$F2,$9B,$FF,$4C,$FD,$53,$C2,$FE,$DE
	.dc.b	$DA,$FE,$D8,$D5,$3F,$EA,$0F,$8C,$7C,$27,$0B,$FE,$EF,$1F,$D6,$8F
	.dc.b	$EB,$77,$C6,$B8,$74,$68,$98,$7F,$79,$AF,$D5,$BD,$B3,$FF,$58,$47
	.dc.b	$E9,$EB,$8C,$EE,$BB,$F4,$F8,$3E,$3A,$4F,$37,$FC,$13,$F5,$5F,$9D
	.dc.b	$C3,$F8,$36,$3D,$27,$75,$DC,$FD,$25,$6E,$33,$8F,$BD,$E7,$5F,$E1
	.dc.b	$4F,$AA,$77,$FF,$3A,$C3,$FC,$D5,$18,$F2,$3F,$CB,$CF,$F1,$1F,$5E
	.dc.b	$D6,$EB,$FC,$EB,$E1,$3E,$73,$FD,$35,$1F,$B5,$8C,$63,$FE,$67,$6C
	.dc.b	$7F,$9D,$DA,$FF,$A7,$7E,$99,$CB,$FE,$C2,$AF,$F5,$2C,$61,$FF,$14
	.dc.b	$EA,$8F,$EE,$51,$87,$55,$FC,$B5,$B3,$FE,$CC,$6C,$F4,$E6,$ED,$5F
	.dc.b	$84,$9C,$BF,$86,$FA,$A7,$97,$AC,$11,$8C,$3C,$7F,$17,$FD,$89,$F2
	.dc.b	$C6,$6F,$F9,$F3,$6B,$FD,$BA,$D9,$FF,$60,$D7,$1A,$DD,$F7,$AD,$47
	.dc.b	$8E,$3C,$53,$BB,$E7,$1B,$25,$F9,$38,$2D,$F4,$8F,$2C,$F2,$9C,$D7
	.dc.b	$29,$C9,$C4,$B5,$C4,$91,$FE,$F6,$BF,$14,$FF,$B6,$B0,$F3,$CF,$C6
	.dc.b	$C3,$AC,$FB,$B0,$B8,$7A,$E3,$AD,$73,$9C,$BC,$63,$27,$12,$CF,$13
	.dc.b	$C6,$E5,$A6,$63,$D8,$FF,$6F,$C3,$CE,$E6,$7D,$17,$E0,$9C,$D7,$86
	.dc.b	$B9,$B4,$F3,$57,$89,$C3,$C8,$F9,$11,$C3,$0F,$86,$7C,$4A,$EA,$E2
	.dc.b	$FC,$18,$7C,$08,$C5,$F2,$C3,$D2,$47,$B6,$1E,$BA,$E1,$7C,$A7,$C9
	.dc.b	$7C,$AB,$51,$89,$D4,$79,$5B,$7D,$63,$E2,$11,$61,$E8,$9E,$F5,$E5
	.dc.b	$3C,$73,$C4,$F8,$75,$FC,$BA,$B8,$4F,$84,$E1,$1D,$6F,$CD,$5F,$32
	.dc.b	$3C,$93,$F3,$6B,$8C,$60,$F5,$96,$7C,$11,$A7,$C5,$5F,$3F,$0F,$3D
	.dc.b	$7D,$AA,$D5,$FA,$09,$1E,$79,$DD,$F4,$6C,$EA,$3E,$1C,$62,$7E,$35
	.dc.b	$74,$7D,$98,$7E,$B6,$FE,$04,$78,$6B,$85,$FC,$63,$9C,$E9,$F0,$34
	.dc.b	$F0,$5F,$58,$E1,$87,$B6,$75,$7E,$2B,$EB,$39,$8F,$8D,$3C,$A7,$38
	.dc.b	$79,$67,$27,$19,$E7,$7E,$5C,$3C,$18,$7D,$69,$FD,$74,$FE,$1E,$9F
	.dc.b	$54,$FA,$1F,$E8,$31,$86,$C6,$6B,$FC,$71,$30,$D3,$5F,$F1,$E4,$6D
	.dc.b	$FE,$B8,$68,$FE,$79,$F3,$FF,$CE,$1C,$5F,$F2,$CF,$A9,$FE,$50,$7F
	.dc.b	$29,$F7,$6E,$1F,$CD,$22,$D7,$F3,$49,$B5,$FD,$17,$E5,$3F,$9F,$6E
	.dc.b	$F6,$2C,$61,$B8,$FF,$4F,$6E,$1F,$D3,$DB,$7F,$D3,$DB,$3F,$D6,$9F
	.dc.b	$47,$D2,$BE,$2B,$FA,$67,$D4,$BE,$A3,$31,$86,$DF,$F9,$7F,$03,$F2
	.dc.b	$AF,$93,$FD,$43,$EA,$77,$97,$F6,$1F,$94,$F5,$E1,$EC,$6D,$7F,$67
	.dc.b	$6C,$EE,$3F,$B5,$4F,$F5,$1F,$32,$47,$F5,$CF,$36,$D7,$F5,$C9,$B9
	.dc.b	$7F,$5C,$9B,$5F,$D0,$E6,$D7,$F6,$B9,$AB,$27,$D2,$FF,$6D,$FA,$A7
	.dc.b	$D5,$7C,$68,$C3,$71,$F8,$47,$C1,$8F,$AF,$F5,$39,$C3,$FB,$7F,$BB
	.dc.b	$FD,$A7,$DD,$FE,$D4,$DB,$FE,$EA,$D9,$DC,$7F,$78,$9F,$C0,$F9,$F2
	.dc.b	$3F,$C1,$9E,$85,$AF,$F0,$64,$DD,$FF,$C1,$93,$6B,$FC,$2F,$36,$BF
	.dc.b	$C1,$93,$57,$D5,$32,$FE,$FD,$F4,$7F,$9B,$8C,$61,$FE,$0A,$8C,$4C
	.dc.b	$8F,$F0,$6F,$75,$BF,$F0,$F7,$C8,$FE,$60,$FC,$CF,$CE,$5A,$FF,$12
	.dc.b	$45,$9F,$F1,$24,$DD,$3F,$C2,$7F,$03,$FE,$0F,$F5,$FF,$C2,$0D,$C3
	.dc.b	$FC,$36,$DB,$FF,$0D,$B6,$B8,$CF,$F8,$82,$B8,$B4,$FF,$1B,$F8,$2F
	.dc.b	$43,$FC,$54,$D8,$CF,$C3,$E2,$7F,$C5,$DF,$4F,$FE,$32,$C0,$FF,$8F
	.dc.b	$B2,$FF,$74,$39,$31,$1F,$A4,$8F,$F3,$EC,$3E,$57,$9F,$9D,$D1,$B8
	.dc.b	$FF,$92,$BE,$9B,$F7,$CD,$CF,$FC,$7D,$18,$C3,$FD,$EC,$A3,$3F,$13
	.dc.b	$2B,$97,$D4,$DB,$FF,$7E,$7D,$33,$AB,$FC,$90,$DA,$FF,$2C,$AE,$AC
	.dc.b	$CF,$F9,$37,$E9,$9F,$55,$4C,$7C,$25,$C3,$FC,$C3,$E1,$7F,$25,$E2
	.dc.b	$F8,$01,$75,$FF,$95,$DD,$4F,$F0,$97,$2F,$F9,$AB,$E9,$9E,$D6,$E3
	.dc.b	$FC,$BD,$18,$7B,$9F,$E6,$1F,$A5,$75,$5E,$AC,$36,$7F,$CE,$BE,$67
	.dc.b	$F1,$4E,$2F,$F4,$A6,$9F,$CA,$FC,$BF,$F3,$54,$BE,$9A,$C6,$3E,$BA
	.dc.b	$61,$CB,$E5,$EE,$3F,$D0,$BE,$99,$E0,$7F,$41,$7F,$C2,$B5,$FC,$12
	.dc.b	$42,$37,$19,$8C,$7A,$FC,$7E,$16,$4E,$7E,$95,$E8,$E2,$FF,$A8,$C6
	.dc.b	$27,$53,$FD,$47,$E9,$9E,$8F,$F5,$3A,$D5,$F0,$F3,$70,$7F,$57,$F7
	.dc.b	$1E,$C6,$32,$F2,$46,$1B,$1F,$D7,$7E,$9B,$EB,$DF,$F6,$2A,$D3,$FD
	.dc.b	$92,$33,$CB,$FA,$18,$4E,$A3,$13,$AB,$FE,$E3,$39,$7F,$B9,$FD,$33
	.dc.b	$E9,$BF,$D3,$DF,$3E,$F8,$9F,$ED,$D3,$9A,$FF,$58,$CE,$5F,$99,$19
	.dc.b	$CB,$FB,$BB,$72,$FC,$AD,$3E,$73,$9E,$33,$0E,$2F,$82,$31,$87,$2E
	.dc.b	$5F,$F5,$87,$D3,$7E,$2F,$1F,$76,$5F,$72,$BF,$1B,$39,$C7,$D7,$7E
	.dc.b	$83,$8F,$C6,$D8,$F3,$FD,$1F,$B0,$F6,$DD,$8C,$91,$F5,$99,$3F,$C2
	.dc.b	$78,$FF,$0D,$8F,$C6,$36,$3F,$C2,$E7,$E4,$BE,$9B,$F2,$7F,$4A,$FA
	.dc.b	$4F,$F1,$0D,$6C,$FB,$05,$BC,$FA,$5F,$E1,$DC,$39,$0B,$E4,$FF,$80
	.dc.b	$E3,$4B,$0E,$F8,$72,$BB,$3F,$E2,$F9,$AB,$0F,$46,$1D,$23,$FC,$6B
	.dc.b	$3A,$9C,$DF,$F8,$D6,$31,$87,$F2,$FB,$E1,$7E,$67,$95,$F0,$8F,$01
	.dc.b	$E9,$BF,$23,$4C,$CF,$0A,$E9,$3D,$E7,$47,$9E,$32,$F0,$8E,$73,$E2
	.dc.b	$8F,$84,$70,$AF,$13,$EC,$3F,$C9,$C7,$CC,$2D,$E1,$C1,$FE,$56,$FA
	.dc.b	$57,$AE,$FA,$6E,$31,$86,$C6,$67,$FC,$B5,$39,$6F,$B3,$F3,$DB,$87
	.dc.b	$28,$C4,$7F,$98,$E5,$18,$8F,$13,$C7,$1F,$AD,$93,$5F,$F3,$5C,$61
	.dc.b	$B1,$C3,$0F,$BF,$87,$47,$F0,$31,$C7,$0B,$5C,$1F,$E5,$35,$B7,$E8
	.dc.b	$B6,$FB,$47,$03,$AE,$1E,$5B,$F0,$47,$D8,$7B,$36,$FB,$2B,$E9,$3C
	.dc.b	$2F,$AC,$EC,$EB,$87,$77,$FD,$07,$3D,$23,$F9,$03,$75,$E4,$FF,$A3
	.dc.b	$8C,$4F,$F4,$45,$F1,$FF,$B2,$9D,$36,$75,$8D,$5F,$4D,$FE,$0B,$7F
	.dc.b	$A5,$35,$FF,$4A,$B7,$0F,$E6,$4F,$DA,$8D,$EF,$D6,$5C,$CF,$93,$C3
	.dc.b	$6B,$67,$F5,$7F,$A5,$FE,$8E,$FF,$59,$FA,$6F,$D6,$E9,$FE,$AA,$6C
	.dc.b	$7F,$60,$9D,$BF,$EA,$86,$E1,$FD,$16,$77,$26,$CE,$AB,$FD,$6A,$FF
	.dc.b	$85,$67,$C5,$EB,$7C,$38,$C4,$7F,$AD,$29,$E3,$E6,$BE,$44,$7F,$76
	.dc.b	$9F,$47,$3E,$5E,$66,$08,$79,$E7,$E8,$7D,$37,$18,$FF,$60,$B7,$2F
	.dc.b	$EB,$CD,$CB,$E1,$75,$F0,$C3,$EC,$7D,$23,$E4,$A9,$9C,$7F,$C0,$51
	.dc.b	$97,$E6,$6F,$E8,$36,$EB,$CE,$E7,$EE,$48,$C1,$FE,$0B,$FA,$6F,$90
	.dc.b	$FF,$84,$1A,$9B,$69,$F8,$BD,$D2,$BE,$E3,$F8,$FA,$FF,$24,$47,$F8
	.dc.b	$6F,$67,$D1,$B0,$FC,$52,$FA,$6A,$62,$3F,$1D,$87,$F8,$0A,$F9,$35
	.dc.b	$FF,$19,$B6,$FF,$C9,$2B,$E9,$BF,$2D,$F3,$1E,$C9,$34,$9F,$2F,$E2
	.dc.b	$FA,$39,$87,$9F,$34,$79,$BE,$6B,$9F,$9F,$D3,$DD,$97,$ED,$C7,$FC
	.dc.b	$52,$DF,$82,$F0,$F4,$C8,$C1,$D7,$E9,$3E,$0F,$D2,$C8,$C1,$CF,$DC
	.dc.b	$F4,$91,$ED,$EE,$FB,$3E,$8E,$47,$85,$F7,$FD,$27,$ED,$7E,$9B,$FC
	.dc.b	$C7,$3F,$45,$AF,$F9,$69,$BC,$7F,$E7,$0E,$2F,$EF,$3D,$FF,$EF,$5D
	.dc.b	$FE,$0D,$FB,$78,$BE,$59,$F3,$38,$FE,$6E,$FF,$3B,$79,$5F,$0E,$3F
	.dc.b	$88,$7D,$37,$F8,$C6,$3F,$CB,$0F,$F0,$57,$FB,$C6,$3F,$E5,$E5,$19
	.dc.b	$CF,$FC,$84,$F1,$8D,$9F,$E6,$47,$11,$AE,$DE,$E7,$83,$D4,$95,$E0
	.dc.b	$8C,$BE,$28,$FE,$49,$93,$E1,$A9,$27,$CF,$F0,$55,$F0,$8F,$A6,$F8
	.dc.b	$EF,$F9,$D9,$AF,$E2,$27,$F9,$84,$38,$E9,$27,$FC,$F5,$0E,$3C,$47
	.dc.b	$68,$E0,$EA,$79,$CF,$4E,$9E,$BD,$32,$B9,$C6,$5E,$B1,$97,$A4,$65
	.dc.b	$ED,$19,$D7,$FC,$E1,$87,$E6,$A7,$DF,$F4,$CE,$7F,$E9,$4D,$8F,$F4
	.dc.b	$93,$5F,$8B,$F4,$DF,$B1,$73,$3F,$D4,$63,$E7,$47,$FA,$65,$AF,$68
	.dc.b	$D7,$BB,$BF,$96,$93,$B8,$CB,$C6,$32,$F8,$23,$B7,$D1,$C8,$EC,$F8
	.dc.b	$7C,$7C,$47,$9A,$2E,$7F,$E8,$FC,$7F,$95,$4E,$4E,$24,$3D,$B5,$FD
	.dc.b	$AF,$E9,$9E,$1D,$24,$FF,$AD,$7E,$55,$F0,$B8,$8F,$CF,$1F,$EB,$CA
	.dc.b	$E2,$FF,$8E,$9A,$E8,$38,$78,$CF,$E8,$3A,$BF,$50,$D8,$FF,$63,$B5
	.dc.b	$3F,$AB,$8D,$F4,$7F,$80,$E3,$44,$E9,$C7,$C6,$4B,$F5,$3F,$E0,$9F
	.dc.b	$A6,$FC,$ED,$7F,$08,$C3,$C3,$3B,$AF,$D2,$D6,$A3,$1B,$7F,$AE,$EF
	.dc.b	$85,$7F,$B7,$2B,$F4,$FE,$07,$F8,$36,$3E,$03,$5F,$8E,$D7,$F8,$33
	.dc.b	$5F,$F7,$2F,$D3,$3C,$1F,$E8,$91,$FC,$2E,$31,$F0,$7F,$E6,$67,$F9
	.dc.b	$DF,$B3,$6B,$75,$FE,$89,$FA,$37,$2B,$6F,$EB,$11,$86,$F2,$66,$BF
	.dc.b	$89,$EC,$AD,$46,$1B,$1B,$35,$76,$FF,$DF,$B1,$C6,$F1,$7E,$3A,$B2
	.dc.b	$FA,$67,$97,$86,$E1,$FD,$B7,$E2,$FF,$DC,$EF,$F6,$F8,$CE,$1E,$66
	.dc.b	$DF,$8D,$B3,$E2,$AE,$35,$BB,$EC,$49,$ED,$B5,$9D,$47,$79,$F3,$D7
	.dc.b	$2C,$3B,$46,$DB,$3C,$8E,$04,$AE,$67,$CC,$AE,$51,$F1,$B0,$F3,$CF
	.dc.b	$FA,$FD,$B5,$F1,$6F,$C3,$87,$7B,$E1,$16,$FF,$D8,$D8,$74,$79,$46
	.dc.b	$6F,$95,$79,$67,$C7,$1E,$86,$C7,$C7,$BE,$18,$78,$8F,$1E,$1D,$6F
	.dc.b	$9D,$F0,$3B,$DF,$58,$E8,$47,$CE,$64,$B1,$C6,$39,$9F,$8B,$9E,$D5
	.dc.b	$E1,$C3,$D5,$18,$7C,$A7,$C4,$BF,$9D,$5D,$EF,$C6,$7C,$FC,$37,$3A
	.dc.b	$3D,$C5,$8D,$CF,$11,$7D,$E7,$51,$E6,$78,$CF,$A7,$0F,$34,$7B,$63
	.dc.b	$9B,$EE,$9E,$95,$CE,$78,$61,$E9,$6D,$76,$9D,$57,$8E,$FE,$8E,$1E
	.dc.b	$AB,$F9,$15,$AB,$EC,$48,$FC,$1C,$EE,$FC,$2D,$9D,$47,$BA,$31,$3E
	.dc.b	$83,$8C,$FC,$0B,$F6,$C7,$CB,$8C,$61,$F5,$2B,$8D,$FC,$5A,$E2,$7B
	.dc.b	$0F,$3C,$6A,$FD,$77,$C8,$F8,$11,$9B,$F8,$97,$F0,$5E,$D3,$BB,$E6
	.dc.b	$D9,$F6,$B4,$F6,$E1,$F1,$67,$D9,$3E,$53,$E0,$4E,$E3,$33,$B9,$D6
	.dc.b	$1E,$69,$F5,$47,$66,$D7,$6C,$3D,$0F,$E0,$67,$8E,$1E,$8A,$D9,$DA
	.dc.b	$B6,$FC,$B3,$F5,$43,$E9,$9F,$37,$FC,$9B,$E8,$1E,$BF,$F2,$AA,$D4
	.dc.b	$E6,$31,$5C,$32,$FF,$4A,$5F,$07,$F9,$99,$FE,$53,$8D,$5F,$F3,$E9
	.dc.b	$D3,$FC,$AF,$E9,$BF,$99,$1F,$CC,$FE,$32,$DF,$F3,$18,$B3,$FC,$C6
	.dc.b	$6E,$9F,$D0,$F8,$5F,$4D,$FC,$F9,$7B,$36,$30,$DC,$7F,$A9,$B7,$0F
	.dc.b	$E9,$CD,$BE,$8D,$9F,$E9,$B7,$D1,$F6,$D0,$7D,$1B,$E5,$3F,$A7,$FD
	.dc.b	$33,$E9,$7E,$54,$E6,$FF,$D7,$53,$97,$9D,$F2,$BF,$EC,$8F,$F8,$7A
	.dc.b	$B8,$3F,$E1,$B6,$B9,$8D,$BE,$6F,$94,$F8,$58,$7C,$36,$DF,$F7,$26
	.dc.b	$CF,$F6,$D6,$E7,$FD,$6F,$CC,$FE,$C8,$7F,$65,$F3,$2D,$FF,$63,$8B
	.dc.b	$7F,$D8,$E6,$D7,$1A,$D1,$FD,$A6,$32,$FF,$A5,$72,$7D,$27,$E3,$FE
	.dc.b	$97,$E2,$B6,$7E,$1C,$E9,$B1,$F7,$7E,$93,$71,$FD,$D7,$DC,$FE,$F5
	.dc.b	$EE,$7F,$7B,$6C,$64,$FE,$F5,$19,$69,$FE,$14,$F3,$7F,$10,$7E,$27
	.dc.b	$CE,$B1,$92,$6B,$FE,$13,$8C,$BF,$E1,$59,$BA,$FE,$4E,$6D,$7F,$AC
	.dc.b	$FE,$95,$97,$F8,$3F,$E7,$7F,$59,$18,$9F,$F0,$74,$62,$64,$7F,$84
	.dc.b	$7B,$6E,$5F,$E2,$3F,$90,$FC,$F1,$F9,$FF,$91,$B5,$FE,$2E,$8B,$3F
	.dc.b	$E2,$99,$B9,$FF,$85,$FE,$03,$FC,$41,$EA,$7F,$88,$5B,$87,$F8,$7D
	.dc.b	$B7,$FE,$1F,$6D,$71,$9F,$F1,$25,$71,$69,$FA,$48,$7D,$17,$F9,$A2
	.dc.b	$30,$DB,$FF,$1F,$AF,$87,$FF,$1A,$B9,$AF,$F1,$F7,$D2,$FF,$90,$2F
	.dc.b	$FC,$7A,$BC,$AF,$F1,$DB,$9C,$14,$8D,$57,$8D,$7C,$77,$0F,$A5,$E2
	.dc.b	$D9,$F4,$7D,$2F,$EF,$70,$FF,$29,$BE,$A8,$CF,$85,$FE,$4A,$8C,$4E
	.dc.b	$BE,$2F,$06,$7E,$96,$DF,$C3,$FA,$57,$5F,$F9,$49,$B1,$9E,$BC,$4F
	.dc.b	$F9,$5B,$E9,$5F,$4B,$4C,$7C,$25,$BF,$9D,$E1,$7F,$2C,$E2,$FF,$32
	.dc.b	$2E,$CF,$F2,$FB,$99,$FF,$32,$B9,$7E,$B7,$D2,$BD,$BD,$C7,$F9,$9A
	.dc.b	$30,$F6,$3F,$CD,$3F,$4A,$EB,$BD,$78,$6C,$FE,$07,$97,$F8,$BE,$7F
	.dc.b	$E7,$76,$C7,$F9,$D9,$7C,$3F,$F9,$D1,$CF,$2F,$DA,$5F,$4B,$93,$5F
	.dc.b	$2B,$6F,$F9,$3F,$D2,$FE,$25,$BE,$CF,$F5,$78,$C6,$1F,$84,$9F,$E6
	.dc.b	$CA,$3F,$C5,$F1,$CB,$E1,$38,$1F,$D7,$FE,$91,$E7,$62,$7F,$9A,$C6
	.dc.b	$1B,$1F,$D9,$7E,$95,$E7,$F1,$8E,$5E,$64,$AF,$E7,$BE,$D3,$D6,$C6
	.dc.b	$9E,$53,$FA,$54,$66,$77,$3F,$DC,$7E,$97,$EC,$56,$C5,$6D,$C4,$6B
	.dc.b	$8F,$FA,$4C,$E5,$FE,$97,$19,$BF,$87,$39,$7E,$2F,$D2,$3E,$97,$FD
	.dc.b	$4F,$F3,$AF,$89,$FF,$01,$4E,$6B,$F7,$D3,$97,$FA,$CC,$67,$2F,$33
	.dc.b	$72,$FF,$5C,$57,$F8,$01,$FE,$BB,$19,$AF,$0E,$1D,$8F,$0F,$BF,$63
	.dc.b	$AF,$9D,$FE,$BA,$FA,$5F,$C3,$E4,$C3,$EF,$E3,$FE,$C5,$9C,$C6,$2F
	.dc.b	$DB,$7C,$7E,$33,$D6,$FB,$3E,$82,$4F,$AB,$1A,$7A,$A2,$E3,$E8,$8B
	.dc.b	$3F,$8C,$8F,$E1,$AF,$F7,$A6,$C7,$D7,$3F,$C0,$DF,$4B,$FE,$07,$FA
	.dc.b	$37,$D1,$F1,$F4,$3F,$C5,$D7,$C0,$FF,$00,$97,$0F,$CC,$E1,$C8,$5F
	.dc.b	$27,$FD,$C3,$5F,$E3,$77,$51,$97,$F9,$2C,$7F,$BB,$E3,$11,$AB,$F1
	.dc.b	$57,$95,$EF,$0C,$30,$F6,$9A,$FE,$B2,$2E,$1F,$02,$FA,$CF,$13,$CF
	.dc.b	$7D,$9A,$7B,$2B,$8E,$1E,$59,$F0,$C6,$30,$EF,$5E,$73,$C4,$72,$3C
	.dc.b	$45,$C3,$C5,$5C,$A3,$17,$D5,$F0,$1E,$E3,$E6,$16,$F0,$E0,$FE,$13
	.dc.b	$F4,$8F,$59,$F4,$BC,$63,$0D,$8C,$CF,$CA,$9C,$B7,$D8,$FE,$AE,$DC
	.dc.b	$B9,$46,$23,$EF,$CA,$31,$1E,$27,$8E,$9F,$77,$36,$BF,$C7,$A3,$0D
	.dc.b	$8E,$18,$7D,$8C,$3A,$3F,$6A,$38,$E1,$6B,$83,$FE,$5D,$AD,$BF,$3D
	.dc.b	$B7,$DA,$38,$1D,$70,$F2,$DF,$82,$3E,$B3,$D9,$B7,$D9,$5F,$49,$E1
	.dc.b	$7D,$67,$67,$5C,$3B,$BF,$C8,$E7,$A4,$7F,$9A,$1B,$AF,$27,$FA,$09
	.dc.b	$89,$FF,$3C,$2F,$8C,$FD,$D4,$E9,$B3,$AC,$6A,$FA,$5F,$F0,$5B,$FE
	.dc.b	$78,$6B,$FC,$D5,$B8,$7F,$9E,$9F,$B5,$1B,$DF,$AC,$B9,$9F,$25,$86
	.dc.b	$D6,$CF,$F4,$2F,$D2,$FF,$49,$7F,$D0,$FF,$4B,$FB,$DB,$FF,$43,$B7
	.dc.b	$4F,$EB,$F5,$BA,$FE,$90,$DA,$DE,$BE,$CB,$E1,$5F,$E9,$BA,$D5,$F2
	.dc.b	$F5,$7E,$0C,$1A,$71,$F1,$ED,$79,$39,$BA,$18,$7C,$33,$87,$47,$B1
	.dc.b	$F9,$BF,$4B,$FE,$43,$8F,$F5,$1B,$5F,$EC,$0D,$CF,$DF,$D7,$C2,$FE
	.dc.b	$AF,$D1,$FF,$91,$6F,$FD,$56,$D7,$8C,$6B,$DE,$F8,$1C,$12,$77,$1C
	.dc.b	$BC,$CE,$7E,$DC,$9E,$71,$5F,$EE,$5F,$4B,$F1,$9F,$F6,$13,$56,$2A
	.dc.b	$FC,$4C,$FF,$90,$A7,$EC,$F7,$7D,$13,$3E,$D7,$E1,$3E,$8B,$53,$FE
	.dc.b	$05,$D9,$18,$5C,$7F,$ED,$17,$35,$C7,$87,$F1,$BF,$46,$FB,$8B,$E7
	.dc.b	$CC,$BF,$CE,$D3,$96,$F6,$F0,$ED,$91,$83,$FC,$F9,$A3,$E1,$3C,$DE
	.dc.b	$DC,$9C,$C7,$E7,$37,$7C,$AD,$24,$FF,$A1,$63,$EA,$6F,$EA,$8F,$F3
	.dc.b	$CC,$7F,$BE,$1D,$72,$73,$3A,$7C,$2F,$C5,$F6,$24,$F9,$1F,$91,$EC
	.dc.b	$F8,$7E,$1B,$C5,$E9,$CA,$F0,$47,$97,$D3,$F3,$7C,$44,$9F,$24,$7D
	.dc.b	$1F,$A0,$FA,$5F,$41,$27,$E0,$C7,$B7,$C8,$91,$EB,$7E,$07,$C4,$7C
	.dc.b	$1E,$E9,$3E,$F7,$E2,$7C,$3F,$F0,$AF,$A2,$FF,$70,$CF,$FC,$00,$6B
	.dc.b	$FE,$3D,$6F,$1F,$C5,$84,$7E,$D9,$BF,$19,$CC,$8F,$3F,$2E,$BC,$E2
	.dc.b	$D7,$3A,$E4,$7F,$8F,$67,$C5,$E2,$F9,$4F,$37,$AF,$23,$24,$3C,$CF
	.dc.b	$F2,$4F,$A2,$7C,$3D,$24,$FB,$7E,$0E,$BF,$E5,$9F,$A2,$FF,$0E,$F8
	.dc.b	$FF,$E7,$46,$C0,$5F,$47,$F3,$D3,$BE,$2F,$E6,$74,$69,$8E,$3B,$5C
	.dc.b	$0E,$93,$B7,$01,$C3,$E5,$8E,$2F,$9E,$6E,$BF,$4E,$95,$E2,$F7,$FC
	.dc.b	$7F,$11,$30,$EB,$5F,$E8,$46,$CE,$5F,$85,$3C,$5F,$1F,$D1,$3A,$FF
	.dc.b	$A7,$B6,$33,$1F,$D4,$BE,$89,$F4,$BA,$8C,$70,$4A,$D9,$EE,$A7,$57
	.dc.b	$1E,$B9,$39,$8E,$86,$1F,$F4,$74,$A7,$31,$B3,$A3,$BD,$BF,$2D,$3D
	.dc.b	$A7,$C1,$E1,$79,$C8,$F8,$FC,$FA,$F6,$8B,$5C,$A7,$FB,$7F,$D1,$BC
	.dc.b	$4A,$4B,$FB,$3F,$46,$F3,$31,$1F,$EA,$58,$FD,$31,$FA,$7A,$E5,$1C
	.dc.b	$3A,$7F,$BD,$CF,$E8,$63,$84,$63,$AB,$2C,$8E,$07,$82,$30,$F0,$32
	.dc.b	$78,$5E,$BF,$1B,$23,$CE,$FF,$82,$BE,$8F,$F3,$F0,$BF,$F7,$93,$A2
	.dc.b	$1F,$EB,$F8,$DF,$D1,$3E,$8B,$FD,$83,$3F,$E7,$68,$C7,$C3,$FF,$96
	.dc.b	$DF,$D4,$FA,$16,$B7,$5F,$E8,$AD,$BC,$1D,$DF,$E6,$16,$C7,$FA,$01
	.dc.b	$AE,$F2,$FE,$24,$FF,$4F,$F9,$CF,$E7,$B5,$FA,$DE,$24,$62,$7F,$AC
	.dc.b	$2E,$2F,$E7,$D3,$23,$4D,$E3,$CD,$F5,$6D,$A7,$AD,$E1,$EF,$36,$F9
	.dc.b	$DF,$13,$C8,$F8,$5B,$1B,$23,$AA,$B7,$E8,$9B,$7E,$79,$B5,$D6,$B4
	.dc.b	$F1,$3A,$91,$C9,$F0,$A3,$8B,$EB,$BF,$3B,$6F,$D5,$3A,$BF,$6D,$F3
	.dc.b	$7C,$4D,$3C,$4F,$C9,$7C,$55,$E0,$8C,$36,$3E,$C4,$F5,$BF,$0D,$76
	.dc.b	$AF,$35,$75,$7C,$8D,$8F,$A2,$D9,$E2,$4A,$E9,$7D,$5B,$3B,$6D,$F8
	.dc.b	$CF,$84,$E2,$FC,$E7,$BA,$34,$F9,$4C,$CF,$A8,$F1,$57,$76,$9D,$CB
	.dc.b	$3C,$AF,$F0,$4D,$9E,$0D,$9E,$0D,$38,$5F,$96,$3E,$79,$E1,$8D,$5F
	.dc.b	$91,$F5,$C7,$82,$30,$7C,$88,$DC,$72,$76,$60,$F9,$E7,$D2,$31,$1F
	.dc.b	$06,$3C,$0F,$4B,$FA,$37,$F8,$89,$F0,$1D,$67,$C0,$6E,$3D,$ED,$8F
	.dc.b	$7B,$6B,$DF,$1C,$9F,$1C,$7D,$33,$47,$86,$FE,$24,$6A,$7C,$A7,$86
	.dc.b	$FC,$AF,$E5,$2B,$8B,$9A,$E3,$1E,$DA,$F8,$6E,$9E,$86,$27,$A5,$FA
	.dc.b	$9A,$7A,$8B,$1E,$A3,$E4,$DF,$E5,$63,$D9,$5D,$CF,$8A,$F0,$7D,$8E
	.dc.b	$A3,$84,$FE,$4B,$E8,$9F,$45,$FC,$45,$B5,$FC,$87,$E4,$1C,$19,$A9
	.dc.b	$3F,$C9,$6B,$77,$FE,$7B,$BD,$CE,$6B,$F8,$FC,$BE,$8B,$FC,$DD,$3F
	.dc.b	$CB,$7E,$89,$EF,$6A,$73,$5F,$E8,$59,$CD,$7F,$87,$27,$3A,$7F,$92
	.dc.b	$5B,$1F,$CE,$7D,$2E,$5E,$A7,$AF,$E8,$A4,$F9,$FE,$42,$E5,$FE,$91
	.dc.b	$D9,$58,$CB,$F9,$DC,$2B,$17,$FE,$93,$8D,$BF,$06,$BE,$44,$EA,$FF
	.dc.b	$91,$7D,$17,$F3,$9F,$5F,$F9,$D3,$6F,$F9,$CB,$67,$FA,$5B,$71,$FF
	.dc.b	$32,$7C,$F3,$E8,$7F,$A2,$7D,$13,$E8,$BE,$5C,$E7,$1F,$EA,$D3,$97
	.dc.b	$FC,$83,$19,$C7,$FA,$74,$7F,$94,$54,$67,$0F,$7F,$D3,$F4,$C7,$FA
	.dc.b	$87,$B7,$FD,$47,$2B,$8F,$F5,$06,$E3,$FD,$89,$B8,$7F,$61,$6C,$FF
	.dc.b	$6B,$6C,$FF,$45,$7F,$1D,$E7,$49,$FC,$57,$9D,$6B,$EA,$4D,$BF,$A7
	.dc.b	$36,$BD,$59,$BD,$5F,$ED,$1C,$CF,$A3,$F8,$B1,$86,$D7,$C4,$8F,$ED
	.dc.b	$9E,$E6,$33,$FC,$17,$C2,$7E,$87,$E1,$3F,$44,$DB,$FC,$F3,$67,$71
	.dc.b	$FA,$19,$DB,$4D,$96,$3F,$C0,$FE,$7F,$F7,$B3,$FB,$E7,$9F,$67,$FB
	.dc.b	$DC,$59,$FF,$06,$CD,$DB,$FD,$4D,$F4,$6C,$BF,$6D,$F0,$3F,$AC,$8C
	.dc.b	$63,$FB,$38,$C4,$C8,$FD,$B7,$65,$C7,$FC,$33,$F2,$3F,$E0,$D3,$FC
	.dc.b	$1D,$F2,$56,$BF,$C1,$B1,$67,$FC,$3F,$37,$2F,$F0,$F4,$DA,$FF,$14
	.dc.b	$4D,$AF,$F0,$E7,$7F,$D3,$9F,$F1,$53,$FB,$A8,$FF,$80,$0E,$5B,$E0
	.dc.b	$FF,$88,$7E,$8D,$F7,$94,$FF,$12,$57,$FA,$A6,$FF,$C6,$93,$A6,$E3
	.dc.b	$C4,$E5,$8D,$8F,$F1,$D7,$D2,$7F,$81,$09,$E5,$7F,$8B,$DC,$7C,$2D
	.dc.b	$F8,$5C,$36,$BF,$C8,$5F,$48,$ED,$FF,$1B,$2E,$DB,$18,$7F,$C7,$5F
	.dc.b	$48,$F8,$7B,$F0,$F8,$6E,$1F,$E4,$DA,$76,$7F,$90,$5B,$19,$EC,$C4
	.dc.b	$E7,$E9,$A4,$F1,$FA,$7B,$CF,$FB,$E3,$FC,$99,$F4,$9F,$C9,$1B,$D5
	.dc.b	$FE,$4D,$87,$65,$8F,$F2,$87,$94,$F6,$B5,$8E,$1A,$FF,$98,$E3,$FC
	.dc.b	$D7,$DB,$FE,$60,$5F,$25,$CB,$D6,$E6,$FF,$40,$FA,$37,$A9,$4C,$45
	.dc.b	$C3,$FD,$AF,$3F,$E9,$F6,$FC,$07,$F4,$8F,$A3,$79,$BC,$E3,$0C,$9F
	.dc.b	$F6,$FF,$D1,$FD,$57,$FD,$C3,$5C,$1B,$E4,$F3,$3F,$A9,$FB,$1F,$CF
	.dc.b	$63,$27,$FB,$A1,$C7,$1F,$CB,$3F,$B0,$FA,$9F,$D6,$BE,$8F,$E4,$D6
	.dc.b	$E3,$1C,$9D,$1B,$19,$3F,$DE,$7F,$42,$FA,$3F,$F3,$F7,$CE,$BE,$17
	.dc.b	$FB,$2C,$E6,$BF,$D2,$13,$97,$8D,$68,$E3,$E6,$D8,$C9,$33,$FC,$17
	.dc.b	$3A,$FF,$60,$A7,$FA,$16,$1D,$EB,$DC,$FD,$A8,$FF,$50,$7D,$1F,$F4
	.dc.b	$57,$7C,$BF,$E9,$58,$C5,$05,$71,$79,$CE,$FE,$07,$DA,$FB,$BE,$5A
	.dc.b	$61,$F1,$2A,$9E,$28,$B8,$F9,$E3,$B6,$1F,$C0,$5D,$57,$86,$75,$1F
	.dc.b	$EA,$BF,$23,$FC,$1F,$07,$FC,$23,$1E,$7F,$A3,$F4,$7D,$0B,$E5,$BF
	.dc.b	$B2,$5E,$F0,$E8,$63,$0E,$67,$2C,$18,$72,$BD,$57,$09,$FE,$DC,$68
	.dc.b	$61,$DD,$F4,$5F,$B6,$31,$1F,$DD,$6F,$A3,$AA,$AE,$4E,$77,$E0,$69
	.dc.b	$E0,$9D,$E1,$EB,$8C,$9C,$6F,$BC,$65,$F0,$C7,$C5,$37,$19,$AF,$5C
	.dc.b	$7A,$E7,$8E,$1E,$99,$CD,$F8,$23,$15,$E3,$39,$B8,$7C,$53,$E9,$AF
	.dc.b	$0C,$7C,$73,$B9,$4E,$A6,$0C,$1F,$91,$FA,$27,$A0,$FA,$3E,$31,$86
	.dc.b	$C6,$67,$D7,$39,$6F,$A5,$FE,$F3,$6E,$7C,$A3,$11,$F7,$25,$18,$8F
	.dc.b	$F1,$64,$62,$7E,$BE,$2D,$BF,$C6,$19,$7D,$E8,$C5,$FD,$CB,$E0,$FD
	.dc.b	$66,$CE,$EB,$8D,$7F,$8E,$4E,$35,$BC,$3D,$15,$F8,$AB,$E3,$5B,$BE
	.dc.b	$37,$FB,$78,$E3,$5B,$BE,$31,$EA,$6C,$EE,$B8,$BF,$A7,$6D,$F6,$8F
	.dc.b	$A5,$3C,$1D,$1F,$0A,$77,$19,$AF,$4E,$1F,$B8,$6E,$BE,$7B,$FD,$50
	.dc.b	$C3,$B3,$D0,$FE,$69,$2F,$89,$FD,$CD,$7C,$56,$CF,$A4,$E7,$F4,$7F
	.dc.b	$C0,$DB,$B6,$94,$65,$FE,$80,$FD,$E7,$67,$F3,$4F,$8D,$EB,$97,$F2
	.dc.b	$D7,$94,$F1,$8F,$E2,$7F,$45,$FE,$89,$AF,$5F,$D1,$FE,$C3,$2F,$F3
	.dc.b	$93,$63,$F9,$54,$ED,$FE,$A0,$DA,$FE,$3B,$3B,$93,$67,$55,$FE,$7D
	.dc.b	$5E,$77,$B5,$B5,$F3,$1A,$FF,$5D,$A7,$D1,$EE,$BE,$24,$7F,$7C,$9F
	.dc.b	$4F,$36,$5E,$53,$A8,$91,$83,$EC,$FD,$1F,$17,$FB,$53,$72,$FE,$6E
	.dc.b	$DC,$BE,$1F,$5F,$D3,$8C,$57,$F4,$DF,$A3,$7B,$5F,$5D,$B1,$DB,$1C
	.dc.b	$4F,$A3,$79,$3F,$8A,$AD,$BF,$DE,$D7,$99,$CF,$D4,$93,$B8,$AF,$F8
	.dc.b	$07,$E8,$FE,$6B,$FD,$7D,$AA,$BF,$C7,$CF,$E6,$23,$FC,$76,$FE,$03
	.dc.b	$BB,$E5,$9F,$8F,$8C,$C7,$F7,$6F,$A1,$BB,$FF,$5F,$0E,$37,$FE,$BC
	.dc.b	$D9,$4E,$B8,$3F,$27,$F4,$4F,$A3,$BE,$74,$BF,$EA,$96,$AD,$46,$1F
	.dc.b	$D5,$37,$B7,$99,$D3,$AB,$F3,$F9,$3D,$3B,$E4,$CB,$EB,$39,$7B,$EE
	.dc.b	$F9,$3F,$D3,$B6,$33,$18,$E5,$CF,$1C,$9E,$2F,$0F,$81,$F0,$FD,$1C
	.dc.b	$9E,$EF,$93,$E6,$FA,$FC,$2F,$6F,$52,$57,$8A,$3C,$3E,$97,$8B,$E1
	.dc.b	$E4,$F9,$A3,$DB,$F4,$1E,$EF,$A0,$93,$EC,$8F,$4F,$8D,$23,$DA,$FB
	.dc.b	$3E,$17,$DB,$DF,$27,$E0,$BF,$03,$E0,$BF,$87,$7D,$1F,$FB,$5E,$7F
	.dc.b	$8E,$B5,$FF,$27,$37,$AF,$FC,$36,$FE,$FC,$C7,$7F,$FB,$E5,$B3,$FE
	.dc.b	$31,$6A,$78,$7C,$5F,$80,$91,$E0,$F2,$FC,$7E,$B7,$93,$C5,$93,$E4
	.dc.b	$8C,$BE,$68,$CB,$E6,$8F,$E5,$1F,$44,$F9,$FA,$49,$F6,$FD,$1D,$7F
	.dc.b	$CD,$5F,$47,$FE,$36,$D7,$FC,$6E,$DD,$7F,$C9,$52,$C3,$A3,$FD,$0A
	.dc.b	$5C,$5F,$CD,$DB,$19,$8C,$71,$E5,$91,$93,$94,$61,$F1,$C5,$C3,$C7
	.dc.b	$3D,$7E,$2B,$B7,$C4,$CB,$F1,$46,$5F,$24,$65,$F6,$C7,$F4,$BF,$A3
	.dc.b	$75,$D2,$4F,$F5,$CF,$A2,$7D,$1E,$23,$FD,$1D,$1F,$01,$BC,$19,$CF
	.dc.b	$FC,$BB,$2E,$AF,$EB,$0B,$AE,$98,$89,$7D,$E3,$2F,$86,$32,$F8,$23
	.dc.b	$2F,$8A,$3A,$F8,$52,$3D,$0F,$83,$C2,$C4,$7A,$22,$E1,$E6,$9F,$ED
	.dc.b	$9F,$46,$F1,$29,$2F,$FB,$EF,$C4,$3C,$CC,$47,$FA,$BE,$3F,$AF,$1F
	.dc.b	$D7,$EB,$94,$70,$E9,$FE,$F1,$3F,$EA,$98,$C1,$D3,$AB,$33,$23,$27
	.dc.b	$58,$C3,$DE,$2E,$5F,$E8,$AA,$57,$6F,$89,$97,$E8,$7F,$C1,$1F,$47
	.dc.b	$F9,$98,$FF,$51,$57,$FB,$A5,$D1,$0F,$F6,$54,$6D,$FE,$65,$F4,$5F
	.dc.b	$E5,$BF,$3C,$D2,$47,$F9,$8B,$E8,$F3,$F1,$16,$4E,$7B,$BF,$9C,$D7
	.dc.b	$FA,$42,$B6,$FF,$A4,$B4,$7D,$1F,$F3,$3F,$03,$FA,$1B,$67,$FD,$8E
	.dc.b	$FF,$6E,$8C,$F2,$7F,$53,$7F,$D5,$D1,$BA,$FE,$AC,$D8,$CF,$26,$EF
	.dc.b	$FA,$AC,$E5,$B7,$9B,$F1,$B6,$7C,$13,$E2,$8C,$12,$FC,$F7,$67,$51
	.dc.b	$E6,$32,$57,$24,$BF,$3C,$F1,$7D,$35,$6B,$D2,$71,$8C,$B4,$CD,$73
	.dc.b	$7E,$1C,$F1,$8F,$7D,$5A,$F6,$D7,$19,$CB,$63,$31,$CC,$C9,$5C,$9E
	.dc.b	$27,$EA,$CF,$5B,$F9,$93,$AA,$F2,$BB,$23,$E6,$3E,$2D,$75,$9E,$24
	.dc.b	$AE,$97,$D5,$B5,$E7,$9F,$AB,$7D,$2F,$ED,$CE,$6B,$71,$D5,$B3,$B3
	.dc.b	$F2,$2E,$27,$C9,$3E,$C7,$B5,$F8,$5A,$75,$2C,$75,$6D,$75,$BF,$2C
	.dc.b	$74,$8C,$57,$E3,$6F,$C1,$7E,$83,$C0,$7C,$38,$F0,$36,$BC,$0D,$3C
	.dc.b	$05,$AF,$05,$74,$7E,$94,$7D,$A7,$C3,$3C,$9B,$1C,$9A,$72,$8F,$59
	.dc.b	$E9,$3E,$D9,$63,$D1,$1B,$7C,$51,$98,$F9,$AF,$28,$DB,$C9,$A4,$8F
	.dc.b	$39,$C1,$E6,$66,$FE,$5B,$6F,$E2,$CF,$A2,$BC,$15,$A8,$E6,$FE,$32
	.dc.b	$7E,$84,$F8,$E7,$37,$F1,$EB,$F2,$13,$96,$DF,$8A,$B7,$39,$BE,$73
	.dc.b	$9A,$DD,$7E,$37,$E8,$5F,$47,$FC,$35,$B8,$7F,$98,$3E,$3F,$F9,$3B
	.dc.b	$FE,$62,$8C,$E1,$FE,$6A,$83,$FE,$6C,$8C,$DF,$F9,$CD,$B3,$A8,$57
	.dc.b	$4A,$E4,$71,$BF,$E4,$4B,$E8,$FF,$CD,$17,$FE,$72,$F9,$C7,$BD,$A9
	.dc.b	$CD,$72,$C5,$5A,$BF,$F7,$1D,$6A,$BF,$C4,$E7,$5F,$59,$38,$8F,$5C
	.dc.b	$ED,$FF,$16,$B5,$F7,$7D,$1F,$C7,$3E,$47,$C7,$5A,$F8,$D1,$67,$C9
	.dc.b	$AA,$B1,$97,$C5,$AB,$7F,$16,$B4,$F2,$3F,$A7,$FD,$1F,$F3,$FF,$5F
	.dc.b	$FA,$03,$63,$27,$F5,$48,$CB,$4F,$A0,$7F,$91,$BE,$7D,$F4,$3F,$D2
	.dc.b	$FE,$8D,$F4,$5F,$42,$73,$5F,$EA,$09,$CB,$CB,$5F,$BF,$3A,$7E,$C8
	.dc.b	$71,$A7,$E0,$7E,$8F,$A6,$3F,$D5,$FD,$AF,$EB,$19,$5C,$7F,$B5,$B7
	.dc.b	$1F,$ED,$6D,$C3,$FB,$3B,$67,$FB,$5B,$67,$FB,$3B,$F9,$9F,$36,$4F
	.dc.b	$E5,$FC,$EB,$5F,$4A,$6D,$FE,$56,$6D,$7C,$5F,$53,$FB,$77,$2B,$E8
	.dc.b	$FE,$24,$61,$B3,$F0,$F1,$F8,$13,$FD,$CF,$DB,$CC,$7E,$03,$E7,$7F
	.dc.b	$59,$F0,$9F,$AD,$6D,$FE,$AD,$B3,$B8,$FD,$74,$ED,$A6,$CB,$1F,$E1
	.dc.b	$0F,$37,$FC,$04,$7F,$80,$FC,$EB,$3F,$E1,$68,$B3,$FE,$16,$9B,$AF
	.dc.b	$E8,$23,$F7,$3F,$46,$CB,$F8,$17,$C1,$7E,$CA,$31,$8F,$F0,$18,$C4
	.dc.b	$C8,$FE,$09,$DB,$6F,$FC,$47,$F1,$FF,$E1,$53,$FC,$2D,$F2,$36,$BF
	.dc.b	$C5,$31,$67,$FC,$53,$37,$2F,$F1,$2C,$DA,$FF,$14,$CD,$AF,$F7,$CF
	.dc.b	$C1,$FD,$59,$FF,$7A,$BF,$BF,$8F,$E0,$4E,$5B,$E0,$FF,$8A,$7E,$89
	.dc.b	$F4,$74,$FF,$16,$57,$F1,$AB,$FF,$7F,$CE,$9B,$8F,$13,$96,$36,$3F
	.dc.b	$49,$93,$E8,$FF,$C1,$A4,$F2,$7F,$C7,$0E,$3E,$16,$FC,$2E,$1B,$7C
	.dc.b	$BE,$8D,$D9,$FE,$3F,$5D,$96,$30,$FF,$90,$FE,$8D,$F0,$F7,$E1,$F0
	.dc.b	$DC,$3C,$B4,$EC,$FF,$25,$36,$33,$D9,$89,$CF,$D1,$C9,$E3,$F4,$77
	.dc.b	$A3,$F9,$77,$D1,$FF,$30,$E3,$FC,$50,$71,$5F,$CF,$D7,$67,$F9,$65
	.dc.b	$CD,$29,$E4,$BD,$AD,$61,$FE,$5F,$8C,$3F,$3A,$FF,$9B,$BF,$42,$1E
	.dc.b	$D6,$FE,$4B,$83,$67,$F6,$DF,$47,$FD,$41,$BE,$A7,$F4,$E8,$55,$8C
	.dc.b	$7C,$04,$AF,$DF,$7D,$13,$CC,$C6,$DF,$C8,$23,$13,$A9,$FE,$B3,$F4
	.dc.b	$7F,$6A,$30,$72,$F2,$7F,$AE,$D7,$FB,$4E,$3F,$85,$7B,$9F,$D0,$30
	.dc.b	$E9,$5F,$D8,$A3,$1C,$3F,$B7,$9F,$F4,$CF,$AD,$FD,$9B,$E8,$FE,$6E
	.dc.b	$1C,$A3,$1C,$9C,$5B,$19,$3F,$8C,$7D,$03,$E8,$FF,$D1,$9F,$46,$F8
	.dc.b	$5F,$EE,$33,$9A,$FC,$EC,$E5,$E3,$5A,$38,$F9,$96,$32,$4C,$FF,$BA
	.dc.b	$57,$ED,$74,$F5,$61,$D6,$7F,$A4,$BF,$4A,$BF,$57,$F4,$7F,$D2,$9F
	.dc.b	$F0,$6C,$A7,$4F,$F7,$F9,$D3,$FE,$12,$8C,$D7,$6E,$5F,$F5,$45,$7F
	.dc.b	$BA,$EF,$E0,$C7,$BB,$E5,$E4,$7C,$4C,$29,$E4,$8B,$7E,$88,$B3,$E4
	.dc.b	$87,$8D,$FE,$18,$6C,$7F,$45,$9F,$F1,$2F,$D1,$FF,$89,$BE,$89,$F2
	.dc.b	$BC,$70,$E0,$4B,$E1,$6B,$E1,$5B,$9C,$C6,$27,$FC,$59,$5F,$CE,$EF
	.dc.b	$AC,$7F,$87,$A7,$4F,$EB,$E7,$57,$FE,$1F,$9C,$E1,$E8,$A5,$F2,$7E
	.dc.b	$8C,$EA,$F8,$C7,$53,$77,$E5,$69,$9A,$E5,$7D,$5E,$91,$F1,$0D,$9E
	.dc.b	$5C,$3C,$EE,$63,$8D,$F8,$23,$C6,$70,$AE,$97,$EB,$8C,$4E,$DE,$C7
	.dc.b	$F8,$F4,$F9,$85,$29,$9C,$35,$F4,$4F,$39,$F4,$78,$6D,$7F,$91,$E7
	.dc.b	$53,$93,$8F,$A1,$FC,$76,$74,$D7,$FC,$A3,$28,$C4,$7E,$7A,$31,$97
	.dc.b	$23,$EC,$E4,$D7,$F4,$7A,$7D,$58,$C5,$FD,$EB,$E2,$FC,$C6,$CE,$EB
	.dc.b	$8D,$7E,$B0,$E3,$5B,$C3,$D1,$5F,$81,$BE,$35,$BB,$E3,$7F,$E8,$08
	.dc.b	$E3,$5B,$BE,$31,$EB,$6C,$EE,$B8,$BF,$A2,$6D,$F6,$8F,$F4,$5B,$5D
	.dc.b	$1F,$0A,$37,$3D,$23,$27,$F9,$4A,$BD,$78,$7F,$27,$6E,$BE,$38,$E5
	.dc.b	$3F,$E3,$99,$7C,$DF,$EE,$EB,$E2,$36,$7D,$07,$F8,$FF,$E8,$FF,$7E
	.dc.b	$DD,$F4,$A3,$2F,$F9,$91,$FB,$CE,$CF,$E5,$7F,$1D,$D7,$17,$22,$39
	.dc.b	$D7,$29,$FE,$81,$F4,$7F,$E8,$9B,$FF,$28,$7D,$17,$EB,$33,$FE,$4A
	.dc.b	$D8,$FF,$59,$CE,$DF,$F3,$AB,$6B,$FA,$54,$EE,$4D,$9D,$57,$F3,$05
	.dc.b	$E6,$7B,$9B,$5F,$E6,$16,$BF,$E8,$2A,$7D,$1E,$EB,$E2,$C7,$F6,$E9
	.dc.b	$F4,$F3,$65,$E5,$3A,$89,$18,$3F,$CD,$DF,$47,$C5,$FF,$47,$B7,$2F
	.dc.b	$EC,$8D,$CB,$E1,$F5,$7D,$48,$C5,$7F,$B5,$BE,$8D,$EC,$7D,$86,$C7
	.dc.b	$6C,$71,$3E,$8D,$E4,$FF,$9F,$EB,$6F,$FA,$81,$FF,$0C,$B9,$F4,$BA
	.dc.b	$1D,$7C,$CB,$1F,$E8,$7F,$A3,$F9,$AF,$FB,$A1,$AA,$BF,$C7,$6E,$FD
	.dc.b	$2F,$93,$F2,$D2,$30,$7F,$A4,$7E,$DB,$FD,$2C,$FF,$A4,$F9,$DF,$27
	.dc.b	$F9,$46,$C6,$FE,$17,$F2,$FF,$6D,$FE,$62,$8E,$1E,$4E,$CC,$FC,$0E
	.dc.b	$BB,$F9,$7C,$8C,$AF,$F5,$05,$FE,$71,$FF,$60,$AF,$6F,$F4,$7F,$6C
	.dc.b	$FA,$9A,$78,$23,$F4,$E7,$E2,$3A,$FF,$C3,$92,$F8,$8E,$26,$BB,$77
	.dc.b	$EE,$74,$EA,$99,$FF,$AC,$71,$4F,$EC,$BE,$DB,$FC,$39,$3F,$EB,$5F
	.dc.b	$B5,$FD,$EB,$7C,$6D,$7A,$1B,$FA,$9F,$DC,$D7,$F9,$2A,$B8,$7B,$B9
	.dc.b	$7F,$D7,$DF,$6D,$FE,$A5,$7A,$75,$F0,$3F,$80,$C6,$63,$FD,$C6,$FF
	.dc.b	$49,$7C,$5F,$01,$D8,$DF,$A3,$FC,$23,$ED,$9F,$2F,$FC,$2B,$ED,$9F
	.dc.b	$4F,$4D,$46,$FE,$C3,$5F,$0F,$32,$FF,$69,$60,$9F,$E2,$7E,$13,$E3
	.dc.b	$A9,$2B,$FD,$B1,$F5,$19,$7F,$DB,$53,$5F,$F7,$DC,$7E,$7B,$B3,$FC
	.dc.b	$A0,$DF,$3F,$FC,$67,$5B,$9F,$F2,$A3,$63,$FC,$6F,$3B,$A0,$70,$FE
	.dc.b	$5A,$FF,$94,$FD,$5B,$E5,$29,$27,$FD,$DD,$1F,$C5,$CD,$38,$8D,$3F
	.dc.b	$E8,$BB,$FF,$23,$CE,$FD,$FE,$BE,$7C,$AE,$71,$E0,$30,$F1,$3C,$F3
	.dc.b	$FC,$CD,$D9,$F1,$BE,$9B,$E6,$91,$F2,$FD,$5D,$79,$0F,$AB,$FF,$3C
	.dc.b	$6D,$FE,$79,$6E,$5F,$2A,$F8,$4E,$63,$FA,$5C,$EA,$BE,$4E,$CB,$CF
	.dc.b	$1C,$9F,$9B,$87,$58,$E3,$BE,$67,$C7,$87,$E3,$1D,$E3,$FE,$6C,$8C
	.dc.b	$46,$A7,$C3,$F2,$1E,$2E,$39,$3C,$0F,$B7,$5F,$ED,$67,$07,$BE,$76
	.dc.b	$F8,$27,$9B,$AF,$AB,$7D,$5F,$F6,$86,$C6,$63,$FB,$5F,$D5,$3E,$AF
	.dc.b	$51,$8D,$A5,$6C,$87,$D0,$AF,$8C,$F9,$1A,$FD,$4B,$EE,$EF,$E2,$FB
	.dc.b	$F9,$32,$38,$47,$88,$C3,$D6,$30,$F5,$FA,$6F,$39,$1F,$07,$93,$5F
	.dc.b	$24,$5E,$2F,$F5,$D6,$3F,$A2,$FA,$B7,$85,$FE,$0A,$6D,$FC,$C6,$BF
	.dc.b	$8C,$F0,$DF,$23,$BA,$FF,$08,$C7,$F7,$B3,$FB,$E5,$70,$C3,$FD,$90
	.dc.b	$DE,$5F,$F5,$5C,$EF,$E1,$39,$63,$F6,$A3,$FD,$84,$DE,$BE,$A7,$6E
	.dc.b	$CB,$19,$21,$E5,$7C,$7E,$24,$8E,$8F,$F8,$9B,$EA,$FF,$39,$1F,$EF
	.dc.b	$88,$D3,$ED,$75,$8F,$F8,$B3,$EA,$5F,$54,$F9,$1F,$F3,$D4,$61,$BD
	.dc.b	$FF,$CF,$B0,$FF,$63,$89,$D3,$81,$F5,$6F,$AC,$FF,$76,$47,$FB,$3A
	.dc.b	$31,$8F,$FA,$1D,$B1,$C5,$AF,$67,$F8,$E4,$6F,$C0,$E7,$3F,$C8,$27
	.dc.b	$2F,$18,$CA,$FB,$87,$C1,$61,$B7,$FE,$97,$6C,$FA,$5A,$FB,$97,$5F
	.dc.b	$12,$4F,$16,$CF,$F1,$49,$DC,$FE,$E1,$FF,$49,$36,$4E,$7D,$4B,$5C
	.dc.b	$7E,$7A,$F3,$E2,$3D,$53,$8A,$FF,$7B,$E0,$BC,$57,$F4,$28,$FF,$03
	.dc.b	$4F,$86,$73,$3C,$A3,$07,$84,$AF,$A6,$B7,$5D,$67,$53,$E5,$6D,$F9
	.dc.b	$67,$53,$C6,$79,$CE,$6B,$9D,$F4,$6C,$F2,$3D,$71,$97,$89,$EE,$8D
	.dc.b	$D7,$53,$C4,$5A,$F1,$57,$18,$CC,$78,$6B,$9C,$7C,$5A,$E7,$39,$8D
	.dc.b	$CF,$73,$6F,$C7,$32,$71,$73,$7F,$3A,$32,$6E,$7C,$73,$ED,$8F,$A9
	.dc.b	$7E,$59,$CC,$7A,$67,$E0,$9B,$7F,$31,$3B,$8C,$C6,$CE,$6E,$EB,$E8
	.dc.b	$D7,$AC,$EA,$58,$E2,$D9,$E2,$73,$3B,$C7,$DE,$7F,$41,$7F,$A7,$24
	.dc.b	$7E,$9D,$B1,$FA,$76,$BC,$23,$E0,$7C,$83,$E4,$DE,$FF,$AE,$BE,$51
	.dc.b	$EB,$9D,$46,$63,$E5,$57,$D3,$7F,$C1,$13,$A8,$F8,$17,$CE,$31,$3E
	.dc.b	$BC,$7E,$A6,$3E,$F9,$CB,$F0,$67,$A9,$E2,$7F,$63,$1A,$BF,$04,$6E
	.dc.b	$B9,$C7,$F8,$BA,$75,$5C,$63,$64,$51,$A7,$B9,$27,$9B,$AC,$7E,$2D
	.dc.b	$7B,$0F,$92,$E2,$3D,$98,$7B,$E3,$18,$79,$63,$18,$FC,$9C,$3A,$DF
	.dc.b	$BB,$1F,$14,$60,$95,$FC,$22,$32,$F1,$31,$87,$6C,$7D,$97,$E9,$9C
	.dc.b	$E1,$B9,$F3,$B6,$7E,$E5,$71,$7E,$31,$FC,$57,$E1,$BC,$1F,$52,$F9
	.dc.b	$DF,$E1,$F1,$86,$E1,$FC,$D6,$3F,$9C,$11,$FE,$71,$E2,$FF,$3B,$23
	.dc.b	$82,$5F,$F3,$38,$F5,$F0,$FF,$1C,$7F,$8E,$FB,$DF,$07,$F4,$17,$56
	.dc.b	$30,$FF,$19,$6E,$9C,$C7,$F3,$BA,$FE,$94,$47,$FA,$53,$FD,$53,$85
	.dc.b	$F5,$CF,$17,$F9,$AC,$61,$B1,$FE,$E6,$AF,$F2,$04,$E5,$B1,$C5,$B3
	.dc.b	$C3,$C6,$E2,$F4,$FA,$D7,$C1,$D3,$BF,$CF,$7F,$7B,$BE,$B3,$A6,$CE
	.dc.b	$DF,$0C,$EF,$83,$FC,$ED,$92,$F8,$47,$F9,$C3,$EB,$9F,$4D,$EC,$FA
	.dc.b	$97,$C9,$7F,$4E,$9C,$E5,$EC,$9C,$BF,$E5,$A8,$CD,$F3,$BF,$87,$EC
	.dc.b	$F1,$65,$6D,$CB,$4F,$5F,$B9,$F0,$4F,$85,$F3,$B6,$32,$4B,$F5,$E6
	.dc.b	$9D,$47,$BB,$85,$F5,$DF,$53,$C2,$FA,$AD,$C3,$E9,$B6,$B8,$37,$7F
	.dc.b	$EE,$DF,$5C,$FA,$DF,$BF,$DC,$FA,$DF,$EE,$33,$9C,$7E,$FC,$E5,$FF
	.dc.b	$3E,$C6,$71,$FF,$4E,$B7,$0F,$F0,$1E,$AF,$AE,$E7,$94,$C3,$8B,$F8
	.dc.b	$38,$C6,$1E,$DF,$A6,$FC,$7E,$7F,$90,$6D,$FE,$39,$B3,$B8,$FC,$8C
	.dc.b	$FE,$37,$DC,$91,$F8,$AF,$72,$D7,$CC,$9B,$9F,$9A,$6D,$7F,$B5,$29
	.dc.b	$A7,$FB,$46,$3F,$09,$F2,$5F,$13,$E0,$BF,$14,$D8,$F8,$71,$FA,$6F
	.dc.b	$AD,$C5,$FC,$48,$B8,$FF,$89,$B8,$D5,$8B,$FF,$5B,$F0,$3E,$5B,$F4
	.dc.b	$DF,$21,$FA,$76,$CF,$E9,$1B,$97,$E1,$9B,$8F,$5F,$3B,$F6,$E7,$EE
	.dc.b	$3D,$EB,$7F,$B6,$8B,$7F,$8B,$9B,$5F,$E3,$9F,$AE,$F2,$DF,$EC,$BC
	.dc.b	$DF,$F1,$9C,$62,$FF,$2B,$18,$D2,$57,$EE,$3B,$2D,$7F,$B8,$2F,$86
	.dc.b	$1D,$6F,$F8,$15,$6A,$3F,$C9,$1F,$87,$FF,$2D,$C3,$FD,$92,$DC,$3F
	.dc.b	$7B,$DC,$F6,$BF,$C8,$B3,$9B,$FF,$28,$7C,$83,$93,$FC,$92,$DC,$7F
	.dc.b	$8B,$E0,$8C,$CF,$F9,$3B,$ED,$DF,$2F,$4F,$E4,$3F,$61,$FE,$64,$D3
	.dc.b	$FD,$F3,$3A,$C7,$FD,$BA,$B1,$E4,$7F,$28,$FB,$0D,$57,$08,$ED,$EA
	.dc.b	$CB,$FE,$53,$5F,$E6,$EA,$EB,$3F,$E6,$0F,$A8,$73,$D3,$F9,$9F,$D8
	.dc.b	$7E,$AF,$0F,$F7,$6C,$E9,$BF,$3D,$C4,$FE,$73,$F6,$0F,$9E,$C4,$FF
	.dc.b	$BD,$3E,$D9,$F3,$5F,$E6,$E6,$C7,$F3,$CF,$09,$E1,$E6,$B5,$F3,$52
	.dc.b	$B5,$F0,$77,$81,$D1,$FE,$81,$C9,$1D,$4F,$E9,$7F,$63,$F9,$DB,$F1
	.dc.b	$47,$F9,$F9,$7C,$CD,$37,$EF,$D8,$E7,$87,$FA,$37,$EF,$5E,$97,$FA
	.dc.b	$19,$B3,$FD,$53,$D1,$FC,$06,$3E,$48,$FD,$E4,$EA,$30,$79,$BE,$C9
	.dc.b	$EB,$FF,$A4,$1B,$5E,$03,$F8,$22,$FA,$5D,$4E,$FE,$46,$56,$EF,$E0
	.dc.b	$FD,$8B,$D2,$CC,$78,$04,$66,$78,$46,$27,$FD,$43,$F6,$5F,$4A,$73
	.dc.b	$5B,$AE,$4F,$E5,$76,$FF,$58,$36,$7F,$8D,$7C,$67,$F4,$97,$8F,$9D
	.dc.b	$FE,$AC,$8C,$5F,$8D,$B7,$FE,$B9,$FB,$2F,$F1,$14,$FF,$BB,$7D,$AF
	.dc.b	$F5,$9C,$62,$B7,$87,$F2,$78,$C4,$FF,$AE,$3E,$23,$F1,$77,$FD,$D3
	.dc.b	$EB,$BF,$71,$7E,$E9,$CE,$3F,$EC,$89,$CD,$7C,$59,$CD,$7F,$33,$6C
	.dc.b	$7E,$3A,$7F,$BA,$BB,$EB,$F0,$12,$BF,$C2,$B3,$AB,$FF,$6A,$5F,$2B
	.dc.b	$FF,$3F,$70,$FF,$87,$BE,$CB,$FA,$C6,$1D,$E3,$1E,$DF,$AD,$91,$FE
	.dc.b	$20,$97,$CB,$62,$FD,$1F,$61,$F0,$4F,$84,$DA,$F8,$11,$63,$DC,$5C
	.dc.b	$BF,$DD,$07,$ED,$E7,$F0,$6D,$8F,$C4,$1F,$AC,$F1,$7F,$5B,$F6,$4F
	.dc.b	$97,$FE,$B1,$5B,$39,$F8,$BF,$EE,$A7,$18,$7F,$5D,$C3,$90,$BE,$4F
	.dc.b	$EB,$AB,$FD,$E2,$EA,$32,$FE,$FA,$3F,$E0,$0C,$62,$35,$7E,$29,$CB
	.dc.b	$DE,$18,$7A,$23,$57,$67,$C9,$35,$ED,$15,$EA,$79,$5C,$3D,$0F,$54
	.dc.b	$F4,$69,$D0,$A7,$42,$9D,$0C,$5F,$AD,$E8,$FC,$0C,$3E,$93,$9B,$F3
	.dc.b	$46,$2B,$DD,$1E,$FA,$FF,$05,$35,$F9,$E6,$5D,$7D,$97,$F2,$0F,$8B
	.dc.b	$BE,$0F,$FA,$CA,$BE,$44,$E5,$B1,$97,$ED,$43,$EA,$FF,$98,$C7,$06
	.dc.b	$BA,$6B,$FE,$1D,$8E,$18,$F8,$A7,$8E,$5F,$7B,$26,$BF,$E2,$38,$C3
	.dc.b	$63,$94,$F0,$6D,$7D,$97,$8E,$16,$B8,$3F,$D1,$30,$F0,$BF,$51,$B8
	.dc.b	$76,$8E,$07,$E0,$E3,$35,$DA,$3E,$DB,$D9,$B5,$D9,$4F,$07,$C1,$87
	.dc.b	$9D,$FE,$AB,$5C,$6F,$C1,$1E,$39,$D3,$FA,$DB,$FD,$6E,$9F,$E4,$5B
	.dc.b	$F0,$62,$8C,$4E,$AB,$BC,$FA,$9F,$E8,$15,$FE,$11,$7F,$9F,$B5,$CC
	.dc.b	$F1,$5F,$65,$FD,$E5,$FE,$8A,$D7,$FA,$53,$6B,$C5,$1F,$DC,$23,$F0
	.dc.b	$8F,$0D,$BC,$92,$E4,$7C,$CE,$1B,$5B,$3F,$A9,$FD,$97,$FA,$45,$FE
	.dc.b	$AB,$F6,$3F,$A3,$D3,$FC,$CC,$D8,$FE,$FF,$3B,$7F,$AF,$37,$0F,$2C
	.dc.b	$27,$83,$74,$F8,$15,$BB,$FF,$39,$3F,$E2,$29,$F1,$FA,$1A,$3E,$35
	.dc.b	$BE,$CB,$13,$E6,$E7,$C9,$87,$C7,$58,$74,$74,$3F,$B8,$FD,$97,$F7
	.dc.b	$D6,$CF,$D3,$8C,$47,$F7,$9B,$FF,$64,$BA,$3F,$BC,$FA,$2F,$90,$FB
	.dc.b	$4D,$8D,$1F,$E0,$17,$0F,$3D,$FD,$67,$09,$E3,$E4,$F2,$F9,$69,$5D
	.dc.b	$A2,$BF,$E0,$BF,$B2,$F9,$0F,$F8,$C1,$AB,$05,$7E,$36,$7F,$D1,$D3
	.dc.b	$F7,$3B,$3E,$59,$9F,$85,$FC,$97,$D9,$7F,$9A,$E7,$FC,$2B,$CC,$F4
	.dc.b	$BF,$2C,$D8,$FF,$0C,$CE,$DF,$F0,$F7,$D9,$7F,$9B,$1E,$3F,$0D,$C0
	.dc.b	$FC,$F4,$66,$37,$D9,$D3,$E4,$E4,$7F,$89,$27,$F3,$6F,$4F,$6F,$F4
	.dc.b	$DF,$64,$F9,$6A,$74,$8F,$D5,$1F,$89,$9F,$F1,$84,$BE,$2B,$89,$AE
	.dc.b	$9D,$FD,$97,$4E,$69,$7F,$E3,$4C,$13,$FB,$5F,$B2,$FF,$34,$CF,$F8
	.dc.b	$E3,$EC,$BF,$75,$E0,$FE,$79,$BE,$07,$FB,$2E,$B7,$5F,$E4,$15,$EE
	.dc.b	$FE,$FE,$9F,$4B,$4E,$D1,$FC,$0F,$EC,$BF,$D8,$F7,$DF,$A7,$81,$9F
	.dc.b	$8E,$D7,$97,$CF,$E7,$65,$7F,$93,$2F,$FC,$DA,$FF,$93,$BE,$C7,$FC
	.dc.b	$D2,$FF,$94,$3E,$C9,$F5,$FF,$C4,$5B,$1B,$F8,$1F,$E2,$94,$FB,$2D
	.dc.b	$7C,$3C,$AE,$91,$FC,$67,$C1,$FF,$52,$47,$F9,$71,$AF,$F9,$85,$BC
	.dc.b	$9F,$C1,$3E,$8E,$F2,$FF,$BC,$66,$D7,$FA,$2A,$7F,$4F,$5B,$E8,$ED
	.dc.b	$F1,$92,$BF,$CD,$35,$AC,$3F,$DF,$4D,$8F,$F4,$9E,$1D,$17,$D9,$3E
	.dc.b	$67,$F9,$83,$63,$FC,$E6,$D7,$FC,$F0,$FF,$18,$31,$97,$F1,$C6,$E1
	.dc.b	$FE,$A0,$9E,$B5,$D3,$D1,$EB,$F2,$BD,$BC,$D9,$5D,$A3,$2F,$82,$32
	.dc.b	$F8,$23,$E6,$7D,$24,$8F,$98,$FA,$3E,$CB,$11,$FE,$8D,$FB,$2F,$F2
	.dc.b	$5E,$5E,$36,$C7,$C2,$AD,$E9,$FC,$D2,$73,$26,$EA,$AF,$EA,$91,$F8
	.dc.b	$49,$ED,$CD,$83,$3B,$D8,$FF,$4D,$CE,$F0,$ED,$5A,$9C,$FC,$7F,$4E
	.dc.b	$49,$18,$21,$EB,$72,$FB,$63,$FB,$37,$D8,$3E,$C6,$92,$7F,$D5,$7F
	.dc.b	$62,$FB,$2C,$47,$FA,$C6,$3D,$8E,$B5,$CE,$7F,$D5,$B0,$E9,$39,$21
	.dc.b	$FE,$B9,$72,$F3,$F8,$9F,$EF,$17,$FC,$36,$3C,$3E,$4E,$26,$46,$08
	.dc.b	$71,$3C,$FF,$53,$C8,$8F,$AB,$AA,$BE,$98,$B8,$FF,$6C,$C3,$F3,$D2
	.dc.b	$FB,$2F,$F4,$AB,$FE,$CC,$6B,$FE,$D1,$6C,$FF,$74,$F1,$9F,$1F,$FE
	.dc.b	$11,$9C,$F3,$7E,$6E,$7F,$BB,$9F,$DE,$23,$37,$FE,$FE,$6B,$FC,$6B
	.dc.b	$AB,$FC,$35,$3B,$F1,$AF,$54,$AF,$F6,$FC,$A3,$7D,$79,$9F,$F7,$34
	.dc.b	$E5,$DC,$E9,$E3,$E2,$76,$23,$FE,$EA,$FB,$2F,$CD,$D2,$73,$3D,$1F
	.dc.b	$F1,$74,$FE,$83,$EB,$DD,$1F,$E7,$76,$D7,$F9,$ED,$AF,$FA,$6D,$AF
	.dc.b	$FA,$AF,$EC,$5F,$64,$F9,$0F,$F4,$14,$61,$BE,$07,$F8,$FE,$BF,$83
	.dc.b	$4E,$5E,$9E,$8F,$F8,$99,$FE,$15,$82,$31,$01,$19,$E2,$F3,$4E,$A7
	.dc.b	$3E,$0D,$A3,$49,$5F,$E9,$7F,$B0,$75,$7F,$92,$B9,$FF,$C9,$7F,$14
	.dc.b	$FA,$BF,$F0,$55,$F8,$67,$FC,$77,$7C,$23,$07,$96,$B8,$7C,$07,$38
	.dc.b	$FE,$49,$3B,$CD,$7D,$B0,$EB,$18,$AD,$E3,$C6,$18,$F4,$9E,$F1,$F0
	.dc.b	$F0,$7A,$BB,$84,$3E,$C7,$0E,$AB,$8C,$EA,$73,$5A,$6C,$EA,$B9,$36
	.dc.b	$74,$D3,$53,$C2,$BD,$47,$95,$C6,$5F,$21,$B3,$F1,$1B,$3B,$8E,$8F
	.dc.b	$23,$CC,$E2,$79,$34,$E4,$58,$E4,$71,$BF,$80,$DA,$EE,$F4,$35,$7E
	.dc.b	$C6,$D7,$B0,$E8,$EE,$FF,$03,$3B,$6C,$EE,$78,$9D,$DE,$37,$F2,$A3
	.dc.b	$33,$EA,$6C,$FA,$9A,$7A,$A3,$AB,$CC,$C9,$E3,$8F,$C6,$D7,$8D,$A7
	.dc.b	$8E,$78,$9D,$CA,$F7,$6C,$F7,$8E,$2F,$78,$CB,$DC,$FC,$84,$F7,$6C
	.dc.b	$F7,$38,$CF,$76,$9D,$CB,$1D,$CE,$27,$77,$2F,$73,$F3,$73,$DD,$B1
	.dc.b	$DD,$E3,$3D,$DB,$1D,$DA,$77,$38,$9D,$CC,$9D,$EB,$F0,$F3,$CE,$36
	.dc.b	$D8,$CC,$F5,$9C,$B6,$33,$1D,$DA,$F5,$AF,$54,$60,$FC,$25,$7C,$E6
	.dc.b	$D7,$C2,$6D,$79,$9A,$F8,$0E,$D5,$D0,$93,$F1,$63,$47,$85,$C4,$F8
	.dc.b	$6F,$F4,$6F,$43,$11,$F9,$88,$C9,$1C,$12,$7F,$27,$3C,$6B,$DB,$F0
	.dc.b	$3F,$C6,$71,$FE,$55,$F3,$5F,$C4,$A3,$1A,$23,$18,$FF,$1C,$AD,$BF
	.dc.b	$C4,$6B,$73,$FC,$6B,$C4,$E3,$34,$E1,$67,$19,$7F,$94,$CF,$E4,$D3
	.dc.b	$FD,$17,$2F,$F1,$E4,$FF,$41,$8C,$7B,$5E,$7F,$B1,$FE,$D1,$3F,$C8
	.dc.b	$63,$1D,$5F,$CA,$9B,$19,$3D,$13,$C3,$0F,$E4,$3B,$BE,$A9,$E2,$7F
	.dc.b	$3B,$8C,$36,$3F,$DD,$17,$FC,$F2,$72,$D8,$E2,$D9,$E1,$E2,$F1,$7A
	.dc.b	$7D,$53,$DF,$A7,$7F,$9B,$FE,$E9,$7D,$67,$4D,$9D,$BE,$19,$DE,$5E
	.dc.b	$AA,$DB,$FC,$CF,$EB,$DF,$55,$EB,$FB,$07,$D8,$FF,$51,$9C,$E3,$EB
	.dc.b	$9C,$BF,$D4,$23,$37,$CE,$FE,$17,$AF,$C5,$95,$B7,$2D,$3D,$5E,$D7
	.dc.b	$C0,$3E,$0F,$CD,$D8,$C9,$2F,$D5,$19,$7E,$AF,$03,$EC,$7E,$97,$C6
	.dc.b	$FD,$36,$DF,$D1,$6D,$70,$6F,$1F,$E3,$E3,$FC,$29,$F5,$EF,$9B,$FB
	.dc.b	$DD,$6F,$B0,$FE,$E9,$39,$C7,$EF,$4E,$5F,$EE,$71,$9B,$FE,$C0,$DC
	.dc.b	$3F,$C2,$9C,$6F,$B1,$E7,$94,$C3,$8B,$F8,$28,$C6,$1E,$EF,$A3,$FC
	.dc.b	$76,$7F,$8F,$6D,$FE,$35,$B3,$B8,$FC,$84,$FE,$33,$CD,$91,$F8,$9F
	.dc.b	$6E,$D7,$E1,$66,$E7,$E6,$9B,$5F,$EC,$3D,$3F,$D9,$CF,$E0,$FE,$BB
	.dc.b	$E1,$FC,$07,$E2,$5B,$1F,$0A,$3F,$49,$F0,$F8,$CB,$E1,$C5,$C7,$FC
	.dc.b	$6D,$C6,$AC,$5F,$F8,$D7,$81,$F2,$5F,$A4,$F3,$BF,$4A,$D9,$FD,$13
	.dc.b	$72,$FC,$F3,$71,$EB,$E5,$FE,$D4,$FD,$B7,$B9,$6F,$F6,$91,$6F,$F1
	.dc.b	$73,$6B,$F0,$90,$FB,$1F,$35,$FE,$C7,$E4,$7F,$C6,$D1,$89,$FD,$74
	.dc.b	$63,$49,$5F,$B7,$EC,$B5,$FE,$23,$BE,$18,$76,$6D,$7C,$1A,$DC,$F8
	.dc.b	$BE,$C7,$F9,$74,$FF,$90,$1B,$7F,$C5,$BB,$BF,$D2,$53,$9F,$2F,$FC
	.dc.b	$95,$F6,$2E,$4A,$7F,$19,$A4,$62,$BF,$C9,$7F,$5E,$FB,$1F,$F2,$93
	.dc.b	$63,$F8,$E7,$D8,$FF,$98,$74,$F8,$B3,$9C,$FF,$C7,$CB,$1E,$27,$F2
	.dc.b	$7F,$B1,$D5,$70,$8E,$BE,$9C,$C7,$F9,$45,$3E,$5F,$18,$FF,$92,$7E
	.dc.b	$97,$FD,$0C,$DE,$8F,$E6,$5F,$63,$FA,$BC,$3F,$CC,$75,$AC,$FE,$CB
	.dc.b	$FE,$74,$6F,$B1,$FC,$E7,$EC,$5F,$39,$FE,$6F,$AD,$5F,$FA,$57,$27
	.dc.b	$D8,$BE,$AF,$FC,$E8,$D9,$FE,$7D,$E1,$FF,$AC,$1F,$F3,$9C,$66,$BF
	.dc.b	$CC,$0B,$E6,$38,$9F,$D2,$69,$F0,$38,$BE,$37,$D8,$FE,$9D,$F6,$3F
	.dc.b	$A2,$8C,$9F,$E8,$01,$F3,$3F,$E8,$76,$9F,$D4,$63,$F0,$BB,$FE,$6A
	.dc.b	$33,$5F,$E7,$3F,$B1,$7A,$5F,$E8,$F6,$CF,$F5,$9F,$47,$F1,$0F,$FA
	.dc.b	$3A,$33,$5F,$E7,$BF,$B1,$7A,$FF,$E9,$76,$D7,$FA,$51,$7D,$2E,$67
	.dc.b	$5F,$21,$2B,$FD,$33,$F6,$0F,$4B,$18,$FF,$A1,$E3,$13,$A9,$FF,$52
	.dc.b	$7D,$8F,$CF,$AD,$47,$FA,$9E,$BF,$29,$5F,$65,$B3,$FE,$CF,$F8,$A7
	.dc.b	$B3,$BF,$3B,$FD,$5D,$18,$BF,$F5,$5B,$6B,$FD,$31,$F6,$2F,$3B,$78
	.dc.b	$7F,$AD,$63,$17,$C2,$31,$3A,$8F,$F5,$AF,$D8,$FE,$0F,$93,$FD,$A9
	.dc.b	$F5,$9F,$B6,$C7,$FD,$49,$39,$AF,$F6,$A4,$E6,$BF,$88,$CE,$6B,$FD
	.dc.b	$D0,$D8,$4F,$F7,$49,$E7,$D9,$FE,$C9,$AE,$10,$78,$4E,$EB,$AD,$6F
	.dc.b	$8B,$FC,$39,$F6,$3F,$D5,$F1,$FD,$16,$1F,$69,$FF,$0E,$46,$71,$F1
	.dc.b	$D7,$19,$CF,$CA,$59,$F6,$FD,$7F,$8C,$F2,$36,$7E,$2C,$58,$F8,$65
	.dc.b	$CB,$F9,$89,$FB,$69,$FF,$6F,$B6,$3F,$D9,$80,$78,$17,$EC,$1F,$2F
	.dc.b	$FB,$BA,$E0,$74,$F0,$FF,$DD,$CE,$30,$FF,$27,$61,$C8,$5F,$27,$B4
	.dc.b	$ED,$FF,$2A,$61,$DF,$0E,$D8,$7F,$82,$EF,$85,$7F,$0B,$AD,$3C,$A7
	.dc.b	$55,$E0,$9C,$CA,$FD,$D7,$CA,$3D,$D3,$A9,$CD,$71,$9F,$3B,$E2,$C3
	.dc.b	$9C,$70,$32,$78,$5C,$BE,$FB,$F3,$D7,$48,$F6,$9C,$0F,$26,$1E,$87
	.dc.b	$33,$C2,$31,$1E,$F3,$84,$F3,$70,$7C,$E3,$E7,$17,$0C,$3E,$73,$FD
	.dc.b	$01,$F6,$3F,$C7,$BE,$3F,$83,$63,$EB,$B5,$DF,$82,$9D,$F9,$5F,$2E
	.dc.b	$76,$D7,$AB,$5F,$E8,$15,$FD,$47,$0E,$27,$D4,$D1,$AF,$F4,$5D,$3E
	.dc.b	$EC,$62,$FE,$CD,$F1,$7E,$43,$67,$75,$C6,$BF,$A7,$1C,$6B,$78,$79
	.dc.b	$EB,$F1,$37,$C6,$B7,$7C,$6F,$F4,$11,$C6,$B7,$7C,$63,$D6,$D9,$DD
	.dc.b	$71,$7F,$4A,$DB,$ED,$1F,$AB,$6B,$A3,$E1,$46,$E7,$A4,$64,$FC,$C6
	.dc.b	$BE,$67,$F4,$ED,$72,$75,$8E,$37,$F0,$71,$FF,$17,$BF,$E1,$49,$E2
	.dc.b	$7B,$BD,$1F,$DD,$5F,$3A,$FF,$1A,$FD,$8F,$F7,$AA,$D8,$70,$F8,$0C
	.dc.b	$6F,$C9,$FE,$11,$1F,$E0,$F7,$8B,$C2,$7F,$9B,$FC,$CE,$DF,$E5,$EF
	.dc.b	$19,$E6,$78,$3E,$C7,$FA,$6B,$E2,$FA,$FF,$D5,$E3,$FE,$0B,$6C,$7F
	.dc.b	$17,$9D,$BE,$56,$DF,$F9,$42,$77,$26,$CE,$AB,$FC,$2C,$BC,$FF,$23
	.dc.b	$6B,$C9,$1E,$7A,$7D,$8E,$A3,$1C,$D2,$B6,$43,$E4,$57,$F7,$A7,$D9
	.dc.b	$1C,$5F,$9F,$F6,$39,$7E,$73,$72,$FF,$34,$B7,$2F,$85,$F1,$9F,$26
	.dc.b	$31,$5F,$CE,$7D,$17,$C7,$FC,$D6,$C7,$6C,$71,$3E,$BD,$A4,$FE,$02
	.dc.b	$B6,$FD,$F5,$E5,$F3,$F9,$59,$3B,$8A,$FE,$17,$EC,$7E,$63,$FD,$2D
	.dc.b	$AA,$95,$F8,$19,$FF,$4A,$CF,$DF,$ED,$F9,$C6,$7E,$0F,$F0,$DF,$63
	.dc.b	$FE,$71,$9F,$CA,$F3,$7F,$90,$1B,$E9,$6B,$DE,$FC,$67,$D8,$FF,$9C
	.dc.b	$67,$7E,$3F,$E3,$EF,$FC,$25,$1F,$9E,$9D,$3F,$9D,$7F,$34,$FE,$7E
	.dc.b	$1F,$25,$C4,$EB,$DD,$DB,$DC,$FC,$D7,$D8,$BE,$5A,$99,$C3,$F5,$73
	.dc.b	$F8,$B7,$F5,$52,$F8,$8E,$27,$E8,$E7,$5D,$7F,$EA,$85,$F6,$3D,$BA
	.dc.b	$26,$1D,$23,$F5,$5F,$63,$FE,$6D,$8F,$DC,$FD,$8F,$F0,$17,$F7,$0D
	.dc.b	$C3,$FD,$67,$3B,$9F,$DE,$3F,$E6,$D7,$8F,$D1,$FE,$D6,$8F,$6F,$7F
	.dc.b	$F7,$1F,$63,$FE,$CF,$AE,$FD,$1B,$33,$F1,$BA,$F3,$B9,$7C,$E4,$AF
	.dc.b	$E1,$B7,$FE,$70,$7F,$87,$FD,$87,$F9,$A9,$FE,$23,$F6,$2F,$AD,$FE
	.dc.b	$0C,$D8,$DF,$C1,$FF,$09,$A7,$D8,$EB,$E1,$A5,$74,$8F,$E1,$DD,$FF
	.dc.b	$EA,$B8,$FE,$3A,$D7,$F8,$E3,$78,$FF,$DA,$FF,$49,$7C,$BF,$F4,$D4
	.dc.b	$DA,$FF,$13,$4F,$E9,$AB,$7C,$FD,$BE,$32,$57,$F2,$AA,$D6,$1F,$EA
	.dc.b	$36,$C7,$F8,$BF,$0E,$8B,$EC,$5F,$2F,$FC,$95,$B1,$FC,$D1,$AF,$F3
	.dc.b	$37,$F8,$E9,$8C,$BF,$DD,$ED,$C3,$FD,$64,$D9,$FF,$1F,$3C,$3E,$AB
	.dc.b	$B7,$99,$27,$B4,$78,$4C,$3D,$E3,$0F,$CA,$FA,$4F,$9A,$47,$CD,$F6
	.dc.b	$35,$FE,$91,$F6,$3F,$E5,$2C,$7F,$D8,$8D,$8F,$85,$5B,$C9,$5A,$3F
	.dc.b	$A3,$56,$EB,$FD,$7F,$58,$E5,$91,$FD,$62,$7F,$0C,$F8,$36,$CB,$E2
	.dc.b	$9B,$87,$68,$FF,$30,$52,$BB,$FC,$EF,$83,$8E,$57,$58,$CB,$DE,$32
	.dc.b	$FB,$A3,$FA,$E7,$D7,$BE,$C6,$92,$7F,$B6,$FD,$83,$EC,$71,$1F,$DB
	.dc.b	$61,$AF,$F6,$56,$C7,$F7,$4B,$E8,$DB,$D5,$48,$C1,$DB,$E2,$F9,$78
	.dc.b	$F2,$30,$43,$C0,$E5,$F1,$47,$87,$EA,$64,$7A,$1F,$27,$4E,$23,$D1
	.dc.b	$17,$83,$E6,$43,$3F,$CF,$AF,$B1,$78,$7F,$E0,$16,$C7,$F8,$41,$AF
	.dc.b	$F8,$3F,$C5,$7C,$8E,$EB,$FC,$11,$1F,$DF,$0F,$EF,$B5,$C3,$0F,$0B
	.dc.b	$79,$BF,$D3,$98,$7F,$1A,$F8,$2E,$27,$37,$FC,$3F,$6E,$AE,$A4,$8C
	.dc.b	$F5,$57,$C9,$17,$1F,$7D,$2B,$CB,$E2,$4B,$EA,$FF,$87,$FE,$C7,$F3
	.dc.b	$54,$8D,$4F,$BD,$D6,$5F,$E2,$5F,$AF,$FE,$A3,$3F,$E7,$F6,$BF,$EA
	.dc.b	$E6,$BE,$E0,$E1,$F8,$2F,$F8,$99,$D3,$5C,$FC,$77,$FA,$22,$31,$1F
	.dc.b	$EC,$3F,$B0,$7D,$8B,$E4,$BF,$D2,$11,$86,$F8,$9F,$D3,$EB,$E7,$CE
	.dc.b	$5F,$37,$D9,$7E,$11,$F1,$B5,$FF,$68,$60,$8C,$47,$A9,$BF,$63,$FD
	.dc.b	$9F,$3F,$F2,$AB,$67,$FC,$C7,$F1,$DF,$EF,$5F,$07,$F7,$9C,$5F,$92
	.dc.b	$AD,$37,$B3,$38,$7E,$12,$72,$FF,$B8,$6F,$93,$DA,$77,$87,$FB,$96
	.dc.b	$9D,$98,$84,$62,$BC,$BE,$BE,$63,$13,$FE,$69,$6E,$3E,$5B,$D5,$6D
	.dc.b	$97,$EE,$72,$D2,$5F,$38,$CB,$63,$35,$D5,$F1,$98,$3E,$04,$F6,$9F
	.dc.b	$25,$5A,$E1,$1C,$CC,$D7,$17,$E1,$56,$E3,$E5,$55,$9C,$C6,$DF,$8B
	.dc.b	$19,$7E,$3C,$DB,$4D,$8C,$36,$B9,$4E,$9B,$3A,$69,$A9,$E1,$5E,$9B
	.dc.b	$FC,$14,$61,$F6,$98,$9E,$33,$B7,$D6,$62,$36,$E2,$34,$7B,$8A,$7B
	.dc.b	$9E,$07,$CE,$3E,$8B,$AB,$FC,$0C,$66,$3E,$69,$9B,$F6,$36,$3D,$87
	.dc.b	$47,$C1,$5E,$97,$47,$5B,$F7,$4F,$59,$D1,$EA,$9E,$31,$93,$D9,$19
	.dc.b	$8F,$AB,$1E,$07,$AC,$66,$FE,$94,$62,$79,$3F,$A5,$AF,$0B,$D8,$C4
	.dc.b	$7A,$5E,$A7,$C0,$AE,$97,$F0,$1B,$3E,$16,$CF,$B6,$39,$5F,$CA,$8E
	.dc.b	$A7,$8C,$F0,$1F,$4C,$C1,$E0,$9F,$87,$3E,$88,$FD,$CF,$AB,$FB,$73
	.dc.b	$F7,$F9,$3E,$45,$EE,$66,$FF,$88,$D5,$BF,$E2,$35,$A9,$C9,$FE,$EA
	.dc.b	$FC,$3F,$FB,$DB,$7C,$2F,$E1,$79,$3E,$67,$FC,$05,$39,$26,$1F,$EE
	.dc.b	$FB,$E2,$FF,$C0,$0A,$E1,$E3,$7F,$11,$9C,$C7,$F1,$FC,$3F,$93,$92
	.dc.b	$FF,$A2,$5F,$F1,$98,$C3,$FC,$4D,$F7,$E3,$FE,$21,$9F,$E5,$B1,$E6
	.dc.b	$FA,$97,$89,$4C,$78,$B6,$98,$79,$6B,$71,$8C,$7F,$99,$C6,$27,$55
	.dc.b	$C5,$F6,$7D,$4F,$FA,$BA,$F8,$1C,$5C,$47,$C1,$D7,$FA,$94,$EA,$18
	.dc.b	$2B,$6F,$F4,$26,$CF,$3F,$7B,$E8,$DF,$2B,$F0,$C7,$F3,$BF,$B0,$7D
	.dc.b	$4F,$D6,$FB,$17,$C6,$E6,$30,$DC,$7C,$58,$2F,$8C,$79,$73,$6A,$F5
	.dc.b	$78,$5F,$A2,$BF,$C2,$B7,$C2,$30,$7C,$C7,$17,$F2,$AB,$8B,$ED,$F6
	.dc.b	$3F,$17,$B1,$BE,$1E,$19,$8F,$8E,$74,$D9,$CE,$3F,$67,$EC,$7F,$01
	.dc.b	$F1,$9F,$81,$6D,$7A,$F4,$46,$2B,$84,$E6,$BF,$B2,$7D,$6B,$EC,$3F
	.dc.b	$3F,$F6,$2F,$B0,$CC,$61,$B7,$F2,$AB,$83,$E7,$FB,$1C,$E3,$F1,$04
	.dc.b	$E9,$C3,$77,$FC,$04,$F0,$7F,$23,$F4,$7F,$9B,$C3,$F3,$8D,$C3,$D8
	.dc.b	$DC,$3F,$34,$DA,$FC,$57,$B1,$F3,$8F,$9F,$E3,$5B,$F9,$91,$6F,$F3
	.dc.b	$73,$6B,$FC,$C7,$97,$FA,$F2,$FF,$09,$CE,$F8,$DF,$8F,$EF,$FE,$3D
	.dc.b	$B3,$F0,$E3,$F6,$5F,$0B,$8C,$BD,$F1,$71,$FD,$6C,$DA,$FD,$47,$C7
	.dc.b	$FE,$13,$E5,$BF,$0A,$D8,$FC,$13,$70,$FD,$8B,$6F,$F6,$FE,$4F,$F0
	.dc.b	$13,$F8,$17,$97,$6F,$F0,$51,$6F,$F8,$04,$DA,$FF,$40,$FD,$8B,$2F
	.dc.b	$CA,$F9,$3F,$E2,$78,$C6,$3F,$93,$8C,$4C,$8F,$CB,$78,$17,$1F,$C0
	.dc.b	$5F,$06,$D7,$E8,$27,$58,$7F,$A4,$3E,$C5,$F6,$3F,$E3,$56,$E3,$FB
	.dc.b	$3F,$03,$FC,$FC,$FE,$E2,$33,$8F,$E3,$7E,$77,$EA,$4F,$EF,$1B,$1F
	.dc.b	$CC,$69,$1F,$E4,$13,$38,$7F,$A8,$3E,$BD,$F6,$3F,$E4,$56,$E1,$FC
	.dc.b	$0F,$EC,$7F,$AC,$62,$9C,$BF,$9D,$7F,$83,$4F,$F0,$C6,$C7,$F4,$1F
	.dc.b	$B1,$DE,$1D,$F2,$ED,$E8,$4C,$3F,$88,$57,$F7,$6A,$F2,$4F,$F9,$5F
	.dc.b	$E9,$5C,$D4,$FE,$A7,$F6,$2F,$03,$18,$FF,$1C,$A7,$CE,$F1,$3F,$AC
	.dc.b	$FD,$8B,$E6,$B1,$5F,$C8,$BE,$C5,$F5,$5F,$E6,$56,$C7,$F2,$3F,$09
	.dc.b	$E1,$E6,$B5,$F2,$F2,$B5,$EF,$DE,$4F,$E0,$18,$FF,$4E,$9E,$D1,$FD
	.dc.b	$AB,$EC,$5E,$17,$78,$FF,$3A,$4B,$E6,$29,$BE,$DB,$1F,$CE,$3E,$C7
	.dc.b	$F9,$1D,$7F,$0D,$6F,$3F,$FA,$2B,$CD,$FB,$B5,$FE,$CA,$9C,$BF,$EC
	.dc.b	$CF,$B1,$FE,$F4,$FF,$42,$6D,$FF,$17,$5F,$49,$FE,$87,$73,$87,$FB
	.dc.b	$99,$FF,$6E,$D7,$1F,$8F,$FF,$46,$7D,$83,$CE,$E1,$18,$64,$FF,$50
	.dc.b	$FB,$1F,$96,$FF,$52,$AE,$0D,$F2,$F9,$9F,$E9,$7F,$88,$7A,$D8,$9F
	.dc.b	$EB,$31,$86,$C7,$FA,$77,$EC,$7F,$C3,$9B,$7F,$31,$AD,$3F,$CC,$E3
	.dc.b	$3A,$FF,$C0,$09,$CA,$BF,$F7,$8C,$E9,$FF,$79,$CE,$9F,$F7,$AF,$C0
	.dc.b	$FE,$1A,$FF,$C3,$DF,$63,$FB,$67,$FA,$04,$67,$0F,$F5,$EC,$E5,$ED
	.dc.b	$3F,$E1,$29,$D5,$74,$EA,$FF,$17,$DF,$F7,$A8,$FE,$EF,$3A,$AE,$15
	.dc.b	$FE,$BB,$8C,$5A,$F8,$57,$8A,$B5,$C3,$FE,$3A,$FB,$1F,$EA,$18,$FA
	.dc.b	$A3,$1E,$BE,$A3,$BC,$6F,$E5,$31,$87,$B3,$EB,$FC,$07,$85,$B5,$E6
	.dc.b	$8B,$1F,$08,$B9,$FC,$07,$FC,$43,$E0,$FF,$B4,$E3,$F6,$95,$FE,$14
	.dc.b	$9F,$F1,$AF,$89,$FE,$36,$FB,$17,$CA,$FF,$8E,$27,$8D,$7F,$86,$AF
	.dc.b	$8C,$7F,$6F,$69,$FE,$F7,$BE,$55,$C6,$33,$5F,$E4,$E9,$CD,$7F,$90
	.dc.b	$70,$FA,$B8,$78,$23,$C0,$27,$6F,$88,$46,$DF,$34,$6A,$FC,$B3,$99
	.dc.b	$F6,$9C,$30,$F2,$5F,$81,$F2,$46,$4F,$26,$1E,$07,$8D,$77,$BE,$B5
	.dc.b	$C2,$31,$7C,$EB,$D3,$87,$8E,$31,$1E,$63,$B1,$EB,$9F,$33,$EF,$33
	.dc.b	$7D,$9E,$67,$F9,$D8,$F5,$96,$B0,$E0,$FB,$1F,$63,$FC,$77,$E2,$F8
	.dc.b	$36,$3F,$CF,$ED,$7E,$7B,$5E,$3F,$1B,$FE,$68,$76,$D7,$6D,$77,$7D
	.dc.b	$A3,$86,$1F,$57,$26,$9F,$8B,$8C,$36,$38,$DF,$C4,$C3,$FD,$FF,$87
	.dc.b	$E0,$27,$8C,$DA,$E0,$FF,$A3,$2B,$6F,$1C,$3C,$35,$F9,$93,$8C,$7D
	.dc.b	$06,$E1,$F8,$9A,$E2,$DA,$E3,$5F,$E9,$8A,$D9,$CE,$FC,$30,$BE,$33
	.dc.b	$EA,$BF,$6C,$F9,$CF,$CC,$CE,$9E,$95,$FA,$5D,$3A,$47,$FA,$1A,$FC
	.dc.b	$F8,$FE,$E6,$F8,$4F,$ED,$2F,$83,$E5,$F4,$EB,$FB,$86,$DF,$2C,$3F
	.dc.b	$75,$F6,$3F,$DC,$EB,$FD,$26,$D7,$FC,$D0,$DC,$7F,$84,$2C,$FA,$B9
	.dc.b	$AF,$F5,$C7,$23,$E5,$BC,$77,$FC,$8D,$DF,$07,$93,$EC,$7F,$A5,$E5
	.dc.b	$FE,$A2,$FB,$0F,$D5,$63,$FC,$35,$B1,$FE,$C6,$9D,$BF,$E7,$B6,$E1
	.dc.b	$FC,$6E,$77,$26,$CE,$AB,$F8,$C3,$FE,$1A,$9F,$F5,$8B,$5E,$9E,$8F
	.dc.b	$C8,$A7,$D8,$F9,$E3,$1C,$D2,$B8,$C6,$5F,$54,$65,$E6,$7F,$AF,$BE
	.dc.b	$C7,$FB,$DB,$63,$EA,$C6,$23,$FD,$2D,$7F,$EC,$87,$47,$FB,$1B,$D1
	.dc.b	$7C,$77,$DC,$6C,$68,$FF,$4F,$38,$79,$ED,$EB,$38,$4F,$1F,$23,$97
	.dc.b	$D8,$CA,$ED,$15,$FF,$54,$7D,$8F,$C9,$7F,$DE,$ED,$7F,$C7,$B3,$F8
	.dc.b	$E8,$FF,$15,$C7,$DB,$EC,$F9,$46,$7C,$DF,$C9,$FD,$8F,$FA,$66,$7F
	.dc.b	$DC,$3C,$8F,$47,$F2,$ED,$8F,$F7,$2C,$ED,$FF,$5D,$7D,$8F,$F9,$99
	.dc.b	$E3,$F0,$7C,$0F,$CF,$C6,$63,$7D,$5D,$3E,$46,$47,$FB,$C2,$7F,$34
	.dc.b	$DF,$6F,$F4,$FF,$60,$FB,$1A,$74,$8F,$D5,$9F,$87,$EC,$FE,$D5,$2F
	.dc.b	$87,$E2,$6B,$AB,$7F,$63,$D3,$9A,$65,$FF,$00,$31,$4F,$ED,$BE,$C7
	.dc.b	$FC,$C3,$23,$FD,$B1,$F6,$3F,$E8,$87,$19,$FF,$79,$8C,$47,$FB,$76
	.dc.b	$78,$B7,$DC,$FE,$03,$2F,$A2,$A7,$78,$FE,$0B,$F6,$3F,$EA,$FA,$EB
	.dc.b	$E1,$FF,$7D,$B7,$4F,$23,$7F,$17,$C3,$DD,$EB,$F4,$73,$0E,$27,$A7
	.dc.b	$1F,$F3,$53,$81,$F6,$0F,$98,$FE,$2B,$F6,$2F,$AE,$A7,$F1,$89,$D5
	.dc.b	$71,$5F,$63,$C3,$E1,$E4,$E6,$3F,$8F,$F8,$2F,$8C,$A4,$9F,$17,$D7
	.dc.b	$65,$F2,$4D,$5C,$DF,$E2,$59,$FD,$53,$7A,$79,$4E,$B9,$3F,$C4,$57
	.dc.b	$C6,$B5,$3D,$AB,$71,$FE,$2F,$FB,$17,$CA,$FF,$34,$6D,$7C,$56,$BD
	.dc.b	$4F,$C1,$3F,$C7,$0C,$CF,$39,$E4,$73,$F4,$FA,$7D,$5F,$5F,$32,$61
	.dc.b	$D6,$32,$F7,$8C,$BD,$C6,$09,$F9,$7F,$47,$27,$E3,$BE,$9F,$B1,$C4
	.dc.b	$71,$FB,$1F,$F2,$66,$7F,$E4,$D6,$ED,$F1,$AF,$73,$98,$FE,$AB,$3A
	.dc.b	$AF,$D0,$6C,$BC,$F2,$49,$FC,$55,$7E,$1D,$BC,$19,$7C,$53,$71,$F1
	.dc.b	$4F,$6F,$8F,$EF,$C9,$2F,$9C,$65,$EB,$19,$7D,$91,$FD,$AB,$EC,$5F
	.dc.b	$63,$49,$3F,$9D,$FB,$17,$D5,$62,$3E,$64,$7A,$DD,$6B,$9C,$FF,$CE
	.dc.b	$58,$74,$9C,$90,$FA,$8E,$5E,$7F,$15,$FD,$F2,$FF,$88,$47,$87,$C9
	.dc.b	$C4,$C8,$C1,$0E,$27,$9F,$EA,$79,$11,$F5,$74,$D7,$D3,$17,$1F,$F4
	.dc.b	$46,$1F,$A4,$E3,$FF,$41,$7D,$8F,$F0,$6B,$FD,$F3,$5F,$C6,$36,$BF
	.dc.b	$D2,$1E,$3B,$E4,$3F,$C2,$B3,$9B,$FF,$46,$4F,$F7,$A3,$FB,$DC,$66
	.dc.b	$FF,$92,$35,$FE,$39,$D7,$FE,$1E,$9D,$F8,$D7,$AA,$57,$F1,$19,$FC
	.dc.b	$E3,$7B,$32,$FE,$86,$32,$D9,$E9,$3D,$7C,$49,$3C,$DF,$F1,$6F,$D8
	.dc.b	$FE,$66,$3F,$97,$46,$9E,$8E,$B3,$FF,$19,$FE,$37,$FE,$F5,$9F,$F3
	.dc.b	$C3,$5F,$F4,$E8,$70,$FF,$A8,$1A,$FF,$AC,$7E,$C1,$E9,$FF,$9F,$9B
	.dc.b	$1F,$E8,$36,$BD,$1A,$FF,$AA,$97,$57,$12,$47,$17,$FD,$13,$F6,$2F
	.dc.b	$B2,$7C,$87,$FA,$3E,$30,$DF,$0B,$FA,$9D,$FF,$B1,$67,$2F,$B3,$EC
	.dc.b	$7F,$B3,$69,$FE,$97,$6C,$67,$AF,$FD,$A8,$FF,$B3,$63,$2F,$FB,$01
	.dc.b	$DF,$3F,$FB,$65,$B5,$FD,$9F,$E3,$1F,$22,$F0,$BF,$DF,$B7,$BC,$6D
	.dc.b	$F5,$EE,$B9,$31,$7D,$53,$E0,$F3,$2D,$17,$C2,$F7,$5A,$9C,$E1,$FD
	.dc.b	$8A,$72,$FF,$61,$8C,$DF,$F3,$D9,$CB,$FC,$C6,$33,$92,$BD,$98,$F8
	.dc.b	$DE,$58,$F8,$63,$17,$C6,$31,$8F,$8D,$B5,$C0,$F2,$57,$0B,$F9,$53
	.dc.b	$AB,$E5,$5A,$BE,$4D,$39,$47,$A6,$F9,$47,$CB,$35,$1C,$8D,$4F,$26
	.dc.b	$C7,$2B,$F7,$47,$D1,$AE,$6F,$F8,$87,$0F,$87,$5F,$0A,$F9,$5F,$0C
	.dc.b	$7E,$4B,$EA,$9F,$0B,$6F,$B9,$E3,$78,$61,$DA,$3E,$E6,$3F,$06,$78
	.dc.b	$1A,$C3,$BD,$F8,$4F,$41,$6B,$A3,$67,$8B,$4F,$89,$1E,$28,$CB,$EF
	.dc.b	$9E,$AD,$9E,$AD,$3A,$96,$78,$E1,$F0,$27,$AC,$79,$DE,$31,$99,$F9
	.dc.b	$57,$FA,$C6,$CF,$EC,$63,$11,$F0,$A3,$04,$7F,$71,$1A,$C7,$F3,$31
	.dc.b	$FB,$38,$F4,$BF,$A5,$F7,$2C,$FD,$88,$EE,$7E,$76,$33,$3A,$6D,$7C
	.dc.b	$09,$E4,$DA,$E5,$1E,$53,$94,$66,$B9,$34,$E1,$5D,$27,$E4,$9D,$0B
	.dc.b	$5D,$09,$1B,$C3,$EA,$CE,$DB,$3B,$BF,$DF,$56,$EF,$E8,$9F,$C5,$DC
	.dc.b	$5F,$C0,$BF,$E2,$D1,$88,$F5,$46,$27,$D1,$1E,$49,$D5,$F9,$27,$35
	.dc.b	$E2,$9C,$DF,$8A,$7E,$37,$B1,$FC,$47,$ED,$3F,$89,$36,$3F,$88,$3F
	.dc.b	$64,$8F,$F3,$17,$F0,$3F,$65,$FE,$2F,$8F,$EA,$39,$23,$0F,$F5,$46
	.dc.b	$BF,$E9,$53,$FD,$2C,$E1,$C3,$C3,$1F,$DE,$C7,$3F,$1B,$F9,$7C,$3E
	.dc.b	$CB,$FC,$4F,$7F,$E8,$5B,$E0,$FF,$41,$AF,$E3,$CF,$F2,$D6,$BF,$CA
	.dc.b	$DF,$CA,$11,$FE,$98,$FE,$7E,$9F,$5E,$F1,$BC,$2D,$AF,$F4,$BC,$EB
	.dc.b	$C7,$CD,$49,$CD,$6F,$2F,$F3,$A5,$EE,$73,$5E,$CF,$B0,$FE,$E7,$1D
	.dc.b	$1B,$5C,$33,$FE,$D3,$39,$ED,$E0,$47,$C7,$1A,$9E,$F5,$CB,$0F,$97
	.dc.b	$F6,$2F,$AB,$F9,$7F,$5C,$FB,$2F,$EA,$F3,$9C,$BD,$B3,$97,$FD,$65
	.dc.b	$19,$C7,$FD,$00,$DC,$3F,$B3,$EB,$FE,$A9,$AD,$E3,$D0,$8F,$C4,$F5
	.dc.b	$AE,$1E,$6F,$6B,$E1,$1F,$0F,$E7,$6D,$7C,$38,$B3,$EC,$9B,$8F,$83
	.dc.b	$EC,$BE,$AF,$C7,$7D,$66,$E1,$F4,$9B,$7F,$4B,$34,$62,$FF,$1F,$F6
	.dc.b	$2F,$9C,$FC,$7F,$7B,$EC,$BF,$BD,$CE,$72,$FC,$0C,$E5,$FF,$6B,$46
	.dc.b	$72,$FF,$52,$37,$0F,$F0,$5F,$1B,$EA,$B9,$E5,$30,$E2,$FE,$0E,$31
	.dc.b	$87,$9F,$E9,$3F,$23,$9F,$E4,$9B,$7F,$8E,$6C,$EE,$3F,$23,$3F,$8D
	.dc.b	$F2,$A4,$7E,$37,$C5,$B5,$F8,$69,$B9,$F9,$A6,$D7,$ED,$F4,$FF,$68
	.dc.b	$BF,$84,$FB,$2F,$87,$F0,$1F,$8A,$6C,$7C,$28,$FD,$3F,$C3,$63,$2F
	.dc.b	$87,$17,$1F,$C3,$CD,$A5,$7F,$BA,$B7,$7C,$97,$E9,$7E,$33,$F4,$CD
	.dc.b	$AF,$D3,$36,$BF,$3E,$DC,$3F,$27,$E2,$FE,$E0,$FD,$CF,$95,$6F,$F6
	.dc.b	$B1,$6F,$F0,$93,$6B,$F8,$8F,$D9,$33,$FD,$B7,$C8,$FF,$8F,$23,$18
	.dc.b	$FE,$DE,$31,$32,$3F,$75,$DB,$71,$FF,$73,$5F,$06,$D7,$F0,$19,$D6
	.dc.b	$1F,$C8,$3E,$C9,$F5,$5F,$E4,$A6,$E3,$FC,$43,$BF,$FD,$DC,$FC,$8A
	.dc.b	$E1,$5F,$E5,$0F,$9F,$71,$D3,$F9,$4D,$A3,$17,$C4,$FE,$69,$F6,$0F
	.dc.b	$B2,$FF,$2C,$B6,$3F,$8F,$FD,$97,$F6,$C8,$C6,$DF,$F0,$05,$FE,$00
	.dc.b	$DC,$3F,$9B,$FD,$96,$B4,$FE,$77,$5A,$6F,$9D,$D8,$DF,$A1,$72,$F2
	.dc.b	$63,$F8,$4F,$A7,$74,$7F,$9B,$1B,$3F,$CC,$BE,$CB,$F5,$8F,$0A,$4E
	.dc.b	$96,$3F,$D3,$23,$FC,$74,$7F,$92,$17,$D9,$3E,$77,$B7,$AF,$FE,$7A
	.dc.b	$FB,$27,$D6,$53,$FA,$9F,$85,$FE,$A2,$CF,$C1,$3A,$6F,$CC,$F1,$37
	.dc.b	$EE,$D8,$DE,$BD,$4F,$EB,$BF,$65,$FA,$08,$C9,$EB,$1F,$33,$FE,$8C
	.dc.b	$69,$FD,$4A,$3F,$0B,$3E,$E8,$FE,$DC,$69,$C1,$FA,$EF,$B2,$7A,$5F
	.dc.b	$E9,$66,$CF,$F5,$CF,$47,$F1,$2F,$CB,$8C,$D7,$21,$F6,$5F,$DC,$E7
	.dc.b	$E8,$36,$34,$BE,$9F,$FD,$46,$E7,$8F,$FB,$EB,$FE,$FB,$AE,$3F,$1F
	.dc.b	$FE,$A8,$FB,$17,$A1,$C2,$30,$C9,$FB,$DF,$65,$F4,$72,$F2,$5F,$09
	.dc.b	$FC,$13,$FE,$42,$3F,$C8,$7F,$0C,$F6,$3B,$79,$BF,$EB,$68,$C5,$7E
	.dc.b	$25,$B7,$EF,$FB,$27,$9B,$BC,$3F,$D7,$B1,$8B,$E1,$18,$9D,$47,$E4
	.dc.b	$FE,$1B,$F0,$97,$FE,$02,$FA,$DF,$DC,$BF,$2E,$33,$C1,$FE,$D0,$9C
	.dc.b	$BF,$E2,$28,$CD,$7F,$8A,$5B,$5F,$9C,$AF,$EE,$CF,$3E,$BF,$F6,$8C
	.dc.b	$EA,$3F,$DB,$B1,$FE,$2A,$9D,$DF,$58,$DF,$17,$F8,$BB,$EC,$BF,$AC
	.dc.b	$63,$DE,$31,$EC,$EA,$3D,$91,$BF,$92,$C6,$5E,$BF,$B0,$F0,$9E,$26
	.dc.b	$D7,$C3,$8B,$1F,$04,$B9,$7F,$90,$0F,$D9,$4F,$EE,$DB,$1F,$91,$3F
	.dc.b	$82,$0F,$B2,$7C,$AF,$F9,$36,$76,$73,$F0,$FF,$DE,$6E,$2F,$FC,$A5
	.dc.b	$7C,$AF,$8B,$FC,$AE,$BF,$DF,$2E,$6F,$A4,$7F,$B4,$27,$73,$FE,$5B
	.dc.b	$F1,$7F,$AC,$C6,$30,$F3,$BA,$9E,$47,$5C,$37,$87,$8A,$3F,$8B,$9E
	.dc.b	$B7,$BC,$6A,$3C,$F7,$5F,$E3,$78,$78,$EB,$94,$E7,$0F,$0C,$62,$BA
	.dc.b	$1E,$12,$9E,$92,$C7,$07,$B9,$8A,$EC,$E3,$0F,$74,$62,$39,$1F,$E8
	.dc.b	$2B,$FF,$41,$BF,$58,$C1,$EF,$FB,$2F,$E3,$7F,$1B,$FE,$88,$6C,$67
	.dc.b	$C1,$4E,$5B,$1C,$7C,$AF,$E7,$6D,$CB,$94,$62,$3E,$ED,$A3,$11,$FE
	.dc.b	$91,$8C,$4F,$CB,$CD,$AF,$FA,$57,$4F,$BB,$18,$BF,$A1,$7C,$1F,$92
	.dc.b	$D9,$DD,$71,$AF,$F4,$F1,$C6,$B7,$87,$A2,$BF,$13,$7C,$6B,$77,$C6
	.dc.b	$FF,$C4,$71,$C6,$B7,$7C,$63,$D4,$D9,$DD,$71,$7F,$46,$DB,$ED,$1F
	.dc.b	$E2,$E6,$BA,$3E,$14,$6E,$7A,$46,$4F,$E8,$15,$EB,$C3,$FD,$6A,$DD
	.dc.b	$7C,$71,$CA,$7F,$B9,$4F,$F8,$3D,$BE,$8F,$ED,$6B,$DF,$1F,$DD,$FE
	.dc.b	$CB,$FB,$83,$77,$D3,$FE,$44,$8C,$BF,$D4,$5F,$B8,$EC,$FF,$59,$F2
	.dc.b	$BE,$57,$8E,$48,$C4,$EC,$FE,$B5,$F6,$5F,$D0,$DF,$F0,$47,$D8,$7E
	.dc.b	$E7,$2F,$F6,$B3,$63,$FC,$AB,$3B,$78,$F5,$F9,$94,$ED,$BA,$E6,$B8
	.dc.b	$63,$FE,$DC,$7F,$C3,$73,$FE,$17,$6B,$E4,$F3,$FA,$5B,$EC,$B9,$46
	.dc.b	$3A,$25,$71,$8C,$BE,$A8,$CB,$CC,$FF,$12,$7D,$97,$F7,$36,$C7,$A6
	.dc.b	$31,$1F,$DF,$6F,$FD,$88,$E8,$FF,$16,$79,$EF,$8F,$F8,$0D,$8D,$1F
	.dc.b	$E0,$57,$0F,$3E,$0F,$59,$C2,$78,$F9,$5C,$BE,$CA,$57,$68,$AF,$F8
	.dc.b	$3F,$EC,$BE,$4F,$17,$FA,$1A,$11,$87,$FD,$13,$8A,$FE,$A4,$FF,$A2
	.dc.b	$A7,$EE,$F7,$7C,$C3,$3F,$05,$F6,$7E,$CB,$FC,$CF,$3F,$E4,$9E,$87
	.dc.b	$A3,$F7,$5B,$1F,$E4,$D9,$DB,$FE,$23,$FB,$2F,$F3,$43,$C7,$C7,$E0
	.dc.b	$7E,$16,$33,$1B,$ED,$E9,$F2,$52,$3F,$CB,$13,$F9,$C7,$A7,$B3,$F8
	.dc.b	$DF,$B1,$7D,$95,$3A,$47,$E4,$4F,$C4,$F6,$38,$FE,$35,$3E,$1F,$91
	.dc.b	$BE,$BE,$1F,$65,$D7,$A2,$67,$FE,$69,$C9,$3F,$9D,$FB,$2F,$F3,$0C
	.dc.b	$FF,$9B,$7E,$CB,$F7,$8F,$F9,$0D,$BA,$78,$63,$FD,$77,$3C,$1F,$F2
	.dc.b	$42,$F6,$FF,$51,$4F,$A4,$A7,$78,$FD,$67,$D9,$7F,$AC,$6F,$C1,$D3
	.dc.b	$C0,$CF,$C5,$EB,$CD,$E7,$F4,$72,$BF,$D0,$57,$FE,$68,$7F,$D0,$9F
	.dc.b	$63,$FE,$60,$7F,$D0,$DF,$64,$FA,$EF,$DD,$36,$37,$F0,$3F,$BD,$A7
	.dc.b	$D9,$6B,$E1,$65,$74,$8F,$E0,$5E,$0B,$E2,$A9,$27,$FD,$29,$F5,$F9
	.dc.b	$7F,$D2,$D3,$5F,$F3,$64,$7E,$A3,$97,$E5,$42,$31,$D9,$FA,$1A,$E1
	.dc.b	$5A,$6C,$7E,$96,$77,$5F,$EE,$C6,$BF,$9C,$BF,$E3,$1F,$64,$F9,$5A
	.dc.b	$49,$FF,$55,$47,$F1,$B3,$4E,$23,$4F,$FA,$0E,$35,$F5,$7D,$3C,$C9
	.dc.b	$3C,$A3,$B9,$87,$84,$61,$F8,$7F,$47,$F2,$88,$F8,$BE,$CA,$BF,$E9
	.dc.b	$1F,$B2,$FF,$23,$69,$FC,$0D,$B0,$3F,$73,$5C,$36,$F9,$73,$A9,$37
	.dc.b	$65,$7F,$3B,$8F,$C1,$4F,$7E,$4C,$19,$DE,$C7,$FB,$3E,$77,$87,$6A
	.dc.b	$D4,$E7,$E6,$FA,$71,$48,$C1,$0F,$63,$97,$DD,$1F,$D3,$7E,$C1,$F6
	.dc.b	$54,$93,$FE,$DF,$FB,$17,$D9,$62,$3F,$D6,$D1,$EC,$75,$AE,$73,$FC
	.dc.b	$6E,$1D,$27,$24,$3F,$D7,$CE,$5E,$7F,$15,$FD,$7E,$FF,$86,$47,$87
	.dc.b	$C8,$C4,$C8,$C1,$0E,$27,$9F,$EA,$79,$11,$F5,$79,$15,$F4,$C5,$C7
	.dc.b	$F4,$78,$7E,$6B,$EC,$9E,$1F,$F7,$16,$DF,$FC,$01,$6B,$FE,$D7,$F1
	.dc.b	$5F,$21,$BA,$FE,$F1,$1F,$DD,$CF,$EF,$15,$C3,$0F,$D8,$B7,$9F,$FA
	.dc.b	$84,$EF,$DE,$E5,$8F,$E3,$EF,$F6,$F2,$EB,$EA,$49,$D5,$F2,$EB,$C4
	.dc.b	$79,$66,$DF,$29,$F2,$F8,$92,$B8,$9F,$9D,$C1,$1F,$E1,$1F,$B2,$FC
	.dc.b	$E4,$62,$76,$7F,$2D,$9E,$4F,$FB,$E6,$7F,$45,$F6,$5F,$DD,$A3,$FD
	.dc.b	$46,$D7,$FB,$73,$5F,$D6,$35,$3F,$C3,$BE,$23,$F5,$7F,$25,$FE,$38
	.dc.b	$6C,$F8,$A7,$F8,$88,$E2,$FF,$14,$36,$35,$1C,$7E,$59,$F4,$1A,$8F
	.dc.b	$F4,$5D,$6E,$BF,$D5,$0D,$7D,$31,$FE,$90,$FB,$07,$D8,$BE,$53,$FD
	.dc.b	$5D,$49,$CC,$61,$FE,$FF,$F1,$76,$B8,$57,$FA,$D7,$ED,$7F,$5B,$C9
	.dc.b	$C9,$B1,$9E,$FF,$E6,$EF,$F8,$56,$32,$FF,$B3,$BE,$C9,$F0,$BF,$D9
	.dc.b	$79,$BF,$9E,$C7,$C3,$C3,$EB,$BA,$6E,$BB,$7F,$C4,$11,$9B,$57,$2A
	.dc.b	$D4,$7F,$21,$9D,$4E,$6B,$AD,$EA,$95,$E1,$79,$CF,$69,$D4,$E6,$BC
	.dc.b	$F1,$8B,$CB,$6B,$A9,$E8,$2D,$F4,$3C,$93,$C2,$35,$19,$3C,$73,$CD
	.dc.b	$C5,$FA,$A7,$E0,$D7,$19,$CB,$6F,$DD,$1C,$2F,$94,$7A,$2B,$BB,$67
	.dc.b	$8C,$79,$A3,$89,$92,$BF,$0C,$E2,$58,$E2,$77,$72,$FD,$19,$F0,$C6
	.dc.b	$E7,$9C,$E6,$36,$FC,$06,$DF,$BE,$3E,$3C,$79,$4F,$82,$57,$27,$84
	.dc.b	$F1,$CF,$A6,$37,$19,$8F,$0B,$DC,$DC,$79,$4D,$C7,$18,$EA,$FC,$77
	.dc.b	$F3,$55,$F9,$E7,$D4,$7F,$9C,$CA,$6A,$7F,$5F,$5F,$97,$8C,$1B,$3C
	.dc.b	$46,$2B,$C1,$C3,$FB,$7F,$81,$FB,$97,$E5,$AE,$F5,$FE,$4D,$AD,$4F
	.dc.b	$C3,$9F,$64,$F2,$DD,$7C,$A7,$8D,$73,$BE,$31,$F1,$63,$B3,$C9,$FF
	.dc.b	$05,$46,$AF,$E2,$4E,$4E,$62,$35,$3C,$B0,$FB,$53,$B9,$CB,$C0,$E4
	.dc.b	$67,$0F,$3D,$7A,$EB,$89,$F8,$3A,$E3,$5C,$23,$38,$FF,$00,$9C,$C6
	.dc.b	$2B,$F0,$D1,$89,$F8,$55,$DF,$1B,$7B,$8F,$DB,$56,$EF,$C5,$7D,$70
	.dc.b	$F2,$D7,$C7,$3D,$57,$E5,$8C,$E3,$F1,$A3,$A3,$F5,$E3,$37,$F1,$6F
	.dc.b	$EA,$61,$DF,$0E,$87,$96,$7E,$17,$D5,$FB,$7E,$AD,$E3,$FF,$14,$8C
	.dc.b	$36,$7F,$1F,$1F,$95,$AF,$E3,$AD,$CC,$DF,$6E,$57,$E5,$9B,$5F,$95
	.dc.b	$A4,$7F,$22,$3E,$77,$CB,$BE,$0F,$17,$FC,$E3,$E1,$FC,$DF,$4B,$FC
	.dc.b	$A1,$B1,$93,$F4,$6F,$8C,$8F,$E8,$F8,$1F,$59,$FC,$BE,$3F,$C1,$8D
	.dc.b	$CF,$F9,$FB,$FE,$F1,$8F,$F0,$1C,$F0,$E3,$FE,$82,$D8,$CC,$F1,$8E
	.dc.b	$BE,$3E,$A3,$1F,$52,$F8,$6E,$2D,$BE,$BF,$61,$FD,$A5,$EF,$19,$CB
	.dc.b	$F9,$E6,$28,$C3,$FD,$75,$B3,$FE,$25,$AE,$55,$C2,$72,$FF,$9D,$FE
	.dc.b	$BB,$FB,$FD,$7B,$3E,$BD,$F6,$1F,$D5,$E7,$35,$E0,$9C,$D7,$F8,$7A
	.dc.b	$73,$8F,$FA,$A1,$B1,$F2,$BD,$BE,$5E,$14,$8E,$0F,$C3,$8C,$61,$E5
	.dc.b	$F7,$BE,$19,$F1,$3E,$56,$DF,$C2,$8B,$3E,$B9,$B8,$FF,$7D,$DF,$F1
	.dc.b	$F7,$F4,$BB,$3E,$87,$4F,$D1,$6C,$64,$FA,$F1,$97,$FC,$73,$F5,$CF
	.dc.b	$9D,$FB,$BD,$AF,$AF,$FE,$F5,$39,$AF,$C1,$CE,$5E,$35,$A3,$EE,$FD
	.dc.b	$85,$8C,$93,$5F,$A9,$19,$7E,$EE,$6F,$2B,$F1,$19,$7E,$25,$B7,$F8
	.dc.b	$E6,$D7,$06,$EB,$F8,$3F,$6B,$F1,$A7,$E3,$BC,$1B,$7F,$8C,$8B,$7F
	.dc.b	$85,$9B,$5F,$EE,$2D,$7F,$DA,$D7,$F8,$5F,$AE,$F8,$3F,$05,$FA,$0E
	.dc.b	$74,$62,$7E,$4C,$FE,$7F,$E2,$31,$97,$F8,$3F,$0E,$6D,$9D,$57,$03
	.dc.b	$F3,$FF,$27,$FA,$4F,$8E,$FD,$2B,$63,$27,$E9,$23,$2D,$3F,$63,$ED
	.dc.b	$FE,$BC,$FD,$87,$B9,$63,$24,$C3,$E7,$C6,$5F,$F1,$27,$D8,$33,$FD
	.dc.b	$AF,$8D,$FE,$39,$8C,$5F,$ED,$23,$13,$23,$F6,$FD,$B6,$F9,$CD,$E0
	.dc.b	$FD,$B4,$DA,$FF,$74,$53,$EC,$1E,$3F,$F9,$1D,$B3,$FC,$43,$BF,$FD
	.dc.b	$3C,$FF,$A3,$A3,$35,$FE,$3B,$F4,$7E,$9C,$FF,$A4,$9B,$1F,$C2,$29
	.dc.b	$1F,$E5,$33,$38,$AB,$C7,$F5,$AF,$B0,$FF,$2C,$36,$3F,$90,$FD,$83
	.dc.b	$87,$35,$AC,$65,$7F,$26,$FB,$0E,$19,$FF,$93,$F1,$57,$4F,$3A,$5E
	.dc.b	$FE,$62,$DF,$F9,$57,$E9,$9C,$DF,$E6,$96,$D7,$F3,$2F,$B0,$78,$59
	.dc.b	$AD,$7C,$EC,$AF,$E6,$FF,$61,$FD,$8B,$10,$C5,$5F,$E7,$5F,$75,$F5
	.dc.b	$54,$FE,$7B,$E1,$3C,$3C,$65,$FE,$BE,$A7,$CB,$F1,$37,$F2,$16,$75
	.dc.b	$9F,$53,$FA,$3F,$D8,$3C,$2C,$4F,$FB,$20,$7C,$C7,$FA,$1D,$A7,$F5
	.dc.b	$09,$FB,$FD,$7F,$5E,$FD,$73,$A7,$FC,$F9,$F6,$0F,$47,$FD,$20,$D9
	.dc.b	$FE,$B3,$E8,$7E,$01,$FC,$8C,$66,$BF,$D0,$9F,$60,$F5,$BF,$D3,$0D
	.dc.b	$AF,$F6,$EA,$FA,$5C,$CE,$BE,$3A,$57,$FB,$8B,$EB,$FF,$8D,$3C,$23
	.dc.b	$FD,$45,$18,$C3,$FD,$24,$DD,$3D,$90,$FB,$0F,$A1,$7C,$FD,$1E,$3E
	.dc.b	$24,$8F,$F7,$7F,$C1,$3D,$8C,$61,$FE,$9D,$8C,$4E,$A7,$FD,$65,$F6
	.dc.b	$1F,$E0,$B8,$C1,$FE,$F8,$71,$5C,$67,$7E,$77,$FA,$E6,$F9,$CF,$FA
	.dc.b	$A7,$E0,$7F,$0F,$87,$F8,$13,$EA,$FF,$6B,$39,$EE,$FF,$02,$C6,$36
	.dc.b	$F6,$DF,$0A,$D5,$F1,$F7,$7F,$D8,$D3,$9A,$DE,$18,$39,$CD,$AF,$CD
	.dc.b	$CF,$4E,$3F,$F0,$9F,$D8,$7F,$55,$C7,$CD,$18,$F6,$FA,$B2,$FF,$C3
	.dc.b	$74,$F9,$3C,$5F,$B7,$EB,$BC,$67,$91,$B5,$E0,$3F,$8C,$E8,$9C,$CF
	.dc.b	$C5,$2E,$5F,$70,$FD,$93,$FE,$EA,$F8,$8F,$F7,$23,$FB,$4F,$1B,$F6
	.dc.b	$BF,$60,$F9,$4F,$E4,$B7,$C0,$97,$C2,$FF,$74,$2B,$83,$F1,$23,$57
	.dc.b	$D2,$F9,$C7,$EA,$6B,$75,$FC,$CA,$B4,$F1,$CA,$B9,$3B,$5F,$28,$CC
	.dc.b	$7A,$4C,$9C,$4A,$71,$AE,$F8,$7F,$9F,$1F,$0C,$71,$73,$87,$96,$FF
	.dc.b	$D7,$18,$78,$AF,$84,$62,$FA,$C6,$1F,$F6,$24,$EE,$B9,$46,$0D,$B8
	.dc.b	$7D,$87,$FB,$14,$F9,$65,$8C,$3E,$D3,$FD,$A1,$F6,$1F,$C5,$BE,$2F
	.dc.b	$83,$63,$FD,$9C,$D7,$7E,$1F,$FA,$DD,$CF,$9E,$C7,$8B,$63,$31,$82
	.dc.b	$3F,$2B,$2F,$C7,$3F,$51,$AF,$CE,$B5,$73,$CF,$EE,$C6,$2F,$86,$BF
	.dc.b	$DC,$27,$F0,$38,$72,$77,$7C,$63,$FB,$AD,$F1,$C3,$C5,$7C,$67,$FB
	.dc.b	$C9,$C4,$DD,$71,$9F,$E9,$F5,$C6,$FB,$DF,$19,$FE,$A7,$7C,$5D,$DF
	.dc.b	$15,$7C,$67,$BD,$71,$9F,$D3,$36,$75,$1D,$23,$F5,$3A,$F0,$8F,$DC
	.dc.b	$19,$29,$DA,$FC,$17,$F3,$31,$FE,$66,$FF,$82,$27,$6D,$7E,$D5,$75
	.dc.b	$69,$FD,$BB,$EC,$1B,$78,$A3,$0D,$51,$A7,$C6,$FD,$A6,$C7,$1B,$73
	.dc.b	$3E,$57,$0D,$AD,$9E,$CF,$B0,$FE,$88,$FB,$BE,$BB,$F6,$39,$7F,$8C
	.dc.b	$1B,$1F,$E0,$49,$DB,$EA,$0E,$30,$FE,$AB,$3C,$24,$D9,$D5,$7F,$8F
	.dc.b	$1F,$F0,$CC,$F8,$FD,$1F,$81,$18,$8E,$D4,$FB,$0C,$D7,$C5,$8F,$ED
	.dc.b	$D3,$E9,$E8,$CB,$CC,$C1,$0F,$45,$F5,$C7,$FB,$5F,$D8,$72,$9F,$63
	.dc.b	$63,$35,$DF,$AF,$EA,$4E,$73,$FF,$2A,$FA,$4F,$8E,$FB,$0D,$C7,$F1
	.dc.b	$35,$B7,$FC,$5B,$19,$7D,$BB,$FA,$CE,$35,$CB,$CB,$E7,$F2,$12,$BF
	.dc.b	$27,$69,$C4,$7D,$3F,$B0,$F9,$AF,$F8,$E9,$AF,$F8,$EA,$BF,$19,$1F
	.dc.b	$E2,$D8,$FC,$07,$77,$CB,$33,$E8,$7E,$47,$EC,$3F,$D3,$33,$FA,$2E
	.dc.b	$67,$A5,$F9,$56,$C7,$E9,$A7,$6F,$E1,$7E,$C3,$FC,$CC,$F1,$F1,$F8
	.dc.b	$1F,$9D,$8C,$C6,$FB,$3A,$7C,$94,$8F,$D8,$CF,$E7,$DE,$9E,$E7,$E9
	.dc.b	$7E,$BD,$F6,$14,$E9,$1F,$A9,$3E,$F7,$37,$FA,$2A,$5F,$15,$C4,$D7
	.dc.b	$6E,$FE,$C3,$A7,$3C,$CB,$F8,$0E,$29,$FD,$A7,$D8,$7F,$98,$A7,$F8
	.dc.b	$2F,$D8,$7E,$FF,$BB,$FC,$6A,$DD,$3F,$D3,$35,$BA,$FD,$43,$FE,$64
	.dc.b	$78,$FD,$0F,$EF,$E8,$F6,$F7,$7F,$81,$7D,$87,$FA,$BE,$BB,$F4,$EC
	.dc.b	$CF,$C7,$6B,$CC,$E5,$F4,$52,$BF,$8C,$DF,$F9,$A1,$FE,$37,$F5,$FF
	.dc.b	$E6,$17,$F8,$EF,$D8,$3E,$AF,$F8,$83,$63,$7F,$05,$FC,$4E,$9F,$61
	.dc.b	$AF,$87,$95,$D2,$3F,$8C,$78,$3F,$E9,$A8,$FE,$52,$D7,$F8,$43,$79
	.dc.b	$3F,$CB,$7F,$5D,$79,$7F,$D7,$F3,$6B,$FD,$27,$3F,$A7,$AD,$F4,$76
	.dc.b	$F8,$C9,$5F,$CD,$AB,$58,$7F,$B2,$1B,$1F,$E9,$CC,$3A,$2F,$B0,$7C
	.dc.b	$BF,$F2,$F6,$C7,$F4,$06,$BF,$C7,$DF,$E3,$86,$32,$FF,$39,$B7,$0F
	.dc.b	$F6,$BB,$67,$FD,$56,$F0,$FA,$9E,$DE,$64,$9E,$D1,$E1,$30,$F7,$8C
	.dc.b	$3F,$2B,$EB,$7E,$69,$1F,$37,$D8,$57,$F9,$A7,$D8,$7F,$92,$F1,$FF
	.dc.b	$74,$B6,$3E,$15,$6F,$2F,$F4,$5C,$E6,$4D,$D1,$5F,$D4,$E3,$F0,$93
	.dc.b	$DB,$97,$06,$76,$B1,$FD,$A2,$77,$87,$6A,$D4,$E7,$E3,$FA,$71,$C8
	.dc.b	$C1,$0F,$5B,$97,$DB,$1F,$D9,$7E,$BD,$F6,$14,$93,$FD,$E3,$EC,$1F
	.dc.b	$57,$88,$FE,$B1,$1E,$C7,$5A,$E7,$3F,$F5,$6E,$1D,$27,$24,$3F,$B0
	.dc.b	$39,$79,$FC,$57,$F7,$7B,$FE,$1B,$1E,$1F,$23,$13,$23,$04,$38,$9E
	.dc.b	$7F,$A9,$E4,$47,$D5,$D3,$5F,$4C,$5C,$7F,$D8,$78,$7E,$7A,$5F,$61
	.dc.b	$FE,$97,$7F,$C3,$2D,$7F,$BC,$36,$7F,$D9,$7E,$2B,$E4,$3F,$C2,$13
	.dc.b	$9E,$65,$CA,$FE,$F8,$7F,$7D,$9D,$46,$9A,$FF,$1A,$F8,$1F,$F0,$D4
	.dc.b	$EF,$C6,$BD,$72,$BF,$C5,$B2,$8D,$F5,$E6,$7F,$C1,$D3,$97,$73,$A7
	.dc.b	$8F,$89,$D8,$8F,$F8,$E3,$EC,$3F,$39,$49,$CC,$F4,$7F,$C5,$D3,$FA
	.dc.b	$1F,$B0,$F9,$44,$E1,$FF,$19,$7D,$EF,$C2,$BF,$F2,$1C,$FF,$0C,$F8
	.dc.b	$8F,$D3,$FC,$8F,$F8,$E5,$B3,$AB,$FF,$1D,$53,$92,$92,$7F,$C9,$F4
	.dc.b	$F9,$87,$D1,$E6,$7F,$C5,$D7,$C5,$F0,$DF,$7A,$DC,$7F,$95,$BE,$BD
	.dc.b	$E9,$FF,$A5,$1B,$1F,$E9,$96,$BE,$B3,$EE,$B8,$71,$F6,$2F,$B1,$7C
	.dc.b	$A7,$FA,$86,$30,$DF,$1F,$FC,$BD,$5F,$16,$72,$F5,$1F,$63,$FE,$17
	.dc.b	$EA,$FF,$35,$BF,$0D,$BF,$08,$73,$AD,$AF,$F2,$75,$71,$6C,$67,$B7
	.dc.b	$51,$DA,$F1,$7F,$5E,$FB,$4F,$29,$CB,$F9,$C8,$CD,$72,$7B,$36,$F7
	.dc.b	$3D,$6F,$CA,$7A,$0A,$F5,$6C,$F3,$BE,$37,$DA,$7C,$91,$8B,$CB,$63
	.dc.b	$A5,$FB,$9B,$7E,$36,$9E,$DA,$E2,$F8,$0E,$A5,$9F,$75,$75,$7C,$C7
	.dc.b	$52,$D7,$56,$CF,$53,$BB,$F1,$EB,$BB,$C6,$7D,$D3,$DE,$3E,$31,$97
	.dc.b	$C6,$D3,$C6,$EE,$FE,$7B,$DE,$3C,$27,$77,$C2,$FE,$2A,$BC,$35,$E7
	.dc.b	$3C,$31,$D4,$F0,$B9,$9F,$0B,$63,$C2,$6E,$BC,$2F,$73,$C2,$77,$9F
	.dc.b	$0C,$7D,$98,$E3,$3C,$DE,$F3,$E9,$8F,$19,$99,$EB,$3C,$CE,$AF,$37
	.dc.b	$89,$E1,$8E,$33,$CC,$F4,$96,$3C,$2D,$78,$9D,$63,$8B,$CE,$38,$B4
	.dc.b	$E2,$53,$8C,$F8,$CF,$39,$6F,$F1,$15,$D2,$7B,$57,$46,$9D,$0A,$74
	.dc.b	$BF,$AF,$1E,$7B,$EB,$18,$9F,$14,$62,$7E,$7C,$E6,$79,$CE,$5F,$93
	.dc.b	$19,$8F,$34,$FE,$2E,$FC,$EF,$96,$33,$5F,$7E,$72,$F9,$E3,$37,$F5
	.dc.b	$E7,$33,$E9,$AF,$E2,$FF,$5F,$FC,$23,$89,$F5,$EF,$23,$F7,$B3,$98
	.dc.b	$C5,$70,$3F,$9E,$B8,$BF,$E3,$73,$BA,$FE,$0B,$C3,$FC,$1B,$EC,$EE
	.dc.b	$1F,$C9,$9B,$7F,$C8,$A7,$C5,$F2,$AF,$83,$CE,$BF,$C6,$7E,$23,$C5
	.dc.b	$F5,$5F,$C4,$1B,$97,$F3,$06,$E3,$FC,$B1,$B5,$FC,$AF,$EC,$3F,$3C
	.dc.b	$9A,$7F,$B7,$6B,$FC,$07,$1B,$E2,$FE,$56,$ED,$B3,$C3,$AE,$C6,$BE
	.dc.b	$C9,$F0,$D4,$E3,$F4,$1F,$D3,$B1,$EA,$76,$9D,$B8,$CD,$92,$73,$18
	.dc.b	$9E,$17,$D3,$A7,$B4,$71,$97,$D9,$7F,$6E,$8E,$98,$3E,$C5,$F2,$3F
	.dc.b	$D0,$2B,$57,$D4,$F0,$CE,$AF,$A4,$EA,$FF,$A9,$B6,$3E,$3E,$CF,$6F
	.dc.b	$A7,$85,$27,$8B,$F2,$23,$18,$F7,$C3,$DD,$EF,$7C,$93,$E5,$7C,$FD
	.dc.b	$C3,$D3,$BA,$71,$7C,$61,$58,$BF,$BD,$D5,$F5,$BB,$7E,$BB,$71,$F8
	.dc.b	$7A,$23,$18,$F8,$E3,$13,$AA,$FE,$CB,$F6,$0F,$9F,$FC,$1F,$73,$EC
	.dc.b	$B3,$18,$6D,$FD,$3A,$E0,$FA,$BE,$A7,$39,$7C,$F1,$3A,$70,$DD,$3F
	.dc.b	$C0,$5B,$2B,$8C,$7E,$6F,$C9,$FC,$9E,$3F,$94,$6E,$3E,$E6,$E3,$ED
	.dc.b	$6D,$74,$F6,$FF,$30,$7E,$67,$C2,$B7,$F4,$22,$DF,$CF,$9B,$5F,$E1
	.dc.b	$4C,$FF,$D9,$B7,$F8,$EE,$67,$D9,$7C,$8F,$82,$FD,$4C,$62,$7E,$24
	.dc.b	$65,$FA,$7F,$11,$72,$F8,$33,$71,$FD,$C4,$DA,$FD,$47,$CA,$7E,$BF
	.dc.b	$E3,$BF,$60,$D8,$FC,$3B,$70,$FC,$33,$6F,$F8,$17,$B7,$FB,$A3,$F7
	.dc.b	$7E,$E5,$BF,$C2,$C5,$BF,$C2,$4D,$AF,$F1,$9F,$D9,$32,$FE,$01,$E3
	.dc.b	$FF,$8A,$A3,$18,$FE,$62,$31,$32,$3F,$37,$DD,$71,$FC,$15,$F0,$6D
	.dc.b	$7F,$0B,$9D,$61,$FE,$42,$FB,$27,$91,$FE,$38,$6E,$3F,$BE,$F0,$3F
	.dc.b	$D1,$CF,$F0,$38,$CE,$3F,$8F,$FB,$2F,$A4,$DE,$3F,$E3,$B6,$9F,$F2
	.dc.b	$19,$AC,$7F,$CA,$3F,$5E,$FB,$1F,$F2,$3B,$72,$FE,$1B,$F6,$4E,$2C
	.dc.b	$D6,$B1,$95,$FC,$4B,$EC,$B8,$69,$FA,$1C,$7F,$84,$4F,$F1,$97,$FC
	.dc.b	$FA,$DF,$3F,$F9,$8D,$3E,$67,$C3,$1F,$E5,$8F,$A8,$74,$53,$F9,$BF
	.dc.b	$D9,$7E,$9F,$0F,$E4,$93,$A6,$FC,$FF,$13,$F9,$EF,$D9,$3E,$7F,$13
	.dc.b	$FC,$AF,$DD,$7D,$67,$F9,$99,$B1,$FC,$B3,$C2,$78,$79,$AD,$7C,$CC
	.dc.b	$AD,$7B,$B7,$8F,$F8,$06,$3F,$ED,$09,$ED,$1F,$D3,$BE,$C9,$E1,$F7
	.dc.b	$8F,$F3,$AC,$BE,$6A,$9B,$EF,$B1,$FD,$07,$EC,$BF,$93,$37,$D3,$FE
	.dc.b	$B3,$E8,$7D,$B8,$FF,$5F,$4E,$B0,$FE,$25,$F6,$4F,$67,$FD,$06,$DA
	.dc.b	$FE,$98,$BE,$9B,$33,$AF,$8F,$95,$FD,$47,$EC,$9E,$86,$31,$FE,$3D
	.dc.b	$18,$9D,$4F,$FA,$43,$EC,$BE,$4C,$60,$E5,$E2,$7F,$A5,$2B,$F1,$91
	.dc.b	$FD,$77,$DC,$FE,$7E,$E8,$C7,$B5,$FE,$9A,$8C,$E3,$FC,$B7,$EC,$BF
	.dc.b	$C1,$55,$A3,$FB,$33,$88,$FE,$65,$3C,$3C,$9F,$F5,$7C,$E5,$FF,$58
	.dc.b	$C6,$6B,$FD,$CD,$F0,$5F,$82,$BF,$EB,$5F,$5B,$FB,$07,$F9,$E4,$66
	.dc.b	$FF,$D7,$53,$97,$FD,$77,$19,$AE,$7D,$9F,$E0,$EA,$FE,$E1,$7C,$A3
	.dc.b	$FD,$6B,$5B,$80,$71,$6A,$E2,$F6,$AE,$3C,$3F,$E1,$AF,$B2,$FE,$99
	.dc.b	$8F,$AA,$31,$ED,$EA,$3F,$C1,$58,$FD,$F7,$FC,$13,$4F,$92,$D6,$7E
	.dc.b	$EF,$B0,$F2,$9E,$66,$D7,$28,$B1,$F1,$0B,$9F,$FB,$70,$FD,$84,$FF
	.dc.b	$86,$5B,$1F,$D9,$4F,$7F,$8D,$F0,$3E,$C9,$F2,$BF,$0A,$7F,$C4,$FE
	.dc.b	$37,$FB,$6A,$F8,$9D,$5F,$F6,$EC,$72,$15,$C9,$FF,$1F,$C6,$95,$F7
	.dc.b	$BE,$D7,$E4,$BE,$15,$E1,$AD,$3C,$B0,$AE,$4E,$F7,$CB,$0F,$F8,$01
	.dc.b	$7C,$A7,$C8,$F4,$C3,$B5,$7B,$67,$2F,$1B,$EB,$18,$3C,$07,$63,$86
	.dc.b	$1E,$2A,$E7,$18,$34,$7C,$89,$E6,$FC,$03,$31,$C1,$F5,$5F,$97,$0F
	.dc.b	$4E,$1E,$59,$E0,$52,$99,$C3,$0F,$D1,$3D,$9F,$65,$FC,$6F,$E2,$F8
	.dc.b	$36,$3A,$35,$DF,$82,$9D,$F9,$DF,$E6,$69,$DB,$5E,$AD,$7F,$0D,$5F
	.dc.b	$D2,$73,$E2,$7D,$6C,$5A,$FE,$2F,$4F,$BB,$18,$BF,$B7,$7C,$5F,$92
	.dc.b	$D9,$DD,$71,$AF,$CA,$1C,$6B,$78,$79,$EB,$F1,$37,$C6,$B7,$7C,$6F
	.dc.b	$F7,$B1,$C6,$B7,$7C,$63,$D6,$D9,$DD,$71,$7F,$4A,$DB,$ED,$1F,$C2
	.dc.b	$1A,$E8,$F8,$51,$B9,$E9,$19,$3F,$CA,$D5,$EB,$C1,$AF,$A2,$7F,$D2
	.dc.b	$18,$74,$AD,$4F,$C2,$F5,$3F,$52,$DA,$FF,$4B,$4E,$9F,$E3,$7F,$65
	.dc.b	$FD,$C5,$FF,$4D,$35,$FF,$33,$37,$0F,$0C,$70,$C7,$AB,$CB,$7B,$3F
	.dc.b	$C8,$FE,$5E,$F3,$9D,$D3,$0E,$99,$7F,$0B,$FB,$0F,$F4,$5C,$FF,$AA
	.dc.b	$FE,$C7,$F6,$93,$FC,$3D,$B1,$FC,$CA,$76,$FF,$9E,$DB,$8F,$9A,$13
	.dc.b	$C1,$BA,$78,$AB,$77,$E1,$F4,$B0,$4A,$F1,$7D,$95,$7F,$D1,$18,$7F
	.dc.b	$73,$7C,$DD,$59,$CB,$C1,$8F,$F8,$92,$72,$43,$CD,$87,$C2,$FB,$2E
	.dc.b	$51,$FE,$93,$6D,$7A,$5B,$97,$C1,$F8,$CE,$18,$FC,$AF,$51,$F1,$F4
	.dc.b	$CE,$3F,$ED,$38,$CB,$F2,$F8,$3E,$8E,$BF,$4E,$35,$5C,$FC,$BE,$9F
	.dc.b	$21,$27,$94,$57,$FD,$51,$F6,$5F,$21,$FE,$C6,$D5,$8A,$BE,$FC,$FF
	.dc.b	$8B,$27,$EE,$77,$78,$8C,$FA,$1F,$84,$FB,$2F,$F4,$CC,$FF,$BA,$FA
	.dc.b	$1E,$A7,$E2,$1B,$1F,$EE,$F9,$DB,$FE,$BB,$FB,$2F,$F3,$33,$C7,$E1
	.dc.b	$B8,$1F,$90,$8C,$C6,$FB,$3A,$7C,$94,$8F,$F7,$D4,$FF,$97,$DB,$EE
	.dc.b	$FE,$63,$EC,$5F,$65,$4E,$91,$F9,$C3,$F0,$13,$FE,$CD,$97,$C5,$71
	.dc.b	$35,$D9,$BF,$B2,$E9,$CD,$2F,$FC,$1B,$DB,$F9,$9C,$55,$FA,$8F,$B2
	.dc.b	$FF,$31,$D6,$BB,$7F,$B9,$7D,$97,$FA,$26,$30,$7F,$85,$9C,$3F,$ED
	.dc.b	$F9,$E2,$DF,$7F,$F6,$72,$FA,$2A,$77,$8F,$DB,$FD,$97,$FA,$BE,$BA
	.dc.b	$8A,$FF,$0A,$38,$6B,$FE,$ED,$8E,$0D,$F3,$7B,$FD,$1C,$8E,$73,$FE
	.dc.b	$6D,$7A,$FD,$8B,$E0,$3F,$83,$FD,$93,$EB,$69,$FC,$2A,$75,$5C,$3E
	.dc.b	$CB,$7F,$11,$23,$07,$A3,$BD,$F1,$BF,$C4,$9B,$19,$8F,$E2,$BF,$61
	.dc.b	$88,$73,$7F,$0B,$AC,$3F,$C3,$67,$F5,$B3,$AE,$4F,$F1,$24,$62,$78
	.dc.b	$5F,$73,$AF,$3C,$BF,$01,$F3,$3E,$C9,$F2,$FF,$C9,$9B,$19,$8C,$7A
	.dc.b	$39,$64,$64,$E5,$1C,$7E,$AB,$97,$9B,$27,$B4,$65,$F0,$46,$5F,$51
	.dc.b	$FE,$5E,$B4,$FC,$7F,$AE,$93,$F2,$DF,$37,$D9,$62,$38,$8F,$B2,$FF
	.dc.b	$27,$6F,$FE,$47,$6C,$7F,$9A,$B8,$17,$C7,$7F,$E7,$15,$AA,$E5,$79
	.dc.b	$39,$72,$7F,$4A,$9F,$C2,$4F,$E5,$1A,$F8,$97,$0E,$E6,$4E,$DE,$51
	.dc.b	$CF,$E4,$3A,$71,$CA,$E9,$19,$7B,$46,$5E,$D1,$FD,$73,$EC,$1F,$65
	.dc.b	$49,$3F,$A8,$FB,$17,$D9,$62,$3E,$3C,$7C,$27,$5B,$67,$6F,$EB,$58
	.dc.b	$74,$9C,$90,$F9,$EE,$5E,$7F,$17,$FD,$BA,$FF,$88,$C7,$87,$C7,$C4
	.dc.b	$C8,$C1,$0E,$27,$9F,$EA,$79,$11,$F5,$74,$D7,$D3,$17,$1F,$EE,$18
	.dc.b	$7E,$87,$8B,$FD,$0D,$F6,$5F,$C1,$27,$FB,$A3,$63,$3E,$9F,$F8,$13
	.dc.b	$C6,$7C,$87,$F8,$1A,$B5,$97,$FA,$C2,$7F,$BD,$1F,$DE,$E3,$30,$6C
	.dc.b	$6B,$B3,$95,$F5,$EA,$97,$FC,$6A,$BF,$3A,$DE,$CC,$BF,$9A,$8C,$B6
	.dc.b	$7C,$D3,$E7,$F1,$24,$F3,$7F,$C3,$BF,$65,$F9,$9B,$46,$A7,$A3,$AD
	.dc.b	$3F,$C4,$7F,$64,$F2,$6F,$C1,$CB,$FE,$5F,$39,$6C,$E5,$FD,$7C,$7F
	.dc.b	$3E,$E0,$FF,$10,$3B,$8F,$D9,$FD,$97,$D3,$7F,$DB,$2D,$4D,$8F,$F5
	.dc.b	$04,$FF,$8A,$63,$87,$C4,$7E,$EF,$E4,$FF,$C7,$ED,$BF,$E8,$73,$CC
	.dc.b	$4F,$F8,$F9,$FC,$8C,$78,$BA,$31,$32,$38,$C6,$CC,$3F,$C2,$BE,$C1
	.dc.b	$F6,$2F,$91,$FF,$4F,$46,$1B,$E2,$FF,$5E,$AF,$F6,$A4,$E5,$FF,$53
	.dc.b	$3F,$E0,$AD,$BF,$DB,$7D,$DA,$AF,$ED,$1F,$1B,$FB,$1B,$FA,$5F,$45
	.dc.b	$FE,$5E,$8C,$61,$C6,$7F,$DA,$B7,$FE,$E1,$E1,$5E,$E3,$A6,$3D,$A8
	.dc.b	$74,$E2,$C2,$C3,$A6,$3E,$1C,$79,$D7,$03,$FD,$A4,$E2,$BF,$CB,$87
	.dc.b	$32,$B9,$8D,$F9,$5F,$E7,$66,$9C,$E7,$D1,$19,$6C,$65,$F3,$CF,$F5
	.dc.b	$28,$F3,$B6,$3E,$09,$AF,$8D,$F8,$0F,$0A,$AF,$E6,$E2,$CF,$1B,$F7
	.dc.b	$46,$2B,$2D,$BE,$ED,$BE,$E7,$38,$F1,$CF,$06,$9C,$0B,$1C,$2F,$FD
	.dc.b	$9B,$18,$BF,$0B,$72,$FB,$58,$77,$9C,$DF,$1C,$3B,$36,$FB,$34,$ED
	.dc.b	$5C,$6F,$9C,$F5,$9F,$8C,$D7,$D4,$71,$BE,$E7,$E0,$6F,$BB,$67,$BB
	.dc.b	$67,$BC,$73,$C3,$DB,$1C,$EF,$C2,$D3,$C2,$53,$C2,$7E,$33,$0F,$C5
	.dc.b	$BA,$31,$5F,$98,$69,$4D,$1D,$8C,$9A,$E3,$FF,$58,$5F,$91,$B9,$FF
	.dc.b	$B2,$1C,$DD,$BF,$3C,$61,$FF,$26,$D2,$B5,$7E,$AA,$D3,$FE,$0E,$9E
	.dc.b	$D5,$FB,$3A,$DC,$FC,$EA,$D9,$27,$47,$6B,$DD,$F2,$AE,$D3,$A7,$B4
	.dc.b	$E8,$91,$E6,$BB,$3A,$8E,$86,$4A,$E4,$97,$F1,$23,$2F,$96,$3C,$27
	.dc.b	$EC,$A7,$AC,$66,$3D,$93,$E5,$8C,$5F,$A6,$72,$71,$9E,$A7,$CE,$23
	.dc.b	$93,$C6,$F3,$32,$FF,$53,$AF,$0C,$7C,$38,$FE,$27,$1F,$88,$75,$1F
	.dc.b	$85,$7A,$1F,$C6,$0D,$4F,$F0,$D7,$F9,$41,$82,$47,$E0,$1F,$B3,$F6
	.dc.b	$2F,$A2,$FA,$B1,$86,$CF,$FA,$4A,$74,$F1,$31,$1E,$58,$CC,$F9,$5A
	.dc.b	$FE,$6D,$AF,$E7,$D7,$C3,$78,$BC,$2F,$1B,$6F,$C0,$E1,$FF,$57,$F3
	.dc.b	$FF,$55,$F8,$2C,$61,$FE,$62,$C1,$E2,$FB,$0F,$6F,$C0,$5C,$3D,$51
	.dc.b	$6B,$84,$DD,$95,$E5,$F6,$FE,$2F,$D8,$FC,$66,$C7,$C3,$70,$F9,$3E
	.dc.b	$4D,$F0,$78,$C7,$FC,$FB,$83,$EB,$DC,$1F,$AE,$8C,$36,$1E,$5D,$89
	.dc.b	$8F,$FB,$3A,$B6,$DA,$DB,$67,$A7,$D7,$3E,$16,$9D,$FE,$73,$FB,$E5
	.dc.b	$73,$9D,$36,$76,$F7,$9D,$E5,$EA,$AD,$BF,$E9,$0F,$B1,$7D,$4F,$AF
	.dc.b	$EB,$1F,$23,$FC,$12,$73,$8F,$A6,$72,$FF,$B7,$23,$37,$CE,$FE,$17
	.dc.b	$BB,$C5,$95,$B7,$2D,$3D,$5F,$03,$F0,$0F,$83,$F3,$B6,$32,$4B,$F5
	.dc.b	$46,$5F,$AB,$BB,$EC,$7E,$97,$67,$D3,$6D,$FC,$F6,$D7,$06,$F1,$7E
	.dc.b	$22,$3F,$C2,$3F,$5C,$F9,$DF,$BD,$F6,$2F,$B0,$FE,$4F,$39,$C7,$EE
	.dc.b	$4E,$5F,$F7,$FC,$66,$FF,$D7,$AD,$C3,$F9,$E7,$1B,$EC,$79,$E5,$30
	.dc.b	$E2,$FE,$0A,$31,$87,$BB,$CF,$FC,$76,$7F,$8F,$6D,$FE,$2D,$B3,$B8
	.dc.b	$FC,$74,$FE,$33,$DF,$91,$F8,$9F,$80,$B5,$F8,$59,$B9,$F9,$A6,$D7
	.dc.b	$F5,$DD,$3F,$C3,$93,$FD,$57,$9B,$FC,$5C,$FC,$9C,$13,$AA,$FF,$1D
	.dc.b	$CA,$B7,$1C,$FC,$BF,$96,$7E,$A7,$EC,$2C,$7C,$A8,$B8,$7D,$E9,$B7
	.dc.b	$F2,$BE,$53,$F4,$FF,$21,$FA,$86,$E1,$FA,$56,$DF,$E8,$5B,$5F,$DB
	.dc.b	$79,$7F,$87,$D7,$EC,$FC,$5F,$DD,$7A,$9F,$BB,$64,$64,$FD,$C4,$63
	.dc.b	$0F,$F7,$CF,$D8,$B3,$FD,$CF,$87,$FD,$B2,$31,$7F,$B7,$8C,$4C,$8F
	.dc.b	$DD,$76,$5B,$E7,$37,$3F,$C3,$CD,$AF,$F0,$77,$D8,$BE,$AB,$FB,$CB
	.dc.b	$63,$F8,$77,$7B,$DE,$CD,$7F,$7E,$F4,$3E,$8E,$7F,$E6,$E6,$E4,$BF
	.dc.b	$F0,$56,$1F,$60,$FF,$0E,$4F,$0A,$7D,$87,$F4,$6B,$FC,$03,$6B,$EF
	.dc.b	$FD,$8F,$CC,$A5,$6B,$D1,$FF,$07,$DF,$7B,$ED,$5F,$E8,$08,$C3,$FC
	.dc.b	$C3,$EC,$7A,$39,$CB,$8F,$9B,$2F,$F9,$65,$3C,$FC,$5F,$82,$7F,$C6
	.dc.b	$9E,$B3,$A7,$FC,$3E,$D8,$FE,$69,$F6,$3F,$B3,$7C,$74,$9D,$3C,$CF
	.dc.b	$F4,$AA,$F9,$EE,$67,$F4,$3F,$B1,$7A,$B8,$BE,$7F,$62,$FA,$BF,$F1
	.dc.b	$7B,$63,$FA,$27,$86,$F1,$33,$5A,$F6,$65,$6B,$DE,$BA,$F8,$C6,$3F
	.dc.b	$E9,$7E,$6F,$D9,$36,$BF,$63,$F6,$2F,$13,$C3,$7F,$E4,$09,$7C,$CD
	.dc.b	$3F,$AE,$57,$E2,$E7,$B4,$E8,$F8,$BF,$62,$F4,$FF,$C9,$2D,$8F,$EB
	.dc.b	$DE,$97,$E1,$E3,$15,$E7,$FB,$17,$AF,$FE,$50,$6D,$FD,$05,$F4,$D9
	.dc.b	$9D,$7C,$7C,$AF,$AB,$F6,$1F,$EE,$87,$84,$7F,$96,$A3,$18,$7C,$56
	.dc.b	$F0,$7F,$15,$FB,$1F,$9D,$7C,$BD,$3E,$1E,$74,$8F,$C0,$FC,$3B,$DA
	.dc.b	$C6,$5F,$3E,$31,$3A,$9F,$F3,$4F,$D8,$FF,$82,$23,$07,$E2,$9C,$57
	.dc.b	$19,$DF,$9F,$FE,$70,$BE,$73,$F6,$FE,$07,$F0,$8F,$F8,$43,$EB,$BF
	.dc.b	$75,$8F,$E5,$27,$3A,$FF,$A1,$67,$35,$C7,$B3,$FC,$F7,$5A,$BF,$E6
	.dc.b	$35,$FD,$E2,$B5,$1C,$EB,$67,$F8,$6E,$75,$7F,$CE,$EF,$95,$FF,$BD
	.dc.b	$1F,$AB,$7F,$E1,$BF,$B1,$FE,$B8,$FF,$A8,$63,$33,$FE,$28,$9C,$BC
	.dc.b	$78,$FC,$A7,$F8,$A6,$1F,$25,$8B,$F7,$7C,$47,$A8,$F5,$DD,$8C,$91
	.dc.b	$F9,$06,$4F,$D2,$63,$FB,$28,$FD,$DB,$63,$F5,$07,$EE,$87,$D8,$BE
	.dc.b	$53,$FD,$75,$3B,$3F,$AE,$16,$B3,$E2,$7F,$AA,$2F,$95,$F1,$7F,$AF
	.dc.b	$C6,$5F,$E1,$47,$6B,$E3,$48,$C1,$FE,$C7,$F1,$F1,$7C,$E3,$F9,$4D
	.dc.b	$FF,$85,$63,$11,$E0,$BE,$6F,$23,$F8,$F9,$70,$F8,$13,$C2,$3C,$6F
	.dc.b	$58,$CC,$FF,$A4,$EF,$C2,$F4,$9E,$11,$8A,$F6,$E1,$5E,$F3,$CF,$0F
	.dc.b	$35,$F3,$9F,$14,$61,$F7,$18,$AE,$4F,$D3,$3F,$D4,$87,$C8,$2D,$E1
	.dc.b	$D1,$FE,$EF,$FB,$1F,$E3,$3F,$17,$C1,$B1,$FE,$AD,$6B,$FE,$F3,$6B
	.dc.b	$C7,$CC,$FE,$AA,$ED,$AE,$DA,$EF,$3E,$D1,$C3,$3F,$AD,$9B,$4F,$B9
	.dc.b	$7F,$4E,$30,$FC,$7C,$3F,$AB,$61,$70,$E5,$5C,$0D,$D7,$1B,$E7,$87
	.dc.b	$E0,$F0,$E3,$3F,$7D,$B1,$E2,$9D,$D7,$36,$D7,$E3,$EB,$73,$EB,$8E
	.dc.b	$C7,$DB,$7C,$35,$DE,$FC,$B3,$C5,$AF,$E9,$EF,$AB,$4F,$66,$1C,$E3
	.dc.b	$F5,$79,$F8,$E7,$FB,$6E,$1E,$EA,$FF,$24,$63,$E4,$9F,$D2,$4E,$6B
	.dc.b	$E2,$7A,$7C,$A7,$2D,$7C,$07,$F6,$3F,$B1,$FE,$E2,$FF,$7A,$6B,$9D
	.dc.b	$B8,$47,$F8,$1E,$35,$1D,$71,$71,$BE,$5B,$0D,$AD,$9F,$E0,$6F,$B0
	.dc.b	$FE,$8C,$FF,$82,$3E,$BF,$F5,$99,$FE,$09,$B1,$FB,$F9,$DB,$FD,$E5
	.dc.b	$B8,$7F,$9A,$67,$72,$6C,$EA,$BF,$1A,$FF,$86,$E7,$C5,$E8,$FC,$38
	.dc.b	$C4,$7F,$81,$69,$F6,$39,$AF,$91,$1F,$DB,$A7,$D1,$CF,$97,$99,$82
	.dc.b	$1E,$79,$FA,$1F,$63,$C6,$3F,$C2,$4D,$CB,$FD,$04,$DC,$BE,$17,$C6
	.dc.b	$70,$C3,$EC,$7A,$7F,$EE,$6C,$3F,$C5,$ED,$79,$C6,$BE,$3F,$CF,$BC
	.dc.b	$9D,$C7,$2F,$2B,$9F,$D8,$C9,$E7,$15,$FF,$13,$7D,$8F,$C9,$7F,$98
	.dc.b	$35,$FF,$1D,$4F,$E2,$E3,$FC,$4D,$1F,$6F,$B7,$E3,$99,$F4,$3F,$21
	.dc.b	$F6,$2F,$97,$FC,$97,$3B,$D4,$A6,$A3,$F2,$DF,$62,$F7,$F7,$E4,$4C
	.dc.b	$FF,$C9,$F3,$96,$F8,$3D,$3E,$4E,$46,$0D,$FB,$DF,$A1,$FB,$07,$D8
	.dc.b	$D3,$BC,$7E,$94,$FC,$1F,$5F,$FA,$D6,$5F,$15,$C4,$D7,$6E,$FE,$C7
	.dc.b	$A7,$44,$CB,$FC,$C7,$8A,$7F,$61,$F6,$3F,$E5,$B9,$FF,$33,$FD,$8B
	.dc.b	$C6,$FD,$A3,$63,$FC,$D7,$3B,$7F,$CB,$0B,$DF,$DF,$D1,$49,$DC,$7E
	.dc.b	$F7,$EC,$7F,$D5,$91,$DB,$AF,$A9,$9F,$8E,$D7,$C0,$F8,$3E,$8E,$57
	.dc.b	$F9,$EA,$FF,$CB,$AF,$F9,$F3,$EC,$3F,$CB,$0F,$F9,$FB,$EC,$5F,$5B
	.dc.b	$FC,$2D,$B1,$BF,$84,$FE,$1D,$4F,$B1,$D7,$C4,$CA,$E9,$1F,$C4,$FC
	.dc.b	$1F,$F5,$54,$7F,$A2,$9A,$FF,$9D,$DB,$D5,$FE,$E7,$FA,$EB,$E6,$FF
	.dc.b	$48,$9B,$5F,$E0,$59,$FD,$25,$6F,$A7,$B7,$C7,$4A,$FF,$4B,$D6,$B0
	.dc.b	$FE,$A2,$D8,$FF,$07,$61,$D1,$7D,$8B,$E6,$3F,$95,$36,$3F,$D4,$2D
	.dc.b	$7F,$D1,$CF,$F1,$43,$19,$7F,$BF,$DB,$87,$F6,$06,$CF,$F8,$75,$E1
	.dc.b	$F2,$9D,$BC,$E9,$3D,$A3,$C2,$61,$EF,$18,$7E,$57,$D2,$FC,$D2,$3E
	.dc.b	$6F,$B1,$AF,$FA,$8F,$EC,$7F,$C6,$F8,$FF,$72,$6C,$7C,$2A,$DE,$4A
	.dc.b	$D1,$FD,$2E,$B7,$5F,$BB,$AC,$72,$C8,$FF,$62,$CF,$E1,$9F,$06,$D9
	.dc.b	$7C,$53,$70,$ED,$1F,$E3,$FA,$57,$7F,$91,$F0,$71,$CA,$EB,$19,$7B
	.dc.b	$C6,$5F,$74,$7F,$61,$FA,$F7,$D8,$D2,$4F,$FB,$6B,$EC,$1F,$63,$88
	.dc.b	$FF,$64,$43,$5F,$ED,$4D,$8F,$F7,$0D,$F4,$6D,$EA,$A4,$60,$ED,$F1
	.dc.b	$9C,$BC,$99,$18,$21,$E0,$72,$F8,$A3,$C3,$F5,$32,$3D,$0F,$93,$A7
	.dc.b	$11,$E8,$8B,$C1,$F3,$21,$9F,$E7,$D7,$D8,$FF,$A3,$1F,$F7,$DB,$5F
	.dc.b	$F7,$43,$67,$97,$8A,$F9,$2F,$F0,$64,$E7,$97,$FC,$77,$3F,$DD,$0F
	.dc.b	$EE,$B1,$9B,$FF,$3D,$35,$46,$FE,$13,$9F,$8D,$FE,$17,$87,$55,$25
	.dc.b	$7F,$A0,$F2,$F1,$F5,$62,$3F,$DF,$D5,$C1,$D6,$5E,$4B,$E9,$5D,$7C
	.dc.b	$49,$5A,$3F,$D1,$D6,$8F,$F1,$47,$D8,$FE,$7A,$31,$3B,$3D,$33,$C9
	.dc.b	$63,$F3,$EB,$F4,$BF,$63,$FC,$E6,$78,$3F,$CF,$DD,$E5,$FE,$62,$8C
	.dc.b	$57,$D7,$78,$4E,$8F,$F5,$13,$88,$D7,$99,$FE,$3D,$8E,$35,$CF,$CA
	.dc.b	$FF,$20,$7D,$8F,$DF,$27,$26,$F1,$CB,$FE,$47,$F8,$B7,$7B,$E4,$F0
	.dc.b	$DA,$F8,$17,$CE,$7A,$F4,$C9,$CC,$78,$0C,$3E,$91,$3F,$DB,$1D,$9F
	.dc.b	$2F,$EC,$1F,$62,$F9,$3F,$F6,$07,$C8,$7E,$BE,$95,$A9,$CF,$77,$FA
	.dc.b	$2A,$31,$2A,$E1,$1F,$D6,$A3,$13,$BB,$ED,$1F,$E9,$23,$FD,$9D,$3C
	.dc.b	$9C,$3F,$E1,$7E,$1F,$F5,$7B,$98,$FF,$4C,$EF,$FE,$1D,$E0,$F1,$C7
	.dc.b	$2F,$27,$FB,$E6,$5E,$9A,$D4,$F8,$8D,$61,$C1,$FE,$FD,$EF,$7F,$22
	.dc.b	$AF,$4D,$F2,$CF,$BC,$E5,$F8,$11,$9A,$FF,$73,$61,$FD,$5F,$1E,$8F
	.dc.b	$FB,$DD,$BE,$57,$19,$DF,$65,$7F,$C1,$D1,$6B,$84,$79,$CC,$D7,$4A
	.dc.b	$D7,$C7,$EE,$EC,$F6,$3C,$31,$96,$99,$AE,$97,$FE,$C3,$8C,$67,$F1
	.dc.b	$EB,$85,$69,$B1,$99,$DE,$5F,$4A,$72,$D8,$CD,$7C,$49,$CB,$63,$2F
	.dc.b	$F8,$B6,$32,$D8,$CD,$F9,$63,$CF,$87,$B9,$B3,$EC,$69,$E7,$78,$D7
	.dc.b	$4A,$E5,$87,$E0,$EF,$93,$6B,$93,$6F,$91,$D1,$F8,$0D,$C3,$DC,$D3
	.dc.b	$D8,$53,$D4,$5B,$F0,$C7,$E6,$2F,$C2,$DB,$F0,$B6,$BC,$37,$E3,$3C
	.dc.b	$25,$8F,$0B,$63,$C2,$D7,$C2,$DB,$F0,$C7,$E5,$AF,$C2,$DB,$F0,$B4
	.dc.b	$F0,$DF,$8E,$FC,$2D,$3C,$25,$8F,$0B,$5F,$0B,$6F,$C3,$1F,$9A,$3A
	.dc.b	$9D,$E7,$AB,$67,$AB,$4E,$A5,$3A,$96,$7A,$9F,$B1,$7A,$9D,$E7,$AB
	.dc.b	$67,$F8,$4E,$9F,$E2,$0E,$7F,$EE,$D7,$F8,$F6,$C6,$70,$FB,$F1,$87
	.dc.b	$FA,$94,$ED,$E6,$63,$2E,$18,$7F,$04,$28,$FB,$37,$D8,$E2,$D3,$A9
	.dc.b	$CD,$F6,$8F,$F0,$51,$FC,$80,$A7,$57,$F9,$53,$4B,$F6,$5F,$EB,$B8
	.dc.b	$E4,$7F,$69,$71,$1F,$D1,$6B,$C3,$39,$F8,$2F,$24,$FA,$2B,$79,$7C
	.dc.b	$68,$C7,$C4,$67,$C6,$B8,$75,$6C,$FF,$51,$B5,$7C,$4E,$BF,$E9,$DF
	.dc.b	$09,$AC,$7F,$C4,$D4,$F2,$3C,$07,$87,$E0,$AE,$7E,$98,$B9,$75,$9B
	.dc.b	$7C,$BD,$EF,$47,$07,$A5,$B8,$7C,$17,$0F,$F8,$47,$37,$CA,$BE,$17
	.dc.b	$39,$7F,$80,$71,$7D,$4B,$83,$EE,$C6,$1B,$0F,$0E,$C4,$CB,$FC,$3D
	.dc.b	$5B,$6D,$6D,$B3,$D3,$EB,$DF,$0D,$4E,$FF,$3B,$FD,$D6,$B9,$CE,$9B
	.dc.b	$3B,$7B,$CE,$F2,$F5,$56,$DF,$F0,$27,$D9,$3E,$AB,$D7,$F5,$8F,$92
	.dc.b	$FC,$BC,$E7,$1F,$4C,$E5,$FF,$19,$46,$6F,$9D,$FC,$2F,$7B,$8B,$2B
	.dc.b	$6E,$5A,$7A,$BE,$07,$E0,$1F,$07,$E7,$6C,$64,$97,$EA,$8C,$BF,$57
	.dc.b	$77,$D9,$7D,$2E,$CF,$A6,$DB,$F9,$ED,$AE,$0D,$E2,$FF,$3B,$7D,$73
	.dc.b	$EC,$7E,$EF,$D9,$3E,$C7,$F7,$33,$9C,$7E,$DC,$E5,$FF,$2A,$46,$71
	.dc.b	$FF,$18,$37,$0F,$E1,$DC,$6F,$B2,$E7,$94,$C3,$8B,$F8,$18,$C6,$1E
	.dc.b	$DF,$47,$F1,$B9,$FE,$39,$B7,$F8,$A6,$CE,$E3,$F1,$B3,$F8,$BF,$81
	.dc.b	$91,$F8,$8F,$82,$B5,$F8,$49,$B9,$F9,$A6,$D7,$F1,$4A,$69,$FC,$B6
	.dc.b	$BF,$94,$73,$7F,$8B,$1F,$93,$E6,$7E,$2B,$CF,$F8,$EC,$C3,$F3,$FF
	.dc.b	$19,$71,$F8,$B3,$6F,$F9,$97,$22,$B1,$7F,$E8,$FE,$17,$CA,$FE,$97
	.dc.b	$E4,$7F,$4C,$D9,$FD,$1B,$72,$FC,$FB,$71,$EB,$E3,$FE,$D8,$FD,$BF
	.dc.b	$AF,$6F,$F6,$71,$6F,$F1,$93,$6B,$FD,$75,$F6,$5E,$5B,$FD,$8F,$89
	.dc.b	$FC,$F2,$31,$7F,$AE,$8C,$69,$2B,$F6,$FD,$B6,$BF,$CE,$97,$C3,$0E
	.dc.b	$B7,$FC,$06,$B5,$1F,$EC,$9F,$B2,$7D,$97,$F4,$A6,$E3,$FC,$33,$C0
	.dc.b	$FF,$5B,$BF,$EA,$58,$CE,$5F,$E8,$1F,$4B,$EB,$4F,$DB,$6C,$7F,$AA
	.dc.b	$29,$1F,$D5,$8C,$C3,$EC,$7F,$A0,$D7,$E0,$DB,$1F,$EB,$0F,$B2,$FE
	.dc.b	$71,$3F,$EB,$59,$CC,$F1,$BF,$EC,$30,$7F,$94,$FD,$96,$A3,$15,$C7
	.dc.b	$CD,$D9,$AF,$3E,$C7,$FB,$07,$D7,$73,$7F,$6C,$6C,$7F,$2E,$FB,$2F
	.dc.b	$D7,$3B,$9D,$7D,$04,$8F,$E6,$7F,$64,$F5,$71,$B7,$FA,$9F,$EC,$9F
	.dc.b	$59,$FD,$E1,$B7,$FC,$EB,$C3,$78,$99,$AD,$7B,$32,$B5,$EF,$5C,$7A
	.dc.b	$1C,$CF,$E9,$3F,$65,$F9,$C9,$FF,$6E,$65,$FD,$CA,$7F,$D7,$10,$F9
	.dc.b	$AF,$F0,$4B,$4F,$E9,$75,$F7,$BA,$D1,$FB,$73,$55,$FE,$C1,$FB,$27
	.dc.b	$A5,$FE,$12,$6C,$FF,$54,$F4,$7E,$FC,$62,$BF,$D9,$1F,$64,$F5,$BF
	.dc.b	$C3,$2D,$BF,$F7,$92,$FA,$6C,$CE,$BE,$46,$57,$FB,$DF,$EC,$BF,$DB
	.dc.b	$8F,$08,$FF,$11,$C6,$30,$FF,$6B,$37,$4F,$5C,$7F,$89,$3E,$CB,$E7
	.dc.b	$DF,$3F,$4F,$8F,$97,$23,$FC,$53,$7C,$2F,$FD,$B5,$EC,$65,$FF,$71
	.dc.b	$7C,$4A,$3F,$C6,$6D,$AE,$27,$F1,$FF,$B2,$FF,$01,$38,$AF,$F1,$AC
	.dc.b	$ED,$FF,$76,$4E,$9E,$B5,$FE,$3E,$8C,$5F,$8E,$B8,$3E,$5F,$80,$FC
	.dc.b	$5D,$FF,$83,$7E,$BB,$F7,$2F,$FB,$DE,$33,$3F,$E5,$59,$CB,$C7,$AB
	.dc.b	$FC,$97,$3A,$CB,$E0,$D7,$F7,$5B,$E5,$5A,$24,$FF,$84,$A7,$57,$CE
	.dc.b	$3D,$D5,$FE,$22,$E1,$FF,$0E,$FD,$97,$F5,$99,$FE,$85,$8F,$DF,$7F
	.dc.b	$C3,$71,$9A,$FF,$2C,$CE,$5F,$55,$72,$F9,$5F,$3B,$E8,$F8,$99,$87
	.dc.b	$82,$FF,$41,$8F,$C5,$9C,$E1,$F0,$DF,$7B,$88,$FB,$C7,$EE,$6B,$D2
	.dc.b	$D8,$3F,$E6,$6F,$17,$FC,$DB,$F6,$4F,$8C,$FE,$B9,$3B,$39,$F8,$7F
	.dc.b	$E7,$67,$18,$7F,$9B,$2F,$8D,$70,$AF,$F3,$DE,$1F,$E7,$89,$CB,$F9
	.dc.b	$28,$FC,$CE,$1D,$E3,$FB,$50,$9D,$BE,$69,$DD,$77,$85,$71,$75,$7C
	.dc.b	$1F,$01,$A9,$F2,$46,$63,$B1,$93,$C8,$E6,$FC,$38,$7C,$7B,$ED,$5D
	.dc.b	$1F,$04,$66,$FC,$91,$8C,$3C,$F5,$EC,$BE,$71,$88,$FF,$04,$46,$30
	.dc.b	$F0,$47,$FA,$58,$F4,$96,$30,$E0,$FF,$08,$7D,$97,$F2,$1F,$88,$E0
	.dc.b	$D8,$FF,$4F,$B5,$F9,$AD,$78,$F9,$5F,$82,$76,$D7,$6D,$77,$9F,$68
	.dc.b	$E1,$97,$D1,$CD,$A7,$F8,$8E,$30,$D8,$E3,$7F,$3C,$60,$BE,$73,$FE
	.dc.b	$2B,$39,$CF,$E0,$9A,$F1,$9E,$18,$7E,$19,$B1,$D6,$77,$87,$8A,$35
	.dc.b	$7D,$9F,$F6,$5D,$70,$9D,$47,$1C,$3D,$0E,$AB,$83,$E8,$6D,$F6,$3D
	.dc.b	$13,$C2,$75,$87,$A6,$3F,$57,$9E,$A7,$F8,$5D,$7B,$2B,$F9,$A6,$3E
	.dc.b	$38,$D4,$FE,$DA,$BE,$27,$A1,$CA,$3F,$DC,$73,$FE,$14,$79,$E1,$95
	.dc.b	$F6,$5F,$DE,$5F,$E3,$2D,$7A,$1F,$EE,$C7,$18,$FF,$45,$7E,$A4,$70
	.dc.b	$DB,$C1,$2E,$37,$CB,$E1,$B5,$B3,$F9,$3F,$D9,$7F,$4B,$7F,$94,$FD
	.dc.b	$8F,$EA,$35,$FF,$34,$36,$3F,$DF,$B3,$B7,$CA,$1C,$61,$FD,$66,$78
	.dc.b	$49,$B3,$AA,$FF,$3A,$3F,$E2,$29,$F1,$F9,$DF,$16,$31,$1D,$A9,$F6
	.dc.b	$59,$AF,$95,$1F,$DE,$67,$D3,$CD,$97,$99,$82,$1E,$8B,$EB,$8F,$F6
	.dc.b	$FF,$B2,$E5,$3E,$C6,$C6,$6B,$BF,$C5,$FD,$99,$CE,$7F,$E8,$DF,$47
	.dc.b	$FD,$D1,$87,$F5,$26,$BC,$EF,$F0,$11,$8C,$3E,$36,$DE,$A3,$3F,$1D
	.dc.b	$AF,$27,$7F,$65,$31,$E9,$15,$FA,$3F,$65,$F2,$D0,$70,$B1,$57,$E4
	.dc.b	$67,$FD,$27,$3F,$80,$EE,$F9,$66,$7E,$0B,$F2,$DF,$64,$F8,$7F,$CC
	.dc.b	$73,$BD,$3A,$6A,$3F,$39,$F6,$4F,$95,$DF,$C2,$CC,$FF,$BA,$CE,$5B
	.dc.b	$E0,$F4,$F9,$29,$18,$37,$EE,$FE,$9F,$EC,$5F,$65,$4E,$F1,$FA,$B3
	.dc.b	$ED,$F4,$FF,$AF,$E5,$F1,$3C,$4D,$77,$EF,$EC,$BA,$74,$CC,$BF,$C1
	.dc.b	$38,$A7,$F6,$DF,$63,$FE,$6A,$9F,$F0,$6F,$D9,$7E,$ED,$FD,$0B,$7B
	.dc.b	$7F,$C7,$13,$B9,$FD,$2A,$F7,$FF,$80,$53,$E9,$29,$DA,$3F,$82,$7D
	.dc.b	$97,$FB,$26,$FB,$F5,$F0,$33,$F1,$BA,$F8,$1E,$7F,$4B,$2B,$FC,$43
	.dc.b	$7F,$E6,$F7,$FC,$47,$F6,$3F,$E6,$87,$FC,$4B,$F6,$4F,$AF,$FE,$24
	.dc.b	$D8,$DF,$C2,$7F,$15,$A7,$D9,$6B,$E2,$65,$74,$8F,$E3,$5E,$0F,$FA
	.dc.b	$8E,$3F,$C6,$2D,$7F,$82,$B7,$8F,$FC,$B9,$F6,$17,$AF,$FD,$E9,$36
	.dc.b	$BF,$33,$3F,$A5,$9F,$F7,$C3,$63,$F3,$B3,$B7,$F2,$5F,$1F,$FC,$AE
	.dc.b	$75,$06,$CE,$FE,$C9,$F3,$1F,$CC,$5B,$5F,$E4,$86,$BF,$C7,$1F,$E3
	.dc.b	$26,$31,$FF,$3A,$37,$1F,$D7,$CA,$30,$FE,$C9,$E3,$F5,$9D,$FC,$E9
	.dc.b	$3D,$E3,$C4,$61,$F0,$46,$1F,$97,$F4,$DF,$38,$8F,$9F,$EC,$AB,$FC
	.dc.b	$CB,$EA,$FF,$C9,$78,$F9,$1B,$1F,$22,$B7,$97,$FA,$36,$73,$26,$E8
	.dc.b	$AF,$EA,$D1,$F8,$59,$ED,$C9,$83,$3B,$58,$FF,$38,$CE,$F0,$ED,$5A
	.dc.b	$9C,$FC,$8F,$4E,$29,$18,$21,$EB,$72,$FB,$63,$FB,$47,$D8,$BE,$C6
	.dc.b	$92,$7F,$D0,$3F,$62,$FB,$2C,$47,$F5,$68,$F6,$3A,$D7,$39,$FF,$AB
	.dc.b	$F0,$E9,$39,$21,$FD,$79,$CB,$CF,$E3,$3F,$BD,$5F,$F0,$E8,$F0,$F9
	.dc.b	$18,$99,$18,$21,$C4,$F3,$FD,$4F,$22,$3E,$AF,$22,$BE,$98,$B8,$FF
	.dc.b	$B1,$70,$FC,$FF,$D9,$3C,$3F,$F0,$63,$6F,$FD,$46,$D7,$FB,$BF,$8A
	.dc.b	$F9,$2D,$D7,$F8,$4E,$3F,$BB,$9F,$DE,$2B,$86,$1F,$ED,$36,$F3,$FE
	.dc.b	$82,$77,$F0,$9C,$B1,$FF,$58,$C3,$AF,$A1,$23,$3D,$75,$F1,$C5,$C7
	.dc.b	$FD,$C1,$5F,$E1,$69,$F3,$F8,$9E,$52,$3F,$EB,$EF,$B2,$FC,$CC,$A7
	.dc.b	$33,$C0,$D6,$5F,$E2,$FF,$B2,$FE,$65,$1F,$EC,$67,$F9,$C4,$7E,$E9
	.dc.b	$B1,$FD,$A9,$DB,$FE,$1E,$6E,$1F,$EF,$57,$70,$69,$A9,$FE,$0B,$F6
	.dc.b	$5F,$C5,$09,$C5,$E3,$AF,$F2,$14,$7F,$1D,$F8,$BF,$D9,$7C,$9F,$F9
	.dc.b	$25,$B3,$FE,$DC,$9E,$53,$C3,$9A,$4E,$63,$99,$87,$FC,$6B,$3F,$95
	.dc.b	$6C,$7F,$94,$7E,$C7,$FA,$24,$7F,$AE,$FA,$7F,$8D,$C3,$EC,$9F,$27
	.dc.b	$FE,$C5,$9C,$C6,$2B,$81,$FE,$10,$AE,$7F,$21,$FA,$E8,$46,$2B,$8D
	.dc.b	$F3,$9E,$17,$CE,$F8,$5E,$E7,$35,$FE,$AF,$8F,$23,$7C,$4F,$EE,$53
	.dc.b	$D0,$FF,$6A,$9E,$07,$FC,$E1,$5D,$1B,$87,$F3,$8F,$43,$FC,$39,$BF
	.dc.b	$8E,$F8,$7C,$75,$9F,$F7,$74,$DE,$83,$9F,$26,$DF,$C2,$3F,$BF,$4E
	.dc.b	$32,$F8,$57,$CA,$74,$D8,$CD,$6F,$1F,$25,$DB,$FF,$1D,$DF,$F5,$8C
	.dc.b	$3C,$07,$97,$E3,$B5,$3F,$E0,$DC,$2D,$6E,$3B,$C6,$27,$2C,$AE,$73
	.dc.b	$E4,$9C,$3C,$23,$38,$FA,$A7,$2F,$A2,$33,$8F,$CB,$9C,$BF,$36,$ED
	.dc.b	$FA,$A6,$BE,$7A,$E2,$E5,$AE,$4F,$11,$E5,$73,$87,$B7,$0F,$14,$E4
	.dc.b	$DB,$82,$61,$E5,$AE,$38,$78,$AF,$FD,$C7,$18,$3A,$E1,$99,$D3,$63
	.dc.b	$26,$F0,$EF,$7D,$5A,$7D,$8A,$E3,$3B,$C3,$C7,$3E,$F3,$67,$BD,$CB
	.dc.b	$FE,$47,$BF,$03,$AC,$3C,$ED,$BE,$AD,$BF,$2B,$4E,$15,$CA,$FA,$DF
	.dc.b	$26,$DF,$26,$DF,$26,$DF,$28,$EB,$CB,$FA,$08,$F3,$47,$FA,$42,$31
	.dc.b	$6C,$FF,$4B,$18,$C6,$D7,$C7,$8C,$3C,$4F,$DA,$4E,$AF,$E3,$4F,$7F
	.dc.b	$B2,$7D,$8E,$65,$18,$BE,$51,$FA,$83,$51,$97,$FD,$AE,$DA,$E9,$5C
	.dc.b	$F1,$FF,$54,$DF,$2B,$EB,$39,$7E,$01,$FC,$24,$A7,$78,$FE,$22,$FF
	.dc.b	$4E,$6C,$F4,$8F,$F7,$3B,$6B,$E5,$7D,$97,$FA,$9E,$BC,$13,$97,$FC
	.dc.b	$9B,$0F,$89,$F8,$0B,$C3,$B1,$86,$CF,$A6,$78,$3F,$13,$A3,$F9,$DF
	.dc.b	$C2,$6B,$2F,$F7,$CD,$3C,$5E,$A7,$6F,$80,$B8,$73,$8B,$5C,$26,$E9
	.dc.b	$EA,$F7,$3C,$BB,$79,$9B,$87,$8D,$C3,$E3,$F9,$47,$C1,$E3,$16,$1F
	.dc.b	$88,$FA,$AF,$75,$FD,$C8,$46,$27,$73,$F9,$18,$78,$1F,$89,$AD,$F0
	.dc.b	$59,$E1,$1D,$63,$13,$97,$87,$D5,$3E,$12,$9C,$FE,$77,$FB,$1E,$3E
	.dc.b	$39,$D3,$67,$6F,$9E,$77,$97,$98,$F6,$D7,$07,$03,$EC,$9F,$55,$E9
	.dc.b	$FA,$C7,$C9,$FE,$7E,$73,$8F,$C3,$9C,$BF,$97,$8C,$DF,$46,$EB,$F2
	.dc.b	$FD,$CE,$58,$CB,$E0,$FC,$58,$C5,$F9,$E9,$93,$DF,$F8,$A7,$C6,$F9
	.dc.b	$EB,$8F,$C3,$8B,$7C,$69,$58,$BF,$B9,$F6,$5F,$4F,$B3,$EA,$37,$1F
	.dc.b	$BD,$C6,$8C,$63,$E3,$8C,$4E,$AB,$F8,$17,$D7,$3E,$C7,$F0,$1F,$64
	.dc.b	$FB,$1C,$C6,$1B,$7F,$BD,$AE,$0F,$DB,$FB,$2C,$E5,$F5,$27,$2D,$BF
	.dc.b	$C2,$CE,$B4,$FC,$B7,$9D,$F8,$DC,$7F,$1C,$DC,$7D,$6D,$C3,$D4,$DA
	.dc.b	$FE,$4B,$BB,$DE,$FC,$89,$F9,$2F,$7E,$E1,$F8,$F8,$B8,$7C,$D9,$B7
	.dc.b	$FC,$CF,$4F,$F0,$03,$FC,$C3,$9B,$FC,$3B,$5E,$BF,$2B,$F1,$9F,$27
	.dc.b	$F2,$19,$97,$D0,$F8,$BB,$5B,$8F,$8D,$3F,$A9,$9C,$E1,$E8,$F9,$5F
	.dc.b	$D3,$FC,$87,$EA,$1B,$7F,$A5,$6E,$1F,$83,$6D,$7E,$F7,$C6,$FD,$A1
	.dc.b	$FB,$5F,$5A,$DF,$E0,$A2,$DF,$E0,$66,$D7,$F5,$1F,$B2,$65,$FB,$8F
	.dc.b	$13,$FA,$24,$63,$1F,$DB,$46,$26,$47,$EE,$7B,$AE,$3F,$B2,$BE,$0D
	.dc.b	$AF,$E0,$73,$AC,$3F,$B0,$7D,$8B,$EC,$BF,$A9,$37,$1F,$E2,$DE,$07
	.dc.b	$FA,$85,$FE,$95,$19,$C7,$F7,$FE,$87,$D4,$C3,$FB,$53,$6B,$FB,$55
	.dc.b	$FF,$69,$7F,$A9,$46,$68,$3E,$C3,$FA,$43,$FC,$29,$B7,$FD,$C3,$EC
	.dc.b	$BF,$9E,$5F,$F5,$B9,$CB,$CA,$B9,$61,$FD,$AE,$4F,$F2,$9F,$B2,$DC
	.dc.b	$F8,$27,$35,$CF,$CC,$E0,$6F,$CE,$B1,$FD,$A7,$D5,$73,$7F,$77,$6C
	.dc.b	$7F,$3B,$FB,$27,$83,$9A,$D7,$CF,$CA,$FE,$81,$F6,$4F,$53,$1A,$FF
	.dc.b	$28,$FB,$27,$D6,$7F,$80,$DB,$7F,$D1,$FC,$2F,$D6,$3F,$DE,$E3,$35
	.dc.b	$C7,$D8,$E0,$7F,$47,$A7,$BB,$8B,$E1,$5D,$4F,$E9,$9F,$64,$F0,$71
	.dc.b	$3F,$E0,$71,$F3,$3F,$E1,$56,$9F,$D6,$67,$F0,$5D,$3F,$E2,$98,$CD
	.dc.b	$FF,$42,$FB,$2F,$E5,$D1,$FE,$11,$6C,$7F,$33,$F3,$BF,$04,$A3,$35
	.dc.b	$FE,$33,$FB,$27,$AD,$FE,$25,$6E,$1F,$E1,$B5,$F4,$D9,$9D,$7C,$7C
	.dc.b	$AF,$F1,$07,$D9,$3D,$0C,$67,$FD,$5A,$31,$3A,$9F,$F1,$97,$D9,$7C
	.dc.b	$EA,$D4,$7F,$8D,$6B,$F2,$31,$C7,$CB,$FF,$1C,$7C,$37,$F4,$4C,$3B
	.dc.b	$C6,$3D,$89,$1F,$E3,$1F,$B2,$FF,$02,$BF,$E3,$3A,$E1,$7F,$DA,$27
	.dc.b	$2F,$F6,$A8,$CE,$FF,$E5,$59,$CB,$FE,$57,$8C,$D7,$E8,$7E,$03,$F0
	.dc.b	$73,$FE,$56,$FA,$EF,$DC,$61,$FD,$D2,$73,$5F,$E5,$D9,$CD,$7F,$98
	.dc.b	$67,$35,$FB,$46,$C2,$7F,$BA,$CF,$3E,$BF,$F2,$C5,$70,$83,$C2,$77
	.dc.b	$5D,$6B,$7C,$3F,$E1,$BF,$B2,$FE,$A7,$3E,$18,$C7,$B3,$AC,$BF,$CB
	.dc.b	$18,$FD,$E7,$FC,$51,$6F,$94,$D6,$7E,$DF,$88,$F1,$9E,$46,$D7,$28
	.dc.b	$B1,$F0,$CB,$97,$FA,$04,$FD,$A4,$FF,$99,$DB,$1F,$E1,$C3,$F8,$BF
	.dc.b	$8D,$FC,$63,$EC,$9F,$17,$FC,$46,$B6,$73,$F0,$FF,$CF,$8E,$30,$FF
	.dc.b	$7D,$DF,$1A,$E1,$5F,$E8,$1C,$3A,$8C,$3A,$BF,$E9,$38,$FF,$7A,$C6
	.dc.b	$23,$F9,$05,$73,$75,$75,$C9,$CE,$FB,$C6,$63,$BD,$73,$C3,$DE,$F5
	.dc.b	$8C,$D7,$07,$91,$FC,$BC,$5F,$7A,$E3,$3E,$E9,$D9,$FE,$62,$C3,$D3
	.dc.b	$19,$BF,$04,$62,$7A,$5F,$38,$C4,$7A,$A7,$78,$7A,$63,$C2,$7F,$A5
	.dc.b	$8A,$7F,$A6,$2B,$0E,$0F,$E8,$DF,$65,$FC,$73,$E2,$BF,$D3,$AD,$8F
	.dc.b	$60,$70,$FB,$DA,$F2,$F2,$FF,$CF,$AF,$08,$46,$23,$83,$5E,$19,$F8
	.dc.b	$23,$8E,$3F,$6F,$36,$9F,$D6,$E3,$0D,$8E,$37,$F3,$30,$EF,$87,$DF
	.dc.b	$9E,$33,$6B,$83,$FD,$BA,$B6,$F1,$C3,$C3,$5F,$DA,$4E,$31,$F6,$1B
	.dc.b	$87,$E3,$EB,$8B,$6B,$8D,$7F,$7A,$AD,$9C,$EF,$C3,$1F,$2E,$76,$DC
	.dc.b	$3C,$CC,$3C,$8D,$3F,$DE,$95,$E4,$6D,$7E,$BF,$3F,$87,$87,$C6,$BF
	.dc.b	$F6,$F5,$FF,$BD,$DB,$E8,$ED,$F3,$5D,$5F,$65,$FD,$CD,$FF,$73,$35
	.dc.b	$FF,$4D,$B7,$16,$3F,$E0,$F8,$FB,$F3,$C3,$6E,$D5,$FC,$9F,$E5,$AF
	.dc.b	$31,$F0,$5F,$3C,$3F,$C3,$1F,$61,$FE,$8C,$C7,$FD,$ED,$F6,$3F,$B3
	.dc.b	$9F,$F0,$E3,$63,$FC,$4F,$3B,$7F,$D5,$ED,$C7,$CB,$1F,$E2,$C9,$E0
	.dc.b	$9B,$3B,$AF,$0F,$9D,$F0,$E3,$FC,$3B,$5F,$E3,$4B,$E6,$A3,$2F,$FA
	.dc.b	$FA,$DF,$65,$C6,$FE,$64,$7F,$74,$9F,$5F,$4E,$5E,$98,$F8,$1C,$90
	.dc.b	$F5,$D7,$D5,$FB,$2E,$51,$FE,$CC,$6E,$5E,$46,$D7,$C6,$F8,$BE,$18
	.dc.b	$7D,$DF,$47,$FD,$D3,$87,$80,$FF,$27,$B8,$7A,$46,$FE,$3B,$D3,$BC
	.dc.b	$9E,$02,$B8,$65,$F1,$17,$95,$DF,$EC,$A5,$70,$9C,$B6,$3F,$21,$F6
	.dc.b	$5F,$46,$3F,$CC,$2D,$7B,$46,$BC,$FF,$97,$DD,$27,$C7,$5F,$E3,$77
	.dc.b	$CB,$F6,$4F,$85,$FC,$E7,$47,$F9,$02,$7D,$2D,$7F,$DE,$B3,$B7,$DB
	.dc.b	$F6,$2F,$B2,$E3,$F0,$B2,$73,$1F,$A7,$9F,$CF,$C7,$FB,$FA,$DF,$25
	.dc.b	$E0,$35,$E0,$EF,$DD,$FD,$77,$D8,$BE,$CA,$99,$C3,$E6,$CF,$E3,$39
	.dc.b	$3F,$1F,$4F,$89,$E2,$7E,$DE,$75,$D5,$FE,$7F,$5F,$65,$DB,$A6,$63
	.dc.b	$D2,$3F,$7D,$F6,$5F,$E6,$C8,$FB,$7F,$65,$FC,$15,$67,$F9,$E8,$C3
	.dc.b	$FA,$09,$E1,$3A,$C7,$FD,$1D,$0F,$7F,$F8,$55,$BE,$8A,$9E,$08,$FE
	.dc.b	$1D,$F6,$5F,$EB,$CC,$3C,$3D,$3C,$0C,$FC,$6E,$BE,$03,$9F,$D1,$CA
	.dc.b	$FC,$8D,$FF,$9C,$9F,$C9,$FD,$8B,$E6,$BF,$8D,$7D,$93,$EB,$69,$FC
	.dc.b	$76,$75,$5C,$3E,$CB,$7F,$11,$23,$07,$E7,$BE,$23,$FD,$4E,$DF,$8F
	.dc.b	$FE,$4D,$18,$8F,$84,$7F,$AE,$3E,$C3,$1C,$DF,$C1,$A6,$F3,$7F,$AC
	.dc.b	$E7,$F4,$ED,$E8,$E7,$84,$C3,$F5,$B5,$A7,$A4,$E5,$EE,$3E,$C9,$F2
	.dc.b	$FF,$CE,$5B,$1F,$B6,$6B,$C1,$FE,$42,$63,$2F,$E3,$6D,$C7,$FD,$90
	.dc.b	$D9,$FF,$69,$BC,$3E,$AF,$B7,$9B,$27,$B4,$78,$4C,$3D,$E3,$0F,$CB
	.dc.b	$FA,$6F,$92,$47,$CD,$F6,$55,$F6,$FD,$97,$F9,$47,$1F,$F6,$D3,$63
	.dc.b	$FD,$CB,$BA,$F8,$67,$FC,$D2,$B5,$46,$D6,$E3,$F8,$CC,$7E,$15,$BC
	.dc.b	$7F,$D8,$63,$11,$E2,$DF,$13,$23,$67,$28,$E1,$F3,$DC,$78,$A4,$F6
	.dc.b	$8C,$BE,$08,$CB,$EE,$8F,$EE,$1F,$60,$FB,$2A,$49,$FE,$5F,$F6,$2F
	.dc.b	$B2,$C4,$7E,$0A,$3D,$8D,$DB,$3B,$7F,$56,$C3,$A4,$F2,$9F,$EF,$D5
	.dc.b	$C2,$B5,$86,$3E,$37,$3E,$3C,$C3,$C3,$19,$7C,$71,$97,$C5,$1D,$3E
	.dc.b	$A6,$47,$A1,$EF,$D1,$88,$F4,$45,$C3,$FB,$5E,$3F,$9F,$97,$D9,$7F
	.dc.b	$A6,$DF,$E9,$ED,$7F,$42,$D9,$FE,$E7,$E2,$BE,$4F,$FC,$35,$39,$E6
	.dc.b	$5C,$AF,$F0,$09,$FE,$02,$9D,$46,$9A,$FF,$19,$F8,$5F,$F1,$24,$EF
	.dc.b	$C6,$BD,$72,$BF,$B1,$CA,$37,$D7,$99,$FD,$BC,$E5,$DC,$E9,$E3,$E2
	.dc.b	$76,$23,$FD,$BB,$EC,$BF,$37,$49,$CC,$F4,$7F,$C7,$13,$FA,$0F,$B2
	.dc.b	$FE,$65,$3F,$DD,$9F,$E7,$3A,$7C,$28,$46,$1F,$8A,$F0,$7F,$86,$B7
	.dc.b	$6F,$F0,$FB,$B8,$E3,$F0,$FF,$E4,$78,$D5,$75,$FB,$2F,$B4,$4E,$3D
	.dc.b	$5F,$F9,$36,$3F,$8E,$7C,$5F,$ED,$7E,$57,$FC,$A4,$D9,$FF,$06,$CF
	.dc.b	$29,$E1,$CF,$27,$31,$CC,$C3,$FC,$92,$7F,$B3,$3A,$3F,$C2,$FF,$63
	.dc.b	$FD,$01,$FF,$1C,$87,$18,$FF,$AE,$5B,$1F,$1D,$AF,$E2,$9E,$5F,$1D
	.dc.b	$FB,$05,$F6,$4F,$94,$FF,$71,$F6,$F1,$8C,$5F,$FB,$12,$3A,$79,$BF
	.dc.b	$ED,$67,$55,$C2,$7B,$57,$0C,$3B,$36,$B8,$CE,$CF,$54,$76,$7D,$4D
	.dc.b	$7D,$5E,$37,$F7,$D7,$B9,$FE,$BC,$F4,$7F,$DE,$F9,$FE,$98,$FF,$6D
	.dc.b	$38,$8F,$8E,$D8,$CF,$57,$F9,$B1,$A7,$8E,$38,$63,$FC,$51,$FE,$AB
	.dc.b	$19,$9F,$F3,$9C,$E7,$CA,$B9,$7E,$56,$F8,$B6,$33,$87,$EE,$A7,$2D
	.dc.b	$8C,$BF,$E1,$1B,$B3,$E0,$8F,$A0,$71,$8C,$CE,$9B,$87,$94,$F5,$46
	.dc.b	$5A,$66,$79,$9F,$E1,$F8,$D3,$8A,$F0,$61,$F0,$6F,$AC,$E5,$B1,$9B
	.dc.b	$EB,$39,$6C,$66,$3A,$5F,$68,$F9,$13,$E3,$6C,$F1,$37,$19,$C3,$E5
	.dc.b	$36,$3E,$43,$5F,$53,$70,$F4,$CF,$D3,$C3,$CE,$D9,$F2,$B6,$BB,$9F
	.dc.b	$86,$C3,$F0,$8F,$80,$C4,$FE,$35,$C3,$F8,$E8,$DC,$79,$0D,$93,$0F
	.dc.b	$69,$F8,$BF,$8E,$F1,$C6,$E3,$37,$F8,$99,$CC,$FB,$0F,$2C,$F1,$BF
	.dc.b	$4C,$71,$AF,$34,$E6,$37,$18,$23,$F9,$1A,$EB,$8F,$C6,$8E,$46,$6F
	.dc.b	$AB,$E1,$BF,$2B,$93,$A9,$C0,$D5,$FA,$67,$8D,$75,$9C,$9C,$6B,$9B
	.dc.b	$82,$42,$FC,$B8,$79,$9B,$7D,$1C,$34,$95,$FC,$0E,$3F,$92,$BE,$73
	.dc.b	$F8,$1B,$85,$F6,$4F,$B1,$C3,$6B,$FA,$74,$FF,$4F,$8C,$38,$38,$CF
	.dc.b	$03,$FD,$EB,$D1,$FE,$EC,$9D,$BF,$EE,$EC,$7F,$10,$E6,$77,$7F,$C1
	.dc.b	$EE,$9F,$DF,$AF,$8C,$64,$FF,$6B,$B8,$8F,$E1,$47,$AE,$FF,$28,$FF
	.dc.b	$20,$30,$7B,$FE,$CB,$FD,$B7,$5F,$EE,$05,$E1,$F8,$25,$1A,$7F,$CE
	.dc.b	$AE,$A7,$87,$C4,$6E,$35,$92,$7F,$C2,$BD,$1F,$D2,$7D,$FD,$5E,$FC
	.dc.b	$69,$5E,$1F,$7A,$E5,$DA,$6E,$3C,$E3,$FC,$3F,$92,$B3,$7F,$E0,$18
	.dc.b	$7B,$9E,$3D,$FC,$8D,$9F,$73,$87,$FC,$53,$87,$5D,$1F,$2A,$F8,$3D
	.dc.b	$67,$FE,$A2,$A7,$D8,$7C,$18,$C3,$63,$EF,$AF,$03,$EC,$C7,$0E,$1B
	.dc.b	$3B,$CB,$C2,$DB,$FF,$6A,$BC,$63,$CA,$6A,$79,$FD,$77,$FB,$CA,$33
	.dc.b	$5F,$ED,$C8,$C3,$E7,$F9,$EF,$EF,$13,$AC,$15,$A7,$F2,$AD,$8F,$86
	.dc.b	$DA,$FF,$73,$D7,$27,$E1,$46,$57,$D9,$3E,$B3,$DF,$F5,$CF,$91,$FC
	.dc.b	$54,$E7,$1F,$14,$E5,$F4,$DF,$2B,$D9,$F1,$BD,$CB,$3A,$BE,$55,$FE
	.dc.b	$C6,$6E,$7F,$42,$B7,$3E,$6F,$0F,$E3,$9F,$23,$E7,$EE,$1F,$0F,$4F
	.dc.b	$F7,$63,$67,$4D,$3E,$46,$EF,$B2,$FB,$3D,$5F,$69,$B3,$A6,$E3,$F8
	.dc.b	$46,$E7,$F9,$98,$7D,$83,$E7,$FF,$0B,$F6,$4F,$B2,$FD,$84,$E7,$1F
	.dc.b	$B1,$39,$7E,$3C,$66,$FF,$DC,$ED,$C3,$ED,$FD,$5F,$2C,$65,$F0,$7F
	.dc.b	$13,$18,$BF,$CC,$F9,$BF,$94,$CB,$F2,$AD,$BF,$82,$D9,$DC,$7C,$49
	.dc.b	$FC,$9F,$BB,$23,$E6,$7B,$D6,$BF,$0F,$37,$2F,$C7,$CD,$AF,$F3,$C6
	.dc.b	$7F,$E1,$B9,$FE,$4F,$CC,$F9,$3F,$87,$E4,$7E,$31,$B7,$F0,$A3,$F5
	.dc.b	$7F,$15,$8C,$FE,$1C,$5C,$3F,$67,$36,$BF,$3B,$F2,$BF,$AA,$F9,$AF
	.dc.b	$D5,$B6,$3F,$08,$DC,$3F,$3E,$DB,$FD,$EF,$8F,$FB,$A3,$F7,$7E,$A5
	.dc.b	$BF,$C3,$C5,$BF,$E0,$B3,$6B,$F0,$5F,$64,$CB,$F7,$7E,$1F,$F2,$F8
	.dc.b	$C6,$3F,$95,$8C,$4C,$8F,$CC,$77,$DC,$58,$74,$AC,$3F,$BC,$9B,$5F
	.dc.b	$EA,$5F,$B1,$7D,$97,$F3,$D6,$C7,$F0,$5F,$03,$FD,$7C,$FE,$D2,$33
	.dc.b	$8F,$68,$FF,$57,$7B,$3F,$55,$FD,$CB,$63,$F9,$35,$23,$FA,$51,$9C
	.dc.b	$7F,$45,$F6,$0F,$98,$FE,$9C,$DC,$3F,$8B,$7D,$93,$87,$35,$AC,$65
	.dc.b	$7F,$1B,$FB,$2E,$19,$F9,$4F,$F7,$1E,$2A,$EB,$E6,$CB,$E1,$E8,$5B
	.dc.b	$E0,$3D,$57,$3F,$F6,$16,$D7,$F2,$9F,$B2,$78,$39,$AD,$7C,$FC,$AF
	.dc.b	$E5,$DF,$64,$F5,$71,$97,$6F,$B2,$7D,$67,$F6,$D6,$DF,$F3,$5F,$0D
	.dc.b	$E2,$66,$B5,$EC,$4A,$D7,$BB,$71,$E8,$73,$3F,$A7,$FD,$93,$C3,$C5
	.dc.b	$7F,$2E,$1F,$33,$FD,$F5,$A7,$F4,$59,$FB,$FD,$9F,$C5,$E3,$37,$F1
	.dc.b	$BE,$CB,$F9,$3B,$7D,$1F,$EB,$BE,$87,$DD,$AF,$F7,$95,$6A,$FF,$90
	.dc.b	$7D,$93,$D7,$FF,$06,$36,$BF,$A0,$AF,$A6,$CC,$EB,$E3,$E5,$7F,$46
	.dc.b	$FB,$17,$A5,$8D,$7E,$CC,$62,$75,$3F,$E1,$9F,$B2,$F9,$71,$83,$97
	.dc.b	$8B,$FE,$1D,$AF,$C7,$47,$F5,$2F,$87,$7B,$59,$AF,$F1,$14,$62,$75
	.dc.b	$3F,$E2,$5F,$B2,$FF,$03,$46,$0F,$EB,$6E,$32,$FC,$6D,$71,$7F,$1F
	.dc.b	$19,$DF,$FA,$24,$E5,$FE,$8D,$F0,$1F,$84,$AF,$F0,$87,$D7,$FE,$D1
	.dc.b	$FC,$BC,$66,$7F,$C9,$B3,$97,$8F,$57,$F8,$E6,$75,$8F,$F8,$7A,$BF
	.dc.b	$B8,$5F,$2A,$D1,$27,$FC,$0D,$3A,$BE,$71,$FD,$DA,$FE,$B3,$FE,$0B
	.dc.b	$FB,$2F,$EA,$0B,$09,$3A,$7F,$C5,$13,$A7,$3C,$5E,$53,$FC,$4B,$0F
	.dc.b	$93,$C5,$F7,$F8,$8F,$51,$EB,$BB,$19,$23,$F1,$0C,$9F,$BC,$C7,$F6
	.dc.b	$51,$FE,$0E,$6C,$7F,$00,$3D,$1E,$3F,$A7,$EC,$5F,$17,$EB,$9F,$E5
	.dc.b	$BE,$27,$F9,$8E,$F8,$9F,$DE,$CB,$87,$FA,$07,$0E,$57,$C6,$73,$7F
	.dc.b	$E8,$89,$CB,$FE,$50,$BF,$F2,$E6,$1E,$08,$D0,$70,$F8,$B0,$E2,$F9
	.dc.b	$6A,$DF,$B2,$B7,$3E,$C3,$96,$1E,$3C,$3B,$BE,$78,$C9,$E3,$C3,$AB
	.dc.b	$C6,$BC,$37,$D6,$B8,$46,$2F,$9D,$7A,$23,$17,$CA,$3F,$96,$1D,$8F
	.dc.b	$51,$4E,$D5,$E5,$8E,$18,$7F,$AA,$E7,$DC,$53,$99,$97,$5F,$65,$FC
	.dc.b	$73,$E2,$2F,$80,$69,$2B,$FD,$6D,$3F,$E8,$9C,$F8,$F9,$9F,$DA,$A7
	.dc.b	$2D,$7F,$D7,$F3,$B7,$33,$F4,$B2,$FF,$6A,$45,$D7,$F1,$1A,$7D,$8B
	.dc.b	$E0,$DC,$3B,$3F,$72,$F8,$DF,$C5,$6C,$6F,$0F,$BF,$7C,$A7,$F2,$86
	.dc.b	$8B,$87,$81,$FC,$C4,$6E,$FF,$68,$D7,$F3,$B3,$B7,$F7,$13,$CD,$DE
	.dc.b	$1E,$68,$E6,$D8,$DC,$F1,$79,$9E,$68,$D3,$83,$FB,$6C,$FB,$AF,$FC
	.dc.b	$C4,$DD,$3A,$C7,$C9,$9F,$F4,$1C,$FF,$82,$5B,$E9,$FE,$AE,$BD,$F1
	.dc.b	$FE,$88,$FB,$2F,$ED,$CD,$DF,$4F,$F1,$38,$CB,$FD,$F9,$FF,$64,$3B
	.dc.b	$3F,$8F,$FC,$B7,$0C,$5C,$A8,$E7,$2A,$E7,$5F,$E7,$BF,$B2,$FE,$8F
	.dc.b	$3F,$E9,$AF,$B1,$FD,$3E,$5F,$C2,$5B,$1F,$C9,$67,$6F,$F8,$41,$B5
	.dc.b	$FE,$85,$9D,$C9,$B3,$AA,$FE,$28,$FF,$85,$A7,$FD,$52,$D7,$A7,$9F
	.dc.b	$E8,$A7,$D9,$7A,$AB,$58,$77,$E8,$D1,$9B,$98,$7A,$E3,$2F,$68,$F8
	.dc.b	$7F,$65,$FD,$DE,$3F,$D7,$2D,$7F,$C5,$6D,$AF,$67,$C6,$70,$BF,$97
	.dc.b	$E9,$7F,$B9,$6F,$FD,$82,$D7,$8C,$6B,$E3,$BD,$5B,$C9,$DC,$72,$F2
	.dc.b	$B9,$FD,$94,$9E,$71,$5F,$F1,$EF,$D9,$7C,$97,$FA,$B3,$56,$2A,$FB
	.dc.b	$F3,$FE,$29,$9F,$B9,$DD,$F2,$CC,$F9,$FF,$84,$FB,$2F,$F4,$B4,$FF
	.dc.b	$B6,$B9,$DE,$8F,$E2,$1B,$1F,$ED,$E9,$DB,$FE,$50,$FB,$2F,$F3,$13
	.dc.b	$C7,$E1,$38,$1F,$90,$8C,$C6,$FB,$7A,$7D,$94,$8F,$F7,$4C,$FE,$71
	.dc.b	$E9,$EE,$7E,$63,$EC,$5F,$65,$4E,$91,$F9,$C3,$F0,$7D,$BF,$D7,$E5
	.dc.b	$F1,$3C,$4D,$75,$EF,$EC,$BA,$73,$CC,$BF,$DF,$38,$A7,$F4,$FF,$65
	.dc.b	$FE,$5E,$9F,$F7,$EF,$D9,$7E,$ED,$FF,$37,$B7,$B7,$FD,$C9,$3B,$9F
	.dc.b	$F3,$9A,$F7,$3F,$63,$4F,$A3,$A7,$68,$FD,$AF,$D9,$7F,$AA,$EF,$BF
	.dc.b	$4F,$03,$3F,$1B,$AF,$3B,$9F,$D2,$4A,$FF,$0C,$69,$FE,$F5,$C3,$FC
	.dc.b	$CB,$18,$E3,$FF,$01,$7D,$8B,$E0,$3F,$81,$FD,$93,$EB,$A9,$FC,$1A
	.dc.b	$75,$7C,$4F,$F1,$22,$FB,$2E,$3F,$0D,$27,$42,$74,$E3,$C1,$FF,$4D
	.dc.b	$37,$E2,$FF,$89,$C6,$23,$FD,$33,$F5,$F7,$66,$BF,$E1,$9A,$C3,$FE
	.dc.b	$1B,$9F,$D7,$4E,$AB,$FC,$71,$18,$9E,$17,$D8,$FE,$43,$19,$9D,$F4
	.dc.b	$E5,$F4,$FD,$93,$E5,$BF,$93,$B6,$33,$18,$F5,$72,$C8,$C9,$CA,$38
	.dc.b	$FD,$5F,$2F,$36,$4F,$68,$CB,$E0,$8C,$BF,$EB,$AB,$4F,$C6,$FA,$69
	.dc.b	$3F,$2D,$F2,$FD,$96,$23,$FD,$83,$F6,$5F,$E3,$BD,$BF,$C8,$0D,$8F
	.dc.b	$F2,$E7,$02,$F8,$70,$7C,$99,$56,$DB,$7A,$39,$F2,$7F,$4A,$9F,$C4
	.dc.b	$4F,$E0,$1A,$F8,$B8,$35,$32,$34,$F1,$8E,$5F,$1D,$CF,$92,$57,$38
	.dc.b	$CB,$D6,$32,$FB,$A3,$FA,$DF,$D8,$BE,$C6,$92,$7F,$27,$F6,$2F,$B2
	.dc.b	$C4,$7F,$B9,$A3,$E1,$3A,$DB,$3B,$7C,$3C,$3A,$4E,$48,$7F,$BB,$9C
	.dc.b	$BC,$FE,$2B,$FB,$6D,$FF,$10,$8F,$0F,$93,$89,$91,$82,$1C,$4F,$3F
	.dc.b	$D4,$F2,$23,$EA,$E9,$AF,$A6,$2E,$3F,$67,$0F,$CE,$4B,$EC,$BF,$D2
	.dc.b	$4F,$ED,$1A,$A7,$F0,$11,$87,$F0,$5E,$33,$E4,$3F,$C0,$F3,$9E,$5F
	.dc.b	$F4,$54,$FF,$7C,$3F,$BE,$C6,$6F,$FD,$50,$D7,$F8,$F7,$5F,$F8,$42
	.dc.b	$77,$E3,$5E,$A9,$5F,$C2,$27,$F3,$ED,$EC,$CB,$E7,$C7,$C3,$58,$CF
	.dc.b	$CF,$5D,$7C,$49,$7C,$DF,$F0,$FF,$D9,$7E,$6F,$14,$6A,$7A,$1F,$EB
	.dc.b	$F8,$CB,$FC,$6A,$7F,$45,$F6,$5F,$CE,$63,$F8,$EB,$FC,$FF,$2F,$F5
	.dc.b	$63,$63,$FD,$90,$ED,$EA,$1C,$63,$FA,$F7,$84,$67,$E1,$3F,$C6,$B1
	.dc.b	$AA,$FF,$5C,$7D,$97,$DC,$27,$1E,$AF,$FC,$77,$1F,$C6,$FE,$27,$F7
	.dc.b	$3F,$27,$FE,$48,$6C,$FF,$34,$9E,$53,$C3,$9E,$4E,$63,$99,$87,$D3
	.dc.b	$3F,$9B,$6C,$78,$3A,$7F,$D9,$98,$FF,$B3,$65,$5F,$AC,$7F,$86,$18
	.dc.b	$9D,$BE,$C6,$BF,$EE,$F8,$E1,$F1,$7C,$5F,$EF,$F1,$FE,$DC,$9F,$87
	.dc.b	$EA,$FF,$B1,$DE,$51,$93,$94,$65,$A6,$4F,$2B,$97,$C8,$66,$7B,$BE
	.dc.b	$86,$DF,$B5,$B1,$E2,$07,$BB,$EC,$9F,$64,$F9,$1F,$F6,$04,$61,$BE
	.dc.b	$37,$F8,$26,$BF,$DE,$93,$97,$E5,$07,$FC,$63,$C3,$E9,$6C,$34,$F9
	.dc.b	$DE,$93,$E9,$BF,$B1,$47,$D1,$AD,$BA,$8C,$ED,$E0,$C7,$FD,$29,$39
	.dc.b	$D7,$B4,$62,$7D,$B8,$FF,$86,$6F,$8B,$F0,$30,$EB,$3A,$C7,$C5,$87
	.dc.b	$2C,$BE,$84,$62,$FD,$15,$E1,$9C,$E1,$D2,$7E,$06,$3E,$17,$E6,$DF
	.dc.b	$1C,$BD,$93,$99,$F2,$63,$DD,$EA,$70,$BE,$99,$79,$A3,$13,$BB,$F8
	.dc.b	$11,$C0,$F0,$E3,$E4,$9D,$65,$ED,$C3,$9E,$3E,$1C,$BC,$93,$EB,$C3
	.dc.b	$94,$61,$F8,$D3,$BA,$E3,$87,$59,$F0,$65,$E5,$7D,$66,$2B,$86,$3E
	.dc.b	$28,$F4,$99,$AE,$59,$7C,$46,$E5,$E3,$8F,$69,$99,$F2,$E3,$E2,$6D
	.dc.b	$FE,$68,$FD,$D6,$5F,$5D,$B3,$FE,$2D,$69,$B9,$F8,$8D,$7C,$AE,$9F
	.dc.b	$E0,$33,$CD,$D9,$E0,$3E,$3F,$C2,$FE,$06,$3E,$44,$7C,$2B,$F4,$E1
	.dc.b	$F8,$DA,$D6,$13,$0F,$7D,$FE,$1A,$70,$F5,$BE,$2E,$67,$6F,$5B,$F1
	.dc.b	$CF,$EA,$AB,$8B,$C2,$74,$71,$3A,$4F,$02,$39,$8E,$18,$7A,$89,$5D
	.dc.b	$2B,$75,$D0,$FF,$42,$46,$5F,$BB,$5C,$A3,$11,$BA,$E7,$7E,$A7,$B6
	.dc.b	$1F,$26,$B7,$06,$92,$7E,$B5,$78,$1B,$3B,$6C,$FD,$F9,$FD,$E7,$C6
	.dc.b	$FF,$3A,$7F,$66,$D7,$FD,$38,$DA,$34,$FF,$3A,$73,$7F,$93,$63,$37
	.dc.b	$FE,$86,$59,$EA,$3B,$99,$71,$7E,$08,$C1,$C6,$F3,$1F,$EA,$D9,$F1
	.dc.b	$12,$7F,$95,$C6,$5B,$F6,$2F,$87,$E2,$BC,$3B,$18,$6C,$FF,$A7,$27
	.dc.b	$83,$FE,$A3,$E7,$FE,$7D,$F0,$7A,$C3,$FB,$FD,$3C,$5E,$A7,$6F,$81
	.dc.b	$B8,$73,$8B,$5C,$26,$ED,$FE,$73,$B7,$BB,$E6,$FB,$1F,$3B,$71,$F2
	.dc.b	$38,$7C,$9F,$26,$F8,$4C,$65,$FE,$1E,$C5,$F5,$FE,$FB,$FF,$5E,$46
	.dc.b	$37,$F9,$EE,$BC,$0B,$5B,$BE,$4D,$9F,$BA,$DB,$FF,$68,$AA,$E3,$7D
	.dc.b	$3E,$C1,$F0,$D4,$EF,$F3,$7F,$D9,$F0,$F1,$CE,$9B,$3B,$7C,$F3,$BC
	.dc.b	$BD,$95,$B7,$FC,$5F,$F6,$2F,$AA,$F5,$FD,$63,$E4,$3F,$0F,$39,$C7
	.dc.b	$E0,$CE,$5F,$F6,$CC,$66,$F9,$DF,$C2,$F7,$B8,$B2,$B6,$E5,$A7,$AB
	.dc.b	$E0,$BE,$01,$F0,$7E,$72,$C6,$49,$7E,$A8,$CB,$F3,$BE,$C7,$E8,$F6
	.dc.b	$7D,$26,$D7,$D8,$E2,$46,$2B,$84,$E6,$BF,$4B,$F5,$CF,$B0,$FB,$BF
	.dc.b	$62,$FB,$1C,$C6,$1B,$7F,$EF,$AA,$E0,$FD,$4F,$A9,$CE,$3F,$E5,$C9
	.dc.b	$CB,$6F,$EF,$4E,$BA,$3F,$0B,$E7,$FE,$2B,$1F,$C5,$B7,$1F,$53,$70
	.dc.b	$FC,$6B,$6B,$F8,$4F,$2B,$DF,$FC,$51,$F8,$BF,$80,$B8,$7E,$22,$2E
	.dc.b	$1F,$84,$9B,$7F,$ED,$DD,$3F,$DF,$97,$F8,$0F,$27,$E2,$FC,$37,$E1
	.dc.b	$9B,$1F,$06,$3F,$43,$F1,$98,$C7,$E0,$45,$C3,$F4,$33,$6B,$FC,$87
	.dc.b	$0C,$DF,$29,$FA,$1F,$91,$FD,$13,$6B,$F3,$ED,$C7,$F3,$6D,$BF,$D6
	.dc.b	$F8,$FF,$B2,$3F,$67,$EB,$5B,$FC,$14,$5B,$FD,$AC,$DA,$FF,$7E,$7D
	.dc.b	$8B,$3F,$D9,$78,$9F,$CA,$23,$18,$FE,$C2,$31,$32,$3F,$67,$D9,$71
	.dc.b	$FF,$4D,$DF,$06,$D7,$EF,$67,$58,$2F,$FC,$93,$F6,$0F,$B1,$FE,$6E
	.dc.b	$D8,$FE,$23,$DE,$F7,$F3,$5F,$CF,$3C,$FF,$F2,$46,$1C,$F3,$FF,$37
	.dc.b	$C6,$30,$F9,$F8,$7F,$4A,$7E,$89,$FE,$58,$8D,$4F,$F9,$DB,$EB,$9F
	.dc.b	$61,$FD,$21,$B5,$FC,$8B,$EC,$7F,$9C,$3C,$AB,$85,$6F,$09,$3F,$CA
	.dc.b	$3E,$C7,$8D,$7F,$AF,$B0,$57,$5F,$3A,$5E,$FD,$1B,$7F,$EC,$4F,$5D
	.dc.b	$D1,$FD,$69,$B5,$FC,$D3,$EC,$5E,$16,$6B,$5F,$3F,$2B,$F9,$CF,$D8
	.dc.b	$BD,$6C,$65,$FE,$D0,$FB,$17,$D5,$FF,$66,$6D,$FF,$40,$F1,$3F,$58
	.dc.b	$FB,$E9,$3A,$79,$7B,$7C,$4F,$E8,$B4,$F7,$B1,$7C,$6F,$FC,$89,$D1
	.dc.b	$F5,$F0,$E6,$7F,$4C,$FB,$1F,$D1,$D6,$E3,$FB,$CA,$F9,$BA,$6F,$AE
	.dc.b	$C7,$97,$EC,$5E,$9F,$F8,$01,$B1,$FD,$6B,$D0,$FB,$F8,$27,$2F,$FB
	.dc.b	$CB,$EC,$5E,$C7,$F8,$21,$B7,$F0,$5F,$F1,$14,$EE,$75,$F2,$12,$BE
	.dc.b	$37,$D8,$3D,$1C,$5F,$FB,$FA,$31,$3A,$9F,$F0,$AF,$D8,$FC,$EA,$D4
	.dc.b	$7F,$86,$2B,$F1,$F1,$C7,$CC,$FF,$0D,$FC,$3B,$D9,$D4,$61,$98,$FD
	.dc.b	$7F,$B1,$FE,$FC,$FD,$9A,$E0,$E7,$CE,$DE,$0E,$5F,$A9,$7C,$23,$15
	.dc.b	$FE,$BE,$BE,$31,$83,$FD,$BB,$F0,$3F,$83,$BF,$F1,$DF,$D7,$7E,$DE
	.dc.b	$FF,$11,$39,$7F,$90,$C6,$6B,$90,$9D,$3D,$3B,$3F,$C1,$F3,$FD,$C6
	.dc.b	$37,$3A,$78,$4F,$F8,$FA,$38,$DA,$B8,$3F,$8F,$8E,$07,$DA,$7F,$CA
	.dc.b	$3F,$63,$FD,$59,$D7,$C4,$7F,$87,$A3,$18,$FF,$B7,$AB,$83,$67,$5F
	.dc.b	$27,$E1,$7C,$5F,$13,$27,$E0,$E7,$4F,$4C,$5C,$7C,$91,$67,$FD,$D7
	.dc.b	$1F,$B7,$7F,$60,$D8,$F8,$E0,$FF,$7F,$0F,$B1,$7C,$AF,$2F,$0F,$FC
	.dc.b	$C3,$5C,$4F,$D8,$16,$FF,$DF,$37,$CF,$C9,$E9,$87,$28,$FE,$AF,$7F
	.dc.b	$E6,$A7,$47,$7C,$3B,$5F,$07,$9C,$79,$2B,$93,$F8,$DA,$FE,$6D,$18
	.dc.b	$8E,$6E,$B0,$F7,$39,$8E,$55,$D9,$F0,$1E,$8C,$3C,$0F,$9B,$0A,$FC
	.dc.b	$53,$89,$93,$89,$98,$DD,$F8,$2B,$D3,$18,$C3,$CF,$18,$AE,$35,$B7
	.dc.b	$C4,$62,$76,$E0,$FF,$46,$D7,$FA,$39,$F9,$26,$0F,$8F,$F6,$3F,$C7
	.dc.b	$7E,$33,$FD,$26,$D8,$CF,$82,$9C,$B6,$38,$F9,$9F,$B9,$6E,$3C,$A3
	.dc.b	$11,$F7,$AD,$18,$8F,$8D,$18,$9F,$93,$97,$FA,$DA,$6E,$BF,$2F,$5F
	.dc.b	$B7,$7B,$9F,$96,$D8,$FB,$F5,$C6,$7E,$7B,$63,$8E,$1E,$76,$E1,$DA
	.dc.b	$A1,$D2,$75,$5E,$66,$C7,$E8,$63,$A1,$F7,$CA,$FE,$96,$FC,$91,$C6
	.dc.b	$BA,$DF,$9B,$0F,$64,$E6,$FC,$F1,$FB,$2D,$7A,$36,$3C,$0D,$58,$76
	.dc.b	$72,$7F,$67,$AF,$F6,$CB,$7D,$3E,$11,$F1,$71,$FE,$C5,$F6,$3F,$DD
	.dc.b	$6B,$FB,$73,$5F,$E4,$AD,$8D,$E7,$E1,$8C,$65,$FC,$B3,$E5,$FC,$D5
	.dc.b	$FC,$85,$D7,$07,$9F,$EB,$FF,$A4,$63,$FD,$EF,$EC,$7F,$61,$87,$E9
	.dc.b	$DB,$1F,$C3,$67,$6F,$F3,$66,$E1,$FD,$E6,$77,$26,$CE,$AB,$F6,$4F
	.dc.b	$F8,$6A,$7F,$C1,$4D,$7A,$79,$FF,$2E,$9F,$63,$E7,$8C,$72,$CA,$E3
	.dc.b	$19,$7D,$51,$97,$99,$FE,$17,$FB,$14,$7D,$76,$C6,$63,$C1,$F1,$B9
	.dc.b	$7E,$E7,$A5,$FE,$D5,$9F,$F1,$03,$5F,$EA,$F5,$C9,$FE,$B3,$19,$D3
	.dc.b	$E2,$6D,$EB,$39,$FC,$87,$4F,$2F,$AF,$D8,$CC,$3A,$C5,$7F,$B1,$FD
	.dc.b	$8F,$C9,$7F,$9F,$35,$62,$AF,$C9,$4F,$FA,$2A,$7E,$E7,$6F,$CB,$33
	.dc.b	$F0,$9F,$97,$FB,$17,$C4,$7E,$67,$9B,$FC,$6D,$7F,$E3,$D6,$BF,$DD
	.dc.b	$27,$6F,$F9,$07,$EC,$7F,$CC,$AE,$FC,$8E,$47,$E8,$E3,$31,$FD,$F1
	.dc.b	$FE,$1E,$F5,$F9,$3F,$01,$BE,$BE,$1E,$EF,$EA,$FE,$C5,$F3,$94,$CD
	.dc.b	$FF,$94,$67,$F1,$1C,$BF,$E4,$9A,$7C,$57,$13,$F6,$93,$AE,$EF,$E8
	.dc.b	$4B,$EC,$7B,$73,$CC,$7A,$47,$EE,$FE,$C7,$FC,$C7,$1F,$E5,$FF,$B1
	.dc.b	$FD,$CB,$FE,$16,$6E,$7F,$E5,$99,$DC,$FF,$86,$97,$C0,$7F,$05,$A7
	.dc.b	$D1,$53,$B4,$7F,$09,$FB,$1F,$F5,$85,$F7,$E8,$E0,$67,$E3,$B5,$E6
	.dc.b	$73,$FA,$49,$5F,$E7,$3B,$FF,$32,$3F,$E7,$5F,$AF,$FF,$30,$BF,$E7
	.dc.b	$7F,$B1,$7D,$6F,$F1,$76,$C6,$FE,$0B,$F8,$D5,$3E,$C7,$5F,$11,$2B
	.dc.b	$A4,$7F,$1F,$F8,$8F,$F4,$CC,$7F,$A0,$DA,$FF,$8E,$9B,$C7,$FE,$7D
	.dc.b	$FA,$EB,$D1,$FD,$CE,$6D,$7F,$86,$67,$F5,$13,$FD,$DD,$B1,$FE,$1D
	.dc.b	$9D,$BF,$E8,$BF,$8D,$FE,$61,$3A,$83,$67,$7F,$62,$F9,$8F,$E6,$CD
	.dc.b	$AF,$F4,$D3,$5F,$F2,$8B,$FC,$68,$C6,$3F,$E9,$56,$E3,$FE,$0B,$6C
	.dc.b	$FF,$8C,$5E,$1F,$29,$DB,$CC,$93,$DA,$3C,$26,$1E,$F1,$87,$E5,$7D
	.dc.b	$2F,$CD,$23,$E6,$FB,$1A,$FF,$99,$BE,$C7,$FC,$73,$8F,$F8,$75,$B1
	.dc.b	$F1,$EB,$79,$7F,$AB,$67,$32,$6E,$8A,$FE,$B5,$1F,$84,$9E,$DC,$98
	.dc.b	$33,$B5,$8F,$F6,$14,$EF,$0E,$D5,$A9,$CF,$CE,$F4,$E2,$91,$82,$1E
	.dc.b	$B7,$2F,$B6,$3F,$B6,$7D,$7B,$EC,$69,$27,$FD,$A5,$F6,$0F,$B1,$C4
	.dc.b	$7F,$A2,$A3,$D8,$EB,$5C,$E7,$FE,$CF,$C3,$A4,$E4,$87,$FA,$49,$CB
	.dc.b	$CF,$E2,$FF,$BF,$5F,$F0,$E8,$F0,$F8,$F8,$99,$18,$21,$C4,$F3,$FD
	.dc.b	$4F,$22,$3E,$AF,$1E,$BE,$98,$B8,$FF,$B9,$70,$FC,$F7,$D8,$BC,$3F
	.dc.b	$F0,$83,$6F,$FD,$E8,$D7,$FD,$57,$E2,$BE,$47,$75,$FE,$17,$8F,$EE
	.dc.b	$E7,$F7,$8A,$E1,$87,$FB,$CD,$BC,$FF,$E6,$99,$DF,$C0,$F2,$C7,$FD
	.dc.b	$21,$7F,$EF,$99,$75,$F5,$24,$EA,$F9,$75,$E2,$3C,$B3,$6F,$94,$F9
	.dc.b	$7C,$49,$5C,$4F,$F4,$F6,$08,$FF,$18,$7D,$8F,$E7,$23,$13,$B3,$F9
	.dc.b	$7C,$F2,$7F,$D9,$B3,$FA,$2F,$B1,$FE,$6B,$1B,$7F,$9D,$E5,$FE,$87
	.dc.b	$6C,$7F,$AB,$5D,$BF,$ED,$66,$E2,$9E,$1F,$09,$FE,$43,$9E,$34,$77
	.dc.b	$3F,$E9,$3F,$B1,$FE,$30,$4E,$3E,$15,$FE,$4B,$8F,$E4,$3F,$13,$FB
	.dc.b	$7F,$93,$FF,$28,$36,$7D,$D3,$CA,$78,$73,$C9,$CC,$73,$30,$FF,$BB
	.dc.b	$A7,$F3,$4D,$8F,$07,$4F,$FA,$8C,$F3,$16,$3F,$D9,$D3,$FA,$C5,$F1
	.dc.b	$76,$3F,$CA,$CD,$9E,$12,$9F,$EF,$6F,$FB,$36,$3E,$17,$A9,$A9,$E9
	.dc.b	$5C,$E3,$04,$AE,$07,$8A,$B0,$F8,$9E,$6E,$8F,$04,$75,$8F,$1F,$D8
	.dc.b	$3E,$C5,$F2,$5B,$8C,$36,$33,$7D,$4F,$F7,$74,$FF,$6C,$70,$1E,$2F
	.dc.b	$FB,$81,$AF,$FB,$B1,$78,$DF,$E0,$A8,$FF,$6A,$46,$9F,$F6,$B7,$A5
	.dc.b	$FE,$34,$E0,$F2,$57,$FA,$23,$3F,$A1,$39,$AF,$0D,$F4,$E2,$F8,$13
	.dc.b	$9A,$FA,$D3,$9A,$F3,$CE,$73,$F0,$E1,$D5,$F0,$FB,$DF,$9B,$AE,$B3
	.dc.b	$94,$7F,$87,$23,$58,$76,$9D,$4E,$72,$F4,$E5,$E4,$BD,$E1,$E0,$8C
	.dc.b	$4E,$AB,$CB,$5A,$7E,$7D,$F4,$C7,$D1,$18,$BE,$2F,$99,$B8,$F8,$DB
	.dc.b	$8F,$79,$EB,$8F,$7A,$F9,$78,$F7,$9F,$B5,$8F,$78,$C6,$5F,$26,$B7
	.dc.b	$7C,$70,$F8,$ED,$C3,$E2,$E5,$E6,$79,$E5,$E2,$7D,$07,$88,$F4,$16
	.dc.b	$FD,$0D,$BF,$44,$78,$23,$6D,$36,$58,$D9,$E4,$8D,$B6,$36,$D3,$71
	.dc.b	$E8,$C7,$DF,$97,$C1,$C7,$D9,$3F,$6C,$E6,$6D,$CE,$3E,$FC,$BE,$36
	.dc.b	$1D,$6B,$DF,$8F,$95,$B1,$97,$E2,$9E,$F3,$27,$53,$B9,$E1,$8F,$4B
	.dc.b	$96,$B9,$8F,$5B,$4E,$65,$3B,$C7,$8D,$B1,$E3,$9F,$4C,$75,$8E,$75
	.dc.b	$ED,$C3,$E2,$4F,$92,$31,$97,$DD,$8E,$67,$E0,$BE,$17,$F1,$93,$F8
	.dc.b	$33,$FD,$A3,$96,$DB,$3F,$04,$FF,$43,$CF,$0C,$F8,$FC,$6F,$2F,$36
	.dc.b	$E1,$FB,$E6,$DB,$9F,$F7,$0F,$FA,$42,$BF,$75,$18,$C5,$96,$30,$FE
	.dc.b	$1F,$56,$FF,$33,$35,$E5,$3F,$DA,$EB,$8B,$6B,$51,$89,$F1,$4F,$43
	.dc.b	$85,$6A,$BA,$7D,$8B,$E1,$AF,$89,$86,$DF,$F7,$99,$D3,$63,$CD,$D1
	.dc.b	$FC,$E3,$E0,$B8,$4E,$FC,$6E,$A7,$6F,$7A,$C6,$49,$A7,$98,$FE,$37
	.dc.b	$8A,$77,$19,$10,$F7,$7D,$3F,$63,$EA,$6D,$F9,$DC,$3F,$E1,$7C,$9F
	.dc.b	$22,$F8,$3C,$E3,$E1,$E2,$7D,$7B,$83,$E9,$46,$1B,$0F,$06,$C4,$CB
	.dc.b	$FC,$49,$5B,$6D,$6D,$B3,$D3,$EA,$9F,$09,$4E,$FF,$37,$FD,$8A,$B9
	.dc.b	$CE,$9B,$3B,$7B,$CE,$F2,$F5,$56,$DF,$91,$F6,$0F,$AA,$F5,$FD,$8B
	.dc.b	$E4,$3F,$19,$39,$C7,$D3,$39,$7F,$C7,$11,$9B,$E7,$7F,$0B,$DE,$E2
	.dc.b	$CA,$DB,$96,$9E,$AF,$81,$F8,$07,$C1,$F9,$CB,$19,$25,$FA,$A3,$2F
	.dc.b	$D5,$DD,$F6,$3F,$4B,$B3,$E9,$B6,$FE,$7B,$6B,$83,$78,$BF,$67,$F5
	.dc.b	$CF,$B0,$FB,$BF,$62,$FB,$1F,$D7,$4E,$71,$FB,$73,$97,$FC,$B5,$19
	.dc.b	$C7,$F2,$8D,$C3,$F7,$BC,$6F,$AA,$E7,$94,$C3,$8B,$F8,$18,$C6,$1E
	.dc.b	$DF,$43,$F1,$B9,$FE,$39,$B7,$F8,$A6,$CE,$E3,$F1,$B3,$F8,$BF,$80
	.dc.b	$91,$F8,$8F,$81,$B5,$F8,$49,$B9,$F9,$A6,$D7,$F8,$2A,$9A,$7F,$BD
	.dc.b	$63,$F0,$7E,$4F,$C4,$F2,$7F,$12,$D8,$F8,$71,$FA,$4F,$8C,$C5,$FC
	.dc.b	$48,$B8,$FF,$20,$E3,$56,$2F,$FD,$2F,$C0,$F9,$3F,$D2,$7C,$C7,$E9
	.dc.b	$5B,$3F,$A2,$6E,$5F,$9E,$6E,$3D,$7C,$8F,$DA,$9F,$B6,$F5,$AD,$FE
	.dc.b	$CA,$2D,$FE,$2E,$6D,$7F,$36,$FB,$1F,$2D,$FE,$C3,$C4,$FE,$61,$18
	.dc.b	$BF,$D6,$C6,$34,$95,$FB,$6E,$CB,$5F,$C5,$EF,$86,$1D,$6F,$F8,$05
	.dc.b	$6A,$3F,$A2,$FD,$8B,$EA,$7F,$9F,$37,$1F,$E1,$7D,$FF,$EB,$A7,$FD
	.dc.b	$55,$19,$CB,$F9,$27,$9F,$FE,$51,$C3,$FA,$7B,$6B,$FD,$87,$7F,$D0
	.dc.b	$E7,$3C,$9F,$D2,$FE,$B9,$F6,$14,$FE,$3F,$F6,$2E,$2C,$63,$FE,$B9
	.dc.b	$A6,$3C,$4F,$E4,$BF,$63,$CA,$75,$1D,$7C,$E9,$1F,$C9,$27,$F9,$E3
	.dc.b	$CE,$3F,$AE,$7B,$2E,$7A,$7F,$30,$FB,$1F,$D5,$57,$FB,$26,$74,$DF
	.dc.b	$9F,$E2,$7F,$36,$FB,$17,$AD,$89,$FF,$68,$7D,$8B,$EA,$FF,$B5,$36
	.dc.b	$3F,$9B,$F8,$4F,$0F,$35,$AF,$6A,$56,$BD,$CB,$D3,$FD,$4F,$1F,$F3
	.dc.b	$8C,$F6,$8F,$E8,$FF,$63,$F9,$B8,$C1,$FE,$E1,$1F,$37,$FD,$ED,$A7
	.dc.b	$F4,$88,$FB,$9D,$5F,$E0,$88,$CD,$7F,$61,$FB,$17,$A3,$FE,$03,$6C
	.dc.b	$FF,$50,$F4,$3F,$01,$5F,$E8,$49,$CB,$FD,$A7,$EC,$5E,$B7,$F8,$2D
	.dc.b	$B7,$FE,$F5,$7F,$C4,$38,$FF,$86,$1C,$D7,$1F,$90,$E0,$7F,$84,$FE
	.dc.b	$C1,$E8,$63,$0F,$F7,$F4,$61,$B1,$FE,$19,$FB,$1F,$9D,$18,$37,$E0
	.dc.b	$73,$3F,$C3,$DF,$0E,$F5,$F1,$3F,$E9,$ED,$3F,$C5,$73,$98,$C1,$FC
	.dc.b	$65,$C6,$7F,$E0,$7F,$B1,$FE,$FD,$3A,$8F,$F5,$3C,$EA,$BF,$C6,$B5
	.dc.b	$A9,$FF,$07,$4E,$6B,$F9,$24,$E6,$BC,$37,$FE,$33,$F7,$FF,$15,$87
	.dc.b	$F8,$27,$EC,$7F,$6C,$FF,$85,$E3,$35,$FE,$40,$9C,$BF,$E4,$48,$CE
	.dc.b	$1F,$CC,$5B,$5F,$E4,$9A,$FE,$E4,$F3,$E9,$F0,$12,$BF,$C1,$F3,$AB
	.dc.b	$E9,$3E,$5F,$86,$FF,$0C,$FD,$8F,$F5,$5B,$ED,$18,$F6,$B5,$1F,$17
	.dc.b	$1F,$F1,$2C,$F0,$F9,$5C,$C7,$AF,$E2,$3E,$01,$F0,$5B,$5C,$74,$4E
	.dc.b	$27,$E1,$17,$3F,$F3,$19,$FB,$6A,$F0,$36,$3F,$C7,$C7,$FB,$73,$C4
	.dc.b	$FF,$6E,$FD,$8B,$E5,$7F,$AB,$56,$CE,$7E,$27,$F9,$95,$C6,$1F,$D6
	.dc.b	$F0,$E4,$2F,$93,$F8,$BA,$FF,$36,$3A,$8C,$BF,$E7,$98,$FF,$78,$46
	.dc.b	$23,$57,$E2,$9C,$BD,$E1,$87,$A2,$35,$76,$7C,$93,$5E,$D1,$5E,$A7
	.dc.b	$95,$C3,$D0,$F5,$4F,$46,$9D,$0A,$74,$29,$D0,$C5,$FA,$DE,$8F,$C0
	.dc.b	$C3,$E9,$39,$BF,$34,$62,$BD,$D1,$EF,$AF,$F0,$43,$5F,$9E,$65,$D7
	.dc.b	$D8,$FF,$20,$F8,$8B,$E0,$9A,$4A,$F9,$53,$FE,$81,$C3,$8F,$C5,$FF
	.dc.b	$AA,$E7,$2D,$7E,$F4,$ED,$CC,$FC,$DC,$7F,$27,$17,$5F,$F0,$F6,$7F
	.dc.b	$7A,$F8,$37,$0E,$CF,$E0,$AF,$8D,$FC,$C6,$C6,$F0,$FB,$57,$CA,$7F
	.dc.b	$C5,$46,$8B,$87,$81,$FF,$17,$46,$EC,$1C,$3F,$9F,$9E,$0E,$A7,$A3
	.dc.b	$C2,$76,$70,$8C,$C7,$13,$CB,$87,$B2,$3C,$71,$FB,$F7,$DB,$7F,$ED
	.dc.b	$66,$E9,$E6,$AE,$15,$E1,$C3,$F6,$33,$EE,$F4,$EC,$F9,$5A,$F8,$4F
	.dc.b	$F5,$C7,$D8,$FF,$74,$7F,$87,$35,$CE,$FC,$23,$FD,$81,$1A,$8E,$B9
	.dc.b	$39,$9F,$2D,$86,$D6,$CF,$E3,$5F,$63,$FD,$25,$FE,$39,$F6,$3F,$9D
	.dc.b	$CF,$FC,$BC,$D8,$FA,$33,$B7,$FD,$9C,$DC,$3F,$DF,$33,$B9,$36,$75
	.dc.b	$5F,$E6,$97,$FC,$35,$3E,$2F,$43,$E1,$C6,$23,$FD,$B5,$4F,$B1,$CD
	.dc.b	$7C,$88,$74,$53,$35,$CC,$F5,$B9,$70,$7F,$B9,$7E,$C7,$8B,$FC,$ED
	.dc.b	$AE,$63,$B7,$C6,$7D,$69,$CE,$6C,$FF,$D0,$5E,$97,$FB,$77,$0F,$E8
	.dc.b	$8D,$79,$18,$F8,$FF,$46,$F2,$30,$70,$F2,$FF,$07,$4F,$B1,$A6,$22
	.dc.b	$CF,$F5,$1F,$B1,$F9,$6F,$E9,$1A,$AB,$93,$BF,$85,$F9,$1E,$04,$8C
	.dc.b	$1F,$D6,$FE,$C5,$EF,$FE,$53,$9D,$EA,$53,$F2,$F3,$AA,$FE,$C5,$F6
	.dc.b	$3F,$E6,$47,$FB,$25,$3E,$1B,$91,$F9,$E9,$D4,$B8,$BE,$53,$FB,$C8
	.dc.b	$FC,$DB,$9E,$2F,$F5,$75,$FF,$C0,$08,$EB,$F2,$5C,$FB,$FA,$7B,$BF
	.dc.b	$AA,$FB,$17,$CD,$53,$37,$FD,$DE,$7F,$19,$9F,$82,$14,$F8,$AE,$47
	.dc.b	$ED,$2B,$79,$FF,$AF,$D7,$D8,$F7,$E9,$98,$75,$8F,$DD,$FD,$8F,$F9
	.dc.b	$A6,$3F,$C0,$DF,$62,$F1,$7F,$7E,$D8,$FF,$05,$4E,$DF,$80,$7F,$1E
	.dc.b	$5E,$FF,$0F,$A2,$93,$C2,$3F,$84,$FD,$8F,$FB,$1A,$3B,$F4,$F6,$33
	.dc.b	$F1,$DA,$F3,$FC,$3F,$47,$2B,$FC,$35,$7F,$E6,$E7,$FC,$39,$F6,$1F
	.dc.b	$E6,$87,$FC,$3D,$F6,$2F,$AC,$FE,$2E,$D8,$DF,$C2,$7F,$1A,$A7,$D8
	.dc.b	$EB,$E2,$65,$74,$8F,$E3,$FE,$0F,$FA,$92,$3F,$C5,$4D,$7E,$09,$F9
	.dc.b	$0F,$8D,$FE,$4B,$F5,$F9,$EC,$FB,$15,$7A,$BF,$DE,$53,$FA,$F9,$FB
	.dc.b	$8D,$8F,$F7,$B4,$EE,$FF,$A4,$B5,$4E,$DF,$E9,$AD,$7C,$47,$F8,$FF
	.dc.b	$EC,$5F,$33,$FC,$E1,$B1,$98,$C7,$AF,$96,$46,$4E,$51,$C7,$EA,$B9
	.dc.b	$79,$D2,$7B,$46,$5F,$04,$65,$F8,$B6,$9F,$97,$F4,$B2,$7E,$5B,$E5
	.dc.b	$FB,$1C,$47,$CC,$FB,$1F,$F2,$5E,$BF,$DA,$A1,$18,$7E,$2D,$70,$DB
	.dc.b	$F4,$D3,$99,$DD,$64,$E5,$1F,$83,$8E,$DC,$BF,$D7,$23,$15,$CB,$6C
	.dc.b	$4C,$AF,$F3,$55,$6A,$F9,$D7,$1F,$91,$E5,$C7,$2F,$71,$97,$8C,$65
	.dc.b	$F7,$47,$F6,$CF,$AF,$7D,$8D,$24,$FF,$9E,$7E,$C1,$F6,$38,$8F,$C8
	.dc.b	$C3,$6F,$EE,$AD,$8F,$F4,$05,$F4,$6D,$EA,$A4,$60,$ED,$F1,$9C,$BC
	.dc.b	$99,$18,$21,$E0,$72,$F8,$A3,$C3,$F5,$32,$3D,$0F,$93,$C9,$C4,$7A
	.dc.b	$22,$F1,$7F,$1D,$CF,$F3,$5F,$62,$F1,$3F,$C1,$ED,$BF,$F4,$D3,$5F
	.dc.b	$D8,$F8,$CF,$92,$DD,$7F,$85,$A3,$FB,$E1,$FD,$F6,$B8,$61,$FC,$AD
	.dc.b	$BC,$9F,$8D,$9D,$F5,$F2,$C7,$FD,$51,$1F,$99,$6F,$4F,$53,$B7,$5D
	.dc.b	$8C,$90,$F2,$BE,$3F,$12,$47,$47,$FC,$57,$F6,$3F,$97,$B4,$6A,$7D
	.dc.b	$AE,$B1,$FF,$18,$FD,$8F,$F2,$E9,$FF,$5E,$3F,$CD,$5F,$E2,$6D,$D3
	.dc.b	$FA,$43,$B8,$CF,$93,$CB,$CC,$91,$CA,$3F,$C7,$BF,$63,$FC,$2C,$9C
	.dc.b	$9E,$09,$CB,$FE,$44,$F8,$AF,$D9,$FC,$A5,$F0,$0E,$AB,$FD,$A1,$78
	.dc.b	$6F,$4F,$63,$BC,$C8,$C9,$0F,$E5,$B3,$C5,$BE,$B7,$FA,$53,$AF,$F8
	.dc.b	$24,$FE,$7E,$7F,$D6,$B1,$83,$4F,$F9,$CA,$3F,$CA,$9F,$19,$B6,$CE
	.dc.b	$E7,$FD,$A8,$D7,$99,$C6,$B9,$A9,$F0,$57,$3A,$CC,$61,$F2,$D6,$1C
	.dc.b	$FA,$B2,$FB,$9E,$58,$EB,$F6,$1F,$D6,$E3,$C2,$17,$D8,$BE,$57,$8C
	.dc.b	$E6,$30,$F0,$9E,$59,$7C,$56,$DF,$F2,$B7,$9D,$F2,$6D,$71,$F3,$FC
	.dc.b	$78,$74,$8F,$F7,$84,$E9,$BE,$67,$E4,$5E,$93,$B7,$B4,$7F,$82,$71
	.dc.b	$F1,$37,$1F,$E7,$9C,$5F,$E3,$AE,$6F,$14,$E6,$FF,$D5,$38,$FF,$8F
	.dc.b	$39,$32,$7C,$0B,$E1,$3E,$FA,$E0,$FC,$76,$D7,$18,$F9,$73,$C3,$1F
	.dc.b	$1D,$F4,$9E,$4F,$5F,$7B,$FC,$7F,$3A,$9C,$91,$FC,$85,$73,$CB,$DF
	.dc.b	$7C,$E7,$81,$AB,$E7,$3A,$BE,$4D,$9D,$65,$C6,$30,$49,$FF,$65,$E1
	.dc.b	$DD,$F8,$B9,$7B,$89,$8F,$A6,$F9,$4E,$A2,$BF,$E3,$6C,$7C,$6E,$4F
	.dc.b	$19,$71,$F1,$B8,$3E,$64,$F2,$C3,$C5,$18,$35,$97,$CF,$AE,$27,$C2
	.dc.b	$CB,$D5,$3B,$8C,$C6,$27,$83,$AA,$EB,$18,$AF,$24,$E7,$0F,$4C,$73
	.dc.b	$8E,$59,$7E,$0A,$F8,$BF,$12,$7D,$51,$8C,$7E,$1E,$1D,$67,$E2,$65
	.dc.b	$8B,$E1,$97,$9A,$7C,$24,$3C,$07,$D4,$CB,$C7,$7C,$C9,$5E,$38,$CB
	.dc.b	$D3,$2F,$56,$1D,$F1,$F9,$95,$A7,$18,$7E,$62,$30,$D9,$D9,$1C,$1F
	.dc.b	$9E,$F8,$6F,$8B,$9F,$D6,$3F,$C8,$73,$FA,$9F,$3B,$3F,$1B,$A6,$46
	.dc.b	$4F,$D7,$46,$2F,$F5,$2D,$D7,$F4,$2D,$CF,$F7,$6F,$FA,$9B,$AB,$FD
	.dc.b	$31,$8A,$31,$87,$F0,$0D,$2E,$3F,$9C,$9A,$BE,$C5,$F0,$B8,$5E,$3D
	.dc.b	$8C,$37,$05,$7F,$9D,$2B,$8C,$63,$2F,$F5,$CF,$5F,$F2,$9E,$7E,$17
	.dc.b	$BF,$23,$B1,$DF,$DE,$B8,$71,$3F,$CA,$58,$A7,$35,$DE,$AE,$3F,$EB
	.dc.b	$7A,$7B,$BE,$9F,$B1,$F5,$37,$1E,$98,$7A,$2B,$71,$88,$F8,$53,$F1
	.dc.b	$B3,$7C,$97,$FB,$5E,$B5,$7C,$2F,$FD,$67,$8B,$EA,$BE,$3D,$76,$8C
	.dc.b	$6D,$F4,$1D,$78,$56,$B7,$7C,$9B,$3F,$6D,$B7,$F2,$D5,$71,$BE,$9F
	.dc.b	$54,$F8,$3A,$77,$F9,$DF,$EB,$18,$78,$E7,$4D,$9D,$BE,$79,$DE,$5E
	.dc.b	$CA,$DB,$FE,$B0,$FB,$07,$D5,$7A,$FE,$C5,$F2,$1F,$8A,$9C,$E3,$F0
	.dc.b	$67,$2F,$CD,$8C,$DF,$3B,$F8,$5E,$F7,$16,$56,$DC,$B4,$F5,$78,$9F
	.dc.b	$00,$F8,$3F,$39,$63,$24,$BF,$54,$65,$F9,$DF,$63,$F4,$7A,$FE,$93
	.dc.b	$6B,$EC,$71,$23,$15,$C2,$73,$5F,$A8,$FA,$E7,$CE,$7D,$DF,$B1,$7D
	.dc.b	$86,$63,$0D,$BF,$C7,$57,$07,$EA,$7D,$8E,$71,$FF,$6E,$CE,$5B,$7F
	.dc.b	$7A,$75,$D1,$F8,$5F,$3F,$F1,$58,$FE,$2D,$B8,$FA,$9B,$87,$E3,$5B
	.dc.b	$5F,$C1,$F9,$5F,$01,$F8,$A3,$F1,$7F,$03,$70,$FC,$44,$5C,$3F,$09
	.dc.b	$36,$D5,$FE,$C7,$D3,$F8,$FD,$7E,$93,$9B,$FC,$0F,$5A,$8F,$E0,$F3
	.dc.b	$A8,$FE,$31,$3A,$AE,$5E,$4F,$CD,$3F,$4D,$F1,$76,$3E,$4C,$5C,$3F
	.dc.b	$51,$36,$FD,$9F,$27,$FA,$4F,$99,$FD,$2B,$70,$FD,$13,$6F,$F3,$CD
	.dc.b	$AF,$D8,$78,$DF,$B5,$3F,$6D,$EA,$DB,$FC,$2C,$5B,$FD,$C4,$DA,$F7
	.dc.b	$43,$EC,$59,$7E,$DB,$E5,$3F,$97,$C6,$31,$FD,$A4,$62,$64,$7E,$DF
	.dc.b	$B2,$DB,$0E,$B8,$71,$8C,$57,$FB,$DE,$31,$3F,$76,$B8,$BE,$4F,$B0
	.dc.b	$7D,$8F,$F4,$06,$DF,$F1,$7E,$F7,$C0,$66,$BF,$A2,$F9,$FF,$57,$3F
	.dc.b	$E7,$ED,$C5,$68,$DF,$A9,$FD,$33,$EB,$DF,$63,$4F,$E4,$5F,$63,$FC
	.dc.b	$EF,$4F,$E6,$13,$A6,$E1,$C4,$FE,$51,$F6,$39,$CF,$DE,$7F,$50,$AF
	.dc.b	$0D,$72,$F4,$3C,$07,$0F,$4A,$C7,$F3,$DF,$5D,$D3,$FD,$81,$B1,$FC
	.dc.b	$DF,$EC,$5E,$1E,$6B,$5F,$3B,$2B,$F9,$E7,$D8,$BD,$7C,$6B,$E4,$1F
	.dc.b	$62,$FA,$BF,$ED,$AD,$BF,$E8,$9E,$27,$EC,$1F,$EA,$11,$9A,$E4,$BD
	.dc.b	$BE,$27,$F4,$9A,$7B,$D8,$BE,$37,$D8,$FE,$9D,$F6,$2F,$13,$13,$FD
	.dc.b	$74,$7C,$DF,$F8,$01,$A7,$F5,$79,$FC,$1D,$7A,$E3,$2F,$B7,$EC,$5E
	.dc.b	$A7,$F8,$21,$B3,$FD,$7F,$D1,$FC,$35,$27,$2F,$C5,$FB,$17,$B1,$FE
	.dc.b	$10,$6D,$FF,$71,$5F,$05,$99,$D7,$C8,$CA,$FE,$ED,$F6,$3F,$EE,$07
	.dc.b	$84,$7F,$86,$A3,$18,$7D,$36,$F8,$9F,$B4,$FB,$1F,$A1,$39,$F5,$38
	.dc.b	$79,$F3,$1F,$F0,$17,$C4,$3D,$BC,$5F,$DD,$8C,$4E,$A7,$FC,$53,$F6
	.dc.b	$3F,$E0,$5A,$D4,$63,$D1,$FF,$16,$DF,$09,$FC,$35,$6F,$3F,$E0,$F3
	.dc.b	$97,$F8,$4F,$C1,$FE,$0D,$FF,$1E,$7D,$6F,$ED,$32,$FF,$0B,$4E,$5F
	.dc.b	$F1,$74,$67,$0F,$C9,$4E,$5F,$F1,$AB,$69,$3F,$DD,$27,$7D,$BF,$E4
	.dc.b	$2A,$E1,$1F,$E2,$08,$FF,$0C,$4E,$EB,$CB,$1B,$E2,$FF,$0E,$7D,$8F
	.dc.b	$F5,$79,$F4,$C6,$3D,$CD,$65,$E8,$8D,$FC,$B6,$32,$F3,$FC,$4F,$84
	.dc.b	$F1,$36,$BD,$F1,$63,$DA,$5C,$3F,$97,$9F,$B3,$9F,$F1,$D3,$63,$F6
	.dc.b	$87,$F9,$5B,$C8,$FF,$2B,$FD,$8B,$E5,$BF,$CB,$33,$FC,$B7,$C5,$FF
	.dc.b	$31,$DF,$13,$AB,$FE,$66,$8E,$42,$B9,$3F,$E5,$F8,$D2,$BE,$F7,$DA
	.dc.b	$FC,$97,$C2,$BC,$35,$A7,$96,$15,$C9,$DE,$F9,$61,$FD,$AE,$F9,$4F
	.dc.b	$91,$E9,$87,$6A,$F6,$CE,$5E,$37,$D6,$30,$78,$0E,$C7,$0C,$3C,$55
	.dc.b	$CE,$30,$68,$F9,$13,$CD,$F8,$06,$63,$83,$EA,$BF,$2E,$1E,$9C,$3C
	.dc.b	$B3,$C0,$A5,$33,$86,$1F,$A2,$7B,$3E,$C7,$F8,$DF,$C5,$F0,$6C,$74
	.dc.b	$6B,$BF,$09,$3B,$F3,$7F,$B6,$4E,$DA,$F5,$6B,$F8,$6A,$46,$23,$FD
	.dc.b	$43,$18,$CF,$EC,$63,$FE,$B1,$9B,$AF,$E3,$B4,$FC,$05,$EE,$7E,$53
	.dc.b	$63,$F0,$95,$C6,$7E,$73,$63,$8E,$1E,$86,$E1,$DA,$A1,$D2,$75,$5E
	.dc.b	$96,$C7,$FA,$E2,$3A,$1F,$A2,$2B,$EC,$8D,$AB,$F3,$4E,$F0,$F6,$36
	.dc.b	$7F,$31,$3A,$7B,$4F,$E8,$75,$D4,$7F,$67,$AF,$76,$1F,$E2,$8B,$E1
	.dc.b	$3F,$B5,$BE,$0F,$9F,$D5,$AF,$EE,$5B,$7C,$B0,$FD,$DF,$D8,$FF,$75
	.dc.b	$AF,$F5,$E3,$5F,$EE,$CD,$C7,$FC,$6E,$FD,$C8,$DE,$FE,$09,$FE,$4F
	.dc.b	$F3,$17,$98,$F8,$AF,$96,$1F,$C2,$BE,$C7,$FA,$46,$3F,$EC,$BF,$B0
	.dc.b	$FD,$9C,$FF,$0E,$6C,$7F,$BA,$27,$6F,$F8,$25,$B8,$FF,$92,$A7,$72
	.dc.b	$6C,$EA,$BC,$1E,$87,$C4,$8F,$F0,$ED,$7F,$B6,$DA,$FF,$85,$69,$F6
	.dc.b	$3B,$AF,$97,$1F,$DC,$27,$D1,$CF,$97,$94,$EA,$24,$60,$FF,$74,$7D
	.dc.b	$8F,$17,$FC,$44,$DC,$BF,$CC,$6D,$CB,$E1,$7C,$6F,$DA,$8C,$50,$7A
	.dc.b	$7F,$ED,$E7,$FD,$E8,$D7,$AE,$1F,$82,$8C,$61,$F1,$F6,$F5,$99,$F9
	.dc.b	$0D,$79,$9B,$FB,$19,$8F,$48,$AF,$F8,$E7,$EC,$7E,$63,$E3,$6B,$E4
	.dc.b	$9F,$C8,$C7,$F8,$B6,$3E,$F7,$77,$CA,$33,$E8,$FE,$5B,$EC,$5F,$01
	.dc.b	$F9,$8E,$8F,$F1,$CD,$FB,$F5,$FF,$43,$C6,$23,$FC,$95,$5C,$1D,$71
	.dc.b	$7F,$56,$FB,$1F,$F3,$5C,$60,$F9,$14,$F2,$7B,$1F,$A6,$9D,$CF,$F9
	.dc.b	$55,$FC,$F3,$FE,$57,$97,$C9,$F9,$4E,$1E,$07,$1F,$77,$F5,$FF,$62
	.dc.b	$F9,$BA,$67,$0F,$10,$9F,$C7,$EB,$FE,$A2,$A7,$C5,$F2,$3F,$73,$3A
	.dc.b	$D7,$FB,$8A,$FB,$1E,$FC,$D3,$1E,$B3,$97,$BF,$D8,$7F,$9B,$5F,$0F
	.dc.b	$D8,$FF,$A4,$5B,$E2,$FF,$06,$9D,$47,$F9,$D5,$FF,$36,$3C,$3E,$8F
	.dc.b	$F8,$5D,$1E,$1F,$01,$FC,$3B,$EC,$7F,$D7,$35,$DB,$A7,$66,$7E,$43
	.dc.b	$5E,$77,$2F,$A4,$95,$F1,$2F,$FC,$DE,$FC,$6F,$B0,$FF,$34,$BF,$23
	.dc.b	$EC,$5F,$5D,$FC,$71,$B1,$BF,$83,$FE,$3F,$4F,$B1,$D7,$C3,$CA,$E9
	.dc.b	$1F,$C9,$3E,$1F,$FD,$67,$1F,$51,$AF,$FA,$55,$BD,$5F,$ED,$6F,$AF
	.dc.b	$BD,$3F,$E2,$29,$B5,$FB,$39,$FD,$6C,$FF,$89,$9B,$1F,$B7,$9D,$BF
	.dc.b	$E9,$4F,$8F,$FE,$6B,$3A,$83,$67,$7F,$62,$F9,$8F,$E7,$6D,$AF,$C5
	.dc.b	$B5,$FF,$56,$3F,$C8,$8C,$63,$FE,$EE,$6E,$3F,$E3,$A6,$CF,$F0,$B7
	.dc.b	$87,$CA,$F6,$F3,$24,$F6,$8F,$09,$87,$BC,$61,$F9,$5F,$4B,$F3,$48
	.dc.b	$F9,$BE,$C6,$BF,$EC,$1F,$AA,$FF,$29,$E3,$FE,$4C,$6C,$7C,$7A,$DE
	.dc.b	$4A,$D1,$FD,$72,$B7,$5F,$EB,$EA,$C7,$24,$8F,$D7,$CF,$E1,$9F,$06
	.dc.b	$D9,$7C,$53,$70,$ED,$1F,$CB,$6B,$E9,$BC,$BE,$7B,$C1,$C5,$23,$04
	.dc.b	$3D,$8E,$5F,$74,$7F,$72,$FB,$07,$D8,$52,$4F,$F0,$4F,$B0,$7D,$8E
	.dc.b	$23,$FD,$CD,$0D,$7F,$BD,$B6,$3F,$85,$DF,$46,$DE,$AA,$46,$0E,$DF
	.dc.b	$19,$CB,$C8,$91,$82,$1E,$07,$2F,$8A,$3C,$3F,$53,$23,$D0,$F9,$3C
	.dc.b	$8C,$47,$A2,$2E,$FF,$32,$19,$FE,$7F,$EC,$5E,$1F,$F8,$59,$B3,$FC
	.dc.b	$99,$AA,$B9,$78,$AF,$92,$E1,$7F,$E1,$E8,$FF,$00,$9F,$E0,$2B,$E3
	.dc.b	$5E,$16,$F3,$7F,$A6,$A7,$7F,$07,$CB,$1F,$E6,$B0,$EA,$E8,$48,$CF
	.dc.b	$55,$7C,$97,$CE,$71,$8F,$9A,$BC,$9E,$24,$BE,$AF,$F8,$D3,$EC,$7F
	.dc.b	$33,$8A,$35,$3D,$0F,$EF,$D1,$97,$FA,$4C,$FE,$83,$EC,$7F,$9A,$C6
	.dc.b	$C3,$FC,$F5,$C3,$70,$FA,$4F,$06,$F9,$3F,$E4,$68,$D6,$7F,$EB,$E6
	.dc.b	$D7,$F8,$3D,$E1,$1F,$EC,$2F,$B1,$FE,$30,$4E,$3E,$55,$FE,$50,$8F
	.dc.b	$E4,$3F,$14,$ED,$7C,$9F,$F9,$5A,$30,$EE,$F8,$3C,$7A,$26,$1F,$DA
	.dc.b	$30,$E2,$D9,$CB,$F0,$63,$FB,$4B,$93,$87,$AD,$FE,$C8,$EA,$F4,$16
	.dc.b	$3F,$C5,$53,$FA,$B7,$FD,$A5,$1E,$D9,$D4,$27,$FC,$C9,$F1,$9B,$78
	.dc.b	$36,$B7,$1A,$7B,$CF,$FB,$AD,$AF,$63,$C2,$F0,$AE,$2F,$2F,$56,$4F
	.dc.b	$13,$65,$1F,$61,$FD,$99,$E8,$63,$EC,$5F,$27,$B9,$46,$27,$71,$C7
	.dc.b	$2F,$84,$DC,$3F,$DF,$EF,$0B,$E5,$7C,$6B,$7E,$77,$C6,$AD,$61,$D2
	.dc.b	$B5,$87,$26,$C7,$F9,$66,$7B,$36,$3C,$8F,$FB,$CA,$39,$B7,$CB,$FC
	.dc.b	$E3,$CF,$1F,$F7,$0C,$A7,$FC,$39,$8F,$A9,$B8,$7F,$99,$ED,$C9,$FE
	.dc.b	$4E,$E3,$F8,$95,$A9,$F2,$35,$F9,$AD,$9E,$11,$F2,$23,$58,$78,$7D
	.dc.b	$5E,$67,$D5,$AD,$E3,$E3,$8C,$DF,$17,$5E,$F7,$F5,$09,$D4,$E4,$8F
	.dc.b	$E5,$6B,$9E,$5F,$53,$1F,$09,$30,$E7,$97,$6B,$E6,$4A,$F8,$57,$CB
	.dc.b	$0F,$7D,$EE,$7A,$3C,$A3,$C3,$18,$69,$FE,$91,$9E,$4F,$C6,$9E,$4F
	.dc.b	$43,$F2,$71,$AC,$BE,$34,$72,$38,$1A,$CB,$E1,$5F,$7C,$3D,$13,$CD
	.dc.b	$F8,$53,$A8,$C4,$EC,$8F,$96,$34,$F6,$9E,$6D,$9F,$87,$3A,$AF,$2D
	.dc.b	$69,$B3,$AC,$7D,$B7,$C4,$F0,$9D,$DC,$CF,$08,$C9,$F9,$0B,$ED,$97
	.dc.b	$D7,$8F,$85,$8F,$C5,$8C,$63,$E6,$8C,$65,$D2,$7F,$25,$7C,$A3,$07
	.dc.b	$D7,$AF,$44,$F2,$CB,$D1,$87,$CF,$C7,$C8,$FA,$4F,$26,$5C,$9F,$1E
	.dc.b	$3D,$5F,$A9,$1F,$03,$2F,$1C,$60,$98,$FE,$57,$E1,$3E,$7C,$FE,$10
	.dc.b	$FF,$2C,$57,$E8,$FC,$EC,$FC,$6E,$99,$19,$3F,$51,$18,$C3,$F1,$2D
	.dc.b	$C3,$F4,$4D,$CF,$F3,$6F,$F8,$E3,$AB,$FC,$1D,$8A,$31,$87,$E8,$36
	.dc.b	$B8,$FE,$66,$6A,$FB,$07,$C2,$E1,$78,$F6,$30,$DC,$3F,$D7,$73,$C1
	.dc.b	$FF,$67,$75,$FF,$30,$E7,$D6,$1F,$DD,$69,$E4,$75,$3B,$7B,$D7,$0E
	.dc.b	$71,$6B,$84,$DD,$BF,$C5,$57,$FE,$C9,$BF,$0F,$D8,$F9,$FD,$DF,$43
	.dc.b	$63,$CB,$1F,$03,$37,$CA,$FF,$B8,$67,$27,$FB,$97,$07,$D8,$7C,$08
	.dc.b	$C3,$73,$FF,$0A,$4E,$BC,$1F,$96,$F0,$DE,$D6,$E3,$EA,$36,$FF,$DF
	.dc.b	$4A,$B7,$5C,$BE,$B7,$FD,$DB,$5E,$36,$E1,$E0,$F9,$CF,$EB,$73,$9F
	.dc.b	$84,$E0,$47,$D5,$1A,$9F,$3D,$72,$CB,$FC,$A5,$F6,$2F,$AB,$F4,$FD
	.dc.b	$6B,$E4,$3F,$07,$39,$CB,$E2,$4E,$5F,$F7,$CC,$67,$1F,$F1,$83,$70
	.dc.b	$F7,$FB,$DC,$B1,$97,$C1,$F8,$B1,$8B,$F2,$F8,$BF,$04,$F8,$5F,$3B
	.dc.b	$6F,$E1,$45,$9F,$5C,$DC,$3E,$6F,$D8,$FD,$0E,$BF,$A2,$DC,$3E,$CB
	.dc.b	$6F,$EC,$F1,$A3,$17,$FE,$75,$FA,$F7,$CE,$7D,$CF,$B1,$7D,$8F,$E9
	.dc.b	$A7,$39,$7E,$1E,$72,$B2,$FF,$29,$4E,$9B,$7F,$77,$EA,$79,$32,$B8
	.dc.b	$63,$E1,$8D,$46,$0F,$EF,$17,$CE,$7F,$17,$E7,$FE,$3B,$2F,$C7,$B6
	.dc.b	$BF,$2E,$DC,$7D,$ED,$AF,$E0,$DC,$AF,$80,$FC,$81,$F9,$1F,$81,$B8
	.dc.b	$7E,$46,$2E,$1F,$8A,$9B,$7F,$1F,$C9,$FC,$4F,$97,$F1,$5B,$87,$A6
	.dc.b	$FE,$26,$9F,$C8,$9F,$8B,$3F,$A6,$F8,$BE,$15,$B9,$56,$F0,$ED,$3F
	.dc.b	$B2,$AD,$D7,$E7,$3E,$4F,$F5,$1F,$31,$FA,$96,$D7,$EA,$5B,$3B,$8F
	.dc.b	$D6,$4F,$ED,$BC,$79,$1F,$BE,$F5,$6D,$7E,$1A,$6E,$3F,$BA,$9B,$5F
	.dc.b	$E2,$BF,$B1,$65,$FB,$9F,$94,$FE,$57,$18,$9F,$DD,$46,$26,$47,$EF
	.dc.b	$7B,$6E,$5F,$E9,$9B,$E0,$DA,$FE,$09,$3A,$C7,$FD,$95,$F6,$2F,$B1
	.dc.b	$FE,$72,$DC,$7F,$8D,$77,$FF,$AE,$5F,$99,$19,$C7,$FD,$4D,$E7,$FF
	.dc.b	$8E,$B0,$FF,$6A,$36,$BF,$27,$68,$FE,$8A,$66,$BF,$3D,$F5,$CF,$B1
	.dc.b	$FE,$92,$DB,$FE,$55,$F6,$3F,$CF,$9F,$B9,$19,$AF,$F5,$B2,$C3,$89
	.dc.b	$FC,$AB,$EC,$75,$97,$FB,$9E,$75,$7C,$BC,$EE,$E6,$FD,$1B,$1F,$89
	.dc.b	$F5,$9C,$DF,$D7,$1B,$1F,$CF,$7E,$C5,$E1,$66,$B5,$F3,$F2,$BF,$A1
	.dc.b	$7D,$8B,$D6,$C7,$07,$FB,$37,$EC,$5F,$57,$FD,$A1,$B7,$FD,$27,$C2
	.dc.b	$FD,$63,$F9,$B8,$CD,$71,$F6,$B8,$1F,$D2,$2B,$FC,$F1,$5B,$8E,$93
	.dc.b	$23,$FA,$A7,$D8,$BC,$1C,$4F,$FB,$7E,$5F,$37,$FD,$E9,$A7,$0E,$CC
	.dc.b	$47,$F7,$EF,$B1,$7A,$34,$FE,$B3,$E8,$7D,$DC,$13,$AC,$3F,$69,$F6
	.dc.b	$2F,$5B,$FC,$0E,$D8,$CF,$C0,$62,$73,$F2,$32,$7F,$79,$F6,$3F,$EE
	.dc.b	$07,$9C,$7F,$84,$23,$18,$FF,$BE,$5B,$DC,$AF,$F4,$C7,$D8,$FC,$CA
	.dc.b	$D7,$A3,$C7,$CC,$95,$FC,$27,$E1,$9E,$C6,$2D,$8F,$37,$F9,$B4,$EA
	.dc.b	$B8,$57,$F8,$8B,$EC,$7F,$C0,$95,$C2,$7F,$D5,$15,$B8,$C1,$FC,$FE
	.dc.b	$77,$7D,$63,$FD,$9F,$7C,$FC,$EF,$F1,$6D,$F1,$AD,$7C,$07,$E1,$1F
	.dc.b	$F1,$C7,$D7,$7E,$DF,$0E,$A7,$F8,$5B,$B7,$FC,$1B,$19,$9E,$31,$FE
	.dc.b	$38,$AE,$59,$FF,$4A,$AF,$EE,$D7,$CA,$72,$FF,$55,$6B,$D6,$7F,$C4
	.dc.b	$13,$C2,$FA,$C7,$F3,$3A,$71,$7F,$87,$BE,$C7,$FA,$94,$FF,$B2,$E7
	.dc.b	$2F,$F8,$A6,$33,$87,$7A,$F3,$1C,$7E,$5A,$D7,$5F,$87,$F5,$1E,$BA
	.dc.b	$B1,$92,$3E,$B3,$27,$C6,$C7,$F6,$D1,$FD,$25,$B1,$F2,$8F,$EE,$9E
	.dc.b	$3F,$F7,$5F,$B1,$7C,$B7,$FB,$8E,$B6,$7D,$92,$DE,$7C,$4F,$F3,$2E
	.dc.b	$1C,$85,$F2,$7F,$BC,$C6,$96,$1D,$F0,$E5,$76,$7F,$DE,$46,$8E,$71
	.dc.b	$A6,$9A,$9E,$0F,$E6,$AB,$F3,$57,$E1,$8F,$F7,$E4,$EB,$0F,$83,$39
	.dc.b	$9F,$43,$B8,$CD,$F8,$A7,$C5,$39,$BE,$B3,$9B,$E9,$18,$AF,$21,$E6
	.dc.b	$3C,$98,$79,$9E,$F5,$EB,$BF,$14,$76,$8C,$1D,$5F,$11,$F3,$4B,$78
	.dc.b	$70,$7F,$8B,$47,$D8,$FF,$20,$F8,$8E,$2D,$8F,$63,$5F,$9E,$D7,$92
	.dc.b	$F3,$3F,$82,$C7,$06,$BC,$1A,$F0,$6B,$FE,$3F,$8E,$59,$FD,$6C,$3F
	.dc.b	$D6,$13,$74,$F4,$65,$F5,$EF,$73,$F2,$5B,$1F,$6E,$B8,$CF,$CD,$6C
	.dc.b	$71,$C3,$CE,$DC,$3B,$54,$3A,$4E,$AB,$D1,$39,$AE,$D8,$79,$27,$37
	.dc.b	$C2,$B5,$1F,$3C,$F0,$9A,$AE,$33,$FE,$37,$8F,$3C,$FF,$21,$6E,$7B
	.dc.b	$79,$5F,$EB,$E3,$A7,$A7,$6F,$FC,$D1,$5F,$E0,$F7,$A5,$65,$7D,$8F
	.dc.b	$F7,$37,$FC,$8C,$D7,$F9,$3B,$70,$FE,$64,$FD,$D8,$DE,$DD,$E5,$C6
	.dc.b	$F9,$7C,$36,$B6,$7F,$94,$FE,$C7,$FA,$5B,$FE,$55,$FB,$0F,$D5,$6B
	.dc.b	$F9,$36,$C7,$FA,$06,$76,$FF,$3A,$6E,$1F,$D1,$67,$72,$6C,$EA,$BF
	.dc.b	$3C,$FF,$86,$A7,$C5,$E7,$FC,$38,$C4,$7F,$46,$A7,$D8,$E6,$BE,$44
	.dc.b	$7F,$38,$9F,$47,$36,$5E,$66,$08,$79,$E7,$E8,$7D,$8F,$18,$FE,$A6
	.dc.b	$DC,$BF,$AF,$37,$2F,$85,$F1,$7C,$30,$FB,$1E,$97,$FB,$77,$0F,$F3
	.dc.b	$E3,$5E,$71,$AF,$8E,$F3,$EF,$27,$71,$CB,$CA,$E7,$F6,$32,$79,$C5
	.dc.b	$7F,$B3,$FD,$8F,$C9,$7F,$D6,$4D,$58,$AB,$F1,$93,$FE,$2B,$9F,$B9
	.dc.b	$DB,$F2,$0C,$FA,$1F,$91,$FB,$17,$C3,$7E,$4F,$9D,$F2,$D4,$D4,$7E
	.dc.b	$5F,$EC,$7F,$CC,$98,$FF,$A6,$E9,$E3,$F1,$35,$F0,$3B,$F0,$7A,$7C
	.dc.b	$94,$BF,$F5,$16,$1F,$99,$7F,$C0,$6B,$DE,$FD,$27,$D8,$BE,$6A,$9E
	.dc.b	$18,$FD,$41,$F8,$6E,$BF,$F0,$A4,$BE,$27,$89,$AE,$DD,$FD,$8F,$4E
	.dc.b	$99,$9F,$FA,$CF,$14,$FE,$CF,$EC,$7F,$CC,$33,$FE,$B7,$FB,$1F,$DC
	.dc.b	$3F,$E1,$86,$E7,$FE,$23,$9D,$CF,$F8,$71,$7B,$FF,$BD,$A7,$D1,$53
	.dc.b	$B4,$7F,$01,$FB,$1F,$F5,$6D,$F7,$EA,$E0,$67,$E3,$75,$F0,$1C,$FE
	.dc.b	$92,$57,$FB,$2E,$FF,$CC,$AF,$FB,$37,$EC,$3F,$CC,$2F,$FB,$3F,$EC
	.dc.b	$5F,$5B,$FC,$3D,$B1,$BF,$84,$FE,$25,$4F,$B1,$D7,$C4,$4A,$E9,$1F
	.dc.b	$C5,$FC,$1F,$F4,$E4,$7F,$B6,$DA,$FF,$8F,$1B,$C7,$FE,$47,$FA,$FB
	.dc.b	$E6,$7F,$5D,$9B,$41,$3F,$A9,$AE,$1D,$1D,$FE,$42,$57,$FB,$A6,$B5
	.dc.b	$87,$F6,$76,$C7,$3C,$3A,$AF,$B1,$7C,$BF,$F2,$F6,$C7,$FB,$CD,$AF
	.dc.b	$F9,$51,$FE,$32,$63,$2F,$F3,$03,$70,$FE,$EC,$D9,$F5,$3C,$3E,$AF
	.dc.b	$B7,$9D,$27,$B4,$78,$4C,$3D,$E3,$0F,$CB,$FA,$5F,$9C,$47,$CD,$F6
	.dc.b	$35,$FF,$34,$FD,$57,$F8,$EF,$1F,$F0,$43,$63,$E1,$56,$F2,$FF,$3C
	.dc.b	$4E,$64,$DD,$15,$FD,$4E,$3F,$0B,$3D,$B9,$70,$67,$6B,$1F,$67,$0E
	.dc.b	$B3,$88,$F2,$CF,$3F,$9C,$E9,$C7,$27,$86,$7E,$89,$D4,$60,$FC,$26
	.dc.b	$2A,$E1,$1F,$02,$7F,$B4,$7D,$83,$EC,$29,$27,$90,$FB,$07,$D8,$E6
	.dc.b	$3F,$D2,$31,$F0,$1D,$B5,$FF,$4A,$D7,$57,$9B,$5E,$3B,$7F,$79,$9E
	.dc.b	$1B,$7F,$A9,$6B,$4F,$FA,$9E,$1F,$19,$BF,$22,$61,$E0,$8E,$46,$1F
	.dc.b	$2C,$61,$F2,$FD,$4F,$72,$3E,$8E,$9A,$FB,$22,$E1,$E1,$9F,$F0,$77
	.dc.b	$D8,$BC,$3A,$4B,$F8,$FE,$2B,$E4,$B1,$1F,$EB,$68,$FE,$F6,$7F,$7C
	.dc.b	$AE,$51,$F3,$9A,$EF,$E1,$3F,$C3,$93,$BB,$BD,$72,$BB,$BA,$EB,$C4
	.dc.b	$77,$8B,$9F,$FB,$3E,$FF,$C3,$13,$E3,$F1,$3A,$91,$FB,$FF,$63,$F9
	.dc.b	$D9,$4E,$67,$81,$AE,$9F,$F1,$87,$D8,$FF,$36,$8F,$C3,$BF,$CF,$23
	.dc.b	$FC,$E4,$D8,$FE,$2C,$ED,$FF,$6C,$B7,$0F,$F7,$23,$B8,$34,$D4,$FF
	.dc.b	$9F,$7E,$C7,$F8,$BC,$9A,$62,$7F,$D5,$35,$FE,$EF,$8F,$E3,$BF,$17
	.dc.b	$FB,$8F,$93,$FF,$26,$36,$FF,$3F,$3C,$A7,$FC,$98,$FF,$63,$8C,$F2
	.dc.b	$62,$64,$60,$87,$FB,$D2,$51,$CB,$D9,$FF,$58,$F4,$FA,$4B,$5F,$CE
	.dc.b	$6B,$F5,$1F,$1B,$FA,$97,$FC,$B0,$DA,$DC,$FB,$EA,$D7,$07,$E2,$4E
	.dc.b	$19,$3F,$ED,$C7,$C0,$D3,$C1,$1A,$7F,$DA,$71,$F3,$7D,$5E,$A6,$6F
	.dc.b	$F7,$9F,$61,$FD,$92,$B6,$7F,$51,$5F,$62,$F9,$3E,$D2,$9C,$C6,$1E
	.dc.b	$55,$D3,$CD,$F6,$C6,$AB,$A9,$C6,$FA,$38,$8E,$38,$FF,$AF,$DE,$17
	.dc.b	$E6,$AD,$B7,$0F,$06,$1D,$A7,$37,$C5,$B8,$7A,$DA,$FA,$67,$85,$7F
	.dc.b	$B4,$B0,$F3,$4F,$FB,$D9,$FF,$1F,$FC,$76,$6F,$CF,$39,$F2,$3F,$D3
	.dc.b	$19,$7C,$A6,$E5,$FE,$D7,$F4,$FF,$C9,$3C,$9F,$16,$73,$7F,$35,$AF
	.dc.b	$CB,$6C,$F0,$8F,$89,$1A,$CB,$C7,$18,$CB,$E9,$D6,$F1,$F1,$47,$72
	.dc.b	$4F,$D0,$9D,$63,$F3,$70,$E8,$4A,$E7,$EF,$B2,$E3,$8F,$4C,$7C,$07
	.dc.b	$39,$D3,$88,$D1,$30,$F8,$53,$A8,$C4,$FC,$1B,$EB,$5A,$BB,$39,$3A
	.dc.b	$46,$5A,$66,$B9,$BF,$E4,$CB,$ED,$3E,$DB,$E7,$3F,$35,$F6,$DE,$32
	.dc.b	$F9,$78,$72,$CB,$E2,$DF,$5C,$39,$3E,$03,$B5,$F1,$CB,$E4,$47,$C3
	.dc.b	$8C,$63,$E0,$BE,$73,$E1,$39,$63,$E0,$C3,$D9,$8F,$AA,$31,$3E,$13
	.dc.b	$FC,$85,$5C,$A3,$31,$CE,$7D,$19,$7C,$08,$E0,$66,$FD,$33,$FD,$E6
	.dc.b	$F9,$65,$E8,$8E,$39,$7A,$2F,$DD,$5A,$CB,$15,$C7,$1E,$D5,$D4,$EF
	.dc.b	$97,$92,$F8,$47,$C3,$39,$CF,$09,$D3,$CA,$7E,$28,$F8,$6C,$46,$7C
	.dc.b	$E9,$97,$E7,$BE,$3A,$E3,$F9,$B9,$B9,$7B,$66,$EB,$FB,$16,$D7,$CA
	.dc.b	$6B,$FE,$92,$7F,$33,$CF,$F9,$A8,$C6,$1F,$E2,$6C,$DA,$EA,$3A,$1F
	.dc.b	$E2,$AF,$B1,$7C,$46,$6F,$FC,$F1,$0F,$17,$10,$8D,$D7,$19,$CC,$62
	.dc.b	$58,$F8,$E3,$0F,$EB,$BA,$BF,$99,$73,$F0,$C7,$8F,$8D,$32,$F2,$7C
	.dc.b	$0D,$CB,$9C,$7F,$A2,$DF,$14,$E7,$1E,$D5,$6B,$59,$7F,$B3,$DF,$77
	.dc.b	$D8,$F9,$BD,$EF,$3B,$63,$E1,$47,$C1,$D1,$F2,$9F,$EE,$09,$C9,$FE
	.dc.b	$59,$C5,$F5,$FF,$16,$30,$DD,$3C,$E7,$C2,$5E,$5F,$CD,$AE,$3C,$16
	.dc.b	$78,$47,$DA,$6D,$FF,$9C,$5E,$2F,$69,$E5,$F5,$6F,$85,$A7,$5F,$9C
	.dc.b	$FE,$BD,$5E,$79,$D3,$67,$6F,$AE,$77,$97,$A3,$15,$F0,$8F,$58,$FB
	.dc.b	$07,$D5,$7A,$FE,$C5,$F2,$3F,$86,$9C,$E5,$F1,$E7,$2F,$F9,$E2,$33
	.dc.b	$7D,$1B,$8F,$C1,$F7,$B9,$63,$2F,$83,$F1,$E3,$17,$E7,$C9,$E2,$FC
	.dc.b	$43,$E2,$FC,$E5,$C3,$E1,$45,$AE,$13,$71,$FA,$1F,$63,$F4,$FA,$BE
	.dc.b	$A3,$70,$FB,$AD,$BF,$8D,$C4,$8C,$5F,$EC,$7E,$B9,$F6,$1F,$7F,$EC
	.dc.b	$5F,$63,$FA,$A9,$CE,$5F,$8B,$9C,$BF,$E9,$C8,$CE,$5F,$41,$B8,$7E
	.dc.b	$FB,$67,$D5,$F3,$CA,$61,$C5,$FC,$2C,$63,$0F,$3F,$9F,$F8,$FC,$FF
	.dc.b	$20,$DB,$FC,$CB,$67,$71,$F9,$C9,$FC,$7F,$BF,$23,$F1,$9F,$01,$6B
	.dc.b	$E9,$4D,$CF,$CD,$36,$BF,$71,$A7,$FB,$D9,$FC,$27,$95,$F0,$FC,$AF
	.dc.b	$C5,$36,$3E,$14,$7E,$97,$E2,$F1,$97,$C3,$8B,$8F,$E9,$E6,$D7,$FB
	.dc.b	$5F,$77,$CA,$FE,$9B,$E4,$3F,$4E,$D9,$FD,$23,$71,$FC,$53,$70,$FD
	.dc.b	$97,$8F,$FB,$53,$F6,$DE,$AD,$BF,$DF,$C5,$BF,$C1,$CD,$AF,$E6,$5F
	.dc.b	$62,$CF,$F6,$FF,$2B,$FC,$AA,$31,$8F,$ED,$63,$13,$23,$F7,$1D,$B7
	.dc.b	$1F,$D8,$DF,$06,$D7,$F0,$29,$D6,$1F,$D0,$7E,$C5,$F5,$5F,$CE,$1B
	.dc.b	$8F,$F1,$9E,$F7,$BD,$9A,$FE,$7B,$E7,$FF,$8D,$F0,$FE,$94,$DB,$67
	.dc.b	$FC,$F3,$0F,$E9,$71,$FE,$F2,$9D,$4E,$BE,$BD,$F3,$1F,$D2,$1B,$5F
	.dc.b	$CA,$7E,$C7,$FD,$CE,$FF,$BE,$A3,$35,$FC,$39,$61,$C4,$FE,$51,$F6
	.dc.b	$3A,$CB,$FA,$F4,$EA,$F9,$79,$FD,$CD,$FA,$56,$3D,$35,$FD,$97,$D7
	.dc.b	$74,$7F,$5C,$6C,$FF,$3D,$FB,$1F,$D6,$BC,$0F,$EA,$11,$AD,$3F,$91
	.dc.b	$AF,$9F,$E4,$7F,$3D,$FB,$17,$AF,$8B,$E5,$7F,$2B,$EC,$5F,$57,$FD
	.dc.b	$B1,$B3,$FD,$37,$C3,$78,$99,$AD,$7B,$52,$B5,$EF,$DC,$7F,$9A,$E3
	.dc.b	$F5,$A7,$BC,$23,$0F,$F3,$DF,$B1,$FD,$2C,$EA,$73,$C3,$FE,$09,$87
	.dc.b	$CD,$FF,$80,$1A,$7F,$61,$AF,$C5,$DF,$F4,$48,$CB,$FD,$1B,$EC,$7F
	.dc.b	$99,$C7,$89,$B1,$FD,$37,$D1,$FC,$3A,$8C,$E1,$FE,$19,$FB,$17,$B1
	.dc.b	$FE,$10,$6E,$1E,$C7,$FC,$5C,$D8,$D7,$C8,$48,$F8,$3F,$63,$FE,$DE
	.dc.b	$78,$47,$F8,$6A,$31,$8F,$F5,$C6,$E1,$FE,$29,$FB,$1F,$A3,$7C,$BD
	.dc.b	$1E,$1E,$5C,$8F,$9B,$F1,$0F,$63,$19,$7F,$66,$8C,$4E,$A7,$FC,$53
	.dc.b	$F6,$3F,$E0,$BA,$D4,$63,$CD,$FF,$16,$DF,$09,$FE,$DF,$5B,$CF,$FC
	.dc.b	$79,$39,$7F,$C7,$DF,$03,$F8,$8A,$FE,$F1,$F5,$BF,$BC,$CB,$F0,$33
	.dc.b	$9A,$FF,$22,$CE,$6B,$FB,$DC,$E6,$BF,$C5,$8D,$8F,$F1,$CC,$FF,$78
	.dc.b	$77,$D3,$FE,$42,$9D,$47,$F8,$9A,$3F,$C3,$93,$BB,$F2,$C6,$FE,$23
	.dc.b	$FC,$41,$F6,$3F,$D6,$B1,$EF,$18,$F6,$B5,$1E,$88,$DF,$C9,$E3,$2F
	.dc.b	$3F,$C4,$F8,$4F,$13,$6B,$DF,$16,$3D,$A5,$CB,$FC,$C8,$7E,$D2,$7F
	.dc.b	$4A,$D8,$FF,$0E,$9F,$C7,$BC,$4F,$E3,$FF,$60,$F9,$5F,$F2,$85,$6C
	.dc.b	$E7,$E2,$7F,$99,$1C,$61,$FE,$55,$C3,$90,$BE,$4F,$F1,$2A,$FF,$35
	.dc.b	$BA,$8C,$BF,$E6,$88,$FF,$69,$C6,$23,$57,$E2,$9C,$BD,$E1,$87,$A2
	.dc.b	$35,$76,$7C,$93,$5E,$D1,$5E,$A7,$95,$C3,$D0,$F5,$4F,$46,$9D,$0A
	.dc.b	$74,$29,$D0,$C5,$FA,$DE,$8F,$C0,$C3,$E9,$39,$BF,$34,$62,$BD,$D1
	.dc.b	$EF,$AF,$E6,$65,$3C,$C6,$5D,$7D,$8F,$F1,$AF,$8A,$BE,$09,$A4,$AF
	.dc.b	$95,$2F,$2F,$FD,$33,$18,$6B,$F7,$63,$58,$F6,$9C,$CF,$FA,$56,$72
	.dc.b	$FC,$4C,$5A,$FF,$A6,$73,$FB,$F1,$89,$FA,$95,$C2,$7F,$06,$D9,$FF
	.dc.b	$50,$1B,$9E,$95,$CE,$3C,$17,$E8,$BE,$2D,$8E,$33,$BA,$FE,$F7,$1C
	.dc.b	$EB,$78,$78,$AF,$85,$6A,$F8,$3F,$A3,$6E,$1E,$58,$FF,$02,$B5,$D1
	.dc.b	$F8,$FC,$3B,$1E,$6B,$C5,$7E,$B3,$4F,$2D,$7F,$58,$AF,$75,$7F,$B6
	.dc.b	$EF,$A4,$62,$BE,$1F,$A5,$C6,$FD,$6D,$7C,$27,$FA,$63,$EC,$7F,$B9
	.dc.b	$BF,$D9,$1A,$E7,$7E,$11,$F8,$18,$D4,$75,$C5,$CA,$F9,$7C,$36,$B6
	.dc.b	$7F,$70,$FB,$1F,$E8,$8F,$F7,$2F,$B0,$FC,$F6,$7F,$ED,$96,$C7,$F8
	.dc.b	$B2,$76,$FF,$AB,$9B,$87,$FB,$EE,$77,$26,$CE,$AB,$FD,$C8,$FF,$87
	.dc.b	$27,$C5,$E8,$7C,$38,$C4,$7F,$AE,$A9,$F6,$39,$AF,$91,$0E,$7A,$66
	.dc.b	$B9,$9E,$B7,$2E,$0F,$F6,$2F,$D8,$F1,$7F,$C2,$2D,$73,$1D,$BE,$37
	.dc.b	$EB,$4E,$73,$67,$FE,$FD,$F4,$BF,$DC,$F8,$7F,$86,$9A,$F2,$31,$F2
	.dc.b	$1E,$8D,$E4,$60,$E1,$E6,$7E,$0E,$9F,$63,$4C,$45,$9F,$F1,$3F,$D8
	.dc.b	$FC,$B7,$FC,$BE,$D7,$FC,$7D,$3F,$8D,$8F,$F4,$6C,$7D,$DE,$EF,$92
	.dc.b	$67,$E1,$3F,$25,$F6,$2F,$88,$FC,$A7,$3B,$D4,$A6,$A3,$F3,$1F,$63
	.dc.b	$FE,$68,$C7,$FC,$73,$4F,$2B,$89,$AF,$80,$DF,$83,$D3,$E4,$E5,$FF
	.dc.b	$90,$30,$FC,$D3,$FE,$F8,$5E,$EF,$E9,$7E,$C5,$F3,$74,$F0,$C7,$EA
	.dc.b	$4F,$C3,$F4,$7D,$4B,$FF,$40,$D3,$E2,$B9,$1B,$EE,$E1,$F6,$3D,$7A
	.dc.b	$A5,$FF,$94,$B1,$4F,$ED,$7E,$C7,$FC,$C3,$3F,$E5,$6F,$B1,$78,$DF
	.dc.b	$B9,$6C,$7F,$96,$A7,$6B,$4F,$22,$87,$BF,$C7,$E8,$A4,$F1,$8F,$E0
	.dc.b	$9F,$63,$FE,$B3,$8F,$07,$5F,$73,$3F,$1D,$AF,$43,$C5,$F4,$72,$BF
	.dc.b	$CD,$77,$FE,$64,$7F,$CD,$BF,$60,$F9,$AF,$E2,$1F,$62,$FA,$EA,$7F
	.dc.b	$13,$9D,$57,$0F,$B1,$DF,$C5,$48,$C1,$FE,$78,$F0,$BF,$D4,$2D,$F8
	.dc.b	$FF,$E3,$91,$88,$F8,$5E,$5F,$FA,$D1,$7D,$7E,$79,$7E,$2D,$5A,$D4
	.dc.b	$FE,$AA,$7E,$4B,$63,$94,$EE,$BF,$D7,$ED,$7A,$F5,$7F,$2D,$AE,$11
	.dc.b	$AC,$3F,$98,$7D,$8B,$E6,$69,$27,$FD,$27,$1F,$C6,$8D,$38,$8D,$3E
	.dc.b	$C3,$FD,$9F,$1B,$F9,$5E,$BE,$7C,$9E,$71,$E0,$30,$F1,$13,$FC,$D5
	.dc.b	$C3,$FE,$E0,$7E,$77,$D2,$FC,$C2,$3E,$6F,$B1,$AF,$3F,$B1,$FF,$24
	.dc.b	$ED,$F3,$DB,$1F,$1A,$B7,$BF,$E5,$27,$32,$6E,$EA,$FE,$AD,$1F,$87
	.dc.b	$9F,$F5,$BB,$5F,$07,$1E,$B7,$DB,$23,$6E,$5E,$3F,$25,$D3,$8A,$46
	.dc.b	$08,$7A,$DC,$BE,$D8,$FE,$D1,$F5,$EF,$B1,$A4,$9F,$F6,$5F,$D8,$3E
	.dc.b	$C7,$11,$F3,$E3,$D8,$EB,$4C,$ED,$FB,$6C,$3A,$4E,$48,$7D,$77,$2F
	.dc.b	$3F,$8D,$FE,$F5,$7F,$C3,$E3,$C3,$E4,$E2,$64,$60,$87,$13,$CF,$F5
	.dc.b	$3C,$88,$FA,$BA,$6B,$E9,$8B,$8F,$F0,$CC,$3F,$3B,$2F,$B1,$78,$7F
	.dc.b	$E0,$D6,$C7,$FB,$B9,$AF,$E4,$7C,$57,$C9,$EE,$BF,$C2,$91,$FD,$E8
	.dc.b	$FE,$F7,$5C,$30,$FE,$36,$DE,$7F,$DE,$4E,$FE,$13,$96,$3F,$EF,$C8
	.dc.b	$FC,$D3,$DF,$AF,$A9,$DB,$B2,$C6,$E2,$BE,$58,$F1,$F8,$92,$3A,$3F
	.dc.b	$E2,$9F,$B1,$FC,$CD,$A3,$53,$ED,$75,$8F,$F8,$BF,$EC,$7F,$98,$C8
	.dc.b	$FD,$93,$FC,$E9,$BE,$66,$FC,$A9,$1B,$DB,$F9,$DD,$6D,$B5,$A9,$FD
	.dc.b	$E7,$D8,$FD,$42,$72,$78,$63,$FC,$81,$1F,$C5,$3E,$2F,$F6,$1F,$29
	.dc.b	$FE,$45,$6C,$F7,$CF,$F9,$45,$73,$AE,$3D,$12,$B4,$7A,$E5,$3D,$8D
	.dc.b	$47,$92,$70,$F0,$13,$FE,$4E,$F8,$DF,$D2,$C6,$3A,$FF,$9C,$46,$27
	.dc.b	$2D,$F6,$F3,$3E,$66,$D7,$8E,$2B,$CC,$FF,$6C,$3B,$69,$B2,$9B,$3F
	.dc.b	$D8,$EF,$C2,$39,$D7,$9D,$C4,$70,$8F,$F0,$3B,$7D,$6D,$CF,$8B,$EC
	.dc.b	$3F,$AE,$CF,$43,$F0,$4B,$EC,$5F,$29,$E2,$8C,$37,$D0,$E4,$73,$9D
	.dc.b	$D7,$31,$5C,$1C,$3E,$0C,$FE,$6B,$96,$CF,$26,$DF,$7A,$F1,$4E,$70
	.dc.b	$FE,$D5,$7C,$5B,$19,$9F,$5E,$1E,$5B,$E2,$48,$F1,$C6,$5F,$F7,$A5
	.dc.b	$F9,$DF,$EE,$C7,$AB,$C9,$FD,$23,$DA,$BF,$DB,$93,$FE,$45,$E2,$F8
	.dc.b	$D5,$FE,$22,$CB,$F1,$3C,$BF,$E4,$BE,$1C,$D4,$C7,$E1,$DE,$F4,$FF
	.dc.b	$02,$5E,$E7,$35,$BC,$3E,$3D,$EE,$7F,$71,$5B,$9C,$D6,$EF,$E2,$DE
	.dc.b	$EF,$FC,$5F,$EF,$FF,$91,$6F,$71,$D8,$8F,$A2,$35,$5C,$AF,$A5,$7E
	.dc.b	$1A,$FA,$61,$E1,$AD,$4F,$26,$E1,$CF,$1E,$33,$F2,$F0,$B7,$C6,$3A
	.dc.b	$46,$B2,$F9,$91,$F8,$B9,$D3,$C4,$EC,$70,$35,$97,$A1,$C6,$1E,$E2
	.dc.b	$63,$F3,$A7,$53,$CE,$72,$F0,$9D,$63,$E3,$CB,$E5,$63,$E4,$BF,$65
	.dc.b	$F8,$8E,$57,$F0,$0E,$2E,$67,$84,$66,$3E,$CE,$1E,$29,$CC,$78,$48
	.dc.b	$7D,$8C,$3C,$59,$7C,$8A,$E1,$8F,$90,$91,$FC,$23,$2F,$8F,$87,$4C
	.dc.b	$3C,$A7,$48,$CD,$70,$8F,$54,$62,$BB,$C7,$CC,$C2,$E5,$F1,$A3,$C8
	.dc.b	$70,$C7,$C7,$5E,$16,$C7,$12,$4F,$E9,$3E,$23,$E8,$E1,$F7,$0F,$F2
	.dc.b	$96,$1E,$9F,$3B,$3F,$1F,$A6,$46,$4F,$74,$63,$0F,$C3,$37,$0F,$D8
	.dc.b	$B7,$5F,$D6,$BF,$EA,$5E,$8F,$F4,$9E,$28,$C6,$1F,$B2,$D2,$E3,$FA
	.dc.b	$E9,$AB,$EC,$5F,$0F,$85,$E3,$58,$C3,$70,$FE,$9D,$3C,$1F,$EA,$FD
	.dc.b	$5F,$CB,$BA,$35,$87,$F9,$8E,$9E,$3F,$53,$B7,$C0,$DC,$39,$C5,$AE
	.dc.b	$13,$76,$FF,$57,$DF,$FB,$2A,$FC,$3F,$63,$E7,$F7,$BD,$0D,$8F,$2C
	.dc.b	$7C,$0C,$DF,$2D,$FE,$DB,$9C,$9F,$DB,$B0,$7D,$7F,$C0,$8C,$37,$3F
	.dc.b	$F4,$24,$EB,$C1,$F9,$6F,$0D,$ED,$6E,$3E,$A3,$6F,$FC,$1C,$AB,$75
	.dc.b	$CB,$EA,$9F,$0B,$4E,$BF,$3B,$FD,$63,$0F,$34,$E9,$B3,$B7,$D5,$3B
	.dc.b	$CB,$CF,$5B,$7F,$D2,$9F,$60,$FA,$AF,$4F,$D8,$BE,$4B,$F0,$73,$9C
	.dc.b	$7E,$24,$E5,$FF,$06,$46,$6F,$9D,$FC,$0F,$7B,$8B,$2B,$6E,$5A,$7A
	.dc.b	$BC,$4F,$80,$7C,$1F,$9E,$B1,$92,$5F,$AA,$32,$FD,$2E,$07,$D8,$FD
	.dc.b	$0E,$BF,$A2,$DB,$FB,$2D,$AE,$0D,$E3,$FF,$7C,$FD,$73,$E7,$7E,$DF
	.dc.b	$D8,$BE,$C3,$F4,$B3,$9C,$7F,$0D,$39,$7F,$C5,$D1,$9C,$7F,$D8,$0D
	.dc.b	$C3,$F6,$9C,$6F,$B1,$E7,$94,$C3,$8B,$F8,$18,$C6,$1E,$DF,$43,$F1
	.dc.b	$79,$FE,$31,$B7,$F9,$46,$CE,$E3,$F2,$F3,$F8,$BF,$7E,$47,$E2,$3E
	.dc.b	$02,$D7,$E1,$26,$E7,$E6,$9B,$5F,$EF,$BD,$3F,$8C,$DF,$E8,$39,$BF
	.dc.b	$BD,$5F,$C8,$F2,$7F,$13,$E5,$FC,$66,$61,$FA,$8F,$8B,$B8,$7C,$49
	.dc.b	$B7,$FA,$49,$B5,$FE,$71,$E0,$7C,$AF,$E8,$FE,$43,$F4,$8D,$9F,$D0
	.dc.b	$B7,$1F,$CE,$B7,$0F,$D7,$F8,$FF,$B2,$3F,$67,$EA,$DB,$FD,$DC,$5B
	.dc.b	$FC,$54,$DA,$57,$FA,$26,$5F,$62,$CF,$F6,$DF,$2B,$FC,$A2,$31,$5F
	.dc.b	$B4,$8C,$4C,$8F,$DB,$F6,$DB,$61,$D9,$FF,$4B,$61,$C9,$CD,$78,$63
	.dc.b	$13,$F8,$AA,$E4,$F1,$A7,$D8,$BE,$C7,$F9,$DB,$70,$FE,$39,$E0,$3D
	.dc.b	$FC,$D7,$F4,$0F,$43,$ED,$D7,$BD,$B8,$79,$AD,$1B,$E4,$FE,$8F,$F5
	.dc.b	$EF,$98,$A7,$F2,$2F,$B1,$FE,$7B,$8F,$FA,$46,$74,$DC,$38,$9F,$CA
	.dc.b	$3E,$C7,$39,$7C,$03,$FD,$71,$49,$EF,$E7,$CB,$FE,$6D,$6F,$4B,$18
	.dc.b	$71,$1E,$BB,$A7,$FA,$EB,$6F,$F9,$F7,$D8,$FE,$CD,$FF,$53,$46,$6B
	.dc.b	$92,$F9,$FE,$27,$F3,$EF,$B1,$7A,$F8,$BF,$F5,$77,$D8,$BE,$AF,$FB
	.dc.b	$5B,$63,$FA,$67,$88,$F1,$73,$5A,$F6,$E5,$6B,$DE,$BA,$FA,$F1,$FF
	.dc.b	$6D,$4F,$68,$FE,$9F,$F6,$3F,$A6,$8C,$1F,$EB,$F1,$F3,$7F,$DF,$1A
	.dc.b	$7F,$5D,$8F,$C2,$4F,$C7,$8C,$BF,$27,$EC,$5E,$A7,$F8,$11,$B3,$FD
	.dc.b	$97,$D2,$FC,$2D,$27,$2F,$D1,$FB,$17,$B1,$FE,$0C,$6D,$FF,$B4,$D7
	.dc.b	$C1,$66,$75,$F2,$12,$BF,$DB,$1F,$63,$FE,$DE,$78,$47,$F8,$5A,$31
	.dc.b	$87,$DF,$6F,$8D,$FA,$3F,$B1,$F9,$D7,$CB,$D4,$E1,$E7,$48,$FF,$72
	.dc.b	$7C,$3B,$DB,$C6,$5F,$89,$8C,$4E,$A7,$FC,$4B,$F6,$3F,$E0,$7A,$D4
	.dc.b	$63,$D0,$FF,$14,$DF,$09,$FC,$95,$6F,$3F,$DB,$4E,$5F,$DC,$7C,$17
	.dc.b	$E1,$6B,$FC,$17,$F5,$BF,$B5,$CB,$FD,$F1,$39,$AF,$F1,$E4,$E6,$BF
	.dc.b	$3D,$39,$AF,$E2,$CD,$8F,$E1,$B3,$FD,$DD,$DF,$5F,$F8,$FA,$75,$1F
	.dc.b	$E2,$28,$FF,$0C,$CE,$EF,$CB,$1B,$E2,$FF,$0E,$FD,$8F,$F5,$BC,$7B
	.dc.b	$C6,$3D,$CD,$47,$A2,$37,$F2,$B8,$CB,$CF,$F1,$3E,$13,$C4,$DA,$F7
	.dc.b	$C5,$8F,$69,$72,$FE,$3C,$7E,$C2,$51,$FB,$F8,$C3,$FC,$04,$CF,$8B
	.dc.b	$AF,$B1,$7C,$AF,$F3,$BA,$D9,$D3,$C5,$FF,$30,$B8,$C3,$FD,$CB,$87
	.dc.b	$F0,$A1,$7C,$9F,$0C,$ED,$FF,$75,$E1,$DF,$0E,$D8,$7F,$3D,$BE,$15
	.dc.b	$FD,$2E,$B4,$F2,$9D,$57,$82,$73,$3F,$D1,$A7,$2F,$63,$C7,$3D,$B0
	.dc.b	$DC,$7C,$03,$27,$57,$2F,$AE,$FC,$78,$7C,$87,$CB,$1C,$5C,$DF,$82
	.dc.b	$3D,$07,$43,$C9,$5D,$30,$F3,$DF,$3C,$3D,$55,$B8,$C3,$E9,$3E,$69
	.dc.b	$F3,$4B,$38,$7D,$47,$F7,$11,$F6,$3F,$C7,$3E,$2B,$8B,$63,$EB,$B5
	.dc.b	$E0,$D7,$FB,$D3,$FD,$1E,$B8,$F9,$9F,$DC,$A7,$8B,$5E,$2D,$78,$BB
	.dc.b	$CF,$ED,$E3,$FE,$AD,$9F,$AE,$D7,$E3,$5F,$76,$D7,$19,$FF,$06,$1B
	.dc.b	$AE,$7A,$7F,$83,$DF,$C3,$DF,$16,$D7,$1B,$FD,$0D,$F1,$39,$DF,$17
	.dc.b	$CF,$87,$E2,$B0,$E5,$1A,$AF,$3D,$6D,$FD,$5C,$E9,$B7,$DD,$FD,$4C
	.dc.b	$ED,$A7,$58,$D5,$7C,$1B,$FF,$0A,$37,$3F,$05,$7F,$75,$C3,$B3,$C8
	.dc.b	$FE,$F1,$3F,$E0,$E7,$F5,$B3,$F1,$BD,$4F,$DC,$3C,$AF,$FC,$41,$F6
	.dc.b	$3F,$DC,$AB,$67,$F0,$47,$0F,$F8,$0D,$B1,$C3,$1E,$73,$BD,$AC,$B9
	.dc.b	$1F,$2F,$86,$D7,$61,$87,$4C,$7F,$C8,$DF,$63,$FD,$2E,$75,$F6,$1F
	.dc.b	$AE,$9F,$F2,$53,$63,$F8,$CC,$ED,$FF,$0C,$37,$1F,$F1,$E4,$EE,$4D
	.dc.b	$9D,$57,$83,$D1,$F5,$D2,$7F,$C4,$0F,$9D,$AF,$F8,$96,$9F,$63,$C2
	.dc.b	$73,$87,$83,$9B,$9D,$7F,$76,$BE,$58,$6C,$CC,$C8,$C9,$F1,$7E,$C7
	.dc.b	$83,$FE,$33,$6D,$FF,$95,$DB,$97,$C4,$F8,$DF,$4C,$62,$BF,$9E,$FA
	.dc.b	$7F,$EE,$57,$E8,$35,$E9,$87,$C0,$8C,$61,$F2,$37,$F4,$99,$F9,$0D
	.dc.b	$79,$7B,$FB,$19,$8F,$48,$AF,$F9,$27,$EC,$7E,$6B,$FD,$39,$AF,$F9
	.dc.b	$0E,$7E,$7C,$7F,$A5,$A3,$EF,$F6,$FC,$83,$3F,$0D,$F5,$3E,$C5,$F0
	.dc.b	$1F,$5B,$A1,$EA,$D3,$51,$F6,$BE,$C7,$FC,$D9,$8F,$E3,$E9,$E4,$F1
	.dc.b	$35,$F0,$3B,$ED,$E9,$F2,$72,$FF,$2D,$87,$F9,$95,$BE,$FF,$E1,$BE
	.dc.b	$C5,$F3,$74,$F0,$47,$E2,$8F,$C1,$F5,$FF,$A8,$65,$F1,$7C,$4D,$76
	.dc.b	$EF,$EC,$7A,$74,$4C,$BF,$4D,$8A,$7F,$2B,$F6,$1F,$E6,$59,$FD,$57
	.dc.b	$D8,$FE,$FD,$FF,$3C,$37,$B3,$FD,$39,$3B,$9F,$F3,$E2,$F8,$0F,$CF
	.dc.b	$D3,$E8,$A9,$DA,$3F,$47,$F6,$3F,$EB,$7B,$EF,$D7,$C0,$CF,$C8,$6B
	.dc.b	$E0,$B9,$FD,$1C,$AF,$DF,$5F,$F9,$A9,$FE,$01,$F6,$1F,$E6,$47,$F8
	.dc.b	$17,$D8,$BE,$B3,$F6,$2D,$8D,$FC,$2F,$ED,$29,$F6,$3A,$F8,$A9,$5D
	.dc.b	$23,$F7,$3E,$17,$FA,$7E,$3F,$87,$35,$FF,$4E,$37,$8F,$FD,$81,$F5
	.dc.b	$D7,$AF,$FD,$AD,$36,$BF,$C4,$B3,$FA,$49,$FF,$6D,$36,$3F,$C5,$33
	.dc.b	$B7,$F2,$7F,$1F,$FC,$2A,$75,$06,$CE,$FE,$C5,$F3,$3F,$C4,$1B,$5F
	.dc.b	$C9,$5A,$FF,$AC,$DF,$E3,$46,$31,$FF,$6C,$37,$1F,$F7,$73,$67,$FC
	.dc.b	$7A,$F0,$F9,$5E,$DE,$6C,$9E,$D1,$E1,$30,$F7,$8C,$3F,$2B,$E9,$7E
	.dc.b	$69,$1F,$37,$D8,$D7,$FD,$91,$F6,$3F,$E4,$EC,$7F,$E0,$03,$63,$E3
	.dc.b	$D6,$F2,$FF,$78,$4E,$64,$DD,$15,$FC,$D2,$3F,$09,$3D,$B9,$30,$67
	.dc.b	$6B,$1F,$D3,$67,$78,$76,$AD,$4E,$7E,$4B,$A7,$14,$8C,$10,$F5,$B9
	.dc.b	$7D,$B1,$FD,$1B,$EB,$DF,$63,$49,$3F,$D7,$7E,$C1,$F6,$38,$8F,$F7
	.dc.b	$64,$7B,$1D,$6B,$9C,$DB,$FF,$9D,$25,$8F,$67,$64,$3F,$DE,$CE,$5E
	.dc.b	$BF,$1B,$FD,$6B,$0F,$E1,$F1,$E3,$F2,$31,$32,$30,$43,$89,$E9,$FA
	.dc.b	$9E,$44,$7D,$9E,$45,$7D,$71,$72,$F4,$C7,$FA,$33,$0F,$D0,$FD,$8B
	.dc.b	$C3,$FE,$DA,$D9,$FF,$01,$35,$57,$27,$F8,$A7,$C9,$F0,$AF,$EE,$BE
	.dc.b	$37,$F7,$B3,$FB,$E5,$F1,$BE,$0D,$DB,$E5,$4E,$FE,$13,$96,$3F,$E1
	.dc.b	$18,$75,$F4,$24,$67,$AE,$BE,$4B,$E7,$38,$C7,$CD,$7F,$E1,$CA,$D1
	.dc.b	$FE,$AC,$5E,$27,$A0,$91,$FE,$21,$FB,$1F,$D0,$4E,$A3,$C9,$3A,$79
	.dc.b	$89,$FD,$37,$D8,$FF,$39,$8F,$F1,$53,$FC,$FF,$1F,$C4,$36,$3F,$D7
	.dc.b	$AE,$DE,$6D,$C3,$E7,$3B,$83,$4D,$4F,$E4,$FE,$C7,$F8,$D9,$38,$BF
	.dc.b	$C4,$F3,$FE,$8D,$BF,$B1,$3F,$C9,$7E,$2F,$F7,$7F,$29,$FE,$46,$6E
	.dc.b	$3F,$E4,$19,$E7,$3F,$E3,$07,$F2,$F1,$F0,$1B,$5D,$27,$46,$F9,$A5
	.dc.b	$73,$8E,$5E,$CF,$FB,$17,$A7,$D0,$59,$FF,$6F,$E1,$FB,$15,$F1,$B6
	.dc.b	$3F,$CB,$ED,$AE,$12,$9E,$13,$BA,$CC,$70,$7B,$CF,$FB,$A1,$AF,$66
	.dc.b	$CF,$58,$D1,$E4,$8E,$7E,$AC,$8C,$9E,$18,$F1,$FD,$87,$F6,$98,$E8
	.dc.b	$7F,$BE,$97,$D8,$BE,$53,$8C,$61,$BE,$6F,$23,$9C,$EE,$B9,$8A,$E0
	.dc.b	$E1,$F0,$63,$F2,$5C,$B6,$79,$36,$FB,$D7,$8A,$73,$87,$F1,$DB,$E2
	.dc.b	$D8,$CC,$FA,$F0,$F2,$DF,$12,$47,$8E,$32,$F9,$67,$8B,$E9,$6D,$FA
	.dc.b	$2F,$B5,$FF,$BA,$23,$DB,$5F,$EE,$57,$FC,$97,$F1,$B9,$BF,$89,$39
	.dc.b	$F2,$3F,$D5,$99,$7C,$16,$E3,$FC,$DF,$D3,$FF,$2A,$F2,$FD,$09,$CD
	.dc.b	$FD,$56,$BF,$35,$B3,$C2,$3E,$74,$6B,$2F,$1C,$63,$2F,$C1,$D6,$F1
	.dc.b	$F1,$47,$72,$4F,$DA,$9D,$63,$F3,$F0,$E8,$4A,$E7,$EF,$FF,$38,$C3
	.dc.b	$9E,$1D,$C8,$FE,$9F,$0E,$71,$EB,$AE,$F3,$C0,$D5,$73,$C3,$C1,$3A
	.dc.b	$6D,$79,$67,$4F,$33,$BC,$6A,$B9,$E5,$C6,$75,$1D,$E7,$E8,$57,$29
	.dc.b	$F9,$97,$B6,$9D,$EF,$19,$7C,$CC,$39,$48,$FF,$32,$5F,$39,$D6,$1E
	.dc.b	$32,$4E,$F1,$F8,$6F,$6B,$E6,$D9,$F2,$E1,$E0,$CB,$E8,$4F,$8C,$F7
	.dc.b	$9E,$C9,$EA,$F8,$A7,$86,$5E,$BC,$3B,$E3,$E4,$AE,$04,$38,$1D,$0F
	.dc.b	$11,$98,$E3,$87,$8B,$1F,$3C,$62,$FB,$D7,$96,$73,$97,$C3,$BF,$CB
	.dc.b	$B6,$BF,$13,$8C,$BF,$C8,$7C,$37,$D3,$C7,$E2,$9F,$E5,$E9,$FC,$CF
	.dc.b	$9D,$9F,$8F,$D3,$23,$27,$E7,$A3,$17,$F9,$B6,$E7,$F9,$36,$EB,$FB
	.dc.b	$17,$FC,$23,$CD,$FE,$09,$C5,$18,$C3,$F6,$D9,$DC,$7E,$D4,$D5,$F6
	.dc.b	$2F,$87,$C2,$F1,$AC,$61,$B8,$7E,$27,$2F,$F0,$C5,$71,$8C,$1F,$8D
	.dc.b	$EA,$FE,$69,$D1,$C3,$1F,$EF,$B4,$F1,$FB,$9E,$0F,$81,$B8,$72,$AF
	.dc.b	$CB,$56,$E3,$11,$E8,$AE,$11,$89,$D4,$F6,$F7,$7D,$7F,$63,$EC,$6D
	.dc.b	$6D,$C3,$DF,$27,$CB,$7F,$BA,$23,$47,$8F,$47,$D8,$7C,$58,$C3,$71
	.dc.b	$FF,$11,$CE,$BC,$2F,$A0,$F0,$DE,$D6,$E3,$EC,$B6,$FE,$62,$AD,$D7
	.dc.b	$2F,$AA,$7C,$2D,$3A,$FC,$EF,$F6,$1C,$3C,$D3,$A6,$CE,$DF,$54,$EF
	.dc.b	$2F,$3D,$6D,$FF,$18,$FD,$8B,$EA,$BD,$3F,$58,$F9,$2F,$C4,$4E,$71
	.dc.b	$F8,$93,$97,$EA,$C6,$6F,$9D,$FC,$0F,$7B,$8B,$2B,$6E,$5A,$7A,$BC
	.dc.b	$5F,$80,$7C,$1F,$9E,$B1,$92,$5F,$AA,$32,$FD,$2E,$07,$D8,$FD,$0E
	.dc.b	$BF,$A2,$DB,$FB,$2D,$AE,$0D,$E3,$FF,$3A,$7D,$73,$E7,$7E,$DF,$D8
	.dc.b	$BE,$C3,$F5,$53,$9C,$7F,$0D,$39,$7F,$2F,$19,$C7,$FC,$BA,$DC,$3F
	.dc.b	$73,$C6,$FB,$1E,$79,$4C,$38,$BF,$81,$8C,$61,$ED,$F3,$FF,$17,$9F
	.dc.b	$E3,$1B,$7F,$94,$6C,$EE,$3F,$2F,$3F,$8B,$F7,$E4,$7E,$23,$E0,$2D
	.dc.b	$7E,$12,$6E,$7E,$69,$B5,$FE,$B7,$D3,$F8,$F5,$FE,$83,$9B,$FB,$E5
	.dc.b	$FC,$8F,$27,$F1,$3E,$5F,$C6,$66,$1F,$A8,$F8,$BB,$87,$C4,$9B,$7F
	.dc.b	$A4,$9B,$5F,$C3,$F8,$1F,$2B,$FA,$3F,$90,$FD,$23,$67,$F4,$2D,$C7
	.dc.b	$F3,$AD,$C3,$F5,$FE,$37,$EC,$8F,$D9,$FA,$B6,$FF,$77,$16,$FF,$15
	.dc.b	$36,$BF,$D6,$DF,$62,$CF,$F6,$9F,$2B,$FC,$AE,$31,$8F,$EC,$A3,$13
	.dc.b	$23,$F6,$BD,$B7,$1F,$F4,$FD,$F0,$6D,$7E,$FE,$75,$87,$FB,$17,$EC
	.dc.b	$5F,$63,$FC,$E5,$B8,$FF,$15,$F0,$3F,$D7,$AF,$F2,$68,$CE,$3F,$EA
	.dc.b	$CF,$3F,$EA,$CF,$FB,$BD,$B1,$FE,$D2,$A4,$7F,$45,$33,$0F,$B0,$7C
	.dc.b	$C7,$F4,$96,$DF,$F2,$5F,$B1,$FE,$63,$9F,$FB,$6E,$73,$7F,$EB,$C5
	.dc.b	$8F,$13,$F9,$2F,$D8,$EA,$B8,$47,$5F,$3A,$63,$FC,$C6,$9E,$8E,$31
	.dc.b	$FF,$63,$FA,$CE,$7F,$EB,$8D,$BF,$E7,$1F,$63,$FA,$D7,$FA,$4C,$66
	.dc.b	$B8,$FC,$FF,$03,$F9,$BF,$D8,$BD,$6C,$5F,$F5,$0F,$B1,$7D,$5F,$F6
	.dc.b	$96,$C7,$F4,$5F,$0D,$E2,$66,$B5,$ED,$4A,$D7,$BB,$74,$FF,$6F,$E3
	.dc.b	$FE,$33,$9E,$D1,$FD,$23,$EC,$7F,$3F,$18,$3F,$B0,$8F,$9B,$FE,$F4
	.dc.b	$D3,$FA,$B4,$7E,$03,$99,$1F,$E3,$FA,$B5,$FE,$01,$FB,$17,$A7,$4F
	.dc.b	$EA,$FE,$8F,$E0,$23,$37,$FD,$C3,$EC,$5E,$BF,$F8,$25,$B1,$9F,$81
	.dc.b	$C4,$E7,$E4,$24,$FF,$78,$FB,$17,$A5,$8D,$7F,$DF,$F1,$89,$D4,$FF
	.dc.b	$85,$7E,$C7,$E5,$C6,$0E,$5E,$67,$F8,$66,$BF,$1F,$1E,$DF,$87,$7B
	.dc.b	$59,$AF,$F0,$F4,$62,$7F,$C1,$2D,$C5,$C5,$F1,$9F,$F3,$27,$D8,$FF
	.dc.b	$80,$5C,$D7,$5C,$3F,$C5,$13,$9A,$E1,$5F,$ED,$19,$CF,$9D,$E0,$7F
	.dc.b	$C5,$DF,$03,$F8,$69,$FF,$02,$FD,$77,$ED,$9E,$A7,$F9,$CE,$35,$7F
	.dc.b	$E3,$C9,$D3,$DA,$33,$8F,$DB,$6D,$7A,$AB,$FB,$93,$D0,$FF,$0D,$F5
	.dc.b	$F8,$89,$5F,$E1,$B9,$DD,$F5,$9F,$3F,$17,$F8,$63,$EC,$7F,$AB,$5F
	.dc.b	$78,$C7,$B7,$A8,$FF,$19,$E4,$9E,$1F,$2B,$9C,$FC,$53,$FE,$54,$AD
	.dc.b	$FC,$47,$C4,$9F,$8A,$C8,$C9,$D2,$30,$7A,$F3,$FD,$A4,$7F,$90,$9B
	.dc.b	$1E,$F3,$FD,$39,$E2,$FF,$A7,$7E,$C5,$F2,$BF,$A1,$AD,$9F,$60,$B7
	.dc.b	$9F,$17,$FC,$C9,$87,$21,$7C,$9F,$F3,$04,$69,$61,$DF,$0E,$57,$67
	.dc.b	$F6,$06,$8E,$71,$A6,$9A,$9E,$0F,$E5,$2B,$F2,$97,$E1,$8F,$DD,$4E
	.dc.b	$B0,$F8,$33,$99,$F4,$3B,$8C,$DF,$8A,$7C,$53,$9B,$EB,$39,$BE,$91
	.dc.b	$8A,$F2,$1E,$63,$C9,$87,$99,$EF,$5E,$BB,$F1,$47,$68,$C1,$D5,$F1
	.dc.b	$1F,$34,$B7,$87,$07,$F1,$7F,$B1,$FE,$3D,$F1,$3C,$1B,$1E,$B6,$BF
	.dc.b	$39,$AF,$1F,$33,$F7,$AE,$DA,$ED,$AE,$F3,$ED,$1C,$31,$FA,$18,$FF
	.dc.b	$AB,$2B,$EA,$B5,$F8,$97,$FD,$52,$FE,$CB,$63,$E4,$4E,$EB,$9E,$9F
	.dc.b	$CC,$DF,$B9,$5B,$6E,$1E,$1A,$FE,$70,$71,$2D,$71,$7D,$17,$E2,$9E
	.dc.b	$0F,$FB,$F6,$38,$36,$78,$3F,$F0,$06,$FC,$13,$CA,$FC,$B3,$EE,$AD
	.dc.b	$AA,$DB,$CC,$FF,$59,$D7,$C3,$C3,$F9,$1B,$73,$E8,$FA,$27,$7E,$9F
	.dc.b	$EC,$E7,$DA,$D8,$E8,$7F,$19,$FB,$1F,$EE,$4F,$FB,$01,$AE,$76,$E0
	.dc.b	$A3,$FC,$0A,$EE,$3A,$60,$E5,$7C,$BE,$1B,$5B,$3F,$D9,$BF,$61,$FD
	.dc.b	$29,$FF,$67,$FD,$87,$EA,$F3,$FE,$D4,$D8,$09,$E0,$FF,$31,$6E,$1F
	.dc.b	$D1,$A7,$84,$9B,$3A,$AF,$EE,$CF,$F8,$6E,$7C,$7E,$8F,$BA,$31,$1F
	.dc.b	$CF,$69,$F6,$39,$AF,$87,$1F,$CC,$E7,$D3,$CF,$97,$99,$82,$1E,$79
	.dc.b	$F9,$7F,$63,$C6,$3F,$A5,$37,$2F,$EB,$ED,$CB,$E1,$FC,$67,$0C,$3E
	.dc.b	$9F,$A7,$FE,$D9,$C3,$FD,$EE,$D7,$9C,$6B,$E3,$FC,$FB,$C9,$DC,$72
	.dc.b	$F2,$F9,$FD,$8C,$9E,$71,$5F,$EB,$DF,$63,$F2,$5F,$A6,$D5,$8A,$BF
	.dc.b	$0F,$3F,$E8,$A9,$FB,$BD,$DF,$20,$CF,$C2,$7E,$33,$EC,$5F,$11,$F8
	.dc.b	$EE,$77,$AB,$4D,$47,$E4,$BE,$C7,$FC,$CB,$8F,$E1,$39,$3F,$C0,$B6
	.dc.b	$F2,$39,$1B,$F8,$1E,$1E,$17,$5F,$92,$98,$71,$E7,$FF,$1C,$65,$F9
	.dc.b	$C8,$FE,$F8,$BD,$FF,$D0,$FD,$8B,$E6,$A9,$E3,$8F,$D2,$9F,$88,$D7
	.dc.b	$FC,$2F,$2F,$8A,$E2,$6B,$B7,$7F,$63,$D3,$AA,$67,$E3,$3F,$3F,$8A
	.dc.b	$7F,$63,$F6,$3F,$E6,$49,$D0,$FB,$1F,$F4,$5B,$86,$BF,$E1,$58,$E0
	.dc.b	$DF,$80,$FD,$DC,$BE,$8A,$9D,$A3,$F7,$FF,$63,$FE,$B1,$AE,$9D,$7C
	.dc.b	$0C,$FC,$76,$BE,$07,$9F,$D1,$CA,$F2,$DF,$F9,$A5,$F3,$FD,$87,$F9
	.dc.b	$89,$F4,$FD,$8B,$EB,$3F,$86,$B6,$37,$F0,$BF,$C4,$29,$F6,$3A,$F8
	.dc.b	$A9,$5D,$23,$F8,$AF,$85,$FE,$A0,$8F,$8A,$D7,$FC,$6C,$DE,$9F,$F2
	.dc.b	$5F,$D7,$DE,$DF,$F3,$C4,$DA,$FE,$1F,$3F,$AB,$AD,$F4,$76,$F8,$F9
	.dc.b	$5F,$4A,$B5,$87,$FA,$05,$B1,$FC,$63,$0E,$8B,$EC,$5F,$31,$FC,$B1
	.dc.b	$B1,$F7,$1A,$FF,$93,$5F,$E3,$66,$32,$FF,$31,$B7,$0F,$F4,$83,$67
	.dc.b	$F9,$3B,$C3,$EA,$BB,$79,$D2,$7B,$47,$84,$C3,$DE,$30,$FC,$AF,$A5
	.dc.b	$F9,$A4,$7C,$DF,$63,$5F,$F3,$17,$D8,$FF,$8F,$F1,$FF,$51,$36,$3E
	.dc.b	$15,$6F,$2F,$F3,$DC,$E6,$4D,$D1,$5F,$D4,$23,$F0,$93,$DB,$97,$06
	.dc.b	$76,$B1,$FA,$49,$DE,$1D,$AB,$53,$9F,$92,$E9,$C7,$23,$04,$3D,$6E
	.dc.b	$5F,$6C,$7F,$62,$FA,$F7,$D8,$D2,$4F,$ED,$3E,$C1,$F6,$38,$8F,$F4
	.dc.b	$34,$7B,$1D,$6B,$9C,$FF,$D3,$B8,$74,$9C,$90,$FF,$47,$39,$79,$FC
	.dc.b	$6F,$F7,$4B,$FE,$1D,$1E,$1F,$23,$13,$23,$04,$38,$9E,$7F,$A9,$E4
	.dc.b	$47,$D5,$D3,$5F,$4C,$5C,$7F,$D6,$78,$7E,$7A,$5F,$62,$F0,$FF,$C0
	.dc.b	$ED,$8F,$E2,$ED,$7F,$D5,$1E,$2B,$E4,$F7,$5F,$E0,$E8,$FE,$F2,$7F
	.dc.b	$7A,$AE,$18,$7F,$B0,$9B,$CF,$FE,$EF,$9D,$FC,$27,$2C,$7F,$92,$C3
	.dc.b	$AF,$A1,$23,$3D,$75,$F1,$C5,$C7,$FD,$9F,$5F,$E1,$89,$F3,$F8,$9E
	.dc.b	$52,$3F,$CC,$BE,$C7,$F3,$52,$9C,$CF,$03,$59,$7F,$8A,$BE,$C5,$E4
	.dc.b	$DF,$86,$95,$FC,$F2,$72,$D9,$CB,$FE,$CF,$8F,$E7,$B9,$7F,$B9,$1D
	.dc.b	$C7,$FB,$4F,$EC,$7E,$A1,$39,$3C,$11,$FE,$3A,$8F,$E2,$BF,$17,$FB
	.dc.b	$0F,$96,$FF,$22,$36,$7F,$A6,$CF,$29,$E1,$D1,$27,$31,$CC,$C3,$E1
	.dc.b	$8C,$3C,$BD,$AF,$F5,$5F,$5F,$F8,$EA,$17,$FA,$37,$FD,$7B,$19,$8C
	.dc.b	$7C,$6E,$5F,$F2,$9B,$6B,$9D,$7F,$91,$27,$75,$C4,$CC,$76,$9B,$3A
	.dc.b	$8E,$13,$83,$33,$63,$FD,$BB,$1E,$3F,$5B,$89,$98,$F3,$9F,$E5,$4F
	.dc.b	$B0,$FE,$B8,$E0,$2F,$B1,$7C,$B7,$49,$E1,$18,$78,$4F,$2C,$BE,$23
	.dc.b	$70,$F9,$0F,$3B,$E5,$5D,$AF,$95,$F0,$F3,$FE,$0D,$EE,$BA,$D6,$B0
	.dc.b	$E6,$D8,$F5,$57,$08,$C4,$75,$C3,$BD,$F3,$C3,$B4,$E2,$B7,$39,$C3
	.dc.b	$AD,$F0,$C3,$A3,$B2,$5F,$C0,$8C,$BE,$68,$CE,$1D,$A3,$07,$FB,$B6
	.dc.b	$FD,$8F,$FC,$00,$7F,$C8,$FC,$7F,$EE,$07,$31,$F2,$3C,$9F,$F1,$2D
	.dc.b	$7C,$C6,$E3,$F9,$3E,$4F,$F2,$8F,$2F,$BA,$73,$7F,$75,$AF,$DC,$6C
	.dc.b	$F0,$8F,$B3,$1A,$C3,$C7,$18,$C7,$EB,$D6,$F1,$F1,$47,$72,$4F,$E2
	.dc.b	$67,$58,$FA,$67,$35,$C2,$B5,$EF,$B2,$CD,$EE,$7C,$06,$A3,$2F,$AE
	.dc.b	$EB,$FC,$4A,$B9,$3C,$2F,$B1,$2F,$E3,$4E,$A3,$13,$CC,$E1,$3A,$BE
	.dc.b	$B8,$FC,$2A,$DD,$77,$9C,$E5,$F1,$CF,$97,$87,$5C,$3D,$57,$C3,$0E
	.dc.b	$51,$D7,$0E,$E7,$86,$7A,$E5,$F2,$23,$13,$C8,$8F,$FB,$2A,$34,$FD
	.dc.b	$BC,$BC,$67,$D6,$C3,$C5,$1C,$88,$6C,$F9,$87,$88,$C5,$F8,$A3,$8C
	.dc.b	$62,$FC,$2D,$3C,$51,$AC,$3D,$B1,$D9,$AF,$C7,$3B,$38,$33,$87,$A9
	.dc.b	$B3,$CF,$1F,$97,$97,$DE,$AF,$79,$3E,$1B,$CA,$E3,$CE,$A6,$3E,$3E
	.dc.b	$D7,$E6,$22,$E5,$F9,$59,$BA,$7F,$39,$B4,$63,$4F,$D6,$36,$BE,$83
	.dc.b	$F8,$EE,$9F,$C7,$C7,$F6,$28,$C6,$3F,$B3,$DA,$CF,$08,$FF,$4F,$FD
	.dc.b	$8B,$E2,$33,$28,$78,$D8,$75,$5C,$23,$15,$23,$FD,$5D,$D5,$FC,$D3
	.dc.b	$9F,$18,$33,$F2,$E2,$F1,$FA,$1D,$7E,$06,$E3,$E4,$8B,$7C,$63,$FA
	.dc.b	$F5,$2B,$37,$D3,$DD,$F3,$7D,$8F,$9D,$B7,$EE,$70,$F9,$BE,$4F,$FD
	.dc.b	$B8,$E4,$FF,$63,$53,$EC,$3E,$0C,$61,$B9,$78,$CF,$62,$F0,$7E,$6D
	.dc.b	$71,$E0,$B3,$C2,$3E,$B3,$6F,$FD,$DC,$F1,$7B,$4F,$2F,$AA,$7C,$2D
	.dc.b	$3A,$FC,$EF,$F6,$0A,$F3,$CE,$9B,$3B,$7D,$73,$BC,$BD,$18,$AF,$84
	.dc.b	$7A,$C7,$D8,$BE,$AB,$D7,$F5,$8F,$91,$FC,$2C,$E7,$2F,$8F,$39,$7F
	.dc.b	$DD,$31,$9B,$E8,$DC,$7E,$0F,$BD,$CB,$19,$7C,$1F,$8F,$18,$BF,$3E
	.dc.b	$4F,$13,$E2,$1F,$17,$E7,$6E,$1F,$0A,$2D,$70,$9B,$8F,$D0,$FB,$1F
	.dc.b	$A7,$D7,$F5,$1B,$87,$DD,$6D,$FC,$6E,$24,$62,$FF,$5B,$F5,$CF,$9D
	.dc.b	$FB,$FF,$62,$FB,$0F,$D4,$CE,$72,$FC,$5C,$E5,$65,$F3,$E7,$4D,$BF
	.dc.b	$C0,$7D,$8F,$26,$57,$0C,$7C,$31,$A8,$C1,$FE,$41,$BE,$73,$F8,$EF
	.dc.b	$43,$F2,$39,$7E,$49,$B5,$F9,$B6,$E3,$EF,$6D,$7F,$0B,$E0,$7B,$FF
	.dc.b	$93,$3F,$29,$F0,$17,$0F,$C8,$C5,$C3,$EB,$CD,$BF,$8F,$E4,$FE,$2B
	.dc.b	$CC,$F8,$AD,$C3,$D3,$7F,$13,$4F,$F7,$CC,$62,$11,$87,$EA,$7C,$5F
	.dc.b	$29,$D4,$E7,$0E,$F1,$FB,$79,$D5,$7E,$2B,$E5,$7F,$59,$F2,$1F,$AD
	.dc.b	$6D,$7E,$A9,$B3,$B8,$FD,$6C,$FE,$EB,$C6,$91,$FC,$0F,$D5,$B5,$F8
	.dc.b	$79,$B8,$FE,$FA,$6D,$7F,$9D,$7E,$C5,$97,$EF,$7E,$57,$F9,$7C,$62
	.dc.b	$7F,$77,$18,$99,$1F,$BE,$ED,$B9,$7E,$DE,$F8,$36,$BF,$84,$4E,$B1
	.dc.b	$FE,$87,$F6,$0F,$B1,$FE,$7A,$DC,$7F,$8F,$F7,$FF,$AE,$DF,$A1,$19
	.dc.b	$C7,$F8,$27,$9D,$FE,$33,$C3,$FA,$6B,$6B,$F2,$96,$8F,$E9,$26,$6B
	.dc.b	$F3,$FF,$60,$F9,$8F,$E9,$AD,$BF,$E5,$FF,$63,$FD,$0D,$FC,$04,$66
	.dc.b	$BF,$8A,$2C,$38,$9F,$CB,$BE,$C7,$59,$7F,$60,$9D,$5F,$2F,$37,$B9
	.dc.b	$BF,$3E,$C7,$E3,$7D,$67,$37,$F6,$06,$C7,$F4,$2F,$B1,$78,$59,$AD
	.dc.b	$7C,$FC,$AF,$E8,$FF,$62,$F5,$71,$C1,$FC,$AF,$EC,$5F,$57,$FD,$B1
	.dc.b	$B7,$FD,$3B,$C2,$FD,$6B,$F9,$F8,$CD,$71,$F6,$F8,$1F,$D3,$29,$EE
	.dc.b	$E2,$F8,$57,$53,$FA,$8F,$D8,$BC,$2C,$4F,$EA,$C7,$CD,$FF,$7E,$69
	.dc.b	$FD,$82,$7F,$01,$CD,$FE,$0A,$8C,$DF,$F4,$5F,$B1,$FE,$65,$1F,$B5
	.dc.b	$6C,$7F,$4C,$F4,$3F,$00,$A3,$35,$FE,$14,$FB,$17,$AF,$FE,$0E,$6E
	.dc.b	$1F,$BF,$5F,$03,$99,$D7,$C8,$4A,$FE,$09,$F6,$0F,$43,$19,$FF,$59
	.dc.b	$8C,$4E,$A7,$FC,$37,$F6,$3F,$3A,$B5,$1F,$E1,$EA,$FC,$84,$71,$F3
	.dc.b	$3F,$C4,$3F,$0C,$F6,$35,$18,$66,$3F,$C5,$7E,$C7,$FC,$13,$7F,$DA
	.dc.b	$6F,$84,$E7,$CE,$DC,$FF,$6D,$AD,$F0,$7F,$8D,$AF,$73,$FD,$C7,$E0
	.dc.b	$7F,$03,$3F,$E3,$BF,$AD,$FD,$A4,$E7,$B3,$FC,$1B,$18,$9F,$F1,$FD
	.dc.b	$70,$CB,$FA,$6D,$A3,$0F,$F8,$C6,$BF,$BB,$3C,$E7,$F9,$74,$EA,$38
	.dc.b	$5F,$6F,$7F,$FC,$87,$39,$3E,$AD,$7F,$89,$BE,$C7,$FA,$7B,$E0,$8C
	.dc.b	$DF,$F8,$3A,$72,$FF,$60,$8C,$E3,$E8,$E4,$FF,$13,$D3,$E5,$7B,$C7
	.dc.b	$83,$E2,$26,$3E,$6C,$69,$F1,$E2,$E1,$F0,$A2,$CF,$C2,$8F,$DA,$3F
	.dc.b	$D2,$DB,$1F,$E1,$F3,$FB,$9F,$8F,$FD,$D3,$EC,$5F,$2B,$C7,$C4,$FF
	.dc.b	$31,$DF,$03,$FC,$DA,$5C,$3E,$7E,$1C,$85,$F2,$7F,$BC,$57,$F9,$B1
	.dc.b	$D4,$65,$FF,$32,$47,$FB,$1A,$31,$1A,$BF,$15,$79,$5E,$F0,$C3,$0F
	.dc.b	$69,$AF,$E0,$A2,$E1,$F0,$2F,$AC,$F1,$3C,$F7,$D9,$A7,$B2,$B8,$E1
	.dc.b	$E5,$9F,$0C,$63,$0E,$F5,$E7,$3C,$47,$23,$C4,$5C,$3C,$55,$CA,$31
	.dc.b	$7D,$5F,$01,$EE,$3E,$61,$6F,$0E,$0F,$CF,$FD,$8F,$F1,$CF,$8B,$E0
	.dc.b	$D8,$F8,$ED,$7E,$BB,$5E,$3E,$67,$FA,$55,$DB,$5D,$B5,$DE,$3D,$A3
	.dc.b	$86,$7F,$6F,$2F,$F5,$BD,$7D,$26,$BF,$0E,$FF,$D3,$17,$F5,$9B,$1F
	.dc.b	$1E,$77,$5C,$F4,$FD,$DB,$F8,$7A,$DB,$70,$F0,$D7,$F0,$03,$89,$6B
	.dc.b	$8B,$E8,$BF,$14,$F0,$7F,$CA,$71,$C1,$B3,$C1,$FF,$2B,$DF,$82,$79
	.dc.b	$5F,$96,$7D,$D5,$B5,$5B,$79,$9F,$E3,$BA,$F8,$78,$7F,$B5,$9B,$A7
	.dc.b	$47,$D1,$3B,$F4,$FF,$75,$3E,$D6,$C7,$43,$FD,$31,$F6,$3F,$DC,$5F
	.dc.b	$F2,$43,$5C,$ED,$C1,$47,$FB,$25,$DC,$74,$C1,$CA,$F9,$7C,$36,$B6
	.dc.b	$7F,$95,$7E,$C7,$FA,$33,$FE,$57,$FB,$0F,$DE,$5F,$FB,$C9,$B9,$FF
	.dc.b	$34,$AD,$D7,$FA,$BD,$B5,$BD,$3C,$D7,$C2,$BF,$D0,$F5,$AB,$E5,$E7
	.dc.b	$FA,$E3,$FC,$33,$3E,$38,$E3,$F6,$39,$79,$F2,$C9,$D1,$0F,$F6,$06
	.dc.b	$1F,$DC,$1F,$54,$71,$7F,$CE,$5F,$63,$97,$FD,$8C,$DC,$D9,$7F,$50
	.dc.b	$8C,$3F,$D4,$AF,$FD,$91,$5C,$AF,$E9,$7A,$3F,$EE,$78,$FF,$3F,$35
	.dc.b	$EB,$1A,$F8,$FF,$87,$BC,$9D,$C7,$2F,$2B,$9F,$D8,$C9,$E7,$15,$FF
	.dc.b	$6E,$7D,$8F,$C9,$7F,$D6,$2D,$58,$AB,$F0,$D3,$FE,$29,$9F,$B9,$D9
	.dc.b	$F5,$0C,$F9,$FF,$8B,$FB,$17,$C3,$7E,$37,$9F,$FC,$6B,$7F,$E9,$96
	.dc.b	$BF,$EE,$D9,$DB,$FE,$9D,$FB,$1F,$F3,$23,$BF,$1F,$91,$F9,$78,$CC
	.dc.b	$7F,$BD,$5F,$CC,$3F,$EF,$7A,$7C,$97,$84,$E1,$E0,$71,$F6,$FF,$41
	.dc.b	$F6,$2F,$9A,$A6,$70,$FF,$57,$4F,$E2,$97,$4F,$DC,$7F,$DA,$F4,$F8
	.dc.b	$8E,$67,$EA,$EB,$85,$6E,$90,$FB,$1F,$0F,$34,$9E,$F1,$FB,$3F,$B1
	.dc.b	$FF,$32,$C7,$FA,$FF,$EC,$7F,$D1,$4D,$F1,$35,$E6,$EF,$DD,$FD,$D5
	.dc.b	$3E,$86,$99,$BF,$F6,$47,$D8,$FF,$AB,$9E,$9D,$3C,$0F,$E0,$51,$98
	.dc.b	$F7,$F8,$3F,$E2,$88,$FF,$32,$C7,$8F,$E8,$BB,$9C,$3D,$EF,$E1,$5F
	.dc.b	$60,$F9,$9F,$E1,$9F,$62,$FA,$CA,$6A,$37,$F6,$3A,$F8,$59,$97,$FB
	.dc.b	$77,$04,$FF,$15,$F8,$5F,$F4,$F4,$FF,$B8,$5A,$F9,$81,$FE,$82,$F8
	.dc.b	$CF,$E3,$9F,$5F,$AE,$AF,$A5,$57,$83,$FC,$87,$3F,$AD,$9F,$AC,$D8
	.dc.b	$FF,$23,$CE,$EF,$AB,$54,$ED,$F0,$B5,$F1,$1F,$EF,$7F,$B1,$7C,$BF
	.dc.b	$F2,$F6,$C6,$63,$1E,$9E,$59,$19,$39,$47,$1F,$AA,$E5,$E7,$49,$ED
	.dc.b	$19,$7C,$11,$97,$E3,$0B,$4F,$CD,$FA,$59,$3F,$35,$F3,$7D,$8E,$23
	.dc.b	$8F,$D8,$FF,$8F,$B5,$FF,$35,$B6,$3E,$2D,$6F,$6F,$D0,$4E,$64,$DD
	.dc.b	$95,$FD,$46,$3F,$0D,$3F,$6A,$B8,$F2,$64,$D6,$F6,$75,$8F,$8A,$71
	.dc.b	$1E,$69,$E9,$F3,$9D,$79,$24,$F0,$3F,$0F,$8A,$77,$19,$3C,$F3,$C1
	.dc.b	$F8,$31,$FD,$9F,$EB,$DF,$63,$49,$3C,$7E,$C1,$F6,$38,$8F,$A3,$1F
	.dc.b	$01,$D6,$D9,$DB,$F8,$36,$1D,$27,$24,$3E,$CB,$97,$9F,$C5,$7F,$79
	.dc.b	$BF,$E2,$51,$E1,$F2,$71,$32,$30,$43,$89,$E7,$FA,$9E,$44,$7D,$5E
	.dc.b	$4D,$7D,$31,$71,$FE,$39,$87,$E7,$FE,$C7,$FD,$2D,$5F,$25,$AF,$E4
	.dc.b	$1B,$3F,$C8,$FC,$47,$C8,$7F,$84,$67,$3C,$BF,$EB,$69,$FE,$FD,$7F
	.dc.b	$C6,$FA,$F5,$E2,$C8,$FA,$D3,$A3,$77,$CA,$3F,$81,$C7,$E7,$9D,$F5
	.dc.b	$7F,$87,$E3,$11,$FA,$47,$FC,$34,$D8,$D1,$1E,$FD,$9C,$7C,$49,$1E
	.dc.b	$27,$FC,$55,$F6,$3F,$9F,$B4,$6A,$7B,$3A,$CF,$FC,$61,$F6,$2F,$26
	.dc.b	$F9,$72,$FF,$25,$39,$6C,$E5,$FD,$DC,$7F,$44,$6C,$7F,$8E,$FE,$C7
	.dc.b	$F8,$91,$39,$3B,$CE,$5F,$F2,$3F,$C4,$FE,$D3,$E4,$AF,$79,$D5,$7E
	.dc.b	$8E,$F0,$DE,$9E,$C7,$79,$91,$92,$1F,$C4,$27,$8B,$7D,$6F,$F5,$97
	.dc.b	$67,$FB,$D2,$7F,$49,$3F,$EB,$98,$C1,$A7,$F9,$EC,$7F,$96,$7E,$2F
	.dc.b	$6D,$9D,$CF,$FB,$61,$AF,$33,$8D,$73,$53,$E0,$AE,$75,$98,$C3,$E5
	.dc.b	$AC,$39,$F5,$65,$F7,$3C,$B1,$D7,$EC,$3F,$B1,$46,$42,$FB,$17,$C8
	.dc.b	$E6,$30,$DF,$3F,$91,$CE,$77,$5E,$B3,$9B,$8C,$FE,$4D,$F8,$27,$4F
	.dc.b	$46,$CE,$F0,$FE,$CF,$1C,$B0,$F0,$3E,$39,$D3,$FD,$B2,$B9,$5F,$69
	.dc.b	$CC,$FB,$30,$F3,$5F,$12,$47,$8A,$32,$F8,$A3,$35,$E8,$9C,$BD,$63
	.dc.b	$38,$79,$B0,$F6,$4E,$6F,$91,$1E,$4F,$80,$F6,$9E,$D2,$9F,$4B,$C9
	.dc.b	$FD,$43,$D7,$1F,$9A,$F0,$A4,$FF,$8C,$2B,$E6,$B6,$FE,$4F,$17,$F9
	.dc.b	$7B,$97,$ED,$4E,$70,$F9,$0D,$7C,$1C,$BF,$83,$AE,$11,$F5,$A7,$58
	.dc.b	$F8,$2B,$99,$BF,$7B,$FC,$A1,$97,$8A,$33,$7C,$5D,$63,$89,$D4,$E4
	.dc.b	$FC,$46,$3E,$08,$C9,$32,$F9,$D9,$79,$E7,$51,$8C,$3A,$36,$F9,$E3
	.dc.b	$E4,$8C,$61,$CE,$7E,$0D,$79,$1F,$14,$65,$F8,$67,$2C,$7A,$BC,$0D
	.dc.b	$D7,$CE,$BE,$D1,$9A,$E5,$97,$C7,$8E,$47,$03,$55,$E4,$39,$5F,$68
	.dc.b	$F0,$D7,$8F,$1F,$7C,$70,$21,$AA,$E2,$F5,$30,$6A,$38,$B8,$25,$F9
	.dc.b	$6B,$F8,$EC,$EA,$31,$5E,$5A,$E5,$87,$B9,$E3,$97,$E1,$0F,$A9,$19
	.dc.b	$AE,$33,$E6,$6B,$FD,$2E,$35,$3C,$31,$F8,$18,$78,$63,$15,$F2,$31
	.dc.b	$F2,$FC,$27,$86,$BD,$67,$F9,$62,$FF,$31,$E7,$67,$E3,$74,$C8,$C9
	.dc.b	$F9,$D8,$C6,$1F,$90,$6E,$1F,$09,$BA,$7E,$A9,$FF,$56,$78,$7F,$EA
	.dc.b	$7C,$51,$8C,$3F,$5F,$B5,$C7,$FC,$FF,$5A,$9F,$F1,$C7,$D8,$BE,$1B
	.dc.b	$54,$87,$8F,$87,$55,$C2,$31,$52,$3F,$C8,$DD,$7F,$CD,$B9,$F1,$87
	.dc.b	$F9,$AB,$07,$91,$CC,$E9,$EF,$5C,$3C,$71,$6B,$84,$DD,$BF,$D8,$97
	.dc.b	$FE,$C4,$BF,$4F,$D8,$F9,$7D,$EF,$33,$63,$DB,$1F,$07,$37,$C9,$3E
	.dc.b	$0F,$59,$FF,$A8,$F2,$7D,$7F,$C0,$C7,$FC,$C1,$18,$E0,$F9,$4E,$BC
	.dc.b	$3B,$5B,$BE,$4D,$9F,$AA,$DB,$FF,$3B,$2A,$E3,$7D,$3E,$AF,$FD,$E3
	.dc.b	$5E,$06,$E1,$E5,$F9,$AF,$EC,$33,$9F,$85,$E0,$47,$CF,$1A,$9F,$65
	.dc.b	$72,$CB,$FD,$D9,$F6,$0F,$A9,$F5,$FD,$8B,$E6,$7F,$03,$39,$CB,$E0
	.dc.b	$CE,$5F,$F3,$F4,$67,$1F,$F5,$2B,$70,$F7,$FB,$FC,$B1,$97,$C1,$F8
	.dc.b	$F1,$8B,$F2,$F8,$DF,$04,$F8,$5F,$37,$6F,$E1,$45,$9F,$5C,$DC,$3E
	.dc.b	$6F,$D8,$FD,$2E,$CF,$A6,$DC,$3E,$CF,$1A,$31,$87,$D6,$6D,$AB,$FB
	.dc.b	$EF,$D7,$3E,$C3,$EF,$FD,$8B,$EC,$7F,$4B,$39,$AF,$C2,$CE,$5F,$F5
	.dc.b	$14,$67,$1F,$F6,$4B,$70,$FD,$BF,$1B,$EA,$39,$E5,$30,$E2,$FE,$0E
	.dc.b	$31,$87,$E3,$3D,$0F,$C7,$E7,$F9,$06,$DF,$B9,$B3,$B8,$F8,$33,$F8
	.dc.b	$DF,$81,$91,$F8,$DF,$82,$B5,$F8,$69,$B9,$F9,$A6,$D7,$C4,$F2,$BF
	.dc.b	$0F,$E6,$FC,$26,$C7,$C0,$70,$FD,$0F,$B0,$B8,$FC,$29,$B8,$7E,$92
	.dc.b	$6D,$7F,$B6,$76,$7C,$A7,$E8,$FE,$6B,$F4,$8D,$9F,$D2,$37,$1F,$CE
	.dc.b	$B7,$0F,$D7,$F9,$1F,$B4,$3F,$6B,$EB,$5B,$FC,$1C,$5B,$FD,$BC,$DA
	.dc.b	$FC,$14,$07,$D8,$B3,$FD,$B7,$89,$FC,$9A,$31,$8F,$ED,$E3,$13,$23
	.dc.b	$F7,$5D,$B6,$D8,$76,$7A,$D6,$63,$F0,$B3,$6B,$FC,$9F,$F6,$0F,$B1
	.dc.b	$FE,$6A,$D8,$FE,$2B,$DE,$F8,$0C,$D7,$F3,$AF,$43,$FC,$4F,$8F,$76
	.dc.b	$DF,$B6,$16,$9E,$1C,$9F,$D0,$BE,$B9,$F6,$34,$FE,$43,$F6,$3F,$CE
	.dc.b	$B1,$FF,$80,$13,$A6,$E1,$C4,$FE,$4F,$F6,$39,$E0,$F8,$33,$FE,$6A
	.dc.b	$AF,$0B,$CB,$CF,$F0,$1C,$3D,$2B,$1E,$73,$FC,$E3,$EB,$BA,$7F,$AB
	.dc.b	$B6,$3F,$9B,$FD,$8F,$EB,$5E,$18,$FD,$B9,$D3,$E2,$31,$F4,$1C,$8F
	.dc.b	$E7,$BF,$62,$F5,$F1,$7C,$A9,$F6,$2F,$9B,$FE,$CE,$D9,$FE,$8F,$E1
	.dc.b	$3C,$3C,$D6,$BD,$A9,$5A,$F7,$AE,$3E,$33,$A9,$D0,$FE,$9F,$F6,$3F
	.dc.b	$A4,$9F,$16,$5F,$95,$7C,$83,$E6,$BF,$BC,$B4,$FE,$B3,$3F,$87,$AE
	.dc.b	$D1,$97,$C1,$F6,$2F,$4F,$FC,$02,$D9,$FE,$C3,$E9,$7E,$22,$BF,$49
	.dc.b	$39,$7D,$1F,$62,$F6,$3F,$C1,$2D,$BF,$90,$BE,$0B,$33,$AF,$91,$95
	.dc.b	$F3,$7E,$C3,$FD,$D4,$F0,$8F,$F0,$9C,$63,$0F,$8C,$DF,$0F,$F5,$5F
	.dc.b	$63,$F3,$EF,$97,$A5,$C3,$CE,$91,$F6,$7E,$21,$ED,$63,$2F,$A1,$18
	.dc.b	$9D,$4F,$F8,$8B,$EC,$7F,$C0,$D1,$83,$F0,$4E,$2B,$8C,$EF,$CF,$FF
	.dc.b	$14,$DF,$39,$FB,$9F,$05,$F8,$57,$FC,$1D,$F5,$BF,$B7,$C7,$F1,$93
	.dc.b	$9D,$7F,$C7,$73,$9A,$E3,$D7,$FE,$34,$AD,$5F,$EE,$6B,$FB,$A5,$6A
	.dc.b	$39,$D6,$CF,$F0,$BC,$EA,$FF,$88,$5F,$2B,$FE,$3C,$FD,$5B,$FF,$0F
	.dc.b	$FD,$8F,$F5,$77,$FD,$BF,$19,$9F,$F1,$34,$E5,$E3,$C7,$E5,$3F,$C4
	.dc.b	$B0,$F9,$5C,$5F,$BB,$E2,$7D,$47,$AE,$EC,$64,$8F,$C8,$32,$7E,$97
	.dc.b	$1F,$DA,$47,$EC,$DB,$1F,$A9,$3F,$7C,$3E,$C5,$F2,$FF,$EF,$89,$D9
	.dc.b	$FC,$C4,$B5,$9F,$0F,$FC,$C3,$7C,$AF,$8B,$FC,$9E,$32,$BC,$6F,$E8
	.dc.b	$D3,$A8,$FE,$7B,$3F,$D2,$A3,$11,$CF,$0E,$D7,$C1,$FF,$76,$DD,$9E
	.dc.b	$D3,$5F,$E3,$E7,$AE,$78,$B4,$FF,$7B,$CE,$F0,$F8,$77,$E0,$9F,$31
	.dc.b	$B3,$CB,$7D,$5F,$21,$E0,$34,$60,$D9,$83,$9C,$7B,$E7,$BD,$FA,$A7
	.dc.b	$B4,$61,$F2,$9F,$DA,$CD,$14,$F6,$94,$FB,$26,$7E,$C7,$F8,$E7,$C5
	.dc.b	$66,$11,$87,$83,$5E,$1E,$0A,$78,$4B,$CB,$FE,$E3,$5C,$1A,$F7,$6B
	.dc.b	$E7,$8E,$17,$CA,$78,$E3,$F7,$F1,$FF,$57,$DF,$D7,$6B,$F3,$EF,$8B
	.dc.b	$6B,$8C,$FC,$33,$75,$CF,$5F,$8C,$FE,$2E,$F8,$B6,$B8,$DF,$F8,$56
	.dc.b	$F8,$9C,$EF,$8B,$E9,$9D,$CF,$C8,$6C,$70,$8E,$53,$C1,$FF,$10,$5F
	.dc.b	$82,$79,$61,$ED,$AD,$57,$EB,$1B,$5B,$8E,$67,$F6,$7A,$F8,$75,$FE
	.dc.b	$1F,$6E,$7D,$27,$F2,$91,$87,$D9,$E9,$FE,$DE,$7E,$33,$63,$A9,$FD
	.dc.b	$9F,$EC,$7F,$BB,$3F,$DD,$5A,$E7,$7E,$0A,$3F,$04,$EE,$3A,$60,$E4
	.dc.b	$7C,$C6,$1B,$5B,$3F,$C0,$5F,$63,$FD,$29,$FF,$02,$7D,$7F,$E8,$F3
	.dc.b	$FC,$F3,$63,$FC,$95,$3B,$7F,$BF,$37,$0F,$F2,$14,$EE,$4D,$9D,$57
	.dc.b	$EA,$5F,$F1,$04,$F8,$BD,$0F,$6C,$62,$3F,$C1,$54,$FB,$1C,$D7,$C2
	.dc.b	$8F,$EE,$53,$E8,$E8,$CB,$CC,$C1,$0F,$3C,$FC,$AF,$B1,$E3,$1F,$E1
	.dc.b	$76,$E5,$FE,$5B,$6E,$5F,$0B,$E3,$78,$61,$F4,$BD,$2F,$F7,$06,$1F
	.dc.b	$E2,$B6,$BC,$E3,$5F,$21,$E7,$DE,$4E,$E3,$97,$97,$CF,$EC,$64,$F3
	.dc.b	$8A,$FF,$8A,$FE,$C7,$E4,$BF,$E7,$B6,$AC,$55,$F8,$69,$FF,$48,$CF
	.dc.b	$DC,$EE,$F9,$06,$7E,$17,$F1,$7F,$62,$F7,$FF,$1B,$CE,$F5,$29,$A8
	.dc.b	$FC,$8F,$D8,$FF,$9A,$B1,$FF,$25,$53,$C8,$E2,$6B,$E0,$37,$E0,$F4
	.dc.b	$F9,$39,$7F,$E5,$1C,$3F,$32,$FF,$91,$D7,$BB,$F9,$EF,$B1,$7C,$DD
	.dc.b	$3C,$31,$FA,$23,$F0,$DD,$7F,$E9,$19,$7C,$57,$13,$5D,$BB,$FB,$1E
	.dc.b	$9D,$33,$3F,$F3,$16,$29,$FD,$6F,$D8,$FF,$98,$E7,$FC,$CD,$F6,$3F
	.dc.b	$E8,$B6,$F8,$DA,$F4,$37,$EF,$FE,$D2,$9F,$47,$4C,$DF,$F9,$BF,$EC
	.dc.b	$7F,$D6,$0F,$4E,$CE,$07,$EF,$23,$31,$FE,$69,$7F,$CC,$4F,$8B,$E9
	.dc.b	$3B,$1B,$F4,$7F,$81,$FD,$87,$F9,$7E,$7F,$CF,$9F,$62,$FA,$EF,$E0
	.dc.b	$ED,$8F,$F4,$04,$ED,$FF,$3B,$2F,$B1,$DF,$C4,$49,$DC,$7F,$10,$F0
	.dc.b	$7F,$D5,$B1,$FE,$89,$6B,$FE,$7F,$6F,$4F,$FB,$0F,$EB,$EF,$99,$FD
	.dc.b	$9E,$6D,$7F,$B4,$E7,$F4,$F5,$BE,$8E,$DF,$1D,$2B,$FD,$2D,$5A,$C3
	.dc.b	$FB,$83,$63,$FD,$B9,$87,$45,$F6,$2F,$98,$FE,$4E,$D8,$FF,$50,$35
	.dc.b	$FF,$4B,$3F,$C5,$CC,$65,$FE,$DB,$6E,$1F,$DF,$DB,$3F,$EE,$B7,$87
	.dc.b	$C9,$F6,$F3,$A4,$F6,$8F,$09,$87,$BC,$61,$F9,$5F,$4B,$F3,$48,$F9
	.dc.b	$BE,$C6,$BF,$EA,$BF,$AA,$FF,$1F,$63,$FE,$11,$6C,$7C,$2A,$DE,$5F
	.dc.b	$EF,$39,$CC,$9B,$A2,$BF,$A5,$C7,$E1,$67,$B7,$26,$0C,$ED,$63,$FD
	.dc.b	$91,$3B,$C3,$B5,$6A,$73,$F2,$1D,$38,$E4,$60,$87,$AD,$CB,$ED,$8F
	.dc.b	$EB,$BF,$5E,$FB,$1A,$49,$FF,$6C,$7D,$83,$EC,$71,$1F,$EC,$F8,$F6
	.dc.b	$3A,$D7,$39,$B7,$FF,$1C,$63,$D6,$B4,$43,$FD,$B0,$E5,$E9,$F1,$5F
	.dc.b	$DC,$30,$FE,$1B,$1E,$2F,$23,$13,$23,$04,$38,$9E,$8F,$A9,$E4,$47
	.dc.b	$D7,$D3,$5F,$54,$5C,$BD,$10,$18,$7E,$85,$7D,$8F,$FA,$55,$FF,$7E
	.dc.b	$35,$FF,$79,$36,$79,$C7,$F1,$6F,$90,$FF,$05,$CE,$74,$FF,$2E,$78
	.dc.b	$7F,$DE,$4F,$EF,$51,$9A,$F1,$35,$FE,$3F,$F0,$7F,$E1,$49,$DF,$89
	.dc.b	$7A,$E5,$7A,$72,$F0,$D2,$78,$75,$EA,$95,$A9,$C3,$C2,$7B,$78,$72
	.dc.b	$BA,$67,$E8,$8F,$F1,$2F,$D8,$FE,$83,$14,$6E,$BA,$D6,$DE,$27,$CD
	.dc.b	$AF,$D3,$FD,$8B,$C8,$FF,$17,$BF,$D0,$67,$89,$F4,$DC,$65,$F3,$1E
	.dc.b	$11,$D0,$38,$78,$98,$F2,$BA,$9D,$AB,$FC,$7B,$F6,$3F,$74,$9C,$9E
	.dc.b	$2A,$CB,$FE,$49,$F8,$AF,$DE,$7C,$8D,$F0,$0E,$AB,$C5,$78,$6F,$4F
	.dc.b	$63,$BC,$C8,$C9,$0F,$3C,$F1,$6F,$B3,$FE,$A1,$F8,$CF,$D7,$C2,$6C
	.dc.b	$61,$BD,$BF,$EA,$77,$FC,$B7,$19,$AE,$53,$92,$D7,$E7,$EB,$8B,$67
	.dc.b	$2F,$79,$B3,$A8,$C1,$FA,$7A,$EC,$D9,$EC,$D3,$B7,$AB,$C8,$E7,$1D
	.dc.b	$BE,$C3,$FB,$35,$7B,$97,$D8,$BE,$47,$96,$09,$CC,$61,$E3,$7F,$EC
	.dc.b	$76,$CF,$EE,$DD,$BF,$EF,$18,$D5,$75,$8C,$1B,$F9,$9F,$24,$67,$0F
	.dc.b	$23,$6B,$8E,$1E,$1A,$DB,$E3,$9D,$5F,$85,$AF,$A6,$76,$F8,$0F,$64
	.dc.b	$6B,$0F,$3E,$1F,$0E,$B6,$F8,$AF,$B1,$E1,$BA,$73,$AE,$38,$79,$A3
	.dc.b	$17,$CA,$31,$7C,$E7,$27,$27,$04,$8F,$99,$3E,$89,$FF,$79,$57,$CC
	.dc.b	$6F,$93,$FB,$97,$C0,$63,$8F,$EB,$47,$F8,$F7,$0F,$CD,$F1,$FF,$98
	.dc.b	$F9,$F3,$52,$FD,$57,$BC,$BF,$9A,$47,$F0,$7F,$7F,$F9,$BE,$3F,$87
	.dc.b	$AD,$4F,$DF,$6B,$F8,$76,$E3,$DE,$3A,$E1,$DE,$B9,$3F,$8E,$9D,$46
	.dc.b	$27,$64,$7D,$37,$E2,$C3,$F2,$18,$78,$32,$FF,$3E,$63,$D6,$7D,$38
	.dc.b	$F4,$77,$19,$C7,$D0,$F8,$70,$E9,$18,$3C,$13,$C1,$FC,$45,$73,$75
	.dc.b	$8F,$9C,$93,$E0,$AE,$73,$F0,$EF,$C0,$D3,$9C,$6B,$0F,$4C,$78,$0E
	.dc.b	$B5,$CF,$2E,$33,$AC,$3B,$D7,$F6,$9C,$7C,$8F,$58,$CB,$6F,$93,$D4
	.dc.b	$C1,$D8,$E5,$87,$8F,$2F,$98,$4C,$BE,$36,$5E,$28,$C5,$7A,$A3,$13
	.dc.b	$EE,$C7,$C7,$97,$9F,$1F,$95,$97,$92,$73,$8F,$E1,$B2,$F0,$61,$DB
	.dc.b	$2F,$35,$7B,$09,$F0,$DF,$31,$C7,$9D,$4C,$7C,$7D,$D7,$F1,$91,$6F
	.dc.b	$EC,$CD,$CF,$F4,$6D,$D3,$F0,$ED,$9F,$98,$FF,$85,$BA,$BF,$C1,$B8
	.dc.b	$A3,$18,$FE,$3B,$3B,$7C,$23,$FB,$37,$D8,$BE,$23,$32,$87,$8F,$87
	.dc.b	$55,$C2,$31,$52,$3F,$BD,$F5,$FF,$35,$E8,$C6,$1F,$DC,$30,$79,$1C
	.dc.b	$CE,$9F,$03,$70,$F1,$C5,$AE,$13,$76,$FF,$14,$DF,$FB,$5A,$FD,$3F
	.dc.b	$63,$E5,$F7,$BC,$CD,$8F,$6C,$7C,$0C,$DF,$29,$FE,$E7,$9C,$9F,$E1
	.dc.b	$2C,$1F,$5F,$F0,$23,$0D,$CF,$FC,$23,$3A,$F0,$BE,$5B,$C3,$82,$D6
	.dc.b	$E3,$EA,$36,$FF,$C6,$4A,$B7,$5C,$BE,$A5,$F0,$B4,$EB,$F3,$7F,$D7
	.dc.b	$F0,$F3,$4E,$9B,$3B,$7D,$53,$BC,$BC,$F5,$B7,$FC,$43,$F6,$0F,$A8
	.dc.b	$F4,$FD,$8B,$E4,$7F,$03,$39,$C7,$E2,$4E,$5F,$F1,$64,$66,$F9,$DF
	.dc.b	$C0,$F7,$B8,$B2,$B6,$E5,$A7,$AB,$C3,$F8,$07,$C1,$F9,$BB,$19,$25
	.dc.b	$FA,$A3,$2F,$D2,$E0,$7D,$8F,$D0,$EB,$FA,$2D,$BF,$B2,$DA,$E0,$DE
	.dc.b	$3F,$F3,$9F,$D8,$3E,$C3,$ED,$FD,$8B,$EC,$3F,$47,$39,$C7,$F0,$D3
	.dc.b	$97,$FC,$9D,$19,$C7,$FC,$9A,$DC,$3F,$69,$C6,$FB,$1E,$79,$4C,$38
	.dc.b	$BF,$81,$8C,$61,$ED,$F3,$FF,$17,$9F,$E3,$1B,$7F,$94,$6C,$EE,$3F
	.dc.b	$2F,$3F,$8B,$F8,$09,$1F,$88,$F8,$1B,$5F,$84,$9B,$9F,$9A,$6D,$7F
	.dc.b	$6C,$D3,$FD,$F4,$FE,$0B,$CB,$F8,$7E,$57,$E2,$1B,$1F,$0A,$3F,$43
	.dc.b	$F6,$18,$CB,$E1,$C5,$C7,$F4,$73,$6B,$FD,$29,$BB,$E5,$BF,$45,$F2
	.dc.b	$3F,$A3,$6C,$FE,$81,$B8,$FE,$71,$B8,$7E,$BB,$C7,$FD,$89,$FB,$2F
	.dc.b	$56,$DF,$EE,$A2,$DF,$E2,$66,$D7,$FA,$D3,$EC,$59,$FE,$CF,$E5,$BF
	.dc.b	$91,$C6,$31,$FD,$8C,$62,$64,$7E,$D3,$B2,$E3,$FE,$8E,$BE,$0D,$AF
	.dc.b	$DF,$4E,$B0,$FF,$61,$FD,$8B,$EC,$7F,$98,$B7,$1F,$E2,$9D,$FF,$EB
	.dc.b	$77,$FD,$39,$19,$C7,$FD,$37,$E7,$7F,$86,$70,$FF,$66,$B6,$BF,$D8
	.dc.b	$16,$8F,$E7,$A6,$6B,$FD,$B9,$F5,$EF,$99,$FE,$82,$DB,$FE,$49,$F6
	.dc.b	$3F,$CE,$1F,$F5,$64,$66,$BF,$D5,$6B,$0E,$27,$F2,$4F,$B1,$D6,$5F
	.dc.b	$EE,$09,$D5,$F2,$F3,$BB,$9B,$F4,$6C,$7F,$AE,$7D,$67,$3F,$F5,$46
	.dc.b	$C7,$F3,$7F,$B1,$FD,$63,$B9,$D7,$D0,$48,$FE,$79,$F6,$2F,$5B,$1B
	.dc.b	$FF,$B0,$BE,$C5,$F3,$7F,$D8,$1B,$7F,$D0,$FC,$27,$87,$9A,$D7,$B5
	.dc.b	$2B,$5E,$ED,$C7,$A1,$CC,$FE,$8B,$F6,$3F,$9D,$9F,$F6,$A6,$49,$FE
	.dc.b	$DF,$0F,$9A,$A7,$0E,$CB,$1F,$ED,$BF,$B1,$7A,$5F,$DD,$9B,$1F,$D6
	.dc.b	$7D,$0F,$B9,$82,$72,$FF,$B8,$3E,$C5,$EB,$FF,$7E,$6D,$FF,$B9,$D7
	.dc.b	$C0,$E6,$75,$F2,$32,$BF,$DD,$9F,$63,$FE,$DE,$78,$47,$F8,$22,$31
	.dc.b	$87,$FB,$B9,$BD,$CA,$FF,$3B,$FD,$8F,$CC,$79,$FA,$5C,$7C,$C9,$1F
	.dc.b	$EF,$9F,$86,$7B,$38,$CB,$FD,$F3,$18,$9D,$4F,$F8,$63,$EC,$7F,$BE
	.dc.b	$56,$A3,$1E,$77,$F8,$72,$F8,$4B,$3F,$23,$F4,$EF,$8C,$62,$3F,$D9
	.dc.b	$D7,$FE,$24,$F8,$1F,$C2,$C7,$F8,$0B,$EB,$7F,$68,$F3,$CB,$FD,$27
	.dc.b	$3A,$BF,$F1,$A5,$69,$E9,$19,$9D,$F6,$7F,$82,$AB,$FB,$75,$F3,$3F
	.dc.b	$0D,$5B,$6C,$F8,$CE,$1E,$FD,$BF,$C6,$61,$DB,$8F,$FC,$2B,$F6,$3F
	.dc.b	$D4,$E7,$CB,$18,$F6,$FD,$8C,$8F,$F1,$14,$BE,$53,$19,$78,$CF,$F1
	.dc.b	$4F,$C4,$7C,$33,$E2,$36,$BC,$27,$C2,$8C,$47,$1D,$3F,$41,$3A,$8F
	.dc.b	$6C,$FF,$94,$32,$FD,$B4,$E6,$7E,$B7,$89,$FE,$54,$7F,$51,$E4,$FE
	.dc.b	$A7,$EC,$5F,$2D,$FE,$C2,$C3,$8D,$79,$DA,$7D,$B2,$E1,$FE,$C8,$8E
	.dc.b	$62,$B9,$BF,$2B,$0F,$02,$C3,$C1,$87,$79,$FC,$4D,$70,$AF,$F6,$9D
	.dc.b	$69,$E5,$75,$C9,$DE,$F9,$46,$63,$C7,$5C,$F0,$F5,$BD,$63,$35,$C1
	.dc.b	$E4,$7A,$EF,$B5,$70,$9F,$74,$FA,$4D,$9F,$32,$FC,$0E,$6F,$C1,$87
	.dc.b	$B2,$B8,$E1,$D2,$31,$5D,$CE,$78,$77,$70,$78,$0F,$01,$67,$0F,$D0
	.dc.b	$3E,$97,$D8,$FF,$21,$F8,$9E,$0D,$8F,$4B,$5D,$F8,$69,$DF,$9D,$FB
	.dc.b	$B9,$DB,$5E,$AD,$7F,$98,$0A,$46,$63,$F9,$A4,$63,$0F,$B1,$97,$FA
	.dc.b	$AE,$6E,$BC,$B4,$FC,$05,$EE,$7E,$6B,$63,$F0,$95,$C6,$7E,$8B,$63
	.dc.b	$8E,$1E,$96,$E1,$DA,$A1,$D2,$75,$5E,$AB,$EC,$FB,$0E,$85,$C3,$CC
	.dc.b	$EE,$78,$D7,$B9,$AF,$EA,$A7,$6F,$F5,$59,$E6,$EE,$FC,$67,$F2,$58
	.dc.b	$F6,$36,$7F,$65,$AF,$78,$DB,$E5,$F5,$73,$7D,$5A,$F8,$4F,$E0,$1F
	.dc.b	$63,$FD,$DD,$BB,$EA,$31,$1F,$C4,$DF,$B2,$E8,$FF,$5D,$73,$BE,$5F
	.dc.b	$BE,$28,$C4,$EC,$FE,$39,$F6,$3F,$D3,$9F,$E2,$1F,$61,$FB,$1C,$FF
	.dc.b	$0E,$D8,$FE,$F1,$3B,$78,$F5,$FA,$A7,$50,$6C,$EF,$0F,$E9,$8F,$F8
	.dc.b	$72,$7F,$8F,$35,$F1,$FA,$5C,$FE,$C7,$E2,$CE,$6B,$FA,$A5,$FF,$76
	.dc.b	$BF,$3F,$56,$B0,$91,$82,$1D,$0F,$E5,$9F,$63,$C5,$FE,$78,$DB,$FE
	.dc.b	$C0,$DC,$FE,$0F,$6F,$D6,$8C,$57,$F8,$67,$D4,$FF,$6D,$3F,$CD,$DA
	.dc.b	$F4,$C7,$EE,$C6,$30,$F4,$EF,$F1,$CC,$FC,$7E,$BC,$CD,$FD,$8C,$C7
	.dc.b	$A4,$57,$FA,$97,$D8,$FC,$BD,$7F,$76,$D8,$F2,$57,$E3,$23,$FD,$19
	.dc.b	$1F,$73,$C0,$FA,$C6,$7E,$13,$F2,$3F,$62,$F8,$8F,$C9,$F4,$BD,$6A
	.dc.b	$6A,$3F,$2F,$F6,$3F,$E6,$7C,$7F,$AB,$53,$CB,$E2,$6B,$E0,$B7,$E0
	.dc.b	$F4,$F9,$29,$7F,$D7,$70,$FC,$CB,$FD,$B5,$7B,$FF,$A4,$FB,$17,$CD
	.dc.b	$53,$C3,$1F,$A8,$3F,$09,$C7,$FC,$72,$5F,$15,$C4,$D7,$6E,$FE,$C7
	.dc.b	$A7,$44,$CF,$FB,$76,$29,$FD,$9F,$D8,$FF,$99,$27,$FB,$97,$D8,$FF
	.dc.b	$A2,$5B,$E3,$EB,$D2,$DF,$C0,$FE,$EA,$9F,$47,$4C,$DF,$F7,$BF,$B1
	.dc.b	$FF,$56,$BD,$3B,$38,$1F,$C0,$A3,$31,$FE,$0F,$7F,$CC,$8F,$8B,$E9
	.dc.b	$3B,$1B,$F4,$BF,$84,$FD,$87,$F9,$92,$7F,$C1,$1F,$62,$FA,$DF,$E1
	.dc.b	$AD,$8F,$F0,$64,$ED,$FF,$0E,$2F,$B1,$DF,$C5,$49,$DC,$7F,$16,$F0
	.dc.b	$7F,$D3,$D1,$FE,$15,$6B,$FE,$25,$6F,$91,$FE,$49,$FA,$FB,$D7,$FD
	.dc.b	$42,$6D,$7F,$A0,$67,$F4,$93,$FD,$51,$B1,$FE,$85,$9D,$BF,$93,$F9
	.dc.b	$0F,$E4,$F3,$A8,$36,$77,$F6,$2F,$98,$FE,$5A,$DA,$FF,$16,$35,$FF
	.dc.b	$1E,$3F,$C6,$0C,$63,$FE,$61,$6E,$3F,$DA,$1B,$3F,$E9,$97,$87,$D5
	.dc.b	$F6,$F3,$64,$F6,$8F,$09,$87,$BC,$61,$F9,$5F,$4B,$F3,$48,$F9,$BE
	.dc.b	$C6,$BF,$E5,$1F,$AA,$FF,$1D,$63,$FD,$E9,$B1,$F1,$EB,$79,$7F,$9E
	.dc.b	$67,$32,$6E,$8A,$FE,$A3,$1F,$85,$9E,$DC,$78,$33,$B5,$8F,$F2,$A4
	.dc.b	$EF,$0E,$D5,$A9,$CF,$C9,$74,$E2,$91,$82,$1E,$B7,$2F,$B6,$3F,$B1
	.dc.b	$FD,$7B,$EC,$69,$27,$FC,$C9,$F6,$0F,$B1,$C4,$7F,$9D,$63,$D8,$EB
	.dc.b	$5C,$E7,$FE,$9B,$C3,$A4,$E4,$87,$F9,$F5,$CB,$CF,$E2,$FF,$BA,$DF
	.dc.b	$F0,$C8,$F0,$F9,$18,$99,$18,$21,$C4,$F3,$FD,$4F,$22,$3E,$AF,$22
	.dc.b	$BE,$98,$B8,$FF,$AC,$70,$FC,$F7,$D8,$FF,$A4,$EB,$FD,$08,$D7,$FD
	.dc.b	$28,$D9,$FF,$5C,$78,$4F,$91,$FF,$06,$CE,$79,$BF,$C8,$33,$FD,$D8
	.dc.b	$FE,$EF,$19,$BF,$F7,$A3,$5F,$E3,$7F,$0B,$FE,$16,$9D,$F8,$97,$AE
	.dc.b	$57,$FA,$5E,$51,$BE,$BC,$CF,$FA,$A6,$72,$EE,$74,$F1,$F0,$FB,$11
	.dc.b	$FF,$51,$7D,$8F,$E6,$A9,$39,$9E,$8F,$F8,$AA,$7F,$3F,$F6,$2F,$23
	.dc.b	$FC,$5A,$FF,$39,$AF,$F5,$D3,$77,$FF,$6F,$3B,$8F,$F5,$83,$5D,$F9
	.dc.b	$5C,$8E,$75,$FE,$38,$FB,$1F,$E2,$44,$E4,$EF,$59,$7F,$C7,$FF,$17
	.dc.b	$FB,$1F,$92,$BE,$01,$D5,$7F,$AF,$EF,$0D,$E9,$EC,$77,$99,$19,$21
	.dc.b	$FE,$D5,$9E,$2D,$F6,$7F,$D6,$7D,$9F,$E8,$09,$FD,$2C,$FF,$AE,$E3
	.dc.b	$06,$90,$8F,$F2,$7F,$C6,$F0,$6C,$F0,$9F,$F6,$9B,$5E,$87,$28,$DB
	.dc.b	$63,$B3,$D5,$B1,$E4,$9E,$D5,$89,$CF,$AB,$2F,$B9,$CC,$A3,$EC,$3F
	.dc.b	$AE,$5F,$B9,$7D,$8B,$E5,$37,$18,$6F,$CD,$79,$23,$8E,$3F,$EA,$67
	.dc.b	$58,$7B,$0E,$D7,$CD,$C4,$70,$7E,$4B,$C3,$0F,$14,$ED,$B8,$78,$2B
	.dc.b	$C7,$39,$BF,$03,$63,$9B,$5F,$76,$1E,$78,$C6,$1D,$EB,$8D,$F5,$BF
	.dc.b	$2D,$68,$91,$9C,$3C,$73,$A8,$F1,$99,$AE,$71,$92,$4F,$CD,$C3,$C5
	.dc.b	$5C,$27,$35,$F2,$67,$31,$83,$D7,$3D,$5A,$7F,$BD,$B0,$F9,$EF,$FB
	.dc.b	$58,$FB,$9E,$4F,$EE,$5F,$19,$98,$FF,$2B,$71,$FD,$C9,$FF,$1B,$63
	.dc.b	$F9,$EE,$3F,$F2,$E7,$76,$6A,$61,$F6,$2F,$79,$FF,$B9,$2F,$73,$9A
	.dc.b	$DE,$1F,$8B,$BD,$CF,$E9,$AB,$73,$9A,$DD,$FE,$26,$F7,$7F,$C4,$FD
	.dc.b	$F6,$3A,$8F,$C3,$CE,$B2,$F5,$C7,$E3,$F2,$EB,$87,$3C,$78,$4F,$C4
	.dc.b	$7F,$43,$3A,$8E,$71,$AA,$E7,$97,$C3,$CB,$C3,$5E,$D2,$47,$A5,$F0
	.dc.b	$46,$A3,$04,$C3,$C5,$3A,$CB,$B6,$5E,$93,$DF,$97,$8E,$7C,$C7,$6A
	.dc.b	$E3,$7E,$3B,$E3,$97,$B2,$F8,$65,$E8,$BD,$93,$0F,$A1,$97,$96,$31
	.dc.b	$1D,$B1,$F3,$3A,$C7,$AB,$E1,$30,$53,$84,$F2,$C3,$C3,$97,$AF,$2F
	.dc.b	$26,$3D,$EF,$E6,$9F,$04,$E5,$5C,$27,$E9,$CE,$E7,$F0,$18,$7D,$6F
	.dc.b	$86,$C5,$FC,$53,$FC,$AD,$1F,$6B,$CE,$CF,$C7,$E9,$91,$93,$EF,$C6
	.dc.b	$2F,$F0,$CD,$D3,$F4,$0D,$D3,$F3,$8F,$E0,$3A,$BF,$03,$5F,$D9,$63
	.dc.b	$17,$FA,$1D,$2C,$FE,$76,$6A,$FB,$17,$C3,$E1,$78,$F6,$30,$DC,$3F
	.dc.b	$A5,$E3,$E0,$AE,$31,$83,$FA,$7F,$5F,$F3,$5E,$7E,$18,$AF,$87,$91
	.dc.b	$DC,$F0,$7C,$0D,$C3,$95,$7F,$58,$C5,$59,$8E,$53,$FD,$71,$F5,$56
	.dc.b	$6F,$C9,$EE,$FA,$FE,$C7,$D8,$DB,$EB,$7F,$06,$33,$1E,$0C,$DF,$29
	.dc.b	$FE,$E1,$8D,$E7,$FE,$DA,$7C,$B8,$BE,$BF,$E3,$C6,$1B,$8F,$98,$F8
	.dc.b	$0B,$C2,$FA,$75,$C7,$7B,$3C,$23,$EE,$36,$FE,$7B,$C5,$ED,$3C,$BE
	.dc.b	$A9,$F0,$B4,$EB,$F3,$9F,$D9,$6B,$CF,$3A,$6C,$ED,$F5,$CE,$F2,$F4
	.dc.b	$62,$BE,$11,$EB,$1F,$60,$FA,$9F,$5F,$D8,$BE,$4B,$F1,$B3,$9C,$BE
	.dc.b	$3C,$E5,$FA,$31,$9B,$E8,$DC,$7E,$0F,$BD,$CB,$19,$7C,$1F,$8F,$18
	.dc.b	$BF,$3E,$4F,$13,$E2,$1F,$17,$E5,$AE,$1F,$0A,$2D,$70,$9B,$8F,$D0
	.dc.b	$FB,$1F,$A7,$D7,$F5,$1B,$87,$DD,$6D,$FC,$6E,$24,$62,$FF,$67,$F5
	.dc.b	$CF,$9E,$FB,$FF,$62,$FB,$0F,$D8,$CE,$72,$FC,$5C,$E5,$FC,$AC,$67
	.dc.b	$2F,$A0,$DC,$3F,$80,$6C,$FB,$1E,$79,$4C,$38,$BF,$85,$8C,$61,$E7
	.dc.b	$F3,$FF,$1F,$9F,$E4,$1B,$7F,$99,$6C,$EE,$3F,$39,$3F,$8F,$F8,$09
	.dc.b	$1F,$8C,$F8,$1B,$5F,$4A,$6E,$7E,$69,$B5,$FB,$7D,$3F,$92,$DF,$E3
	.dc.b	$B9,$BF,$BA,$DF,$C8,$F2,$3F,$17,$E5,$FC,$66,$61,$FA,$DF,$B0,$B8
	.dc.b	$7C,$49,$B7,$FA,$89,$B5,$FC,$5B,$81,$F2,$DF,$A7,$F9,$1F,$D4,$36
	.dc.b	$7F,$4A,$DC,$7F,$16,$DC,$3F,$67,$E3,$FE,$D8,$FD,$BF,$AB,$6F,$F8
	.dc.b	$04,$5B,$FC,$24,$DA,$FE,$6B,$F6,$2C,$FF,$71,$F2,$DF,$CC,$A3,$18
	.dc.b	$FE,$DA,$31,$32,$3F,$73,$DD,$71,$FD,$95,$F0,$6D,$7F,$03,$9D,$61
	.dc.b	$FD,$13,$EC,$5F,$55,$FD,$01,$B8,$FF,$1A,$F0,$3F,$D7,$AF,$F2,$48
	.dc.b	$CE,$3F,$BF,$F3,$FE,$AC,$FF,$2F,$6C,$7F,$3B,$A4,$7F,$4C,$33,$0F
	.dc.b	$AF,$7C,$CF,$F5,$06,$DF,$F2,$AF,$B1,$FE,$7A,$FF,$34,$8C,$D7,$F0
	.dc.b	$F5,$87,$13,$F9,$4F,$D8,$EB,$4F,$EC,$33,$AB,$E5,$E7,$F7,$37,$E8
	.dc.b	$D8,$FE,$83,$EA,$B9,$FF,$B1,$B6,$3F,$9E,$FD,$8F,$EB,$5D,$CE,$BE
	.dc.b	$82,$47,$F4,$2F,$B1,$7A,$D8,$DF,$F9,$37,$D8,$3E,$6F,$FB,$7B,$6F
	.dc.b	$FA,$4F,$86,$F1,$33,$5A,$F6,$E5,$6B,$DD,$B8,$F4,$39,$9F,$D2,$BE
	.dc.b	$C7,$F3,$B3,$FD,$73,$2F,$F5,$FB,$FD,$78,$7C,$D7,$F8,$01,$A7,$F5
	.dc.b	$A9,$FB,$FC,$BF,$E0,$A8,$CD,$FF,$3F,$FB,$1F,$E5,$91,$FD,$A1,$B1
	.dc.b	$FD,$1B,$D1,$FB,$F1,$8A,$FF,$09,$FD,$8B,$D7,$FF,$07,$B7,$0F,$EE
	.dc.b	$2B,$E0,$73,$3A,$F9,$19,$5F,$DD,$BE,$C7,$FD,$D8,$F0,$8F,$F0,$CC
	.dc.b	$63,$3F,$EA,$8D,$C7,$FC,$47,$F6,$3F,$36,$F9,$7A,$5C,$3C,$B9,$1F
	.dc.b	$E0,$2F,$86,$7B,$38,$CB,$FA,$F4,$62,$75,$3F,$E2,$8F,$B1,$FF,$02
	.dc.b	$56,$A3,$1E,$6F,$F8,$B2,$F8,$4F,$F6,$7A,$DE,$7F,$E3,$69,$CB,$FE
	.dc.b	$38,$F8,$2F,$C2,$57,$F7,$0F,$AD,$FD,$9B,$FD,$BE,$33,$3F,$E3,$F9
	.dc.b	$CB,$C7,$AB,$FC,$77,$3A,$CF,$FC,$6B,$5F,$DB,$EF,$95,$68,$93,$FE
	.dc.b	$1A,$9D,$5F,$38,$FF,$12,$DF,$D5,$7F,$C3,$FF,$63,$FD,$45,$FF,$18
	.dc.b	$46,$70,$FF,$03,$CE,$51,$FE,$2A,$E3,$F2,$9F,$E2,$08,$7C,$A6,$2F
	.dc.b	$BF,$C4,$7A,$8F,$5D,$D8,$C9,$1F,$88,$64,$FF,$0A,$63,$FB,$28,$FF
	.dc.b	$1E,$B6,$3F,$C3,$07,$93,$C5,$F2,$FD,$8B,$E5,$BF,$C9,$55,$B3,$FC
	.dc.b	$D6,$5B,$CF,$87,$FE,$68,$C3,$90,$BE,$4F,$F9,$92,$34,$B0,$EF,$87
	.dc.b	$2B,$B3,$FE,$E3,$9A,$FF,$99,$6B,$FD,$77,$7D,$23,$FC,$BD,$3A,$C3
	.dc.b	$D9,$39,$9F,$1B,$B8,$CE,$1F,$EE,$F9,$F1,$1E,$77,$C4,$D3,$C4,$53
	.dc.b	$C5,$87,$91,$EB,$1C,$2B,$C1,$1D,$1F,$06,$1E,$89,$F0,$61,$E7,$8C
	.dc.b	$57,$6A,$DD,$F5,$8F,$29,$E7,$29,$C8,$C1,$83,$F1,$BF,$63,$FC,$73
	.dc.b	$E2,$B8,$36,$3E,$03,$5F,$F4,$30,$70,$FF,$A2,$57,$95,$FE,$95,$8E
	.dc.b	$2D,$78,$C6,$B3,$EF,$39,$9E,$4F,$1C,$3E,$C6,$6D,$3F,$D2,$D1,$86
	.dc.b	$C7,$0B,$FC,$0E,$1D,$70,$FA,$53,$C5,$FC,$2B,$5F,$AD,$1B,$BF,$C3
	.dc.b	$36,$7F,$D4,$86,$E3,$C1,$3D,$B0,$F3,$4F,$86,$3B,$E1,$E4,$7F,$73
	.dc.b	$1C,$4D,$57,$6A,$E3,$3A,$AE,$0F,$EA,$DB,$7E,$38,$D1,$FE,$B0,$BF
	.dc.b	$56,$1F,$ED,$86,$E5,$D9,$F4,$DF,$ED,$23,$D7,$E8,$DB,$F2,$B5,$F0
	.dc.b	$1F,$EB,$CF,$B1,$FE,$E6,$DD,$75,$18,$8F,$F4,$D3,$F8,$07,$47,$F2
	.dc.b	$3F,$98,$ED,$93,$91,$1C,$AF,$C1,$F6,$1F,$D2,$2F,$FD,$97,$F6,$1F
	.dc.b	$B1,$CF,$FD,$A8,$D8,$FE,$47,$3B,$7F,$D5,$4D,$C7,$FD,$F3,$3B,$93
	.dc.b	$67,$55,$FE,$DF,$7F,$C3,$13,$FE,$DA,$6B,$D3,$CE,$F8,$94,$FB,$1F
	.dc.b	$3C,$63,$9E,$57,$18,$CB,$EA,$8C,$BC,$CF,$F7,$3F,$D8,$FF,$77,$6C
	.dc.b	$7D,$18,$C4,$7F,$B1,$6F,$FD,$84,$E8,$FF,$77,$FA,$1F,$EE,$16,$FC
	.dc.b	$7F,$1B,$FB,$53,$99,$E5,$B7,$A8,$CD,$6B,$C9,$DF,$D8,$CA,$EB,$15
	.dc.b	$FF,$6B,$7D,$8F,$C8,$7F,$A9,$B5,$60,$AF,$C5,$4F,$FA,$1A,$7E,$DF
	.dc.b	$67,$CA,$33,$F0,$3F,$8F,$FB,$17,$C2,$FE,$47,$95,$E9,$53,$51,$F9
	.dc.b	$5F,$B1,$FF,$30,$E3,$FD,$93,$9D,$CF,$FD,$9A,$3F,$C9,$3E,$3F,$33
	.dc.b	$87,$BD,$C7,$C0,$ED,$F2,$52,$B9,$5F,$1C,$7F,$35,$1F,$EF,$55,$ED
	.dc.b	$FE,$97,$EC,$5F,$35,$4F,$24,$7E,$A4,$FC,$4E,$9E,$4B,$FE,$F3,$4F
	.dc.b	$87,$E4,$6F,$B7,$87,$D8,$F5,$EA,$97,$E4,$3F,$C0,$38,$A7,$F6,$DF
	.dc.b	$63,$FE,$63,$9D,$0F,$B1,$FF,$44,$B8,$6A,$CF,$C8,$A7,$93,$9D,$BE
	.dc.b	$54,$3D,$DF,$E0,$96,$FA,$3A,$78,$A3,$F8,$47,$D8,$FF,$AC,$F0,$F0
	.dc.b	$75,$F0,$33,$F1,$7A,$F7,$F9,$FD,$24,$AF,$4D,$FF,$9B,$1F,$5F,$D8
	.dc.b	$3E,$67,$F8,$A7,$D8,$BE,$B6,$9F,$C5,$E7,$55,$C3,$EC,$77,$F1,$12
	.dc.b	$30,$7C,$6F,$07,$FD,$42,$DF,$8D,$FE,$41,$18,$8F,$84,$7F,$8C,$3E
	.dc.b	$BF,$1C,$DF,$22,$6F,$37,$E3,$E7,$F5,$D3,$F3,$1B,$1F,$93,$9D,$D7
	.dc.b	$F8,$E9,$AA,$76,$F7,$E9,$F1,$E1,$FC,$CF,$EC,$5F,$2D,$49,$5F,$80
	.dc.b	$8F,$E3,$C6,$9C,$46,$9F,$58,$8D,$FD,$5F,$5F,$3A,$4F,$38,$F0,$18
	.dc.b	$78,$CF,$F3,$87,$47,$CB,$FA,$5F,$92,$47,$C9,$F6,$35,$F0,$FD,$57
	.dc.b	$F9,$0B,$6F,$DB,$36,$3E,$25,$6F,$7F,$CA,$CE,$64,$DD,$D5,$FD,$5E
	.dc.b	$3F,$09,$3F,$A2,$6B,$E0,$E3,$D7,$06,$D9,$1B,$72,$F1,$F9,$0E,$9C
	.dc.b	$52,$30,$43,$D6,$E5,$F6,$C7,$F6,$9F,$AF,$7D,$8D,$24,$FE,$DB,$EC
	.dc.b	$1F,$63,$88,$FA,$B1,$EC,$75,$A6,$76,$FD,$BE,$1D,$27,$24,$3E,$EB
	.dc.b	$97,$9F,$C5,$FF,$7B,$BF,$E2,$11,$E1,$F2,$31,$32,$30,$43,$89,$E7
	.dc.b	$FA,$9E,$44,$7D,$5D,$35,$F4,$C5,$C7,$F8,$6E,$1F,$9C,$97,$D8,$FF
	.dc.b	$A5,$9F,$E2,$ED,$7F,$2A,$D9,$FE,$2B,$E2,$BE,$47,$FC,$29,$39,$E6
	.dc.b	$FE,$7D,$3F,$DF,$4F,$EF,$D1,$9B,$FF,$53,$B5,$FE,$31,$F0,$9F,$E1
	.dc.b	$D9,$DF,$8D,$7A,$E5,$7F,$2A,$94,$6F,$B3,$33,$FA,$A9,$CB,$B9,$D3
	.dc.b	$C7,$C4,$EC,$47,$F9,$AF,$D8,$FE,$6E,$93,$99,$E8,$FF,$8C,$67,$F4
	.dc.b	$1F,$63,$FC,$C6,$7F,$9E,$BF,$CE,$B6,$FE,$72,$DD,$7F,$B0,$3B,$9F
	.dc.b	$DF,$B6,$77,$F0,$BF,$E3,$D8,$E1,$3F,$EC,$C7,$55,$FD,$A7,$EC,$7E
	.dc.b	$C1,$38,$F8,$57,$F9,$16,$3F,$8B,$FC,$57,$ED,$3E,$4B,$FC,$94,$D9
	.dc.b	$FE,$A7,$3C,$A7,$87,$3C,$9C,$C7,$33,$0F,$F1,$69,$FC,$BB,$63,$C1
	.dc.b	$ED,$FF,$AE,$3A,$BC,$C5,$9F,$F7,$2D,$7E,$A5,$FF,$63,$47,$AE,$75
	.dc.b	$09,$FF,$2C,$7C,$66,$DE,$0D,$AD,$C6,$9E,$F3,$FE,$DF,$6B,$D8,$F0
	.dc.b	$BC,$2B,$8B,$CB,$D6,$93,$C4,$D9,$47,$D8,$7F,$5A,$7A,$18,$FB,$17
	.dc.b	$C9,$6E,$51,$89,$DC,$71,$CB,$E1,$37,$0F,$F1,$23,$C2,$F9,$5F,$1A
	.dc.b	$DF,$9F,$F1,$AB,$58,$74,$AD,$61,$C9,$B0,$3D,$35,$C6,$33,$1E,$9C
	.dc.b	$3B,$9E,$99,$D3,$88,$EC,$E5,$C6,$1E,$AA,$E7,$87,$92,$3C,$F3,$AC
	.dc.b	$3C,$97,$63,$C9,$5F,$05,$B7,$C8,$93,$DE,$3C,$24,$BF,$04,$65,$B1
	.dc.b	$9B,$EE,$FD,$0C,$3D,$6E,$4E,$F5,$C6,$77,$1F,$EE,$8C,$3E,$6B,$7C
	.dc.b	$8F,$F1,$AB,$A3,$EB,$BF,$E3,$AC,$3E,$DB,$6B,$FC,$BD,$E9,$FF,$9A
	.dc.b	$38,$BF,$09,$39,$BF,$82,$D7,$F1,$AD,$9E,$11,$F8,$88,$D6,$1F,$E7
	.dc.b	$0E,$4E,$67,$E5,$6B,$78,$F8,$A3,$37,$C5,$D7,$BF,$FE,$5F,$9D,$4E
	.dc.b	$48,$FE,$D6,$B9,$E5,$EE,$BE,$73,$C0,$D5,$F3,$9D,$5F,$26,$CE,$B2
	.dc.b	$E3,$18,$24,$FE,$57,$0E,$EF,$C4,$C3,$6D,$3F,$5F,$1A,$7E,$3C,$69
	.dc.b	$A6,$B0,$F2,$3C,$70,$F4,$5F,$82,$3F,$CD,$13,$A7,$8C,$E9,$E1,$3F
	.dc.b	$6A,$75,$8F,$D1,$9F,$71,$A3,$D5,$1A,$CB,$EB,$93,$0F,$A3,$97,$9E
	.dc.b	$F9,$63,$DF,$0F,$1C,$E6,$F8,$47,$AE,$3D,$D9,$7C,$1B,$E3,$97,$96
	.dc.b	$73,$97,$AE,$BE,$30,$F8,$6C,$46,$7C,$E9,$5F,$89,$F8,$DB,$97,$CE
	.dc.b	$9B,$A7,$E0,$E6,$E5,$F9,$C6,$D7,$D9,$6B,$FE,$93,$7F,$D5,$9D,$3F
	.dc.b	$E9,$AC,$D1,$8C,$7F,$4B,$B5,$BE,$51,$FE,$A6,$FB,$07,$C4,$66,$50
	.dc.b	$F1,$B0,$EA,$B8,$46,$2A,$47,$FA,$D7,$AB,$F9,$AF,$3E,$30,$FF,$37
	.dc.b	$E0,$F1,$F9,$9D,$3E,$06,$E1,$E3,$8B,$5C,$26,$ED,$FE,$B6,$BF,$F6
	.dc.b	$8D,$FA,$7E,$C7,$CB,$EF,$79,$9B,$1E,$D8,$F8,$19,$BE,$4D,$F0,$BA
	.dc.b	$CF,$FD,$3B,$93,$EB,$FE,$06,$3F,$EC,$F8,$C7,$07,$C9,$75,$E1,$5A
	.dc.b	$DD,$F2,$6C,$FD,$46,$DF,$FB,$D5,$57,$1B,$E9,$F5,$0F,$86,$A7,$7F
	.dc.b	$9B,$FE,$BF,$87,$8E,$74,$D9,$DB,$E7,$9D,$E5,$EC,$AD,$BF,$EA,$0F
	.dc.b	$B0,$7D,$47,$AF,$EA,$9F,$63,$F8,$19,$CE,$3F,$06,$72,$FF,$BB,$23
	.dc.b	$37,$CE,$FE,$17,$BF,$C5,$95,$B7,$2D,$3D,$5E,$2F,$C0,$3E,$0F,$CE
	.dc.b	$D8,$C9,$2F,$D5,$19,$7E,$77,$D8,$FD,$1E,$CF,$A4,$DA,$FB,$1C,$48
	.dc.b	$C5,$70,$9C,$D7,$E7,$FE,$B5,$F6,$1F,$77,$EC,$5F,$61,$98,$C3,$6F
	.dc.b	$FB,$F7,$39,$FF,$00,$62,$C3,$9C,$7D,$7F,$B1,$E1,$3F,$ED,$09,$CB
	.dc.b	$70,$FC,$0C,$EA,$FF,$11,$E8,$7E,$37,$1F,$C7,$37,$1F,$6B,$70,$FC
	.dc.b	$83,$6B,$F8,$1F,$2B,$E0,$3F,$1A,$7E,$3B,$E0,$6E,$1F,$8B,$8B,$87
	.dc.b	$E1,$E6,$DF,$B6,$3F,$C3,$9A,$7F,$BC,$2F,$F0,$9E,$5F,$C9,$F2,$FF
	.dc.b	$14,$D8,$F8,$A7,$C4,$9E,$15,$FA,$5F,$8C,$D3,$F0,$EE,$E1,$FA,$79
	.dc.b	$B7,$F3,$B3,$7C,$AF,$E9,$BE,$47,$F4,$ED,$9F,$D2,$37,$1F,$D0,$37
	.dc.b	$0F,$D9,$79,$1F,$B7,$3F,$71,$EB,$5B,$FC,$3C,$5B,$FD,$D4,$DA,$F8
	.dc.b	$50,$FB,$16,$7F,$B8,$F1,$3F,$94,$46,$31,$FD,$B4,$62,$64,$7E,$E7
	.dc.b	$B2,$DB,$0E,$B8,$71,$8C,$57,$E3,$23,$13,$F7,$2B,$8B,$E4,$FB,$07
	.dc.b	$D8,$FF,$39,$6D,$FF,$19,$EF,$FF,$5C,$BF,$53,$1F,$F2,$B4,$E9,$EE
	.dc.b	$7F,$96,$7C,$EF,$F1,$1E,$1F,$01,$B5,$FA,$0B,$47,$F4,$73,$35,$FE
	.dc.b	$61,$FA,$F7,$CC,$7F,$4B,$6D,$FF,$2B,$FB,$1F,$E8,$4F,$8E,$93,$A7
	.dc.b	$C6,$63,$0E,$47,$F3,$1F,$B1,$D6,$5F,$52,$93,$DF,$CE,$98,$7F,$39
	.dc.b	$B7,$A3,$8C,$7A,$FA,$CE,$7F,$EC,$0D,$BF,$E8,$1F,$62,$F0,$73,$5A
	.dc.b	$F9,$F9,$5F,$D1,$7E,$C5,$EB,$63,$2F,$4F,$D8,$BE,$AF,$FB,$63,$6F
	.dc.b	$FA,$67,$87,$FB,$57,$E1,$46,$6B,$8F,$B5,$C0,$FE,$A1,$4F,$7B,$17
	.dc.b	$C2,$BA,$9F,$D5,$7E,$C7,$F4,$F1,$93,$E7,$0F,$9B,$FE,$FC,$D3,$FA
	.dc.b	$F4,$7E,$0F,$9F,$F3,$F1,$9A,$FA,$7F,$62,$F4,$BF,$C0,$CD,$9F,$EC
	.dc.b	$DE,$7F,$E0,$A9,$39,$7E,$DF,$D8,$BD,$7F,$F0,$73,$6F,$F0,$4F,$F8
	.dc.b	$A5,$B1,$AF,$90,$91,$F8,$7F,$B1,$7A,$18,$C3,$F0,$F1,$89,$D4,$FF
	.dc.b	$86,$FE,$C7,$E7,$56,$A3,$FC,$3D,$5F,$92,$8E,$3E,$5F,$F8,$87,$E1
	.dc.b	$9E,$C6,$A3,$0C,$C7,$F3,$1F,$63,$FE,$0A,$7F,$35,$5C,$1C,$F9,$BB
	.dc.b	$C3,$F3,$75,$B7,$F3,$B5,$B7,$F3,$FF,$01,$F8,$1E,$3F,$F1,$DF,$D6
	.dc.b	$FE,$E2,$73,$D7,$FE,$11,$8C,$4F,$F1,$4A,$E1,$97,$F5,$26,$DF,$F0
	.dc.b	$DA,$FE,$F2,$F2,$9F,$D8,$4E,$A3,$FC,$49,$7D,$7D,$EF,$F2,$14,$E4
	.dc.b	$FA,$93,$FE,$26,$FB,$1F,$EB,$38,$F8,$7E,$23,$FC,$49,$18,$C7,$C5
	.dc.b	$1C,$8C,$FC,$9D,$8F,$0F,$C4,$76,$3B,$E1,$63,$24,$7E,$29,$93,$F8
	.dc.b	$2E,$1F,$B0,$8F,$E1,$4D,$8F,$E1,$27,$FB,$CF,$C7,$FF,$7A,$7D,$8B
	.dc.b	$E5,$7F,$BB,$D6,$CF,$E7,$A5,$BC,$F8,$9F,$E6,$3C,$39,$0B,$E4,$FF
	.dc.b	$80,$A3,$4B,$0E,$F8,$72,$BB,$3F,$E0,$63,$47,$38,$D3,$4D,$4F,$07
	.dc.b	$FA,$35,$7F,$73,$BF,$0C,$7F,$84,$27,$58,$7C,$19,$CC,$FA,$1D,$C6
	.dc.b	$6F,$C5,$3E,$29,$CD,$F5,$9C,$DF,$48,$C5,$79,$0F,$31,$E4,$C3,$CC
	.dc.b	$F7,$AF,$5D,$F8,$A3,$B4,$60,$EA,$F8,$8F,$9A,$5B,$C3,$83,$FC,$5B
	.dc.b	$F6,$3F,$C6,$7E,$27,$83,$63,$D6,$D7,$E7,$35,$E3,$E6,$7F,$6F,$76
	.dc.b	$D7,$6D,$77,$9F,$68,$E1,$8F,$D0,$C7,$FD,$5B,$5F,$55,$AF,$C4,$BF
	.dc.b	$E9,$F7,$F6,$5B,$1F,$22,$77,$5C,$F4,$FF,$24,$BF,$72,$B6,$DC,$3C
	.dc.b	$35,$FE,$4E,$38,$96,$B8,$BE,$8B,$F1,$4F,$04,$7E,$86,$76,$E2,$79
	.dc.b	$BA,$BF,$0C,$73,$BF,$34,$FB,$EB,$8A,$AD,$BC,$E3,$F5,$B9,$F9,$63
	.dc.b	$FB,$F5,$FC,$6C,$7F,$C5,$77,$C2,$7C,$86,$BD,$3E,$2F,$95,$AE,$8F
	.dc.b	$F0,$2F,$D8,$FF,$70,$7F,$C1,$4D,$73,$B7,$08,$FF,$00,$C6,$A3,$AD
	.dc.b	$B9,$5F,$2F,$86,$D6,$CF,$F0,$B7,$D8,$FF,$44,$7F,$C3,$1F,$5F,$FA
	.dc.b	$CD,$7F,$D0,$2D,$8F,$9F,$3B,$7F,$C3,$6D,$C3,$FC,$97,$3B,$93,$67
	.dc.b	$55,$FE,$8C,$7F,$C2,$D3,$E2,$F3,$FE,$1C,$62,$3F,$C4,$F4,$FB,$1C
	.dc.b	$D7,$C8,$87,$35,$33,$5C,$CF,$5B,$97,$07,$F8,$CB,$EC,$78,$BF,$E3
	.dc.b	$76,$B9,$8E,$DD,$5F,$5A,$73,$9F,$F9,$A3,$D1,$FF,$6F,$DF,$F8,$FD
	.dc.b	$AF,$13,$1F,$1F,$E7,$DE,$46,$0E,$1E,$57,$E0,$A9,$F6,$34,$C4,$59
	.dc.b	$FF,$26,$7D,$8F,$CA,$7F,$42,$D5,$5C,$9D,$FC,$1F,$C7,$EE,$91,$83
	.dc.b	$FC,$AD,$F6,$2F,$88,$FC,$9F,$3F,$F8,$CE,$7F,$CB,$4D,$7F,$CB,$33
	.dc.b	$B7,$FC,$BD,$F6,$3F,$E6,$47,$7E,$3F,$23,$F3,$B1,$98,$FF,$30,$BF
	.dc.b	$98,$7F,$CC,$74,$F9,$2F,$09,$C3,$C1,$E3,$EE,$FE,$9B,$EC,$5F,$35
	.dc.b	$4C,$E1,$FE,$6F,$9F,$C4,$72,$FF,$A7,$29,$F1,$3C,$4F,$D7,$CE,$BB
	.dc.b	$BF,$DB,$AB,$EC,$7B,$74,$4C,$7A,$47,$ED,$BE,$C7,$FC,$C1,$1F,$E7
	.dc.b	$DF,$B1,$FF,$43,$B7,$C4,$D7,$9B,$BF,$7B,$F7,$B4,$FA,$2A,$66,$FF
	.dc.b	$D0,$DF,$63,$FE,$B0,$7A,$74,$F0,$3F,$82,$C6,$63,$FD,$12,$FF,$97
	.dc.b	$DF,$17,$D1,$F6,$37,$EF,$FF,$0C,$FB,$0F,$F2,$E4,$FF,$A4,$FE,$C5
	.dc.b	$F5,$9F,$C4,$1B,$1F,$E9,$79,$DB,$FE,$96,$5F,$63,$BF,$85,$93,$B8
	.dc.b	$FE,$33,$F0,$BF,$E9,$68,$FF,$50,$35,$FF,$4F,$B7,$A7,$FD,$9F,$F5
	.dc.b	$D7,$A3,$F9,$CD,$7F,$A0,$7E,$33,$F9,$1C,$FE,$AA,$B7,$D1,$DA,$65
	.dc.b	$7F,$AB,$6B,$57,$C9,$B1,$FD,$13,$0E,$CF,$F4,$6F,$B1,$7C,$BF,$F3
	.dc.b	$06,$C7,$FA,$DD,$AF,$FA,$D9,$FE,$34,$63,$2F,$F7,$3B,$70,$F4,$B6
	.dc.b	$7B,$7A,$DD,$7E,$AB,$B7,$9B,$27,$B4,$65,$F0,$46,$5F,$04,$7C,$CF
	.dc.b	$A5,$91,$F3,$1F,$47,$D8,$E2,$3F,$D9,$7F,$63,$FE,$3E,$CB,$E5,$B6
	.dc.b	$3E,$15,$6F,$3F,$F7,$FC,$E6,$4D,$D5,$5F,$D5,$23,$F0,$B3,$DB,$97
	.dc.b	$06,$77,$B1,$FE,$DF,$9D,$E1,$DA,$B5,$39,$F9,$DE,$9C,$92,$30,$43
	.dc.b	$D6,$E5,$F6,$C7,$F6,$6F,$AF,$7D,$8D,$24,$FF,$BB,$FE,$C1,$F6,$38
	.dc.b	$8F,$F7,$7C,$7B,$1D,$6B,$9C,$DB,$FE,$4B,$1E,$B5,$A2,$1F,$EF,$87
	.dc.b	$2F,$4F,$89,$FE,$F3,$87,$F0,$E8,$F1,$79,$18,$99,$18,$21,$C4,$F4
	.dc.b	$7D,$4F,$22,$3E,$BE,$AA,$FA,$A2,$E5,$E8,$8F,$F1,$0E,$1F,$A0,$5F
	.dc.b	$62,$F0,$BF,$C1,$AD,$8F,$F1,$3D,$71,$77,$3F,$E1,$2F,$0D,$F1,$FC
	.dc.b	$23,$1D,$72,$B8,$EF,$FB,$8C,$7A,$43,$1E,$91,$87,$B6,$3F,$E0,$33
	.dc.b	$FC,$09,$19,$AF,$E0,$30,$EA,$F1,$12,$FB,$9F,$E3,$E8,$FF,$0C,$BD
	.dc.b	$E7,$8C,$EE,$30,$6D,$F6,$75,$FB,$7C,$39,$3D,$0F,$E6,$8F,$F8,$CB
	.dc.b	$EC,$7F,$49,$82,$37,$3D,$0D,$9C,$67,$F5,$5F,$62,$F2,$3F,$C7,$6F
	.dc.b	$F4,$49,$E0,$DD,$3E,$83,$B8,$F1,$35,$DF,$95,$C8,$E7,$5F,$E4,$6F
	.dc.b	$B1,$FB,$C4,$E4,$EF,$59,$7F,$C9,$BF,$13,$FB,$CF,$91,$BE,$01,$D5
	.dc.b	$7C,$1B,$C3,$7A,$7B,$1D,$E6,$46,$48,$7C,$39,$E2,$DF,$57,$FD,$71
	.dc.b	$D9,$FC,$F2,$7F,$61,$3F,$EC,$98,$C1,$A7,$FC,$F3,$1F,$E5,$DF,$8B
	.dc.b	$DB,$67,$73,$FE,$DF,$6B,$CC,$E3,$5C,$D4,$F8,$2B,$9D,$66,$30,$F9
	.dc.b	$6B,$0E,$7D,$49,$7D,$CF,$2C,$75,$FB,$0F,$EC,$B8,$7B,$17,$D8,$BE
	.dc.b	$43,$85,$A7,$31,$87,$8D,$FF,$60,$6C,$FF,$A6,$9D,$BE,$D3,$95,$75
	.dc.b	$71,$1C,$1C,$D7,$59,$D5,$F9,$5B,$7D,$FE,$63,$D5,$18,$BF,$03,$5E
	.dc.b	$67,$68,$DD,$F5,$24,$64,$B7,$DE,$2D,$F7,$9A,$76,$8A,$FC,$5A,$EA
	.dc.b	$FB,$2F,$C5,$87,$C2,$C3,$E7,$56,$9E,$D5,$BB,$F9,$93,$9C,$3E,$66
	.dc.b	$1C,$8E,$71,$9A,$F1,$1C,$4F,$51,$6F,$A9,$2B,$D5,$1B,$3F,$C2,$E7
	.dc.b	$C6,$2E,$0F,$27,$F8,$04,$78,$B1,$DB,$F7,$63,$FC,$8F,$7F,$89,$6D
	.dc.b	$FF,$88,$B8,$BF,$CE,$5D,$7F,$87,$9C,$DF,$1E,$4C,$1F,$8E,$AD,$C7
	.dc.b	$E1,$27,$6F,$E4,$5B,$97,$96,$B8,$7B,$DF,$D2,$72,$F1,$47,$4B,$E4
	.dc.b	$EB,$1C,$4E,$A7,$27,$E5,$71,$F0,$C6,$49,$97,$E7,$72,$F1,$CE,$A3
	.dc.b	$18,$74,$6D,$F3,$C7,$C9,$18,$C3,$9C,$FC,$13,$58,$F1,$71,$1A,$7F
	.dc.b	$99,$57,$27,$35,$CC,$88,$DD,$73,$7E,$45,$DC,$3D,$11,$C8,$86,$8F
	.dc.b	$19,$F1,$30,$F1,$65,$F3,$E3,$31,$8C,$BE,$99,$F1,$32,$F3,$4F,$39
	.dc.b	$D6,$1F,$46,$3A,$65,$F8,$5A,$DE,$1D,$6F,$F9,$6E,$3E,$6B,$E5,$7B
	.dc.b	$CB,$95,$6B,$E1,$BB,$DE,$CF,$F3,$1C,$FE,$2F,$CE,$CF,$C7,$69,$91
	.dc.b	$93,$F2,$11,$8C,$7F,$04,$DC,$3C,$6D,$C7,$F3,$8F,$F8,$5F,$AB,$FC
	.dc.b	$21,$8A,$31,$87,$E8,$75,$B8,$FD,$5A,$D4,$FF,$77,$FB,$17,$C2,$EA
	.dc.b	$90,$F1,$F0,$EA,$B8,$46,$2A,$47,$F8,$33,$AF,$F9,$C7,$3E,$30,$FE
	.dc.b	$FD,$83,$C8,$E6,$74,$F7,$AE,$1E,$38,$B5,$C2,$6E,$DF,$E2,$DB,$7B
	.dc.b	$BE,$4F,$B1,$F2,$B7,$1F,$63,$87,$C9,$F2,$6F,$83,$C6,$5F,$E1,$CC
	.dc.b	$5F,$5F,$EF,$BF,$F1,$2C,$63,$83,$E4,$BA,$F0,$ED,$6E,$F9,$36,$7E
	.dc.b	$A3,$6F,$FC,$8A,$AB,$8D,$F4,$FA,$AF,$F7,$B5,$78,$1B,$87,$97,$E6
	.dc.b	$BF,$B1,$4E,$7E,$17,$81,$1F,$3C,$6A,$7D,$95,$CB,$2F,$F2,$A7,$D8
	.dc.b	$3E,$A7,$D7,$F6,$2F,$90,$FC,$04,$E7,$2F,$83,$39,$7F,$C7,$91,$9C
	.dc.b	$7F,$C7,$0D,$C3,$DF,$EF,$F2,$C6,$5F,$07,$E3,$C6,$2F,$CB,$E2,$FC
	.dc.b	$13,$E1,$7C,$DD,$BF,$85,$16,$7D,$73,$70,$F9,$BF,$63,$F4,$BB,$3E
	.dc.b	$9B,$70,$FB,$3C,$68,$C6,$1F,$59,$B7,$FE,$77,$FA,$E7,$D8,$7D,$EF
	.dc.b	$B1,$7D,$8F,$E8,$E7,$39,$7E,$12,$72,$FF,$97,$E3,$38,$FF,$96,$DB
	.dc.b	$87,$ED,$78,$DF,$51,$CF,$29,$87,$17,$F0,$51,$8C,$3F,$17,$E8,$7E
	.dc.b	$3B,$3F,$C7,$B6,$FD,$AD,$9D,$C7,$C0,$9F,$C6,$7C,$0C,$8F,$C6,$7C
	.dc.b	$15,$AF,$C2,$CD,$CF,$CD,$36,$BF,$9D,$53,$4F,$E3,$6F,$E9,$F9,$BF
	.dc.b	$B2,$DF,$C9,$F2,$BF,$17,$E6,$FC,$76,$61,$F4,$BE,$C6,$E1,$F1,$66
	.dc.b	$DF,$E9,$A6,$D7,$FA,$AB,$81,$F2,$9F,$A5,$F9,$AF,$D3,$36,$7F,$4C
	.dc.b	$DC,$7F,$3E,$DC,$3F,$63,$E3,$FE,$D8,$FD,$BF,$AD,$6F,$F0,$B1,$6F
	.dc.b	$F7,$33,$6B,$FD,$75,$F6,$2C,$FF,$6D,$E2,$7F,$25,$8C,$63,$FB,$78
	.dc.b	$C4,$C8,$FD,$D7,$75,$C7,$FD,$3D,$7C,$1B,$5F,$C0,$67,$58,$7F,$B2
	.dc.b	$7E,$C1,$F6,$3F,$CC,$DB,$8F,$F1,$4F,$03,$FD,$78,$FF,$AB,$E3,$38
	.dc.b	$FF,$AA,$FC,$FF,$AB,$3F,$69,$B1,$FE,$CB,$A4,$7F,$3F,$33,$0F,$AF
	.dc.b	$7D,$8F,$F4,$36,$DF,$F2,$4F,$B1,$FE,$74,$FF,$AF,$E3,$35,$FE,$BB
	.dc.b	$58,$71,$3F,$94,$FD,$8E,$B4,$FF,$73,$4E,$AF,$97,$9B,$DC,$DF,$A1
	.dc.b	$63,$FD,$99,$EB,$39,$FF,$AB,$36,$3F,$9B,$FD,$8F,$EB,$9D,$CE,$BE
	.dc.b	$7A,$47,$F3,$CF,$B1,$7A,$B8,$DF,$FD,$A5,$F6,$2F,$9B,$FE,$C4,$DB
	.dc.b	$FE,$87,$E1,$BC,$4C,$D6,$BD,$99,$5A,$F7,$AE,$3D,$0E,$67,$F4,$AF
	.dc.b	$B1,$FC,$EC,$FF,$B9,$F2,$FF,$14,$3F,$EE,$91,$F3,$5F,$DC,$DA,$7F
	.dc.b	$54,$9F,$BF,$CA,$8F,$DB,$9A,$BF,$F7,$67,$D8,$BD,$3F,$EF,$8D,$9F
	.dc.b	$EB,$BE,$97,$E0,$23,$15,$FE,$F3,$FB,$17,$AF,$FE,$03,$6D,$FF,$BF
	.dc.b	$57,$C0,$E6,$75,$F2,$32,$BF,$C8,$59,$F9,$70,$FF,$22,$7D,$87,$FB
	.dc.b	$9E,$79,$57,$F8,$3E,$31,$2A,$E0,$A3,$31,$8F,$B1,$FA,$57,$D8,$FE
	.dc.b	$2F,$3A,$6F,$97,$D4,$FF,$0E,$7C,$33,$DA,$C4,$F1,$C7,$FC,$B1,$39
	.dc.b	$8C,$47,$FB,$46,$7F,$C4,$9F,$63,$FE,$09,$9C,$D6,$EB,$D1,$1F,$CA
	.dc.b	$27,$51,$83,$AD,$72,$F3,$7C,$6F,$F8,$BF,$E0,$7F,$1D,$3F,$E1,$3F
	.dc.b	$AD,$FD,$EB,$C4,$46,$B5,$FF,$20,$4E,$9F,$8F,$19,$C7,$F9,$CB,$6B
	.dc.b	$E6,$57,$F7,$B7,$A4,$65,$EB,$D5,$E3,$3F,$C3,$F3,$BB,$EB,$1E,$7E
	.dc.b	$1F,$F1,$1F,$D8,$FF,$5E,$C7,$BC,$63,$DB,$D4,$7C,$AC,$BF,$0E,$FF
	.dc.b	$8B,$29,$F2,$DA,$CF,$C7,$F1,$1F,$0C,$F8,$8D,$AE,$51,$63,$D2,$5C
	.dc.b	$FF,$0E,$7E,$F2,$7E,$CB,$63,$E5,$1F,$E9,$5F,$1B,$FD,$2D,$F6,$2F
	.dc.b	$97,$FE,$BF,$5B,$39,$F8,$9F,$E6,$37,$18,$7F,$B4,$2F,$8D,$70,$AF
	.dc.b	$F3,$46,$1D,$46,$1D,$5F,$D1,$C7,$E8,$E3,$11,$FD,$BA,$B9,$BA,$BA
	.dc.b	$E4,$E7,$7D,$E3,$31,$DE,$B9,$E1,$EF,$7A,$C6,$6B,$83,$C8,$FF,$77
	.dc.b	$DF,$6A,$E1,$3E,$D9,$D1,$FB,$3C,$3D,$11,$9B,$EF,$18,$9E,$67,$0A
	.dc.b	$E6,$E2,$77,$87,$A2,$18,$7F,$BE,$9F,$71,$5C,$3F,$2C,$FF,$09,$8F
	.dc.b	$B1,$FE,$4B,$F1,$7C,$5B,$1A,$35,$E0,$A7,$4D,$8D,$4B,$CB,$FE,$21
	.dc.b	$18,$6B,$E0,$8D,$65,$DE,$73,$3F,$E2,$69,$CB,$F2,$34,$FF,$59,$3F
	.dc.b	$69,$AF,$CC,$8D,$E7,$F8,$26,$CF,$F8,$C0,$DC,$7E,$19,$A7,$7A,$F3
	.dc.b	$C7,$16,$C7,$1D,$7F,$C7,$33,$C5,$B1,$C7,$0F,$C6,$DF,$83,$0F,$F7
	.dc.b	$C4,$F6,$6D,$70,$7F,$DF,$93,$B7,$E9,$61,$E0,$76,$7B,$E7,$83,$AB
	.dc.b	$C3,$FC,$B2,$FE,$1C,$7F,$3E,$6E,$9E,$59,$F9,$55,$FB,$48,$F8,$9E
	.dc.b	$95,$9F,$3B,$5F,$DD,$4D,$FB,$1F,$EF,$B1,$FC,$E9,$AF,$F2,$B6,$E1
	.dc.b	$FD,$39,$FC,$14,$6F,$7E,$D2,$E6,$7C,$BE,$1B,$5B,$3F,$A3,$FD,$8F
	.dc.b	$F4,$D7,$FA,$4F,$D8,$7E,$D3,$4F,$F3,$5B,$60,$27,$83,$FC,$FD,$B8
	.dc.b	$7F,$62,$9E,$12,$6C,$EA,$BF,$CE,$EF,$F8,$96,$7C,$7E,$7F,$C5,$8C
	.dc.b	$47,$F4,$BA,$7D,$8E,$6B,$E5,$47,$F7,$C9,$F4,$F5,$65,$E6,$60,$87
	.dc.b	$9E,$7E,$97,$D8,$F1,$8F,$EB,$0D,$CB,$FB,$DB,$72,$F8,$7F,$19,$C3
	.dc.b	$0F,$B5,$E9,$7F,$B9,$70,$FE,$E2,$D7,$9C,$6B,$E3,$FC,$FB,$C9,$DC
	.dc.b	$72,$F2,$F9,$FD,$8C,$9E,$71,$5F,$ED,$9F,$63,$F2,$5F,$A6,$D5,$8A
	.dc.b	$BF,$1D,$3F,$E9,$49,$FB,$BD,$DF,$20,$CF,$C1,$FE,$4F,$EC,$5F,$0F
	.dc.b	$F9,$5E,$77,$A9,$4D,$47,$E6,$7E,$C7,$FC,$DF,$8F,$F8,$42,$9E,$4F
	.dc.b	$13,$5F,$01,$BF,$03,$A7,$C9,$4B,$FF,$0B,$E1,$F9,$A7,$FC,$10,$BD
	.dc.b	$DF,$D3,$7D,$8B,$E6,$A9,$E1,$8F,$D5,$1F,$86,$EB,$FF,$18,$CB,$E2
	.dc.b	$78,$9A,$ED,$DF,$D8,$F4,$E9,$99,$FF,$8A,$31,$4F,$ED,$7E,$C7,$FC
	.dc.b	$D9,$3F,$E2,$CF,$B1,$FF,$48,$B7,$C7,$D7,$A3,$BF,$7F,$F7,$94,$FA
	.dc.b	$3A,$66,$FF,$C6,$BF,$63,$FE,$B5,$7A,$76,$70,$3F,$82,$46,$63,$FC
	.dc.b	$54,$FF,$9B,$1F,$17,$D2,$76,$37,$E8,$7F,$0B,$FB,$0F,$F3,$44,$FF
	.dc.b	$90,$BE,$C5,$F5,$BF,$C3,$DB,$1F,$E4,$69,$DB,$FE,$37,$5F,$63,$BF
	.dc.b	$88,$93,$B8,$FE,$31,$E0,$FF,$A9,$A3,$FC,$9E,$D7,$FC,$80,$DE,$9F
	.dc.b	$F3,$1F,$D7,$DF,$33,$FD,$EB,$36,$BF,$85,$CF,$E9,$EB,$7D,$1D,$BE
	.dc.b	$3A,$57,$F9,$72,$B5,$87,$FB,$F9,$B1,$FC,$53,$0E,$8B,$EC,$5F,$2F
	.dc.b	$FC,$BD,$B1,$FE,$66,$6B,$FE,$57,$7F,$8B,$98,$CB,$FC,$F6,$DC,$3F
	.dc.b	$90,$CF,$5A,$E9,$EB,$F5,$FA,$BE,$DE,$7C,$AE,$D1,$97,$C1,$19,$7C
	.dc.b	$11,$F3,$3E,$96,$47,$CC,$7D,$1F,$63,$88,$FF,$36,$7D,$57,$F9,$23
	.dc.b	$2F,$1B,$63,$E1,$56,$F4,$FF,$49,$CE,$64,$DD,$55,$FD,$4E,$3F,$0B
	.dc.b	$3D,$B9,$70,$67,$7B,$1F,$E8,$E9,$DE,$1D,$AB,$53,$9F,$90,$E9,$C9
	.dc.b	$23,$04,$3D,$6E,$5F,$6C,$7F,$65,$FA,$F7,$D8,$D2,$4F,$FA,$7B,$EC
	.dc.b	$1F,$63,$88,$FF,$48,$47,$B1,$D6,$B9,$CF,$FD,$6B,$87,$49,$C9,$0F
	.dc.b	$F4,$C3,$97,$9F,$C5,$FF,$77,$BF,$E1,$91,$E1,$F2,$71,$32,$30,$43
	.dc.b	$89,$E7,$FA,$9E,$44,$7D,$5D,$55,$F4,$C5,$C7,$FD,$97,$87,$E7,$A5
	.dc.b	$F6,$3F,$E9,$27,$FD,$7E,$D7,$FD,$60,$D9,$FF,$69,$F8,$6F,$90,$FF
	.dc.b	$08,$4E,$79,$BF,$37,$3F,$DD,$8F,$EE,$F1,$9B,$FE,$FE,$D7,$F8,$E7
	.dc.b	$57,$F8,$66,$77,$E2,$5E,$A9,$5F,$ED,$29,$46,$FA,$F3,$3F,$EC,$59
	.dc.b	$CB,$B9,$D3,$C7,$C3,$EC,$47,$FD,$BB,$F6,$3F,$9B,$A4,$E6,$7A,$3F
	.dc.b	$E2,$D9,$FD,$07,$D8,$BC,$8F,$F1,$83,$FC,$EA,$BF,$DA,$CD,$DF,$FD
	.dc.b	$F2,$EE,$3F,$DD,$6D,$77,$E5,$72,$39,$D7,$F8,$EF,$EC,$7E,$B1,$39
	.dc.b	$3B,$D6,$5F,$F2,$17,$C5,$7E,$CB,$EA,$2F,$80,$75,$5F,$EF,$AB,$C3
	.dc.b	$7A,$7B,$1D,$E6,$46,$48,$7F,$BC,$67,$8B,$7D,$9F,$F5,$C7,$67,$F1
	.dc.b	$A9,$FD,$34,$FF,$A7,$23,$06,$9F,$F1,$FC,$7F,$94,$BE,$33,$6D,$9D
	.dc.b	$CF,$FB,$55,$AF,$33,$8D,$73,$53,$E0,$AE,$75,$98,$C3,$E5,$AC,$39
	.dc.b	$F5,$65,$F7,$3C,$B1,$D7,$EC,$3F,$AD,$E1,$EC,$5F,$62,$F9,$4E,$16
	.dc.b	$9C,$C6,$1E,$37,$FC,$F9,$B3,$FE,$60,$76,$FB,$4E,$55,$D5,$C4,$70
	.dc.b	$73,$5D,$67,$57,$E5,$6D,$F7,$F9,$AF,$54,$62,$FC,$0D,$79,$9D,$A3
	.dc.b	$77,$D4,$91,$92,$DF,$78,$B7,$DE,$69,$DA,$2B,$F1,$6B,$AB,$EC,$BF
	.dc.b	$16,$1F,$0B,$0F,$9D,$5A,$7B,$56,$EF,$E6,$4E,$70,$F9,$98,$72,$39
	.dc.b	$C6,$6B,$C4,$71,$3D,$45,$BE,$A4,$AF,$54,$65,$EB,$7E,$AC,$3E,$91
	.dc.b	$F6,$27,$FD,$E6,$D7,$EF,$79,$3F,$C0,$DE,$27,$FB,$CE,$3F,$CC,$3C
	.dc.b	$7F,$85,$9F,$F2,$0E,$1F,$A8,$E3,$FF,$33,$F7,$66,$A6,$1F,$82,$BD
	.dc.b	$E5,$FE,$08,$BD,$CE,$6B,$78,$7E,$5E,$F7,$3F,$EA,$EA,$DC,$E6,$B7
	.dc.b	$7E,$87,$F3,$73,$A8,$C5,$75,$25,$79,$FD,$F6,$3C,$A7,$F3,$98,$F8
	.dc.b	$1F,$D2,$37,$0E,$91,$AC,$78,$3A,$7C,$37,$D2,$3B,$E3,$E2,$C7,$D1
	.dc.b	$1F,$A7,$BE,$6F,$13,$67,$53,$55,$CE,$78,$92,$BA,$1F,$2E,$34,$E2
	.dc.b	$79,$4E,$DF,$95,$3A,$73,$3B,$7C,$53,$F1,$71,$F6,$E5,$F8,$09,$F2
	.dc.b	$9F,$D2,$6F,$93,$C2,$F9,$4E,$EF,$94,$F0,$C3,$DF,$39,$AF,$46,$1F
	.dc.b	$2A,$7E,$19,$3E,$1B,$AB,$8F,$36,$98,$F8,$EB,$5F,$80,$8B,$8F,$CD
	.dc.b	$9B,$97,$E5,$5B,$8F,$E4,$5B,$3F,$8D,$7E,$C7,$57,$D9,$8F,$EC,$D1
	.dc.b	$8C,$3F,$31,$AD,$9E,$10,$55,$C7,$EC,$5F,$11,$B9,$43,$C7,$C3,$AB
	.dc.b	$E5,$7C,$67,$37,$27,$27,$7E,$BF,$E7,$5C,$F9,$C7,$C3,$3F,$E7,$4C
	.dc.b	$5E,$47,$63,$BF,$C0,$D9,$E7,$5E,$9A,$DC,$62,$3C,$F5,$C2,$31,$3A
	.dc.b	$9E,$E7,$1F,$77,$D7,$F6,$3E,$C6,$D7,$07,$0F,$83,$27,$CA,$7F,$B9
	.dc.b	$A3,$47,$93,$67,$D8,$7C,$78,$C3,$71,$F0,$CE,$BC,$2F,$A8,$F0,$E0
	.dc.b	$B5,$B8,$FB,$AD,$BF,$9E,$AB,$75,$CB,$EA,$9F,$0B,$4E,$BF,$39,$FD
	.dc.b	$A7,$0F,$34,$E9,$B3,$B7,$D5,$3B,$CB,$CF,$5B,$7E,$48,$FB,$17,$D4
	.dc.b	$FA,$BE,$B1,$F2,$3F,$89,$9C,$E3,$F1,$67,$2F,$D7,$8C,$DF,$46,$ED
	.dc.b	$F0,$3D,$EE,$58,$CB,$E0,$FC,$68,$C5,$F9,$BC,$4F,$84,$7C,$3F,$9C
	.dc.b	$B7,$F0,$22,$CF,$B2,$6E,$1F,$3B,$EC,$7E,$8F,$5F,$D2,$6E,$1F,$69
	.dc.b	$B7,$F6,$B8,$D1,$8B,$FD,$97,$D7,$3E,$7B,$EE,$FD,$8B,$EC,$3F,$59
	.dc.b	$39,$CB,$F1,$13,$97,$F3,$51,$9C,$BE,$7B,$70,$FD,$FF,$1B,$EC,$79
	.dc.b	$E5,$30,$E2,$FE,$0E,$31,$87,$9F,$CF,$FC,$6E,$7F,$8E,$6D,$FE,$59
	.dc.b	$B3,$B8,$FC,$CC,$FE,$37,$E0,$24,$7E,$2B,$E0,$6D,$7E,$1A,$6E,$7E
	.dc.b	$69,$B5,$FE,$F0,$D3,$FD,$F6,$FE,$0F,$CC,$F8,$7E,$4F,$E2,$5B,$1F
	.dc.b	$0A,$3F,$47,$F6,$18,$CB,$E1,$C5,$C7,$F4,$B3,$6B,$F8,$9E,$EF,$96
	.dc.b	$FD,$27,$C8,$FE,$95,$B3,$FA,$26,$E3,$F9,$E6,$E1,$FB,$0F,$1B,$F6
	.dc.b	$67,$ED,$3D,$5B,$7F,$BC,$8B,$7F,$83,$9B,$5F,$CC,$FE,$C5,$9F,$ED
	.dc.b	$7E,$5B,$F9,$64,$63,$1F,$D9,$C6,$26,$47,$ED,$BB,$6E,$3F,$B0,$BE
	.dc.b	$0D,$AF,$E0,$13,$AC,$3F,$A1,$7D,$83,$EC,$7F,$9D,$37,$1F,$E2,$DD
	.dc.b	$FF,$EB,$C7,$F9,$44,$67,$1F,$DE,$F9,$DF,$E1,$EC,$3F,$A6,$36,$BF
	.dc.b	$9F,$DA,$3F,$A3,$19,$AF,$E9,$9F,$5E,$F9,$9F,$E9,$4D,$BF,$E4,$DF
	.dc.b	$63,$FC,$CF,$1F,$EB,$13,$9B,$FE,$1A,$B1,$E2,$7F,$26,$FB,$1D,$57
	.dc.b	$08,$EB,$E7,$4C,$7F,$99,$53,$D1,$C6,$3F,$C6,$3D,$57,$3F,$F5,$D6
	.dc.b	$DF,$F3,$9F,$B1,$FD,$63,$FD,$2E,$33,$5C,$7E,$83,$81,$FC,$E3,$EC
	.dc.b	$5E,$B6,$2F,$FA,$97,$D8,$BE,$6F,$FB,$53,$63,$FA,$37,$84,$F0,$F3
	.dc.b	$5A,$F6,$E5,$6B,$DE,$BA,$7F,$30,$C7,$FB,$C4,$F6,$8F,$E9,$3F,$63
	.dc.b	$F9,$F8,$C1,$FD,$8C,$7C,$D7,$F7,$B6,$9F,$D5,$E3,$F0,$1C,$DF,$E0
	.dc.b	$98,$CD,$7F,$3E,$FB,$17,$A5,$FE,$03,$6C,$FF,$5F,$F4,$7F,$01,$49
	.dc.b	$CB,$FD,$1B,$EC,$5E,$BF,$F8,$2D,$B7,$FD,$DD,$7C,$0E,$67,$5F,$23
	.dc.b	$2B,$FB,$E7,$D8,$7F,$BB,$5E,$11,$FE,$15,$8C,$61,$FD,$51,$BA,$7F
	.dc.b	$88,$FE,$C7,$E7,$5F,$2F,$4B,$87,$99,$23,$FC,$11,$F0,$CF,$67,$19
	.dc.b	$7F,$5E,$8C,$4E,$A7,$FC,$49,$F6,$3F,$E0,$1A,$D4,$63,$CE,$FF,$14
	.dc.b	$5F,$09,$FE,$CF,$5B,$CF,$FC,$6D,$39,$7F,$C7,$1F,$03,$F8,$5A,$FF
	.dc.b	$02,$FD,$6F,$ED,$5F,$ED,$F1,$99,$FF,$1F,$CE,$5E,$3D,$7F,$E3,$79
	.dc.b	$D6,$7F,$E3,$4A,$FE,$DF,$7C,$AB,$44,$9F,$F0,$C4,$EA,$F9,$C7,$F8
	.dc.b	$AE,$FE,$AB,$FE,$1D,$FB,$1F,$EA,$4F,$F9,$27,$0F,$F1,$1F,$1F,$F8
	.dc.b	$6A,$73,$1F,$E4,$DA,$E2,$FA,$A7,$5F,$29,$E7,$7D,$1F,$11,$30,$F0
	.dc.b	$55,$3E,$0C,$5C,$7E,$3C,$59,$F1,$47,$EC,$9F,$F2,$13,$63,$FC,$2E
	.dc.b	$7F,$97,$FC,$4F,$F3,$07,$D8,$BE,$5B,$8F,$89,$FE,$5F,$BE,$07,$F9
	.dc.b	$AC,$B8,$7B,$F0,$E4,$2F,$93,$FE,$62,$AF,$F3,$43,$A8,$CB,$FE,$6E
	.dc.b	$8F,$F5,$74,$62,$35,$7E,$2A,$F2,$BD,$E1,$86,$1E,$D3,$5F,$A5,$17
	.dc.b	$0F,$81,$7D,$67,$89,$E7,$BE,$CD,$3D,$95,$C7,$0F,$2C,$F8,$63,$18
	.dc.b	$77,$AF,$39,$E2,$39,$1E,$22,$E1,$E2,$AE,$51,$8B,$EA,$F8,$0F,$71
	.dc.b	$F3,$0B,$78,$70,$7E,$3F,$EC,$7F,$8D,$FC,$4F,$06,$C7,$C7,$6B,$F5
	.dc.b	$DA,$F1,$F2,$FF,$D3,$EE,$DA,$ED,$AE,$F1,$ED,$1C,$33,$FB,$79,$34
	.dc.b	$FD,$14,$61,$B1,$C6,$FE,$2E,$1F,$D4,$F0,$FB,$13,$C6,$6D,$70,$7F
	.dc.b	$D6,$95,$B7,$8E,$1E,$8F,$9B,$ED,$1F,$EB,$96,$A7,$F1,$F3,$C5,$B3
	.dc.b	$E9,$BF,$DC,$1E,$1A,$FB,$2D,$C7,$78,$FE,$37,$0F,$14,$76,$24,$FB
	.dc.b	$5F,$0C,$FF,$B4,$63,$A3,$71,$FD,$AE,$BB,$9F,$0C,$7D,$D7,$C5,$87
	.dc.b	$EB,$5A,$E8,$FC,$2F,$D8,$FF,$6E,$7F,$D7,$CD,$7F,$D2,$6D,$BF,$F7
	.dc.b	$13,$FE,$01,$8D,$ED,$F8,$B8,$E1,$1F,$BB,$6B,$FC,$6D,$AF,$FB,$97
	.dc.b	$EC,$7F,$A2,$3F,$96,$6B,$FB,$56,$BF,$C1,$5A,$FF,$25,$77,$5F,$EF
	.dc.b	$5F,$B0,$FE,$06,$DC,$3E,$2D,$77,$EA,$AA,$FF,$C2,$EF,$FB,$75,$AF
	.dc.b	$FA,$C9,$B5,$D7,$EC,$78,$C6,$A7,$C9,$E8,$7E,$AE,$31,$1E,$58,$C1
	.dc.b	$E0,$C7,$FB,$9C,$7F,$BA,$DA,$F7,$8F,$2F,$D8,$E3,$06,$DF,$D0,$E3
	.dc.b	$0D,$E5,$E9,$F1,$9C,$1F,$DE,$4F,$4C,$3F,$96,$35,$E0,$7F,$04,$F4
	.dc.b	$7F,$DB,$CD,$F8,$FF,$E0,$91,$88,$FE,$4C,$D7,$F9,$B2,$F2,$76,$48
	.dc.b	$E8,$F2,$FB,$19,$1D,$DA,$FF,$31,$FB,$1F,$F4,$53,$FC,$59,$AE,$7E
	.dc.b	$0F,$F8,$B4,$62,$3F,$A3,$7D,$8B,$E1,$FF,$8D,$36,$3F,$8F,$B5,$FE
	.dc.b	$7C,$D7,$FA,$77,$D8,$FF,$9A,$5F,$E4,$AD,$7F,$A2,$35,$DF,$BF,$FC
	.dc.b	$9E,$3F,$41,$1A,$ED,$C1,$1E,$7E,$E6,$C8,$E0,$FE,$66,$FE,$19,$07
	.dc.b	$0D,$E4,$FE,$6D,$18,$1F,$61,$82,$4F,$47,$F9,$E7,$D8,$7F,$99,$63
	.dc.b	$FA,$0B,$5F,$EC,$0D,$7F,$B8,$35,$FF,$08,$2F,$67,$99,$23,$9B,$FD
	.dc.b	$27,$EB,$FF,$D6,$31,$FD,$39,$AF,$F6,$D6,$BF,$DF,$5A,$FF,$87,$5F
	.dc.b	$F3,$4B,$C7,$D8,$E8,$47,$FA,$CF,$D7,$7F,$99,$9B,$EC,$7F,$59,$8C
	.dc.b	$47,$F7,$B6,$BF,$E0,$B5,$F5,$DB,$24,$74,$7F,$B1,$7B,$9F,$EA,$38
	.dc.b	$FA,$7D,$BF,$B0,$7E,$AC,$65,$46,$54,$66,$31,$F5,$97,$BB,$FC,$75
	.dc.b	$3C,$6B,$FC,$3C,$D7,$3F,$07,$E2,$23,$E2,$9F,$F2,$15,$7E,$CE,$FA
	.dc.b	$F6,$7F,$1A,$C3,$B5,$FE,$2B,$EB,$9F,$1F,$FC,$79,$B1,$CC,$D4,$67
	.dc.b	$C6,$D3,$23,$42,$78,$BF,$03,$2F,$F2,$A4,$7F,$2E,$6F,$D3,$F7,$FA
	.dc.b	$89,$5C,$A9,$3E,$33,$31,$CC,$F4,$4E,$5F,$0E,$2F,$9F,$F7,$12,$3A
	.dc.b	$7D,$75,$7F,$C7,$53,$B7,$FC,$3E,$D7,$33,$F7,$8F,$E8,$3F,$4B,$FE
	.dc.b	$51,$CB,$FC,$DC,$D9,$FB,$F3,$AF,$AE,$91,$F1,$25,$1C,$B8,$33,$3E
	.dc.b	$B1,$3A,$78,$4E,$DE,$5F,$0F,$E2,$F8,$89,$18,$21,$CE,$91,$A8,$C4
	.dc.b	$FF,$5A,$FA,$D7,$D7,$52,$4F,$DA,$FA,$EF,$F0,$C3,$F7,$27,$6F,$F9
	.dc.b	$11,$AF,$F9,$8D,$AF,$F9,$CD,$AF,$FA,$4D,$DF,$2F,$E8,$67,$E7,$37
	.dc.b	$4F,$0D,$7C,$39,$7C,$17,$F7,$05,$F2,$F4,$C4,$49,$F4,$46,$5F,$54
	.dc.b	$71,$FA,$39,$1E,$D7,$A7,$46,$23,$C1,$17,$1F,$34,$FF,$80,$3E,$B9
	.dc.b	$EF,$52,$5F,$E8,$3E,$B9,$F1,$58,$8F,$F4,$35,$70,$7F,$CD,$6D,$73
	.dc.b	$5F,$E6,$B8,$C6,$1F,$C1,$6F,$F4,$D5,$FC,$F9,$AA,$77,$CB,$FE,$BF
	.dc.b	$9F,$D3,$CF,$8F,$A7,$FC,$2F,$18,$8F,$C7,$2E,$9C,$CC,$9E,$07,$AE
	.dc.b	$38,$FD,$74,$8F,$73,$FE,$22,$FA,$EF,$D0,$42,$D3,$B8,$EE,$EF,$0F
	.dc.b	$F1,$4F,$D7,$7F,$45,$9F,$F4,$6B,$5F,$F5,$63,$5F,$F5,$E3,$5F,$F6
	.dc.b	$E3,$BC,$FF,$95,$2F,$98,$EA,$76,$F1,$FF,$C6,$FF,$5D,$FC,$87,$0F
	.dc.b	$F7,$23,$70,$FE,$D2,$D5,$7F,$EC,$BC,$3F,$9B,$4F,$F8,$66,$78,$F0
	.dc.b	$FF,$A8,$63,$11,$CA,$31,$F1,$B2,$3F,$D5,$32,$8C,$47,$FB,$26,$30
	.dc.b	$FF,$28,$7F,$3A,$DE,$5F,$4C,$FA,$BE,$03,$C0,$49,$EA,$D7,$FD,$A7
	.dc.b	$D1,$EE,$3F,$DA,$25,$7F,$DD,$CF,$C2,$8C,$C7,$C2,$8A,$AF,$F9,$17
	.dc.b	$C3,$7F,$23,$6C,$F1,$9E,$67,$13,$FC,$27,$A7,$FB,$CE,$7F,$95,$36
	.dc.b	$7F,$C2,$F3,$FD,$84,$EC,$E2,$3F,$DF,$5F,$59,$FD,$D1,$BE,$56,$6B
	.dc.b	$8C,$E5,$B1,$97,$FC,$47,$1F,$DF,$30,$FE,$74,$BE,$B5,$F1,$BE,$8A
	.dc.b	$4E,$63,$0F,$1C,$7B,$D7,$0A,$F9,$A7,$67,$17,$D9,$B3,$E4,$9E,$71
	.dc.b	$87,$85,$E0,$EF,$87,$D0,$C3,$C5,$3D,$8F,$2D,$71,$6D,$70,$79,$1B
	.dc.b	$9E,$38,$78,$E7,$53,$9A,$E5,$5A,$C3,$DB,$5A,$BE,$17,$B8,$EB,$1F
	.dc.b	$2B,$0F,$36,$1C,$89,$3C,$6F,$9D,$F1,$25,$F1,$79,$C6,$A3,$B9,$98
	.dc.b	$DC,$64,$8F,$DC,$BF,$0D,$72,$8F,$96,$7C,$18,$CB,$C8,$CD,$75,$3D
	.dc.b	$AF,$E3,$AF,$EE,$3F,$90,$F1,$7F,$85,$3F,$96,$3A,$99,$E3,$C3,$F8
	.dc.b	$E8,$DB,$93,$39,$7F,$8F,$27,$71,$F8,$E9,$EB,$7F,$94,$6E,$5F,$A4
	.dc.b	$E3,$FF,$4A,$69,$F9,$29,$CD,$F7,$C3,$07,$E6,$6B,$71,$F8,$19,$DB
	.dc.b	$DE,$76,$FE,$2E,$77,$87,$86,$72,$D9,$D7,$AF,$FE,$59,$8C,$13,$1F
	.dc.b	$D8,$E1,$DB,$0E,$71,$8C,$39,$B7,$0E,$78,$E1,$F5,$1E,$79,$F5,$63
	.dc.b	$DE,$B8,$63,$DE,$B6,$FF,$8C,$F0,$EF,$18,$9D,$91,$FF,$51,$C6,$9F
	.dc.b	$84,$77,$24,$F3,$AF,$04,$73,$21,$C2,$35,$87,$A6,$31,$7D,$DF,$F4
	.dc.b	$F5,$72,$72,$78,$9C,$65,$F8,$38,$F8,$67,$8A,$7F,$01,$3F,$E8,$7A
	.dc.b	$FC,$86,$1F,$71,$B1,$9A,$DC,$7F,$2E,$7F,$11,$E4,$E5,$F8,$2D,$7F
	.dc.b	$D6,$4D,$AF,$F3,$EC,$F1,$9D,$BF,$8B,$9F,$54,$BE,$16,$C6,$1B,$3F
	.dc.b	$91,$6D,$7E,$35,$FE,$78,$FD,$26,$BF,$EC,$36,$D7,$59,$FD,$2F,$D6
	.dc.b	$BD,$DF,$F5,$47,$3F,$FA,$96,$1E,$0E,$2B,$FA,$4C,$6D,$FF,$47,$BA
	.dc.b	$6D,$78,$23,$58,$FE,$E1,$AF,$EF,$5A,$F4,$F0,$B8,$12,$BC,$AD,$8F
	.dc.b	$F3,$B4,$FF,$95,$27,$C3,$EB,$F9,$67,$FD,$E7,$39,$77,$EB,$65,$43
	.dc.b	$EB,$70,$FF,$BE,$6F,$D7,$18,$AF,$51,$23,$E3,$BF,$3B,$37,$D5,$7C
	.dc.b	$1C,$3F,$D4,$91,$8D,$3C,$8E,$BB,$6D,$6E,$3E,$7B,$70,$FF,$6C,$B6
	.dc.b	$FF,$DA,$EA,$B8,$DF,$4F,$A4,$7D,$6D,$3B,$FC,$BF,$F6,$EC,$7F,$D9
	.dc.b	$19,$2B,$71,$86,$E5,$E8,$6C,$FE,$0A,$B9,$4E,$6F,$E5,$7D,$63,$E9
	.dc.b	$3D,$BF,$5A,$F8,$9C,$C6,$1B,$9F,$F3,$2A,$E0,$D9,$FC,$AC,$63,$0F
	.dc.b	$E7,$AD,$CB,$FD,$EC,$DA,$F8,$BE,$CF,$4A,$F4,$78,$1F,$0C,$F8,$9F
	.dc.b	$1B,$6B,$DD,$16,$7D,$D3,$70,$FA,$3F,$5B,$F5,$3A,$3E,$AB,$70,$F9
	.dc.b	$3C,$68,$C6,$1F,$7A,$31,$3A,$AF,$C7,$7D,$4B,$EB,$3F,$03,$F5,$AF
	.dc.b	$AC,$CC,$61,$B7,$FD,$5A,$B8,$36,$7F,$73,$18,$C7,$FB,$0B,$73,$57
	.dc.b	$F8,$F2,$30,$A5,$3F,$81,$FA,$DE,$F1,$E9,$F2,$FF,$27,$8F,$E5,$1B
	.dc.b	$7F,$01,$B3,$B8,$F8,$73,$F9,$1F,$66,$47,$E2,$FD,$AB,$5F,$8B,$9B
	.dc.b	$8F,$F0,$CE,$9F,$F7,$1D,$FE,$46,$7E,$3F,$87,$F8,$AF,$12,$CF,$E6
	.dc.b	$3A,$7F,$B6,$E1,$EE,$C7,$D9,$B7,$F2,$A7,$E1,$E1,$F1,$89,$7F,$2D
	.dc.b	$E7,$1F,$5F,$EB,$2D,$F8,$FE,$37,$F5,$5F,$13,$FA,$B6,$E1,$FA,$56
	.dc.b	$DF,$E9,$5B,$5F,$B8,$F1,$7F,$76,$7E,$F3,$D1,$B7,$F8,$88,$B7,$F8
	.dc.b	$79,$B5,$EB,$87,$D6,$B2,$FD,$E7,$D0,$7F,$17,$8C,$63,$FB,$78,$C4
	.dc.b	$FF,$3C,$9C,$D7,$F8,$91,$AF,$F8,$A5,$AF,$FA,$01,$DD,$F8,$5B,$7F
	.dc.b	$87,$8F,$E0,$DD,$FD,$A3,$07,$FA,$23,$EB,$1F,$5B,$FD,$31,$B1,$FE
	.dc.b	$2E,$6E,$1F,$E3,$66,$BF,$E3,$BF,$43,$FC,$4E,$DE,$3F,$EA,$B1,$88
	.dc.b	$FF,$4A,$B5,$FC,$D7,$D6,$3E,$43,$FA,$DB,$70,$FF,$1E,$B6,$7F,$C8
	.dc.b	$AD,$7F,$C9,$4B,$C3,$E2,$49,$E2,$F4,$F2,$A4,$74,$6D,$7F,$94,$3D
	.dc.b	$27,$2F,$F6,$C6,$D7,$F9,$39,$B3,$B9,$D7,$CA,$4A,$D3,$67,$FC,$B7
	.dc.b	$F5,$2F,$A3,$FE,$EE,$DA,$FF,$2C,$B6,$77,$3A,$F4,$A5,$69,$B3,$D0
	.dc.b	$E6,$48,$E6,$FF,$81,$17,$C7,$D2,$3F,$E6,$83,$27,$F9,$AB,$EA,$1E
	.dc.b	$5F,$F8,$31,$B1,$98,$C1,$FE,$10,$FA,$87,$A1,$49,$3F,$E7,$48,$CB
	.dc.b	$7E,$1B,$81,$1C,$1F,$E7,$AF,$A8,$FF,$7D,$37,$C8,$FF,$0E,$C6,$33
	.dc.b	$FF,$42,$B5,$FF,$45,$3F,$E0,$36,$F8,$1C,$09,$5F,$E8,$6F,$75,$E8
	.dc.b	$7F,$8A,$1B,$19,$8C,$11,$FD,$2D,$7F,$2F,$9D,$BF,$CC,$23,$FD,$DF
	.dc.b	$F4,$EF,$0B,$FC,$25,$5A,$F0,$3F,$81,$7D,$2F,$DD,$7F,$C6,$3F,$4E
	.dc.b	$F8,$8F,$F0,$C4,$E7,$1F,$DB,$4E,$5F,$0E,$49,$FC,$33,$FE,$A9,$6B
	.dc.b	$FF,$00,$1A,$FF,$45,$6B,$C7,$9B,$B9,$FE,$18,$AF,$EF,$95,$DA,$BC
	.dc.b	$55,$C7,$E7,$BF,$C3,$BF,$4F,$FD,$42,$31,$5F,$E1,$08,$C4,$79,$67
	.dc.b	$FC,$A7,$46,$92,$BF,$D6,$F1,$FE,$55,$9F,$F5,$FB,$5F,$67,$BF,$C8
	.dc.b	$93,$E5,$7E,$3C,$7C,$87,$FC,$81,$9B,$E1,$7F,$CC,$6D,$AF,$F6,$2B
	.dc.b	$6B,$FD,$9A,$D7,$FD,$A4,$E4,$FE,$70,$58,$FE,$09,$0F,$A4,$7C,$B7
	.dc.b	$F7,$18,$FE,$E3,$7F,$E5,$17,$FB,$A7,$81,$FD,$D6,$B7,$1F,$CC,$67
	.dc.b	$FA,$F4,$62,$39,$4E,$EB,$A6,$1E,$7A,$D5,$74,$3C,$53,$AB,$E6,$F9
	.dc.b	$A7,$A5,$71,$8F,$4F,$D3,$62,$BF,$B7,$3D,$2D,$87,$F8,$2E,$35,$8D
	.dc.b	$BE,$87,$77,$CD,$5F,$EF,$9A,$AF,$C2,$8A,$F9,$F1,$E6,$FA,$0E,$05
	.dc.b	$38,$14,$F7,$1F,$CB,$CF,$F7,$E9,$E6,$2C,$FF,$C0,$0F,$A6,$FE,$3B
	.dc.b	$F4,$F7,$BC,$D2,$57,$D7,$3F,$C5,$12,$F8,$5F,$F1,$4C,$64,$46,$75
	.dc.b	$ED,$3A,$9E,$53,$97,$83,$71,$FB,$99,$FF,$AF,$67,$E2,$B5,$F8,$13
	.dc.b	$B8,$FC,$53,$63,$FC,$79,$1B,$9E,$7B,$7E,$3D,$FC,$45,$6E,$BC,$F7
	.dc.b	$EA,$D7,$71,$F6,$4E,$77,$F9,$78,$E3,$3B,$C3,$C2,$FC,$2C,$38,$DF
	.dc.b	$D5,$9F,$11,$AC,$7E,$FE,$3E,$18,$CC,$F6,$33,$7C,$E3,$F4,$C7,$C1
	.dc.b	$AF,$F1,$23,$75,$CE,$5F,$A0,$7B,$5F,$86,$3F,$03,$3F,$26,$3E,$D3
	.dc.b	$5E,$27,$D9,$1F,$4F,$FC,$0D,$FE,$52,$D7,$F9,$C1,$FE,$69,$71,$7F
	.dc.b	$E3,$77,$F0,$91,$C3,$7F,$C8,$C7,$18,$FC,$53,$5F,$E0,$4D,$7F,$CE
	.dc.b	$9F,$4D,$FD,$39,$F2,$35,$FC,$73,$5F,$CA,$35,$FE,$18,$EE,$BF,$C9
	.dc.b	$5F,$4D,$FB,$CC,$BF,$88,$36,$BB,$F5,$FC,$97,$C3,$E3,$D5,$87,$F8
	.dc.b	$71,$FE,$96,$D7,$E1,$8A,$7D,$3F,$19,$D4,$F9,$AB,$F6,$71,$88,$F3
	.dc.b	$C6,$0F,$05,$7F,$41,$8F,$EB,$2D,$7B,$CF,$87,$1F,$F3,$37,$D3,$E5
	.dc.b	$BC,$FE,$98,$C6,$7F,$03,$E1,$71,$3F,$C1,$27,$A6,$1F,$BD,$6B,$FC
	.dc.b	$D5,$AF,$FA,$9F,$CA,$FF,$77,$3F,$87,$6B,$FC,$05,$AF,$F0,$76,$BF
	.dc.b	$CF,$D7,$C1,$7F,$14,$6C,$73,$8C,$19,$FA,$79,$1D,$DA,$FF,$0F,$FA
	.dc.b	$77,$B5,$FC,$79,$B1,$F9,$B6,$BB,$8F,$E4,$7F,$4E,$F5,$69,$27,$F4
	.dc.b	$51,$97,$F4,$9F,$4F,$FE,$6D,$6F,$AB,$FC,$B2,$31,$1F,$C8,$5A,$FF
	.dc.b	$27,$7F,$46,$E7,$AB,$81,$1E,$BE,$9F,$32,$38,$3F,$68,$FE,$1D,$FF
	.dc.b	$6C,$35,$34,$9F,$4F,$FD,$01,$34,$91,$FB,$FF,$A7,$FF,$36,$37,$C7
	.dc.b	$FE,$8F,$18,$8F,$E7,$8D,$7F,$A1,$2F,$53,$64,$8E,$8F,$F5,$0F,$A7
	.dc.b	$FF,$5C,$47,$F0,$D6,$BF,$D1,$DA,$FF,$4D,$6B,$FE,$05,$7F,$CE,$2F
	.dc.b	$1F,$17,$A1,$1F,$E3,$1F,$4F,$FE,$70,$6F,$AB,$FD,$82,$31,$1F,$D5
	.dc.b	$5A,$FF,$5B,$5F,$4F,$B2,$47,$47,$FB,$4F,$CB,$7F,$A9,$63,$FC,$96
	.dc.b	$D7,$F8,$0B,$7B,$3F,$DB,$1F,$4B,$7B,$7F,$C3,$D3,$BA,$FE,$D4,$D7
	.dc.b	$3E,$DF,$62,$3F,$E5,$8A,$FD,$6D,$F4,$F9,$3F,$E2,$97,$D6,$83,$E9
	.dc.b	$DF,$0D,$FC,$65,$B1,$FE,$62,$6B,$FC,$61,$FE,$44,$F1,$8E,$04,$78
	.dc.b	$79,$3C,$FD,$3E,$9F,$3F,$27,$A4,$65,$ED,$19,$7B,$47,$A3,$E6,$24
	.dc.b	$7C,$A7,$CD,$F4,$F8,$8F,$F0,$56,$49,$FA,$24,$8F,$EA,$D1,$97,$FC
	.dc.b	$FB,$F4,$DF,$E5,$0E,$0F,$F2,$63,$63,$E2,$CE,$DF,$E6,$F0,$FA,$7F
	.dc.b	$E7,$CD,$8E,$71,$8E,$0C,$B2,$32,$78,$23,$8F,$C7,$72,$F9,$09,$3E
	.dc.b	$48,$CB,$E6,$8C,$BE,$28,$FE,$9D,$F4,$CF,$A7,$A4,$9F,$F4,$CF,$D3
	.dc.b	$BD,$8C,$47,$F8,$BE,$B8,$3F,$DE,$5A,$E6,$BF,$C8,$31,$8E,$66,$DF
	.dc.b	$E7,$1C,$3E,$24,$62,$FF,$17,$18,$9F,$F3,$B5,$71,$78,$7C,$0F,$F6
	.dc.b	$59,$7C,$85,$31,$12,$FD,$31,$97,$D7,$1C,$7E,$72,$47,$B9,$E9,$D3
	.dc.b	$88,$FE,$DB,$69,$C6,$5E,$79,$4F,$E9,$B1,$FF,$46,$7D,$33,$E9,$FF
	.dc.b	$BD,$36,$BF,$D8,$6D,$7F,$BC,$FD,$33,$E2,$77,$5F,$E6,$8A,$D5,$7F
	.dc.b	$98,$5A,$F6,$8D,$1F,$E0,$7A,$FD,$59,$FA,$C9,$ED,$8F,$9E,$3F,$DB
	.dc.b	$15,$FE,$D4,$6B,$E2,$E9,$E5,$D5,$CD,$91,$B7,$2F,$1F,$A6,$F0,$11
	.dc.b	$FF,$6F,$FD,$3F,$E8,$55,$A7,$51,$F0,$1F,$F0,$F4,$EF,$E9,$FF,$9E
	.dc.b	$4F,$F9,$0D,$AF,$F9,$F5,$AF,$FA,$0D,$AF,$FA,$ED,$DE,$3F,$55,$7C
	.dc.b	$97,$F8,$AD,$B3,$FE,$F4,$77,$B7,$FB,$0F,$E9,$FE,$D4,$FF,$1C,$6A
	.dc.b	$6C,$FF,$B4,$27,$87,$17,$FA,$7A,$7F,$C1,$F5,$C6,$30,$67,$E3,$A4
	.dc.b	$7F,$97,$E5,$18,$8F,$F4,$D4,$61,$FE,$5B,$8F,$E3,$E3,$F3,$F1,$8B
	.dc.b	$F4,$72,$FF,$91,$E7,$84,$FF,$91,$1F,$F2,$E3,$E9,$F8,$2F,$31,$1F
	.dc.b	$13,$FE,$DA,$F0,$3E,$11,$63,$E1,$35,$E4,$7F,$B9,$9D,$B4,$D9,$4F
	.dc.b	$90,$F1,$9F,$F3,$D7,$C3,$FF,$23,$7F,$92,$61,$FE,$BD,$6D,$74,$3F
	.dc.b	$5C,$72,$3F,$A7,$C7,$FB,$0F,$E9,$BF,$B6,$B7,$CA,$FF,$7E,$36,$76
	.dc.b	$7F,$BF,$E7,$6D,$9F,$64,$3E,$9D,$F1,$BE,$98,$C3,$67,$73,$AA,$FF
	.dc.b	$78,$E5,$D6,$31,$97,$F3,$FC,$38,$D6,$8F,$EC,$EF,$F8,$12,$35,$3C
	.dc.b	$2B,$96,$1E,$18,$C5,$F0,$C3,$C3,$39,$24,$75,$18,$75,$BE,$71,$98
	.dc.b	$F1,$E1,$E0,$AF,$64,$F2,$8C,$3B,$C3,$C2,$F4,$C3,$C2,$E6,$B8,$61
	.dc.b	$F2,$E2,$9E,$EA,$E3,$87,$82,$75,$3C,$49,$87,$80,$F9,$18,$76,$71
	.dc.b	$3E,$A7,$33,$D8,$F7,$DF,$63,$BD,$71,$AE,$04,$9F,$94,$71,$BF,$23
	.dc.b	$E7,$C3,$E1,$BE,$53,$F1,$D7,$F7,$DB,$1F,$ED,$79,$FB,$B1,$83,$FC
	.dc.b	$46,$7D,$03,$F2,$DE,$2F,$F1,$07,$46,$38,$BF,$1D,$1F,$C5,$32,$FF
	.dc.b	$03,$79,$FF,$E8,$5D,$33,$52,$FE,$D5,$EF,$3E,$F7,$B9,$CD,$6F,$0F
	.dc.b	$C9,$5E,$E7,$FC,$6F,$5B,$9C,$D6,$EF,$F4,$D7,$BB,$FF,$37,$7B,$1F
	.dc.b	$DD,$6F,$71,$D8,$8F,$A2,$35,$5C,$AF,$A5,$7E,$82,$FA,$61,$E1,$AD
	.dc.b	$4F,$26,$E1,$CF,$1E,$33,$F9,$CB,$F2,$64,$79,$E5,$E0,$9E,$D9,$70
	.dc.b	$AD,$3F,$C4,$F2,$F0,$4E,$8F,$01,$71,$F0,$38,$3C,$13,$E2,$9E,$31
	.dc.b	$83,$55,$D9,$FD,$94,$6A,$31,$5D,$0F,$92,$5C,$BD,$E7,$DA,$C3,$FC
	.dc.b	$45,$97,$E2,$B0,$FA,$2F,$F1,$A7,$EC,$C6,$9E,$11,$97,$FD,$49,$E5
	.dc.b	$65,$FB,$ED,$73,$1B,$C7,$FC,$EE,$D9,$F9,$53,$A3,$D7,$F0,$F8,$AD
	.dc.b	$C6,$2B,$E6,$56,$6B,$4D,$8F,$54,$FE,$42,$7F,$9D,$B6,$32,$FF,$AE
	.dc.b	$E3,$2B,$E9,$FF,$DE,$F3,$FD,$7E,$3F,$D1,$F1,$FE,$33,$8C,$46,$FD
	.dc.b	$EF,$F2,$57,$2F,$F8,$D6,$B5,$28,$C4,$F6,$9E,$B3,$E6,$6B,$F4,$5A
	.dc.b	$FE,$F1,$78,$7C,$49,$7E,$A8,$CB,$63,$C6,$F1,$6C,$FA,$BD,$8A,$FB
	.dc.b	$E2,$D7,$FB,$B2,$BD,$AB,$E9,$F3,$EC,$E9,$B3,$E1,$8F,$61,$1F,$A3
	.dc.b	$93,$E9,$3E,$1B,$F1,$9B,$1F,$E0,$E5,$D9,$E6,$9E,$1E,$3D,$9D,$E7
	.dc.b	$F2,$DB,$3D,$6B,$55,$C7,$E8,$1F,$4F,$4E,$9F,$2F,$FD,$BB,$2F,$B9
	.dc.b	$82,$B7,$18,$6E,$5F,$66,$3F,$C4,$71,$89,$FB,$F5,$CE,$B4,$7D,$F9
	.dc.b	$D5,$7F,$34,$FA,$77,$CF,$7B,$7E,$89,$F1,$5F,$4A,$73,$5E,$49,$CB
	.dc.b	$F8,$E8,$CE,$1F,$9C,$6B,$FB,$76,$BD,$EF,$F3,$71,$C3,$2F,$F4,$43
	.dc.b	$6B,$E8,$7B,$5E,$1A,$F9,$1E,$0C,$9F,$85,$F1,$F7,$1F,$6E,$3F,$1E
	.dc.b	$B0,$FE,$83,$45,$62,$FE,$E7,$D3,$FD,$9E,$8F,$B4,$D8,$FA,$3C,$68
	.dc.b	$C6,$5D,$A3,$13,$AA,$FC,$BF,$D1,$BE,$9B,$F0,$DF,$4E,$FA,$6C,$C6
	.dc.b	$1B,$7F,$BE,$AE,$0D,$9F,$DF,$C6,$31,$FE,$0E,$DD,$7F,$D5,$CD,$AF
	.dc.b	$C0,$FD,$3F,$4A,$F3,$79,$9F,$95,$C7,$F2,$CD,$AF,$6B,$67,$71,$F0
	.dc.b	$27,$F2,$DE,$D4,$8F,$C8,$7B,$76,$BE,$7F,$0A,$B1,$9F,$CF,$AB,$7F
	.dc.b	$E0,$0C,$FF,$92,$CF,$F1,$2E,$6F,$EE,$8F,$AB,$C2,$FC,$87,$89,$F2
	.dc.b	$9F,$82,$4B,$F8,$CD,$CF,$F3,$63,$0E,$B3,$FB,$7F,$A7,$E4,$FE,$9F
	.dc.b	$E3,$7F,$61,$F2,$BF,$B1,$6C,$64,$FD,$64,$65,$A7,$E0,$BC,$4F,$DF
	.dc.b	$1F,$BF,$F4,$6C,$64,$98,$FE,$16,$32,$F9,$BE,$9D,$8F,$F0,$1F,$9F
	.dc.b	$FE,$3B,$18,$BF,$DE,$46,$27,$F9,$FC,$E6,$BF,$A3,$B5,$FE,$C4,$D7
	.dc.b	$FC,$5A,$EF,$6F,$F7,$93,$6F,$F8,$2F,$6F,$57,$FA,$2F,$D3,$3E,$9F
	.dc.b	$FA,$6B,$63,$FA,$8B,$70,$FE,$AE,$D7,$FB,$97,$A1,$FE,$2D,$6F,$17
	.dc.b	$F5,$68,$C4,$7F,$90,$1A,$87,$D3,$BE,$4B,$FA,$EB,$70,$FE,$C2,$D9
	.dc.b	$FE,$CE,$D7,$FC,$04,$BC,$3E,$24,$9E,$2F,$4F,$2A,$47,$46,$D7,$F7
	.dc.b	$2F,$45,$C9,$FD,$B9,$B5,$FD,$C9,$B3,$B9,$D7,$CB,$4A,$D3,$67,$FB
	.dc.b	$EF,$D3,$BE,$4B,$FB,$CB,$6B,$FB,$EB,$67,$73,$AF,$4E,$56,$9B,$3D
	.dc.b	$0E,$64,$8E,$6F,$F8,$19,$7C,$85,$23,$FE,$0C,$32,$7F,$83,$BE,$9D
	.dc.b	$E5,$FF,$83,$9B,$19,$8C,$1F,$E1,$2F,$A7,$79,$F4,$93,$FE,$18,$8C
	.dc.b	$B7,$E1,$B8,$11,$C1,$FE,$1E,$FA,$7F,$F7,$FB,$7C,$8F,$F0,$FC,$63
	.dc.b	$3F,$F1,$3B,$5F,$F2,$13,$FE,$0C,$73,$E0,$70,$25,$7F,$8A,$BD,$CF
	.dc.b	$E8,$2D,$F4,$3F,$C5,$71,$8A,$FF,$19,$B5,$FF,$27,$39,$EF,$57,$FA
	.dc.b	$6A,$30,$FF,$A7,$3E,$9D,$E1,$7F,$86,$27,$37,$F9,$FF,$A5,$FB,$D8
	.dc.b	$7F,$8C,$BE,$9D,$F0,$FF,$E1,$E9,$CE,$5F,$A5,$A5,$6A,$3F,$C7,$73
	.dc.b	$F8,$7B,$FF,$31,$35,$FF,$56,$B5,$F0,$73,$7F,$92,$23,$AC,$62,$7F
	.dc.b	$C1,$B5,$FD,$EA,$3F,$95,$CE,$92,$F9,$FF,$F1,$07,$D3,$FF,$4F,$9F
	.dc.b	$64,$63,$D2,$F5,$3D,$C9,$87,$F9,$6A,$F9,$37,$DF,$F2,$11,$C1,$2F
	.dc.b	$C7,$1F,$15,$F8,$CF,$F9,$13,$37,$C2,$FF,$99,$1B,$5F,$E6,$86,$D7
	.dc.b	$F8,$FA,$74,$71,$3F,$C9,$AF,$F1,$3E,$FF,$E2,$9F,$4C,$F9,$6F,$F6
	.dc.b	$BD,$F2,$3F,$CE,$B3,$A8,$CD,$FE,$12,$D7,$C5,$FF,$4B,$56,$DF,$B9
	.dc.b	$3B,$9D,$47,$FB,$92,$74,$FF,$9C,$E7,$55,$FC,$56,$3A,$BE,$0C,$37
	.dc.b	$1E,$91,$F2,$5D,$31,$F8,$2F,$68,$D3,$9B,$F4,$39,$86,$1E,$2A,$F9
	.dc.b	$38,$E6,$31,$7E,$4A,$F6,$9E,$43,$DB,$87,$91,$E6,$64,$F6,$39,$9F
	.dc.b	$F4,$EC,$7A,$4F,$E5,$C6,$1F,$F5,$0F,$D3,$7F,$1A,$FA,$7B,$E0,$1A
	.dc.b	$4A,$FF,$55,$CB,$E1,$7F,$D3,$91,$86,$BF,$EC,$98,$D6,$7D,$A7,$33
	.dc.b	$FD,$54,$4E,$9F,$56,$AD,$B5,$AF,$DA,$8C,$4F,$DA,$AE,$13,$F7,$9B
	.dc.b	$3D,$4D,$CF,$4A,$E7,$1E,$19,$DC,$FF,$67,$6B,$F9,$59,$DB,$FD,$AE
	.dc.b	$79,$D7,$7B,$F8,$2D,$C3,$F1,$B3,$A3,$C5,$7C,$AF,$DF,$7C,$27,$C4
	.dc.b	$D8,$ED,$3B,$9E,$51,$E5,$9D,$3E,$53,$FD,$A3,$1E,$EA,$FD,$34,$64
	.dc.b	$FF,$1C,$E1,$EC,$9D,$69,$E4,$9E,$38,$7E,$CA,$3C,$77,$FA,$6F,$A7
	.dc.b	$FE,$03,$5F,$ED,$E6,$BF,$E4,$46,$C7,$81,$FC,$24,$6F,$6E,$8F,$F8
	.dc.b	$4E,$37,$5F,$81,$6C,$67,$2D,$AF,$A5,$FF,$45,$BF,$EE,$E6,$BF,$E5
	.dc.b	$46,$E1,$F0,$3E,$99,$E0,$72,$9C,$E3,$F8,$F9,$FD,$C3,$74,$D7,$6E
	.dc.b	$FC,$89,$97,$77,$C7,$6F,$A7,$CC,$F8,$66,$E7,$E1,$9B,$8F,$7E,$7E
	.dc.b	$26,$2F,$FC,$69,$87,$FB,$01,$FE,$A5,$3F,$A3,$FA,$7F,$E8,$51,$83
	.dc.b	$FC,$74,$E1,$FF,$39,$CF,$17,$FC,$EB,$19,$C3,$F0,$2D,$AE,$97,$EA
	.dc.b	$F8,$5F,$9B,$E5,$FF,$BB,$B0,$EC,$1C,$3F,$E8,$06,$E1,$D2,$B8,$FC
	.dc.b	$14,$AE,$34,$8F,$8F,$E4,$76,$FA,$79,$5A,$78,$3F,$E8,$CF,$A7,$7B
	.dc.b	$7F,$80,$6C,$79,$1A,$EF,$87,$F0,$BF,$4E,$F5,$A9,$27,$D7,$D2,$FA
	.dc.b	$6F,$C5,$FD,$3F,$F9,$86,$7D,$ED,$7F,$D3,$AD,$E0,$FD,$4D,$3E,$7B
	.dc.b	$8F,$5C,$BE,$2F,$E5,$A7,$F1,$B1,$FE,$6F,$B7,$AB,$E0,$24,$7C,$EF
	.dc.b	$B1,$FE,$92,$FD,$08,$FF,$31,$B7,$AB,$F4,$31,$88,$FF,$58,$CF,$F9
	.dc.b	$39,$E1,$9E,$88,$F5,$F8,$EF,$D4,$FD,$47,$F9,$8E,$3E,$EB,$5F,$F5
	.dc.b	$DB,$75,$FE,$0F,$F6,$3F,$CC,$E7,$3E,$C7,$22,$3F,$86,$8F,$D1,$3F
	.dc.b	$88,$FA,$9F,$F5,$7B,$7E,$0F,$F7,$31,$88,$FF,$64,$CF,$F9,$69,$E1
	.dc.b	$E4,$E8,$8F,$5F,$A6,$FE,$03,$F5,$3F,$E6,$68,$FC,$AB,$5F,$F6,$9B
	.dc.b	$72,$FE,$3F,$F6,$2F,$AA,$E2,$4A,$E2,$FF,$0C,$8F,$99,$3F,$C9,$BE
	.dc.b	$6F,$FD,$39,$5F,$A2,$6B,$FE,$E0,$6C,$7F,$2C,$FA,$7B,$E2,$7F,$59
	.dc.b	$9B,$5F,$EA,$F6,$A6,$CE,$1B,$F3,$1F,$C7,$A3,$F5,$B5,$CF,$E0,$FF
	.dc.b	$90,$FD,$53,$E3,$A9,$31,$FD,$BC,$7F,$1A,$7F,$D7,$CE,$DC,$C6,$DF
	.dc.b	$F7,$CC,$6D,$3F,$E6,$47,$9F,$D2,$78,$0F,$0F,$B9,$23,$24,$39,$BF
	.dc.b	$CD,$B3,$7C,$ED,$24,$F9,$B8,$6B,$F2,$E2,$ED,$FD,$5B,$15,$63,$4F
	.dc.b	$F0,$27,$D5,$7F,$5A,$AF,$EB,$6D,$8F,$A1,$7F,$D1,$A1,$F5,$54,$98
	.dc.b	$FF,$1C,$86,$F8,$86,$1D,$2B,$0F,$69,$E9,$F2,$BD,$7E,$56,$5F,$38
	.dc.b	$CB,$D6,$32,$F4,$CB,$D3,$3F,$D6,$BE,$AB,$F8,$54,$FF,$2E,$6B,$B3
	.dc.b	$FD,$EC,$E3,$0F,$EF,$3F,$55,$FC,$1A,$74,$27,$57,$FE,$29,$6C,$72
	.dc.b	$F1,$BF,$B4,$D7,$1B,$FF,$01,$D6,$A9,$3F,$61,$FE,$82,$D7,$BC,$6E
	.dc.b	$1F,$0F,$D3,$E5,$65,$70,$8E,$E6,$1F,$3C,$A7,$31,$DB,$C1,$F2,$91
	.dc.b	$F0,$F4,$57,$8C,$5C,$7E,$1C,$FF,$81,$3E,$A5,$F5,$54,$97,$FD,$6B
	.dc.b	$EA,$BF,$C0,$2F,$F5,$C9,$DB,$67,$6F,$CC,$9D,$BF,$3A,$76,$E7,$DA
	.dc.b	$FF,$09,$CF,$EA,$67,$38,$7F,$8B,$A3,$F5,$55,$CF,$A7,$67,$0E,$AB
	.dc.b	$1E,$58,$AF,$86,$3A,$FD,$54,$8F,$53,$FE,$22,$FA,$AF,$D0,$40,$5A
	.dc.b	$78,$47,$AD,$E1,$C7,$FE,$2A,$FA,$A7,$8F,$49,$5F,$E0,$DF,$84,$7C
	.dc.b	$A3,$A3,$33,$F9,$0B,$7C,$BF,$33,$3B,$FF,$93,$BE,$AB,$F8,$CD,$79
	.dc.b	$5A,$9B,$5F,$D3,$67,$87,$27,$F8,$FA,$76,$6A,$B1,$F2,$71,$FF,$0A
	.dc.b	$C2,$3F,$C9,$79,$FA,$35,$D5,$7E,$9A,$BF,$20,$F0,$E7,$CE,$3E,$7F
	.dc.b	$4F,$C0,$49,$EB,$F4,$F5,$FF,$5A,$F8,$5F,$E9,$5C,$3F,$49,$F1,$5F
	.dc.b	$A5,$6C,$66,$FF,$D9,$8E,$CB,$5F,$0D,$AF,$C3,$6C,$FF,$9A,$AB,$A4
	.dc.b	$60,$93,$FE,$E7,$7B,$34,$FC,$17,$D4,$FF,$63,$9F,$DC,$36,$7F,$D5
	.dc.b	$8D,$9D,$FB,$5E,$28,$C4,$17,$D5,$3E,$4F,$B4,$E6,$30,$FC,$43,$C3
	.dc.b	$F3,$1C,$E3,$38,$F8,$9B,$8F,$89,$B5,$C0,$EB,$87,$83,$0B,$3A,$9F
	.dc.b	$04,$6B,$0E,$CE,$89,$1E,$33,$D5,$1A,$71,$5D,$5B,$5D,$63,$E0,$99
	.dc.b	$C3,$D1,$7D,$1F,$97,$7C,$DC,$DF,$29,$F2,$3B,$C3,$DD,$16,$3D,$C6
	.dc.b	$C9,$3C,$8E,$AE,$CF,$79,$A9,$FB,$35,$E1,$7A,$C6,$4E,$B3,$CE,$32
	.dc.b	$D8,$CC,$60,$8F,$E2,$30,$F8,$55,$D6,$31,$7E,$3C,$3E,$2D,$7C,$72
	.dc.b	$47,$C0,$9F,$61,$FE,$47,$9F,$9C,$DF,$2F,$F8,$3B,$FE,$4B,$3D,$A6
	.dc.b	$79,$7F,$03,$1A,$CB,$FA,$87,$47,$F9,$EF,$8B,$35,$2F,$ED,$DE,$F4
	.dc.b	$FE,$45,$1F,$C7,$3D,$C5,$EF,$1F,$D1,$DE,$EB,$F4,$ED,$78,$4E,$EF
	.dc.b	$F4,$F3,$9C,$BF,$C7,$18,$FE,$8B,$2F,$26,$1E,$28,$E0,$F4,$75,$97
	.dc.b	$8A,$34,$4C,$BC,$73,$AC,$B8,$CF,$EC,$71,$F0,$1E,$5C,$3B,$DF,$09
	.dc.b	$E5,$97,$9B,$0E,$51,$83,$C4,$53,$DA,$5C,$BD,$6E,$6F,$AD,$73,$8D
	.dc.b	$61,$E1,$CB,$D5,$8F,$84,$FA,$AE,$2B,$84,$63,$2F,$5E,$1D,$30,$F1
	.dc.b	$CE,$5F,$07,$9B,$C5,$FE,$B5,$2F,$07,$11,$FD,$5A,$6C,$E6,$FE,$94
	.dc.b	$60,$CF,$C6,$58,$FA,$CD,$AF,$E0,$94,$8C,$61,$F8,$E6,$E3,$F8,$A9
	.dc.b	$D5,$F6,$3F,$90,$FD,$57,$FB,$6D,$FF,$72,$AF,$17,$EF,$73,$FD,$FC
	.dc.b	$93,$B7,$FC,$6E,$EE,$B8,$FC,$27,$07,$F3,$31,$99,$DC,$E5,$FE,$FD
	.dc.b	$19,$AE,$CF,$F8,$CA,$BF,$BC,$7D,$37,$FB,$DD,$F2,$F9,$3C,$7B,$13
	.dc.b	$9C,$3C,$73,$D5,$F3,$B6,$DF,$01,$63,$1E,$F6,$7B,$6E,$5F,$09,$B8
	.dc.b	$7B,$9B,$9F,$99,$F5,$2F,$A7,$F5,$57,$F8,$66,$31,$B6,$A5,$E0,$D8
	.dc.b	$95,$B3,$84,$F2,$C3,$FC,$F2,$F0,$8D,$46,$27,$C3,$F4,$AF,$AA,$A7
	.dc.b	$93,$E2,$3F,$AF,$5F,$0C,$95,$B8,$C3,$70,$FF,$3B,$4E,$AF,$97,$6F
	.dc.b	$4B,$F0,$D3,$EA,$9F,$49,$F0,$BE,$9D,$F1,$DF,$46,$73,$7E,$19,$CB
	.dc.b	$CA,$33,$7E,$09,$CB,$FE,$82,$8C,$E5,$F9,$29,$CB,$F9,$38,$CD,$7F
	.dc.b	$26,$6D,$7C,$8F,$73,$C5,$8F,$C6,$F1,$25,$FC,$0F,$16,$DF,$C0,$D9
	.dc.b	$58,$CB,$F4,$30,$AC,$5F,$D6,$FA,$AF,$B9,$D7,$F7,$5B,$1F,$3F,$8D
	.dc.b	$18,$CB,$E6,$C6,$27,$55,$F9,$7F,$A8,$7C,$DF,$E3,$3B,$1F,$53,$98
	.dc.b	$C3,$6F,$EE,$4E,$9F,$AB,$5C,$9C,$CE,$F1,$FF,$56,$CE,$5F,$81,$F5
	.dc.b	$59,$DB,$F7,$F9,$FF,$94,$70,$F5,$61,$F9,$9F,$3B,$F3,$4D,$8C,$9E
	.dc.b	$F8,$CB,$F8,$CF,$43,$F2,$A7,$E5,$BD,$CB,$19,$26,$5F,$46,$32,$F9
	.dc.b	$BC,$7F,$C6,$79,$3F,$15,$B5,$EE,$6C,$EE,$3F,$1D,$3B,$74,$7D,$3F
	.dc.b	$88,$C4,$FD,$FF,$90,$FD,$4F,$C6,$FE,$A9,$B5,$FA,$46,$CE,$E3,$F4
	.dc.b	$F3,$FB,$BF,$12,$47,$DF,$F4,$ED,$7D,$E9,$BC,$1F,$32,$6D,$7F,$7C
	.dc.b	$FA,$A6,$5F,$BA,$FA,$2F,$E3,$11,$89,$FD,$AC,$62,$75,$1F,$B2,$9D
	.dc.b	$63,$F9,$9A,$D4,$E6,$BF,$82,$F6,$62,$BF,$DD,$BF,$52,$FA,$AF,$E3
	.dc.b	$3D,$EF,$77,$F9,$44,$E7,$1F,$E1,$13,$97,$F8,$0F,$AB,$FE,$06,$DB
	.dc.b	$FA,$2C,$FF,$4C,$C7,$FD,$F3,$39,$AF,$E4,$53,$9A,$FE,$95,$F5,$2F
	.dc.b	$AA,$FE,$53,$F6,$1F,$EE,$9B,$FE,$1B,$39,$7F,$8C,$46,$71,$5F,$2C
	.dc.b	$3F,$98,$FE,$2B,$FE,$C7,$BE,$31,$8A,$FE,$85,$83,$CF,$FE,$6F,$F6
	.dc.b	$2F,$91,$CC,$61,$FE,$91,$EC,$3A,$3F,$9E,$FD,$8F,$F0,$4C,$FF,$90
	.dc.b	$CE,$5F,$E5,$91,$9C,$BC,$B1,$FE,$38,$9F,$EA,$9A,$FF,$6A,$FB,$17
	.dc.b	$A9,$DB,$E7,$7F,$A4,$C6,$2F,$F9,$6F,$C1,$BE,$AF,$FA,$6F,$84,$3F
	.dc.b	$AE,$46,$67,$F9,$E4,$65,$E3,$ED,$F0,$F8,$1D,$B6,$3A,$EB,$FD,$63
	.dc.b	$EC,$7D,$4F,$A0,$FF,$78,$61,$F2,$DF,$E8,$71,$AD,$7F,$C0,$55,$F2
	.dc.b	$4F,$E9,$90,$F9,$AF,$EE,$11,$9A,$E3,$1F,$DC,$FE,$C3,$E3,$5F,$F8
	.dc.b	$33,$D0,$FE,$77,$7F,$E5,$E9,$CB,$FD,$A6,$33,$61,$F6,$0F,$6F,$FB
	.dc.b	$F7,$2B,$E8,$7F,$C0,$13,$9C,$3F,$AB,$CE,$5F,$EB,$30,$FA,$7D,$FD
	.dc.b	$1F,$06,$CF,$F5,$EF,$B0,$FE,$45,$97,$C5,$3C,$33,$FE,$0D,$8C,$EB
	.dc.b	$FE,$25,$C3,$A5,$F1,$F2,$FF,$C2,$5F,$60,$F3,$78,$46,$1F,$F0,$B7
	.dc.b	$BF,$FD,$36,$B8,$1F,$DB,$9C,$67,$E1,$CB,$FC,$48,$F8,$E3,$39,$7F
	.dc.b	$8D,$5B,$5F,$E8,$8C,$5E,$67,$6F,$6B,$FC,$49,$F5,$DF,$89,$C7,$FB
	.dc.b	$FF,$D8,$3E,$4F,$3D,$BF,$E1,$C8,$C6,$5F,$E0,$0B,$E1,$A7,$F9,$12
	.dc.b	$73,$87,$C5,$9C,$D7,$F8,$86,$73,$5F,$E4,$99,$CD,$7D,$B6,$C2,$7E
	.dc.b	$FC,$A7,$6D,$E2,$FF,$17,$FD,$87,$F5,$9B,$E9,$18,$F7,$7D,$AD,$C7
	.dc.b	$D0,$DB,$F3,$7C,$5F,$F8,$97,$E4,$3E,$1B,$6B,$DC,$E1,$F8,$13,$4F
	.dc.b	$C1,$60,$AF,$F2,$73,$FB,$98,$CC,$63,$C4,$CB,$2B,$FC,$53,$3F,$9B
	.dc.b	$F1,$7F,$39,$F5,$FF,$E8,$58,$FF,$24,$57,$F9,$65,$D0,$8D,$3F,$C8
	.dc.b	$A3,$FD,$65,$18,$8D,$BA,$C3,$CD,$7D,$BE,$3B,$FD,$AD,$85,$BE,$F3
	.dc.b	$5F,$E3,$07,$79,$FF,$68,$D7,$28,$C1,$FE,$D4,$15,$C6,$35,$7E,$39
	.dc.b	$C9,$C0,$F4,$5F,$77,$CF,$7D,$B0,$F5,$BE,$68,$E1,$7D,$23,$17,$ED
	.dc.b	$AD,$D7,$8F,$0F,$14,$62,$3D,$67,$FB,$D4,$F1,$96,$30,$FC,$E3,$D9
	.dc.b	$F5,$FF,$D1,$7E,$3B,$83,$63,$FD,$FC,$D7,$7E,$5A,$77,$F1,$BF,$D5
	.dc.b	$A7,$6D,$7A,$B5,$FF,$7E,$D7,$FA,$53,$3E,$27,$D3,$CF,$FD,$79,$37
	.dc.b	$6F,$E4,$4F,$CB,$AE,$0D,$BF,$06,$9F,$86,$BE,$33,$F4,$5B,$1F,$8A
	.dc.b	$AE,$2D,$AE,$33,$D8,$D9,$6F,$C3,$3B,$C3,$F1,$17,$CA,$B5,$87,$9C
	.dc.b	$CE,$1D,$1F,$E7,$53,$B7,$F9,$E4,$F6,$6E,$1D,$A3,$E2,$D7,$19,$FF
	.dc.b	$75,$4F,$E5,$35,$EF,$1F,$CD,$6B,$DF,$87,$FB,$5F,$0E,$D3,$C0,$FF
	.dc.b	$4C,$CE,$AB,$8D,$6F,$0E,$DF,$5F,$FB,$DB,$FF,$4F,$B5,$FE,$84,$D9
	.dc.b	$FF,$72,$3F,$EC,$C8,$DE,$DE,$57,$FA,$C4,$6E,$FE,$CB,$63,$37,$B5
	.dc.b	$F5,$FF,$E8,$F7,$FD,$5E,$D7,$FA,$93,$70,$FF,$7A,$7D,$77,$F5,$B9
	.dc.b	$DF,$4F,$38,$FE,$05,$8F,$E2,$1B,$3F,$DA,$EF,$94,$FE,$31,$FF,$10
	.dc.b	$47,$83,$C2,$F2,$46,$23,$FB,$15,$3E,$BF,$35,$27,$81,$FE,$C6,$9D
	.dc.b	$38,$38,$F2,$F9,$4E,$0F,$4F,$8E,$F8,$5F,$5E,$9A,$7A,$E2,$CF,$FB
	.dc.b	$4A,$74,$D9,$D3,$FE,$D6,$F4,$7F,$DD,$ED,$F9,$2F,$9D,$18,$8F,$EF
	.dc.b	$95,$C9,$FE,$FD,$0F,$2B,$64,$8F,$67,$95,$CB,$CA,$91,$DD,$D3,$FE
	.dc.b	$07,$FA,$FF,$F0,$03,$FE,$E8,$6B,$9F,$23,$F0,$11,$F4,$A3,$FC,$3D
	.dc.b	$F5,$EF,$85,$FC,$23,$63,$FD,$E2,$D7,$FC,$29,$D1,$FE,$31,$7F,$DE
	.dc.b	$BF,$5F,$FE,$6D,$6F,$BD,$F8,$D8,$C4,$70,$FA,$3C,$75,$48,$E2,$FE
	.dc.b	$4E,$FC,$90,$AF,$F3,$53,$DB,$DA,$E8,$7E,$67,$EB,$DF,$61,$F9,$B8
	.dc.b	$FC,$47,$6B,$B3,$F8,$CC,$61,$BD,$5C,$7C,$4E,$1E,$AC,$CB,$FC,$81
	.dc.b	$87,$17,$FC,$85,$87,$BD,$FF,$1A,$64,$F9,$8F,$D5,$7D,$87,$F9,$A6
	.dc.b	$78,$E7,$FC,$A2,$31,$1F,$E3,$B6,$E1,$D6,$1F,$5F,$FD,$76,$31,$2F
	.dc.b	$6F,$A9,$2B,$A9,$FE,$55,$AF,$CF,$BF,$E4,$7F,$B0,$FF,$5E,$BA,$0E
	.dc.b	$1D,$7C,$5F,$F0,$19,$FF,$2E,$46,$E7,$FD,$2C,$DD,$BF,$83,$46,$23
	.dc.b	$FC,$A9,$2F,$AC,$FE,$15,$F6,$0F,$9A,$A4,$AF,$17,$D7,$BE,$C3,$64
	.dc.b	$8F,$35,$65,$DF,$27,$F1,$5F,$A2,$7C,$55,$24,$FB,$7E,$B7,$2F,$BE
	.dc.b	$6B,$FE,$69,$6F,$17,$CF,$6F,$1F,$F5,$CA,$B5,$FE,$7F,$9A,$A3,$F7
	.dc.b	$15,$F5,$DB,$5F,$D9,$2B,$7F,$60,$F9,$6F,$E5,$8D,$C3,$E9,$35,$FF
	.dc.b	$3E,$3F,$C9,$13,$9C,$3F,$08,$DB,$FE,$E2,$D9,$FF,$4A,$3C,$4F,$F3
	.dc.b	$53,$DB,$E4,$BC,$27,$8B,$DE,$91,$92,$1E,$27,$E7,$7D,$24,$8F,$94
	.dc.b	$FA,$B8,$B1,$1F,$2E,$2E,$3E,$69,$B8,$FF,$81,$BE,$C3,$F3,$93,$F9
	.dc.b	$D6,$C7,$C7,$AF,$EA,$10,$FB,$0A,$4C,$3F,$31,$1F,$45,$B1,$97,$FD
	.dc.b	$49,$1C,$5D,$ED,$D3,$E6,$BA,$FC,$74,$9E,$71,$97,$AC,$65,$E7,$1F
	.dc.b	$D8,$FE,$C1,$E3,$52,$4F,$EA,$FE,$C3,$F8,$7B,$FA,$D9,$DB,$67,$6F
	.dc.b	$FA,$E6,$76,$FF,$AE,$E7,$6E,$73,$FE,$E5,$27,$8E,$BF,$C1,$61,$F1
	.dc.b	$5A,$F9,$C9,$7E,$38,$E2,$61,$F0,$46,$1F,$47,$85,$D8,$8F,$AF,$D4
	.dc.b	$AF,$9E,$2E,$1E,$19,$FF,$03,$7D,$7B,$EC,$29,$2F,$F8,$77,$D7,$FF
	.dc.b	$80,$DF,$F6,$C4,$ED,$FE,$24,$D7,$33,$FE,$11,$9E,$17,$FC,$8E,$B4
	.dc.b	$FF,$25,$8F,$D3,$1F,$A7,$8C,$F2,$7F,$95,$23,$F4,$37,$FE,$E7,$6D
	.dc.b	$71,$E6,$ED,$CF,$DD,$95,$E8,$72,$FA,$BE,$BF,$C8,$47,$F9,$57,$D8
	.dc.b	$7E,$69,$0C,$13,$B8,$F1,$3F,$E2,$D9,$E1,$0F,$B0,$FE,$78,$FF,$35
	.dc.b	$6B,$FE,$FF,$6E,$3F,$D4,$7E,$19,$F2,$CE,$7F,$F1,$C5,$6B,$5F,$F3
	.dc.b	$AA,$F9,$EC,$9A,$F2,$FF,$C7,$DF,$61,$FC,$63,$3F,$4B,$53,$6B,$FA
	.dc.b	$95,$FF,$A0,$23,$44,$E3,$F0,$D7,$4F,$A4,$8F,$F4,$F9,$47,$F9,$4F
	.dc.b	$4F,$4D,$EA,$9A,$FF,$61,$B7,$37,$98,$D4,$67,$D4,$F4,$92,$BB,$7C
	.dc.b	$65,$7F,$D8,$7D,$7E,$73,$FD,$2F,$F1,$9F,$AB,$8F,$7C,$60,$F6,$55
	.dc.b	$8C,$9F,$EE,$58,$E4,$D3,$27,$63,$E4,$C7,$17,$E4,$C5,$C7,$C1,$1B
	.dc.b	$32,$73,$3E,$7C,$F1,$FB,$0F,$ED,$11,$96,$BC,$5B,$7D,$1F,$F3,$A3
	.dc.b	$7C,$DE,$5F,$36,$F9,$6E,$4D,$BE,$53,$A9,$C9,$23,$25,$AE,$D5,$C5
	.dc.b	$B3,$E5,$AE,$55,$E5,$AD,$65,$FE,$C6,$C1,$18,$9F,$D3,$C7,$AA,$75
	.dc.b	$39,$94,$F0,$BF,$15,$73,$BE,$F8,$78,$AF,$C1,$5E,$37,$67,$6C,$3A
	.dc.b	$61,$CE,$3D,$A7,$02,$57,$33,$E5,$3C,$8F,$1C,$F2,$39,$DF,$8B,$0F
	.dc.b	$4E,$1B,$C3,$D9,$7E,$18,$C5,$F0,$8B,$1B,$BE,$51,$B3,$C7,$7E,$F7
	.dc.b	$66,$0F,$93,$1C,$1F,$2C,$F3,$77,$87,$AE,$B9,$47,$BC,$FF,$7C,$9C
	.dc.b	$0A,$7D,$0F,$2B,$F8,$43,$8C,$7E,$43,$C1,$46,$AB,$F2,$4D,$AF,$E4
	.dc.b	$75,$FE,$57,$F7,$7F,$CD,$15,$AE,$6F,$CE,$5E,$EB,$F0,$6D,$78,$70
	.dc.b	$EA,$64,$7E,$86,$75,$CD,$FE,$95,$AC,$DF,$5A,$D5,$FE,$B6,$B5,$18
	.dc.b	$24,$EC,$FD,$AC,$69,$C6,$3F,$99,$C3,$C9,$97,$C0,$C3,$A6,$5E,$CA
	.dc.b	$D4,$78,$9B,$8F,$EF,$F1,$E9,$18,$C7,$AC,$FC,$0A,$E7,$97,$18,$C1
	.dc.b	$E6,$AF,$97,$7C,$A3,$37,$C9,$FB,$17,$C9,$E0,$7F,$54,$9E,$0E,$8F
	.dc.b	$26,$1E,$97,$9D,$F6,$8C,$46,$EF,$E2,$C6,$31,$F8,$53,$ED,$F2,$F7
	.dc.b	$83,$E3,$2C,$61,$B9,$FD,$38,$DF,$C7,$62,$FE,$33,$72,$FA,$6D,$CF
	.dc.b	$E0,$B6,$7D,$8D,$5F,$61,$FE,$E7,$C3,$FD,$3C,$FF,$AC,$DB,$F0,$BF
	.dc.b	$EB,$9E,$6F,$F5,$74,$E6,$BF,$A8,$D7,$E6,$23,$94,$60,$FE,$FA,$EE
	.dc.b	$78,$D7,$3A,$E1,$E2,$DB,$DF,$8D,$7A,$9F,$56,$F8,$7D,$57,$A2,$B5
	.dc.b	$39,$C3,$D3,$1A,$CF,$FC,$7F,$1E,$CF,$81,$C4,$E7,$DE,$D7,$6D,$AF
	.dc.b	$82,$DC,$7D,$47,$C6,$71,$B7,$C2,$7D,$0B,$EA,$7D,$D5,$FE,$E0,$6D
	.dc.b	$7F,$76,$5E,$0F,$07,$35,$C7,$7C,$19,$9B,$7F,$ED,$69,$7D,$6B,$EC
	.dc.b	$3C,$2D,$9D,$7C,$57,$F5,$97,$84,$67,$2F,$A9,$7C,$AF,$89,$EA,$BE
	.dc.b	$13,$AA,$E1,$5F,$EE,$DF,$AF,$7D,$57,$C0,$FB,$07,$C8,$7D,$49,$CE
	.dc.b	$7E,$79,$CB,$FE,$E9,$8C,$E3,$E2,$9C,$BF,$3A,$33,$87,$E3,$E7,$2F
	.dc.b	$87,$0E,$C7,$C9,$F7,$2E,$1E,$EF,$13,$E1,$9F,$13,$E5,$2C,$64,$99
	.dc.b	$7B,$63,$2F,$C4,$E1,$7D,$87,$D5,$EB,$FA,$CD,$BF,$A5,$3A,$6C,$E9
	.dc.b	$A7,$F9,$73,$33,$FC,$F3,$EC,$1F,$35,$F8,$8E,$B7,$D7,$FE,$72,$73
	.dc.b	$5F,$7E,$72,$F3,$39,$57,$1A,$E0,$D8,$D6,$7F,$E6,$CB,$E6,$FE,$69
	.dc.b	$B8,$7E,$EF,$8D,$F6,$1C,$AF,$DB,$E7,$FE,$57,$2F,$CB,$36,$32,$7C
	.dc.b	$28,$CB,$4F,$C7,$FA,$3F,$94,$3F,$2B,$EE,$D8,$C9,$33,$FC,$74,$65
	.dc.b	$F3,$78,$DF,$8D,$F2,$BE,$23,$6B,$DB,$2D,$53,$98,$E3,$3F,$91,$AE
	.dc.b	$2F,$0A,$DB,$F6,$BE,$2B,$35,$F9,$FF,$93,$FD,$5F,$C2,$FE,$B1,$B5
	.dc.b	$FA,$A6,$CE,$E3,$F5,$B3,$FB,$BF,$0E,$47,$E1,$BD,$4B,$5F,$8D,$9B
	.dc.b	$C5,$FC,$AF,$9B,$FC,$EF,$7F,$8C,$9F,$E3,$BF,$60,$CF,$F7,$BF,$45
	.dc.b	$86,$D7,$C2,$9D,$3F,$BE,$9D,$34,$D4,$E3,$1F,$DE,$76,$DC,$BF,$2B
	.dc.b	$1F,$EB,$2F,$AF,$7D,$87,$F1,$CE,$F7,$BF,$FC,$C2,$73,$97,$F1,$39
	.dc.b	$CB,$F5,$3D,$4F,$F0,$3E,$3F,$D1,$E5,$E7,$FF,$38,$9C,$D7,$E9,$67
	.dc.b	$2F,$DC,$FA,$E7,$CB,$7F,$2C,$FB,$0F,$8B,$AF,$F3,$29,$CD,$7F,$23
	.dc.b	$9C,$E5,$EA,$A5,$30,$FE,$6B,$F6,$1F,$EC,$EC,$3B,$46,$2B,$FA,$5D
	.dc.b	$3C,$FF,$E7,$7F,$60,$F9,$4C,$C6,$1F,$EA,$3E,$BB,$AB,$FA,$0F,$D8
	.dc.b	$3C,$BC,$C6,$1D,$FA,$DF,$D1,$BE,$C1,$EA,$E3,$7F,$E7,$91,$87,$F3
	.dc.b	$BF,$06,$FA,$AF,$E9,$BE,$30,$FE,$C3,$19,$9F,$E8,$91,$97,$8F,$BB
	.dc.b	$C3,$E0,$76,$D8,$EB,$C7,$FD,$63,$EC,$3D,$51,$83,$FA,$73,$88,$FE
	.dc.b	$C3,$1F,$0C,$CF,$CD,$FF,$72,$71,$1F,$D3,$25,$3F,$DD,$BE,$C3,$E2
	.dc.b	$57,$F5,$9F,$4F,$F9,$DB,$FE,$5E,$8C,$DF,$F5,$49,$CA,$1F,$5E,$F6
	.dc.b	$BF,$C0,$3C,$6F,$A2,$FF,$01,$CE,$71,$FE,$0B,$39,$7F,$83,$C3,$EC
	.dc.b	$37,$F4,$BC,$1B,$3F,$C3,$3E,$C3,$FD,$DE,$F0,$8F,$F0,$7C,$63,$2F
	.dc.b	$F0,$1D,$F3,$AE,$1E,$67,$F8,$53,$EC,$1E,$87,$08,$C3,$FE,$18,$F8
	.dc.b	$A7,$BB,$9B,$FE,$37,$18,$79,$98,$8E,$07,$5F,$3E,$C7,$F2,$2F,$AF
	.dc.b	$FC,$16,$7F,$DF,$3E,$C1,$F2,$7F,$E2,$69,$CF,$93,$FE,$39,$9C,$DF
	.dc.b	$1E,$DF,$F0,$CC,$63,$0F,$F0,$FD,$F0,$AF,$EB,$53,$9B,$FD,$74,$E6
	.dc.b	$BE,$B3,$61,$3F,$6A,$53,$B6,$F1,$7F,$8A,$3E,$C3,$FA,$9E,$1F,$EA
	.dc.b	$0A,$4E,$7D,$FF,$53,$67,$E0,$36,$FC,$FE,$7F,$F8,$67,$E3,$FE,$03
	.dc.b	$6B,$E3,$38,$7E,$44,$D3,$FC,$7B,$9A,$BF,$C9,$AF,$EC,$E3,$31,$8F
	.dc.b	$0F,$2C,$AF,$EA,$73,$F9,$7F,$1F,$F3,$1F,$61,$FE,$7B,$8F,$F0,$15
	.dc.b	$7F,$96,$1D,$08,$D2,$C3,$C5,$7C,$BE,$3B,$FD,$9B,$85,$AE,$73,$5F
	.dc.b	$F6,$3C,$5A,$E3,$1D,$6B,$8B,$4E,$25,$38,$96,$38,$DF,$92,$3C,$EF
	.dc.b	$9A,$33,$3D,$23,$18,$7F,$4A,$78,$52,$FC,$93,$9A,$E0,$7F,$B7,$AF
	.dc.b	$9D,$7F,$BB,$2B,$78,$7B,$63,$E3,$9F,$EE,$E3,$A1,$4A,$66,$35,$F6
	.dc.b	$1F,$CE,$7E,$3A,$F9,$86,$92,$BF,$DF,$32,$F8,$DF,$A5,$18,$6B,$FF
	.dc.b	$00,$23,$58,$F6,$9C,$CF,$FB,$E2,$72,$FC,$DC,$BF,$B0,$BF,$5D,$AF
	.dc.b	$D1,$8D,$E9,$F8,$26,$CF,$F2,$19,$FF,$62,$6D,$6B,$F0,$F3,$C9,$E1
	.dc.b	$7E,$86,$E1,$DE,$B8,$57,$33,$81,$E7,$AE,$0D,$9E,$17,$0F,$34,$E9
	.dc.b	$FE,$69,$87,$99,$FE,$6D,$5D,$1B,$7E,$07,$75,$C5,$F5,$9F,$E3,$A9
	.dc.b	$F8,$77,$FD,$E5,$BA,$78,$E7,$E5,$5F,$E1,$E3,$D4,$D8,$EE,$7F,$91
	.dc.b	$30,$F2,$2F,$B0,$FD,$E3,$FE,$48,$6B,$FE,$BC,$6E,$1F,$E0,$97,$F0
	.dc.b	$51,$BD,$BB,$3C,$DB,$1C,$30,$FA,$CF,$09,$FF,$63,$FD,$87,$F3,$E7
	.dc.b	$FC,$AE,$D7,$3D,$3D,$3E,$BF,$FA,$A4,$6A,$31,$A4,$8C,$1C,$BC,$2F
	.dc.b	$1D,$FF,$84,$63,$07,$F9,$92,$DF,$61,$AA,$93,$A3,$FD,$B7,$3A,$71
	.dc.b	$A7,$F8,$A6,$3F,$98,$C7,$3E,$5E,$2F,$0F,$8E,$F8,$7F,$61,$A8,$FF
	.dc.b	$72,$B6,$F9,$4E,$AA,$57,$F9,$E2,$B5,$B7,$E7,$70,$FF,$64,$CF,$F9
	.dc.b	$F3,$D1,$FF,$72,$B7,$E4,$7E,$8C,$62,$3B,$E1,$8F,$2A,$47,$17,$E6
	.dc.b	$79,$5C,$3C,$A9,$3D,$5D,$BF,$EF,$CF,$B0,$FE,$EE,$FF,$A3,$1A,$E7
	.dc.b	$C6,$FC,$0C,$7D,$9D,$FF,$C9,$DF,$60,$F8,$8F,$C2,$B6,$FF,$D2,$AD
	.dc.b	$7F,$9E,$6F,$FB,$BE,$8F,$F1,$2C,$7F,$A6,$BE,$C3,$FC,$BE,$DF,$7F
	.dc.b	$F1,$F1,$88,$E2,$7F,$88,$E5,$F4,$7A,$EE,$93,$CD,$FC,$BC,$F9,$F0
	.dc.b	$FE,$15,$7F,$E5,$DA,$F0,$7B,$7D,$8F,$CE,$FD,$7B,$EC,$3F,$3F,$1F
	.dc.b	$8D,$9F,$09,$FE,$65,$EA,$C9,$C3,$C4,$DF,$AB,$27,$FD,$6B,$5E,$E7
	.dc.b	$8E,$0F,$98,$FD,$5F,$D8,$7F,$97,$67,$FD,$76,$D7,$BB,$78,$FF,$CF
	.dc.b	$9F,$5F,$FC,$8E,$73,$ED,$72,$23,$FE,$C4,$8F,$CE,$3E,$8F,$B0,$FF
	.dc.b	$55,$3F,$EC,$86,$B9,$F8,$CF,$DE,$CF,$F9,$36,$35,$AF,$F0,$0A,$FF
	.dc.b	$41,$46,$0E,$DF,$59,$FC,$13,$EB,$BF,$D0,$11,$FE,$D4,$6B,$F2,$1B
	.dc.b	$97,$FA,$57,$EC,$1F,$61,$C4,$95,$C5,$FE,$1F,$1F,$46,$7F,$D3,$9F
	.dc.b	$47,$FE,$9B,$AF,$F7,$0B,$5F,$AA,$D8,$FF,$51,$7D,$6D,$E7,$FE,$95
	.dc.b	$36,$BF,$CD,$8D,$4D,$9C,$37,$C5,$FE,$43,$1F,$B1,$BF,$1F,$C5,$FF
	.dc.b	$24,$FB,$0F,$F7,$8E,$1F,$EF,$46,$BF,$88,$6C,$FF,$AD,$5F,$F0,$56
	.dc.b	$3F,$D7,$5C,$B7,$E6,$3A,$1D,$7C,$EE,$DE,$E7,$7F,$92,$98,$F7,$8C
	.dc.b	$BE,$18,$CB,$E1,$8F,$E7,$3A,$BE,$92,$92,$7D,$1C,$55,$F9,$71,$74
	.dc.b	$F4,$4D,$CF,$FB,$A7,$D7,$FF,$B3,$E7,$FD,$A0,$D8,$F9,$F5,$FD,$22
	.dc.b	$1F,$61,$49,$87,$FA,$76,$F8,$F0,$62,$3F,$5F,$5C,$1D,$61,$E0,$BE
	.dc.b	$95,$D7,$E7,$3B,$7C,$64,$BD,$9F,$EA,$9C,$13,$B8,$C8,$8F,$F6,$63
	.dc.b	$86,$BE,$28,$FE,$C3,$F6,$1F,$C4,$23,$8B,$5F,$E0,$AD,$CB,$FD,$E5
	.dc.b	$F6,$1F,$C2,$E7,$7E,$47,$22,$3E,$28,$D3,$CA,$3F,$B7,$C7,$D7,$8D
	.dc.b	$6B,$C1,$FE,$27,$0F,$87,$DF,$CE,$C9,$E9,$1C,$8C,$3C,$23,$0F,$87
	.dc.b	$C2,$EE,$47,$C9,$D3,$5E,$F1,$72,$F6,$CF,$F8,$13,$EB,$DF,$61,$49
	.dc.b	$7F,$3B,$EC,$3F,$BE,$BF,$CA,$A7,$6F,$D2,$6B,$99,$FF,$07,$CF,$0B
	.dc.b	$72,$7F,$90,$EF,$71,$8B,$FF,$73,$57,$EB,$1B,$D3,$E2,$7F,$9C,$C7
	.dc.b	$FB,$11,$F2,$74,$76,$3B,$F4,$D8,$C9,$0E,$2F,$93,$EC,$24,$7B,$5F
	.dc.b	$F1,$2F,$D8,$7E,$86,$3C,$71,$A7,$C2,$EB,$3F,$F1,$5F,$D8,$3C,$7A
	.dc.b	$4A,$FC,$CF,$C2,$BE,$5B,$F3,$31,$F9,$8A,$E0,$DF,$9E,$E6,$71,$7F
	.dc.b	$AB,$FD,$87,$F1,$7C,$7F,$CD,$4D,$80,$70,$AB,$FA,$77,$07,$CD,$8D
	.dc.b	$71,$FF,$90,$AB,$67,$0A,$DD,$F6,$FA,$7D,$12,$FD,$79,$7C,$57,$C1
	.dc.b	$1C,$F9,$A4,$7E,$D3,$04,$F7,$F5,$24,$E5,$F1,$7C,$5E,$23,$FD,$5D
	.dc.b	$D7,$EE,$3F,$D8,$65,$7F,$DA,$8F,$C2,$9C,$C7,$C2,$BF,$D7,$3C,$8C
	.dc.b	$4E,$9F,$8C,$D8,$FF,$34,$FC,$5F,$07,$8B,$63,$46,$EB,$97,$D8,$7F
	.dc.b	$64,$9E,$ED,$9F,$F4,$F3,$6F,$EA,$BF,$E6,$79,$DF,$CE,$72,$F9,$B7
	.dc.b	$CA,$FB,$2F,$71,$87,$D4,$2F,$A3,$63,$57,$E2,$6E,$1E,$26,$CF,$8B
	.dc.b	$0F,$06,$1D,$E3,$93,$70,$57,$09,$F0,$57,$07,$C7,$3A,$AE,$15,$B8
	.dc.b	$F2,$99,$9D,$C6,$4D,$57,$3C,$3D,$17,$C8,$F1,$46,$70,$F5,$46,$6F
	.dc.b	$C3,$39,$7E,$3C,$66,$BE,$6C,$5A,$E2,$F8,$A7,$AB,$93,$89,$93,$A9
	.dc.b	$93,$E2,$99,$9E,$AD,$9E,$B1,$EA,$8E,$26,$4E,$AE,$4E,$25,$8E,$24
	.dc.b	$8F,$AC,$78,$70,$F9,$53,$B8,$C9,$D4,$F2,$B9,$7E,$B9,$C7,$0F,$A1
	.dc.b	$7F,$02,$3D,$C7,$F9,$46,$BE,$B3,$7C,$97,$27,$F9,$56,$3E,$29,$97
	.dc.b	$F0,$D1,$A8,$FC,$AF,$37,$FA,$27,$8B,$35,$27,$F0,$F5,$BC,$BF,$6B
	.dc.b	$7B,$9C,$D6,$E7,$F5,$75,$BC,$FF,$8E,$D7,$F2,$DF,$7B,$FA,$35,$6A
	.dc.b	$F8,$57,$EC,$2F,$85,$FF,$24,$C3,$B5,$EE,$3B,$91,$F1,$E1,$D2,$F9
	.dc.b	$61,$D2,$BF,$73,$87,$47,$C5,$8E,$5F,$DA,$46,$72,$FF,$28,$D7,$EB
	.dc.b	$72,$F0,$E1,$D2,$FC,$0D,$CB,$C3,$8F,$3C,$38,$D6,$8F,$56,$5E,$57
	.dc.b	$C7,$8F,$58,$C4,$78,$A3,$15,$E5,$C7,$C3,$3C,$0F,$AF,$5C,$2F,$BE
	.dc.b	$1F,$53,$CB,$E7,$6E,$EB,$18,$6E,$3E,$D8,$DF,$C7,$63,$2F,$C1,$B7
	.dc.b	$0F,$BE,$DD,$3E,$DE,$28,$C5,$7D,$76,$AF,$B0,$7C,$2F,$F8,$17,$AF
	.dc.b	$FA,$D4,$7F,$AC,$73,$FF,$20,$BA,$6F,$89,$B7,$F1,$31,$9B,$FF,$60
	.dc.b	$BC,$5B,$7B,$7F,$B5,$C6,$6F,$BF,$8B,$9E,$E7,$D6,$3E,$23,$86,$1E
	.dc.b	$7A,$D4,$E7,$2F,$2C,$6B,$1F,$F1,$F4,$7A,$3E,$07,$13,$9F,$77,$5D
	.dc.b	$B6,$BD,$AD,$C7,$D2,$DD,$3E,$1B,$E9,$5F,$51,$E9,$AF,$F3,$A3,$69
	.dc.b	$3A,$F0,$F8,$4E,$63,$8F,$06,$0C,$CD,$BF,$F3,$9C,$BE,$B1,$F6,$1E
	.dc.b	$16,$CE,$BE,$2B,$FA,$EB,$C2,$33,$97,$8A,$3F,$D1,$11,$8B,$FB,$CD
	.dc.b	$8F,$F4,$65,$72,$AE,$07,$AA,$74,$FF,$88,$BE,$C1,$F5,$5F,$0F,$EC
	.dc.b	$1F,$23,$F5,$A7,$38,$FA,$A7,$2F,$FA,$2A,$33,$7D,$B2,$56,$A3,$C2
	.dc.b	$27,$73,$F9,$59,$D3,$FE,$2F,$8C,$D7,$F1,$A6,$D7,$D2,$D9,$EE,$79
	.dc.b	$72,$F9,$5E,$2C,$C3,$E4,$7C,$A5,$C3,$C9,$AA,$B1,$9F,$C8,$AB,$7F
	.dc.b	$80,$FB,$0F,$BB,$D7,$F7,$9B,$1F,$6B,$44,$63,$2F,$81,$18,$9D,$57
	.dc.b	$E7,$BE,$B9,$F3,$5F,$8F,$EB,$7D,$7E,$63,$0D,$BF,$C3,$CE,$9F,$F5
	.dc.b	$BD,$72,$73,$3B,$C7,$E6,$CE,$5F,$7F,$D8,$67,$5F,$E0,$59,$FF,$93
	.dc.b	$30,$F5,$61,$F9,$AF,$3B,$F3,$6D,$8C,$9E,$F8,$CB,$F9,$3F,$73,$F3
	.dc.b	$07,$E6,$7D,$EB,$19,$26,$5F,$42,$32,$F9,$BC,$7F,$C7,$79,$5F,$15
	.dc.b	$B5,$EE,$6C,$EE,$3F,$21,$3B,$74,$7D,$3F,$89,$C4,$FD,$EF,$93,$FD
	.dc.b	$67,$C7,$7E,$B5,$B5,$FA,$B6,$CE,$E3,$F5,$D3,$FB,$CF,$0E,$47,$DF
	.dc.b	$F5,$2D,$7D,$E9,$BC,$1F,$32,$6D,$7F,$70,$FB,$06,$5F,$BD,$FA,$2F
	.dc.b	$E3,$F1,$89,$FD,$E4,$62,$75,$1F,$EF,$49,$D6,$3F,$98,$AD,$4E,$6B
	.dc.b	$F8,$47,$66,$2B,$F5,$BF,$5E,$FB,$0F,$E3,$9D,$EF,$77,$F9,$7C,$E7
	.dc.b	$1F,$E2,$73,$97,$FD,$47,$AF,$FA,$7F,$D4,$FF,$07,$3F,$D2,$67,$FA
	.dc.b	$84,$E8,$FF,$06,$CE,$9E,$31,$AC,$7F,$7F,$F5,$CF,$96,$FE,$5D,$F6
	.dc.b	$0E,$1F,$E8,$33,$9C,$3F,$92,$CE,$5E,$D7,$FE,$B6,$9F,$DF,$C7,$F2
	.dc.b	$AF,$B0,$FF,$5C,$BA,$94,$E9,$FE,$25,$3A,$7C,$7E,$77,$F3,$FF,$B0
	.dc.b	$7C,$9F,$8A,$30,$FF,$54,$F6,$1D,$1F,$D1,$7E,$C3,$F8,$46,$7D,$E7
	.dc.b	$2F,$F3,$C8,$CE,$3F,$C7,$A7,$FA,$D3,$FC,$8F,$EC,$1E,$9F,$5F,$5B
	.dc.b	$FA,$84,$63,$0F,$57,$C1,$3E,$AB,$FA,$A7,$84,$3F,$B3,$C6,$67,$FA
	.dc.b	$5C,$65,$E3,$ED,$F0,$F8,$1D,$B6,$3A,$E7,$FD,$7F,$EC,$3D,$2F,$F3
	.dc.b	$4C,$3E,$37,$57,$F8,$1E,$72,$FF,$B8,$6B,$E3,$9F,$D6,$A1,$F3,$7F
	.dc.b	$DE,$A3,$37,$FD,$06,$73,$1F,$57,$EC,$3E,$2D,$FF,$84,$FD,$0F,$E7
	.dc.b	$F3,$FE,$56,$9C,$BF,$92,$8C,$DF,$FB,$DB,$EB,$9E,$CF,$F8,$23,$95
	.dc.b	$F4,$5F,$E0,$A9,$CE,$1F,$83,$9C,$BF,$85,$87,$D8,$6F,$E9,$38,$36
	.dc.b	$7F,$15,$F6,$1F,$C9,$72,$FE,$AF,$7C,$27,$3E,$87,$07,$7E,$67,$F8
	.dc.b	$6B,$EC,$1E,$77,$F8,$73,$1F,$F0,$BE,$3F,$D8,$27,$37,$F9,$7F,$8A
	.dc.b	$7B,$5B,$C7,$FC,$47,$18,$9E,$A6,$E7,$27,$7F,$3E,$C7,$E8,$7E,$BB
	.dc.b	$F0,$98,$FF,$71,$FB,$07,$C9,$7F,$8B,$A7,$3C,$FF,$E4,$39,$CD,$F1
	.dc.b	$EE,$FF,$11,$46,$30,$FE,$EF,$7C,$2B,$F8,$2C,$E6,$D1,$FE,$8E,$9D
	.dc.b	$50,$38,$95,$7D,$C9,$CD,$F9,$78,$FF,$C6,$5F,$61,$FD,$66,$FA,$D7
	.dc.b	$3C,$15,$AF,$77,$86,$5F,$E2,$CC,$38,$3E,$39,$D3,$F1,$7E,$2B,$FC
	.dc.b	$7D,$D2,$F8,$FF,$89,$18,$6C,$EC,$F9,$99,$60,$FF,$0C,$E7,$C6,$3F
	.dc.b	$88,$3F,$B9,$8C,$CF,$F9,$72,$30,$7A,$A3,$FC,$93,$19,$3F,$CC,$2F
	.dc.b	$C7,$F1,$BE,$47,$D8,$7F,$A2,$63,$F9,$9C,$E9,$4E,$BE,$3B,$FD,$7B
	.dc.b	$7E,$39,$5F,$17,$FD,$83,$5B,$79,$E1,$5C,$9C,$2F,$9E,$1F,$42,$F9
	.dc.b	$4F,$94,$D1,$C8,$A7,$27,$C1,$87,$8D,$E3,$1E,$83,$71,$9A,$FF,$73
	.dc.b	$61,$E3,$8C,$4F,$FB,$60,$E1,$5D,$2F,$C9,$7C,$27,$9C,$61,$E1,$7C
	.dc.b	$E3,$11,$EE,$3F,$DE,$87,$8C,$B1,$87,$DC,$78,$7E,$C3,$FA,$2F,$C7
	.dc.b	$70,$6C,$7F,$BF,$5A,$EF,$CC,$4E,$FE,$37,$FD,$0F,$3B,$6B,$D5,$AF
	.dc.b	$FB,$DA,$91,$88,$FD,$54,$63,$3F,$B1,$97,$FB,$06,$6E,$BF,$EF,$ED
	.dc.b	$3F,$01,$7B,$9F,$AA,$D8,$FC,$25,$71,$9F,$B2,$D8,$E3,$87,$A1,$B8
	.dc.b	$76,$A8,$74,$9D,$57,$A5,$B1,$FC,$06,$3A,$1F,$C0,$DF,$F7,$26,$1E
	.dc.b	$78,$E1,$5F,$93,$C3,$CD,$1B,$38,$D7,$3A,$E0,$FC,$4B,$EE,$F4,$3F
	.dc.b	$C0,$53,$ED,$BF,$D8,$37,$5E,$F3,$F0,$2F,$F5,$31,$DD,$B1,$DC,$FF
	.dc.b	$06,$61,$E4,$7F,$8E,$7D,$87,$F0,$97,$FC,$20,$D7,$FD,$3A,$DC,$3F
	.dc.b	$80,$3F,$72,$37,$B7,$65,$1B,$C3,$E3,$36,$3B,$E5,$BF,$B0,$FE,$8F
	.dc.b	$1F,$E1,$F6,$BF,$EA,$C6,$E1,$FC,$3F,$EB,$BF,$AF,$CE,$FA,$BC,$11
	.dc.b	$FB,$6C,$7E,$A3,$67,$F9,$C5,$F2,$9F,$B0,$FF,$88,$A3,$C1,$E1,$F8
	.dc.b	$CE,$7F,$61,$63,$24,$3F,$D7,$F6,$8D,$46,$39,$7C,$66,$32,$F4,$7C
	.dc.b	$77,$C5,$FB,$0F,$E7,$F5,$FE,$3D,$6B,$D2,$35,$52,$34,$FB,$A3,$E7
	.dc.b	$7A,$2F,$91,$A4,$AD,$9F,$E4,$98,$79,$5C,$09,$1F,$E4,$DC,$FF,$C2
	.dc.b	$2E,$FC,$AE,$A4,$79,$4F,$DD,$FB,$0F,$F0,$1C,$7F,$95,$5A,$FF,$B8
	.dc.b	$9B,$A7,$F4,$57,$E8,$47,$F6,$9F,$B0,$7C,$37,$E1,$DB,$1F,$E5,$F6
	.dc.b	$BF,$EE,$CE,$6F,$F1,$9B,$FE,$63,$FB,$0F,$F3,$83,$7D,$FF,$C8,$46
	.dc.b	$23,$87,$D1,$63,$A2,$47,$17,$F2,$D7,$E4,$85,$3D,$BE,$64,$8F,$F3
	.dc.b	$97,$D7,$BE,$C3,$F3,$D3,$F8,$99,$FF,$80,$2D,$DB,$FB,$23,$7A,$BF
	.dc.b	$83,$5F,$F9,$D2,$BF,$CF,$93,$EC,$78,$FA,$DD,$0F,$D5,$4B,$E6,$3F
	.dc.b	$59,$F6,$1F,$E6,$9A,$FF,$42,$35,$63,$FD,$CE,$30,$FF,$75,$FB,$0F
	.dc.b	$EA,$0E,$7D,$BE,$64,$AF,$F4,$74,$FE,$6D,$FF,$48,$7D,$87,$FA,$F9
	.dc.b	$BF,$1B,$FB,$D8,$C4,$78,$35,$FE,$45,$5F,$8A,$8C,$1F,$E2,$79,$F8
	.dc.b	$0E,$03,$8C,$3F,$C1,$4D,$AE,$7F,$59,$FC,$2F,$EC,$1F,$33,$49,$87
	.dc.b	$FA,$97,$EB,$DF,$61,$B2,$47,$FA,$A6,$B2,$EF,$C5,$FE,$2D,$F4,$6F
	.dc.b	$8E,$A4,$9F,$F5,$87,$D7,$65,$FF,$59,$4D,$7D,$4D,$DB,$FC,$44,$DD
	.dc.b	$7F,$A2,$55,$AF,$F2,$44,$D5,$1F,$B9,$7B,$F8,$9E,$17,$E3,$7D,$83
	.dc.b	$E5,$FF,$96,$36,$3F,$D8,$8D,$77,$E8,$6A,$64,$69,$E3,$1C,$BD,$BE
	.dc.b	$7F,$29,$2B,$9C,$65,$EB,$19,$7C,$71,$F2,$3E,$92,$47,$C4,$7C,$BC
	.dc.b	$98,$8F,$8B,$17,$6F,$F2,$36,$4A,$C6,$DF,$DA,$7E,$BF,$FC,$95,$5F
	.dc.b	$E4,$B6,$C7,$F6,$DD,$D8,$7F,$4D,$97,$D8,$52,$65,$FE,$E7,$8F,$9A
	.dc.b	$D8,$CB,$F9,$18,$E2,$EF,$83,$A7,$CD,$F5,$F8,$E9,$3C,$E3,$2F,$58
	.dc.b	$CB,$CE,$3F,$B0,$7D,$83,$C7,$A4,$9F,$F7,$D7,$D8,$7F,$09,$7F,$DF
	.dc.b	$93,$B6,$CE,$DF,$D3,$4E,$DF,$D4,$4E,$DC,$E7,$FD,$BE,$7E,$2C,$72
	.dc.b	$3B,$FC,$36,$7E,$62,$46,$08,$76,$72,$F8,$23,$A7,$87,$23,$D2,$F7
	.dc.b	$E8,$C4,$79,$A2,$F1,$7F,$A0,$F0,$FC,$E4,$FE,$FF,$EB,$DF,$61,$FE
	.dc.b	$05,$6C,$7F,$AA,$EB,$FD,$15,$18,$8F,$E0,$DF,$61,$FD,$F9,$E5,$5C
	.dc.b	$48,$F1,$E5,$FF,$19,$4E,$EF,$9D,$7F,$84,$E3,$FC,$05,$46,$DF,$7A
	.dc.b	$FF,$4D,$47,$FB,$32,$7A,$1F,$EB,$F7,$0F,$93,$A3,$A7,$4F,$56,$47
	.dc.b	$03,$B3,$A7,$12,$FA,$FF,$21,$23,$87,$D8,$7E,$81,$F0,$46,$5F,$86
	.dc.b	$FF,$8B,$63,$50,$FB,$07,$91,$FE,$31,$6C,$78,$DA,$FF,$34,$F8,$37
	.dc.b	$CB,$3A,$38,$5F,$1F,$9F,$97,$C6,$3F,$C7,$9F,$61,$F7,$37,$FF,$60
	.dc.b	$37,$8B,$FC,$C4,$DA,$FF,$73,$67,$FD,$42,$73,$CB,$C8,$93,$E5,$3A
	.dc.b	$FC,$A4,$7E,$55,$A3,$FC,$B3,$AF,$A2,$F5,$5F,$D2,$E9,$D1,$E6,$33
	.dc.b	$97,$9B,$D4,$F4,$92,$BA,$FC,$DD,$7F,$DA,$3D,$7E,$A3,$FD,$9A,$57
	.dc.b	$FD,$B4,$FB,$A7,$30,$18,$7E,$C9,$E6,$65,$C7,$C5,$65,$B5,$CE,$7A
	.dc.b	$9B,$7A,$BD,$1B,$1C,$23,$A5,$6B,$EC,$3F,$B2,$C7,$20,$E1,$F1,$98
	.dc.b	$F3,$B8,$7C,$37,$17,$97,$CD,$BE,$53,$95,$A3,$17,$CA,$3A,$67,$EF
	.dc.b	$75,$5C,$2B,$73,$96,$E7,$FE,$A7,$AE,$2D,$8C,$DF,$CB,$9E,$6F,$78
	.dc.b	$CB,$E9,$3B,$5C,$74,$FA,$30,$F5,$B9,$C3,$C4,$E6,$71,$87,$86,$2C
	.dc.b	$78,$6B,$8B,$D0,$95,$EF,$8C,$BD,$23,$35,$ED,$9C,$BD,$63,$27,$53
	.dc.b	$B9,$F5,$AF,$DB,$87,$C0,$BE,$D3,$A8,$C5,$73,$9C,$CF,$43,$96,$1B
	.dc.b	$C3,$C3,$1F,$2F,$0E,$A7,$68,$CC,$70,$39,$61,$F5,$1E,$15,$E1,$AE
	.dc.b	$11,$63,$B3,$4E,$35,$D2,$BC,$F1,$C8,$FC,$44,$7C,$C7,$FD,$D0,$78
	.dc.b	$4B,$0F,$21,$CD,$A9,$E4,$6A,$32,$FE,$4D,$B8,$FE,$C7,$9B,$FD,$1B
	.dc.b	$C5,$F9,$A9,$CE,$9F,$91,$6D,$FE,$3D,$AF,$F2,$AF,$79,$5C,$71,$93
	.dc.b	$DD,$FD,$76,$1D,$A3,$04,$C7,$F4,$98,$74,$9D,$46,$2B,$9B,$6B,$F6
	.dc.b	$B3,$AC,$7C,$D5,$A6,$CE,$B1,$F2,$BD,$E3,$37,$CC,$F8,$53,$A7,$11
	.dc.b	$A2,$4F,$F1,$9B,$F0,$3F,$0E,$FD,$B7,$C6,$B7,$3F,$17,$1F,$23,$C6
	.dc.b	$74,$F1,$3C,$D1,$A7,$FA,$AF,$9B,$A7,$FA,$BF,$9B,$71,$FF,$2F,$57
	.dc.b	$CC,$8C,$54,$8F,$9F,$F1,$76,$FE,$3B,$73,$FC,$03,$72,$FE,$69,$3F
	.dc.b	$D9,$23,$0F,$BD,$AB,$EC,$3F,$DC,$B7,$FD,$E5,$FF,$58,$37,$E0,$BE
	.dc.b	$7F,$3F,$D0,$AE,$1E,$3E,$2B,$BD,$F2,$9F,$F4,$F3,$C9,$B7,$B7,$FA
	.dc.b	$44,$66,$FA,$2F,$23,$5D,$6F,$AB,$7C,$37,$1C,$3D,$15,$A9,$CE,$5E
	.dc.b	$98,$D6,$1C,$E3,$D9,$F0,$38,$9C,$FB,$FA,$EE,$B5,$F0,$5B,$8F,$A9
	.dc.b	$BA,$79,$9F,$52,$FA,$9F,$55,$7F,$8E,$DB,$5F,$ED,$45,$E1,$70,$73
	.dc.b	$5C,$78,$30,$66,$6D,$FF,$8C,$A5,$F5,$AF,$B0,$F0,$B6,$75,$F2,$7F
	.dc.b	$D6,$5E,$11,$9C,$BF,$DA,$77,$CA,$F8,$9E,$AB,$E1,$3A,$AE,$15,$F7
	.dc.b	$FE,$BD,$F5,$5F,$03,$EC,$1F,$1D,$F4,$E7,$39,$F9,$E7,$2F,$F9,$0E
	.dc.b	$33,$8F,$8A,$72,$FF,$B8,$E3,$38,$7E,$3A,$72,$F8,$70,$EC,$7C,$9F
	.dc.b	$76,$E1,$EE,$F0,$FE,$19,$F1,$3C,$6B,$19,$26,$5E,$D8,$CB,$F1,$38
	.dc.b	$5F,$61,$F5,$7A,$FE,$B3,$6F,$E9,$4E,$9B,$3A,$69,$FE,$61,$FB,$07
	.dc.b	$CD,$7E,$1B,$AD,$F5,$FF,$99,$9C,$D7,$DD,$9C,$BC,$6B,$85,$6D,$B1
	.dc.b	$92,$65,$F5,$FE,$C3,$51,$E6,$F3,$FF,$1D,$7F,$8F,$6C,$64,$F5,$C6
	.dc.b	$5A,$7E,$23,$D1,$FC,$69,$F8,$EF,$76,$C6,$49,$87,$E1,$E3,$2B,$8F
	.dc.b	$F8,$36,$9F,$C7,$23,$F0,$1E,$4F,$C1,$F1,$FF,$14,$DC,$3F,$10,$D9
	.dc.b	$DC,$7C,$B9,$C5,$FF,$12,$E0,$7C,$4E,$2B,$FD,$55,$D4,$FA,$3F,$D2
	.dc.b	$FC,$2F,$E9,$9B,$3F,$A4,$6E,$5F,$9F,$6E,$3D,$7C,$3F,$DB,$1F,$B7
	.dc.b	$F5,$2D,$FE,$0A,$2D,$FE,$26,$6D,$7F,$25,$FB,$0F,$2D,$FE,$CB,$E8
	.dc.b	$BF,$8C,$46,$2F,$F5,$D1,$8D,$3F,$76,$DA,$E0,$7F,$AA,$EB,$85,$F5
	.dc.b	$AF,$E0,$3D,$BB,$C3,$E8,$5F,$F4,$0F,$AF,$7D,$87,$F1,$6E,$F7,$BF
	.dc.b	$FC,$A6,$73,$5F,$C3,$67,$2F,$FA,$E7,$D3,$FE,$F9,$9F,$F4,$09,$79
	.dc.b	$FF,$CC,$67,$35,$FE,$D9,$9C,$BF,$EC,$1F,$AF,$7C,$AF,$F2,$6F,$B0
	.dc.b	$FF,$4A,$CE,$63,$1C,$5F,$CE,$E9,$7F,$CB,$3E,$C3,$FD,$19,$7C,$E3
	.dc.b	$1B,$EF,$CF,$FE,$63,$F6,$0F,$A4,$C5,$FF,$27,$8C,$3F,$ED,$3F,$5D
	.dc.b	$D5,$FC,$E3,$EC,$1E,$5F,$F4,$B9,$CD,$7F,$2F,$9C,$BC,$7D,$6F,$E7
	.dc.b	$FF,$60,$F5,$78,$46,$1F,$EA,$9F,$06,$FA,$9F,$E8,$BE,$38,$CB,$87
	.dc.b	$7E,$EE,$BE,$06,$E5,$FC,$FE,$7F,$D2,$0F,$FB,$AE,$FE,$1F,$17,$F3
	.dc.b	$AF,$B0,$F3,$BF,$80,$3F,$D6,$51,$AE,$CF,$EE,$F3,$A7,$FB,$C5,$7C
	.dc.b	$4A,$E5,$F3,$7F,$DA,$E3,$13,$C2,$3F,$B7,$7D,$87,$C2,$AF,$E9,$3E
	.dc.b	$9F,$F3,$4B,$FF,$06,$CE,$6F,$FC,$0B,$39,$AF,$F6,$27,$D7,$BD,$AF
	.dc.b	$EF,$1C,$CF,$A2,$FE,$F5,$39,$CB,$FD,$95,$8F,$F8,$36,$B5,$18,$3F
	.dc.b	$C3,$72,$FB,$0E,$3F,$4B,$C9,$B3,$BC,$BF,$DA,$9F,$61,$FE,$E9,$8C
	.dc.b	$42,$F9,$46,$27,$53,$F2,$CD,$57,$F8,$3E,$3F,$C1,$51,$9F,$B0,$FE
	.dc.b	$1D,$5C,$7C,$EF,$F0,$A4,$62,$79,$7C,$53,$DD,$E5,$5F,$E1,$98,$C5
	.dc.b	$F1,$3C,$AE,$2B,$8D,$74,$F4,$38,$3F,$E2,$1F,$AF,$FC,$75,$FF,$81
	.dc.b	$7E,$C1,$F2,$79,$EC,$FF,$0A,$C6,$27,$E0,$D7,$0C,$7F,$C0,$73,$97
	.dc.b	$E0,$46,$78,$BF,$C7,$F3,$97,$FC,$81,$19,$AF,$F2,$6B,$6B,$FA,$05
	.dc.b	$7E,$05,$46,$5B,$C3,$FE,$28,$FB,$0F,$EC,$31,$E0,$8C,$7B,$FE,$D6
	.dc.b	$E1,$E8,$6C,$F9,$BC,$FF,$F0,$C7,$C7,$FC,$36,$D7,$B9,$C3,$F0,$26
	.dc.b	$9F,$77,$35,$7F,$92,$DF,$DE,$4F,$E1,$9C,$B7,$C1,$D4,$C8,$FC,$5D
	.dc.b	$07,$8B,$7E,$C1,$F1,$DF,$E7,$2C,$79,$D7,$E3,$4E,$15,$B9,$D6,$1F
	.dc.b	$E9,$CB,$E3,$5F,$EC,$6A,$D3,$FA,$69,$D5,$7F,$B1,$67,$33,$FD,$C6
	.dc.b	$72,$F1,$15,$D1,$D6,$1C,$1F,$19,$AB,$F1,$C6,$63,$DA,$64,$F1,$39
	.dc.b	$BF,$0E,$1E,$EB,$ED,$5D,$1F,$04,$66,$FD,$13,$CC,$F1,$94,$E0,$F8
	.dc.b	$9C,$1E,$23,$FD,$C8,$53,$99,$8A,$CF,$D8,$7F,$45,$F8,$EB,$E6,$1A
	.dc.b	$4A,$FF,$77,$4B,$E3,$7F,$C3,$51,$86,$BF,$EF,$88,$D6,$3D,$A7,$33
	.dc.b	$FE,$B0,$9C,$BE,$DC,$7F,$B4,$BF,$59,$AF,$C1,$8D,$E5,$F5,$5B,$3F
	.dc.b	$EB,$A3,$71,$F6,$5A,$77,$AF,$3C,$71,$6C,$71,$DB,$FD,$87,$3C,$5B
	.dc.b	$1C,$70,$FC,$4D,$F8,$30,$FC,$E4,$F6,$6D,$70,$7F,$41,$3B,$4D,$9E
	.dc.b	$11,$FA,$29,$E0,$EB,$0F,$2B,$D0,$FE,$6B,$3F,$12,$FF,$D9,$CD,$D3
	.dc.b	$CD,$3F,$2E,$FF,$59,$1E,$D8,$E3,$18,$7C,$B7,$D9,$7D,$87,$F0,$07
	.dc.b	$FA,$1B,$5F,$E8,$0D,$C3,$FD,$BC,$FE,$0E,$37,$B7,$57,$9B,$63,$86
	.dc.b	$1F,$EF,$07,$84,$FF,$4B,$FB,$0F,$E8,$4F,$F5,$26,$B9,$E8,$E9,$F5
	.dc.b	$DF,$D7,$63,$51,$8D,$24,$60,$E5,$E0,$F8,$EF,$FC,$3B,$18,$3F,$AF
	.dc.b	$DB,$EC,$35,$52,$74,$7F,$5F,$9D,$38,$D3,$FD,$FD,$0E,$5F,$29,$C6
	.dc.b	$78,$7C,$77,$C3,$FB,$0D,$47,$F6,$E6,$BA,$31,$3B,$23,$E6,$9F,$99
	.dc.b	$E7,$FF,$B9,$A3,$FB,$BB,$5F,$EE,$8D,$C1,$BF,$F3,$19,$61,$E0,$9C
	.dc.b	$C7,$F3,$5A,$79,$1D,$89,$1F,$1B,$C8,$F0,$F9,$12,$3C,$2E,$9F,$F0
	.dc.b	$27,$D8,$7F,$A3,$5F,$F0,$5B,$5C,$FC,$1F,$E1,$63,$EE,$E3,$E2,$97
	.dc.b	$D8,$3E,$23,$F1,$4D,$8F,$F0,$AB,$5F,$F0,$97,$3B,$CE,$FC,$87,$D8
	.dc.b	$7F,$9B,$27,$FC,$36,$D7,$FC,$30,$DC,$3C,$14,$FA,$3E,$3D,$72,$F8
	.dc.b	$BF,$9B,$9F,$3F,$17,$CB,$C1,$ED,$78,$09,$7F,$E2,$AF,$AF,$7D,$87
	.dc.b	$E8,$E7,$F2,$33,$FE,$29,$6E,$7F,$05,$B8,$7D,$7B,$FF,$3C,$57,$F8
	.dc.b	$D2,$7D,$CD,$D3,$A1,$FA,$F9,$7C,$C7,$EC,$7E,$C3,$FC,$DD,$5F,$E3
	.dc.b	$C6,$BF,$E3,$96,$E3,$F4,$7E,$C3,$F0,$D3,$9F,$6F,$91,$1F,$F2,$2C
	.dc.b	$7E,$89,$FF,$21,$7D,$87,$FB,$39,$FF,$24,$B5,$CF,$C6,$7F,$03,$9F
	.dc.b	$F2,$F4,$6B,$C9,$FE,$0D,$5E,$E8,$E7,$1B,$FA,$CF,$E1,$5F,$5D,$FE
	.dc.b	$92,$9F,$F2,$AB,$5F,$F2,$93,$72,$FC,$57,$D8,$3E,$C3,$89,$2B,$8B
	.dc.b	$FC,$56,$3E,$7C,$7F,$85,$7E,$8B,$FD,$5A,$FF,$98,$9A,$FF,$97,$DB
	.dc.b	$5F,$97,$FA,$DB,$C9,$FA,$D9,$B5,$CF,$E2,$3F,$91,$46,$21,$56,$30
	.dc.b	$FE,$DE,$BF,$6C,$D9,$E4,$7F,$9D,$3E,$C1,$F2,$FF,$CB,$5B,$19,$8C
	.dc.b	$79,$99,$64,$64,$E5,$1C,$7D,$AE,$5F,$25,$27,$B4,$65,$F0,$46,$5F
	.dc.b	$F4,$05,$FF,$9A,$9F,$8D,$F4,$9E,$E2,$3E,$7E,$1A,$FC,$98,$BB,$7E
	.dc.b	$FE,$95,$8D,$3F,$8B,$7D,$7F,$F0,$3A,$FE,$08,$D8,$F9,$D7,$FD,$36
	.dc.b	$1F,$61,$49,$8F,$FA,$6A,$3E,$7B,$63,$2F,$FA,$66,$38,$BB,$DB,$A7
	.dc.b	$CE,$75,$F8,$A9,$3C,$E3,$2F,$58,$CB,$CE,$3F,$B0,$7D,$87,$F0,$B8
	.dc.b	$FF,$55,$B5,$FF,$54,$37,$3F,$E4,$1F,$61,$FC,$1A,$77,$E3,$72,$23
	.dc.b	$FE,$B4,$8D,$3C,$A3,$FB,$74,$34,$A7,$09,$FF,$5B,$43,$E1,$77,$F2
	.dc.b	$92,$7A,$47,$23,$0F,$08,$C3,$E1,$F0,$7B,$91,$F2,$7D,$3D,$7B,$C5
	.dc.b	$CB,$DB,$3F,$E0,$3F,$AF,$7D,$85,$25,$FF,$B4,$3E,$C3,$FB,$E3,$FE
	.dc.b	$CD,$9D,$BF,$ED,$46,$B9,$9F,$F0,$74,$F0,$BF,$E9,$55,$A7,$FA,$64
	.dc.b	$7F,$9B,$39,$3F,$B1,$36,$FF,$D2,$95,$B9,$CF,$47,$42,$4F,$97,$A2
	.dc.b	$BF,$EE,$1B,$FD,$13,$67,$C7,$3E,$4F,$B0,$93,$ED,$7F,$C4,$9F,$61
	.dc.b	$F9,$B8,$FF,$52,$5A,$77,$1D,$5D,$E1,$FE,$2B,$FB,$07,$8F,$49,$5F
	.dc.b	$EF,$BF,$84,$7C,$A3,$A3,$33,$FE,$F7,$B7,$CF,$73,$33,$C1,$FD,$C7
	.dc.b	$EC,$3F,$88,$51,$A6,$2B,$FA,$2C,$EF,$8F,$FC,$7D,$3A,$7C,$D5,$8F
	.dc.b	$A8,$8F,$FB,$06,$BB,$47,$F9,$47,$3F,$3E,$BA,$A5,$78,$F9,$BA,$99
	.dc.b	$7C,$DE,$9F,$80,$93,$D6,$BF,$C3,$D9,$3E,$2B,$33,$FE,$D0,$EB,$C9
	.dc.b	$FE,$23,$F8,$AF,$D4,$47,$FB,$56,$AB,$B3,$18,$6E,$B8,$C7,$FB,$51
	.dc.b	$AE,$CF,$90,$ED,$E2,$6C,$EC,$E5,$D9,$92,$9C,$E7,$FC,$E1,$5E,$4F
	.dc.b	$B0,$FE,$C0,$DF,$9C,$E5,$18,$BF,$F1,$F3,$FE,$68,$9D,$5F,$58,$7C
	.dc.b	$D3,$E5,$30,$D9,$F5,$4F,$37,$FD,$F0,$F0,$9D,$92,$37,$87,$8A,$B9
	.dc.b	$5F,$9D,$B8,$79,$1B,$5D,$2B,$85,$F3,$BE,$0F,$9E,$33,$3E,$4A,$E2
	.dc.b	$B0,$F2,$3E,$23,$96,$1D,$EE,$C7,$72,$9C,$6B,$9D,$72,$C3,$E3,$CF
	.dc.b	$7A,$DC,$E6,$FE,$6E,$1E,$4B,$F3,$DE,$CF,$11,$5C,$92,$78,$C7,$4C
	.dc.b	$3C,$27,$8A,$BA,$4F,$A6,$79,$3C,$1A,$7D,$49,$E3,$87,$C7,$8E,$8F
	.dc.b	$A2,$B9,$C6,$E3,$0E,$8E,$67,$CB,$8D,$BF,$16,$36,$EB,$0F,$84,$F3
	.dc.b	$3D,$F8,$7B,$5C,$35,$E8,$7E,$1C,$FF,$26,$C7,$D8,$6F,$97,$FC,$51
	.dc.b	$FF,$28,$9E,$12,$C2,$35,$3F,$A3,$6E,$3F,$E9,$DE,$7F,$F4,$46,$DF
	.dc.b	$9F,$9C,$E1,$F9,$A6,$BF,$89,$6C,$F0,$8F,$D0,$C6,$B0,$F0,$72,$F3
	.dc.b	$3F,$6D,$5B,$C7,$C5,$19,$BE,$2E,$BD,$CF,$F5,$6C,$EA,$72,$47,$F8
	.dc.b	$7D,$73,$CB,$F7,$38,$F8,$49,$87,$3C,$BB,$5F,$32,$57,$C1,$C7,$BB
	.dc.b	$CA,$B8,$63,$DE,$35,$7C,$9B,$19,$07,$F4,$EA,$E6,$FC,$9A,$F8,$3E
	.dc.b	$66,$23,$FA,$9F,$99,$73,$FF,$2A,$E3,$F1,$A3,$15,$23,$E4,$FC,$65
	.dc.b	$BF,$E5,$16,$8C,$67,$F1,$1B,$9F,$DD,$6D,$79,$AB,$85,$F9,$0F,$E5
	.dc.b	$5F,$61,$FE,$DA,$7F,$A2,$4F,$B2,$3F,$D4,$B1,$88,$46,$DF,$EA,$31
	.dc.b	$FE,$E7,$8D,$C6,$0E,$5F,$07,$FE,$E8,$E6,$FF,$6C,$4E,$9B,$19,$AE
	.dc.b	$5E,$2D,$BF,$17,$8D,$7A,$9F,$58,$F8,$5D,$57,$92,$B5,$39,$C3,$D1
	.dc.b	$9F,$F8,$ED,$B3,$C2,$1F,$01,$63,$1E,$F6,$7B,$6D,$F2,$6E,$3E,$D3
	.dc.b	$E1,$B8,$DB,$DA,$FA,$57,$D4,$FB,$EB,$FB,$E3,$6B,$FB,$CA,$F0,$78
	.dc.b	$39,$AE,$3B,$E0,$CC,$DB,$FA,$F2,$FA,$97,$D8,$78,$5B,$3A,$F8,$8F
	.dc.b	$EB,$4F,$08,$CE,$5F,$5E,$F9,$5F,$13,$D5,$7C,$27,$55,$C2,$97,$FA
	.dc.b	$29,$7D,$83,$EA,$BE,$17,$D8,$3E,$37,$EB,$CE,$6F,$D3,$39,$7F,$1D
	.dc.b	$19,$BF,$24,$E5,$FA,$11,$9C,$3F,$19,$39,$7C,$73,$FB,$49,$D5,$79
	.dc.b	$5B,$1F,$43,$DC,$F1,$E3,$F1,$FC,$49,$7F,$17,$E4,$ED,$FC,$09,$B8
	.dc.b	$FE,$6F,$45,$62,$FE,$EF,$D8,$7D,$AE,$BF,$B6,$D8,$FA,$FC,$68,$C6
	.dc.b	$5F,$46,$31,$3A,$AF,$CE,$FD,$73,$E6,$7F,$19,$D6,$FA,$FC,$C6,$1B
	.dc.b	$7F,$84,$9D,$3F,$B9,$AE,$4E,$67,$78,$FE,$6E,$72,$FD,$EF,$B0,$CE
	.dc.b	$DF,$B9,$CD,$E7,$FE,$5B,$0F,$CB,$B7,$1F,$6B,$70,$F6,$36,$FE,$47
	.dc.b	$A3,$F9,$53,$F2,$DE,$ED,$BF,$CA,$45,$BF,$C7,$CD,$AF,$BB,$0C,$FF
	.dc.b	$94,$BF,$86,$F2,$BE,$3F,$91,$F9,$06,$E3,$F8,$E6,$CE,$E3,$E2,$4E
	.dc.b	$2F,$DD,$F1,$36,$7F,$CE,$5C,$8F,$91,$FD,$67,$CB,$7E,$B5,$B3,$FA
	.dc.b	$B6,$E3,$FA,$66,$E1,$F2,$7C,$3F,$DE,$9F,$BE,$F5,$2D,$FE,$12,$2D
	.dc.b	$FE,$0E,$6D,$7C,$0F,$B0,$67,$FB,$DF,$A1,$FE,$43,$18,$C7,$F7,$91
	.dc.b	$89,$D4,$7F,$39,$9D,$61,$FB,$AA,$D4,$E6,$BF,$84,$76,$62,$BF,$8F
	.dc.b	$7D,$7B,$EC,$3F,$8E,$77,$BD,$EF,$E6,$33,$9C,$7F,$89,$CE,$5F,$E9
	.dc.b	$1E,$9F,$F8,$1F,$5F,$E8,$53,$FD,$4B,$1F,$E5,$53,$9A,$FF,$50,$4E
	.dc.b	$6B,$EF,$FD,$73,$EB,$FF,$96,$7D,$83,$87,$FA,$0C,$E7,$0F,$E4,$73
	.dc.b	$97,$FA,$C4,$30,$FE,$69,$F6,$1F,$DE,$27,$31,$8C,$F9,$F9,$DF,$CE
	.dc.b	$7E,$C1,$F2,$58,$BF,$E5,$F1,$87,$FB,$27,$B0,$E9,$FE,$81,$F6,$0F
	.dc.b	$2B,$FA,$AC,$E6,$BF,$9B,$CE,$5E,$3E,$AF,$F4,$7F,$B0,$FF,$59,$46
	.dc.b	$0F,$E7,$AE,$34,$FE,$A1,$F0,$6F,$AA,$FE,$9D,$E2,$0F,$EC,$71,$9B
	.dc.b	$FE,$8B,$19,$7F,$BD,$47,$FB,$91,$CF,$C0,$F0,$F6,$F9,$BC,$B0,$FE
	.dc.b	$B5,$F6,$1E,$97,$C2,$63,$7F,$EE,$37,$F0,$E3,$FA,$54,$BE,$6F,$3C
	.dc.b	$5C,$9F,$EE,$FF,$61,$F0,$6B,$FA,$87,$A3,$FC,$E7,$0F,$E9,$53,$9E
	.dc.b	$CF,$F0,$44,$E6,$BF,$C5,$FF,$60,$F6,$3F,$C0,$3A,$3E,$8B,$FC,$07
	.dc.b	$39,$CB,$FC,$31,$39,$7F,$C3,$50,$FB,$0D,$FD,$2F,$06,$CF,$F8,$7F
	.dc.b	$EC,$3F,$8F,$BC,$23,$FC,$1F,$18,$CB,$FC,$05,$7C,$EB,$87,$97,$FE
	.dc.b	$14,$FB,$0F,$E1,$11,$83,$FC,$54,$E2,$BF,$C1,$5F,$15,$FD,$2A,$75
	.dc.b	$18,$F6,$FB,$DF,$F8,$7A,$B5,$39,$AE,$7E,$6E,$6F,$FC,$47,$F5,$FF
	.dc.b	$81,$AF,$EF,$9F,$60,$F9,$3C,$F6,$FF,$85,$63,$18,$FF,$8F,$6F,$87
	.dc.b	$2F,$F8,$DE,$73,$87,$F8,$7E,$73,$5F,$E6,$0A,$56,$9F,$E6,$B3,$AA
	.dc.b	$FF,$32,$B6,$37,$F0,$3E,$2B,$F0,$71,$7F,$89,$FE,$C3,$FA,$8D,$72
	.dc.b	$8C,$7B,$FE,$C6,$D7,$03,$FC,$43,$5C,$1F,$F0,$96,$1F,$C2,$5F,$37
	.dc.b	$3B,$E4,$3E,$2F,$C5,$FF,$87,$1B,$3F,$06,$3E,$4E,$98,$3F,$C7,$D9
	.dc.b	$F1,$9F,$F3,$13,$FB,$33,$FA,$99,$98,$DF,$C4,$7F,$95,$23,$06,$AF
	.dc.b	$FC,$AF,$3F,$E8,$8F,$1F,$FD,$13,$F6,$0F,$8E,$FF,$15,$5F,$0C,$7F
	.dc.b	$D7,$B5,$A9,$CE,$1F,$E1,$EB,$E1,$87,$F8,$C4,$56,$DF,$AD,$3B,$AE
	.dc.b	$70,$AE,$2B,$0F,$54,$62,$3C,$C6,$E7,$C3,$87,$58,$FF,$20,$61,$6F
	.dc.b	$8D,$76,$BE,$B1,$8A,$F3,$56,$F0,$F2,$46,$0E,$C7,$A3,$0E,$CF,$8A
	.dc.b	$F9,$CF,$08,$F0,$9C,$0F,$53,$97,$D2,$7F,$B9,$8F,$8A,$52,$99,$AD
	.dc.b	$7D,$87,$F4,$0F,$8E,$BE,$19,$A4,$AF,$F7,$8C,$FF,$A2,$B2,$E3,$F1
	.dc.b	$BF,$2E,$72,$D7,$EB,$4E,$DC,$CF,$CB,$D3,$FD,$79,$17,$6F,$F3,$66
	.dc.b	$9F,$5A,$F8,$37,$0E,$CF,$DA,$BE,$37,$F0,$DB,$1B,$C3,$EC,$5F,$29
	.dc.b	$FF,$3C,$1A,$2E,$1E,$07,$FC,$F9,$1B,$BF,$F6,$FB,$5F,$CC,$4E,$DF
	.dc.b	$F7,$1C,$F3,$77,$87,$9A,$39,$B6,$37,$3C,$5E,$67,$9A,$34,$E0,$FF
	.dc.b	$58,$4F,$BA,$FF,$C8,$6D,$D7,$AC,$7C,$99,$FF,$6C,$4E,$5E,$E7,$FB
	.dc.b	$6F,$0F,$57,$D8,$7F,$05,$8F,$F6,$F3,$5F,$F5,$F3,$6B,$FC,$A2,$FF
	.dc.b	$B2,$23,$7B,$7E,$E3,$0F,$F5,$14,$ED,$B1,$99,$09,$E4,$FF,$B2,$FE
	.dc.b	$C3,$FD,$16,$FF,$BB,$DA,$E7,$AB,$92,$FA,$FF,$EB,$4E,$E3,$31,$FB
	.dc.b	$0C,$3F,$D6,$8D,$9F,$0D,$74,$C7,$FC,$D6,$FF,$86,$63,$C1,$E1,$F6
	.dc.b	$8C,$47,$FB,$7E,$9F,$61,$9A,$93,$C0,$FE,$89,$87,$82,$73,$1C,$7A
	.dc.b	$3C,$A6,$31,$FF,$3D,$E1,$FE,$C5,$9E,$27,$C6,$FB,$05,$7C,$06,$C6
	.dc.b	$8F,$85,$D7,$99,$91,$99,$DB,$E9,$9F,$93,$E8,$7F,$BB,$E3,$93,$5F
	.dc.b	$F7,$E3,$70,$FF,$47,$D7,$27,$FD,$23,$0F,$23,$99,$2B,$C9,$8F,$F8
	.dc.b	$55,$EB,$E4,$78,$08,$F8,$27,$ED,$FD,$87,$FA,$36,$3D,$8D,$7F,$9A
	.dc.b	$6B,$FE,$ED,$8C,$63,$FE,$A2,$7E,$D4,$7F,$BC,$7E,$C1,$F0,$FF,$86
	.dc.b	$6D,$FC,$76,$BD,$0F,$C9,$F2,$7F,$8D,$DF,$99,$F6,$1F,$E6,$F6,$FB
	.dc.b	$DF,$90,$8C,$47,$13,$FB,$67,$D1,$67,$AA,$47,$27,$F2,$F7,$E6,$8F
	.dc.b	$D0,$D3,$DA,$E8,$48,$FB,$9F,$5E,$FB,$0F,$CF,$CF,$E4,$B7,$FF,$60
	.dc.b	$B7,$6F,$D5,$DF,$3E,$8F,$D2,$E1,$FE,$78,$8F,$C3,$57,$B5,$F4,$36
	.dc.b	$F9,$FA,$7C,$3E,$63,$F5,$BF,$61,$FE,$6D,$BF,$C7,$35,$F8,$0D,$AF
	.dc.b	$F6,$9F,$D8,$7F,$71,$7F,$DA,$B0,$F6,$79,$92,$BF,$31,$5F,$AA,$7F
	.dc.b	$35,$F6,$1F,$EB,$C6,$FC,$5F,$EF,$A3,$11,$F3,$E7,$F1,$8F,$D1,$8F
	.dc.b	$73,$7C,$8E,$07,$F0,$69,$7D,$67,$F0,$8F,$B0,$FF,$4A,$57,$EA,$5A
	.dc.b	$FD,$B6,$E7,$FE,$EE,$FA,$F7,$D8,$71,$25,$71,$7F,$89,$C7,$D3,$8F
	.dc.b	$E3,$BF,$45,$FE,$AF,$7F,$6C,$D7,$F1,$0D,$AF,$F7,$DF,$D6,$DE,$5F
	.dc.b	$B7,$36,$B9,$FC,$47,$F2,$08,$C4,$2A,$C6,$1F,$DC,$57,$E1,$1B,$3C
	.dc.b	$8F,$E1,$1F,$60,$F9,$6F,$E5,$6D,$8C,$C6,$3C,$EC,$B2,$32,$72,$8E
	.dc.b	$3E,$D7,$2F,$91,$93,$DA,$32,$F8,$23,$2F,$EA,$6F,$FC,$E8,$FC,$6F
	.dc.b	$A4,$F7,$11,$F3,$F0,$D7,$E4,$C5,$D9,$A7,$E8,$AF,$EF,$D7,$33,$F9
	.dc.b	$37,$D7,$FF,$95,$A7,$F4,$ED,$8F,$AF,$7F,$D3,$21,$F6,$14,$93,$CB
	.dc.b	$6A,$FF,$01,$9D,$BD,$A7,$94,$F3,$F9,$BE,$9F,$17,$2B,$C7,$19,$7C
	.dc.b	$B1,$97,$A4,$7F,$5E,$FB,$07,$8D,$49,$3F,$D0,$7E,$C3,$F8,$33,$FD
	.dc.b	$0E,$76,$D9,$DB,$FC,$5E,$76,$FF,$19,$9D,$B9,$CF,$FB,$6C,$9E,$3B
	.dc.b	$7C,$E9,$FE,$F3,$2F,$85,$DF,$CA,$C9,$F2,$47,$23,$0F,$86,$30,$FA
	.dc.b	$7C,$1E,$E4,$7D,$9F,$4F,$5F,$44,$5C,$3C,$50,$C3,$F3,$AB,$EB,$DF
	.dc.b	$61,$FE,$06,$6C,$7F,$67,$6B,$FC,$E7,$EB,$FF,$C1,$2F,$F3,$B8,$CD
	.dc.b	$F6,$6F,$27,$F2,$E6,$CF,$F8,$6A,$B8,$D6,$EF,$FC,$2B,$1F,$E0,$F9
	.dc.b	$36,$B8,$C6,$3A,$36,$47,$FB,$E4,$BA,$31,$1E,$79,$B8,$7C,$2A,$D4
	.dc.b	$F0,$FA,$FF,$09,$1F,$F0,$37,$D8,$7E,$7D,$F1,$C6,$5F,$84,$FF,$8B
	.dc.b	$23,$5F,$61,$FC,$E6,$7F,$C1,$CD,$7F,$B0,$36,$FF,$09,$F0,$6F,$95
	.dc.b	$73,$FF,$8D,$AB,$5B,$FF,$56,$5F,$3B,$93,$5E,$57,$F8,$F3,$EC,$3E
	.dc.b	$D9,$39,$79,$69,$96,$3F,$E2,$18,$FC,$0D,$FF,$51,$C3,$FB,$1C,$3E
	.dc.b	$9B,$A1,$2B,$FC,$5B,$A7,$9E,$B1,$1F,$DE,$E9,$CF,$E3,$33,$B7,$E4
	.dc.b	$23,$FC,$B9,$3F,$E5,$A6,$A6,$8F,$89,$F6,$7A,$BF,$EC,$6E,$BF,$29
	.dc.b	$6B,$93,$5E,$51,$6B,$F6,$51,$FE,$E7,$9E,$2D,$38,$94,$E2,$58,$E2
	.dc.b	$FF,$B9,$23,$8B,$4E,$25,$38,$97,$EC,$3F,$B2,$FC,$57,$26,$CF,$F8
	.dc.b	$41,$B3,$FE,$7C,$7F,$CE,$F3,$BF,$9C,$F6,$43,$E6,$9F,$29,$E7,$8C
	.dc.b	$36,$77,$3A,$BF,$13,$6F,$C4,$DC,$3C,$55,$C2,$FA,$4E,$6F,$C8,$D5
	.dc.b	$87,$A1,$B1,$98,$F1,$C7,$82,$7C,$8F,$46,$39,$3E,$35,$78,$5C,$61
	.dc.b	$E6,$9E,$31,$8C,$3E,$46,$1D,$CF,$05,$F1,$BE,$15,$B3,$E6,$46,$6B
	.dc.b	$8D,$F8,$AB,$E7,$55,$9C,$BC,$4F,$7D,$71,$7E,$1D,$6D,$EA,$53,$E1
	.dc.b	$B7,$0F,$8C,$7C,$4B,$F2,$92,$FE,$8D,$6F,$0F,$65,$FC,$0A,$D4,$F8
	.dc.b	$0F,$25,$F4,$AD,$C6,$30,$F3,$3E,$23,$07,$5B,$F2,$46,$AB,$C9,$87
	.dc.b	$C0,$C3,$E1,$DF,$6C,$3A,$C6,$0F,$9B,$1F,$E6,$83,$B1,$69,$E5,$FF
	.dc.b	$16,$8D,$1A,$3F,$40,$E6,$7F,$48,$DB,$FF,$39,$F2,$7F,$A4,$B8,$3F
	.dc.b	$3D,$39,$BF,$D3,$35,$FC,$9B,$67,$84,$7E,$C6,$35,$87,$93,$97,$99
	.dc.b	$FB,$FA,$DE,$3E,$28,$CD,$F1,$75,$EF,$7F,$79,$9D,$4E,$48,$FF,$06
	.dc.b	$AE,$79,$7B,$AF,$9C,$F0,$35,$7C,$E7,$57,$FC,$26,$F8,$BE,$EB,$F0
	.dc.b	$E5,$CA,$BE,$2D,$F1,$AD,$1C,$5C,$3F,$12,$BD,$DE,$66,$30,$79,$96
	.dc.b	$30,$DC,$76,$7C,$5F,$8C,$C6,$1E,$56,$E7,$EF,$6E,$7F,$61,$B2,$3F
	.dc.b	$D3,$4E,$17,$D8,$7F,$B8,$6F,$FD,$E4,$BC,$5F,$F1,$37,$37,$F8,$5A
	.dc.b	$35,$5F,$EC,$77,$55,$C3,$E0,$B6,$FF,$AE,$A3,$33,$B9,$CB,$E3,$8C
	.dc.b	$D7,$5F,$1F,$3D,$4F,$AC,$7C,$37,$0C,$3C,$F5,$A9,$CE,$5E,$88,$D6
	.dc.b	$3F,$E3,$B8,$F5,$FC,$0E,$27,$3E,$EE,$BB,$6D,$7C,$06,$E3,$ED,$6E
	.dc.b	$9F,$05,$F4,$AF,$AB,$F4,$D7,$CD,$8C,$6F,$A9,$78,$56,$25,$6C,$E1
	.dc.b	$3C,$B0,$5F,$5B,$DC,$7F,$BF,$61,$F5,$2F,$B0,$F1,$B7,$0E,$95,$CB
	.dc.b	$E4,$FF,$AA,$C6,$2F,$8D,$F0,$24,$66,$77,$5D,$EF,$A3,$E4,$9F,$F7
	.dc.b	$AF,$D7,$BE,$A7,$E1,$FD,$83,$E3,$FE,$B4,$E6,$BC,$13,$97,$CB,$69
	.dc.b	$D3,$ED,$9D,$3D,$CF,$CC,$46,$B3,$FC,$9C,$E9,$FF,$34,$C6,$6B,$CB
	.dc.b	$87,$D1,$F7,$3C,$6F,$C6,$F1,$24,$FC,$3C,$DF,$27,$8C,$FE,$05,$5C
	.dc.b	$BF,$43,$5A,$BE,$13,$F7,$7E,$C3,$EF,$75,$D9,$F9,$DC,$5F,$A6,$C3
	.dc.b	$F5,$2A,$B7,$3A,$6E,$3E,$F1,$F5,$CF,$9B,$FC,$7F,$5B,$EB,$FF,$45
	.dc.b	$39,$C3,$EE,$4E,$5F,$9B,$19,$BF,$C4,$4E,$5E,$D1,$9C,$BF,$7B,$39
	.dc.b	$7F,$D2,$71,$9A,$FE,$11,$5E,$4C,$3F,$1F,$F6,$1E,$27,$F3,$DE,$7F
	.dc.b	$E7,$F2,$B7,$F0,$9B,$7F,$99,$D3,$B4,$FE,$82,$B8,$3F,$4F,$D1,$FC
	.dc.b	$D9,$F9,$CF,$76,$C6,$49,$7F,$95,$8C,$BE,$CF,$1F,$F2,$7E,$57,$C6
	.dc.b	$6D,$7C,$46,$CE,$E3,$F2,$B3,$B7,$47,$DA,$F8,$9C,$4F,$E8,$FE,$C3
	.dc.b	$F5,$FF,$1D,$FB,$06,$D7,$E2,$DB,$3B,$8F,$C7,$4F,$F0,$3F,$0E,$47
	.dc.b	$E1,$BD,$4B,$5F,$16,$6F,$1F,$E2,$26,$D7,$D2,$FB,$06,$5F,$C0,$3E
	.dc.b	$8B,$F9,$24,$62,$7F,$39,$18,$9D,$47,$EC,$27,$58,$FE,$F6,$B5,$39
	.dc.b	$AF,$E1,$7D,$98,$AF,$CA,$7D,$7B,$EC,$3F,$90,$F7,$FF,$AE,$F0,$FD
	.dc.b	$DC,$E5,$FE,$1D,$19,$AF,$CE,$7A,$5F,$E0,$2D,$7F,$A5,$CB,$CE,$FE
	.dc.b	$79,$39,$BF,$E5,$53,$97,$F8,$3F,$D7,$BE,$5B,$F9,$8F,$D8,$7F,$A7
	.dc.b	$27,$31,$8E,$1F,$E8,$D4,$BF,$E6,$FF,$61,$FD,$86,$F9,$C6,$34,$DF
	.dc.b	$9D,$FC,$F3,$EC,$1F,$27,$8B,$FE,$51,$18,$7F,$8C,$FA,$CE,$8F,$E8
	.dc.b	$7F,$60,$F2,$3F,$AB,$4E,$6B,$F9,$84,$E5,$E3,$EA,$FF,$4A,$FB,$0F
	.dc.b	$F5,$64,$60,$FE,$6C,$E3,$4F,$E0,$9F,$04,$FA,$AF,$EA,$3E,$10,$FE
	.dc.b	$C9,$19,$BF,$E7,$D1,$97,$F9,$7C,$3D,$CE,$3E,$FF,$06,$CE,$B0,$FE
	.dc.b	$BB,$F6,$1E,$97,$C2,$63,$7F,$EE,$53,$F0,$E3,$FA,$A4,$BE,$6B,$3C
	.dc.b	$5C,$9F,$EF,$1F,$61,$F0,$6B,$FA,$D7,$A3,$FC,$E7,$0F,$E1,$B3,$9E
	.dc.b	$EF,$F0,$7C,$E6,$BF,$C3,$1F,$5C,$F6,$3F,$C0,$5A,$3E,$8B,$FC,$09
	.dc.b	$39,$CB,$FA,$6C,$E5,$FE,$9F,$0F,$B0,$DF,$D3,$70,$6C,$FF,$54,$FB
	.dc.b	$0F,$F7,$73,$C2,$3F,$C2,$11,$8C,$BF,$9B,$5F,$3A,$E1,$E5,$FF,$85
	.dc.b	$7E,$C1,$E6,$F0,$8C,$3F,$E1,$9F,$89,$FE,$95,$87,$F3,$FC,$7F,$C2
	.dc.b	$CF,$F4,$18,$CF,$03,$1F,$E6,$35,$A8,$C5,$7F,$8E,$EB,$75,$FC,$D9
	.dc.b	$B1,$FE,$3F,$C5,$E6,$F8,$7D,$BF,$F1,$57,$D7,$7E,$1E,$BF,$C1,$9F
	.dc.b	$60,$F9,$2C,$F7,$7F,$87,$A3,$19,$FF,$78,$BE,$1A,$7F,$91,$27,$38
	.dc.b	$7F,$59,$9C,$D7,$F8,$8E,$73,$5E,$8A,$56,$9F,$F2,$A3,$61,$3F,$81
	.dc.b	$9D,$DF,$A3,$8B,$FC,$63,$F6,$1F,$D5,$67,$8C,$63,$DA,$D6,$5F,$E1
	.dc.b	$FA,$D5,$78,$2B,$53,$EB,$F3,$FF,$C7,$7D,$0F,$8F,$F8,$B1,$86,$CE
	.dc.b	$CF,$85,$78,$3F,$BC,$69,$C6,$3F,$C3,$AF,$EE,$4F,$80,$66,$37,$F0
	.dc.b	$FF,$E5,$B8,$C1,$AA,$FF,$2F,$4F,$F9,$DF,$C7,$FF,$3C,$7D,$87,$FA
	.dc.b	$1A,$3F,$C9,$D3,$A5,$3A,$F8,$DF,$EE,$F7,$DE,$F8,$57,$F7,$AA,$D3
	.dc.b	$CB,$1A,$E4,$4F,$0A,$E5,$7F,$EC,$8B,$DC,$62,$BF,$DA,$11,$8C,$3F
	.dc.b	$2B,$7E,$09,$F5,$1D,$2F,$CB,$3A,$7A,$46,$4F,$21,$4F,$24,$F8,$E7
	.dc.b	$84,$78,$0D,$5F,$8D,$DD,$76,$C3,$DB,$1E,$23,$FD,$C2,$7C,$82,$94
	.dc.b	$CE,$1A,$FB,$0F,$E8,$7F,$1B,$7C,$33,$49,$5F,$EE,$C9,$7C,$67,$F9
	.dc.b	$EA,$30,$D7,$FD,$ED,$1A,$CB,$B4,$E6,$7F,$C5,$F3,$97,$E1,$65,$FD
	.dc.b	$9D,$FA,$ED,$7E,$34,$6F,$4F,$BE,$D9,$FF,$1C,$9B,$8F,$C1,$B4,$EF
	.dc.b	$5E,$78,$E2,$D8,$E3,$B7,$F9,$0A,$78,$B6,$38,$E1,$F8,$AB,$F0,$61
	.dc.b	$FE,$CF,$9E,$CD,$AE,$0F,$FB,$4E,$76,$9B,$3C,$23,$FD,$CB,$3C,$1D
	.dc.b	$61,$E5,$7A,$1F,$E8,$C9,$F8,$97,$FC,$3D,$BA,$79,$A7,$E5,$DF,$E9
	.dc.b	$A3,$DB,$1C,$63,$0F,$96,$FB,$2F,$B0,$FE,$0C,$FF,$A6,$9A,$FF,$9E
	.dc.b	$5B,$87,$F1,$F7,$F0,$71,$BD,$BA,$BC,$DB,$1C,$30,$FF,$37,$BC,$27
	.dc.b	$FD,$09,$F6,$1F,$E9,$67,$FD,$52,$D7,$3D,$3D,$3E,$BB,$FA,$F4,$6A
	.dc.b	$31,$A4,$8C,$1C,$BC,$3F,$1D,$FF,$86,$E3,$07,$FA,$DA,$DF,$61,$AA
	.dc.b	$93,$A3,$FD,$31,$3A,$71,$A7,$F3,$F8,$73,$79,$4E,$33,$C3,$E3,$7E
	.dc.b	$1F,$D8,$6A,$3F,$D8,$AD,$74,$62,$76,$47,$CD,$3F,$33,$D1,$FF,$77
	.dc.b	$C7,$FB,$39,$AF,$FA,$B5,$B8,$7F,$53,$BE,$CF,$F5,$58,$79,$3C,$C9
	.dc.b	$5F,$0F,$AB,$B7,$93,$23,$B3,$A7,$FD,$75,$F6,$1F,$E0,$27,$FD,$BE
	.dc.b	$D7,$3E,$3F,$E0,$A3,$ED,$63,$FD,$97,$EC,$1F,$09,$F8,$66,$DF,$FB
	.dc.b	$A1,$AF,$FB,$1F,$9F,$FC,$60,$FF,$BB,$3E,$C3,$FC,$CC,$DF,$7B,$F1
	.dc.b	$F1,$88,$E1,$F4,$78,$EB,$91,$C5,$FC,$AD,$F9,$3A,$FF,$D8,$36,$F6
	.dc.b	$F9,$92,$FF,$DF,$7F,$5E,$FB,$0F,$CE,$CF,$E2,$E7,$FD,$B8,$DD,$BF
	.dc.b	$C0,$4D,$C3,$FD,$9B,$7F,$E7,$0A,$1C,$3D,$4E,$64,$7B,$FC,$BF,$EA
	.dc.b	$BE,$C3,$FC,$C9,$1C,$31,$FF,$07,$46,$23,$FD,$D8,$D8,$FF,$09,$7D
	.dc.b	$7F,$FB,$72,$73,$EC,$F3,$23,$CF,$2F,$F7,$0C,$FE,$76,$3F,$DE,$DF
	.dc.b	$61,$FE,$B2,$74,$7A,$1C,$3A,$F8,$BF,$DF,$CF,$F9,$66,$36,$27,$E0
	.dc.b	$3D,$3C,$AF,$E0,$B1,$8C,$7C,$FF,$5B,$FC,$23,$EC,$3F,$D1,$D7,$C9
	.dc.b	$AA,$7F,$C5,$D1,$87,$FC,$61,$F5,$EF,$B0,$E4,$4A,$E4,$FF,$14,$8F
	.dc.b	$A9,$1F,$4B,$E8,$BF,$D5,$4F,$A1,$AF,$7D,$7E,$BA,$8C,$DF,$F8,$FF
	.dc.b	$EB,$6D,$6A,$6D,$75,$F8,$9F,$E4,$71,$88,$55,$8C,$3F,$B8,$AE,$ED
	.dc.b	$9E,$47,$CC,$FB,$07,$CA,$7F,$2E,$6C,$66,$31,$E6,$E5,$91,$93,$94
	.dc.b	$71,$F6,$F9,$7D,$54,$9E,$D1,$97,$C1,$19,$7E,$21,$F9,$AB,$FF,$36
	.dc.b	$3F,$27,$E9,$3D,$E4,$7D,$1C,$35,$E2,$7C,$B8,$C6,$DF,$E6,$9A,$56
	.dc.b	$34,$FA,$3F,$61,$FE,$4E,$69,$9C,$3F,$57,$0F,$B0,$FE,$A2,$D8,$FC
	.dc.b	$9B,$5F,$0B,$F4,$9C,$C7,$02,$3C,$36,$E7,$F3,$3D,$3E,$32,$4F,$48
	.dc.b	$CB,$DA,$32,$F6,$8F,$EC,$7F,$60,$F1,$69,$27,$F4,$FF,$61,$FC,$35
	.dc.b	$FD,$4C,$ED,$B3,$B7,$E4,$CE,$DF,$97,$3B,$73,$BF,$F7,$29,$3C,$77
	.dc.b	$FF,$4B,$43,$E2,$35,$F2,$B2,$FC,$71,$C4,$C3,$E0,$8C,$3E,$8F,$07
	.dc.b	$B1,$1F,$5F,$D3,$D7,$CF,$17,$0F,$0C,$FF,$81,$BE,$BD,$F6,$14,$97
	.dc.b	$FC,$2B,$EC,$3F,$BF,$BF,$88,$9D,$BF,$C3,$9A,$E6,$7F,$C2,$33,$C2
	.dc.b	$FF,$D6,$95,$A7,$FD,$6D,$1F,$E8,$1E,$4F,$DE,$B6,$FF,$9A,$D6,$E7
	.dc.b	$3D,$1D,$09,$3E,$5E,$8A,$FE,$72,$FF,$46,$D9,$F1,$CF,$93,$EC,$24
	.dc.b	$FB,$5F,$F1,$3F,$D8,$7E,$76,$3F,$A2,$DA,$77,$1D,$5D,$E1,$FE,$2E
	.dc.b	$FB,$07,$8F,$49,$5F,$CE,$3E,$15,$F2,$CE,$8C,$CF,$EC,$AD,$F3,$DC
	.dc.b	$CC,$F0,$7F,$B6,$FE,$C3,$F8,$95,$1A,$62,$BF,$A7,$4E,$F8,$FF,$C8
	.dc.b	$33,$A7,$CD,$58,$FA,$48,$FF,$46,$94,$7F,$95,$33,$F3,$EB,$AA,$FE
	.dc.b	$0F,$4E,$6E,$A6,$71,$F3,$7A,$7E,$02,$57,$5F,$8B,$AF,$FB,$2B,$AF
	.dc.b	$F9,$DE,$1F,$A4,$8F,$F6,$94,$60,$FF,$65,$1A,$AB,$1F,$DE,$27,$83
	.dc.b	$CC,$C4,$E9,$F8,$CD,$8F,$F3,$4F,$C5,$F0,$78,$B6,$34,$6E,$B9,$7D
	.dc.b	$87,$F6,$09,$EE,$D9,$FE,$A6,$DB,$FF,$05,$BF,$E6,$79,$DF,$CE,$F2
	.dc.b	$F9,$B7,$CB,$7B,$AF,$71,$87,$D4,$2F,$A3,$63,$57,$E2,$6E,$1E,$26
	.dc.b	$CF,$8B,$0F,$06,$1D,$E3,$93,$70,$57,$09,$F0,$57,$07,$C7,$3A,$AE
	.dc.b	$15,$B8,$F2,$99,$9D,$C6,$4D,$57,$3C,$3D,$17,$C8,$F1,$46,$70,$F5
	.dc.b	$46,$6F,$C3,$39,$7E,$3C,$66,$BE,$6C,$5A,$E2,$F8,$A7,$AB,$93,$89
	.dc.b	$93,$A9,$93,$E2,$99,$9E,$AD,$9E,$B1,$EA,$8E,$26,$4E,$AE,$4E,$25
	.dc.b	$8E,$24,$8F,$AC,$78,$70,$F9,$53,$B8,$C9,$D4,$F2,$B9,$7E,$B9,$C7
	.dc.b	$0F,$A1,$7F,$02,$38,$57,$4B,$F0,$47,$03,$35,$C1,$F5,$DF,$5A,$E1
	.dc.b	$1E,$CA,$FC,$FE,$1F,$4E,$3C,$71,$D2,$7F,$07,$7F,$9B,$9F,$F3,$04
	.dc.b	$7E,$15,$BE,$5F,$F1,$E7,$FC,$C8,$7D,$03,$3C,$9F,$95,$8D,$63,$FB
	.dc.b	$EE,$7F,$F4,$BF,$0E,$6A,$5F,$E5,$AF,$79,$7F,$91,$AF,$73,$9A,$DE
	.dc.b	$1F,$BA,$BD,$CF,$FA,$0A,$B7,$39,$AD,$DF,$EF,$AF,$77,$FC,$FB,$DD
	.dc.b	$63,$A8,$FE,$03,$3A,$CB,$FC,$AD,$87,$EC,$72,$EB,$8F,$38,$C6,$3C
	.dc.b	$B0,$F3,$1F,$C0,$AB,$6F,$49,$DE,$3D,$E3,$C4,$D9,$FF,$29,$BD,$FC
	.dc.b	$BE,$99,$FC,$9F,$32,$E5,$F1,$66,$E1,$F0,$66,$BF,$D6,$9B,$83,$E3
	.dc.b	$6C,$61,$B8,$FB,$5B,$5E,$96,$E3,$F6,$5B,$8F,$F9,$EB,$EC,$3F,$DC
	.dc.b	$37,$FE,$96,$5E,$27,$C3,$E6,$F8,$98,$A7,$6F,$F9,$61,$DD,$71,$F8
	.dc.b	$4E,$0F,$5B,$B5,$FE,$BE,$AE,$37,$D1,$E3,$1C,$FC,$6D,$F5,$3E,$AD
	.dc.b	$F0,$DC,$AB,$C5,$5A,$9C,$E5,$E9,$8D,$61,$FE,$3B,$8F,$67,$C0,$62
	.dc.b	$73,$EF,$6B,$B6,$D7,$C1,$6E,$3E,$D6,$E9,$E7,$7D,$6B,$EA,$3D,$2F
	.dc.b	$C7,$6D,$7F,$B5,$57,$83,$C1,$CD,$FF,$AE,$5B,$5C,$F7,$D5,$6D,$EF
	.dc.b	$F5,$AF,$B0,$A7,$8B,$E2,$7F,$AD,$E5,$FE,$D4,$C1,$5B,$8C,$37,$2F
	.dc.b	$F4,$BC,$EA,$F9,$78,$1D,$2F,$FD,$AF,$F5,$EF,$AA,$F7,$FD,$83,$E3
	.dc.b	$BE,$94,$E6,$BC,$13,$97,$8D,$70,$AD,$B6,$32,$4C,$BC,$7B,$BD,$CD
	.dc.b	$CF,$A3,$C5,$F8,$27,$C2,$F9,$4B,$19,$25,$FB,$37,$4E,$A3,$AF,$D8
	.dc.b	$7D,$1E,$CF,$A4,$DB,$F9,$19,$23,$15,$C2,$73,$5F,$91,$FB,$07,$CD
	.dc.b	$FE,$0B,$B1,$F5,$F9,$8C,$36,$FE,$A4,$E9,$61,$FE,$C0,$BE,$71,$AA
	.dc.b	$E1,$39,$BF,$C0,$FD,$86,$39,$BF,$1D,$E7,$FE,$2E,$FF,$18,$DC,$3D
	.dc.b	$6D,$C3,$E6,$B6,$BF,$79,$BF,$F9,$86,$BF,$0B,$E8,$FE,$3B,$DD,$FC
	.dc.b	$7B,$23,$27,$E2,$A3,$18,$7A,$7C,$7F,$C2,$F9,$7F,$15,$B5,$EE,$C7
	.dc.b	$F8,$4E,$DF,$C9,$2F,$F2,$3D,$1F,$D1,$27,$37,$C2,$7A,$5F,$13,$E3
	.dc.b	$AB,$E5,$5F,$AF,$F8,$9F,$05,$7E,$7B,$E4,$BF,$4F,$F2,$FF,$A8,$6C
	.dc.b	$64,$FD,$1C,$65,$A7,$E0,$BC,$4F,$DB,$1F,$B7,$F5,$2C,$64,$99,$7E
	.dc.b	$62,$32,$FF,$BE,$BE,$C1,$8F,$EE,$3E,$8B,$F8,$C4,$62,$FF,$67,$18
	.dc.b	$9D,$47,$BE,$BF,$0B,$5B,$7F,$39,$5B,$9D,$42,$75,$9F,$DC,$EE,$D4
	.dc.b	$A3,$F9,$CF,$D7,$BE,$C3,$F8,$CF,$80,$F7,$BF,$96,$4E,$72,$FE,$15
	.dc.b	$39,$7B,$FA,$9F,$5D,$FB,$91,$FD,$4D,$CF,$2F,$4F,$47,$F9,$B7,$D7
	.dc.b	$BE,$5B,$F9,$3F,$D8,$7F,$A5,$EF,$D3,$39,$7F,$8C,$46,$76,$F4,$47
	.dc.b	$E0,$27,$F8,$23,$FC,$77,$EC,$3F,$52,$F2,$CB,$F9,$9C,$62,$B8,$9F
	.dc.b	$AE,$A7,$A1,$FC,$E3,$EC,$1F,$27,$E6,$8C,$3F,$D3,$7D,$87,$57,$F3
	.dc.b	$EF,$B0,$79,$99,$8C,$3B,$F5,$FF,$A1,$FD,$83,$D6,$C7,$07,$F3,$38
	.dc.b	$C3,$F6,$3E,$0D,$F5,$1F,$D2,$BC,$61,$FD,$76,$33,$3F,$CE,$E3,$2F
	.dc.b	$1F,$77,$87,$C0,$ED,$B1,$D7,$8F,$FA,$AF,$D8,$7A,$A3,$07,$F4,$57
	.dc.b	$11,$F1,$8F,$F1,$4F,$0F,$0F,$9A,$FE,$DD,$5F,$18,$E0,$FF,$71,$FB
	.dc.b	$0F,$87,$1F,$D3,$FD,$3F,$E8,$17,$CE,$72,$FF,$4E,$8C,$F8,$5F,$E0
	.dc.b	$AF,$AF,$7B,$9F,$DE,$F9,$5F,$45,$FD,$F6,$73,$97,$E7,$A7,$2F,$E8
	.dc.b	$21,$F6,$1B,$FA,$5E,$0D,$9F,$D2,$FD,$87,$FB,$A9,$E1,$1F,$E0,$98
	.dc.b	$C6,$9F,$06,$F9,$D7,$0F,$3B,$FC,$1D,$F6,$0F,$43,$84,$61,$FF,$09
	.dc.b	$7C,$07,$F4,$EA,$E0,$7E,$D1,$C6,$1F,$37,$1F,$F0,$93,$F3,$E3,$3B
	.dc.b	$FF,$8B,$1B,$5F,$C7,$B1,$79,$FD,$BD,$EF,$F1,$07,$D8,$7E,$13,$1F
	.dc.b	$EF,$3F,$60,$F9,$3C,$F7,$7F,$85,$23,$19,$7F,$06,$BE,$1A,$7F,$8E
	.dc.b	$E7,$38,$7E,$06,$73,$5F,$CD,$E7,$35,$FE,$40,$9C,$D7,$F5,$86,$C2
	.dc.b	$7E,$ED,$7F,$36,$9D,$D1,$71,$7F,$8A,$3E,$C3,$FA,$FC,$F8,$63,$1E
	.dc.b	$FF,$B9,$B7,$E9,$6C,$F9,$FC,$EF,$F1,$07,$C7,$FC,$46,$D7,$BD,$C3
	.dc.b	$F0,$66,$9F,$9F,$CD,$5F,$E4,$B7,$F6,$B1,$98,$C7,$89,$96,$57,$F3
	.dc.b	$09,$FE,$F3,$E4,$7F,$7A,$FB,$07,$C7,$7F,$56,$C7,$AD,$EE,$D7,$BA
	.dc.b	$D6,$1F,$E8,$FC,$38,$D7,$F5,$E3,$61,$CB,$FC,$EE,$38,$4F,$FA,$E2
	.dc.b	$BA,$4F,$F6,$F9,$CB,$CE,$32,$F1,$BA,$FA,$8E,$D5,$E4,$8C,$C7,$8A
	.dc.b	$FC,$38,$7B,$23,$67,$52,$9D,$4A,$75,$8E,$38,$79,$23,$FB,$DE,$1E
	.dc.b	$4A,$E9,$5E,$E9,$E1,$87,$8D,$E6,$7C,$43,$FD,$AE,$53,$E1,$18,$30
	.dc.b	$7F,$8B,$BE,$C3,$F9,$FF,$C7,$70,$6C,$7F,$B8,$1A,$FD,$16,$BF,$E1
	.dc.b	$05,$F1,$BF,$DA,$E3,$83,$5E,$0D,$78,$35,$F1,$3C,$6F,$ED,$DF,$FA
	.dc.b	$EE,$7E,$AB,$5F,$97,$5E,$0C,$3E,$CB,$63,$E7,$CE,$EB,$9E,$BF,$E2
	.dc.b	$67,$F0,$F5,$B6,$E1,$E1,$AF,$F1,$59,$C4,$B5,$C5,$F4,$E1,$E0,$9E
	.dc.b	$18,$7F,$96,$27,$83,$6B,$83,$EE,$AE,$07,$B2,$33,$1F,$A6,$6E,$1E
	.dc.b	$5A,$F0,$CF,$F8,$71,$F8,$55,$FE,$D0,$6E,$9E,$49,$F8,$37,$FA,$F8
	.dc.b	$E3,$1F,$D6,$E7,$6F,$8C,$FC,$AF,$D8,$7F,$00,$7F,$C5,$2D,$7F,$B1
	.dc.b	$36,$FF,$DB,$EF,$E0,$23,$7A,$F7,$7F,$CF,$31,$BB,$FF,$2A,$B6,$33
	.dc.b	$96,$D7,$D8,$7F,$A5,$1F,$F1,$CB,$5F,$EE,$4D,$C3,$FD,$DF,$F5,$FF
	.dc.b	$D7,$A7,$7D,$1C,$E3,$F6,$78,$FF,$99,$1B,$3F,$E8,$CB,$E5,$3F,$E6
	.dc.b	$87,$FC,$37,$1E,$0F,$07,$C9,$18,$8F,$F0,$15,$3E,$C3,$35,$27,$81
	.dc.b	$FE,$50,$9D,$38,$38,$F2,$79,$4E,$0F,$4F,$8D,$F8,$5F,$61,$FD,$0A
	.dc.b	$3F,$CB,$2D,$74,$62,$76,$47,$C1,$3F,$2F,$D1,$FF,$75,$C7,$F9,$89
	.dc.b	$AF,$F8,$6D,$BA,$B4,$FF,$44,$E1,$DE,$31,$5F,$EB,$2A,$79,$1D,$49
	.dc.b	$1E,$FF,$23,$C1,$E4,$48,$F0,$3A,$7F,$C5,$5F,$61,$FE,$07,$7F,$CE
	.dc.b	$AD,$73,$E3,$7E,$0E,$3E,$AC,$7F,$B0,$BE,$C1,$EE,$FE,$1D,$B1,$FE
	.dc.b	$7E,$6B,$FE,$37,$E7,$79,$BF,$8D,$FB,$0F,$F3,$74,$FF,$A1,$1A,$FF
	.dc.b	$8F,$5B,$B7,$C0,$8F,$EC,$3F,$47,$C7,$A6,$47,$17,$F3,$18,$79,$A1
	.dc.b	$7F,$E6,$A7,$B7,$B3,$E1,$3F,$3D,$F5,$EF,$B0,$FD,$04,$7E,$37,$2F
	.dc.b	$03,$79,$7F,$D7,$6D,$BF,$F7,$1D,$FF,$9D,$1F,$F4,$EC,$7B,$9B,$A7
	.dc.b	$43,$BF,$CB,$FE,$B7,$EC,$3F,$CD,$73,$FE,$A4,$6B,$FE,$5D,$6E,$3F
	.dc.b	$07,$EC,$3F,$B7,$3F,$0E,$1E,$C7,$32,$57,$FA,$BA,$BF,$3E,$FF,$99
	.dc.b	$FE,$C3,$FD,$7A,$FF,$AC,$9A,$E7,$E3,$3F,$80,$4F,$F9,$7A,$35,$EC
	.dc.b	$FF,$03,$AF,$74,$73,$8D,$FD,$67,$F0,$7F,$B0,$FF,$47,$CF,$FA,$F9
	.dc.b	$AF,$F9,$DD,$B9,$FD,$BF,$AF,$7D,$87,$12,$57,$17,$F8,$94,$7C,$F8
	.dc.b	$7D,$1B,$E2,$7F,$8B,$36,$3F,$D9,$AD,$7F,$D0,$DF,$5B,$70,$FC,$54
	.dc.b	$DE,$5F,$F7,$AB,$68,$76,$BF,$E4,$57,$98,$C1,$FE,$DA,$7F,$6F,$5F
	.dc.b	$EF,$E6,$C7,$2A,$E1,$F6,$0F,$96,$FE,$58,$D9,$FF,$71,$B5,$FF,$4E
	.dc.b	$3F,$C9,$13,$9C,$7F,$44,$DC,$3C,$F3,$AA,$EB,$E7,$76,$F6,$BB,$FC
	.dc.b	$8C,$AE,$F1,$97,$C3,$19,$7C,$31,$F3,$3E,$92,$47,$CC,$7D,$3C,$38
	.dc.b	$8F,$F5,$96,$89,$C6,$9E,$6A,$B9,$F9,$BE,$C3,$F7,$53,$FB,$B6,$C7
	.dc.b	$D2,$AF,$EA,$10,$FB,$0A,$4C,$7E,$AD,$F1,$DF,$11,$FE,$C3,$AE,$0E
	.dc.b	$B0,$F0,$5F,$4A,$EB,$F3,$7D,$BE,$2A,$5E,$CF,$C0,$E0,$9D,$C6,$44
	.dc.b	$7F,$B8,$1C,$35,$F1,$47,$F6,$9F,$B0,$78,$B4,$93,$CF,$EC,$3F,$88
	.dc.b	$3D,$67,$6D,$9D,$BF,$ED,$F9,$DB,$FE,$E1,$9D,$B9,$D7,$FB,$C4,$FD
	.dc.b	$38,$E4,$77,$F8,$7C,$FC,$DC,$8C,$10,$EC,$E5,$F0,$47,$4F,$06,$47
	.dc.b	$A5,$EF,$F4,$F8,$8F,$34,$5E,$1F,$E7,$18,$7E,$92,$5F,$5E,$FB,$0F
	.dc.b	$F0,$73,$63,$E7,$35,$FF,$7F,$7D,$7F,$F8,$2D,$FF,$80,$11,$9B,$FE
	.dc.b	$88,$DE,$4F,$D1,$B6,$7F,$77,$5C,$6B,$77,$FE,$1D,$8F,$D7,$1F,$AF
	.dc.b	$AE,$14,$8F,$F6,$B3,$F8,$16,$BE,$4E,$7E,$5D,$1C,$D9,$1B,$72,$F1
	.dc.b	$FA,$FF,$01,$1F,$C6,$7D,$87,$E8,$9F,$3C,$65,$F8,$2F,$F8,$CE,$35
	.dc.b	$0F,$B0,$FE,$82,$FE,$55,$AB,$3F,$ED,$11,$87,$FC,$13,$F0,$8F,$95
	.dc.b	$FC,$F5,$71,$F2,$7F,$C8,$16,$F9,$EA,$62,$3F,$C8,$9F,$61,$F6,$C9
	.dc.b	$C9,$C2,$F2,$C7,$F5,$11,$F8,$1B,$FE,$B3,$C5,$FC,$62,$1F,$4F,$D0
	.dc.b	$95,$FB,$3C,$FC,$D5,$88,$F6,$3C,$79,$FC,$66,$75,$FF,$03,$61,$FE
	.dc.b	$5B,$A3,$49,$F1,$5E,$AF,$53,$FD,$77,$D5,$E3,$2D,$7F,$42,$AF,$D5
	.dc.b	$3F,$EE,$28,$F4,$4E,$A3,$AC,$59,$DF,$C5,$70,$6C,$70,$9F,$F7,$8B
	.dc.b	$DD,$A7,$77,$91,$CE,$BB,$7D,$87,$F6,$AB,$F1,$B6,$7F,$A0,$B6,$BF
	.dc.b	$AC,$BF,$E7,$59,$DF,$CD,$F2,$F9,$A7,$CA,$F8,$2D,$69,$CC,$7B,$CE
	.dc.b	$75,$C6,$B8,$56,$DB,$5C,$2F,$BD,$E2,$63,$CC,$EF,$3B,$C3,$D3,$87
	.dc.b	$59,$D6,$1D,$5B,$5B,$7C,$67,$4C,$3E,$1C,$66,$3A,$E1,$E6,$BE,$95
	.dc.b	$A8,$E1,$87,$82,$A9,$F2,$2F,$96,$1E,$1B,$F5,$CF,$8A,$F6,$F6,$3C
	.dc.b	$F8,$76,$C3,$E9,$E1,$C8,$E7,$19,$9E,$D5,$CB,$0F,$B1,$1C,$F0,$F3
	.dc.b	$3D,$CE,$A6,$EF,$A5,$7B,$63,$0F,$3B,$ED,$1C,$F0,$F1,$CE,$E3,$A5
	.dc.b	$F6,$24,$FC,$38,$CB,$C1,$A7,$D0,$29,$D4,$F5,$9E,$63,$38,$7C,$98
	.dc.b	$E5,$3F,$28,$EF,$3C,$1A,$FE,$4A,$FD,$35,$C6,$35,$5E,$5A,$DC,$E6
	.dc.b	$7C,$47,$13,$E1,$9E,$2A,$FC,$63,$FE,$26,$3E,$C1,$63,$FC,$97,$C9
	.dc.b	$98,$F5,$99,$F2,$3F,$91,$3F,$94,$6D,$FF,$04,$E5,$FF,$50,$F0,$FE
	.dc.b	$92,$73,$97,$ED,$1B,$7F,$B4,$6B,$FE,$78,$F7,$3F,$D7,$78,$7E,$F2
	.dc.b	$73,$97,$92,$BC,$19,$7E,$DF,$2E,$F1,$82,$5F,$19,$D6,$1D,$31,$F1
	.dc.b	$46,$32,$F3,$36,$F8,$CF,$F9,$2B,$0E,$B8,$73,$9F,$DC,$46,$27,$9C
	.dc.b	$63,$CB,$EC,$FF,$9A,$25,$E5,$E2,$3F,$D2,$B3,$67,$38,$FA,$E3,$06
	.dc.b	$7E,$32,$C7,$BD,$B5,$E2,$6E,$3F,$45,$B8,$FA,$5B,$3F,$E7,$AF,$B0
	.dc.b	$FF,$72,$DF,$F8,$E5,$78,$9F,$EE,$AE,$6F,$F6,$64,$6A,$BF,$A6,$9C
	.dc.b	$67,$55,$FE,$6E,$9D,$57,$09,$CD,$77,$F8,$3A,$AB,$BD,$F8,$7C,$6D
	.dc.b	$F5,$3E,$B1,$F0,$BC,$2D,$5A,$9C,$DF,$12,$61,$E6,$3B,$7C,$07,$1A
	.dc.b	$E5,$EE,$F3,$ED,$B7,$F0,$5B,$7C,$71,$46,$36,$F8,$6F,$B1,$7D,$77
	.dc.b	$AA,$BE,$2C,$63,$7D,$4B,$C2,$B1,$31,$FF,$36,$56,$DB,$5B,$6C,$F7
	.dc.b	$FA,$B7,$D8,$53,$C5,$F1,$1F,$D6,$2B,$9C,$E9,$B3,$B7,$BC,$EF,$2F
	.dc.b	$3D,$6D,$F9,$1F,$60,$FA,$AF,$77,$D6,$BE,$37,$EC,$CE,$71,$F4,$CE
	.dc.b	$5F,$F4,$14,$66,$FE,$1C,$E5,$EB,$7D,$2F,$9B,$2F,$E0,$FB,$96,$BC
	.dc.b	$BE,$1F,$C1,$3E,$17,$C9,$DA,$F7,$C5,$9F,$5C,$DE,$0F,$CD,$70,$3E
	.dc.b	$C3,$E9,$F5,$FD,$46,$E3,$F4,$1B,$87,$DA,$AE,$13,$F5,$2B,$F3,$3F
	.dc.b	$5C,$F9,$9F,$C1,$75,$BE,$BF,$31,$86,$DF,$DE,$9D,$3F,$EA,$7A,$E4
	.dc.b	$E6,$77,$8F,$E6,$67,$2F,$D6,$FB,$0C,$E9,$EC,$F3,$FF,$1B,$7F,$8E
	.dc.b	$6E,$1F,$8A,$6D,$FA,$DB,$5F,$BD,$E5,$FF,$5D,$57,$E2,$3D,$CF,$C7
	.dc.b	$7B,$DF,$8F,$64,$64,$FC,$6C,$63,$0F,$37,$8D,$F8,$6F,$2B,$E2,$B6
	.dc.b	$BD,$CD,$9D,$C7,$E2,$27,$6E,$8F,$D0,$FC,$4E,$27,$F2,$BF,$23,$FA
	.dc.b	$1F,$96,$FD,$13,$6B,$F3,$ED,$9D,$C7,$E8,$A7,$F6,$3E,$24,$8F,$D6
	.dc.b	$FA,$96,$BE,$F4,$DD,$BF,$0F,$36,$95,$3E,$C1,$97,$EC,$FE,$87,$F8
	.dc.b	$DC,$62,$BF,$63,$18,$9D,$47,$FB,$96,$75,$5F,$B0,$AD,$4E,$6B,$F7
	.dc.b	$DD,$B8,$AF,$E3,$5F,$5E,$FB,$0F,$E2,$1D,$EF,$7B,$F9,$5C,$E7,$1F
	.dc.b	$E1,$73,$97,$FD,$E9,$E9,$FF,$64,$DB,$F9,$E4,$BC,$FF,$E6,$73,$9A
	.dc.b	$F7,$D2,$B5,$1F,$EF,$DF,$AD,$7D,$87,$F2,$1F,$B0,$71,$7F,$3C,$9C
	.dc.b	$D7,$F1,$C9,$CB,$C6,$FF,$94,$7D,$87,$CC,$8C,$1F,$C8,$9C,$6F,$FC
	.dc.b	$E6,$DE,$87,$F2,$FF,$B0,$7C,$97,$78,$C3,$FD,$33,$D7,$74,$FF,$35
	.dc.b	$FB,$07,$97,$98,$C3,$BF,$5B,$F9,$D7,$D8,$7F,$3E,$DB,$E0,$9F,$80
	.dc.b	$D5,$8F,$F4,$E7,$F9,$EC,$ED,$FE,$9D,$F0,$6F,$AA,$FE,$89,$E1,$8F
	.dc.b	$EB,$F1,$9C,$3F,$9F,$C6,$5E,$F5,$F8,$E9,$7B,$9C,$BE,$07,$8B,$6B
	.dc.b	$75,$FE,$D5,$3F,$A9,$7D,$87,$C0,$8C,$DF,$CA,$E9,$FE,$F5,$39,$AF
	.dc.b	$EC,$95,$F2,$8F,$EA,$30,$F9,$AF,$EE,$B1,$99,$FE,$B7,$39,$8F,$0F
	.dc.b	$D8,$7C,$C7,$FA,$CF,$A3,$FC,$F7,$04,$7C,$99,$D5,$FF,$64,$AD,$4F
	.dc.b	$CD,$FA,$F7,$B3,$FE,$04,$E4,$7D,$17,$F8,$1A,$73,$5F,$02,$72,$FC
	.dc.b	$28,$7D,$86,$FE,$97,$83,$67,$E3,$FD,$87,$FB,$A9,$E1,$1F,$E1,$28
	.dc.b	$C6,$7F,$E0,$7B,$E7,$5C,$3C,$CF,$F0,$BF,$D8,$3C,$EE,$11,$87,$FC
	.dc.b	$37,$F1,$4F,$73,$37,$F5,$A3,$0F,$33,$11,$C0,$EB,$E7,$D8,$FB,$BF
	.dc.b	$5F,$F8,$7C,$FF,$C1,$7F,$60,$F9,$2F,$F1,$4C,$E7,$9B,$FC,$6D,$39
	.dc.b	$BE,$3D,$BF,$E1,$58,$C6,$1F,$E2,$3B,$E1,$5F,$BE,$9C,$DF,$D8,$AF
	.dc.b	$E3,$F5,$A6,$DF,$3F,$80,$F0,$BE,$2E,$3F,$F1,$17,$D8,$7F,$56,$C3
	.dc.b	$95,$B0,$56,$BD,$FE,$19,$FF,$88,$30,$E0,$FE,$BE,$B6,$E6,$7D,$FE
	.dc.b	$87,$F8,$7B,$E3,$EC,$FF,$86,$9B,$8F,$B6,$7E,$2B,$8C,$BC,$62,$7F
	.dc.b	$CA,$2F,$EE,$4C,$B8,$F8,$8C,$B2,$FF,$5F,$5F,$D0,$BC,$5F,$E8,$7F
	.dc.b	$61,$FE,$7A,$8F,$E4,$F5,$FE,$5C,$74,$23,$4B,$0F,$15,$F2,$F8,$EF
	.dc.b	$F5,$66,$16,$B9,$CD,$7F,$89,$45,$AE,$31,$D6,$B8,$B4,$E2,$53,$89
	.dc.b	$63,$8D,$F9,$23,$CE,$F9,$A3,$33,$D2,$31,$87,$E1,$5E,$15,$FC,$90
	.dc.b	$E9,$19,$9E,$07,$F2,$9A,$E7,$3F,$EC,$DB,$F5,$56,$E3,$07,$FB,$34
	.dc.b	$F9,$25,$8C,$38,$3F,$9A,$7D,$87,$F4,$2F,$8E,$E0,$D8,$FF,$6B,$35
	.dc.b	$FF,$6E,$35,$E3,$F1,$DF,$4D,$DB,$5D,$B5,$DE,$1D,$A3,$86,$5F,$73
	.dc.b	$2F,$ED,$55,$F5,$DA,$FD,$1B,$57,$09,$E7,$3F,$55,$DD,$73,$D7,$FA
	.dc.b	$83,$F8,$8B,$E3,$87,$E0,$DB,$7C,$6B,$77,$FD,$5C,$E6,$76,$BE,$58
	.dc.b	$78,$6B,$FC,$37,$3A,$6D,$72,$7F,$C3,$F7,$DA,$F8,$E1,$ED,$AD,$57
	.dc.b	$FB,$69,$B5,$BB,$F1,$47,$E0,$74,$B5,$FD,$62,$7E,$2D,$FF,$06,$BE
	.dc.b	$13,$E2,$35,$7C,$5C,$1F,$EB,$B8,$DA,$FB,$0F,$DE,$3F,$D9,$1A,$FF
	.dc.b	$B0,$5B,$5F,$C5,$1F,$A7,$1B,$D7,$B3,$CD,$B1,$C3,$1F,$F0,$13,$C2
	.dc.b	$7F,$D9,$9F,$61,$FC,$FD,$FE,$E6,$D7,$3D,$3D,$3E,$BF,$FA,$AC,$6A
	.dc.b	$31,$A4,$8C,$1C,$BC,$2F,$1D,$FF,$86,$A3,$07,$F8,$02,$DF,$61,$AA
	.dc.b	$93,$A3,$FD,$C1,$3A,$71,$A7,$F3,$48,$FE,$67,$1C,$F9,$F8,$BC,$3A
	.dc.b	$7E,$1F,$D8,$6A,$3F,$DD,$4D,$BE,$53,$AA,$95,$FE,$14,$AD,$6D,$FE
	.dc.b	$29,$C3,$FD,$91,$3F,$E1,$8F,$43,$FD,$C4,$DF,$92,$FA,$31,$88,$EF
	.dc.b	$86,$3C,$99,$1C,$5F,$99,$E4,$F0,$F2,$64,$F5,$76,$FF,$C0,$1F,$B0
	.dc.b	$FE,$F0,$FF,$8A,$1A,$E7,$C5,$FC,$0C,$7D,$9D,$FF,$AC,$FD,$83,$E1
	.dc.b	$BF,$0A,$DB,$FF,$18,$B5,$FE,$79,$BF,$F9,$1F,$A3,$FC,$55,$1F,$E3
	.dc.b	$6F,$B0,$FF,$33,$B7,$E0,$3F,$1F,$18,$8E,$27,$FA,$3A,$5F,$47,$AE
	.dc.b	$F9,$3C,$DF,$CB,$CF,$9F,$0F,$F2,$B5,$FF,$99,$AB,$C1,$ED,$F6,$3F
	.dc.b	$3B,$F5,$EF,$B0,$FC,$FC,$7E,$36,$7C,$27,$F9,$BB,$AB,$27,$0F,$77
	.dc.b	$7E,$B4,$9F,$F2,$8D,$7B,$9E,$38,$3E,$63,$F5,$7F,$61,$FE,$63,$9F
	.dc.b	$F2,$BB,$5E,$ED,$E3,$FF,$03,$FD,$87,$F9,$BE,$73,$ED,$F2,$23,$FE
	.dc.b	$60,$8F,$CE,$3E,$8F,$B0,$FF,$59,$BF,$E6,$46,$B9,$F8,$DF,$DE,$CF
	.dc.b	$F9,$3A,$35,$AF,$F0,$0A,$FF,$45,$C6,$0E,$DF,$59,$FC,$13,$EC,$3F
	.dc.b	$D1,$31,$FE,$70,$6B,$F2,$1B,$97,$F8,$83,$EB,$DF,$61,$C4,$95,$C5
	.dc.b	$FE,$1F,$1F,$46,$7F,$C4,$FF,$47,$FE,$AF,$AF,$F3,$EB,$5F,$AA,$D8
	.dc.b	$FF,$16,$FD,$75,$E7,$FF,$4B,$4D,$AF,$F6,$BB,$53,$67,$0D,$F1,$7F
	.dc.b	$90,$C7,$EC,$6F,$C7,$F1,$9F,$C9,$3E,$C1,$F2,$D4,$98,$FF,$A4,$A3
	.dc.b	$FC,$11,$97,$FA,$9D,$DC,$66,$76,$FE,$3A,$36,$FF,$BB,$1F,$F2,$E3
	.dc.b	$CF,$E4,$FC,$07,$87,$DD,$91,$92,$1C,$DF,$E7,$3A,$3E,$92,$92,$7C
	.dc.b	$DC,$35,$F9,$71,$74,$FF,$2B,$E2,$AC,$67,$FE,$C1,$FA,$FF,$F2,$4B
	.dc.b	$4C,$E0,$10,$FB,$0F,$E9,$6D,$8F,$F5,$BB,$5F,$D7,$AD,$F7,$32,$76
	.dc.b	$F2,$8E,$7F,$35,$D3,$E3,$65,$74,$8C,$BD,$A3,$2F,$68,$FE,$BB,$F6
	.dc.b	$1F,$C3,$23,$FD,$92,$D7,$F8,$0B,$74,$FF,$3C,$FD,$87,$F0,$79,$DF
	.dc.b	$8F,$C8,$8F,$FB,$4A,$34,$F2,$8F,$ED,$71,$F4,$E3,$5B,$F0,$7F,$87
	.dc.b	$C3,$E1,$F7,$F2,$F2,$7A,$47,$23,$0F,$08,$C3,$E1,$F0,$7B,$91,$F2
	.dc.b	$74,$57,$BC,$5C,$BD,$B3,$FE,$00,$FA,$F7,$D8,$52,$5F,$FB,$BB,$EC
	.dc.b	$3F,$BC,$BF,$C9,$A7,$6F,$FB,$C9,$AE,$67,$FC,$17,$3C,$2F,$FD,$47
	.dc.b	$5A,$7F,$D4,$B1,$FA,$2A,$36,$75,$C9,$F8,$D9,$CF,$47,$42,$3E,$5E
	.dc.b	$9A,$FF,$3A,$BF,$D5,$B6,$7C,$73,$E4,$FB,$09,$3E,$D7,$FC,$43,$F6
	.dc.b	$1F,$9E,$8F,$CE,$5A,$77,$1D,$5D,$E1,$FE,$28,$FB,$07,$8D,$49,$5F
	.dc.b	$A5,$D3,$FD,$99,$F0,$AF,$95,$74,$6A,$BF,$A8,$E0,$F9,$EE,$86,$70
	.dc.b	$FF,$65,$FD,$87,$F1,$5C,$FE,$23,$6F,$F6,$4D,$57,$FD,$2E,$B8,$F1
	.dc.b	$FF,$8F,$AB,$46,$E7,$57,$D3,$E9,$74,$4C,$3D,$51,$FB,$FD,$3D,$4E
	.dc.b	$63,$FB,$4D,$B9,$BB,$19,$CB,$D3,$E9,$F8,$49,$5D,$0F,$4F,$C5,$E1
	.dc.b	$7C,$5F,$EA,$56,$16,$30,$DE,$CF,$87,$7F,$EC,$67,$E2,$96,$BE,$2B
	.dc.b	$5E,$6D,$9F,$F3,$3D,$74,$8C,$12,$7F,$DC,$0F,$66,$9F,$C8,$7E,$C3
	.dc.b	$FB,$1C,$F9,$DB,$3F,$99,$6C,$EF,$E2,$3C,$51,$88,$2F,$9A,$7C,$A7
	.dc.b	$69,$CC,$61,$F8,$B7,$D7,$E7,$39,$CE,$71,$F1,$36,$7C,$4D,$AF,$45
	.dc.b	$70,$BE,$D8,$78,$5B,$5B,$AF,$2D,$6D,$F0,$9F,$06,$35,$5C,$63,$27
	.dc.b	$03,$9D,$47,$31,$BC,$3E,$45,$F5,$7D,$35,$B7,$35,$70,$F5,$CD,$8F
	.dc.b	$59,$4F,$8E,$DB,$F9,$C7,$79,$E3,$3E,$2C,$3D,$38,$78,$67,$38,$78
	.dc.b	$AB,$CF,$39,$7E,$06,$1D,$63,$11,$E0,$3A,$47,$CC,$8C,$E1,$E1,$9E
	.dc.b	$27,$43,$CD,$87,$CA,$3C,$6E,$2B,$B3,$4E,$A7,$12,$DF,$9C,$98,$7E
	.dc.b	$16,$38,$46,$1F,$8E,$70,$8F,$79,$1C,$C6,$DF,$63,$6F,$E5,$DF,$9E
	.dc.b	$FC,$B5,$CC,$F6,$D6,$F0,$F8,$35,$B9,$F4,$4F,$CD,$7F,$CC,$17,$F8
	.dc.b	$96,$F9,$5F,$C8,$DF,$F3,$21,$F3,$4C,$F2,$FE,$62,$35,$5F,$C1,$FA
	.dc.b	$3F,$D3,$1C,$59,$A9,$7F,$90,$BD,$E9,$FD,$36,$F7,$39,$AD,$E1,$FB
	.dc.b	$BB,$DC,$FF,$87,$EB,$73,$9A,$FD,$B4,$62,$FF,$80,$5F,$0B,$FE,$8B
	.dc.b	$EF,$31,$EB,$5E,$18,$C4,$F8,$4F,$D4,$DF,$37,$F8,$75,$F3,$8F,$F2
	.dc.b	$9D,$F8,$3E,$2F,$8E,$5F,$07,$CD,$B9,$7B,$66,$E3,$E9,$9A,$FF,$58
	.dc.b	$6E,$0F,$8E,$B1,$86,$E3,$EE,$6D,$79,$9B,$8F,$D1,$6E,$3F,$DE,$7E
	.dc.b	$C1,$F0,$DE,$EE,$7F,$7E,$09,$FF,$54,$63,$FD,$0E,$31,$5B,$8E,$5E
	.dc.b	$35,$9F,$F2,$85,$74,$9E,$0F,$28,$FF,$5D,$B9,$6C,$F1,$7C,$1E,$3F
	.dc.b	$3E,$B7,$D5,$BE,$27,$58,$7F,$8D,$71,$FE,$AD,$5A,$9C,$DC,$9F,$0F
	.dc.b	$C1,$5C,$7C,$3E,$F7,$2E,$EE,$6D,$AF,$3B,$67,$71,$EA,$9F,$76,$CF
	.dc.b	$AC,$F6,$3F,$25,$B3,$FE,$0D,$5E,$1F,$07,$36,$AE,$5B,$E4,$D5,$5B
	.dc.b	$FE,$E5,$2F,$A9,$7D,$87,$89,$B3,$AF,$8B,$FE,$AA,$F0,$8C,$E3,$FE
	.dc.b	$FA,$BE,$57,$C4,$F5,$DF,$09,$D5,$70,$AF,$C1,$FD,$83,$EA,$BE,$0F
	.dc.b	$D8,$3E,$4B,$EC,$4E,$73,$F3,$CE,$5F,$C4,$46,$71,$F1,$4E,$5F,$F7
	.dc.b	$EC,$67,$0F,$C9,$4E,$5F,$0E,$1D,$8F,$95,$EE,$DC,$3D,$DE,$27,$C3
	.dc.b	$3E,$27,$8D,$63,$24,$CB,$DB,$19,$7E,$27,$0B,$EC,$3E,$B7,$5F,$D7
	.dc.b	$6D,$FD,$29,$D3,$67,$4D,$3F,$4D,$F6,$0F,$9A,$FC,$3F,$6B,$EB,$FF
	.dc.b	$3D,$39,$AF,$BB,$39,$78,$D7,$0A,$DB,$63,$24,$CB,$EB,$FD,$86,$A3
	.dc.b	$CD,$E7,$FE,$3E,$FF,$20,$D8,$C9,$EB,$8C,$B4,$FC,$47,$A3,$F8,$D3
	.dc.b	$F1,$DE,$ED,$8C,$93,$0F,$C3,$C6,$5F,$CE,$47,$FA,$17,$3F,$E4,$71
	.dc.b	$F8,$0F,$2B,$E0,$F8,$FF,$8A,$6D,$7E,$21,$B3,$B8,$F9,$73,$8E,$5F
	.dc.b	$D7,$7C,$4D,$9F,$F3,$D7,$53,$E4,$FF,$49,$F1,$DF,$A5,$6C,$FE,$8D
	.dc.b	$B8,$FE,$79,$B8,$7E,$E7,$C4,$FD,$B1,$FB,$7F,$52,$DF,$E0,$62,$DF
	.dc.b	$E2,$26,$D7,$F2,$8F,$B0,$67,$FB,$5F,$A2,$FE,$39,$18,$C7,$F6,$91
	.dc.b	$89,$D4,$7F,$2D,$9D,$61,$FB,$2A,$D4,$E6,$BF,$80,$76,$E2,$BE,$79
	.dc.b	$FE,$FF,$FA,$F7,$D8,$7F,$16,$EF,$7B,$DF,$CB,$67,$38,$FF,$0C,$9C
	.dc.b	$BF,$CF,$7D,$4F,$ED,$FA,$FF,$43,$9F,$EA,$13,$9E,$4E,$9E,$87,$F3
	.dc.b	$7F,$AE,$7C,$B7,$F2,$4F,$B0,$FF,$47,$5F,$F4,$79,$CB,$FC,$6A,$33
	.dc.b	$B7,$A0,$4B,$1F,$E5,$BF,$60,$9E,$D1,$87,$C1,$E7,$FF,$32,$FB,$07
	.dc.b	$C9,$E3,$2F,$E4,$F1,$87,$FA,$C7,$AE,$EB,$FE,$73,$F6,$0F,$2F,$FA
	.dc.b	$84,$E6,$BF,$97,$CE,$5E,$3E,$AF,$F4,$0F,$B0,$7A,$BC,$23,$0F,$F5
	.dc.b	$9F,$83,$7D,$4F,$F4,$6F,$18,$65,$C3,$BF,$73,$5F,$01,$72,$FE,$7F
	.dc.b	$3F,$E8,$F7,$FB,$85,$FC,$37,$FA,$1F,$D8,$79,$F8,$FE,$54,$E5,$FE
	.dc.b	$7D,$19,$EB,$FE,$F7,$5F,$0E,$73,$F3,$7F,$DB,$E3,$13,$FE,$01,$A4
	.dc.b	$FF,$73,$FB,$0F,$85,$7F,$D2,$7D,$2F,$E6,$98,$7F,$C0,$19,$CE,$1F
	.dc.b	$E0,$99,$CD,$7D,$9F,$AE,$7B,$5F,$DF,$79,$9F,$45,$FD,$FE,$73,$97
	.dc.b	$F8,$36,$72,$FF,$83,$E1,$F6,$1B,$FA,$4E,$0D,$9F,$F0,$A7,$D8,$3D
	.dc.b	$5C,$C6,$1B,$9F,$E3,$E7,$FC,$10,$FE,$47,$EC,$3F,$84,$57,$2F,$43
	.dc.b	$FC,$25,$18,$9F,$CB,$7C,$53,$DD,$FF,$0B,$61,$CB,$87,$FC,$4F,$39
	.dc.b	$C7,$91,$FE,$27,$71,$7C,$AF,$AF,$A1,$FE,$1F,$9C,$DF,$F8,$C3,$EB
	.dc.b	$BF,$13,$87,$F8,$03,$EC,$1F,$27,$FE,$25,$9C,$E5,$FD,$EE,$73,$5F
	.dc.b	$E3,$59,$CD,$FF,$8E,$E7,$35,$FA,$F9,$CD,$7E,$B6,$95,$A7,$C5,$DB
	.dc.b	$DC,$FF,$0F,$C7,$DC,$B5,$F2,$7F,$6C,$B8,$7F,$C4,$BF,$61,$FD,$76
	.dc.b	$7C,$11,$8F,$7B,$D4,$DC,$BD,$0D,$9F,$37,$9D,$FE,$14,$F9,$0F,$80
	.dc.b	$DA,$F7,$38,$7E,$04,$D3,$F8,$66,$8A,$FF,$26,$3F,$B8,$9F,$F2,$D3
	.dc.b	$96,$F8,$5A,$99,$1F,$E5,$FA,$FE,$25,$E3,$7F,$13,$FB,$07,$C7,$7F
	.dc.b	$15,$9F,$F1,$D5,$FF,$96,$5D,$89,$DB,$F3,$E3,$FD,$17,$18,$8E,$55
	.dc.b	$D6,$BA,$63,$E5,$9E,$51,$FE,$AE,$9D,$47,$3C,$2B,$C8,$D9,$92,$99
	.dc.b	$BF,$05,$F5,$7C,$51,$E8,$73,$87,$9B,$0F,$04,$F2,$AE,$11,$8A,$DD
	.dc.b	$F8,$AB,$94,$FA,$AF,$94,$7C,$71,$87,$A3,$0E,$71,$98,$FF,$65,$BE
	.dc.b	$B2,$B8,$7E,$69,$CF,$EC,$3F,$A0,$FC,$6F,$06,$C7,$FB,$51,$AE,$FC
	.dc.b	$D4,$EF,$E3,$7F,$CE,$53,$B6,$BD,$5A,$FD,$FA,$46,$23,$FA,$B4,$63
	.dc.b	$0F,$A7,$97,$FA,$DE,$6E,$DF,$87,$D7,$F0,$17,$B9,$FA,$AD,$8F,$C2
	.dc.b	$57,$19,$FB,$2D,$8E,$38,$7A,$1B,$87,$6A,$87,$49,$D5,$7A,$5B,$1F
	.dc.b	$DB,$63,$A1,$F9,$B2,$BE,$C8,$DA,$BF,$34,$EF,$0F,$63,$67,$F2,$73
	.dc.b	$A7,$B4,$FE,$73,$5D,$47,$F8,$06,$BD,$D8,$7F,$B4,$2F,$84,$FE,$BA
	.dc.b	$F8,$3C,$8F,$F6,$95,$70,$BF,$65,$FE,$8B,$EC,$3F,$80,$57,$FB,$5D
	.dc.b	$AF,$F8,$3D,$B1,$FE,$DE,$7E,$DC,$6F,$6E,$EA,$37,$5F,$BF,$6C,$7F
	.dc.b	$85,$5E,$2F,$F8,$6F,$EC,$3F,$A3,$3F,$EE,$76,$B9,$EA,$EB,$F5,$DF
	.dc.b	$D7,$23,$51,$8C,$A4,$60,$E5,$E1,$79,$2F,$FC,$35,$18,$3F,$DE,$96
	.dc.b	$FB,$0D,$54,$9D,$1F,$E2,$F9,$D3,$8D,$3F,$DF,$F1,$FC,$E6,$39,$F2
	.dc.b	$F1,$78,$7C,$77,$C4,$FB,$0D,$47,$F8,$E5,$B7,$CA,$75,$52,$BF,$A7
	.dc.b	$6D,$FC,$9E,$F6,$FF,$29,$C3,$FD,$95,$3A,$3F,$C8,$1E,$87,$FB,$A1
	.dc.b	$C1,$FF,$00,$7E,$47,$EA,$C6,$63,$C5,$86,$3C,$A9,$1C,$25,$1F,$43
	.dc.b	$CA,$E3,$E5,$4A,$D3,$C1,$FF,$2A,$7D,$87,$FA,$2D,$F0,$B5,$CF,$C3
	.dc.b	$7E,$1A,$3E,$EE,$ED,$7F,$A2,$7D,$83,$DF,$FC,$5B,$70,$F5,$B5,$FF
	.dc.b	$31,$74,$7F,$8C,$1F,$7F,$D8,$7F,$9B,$1B,$F0,$1F,$94,$8C,$47,$0F
	.dc.b	$A2,$C7,$74,$8E,$2F,$E6,$AF,$CB,$97,$F5,$7B,$7B,$BC,$C9,$7F,$3B
	.dc.b	$EB,$DF,$61,$FA,$29,$FC,$7C,$FF,$9D,$DB,$B7,$BE,$3F,$CE,$D1,$8C
	.dc.b	$3F,$B2,$5F,$F9,$D2,$7E,$CC,$FB,$5E,$5E,$B7,$53,$F5,$F2,$F9,$8F
	.dc.b	$D8,$FD,$87,$F9,$AA,$BF,$02,$D7,$FD,$10,$DC,$7C,$50,$FB,$0F,$E9
	.dc.b	$D1,$89,$7B,$DD,$09,$5F,$8C,$AF,$D1,$BF,$E9,$3F,$B0,$FF,$5C,$BF
	.dc.b	$90,$6B,$9F,$8D,$FE,$0B,$3F,$E5,$C8,$D7,$97,$FC,$22,$BE,$04,$73
	.dc.b	$8D,$FD,$67,$F0,$CF,$B0,$FF,$48,$4F,$E7,$1A,$FF,$A8,$DB,$9F,$B3
	.dc.b	$EB,$DF,$61,$C4,$95,$C5,$FE,$2F,$1F,$46,$3F,$D5,$1F,$45,$FE,$AB
	.dc.b	$7F,$52,$D7,$FD,$60,$DA,$F9,$3F,$5B,$79,$3F,$C3,$93,$6B,$9F,$C5
	.dc.b	$FF,$24,$8C,$42,$AC,$61,$FD,$A5,$F4,$F1,$78,$5F,$F2,$CF,$B0,$7C
	.dc.b	$C5,$26,$3F,$BE,$8F,$F0,$AE,$5F,$E2,$D7,$71,$99,$DB,$FE,$C9,$8D
	.dc.b	$BF,$EC,$77,$FC,$CE,$F3,$F9,$3F,$01,$E1,$F7,$64,$64,$87,$37,$DD
	.dc.b	$F4,$92,$3E,$4B,$E9,$E1,$C4,$7C,$A8,$BA,$7E,$4B,$05,$63,$2F,$F2
	.dc.b	$37,$D8,$7E,$5A,$BF,$2E,$D8,$F9,$F7,$FD,$42,$1F,$61,$49,$8F,$F2
	.dc.b	$28,$FA,$0D,$8C,$BF,$EE,$B8,$E2,$EF,$7E,$9F,$3B,$D7,$E3,$64,$F3
	.dc.b	$8C,$BD,$63,$2F,$38,$FE,$C7,$F6,$0F,$16,$92,$7F,$9A,$FD,$87,$F0
	.dc.b	$87,$F9,$BC,$ED,$B3,$B7,$FE,$00,$4E,$D6,$BF,$BA,$AE,$11,$AD,$3F
	.dc.b	$B9,$D1,$E5,$5F,$C0,$21,$F1,$3A,$F9,$D9,$7E,$48,$E2,$61,$F0,$C6
	.dc.b	$1F,$4F,$83,$D8,$8F,$B3,$9E,$BE,$88,$B8,$78,$67,$FC,$0F,$F5,$EF
	.dc.b	$B0,$A4,$BF,$EB,$3F,$5F,$FD,$FD,$F7,$F2,$7F,$A2,$6B,$84,$7F,$5E
	.dc.b	$6B,$A0,$1C,$E1,$FC,$82,$BA,$3F,$C8,$A3,$35,$FE,$95,$9F,$D2,$D1
	.dc.b	$B3,$A8,$EB,$1F,$A8,$7C,$06,$F9,$FF,$C4,$71,$9B,$F0,$74,$62,$65
	.dc.b	$7B,$9D,$C7,$0F,$AF,$93,$F0,$9F,$F1,$6F,$D8,$7E,$86,$3C,$31,$A7
	.dc.b	$C3,$82,$37,$97,$F8,$D3,$EC,$3F,$9D,$D7,$F8,$11,$AF,$A1,$B9,$7F
	.dc.b	$55,$F8,$57,$CB,$39,$F9,$79,$1C,$FE,$7E,$63,$CE,$3F,$C8,$3F,$61
	.dc.b	$FC,$6B,$3F,$F5,$F3,$67,$E9,$B5,$67,$FD,$42,$77,$C7,$C8,$93,$E7
	.dc.b	$3A,$FC,$A4,$7F,$C3,$B6,$8F,$F2,$FE,$9E,$8B,$D5,$7D,$4A,$73,$79
	.dc.b	$8C,$ED,$FD,$8A,$3F,$CA,$B2,$69,$3E,$2F,$DB,$E9,$FF,$B4,$3A,$FD
	.dc.b	$A5,$AF,$6B,$5E,$51,$6B,$F6,$51,$FE,$E4,$9E,$2D,$38,$94,$E2,$58
	.dc.b	$E2,$FF,$B0,$A3,$8B,$4E,$25,$38,$97,$EC,$3F,$B3,$FC,$67,$26,$CF
	.dc.b	$FB,$CD,$B3,$FA,$87,$FC,$E3,$3B,$F9,$DF,$64,$3E,$6D,$F2,$9E,$78
	.dc.b	$C3,$67,$73,$AB,$F1,$36,$FC,$4D,$C3,$C5,$5C,$2F,$A4,$E6,$FC,$8D
	.dc.b	$58,$7A,$1B,$19,$8F,$1C,$78,$27,$C8,$F4,$63,$93,$E3,$57,$85,$C6
	.dc.b	$1E,$69,$E3,$18,$C3,$E4,$61,$DC,$F0,$5F,$1B,$E1,$5B,$3E,$64,$66
	.dc.b	$B8,$DF,$8A,$BE,$75,$59,$CB,$C4,$F7,$D7,$17,$E1,$D6,$DE,$A5,$3E
	.dc.b	$1B,$70,$F8,$C7,$C4,$BF,$29,$2F,$E8,$D6,$F0,$F6,$5F,$C0,$AD,$4F
	.dc.b	$80,$F2,$5F,$4A,$DC,$63,$0F,$33,$E2,$30,$75,$BF,$24,$6A,$BC,$98
	.dc.b	$7C,$0C,$3E,$1D,$F6,$C3,$AC,$60,$F1,$C7,$56,$D7,$8F,$0E,$D3,$F7
	.dc.b	$0B,$3A,$C3,$C1,$7D,$30,$F8,$35,$F8,$33,$FA,$49,$EF,$29,$F9,$0F
	.dc.b	$2D,$CB,$88,$D1,$FA,$B8,$CB,$FB,$16,$E1,$FB,$AE,$5F,$F4,$F7,$0F
	.dc.b	$EA,$67,$38,$71,$C3,$07,$F0,$3A,$DE,$1C,$E3,$1E,$E7,$FA,$2F,$1F
	.dc.b	$DE,$61,$CA,$BF,$B0,$DF,$09,$CD,$F0,$C7,$AD,$7F,$07,$AD,$63,$E1
	.dc.b	$8C,$79,$7D,$63,$E1,$CB,$E2,$B1,$5D,$AA,$BF,$E8,$88,$B3,$FE,$58
	.dc.b	$BF,$55,$63,$E3,$72,$D8,$C9,$EF,$8C,$B4,$CD,$F9,$E3,$2F,$F1,$FF
	.dc.b	$B0,$FF,$6B,$CF,$FB,$81,$FF,$52,$37,$E1,$3F,$D6,$1C,$DF,$EA,$09
	.dc.b	$CD,$7F,$78,$78,$B7,$C5,$DD,$FE,$0A,$73,$5C,$67,$35,$FE,$0E,$9F
	.dc.b	$F1,$5D,$7F,$BC,$FA,$5F,$56,$F8,$6D,$DA,$F6,$FF,$85,$EB,$73,$A9
	.dc.b	$CE,$3E,$7F,$23,$1F,$01,$97,$5E,$F6,$FB,$6E,$9F,$05,$B9,$FB,$5B
	.dc.b	$87,$C1,$7D,$2B,$EB,$3E,$1B,$F1,$9B,$5F,$E2,$A8,$D7,$83,$C2,$73
	.dc.b	$7F,$E5,$26,$CF,$3D,$F5,$5B,$7B,$FD,$4B,$EC,$29,$E2,$F8,$9F,$E9
	.dc.b	$79,$77,$3C,$13,$B7,$19,$7F,$98,$23,$0F,$F8,$FF,$14,$61,$FB,$F5
	.dc.b	$D2,$75,$39,$AF,$DE,$FD,$83,$EA,$BE,$1F,$D8,$3E,$43,$EB,$CE,$6B
	.dc.b	$C3,$92,$B5,$1E,$4C,$55,$B7,$C9,$3A,$7F,$C8,$B1,$9A,$FC,$9C,$E5
	.dc.b	$FF,$79,$C6,$6B,$CB,$87,$CF,$F7,$3C,$6F,$CA,$F0,$E4,$F9,$33,$7C
	.dc.b	$A6,$33,$F7,$D5,$CB,$F4,$35,$AB,$E1,$3F,$6F,$EC,$3E,$E7,$5D,$9F
	.dc.b	$7F,$1F,$E8,$1F,$C1,$29,$DC,$69,$B8,$FF,$98,$BE,$B9,$F3,$5F,$8E
	.dc.b	$ED,$7D,$7F,$E8,$A7,$38,$7C,$C9,$CB,$F2,$E3,$37,$F7,$67,$2F,$5B
	.dc.b	$E9,$7C,$D9,$7F,$0F,$EC,$2D,$7E,$17,$2F,$F2,$34,$FB,$2F,$F3,$5E
	.dc.b	$7F,$E6,$DB,$19,$3E,$0C,$65,$F9,$DE,$8F,$E6,$8F,$CD,$FB,$B6,$32
	.dc.b	$4C,$3F,$1F,$19,$7E,$27,$89,$F8,$EF,$2F,$E1,$36,$BE,$03,$67,$71
	.dc.b	$F9,$0E,$8F,$E8,$3C,$88,$FE,$AB,$C1,$FC,$B6,$B6,$7B,$FE,$27,$AC
	.dc.b	$FE,$17,$E4,$BF,$63,$F2,$FF,$B2,$6C,$64,$FC,$54,$65,$A7,$E0,$BC
	.dc.b	$3F,$E0,$07,$F0,$1F,$52,$C6,$49,$87,$CF,$8C,$BF,$EA,$4F,$B0,$63
	.dc.b	$FC,$0B,$E8,$BF,$91,$C6,$2F,$F3,$51,$89,$D4,$7E,$B6,$75,$5F,$BA
	.dc.b	$AD,$4E,$6B,$F8,$6F,$6E,$2B,$FD,$67,$F5,$EF,$B0,$FE,$3F,$E0,$3D
	.dc.b	$EF,$E6,$73,$9C,$7F,$7B,$39,$7F,$79,$E9,$7D,$07,$F9,$14,$3C,$FF
	.dc.b	$E7,$73,$9C,$7F,$DE,$73,$9A,$FE,$09,$F5,$EF,$96,$FE,$5F,$F6,$1F
	.dc.b	$E9,$29,$CC,$63,$87,$FA,$2D,$2F,$F9,$AF,$D8,$7F,$B6,$AF,$9C,$63
	.dc.b	$1D,$F9,$FF,$CE,$BE,$C1,$F2,$58,$BF,$E2,$F1,$87,$F8,$B7,$B2,$E8
	.dc.b	$FE,$83,$F6,$0F,$2B,$FA,$AC,$E6,$BF,$8F,$CE,$5E,$3E,$B7,$F4,$8F
	.dc.b	$B0,$7A,$BC,$23,$0F,$F5,$EF,$83,$7D,$57,$F4,$EF,$18,$65,$C3,$BF
	.dc.b	$6F,$5E,$FD,$CB,$F9,$7C,$FF,$A3,$DF,$E5,$D7,$F0,$F8,$BF,$A0,$7D
	.dc.b	$87,$9D,$FF,$78,$C6,$7C,$6F,$EF,$F3,$97,$F5,$13,$F2,$F8,$FF,$B0
	.dc.b	$CB,$E6,$FF,$BA,$E1,$F1,$EB,$AC,$7F,$78,$FB,$0F,$85,$1F,$D3,$FD
	.dc.b	$4F,$E7,$15,$FE,$46,$9C,$E3,$FE,$0E,$9C,$D0,$7D,$83,$DA,$FF,$01
	.dc.b	$F3,$BE,$8B,$FC,$0B,$39,$CB,$FA,$64,$E5,$FE,$9D,$0F,$B0,$DF,$D2
	.dc.b	$70,$6C,$FF,$53,$FA,$FF,$E3,$F9,$34,$F8,$2F,$8A,$BF,$C2,$73,$A9
	.dc.b	$CE,$1D,$6F,$97,$9B,$FE,$18,$FB,$07,$9F,$C2,$30,$FF,$87,$3E,$27
	.dc.b	$FA,$66,$7D,$EF,$FC,$D3,$97,$F8,$69,$FF,$35,$CE,$B2,$F1,$CE,$6B
	.dc.b	$E3,$4E,$6B,$7E,$F7,$87,$D0,$EC,$FF,$8A,$BE,$BF,$F1,$B3,$FD,$D3
	.dc.b	$EC,$1F,$27,$9E,$EF,$F0,$EC,$62,$7F,$BB,$57,$0C,$BF,$C1,$13,$97
	.dc.b	$D0,$7D,$78,$D6,$9F,$E4,$89,$D3,$F2,$A3,$35,$FE,$86,$6D,$7C,$FA
	.dc.b	$FC,$0A,$8C,$B7,$8B,$FC,$63,$F6,$1F,$D7,$B0,$E9,$18,$F7,$FD,$CD
	.dc.b	$C7,$D2,$DB,$F3,$9D,$7D,$0F,$F1,$57,$C8,$7C,$56,$D7,$C0,$70,$FC
	.dc.b	$29,$A7,$2C,$15,$FE,$56,$7F,$7B,$3F,$E1,$D7,$2D,$F0,$F5,$32,$3F
	.dc.b	$C4,$55,$F9,$FF,$1F,$F4,$1F,$61,$FE,$87,$8F,$F2,$55,$7F,$98,$9D
	.dc.b	$08,$D3,$F9,$78,$FD,$BC,$62,$36,$EB,$0F,$1C,$71,$F8,$DF,$F5,$95
	.dc.b	$F8,$1F,$F5,$CC,$EA,$B9,$52,$B4,$F9,$30,$AF,$FB,$02,$2D,$71,$BE
	.dc.b	$B3,$DC,$E0,$53,$81,$63,$83,$FE,$C8,$9E,$78,$7B,$2B,$9E,$1E,$B8
	.dc.b	$C5,$6C,$E4,$E1,$EA,$7F,$B3,$CF,$51,$4A,$66,$B5,$F6,$1F,$D1,$3E
	.dc.b	$3A,$F8,$86,$92,$BF,$DB,$52,$F8,$DF,$E1,$11,$86,$BF,$EE,$28,$D6
	.dc.b	$7D,$A7,$33,$FE,$E8,$9C,$BF,$17,$3F,$ED,$0F,$CE,$6B,$F2,$63,$79
	.dc.b	$FE,$09,$B3,$FE,$F2,$37,$1F,$86,$69,$DE,$BC,$F1,$C5,$B1,$C7,$6F
	.dc.b	$F7,$DC,$F1,$6C,$71,$C3,$F0,$37,$E0,$C3,$F9,$C4,$F6,$6D,$70,$7F
	.dc.b	$9E,$4E,$D3,$67,$84,$7F,$B9,$E7,$83,$AC,$3C,$AF,$43,$FD,$15,$3F
	.dc.b	$12,$FF,$9B,$B7,$5F,$34,$FC,$BB,$FC,$EC,$7B,$63,$8C,$61,$F2,$DF
	.dc.b	$65,$F6,$1F,$BE,$7F,$D3,$0D,$7F,$85,$B7,$0F,$E8,$EF,$E0,$E3,$7B
	.dc.b	$75,$79,$B6,$38,$61,$FD,$7E,$7B,$DE,$FE,$C3,$FD,$2B,$1F,$EA,$76
	.dc.b	$BF,$C6,$9B,$87,$F5,$6F,$AE,$FE,$BD,$3B,$E9,$F0,$47,$F0,$1C,$3B
	.dc.b	$36,$7F,$B7,$5F,$29,$F1,$3F,$E1,$F8,$F0,$78,$5E,$33,$9F,$D8,$58
	.dc.b	$C9,$0F,$E5,$B6,$8D,$46,$39,$7C,$66,$32,$F4,$7C,$77,$C5,$FB,$04
	.dc.b	$D3,$D3,$16,$BF,$D9,$13,$A6,$CE,$9F,$F6,$67,$A0,$F9,$2F,$A0,$D8
	.dc.b	$CC,$73,$7F,$A1,$C3,$CA,$D9,$23,$FD,$AD,$9F,$F8,$45,$EB,$E4,$F4
	.dc.b	$23,$E0,$9F,$BD,$F6,$1F,$E0,$08,$FF,$6F,$35,$FE,$A2,$DD,$3F,$C1
	.dc.b	$AF,$D1,$8F,$F1,$07,$D8,$3E,$27,$F1,$0D,$8F,$F7,$43,$5F,$EB,$7C
	.dc.b	$DF,$E3,$37,$FD,$D9,$F6,$1F,$E6,$96,$FB,$FF,$91,$8C,$47,$0F,$A3
	.dc.b	$C7,$44,$8E,$2F,$E5,$EF,$C9,$0A,$7B,$9C,$C9,$1F,$EF,$BF,$AF,$7D
	.dc.b	$87,$E7,$E7,$F1,$53,$FD,$C1,$BB,$7F,$8B,$1B,$D5,$FA,$1B,$FF,$39
	.dc.b	$D7,$F9,$06,$B5,$EA,$F3,$24,$77,$F9,$8F,$D6,$7D,$87,$F9,$A2,$38
	.dc.b	$61,$FA,$B8,$C4,$7F,$80,$1B,$7F,$E3,$CF,$B0,$FD,$84,$E7,$DD,$E6
	.dc.b	$47,$99,$FE,$53,$8F,$CE,$38,$1F,$61,$FE,$B6,$70,$D7,$FC,$1E,$DA
	.dc.b	$F0,$7B,$3F,$C0,$E3,$DF,$3A,$F2,$70,$7F,$07,$B7,$D6,$7F,$09,$FB
	.dc.b	$0F,$F4,$85,$78,$5A,$FF,$87,$1B,$8F,$F9,$63,$EB,$DF,$61,$C4,$95
	.dc.b	$C5,$FE,$29,$1F,$4A,$3F,$CE,$1F,$45,$FE,$AB,$7D,$CD,$7F,$C5,$2D
	.dc.b	$AF,$F3,$2F,$D6,$DE,$0F,$E3,$53,$6B,$9F,$C4,$FF,$21,$8C,$42,$AC
	.dc.b	$61,$FD,$B4,$7F,$8D,$9B,$7F,$E7,$29,$E3,$F6,$0F,$98,$FE,$58,$DB
	.dc.b	$FA,$8D,$7F,$C7,$EF,$F1,$F4,$E6,$BF,$CF,$AD,$C7,$F9,$93,$67,$FD
	.dc.b	$24,$F1,$3F,$CC,$EF,$6F,$92,$F0,$9E,$2F,$6E,$46,$48,$78,$9F,$85
	.dc.b	$F4,$92,$3E,$6B,$EA,$E0,$C4,$7C,$E8,$B8,$F9,$A6,$E3,$FD,$1F,$EB
	.dc.b	$FF,$D2,$B3,$FE,$96,$6C,$7D,$2A,$FE,$A1,$0F,$B0,$A4,$C3,$F3,$51
	.dc.b	$F4,$DB,$19,$7F,$CC,$F1,$C5,$DE,$DD,$3E,$73,$AF,$C5,$C9,$E7,$19
	.dc.b	$7A,$C6,$5E,$71,$FD,$8F,$EC,$1E,$25,$24,$FE,$B7,$EC,$3F,$85,$BF
	.dc.b	$AF,$9D,$B6,$76,$FF,$9F,$27,$6F,$F9,$FA,$76,$E7,$3F,$EE,$52,$78
	.dc.b	$EB,$FE,$BE,$87,$C3,$EB,$E5,$65,$F8,$E3,$89,$87,$C1,$18,$7D,$1E
	.dc.b	$07,$62,$3E,$BF,$42,$BE,$78,$B8,$78,$67,$FC,$0D,$F5,$EF,$B0,$A4
	.dc.b	$BF,$E2,$1F,$5F,$FD,$F9,$FF,$4D,$CE,$DF,$E2,$8D,$73,$3F,$E1,$19
	.dc.b	$E1,$7F,$ED,$CA,$D3,$FE,$DE,$8F,$D3,$1F,$A7,$8C,$F2,$7F,$85,$63
	.dc.b	$F4,$37,$FE,$AA,$6D,$71,$E6,$ED,$CF,$DD,$95,$E8,$72,$FA,$BE,$BF
	.dc.b	$C8,$47,$F9,$67,$D8,$7E,$69,$0C,$13,$B8,$F1,$3F,$E2,$D9,$E1,$0F
	.dc.b	$B0,$78,$DF,$E3,$16,$C7,$F3,$96,$BF,$ED,$AF,$83,$7C,$A3,$A3,$85
	.dc.b	$F1,$F9,$D9,$7C,$63,$FC,$79,$F6,$1F,$C5,$B8,$3F,$C7,$4D,$9F,$4B
	.dc.b	$5F,$21,$C3,$8E,$3F,$D2,$E3,$59,$FF,$54,$E1,$5F,$4F,$A6,$EE,$4B
	.dc.b	$FE,$AB,$AF,$A2,$71,$1F,$ED,$EC,$1C,$DE,$63,$38,$74,$9F,$2F,$A5
	.dc.b	$EA,$23,$E3,$F8,$9A,$FF,$B0,$7A,$FD,$A7,$FA,$FC,$AF,$FB,$65,$F8
	.dc.b	$33,$98,$F8,$31,$56,$1F,$B1,$78,$1A,$29,$A9,$E1,$1B,$8F,$11,$B2
	.dc.b	$9D,$27,$73,$DF,$EC,$3F,$B2,$FC,$57,$16,$CF,$F9,$BD,$B5,$F8,$E7
	.dc.b	$FC,$DB,$3B,$F9,$CE,$5F,$34,$F9,$3F,$2D,$23,$17,$E2,$8E,$69,$D4
	.dc.b	$F0,$9D,$DF,$8D,$B8,$78,$F0,$EF,$5B,$26,$1D,$63,$2F,$7A,$E2,$FA
	.dc.b	$5A,$F8,$63,$77,$DE,$B9,$BE,$D9,$B7,$F0,$EF,$8E,$1E,$89,$F0,$1A
	.dc.b	$3B,$5D,$7B,$16,$38,$5F,$78,$E2,$79,$E7,$55,$DB,$0F,$2E,$1E,$19
	.dc.b	$E5,$18,$AE,$E7,$13,$9D,$F5,$25,$71,$8C,$B6,$33,$1C,$9F,$80,$F3
	.dc.b	$72,$FD,$7B,$F2,$3E,$6B,$F1,$B4,$E1,$5F,$5D,$C5,$72,$C3,$F0,$27
	.dc.b	$C2,$9F,$11,$27,$F1,$D8,$7B,$5F,$05,$F6,$7A,$9E,$79,$E0,$FC,$E3
	.dc.b	$27,$02,$D7,$83,$0F,$B2,$4A,$FC,$BD,$F7,$9F,$25,$75,$3C,$67,$AA
	.dc.b	$FB,$C7,$3A,$C4,$E8,$FA,$47,$F9,$A2,$3F,$0C,$DF,$29,$CB,$FE,$6C
	.dc.b	$8F,$B2,$65,$FD,$4C,$6A,$3F,$73,$CB,$FE,$9F,$E1,$CD,$7F,$02,$8C
	.dc.b	$4F,$EB,$6B,$86,$9F,$C0,$E7,$2F,$EF,$63,$39,$7F,$BA,$E7,$2F,$EF
	.dc.b	$E3,$35,$FD,$32,$72,$FF,$01,$8C,$BF,$D2,$3C,$AF,$E4,$D7,$DF,$E2
	.dc.b	$EE,$1D,$66,$D7,$39,$AF,$1C,$BD,$53,$8C,$5F,$17,$63,$0D,$9F,$80
	.dc.b	$DA,$F8,$AD,$C7,$E0,$37,$1F,$EC,$B2,$FB,$0F,$F6,$D5,$FF,$A5,$57
	.dc.b	$87,$FC,$FF,$97,$FA,$05,$6F,$E0,$71,$7C,$AF,$6F,$18,$CC,$FF,$72
	.dc.b	$9C,$AA,$E7,$1E,$08,$C5,$73,$BE,$31,$E5,$F1,$79,$F4,$BE,$B1,$F0
	.dc.b	$BA,$C7,$FC,$63,$8F,$F5,$C9,$CB,$EB,$BD,$1B,$F7,$EC,$7B,$FD,$DE
	.dc.b	$5D,$9C,$DB,$19,$3E,$24,$65,$A7,$FA,$FB,$57,$D5,$FB,$9F,$90,$D8
	.dc.b	$FF,$6A,$2F,$07,$E2,$CF,$0D,$EC,$EF,$7F,$8C,$D9,$EB,$5A,$AE,$3F
	.dc.b	$5C,$FB,$0A,$74,$F7,$7F,$A8,$65,$E5,$9D,$36,$76,$FA,$67,$79,$7D
	.dc.b	$23,$EB,$57,$07,$03,$EC,$1F,$53,$EA,$FA,$D7,$C7,$7D,$89,$CE,$3F
	.dc.b	$16,$72,$FF,$B8,$63,$37,$DF,$55,$6A,$3B,$CE,$A7,$F2,$53,$97,$F2
	.dc.b	$71,$9A,$FC,$DB,$6B,$E4,$FB,$7E,$3C,$7E,$3F,$87,$2F,$E2,$7C,$95
	.dc.b	$BF,$16,$8A,$C6,$5F,$A1,$85,$62,$FE,$DF,$D8,$7D,$8E,$BF,$B2,$D8
	.dc.b	$FC,$27,$1A,$31,$97,$BE,$31,$3A,$AF,$CF,$FD,$73,$E6,$3F,$0B,$D4
	.dc.b	$FA,$FC,$C6,$1B,$7F,$91,$9D,$3F,$E6,$0D,$7F,$6F,$7C,$E3,$55,$C3
	.dc.b	$2F,$B7,$39,$7E,$0F,$D8,$66,$FF,$33,$E7,$7E,$52,$FF,$2A,$DC,$3D
	.dc.b	$AD,$BF,$63,$6B,$E4,$7A,$1F,$96,$3F,$2F,$EE,$5B,$FC,$A4,$5B,$FA
	.dc.b	$33,$6B,$D3,$E3,$7E,$37,$E3,$FE,$2B,$70,$F7,$1F,$C6,$B4,$FF,$7E
	.dc.b	$3F,$8F,$9D,$B4,$DE,$1F,$E8,$49,$DB,$F0,$AB,$F6,$3F,$5F,$C2,$3F
	.dc.b	$07,$F4,$5F,$AE,$F9,$8F,$D7,$B7,$0F,$D6,$36,$77,$1F,$AF,$9F,$DD
	.dc.b	$F8,$92,$3F,$07,$EA,$5A,$FC,$14,$DD,$3F,$83,$CD,$AF,$E6,$BF,$60
	.dc.b	$CB,$F7,$FF,$45,$FC,$92,$31,$3F,$BD,$8C,$4E,$A3,$DF,$8A,$B6,$FE
	.dc.b	$7A,$B7,$3A,$9C,$E1,$F9,$FE,$CC,$DF,$F0,$0F,$AF,$7D,$87,$F1,$8E
	.dc.b	$F7,$BF,$FC,$DA,$73,$97,$F2,$39,$CB,$E2,$F5,$3F,$B9,$E5,$FD,$2E
	.dc.b	$7F,$AD,$CE,$79,$3A,$7A,$1F,$D0,$3E,$BD,$F2,$DF,$C9,$BE,$C3,$FD
	.dc.b	$27,$7E,$C9,$CB,$FC,$C6,$33,$B7,$F1,$69,$63,$FC,$BB,$EC,$3F,$D8
	.dc.b	$77,$C6,$31,$5D,$FC,$FF,$E6,$7F,$60,$F9,$3C,$5F,$F3,$D8,$C3,$F3
	.dc.b	$3D,$77,$57,$F3,$AF,$B0,$FE,$11,$39,$8C,$79,$7F,$D6,$AB,$FA,$9E
	.dc.b	$5F,$C7,$FE,$C1,$EA,$67,$D7,$FE,$87,$18,$C7,$EC,$FC,$1B,$EA,$BF
	.dc.b	$A3,$F8,$A3,$FB,$2C,$66,$7F,$A9,$46,$5E,$3E,$E7,$0F,$81,$DB,$63
	.dc.b	$AE,$7F,$D4,$BE,$C3,$D7,$18,$3F,$AE,$38,$8F,$E8,$31,$F0,$9F,$E7
	.dc.b	$10,$F9,$AF,$EE,$F3,$9E,$CF,$F0,$44,$E6,$3F,$27,$F6,$1F,$1A,$FF
	.dc.b	$A0,$7A,$3F,$CE,$E7,$FC,$27,$39,$BF,$F0,$84,$E6,$BF,$DF,$7F,$5E
	.dc.b	$F6,$BF,$C0,$BC,$AF,$A2,$FF,$03,$CE,$72,$FD,$0C,$E5,$FD,$1C,$3E
	.dc.b	$C3,$7F,$4B,$C1,$B3,$FA,$8F,$AF,$7A,$79,$8C,$37,$3F,$EC,$73,$FE
	.dc.b	$10,$7F,$B2,$FD,$83,$D1,$FF,$0C,$CE,$77,$FF,$14,$CE,$70,$FD,$AF
	.dc.b	$C5,$7F,$4C,$7C,$1E,$97,$F8,$82,$31,$8F,$68,$C7,$B9,$2B,$F7,$B6
	.dc.b	$F4,$71,$7F,$DD,$BE,$BF,$F0,$8F,$F6,$DF,$B0,$7C,$9F,$F8,$B2,$73
	.dc.b	$A7,$F8,$F6,$73,$8F,$F0,$89,$CB,$FD,$2E,$33,$7D,$7B,$BF,$C3,$31
	.dc.b	$8B,$FF,$22,$DF,$0A,$07,$10,$9F,$B9,$39,$AF,$27,$17,$F8,$93,$EC
	.dc.b	$3F,$AA,$5F,$48,$C7,$BD,$AC,$BF,$C6,$15,$A7,$FC,$B7,$92,$B7,$1E
	.dc.b	$BF,$3F,$FC,$49,$F1,$FF,$05,$B1,$92,$67,$F2,$4D,$3F,$CA,$B2,$57
	.dc.b	$F9,$51,$FD,$B4,$FF,$89,$1C,$CE,$FC,$5D,$54,$8F,$E6,$75,$F1,$3C
	.dc.b	$6F,$8B,$F6,$0F,$8E,$FF,$39,$E3,$CA,$BF,$C5,$E6,$C4,$ED,$FE,$FD
	.dc.b	$83,$0E,$77,$FE,$A9,$C2,$D6,$A6,$BF,$E8,$08,$B8,$77,$C3,$A4,$E4
	.dc.b	$EF,$7E,$4C,$3B,$46,$0F,$F4,$48,$BE,$71,$AA,$E7,$3E,$97,$AD,$F7
	.dc.b	$BF,$34,$F3,$3F,$D0,$95,$C6,$33,$5C,$23,$13,$E0,$BE,$F8,$7B,$23
	.dc.b	$B5,$F9,$E3,$11,$E5,$3F,$D9,$87,$CB,$2C,$E1,$C1,$FE,$A4,$FB,$0F
	.dc.b	$E8,$5F,$1D,$C1,$B1,$FE,$D5,$6B,$F0,$9A,$F1,$F8,$DF,$F4,$F3,$B6
	.dc.b	$BB,$6B,$BC,$FB,$47,$0C,$FE,$9E,$5F,$EC,$0A,$FB,$2D,$7E,$45,$AB
	.dc.b	$84,$F3,$9F,$98,$EE,$B9,$EB,$FE,$C0,$7E,$FD,$F1,$C3,$F0,$AD,$BE
	.dc.b	$35,$BB,$FF,$63,$9C,$CE,$D7,$CB,$0F,$0D,$7F,$00,$9D,$36,$B9,$3F
	.dc.b	$C1,$2F,$B5,$F1,$C3,$DB,$5A,$AF,$D2,$36,$B7,$7E,$28,$FD,$3E,$76
	.dc.b	$BF,$C3,$F3,$F1,$6F,$FD,$AF,$7C,$27,$C4,$6A,$F8,$B8,$3F,$D3,$B1
	.dc.b	$B5,$F6,$1F,$C0,$1F,$F1,$5B,$5F,$F5,$1B,$6B,$FD,$CA,$FD,$38,$DE
	.dc.b	$BD,$9E,$6D,$8E,$18,$FF,$BC,$9E,$13,$FE,$AE,$FB,$0F,$F4,$7B,$FE
	.dc.b	$3A,$6B,$9E,$9E,$9F,$5D,$FD,$6E,$35,$18,$D2,$46,$0E,$5E,$17,$8E
	.dc.b	$FF,$C3,$31,$83,$FC,$8F,$6F,$B0,$D5,$49,$D1,$FE,$C2,$9D,$38,$D3
	.dc.b	$F8,$2A,$8F,$E6,$4F,$4E,$4E,$4F,$1F,$8E,$F8,$9F,$61,$A8,$FF,$66
	.dc.b	$B6,$F9,$DF,$FB,$32,$7F,$CB,$73,$9D,$BF,$A6,$C2,$30,$FF,$50,$6B
	.dc.b	$E3,$BF,$A5,$E8,$FF,$BC,$63,$FC,$CA,$D7,$FD,$BA,$DC,$3D,$75,$CA
	.dc.b	$1E,$4F,$22,$57,$C8,$F2,$7A,$F9,$32,$3A,$BA,$7F,$DD,$3F,$61,$FE
	.dc.b	$05,$7F,$CE,$AD,$73,$E3,$FE,$1E,$3E,$FE,$5F,$33,$EC,$1E,$F7,$E2
	.dc.b	$DB,$7F,$E7,$E6,$BF,$EF,$6E,$7F,$F1,$A3,$FE,$82,$FB,$0F,$F3,$5B
	.dc.b	$7D,$FF,$CA,$46,$23,$87,$D1,$E3,$BA,$47,$17,$F3,$57,$E4,$E3,$FA
	.dc.b	$36,$F7,$39,$92,$FF,$D2,$3F,$5E,$FB,$0F,$D1,$4F,$E4,$25,$B7,$E2
	.dc.b	$23,$0D,$E9,$FD,$45,$FF,$9C,$EB,$FD,$37,$5E,$C7,$8E,$3F,$82,$8C
	.dc.b	$5F,$3F,$57,$8F,$CB,$FE,$C7,$EC,$3F,$CD,$58,$7F,$A9,$1A,$F2,$3F
	.dc.b	$C2,$EE,$2B,$F3,$7F,$5F,$FD,$79,$FC,$EC,$3D,$9E,$84,$AF,$F5,$7D
	.dc.b	$7E,$91,$E8,$3E,$C3,$FD,$74,$FF,$AD,$1A,$E9,$B7,$E0,$E9,$FE,$11
	.dc.b	$1F,$06,$75,$F0,$58,$3F,$86,$5B,$EB,$3F,$87,$7D,$83,$E0,$A9,$2F
	.dc.b	$FD,$87,$F5,$EF,$B0,$D9,$23,$FD,$8F,$59,$77,$C9,$FC,$67,$E8,$BF
	.dc.b	$D5,$71,$FE,$CC,$6B,$EE,$6E,$DF,$C0,$BE,$B6,$F2,$FE,$DA,$6D,$7F
	.dc.b	$91,$1A,$9B,$38,$6F,$D0,$FF,$29,$8F,$DA,$57,$3F,$8A,$FE,$59,$F6
	.dc.b	$0F,$97,$A4,$C7,$FD,$C3,$1F,$E1,$0A,$FE,$0E,$EE,$33,$3B,$7E,$BC
	.dc.b	$6D,$FF,$2C,$BF,$E6,$87,$9F,$C9,$78,$0F,$0F,$BB,$23,$24,$39,$BF
	.dc.b	$27,$E9,$24,$7C,$97,$D3,$C7,$88,$F9,$51,$76,$FE,$57,$87,$DD,$9E
	.dc.b	$07,$FB,$F3,$EB,$FF,$97,$CF,$F3,$06,$E7,$F3,$EF,$FA,$84,$3E,$C2
	.dc.b	$92,$F9,$6D,$5F,$CB,$CE,$DE,$D3,$CA,$79,$FC,$E7,$4F,$8D,$95,$E3
	.dc.b	$8C,$BE,$58,$CB,$D2,$3F,$B1,$7D,$83,$C7,$A4,$9F,$F4,$3E,$9F,$D2
	.dc.b	$BE,$C3,$F8,$44,$63,$5F,$E7,$F7,$C6,$30,$7F,$A3,$9C,$3F,$B0,$6D
	.dc.b	$73,$F1,$7B,$4F,$ED,$2D,$B7,$F7,$56,$C7,$41,$7C,$B0,$FE,$9F,$2F
	.dc.b	$87,$E1,$F3,$B2,$F8,$4E,$5B,$39,$7C,$13,$DE,$3C,$1E,$24,$9E,$2F
	.dc.b	$8F,$A3,$11,$E7,$8B,$A7,$C5,$9F,$F0,$67,$D7,$BE,$C2,$92,$FD,$5F
	.dc.b	$61,$FE,$05,$7F,$89,$CE,$DF,$73,$5C,$CF,$F8,$5E,$78,$5F,$F7,$CA
	.dc.b	$D3,$FD,$FA,$3F,$4D,$46,$CE,$B4,$FE,$EB,$1F,$A9,$9F,$E4,$6D,$AE
	.dc.b	$5D,$1D,$FA,$7C,$0C,$AF,$4B,$97,$D7,$F6,$1E,$52,$3F,$5F,$EC,$3F
	.dc.b	$42,$FF,$B4,$30,$4E,$A3,$C2,$FF,$8C,$E7,$70,$FB,$0F,$E7,$8F,$E0
	.dc.b	$5A,$FF,$3A,$6E,$3F,$EA,$2F,$85,$7C,$BB,$A3,$FC,$7D,$5A,$D7,$FC
	.dc.b	$2A,$BE,$7B,$26,$BC,$CF,$F2,$27,$D8,$7F,$1C,$CF,$FD,$CE,$D4,$DA
	.dc.b	$FE,$A5,$7F,$E2,$08,$D1,$38,$FC,$35,$D3,$E5,$23,$F9,$B9,$47,$F9
	.dc.b	$83,$4F,$4D,$EA,$BF,$AB,$D3,$9F,$CA,$67,$6F,$F1,$EC,$7F,$96,$27
	.dc.b	$CF,$EA,$7A,$8F,$5F,$C5,$D8,$FF,$6F,$76,$79,$CF,$F6,$81,$5C,$9C
	.dc.b	$8A,$7C,$07,$6F,$C4,$6B,$93,$75,$FE,$73,$F8,$BF,$DC,$C6,$4F,$F6
	.dc.b	$A4,$F7,$69,$CC,$E3,$7D,$3E,$C3,$FB,$44,$78,$9A,$F0,$0E,$2B,$BB
	.dc.b	$FE,$74,$6F,$9D,$CF,$E6,$9F,$2B,$CD,$B7,$CE,$F9,$63,$FE,$5A,$53
	.dc.b	$AC,$3A,$B6,$74,$DA,$1E,$4C,$3C,$18,$77,$24,$77,$8C,$BE,$19,$E4
	.dc.b	$FB,$1B,$87,$92,$3A,$BE,$4A,$E7,$87,$91,$F8,$D7,$D2,$31,$5D,$67
	.dc.b	$2F,$0C,$3D,$31,$63,$AD,$71,$9C,$C7,$13,$D0,$6C,$F3,$4D,$BF,$34
	.dc.b	$60,$E1,$7E,$23,$E9,$BB,$C3,$DA,$7B,$B0,$F0,$47,$BA,$33,$7E,$88
	.dc.b	$E7,$87,$92,$FA,$3F,$5E,$F8,$CF,$86,$72,$FB,$0F,$05,$79,$E7,$78
	.dc.b	$7C,$B9,$EA,$6C,$A7,$3B,$F0,$1D,$DE,$66,$A3,$E0,$9C,$EF,$C3,$1C
	.dc.b	$4E,$8E,$23,$53,$D2,$72,$D7,$81,$23,$AE,$1F,$12,$3C,$17,$EA,$8C
	.dc.b	$5F,$CE,$9F,$24,$62,$FD,$11,$8C,$3E,$13,$EA,$AF,$25,$FE,$52,$3F
	.dc.b	$D7,$67,$DB,$2E,$1F,$E4,$5E,$4C,$C7,$D7,$BA,$FF,$9C,$FD,$D7,$07
	.dc.b	$EE,$7C,$DF,$E6,$13,$FB,$9E,$6F,$F5,$4C,$E7,$0F,$E1,$F1,$89,$FD
	.dc.b	$A5,$70,$CF,$F8,$5C,$E5,$FD,$FC,$67,$4F,$F6,$8C,$FF,$94,$70,$E7
	.dc.b	$F1,$3E,$2C,$3C,$1E,$6D,$C7,$9C,$DC,$38,$CD,$AF,$DD,$E1,$CD,$F4
	.dc.b	$7C,$6E,$5F,$F4,$84,$D9,$4D,$8C,$37,$1F,$6B,$6B,$9C,$7C,$09,$E9
	.dc.b	$F6,$1F,$ED,$C8,$FF,$10,$2F,$13,$D5,$CD,$FE,$FC,$8D,$5F,$FA,$71
	.dc.b	$D5,$70,$F8,$4D,$BF,$EF,$48,$CC,$EE,$72,$FF,$18,$8C,$D7,$5F,$1B
	.dc.b	$3D,$4F,$AA,$7C,$37,$0C,$18,$7F,$8C,$72,$FF,$20,$DE,$EB,$53,$9A
	.dc.b	$F3,$61,$8F,$81,$CB,$AF,$7F,$7D,$B7,$2F,$82,$DB,$F7,$37,$4F,$84
	.dc.b	$FA,$97,$D5,$FA,$AB,$FC,$B2,$DA,$FF,$6E,$AF,$07,$83,$9A,$E3,$C1
	.dc.b	$83,$33,$6F,$FC,$BB,$2F,$AD,$7D,$87,$85,$B3,$AF,$80,$FE,$9E,$F0
	.dc.b	$8C,$E5,$FE,$DD,$BE,$57,$C4,$F5,$5F,$09,$D5,$70,$AF,$CD,$43,$EB
	.dc.b	$DF,$53,$F0,$7E,$C1,$F1,$DF,$56,$73,$9F,$A2,$72,$FF,$9A,$E3,$37
	.dc.b	$E3,$9C,$BF,$EE,$68,$CE,$1F,$90,$9C,$BE,$28,$CD,$7F,$60,$AF,$26
	.dc.b	$1F,$33,$DD,$F1,$3F,$17,$C4,$95,$F0,$FE,$4E,$DF,$BA,$6E,$3F,$9E
	.dc.b	$D1,$58,$BF,$B7,$F6,$1F,$63,$AF,$EC,$B6,$3E,$AF,$1A,$31,$97,$CF
	.dc.b	$8C,$4E,$AB,$F3,$5F,$5C,$F9,$BF,$C5,$76,$BE,$BF,$31,$86,$DF,$E0
	.dc.b	$A7,$4F,$FA,$56,$B9,$39,$9D,$E2,$D7,$EC,$D6,$A3,$EE,$FD,$86,$A7
	.dc.b	$F7,$B9,$FF,$91,$30,$F5,$E1,$F9,$6F,$3F,$F2,$ED,$8C,$9F,$02,$32
	.dc.b	$FE,$3F,$D1,$FC,$A1,$F9,$5F,$7A,$C6,$49,$97,$E3,$A3,$2F,$9E,$3E
	.dc.b	$E6,$7F,$C8,$72,$FC,$17,$97,$F2,$3C,$4F,$C8,$36,$BF,$1C,$DB,$FC
	.dc.b	$63,$6B,$E0,$D7,$EC,$3E,$C3,$17,$FE,$CE,$E4,$7C,$9F,$EA,$FE,$1B
	.dc.b	$F5,$8D,$9F,$D5,$37,$2F,$D2,$B7,$1F,$93,$E1,$FE,$F0,$FD,$EF,$A9
	.dc.b	$6F,$F0,$B1,$6F,$F0,$93,$6B,$F5,$BF,$60,$CF,$F7,$9F,$45,$FC,$7E
	.dc.b	$31,$8F,$EE,$E3,$13,$A8,$FF,$6B,$4E,$B0,$FD,$CD,$6A,$73,$5F,$C1
	.dc.b	$FB,$31,$5F,$3C,$FF,$24,$FD,$7B,$EC,$3F,$8E,$77,$BD,$FF,$E6,$13
	.dc.b	$9C,$7F,$89,$CE,$5F,$F7,$37,$A9,$FD,$A7,$5F,$E8,$F3,$FD,$52,$73
	.dc.b	$C9,$D3,$CF,$FE,$75,$F5,$EF,$96,$FE,$55,$F6,$1F,$E8,$FB,$FF,$78
	.dc.b	$4E,$5F,$E4,$31,$9D,$BD,$02,$58,$FF,$33,$FB,$0F,$F3,$1D,$F2,$8C
	.dc.b	$57,$83,$CF,$FE,$71,$F6,$0F,$94,$C5,$FF,$2E,$8C,$3F,$F0,$07,$D9
	.dc.b	$75,$FF,$3F,$FB,$07,$97,$FD,$4E,$73,$5F,$CD,$A7,$2F,$1F,$5B,$FA
	.dc.b	$37,$D8,$3D,$5E,$11,$87,$FA,$E7,$C1,$BE,$A7,$FA,$67,$8C,$32,$E1
	.dc.b	$DF,$B7,$AF,$81,$B9,$7F,$46,$9F,$F4,$4B,$FE,$84,$EA,$FE,$73,$87
	.dc.b	$C4,$E3,$FE,$85,$F6,$1E,$87,$E5,$C6,$67,$FB,$FC,$E5,$FE,$83,$5F
	.dc.b	$12,$B8,$FC,$DF,$F7,$38,$C4,$F0,$8F,$EE,$DF,$61,$F0,$AB,$FA,$87
	.dc.b	$A7,$FC,$EA,$FF,$9C,$4E,$6F,$FC,$1B,$39,$AF,$F4,$DF,$D8,$3D,$AF
	.dc.b	$F0,$07,$33,$E8,$BF,$C0,$53,$9C,$BD,$C7,$F4,$39,$D3,$8B,$E7,$F6
	.dc.b	$1C,$3E,$9B,$8B,$67,$52,$FB,$0F,$E4,$4F,$E3,$B0,$57,$F8,$46,$73
	.dc.b	$A7,$F6,$49,$CD,$F2,$9F,$F0,$84,$FF,$68,$FB,$0F,$E1,$B3,$9F,$3F
	.dc.b	$FC,$35,$18,$C7,$F3,$7F,$13,$FD,$42,$B4,$79,$1C,$47,$E8,$30,$EB
	.dc.b	$3F,$D8,$67,$38,$7F,$76,$6C,$7F,$65,$B7,$9D,$FE,$29,$9C,$E1,$FE
	.dc.b	$C8,$FA,$EF,$C8,$5F,$F8,$3B,$EC,$1F,$27,$FE,$2F,$9C,$D7,$F8,$26
	.dc.b	$73,$5F,$22,$73,$97,$F9,$1A,$73,$5F,$B6,$9C,$D7,$E4,$E7,$35,$FE
	.dc.b	$0C,$9C,$D7,$8F,$AF,$FC,$59,$5F,$83,$84,$E9,$BC,$1F,$E3,$1F,$B0
	.dc.b	$FE,$B9,$87,$38,$C7,$BF,$E9,$6E,$5E,$76,$DF,$97,$CC,$FF,$10,$FC
	.dc.b	$7F,$BD,$B5,$ED,$70,$FB,$E6,$9F,$C4,$B0,$57,$F9,$55,$FD,$F4,$FE
	.dc.b	$31,$CB,$7C,$0D,$4C,$8F,$C8,$D7,$FB,$EB,$C4,$FF,$7D,$FD,$83,$E3
	.dc.b	$BF,$C3,$D8,$F2,$AF,$E4,$26,$C4,$ED,$FE,$01,$87,$FA,$46,$31,$1F
	.dc.b	$E2,$7B,$EB,$7D,$2B,$4F,$FA,$B2,$33,$3C,$B0,$AF,$13,$BC,$F1,$8C
	.dc.b	$C7,$8E,$FB,$E1,$E7,$7C,$31,$9A,$E0,$F2,$3F,$D7,$F7,$DA,$B8,$4F
	.dc.b	$8E,$74,$7F,$86,$F0,$F3,$BC,$0C,$CF,$07,$17,$E2,$9D,$C7,$A4,$FF
	.dc.b	$63,$DF,$FB,$21,$F9,$46,$0F,$F6,$6F,$D8,$7F,$45,$F8,$EF,$F6,$6B
	.dc.b	$63,$3E,$5A,$72,$D8,$E3,$F1,$BF,$E3,$16,$E5,$CA,$31,$1F,$ED,$EB
	.dc.b	$46,$23,$FC,$B9,$18,$9F,$A3,$9F,$FA,$FA,$6E,$DF,$E6,$1D,$3E,$ED
	.dc.b	$EE,$7E,$3B,$63,$F0,$35,$C6,$7E,$5B,$63,$8E,$1E,$76,$E1,$DA,$A1
	.dc.b	$D2,$75,$5E,$66,$C0,$FC,$C6,$1E,$59,$FF,$3B,$1D,$CE,$15,$F8,$CC
	.dc.b	$3C,$D1,$A9,$E1,$5C,$E3,$83,$D6,$76,$E6,$7F,$4B,$AF,$18,$FE,$77
	.dc.b	$5E,$FB,$FF,$4C,$DF,$09,$F2,$9A,$BE,$2F,$33,$FA,$1C,$6F,$EC,$3F
	.dc.b	$7B,$1F,$D1,$9A,$FF,$84,$9B,$5F,$EA,$67,$ED,$C6,$F6,$EA,$A3,$78
	.dc.b	$FF,$A6,$5B,$1F,$25,$E2,$FF,$87,$BE,$C3,$FD,$26,$FF,$56,$6B,$9F
	.dc.b	$89,$EB,$F5,$FF,$D5,$A3,$51,$8D,$24,$60,$E5,$E0,$F9,$2F,$FC,$3B
	.dc.b	$18,$3F,$B1,$DB,$EC,$35,$52,$74,$7F,$8C,$A7,$4E,$34,$FF,$63,$C3
	.dc.b	$97,$CA,$71,$9E,$1F,$1D,$F1,$3E,$C3,$51,$FD,$C9,$AE,$8C,$4E,$C8
	.dc.b	$F9,$A7,$E6,$F9,$FF,$EE,$D8,$FE,$F6,$D7,$FC,$8E,$DC,$3F,$DB,$77
	.dc.b	$D9,$FF,$6E,$C3,$C8,$E6,$4A,$F8,$7E,$47,$6F,$22,$47,$67,$4F,$F9
	.dc.b	$53,$EC,$3F,$D1,$6F,$F8,$31,$AE,$7E,$13,$F0,$71,$F6,$B1,$FF,$76
	.dc.b	$FD,$83,$DD,$FC,$3B,$6F,$FC,$2A,$D7,$FC,$C3,$CE,$F3,$BF,$1B,$F6
	.dc.b	$1F,$E6,$B9,$FF,$0D,$B5,$FF,$33,$B7,$0F,$F7,$DD,$3E,$93,$8F,$6C
	.dc.b	$BE,$2F,$E6,$27,$CD,$D7,$FE,$EF,$C1,$ED,$78,$09,$7F,$E2,$AF,$B0
	.dc.b	$7D,$2F,$E8,$27,$F1,$B3,$FE,$78,$6E,$6C,$3F,$DF,$51,$87,$FD,$F9
	.dc.b	$7F,$E7,$5A,$FF,$1A,$CF,$BD,$BA,$75,$3F,$5B,$2F,$98,$FD,$7F,$D8
	.dc.b	$7F,$9A,$EB,$FC,$7A,$D7,$FD,$12,$DC,$7C,$51,$FC,$67,$EB,$FF,$8D
	.dc.b	$3C,$3D,$BE,$64,$7F,$C8,$F1,$FA,$07,$FD,$27,$F6,$1F,$EB,$97,$FC
	.dc.b	$96,$D7,$3F,$17,$FC,$0A,$7F,$CB,$91,$AF,$23,$F8,$2D,$7B,$E3,$9C
	.dc.b	$6F,$EB,$7F,$84,$FD,$77,$FA,$42,$7F,$CA,$ED,$7F,$D4,$6D,$CB,$D7
	.dc.b	$0F,$B0,$7D,$87,$22,$47,$27,$F8,$AC,$7D,$18,$FE,$75,$F4,$5F,$EA
	.dc.b	$C7,$FC,$CA,$D7,$FD,$62,$DA,$EF,$F5,$B7,$93,$E4,$CD,$AE,$7F,$11
	.dc.b	$FC,$8A,$31,$0A,$B1,$87,$F6,$D5,$F4,$9B,$3C,$8F,$F3,$BF,$D8,$3E
	.dc.b	$5B,$F9,$6B,$63,$31,$8F,$33,$2C,$8C,$9C,$A3,$8F,$B5,$CB,$E4,$64
	.dc.b	$F6,$8C,$BE,$08,$CB,$FE,$D2,$BF,$F3,$3B,$F1,$BE,$93,$DC,$47,$CF
	.dc.b	$C3,$5F,$93,$17,$6F,$AF,$4A,$C6,$9F,$92,$FB,$0F,$B7,$5F,$71,$B1
	.dc.b	$F3,$AF,$FA,$6C,$3E,$C2,$93,$1F,$F4,$EC,$7C,$F6,$C6,$5F,$F7,$5C
	.dc.b	$71,$77,$B7,$4F,$9B,$EB,$F1,$52,$79,$C6,$5E,$B1,$97,$9C,$7F,$60
	.dc.b	$FB,$0F,$E1,$D1,$FE,$AE,$6B,$FE,$FA,$6E,$7F,$98,$FB,$0F,$E1,$33
	.dc.b	$BF,$1B,$91,$1F,$F5,$BC,$69,$E5,$1F,$DB,$A1,$A5,$38,$4A,$BB,$7C
	.dc.b	$3E,$FE,$6E,$4F,$48,$E4,$61,$E1,$BF,$F0,$4C,$15,$A9,$F1,$F8,$3E
	.dc.b	$12,$3E,$6F,$A7,$AF,$23,$FC,$4D,$5F,$C5,$DE,$26,$CF,$F6,$9F,$D7
	.dc.b	$BE,$C3,$FC,$12,$D8,$CC,$7F,$83,$3E,$C3,$FC,$01,$1D,$44,$ED,$C3
	.dc.b	$74,$FE,$0A,$DC,$FF,$8F,$D7,$29,$CD,$F5,$E9,$98,$78,$0F,$F7,$3D
	.dc.b	$7E,$81,$F0,$36,$BF,$88,$B6,$FA,$F4,$79,$FA,$65,$F2,$7D,$51,$EB
	.dc.b	$FB,$09,$3D,$9F,$F1,$4F,$D8,$7E,$82,$3F,$C9,$D8,$27,$71,$E2,$77
	.dc.b	$87,$F8,$C3,$EC,$1E,$3D,$25,$7F,$95,$F3,$FE,$83,$F0,$8F,$94,$74
	.dc.b	$EA,$BE,$66,$0F,$9E,$E8,$67,$0F,$E5,$FF,$61,$FC,$5A,$8D,$31,$5F
	.dc.b	$D3,$23,$EB,$46,$79,$7F,$C8,$AE,$A2,$FD,$34,$7C,$DA,$FF,$54,$94
	.dc.b	$FF,$96,$34,$F4,$4E,$EB,$AF,$47,$43,$19,$79,$BE,$2B,$BF,$A9,$2B
	.dc.b	$A9,$FE,$79,$C9,$3E,$43,$51,$E8,$9C,$3C,$0F,$7B,$AA,$F8,$6E,$A3
	.dc.b	$81,$FE,$DD,$EC,$F5,$1F,$ED,$32,$B9,$3C,$51,$97,$FD,$19,$F6,$1F
	.dc.b	$E6,$0F,$8B,$FD,$EC,$63,$E7,$34,$A7,$FB,$24,$EA,$31,$F1,$1A,$9E
	.dc.b	$58,$7F,$7E,$76,$DB,$D4,$ED,$E1,$F3,$4F,$93,$F3,$61,$DE,$30,$B1
	.dc.b	$F0,$63,$D2,$3C,$C7,$05,$3C,$27,$6D,$C7,$C5,$5B,$C3,$C3,$87,$1A
	.dc.b	$D4,$F4,$3C,$15,$C1,$A7,$C8,$8D,$BE,$79,$E8,$C3,$66,$4B,$87,$9A
	.dc.b	$35,$87,$B3,$0F,$1D,$6A,$FA,$DF,$BE,$3E,$94,$EA,$BD,$B7,$CB,$0F
	.dc.b	$34,$62,$B9,$4E,$5F,$64,$64,$93,$C0,$91,$ED,$9F,$88,$71,$7C,$26
	.dc.b	$0E,$A7,$B2,$BE,$E3,$9A,$EA,$F8,$2B,$AB,$F0,$E3,$2F,$43,$91,$C6
	.dc.b	$3F,$15,$87,$C2,$AE,$B3,$E0,$AE,$38,$7C,$52,$57,$C8,$38,$4F,$A6
	.dc.b	$3E,$83,$C6,$FC,$71,$C3,$0F,$04,$F9,$C8,$79,$E7,$C7,$1C,$F0,$FA
	.dc.b	$51,$89,$F4,$93,$0F,$2E,$1E,$28,$C4,$EE,$7D,$35,$EC,$77,$7D,$6B
	.dc.b	$C6,$78,$4F,$1E,$1E,$77,$D4,$62,$BC,$6F,$D9,$3B,$94,$FE,$4D,$1F
	.dc.b	$09,$BE,$67,$F3,$17,$F9,$51,$F9,$73,$3C,$BF,$AD,$8D,$61,$FC,$07
	.dc.b	$A3,$FD,$59,$C5,$9A,$FD,$CC,$3B,$BF,$D0,$35,$B6,$E9,$FE,$82,$9F
	.dc.b	$F2,$75,$F0,$9C,$E2,$3F,$A6,$B8,$6C,$76,$F3,$31,$5D,$E3,$07,$32
	.dc.b	$CF,$EC,$A3,$4D,$9E,$27,$2F,$8D,$B5,$FC,$4B,$EC,$3F,$DB,$0F,$F3
	.dc.b	$37,$FD,$44,$DF,$84,$FF,$4E,$F2,$7F,$A1,$27,$38,$7F,$81,$DE,$2D
	.dc.b	$F1,$B7,$7F,$EA,$29,$CD,$71,$9C,$D7,$C1,$97,$8F,$8E,$A7,$D6,$3E
	.dc.b	$1B,$76,$BD,$E5,$FD,$42,$F7,$5A,$9C,$D7,$97,$0C,$7C,$06,$5D,$7B
	.dc.b	$DB,$EC,$B9,$7C,$16,$DF,$B5,$BA,$FC,$17,$D2,$BE,$AF,$E1,$BF,$19
	.dc.b	$B5,$FE,$05,$5E,$0F,$07,$37,$FE,$4B,$6D,$73,$DF,$55,$B7,$BF,$D4
	.dc.b	$BE,$C2,$9E,$2F,$89,$FE,$99,$97,$F7,$2C,$15,$B8,$C3,$72,$FF,$25
	.dc.b	$4E,$AF,$97,$77,$4B,$FE,$E9,$F5,$EF,$AA,$F7,$FD,$83,$E3,$BE,$9C
	.dc.b	$E6,$BC,$13,$97,$8D,$70,$AD,$B6,$32,$4C,$BC,$7B,$BD,$CD,$CF,$A3
	.dc.b	$C3,$F8,$27,$C2,$F1,$6C,$64,$97,$EC,$DD,$3A,$8E,$BF,$61,$F4,$7B
	.dc.b	$3E,$93,$6F,$E4,$64,$8C,$57,$09,$CD,$7E,$4B,$EB,$9F,$33,$F8,$2E
	.dc.b	$E7,$D7,$E6,$30,$DB,$FA,$93,$A7,$FC,$41,$5C,$9C,$CE,$F0,$FF,$39
	.dc.b	$4E,$5F,$67,$D8,$67,$9B,$F1,$DE,$77,$E2,$EF,$F1,$8D,$C3,$D6,$DB
	.dc.b	$F9,$AD,$AF,$DD,$EF,$FE,$56,$AF,$C2,$FA,$1F,$8E,$F7,$3F,$1E,$C8
	.dc.b	$C9,$F8,$A8,$C6,$1E,$6F,$17,$F0,$BF,$1F,$F1,$5B,$5E,$E6,$CE,$E3
	.dc.b	$F0,$F3,$B7,$47,$D2,$F8,$9C,$4F,$DB,$F9,$1F,$D0,$FC,$B7,$E8,$9B
	.dc.b	$5F,$9C,$6C,$EE,$3F,$3F,$3F,$AF,$F1,$24,$7D,$EF,$52,$D7,$E4,$26
	.dc.b	$ED,$FB,$99,$B5,$FC,$D6,$9F,$60,$CB,$F6,$7F,$43,$FC,$52,$31,$5F
	.dc.b	$AE,$8C,$4E,$A3,$FC,$BD,$3A,$AF,$CB,$56,$A7,$35,$FB,$EE,$DC,$57
	.dc.b	$FA,$CF,$EB,$DF,$61,$FC,$43,$BD,$EF,$7F,$24,$9C,$E3,$FC,$06,$72
	.dc.b	$FF,$9B,$7D,$4F,$EB,$38,$7F,$00,$97,$A1,$FC,$B2,$73,$BF,$F3,$E9
	.dc.b	$CD,$FF,$9D,$7E,$B9,$F6,$1F,$C8,$3E,$C3,$FD,$B3,$39,$8C,$71,$7F
	.dc.b	$36,$AF,$DD,$4F,$F1,$3F,$B0,$FE,$BC,$F3,$C3,$F9,$4C,$63,$4F,$F6
	.dc.b	$85,$3D,$0F,$E5,$BF,$60,$F9,$2E,$91,$87,$FA,$1F,$B4,$E9,$FE,$67
	.dc.b	$F6,$0F,$2F,$31,$87,$7E,$BF,$F3,$8F,$B0,$FF,$4A,$EF,$FE,$91,$C3
	.dc.b	$FA,$73,$FC,$A6,$33,$AF,$FB,$8F,$E0,$DF,$55,$FD,$03,$C3,$1F,$D5
	.dc.b	$23,$37,$FC,$C6,$32,$FF,$A7,$63,$FD,$92,$E7,$E0,$78,$7B,$BC,$DE
	.dc.b	$58,$7F,$4C,$FB,$0F,$65,$71,$8C,$6D,$FD,$8A,$3E,$21,$FC,$FE,$5F
	.dc.b	$37,$9E,$1E,$4F,$F6,$8F,$B0,$F8,$95,$FD,$17,$D1,$FE,$6D,$8F,$FB
	.dc.b	$FE,$72,$FF,$44,$8C,$F6,$2B,$F4,$5F,$5E,$F6,$FF,$B9,$E8,$FA,$2F
	.dc.b	$EE,$B3,$9A,$FF,$5E,$4E,$5F,$F5,$F4,$3E,$C3,$7F,$4B,$C1,$B3,$FE
	.dc.b	$C3,$FB,$0F,$F6,$AB,$C2,$3F,$C0,$31,$8D,$19,$F9,$5F,$F1,$56,$1D
	.dc.b	$6F,$97,$99,$FE,$08,$FB,$07,$9D,$CA,$30,$FF,$83,$3D,$FF,$E8,$F5
	.dc.b	$C0,$FF,$6A,$38,$C3,$BC,$7F,$03,$CB,$FC,$1B,$97,$86,$7F,$D7,$9E
	.dc.b	$E7,$33,$C1,$E7,$DB,$FF,$6F,$7D,$7F,$E1,$AB,$FB,$8F,$D8,$3E,$4F
	.dc.b	$FC,$3B,$39,$C7,$FC,$5F,$39,$BE,$3D,$DF,$E1,$08,$C6,$1E,$43,$FC
	.dc.b	$95,$7C,$6B,$FC,$99,$6A,$DB,$E7,$9D,$57,$0B,$E9,$F0,$1E,$27,$C7
	.dc.b	$C7,$FE,$1E,$FB,$0F,$EA,$58,$71,$8C,$7B,$FA,$CF,$FC,$21,$5A,$7F
	.dc.b	$94,$64,$AD,$C7,$C2,$F4,$3F,$C3,$1F,$1F,$F0,$1B,$19,$26,$9E,$D1
	.dc.b	$7F,$C1,$F2,$F5,$57,$F9,$11,$5D,$F8,$8C,$37,$0F,$86,$7F,$3B,$9D
	.dc.b	$B8,$C3,$E3,$CF,$F9,$39,$FF,$3A,$78,$DF,$E7,$5F,$B0,$FF,$3F,$46
	.dc.b	$AB,$83,$FD,$16,$75,$F1,$BD,$6F,$F2,$97,$C2,$B4,$1C,$61,$E2,$8C
	.dc.b	$57,$FA,$52,$F9,$C7,$73,$27,$F5,$30,$E9,$FD,$1D,$F8,$27,$94,$70
	.dc.b	$33,$86,$DF,$69,$9B,$F1,$61,$E7,$8C,$BE,$6B,$E9,$1E,$1C,$3C,$15
	.dc.b	$C3,$0E,$F7,$BC,$3B,$5F,$98,$ED,$87,$27,$E1,$9F,$EB,$23,$45,$C3
	.dc.b	$0F,$98,$F8,$1F,$61,$FD,$0F,$E3,$B8,$36,$3F,$D7,$6D,$77,$E2,$27
	.dc.b	$7F,$1B,$FC,$22,$76,$D7,$AB,$5F,$F0,$05,$23,$11,$FE,$B7,$8C,$69
	.dc.b	$F4,$74,$FE,$D7,$37,$6F,$F0,$46,$BF,$81,$BD,$CF,$D3,$6C,$7E,$16
	.dc.b	$B8,$CF,$D7,$6C,$71,$C3,$D0,$DC,$3B,$54,$3A,$4E,$AB,$D2,$D8,$FF
	.dc.b	$67,$47,$43,$FC,$38,$57,$D9,$1B,$57,$E6,$9D,$E1,$EC,$6C,$FE,$46
	.dc.b	$74,$F6,$9F,$D7,$6B,$A8,$FD,$D5,$7B,$B0,$FE,$51,$7C,$27,$FD,$D5
	.dc.b	$7C,$1E,$47,$EB,$AB,$85,$FB,$2F,$F4,$1F,$61,$FC,$02,$BF,$66,$D7
	.dc.b	$F8,$43,$63,$F9,$B3,$F6,$E3,$7B,$77,$51,$BA,$FF,$20,$36,$3F,$DE
	.dc.b	$CF,$17,$F8,$A7,$D8,$7F,$A3,$9F,$E0,$6D,$73,$D5,$D7,$EB,$FF,$B0
	.dc.b	$46,$A3,$19,$48,$C1,$CB,$C3,$F2,$5F,$F8,$76,$30,$7F,$10,$B7,$D8
	.dc.b	$6A,$A4,$E8,$FE,$51,$3A,$71,$A7,$F5,$78,$FE,$69,$1C,$F9,$B8,$BC
	.dc.b	$3A,$BE,$27,$D8,$6A,$3F,$99,$36,$F9,$4E,$AA,$57,$F2,$4A,$D6,$DF
	.dc.b	$E6,$9C,$3F,$D9,$13,$FC,$A3,$D1,$FF,$71,$B7,$E4,$BE,$94,$62,$3B
	.dc.b	$E1,$8F,$2A,$47,$17,$E6,$79,$5C,$3C,$A9,$3D,$5D,$BF,$D2,$7E,$C3
	.dc.b	$FC,$02,$FF,$38,$6B,$9F,$1F,$F0,$51,$F6,$77,$FE,$F9,$F6,$0F,$86
	.dc.b	$FC,$33,$6F,$FA,$0B,$5F,$EA,$DD,$1F,$E3,$27,$FA,$2F,$D8,$7F,$9A
	.dc.b	$1B,$EF,$FE,$3E,$31,$1C,$3E,$8B,$1D,$D2,$38,$BF,$95,$BF,$27,$27
	.dc.b	$FA,$4E,$FF,$CC,$D5,$DB,$DB,$E8,$7E,$6F,$EB,$DF,$61,$F9,$D8,$FC
	.dc.b	$54,$FF,$6B,$6E,$DF,$E1,$96,$E1,$FE,$9F,$BF,$F3,$85,$7F,$5E,$9F
	.dc.b	$63,$C7,$D5,$E8,$77,$F9,$8F,$D5,$7D,$87,$F9,$9A,$7F,$B2,$B5,$FE
	.dc.b	$F4,$DC,$7F,$C5,$1F,$61,$FD,$6D,$FF,$15,$43,$DB,$E6,$4A,$FE,$DF
	.dc.b	$5F,$9B,$7F,$C0,$9F,$61,$FE,$B3,$7F,$B9,$B5,$CF,$C6,$FE,$F6,$7F
	.dc.b	$CB,$11,$AF,$2F,$F8,$05,$7B,$A3,$9C,$6F,$EB,$3F,$82,$7D,$87,$FA
	.dc.b	$2E,$7F,$BF,$B5,$FF,$0A,$37,$3F,$F2,$17,$D7,$BE,$C3,$89,$2B,$8B
	.dc.b	$FC,$3E,$3E,$7C,$07,$D1,$3E,$2F,$F8,$A3,$63,$FC,$1E,$D7,$FC,$49
	.dc.b	$F5,$B7,$0F,$F2,$94,$DE,$5F,$F6,$C3,$6B,$B3,$55,$58,$C3,$FB,$1A
	.dc.b	$FF,$2E,$36,$B9,$DF,$F2,$7F,$B0,$7C,$B5,$26,$3F,$E2,$48,$FF,$0E
	.dc.b	$D7,$FB,$A5,$DC,$66,$76,$FF,$8F,$63,$6F,$C3,$7F,$CC,$8F,$3F,$93
	.dc.b	$F0,$1E,$1F,$76,$46,$48,$73,$7E,$4F,$D2,$48,$F9,$2F,$A7,$83,$11
	.dc.b	$FE,$4D,$D5,$38,$D7,$FC,$F9,$8A,$F1,$9F,$D7,$C1,$F5,$FF,$E8,$1B
	.dc.b	$FF,$41,$36,$7E,$8E,$1F,$D3,$61,$F6,$14,$97,$FE,$4A,$8F,$A6,$D8
	.dc.b	$CB,$FE,$61,$8E,$2E,$F7,$E9,$F3,$5D,$7E,$36,$4F,$38,$CB,$D6,$32
	.dc.b	$F3,$8F,$EC,$1F,$60,$F1,$29,$27,$FC,$BB,$F6,$1F,$C2,$9F,$F2,$FC
	.dc.b	$ED,$B3,$B7,$FC,$ED,$3B,$7F,$CF,$13,$B7,$39,$FF,$6F,$9F,$9F,$1C
	.dc.b	$8E,$FF,$11,$9F,$96,$91,$82,$1D,$9C,$BE,$08,$E9,$E0,$48,$F4,$BD
	.dc.b	$FD,$2C,$47,$9A,$2F,$0F,$FA,$E7,$0F,$CF,$4F,$EC,$FE,$C1,$F6,$1F
	.dc.b	$E0,$56,$C7,$FA,$01,$AF,$FA,$57,$EB,$FF,$BE,$BF,$E9,$78,$CD,$FF
	.dc.b	$B1,$5B,$C9,$F8,$96,$CF,$F0,$1A,$E3,$5B,$BF,$F0,$A4,$7E,$92,$4D
	.dc.b	$AE,$31,$8E,$8D,$91,$FF,$4A,$CB,$A7,$11,$E7,$9B,$87,$FB,$66,$B5
	.dc.b	$3C,$3E,$BF,$C2,$47,$FD,$3F,$F6,$1F,$9D,$7D,$11,$97,$E1,$3F,$E2
	.dc.b	$B8,$D4,$3E,$C3,$F9,$E3,$FE,$A8,$6B,$FE,$BD,$6D,$FF,$BE,$FE,$15
	.dc.b	$F2,$CE,$7F,$F1,$B5,$6B,$7F,$D9,$2F,$9E,$C9,$AF,$27,$FC,$79,$F6
	.dc.b	$1F,$73,$33,$4C,$57,$F5,$1B,$FD,$E4,$67,$97,$89,$E0,$AE,$7F,$4B
	.dc.b	$1F,$F6,$0C,$A3,$FC,$B1,$A7,$A2,$F5,$5F,$ED,$9A,$74,$79,$0C,$EC
	.dc.b	$AF,$0F,$C5,$FF,$A8,$28,$D2,$5F,$FA,$9E,$32,$F8,$BD,$3F,$79,$F0
	.dc.b	$26,$C6,$48,$78,$DC,$BF,$EC,$B3,$E2,$C7,$84,$8F,$FA,$D7,$EC,$3F
	.dc.b	$CB,$97,$FE,$D1,$6C,$7F,$23,$6B,$FD,$6D,$A9,$FF,$32,$7C,$67,$F0
	.dc.b	$08,$C7,$CE,$7F,$BA,$7B,$B9,$36,$39,$7C,$4D,$76,$7F,$BD,$27,$6D
	.dc.b	$3C,$13,$BB,$EF,$3A,$9F,$F7,$3B,$6F,$C5,$3F,$DE,$DB,$7C,$6E,$5F
	.dc.b	$53,$C2,$72,$3C,$A6,$6B,$95,$F5,$24,$68,$E0,$72,$BF,$0D,$76,$8C
	.dc.b	$D7,$A1,$F5,$9D,$63,$53,$E6,$8C,$10,$E3,$1F,$06,$F7,$7E,$29,$F0
	.dc.b	$3F,$5E,$33,$1D,$8E,$27,$AD,$CB,$C0,$D5,$78,$5D,$91,$F5,$BC,$4F
	.dc.b	$23,$87,$53,$E1,$7D,$18,$7E,$2E,$FF,$B6,$36,$3F,$C4,$8D,$7F,$4C
	.dc.b	$7D,$B7,$04,$70,$7B,$F0,$7C,$DB,$E5,$BE,$B3,$67,$C6,$7B,$30,$FC
	.dc.b	$D4,$67,$1F,$C2,$B6,$3A,$63,$F8,$D8,$C5,$7E,$7D,$AF,$C7,$26,$1E
	.dc.b	$2C,$3C,$27,$C4,$BF,$35,$70,$9D,$57,$90,$E3,$19,$35,$7F,$52,$77
	.dc.b	$87,$9B,$0E,$97,$EB,$3C,$8F,$81,$E9,$3F,$96,$72,$7E,$42,$FA,$BC
	.dc.b	$8F,$DC,$3E,$19,$C4,$F1,$9C,$C7,$5A,$FB,$C4,$C3,$E5,$C6,$5F,$7B
	.dc.b	$67,$C3,$3B,$7C,$67,$F0,$DB,$FA,$CF,$F0,$FF,$2B,$F9,$F3,$FC,$5F
	.dc.b	$0F,$8E,$FF,$8C,$CF,$F4,$47,$C5,$FE,$1B,$3C,$73,$23,$13,$FE,$39
	.dc.b	$8C,$4F,$E3,$A3,$13,$A7,$FC,$E9,$3A,$7F,$72,$DC,$7F,$8D,$3F,$E6
	.dc.b	$2A,$FE,$0C,$E7,$3E,$DE,$67,$E6,$2B,$78,$59,$CD,$F0,$8C,$61,$E0
	.dc.b	$BF,$13,$FE,$A3,$6E,$1D,$E3,$F4,$66,$EB,$D4,$FF,$A2,$FE,$C1,$F0
	.dc.b	$9F,$A8,$E4,$DC,$3C,$5C,$5F,$F9,$6E,$36,$FF,$A4,$E7,$73,$AB,$E9
	.dc.b	$75,$5F,$3B,$E0,$73,$9D,$46,$65,$7E,$2F,$1B,$A7,$43,$EA,$DF,$0F
	.dc.b	$C2,$73,$18,$CB,$AE,$1E,$D9,$D4,$62,$3C,$7E,$FE,$1E,$F1,$ED,$F7
	.dc.b	$3C,$7D,$5E,$46,$C7,$1D,$11,$8A,$E2,$7C,$B7,$1A,$7C,$BF,$AD,$F6
	.dc.b	$37,$7F,$8C,$A1,$DF,$87,$78,$FC,$D6,$E1,$F1,$1B,$5C,$6B,$85,$F4
	.dc.b	$FA,$97,$D8,$53,$BF,$C4,$FF,$1D,$CB,$E2,$E0,$AD,$C6,$1B,$97,$B2
	.dc.b	$75,$7C,$BB,$3A,$5F,$CB,$FB,$07,$D5,$7B,$3E,$C1,$F2,$1F,$81,$9C
	.dc.b	$D7,$82,$72,$F1,$AE,$15,$B6,$C6,$49,$97,$8F,$7F,$F7,$04,$63,$3F
	.dc.b	$CC,$68,$F0,$7E,$11,$F0,$FC,$2B,$5E,$D8,$B8,$7B,$69,$58,$BF,$B1
	.dc.b	$F6,$1F,$4F,$AF,$EA,$37,$1F,$93,$92,$31,$8F,$95,$B7,$FB,$0F,$B0
	.dc.b	$7C,$E7,$E0,$3A,$DF,$5F,$FA,$59,$CE,$5F,$52,$72,$FE,$5E,$33,$8F
	.dc.b	$D3,$9C,$BF,$90,$11,$AC,$3F,$75,$3A,$7C,$58,$77,$3F,$0D,$F6,$17
	.dc.b	$0F,$5F,$97,$F9,$1C,$BF,$24,$D8,$C9,$F0,$23,$2D,$3F,$11,$E7,$7E
	.dc.b	$44,$FC,$97,$B5,$63,$24,$D3,$E8,$46,$5F,$56,$9F,$C5,$F3,$FE,$25
	.dc.b	$CA,$F8,$FF,$8D,$E2,$FE,$35,$B8,$FE,$2D,$B8,$7E,$5E,$B8,$4F,$C9
	.dc.b	$AC,$DF,$EC,$FE,$22,$CF,$F1,$9E,$57,$C8,$FE,$A7,$E5,$BF,$54,$D9
	.dc.b	$FD,$23,$72,$FC,$43,$70,$FD,$F7,$8B,$FB,$93,$F7,$5E,$8D,$BF,$C3
	.dc.b	$C5,$BF,$CD,$4D,$AF,$E6,$BF,$60,$CF,$F7,$5F,$45,$FC,$A6,$31,$8F
	.dc.b	$ED,$63,$13,$A8,$FE,$39,$3A,$C3,$F3,$95,$A9,$CD,$7F,$05,$ED,$C5
	.dc.b	$7E,$27,$EB,$DF,$61,$FC,$53,$E2,$3F,$D8,$58,$7F,$27,$9C,$BF,$C2
	.dc.b	$23,$35,$F9,$1F,$63,$E7,$BF,$C1,$E1,$F3,$DF,$D0,$67,$3A,$FF,$4D
	.dc.b	$9C,$DF,$F3,$3F,$AE,$7C,$B7,$F2,$4F,$B0,$FF,$50,$CE,$63,$1C,$3F
	.dc.b	$D2,$EB,$F8,$0E,$7F,$C5,$7E,$C3,$FB,$A6,$5F,$D7,$A7,$2F,$F5,$B8
	.dc.b	$CE,$0A,$95,$FC,$EA,$3F,$91,$FD,$83,$E4,$BB,$7A,$1F,$CD,$A3,$13
	.dc.b	$FD,$27,$D9,$74,$FF,$3A,$FB,$0F,$E2,$13,$98,$C7,$97,$FD,$86,$BF
	.dc.b	$AA,$67,$FC,$AF,$EC,$3F,$D6,$F7,$CF,$D7,$FE,$87,$18,$9F,$DE,$7C
	.dc.b	$13,$EA,$7F,$A3,$F8,$A3,$FB,$64,$67,$0F,$E7,$71,$97,$FA,$F4,$3D
	.dc.b	$BE,$3F,$01,$C1,$B3,$AC,$3F,$A9,$FD,$87,$B2,$30,$7F,$46,$71,$1F
	.dc.b	$C3,$E3,$E1,$BF,$D1,$61,$F3,$5F,$DF,$E7,$3C,$DF,$E0,$B9,$CC,$7F
	.dc.b	$71,$FB,$0F,$8F,$7F,$D3,$7D,$3F,$E8,$6F,$F1,$B8,$CD,$FF,$84,$E7
	.dc.b	$2F,$F8,$6B,$EB,$DE,$D7,$F8,$37,$91,$F4,$5F,$E0,$F9,$CE,$5F,$DF
	.dc.b	$E7,$2F,$F8,$06,$1F,$61,$BF,$A4,$E0,$D9,$FF,$03,$7D,$7F,$FB,$DD
	.dc.b	$E1,$1F,$E1,$B8,$C6,$BF,$CC,$6F,$9D,$70,$F2,$FF,$C4,$1F,$61,$FC
	.dc.b	$32,$30,$7F,$84,$5C,$57,$F3,$9F,$7B,$FA,$8C,$EA,$31,$EE,$77,$BF
	.dc.b	$F1,$55,$6A,$73,$5C,$FC,$DC,$DF,$F8,$BF,$EB,$BF,$09,$5F,$DE,$7E
	.dc.b	$C1,$F2,$59,$EC,$FF,$08,$C6,$31,$FF,$12,$5F,$0E,$4F,$F2,$14,$E7
	.dc.b	$0F,$E9,$D3,$9A,$FF,$20,$4E,$6B,$FC,$9B,$39,$AF,$D7,$36,$13,$F6
	.dc.b	$AB,$F6,$54,$AE,$0D,$E1,$FF,$12,$FD,$87,$F5,$BA,$DD,$27,$3E,$EF
	.dc.b	$BD,$B3,$EA,$6D,$FA,$3C,$EF,$F1,$0F,$C7,$7C,$56,$D7,$C0,$70,$FC
	.dc.b	$29,$A7,$F6,$DC,$D5,$FE,$5B,$7F,$6C,$7F,$9B,$0C,$CF,$F9,$35,$CB
	.dc.b	$EE,$73,$3C,$E7,$FC,$C8,$FF,$13,$F0,$FF,$8A,$7D,$87,$FA,$46,$3F
	.dc.b	$CD,$93,$A5,$3A,$F8,$CE,$93,$FE,$DE,$AE,$15,$F3,$2B,$4F,$2C,$2B
	.dc.b	$91,$3C,$30,$F2,$BC,$A3,$C8,$6B,$0E,$F5,$CA,$33,$3C,$70,$F1,$57
	.dc.b	$FA,$B6,$7C,$E7,$13,$C4,$53,$C5,$3C,$6B,$A4,$70,$8C,$3B,$3B,$CF
	.dc.b	$89,$C1,$B7,$11,$E9,$8F,$09,$E3,$29,$E5,$C3,$0F,$B8,$FF,$62,$7D
	.dc.b	$87,$F4,$1F,$8D,$E0,$D8,$F6,$35,$DF,$98,$9D,$FC,$6F,$FA,$62,$76
	.dc.b	$D7,$AB,$5F,$D0,$D2,$31,$1F,$D5,$63,$1A,$7D,$8C,$BF,$B5,$CD,$D7
	.dc.b	$F5,$39,$FE,$02,$F7,$3F,$35,$B1,$F8,$4A,$E3,$3F,$45,$B1,$C7,$0F
	.dc.b	$43,$70,$ED,$50,$E9,$3A,$AF,$4B,$63,$FB,$64,$74,$3F,$7A,$57,$D9
	.dc.b	$1B,$57,$E6,$9D,$E1,$EC,$6C,$FE,$62,$74,$F6,$9F,$D6,$69,$A8,$FF
	.dc.b	$31,$57,$BB,$0F,$F6,$FD,$F0,$9F,$F7,$E5,$F0,$79,$1F,$EC,$3A,$E1
	.dc.b	$7E,$CB,$FD,$3F,$D8,$7F,$00,$AF,$F6,$43,$5F,$F3,$83,$63,$FD,$DA
	.dc.b	$FD,$B8,$DE,$DD,$D4,$6E,$BF,$90,$B6,$3F,$C2,$8F,$17,$FC,$F9,$F6
	.dc.b	$1F,$E9,$47,$FD,$B0,$D7,$3C,$FD,$7E,$BB,$FA,$DC,$6A,$31,$94,$8C
	.dc.b	$1C,$BC,$0F,$25,$FF,$86,$A3,$07,$FB,$96,$DF,$61,$AA,$93,$A3,$FD
	.dc.b	$21,$3A,$71,$A3,$3F,$E7,$B3,$FC,$DE,$7A,$72,$72,$78,$FC,$5F,$C5
	.dc.b	$FB,$0D,$47,$FA,$6D,$B8,$73,$8F,$F1,$C5,$FF,$B3,$1F,$F7,$D4,$EB
	.dc.b	$0F,$E9,$6D,$8F,$F1,$F3,$54,$F1,$7F,$D5,$1E,$7F,$FB,$C1,$FF,$22
	.dc.b	$5F,$75,$19,$77,$F2,$1E,$7C,$BE,$E4,$BC,$8A,$4A,$F9,$5E,$46,$BC
	.dc.b	$89,$1D,$B4,$FE,$C3,$7F,$E1,$88,$C1,$FE,$55,$7A,$38,$1F,$61,$FE
	.dc.b	$96,$70,$D7,$FD,$84,$DC,$FD,$4F,$E1,$DF,$67,$D8,$7F,$AA,$E7,$A3
	.dc.b	$5F,$F6,$53,$63,$E1,$73,$7F,$8F,$27,$C5,$F6,$1F,$E7,$66,$FB,$7F
	.dc.b	$9D,$8C,$47,$FB,$4E,$5F,$47,$9E,$C9,$3C,$9F,$D2,$4F,$97,$4F,$F0
	.dc.b	$1D,$BD,$9E,$84,$BF,$85,$F5,$EF,$B0,$FD,$64,$FE,$66,$7F,$DC,$4D
	.dc.b	$D3,$ED,$37,$0F,$F0,$7D,$FF,$9F,$6B,$E6,$4F,$B5,$E3,$E9,$F4,$3F
	.dc.b	$71,$2F,$96,$FD,$D7,$D8,$7F,$9C,$EB,$EA,$35,$FF,$79,$37,$1F,$C5
	.dc.b	$7D,$87,$F8,$7A,$73,$EB,$F2,$23,$F7,$63,$F5,$4F,$FB,$EF,$EC,$3F
	.dc.b	$D8,$4F,$E0,$5A,$E7,$E2,$7F,$86,$4F,$F9,$8A,$35,$E4,$FF,$0F,$AF
	.dc.b	$74,$73,$8D,$FD,$67,$F1,$3F,$B0,$7C,$05,$25,$7E,$3F,$EB,$DF,$61
	.dc.b	$B2,$47,$E4,$EB,$2E,$FC,$5F,$E3,$DF,$44,$F8,$AA,$49,$FC,$D7,$D6
	.dc.b	$E5,$FC,$E4,$D5,$C3,$FB,$18,$C3,$6B,$F9,$55,$63,$9F,$FC,$93,$56
	.dc.b	$FF,$D7,$93,$FB,$7B,$FD,$D3,$6B,$77,$FC,$CF,$EC,$1F,$29,$49,$87
	.dc.b	$EB,$E3,$FC,$49,$5F,$E5,$67,$71,$99,$DB,$E8,$9D,$79,$BD,$7D,$AE
	.dc.b	$DF,$55,$2B,$A4,$65,$ED,$19,$79,$1F,$ED,$5C,$3F,$CE,$EF,$C6,$FA
	.dc.b	$4F,$9A,$47,$CD,$C7,$5E,$23,$64,$E7,$6F,$E3,$96,$BC,$67,$FE,$72
	.dc.b	$FA,$FF,$F2,$D3,$4C,$E3,$FE,$E7,$87,$D8,$7F,$5B,$6C,$7F,$15,$6B
	.dc.b	$E2,$5C,$1B,$99,$3B,$79,$47,$3F,$9C,$E9,$F1,$72,$BA,$46,$5E,$D1
	.dc.b	$97,$B4,$7F,$6D,$FB,$0F,$E1,$F1,$FC,$9D,$AF,$C4,$6E,$9F,$CF,$7E
	.dc.b	$C3,$F8,$54,$EF,$C8,$E4,$47,$F9,$84,$69,$E5,$1F,$DE,$E3,$E6,$C6
	.dc.b	$B7,$E0,$FD,$28,$7C,$2E,$FE,$4E,$4F,$48,$E4,$61,$E1,$18,$7C,$3E
	.dc.b	$1F,$72,$3E,$4F,$4E,$BD,$E2,$E5,$ED,$9F,$F0,$87,$D7,$BE,$C2,$92
	.dc.b	$FF,$A6,$7D,$87,$F8,$21,$FC,$4C,$ED,$FE,$A0,$D7,$33,$FE,$1B,$9E
	.dc.b	$17,$FD,$9E,$B4,$FF,$69,$8F,$D4,$D0,$38,$53,$C3,$93,$FD,$87,$3F
	.dc.b	$A0,$9F,$CD,$B6,$BA,$74,$F8,$7A,$7C,$4C,$AF,$5B,$97,$DB,$F6,$1E
	.dc.b	$72,$3F,$DA,$FE,$C3,$F3,$EF,$97,$04,$EA,$3C,$2F,$F8,$E6,$77,$0F
	.dc.b	$B0,$79,$3F,$E3,$D6,$C7,$F7,$76,$BF,$BC,$F8,$47,$CA,$3A,$38,$5F
	.dc.b	$1F,$9B,$97,$C6,$3F,$C9,$1F,$61,$FC,$66,$3E,$53,$53,$71,$FE,$9D
	.dc.b	$BF,$F8,$66,$34,$4E,$3F,$35,$74,$FA,$78,$FF,$83,$65,$1F,$E6,$5D
	.dc.b	$3C,$FC,$5A,$AF,$E1,$B4,$E6,$F2,$99,$C7,$C1,$EB,$7F,$AB,$27,$FC
	.dc.b	$B5,$5F,$D8,$5A,$FF,$C0,$06,$BE,$C8,$FF,$33,$7E,$09,$F2,$2B,$FC
	.dc.b	$64,$DC,$38,$4E,$B5,$97,$FE,$25,$AD,$3C,$CC,$47,$FA,$FA,$3E,$9C
	.dc.b	$7F,$73,$6B,$F9,$50,$E1,$F9,$FE,$BF,$AC,$8F,$FB,$2F,$1F,$F4,$13
	.dc.b	$ED,$8F,$F7,$87,$67,$B4,$B1,$F6,$1A,$FD,$88,$B3,$FC,$2A,$3F,$DF
	.dc.b	$B3,$C5,$FD,$47,$C5,$66,$74,$DA,$E3,$52,$B6,$F1,$79,$9E,$A3,$27
	.dc.b	$8C,$A7,$E2,$9E,$4F,$03,$CC,$6A,$BC,$31,$D6,$BC,$8F,$AA,$B8,$C6
	.dc.b	$2B,$84,$E5,$A4,$3B,$57,$47,$CC,$62,$39,$9D,$CF,$C4,$56,$23,$D4
	.dc.b	$6C,$F3,$39,$7D,$87,$7A,$F9,$0D,$9E,$8F,$28,$CB,$CC,$8F,$AD,$F3
	.dc.b	$CF,$5A,$F4,$47,$63,$27,$B7,$2F,$D0,$5F,$E3,$5B,$1F,$C8,$DA,$FF
	.dc.b	$A3,$CE,$4E,$08,$FD,$B7,$77,$F8,$3F,$B3,$FC,$49,$E3,$AB,$5F,$5E
	.dc.b	$30,$D2,$1E,$C8,$F8,$2F,$F0,$A8,$E2,$E4,$F8,$91,$FB,$78,$CC,$FC
	.dc.b	$32,$46,$67,$AC,$61,$E2,$7F,$18,$9E,$33,$E0,$BF,$92,$23,$33,$C0
	.dc.b	$F9,$95,$EF,$7A,$D7,$13,$6F,$A4,$FB,$87,$1A,$E8,$F0,$31,$1C,$23
	.dc.b	$31,$EF,$39,$1F,$94,$AD,$4F,$C4,$BF,$E3,$51,$F6,$E7,$E0,$1F,$66
	.dc.b	$B9,$CF,$F9,$FF,$EC,$9F,$34,$F9,$4F,$F4,$15,$FF,$63,$9C,$C6,$21
	.dc.b	$1F,$E3,$9E,$5C,$3A,$8E,$2E,$5F,$DB,$37,$3F,$E1,$5B,$3E,$A3,$FD
	.dc.b	$D1,$1F,$C3,$67,$5A,$FF,$1F,$AD,$67,$FC,$0A,$73,$87,$F7,$79,$CB
	.dc.b	$DB,$CB,$65,$FC,$22,$73,$7F,$E9,$D6,$D7,$5C,$BF,$DF,$13,$96,$C7
	.dc.b	$F2,$FF,$B1,$FF,$74,$4F,$FB,$75,$FF,$57,$37,$E1,$7F,$E0,$07,$27
	.dc.b	$FB,$CE,$73,$5F,$E5,$E7,$8B,$7C,$3D,$DF,$5B,$B3,$FE,$11,$AE,$37
	.dc.b	$CF,$C4,$CF,$43,$EB,$7F,$DF,$F8,$F8,$E7,$31,$8C,$3C,$11,$D1,$61
	.dc.b	$FE,$3E,$AF,$57,$46,$63,$13,$E7,$F7,$B1,$F6,$59,$6D,$FC,$16,$CE
	.dc.b	$E3,$E2,$4F,$AB,$37,$D8,$7A,$A7,$FC,$59,$18,$DB,$52,$EE,$B1,$2B
	.dc.b	$67,$09,$E5,$87,$FA,$C1,$E1,$1A,$8C,$4F,$87,$EA,$DF,$65,$4F,$27
	.dc.b	$C5,$7E,$76,$F8,$64,$AD,$C6,$1B,$87,$FA,$7A,$75,$7C,$BA,$BA,$5F
	.dc.b	$86,$9F,$64,$FA,$BF,$85,$F5,$CF,$91,$FA,$53,$9B,$F0,$CE,$5E,$51
	.dc.b	$9B,$F0,$4E,$5F,$F5,$3C,$67,$2F,$C7,$CE,$55,$7F,$81,$67,$4D,$C3
	.dc.b	$E7,$FB,$7E,$27,$E3,$7C,$04,$9F,$81,$F2,$97,$2F,$81,$B7,$FB,$C2
	.dc.b	$7E,$25,$7C,$DA,$D6,$5F,$A0,$E1,$7D,$97,$DC,$EB,$FB,$AD,$C3,$E7
	.dc.b	$CE,$9B,$3A,$69,$F2,$FE,$C1,$F3,$7F,$8B,$EC,$7D,$8F,$E6,$E7,$35
	.dc.b	$F6,$A7,$2F,$1A,$E1,$5B,$6C,$64,$9B,$7B,$BE,$CB,$51,$E6,$F3,$7F
	.dc.b	$2B,$7F,$96,$6C,$64,$F5,$C6,$5A,$7E,$27,$CF,$FC,$91,$F9,$3F,$6E
	.dc.b	$C6,$49,$87,$CE,$8C,$BF,$EE,$4C,$FF,$DF,$F3,$F8,$6F,$0F,$E0,$7C
	.dc.b	$17,$E3,$1B,$19,$3E,$44,$65,$A4,$AF,$CD,$6E,$F8,$AE,$31,$F7,$BE
	.dc.b	$CB,$F5,$1F,$0B,$FA,$96,$DF,$E8,$DB,$3B,$8F,$D3,$4F,$EE,$BC,$69
	.dc.b	$1F,$7B,$D3,$B5,$E0,$9B,$CD,$F8,$79,$B5,$FE,$5B,$FB,$2F,$23,$FB
	.dc.b	$1F,$A4,$FE,$31,$18,$C3,$F2,$71,$8C,$FF,$7C,$DA,$E0,$7E,$CE,$B8
	.dc.b	$61,$FA,$4B,$FF,$5E,$46,$27,$F8,$9F,$5F,$E9,$A7,$38,$7F,$27,$FB
	.dc.b	$17,$D9,$7F,$1B,$EF,$7B,$D9,$8C,$3F,$CB,$7C,$FF,$EA,$3A,$7F,$42
	.dc.b	$A7,$D0,$67,$97,$93,$FC,$CF,$EC,$1F,$2F,$FC,$9F,$EC,$BF,$D2,$DC
	.dc.b	$1F,$C2,$E7,$2F,$F1,$68,$CE,$EC,$BD,$8A,$99,$7F,$31,$FB,$2F,$F1
	.dc.b	$C4,$EF,$0F,$F4,$45,$69,$FE,$43,$3A,$BE,$72,$F4,$3F,$9D,$7D,$93
	.dc.b	$E4,$FD,$11,$87,$FA,$67,$B4,$EA,$FE,$81,$F6,$5F,$C1,$F4,$F0,$1F
	.dc.b	$E8,$48,$CF,$95,$C3,$E7,$BF,$A3,$7D,$93,$D6,$C7,$07,$F3,$48,$C3
	.dc.b	$FC,$B7,$E1,$1F,$59,$FD,$37,$C6,$1F,$D7,$63,$33,$FC,$F2,$32,$F1
	.dc.b	$F7,$38,$7C,$06,$DB,$1D,$75,$FE,$B1,$F6,$5E,$B8,$C1,$FD,$19,$C4
	.dc.b	$7B,$0F,$E5,$FC,$3C,$3E,$77,$FB,$75,$7C,$73,$83,$FD,$C7,$EC,$BE
	.dc.b	$27,$3F,$F8,$1B,$D2,$FE,$71,$5F,$EB,$B9,$CB,$FD,$9A,$33,$80,$7D
	.dc.b	$83,$DC,$FE,$F9,$CA,$FA,$4F,$EF,$D3,$9C,$3F,$AA,$4E,$5F,$EA,$D0
	.dc.b	$FB,$2D,$FD,$37,$06,$CF,$F5,$BF,$B2,$FF,$76,$BC,$23,$FC,$15,$18
	.dc.b	$CB,$E1,$5F,$3A,$E1,$E7,$7F,$83,$FE,$C9,$E8,$70,$8C,$3F,$E1,$3F
	.dc.b	$80,$7B,$99,$BF,$ED,$51,$87,$99,$88,$E0,$75,$F4,$6C,$7F,$71,$FB
	.dc.b	$0F,$C0,$E7,$FD,$DF,$EC,$9F,$29,$FE,$21,$9C,$F3,$7F,$8D,$A7,$37
	.dc.b	$C7,$BB,$FC,$31,$18,$C3,$EF,$5F,$0A,$FF,$0C,$CE,$6F,$F6,$33,$9A
	.dc.b	$FA,$CD,$84,$FD,$A9,$4E,$DB,$C5,$FE,$27,$FB,$2F,$EB,$71,$E0,$8C
	.dc.b	$7B,$DE,$96,$E3,$EF,$6C,$F9,$BC,$EF,$F0,$EF,$C8,$7B,$DB,$5F,$15
	.dc.b	$C3,$F1,$E6,$9F,$9B,$CD,$5F,$E4,$77,$F6,$31,$98,$C7,$89,$96,$57
	.dc.b	$F8,$86,$7F,$E0,$08,$FB,$27,$D0,$FE,$66,$3F,$8C,$57,$F9,$4D,$DC
	.dc.b	$E9,$FF,$22,$C7,$F3,$78,$C4,$71,$F2,$BF,$45,$18,$C7,$CF,$86,$30
	.dc.b	$FF,$4B,$D5,$BE,$73,$5E,$47,$96,$7C,$95,$CA,$30,$79,$C5,$71,$8D
	.dc.b	$5F,$92,$7D,$35,$CB,$0F,$43,$E2,$8E,$67,$C0,$C3,$C4,$F0,$3C,$A6
	.dc.b	$8C,$1B,$30,$78,$8E,$77,$E2,$9E,$D1,$FE,$B3,$C3,$FD,$68,$FD,$23
	.dc.b	$07,$FA,$EB,$EC,$BF,$9D,$7C,$77,$FA,$E5,$B1,$9F,$0D,$39,$6C,$71
	.dc.b	$F8,$EF,$EA,$0D,$C7,$94,$62,$3E,$45,$A3,$11,$F8,$18,$C4,$FD,$3C
	.dc.b	$7F,$B2,$CD,$DB,$F0,$DA,$FE,$02,$F7,$3F,$29,$B1,$F8,$4A,$E3,$3F
	.dc.b	$39,$B1,$C7,$0F,$3B,$70,$ED,$50,$E9,$3A,$AF,$33,$63,$F9,$84,$74
	.dc.b	$3F,$34,$57,$F9,$A5,$F9,$23,$8D,$75,$BF,$36,$1E,$C9,$CD,$F9,$E3
	.dc.b	$F2,$3A,$74,$6C,$78,$1A,$B0,$EC,$E4,$FF,$48,$57,$C2,$C3,$FC,$06
	.dc.b	$DB,$EF,$87,$EA,$BE,$CB,$F7,$B7,$FE,$97,$6B,$FD,$6D,$B7,$FE,$0C
	.dc.b	$7E,$D4,$6F,$6F,$1C,$6A,$FF,$78,$D9,$FE,$AC,$F0,$7F,$B3,$7D,$97
	.dc.b	$F3,$E7,$FD,$4C,$D7,$3D,$3E,$EF,$B0,$FE,$A9,$1A,$8C,$67,$23,$07
	.dc.b	$2F,$0B,$E2,$DF,$F8,$46,$30,$7F,$AC,$ED,$F6,$5A,$A9,$3A,$3F,$BD
	.dc.b	$4E,$9C,$69,$FE,$2C,$87,$2F,$94,$E3,$3C,$3E,$3B,$EA,$FD,$96,$A3
	.dc.b	$FD,$84,$D7,$46,$27,$64,$7C,$D3,$F7,$BD,$1F,$F7,$14,$7F,$B2,$DA
	.dc.b	$FF,$83,$DB,$87,$F9,$02,$FB,$3F,$E4,$28,$79,$3C,$C9,$5F,$0F,$C9
	.dc.b	$ED,$E4,$C8,$EC,$E9,$FF,$0E,$7D,$97,$F7,$87,$FD,$B8,$D7,$3E,$2F
	.dc.b	$E5,$23,$ED,$63,$FE,$52,$FB,$27,$BB,$F9,$86,$DF,$FB,$95,$AF,$F8
	.dc.b	$A7,$9D,$E7,$7E,$7B,$EC,$BF,$CC,$73,$FE,$EB,$6B,$FE,$2F,$6E,$1F
	.dc.b	$E6,$0A,$7D,$27,$1E,$E9,$7C,$5F,$D4,$CF,$9B,$AF,$FA,$96,$1F,$E5
	.dc.b	$DA,$ED,$EC,$F8,$4F,$D7,$FD,$8B,$EC,$BF,$63,$1F,$8D,$9F,$F2,$03
	.dc.b	$73,$FF,$38,$B7,$0F,$EB,$D7,$FE,$67,$AF,$F1,$C5,$6B,$4E,$64,$8C
	.dc.b	$D7,$0F,$97,$FD,$FF,$D9,$7F,$97,$A7,$27,$F9,$05,$C3,$FE,$51,$6D
	.dc.b	$FF,$A0,$BE,$CB,$FA,$C3,$FE,$84,$87,$B1,$D0,$95,$D0,$57,$E8,$5F
	.dc.b	$F2,$E7,$D9,$7F,$B1,$5D,$35,$CF,$C5,$7F,$14,$9F,$F2,$94,$6B,$C8
	.dc.b	$FE,$2F,$5F,$06,$39,$C6,$FE,$BB,$F8,$DF,$D8,$3E,$06,$92,$BC,$9F
	.dc.b	$64,$FB,$2D,$92,$3D,$15,$97,$7E,$2F,$F2,$6F,$A1,$FF,$50,$C7,$B5
	.dc.b	$AF,$F9,$DD,$BC,$1F,$EA,$8F,$AF,$BC,$7F,$E1,$59,$B5,$FE,$6F,$6A
	.dc.b	$6C,$E1,$BF,$43,$FC,$DA,$3F,$6B,$5C,$FE,$2F,$F9,$CF,$D9,$3E,$5E
	.dc.b	$93,$1F,$A5,$1F,$E1,$BA,$FF,$13,$3B,$8C,$CE,$DF,$F4,$84,$6D,$FF
	.dc.b	$42,$3F,$E5,$E7,$9F,$C8,$F8,$0F,$0F,$BB,$23,$24,$39,$BF,$27,$E9
	.dc.b	$64,$7C,$97,$D3,$C7,$88,$F9,$51,$76,$FF,$68,$E0,$AC,$65,$FE,$3F
	.dc.b	$FB,$2F,$F6,$AD,$7F,$B5,$9B,$1F,$3E,$FF,$AE,$C3,$EC,$A9,$31,$FC
	.dc.b	$B4,$36,$C4,$7F,$AC,$63,$83,$A9,$E7,$3D,$3E,$73,$AF,$C6,$4A,$E7
	.dc.b	$19,$7A,$C6,$5E,$91,$FD,$C3,$EC,$BF,$86,$C7,$E9,$DA,$FF,$B0,$5B
	.dc.b	$9F,$FB,$CB,$EC,$BF,$84,$4E,$FC,$8E,$44,$7F,$61,$1A,$79,$47,$F7
	.dc.b	$E8,$69,$4E,$13,$FE,$CF,$87,$C3,$EF,$E7,$24,$F4,$8E,$46,$1E,$11
	.dc.b	$87,$C3,$E1,$F7,$23,$E4,$E7,$AF,$78,$B9,$7B,$67,$FC,$27,$F6,$2F
	.dc.b	$B2,$A4,$BF,$E1,$9F,$65,$FD,$E9,$FF,$73,$4E,$DF,$E2,$0D,$73,$3F
	.dc.b	$E1,$E9,$E1,$6E,$4F,$F4,$45,$EE,$31,$7F,$EE,$3A,$FD,$12,$6C,$EE
	.dc.b	$BF,$D2,$11,$FA,$59,$FF,$7C,$36,$B9,$C7,$1E,$7F,$17,$47,$8D,$91
	.dc.b	$EC,$72,$FB,$BE,$CB,$D0,$47,$F9,$67,$D9,$7E,$79,$F3,$46,$5F,$0B
	.dc.b	$FE,$3C,$8D,$43,$EC,$BF,$9B,$BF,$CD,$5A,$B0,$F3,$C6,$1F,$87,$F0
	.dc.b	$AF,$96,$73,$FF,$92,$2B,$5A,$FF,$AC,$57,$CF,$64,$D7,$97,$FE,$4E
	.dc.b	$FB,$0F,$E3,$39,$83,$86,$F1,$F5,$78,$31,$FE,$99,$1B,$BF,$EA,$78
	.dc.b	$7F,$B0,$61,$F4,$FD,$C9,$5F,$D7,$B4,$F5,$56,$23,$E0,$3C,$79,$BC
	.dc.b	$C6,$76,$FB,$38,$7F,$93,$DB,$EA,$FA,$CF,$67,$97,$70,$FF,$5E,$1E
	.dc.b	$DA,$FF,$2B,$3F,$EB,$F6,$BE,$79,$FF,$55,$FD,$97,$CA,$8F,$88,$D4
	.dc.b	$D9,$F1,$63,$FE,$DF,$8D,$13,$6F,$15,$7B,$0F,$17,$AD,$F4,$48,$FC
	.dc.b	$AF,$8C,$AE,$CF,$A7,$18,$CF,$E0,$CF,$D9,$EE,$FF,$73,$1F,$3C,$B5
	.dc.b	$EF,$6B,$C8,$FB,$DF,$19,$FC,$41,$D1,$88,$E4,$D3,$27,$FB,$91,$FB
	.dc.b	$97,$D3,$3E,$53,$93,$75,$88,$CB,$67,$85,$4A,$EC,$7E,$41,$E9,$19
	.dc.b	$8E,$86,$4D,$D7,$49,$F1,$C6,$1E,$15,$E1,$79,$C6,$48,$74,$9E,$4F
	.dc.b	$D4,$33,$3C,$5B,$3D,$9E,$0F,$63,$E3,$BC,$EB,$0F,$C2,$8C,$CF,$C1
	.dc.b	$8E,$66,$48,$7D,$53,$A9,$A3,$E0,$1D,$4C,$57,$A1,$CC,$75,$78,$91
	.dc.b	$F5,$3C,$4C,$4F,$90,$8F,$C1,$32,$79,$27,$B1,$FE,$C8,$9E,$64,$9E
	.dc.b	$33,$E9,$8C,$3C,$63,$E8,$BD,$0F,$3B,$C2,$7A,$34,$F9,$31,$CD,$C3
	.dc.b	$F0,$2B,$33,$E4,$9E,$B5,$E0,$AE,$A7,$A6,$B7,$3C,$1B,$1C,$CF,$97
	.dc.b	$19,$7F,$1B,$F6,$5F,$9D,$23,$F4,$1A,$FE,$25,$F3,$FD,$93,$E7,$1F
	.dc.b	$2D,$F9,$C8,$C3,$63,$35,$F9,$C9,$FF,$32,$39,$9F,$E2,$EA,$38,$BF
	.dc.b	$C6,$9F,$F2,$A7,$3E,$9C,$C7,$F3,$93,$F9,$D6,$DF,$EB,$4E,$0F,$17
	.dc.b	$9D,$FD,$0A,$7F,$57,$39,$7C,$51,$9D,$7F,$77,$39,$7F,$7B,$7E,$0A
	.dc.b	$E3,$EF,$7F,$74,$9C,$DF,$1D,$7C,$5F,$65,$FE,$D4,$CB,$FD,$20,$BC
	.dc.b	$4F,$EF,$7C,$7F,$DC,$A3,$59,$7F,$1F,$75,$5C,$3E,$13,$6F,$4B,$B5
	.dc.b	$FD,$5E,$B8,$DF,$39,$CC,$A7,$FC,$53,$5F,$D3,$FE,$B5,$F0,$DA,$9C
	.dc.b	$D7,$F3,$5E,$B5,$6A,$FD,$3E,$46,$CE,$1D,$56,$3D,$BE,$FF,$3E,$AE
	.dc.b	$8D,$8C,$9F,$0A,$32,$D3,$FD,$7F,$AB,$EA,$BE,$0E,$7F,$D1,$E3,$1B
	.dc.b	$7C,$07,$5E,$0D,$AD,$C7,$CC,$6D,$FF,$B4,$9B,$7F,$E0,$B5,$5C,$6F
	.dc.b	$A7,$D4,$BE,$CA,$9D,$FE,$27,$F3,$78,$F9,$A7,$4D,$9D,$BE,$A9,$DE
	.dc.b	$3E,$9A,$DB,$FD,$6F,$EC,$5F,$63,$EB,$FB,$27,$C7,$7D,$89,$CE,$3F
	.dc.b	$12,$72,$FF,$B6,$63,$37,$F1,$67,$2F,$5B,$E9,$7C,$D9,$7F,$3B,$57
	.dc.b	$B9,$8B,$F3,$7C,$0F,$C2,$3E,$1F,$CA,$5B,$F8,$11,$67,$D9,$37,$87
	.dc.b	$E8,$7D,$97,$D3,$EB,$FA,$8D,$C3,$C0,$DB,$FB,$99,$23,$17,$FE,$54
	.dc.b	$FB,$07,$CD,$FD,$FE,$A7,$D8,$FE,$72,$73,$97,$E2,$27,$2F,$FB,$F6
	.dc.b	$33,$97,$E2,$67,$2F,$F8,$2E,$33,$87,$EC,$A7,$2F,$87,$0E,$C7,$E1
	.dc.b	$3E,$CA,$E1,$F2,$7C,$FF,$C8,$65,$F9,$16,$C6,$4F,$9D,$19,$69,$F8
	.dc.b	$9F,$47,$F1,$E7,$E4,$3D,$DB,$19,$26,$9F,$89,$8C,$BE,$9F,$83,$FC
	.dc.b	$47,$8F,$F0,$DB,$5F,$05,$B3,$B8,$FC,$54,$ED,$D1,$F8,$3F,$8A,$C4
	.dc.b	$FE,$5F,$E4,$BF,$47,$F0,$DF,$A4,$6D,$7E,$85,$B3,$B8,$FD,$24,$FE
	.dc.b	$CF,$C4,$91,$F8,$FF,$56,$D7,$E0,$26,$F0,$FE,$DE,$6D,$7F,$39,$FB
	.dc.b	$26,$5F,$B4,$FA,$3F,$E3,$71,$89,$FD,$94,$62,$75,$1F,$E7,$9B,$FF
	.dc.b	$4E,$D6,$F1,$FD,$95,$EE,$B4,$77,$BF,$E0,$3D,$99,$AF,$F2,$87,$D9
	.dc.b	$3E,$CB,$F8,$87,$73,$DF,$FE,$5B,$39,$C7,$F8,$CC,$E5,$EB,$7F,$E8
	.dc.b	$BF,$4F,$FA,$76,$BF,$D0,$29,$F4,$3F,$CD,$A7,$37,$E9,$9C,$BC,$AD
	.dc.b	$F5,$EF,$B2,$FE,$45,$F6,$4E,$2F,$E7,$F3,$9B,$FE,$57,$39,$79,$47
	.dc.b	$F0,$07,$F9,$7F,$D9,$7F,$A7,$DC,$E3,$FC,$B6,$31,$8F,$F9,$C6,$9E
	.dc.b	$7F,$F3,$1F,$B2,$7C,$07,$82,$30,$FF,$50,$F6,$1D,$3F,$CD,$FE,$C9
	.dc.b	$E5,$66,$30,$EF,$D7,$FE,$79,$F6,$4F,$57,$1C,$1F,$D1,$23,$0F,$C3
	.dc.b	$F8,$47,$D6,$7F,$43,$F1,$07,$F6,$08,$CC,$FF,$4C,$8C,$BC,$7D,$EE
	.dc.b	$1F,$01,$B6,$C7,$5E,$3F,$E9,$DF,$65,$EA,$8C,$1F,$D5,$9C,$47,$FA
	.dc.b	$56,$3E,$19,$9F,$9B,$FE,$E2,$E2,$3F,$AD,$4A,$7F,$BA,$FD,$97,$C4
	.dc.b	$AF,$EC,$3E,$9F,$F3,$AE,$FF,$F7,$AC,$E7,$0F,$EC,$53,$9A,$FF,$0F
	.dc.b	$7D,$83,$DA,$FE,$FF,$C8,$FA,$3F,$F0,$0C,$E6,$BF,$15,$39,$7F,$19
	.dc.b	$0F,$B2,$DF,$D3,$70,$6C,$FE,$47,$EC,$BF,$DB,$2F,$08,$FF,$06,$C6
	.dc.b	$33,$FF,$5B,$5F,$3A,$E1,$E6,$FF,$84,$BE,$C9,$E7,$F0,$8C,$3F,$E1
	.dc.b	$6F,$89,$7B,$99,$BF,$D0,$46,$1E,$66,$23,$81,$D7,$D0,$B1,$FA,$7F
	.dc.b	$B1,$FC,$0E,$7F,$E0,$9F,$B2,$7C,$9F,$F8,$92,$73,$CD,$FE,$32,$9C
	.dc.b	$DF,$1E,$DF,$F0,$74,$63,$0F,$F6,$95,$F0,$AF,$E1,$D3,$9B,$47,$F9
	.dc.b	$46,$75,$40,$E2,$55,$F6,$E7,$53,$C1,$DF,$17,$F8,$7F,$EC,$BF,$AB
	.dc.b	$E1,$D6,$31,$EE,$7A,$DB,$1F,$05,$B5,$E8,$F4,$3F,$C3,$5F,$1F,$F0
	.dc.b	$5B,$5F,$1D,$C3,$F2,$66,$9F,$EF,$1C,$D5,$FE,$4C,$7F,$67,$19,$8C
	.dc.b	$78,$79,$65,$7F,$19,$9F,$F7,$78,$FB,$17,$D8,$7F,$9D,$F0,$EB,$5B
	.dc.b	$A7,$93,$F2,$6F,$71,$FC,$F7,$0F,$9E,$F3,$C3,$FD,$00,$F6,$9D,$AC
	.dc.b	$3A,$BB,$37,$7D,$6B,$C5,$7D,$2F,$8D,$73,$9C,$E1,$B8,$E2,$73,$3B
	.dc.b	$57,$19,$F1,$C7,$A8,$D6,$1E,$2B,$EB,$19,$7C,$F1,$99,$E3,$87,$7A
	.dc.b	$EB,$85,$7C,$87,$C3,$70,$F3,$9F,$3B,$E5,$3E,$89,$FE,$AC,$9F,$F5
	.dc.b	$6B,$83,$B9,$E8,$FB,$2F,$E7,$BF,$1B,$FE,$B2,$6C,$67,$CC,$4E,$5B
	.dc.b	$1C,$7E,$37,$E8,$37,$2E,$51,$88,$FA,$F6,$8C,$47,$FA,$E6,$31,$3F
	.dc.b	$07,$1F,$EC,$F3,$76,$FF,$5F,$69,$F7,$AF,$73,$F4,$DB,$1F,$82,$AE
	.dc.b	$33,$F5,$DB,$1C,$70,$F3,$B7,$0E,$D5,$0E,$93,$AA,$F3,$36,$3F,$7D
	.dc.b	$1D,$0F,$F6,$91,$5F,$E0,$57,$E4,$8E,$35,$D6,$FC,$D8,$7B,$27,$37
	.dc.b	$E7,$8F,$D3,$69,$D1,$B1,$E0,$6A,$C3,$B3,$93,$FC,$15,$5F,$0B,$0F
	.dc.b	$D4,$C7,$EC,$AB,$8D,$70,$87,$D9,$7F,$00,$7F,$C2,$4D,$7F,$59,$87
	.dc.b	$FB,$BE,$31,$7D,$1F,$B9,$3C,$36,$F2,$C6,$AB,$84,$63,$3F,$4F,$D9
	.dc.b	$7F,$42,$BF,$F1,$03,$5E,$E7,$F2,$B7,$18,$7A,$BE,$C7,$FA,$D4,$F0
	.dc.b	$EA,$F8,$71,$FA,$CC,$7F,$99,$4A,$30,$F3,$9F,$0F,$85,$23,$C2,$F2
	.dc.b	$FB,$2C,$47,$88,$63,$FE,$1E,$76,$7F,$8D,$ED,$3C,$31,$F3,$72,$F9
	.dc.b	$0C,$65,$E4,$AF,$F6,$0D,$FF,$8F,$FE,$C9,$5F,$71,$B1,$D7,$E3,$31
	.dc.b	$32,$38,$3F,$02,$3F,$07,$E8,$FF,$B7,$A7,$FC,$94,$D7,$D4,$DC,$7F
	.dc.b	$07,$7D,$A7,$3E,$4F,$22,$3E,$DF,$27,$AF,$93,$23,$AB,$A7,$E3,$7D
	.dc.b	$97,$F8,$01,$FF,$2D,$35,$CF,$8B,$F9,$88,$FB,$7B,$7D,$6F,$B2,$7C
	.dc.b	$2F,$E7,$1B,$7F,$E6,$36,$BF,$4F,$9B,$FC,$58,$FF,$99,$FE,$CB,$FC
	.dc.b	$CC,$DF,$7F,$F4,$B1,$88,$E1,$F4,$98,$E9,$91,$C5,$FD,$5D,$F9,$23
	.dc.b	$FB,$D5,$7F,$98,$9E,$DE,$D7,$43,$F6,$3F,$62,$FB,$2F,$D9,$C7,$E1
	.dc.b	$67,$F1,$0D,$E4,$FD,$8B,$76,$FC,$AD,$FF,$9B,$2B,$FD,$01,$3E,$C7
	.dc.b	$8F,$A9,$D0,$EF,$F2,$FF,$C0,$3E,$CB,$FC,$C7,$3F,$E8,$76,$BF,$97
	.dc.b	$6E,$3F,$C0,$3E,$CB,$FB,$3B,$FC,$0A,$1E,$CF,$32,$57,$FA,$46,$BF
	.dc.b	$36,$FF,$A4,$FE,$CB,$FD,$6A,$DF,$8C,$FE,$23,$18,$8F,$D3,$4F,$F9
	.dc.b	$55,$E1,$E4,$E8,$8F,$5F,$AE,$FE,$33,$F6,$1F,$E8,$68,$FF,$50,$35
	.dc.b	$FD,$83,$73,$FE,$33,$F6,$4F,$B2,$E2,$4A,$E2,$FF,$24,$8F,$9D,$1F
	.dc.b	$E3,$3F,$A2,$7C,$57,$F2,$86,$C7,$FA,$BD,$AF,$EF,$FE,$BE,$E1,$FC
	.dc.b	$A2,$6F,$2F,$F0,$C6,$D7,$5B,$FE,$67,$78,$84,$D8,$FD,$85,$7F,$14
	.dc.b	$6C,$F1,$3F,$D8,$1F,$64,$F9,$5F,$E7,$AD,$8C,$C6,$3D,$1C,$B2,$32
	.dc.b	$72,$8E,$3E,$DF,$2F,$92,$93,$DA,$32,$F8,$23,$2F,$F1,$FB,$FF,$34
	.dc.b	$BF,$17,$E9,$3E,$41,$1F,$3F,$0D,$7F,$93,$68,$9C,$6D,$FD,$42,$D7
	.dc.b	$8D,$7F,$9A,$FD,$97,$F5,$4A,$FE,$AA,$D8,$F9,$D7,$FD,$7A,$1F,$65
	.dc.b	$49,$97,$FB,$9E,$1B,$E2,$3F,$9C,$47,$07,$53,$CE,$7A,$7C,$CF,$5F
	.dc.b	$8B,$95,$CE,$32,$F5,$8C,$BD,$23,$FB,$8F,$D9,$3C,$5A,$49,$FF,$7D
	.dc.b	$7D,$97,$F0,$57,$FD,$F9,$3B,$6C,$ED,$FE,$9D,$3B,$7F,$A8,$4E,$DC
	.dc.b	$E7,$FE,$00,$9F,$9D,$1C,$8E,$FF,$0D,$9F,$96,$91,$82,$1D,$9C,$BE
	.dc.b	$08,$E9,$E0,$C8,$F4,$BD,$FE,$9F,$11,$E6,$8B,$C3,$FE,$0C,$C3,$F3
	.dc.b	$72,$FB,$17,$D9,$7F,$85,$9B,$1F,$EA,$5A,$FF,$09,$C6,$23,$FB,$77
	.dc.b	$D9,$7F,$7C,$7F,$B8,$46,$67,$75,$CC,$FF,$10,$D6,$E3,$35,$D7,$93
	.dc.b	$FC,$03,$3F,$E0,$08,$36,$FA,$4E,$BA,$79,$92,$B9,$1F,$EB,$BA,$74
	.dc.b	$E6,$3C,$F3,$70,$FF,$16,$D6,$A7,$89,$E0,$FB,$2F,$39,$1C,$FD,$97
	.dc.b	$E7,$DE,$D1,$97,$E1,$BF,$E3,$E8,$D4,$3E,$CB,$F9,$CB,$D9,$AF,$F8
	.dc.b	$5D,$B7,$FE,$A8,$F8,$57,$CB,$39,$FF,$C9,$35,$AD,$7F,$C4,$EB,$E7
	.dc.b	$B2,$6B,$CA,$FF,$27,$FD,$8F,$DE,$27,$2F,$2D,$32,$C7,$E0,$C7,$E1
	.dc.b	$2F,$FA,$8E,$1F,$E3,$48,$7D,$2F,$42,$57,$D0,$D3,$CF,$58,$8F,$F1
	.dc.b	$95,$3A,$3C,$66,$76,$FF,$2E,$C7,$F6,$B9,$DF,$A9,$E8,$3D,$3E,$54
	.dc.b	$8F,$2C,$65,$F3,$C5,$7F,$D7,$73,$FE,$5C,$6C,$FF,$AA,$A3,$11,$FE
	.dc.b	$D2,$75,$5F,$E6,$EF,$B2,$F9,$92,$0E,$14,$F0,$C7,$E2,$7C,$67,$CE
	.dc.b	$F5,$A5,$7C,$A8,$FA,$26,$1F,$87,$18,$7E,$7B,$FE,$60,$F8,$CF,$E0
	.dc.b	$B1,$DB,$BF,$07,$D1,$2B,$E5,$C3,$FC,$FD,$3C,$1B,$3B,$9F,$54,$E6
	.dc.b	$38,$C7,$AA,$AD,$70,$8E,$93,$83,$33,$23,$9C,$75,$7F,$E0,$0B,$6B
	.dc.b	$C5,$3B,$9C,$9D,$0E,$25,$32,$78,$CC,$94,$CD,$7C,$49,$F5,$46,$A3
	.dc.b	$D0,$EE,$32,$F0,$AF,$14,$79,$23,$84,$7B,$5E,$55,$A8,$F5,$12,$3E
	.dc.b	$74,$65,$F4,$4F,$63,$A3,$E6,$8E,$8F,$84,$EA,$43,$04,$8E,$24,$9F
	.dc.b	$04,$7E,$04,$E1,$19,$7C,$31,$AA,$F8,$24,$7C,$33,$B9,$F8,$91,$B3
	.dc.b	$D4,$E0,$E6,$43,$04,$9F,$A1,$1D,$EB,$8C,$62,$78,$3E,$28,$F0,$CE
	.dc.b	$E7,$D2,$F8,$5A,$7C,$C3,$E9,$90,$F9,$46,$8F,$84,$7C,$77,$F8,$94
	.dc.b	$78,$9E,$F1,$CC,$E2,$6C,$F2,$C6,$DE,$D5,$F1,$6B,$D5,$1F,$BC,$FB
	.dc.b	$2F,$D1,$9D,$1F,$0C,$FE,$B6,$33,$1C,$AF,$FD,$AF,$F6,$5F,$E6,$1A
	.dc.b	$FD,$CA,$F9,$C7,$CB,$7E,$72,$30,$DF,$9E,$FD,$9C,$71,$BE,$75,$C3
	.dc.b	$0F,$D2,$CF,$06,$D7,$09,$D4,$7F,$BA,$04,$ED,$C3,$FE,$5C,$E8,$D4
	.dc.b	$7F,$1D,$8D,$3E,$AF,$37,$FA,$53,$E0,$6E,$5F,$C5,$3A,$7F,$D7,$FC
	.dc.b	$1E,$B9,$CE,$7F,$EF,$8F,$B2,$7C,$2A,$BF,$B5,$C3,$C6,$C6,$5F,$EF
	.dc.b	$58,$DC,$EA,$73,$CB,$AA,$56,$9E,$13,$AA,$F0,$5D,$9F,$E7,$15,$DE
	.dc.b	$7C,$3E,$3E,$7A,$9F,$56,$F8,$8E,$18,$76,$AD,$4E,$6F,$89,$27,$D1
	.dc.b	$D5,$C3,$15,$DF,$77,$1D,$57,$0F,$82,$DC,$3E,$03,$76,$F7,$60,$8F
	.dc.b	$54,$3E,$BB,$D5,$48,$F8,$B1,$86,$F8,$3C,$23,$18,$7F,$8D,$DB,$5C
	.dc.b	$5B,$5D,$6B,$87,$07,$83,$EC,$5F,$65,$4F,$1F,$C5,$7E,$6B,$2F,$F1
	.dc.b	$6E,$0A,$DC,$61,$B9,$79,$E3,$FC,$91,$18,$9F,$BB,$5C,$EB,$47,$B6
	.dc.b	$75,$5F,$DC,$61,$F6,$4F,$AC,$F8,$BF,$64,$F8,$FF,$B3,$39,$BF,$2C
	.dc.b	$E5,$FF,$1D,$C6,$6F,$BE,$4A,$D4,$78,$8E,$F8,$76,$C3,$AB,$30,$F8
	.dc.b	$3B,$7F,$B5,$1E,$B1,$F2,$7B,$E4,$7C,$3F,$95,$B8,$78,$34,$56,$33
	.dc.b	$FC,$FC,$2B,$17,$F6,$7E,$CB,$EE,$75,$7D,$D6,$C7,$89,$B8,$FB,$33
	.dc.b	$46,$31,$FD,$57,$D9,$3E,$73,$F1,$BD,$4F,$B1,$FD,$0C,$E7,$2F,$BB
	.dc.b	$39,$7F,$CC,$71,$9C,$BE,$4C,$E5,$F9,$11,$9C,$3F,$73,$39,$7C,$38
	.dc.b	$76,$3F,$1D,$F6,$57,$0F,$97,$E6,$FE,$67,$2F,$CD,$36,$32,$7D,$08
	.dc.b	$CB,$4F,$C8,$79,$FF,$99,$3F,$35,$ED,$D8,$C9,$34,$FA,$51,$97,$D3
	.dc.b	$F0,$5F,$8C,$F9,$1F,$86,$DA,$F8,$2D,$9D,$C7,$E3,$A7,$6E,$8F,$C2
	.dc.b	$FC,$56,$27,$F0,$1F,$27,$FA,$DF,$9A,$FD,$73,$6B,$F5,$0D,$9D,$C7
	.dc.b	$EA,$E7,$F7,$9E,$1C,$8F,$C8,$FA,$96,$BF,$01,$37,$87,$F8,$2C,$DA
	.dc.b	$FD,$57,$D9,$32,$FD,$F7,$D2,$7F,$20,$8C,$4F,$EE,$63,$13,$A8,$FF
	.dc.b	$51,$CE,$B1,$FC,$D5,$6A,$73,$5F,$C2,$7B,$31,$5F,$AB,$FB,$17,$D9
	.dc.b	$7F,$1C,$F8,$BF,$F5,$9E,$1F,$EA,$E9,$CB,$FC,$32,$33,$5F,$B3,$F4
	.dc.b	$BF,$A7,$BF,$B8,$87,$D0,$FF,$38,$9C,$ED,$FD,$22,$73,$7F,$EB,$AF
	.dc.b	$B0,$7C,$C7,$F2,$CF,$B2,$FF,$49,$4E,$63,$1C,$3F,$D0,$AB,$F7,$99
	.dc.b	$FF,$06,$FB,$2F,$F4,$15,$F3,$C7,$F9,$AC,$62,$7F,$84,$53,$CF,$FE
	.dc.b	$73,$F6,$4F,$7F,$A4,$61,$FE,$9F,$EC,$3A,$3F,$9F,$FD,$97,$F0,$4D
	.dc.b	$3F,$DA,$33,$97,$F9,$7C,$67,$2F,$E2,$D3,$FD,$55,$FE,$37,$F6,$5F
	.dc.b	$D2,$6B,$8F,$AD,$FD,$2A,$31,$3F,$C7,$7E,$0D,$F5,$9F,$D3,$BC,$11
	.dc.b	$FD,$8A,$33,$87,$F3,$E8,$CB,$FE,$E4,$87,$B9,$C7,$DE,$E0,$D9,$D6
	.dc.b	$1F,$D6,$BE,$CB,$D4,$F8,$4C,$6D,$FD,$C2,$7E,$31,$FD,$32,$5F,$39
	.dc.b	$9E,$0E,$4F,$F7,$6F,$B2,$F8,$5D,$5F,$E0,$6F,$43,$F9,$CB,$FE,$07
	.dc.b	$8C,$E1,$FE,$B9,$8F,$F0,$9F,$A9,$FD,$FB,$EC,$1E,$C7,$F8,$03,$47
	.dc.b	$D2,$66,$30,$EF,$EC,$B8,$7D,$3D,$CB,$A1,$FD,$57,$04,$7F,$82,$BE
	.dc.b	$CB,$FD,$CB,$3A,$E1,$FF,$65,$52,$F5,$E8,$F3,$9F,$EA,$17,$FE,$0A
	.dc.b	$BE,$10,$FB,$27,$9D,$D7,$CE,$FF,$0B,$C6,$2B,$7F,$11,$FD,$25,$E3
	.dc.b	$87,$3C,$3F,$AE,$4E,$5E,$F7,$FE,$20,$8C,$70,$FF,$8B,$DB,$7F,$DC
	.dc.b	$70,$79,$DB,$F6,$BF,$C4,$FF,$63,$F8,$6C,$BF,$B5,$FD,$93,$E4,$F3
	.dc.b	$E0,$7F,$86,$A3,$19,$7B,$2F,$86,$9F,$E3,$E9,$CE,$1F,$DC,$67,$35
	.dc.b	$F6,$67,$35,$FE,$43,$9C,$D7,$FB,$D5,$B0,$9F,$C1,$D7,$F8,$26,$77
	.dc.b	$47,$EB,$C7,$F8,$6B,$EC,$BF,$AC,$3B,$E4,$FF,$18,$C6,$2B,$B4,$63
	.dc.b	$DA,$F8,$2D,$AF,$67,$A1,$FE,$26,$F8,$FF,$8A,$D8,$C9,$2F,$DC,$69
	.dc.b	$FB,$F9,$AB,$FC,$A6,$FE,$E0,$FF,$31,$19,$8D,$FC,$3E,$A6,$47,$E2
	.dc.b	$2B,$FD,$FE,$3E,$C9,$F4,$3F,$E2,$BC,$39,$4F,$F8,$7C,$DC,$E9,$FD
	.dc.b	$45,$FF,$9D,$A3,$11,$FE,$33,$F2,$8F,$5B,$E9,$5B,$7F,$D1,$B3,$AA
	.dc.b	$E3,$7E,$07,$9D,$F8,$2B,$8E,$1B,$78,$C7,$53,$C7,$19,$7F,$D3,$D7
	.dc.b	$5F,$F3,$26,$1E,$17,$27,$13,$33,$B7,$C0,$74,$34,$60,$D9,$81,$3D
	.dc.b	$F0,$F5,$5F,$9A,$7F,$D5,$66,$8A,$53,$33,$AF,$B2,$FE,$89,$F1,$B7
	.dc.b	$C5,$34,$95,$FE,$B6,$97,$C6,$FF,$A0,$63,$0D,$7F,$D8,$11,$AC,$7B
	.dc.b	$4E,$67,$FC,$D5,$39,$7E,$26,$3F,$DA,$9F,$94,$D7,$E4,$46,$F4,$FC
	.dc.b	$0B,$67,$FC,$E8,$6E,$3F,$0A,$D3,$BD,$79,$E3,$8B,$63,$8E,$DF,$E7
	.dc.b	$D9,$E2,$D8,$E3,$87,$DD,$BF,$06,$1F,$90,$9E,$CD,$AE,$0F,$E5,$27
	.dc.b	$69,$B3,$C2,$3F,$47,$3C,$1D,$61,$E5,$7A,$1F,$CC,$A7,$E2,$5F,$FB
	.dc.b	$4D,$BA,$79,$A7,$E5,$DF,$EB,$63,$DB,$1C,$63,$0F,$96,$FB,$2F,$B2
	.dc.b	$FE,$00,$FF,$40,$6B,$FE,$93,$6E,$1F,$EE,$37,$F0,$71,$BD,$BA,$BC
	.dc.b	$DB,$1C,$30,$FF,$55,$3C,$27,$FD,$3D,$F6,$5F,$E8,$F7,$FA,$7B,$5C
	.dc.b	$F4,$74,$FB,$1F,$EB,$11,$A8,$C6,$92,$30,$72,$F1,$7C,$77,$FE,$1D
	.dc.b	$8C,$1F,$D7,$2D,$F6,$5A,$A9,$3A,$3F,$D6,$53,$A7,$1A,$33,$FF,$62
	.dc.b	$4F,$F7,$09,$E9,$CB,$C9,$E3,$E3,$FC,$4F,$B2,$D4,$7F,$AF,$5B,$87
	.dc.b	$38,$FE,$37,$7F,$EC,$B7,$FB,$8C,$EB,$0F,$F6,$93,$63,$F9,$0B,$54
	.dc.b	$F1,$7F,$D9,$5E,$93,$E4,$7E,$AB,$63,$FB,$EB,$5F,$26,$5A,$F3,$24
	.dc.b	$69,$FA,$5E,$5F,$2F,$2E,$4F,$27,$6F,$FB,$67,$EC,$BF,$C0,$EF,$F8
	.dc.b	$29,$AE,$7C,$9F,$C3,$C7,$DF,$DB,$E1,$4B,$EC,$9F,$0B,$F8,$C6,$C7
	.dc.b	$F8,$51,$AF,$FB,$9F,$A3,$FC,$6E,$FF,$86,$3E,$CB,$FC,$DE,$DF,$7F
	.dc.b	$F2,$B1,$88,$E1,$F4,$B8,$EC,$91,$C5,$FC,$DD,$F9,$79,$7F,$A7,$CA
	.dc.b	$FF,$CD,$CF,$7F,$6F,$A9,$FA,$2F,$B1,$7D,$97,$E9,$23,$F2,$53,$FE
	.dc.b	$FD,$6E,$DF,$01,$B8,$76,$BF,$F3,$C5,$7F,$8C,$A7,$DA,$F1,$F5,$BA
	.dc.b	$1D,$FE,$5F,$F6,$3F,$65,$FE,$6C,$9F,$F1,$CB,$56,$3F,$3A,$30,$FD
	.dc.b	$0F,$B1,$F6,$CE,$B2,$F7,$D7,$F9,$BA,$39,$FB,$7D,$8F,$DF,$D7,$E9
	.dc.b	$63,$81,$FD,$D3,$EC,$BF,$D7,$CF,$F9,$29,$AE,$83,$8C,$3C,$5D,$1F
	.dc.b	$C2,$A3,$E0,$CE,$DB,$3F,$86,$6E,$3F,$52,$DF,$5D,$FC,$4B,$EC,$9F
	.dc.b	$33,$49,$8F,$F9,$73,$EC,$5F,$65,$B2,$47,$F9,$82,$B2,$EF,$C5,$FE
	.dc.b	$3B,$F4,$2F,$8B,$A4,$9F,$F3,$3F,$D7,$E5,$FF,$34,$CD,$7E,$03,$77
	.dc.b	$FC,$FB,$76,$FC,$6D,$5A,$FF,$0E,$CD,$51,$FB,$7A,$FD,$3B,$6B,$F2
	.dc.b	$B5,$BF,$B2,$7C,$B7,$F3,$46,$E1,$FE,$7C,$6B,$F4,$DF,$F1,$2D,$7E
	.dc.b	$79,$CE,$1F,$B4,$6C,$66,$BA,$79,$BD,$7D,$CE,$DF,$59,$31,$ED,$19
	.dc.b	$7C,$11,$97,$C1,$1F,$33,$E9,$24,$7C,$87,$D1,$C7,$88,$F9,$31,$73
	.dc.b	$F2,$4D,$CB,$F6,$BF,$65,$FE,$56,$69,$9B,$FF,$23,$C3,$EC,$BF,$AB
	.dc.b	$B6,$3F,$D3,$ED,$7F,$2C,$FD,$07,$31,$C0,$8F,$0D,$F9,$FC,$DF,$4F
	.dc.b	$8E,$93,$D2,$32,$F6,$8C,$BD,$A3,$FB,$4F,$D9,$3C,$6A,$49,$FF,$59
	.dc.b	$FD,$97,$F0,$C7,$FD,$6B,$3B,$6C,$ED,$FD,$74,$ED,$FD,$84,$ED,$CE
	.dc.b	$DF,$DE,$24,$F1,$DB,$F9,$9C,$3E,$23,$5F,$2D,$2F,$C7,$1C,$4C,$3E
	.dc.b	$08,$C3,$E8,$F0,$BB,$11,$F5,$FA,$75,$F3,$C5,$C3,$C3,$3F,$E0,$CF
	.dc.b	$B1,$7D,$95,$25,$FF,$B5,$BE,$CB,$FC,$0A,$FF,$0D,$9D,$BF,$ED,$A6
	.dc.b	$B9,$9F,$F0,$BC,$F0,$BF,$EA,$35,$A7,$FA,$9C,$7E,$92,$BF,$D1,$ED
	.dc.b	$53,$BE,$4F,$E8,$B3,$FE,$D1,$9F,$27,$37,$63,$BF,$3D,$8C,$90,$FE
	.dc.b	$49,$28,$F2,$FD,$94,$9F,$73,$FE,$2E,$FB,$2F,$CF,$47,$92,$34,$F2
	.dc.b	$75,$87,$F8,$D3,$EC,$BF,$9D,$4F,$F3,$16,$E7,$FD,$C1,$B3,$FE,$AC
	.dc.b	$F8,$47,$CB,$B9,$F9,$79,$5C,$FE,$7A,$65,$FE,$B2,$CB,$94,$FF,$90
	.dc.b	$BE,$C7,$F8,$A4,$FF,$AD,$9A,$9B,$9F,$F5,$0A,$FF,$00,$C6,$89,$C7
	.dc.b	$E9,$AE,$9F,$29,$1F,$06,$BF,$DD,$29,$3F,$E5,$ED,$7D,$33,$BA,$FE
	.dc.b	$A1,$4E,$6F,$31,$9C,$7C,$3F,$13,$FE,$A8,$9F,$EA,$35,$FD,$91,$AF
	.dc.b	$78,$D1,$EF,$F5,$A4,$79,$0F,$81,$18,$7C,$31,$6B,$BC,$6C,$F8,$91
	.dc.b	$E1,$23,$FE,$BC,$AF,$F3,$43,$E3,$F8,$BF,$31,$1D,$35,$F7,$FA,$FF
	.dc.b	$4C,$FA,$91,$63,$C1,$3F,$E7,$57,$EB,$4E,$9C,$BB,$79,$9C,$27,$17
	.dc.b	$F6,$26,$E5,$F4,$E6,$D7,$43,$99,$C4,$A7,$42,$57,$E0,$A3,$8C,$7E
	.dc.b	$1E,$AD,$A6,$C6,$1B,$3D,$2B,$93,$6F,$F0,$D5,$B8,$C1,$E5,$3C,$31
	.dc.b	$E9,$72,$78,$4B,$7E,$16,$CF,$37,$D4,$7E,$1D,$FC,$51,$1F,$B4,$D7
	.dc.b	$FC,$5E,$75,$2C,$7F,$AC,$1E,$40,$FF,$80,$3D,$BF,$9F,$3F,$CA,$7F
	.dc.b	$17,$FC,$E2,$3F,$5D,$18,$3D,$EF,$18,$D3,$DD,$B3,$86,$CF,$19,$F8
	.dc.b	$91,$AA,$C4,$7E,$E6,$6D,$71,$7B,$C7,$98,$E2,$F9,$DF,$13,$5D,$1E
	.dc.b	$37,$8D,$78,$A3,$67,$B0,$A6,$CE,$A6,$4E,$73,$E3,$72,$6C,$F4,$1C
	.dc.b	$9D,$39,$7F,$DB,$F1,$E1,$8E,$73,$EE,$A9,$1C,$63,$E0,$98,$7C,$31
	.dc.b	$87,$47,$77,$53,$D4,$93,$F0,$09,$3E,$B8,$CB,$EA,$AF,$39,$27,$99
	.dc.b	$F2,$EF,$F8,$B3,$FE,$96,$6C,$FE,$25,$AF,$FA,$B8,$7D,$87,$F9,$99
	.dc.b	$D1,$FE,$9E,$70,$ED,$B1,$FE,$86,$6C,$6F,$D0,$FC,$2C,$3E,$69,$F2
	.dc.b	$BF,$62,$30,$D9,$DC,$EA,$BF,$07,$5A,$6C,$67,$2F,$CC,$CF,$91,$B1
	.dc.b	$C6,$7C,$8D,$7D,$5E,$6B,$9F,$F5,$F1,$F0,$27,$4F,$EB,$7E,$C1,$F0
	.dc.b	$7F,$EF,$3E,$0F,$F6,$9C,$3C,$3C,$6D,$FE,$C2,$8D,$BF,$CE,$5D,$36
	.dc.b	$B7,$7D,$6A,$AA,$E2,$FF,$A8,$DE,$2D,$F1,$38,$F3,$BE,$A5,$F0,$DC
	.dc.b	$AF,$FC,$69,$97,$F4,$FA,$D4,$E6,$E4,$F9,$79,$EE,$3D,$FD,$CE,$5D
	.dc.b	$1C,$DB,$5E,$B6,$CE,$E3,$DD,$3E,$CD,$9F,$59,$E8,$7E,$43,$67,$FC
	.dc.b	$08,$BB,$78,$39,$BF,$F6,$9B,$6B,$9E,$FA,$AD,$BD,$FE,$B5,$F6,$14
	.dc.b	$F1,$7B,$BF,$8E,$CB,$FB,$86,$0A,$DC,$61,$B9,$7F,$B5,$A7,$57,$CB
	.dc.b	$AB,$A5,$FF,$73,$FB,$07,$D4,$FB,$FE,$C1,$F1,$FF,$6A,$73,$5E,$09
	.dc.b	$CB,$C6,$B8,$56,$DB,$19,$26,$5E,$3D,$FF,$D9,$91,$8C,$FF,$25,$A7
	.dc.b	$FB,$6A,$3D,$5D,$DF,$0F,$E4,$BE,$23,$23,$27,$BE,$31,$87,$5F,$B0
	.dc.b	$FA,$3D,$7F,$49,$B5,$F2,$32,$46,$2B,$84,$E6,$BF,$33,$F6,$0F,$99
	.dc.b	$FC,$17,$6B,$EB,$F3,$18,$6D,$FD,$49,$D3,$FE,$1F,$AE,$4E,$67,$78
	.dc.b	$36,$4F,$F6,$18,$57,$07,$ED,$6C,$FB,$0E,$37,$F9,$2F,$2B,$F1,$F8
	.dc.b	$FE,$41,$B8,$FB,$DB,$87,$D1,$6D,$7D,$0F,$37,$F2,$07,$E4,$7D,$9B
	.dc.b	$7F,$8B,$8B,$7E,$29,$B5,$E8,$8F,$EF,$19,$FF,$BC,$AF,$F0,$DF,$1F
	.dc.b	$F2,$BE,$03,$F1,$8D,$8C,$9F,$16,$32,$D2,$57,$E1,$FE,$BF,$83,$F0
	.dc.b	$BE,$4B,$F4,$DF,$0B,$FA,$76,$D7,$E8,$5B,$3B,$8F,$D2,$4F,$ED,$3C
	.dc.b	$49,$1F,$83,$F4,$6D,$7E,$56,$6E,$9F,$BE,$9B,$5F,$57,$EC,$19,$7E
	.dc.b	$DF,$E8,$BF,$8F,$C6,$27,$F6,$51,$89,$D4,$7F,$98,$67,$58,$FD,$CA
	.dc.b	$D4,$E6,$BF,$81,$76,$62,$BF,$0D,$3F,$85,$FA,$F7,$D8,$7F,$15,$F8
	.dc.b	$9F,$F5,$54,$FF,$9B,$67,$2F,$F0,$68,$CD,$72,$3F,$C5,$DE,$7B,$8B
	.dc.b	$97,$CF,$7F,$38,$8C,$5F,$18,$FE,$77,$F5,$CF,$96,$FE,$49,$F6,$1F
	.dc.b	$E8,$DA,$FF,$3E,$4E,$5F,$E2,$B1,$9C,$7C,$02,$7F,$80,$BF,$C5,$7E
	.dc.b	$C3,$FB,$9F,$1F,$E2,$E3,$5A,$7F,$31,$9C,$E1,$FA,$5B,$79,$FF,$CD
	.dc.b	$7E,$C1,$EE,$79,$A3,$0F,$F5,$2F,$61,$D3,$FC,$EF,$EC,$3F,$87,$69
	.dc.b	$E4,$9C,$E5,$FC,$B2,$73,$AF,$8E,$7F,$A9,$CF,$F3,$0F,$B0,$FF,$3A
	.dc.b	$3F,$E9,$B8,$CD,$FF,$2F,$9C,$BF,$03,$E0,$5F,$55,$FD,$2F,$C1,$1F
	.dc.b	$D9,$A3,$38,$7F,$3C,$8C,$BF,$EA,$78,$7C,$17,$1F,$73,$83,$67,$58
	.dc.b	$7F,$57,$FB,$0F,$5B,$E1,$31,$B7,$F7,$49,$F8,$E7,$F4,$A9,$7C,$D6
	.dc.b	$78,$39,$3F,$DE,$BE,$C3,$E1,$D7,$FB,$87,$CE,$FE,$83,$8F,$DC,$9C
	.dc.b	$BF,$C6,$E3,$3C,$3F,$6F,$EC,$1E,$CF,$F8,$13,$47,$D1,$7F,$81,$A7
	.dc.b	$39,$7F,$B1,$67,$2F,$FB,$1E,$1F,$61,$BF,$A4,$E0,$D9,$FF,$66,$7D
	.dc.b	$87,$FB,$91,$E1,$1F,$E1,$28,$C6,$9F,$92,$BE,$75,$C3,$CA,$FF,$0B
	.dc.b	$FD,$83,$CC,$E1,$18,$7F,$C3,$7E,$EB,$D9,$CD,$FF,$B6,$E3,$0F,$33
	.dc.b	$11,$C0,$EB,$E6,$58,$FF,$71,$7D,$7F,$E0,$33,$FE,$ED,$F6,$0F,$92
	.dc.b	$FF,$14,$CE,$79,$3F,$C6,$D3,$9B,$E3,$DD,$FE,$15,$8C,$61,$FA,$EB
	.dc.b	$E1,$5F,$EA,$BB,$56,$9F,$D7,$4E,$AB,$FD,$60,$D8,$DF,$BD,$E2,$C3
	.dc.b	$C7,$C7,$FE,$23,$FB,$0F,$EA,$95,$CA,$31,$EE,$6B,$3F,$F0,$AD,$69
	.dc.b	$FE,$B9,$8A,$B7,$1F,$0B,$CC,$FF,$0E,$FC,$77,$C0,$6C,$64,$9A,$7B
	.dc.b	$4D,$3E,$D3,$FB,$46,$4A,$FF,$28,$BF,$B2,$3F,$86,$19,$8E,$13,$F1
	.dc.b	$7E,$1B,$75,$27,$33,$FE,$D0,$1F,$60,$F9,$EE,$F5,$FE,$5E,$9F,$F6
	.dc.b	$81,$A8,$CD,$7F,$B6,$EB,$84,$FF,$7D,$1E,$46,$6B,$71,$C4,$D4,$65
	.dc.b	$FB,$57,$FD,$BE,$B9,$C7,$95,$B8,$FB,$63,$91,$9B,$EB,$3C,$23,$CE
	.dc.b	$F0,$AE,$93,$FE,$F6,$BF,$05,$7F,$85,$67,$2F,$90,$EB,$7D,$EF,$A4
	.dc.b	$7B,$0C,$CF,$07,$C4,$7A,$B0,$EA,$FC,$83,$FD,$2F,$7F,$E9,$87,$07
	.dc.b	$8C,$F9,$03,$EC,$3F,$9F,$FC,$67,$FA,$79,$B1,$A6,$BF,$3D,$AF,$F8
	.dc.b	$B5,$7C,$5F,$F8,$C6,$38,$35,$E0,$D7,$83,$51,$E3,$79,$63,$F5,$33
	.dc.b	$6B,$FE,$3B,$8C,$36,$39,$E9,$F3,$B0,$F1,$BF,$42,$38,$E1,$6B,$83
	.dc.b	$E0,$AD,$BF,$79,$B7,$DA,$38,$1D,$70,$F2,$DF,$82,$3F,$0A,$F6,$6D
	.dc.b	$F6,$7D,$98,$71,$39,$BA,$30,$53,$FB,$EC,$F3,$8F,$EF,$4D,$CF,$C1
	.dc.b	$87,$96,$B7,$7F,$B8,$9F,$F3,$05,$6E,$7D,$D8,$7A,$1F,$9F,$F6,$1F
	.dc.b	$BD,$7F,$6A,$D7,$FC,$1A,$DC,$7F,$C1,$4F,$DE,$8D,$EB,$DD,$E6,$D5
	.dc.b	$1C,$70,$FF,$39,$3C,$63,$FC,$33,$F6,$1F,$D0,$5F,$E0,$AD,$73,$D3
	.dc.b	$CF,$EB,$BF,$AB,$C6,$A3,$19,$C8,$C1,$CB,$C2,$F1,$5F,$F8,$5A,$30
	.dc.b	$7F,$12,$B7,$D8,$6A,$A4,$E8,$FF,$16,$CE,$9C,$69,$FE,$2B,$8F,$E6
	.dc.b	$11,$CF,$93,$8B,$C3,$A7,$E1,$7D,$86,$A3,$FC,$6E,$DB,$E5,$3A,$A9
	.dc.b	$5F,$C9,$AB,$5B,$7F,$A6,$70,$FF,$64,$CF,$F2,$AF,$91,$FF,$72,$37
	.dc.b	$E4,$7E,$84,$62,$3B,$E1,$8F,$26,$47,$17,$E6,$79,$3C,$3C,$99,$3D
	.dc.b	$5D,$BF,$E4,$EF,$B0,$FE,$F4,$FF,$3A,$6B,$9F,$17,$F0,$11,$F6,$77
	.dc.b	$FF,$28,$7D,$87,$F9,$8E,$BF,$A0,$B5,$FF,$2C,$B6,$3F,$CA,$DC,$FF
	.dc.b	$E2,$89,$FE,$8F,$F6,$1F,$E6,$56,$FB,$BF,$8D,$8C,$47,$F9,$8A,$5F
	.dc.b	$47,$9E,$B9,$3C,$9F,$CA,$4F,$97,$8F,$FD,$89,$7F,$E6,$0A,$EF,$EC
	.dc.b	$F5,$3F,$35,$F5,$EF,$B0,$FC,$E4,$7E,$2E,$7F,$CE,$2D,$D3,$FC,$DC
	.dc.b	$DC,$3F,$DA,$17,$FE,$6A,$AF,$EC,$33,$ED,$78,$FA,$7D,$0E,$FF,$2F
	.dc.b	$FA,$9F,$B0,$FF,$2E,$CF,$F6,$86,$BF,$E8,$06,$E3,$FE,$7D,$FA,$FF
	.dc.b	$EC,$4F,$F9,$FE,$3F,$CB,$B3,$C7,$D9,$E8,$7E,$D6,$BF,$39,$1F,$E8
	.dc.b	$AF,$B0,$FF,$58,$BF,$DD,$9A,$E7,$E2,$B8,$74,$7E,$F6,$7D,$93,$C7
	.dc.b	$DA,$C1,$D7,$EB,$3F,$81,$7D,$73,$E6,$69,$2B,$FC,$07,$F6,$0F,$B0
	.dc.b	$D9,$23,$FC,$0F,$59,$77,$E3,$7F,$0D,$FA,$07,$C5,$52,$4F,$F8,$3B
	.dc.b	$EB,$72,$FF,$84,$26,$BF,$EA,$86,$F1,$FF,$A9,$5B,$D1,$FD,$4A,$55
	.dc.b	$87,$FA,$AC,$D7,$3E,$37,$F1,$F9,$FD,$85,$F3,$C3,$F9,$17,$D8,$3E
	.dc.b	$56,$93,$0F,$F1,$1C,$7F,$83,$B2,$F0,$3B,$8C,$CE,$DF,$F5,$FC,$6D
	.dc.b	$FE,$CE,$FF,$97,$5E,$7F,$53,$E0,$3C,$3E,$EC,$8C,$90,$E6,$FC,$BF
	.dc.b	$A5,$91,$F1,$DF,$4F,$1E,$23,$E4,$45,$CF,$FD,$99,$82,$B1,$97,$C9
	.dc.b	$FB,$0F,$F2,$2B,$4C,$E1,$FD,$FA,$1F,$61,$FD,$1D,$B1,$FE,$46,$6B
	.dc.b	$FE,$DE,$7E,$9B,$98,$E0,$47,$86,$DC,$FE,$6F,$A7,$C6,$C9,$E9,$19
	.dc.b	$7B,$46,$5E,$D1,$FD,$67,$EC,$3F,$85,$C7,$F9,$61,$AF,$FB,$C1,$BA
	.dc.b	$7F,$BB,$7E,$C3,$F8,$34,$EF,$C9,$E4,$47,$FC,$C3,$1A,$79,$47,$F6
	.dc.b	$78,$6F,$4E,$13,$FE,$FE,$87,$C3,$EF,$E7,$24,$F4,$8E,$46,$1E,$11
	.dc.b	$87,$C3,$E2,$77,$23,$E4,$E7,$AF,$78,$B9,$7B,$67,$FB,$E7,$D7,$BE
	.dc.b	$C2,$92,$FF,$CF,$BF,$61,$FD,$D9,$5B,$93,$F5,$97,$C6,$7F,$D0,$6D
	.dc.b	$77,$1A,$BF,$F0,$55,$F3,$BE,$55,$FE,$46,$BE,$11,$88,$4F,$E9,$5B
	.dc.b	$EE,$79,$5E,$A7,$F9,$3A,$3F,$50,$E0,$E3,$D1,$E0,$3C,$3D,$36,$3D
	.dc.b	$B1,$5E,$12,$9F,$3F,$D7,$C9,$F8,$6F,$F8,$8F,$EC,$3F,$47,$1E,$18
	.dc.b	$D3,$C9,$D6,$9F,$E2,$9F,$B0,$F8,$0F,$F9,$93,$E1,$5D,$7F,$C4,$DB
	.dc.b	$DF,$BF,$92,$98,$FF,$AC,$A7,$2D,$AF,$E3,$18,$74,$8F,$F3,$7B,$FD
	.dc.b	$19,$F8,$4D,$DF,$E6,$37,$1F,$17,$CB,$3A,$FC,$DE,$6F,$9F,$E7,$66
	.dc.b	$3F,$EB,$EC,$38,$3F,$E7,$CF,$B0,$7D,$37,$F9,$1F,$5F,$5F,$D8,$DD
	.dc.b	$8F,$F9,$26,$B7,$7F,$EC,$86,$BF,$6A,$B8,$BE,$D9,$E1,$7C,$7E,$9F
	.dc.b	$97,$64,$AE,$51,$97,$A4,$65,$FC,$16,$09,$D4,$62,$75,$C1,$F8,$FA
	.dc.b	$D4,$F4,$39,$90,$E7,$5C,$A7,$85,$72,$E0,$FE,$9D,$7C,$AF,$67,$1A
	.dc.b	$DB,$D4,$C9,$E3,$35,$3C,$AF,$B3,$D2,$FB,$4F,$D1,$9C,$BF,$4E,$32
	.dc.b	$AF,$C5,$7E,$DA,$CC,$2B,$8D,$6E,$73,$18,$9E,$37,$E1,$AF,$29,$E1
	.dc.b	$3D,$57,$EC,$7D,$47,$B1,$F0,$3C,$0F,$25,$75,$BE,$F1,$CA,$BE,$03
	.dc.b	$BB,$F3,$39,$3C,$07,$6B,$F0,$1D,$9F,$04,$F9,$DD,$3D,$1C,$9F,$1C
	.dc.b	$ED,$5E,$FB,$F8,$AD,$8C,$9E,$13,$D2,$60,$E6,$60,$EE,$79,$AB,$C8
	.dc.b	$79,$AB,$C8,$E1,$F2,$39,$8F,$21,$D2,$FE,$8B,$88,$D3,$F5,$CF,$24
	.dc.b	$7C,$67,$C2,$64,$E5,$7E,$C7,$6D,$72,$7A,$E7,$C8,$F7,$AF,$39,$E6
	.dc.b	$23,$92,$1D,$E3,$8B,$F0,$DF,$3D,$7A,$A3,$53,$D1,$E3,$19,$25,$79
	.dc.b	$6B,$9C,$7B,$5E,$77,$F5,$6B,$DF,$3C,$63,$F1,$71,$F5,$3E,$C5,$F3
	.dc.b	$8F,$97,$FF,$21,$4F,$F9,$0E,$73,$18,$80,$87,$96,$E1,$D4,$E6,$FA
	.dc.b	$3C,$7E,$D7,$FD,$19,$AE,$9B,$19,$F6,$FF,$AE,$3F,$E5,$38,$CB,$E4
	.dc.b	$FB,$2F,$DA,$57,$8B,$6F,$EC,$51,$89,$FE,$4D,$F2,$8F,$2B,$FB,$4D
	.dc.b	$6B,$86,$E1,$FE,$5F,$C3,$84,$E6,$3F,$A5,$5F,$17,$B1,$BA,$E3,$87
	.dc.b	$C2,$BF,$F0,$AC,$FF,$51,$FA,$D7,$C4,$EA,$95,$A7,$F9,$D4,$EA,$BA
	.dc.b	$C7,$3F,$23,$85,$6F,$2F,$6F,$C1,$62,$7C,$30,$FA,$F7,$95,$86,$E3
	.dc.b	$D2,$75,$E5,$66,$A4,$E6,$B7,$97,$F8,$1A,$F7,$39,$AE,$FF,$5C,$FB
	.dc.b	$2A,$78,$BD,$DF,$C4,$57,$79,$D3,$67,$6F,$8E,$77,$87,$7A,$DB,$FE
	.dc.b	$13,$FB,$27,$D6,$FB,$BE,$C9,$F2,$5F,$2E,$73,$8F,$B6,$72,$FC,$AB
	.dc.b	$4E,$9F,$64,$E9,$ED,$7D,$6F,$A3,$2F,$E2,$79,$36,$BC,$DE,$FF,$C2
	.dc.b	$3E,$1F,$CA,$DA,$F8,$11,$67,$D9,$37,$7F,$C8,$F0,$3E,$CB,$EB,$7C
	.dc.b	$8F,$D7,$6E,$3F,$4D,$B8,$7D,$3A,$E1,$3F,$5E,$BF,$19,$F6,$2F,$9F
	.dc.b	$FC,$1F,$4F,$F8,$3A,$BE,$14,$E5,$FC,$14,$67,$0F,$F2,$04,$E5,$FC
	.dc.b	$04,$67,$2F,$F2,$64,$E5,$FD,$4C,$66,$BF,$D8,$D3,$96,$DF,$E1,$BE
	.dc.b	$CB,$CA,$FE,$4F,$C6,$FC,$A7,$91,$6B,$F2,$0D,$CB,$DC,$DC,$BE,$3B
	.dc.b	$F9,$0F,$3A,$57,$E3,$7C,$1B,$7F,$89,$9B,$7F,$BE,$E7,$FD,$C4,$E6
	.dc.b	$FE,$1F,$C0,$FE,$2F,$EB,$FE,$4B,$63,$27,$E2,$E3,$2D,$33,$3F,$22
	.dc.b	$7F,$5D,$F1,$3A,$8F,$CD,$7C,$A7,$E9,$BE,$17,$F4,$ED,$BF,$D2,$36
	.dc.b	$77,$1F,$A7,$9F,$DC,$78,$32,$3F,$69,$E2,$DA,$FC,$14,$DD,$3F,$17
	.dc.b	$36,$BF,$DF,$1F,$64,$CB,$F6,$FF,$4B,$FC,$46,$31,$3F,$B5,$8C,$4E
	.dc.b	$A3,$F5,$13,$AC,$7F,$69,$5A,$9C,$D7,$F0,$2E,$DC,$57,$FA,$27,$EC
	.dc.b	$5F,$65,$FC,$53,$BF,$FD,$09,$87,$EC,$E7,$2F,$F1,$38,$CD,$7F,$A4
	.dc.b	$7C,$E7,$17,$1F,$A0,$FE,$51,$18,$BE,$11,$FC,$AF,$EC,$1F,$31,$FC
	.dc.b	$8B,$EC,$BF,$9C,$D7,$F5,$4C,$55,$A8,$FE,$45,$3A,$9F,$EB,$32,$CB
	.dc.b	$F9,$5F,$D9,$7F,$5C,$BF,$E1,$53,$97,$F9,$54,$66,$BF,$D5,$12,$F9
	.dc.b	$7F,$E6,$7F,$64,$F9,$4F,$34,$61,$FE,$85,$F4,$EE,$9F,$E7,$3F,$64
	.dc.b	$F2,$B3,$18,$77,$EA,$7F,$3E,$FB,$27,$D5,$63,$7F,$E7,$91,$87,$F9
	.dc.b	$07,$C2,$3E,$B3,$FA,$2F,$8B,$FA,$2F,$03,$FD,$8D,$39,$AF,$E8,$93
	.dc.b	$97,$E4,$43,$E1,$78,$FB,$BF,$D6,$63,$13,$C2,$31,$C9,$FD,$4B,$EC
	.dc.b	$BD,$93,$E2,$8C,$6F,$FD,$86,$3E,$21,$FD,$4E,$5F,$3D,$9E,$3E,$4F
	.dc.b	$F6,$7F,$B2,$F8,$95,$FD,$67,$D1,$FE,$71,$B7,$FB,$7E,$73,$97,$F5
	.dc.b	$79,$CD,$7E,$07,$EC,$5E,$CF,$F7,$1D,$1F,$4D,$FD,$CE,$73,$5F,$D1
	.dc.b	$E7,$2F,$F4,$98,$7C,$5E,$FE,$9F,$83,$67,$FA,$77,$D9,$7F,$A9,$DE
	.dc.b	$11,$FD,$FE,$31,$9F,$FB,$B6,$F9,$D7,$0F,$37,$FC,$09,$F6,$4F,$3F
	.dc.b	$84,$61,$FF,$04,$7C,$4B,$DC,$CD,$FF,$5D,$8C,$3C,$CC,$47,$03,$AF
	.dc.b	$9F,$63,$FB,$27,$C2,$FE,$07,$3F,$EF,$DF,$64,$F9,$4F,$F0,$B4,$E7
	.dc.b	$9B,$FC,$4F,$39,$BE,$3D,$DF,$E1,$08,$C6,$0A,$57,$EC,$6F,$93,$FE
	.dc.b	$35,$AD,$CE,$AB,$FC,$93,$5A,$6D,$F5,$F8,$0F,$1B,$E4,$E3,$FF,$0F
	.dc.b	$FD,$97,$F4,$FA,$E3,$18,$F7,$35,$9F,$F8,$7E,$B5,$4C,$57,$BA,$F8
	.dc.b	$7E,$7F,$F8,$67,$EA,$7E,$0B,$63,$24,$BF,$73,$ED,$76,$FF,$83,$F4
	.dc.b	$57,$F8,$FD,$FD,$A4,$F0,$3F,$CA,$D1,$9F,$17,$4C,$AF,$F0,$C4,$FF
	.dc.b	$9B,$3C,$9F,$F3,$6F,$D8,$BE,$63,$F8,$E6,$3D,$AF,$76,$BD,$D6,$B0
	.dc.b	$F9,$D8,$71,$AF,$F3,$DD,$69,$F1,$DD,$7F,$94,$45,$C3,$B4,$2B,$93
	.dc.b	$A3,$18,$77,$BE,$51,$E9,$32,$76,$29,$C6,$79,$D7,$67,$AC,$71,$73
	.dc.b	$87,$AA,$FF,$D2,$18,$79,$63,$15,$EC,$3A,$DF,$3A,$DC,$62,$76,$FA
	.dc.b	$4F,$F4,$AD,$7F,$A5,$9F,$41,$83,$FD,$3D,$F6,$5F,$CE,$FE,$4F,$FD
	.dc.b	$36,$D8,$CF,$88,$9C,$B6,$38,$FC,$67,$E9,$5B,$97,$28,$C4,$7F,$AB
	.dc.b	$AD,$18,$8F,$EB,$51,$89,$FA,$D9,$36,$FE,$BD,$A7,$D4,$8C,$5F,$DB
	.dc.b	$BE,$0F,$D0,$6C,$EE,$B8,$D7,$F6,$73,$8D,$6F,$0F,$45,$7E,$02,$F8
	.dc.b	$D6,$EF,$8D,$FF,$B0,$A3,$8D,$6E,$F8,$C7,$A9,$B3,$BA,$E2,$FE,$75
	.dc.b	$B7,$DA,$3F,$D9,$AD,$74,$7C,$28,$DC,$F4,$8C,$9F,$AF,$2E,$1F,$C3
	.dc.b	$DB,$AF,$8E,$39,$47,$FA,$1F,$0F,$83,$1B,$6C,$EE,$7F,$33,$F6,$5F
	.dc.b	$BE,$7F,$D1,$CD,$7C,$0D,$AF,$03,$F7,$23,$7B,$7E,$EA,$BF,$C2,$D3
	.dc.b	$B6,$C6,$67,$FD,$DB,$3C,$5F,$67,$D9,$7F,$43,$7F,$D3,$CD,$73,$D5
	.dc.b	$C5,$7D,$87,$F5,$A7,$71,$98,$FD,$8E,$1F,$E2,$76,$CF,$FB,$F2,$BA
	.dc.b	$63,$F2,$5F,$F0,$C4,$78,$3C,$3E,$D1,$88,$FA,$14,$FB,$2C,$D4,$9E
	.dc.b	$07,$FA,$D6,$74,$E0,$E3,$CB,$E5,$38,$3D,$3A,$BD,$9F,$65,$FC,$EA
	.dc.b	$3F,$D7,$CD,$74,$62,$76,$47,$C1,$3F,$17,$D1,$7C,$8D,$24,$F1,$3F
	.dc.b	$D9,$50,$F2,$B8,$12,$3D,$5E,$57,$3F,$2A,$47,$07,$4F,$E4,$FE,$CB
	.dc.b	$FB,$F3,$FE,$D5,$6B,$9F,$1B,$EC,$47,$D0,$E2,$FF,$2D,$7D,$97,$F9
	.dc.b	$92,$BF,$DB,$8D,$7F,$3C,$DE,$3F,$CF,$73,$7F,$8B,$27,$FD,$C5,$F6
	.dc.b	$5F,$E6,$66,$FB,$DF,$85,$8C,$47,$E9,$E5,$F4,$D9,$EB,$93,$C9,$FC
	.dc.b	$64,$F9,$60,$2B,$FC,$C4,$F8,$3D,$AE,$C7,$E4,$FE,$C9,$F3,$3F,$95
	.dc.b	$8F,$C4,$67,$FB,$56,$EB,$FE,$79,$6D,$F6,$BF,$F3,$63,$FE,$FD,$8F
	.dc.b	$73,$FB,$F6,$D7,$3F,$53,$C1,$F2,$FF,$A3,$FB,$2F,$F3,$25,$FA,$2B
	.dc.b	$F8,$24,$62,$3F,$84,$36,$3F,$84,$7D,$97,$FA,$2E,$73,$ED,$F3,$23
	.dc.b	$CF,$3F,$F4,$84,$FE,$72,$3F,$89,$7D,$97,$FA,$CD,$D1,$F2,$1C,$3A
	.dc.b	$F8,$BF,$DC,$4F,$F9,$62,$36,$27,$E0,$BD,$3C,$9F,$DF,$46,$31,$F3
	.dc.b	$FD,$77,$F0,$1F,$B0,$FF,$46,$5F,$26,$BF,$C9,$5B,$3F,$C9,$7E,$C9
	.dc.b	$F6,$5C,$49,$5C,$5F,$E1,$91,$F5,$21,$F4,$3F,$EA,$97,$CE,$D7,$F9
	.dc.b	$93,$6B,$F9,$97,$D8,$5D,$BF,$D6,$D3,$6B,$F1,$6D,$7A,$FC,$5F,$F1
	.dc.b	$BA,$C4,$E5,$FD,$CC,$FC,$66,$BE,$09,$FE,$45,$F6,$4F,$95,$A4,$9F
	.dc.b	$9B,$1F,$C8,$13,$B7,$31,$B7,$FA,$64,$6C,$FF,$34,$3C,$BE,$AF,$B9
	.dc.b	$E0,$F7,$24,$64,$87,$27,$F9,$B6,$2F,$B0,$A4,$9F,$2F,$05,$7E,$5C
	.dc.b	$5D,$7F,$AE,$62,$AC,$6B,$FE,$E1,$FA,$9F,$EC,$15,$FD,$85,$B1,$F4
	.dc.b	$2F,$FA,$34,$3E,$CA,$93,$1F,$C8,$C3,$7C,$47,$F6,$D8,$E0,$EA,$79
	.dc.b	$CF,$4F,$9B,$EB,$F1,$92,$B9,$C6,$5E,$B1,$97,$A4,$7F,$59,$FB,$2F
	.dc.b	$E2,$71,$FA,$26,$BF,$DF,$5B,$9F,$F7,$DF,$B2,$FE,$1B,$3B,$F1,$79
	.dc.b	$11,$FD,$5C,$69,$E5,$1F,$D9,$E3,$E9,$46,$B4,$E0,$FF,$82,$E1,$F0
	.dc.b	$DB,$F9,$C9,$3D,$23,$91,$87,$84,$61,$F0,$F8,$1D,$C8,$F9,$3D,$0A
	.dc.b	$F7,$8B,$97,$B6,$7F,$BE,$7D,$8B,$EC,$A9,$2F,$F8,$3F,$D9,$7F,$7F
	.dc.b	$7F,$C4,$53,$B7,$F8,$5B,$5C,$CF,$F8,$1E,$78,$5F,$F8,$9E,$B4,$FF
	.dc.b	$8A,$63,$F4,$C7,$E9,$E3,$3C,$9F,$D0,$AD,$3F,$A4,$7F,$C6,$4D,$AE
	.dc.b	$5D,$1D,$FA,$7C,$0C,$AF,$4B,$97,$D7,$F6,$3E,$52,$3F,$C9,$3E,$CB
	.dc.b	$F3,$6F,$F5,$2C,$13,$A8,$F0,$BF,$E2,$39,$DC,$3E,$CB,$E3,$3F,$CF
	.dc.b	$5A,$FA,$DA,$FF,$5E,$6B,$C7,$E4,$7F,$C6,$DF,$0A,$EB,$FE,$34,$DF
	.dc.b	$03,$FC,$65,$7F,$33,$6E,$9E,$46,$09,$8F,$F4,$4B,$7C,$AB,$AF,$33
	.dc.b	$FD,$44,$F4,$57,$FA,$5F,$03,$4F,$F4,$D7,$D8,$BE,$96,$92,$B9,$08
	.dc.b	$D3,$86,$BF,$ED,$7C,$7D,$9F,$65,$EE,$D4,$FF,$93,$6B,$7C,$9F,$E6
	.dc.b	$D5,$F4,$F9,$26,$1E,$DD,$7E,$8C,$EA,$3F,$AE,$B5,$FF,$3C,$36,$9B
	.dc.b	$7F,$83,$B2,$F8,$AF,$4A,$DE,$16,$B7,$1F,$EB,$3C,$3B,$CE,$6B,$F2
	.dc.b	$D8,$75,$7A,$4E,$AB,$C7,$5B,$79,$4D,$7E,$44,$67,$0F,$1C,$70,$21
	.dc.b	$E4,$3E,$04,$F1,$BF,$04,$7F,$8A,$1B,$3C,$A3,$81,$30,$F9,$97,$F8
	.dc.b	$EC,$3C,$D3,$9C,$3D,$73,$97,$B3,$E5,$9D,$D7,$2A,$F2,$BF,$1A,$FC
	.dc.b	$0E,$63,$A1,$E5,$AE,$8F,$0B,$EE,$D3,$E2,$94,$F8,$A6,$6B,$AC,$70
	.dc.b	$8E,$A6,$8F,$C7,$5F,$9A,$35,$7E,$33,$CD,$19,$9F,$21,$D5,$F8,$6F
	.dc.b	$F8,$D7,$0F,$C5,$37,$0F,$D1,$91,$FC,$F3,$5F,$D6,$47,$07,$C3,$7F
	.dc.b	$86,$EA,$FC,$F1,$F3,$23,$F2,$EE,$1F,$81,$3F,$71,$FD,$64,$65,$F7
	.dc.b	$C6,$44,$67,$0F,$75,$78,$2B,$F0,$93,$98,$C4,$7C,$79,$FD,$74,$E6
	.dc.b	$3E,$19,$D9,$C3,$B9,$E3,$5E,$08,$F4,$9E,$8B,$DC,$66,$3C,$F1,$8A
	.dc.b	$F8,$11,$FB,$47,$71,$F9,$B6,$C7,$C2,$6B,$E3,$3B,$9D,$0B,$7E,$A7
	.dc.b	$C4,$62,$FE,$83,$83,$C0,$64,$F7,$C7,$C8,$7E,$39,$93,$E0,$1F,$E9
	.dc.b	$4F,$B1,$FF,$49,$BF,$CA,$5A,$FF,$9E,$DA,$FF,$59,$6B,$FE,$9C,$FB
	.dc.b	$27,$CD,$BE,$5B,$F9,$4C,$61,$BE,$4B,$E8,$F3,$5B,$BF,$E7,$1F,$65
	.dc.b	$F8,$0C,$FF,$98,$B6,$33,$A7,$F1,$87,$FD,$1D,$19,$6C,$7F,$1C,$D3
	.dc.b	$53,$C5,$FD,$D7,$C9,$BC,$8D,$61,$FD,$CE,$B4,$DF,$13,$38,$B8,$DF
	.dc.b	$5A,$F8,$5C,$CB,$0F,$F0,$FE,$1F,$D5,$6B,$4F,$FA,$52,$33,$3E,$EC
	.dc.b	$3D,$4F,$79,$F7,$C6,$21,$9B,$EB,$9E,$3F,$92,$30,$D8,$FF,$7A,$DF
	.dc.b	$FA,$A2,$72,$D8,$E2,$D9,$E1,$E3,$F1,$7A,$7D,$63,$EC,$A9,$DF,$E1
	.dc.b	$FE,$E5,$F5,$9D,$36,$76,$F8,$67,$79,$7A,$A3,$FD,$95,$5C,$23,$07
	.dc.b	$FB,$93,$EC,$9F,$5B,$ED,$FB,$27,$C6,$7C,$E9,$CE,$1E,$D9,$CB,$FE
	.dc.b	$C0,$8C,$DF,$4D,$95,$A8,$EE,$27,$73,$F8,$79,$D3,$F8,$D8,$CD,$7F
	.dc.b	$8D,$DB,$5F,$53,$67,$8D,$E5,$CB,$E4,$FD,$34,$C3,$E3,$7C,$A5,$C3
	.dc.b	$C9,$AA,$B1,$9F,$E5,$E1,$58,$BF,$F7,$86,$4F,$B2,$FB,$DF,$1F,$F7
	.dc.b	$DB,$3F,$75,$B9,$7C,$1A,$E1,$3F,$81,$AF,$CB,$FD,$93,$E7,$FF,$15
	.dc.b	$D1,$FE,$21,$AF,$F7,$4C,$E5,$FC,$54,$67,$0F,$93,$39,$7E,$74,$67
	.dc.b	$4F,$DA,$4E,$5F,$D9,$46,$6B,$FC,$C1,$39,$6D,$FE,$43,$EC,$BC,$AF
	.dc.b	$E6,$FC,$5F,$CE,$79,$56,$BF,$34,$DC,$BD,$CD,$CB,$F8,$06,$9D,$E7
	.dc.b	$F5,$3B,$FF,$91,$6F,$F3,$1E,$6F,$E7,$BB,$FF,$3E,$C8,$C9,$F9,$D8
	.dc.b	$C5,$7C,$0F,$7B,$F2,$9F,$5F,$F2,$9B,$5F,$21,$B3,$B8,$FC,$B4,$ED
	.dc.b	$D1,$FB,$3F,$86,$C4,$FE,$13,$E4,$BF,$63,$F2,$FF,$B2,$6D,$7E,$BD
	.dc.b	$B3,$B8,$FD,$94,$FF,$01,$F0,$64,$7E,$FF,$C6,$B5,$F8,$39,$BB,$FC
	.dc.b	$69,$B5,$FE,$93,$FB,$26,$5F,$C0,$BE,$97,$F8,$EC,$62,$7F,$80,$46
	.dc.b	$27,$50,$C7,$F5,$B5,$B7,$F3,$F5,$B9,$D6,$1F,$CE,$6B,$4F,$E8,$7B
	.dc.b	$35,$7F,$B6,$FB,$17,$D9,$7F,$1F,$EF,$7B,$9F,$CC,$27,$39,$7F,$1F
	.dc.b	$9C,$BD,$8F,$F5,$77,$9A,$E2,$E5,$F4,$1F,$CE,$23,$13,$C6,$3F,$9D
	.dc.b	$FD,$83,$E5,$FF,$97,$7D,$97,$F4,$4D,$7D,$D3,$9B,$FE,$5D,$39,$94
	.dc.b	$FF,$04,$9F,$E6,$5F,$65,$F1,$5F,$24,$F2,$CF,$F9,$D4,$E6,$3F,$8A
	.dc.b	$D3,$E5,$3F,$9F,$7D,$93,$E4,$BC,$B1,$87,$FA,$8F,$D3,$BA,$BF,$A2
	.dc.b	$7D,$97,$F0,$DD,$3B,$E4,$AD,$47,$F4,$59,$D4,$FF,$24,$9F,$EA,$EF
	.dc.b	$F2,$9F,$B2,$7D,$57,$6F,$5B,$FA,$84,$63,$0E,$BE,$FB,$EB,$3F,$AA
	.dc.b	$78,$7F,$DB,$67,$31,$88,$FE,$CF,$3F,$EB,$29,$FE,$AD,$1F,$E0,$87
	.dc.b	$9F,$C1,$F2,$F7,$37,$9F,$F5,$FF,$B2,$F4,$BF,$CE,$B0,$F9,$1D,$FF
	.dc.b	$DF,$E7,$2F,$F8,$17,$0F,$90,$7F,$63,$87,$CD,$FF,$78,$8C,$DF,$F4
	.dc.b	$79,$CC,$7C,$8F,$B2,$F8,$B7,$FE,$0B,$F4,$BF,$A2,$CF,$FB,$E6,$72
	.dc.b	$FE,$72,$33,$7F,$A3,$FB,$17,$B5,$FE,$06,$E5,$7D,$2F,$F8,$22,$73
	.dc.b	$87,$D6,$9C,$BF,$62,$1F,$15,$BF,$A6,$E0,$D9,$FB,$BF,$65,$FE,$CA
	.dc.b	$78,$47,$F8,$52,$31,$97,$F6,$0B,$E7,$5C,$3C,$DF,$F0,$CF,$D9,$3C
	.dc.b	$FE,$11,$87,$FC,$3B,$F0,$FF,$D5,$2B,$81,$F8,$C7,$18,$7F,$6C,$C7
	.dc.b	$FC,$3B,$C3,$FE,$35,$9C,$B7,$DC,$EF,$E7,$F4,$7F,$C5,$3F,$0D,$F8
	.dc.b	$9C,$7F,$C3,$1F,$64,$F9,$2C,$F6,$FF,$87,$E3,$13,$F9,$CA,$E1,$A7
	.dc.b	$F7,$F9,$CB,$FD,$FA,$33,$97,$F9,$0E,$72,$FF,$92,$63,$35,$FC,$66
	.dc.b	$91,$87,$FC,$63,$5F,$71,$CD,$FA,$38,$BF,$C6,$5F,$65,$FD,$6E,$7F
	.dc.b	$91,$5A,$73,$EE,$6F,$1F,$F1,$A5,$EE,$FB,$D6,$A7,$D9,$E6,$FF,$8B
	.dc.b	$FE,$9F,$E2,$B6,$32,$4A,$F7,$1A,$7F,$7F,$A2,$BF,$CA,$AF,$ED,$E7
	.dc.b	$FC,$42,$E6,$77,$E2,$6A,$A4,$7F,$08,$AF,$E7,$9E,$2F,$F3,$DF,$B2
	.dc.b	$7D,$8F,$C7,$C7,$95,$7F,$8B,$0D,$89,$DB,$FD,$03,$06,$1C,$EF,$E8
	.dc.b	$4F,$0A,$DB,$FE,$60,$AD,$BC,$E7,$6F,$F1,$B3,$07,$5A,$E8,$F4,$C2
	.dc.b	$BC,$E2,$BD,$E7,$D5,$5D,$AF,$C3,$39,$3A,$1D,$8F,$C3,$5F,$4C,$3E
	.dc.b	$0C,$F0,$9F,$23,$EA,$8C,$DF,$A2,$B8,$5F,$78,$C5,$F5,$AD,$BE,$93
	.dc.b	$DE,$7C,$C2,$D6,$1C,$1F,$99,$FB,$2F,$E8,$DF,$1B,$C1,$B1,$F2,$1A
	.dc.b	$FD,$06,$BC,$7E,$AF,$FD,$1E,$ED,$AE,$DA,$EE,$FB,$47,$0C,$BE,$CE
	.dc.b	$2D,$7F,$59,$18,$6C,$71,$BF,$99,$82,$F9,$4F,$EC,$CE,$53,$F6,$9A
	.dc.b	$F0,$8D,$BF,$E1,$19,$E5,$86,$EF,$BC,$FF,$85,$6F,$8E,$1F,$8E,$BD
	.dc.b	$CF,$3B,$E0,$C3,$04,$9F,$53,$97,$96,$1E,$28,$FD,$2E,$1F,$A5,$D3
	.dc.b	$D4,$D8,$F1,$35,$61,$E2,$BE,$B1,$FE,$96,$C7,$DF,$8F,$FB,$7D,$B8
	.dc.b	$7C,$9A,$FD,$3F,$D9,$7E,$FE,$FF,$BF,$B5,$FF,$4F,$B6,$FF,$DD,$0F
	.dc.b	$CF,$8D,$EB,$E3,$8D,$5F,$F2,$36,$CF,$F8,$F5,$E0,$FF,$AA,$BE,$CB
	.dc.b	$FD,$2E,$FF,$83,$9A,$E7,$9F,$DD,$F6,$3F,$D6,$A3,$51,$8C,$E4,$60
	.dc.b	$E5,$E0,$FC,$5B,$FF,$0E,$46,$0F,$F0,$DD,$BE,$CB,$55,$27,$47,$FA
	.dc.b	$FA,$74,$E3,$4F,$E1,$6A,$3F,$9B,$3D,$39,$39,$3C,$7E,$3B,$EB,$7D
	.dc.b	$96,$A3,$FD,$92,$DB,$E7,$7F,$ED,$69,$FF,$17,$CE,$76,$FE,$98,$D8
	.dc.b	$FF,$33,$35,$F1,$5F,$E0,$FD,$07,$C9,$52,$4F,$4C,$B5,$E4,$C8,$FF
	.dc.b	$1E,$D2,$3E,$7F,$93,$CF,$C9,$95,$A7,$83,$FE,$E2,$FB,$2F,$F0,$03
	.dc.b	$FE,$46,$6B,$9F,$1B,$F2,$D1,$F7,$37,$F9,$3F,$65,$FE,$6B,$AF,$F2
	.dc.b	$6B,$5F,$F7,$6B,$63,$E7,$F3,$FF,$8C,$27,$FC,$A7,$F6,$5F,$E6,$D6
	.dc.b	$FB,$BF,$A3,$8C,$47,$FB,$DA,$5F,$4B,$9E,$E9,$3C,$9F,$D5,$4F,$97
	.dc.b	$93,$FB,$D5,$FF,$9A,$6B,$BF,$B3,$D4,$FD,$87,$D8,$BE,$CB,$F6,$51
	.dc.b	$F8,$E9,$69,$F8,$68,$C3,$7A,$78,$FB,$7C,$3D,$59,$7F,$E6,$EB,$F7
	.dc.b	$3C,$B3,$FF,$06,$64,$F9,$8F,$E0,$1F,$65,$FE,$67,$AF,$F3,$AB,$5E
	.dc.b	$07,$FA,$AD,$C6,$7F,$98,$FB,$2F,$EC,$EF,$E6,$A1,$ED,$F4,$25,$7F
	.dc.b	$A0,$2B,$F4,$2F,$41,$F6,$5F,$EB,$E7,$FD,$0A,$D7,$4D,$BF,$07,$4F
	.dc.b	$F1,$78,$FF,$49,$4F,$0D,$B0,$7F,$1B,$B7,$D7,$7F,$1D,$FB,$27,$CD
	.dc.b	$52,$5F,$FA,$53,$EC,$5F,$65,$B2,$47,$FA,$62,$B2,$EF,$93,$F9,$47
	.dc.b	$D0,$BE,$26,$92,$7F,$D3,$FF,$61,$97,$FD,$43,$35,$F8,$4D,$E0,$FE
	.dc.b	$06,$DE,$5F,$F2,$15,$5A,$FF,$6E,$4D,$51,$FB,$67,$BF,$8B,$E1,$7E
	.dc.b	$77,$D9,$3E,$5B,$F9,$F3,$63,$FD,$66,$D7,$7E,$6E,$A6,$46,$9E,$31
	.dc.b	$CB,$DB,$E7,$F2,$32,$B9,$C6,$5E,$B1,$97,$C7,$1F,$23,$EC,$24,$7C
	.dc.b	$47,$CB,$C9,$88,$F8,$B1,$76,$FE,$4F,$97,$DC,$9E,$07,$FB,$33,$EA
	.dc.b	$7F,$95,$CF,$F2,$C6,$EF,$FE,$6B,$DD,$87,$F5,$F9,$7D,$95,$26,$1C
	.dc.b	$F8,$2B,$F9,$79,$DB,$DE,$79,$4F,$3F,$9A,$E9,$F1,$92,$BC,$91,$97
	.dc.b	$CD,$19,$7A,$47,$F7,$1F,$B2,$FE,$1B,$1F,$EE,$66,$BF,$A7,$6E,$9F
	.dc.b	$D2,$3E,$CB,$F8,$44,$EF,$C9,$E4,$47,$FD,$DF,$1A,$79,$47,$F7,$F8
	.dc.b	$F8,$71,$AD,$38,$3F,$B5,$87,$C3,$EF,$E7,$24,$F4,$8E,$46,$1E,$11
	.dc.b	$87,$C3,$E2,$77,$23,$E4,$E8,$AF,$78,$B9,$7B,$67,$FC,$29,$F6,$2F
	.dc.b	$B2,$A4,$B1,$FC,$2F,$EC,$BF,$BE,$3F,$C3,$67,$83,$AB,$FE,$DF,$18
	.dc.b	$8D,$57,$F8,$86,$B8,$D6,$A3,$1D,$52,$B9,$73,$7F,$AD,$70,$E5,$7F
	.dc.b	$DE,$B0,$FF,$63,$56,$8F,$2B,$87,$CD,$D3,$DF,$AB,$C0,$C8,$E0,$27
	.dc.b	$85,$75,$AF,$37,$D9,$4A,$DD,$23,$FC,$75,$F6,$5F,$A0,$9F,$0C,$6D
	.dc.b	$F8,$CE,$B2,$FF,$20,$7D,$97,$F0,$19,$FF,$50,$B5,$FF,$69,$35,$F9
	.dc.b	$ED,$4F,$F0,$9C,$7F,$D4,$BF,$08,$EB,$FE,$9D,$5F,$0E,$31,$C3,$E0
	.dc.b	$EF,$FF,$28,$7D,$9B,$EE,$69,$2F,$E5,$62,$F2,$DF,$69,$99,$FF,$58
	.dc.b	$DF,$17,$FD,$2A,$D7,$37,$FE,$95,$D3,$FA,$AC,$7F,$A6,$9A,$FF,$AE
	.dc.b	$9A,$FC,$2F,$37,$81,$F1,$BE,$7E,$47,$D6,$9E,$39,$7F,$8E,$BE,$D3
	.dc.b	$F9,$94,$FF,$5D,$6C,$67,$A7,$E7,$CF,$F9,$BE,$73,$C3,$FE,$FF,$6C
	.dc.b	$7E,$71,$AF,$CE,$38,$4C,$7F,$0F,$1A,$BF,$EE,$75,$CE,$B5,$5E,$37
	.dc.b	$C2,$F0,$73,$7E,$58,$F6,$18,$9E,$4E,$CC,$94,$E8,$58,$FF,$2D,$D7
	.dc.b	$8A,$3D,$55,$CD,$A7,$23,$89,$27,$C4,$FF,$81,$67,$92,$BF,$86,$4A
	.dc.b	$F5,$C7,$1B,$F3,$B4,$FC,$B9,$E1,$39,$45,$AE,$B5,$CE,$7D,$55,$FA
	.dc.b	$07,$84,$65,$F2,$18,$AE,$4D,$38,$9E,$02,$9E,$02,$D7,$82,$38,$47
	.dc.b	$43,$E6,$1F,$C7,$0F,$8E,$52,$96,$7F,$1E,$D8,$FF,$79,$46,$23,$E1
	.dc.b	$99,$3F,$DF,$2F,$E3,$B8,$3F,$6D,$F5,$5F,$EC,$36,$CF,$A2,$75,$C4
	.dc.b	$AD,$36,$75,$78,$99,$18,$3F,$80,$4E,$DF,$F2,$5D,$F6,$7E,$B1,$F1
	.dc.b	$4C,$C6,$F2,$FE,$16,$DC,$BE,$EC,$7E,$18,$CD,$F8,$2B,$9E,$3E,$88
	.dc.b	$C4,$EE,$B8,$9F,$D2,$30,$F1,$47,$2B,$F4,$1E,$2A,$E3,$5C,$23,$E9
	.dc.b	$CE,$DD,$1C,$1C,$12,$38,$47,$B6,$30,$AB,$C3,$3A,$BF,$44,$FC,$8B
	.dc.b	$FC,$2D,$F0,$BD,$CF,$8A,$FC,$91,$83,$31,$DE,$33,$87,$C9,$8F,$03
	.dc.b	$5D,$1C,$CC,$DF,$C0,$70,$6A,$BB,$4F,$1C,$3F,$67,$C5,$FC,$6C,$95
	.dc.b	$FC,$FF,$E6,$9F,$4B,$88,$FC,$D1,$FB,$B9,$FC,$B9,$1C,$8F,$B3,$7C
	.dc.b	$EB,$E5,$F3,$18,$6F,$94,$E2,$CD,$6E,$FB,$FD,$BB,$EE,$3F,$DD,$F1
	.dc.b	$FA,$D8,$C7,$0F,$81,$B1,$C2,$76,$FF,$83,$A3,$FC,$DF,$3B,$F2,$F3
	.dc.b	$29,$CB,$E0,$8C,$FD,$AF,$FB,$8E,$BD,$AD,$9F,$86,$D7,$F7,$CD,$7F
	.dc.b	$C3,$1F,$5B,$FE,$B7,$9F,$EC,$6D,$8F,$F5,$AF,$2A,$3F,$DE,$F3,$AD
	.dc.b	$CD,$25,$70,$9D,$B6,$76,$F4,$FA,$D7,$DA,$E3,$09,$18,$C3,$FB,$AF
	.dc.b	$77,$D5,$7F,$BB,$C6,$6F,$95,$F1,$9F,$F1,$94,$E6,$FA,$5F,$1B,$DD
	.dc.b	$7F,$88,$BE,$D3,$F1,$2F,$F8,$0B,$ED,$5F,$1D,$CE,$BF,$C8,$35,$A7
	.dc.b	$E1,$CE,$AB,$97,$36,$F1,$4E,$6B,$FD,$A5,$7C,$67,$DD,$39,$7E,$1C
	.dc.b	$A7,$57,$FC,$D5,$B1,$FE,$F6,$D7,$FD,$85,$3E,$03,$E1,$FB,$7E,$1D
	.dc.b	$3F,$DA,$31,$F1,$BE,$4F,$13,$23,$66,$4B,$7F,$5F,$ED,$7D,$58,$7E
	.dc.b	$27,$2F,$A8,$A3,$3F,$1D,$A6,$DF,$E2,$1B,$7F,$EF,$1F,$B4,$7C,$EF
	.dc.b	$C5,$FB,$57,$DA,$7C,$DA,$56,$B0,$FF,$19,$D6,$9F,$F2,$BC,$66,$7F
	.dc.b	$27,$39,$7B,$5F,$5B,$E9,$5B,$CF,$E6,$B7,$0F,$C2,$FD,$AE,$AF,$DF
	.dc.b	$E2,$7C,$0C,$2D,$7E,$89,$B8,$FE,$6D,$B8,$FC,$47,$E8,$7B,$52,$BF
	.dc.b	$2B,$ED,$DB,$FC,$F4,$DB,$FC,$EC,$DA,$FE,$E5,$9F,$FB,$A5,$FC,$97
	.dc.b	$D8,$FA,$BC,$EF,$C6,$73,$23,$13,$A3,$E2,$CE,$A3,$39,$7D,$18,$FB
	.dc.b	$7F,$6B,$8C,$BF,$5B,$F2,$9F,$87,$F8,$3F,$C4,$37,$1F,$D6,$B7,$0F
	.dc.b	$DB,$36,$BF,$07,$ED,$FE,$18,$FC,$3F,$B7,$6F,$F8,$4C,$5B,$FE,$05
	.dc.b	$36,$BF,$3F,$F6,$AC,$BF,$31,$F4,$FF,$83,$8C,$63,$FB,$E8,$C4,$EA
	.dc.b	$3F,$D3,$33,$AC,$3F,$85,$56,$A7,$35,$FA,$3E,$EC,$53,$0F,$D7,$FD
	.dc.b	$A3,$ED,$7F,$71,$F0,$EF,$73,$F2,$93,$9C,$BF,$D0,$B3,$97,$FD,$53
	.dc.b	$E6,$7D,$37,$3F,$47,$F9,$B8,$C6,$5F,$CA,$29,$3F,$9F,$FB,$37,$CC
	.dc.b	$7F,$07,$FB,$57,$16,$63,$0E,$EF,$F8,$67,$DA,$FF,$B4,$B3,$FF,$5C
	.dc.b	$E1,$FC,$0B,$2F,$E6,$B3,$97,$DD,$3F,$E3,$D9,$7C,$AF,$F1,$5F,$B5
	.dc.b	$7C,$9F,$8E,$30,$FE,$CF,$EA,$1D,$5F,$C6,$FE,$D7,$F8,$5E,$5D,$0F
	.dc.b	$F7,$1C,$EB,$2F,$F5,$4C,$EB,$97,$F8,$DC,$BE,$87,$F9,$27,$DA,$BE
	.dc.b	$AF,$B4,$61,$FE,$05,$EE,$BE,$BB,$F9,$57,$8A,$32,$E1,$DF,$C2,$6B
	.dc.b	$DC,$BA,$FF,$AE,$A7,$FB,$DB,$FE,$DE,$BF,$89,$A7,$F4,$DF,$B5,$F4
	.dc.b	$6F,$EB,$13,$AA,$FF,$7D,$4E,$AB,$F9,$AC,$7C,$83,$FD,$8F,$0F,$9E
	.dc.b	$FE,$35,$19,$BF,$EC,$F3,$98,$FF,$77,$FD,$AF,$C5,$53,$F9,$3F,$4B
	.dc.b	$FA,$14,$7F,$9C,$E7,$53,$9F,$63,$F9,$37,$DA,$7F,$B5,$6B,$FD,$F3
	.dc.b	$CA,$FA,$7C,$FB,$7F,$CA,$E3,$13,$FE,$FE,$A7,$DA,$F2,$FA,$9E,$6D
	.dc.b	$9F,$1E,$FF,$D4,$BE,$D3,$FC,$E5,$1C,$67,$F9,$B4,$63,$4F,$91,$87
	.dc.b	$4A,$E2,$7F,$A2,$E5,$E6,$FF,$3D,$FB,$57,$9F,$CA,$30,$FF,$41,$F8
	.dc.b	$77,$BD,$9C,$39,$E5,$F1,$A7,$31,$88,$63,$DA,$31,$3A,$9F,$B8,$DD
	.dc.b	$3F,$B6,$E4,$F4,$3C,$18,$FF,$51,$FB,$4F,$C8,$63,$E0,$B7,$DA,$BE
	.dc.b	$5B,$5D,$DF,$DF,$63,$15,$E3,$BE,$3A,$7F,$89,$A7,$2F,$E3,$63,$38
	.dc.b	$7F,$87,$67,$2F,$F8,$02,$33,$5F,$EB,$36,$D7,$F1,$7A,$FC,$2B,$F8
	.dc.b	$88,$CA,$5C,$5F,$E1,$1F,$B5,$FE,$BD,$3B,$8C,$7B,$BE,$F6,$E1,$E9
	.dc.b	$6C,$F9,$FC,$FF,$F0,$D7,$3B,$EA,$FE,$34,$61,$B7,$F0,$63,$E5,$56
	.dc.b	$0F,$D6,$69,$C6,$7E,$C3,$FC,$05,$B5,$FE,$01,$8C,$4F,$08,$C1,$FE
	.dc.b	$05,$7F,$87,$8F,$B4,$7C,$DF,$FB,$66,$BB,$4F,$EE,$CD,$CE,$9F,$CE
	.dc.b	$D7,$F9,$42,$31,$1F,$EE,$1F,$23,$F6,$71,$8A,$E6,$EA,$79,$DF,$7B
	.dc.b	$E4,$FF,$A3,$AF,$8B,$9C,$2D,$7F,$BA,$E6,$BF,$D4,$6B,$8B,$93,$D9
	.dc.b	$5C,$5A,$66,$FC,$93,$C2,$3D,$06,$8E,$97,$D9,$F3,$9F,$01,$C2,$BF
	.dc.b	$1E,$1E,$EC,$3D,$53,$FE,$A4,$34,$59,$C3,$83,$F8,$C9,$FC,$6B,$ED
	.dc.b	$7F,$A5,$FC,$9F,$FA,$B1,$B1,$FE,$B2,$6B,$C0,$38,$79,$AF,$AE,$FE
	.dc.b	$ED,$1C,$5A,$F1,$8D,$61,$DE,$73,$3C,$9E,$39,$FD,$EC,$7F,$B8,$CF
	.dc.b	$D2,$6B,$F2,$6F,$9B,$6B,$8C,$FB,$8D,$D7,$3D,$3E,$0B,$F8,$AB,$E2
	.dc.b	$DA,$E3,$7F,$CF,$AF,$89,$CE,$F8,$BE,$7C,$3F,$0B,$87,$28,$D5,$79
	.dc.b	$EB,$6F,$CF,$BE,$AF,$CA,$6C,$F6,$8F,$04,$F0,$75,$7C,$DF,$E5,$53
	.dc.b	$F0,$A3,$FD,$22,$DC,$FC,$93,$F8,$D8,$C3,$D0,$E3,$5F,$B1,$9D,$E3
	.dc.b	$FA,$CF,$B5,$FE,$11,$5F,$CD,$DA,$FF,$6F,$6E,$1F,$EA,$07,$F0,$11
	.dc.b	$BD,$7A,$3F,$D9,$A3,$77,$F9,$06,$C6,$72,$DA,$FB,$5F,$E9,$6F,$F4
	.dc.b	$76,$BF,$DF,$DB,$87,$FA,$C3,$ED,$3F,$B2,$CE,$FA,$39,$C7,$F0,$3C
	.dc.b	$7F,$3E,$D9,$FE,$F3,$7C,$A7,$F4,$6F,$F8,$A2,$3C,$1E,$0F,$92,$31
	.dc.b	$1F,$E0,$FA,$7D,$AE,$6A,$4F,$03,$FA,$F4,$E9,$C1,$C7,$93,$CA,$70
	.dc.b	$7A,$78,$9F,$0B,$ED,$7F,$A1,$C7,$F6,$76,$BA,$31,$3B,$23,$E0,$9F
	.dc.b	$97,$E7,$BE,$6A,$92,$78,$9F,$DC,$E1,$E4,$F0,$24,$7A,$BA,$79,$F9
	.dc.b	$12,$38,$3A,$7F,$C6,$5F,$6B,$FE,$06,$7F,$BF,$35,$CF,$8D,$F7,$A3
	.dc.b	$E8,$47,$F8,$AB,$ED,$7F,$D4,$8F,$F8,$15,$AF,$F8,$F5,$BC,$9F,$EE
	.dc.b	$EE,$7F,$F1,$C4,$FF,$82,$FE,$D7,$FC,$E6,$DF,$77,$F1,$71,$88,$FF
	.dc.b	$23,$4B,$EA,$73,$D3,$27,$93,$F9,$29,$F2,$C2,$BF,$CE,$2F,$7F,$67
	.dc.b	$A9,$F9,$8F,$B4,$7D,$AF,$E6,$A3,$F0,$F9,$BA,$FF,$97,$C6,$1B,$D3
	.dc.b	$C7,$DB,$E1,$EA,$4C,$BF,$C4,$D7,$EE,$7F,$CB,$D9,$3E,$63,$F4,$FF
	.dc.b	$6B,$FE,$6F,$9F,$F1,$63,$5F,$F3,$23,$73,$F1,$47,$F9,$63,$ED,$3F
	.dc.b	$9D,$46,$21,$2F,$6F,$A9,$23,$FC,$71,$5F,$9B,$7F,$CE,$1F,$6B,$FE
	.dc.b	$C1,$7F,$C7,$8D,$73,$F1,$7F,$BC,$97,$57,$EF,$AB,$FD,$33,$5C,$36
	.dc.b	$C1,$BF,$B2,$FE,$07,$F6,$6F,$9A,$A4,$AF,$F2,$57,$DA,$BE,$D7,$64
	.dc.b	$8F,$F2,$75,$65,$DF,$27,$F0,$EF,$A4,$7C,$55,$24,$FF,$95,$7E,$8B
	.dc.b	$2F,$F9,$5E,$6B,$FE,$8F,$6F,$1F,$C9,$6F,$47,$F6,$CA,$B5,$FE,$86
	.dc.b	$9A,$A3,$F6,$8F,$7F,$17,$C2,$FF,$A6,$FE,$D5,$F2,$BF,$C9,$1B,$1F
	.dc.b	$E6,$96,$BB,$F4,$B5,$32,$34,$F1,$8E,$5E,$DF,$3F,$92,$95,$CE,$32
	.dc.b	$F5,$8C,$BE,$38,$F9,$1F,$67,$23,$E2,$3E,$5E,$4C,$47,$C5,$8B,$B7
	.dc.b	$E3,$72,$56,$36,$FF,$08,$FD,$4F,$E4,$6B,$F2,$4D,$8F,$F0,$AE,$EC
	.dc.b	$3F,$A2,$CB,$ED,$69,$32,$FF,$46,$47,$CD,$6C,$65,$FF,$63,$47,$17
	.dc.b	$7C,$1D,$3E,$73,$AF,$C5,$C9,$E7,$19,$7A,$C6,$5E,$71,$FD,$63,$ED
	.dc.b	$7F,$86,$47,$FA,$75,$AF,$FB,$55,$BA,$7E,$BF,$ED,$7F,$83,$CE,$FC
	.dc.b	$AE,$44,$7F,$D4,$D1,$A7,$94,$7F,$66,$8F,$89,$1A,$D3,$83,$FE,$E3
	.dc.b	$8F,$F5,$CB,$CB,$E5,$78,$1C,$7E,$26,$46,$48,$71,$7B,$78,$B2,$3C
	.dc.b	$6F,$87,$A3,$11,$DE,$2E,$5E,$D9,$FE,$F7,$F6,$8F,$B5,$A4,$BF,$F6
	.dc.b	$17,$DA,$FF,$78,$7F,$DF,$53,$B7,$FD,$8C,$D7,$33,$FE,$06,$9E,$17
	.dc.b	$FC,$66,$B4,$FF,$1B,$8F,$EF,$D4,$6C,$EB,$97,$FC,$C3,$1F,$EC,$59
	.dc.b	$F1,$F4,$75,$3B,$74,$D8,$C9,$06,$1D,$E3,$C9,$F6,$92,$7D,$AF,$F8
	.dc.b	$73,$ED,$7F,$39,$03,$C9,$1B,$79,$E3,$E4,$8E,$0F,$F8,$93,$EC,$BF
	.dc.b	$7F,$87,$F3,$76,$C7,$F9,$E9,$B4,$F8,$57,$CA,$7F,$5A,$9F,$E0,$F3
	.dc.b	$FE,$79,$6B,$C7,$C0,$FF,$44,$46,$23,$CD,$F6,$8F,$B4,$FF,$1A,$B6
	.dc.b	$3F,$DF,$4D,$7D,$27,$C4,$F2,$DF,$6B,$C2,$A4,$F0,$7F,$C8,$BA,$7F
	.dc.b	$44,$9F,$77,$9F,$2B,$E3,$4F,$BE,$3F,$C9,$7A,$7B,$3B,$75,$5E,$01
	.dc.b	$9F,$F5,$97,$8E,$DF,$16,$F8,$EC,$BF,$F2,$7F,$DA,$FF,$32,$BE,$4D
	.dc.b	$CB,$FB,$23,$70,$F9,$D3,$C3,$9E,$3D,$8F,$B9,$3B,$7E,$F6,$5F,$DB
	.dc.b	$AB,$FD,$5F,$39,$93,$4C,$4E,$E3,$9D,$76,$7C,$AF,$E0,$DC,$5F,$9A
	.dc.b	$2D,$F9,$A7,$C8,$76,$8C,$BD,$0C,$D7,$77,$47,$0B,$F2,$4F,$78,$C3
	.dc.b	$C9,$D9,$F4,$67,$94,$F0,$8D,$9D,$8E,$0E,$5E,$06,$27,$A4,$F2,$71
	.dc.b	$1C,$4E,$67,$C2,$2D,$FE,$8D,$D9,$93,$A9,$AB,$FA,$33,$DA,$37,$3F
	.dc.b	$25,$EC,$71,$70,$D9,$F9,$8D,$3E,$51,$4F,$C8,$9D,$1E,$EF,$48,$E6
	.dc.b	$48,$CC,$79,$8E,$57,$F1,$67,$D9,$1E,$93,$BB,$C5,$FC,$69,$DC,$CC
	.dc.b	$F9,$A3,$51,$F5,$9B,$7F,$36,$3A,$19,$9E,$07,$92,$78,$D7,$E5,$27
	.dc.b	$B1,$B3,$C2,$72,$7D,$13,$CA,$3C,$64,$38,$1F,$18,$E1,$7F,$4D,$A7
	.dc.b	$92,$FC,$CF,$C4,$8F,$C9,$34,$D1,$E4,$AF,$33,$C8,$FA,$93,$CA,$FD
	.dc.b	$75,$C9,$B3,$C8,$E0,$FC,$67,$F8,$2F,$44,$7F,$48,$FF,$0A,$FA,$4C
	.dc.b	$47,$F0,$88,$B3,$FE,$6E,$D7,$F9,$8D,$7F,$B7,$3E,$CF,$F8,$85,$FF
	.dc.b	$41,$20,$72,$E0,$C6,$95,$FE,$91,$1A,$78,$FD,$AB,$E7,$1F,$2F,$B8
	.dc.b	$C3,$7C,$87,$D2,$66,$B7,$7D,$F0,$79,$DF,$E8,$38,$F3,$46,$38,$BC
	.dc.b	$4D,$8E,$13,$B7,$CB,$E4,$57,$D8,$D9,$FF,$30,$57,$28,$E1,$BB,$EC
	.dc.b	$3F,$D3,$93,$FC,$AE,$31,$C3,$FE,$98,$64,$FF,$B6,$E7,$4F,$0C,$3F
	.dc.b	$A8,$CF,$09,$CD,$71,$7F,$A5,$D7,$1A,$DB,$C6,$BB,$7D,$8B,$EC,$F8
	.dc.b	$57,$F7,$E8,$C5,$F9,$7A,$FE,$7B,$E7,$8C,$CF,$89,$B7,$D2,$F9,$BF
	.dc.b	$EA,$2E,$0B,$3A,$BF,$C2,$7D,$9B,$93,$7F,$5A,$F8,$BF,$1E,$6A,$D5
	.dc.b	$FC,$7A,$D3,$FD,$02,$33,$3F,$12,$72,$FF,$67,$8C,$E5,$F3,$67,$2F
	.dc.b	$FB,$0E,$33,$5E,$4C,$3D,$3A,$FF,$AA,$23,$B4,$7A,$BD,$89,$1F,$0F
	.dc.b	$E4,$6E,$5F,$0E,$6E,$3F,$3A,$6D,$7D,$4F,$B3,$EB,$F1,$9D,$9B,$1F
	.dc.b	$79,$B8,$7D,$86,$DF,$FB,$67,$EB,$DF,$33,$F3,$BE,$CD,$F6,$3F,$22
	.dc.b	$73,$8F,$E3,$67,$2F,$F6,$68,$CE,$3F,$8A,$9C,$BF,$E0,$D8,$CE,$1F
	.dc.b	$93,$9C,$BE,$1C,$3B,$1F,$73,$EC,$EE,$1F,$96,$FB,$3F,$93,$97,$CA
	.dc.b	$6C,$64,$FC,$DC,$65,$A7,$E3,$7C,$0F,$C3,$9F,$88,$F6,$2C,$64,$9A
	.dc.b	$7E,$46,$32,$FA,$7C,$BF,$C5,$FD,$9F,$C3,$6D,$7C,$16,$CE,$E3,$F1
	.dc.b	$93,$B7,$47,$EC,$3E,$1B,$13,$F9,$FF,$92,$FC,$DF,$83,$F9,$C6,$D7
	.dc.b	$E9,$5B,$3B,$8F,$D4,$4F,$E3,$3D,$69,$1F,$BC,$F5,$ED,$7E,$DA,$6F
	.dc.b	$0F,$F0,$29,$B5,$FC,$AF,$EC,$D9,$7E,$B3,$E9,$3F,$4F,$18,$9F,$DB
	.dc.b	$46,$27,$51,$FE,$26,$9D,$63,$FB,$9A,$D4,$E6,$BF,$6F,$D5,$8A,$FF
	.dc.b	$28,$7D,$93,$EC,$FF,$7B,$F0,$AF,$63,$F6,$F3,$9C,$7F,$8E,$4E,$5F
	.dc.b	$F1,$AF,$96,$E2,$E3,$F4,$1F,$BE,$8C,$4F,$08,$FE,$03,$F6,$2F,$96
	.dc.b	$FE,$17,$F6,$6E,$2C,$C6,$1D,$E5,$FC,$43,$EC,$FF,$D2,$18,$FF,$91
	.dc.b	$30,$FD,$D6,$FF,$CA,$27,$2F,$F9,$AE,$1F,$27,$FC,$63,$EC,$DF,$21
	.dc.b	$E2,$8C,$3F,$C5,$3E,$9B,$FA,$27,$47,$F3,$7F,$B3,$79,$39,$EC,$FE
	.dc.b	$43,$18,$D7,$FC,$AB,$6F,$9F,$FE,$4B,$F6,$6F,$B3,$E9,$18,$7F,$90
	.dc.b	$FB,$EF,$AD,$FE,$57,$E2,$0C,$B8,$77,$F0,$3A,$F6,$EE,$7F,$D3,$67
	.dc.b	$FB,$43,$FE,$66,$BF,$83,$BF,$F4,$FF,$B3,$F5,$CF,$F7,$79,$CB,$FE
	.dc.b	$94,$8C,$F2,$FC,$CE,$2F,$E7,$B0,$F9,$CF,$E6,$F5,$F2,$0F,$EB,$91
	.dc.b	$97,$FC,$EB,$F6,$7F,$0E,$FF,$AC,$FA,$5F,$CC,$A7,$FC,$0D,$39,$CB
	.dc.b	$FB,$14,$E6,$BF,$D9,$5F,$64,$F7,$3F,$A2,$F3,$3E,$97,$FA,$3C,$E7
	.dc.b	$2F,$F4,$1C,$E5,$FF,$42,$C3,$EC,$F7,$F4,$FC,$1B,$3F,$E8,$AF,$B3
	.dc.b	$7A,$79,$8C,$37,$3F,$F5,$94,$BC,$FF,$EA,$FF,$66,$F4,$7F,$AC,$CE
	.dc.b	$70,$FF,$4A,$CE,$5F,$F4,$BF,$C0,$3D,$BD,$E1,$FD,$82,$31,$3E,$38
	.dc.b	$C4,$C8,$E7,$E9,$5C,$BF,$D8,$9F,$65,$F8,$79,$FF,$06,$7D,$9B,$E4
	.dc.b	$BF,$B5,$CE,$6B,$FC,$13,$39,$AF,$F5,$4C,$E7,$9B,$FC,$25,$39,$AE
	.dc.b	$BD,$FF,$DF,$23,$12,$BF,$F1,$F5,$F1,$6D,$6F,$E0,$BC,$2F,$8B,$8F
	.dc.b	$FC,$0B,$F6,$7F,$D3,$AB,$8C,$63,$DD,$D4,$7F,$89,$67,$59,$86,$2B
	.dc.b	$E1,$5E,$AF,$47,$FC,$23,$D4,$FA,$9F,$83,$18,$6C,$EC,$F8,$38,$60
	.dc.b	$FF,$77,$6B,$C6,$3F,$D9,$AF,$ED,$DB,$3F,$E1,$68,$C4,$F0,$8C,$1F
	.dc.b	$E1,$A7,$D1,$E4,$7A,$7E,$C9,$F4,$3F,$E6,$5B,$ED,$5F,$EF,$D3,$62
	.dc.b	$76,$FC,$18,$DC,$62,$3F,$CD,$D5,$D6,$BA,$5F,$D5,$9E,$51,$FE,$6F
	.dc.b	$9D,$47,$39,$D5,$73,$8F,$0B,$D6,$FB,$DF,$19,$E6,$71,$C3,$9B,$E2
	.dc.b	$32,$27,$8B,$FE,$82,$C3,$71,$88,$E1,$7D,$F0,$F1,$E1,$DE,$31,$87
	.dc.b	$6A,$F2,$CF,$BA,$31,$0C,$3F,$D1,$4F,$CA,$3F,$D1,$8E,$1D,$1F,$E9
	.dc.b	$1F,$B3,$FE,$7F,$F1,$FF,$E9,$16,$C6,$7C,$54,$E5,$B1,$C7,$EB,$3D
	.dc.b	$CD,$D3,$94,$62,$3F,$D4,$36,$8C,$47,$FA,$96,$31,$3F,$2B,$26,$BF
	.dc.b	$EA,$9D,$3E,$F4,$62,$FE,$7D,$F0,$7E,$7B,$67,$75,$C6,$BF,$D6,$47
	.dc.b	$1A,$DE,$1E,$8A,$FC,$55,$F1,$AD,$DF,$1B,$FD,$84,$71,$AD,$DF,$18
	.dc.b	$F5,$36,$77,$5C,$5F,$CF,$B6,$FB,$47,$EE,$DA,$E8,$F8,$51,$B9,$E9
	.dc.b	$19,$3F,$19,$A7,$99,$FD,$23,$5C,$9D,$63,$8D,$FC,$1C,$7F,$49,$3C
	.dc.b	$DB,$19,$AF,$5C,$7E,$73,$EC,$FF,$85,$57,$63,$F8,$93,$87,$B8,$71
	.dc.b	$1D,$9F,$C0,$47,$13,$5B,$7F,$01,$AF,$F7,$34,$F0,$6C,$66,$7B,$E1
	.dc.b	$57,$D9,$FF,$49,$7A,$35,$F1,$B7,$0F,$7F,$D9,$7F,$61,$9D,$F5,$F3
	.dc.b	$8F,$DF,$63,$FE,$F7,$6C,$FF,$2F,$BE,$53,$FE,$FB,$7F,$DD,$31,$E0
	.dc.b	$F1,$3C,$91,$88,$F8,$74,$FB,$3C,$D4,$9E,$07,$C3,$9D,$38,$38,$F2
	.dc.b	$F9,$4E,$0F,$4E,$AF,$85,$F6,$7F,$D0,$63,$E6,$35,$D1,$89,$D9,$1F
	.dc.b	$04,$FC,$BF,$45,$F2,$D4,$93,$C4,$FB,$10,$F2,$B8,$12,$3D,$5D,$5C
	.dc.b	$FC,$A9,$1C,$1D,$3F,$86,$FB,$3F,$F0,$33,$F8,$36,$B9,$F2,$3E,$F4
	.dc.b	$7D,$08,$FE,$C1,$F6,$7F,$E6,$97,$F1,$4D,$7F,$22,$DE,$4F,$CE,$F3
	.dc.b	$7F,$8D,$67,$F2,$1F,$67,$FE,$6B,$6F,$BF,$F8,$B8,$C4,$7E,$6A,$5F
	.dc.b	$4D,$9E,$99,$3C,$9F,$C9,$4F,$96,$15,$FE,$6A,$7B,$FB,$7D,$4F,$CC
	.dc.b	$7D,$93,$EC,$FF,$35,$1F,$87,$CF,$F6,$6D,$EB,$FE,$FD,$7F,$E7,$4E
	.dc.b	$9E,$3E,$17,$0F,$56,$4F,$EA,$EF,$DC,$FE,$CB,$27,$CC,$7E,$9F,$EC
	.dc.b	$FF,$CD,$53,$FB,$16,$BF,$B8,$6E,$7F,$C1,$3E,$CF,$C3,$39,$F7,$39
	.dc.b	$11,$FD,$D4,$7E,$61,$FE,$03,$F6,$7F,$EB,$47,$F7,$CD,$73,$F1,$DF
	.dc.b	$B9,$9F,$F2,$D4,$6B,$5F,$DE,$57,$FA,$42,$30,$76,$FA,$FF,$E0,$1F
	.dc.b	$63,$FE,$8E,$8F,$E1,$4D,$7F,$88,$B7,$2F,$E3,$BF,$66,$FB,$3E,$24
	.dc.b	$AE,$2F,$F0,$B8,$FA,$11,$FE,$2D,$FA,$27,$C5,$FF,$0F,$6C,$7F,$19
	.dc.b	$6B,$FC,$83,$E8,$6E,$1F,$CB,$26,$F3,$7D,$A6,$D7,$5B,$FE,$35,$78
	.dc.b	$84,$D8,$FD,$9C,$7F,$27,$6B,$FC,$BA,$38,$7D,$9B,$E5,$FF,$92,$36
	.dc.b	$7F,$97,$35,$FE,$6A,$FF,$1C,$4E,$71,$FE,$86,$DC,$7F,$22,$D9,$FF
	.dc.b	$28,$3C,$4F,$F3,$63,$DB,$E5,$3C,$27,$8B,$DB,$91,$92,$1E,$27,$E6
	.dc.b	$7D,$84,$8F,$98,$FA,$B8,$31,$1E,$F8,$B8,$F9,$A6,$E3,$FA,$6F,$B2
	.dc.b	$FE,$B7,$3F,$D7,$1B,$1F,$46,$BF,$A4,$43,$EC,$E9,$30,$FE,$A9,$0D
	.dc.b	$B1,$1F,$D7,$23,$83,$A9,$E7,$3D,$3E,$6F,$AF,$C5,$CA,$E7,$19,$7A
	.dc.b	$C6,$5E,$91,$FD,$6F,$EC,$FF,$87,$C7,$F6,$56,$BF,$DB,$1B,$9F,$F7
	.dc.b	$8F,$B3,$FE,$15,$3B,$F1,$B9,$11,$FE,$DD,$1A,$79,$47,$F6,$98,$69
	.dc.b	$4E,$13,$FD,$EA,$3F,$D7,$8F,$2F,$96,$E0,$71,$F8,$59,$19,$21,$C5
	.dc.b	$ED,$E0,$48,$F1,$BE,$1F,$47,$11,$DE,$2E,$5E,$D9,$FE,$FD,$F6,$4F
	.dc.b	$B3,$A4,$BF,$F0,$57,$D9,$FF,$7B,$7F,$C2,$33,$B7,$FC,$1C,$D7,$33
	.dc.b	$FE,$09,$9E,$17,$FE,$22,$AD,$3F,$E2,$48,$FD,$31,$FA,$78,$CF,$27
	.dc.b	$F3,$48,$FD,$0D,$FF,$88,$1B,$5C,$79,$BB,$73,$F7,$65,$7A,$1C,$BE
	.dc.b	$AF,$B2,$F2,$11,$FF,$13,$7D,$9F,$E6,$9F,$F6,$06,$09,$D4,$78,$5F
	.dc.b	$F1,$24,$EE,$1F,$5F,$FE,$1A,$F8,$47,$57,$F0,$AC,$FF,$D4,$91,$8D
	.dc.b	$FF,$C5,$95,$FC,$14,$E1,$5B,$D7,$FD,$A1,$5B,$C3,$FC,$75,$5B,$ED
	.dc.b	$FF,$47,$BF,$C3,$E3,$FA,$4B,$5F,$EB,$6D,$7F,$DB,$4F,$F0,$D9,$F2
	.dc.b	$F0,$FF,$8F,$DB,$3D,$AB,$CB,$F2,$19,$C7,$F9,$5D,$7B,$3C,$0C,$DF
	.dc.b	$92,$B5,$12,$30,$7A,$3C,$8F,$7F,$CF,$48,$C1,$0F,$F2,$DE,$68,$D4
	.dc.b	$62,$7E,$39,$8D,$BF,$CD,$75,$B9,$E5,$18,$9D,$6B,$F0,$AF,$8D,$6A
	.dc.b	$30,$FC,$C8,$CB,$DC,$C1,$0C,$12,$BC,$31,$DA,$3E,$6D,$72,$9C,$C7
	.dc.b	$17,$C8,$F8,$4F,$A7,$3C,$5F,$C4,$CF,$1A,$F2,$C7,$C4,$9C,$3D,$8E
	.dc.b	$91,$AA,$EB,$1E,$58,$C1,$1E,$06,$4F,$09,$C8,$CC,$72,$3D,$2F,$84
	.dc.b	$F4,$1E,$63,$81,$FA,$09,$F2,$B9,$38,$12,$3A,$4F,$E5,$2B,$E0,$5F
	.dc.b	$D9,$AF,$21,$C6,$BC,$24,$AE,$11,$E2,$9E,$13,$98,$F4,$CE,$9F,$03
	.dc.b	$F0,$5E,$8E,$63,$84,$7A,$4E,$95,$E4,$BF,$5D,$A3,$13,$A3,$E0,$C7
	.dc.b	$8A,$FA,$12,$3E,$69,$E2,$3C,$A4,$3C,$06,$4F,$29,$98,$E8,$F2,$BF
	.dc.b	$92,$E6,$FE,$4D,$F9,$9B,$1F,$15,$AF,$CC,$AE,$6F,$82,$B9,$BE,$03
	.dc.b	$17,$F1,$E7,$E0,$47,$33,$DF,$7D,$DA,$64,$F5,$47,$07,$C1,$7E,$2B
	.dc.b	$EF,$3E,$22,$57,$3A,$E2,$FF,$B3,$FE,$C3,$F9,$14,$33,$FE,$57,$F2
	.dc.b	$5F,$C4,$CD,$70,$49,$FF,$73,$4E,$9B,$B2,$B5,$1F,$EE,$8F,$AF,$7D
	.dc.b	$7F,$F9,$C2,$3F,$D0,$31,$87,$01,$C3,$B6,$BF,$E7,$87,$FD,$C4,$DF
	.dc.b	$85,$E2,$F1,$6C,$6A,$3F,$D1,$FF,$60,$F9,$A7,$CA,$78,$23,$0D,$F2
	.dc.b	$BF,$4D,$5C,$2B,$6F,$8F,$ED,$5C,$9E,$8C,$51,$89,$D4,$E6,$31,$1F
	.dc.b	$EA,$0E,$37,$D7,$7F,$A6,$A9,$29,$E4,$6B,$A6,$CE,$A3,$8F,$16,$2B
	.dc.b	$85,$F3,$79,$56,$E3,$15,$D1,$B1,$9A,$E5,$3D,$E3,$B1,$DF,$EC,$7E
	.dc.b	$CB,$8B,$F3,$46,$1E,$E7,$A7,$E0,$FE,$6C,$EE,$95,$B7,$DA,$D8,$ED
	.dc.b	$C3,$E2,$AE,$37,$D2,$F8,$CF,$BE,$1F,$60,$FA,$AF,$4F,$D8,$BE,$27
	.dc.b	$53,$9C,$BC,$53,$97,$D5,$7C,$AF,$8D,$F0,$6C,$EA,$F9,$57,$B8,$E1
	.dc.b	$1E,$DD,$7F,$D5,$6E,$EB,$E5,$E8,$FA,$4A,$63,$9A,$D7,$C5,$9D,$36
	.dc.b	$74,$D3,$B7,$D8,$FD,$6C,$1F,$15,$63,$0D,$9F,$23,$72,$F9,$55,$C2
	.dc.b	$7E,$0F,$D8,$BE,$6B,$F0,$3F,$62,$FB,$0F,$9D,$39,$AF,$B5,$39,$78
	.dc.b	$D7,$0A,$DB,$63,$24,$C3,$ED,$68,$FB,$1D,$CF,$C9,$EE,$F6,$61,$ED
	.dc.b	$6C,$64,$F9,$D1,$96,$9F,$27,$D5,$F9,$E7,$D0,$F5,$AC,$64,$98,$F8
	.dc.b	$A3,$2F,$DB,$D3,$FD,$9D,$3F,$86,$FA,$EF,$83,$E5,$7E,$31,$B1,$93
	.dc.b	$E4,$C6,$5A,$4A,$F9,$FF,$05,$C1,$F8,$3F,$1F,$F8,$4F,$88,$FC,$2B
	.dc.b	$6B,$F0,$4D,$9D,$C7,$E1,$67,$F0,$1E,$AC,$8F,$C8,$7A,$D6,$BE,$EC
	.dc.b	$DE,$EF,$DF,$4D,$AF,$E1,$B4,$FB,$16,$5F,$96,$FA,$2F,$D6,$46,$2B
	.dc.b	$F2,$91,$89,$D4,$7E,$A2,$75,$5F,$72,$B5,$39,$AF,$D0,$F6,$62,$BF
	.dc.b	$51,$F6,$0F,$B1,$FE,$1F,$E9,$3D,$8F,$DE,$4E,$71,$FE,$11,$39,$7F
	.dc.b	$6F,$E9,$B8,$F8,$FC,$EF,$F0,$38,$C4,$F0,$8F,$E0,$DF,$5E,$F9,$3F
	.dc.b	$E3,$BF,$62,$E2,$CC,$61,$DE,$5F,$C8,$BE,$C7,$F9,$9E,$3F,$C1,$B0
	.dc.b	$FD,$EE,$FF,$C4,$E7,$2F,$F0,$38,$7C,$87,$F2,$AF,$B1,$7C,$77,$8A
	.dc.b	$30,$FF,$19,$FA,$47,$57,$F3,$0F,$B1,$FE,$03,$8F,$F0,$D9,$CE,$7F
	.dc.b	$CA,$67,$3D,$5F,$D2,$A5,$F3,$9F,$CE,$3E,$C5,$F6,$3D,$63,$0F,$F2
	.dc.b	$7F,$B1,$7D,$5F,$F3,$EF,$10,$65,$C3,$BF,$81,$D7,$B7,$75,$FE,$75
	.dc.b	$3F,$DA,$5F,$E5,$37,$F0,$B4,$FE,$AF,$F6,$3E,$8D,$FF,$93,$4E,$6B
	.dc.b	$FA,$C4,$E6,$BF,$B1,$C7,$C6,$3F,$A3,$43,$E6,$BF,$9F,$46,$6F,$FA
	.dc.b	$14,$E6,$3F,$9C,$FD,$8F,$C4,$7F,$B2,$7A,$5F,$CF,$6B,$F9,$B4,$E5
	.dc.b	$FE,$D9,$19,$C3,$FB,$9F,$D8,$BD,$BF,$E9,$9C,$AF,$A3,$FE,$9D,$39
	.dc.b	$CB,$FA,$3C,$E5,$FE,$93,$0F,$88,$DF,$D2,$70,$6C,$FF,$4E,$FB,$1F
	.dc.b	$F1,$23,$C2,$3F,$AD,$46,$34,$FE,$99,$7C,$EB,$87,$9B,$FD,$7F,$EC
	.dc.b	$5E,$7F,$08,$C3,$FD,$8F,$E1,$9E,$EE,$6F,$FA,$EC,$61,$E6,$62,$38
	.dc.b	$1D,$7D,$0B,$1F,$D9,$3E,$BF,$F0,$EF,$F7,$9F,$B1,$7C,$8F,$F7,$19
	.dc.b	$CE,$BF,$DD,$67,$35,$C7,$B7,$FC,$17,$18,$9F,$ED,$15,$C3,$97,$FC
	.dc.b	$3B,$39,$7F,$C4,$B1,$9B,$FE,$2E,$DA,$FF,$12,$57,$D9,$9C,$DF,$8F
	.dc.b	$8B,$FC,$33,$F6,$3F,$D5,$EF,$BC,$63,$DD,$D5,$FF,$84,$EB,$59,$78
	.dc.b	$2B,$53,$F0,$7C,$EF,$F0,$97,$D3,$7B,$DB,$19,$26,$3E,$C3,$4F,$F8
	.dc.b	$32,$D5,$FE,$17,$7F,$6B,$19,$8C,$78,$59,$64,$7F,$85,$2B,$FC,$87
	.dc.b	$E2,$FF,$91,$3E,$C5,$F6,$1F,$E4,$9B,$FF,$11,$56,$84,$E9,$FF,$23
	.dc.b	$47,$F9,$4E,$31,$1F,$D0,$AB,$A5,$73,$BB,$3F,$D1,$84,$E1,$FF,$2A
	.dc.b	$5F,$3A,$FF,$3D,$4E,$71,$F7,$DE,$EF,$C9,$87,$82,$73,$1E,$6B,$E2
	.dc.b	$D3,$CF,$5C,$5F,$09,$93,$C2,$E6,$FC,$18,$78,$E7,$D1,$18,$9D,$47
	.dc.b	$1B,$F0,$E1,$F0,$A3,$E4,$1F,$E9,$0A,$FF,$44,$BE,$03,$07,$FA,$5F
	.dc.b	$EC,$7F,$9D,$FC,$8F,$FA,$59,$B1,$9F,$2F,$FD,$50,$D9,$D7,$D5,$7F
	.dc.b	$9E,$E3,$93,$5F,$9D,$48,$C4,$7F,$76,$8C,$63,$C8,$FA,$5A,$36,$FE
	.dc.b	$F7,$9F,$DC,$8C,$5F,$D7,$BE,$2F,$C9,$6C,$EE,$B8,$D7,$F8,$10,$E3
	.dc.b	$5B,$C3,$D1,$5F,$88,$BE,$35,$BB,$E3,$7F,$E9,$88,$E3,$5B,$BE,$31
	.dc.b	$EB,$6C,$EE,$B8,$BF,$9C,$6D,$F6,$8F,$F5,$1B,$5D,$1F,$0A,$37,$3D
	.dc.b	$23,$27,$F9,$D6,$BD,$78,$7F,$AC,$1B,$A7,$8E,$39,$4F,$FA,$0A,$72
	.dc.b	$F9,$4F,$F4,$2E,$1E,$1F,$B1,$FD,$DC,$7F,$A2,$1A,$FF,$A0,$DB,$5F
	.dc.b	$EB,$E7,$EE,$46,$F6,$FF,$60,$61,$FE,$32,$9D,$B6,$33,$3F,$EC,$29
	.dc.b	$E2,$FF,$A3,$FE,$C7,$F9,$EB,$FE,$98,$6B,$9E,$9E,$2B,$EC,$3F,$AA
	.dc.b	$3B,$8C,$C7,$EC,$F0,$FF,$21,$B6,$7F,$DA,$35,$D3,$1F,$F6,$BB,$FE
	.dc.b	$16,$8F,$07,$85,$DA,$31,$1F,$EA,$3A,$7D,$8E,$6A,$4F,$03,$FD,$5F
	.dc.b	$3A,$70,$71,$E5,$F2,$9C,$1E,$9F,$1B,$EC,$FB,$14,$D3,$D7,$16,$7F
	.dc.b	$D7,$73,$A6,$CE,$9F,$F5,$FF,$A0,$F9,$1F,$8E,$D8,$CC,$73,$7F,$D7
	.dc.b	$D1,$FE,$AE,$78,$F9,$5C,$0F,$97,$E5,$72,$F2,$A4,$F6,$74,$FF,$B3
	.dc.b	$BE,$C7,$FB,$CB,$7C,$6F,$AF,$18,$8F,$F6,$63,$F4,$78,$BF,$CE,$FF
	.dc.b	$63,$FE,$5F,$9F,$F6,$BB,$5F,$F6,$93,$78,$D3,$FE,$7E,$E6,$FF,$15
	.dc.b	$57,$FB,$7B,$EC,$7F,$CC,$4D,$F7,$7F,$0F,$18,$8F,$F6,$E4,$BE,$93
	.dc.b	$3D,$92,$79,$3F,$8E,$9F,$34,$05,$7F,$98,$9F,$07,$B3,$D8,$FC,$AF
	.dc.b	$D8,$3E,$C7,$F2,$F1,$F8,$AD,$3D,$35,$CE,$31,$A7,$FA,$61,$B1,$DE
	.dc.b	$FF,$CD,$AF,$FB,$EA,$3D,$EF,$FB,$D5,$B5,$D2,$3F,$CE,$BE,$AF,$8B
	.dc.b	$E6,$3F,$4F,$F6,$3F,$E6,$49,$F5,$56,$E3,$11,$FE,$FF,$6C,$73,$FB
	.dc.b	$1F,$EC,$CE,$FD,$CE,$64,$AE,$79,$FF,$AC,$2B,$F4,$10,$AF,$27,$D8
	.dc.b	$FF,$AC,$A3,$67,$CB,$70,$EE,$BF,$D7,$31,$89,$F1,$73,$FF,$00,$8F
	.dc.b	$87,$5B,$39,$79,$39,$3F,$82,$E0,$FA,$DF,$E0,$FF,$62,$F9,$AA,$4B
	.dc.b	$F1,$8F,$B0,$7D,$8F,$02,$46,$AB,$2F,$09,$FA,$EB,$E8,$9F,$13,$FC
	.dc.b	$59,$B1,$D9,$AF,$A4,$7D,$7E,$37,$FB,$53,$78,$3F,$DB,$6D,$AF,$C9
	.dc.b	$35,$E3,$7D,$CF,$E4,$B7,$FB,$F8,$EC,$DB,$EB,$3F,$CA,$BE,$C5,$F2
	.dc.b	$D4,$97,$F1,$23,$F9,$42,$76,$E6,$36,$FA,$E3,$67,$F9,$A9,$E5,$F2
	.dc.b	$5D,$CF,$07,$B9,$23,$24,$39,$3F,$27,$EB,$E4,$7C,$97,$D1,$C3,$88
	.dc.b	$F9,$5A,$27,$19,$FE,$8F,$1F,$C1,$57,$13,$EF,$FD,$3F,$E9,$E7,$F5
	.dc.b	$0D,$DB,$E7,$DF,$F4,$E8,$7D,$8D,$25,$F2,$DE,$BF,$6E,$76,$F6,$9E
	.dc.b	$53,$CF,$E6,$FA,$7C,$5C,$AF,$1C,$65,$F2,$C6,$5E,$91,$FD,$83,$EC
	.dc.b	$7F,$8C,$47,$E6,$1A,$FE,$2D,$B9,$FF,$03,$FB,$1F,$E2,$53,$BF,$1B
	.dc.b	$91,$1F,$D0,$C6,$9E,$51,$FD,$BA,$3E,$64,$6B,$4E,$0F,$E6,$61,$F0
	.dc.b	$DB,$F9,$49,$3D,$23,$91,$87,$84,$61,$F0,$F8,$3D,$C8,$F9,$3D,$0A
	.dc.b	$F7,$8B,$97,$B6,$7F,$C0,$7F,$60,$FB,$1A,$4B,$FD,$EF,$D8,$FF,$81
	.dc.b	$9F,$D8,$4E,$DF,$E0,$2D,$73,$3F,$E0,$E9,$E1,$7F,$CB,$6B,$4F,$F2
	.dc.b	$F8,$FD,$29,$FA,$68,$CF,$27,$F4,$7B,$4F,$E8,$DF,$E0,$0D,$AE,$5D
	.dc.b	$1D,$FA,$7C,$0C,$AF,$4B,$97,$D7,$F6,$3E,$52,$3F,$C6,$BE,$C7,$F3
	.dc.b	$4F,$F5,$7C,$13,$A8,$F0,$BF,$E2,$A9,$DC,$3E,$C7,$FD,$10,$FA,$5A
	.dc.b	$FF,$61,$6B,$BF,$81,$FF,$55,$3B,$7F,$C7,$BF,$08,$EB,$FE,$21,$AF
	.dc.b	$F6,$98,$C7,$07,$F8,$EB,$0F,$E1,$A7,$F0,$CC,$3F,$92,$37,$6F,$EB
	.dc.b	$8D,$BE,$5F,$29,$FD,$D2,$BD,$1E,$07,$89,$F3,$35,$C9,$9A,$EC,$FF
	.dc.b	$B0,$5F,$C2,$37,$E4,$B8,$46,$23,$E1,$46,$0F,$6E,$7F,$CC,$F3,$F7
	.dc.b	$78,$5E,$B9,$F6,$79,$72,$FF,$A6,$E0,$8F,$8B,$93,$B7,$55,$F0,$26
	.dc.b	$EB,$FE,$06,$C9,$7B,$AE,$75,$AD,$7E,$B3,$D3,$0C,$4F,$D7,$8B,$7C
	.dc.b	$A7,$33,$F7,$E7,$37,$D6,$3C,$47,$62,$1F,$DB,$EB,$93,$8C,$3E,$AD
	.dc.b	$72,$BE,$B8,$FB,$EB,$58,$FD,$89,$CD,$6E,$72,$79,$CB,$87,$67,$C6
	.dc.b	$4C,$3B,$BF,$E0,$A8,$D4,$62,$35,$7F,$11,$D6,$3E,$D9,$D1,$27,$53
	.dc.b	$ED,$9D,$46,$27,$53,$E5,$35,$87,$A9,$F3,$D7,$07,$53,$C1,$F9,$C7
	.dc.b	$8A,$78,$3E,$61,$7E,$AB,$F0,$12,$39,$BF,$25,$FA,$67,$E6,$70,$FC
	.dc.b	$B9,$1F,$CF,$35,$FD,$AE,$3F,$4D,$B1,$9C,$3F,$5B,$F3,$9F,$B3,$3C
	.dc.b	$AE,$27,$F3,$31,$8B,$FE,$08,$EE,$3E,$C9,$EE,$3F,$72,$E5,$F7,$19
	.dc.b	$9E,$8E,$8E,$37,$E5,$8C,$92,$BE,$14,$F1,$7C,$2F,$90,$E3,$7D,$E7
	.dc.b	$84,$7C,$82,$1F,$C7,$6F,$55,$D2,$FC,$0F,$49,$F2,$1E,$88,$CB,$FC
	.dc.b	$86,$78,$BD,$23,$2F,$53,$C8,$71,$BD,$B9,$30,$6A,$FC,$EF,$B4,$E9
	.dc.b	$5C,$2B,$A9,$1C,$57,$EB,$9B,$3F,$29,$AF,$F9,$37,$EC,$7F,$98,$BF
	.dc.b	$E9,$96,$A6,$D7,$F9,$FB,$0F,$EB,$11,$FE,$49,$9F,$EB,$46,$F8,$3F
	.dc.b	$96,$46,$63,$71,$8F,$A6,$91,$FC,$CE,$51,$88,$FE,$8F,$F6,$0F,$AF
	.dc.b	$FE,$6C,$D8,$FE,$74,$D7,$71,$AD,$A4,$69,$AF,$F5,$2F,$B1,$7C,$DB
	.dc.b	$E5,$BF,$A2,$46,$1B,$E5,$37,$CD,$6E,$FF,$D3,$FF,$5A,$F0,$7F,$D6
	.dc.b	$FC,$72,$31,$C3,$FE,$9A,$84,$7F,$BB,$E3,$0E,$AB,$FD,$2D,$3A,$AE
	.dc.b	$75,$FE,$B2,$AD,$4F,$5E,$1C,$DF,$FA,$A6,$73,$5F,$EB,$8F,$AC,$F9
	.dc.b	$0F,$85,$B3,$DB,$D5,$F8,$13,$9E,$0E,$78,$AF,$6F,$B1,$B3,$FE,$C4
	.dc.b	$AE,$37,$E8,$9C,$BF,$EE,$1F,$AF,$7D,$47,$4F,$B0,$7C,$57,$A2,$95
	.dc.b	$AC,$7E,$35,$69,$FF,$57,$C6,$67,$E2,$CE,$5F,$F5,$24,$67,$2F,$A3
	.dc.b	$39,$7F,$DB,$D1,$9A,$FF,$17,$36,$BE,$47,$6F,$8E,$31,$5F,$07,$D6
	.dc.b	$F8,$79,$FF,$B0,$F0,$F9,$5F,$1F,$89,$91,$F3,$5D,$C6,$31,$FB,$7B
	.dc.b	$BE,$C3,$E6,$AF,$89,$B1,$86,$E1,$F8,$46,$D7,$27,$E4,$7D,$6B,$E6
	.dc.b	$7E,$BF,$D8,$3E,$BF,$31,$86,$E1,$F9,$19,$D3,$FE,$DA,$AE,$4E,$67
	.dc.b	$78,$7F,$B3,$E7,$2F,$DD,$FB,$0C,$EF,$F9,$DF,$03,$D7,$7E,$C6,$E1
	.dc.b	$F9,$C6,$DF,$C2,$6D,$7E,$2F,$BF,$E6,$9F,$3B,$D7,$B7,$F9,$78,$B7
	.dc.b	$F8,$D9,$B4,$AF,$F3,$46,$5F,$ED,$47,$F1,$3F,$59,$F1,$BC,$AF,$C8
	.dc.b	$36,$32,$7C,$18,$CB,$49,$87,$D8,$FA,$FE,$0F,$E7,$7E,$3F,$F0,$7F
	.dc.b	$2B,$F8,$46,$D7,$EB,$5B,$3B,$8F,$D8,$4F,$E0,$3D,$79,$1F,$C0,$3D
	.dc.b	$7B,$5F,$86,$9B,$A7,$F0,$99,$B5,$FE,$85,$FB,$06,$5F,$94,$FA,$3F
	.dc.b	$D5,$46,$27,$F7,$D1,$89,$D4,$31,$61,$F2,$5C,$5F,$17,$F7,$57,$C6
	.dc.b	$B8,$56,$E5,$5B,$8F,$E0,$9E,$06,$E7,$AD,$7F,$A6,$FE,$BD,$F6,$1F
	.dc.b	$C3,$3C,$07,$B3,$FC,$06,$73,$5F,$C9,$67,$2F,$80,$FF,$52,$79,$2E
	.dc.b	$3E,$5F,$3D,$FC,$26,$31,$3C,$63,$F8,$67,$D7,$3E,$53,$F8,$DF,$D8
	.dc.b	$38,$F3,$18,$77,$97,$F2,$0F,$B0,$F8,$D8,$F8,$77,$F9,$18,$AB,$51
	.dc.b	$DB,$E4,$7F,$93,$FD,$83,$E3,$F1,$87,$F3,$F8,$C3,$CE,$9F,$48,$EA
	.dc.b	$FE,$5F,$F6,$1F,$C3,$67,$E4,$9F,$5E,$75,$5F,$D2,$67,$5C,$FF,$4E
	.dc.b	$5F,$3B,$FC,$E3,$EC,$1F,$4F,$FC,$B6,$73,$87,$F5,$19,$CB,$E3,$F7
	.dc.b	$DF,$53,$FD,$03,$C4,$1F,$CD,$63,$33,$FD,$66,$32,$F1,$F7,$F8,$7B
	.dc.b	$3B,$6C,$75,$DB,$FA,$5F,$D8,$7A,$9F,$69,$FE,$96,$33,$C1,$FD,$0E
	.dc.b	$7E,$3C,$7F,$62,$97,$CC,$E7,$8F,$9B,$FD,$27,$EC,$3E,$1D,$FE,$5B
	.dc.b	$D0,$FE,$80,$FF,$BA,$E3,$38,$7E,$82,$72,$FE,$8F,$EB,$DE,$D7,$F5
	.dc.b	$3C,$DF,$47,$FD,$56,$73,$8F,$D9,$9C,$BF,$6E,$1F,$0D,$BF,$A4,$E0
	.dc.b	$D9,$FC,$07,$D8,$7F,$87,$1E,$11,$FD,$86,$31,$97,$C4,$BE,$75,$C3
	.dc.b	$CD,$FE,$CD,$F6,$0F,$3F,$84,$61,$FE,$D5,$F0,$8F,$6F,$37,$F8,$F8
	.dc.b	$C3,$CC,$C4,$70,$3A,$FA,$16,$3F,$2D,$F6,$1F,$80,$CF,$FC,$27,$F6
	.dc.b	$0F,$91,$FE,$F1,$39,$F0,$BF,$C3,$F3,$9B,$E3,$DD,$FE,$0A,$8C,$61
	.dc.b	$F8,$0B,$E1,$5F,$C3,$27,$37,$FC,$36,$73,$5F,$C5,$DB,$09,$FB,$32
	.dc.b	$9D,$B7,$8B,$FC,$33,$F6,$1F,$D6,$2F,$A4,$63,$DD,$F4,$B7,$1F,$7B
	.dc.b	$6F,$CD,$E7,$7F,$88,$BA,$1F,$4D,$F0,$23,$0D,$BF,$8F,$1F,$12,$B0
	.dc.b	$7E,$7B,$1E,$33,$FC,$05,$FD,$8C,$66,$BF,$C4,$51,$83,$F9,$71,$C2
	.dc.b	$35,$3F,$C2,$5F,$E0,$83,$EC,$1F,$23,$FC,$F6,$3F,$9E,$DF,$F8,$B5
	.dc.b	$DC,$E9,$FE,$77,$1F,$D0,$63,$11,$C7,$C6,$FE,$8D,$18,$C3,$C4,$EA
	.dc.b	$79,$DF,$79,$CB,$FE,$80,$8F,$E3,$A7,$95,$D5,$F7,$75,$7D,$9C,$E1
	.dc.b	$6B,$FD,$13,$35,$F5,$C5,$7C,$27,$A9,$C3,$E0,$C3,$D1,$7D,$E3,$99
	.dc.b	$E3,$38,$14,$F8,$05,$C3,$D6,$F0,$3D,$87,$B0,$B7,$87,$E9,$9F,$EA
	.dc.b	$0F,$B0,$FE,$85,$F2,$3C,$1B,$1F,$15,$AE,$FC,$2F,$F7,$DB,$9F,$A9
	.dc.b	$FE,$EA,$DC,$39,$46,$23,$B3,$5F,$C0,$52,$38,$E5,$F7,$34,$6B,$F8
	.dc.b	$58,$C3,$63,$8C,$F0,$6E,$1F,$45,$E3,$85,$AE,$0F,$F8,$2F,$0F,$05
	.dc.b	$7D,$76,$C7,$CD,$AE,$2D,$AE,$2A,$77,$87,$E4,$9B,$5F,$9B,$38,$C7
	.dc.b	$A2,$7B,$61,$E7,$7C,$4D,$7F,$47,$7D,$AB,$89,$EC,$8F,$11,$FD,$CD
	.dc.b	$F2,$E1,$FE,$1D,$6E,$BB,$C3,$C1,$5F,$D6,$27,$2F,$F8,$AD,$AF,$AF
	.dc.b	$0F,$17,$D8,$7F,$08,$8F,$F0,$03,$5F,$EC,$2D,$C3,$FC,$58,$FD,$98
	.dc.b	$DE,$DF,$AD,$C7,$F7,$93,$B6,$C6,$67,$FC,$79,$3C,$5F,$ED,$BF,$61
	.dc.b	$FD,$0D,$FF,$08,$35,$CF,$4F,$15,$F5,$FF,$D6,$9D,$C6,$63,$F8,$16
	.dc.b	$1F,$C3,$5B,$3F,$E4,$BA,$E9,$8F,$F9,$15,$FF,$0E,$47,$83,$C2,$ED
	.dc.b	$18,$8F,$F0,$0D,$3E,$C3,$35,$27,$81,$FE,$27,$9D,$38,$38,$F2,$F9
	.dc.b	$4E,$0F,$4F,$8E,$F6,$7D,$82,$69,$EB,$8B,$3F,$E3,$39,$D3,$67,$4F
	.dc.b	$F8,$DF,$D1,$7C,$8F,$C7,$6C,$66,$39,$BF,$E1,$C8,$FF,$5E,$3C,$7C
	.dc.b	$9E,$07,$CB,$EA,$E5,$E4,$C9,$EC,$E9,$FF,$23,$7D,$87,$F7,$F6,$F8
	.dc.b	$BF,$5E,$31,$1F,$E2,$A7,$E8,$F1,$7F,$9D,$BF,$67,$78,$DF,$85,$E4
	.dc.b	$43,$EB,$F0,$DC,$3F,$CB,$D3,$AA,$E1,$F3,$1F,$E6,$F9,$CB,$FE,$F8
	.dc.b	$6F,$16,$BE,$07,$B7,$E0,$5E,$09,$E6,$D7,$FC,$B7,$82,$30,$FF,$A4
	.dc.b	$4C,$65,$E7,$FB,$8F,$5D,$7F,$A5,$9B,$26,$8F,$97,$F3,$FD,$04,$AF
	.dc.b	$3B,$EA,$E6,$F8,$5F,$22,$F8,$FB,$ED,$4B,$F8,$33,$97,$E1,$E2,$8F
	.dc.b	$27,$CA,$C9,$CC,$7C,$6F,$B6,$F8,$FD,$F2,$7E,$3C,$78,$3E,$D6,$47
	.dc.b	$67,$C7,$F4,$FE,$FC,$E4,$F8,$1F,$85,$9F,$D2,$F9,$29,$3F,$49,$FA
	.dc.b	$BE,$47,$B3,$E7,$7D,$BF,$01,$2B,$D7,$1F,$BD,$FA,$97,$B3,$49,$3F
	.dc.b	$E9,$38,$FF,$52,$36,$32,$FF,$9D,$A3,$8B,$BF,$8B,$CF,$37,$FB,$0E
	.dc.b	$73,$3B,$AF,$05,$7F,$0D,$BF,$F5,$44,$FF,$A8,$5A,$FF,$A0,$9B,$7C
	.dc.b	$FE,$27,$81,$E9,$F1,$A5,$71,$8F,$E3,$1F,$56,$D6,$92,$7F,$D5,$D1
	.dc.b	$FC,$59,$3B,$73,$1B,$7F,$D2,$71,$B3,$FC,$BC,$F2,$F9,$1E,$E7,$83
	.dc.b	$DC,$91,$92,$1C,$9C,$BD,$23,$8B,$E4,$8F,$89,$F5,$32,$3E,$23,$EC
	.dc.b	$FA,$CC,$47,$FA,$9B,$E9,$7F,$C8,$7E,$37,$F8,$F5,$B1,$FE,$FB,$D9
	.dc.b	$7F,$CE,$E5,$F5,$94,$95,$FE,$D1,$87,$16,$23,$B4,$DE,$0F,$F7,$F5
	.dc.b	$6A,$78,$73,$75,$E6,$91,$82,$1E,$77,$2F,$A6,$3C,$86,$2B,$AC,$E2
	.dc.b	$7F,$DC,$9F,$58,$E8,$FE,$A8,$D8,$CC,$7F,$57,$FA,$CF,$E3,$11,$FE
	.dc.b	$CA,$9D,$3F,$EC,$C8,$F8,$6F,$FB,$BD,$AE,$E7,$97,$3F,$3E,$79,$3B
	.dc.b	$8C,$BC,$63,$2F,$58,$F1,$7D,$2C,$8E,$CF,$97,$A3,$11,$E3,$8B,$9B
	.dc.b	$8B,$FC,$ED,$97,$F4,$AA,$D1,$C8,$87,$AE,$FF,$BB,$7D,$5B,$EB,$3F
	.dc.b	$CA,$0D,$9F,$F2,$AB,$5F,$F3,$93,$5F,$F4,$20,$75,$87,$8E,$03,$C4
	.dc.b	$FD,$41,$FA,$99,$DE,$3F,$E8,$D8,$FD,$2C,$FF,$BE,$9B,$5C,$E3,$97
	.dc.b	$3F,$F8,$36,$5D,$16,$24,$F8,$2B,$85,$72,$BE,$D7,$C7,$EB,$25,$EE
	.dc.b	$7C,$51,$FE,$19,$FA,$CF,$D1,$CF,$A6,$36,$F9,$5D,$4E,$30,$EF,$16
	.dc.b	$3F,$D5,$15,$CB,$04,$E6,$FE,$2C,$61,$F1,$1F,$22,$78,$4E,$E3,$C4
	.dc.b	$68,$46,$9C,$4F,$37,$CC,$6E,$34,$E0,$97,$F3,$63,$2F,$AA,$39,$D7
	.dc.b	$B9,$A4,$38,$DF,$B8,$F3,$5F,$86,$B9,$49,$ED,$3A,$AF,$65,$F8,$23
	.dc.b	$E8,$D7,$5A,$DB,$63,$2F,$80,$F8,$CF,$37,$46,$0A,$7C,$83,$81,$F2
	.dc.b	$1E,$0E,$27,$C4,$FD,$4A,$F2,$CE,$AB,$D6,$79,$27,$9D,$FB,$E7,$CB
	.dc.b	$39,$36,$E0,$93,$E5,$38,$D7,$67,$CF,$5C,$DF,$11,$F6,$67,$C4,$72
	.dc.b	$7D,$47,$09,$E4,$F4,$3E,$0D,$FB,$1F,$85,$7E,$47,$C4,$D7,$35,$D6
	.dc.b	$78,$5F,$8E,$72,$70,$9F,$14,$64,$D4,$F0,$BF,$1D,$F5,$9D,$47,$E2
	.dc.b	$EF,$CC,$7D,$1B,$F6,$BF,$00,$FC,$43,$F8,$D7,$DC,$7B,$1D,$BE,$43
	.dc.b	$E1,$9D,$2B,$D9,$7E,$38,$E0,$72,$3C,$05,$8D,$CF,$4B,$F5,$B6,$FD
	.dc.b	$13,$EB,$7D,$B1,$F0,$9B,$5E,$99,$D3,$F6,$4E,$D3,$F3,$4E,$8E,$4F
	.dc.b	$43,$97,$67,$37,$E1,$3A,$30,$7D,$32,$9F,$0E,$BD,$2F,$F0,$7F,$AA
	.dc.b	$77,$FF,$9A,$63,$F9,$A8,$8C,$F9,$DF,$CF,$23,$F9,$CC,$3B,$1F,$39
	.dc.b	$99,$E1,$7C,$DF,$E7,$FC,$1F,$CF,$F2,$D3,$63,$A3,$5F,$E9,$2D,$7F
	.dc.b	$9C,$7D,$3B,$93,$FC,$F5,$58,$6C,$FF,$0B,$9F,$F5,$4C,$FF,$A3,$9E
	.dc.b	$0D,$9D,$EA,$9C,$D7,$FA,$21,$E2,$D9,$DF,$D4,$7F,$A0,$6F,$FD,$32
	.dc.b	$D9,$FF,$47,$FA,$DF,$EA,$29,$CF,$9B,$C0,$8F,$FA,$82,$35,$3F,$E9
	.dc.b	$6A,$E5,$8F,$FA,$55,$B5,$FD,$5D,$B3,$FC,$9E,$B8,$D6,$EF,$E0,$CB
	.dc.b	$EA,$5F,$45,$EE,$FA,$A7,$A9,$A9,$CE,$7F,$EB,$29,$CB,$FD,$26,$33
	.dc.b	$7F,$AE,$3F,$BB,$B8,$C7,$FD,$55,$AB,$C9,$E9,$97,$C6,$F3,$E6,$3F
	.dc.b	$1B,$E0,$2E,$5E,$F9,$B8,$FB,$E6,$D7,$DE,$FA,$AF,$AB,$93,$DC,$B1
	.dc.b	$86,$CF,$DD,$6E,$3F,$65,$B8,$79,$9B,$3F,$8A,$7E,$3F,$D3,$BE,$33
	.dc.b	$F0,$7F,$54,$FA,$AC,$C6,$1B,$7F,$DB,$6B,$83,$F8,$2F,$5B,$39,$7E
	.dc.b	$27,$EA,$BD,$37,$EA,$6D,$7E,$4D,$B3,$B8,$FC,$B4,$FE,$3B,$CC,$91
	.dc.b	$F8,$EF,$32,$D7,$E1,$E6,$EB,$F8,$79,$B5,$F8,$79,$B5,$FD,$B7,$A5
	.dc.b	$58,$86,$7F,$E9,$BB,$FC,$17,$B9,$F1,$3D,$4F,$C4,$36,$3E,$04,$7D
	.dc.b	$4F,$AA,$C5,$FE,$67,$E0,$7F,$4B,$F0,$DF,$A6,$6D,$FE,$99,$B3,$B8
	.dc.b	$FD,$4C,$FD,$EF,$3A,$47,$ED,$FC,$FB,$5F,$B3,$9B,$BF,$E4,$26,$D7
	.dc.b	$F0,$49,$B5,$FC,$06,$6D,$7E,$57,$E8,$BF,$7F,$D8,$6F,$87,$87,$3F
	.dc.b	$3D,$32,$FF,$17,$DF,$E7,$7E,$6F,$19,$7E,$8E,$1F,$4E,$DF,$F8,$5A
	.dc.b	$E7,$B1,$87,$F8,$7F,$D3,$32,$B8,$E1,$B9,$78,$33,$FF,$74,$B7,$6F
	.dc.b	$E2,$F1,$8D,$3C,$4A,$3F,$8D,$FD,$23,$C0,$A6,$3C,$6B,$87,$B3,$E8
	.dc.b	$F9,$70,$7F,$9C,$C3,$88,$FF,$33,$3D,$5E,$1C,$7F,$CA,$3A,$5D,$B4
	.dc.b	$C7,$87,$6B,$E4,$CB,$E2,$F8,$7D,$7F,$16,$CF,$CE,$FA,$3E,$6D,$BF
	.dc.b	$E6,$B1,$8E,$1F,$A0,$A3,$3E,$7E,$BC,$EB,$7C,$DF,$E7,$BE,$33,$D6
	.dc.b	$A6,$3D,$6B,$5F,$E6,$7F,$A0,$7A,$1F,$CA,$BC,$17,$C7,$7F,$2C,$9C
	.dc.b	$D7,$F9,$B6,$72,$F1,$EC,$D7,$C4,$ED,$B1,$FE,$73,$FA,$1F,$C2,$3B
	.dc.b	$31,$E0,$6D,$D7,$3F,$F3,$AF,$A1,$78,$DF,$CF,$2F,$EA,$72,$FF,$AA
	.dc.b	$27,$37,$FE,$7F,$F6,$3F,$00,$F2,$31,$DD,$C7,$1F,$E8,$B3,$AB,$FE
	.dc.b	$78,$1C,$76,$FF,$AA,$F0,$7D,$0F,$0C,$3F,$A6,$FC,$FB,$E8,$7F,$B4
	.dc.b	$FD,$0F,$CD,$D3,$FD,$21,$39,$7F,$53,$19,$CF,$BC,$E5,$FD,$54,$67
	.dc.b	$2F,$F5,$D4,$E5,$FC,$A4,$66,$BF,$C2,$73,$96,$DF,$8B,$D9,$F2,$BE
	.dc.b	$6E,$2F,$EF,$BF,$43,$FC,$DB,$2F,$0C,$63,$D4,$D6,$1F,$BE,$AD,$67
	.dc.b	$E2,$AD,$4F,$B3,$E7,$7F,$C0,$FD,$2F,$9B,$F8,$71,$86,$CE,$CF,$87
	.dc.b	$96,$0F,$9F,$87,$18,$FF,$5B,$2F,$A5,$FE,$F2,$E6,$77,$18,$A9,$5F
	.dc.b	$EB,$D9,$FF,$74,$FD,$17,$FB,$AB,$E7,$DF,$23,$FE,$38,$C7,$91,$FE
	.dc.b	$E5,$AD,$89,$DB,$FE,$E3,$C3,$FD,$E9,$18,$8F,$F2,$0D,$F5,$BE,$95
	.dc.b	$A7,$FC,$85,$19,$9E,$58,$57,$89,$DE,$78,$C6,$63,$C7,$7D,$F0,$F8
	.dc.b	$0F,$86,$33,$5C,$1E,$47,$F9,$50,$5F,$7A,$E3,$3E,$49,$D9,$FE,$CD
	.dc.b	$C3,$D0,$F1,$AE,$11,$88,$F3,$DF,$8E,$31,$1F,$20,$FF,$2E,$4F,$F9
	.dc.b	$75,$C1,$DC,$F2,$7D,$03,$E8,$5F,$45,$FE,$63,$8C,$36,$33,$3F,$E6
	.dc.b	$B9,$CB,$7E,$8B,$FD,$C8,$DC,$F9,$46,$23,$BB,$5F,$F3,$BC,$A3,$8E
	.dc.b	$5F,$6F,$36,$DF,$E7,$C8,$C3,$63,$8C,$F0,$6E,$1F,$6D,$E3,$85,$AE
	.dc.b	$08,$61,$E1,$AF,$C1,$36,$3E,$9D,$72,$6D,$72,$78,$DF,$85,$E9,$5C
	.dc.b	$70,$F3,$C7,$84,$EB,$3B,$78,$E1,$97,$A9,$A7,$03,$5F,$F4,$EE,$3E
	.dc.b	$38,$C4,$FA,$E3,$13,$F0,$A3,$99,$FC,$A2,$32,$EF,$1F,$D8,$4E,$F0
	.dc.b	$FD,$57,$D1,$7F,$07,$AF,$8D,$87,$65,$19,$7F,$99,$B6,$34,$FD,$B8
	.dc.b	$E3,$5F,$EB,$48,$D3,$76,$F0,$CF,$D3,$9C,$B7,$2C,$AF,$A2,$FF,$03
	.dc.b	$3D,$0F,$B4,$E1,$FE,$8A,$DC,$3D,$1F,$44,$F0,$B9,$CE,$72,$FF,$63
	.dc.b	$CF,$EE,$DB,$AE,$BA,$F7,$E1,$CC,$BC,$07,$8D,$F2,$DB,$E6,$F5,$3E
	.dc.b	$29,$B9,$F8,$A6,$E3,$E0,$E7,$E2,$62,$FF,$29,$87,$FA,$81,$EF,$F4
	.dc.b	$49,$F8,$6D,$8F,$1C,$ED,$FE,$CB,$19,$6C,$F0,$3C,$FD,$9F,$33,$CF
	.dc.b	$7C,$8D,$25,$78,$2B,$5E,$64,$8F,$83,$5F,$E9,$D7,$C3,$D9,$CF,$CA
	.dc.b	$91,$A7,$83,$FD,$E3,$E8,$BF,$C0,$AF,$CA,$6B,$9F,$27,$EF,$C7,$D5
	.dc.b	$87,$03,$E8,$BF,$C0,$11,$FE,$5E,$5F,$41,$F8,$A9,$D3,$63,$73,$FE
	.dc.b	$61,$FA,$27,$C2,$6D,$F9,$17,$FE,$4B,$6F,$1F,$E8,$1D,$D7,$F9,$B3
	.dc.b	$E8,$B9,$D7,$F9,$45,$B5,$FA,$66,$CF,$A1,$F2,$BF,$E7,$4F,$A2,$F6
	.dc.b	$51,$A4,$F7,$3C,$9F,$29,$2B,$FC,$F7,$8A,$3D,$1C,$7E,$FF,$71,$F1
	.dc.b	$57,$DD,$98,$7F,$A0,$A7,$2F,$85,$F8,$D9,$7F,$94,$E3,$C9,$ED,$E2
	.dc.b	$7E,$27,$D1,$7C,$5E,$D9,$5E,$28,$F0,$7D,$14,$8E,$CF,$8F,$E6,$FA
	.dc.b	$E7,$27,$E0,$3E,$0C,$FE,$87,$C5,$C9,$F6,$BF,$4F,$C3,$F6,$7B,$FE
	.dc.b	$DF,$7E,$57,$AE,$3F,$81,$FC,$FB,$DB,$A4,$9F,$DA,$C7,$FA,$35,$B1
	.dc.b	$97,$FC,$93,$1C,$5D,$FC,$2E,$79,$FF,$8F,$4E,$67,$75,$E0,$AF,$E2
	.dc.b	$76,$F8,$CA,$4A,$F0,$B5,$EF,$E2,$7A,$0C,$63,$FC,$AF,$E8,$9A,$FF
	.dc.b	$1E,$6D,$7F,$0E,$6B,$FE,$60,$7F,$8B,$3F,$D6,$5C,$CF,$39,$E4,$73
	.dc.b	$D7,$A7,$C1,$F5,$F8,$39,$87,$58,$CB,$DE,$32,$F7,$8C,$BE,$18,$CD
	.dc.b	$FA,$67,$E1,$FD,$0C,$AF,$E4,$BA,$23,$D9,$F4,$59,$9F,$F3,$CF,$D0
	.dc.b	$FF,$4B,$C3,$FA,$63,$7C,$1F,$EE,$92,$C3,$93,$FC,$C2,$5F,$45,$FD
	.dc.b	$11,$B1,$98,$C7,$1E,$59,$19,$39,$47,$1E,$6E,$5C,$D2,$7B,$46,$5F
	.dc.b	$04,$65,$F4,$C6,$5F,$5C,$67,$2F,$5C,$FF,$58,$FA,$27,$45,$25,$FF
	.dc.b	$4A,$FA,$2F,$E0,$AF,$FA,$72,$76,$E7,$7F,$EC,$33,$F0,$A3,$55,$8E
	.dc.b	$8C,$F4,$49,$E9,$19,$7B,$46,$5F,$F5,$4D,$7E,$77,$2F,$EF,$8B,$E7
	.dc.b	$7B,$92,$7C,$DD,$15,$E3,$16,$B8,$CD,$E2,$FB,$37,$CE,$FD,$D7,$EA
	.dc.b	$AF,$EF,$7F,$44,$F9,$5A,$4A,$FE,$DB,$E2,$7F,$77,$7F,$B7,$CE,$DC
	.dc.b	$F4,$FF,$81,$27,$F4,$70,$9D,$3F,$EC,$68,$FF,$75,$3D,$3A,$38,$1C
	.dc.b	$7A,$6C,$64,$87,$17,$B7,$D0,$C8,$F2,$BF,$E1,$4F,$A2,$FC,$C4,$7F
	.dc.b	$92,$ED,$3B,$8F,$13,$BD,$B1,$BF,$F8,$AE,$B7,$5D,$0F,$F0,$6D,$F4
	.dc.b	$7F,$A0,$5F,$0B,$FD,$2D,$70,$9F,$F7,$1D,$6E,$31,$7D,$6F,$85,$7F
	.dc.b	$97,$EB,$51,$88,$EA,$78,$2F,$17,$C9,$EF,$3E,$99,$F3,$1E,$C2,$DF
	.dc.b	$F9,$16,$FA,$C6,$2B,$8B,$D4,$C1,$F1,$EB,$C5,$3C,$23,$A1,$2B,$A5
	.dc.b	$79,$2B,$6D,$9E,$62,$B9,$57,$C1,$8C,$C6,$27,$C7,$1D,$5F,$E1,$F1
	.dc.b	$98,$F2,$1C,$23,$B3,$B3,$E5,$38,$9E,$D1,$F4,$A3,$07,$68,$F4,$C7
	.dc.b	$9A,$FE,$63,$D8,$E0,$66,$FD,$AD,$3C,$45,$3C,$47,$BE,$BB,$BC,$EB
	.dc.b	$BB,$AA,$EE,$FD,$E3,$B9,$9A,$EE,$D3,$B9,$4E,$F7,$F1,$5E,$55,$F0
	.dc.b	$0E,$E5,$BF,$80,$48,$F8,$E7,$A6,$34,$E0,$EC,$E1,$F6,$1B,$8E,$4E
	.dc.b	$0D,$B8,$9F,$25,$FB,$8F,$7B,$8B,$F6,$DF,$59,$D4,$7E,$6E,$F9,$9F
	.dc.b	$26,$FF,$06,$7C,$C7,$DE,$E1,$F2,$1B,$3D,$55,$D6,$7B,$CF,$A2,$BB
	.dc.b	$D7,$E8,$4E,$E7,$43,$8C,$6E,$78,$C7,$53,$44,$BF,$63,$F5,$AF,$C0
	.dc.b	$DA,$E4,$D3,$95,$F8,$9F,$C0,$CF,$07,$EE,$CF,$FA,$BB,$C7,$78,$1F
	.dc.b	$DC,$A6,$46,$1C,$1F,$EE,$97,$0E,$1A,$F1,$FA,$2E,$7F,$19,$C2,$11
	.dc.b	$87,$9B,$FD,$FF,$E8,$B4,$E6,$3F,$C0,$78,$BE,$7F,$FA,$B6,$5F,$EC
	.dc.b	$18,$C7,$27,$FB,$89,$93,$FE,$9C,$9D,$3C,$2B,$FD,$17,$1F,$ED,$C8
	.dc.b	$C4,$F0,$AE,$53,$9E,$4B,$7F,$E9,$2B,$DC,$FF,$B9,$3E,$71,$E6,$EB
	.dc.b	$14,$62,$7C,$DE,$AB,$5D,$4E,$71,$F1,$9F,$E9,$E7,$17,$EB,$6D,$FF
	.dc.b	$BC,$AF,$93,$F0,$63,$35,$C7,$0E,$A7,$8E,$B7,$1D,$AF,$CB,$3E,$C8
	.dc.b	$7D,$0B,$E8,$7D,$5F,$44,$F6,$F5,$39,$CF,$E8,$4E,$5F,$F7,$1C,$66
	.dc.b	$FD,$67,$4B,$F8,$9A,$FF,$99,$EB,$71,$F0,$BD,$09,$1E,$BF,$7E,$DF
	.dc.b	$9F,$65,$63,$5F,$89,$56,$FE,$D7,$D1,$76,$F7,$7B,$B6,$3F,$02,$DC
	.dc.b	$7E,$16,$68,$C6,$1F,$88,$6D,$74,$3F,$13,$7F,$1F,$E8,$1F,$1F,$F7
	.dc.b	$FE,$89,$F4,$59,$8C,$36,$D9,$FC,$4B,$E3,$1F,$67,$D7,$D4,$F9,$3E
	.dc.b	$63,$E3,$DF,$C8,$6D,$7E,$69,$B3,$B8,$FC,$EC,$FE,$33,$CC,$91,$F8
	.dc.b	$5F,$36,$D7,$CB,$9B,$B7,$E3,$66,$D7,$E6,$E6,$D7,$E4,$E6,$AC,$BF
	.dc.b	$D2,$F5,$F8,$2F,$73,$E2,$7A,$9F,$88,$6C,$7C,$0C,$4F,$F9,$F2,$B8
	.dc.b	$DF,$D2,$FA,$1D,$BF,$7F,$E0,$7F,$4B,$F0,$DF,$A6,$6D,$FE,$7D,$B3
	.dc.b	$B8,$FD,$14,$FE,$3B,$CF,$91,$FC,$03,$CF,$B5,$F8,$09,$B9,$7F,$06
	.dc.b	$9B,$5F,$C0,$E6,$D7,$EF,$E6,$D7,$FA,$FF,$E7,$FF,$7D,$ED,$1B,$E8
	.dc.b	$61,$CF,$D0,$4C,$BF,$8B,$53,$E2,$B1,$97,$EC,$FE,$89,$BF,$F0,$95
	.dc.b	$CD,$63,$0F,$F0,$DF,$A2,$65,$70,$C3,$72,$FE,$49,$9F,$FB,$95,$BB
	.dc.b	$7F,$15,$8C,$67,$FC,$0D,$47,$F1,$9F,$A2,$77,$D3,$1E,$35,$C3,$F9
	.dc.b	$87,$D1,$72,$E2,$FE,$60,$D8,$FC,$03,$D1,$DF,$1F,$F2,$5E,$97,$65
	.dc.b	$31,$E1,$5A,$FE,$79,$2F,$8B,$E1,$F4,$5C,$5B,$3F,$D0,$7E,$8B,$9B
	.dc.b	$6F,$F9,$94,$63,$87,$F9,$03,$FC,$2A,$78,$7C,$0E,$BD,$3D,$B6,$3F
	.dc.b	$A6,$78,$AF,$63,$F9,$E3,$63,$3E,$C6,$27,$F9,$57,$D1,$3D,$0F,$E5
	.dc.b	$9E,$0B,$E3,$B3,$18,$77,$D9,$C3,$E2,$AE,$5F,$A7,$3F,$DC,$38,$23
	.dc.b	$F9,$B7,$D1,$7E,$12,$75,$C1,$FC,$32,$BF,$19,$18,$84,$ED,$FE,$67
	.dc.b	$3A,$7F,$9A,$C7,$D7,$7F,$8A,$FD,$13,$C6,$EF,$D1,$FD,$16,$31,$87
	.dc.b	$5F,$57,$F0,$AF,$7C,$39,$61,$FD,$77,$B7,$8C,$ED,$FE,$9D,$5C,$27
	.dc.b	$FA,$2D,$BE,$8B,$35,$FD,$4F,$E8,$5F,$43,$FD,$AF,$E8,$BE,$86,$1F
	.dc.b	$D3,$27,$39,$7F,$80,$67,$3E,$07,$F0,$EA,$FE,$5F,$5A,$7F,$B0,$4E
	.dc.b	$AB,$F9,$9C,$27,$55,$FE,$09,$AD,$4F,$2B,$F2,$1E,$3F,$66,$E1,$F3
	.dc.b	$F0,$FA,$CF,$F8,$4F,$E8,$BF,$9E,$DF,$9F,$8B,$FC,$09,$18,$CB,$CD
	.dc.b	$3E,$19,$CF,$A7,$F0,$5B,$7D,$7E,$77,$FB,$D7,$CE,$7C,$56,$C6,$49
	.dc.b	$7F,$14,$D3,$F8,$3C,$15,$FD,$FD,$7D,$15,$8C,$36,$78,$47,$F8,$16
	.dc.b	$BF,$AC,$FD,$17,$F5,$AF,$9F,$7C,$4F,$F8,$D2,$B5,$4A,$D4,$E6,$EC
	.dc.b	$FF,$92,$E6,$AC,$7C,$B3,$88,$EE,$30,$EE,$FF,$4B,$C3,$EF,$46,$23
	.dc.b	$ED,$DF,$4B,$E7,$1C,$70,$FF,$2B,$CE,$6F,$8C,$62,$74,$75,$C3,$C4
	.dc.b	$FA,$CC,$9E,$79,$E2,$FA,$CC,$DF,$83,$0F,$24,$7A,$9F,$34,$66,$FC
	.dc.b	$11,$8A,$E7,$5E,$A3,$FC,$CC,$78,$4F,$F3,$46,$18,$70,$7E,$6B,$E8
	.dc.b	$1F,$42,$FA,$2F,$F3,$74,$61,$B1,$99,$FF,$3C,$CE,$5B,$F4,$3F,$4D
	.dc.b	$B9,$F2,$8C,$47,$FA,$0E,$51,$88,$F1,$3C,$72,$FA,$99,$B6,$FD,$7C
	.dc.b	$61,$B1,$C3,$0F,$C1,$E1,$D1,$FC,$2C,$71,$C2,$D7,$07,$FC,$57,$5B
	.dc.b	$7E,$AB,$6F,$B4,$70,$3A,$E1,$E5,$BF,$04,$7D,$C7,$B3,$6F,$B2,$BE
	.dc.b	$93,$C2,$FA,$CE,$CE,$B8,$77,$7F,$CC,$33,$D2,$3F,$7E,$DD,$79,$3F
	.dc.b	$C4,$8C,$4F,$F9,$19,$AF,$EB,$B0,$FF,$50,$CF,$08,$FF,$25,$FD,$17
	.dc.b	$EF,$1F,$E8,$0D,$7F,$CE,$2D,$BF,$E1,$EF,$D8,$8D,$ED,$D1,$E6,$D5
	.dc.b	$1C,$6B,$F9,$3B,$C6,$3F,$CF,$BF,$45,$FE,$06,$7F,$A8,$35,$CF,$47
	.dc.b	$3F,$A2,$FF,$10,$46,$A3,$19,$C8,$C1,$CB,$C4,$F1,$59,$A7,$3F,$9C
	.dc.b	$B3,$B2,$1F,$E9,$0C,$11,$A8,$C7,$2F,$8C,$C6,$BF,$CC,$AB,$FD,$3D
	.dc.b	$3F,$DA,$7E,$8B,$FB,$83,$6B,$D5,$D5,$F1,$67,$33,$FE,$9E,$6D,$FF
	.dc.b	$3C,$6E,$DF,$D3,$B0,$EB,$5F,$EA,$4F,$39,$F1,$5F,$3D,$B1,$FD,$E5
	.dc.b	$AF,$5C,$B5,$E5,$C8,$D3,$F3,$7C,$9E,$5E,$4C,$9E,$4E,$DF,$F5,$9F
	.dc.b	$D1,$7D,$5D,$7F,$A8,$BF,$DF,$E3,$FC,$10,$D7,$7C,$7F,$85,$97,$D0
	.dc.b	$FE,$05,$FF,$77,$37,$6F,$F3,$F3,$6B,$F9,$E3,$FC,$51,$FF,$7B,$7C
	.dc.b	$87,$FA,$A1,$BE,$07,$58,$C4,$7F,$B5,$57,$CE,$78,$7F,$01,$FF,$2F
	.dc.b	$CF,$FB,$FD,$FC,$5B,$FE,$DB,$6E,$3F,$F0,$05,$DC,$7F,$B7,$F0,$FC
	.dc.b	$D3,$E4,$F0,$BD,$91,$FD,$8A,$3F,$C5,$75,$EE,$EB,$F2,$7B,$9E,$5F
	.dc.b	$83,$95,$EA,$8F,$89,$F4,$4F,$91,$A4,$9E,$71,$F6,$B0,$F8,$FC,$5E
	.dc.b	$02,$4F,$83,$27,$D3,$FC,$3F,$2A,$4F,$C8,$7E,$37,$2F,$83,$E1,$3C
	.dc.b	$3F,$09,$2B,$BC,$7B,$3E,$DF,$DB,$F3,$92,$7B,$47,$CF,$F6,$3E,$87
	.dc.b	$B3,$27,$EA,$C7,$B3,$8E,$47,$A9,$F7,$F2,$7F,$05,$F9,$E6,$17,$19
	.dc.b	$7F,$E5,$19,$CB,$67,$2F,$F7,$FE,$AF,$F1,$8C,$AB,$A5,$4A,$E5,$1C
	.dc.b	$38,$BF,$C4,$F5,$D2,$79,$1F,$EE,$C5,$EB,$F9,$49,$1E,$3F,$85,$AF
	.dc.b	$01,$3C,$1F,$44,$FB,$67,$F9,$17,$B8,$F8,$9A,$4A,$F7,$35,$E9,$7F
	.dc.b	$7F,$3F,$6F,$89,$9C,$BF,$C7,$FF,$42,$F8,$DF,$E5,$CD,$C3,$FC,$F0
	.dc.b	$D7,$CC,$FF,$82,$2F,$FC,$AE,$E6,$B9,$D7,$23,$9F,$9B,$6F,$9C,$57
	.dc.b	$9C,$78,$38,$BC,$3C,$72,$7C,$31,$97,$C7,$19,$7C,$71,$F1,$38,$24
	.dc.b	$7F,$A3,$F6,$FB,$4F,$AB,$C0,$CB,$F3,$FE,$87,$FC,$BF,$9F,$F9,$81
	.dc.b	$BA,$FF,$9C,$AF,$8D,$6F,$E8,$7F,$A9,$51,$A6,$38,$2C,$FF,$A7,$E7
	.dc.b	$4D,$9D,$34,$87,$53,$84,$F6,$F9,$0E,$FF,$0F,$27,$A4,$65,$ED,$19
	.dc.b	$7D,$91,$FD,$9F,$E8,$5F,$43,$FE,$38,$6C,$FF,$A3,$8F,$F4,$7B,$87
	.dc.b	$01,$C3,$B7,$5A,$7F,$A1,$BE,$87,$E2,$B7,$7F,$05,$FF,$77,$9F,$CD
	.dc.b	$5F,$E7,$5F,$F7,$DB,$FE,$C3,$6B,$9E,$8E,$07,$1E,$89,$1B,$8C,$BC
	.dc.b	$63,$2F,$13,$A7,$89,$E1,$23,$DF,$A2,$BE,$38,$B9,$7A,$E7,$FC,$1D
	.dc.b	$F4,$2F,$1F,$FC,$21,$F3,$FF,$9F,$C0,$1C,$29,$E0,$FF,$B7,$DA,$FE
	.dc.b	$F6,$36,$70,$8C,$DF,$FA,$C2,$3F,$DF,$93,$FE,$E7,$6B,$C7,$9F,$C7
	.dc.b	$D1,$E4,$64,$6D,$CB,$48,$71,$3A,$47,$07,$04,$AF,$14,$65,$F2,$4D
	.dc.b	$E7,$FF,$6A,$DA,$B1,$7F,$12,$3C,$F7,$C2,$3B,$9E,$27,$0F,$7A,$E7
	.dc.b	$5C,$A3,$E6,$5D,$BE,$D1,$C6,$BC,$33,$9A,$D3,$67,$47,$8D,$C3,$DC
	.dc.b	$C1,$A3,$ED,$47,$27,$15,$E0,$78,$1E,$43,$31,$B7,$31,$BB,$F6,$92
	.dc.b	$FC,$F3,$F0,$8D,$DF,$92,$75,$1E,$43,$9D,$F5,$8C,$C7,$98,$F2,$CF
	.dc.b	$1A,$EB,$1D,$0E,$D3,$C2,$FE,$9C,$EA,$3E,$EC,$F2,$3D,$65,$3C,$25
	.dc.b	$79,$5F,$8E,$79,$1F,$62,$74,$E1,$FC,$14,$6D,$F6,$4F,$81,$F3,$1D
	.dc.b	$8A,$F6,$79,$9C,$27,$B3,$A7,$B3,$F0,$CC,$1B,$3C,$27,$23,$EC,$1F
	.dc.b	$B1,$9F,$23,$89,$F2,$1F,$1E,$FE,$64,$73,$35,$5E,$0B,$FB,$CD,$9D
	.dc.b	$CE,$8E,$13,$F4,$1F,$19,$E6,$34,$7A,$6B,$C0,$FD,$23,$51,$C5,$D5
	.dc.b	$77,$75,$3C,$1F,$94,$D3,$E5,$3C,$AB,$F8,$1B,$4F,$C2,$CF,$37,$8B
	.dc.b	$67,$89,$CD,$F8,$17,$FC,$8B,$E8,$65,$7F,$15,$7F,$D6,$9F,$43,$FD
	.dc.b	$A2,$3F,$81,$46,$9F,$E4,$0E,$AB,$FD,$75,$F4,$0F,$8C,$7C,$47,$F5
	.dc.b	$0B,$4E,$63,$0B,$B1,$F4,$38,$BE,$37,$FD,$63,$EF,$9E,$2B,$2E,$91
	.dc.b	$86,$C6,$67,$83,$5F,$E7,$A3,$67,$CF,$7F,$9A,$EB,$FC,$E6,$D9,$FF
	.dc.b	$1F,$CE,$B9,$3A,$32,$74,$DE,$5C,$DF,$F4,$A8,$EB,$18,$BF,$05,$7F
	.dc.b	$AC,$AB,$BC,$E6,$B7,$18,$3C,$77,$9A,$EF,$1D,$7E,$7D,$F2,$DC,$5B
	.dc.b	$1F,$E8,$CF,$33,$E8,$BF,$E8,$E8,$CE,$5E,$C6,$E3,$EC,$6C,$FA,$6B
	.dc.b	$8D,$6E,$FD,$4B,$E8,$5F,$43,$F3,$63,$E6,$C6,$2F,$F0,$0D,$8F,$F0
	.dc.b	$8B,$5F,$F0,$B3,$9C,$7E,$B3,$67,$E5,$65,$FE,$63,$7E,$61,$DA,$FE
	.dc.b	$07,$CA,$C8,$F7,$7B,$37,$1F,$74,$DC,$7D,$D3,$6B,$E5,$4D,$AF,$C2
	.dc.b	$68,$AC,$5F,$E1,$BE,$87,$E9,$7B,$9F,$4D,$B8,$FE,$21,$B8,$7E,$23
	.dc.b	$34,$62,$FF,$25,$F3,$CF,$73,$F0,$1F,$42,$FA,$1F,$9B,$39,$CB,$FD
	.dc.b	$8B,$39,$7E,$0E,$5F,$E1,$69,$D3,$FE,$8D,$6E,$1F,$6F,$E6,$F9,$E3
	.dc.b	$F8,$9F,$5F,$F1,$59,$5C,$7F,$3E,$DC,$7D,$0D,$C3,$F3,$CD,$9F,$A6
	.dc.b	$D9,$FC,$DB,$F9,$1F,$2E,$4F,$E3,$FC,$CB,$5F,$8F,$9B,$7F,$8F,$9B
	.dc.b	$5F,$B4,$CF,$FD,$2C,$FF,$B9,$3E,$89,$F4,$5F,$13,$D4,$FD,$34,$62
	.dc.b	$7E,$1E,$4B,$E2,$FC,$3A,$FD,$47,$B3,$C6,$BF,$45,$F3,$3F,$A7,$F8
	.dc.b	$AF,$D4,$36,$32,$7E,$9E,$32,$D3,$37,$FA,$98,$CB,$63,$FD,$F7,$EC
	.dc.b	$FD,$27,$F8,$12,$F5,$FF,$7C,$E6,$7F,$89,$39,$9E,$38,$EB,$D8,$DB
	.dc.b	$63,$F0,$3B,$BE,$8B,$F0,$98,$7F,$16,$6C,$7F,$16,$7F,$23,$3C,$1C
	.dc.b	$BF,$C5,$7E,$8B,$E3,$CF,$F1,$C6,$C6,$70,$FE,$29,$18,$BF,$E5,$1F
	.dc.b	$06,$DB,$F8,$C3,$6F,$CA,$B9,$73,$3A,$F2,$F7,$F4,$D6,$F6,$FF,$21
	.dc.b	$FA,$1E,$58,$FF,$30,$79,$4F,$C0,$73,$A7,$F3,$27,$33,$FC,$DB,$9B
	.dc.b	$F7,$15,$F1,$9F,$E3,$71,$FC,$FD,$CB,$BF,$13,$F9,$84,$7F,$AF,$E7
	.dc.b	$87,$B9,$8F,$73,$2F,$F3,$5F,$A0,$E3,$97,$F4,$26,$D7,$F3,$F7,$BC
	.dc.b	$1C,$E3,$FD,$3D,$7C,$5F,$33,$A7,$D0,$DB,$FB,$BE,$23,$E1,$3F,$A2
	.dc.b	$2F,$2E,$C6,$1F,$E6,$7D,$6F,$87,$A7,$F3,$6F,$A0,$E1,$87,$E9,$A7
	.dc.b	$4F,$EA,$21,$F1,$BB,$38,$7D,$0D,$9F,$D6,$FD,$0F,$CE,$E0,$FE,$D5
	.dc.b	$6A,$FE,$87,$39,$AF,$ED,$B3,$97,$8F,$C0,$FF,$48,$FA,$17,$87,$C2
	.dc.b	$30,$FF,$4C,$EA,$F8,$FC,$8C,$55,$9F,$3B,$86,$3F,$C0,$2F,$67,$0C
	.dc.b	$5F,$43,$B9,$FE,$0B,$F4,$1F,$2D,$FD,$4F,$D0,$BE,$07,$FA,$DC,$E7
	.dc.b	$4F,$EF,$F3,$9A,$E3,$CB,$FD,$D6,$31,$3F,$E0,$99,$FF,$05,$D7,$1E
	.dc.b	$1F,$F0,$2D,$69,$FF,$08,$46,$67,$77,$CF,$E5,$BC,$2F,$8B,$8F,$FC
	.dc.b	$05,$F4,$3F,$C1,$72,$61,$F5,$B0,$FE,$03,$39,$8C,$4F,$F8,$5A,$31
	.dc.b	$5C,$BD,$0F,$6F,$CD,$FF,$83,$7A,$5F,$21,$8C,$BF,$C0,$CD,$C7,$D8
	.dc.b	$7F,$B8,$B2,$5F,$C6,$8D,$6B,$FE,$2B,$39,$DF,$F3,$27,$FC,$65,$1C
	.dc.b	$8F,$F1,$75,$FF,$82,$23,$27,$17,$07,$F8,$32,$3F,$DE,$3F,$43,$FE
	.dc.b	$F2,$F9,$D7,$97,$FE,$38,$BE,$F5,$D4,$D8,$9D,$BF,$E3,$1B,$FF,$21
	.dc.b	$46,$23,$FC,$83,$7D,$6F,$A5,$69,$FF,$22,$CF,$4A,$E1,$7E,$07,$94
	.dc.b	$66,$78,$E1,$CE,$38,$E1,$5F,$83,$15,$E0,$79,$5C,$28,$E1,$7D,$EB
	.dc.b	$C7,$39,$3D,$CF,$81,$A7,$80,$A7,$80,$A7,$80,$C5,$FA,$9E,$87,$F9
	.dc.b	$D6,$BF,$CE,$CF,$CC,$30,$78,$BE,$79,$F4,$0F,$A1,$FF,$3E,$46,$1B
	.dc.b	$19,$9E,$0D,$7E,$6C,$BE,$87,$FC,$E5,$1D,$5A,$E9,$AF,$F9,$E2,$38
	.dc.b	$65,$E0,$9E,$38,$FD,$DC,$BF,$2F,$7F,$45,$AF,$C4,$BE,$2D,$AE,$33
	.dc.b	$FE,$84,$37,$5C,$F5,$FF,$43,$BF,$89,$BE,$2D,$AE,$36,$3F,$2F,$5C
	.dc.b	$2F,$A3,$63,$57,$CA,$7D,$95,$CA,$78,$9E,$9A,$E0,$AB,$83,$CA,$78
	.dc.b	$57,$A0,$EF,$1F,$91,$9D,$BD,$0F,$DA,$47,$C4,$C3,$FD,$48,$DD,$3C
	.dc.b	$71,$E1,$9F,$F1,$6C,$E5,$F7,$35,$F2,$5F,$A7,$E8,$7F,$86,$C7,$F8
	.dc.b	$D1,$AF,$F0,$36,$D7,$FA,$CD,$FC,$14,$6F,$6F,$DF,$61,$FE,$B6,$9D
	.dc.b	$B6,$33,$3F,$3E,$78,$BF,$C3,$FE,$87,$FC,$F0,$FF,$91,$1A,$E7,$B3
	.dc.b	$8A,$F9,$FF,$EB,$4E,$E3,$31,$FC,$03,$0F,$F6,$3B,$67,$EF,$57,$4C
	.dc.b	$7F,$D9,$8B,$C3,$EA,$D9,$F1,$47,$4F,$A1,$CB,$23,$27,$F9,$6A,$D3
	.dc.b	$98,$FE,$55,$1F,$DD,$9D,$73,$E7,$0F,$F6,$E4,$BA,$BD,$BF,$43,$BB
	.dc.b	$FE,$64,$DA,$DC,$EA,$E5,$7F,$9A,$AB,$59,$FC,$0A,$F8,$FE,$4B,$E1
	.dc.b	$A9,$2F,$86,$1A,$F2,$64,$7F,$9D,$29,$1F,$33,$AF,$9F,$93,$2B,$4F
	.dc.b	$07,$FA,$47,$D0,$FC,$DC,$FF,$DE,$EF,$F8,$22,$3F,$CF,$ED,$77,$C5
	.dc.b	$F8,$4D,$9F,$3D,$F8,$47,$FD,$EE,$DB,$FF,$7A,$36,$93,$C3,$AF,$97
	.dc.b	$D0,$FF,$AD,$67,$FC,$DE,$D7,$FB,$2C,$A3,$0A,$FF,$E0,$04,$7F,$9F
	.dc.b	$E7,$FC,$EF,$F4,$2F,$A3,$F3,$3F,$90,$8E,$6D,$BE,$51,$E4,$8E,$0F
	.dc.b	$FA,$03,$1F,$D1,$BE,$85,$F1,$BF,$05,$B1,$BE,$AF,$5F,$B5,$EC,$F9
	.dc.b	$89,$3E,$C8,$F9,$1F,$44,$F8,$FA,$49,$FF,$47,$C7,$E0,$5F,$17,$17
	.dc.b	$72,$3E,$5F,$A2,$F9,$3E,$54,$8F,$92,$FC,$CE,$5E,$FE,$EF,$83,$DE
	.dc.b	$95,$DA,$3D,$5F,$45,$EB,$F9,$F9,$3F,$06,$3E,$AF,$A1,$F5,$BE,$6A
	.dc.b	$4F,$D6,$8F,$5F,$0C,$8F,$4B,$EE,$E3,$FE,$03,$EE,$B0,$B8,$CB,$FF
	.dc.b	$65,$CE,$5B,$39,$7F,$C3,$90,$9E,$73,$27,$8C,$6F,$9B,$ED,$C2,$BA
	.dc.b	$3C,$BD,$6F,$19,$1F,$0F,$CE,$D7,$47,$92,$3D,$31,$FC,$6B,$E6,$1F
	.dc.b	$0F,$49,$5E,$A6,$BE,$8F,$13,$E0,$98,$C7,$CB,$F4,$4F,$88,$FE,$4A
	.dc.b	$DA,$FF,$76,$B5,$FF,$1E,$3F,$C8,$D3,$99,$E7,$3C,$8E,$7E,$7D,$AE
	.dc.b	$71,$5E,$71,$E0,$E1,$F0,$F1,$49,$F0,$C6,$5F,$1C,$65,$F1,$C7,$C3
	.dc.b	$DE,$47,$F7,$3C,$3E,$8E,$9F,$5E,$3D,$7F,$13,$A7,$FC,$AD,$F4,$5F
	.dc.b	$E5,$BF,$0F,$FB,$D4,$23,$0F,$F7,$DA,$E4,$F0,$F8,$4F,$E9,$72,$69
	.dc.b	$8E,$3B,$3E,$3B,$EB,$5B,$8C,$1F,$E0,$97,$0F,$92,$78,$BD,$CD,$9E
	.dc.b	$5A,$EC,$E2,$5F,$1D,$E4,$F8,$69,$39,$8F,$39,$87,$E0,$C6,$1E,$BF
	.dc.b	$44,$FA,$2F,$EC,$ED,$8C,$C7,$F6,$AF,$A2,$7C,$66,$A3,$1C,$12,$B6
	.dc.b	$7A,$2B,$FD,$76,$DE,$8E,$47,$3E,$99,$19,$21,$FE,$8B,$9E,$2D,$8E
	.dc.b	$2E,$B8,$BF,$5F,$3A,$6F,$89,$E7,$24,$F8,$FA,$6B,$D6,$2C,$F2,$9F
	.dc.b	$F0,$3F,$D1,$3C,$7A,$4B,$FA,$7F,$45,$FC,$E7,$97,$F1,$35,$B9,$FF
	.dc.b	$3C,$87,$0F,$F9,$F1,$AF,$FA,$41,$AF,$E4,$A7,$FD,$AC,$F6,$E8,$FF
	.dc.b	$0C,$C6,$23,$FD,$5C,$BA,$73,$32,$78,$1E,$68,$C3,$C2,$2E,$5F,$C5
	.dc.b	$AB,$94,$F0,$34,$58,$D5,$70,$C3,$CD,$76,$BF,$CC,$57,$C9,$E2,$4B
	.dc.b	$ED,$3E,$AB,$E1,$2A,$F6,$56,$A3,$D0,$78,$6F,$BB,$4E,$F5,$CA,$B8
	.dc.b	$CE,$A7,$E2,$CE,$6F,$4D,$8E,$47,$85,$C3,$E7,$9F,$A5,$18,$BF,$24
	.dc.b	$F4,$38,$DF,$81,$F8,$37,$ED,$72,$F8,$1A,$70,$AE,$97,$E7,$BF,$9B
	.dc.b	$3A,$3C,$47,$B8,$C3,$E2,$36,$E8,$F9,$55,$DD,$C5,$F9,$5C,$9F,$30
	.dc.b	$CC,$6E,$7A,$57,$8C,$93,$E1,$36,$7A,$4F,$1B,$A7,$E5,$18,$3E,$19
	.dc.b	$DE,$3A,$B8,$3C,$B7,$F0,$9C,$47,$B0,$DB,$E8,$36,$75,$30,$EE,$7C
	.dc.b	$57,$EF,$25,$75,$AD,$CF,$DC,$79,$4E,$9B,$7F,$65,$F7,$9E,$02,$DF
	.dc.b	$92,$7F,$3C,$D9,$F9,$55,$E2,$6D,$7B,$23,$0F,$C3,$6B,$D6,$79,$3F
	.dc.b	$34,$EF,$1E,$CB,$F5,$5F,$9E,$72,$75,$3E,$B5,$72,$7A,$57,$68,$F4
	.dc.b	$57,$2B,$F7,$46,$CC,$D7,$C2,$3C,$C7,$53,$E4,$96,$FE,$0D,$FF,$17
	.dc.b	$FA,$1F,$E2,$A4,$9F,$EA,$FF,$45,$FD,$A9,$FD,$F4,$65,$FE,$32,$FF
	.dc.b	$AD,$23,$5F,$3F,$FE,$BC,$9F,$EE,$61,$C3,$FD,$D9,$AF,$FB,$C5,$AB
	.dc.b	$E8,$9F,$42,$F8,$BF,$E6,$71,$86,$F7,$3E,$8B,$35,$BB,$F0,$EA,$F9
	.dc.b	$EF,$F3,$36,$BF,$DE,$23,$1C,$9F,$E7,$86,$4F,$FA,$9E,$74,$F0,$AF
	.dc.b	$F4,$4B,$67,$75,$C6,$BF,$C3,$13,$C6,$B7,$39,$7F,$D2,$F1,$C6,$77
	.dc.b	$39,$AF,$37,$25,$AD,$5F,$FA,$7B,$E7,$DF,$2B,$B6,$E3,$EB,$F3,$7E
	.dc.b	$33,$ED,$8C,$D7,$FA,$81,$B8,$FB,$DB,$3F,$E2,$FA,$E3,$5B,$BF,$42
	.dc.b	$FA,$27,$D1,$79,$FE,$73,$FD,$B1,$87,$F8,$72,$72,$FC,$D8,$CD,$7F
	.dc.b	$AD,$A3,$FC,$6B,$3A,$8C,$03,$B5,$7C,$1F,$2F,$C3,$3B,$CF,$D1,$F4
	.dc.b	$5F,$04,$F8,$5E,$E5,$C7,$E0,$45,$C7,$E0,$4D,$AE,$35,$A3,$E3,$46
	.dc.b	$5F,$B3,$F4,$5F,$47,$DF,$FA,$4D,$8F,$C1,$B6,$FF,$06,$DB,$FF,$43
	.dc.b	$FC,$F3,$E4,$7E,$E7,$D1,$3E,$87,$E5,$CE,$71,$FC,$94,$E5,$FF,$22
	.dc.b	$C6,$71,$F8,$A1,$C6,$1F,$67,$E8,$B9,$E3,$F8,$7F,$67,$F1,$19,$5C
	.dc.b	$7F,$32,$DC,$7F,$32,$DC,$3F,$3E,$D9,$FC,$CB,$67,$F2,$EF,$E3,$FC
	.dc.b	$E9,$3F,$8D,$F3,$AD,$7E,$36,$6D,$FE,$36,$6D,$7F,$9D,$33,$FF,$4B
	.dc.b	$3F,$85,$FA,$2F,$87,$EA,$FE,$2D,$B1,$F0,$A3,$F4,$5E,$D6,$32,$FC
	.dc.b	$D7,$CD,$FE,$87,$E3,$7F,$44,$DB,$FD,$03,$67,$71,$FA,$39,$DB,$4D
	.dc.b	$96,$3F,$27,$9F,$FA,$D3,$7F,$DA,$73,$BD,$7B,$18,$73,$9E,$BD,$8B
	.dc.b	$8F,$F0,$DB,$47,$EF,$7E,$8B,$EF,$65,$D9,$B9,$7F,$11,$5C,$DB,$E3
	.dc.b	$FE,$09,$F4,$5F,$22,$FC,$AD,$BF,$E3,$0F,$6C,$BF,$8E,$39,$7F,$D5
	.dc.b	$5F,$06,$E1,$FE,$1C,$D8,$FE,$44,$B9,$B3,$3A,$F2,$37,$EB,$DB,$DB
	.dc.b	$FC,$5F,$E8,$B9,$67,$F2,$1E,$53,$FC,$B5,$CE,$7F,$CB,$5C,$CF,$CE
	.dc.b	$E8,$76,$FF,$21,$7F,$89,$C2,$91,$AB,$FE,$70,$EA,$33,$ED,$F1,$39
	.dc.b	$7B,$96,$BF,$9F,$7D,$17,$1C,$3E,$EB,$6F,$FA,$03,$CC,$C4,$FF,$B4
	.dc.b	$57,$C6,$7F,$35,$6C,$74,$FA,$2C,$4F,$F3,$9F,$09,$F0,$97,$C9,$C3
	.dc.b	$6F,$FD,$83,$DC,$F8,$6F,$E5,$AD,$8F,$F6,$2F,$D0,$AF,$35,$FC,$CA
	.dc.b	$1F,$1F,$4C,$7D,$15,$BF,$F6,$6F,$D1,$7C,$ED,$D1,$FE,$E4,$B5,$7F
	.dc.b	$3C,$AD,$4E,$63,$1F,$01,$FD,$03,$E8,$9E,$36,$2B,$FD,$B3,$18,$7F
	.dc.b	$DB,$7D,$5F,$23,$93,$F9,$D6,$2C,$3F,$A4,$CE,$65,$39,$AE,$3D,$1A
	.dc.b	$3C,$1F,$45,$63,$FD,$CF,$F4,$1F,$31,$FE,$F1,$F4,$4F,$80,$FE,$AB
	.dc.b	$39,$C7,$F7,$33,$9A,$E3,$CB,$FD,$B6,$31,$3F,$CB,$2B,$87,$0F,$F8
	.dc.b	$06,$72,$FC,$89,$FE,$A9,$3A,$BF,$EA,$D0,$8C,$3F,$97,$AF,$EB,$F5
	.dc.b	$C6,$35,$7E,$6E,$2F,$EF,$DF,$45,$FC,$EB,$1E,$78,$FD,$C7,$FC,$29
	.dc.b	$19,$BF,$0D,$70,$F2,$F9,$56,$9F,$83,$F3,$FF,$E0,$9E,$77,$C7,$E2
	.dc.b	$BF,$C0,$8D,$C3,$E0,$57,$C8,$71,$A7,$9C,$E3,$5F,$DC,$DC,$7C,$7F
	.dc.b	$F8,$09,$CC,$FF,$54,$51,$9B,$95,$C4,$FF,$20,$D3,$E8,$BF,$BF,$7D
	.dc.b	$0F,$F2,$48,$ED,$5F,$E0,$E7,$FC,$05,$3A,$F3,$7E,$7E,$3E,$0B,$E1
	.dc.b	$5F,$4A,$B4,$F2,$C2,$B9,$13,$C2,$B9,$5F,$92,$F7,$18,$AF,$F2,$BC
	.dc.b	$63,$0F,$C1,$5F,$82,$7C,$A7,$4B,$F2,$CE,$9E,$91,$93,$C8,$53,$C9
	.dc.b	$3E,$39,$E1,$1E,$03,$57,$E3,$77,$5D,$B0,$F6,$C7,$88,$F9,$A7,$CD
	.dc.b	$29,$4C,$E1,$AF,$A1,$7C,$CB,$E8,$B0,$DA,$FA,$F3,$A9,$C9,$C7,$E8
	.dc.b	$BF,$9A,$4E,$9A,$FF,$9F,$A5,$18,$8F,$D3,$C6,$32,$E4,$7D,$6C,$9B
	.dc.b	$7E,$B3,$5F,$BB,$18,$BF,$B7,$7C,$5F,$94,$D9,$DD,$71,$AF,$DA,$9C
	.dc.b	$6B,$78,$7A,$2B,$F1,$37,$C6,$B7,$7C,$6F,$FC,$A5,$1C,$6B,$77,$C6
	.dc.b	$3D,$6D,$9D,$D7,$17,$FB,$FB,$6F,$B4,$7F,$97,$5A,$E8,$F8,$51,$B9
	.dc.b	$E9,$19,$3F,$C8,$35,$EB,$C3,$F8,$63,$74,$F1,$C7,$29,$FD,$81,$E4
	.dc.b	$29,$CD,$FE,$29,$87,$8B,$E8,$BF,$83,$CE,$9A,$FF,$94,$1B,$3F,$C6
	.dc.b	$DF,$BB,$1B,$DB,$F6,$B7,$FC,$92,$76,$D8,$CC,$FF,$9F,$67,$8B,$FE
	.dc.b	$5C,$FA,$2F,$F0,$0B,$E6,$6B,$9E,$AE,$2B,$E8,$7F,$AC,$BB,$8C,$C7
	.dc.b	$EF,$70,$FE,$6A,$D9,$FF,$46,$57,$4C,$7F,$9A,$2F,$13,$AB,$67,$C5
	.dc.b	$1D,$3E,$8B,$2C,$8C,$9F,$32,$D3,$98,$FF,$3B,$C7,$F7,$67,$5C,$F9
	.dc.b	$C3,$FA,$24,$BE,$87,$DB,$F4,$5B,$BF,$F3,$FB,$6B,$73,$AB,$95,$F7
	.dc.b	$2B,$59,$FC,$0A,$F8,$FE,$4B,$E1,$E9,$2F,$86,$1A,$F2,$A4,$7E,$1A
	.dc.b	$91,$F3,$3A,$F9,$F9,$52,$B4,$F0,$7F,$D2,$BF,$45,$F3,$73,$FE,$CA
	.dc.b	$FF,$82,$23,$F2,$2D,$77,$C5,$F7,$3E,$0B,$F0,$7B,$FF,$BC,$7E,$7F
	.dc.b	$FC,$A5,$5F,$EF,$56,$BF,$EA,$29,$DE,$1F,$DC,$BE,$E1,$EF,$7E,$1B
	.dc.b	$A1,$E9,$D3,$F1,$33,$AC,$3F,$47,$F7,$1F,$EB,$37,$F4,$B5,$FE,$4E
	.dc.b	$6F,$AB,$F9,$09,$D4,$78,$67,$F3,$7A,$34,$FF,$36,$4F,$9F,$D8,$F6
	.dc.b	$57,$E6,$3E,$83,$FD,$EF,$3F,$B2,$FB,$8F,$D2,$BF,$ED,$F6,$A8,$DB
	.dc.b	$FE,$C0,$5E,$C7,$E8,$A7,$FD,$0B,$3C,$FD,$FE,$D3,$F2,$F9,$3C,$B9
	.dc.b	$3E,$6E,$99,$BF,$E0,$7F,$70,$F9,$8F,$D6,$FD,$BF,$C5,$9F,$F6,$7B
	.dc.b	$77,$FF,$0D,$56,$E7,$F8,$5F,$DC,$3E,$4F,$8F,$AB,$27,$31,$F7,$7E
	.dc.b	$93,$C7,$ED,$49,$FE,$27,$D6,$9F,$DE,$E0,$FA,$0F,$DF,$FD,$BB,$E6
	.dc.b	$69,$A8,$FE,$07,$F7,$0F,$73,$7F,$29,$33,$FE,$3F,$39,$7F,$90,$FE
	.dc.b	$4F,$FA,$64,$1C,$37,$C8,$FE,$1F,$FA,$1F,$F0,$38,$FF,$78,$CE,$A3
	.dc.b	$F0,$D1,$FE,$F4,$6F,$1F,$F8,$FE,$79,$9B,$F0,$39,$9F,$C6,$9E,$FD
	.dc.b	$5F,$06,$FF,$9A,$BF,$05,$BA,$7F,$20,$8D,$47,$FA,$8F,$C4,$FF,$28
	.dc.b	$C6,$30,$FF,$27,$3C,$67,$47,$B9,$C4,$6B,$87,$A7,$1C,$AF,$0B,$DF
	.dc.b	$A7,$11,$C8,$FC,$8C,$62,$BB,$CF,$F3,$4E,$17,$C7,$D2,$5F,$B1,$AF
	.dc.b	$3C,$7F,$8A,$3F,$03,$CB,$CE,$1F,$E6,$7B,$E7,$39,$21,$E3,$72,$F9
	.dc.b	$7E,$A3,$F2,$DA,$7F,$9D,$1B,$1F,$79,$A9,$FE,$88,$DF,$1F,$FA,$6C
	.dc.b	$68,$9E,$AE,$3A,$F3,$F4,$52,$FF,$B5,$D7,$FA,$45,$B3,$97,$C9,$3C
	.dc.b	$5F,$EE,$38,$FD,$2C,$7F,$D1,$2D,$8F,$C9,$B5,$F3,$7C,$97,$CD,$E2
	.dc.b	$FE,$C3,$F5,$1F,$E7,$5B,$FF,$48,$B6,$B8,$CE,$DF,$EF,$F1,$94,$DB
	.dc.b	$EC,$FF,$26,$8D,$F9,$7D,$8E,$FC,$16,$32,$43,$F0,$93,$FC,$51,$EF
	.dc.b	$F2,$DE,$6E,$79,$18,$21,$C1,$CB,$CA,$3F,$BE,$FD,$43,$EA,$29,$27
	.dc.b	$FC,$33,$F5,$0F,$92,$C4,$7E,$56,$3E,$41,$F2,$6B,$94,$7F,$88,$1A
	.dc.b	$EF,$5F,$F0,$5C,$EE,$EF,$4E,$3A,$65,$F8,$23,$2F,$8A,$32,$F8,$63
	.dc.b	$9F,$89,$23,$CE,$F6,$E9,$C4,$79,$E2,$F4,$FF,$B1,$71,$FD,$1C,$FE
	.dc.b	$BF,$E9,$DF,$51,$FE,$20,$F2,$7F,$2E,$7E,$62,$33,$3F,$E3,$C6,$BF
	.dc.b	$B7,$8D,$F5,$FF,$8A,$27,$38,$7F,$B4,$DB,$CF,$FE,$DA,$AD,$DF,$3C
	.dc.b	$38,$1E,$1C,$24,$64,$87,$77,$2F,$A6,$38,$BF,$E5,$09,$EA,$E4,$E3
	.dc.b	$87,$FB,$96,$32,$F9,$67,$8C,$F5,$39,$56,$27,$51,$82,$1D,$48,$F5
	.dc.b	$8F,$2B,$C2,$33,$7F,$06,$72,$FF,$BF,$30,$F8,$B3,$FD,$19,$AF,$F6
	.dc.b	$79,$DB,$9B,$F2,$57,$38,$F1,$DF,$72,$46,$63,$D8,$78,$8E,$47,$07
	.dc.b	$51,$C2,$32,$70,$7D,$8F,$CA,$8F,$44,$F9,$23,$CC,$64,$E3,$3C,$1E
	.dc.b	$53,$C1,$D1,$C0,$C9,$C1,$D3,$83,$51,$E5,$57,$F1,$B0,$FB,$75,$C2
	.dc.b	$B5,$18,$AE,$84,$39,$CF,$C0,$38,$11,$FE,$EF,$1C,$5E,$F7,$F4,$1C
	.dc.b	$E1,$F0,$70,$FA,$E7,$8D,$C4,$EE,$7B,$9C,$0F,$83,$3B,$7E,$91,$E5
	.dc.b	$70,$FC,$17,$AC,$7C,$FA,$FD,$A3,$D5,$F3,$46,$A7,$99,$E0,$3D,$27
	.dc.b	$D0,$8D,$CF,$1A,$F3,$CE,$5A,$FE,$1C,$E5,$3E,$07,$47,$6A,$F0,$3F
	.dc.b	$5E,$34,$F1,$7B,$BC,$0F,$65,$78,$5E,$4D,$72,$6F,$0F,$C5,$1F,$81
	.dc.b	$AF,$8A,$E3,$0F,$B1,$3E,$E3,$C5,$87,$D1,$AD,$CF,$DF,$9F,$9D,$7F
	.dc.b	$2E,$B4,$D8,$FA,$AD,$7E,$15,$72,$7E,$D1,$ED,$EA,$FE,$50,$47,$F8
	.dc.b	$97,$55,$7F,$93,$45,$C7,$F8,$94,$FF,$2F,$FA,$89,$5F,$C6,$5F,$F6
	.dc.b	$8F,$D4,$7F,$6B,$D1,$1A,$9F,$E6,$0E,$B4,$FF,$6B,$FD,$3B,$E9,$9F
	.dc.b	$23,$FE,$50,$B8,$27,$51,$88,$FF,$2A,$C7,$F9,$87,$E9,$F1,$1B,$AE
	.dc.b	$6E,$FE,$A1,$F1,$AC,$77,$18,$6C,$66,$78,$35,$F1,$CF,$F9,$C6,$78
	.dc.b	$F4,$E6,$7D,$13,$97,$C1,$19,$5B,$BE,$A3,$FD,$09,$7F,$E6,$E8,$F0
	.dc.b	$19,$E6,$B1,$99,$FF,$6B,$3B,$31,$1A,$AE,$B3,$D1,$FF,$2E,$D7,$3A
	.dc.b	$E1,$1A,$E7,$E1,$39,$8D,$FD,$43,$EA,$32,$DB,$E3,$4F,$80,$FA,$F1
	.dc.b	$89,$D5,$F2,$BE,$27,$6A,$DC,$66,$BC,$6B,$E9,$DE,$87,$93,$EA,$3F
	.dc.b	$E0,$05,$6A,$72,$F8,$A3,$35,$FE,$71,$9C,$B6,$BD,$BA,$7F,$9C,$2F
	.dc.b	$94,$7C,$79,$F0,$36,$B8,$E7,$F4,$AF,$85,$73,$BF,$77,$AB,$F1,$0F
	.dc.b	$8B,$F0,$36,$32,$49,$F9,$71,$97,$E9,$D3,$DB,$F2,$FC,$1F,$99,$B7
	.dc.b	$F3,$A7,$4D,$9D,$34,$F9,$3F,$4C,$F9,$4F,$B7,$F5,$0F,$A5,$F7,$4E
	.dc.b	$6B,$EC,$CE,$5E,$35,$A3,$EF,$FD,$45,$8C,$93,$3F,$C2,$C6,$5F,$5F
	.dc.b	$D2,$7C,$EC,$7E,$7B,$6B,$D4,$D9,$DC,$7B,$67,$F1,$BE,$6C,$8F,$C5
	.dc.b	$79,$D6,$BF,$27,$37,$87,$F4,$BC,$EA,$C5,$FF,$07,$D5,$EA,$FE,$27
	.dc.b	$EA,$3E,$0B,$67,$E3,$B8,$7E,$77,$C3,$DC,$FF,$D7,$57,$CA,$76,$68
	.dc.b	$B1,$FE,$BF,$E1,$72,$FE,$2D,$B3,$FA,$6F,$63,$F9,$0B,$F9,$E8,$CD
	.dc.b	$7F,$B1,$FC,$27,$D3,$7E,$4D,$B5,$C7,$EB,$F1,$39,$D2,$4F,$EC,$BE
	.dc.b	$A5,$8E,$31,$FF,$69,$77,$3E,$63,$F3,$CD,$BF,$DC,$FD,$4B,$3C,$D7
	.dc.b	$E8,$FE,$D1,$EA,$D3,$F8,$2E,$BF,$87,$CB,$FD,$BD,$3A,$C7,$F7,$3F
	.dc.b	$3D,$FC,$03,$9F,$F9,$33,$63,$F8,$E5,$BE,$03,$F6,$13,$9C,$03,$F0
	.dc.b	$7F,$8E,$FF,$BA,$DB,$87,$F2,$4F,$AB,$F1,$3F,$CC,$E3,$38,$FF,$47
	.dc.b	$FB,$5B,$7D,$1B,$4B,$FE,$22,$FF,$3B,$8D,$C9,$7A,$5F,$C1,$5B,$7F
	.dc.b	$C8,$3A,$DF,$65,$9A,$FE,$15,$F8,$93,$BE,$9F,$CB,$3F,$16,$EF,$87
	.dc.b	$F5,$B9,$D3,$FC,$80,$FD,$EE,$5F,$D1,$1F,$EC,$06,$33,$FE,$87,$F5
	.dc.b	$EF,$84,$FE,$31,$39,$37,$E8,$E0,$FE,$77,$D5,$F8,$AC,$3A,$1F,$3A
	.dc.b	$7F,$AA,$61,$FD,$6F,$E7,$9F,$27,$FC,$8D,$B3,$FC,$D3,$EB,$BF,$9F
	.dc.b	$3F,$CD,$A3,$37,$D0,$4B,$E3,$7F,$96,$36,$B9,$7D,$76,$27,$F9,$87
	.dc.b	$D6,$3C,$BD,$CE,$1B,$5F,$D1,$25,$F0,$BF,$CD,$FE,$B5,$F5,$DC,$A3
	.dc.b	$0F,$F3,$BF,$7D,$EA,$E7,$0F,$E9,$B1,$87,$A5,$6E,$64,$7F,$52,$B7
	.dc.b	$C9,$E3,$6F,$77,$D6,$3E,$B7,$FB,$3F,$D7,$3E,$1F,$FA,$54,$E7,$1F
	.dc.b	$EC,$B3,$97,$FA,$EC,$66,$FF,$B9,$CE,$5E,$B7,$D2,$F9,$B2,$F9,$FD
	.dc.b	$1D,$AF,$1F,$17,$F7,$8F,$AE,$FE,$79,$18,$AF,$EE,$B1,$8E,$7F,$F0
	.dc.b	$C6,$7F,$4E,$BF,$C0,$73,$9A,$FF,$17,$4E,$6B,$FC,$23,$39,$AF,$DA
	.dc.b	$CF,$F0,$97,$FC,$17,$E6,$BE,$83,$E0,$F9,$9F,$DF,$5B,$8F,$C0,$8F
	.dc.b	$8B,$58,$3F,$11,$8F,$19,$FF,$00,$BF,$E3,$18,$CD,$7F,$75,$8C,$1F
	.dc.b	$E3,$13,$84,$6A,$7F,$C1,$2F,$F0,$AF,$23,$F8,$5F,$D6,$3E,$3B,$FC
	.dc.b	$95,$7C,$0F,$F1,$94,$EA,$33,$87,$F8,$FE,$B8,$5F,$F9,$42,$B4,$FF
	.dc.b	$8D,$67,$55,$FE,$51,$9C,$CF,$F8,$F2,$72,$F1,$15,$D1,$D6,$1C,$1F
	.dc.b	$41,$AB,$F1,$C6,$61,$7D,$B0,$F8,$37,$CE,$78,$BD,$63,$37,$E5,$9E
	.dc.b	$87,$A6,$FA,$5F,$3C,$3C,$F3,$FE,$6F,$8C,$4E,$A3,$FC,$D6,$78,$8B
	.dc.b	$18,$70,$7F,$9D,$BE,$AD,$F5,$CF,$AE,$FF,$39,$C6,$1B,$19,$9F,$F3
	.dc.b	$D4,$E5,$BF,$5D,$FE,$66,$6E,$7C,$A3,$11,$DD,$AF,$FA,$1E,$51,$C7
	.dc.b	$4F,$A1,$A7,$F6,$3A,$FA,$ED,$7E,$28,$AE,$37,$F7,$1B,$1A,$9E,$15
	.dc.b	$D3,$5F,$F4,$B3,$F7,$6B,$6D,$C3,$C5,$5F,$E9,$B3,$89,$6B,$8B,$E9
	.dc.b	$8D,$DF,$F6,$4B,$E7,$3B,$8F,$4D,$70,$75,$5C,$1F,$53,$67,$BC,$7F
	.dc.b	$6E,$6B,$AC,$39,$BF,$E7,$28,$F8,$91,$FE,$9F,$6E,$9E,$58,$DC,$FF
	.dc.b	$41,$9C,$BD,$6F,$F5,$F5,$B8,$F9,$BF,$5D,$FC,$1D,$FF,$3F,$B5,$FE
	.dc.b	$94,$DB,$FF,$57,$3F,$6A,$37,$B7,$47,$FC,$0F,$1B,$AF,$F6,$33,$63
	.dc.b	$38,$ED,$7D,$77,$F4,$47,$FD,$1E,$D7,$FA,$D3,$70,$FF,$5F,$7D,$6B
	.dc.b	$C1,$E5,$39,$C7,$FD,$A9,$3F,$AF,$6E,$9A,$EC,$DF,$87,$32,$EE,$F8
	.dc.b	$ED,$F5,$D9,$9F,$0C,$DC,$FC,$33,$71,$EF,$D1,$C4,$C5,$FF,$8A,$30
	.dc.b	$74,$FB,$FE,$BB,$FB,$94,$FF,$AB,$1A,$FF,$77,$9D,$B9,$AE,$04,$7F
	.dc.b	$D6,$71,$A7,$FB,$F4,$F2,$C7,$FD,$C5,$F5,$CF,$8E,$F9,$CD,$AE,$87
	.dc.b	$2E,$AC,$F9,$32,$30,$7F,$B0,$27,$FC,$22,$DF,$27,$99,$B7,$FC,$1D
	.dc.b	$F5,$DF,$5D,$FF,$23,$9F,$E4,$87,$0E,$1A,$E7,$87,$F8,$CF,$C0,$FE
	.dc.b	$2E,$38,$9F,$E5,$07,$0E,$1A,$F6,$6B,$FE,$56,$FA,$EF,$F4,$6B,$C8
	.dc.b	$C7,$C0,$7F,$26,$8C,$C6,$5A,$F8,$3E,$BB,$FD,$68,$F1,$0E,$1E,$8D
	.dc.b	$78,$7C,$47,$AB,$B2,$47,$1C,$3F,$54,$FA,$63,$F9,$BF,$D7,$7F,$BE
	.dc.b	$63,$B3,$5F,$43,$5F,$86,$D7,$E6,$AF,$A2,$E6,$48,$E6,$FC,$CD,$64
	.dc.b	$78,$DA,$FC,$5F,$AB,$FF,$35,$BE,$F6,$B9,$F6,$7F,$A6,$C6,$23,$EA
	.dc.b	$FD,$5B,$E5,$BF,$A8,$B6,$3E,$43,$5F,$A4,$D7,$ED,$7C,$0F,$E0,$08
	.dc.b	$FC,$77,$9F,$AF,$23,$9B,$5F,$B9,$F5,$5F,$E4,$27,$EB,$35,$CF,$A3
	.dc.b	$FD,$92,$31,$1F,$8C,$FA,$9F,$D3,$78,$1F,$96,$6C,$7F,$A7,$9A,$BE
	.dc.b	$A9,$E1,$FF,$17,$9F,$DB,$CF,$3F,$07,$FB,$8C,$62,$3A,$46,$0F,$E3
	.dc.b	$D5,$FE,$E3,$8F,$F5,$5F,$7B,$0F,$D3,$CE,$67,$FC,$F1,$7E,$2E,$2F
	.dc.b	$F5,$A4,$70,$AF,$D6,$C3,$4F,$E5,$4D,$9D,$1F,$EB,$87,$8B,$FE,$83
	.dc.b	$6E,$1E,$18,$D4,$F2,$39,$EB,$D3,$D5,$93,$D0,$FC,$83,$E0,$F8,$7C
	.dc.b	$C7,$58,$B7,$D6,$3F,$2D,$3F,$CF,$FE,$A1,$F1,$B4,$93,$D7,$2C,$BD
	.dc.b	$E6,$BF,$BB,$6B,$FC,$19,$AF,$F1,$06,$BF,$ED,$87,$8D,$F8,$E7,$F9
	.dc.b	$14,$FC,$3F,$0F,$C2,$FF,$A9,$FE,$AB,$F8,$2D,$7A,$5B,$AF,$F1,$C6
	.dc.b	$CF,$FB,$91,$FE,$9C,$DE,$FF,$EB,$91,$A8,$34,$DB,$8F,$5F,$7F,$45
	.dc.b	$23,$E7,$4F,$FA,$65,$B3,$97,$CD,$3D,$A3,$E6,$70,$49,$F9,$0F,$8B
	.dc.b	$EA,$B1,$1F,$26,$3F,$C8,$FA,$FE,$1E,$30,$7F,$1F,$71,$9F,$D9,$FA
	.dc.b	$AF,$F3,$E6,$9F,$CF,$1B,$5F,$36,$FF,$BD,$47,$FB,$86,$3F,$0C,$D7
	.dc.b	$FD,$A4,$DA,$E9,$B7,$0F,$82,$95,$C1,$E7,$1D,$3E,$1B,$AF,$C4,$4A
	.dc.b	$EB,$19,$7B,$C6,$5E,$F1,$FE,$0D,$FA,$97,$D4,$D2,$4F,$E6,$FE,$A5
	.dc.b	$F5,$58,$8F,$E9,$15,$C1,$FE,$84,$D7,$35,$FE,$70,$8C,$67,$F9,$FB
	.dc.b	$F9,$3B,$7F,$62,$6D,$85,$70,$9D,$74,$7F,$88,$E9,$D1,$4C,$44,$BF
	.dc.b	$44,$65,$F5,$47,$1F,$06,$47,$B5,$E9,$D1,$88,$65,$E7,$9C,$3E,$99
	.dc.b	$FF,$1A,$7D,$4B,$EA,$A9,$2F,$F8,$17,$8B,$FE,$31,$7F,$82,$4E,$DB
	.dc.b	$3B,$7B,$E9,$FE,$04,$AE,$11,$83,$E2,$57,$17,$01,$C6,$1F,$C2,$DB
	.dc.b	$5C,$FD,$8E,$D3,$9C,$2C,$E6,$FC,$64,$3A,$DF,$27,$EA,$61,$E4,$BE
	.dc.b	$51,$B6,$BE,$48,$D5,$7F,$87,$AB,$94,$F9,$EF,$C2,$F4,$C2,$B9,$3A
	.dc.b	$57,$95,$F1,$CE,$AB,$9D,$F6,$9F,$8B,$39,$3F,$CD,$11,$98,$E8,$72
	.dc.b	$3A,$9E,$12,$9A,$C3,$D6,$7F,$35,$8E,$55,$EA,$25,$76,$C3,$C1,$7C
	.dc.b	$AB,$73,$88,$F8,$93,$6F,$AE,$1E,$D6,$CF,$A8,$E5,$3B,$9F,$95,$5D
	.dc.b	$E7,$CF,$87,$89,$FC,$7D,$6D,$E4,$78,$0D,$CF,$CE,$C3,$CB,$18,$38
	.dc.b	$94,$F8,$C7,$12,$E1,$E6,$7E,$36,$1E,$93,$CA,$FF,$90,$F0,$FC,$B3
	.dc.b	$FE,$56,$6B,$A3,$FC,$A0,$E1,$F3,$9B,$BF,$F7,$BC,$66,$3B,$9F,$92
	.dc.b	$F5,$FE,$E9,$EF,$58,$7F,$48,$3F,$A4,$CF,$C5,$8F,$CA,$4F,$07,$DD
	.dc.b	$87,$DD,$8C,$47,$BB,$0F,$34,$E6,$3F,$77,$3D,$EF,$B3,$4E,$33,$F1
	.dc.b	$6F,$F2,$B1,$8B,$E6,$4B,$F5,$1D,$0F,$1B,$FC,$06,$38,$47,$DA,$AF
	.dc.b	$A1,$7E,$08,$D9,$C8,$F6,$98,$3E,$1C,$F2,$7E,$49,$23,$93,$EF,$3B
	.dc.b	$57,$85,$F2,$19,$38,$4F,$3B,$F9,$97,$F1,$67,$C2,$6E,$3E,$0B,$CA
	.dc.b	$3E,$26,$1E,$3C,$3E,$AB,$F9,$3B,$E8,$DA,$E1,$87,$C7,$7C,$26,$0F
	.dc.b	$39,$0F,$79,$93,$DF,$87,$89,$F2,$E1,$F2,$BA,$BF,$82,$92,$3F,$9B
	.dc.b	$F5,$D7,$F8,$D4,$59,$FE,$05,$3F,$C3,$3E,$A6,$57,$F3,$E7,$FD,$ED
	.dc.b	$F5,$5F,$E0,$08,$FC,$9C,$69,$FE,$48,$EB,$1F,$F7,$E7,$D3,$FF,$AD
	.dc.b	$24,$D2,$7B,$9F,$EA,$D8,$7D,$53,$EA,$9F,$17,$FE,$B1,$8C,$37,$C1
	.dc.b	$7C,$E6,$6B,$77,$FE,$BF,$FB,$FE,$39,$FF,$AE,$5B,$19,$C3,$FC,$A8
	.dc.b	$FF,$B5,$A3,$2F,$FB,$63,$77,$D2,$FF,$A7,$6F,$FC,$D7,$18,$E8,$FF
	.dc.b	$34,$32,$7F,$DD,$D3,$A7,$85,$7F,$5D,$9E,$13,$9A,$E2,$FF,$A0,$AB
	.dc.b	$8D,$6D,$E3,$5D,$BE,$89,$F5,$5C,$2B,$FB,$CC,$62,$FC,$BE,$FF,$BA
	.dc.b	$F8,$5E,$D5,$6E,$FC,$8D,$8F,$5D,$71,$BF,$F4,$AC,$E5,$FF,$7D,$FD
	.dc.b	$43,$EA,$7A,$7D,$53,$E1,$7C,$99,$2B,$57,$F2,$2B,$4F,$FA,$26,$33
	.dc.b	$3F,$EF,$56,$BF,$F0,$01,$AF,$82,$FE,$BB,$FE,$29,$8E,$39,$7F,$A3
	.dc.b	$5B,$5F,$2F,$CF,$F1,$C6,$2B,$E3,$FB,$9F,$0F,$33,$7E,$03,$0C,$C3
	.dc.b	$EF,$CE,$9B,$39,$AF,$AD,$F5,$5F,$37,$81,$F0,$96,$30,$DB,$FC,$23
	.dc.b	$6B,$EF,$3F,$23,$E9,$5F,$53,$F5,$FE,$A9,$F0,$79,$8C,$37,$0F,$F2
	.dc.b	$56,$10,$5F,$28,$C4,$7F,$93,$E3,$0E,$1A,$F1,$73,$C1,$FE,$B7,$6C
	.dc.b	$FE,$0B,$EA,$BC,$15,$F9,$5F,$9B,$F8,$18,$7C,$16,$D7,$E8,$5B,$3B
	.dc.b	$8F,$D2,$4F,$D1,$F4,$64,$7C,$EF,$4A,$D7,$E5,$26,$E5,$FA,$19,$B5
	.dc.b	$F0,$32,$7B,$3F,$92,$FA,$AF,$94,$D9,$FF,$6B,$4E,$9F,$87,$8F,$C6
	.dc.b	$9F,$C5,$B6,$33,$3B,$8F,$8D,$3D,$27,$F7,$9F,$55,$67,$FC,$AB,$C1
	.dc.b	$F8,$57,$E7,$B5,$FA,$8D,$7E,$BB,$5F,$F4,$17,$82,$FA,$6F,$1B,$6F
	.dc.b	$CE,$D9,$FB,$6D,$7F,$00,$F9,$DB,$87,$22,$4F,$C0,$ED,$7C,$97,$C0
	.dc.b	$6C,$66,$30,$7C,$4F,$A7,$FF,$48,$C7,$C8,$6D,$FE,$31,$AF,$E4,$5A
	.dc.b	$FE,$53,$E6,$3F,$87,$3F,$95,$6B,$9F,$07,$E8,$C6,$23,$FD,$49,$F4
	.dc.b	$EF,$2F,$EA,$B7,$0F,$B2,$D9,$FC,$EB,$5F,$D1,$7D,$35,$7F,$4A,$D7
	.dc.b	$FD,$60,$D4,$D3,$FC,$5B,$3F,$86,$6C,$7E,$A9,$AF,$6F,$4F,$F1,$3F
	.dc.b	$48,$E7,$A4,$BF,$C7,$C6,$5F,$C8,$8E,$DF,$C8,$B4,$CB,$83,$3F,$01
	.dc.b	$23,$8B,$70,$FD,$EF,$C3,$BE,$33,$F3,$6D,$AF,$CF,$B6,$77,$3F,$A0
	.dc.b	$FA,$37,$C4,$D2,$5F,$E9,$A3,$2D,$F6,$BF,$9E,$C1,$FE,$7F,$F4,$6F
	.dc.b	$07,$31,$FD,$0B,$E8,$9F,$47,$4D,$FB,$F7,$1F,$F7,$17,$86,$7F,$8D
	.dc.b	$FA,$6F,$2B,$71,$A3,$2E,$37,$FF,$62,$CE,$E7,$51,$FE,$ED,$8C,$3F
	.dc.b	$EC,$AC,$5F,$1D,$FD,$5A,$73,$87,$86,$BF,$DE,$9F,$40,$FA,$2F,$ED
	.dc.b	$1F,$46,$F7,$FF,$AE,$CE,$73,$FF,$01,$CE,$5F,$0D,$A7,$4F,$F8,$06
	.dc.b	$74,$FF,$B7,$23,$38,$7F,$82,$A7,$2F,$8E,$79,$61,$E0,$3C,$7E,$B5
	.dc.b	$C3,$BE,$1F,$5A,$BF,$C1,$DF,$47,$FC,$E1,$F5,$F2,$FF,$80,$63,$11
	.dc.b	$E5,$8C,$79,$FE,$C6,$DF,$4F,$13,$FC,$11,$E2,$BE,$5B,$E0,$C6,$1B
	.dc.b	$3B,$3C,$78,$60,$FF,$7C,$EB,$C6,$3E,$2A,$FA,$BF,$F0,$3B,$99,$DC
	.dc.b	$62,$A5,$7C,$C9,$0F,$16,$FD,$13,$E1,$BF,$C7,$17,$FE,$12,$AD,$D6
	.dc.b	$A7,$37,$67,$F8,$A0,$BE,$4F,$E5,$2F,$F0,$91,$88,$FE,$38,$6E,$B9
	.dc.b	$B4,$E7,$1B,$C7,$D8,$F1,$8F,$19,$B1,$1A,$9F,$EE,$F8,$79,$67,$37
	.dc.b	$E2,$9C,$C7,$1A,$E0,$F8,$23,$C7,$7D,$E3,$E3,$1B,$AE,$4E,$CF,$76
	.dc.b	$1D,$DF,$29,$FD,$9B,$0F,$F3,$A3,$C8,$C1,$FE,$0C,$FA,$27,$D1,$BE
	.dc.b	$93,$FB,$6C,$61,$B1,$99,$E0,$D7,$FC,$31,$2F,$84,$F8,$71,$D5,$AE
	.dc.b	$9A,$FF,$88,$23,$86,$9E,$09,$E3,$9F,$D7,$D3,$FA,$FD,$FC,$B6,$BF
	.dc.b	$46,$F8,$B6,$B8,$CF,$F8,$B0,$DD,$73,$DB,$FC,$5E,$FE,$0E,$F8,$B6
	.dc.b	$B8,$DF,$E5,$AF,$89,$CE,$F8,$BE,$99,$DC,$AF,$9C,$EE,$3D,$15,$C1
	.dc.b	$FD,$0C,$E9,$B5,$CA,$BB,$C7,$86,$B8,$3D,$67,$51,$FE,$9D,$D3,$B4
	.dc.b	$FE,$49,$B8,$FF,$11,$C3,$B4,$FE,$B6,$B8,$3D,$4F,$E3,$F8,$7C,$CA
	.dc.b	$E1,$87,$B3,$E9,$3F,$84,$57,$F2,$46,$BD,$CF,$DD,$B8,$9F,$E4,$0F
	.dc.b	$E0,$E3,$86,$DE,$05,$1B,$CB,$FC,$C2,$D8,$E7,$7B,$FA,$4F,$E9,$51
	.dc.b	$FC,$D1,$AF,$84,$38,$C3,$F9,$87,$D1,$FF,$5D,$9E,$1D,$DE,$18,$FD
	.dc.b	$E6,$3F,$E7,$06,$CF,$F1,$4B,$E5,$3F,$E7,$45,$E3,$78,$C9,$3F,$D1
	.dc.b	$A9,$F4,$98,$8F,$25,$70,$7D,$15,$E1,$9D,$63,$E6,$E6,$F2,$18,$C7
	.dc.b	$C9,$D5,$F1,$FE,$93,$FB,$EE,$3F,$D5,$1A,$FC,$1C,$FF,$D7,$0E,$6F
	.dc.b	$81,$1F,$35,$57,$E4,$7D,$23,$E4,$3E,$A3,$63,$91,$ED,$EA,$CF,$B1
	.dc.b	$23,$07,$F6,$69,$FF,$53,$B7,$CB,$E6,$6D,$FA,$DF,$49,$F3,$1F,$E8
	.dc.b	$2D,$4D,$9F,$F1,$14,$7F,$72,$F8,$3F,$C5,$3F,$7D,$BB,$FF,$6D,$6E
	.dc.b	$FF,$E1,$2F,$AE,$7C,$37,$E3,$FE,$97,$FD,$1B,$87,$F7,$C6,$BF,$89
	.dc.b	$6E,$1F,$DD,$FA,$5F,$61,$F9,$7F,$A6,$FF,$5C,$DF,$F8,$11,$AF,$E4
	.dc.b	$5B,$5C,$7D,$6F,$CF,$69,$F9,$BA,$FF,$6A,$E1,$83,$F6,$34,$CF,$E4
	.dc.b	$CF,$E6,$FE,$99,$EE,$7E,$9D,$B1,$FE,$14,$6B,$FA,$1F,$B0,$46,$FE
	.dc.b	$76,$4E,$DF,$D8,$4F,$9B,$83,$FD,$9F,$9F,$CE,$AF,$99,$C7,$E8,$3C
	.dc.b	$B9,$7F,$97,$63,$FC,$47,$F4,$EF,$73,$F7,$4D,$8C,$C7,$EF,$7E,$C3
	.dc.b	$F6,$11,$FB,$5F,$A8,$7C,$97,$F0,$16,$C7,$F8,$B9,$AF,$EE,$E7,$F3
	.dc.b	$6F,$DF,$FA,$5F,$96,$47,$FC,$8B,$C3,$F2,$7E,$33,$A7,$BF,$27,$A3
	.dc.b	$FC,$3F,$EC,$1F,$71,$FC,$47,$EA,$BF,$C9,$35,$FE,$3E,$6B,$FC,$2D
	.dc.b	$B8,$7F,$8E,$69,$F4,$7F,$C6,$7E,$A7,$F6,$33,$FE,$3D,$6F,$27,$FB
	.dc.b	$E9,$B8,$7F,$87,$BE,$AB,$F8,$13,$FC,$5E,$32,$9A,$62,$37,$5D,$7C
	.dc.b	$6F,$E5,$10,$F9,$2A,$78,$E3,$F9,$67,$4B,$82,$EF,$32,$FF,$2D,$CE
	.dc.b	$5B,$39,$7F,$95,$42,$79,$CF,$4F,$81,$95,$C9,$EF,$F4,$78,$8E,$31
	.dc.b	$73,$FF,$30,$60,$AF,$E8,$1F,$3B,$FB,$4B,$FF,$36,$B5,$FE,$74,$DA
	.dc.b	$E3,$E2,$70,$24,$F0,$7D,$E6,$1F,$4F,$93,$F0,$8C,$4F,$F9,$C7,$EA
	.dc.b	$BF,$7B,$C3,$FE,$50,$A4,$61,$FF,$29,$B5,$CF,$A9,$FD,$5E,$30,$4F
	.dc.b	$90,$C4,$67,$E0,$B5,$F4,$52,$FF,$D1,$37,$FE,$8F,$6C,$E5,$FE,$B1
	.dc.b	$09,$E5,$1F,$13,$87,$E3,$91,$E7,$F5,$55,$F8,$F1,$73,$F7,$4E,$67
	.dc.b	$DC,$36,$FB,$8E,$0F,$F3,$8B,$87,$0D,$78,$FD,$53,$EA,$BF,$BC,$36
	.dc.b	$3F,$D4,$AD,$7F,$B9,$2E,$2D,$CC,$9D,$BC,$A3,$9F,$95,$D3,$9E,$57
	.dc.b	$48,$CB,$DA,$32,$F6,$8F,$F0,$57,$D4,$FE,$D2,$3F,$D6,$ED,$7F,$C0
	.dc.b	$8D,$CF,$FD,$65,$F5,$2F,$AA,$E3,$DB,$2F,$8B,$FE,$15,$9D,$46,$08
	.dc.b	$74,$8D,$4E,$FA,$78,$74,$CA,$FF,$64,$DA,$75,$18,$9D,$3F,$E1,$99
	.dc.b	$D3,$C3,$C4,$F1,$11,$E7,$D3,$5E,$11,$72,$F6,$CF,$F8,$B7,$EA,$5F
	.dc.b	$55,$49,$7F,$ED,$CF,$1F,$F9,$AD,$FF,$B5,$6B,$73,$FE,$E0,$6B,$FE
	.dc.b	$2D,$6C,$7F,$B6,$E7,$8F,$67,$EA,$E7,$2D,$BE,$F7,$FE,$35,$AD,$BF
	.dc.b	$EC,$0A,$DB,$67,$77,$D7,$E4,$2E,$1C,$EB,$C9,$5F,$E4,$9C,$39,$C3
	.dc.b	$0E,$6D,$9E,$2F,$FB,$3A,$77,$3A,$24,$F3,$BD,$D7,$13,$B9,$F1,$DC
	.dc.b	$CF,$08,$C1,$A3,$E3,$4F,$2A,$EE,$F0,$3B,$46,$F0,$FA,$11,$BB,$D4
	.dc.b	$E7,$0F,$E2,$D7,$CA,$30,$7C,$73,$C2,$78,$0C,$9E,$93,$E8,$4F,$47
	.dc.b	$46,$0D,$E1,$EB,$70,$6B,$0F,$2B,$D2,$BC,$4E,$6F,$EA,$3C,$27,$6E
	.dc.b	$4F,$31,$98,$DC,$68,$8F,$BA,$FC,$D1,$89,$FB,$11,$89,$E2,$6E,$3D
	.dc.b	$38,$7C,$96,$9E,$79,$E0,$F8,$AB,$F1,$8E,$8F,$31,$FE,$1B,$8D,$BF
	.dc.b	$26,$36,$EB,$0F,$8D,$18,$3C,$D1,$BB,$F2,$BE,$83,$B4,$6B,$0F,$92
	.dc.b	$F0,$AF,$CB,$C6,$2B,$8D,$73,$39,$1E,$1C,$3D,$CD,$3D,$C7,$68,$E3
	.dc.b	$3C,$E3,$E0,$46,$30,$F6,$DF,$81,$B7,$F9,$4A,$E8,$7C,$4B,$E4,$F8
	.dc.b	$C8,$E2,$BE,$1D,$FD,$39,$E3,$87,$BF,$0E,$B1,$F9,$68,$F8,$18,$7A
	.dc.b	$9B,$1E,$A3,$8D,$79,$DF,$61,$EA,$3E,$1C,$70,$7C,$D5,$D9,$FC,$71
	.dc.b	$FB,$FE,$BF,$E0,$C4,$7F,$90,$75,$57,$F8,$A4,$5A,$FE,$39,$3F,$C3
	.dc.b	$7E,$62,$57,$F2,$77,$FD,$B1,$F5,$4C,$73,$3F,$C7,$66,$DE,$60,$2B
	.dc.b	$0F,$FB,$5F,$EA,$9F,$0F,$FD,$CE,$37,$18,$70,$D7,$FB,$C3,$5F,$EF
	.dc.b	$BF,$7D,$FE,$D0,$79,$35,$FE,$FC,$D7,$7F,$01,$DD,$7D,$5B,$EA,$DF
	.dc.b	$27,$E3,$8C,$37,$C1,$7C,$FE,$6B,$77,$FE,$0E,$D5,$F4,$DF,$E7,$EE
	.dc.b	$0F,$F3,$AB,$65,$3F,$14,$DF,$3D,$AF,$8F,$1C,$BA,$31,$5C,$2F,$94
	.dc.b	$F1,$6D,$70,$7F,$DB,$4F,$82,$B8,$CB,$E9,$FF,$D5,$73,$FE,$88,$6D
	.dc.b	$F9,$23,$C3,$F0,$7F,$09,$D7,$B9,$CA,$B7,$7E,$96,$CF,$0D,$B5,$5E
	.dc.b	$8C,$1F,$50,$FA,$9F,$1F,$D5,$3E,$1F,$53,$9C,$7F,$D3,$73,$97,$E1
	.dc.b	$9F,$E5,$78,$D5,$F9,$CF,$97,$3B,$70,$78,$2F,$BD,$F6,$65,$FD,$4F
	.dc.b	$42,$D7,$3D,$3E,$37,$BD,$EF,$3E,$07,$C6,$5B,$F9,$11,$6B,$86,$CA
	.dc.b	$C5,$FD,$5F,$A0,$FA,$3C,$4F,$89,$B1,$86,$E5,$F8,$66,$E3,$E3,$AE
	.dc.b	$13,$F0,$A3,$F5,$7F,$4E,$F9,$7F,$BB,$F5,$4F,$7F,$E8,$4E,$6F,$FD
	.dc.b	$8B,$39,$79,$5A,$74,$FC,$19,$D3,$F2,$A3,$39,$FE,$5E,$72,$FE,$EA
	.dc.b	$33,$5F,$E8,$B6,$D7,$E7,$B5,$7D,$57,$93,$2F,$87,$F4,$3F,$13,$3B
	.dc.b	$8F,$E9,$9B,$87,$C1,$6E,$5F,$9F,$7F,$25,$E8,$CA,$FC,$D7,$A5,$6F
	.dc.b	$E9,$CD,$BF,$CB,$4D,$AF,$37,$B5,$F9,$2F,$AA,$F8,$ED,$8F,$8B,$B2
	.dc.b	$31,$8F,$C3,$6E,$1F,$05,$E3,$5F,$BE,$FA,$AB,$5F,$92,$E2,$72,$FE
	.dc.b	$CB,$F1,$6F,$E5,$17,$F4,$A7,$2F,$FB,$EA,$33,$5F,$9A,$F0,$9F,$57
	.dc.b	$FB,$AE,$97,$8D,$FA,$D9,$CE,$1F,$85,$D5,$5A,$8F,$BF,$3E,$B7,$5F
	.dc.b	$67,$A7,$3B,$7F,$06,$8C,$65,$FA,$AF,$09,$F3,$5F,$C2,$BE,$CF,$A4
	.dc.b	$E6,$31,$A7,$EF,$FE,$B7,$FD,$BD,$97,$8F,$67,$76,$7D,$7F,$E2,$71
	.dc.b	$89,$FC,$9F,$D0,$7F,$0A,$CB,$F9,$2C,$FF,$BC,$67,$3E,$37,$0F,$81
	.dc.b	$FE,$19,$F5,$AF,$4F,$F8,$EF,$D9,$73,$C7,$F8,$0C,$E7,$0F,$89,$39
	.dc.b	$E6,$FE,$A3,$F5,$D9,$BF,$E0,$13,$FE,$A3,$AF,$D2,$CE,$6B,$F9,$B4
	.dc.b	$E6,$BF,$86,$CB,$DB,$FE,$5B,$DE,$F7,$FF,$90,$4E,$71,$FA,$93,$97
	.dc.b	$F6,$3F,$5C,$F0,$BF,$9A,$FD,$8F,$5D,$3F,$B2,$CE,$6B,$EE,$4E,$5F
	.dc.b	$F7,$C1,$EB,$3F,$A3,$FD,$8F,$FB,$3D,$FD,$E9,$9D,$3F,$A7,$C6,$5F
	.dc.b	$E4,$B1,$F3,$5F,$E4,$1E,$07,$FA,$9A,$38,$F0,$FF,$48,$8C,$4F,$F2
	.dc.b	$CF,$A0,$7C,$B7,$F3,$DF,$1D,$F3,$BF,$CF,$E7,$38,$FF,$0E,$9C,$BF
	.dc.b	$C4,$3E,$C1,$F1,$FF,$D1,$57,$83,$63,$0E,$7E,$0A,$57,$F1,$8F,$B0
	.dc.b	$79,$B8,$D7,$F9,$FF,$D7,$BE,$C3,$FA,$83,$6F,$87,$C4,$E2,$7F,$AA
	.dc.b	$61,$FE,$D3,$9C,$9F,$C9,$3D,$BF,$E7,$4E,$30,$FE,$97,$18,$6F,$AF
	.dc.b	$DB,$EC,$3F,$AF,$4E,$7A,$7F,$C3,$BF,$5C,$FA,$FF,$EE,$3F,$60,$F8
	.dc.b	$AF,$EC,$93,$9C,$FF,$71,$39,$7F,$9A,$C6,$73,$FE,$A9,$39,$7F,$AE
	.dc.b	$C6,$70,$FF,$05,$4E,$5F,$0E,$1D,$8F,$0F,$BD,$70,$EB,$C9,$FE,$04
	.dc.b	$FB,$0F,$E6,$F9,$32,$FA,$B5,$FC,$2E,$73,$18,$9F,$EC,$D1,$8A,$E5
	.dc.b	$ED,$7B,$3D,$0F,$F0,$7F,$3B,$E8,$F1,$97,$F1,$58,$C3,$EE,$9C,$E3
	.dc.b	$EC,$2E,$BF,$E4,$63,$6F,$F7,$C5,$F6,$1F,$E0,$D7,$26,$DC,$61,$2F
	.dc.b	$FA,$FD,$7F,$94,$BC,$EF,$F2,$9F,$D7,$BE,$67,$FC,$AB,$3F,$E3,$4B
	.dc.b	$FF,$0C,$3B,$13,$B7,$FC,$6B,$1F,$E5,$D8,$C4,$72,$AE,$B5,$D3,$1F
	.dc.b	$4D,$6A,$BA,$3C,$23,$33,$E2,$9C,$BC,$6F,$B3,$98,$5F,$67,$C7,$85
	.dc.b	$7C,$47,$A2,$79,$46,$63,$A9,$C0,$D1,$E1,$9E,$57,$E4,$8C,$1A,$3E
	.dc.b	$34,$F3,$7C,$66,$63,$83,$E2,$3F,$CF,$93,$FE,$6F,$76,$60,$F6,$7D
	.dc.b	$7F,$FA,$0B,$EC,$3F,$D0,$4D,$8C,$FC,$7A,$72,$D8,$E3,$F5,$FF,$E6
	.dc.b	$C6,$E1,$CA,$31,$1F,$36,$D1,$88,$FC,$0C,$62,$7E,$56,$5F,$D8,$26
	.dc.b	$ED,$F8,$6D,$7E,$FD,$EE,$7E,$23,$63,$F0,$75,$C6,$7E,$43,$63,$8E
	.dc.b	$1E,$76,$E1,$DA,$A1,$D2,$75,$5E,$66,$C7,$FA,$B2,$3A,$1F,$9A,$2B
	.dc.b	$FE,$B0,$BF,$24,$71,$AE,$B7,$E6,$C3,$D9,$39,$BF,$3C,$7E,$67,$4E
	.dc.b	$8D,$8F,$03,$56,$1D,$9C,$9F,$E4,$EA,$F8,$58,$7F,$B0,$1B,$7D,$F0
	.dc.b	$FD,$4F,$D8,$7F,$07,$BF,$F2,$AB,$5F,$F3,$DB,$6F,$FD,$92,$FD,$A8
	.dc.b	$DE,$DE,$38,$D5,$FE,$F1,$B3,$FE,$D2,$78,$3F,$E8,$6F,$B0,$FF,$3C
	.dc.b	$3F,$E6,$46,$B9,$EC,$F7,$7D,$7B,$C3,$F8,$15,$AC,$BF,$DB,$73,$FB
	.dc.b	$BC,$7F,$85,$B7,$0F,$F7,$0D,$6E,$7F,$87,$AF,$1B,$99,$27,$FC,$EB
	.dc.b	$5F,$ED,$9B,$FF,$73,$D6,$E7,$CD,$5B,$7F,$D3,$F5,$BB,$EB,$3E,$8E
	.dc.b	$79,$5C,$0E,$BD,$5F,$67,$EC,$3F,$B9,$C7,$FA,$11,$AF,$48,$D4,$C8
	.dc.b	$D3,$F1,$23,$11,$FE,$B0,$F2,$1F,$1F,$F8,$56,$C7,$03,$DB,$D5,$9F
	.dc.b	$2E,$46,$0F,$F4,$AC,$FF,$A9,$DB,$E5,$F3,$36,$FF,$AF,$BE,$C3,$E6
	.dc.b	$20,$E1,$BC,$5C,$5F,$F3,$37,$C8,$3E,$DB,$F3,$9C,$68,$FB,$D3,$AE
	.dc.b	$4C,$C6,$E3,$FC,$DD,$F7,$3F,$E2,$06,$FD,$8F,$28,$37,$4E,$BF,$36
	.dc.b	$F2,$69,$2F,$FC,$F1,$0F,$9E,$EF,$F1,$12,$7C,$EF,$8B,$C0,$ED,$F2
	.dc.b	$F2,$7C,$6F,$87,$E5,$FD,$7F,$35,$FE,$4D,$9F,$F4,$22,$F7,$32,$6B
	.dc.b	$DC,$F8,$3F,$5C,$FB,$BF,$7D,$C3,$FD,$19,$19,$7E,$0F,$DE,$FB,$FE
	.dc.b	$B6,$46,$0E,$FF,$6D,$E2,$23,$E2,$F8,$3E,$FC,$D2,$3A,$BE,$2E,$FF
	.dc.b	$A7,$F7,$5F,$E7,$D9,$FF,$4E,$35,$FF,$7B,$37,$A9,$E2,$FE,$05,$75
	.dc.b	$7F,$82,$63,$1C,$2E,$3F,$8B,$C7,$31,$F8,$B1,$FC,$6F,$EB,$1F,$51
	.dc.b	$49,$3F,$EE,$88,$7D,$46,$21,$9F,$89,$FC,$94,$F2,$78,$57,$5A,$E1
	.dc.b	$5B,$95,$6D,$E7,$97,$F8,$52,$31,$3B,$7C,$33,$FE,$FA,$FA,$DF,$DB
	.dc.b	$BF,$E0,$96,$BF,$E1,$C6,$BF,$A5,$6A,$7F,$7F,$3D,$79,$FE,$11,$FC
	.dc.b	$F6,$33,$7F,$E2,$6C,$1F,$5B,$FD,$05,$7C,$2D,$31,$13,$1F,$A9,$5C
	.dc.b	$67,$51,$83,$F7,$15,$D1,$C1,$F3,$BE,$07,$B1,$1C,$C3,$E4,$31,$1F
	.dc.b	$20,$E7,$38,$CF,$E7,$56,$E7,$8C,$E7,$0F,$63,$EB,$BF,$EB,$9B,$7F
	.dc.b	$BA,$AB,$FC,$60,$7E,$27,$E5,$EB,$4C,$C6,$DF,$F1,$B3,$71,$F0,$F2
	.dc.b	$7C,$73,$AD,$7F,$69,$FA,$D7,$CC,$52,$61,$F0,$A3,$FA,$52,$76,$E6
	.dc.b	$36,$FC,$13,$5E,$B7,$5F,$82,$ED,$F5,$12,$7A,$46,$5E,$D1,$97,$95
	.dc.b	$FF,$A5,$DD,$9A,$21,$A9,$F8,$FC,$1F,$13,$4F,$C1,$CF,$C4,$8C,$11
	.dc.b	$FB,$BF,$59,$FB,$57,$F9,$C3,$5F,$DB,$47,$F8,$36,$3F,$43,$C1,$FE
	.dc.b	$76,$6C,$F0,$E6,$D7,$16,$D9,$1D,$9C,$BE,$0F,$0B,$A7,$34,$8C,$10
	.dc.b	$F9,$57,$FC,$9D,$F3,$C7,$8E,$64,$60,$87,$07,$FC,$49,$F5,$8E,$EF
	.dc.b	$F1,$37,$D6,$7C,$4A,$FA,$CD,$E5,$FE,$05,$5B,$9F,$D0,$B5,$DE,$BC
	.dc.b	$8E,$B7,$D3,$9F,$AF,$3C,$AF,$D4,$E2,$9D,$46,$27,$4F,$8A,$3C,$BD
	.dc.b	$F2,$38,$BE,$8E,$8C,$47,$A2,$2E,$7F,$C6,$70,$FE,$FB,$39,$38,$90
	.dc.b	$F7,$D7,$0E,$1F,$F5,$7C,$FF,$6F,$9D,$57,$3A,$D4,$7F,$7B,$9D,$3D
	.dc.b	$4E,$67,$0A,$E7,$7C,$21,$3E,$4B,$F3,$4E,$63,$9C,$77,$9E,$47,$79
	.dc.b	$D5,$ED,$B1,$A3,$AB,$87,$C9,$3D,$9D,$57,$86,$FC,$B1,$8A,$F9,$8D
	.dc.b	$37,$7D,$DA,$64,$EB,$3F,$22,$35,$18,$3E,$33,$8B,$F1,$47,$C6,$34
	.dc.b	$7A,$0E,$A6,$2F,$E1,$D7,$9A,$38,$4F,$22,$46,$5F,$91,$1C,$6F,$E1
	.dc.b	$46,$A3,$E5,$1D,$9C,$3E,$13,$C1,$7E,$7A,$E8,$7B,$63,$E0,$4F,$0B
	.dc.b	$F9,$51,$D0,$F0,$1F,$9E,$AE,$4E,$67,$CA,$71,$BF,$2D,$F6,$9D,$35
	.dc.b	$FC,$3D,$72,$7C,$D5,$CA,$75,$7F,$1A,$78,$9F,$40,$F3,$C6,$AB,$94
	.dc.b	$EE,$7A,$9A,$38,$98,$3D,$67,$C4,$34,$78,$A7,$E2,$3A,$AF,$89,$5D
	.dc.b	$AF,$F1,$73,$9A,$F1,$3C,$AF,$C4,$7C,$A7,$A1,$E2,$73,$5C,$E3,$E6
	.dc.b	$1C,$6B,$AD,$FC,$B8,$FB,$97,$F1,$27,$E6,$C7,$36,$BD,$89,$7D,$9E
	.dc.b	$87,$63,$D0,$78,$EF,$C4,$FE,$4A,$BF,$CF,$8B,$EB,$1E,$4F,$FB,$47
	.dc.b	$E4,$FF,$B2,$9F,$D9,$A3,$33,$FE,$0A,$6B,$FD,$96,$37,$D3,$FE,$D9
	.dc.b	$9C,$E7,$FE,$33,$9F,$EC,$15,$BE,$7E,$C7,$7E,$9B,$1F,$E1,$69,$DD
	.dc.b	$71,$AD,$4E,$7E,$5B,$CC,$47,$FC,$3B,$F5,$9F,$D7,$EF,$FC,$B5,$A7
	.dc.b	$B9,$B3,$FE,$9B,$6A,$9E,$15,$CE,$3D,$8F,$FB,$CE,$3F,$B5,$7D,$5B
	.dc.b	$C2,$FF,$48,$57,$CA,$3F,$D4,$D1,$9F,$3B,$FC,$3A,$FF,$99,$7C,$DB
	.dc.b	$5C,$2B,$87,$DB,$FF,$A6,$73,$FF,$11,$B6,$3A,$07,$0F,$0C,$3F,$CB
	.dc.b	$CF,$F9,$D7,$EA,$7F,$D5,$D5,$FE,$41,$D9,$18,$9F,$F1,$D4,$63,$8F
	.dc.b	$FC,$D8,$C9,$D3,$79,$73,$7D,$5B,$7F,$E7,$4A,$FF,$69,$CF,$77,$8B
	.dc.b	$FE,$87,$FA,$87,$CE,$70,$C1,$18,$9F,$37,$B7,$EF,$9F,$1C,$E7,$0F
	.dc.b	$3B,$63,$37,$CE,$B8,$5F,$78,$ED,$5B,$8C,$1D,$B3,$CC,$FA,$B0,$7D
	.dc.b	$63,$EA,$FE,$5C,$7F,$B7,$E3,$17,$FE,$56,$6C,$7D,$46,$BF,$81,$73
	.dc.b	$97,$FA,$45,$B3,$F5,$7C,$8F,$04,$62,$BF,$D3,$B9,$3E,$9B,$D2,$7A
	.dc.b	$BD,$BB,$87,$DB,$9D,$36,$74,$D3,$E9,$7D,$0F,$CF,$DD,$EF,$D8,$C3
	.dc.b	$67,$EC,$37,$1F,$C2,$D7,$09,$E6,$68,$B1,$A9,$F9,$5F,$50,$F9,$0F
	.dc.b	$93,$28,$C5,$7F,$9F,$9B,$1F,$9C,$6B,$FA,$97,$3B,$7F,$AE,$5B,$3F
	.dc.b	$8B,$FA,$AE,$F7,$ED,$F9,$AF,$77,$CE,$5A,$FC,$93,$71,$F5,$B7,$1F
	.dc.b	$CF,$3F,$86,$F2,$A5,$7C,$DF,$2E,$DF,$E4,$26,$DF,$E5,$26,$D7,$E7
	.dc.b	$66,$D7,$DA,$9B,$4F,$A8,$7B,$7F,$A1,$6E,$1F,$B5,$5E,$DE,$67,$F4
	.dc.b	$DC,$CF,$0E,$98,$F8,$7B,$87,$EF,$7E,$A3,$F5,$FA,$7F,$B9,$9F,$D9
	.dc.b	$BF,$BE,$73,$99,$CB,$FC,$03,$EA,$3B,$57,$F0,$75,$7F,$B7,$73,$7F
	.dc.b	$C2,$1C,$CF,$F0,$DF,$A6,$FC,$34,$FF,$0C,$59,$7E,$FD,$CE,$1F,$C3
	.dc.b	$9C,$D7,$1F,$89,$D1,$BF,$B5,$B1,$FC,$67,$E9,$B8,$AC,$58,$27,$31
	.dc.b	$87,$AB,$FC,$79,$CC,$6F,$DC,$FE,$25,$1F,$EE,$EA,$ED,$EA,$E2,$7F
	.dc.b	$8B,$71,$FE,$D6,$CB,$E7,$B3,$E1,$69,$FE,$39,$F4,$9C,$F3,$FE,$5A
	.dc.b	$FF,$BE,$32,$FE,$64,$E7,$0F,$E5,$EE,$67,$F9,$9F,$D2,$3E,$17,$F9
	.dc.b	$33,$FE,$9D,$BF,$E7,$4E,$5B,$EC,$FF,$2B,$F0,$3F,$4B,$7F,$11,$B8
	.dc.b	$FF,$00,$FA,$FD,$77,$FF,$1F,$F9,$27,$27,$F2,$1E,$F7,$C6,$66,$30
	.dc.b	$FF,$25,$FA,$5F,$BF,$9F,$FB,$85,$79,$B9,$EC,$D3,$BF,$8F,$9C,$1F
	.dc.b	$5F,$63,$FE,$0E,$FA,$37,$8B,$AC,$BF,$DD,$9F,$4A,$FA,$5F,$E6,$8D
	.dc.b	$BE,$47,$1F,$80,$CC,$FF,$39,$C3,$FD,$65,$3A,$BF,$C3,$46,$1F,$F7
	.dc.b	$C7,$A4,$F4,$38,$61,$FD,$06,$31,$7C,$69,$18,$7C,$BD,$7B,$BF,$E9
	.dc.b	$1F,$46,$FA,$4F,$EB,$9F,$4A,$F7,$B3,$18,$6E,$5F,$DC,$E7,$4F,$7A
	.dc.b	$E4,$E6,$77,$8F,$E7,$A3,$FC,$6B,$3A,$8F,$EF,$34,$F9,$8D,$F0,$FC
	.dc.b	$E9,$FD,$35,$7D,$88,$FE,$FB,$F4,$BF,$86,$94,$E7,$0F,$AD,$39,$8C
	.dc.b	$4F,$EB,$A3,$1C,$BF,$DE,$5D,$7C,$AF,$F7,$1F,$9E,$F8,$6D,$C3,$D8
	.dc.b	$E1,$F7,$5D,$3D,$87,$E1,$67,$FB,$96,$2F,$A5,$B1,$86,$D7,$D5,$94
	.dc.b	$FF,$79,$AF,$E0,$9F,$4B,$FC,$17,$E8,$5F,$11,$FE,$5C,$BD,$E1,$F7
	.dc.b	$6F,$62,$B6,$F8,$E3,$95,$7F,$8C,$2E,$CE,$4E,$11,$96,$99,$AE,$6F
	.dc.b	$F9,$B2,$78,$BF,$C6,$30,$FA,$D3,$A8,$F1,$BC,$30,$F3,$C7,$43,$91
	.dc.b	$5C,$C7,$18,$57,$69,$CD,$F9,$6B,$A5,$F8,$23,$07,$C5,$71,$7F,$E5
	.dc.b	$9B,$E5,$18,$8F,$09,$FE,$52,$3B,$94,$A6,$63,$5F,$44,$FA,$47,$D2
	.dc.b	$E1,$B5,$FE,$5C,$9D,$4E,$4E,$3F,$4B,$FC,$FA,$74,$D7,$FC,$DD,$28
	.dc.b	$C4,$7F,$A6,$A3,$19,$F2,$3E,$8E,$4D,$BF,$D3,$FA,$7D,$68,$C5,$FD
	.dc.b	$6B,$E2,$FD,$16,$CE,$EB,$8D,$7F,$AA,$0E,$35,$BC,$3D,$15,$F8,$2B
	.dc.b	$E3,$5B,$BE,$37,$FD,$6E,$38,$D6,$EF,$8C,$7A,$DB,$3B,$AE,$2F,$E7
	.dc.b	$9B,$7D,$A3,$FB,$33,$5D,$1F,$0A,$37,$3D,$23,$27,$ED,$0B,$87,$F0
	.dc.b	$B6,$EB,$E3,$8E,$51,$FC,$9F,$0F,$83,$1B,$6C,$EE,$7F,$35,$F4,$BF
	.dc.b	$B8,$7F,$97,$35,$F0,$36,$BC,$0F,$DD,$8D,$ED,$FB,$BA,$FF,$6B,$CE
	.dc.b	$DB,$19,$9F,$F0,$34,$F1,$7D,$9F,$4B,$FD,$05,$FE,$7C,$D7,$3D,$1C
	.dc.b	$57,$D2,$BC,$1C,$56,$F2,$F8,$93,$FB,$7C,$BF,$DD,$2D,$BF,$F0,$B5
	.dc.b	$6E,$7F,$DD,$AB,$C7,$E6,$49,$FE,$A1,$4F,$AA,$C4,$7A,$26,$E1,$E0
	.dc.b	$9B,$87,$83,$9F,$D1,$87,$F8,$E3,$0F,$6E,$18,$E9,$F6,$7D,$36,$67
	.dc.b	$EF,$37,$2F,$5C,$EB,$09,$5F,$D8,$EB,$58,$FB,$AB,$33,$F8,$7F,$1D
	.dc.b	$F1,$3F,$25,$B1,$C7,$0F,$2F,$8F,$9F,$36,$4F,$C0,$8A,$F3,$F2,$B9
	.dc.b	$18,$D3,$F2,$5F,$4D,$F2,$E7,$FC,$84,$D4,$DC,$7F,$C3,$B1,$FE,$E1
	.dc.b	$F8,$A7,$D3,$7E,$16,$D1,$F7,$67,$5C,$99,$8D,$C7,$FB,$A7,$E8,$7F
	.dc.b	$C4,$AD,$FA,$6E,$50,$6E,$3D,$7E,$5B,$FD,$F9,$8F,$FB,$B1,$BC,$5F
	.dc.b	$E2,$36,$E1,$FE,$5A,$5F,$4D,$E1,$F9,$29,$3E,$97,$C9,$DF,$E0,$F9
	.dc.b	$E9,$3E,$57,$C7,$F2,$7E,$DF,$A3,$7C,$5D,$F7,$E5,$FF,$89,$75,$56
	.dc.b	$A3,$1F,$4A,$FA,$6F,$85,$10,$CB,$F0,$3E,$2F,$E0,$FC,$64,$9F,$8D
	.dc.b	$1E,$1F,$81,$91,$DF,$0F,$F2,$E3,$7E,$03,$E9,$63,$F4,$A3,$C3,$CF
	.dc.b	$83,$3D,$FF,$5F,$E9,$9F,$35,$49,$5F,$57,$9F,$E0,$71,$FC,$1E,$39
	.dc.b	$3E,$18,$FD,$E7,$D3,$3E,$8A,$92,$7F,$C8,$10,$FA,$3C,$47,$F2,$38
	.dc.b	$E0,$EA,$79,$CE,$B9,$3F,$93,$56,$AB,$85,$69,$FF,$26,$7D,$37,$EA
	.dc.b	$9F,$F4,$D3,$53,$67,$F8,$34,$70,$F1,$7D,$3C,$F2,$30,$7F,$95,$F1
	.dc.b	$7C,$57,$F1,$95,$F1,$74,$C4,$49,$E1,$19,$79,$47,$1F,$80,$91,$D5
	.dc.b	$E9,$F4,$58,$8F,$E8,$D6,$9C,$6B,$FD,$1E,$15,$8E,$0F,$F5,$C5,$FC
	.dc.b	$DB,$C3,$FE,$BB,$AD,$BF,$E7,$1D,$FF,$D9,$2D,$F9,$0F,$E6,$91,$88
	.dc.b	$F0,$C6,$0F,$81,$C9,$23,$81,$FE,$7B,$FA,$57,$D3,$7F,$3D,$6C,$66
	.dc.b	$31,$EB,$65,$91,$93,$94,$71,$F8,$1E,$5F,$45,$27,$B4,$65,$F0,$46
	.dc.b	$5F,$ED,$97,$FE,$85,$76,$68,$86,$A7,$DF,$C3,$23,$E2,$3E,$9F,$8D
	.dc.b	$C4,$7C,$C3,$FD,$C9,$A7,$E0,$1F,$F1,$B3,$5D,$07,$0F,$1F,$A6,$FC
	.dc.b	$C7,$17,$F7,$F6,$F0,$FD,$86,$CF,$FB,$C5,$FD,$33,$FE,$A7,$72,$9A
	.dc.b	$4E,$0F,$ED,$8F,$F0,$F8,$FF,$07,$3F,$EE,$C7,$97,$26,$BC,$49,$18
	.dc.b	$21,$C1,$CB,$CA,$3A,$98,$C3,$FC,$31,$F4,$CF,$03,$FB,$E3,$6B,$FD
	.dc.b	$7C,$D7,$FC,$41,$F4,$DF,$CC,$9C,$CF,$F9,$76,$75,$1F,$3D,$AF,$E7
	.dc.b	$1A,$84,$FE,$95,$FF,$66,$3F,$9E,$7B,$F3,$FF,$83,$E3,$51,$E2,$E7
	.dc.b	$C4,$C8,$E3,$1B,$F0,$24,$74,$79,$73,$E2,$3D,$71,$72,$FF,$1D,$61
	.dc.b	$FD,$D2,$72,$71,$21,$F0,$AB,$86,$FF,$B8,$9F,$E6,$D3,$C0,$D1,$C2
	.dc.b	$BA,$DF,$08,$F8,$53,$DE,$73,$1B,$AE,$95,$E8,$8C,$1C,$1E,$4F,$CA
	.dc.b	$BB,$7E,$2B,$E7,$3E,$23,$C8,$E0,$97,$E9,$AE,$37,$C9,$B5,$E0,$3B
	.dc.b	$16,$FB,$57,$9A,$3E,$CC,$F0,$7B,$C7,$57,$85,$75,$69,$CE,$78,$5F
	.dc.b	$92,$3E,$2C,$62,$BB,$47,$AE,$74,$FE,$2A,$BA,$3E,$03,$E3,$1D,$4E
	.dc.b	$E6,$4F,$59,$6F,$D6,$F0,$3B,$57,$82,$BB,$CE,$A7,$8C,$62,$BC,$2D
	.dc.b	$8F,$01,$D6,$33,$5C,$C9,$1D,$E7,$F2,$37,$EF,$8C,$4F,$C9,$79,$1D
	.dc.b	$27,$B3,$E9,$9E,$0F,$C3,$3E,$99,$F0,$23,$D3,$1A,$BF,$51,$F9,$1A
	.dc.b	$F1,$D7,$13,$A1,$C5,$CC,$6E,$3C,$A4,$38,$57,$A9,$E0,$66,$38,$3A
	.dc.b	$9E,$77,$F9,$D8,$F6,$9E,$02,$D7,$A1,$E6,$71,$3F,$0A,$62,$BD,$77
	.dc.b	$ED,$AE,$95,$F1,$8E,$55,$D2,$FD,$75,$B8,$F8,$E6,$4F,$43,$97,$F0
	.dc.b	$67,$AE,$BE,$DB,$4F,$A9,$3E,$C6,$9F,$C7,$FE,$99,$E5,$FF,$AA,$5B
	.dc.b	$1F,$C8,$F0,$FF,$65,$46,$23,$FD,$9D,$F3,$5F,$E7,$17,$FD,$A1,$19
	.dc.b	$9D,$D7,$33,$FD,$69,$5B,$8F,$F6,$A3,$FD,$AE,$B9,$F3,$68,$DF,$45
	.dc.b	$8F,$16,$3F,$E1,$2A,$E1,$7C,$AF,$75,$A3,$F9,$D4,$FF,$94,$1C,$07
	.dc.b	$0F,$AB,$C2,$FF,$64,$7D,$37,$FB,$9E,$FC,$B1,$E8,$24,$F2,$9D,$61
	.dc.b	$E3,$9C,$CF,$33,$51,$FE,$D3,$FA,$FF,$F6,$24,$7F,$7A,$0E,$1F,$EF
	.dc.b	$AD,$7F,$C0,$8D,$7F,$C6,$3F,$4E,$F1,$5F,$1D,$CE,$30,$DE,$B7,$9F
	.dc.b	$9A,$DD,$FF,$83,$F4,$7D,$0F,$FA,$AB,$3F,$F4,$63,$65,$29,$FC,$14
	.dc.b	$FF,$B1,$9D,$F1,$E6,$B7,$1A,$8C,$54,$AE,$8D,$E4,$CD,$6C,$FF,$4E
	.dc.b	$7D,$23,$E9,$B3,$8F,$8A,$31,$3F,$EA,$1F,$6F,$E0,$BC,$2B,$78,$2B
	.dc.b	$77,$E4,$39,$4E,$AF,$8C,$66,$B8,$4F,$2A,$E3,$1E,$A6,$DF,$E8,$EF
	.dc.b	$8C,$E6,$BC,$F6,$FA,$67,$CE,$79,$3E,$BD,$F0,$1A,$9C,$E5,$FE,$BB
	.dc.b	$9C,$BF,$96,$8C,$DF,$A4,$E9,$7E,$ED,$5E,$57,$38,$F4,$7D,$3C,$C3
	.dc.b	$E5,$FC,$3D,$C3,$DD,$BA,$B1,$B7,$B6,$AD,$FD,$EF,$A7,$ED,$F0,$7D
	.dc.b	$DB,$1F,$86,$6E,$3F,$27,$44,$63,$0F,$C0,$36,$BA,$1F,$8E,$BF,$8F
	.dc.b	$F4,$EF,$9E,$FB,$3F,$5E,$F9,$FC,$C6,$1B,$7F,$BA,$AE,$0F,$E2,$3E
	.dc.b	$A3,$3A,$78,$FE,$7F,$E3,$5F,$C7,$6D,$7E,$79,$B3,$B8,$FD,$0C,$FE
	.dc.b	$07,$CD,$91,$F9,$4F,$36,$D7,$CA,$9B,$B7,$E1,$66,$D7,$F2,$7E,$2F
	.dc.b	$F7,$CB,$F9,$CA,$FC,$64,$EA,$FF,$5B,$F5,$0F,$7F,$F3,$2D,$BF,$F7
	.dc.b	$F2,$F8,$0C,$CF,$E7,$B9,$FF,$8D,$57,$F1,$FC,$FF,$07,$E3,$E7,$D9
	.dc.b	$D3,$FA,$4F,$A8,$FD,$8D,$7F,$3B,$7F,$75,$9F,$E8,$5C,$BF,$C0,$1C
	.dc.b	$DF,$EE,$BE,$A3,$BC,$FE,$9D,$61,$FB,$07,$33,$B8,$FD,$9F,$D4,$39
	.dc.b	$EE,$78,$6E,$5E,$99,$7C,$6F,$03,$8F,$D4,$59,$F7,$7D,$47,$36,$DF
	.dc.b	$F0,$08,$FD,$1C,$60,$DF,$BF,$93,$5E,$E5,$8F,$91,$C9,$FA,$ED,$7F
	.dc.b	$8B,$AF,$A2,$FE,$16,$E6,$7F,$B1,$39,$7E,$87,$D4,$7F,$BE,$1B,$7F
	.dc.b	$C4,$A3,$1C,$3F,$C3,$5F,$F7,$DC,$FD,$7F,$A8,$7C,$4F,$F1,$87,$FD
	.dc.b	$3D,$19,$8C,$7B,$BF,$C7,$36,$FC,$FE,$1F,$C5,$DB,$A7,$F2,$E5,$F0
	.dc.b	$1B,$F0,$7F,$89,$FC,$A7,$CC,$CB,$FB,$0F,$7B,$E3,$B3,$CB,$FC,$62
	.dc.b	$31,$3F,$D9,$BE,$A3,$EE,$B7,$CF,$FE,$3B,$18,$D7,$F8,$A2,$8C,$FC
	.dc.b	$94,$AF,$ED,$FF,$4C,$F1,$71,$87,$F1,$BF,$A8,$7D,$47,$F2,$86,$DF
	.dc.b	$0F,$83,$C4,$FF,$2C,$C1,$14,$FE,$5F,$ED,$3C,$FC,$5F,$F7,$E8,$C3
	.dc.b	$63,$B7,$67,$07,$F9,$BF,$D3,$3E,$A3,$FA,$4F,$D4,$3E,$0B,$31,$86
	.dc.b	$E1,$FA,$F9,$D3,$FE,$0C,$AE,$4E,$67,$7B,$7F,$39,$9C,$BF,$D3,$EB
	.dc.b	$F9,$FB,$9B,$F1,$F8,$DF,$D6,$BE,$A3,$EE,$E3,$FE,$D3,$A6,$1F,$47
	.dc.b	$B3,$FB,$6D,$6A,$7F,$DB,$13,$9A,$FF,$1C,$4E,$67,$8D,$F5,$EA,$F8
	.dc.b	$3F,$2D,$FD,$8F,$E8,$6E,$1F,$DB,$5B,$7E,$99,$F8,$EE,$32,$FF,$13
	.dc.b	$4F,$F6,$05,$F5,$16,$30,$DB,$FF,$18,$CA,$7F,$B3,$D7,$FB,$BF,$EA
	.dc.b	$3F,$DE,$1F,$4C,$F9,$2F,$81,$7B,$C3,$FC,$73,$7B,$15,$B7,$FC,$AD
	.dc.b	$87,$F8,$4E,$31,$1F,$27,$0E,$B8,$74,$AD,$3F,$36,$7A,$57,$0C,$3C
	.dc.b	$11,$DA,$74,$AF,$9B,$E0,$C2,$BD,$CE,$06,$4A,$66,$FB,$61,$E2,$8F
	.dc.b	$15,$72,$C3,$C3,$5E,$A9,$F2,$E1,$E9,$C3,$BC,$60,$ED,$3B,$7C,$C7
	.dc.b	$82,$B9,$D7,$AA,$FF,$CA,$53,$F2,$CF,$F2,$A3,$87,$89,$F1,$7E,$99
	.dc.b	$F5,$0F,$A8,$FF,$2D,$46,$1B,$19,$9F,$8F,$39,$6F,$D4,$7F,$93,$1B
	.dc.b	$8F,$28,$C4,$77,$6B,$FA,$19,$47,$1C,$BE,$EE,$4D,$BF,$4D,$18,$6C
	.dc.b	$71,$9E,$0D,$C3,$EB,$BC,$70,$B5,$C1,$0C,$3C,$35,$F7,$9B,$1F,$72
	.dc.b	$B9,$36,$B9,$3C,$6F,$C2,$F4,$AE,$38,$79,$E3,$C2,$75,$9D,$BC,$70
	.dc.b	$CB,$D4,$D3,$83,$FC,$E4,$FE,$A6,$FF,$53,$A7,$F0,$9C,$7C,$B5,$AC
	.dc.b	$7C,$31,$9A,$DD,$FC,$08,$DD,$E6,$3F,$D5,$35,$C1,$F8,$BF,$51,$FB
	.dc.b	$B7,$FD,$5A,$D7,$FC,$B0,$DC,$3C,$AF,$DA,$8D,$ED,$E0,$51,$BB,$FE
	.dc.b	$40,$D8,$EF,$96,$FE,$A3,$FA,$04,$7F,$AF,$1A,$FF,$99,$DB,$87,$C4
	.dc.b	$FA,$7F,$F6,$86,$1F,$CA,$E7,$2F,$F2,$D8,$FE,$00,$DC,$FB,$74,$F7
	.dc.b	$F1,$E6,$3D,$DF,$45,$FF,$AE,$71,$FA,$38,$A7,$75,$E3,$77,$F5,$18
	.dc.b	$AF,$27,$37,$23,$19,$F8,$FE,$1F,$E2,$7D,$47,$F8,$BB,$1F,$F6,$CB
	.dc.b	$5F,$F4,$26,$7F,$E8,$C7,$37,$C0,$8F,$FB,$82,$34,$FF,$A2,$E7,$96
	.dc.b	$1F,$88,$F2,$9F,$15,$F4,$DB,$5D,$0E,$5E,$46,$7C,$D9,$18,$3F,$DD
	.dc.b	$B3,$FE,$3F,$6F,$99,$CC,$DB,$FE,$9B,$8F,$F1,$FB,$FE,$F4,$FA,$8F
	.dc.b	$A6,$FE,$C1,$AF,$F8,$5E,$7F,$2B,$8B,$E3,$1F,$5F,$86,$D7,$F9,$8D
	.dc.b	$AF,$FA,$F5,$F8,$33,$FE,$D4,$FA,$8F,$F0,$8C,$FF,$9A,$5A,$FF,$B0
	.dc.b	$DB,$1B,$E3,$F0,$FD,$1B,$EA,$6F,$A1,$30,$FF,$39,$4E,$5F,$17,$C0
	.dc.b	$F9,$79,$24,$60,$F0,$7C,$6F,$B3,$3F,$F2,$84,$78,$FC,$CC,$4F,$C0
	.dc.b	$F9,$9F,$81,$3F,$E7,$F6,$BF,$ED,$96,$EB,$FE,$F0,$7E,$CC,$7C,$1F
	.dc.b	$B8,$F8,$C4,$7E,$37,$5F,$87,$D9,$91,$F0,$DF,$27,$B3,$E1,$FB,$0F
	.dc.b	$17,$AF,$2B,$C5,$1F,$47,$E8,$FE,$97,$87,$27,$E9,$47,$B7,$E9,$BD
	.dc.b	$DF,$47,$27,$E8,$47,$B7,$C8,$91,$ED,$7E,$07,$9F,$FC,$17,$E4,$1F
	.dc.b	$39,$FC,$1F,$E9,$7F,$8F,$F1,$FF,$20,$6F,$91,$FE,$27,$AD,$E3,$FC
	.dc.b	$EA,$BF,$92,$C6,$23,$FD,$D9,$5C,$1F,$E7,$ED,$51,$B9,$E3,$7E,$1B
	.dc.b	$95,$D5,$F0,$47,$97,$29,$39,$3A,$BE,$9C,$F3,$1E,$28,$B7,$EB,$9F
	.dc.b	$E4,$9F,$4D,$FB,$6C,$3F,$C8,$4D,$BF,$EA,$AD,$59,$7F,$11,$7C,$02
	.dc.b	$35,$0F,$85,$F9,$47,$6E,$AF,$E6,$58,$3E,$1A,$FC,$34,$C7,$9C,$E5
	.dc.b	$B3,$95,$9F,$F5,$29,$57,$4A,$95,$CA,$38,$56,$E7,$C3,$EF,$4A,$EE
	.dc.b	$F9,$3E,$8F,$11,$DB,$5F,$F3,$26,$3F,$E8,$48,$D9,$E6,$9F,$E9,$9F
	.dc.b	$3A,$F8,$CA,$49,$F5,$B5,$E8,$7F,$80,$2B,$F5,$39,$7A,$FC,$DD,$3F
	.dc.b	$DB,$BE,$99,$CF,$FD,$65,$B7,$F4,$1A,$F4,$0F,$F4,$E7,$FC,$1A,$EA
	.dc.b	$7A,$4F,$33,$A7,$45,$AE,$91,$5E,$95,$CD,$EB,$F0,$3E,$3F,$A2,$91
	.dc.b	$E3,$8F,$31,$87,$8C,$61,$FC,$3E,$9F,$61,$FF,$2F,$87,$0F,$F9,$89
	.dc.b	$AF,$F9,$A5,$AF,$FA,$5D,$DE,$3F,$E1,$4F,$A6,$FD,$61,$FA,$D8,$CE
	.dc.b	$FF,$E9,$D8,$73,$FF,$80,$DB,$7F,$9A,$6B,$F2,$D6,$DB,$99,$3B,$79
	.dc.b	$46,$23,$84,$5C,$3C,$93,$DF,$E5,$7C,$1C,$D2,$FC,$11,$97,$C5,$19
	.dc.b	$7C,$51,$FE,$19,$FA,$67,$81,$FE,$1B,$FA,$6F,$6D,$7F,$92,$9A,$A7
	.dc.b	$6F,$ED,$9A,$FE,$1E,$36,$70,$8C,$E3,$FE,$3E,$8F,$D0,$4E,$F9,$7C
	.dc.b	$07,$87,$9A,$47,$F0,$29,$4E,$A3,$13,$A7,$91,$B8,$C5,$77,$9E,$13
	.dc.b	$A3,$CC,$EB,$6F,$F6,$B5,$F6,$9E,$47,$62,$1C,$6B,$D5,$7C,$63,$E3
	.dc.b	$5F,$69,$CC,$70,$7C,$4F,$43,$95,$F8,$27,$94,$2F,$C4,$E8,$EA,$73
	.dc.b	$9F,$3D,$F7,$79,$46,$4F,$B1,$19,$78,$9F,$00,$F1,$5F,$99,$C1,$1C
	.dc.b	$12,$7A,$C6,$5F,$0B,$C9,$CC,$7B,$A3,$D0,$7A,$E7,$DB,$5F,$80,$9E
	.dc.b	$93,$E6,$3D,$07,$C1,$BF,$43,$C4,$CC,$F1,$BF,$0C,$76,$8C,$3E,$EA
	.dc.b	$EA,$F1,$3D,$06,$67,$8C,$6A,$78,$CF,$84,$F6,$1E,$53,$89,$93,$C1
	.dc.b	$7E,$CB,$F3,$CF,$D8,$AF,$84,$4B,$EE,$4B,$F1,$10,$F3,$47,$8E,$BE
	.dc.b	$3D,$F4,$6C,$F9,$0F,$14,$F5,$7C,$CF,$17,$E3,$CF,$16,$9C,$6F,$DE
	.dc.b	$48,$D4,$7B,$A3,$D5,$7E,$78,$EE,$FD,$E3,$11,$A8,$F5,$3C,$DE,$D7
	.dc.b	$E8,$7E,$B3,$6F,$CD,$1D,$EF,$C3,$1E,$B8,$F8,$77,$DA,$B8,$36,$3D
	.dc.b	$67,$42,$CF,$43,$95,$F9,$63,$99,$E6,$AF,$AF,$7F,$66,$3E,$23,$5F
	.dc.b	$BA,$7E,$42,$36,$D3,$65,$36,$7C,$97,$6D,$36,$7C,$DF,$3B,$F9,$71
	.dc.b	$1F,$E3,$DC,$D5,$FE,$55,$17,$0F,$E5,$13,$FE,$CF,$FA,$67,$8F,$FD
	.dc.b	$7A,$D1,$87,$FB,$0B,$51,$0A,$E5,$C1,$CA,$3F,$DB,$9E,$5F,$F7,$53
	.dc.b	$FB,$B4,$F3,$BF,$F7,$BC,$39,$B6,$4B,$FF,$1C,$C3,$9B,$11,$FE,$0B
	.dc.b	$8E,$0E,$A7,$BC,$EA,$BA,$C5,$DB,$D1,$7C,$EF,$AF,$81,$E6,$23,$FE
	.dc.b	$48,$FA,$6F,$F3,$14,$FF,$98,$6B,$8D,$6A,$BF,$C3,$95,$FE,$67,$FA
	.dc.b	$57,$D2,$3E,$2F,$FC,$13,$18,$6F,$5B,$E9,$73,$5B,$C3,$E4,$8F,$AC
	.dc.b	$E5,$9F,$F8,$49,$B1,$A6,$BF,$41,$AF,$F9,$FF,$E8,$BF,$D8,$D3,$FE
	.dc.b	$71,$DD,$18,$9F,$F1,$FC,$63,$8D,$5B,$AD,$4E,$2B,$FC,$BB,$18,$9E
	.dc.b	$DC,$99,$A9,$3F,$EE,$A9,$E5,$39,$76,$60,$ED,$79,$BE,$B3,$6D,$F3
	.dc.b	$EF,$A6,$C3,$70,$F6,$7B,$DF,$12,$BC,$53,$9C,$3D,$AD,$8C,$DF,$5A
	.dc.b	$E1,$7F,$EA,$5A,$D3,$F9,$DF,$A5,$7A,$9E,$3F,$A6,$7C,$1F,$9F,$35
	.dc.b	$6A,$FE,$75,$69,$FF,$48,$C6,$67,$FC,$C4,$DC,$FE,$3F,$93,$CE,$31
	.dc.b	$5E,$3F,$A6,$F7,$9F,$03,$E1,$AC,$7D,$BD,$13,$8C,$7D,$F5,$71,$F8
	.dc.b	$73,$6F,$DD,$36,$BE,$DF,$D3,$7C,$FE,$27,$C2,$58,$C3,$6F,$F0,$8D
	.dc.b	$AF,$C4,$D7,$09,$FF,$57,$FD,$33,$E4,$7E,$D7,$AA,$F9,$AF,$91,$39
	.dc.b	$AF,$CA,$4E,$5E,$35,$A3,$EF,$FD,$35,$8F,$CA,$7B,$7E,$7C,$3D,$0D
	.dc.b	$8C,$9F,$99,$8C,$B4,$CE,$5E,$88,$CB,$63,$F4,$DF,$4D,$FE,$7F,$6F
	.dc.b	$67,$E4,$E3,$13,$FA,$87,$F7,$F3,$FA,$DE,$47,$C3,$7E,$65,$FC,$15
	.dc.b	$7E,$B9,$CC,$EF,$9F,$F3,$DF,$48,$FA,$AA,$71,$BB,$5F,$B9,$FA,$5E
	.dc.b	$31,$FB,$86,$F3,$FE,$C5,$EE,$DC,$BF,$55,$F4,$BF,$B7,$AF,$F6,$AB
	.dc.b	$64,$FE,$DE,$3F,$81,$B9,$1F,$01,$CB,$E0,$39,$B7,$1F,$E1,$9F,$49
	.dc.b	$CC,$E5,$87,$F0,$37,$1E,$1F,$EE,$D7,$C6,$5F,$A5,$C3,$6F,$F8,$B7
	.dc.b	$B0,$FA,$0F,$E0,$4D,$8C,$FC,$76,$27,$F8,$37,$D2,$F4,$9F,$E3,$CD
	.dc.b	$DF,$F8,$EA,$C7,$78,$FF,$0D,$FA,$5F,$D7,$D7,$FB,$E1,$BA,$97,$C6
	.dc.b	$6F,$C8,$FE,$29,$E1,$3D,$2E,$5A,$61,$AF,$F6,$6C,$3F,$94,$63,$FA
	.dc.b	$1F,$92,$FE,$11,$DC,$7F,$85,$70,$79,$EB,$73,$9B,$E4,$7F,$6F,$FA
	.dc.b	$47,$99,$FC,$45,$B1,$AC,$7F,$BA,$C3,$E2,$35,$5B,$F9,$19,$7B,$B7
	.dc.b	$D1,$BC,$4C,$C7,$FB,$DF,$E9,$3E,$83,$79,$FF,$90,$53,$DE,$FE,$45
	.dc.b	$5A,$DB,$DF,$39,$C7,$B3,$70,$4F,$C4,$F5,$DE,$77,$4C,$7F,$96,$46
	.dc.b	$30,$F3,$36,$FC,$9C,$DA,$BF,$E6,$5F,$44,$FA,$3F,$E9,$1F,$49,$FE
	.dc.b	$0B,$C7,$DD,$39,$7F,$B1,$46,$70,$F1,$C7,$F8,$6E,$75,$1F,$D9,$27
	.dc.b	$59,$FD,$A9,$CB,$FD,$7E,$33,$5C,$6F,$C6,$78,$BC,$8B,$1E,$2E,$3F
	.dc.b	$EB,$9F,$49,$FC,$87,$2F,$09,$FE,$2C,$CB,$EC,$57,$F7,$79,$D6,$5E
	.dc.b	$1B,$E3,$E7,$7C,$06,$D7,$BB,$E8,$3F,$AE,$78,$BF,$11,$B1,$92,$65
	.dc.b	$E0,$19,$7F,$8F,$23,$38,$FE,$60,$FE,$02,$FE,$1D,$78,$DF,$D8,$5C
	.dc.b	$D7,$37,$33,$BC,$34,$7F,$67,$9F,$F2,$4F,$D2,$7F,$92,$BE,$89,$F2
	.dc.b	$1F,$E4,$BC,$39,$9F,$A8,$AD,$89,$DB,$FA,$DB,$FF,$0A,$46,$23,$FC
	.dc.b	$A9,$7D,$6F,$A5,$69,$FF,$16,$C6,$67,$96,$15,$E2,$77,$9E,$31,$98
	.dc.b	$F1,$DF,$7C,$3E,$0B,$E1,$8C,$D7,$07,$91,$FE,$64,$BE,$D5,$C2,$7C
	.dc.b	$73,$A3,$F4,$B8,$79,$DE,$06,$67,$83,$8B,$F1,$4E,$E3,$D2,$7F,$92
	.dc.b	$85,$FF,$9F,$5A,$7F,$9F,$DD,$7D,$13,$E8,$DF,$49,$86,$D7,$F9,$52
	.dc.b	$75,$39,$38,$FD,$1F,$EF,$27,$4D,$7F,$D0,$72,$8C,$47,$FA,$26,$31
	.dc.b	$97,$23,$E9,$66,$DF,$FD,$1D,$A7,$DD,$8C,$5F,$D7,$BE,$2F,$C9,$6C
	.dc.b	$EE,$B8,$D7,$FA,$5C,$E3,$5B,$C3,$D1,$5F,$89,$BE,$35,$BB,$E3,$7F
	.dc.b	$EA,$08,$E3,$5B,$BE,$31,$EB,$6C,$EE,$B8,$BF,$9E,$6D,$F6,$8F,$F5
	.dc.b	$5B,$5D,$1F,$0A,$37,$3D,$23,$27,$EE,$6B,$D7,$87,$F4,$E6,$EB,$E3
	.dc.b	$8E,$53,$FC,$FE,$72,$F9,$4F,$E8,$58,$78,$7E,$93,$FB,$FC,$7F,$45
	.dc.b	$6B,$FC,$1D,$B5,$FD,$71,$FB,$91,$BD,$BF,$75,$87,$FB,$16,$76,$D8
	.dc.b	$CC,$FF,$B2,$A7,$8B,$FC,$4F,$E9,$3F,$A1,$BF,$D5,$5A,$E7,$A3,$8A
	.dc.b	$FA,$47,$83,$8A,$DE,$5F,$DC,$27,$F6,$D9,$7F,$B5,$DB,$7F,$ED,$AA
	.dc.b	$DC,$FF,$B6,$D7,$8B,$CC,$93,$FD,$9E,$9F,$3D,$88,$F4,$4D,$C3,$C1
	.dc.b	$37,$0F,$07,$3F,$A3,$0F,$EE,$98,$7B,$70,$C7,$57,$B3,$E9,$33,$3F
	.dc.b	$CD,$5B,$97,$AE,$75,$84,$AF,$EF,$55,$AC,$7D,$D5,$99,$FE,$7D,$E4
	.dc.b	$3D,$2F,$92,$D8,$E3,$87,$97,$C7,$CF,$97,$27,$E0,$45,$79,$F9,$DC
	.dc.b	$8C,$69,$FE,$1E,$C3,$FC,$7B,$3F,$E0,$DF,$A4,$F9,$A8,$38,$7F,$C4
	.dc.b	$55,$F8,$4A,$7C,$37,$F9,$AA,$7F,$DB,$0D,$7F,$DA,$ED,$8F,$F6,$83
	.dc.b	$FE,$B4,$8F,$F6,$F7,$D1,$7F,$92,$5B,$E5,$F5,$8C,$47,$FB,$81,$FB
	.dc.b	$EF,$FB,$97,$E8,$DF,$49,$E3,$7F,$1B,$19,$E7,$C4,$F9,$F0,$7C,$7D
	.dc.b	$E0,$97,$E8,$8F,$07,$A7,$E1,$F2,$E4,$F9,$A3,$E0,$7C,$9F,$C0,$8F
	.dc.b	$F7,$B3,$5F,$F7,$AB,$76,$FF,$77,$3F,$66,$3E,$1F,$D2,$78,$C8,$FC
	.dc.b	$8E,$BF,$81,$F4,$92,$3E,$2B,$F0,$FE,$67,$C3,$F4,$9E,$2F,$56,$57
	.dc.b	$8A,$3D,$DF,$41,$EF,$F1,$A4,$FD,$48,$F9,$9F,$3D,$F3,$7E,$7A,$4F
	.dc.b	$D2,$8F,$6F,$8F,$23,$DA,$FC,$0F,$3F,$F7,$7F,$15,$FE,$6B,$9F,$F2
	.dc.b	$A7,$D1,$FF,$84,$7C,$4F,$C4,$42,$30,$FE,$2E,$78,$3F,$E1,$B6,$F1
	.dc.b	$FF,$9C,$27,$73,$C6,$BB,$FB,$B2,$3B,$C7,$F0,$C9,$DD,$F2,$8F,$0C
	.dc.b	$32,$D1,$2B,$CF,$9D,$7F,$C5,$93,$B7,$D7,$3D,$67,$F8,$CF,$D2,$7E
	.dc.b	$DE,$7F,$48,$D4,$DA,$FC,$AC,$7F,$8D,$E3,$F8,$2B,$F1,$0D,$F4,$FF
	.dc.b	$7F,$69,$FE,$01,$8D,$BF,$2A,$9F,$47,$FC,$A9,$7D,$25,$31,$12,$7F
	.dc.b	$D0,$D3,$A6,$CE,$5F,$F2,$54,$74,$72,$43,$33,$DF,$DC,$91,$D1,$F1
	.dc.b	$7C,$FE,$23,$9C,$5D,$FF,$D5,$36,$AF,$E8,$3F,$34,$F8,$DA,$4C,$3C
	.dc.b	$ED,$7E,$30,$8E,$0F,$F8,$95,$AE,$9A,$FC,$DF,$A4,$72,$FF,$50,$6C
	.dc.b	$7F,$AA,$5A,$FF,$9A,$DF,$E9,$A9,$CE,$3F,$EC,$26,$F0,$FE,$71,$B3
	.dc.b	$FC,$95,$E2,$61,$E5,$5E,$13,$C5,$CD,$E3,$F9,$EF,$27,$CF,$CA,$F2
	.dc.b	$46,$5F,$34,$65,$F3,$47,$CA,$E0,$91,$FE,$C3,$CF,$EF,$3E,$BF,$A5
	.dc.b	$CB,$F1,$67,$F1,$D3,$A8,$FC,$93,$5F,$DB,$FD,$2F,$E8,$5F,$F4,$73
	.dc.b	$75,$FF,$71,$B7,$5D,$F0,$7F,$7C,$8C,$57,$F4,$76,$A6,$88,$E1,$CD
	.dc.b	$27,$FD,$B9,$3A,$9E,$9F,$4B,$D7,$E3,$65,$6A,$32,$F0,$8C,$BE,$88
	.dc.b	$FF,$07,$FD,$2B,$C0,$A4,$9F,$F7,$57,$D2,$BE,$AF,$11,$FE,$AD,$AE
	.dc.b	$0F,$FB,$BD,$AF,$FA,$C6,$78,$3C,$BA,$3F,$C3,$B1,$F9,$F8,$EB,$D1
	.dc.b	$89,$91,$82,$1C,$DC,$BD,$63,$99,$8D,$DB,$FF,$2C,$BF,$F7,$D5,$72
	.dc.b	$9C,$BF,$DD,$E3,$C4,$F5,$3C,$C6,$E7,$95,$75,$7A,$5F,$82,$74,$73
	.dc.b	$74,$F5,$9D,$3D,$49,$7E,$25,$7E,$28,$C3,$6F,$BC,$F9,$0F,$80,$78
	.dc.b	$0E,$37,$E0,$79,$DF,$82,$F9,$CE,$6F,$A5,$79,$4E,$93,$A7,$A4,$69
	.dc.b	$EA,$6A,$FE,$1B,$AA,$E9,$7D,$E7,$CF,$3E,$E7,$6F,$A0,$C1,$F1,$CE
	.dc.b	$F1,$E5,$9E,$51,$E5,$36,$78,$8E,$57,$F1,$A7,$84,$72,$3C,$C7,$53
	.dc.b	$75,$E2,$9F,$74,$74,$BF,$71,$2F,$D7,$1E,$D3,$99,$F3,$0C,$1C,$6F
	.dc.b	$CE,$F9,$6B,$A9,$E5,$76,$E2,$FC,$44,$8F,$34,$65,$F8,$0F,$17,$CD
	.dc.b	$5D,$AF,$CD,$1E,$47,$D3,$7F,$0A,$3D,$E7,$2B,$FA,$77,$D6,$3D,$46
	.dc.b	$8F,$44,$68,$F2,$3A,$8E,$15,$FA,$A8,$E0,$F8,$9F,$84,$7B,$A7,$93
	.dc.b	$E1,$21,$D8,$F7,$BD,$A3,$94,$F6,$AF,$75,$FD,$2B,$EE,$D9,$F8,$87
	.dc.b	$91,$F8,$93,$F0,$63,$37,$D6,$78,$36,$BE,$45,$76,$69,$E9,$3C,$A7
	.dc.b	$A2,$FC,$B3,$EC,$8F,$88,$F3,$7E,$09,$E6,$39,$94,$CD,$F9,$1E,$A7
	.dc.b	$C3,$39,$9D,$EF,$F9,$2F,$95,$23,$F9,$8B,$FC,$A3,$D5,$C4,$7F,$2D
	.dc.b	$8B,$5F,$C2,$27,$FD,$93,$F4,$CF,$26,$92,$FF,$D2,$D5,$FC,$47,$CC
	.dc.b	$FF,$40,$C2,$3F,$8A,$57,$18,$DE,$5F,$E5,$68,$C4,$6E,$31,$7F,$ED
	.dc.b	$8B,$E7,$7D,$7A,$30,$4B,$E6,$7F,$A9,$E5,$D3,$98,$E8,$66,$71,$5E
	.dc.b	$79,$B8,$7F,$9A,$2D,$58,$BE,$95,$E3,$F0,$E4,$78,$0F,$F4,$DB,$FE
	.dc.b	$EE,$FA,$6F,$F3,$4D,$FF,$98,$E7,$C9,$3C,$1F,$18,$9F,$F3,$7F,$D4
	.dc.b	$FF,$B9,$A3,$FC,$62,$1C,$3F,$E3,$46,$BF,$E5,$B6,$BF,$E8,$0F,$A7
	.dc.b	$7C,$E3,$E3,$B9,$C6,$1B,$DA,$FA,$7C,$D6,$EF,$FC,$CD,$BB,$E8,$3F
	.dc.b	$D9,$BC,$1F,$3A,$31,$C9,$FD,$81,$93,$FE,$E7,$9D,$3C,$2B,$E8,$CF
	.dc.b	$09,$CD,$71,$7F,$13,$5C,$6B,$6F,$1A,$EB,$5B,$9F,$C6,$4E,$6B,$8F
	.dc.b	$26,$6F,$B5,$78,$3E,$8D,$F4,$D9,$6D,$FA,$FD,$EF,$8C,$FB,$63,$35
	.dc.b	$C6,$B8,$1D,$6B,$71,$9A,$FC,$17,$D2,$BD,$6F,$17,$D3,$3E,$0B,$D1
	.dc.b	$6A,$D5,$FC,$BA,$D3,$CA,$FF,$BD,$57,$FA,$EA,$B6,$FF,$9D,$1A,$FF
	.dc.b	$A6,$DA,$FE,$CB,$27,$D3,$7B,$CF,$81,$F1,$17,$0F,$74,$5A,$F8,$D3
	.dc.b	$75,$FB,$97,$CA,$F6,$7C,$59,$D3,$F6,$3E,$9B,$E8,$70,$3E,$16,$C6
	.dc.b	$1B,$3F,$84,$6E,$3F,$79,$F8,$BF,$4C,$F9,$4F,$B5,$EB,$BE,$6F,$31
	.dc.b	$86,$E1,$FA,$6A,$E0,$D9,$F9,$38,$04,$E9,$FC,$EE,$4F,$73,$D3,$87
	.dc.b	$A9,$B8,$7E,$69,$B5,$F9,$46,$ED,$FA,$06,$CF,$E6,$9B,$3F,$C8,$7E
	.dc.b	$97,$FD,$55,$3F,$A9,$7F,$6A,$DE,$BF,$CC,$46,$27,$F4,$FE,$B3,$E8
	.dc.b	$BF,$38,$FE,$1A,$33,$18,$E7,$FD,$07,$D2,$BE,$96,$9C,$2A,$E5,$FB
	.dc.b	$9F,$A5,$E8,$DC,$3F,$51,$1D,$75,$FD,$E3,$AA,$FD,$CF,$D2,$FE,$BD
	.dc.b	$FE,$78,$D9,$3F,$B9,$7F,$82,$39,$1F,$03,$FB,$55,$F0,$36,$30,$FE
	.dc.b	$E7,$E9,$7A,$1C,$F2,$FE,$9A,$E3,$C1,$D7,$C6,$EF,$E9,$2D,$7F,$15
	.dc.b	$A4,$7F,$03,$F6,$1F,$1D,$4C,$7C,$95,$BF,$E3,$9F,$4A,$C7,$F8,$53
	.dc.b	$63,$38,$62,$7F,$87,$7D,$2F,$ED,$37,$FE,$44,$DA,$FE,$C6,$BE,$37
	.dc.b	$7E,$37,$F1,$5F,$05,$E8,$F2,$BC,$35,$FE,$99,$DC,$F9,$3F,$E1,$2D
	.dc.b	$8F,$EA,$1B,$79,$79,$BF,$A5,$CE,$5F,$E9,$BF,$4A,$F2,$FF,$88,$B6
	.dc.b	$FF,$AC,$2F,$8A,$CC,$EB,$E4,$65,$7F,$5D,$FA,$4F,$D3,$3C,$23,$F8
	.dc.b	$DF,$D2,$FC,$C8,$FE,$C4,$DC,$7F,$AE,$D3,$DD,$DF,$3F,$F2,$28,$C5
	.dc.b	$F5,$9F,$E4,$BE,$AB,$CD,$D4,$61,$93,$DB,$92,$E3,$FD,$AB,$E8,$9F
	.dc.b	$49,$FD,$07,$E9,$7F,$BC,$DF,$F7,$59,$CB,$FD,$66,$33,$5C,$6B,$87
	.dc.b	$BD,$FC,$DA,$F7,$3F,$D6,$2B,$7D,$5F,$EA,$D9,$CB,$6F,$C1,$E4,$71
	.dc.b	$79,$71,$FF,$53,$FA,$5F,$E2,$15,$C6,$31,$E8,$6B,$4F,$ED,$95,$A7
	.dc.b	$F8,$C6,$6C,$BE,$B4,$6A,$BC,$DF,$43,$FD,$4F,$C6,$B3,$FD,$6D,$B7
	.dc.b	$F1,$67,$D8,$E3,$2F,$F6,$31,$FB,$4A,$FF,$0D,$3F,$B4,$6D,$7F,$58
	.dc.b	$8C,$4F,$F8,$61,$E0,$D8,$FE,$BA,$FF,$2F,$FA,$5F,$E6,$1F,$46,$F9
	.dc.b	$0F,$5D,$F0,$3F,$9A,$CE,$A3,$37,$FE,$EA,$AE,$17,$F0,$85,$6D,$FF
	.dc.b	$27,$CE,$EB,$51,$A1,$5C,$9F,$F2,$94,$ED,$E9,$5C,$DC,$E1,$5F,$51
	.dc.b	$CA,$FC,$B1,$98,$ED,$7D,$70,$F8,$6F,$78,$E2,$E7,$0F,$04,$7D,$CC
	.dc.b	$3C,$15,$C3,$0F,$4C,$62,$B7,$5C,$A7,$C2,$F8,$CF,$49,$FE,$46,$3E
	.dc.b	$01,$FE,$48,$9C,$38,$3F,$17,$F4,$4F,$A5,$7D,$2F,$F9,$3A,$30,$D8
	.dc.b	$CC,$FC,$C9,$CB,$7E,$97,$FC,$F8,$DC,$B9,$46,$23,$EC,$4A,$31,$1E
	.dc.b	$27,$8E,$7F,$57,$26,$BF,$A0,$8C,$36,$38,$61,$F5,$F0,$E8,$FD,$98
	.dc.b	$E3,$85,$AE,$0F,$F8,$06,$B6,$FC,$90,$E2,$FB,$C7,$13,$B6,$1E,$6B
	.dc.b	$F0,$C7,$17,$BB,$6F,$BB,$D2,$78,$3E,$16,$BF,$A3,$8D,$CF,$2A,$FC
	.dc.b	$B5,$F8,$E3,$F8,$09,$DB,$06,$B8,$BF,$D5,$B7,$0F,$F4,$95,$F1,$7F
	.dc.b	$C4,$17,$C6,$75,$1F,$E9,$08,$C3,$E5,$BF,$D1,$7D,$2F,$F7,$68,$FF
	.dc.b	$4E,$35,$FF,$4B,$B6,$3C,$71,$FE,$2D,$8F,$BC,$F0,$DB,$CA,$E5,$AA
	.dc.b	$39,$5F,$F1,$E7,$8C,$7F,$A9,$3E,$97,$FA,$13,$FE,$AD,$6B,$9E,$9E
	.dc.b	$7F,$4B,$FE,$C4,$BF,$E4,$F5,$AC,$7C,$B6,$AF,$DB,$57,$F2,$C6,$C7
	.dc.b	$F9,$16,$B8,$4F,$2F,$13,$CF,$06,$9E,$0F,$9F,$B5,$E5,$8A,$FF,$AF
	.dc.b	$2F,$FD,$6F,$87,$8A,$F9,$4F,$FB,$1E,$7F,$B8,$37,$9F,$8D,$7A,$27
	.dc.b	$FD,$11,$3F,$EC,$DF,$A5,$DB,$FE,$C8,$6C,$EF,$AB,$E6,$C6,$33,$FE
	.dc.b	$92,$DA,$FF,$31,$35,$F8,$93,$C3,$0F,$AB,$E3,$BD,$1A,$66,$BD,$1E
	.dc.b	$36,$BC,$A9,$18,$3F,$DC,$55,$FE,$81,$6F,$9B,$D0,$DB,$FE,$E0,$8F
	.dc.b	$F1,$73,$FE,$E9,$FA,$5F,$A2,$FF,$9D,$DA,$FF,$88,$67,$F1,$3F,$80
	.dc.b	$7F,$BD,$23,$FD,$A5,$F0,$FF,$E1,$D6,$E3,$AF,$13,$F1,$FF,$73,$FE
	.dc.b	$69,$9F,$F7,$B7,$43,$DE,$FC,$9B,$63,$FD,$F5,$3B,$7F,$DE,$BF,$73
	.dc.b	$FE,$6B,$78,$F9,$1C,$0F,$CD,$C6,$63,$FD,$F7,$F3,$0F,$95,$FD,$0F
	.dc.b	$3B,$5F,$F0,$AD,$7A,$3C,$9F,$D1,$DE,$E3,$FD,$2D,$5F,$98,$C3,$FC
	.dc.b	$B3,$18,$E8,$FF,$01,$FD,$CB,$E8,$7F,$56,$D9,$D1,$FE,$9F,$AE,$2A
	.dc.b	$FF,$47,$EC,$7E,$4A,$31,$87,$4D,$39,$9F,$B6,$9D,$57,$EA,$B0,$F7
	.dc.b	$3F,$EF,$FD,$1F,$29,$EE,$6E,$3E,$21,$5C,$31,$FF,$0D,$FD,$CF,$C2
	.dc.b	$9F,$09,$FE,$B5,$FB,$87,$C9,$7F,$06,$6C,$73,$9E,$19,$7E,$F1,$7D
	.dc.b	$4E,$FE,$5E,$56,$E3,$97,$A1,$E5,$E8,$93,$E5,$C7,$FD,$9C,$F9,$FE
	.dc.b	$E7,$FB,$23,$9F,$77,$C4,$7F,$18,$8C,$C7,$A0,$7D,$CF,$FB,$B9,$F8
	.dc.b	$1F,$72,$FA,$3F,$E3,$ED,$8F,$89,$3B,$79,$7E,$52,$FB,$3F,$E4,$AD
	.dc.b	$8F,$94,$D7,$C1,$F2,$7F,$C1,$1E,$33,$FC,$B2,$3E,$8C,$6B,$9B,$9C
	.dc.b	$C8,$FA,$93,$AE,$2F,$E4,$AD,$E4,$FF,$27,$E1,$DE,$73,$AF,$F3,$8F
	.dc.b	$AB,$FF,$23,$D7,$F9,$51,$B8,$7F,$BB,$70,$FF,$4D,$BD,$AB,$85,$48
	.dc.b	$C1,$F8,$5A,$F8,$8F,$CB,$6D,$FF,$38,$6C,$F7,$FA,$5F,$16,$D2,$F9
	.dc.b	$3E,$58,$FE,$9B,$F5,$7F,$C7,$49,$E5,$FA,$7D,$0C,$BE,$BF,$8A,$98
	.dc.b	$7E,$5E,$B6,$F9,$63,$2F,$DF,$8F,$E5,$17,$FA,$F8,$FC,$12,$FA,$8F
	.dc.b	$EB,$AD,$8F,$D1,$3B,$D3,$FA,$9B,$73,$FF,$3E,$BB,$9E,$75,$CB,$C1
	.dc.b	$E7,$F0,$32,$79,$BD,$BC,$1C,$47,$8A,$2E,$3D,$A7,$FB,$7F,$90,$F9
	.dc.b	$3A,$4B,$F2,$B5,$F5,$FA,$1E,$A3,$17,$FD,$AA,$B9,$4F,$B2,$74,$FE
	.dc.b	$FE,$32,$FE,$97,$EA,$DF,$57,$FE,$00,$6C,$7F,$05,$6B,$BF,$6B,$53
	.dc.b	$23,$4F,$18,$E5,$DF,$CF,$AE,$57,$38,$CB,$D6,$32,$F8,$E3,$E1,$EF
	.dc.b	$FE,$67,$C7,$E8,$4F,$FA,$35,$AF,$FA,$C1,$A8,$FE,$91,$1F,$E1,$AF
	.dc.b	$AA,$FD,$0C,$1B,$3F,$6F,$4F,$91,$5B,$E2,$97,$F1,$E3,$97,$3C,$8F
	.dc.b	$E5,$18,$7E,$81,$F0,$2E,$3D,$47,$F1,$18,$D3,$E1,$9F,$24,$F9,$7C
	.dc.b	$AF,$37,$34,$AE,$B1,$97,$BC,$65,$E9,$1F,$E3,$6F,$A8,$7D,$57,$F8
	.dc.b	$E3,$EA,$7D,$9A,$7F,$B5,$1B,$7F,$16,$B6,$FF,$41,$6B,$FC,$9E,$36
	.dc.b	$9B,$1A,$D7,$FC,$6E,$DA,$ED,$AF,$84,$F2,$61,$30,$FE,$9D,$87,$43
	.dc.b	$9C,$66,$3C,$13,$C5,$F2,$46,$70,$FF,$22,$D7,$1A,$C5,$FF,$BC,$30
	.dc.b	$F1,$61,$F8,$1A,$D3,$C8,$87,$63,$C7,$5C,$9F,$89,$87,$85,$D1,$FD
	.dc.b	$26,$BA,$A6,$D7,$68,$D4,$F3,$7E,$45,$73,$C3,$E3,$CE,$6B,$84,$62
	.dc.b	$BC,$AF,$BA,$B8,$61,$ED,$BE,$11,$8A,$F3,$9C,$27,$E8,$BE,$C3,$06
	.dc.b	$8F,$5D,$75,$72,$7A,$EB,$A3,$E8,$33,$87,$BA,$F2,$FC,$23,$11,$DC
	.dc.b	$F8,$27,$4C,$3D,$B8,$7A,$63,$15,$CA,$FD,$71,$C7,$0F,$84,$F2,$3C
	.dc.b	$F5,$EC,$71,$1A,$3D,$4E,$8F,$73,$98,$E0,$E8,$C4,$6A,$32,$47,$04
	.dc.b	$3E,$11,$D4,$F8,$C7,$A8,$F4,$3E,$D7,$CA,$F3,$78,$C7,$81,$E2,$73
	.dc.b	$BF,$D0,$BC,$CE,$D8,$7C,$DB,$FA,$73,$A9,$EA,$7E,$77,$0F,$94,$F1
	.dc.b	$BF,$76,$1E,$37,$C1,$5F,$4A,$7D,$98,$7D,$87,$BE,$1F,$49,$A7,$3B
	.dc.b	$F5,$4F,$8E,$3D,$E7,$84,$FC,$2C,$F9,$B0,$FD,$04,$F5,$C3,$C7,$5E
	.dc.b	$0A,$D3,$C0,$E6,$7C,$6B,$F4,$11,$EC,$79,$0F,$82,$6C,$CC,$FC,$DA
	.dc.b	$F0,$4F,$A2,$BD,$93,$E2,$AF,$A2,$FC,$4B,$F9,$31,$D3,$0F,$8C,$EA
	.dc.b	$FC,$4F,$72,$61,$FC,$03,$9F,$F9,$77,$3C,$9F,$E3,$B1,$FC,$C4,$C3
	.dc.b	$FC,$AE,$30,$FF,$22,$F2,$FF,$9B,$11,$FE,$4B,$CF,$5F,$E6,$51,$73
	.dc.b	$FE,$43,$3F,$EF,$4F,$A9,$7D,$4F,$FB,$DB,$CA,$FF,$72,$E0,$6C,$EA
	.dc.b	$3F,$CF,$CD,$7F,$CC,$91,$C0,$DC,$F5,$3A,$74,$49,$E3,$1D,$F9,$E4
	.dc.b	$7F,$A2,$6B,$FD,$7C,$F4,$E8,$CB,$DA,$6E,$5F,$EF,$EA,$FF,$2D,$4F
	.dc.b	$F7,$77,$FC,$A6,$F9,$3C,$1F,$61,$FD,$ED,$FF,$3D,$7D,$4F,$F7,$A9
	.dc.b	$DD,$73,$AD,$D7,$FA,$6E,$BF,$BF,$FD,$43,$EA,$1F,$1D,$CE,$BA,$4E
	.dc.b	$63,$11,$FE,$09,$8F,$F6,$87,$A1,$88,$46,$DD,$FD,$97,$4E,$6D,$36
	.dc.b	$33,$69,$CB,$63,$BF,$D9,$7F,$B3,$E7,$B3,$63,$3C,$9B,$7D,$11,$97
	.dc.b	$E2,$6A,$FA,$3E,$0F,$53,$15,$6F,$FD,$AF,$2E,$5C,$5F,$FB,$3E,$38
	.dc.b	$CE,$AB,$37,$CA,$3F,$DA,$EF,$7B,$CD,$70,$BF,$F6,$C7,$D3,$3E,$AB
	.dc.b	$8B,$70,$E3,$EE,$FC,$5C,$7C,$D3,$9B,$FF,$66,$C2,$30,$EA,$FC,$75
	.dc.b	$C6,$FF,$DE,$35,$A7,$F0,$DF,$52,$F9,$BF,$4F,$D5,$3E,$1B,$DF,$9A
	.dc.b	$B5,$7E,$CA,$D3,$FE,$E3,$8C,$CF,$FB,$25,$B9,$FC,$DD,$3F,$CE,$D5
	.dc.b	$BC,$7A,$D6,$A3,$C7,$1F,$1B,$CB,$CB,$F1,$3E,$AA,$4F,$CC,$D1,$F1
	.dc.b	$98,$CF,$E1,$C2,$B1,$9F,$C8,$AB,$7F,$73,$EA,$BE,$9F,$C4,$FD,$46
	.dc.b	$C7,$D8,$D5,$18,$CB,$E7,$C6,$27,$55,$F1,$BE,$A9,$EE,$FE,$0B,$D7
	.dc.b	$7C,$FE,$63,$0D,$BF,$F3,$86,$9F,$97,$BE,$31,$ED,$E8,$D6,$7F,$EF
	.dc.b	$39,$CB,$70,$FC,$3C,$EA,$FF,$41,$F3,$5F,$E5,$CC,$3A,$9F,$C0,$DC
	.dc.b	$5F,$E2,$3E,$A5,$F1,$BF,$7A,$73,$7F,$A9,$FB,$F7,$DA,$7E,$09,$B8
	.dc.b	$FE,$73,$99,$E3,$66,$BF,$0F,$E1,$3E,$A6,$9F,$A4,$A4,$62,$FC,$BA
	.dc.b	$FF,$A4,$7F,$03,$FC,$B3,$7C,$2F,$D5,$FD,$57,$6A,$DC,$AB,$6B,$EF
	.dc.b	$3F,$1A,$B6,$FE,$05,$18,$7F,$7D,$F5,$6F,$AC,$FC,$EC,$E7,$0F,$E5
	.dc.b	$0B,$E3,$72,$7E,$F2,$9F,$3D,$8A,$F1,$FE,$20,$FB,$5F,$D3,$B6,$3F
	.dc.b	$80,$FD,$6F,$77,$73,$FA,$CF,$A4,$FA,$31,$EC,$6E,$BF,$BC,$A7,$CB
	.dc.b	$EF,$A7,$F6,$7F,$78,$F9,$6A,$7F,$10,$F8,$17,$C0,$E3,$3F,$8F,$F5
	.dc.b	$4F,$03,$F7,$6D,$8F,$E2,$BF,$5D,$D9,$DC,$F6,$D6,$47,$F1,$AF,$AE
	.dc.b	$7D,$66,$27,$E7,$4F,$F4,$F1,$F5,$1F,$C2,$1A,$71,$F0,$31,$1F,$C3
	.dc.b	$3C,$7F,$7F,$5F,$F3,$76,$D7,$F5,$9F,$90,$7C,$FE,$F8,$FF,$89,$FD
	.dc.b	$73,$D6,$A6,$3E,$4A,$E1,$F8,$69,$7D,$77,$03,$F8,$DF,$D7,$3E,$C7
	.dc.b	$15,$F1,$B8,$7E,$29,$AF,$89,$B1,$F9,$0A,$46,$63,$F2,$7E,$3F,$F1
	.dc.b	$27,$51,$FC,$A2,$31,$C7,$F1,$8F,$F7,$73,$8E,$1F,$EA,$DC,$EF,$A1
	.dc.b	$E1,$EB,$7F,$30,$FA,$EF,$BD,$97,$F4,$AF,$43,$F4,$59,$73,$13,$A7
	.dc.b	$FA,$1C,$6B,$1F,$EC,$53,$97,$FA,$F4,$66,$BF,$C0,$D3,$97,$F8,$9C
	.dc.b	$66,$87,$6E,$EF,$21,$FD,$7A,$3F,$A0,$C7,$8E,$B9,$3C,$17,$17,$F6
	.dc.b	$4F,$AF,$FE,$33,$8F,$92,$31,$E9,$FB,$DB,$97,$EF,$AF,$8B,$FD,$AA
	.dc.b	$F8,$BE,$8F,$AD,$FE,$BB,$E1,$FC,$76,$C7,$17,$0F,$3C,$A9,$F2,$72
	.dc.b	$FD,$03,$7E,$C3,$FA,$FC,$60,$FF,$05,$07,$37,$2B,$F8,$9D,$7C,$6F
	.dc.b	$1B,$E3,$FD,$7B,$E3,$7F,$C1,$F7,$FD,$AA,$B4,$27,$4F,$F4,$28,$FF
	.dc.b	$0F,$46,$23,$FC,$75,$5D,$2B,$9D,$D9,$FF,$1A,$4D,$7F,$90,$5F,$2A
	.dc.b	$FF,$1A,$CE,$71,$F7,$E1,$D6,$FC,$6F,$78,$CD,$70,$9F,$33,$E0,$69
	.dc.b	$E0,$29,$E0,$C3,$C4,$F4,$8E,$15,$DE,$17,$E3,$C3,$CD,$1C,$E7,$C5
	.dc.b	$18,$8F,$82,$7F,$92,$CE,$25,$9C,$38,$3F,$CA,$DF,$5F,$FC,$4B,$EC
	.dc.b	$3F,$CA,$4D,$8F,$F2,$BB,$5D,$F8,$49,$DF,$9B,$FD,$26,$76,$D7,$AB
	.dc.b	$5F,$F3,$75,$7F,$44,$D3,$89,$F4,$F3,$6D,$FE,$73,$D3,$EE,$46,$2F
	.dc.b	$EC,$5F,$17,$E5,$B6,$77,$5C,$6B,$FC,$F8,$71,$AD,$E1,$E7,$AF,$C4
	.dc.b	$5F,$1A,$DD,$F1,$BF,$DD,$C7,$1A,$DD,$F1,$8F,$5B,$67,$75,$C5,$FC
	.dc.b	$E3,$6F,$B4,$7F,$06,$6B,$A3,$E1,$46,$E7,$A4,$64,$FE,$35,$5E,$BC
	.dc.b	$3F,$CF,$0D,$D7,$C7,$1C,$A7,$FC,$03,$39,$7C,$A7,$F8,$13,$0F,$0F
	.dc.b	$D8,$7F,$68,$8F,$F0,$3B,$5F,$E5,$0D,$AF,$F4,$33,$FD,$F2,$37,$B7
	.dc.b	$FA,$1B,$0F,$F5,$54,$ED,$B1,$99,$FE,$4B,$3C,$5F,$E6,$9F,$5F,$FE
	.dc.b	$54,$7F,$C2,$ED,$73,$CF,$C5,$7D,$7F,$F5,$67,$71,$F9,$F9,$E1,$1F
	.dc.b	$EB,$16,$BF,$EB,$D6,$BF,$CD,$97,$8B,$D5,$B1,$E2,$8F,$05,$3E,$C3
	.dc.b	$51,$27,$47,$F8,$BA,$74,$E0,$FF,$5F,$D7,$46,$CF,$9C,$FD,$B4,$FF
	.dc.b	$A2,$63,$FD,$84,$D7,$FD,$8E,$D7,$FD,$98,$D7,$FD,$B1,$F6,$1F,$D2
	.dc.b	$1F,$F1,$FB,$5F,$EC,$0D,$D3,$FD,$7D,$3D,$5F,$F6,$04,$6E,$7F,$DB
	.dc.b	$0D,$7A,$F4,$7F,$0B,$8C,$47,$F5,$EA,$FF,$06,$3E,$8F,$17,$D6,$47
	.dc.b	$FD,$B9,$E3,$3E,$07,$F8,$A3,$63,$31,$83,$3E,$44,$8E,$2D,$7F,$DD
	.dc.b	$3F,$5F,$FE,$6A,$7F,$DD,$0D,$73,$EF,$7F,$21,$8C,$47,$FB,$CF,$EB
	.dc.b	$FF,$CD,$CF,$FB,$BD,$AF,$FB,$D1,$AE,$FD,$FF,$E5,$5F,$5F,$FE,$61
	.dc.b	$8F,$F7,$B3,$5F,$F7,$DB,$5F,$F8,$00,$D7,$FC,$06,$7F,$93,$BE,$67
	.dc.b	$F9,$A7,$A8,$69,$23,$AF,$B7,$E0,$23,$83,$99,$D1,$FC,$63,$86,$B9
	.dc.b	$E3,$FE,$85,$18,$8F,$F0,$AF,$D7,$F9,$5B,$EF,$7F,$47,$8C,$41,$BF
	.dc.b	$5F,$FE,$69,$8F,$09,$FE,$1E,$70,$E1,$AF,$99,$AF,$C2,$5F,$5F,$FD
	.dc.b	$55,$B1,$D0,$38,$7A,$47,$F5,$BF,$AF,$FF,$63,$C6,$DA,$F9,$1A,$FC
	.dc.b	$66,$BF,$4B,$EB,$FF,$CD,$EF,$76,$BE,$C6,$BF,$39,$AF,$DA,$FA,$EF
	.dc.b	$F7,$33,$E7,$6B,$F1,$1A,$FD,$66,$BF,$82,$FA,$D7,$D7,$7F,$1B,$6C
	.dc.b	$7F,$AD,$9A,$FF,$95,$FE,$37,$F6,$DE,$1F,$1A,$FE,$F9,$5A,$AF,$C1
	.dc.b	$B5,$D1,$F2,$DE,$7D,$CA,$FC,$9D,$BE,$7F,$F9,$63,$6F,$C5,$86,$3E
	.dc.b	$7B,$F9,$87,$D7,$7F,$8F,$30,$FC,$DB,$6B,$FC,$BF,$6A,$E1,$1F,$ED
	.dc.b	$09,$D2,$6C,$EE,$35,$F5,$1D,$88,$F1,$FA,$8A,$FF,$9E,$69,$38,$CF
	.dc.b	$D9,$3F,$E7,$09,$57,$F4,$7F,$AE,$FE,$41,$97,$EC,$9B,$5F,$E7,$66
	.dc.b	$BE,$5E,$6E,$1E,$27,$F8,$A6,$7F,$8F,$C9,$A4,$8F,$8B,$F0,$D2,$7F
	.dc.b	$DD,$F3,$D3,$07,$D7,$7F,$59,$5F,$53,$4C,$44,$AF,$F7,$BC,$E9,$B3
	.dc.b	$97,$FD,$35,$1F,$EA,$C7,$7E,$F7,$53,$B7,$65,$8C,$90,$E2,$FF,$6B
	.dc.b	$FA,$C7,$C9,$D2,$4F,$89,$AF,$F0,$08,$DB,$FB,$76,$B9,$8F,$F2,$84
	.dc.b	$63,$83,$EA,$CF,$D4,$FA,$0F,$84,$74,$8C,$FD,$73,$EB,$BF,$BF,$36
	.dc.b	$BF,$D5,$73,$A9,$CF,$A9,$A6,$4E,$B8,$FF,$97,$DF,$18,$FF,$60,$3F
	.dc.b	$E7,$78,$CF,$91,$DB,$AA,$57,$13,$FD,$73,$5F,$D2,$F1,$F2,$D6,$A7
	.dc.b	$9D,$79,$A3,$E5,$70,$49,$E4,$34,$FB,$4F,$8B,$ED,$34,$FC,$B9,$F8
	.dc.b	$E4,$7E,$63,$FE,$1F,$FA,$FF,$D4,$9F,$AA,$AE,$78,$FC,$C8,$FD,$1D
	.dc.b	$78,$1A,$FF,$B7,$DB,$1B,$E2,$E9,$D3,$2B,$83,$DE,$3C,$1E,$37,$87
	.dc.b	$9E,$57,$58,$CB,$DE,$32,$F7,$8F,$F1,$B7,$D6,$BE,$BE,$92,$7E,$2F
	.dc.b	$D7,$BE,$D3,$11,$FE,$F5,$AE,$0F,$CA,$6B,$FE,$F9,$9E,$0E,$6B,$8E
	.dc.b	$7F,$8E,$AD,$BF,$90,$AD,$D7,$5A,$DF,$1F,$F7,$4B,$DB,$FD,$DA,$B7
	.dc.b	$7C,$EB,$77,$FE,$FC,$BD,$BF,$EF,$EA,$DD,$F3,$AD,$CA,$B6,$AB,$6F
	.dc.b	$39,$DC,$EA,$73,$3C,$67,$27,$F9,$86,$33,$3D,$09,$19,$C3,$75,$F1
	.dc.b	$4F,$9E,$E1,$F8,$46,$0D,$18,$26,$1F,$26,$38,$3F,$0E,$38,$1A,$C3
	.dc.b	$E0,$3C,$A3,$83,$63,$2E,$CF,$91,$3D,$23,$33,$C2,$74,$7C,$13,$91
	.dc.b	$C0,$A6,$4D,$E1,$F1,$2F,$A4,$6C,$E7,$87,$C6,$7D,$35,$EB,$73,$3F
	.dc.b	$25,$EB,$3B,$7F,$21,$87,$47,$C1,$5E,$73,$91,$30,$F5,$9F,$8B,$9F
	.dc.b	$CE,$CF,$F8,$E2,$BF,$36,$DB,$FF,$7B,$46,$23,$E3,$99,$3F,$C4,$8E
	.dc.b	$5F,$C4,$18,$8F,$C2,$6F,$23,$D3,$15,$FD,$44,$7A,$71,$FE,$13,$87
	.dc.b	$E7,$E3,$06,$F1,$F5,$57,$63,$89,$FB,$88,$DB,$4E,$E5,$36,$7C,$C3
	.dc.b	$11,$FB,$38,$E3,$1A,$8E,$EF,$CC,$9F,$23,$C1,$F9,$6F,$8E,$39,$39
	.dc.b	$31,$1A,$7F,$5D,$3E,$B8,$C6,$1F,$AD,$8C,$4F,$02,$1C,$0E,$87,$8C
	.dc.b	$CD,$FC,$87,$F8,$04,$7A,$AF,$C7,$1E,$F3,$D9,$7E,$A7,$11,$A2,$47
	.dc.b	$93,$0F,$1C,$F8,$4D,$C7,$19,$DC,$79,$B0,$F5,$3E,$67,$F2,$D1,$A8
	.dc.b	$E4,$FC,$C8,$DC,$79,$0E,$55,$E0,$C3,$D9,$1E,$F3,$F8,$B7,$4F,$F0
	.dc.b	$FE,$99,$1F,$CE,$23,$F8,$91,$87,$F9,$A4,$61,$FE,$7B,$F5,$3F,$C5
	.dc.b	$C8,$FF,$40,$E9,$AF,$F3,$B8,$B9,$7F,$37,$9F,$F7,$F7,$D8,$3E,$C2
	.dc.b	$92,$FF,$B6,$FD,$87,$FC,$01,$57,$FE,$0E,$AE,$11,$FE,$8B,$6B,$FE
	.dc.b	$98,$6B,$FE,$AA,$6B,$FE,$B6,$77,$1F,$E1,$89,$FF,$04,$48,$38,$53
	.dc.b	$C3,$6F,$F5,$FC,$FF,$82,$5B,$D1,$FE,$C2,$BE,$8D,$8D,$F4,$DC,$3B
	.dc.b	$5F,$16,$D6,$5F,$0B,$63,$FC,$F1,$1F,$0A,$BC,$3E,$26,$49,$1F,$E1
	.dc.b	$4F,$B0,$FF,$39,$D7,$F9,$C6,$38,$1E,$09,$DB,$81,$3F,$E7,$AF,$B4
	.dc.b	$7B,$9F,$EB,$79,$46,$1F,$F5,$D3,$5F,$F6,$3B,$5F,$F6,$CF,$D8,$3E
	.dc.b	$C5,$F2,$3F,$EC,$18,$C3,$7B,$DF,$49,$9A,$DD,$FF,$B4,$BF,$0D,$7D
	.dc.b	$8F,$F9,$3A,$3F,$DD,$F1,$8D,$FF,$D9,$8D,$8E,$13,$B7,$F1,$3F,$54
	.dc.b	$E7,$FF,$45,$F2,$48,$C6,$1F,$E4,$58,$43,$9F,$73,$FE,$56,$8D,$DF
	.dc.b	$CC,$AD,$DE,$A3,$BB,$63,$9D,$DA,$E3,$1E,$2F,$AD,$F4,$37,$1E,$1F
	.dc.b	$03,$F2,$32,$F2,$56,$9F,$F3,$3B,$72,$FD,$2B,$6D,$5C,$67,$CD,$3B
	.dc.b	$5F,$5E,$FA,$FF,$37,$D8,$3E,$C7,$D5,$5A,$9C,$C6,$2B,$86,$5F,$EF
	.dc.b	$B6,$D7,$EA,$48,$E1,$C1,$CE,$38,$57,$33,$E2,$79,$D6,$3D,$FF,$63
	.dc.b	$F2,$72,$37,$E2,$B0,$C9,$D9,$F1,$A3,$18,$7D,$5D,$DF,$63,$F4,$7E
	.dc.b	$1F,$E9,$36,$FF,$01,$3A,$6C,$E9,$A7,$FA,$03,$EC,$5F,$33,$F7,$7D
	.dc.b	$A7,$D1,$FC,$89,$CD,$7E,$2E,$72,$F1,$AF,$CA,$64,$AD,$BF,$18,$C4
	.dc.b	$6A,$33,$1F,$29,$DF,$27,$FA,$C1,$B5,$F8,$4F,$A6,$F1,$5F,$DA,$F9
	.dc.b	$CF,$F3,$B4,$F5,$C1,$18,$7C,$0D,$7E,$BB,$5F,$B9,$F6,$0F,$A6,$F4
	.dc.b	$B6,$BA,$36,$7D,$4D,$7F,$03,$DC,$F8,$BF,$82,$DA,$F2,$36,$7E,$13
	.dc.b	$5F,$C4,$FD,$7F,$E6,$DB,$E1,$7C,$A8,$C4,$7E,$45,$AF,$F3,$BF,$AF
	.dc.b	$FF,$70,$DF,$C1,$6C,$7C,$F6,$BF,$96,$6B,$F9,$B5,$F1,$9F,$61,$B3
	.dc.b	$CE,$30,$7D,$CF,$AE,$7D,$75,$24,$FD,$18,$CB,$F4,$FE,$7D,$F2,$9F
	.dc.b	$84,$6C,$66,$30,$7E,$23,$EB,$5F,$49,$49,$3F,$76,$32,$FD,$FF,$A6
	.dc.b	$72,$FE,$41,$B1,$98,$C1,$D7,$59,$1C,$DA,$7E,$55,$7D,$17,$E6,$9A
	.dc.b	$7E,$34,$B1,$B8,$FC,$FF,$57,$B2,$3F,$3C,$D7,$F8,$2B,$5F,$E1,$4D
	.dc.b	$7F,$BA,$7D,$5F,$F1,$CA,$FC,$D3,$63,$F5,$0D,$7F,$87,$B5,$FE,$28
	.dc.b	$BE,$AF,$F6,$2D,$9E,$71,$83,$E3,$7D,$54,$8F,$94,$DA,$FD,$B7,$D5
	.dc.b	$3D,$CF,$DD,$B6,$BF,$5C,$D9,$DC,$FF,$00,$E6,$78,$74,$97,$FB,$58
	.dc.b	$CB,$FB,$7F,$A9,$FB,$77,$FD,$3F,$B5,$EC,$FF,$36,$9C,$F8,$3F,$D7
	.dc.b	$27,$37,$FE,$1B,$83,$FC,$D6,$76,$E6,$77,$7F,$E0,$39,$CB,$FE,$30
	.dc.b	$8C,$D7,$28,$FF,$13,$4F,$91,$FE,$7A,$D7,$B7,$27,$F1,$E8,$C4,$70
	.dc.b	$F8,$EF,$EC,$9F,$53,$FC,$62,$31,$8F,$F7,$18,$C4,$29,$5C,$3C,$2F
	.dc.b	$80,$DB,$F7,$1E,$DF,$A6,$FE,$D3,$F4,$9F,$15,$B1,$AC,$3D,$31,$98
	.dc.b	$C4,$77,$30,$7F,$89,$73,$57,$F6,$45,$F5,$36,$30,$DA,$E7,$3C,$1E
	.dc.b	$B5,$FE,$4D,$EC,$FF,$27,$7D,$43,$E4,$7F,$C9,$18,$FF,$70,$AD,$09
	.dc.b	$D3,$F6,$E3,$E9,$46,$23,$FB,$85,$74,$AE,$77,$67,$FB,$8C,$D7,$FC
	.dc.b	$D7,$7C,$AB,$FA,$84,$E7,$0F,$45,$78,$67,$2D,$3C,$57,$CD,$F3,$E1
	.dc.b	$E3,$AE,$2F,$96,$33,$5E,$47,$91,$E8,$2C,$F0,$8E,$F3,$CF,$0F,$35
	.dc.b	$7A,$8D,$B8,$79,$1F,$CA,$CF,$F2,$C6,$1F,$E5,$94,$FF,$61,$FA,$7F
	.dc.b	$E3,$3F,$53,$7B,$0D,$25,$7F,$73,$97,$D4,$FF,$A4,$A3,$0D,$7F,$C3
	.dc.b	$B1,$AC,$FB,$4E,$67,$FC,$49,$39,$7E,$0E,$6D,$BF,$C5,$1A,$7D,$B8
	.dc.b	$C4,$FD,$7A,$E1,$3F,$7D,$B3,$FE,$2F,$37,$3D,$2B,$9C,$78,$2F,$D1
	.dc.b	$7C,$5B,$1C,$67,$75,$FE,$A1,$8E,$75,$BC,$3C,$57,$C2,$B5,$7C,$1F
	.dc.b	$CD,$37,$0F,$2C,$7F,$AA,$DA,$E8,$FC,$96,$1D,$8F,$35,$E2,$BF,$D4
	.dc.b	$51,$EC,$8F,$E2,$13,$C4,$FF,$4C,$57,$49,$D6,$BE,$B9,$FD,$DC,$61
	.dc.b	$F1,$61,$E9,$5F,$53,$FB,$C7,$F8,$0B,$5F,$F5,$23,$70,$F0,$BF,$82
	.dc.b	$8D,$ED,$FE,$B7,$8D,$61,$FE,$61,$6C,$FF,$B2,$DE,$0F,$FA,$BF,$EA
	.dc.b	$7F,$A0,$BF,$C4,$1A,$E7,$A7,$FD,$85,$F5,$1F,$EE,$1C,$7E,$0D,$69
	.dc.b	$B3,$FE,$C7,$8C,$61,$FE,$71,$AE,$0F,$F9,$CE,$1F,$35,$B2,$61,$FC
	.dc.b	$82,$BF,$D7,$58,$76,$9C,$CF,$9D,$F1,$FC,$26,$27,$53,$E9,$9C,$BC
	.dc.b	$FD,$AF,$49,$1F,$F5,$67,$BF,$FE,$D3,$AF,$B7,$87,$FA,$3E,$7F,$98
	.dc.b	$FD,$4F,$F3,$E6,$F4,$71,$7F,$DA,$D1,$96,$C7,$FB,$9E,$31,$5D,$30
	.dc.b	$F9,$3E,$F4,$BF,$73,$FE,$ED,$FD,$BF,$FD,$07,$1F,$D0,$3C,$B7,$AF
	.dc.b	$FE,$F1,$6C,$66,$39,$3F,$EE,$78,$7D,$56,$C9,$1B,$3F,$A5,$FD,$57
	.dc.b	$FB,$F5,$34,$FF,$80,$19,$7F,$B4,$FE,$6B,$FE,$00,$7D,$57,$F9,$96
	.dc.b	$BF,$A9,$B5,$FF,$7B,$36,$7F,$47,$CE,$FA,$1F,$F5,$84,$7E,$9B,$EA
	.dc.b	$9F,$55,$8E,$8F,$F5,$A4,$63,$1F,$EB,$ED,$55,$D3,$E5,$B8,$40,$7C
	.dc.b	$D3,$E5,$B3,$18,$6C,$66,$3E,$FF,$8F,$AC,$3F,$CD,$31,$E0,$F6,$70
	.dc.b	$72,$FA,$A7,$CD,$F4,$8F,$CB,$CF,$F7,$46,$BE,$6D,$3F,$D9,$D1,$8D
	.dc.b	$3F,$82,$DF,$F9,$52,$BF,$BD,$BE,$D8,$CE,$BD,$CF,$6A,$EC,$7C,$C7
	.dc.b	$B3,$EA,$BF,$CD,$B8,$7F,$81,$1A,$F8,$4F,$88,$E2,$BF,$8A,$FD,$4F
	.dc.b	$FB,$79,$FE,$2F,$0F,$AA,$E8,$4A,$FF,$07,$D7,$EC,$34,$FF,$74,$3F
	.dc.b	$EC,$C8,$CE,$FF,$3A,$31,$1E,$1C,$3D,$1F,$61,$F4,$BE,$AF,$E0,$4F
	.dc.b	$94,$38,$9F,$E5,$2D,$C3,$BF,$47,$DA,$FA,$BF,$F7,$6D,$FF,$88,$9A
	.dc.b	$F3,$6D,$FF,$32,$FB,$47,$DE,$FE,$0B,$EB,$7F,$25,$7F,$E2,$A6,$BE
	.dc.b	$36,$D7,$F3,$AF,$8D,$FE,$01,$39,$EB,$E4,$47,$FC,$67,$1A,$9F,$E8
	.dc.b	$35,$D2,$7F,$28,$7E,$55,$C3,$86,$B9,$AF,$F7,$02,$AF,$E2,$59,$F2
	.dc.b	$FA,$DF,$E9,$B8,$7F,$4E,$6C,$76,$9D,$D7,$E8,$DA,$A7,$6F,$E9,$DA
	.dc.b	$F7,$F9,$FF,$D3,$C3,$EA,$29,$27,$FC,$A3,$0F,$A8,$C4,$7D,$B8,$E0
	.dc.b	$EA,$79,$CF,$EC,$BE,$B7,$F9,$0E,$3F,$11,$B5,$FB,$96,$BE,$7F,$37
	.dc.b	$87,$99,$23,$FC,$C3,$3A,$C7,$F5,$97,$CE,$BE,$4C,$FF,$31,$AF,$A2
	.dc.b	$1C,$3B,$32,$7F,$09,$C1,$F5,$BF,$C2,$17,$D4,$D3,$11,$27,$FC,$E7
	.dc.b	$3A,$6C,$E5,$FC,$F4,$7F,$AC,$5D,$FC,$1F,$53,$B7,$89,$63,$24,$38
	.dc.b	$BF,$C6,$BE,$85,$F3,$54,$93,$E2,$6B,$FA,$D8,$DB,$FE,$88,$6B,$E4
	.dc.b	$8D,$F1,$7F,$86,$27,$2D,$8F,$84,$7F,$2A,$FA,$C7,$D6,$D2,$5F,$FA
	.dc.b	$52,$3F,$CD,$9B,$FE,$81,$DC,$66,$76,$FF,$01,$8D,$BF,$CE,$5F,$EA
	.dc.b	$43,$EA,$3C,$07,$87,$E1,$E4,$F7,$8C,$BE,$18,$F8,$DB,$48,$F8,$0F
	.dc.b	$A7,$B3,$11,$F1,$62,$E7,$FE,$38,$C1,$58,$CB,$F7,$3A,$7D,$DC,$38
	.dc.b	$1E,$F9,$FE,$9B,$F5,$BF,$A0,$8F,$E3,$8D,$C3,$FC,$86,$DB,$FE,$06
	.dc.b	$FE,$A6,$7F,$D7,$6E,$5F,$EB,$4D,$4D,$38,$3D,$BC,$BF,$0F,$D6,$C8
	.dc.b	$E7,$1E,$43,$0F,$92,$30,$FF,$B2,$BE,$B5,$E7,$7F,$66,$6C,$66,$3F
	.dc.b	$B4,$FD,$6F,$F6,$08,$C1,$0F,$E7,$53,$C1,$FE,$79,$1B,$73,$3D,$48
	.dc.b	$FF,$B6,$A3,$59,$7F,$9A,$2B,$93,$6B,$75,$D3,$CE,$B3,$98,$F1,$3F
	.dc.b	$E7,$0B,$E7,$9F,$F3,$0C,$39,$B7,$0E,$2F,$F3,$49,$DD,$FF,$82,$6B
	.dc.b	$6D,$AD,$BF,$E0,$D9,$DA,$9D,$B6,$76,$A7,$73,$A2,$4F,$CC,$C3,$71
	.dc.b	$8A,$F8,$91,$89,$E7,$5C,$9C,$4E,$8F,$83,$7E,$D6,$9E,$7A,$DB,$F4
	.dc.b	$8E,$05,$26,$1F,$2D,$F9,$C7,$23,$D8,$5B,$F1,$39,$C3,$E0,$C6,$23
	.dc.b	$9D,$F8,$DE,$66,$0F,$98,$73,$BF,$2D,$F8,$4F,$B4,$E2,$79,$46,$27
	.dc.b	$C1,$87,$C4,$AD,$47,$43,$BC,$F3,$AE,$24,$8F,$88,$72,$3C,$45,$3C
	.dc.b	$C7,$D9,$9E,$AF,$C2,$3A,$C7,$37,$13,$DD,$EA,$72,$C3,$EE,$3C,$0F
	.dc.b	$34,$6D,$E3,$3F,$93,$AE,$F5,$F3,$1B,$7F,$87,$AF,$35,$F1,$C3,$E8
	.dc.b	$3C,$EF,$84,$EB,$0F,$35,$69,$F4,$4E,$A7,$CD,$7F,$2D,$E0,$79,$A7
	.dc.b	$C8,$E0,$F3,$9E,$EB,$F0,$4F,$18,$E6,$7C,$C2,$9F,$32,$FC,$4F,$52
	.dc.b	$1F,$3C,$ED,$7E,$5A,$EB,$1C,$DA,$F3,$C3,$E2,$BF,$9C,$8C,$9F,$16
	.dc.b	$33,$1F,$44,$F4,$92,$7B,$61,$FC,$0F,$97,$F8,$FF,$2C,$AF,$E2,$D1
	.dc.b	$FC,$88,$C3,$FC,$7A,$30,$FF,$1B,$F2,$FF,$93,$91,$FE,$3D,$F4,$D5
	.dc.b	$FE,$4B,$17,$2F,$E3,$13,$FE,$96,$FA,$DF,$F3,$ED,$7F,$A6,$E5,$8F
	.dc.b	$F9,$2A,$73,$3F,$E6,$96,$CE,$A3,$FD,$45,$F5,$BF,$EE,$C7,$FD,$49
	.dc.b	$3A,$AE,$5E,$77,$73,$FD,$4B,$5B,$A5,$6D,$E3,$3B,$7C,$95,$FE,$5C
	.dc.b	$AE,$07,$FB,$24,$9C,$FD,$3A,$3A,$B2,$B8,$9E,$67,$4E,$04,$FF,$96
	.dc.b	$09,$E1,$7F,$AE,$BE,$B7,$FC,$AF,$18,$9E,$0F,$F9,$7A,$3F,$CB,$7F
	.dc.b	$80,$3E,$7B,$FC,$D6,$DA,$FF,$1F,$87,$0F,$F9,$09,$AF,$F9,$23,$EB
	.dc.b	$BC,$6F,$06,$BF,$E4,$86,$BB,$BF,$F3,$CF,$D7,$3E,$B9,$F2,$BE,$08
	.dc.b	$C3,$7B,$BB,$DF,$FB,$66,$72,$FF,$AD,$7E,$9B,$FD,$81,$C1,$F2,$38
	.dc.b	$D1,$89,$FF,$57,$2E,$4F,$C1,$4C,$8D,$37,$97,$37,$F8,$D9,$CB,$6B
	.dc.b	$37,$FE,$EE,$9C,$BF,$EF,$0F,$A9,$7D,$6E,$B1,$46,$30,$F3,$7C,$27
	.dc.b	$C0,$7D,$11,$9B,$FF,$62,$36,$BC,$ED,$9F,$D0,$57,$1A,$DD,$F9,$17
	.dc.b	$D6,$3E,$B3,$C5,$F5,$AF,$88,$D4,$E7,$1F,$91,$39,$7F,$1D,$19,$BF
	.dc.b	$F6,$B4,$E5,$FF,$6A,$46,$73,$F9,$73,$97,$F5,$D3,$E2,$C2,$BF,$EB
	.dc.b	$EA,$E8,$FF,$A4,$1A,$E6,$36,$FC,$BF,$AD,$F9,$5A,$1B,$F1,$D8,$66
	.dc.b	$5F,$29,$CB,$4F,$C2,$6C,$FA,$DF,$A8,$BE,$2A,$C6,$1B,$3C,$1B,$8F
	.dc.b	$E2,$9F,$9D,$F5,$AF,$80,$FB,$DE,$C3,$E8,$F3,$18,$6E,$1F,$95,$9D
	.dc.b	$3F,$BB,$AE,$4E,$67,$78,$7F,$C0,$19,$CB,$F8,$4F,$A4,$CF,$0F,$CE
	.dc.b	$FA,$0F,$F2,$AE,$1F,$A1,$FB,$5F,$F7,$DB,$FC,$1E,$33,$6A,$FD,$0D
	.dc.b	$6A,$31,$5F,$EF,$2F,$AD,$7D,$77,$E6,$7D,$7F,$F4,$74,$FF,$11,$9C
	.dc.b	$BF,$AF,$8C,$D7,$20,$7F,$0D,$FA,$EE,$5F,$39,$F9,$5F,$19,$18,$BF
	.dc.b	$E3,$9F,$5D,$F9,$BD,$3F,$6B,$F6,$BD,$1F,$C5,$C6,$67,$FA,$04,$E5
	.dc.b	$EF,$8B,$EB,$FF,$D8,$31,$FC,$0E,$7C,$6F,$F2,$98,$CC,$F9,$ED,$5A
	.dc.b	$8F,$51,$CF,$E4,$3F,$75,$F1,$6F,$A2,$FD,$76,$09,$CC,$7E,$C7,$EB
	.dc.b	$9F,$5F,$FC,$07,$F1,$1E,$99,$FE,$92,$73,$A7,$F1,$79,$CB,$EA,$FA
	.dc.b	$67,$CD,$FF,$09,$FB,$67,$D4,$7E,$F6,$73,$5F,$C7,$E7,$2F,$F4,$5F
	.dc.b	$B1,$7D,$67,$F1,$1F,$82,$1F,$C1,$63,$33,$FC,$9E,$32,$FF,$4D,$FB
	.dc.b	$17,$77,$F1,$9F,$B6,$EB,$CB,$F1,$27,$35,$FC,$BE,$73,$C1,$FC,$03
	.dc.b	$1F,$5D,$FE,$3F,$ED,$9F,$23,$E1,$8F,$E4,$91,$8C,$3F,$AE,$1F,$34
	.dc.b	$FE,$74,$BE,$4F,$F8,$E4,$66,$7E,$C4,$E6,$3F,$B2,$78,$9F,$12,$7E
	.dc.b	$CF,$E2,$CF,$AB,$FE,$49,$39,$CF,$F8,$AC,$E6,$FF,$B6,$7D,$A3,$D5
	.dc.b	$FE,$55,$A3,$E9,$7F,$96,$4E,$6B,$FB,$94,$E5,$E3,$F6,$9A,$F9,$6D
	.dc.b	$B6,$3E,$67,$DA,$7F,$35,$FC,$5B,$FD,$D3,$18,$3F,$BE,$38,$DF,$F2
	.dc.b	$3F,$6A,$F8,$CF,$E7,$9C,$CC,$3F,$9F,$4E,$70,$E5,$7F,$98,$F8,$17
	.dc.b	$AF,$AC,$7F,$A2,$46,$2B,$FC,$15,$7C,$6F,$F3,$FF,$69,$F8,$47,$F4
	.dc.b	$9E,$8B,$E2,$7F,$A6,$4E,$79,$7F,$B4,$CE,$70,$FF,$0A,$4E,$5F,$F1
	.dc.b	$A4,$66,$FA,$F7,$7F,$63,$8C,$5F,$F4,$6B,$E1,$5F,$C5,$67,$37,$FB
	.dc.b	$C9,$CD,$27,$35,$FC,$3A,$5C,$3F,$DB,$FE,$D7,$F9,$3D,$F5,$8C,$7A
	.dc.b	$DE,$A6,$E3,$E8,$6E,$1F,$0B,$D0,$FE,$F1,$D0,$FA,$9F,$83,$18,$6D
	.dc.b	$FC,$08,$F8,$B5,$83,$F8,$4E,$3C,$67,$FC,$70,$FE,$99,$FE,$2E,$E6
	.dc.b	$37,$E2,$FF,$79,$8C,$1F,$DA,$8B,$3F,$C7,$A7,$F9,$6F,$8B,$FC,$BB
	.dc.b	$ED,$3F,$3D,$1F,$E6,$3C,$7F,$C0,$4E,$84,$69,$FE,$85,$1F,$E6,$38
	.dc.b	$C4,$6D,$D5,$F9,$5C,$C2,$BC,$1D,$FE,$AB,$E4,$F9,$67,$47,$81,$D6
	.dc.b	$1C,$23,$D5,$55,$F1,$C5,$7C,$53,$C1,$E4,$70,$29,$C0,$A7,$03,$E1
	.dc.b	$47,$37,$C7,$7E,$CB,$F3,$C7,$38,$C4,$EE,$FF,$CC,$33,$F2,$4F,$F3
	.dc.b	$1B,$87,$DA,$7D,$3F,$B4,$FE,$89,$F6,$BF,$E6,$86,$C7,$C3,$6B,$BF
	.dc.b	$35,$3B,$FB,$5F,$EE,$93,$B6,$BD,$5A,$FF,$5E,$A4,$62,$3F,$B7,$C6
	.dc.b	$34,$F9,$9A,$36,$FE,$CB,$A7,$E0,$63,$17,$F4,$6F,$83,$F3,$1B,$3B
	.dc.b	$AE,$35,$FD,$B8,$E3,$5B,$C3,$D3,$5F,$8D,$BE,$35,$BB,$E3,$7F,$E0
	.dc.b	$48,$E3,$5B,$BE,$31,$ED,$6C,$EE,$B8,$BF,$CE,$DB,$7D,$A3,$E0,$CF
	.dc.b	$07,$47,$C2,$9D,$C6,$6B,$D3,$87,$F9,$59,$BA,$79,$EF,$FD,$5D,$87
	.dc.b	$57,$99,$FE,$04,$9D,$BE,$53,$77,$E1,$FB,$5F,$E1,$11,$FE,$B7,$6B
	.dc.b	$FE,$0C,$6E,$1F,$E6,$67,$ED,$C6,$F6,$FE,$0B,$8F,$F8,$72,$76,$D8
	.dc.b	$CC,$FF,$8B,$A7,$8B,$FE,$18,$FB,$5F,$E8,$2F,$FB,$1D,$AE,$7B,$38
	.dc.b	$AF,$B4,$FF,$6E,$4E,$B1,$FF,$3C,$B7,$1F,$F1,$6B,$6B,$FC,$7B,$5C
	.dc.b	$3C,$2E,$73,$FD,$FA,$7C,$11,$DA,$5F,$6B,$A8,$93,$A3,$FD,$B9,$3A
	.dc.b	$70,$7A,$E3,$6F,$F8,$D2,$34,$DF,$88,$EC,$47,$FD,$85,$F1,$6F,$93
	.dc.b	$C4,$7F,$8F,$3E,$D7,$FA,$3B,$FE,$EB,$6B,$CC,$C7,$57,$C6,$8C,$EB
	.dc.b	$FE,$9C,$6E,$BF,$E5,$0C,$BF,$D5,$97,$EA,$EA,$F8,$04,$7F,$DF,$1F
	.dc.b	$6B,$FE,$91,$7F,$DF,$5E,$A7,$FA,$45,$BF,$0B,$F5,$23,$11,$EF,$AC
	.dc.b	$79,$92,$38,$BB,$AF,$C1,$E6,$FD,$A5,$F5,$78,$5F,$55,$4C,$47,$0F
	.dc.b	$C0,$7F,$DE,$38,$FF,$BA,$57,$CF,$E4,$D7,$97,$DB,$F0,$0F,$F3,$CD
	.dc.b	$7F,$BB,$DF,$82,$FF,$9A,$DB,$97,$FB,$B9,$DC,$7F,$9B,$F2,$FF,$80
	.dc.b	$0F,$FB,$DD,$AE,$7E,$5F,$D1,$BF,$B3,$F0,$1F,$C8,$60,$69,$3E,$57
	.dc.b	$C9,$F3,$33,$0F,$23,$E7,$E7,$F8,$5E,$D7,$C3,$F9,$99,$5E,$18,$ED
	.dc.b	$EF,$77,$F9,$39,$3F,$E9,$AC,$DD,$57,$4F,$F3,$2C,$FC,$1F,$84,$F8
	.dc.b	$9E,$0C,$8C,$1E,$DA,$FA,$44,$7E,$07,$C7,$7B,$7E,$92,$47,$79,$ED
	.dc.b	$1F,$07,$E0,$FE,$87,$D4,$CA,$FA,$0F,$D3,$F2,$FF,$77,$F2,$68,$A4
	.dc.b	$9E,$87,$FA,$D2,$3F,$BE,$E1,$FE,$A8,$78,$46,$83,$87,$FD,$46,$DD
	.dc.b	$7F,$DF,$AF,$28,$E9,$5D,$DD,$BE,$4B,$FE,$1F,$F5,$DF,$E7,$B9,$FF
	.dc.b	$03,$07,$0F,$F8,$25,$AF,$F8,$7D,$AF,$FA,$F9,$DD,$F8,$BE,$83,$E0
	.dc.b	$9D,$B8,$BE,$1F,$8D,$FC,$7B,$EB,$BF,$C5,$5A,$7F,$B1,$9B,$8F,$C3
	.dc.b	$9D,$BF,$06,$32,$9B,$A7,$F0,$B8,$D7,$83,$D4,$8F,$1F,$AC,$AF,$FB
	.dc.b	$2A,$BF,$62,$F0,$34,$53,$53,$C3,$3F,$C7,$C7,$FB,$AA,$55,$FC,$E3
	.dc.b	$EB,$BF,$8B,$CF,$F9,$0D,$AF,$FB,$89,$AF,$FB,$C9,$A9,$FE,$32,$F5
	.dc.b	$E0,$F8,$C7,$5E,$9F,$E8,$B4,$FA,$EB,$F5,$D2,$FF,$0D,$39,$6C,$E5
	.dc.b	$FF,$74,$C2,$79,$CF,$4F,$82,$95,$C9,$EF,$F5,$18,$8E,$31,$74,$F8
	.dc.b	$71,$F0,$A7,$1C,$1F,$F0,$03,$1F,$F5,$34,$F3,$3C,$B3,$FD,$7B,$B5
	.dc.b	$F3,$1F,$E3,$C6,$CF,$E4,$9A,$FF,$96,$5A,$FF,$9B,$DC,$E1,$E2,$B7
	.dc.b	$D1,$7C,$63,$B5,$FF,$6C,$FA,$EF,$D3,$64,$C6,$46,$1B,$D3,$BE,$9E
	.dc.b	$0C,$8F,$D9,$4E,$D3,$FA,$F6,$FC,$2F,$7E,$99,$1C,$23,$C4,$61,$69
	.dc.b	$F2,$A9,$5A,$9E,$07,$81,$E1,$3B,$AF,$95,$C3,$FE,$08,$FA,$DF,$D2
	.dc.b	$9F,$A6,$DF,$EF,$61,$FC,$25,$AF,$83,$4F,$84,$A7,$8C,$74,$8F,$E2
	.dc.b	$11,$FA,$67,$81,$F2,$5C,$61,$F8,$26,$CF,$47,$F4,$B3,$CB,$9F,$CC
	.dc.b	$47,$AF,$07,$A7,$B3,$D5,$CD,$27,$C3,$19,$7C,$71,$97,$C6,$7A,$63
	.dc.b	$4E,$2D,$3B,$C7,$CD,$5A,$7D,$11,$9C,$7D,$13,$97,$D5,$19,$BF,$81
	.dc.b	$39,$7E,$14,$65,$F1,$1E,$83,$CE,$43,$C9,$1C,$5F,$A1,$1C,$6D,$39
	.dc.b	$7D,$33,$C6,$FD,$2F,$53,$DD,$5C,$6B,$83,$63,$53,$EE,$79,$5F,$B9
	.dc.b	$CB,$EE,$72,$7B,$8C,$9E,$E3,$27,$B8,$F5,$98,$BF,$3B,$82,$4F,$85
	.dc.b	$F3,$9D,$0E,$4E,$49,$3E,$58,$E2,$F9,$5E,$26,$0F,$31,$88,$E2,$E1
	.dc.b	$F8,$EE,$63,$94,$60,$8E,$09,$1C,$5F,$14,$74,$8F,$AE,$7A,$67,$EB
	.dc.b	$BF,$06,$BE,$C5,$FD,$B6,$C6,$4E,$11,$96,$99,$3D,$87,$C4,$BF,$1B
	.dc.b	$E0,$33,$5E,$07,$D8,$7D,$53,$33,$D5,$A7,$5B,$F7,$BE,$D3,$83,$ED
	.dc.b	$9E,$8E,$A3,$C0,$E6,$7F,$1B,$1F,$7D,$AF,$88,$F4,$1D,$4E,$35,$EA
	.dc.b	$AE,$AF,$CC,$31,$7F,$1A,$79,$4F,$E2,$AF,$C4,$F0,$8F,$3D,$73,$BF
	.dc.b	$0D,$79,$67,$EA,$5F,$E8,$5F,$8E,$D8,$E6,$DA,$F1,$4F,$42,$1E,$A2
	.dc.b	$5F,$8E,$3C,$11,$D8,$F8,$45,$3D,$37,$E4,$8D,$C7,$88,$D9,$4D,$9F
	.dc.b	$E3,$DF,$AB,$7D,$6F,$FB,$0B,$EB,$7F,$9B,$E5,$FD,$B9,$B7,$FC,$E2
	.dc.b	$B7,$7C,$6F,$86,$D2,$30,$7F,$92,$6F,$FB,$9B,$FC,$A1,$73,$77,$3C
	.dc.b	$1C,$F2,$78,$46,$5E,$51,$E5,$F5,$A4,$75,$7D,$1D,$18,$8E,$D1,$74
	.dc.b	$F5,$4D,$D3,$FA,$8E,$2A,$C5,$FF,$75,$8F,$EF,$6F,$F9,$87,$EB,$5F
	.dc.b	$5B,$FE,$56,$84,$61,$FF,$2C,$B5,$FE,$BA,$7F,$7F,$70,$E1,$D6,$3F
	.dc.b	$E1,$AF,$27,$FB,$D1,$FD,$EE,$33,$8F,$F6,$68,$73,$7F,$BD,$9B,$7F
	.dc.b	$E7,$56,$BF,$D5,$1F,$EF,$6E,$63,$81,$1E,$1C,$FC,$FC,$19,$1C,$DF
	.dc.b	$EE,$B8,$FF,$8B,$BE,$B7,$FC,$C3,$87,$F9,$7E,$BB,$57,$07,$FB,$1C
	.dc.b	$EA,$31,$5E,$D7,$9E,$3F,$E0,$09,$FF,$35,$FD,$5B,$CE,$FF,$67,$E1
	.dc.b	$F2,$23,$1E,$8F,$FA,$11,$76,$3E,$8B,$17,$C2,$FF,$C1,$7F,$74,$F4
	.dc.b	$3F,$C8,$B1,$86,$BF,$E3,$18,$D6,$7F,$E9,$06,$C7,$D8,$D5,$F5,$1F
	.dc.b	$85,$AF,$F5,$3C,$63,$97,$FC,$E6,$C9,$FF,$20,$CE,$9E,$15,$FE,$18
	.dc.b	$9E,$13,$9A,$E2,$FF,$B2,$EB,$8D,$6D,$E3,$5D,$BE,$A7,$CA,$DC,$78
	.dc.b	$D7,$F8,$EB,$DB,$6F,$8A,$5F,$0C,$7F,$DB,$0D,$BF,$F2,$0B,$65,$7C
	.dc.b	$AB,$DF,$39,$26,$1D,$67,$57,$ED,$9E,$8B,$EA,$DF,$57,$F4,$23,$0D
	.dc.b	$8C,$E1,$FE,$C0,$6C,$FE,$A0,$DC,$7F,$B4,$63,$38,$FC,$FF,$23,$E0
	.dc.b	$D6,$8F,$17,$A9,$F0,$71,$37,$E1,$30,$CC,$7D,$AE,$5A,$7D,$9F,$AB
	.dc.b	$F9,$5B,$BD,$EB,$18,$6C,$FD,$F6,$E3,$F6,$5B,$9F,$E1,$5B,$3F,$7D
	.dc.b	$FA,$DF,$55,$FE,$BA,$8F,$C6,$B6,$BF,$10,$BF,$01,$DF,$95,$F7,$76
	.dc.b	$7D,$3D,$31,$E0,$DC,$7F,$2F,$2F,$B5,$E0,$71,$F9,$DB,$3F,$9C,$FA
	.dc.b	$9F,$C0,$B7,$0F,$C5,$46,$34,$FC,$D3,$E0,$9E,$17,$AB,$DB,$63,$F4
	.dc.b	$F6,$F1,$FF,$28,$FF,$40,$8C,$C6,$3C,$8F,$CC,$7D,$33,$E4,$69,$8C
	.dc.b	$2D,$FE,$CB,$E9,$BA,$37,$0F,$D0,$47,$5D,$3F,$60,$E9,$FD,$97,$A2
	.dc.b	$F8,$1F,$D3,$36,$FF,$7A,$FE,$21,$B1,$FA,$BF,$A6,$7C,$B5,$C7,$0D
	.dc.b	$AF,$E0,$BF,$4B,$CF,$1F,$E0,$0F,$36,$E1,$FB,$58,$C7,$47,$F0,$A7
	.dc.b	$F8,$ED,$72,$F1,$75,$E3,$ED,$FC,$FF,$D2,$BE,$1B,$F4,$3F,$4A,$F7
	.dc.b	$33,$18,$7F,$4B,$F4,$BF,$67,$3F,$F7,$3F,$C9,$BE,$3F,$3E,$4F,$EA
	.dc.b	$A3,$1A,$FF,$B5,$6B,$FD,$DB,$6E,$BE,$7F,$03,$D1,$B5,$B3,$FD,$77
	.dc.b	$F4,$8F,$A4,$FD,$AF,$D2,$BE,$C7,$F6,$F3,$9A,$E1,$CD,$FE,$E3,$AD
	.dc.b	$46,$21,$F4,$DF,$67,$4F,$F7,$FD,$F1,$9F,$F5,$8C,$E6,$BC,$38,$FF
	.dc.b	$04,$F4,$5E,$6F,$F0,$6A,$D5,$FF,$BC,$67,$37,$DE,$73,$2F,$A4,$7D
	.dc.b	$37,$F2,$BF,$A6,$FE,$E3,$7E,$59,$CB,$FD,$12,$33,$5F,$EC,$49,$CB
	.dc.b	$FD,$06,$33,$A2,$BF,$DC,$35,$A8,$FE,$97,$3A,$9C,$DF,$87,$DD,$FE
	.dc.b	$3F,$7B,$9F,$AD,$5B,$C0,$7A,$78,$7F,$A1,$7D,$33,$E9,$B8,$C6,$1B
	.dc.b	$7F,$D7,$27,$4F,$B6,$74,$FB,$FC,$6F,$E8,$1F,$41,$E2,$6C,$64,$98
	.dc.b	$7C,$13,$4F,$DB,$D5,$5F,$CE,$17,$D3,$58,$C3,$67,$84,$7F,$3E,$AF
	.dc.b	$6F,$D3,$7B,$BE,$97,$F4,$11,$F9,$79,$D2,$9D,$76,$FF,$86,$A7,$E8
	.dc.b	$4A,$B8,$BF,$20,$E0,$66,$78,$5F,$28,$FF,$1D,$DD,$58,$FB,$27,$11
	.dc.b	$E6,$39,$1E,$57,$2F,$6B,$F0,$CF,$18,$D9,$E3,$AE,$D1,$99,$F3,$CF
	.dc.b	$57,$8E,$1E,$C7,$31,$D7,$0F,$4B,$D0,$E0,$E2,$37,$7D,$AB,$94,$FF
	.dc.b	$71,$8F,$F2,$F5,$7C,$43,$FB,$A1,$4D,$1F,$DD,$FE,$97,$F9,$2F,$D3
	.dc.b	$69,$B1,$9F,$1D,$39,$6C,$71,$FA,$6F,$96,$DC,$F9,$46,$23,$FC,$D7
	.dc.b	$68,$C4,$7D,$98,$C4,$FD,$5C,$7F,$A5,$4D,$DB,$EF,$69,$F8,$0B,$DC
	.dc.b	$FD,$06,$C7,$E1,$2B,$8C,$FD,$46,$C7,$1C,$3C,$ED,$C3,$B5,$43,$A4
	.dc.b	$EA,$BC,$CD,$8F,$DB,$C7,$43,$F2,$45,$7F,$77,$7E,$48,$E3,$5D,$6F
	.dc.b	$CD,$87,$B2,$73,$7E,$78,$FF,$37,$BD,$9A,$FF,$47,$6D,$F5,$8D,$1F
	.dc.b	$CC,$67,$E0,$DF,$EA,$23,$81,$FC,$D6,$78,$2F,$A6,$FD,$EB,$FC,$E5
	.dc.b	$AF,$E9,$B6,$FD,$74,$62,$BA,$3F,$6E,$78,$6D,$E6,$8D,$57,$08,$C6
	.dc.b	$7E,$AF,$A6,$FE,$7B,$7F,$D1,$DA,$F7,$3F,$91,$B8,$C3,$D5,$F4,$9F
	.dc.b	$D6,$E7,$87,$5F,$32,$3F,$D4,$63,$4F,$31,$4F,$0B,$C4,$48,$E1,$F3
	.dc.b	$95,$E1,$3B,$7C,$D3,$F4,$A9,$5B,$C3,$E7,$5E,$EF,$8D,$EE,$B8,$7D
	.dc.b	$34,$9F,$33,$F7,$AF,$FD,$B5,$F1,$39,$9F,$47,$D3,$3A,$3F,$02,$D8
	.dc.b	$E2,$7F,$6F,$70,$F0,$8D,$9C,$EB,$E4,$F8,$FE,$92,$4F,$F7,$7F,$A6
	.dc.b	$7C,$07,$E3,$BC,$BF,$F3,$44,$FF,$7B,$6B,$F2,$DB,$7D,$23,$5E,$5F
	.dc.b	$12,$3C,$7E,$03,$DD,$F4,$4F,$A6,$BF,$31,$2B,$FD,$6B,$39,$7F,$D6
	.dc.b	$FF,$4D,$FE,$CC,$C7,$FD,$6E,$BE,$9B,$E4,$36,$77,$E5,$FC,$BF,$A6
	.dc.b	$7D,$65,$E5,$95,$FE,$C2,$9C,$BE,$0F,$5B,$CA,$47,$C7,$C1,$F5,$7E
	.dc.b	$9F,$F2,$5D,$FF,$E2,$D6,$CF,$F4,$A6,$EA,$7F,$DB,$51,$DF,$E5,$3C
	.dc.b	$84,$7C,$5C,$FF,$13,$E9,$3E,$2F,$B5,$27,$E1,$C7,$87,$DC,$F1,$7C
	.dc.b	$64,$9F,$F6,$DE,$3F,$E6,$37,$D9,$F4,$FF,$07,$BE,$46,$0F,$9B,$AF
	.dc.b	$D2,$23,$F4,$3E,$9F,$DD,$F3,$F2,$3A,$BF,$07,$E0,$FE,$67,$D1,$49
	.dc.b	$F9,$4F,$CF,$F3,$3F,$8A,$FD,$32,$29,$27,$FC,$89,$1F,$DA,$31,$FE
	.dc.b	$F0,$EE,$33,$3B,$7F,$71,$1B,$7F,$C4,$AE,$DC,$C6,$F8,$FF,$8F,$4E
	.dc.b	$E5,$F4,$EE,$1F,$E5,$2D,$8F,$F2,$B3,$5E,$FD,$1A,$9F,$EC,$F1,$C9
	.dc.b	$B3,$B8,$FF,$08,$19,$3F,$B6,$D3,$C6,$FE,$6F,$F4,$FF,$CC,$71,$FE
	.dc.b	$64,$DC,$FE,$0C,$FC,$83,$6C,$BF,$8F,$1C,$BC,$09,$1F,$E6,$DA,$47
	.dc.b	$6F,$90,$CC,$FF,$1F,$9C,$BB,$9D,$34,$86,$DC,$BF,$C9,$FE,$9F,$F8
	.dc.b	$5E,$3F,$D2,$5B,$8F,$F8,$BD,$B3,$FE,$58,$BF,$E4,$15,$BE,$9F,$93
	.dc.b	$5F,$C5,$2B,$E3,$B5,$E5,$1F,$3B,$17,$D3,$FF,$5E,$5F,$47,$4C,$44
	.dc.b	$AF,$F4,$74,$E9,$B3,$97,$FA,$04,$7F,$B6,$5D,$FC,$57,$53,$B7,$C7
	.dc.b	$D8,$C9,$0E,$2E,$5E,$71,$72,$FE,$D9,$4A,$FE,$EB,$DA,$F9,$2A,$4C
	.dc.b	$3C,$ED,$7F,$AA,$46,$DF,$F5,$43,$5F,$4C,$7F,$80,$3E,$9F,$F4,$18
	.dc.b	$7F,$7B,$6E,$FF,$E6,$16,$D7,$FA,$29,$FD,$11,$84,$D2,$73,$F5,$F8
	.dc.b	$2E,$DF,$43,$2B,$FD,$75,$49,$D4,$62,$74,$FF,$6D,$9D,$3C,$27,$38
	.dc.b	$79,$EB,$8C,$FC,$0D,$BE,$39,$1E,$FF,$13,$5D,$1F,$12,$3E,$14,$7F
	.dc.b	$88,$BE,$9F,$F3,$B3,$FD,$F5,$B7,$FE,$1F,$6E,$7F,$E8,$55,$BF,$1E
	.dc.b	$59,$7F,$ED,$79,$DB,$FE,$B2,$76,$E7,$B7,$C1,$DB,$23,$04,$3C,$4E
	.dc.b	$5F,$24,$79,$CC,$63,$E5,$9C,$4F,$18,$C3,$E3,$8C,$BC,$8C,$10,$C1
	.dc.b	$23,$8B,$E2,$3A,$63,$F0,$6B,$C2,$FC,$3A,$EB,$5F,$1E,$73,$5F,$06
	.dc.b	$73,$5C,$63,$C1,$5D,$8E,$0F,$63,$81,$A8,$E0,$F9,$5F,$03,$93,$E7
	.dc.b	$57,$67,$89,$E9,$38,$1A,$38,$0B,$5F,$18,$C4,$FA,$9D,$46,$65,$39
	.dc.b	$78,$CF,$A4,$E2,$76,$72,$F8,$8C,$9E,$5B,$F8,$57,$EE,$7E,$1C,$66
	.dc.b	$3D,$EF,$19,$F0,$BD,$6F,$C9,$7E,$36,$C6,$4F,$24,$F0,$7A,$1F,$1A
	.dc.b	$FB,$92,$FB,$BE,$48,$F9,$07,$86,$BE,$05,$FC,$36,$C7,$66,$BB,$3E
	.dc.b	$14,$6D,$A6,$CA,$6C,$F2,$9F,$1A,$7C,$8F,$9E,$BC,$A7,$C7,$23,$92
	.dc.b	$1C,$5C,$BF,$68,$EF,$1D,$1C,$1E,$0B,$F8,$D1,$97,$EC,$BF,$0E,$3A
	.dc.b	$1F,$11,$C4,$6A,$7B,$BC,$9F,$3C,$F9,$23,$0F,$18,$CB,$E2,$9F,$05
	.dc.b	$7D,$B8,$FB,$2D,$7F,$04,$D8,$DC,$7D,$F3,$25,$72,$FB,$6F,$F0,$AD
	.dc.b	$3A,$9B,$AF,$4D,$75,$6C,$64,$F4,$10,$F4,$3E,$AB,$F6,$C7,$C6,$32
	.dc.b	$7C,$19,$FF,$68,$FD,$3B,$E9,$E9,$24,$7F,$AB,$FE,$9B,$FC,$E8,$FF
	.dc.b	$AC,$67,$83,$AC,$BF,$DA,$91,$88,$D5,$7F,$B7,$AB,$8D,$7F,$B6,$5F
	.dc.b	$EF,$53,$D7,$97,$46,$F9,$A4,$75,$CF,$C6,$AB,$FB,$B3,$93,$CE,$E1
	.dc.b	$F4,$4F,$87,$E8,$64,$6C,$E4,$F9,$7E,$9F,$31,$EF,$8B,$8F,$9E,$6E
	.dc.b	$1E,$2C,$15,$8B,$F0,$D7,$FB,$F3,$E9,$DF,$39,$49,$3E,$1F,$2F,$FD
	.dc.b	$B7,$97,$FB,$EE,$B7,$3E,$56,$BF,$ED,$F6,$C7,$FC,$01,$9E,$39,$7C
	.dc.b	$47,$FB,$BD,$7F,$B9,$5B,$3B,$E7,$F9,$2F,$1E,$7B,$7C,$66,$DF,$E0
	.dc.b	$2F,$8C,$E6,$3D,$15,$A5,$3F,$E6,$57,$D9,$E0,$66,$31,$1F,$EF,$7F
	.dc.b	$A7,$FF,$33,$3D,$67,$44,$7E,$A4,$69,$EB,$19,$7D,$8F,$86,$35,$F6
	.dc.b	$1E,$29,$FF,$3C,$9F,$98,$70,$E1,$AF,$F9,$F9,$AF,$E3,$3E,$A1,$E8
	.dc.b	$3D,$3E,$72,$9C,$C6,$17,$82,$FA,$8C,$5F,$1B,$FF,$45,$EE,$FA,$3F
	.dc.b	$C3,$EF,$F9,$78,$C7,$22,$31,$2E,$4C,$5A,$E9,$C2,$B7,$78,$7F,$65
	.dc.b	$7D,$AB,$84,$62,$F5,$1C,$E7,$57,$D3,$E9,$1F,$4F,$A8,$46,$30,$F4
	.dc.b	$7C,$07,$C5,$7D,$51,$9C,$3B,$9E,$77,$15,$D6,$FA,$5F,$1A,$DC,$F6
	.dc.b	$9F,$05,$70,$BF,$E1,$8D,$AF,$85,$7C,$6F,$A5,$7E,$EF,$E9,$9F,$4F
	.dc.b	$E9,$FB,$07,$A3,$F0,$31,$56,$AF,$E9,$56,$9E,$57,$F4,$2B,$FD,$87
	.dc.b	$5B,$7F,$1A,$D7,$31,$B7,$CF,$EB,$FA,$8F,$5F,$C4,$5B,$F8,$F1,$67
	.dc.b	$E0,$E8,$AC,$67,$F7,$B6,$7C,$1F,$D3,$5F,$05,$63,$0D,$CB,$F1,$2D
	.dc.b	$C7,$E6,$D7,$09,$FC,$6D,$70,$69,$C0,$FC,$A7,$D4,$7F,$9B,$1B,$E5
	.dc.b	$7E,$0E,$1F,$51,$63,$F0,$DE,$83,$E9,$E9,$8F,$12,$E1,$F9,$F9,$7D
	.dc.b	$37,$03,$8F,$D0,$59,$FD,$27,$D4,$3A,$3F,$24,$F5,$8C,$C6,$31,$CD
	.dc.b	$EA,$ED,$72,$7F,$35,$2F,$26,$F9,$58,$6D,$7E,$CF,$EA,$1F,$29,$FA
	.dc.b	$16,$C6,$71,$C4,$FE,$97,$EA,$3A,$4F,$EE,$DB,$C9,$F9,$C7,$FD,$95
	.dc.b	$19,$C3,$F5,$9E,$93,$E1,$29,$8E,$BB,$5F,$C1,$7E,$A1,$F3,$1F,$B3
	.dc.b	$7A,$F5,$FF,$02,$73,$3B,$C7,$F7,$1F,$51,$D6,$3F,$87,$2C,$33,$8E
	.dc.b	$9D,$F9,$1C,$3C,$7B,$7C,$1F,$D4,$7D,$43,$E2,$7F,$55,$F5,$0F,$7B
	.dc.b	$31,$87,$F5,$DF,$51,$F6,$70,$FE,$DD,$F2,$AF,$93,$CF,$67,$EC,$E3
	.dc.b	$13,$FD,$CE,$9D,$9C,$BE,$0F,$9B,$67,$FB,$C7,$D3,$BE,$9F,$F7,$9F
	.dc.b	$50,$F8,$3F,$DF,$4E,$6B,$FB,$F4,$E5,$FF,$00,$7D,$47,$CA,$7F,$C1
	.dc.b	$73,$BE,$DF,$E3,$22,$74,$FF,$88,$A3,$5D,$5F,$C1,$7D,$7F,$DC,$5F
	.dc.b	$4E,$AF,$E1,$D1,$8B,$E7,$87,$25,$F4,$EF,$A8,$FE,$5D,$F5,$1F,$EF
	.dc.b	$3C,$7F,$C2,$93,$97,$FA,$34,$66,$BC,$B3,$97,$F9,$4C,$67,$2F,$E3
	.dc.b	$13,$97,$FA,$3C,$66,$BF,$C5,$33,$97,$FC,$71,$19,$AF,$F1,$FC,$E5
	.dc.b	$FE,$A1,$19,$7D,$3C,$1F,$D1,$7E,$9D,$F5,$1F,$CB,$6D,$39,$8C,$37
	.dc.b	$0F,$6B,$6F,$D7,$E2,$7F,$44,$FA,$2E,$4D,$AF,$86,$E1,$F8,$B3,$4F
	.dc.b	$AF,$9A,$BF,$9F,$2F,$A8,$B1,$86,$D7,$F9,$06,$78,$3F,$E4,$3A,$FF
	.dc.b	$2F,$FD,$47,$F9,$83,$E9,$FF,$3D,$1F,$E5,$DC,$3F,$A5,$3A,$11,$A7
	.dc.b	$FB,$B4,$7F,$78,$8C,$46,$DD,$5F,$82,$38,$F6,$7F,$86,$2F,$C0,$FF
	.dc.b	$8D,$AF,$93,$99,$D9,$E3,$C3,$6F,$F9,$90,$56,$1F,$1C,$57,$C5,$3D
	.dc.b	$63,$11,$C0,$EA,$70,$29,$C0,$F0,$18,$38,$61,$EC,$BF,$3C,$78,$EF
	.dc.b	$C5,$87,$BE,$BE,$1C,$F7,$3F,$C5,$B3,$FE,$2E,$7E,$61,$83,$FC,$77
	.dc.b	$F4,$FF,$C8,$BE,$A3,$FC,$66,$D8,$CF,$90,$9C,$B6,$38,$FD,$3F,$F8
	.dc.b	$A5,$B8,$F2,$8C,$47,$C9,$B4,$62,$3F,$D1,$F1,$89,$F9,$9A,$7F,$46
	.dc.b	$9B,$B7,$FA,$53,$3F,$C1,$5E,$E7,$E5,$B6,$3F,$0D,$5C,$67,$E7,$B6
	.dc.b	$38,$E1,$E7,$6E,$1D,$AA,$1D,$27,$55,$E6,$6C,$7F,$AE,$23,$A1,$FE
	.dc.b	$A8,$2B,$FE,$BC,$BF,$24,$71,$AE,$B7,$E6,$C3,$D9,$39,$BF,$3C,$7F
	.dc.b	$A0,$63,$D5,$5F,$E2,$76,$EB,$BB,$E5,$7F,$EA,$F9,$CB,$E5,$BF,$F4
	.dc.b	$56,$1E,$28,$FF,$67,$FD,$47,$EE,$DF,$F5,$B3,$5F,$E6,$AD,$C7,$FC
	.dc.b	$6C,$FD,$88,$DE,$DF,$E9,$58,$D6,$5F,$EC,$86,$CF,$6C,$FF,$D3,$5F
	.dc.b	$4F,$FE,$31,$8F,$F6,$23,$5F,$E8,$6D,$C3,$FC,$89,$F5,$1F,$EB,$09
	.dc.b	$FF,$72,$CE,$70,$FF,$6A,$B6,$33,$EE,$7F,$A9,$EB,$84,$25,$E1,$78
	.dc.b	$89,$1F,$ED,$59,$7C,$F6,$23,$C7,$37,$1F,$1C,$DC,$7C,$33,$6B,$FD
	.dc.b	$F3,$7C,$9E,$9F,$05,$E3,$23,$FE,$93,$F8,$5F,$F5,$BE,$3E,$E9,$FF
	.dc.b	$5F,$7D,$47,$F3,$CA,$FF,$74,$35,$FE,$CF,$7F,$E6,$E7,$3D,$1C,$08
	.dc.b	$F8,$0F,$67,$8F,$F0,$C8,$FF,$BC,$FE,$A3,$FC,$BE,$FF,$BD,$7C,$AF
	.dc.b	$F2,$FB,$7E,$07,$FD,$A1,$18,$8F,$0D,$63,$CB,$91,$C5,$DD,$7F,$B7
	.dc.b	$34,$7D,$3F,$D0,$AF,$F5,$93,$5F,$F4,$9B,$63,$FD,$5E,$BE,$57,$97
	.dc.b	$D3,$7E,$3E,$BF,$D3,$2D,$AF,$F5,$AB,$78,$93,$FE,$B9,$8F,$F1,$DB
	.dc.b	$FE,$C1,$FA,$5F,$ED,$2D,$F0,$BC,$D1,$88,$FF,$52,$BF,$E1,$D7,$CF
	.dc.b	$F7,$9E,$BC,$BF,$C4,$51,$EA,$F3,$B1,$3E,$6E,$CF,$3F,$89,$2B,$FD
	.dc.b	$A1,$97,$E8,$B4,$65,$FE,$79,$9F,$27,$C0,$F8,$8F,$1F,$BF,$DF,$E5
	.dc.b	$E4,$FC,$87,$C5,$F1,$3F,$13,$8E,$4F,$67,$E3,$FD,$6F,$B3,$CD,$93
	.dc.b	$F0,$5F,$7F,$37,$AF,$E1,$7D,$9F,$0D,$2B,$EA,$C7,$D0,$F9,$EF,$A3
	.dc.b	$F3,$B2,$7D,$71,$F2,$3E,$63,$E4,$FC,$C4,$9F,$9D,$1F,$49,$FF,$1E
	.dc.b	$FE,$16,$F1,$31,$C9,$FE,$BF,$8F,$F7,$84,$FD,$88,$C7,$C9,$EB,$B7
	.dc.b	$95,$71,$AD,$46,$2F,$51,$F8,$09,$E6,$F8,$2B,$2A,$39,$3E,$49,$CC
	.dc.b	$70,$BF,$0C,$F6,$3A,$4F,$2F,$A2,$7D,$2D,$24,$EC,$47,$F8,$43,$1F
	.dc.b	$84,$F0,$8D,$35,$FF,$30,$B6,$FF,$DC,$4F,$18,$FC,$DF,$D2,$3E,$97
	.dc.b	$FB,$C3,$63,$FC,$24,$6A,$32,$47,$BE,$BE,$93,$E1,$CF,$18,$F0,$FC
	.dc.b	$E7,$C6,$FA,$5F,$F7,$96,$3F,$EF,$36,$E5,$EA,$9F,$84,$6D,$95,$EC
	.dc.b	$8C,$BF,$0E,$78,$BF,$B4,$6B,$F1,$A3,$C1,$F4,$72,$38,$3E,$3E,$3C
	.dc.b	$47,$FA,$1A,$93,$8C,$3C,$D5,$F7,$BE,$97,$F9,$35,$FF,$8D,$5A,$FE
	.dc.b	$4D,$AF,$F0,$66,$A7,$F9,$74,$FC,$1F,$8E,$F8,$87,$5F,$2F,$F1,$76
	.dc.b	$E5,$BC,$D3,$1F,$C1,$4E,$5B,$39,$7F,$D3,$F0,$9E,$73,$27,$8C,$6F
	.dc.b	$06,$BF,$AB,$BE,$95,$CB,$E0,$7C,$64,$7C,$3F,$25,$5D,$1E,$48,$F4
	.dc.b	$C7,$E9,$3E,$2D,$F2,$9E,$CE,$99,$7F,$9B,$9C,$B6,$72,$FA,$A7,$94
	.dc.b	$7E,$BB,$E9,$7F,$35,$97,$F0,$16,$D7,$F3,$46,$A7,$F3,$CD,$E8,$E4
	.dc.b	$47,$F5,$91,$A3,$89,$23,$C1,$EE,$78,$7B,$3C,$5D,$12,$F8,$C6,$5E
	.dc.b	$71,$97,$FD,$A3,$69,$F8,$9C,$1F,$C2,$3E,$97,$F2,$E7,$E6,$35,$FB
	.dc.b	$17,$F1,$67,$59,$C8,$C1,$FC,$0E,$BF,$2E,$DE,$6E,$87,$5E,$1B,$19
	.dc.b	$21,$FE,$E4,$94,$6A,$31,$37,$8B,$F0,$B1,$FD,$82,$B6,$F2,$8C,$A8
	.dc.b	$F1,$BC,$8F,$35,$73,$BF,$E7,$13,$AA,$E2,$D8,$CC,$EE,$BC,$17,$5E
	.dc.b	$D1,$6B,$84,$68,$EE,$79,$6B,$BD,$6E,$B8,$36,$74,$71,$3D,$07,$C5
	.dc.b	$8D,$BA,$3C,$25,$AF,$0B,$F0,$67,$E8,$39,$7B,$3D,$2F,$DF,$18,$6C
	.dc.b	$6C,$F4,$9B,$BF,$4B,$DE,$FC,$6E,$4F,$A8,$E2,$79,$C7,$A4,$E8,$6C
	.dc.b	$F4,$96,$FD,$24,$9E,$55,$E0,$9E,$46,$EF,$D1,$19,$BE,$E4,$7E,$CC
	.dc.b	$F8,$23,$0F,$98,$C4,$71,$36,$F5,$3D,$C6,$8F,$A9,$1F,$86,$AF,$2C
	.dc.b	$62,$78,$CF,$68,$FA,$47,$23,$A1,$EF,$9E,$2E,$8F,$6C,$71,$BF,$87
	.dc.b	$7F,$6A,$78,$CE,$6B,$8D,$7C,$79,$F3,$1C,$E7,$CC,$E2,$34,$F2,$BF
	.dc.b	$80,$FC,$33,$07,$AC,$C4,$F8,$EF,$E7,$C6,$4E,$D3,$C9,$FB,$73,$B7
	.dc.b	$E0,$35,$F8,$07,$8E,$7D,$91,$93,$D9,$19,$3F,$22,$E6,$7B,$57,$F0
	.dc.b	$09,$F3,$9E,$23,$CF,$1A,$AF,$3C,$64,$93,$F0,$AB,$A4,$7A,$E3,$A4
	.dc.b	$7F,$02,$F3,$FF,$92,$73,$4A,$FE,$3D,$1F,$C9,$CC,$3F,$C6,$63,$0F
	.dc.b	$F8,$97,$E9,$5E,$37,$F8,$D3,$E9,$7F,$95,$CB,$0F,$E2,$F1,$87,$F8
	.dc.b	$CD,$70,$9E,$55,$C7,$59,$18,$3F,$C6,$56,$E7,$ED,$CF,$27,$31,$E1
	.dc.b	$30,$F8,$76,$FF,$2C,$52,$B5,$3C,$0E,$AF,$0C,$BC,$F7,$EB,$F0,$25
	.dc.b	$78,$9F,$77,$46,$23,$B1,$FE,$96,$C3,$FC,$A7,$7F,$02,$B7,$3F,$E4
	.dc.b	$BF,$A5,$79,$5F,$E5,$AF,$A5,$FE,$CE,$7F,$68,$8C,$CE,$E7,$5D,$32
	.dc.b	$30,$7F,$95,$6B,$FA,$F3,$79,$FA,$1D,$7A,$6C,$64,$87,$A0,$4A,$37
	.dc.b	$19,$8E,$AD,$9E,$8F,$87,$C0,$F5,$11,$FF,$34,$7D,$2F,$F9,$12,$BF
	.dc.b	$DC,$7C,$3F,$E8,$96,$CF,$FA,$F1,$AA,$7F,$C9,$7F,$48,$F9,$47,$C6
	.dc.b	$FF,$99,$29,$39,$8C,$2E,$D7,$D2,$62,$F8,$DF,$F9,$B3,$EB,$1F,$4B
	.dc.b	$FD,$AA,$3F,$D3,$91,$8D,$3B,$36,$38,$4E,$DF,$CE,$7D,$17,$FA,$86
	.dc.b	$BF,$D1,$9C,$A8,$C4,$FE,$12,$31,$CF,$FD,$F1,$93,$A6,$F3,$66,$FF
	.dc.b	$D5,$13,$96,$D6,$6F,$C3,$56,$FC,$0D,$BE,$13,$E6,$AC,$4E,$FE,$8F
	.dc.b	$D9,$7E,$36,$E1,$FE,$B7,$F3,$FE,$7B,$EC,$8C,$E1,$FB,$B6,$D1,$B5
	.dc.b	$C2,$7E,$04,$70,$5F,$48,$F5,$3E,$54,$61,$B1,$9C,$3E,$F3,$67,$FD
	.dc.b	$46,$DC,$3F,$C4,$2D,$9F,$8F,$E4,$7C,$5A,$E9,$1E,$9F,$9B,$F8,$07
	.dc.b	$C1,$F8,$8B,$19,$24,$FA,$23,$2F,$AA,$6E,$1F,$1E,$6D,$7F,$B7,$7E
	.dc.b	$8F,$F9,$0B,$F8,$27,$C9,$7F,$C8,$5C,$BF,$83,$73,$5F,$E7,$6F,$80
	.dc.b	$FB,$58,$BE,$62,$C6,$1C,$F9,$BA,$EF,$B9,$7E,$42,$D1,$F8,$3F,$A3
	.dc.b	$FE,$1B,$3F,$94,$6E,$7F,$93,$58,$EF,$C1,$E1,$8C,$BE,$18,$5C,$3F
	.dc.b	$35,$ED,$BF,$02,$FC,$83,$6B,$F4,$2B,$EF,$F3,$3F,$95,$FA,$57,$CE
	.dc.b	$DC,$B0,$DC,$3F,$53,$F4,$9C,$F2,$FD,$43,$CE,$33,$87,$E7,$E3,$13
	.dc.b	$FA,$DF,$3D,$C9,$FA,$37,$F2,$A6,$2F,$F6,$AF,$E6,$3E,$93,$FA,$14
	.dc.b	$7E,$E1,$B5,$FB,$77,$C5,$C9,$FB,$77,$2F,$EE,$7E,$93,$A4,$FE,$F5
	.dc.b	$B1,$9E,$FF,$DA,$46,$2D,$5F,$D8,$E1,$F1,$7A,$37,$F5,$D6,$B7,$87
	.dc.b	$F9,$FB,$E8,$FE,$EC,$60,$E7,$82,$32,$FF,$3B,$FA,$4E,$B3,$B8,$F0
	.dc.b	$CE,$AB,$FD,$13,$F2,$6F,$87,$FD,$43,$6F,$85,$BE,$91,$E1,$EA,$FF
	.dc.b	$5D,$0F,$7E,$98,$E5,$B8,$72,$FA,$37,$C7,$FE,$D7,$E9,$3F,$3D,$39
	.dc.b	$8C,$78,$5F,$B9,$FA,$4F,$EF,$D3,$93,$C6,$E3,$5F,$EB,$18,$74,$9F
	.dc.b	$3F,$A3,$F3,$E7,$E7,$D6,$F8,$3F,$D4,$F3,$99,$4E,$6B,$E5,$FD,$13
	.dc.b	$E3,$3F,$96,$7D,$27,$FB,$26,$7E,$1C,$E5,$FE,$7F,$19,$BF,$ED,$53
	.dc.b	$97,$FA,$04,$67,$1F,$F5,$BC,$E5,$FE,$85,$19,$AF,$0D,$F8,$3D,$8F
	.dc.b	$E3,$57,$B9,$FC,$1D,$6F,$0F,$C5,$CF,$D1,$9F,$EA,$3F,$49,$FB,$C7
	.dc.b	$CF,$C3,$FD,$16,$31,$5E,$68,$C7,$8F,$D9,$B5,$EE,$FA,$0F,$E9,$9C
	.dc.b	$CF,$94,$F1,$C6,$1B,$3B,$3E,$2E,$18,$3F,$C1,$9B,$71,$8F,$C5,$AF
	.dc.b	$94,$FE,$6C,$E6,$77,$18,$A9,$5F,$92,$9F,$D4,$FD,$27,$EA,$BE,$8B
	.dc.b	$F8,$FC,$7E,$9A,$BF,$A0,$3A,$11,$A7,$FC,$A5,$1F,$EE,$C8,$C4,$6D
	.dc.b	$D6,$1E,$37,$30,$AF,$07,$97,$FB,$0A,$D3,$FB,$29,$E4,$E6,$FB,$39
	.dc.b	$8E,$47,$91,$D6,$1D,$23,$95,$D7,$CF,$15,$EA,$7B,$DC,$3D,$30,$F4
	.dc.b	$DF,$82,$39,$9D,$CE,$05,$3F,$84,$D7,$7C,$3D,$71,$C0,$FF,$17,$47
	.dc.b	$F8,$BD,$F9,$86,$0F,$67,$D1,$3E,$8D,$F4,$9F,$E3,$58,$C3,$63,$33
	.dc.b	$C1,$AF,$C9,$97,$D1,$FF,$A1,$E3,$AB,$5D,$35,$FE,$43,$1C,$30,$F0
	.dc.b	$4F,$1C,$BF,$01,$A3,$6F,$E4,$F1,$86,$C7,$29,$E0,$DA,$FA,$4F,$1C
	.dc.b	$2D,$70,$7F,$9D,$E1,$E1,$7E,$63,$70,$ED,$1C,$0F,$B9,$19,$AE,$D1
	.dc.b	$F4,$DE,$CD,$AE,$CA,$78,$3E,$0C,$3C,$EF,$F4,$CA,$E3,$7E,$08,$F1
	.dc.b	$CE,$9F,$F0,$2D,$FE,$67,$5F,$E9,$37,$E0,$C5,$18,$9D,$57,$79,$F5
	.dc.b	$3F,$E8,$7A,$E1,$5B,$8F,$E8,$F8,$F9,$97,$D2,$7F,$09,$7F,$A5,$B5
	.dc.b	$FF,$48,$36,$FC,$31,$FD,$9A,$3F,$06,$F0,$DB,$C6,$F4,$6C,$71,$BF
	.dc.b	$EC,$6F,$09,$FF,$4E,$7D,$27,$F8,$11,$FE,$B6,$D7,$3E,$5F,$4F,$A1
	.dc.b	$FF,$0A,$C6,$A3,$BC,$EF,$1F,$2C,$A3,$18,$7F,$72,$BE,$33,$E0,$F1
	.dc.b	$38,$91,$FE,$D7,$0F,$A4,$C4,$6C,$F2,$CF,$8E,$33,$87,$83,$04,$E3
	.dc.b	$0F,$05,$7A,$7E,$92,$61,$D9,$FB,$38,$3D,$DF,$F5,$6E,$49,$F9,$7F
	.dc.b	$49,$FD,$E6,$FF,$BF,$B5,$FF,$62,$4E,$DC,$F4,$F0,$23,$DC,$F6,$F4
	.dc.b	$FC,$B2,$3F,$E0,$AF,$A4,$7D,$3F,$DA,$F3,$9F,$4B,$49,$5C,$8F,$F0
	.dc.b	$8C,$7F,$85,$5B,$E6,$71,$36,$FF,$AB,$7E,$97,$E2,$BF,$EB,$26,$B9
	.dc.b	$E2,$F8,$70,$F9,$5F,$8B,$F4,$9F,$87,$E2,$FF,$5A,$37,$93,$FC,$62
	.dc.b	$DC,$12,$8F,$E4,$11,$FE,$C0,$FA,$4F,$F3,$5B,$7C,$3F,$A3,$18,$8F
	.dc.b	$F6,$13,$FE,$7D,$7C,$DF,$4B,$E4,$F3,$A4,$60,$FF,$65,$DF,$E7,$5F
	.dc.b	$0B,$FD,$A5,$CF,$8B,$EA,$ED,$EF,$EF,$78,$3D,$F9,$5E,$58,$F8,$3F
	.dc.b	$2F,$23,$E3,$3F,$13,$E1,$BE,$1F,$1C,$9F,$0B,$F1,$BE,$97,$D7,$E7
	.dc.b	$C9,$F8,$0F,$BB,$AB,$E6,$FC,$E7,$CE,$FA,$59,$5F,$4E,$3E,$7F,$CF
	.dc.b	$FD,$0F,$9B,$93,$EE,$8F,$8F,$E9,$FC,$8F,$97,$93,$F3,$63,$F8,$0F
	.dc.b	$D2,$FE,$6E,$3F,$CA,$AB,$E3,$F2,$6B,$C7,$93,$A8,$CB,$C2,$32,$FD
	.dc.b	$CE,$CF,$F3,$5D,$73,$8E,$06,$CF,$F3,$1F,$D2,$BE,$23,$F8,$9B,$63
	.dc.b	$31,$8F,$3F,$2C,$8C,$9C,$A3,$F8,$DF,$D2,$3E,$96,$92,$7B,$F9,$95
	.dc.b	$FF,$05,$4E,$DF,$F0,$60,$70,$EA,$7F,$BC,$60,$F9,$BF,$E5,$3F,$4B
	.dc.b	$FC,$6B,$C4,$FE,$36,$DD,$7F,$D1,$75,$BC,$3D,$F7,$C2,$64,$60,$87
	.dc.b	$C9,$3F,$D3,$B5,$CD,$B1,$F1,$A7,$C3,$F4,$72,$7F,$D1,$B4,$8F,$2F
	.dc.b	$2E,$67,$A0,$9D,$3C,$27,$6F,$FA,$5B,$E9,$7F,$8C,$65,$FC,$D9,$B8
	.dc.b	$FF,$A9,$9B,$3E,$6F,$27,$FA,$5C,$7F,$1F,$8F,$F1,$D3,$5E,$1E,$77
	.dc.b	$F8,$86,$31,$1F,$EB,$AC,$9C,$FF,$D5,$17,$45,$31,$12,$7F,$D5,$73
	.dc.b	$A6,$CE,$5F,$7C,$74,$72,$43,$33,$DF,$E8,$64,$74,$7C,$5F,$45,$88
	.dc.b	$E7,$17,$4F,$EB,$36,$AF,$ED,$7F,$3A,$F9,$5A,$4C,$3C,$ED,$7F,$CB
	.dc.b	$51,$B7,$F2,$2D,$73,$1F,$E4,$7F,$A5,$FD,$24,$7D,$A6,$E1,$FD,$A9
	.dc.b	$BB,$FF,$B6,$9F,$F3,$8B,$7A,$3F,$BF,$C6,$AB,$FD,$D2,$D4,$F0,$73
	.dc.b	$1C,$08,$FF,$B6,$A3,$4F,$0E,$CE,$BD,$B2,$35,$1E,$03,$0F,$A6,$30
	.dc.b	$FF,$B9,$32,$FB,$4F,$E5,$9B,$8F,$F8,$19,$B8,$FB,$7E,$1F,$E7,$F0
	.dc.b	$4B,$E4,$FB,$A3,$FC,$3B,$F4,$BF,$9E,$9F,$D0,$B7,$0F,$F0,$8B,$6F
	.dc.b	$FE,$00,$AD,$F8,$F2,$CB,$FF,$7C,$CE,$D0,$78,$3A,$8E,$0F,$81,$F1
	.dc.b	$CE,$2F,$C3,$16,$FC,$33,$6B,$AD,$F7,$7B,$57,$27,$B1,$0E,$C7,$9A
	.dc.b	$79,$3C,$0D,$5F,$C0,$8C,$D7,$52,$38,$AC,$B6,$B8,$DF,$99,$A7,$9A
	.dc.b	$76,$E0,$F0,$1E,$43,$E8,$CF,$37,$17,$E3,$78,$5F,$8E,$BD,$D1,$AB
	.dc.b	$FA,$B1,$DA,$30,$78,$4E,$37,$D6,$FE,$6C,$7C,$33,$AC,$F1,$8C,$C7
	.dc.b	$23,$CE,$F8,$DC,$5F,$8D,$E4,$62,$79,$5F,$92,$3A,$9A,$AE,$A4,$8D
	.dc.b	$47,$C2,$36,$76,$BF,$44,$6A,$7C,$47,$19,$F7,$1B,$38,$B9,$7F,$25
	.dc.b	$3C,$9D,$C7,$29,$E8,$48,$CB,$F6,$DF,$61,$F2,$8E,$AE,$DE,$A6,$0E
	.dc.b	$27,$E5,$A7,$C0,$FC,$A8,$E9,$3E,$5B,$F9,$53,$F0,$4E,$E7,$AC,$B1
	.dc.b	$BA,$E7,$3F,$94,$9F,$01,$D4,$F0,$18,$74,$F2,$AF,$44,$7E,$D2,$30
	.dc.b	$7A,$0F,$59,$83,$A1,$8A,$F4,$5F,$DE,$8E,$13,$EA,$9F,$99,$18,$9D
	.dc.b	$3C,$A7,$55,$F0,$8F,$1C,$EA,$33,$3D,$2F,$DF,$3F,$A0,$BF,$0C,$FD
	.dc.b	$33,$F8,$CF,$A1,$FC,$73,$E4,$A4,$7F,$14,$8F,$E3,$E6,$1F,$E3,$F1
	.dc.b	$87,$F8,$BC,$FF,$AC,$3E,$95,$E4,$FF,$B0,$9B,$3A,$CF,$F8,$EA,$8C
	.dc.b	$BF,$E0,$AF,$A5,$7D,$77,$1B,$93,$C4,$FE,$4B,$5C,$30,$FF,$5F,$DF
	.dc.b	$09,$D0,$5C,$FF,$ED,$47,$FB,$64,$69,$AF,$86,$34,$6F,$9E,$46,$E3
	.dc.b	$2F,$18,$CB,$C4,$E9,$E2,$77,$23,$DF,$C5,$AF,$BE,$2E,$1E,$A9,$FF
	.dc.b	$74,$FD,$33,$CE,$A4,$BF,$6F,$D3,$7F,$B6,$AF,$FD,$BF,$5B,$9F,$F2
	.dc.b	$7B,$5F,$9E,$7F,$4C,$70,$E1,$AF,$D4,$97,$4F,$FB,$E9,$B1,$E0,$8C
	.dc.b	$74,$E5,$91,$93,$CB,$18,$7F,$C8,$B3,$C9,$B3,$C9,$F5,$F8,$5D,$C8
	.dc.b	$FD,$AF,$A6,$FF,$2B,$2A,$FF,$30,$3D,$27,$6F,$F9,$4A,$77,$F5,$5B
	.dc.b	$8F,$F3,$8B,$5F,$86,$1C,$3F,$19,$AF,$E5,$BE,$9D,$F4,$CF,$91,$FF
	.dc.b	$3D,$C6,$1B,$DE,$FA,$7C,$D6,$EF,$E8,$EE,$FA,$6F,$F7,$F7,$57,$E6
	.dc.b	$E3,$1C,$88,$C4,$FF,$A9,$DB,$19,$AF,$F3,$D5,$70,$9C,$9C,$39,$35
	.dc.b	$12,$FB,$D7,$06,$C6,$6F,$77,$FB,$A9,$F0,$56,$E7,$2F,$FB,$B3,$E9
	.dc.b	$9F,$3B,$8C,$11,$8C,$3D,$9E,$9F,$C7,$7D,$D1,$9B,$F4,$B6,$BC,$ED
	.dc.b	$9F,$E1,$75,$C6,$B7,$7E,$25,$F4,$CF,$63,$C5,$F5,$4F,$4F,$53,$9C
	.dc.b	$7E,$74,$E5,$FE,$0B,$19,$BF,$C1,$B5,$FD,$3B,$5F,$E2,$ED,$7E,$17
	.dc.b	$D3,$FC,$1C,$CD,$F8,$CC,$32,$FE,$0B,$96,$92,$BE,$1C,$65,$B1,$F8
	.dc.b	$5F,$A7,$FE,$56,$DC,$FE,$84,$62,$7F,$06,$FA,$F6,$FC,$0F,$9C,$F8
	.dc.b	$4F,$AC,$BC,$CB,$18,$73,$E7,$EB,$D8,$B9,$FE,$56,$D1,$F7,$FE,$9F
	.dc.b	$F8,$A4,$FE,$65,$B9,$FE,$51,$F0,$46,$7C,$4E,$18,$48,$E1,$76,$FF
	.dc.b	$37,$EF,$BE,$97,$F1,$8D,$AF,$D3,$2F,$A7,$CC,$FE,$47,$E9,$DF,$41
	.dc.b	$72,$C3,$70,$FD,$77,$D3,$F3,$CB,$F5,$2F,$38,$CE,$1F,$9B,$8C,$4F
	.dc.b	$EB,$BD,$17,$37,$E7,$DF,$CB,$43,$8D,$1A,$C3,$F7,$8E,$A1,$F4,$FF
	.dc.b	$EB,$B9,$FE,$00,$DA,$FD,$DB,$E5,$9F,$DC,$B9,$7F,$77,$F4,$FD,$67
	.dc.b	$F8,$0B,$63,$3D,$BF,$B3,$8C,$5F,$F7,$45,$F1,$D9,$35,$F4,$F6,$7F
	.dc.b	$B3,$FD,$37,$DE,$6F,$27,$EB,$FE,$9F,$8C,$60,$FE,$DB,$F0,$2F,$8A
	.dc.b	$FD,$A3,$63,$FB,$8F,$D3,$BC,$5C,$D7,$EE,$61,$F0,$74,$C7,$35,$BF
	.dc.b	$EF,$3F,$4C,$F9,$0F,$E0,$1F,$4F,$F9,$99,$CC,$63,$C1,$FE,$09,$F4
	.dc.b	$FF,$DF,$A7,$27,$F8,$09,$C7,$37,$D7,$8F,$F0,$8E,$1D,$5F,$F0,$3F
	.dc.b	$A3,$F2,$27,$FD,$63,$5C,$39,$FF,$98,$47,$F4,$1E,$AE,$FE,$DF,$F1
	.dc.b	$3F,$A6,$7C,$7F,$F2,$3F,$A7,$FE,$D3,$9F,$F8,$4A,$72,$FF,$3F,$8C
	.dc.b	$E1,$E3,$3F,$D7,$F3,$A7,$F9,$EC,$6B,$1F,$0C,$E5,$FE,$7B,$19,$AF
	.dc.b	$F1,$5C,$E5,$FF,$19,$C6,$6B,$AD,$F9,$7D,$BF,$47,$17,$F3,$9F,$A7
	.dc.b	$FD,$BD,$78,$C6,$3F,$5B,$1F,$EA,$D3,$AC,$BC,$77,$C7,$CA,$E7,$5A
	.dc.b	$7E,$27,$D1,$7F,$45,$E8,$7C,$A6,$2B,$FA,$2B,$73,$F0,$66,$BF,$99
	.dc.b	$19,$CB,$E8,$9C,$AF,$FC,$7A,$BE,$57,$FA,$23,$99,$F5,$39,$9D,$E1
	.dc.b	$A3,$FA,$54,$FF,$BB,$BE,$9F,$FD,$DF,$F4,$BF,$C7,$A3,$FC,$B7,$3A
	.dc.b	$53,$AE,$FF,$F2,$CE,$1D,$EF,$85,$7F,$97,$2B,$4F,$2C,$2B,$91,$3C
	.dc.b	$2B,$95,$FF,$85,$2F,$71,$8A,$FF,$1C,$C6,$30,$FF,$01,$5F,$82,$7D
	.dc.b	$47,$4B,$F2,$CE,$9E,$91,$93,$C8,$53,$C9,$3E,$39,$E1,$1E,$03,$57
	.dc.b	$E3,$77,$5D,$B0,$F6,$C7,$88,$FF,$16,$9F,$20,$A5,$33,$86,$BE,$91
	.dc.b	$F4,$CF,$A7,$C3,$6B,$FC,$6F,$3A,$9C,$9C,$7E,$9F,$F9,$94,$E9,$AF
	.dc.b	$F9,$82,$51,$88,$FF,$48,$C6,$32,$E4,$7D,$6C,$BF,$25,$37,$6F,$F4
	.dc.b	$BE,$BF,$7A,$F8,$4F,$C9,$6C,$7E,$0A,$B8,$CF,$CD,$6C,$71,$C3,$CE
	.dc.b	$DC,$3B,$54,$3A,$4E,$AB,$CE,$D8,$1F,$DD,$30,$F2,$CF,$FA,$B0,$EE
	.dc.b	$70,$AF,$C8,$E1,$E6,$8D,$4F,$0A,$E7,$1C,$1E,$B3,$B7,$33,$FA,$0D
	.dc.b	$38,$C7,$FA,$BA,$BD,$F7,$FC,$9A,$F8,$4F,$94,$D5,$F1,$79,$9F,$EB
	.dc.b	$68,$DF,$D3,$FF,$06,$8F,$F5,$CB,$5F,$E2,$8D,$AF,$E6,$6F,$DC,$8D
	.dc.b	$ED,$D5,$46,$F1,$FF,$67,$B6,$3E,$4B,$C5,$FE,$41,$F4,$FF,$DF,$9F
	.dc.b	$F6,$4B,$5C,$F5,$75,$FA,$5F,$F5,$85,$FF,$B6,$AB,$58,$FF,$46,$6C
	.dc.b	$66,$B8,$78,$5E,$59,$78,$B4,$98,$FF,$B6,$31,$7D,$3E,$23,$CB,$37
	.dc.b	$2F,$EA,$75,$A9,$F3,$CE,$72,$F2,$3E,$7B,$F4,$7B,$D2,$BB,$3F,$4F
	.dc.b	$07,$BF,$FE,$AC,$7A,$C7,$CB,$FA,$7F,$EE,$F5,$FE,$EC,$6B,$FD,$1F
	.dc.b	$2F,$F3,$83,$9E,$CE,$04,$7B,$9E,$DE,$BF,$94,$47,$FD,$EF,$F4,$EF
	.dc.b	$A7,$FB,$5E,$7B,$E9,$E9,$2B,$91,$FE,$FE,$8F,$F0,$93,$7C,$DE,$26
	.dc.b	$DF,$EB,$9E,$0F,$A3,$6F,$F5,$2F,$D3,$FD,$69,$FF,$1D,$35,$FF,$34
	.dc.b	$69,$F8,$BF,$81,$FC,$35,$7F,$AB,$3E,$97,$FC,$A5,$5F,$EB,$96,$BF
	.dc.b	$DA,$67,$7C,$BF,$B6,$87,$DF,$3D,$BF,$CB,$76,$3E,$66,$9F,$99,$9D
	.dc.b	$57,$EF,$B2,$FF,$08,$EB,$F7,$E3,$FD,$83,$AF,$FA,$41,$FF,$65,$35
	.dc.b	$F7,$F4,$7A,$71,$FB,$DA,$FF,$9B,$67,$D1,$EE,$65,$ED,$AF,$9F,$E9
	.dc.b	$BF,$D7,$B1,$83,$F8,$7F,$DE,$FE,$A5,$36,$37,$1F,$3B,$C3,$FC,$66
	.dc.b	$6F,$AB,$A7,$ED,$75,$FE,$C1,$8F,$C5,$E2,$FB,$38,$7F,$B7,$34,$7C
	.dc.b	$BF,$A0,$FA,$5C,$BF,$BD,$FB,$CF,$F7,$D7,$17,$F8,$D3,$EF,$3E,$3E
	.dc.b	$1F,$E1,$96,$C7,$F8,$E2,$77,$97,$D3,$CF,$F7,$37,$F8,$0F,$81,$F3
	.dc.b	$9E,$5F,$13,$F0,$1F,$47,$F2,$3D,$59,$5F,$0C,$FF,$5F,$C7,$F0,$FF
	.dc.b	$BC,$7D,$1F,$F1,$1F,$BB,$7C,$B5,$35,$1F,$C5,$7E,$AB,$FD,$0F,$87
	.dc.b	$36,$BF,$E3,$16,$EF,$FE,$50,$FA,$BF,$EC,$B3,$BD,$7F,$8F,$47,$C6
	.dc.b	$8F,$23,$5E,$51,$DC,$DF,$1C,$8E,$47,$0F,$2B,$9F,$CB,$48,$C1,$0F
	.dc.b	$F2,$25,$A3,$51,$89,$FE,$61,$F5,$5F,$43,$8F,$F4,$0D,$9F,$EA,$ED
	.dc.b	$4F,$E0,$9C,$F8,$7C,$88,$FC,$F8,$D2,$D1,$F5,$7F,$E0,$C6,$C7,$F8
	.dc.b	$41,$AF,$EC,$4F,$EC,$EE,$1C,$3A,$D7,$FC,$FB,$4F,$A0,$F5,$9F,$5E
	.dc.b	$E5,$F8,$E3,$2F,$D9,$8F,$83,$F3,$FF,$D4,$7E,$AF,$FC,$AF,$9F,$F0
	.dc.b	$16,$D7,$C2,$9D,$BF,$8B,$8C,$A6,$E1,$FC,$CE,$B9,$78,$FD,$49,$5C
	.dc.b	$7E,$A2,$BF,$E7,$9B,$FE,$18,$F5,$9F,$24,$FF,$65,$FA,$BF,$E7,$33
	.dc.b	$F1,$5A,$FF,$17,$6B,$FE,$0E,$6A,$7F,$95,$5F,$B7,$8F,$E0,$1D,$7C
	.dc.b	$A9,$5E,$D8,$CB,$FA,$9A,$E2,$FF,$86,$AD,$D3,$FD,$E5,$FD,$5C,$7E
	.dc.b	$BD,$AF,$FA,$55,$DB,$98,$E1,$CD,$23,$07,$1E,$AE,$84,$7A,$73,$D7
	.dc.b	$A4,$5E,$2F,$F5,$FE,$1F,$AB,$7F,$D8,$3D,$0F,$97,$F3,$74,$4C,$3F
	.dc.b	$82,$4E,$5B,$39,$7D,$13,$FE,$15,$9D,$47,$FA,$BD,$B5,$E0,$9E,$0F
	.dc.b	$82,$BF,$C3,$9F,$57,$FA,$92,$74,$63,$C6,$CB,$31,$FE,$2D,$3B,$9E
	.dc.b	$3F,$07,$CB,$A6,$56,$A3,$2F,$08,$CB,$FE,$C3,$A4,$FA,$77,$FF,$17
	.dc.b	$7D,$5F,$E6,$0F,$CC,$EB,$F5,$B4,$F8,$F3,$AE,$59,$8F,$C6,$8E,$4F
	.dc.b	$F2,$BB,$E5,$38,$8F,$89,$3C,$1F,$F6,$A4,$F6,$8D,$57,$FB,$D2,$38
	.dc.b	$CF,$52,$1C,$2B,$94,$F9,$23,$2F,$18,$C9,$FD,$DE,$FD,$4D,$BC,$D7
	.dc.b	$F9,$C2,$FC,$4F,$29,$FE,$FD,$1E,$28,$F4,$99,$38,$47,$27,$FA,$2C
	.dc.b	$FA,$A3,$89,$9A,$F0,$CF,$F9,$F9,$AA,$BB,$5F,$AE,$3B,$9E,$F3,$E6
	.dc.b	$57,$57,$95,$F8,$6B,$D3,$1C,$2B,$E1,$3C,$0F,$41,$E6,$30,$75,$38
	.dc.b	$57,$C5,$73,$1C,$9C,$9E,$03,$31,$EE,$7A,$DF,$B2,$BD,$73,$CE,$78
	.dc.b	$9E,$F7,$E2,$5F,$AE,$30,$74,$3B,$1C,$27,$9B,$8B,$F8,$C7,$94,$95
	.dc.b	$DD,$CB,$E5,$8F,$74,$61,$E5,$19,$8E,$31,$82,$1E,$43,$8C,$75,$23
	.dc.b	$F8,$23,$D0,$7C,$D3,$D0,$FB,$9F,$A3,$3E,$A7,$53,$E3,$7B,$3E,$27
	.dc.b	$33,$D9,$CB,$E7,$31,$7F,$3D,$ED,$1E,$03,$C7,$1F,$8E,$34,$E2,$FE
	.dc.b	$85,$F8,$E7,$51,$F8,$BA,$FA,$91,$C2,$BE,$09,$F1,$1C,$5F,$9D,$E3
	.dc.b	$1F,$0A,$FF,$27,$3E,$D7,$B5,$F8,$2F,$CF,$39,$69,$EF,$9D,$D7,$C8
	.dc.b	$8C,$BF,$52,$32,$FB,$CF,$29,$F2,$0D,$47,$07,$07,$CA,$3E,$C1,$F9
	.dc.b	$4B,$F5,$C7,$CD,$8F,$75,$FC,$37,$EC,$F3,$7F,$26,$23,$FC,$7F,$82
	.dc.b	$BF,$C7,$22,$E3,$FC,$6E,$7F,$96,$FD,$5F,$F2,$EE,$69,$7F,$C9,$E3
	.dc.b	$2F,$F2,$98,$CB,$FC,$9A,$3F,$DC,$5F,$52,$FA,$AF,$F7,$27,$D5,$7F
	.dc.b	$0A,$3F,$85,$C6,$6B,$FC,$B8,$D5,$8B,$4F,$F4,$D5,$71,$D3,$AE,$12
	.dc.b	$FF,$CC,$93,$97,$9C,$78,$39,$BC,$3C,$F2,$77,$19,$78,$C6,$5F,$0E
	.dc.b	$1E,$09,$F4,$F8,$12,$7B,$BE,$CE,$8C,$47,$00,$31,$FF,$6E,$BC,$0F
	.dc.b	$3C,$7F,$AD,$73,$F7,$47,$0F,$AA,$7C,$EE,$72,$FF,$5D,$7D,$57,$F9
	.dc.b	$48,$FF,$29,$CE,$AB,$FD,$08,$D7,$C4,$23,$87,$4F,$38,$ED,$75,$57
	.dc.b	$1C,$3C,$71,$C6,$B5,$7E,$0B,$BC,$F2,$BF,$D2,$93,$FD,$D9,$BD,$39
	.dc.b	$7C,$71,$96,$CF,$49,$EB,$E0,$C9,$E6,$F9,$7E,$AB,$FC,$AB,$7F,$E5
	.dc.b	$39,$F9,$AD,$7D,$D3,$BD,$3D,$9F,$52,$FA,$87,$C9,$7F,$9C,$EF,$EA
	.dc.b	$CE,$63,$10,$8F,$F5,$4F,$D4,$E1,$D4,$71,$FA,$F7,$CF,$7C,$68,$FF
	.dc.b	$77,$46,$31,$E6,$D8,$E1,$3B,$7F,$0B,$0F,$87,$C4,$23,$15,$CA,$73
	.dc.b	$18,$9F,$F4,$67,$13,$EA,$9C,$FF,$8A,$89,$18,$C3,$FD,$D5,$1B,$E6
	.dc.b	$D4,$7F,$BF,$67,$57,$F9,$9A,$E9,$7A,$8F,$D9,$CE,$A7,$AD,$E5,$DD
	.dc.b	$F6,$FA,$8F,$81,$1E,$36,$C7,$E9,$21,$F0,$9F,$52,$38,$69,$CE,$B7
	.dc.b	$8F,$F0,$26,$DF,$0D,$F5,$5E,$5A,$7D,$4B,$EA,$7E,$5B,$71,$FB,$AD
	.dc.b	$9F,$C1,$35,$FF,$5B,$39,$C3,$F4,$2D,$9F,$8D,$E4,$F7,$8C,$57,$EF
	.dc.b	$9A,$F8,$AF,$E2,$6B,$FE,$77,$BF,$F7,$C5,$6B,$E8,$9F,$0B,$E9,$6C
	.dc.b	$FD,$6F,$A9,$7D,$76,$6B,$DC,$3D,$EA,$7D,$1D,$1D,$38,$C3,$E7,$1F
	.dc.b	$C4,$B5,$74,$3E,$A7,$E4,$36,$7F,$0B,$6F,$B2,$CD,$7C,$DF,$9D,$FE
	.dc.b	$65,$5D,$30,$FE,$3B,$18,$7F,$11,$F5,$6F,$A1,$E1,$5F,$57,$A1,$F2
	.dc.b	$D4,$FC,$45,$77,$9E,$97,$F8,$49,$FB,$9F,$6E,$F2,$A9,$F8,$FC,$DF
	.dc.b	$4D,$8C,$38,$FE,$33,$F8,$8C,$4F,$E2,$3E,$B1,$E5,$7E,$1E,$35,$F8
	.dc.b	$37,$C8,$BF,$0B,$6F,$F4,$5F,$5B,$DA,$73,$CB,$F8,$FF,$B8,$D7,$7F
	.dc.b	$F2,$66,$CF,$E7,$6B,$F3,$97,$FC,$CA,$72,$FF,$44,$FC,$67,$EF,$BF
	.dc.b	$CD,$9B,$7F,$B9,$FA,$F7,$D8,$7E,$76,$73,$8F,$DB,$FC,$1D,$EC,$7E
	.dc.b	$85,$B1,$FB,$EB,$F0,$E1,$FC,$EA,$73,$8F,$F4,$4A,$7C,$CF,$EA,$5A
	.dc.b	$7F,$06,$F0,$9E,$E6,$67,$F5,$DF,$61,$F2,$AB,$EB,$B6,$FF,$A7,$F2
	.dc.b	$BE,$6B,$7C,$BF,$B5,$F9,$F7,$C9,$53,$F7,$1F,$61,$FC,$0B,$2F,$C0
	.dc.b	$CE,$9F,$C1,$C3,$E2,$36,$70,$F9,$0B,$3F,$88,$FA,$E7,$D7,$FF,$03
	.dc.b	$FB,$0F,$C9,$4E,$63,$1E,$2F,$F0,$7F,$B0,$7C,$17,$F0,$9C,$38,$DF
	.dc.b	$E4,$A7,$2F,$13,$F2,$BE,$A7,$F3,$1E,$8F,$EC,$B3,$AE,$3F,$ED,$F3
	.dc.b	$9A,$FD,$84,$E6,$BF,$B5,$FD,$73,$EC,$3F,$94,$79,$7F,$ED,$CB,$FD
	.dc.b	$0C,$E5,$FE,$6F,$19,$C7,$FB,$DC,$E5,$FE,$6D,$19,$BE,$B7,$D3,$E0
	.dc.b	$7F,$92,$5E,$E7,$FA,$25,$6F,$0F,$E0,$93,$97,$F9,$A4,$67,$0F,$F0
	.dc.b	$CC,$FD,$19,$FE,$93,$F6,$0F,$98,$EB,$C7,$FD,$06,$31,$3E,$A8,$C4
	.dc.b	$EA,$38,$BE,$4F,$33,$FA,$47,$9C,$FA,$1C,$57,$F5,$68,$C3,$F1,$27
	.dc.b	$39,$7C,$22,$E9,$FE,$16,$36,$FF,$88,$17,$D8,$7F,$45,$72,$6D,$C6
	.dc.b	$12,$FF,$88,$D7,$F8,$E3,$EC,$3F,$C7,$3F,$5D,$FC,$22,$3F,$AC,$57
	.dc.b	$F4,$F7,$42,$34,$FF,$96,$23,$FC,$77,$18,$8D,$BA,$C3,$CE,$E6,$15
	.dc.b	$E0,$ED,$FF,$22,$5F,$27,$FC,$31,$19,$9F,$1E,$15,$FE,$7D,$17,$0E
	.dc.b	$B5,$C2,$7C,$CF,$4A,$E1,$1E,$48,$C3,$A3,$FA,$41,$E4,$30,$7B,$63
	.dc.b	$8B,$9C,$3D,$98,$79,$E3,$8E,$1E,$38,$C4,$79,$8F,$F1,$79,$D4,$FF
	.dc.b	$18,$C6,$1C,$1F,$D6,$7E,$B5,$F6,$0F,$B0,$FF,$1B,$C6,$1B,$19,$9F
	.dc.b	$F1,$F4,$E5,$BF,$61,$FE,$8C,$6E,$9C,$A3,$11,$F0,$25,$18,$8F,$13
	.dc.b	$C7,$2F,$BF,$A7,$F5,$C9,$FA,$4D,$7E,$3D,$72,$BF,$AC,$D8,$F9,$73
	.dc.b	$BA,$E7,$B7,$F7,$77,$F1,$75,$B6,$E1,$E1,$AF,$EF,$87,$12,$D7,$17
	.dc.b	$CF,$87,$82,$78,$61,$FE,$94,$9E,$0D,$AE,$0F,$FA,$63,$0F,$04,$F2
	.dc.b	$8F,$3C,$E9,$FE,$D6,$D9,$DC,$73,$3F,$BD,$CF,$C3,$BF,$F4,$5B,$75
	.dc.b	$E9,$8F,$C0,$AD,$E1,$FE,$93,$AD,$47,$8F,$1F,$4A,$FB,$0F,$E1,$0F
	.dc.b	$F8,$21,$AF,$FA,$81,$B8,$FF,$A6,$9F,$B7,$1B,$DB,$FD,$41,$1A,$CB
	.dc.b	$FC,$4C,$D9,$FF,$5C,$BC,$1F,$F5,$5F,$D8,$7F,$9D,$DF,$F0,$BB,$5C
	.dc.b	$FC,$47,$FA,$AF,$EB,$9E,$17,$FA,$B6,$B5,$97,$FA,$B1,$B8,$FF,$8D
	.dc.b	$5B,$87,$FB,$22,$F8,$BD,$BC,$4E,$44,$7F,$C4,$F0,$FA,$CC,$47,$9A
	.dc.b	$6E,$1F,$EB,$AA,$D4,$F3,$21,$B3,$9D,$7A,$7E,$C2,$47,$67,$FD,$91
	.dc.b	$5F,$E8,$EF,$57,$33,$FE,$CE,$FB,$0F,$EE,$CF,$F8,$E9,$AF,$FB,$4B
	.dc.b	$05,$FF,$B3,$E7,$37,$FE,$CE,$6E,$7F,$E5,$76,$DF,$C3,$AF,$99,$D3
	.dc.b	$23,$E0,$BF,$ED,$9F,$B0,$FF,$0D,$C6,$0F,$F2,$5F,$97,$FE,$6E,$6F
	.dc.b	$C1,$F1,$7F,$DC,$B1,$FE,$14,$77,$E6,$7F,$B8,$E3,$11,$A9,$FF,$73
	.dc.b	$77,$7B,$23,$F9,$2B,$5F,$F3,$73,$5F,$F8,$02,$D7,$FC,$F1,$CD,$FC
	.dc.b	$83,$E0,$FF,$19,$1F,$CB,$9A,$FF,$9D,$DA,$87,$D8,$3E,$7F,$F9,$4B
	.dc.b	$63,$FD,$B2,$D7,$FD,$BC,$D7,$CF,$C7,$FC,$C3,$0F,$C1,$C7,$FB,$8D
	.dc.b	$AE,$8F,$F7,$1D,$FE,$B1,$F3,$74,$FA,$48,$FF,$BA,$3E,$BB,$FD,$C6
	.dc.b	$DF,$8C,$FE,$7B,$18,$8F,$F7,$83,$5F,$93,$D9,$E3,$23,$FE,$F2,$C3
	.dc.b	$FB,$23,$F0,$70,$FA,$6F,$C7,$DF,$E5,$11,$FF,$7C,$FD,$53,$E5,$3F
	.dc.b	$A6,$B6,$33,$18,$3E,$17,$64,$8F,$03,$5F,$EC,$27,$FA,$BF,$E8,$BE
	.dc.b	$C1,$1F,$8A,$FF,$5A,$FA,$A7,$CE,$D2,$4F,$21,$1A,$71,$F5,$0F,$A0
	.dc.b	$FF,$5B,$76,$BB,$7F,$D7,$13,$9A,$FF,$52,$B5,$4E,$DF,$C3,$7D,$4F
	.dc.b	$F6,$67,$77,$FD,$BA,$B8,$47,$F7,$C6,$BD,$A3,$47,$8A,$FF,$D9,$15
	.dc.b	$FE,$B1,$9F,$F5,$AB,$5F,$0C,$71,$73,$F2,$FC,$5F,$1F,$CB,$D5,$48
	.dc.b	$C2,$9A,$62,$24,$F4,$3F,$D8,$BF,$51,$F9,$DC,$BE,$04,$23,$18,$7F
	.dc.b	$B7,$DB,$1F,$A9,$7F,$C6,$11,$9E,$AF,$24,$6A,$79,$46,$39,$7D,$1A
	.dc.b	$3E,$A6,$92,$BC,$7E,$0D,$7F,$0F,$3B,$7E,$EC,$FD,$12,$3E,$AF,$9E
	.dc.b	$F0,$9C,$73,$F5,$E5,$FE,$FA,$9F,$F7,$1F,$D4,$FF,$97,$74,$FF,$7D
	.dc.b	$36,$3D,$9E,$37,$A8,$F9,$D1,$98,$36,$BA,$47,$D8,$3F,$86,$47,$FA
	.dc.b	$F9,$EF,$F0,$7C,$8E,$7F,$4B,$63,$24,$39,$1D,$9F,$BD,$F5,$3F,$CE
	.dc.b	$20,$D3,$14,$F2,$B7,$EA,$FF,$90,$A7,$F9,$31,$1C,$1E,$CE,$69,$7F
	.dc.b	$C8,$F5,$FC,$0D,$77,$C5,$D3,$F8,$F7,$F5,$D5,$E6,$6D,$7E,$21,$DC
	.dc.b	$73,$DB,$F9,$72,$8C,$BC,$7A,$7A,$1D,$6F,$B7,$BD,$27,$91,$FC,$DE
	.dc.b	$BF,$D9,$AE,$0E,$9F,$25,$B7,$C9,$37,$1F,$24,$FE,$9F,$E9,$DF,$2F
	.dc.b	$49,$7C,$F0,$CB,$D6,$6B,$FC,$61,$AF,$F9,$E9,$AF,$B9,$AF,$F4,$D7
	.dc.b	$8D,$FC,$5E,$AF,$94,$78,$27,$EF,$FD,$4F,$E9,$C9,$D7,$FB,$D7,$F5
	.dc.b	$B1,$FD,$5D,$CB,$F0,$5E,$1D,$7B,$F7,$B8,$74,$CA,$D4,$65,$E1,$19
	.dc.b	$7C,$31,$F1,$78,$24,$7C,$17,$C5,$F0,$78,$8F,$CA,$E9,$F8,$0E,$2F
	.dc.b	$81,$5A,$9E,$04,$36,$E5,$A6,$75,$FE,$1F,$5C,$E3,$DD,$7D,$E7,$35
	.dc.b	$6F,$F1,$79,$AB,$17,$F3,$EA,$A9,$D3,$DE,$BC,$ED,$3B,$94,$61,$E3
	.dc.b	$6D,$7A,$70,$F1,$C6,$E7,$57,$DA,$3D,$A4,$9D,$BF,$7B,$0F,$2D,$75
	.dc.b	$38,$B8,$BF,$4D,$F3,$BE,$24,$C3,$D3,$87,$33,$C5,$5C,$2F,$94,$7A
	.dc.b	$EF,$95,$70,$BE,$53,$F1,$CF,$C5,$57,$07,$D1,$3F,$3A,$31,$3C,$27
	.dc.b	$E6,$46,$27,$A1,$B8,$F1,$9E,$18,$D3,$DC,$F4,$1D,$8C,$4F,$5A,$F3
	.dc.b	$C6,$2B,$AB,$4F,$AE,$71,$3E,$A9,$DC,$FC,$81,$FB,$5B,$FF,$71,$E1
	.dc.b	$FC,$62,$30,$7F,$82,$5C,$3F,$30,$C9,$FB,$F8,$E2,$7F,$6F,$FA,$9F
	.dc.b	$E8,$47,$F4,$39,$FB,$51,$9A,$95,$F0,$8F,$E7,$75,$C2,$3E,$1C,$23
	.dc.b	$0E,$F0,$FE,$E3,$18,$8F,$75,$FE,$4E,$31,$3F,$4E,$31,$5E,$0A,$DD
	.dc.b	$DA,$E7,$3E,$0B,$E2,$FC,$CA,$F9,$27,$99,$E0,$66,$FA,$C7,$F2,$F9
	.dc.b	$F8,$17,$FA,$E7,$9C,$7D,$53,$9D,$F8,$1E,$17,$E1,$3A,$BB,$BE,$AF
	.dc.b	$84,$F8,$51,$FD,$22,$BA,$C7,$48,$F4,$46,$27,$E8,$CF,$62,$1F,$08
	.dc.b	$B5,$E3,$70,$77,$23,$F2,$EF,$CB,$1F,$56,$FD,$0F,$27,$E2,$46,$AF
	.dc.b	$E1,$CF,$F0,$97,$C9,$1F,$87,$E9,$93,$F2,$5F,$C5,$70,$62,$3F,$0B
	.dc.b	$17,$0F,$C0,$CF,$E3,$FE,$3B,$F2,$1C,$F2,$FE,$94,$65,$FA,$91,$97
	.dc.b	$F1,$F1,$FE,$45,$FA,$97,$D4,$D2,$4F,$F6,$9F,$A9,$7D,$A6,$23,$FC
	.dc.b	$EB,$5C,$1F,$F3,$73,$5C,$D7,$F9,$C5,$46,$72,$FF,$24,$DF,$F2,$93
	.dc.b	$F9,$54,$67,$6F,$F0,$1C,$7F,$7A,$AF,$F0,$FB,$FE,$61,$AE,$3D,$1F
	.dc.b	$E5,$D8,$D4,$74,$E9,$C4,$CA,$F5,$46,$FC,$49,$1E,$D7,$97,$4E,$23
	.dc.b	$BC,$5B,$F3,$CF,$F9,$AF,$EA,$9F,$55,$49,$7F,$E0,$FF,$29,$E5,$62
	.dc.b	$3F,$D3,$75,$C1,$FF,$4A,$35,$CD,$7C,$78,$C6,$1F,$E6,$EB,$FE,$FB
	.dc.b	$5F,$EA,$DE,$CE,$93,$9E,$BF,$F3,$F5,$1A,$63,$B2,$DF,$96,$2B,$FE
	.dc.b	$4A,$94,$F2,$F1,$64,$FA,$DF,$F4,$57,$D5,$7F,$97,$6B,$FC,$B7,$87
	.dc.b	$96,$77,$5F,$E5,$4A,$FF,$31,$7D,$A7,$FB,$06,$3F,$D7,$6D,$7F,$C7
	.dc.b	$E1,$C3,$FE,$42,$6B,$FE,$F1,$FA,$A7,$D5,$BE,$3B,$FD,$79,$18,$6F
	.dc.b	$80,$FA,$4C,$D6,$EF,$FD,$97,$F5,$6F,$5F,$FD,$33,$1F,$E5,$08,$C7
	.dc.b	$57,$FB,$1D,$B1,$C2,$76,$FF,$B0,$3E,$A7,$FD,$51,$5F,$E5,$0D,$D1
	.dc.b	$89,$FF,$3F,$46,$39,$FF,$D8,$CC,$9D,$37,$A3,$37,$FE,$6D,$9C,$B6
	.dc.b	$B3,$7E,$1A,$B6,$FA,$BF,$3B,$D9,$B7,$FE,$77,$F8,$1F,$90,$F9,$E3
	.dc.b	$37,$FE,$DA,$6D,$73,$CB,$AD,$F0,$AF,$5D,$70,$5F,$54,$FA,$BF,$1F
	.dc.b	$DA,$3D,$6F,$3D,$6A,$73,$18,$AE,$19,$FF,$BD,$5B,$5F,$F0,$01,$AF
	.dc.b	$E1,$9A,$FF,$A1,$9D,$E7,$FE,$D2,$6C,$FC,$9F,$3B,$B6,$1E,$2F,$9D
	.dc.b	$7C,$1F,$DA,$C5,$18,$BF,$C9,$C0,$CC,$EE,$75,$1A,$87,$BB,$CD,$B5
	.dc.b	$CA,$7B,$CE,$4F,$07,$2B,$E9,$E9,$2B,$84,$65,$E5,$F2,$A7,$D0,$D7
	.dc.b	$26,$0F,$67,$0B,$E3,$69,$23,$C9,$19,$7C,$DF,$50,$F2,$3E,$23,$63
	.dc.b	$31,$83,$E4,$7D,$3F,$E9,$34,$FD,$CB,$5F,$F5,$CB,$5F,$94,$D5,$F4
	.dc.b	$EF,$0B,$E9,$36,$FE,$43,$67,$E9,$B5,$FE,$0F,$F4,$D5,$FF,$63,$B5
	.dc.b	$FB,$0D,$4D,$FA,$5F,$B3,$1F,$C4,$1A,$FF,$B4,$1A,$FE,$05,$AB,$E9
	.dc.b	$1E,$97,$E2,$5B,$87,$DF,$6C,$FE,$29,$AF,$F1,$E8,$7D,$27,$E4,$5A
	.dc.b	$7E,$1C,$B1,$F9,$26,$BF,$C9,$BE,$91,$E6,$7E,$65,$B3,$F8,$F6,$CF
	.dc.b	$E6,$9A,$FF,$2E,$F8,$07,$C4,$7E,$85,$B5,$F9,$76,$CF,$E8,$9A,$FF
	.dc.b	$36,$FA,$37,$C1,$7E,$A5,$B5,$F9,$F6,$CF,$EA,$9A,$FF,$3E,$FA,$17
	.dc.b	$D1,$7F,$04,$FA,$3F,$F5,$A5,$FF,$BD,$67,$2F,$FB,$DE,$33,$4F,$0B
	.dc.b	$F4,$4F,$ED,$6B,$6E,$6B,$F7,$0A,$33,$3C,$21,$F4,$6F,$63,$F8,$AD
	.dc.b	$6A,$F8,$F9,$EF,$99,$FE,$33,$7B,$9F,$E0,$98,$F3,$AD,$46,$0F,$0C
	.dc.b	$EF,$1F,$21,$FC,$2F,$E8,$9F,$47,$FC,$E3,$C0,$FE,$ED,$7C,$6B,$53
	.dc.b	$9F,$63,$F9,$4D,$F0,$AF,$E9,$D7,$C2,$FA,$89,$D3,$FD,$42,$35,$97
	.dc.b	$C1,$9C,$BF,$D4,$A3,$35,$F2,$27,$2F,$BA,$33,$5E,$9E,$1F,$E9,$FF
	.dc.b	$46,$F9,$2F,$2C,$61,$B7,$FD,$8E,$74,$FB,$6B,$CD,$5B,$8C,$61,$FC
	.dc.b	$4E,$7F,$B3,$F8,$4F,$94,$F8,$3E,$27,$F5,$26,$DF,$08,$F8,$B5,$83
	.dc.b	$E1,$E9,$C6,$7C,$0B,$E8,$FF,$A9,$B9,$9D,$C6,$2E,$57,$9E,$7F,$A1
	.dc.b	$7D,$1F,$F4,$3F,$A0,$FD,$8C,$7E,$0A,$BF,$AE,$BA,$11,$A7,$F3,$11
	.dc.b	$F9,$88,$C4,$6D,$D6,$1E,$37,$30,$AF,$07,$6F,$E7,$2B,$4F,$F5,$68
	.dc.b	$CC,$F8,$F0,$AF,$E8,$A2,$D7,$08,$FF,$17,$C6,$30,$F2,$5F,$39,$F4
	.dc.b	$BD,$E3,$31,$C0,$C9,$DC,$A7,$72,$D7,$77,$18,$7B,$A7,$9C,$6C,$FE
	.dc.b	$CD,$7F,$E8,$57,$E0,$1F,$E8,$77,$FB,$57,$D0,$BE,$89,$F4,$78,$6D
	.dc.b	$7F,$78,$9D,$4E,$4E,$3F,$3B,$FC,$26,$74,$D7,$FC,$A3,$28,$C4,$7F
	.dc.b	$0A,$8C,$65,$C8,$F9,$B9,$FF,$6D,$9B,$B7,$F1,$0D,$3F,$01,$7C,$27
	.dc.b	$E2,$B6,$3F,$09,$5C,$67,$E4,$B6,$38,$E1,$E7,$6E,$1D,$AA,$1D,$27
	.dc.b	$55,$E7,$6C,$7F,$86,$63,$A1,$FC,$90,$AF,$F8,$76,$FC,$91,$C6,$BA
	.dc.b	$DF,$9B,$0F,$64,$E6,$FC,$F1,$F9,$3D,$7A,$36,$3C,$0D,$58,$76,$72
	.dc.b	$7F,$03,$AF,$85,$87,$F3,$56,$DF,$7C,$3F,$55,$F4,$7F,$C0,$EF,$F8
	.dc.b	$53,$5F,$E4,$2D,$BF,$E7,$CF,$DB,$8D,$ED,$E3,$8D,$5F,$F4,$66,$CF
	.dc.b	$F8,$FD,$E0,$FF,$2C,$FA,$3F,$E8,$4F,$F1,$96,$B9,$EB,$F7,$7D,$1F
	.dc.b	$FA,$FA,$35,$1F,$06,$77,$8F,$F5,$16,$E1,$FD,$5A,$F8,$5F,$7F,$0B
	.dc.b	$81,$1F,$E5,$10,$F9,$FC,$46,$CF,$24,$F8,$A3,$38,$77,$8B,$7D,$E7
	.dc.b	$CF,$EC,$4B,$EA,$FD,$9B,$7B,$9F,$EB,$CB,$F3,$4F,$DB,$FA,$3F,$F0
	.dc.b	$0D,$FF,$3B,$6B,$FD,$4B,$2F,$F4,$03,$9E,$AE,$04,$7B,$9F,$D0,$FE
	.dc.b	$8F,$FC,$46,$FB,$BA,$BE,$59,$F8,$98,$CC,$75,$8F,$C6,$7D,$1F,$F9
	.dc.b	$EA,$3F,$A6,$35,$D1,$8F,$63,$7E,$54,$8C,$1F,$D4,$FC,$0F,$4E,$FF
	.dc.b	$E0,$56,$ED,$FE,$0A,$6C,$FF,$84,$7F,$6A,$7C,$D7,$E7,$7D,$AF,$C3
	.dc.b	$E1,$FE,$0D,$6E,$1F,$E1,$26,$DF,$F8,$67,$EC,$1E,$F7,$E9,$7E,$95
	.dc.b	$F3,$94,$95,$FD,$9F,$B1,$F6,$BF,$AB,$DB,$F0,$53,$E9,$E9,$C1,$F6
	.dc.b	$77,$7C,$B7,$97,$09,$5F,$DC,$75,$FC,$FE,$ED,$BE,$43,$FE,$25,$FA
	.dc.b	$6F,$F7,$25,$7F,$77,$6B,$FE,$0E,$6C,$7F,$8A,$FE,$D1,$1E,$5F,$0A
	.dc.b	$57,$CD,$7C,$9E,$87,$C7,$E5,$93,$E3,$7F,$21,$8F,$DE,$C7,$FC,$6D
	.dc.b	$F4,$EF,$92,$FE,$14,$DB,$FF,$05,$35,$FF,$10,$FD,$97,$EA,$AB,$ED
	.dc.b	$7C,$BF,$C7,$24,$FF,$BD,$7E,$D3,$D3,$D3,$F5,$7A,$BE,$D1,$27,$FD
	.dc.b	$F5,$9B,$EA,$BF,$8D,$7D,$53,$E8,$69,$2B,$FC,$39,$F6,$6F,$A9,$FE
	.dc.b	$41,$F5,$4F,$A2,$A4,$AF,$F1,$17,$D5,$FF,$5B,$7F,$C4,$93,$B6,$CE
	.dc.b	$DF,$F1,$EC,$ED,$FF,$1F,$CE,$DF,$F6,$BB,$5F,$F3,$08,$70,$FF,$99
	.dc.b	$1A,$FF,$B7,$63,$F7,$EF,$93,$E5,$BF,$9C,$41,$BE,$64,$9D,$10,$F2
	.dc.b	$BF,$CF,$FE,$AF,$F3,$71,$FE,$3F,$6B,$FE,$56,$6F,$1F,$F9,$CD,$FF
	.dc.b	$10,$F0,$FF,$9E,$1C,$B7,$C3,$E8,$75,$EA,$F9,$BA,$CC,$3E,$4B,$E1
	.dc.b	$F3,$71,$1F,$1A,$2E,$3D,$E7,$33,$F1,$B5,$FB,$0F,$FB,$DD,$AF,$CC
	.dc.b	$AF,$F7,$D2,$CB,$F9,$22,$1F,$57,$FE,$8D,$C3,$FD,$1C,$DC,$FD,$9F
	.dc.b	$1B,$F1,$4F,$EC,$91,$9A,$E4,$D7,$B4,$6B,$E9,$E4,$7F,$9A,$65,$1C
	.dc.b	$BE,$9F,$33,$FE,$82,$9C,$BB,$9D,$3F,$E7,$1F,$AB,$FE,$36,$9A,$62
	.dc.b	$5E,$9F,$F7,$98,$CF,$A3,$4F,$4C,$5A,$E9,$15,$F1,$47,$F8,$07,$0F
	.dc.b	$E4,$F1,$FE,$81,$7F,$54,$DE,$7D,$75,$C8,$C1,$0F,$F4,$EC,$F2,$6F
	.dc.b	$C3,$75,$23,$D7,$E4,$AB,$A3,$C1,$1E,$38,$FF,$0A,$7C,$FF,$F3,$29
	.dc.b	$FF,$4A,$35,$FF,$55,$B7,$83,$FD,$91,$77,$83,$FD,$9B,$36,$BF,$04
	.dc.b	$D4,$D9,$C3,$3D,$0F,$09,$F1,$4F,$F1,$37,$D5,$7E,$82,$BF,$18,$D4
	.dc.b	$D3,$F4,$EE,$BA,$38,$91,$FF,$57,$46,$9E,$3F,$0D,$DB,$E9,$24,$6A
	.dc.b	$3C,$26,$1F,$F6,$5C,$A7,$31,$F1,$B7,$F9,$24,$7B,$FD,$5D,$7E,$4C
	.dc.b	$5D,$7F,$DE,$16,$AC,$6B,$F2,$AA,$DF,$C0,$9A,$FC,$A8,$B5,$C6,$3C
	.dc.b	$44,$3B,$9E,$23,$17,$FE,$FF,$D3,$F1,$F1,$A3,$DB,$7D,$A3,$33,$BB
	.dc.b	$FA,$73,$97,$C2,$DA,$EB,$87,$F0,$17,$EA,$C6,$5E,$D5,$F0,$23,$0F
	.dc.b	$58,$F3,$1A,$9F,$84,$F4,$3E,$11,$4F,$84,$5C,$3E,$13,$C4,$FC,$8D
	.dc.b	$F9,$9E,$18,$7B,$2B,$AC,$7C,$03,$33,$B7,$91,$D8,$B5,$DA,$FD,$11
	.dc.b	$AB,$E8,$F9,$8C,$57,$39,$D9,$2F,$C2,$FC,$39,$F0,$9E,$BC,$3E,$2C
	.dc.b	$E8,$F9,$98,$78,$5C,$4F,$B8,$E0,$72,$3C,$F3,$C9,$D1,$EF,$3C,$A7
	.dc.b	$02,$9C,$0E,$55,$E7,$7D,$8D,$7E,$23,$6B,$C9,$1D,$A3,$58,$7C,$48
	.dc.b	$FC,$19,$D2,$FD,$2E,$0F,$0C,$F3,$BF,$44,$76,$9F,$83,$18,$AE,$97
	.dc.b	$E0,$AD,$1E,$1A,$EB,$19,$9F,$11,$30,$FC,$DD,$78,$CE,$07,$BA,$BC
	.dc.b	$B3,$B8,$FC,$AE,$1E,$CA,$FB,$95,$A6,$CF,$23,$C6,$4B,$F1,$47,$B1
	.dc.b	$B1,$E1,$3E,$15,$75,$7E,$A3,$5F,$A8,$7C,$6A,$F6,$E1,$F0,$EF,$D7
	.dc.b	$39,$3C,$57,$D9,$C1,$DA,$BD,$0F,$67,$F7,$1F,$57,$FD,$9A,$18,$27
	.dc.b	$3C,$DF,$ED,$9A,$D4,$E6,$BF,$C0,$10,$DF,$55,$27,$FC,$09,$5B,$55
	.dc.b	$CA,$FC,$37,$D7,$E4,$FB,$78,$32,$F8,$D7,$9A,$75,$18,$3C,$33,$B7
	.dc.b	$D3,$3B,$74,$3E,$A9,$F5,$7F,$EF,$36,$C6,$9A,$F2,$3F,$C3,$9F,$57
	.dc.b	$FE,$8A,$75,$3C,$48,$F6,$8D,$3C,$4F,$04,$69,$C4,$F6,$CF,$E2,$57
	.dc.b	$06,$F3,$7F,$2B,$73,$CD,$71,$FE,$5D,$48,$C6,$1E,$6B,$AA,$51,$87
	.dc.b	$84,$79,$E7,$13,$9F,$0A,$4F,$AD,$E1,$E1,$E2,$3D,$97,$FE,$66,$6C
	.dc.b	$E9,$F9,$FF,$54,$FA,$BF,$1B,$63,$31,$E6,$F3,$FF,$D2,$D9,$FF,$49
	.dc.b	$AD,$57,$C9,$6E,$9F,$89,$6C,$FF,$4F,$AE,$35,$FE,$51,$9E,$2F,$FA
	.dc.b	$01,$AF,$5B,$FF,$2C,$C6,$23,$FC,$B1,$2E,$9F,$21,$23,$F1,$31,$FE
	.dc.b	$01,$6C,$65,$FB,$D1,$C5,$DF,$4F,$B7,$C1,$91,$C9,$FA,$BF,$57,$FE
	.dc.b	$5D,$AF,$F3,$4C,$FF,$AF,$67,$6F,$E2,$67,$FC,$C5,$F5,$6F,$07,$FC
	.dc.b	$E2,$D9,$FF,$4E,$9F,$E7,$17,$0E,$1A,$FF,$A8,$9F,$F6,$33,$6F,$99
	.dc.b	$1E,$63,$EA,$DF,$56,$F5,$73,$18,$6F,$80,$FA,$2C,$D6,$EF,$FD,$63
	.dc.b	$C4,$FA,$A7,$27,$F0,$98,$91,$8E,$4F,$DE,$46,$F9,$B5,$1F,$C2,$A7
	.dc.b	$57,$FB,$5A,$E9,$7A,$8F,$E3,$33,$A9,$EB,$79,$77,$7D,$BE,$A5,$F5
	.dc.b	$59,$6C,$78,$BE,$1B,$E0,$3B,$9D,$11,$D6,$D7,$2F,$E0,$1F,$52,$FA
	.dc.b	$9F,$07,$D5,$3D,$2F,$35,$2B,$58,$7C,$0A,$D3,$FC,$62,$33,$3E,$F9
	.dc.b	$CB,$DA,$FA,$DF,$4A,$DE,$9F,$CB,$DB,$87,$C9,$F2,$F5,$7E,$2F,$AA
	.dc.b	$7C,$4F,$C5,$FB,$37,$D0,$E6,$30,$FA,$17,$C2,$6F,$1C,$35,$FE,$5F
	.dc.b	$CE,$FA,$BF,$A1,$9B,$E2,$7E,$0C,$E6,$BE,$FC,$E5,$FE,$6F,$F3,$DF
	.dc.b	$D4,$5F,$C1,$FD,$8B,$C2,$F9,$33,$9E,$3F,$C1,$4E,$6B,$F9,$FF,$3B
	.dc.b	$E6,$3F,$01,$3D,$6B,$F9,$A4,$E6,$BA,$E7,$F5,$3E,$B3,$F9,$64,$7E
	.dc.b	$4B,$57,$D3,$E7,$CC,$FC,$54,$62,$7F,$A6,$FD,$67,$E2,$32,$FC,$A7
	.dc.b	$D8,$BE,$CB,$EF,$CE,$7A,$BF,$41,$39,$BF,$EA,$BF,$5D,$E9,$C3,$F3
	.dc.b	$9F,$65,$D5,$FE,$95,$19,$BF,$D4,$4E,$5F,$ED,$9F,$5F,$9B,$FE,$DF
	.dc.b	$2F,$93,$FC,$7C,$E7,$0F,$EF,$B3,$9A,$FE,$CB,$F5,$FF,$82,$7F,$63
	.dc.b	$F6,$2F,$8C,$FC,$BC,$E6,$FF,$61,$39,$AF,$ED,$BF,$60,$F6,$FF,$5D
	.dc.b	$3D,$6B,$FB,$14,$E6,$BA,$E7,$FA,$27,$D4,$7F,$81,$BC,$1F,$F0,$53
	.dc.b	$FD,$E0,$CD,$7F,$85,$23,$2F,$F6,$DF,$B0,$7C,$5F,$EB,$1F,$E7,$55
	.dc.b	$FE,$00,$72,$DF,$5B,$F6,$3F,$0C,$F9,$4F,$D9,$EE,$F9,$CC,$C6,$1F
	.dc.b	$DB,$FD,$87,$F3,$1C,$3F,$C1,$6B,$BB,$3E,$C6,$9F,$DE,$FD,$87,$FA
	.dc.b	$F6,$BF,$C2,$2D,$DB,$FC,$09,$F6,$1F,$AE,$9C,$FC,$4F,$F0,$48,$C4
	.dc.b	$EB,$CB,$FE,$0F,$F6,$1F,$DE,$AF,$FC,$37,$5A,$EB,$FF,$16,$4E,$6F
	.dc.b	$FC,$20,$D8,$FF,$18,$7A,$4F,$A5,$E7,$F0,$7F,$C4,$BE,$B9,$F6,$1F
	.dc.b	$C9,$7C,$CF,$93,$97,$F8,$66,$73,$A7,$F3,$49,$CE,$5F,$E3,$B9,$CD
	.dc.b	$7F,$33,$9C,$D7,$FB,$0E,$73,$5F,$CF,$E7,$34,$13,$AA,$FF,$23,$CE
	.dc.b	$A7,$37,$E5,$E7,$F4,$71,$7F,$3B,$FB,$07,$D1,$63,$1F,$E9,$31,$87
	.dc.b	$FC,$83,$97,$D9,$C7,$FA,$44,$E5,$F7,$5F,$3B,$F8,$9E,$77,$F4,$7E
	.dc.b	$57,$D0,$E3,$2F,$E8,$8D,$C3,$E3,$57,$CB,$71,$B7,$F9,$7C,$E3,$5F
	.dc.b	$E5,$25,$F6,$1F,$D0,$9C,$CF,$F9,$1D,$CC,$EE,$F4,$7F,$48,$9F,$91
	.dc.b	$F6,$1F,$27,$EB,$3F,$43,$1F,$E6,$E9,$D2,$9D,$78,$1F,$E4,$B9,$FF
	.dc.b	$3C,$57,$0A,$FF,$7C,$56,$9E,$57,$5C,$89,$E1,$87,$A1,$E5,$1E,$43
	.dc.b	$58,$77,$AE,$51,$99,$E3,$87,$8A,$BF,$C3,$38,$77,$C3,$A4,$62,$F9
	.dc.b	$9E,$42,$9E,$4B,$E7,$87,$68,$F8,$06,$AB,$9B,$F8,$F3,$C5,$5E,$57
	.dc.b	$99,$E9,$3F,$C4,$37,$FE,$22,$70,$78,$CE,$43,$EB,$5F,$60,$FB,$0F
	.dc.b	$F1,$4C,$61,$B1,$99,$FF,$1B,$CE,$5B,$F6,$1F,$EB,$06,$E3,$CA,$31
	.dc.b	$1E,$09,$46,$23,$D9,$18,$9F,$B1,$8F,$F6,$19,$BB,$7C,$1D,$3F,$01
	.dc.b	$7B,$9F,$96,$D8,$FC,$25,$71,$9F,$9E,$D8,$E3,$87,$9D,$B8,$76,$A8
	.dc.b	$74,$9D,$57,$99,$B1,$FC,$06,$3A,$1F,$54,$AF,$F0,$5B,$F2,$47,$1A
	.dc.b	$EB,$7E,$6C,$3D,$93,$9B,$F3,$C7,$FA,$8A,$3D,$55,$FE,$D1,$6E,$9B
	.dc.b	$BE,$57,$FE,$C6,$9C,$BE,$5B,$FF,$55,$E1,$E2,$8F,$E3,$1F,$61,$FC
	.dc.b	$11,$FF,$66,$B5,$FF,$60,$37,$1F,$F7,$03,$F6,$23,$7B,$7F,$AD,$A3
	.dc.b	$59,$7E,$55,$B3,$DB,$3F,$F5,$DF,$D8,$7F,$39,$8F,$F6,$DB,$5F,$F6
	.dc.b	$73,$70,$FF,$77,$7D,$7F,$FA,$92,$53,$9C,$3F,$44,$D8,$FE,$64,$D5
	.dc.b	$3C,$61,$E3,$78,$89,$3F,$EE,$C9,$FF,$1E,$3E,$48,$CB,$E4,$7F,$DA
	.dc.b	$71,$A8,$C1,$B2,$1F,$EE,$77,$57,$EA,$F8,$2E,$24,$9F,$F4,$2F,$C4
	.dc.b	$7F,$AA,$2B,$FC,$D9,$1F,$EF,$EF,$B0,$FF,$50,$DB,$2F,$E0,$11,$87
	.dc.b	$F8,$15,$70,$AF,$F7,$B4,$E5,$FE,$A8,$D4,$D8,$E1,$5D,$67,$FA,$D7
	.dc.b	$D8,$7F,$91,$9F,$89,$CF,$F0,$0F,$F7,$9C,$66,$3D,$D1,$FE,$F6,$F2
	.dc.b	$3F,$CE,$31,$FD,$8E,$BA,$3F,$E1,$3F,$82,$E4,$E7,$2F,$E3,$55,$C2
	.dc.b	$15,$C2,$3F,$C3,$91,$B3,$FB,$73,$87,$73,$CA,$BF,$92,$57,$FA,$CF
	.dc.b	$EB,$BF,$8D,$D7,$FA,$C9,$78,$5A,$BF,$34,$62,$75,$D1,$97,$C7,$B3
	.dc.b	$EE,$31,$8F,$FB,$03,$EA,$EC,$FF,$BE,$23,$FD,$87,$3A,$FB,$AF,$F1
	.dc.b	$95,$FF,$B0,$97,$D8,$E4,$D6,$DD,$33,$FE,$D7,$5E,$7E,$FC,$19,$E7
	.dc.b	$EF,$F0,$D2,$7F,$12,$F8,$BB,$3C,$1F,$23,$27,$C0,$F8,$FE,$47,$B7
	.dc.b	$D8,$77,$EE,$95,$F1,$63,$DB,$F5,$AF,$C0,$BD,$B3,$0F,$F6,$DC,$65
	.dc.b	$F6,$7D,$DF,$BF,$EA,$E4,$60,$FA,$1F,$4D,$ED,$23,$F4,$FD,$8E,$FC
	.dc.b	$F2,$3A,$BE,$2F,$96,$F9,$FF,$77,$FE,$A7,$9F,$F7,$5B,$5F,$F7,$43
	.dc.b	$7A,$7F,$DE,$5D,$1F,$E1,$7E,$CF,$EC,$DF,$5C,$FA,$AF,$B9,$82,$30
	.dc.b	$FE,$34,$C4,$6A,$1A,$58,$93,$B9,$E2,$FE,$48,$FF,$21,$3C,$1D,$47
	.dc.b	$07,$47,$63,$94,$F9,$3E,$B9,$F4,$DF,$DD,$DB,$3F,$DD,$8F,$F2,$73
	.dc.b	$87,$0D,$7F,$BD,$B9,$CB,$F6,$38,$3E,$3B,$D8,$77,$BF,$76,$DE,$FD
	.dc.b	$A6,$3E,$D8,$CB,$EF,$8C,$BE,$AB,$4F,$CC,$FA,$E7,$D6,$52,$57,$CD
	.dc.b	$8F,$F1,$7E,$DF,$E1,$97,$71,$99,$DB,$F1,$A3,$6F,$F9,$B5,$DB,$98
	.dc.b	$DE,$7F,$C3,$A7,$72,$D1,$C5,$F2,$C9,$1F,$2C,$D3,$87,$C1,$F4,$7E
	.dc.b	$B3,$EA,$47,$B7,$E5,$BF,$13,$F5,$DF,$8D,$AF,$F4,$03,$56,$1F,$93
	.dc.b	$CF,$DD,$5C,$19,$3F,$0E,$39,$FD,$64,$8F,$C9,$5A,$3B,$FD,$5E,$67
	.dc.b	$F1,$33,$97,$73,$A6,$90,$DB,$97,$F2,$3F,$5C,$F6,$BF,$C8,$8D,$8F
	.dc.b	$88,$D7,$F2,$0D,$7F,$C8,$EF,$F4,$BA,$FC,$76,$3F,$E5,$36,$CF,$FA
	.dc.b	$8E,$7F,$A8,$61,$FA,$6D,$BF,$A1,$BF,$A5,$9F,$D1,$B6,$33,$CD,$FB
	.dc.b	$68,$E1,$87,$F9,$7D,$D3,$79,$B5,$3B,$F2,$E6,$1F,$BB,$BF,$E7,$8F
	.dc.b	$6E,$7C,$BE,$09,$B8,$F9,$26,$E3,$DE,$6D,$7F,$AF,$6F,$F5,$4F,$F0
	.dc.b	$7F,$05,$F3,$7E,$7E,$89,$39,$8F,$E1,$F3,$A8,$F2,$4E,$9F,$E0,$71
	.dc.b	$97,$F8,$2F,$D7,$7E,$7B,$3F,$ED,$2D,$CB,$FD,$00,$D9,$FF,$68,$3F
	.dc.b	$A0,$30,$9A,$4F,$63,$F9,$0A,$F7,$EF,$3C,$AF,$E4,$53,$96,$CE,$5F
	.dc.b	$2C,$F2,$89,$1C,$23,$58,$7A,$27,$33,$C0,$D1,$D0,$86,$49,$1C,$5C
	.dc.b	$BF,$0A,$7A,$E1,$F0,$AB,$AC,$E6,$BA,$9F,$2A,$33,$3F,$22,$B8,$CF
	.dc.b	$C8,$9C,$BF,$2A,$32,$FC,$98,$CB,$F2,$E3,$37,$F1,$63,$B1,$E0,$BF
	.dc.b	$13,$82,$38,$25,$FA,$63,$DA,$F9,$5E,$4E,$5E,$2E,$5F,$8A,$E6,$3E
	.dc.b	$34,$60,$F5,$1D,$0F,$71,$D0,$C9,$D0,$C9,$C4,$F7,$1C,$4F,$84,$71
	.dc.b	$32,$71,$32,$71,$95,$7C,$87,$E3,$4F,$C0,$9F,$0B,$E8,$25,$7B,$5D
	.dc.b	$47,$47,$2F,$08,$D1,$F0,$48,$E4,$87,$17,$2F,$A2,$3A,$CF,$92,$72
	.dc.b	$FA,$27,$DC,$E5,$F8,$B3,$C2,$BD,$97,$F8,$26,$CF,$06,$BF,$0D,$B3
	.dc.b	$C6,$3E,$65,$75,$7C,$97,$EC,$7B,$BE,$4A,$FC,$7D,$FC,$B6,$CF,$26
	.dc.b	$BF,$4E,$BD,$4F,$D3,$9E,$0F,$B1,$AE,$4E,$46,$4A,$64,$E9,$3D,$AF
	.dc.b	$E9,$C7,$A0,$F9,$F5,$EB,$24,$EE,$FD,$53,$D4,$E6,$5A,$E7,$5E,$87
	.dc.b	$EA,$DF,$99,$B3,$E4,$6B,$C8,$F3,$3C,$AF,$C1,$1F,$28,$F2,$BF,$BA
	.dc.b	$C7,$FA,$4C,$07,$D7,$7F,$79,$3F,$BD,$46,$A0,$D8,$D6,$3F,$CF,$38
	.dc.b	$7F,$C9,$71,$FD,$CE,$BF,$C7,$AD,$C3,$FD,$42,$DA,$F0,$4F,$83,$6E
	.dc.b	$3C,$B2,$3F,$CA,$55,$B7,$CC,$A7,$BF,$C8,$F8,$3B,$A5,$70,$8C,$BC
	.dc.b	$A3,$2F,$92,$32,$F9,$A3,$38,$F9,$A7,$FC,$CB,$F5,$AF,$AE,$FF,$13
	.dc.b	$5A,$30,$FF,$8A,$5A,$FF,$95,$5A,$FB,$8E,$11,$FE,$70,$FA,$EF,$98
	.dc.b	$7C,$D9,$E3,$7F,$EC,$68,$FE,$EE,$DE,$6E,$07,$1E,$69,$87,$F9,$DE
	.dc.b	$BF,$B9,$BF,$E6,$C6,$D7,$39,$F0,$78,$12,$B8,$3E,$3E,$6C,$47,$18
	.dc.b	$B5,$E3,$9B,$C1,$F6,$30,$FF,$2F,$DF,$23,$C5,$3F,$E8,$CF,$AD,$7D
	.dc.b	$77,$C8,$6C,$FF,$93,$DA,$FF,$9E,$DA,$FC,$77,$38,$7F,$BA,$3E,$B3
	.dc.b	$FC,$B6,$7F,$97,$23,$38,$FE,$2A,$3F,$85,$13,$9F,$FD,$3A,$FE,$AE
	.dc.b	$35,$18,$25,$FF,$A7,$E7,$4F,$E8,$DF,$F2,$C3,$7B,$F9,$9F,$EA,$8F
	.dc.b	$AE,$FF,$2F,$4E,$A4,$D3,$11,$A9,$E3,$3D,$DE,$57,$FE,$B2,$FA,$D7
	.dc.b	$9E,$F8,$DF,$F3,$A4,$61,$BD,$6F,$4B,$35,$BC,$7F,$D4,$9F,$6C,$FA
	.dc.b	$C6,$5F,$E7,$D8,$C3,$63,$33,$C1,$AF,$ED,$38,$1F,$5D,$FE,$9C,$AF
	.dc.b	$E0,$D1,$C0,$CF,$2D,$8C,$CB,$93,$16,$BD,$D7,$23,$77,$A8,$CC,$FF
	.dc.b	$B2,$5F,$0D,$66,$FC,$53,$9F,$AE,$F5,$B7,$1E,$5F,$01,$F2,$72,$ED
	.dc.b	$3E,$1A,$DE,$3E,$96,$CF,$B1,$B3,$E7,$AE,$4F,$59,$E7,$4A,$DB,$6F
	.dc.b	$C3,$BF,$AE,$73,$F5,$6F,$86,$F7,$37,$0E,$6B,$BB,$33,$AF,$52,$56
	.dc.b	$BC,$EB,$A7,$C5,$F2,$7F,$1F,$3E,$46,$C6,$7C,$DF,$99,$18,$DB,$EC
	.dc.b	$7E,$29,$F8,$4B,$F5,$AC,$7E,$93,$9C,$3E,$93,$99,$F8,$3F,$57,$CD
	.dc.b	$FC,$1B,$FE,$A9,$73,$D9,$F6,$E3,$15,$F6,$FE,$A1,$F0,$BF,$7D,$FF
	.dc.b	$43,$DF,$CC,$72,$DC,$FF,$09,$F4,$CF,$AB,$A6,$32,$B5,$F5,$3E,$97
	.dc.b	$8E,$5F,$89,$6E,$DF,$97,$7B,$C6,$73,$FC,$84,$3D,$2A,$63,$CF,$B8
	.dc.b	$7E,$03,$E9,$1F,$55,$F9,$77,$AC,$7E,$91,$CB,$BC,$35,$7B,$BB,$3C
	.dc.b	$DF,$D0,$7D,$1B,$EF,$2F,$9F,$86,$D7,$E6,$B4,$7D,$23,$D7,$FD,$6C
	.dc.b	$63,$CB,$B3,$9A,$E0,$FF,$A7,$5E,$5D,$6F,$80,$E3,$F4,$8F,$8A,$E7
	.dc.b	$39,$AF,$04,$E5,$F1,$7D,$27,$DC,$E5,$FE,$0E,$BC,$5F,$1B,$99,$FE
	.dc.b	$FA,$E5,$F5,$7D,$23,$E9,$3D,$2D,$8F,$77,$D1,$BE,$93,$35,$EE,$8C
	.dc.b	$4E,$A7,$E0,$FD,$23,$D8,$D4,$61,$93,$F2,$7C,$A7,$D2,$63,$93,$F8
	.dc.b	$A4,$FF,$84,$BE,$89,$F4,$7F,$C4,$FE,$93,$E4,$56,$8F,$F1,$34,$EA
	.dc.b	$BF,$81,$CE,$B8,$7F,$94,$CE,$6B,$F8,$D4,$E6,$BB,$5F,$5E,$7F,$B3
	.dc.b	$7B,$C3,$F9,$65,$EF,$0E,$F7,$BC,$3F,$05,$7F,$36,$7F,$96,$7D,$23
	.dc.b	$E9,$39,$F9,$9F,$CA,$E3,$17,$C2,$31,$3A,$8F,$E3,$73,$AA,$F7,$0B
	.dc.b	$7C,$9F,$F2,$CF,$93,$F3,$B6,$FC,$6E,$1F,$2D,$53,$9E,$8A,$FC,$BA
	.dc.b	$F9,$5B,$18,$6D,$7E,$7E,$78,$3F,$A1,$AF,$F2,$B7,$D2,$7F,$95,$FE
	.dc.b	$8F,$F4,$51,$FD,$53,$0F,$D1,$BA,$11,$A7,$FA,$94,$7F,$60,$8C,$46
	.dc.b	$DD,$5F,$82,$38,$F8,$1F,$E3,$95,$7E,$18,$FE,$CD,$7C,$DD,$46,$67
	.dc.b	$C9,$85,$7F,$CB,$F1,$6F,$B6,$1E,$29,$EA,$78,$0E,$25,$38,$96,$38
	.dc.b	$DF,$82,$31,$87,$8D,$FF,$28,$5F,$78,$C4,$72,$BF,$7C,$62,$3E,$41
	.dc.b	$EF,$3C,$65,$9F,$F3,$0B,$83,$FC,$F3,$F4,$6F,$9D,$7D,$27,$18,$C3
	.dc.b	$63,$33,$F2,$E7,$2D,$F9,$AF,$F0,$33,$73,$E5,$18,$8F,$F3,$54,$A3
	.dc.b	$11,$E2,$78,$E5,$F7,$F3,$FC,$CC,$FD,$16,$BF,$2E,$B9,$5F,$D5,$6C
	.dc.b	$7C,$F9,$DD,$73,$DB,$FD,$2E,$FE,$2E,$B6,$DC,$3C,$35,$FE,$9C,$38
	.dc.b	$96,$B8,$BE,$7C,$3C,$13,$C3,$0F,$F5,$94,$F0,$6D,$70,$7F,$D6,$D8
	.dc.b	$78,$27,$94,$79,$E7,$4F,$F4,$56,$CE,$E3,$99,$FD,$2E,$7E,$1D,$FF
	.dc.b	$90,$1B,$AF,$4C,$7E,$05,$6F,$0F,$F2,$24,$FA,$E3,$FA,$A6,$1D,$57
	.dc.b	$D2,$7E,$FD,$FE,$B0,$D7,$F8,$6B,$71,$FF,$27,$BF,$6E,$37,$B7,$67
	.dc.b	$FD,$A9,$1B,$BF,$F6,$3B,$63,$39,$ED,$7D,$27,$F3,$E7,$FB,$43,$5F
	.dc.b	$E3,$AD,$C3,$FC,$BF,$F4,$4F,$A1,$E5,$39,$C7,$FD,$AA,$DB,$FF,$73
	.dc.b	$35,$57,$17,$BF,$8B,$C8,$8F,$F7,$98,$FF,$0D,$3A,$21,$E4,$8F,$54
	.dc.b	$6A,$30,$6B,$0F,$F3,$AD,$7F,$92,$E3,$A7,$A7,$C3,$E9,$24,$77,$7E
	.dc.b	$55,$3E,$0B,$FD,$2D,$82,$7E,$47,$D2,$3A,$E9,$CB,$0F,$E8,$71,$97
	.dc.b	$FC,$24,$D7,$31,$BC,$3F,$D1,$3F,$49,$FE,$26,$9F,$81,$E4,$FB,$0F
	.dc.b	$AD,$19,$8F,$E9,$F8,$FF,$82,$5F,$F0,$F7,$97,$FE,$46,$6F,$A7,$C5
	.dc.b	$A6,$BC,$DF,$C0,$4F,$F8,$6E,$32,$7F,$A2,$7E,$8B,$F8,$A3,$FE,$78
	.dc.b	$7F,$DD,$7C,$5F,$E9,$A6,$C7,$FA,$55,$DB,$FE,$91,$DD,$F4,$9F,$29
	.dc.b	$FB,$D1,$9E,$4C,$4F,$CE,$F9,$FF,$F1,$54,$FF,$A1,$97,$D2,$64,$D6
	.dc.b	$5F,$53,$1F,$EC,$55,$E5,$F0,$70,$67,$A3,$C1,$F1,$72,$7F,$D2,$3A
	.dc.b	$7F,$B0,$5B,$DB,$E1,$F8,$C9,$1E,$17,$C9,$F1,$9D,$FE,$93,$C1,$DF
	.dc.b	$2B,$E2,$47,$E1,$BE,$8D,$F4,$9E,$F9,$87,$FA,$82,$32,$FB,$7E,$1F
	.dc.b	$E0,$7C,$44,$8C,$1F,$3F,$DE,$F7,$11,$FA,$5D,$1D,$F6,$91,$D5,$F1
	.dc.b	$7C,$77,$E6,$3E,$93,$FD,$E3,$DE,$69,$8E,$8F,$F6,$BC,$7F,$B6,$3E
	.dc.b	$93,$FC,$F5,$87,$F9,$9D,$AF,$F8,$71,$B1,$BC,$BF,$84,$46,$3B,$15
	.dc.b	$23,$0D,$F8,$2D,$D7,$06,$47,$77,$2F,$F9,$DF,$E8,$DF,$37,$FC,$55
	.dc.b	$B1,$98,$F0,$F9,$D9,$7F,$B8,$42,$76,$FF,$84,$5A,$F9,$8F,$3B,$87
	.dc.b	$18,$7E,$FD,$F8,$5A,$FB,$8F,$7E,$D2,$32,$43,$E3,$3F,$CA,$FE,$8D
	.dc.b	$F4,$94,$93,$FE,$93,$8F,$F1,$3E,$BE,$37,$71,$99,$DB,$FE,$4C,$8D
	.dc.b	$BF,$39,$DB,$98,$DE,$CA,$E9,$E8,$7C,$9E,$39,$7F,$35,$F3,$79,$58
	.dc.b	$8F,$8D,$1F,$2B,$53,$4F,$9B,$93,$E5,$FF,$A4,$7D,$27,$E2,$B2,$F9
	.dc.b	$8D,$C3,$CD,$3F,$0C,$DC,$C8,$F8,$51,$CB,$E8,$64,$7F,$AD,$E9,$1D
	.dc.b	$BE,$83,$33,$FE,$6E,$9C,$BB,$9D,$34,$86,$DC,$BF,$E7,$6F,$A4,$FE
	.dc.b	$65,$B7,$C7,$8F,$CC,$C6,$31,$FC,$33,$63,$F3,$B7,$FD,$25,$E5,$E7
	.dc.b	$7F,$69,$9D,$41,$B3,$BB,$ED,$2C,$7F,$4A,$DE,$7F,$EE,$50,$6F,$3C
	.dc.b	$9F,$F6,$D4,$E5,$B3,$97,$FD,$23,$1F,$CD,$1E,$DF,$01,$D0,$EB,$F4
	.dc.b	$16,$39,$C5,$78,$C5,$C3,$CF,$5C,$C9,$1F,$E0,$7F,$0D,$F2,$54,$97
	.dc.b	$E6,$6B,$FE,$A4,$8D,$BF,$EF,$36,$BE,$88,$FF,$09,$7D,$27,$E7,$30
	.dc.b	$F8,$ED,$DF,$F7,$0D,$AF,$E1,$EF,$E9,$13,$96,$F3,$F4,$3A,$F3,$F6
	.dc.b	$F8,$1E,$FF,$49,$2B,$F8,$CE,$3E,$09,$D4,$63,$1F,$0D,$6E,$3F,$D7
	.dc.b	$F8,$AB,$71,$89,$D5,$F3,$8E,$47,$A2,$73,$7D,$23,$27,$AA,$39,$CF
	.dc.b	$C2,$AD,$CF,$13,$77,$E1,$6C,$69,$E4,$6A,$B1,$29,$EB,$7F,$32,$73
	.dc.b	$18,$AE,$C7,$AC,$EA,$73,$21,$92,$5F,$BD,$F8,$B3,$D1,$DC,$64,$F6
	.dc.b	$3D,$0D,$9A,$AF,$88,$F4,$7C,$CE,$4F,$C0,$D7,$B2,$BD,$11,$F8,$28
	.dc.b	$C3,$B3,$07,$7A,$F1,$57,$48,$EC,$74,$38,$5F,$CD,$72,$72,$3C,$E4
	.dc.b	$9F,$5B,$A8,$E1,$1D,$A3,$A1,$9B,$F3,$CF,$D3,$8F,$25,$78,$EB,$84
	.dc.b	$77,$7A,$98,$9E,$B3,$F1,$A3,$2F,$12,$4F,$82,$32,$FD,$C7,$95,$FC
	.dc.b	$68,$F5,$4F,$32,$3B,$79,$BE,$37,$2F,$E3,$8E,$27,$77,$35,$E6,$8E
	.dc.b	$46,$48,$71,$AF,$1B,$F0,$8D,$99,$9E,$75,$F2,$2F,$F1,$93,$E0,$34
	.dc.b	$74,$BF,$98,$E5,$F1,$3E,$93,$51,$F4,$A7,$E1,$5F,$8A,$7E,$B1,$F8
	.dc.b	$28,$CD,$FC,$2B,$EB,$5F,$32,$BC,$24,$8C,$CF,$C2,$BE,$D3,$ED,$34
	.dc.b	$FA,$AB,$EE,$36,$7B,$36,$7E,$1D,$7A,$8E,$AF,$A8,$F3,$1D,$5D,$47
	.dc.b	$18,$EE,$66,$FE,$53,$4F,$CF,$C6,$E7,$8C,$64,$F9,$D9,$FF,$51,$7F
	.dc.b	$9E,$6F,$FD,$0C,$FE,$73,$F1,$16,$3F,$9F,$45,$7F,$93,$C7,$FB,$3F
	.dc.b	$E9,$3F,$BC,$C3,$3F,$EC,$B1,$86,$F3,$6F,$87,$83,$2B,$F9,$3D,$6F
	.dc.b	$5F,$F1,$E3,$FD,$B6,$33,$F3,$3D,$FC,$09,$18,$21,$E1,$AE,$31,$A8
	.dc.b	$C1,$FE,$5C,$A5,$7A,$4D,$63,$E8,$AC,$57,$F3,$AF,$A4,$7C,$8F,$FB
	.dc.b	$B5,$B1,$98,$FF,$78,$7D,$23,$E9,$75,$12,$74,$7F,$44,$9D,$37,$A3
	.dc.b	$8F,$44,$8C,$10,$F4,$89,$46,$E3,$31,$FE,$07,$3F,$CC,$0E,$3C,$2F
	.dc.b	$29,$E6,$F0,$EC,$78,$E2,$BC,$A2,$E9,$FE,$1E,$AE,$6D,$AE,$55,$FD
	.dc.b	$7B,$E9,$1F,$4B,$FE,$73,$AF,$F1,$2C,$62,$75,$39,$E1,$FF,$3C,$7C
	.dc.b	$D7,$FB,$2A,$F7,$3F,$E8,$0B,$E1,$7F,$E7,$D3,$FD,$26,$E1,$C3,$5E
	.dc.b	$7D,$3D,$5F,$F4,$CD,$FF,$06,$27,$56,$61,$19,$9C,$3A,$8F,$EF,$2F
	.dc.b	$F9,$95,$BE,$16,$8E,$5F,$4B,$FE,$65,$8E,$71,$97,$67,$F8,$0A,$35
	.dc.b	$3C,$E3,$ED,$B7,$2F,$F2,$5B,$73,$FF,$4E,$57,$5F,$B1,$F3,$DF,$C3
	.dc.b	$3F,$D6,$AE,$1C,$35,$F9,$0D,$7F,$D4,$DF,$4C,$F5,$1F,$23,$CE,$53
	.dc.b	$98,$C2,$EF,$FF,$77,$4F,$D3,$8D,$3F,$57,$37,$D2,$B8,$BF,$CE,$51
	.dc.b	$23,$1A,$7F,$8A,$23,$7C,$9A,$8F,$F1,$BC,$EA,$FF,$6F,$5D,$2F,$51
	.dc.b	$FC,$36,$75,$3D,$6F,$2E,$EF,$B7,$D2,$3E,$97,$2D,$8F,$17,$C2,$FC
	.dc.b	$27,$73,$A2,$3A,$D6,$E5,$E4,$BE,$4D,$AF,$0D,$F3,$8C,$5F,$BF,$E9
	.dc.b	$5F,$15,$E6,$6C,$67,$C0,$C4,$E7,$D7,$93,$C7,$D0,$B8,$7C,$4F,$2F
	.dc.b	$F3,$33,$F2,$DB,$19,$F3,$FE,$1C,$63,$7F,$A7,$F4,$BF,$EC,$BB,$FA
	.dc.b	$4B,$1F,$92,$E7,$0F,$A2,$E6,$7E,$C7,$D2,$F3,$7F,$CD,$AB,$1F,$A2
	.dc.b	$E6,$BF,$A1,$39,$9F,$C0,$7D,$2B,$E2,$3E,$BB,$FE,$90,$8C,$C6,$33
	.dc.b	$FB,$7F,$4A,$F8,$DA,$63,$BA,$DF,$E2,$FE,$97,$A3,$70,$FC,$1C,$75
	.dc.b	$CF,$F1,$6E,$9B,$EA,$68,$DF,$A3,$6F,$F2,$DF,$48,$FA,$5F,$C6,$BC
	.dc.b	$78,$FF,$3A,$E6,$77,$96,$B0,$DE,$16,$79,$BF,$96,$FA,$57,$D2,$DF
	.dc.b	$3B,$0D,$AF,$F4,$4F,$D2,$BD,$8F,$D0,$EC,$8F,$F4,$B4,$EA,$73,$18
	.dc.b	$9F,$E8,$FD,$4F,$83,$E3,$F4,$AF,$8E,$E7,$39,$C7,$FD,$2F,$39,$7F
	.dc.b	$D3,$3F,$4B,$F6,$70,$FE,$A2,$BC,$1F,$1B,$9E,$3F,$F7,$1B,$9A,$FF
	.dc.b	$50,$FD,$2B,$DC,$F4,$B6,$3F,$D4,$BF,$4A,$FA,$4C,$D7,$BA,$31,$3A
	.dc.b	$9F,$83,$F4,$AF,$67,$51,$86,$4F,$FA,$C3,$CA,$7D,$2E,$36,$FE,$D3
	.dc.b	$F4,$2F,$A4,$FE,$0F,$F4,$BF,$1F,$83,$FD,$F7,$39,$C7,$F8,$B4,$E5
	.dc.b	$5C,$9D,$3F,$C5,$A7,$4E,$6F,$AF,$B1,$F6,$2B,$78,$7F,$24,$BD,$DA
	.dc.b	$73,$87,$FB,$62,$73,$1F,$36,$3F,$DB,$7F,$48,$FA,$5E,$9E,$57,$F2
	.dc.b	$08,$C4,$FF,$B4,$2B,$86,$1F,$E3,$39,$CB,$F0,$AB,$85,$F9,$7E,$57
	.dc.b	$F8,$CF,$CA,$DC,$7F,$92,$B6,$FE,$4C,$FA,$DC,$66,$AF,$F0,$05,$7E
	.dc.b	$5A,$1F,$31,$63,$0D,$9F,$F7,$2C,$A7,$F3,$B5,$F3,$BE,$97,$E7,$FD
	.dc.b	$27,$E8,$63,$FD,$F7,$3A,$53,$AF,$03,$C3,$85,$9F,$F0,$44,$D7,$FC
	.dc.b	$A9,$5E,$75,$7D,$5F,$BA,$74,$33,$7D,$1A,$74,$29,$D2,$FB,$3E,$93
	.dc.b	$38,$7B,$63,$31,$D6,$BC,$13,$98,$E6,$71,$8C,$DF,$9D,$F3,$46,$6F
	.dc.b	$C9,$18,$8E,$91,$83,$FC,$0D,$87,$C1,$BE,$D5,$B8,$FF,$00,$57,$9C
	.dc.b	$FF,$01,$38,$7E,$11,$FE,$31,$FA,$47,$CA,$3E,$97,$FC,$11,$18,$6C
	.dc.b	$66,$7F,$C2,$D3,$96,$FC,$E7,$FA,$15,$B8,$F2,$8C,$47,$76,$BF,$52
	.dc.b	$51,$C7,$1F,$B1,$93,$6F,$B3,$18,$6C,$71,$9E,$0D,$C3,$F0,$8F,$1C
	.dc.b	$2D,$70,$7F,$7F,$87,$82,$BF,$14,$D8,$FC,$0D,$71,$6D,$71,$53,$BC
	.dc.b	$3F,$1C,$DA,$FC,$79,$C6,$3D,$13,$DB,$0F,$3B,$E2,$6B,$FD,$0E,$FB
	.dc.b	$57,$13,$D9,$1E,$23,$F8,$BE,$1F,$9F,$78,$18,$D7,$C3,$5F,$E6,$3C
	.dc.b	$7C,$93,$FC,$7A,$FD,$97,$C5,$B7,$C2,$79,$1E,$59,$FD,$77,$D2,$FE
	.dc.b	$EE,$32,$1C,$3F,$E9,$A6,$C7,$95,$FB,$B1,$C1,$CC,$FE,$D1,$B1,$FC
	.dc.b	$C9,$AF,$5C,$7F,$CF,$9F,$4B,$FD,$02,$3C,$2D,$7F,$D5,$2D,$BF,$89
	.dc.b	$F4,$7F,$E9,$1C,$3F,$80,$CE,$5B,$D3,$D0,$FF,$44,$4E,$EF,$FA,$1D
	.dc.b	$7F,$8F,$1E,$FE,$27,$88,$DF,$A5,$63,$24,$3F,$D7,$75,$FE,$1D,$71
	.dc.b	$8F,$D7,$BE,$91,$E8,$BE,$6F,$CC,$A7,$CC,$79,$09,$1F,$E6,$DF,$81
	.dc.b	$FF,$45,$3F,$D6,$27,$6F,$D4,$FA,$5F,$F1,$C3,$7A,$78,$BF,$EC,$E8
	.dc.b	$CB,$63,$FD,$51,$18,$CB,$E4,$E1,$FE,$AB,$FA,$5F,$F1,$35,$F9,$7C
	.dc.b	$8F,$8E,$67,$D4,$B1,$F8,$4F,$2F,$FC,$84,$DF,$4F,$2F,$FB,$7A,$32
	.dc.b	$DF,$37,$FD,$71,$1F,$E1,$B8,$E0,$7F,$A2,$FE,$89,$EC,$FF,$B0,$74
	.dc.b	$46,$27,$53,$98,$FF,$6D,$F2,$FE,$9B,$85,$F7,$BC,$5B,$7F,$2E,$31
	.dc.b	$BF,$5F,$9E,$FF,$0C,$4F,$FA,$6D,$AF,$FA,$61,$B9,$7F,$A8,$9F,$B6
	.dc.b	$FF,$A9,$31,$F8,$F3,$FE,$9E,$7D,$F1,$9D,$7C,$DB,$78,$FD,$AF,$27
	.dc.b	$1C,$C3,$EF,$47,$8F,$EC,$3C,$9F,$77,$27,$C9,$1E,$FF,$BB,$91,$F1
	.dc.b	$DF,$85,$EA,$FB,$FE,$C2,$4F,$E2,$5F,$85,$D9,$EF,$F6,$24,$FB,$DF
	.dc.b	$85,$EC,$FC,$BF,$95,$F9,$9F,$7B,$2B,$EA,$47,$CB,$F9,$EF,$99,$E2
	.dc.b	$C9,$FC,$B4,$7C,$BF,$9C,$F9,$9F,$3B,$27,$E6,$47,$FB,$DF,$C6,$7C
	.dc.b	$55,$F9,$C9,$5F,$C3,$A7,$2D,$9C,$AE,$8F,$E4,$52,$AE,$95,$2B,$94
	.dc.b	$70,$AD,$CF,$F8,$11,$FE,$55,$F4,$6F,$84,$C7,$81,$FE,$08,$8C,$61
	.dc.b	$FC,$81,$AF,$86,$35,$F0,$9C,$23,$FC,$97,$3A,$BE,$B7,$99,$4E,$6F
	.dc.b	$8C,$7F,$2C,$FA,$3F,$53,$FE,$F9,$6B,$FE,$1A,$3F,$CA,$EE,$1C,$35
	.dc.b	$FF,$0F,$BC,$2F,$DB,$C1,$E1,$B3,$E3,$9D,$5F,$BF,$C2,$D7,$C8,$CC
	.dc.b	$7E,$04,$71,$30,$FC,$13,$DF,$19,$76,$7C,$77,$78,$FC,$E9,$F8,$5F
	.dc.b	$46,$E2,$A4,$BF,$E9,$B1,$FE,$2A,$C3,$FC,$72,$EE,$33,$3B,$7C,$42
	.dc.b	$38,$3F,$1B,$E3,$FF,$C4,$79,$1A,$48,$F9,$CF,$93,$E6,$A4,$7D,$38
	.dc.b	$F3,$FC,$D7,$D9,$FA,$2F,$EE,$1A,$7F,$71,$6E,$7E,$69,$DD,$FF,$66
	.dc.b	$9C,$BF,$E8,$E6,$BE,$78,$DB,$EA,$8C,$9A,$CB,$F9,$FC,$21,$CB,$E3
	.dc.b	$24,$77,$E6,$AF,$38,$B8,$F9,$E7,$F2,$5F,$47,$FC,$FF,$2F,$F3,$2E
	.dc.b	$1F,$82,$C3,$CD,$CD,$AF,$0E,$47,$AE,$3E,$11,$87,$E8,$E1,$F9,$B7
	.dc.b	$FD,$56,$B9,$BF,$40,$D8,$FF,$04,$BB,$4D,$34,$F4,$8E,$3C,$D2,$38
	.dc.b	$C7,$43,$0F,$8B,$E1,$3B,$91,$EF,$CD,$5D,$1E,$28,$F2,$C7,$ED,$7E
	.dc.b	$67,$FA,$64,$FB,$39,$FD,$26,$7C,$C9,$1E,$A8,$CB,$FE,$24,$9E,$3C
	.dc.b	$1F,$E3,$5F,$A3,$FC,$C4,$FE,$59,$B1,$9E,$8F,$E0,$91,$F9,$89,$3A
	.dc.b	$E1,$FF,$4C,$B6,$3F,$DA,$ED,$77,$DB,$CA,$A4,$7F,$8D,$A7,$47,$52
	.dc.b	$C6,$A3,$C9,$36,$FC,$91,$C6,$7C,$B3,$9B,$F1,$C5,$9F,$1C,$D7,$D3
	.dc.b	$16,$FC,$0D,$39,$C7,$52,$1F,$15,$CB,$4C,$DF,$91,$F3,$98,$BF,$1B
	.dc.b	$4F,$19,$4F,$19,$6B,$C6,$DA,$EF,$5D,$49,$5D,$27,$B9,$E3,$24,$69
	.dc.b	$EE,$62,$38,$46,$63,$EA,$9D,$CF,$15,$FC,$36,$99,$39,$1D,$23,$95
	.dc.b	$FD,$88,$FB,$93,$C5,$D9,$D8,$B1,$B7,$D3,$3C,$AF,$EA,$57,$18,$EA
	.dc.b	$DB,$F4,$9F,$90,$71,$5F,$90,$AF,$8F,$3A,$8C,$1E,$A3,$C3,$1D,$9F
	.dc.b	$65,$76,$6C,$F1,$7D,$11,$E5,$79,$35,$F5,$D7,$89,$FB,$67,$B2,$79
	.dc.b	$5F,$BE,$BF,$61,$5B,$8F,$D0,$B6,$FD,$37,$E8,$3C,$E7,$CC,$3C,$A6
	.dc.b	$6B,$9B,$83,$C2,$E1,$E2,$79,$8D,$57,$81,$B3,$E0,$78,$DF,$A2,$31
	.dc.b	$3E,$98,$F5,$DF,$9E,$BC,$87,$6B,$F9,$4D,$9D,$BD,$8D,$C7,$C9,$36
	.dc.b	$58,$DC,$7F,$13,$F8,$4F,$E2,$3D,$32,$3F,$84,$47,$F1,$43,$0F,$F1
	.dc.b	$98,$C3,$FC,$7F,$7F,$F0,$E7,$D1,$FF,$5C,$3F,$AE,$E9,$FC,$EB,$1F
	.dc.b	$F6,$B3,$71,$5C,$AE,$11,$FC,$C6,$F9,$F3,$CA,$FF,$67,$D7,$F6,$06
	.dc.b	$F1,$E5,$FF,$00,$C6,$5B,$3E,$49,$D6,$3F,$EE,$7A,$D3,$67,$55,$DB
	.dc.b	$BB,$D7,$CB,$23,$04,$3B,$B9,$7C,$31,$FE,$3D,$FA,$37,$85,$FE,$3F
	.dc.b	$FA,$2F,$81,$5F,$01,$AA,$76,$FF,$BB,$1A,$FF,$87,$23,$67,$08,$CE
	.dc.b	$3F,$F0,$06,$1C,$FD,$F9,$E5,$FF,$BD,$A7,$2D,$9C,$BD,$67,$CD,$12
	.dc.b	$38,$46,$AB,$B4,$FA,$FC,$09,$5D,$5F,$77,$3E,$23,$D1,$16,$BC,$D3
	.dc.b	$FE,$61,$FA,$17,$D1,$7F,$96,$EF,$F0,$62,$32,$E4,$C4,$56,$99,$8D
	.dc.b	$BF,$E4,$2F,$37,$FB,$51,$FD,$AE,$33,$6E,$4E,$F2,$E7,$FF,$3A,$36
	.dc.b	$FF,$2B,$5B,$7F,$2F,$19,$7B,$73,$F0,$A9,$1C,$2F,$AC,$74,$9C,$CF
	.dc.b	$96,$2D,$F5,$85,$AB,$C7,$E0,$C9,$E8,$78,$9F,$F4,$57,$D1,$7F,$92
	.dc.b	$AB,$FC,$8A,$48,$CE,$3F,$E8,$FF,$A0,$7D,$0B,$E3,$BE,$64,$61,$BD
	.dc.b	$8F,$A8,$CD,$6F,$3F,$AD,$F5,$8F,$A1,$F4,$47,$D8,$8C,$65,$F5,$DB
	.dc.b	$1C,$27,$6F,$FA,$DF,$E8,$DC,$BF,$37,$86,$46,$30,$F7,$C2,$1C,$BB
	.dc.b	$9F,$AB,$1B,$BF,$F4,$95,$6E,$F5,$1D,$DB,$1C,$EE,$D7,$18,$F0,$CE
	.dc.b	$0C,$CD,$BF,$1F,$D1,$7F,$B3,$5B,$F7,$BD,$23,$18,$F8,$94,$67,$6D
	.dc.b	$6B,$6F,$9B,$E4,$F1,$9E,$ED,$F7,$70,$DA,$F8,$7F,$45,$FE,$99,$6F
	.dc.b	$9F,$EC,$86,$36,$3D,$FF,$45,$D2,$BE,$5B,$75,$F8,$EB,$1D,$E3,$F1
	.dc.b	$BE,$8B,$E2,$5F,$E8,$9B,$7F,$EE,$F7,$E9,$C7,$D6,$73,$C2,$FA,$27
	.dc.b	$9F,$F4,$1F,$04,$7D,$F7,$35,$F7,$9C,$BF,$B0,$FA,$2E,$67,$6C,$3E
	.dc.b	$EB,$8E,$3F,$B6,$BE,$12,$FD,$56,$1B,$7F,$8A,$FA,$37,$1F,$E0,$9B
	.dc.b	$19,$E7,$C4,$E6,$A4,$F1,$BB,$97,$E3,$BE,$8D,$F0,$3F,$8C,$6D,$7E
	.dc.b	$61,$7C,$16,$67,$F2,$74,$FA,$3F,$E2,$17,$FB,$C8,$C7,$8D,$A9,$7A
	.dc.b	$96,$38,$FC,$13,$E2,$29,$D3,$E3,$7E,$16,$3C,$E7,$4F,$5F,$A3,$7C
	.dc.b	$5F,$89,$B1,$9E,$8C,$4F,$9B,$E8,$DF,$47,$4F,$4F,$D1,$BE,$8B,$18
	.dc.b	$7A,$63,$0D,$8F,$7F,$D1,$FF,$54,$8C,$1E,$F7,$1C,$3F,$EF,$86,$F6
	.dc.b	$FF,$23,$F3,$1F,$2B,$CB,$D5,$F9,$3F,$42,$FA,$3F,$DD,$FD,$1F,$CA
	.dc.b	$CB,$FE,$00,$CE,$73,$FE,$13,$39,$CB,$F9,$74,$E6,$BF,$83,$CE,$6B
	.dc.b	$FA,$8C,$E6,$BF,$8A,$4E,$6B,$FC,$2F,$39,$AF,$E8,$B3,$99,$4E,$55
	.dc.b	$27,$53,$FA,$09,$FE,$2F,$F4,$6F,$A3,$F6,$F1,$FF,$16,$8C,$63,$F8
	.dc.b	$5B,$E1,$8F,$F2,$79,$CE,$1F,$02,$F8,$61,$F1,$BE,$77,$F8,$FF,$23
	.dc.b	$E8,$B1,$9F,$F2,$C8,$C3,$D6,$73,$87,$CA,$2E,$5F,$D5,$4D,$BE,$C9
	.dc.b	$FE,$EF,$0F,$A3,$FC,$EB,$A3,$83,$8C,$24,$FE,$7E,$BF,$C8,$3F,$47
	.dc.b	$FE,$42,$FA,$1F,$CF,$C7,$F7,$0A,$FD,$3B,$A1,$1A,$7E,$6C,$7F,$75
	.dc.b	$8C,$46,$DD,$61,$E8,$73,$0A,$F0,$77,$7A,$AF,$93,$FE,$13,$8C,$CF
	.dc.b	$8F,$0A,$FF,$85,$A2,$E1,$D6,$B8,$4F,$99,$E9,$5C,$23,$C9,$18,$74
	.dc.b	$7F,$88,$0F,$21,$83,$DB,$1C,$5C,$E1,$EC,$C3,$CF,$1C,$70,$F1,$C6
	.dc.b	$23,$CC,$7F,$89,$E3,$FC,$50,$F6,$3F,$C5,$8F,$F8,$B7,$E8,$5F,$44
	.dc.b	$FA,$3C,$36,$BF,$C6,$D3,$A9,$C9,$C7,$E5,$7F,$3F,$3A,$6B,$ED,$94
	.dc.b	$62,$3F,$C7,$F1,$8D,$39,$1F,$3B,$46,$DF,$E4,$4D,$3F,$01,$18,$BF
	.dc.b	$A7,$7C,$5F,$90,$D9,$DD,$71,$AF,$F2,$69,$C6,$B7,$87,$A2,$BF,$19
	.dc.b	$7C,$6B,$77,$C6,$FF,$D2,$F1,$C6,$B7,$7C,$63,$D6,$D9,$DD,$71,$7F
	.dc.b	$9D,$B6,$FB,$47,$FA,$89,$AE,$8F,$85,$1B,$9E,$91,$93,$F6,$47,$63
	.dc.b	$FA,$55,$7B,$70,$FF,$26,$5F,$48,$C6,$BE,$CA,$DE,$1F,$BA,$8F,$4D
	.dc.b	$FE,$7B,$E8,$FF,$84,$57,$F5,$56,$BE,$00,$E2,$3F,$CB,$4F,$DE,$8E
	.dc.b	$1B,$75,$7F,$D7,$F1,$BA,$FF,$3F,$B6,$33,$96,$D7,$D1,$FF,$42,$7F
	.dc.b	$B3,$35,$F1,$37,$0F,$F3,$57,$D1,$BE,$B7,$94,$E7,$1F,$F4,$6B,$6F
	.dc.b	$FD,$A2,$D5,$5C,$5E,$FE,$27,$22,$3F,$DD,$E1,$F3,$98,$8F,$1C,$DC
	.dc.b	$7C,$73,$70,$F0,$CD,$AF,$F7,$0D,$F2,$7A,$7A,$DF,$06,$3F,$C6,$91
	.dc.b	$83,$FC,$C7,$E8,$3E,$0F,$55,$F1,$FE,$93,$FC,$5F,$5F,$E0,$B6,$BF
	.dc.b	$5E,$F9,$BF,$66,$33,$9F,$FA,$45,$B9,$7F,$AC,$9B,$3E,$6B,$FA,$9F
	.dc.b	$48,$FA,$1F,$37,$5C,$AF,$F0,$D4,$E5,$F8,$D1,$F6,$FC,$C7,$D1,$53
	.dc.b	$84,$FE,$2E,$32,$FF,$88,$DF,$F1,$E3,$7D,$4E,$0F,$FA,$13,$E8,$FF
	.dc.b	$C7,$6D,$F5,$BF,$1D,$B2,$73,$3F,$E8,$47,$FD,$1E,$FC,$AF,$73,$7C
	.dc.b	$5F,$E7,$08,$C1,$F3,$B6,$E7,$F4,$9F,$E1,$F8,$FF,$49,$B5,$FF,$47
	.dc.b	$B7,$0F,$F4,$D2,$E4,$F1,$E3,$F0,$EB,$C5,$AE,$0C,$F0,$78,$BE,$2B
	.dc.b	$C7,$E2,$CA,$FB,$F1,$E2,$F5,$BC,$7E,$BC,$9F,$1C,$76,$FA,$49,$1F
	.dc.b	$15,$F0,$FA,$DE,$EF,$A4,$93,$F8,$A7,$E0,$F6,$7B,$BE,$92,$4F,$B9
	.dc.b	$F8,$3F,$2B,$EA,$F9,$3F,$5F,$D2,$4A,$FA,$51,$F2,$BE,$7B,$E5,$F9
	.dc.b	$12,$7F,$2F,$1F,$2B,$E6,$BE,$5F,$CE,$49,$F9,$71,$FB,$FF,$1D,$F3
	.dc.b	$57,$E6,$E5,$7F,$99,$67,$2D,$9C,$BF,$C4,$E1,$3C,$E6,$4F,$18,$DF
	.dc.b	$5F,$FB,$BE,$B9,$CF,$F1,$BF,$A4,$7D,$07,$F1,$16,$C7,$F9,$D5,$AE
	.dc.b	$FE,$7F,$53,$23,$4F,$18,$FE,$35,$F4,$9E,$49,$FF,$3F,$35,$FE,$54
	.dc.b	$DC,$1E,$4A,$66,$1F,$DD,$1E,$39,$7F,$81,$DD,$0B,$E8,$FF,$82,$DA
	.dc.b	$E9,$AF,$4F,$0B,$DD,$F1,$72,$3D,$D1,$F0,$4C,$3F,$16,$30,$E8,$F8
	.dc.b	$8E,$B1,$F4,$C7,$53,$FD,$39,$F4,$8E,$1F,$E7,$8D,$8C,$C6,$38,$B2
	.dc.b	$C8,$C9,$CA,$3E,$5F,$C8,$48,$F9,$6F,$4F,$17,$11,$F1,$73,$4F,$C7
	.dc.b	$8F,$99,$18,$70,$D7,$E6,$61,$FC,$B5,$FF,$5A,$7D,$1F,$F9,$87,$7F
	.dc.b	$9E,$D8,$FA,$D3,$B7,$FB,$23,$73,$F8,$70,$9E,$0D,$AF,$87,$E7,$C8
	.dc.b	$EF,$1F,$D8,$67,$77,$CE,$3C,$3C,$B9,$25,$79,$B9,$AB,$FD,$D6,$D3
	.dc.b	$8C,$7C,$70,$AF,$ED,$DF,$49,$FC,$FA,$CD,$31,$3F,$C6,$E3,$FD,$AD
	.dc.b	$1F,$D3,$1F,$F2,$C3,$5F,$C9,$B5,$FC,$D3,$5F,$07,$2F,$F7,$DA,$72
	.dc.b	$D3,$11,$27,$FD,$C9,$3A,$6C,$E5,$FF,$09,$47,$ED,$30,$F9,$EE,$E3
	.dc.b	$33,$B9,$CF,$C0,$F8,$49,$3D,$BA,$6B,$D2,$2C,$F4,$9F,$F0,$A7,$C9
	.dc.b	$FF,$4C,$8F,$BA,$DD,$7F,$66,$DB,$FD,$D6,$35,$45,$27,$99,$F0,$CF
	.dc.b	$F1,$07,$D2,$7E,$96,$7F,$C5,$CD,$C3,$F1,$6D,$C3,$F8,$0B,$FE,$9E
	.dc.b	$6F,$37,$22,$57,$F0,$BA,$E3,$CD,$88,$EB,$36,$73,$1B,$AE,$6D,$3F
	.dc.b	$8A,$5F,$85,$B5,$E8,$BE,$55,$C2,$7A,$9C,$45,$79,$DE,$26,$5E,$8E
	.dc.b	$AB,$A3,$97,$AB,$C0,$F8,$B5,$87,$81,$B3,$A3,$99,$EA,$47,$13,$96
	.dc.b	$CF,$13,$A9,$6F,$E8,$CE,$E3,$17,$EF,$9C,$C7,$80,$C9,$E7,$72,$F4
	.dc.b	$3E,$29,$1C,$9F,$49,$C4,$75,$23,$F4,$AB,$CE,$E8,$C1,$E0,$3E,$19
	.dc.b	$93,$E8,$BC,$27,$BC,$FE,$2A,$30,$F6,$9D,$9C,$EF,$E4,$DF,$D4,$8E
	.dc.b	$91,$87,$E0,$18,$8F,$8C,$D3,$95,$76,$7E,$25,$78,$5E,$D3,$C2,$FC
	.dc.b	$B1,$A2,$1E,$43,$A1,$E2,$AE,$8F,$C0,$8F,$94,$72,$71,$5E,$DB,$FC
	.dc.b	$94,$60,$EA,$FE,$15,$F1,$9E,$82,$9D,$CB,$3D,$63,$99,$F0,$E3,$47
	.dc.b	$9A,$35,$5E,$69,$ED,$39,$AF,$24,$E6,$FE,$75,$F5,$7E,$24,$FA,$5B
	.dc.b	$5E,$86,$BE,$B2,$5F,$6A,$E5,$1E,$B3,$C0,$7C,$0B,$F0,$3F,$6A,$36
	.dc.b	$FA,$9A,$7A,$9F,$83,$7E,$86,$9C,$CF,$DE,$7C,$67,$F1,$CF,$9F,$91
	.dc.b	$82,$1F,$C7,$9C,$BF,$C8,$23,$F9,$7E,$F2,$3F,$85,$69,$FC,$F7,$0F
	.dc.b	$E4,$7F,$17,$99,$FF,$6A,$4E,$71,$FE,$67,$7F,$CB,$2B,$FD,$99,$F4
	.dc.b	$9F,$D9,$2F,$F8,$A3,$5F,$F6,$E3,$70,$FE,$F2,$FF,$67,$8D,$F2,$F2
	.dc.b	$23,$CB,$82,$BC,$A2,$D7,$29,$F0,$F8,$1E,$2E,$E9,$7E,$28,$CB,$E4
	.dc.b	$8C,$BE,$48,$FF,$72,$7D,$23,$C3,$A4,$9F,$E5,$DF,$48,$FA,$8C,$47
	.dc.b	$F9,$1E,$B8,$3F,$E2,$36,$B9,$AF,$F1,$32,$A7,$47,$FB,$CD,$74,$53
	.dc.b	$11,$27,$C3,$19,$7C,$71,$95,$A7,$F8,$92,$B9,$C7,$03,$67,$AB,$C2
	.dc.b	$F0,$11,$F6,$F8,$55,$E7,$C7,$FE,$62,$AF,$F5,$EC,$60,$FF,$35,$38
	.dc.b	$BF,$05,$7F,$9A,$1F,$F1,$AF,$D2,$3E,$97,$18,$7F,$9B,$23,$18,$7F
	.dc.b	$59,$6B,$D4,$79,$EF,$7B,$8C,$7F,$9E,$67,$6F,$F9,$D8,$38,$7B,$5F
	.dc.b	$D0,$8C,$47,$FA,$06,$9D,$3D,$23,$BB,$63,$31,$FD,$B5,$75,$66,$64
	.dc.b	$E5,$E1,$18,$8F,$54,$F2,$6C,$E9,$E9,$E1,$F7,$23,$FD,$EF,$E9,$7F
	.dc.b	$CB,$75,$FE,$55,$8F,$44,$E9,$F8,$D3,$FE,$60,$FA,$8F,$8F,$1F,$E7
	.dc.b	$36,$BF,$E9,$B6,$BF,$1D,$AF,$FA,$7B,$E9,$9F,$40,$F9,$2F,$F3,$E4
	.dc.b	$61,$BE,$03,$E9,$B3,$5B,$BF,$A5,$C4,$FA,$47,$37,$FA,$62,$24,$63
	.dc.b	$93,$FC,$41,$1B,$E5,$D4,$7F,$8A,$E7,$57,$F8,$FA,$E9,$7A,$8F,$CF
	.dc.b	$CE,$A7,$AD,$E5,$DD,$F6,$9D,$3D,$6F,$50,$AB,$1C,$7E,$93,$FD,$BF
	.dc.b	$5E,$65,$F4,$99,$EF,$D3,$BD,$78,$69,$6F,$83,$E3,$F2,$1F,$01,$7D
	.dc.b	$0C,$36,$BE,$1F,$D2,$7F,$AB,$1B,$E8,$FA,$E3,$FC,$BB,$18,$3E,$4F
	.dc.b	$D2,$74,$6D,$7C,$18,$5D,$8F,$89,$F4,$9F,$0B,$8F,$F8,$3B,$78,$CF
	.dc.b	$C3,$C3,$EA,$B9,$FA,$47,$BF,$F3,$5F,$05,$FD,$57,$35,$F7,$1C,$DF
	.dc.b	$E0,$7E,$93,$B1,$D6,$BE,$E3,$8E,$5F,$B0,$BE,$1A,$FD,$1E,$1B,$8F
	.dc.b	$E2,$3E,$91,$C5,$F7,$DB,$19,$E8,$C4,$E6,$A4,$F1,$BB,$97,$E3,$7E
	.dc.b	$91,$F0,$7F,$89,$6D,$7E,$59,$7C,$26,$67,$F2,$54,$FA,$4F,$E1,$77
	.dc.b	$FE,$73,$6C,$7F,$1B,$5E,$B7,$07,$3C,$1F,$ED,$AF,$84,$7C,$5F,$46
	.dc.b	$E1,$FE,$7D,$F8,$AF,$88,$FF,$9F,$E3,$38,$FF,$27,$FA,$47,$C6,$F9
	.dc.b	$1B,$5F,$E8,$57,$EA,$CE,$E7,$D3,$F4,$8F,$6E,$9E,$CF,$A4,$7D,$26
	.dc.b	$31,$FF,$47,$46,$1B,$1F,$0B,$E9,$3F,$AD,$C6,$0F,$F4,$A3,$8C,$7F
	.dc.b	$9E,$37,$6F,$F7,$CF,$94,$FA,$4E,$5E,$BF,$CC,$FA,$27,$D2,$7E,$EF
	.dc.b	$E9,$3E,$3E,$6C,$E5,$6A,$3F,$85,$CE,$A7,$35,$C7,$D5,$E8,$DA,$FF
	.dc.b	$54,$5F,$1A,$FE,$1D,$7C,$54,$67,$3F,$E3,$53,$98,$F9,$71,$FC,$77
	.dc.b	$E9,$1F,$49,$E9,$F1,$7F,$89,$46,$31,$EF,$18,$9D,$47,$F2,$29,$D6
	.dc.b	$3F,$D8,$70,$7C,$F7,$F1,$BE,$57,$CF,$7C,$58,$C3,$70,$F5,$C7,$C9
	.dc.b	$BC,$1F,$DA,$F3,$E3,$3F,$EC,$75,$F4,$9F,$9A,$73,$3B,$8C,$5C,$AF
	.dc.b	$F6,$74,$FD,$1F,$A4,$FA,$5F,$45,$F9,$C8,$FF,$71,$57,$E9,$1D,$08
	.dc.b	$D3,$FE,$16,$85,$72,$EF,$FE,$F5,$2B,$C3,$D2,$6B,$F8,$13,$47,$59
	.dc.b	$FF,$08,$57,$48,$C1,$FE,$05,$AE,$38,$FB,$B0,$F1,$CF,$99,$F2,$C6
	.dc.b	$6F,$C7,$5C,$2F,$A5,$70,$BF,$0B,$C2,$33,$5C,$27,$C2,$7B,$F0,$F4
	.dc.b	$BD,$E3,$CF,$18,$8E,$71,$87,$89,$F2,$8F,$8A,$7C,$02,$DF,$F9,$31
	.dc.b	$C1,$F1,$7E,$89,$F4,$8F,$A4,$E3,$18,$6C,$66,$7E,$8C,$E5,$BF,$2D
	.dc.b	$FE,$4C,$6E,$9C,$A3,$11,$FE,$31,$94,$62,$3C,$4F,$1D,$3F,$03,$97
	.dc.b	$E7,$27,$ED,$B5,$F8,$F5,$CA,$FE,$FB,$63,$E5,$CE,$EB,$9E,$DF,$83
	.dc.b	$7F,$1B,$5B,$6E,$1E,$1A,$FC,$41,$C4,$B5,$C5,$F3,$E1,$E0,$9E,$18
	.dc.b	$7F,$0F,$9E,$0D,$AE,$0F,$F1,$3C,$3C,$13,$CA,$3C,$F3,$A7,$FA,$0B
	.dc.b	$67,$71,$CC,$FE,$39,$7F,$99,$76,$63,$4F,$3B,$4F,$06,$1E,$67,$67
	.dc.b	$F9,$A7,$0F,$99,$5C,$DB,$5E,$5C,$7F,$67,$F4,$9F,$C3,$2F,$A0,$70
	.dc.b	$FF,$9C,$DB,$7E,$57,$EF,$C7,$07,$35,$FB,$56,$C7,$F3,$76,$BD,$70
	.dc.b	$FF,$38,$7D,$27,$F4,$18,$F0,$B5,$FF,$41,$36,$FE,$27,$D2,$7F,$A0
	.dc.b	$E7,$7D,$7C,$88,$FA,$A3,$4F,$2F,$13,$B9,$1D,$7C,$E5,$74,$78,$A3
	.dc.b	$FD,$09,$29,$D5,$FD,$5C,$15,$8C,$FF,$8A,$5E,$F0,$F1,$7A,$BC,$88
	.dc.b	$FF,$98,$3C,$FF,$F3,$E2,$9F,$F1,$BB,$F4,$BE,$93,$FC,$5E,$DE,$CE
	.dc.b	$2F,$FA,$96,$32,$D8,$FF,$4D,$46,$33,$F8,$F8,$7F,$A7,$7E,$91,$F4
	.dc.b	$3E,$CE,$B9,$8F,$E0,$27,$2F,$C9,$8F,$F5,$27,$96,$FA,$2A,$70,$9F
	.dc.b	$F5,$B4,$65,$FC,$4B,$FE,$39,$6F,$A5,$C1,$ED,$F4,$5F,$EB,$AC,$DF
	.dc.b	$B5,$FF,$A2,$E7,$FD,$06,$D7,$FD,$85,$92,$BF,$D6,$DF,$83,$7F,$AA
	.dc.b	$EF,$FD,$19,$C3,$FE,$20,$BF,$99,$A6,$BC,$7F,$F5,$E7,$DF,$FF,$9C
	.dc.b	$27,$F3,$BE,$13,$E0,$BF,$D8,$0D,$8F,$D0,$CE,$DF,$0F,$C9,$79,$F6
	.dc.b	$FF,$4E,$F3,$FE,$D9,$B3,$FE,$13,$DF,$FD,$DE,$FF,$B6,$7E,$5A,$ED
	.dc.b	$E6,$F2,$FE,$06,$FF,$E1,$CA,$FD,$7D,$6B,$1F,$EB,$9F,$83,$3D,$AF
	.dc.b	$F6,$A7,$E0,$0F,$9F,$A7,$FB,$5E,$75,$87,$D9,$F2,$3E,$67,$D4,$3E
	.dc.b	$C2,$9F,$81,$8E,$6F,$8B,$E5,$25,$7F,$00,$C7,$83,$FC,$0B,$F0,$0F
	.dc.b	$33,$FE,$F3,$FC,$03,$EA,$70,$FF,$15,$6E,$DF,$DC,$EB,$73,$FC,$67
	.dc.b	$8B,$FC,$1B,$7F,$EB,$ED,$3F,$D5,$CD,$F8,$5F,$CA,$7A,$3F,$EF,$1C
	.dc.b	$BF,$C5,$33,$FA,$5E,$AF,$8E,$67,$D4,$FF,$7B,$7D,$FF,$F5,$C9,$FE
	.dc.b	$2F,$5F,$AA,$3F,$57,$19,$86,$7F,$CA,$FF,$2E,$C3,$78,$BF,$E0,$04
	.dc.b	$63,$7F,$F0,$67,$9A,$FB,$5C,$D7,$F8,$3E,$FF,$9A,$4B,$E8,$71,$7C
	.dc.b	$1D,$E3,$FC,$AB,$1E,$8F,$F9,$5E,$B8,$3F,$CB,$9A,$F9,$AB,$FC,$39
	.dc.b	$3C,$23,$FB,$5F,$4F,$67,$97,$C9,$57,$FB,$7B,$6D,$34,$C4,$49,$F0
	.dc.b	$47,$1F,$AB,$FC,$64,$7F,$40,$6B,$E3,$3F,$CD,$CE,$32,$FE,$A8,$BE
	.dc.b	$97,$97,$5C,$AE,$4F,$58,$F3,$7D,$5B,$1F,$F0,$CD,$23,$0F,$F8,$71
	.dc.b	$AF,$F9,$09,$AF,$F8,$FC,$E1,$1E,$9F,$06,$47,$F5,$89,$E3,$8F,$F6
	.dc.b	$96,$EB,$FE,$47,$CB,$F5,$D5,$EC,$BF,$2E,$1F,$0F,$BF,$E2,$73,$CA
	.dc.b	$F8,$91,$97,$E3,$46,$5F,$8C,$2D,$3F,$57,$EA,$DF,$53,$49,$5F,$DB
	.dc.b	$A3,$FC,$9F,$97,$F9,$6D,$DC,$66,$76,$F6,$8D,$BF,$EA,$1C,$DF,$2B
	.dc.b	$F3,$1B,$1F,$2E,$30,$47,$C1,$DF,$E5,$3E,$AC,$48,$C9,$0F,$6C,$F4
	.dc.b	$7C,$5F,$3D,$F9,$1F,$AB,$FF,$0B,$70,$7F,$85,$DB,$9F,$AE,$7E,$11
	.dc.b	$B6,$57,$C1,$8E,$5F,$51,$23,$FC,$23,$48,$ED,$CF,$99,$F9,$F3,$97
	.dc.b	$73,$A7,$FC,$33,$F5,$7F,$D8,$70,$FF,$43,$57,$E0,$A7,$FD,$90,$D5
	.dc.b	$3B,$7F,$D9,$8D,$51,$B3,$84,$64,$D3,$9E,$2F,$DA,$47,$1C,$3F,$51
	.dc.b	$7F,$E9,$77,$F5,$17,$FE,$2E,$6B,$F8,$A7,$6F,$FB,$75,$AF,$F0,$FE
	.dc.b	$4F,$E0,$11,$C3,$3F,$F2,$33,$A1,$F0,$BA,$25,$77,$E5,$AF,$78,$B5
	.dc.b	$CA,$7F,$86,$79,$8F,$92,$EF,$CF,$30,$FF,$22,$CE,$5B,$39,$7D,$53
	.dc.b	$CA,$24,$7B,$E3,$57,$FE,$5D,$AE,$53,$FA,$A6,$C7,$B2,$77,$9F,$FA
	.dc.b	$FE,$72,$DB,$CD,$FC,$7A,$EE,$F3,$8C,$BF,$CD,$E3,$37,$E1,$9C,$E5
	.dc.b	$FD,$52,$B8,$1E,$C3,$D0,$43,$24,$9F,$33,$97,$A3,$C4,$F8,$66,$6F
	.dc.b	$D0,$E6,$38,$BA,$3A,$46,$9C,$46,$8E,$AE,$8F,$80,$75,$07,$F8,$76
	.dc.b	$BD,$F3,$F9,$1B,$F8,$78,$7C,$AA,$DB,$5F,$8C,$73,$39,$1E,$79,$E4
	.dc.b	$E6,$38,$1C,$70,$F9,$58,$79,$EB,$D4,$FA,$0C,$57,$A0,$F5,$1D,$8F
	.dc.b	$7C,$76,$73,$5D,$6F,$D3,$87,$B2,$79,$5F,$47,$93,$DE,$F9,$39,$BF
	.dc.b	$3D,$7E,$02,$F9,$5F,$0B,$D9,$1F,$8B,$3A,$C3,$DB,$39,$9E,$06,$8F
	.dc.b	$91,$1A,$70,$7B,$9C,$3F,$38,$F6,$CF,$8D,$F9,$C7,$0C,$3D,$93,$C3
	.dc.b	$0F,$4C,$F6,$C3,$D3,$1F,$16,$3E,$A5,$F9,$E7,$E3,$C7,$43,$C1,$7F
	.dc.b	$0E,$75,$1C,$0D,$5F,$98,$F0,$46,$A3,$EA,$9E,$5C,$3E,$E9,$F0,$E7
	.dc.b	$71,$CA,$FC,$2F,$C1,$6D,$FD,$58,$FB,$35,$F2,$F0,$F6,$C7,$63,$CF
	.dc.b	$1D,$9F,$01,$C0,$87,$D2,$25,$FA,$6F,$E2,$E1,$F3,$67,$E7,$9F,$24
	.dc.b	$F8,$97,$F2,$5F,$03,$E2,$C3,$E8,$D7,$29,$E0,$72,$3C,$D7,$F0,$1F
	.dc.b	$71,$F2,$CF,$F0,$2F,$D5,$FF,$7C,$6F,$37,$F8,$E6,$31,$7F,$EF,$F7
	.dc.b	$FB,$EC,$9C,$E5,$FB,$C8,$C5,$71,$39,$7C,$07,$3F,$0F,$A7,$3C,$AF
	.dc.b	$F7,$DE,$1F,$EA,$7C,$39,$36,$3F,$83,$A9,$E1,$1E,$8A,$FE,$87,$C1
	.dc.b	$FE,$4F,$FA,$BF,$EF,$B3,$FC,$39,$AF,$FA,$AE,$36,$FF,$12,$CD,$8F
	.dc.b	$A6,$B6,$70,$DB,$FA,$D4,$7F,$3D,$BE,$CC,$BF,$E9,$F1,$E1,$E5,$91
	.dc.b	$D6,$7A,$47,$97,$6C,$CF,$F8,$D3,$0F,$F0,$33,$67,$4F,$13,$D5,$DB
	.dc.b	$ED,$E5,$91,$91,$5F,$E6,$C7,$0D,$7B,$47,$07,$C1,$3F,$E7,$5F,$AA
	.dc.b	$7D,$57,$F9,$DF,$EA,$BD,$F5,$FC,$D9,$AA,$76,$F8,$5A,$FF,$93,$A3
	.dc.b	$67,$08,$CE,$DF,$3A,$7F,$A0,$C6,$34,$FC,$45,$70,$D7,$C9,$87,$8F
	.dc.b	$9F,$C9,$CF,$31,$F8,$17,$FE,$0A,$6C,$E5,$F1,$4F,$17,$7D,$DE,$D2
	.dc.b	$3C,$B9,$EB,$E9,$8B,$A7,$9E,$18,$7F,$24,$FA,$87,$D4,$FF,$A1,$9B
	.dc.b	$3F,$E8,$13,$FA,$EB,$87,$0D,$7F,$D0,$CE,$6B,$B7,$D4,$FE,$B0,$FD
	.dc.b	$6C,$67,$6F,$D6,$36,$BF,$B4,$CE,$FA,$7F,$D5,$D3,$99,$4F,$EA,$5B
	.dc.b	$CF,$FE,$B3,$8C,$1F,$AB,$8F,$C5,$38,$8D,$1B,$E8,$E1,$DD,$2B,$6F
	.dc.b	$FA,$F7,$EA,$7F,$5B,$68,$D4,$FF,$A4,$23,$0F,$E5,$E7,$F6,$7F,$50
	.dc.b	$FA,$47,$C9,$7C,$9A,$4E,$63,$0B,$B5,$F5,$18,$BE,$37,$F4,$7E,$C5
	.dc.b	$F6,$3F,$66,$3F,$8A,$46,$38,$7B,$36,$38,$4E,$DF,$F0,$AF,$6F,$FB
	.dc.b	$26,$BC,$8D,$8F,$BC,$D7,$FD,$80,$D7,$F9,$17,$D0,$BE,$47,$F4,$1C
	.dc.b	$BF,$8B,$8C,$5F,$EF,$61,$F5,$2F,$BD,$DC,$F2,$BB,$87,$F8,$A3,$E6
	.dc.b	$3F,$6F,$3F,$C0,$5B,$1F,$EE,$3C,$39,$FC,$3F,$08,$4E,$A7,$F3,$9F
	.dc.b	$82,$3E,$D3,$4D,$C3,$C9,$E5,$BA,$B3,$5C,$F0,$FF,$50,$57,$87,$C8
	.dc.b	$F2,$E2,$FA,$BF,$B9,$39,$3F,$8B,$7E,$09,$F7,$9B,$D3,$C3,$EB,$3B
	.dc.b	$46,$0F,$E3,$9F,$7F,$FE,$7D,$6E,$BD,$BE,$B1,$F0,$9E,$DA,$D7,$0F
	.dc.b	$FB,$7B,$CD,$7C,$57,$C1,$6E,$3F,$3F,$EB,$1F,$25,$9A,$F8,$DF,$82
	.dc.b	$3E,$DA,$9E,$CF,$AE,$EF,$87,$F2,$F9,$D3,$FC,$C7,$E8,$5D,$5F,$3D
	.dc.b	$B1,$F7,$6B,$C3,$A7,$FB,$CA,$72,$FF,$BD,$2F,$F0,$6D,$EB,$F9,$7D
	.dc.b	$CF,$3F,$CB,$1F,$6F,$E9,$9F,$61,$4D,$ED,$74,$FE,$89,$F7,$DF,$E6
	.dc.b	$E6,$E5,$F5,$BE,$C1,$F5,$3F,$85,$AD,$4B,$B7,$F9,$07,$D7,$7D,$48
	.dc.b	$C6,$1F,$DF,$63,$13,$AF,$4B,$F1,$D1,$89,$FF,$01,$7D,$77,$BA,$BF
	.dc.b	$95,$36,$3F,$C0,$DF,$42,$F9,$CF,$CB,$4E,$71,$FF,$42,$FD,$73,$E6
	.dc.b	$FF,$34,$DC,$3F,$B0,$2F,$A1,$CC,$FE,$7F,$EB,$9F,$5D,$4F,$91,$C1
	.dc.b	$70,$FE,$D1,$C9,$F0,$9B,$1F,$A7,$FA,$EF,$F6,$AC,$60,$FE,$DC,$E3
	.dc.b	$3F,$86,$7F,$A6,$9C,$6F,$FD,$1F,$C6,$7D,$47,$3F,$81,$FD,$97,$D6
	.dc.b	$BE,$7F,$F4,$FE,$77,$C7,$CB,$FD,$47,$39,$CF,$F8,$24,$E6,$53,$97
	.dc.b	$F7,$D1,$98,$C6,$5F,$C3,$A3,$17,$DB,$D0,$F0,$B7,$0F,$F0,$45,$FC
	.dc.b	$97,$F8,$97,$D7,$3E,$67,$CF,$C9,$FB,$C8,$C4,$F0,$8C,$4E,$A3,$8B
	.dc.b	$EE,$EF,$FE,$07,$CB,$FA,$0D,$3F,$B5,$63,$F9,$E7,$F9,$1C,$66,$FF
	.dc.b	$C9,$13,$96,$D7,$C8,$D7,$A4,$7F,$87,$EF,$84,$7F,$8E,$7E,$6B,$F8
	.dc.b	$F4,$62,$3F,$35,$18,$EF,$95,$FE,$27,$9F,$F6,$8F,$89,$FE,$D2,$FA
	.dc.b	$C7,$7F,$F9,$67,$1E,$55,$A1,$3A,$7F,$C9,$18,$7F,$86,$E3,$11,$FE
	.dc.b	$61,$BE,$97,$CE,$B4,$FF,$81,$63,$33,$CB,$0A,$F1,$3B,$CF,$18,$CC
	.dc.b	$78,$EF,$BE,$1F,$C3,$1F,$0C,$66,$B8,$3C,$8F,$F3,$A0,$BE,$F5,$C6
	.dc.b	$7C,$93,$B3,$F8,$16,$1E,$87,$8D,$70,$8C,$47,$9E,$FC,$71,$88,$F9
	.dc.b	$07,$F7,$B9,$FE,$F8,$E0,$EE,$79,$3E,$B5,$F3,$2F,$AE,$FF,$01,$46
	.dc.b	$1B,$19,$9F,$F1,$74,$E5,$BF,$33,$FC,$65,$B9,$F2,$8C,$47,$76,$BF
	.dc.b	$E9,$79,$47,$1C,$BE,$DE,$7F,$DE,$6B,$EA,$B5,$F8,$D5,$C2,$FE,$CB
	.dc.b	$63,$E5,$4E,$EB,$9E,$DF,$EA,$57,$F0,$F5,$B6,$E1,$E1,$AF,$F5,$51
	.dc.b	$C4,$B5,$C5,$F4,$46,$EC,$2F,$A4,$F0,$8F,$4D,$71,$77,$3A,$6D,$73
	.dc.b	$7F,$47,$3B,$69,$EE,$8D,$57,$C3,$C3,$FC,$CC,$DD,$7C,$15,$FC,$98
	.dc.b	$F2,$57,$0B,$FD,$74,$EA,$F8,$C6,$2B,$C7,$1F,$A4,$FA,$EF,$EF,$71
	.dc.b	$FE,$9F,$6B,$D5,$B1,$FE,$77,$7E,$DC,$6F,$6E,$8A,$37,$7F,$ED,$76
	.dc.b	$C7,$CA,$78,$BE,$9F,$AD,$FE,$76,$FF,$AB,$DA,$E7,$AB,$AF,$D6,$BE
	.dc.b	$D3,$BD,$6B,$2F,$F4,$53,$71,$FF,$73,$37,$0F,$B3,$7C,$5E,$DE,$37
	.dc.b	$22,$3F,$EB,$F8,$7D,$7E,$23,$CD,$37,$0F,$F4,$C5,$6A,$79,$90,$D9
	.dc.b	$CE,$BE,$84,$BE,$3F,$FC,$EF,$3F,$66,$BF,$C4,$EF,$87,$D3,$DC,$EA
	.dc.b	$BB,$C7,$EF,$AB,$75,$FE,$A3,$6B,$FC,$49,$AF,$E6,$07,$D7,$FF,$7B
	.dc.b	$7F,$DB,$8D,$7F,$14,$DA,$EF,$1F,$EE,$1F,$A5,$7D,$7F,$F0,$E6,$C7
	.dc.b	$FA,$2D,$AF,$FA,$D5,$AF,$8F,$B3,$D0,$47,$FD,$25,$87,$F8,$45,$FF
	.dc.b	$4A,$7A,$AF,$AF,$FE,$36,$D8,$CC,$60,$F1,$7D,$7C,$8F,$13,$5F,$F6
	.dc.b	$37,$D7,$FF,$9B,$5F,$F5,$03,$5C,$F8,$BF,$CA,$23,$11,$FC,$EB,$EB
	.dc.b	$DF,$65,$FC,$B1,$B1,$FE,$A9,$6B,$FE,$D3,$6B,$FD,$0B,$47,$CC,$7C
	.dc.b	$72,$47,$C7,$7F,$9B,$FB,$DA,$8F,$F5,$A3,$5F,$F6,$F3,$5F,$E9,$AD
	.dc.b	$7E,$E7,$D7,$3E,$BF,$FA,$0B,$63,$FD,$7C,$D7,$FD,$D0,$D7,$CD,$E1
	.dc.b	$FD,$B2,$3F,$6D,$FE,$97,$F5,$FF,$E8,$28,$FF,$64,$35,$FF,$79,$35
	.dc.b	$FE,$C2,$D7,$F2,$1F,$5F,$FE,$33,$7F,$DA,$0D,$7F,$DF,$4D,$7F,$B4
	.dc.b	$B5,$FC,$C6,$4F,$23,$F2,$04,$8F,$C8,$3F,$D8,$3E,$BB,$FC,$59,$1F
	.dc.b	$ED,$D6,$BF,$DD,$03,$87,$FB,$BB,$5F,$D3,$7D,$77,$F6,$67,$FD,$CC
	.dc.b	$D7,$93,$5F,$EF,$ED,$7F,$5D,$F5,$BB,$E9,$78,$18,$D3,$F9,$2C,$E6
	.dc.b	$3F,$69,$E3,$3E,$BB,$35,$FD,$EA,$BA,$61,$FD,$41,$A9,$A7,$37,$FA
	.dc.b	$AD,$3E,$97,$F9,$9B,$63,$FA,$CE,$3E,$47,$A5,$EB,$8B,$7F,$29,$27
	.dc.b	$79,$F9,$A7,$51,$83,$BC,$ED,$FE,$A5,$82,$7F,$A1,$FD,$6B,$EB,$A9
	.dc.b	$2B,$90,$87,$D2,$E6,$3F,$AE,$47,$17,$6D,$73,$1F,$D4,$3E,$B1,$8D
	.dc.b	$24,$F0,$CE,$BF,$83,$9D,$BF,$85,$0E,$1D,$4F,$F8,$B2,$31,$A7,$C9
	.dc.b	$9F,$47,$47,$B8,$EB,$19,$C7,$F9,$F4,$FA,$3E,$1F,$E1,$9F,$13,$C6
	.dc.b	$91,$92,$1F,$01,$FE,$D7,$F5,$CF,$AB,$A4,$9F,$93,$1F,$E5,$6E,$1F
	.dc.b	$E7,$CE,$E3,$33,$B7,$FC,$0D,$1B,$7F,$C3,$79,$BE,$BB,$E5,$12,$3A
	.dc.b	$F9,$35,$F9,$11,$F0,$C8,$FC,$99,$E4,$F9,$3E,$97,$D2,$71,$D7,$E4
	.dc.b	$61,$FD,$6F,$2F,$EA,$9F,$5D,$F2,$E7,$E6,$36,$3D,$7F,$4D,$F0,$8F
	.dc.b	$F0,$6C,$67,$0F,$F1,$9B,$55,$F5,$73,$F3,$5C,$08,$FE,$42,$7F,$48
	.dc.b	$FF,$8A,$54,$6A,$30,$F8,$3A,$3F,$C3,$BF,$5D,$FD,$97,$1F,$ED,$8D
	.dc.b	$AF,$F2,$2B,$55,$BD,$0E,$3E,$F7,$F9,$DA,$BF,$B8,$57,$FA,$21,$AF
	.dc.b	$48,$D1,$F0,$39,$A4,$7E,$A2,$7B,$63,$FA,$3D,$7F,$28,$FE,$AA,$7F
	.dc.b	$5A,$D7,$FC,$8C,$EF,$2F,$F0,$23,$63,$FC,$BA,$EE,$79,$F4,$7F,$8F
	.dc.b	$23,$53,$FC,$37,$E0,$B0,$47,$BF,$45,$7B,$C5,$BF,$2C,$FF,$91,$BE
	.dc.b	$91,$F3,$34,$97,$E8,$6B,$FE,$5D,$8D,$BF,$C2,$1A,$FB,$23,$75,$FB
	.dc.b	$59,$CB,$63,$D4,$6B,$2F,$F1,$24,$E9,$B1,$AA,$EC,$ED,$FF,$40,$CE
	.dc.b	$DF,$8F,$19,$51,$96,$C7,$37,$D4,$EE,$35,$8E,$D9,$1C,$E2,$CF,$88
	.dc.b	$87,$BC,$B7,$ED,$6D,$FD,$CA,$EC,$73,$2C,$68,$F4,$94,$F3,$96,$BC
	.dc.b	$2F,$A0,$C4,$72,$70,$78,$CB,$5E,$36,$D7,$8D,$ED,$7D,$27,$F2,$F5
	.dc.b	$CD,$E0,$62,$77,$1E,$96,$9E,$22,$9D,$8F,$59,$4F,$59,$6F,$C1,$7F
	.dc.b	$7A,$39,$1B,$BC,$BC,$AF,$D4,$FF,$BF,$23,$F1,$ED,$BF,$F7,$7C,$62
	.dc.b	$3A,$1F,$E0,$F7,$0F,$E2,$23,$64,$7F,$6B,$D3,$FA,$D3,$E1,$5F,$ED
	.dc.b	$1C,$47,$BA,$FF,$7B,$18,$8F,$45,$59,$CC,$6E,$B9,$3D,$8C,$1F,$BE
	.dc.b	$AE,$11,$E8,$9F,$A2,$F0,$9F,$E0,$F3,$E4,$78,$1C,$CF,$09,$67,$C2
	.dc.b	$E0,$E2,$66,$7B,$11,$E6,$68,$E1,$1C,$89,$1B,$8F,$B7,$18,$78,$BE
	.dc.b	$87,$E7,$1D,$CC,$C7,$1B,$FD,$7D,$7C,$09,$D9,$F8,$57,$0F,$69,$F0
	.dc.b	$60,$9C,$D7,$E1,$AB,$FC,$5D,$18,$C3,$F7,$0D,$8F,$D5,$35,$EC,$79
	.dc.b	$4E,$C7,$8B,$0F,$63,$D0,$FA,$E5,$3E,$33,$EF,$78,$99,$78,$D7,$F1
	.dc.b	$1F,$AE,$FF,$01,$4A,$7F,$C3,$11,$86,$F4,$FF,$BE,$23,$FC,$09,$27
	.dc.b	$59,$7F,$B2,$5B,$03,$53,$C7,$DC,$E5,$E7,$73,$E8,$95,$FE,$3E,$BF
	.dc.b	$F5,$86,$3E,$2A,$CC,$F9,$71,$ED,$7C,$A3,$F9,$F7,$07,$F4,$72,$3C
	.dc.b	$FE,$2A,$BF,$AF,$8F,$D9,$91,$FF,$1E,$69,$FD,$6E,$FF,$DC,$5F,$5D
	.dc.b	$FE,$07,$3F,$C1,$11,$99,$F6,$47,$F8,$09,$FF,$2A,$B5,$F3,$1A,$E8
	.dc.b	$FF,$76,$67,$FE,$43,$9F,$F0,$5B,$63,$37,$E0,$6D,$71,$3A,$F0,$E1
	.dc.b	$5F,$F9,$A5,$F0,$F2,$E6,$3F,$CA,$37,$EA,$F0,$B0,$66,$65,$7A,$23
	.dc.b	$A7,$D7,$3E,$BA,$92,$7F,$CE,$3F,$5C,$FB,$5C,$47,$FA,$9E,$B8,$3F
	.dc.b	$E6,$56,$B9,$AF,$F3,$42,$8C,$EB,$FE,$65,$BF,$99,$26,$CE,$B2,$FC
	.dc.b	$54,$E7,$A7,$E1,$57,$FB,$86,$F9,$74,$E2,$65,$7A,$23,$66,$1F,$AA
	.dc.b	$77,$F0,$F9,$91,$F7,$F4,$D7,$78,$7F,$9D,$9B,$3E,$29,$FA,$9F,$5C
	.dc.b	$FA,$FA,$4A,$FF,$4C,$FD,$7B,$E0,$71,$1F,$EC,$FA,$E0,$FF,$A3,$5A
	.dc.b	$E6,$BE,$4C,$63,$4F,$F4,$65,$FE,$C8,$FD,$9C,$67,$5F,$8F,$1F,$AA
	.dc.b	$6F,$57,$E2,$63,$18,$7E,$D5,$75,$66,$64,$FA,$1E,$11,$C7,$C4,$93
	.dc.b	$EC,$7F,$29,$F5,$FF,$E6,$DB,$FF,$36,$49,$A6,$23,$FC,$E5,$F6,$B9
	.dc.b	$9F,$F5,$63,$5F,$F7,$93,$5F,$C0,$87,$0F,$E1,$3E,$C1,$F5,$EF,$93
	.dc.b	$FF,$5B,$46,$1B,$DC,$FB,$0C,$D6,$EF,$FE,$00,$7D,$83,$E2,$7F,$D7
	.dc.b	$11,$F9,$08,$C7,$07,$FB,$05,$B1,$C2,$76,$FF,$1F,$FA,$17,$C6,$E2
	.dc.b	$11,$8B,$FF,$65,$36,$3D,$A6,$A7,$87,$CB,$7E,$CE,$BF,$DA,$0D,$9D
	.dc.b	$B5,$F8,$AD,$7F,$4F,$F5,$AF,$AD,$F1,$36,$BF,$DB,$4D,$9F,$0B,$5E
	.dc.b	$7E,$37,$88,$93,$E2,$7D,$9F,$59,$F3,$E3,$FD,$C8,$D9,$F5,$B5,$FB
	.dc.b	$2D,$7F,$77,$F5,$3F,$E7,$C9,$FF,$76,$36,$3E,$1B,$5F,$C0,$B5,$FE
	.dc.b	$07,$E2,$BE,$4B,$E7,$36,$BF,$DE,$CD,$9F,$9A,$D7,$F1,$7F,$4C,$FA
	.dc.b	$8F,$AC,$DA,$FF,$7F,$36,$7E,$AB,$5F,$C9,$FC,$C3,$E3,$BE,$F3,$6B
	.dc.b	$ED,$CA,$30,$FD,$E6,$BE,$2E,$0F,$09,$27,$C2,$FE,$27,$E7,$DF,$4F
	.dc.b	$49,$3D,$A3,$2F,$83,$E9,$BF,$B8,$36,$FF,$25,$18,$7F,$1E,$D7,$F5
	.dc.b	$BF,$4C,$F0,$BF,$30,$DA,$F5,$B6,$7F,$2E,$D7,$F6,$BF,$4B,$EC,$7F
	.dc.b	$9B,$B5,$FF,$01,$35,$37,$E9,$5F,$2D,$FA,$76,$DF,$CA,$6D,$7E,$99
	.dc.b	$AF,$F0,$4F,$A5,$7A,$FF,$AF,$6D,$7D,$26,$CF,$EB,$9A,$FF,$0C,$D1
	.dc.b	$F4,$BF,$3C,$95,$F3,$5F,$DD,$7D,$2B,$D2,$A4,$9F,$BF,$19,$7F,$03
	.dc.b	$F4,$4F,$A5,$47,$F8,$8F,$CE,$B9,$7F,$8B,$FC,$0F,$99,$18,$85,$F1
	.dc.b	$AE,$07,$1A,$DB,$FD,$92,$35,$3F,$E1,$0A,$56,$A3,$FC,$65,$3A,$9F
	.dc.b	$09,$AE,$2B,$7C,$F0,$B7,$E5,$9A,$F9,$23,$35,$C6,$BC,$B3,$DB,$E0
	.dc.b	$73,$F4,$AF,$A5,$E3,$92,$73,$18,$6D,$F2,$BF,$83,$18,$7F,$BB,$D7
	.dc.b	$F2,$7C,$3D,$3C,$7F,$BD,$7E,$44,$66,$FE,$74,$E5,$F8,$DD,$1E,$23
	.dc.b	$C7,$76,$3C,$17,$B8,$FA,$DF,$25,$F0,$23,$11,$FE,$65,$8C,$77,$CA
	.dc.b	$F4,$4F,$F9,$7F,$AF,$FC,$C1,$F4,$8F,$82,$FF,$20,$DF,$CA,$AD,$09
	.dc.b	$D3,$F8,$58,$FC,$2C,$62,$3F,$C4,$35,$D2,$B9,$DD,$9F,$E2,$67,$38
	.dc.b	$CB,$4C,$D7,$3C,$7F,$92,$3D,$63,$C6,$E7,$0F,$0D,$7F,$17,$9E,$47
	.dc.b	$FA,$16,$32,$F4,$8E,$F3,$CE,$3F,$90,$1C,$2B,$BC,$F3,$9E,$D7,$E9
	.dc.b	$AF,$69,$BB,$EA,$E0,$FE,$57,$87,$F9,$29,$EA,$60,$FE,$D3,$F4,$8F
	.dc.b	$8E,$7D,$2F,$F3,$28,$C3,$63,$33,$C1,$AF,$F8,$3A,$5F,$23,$FE,$A3
	.dc.b	$8E,$AD,$74,$D7,$FC,$AB,$1C,$34,$F0,$4F,$1C,$FE,$D6,$9F,$E0,$4B
	.dc.b	$F8,$0D,$7E,$6D,$F1,$6D,$71,$9F,$F3,$09,$BA,$E7,$B7,$F9,$91,$FC
	.dc.b	$35,$F1,$6D,$71,$BF,$F5,$95,$F1,$39,$DF,$17,$D3,$3B,$95,$F3,$9D
	.dc.b	$C7,$A2,$B8,$3F,$EB,$D9,$D3,$6B,$95,$77,$8F,$0D,$70,$7A,$CE,$A3
	.dc.b	$FD,$39,$3F,$0F,$0F,$E4,$4D,$D7,$CB,$1E,$1B,$FE,$0D,$39,$7A,$9F
	.dc.b	$C2,$70,$F1,$AF,$A5,$FF,$03,$3F,$C3,$5A,$FF,$A0,$9B,$5F,$CC,$5F
	.dc.b	$C0,$C6,$F6,$FF,$58,$46,$B0,$FF,$49,$B6,$7F,$DB,$CF,$07,$FD,$1B
	.dc.b	$F4,$BF,$E4,$B7,$F8,$E3,$5C,$F9,$DF,$EB,$BF,$A2,$FF,$42,$46,$A3
	.dc.b	$FD,$7D,$3B,$C7,$FA,$23,$70,$FF,$51,$DF,$0B,$EF,$F1,$BC,$08,$FF
	.dc.b	$2B,$87,$D2,$E2,$36,$79,$27,$C5,$19,$C3,$BC,$5B,$EF,$3F,$ED,$49
	.dc.b	$7C,$0F,$F9,$FB,$07,$C9,$FA,$CD,$46,$5C,$63,$E3,$73,$EE,$E2,$3F
	.dc.b	$D6,$7F,$4B,$FD,$15,$FE,$84,$D7,$FD,$6F,$7D,$9F,$E8,$DF,$36,$FA
	.dc.b	$5F,$F7,$23,$63,$31,$E7,$FB,$2F,$E4,$1A,$FC,$8E,$CF,$A0,$4A,$D9
	.dc.b	$FD,$47,$CF,$7D,$37,$FB,$BD,$B1,$98,$E0,$FF,$B2,$B1,$7D,$37,$88
	.dc.b	$91,$FD,$93,$81,$93,$E4,$FF,$DF,$5F,$4D,$FE,$B2,$A3,$4F,$F5,$AD
	.dc.b	$7C,$EF,$3B,$FE,$00,$7D,$2B,$E9,$A9,$2F,$FB,$47,$5B,$E7,$3E,$1F
	.dc.b	$C7,$C9,$F8,$EF,$C3,$F8,$DF,$F7,$5E,$5F,$E0,$FF,$6F,$FD,$D5,$C9
	.dc.b	$FE,$6C,$AC,$B6,$FD,$5D,$5F,$DE,$DF,$EE,$DF,$4A,$FA,$6D,$C0,$46
	.dc.b	$5D,$C6,$97,$D3,$3E,$57,$EC,$F9,$32,$7E,$EB,$F6,$7E,$F3,$85,$3E
	.dc.b	$B2,$92,$BC,$BF,$27,$D7,$EA,$3F,$C6,$11,$FE,$0B,$6B,$FF,$00,$1B
	.dc.b	$C1,$FE,$2F,$FA,$8F,$F6,$6B,$FE,$31,$EC,$7A,$3F,$8C,$25,$FD,$BE
	.dc.b	$CF,$E2,$0A,$FE,$22,$BE,$9E,$A7,$BE,$73,$1F,$E3,$0E,$27,$D6,$FC
	.dc.b	$3F,$A9,$FE,$B1,$87,$F8,$85,$AF,$3E,$5F,$98,$A3,$37,$E9,$ED,$F9
	.dc.b	$9F,$51,$8B,$DB,$6F,$EA,$35,$FE,$BB,$F5,$3C,$39,$64,$7F,$8C,$67
	.dc.b	$5E,$9A,$B3,$3E,$33,$9D,$63,$1F,$F2,$C5,$6E,$7F,$C7,$4B,$AF,$F0
	.dc.b	$0B,$E7,$A9,$88,$93,$C2,$32,$F2,$8F,$C4,$7D,$43,$EA,$69,$27,$FC
	.dc.b	$91,$0F,$A3,$C4,$79,$C4,$71,$77,$3D,$27,$F2,$9F,$53,$E5,$9F,$F2
	.dc.b	$93,$5E,$4D,$CB,$FC,$ED,$7F,$2E,$31,$B7,$D6,$6D,$7E,$8D,$AF,$8B
	.dc.b	$4E,$F9,$C8,$EE,$FB,$A3,$DF,$C5,$F0,$3E,$7E,$57,$BA,$32,$FC,$08
	.dc.b	$CB,$EF,$C7,$F9,$AB,$FE,$6B,$FA,$9F,$E2,$0D,$F2,$7F,$65,$18,$8F
	.dc.b	$8C,$FF,$26,$7F,$74,$E7,$4F,$F4,$A3,$74,$FC,$9B,$67,$7E,$CF,$CA
	.dc.b	$F1,$A4,$7C,$A7,$BF,$A7,$88,$F9,$71,$70,$EB,$37,$0F,$95,$36,$BC
	.dc.b	$73,$E8,$F4,$FF,$85,$7D,$4F,$FA,$9F,$0F,$F5,$43,$70,$F2,$CE,$EF
	.dc.b	$FD,$19,$39,$7F,$8B,$35,$64,$9E,$BF,$27,$27,$8B,$E0,$FA,$6C,$47
	.dc.b	$E3,$6D,$38,$C7,$C7,$5F,$C7,$FE,$A7,$FA,$D4,$7F,$AF,$5B,$97,$EF
	.dc.b	$1B,$87,$93,$D9,$E1,$F0,$72,$3F,$D4,$93,$A5,$92,$B5,$1F,$9F,$9D
	.dc.b	$1B,$9D,$47,$FA,$B7,$1F,$D1,$3F,$A5,$7F,$50,$E7,$9F,$F9,$B4,$6B
	.dc.b	$2F,$F6,$A3,$72,$FE,$26,$F0,$AE,$55,$C3,$A3,$8F,$C4,$C9,$F0,$3D
	.dc.b	$3C,$EC,$47,$86,$2E,$3D,$A7,$FA,$37,$CE,$3E,$A6,$92,$FD,$4D,$7F
	.dc.b	$7D,$1B,$7F,$D9,$4D,$7D,$D1,$BA,$FF,$77,$4E,$5B,$1F,$00,$D6,$7F
	.dc.b	$CC,$27,$4D,$8D,$57,$67,$6F,$F6,$19,$DB,$FD,$8F,$3A,$A2,$92,$79
	.dc.b	$C7,$A9,$E6,$ED,$CD,$F9,$A3,$C0,$43,$DA,$57,$4F,$33,$E3,$99,$BF
	.dc.b	$53,$4F,$34,$F8,$2B,$C9,$1C,$0E,$44,$AE,$A7,$E1,$27,$81,$E8,$BE
	.dc.b	$33,$BB,$EE,$F9,$8C,$1F,$1E,$78,$1C,$9F,$20,$9E,$75,$E0,$AE,$6F
	.dc.b	$86,$77,$7A,$9F,$CA,$DF,$39,$EE,$73,$71,$5C,$E3,$E0,$9C,$A3,$C6
	.dc.b	$D3,$C7,$7D,$5F,$41,$88,$E7,$5E,$43,$C3,$1C,$4F,$0C,$F1,$7E,$E1
	.dc.b	$C4,$D5,$78,$DF,$31,$89,$E8,$F9,$8F,$5D,$F3,$AE,$8D,$BF,$1C,$F9
	.dc.b	$8F,$A1,$7D,$67,$51,$D4,$D4,$7C,$B7,$AC,$FC,$26,$9E,$72,$9E,$C7
	.dc.b	$BB,$EC,$3B,$DF,$A1,$C4,$FB,$6F,$CF,$5D,$63,$C8,$DA,$E1,$7E,$6B
	.dc.b	$EF,$5C,$6B,$BC,$FE,$B2,$F9,$D7,$96,$B9,$B6,$F9,$DF,$58,$F7,$C6
	.dc.b	$1F,$CF,$4F,$07,$D8,$7C,$53,$D3,$1F,$71,$F1,$BE,$DA,$FD,$EF,$D4
	.dc.b	$FF,$63,$93,$49,$CD,$FE,$61,$71,$CF,$4C,$46,$7E,$2B,$5F,$49,$2F
	.dc.b	$FC,$D9,$93,$2F,$F2,$1E,$3F,$E8,$92,$56,$EB,$FC,$41,$1F,$E8,$DD
	.dc.b	$3F,$9B,$5F,$F2,$0E,$0F,$E5,$A7,$8F,$C4,$B7,$FC,$A2,$2B,$CA,$2C
	.dc.b	$F1,$8D,$1E,$4C,$7C,$73,$FE,$7D,$FA,$9F,$EC,$73,$93,$FD,$08,$E1
	.dc.b	$FF,$1A,$36,$BD,$91,$AD,$B9,$73,$C8,$E4,$2B,$86,$1F,$E6,$47,$F8
	.dc.b	$34,$69,$FF,$28,$37,$C0,$F2,$1E,$5F,$0A,$46,$48,$79,$1F,$F4,$B7
	.dc.b	$D4,$3E,$A6,$92,$7C,$7F,$53,$FE,$99,$7C,$B3,$B6,$CE,$DF,$F2,$94
	.dc.b	$ED,$FF,$2A,$4E,$DB,$3B,$D3,$E1,$C2,$B8,$46,$17,$3F,$FA,$AD,$FF
	.dc.b	$5C,$CF,$FA,$C5,$AF,$18,$D1,$BE,$79,$1E,$A8,$E5,$E1,$C8,$F6,$BD
	.dc.b	$7C,$3C,$47,$AA,$2E,$1E,$39,$FF,$5E,$7D,$4B,$D2,$A4,$BF,$B3,$F5
	.dc.b	$2F,$6B,$11,$FE,$75,$AE,$0F,$DE,$6B,$FE,$79,$9E,$0E,$6B,$8E,$1F
	.dc.b	$1A,$B6,$FC,$8A,$DD,$71,$EA,$FF,$68,$47,$FB,$BC,$9D,$58,$89,$3E
	.dc.b	$67,$71,$C3,$C3,$93,$EA,$7F,$DB,$5F,$53,$FE,$49,$BF,$F2,$44,$E4
	.dc.b	$87,$FA,$46,$3F,$C9,$DF,$5B,$FA,$18,$FF,$5D,$87,$0F,$FA,$F9,$AF
	.dc.b	$FB,$AD,$AB,$EA,$9E,$37,$28,$FF,$62,$C6,$1C,$1B,$8A,$D3,$3F,$54
	.dc.b	$FA,$87,$D5,$61,$B3,$D6,$7F,$C3,$1F,$55,$9A,$EB,$F5,$2F,$93,$FD
	.dc.b	$CF,$0C,$FB,$17,$C9,$63,$C1,$FF,$6C,$DE,$DF,$E0,$94,$AD,$C7,$EF
	.dc.b	$BE,$6F,$F6,$0E,$A3,$FD,$A4,$E7,$BB,$A6,$BF,$EE,$F9,$CB,$FE,$BF
	.dc.b	$FA,$97,$D5,$6B,$1F,$F2,$57,$3F,$0A,$56,$B1,$F1,$F9,$AA,$D6,$7F
	.dc.b	$E7,$C9,$F3,$FD,$83,$E4,$33,$E4,$75,$8C,$4F,$F1,$6F,$AB,$F9,$6F
	.dc.b	$BB,$EC,$3A,$39,$E4,$F2,$46,$35,$FF,$6B,$7D,$67,$F9,$7E,$B8,$47
	.dc.b	$9E,$30,$FF,$BC,$77,$52,$3D,$9E,$4B,$E3,$BE,$07,$6B,$C1,$CC,$61
	.dc.b	$F8,$5F,$56,$FA,$CD,$E1,$86,$E9,$FC,$C3,$E8,$3D,$B3,$E1,$FB,$07
	.dc.b	$C5,$7C,$B9,$CE,$DF,$82,$9C,$DF,$F3,$8C,$7F,$01,$87,$A7,$B1,$F1
	.dc.b	$BE,$2E,$FF,$AD,$18,$BF,$E7,$FF,$4E,$FA,$DC,$EB,$F6,$63,$13,$F8
	.dc.b	$AA,$EF,$8F,$FB,$FF,$EB,$7E,$24,$FE,$3B,$EC,$3F,$D4,$CF,$F4,$A8
	.dc.b	$CC,$FC,$B9,$CA,$DB,$F9,$8F,$D7,$7C,$9B,$FC,$BF,$E2,$FF,$35,$FE
	.dc.b	$A7,$19,$AF,$A9,$39,$79,$4F,$D5,$5F,$5F,$F1,$A3,$14,$F7,$FF,$29
	.dc.b	$39,$AF,$B9,$39,$AF,$EC,$3F,$5E,$FA,$0F,$CD,$7C,$33,$EC,$7F,$39
	.dc.b	$39,$AF,$EC,$F3,$97,$FB,$4F,$D8,$3E,$07,$F4,$9A,$FE,$5B,$3F,$39
	.dc.b	$F8,$B9,$D5,$7C,$0F,$07,$E8,$7D,$87,$EA,$67,$35,$C0,$72,$27,$58
	.dc.b	$FE,$BF,$EC,$3F,$AB,$E3,$F7,$27,$1E,$EF,$48,$FE,$FB,$F5,$CF,$B0
	.dc.b	$FD,$C7,$97,$F3,$79,$BF,$B1,$CE,$6B,$F8,$4C,$E5,$F9,$D1,$9D,$3F
	.dc.b	$4D,$39,$7F,$4D,$19,$AF,$E2,$F3,$99,$4E,$6B,$FC,$5F,$39,$8C,$4F
	.dc.b	$E1,$63,$1E,$B7,$F0,$DF,$B0,$7C,$D7,$F0,$DB,$FE,$6B,$87,$F1,$E9
	.dc.b	$CB,$EE,$AE,$15,$B6,$C7,$C2,$EF,$FE,$31,$CF,$F8,$BC,$FE,$0C,$63
	.dc.b	$C6,$EC,$CA,$F9,$3A,$53,$EA,$57,$18,$FC,$93,$FA,$63,$FC,$1A,$66
	.dc.b	$3F,$C9,$BD,$FF,$C9,$63,$11,$FA,$3B,$FE,$50,$FF,$0F,$F1,$3F,$88
	.dc.b	$7D,$7F,$E6,$63,$FC,$AB,$3A,$53,$AE,$EF,$0C,$FF,$8F,$AB,$85,$7E
	.dc.b	$96,$B4,$F2,$C6,$B9,$13,$C3,$0F,$D7,$3C,$A3,$C8,$6B,$0E,$F5,$CA
	.dc.b	$33,$3C,$70,$F1,$57,$F6,$9C,$3B,$E1,$D2,$31,$7C,$CF,$21,$4F,$25
	.dc.b	$F3,$C3,$B4,$7C,$03,$55,$CD,$FE,$FF,$7D,$B0,$F0,$C6,$0F,$EE,$B1
	.dc.b	$FD,$D9,$E4,$7F,$7A,$7F,$BD,$FD,$7B,$E5,$5F,$61,$86,$D7,$F8,$0E
	.dc.b	$75,$39,$38,$FC,$C7,$F1,$D9,$D3,$5F,$F0,$9C,$A3,$11,$FC,$66,$31
	.dc.b	$8F,$23,$EA,$E9,$FE,$01,$9B,$B7,$F1,$FD,$7E,$ED,$F0,$9F,$90,$D8
	.dc.b	$FC,$0D,$71,$9F,$98,$D8,$E3,$87,$9D,$B8,$76,$A8,$74,$9D,$57,$9D
	.dc.b	$B1,$FD,$96,$3A,$1F,$CC,$CA,$FF,$69,$BF,$24,$71,$AE,$B7,$E6,$C3
	.dc.b	$D9,$39,$BF,$3C,$7E,$B3,$5E,$8D,$8F,$03,$56,$1D,$9C,$9F,$E9,$CA
	.dc.b	$F8,$58,$7F,$A0,$1B,$7D,$F0,$FD,$3F,$D8,$7F,$7D,$BF,$F5,$1B,$5F
	.dc.b	$E6,$4D,$BF,$F4,$4B,$F6,$E3,$7B,$78,$E3,$57,$FD,$55,$B3,$FE,$0B
	.dc.b	$78,$3F,$CF,$7E,$19,$F1,$FE,$DC,$1F,$3E,$FA,$FC,$36,$FF,$D6,$F7
	.dc.b	$C1,$B7,$AA,$91,$83,$7E,$E7,$F9,$5A,$BF,$C7,$31,$FE,$5A,$6B,$FE
	.dc.b	$5F,$6B,$FE,$6A,$73,$87,$F9,$4B,$0F,$F0,$E4,$FF,$93,$9A,$FA,$EB
	.dc.b	$EA,$57,$0C,$BF,$DD,$B5,$AB,$E5,$E8,$6A,$FE,$D7,$EE,$1F,$DA,$71
	.dc.b	$FF,$36,$B5,$FF,$39,$35,$FF,$3E,$B5,$FF,$01,$3B,$AF,$F5,$EB,$FE
	.dc.b	$98,$C3,$3F,$13,$DB,$C2,$EE,$FF,$9E,$27,$FD,$47,$7D,$23,$06,$7E
	.dc.b	$82,$47,$91,$AF,$93,$AB,$91,$1F,$4B,$D7,$C0,$91,$EC,$6B,$EC,$F7
	.dc.b	$7C,$44,$7E,$0B,$E5,$F5,$64,$7C,$46,$BF,$13,$E6,$BD,$44,$7E,$4B
	.dc.b	$ED,$FA,$99,$1F,$31,$AF,$CC,$F0,$3E,$11,$1F,$A2,$FC,$5F,$8A,$91
	.dc.b	$F5,$1A,$FD,$4E,$5F,$94,$47,$EC,$BF,$37,$8E,$47,$DC,$6B,$F7,$3A
	.dc.b	$BE,$91,$1F,$C0,$BF,$EA,$CF,$A4,$7D,$35,$24,$FF,$AC,$A3,$2F,$FA
	.dc.b	$D3,$E8,$DF,$73,$FE,$B4,$6C,$66,$30,$7F,$AE,$3E,$95,$C7,$49,$3F
	.dc.b	$EB,$E8,$CB,$D7,$C4,$FE,$21,$7F,$D1,$E3,$FD,$0D,$D2,$E9,$FF,$2A
	.dc.b	$4E,$67,$FC,$75,$7F,$95,$9C,$DF,$94,$93,$D4,$87,$0F,$90,$FE,$3D
	.dc.b	$5F,$EE,$28,$FF,$22,$5F,$5C,$3F,$D8,$BF,$76,$F8,$6F,$E1,$4B,$E6
	.dc.b	$A9,$88,$97,$FE,$C6,$9D,$36,$72,$FF,$A8,$FE,$F7,$C1,$7F,$ED,$26
	.dc.b	$E9,$F0,$1B,$3F,$E7,$DB,$FF,$10,$CE,$73,$FE,$39,$5A,$B3,$6B,$77
	.dc.b	$FE,$AF,$C9,$F3,$9F,$38,$91,$F3,$B1,$FF,$40,$7E,$0A,$FA,$4C,$CF
	.dc.b	$FB,$DE,$F8,$BF,$83,$BF,$C1,$11,$E4,$7F,$AF,$6B,$F9,$53,$FE,$6F
	.dc.b	$6E,$5F,$6E,$3C,$93,$EA,$EB,$F5,$F6,$49,$FF,$77,$E9,$FE,$C3,$6C
	.dc.b	$E5,$F9,$53,$C5,$DF,$D0,$7C,$42,$3C,$B9,$EB,$F4,$62,$E3,$F4,$67
	.dc.b	$FB,$07,$E1,$2F,$B8,$A4,$BF,$68,$D1,$F5,$39,$8F,$F5,$41,$FE,$DD
	.dc.b	$AE,$4E,$0C,$55,$69,$63,$6F,$C7,$E4,$FE,$E5,$7F,$91,$8F,$0C,$EB
	.dc.b	$1F,$94,$3C,$3F,$EC,$6D,$F2,$FF,$B0,$46,$0F,$F7,$0D,$7D,$69,$D3
	.dc.b	$63,$53,$C3,$D2,$FE,$D1,$F8,$6F,$D8,$E5,$FD,$9A,$F9,$9C,$9A,$F3
	.dc.b	$66,$3D,$67,$4D,$9C,$BF,$EB,$1C,$DF,$7B,$F4,$09,$1F,$3B,$57,$CB
	.dc.b	$E2,$3E,$55,$70,$7E,$45,$7C,$B5,$19,$77,$1F,$2B,$EF,$64,$7C,$A8
	.dc.b	$FF,$02,$7E,$15,$F2,$09,$B6,$36,$CB,$2B,$E4,$4E,$E7,$FC,$19,$F1
	.dc.b	$9F,$16,$7F,$DE,$4F,$C0,$7F,$DA,$BC,$28,$C3,$F6,$1E,$0E,$77,$E7
	.dc.b	$B4,$AF,$AD,$5D,$5F,$6F,$27,$E5,$88,$FC,$7F,$85,$AF,$CD,$8B,$87
	.dc.b	$EB,$E3,$F3,$13,$8E,$2F,$F7,$F6,$FF,$9A,$9F,$C6,$DF,$E4,$2B,$51
	.dc.b	$F9,$B8,$AF,$23,$64,$3C,$07,$8C,$C5,$F8,$6B,$B4,$68,$FE,$09,$5D
	.dc.b	$23,$33,$BA,$F1,$57,$2B,$F3,$C7,$8B,$85,$87,$18,$C3,$6B,$81,$FC
	.dc.b	$4D,$C3,$E1,$BF,$89,$5C,$A3,$67,$CE,$2B,$97,$85,$F8,$9C,$D7,$4A
	.dc.b	$F4,$47,$C1,$36,$7C,$93,$67,$6C,$3D,$F5,$CE,$3E,$19,$9C,$3D,$6D
	.dc.b	$38,$9D,$0F,$31,$70,$F3,$5F,$8A,$35,$7D,$9F,$84,$73,$38,$61,$E9
	.dc.b	$C3,$C5,$5F,$12,$31,$3E,$93,$CA,$74,$C3,$CA,$FD,$B3,$F9,$0D,$FD
	.dc.b	$E6,$C7,$ED,$63,$11,$EF,$32,$7F,$21,$8D,$11,$FE,$05,$F4,$BF,$C8
	.dc.b	$8F,$1B,$FC,$C5,$C3,$EB,$BF,$B7,$18,$8F,$31,$C8,$86,$0F,$EA,$8E
	.dc.b	$E3,$E4,$13,$0F,$EC,$71,$97,$E2,$92,$7F,$AF,$46,$5F,$8D,$3C,$5C
	.dc.b	$9E,$07,$0F,$D8,$7C,$EE,$67,$A3,$97,$C2,$E4,$EA,$EB,$15,$78,$67
	.dc.b	$81,$A3,$C2,$7B,$8F,$8D,$3D,$1C,$1F,$0E,$BC,$6E,$1F,$3E,$1F,$89
	.dc.b	$8F,$A0,$78,$CE,$E4,$BF,$1B,$98,$E0,$71,$8C,$C7,$5A,$F7,$CE,$6F
	.dc.b	$D4,$F0,$C3,$D9,$7C,$30,$F1,$C6,$2B,$F0,$73,$9B,$F6,$56,$B0,$F3
	.dc.b	$4E,$5F,$88,$79,$EB,$BD,$FF,$0B,$6C,$7F,$63,$F8,$8F,$ED,$24,$7F
	.dc.b	$B3,$7D,$95,$7F,$9E,$F3,$27,$18,$FF,$74,$85,$7F,$6A,$F5,$3F,$B5
	.dc.b	$F8,$33,$0F,$EC,$D1,$97,$FB,$44,$65,$FE,$93,$D1,$FE,$BF,$7F,$C5
	.dc.b	$9F,$27,$FE,$04,$23,$FD,$EF,$F0,$E7,$7E,$23,$FC,$21,$BA,$BF,$B3
	.dc.b	$36,$3F,$C2,$4D,$77,$1F,$EE,$7A,$77,$53,$A6,$BF,$E1,$0E,$8F,$81
	.dc.b	$7F,$D9,$37,$FE,$D6,$7F,$6C,$DA,$C6,$48,$7F,$64,$9F,$8A,$FF,$5E
	.dc.b	$DB,$FB,$5E,$D2,$30,$43,$FB,$5F,$2F,$F9,$55,$FE,$CB,$1F,$DC,$A5
	.dc.b	$1F,$E3,$6F,$8A,$FF,$6B,$38,$6A,$73,$93,$F0,$9F,$93,$3C,$78,$BF
	.dc.b	$C1,$95,$FE,$DA,$8E,$A7,$18,$C9,$1F,$F1,$64,$FF,$90,$AB,$FC,$6A
	.dc.b	$DE,$5E,$37,$E0,$B7,$D5,$FF,$8D,$89,$87,$F8,$CF,$C5,$AF,$E3,$A2
	.dc.b	$E3,$F8,$D9,$FA,$DE,$47,$F8,$DB,$19,$7F,$83,$8C,$BF,$E3,$38,$CB
	.dc.b	$FE,$14,$E2,$FF,$46,$AF,$C3,$9F,$01,$EE,$C7,$FC,$11,$18,$9F,$F3
	.dc.b	$0B,$5E,$5E,$3E,$F0,$F8,$B3,$99,$FF,$31,$B5,$CD,$FF,$8D,$E3,$11
	.dc.b	$F6,$1E,$97,$FE,$5C,$C1,$E3,$FE,$4C,$ED,$5B,$BF,$F1,$CE,$92,$3F
	.dc.b	$C6,$AF,$F8,$F3,$4C,$47,$E3,$E2,$E1,$F9,$99,$E6,$FE,$07,$4F,$F1
	.dc.b	$DE,$72,$FE,$DC,$6B,$7F,$F5,$77,$17,$FA,$B2,$FF,$C7,$97,$9A,$F2
	.dc.b	$DF,$F3,$4E,$5B,$D7,$FA,$68,$5F,$F8,$CE,$30,$FE,$1B,$83,$75,$FE
	.dc.b	$5B,$9E,$15,$FE,$8C,$6B,$F9,$26,$BF,$A0,$79,$E1,$FE,$E8,$FC,$09
	.dc.b	$F2,$BE,$E6,$D7,$FA,$6A,$B6,$FE,$A3,$E8,$FF,$CF,$9F,$21,$FE,$95
	.dc.b	$6C,$7F,$A6,$57,$88,$FA,$3D,$78,$AB,$FF,$4E,$7E,$18,$F8,$AF,$F2
	.dc.b	$86,$3F,$A8,$8C,$6B,$A9,$CC,$FE,$E2,$72,$FF,$A9,$FF,$11,$7A,$FA
	.dc.b	$84,$62,$FF,$D5,$4D,$8E,$13,$BC,$9E,$F5,$25,$FF,$AC,$63,$2F,$FA
	.dc.b	$CF,$E7,$BE,$3B,$78,$7C,$11,$8C,$3F,$D7,$4D,$7F,$D7,$FF,$1A,$FC
	.dc.b	$0F,$CC,$DB,$FF,$5F,$36,$7F,$D8,$4D,$7F,$D8,$FF,$90,$73,$7F,$86
	.dc.b	$B5,$FE,$36,$D4,$D1,$F1,$3F,$15,$B5,$FE,$CD,$6C,$FF,$B4,$5A,$F8
	.dc.b	$3E,$B3,$81,$27,$83,$F3,$FE,$19,$D3,$49,$3F,$ED,$98,$CB,$7E,$1B
	.dc.b	$81,$1C,$1B,$DF,$A1,$1C,$1F,$EE,$3F,$C6,$5E,$47,$DF,$6C,$66,$30
	.dc.b	$65,$87,$12,$E5,$FE,$EF,$FC,$85,$DF,$F8,$96,$CF,$FB,$C5,$B3,$B9
	.dc.b	$FC,$77,$E2,$CF,$A7,$A4,$BF,$F7,$C4,$65,$BD,$7C,$08,$E0,$DF,$67
	.dc.b	$42,$38,$3F,$A5,$E1,$FD,$33,$E6,$1E,$C7,$E7,$DB,$3A,$9F,$EB,$B1
	.dc.b	$87,$0D,$79,$B9,$F0,$FF,$D8,$0D,$9F,$EC,$B0,$F9,$7F,$82,$EA,$D3
	.dc.b	$A3,$C8,$E2,$FF,$8C,$C6,$0D,$3F,$C7,$2B,$9C,$66,$36,$7F,$74,$FC
	.dc.b	$93,$FA,$6C,$7A,$E1,$18,$7D,$CD,$7E,$2B,$5F,$EE,$2E,$EF,$FD,$B8
	.dc.b	$D9,$F6,$D7,$F8,$09,$F2,$D7,$6F,$2B,$D5,$7F,$EE,$3F,$98,$E1,$3F
	.dc.b	$CD,$3E,$3D,$EE,$EB,$0F,$EF,$F5,$AB,$56,$A7,$AC,$F2,$B3,$4F,$13
	.dc.b	$F0,$8C,$E1,$D5,$FA,$78,$57,$FA,$2E,$77,$1E,$D1,$C5,$FB,$4F,$0C
	.dc.b	$3D,$B1,$A3,$F9,$5F,$CC,$7F,$88,$BE,$57,$FC,$73,$5B,$3C,$47,$F8
	.dc.b	$9A,$35,$EC,$7F,$7F,$C7,$EF,$57,$0A,$FF,$0F,$D6,$9E,$51,$97,$90
	.dc.b	$78,$5F,$A1,$D4,$27,$93,$FD,$8A,$BB,$DF,$E7,$67,$37,$C6,$31,$5D
	.dc.b	$5F,$36,$1C,$A3,$C3,$87,$83,$0E,$71,$FA,$CC,$39,$C6,$23,$D6,$73
	.dc.b	$3D,$77,$DA,$B9,$5F,$59,$F4,$1F,$97,$29,$F9,$9B,$C3,$83,$FC,$8B
	.dc.b	$F2,$CF,$8B,$7C,$C7,$E7,$E3,$0D,$8C,$CF,$EF,$67,$2D,$F9,$8F,$F0
	.dc.b	$3B,$74,$E5,$18,$8F,$E1,$32,$8C,$47,$89,$E3,$97,$DB,$CF,$FC,$70
	.dc.b	$DD,$BF,$CB,$6F,$DF,$BE,$58,$7D,$66,$E1,$F1,$CF,$F4,$1D,$70,$BF
	.dc.b	$F3,$19,$D0,$F0,$57,$1C,$3C,$31,$A9,$E1,$5F,$E7,$68,$E0,$FF,$9E
	.dc.b	$2F,$BE,$1E,$3B,$E3,$3B,$BE,$2F,$88,$ED,$84,$7C,$D8,$74,$74,$7A
	.dc.b	$70,$EB,$1E,$38,$FE,$0A,$FF,$94,$34,$C4,$7E,$DE,$76,$F8,$4D,$DF
	.dc.b	$F8,$FA,$B8,$47,$A7,$0F,$3A,$F9,$8F,$E1,$6F,$F1,$26,$BF,$C0,$DB
	.dc.b	$8F,$F2,$D8,$E1,$3F,$E4,$A9,$CE,$3F,$E9,$06,$CF,$FA,$79,$AF,$03
	.dc.b	$F8,$FF,$87,$FE,$36,$7F,$97,$35,$FE,$94,$D7,$FA,$8B,$5F,$F4,$EB
	.dc.b	$BB,$FE,$4F,$F3,$0F,$94,$FF,$33,$B6,$BB,$F4,$7F,$98,$2F,$FD,$AD
	.dc.b	$1F,$E1,$4E,$EC,$1C,$3E,$56,$C7,$92,$3F,$59,$3F,$E0,$08,$C1,$0F
	.dc.b	$E9,$38,$B0,$7C,$AF,$88,$E7,$3F,$CB,$7E,$5D,$F3,$1F,$E7,$26,$C7
	.dc.b	$F4,$26,$BF,$CC,$F0,$7C,$56,$EB,$F7,$D5,$AA,$FE,$EA,$D7,$FD,$82
	.dc.b	$D7,$FD,$9E,$F4,$C7,$ED,$3F,$EC,$E2,$4F,$C5,$F8,$EF,$79,$2B,$EA
	.dc.b	$BE,$0F,$95,$91,$FD,$96,$91,$88,$FF,$01,$47,$F5,$27,$5D,$BE,$82
	.dc.b	$3F,$80,$F8,$BF,$9A,$47,$07,$5F,$8A,$F7,$91,$C1,$E1,$F3,$3E,$19
	.dc.b	$1C,$1F,$8C,$F1,$BE,$C9,$1C,$1E,$9F,$2B,$E5,$91,$C1,$ED,$F9,$8F
	.dc.b	$9E,$47,$07,$E6,$3C,$7F,$C2,$91,$C1,$F1,$79,$FE,$B9,$1C,$1F,$27
	.dc.b	$8F,$ED,$91,$C1,$FE,$11,$F9,$77,$9D,$FC,$F5,$B1,$98,$C1,$9F,$98
	.dc.b	$91,$C5,$AF,$F8,$D7,$E5,$7F,$D2,$EF,$F8,$75,$AE,$7D,$8F,$E9,$51
	.dc.b	$88,$FF,$20,$FC,$C7,$FA,$9D,$FF,$0B,$B5,$FF,$04,$37,$9F,$FC,$33
	.dc.b	$E9,$FE,$12,$77,$EE,$7F,$55,$AE,$11,$D2,$30,$47,$FC,$6F,$4A,$FF
	.dc.b	$3E,$C6,$1A,$F8,$27,$FD,$CB,$1F,$EF,$28,$D7,$87,$83,$3F,$0D,$23
	.dc.b	$24,$3F,$73,$3C,$5F,$DD,$FE,$19,$E1,$7C,$6D,$73,$1F,$EF,$48,$FF
	.dc.b	$4E,$C6,$A3,$04,$78,$F7,$7F,$BC,$3E,$59,$F1,$14,$93,$DB,$E8,$2B
	.dc.b	$FE,$6B,$9D,$BF,$E5,$36,$B9,$9F,$F0,$04,$63,$4F,$F2,$4E,$2F,$9B
	.dc.b	$F3,$1D,$23,$35,$E9,$F8,$A9,$87,$BA,$3D,$86,$1F,$87,$18,$7E,$AF
	.dc.b	$CD,$3E,$4D,$18,$6B,$FC,$92,$34,$F0,$F8,$8C,$E1,$FE,$5E,$9C,$CE
	.dc.b	$EB,$3E,$67,$FA,$0A,$73,$5B,$BC,$DF,$83,$93,$FC,$34,$FF,$9A,$9A
	.dc.b	$FF,$96,$DB,$3F,$D6,$79,$3F,$91,$61,$E6,$9D,$D7,$FA,$79,$AE,$7D
	.dc.b	$1F,$D1,$91,$FD,$0F,$B7,$E2,$57,$FE,$BA,$7C,$53,$E6,$C1,$93,$C7
	.dc.b	$C2,$F9,$BA,$62,$25,$FC,$E8,$CB,$F4,$23,$A7,$CD,$F7,$8F,$EA,$2D
	.dc.b	$BF,$F4,$4B,$74,$F9,$47,$0C,$E5,$7E,$BA,$74,$AB,$CF,$5E,$DD,$1B
	.dc.b	$D2,$57,$FA,$63,$E5,$9F,$29,$88,$FE,$1D,$3F,$C2,$88,$FF,$0F,$7E
	.dc.b	$44,$FF,$79,$BF,$C3,$B6,$7E,$E4,$FB,$BC,$6F,$7F,$8F,$27,$D1,$19
	.dc.b	$7D,$51,$97,$E1,$C7,$C6,$EB,$91,$EF,$7E,$4F,$66,$23,$E4,$C5,$D3
	.dc.b	$E4,$CF,$DD,$F9,$87,$55,$25,$FC,$1F,$26,$BF,$EA,$99,$DB,$FE,$BC
	.dc.b	$6B,$FE,$AD,$D1,$3C,$30,$FF,$16,$56,$9B,$7F,$1A,$75,$AF,$EB,$2B
	.dc.b	$4D,$9D,$61,$DD,$DB,$FB,$49,$DB,$FB,$6C,$3C,$2E,$C8,$7A,$67,$9B
	.dc.b	$4C,$E1,$E4,$7E,$01,$8B,$F1,$34,$ED,$87,$89,$61,$E2,$6D,$F6,$BE
	.dc.b	$91,$DC,$8F,$F9,$3F,$0F,$3C,$63,$0E,$ED,$3B,$9E,$13,$A9,$63,$8C
	.dc.b	$F6,$24,$F2,$7B,$61,$E7,$C3,$CC,$79,$EB,$8D,$6C,$97,$F3,$70,$F5
	.dc.b	$46,$23,$C2,$79,$44,$F2,$78,$1E,$7A,$ED,$7E,$1B,$F9,$15,$BA,$E7
	.dc.b	$87,$A1,$E2,$7E,$2E,$B8,$3E,$F6,$BE,$19,$E5,$7E,$CB,$EB,$5A,$BE
	.dc.b	$97,$E6,$8C,$61,$EB,$C3,$8D,$FF,$B3,$70,$FC,$DD,$FF,$3E,$3D,$E5
	.dc.b	$72,$FD,$63,$11,$C4,$8F,$13,$D7,$87,$DF,$E2,$FE,$0A,$78,$A3,$F3
	.dc.b	$0E,$1F,$91,$35,$FD,$64,$6D,$F9,$EE,$67,$DB,$5F,$6E,$31,$87,$D7
	.dc.b	$8C,$57,$F1,$B9,$E5,$87,$F3,$78,$C1,$E0,$29,$E0,$30,$52,$5F,$83
	.dc.b	$0E,$F1,$EA,$37,$1E,$F7,$9B,$B6,$E1,$FC,$92,$FC,$67,$28,$FC,$1C
	.dc.b	$7D,$58,$C6,$1F,$76,$78,$9D,$CF,$C9,$3E,$E7,$CB,$3F,$06,$FA,$61
	.dc.b	$EF,$9D,$C6,$2F,$C3,$1D,$0F,$09,$F6,$2B,$CC,$E6,$FC,$4F,$63,$C5
	.dc.b	$87,$B5,$C1,$F5,$8F,$87,$3C,$9C,$CF,$29,$F8,$D1,$8B,$FE,$D9,$F2
	.dc.b	$9F,$EC,$E9,$34,$9E,$E6,$BD,$C9,$7F,$E4,$C9,$CB,$FE,$4E,$F9,$8F
	.dc.b	$EA,$16,$5E,$57,$F7,$96,$CE,$FD,$6C,$19,$99,$19,$3F,$A0,$7C,$CF
	.dc.b	$F4,$92,$3F,$D9,$FE,$6E,$BF,$A1,$D5,$3F,$AD,$24,$7E,$94,$FE,$52
	.dc.b	$FF,$68,$F9,$B9,$1D,$8F,$F3,$3F,$CD,$7C,$32,$6B,$FE,$11,$5B,$58
	.dc.b	$93,$C1,$D4,$7F,$85,$FE,$5B,$E1,$9F,$02,$7F,$DB,$AD,$D9,$C3,$FE
	.dc.b	$40,$51,$C3,$6E,$5B,$4B,$FF,$3D,$CF,$F7,$D7,$FA,$1F,$27,$F6,$D2
	.dc.b	$3F,$E0,$BF,$62,$BF,$16,$3F,$88,$11,$F8,$9B,$FF,$25,$E5,$78,$7F
	.dc.b	$E0,$C3,$8E,$7E,$6A,$FF,$0E,$72,$3E,$5B,$F1,$24,$8F,$C2,$7C,$DD
	.dc.b	$7F,$DF,$73,$B7,$F1,$13,$F9,$09,$FC,$27,$B9,$F8,$5E,$F9,$5F,$8B
	.dc.b	$8C,$BF,$8D,$8C,$BF,$E2,$DE,$1F,$91,$D0,$AF,$F1,$1F,$8F,$2B,$FC
	.dc.b	$58,$FF,$92,$7E,$65,$E0,$66,$7F,$94,$6F,$FE,$68,$7F,$99,$35,$FF
	.dc.b	$15,$D7,$E2,$5F,$EA,$2F,$3E,$E9,$1D,$8F,$C4,$EB,$F8,$72,$3F,$8C
	.dc.b	$D6,$BE,$93,$FC,$8D,$D3,$F1,$5C,$70,$23,$0F,$D1,$31,$3F,$8D,$D3
	.dc.b	$F1,$DB,$CB,$FC,$2C,$65,$FC,$3C,$65,$FC,$8C,$7F,$9B,$FE,$6F,$F7
	.dc.b	$F1,$F9,$1E,$0F,$F4,$7C,$62,$3C,$C1,$C6,$1E,$3F,$9A,$F9,$73,$C4
	.dc.b	$FF,$4C,$CF,$17,$FA,$E3,$5E,$BC,$5F,$2E,$31,$1C,$23,$FC,$FD,$3F
	.dc.b	$D8,$EB,$B4,$F9,$E7,$FD,$0D,$F6,$8E,$FF,$A8,$D9,$FD,$F3,$5F,$EC
	.dc.b	$8D,$7F,$B7,$7C,$D7,$F1,$4F,$9A,$FE,$08,$D8,$FE,$D4,$BB,$5F,$31
	.dc.b	$AE,$E5,$FF,$0B,$F9,$A7,$AD,$FA,$5C,$7F,$BD,$46,$38,$35,$39,$9F
	.dc.b	$F5,$DC,$E5,$FE,$F1,$F5,$9F,$99,$AF,$E2,$CD,$8C,$F1,$71,$7F,$D8
	.dc.b	$B1,$97,$FC,$11,$93,$DF,$D4,$7A,$1B,$19,$BF,$E4,$4D,$9F,$F0,$47
	.dc.b	$CC,$BE,$6B,$8B,$6B,$F9,$43,$67,$73,$DB,$F1,$0F,$F5,$85,$7F,$2D
	.dc.b	$6C,$FF,$84,$DA,$FF,$87,$5A,$FF,$B7,$3E,$67,$A3,$FE,$21,$6B,$9A
	.dc.b	$F5,$C6,$23,$FD,$CA,$BD,$BC,$92,$FF,$9F,$46,$5F,$E8,$1F,$87,$FF
	.dc.b	$AC,$1B,$E8,$FC,$68,$C5,$FF,$8A,$5A,$FF,$8C,$FE,$01,$E9,$7C,$C6
	.dc.b	$DF,$F4,$B6,$CF,$F8,$C5,$AF,$1F,$9A,$E0,$49,$E0,$FD,$6F,$2D,$F8
	.dc.b	$2D,$24,$FF,$57,$8C,$BF,$D6,$7E,$6F,$F0,$16,$23,$86,$F2,$7E,$0A
	.dc.b	$31,$07,$FA,$7C,$F3,$34,$43,$FB,$37,$E4,$7F,$26,$7F,$C9,$8D,$7F
	.dc.b	$CA,$CD,$7C,$41,$C3,$E5,$F9,$B7,$C5,$7E,$51,$B5,$FD,$CD,$B3,$FE
	.dc.b	$59,$6B,$FE,$63,$7F,$C1,$31,$BE,$BE,$44,$7F,$BE,$7C,$DF,$F7,$47
	.dc.b	$F8,$97,$84,$FC,$71,$EB,$7F,$29,$9F,$3D,$6F,$1F,$E5,$F5,$A9,$FE
	.dc.b	$35,$09,$D3,$FE,$75,$6B,$CB,$D5,$FD,$7C,$62,$3E,$A3,$D7,$0F,$E0
	.dc.b	$8D,$AF,$EF,$DF,$36,$F7,$BC,$97,$C3,$0A,$BE,$71,$F2,$FA,$C9,$87
	.dc.b	$F1,$4C,$7C,$F5,$A7,$E0,$C6,$6F,$F9,$CC,$E5,$FF,$43,$35,$FF,$48
	.dc.b	$35,$FB,$6D,$7F,$1D,$F3,$9C,$6F,$FA,$97,$C8,$BD,$CF,$3C,$E7,$1F
	.dc.b	$F0,$D4,$E5,$F9,$71,$9B,$E7,$97,$07,$8B,$F0,$30,$E1,$1F,$D4,$1F
	.dc.b	$90,$FF,$89,$70,$D6,$7B,$66,$1F,$62,$7F,$49,$F3,$9F,$A5,$F9,$B7
	.dc.b	$AB,$F9,$6C,$7A,$9F,$CA,$E7,$51,$9C,$3F,$C3,$D5,$C2,$FF,$D6,$95
	.dc.b	$A7,$C5,$85,$7F,$43,$17,$0F,$4D,$71,$9C,$87,$11,$AC,$3F,$D7,$15
	.dc.b	$C7,$0F,$0C,$62,$F8,$3E,$5C,$3C,$11,$EE,$32,$79,$0A,$78,$30,$EF
	.dc.b	$5D,$23,$D6,$6B,$0F,$13,$89,$DD,$76,$86,$1F,$E2,$16,$9F,$E2,$37
	.dc.b	$07,$F9,$B7,$E6,$DF,$10,$F9,$CC,$36,$BF,$DA,$D3,$A9,$C9,$C7,$E7
	.dc.b	$19,$52,$57,$08,$CB,$63,$E9,$69,$FA,$68,$BB,$7F,$14,$D3,$E9,$62
	.dc.b	$C3,$8C,$61,$B7,$FC,$68,$E5,$7F,$29,$B3,$BC,$3F,$01,$7C,$DE,$F5
	.dc.b	$0F,$A3,$19,$6D,$F6,$7F,$9C,$5F,$43,$C1,$87,$9D,$FE,$57,$5C,$70
	.dc.b	$F0,$57,$8E,$B4,$F2,$6A,$9E,$15,$D2,$3F,$8D,$BE,$CC,$3F,$C2,$CD
	.dc.b	$D3,$BE,$3E,$5A,$FD,$3E,$1F,$A7,$9E,$51,$FE,$42,$AE,$0F,$F5,$1C
	.dc.b	$3B,$BF,$D4,$FE,$73,$F8,$23,$FC,$A9,$AF,$F9,$29,$B8,$FF,$89,$DF
	.dc.b	$A9,$1B,$DB,$C0,$F3,$6A,$8E,$38,$7F,$50,$78,$C7,$F9,$5B,$C6,$FF
	.dc.b	$52,$13,$D0,$E7,$83,$E6,$DF,$31,$D2,$BF,$C7,$D7,$FE,$C5,$6B,$FD
	.dc.b	$70,$FE,$DC,$E1,$C3,$5F,$EC,$2F,$FA,$45,$FE,$92,$D7,$C9,$E0,$F8
	.dc.b	$A9,$F3,$9E,$18,$93,$C1,$D4,$62,$3C,$47,$F8,$7E,$3F,$D2,$55,$D6
	.dc.b	$3C,$71,$8F,$0E,$47,$AD,$AF,$2F,$9A,$7C,$E7,$CC,$6C,$7F,$60,$6B
	.dc.b	$F2,$E7,$53,$FE,$A1,$9C,$CF,$C5,$6B,$FE,$02,$6B,$F2,$3B,$BC,$E4
	.dc.b	$7E,$73,$EA,$F8,$B9,$1F,$45,$AF,$9F,$E7,$3C,$04,$7F,$DF,$54,$8F
	.dc.b	$87,$F1,$72,$7E,$D3,$5F,$93,$9B,$E2,$BE,$49,$23,$F0,$0F,$B7,$E1
	.dc.b	$A4,$76,$6B,$CB,$C5,$FA,$24,$7F,$10,$FD,$4F,$22,$47,$E2,$DA,$FD
	.dc.b	$1F,$5B,$E5,$11,$F5,$BF,$73,$DD,$91,$F9,$36,$BE,$AF,$27,$F0,$24
	.dc.b	$7F,$30,$FD,$5F,$9C,$91,$F1,$9A,$FE,$2B,$F0,$3F,$F3,$F3,$F8,$C0
	.dc.b	$E1,$D7,$AD,$FE,$DF,$8C,$47,$E8,$FE,$71,$E9,$FF,$B8,$DB,$1C,$DA
	.dc.b	$FE,$61,$AF,$1F,$9C,$E0,$47,$83,$FE,$ED,$FC,$0D,$E5,$7F,$00,$6C
	.dc.b	$FF,$8D,$1A,$FF,$B9,$BB,$3F,$89,$A8,$CB,$FB,$96,$BF,$E5,$66,$BF
	.dc.b	$E6,$96,$BC,$67,$F8,$3F,$E1,$DF,$E8,$D8,$FF,$25,$B5,$FF,$71,$CF
	.dc.b	$8B,$BF,$FD,$E9,$2F,$BB,$E3,$F3,$F2,$F8,$C7,$43,$0F,$FB,$FA,$53
	.dc.b	$98,$EF,$ED,$F8,$88,$F8,$BE,$32,$BC,$62,$E1,$E2,$9F,$E4,$FF,$06
	.dc.b	$FB,$BA,$4B,$FF,$32,$FE,$28,$FA,$5C,$47,$F0,$EA,$E0,$FE,$15,$AE
	.dc.b	$6B,$F1,$0F,$FB,$81,$FE,$5B,$4F,$94,$A7,$4C,$5E,$5F,$FA,$5D,$4B
	.dc.b	$E6,$F8,$FE,$21,$31,$F2,$C7,$43,$0F,$F8,$12,$F9,$52,$B5,$3E,$C9
	.dc.b	$F1,$91,$F7,$F9,$35,$E4,$7F,$83,$AB,$FD,$F1,$87,$9A,$31,$5E,$CA
	.dc.b	$FE,$A3,$F3,$DF,$EF,$E9,$FF,$42,$B5,$E6,$1C,$57,$31,$F4,$7F,$C5
	.dc.b	$E7,$8E,$BF,$2A,$B9,$47,$F4,$F6,$BF,$EA,$C6,$BF,$EB,$F7,$AE,$1F
	.dc.b	$2E,$BF,$DA,$B3,$FE,$A0,$6B,$E4,$9D,$BF,$E2,$C8,$7C,$E7,$F6,$47
	.dc.b	$FC,$A7,$5C,$FD,$3C,$4C,$AE,$91,$B3,$0F,$CA,$C5,$F4,$5E,$E8,$FF
	.dc.b	$58,$B5,$D3,$63,$E2,$3F,$AB,$6E,$1F,$DD,$A3,$58,$7E,$1D,$B5,$FE
	.dc.b	$44,$78,$47,$FB,$43,$E7,$5F,$43,$FE,$06,$6D,$7F,$AE,$9A,$F8,$BE
	.dc.b	$8F,$53,$F9,$A9,$CC,$FF,$7F,$6B,$D6,$34,$79,$7E,$8E,$47,$58,$F4
	.dc.b	$1E,$9E,$79,$1F,$0A,$32,$FC,$48,$CB,$F1,$23,$FC,$2B,$F4,$6F,$9C
	.dc.b	$A4,$9F,$F6,$DC,$7E,$B5,$B1,$97,$E9,$47,$17,$7C,$DF,$E2,$8F,$A2
	.dc.b	$FF,$39,$C7,$FB,$71,$AF,$E6,$1B,$BF,$E8,$63,$F3,$5E,$A7,$1F,$80
	.dc.b	$91,$C5,$FD,$1C,$62,$39,$45,$9E,$5C,$9F,$E0,$A7,$A5,$F6,$E0,$5C
	.dc.b	$8C,$36,$FA,$BD,$C8,$76,$3F,$63,$87,$7A,$E7,$0A,$E6,$DA,$E6,$F1
	.dc.b	$25,$F9,$23,$2D,$8E,$ED,$7F,$55,$7D,$DF,$31,$AB,$FD,$B4,$66,$3C
	.dc.b	$06,$4F,$DC,$39,$BF,$29,$E6,$C3,$E3,$57,$AA,$33,$87,$39,$CD,$F8
	.dc.b	$0F,$4D,$6E,$7D,$2D,$8F,$01,$E2,$3D,$26,$2B,$AB,$F2,$8C,$4F,$7B
	.dc.b	$F3,$36,$32,$74,$BF,$1B,$D7,$0F,$8C,$4C,$3E,$1C,$70,$3C,$E7,$C2
	.dc.b	$31,$1B,$70,$68,$E9,$1E,$B7,$D2,$67,$0F,$4B,$89,$E4,$F0,$3E,$33
	.dc.b	$F9,$2C,$3F,$2D,$1F,$E7,$B9,$FD,$0B,$70,$FE,$B7,$18,$8F,$01,$93
	.dc.b	$FC,$4C,$E5,$FA,$46,$4F,$E5,$DF,$0D,$FD,$18,$F9,$2F,$F4,$67,$0F
	.dc.b	$C0,$C3,$F9,$E4,$62,$3D,$95,$70,$FD,$4C,$F2,$9E,$A7,$33,$91,$FD
	.dc.b	$0D,$E1,$5F,$1E,$7E,$1E,$29,$CD,$7C,$4A,$D1,$E2,$7A,$1E,$22,$D7
	.dc.b	$89,$EE,$7F,$5D,$9E,$6F,$CF,$25,$77,$78,$D7,$67,$C0,$79,$CC,$98
	.dc.b	$38,$CF,$4C,$3E,$ED,$F8,$67,$E0,$9E,$52,$B9,$3F,$03,$1F,$AB,$9E
	.dc.b	$58,$7E,$A5,$B5,$F9,$96,$BF,$8F,$C3,$D0,$F4,$C3,$D0,$F9,$0F,$39
	.dc.b	$EF,$BE,$CE,$1F,$3C,$7C,$B8,$E4,$7B,$63,$6E,$0E,$57,$F2,$AB,$C3
	.dc.b	$87,$E8,$C6,$6F,$57,$FA,$2F,$5C,$C3,$FA,$1C,$7F,$48,$30,$FF,$42
	.dc.b	$8C,$3F,$E2,$FC,$BF,$C6,$1F,$47,$D9,$BA,$7F,$B9,$23,$12,$C7,$C3
	.dc.b	$59,$A9,$3C,$CD,$57,$0A,$DC,$7F,$4D,$E1,$FE,$A0,$47,$80,$F9,$F7
	.dc.b	$CE,$E6,$3F,$DB,$BB,$AB,$F6,$64,$8F,$D8,$BF,$CF,$AF,$FA,$B5,$FF
	.dc.b	$BD,$9B,$EE,$F3,$7F,$C6,$18,$FF,$88,$9B,$EA,$7B,$CF,$81,$F2,$13
	.dc.b	$1F,$64,$A9,$5B,$9C,$C7,$3D,$7E,$5F,$6F,$C9,$24,$EB,$B6,$BF,$E2
	.dc.b	$5A,$FF,$17,$BC,$A7,$E8,$4F,$AE,$BF,$AC,$FD,$13,$AF,$17,$FD,$9A
	.dc.b	$30,$FC,$A6,$BF,$E5,$0D,$7F,$DB,$8E,$67,$FC,$EB,$3A,$8F,$EC,$A1
	.dc.b	$C3,$E0,$8D,$9E,$07,$EB,$71,$DA,$FC,$94,$75,$3E,$3F,$CE,$3E,$0B
	.dc.b	$CF,$B2,$31,$3A,$9C,$C7,$FA,$9F,$2F,$F4,$13,$A9,$FF,$25,$36,$BF
	.dc.b	$CD,$0D,$9D,$F9,$9D,$BE,$8B,$F0,$11,$FE,$8B,$5E,$AE,$4D,$71,$CC
	.dc.b	$7E,$EC,$E9,$B3,$97,$FC,$F3,$B3,$DC,$F8,$4D,$8F,$A1,$1F,$0F,$D7
	.dc.b	$CB,$F3,$89,$1F,$EE,$B9,$E0,$FD,$03,$E8,$47,$CB,$F6,$A4,$74,$3F
	.dc.b	$23,$F4,$3E,$F7,$F6,$4D,$D8,$D3,$E5,$46,$B7,$FA,$31,$96,$ED,$F5
	.dc.b	$3E,$8F,$E3,$D7,$EF,$5F,$84,$FF,$9E,$5B,$7F,$E9,$77,$69,$A6,$B7
	.dc.b	$FF,$5D,$36,$7A,$6F,$F2,$3D,$D9,$1F,$97,$7F,$15,$F0,$58,$8F,$89
	.dc.b	$B7,$E6,$1B,$3F,$98,$23,$FC,$23,$8F,$FC,$C0,$FE,$7F,$C4,$E5,$9F
	.dc.b	$F9,$5F,$17,$DC,$7F,$B5,$1B,$5F,$44,$38,$7E,$A3,$5F,$D2,$FD,$03
	.dc.b	$8F,$FD,$BA,$D8,$E4,$D7,$EE,$35,$FD,$6F,$D0,$7E,$4B,$E8,$3C,$0D
	.dc.b	$8F,$C1,$3E,$2F,$9E,$CF,$3A,$BC,$DF,$29,$E1,$E3,$B9,$7F,$9C,$F1
	.dc.b	$FF,$2E,$CE,$5F,$C6,$35,$FD,$EF,$CF,$BD,$3E,$11,$FE,$6E,$8C,$5F
	.dc.b	$C3,$6C,$7E,$55,$AF,$F0,$6C,$9F,$09,$CA,$11,$8A,$F9,$8D,$8F,$CF
	.dc.b	$35,$FE,$1F,$E7,$3C,$DE,$4D,$AF,$A8,$D9,$FD,$33,$5F,$E2,$FF,$3C
	.dc.b	$F3,$3C,$4D,$AF,$B8,$D9,$FD,$73,$5F,$E3,$FF,$3C,$F8,$5F,$53,$6B
	.dc.b	$F0,$8D,$9F,$DB,$35,$FE,$4F,$F6,$7F,$EE,$69,$FC,$53,$63,$F7,$8D
	.dc.b	$77,$E4,$7C,$7F,$9F,$FE,$B5,$5F,$91,$6C,$FF,$03,$6B,$FC,$D1,$AF
	.dc.b	$FA,$6B,$E1,$9E,$6F,$D1,$6D,$7E,$6D,$B3,$FC,$35,$AF,$F3,$EF,$B0
	.dc.b	$7C,$CF,$D9,$6D,$7E,$8D,$B3,$FC,$55,$AF,$F4,$6F,$09,$F3,$FF,$81
	.dc.b	$6D,$7E,$AD,$B3,$FC,$75,$AF,$F4,$EF,$9F,$FC,$5B,$FE,$B4,$6B,$FE
	.dc.b	$C0,$6A,$6F,$B7,$3A,$3F,$20,$DB,$FD,$D7,$CF,$BE,$3F,$F2,$8D,$8C
	.dc.b	$C6,$0F,$CC,$7C,$FB,$E4,$29,$27,$F8,$24,$65,$FE,$0D,$F2,$6F,$31
	.dc.b	$1F,$E0,$4F,$9F,$7A,$18,$DB,$FC,$4D,$3F,$ED,$0B,$E0,$EE,$75,$E4
	.dc.b	$FF,$BA,$E7,$35,$C0,$FF,$17,$5F,$4B,$FF,$12,$CE,$AB,$8C,$E6,$B8
	.dc.b	$09,$E6,$FF,$7C,$6B,$E3,$C3,$F7,$11,$88,$FF,$77,$FC,$FB,$E6,$BC
	.dc.b	$B8,$FF,$BC,$B1,$78,$9F,$BE,$8F,$F1,$BD,$6D,$FF,$12,$C6,$6F,$C7
	.dc.b	$39,$7D,$95,$CB,$C2,$EB,$1E,$0C,$2C,$E7,$0F,$F2,$1F,$CF,$F1,$8F
	.dc.b	$8F,$D1,$FE,$0B,$9C,$D7,$08,$C5,$FC,$2B,$E1,$5F,$F0,$01,$B0,$0D
	.dc.b	$8E,$67,$F9,$4A,$F9,$9F,$6E,$7F,$CC,$1A,$7A,$1F,$F3,$17,$A5,$67
	.dc.b	$EA,$57,$F9,$46,$30,$6E,$7C,$1F,$3F,$E1,$F9,$E7,$B7,$BB,$FF,$30
	.dc.b	$D6,$84,$E9,$FA,$57,$F8,$D8,$C4,$7A,$C5,$F5,$BE,$91,$C9,$FF,$07
	.dc.b	$46,$63,$9D,$D7,$94,$5C,$3E,$5D,$F6,$9F,$23,$E0,$8C,$DF,$49,$EC
	.dc.b	$74,$29,$D0,$B1,$C1,$FF,$27,$D7,$7C,$3D,$95,$CE,$FB,$46,$2B,$67
	.dc.b	$27,$0F,$53,$FB,$D5,$7F,$7B,$7E,$51,$83,$FC,$5D,$F3,$BF,$D3,$FE
	.dc.b	$7F,$FC,$00,$D8,$CF,$92,$9C,$B6,$38,$FC,$FF,$FA,$85,$B9,$F2,$8C
	.dc.b	$47,$F9,$AA,$D1,$88,$FA,$D1,$89,$FA,$19,$FE,$BE,$6F,$07,$DB,$D3
	.dc.b	$EE,$DE,$E7,$E5,$B6,$3F,$03,$5C,$67,$E7,$B6,$38,$E1,$E7,$6E,$1D
	.dc.b	$AA,$1D,$27,$55,$E6,$6C,$7E,$9E,$3A,$1F,$8F,$2B,$FA,$BB,$F2,$47
	.dc.b	$1A,$EB,$7E,$6C,$3D,$93,$9B,$F3,$C7,$FA,$2E,$3D,$55,$FA,$F6,$EB
	.dc.b	$BB,$E5,$7F,$EC,$69,$CB,$E5,$BF,$F4,$A6,$1E,$28,$FD,$F7,$CF,$FF
	.dc.b	$99,$D3,$49,$97,$FA,$6D,$C6,$D4,$C4,$7F,$A7,$FF,$71,$FE,$BF,$5F
	.dc.b	$EC,$2F,$98,$7D,$0F,$FA,$71,$B1,$FE,$DA,$9D,$BF,$CD,$1B,$B7,$F4
	.dc.b	$09,$DC,$FF,$24,$6C,$7F,$0F,$9D,$ED,$FB,$D5,$F3,$1F,$9D,$6D,$FF
	.dc.b	$96,$5A,$FF,$A2,$DA,$FF,$AF,$1C,$D7,$87,$D0,$F4,$1E,$DA,$95,$ED
	.dc.b	$8C,$BE,$B8,$F8,$59,$78,$49,$1E,$16,$BE,$FC,$FE,$41,$C1,$F8,$8F
	.dc.b	$A4,$8F,$A5,$F6,$7D,$34,$8F,$63,$5F,$67,$CB,$FC,$12,$3F,$05,$F8
	.dc.b	$9F,$11,$23,$E2,$35,$F8,$9F,$1F,$F2,$48,$FC,$97,$E6,$7E,$11,$23
	.dc.b	$E6,$35,$F9,$9E,$C7,$D1,$23,$F4,$5F,$E2,$1E,$2B,$E6,$29,$27,$FD
	.dc.b	$1D,$19,$6F,$DE,$70,$23,$83,$7F,$15,$D0,$8E,$0F,$F4,$C7,$CF,$3E
	.dc.b	$87,$F9,$03,$63,$31,$83,$CB,$F0,$F2,$3C,$6D,$7F,$DA,$5F,$85,$BE
	.dc.b	$97,$F9,$53,$63,$FD,$4C,$D7,$71,$FC,$BF,$E9,$1F,$49,$49,$3F,$EA
	.dc.b	$D8,$CB,$7E,$AF,$81,$1C,$1B,$F0,$7A,$11,$C1,$F0,$3D,$8F,$EC,$1D
	.dc.b	$7F,$DD,$75,$46,$3B,$7F,$B2,$46,$2B,$FC,$35,$1F,$D9,$FE,$63,$FB
	.dc.b	$44,$7F,$8A,$27,$4F,$FB,$CD,$AE,$9A,$F4,$AF,$9E,$77,$7F,$C1,$93
	.dc.b	$FE,$A2,$7B,$98,$F2,$F4,$AB,$BE,$1F,$E4,$2F,$98,$7C,$FF,$38,$46
	.dc.b	$2B,$9C,$62,$75,$3F,$EC,$F8,$E5,$18,$24,$F0,$9D,$65,$FE,$3B,$7F
	.dc.b	$D8,$51,$9F,$98,$E6,$74,$F6,$24,$64,$87,$F8,$EA,$78,$BF,$27,$E9
	.dc.b	$BD,$B9,$1A,$7C,$0D,$1F,$43,$8A,$FC,$CC,$FE,$60,$8E,$0F,$92,$25
	.dc.b	$F0,$3C,$CE,$51,$FE,$4C,$F9,$8F,$97,$A7,$F9,$5D,$B0,$D9,$18,$6F
	.dc.b	$A7,$F3,$E3,$04,$E1,$C4,$67,$8F,$5C,$B3,$0F,$A7,$69,$D4,$62,$74
	.dc.b	$F9,$63,$E2,$FD,$2C,$8F,$6E,$E8,$F8,$D5,$99,$FC,$94,$7E,$F0,$8F
	.dc.b	$F0,$07,$E6,$6D,$23,$9C,$7E,$27,$E6,$1F,$47,$7E,$0A,$57,$E2,$A7
	.dc.b	$2D,$9C,$BF,$E7,$68,$FE,$A2,$FC,$97,$6F,$F9,$EB,$E6,$7F,$28,$DE
	.dc.b	$7F,$CC,$46,$38,$3F,$D3,$AF,$E4,$27,$E9,$39,$E1,$FF,$46,$36,$33
	.dc.b	$5C,$BE,$3B,$F4,$FF,$38,$FA,$1A,$4A,$FC,$2F,$37,$F6,$8C,$7F,$CC
	.dc.b	$55,$B9,$FD,$E9,$FE,$7F,$70,$E1,$AF,$F0,$26,$BC,$AB,$F3,$1F,$11
	.dc.b	$23,$F2,$73,$CB,$3F,$F5,$CF,$CD,$7C,$46,$EF,$FB,$98,$F9,$30,$75
	.dc.b	$8F,$FA,$AD,$B7,$F9,$57,$64,$DF,$F8,$17,$85,$F1,$6C,$D2,$6B,$AD
	.dc.b	$A6,$1F,$C1,$A7,$2F,$A2,$FE,$59,$1F,$93,$CF,$5F,$F6,$4E,$9E,$44
	.dc.b	$DD,$D3,$BC,$BF,$D4,$93,$B3,$51,$9D,$3E,$9E,$17,$0F,$E4,$B3,$5E
	.dc.b	$B1,$70,$EB,$1E,$72,$1E,$48,$FE,$15,$3C,$6F,$C1,$18,$9D,$E1,$E4
	.dc.b	$8E,$95,$B6,$CE,$B0,$F2,$B4,$F2,$94,$F2,$97,$0F,$2B,$E3,$35,$87
	.dc.b	$C0,$6D,$76,$74,$4C,$3C,$8F,$D0,$36,$75,$C3,$CC,$D3,$27,$1C,$3F
	.dc.b	$AC,$46,$A3,$07,$A0,$A7,$A0,$B8,$7A,$2B,$C3,$1A,$C3,$D0,$DC,$3D
	.dc.b	$13,$A2,$61,$F1,$A7,$F9,$53,$E2,$7C,$B1,$A7,$C4,$D7,$C4,$DB,$F1
	.dc.b	$4F,$39,$FD,$85,$7F,$AD,$EF,$F8,$6C,$60,$FD,$C9,$EB,$30,$7E,$E6
	.dc.b	$38,$11,$FE,$07,$C3,$FD,$3C,$F8,$D8,$7F,$BE,$70,$46,$67,$D3,$87
	.dc.b	$F8,$0E,$31,$1F,$0C,$E6,$43,$07,$C2,$3F,$BF,$BC,$2B,$E0,$61,$FA
	.dc.b	$78,$C5,$7C,$99,$F8,$51,$8A,$FE,$29,$87,$E8,$63,$11,$F2,$2B,$81
	.dc.b	$E5,$7C,$55,$DA,$B7,$87,$B2,$7D,$67,$90,$DD,$72,$7B,$98,$34,$7B
	.dc.b	$2B,$E3,$5F,$8B,$0F,$D6,$CE,$CF,$EE,$51,$F3,$2B,$C9,$1F,$0C,$C9
	.dc.b	$E2,$29,$9B,$F3,$3D,$E3,$8D,$F8,$23,$84,$EA,$31,$1B,$3C,$EE,$0F
	.dc.b	$84,$F9,$63,$84,$FB,$E3,$E5,$1C,$1E,$E6,$6F,$EB,$47,$B5,$F7,$BE
	.dc.b	$7B,$F8,$58,$7D,$58,$D3,$FD,$B7,$E8,$3F,$B7,$7A,$32,$7F,$B6,$46
	.dc.b	$5F,$ED,$D1,$97,$FB,$64,$7F,$66,$F9,$D9,$1F,$E3,$CE,$44,$7F,$6A
	.dc.b	$F9,$1C,$CF,$F6,$F8,$BA,$FF,$6C,$9F,$F2,$5F,$D8,$7F,$84,$6B,$FC
	.dc.b	$4E,$D7,$EA,$36,$5A,$A7,$FC,$49,$F5,$9F,$B4,$71,$5C,$5F,$F3,$2B
	.dc.b	$5E,$01,$C3,$C3,$B3,$F6,$EE,$5F,$F2,$D4,$FF,$A0,$23,$F5,$16,$8C
	.dc.b	$3E,$48,$E4,$7F,$97,$B9,$5A,$7F,$6B,$D6,$4E,$63,$FB,$71,$87,$AC
	.dc.b	$FF,$B7,$B2,$FE,$DD,$3F,$DA,$B4,$95,$FD,$95,$FE,$DB,$A6,$23,$CB
	.dc.b	$C7,$F4,$9D,$1F,$DA,$E3,$FC,$B5,$E4,$C8,$FF,$08,$BF,$E7,$7F,$AB
	.dc.b	$F9,$51,$FE,$57,$8F,$87,$C9,$FE,$B3,$8F,$F0,$9B,$5F,$F4,$8B,$5F
	.dc.b	$F4,$A4,$73,$F7,$64,$76,$3F,$C7,$1E,$BF,$F8,$F7,$DF,$91,$82,$1F
	.dc.b	$E3,$97,$2F,$F8,$EE,$3F,$C6,$BF,$15,$23,$F2,$0F,$F8,$E3,$E7,$B1
	.dc.b	$1F,$E3,$D8,$BC,$5F,$E3,$A9,$FC,$7F,$8D,$2B,$F1,$EF,$DE,$F1,$B1
	.dc.b	$1F,$E3,$0E,$7F,$F3,$BB,$FC,$E9,$AE,$67,$F0,$F1,$8B,$FC,$11,$F8
	.dc.b	$5E,$7F,$C4,$9D,$1F,$F1,$7E,$5F,$E2,$DC,$E6,$1F,$87,$8F,$F1,$81
	.dc.b	$87,$F0,$71,$87,$FC,$49,$9F,$F8,$B0,$8F,$F8,$9F,$3A,$FE,$0A,$2D
	.dc.b	$F9,$38,$55,$FE,$BA,$F9,$D7,$C2,$D2,$61,$FE,$B6,$E2,$FF,$48,$A9
	.dc.b	$DB,$FD,$71,$AF,$F9,$59,$AF,$FB,$51,$AF,$EC,$1D,$E1,$F1,$A7,$FD
	.dc.b	$9D,$3F,$AC,$6C,$7E,$AE,$3C,$15,$FE,$D0,$F9,$F7,$A7,$FB,$66,$CF
	.dc.b	$F6,$E6,$BF,$E6,$86,$BF,$EE,$67,$E0,$BB,$C7,$91,$1F,$EE,$FF,$59
	.dc.b	$FC,$F9,$FF,$76,$B5,$FE,$12,$D7,$F8,$93,$57,$CE,$3E,$71,$ED,$7F
	.dc.b	$80,$A3,$0D,$F9,$EF,$83,$5F,$C5,$27,$2F,$FA,$03,$EB,$1F,$4F,$F0
	.dc.b	$E3,$F8,$CC,$63,$BF,$FC,$1A,$D8,$E1,$3B,$7F,$93,$7D,$7F,$FA,$AA
	.dc.b	$BF,$C2,$6D,$8F,$F4,$6B,$5F,$E4,$A1,$C3,$FC,$A3,$F1,$47,$67,$58
	.dc.b	$46,$2B,$FC,$3E,$D8,$FF,$4C,$B5,$EA,$BD,$8E,$4D,$9E,$31,$83,$33
	.dc.b	$23,$AB,$4F,$F5,$2F,$6F,$F9,$F2,$3F,$C5,$ED,$8C,$F9,$5E,$78,$C4
	.dc.b	$7B,$91,$92,$4F,$27,$87,$D3,$48,$EA,$D9,$FF,$5A,$FC,$CF,$DF,$9F
	.dc.b	$F1,$FB,$63,$3C,$3F,$1E,$31,$1F,$39,$70,$64,$95,$C9,$E1,$ED,$48
	.dc.b	$EA,$DA,$FF,$62,$FC,$BB,$C5,$FA,$8D,$AF,$F2,$8B,$67,$73,$AF,$1A
	.dc.b	$56,$9B,$3D,$39,$F9,$92,$79,$BF,$81,$F9,$5F,$EB,$B1,$FE,$5E,$6C
	.dc.b	$FF,$B6,$1A,$FE,$31,$AF,$F7,$27,$7D,$9F,$BF,$6C,$FF,$7D,$87,$CB
	.dc.b	$7F,$80,$23,$07,$1F,$27,$F9,$7B,$4C,$DF,$F8,$16,$F9,$AB,$E7,$3C
	.dc.b	$23,$55,$57,$CA,$BC,$8F,$CF,$36,$BF,$CF,$6D,$9F,$F7,$AB,$5F,$D4
	.dc.b	$B9,$BF,$E3,$4D,$9F,$67,$CB,$7E,$96,$73,$7E,$1F,$96,$C5,$7C,$1E
	.dc.b	$67,$C6,$E7,$AB,$D1,$87,$FB,$6A,$31,$5F,$03,$36,$1C,$9F,$F1,$33
	.dc.b	$FE,$27,$8D,$39,$8F,$89,$3F,$06,$BF,$A1,$D7,$37,$C2,$DF,$0F,$53
	.dc.b	$23,$C1,$5F,$E3,$3F,$96,$FF,$1A,$7C,$AB,$E1,$3F,$C5,$F8,$F2,$AF
	.dc.b	$F3,$71,$B1,$3B,$7F,$01,$87,$F8,$AE,$31,$1F,$E3,$8B,$EB,$7D,$2B
	.dc.b	$4F,$F1,$89,$4E,$9E,$78,$57,$91,$E0,$9E,$51,$98,$F2,$5F,$83,$0F
	.dc.b	$43,$E2,$8C,$D7,$07,$91,$FC,$A2,$FB,$57,$09,$F2,$46,$30,$F3,$D6
	.dc.b	$CE,$27,$D2,$C3,$C4,$FA,$8C,$DF,$37,$81,$FE,$CC,$AF,$F2,$BB,$B3
	.dc.b	$07,$C3,$F9,$57,$98,$F9,$6F,$F6,$9C,$61,$B1,$99,$E0,$D7,$FA,$1C
	.dc.b	$BC,$EF,$D5,$47,$56,$BA,$6B,$FE,$6E,$8E,$18,$78,$27,$8E,$9F,$73
	.dc.b	$4F,$E9,$77,$F5,$5A,$FC,$4B,$E2,$DA,$E3,$3F,$E7,$A3,$75,$CF,$7F
	.dc.b	$F3,$EB,$F8,$8B,$E2,$DA,$E3,$7F,$E8,$9B,$E2,$73,$BE,$2F,$A6,$77
	.dc.b	$2B,$E7,$3B,$8F,$45,$70,$7F,$D2,$93,$A6,$D7,$2A,$EF,$1E,$1A,$E0
	.dc.b	$F5,$9D,$47,$EA,$34,$ED,$3F,$ED,$5A,$F8,$B8,$FF,$89,$70,$ED,$3F
	.dc.b	$AF,$BE,$0F,$53,$F8,$25,$70,$BF,$65,$FE,$8F,$E5,$BF,$86,$D7,$F0
	.dc.b	$96,$BF,$EE,$36,$C7,$F8,$CD,$FC,$14,$6F,$6E,$EA,$37,$5F,$EA,$E6
	.dc.b	$C7,$FA,$DD,$E2,$FF,$BB,$FE,$51,$EF,$7F,$AC,$1B,$1F,$BB,$0E,$1F
	.dc.b	$DF,$35,$FF,$0B,$B9,$C3,$FC,$E3,$F2,$AF,$96,$F8,$70,$8C,$57,$F2
	.dc.b	$7A,$E1,$8F,$FB,$11,$FF,$58,$47,$FB,$35,$AF,$8B,$D9,$E7,$4F,$81
	.dc.b	$F8,$34,$9F,$51,$2B,$D5,$19,$6C,$72,$F9,$69,$1E,$38,$E7,$C3,$23
	.dc.b	$CC,$FC,$EF,$DC,$BE,$D4,$79,$63,$E4,$11,$EE,$D7,$E7,$6F,$E6,$3A
	.dc.b	$3F,$47,$E9,$7E,$81,$1F,$3B,$F4,$FE,$62,$47,$A9,$AF,$D2,$F9,$8F
	.dc.b	$B0,$47,$DE,$FD,$BF,$C3,$A4,$7C,$26,$BF,$6B,$E6,$3F,$00,$47,$E3
	.dc.b	$BF,$83,$F9,$89,$1F,$29,$AF,$E0,$BF,$0F,$FC,$41,$1F,$9E,$FF,$BE
	.dc.b	$BE,$61,$F3,$34,$93,$F8,$D8,$CB,$7E,$67,$81,$1C,$1F,$6B,$E7,$7F
	.dc.b	$2C,$47,$07,$E5,$FE,$4F,$FD,$C4,$83,$86,$FC,$37,$E7,$5A,$78,$FE
	.dc.b	$0F,$D0,$47,$07,$9F,$D5,$F6,$11,$C1,$FA,$7F,$98,$7C,$A7,$56,$C6
	.dc.b	$63,$06,$7E,$1E,$47,$16,$BF,$DA,$FE,$C1,$F4,$9F,$BE,$6C,$7F,$83
	.dc.b	$DA,$FF,$8B,$7B,$FF,$C7,$1F,$28,$F9,$1E,$17,$EF,$BD,$DF,$FA,$39
	.dc.b	$AF,$C7,$0E,$1F,$94,$F6,$F0,$BF,$DD,$B6,$F7,$7F,$7A,$DA,$F2,$5E
	.dc.b	$32,$91,$FE,$44,$B5,$27,$8E,$7A,$AF,$51,$FE,$47,$9D,$3C,$5A,$EA
	.dc.b	$B9,$CF,$F0,$EF,$AD,$F3,$69,$F0,$AB,$CA,$67,$18,$F3,$32,$79,$67
	.dc.b	$8B,$F8,$9F,$89,$75,$FF,$2E,$6C,$7F,$8F,$DA,$F8,$E5,$D1,$BA,$FC
	.dc.b	$75,$6A,$BF,$D7,$AD,$7E,$83,$5F,$04,$7A,$FC,$19,$1D,$E3,$DC,$47
	.dc.b	$D0,$FC,$1E,$DC,$47,$8E,$2E,$8C,$FF,$CF,$99,$2F,$F9,$AF,$CD,$39
	.dc.b	$6E,$73,$2F,$B1,$39,$6C,$E5,$FA,$51,$FF,$00,$31,$ED,$1F,$E7,$F8
	.dc.b	$E0,$FD,$C3,$F6,$7F,$03,$FD,$63,$A3,$FD,$83,$1F,$E6,$96,$BB,$F8
	.dc.b	$FF,$F7,$1D,$F4,$8E,$27,$EF,$EB,$F7,$55,$C2,$3F,$DD,$2D,$7F,$38
	.dc.b	$F4,$9E,$55,$FE,$E1,$9F,$F3,$C3,$5F,$34,$EF,$4F,$F5,$16,$0F,$95
	.dc.b	$F8,$E4,$AF,$8B,$F2,$D5,$F8,$71,$71,$F8,$62,$95,$F4,$3E,$6F,$E8
	.dc.b	$FC,$1C,$C3,$E7,$C6,$5F,$A3,$19,$7E,$7C,$7C,$9F,$99,$91,$F2,$5F
	.dc.b	$F0,$1E,$8F,$9B,$CC,$FF,$9D,$75,$57,$C8,$24,$7F,$29,$78,$FB,$92
	.dc.b	$3A,$1F,$B9,$F9,$BF,$88,$FF,$AE,$DB,$A9,$A3,$DD,$FF,$00,$B8,$E1
	.dc.b	$A6,$23,$3D,$9A,$EB,$98,$77,$8C,$BE,$18,$CB,$FC,$06,$FF,$81,$E1
	.dc.b	$F8,$5C,$DF,$37,$F8,$D2,$57,$DE,$BA,$FD,$E8,$B5,$F8,$49,$BC,$9F
	.dc.b	$88,$9B,$5F,$D2,$F7,$FE,$76,$F1,$3F,$31,$1A,$51,$AB,$F2,$4E,$70
	.dc.b	$F2,$D7,$8C,$E2,$43,$A1,$FA,$1C,$38,$4F,$5B,$E9,$87,$96,$B8,$61
	.dc.b	$D9,$E8,$72,$70,$D8,$FE,$7D,$3C,$A3,$FB,$36,$1E,$47,$46,$2F,$F1
	.dc.b	$78,$78,$A3,$BD,$F2,$8C,$47,$03,$E0,$B8,$C3,$C2,$7C,$D8,$E2,$75
	.dc.b	$29,$C0,$B7,$E3,$8E,$35,$DE,$73,$7F,$0E,$FB,$4F,$C3,$9F,$90,$68
	.dc.b	$EB,$87,$BD,$FF,$01,$D7,$07,$EE,$55,$72,$7C,$AB,$F0,$BE,$73,$DD
	.dc.b	$5C,$F0,$F3,$61,$E1,$AD,$D7,$AC,$F8,$93,$A7,$94,$7C,$37,$38,$7A
	.dc.b	$DF,$C1,$1F,$C6,$9F,$E3,$A4,$7F,$C3,$CD,$7F,$78,$D7,$81,$FC,$73
	.dc.b	$6F,$E0,$67,$B7,$0F,$E0,$31,$93,$E3,$1F,$CB,$E3,$6E,$2B,$F9,$3C
	.dc.b	$62,$3C,$E7,$13,$81,$BC,$3F,$1B,$1B,$9F,$87,$87,$C1,$AF,$D2,$CE
	.dc.b	$6B,$E4,$12,$3F,$5D,$19,$7E,$3C,$65,$E2,$78,$6B,$AE,$1B,$8C,$9E
	.dc.b	$63,$CA,$48,$DB,$C6,$7C,$EF,$A4,$F0,$99,$3C,$24,$3A,$C7,$C4,$78
	.dc.b	$18,$8E,$6F,$86,$34,$F8,$63,$D5,$7F,$0F,$0F,$A7,$39,$3B,$46,$70
	.dc.b	$F6,$3F,$24,$C4,$79,$4F,$CB,$9F,$94,$8D,$E1,$FC,$1D,$B1,$FA,$26
	.dc.b	$BB,$39,$4F,$59,$F9,$55,$E4,$C3,$F4,$27,$F9,$2F,$E6,$BF,$D7,$7E
	.dc.b	$0F,$FA,$E3,$04,$FF,$93,$FD,$5C,$D4,$9F,$F2,$05,$6E,$B5,$18,$9F
	.dc.b	$F6,$2F,$CD,$7F,$6D,$8F,$F2,$33,$5F,$F1,$F3,$67,$FB,$5B,$FD,$AA
	.dc.b	$37,$D1,$C8,$8F,$2E,$7F,$F7,$2F,$CD,$FF,$94,$23,$FD,$09,$7F,$E0
	.dc.b	$B8,$C4,$7F,$85,$5B,$1F,$E0,$6F,$5F,$8F,$87,$FE,$87,$9D,$D7,$F9
	.dc.b	$9C,$38,$7F,$CD,$4D,$7F,$D2,$4F,$4E,$1F,$E6,$3E,$B7,$F7,$63,$AC
	.dc.b	$FF,$99,$7E,$69,$AF,$FB,$A1,$F9,$04,$70,$67,$79,$3C,$63,$FD,$DF
	.dc.b	$DD,$F1,$F0,$F7,$B6,$7E,$BB,$74,$FF,$20,$6B,$C3,$69,$5F,$E7,$7A
	.dc.b	$D1,$F3,$DF,$8D,$87,$CC,$23,$FE,$2A,$E8,$AF,$C5,$3E,$2C,$7C,$92
	.dc.b	$39,$32,$7F,$8C,$B8,$7F,$C3,$27,$2C,$7F,$C5,$2B,$E1,$7F,$23,$E2
	.dc.b	$4A,$FC,$6C,$7E,$50,$C3,$FE,$85,$9F,$F8,$00,$D8,$FC,$7F,$C4,$C8
	.dc.b	$FC,$7B,$F9,$2F,$8A,$C4,$7E,$42,$2E,$1F,$90,$9F,$F5,$33,$FE,$53
	.dc.b	$F9,$AF,$F0,$8E,$FF,$EA,$96,$DF,$FB,$45,$B2,$F9,$BF,$F1,$24,$FE
	.dc.b	$4A,$B8,$77,$7F,$B4,$EF,$FD,$3B,$3F,$8D,$6B,$A3,$E4,$3D,$2B,$FB
	.dc.b	$D5,$F5,$ED,$F2,$E1,$BB,$FF,$53,$72,$7C,$97,$F1,$DB,$7E,$34,$FC
	.dc.b	$76,$B2,$32,$43,$8B,$F8,$CD,$64,$7E,$15,$FC,$7E,$B8,$8F,$F5,$A7
	.dc.b	$0A,$71,$9F,$F9,$8A,$15,$FE,$56,$F3,$E5,$FF,$9C,$9F,$4B,$FC,$33
	.dc.b	$E6,$7E,$3B,$FE,$83,$8F,$F7,$C7,$67,$FB,$EC,$8E,$0E,$38,$29,$F0
	.dc.b	$38,$6D,$F7,$3E,$91,$FB,$6F,$84,$FD,$F1,$1F,$CB,$75,$A3,$F3,$10
	.dc.b	$F9,$97,$95,$86,$C7,$E7,$5A,$FE,$8D,$AF,$F0,$95,$F1,$DC,$49,$1C
	.dc.b	$5E,$FF,$53,$FB,$18,$FE,$20,$D7,$FD,$F0,$D4,$1C,$2F,$98,$7A,$CF
	.dc.b	$99,$FD,$94,$61,$BF,$27,$FE,$A3,$AE,$67,$1A,$E0,$FE,$F7,$E6,$5F
	.dc.b	$33,$FE,$B5,$8D,$C6,$3A,$BF,$7C,$D8,$E1,$3B,$7C,$FF,$4F,$FE,$AD
	.dc.b	$AF,$E0,$8D,$8F,$E1,$4D,$7F,$98,$35,$F2,$FC,$CF,$DE,$9F,$E1,$8D
	.dc.b	$8F,$E2,$4D,$7F,$9C,$35,$E3,$F3,$3E,$35,$08,$C3,$67,$AC,$6F,$E9
	.dc.b	$5E,$CD,$24,$FF,$1D,$8C,$BF,$C7,$FD,$2F,$F5,$23,$7C,$7F,$04,$63
	.dc.b	$1F,$E5,$0D,$7F,$A7,$23,$64,$9E,$8F,$A7,$E9,$1F,$01,$49,$3F,$CB
	.dc.b	$E3,$2F,$F3,$1F,$9B,$FC,$1B,$78,$3E,$14,$63,$0F,$E7,$4D,$7F,$B0
	.dc.b	$2E,$1D,$92,$BA,$3F,$2F,$E9,$1E,$FD,$24,$FF,$41,$8C,$BF,$D0,$FE
	.dc.b	$73,$F8,$2B,$7C,$2F,$A9,$18,$C7,$FA,$63,$5F,$EE,$4F,$F0,$F9,$E5
	.dc.b	$E2,$70,$23,$FD,$3F,$E6,$FF,$A5,$3F,$C0,$3C,$27,$CE,$3E,$AF,$FD
	.dc.b	$EF,$3F,$A0,$AD,$E3,$FE,$1E,$8F,$F0,$4D,$6D,$FE,$13,$09,$DB,$FE
	.dc.b	$05,$6B,$CF,$C8,$FC,$74,$62,$3F,$6E,$F6,$C1,$5B,$3C,$D7,$87,$BC
	.dc.b	$ED,$FD,$AF,$CE,$BD,$DF,$4E,$1F,$E3,$4F,$9D,$FF,$71,$D7,$9A,$B4
	.dc.b	$FC,$18,$CC,$FB,$4F,$E1,$33,$A7,$FB,$D3,$5F,$F1,$03,$5F,$0D,$7E
	.dc.b	$A7,$E7,$7F,$61,$97,$AA,$7F,$C8,$BD,$0F,$91,$F2,$C2,$B7,$97,$F8
	.dc.b	$D2,$B4,$FC,$C8,$CC,$FF,$1C,$6B,$C4,$E6,$E2,$F8,$C7,$AC,$FF,$2A
	.dc.b	$FC,$EF,$82,$32,$79,$63,$35,$F6,$A3,$2F,$F7,$EC,$3C,$F3,$FC,$B3
	.dc.b	$E7,$7F,$96,$FC,$E3,$E6,$F5,$7F,$82,$AD,$09,$D3,$FE,$50,$BF,$89
	.dc.b	$18,$8F,$E5,$B7,$D2,$F9,$C7,$17,$FC,$55,$19,$8E,$57,$5E,$26,$CC
	.dc.b	$94,$CD,$77,$C3,$E2,$BE,$18,$F3,$B9,$C3,$CB,$87,$82,$79,$57,$08
	.dc.b	$C5,$6E,$FC,$55,$CA,$7D,$33,$D6,$18,$79,$EF,$94,$7F,$9A,$63,$FB
	.dc.b	$5B,$A3,$FB,$6B,$83,$FC,$8D,$F3,$9F,$DC,$7E,$76,$F9,$A6,$92,$BF
	.dc.b	$CA,$12,$F9,$8F,$F4,$AC,$61,$AF,$F9,$E6,$35,$87,$69,$CC,$FF,$60
	.dc.b	$9C,$BF,$2F,$2F,$F0,$4B,$7D,$1F,$EC,$B9,$FD,$BA,$DB,$6F,$C2,$FD
	.dc.b	$FB,$E2,$DA,$E3,$1D,$E7,$6D,$C3,$EF,$3C,$0F,$EE,$65,$C3,$FB,$FE
	.dc.b	$1E,$08,$E1,$1E,$53,$8D,$F6,$77,$7D,$9B,$7A,$61,$F8,$9B,$E8,$FF
	.dc.b	$83,$2F,$E0,$CF,$EA,$5B,$A7,$93,$0F,$F1,$CD,$2F,$8B,$FA,$78,$CD
	.dc.b	$70,$8E,$91,$C2,$77,$F3,$BF,$C1,$E7,$F9,$6B,$5F,$DE,$37,$0F,$DD
	.dc.b	$3F,$7E,$37,$B7,$8A,$35,$8F,$F8,$69,$B3,$FE,$26,$78,$3F,$C1,$FE
	.dc.b	$73,$FD,$B8,$9A,$7F,$B6,$7E,$75,$F2,$5E,$E8,$7C,$1F,$F9,$11,$B5
	.dc.b	$FE,$9D,$6B,$FE,$A3,$6B,$FE,$C9,$7F,$D2,$B9,$7F,$13,$6E,$3F,$E3
	.dc.b	$06,$CF,$FB,$BB,$07,$C2,$78,$7D,$B9,$3F,$D4,$AB,$78,$2F,$73,$9A
	.dc.b	$F3,$11,$F3,$3F,$E5,$CF,$9C,$FC,$EC,$7F,$B7,$5A,$FF,$90,$9A,$FF
	.dc.b	$93,$9A,$FC,$7E,$1F,$88,$47,$FB,$2C,$7D,$F7,$47,$C5,$8F,$90,$47
	.dc.b	$FD,$89,$1F,$E5,$17,$C1,$E6,$F9,$48,$FC,$DF,$9C,$FA,$04,$70,$7D
	.dc.b	$1F,$92,$FA,$84,$70,$76,$3E,$01,$0C,$1F,$67,$E7,$3E,$E1,$1C,$1F
	.dc.b	$77,$E7,$3F,$00,$47,$07,$E0,$7C,$9F,$C2,$11,$C1,$F8,$5F,$96,$FC
	.dc.b	$41,$1C,$1F,$EE,$5F,$9A,$7C,$DF,$FA,$41,$B1,$98,$C1,$FE,$94,$FA
	.dc.b	$37,$CF,$52,$4F,$FB,$BE,$32,$FE,$57,$D3,$FC,$C1,$1C,$1F,$EF,$5F
	.dc.b	$93,$7C,$27,$FA,$81,$B1,$98,$C1,$E5,$F8,$F9,$1E,$96,$BF,$E8,$A3
	.dc.b	$FD,$2B,$97,$FA,$49,$F7,$7A,$BE,$D2,$3B,$3F,$D3,$7F,$38,$F8,$AF
	.dc.b	$F5,$8B,$63,$41,$C3,$C5,$AF,$87,$EC,$3F,$D7,$4F,$F8,$E1,$AF,$F8
	.dc.b	$1D,$BD,$5F,$E1,$93,$F5,$5F,$2B,$FC,$72,$37,$3C,$FC,$8F,$F6,$1C
	.dc.b	$62,$3F,$D4,$ED,$7F,$D6,$6F,$4E,$FF,$F5,$AD,$BE,$C3,$FC,$CB,$5F
	.dc.b	$E5,$36,$BE,$3B,$FF,$6D,$C7,$F0,$A9,$F2,$36,$B9,$B6,$93,$D7,$83
	.dc.b	$8F,$83,$27,$B3,$D2,$3A,$F4,$F6,$F1,$25,$78,$E3,$2F,$96,$32,$FF
	.dc.b	$91,$EF,$D5,$3F,$D0,$3E,$07,$FC,$1D,$97,$FB,$21,$B5,$FE,$D4,$6D
	.dc.b	$7C,$1F,$83,$E1,$E4,$FF,$BB,$27,$FC,$11,$5B,$8C,$11,$FF,$3C,$7D
	.dc.b	$83,$E3,$7F,$DF,$ED,$8F,$F3,$9D,$76,$87,$36,$89,$3C,$BE,$A2,$BE
	.dc.b	$B9,$DB,$DA,$7A,$CF,$6F,$9D,$EF,$EF,$CA,$F1,$C6,$5F,$2C,$65,$E9
	.dc.b	$1F,$D9,$3E,$83,$EE,$EB,$FE,$51,$6E,$DF,$EE,$E6,$D7,$DA,$F9,$87
	.dc.b	$CB,$71,$F0,$FF,$3F,$2A,$FC,$3B,$FA,$D6,$BC,$E3,$71,$D3,$05,$FF
	.dc.b	$5A,$AF,$F5,$13,$5F,$F3,$44,$EE,$1F,$2C,$9D,$B5,$FC,$14,$69,$E7
	.dc.b	$D9,$9B,$FD,$54,$E6,$77,$59,$CB,$F4,$73,$9A,$DD,$E6,$FF,$37,$F4
	.dc.b	$1E,$37,$13,$EA,$6C,$7B,$4D,$3E,$C3,$39,$7F,$85,$3E,$47,$FD,$EF
	.dc.b	$1F,$EB,$86,$BF,$E8,$76,$D7,$F0,$1A,$78,$DC,$7E,$2B,$E8,$D7,$0B
	.dc.b	$FE,$38,$D7,$E2,$35,$F9,$0F,$4C,$FF,$80,$DF,$F7,$63,$FB,$EC,$A7
	.dc.b	$4F,$C0,$AF,$69,$1F,$F6,$C4,$BA,$31,$1F,$B3,$8E,$38,$9A,$6E,$3F
	.dc.b	$DC,$1F,$41,$FA,$77,$F7,$4F,$85,$CE,$1F,$3A,$35,$5F,$C8,$DB,$8F
	.dc.b	$F1,$A7,$85,$72,$AE,$1C,$FF,$E4,$2F,$15,$F1,$54,$95,$FE,$E8,$F9
	.dc.b	$5F,$E8,$99,$7F,$AC,$EB,$73,$FE,$EC,$6B,$FE,$BB,$6C,$7F,$AE,$67
	.dc.b	$8E,$5F,$93,$9C,$B6,$FB,$DF,$FB,$0E,$B6,$FE,$6E,$B6,$D9,$DD,$F5
	.dc.b	$8B,$7C,$DF,$24,$FF,$97,$AF,$9C,$2F,$9B,$67,$8B,$FA,$99,$DC,$E8
	.dc.b	$93,$CE,$76,$71,$3B,$9F,$1C,$CC,$70,$70,$68,$F8,$D1,$CA,$7B,$BC
	.dc.b	$0E,$D1,$BB,$FA,$11,$B8,$D1,$9B,$FD,$F5,$72,$8C,$1F,$1C,$F0,$9E
	.dc.b	$03,$27,$A4,$FA,$07,$43,$46,$0D,$DF,$AD,$C1,$AB,$F2,$BD,$27,$C4
	.dc.b	$E6,$BE,$A3,$C2,$76,$E4,$F3,$19,$8D,$C6,$88,$FB,$A3,$CD,$18,$7E
	.dc.b	$C1,$88,$E2,$6D,$F4,$DF,$C9,$7E,$91,$FE,$1F,$BF,$CC,$B6,$3F,$C5
	.dc.b	$44,$7F,$C5,$CD,$7F,$C4,$11,$C1,$EE,$7E,$27,$CE,$FC,$81,$E3,$70
	.dc.b	$7E,$AD,$C2,$C1,$1C,$23,$CE,$6C,$87,$C7,$3E,$C6,$1F,$AD,$8E,$11
	.dc.b	$9A,$FC,$ED,$74,$8E,$35,$F1,$DE,$38,$7E,$22,$F8,$CE,$6F,$8C,$62
	.dc.b	$7D,$B5,$C0,$F5,$9F,$B7,$8F,$5D,$F7,$8C,$1E,$A2,$92,$FB,$61,$D6
	.dc.b	$39,$C6,$1F,$3D,$F9,$5E,$51,$EF,$72,$7E,$21,$F6,$57,$85,$D4,$7C
	.dc.b	$3A,$EC,$FC,$FB,$F3,$98,$C3,$F3,$95,$C0,$98,$7C,$08,$F8,$75,$E0
	.dc.b	$3E,$11,$30,$F0,$4F,$9E,$3C,$06,$E3,$2F,$9E,$3E,$5C,$62,$3C,$87
	.dc.b	$BA,$39,$57,$8A,$38,$10,$D9,$F2,$E3,$F2,$30,$CD,$F3,$5F,$D1,$49
	.dc.b	$5F,$D3,$3A,$AB,$FD,$32,$2D,$7F,$49,$9F,$E9,$9F,$41,$FD,$37,$A6
	.dc.b	$5F,$F5,$48,$CB,$FD,$5A,$32,$FF,$56,$8F,$E9,$9F,$3D,$23,$FA,$33
	.dc.b	$FC,$CF,$E8,$31,$1F,$AA,$8F,$F4,$C1,$1F,$F4,$E3,$FC,$E7,$E0,$E4
	.dc.b	$72,$3F,$CE,$BA,$7F,$9D,$BE,$83,$FA,$2E,$66,$93,$E1,$3F,$C2,$6E
	.dc.b	$38,$A9,$88,$CF,$66,$BA,$E6,$3D,$8F,$F3,$BD,$FF,$88,$71,$F0,$56
	.dc.b	$A7,$FC,$DD,$3F,$E9,$CC,$7F,$D2,$8E,$E3,$FA,$D7,$97,$FD,$BC,$8F
	.dc.b	$F6,$DC,$2B,$FC,$32,$3F,$D7,$44,$7F,$D7,$CF,$F5,$BD,$24,$7F,$71
	.dc.b	$8F,$F1,$7F,$7B,$E7,$AF,$93,$2B,$D5,$97,$F9,$5A,$76,$66,$3F,$D0
	.dc.b	$D5,$FE,$95,$9D,$47,$0F,$A0,$FC,$5B,$79,$7F,$C7,$B1,$88,$FF,$47
	.dc.b	$3F,$E7,$46,$C6,$88,$F3,$E5,$FF,$24,$2F,$9B,$FF,$23,$3F,$EA,$4E
	.dc.b	$0F,$F2,$C4,$63,$04,$63,$C2,$FE,$F9,$F4,$1C,$AB,$8D,$FF,$B6,$5A
	.dc.b	$EC,$FF,$68,$B8,$7F,$CC,$CF,$5A,$FE,$D3,$EF,$FF,$81,$CE,$97,$FE
	.dc.b	$5A,$F9,$EF,$99,$7F,$29,$F8,$C4,$DB,$46,$F7,$93,$B8,$FF,$31,$F7
	.dc.b	$FC,$7D,$3F,$DB,$0D,$CB,$FD,$9E,$DA,$FF,$62,$BF,$2E,$77,$B7,$22
	.dc.b	$4F,$DC,$9D,$57,$C8,$BF,$88,$49,$FF,$1D,$73,$57,$FD,$F1,$AF,$B1
	.dc.b	$B3,$FE,$98,$8C,$47,$CB,$8F,$F1,$57,$0C,$8F,$D2,$1F,$66,$3F,$C5
	.dc.b	$DE,$57,$FA,$83,$B8,$D2,$4F,$D2,$F4,$3B,$11,$C1,$FE,$97,$3F,$D3
	.dc.b	$1F,$3B,$FE,$58,$70,$D7,$3E,$1F,$FA,$BA,$31,$1F,$EA,$1F,$29,$EE
	.dc.b	$7F,$AC,$9B,$1D,$4D,$46,$4F,$F5,$BF,$CE,$7E,$66,$3C,$4D,$7F,$D5
	.dc.b	$4D,$7F,$D5,$CD,$5F,$38,$F5,$DE,$97,$79,$CC,$61,$F1,$4B,$0F,$ED
	.dc.b	$11,$FE,$B7,$8E,$55,$C7,$EB,$1F,$25,$FA,$2A,$F5,$46,$39,$35,$39
	.dc.b	$9F,$95,$39,$7E,$07,$CE,$7F,$44,$AF,$23,$63,$3C,$FC,$5F,$F6,$44
	.dc.b	$65,$FA,$9F,$38,$F8,$BD,$47,$FB,$3E,$31,$7E,$F6,$C7,$CF,$6B,$F7
	.dc.b	$7E,$AD,$EC,$72,$84,$62,$BE,$43,$63,$EC,$35,$FC,$37,$CE,$BD,$4E
	.dc.b	$4D,$AF,$A0,$D9,$FC,$03,$5F,$C7,$79,$6E,$DF,$13,$6B,$EC,$36,$7F
	.dc.b	$10,$D7,$F2,$DF,$56,$F8,$1F,$53,$6B,$F0,$0D,$9F,$C8,$35,$FC,$F7
	.dc.b	$CF,$3D,$EF,$84,$DA,$FC,$43,$67,$F3,$0D,$7F,$4D,$F3,$DF,$8F,$9F
	.dc.b	$C7,$36,$3F,$3C,$D7,$7D,$1F,$3E,$DE,$A5,$25,$FE,$5E,$32,$FE,$67
	.dc.b	$E7,$9E,$E7,$D7,$6C,$66,$30,$7D,$BF,$9D,$7C,$F5,$24,$FE,$8E,$32
	.dc.b	$FE,$97,$E6,$FF,$2D,$E5,$7E,$74,$7C,$E3,$E7,$71,$4D,$A5,$F0,$8C
	.dc.b	$5F,$0A,$DD,$6B,$BA,$CE,$6B,$F8,$CD,$F1,$A4,$E5,$FE,$A4,$71,$9F
	.dc.b	$35,$7E,$5E,$BF,$59,$3F,$C5,$DA,$F9,$47,$CE,$3D,$AF,$0F,$CE,$BD
	.dc.b	$7F,$2D,$6A,$73,$18,$AE,$18,$FF,$78,$BE,$0F,$F3,$08,$CD,$F9,$B2
	.dc.b	$56,$A3,$D2,$3E,$77,$F6,$73,$F3,$7A,$9F,$11,$E2,$9C,$E3,$F2,$A7
	.dc.b	$2F,$F3,$E8,$CD,$FD,$16,$BD,$4E,$5E,$0F,$B3,$0E,$31,$E2,$5F,$3B
	.dc.b	$9C,$74,$DC,$3F,$A5,$CF,$03,$F8,$6F,$CE,$FD,$FF,$9C,$FE,$0F,$87
	.dc.b	$DE,$C3,$E4,$BF,$83,$9D,$78,$7F,$46,$FC,$37,$C2,$BE,$A5,$69,$E5
	.dc.b	$85,$72,$27,$85,$72,$BF,$E9,$17,$B8,$C5,$7F,$57,$8C,$61,$FC,$D6
	.dc.b	$FC,$13,$E5,$3A,$5F,$96,$74,$F4,$8C,$9E,$42,$9E,$49,$F1,$CF,$08
	.dc.b	$F0,$1A,$BF,$1B,$BA,$ED,$87,$B6,$3C,$47,$F5,$3C,$3F,$AA,$38,$3C
	.dc.b	$47,$F7,$1F,$9C,$7A,$0F,$9D,$FE,$B5,$18,$6C,$66,$7F,$B6,$4E,$5B
	.dc.b	$F3,$1F,$AC,$6E,$5C,$A3,$11,$DD,$AF,$EA,$65,$1C,$74,$FB,$59,$FF
	.dc.b	$51,$AF,$B0,$D7,$E3,$D7,$0B,$FB,$8D,$8F,$97,$3B,$AE,$7B,$7E,$DD
	.dc.b	$FC,$35,$6D,$B8,$78,$6B,$F7,$87,$12,$D7,$17,$D1,$1B,$BF,$E2,$D7
	.dc.b	$CE,$77,$1E,$8A,$E0,$EA,$B8,$2A,$E0,$FF,$1F,$AF,$03,$6B,$8B,$BA
	.dc.b	$E3,$3F,$0F,$0F,$F0,$13,$74,$F0,$D7,$F0,$4C,$3A,$BF,$DC,$B0,$FF
	.dc.b	$39,$D6,$E3,$B1,$FE,$3E,$9D,$FC,$EF,$F0,$88,$FF,$20,$B5,$FE,$1E
	.dc.b	$DC,$3F,$C2,$4F,$DA,$8D,$ED,$FE,$80,$8D,$63,$FC,$9D,$B3,$FC,$F5
	.dc.b	$E0,$FF,$1A,$F9,$97,$B1,$FB,$DA,$FE,$7F,$F3,$7F,$DE,$6B,$FD,$3A
	.dc.b	$D7,$84,$FF,$A3,$AB,$8C,$6B,$93,$93,$FD,$02,$B8,$65,$FE,$29,$AD
	.dc.b	$CF,$86,$BF,$4F,$1D,$63,$FC,$AC,$D7,$FA,$03,$5F,$EA,$CF,$4C,$3C
	.dc.b	$FE,$47,$82,$D8,$23,$CD,$84,$C3,$D7,$1B,$9D,$47,$F9,$C3,$0E,$8F
	.dc.b	$6F,$7F,$CA,$47,$FC,$A1,$08,$C4,$7C,$BF,$97,$F8,$87,$C5,$8B,$1C
	.dc.b	$23,$DE,$47,$9F,$D5,$7C,$22,$38,$3B,$FC,$2F,$C6,$23,$83,$C7,$F0
	.dc.b	$DF,$28,$8E,$0F,$3F,$C4,$7C,$E2,$38,$3D,$7C,$7F,$48,$8E,$0F,$7F
	.dc.b	$7F,$D6,$23,$83,$FC,$E5,$F8,$77,$F8,$69,$BE,$3F,$F2,$28,$C4,$7F
	.dc.b	$A1,$5A,$FF,$A4,$DF,$C9,$B7,$A3,$81,$1F,$F3,$F7,$CC,$BE,$1F,$F9
	.dc.b	$6B,$63,$31,$83,$F9,$8F,$CC,$BE,$6A,$92,$7F,$D1,$11,$96,$FD,$37
	.dc.b	$02,$38,$37,$F0,$5D,$08,$E0,$FF,$49,$7C,$DB,$8F,$FA,$0B,$63,$31
	.dc.b	$83,$3C,$72,$38,$B5,$FF,$55,$4B,$8F,$99,$23,$BB,$F3,$7E,$57,$FD
	.dc.b	$FF,$B9,$A7,$EC,$A5,$F2,$EF,$A7,$CC,$7D,$89,$E4,$D2,$1F,$E1,$E9
	.dc.b	$7C,$EF,$36,$C7,$FB,$5A,$FC,$0F,$F9,$1B,$BB,$FC,$8F,$F2,$AF,$9A
	.dc.b	$FD,$A3,$FE,$4F,$9D,$1F,$E4,$47,$0F,$F8,$BD,$E0,$EB,$C0,$46,$23
	.dc.b	$5E,$8F,$43,$AD,$76,$F8,$69,$3C,$6B,$FD,$84,$DF,$67,$2F,$79,$B8
	.dc.b	$7F,$99,$6B,$FD,$9D,$3F,$EF,$2D,$7F,$0C,$FF,$90,$1B,$3F,$EC,$36
	.dc.b	$E5,$FE,$55,$D3,$E1,$D6,$F9,$BF,$C0,$75,$CA,$3F,$DE,$4D,$7F,$CB
	.dc.b	$8D,$7F,$CE,$4F,$47,$FC,$E9,$3F,$CE,$9F,$B7,$E3,$7D,$D9,$D6,$0F
	.dc.b	$94,$FF,$54,$57,$C9,$7F,$97,$C7,$FA,$05,$B8,$FF,$9F,$1D,$CF,$F9
	.dc.b	$7D,$B3,$BF,$1B,$F9,$64,$70,$9F,$F3,$DB,$AA,$E5,$C7,$A2,$4F,$87
	.dc.b	$96,$BD,$62,$E1,$E3,$9F,$57,$B9,$2B,$E8,$3F,$8F,$F3,$31,$1F,$E8
	.dc.b	$13,$E2,$C7,$CC,$23,$93,$27,$E3,$8F,$D0,$FC,$FF,$E6,$4E,$77,$FE
	.dc.b	$A3,$CD,$F2,$7F,$D2,$5B,$5F,$91,$6B,$FE,$83,$5D,$BB,$99,$3B,$79
	.dc.b	$47,$F5,$3F,$95,$7C,$45,$FA,$A9,$7F,$9D,$9C,$B6,$72,$FF,$A4,$61
	.dc.b	$3C,$E7,$F0,$1E,$EC,$AF,$CB,$3F,$B5,$F7,$F1,$1F,$0F,$8F,$FD,$8E
	.dc.b	$FC,$53,$F8,$34,$FE,$50,$8F,$C7,$E5,$7C,$E7,$EF,$CE,$7B,$7F,$B2
	.dc.b	$7E,$5B,$E1,$B7,$6F,$EE,$11,$8E,$5F,$F6,$3B,$F1,$B0,$FD,$D3,$97
	.dc.b	$FD,$86,$DA,$34,$E1,$1F,$EB,$1F,$95,$F8,$AD,$DB,$FB,$F4,$7C,$48
	.dc.b	$FE,$0C,$EA,$BF,$DB,$CD,$BF,$F6,$CB,$B9,$E4,$70,$DB,$F5,$5F,$07
	.dc.b	$23,$F8,$03,$FC,$7F,$2C,$47,$F0,$18,$B8,$30,$FE,$01,$58,$7F,$77
	.dc.b	$36,$BF,$73,$35,$FE,$6D,$17,$4F,$89,$18,$3C,$24,$3F,$99,$4F,$1D
	.dc.b	$95,$C6,$30,$D9,$E1,$87,$F1,$58,$C4,$7A,$4C,$E1,$E6,$7F,$01,$5C
	.dc.b	$9C,$9E,$82,$9E,$53,$91,$E4,$33,$7E,$47,$78,$7B,$23,$11,$C3,$0F
	.dc.b	$23,$E0,$38,$E1,$E6,$8C,$C6,$F0,$F3,$3D,$CC,$1F,$C9,$A7,$84,$6A
	.dc.b	$78,$91,$FA,$B3,$D8,$F2,$E1,$EF,$79,$9F,$D9,$9C,$4F,$18,$F8,$46
	.dc.b	$AF,$D7,$19,$8E,$07,$53,$9E,$1F,$07,$0E,$78,$78,$A7,$94,$7A,$1F
	.dc.b	$51,$FB,$5B,$F8,$EF,$F4,$68,$FE,$92,$47,$4D,$7F,$7C,$DB,$FC,$A4
	.dc.b	$7F,$48,$E1,$FE,$F0,$78,$9C,$1F,$BF,$70,$FF,$81,$9D,$BE,$B8,$E6
	.dc.b	$E4,$E0,$66,$FF,$B0,$4F,$08,$F6,$9A,$3F,$C1,$4F,$6C,$3F,$7B,$5D
	.dc.b	$1B,$5F,$64,$95,$E6,$38,$9D,$CC,$4F,$1A,$E8,$78,$5F,$A9,$87,$D6
	.dc.b	$AF,$2D,$78,$E7,$38,$6E,$73,$3E,$83,$B9,$27,$6F,$18,$DC,$62,$35
	.dc.b	$18,$3D,$2F,$E9,$E3,$CC,$F1,$3E,$F9,$F8,$6B,$F1,$38,$8F,$C6,$61
	.dc.b	$F1,$23,$DF,$5D,$30,$F6,$46,$08,$77,$8F,$0C,$F1,$8F,$B0,$7B,$6B
	.dc.b	$C6,$FA,$0F,$04,$F0,$BF,$1C,$73,$32,$76,$AE,$97,$E2,$8D,$1D,$AF
	.dc.b	$A9,$FA,$93,$FD,$9B,$F4,$1F,$DB,$49,$1F,$DB,$7E,$46,$BF,$DB,$62
	.dc.b	$CF,$F7,$59,$FE,$D5,$F0,$3F,$DA,$F9,$E5,$FF,$69,$8C,$BF,$DA,$E3
	.dc.b	$2F,$F6,$98,$FF,$31,$7D,$47,$F9,$CE,$17,$F9,$C8,$C3,$7C,$8F,$F6
	.dc.b	$57,$D4,$7F,$87,$E9,$5C,$5F,$F0,$93,$51,$1C,$36,$7F,$B8,$3C,$9E
	.dc.b	$BD,$92,$3F,$C2,$D3,$E2,$DF,$FB,$4E,$9F,$DB,$09,$3F,$DA,$B4,$AF
	.dc.b	$AA,$76,$FF,$6E,$9F,$EE,$73,$FD,$9B,$3F,$EC,$FA,$4A,$FE,$CF,$19
	.dc.b	$7F,$B4,$C6,$5F,$EE,$11,$FE,$89,$F8,$E7,$C2,$52,$4F,$F9,$D7,$EA
	.dc.b	$3E,$34,$EA,$3F,$C1,$CD,$7F,$C6,$8D,$7F,$C7,$4D,$7E,$48,$78,$6F
	.dc.b	$FC,$68,$4B,$E7,$3F,$DF,$CD,$AF,$F1,$F5,$70,$D3,$E9,$DB,$B9,$DB
	.dc.b	$FE,$3D,$25,$FF,$8F,$7B,$2B,$FE,$3D,$8B,$5F,$E4,$59,$FF,$1D,$79
	.dc.b	$1F,$E3,$BF,$26,$5F,$F8,$E6,$32,$FF,$8E,$E3,$2F,$F8,$E6,$3F,$23
	.dc.b	$E1,$C8,$FB,$EF,$E3,$BE,$37,$11,$FE,$59,$E4,$FF,$60,$3F,$E3,$76
	.dc.b	$B9,$9F,$C5,$C6,$31,$F8,$DC,$F7,$5F,$82,$DB,$FC,$A5,$FF,$8D,$B4
	.dc.b	$91,$FE,$34,$7F,$C7,$3A,$62,$3F,$18,$7F,$8D,$A3,$18,$7F,$90,$67
	.dc.b	$FC,$63,$9F,$F8,$CB,$49,$7F,$E3,$28,$CB,$FE,$34,$8C,$BF,$E3,$B8
	.dc.b	$FF,$6A,$77,$BE,$22,$92,$7F,$19,$B5,$7F,$CF,$93,$B7,$FC,$B4,$D7
	.dc.b	$33,$FA,$E8,$C6,$3F,$89,$E1,$FC,$25,$7F,$00,$6B,$FE,$6C,$AF,$F7
	.dc.b	$37,$D5,$3E,$C3,$F6,$ED,$9F,$F3,$43,$5F,$F4,$73,$5F,$F4,$A2,$FB
	.dc.b	$1E,$24,$8E,$2F,$4F,$9A,$91,$D1,$AF,$FA,$67,$F1,$4F,$F7,$6B,$7D
	.dc.b	$2F,$E1,$31,$88,$FF,$51,$35,$FD,$4B,$FE,$FD,$7F,$CF,$CD,$7A,$7A
	.dc.b	$7C,$8E,$7F,$31,$23,$6D,$7F,$D5,$5F,$32,$F5,$13,$FE,$89,$8C,$37
	.dc.b	$D4,$7D,$1E,$6B,$77,$FE,$B3,$FA,$B6,$9F,$E0,$A8,$FD,$FC,$63,$AF
	.dc.b	$FD,$28,$D8,$E1,$3B,$7F,$96,$FD,$6B,$E8,$31,$08,$C5,$FF,$A7,$5B
	.dc.b	$1F,$EC,$36,$BF,$EC,$9F,$BC,$7C,$4F,$26,$D7,$FA,$95,$B3,$FE,$CB
	.dc.b	$6B,$C7,$C4,$E0,$49,$E0,$F9,$29,$D5,$49,$3F,$EB,$18,$CB,$FE,$B3
	.dc.b	$F9,$87,$D6,$FB,$1B,$19,$8C,$19,$FA,$E9,$1C,$5B,$8F,$FB,$86,$9F
	.dc.b	$5F,$CC,$95,$DD,$F9,$3F,$34,$F0,$29,$27,$FD,$8D,$19,$6F,$D3,$F0
	.dc.b	$23,$83,$7F,$45,$D0,$8E,$0F,$F6,$87,$98,$FA,$7F,$B2,$D8,$CC,$60
	.dc.b	$8F,$FB,$E6,$35,$9F,$F1,$4F,$0D,$F3,$AF,$A8,$FE,$8D,$7F,$D2,$AF
	.dc.b	$75,$FE,$58,$6C,$67,$0D,$E1,$FE,$02,$C3,$94,$F7,$9C,$CB,$E7,$9F
	.dc.b	$23,$F9,$06,$D7,$FB,$AD,$B3,$FD,$68,$FF,$01,$B8,$70,$EB,$1F,$F3
	.dc.b	$5B,$67,$FA,$BF,$CF,$7F,$5D,$7F,$AC,$9F,$D9,$23,$55,$FE,$0E,$F9
	.dc.b	$E7,$CD,$7A,$B0,$7C,$C3,$E2,$3D,$35,$AC,$79,$E1,$FE,$20,$AD,$46
	.dc.b	$21,$3B,$7F,$A0,$98,$C3,$9B,$D1,$F8,$B8,$72,$8D,$AF,$9E,$CE,$3A
	.dc.b	$6E,$1D,$E7,$81,$FB,$DF,$9E,$FF,$14,$FC,$EB,$E5,$FF,$C0,$73,$AA
	.dc.b	$F9,$98,$AF,$75,$AC,$7F,$D3,$F8,$71,$AF,$A5,$5A,$7F,$C0,$73,$AA
	.dc.b	$E7,$39,$9F,$F2,$1C,$E5,$E2,$2B,$A3,$AB,$E0,$FF,$90,$2B,$96,$1F
	.dc.b	$0E,$B7,$39,$39,$61,$E7,$BE,$31,$EB,$32,$79,$9C,$DF,$4C,$3C,$93
	.dc.b	$FE,$26,$8C,$4E,$A3,$89,$E5,$29,$E1,$3E,$49,$DC,$B1,$FE,$54,$70
	.dc.b	$7F,$8C,$BE,$75,$F1,$8F,$9E,$E3,$18,$6C,$66,$7E,$AC,$E5,$BF,$2F
	.dc.b	$FE,$6E,$6E,$9C,$A3,$11,$FA,$79,$46,$23,$C4,$F1,$D3,$E8,$E9,$FA
	.dc.b	$09,$FB,$0D,$7E,$0D,$72,$BF,$B8,$D8,$F8,$B3,$BA,$E7,$B7,$F9,$41
	.dc.b	$FB,$B5,$B6,$E1,$E1,$AF,$F2,$A9,$C4,$B5,$C5,$F3,$E1,$E0,$9E,$18
	.dc.b	$7F,$A1,$27,$83,$6B,$83,$FE,$88,$C3,$C1,$3C,$A3,$CF,$3A,$7F,$48
	.dc.b	$D9,$DC,$73,$3F,$75,$3F,$0E,$FF,$DE,$ED,$D3,$A6,$3F,$02,$B7,$87
	.dc.b	$EB,$A7,$D7,$1F,$C0,$F0,$EA,$BE,$7B,$F8,$63,$FC,$1D,$AF,$F0,$66
	.dc.b	$E2,$AF,$F3,$DC,$7D,$C9,$E1,$B7,$77,$FD,$4F,$1C,$27,$73,$99,$F0
	.dc.b	$67,$7E,$77,$FC,$F5,$1F,$9A,$6B,$FC,$11,$AF,$F3,$16,$BF,$E8,$C7
	.dc.b	$78,$7F,$A2,$FE,$7B,$F9,$3C,$FF,$24,$84,$62,$3F,$90,$CF,$FB,$04
	.dc.b	$C4,$FB,$63,$FD,$7F,$18,$C3,$B7,$1F,$B6,$B9,$F8,$9E,$98,$FF,$60
	.dc.b	$D7,$5B,$F3,$46,$08,$F8,$F0,$FF,$0A,$B7,$C4,$E4,$FF,$37,$A7,$CF
	.dc.b	$7C,$52,$47,$F3,$BF,$DC,$DE,$7E,$23,$E3,$4F,$88,$8F,$CB,$7E,$77
	.dc.b	$C4,$48,$ED,$1F,$47,$D6,$F4,$92,$3E,$93,$EC,$F8,$89,$1F,$55,$AF
	.dc.b	$D5,$F8,$0F,$82,$47,$ED,$3F,$13,$D0,$91,$F7,$5A,$FD,$DF,$9F,$F9
	.dc.b	$24,$7F,$04,$FC,$CE,$99,$1F,$85,$6B,$FD,$23,$F1,$1F,$F7,$9A,$69
	.dc.b	$33,$FF,$7A,$FC,$CB,$E7,$E9,$2B,$F9,$4C,$65,$BC,$FC,$08,$E0,$FE
	.dc.b	$5F,$F6,$0F,$84,$FF,$80,$2D,$8C,$C6,$0F,$E6,$AB,$E4,$1F,$35,$86
	.dc.b	$C7,$F3,$A6,$BF,$D9,$1A,$FF,$84,$57,$CF,$71,$24,$71,$7A,$7A,$52
	.dc.b	$3A,$35,$FE,$DF,$F3,$BF,$20,$9A,$F9,$9C,$71,$D2,$4F,$F4,$E8,$47
	.dc.b	$B3,$E5,$9F,$37,$FB,$D6,$D7,$F8,$45,$AF,$F8,$F7,$E5,$FF,$D0,$EE
	.dc.b	$67,$91,$1F,$EB,$73,$C1,$FF,$18,$3C,$BC,$0F,$DB,$0A,$FD,$5C,$F5
	.dc.b	$ED,$F1,$5F,$ED,$E7,$FD,$CD,$5B,$C3,$06,$7A,$26,$1F,$C1,$EB,$DA
	.dc.b	$D9,$CB,$E9,$F8,$0E,$84,$78,$78,$D5,$D1,$CE,$3C,$51,$FC,$87,$E7
	.dc.b	$BF,$92,$48,$38,$6F,$8B,$FC,$3F,$EB,$5F,$2D,$C6,$3F,$17,$5F,$CD
	.dc.b	$AB,$FC,$A8,$D7,$A1,$C8,$FF,$02,$BE,$0D,$17,$FE,$AF,$9F,$E4,$CD
	.dc.b	$7F,$C8,$B3,$B8,$9E,$77,$8C,$F2,$5C,$AE,$31,$97,$9C,$7F,$27,$F9
	.dc.b	$FF,$CB,$C7,$F3,$36,$BF,$E7,$F6,$E1,$F7,$1F,$D0,$B7,$97,$91,$27
	.dc.b	$97,$3F,$F4,$BE,$B7,$CC,$52,$4F,$F9,$87,$E6,$7F,$DC,$D8,$7F,$3D
	.dc.b	$85,$70,$7F,$80,$D7,$F0,$62,$38,$39,$7C,$5F,$F5,$9B,$FF,$7C,$DF
	.dc.b	$3B,$E5,$5E,$BF,$7B,$E0,$7C,$8C,$8F,$81,$1F,$0C,$C3,$F0,$63,$0F
	.dc.b	$C5,$F2,$3E,$41,$1F,$91,$F1,$15,$F9,$11,$70,$F9,$13,$FD,$B7,$3F
	.dc.b	$93,$E7,$FE,$09,$B7,$F8,$F6,$AC,$FD,$37,$F1,$67,$F8,$7C,$F1,$D7
	.dc.b	$F8,$CC,$62,$39,$46,$0F,$89,$C1,$23,$FD,$21,$3E,$0A,$7C,$FF,$FB
	.dc.b	$A2,$7F,$BA,$3F,$DE,$31,$FD,$14,$EB,$E5,$BF,$B7,$4E,$EB,$FD,$88
	.dc.b	$D9,$FE,$02,$F1,$84,$71,$7A,$FA,$3C,$BB,$24,$78,$1E,$BD,$78,$8E
	.dc.b	$B1,$6B,$AC,$FC,$BF,$99,$95,$EC,$7F,$05,$EB,$62,$3F,$D5,$7A,$27
	.dc.b	$17,$F8,$FD,$FF,$CD,$F5,$C4,$87,$09,$E6,$E2,$7E,$94,$72,$D7,$F7
	.dc.b	$75,$CA,$B8,$1B,$8F,$BD,$3C,$A3,$24,$AE,$6F,$E8,$F0,$E8,$FE,$0E
	.dc.b	$6B,$F8,$38,$B8,$7A,$E3,$B1,$FC,$5F,$0F,$15,$F4,$8E,$D8,$78,$9E
	.dc.b	$C7,$03,$F6,$B8,$76,$AE,$33,$B9,$E2,$78,$CE,$26,$0E,$27,$38,$E2
	.dc.b	$EC,$D5,$F9,$DF,$F6,$EC,$69,$E2,$6A,$BE,$35,$FA,$49,$7E,$37,$EC
	.dc.b	$61,$F0,$A7,$D0,$DA,$F1,$35,$F3,$5F,$7C,$3C,$57,$D2,$7C,$CF,$47
	.dc.b	$E2,$46,$B0,$F0,$1F,$7A,$FA,$57,$1A,$FA,$87,$0C,$3C,$6D,$3C,$65
	.dc.b	$3C,$73,$F9,$57,$FA,$19,$F0,$8B,$7F,$1D,$B3,$FB,$E2,$3F,$EC,$C8
	.dc.b	$D3,$F5,$CD,$9F,$CB,$7B,$FF,$92,$1E,$8C,$3F,$2E,$E2,$3C,$98,$7E
	.dc.b	$8E,$31,$1F,$4C,$E4,$43,$07,$DA,$77,$1F,$0E,$BE,$44,$FE,$4E,$73
	.dc.b	$5F,$2A,$3E,$7B,$E2,$7A,$57,$1B,$DC,$71,$3C,$B3,$B7,$C0,$60,$D1
	.dc.b	$83,$CB,$3E,$B7,$AC,$78,$5F,$C7,$C6,$5E,$0F,$85,$D9,$D1,$C3,$A8
	.dc.b	$F1,$46,$0E,$44,$7B,$5F,$C2,$9F,$E7,$57,$EC,$7C,$33,$E3,$C7,$FC
	.dc.b	$50,$F4,$38,$E1,$B8,$CC,$FC,$82,$1D,$0F,$39,$D0,$CB,$C1,$D1,$F1
	.dc.b	$27,$B3,$ED,$7B,$BC,$A3,$85,$FA,$4F,$53,$CC,$CD,$7F,$07,$BF,$64
	.dc.b	$FE,$03,$06,$DF,$E4,$1C,$3F,$D7,$CD,$7F,$A1,$E3,$FD,$32,$31,$3F
	.dc.b	$E4,$87,$FC,$E5,$3C,$3B,$FF,$D7,$91,$AA,$E1,$18,$F1,$7F,$D8,$1F
	.dc.b	$3E,$F0,$EF,$CA,$CC,$7F,$D9,$D3,$96,$CE,$5F,$19,$FD,$5A,$13,$D2
	.dc.b	$7F,$BC,$7C,$FC,$AF,$ED,$2F,$F8,$3F,$E5,$F1,$1F,$BA,$3F,$C3,$F1
	.dc.b	$FD,$D0,$8F,$ED,$DF,$F0,$B7,$95,$23,$99,$FE,$E3,$F9,$FF,$73,$FD
	.dc.b	$A1,$BB,$1A,$7C,$68,$F4,$C6,$74,$FF,$71,$4E,$9B,$A7,$FB,$97,$E7
	.dc.b	$FE,$15,$F0,$7D,$EF,$FB,$CD,$AE,$73,$E0,$71,$D2,$46,$E3,$FC,$39
	.dc.b	$F3,$F2,$3F,$C1,$AF,$F8,$CF,$0C,$47,$C5,$E1,$4F,$F8,$60,$91,$FC
	.dc.b	$6D,$FF,$1D,$63,$23,$99,$F8,$AF,$1F,$F0,$64,$7F,$1B,$F3,$F5,$F8
	.dc.b	$7C,$7F,$E7,$47,$81,$A3,$F0,$7F,$0F,$F8,$8E,$69,$18,$21,$F3,$39
	.dc.b	$3F,$38,$49,$FF,$3C,$F0,$FF,$95,$FE,$75,$E7,$E3,$9F,$FC,$FB,$18
	.dc.b	$C3,$FC,$53,$18,$9D,$47,$F8,$93,$E7,$7F,$C4,$CA,$BF,$DC,$E4,$9F
	.dc.b	$F6,$AB,$63,$99,$C7,$B3,$9E,$0B,$FF,$52,$CF,$E7,$0E,$97,$F8,$4D
	.dc.b	$A4,$7E,$3D,$FC,$3E,$D8,$8F,$C5,$F3,$7C,$96,$CE,$9F,$C3,$69,$F8
	.dc.b	$BD,$A4,$60,$87,$E3,$5C,$BF,$8F,$8F,$82,$BE,$79,$F2,$58,$6C,$7B
	.dc.b	$8D,$47,$BB,$E7,$BE,$3C,$EA,$7F,$60,$D7,$FB,$5B,$5F,$F5,$7B,$5F
	.dc.b	$F5,$E3,$C7,$5F,$F6,$BC,$BE,$7F,$B1,$E3,$BF,$3F,$DF,$7F,$AE,$70
	.dc.b	$FF,$62,$35,$34,$C7,$CF,$3E,$7F,$F8,$24,$FF,$B2,$E3,$0E,$1A,$FF
	.dc.b	$09,$6B,$FE,$04,$5F,$01,$CA,$11,$87,$94,$66,$31,$1F,$C4,$FE,$17
	.dc.b	$F2,$4D,$E2,$E5,$18,$8F,$F0,$8B,$5F,$F6,$AF,$CF,$BD,$1F,$03,$63
	.dc.b	$D6,$D7,$F9,$03,$5F,$F0,$DA,$F9,$FE,$24,$8E,$2F,$AF,$EA,$9F,$2D
	.dc.b	$49,$3F,$22,$32,$FC,$AF,$A0,$7A,$2F,$A0,$C3,$67,$F9,$94,$BD,$37
	.dc.b	$BB,$AB,$FA,$7F,$40,$F8,$8F,$F6,$86,$1F,$EF,$48,$C7,$85,$A9,$CC
	.dc.b	$84,$E9,$FE,$81,$F2,$2F,$83,$DC,$23,$17,$F8,$26,$C7,$09,$DF,$CF
	.dc.b	$3E,$0E,$92,$FF,$13,$19,$6F,$57,$02,$38,$3F,$1F,$E0,$3D,$0F,$1B
	.dc.b	$63,$31,$83,$CF,$F3,$8F,$9E,$A4,$9F,$CC,$C6,$5B,$F3,$DC,$08,$E0
	.dc.b	$FC,$FF,$DE,$FC,$76,$EB,$F1,$23,$0F,$F6,$76,$BF,$E6,$2F,$9F,$7C
	.dc.b	$0F,$CA,$6D,$7E,$A9,$B3,$FD,$BD,$AF,$1E,$9E,$04,$9E,$0F,$D4,$F1
	.dc.b	$DF,$38,$9F,$E5,$7E,$3B,$E7,$55,$9C,$3F,$95,$3C,$2B,$73,$A7,$F8
	.dc.b	$B4,$66,$7F,$C0,$4D,$7F,$CF,$AD,$7F,$14,$76,$BF,$21,$DE,$FC,$11
	.dc.b	$9A,$EF,$F3,$D9,$78,$DF,$7A,$DF,$CE,$3E,$7B,$58,$7F,$BC,$3E,$79
	.dc.b	$F4,$B8,$85,$6F,$1F,$83,$5A,$7F,$98,$46,$67,$FC,$38,$D7,$FD,$2E
	.dc.b	$D7,$D6,$D7,$F4,$DF,$3E,$F9,$0F,$1E,$4F,$21,$F1,$1E,$3A,$D6,$3F
	.dc.b	$4A,$72,$FF,$41,$8C,$E1,$FD,$6D,$B7,$F1,$9F,$C5,$B7,$0F,$73,$CA
	.dc.b	$7F,$B1,$2F,$8E,$CE,$3A,$6E,$1F,$D3,$E7,$81,$FC,$03,$BF,$F8,$07
	.dc.b	$CF,$3E,$5F,$FC,$03,$87,$CB,$C7,$F8,$25,$6A,$73,$85,$9F,$F5,$EC
	.dc.b	$D5,$87,$97,$0E,$91,$A0,$E1,$DE,$1F,$93,$8C,$47,$80,$DE,$1E,$06
	.dc.b	$9C,$E3,$95,$77,$9F,$E2,$F3,$A3,$AE,$1E,$57,$A5,$70,$BF,$05,$70
	.dc.b	$8C,$56,$EB,$F8,$F1,$CE,$FB,$3E,$1A,$F2,$57,$A2,$36,$7F,$23,$3F
	.dc.b	$AE,$61,$FD,$75,$C9,$FA,$6F,$9E,$7C,$4B,$E7,$FF,$B5,$C6,$1B,$19
	.dc.b	$9F,$F3,$94,$E5,$BF,$2F,$F4,$1B,$97,$28,$C4,$7F,$BC,$A5,$18,$8F
	.dc.b	$13,$C7,$1F,$A9,$97,$E7,$67,$EE,$35,$F9,$95,$CA,$FF,$00,$D8,$FA
	.dc.b	$13,$BA,$E7,$B7,$F0,$47,$F0,$15,$B6,$E1,$E1,$AF,$E1,$27,$12,$D7
	.dc.b	$17,$CF,$87,$82,$78,$61,$FD,$66,$78,$36,$B8,$3F,$D7,$70,$F0,$4F
	.dc.b	$28,$F3,$CE,$9F,$D1,$36,$77,$1C,$CF,$F1,$C4,$FC,$3B,$FE,$36,$DD
	.dc.b	$3A,$63,$F0,$2B,$78,$7F,$91,$2B,$51,$E3,$C7,$D2,$BE,$7F,$F8,$4B
	.dc.b	$FE,$45,$6B,$FE,$5F,$6E,$3F,$CA,$9F,$B7,$1B,$DB,$FC,$A3,$1A,$CB
	.dc.b	$F9,$9B,$67,$FB,$F3,$C1,$FF,$35,$7C,$F7,$F9,$D8,$9F,$05,$FE,$70
	.dc.b	$F9,$FB,$3F,$E3,$E9,$FF,$70,$B6,$7F,$C7,$11,$88,$FF,$63,$3F,$C8
	.dc.b	$DF,$F2,$F3,$5E,$FC,$9D,$7E,$43,$FC,$CF,$39,$CF,$FA,$2B,$72,$FE
	.dc.b	$96,$DB,$FE,$1B,$8A,$F8,$E1,$E2,$24,$F9,$DF,$F2,$67,$CF,$FF,$89
	.dc.b	$E3,$FC,$AE,$D7,$FD,$B8,$D7,$FA,$C8,$70,$F8,$BE,$7F,$CE,$47,$FC
	.dc.b	$F5,$1C,$F2,$F4,$CF,$C4,$24,$FC,$97,$C7,$F1,$12,$3C,$6D,$7E,$4F
	.dc.b	$B7,$F3,$08,$FD,$17,$E7,$FC,$2C,$8F,$A8,$D7,$D7,$E0,$7C,$02,$3F
	.dc.b	$01,$FA,$FE,$8C,$8F,$B8,$D7,$EC,$FC,$FF,$DC,$23,$F8,$17,$FC,$ED
	.dc.b	$F4,$AC,$A9,$27,$FD,$07,$19,$7F,$D0,$BF,$41,$FE,$71,$7F,$C9,$CD
	.dc.b	$4D,$31,$F3,$2F,$A0,$FF,$42,$36,$7F,$D1,$ED,$7E,$B8,$70,$FD,$AF
	.dc.b	$A0,$AF,$F8,$51,$AF,$F9,$7D,$A9,$A7,$F9,$35,$FF,$4E,$B5,$EE,$D7
	.dc.b	$B7,$A3,$FE,$99,$F8,$E7,$CD,$52,$4F,$FA,$9A,$32,$DF,$3F,$81,$1C
	.dc.b	$1B,$F6,$3A,$11,$C1,$FE,$B0,$FA,$0F,$94,$4F,$84,$FF,$55,$3F,$37
	.dc.b	$EC,$9F,$33,$FE,$D4,$6D,$7F,$B4,$5A,$FF,$78,$EE,$FF,$41,$74,$BB
	.dc.b	$7F,$D9,$13,$BB,$FF,$5B,$5F,$08,$FD,$3B,$5F,$F2,$0B,$5F,$F4,$63
	.dc.b	$DA,$BF,$C8,$55,$FA,$E6,$CF,$96,$BF,$DF,$FF,$65,$F3,$E7,$FC,$90
	.dc.b	$FF,$BD,$DB,$F1,$DA,$E0,$97,$FE,$5E,$AD,$65,$EA,$6C,$7F,$85,$1E
	.dc.b	$91,$9F,$33,$B9,$27,$97,$8B,$5F,$88,$71,$8C,$63,$E4,$9F,$F7,$FF
	.dc.b	$57,$F9,$56,$BF,$DF,$8D,$78,$9F,$EB,$3F,$47,$DB,$F0,$DC,$1F,$F7
	.dc.b	$04,$EE,$7F,$88,$B5,$FF,$35,$B5,$FF,$5B,$3D,$6F,$FC,$D9,$F6,$4F
	.dc.b	$A3,$FF,$38,$C6,$1E,$D5,$FE,$82,$97,$D0,$6B,$4F,$BF,$1F,$E8,$49
	.dc.b	$D7,$AD,$97,$FC,$77,$1C,$1E,$55,$9C,$13,$C6,$7E,$3C,$62,$B5,$3C
	.dc.b	$67,$2F,$FB,$79,$F2,$7C,$46,$8D,$F3,$D8,$C9,$0E,$2F,$95,$7C,$CF
	.dc.b	$E3,$72,$FF,$48,$B6,$3F,$DB,$8D,$C5,$F3,$3F,$EE,$59,$FC,$FD,$70
	.dc.b	$E5,$FE,$17,$1C,$1A,$F6,$6D,$FE,$76,$30,$7F,$AB,$AB,$FD,$BC,$DF
	.dc.b	$85,$F3,$4F,$86,$5F,$3F,$F2,$57,$8F,$4C,$44,$BE,$91,$97,$B4,$71
	.dc.b	$F8,$99,$1E,$17,$A7,$A7,$88,$FA,$76,$9C,$69,$FE,$B3,$8F,$F0,$9C
	.dc.b	$FE,$27,$D2,$F9,$9C,$7F,$E0,$78,$C1,$FE,$CA,$FA,$3F,$D5,$B8,$3F
	.dc.b	$98,$C7,$F8,$59,$FE,$D2,$D7,$FD,$86,$D4,$71,$BF,$F6,$24,$BE,$9B
	.dc.b	$F1,$8D,$9F,$AA,$70,$AF,$F0,$D5,$26,$9F,$EA,$19,$F3,$75,$72,$C2
	.dc.b	$46,$08,$7E,$5A,$79,$FD,$23,$E9,$3F,$38,$D9,$FF,$74,$B5,$FC,$EB
	.dc.b	$FC,$E9,$B1,$C0,$8F,$0F,$1B,$F2,$DE,$43,$E7,$A9,$27,$DA,$3E,$8D
	.dc.b	$DF,$98,$FE,$13,$5C,$5D,$B5,$FE,$1B,$3C,$1C,$D7,$1D,$7F,$DD,$F5
	.dc.b	$B7,$FD,$E3,$5B,$AE,$B5,$BE,$8F,$F7,$CD,$ED,$FF,$7D,$D6,$EF,$9D
	.dc.b	$6E,$FE,$DD,$ED,$FB,$B5,$BB,$E7,$5B,$9F,$EE,$CF,$EE,$67,$51,$CA
	.dc.b	$35,$5C,$E3,$E2,$61,$89,$E5,$5C,$27,$99,$F0,$08,$7C,$49,$E0,$E3
	.dc.b	$0F,$83,$18,$8F,$82,$7D,$29,$E9,$7E,$12,$5F,$9E,$B8,$5F,$C2,$BE
	.dc.b	$11,$8C,$3D,$F1,$83,$C0,$E1,$EE,$79,$4C,$9F,$04,$A7,$C0,$2D,$F8
	.dc.b	$9E,$87,$2A,$F1,$36,$7E,$BD,$78,$E3,$A1,$DC,$D1,$89,$E2,$7C,$47
	.dc.b	$BD,$FC,$E9,$E5,$1F,$38,$E5,$3C,$2B,$D8,$D8,$FC,$0C,$F0,$BF,$B5
	.dc.b	$87,$CC,$C3,$FC,$76,$4A,$FF,$1D,$B5,$FF,$1D,$CF,$E7,$5F,$F3,$9B
	.dc.b	$5F,$94,$62,$3F,$15,$F2,$92,$3D,$27,$DB,$70,$FC,$03,$D5,$87,$E8
	.dc.b	$A3,$2F,$A6,$33,$8F,$E8,$27,$2F,$B8,$91,$F9,$E8,$CB,$F2,$E3,$2F
	.dc.b	$EE,$E3,$2F,$43,$47,$13,$CC,$7A,$0F,$DD,$BC,$2F,$F0,$11,$83,$67
	.dc.b	$AD,$F9,$B1,$BB,$FD,$2B,$67,$F1,$78,$23,$11,$F0,$E3,$0F,$4C,$3E
	.dc.b	$15,$F4,$8C,$61,$EC,$8C,$47,$8C,$F1,$61,$EE,$BE,$71,$83,$89,$EF
	.dc.b	$30,$71,$BF,$A4,$FC,$16,$B9,$BF,$05,$75,$C3,$F2,$77,$DA,$F6,$7A
	.dc.b	$27,$57,$CE,$FE,$D3,$E2,$3E,$91,$2B,$94,$6A,$3F,$4B,$4D,$1E,$4F
	.dc.b	$F4,$4E,$F9,$5F,$D1,$23,$FA,$39,$87,$FA,$2C,$61,$FE,$93,$F2,$3F
	.dc.b	$D3,$08,$FF,$4C,$EE,$AF,$F4,$C8,$B8,$31,$FF,$39,$E3,$FE,$27,$AF
	.dc.b	$F2,$5E,$7F,$D4,$6B,$51,$FD,$B2,$31,$5F,$DB,$A3,$FA,$86,$FF,$CB
	.dc.b	$E7,$FC,$50,$7F,$59,$8F,$F6,$FB,$5E,$0D,$7A,$47,$F8,$8A,$3A,$C7
	.dc.b	$F3,$EE,$2F,$E7,$47,$49,$FF,$22,$7D,$27,$FB,$D6,$7F,$BC,$E0,$87
	.dc.b	$B5,$A3,$7E,$CC,$AD,$CE,$9B,$39,$7F,$C9,$B1,$FE,$2A,$77,$D9,$E0
	.dc.b	$3C,$3D,$B6,$32,$43,$8B,$F0,$BE,$3A,$47,$F7,$2D,$11,$FD,$57,$E0
	.dc.b	$33,$3F,$EC,$E9,$CB,$FE,$46,$6B,$99,$F8,$C7,$F5,$DF,$0A,$47,$58
	.dc.b	$FF,$34,$65,$FE,$52,$8F,$8A,$D7,$FC,$D2,$DD,$3E,$E9,$F2,$3D,$0E
	.dc.b	$3D,$F2,$B8,$BD,$23,$FC,$E9,$F4,$1F,$E6,$49,$FA,$2B,$E8,$72,$6B
	.dc.b	$C4,$93,$A8,$CB,$C2,$3F,$0F,$F4,$72,$3E,$33,$FD,$E7,$E9,$31,$1F
	.dc.b	$16,$3F,$0C,$47,$E3,$3F,$DF,$FC,$F9,$1C,$8F,$C2,$FD,$07,$BF,$A3
	.dc.b	$FD,$A2,$D8,$6A,$8C,$37,$4F,$F4,$94,$60,$9A,$E2,$3F,$D2,$DF,$49
	.dc.b	$F0,$EF,$FD,$2A,$FC,$37,$F2,$AD,$73,$A7,$03,$8E,$B2,$37,$1F,$AA
	.dc.b	$F7,$E4,$7E,$25,$FF,$0E,$61,$88,$FC,$44,$7E,$A4,$8F,$E2,$5F,$F1
	.dc.b	$16,$12,$39,$1F,$E4,$4F,$BF,$7C,$6F,$FA,$49,$B1,$FE,$96,$6B,$FB
	.dc.b	$F3,$FC,$DA,$E1,$C7,$CF,$FF,$9C,$53,$49,$DB,$FE,$9E,$F9,$E7,$CF
	.dc.b	$D2,$57,$FA,$92,$32,$FF,$A9,$BE,$43,$F9,$2B,$7C,$5F,$F5,$3C,$62
	.dc.b	$3C,$81,$C3,$E7,$F8,$77,$BD,$FE,$AE,$6C,$7F,$AC,$9A,$F2,$6B,$EF
	.dc.b	$F9,$FF,$E9,$ED,$F2,$BF,$D6,$F1,$88,$FE,$44,$D7,$FD,$23,$E3,$BE
	.dc.b	$8F,$FD,$7A,$D8,$FF,$60,$B5,$F4,$B5,$F9,$DF,$40,$F9,$6F,$F6,$2B
	.dc.b	$63,$FD,$92,$D7,$E0,$B5,$FA,$DF,$40,$F3,$1F,$41,$FE,$D0,$8C,$37
	.dc.b	$0F,$F5,$C5,$7C,$A9,$CB,$F3,$3E,$B9,$F1,$7F,$D0,$A3,$F0,$51,$8F
	.dc.b	$03,$FD,$B6,$D8,$E1,$3B,$7F,$A3,$7C,$AB,$DD,$C4,$7F,$AD,$A3,$17
	.dc.b	$FE,$E4,$6C,$7D,$C6,$BF,$91,$FA,$07,$BD,$CA,$11,$8A,$FF,$76,$B6
	.dc.b	$3F,$0A,$D7,$F3,$5F,$59,$FD,$1A,$7F,$DE,$6D,$8F,$C5,$B5,$DF,$95
	.dc.b	$DF,$CB,$79,$14,$97,$FE,$FC,$8C,$BF,$EF,$EF,$9F,$7D,$07,$9D,$B1
	.dc.b	$98,$C1,$EB,$F3,$1C,$54,$93,$F9,$EC,$93,$A8,$C7,$CF,$FF,$B8,$1B
	.dc.b	$3F,$16,$30,$FE,$99,$AF,$EE,$FE,$7F,$FC,$23,$3D,$5B,$1F,$AC,$6B
	.dc.b	$BE,$FF,$A1,$F0,$AF,$9E,$FE,$93,$7F,$EF,$0F,$3D,$F3,$EA,$CE,$1F
	.dc.b	$EF,$47,$85,$7F,$78,$9D,$3F,$ED,$68,$CC,$FF,$34,$8E,$93,$A8,$C1
	.dc.b	$CE,$FA,$DF,$82,$FB,$CF,$86,$7C,$E4,$BF,$95,$3D,$A7,$F8,$E7,$CF
	.dc.b	$BE,$2B,$9F,$8B,$FC,$92,$FF,$DD,$33,$97,$FC,$1F,$19,$AF,$18,$9D
	.dc.b	$3E,$6C,$7F,$8A,$79,$7E,$49,$CC,$7F,$70,$AE,$4F,$EB,$7E,$7D,$F1
	.dc.b	$3E,$CF,$1D,$E5,$FB,$27,$39,$FF,$88,$27,$2F,$FB,$FA,$33,$9F,$91
	.dc.b	$B8,$75,$8E,$58,$7C,$68,$E3,$1E,$A5,$F0,$D9,$C7,$4D,$C3,$EF,$1F
	.dc.b	$E4,$19,$E2,$7E,$F3,$B9,$6F,$9D,$7C,$EE,$F1,$FA,$B3,$FC,$04,$D4
	.dc.b	$67,$0E,$35,$C2,$7F,$A8,$5E,$2B,$51,$C0,$C9,$4F,$EF,$77,$FE,$58
	.dc.b	$AE,$51,$FC,$48,$D6,$1F,$02,$33,$1E,$37,$71,$9A,$FE,$25,$3C,$2B
	.dc.b	$A4,$62,$B7,$18,$9F,$1C,$63,$0F,$44,$62,$BD,$33,$E0,$86,$1A,$7D
	.dc.b	$35,$5F,$8C,$7F,$25,$3F,$81,$1C,$0B,$1F,$C1,$7E,$79,$EE,$BE,$7F
	.dc.b	$0D,$AF,$EC,$F3,$A9,$C9,$C7,$E4,$3F,$C4,$73,$A6,$BF,$E0,$19,$46
	.dc.b	$23,$FD,$05,$18,$D3,$91,$EB,$CF,$F3,$73,$76,$FF,$44,$69,$F5,$2F
	.dc.b	$84,$FD,$96,$C7,$D8,$AE,$33,$F7,$5B,$1C,$70,$F3,$B7,$0E,$D5,$0E
	.dc.b	$93,$AA,$F3,$B6,$3F,$2F,$1D,$0F,$F4,$E1,$5F,$CD,$DF,$92,$38,$D7
	.dc.b	$5B,$F3,$61,$EC,$9C,$DF,$9E,$3F,$4D,$A7,$46,$C7,$81,$AB,$0E,$CE
	.dc.b	$4F,$D7,$94,$FE,$1F,$5F,$13,$0F,$EE,$93,$C2,$5F,$3F,$FC,$21,$E2
	.dc.b	$63,$6E,$F1,$98,$FE,$32,$FD,$C7,$6E,$6B,$FC,$00,$DC,$7F,$D8,$0D
	.dc.b	$9F,$DF,$BC,$5F,$E4,$7F,$3C,$F7,$3D,$F6,$F9,$57,$CF,$61,$B7,$E6
	.dc.b	$BE,$0D,$BD,$54,$8C,$1B,$F7,$7F,$47,$5F,$E7,$18,$FF,$2B,$35,$FF
	.dc.b	$6B,$35,$FE,$2A,$FF,$9D,$F0,$FF,$50,$E0,$F9,$4F,$21,$F5,$27,$59
	.dc.b	$F2,$C3,$38,$7F,$B4,$2F,$83,$E3,$3D,$5F,$17,$E4,$23,$FD,$E6,$D1
	.dc.b	$88,$F8,$FE,$A7,$73,$C1,$16,$38,$47,$B8,$8F,$3F,$A8,$F8,$24,$70
	.dc.b	$77,$F8,$3F,$8A,$47,$07,$8F,$C8,$F9,$24,$70,$7F,$83,$FE,$71,$F3
	.dc.b	$DF,$C4,$1B,$19,$8C,$1F,$4B,$E8,$64,$7B,$DA,$FF,$9D,$BE,$7D,$F6
	.dc.b	$9F,$C6,$9B,$1F,$E1,$D6,$BB,$8F,$E3,$FF,$45,$FE,$52,$8F,$F1,$13
	.dc.b	$5F,$F4,$0B,$5F,$F4,$B3,$5F,$EA,$61,$FE,$60,$FF,$56,$6B,$9E,$9E
	.dc.b	$E7,$83,$C5,$91,$E0,$6B,$FE,$90,$FA,$0F,$F4,$02,$69,$3E,$03,$F9
	.dc.b	$BA,$F9,$6A,$49,$FF,$1E,$46,$5B,$EE,$F0,$23,$83,$FC,$85,$F4,$1F
	.dc.b	$19,$FF,$5A,$35,$F7,$B5,$FE,$E6,$D4,$FD,$A7,$FD,$52,$D7,$9E,$DF
	.dc.b	$D3,$A3,$11,$D3,$EB,$30,$47,$C6,$FE,$17,$E1,$3F,$8D,$F0,$FF,$AB
	.dc.b	$3B,$3F,$D1,$0D,$EB,$FA,$7E,$03,$E2,$B7,$7C,$71,$FC,$83,$5F,$F5
	.dc.b	$ED,$F1,$7F,$C1,$B5,$FE,$66,$7F,$8E,$B5,$FF,$19,$C6,$F0,$7C,$B7
	.dc.b	$F9,$4E,$7F,$95,$C6,$3D,$0F,$E2,$B3,$99,$FF,$1C,$2F,$67,$33,$27
	.dc.b	$8C,$EE,$99,$FF,$98,$63,$84,$EB,$DF,$EB,$D3,$23,$04,$3B,$B9,$7C
	.dc.b	$31,$F2,$B9,$64,$7F,$3A,$CF,$FD,$E7,$8F,$FA,$CF,$46,$1A,$AF,$8F
	.dc.b	$7F,$B7,$23,$FB,$B7,$FA,$7C,$FF,$A5,$23,$E4,$F5,$71,$BF,$5C,$17
	.dc.b	$CA,$7F,$1E,$8F,$F2,$E3,$63,$66,$BC,$3F,$E7,$F1,$FC,$F2,$77,$18
	.dc.b	$23,$F4,$A3,$53,$C9,$FB,$03,$8B,$99,$D3,$86,$57,$18,$CB,$CE,$3F
	.dc.b	$B5,$FC,$C3,$E8,$A9,$27,$F0,$FE,$0D,$7F,$82,$9F,$E5,$89,$E0,$FF
	.dc.b	$AB,$DA,$E8,$38,$78,$BA,$C7,$EA,$78,$5F,$38,$EC,$FB,$7E,$93,$E2
	.dc.b	$FD,$34,$BF,$85,$1F,$20,$C3,$F3,$A3,$0F,$E6,$3E,$07,$FD,$F6,$DE
	.dc.b	$CF,$ED,$B1,$88,$FF,$48,$3F,$C2,$DB,$1A,$23,$CF,$BB,$E3,$FB,$D2
	.dc.b	$3F,$C0,$F9,$FF,$B3,$9F,$D6,$F9,$39,$7E,$14,$65,$B3,$FE,$54,$8C
	.dc.b	$47,$C4,$8F,$C7,$FC,$94,$8E,$B1,$FE,$04,$FA,$96,$97,$49,$5F,$B6
	.dc.b	$9C,$B6,$72,$FF,$AB,$61,$3C,$E7,$FC,$1F,$F2,$9F,$0F,$6F,$CA,$B7
	.dc.b	$B7,$FD,$AC,$DC,$3F,$3C,$FC,$79,$DE,$DC,$89,$3F,$C2,$27,$55,$FD
	.dc.b	$33,$4F,$F1,$05,$FE,$A2,$BF,$0B,$E9,$FE,$48,$FD,$7F,$CE,$D8,$F9
	.dc.b	$B1,$5F,$F6,$4F,$52,$71,$5C,$EF,$AE,$1D,$F0,$E7,$87,$22,$1F,$ED
	.dc.b	$29,$DB,$FC,$9A,$F9,$D7,$18,$DD,$FF,$09,$9C,$C7,$E8,$6E,$BF,$ED
	.dc.b	$08,$B7,$D2,$3D,$A7,$F3,$08,$CC,$F0,$C3,$D5,$3D,$B0,$F0,$3E,$59
	.dc.b	$F4,$B4,$D1,$E1,$3A,$18,$3B,$9B,$3E,$3D,$78,$1D,$57,$81,$D6,$1F
	.dc.b	$26,$BC,$95,$B9,$CD,$FF,$52,$8C,$56,$E7,$C2,$68,$F2,$14,$F8,$87
	.dc.b	$97,$0F,$88,$7C,$5C,$3C,$53,$CE,$78,$4E,$6B,$BD,$FC,$4A,$FA,$55
	.dc.b	$CF,$0F,$3B,$4F,$6D,$70,$7C,$73,$C1,$EB,$3C,$1A,$70,$3F,$0A,$7F
	.dc.b	$16,$7F,$6C,$47,$F8,$F3,$5F,$DE,$B5,$E0,$7F,$6E,$F9,$CF,$E9,$27
	.dc.b	$8D,$62,$8C,$C7,$A4,$FC,$EB,$87,$CA,$60,$E8,$6E,$BF,$AE,$46,$DF
	.dc.b	$94,$4C,$3F,$C1,$31,$97,$E5,$12,$7F,$C0,$B1,$97,$E3,$C6,$5F,$3C
	.dc.b	$65,$FC,$6C,$F5,$78,$9E,$43,$CE,$F9,$1C,$3B,$78,$CF,$91,$CB,$DF
	.dc.b	$1F,$C2,$57,$77,$F0,$F8,$78,$70,$F9,$17,$F0,$6F,$93,$F1,$4F,$0B
	.dc.b	$8C,$3C,$51,$83,$CE,$6D,$F2,$5F,$82,$3C,$67,$C2,$21,$F0,$8C,$9E
	.dc.b	$A3,$35,$E4,$7A,$57,$AE,$FD,$18,$7E,$06,$B5,$87,$90,$F0,$CE,$AF
	.dc.b	$D8,$79,$A3,$57,$CE,$3C,$35,$CA,$3E,$D4,$FF,$84,$3E,$87,$FC,$17
	.dc.b	$EA,$CB,$FF,$04,$47,$F8,$38,$C3,$FE,$14,$8C,$3F,$E3,$EF,$98,$FD
	.dc.b	$9D,$A3,$FC,$39,$18,$6E,$9F,$ED,$68,$F1,$46,$0E,$51,$93,$59,$FF
	.dc.b	$BE,$3E,$A9,$F1,$94,$95,$FD,$67,$D5,$AF,$F8,$26,$76,$FF,$82,$C3
	.dc.b	$87,$53,$97,$2F,$F6,$7E,$E9,$1F,$EF,$C9,$E1,$AF,$F8,$CB,$95,$B7
	.dc.b	$F6,$DD,$A5,$FF,$6B,$8F,$EE,$06,$1F,$ED,$51,$87,$FC,$CB,$EB,$7C
	.dc.b	$5B,$FF,$7D,$6B,$FE,$FA,$8C,$1F,$1A,$33,$3C,$27,$46,$F5,$FF,$58
	.dc.b	$CF,$F6,$5F,$4B,$FC,$79,$5F,$4E,$31,$A6,$B8,$3F,$D6,$F5,$AA,$FF
	.dc.b	$29,$46,$FE,$A7,$FC,$8E,$EA,$BF,$D9,$44,$9F,$F6,$53,$5E,$27,$F9
	.dc.b	$6E,$5E,$8F,$78,$C3,$55,$F8,$70,$FF,$2E,$F7,$7F,$8F,$71,$F4,$78
	.dc.b	$7F,$E4,$89,$FF,$2C,$5F,$F8,$CB,$C2,$8F,$F9,$74,$FF,$31,$98,$3C
	.dc.b	$70,$E5,$6F,$F9,$12,$4F,$E6,$7E,$8A,$BC,$27,$6F,$E7,$67,$F3,$13
	.dc.b	$F9,$3F,$42,$4F,$E1,$5F,$F1,$77,$8D,$88,$FC,$94,$7E,$37,$9B,$FD
	.dc.b	$7B,$1F,$A5,$6B,$F9,$48,$FF,$1B,$78,$D2,$3A,$1F,$E2,$3C,$FF,$C7
	.dc.b	$4C,$30,$43,$FC,$6C,$64,$F8,$3D,$BF,$2D,$FF,$47,$65,$EC,$7E,$2B
	.dc.b	$74,$FF,$46,$B7,$6F,$C4,$2D,$3F,$28,$CA,$FC,$CB,$F9,$78,$FB,$9E
	.dc.b	$13,$E1,$69,$2B,$FD,$87,$C9,$FE,$9B,$C3,$F8,$34,$2B,$83,$FB,$76
	.dc.b	$BF,$E8,$36,$BF,$E8,$A6,$BC,$9D,$E3,$FE,$A6,$AF,$F5,$CD,$7E,$6F
	.dc.b	$C9,$F1,$5F,$E5,$87,$D3,$7F,$B9,$5F,$E0,$CD,$7F,$D2,$4D,$7F,$D3
	.dc.b	$2D,$7D,$2B,$EA,$3F,$85,$36,$39,$C6,$0C,$D4,$8E,$ED,$7F,$D4,$1F
	.dc.b	$2E,$F9,$FF,$E2,$8D,$8F,$E3,$4D,$77,$1A,$FA,$F9,$1A,$6B,$D3,$ED
	.dc.b	$79,$91,$E6,$FF,$21,$9F,$F7,$5C,$7F,$AC,$5A,$FF,$AE,$1A,$FD,$86
	.dc.b	$BF,$CF,$DF,$88,$DF,$5B,$A1,$1F,$E6,$5F,$59,$FE,$8E,$7F,$D7,$CD
	.dc.b	$7F,$D8,$AD,$7F,$0A,$D7,$FA,$67,$CC,$3E,$67,$FD,$65,$1F,$8A,$8C
	.dc.b	$79,$7F,$CF,$D7,$C9,$FF,$36,$9F,$F6,$64,$69,$FF,$67,$7D,$63,$B7
	.dc.b	$FD,$19,$1F,$94,$8C,$6F,$FD,$25,$B1,$C2,$76,$FF,$5F,$FA,$D7,$D5
	.dc.b	$62,$11,$8B,$FE,$A4,$D8,$FF,$6F,$35,$FF,$72,$FC,$BB,$D7,$E4,$DA
	.dc.b	$FE,$B4,$D9,$FF,$73,$35,$E3,$F0,$BC,$09,$3C,$1E,$7F,$0B,$23,$9B
	.dc.b	$67,$FD,$E1,$F2,$CF,$81,$F5,$36,$BF,$B5,$36,$77,$3A,$F8,$19,$5A
	.dc.b	$6C,$F4,$F7,$39,$92,$79,$BF,$1E,$7F,$67,$1F,$DD,$DB,$3F,$C1,$8F
	.dc.b	$F0,$63,$87,$0D,$7F,$85,$BC,$3B,$BF,$D8,$CD,$9F,$EC,$5F,$2A,$E7
	.dc.b	$47,$F6,$53,$8F,$6E,$30,$FE,$0D,$7C,$5F,$0C,$67,$09,$87,$3B,$E5
	.dc.b	$5A,$6A,$F9,$47,$8F,$F8,$56,$CF,$F8,$55,$B3,$EF,$38,$4E,$CF,$04
	.dc.b	$FF,$28,$9D,$DF,$8F,$E5,$3F,$B7,$CF,$82,$72,$FF,$2E,$F9,$37,$A9
	.dc.b	$CB,$16,$0F,$94,$FF,$46,$DF,$A2,$F6,$FB,$27,$5E,$D7,$1B,$56,$9F
	.dc.b	$E5,$8D,$9F,$F7,$B3,$FE,$43,$AE,$98,$71,$9E,$F5,$FC,$CD,$7C,$A6
	.dc.b	$70,$D3,$6F,$E5,$CF,$03,$F6,$3F,$29,$FE,$4F,$F9,$2F,$F0,$7E,$1F
	.dc.b	$E1,$FC,$3C,$CF,$F9,$4E,$75,$EB,$7F,$89,$A7,$EF,$57,$0A,$FE,$A7
	.dc.b	$5A,$79,$61,$5C,$89,$E1,$87,$A9,$E5,$1D,$CD,$61,$DE,$B9,$46,$67
	.dc.b	$8E,$1E,$2A,$FC,$06,$1D,$F0,$E9,$18,$BE,$67,$90,$A7,$92,$F9,$E1
	.dc.b	$DA,$17,$CF,$0E,$D1,$FA,$B8,$C4,$ED,$FF,$31,$47,$03,$FC,$C8,$E0
	.dc.b	$FF,$35,$7C,$93,$E5,$1F,$29,$86,$D7,$E6,$67,$53,$93,$8F,$C0,$FE
	.dc.b	$42,$74,$D7,$F7,$D2,$8C,$47,$F8,$1A,$31,$A7,$23,$EA,$69,$F9,$C9
	.dc.b	$BB,$7F,$83,$35,$FB,$97,$C2,$7E,$3B,$63,$F0,$15,$C6,$7E,$5B,$63
	.dc.b	$8E,$1E,$76,$E1,$DA,$A1,$D2,$75,$5E,$76,$C7,$FA,$8E,$3A,$1F,$E2
	.dc.b	$02,$BF,$EA,$7B,$F2,$47,$1A,$EB,$7E,$6C,$3D,$93,$9B,$F3,$C7,$F9
	.dc.b	$AE,$3D,$55,$FC,$66,$7F,$CE,$33,$FE,$6F,$BE,$B1,$E3,$8E,$04,$7F
	.dc.b	$8B,$C6,$97,$CA,$36,$FF,$3C,$36,$3F,$8E,$B5,$FF,$16,$3F,$65,$CC
	.dc.b	$70,$23,$C3,$2F,$F4,$17,$C9,$FF,$94,$E3,$F9,$F3,$5F,$F2,$2B,$5F
	.dc.b	$F6,$63,$53,$FE,$57,$7F,$8D,$FC,$A7,$FA,$61,$BF,$09,$DB,$D6,$EF
	.dc.b	$F1,$32,$7F,$9A,$DA,$3F,$D2,$DA,$B0,$7D,$06,$EF,$F9,$A5,$ED,$FF
	.dc.b	$66,$35,$CD,$FF,$B2,$E3,$17,$FD,$1E,$FF,$CC,$15,$E6,$F2,$3A,$4F
	.dc.b	$D7,$A7,$81,$F1,$09,$5F,$D2,$BE,$0D,$F2,$78,$8F,$1C,$F7,$23,$E4
	.dc.b	$7E,$6F,$91,$23,$DF,$1F,$42,$BC,$C4,$8F,$5B,$E9,$F1,$E4,$7B,$9A
	.dc.b	$FC,$2F,$1F,$DC,$47,$E1,$BF,$07,$DA,$91,$F1,$9A,$FF,$5D,$F8,$FF
	.dc.b	$F5,$09,$3D,$EF,$F7,$93,$FE,$A4,$D1,$F1,$FF,$2C,$95,$F4,$1F,$F7
	.dc.b	$C7,$CA,$7F,$8E,$63,$FD,$FC,$D7,$FB,$53,$5F,$F0,$53,$5F,$F4,$8A
	.dc.b	$F9,$CE,$64,$8E,$6F,$6E,$C9,$1D,$9A,$FF,$77,$F9,$5F,$F4,$5B,$FD
	.dc.b	$E4,$38,$75,$E5,$65,$CB,$63,$EE,$C7,$16,$BD,$78,$B4,$6F,$E4,$64
	.dc.b	$6D,$AF,$F8,$27,$E8,$FE,$4B,$E3,$6B,$9C,$FC,$51,$88,$FF,$17,$7C
	.dc.b	$B3,$E5,$FC,$CD,$8F,$63,$5F,$F0,$B3,$5E,$3C,$1C,$08,$F0,$7E,$07
	.dc.b	$C5,$9F,$F6,$0F,$47,$F1,$78,$C4,$FF,$8A,$1A,$FF,$BE,$BE,$93,$F7
	.dc.b	$8A,$3F,$DD,$24,$AE,$A4,$8F,$F0,$C4,$FF,$87,$5B,$E6,$78,$3C,$AF
	.dc.b	$F6,$8D,$F6,$C3,$FC,$79,$7F,$E3,$6B,$E3,$E7,$FF,$04,$8C,$47,$8F
	.dc.b	$E3,$71,$32,$36,$78,$23,$87,$C9,$F1,$F9,$49,$3E,$48,$CB,$E6,$8C
	.dc.b	$BE,$F8,$F3,$23,$FD,$E5,$E3,$61,$A7,$F9,$3B,$AF,$F9,$83,$FE,$3C
	.dc.b	$D3,$86,$6F,$AA,$7D,$17,$1B,$DD,$FF,$9B,$5A,$FF,$BD,$1A,$F8,$5A
	.dc.b	$F8,$0E,$C7,$F9,$2A,$7F,$DE,$37,$E0,$6C,$71,$C3,$F9,$6C,$79,$B0
	.dc.b	$FF,$35,$36,$B8,$4E,$BC,$FD,$B2,$FF,$CA,$55,$B7,$C1,$3D,$3E,$03
	.dc.b	$AF,$3C,$9D,$C6,$5E,$31,$97,$D3,$8F,$C5,$9F,$99,$DF,$F7,$A3,$FB
	.dc.b	$53,$72,$FA,$54,$8C,$3F,$57,$CB,$FF,$46,$3F,$E8,$08,$FC,$3B,$FE
	.dc.b	$99,$6B,$C9,$AF,$DD,$6B,$D7,$87,$EF,$4F,$FB,$6A,$3F,$D1,$6D,$7E
	.dc.b	$89,$B3,$FC,$ED,$F3,$4F,$63,$FA,$DB,$63,$31,$8F,$6B,$2C,$8C,$9C
	.dc.b	$A3,$8F,$CD,$72,$ED,$93,$DA,$32,$F8,$23,$2F,$CC,$B4,$FC,$CF,$8A
	.dc.b	$93,$F1,$5F,$6D,$62,$3F,$A2,$47,$F4,$22,$3F,$CA,$F3,$47,$C4,$F9
	.dc.b	$A9,$3C,$CF,$F4,$B7,$C9,$FF,$6E,$DC,$D2,$7A,$5F,$DF,$9C,$79,$D4
	.dc.b	$C4,$67,$83,$5D,$73,$0E,$B1,$97,$BC,$65,$FC,$5D,$FE,$11,$7C,$CB
	.dc.b	$E4,$BF,$70,$D8,$FF,$62,$B5,$FB,$7C,$7F,$F0,$05,$CC,$F0,$23,$F8
	.dc.b	$2F,$42,$BF,$82,$AE,$5A,$7F,$90,$5B,$CB,$F9,$8C,$3A,$56,$F0,$E3
	.dc.b	$7B,$7F,$1F,$86,$2F,$F3,$F8,$72,$BF,$E2,$4D,$7F,$C8,$93,$B7,$35
	.dc.b	$5C,$9E,$A2,$1E,$13,$D4,$62,$FC,$4D,$BF,$45,$73,$BF,$01,$E1,$C3
	.dc.b	$D7,$7C,$2F,$A1,$1F,$F3,$14,$6A,$FD,$78,$73,$3D,$F5,$DC,$F4,$94
	.dc.b	$F4,$9C,$E3,$8B,$99,$E8,$F0,$34,$5A,$F0,$E1,$DE,$38,$35,$D9,$DA
	.dc.b	$3D,$44,$AE,$F1,$F8,$49,$E1,$39,$8F,$E2,$11,$89,$DC,$F8,$27,$D5
	.dc.b	$87,$C8,$C3,$DF,$1E,$08,$C3,$F4,$8C,$47,$D9,$C3,$DB,$1F,$9A,$8E
	.dc.b	$CF,$80,$EA,$EF,$0F,$C9,$C7,$F2,$63,$81,$FE,$23,$AF,$CF,$39,$6B
	.dc.b	$EE,$31,$3F,$E2,$58,$CC,$7C,$D3,$27,$F2,$2F,$43,$F9,$29,$E4,$7F
	.dc.b	$D2,$CE,$1F,$29,$B2,$1E,$E2,$E1,$F8,$68,$E5,$87,$F9,$CA,$73,$8F
	.dc.b	$CC,$23,$FC,$FE,$7B,$3C,$CF,$DC,$B8,$AF,$AD,$1F,$08,$C9,$D0,$FC
	.dc.b	$34,$F4,$7D,$26,$4E,$11,$CA,$BC,$91,$C8,$F0,$47,$A5,$E1,$7F,$71
	.dc.b	$D4,$65,$DD,$7F,$03,$BF,$33,$E1,$AF,$D6,$46,$27,$CA,$43,$CA,$74
	.dc.b	$9E,$18,$6E,$33,$7F,$02,$F9,$C7,$92,$7F,$19,$7B,$9F,$35,$79,$49
	.dc.b	$7D,$2F,$E1,$61,$F0,$27,$A6,$1D,$E3,$15,$BA,$F2,$1F,$54,$FC,$09
	.dc.b	$83,$CA,$EE,$BB,$C6,$CF,$DC,$BF,$D2,$3B,$BF,$A3,$77,$C9,$FE,$9B
	.dc.b	$1F,$D2,$4C,$3F,$D0,$23,$0F,$F8,$C2,$BF,$C6,$3F,$2A,$EA,$FF,$66
	.dc.b	$B6,$74,$30,$46,$63,$FC,$2A,$BA,$F8,$D4,$9E,$35,$C2,$3A,$CF,$F4
	.dc.b	$EF,$9A,$95,$FD,$39,$FF,$05,$FA,$98,$8F,$DB,$1F,$D5,$A3,$F6,$04
	.dc.b	$7F,$70,$FF,$84,$BE,$4A,$47,$33,$C3,$F3,$2D,$7F,$DD,$2B,$5A,$62
	.dc.b	$24,$F0,$8C,$BC,$A3,$FD,$E3,$F2,$BF,$07,$46,$DF,$E5,$A8,$C3,$75
	.dc.b	$DE,$BC,$19,$87,$C4,$AD,$BF,$35,$F9,$4F,$F7,$0F,$0F,$FB,$99,$1F
	.dc.b	$F1,$6F,$AD,$5F,$8B,$C7,$FE,$60,$24,$FF,$15,$3F,$C6,$BF,$31,$27
	.dc.b	$A1,$FE,$12,$F9,$9F,$F0,$77,$BD,$23,$04,$3F,$C0,$EE,$5F,$F0,$4C
	.dc.b	$7F,$A0,$35,$7C,$CB,$2C,$36,$BF,$02,$D7,$FC,$B6,$B0,$DC,$C9,$DB
	.dc.b	$CA,$3F,$D1,$5F,$32,$F9,$AA,$4A,$FB,$BD,$F5,$FC,$3C,$ED,$FF,$6F
	.dc.b	$35,$CC,$FF,$B7,$63,$18,$FD,$A9,$FC,$F7,$23,$BB,$F0,$E7,$59,$D0
	.dc.b	$86,$9F,$8C,$D6,$5F,$E4,$23,$F1,$E6,$1F,$F3,$DC,$A7,$31,$FA,$6F
	.dc.b	$7F,$E3,$5F,$66,$F2,$7F,$A9,$9B,$3F,$B3,$7E,$53,$C7,$6F,$A9,$1A
	.dc.b	$A3,$4D,$CF,$F9,$D3,$CF,$F7,$3F,$ED,$BC,$FF,$D7,$11,$88,$FF,$48
	.dc.b	$FC,$CF,$FB,$72,$31,$1F,$DA,$5A,$FF,$B5,$A7,$FD,$D2,$47,$F8,$0C
	.dc.b	$BD,$9F,$C4,$36,$3C,$07,$F0,$E1,$F4,$8F,$87,$FE,$03,$1F,$EC,$D8
	.dc.b	$C3,$86,$BF,$C1,$9A,$FF,$7D,$F6,$FF,$D3,$EF,$F0,$A6,$B9,$F8,$CE
	.dc.b	$AA,$32,$FF,$AA,$97,$CD,$69,$B1,$E4,$6B,$D6,$39,$FA,$FF,$12,$3F
	.dc.b	$DC,$8D,$4D,$31,$F3,$4F,$98,$F1,$B6,$7E,$0B,$5F,$E4,$6D,$7F,$C3
	.dc.b	$4B,$E6,$B8,$92,$38,$BF,$E1,$79,$91,$FE,$1F,$6B,$FC,$B9,$7A,$9F
	.dc.b	$11,$B1,$F4,$9A,$EE,$35,$94,$8D,$35,$FE,$73,$F4,$7F,$E7,$C7,$FC
	.dc.b	$60,$D7,$FD,$AE,$D4,$1C,$2F,$97,$7C,$CF,$F5,$18,$FF,$6F,$C6,$3C
	.dc.b	$3F,$C3,$AF,$99,$FD,$AC,$FF,$4A,$8D,$3F,$D3,$3E,$65,$E9,$FF,$01
	.dc.b	$8F,$F2,$24,$63,$AF,$F2,$2D,$8E,$13,$B7,$FD,$DF,$A7,$E2,$6B,$F2
	.dc.b	$ED,$8F,$EB,$2D,$7F,$CA,$0D,$7F,$DE,$FF,$2E,$F4,$BA,$47,$C2,$8C
	.dc.b	$57,$E8,$9B,$1F,$D9,$1A,$FF,$96,$D7,$C5,$F2,$84,$62,$79,$4E,$63
	.dc.b	$13,$FD,$BB,$E7,$FF,$8D,$D7,$EC,$1B,$19,$F0,$BA,$46,$23,$FC,$D5
	.dc.b	$F3,$1F,$88,$AF,$DB,$B6,$3F,$BC,$B5,$DE,$5E,$58,$7B,$D4,$95,$C9
	.dc.b	$F5,$1F,$2F,$C1,$FE,$63,$F3,$1D,$5F,$B1,$19,$9F,$C4,$C2,$75,$1F
	.dc.b	$8C,$C3,$BF,$5F,$42,$47,$F0,$EB,$FA,$93,$F9,$27,$A3,$67,$FA,$0C
	.dc.b	$F6,$A7,$36,$2B,$53,$FC,$6F,$E5,$DE,$D6,$72,$7C,$C3,$E3,$78,$60
	.dc.b	$BD,$E1,$FE,$28,$AD,$3C,$F0,$FB,$93,$AA,$FF,$12,$B5,$FF,$4E,$35
	.dc.b	$F8,$1F,$31,$FD,$3E,$BF,$C6,$7D,$EF,$6B,$C3,$39,$AF,$F1,$F5,$7C
	.dc.b	$AA,$D3,$FC,$D6,$75,$9F,$E9,$DB,$1D,$AB,$9F,$95,$EC,$AE,$31,$D0
	.dc.b	$FA,$BE,$26,$71,$D3,$6F,$FA,$15,$70,$3F,$47,$F3,$1F,$C1,$3E,$5F
	.dc.b	$FA,$46,$1F,$E5,$3C,$3E,$4B,$FC,$22,$75,$E4,$7D,$FB,$F0,$DF,$0A
	.dc.b	$D0,$79,$C6,$A7,$FC,$9F,$5C,$AB,$F1,$77,$FE,$05,$BE,$50,$9D,$3E
	.dc.b	$3C,$2B,$DA,$7A,$D7,$1B,$EB,$3E,$17,$CD,$19,$8F,$F6,$A5,$F4,$C3
	.dc.b	$C9,$18,$38,$1E,$4C,$3B,$BE,$33,$27,$8C,$A7,$8E,$FB,$C7,$0A,$FE
	.dc.b	$C3,$3C,$4F,$ED,$EE,$0F,$EE,$3F,$2E,$F8,$77,$CC,$61,$B5,$FE,$76
	.dc.b	$9D,$4E,$4E,$3E,$B7,$FA,$26,$74,$D7,$FD,$E5,$28,$C4,$7E,$DE,$31
	.dc.b	$97,$23,$E7,$E5,$FE,$29,$9B,$B7,$EF,$75,$FC,$2D,$F0,$9F,$86,$D8
	.dc.b	$FC,$4D,$71,$9F,$8E,$D8,$E3,$87,$9D,$B8,$76,$A8,$74,$9D,$57,$9D
	.dc.b	$B1,$FD,$6A,$3A,$1F,$C4,$4A,$FF,$5E,$BF,$24,$71,$AE,$B7,$E6,$C3
	.dc.b	$D9,$39,$BF,$3C,$7E,$77,$4E,$8D,$8F,$03,$56,$1D,$9C,$9F,$E2,$FA
	.dc.b	$F8,$58,$7F,$61,$6D,$F7,$C3,$F5,$FF,$31,$FC,$1E,$FF,$C6,$CD,$7F
	.dc.b	$78,$1C,$5F,$F6,$B7,$EE,$47,$0D,$BC,$91,$AB,$FE,$68,$D9,$FF,$00
	.dc.b	$BC,$1E,$BF,$2B,$FE,$5A,$7F,$C8,$CD,$73,$E8,$FF,$A4,$FE,$33,$FB
	.dc.b	$D6,$3F,$DE,$EB,$55,$FD,$04,$38,$7F,$A2,$35,$FF,$08,$B5,$DF,$C0
	.dc.b	$FC,$B8,$34,$FF,$51,$7C,$BB,$E6,$31,$5E,$69,$AF,$F2,$1B,$56,$E3
	.dc.b	$47,$F9,$7A,$FA,$B6,$BF,$CA,$B1,$88,$FF,$71,$FA,$CF,$95,$FF,$2C
	.dc.b	$B6,$3F,$CC,$ED,$77,$1C,$BE,$1F,$D4,$48,$FF,$38,$4E,$F1,$F9,$13
	.dc.b	$B9,$F2,$35,$F0,$FC,$0F,$80,$8F,$81,$F3,$D4,$8F,$63,$5F,$F7,$DE
	.dc.b	$1F,$E1,$67,$E0,$79,$3E,$E2,$3F,$E8,$0F,$67,$FD,$70,$FF,$A3,$FD
	.dc.b	$CF,$F3,$BC,$62,$3F,$BE,$35,$FB,$3B,$FC,$82,$3F,$35,$FF,$3F,$FC
	.dc.b	$5B,$E4,$A9,$27,$FD,$1F,$19,$7E,$A7,$B3,$F4,$88,$E0,$FF,$4A,$FC
	.dc.b	$BF,$F9,$29,$BE,$8F,$FA,$32,$31,$1F,$91,$0E,$1F,$CA,$3F,$E0,$37
	.dc.b	$9F,$8F,$C4,$8F,$63,$A1,$0C,$1F,$EA,$6F,$97,$FF,$4D,$37,$CB,$FF
	.dc.b	$4E,$C6,$23,$C4,$D7,$F4,$8F,$E1,$DB,$C9,$C0,$8F,$FA,$CB,$E5,$DC
	.dc.b	$9F,$EA,$86,$C6,$63,$06,$7E,$42,$47,$16,$BF,$0F,$E5,$9F,$2F,$FE
	.dc.b	$B2,$6C,$7F,$B0,$1A,$EE,$3F,$D6,$FF,$5A,$F9,$EF,$E3,$CD,$9F,$F6
	.dc.b	$53,$5F,$D8,$76,$3E,$53,$75,$C8,$93,$FE,$CE,$AE,$0F,$F0,$77,$97
	.dc.b	$67,$FC,$00,$BF,$F4,$5C,$F4,$F5,$F9,$BF,$E5,$0F,$0B,$FC,$A3,$F5
	.dc.b	$AF,$99,$FE,$69,$82,$31,$5B,$11,$99,$DC,$BE,$6A,$C4,$AE,$15,$B6
	.dc.b	$D6,$DB,$E6,$F2,$FA,$E9,$18,$21,$FC,$4E,$51,$A8,$C4,$FF,$49,$E9
	.dc.b	$7D,$55,$24,$F8,$39,$AB,$F9,$D9,$DB,$FE,$F8,$6B,$99,$FF,$77,$C6
	.dc.b	$31,$69,$FE,$78,$FA,$E7,$89,$F2,$1B,$87,$C7,$3F,$D6,$F8,$67,$DC
	.dc.b	$91,$F1,$27,$FD,$AB,$87,$48,$F9,$B1,$FC,$A6,$30,$1D,$E1,$E1,$AE
	.dc.b	$75,$FC,$DE,$78,$4F,$97,$9F,$C9,$F3,$52,$BC,$11,$E7,$30,$F1,$8C
	.dc.b	$3E,$9F,$0F,$F8,$63,$FD,$A1,$B8,$FF,$B4,$5B,$7F,$E9,$AE,$27,$C7
	.dc.b	$FF,$7B,$9E,$37,$FD,$6D,$AF,$F2,$2F,$0F,$F9,$29,$1F,$C3,$D3,$D7
	.dc.b	$FC,$9B,$63,$C9,$38,$F6,$64,$70,$77,$F1,$B8,$8F,$EB,$55,$FE,$9C
	.dc.b	$74,$76,$8E,$BF,$29,$DB,$E5,$64,$F8,$A3,$2F,$92,$32,$F4,$8F,$F0
	.dc.b	$C7,$93,$FE,$01,$DB,$FD,$E8,$DD,$BF,$D7,$4D,$AF,$D4,$FC,$EF,$EB
	.dc.b	$1F,$F0,$9C,$7F,$82,$5F,$F0,$03,$5E,$3E,$C7,$F3,$34,$23,$4E,$1E
	.dc.b	$D6,$AF,$E0,$77,$FE,$CE,$6D,$6E,$FF,$7F,$1C,$08,$E3,$A7,$2C,$9F
	.dc.b	$DD,$CE,$E7,$8F,$C4,$F2,$F4,$25,$6A,$32,$F0,$8C,$BF,$E0,$DA,$4F
	.dc.b	$F9,$07,$CD,$FA,$D3,$FE,$36,$6E,$5F,$A5,$84,$63,$2F,$F7,$37,$AB
	.dc.b	$FE,$DC,$9F,$D8,$4E,$5F,$8E,$D7,$8F,$0F,$F9,$3E,$78,$47,$78,$F1
	.dc.b	$11,$FD,$D4,$6A,$FC,$55,$CC,$95,$CF,$E0,$38,$5F,$33,$C7,$19,$69
	.dc.b	$0D,$1F,$E5,$4B,$ED,$87,$FC,$01,$BE,$B3,$BC,$3E,$15,$6A,$3D,$B7
	.dc.b	$E0,$9F,$2B,$DE,$3A,$C6,$1E,$57,$F2,$30,$E0,$F7,$AB,$3F,$8E,$BE
	.dc.b	$CF,$C8,$38,$9E,$03,$E2,$D7,$07,$47,$82,$BB,$3C,$0C,$DF,$8D,$F8
	.dc.b	$13,$C7,$0F,$7C,$60,$E0,$E2,$7E,$0C,$EF,$0F,$06,$1E,$6A,$D6,$1E
	.dc.b	$BC,$3C,$A7,$2A,$DD,$F1,$8F,$25,$F1,$9C,$47,$84,$F5,$1E,$F2,$88
	.dc.b	$E0,$F5,$3B,$9A,$BF,$1B,$F2,$EB,$AE,$1F,$04,$F5,$C7,$E0,$CF,$F4
	.dc.b	$04,$FE,$89,$B7,$FE,$42,$23,$FE,$40,$6B,$FE,$37,$8E,$0F,$C1,$3F
	.dc.b	$19,$F1,$3F,$9F,$3C,$38,$FE,$A1,$C4,$7A,$4D,$9F,$A0,$76,$FE,$0D
	.dc.b	$DB,$ED,$AB,$7F,$94,$9D,$CF,$C5,$73,$5E,$39,$F8,$0F,$86,$7B,$1E
	.dc.b	$18,$E4,$E0,$EC,$57,$2E,$0F,$39,$1E,$AF,$D7,$76,$48,$F1,$12,$3D
	.dc.b	$87,$57,$0F,$E5,$9F,$33,$D5,$CD,$FF,$0B,$C5,$7E,$EA,$F0,$57,$29
	.dc.b	$CC,$75,$32,$43,$E3,$19,$3C,$06,$5E,$2F,$84,$E1,$5E,$1A,$F0,$47
	.dc.b	$02,$4F,$0B,$F8,$11,$C3,$0D,$CE,$70,$F5,$C6,$2F,$EC,$CE,$6B,$E2
	.dc.b	$9F,$3E,$F9,$BE,$83,$17,$F1,$A3,$38,$7F,$31,$5F,$3A,$F8,$4F,$F2
	.dc.b	$6B,$6F,$FC,$8A,$D7,$FD,$A0,$FF,$87,$9F,$F2,$7B,$99,$FF,$11,$B6
	.dc.b	$FF,$B1,$C7,$6A,$EB,$DB,$DB,$E7,$7B,$F6,$CC,$7B,$C6,$5F,$0C,$65
	.dc.b	$F0,$C7,$F5,$3F,$2A,$47,$F4,$C7,$FA,$44,$E2,$3F,$61,$1F,$EB,$62
	.dc.b	$3F,$EB,$47,$FA,$77,$74,$8E,$47,$F9,$93,$E7,$7F,$D5,$F9,$7F,$8F
	.dc.b	$9B,$9F,$B1,$B3,$FE,$70,$7F,$C5,$AF,$1F,$0F,$FC,$EF,$1A,$A3,$4D
	.dc.b	$CF,$FB,$ED,$FA,$83,$AF,$89,$CB,$8A,$4F,$96,$32,$F9,$E3,$FC,$22
	.dc.b	$FF,$A0,$B0,$BF,$30,$F9,$DC,$E3,$FD,$DE,$31,$5F,$E6,$36,$BF,$D7
	.dc.b	$FE,$33,$77,$CA,$FF,$C3,$D2,$AD,$BE,$03,$11,$A8,$CC,$71,$3F,$BB
	.dc.b	$FA,$3F,$DF,$4E,$AF,$F8,$A2,$3C,$38,$7C,$FE,$EF,$93,$DF,$2B,$EA
	.dc.b	$47,$CC,$30,$FC,$98,$C3,$FE,$9C,$F9,$DF,$F0,$CB,$7A,$FF,$D5,$31
	.dc.b	$88,$F8,$C7,$FA,$A1,$76,$EA,$64,$F1,$3A,$47,$A1,$F8,$5F,$3B,$72
	.dc.b	$34,$FF,$0D,$7B,$B6,$7F,$C8,$D3,$A6,$CF,$C7,$23,$FE,$1C,$F9,$DF
	.dc.b	$F1,$11,$C5,$F3,$E8,$F9,$96,$BF,$EC,$45,$AD,$31,$12,$BF,$D7,$F3
	.dc.b	$A6,$CE,$5E,$BF,$3B,$F1,$6F,$FD,$3A,$DE,$1F,$DD,$B6,$7F,$D9,$EF
	.dc.b	$C8,$7F,$D9,$6E,$64,$D2,$6B,$F1,$FC,$49,$5F,$91,$7F,$C8,$1E,$6E
	.dc.b	$23,$F1,$87,$E3,$E3,$F1,$84,$7E,$5B,$FE,$47,$F9,$C9,$1C,$CF,$F5
	.dc.b	$F7,$D5,$3E,$1F,$F3,$2D,$8F,$F3,$C3,$5E,$ED,$7F,$39,$F3,$6F,$8D
	.dc.b	$FD,$0B,$63,$FD,$00,$D7,$CE,$D7,$F4,$9E,$FF,$F9,$69,$BE,$4F,$EA
	.dc.b	$A3,$11,$FB,$86,$BF,$E8,$AF,$90,$75,$7E,$C1,$B1,$FE,$8F,$6B,$F1
	.dc.b	$9A,$FE,$CF,$D0,$FE,$14,$FF,$A5,$5A,$FC,$C6,$BB,$EF,$FD,$EF,$CD
	.dc.b	$BE,$6A,$92,$7F,$D3,$D1,$97,$FD,$41,$D8,$FA,$BF,$E0,$AD,$8C,$C6
	.dc.b	$0F,$E0,$95,$23,$F4,$6D,$7E,$FA,$F2,$7F,$87,$B6,$3F,$D5,$CD,$77
	.dc.b	$1F,$C5,$3C,$EF,$F5,$C4,$7F,$13,$6B,$FC,$91,$AF,$FA,$D9,$AB,$E6
	.dc.b	$DE,$27,$F1,$E6,$C7,$FA,$F5,$AF,$E5,$1A,$FF,$21,$F9,$A7,$CD,$FF
	.dc.b	$0D,$8F,$E6,$71,$8F,$13,$FD,$8C,$FF,$85,$3D,$FB,$5B,$AF,$F6,$4F
	.dc.b	$D4,$3C,$BF,$F5,$64,$7F,$3C,$8C,$78,$3F,$EC,$F6,$C7,$E9,$9A,$FF
	.dc.b	$33,$D5,$CD,$CA,$3F,$DA,$B1,$8A,$FF,$6C,$36,$3F,$5E,$D7,$F9,$E7
	.dc.b	$CD,$FC,$49,$FF,$6F,$36,$3F,$6A,$D7,$7B,$77,$7F,$91,$E5,$7C,$7F
	.dc.b	$8B,$7F,$EE,$76,$CF,$EF,$9A,$FF,$4B,$F9,$B6,$BC,$DB,$5F,$EE,$F6
	.dc.b	$CF,$F0,$66,$BF,$D4,$FE,$6F,$F1,$53,$FE,$F5,$6C,$7F,$0C,$6B,$BC
	.dc.b	$FD,$1E,$9F,$C3,$F1,$97,$FE,$1A,$F9,$BE,$31,$FE,$DF,$9D,$4E,$6B
	.dc.b	$77,$FD,$30,$FF,$16,$6F,$8B,$98,$72,$BF,$EE,$F8,$72,$B5,$2B,$50
	.dc.b	$39,$D7,$F2,$46,$9F,$C9,$AB,$EC,$3D,$2F,$E9,$7C,$DB,$D0,$F1,$5F
	.dc.b	$F8,$D3,$B9,$F1,$B8,$C3,$B5,$EE,$BF,$C7,$26,$E7,$52,$9D,$3F,$D5
	.dc.b	$2B,$9D,$70,$8E,$77,$D7,$35,$6A,$7E,$37,$CD,$BE,$67,$5E,$0F,$FA
	.dc.b	$E7,$1E,$54,$AD,$47,$5C,$15,$BA,$E7,$7E,$6B,$DC,$76,$31,$1E,$AA
	.dc.b	$E0,$7D,$88,$FA,$3E,$2E,$7D,$BD,$37,$1F,$1D,$70,$7F,$AA,$FC,$DF
	.dc.b	$D5,$F9,$9F,$E9,$53,$9C,$3F,$C4,$93,$9F,$2B,$FC,$63,$8F,$E1,$AF
	.dc.b	$FA,$75,$2B,$4F,$F9,$A2,$75,$5D,$A1,$5D,$A7,$31,$D8,$CC,$F6,$C3
	.dc.b	$9C,$7F,$91,$2E,$E1,$F1,$F0,$F0,$C6,$30,$E7,$5B,$39,$9E,$3B,$E6
	.dc.b	$FE,$4E,$78,$46,$23,$D4,$70,$3F,$37,$3D,$1E,$06,$6F,$C5,$87,$B6
	.dc.b	$3F,$39,$7F,$E4,$E7,$99,$83,$FC,$07,$F3,$4F,$80,$7C,$DF,$E8,$E3
	.dc.b	$0D,$8C,$CF,$06,$BF,$E5,$79,$79,$FF,$B0,$8E,$AD,$74,$D7,$FC,$DB
	.dc.b	$1C,$31,$F0,$4F,$1D,$3F,$C1,$59,$7E,$6E,$FE,$8B,$5F,$89,$7C,$5B
	.dc.b	$5C,$67,$FC,$F2,$6E,$B9,$ED,$FE,$7C,$7F,$C2,$D7,$C5,$B5,$C6,$FF
	.dc.b	$83,$5F,$13,$9D,$F1,$7D,$33,$B9,$5F,$39,$DC,$7A,$2B,$83,$FC,$46
	.dc.b	$74,$DA,$E5,$5D,$E3,$C3,$5C,$1E,$B3,$A8,$FF,$34,$4F,$C3,$C3,$F6
	.dc.b	$0D,$D7,$CB,$1E,$1B,$FE,$03,$39,$7A,$9F,$C1,$30,$F1,$AF,$9B,$6D
	.dc.b	$FE,$75,$6C,$7F,$0A,$6B,$FC,$D5,$FC,$0B,$98,$E0,$47,$86,$5F,$E7
	.dc.b	$FF,$45,$F4,$7F,$E9,$8C,$9F,$08,$F4,$30,$DC,$3F,$8C,$DF,$06,$DE
	.dc.b	$AA,$46,$0D,$FC,$67,$FA,$BA,$8D,$24,$7F,$41,$8C,$BF,$C8,$ED,$F2
	.dc.b	$FE,$23,$D1,$52,$75,$19,$7B,$9E,$6E,$5F,$01,$1E,$CD,$7F,$AE,$5F
	.dc.b	$F9,$05,$BE,$6F,$C5,$9F,$8C,$FA,$49,$1E,$77,$FD,$93,$F8,$9B,$E3
	.dc.b	$69,$27,$FA,$E4,$65,$FE,$BD,$F0,$AF,$96,$FF,$68,$B6,$33,$18,$3F
	.dc.b	$DA,$BF,$89,$BE,$5A,$92,$7F,$B4,$46,$5F,$ED,$5F,$38,$F9,$CF,$F6
	.dc.b	$EB,$63,$31,$83,$3F,$49,$23,$8B,$5F,$F4,$BC,$BE,$27,$99,$23,$BB
	.dc.b	$FE,$EA,$F9,$CF,$74,$7F,$A7,$9A,$FF,$BA,$DA,$FF,$BF,$1A,$FF,$80
	.dc.b	$D7,$D7,$73,$24,$7F,$81,$63,$2D,$FA,$9F,$01,$1C,$1F,$E0,$BF,$9E
	.dc.b	$FB,$8F,$FB,$ED,$AF,$F8,$44,$38,$7F,$C2,$8D,$73,$F4,$7F,$E1,$56
	.dc.b	$9D,$5C,$41,$BF,$53,$27,$C4,$D7,$FD,$83,$F4,$0F,$AF,$E8,$D8,$FF
	.dc.b	$11,$35,$DC,$6B,$CA,$91,$A6,$BD,$3E,$8B,$99,$1E,$6F,$E3,$BD,$0F
	.dc.b	$F7,$F7,$2F,$F9,$AA,$7F,$C2,$6D,$7F,$C7,$DC,$7F,$E1,$27,$33,$C5
	.dc.b	$FC,$CB,$5F,$F6,$DC,$F1,$F0,$3F,$C2,$0F,$C3,$8E,$BA,$78,$2A,$46
	.dc.b	$08,$7F,$93,$67,$97,$8D,$F5,$5B,$1F,$E4,$77,$A3,$BA,$E6,$F1,$AF
	.dc.b	$6C,$48,$F4,$90,$E0,$78,$2F,$CF,$3C,$9F,$4C,$F1,$9E,$86,$EF,$C9
	.dc.b	$19,$9E,$33,$93,$CD,$19,$9F,$55,$59,$C9,$B1,$7E,$E6,$D6,$DA,$7A
	.dc.b	$6F,$CA,$F1,$9F,$87,$1E,$28,$F6,$91,$DB,$EA,$31,$7E,$A7,$95,$FB
	.dc.b	$DF,$D2,$1E,$03,$EE,$DF,$F1,$62,$3F,$E5,$AF,$C6,$5D,$BF,$AB,$C3
	.dc.b	$FC,$A1,$18,$9F,$35,$67,$0F,$D8,$4E,$67,$FD,$18,$DA,$F3,$92,$77
	.dc.b	$7C,$6B,$FD,$11,$5B,$9F,$54,$72,$9D,$29,$D3,$C2,$3E,$64,$62,$35
	.dc.b	$1C,$23,$18,$FC,$CA,$F0,$D7,$BE,$B4,$72,$3F,$82,$57,$AE,$3E,$4B
	.dc.b	$5E,$27,$CC,$3B,$9F,$27,$D2,$F0,$1D,$1E,$C6,$48,$7B,$4C,$9F,$21
	.dc.b	$E2,$E3,$0F,$A3,$18,$8D,$19,$BF,$A3,$19,$9D,$CE,$4F,$9B,$19,$9D
	.dc.b	$C6,$4F,$99,$19,$76,$66,$91,$98,$D9,$97,$11,$A8,$C1,$EE,$AF,$1B
	.dc.b	$EF,$3D,$24,$72,$7E,$3A,$F9,$D7,$46,$CF,$3B,$FA,$35,$C1,$F5,$C7
	.dc.b	$37,$26,$23,$51,$EC,$3C,$4E,$1F,$59,$C4,$FC,$61,$DC,$E0,$EE,$FF
	.dc.b	$1C,$7B,$E3,$81,$CA,$3D,$4F,$13,$13,$E3,$3C,$AF,$47,$0D,$3C,$27
	.dc.b	$49,$ED,$3E,$1A,$F8,$73,$96,$C6,$63,$C1,$3C,$2B,$AC,$7E,$58,$E9
	.dc.b	$7F,$35,$DB,$5C,$9C,$AF,$AC,$F2,$AF,$83,$58,$AF,$3D,$FD,$E7,$F4
	.dc.b	$F1,$FA,$A2,$3F,$A7,$3E,$79,$A3,$13,$F2,$DC,$1F,$AD,$7D,$48,$66
	.dc.b	$F9,$1F,$E0,$AD,$8F,$59,$FE,$F4,$70,$FF,$62,$8F,$43,$C9,$E2,$62
	.dc.b	$37,$7F,$EE,$F6,$E1,$F0,$70,$DD,$F5,$BE,$06,$1F,$92,$77,$34,$79
	.dc.b	$0C,$12,$FF,$19,$3C,$5F,$04,$F4,$9E,$27,$48,$CB,$87,$8B,$6B,$FC
	.dc.b	$45,$7E,$88,$FC,$61,$F9,$23,$E5,$39,$7E,$11,$E5,$9E,$AF,$96,$7E
	.dc.b	$03,$D4,$E3,$5E,$57,$47,$6C,$3E,$BB,$D0,$C9,$E1,$25,$F8,$4F,$14
	.dc.b	$70,$38,$B9,$9E,$D7,$E4,$3D,$F5,$D2,$78,$E1,$FC,$55,$E1,$1E,$BC
	.dc.b	$3E,$1D,$FC,$D7,$D7,$1C,$63,$AC,$6A,$38,$1E,$E3,$E0,$1E,$88,$E4
	.dc.b	$E4,$FE,$03,$3E,$07,$DF,$87,$AA,$77,$1F,$0F,$0F,$E2,$15,$F9,$48
	.dc.b	$FE,$EA,$D7,$C4,$7F,$4D,$70,$F7,$85,$FF,$97,$FD,$C7,$D6,$7F,$66
	.dc.b	$6D,$7A,$2F,$EF,$46,$21,$7C,$9F,$D6,$46,$63,$E7,$5F,$89,$D4,$A7
	.dc.b	$53,$E1,$BF,$6C,$62,$79,$C7,$8C,$C9,$5F,$DD,$57,$49,$E1,$1F,$16
	.dc.b	$BA,$3A,$6B,$E8,$6C,$F9,$A3,$B9,$FC,$36,$7D,$CF,$F8,$81,$AF,$23
	.dc.b	$15,$FE,$0C,$FA,$9F,$F3,$A5,$7B,$9B,$7D,$89,$5F,$C4,$63,$2D,$E2
	.dc.b	$F6,$3E,$53,$C6,$51,$1C,$1B,$1C,$2B,$C0,$F8,$C9,$1F,$04,$E7,$1A
	.dc.b	$70,$72,$2C,$F5,$C3,$E5,$BD,$C9,$5D,$EF,$FD,$4D,$1F,$E9,$3A,$FF
	.dc.b	$22,$B5,$CC,$FF,$A9,$63,$11,$EF,$3C,$3F,$25,$FE,$5D,$3F,$CB,$D1
	.dc.b	$DA,$33,$C3,$FE,$38,$9C,$D7,$B2,$7E,$24,$71,$9C,$63,$F8,$E8,$C4
	.dc.b	$F2,$3D,$B5,$CE,$A4,$64,$F9,$D7,$F3,$6B,$8C,$76,$9E,$B8,$7F,$B1
	.dc.b	$67,$FD,$49,$3F,$E6,$46,$B9,$9F,$F6,$24,$62,$3F,$A8,$3F,$E9,$8C
	.dc.b	$1F,$52,$F6,$3F,$CB,$51,$87,$CB,$5F,$81,$BF,$F4,$1B,$FD,$85,$AF
	.dc.b	$F9,$25,$AF,$F9,$D9,$AF,$FB,$6D,$7B,$FF,$ED,$76,$C7,$38,$C1,$9F
	.dc.b	$86,$91,$DD,$AF,$F9,$5F,$E2,$5E,$77,$FB,$85,$B1,$FD,$D5,$AE,$E3
	.dc.b	$5E,$4C,$8D,$35,$E9,$E7,$F3,$23,$CD,$FF,$77,$7D,$63,$D3,$A4,$9F
	.dc.b	$F0,$24,$65,$BF,$31,$C0,$8E,$0D,$FC,$9F,$42,$38,$3F,$C1,$DD,$CF
	.dc.b	$A0,$FF,$7F,$36,$33,$18,$33,$F1,$12,$38,$B5,$FF,$3F,$4F,$F3,$24
	.dc.b	$D2,$75,$FF,$87,$61,$F1,$3F,$63,$E2,$BF,$C4,$0D,$8F,$F4,$42,$E1
	.dc.b	$7C,$3E,$B8,$D7,$FE,$27,$FA,$A6,$7F,$CC,$B1,$FF,$11,$08,$CF,$56
	.dc.b	$E7,$53,$FE,$27,$9C,$BF,$E9,$7F,$AC,$7B,$5A,$84,$62,$FF,$C6,$CD
	.dc.b	$8E,$13,$BF,$C0,$9C,$34,$97,$FE,$3D,$8C,$BF,$E3,$FF,$C0,$5F,$57
	.dc.b	$DD,$B1,$98,$C1,$E3,$F8,$B7,$C9,$52,$4F,$F9,$26,$32,$FF,$92,$FC
	.dc.b	$87,$B1,$EB,$6C,$66,$30,$47,$F9,$4C,$6B,$4F,$E7,$1E,$0B,$E2,$FE
	.dc.b	$A5,$6A,$FF,$B3,$4F,$F6,$2B,$E0,$D8,$CC,$FF,$5C,$AE,$37,$FD,$06
	.dc.b	$F7,$3F,$63,$E2,$DE,$D7,$D1,$6C,$7F,$99,$5B,$5F,$EC,$D6,$BF,$EF
	.dc.b	$A7,$38,$FF,$74,$6C,$FF,$6C,$F8,$A7,$AB,$E4,$BF,$F3,$45,$69,$FF
	.dc.b	$35,$7C,$53,$E1,$B5,$8B,$D4,$78,$9C,$6D,$7B,$CF,$FA,$A5,$EE,$B5
	.dc.b	$39,$AF,$EC,$0D,$9F,$F0,$4B,$FE,$79,$AF,$EB,$2E,$5F,$CD,$C7,$81
	.dc.b	$FE,$BA,$BC,$5F,$13,$9A,$FE,$D8,$E6,$77,$7A,$3D,$13,$FE,$87,$F8
	.dc.b	$AF,$F4,$47,$C4,$7F,$0D,$8F,$F1,$0C,$E9,$4E,$BC,$7F,$1C,$FF,$89
	.dc.b	$6B,$85,$73,$0F,$3C,$3C,$8E,$A3,$99,$AA,$E7,$87,$58,$FD,$C9,$A9
	.dc.b	$E5,$7D,$EB,$33,$97,$8B,$CC,$E3,$87,$37,$D7,$75,$F3,$5F,$48,$C4
	.dc.b	$73,$3F,$C6,$37,$DD,$E6,$75,$39,$94,$E7,$7E,$2A,$F7,$5F,$F1,$F9
	.dc.b	$F7,$9F,$EB,$D7,$07,$FA,$FF,$E2,$1F,$12,$F8,$AC,$36,$BD,$F3,$A9
	.dc.b	$C9,$C7,$CB,$FF,$2B,$CE,$9A,$FC,$A9,$FE,$B7,$A6,$C9,$1E,$BC,$BF
	.dc.b	$3C,$FC,$76,$BE,$F9,$E2,$DA,$EF,$97,$E1,$EB,$6D,$E0,$FE,$62,$FE
	.dc.b	$2E,$F8,$B6,$B8,$D7,$F3,$63,$89,$6B,$8B,$FC,$EA,$B8,$B6,$78,$DF
	.dc.b	$FB,$7E,$38,$B6,$B8,$BF,$EE,$3A,$E2,$DA,$E2,$FF,$6F,$9D,$B4,$FC
	.dc.b	$44,$F0,$70,$7F,$95,$EB,$E2,$E1,$FE,$80,$6E,$BE,$58,$F9,$57,$FB
	.dc.b	$88,$F6,$57,$F9,$52,$B6,$F8,$EB,$F2,$DF,$15,$FC,$32,$3F,$CB,$2D
	.dc.b	$7F,$CC,$CD,$8F,$F4,$7B,$FE,$12,$6E,$DD,$E4,$27,$85,$FF,$64,$6C
	.dc.b	$69,$E5,$1F,$DA,$3D,$8F,$EC,$33,$FA,$E0,$E1,$FD,$8B,$5F,$DB,$B5
	.dc.b	$E5,$E0,$7F,$90,$3E,$25,$F1,$54,$F3,$4F,$49,$B4,$F8,$AA,$49,$FE
	.dc.b	$57,$7F,$E0,$37,$FC,$FB,$3B,$7C,$CD,$7C,$B3,$FE,$0F,$8C,$47,$9A
	.dc.b	$3C,$9F,$15,$23,$9C,$78,$B9,$64,$76,$7C,$BF,$5F,$88,$F8,$31,$D4
	.dc.b	$8F,$89,$F5,$FC,$5C,$8E,$47,$FA,$53,$E2,$9F,$17,$F0,$9B,$19,$8F
	.dc.b	$8B,$EF,$BE,$53,$51,$FE,$29,$AD,$D7,$F2,$26,$BF,$CA,$1A,$FF,$88
	.dc.b	$BE,$0B,$FC,$02,$FD,$B6,$BF,$CA,$9A,$FF,$30,$6B,$FE,$2A,$F8,$AF
	.dc.b	$C0,$BF,$83,$6B,$FC,$C9,$AF,$F3,$86,$BC,$7E,$33,$FC,$81,$06,$92
	.dc.b	$3A,$7C,$66,$88,$E0,$FC,$97,$C5,$7F,$95,$5F,$E8,$2D,$7F,$A3,$B5
	.dc.b	$FF,$20,$B5,$FD,$1B,$F1,$1B,$8F,$F9,$4E,$31,$06,$8F,$8B,$D1,$23
	.dc.b	$C6,$FF,$96,$FE,$2D,$F5,$54,$93,$FA,$98,$CB,$77,$E0,$47,$06,$FE
	.dc.b	$2B,$A1,$1C,$1F,$B3,$F3,$5F,$1B,$FE,$6B,$6C,$66,$30,$7F,$9B,$FE
	.dc.b	$33,$FD,$AB,$06,$93,$E2,$75,$E5,$49,$FE,$01,$3F,$E1,$F6,$C7,$F9
	.dc.b	$E7,$97,$11,$FE,$F9,$6F,$82,$EF,$FF,$38,$C6,$1F,$F3,$9F,$D6,$3A
	.dc.b	$BF,$3B,$87,$32,$61,$FC,$3E,$FA,$BF,$E0,$27,$93,$FC,$73,$9D,$F0
	.dc.b	$DF,$C2,$9B,$1F,$F0,$02,$76,$9B,$1A,$86,$5E,$77,$A5,$79,$1E,$57
	.dc.b	$D6,$24,$60,$87,$2C,$BF,$5D,$1A,$8F,$89,$3C,$DD,$1D,$8F,$E6,$B5
	.dc.b	$E1,$57,$E5,$8C,$61,$EA,$3D,$37,$D6,$33,$7D,$5E,$55,$C7,$0F,$1E
	.dc.b	$1E,$69,$CD,$F1,$9C,$CF,$98,$F1,$B8,$9D,$C6,$0F,$41,$B3,$D6,$7A
	.dc.b	$27,$A3,$A8,$F2,$BC,$4D,$16,$BC,$CE,$EB,$F0,$F7,$CA,$36,$76,$9D
	.dc.b	$BF,$6E,$35,$18,$33,$87,$C6,$9F,$E3,$D8,$7F,$B3,$E7,$F0,$11,$FC
	.dc.b	$76,$30,$7E,$27,$79,$1E,$58,$EB,$18,$8F,$14,$61,$F5,$47,$8A,$35
	.dc.b	$8F,$D2,$8C,$4E,$88,$72,$3B,$18,$8E,$0E,$0F,$81,$1D,$31,$FE,$74
	.dc.b	$F8,$63,$27,$18,$FB,$06,$9C,$1F,$A7,$8E,$13,$DA,$38,$9E,$32,$B9
	.dc.b	$24,$F4,$8E,$47,$43,$F2,$F3,$D2,$3C,$11,$C9,$E0,$E8,$E1,$69,$E1
	.dc.b	$1D,$0E,$07,$E0,$CC,$57,$89,$B5,$D2,$B9,$47,$4A,$E6,$F7,$39,$9F
	.dc.b	$8A,$AE,$4E,$5B,$3F,$89,$9F,$9D,$3D,$8F,$11,$BA,$F6,$D7,$0B,$DC
	.dc.b	$E4,$E0,$77,$9D,$91,$CD,$F7,$7D,$07,$DF,$AF,$1D,$F4,$9D,$4F,$A9
	.dc.b	$F7,$BF,$5E,$76,$FC,$CB,$E1,$5C,$5B,$19,$3B,$9C,$DC,$B4,$CC,$FB
	.dc.b	$5B,$7E,$97,$D8,$7A,$AB,$B3,$A3,$DB,$5D,$2B,$94,$7E,$38,$FF,$06
	.dc.b	$4F,$EC,$C8,$FF,$84,$DA,$FF,$84,$DA,$F0,$3F,$85,$7C,$6F,$F8,$42
	.dc.b	$7F,$94,$36,$FC,$D8,$7C,$69,$CF,$7F,$F9,$3A,$B5,$7E,$93,$9C,$64
	.dc.b	$F4,$5F,$F9,$A2,$13,$C1,$B3,$A7,$FC,$DB,$19,$35,$7E,$1A,$F9,$11
	.dc.b	$9A,$E1,$5C,$8F,$09,$F2,$4C,$1F,$CB,$23,$9E,$1F,$8E,$79,$46,$DA
	.dc.b	$FE,$16,$3D,$4F,$43,$05,$C3,$E1,$BC,$6B,$6E,$8F,$03,$88,$FC,$6C
	.dc.b	$60,$E5,$7F,$3A,$B4,$FC,$7B,$F3,$CF,$F0,$57,$CB,$5C,$8F,$0C,$73
	.dc.b	$3C,$0E,$09,$3D,$63,$B9,$F3,$AF,$E4,$D7,$49,$D4,$F3,$AE,$2F,$DD
	.dc.b	$31,$1E,$5A,$E2,$7A,$23,$53,$C1,$FA,$2F,$9E,$78,$47,$88,$FD,$F1
	.dc.b	$A7,$2F,$63,$F8,$0D,$79,$5F,$41,$F0,$EB,$C0,$F1,$36,$FF,$AB,$5F
	.dc.b	$C0,$9F,$91,$29,$D8,$C4,$7E,$3F,$D2,$FC,$E1,$D1,$C4,$7E,$9A,$31
	.dc.b	$3A,$8C,$1E,$B9,$4E,$63,$F2,$0F,$33,$17,$FA,$E7,$11,$C8,$D9,$CD
	.dc.b	$F5,$4F,$48,$FF,$64,$35,$F6,$1F,$AC,$8D,$9F,$E9,$EF,$82,$91,$D6
	.dc.b	$BC,$0F,$F8,$EE,$32,$F8,$09,$1F,$E2,$68,$CB,$C2,$3F,$CA,$91,$8C
	.dc.b	$3F,$CA,$6F,$B1,$B1,$D4,$87,$47,$2F,$23,$33,$D4,$F0,$3F,$E5,$79
	.dc.b	$FF,$32,$11,$C1,$FE,$21,$F8,$C7,$C6,$FF,$B8,$1B,$1D,$2F,$BD,$FF
	.dc.b	$90,$E7,$33,$FE,$2E,$6C,$78,$23,$C8,$66,$BA,$1E,$F7,$FE,$00,$CF
	.dc.b	$0A,$FE,$38,$D8,$CC,$ED,$46,$5B,$1D,$1F,$31,$C0,$F6,$39,$7C,$75
	.dc.b	$E9,$8F,$CE,$4F,$AA,$7F,$38,$D7,$C7,$1A,$3F,$98,$4F,$F9,$8F,$E3
	.dc.b	$5C,$98,$C3,$F9,$A4,$62,$FC,$27,$F9,$37,$E6,$1E,$77,$EA,$DB,$1F
	.dc.b	$90,$0E,$1F,$C9,$B5,$FD,$87,$0F,$F9,$C9,$E0,$D7,$F3,$2D,$77,$C3
	.dc.b	$FB,$98,$7C,$1D,$24,$F2,$7F,$7D,$C8,$FA,$7A,$49,$F3,$46,$5F,$47
	.dc.b	$C6,$79,$5F,$EA,$8D,$41,$C3,$72,$FE,$19,$0F,$1E,$92,$7A,$BF,$C4
	.dc.b	$7E,$D9,$F2,$D4,$93,$F2,$63,$2F,$CB,$F8,$E7,$A5,$FC,$65,$B1,$98
	.dc.b	$C1,$9F,$8E,$91,$C5,$AF,$F0,$4F,$B6,$78,$7F,$C8,$DB,$1F,$65,$AE
	.dc.b	$E3,$F9,$47,$C7,$7F,$BF,$60,$D2,$7C,$7E,$BE,$2E,$4F,$E0,$E5,$18
	.dc.b	$8F,$E2,$BF,$33,$EE,$6E,$BF,$CD,$63,$11,$FC,$D5,$AF,$F8,$1F,$E4
	.dc.b	$3F,$03,$F1,$DF,$8F,$6C,$7F,$20,$5F,$20,$F9,$0D,$73,$AF,$F2,$DF
	.dc.b	$21,$E4,$ED,$BA,$7F,$B1,$72,$FF,$0A,$4E,$5F,$E5,$6D,$7F,$A4,$7C
	.dc.b	$73,$A7,$84,$7E,$1A,$31,$7F,$A3,$6C,$7F,$34,$6B,$CB,$E3,$F8,$A8
	.dc.b	$46,$1B,$3C,$E3,$7F,$1E,$F8,$FA,$49,$FD,$94,$65,$FD,$A7,$8E,$F4
	.dc.b	$3F,$DE,$5E,$2C,$F8,$D7,$C7,$E2,$FF,$C6,$97,$B7,$FD,$A9,$4A,$DC
	.dc.b	$7F,$B3,$E7,$5A,$7F,$50,$6B,$FD,$B5,$AF,$4C,$3D,$F1,$C2,$FF,$DE
	.dc.b	$8D,$AF,$F7,$BD,$FD,$A7,$B5,$79,$6F,$F6,$7F,$16,$F8,$EC,$62,$F8
	.dc.b	$E7,$B1,$BB,$5E,$F1,$F5,$D6,$9F,$F7,$44,$66,$7F,$B4,$B5,$FF,$03
	.dc.b	$35,$FF,$2C,$35,$FE,$0B,$F1,$AF,$88,$F1,$F1,$BE,$23,$C5,$39,$C3
	.dc.b	$E5,$CE,$5F,$F7,$C4,$66,$FF,$DF,$ED,$CD,$3F,$E6,$78,$FE,$26,$F0
	.dc.b	$C3,$DD,$3C,$E8,$F0,$7F,$8C,$BF,$C5,$66,$9F,$D0,$B0,$E3,$E3,$72
	.dc.b	$64,$6A,$BF,$C4,$DF,$1B,$FC,$8B,$E2,$FF,$88,$C7,$6A,$F9,$CF,$F2
	.dc.b	$69,$D7,$8D,$D6,$EC,$F2,$17,$CB,$1F,$43,$D2,$39,$1A,$AE,$4D,$39
	.dc.b	$47,$1C,$3D,$CF,$18,$61,$E0,$7C,$D8,$57,$AC,$F1,$8C,$47,$03,$B1
	.dc.b	$C0,$A7,$0B,$F1,$CF,$7A,$ED,$87,$AA,$7F,$C6,$31,$8C,$3C,$B1,$8B
	.dc.b	$FF,$29,$4F,$F9,$4D,$F0,$1F,$DA,$5F,$ED,$5F,$16,$F8,$D7,$C6,$E1
	.dc.b	$B5,$FE,$3E,$9D,$4E,$4E,$3E,$6F,$E0,$A7,$4D,$7F,$CC,$13,$FD,$EB
	.dc.b	$2D,$92,$3E,$0E,$3F,$9B,$7E,$53,$5F,$8B,$3C,$5B,$5D,$F3,$FC,$2D
	.dc.b	$6D,$BB,$7F,$AF,$DF,$C4,$DF,$16,$D7,$1A,$FF,$62,$1C,$4B,$5C,$5F
	.dc.b	$F6,$45,$71,$6C,$F1,$BF,$EE,$F1,$C5,$B5,$C5,$FE,$F7,$5C,$5B,$5C
	.dc.b	$5F,$ED,$F3,$B6,$9F,$8B,$9E,$0E,$0F,$D7,$EB,$E6,$7F,$50,$67,$1F
	.dc.b	$D9,$29,$E3,$87,$CB,$9C,$B6,$B9,$57,$43,$CB,$3F,$9B,$F8,$DF,$E1
	.dc.b	$91,$D9,$AE,$C3,$88,$F0,$BF,$E3,$36,$ED,$FE,$3F,$9F,$F0,$EC,$EE
	.dc.b	$BF,$DE,$8D,$8C,$E5,$B5,$EC,$BE,$33,$0D,$8F,$82,$D7,$D5,$8B,$E3
	.dc.b	$3F,$C7,$31,$AA,$FF,$5A,$56,$E3,$FD,$68,$D7,$FC,$62,$78,$23,$7F
	.dc.b	$17,$E0,$24,$7F,$56,$AF,$F0,$0B,$E2,$8C,$BF,$16,$91,$F0,$6F,$C9
	.dc.b	$E3,$EA,$BC,$33,$D9,$FF,$2E,$57,$06,$FB,$FE,$82,$38,$37,$CB,$F1
	.dc.b	$08,$FC,$48,$AF,$FA,$DA,$7F,$C9,$0F,$02,$1F,$EB,$AF,$7D,$F1,$7F
	.dc.b	$ED,$96,$C7,$F9,$BD,$AE,$E3,$FD,$BD,$F1,$6F,$88,$FB,$0D,$9F,$CA
	.dc.b	$B5,$FC,$45,$72,$D3,$F2,$F3,$99,$FF,$73,$35,$FA,$6D,$7F,$CC,$AD
	.dc.b	$7F,$0B,$F3,$DF,$70,$8F,$CA,$7F,$DE,$5F,$19,$F8,$A8,$FF,$44,$35
	.dc.b	$FF,$6C,$B5,$FF,$7B,$35,$FC,$1A,$F8,$CE,$64,$8E,$6F,$6F,$8E,$91
	.dc.b	$D9,$AF,$FB,$9B,$E3,$3F,$CB,$84,$F5,$BF,$D3,$50,$8C,$E9,$87,$DC
	.dc.b	$47,$1E,$94,$9F,$F5,$14,$FF,$94,$5B,$1D,$3E,$2D,$F1,$94,$93,$FE
	.dc.b	$A9,$8C,$B7,$D9,$E0,$47,$07,$FA,$BF,$CA,$FF,$80,$0F,$E9,$8F,$F1
	.dc.b	$F7,$99,$EA,$8C,$C7,$EA,$9A,$FF,$A7,$17,$C6,$6C,$91,$D5,$F8,$7F
	.dc.b	$19,$FE,$C3,$8E,$C1,$C3,$E1,$6B,$FB,$86,$BF,$EA,$AF,$8C,$74,$7F
	.dc.b	$0B,$6C,$7F,$28,$6B,$FC,$23,$EB,$BF,$C0,$DE,$0F,$F0,$29,$CC,$FB
	.dc.b	$DA,$E6,$BE,$CC,$62,$3C,$11,$FC,$67,$A3,$FC,$F1,$1F,$CD,$9A,$FF
	.dc.b	$29,$9F,$0F,$47,$F3,$86,$DF,$F1,$2A,$DC,$FF,$3C,$6C,$78,$6B,$D5
	.dc.b	$7D,$FC,$B9,$18,$21,$FB,$07,$2F,$86,$3E,$21,$8B,$ED,$38,$9E,$31
	.dc.b	$87,$D5,$19,$7A,$18,$21,$D4,$B1,$C5,$D9,$92,$B9,$25,$7B,$63,$2F
	.dc.b	$C3,$9F,$33,$99,$DC,$FF,$BA,$2F,$E2,$B6,$BC,$D3,$A8,$4E,$9B,$3A
	.dc.b	$AE,$D1,$E8,$6B,$E7,$AF,$53,$FA,$1A,$FB,$6D,$AF,$E5,$04,$7F,$BA
	.dc.b	$FC,$2B,$E1,$7F,$AE,$36,$3B,$1E,$73,$71,$A3,$DE,$E4,$FE,$E1,$1C
	.dc.b	$5B,$1A,$7D,$67,$22,$1F,$28,$E5,$7F,$E0,$69,$E4,$D9,$E5,$7F,$16
	.dc.b	$73,$87,$F6,$46,$DF,$28,$F8,$B7,$C6,$3B,$BF,$3E,$77,$06,$98,$AE
	.dc.b	$D7,$EF,$9E,$37,$D0,$F0,$46,$63,$DE,$7C,$83,$A3,$9B,$F1,$47,$BC
	.dc.b	$C9,$D0,$18,$35,$7E,$49,$E5,$5E,$19,$DC,$E5,$E4,$7C,$E3,$C6,$4A
	.dc.b	$F0,$9F,$06,$B9,$C7,$27,$EB,$4F,$38,$E1,$5D,$DF,$25,$72,$8F,$83
	.dc.b	$1E,$F9,$E1,$39,$8F,$A4,$75,$3C,$B1,$F3,$70,$F0,$C7,$4A,$E6,$FC
	.dc.b	$0B,$CD,$6E,$38,$C7,$0C,$3E,$AC,$60,$F1,$47,$03,$D6,$57,$26,$CE
	.dc.b	$C5,$C3,$E2,$36,$FE,$05,$FB,$1F,$2C,$EA,$7E,$29,$ED,$39,$98,$3D
	.dc.b	$87,$C2,$29,$27,$D8,$ED,$EC,$ED,$AF,$0B,$F7,$BB,$BF,$6B,$FD,$48
	.dc.b	$FC,$41,$FD,$86,$7F,$64,$D8,$FE,$F7,$18,$8F,$9E,$64,$E4,$E5,$F6
	.dc.b	$57,$EC,$FE,$37,$F8,$29,$F2,$E3,$F7,$EE,$1F,$90,$7E,$51,$C3,$E7
	.dc.b	$3F,$87,$B8,$8F,$C3,$47,$C3,$8C,$61,$F3,$89,$1C,$CE,$27,$63,$F7
	.dc.b	$11,$C1,$F3,$C7,$07,$27,$03,$C4,$7A,$0C,$4F,$8A,$BF,$80,$D7,$39
	.dc.b	$FA,$D7,$65,$3F,$06,$BE,$25,$72,$BF,$05,$72,$8C,$1E,$42,$1E,$43
	.dc.b	$C4,$79,$27,$9B,$87,$C9,$7E,$1B,$F0,$5F,$F1,$F7,$C9,$1F,$A6,$6D
	.dc.b	$7E,$B9,$AF,$E3,$5B,$5E,$F6,$D7,$A1,$B3,$E6,$BF,$7B,$EA,$3E,$19
	.dc.b	$5C,$BE,$99,$F2,$39,$BF,$23,$4F,$21,$4F,$21,$6F,$C8,$D9,$E6,$FC
	.dc.b	$36,$C6,$A3,$07,$E1,$4E,$66,$0F,$9C,$EC,$F9,$A4,$8F,$81,$1E,$BB
	.dc.b	$F3,$BC,$6F,$DB,$5C,$6B,$C2,$49,$F8,$D5,$FA,$C2,$3F,$AA,$6B,$FE
	.dc.b	$D3,$F8,$DF,$F3,$E5,$7F,$B8,$9B,$5C,$E7,$CE,$FF,$AF,$A3,$31,$FD
	.dc.b	$11,$AE,$8F,$81,$3C,$FB,$FF,$DA,$93,$AC,$3D,$37,$CE,$72,$4B,$FB
	.dc.b	$95,$D0,$91,$FA,$16,$C7,$FB,$E9,$E9,$87,$18,$FF,$30,$C7,$FA,$4D
	.dc.b	$F3,$4F,$FA,$22,$31,$07,$B7,$17,$98,$E5,$8F,$A6,$7B,$46,$5F,$05
	.dc.b	$6B,$0F,$89,$39,$9F,$99,$36,$BE,$64,$DA,$F8,$15,$D0,$87,$D3,$8E
	.dc.b	$8F,$CD,$78,$34,$F0,$9C,$0F,$F4,$65,$FF,$A2,$1F,$64,$7B,$C8,$FF
	.dc.b	$9A,$FE,$37,$FC,$DC,$74,$5F,$3C,$F9,$4F,$F5,$6B,$63,$F9,$D3,$5F
	.dc.b	$F4,$6B,$5F,$F4,$E7,$C5,$BC,$9F,$F5,$AB,$63,$FA,$13,$5F,$F4,$AB
	.dc.b	$5F,$F5,$26,$FF,$E1,$F7,$FA,$43,$5F,$F4,$DB,$5D,$EF,$FE,$C2,$F5
	.dc.b	$7E,$01,$33,$C1,$FE,$C6,$E2,$7C,$5D,$25,$FF,$55,$8C,$BF,$D5,$FE
	.dc.b	$2F,$C8,$FF,$88,$03,$87,$FC,$46,$D7,$39,$FF,$B5,$3D,$B7,$9B,$49
	.dc.b	$3F,$D8,$E3,$2F,$F6,$5F,$8B,$7C,$3F,$FB,$71,$B1,$98,$C1,$FE,$E0
	.dc.b	$F8,$B7,$A7,$49,$3F,$DB,$E3,$2F,$F7,$1F,$B3,$7C,$9F,$FB,$A1,$B1
	.dc.b	$98,$C1,$FE,$EC,$F8,$C7,$89,$49,$3F,$DE,$E3,$2F,$F7,$DF,$8C,$79
	.dc.b	$7F,$EF,$46,$C6,$63,$07,$FB,$E3,$D4,$FF,$74,$C7,$FB,$59,$AF,$FB
	.dc.b	$95,$AF,$D1,$6B,$FE,$52,$F8,$CF,$8A,$FF,$B7,$9A,$FF,$BB,$5A,$FD
	.dc.b	$96,$BF,$E5,$AF,$8C,$7C,$63,$E3,$3F,$C2,$D1,$86,$DB,$CE,$CD,$6E
	.dc.b	$FF,$DD,$BF,$39,$FE,$75,$F0,$FF,$C3,$CD,$8C,$F6,$A7,$2D,$8F,$F7
	.dc.b	$AF,$C5,$BA,$F1,$1F,$8F,$11,$9A,$FF,$15,$36,$38,$CF,$07,$F2,$D0
	.dc.b	$F8,$BC,$47,$F9,$E2,$31,$5C,$A7,$33,$D9,$BD,$CF,$3F,$CD,$E3,$CF
	.dc.b	$89,$76,$E3,$1F,$D2,$5E,$DF,$F4,$2D,$6E,$75,$8F,$FA,$22,$B4,$A7
	.dc.b	$55,$C0,$E7,$87,$08,$61,$E0,$9D,$57,$3A,$F3,$92,$3F,$C9,$D7,$DA
	.dc.b	$7E,$57,$C3,$BE,$27,$A7,$88,$F8,$9F,$2D,$2B,$58,$76,$CD,$7B,$8F
	.dc.b	$0D,$6F,$0E,$78,$72,$C3,$8C,$FB,$6B,$E1,$DF,$18,$C1,$FE,$AA,$F8
	.dc.b	$97,$C2,$FB,$F8,$5F,$09,$E4,$9C,$E1,$E3,$9C,$BE,$28,$CD,$F4,$0E
	.dc.b	$34,$FC,$5B,$FE,$B8,$9E,$18,$7C,$09,$E7,$5C,$57,$C0,$E7,$1D,$36
	.dc.b	$FE,$0C,$F0,$3F,$1B,$F1,$3F,$EC,$4F,$87,$FE,$1F,$87,$DD,$C3,$E6
	.dc.b	$BF,$EC,$89,$D7,$8B,$FC,$AE,$FC,$37,$C2,$B4,$1E,$71,$A9,$FD,$24
	.dc.b	$AB,$9B,$F9,$0B,$FD,$65,$F3,$8D,$1C,$EB,$C1,$3A,$79,$D7,$29,$F4
	.dc.b	$61,$5F,$1D,$F8,$AF,$84,$7E,$D6,$B8,$61,$E5,$8C,$1D,$CF,$25,$F3
	.dc.b	$7C,$77,$CE,$31,$5C,$AF,$AD,$7F,$4F,$8F,$F5,$43,$F1,$0F,$F5,$53
	.dc.b	$87,$DA,$7F,$07,$F8,$77,$C4,$BE,$27,$FD,$61,$18,$6C,$66,$7E,$5C
	.dc.b	$E5,$BE,$67,$ED,$5B,$9F,$28,$C4,$77,$6B,$FD,$96,$7F,$C0,$13,$F2
	.dc.b	$F4,$FC,$EC,$DD,$BF,$B5,$6B,$F8,$6B,$DC,$FC,$70,$E2,$3F,$17,$5C
	.dc.b	$A7,$4D,$8E,$58,$79,$DB,$87,$7A,$87,$5B,$E9,$7E,$7B,$E0,$FF,$7D
	.dc.b	$3A,$17,$0F,$3B,$BB,$FC,$35,$72,$B1,$FD,$CA,$B6,$E2,$B9,$BA,$9E
	.dc.b	$0E,$CF,$E1,$71,$ED,$AF,$E8,$2D,$D7,$C1,$3F,$02,$FF,$51,$1E,$58
	.dc.b	$E3,$18,$7B,$61,$EC,$7C,$BF,$13,$FC,$31,$FE,$32,$D7,$F8,$43,$70
	.dc.b	$FE,$A2,$FF,$8D,$5B,$B7,$59,$4E,$F0,$FF,$12,$36,$3B,$DE,$FF,$74
	.dc.b	$F5,$47,$F3,$FF,$8A,$7D,$47,$9F,$24,$62,$7F,$96,$D7,$04,$F2,$9D
	.dc.b	$D7,$1A,$E1,$5F,$C8,$6B,$81,$FA,$52,$3F,$AA,$6B,$FC,$F6,$3F,$C9
	.dc.b	$95,$FD,$79,$B8,$FF,$8F,$EB,$FB,$94,$FF,$74,$6C,$7F,$5E,$9D,$D7
	.dc.b	$B7,$BF,$E0,$3F,$1F,$F1,$67,$C3,$7E,$E9,$B1,$FC,$68,$38,$7F,$8E
	.dc.b	$B5,$FF,$01,$39,$C7,$FC,$0B,$F5,$DF,$E0,$79,$FE,$A6,$D7,$FA,$14
	.dc.b	$EF,$1F,$F2,$C4,$E6,$7F,$C1,$8D,$73,$D1,$FC,$3A,$31,$1F,$9C,$F7
	.dc.b	$3F,$A1,$BE,$76,$BF,$CC,$9A,$FF,$85,$DA,$FF,$99,$DD,$E1,$FE,$1C
	.dc.b	$7F,$C8,$F1,$FD,$5D,$B1,$9E,$8F,$C7,$46,$A7,$A7,$9D,$FC,$8A,$31
	.dc.b	$1D,$63,$F9,$2F,$C4,$BE,$8E,$92,$7E,$74,$65,$B1,$F9,$B8,$FF,$24
	.dc.b	$E1,$FE,$30,$6E,$7F,$E7,$F7,$85,$7F,$6D,$7F,$C9,$29,$A4,$F5,$3F
	.dc.b	$9B,$FC,$47,$FA,$2A,$7F,$AB,$B5,$FF,$1E,$35,$FF,$46,$35,$FE,$16
	.dc.b	$F1,$BF,$F2,$1A,$F9,$FF,$D9,$36,$7F,$C2,$6E,$F0,$E7,$3E,$22,$4F
	.dc.b	$E3,$E7,$83,$FD,$AB,$E2,$3F,$D3,$04,$E6,$FE,$9E,$D8,$FF,$0F,$BF
	.dc.b	$BD,$4D,$3F,$83,$5F,$FA,$95,$FE,$19,$1F,$E5,$96,$B9,$E8,$EA,$7F
	.dc.b	$5C,$F8,$7F,$F4,$2C,$7F,$7C,$6B,$FE,$61,$6B,$FE,$B1,$6B,$FC,$C1
	.dc.b	$E3,$97,$F9,$A1,$7C,$0F,$F1,$76,$CF,$F8,$E1,$DE,$1C,$E7,$C4,$49
	.dc.b	$FD,$A4,$F0,$7F,$C1,$FF,$08,$F8,$EF,$E4,$5F,$1C,$F8,$CA,$7F,$25
	.dc.b	$9D,$57,$F9,$1A,$74,$D9,$DB,$FE,$4A,$9D,$BF,$E3,$99,$DB,$67,$6F
	.dc.b	$F8,$F6,$77,$97,$FA,$1A,$B6,$DA,$DB,$FE,$8A,$9D,$E5,$FE,$D9,$AD
	.dc.b	$B6,$B6,$FF,$B7,$67,$77,$FD,$72,$B6,$DA,$DB,$FD,$82,$76,$A7,$6D
	.dc.b	$9D,$A9,$DC,$E8,$93,$F2,$70,$DC,$62,$BE,$34,$62,$79,$D7,$27,$13
	.dc.b	$A3,$E7,$10,$F9,$46,$4A,$4C,$3E,$43,$96,$B9,$C3,$E3,$39,$6B,$93
	.dc.b	$E2,$19,$29,$FE,$56,$C3,$ED,$BF,$EB,$B6,$BC,$8C,$57,$F6,$DE,$AF
	.dc.b	$F5,$84,$7F,$A3,$9A,$F7,$3E,$61,$9F,$33,$81,$F0,$9F,$F4,$14,$70
	.dc.b	$C3,$FD,$96,$DC,$3D,$31,$EE,$33,$5C,$AB,$E4,$46,$67,$FC,$3B,$0A
	.dc.b	$E4,$FF,$88,$1A,$F9,$0F,$33,$83,$38,$7D,$B8,$F2,$DF,$77,$6D,$BF
	.dc.b	$1C,$78,$AF,$C3,$7C,$EB,$51,$EA,$32,$E2,$7C,$8F,$E4,$27,$84,$7C
	.dc.b	$73,$A1,$A3,$07,$90,$C9,$EF,$8E,$8D,$33,$1D,$1E,$07,$A8,$E7,$5E
	.dc.b	$A7,$06,$9C,$1F,$35,$F1,$19,$AF,$5B,$DA,$BC,$35,$F2,$B0,$B1,$E4
	.dc.b	$C3,$F5,$F3,$E3,$C3,$AE,$1F,$17,$0E,$B1,$86,$C6,$5F,$61,$E1,$32
	.dc.b	$7E,$6A,$38,$3A,$6D,$78,$2F,$C7,$3E,$73,$E0,$D7,$A9,$F8,$35,$CA
	.dc.b	$FD,$51,$A3,$B9,$D6,$7C,$91,$EB,$79,$4F,$C7,$9D,$1C,$27,$9B,$E9
	.dc.b	$32,$4B,$F3,$3D,$CC,$61,$E8,$78,$46,$9E,$53,$D2,$BD,$18,$7C,$FB
	.dc.b	$E1,$1F,$09,$B8,$79,$8F,$4C,$FF,$AE,$2B,$F7,$37,$FA,$A3,$E7,$95
	.dc.b	$CB,$F1,$CC,$47,$12,$3F,$E4,$78,$FD,$7E,$2F,$2F,$FA,$81,$E2,$9F
	.dc.b	$E6,$8E,$1F,$81,$3C,$23,$11,$EE,$AB,$8F,$F4,$79,$E5,$3F,$CD,$DC
	.dc.b	$E1,$ED,$8F,$AF,$87,$85,$FE,$B5,$3C,$23,$11,$FD,$5E,$30,$72,$7E
	.dc.b	$C5,$D7,$B1,$E6,$71,$3C,$AF,$E8,$CF,$62,$1E,$92,$BE,$97,$91,$8B
	.dc.b	$FE,$B9,$7E,$28,$C1,$E2,$29,$E2,$3A,$1E,$93,$38,$7E,$4D,$F9,$A6
	.dc.b	$30,$FE,$0B,$87,$9E,$FB,$CF,$88,$E4,$4C,$3D,$6D,$8D,$BE,$08,$F5
	.dc.b	$39,$3C,$C5,$3C,$CF,$23,$64,$8F,$AD,$1F,$A2,$9F,$23,$E8,$3B,$47
	.dc.b	$82,$78,$47,$66,$BD,$A3,$9D,$FE,$26,$3E,$01,$FA,$BA,$EA,$EC,$E5
	.dc.b	$3D,$48,$FE,$23,$0F,$BA,$7C,$82,$BC,$DA,$FE,$EE,$7A,$92,$38,$47
	.dc.b	$D8,$8F,$0D,$79,$63,$E3,$46,$30,$E7,$87,$B2,$B5,$5F,$B3,$AF,$CB
	.dc.b	$B6,$FF,$D6,$8F,$E0,$CC,$1F,$EF,$EA,$61,$FD,$93,$E7,$BF,$B0,$9E
	.dc.b	$18,$F9,$2E,$31,$F0,$CE,$DF,$F7,$54,$6E,$3E,$2D,$23,$83,$ED,$BD
	.dc.b	$D4,$8C,$1E,$3A,$E8,$D0,$73,$8F,$0B,$8A,$EA,$FE,$2D,$B7,$EC,$8D
	.dc.b	$4F,$ED,$67,$35,$DA,$78,$4F,$8D,$57,$17,$C0,$7E,$B2,$B8,$BD,$0C
	.dc.b	$4F,$19,$E1,$3F,$0E,$74,$74,$AF,$04,$EA,$BC,$87,$D6,$AF,$E5,$35
	.dc.b	$C0,$FE,$5C,$E1,$F1,$9C,$CF,$3F,$CF,$3D,$DF,$F0,$ED,$FC,$11,$19
	.dc.b	$9F,$98,$72,$8F,$77,$C2,$3E,$6F,$F0,$ED,$9F,$F7,$E3,$5F,$E8,$01
	.dc.b	$C3,$FD,$0D,$7C,$77,$22,$47,$27,$AF,$8B,$23,$AB,$5E,$BF,$0A,$FA
	.dc.b	$3F,$CC,$B6,$3C,$21,$C3,$C2,$37,$F2,$72,$37,$28,$C4,$76,$F3,$BA
	.dc.b	$91,$DB,$FA,$8F,$85,$7A,$B4,$93,$E2,$8C,$B7,$7E,$04,$70,$6F,$E2
	.dc.b	$FA,$11,$C1,$EC,$FC,$61,$F3,$DF,$B8,$6C,$66,$30,$67,$96,$47,$16
	.dc.b	$BF,$5B,$E1,$9F,$67,$FC,$09,$B1,$F2,$9A,$EE,$3F,$BF,$7C,$84,$8F
	.dc.b	$EE,$CD,$7F,$03,$F2,$0F,$77,$F8,$6B,$63,$EA,$B5,$DC,$7F,$12,$F9
	.dc.b	$F7,$CF,$D2,$4F,$DC,$8C,$BF,$7B,$E2,$5F,$0C,$F8,$9C,$36,$7F,$27
	.dc.b	$2F,$89,$7C,$4E,$AF,$F0,$FF,$36,$F9,$CF,$DE,$61,$FE,$82,$8C,$70
	.dc.b	$EA,$73,$3F,$C4,$27,$2F,$E8,$3E,$85,$F2,$5A,$84,$62,$FF,$2A,$D8
	.dc.b	$E1,$3B,$FC,$A3,$CB,$5F,$9A,$6C,$FE,$B1,$AF,$F8,$B5,$AF,$FA,$61
	.dc.b	$DF,$87,$FC,$55,$B3,$FC,$EF,$E2,$BE,$3B,$FB,$E8,$CD,$7F,$2A,$6A
	.dc.b	$73,$4A,$E3,$3F,$C3,$4D,$69,$FC,$2E,$FF,$D5,$3F,$14,$FA,$1F,$84
	.dc.b	$D8,$FD,$AB,$6B,$F8,$3B,$5F,$F2,$43,$9C,$3F,$97,$36,$7F,$AA,$7C
	.dc.b	$5B,$CE,$F0,$5F,$F4,$2A,$D3,$FD,$13,$E2,$DF,$0D,$AC,$5F,$02,$F5
	.dc.b	$B8,$DA,$F7,$9F,$F4,$DB,$DD,$6A,$73,$5F,$05,$B3,$FD,$0D,$FE,$A7
	.dc.b	$5F,$CA,$9C,$BF,$CC,$2B,$BC,$BE,$02,$C6,$1B,$3F,$CC,$A7,$CC,$FF
	.dc.b	$34,$BF,$EB,$FF,$17,$FD,$83,$E2,$9F,$13,$D6,$B5,$4A,$D4,$E6,$15
	.dc.b	$C1,$FF,$3B,$4E,$9F,$F0,$06,$3F,$D9,$67,$51,$FE,$E4,$9D,$3F,$D8
	.dc.b	$A7,$55,$FD,$02,$73,$87,$A2,$BB,$CE,$5A,$76,$C3,$9B,$EC,$32,$75
	.dc.b	$72,$FA,$CC,$CF,$5C,$3C,$B1,$C5,$F3,$C6,$6B,$C1,$18,$C3,$CF,$5F
	.dc.b	$00,$E6,$E1,$F7,$1F,$E2,$4C,$3F,$C4,$AF,$84,$FE,$9A,$FF,$4E,$F8
	.dc.b	$A7,$94,$F8,$BC,$36,$BF,$B1,$CE,$A7,$27,$1F,$7B,$FC,$17,$3A,$6B
	.dc.b	$FE,$A9,$94,$62,$3F,$C3,$E2,$33,$9F,$33,$EA,$E7,$FA,$09,$BC,$1C
	.dc.b	$34,$FB,$77,$C2,$7E,$7B,$63,$EF,$D7,$19,$FA,$6D,$8E,$38,$7A,$1B
	.dc.b	$87,$6A,$87,$49,$D5,$7A,$2F,$B3,$EA,$3A,$17,$0F,$33,$B9,$E3,$5F
	.dc.b	$E3,$EA,$E2,$FF,$90,$27,$8B,$6B,$8B,$BA,$E2,$FE,$9A,$3F,$C4,$3A
	.dc.b	$F7,$8F,$F0,$FD,$7B,$F0,$FF,$1E,$61,$DA,$78,$1A,$BE,$2F,$73,$FC
	.dc.b	$4F,$1B,$F8,$BF,$E1,$91,$FE,$2A,$6B,$FC,$9D,$B5,$FE,$49,$7E,$EC
	.dc.b	$6F,$6E,$AA,$37,$8F,$E0,$1B,$1F,$E6,$07,$8B,$FC,$CF,$E2,$FE,$C3
	.dc.b	$FE,$F1,$6B,$F8,$86,$BF,$E6,$86,$A7,$E1,$4F,$FA,$07,$E2,$DE,$9F
	.dc.b	$C0,$C5,$18,$BF,$F2,$3D,$F0,$7A,$1E,$A7,$F2,$EF,$94,$E5,$1F,$E4
	.dc.b	$27,$89,$FE,$92,$8F,$D7,$91,$C1,$C3,$DC,$8F,$03,$FC,$AF,$F1,$4F
	.dc.b	$8B,$F9,$2D,$8C,$C7,$A7,$9D,$F4,$7A,$8F,$F5,$0D,$6E,$BF,$CB,$2D
	.dc.b	$7D,$2D,$7F,$62,$EF,$2F,$F4,$17,$C6,$7F,$4C,$9F,$F3,$4B,$5F,$EC
	.dc.b	$D3,$BD,$7F,$73,$F1,$9F,$D0,$67,$FC,$DE,$D7,$FB,$63,$6F,$FD,$1C
	.dc.b	$BE,$67,$F0,$AE,$E7,$E6,$B5,$EB,$D5,$FE,$BC,$8C,$47,$FA,$99,$E9
	.dc.b	$7E,$2F,$8D,$E6,$74,$AF,$C8,$FC,$63,$E3,$69,$30,$F3,$FD,$9D,$7F
	.dc.b	$CF,$F3,$B7,$FC,$E2,$D7,$33,$FE,$CE,$8C,$67,$FE,$A9,$5F,$1D,$E0
	.dc.b	$3A,$47,$FA,$B7,$E3,$5F,$31,$FA,$46,$E1,$FE,$96,$6B,$FE,$13,$BF
	.dc.b	$F1,$F3,$99,$FF,$6E,$CE,$A3,$F2,$6D,$7F,$95,$B5,$FF,$65,$BD,$2B
	.dc.b	$C5,$F1,$FE,$13,$C5,$5F,$B6,$F8,$FF,$D8,$CF,$FB,$0D,$BB,$7F,$36
	.dc.b	$6E,$1D,$3B,$78,$77,$FF,$BC,$27,$F7,$33,$FA,$66,$BC,$63,$47,$7E
	.dc.b	$D9,$1F,$EB,$69,$EA,$BE,$39,$F2,$BF,$C2,$9B,$1F,$EB,$C6,$BF,$E3
	.dc.b	$9F,$8B,$7D,$2E,$EB,$FD,$53,$3C,$DB,$5C,$23,$F7,$6D,$7F,$A9,$BD
	.dc.b	$34,$FF,$72,$2F,$91,$F1,$1D,$67,$FD,$9B,$F1,$EF,$91,$FE,$3E,$D8
	.dc.b	$CC,$7F,$0F,$FB,$2F,$F4,$8E,$1F,$EE,$FA,$D5,$7F,$B3,$9A,$FF,$0E
	.dc.b	$6B,$FD,$99,$A8,$3C,$A3,$FD,$F5,$F2,$5F,$E4,$B9,$FF,$70,$B5,$FF
	.dc.b	$2F,$4E,$F8,$7A,$61,$FC,$DA,$B5,$7C,$BC,$FF,$5C,$CC,$39,$C6,$CC
	.dc.b	$3D,$23,$0E,$8F,$82,$43,$FC,$DD,$5D,$AF,$FC,$02,$D8,$F6,$61,$E5
	.dc.b	$6D,$74,$78,$DF,$83,$0F,$95,$39,$7D,$13,$DD,$CB,$E3,$38,$99,$C3
	.dc.b	$E0,$B9,$8E,$2F,$AC,$D1,$F3,$F0,$F3,$26,$E1,$E7,$8F,$A7,$7E,$3B
	.dc.b	$F6,$E1,$E7,$7B,$CF,$88,$98,$79,$9F,$01,$88,$E2,$F8,$8F,$04,$F5
	.dc.b	$7C,$8F,$56,$9E,$B8,$E2,$F4,$9F,$69,$23,$2F,$84,$C4,$72,$3C,$47
	.dc.b	$D2,$3A,$19,$76,$FC,$03,$27,$98,$B7,$F1,$70,$F9,$24,$BF,$44,$EC
	.dc.b	$EE,$5A,$F2,$5F,$98,$91,$E1,$7E,$5C,$78,$AB,$A4,$FC,$76,$BD,$2F
	.dc.b	$C5,$7F,$26,$3B,$5F,$23,$AB,$83,$B1,$F0,$4F,$2D,$74,$BF,$2C,$68
	.dc.b	$E0,$4B,$EE,$FC,$49,$FC,$3E,$1E,$17,$95,$F8,$67,$94,$61,$F0,$5F
	.dc.b	$B7,$0F,$46,$1D,$27,$C7,$1F,$2A,$B8,$47,$87,$0E,$D3,$C1,$B1,$F5
	.dc.b	$1A,$FC,$56,$DF,$DC,$BE,$CF,$86,$7E,$65,$6B,$0F,$44,$7C,$0C,$3C
	.dc.b	$37,$C2,$7E,$5C,$7C,$09,$F8,$55,$E3,$69,$E3,$29,$E3,$2D,$F8,$F0
	.dc.b	$F0,$47,$CF,$C3,$B3,$F2,$2B,$E1,$1A,$7F,$28,$7C,$63,$B1,$FB,$13
	.dc.b	$FC,$77,$7F,$C9,$5B,$1F,$D2,$23,$11,$FA,$A3,$27,$F9,$08,$FD,$31
	.dc.b	$83,$0F,$F0,$6F,$0D,$BF,$F9,$89,$B3,$EC,$CB,$FD,$43,$18,$86,$49
	.dc.b	$CD,$7A,$EB,$53,$9A,$FE,$85,$18,$AF,$9F,$87,$7B,$95,$DC,$DB,$87
	.dc.b	$E2,$1E,$09,$EE,$FF,$4D,$BE,$CA,$FB,$3F,$3C,$D4,$F0,$79,$1D,$8F
	.dc.b	$01,$C6,$BC,$15,$F8,$B8,$D9,$93,$99,$C4,$F2,$BE,$2A,$F0,$C6,$CC
	.dc.b	$D7,$86,$75,$5F,$1C,$EF,$1F,$12,$7A,$9C,$67,$6F,$9A,$BA,$9D,$B0
	.dc.b	$FD,$5B,$8B,$E9,$3C,$A3,$58,$7F,$0B,$75,$1C,$E7,$C2,$FC,$73,$DF
	.dc.b	$1A,$7D,$27,$BE,$FC,$6E,$0F,$36,$1E,$D3,$D4,$F2,$AE,$AF,$EC,$6E
	.dc.b	$DF,$92,$3F,$99,$BD,$EF,$E8,$1D,$89,$7D,$1B,$3E,$38,$C1,$E1,$70
	.dc.b	$7A,$E3,$6F,$98,$C4,$6D,$C1,$E5,$3E,$45,$FA,$30,$FD,$33,$E8,$3E
	.dc.b	$51,$63,$D2,$D7,$FA,$2D,$FA,$EB,$B5,$72,$9E,$A4,$37,$5C,$6F,$FA
	.dc.b	$BD,$FE,$BE,$7F,$D7,$6D,$7F,$34,$62,$7F,$C9,$31,$F9,$0C,$5F,$25
	.dc.b	$FE,$3E,$3A,$63,$FE,$35,$71,$7E,$1C,$3E,$2D,$70,$8F,$F2,$3C,$F0
	.dc.b	$7F,$C3,$EF,$07,$EE,$1C,$6B,$A3,$C0,$C1,$F4,$8F,$F2,$0C,$71,$AE
	.dc.b	$ED,$9E,$33,$D0,$FF,$20,$C6,$5F,$09,$D0,$E6,$77,$36,$66,$BD,$4E
	.dc.b	$09,$5E,$07,$83,$FE,$5D,$9F,$CA,$CF,$08,$C4,$EA,$BF,$0D,$39,$7C
	.dc.b	$F3,$A9,$F4,$35,$FF,$32,$57,$F9,$F2,$BC,$71,$83,$F9,$A1,$EA,$3F
	.dc.b	$86,$FC,$97,$F0,$83,$ED,$4F,$F9,$A6,$3F,$88,$FC,$DB,$CF,$FF,$43
	.dc.b	$B6,$39,$B5,$FE,$80,$D7,$90,$F9,$17,$C9,$7F,$A4,$1B,$1E,$46,$BF
	.dc.b	$D2,$5A,$F2,$5F,$25,$C4,$91,$C5,$FF,$4D,$F9,$5F,$02,$3D,$4D,$7F
	.dc.b	$D0,$8D,$4D,$F9,$17,$C9,$7F,$A9,$1B,$3F,$21,$AF,$F5,$D6,$BE,$E5
	.dc.b	$F2,$5C,$49,$1C,$5F,$F5,$7F,$CD,$BE,$22,$92,$7E,$AC,$65,$FA,$FF
	.dc.b	$24,$F9,$3F,$F5,$BB,$63,$31,$83,$3F,$27,$23,$8B,$5F,$EE,$9F,$34
	.dc.b	$F0,$BF,$D8,$2D,$8F,$C2,$B5,$DC,$7F,$B1,$BE,$4D,$F2,$74,$93,$F8
	.dc.b	$C8,$CB,$7A,$78,$11,$C1,$F9,$2E,$AF,$C2,$37,$8F,$F8,$59,$1D,$7C
	.dc.b	$8F,$F0,$E2,$38,$3F,$37,$F2,$2F,$91,$FF,$6D,$B6,$33,$18,$3F,$DB
	.dc.b	$FF,$34,$F3,$E9,$27,$F4,$B1,$97,$F4,$FF,$20,$F9,$17,$C8,$61,$B3
	.dc.b	$FE,$1F,$97,$C8,$BE,$2F,$57,$FB,$0F,$90,$7C,$8F,$F8,$23,$0F,$D7
	.dc.b	$46,$38,$75,$39,$9F,$F6,$D4,$E5,$FF,$16,$7C,$7F,$5A,$FF,$01,$F8
	.dc.b	$EF,$8F,$78,$FF,$E2,$18,$FF,$0A,$D6,$EF,$FC,$23,$1B,$9D,$5F,$EF
	.dc.b	$EB,$55,$FB,$86,$B9,$CF,$C8,$A3,$2F,$FB,$BD,$ED,$7E,$09,$EB,$3F
	.dc.b	$E3,$8C,$3E,$55,$FF,$83,$AF,$6F,$F0,$FF,$8D,$7C,$7F,$0C,$5F,$1E
	.dc.b	$F6,$79,$E4,$BD,$E1,$F0,$AB,$4F,$F8,$6A,$33,$3F,$E5,$36,$BF,$C5
	.dc.b	$1A,$F8,$70,$F6,$FC,$7B,$D9,$F1,$79,$1F,$A7,$C3,$FC,$4B,$39,$7E
	.dc.b	$8C,$66,$FF,$C6,$F3,$96,$D7,$A2,$11,$63,$D7,$9F,$13,$FC,$96,$3D
	.dc.b	$FC,$F8,$1A,$6E,$3F,$E3,$69,$E0,$FF,$9B,$FE,$3F,$AF,$C7,$3E,$3B
	.dc.b	$F9,$D5,$6B,$2F,$E6,$75,$A9,$CE,$3F,$E6,$7B,$E1,$5A,$0E,$1F,$F3
	.dc.b	$8C,$6E,$7C,$11,$A7,$9E,$1D,$63,$95,$71,$C3,$E3,$3C,$E3,$E1,$E1
	.dc.b	$B7,$AD,$57,$27,$53,$FC,$E1,$87,$85,$EB,$87,$81,$D1,$FD,$66,$BA
	.dc.b	$61,$E0,$9F,$3E,$1E,$3B,$E5,$7B,$38,$D6,$DF,$84,$7F,$68,$BF,$ED
	.dc.b	$2F,$90,$C1,$FE,$9F,$F8,$EF,$EC,$DF,$1F,$FD,$B9,$B1,$9F,$49,$39
	.dc.b	$6C,$71,$F7,$FF,$D3,$AD,$C3,$94,$62,$3F,$B5,$DA,$31,$1F,$DD,$E3
	.dc.b	$13,$F2,$F3,$FC,$D4,$DD,$BF,$BE,$69,$F6,$EF,$73,$F2,$1B,$1F,$7E
	.dc.b	$B8,$CF,$CC,$6C,$71,$C3,$CE,$DC,$3B,$54,$3A,$4E,$AB,$CC,$D8,$FD
	.dc.b	$1C,$74,$3F,$C2,$05,$7F,$4F,$7E,$48,$E3,$5D,$6F,$CD,$87,$B2,$73
	.dc.b	$7E,$78,$FD,$0E,$BD,$1B,$1E,$06,$AC,$3B,$39,$3F,$D4,$75,$F0,$B0
	.dc.b	$FF,$5F,$B6,$FB,$E1,$FA,$5F,$8F,$FE,$D1,$5F,$EC,$46,$CF,$F8,$AD
	.dc.b	$B7,$FC,$09,$FB,$4F,$FA,$C1,$CA,$69,$32,$F3,$FC,$6F,$86,$7F,$85
	.dc.b	$B5,$39,$B7,$C7,$7F,$8B,$1F,$F6,$BC,$7D,$C9,$FF,$1A,$B5,$FF,$20
	.dc.b	$35,$EB,$A7,$E8,$23,$FC,$77,$3C,$A7,$C6,$7C,$C7,$BF,$A9,$88,$FF
	.dc.b	$65,$CF,$0C,$BA,$C6,$F0,$FF,$1D,$BA,$3B,$10,$FF,$2B,$CF,$EC,$17
	.dc.b	$C5,$BE,$37,$EC,$B6,$3F,$DB,$2D,$7F,$AF,$F3,$3E,$3B,$75,$27,$6F
	.dc.b	$A6,$75,$18,$3F,$CC,$F5,$CD,$FF,$2D,$B5,$FF,$33,$35,$FE,$7B,$F1
	.dc.b	$8F,$8D,$FE,$0E,$D8,$FF,$38,$B5,$FF,$32,$87,$0F,$F9,$A1,$FE,$84
	.dc.b	$FF,$BD,$1A,$B5,$69,$FE,$77,$9C,$C6,$27,$C7,$5F,$9E,$9E,$31,$D7
	.dc.b	$A3,$CA,$47,$CB,$F1,$9E,$82,$38,$3F,$D1,$3F,$16,$F8,$CF,$D4,$B6
	.dc.b	$3F,$AE,$E3,$E1,$51,$97,$D1,$97,$9E,$F8,$DF,$F2,$CA,$D5,$7F,$A2
	.dc.b	$1A,$FF,$A4,$DA,$FF,$6A,$7F,$C5,$EF,$93,$C8,$E4,$47,$FD,$3F,$F1
	.dc.b	$8F,$8C,$FD,$FB,$63,$CF,$8F,$FA,$7E,$31,$1E,$63,$B5,$70,$AF,$E8
	.dc.b	$33,$A9,$FF,$4E,$B5,$E5,$1A,$E5,$91,$E2,$6B,$F8,$0F,$8C,$FE,$02
	.dc.b	$F7,$0E,$34,$F9,$4D,$C7,$CD,$5F,$C6,$23,$8D,$7F,$AB,$9A,$FF,$AE
	.dc.b	$5A,$FF,$84,$1A,$BE,$11,$F1,$9F,$D7,$1B,$1F,$EC,$66,$BF,$96,$6B
	.dc.b	$FE,$BC,$F8,$B7,$C6,$7F,$2A,$6C,$7C,$C3,$FC,$3E,$E1,$F7,$4F,$4C
	.dc.b	$BF,$0D,$39,$9E,$95,$FD,$BA,$31,$1F,$ED,$36,$BF,$A1,$F8,$B7,$C9
	.dc.b	$FF,$3A,$F8,$97,$C6,$53,$F8,$A4,$EA,$BF,$CC,$ED,$7E,$84,$F0,$7D
	.dc.b	$01,$C3,$FE,$68,$8E,$18,$FE,$76,$72,$DA,$F0,$57,$6B,$AF,$68,$B3
	.dc.b	$FE,$75,$BE,$7B,$7F,$BB,$F0,$E6,$DA,$DB,$FA,$C9,$DD,$FF,$92,$AB
	.dc.b	$6D,$AD,$BF,$17,$0A,$FA,$E3,$38,$2F,$9C,$E4,$F8,$37,$D1,$C1,$F0
	.dc.b	$6F,$9C,$64,$F0,$D7,$38,$CD,$F6,$BE,$B5,$B8,$C4,$EA,$FB,$C7,$D6
	.dc.b	$8C,$47,$C7,$3E,$51,$A2,$19,$3C,$86,$2F,$E9,$37,$0F,$91,$3F,$4A
	.dc.b	$3F,$CE,$27,$C7,$3F,$98,$9E,$E3,$25,$3E,$91,$F8,$4F,$8B,$7C,$67
	.dc.b	$F8,$21,$B1,$DC,$E8,$7E,$ED,$CB,$63,$DB,$7F,$C9,$67,$2D,$8C,$DF
	.dc.b	$9A,$72,$D8,$CD,$7F,$A2,$A7,$2D,$8C,$A8,$CB,$63,$31,$83,$38,$FD
	.dc.b	$59,$F7,$4F,$B4,$D9,$5F,$71,$F8,$4B,$F4,$3C,$0C,$CF,$E2,$6B,$B5
	.dc.b	$72,$9D,$1F,$66,$BE,$BB,$F1,$4C,$B8,$25,$FC,$18,$CB,$F5,$1E,$4E
	.dc.b	$5E,$46,$0F,$A7,$1E,$67,$2F,$47,$31,$C6,$34,$66,$3D,$D1,$92,$3C
	.dc.b	$08,$7C,$D3,$CA,$62,$FE,$15,$7C,$98,$F2,$1D,$0F,$B0,$70,$30,$4B
	.dc.b	$F9,$2F,$CC,$C7,$96,$1E,$6C,$3E,$13,$E7,$AD,$D7,$6C,$3C,$2F,$57
	.dc.b	$EF,$3F,$0A,$BD,$6F,$82,$36,$E4,$F5,$5F,$82,$77,$3C,$AF,$C5,$5E
	.dc.b	$A3,$C1,$5B,$71,$1F,$2E,$7A,$BC,$23,$C1,$3B,$7E,$51,$C8,$F8,$97
	.dc.b	$F9,$68,$FE,$36,$FB,$A7,$A4,$71,$8D,$4F,$E5,$9D,$5F,$C6,$3D,$EE
	.dc.b	$23,$EC,$B5,$F0,$D7,$CB,$C3,$A1,$E0,$2C,$75,$6B,$F7,$27,$91,$E2
	.dc.b	$7E,$51,$D2,$BE,$4D,$F0,$BF,$85,$87,$C4,$9F,$2E,$1D,$EF,$E0,$B5
	.dc.b	$E7,$87,$5B,$F9,$4F,$A6,$F9,$4F,$D0,$BE,$EF,$23,$F0,$58,$78,$63
	.dc.b	$31,$D3,$0F,$A5,$5C,$A3,$D0,$73,$3E,$3C,$71,$69,$D4,$F8,$4E,$5F
	.dc.b	$CD,$36,$3A,$5F,$9E,$F4,$FA,$EF,$E0,$DF,$98,$E9,$87,$C6,$AD,$E1
	.dc.b	$F0,$E3,$E0,$B6,$B8,$D7,$88,$E2,$FE,$05,$F5,$1E,$C2,$61,$E4,$9F
	.dc.b	$83,$7D,$A7,$EF,$1C,$8F,$79,$F1,$CC,$94,$F0,$16,$FF,$14,$F1,$3E
	.dc.b	$19,$6B,$F5,$EF,$F5,$0C,$3F,$A3,$B6,$FF,$AA,$BF,$A4,$30,$7F,$85
	.dc.b	$BD,$EF,$F1,$D3,$76,$FA,$11,$88,$FF,$26,$37,$0F,$8F,$1B,$21,$E7
	.dc.b	$CB,$FB,$2C,$ED,$B5,$D0,$FF,$73,$C6,$67,$E5,$E1,$FD,$3A,$31,$06
	.dc.b	$C7,$EB,$5C,$98,$C3,$9D,$4A,$F1,$E1,$FC,$16,$B7,$39,$9F,$D5,$C6
	.dc.b	$1D,$E1,$FC,$6A,$31,$1F,$02,$7F,$86,$46,$2B,$FD,$28,$F8,$67,$88
	.dc.b	$3D,$13,$DE,$7C,$4F,$D2,$9F,$99,$3F,$E0,$66,$BF,$18,$FF,$05,$BF
	.dc.b	$0B,$17,$C6,$7F,$95,$4E,$98,$7F,$84,$BE,$71,$F0,$9F,$79,$B8,$7F
	.dc.b	$85,$1A,$FF,$B1,$9A,$FF,$85,$BE,$35,$E5,$FE,$19,$B1,$FE,$1C,$6B
	.dc.b	$FE,$CE,$6B,$FE,$1E,$F8,$27,$C6,$FE,$39,$B1,$FE,$24,$6B,$FE,$D6
	.dc.b	$6B,$FE,$26,$F5,$3F,$D0,$6F,$F8,$B1,$AF,$CB,$6A,$6F,$89,$F1,$23
	.dc.b	$FC,$62,$D7,$FD,$C6,$D7,$FC,$66,$D7,$FC,$73,$F1,$8F,$8D,$FD,$2B
	.dc.b	$63,$FC,$76,$D7,$FD,$DA,$D7,$FC,$7B,$F3,$0F,$A3,$FD,$6B,$63,$FC
	.dc.b	$86,$D7,$FD,$EA,$D7,$FC,$8B,$F1,$AF,$8E,$FD,$AB,$63,$FC,$96,$D7
	.dc.b	$FD,$FA,$D7,$FC,$9B,$F1,$CF,$8E,$FD,$EB,$63,$FC,$A6,$D7,$FC,$A6
	.dc.b	$1C,$3F,$E5,$6F,$9D,$AF,$F9,$71,$AF,$E6,$DA,$9B,$F1,$EF,$6F,$F8
	.dc.b	$6B,$67,$FC,$C6,$D7,$C2,$D7,$FC,$CB,$F1,$CF,$8F,$FE,$2A,$D8,$FF
	.dc.b	$35,$B5,$F4,$B5,$FF,$36,$F5,$38,$7F,$8E,$B6,$3F,$CE,$6D,$7E,$0B
	.dc.b	$5F,$D8,$71,$FE,$CC,$8F,$EC,$1F,$DB,$7C,$7C,$8F,$D9,$B5,$F9,$7E
	.dc.b	$5B,$E8,$3F,$96,$B6,$3F,$D0,$6D,$77,$1F,$CC,$BE,$35,$F1,$D4,$93
	.dc.b	$FE,$8A,$8C,$BF,$E8,$CF,$8D,$7C,$73,$E3,$B0,$D9,$FB,$B2,$C5,$F1
	.dc.b	$7A,$BF,$F4,$AF,$C5,$63,$C3,$FF,$76,$78,$EF,$8B,$7C,$62,$7F,$DB
	.dc.b	$35,$BC,$D7,$96,$DF,$0A,$DE,$39,$A9,$39,$AE,$35,$FE,$A8,$BE,$95
	.dc.b	$AC,$38,$4E,$70,$ED,$7F,$EA,$AB,$F1,$36,$32,$EE,$7F,$D6,$3F,$16
	.dc.b	$F8,$BE,$73,$E8,$F0,$DF,$19,$88,$56,$EF,$E1,$D6,$9F,$F7,$CC,$66
	.dc.b	$7D,$27,$3A,$E5,$19,$CF,$E0,$5F,$1A,$D8,$F8,$C7,$AF,$E4,$AF,$CC
	.dc.b	$76,$F8,$E7,$39,$7D,$4E,$AF,$D6,$3F,$D2,$EB,$E1,$CE,$B0,$EA,$D8
	.dc.b	$CC,$73,$C3,$DD,$1C,$A3,$C6,$BD,$DC,$E3,$A6,$E1,$DC,$FF,$6E,$CF
	.dc.b	$13,$E7,$D3,$E3,$3F,$AF,$7C,$4B,$E2,$B7,$8F,$CD,$AF,$EC,$55,$A9
	.dc.b	$CE,$16,$7F,$DC,$B3,$5F,$C6,$D7,$E3,$6F,$A4,$68,$38,$77,$5C,$5E
	.dc.b	$E6,$AF,$BB,$4E,$53,$C7,$0F,$88,$F3,$8E,$2E,$6F,$C5,$5F,$CA,$30
	.dc.b	$F3,$5F,$08,$C5,$F5,$3C,$65,$39,$BE,$2B,$ED,$1F,$24,$E1,$3C,$DF
	.dc.b	$71,$FC,$D8,$D1,$6B,$FC,$72,$E0,$FF,$07,$0F,$8B,$FE,$F3,$F1,$9C
	.dc.b	$5B,$1F,$D0,$1A,$F0,$38,$79,$E9,$E1,$F1,$7F,$BE,$9E,$0D,$78,$52
	.dc.b	$31,$1E,$28,$C6,$1C,$8F,$B9,$97,$E8,$26,$ED,$E8,$CF,$ED,$5F,$09
	.dc.b	$F9,$4D,$8F,$BD,$5C,$67,$E7,$36,$38,$E1,$E7,$6E,$1D,$AA,$1D,$27
	.dc.b	$55,$E7,$6C,$7F,$8D,$E3,$A1,$F3,$0A,$FF,$8E,$EF,$C9,$1C,$6B,$AD
	.dc.b	$F9,$B0,$F6,$4E,$6F,$CF,$1F,$AA,$D7,$A3,$63,$C0,$D7,$ED,$63,$E4
	.dc.b	$7F,$59,$19,$AF,$85,$87,$F2,$46,$DF,$7C,$3F,$49,$F1,$9F,$DB,$EB
	.dc.b	$F9,$4B,$67,$F0,$ED,$C3,$FC,$AC,$FD,$B7,$F8,$01,$8D,$BC,$91,$B8
	.dc.b	$34,$E0,$FF,$1E,$F8,$AF,$F2,$6D,$FF,$41,$F8,$B7,$D0,$FB,$E1,$18
	.dc.b	$AE,$E2,$B8,$E3,$F9,$C6,$C7,$F5,$09,$E0,$FF,$32,$AE,$77,$F2,$6B
	.dc.b	$72,$F8,$A7,$C4,$FF,$A8,$1B,$1F,$AD,$6B,$FE,$24,$6B,$FA,$57,$37
	.dc.b	$FD,$4B,$C8,$FE,$11,$97,$EC,$1B,$5F,$E7,$FA,$DC,$FE,$D1,$B1,$FE
	.dc.b	$83,$9D,$A6,$C6,$A3,$E1,$1F,$EB,$6C,$3F,$7E,$47,$07,$CB,$CB,$FD
	.dc.b	$48,$FF,$7A,$6B,$F4,$CF,$88,$6F,$DA,$8F,$D0,$9F,$E0,$58,$FF,$6E
	.dc.b	$C9,$F2,$5F,$EC,$66,$D7,$F1,$36,$BF,$E4,$A6,$BF,$C2,$DC,$CF,$E0
	.dc.b	$7C,$CF,$31,$F9,$D8,$CE,$9F,$C6,$1B,$5F,$76,$76,$FE,$12,$32,$BE
	.dc.b	$21,$EA,$7F,$B6,$9B,$1F,$C9,$DA,$FF,$97,$9A,$FF,$1F,$73,$87,$F8
	.dc.b	$57,$E8,$5F,$23,$F9,$56,$D7,$E5,$2B,$79,$7F,$2E,$6C,$7F,$88,$A7
	.dc.b	$6F,$E4,$DB,$7F,$E1,$89,$DC,$BE,$2B,$FC,$A0,$FF,$38,$6B,$FE,$C7
	.dc.b	$6B,$BE,$FF,$F7,$BC,$7F,$96,$6B,$D1,$F3,$7E,$73,$F8,$44,$67,$1E
	.dc.b	$95,$9B,$FF,$1E,$E2,$BD,$CF,$F0,$FF,$8A,$FF,$4D,$57,$F4,$A6,$BF
	.dc.b	$ED,$96,$A6,$9F,$EA,$3C,$7E,$57,$89,$F2,$4E,$9E,$E4,$AF,$0C,$65
	.dc.b	$FD,$FD,$71,$CB,$FA,$CF,$C1,$7F,$9E,$E6,$F0,$FF,$61,$51,$97,$FD
	.dc.b	$D8,$D7,$31,$F6,$7E,$72,$47,$F0,$C9,$FF,$44,$65,$FE,$4A,$6D,$FF
	.dc.b	$8A,$F5,$5F,$0B,$FF,$1B,$D6,$9F,$DE,$7C,$5F,$F3,$5E,$83,$4F,$E7
	.dc.b	$1D,$3F,$C4,$23,$33,$FC,$71,$AF,$F9,$09,$B9,$7F,$0B,$F8,$3E,$53
	.dc.b	$BB,$AF,$F9,$22,$FA,$46,$2B,$FC,$C7,$7C,$B3,$FE,$25,$86,$5F,$E2
	.dc.b	$15,$70,$FE,$21,$35,$FE,$49,$17,$0E,$B1,$E1,$21,$E1,$9E,$17,$DE
	.dc.b	$3E,$19,$FC,$E2,$78,$BE,$DA,$E9,$5C,$1B,$1D,$30,$F4,$C6,$23,$81
	.dc.b	$9C,$3D,$2F,$5A,$E4,$E4,$F5,$94,$F5,$96,$FC,$77,$DA,$31,$3F,$31
	.dc.b	$FE,$79,$7F,$D0,$C8,$FF,$43,$7E,$DF,$C2,$C8,$ED,$8F,$C1,$7F,$9C
	.dc.b	$46,$5F,$49,$2F,$FA,$34,$65,$F1,$46,$71,$FE,$91,$39,$7D,$F1,$93
	.dc.b	$FA,$34,$65,$F9,$E6,$4F,$09,$F4,$5F,$8A,$FE,$1D,$E0,$FC,$43,$86
	.dc.b	$28,$CD,$FA,$CF,$64,$66,$7A,$1F,$03,$0E,$71,$98,$F1,$4F,$0A,$E7
	.dc.b	$5F,$8B,$6C,$FD,$36,$BE,$26,$E1,$F1,$0F,$8F,$1E,$03,$AC,$F8,$27
	.dc.b	$DB,$3A,$9F,$71,$E5,$38,$39,$3A,$61,$F3,$1C,$3F,$42,$38,$1C,$89
	.dc.b	$1C,$1F,$9E,$F0,$7E,$99,$C1,$C3,$F4,$8F,$39,$E2,$8E,$CE,$0E,$AE
	.dc.b	$23,$B4,$7D,$57,$9E,$1F,$35,$CE,$1F,$36,$7A,$1E,$4A,$F6,$1C,$C8
	.dc.b	$E4,$87,$96,$7C,$F3,$F7,$F0,$F9,$93,$EB,$AE,$2E,$DE,$AF,$2B,$F2
	.dc.b	$46,$AB,$94,$F9,$0E,$07,$81,$F4,$3F,$00,$C5,$78,$1F,$89,$7F,$03
	.dc.b	$0F,$BB,$7E,$2C,$3B,$D7,$03,$B3,$87,$E4,$47,$99,$F0,$93,$0F,$2B
	.dc.b	$E1,$31,$87,$E5,$5F,$44,$6B,$0F,$14,$7C,$33,$EF,$9C,$63,$55,$E5
	.dc.b	$C3,$CF,$3A,$AF,$9B,$7D,$63,$83,$63,$31,$FA,$33,$C3,$19,$AF,$98
	.dc.b	$D8,$E9,$1E,$DB,$F9,$75,$CA,$7F,$54,$72,$3A,$1E,$02,$BB,$3F,$1F
	.dc.b	$87,$C9,$3E,$54,$FB,$27,$B9,$0F,$9A,$7E,$07,$0F,$DB,$DF,$CF,$AE
	.dc.b	$51,$F0,$A7,$E8,$C7,$4C,$3D,$51,$F4,$0F,$1D,$7B,$9A,$77,$BF,$13
	.dc.b	$CA,$FA,$5F,$68,$FA,$86,$4A,$E5,$FC,$5D,$7D,$57,$33,$E8,$7D,$47
	.dc.b	$C7,$3E,$49,$F1,$CB,$87,$C7,$9F,$74,$F1,$9F,$86,$78,$4E,$E7,$C7
	.dc.b	$AF,$80,$FB,$0F,$80,$74,$9F,$80,$D3,$E0,$61,$F1,$A7,$8C,$64,$E3
	.dc.b	$7F,$62,$FC,$58,$7E,$0B,$0E,$73,$9A,$F6,$35,$F6,$34,$F6,$3F,$08
	.dc.b	$CC,$7B,$9E,$2D,$72,$6E,$38,$BF,$0A,$FE,$84,$7F,$16,$C3,$F7,$8D
	.dc.b	$9F,$DB,$E1,$FE,$0F,$F6,$1E,$FF,$F8,$ED,$B2,$21,$1C,$B8,$34,$E1
	.dc.b	$C3,$AA,$63,$CD,$7B,$1F,$E5,$FA,$FF,$6D,$C6,$39,$7F,$DD,$D7,$C6
	.dc.b	$FC,$18,$7F,$B5,$70,$E1,$87,$FB,$35,$B8,$76,$8F,$1C,$FF,$BD,$FE
	.dc.b	$05,$F4,$1F,$E5,$76,$CF,$A4,$F7,$38,$70,$D7,$DC,$E6,$BB,$0F,$C2
	.dc.b	$5E,$97,$FB,$A5,$B5,$FE,$E8,$AE,$19,$7C,$A6,$C2,$F9,$5F,$13,$D1
	.dc.b	$87,$D0,$8F,$64,$BE,$01,$F1,$3F,$E7,$16,$C7,$AC,$D4,$64,$8F,$CE
	.dc.b	$FC,$21,$E6,$7F,$BF,$DB,$1F,$EF,$EA,$E6,$F9,$73,$FC,$0C,$63,$0F
	.dc.b	$83,$5C,$2B,$FD,$FC,$D8,$F3,$CE,$EB,$F1,$3F,$00,$F7,$7F,$D1,$0D
	.dc.b	$C3,$D0,$D7,$EA,$35,$FC,$73,$9B,$FA,$3F,$0F,$FE,$BE,$4F,$FA,$EA
	.dc.b	$35,$7E,$0B,$F9,$4F,$E4,$67,$81,$B8,$CD,$FE,$7F,$E7,$FF,$D0,$6F
	.dc.b	$E7,$5A,$FE,$89,$AE,$F9,$BF,$D4,$51,$FE,$71,$C5,$F7,$4F,$96,$EB
	.dc.b	$18,$7B,$DA,$BF,$CB,$07,$1A,$E1,$1F,$A4,$6B,$F4,$CF,$01,$FA,$BC
	.dc.b	$BF,$6B,$F5,$2F,$37,$FD,$68,$D9,$FC,$23,$5F,$D0,$35,$FD,$D3,$9C
	.dc.b	$7F,$3B,$EF,$DB,$FD,$C3,$6B,$F8,$25,$54,$E6,$93,$4D,$46,$FE,$0B
	.dc.b	$F0,$5F,$03,$F9,$76,$DF,$EC,$D7,$C1,$7E,$AA,$3F,$87,$46,$9F,$E2
	.dc.b	$1F,$32,$F3,$7F,$4B,$1F,$C5,$A3,$19,$FE,$8D,$B1,$FB,$F6,$BF,$C6
	.dc.b	$7E,$85,$F4,$3C,$A3,$F9,$04,$62,$BF,$58,$D8,$FE,$10,$D7,$F9,$17
	.dc.b	$C0,$7F,$BE,$67,$F6,$4D,$8F,$E1,$CD,$77,$E1,$77,$51,$BD,$FF,$7E
	.dc.b	$DA,$FE,$51,$F0,$3F,$C0,$1F,$C0,$57,$F3,$59,$D3,$7D,$CF,$E2,$B3
	.dc.b	$FC,$BF,$D0,$7B,$58,$AD,$4F,$F3,$09,$CB,$D2,$3C,$75,$B6,$C6,$6F
	.dc.b	$FA,$23,$67,$58,$F2,$AF,$CB,$1F,$C7,$7E,$07,$38,$E9,$B7,$DE,$B8
	.dc.b	$1F,$47,$E0,$7F,$A8,$7B,$AF,$7B,$F9,$16,$1E,$3C,$7F,$AA,$56,$A7
	.dc.b	$38,$59,$FE,$A9,$35,$FE,$5D,$5F,$D1,$6F,$A4,$68,$38,$77,$5C,$5E
	.dc.b	$E6,$AF,$BB,$4E,$53,$C7,$0F,$84,$F3,$8E,$2E,$6F,$C5,$5F,$D4,$70
	.dc.b	$F3,$5F,$08,$C5,$F5,$3C,$65,$39,$BE,$2B,$ED,$0A,$E7,$7D,$A3,$DC
	.dc.b	$7F,$45,$8F,$E8,$CF,$B0,$C1,$FD,$8F,$DF,$7C,$0B,$E0,$7F,$A6,$46
	.dc.b	$1B,$19,$9E,$0D,$7F,$C0,$93,$FD,$3B,$2E,$3E,$2F,$F5,$F9,$EC,$D7
	.dc.b	$6D,$7F,$C1,$F3,$C5,$F8,$59,$B6,$FF,$09,$E9,$F5,$E3,$13,$F7,$EB
	.dc.b	$84,$FD,$B6,$CF,$F8,$68,$DC,$F4,$AE,$71,$E0,$BF,$35,$F1,$6C,$71
	.dc.b	$9D,$D7,$F8,$96,$39,$D6,$F0,$F1,$5F,$0A,$D5,$F0,$7F,$38,$DC,$3C
	.dc.b	$B1,$FE,$2F,$6B,$A3,$F3,$58,$76,$3C,$D7,$8A,$FE,$97,$1E,$C8,$FF
	.dc.b	$03,$B7,$5D,$D7,$BC,$FD,$F4,$65,$F2,$9F,$C0,$6F,$C3,$F0,$3F,$C2
	.dc.b	$23,$F8,$23,$5F,$DE,$B6,$FF,$C2,$AF,$F8,$06,$FF,$C8,$EE,$5B,$B7
	.dc.b	$F6,$08,$C5,$7F,$93,$5E,$0F,$F9,$3F,$DF,$7A,$7F,$96,$6C,$7F,$87
	.dc.b	$5A,$FF,$8E,$5A,$FF,$94,$DC,$E1,$FE,$36,$EF,$7C,$0F,$F7,$26,$D7
	.dc.b	$F1,$FA,$DE,$5F,$E5,$EF,$51,$EF,$FF,$76,$6D,$7F,$25,$6B,$FC,$7B
	.dc.b	$D5,$7D,$1E,$EB,$F5,$B5,$AA,$FF,$33,$35,$ED,$1A,$3F,$C0,$75,$FC
	.dc.b	$26,$FF,$CD,$ED,$9F,$F3,$95,$78,$33,$FF,$21,$E3,$FE,$71,$6C,$FF
	.dc.b	$83,$23,$17,$FE,$6F,$C9,$F3,$99,$AF,$98,$43,$E6,$14,$FF,$07,$5F
	.dc.b	$F8,$2F,$E0,$FE,$41,$C6,$3F,$A3,$66,$F8,$3F,$F0,$DB,$63,$31,$F3
	.dc.b	$FC,$4C,$BF,$C4,$27,$53,$FE,$71,$6B,$FE,$8B,$6B,$FE,$90,$77,$7F
	.dc.b	$6F,$DD,$FB,$47,$49,$FE,$B1,$F0,$4F,$83,$FF,$16,$B6,$33,$1F,$E3
	.dc.b	$2F,$7D,$F2,$9A,$8F,$E4,$75,$BA,$FF,$42,$35,$FF,$4F,$35,$FF,$51
	.dc.b	$BF,$E5,$1A,$FF,$09,$47,$FA,$69,$AF,$F6,$28,$E9,$87,$F9,$EB,$E1
	.dc.b	$1E,$1F,$F9,$09,$B5,$FD,$C5,$AF,$D2,$B7,$C3,$6E,$BF,$9C,$D6,$AB
	.dc.b	$FD,$64,$D7,$FD,$6E,$D7,$FD,$74,$F4,$BF,$F4,$CD,$FF,$AA,$38,$BF
	.dc.b	$D6,$8D,$AD,$DF,$F9,$5B,$E0,$DF,$0B,$49,$3E,$FF,$98,$AF,$FA,$AE
	.dc.b	$76,$FF,$A4,$DA,$E6,$7F,$A8,$C6,$31,$F1,$FC,$37,$D5,$39,$BF,$EA
	.dc.b	$0F,$85,$70,$FF,$9A,$1B,$87,$F8,$5D,$AF,$C7,$F8,$FD,$4F,$F5,$E9
	.dc.b	$CC,$FF,$B5,$DA,$FF,$B7,$1A,$FF,$B7,$DE,$75,$F1,$BE,$1F,$F1,$07
	.dc.b	$49,$FE,$05,$EC,$FF,$05,$23,$FE,$29,$ED,$FE,$16,$FE,$E2,$32,$FF
	.dc.b	$89,$3B,$51,$FC,$03,$C0,$CE,$1F,$EF,$5A,$FE,$19,$18,$E2,$FE,$29
	.dc.b	$56,$7F,$DE,$98,$76,$58,$74,$7C,$A4,$39,$97,$0E,$C7,$FB,$32,$79
	.dc.b	$4F,$19,$F3,$57,$75,$3A,$8C,$4E,$EB,$96,$18,$9E,$15,$BB,$F0,$61
	.dc.b	$EA,$7E,$35,$6D,$C9,$EC,$70,$F8,$8D,$9A,$3D,$F5,$DA,$37,$7F,$EE
	.dc.b	$6C,$3C,$97,$FC,$AA,$72,$F8,$6F,$E1,$9D,$49,$5C,$6F,$EA,$47,$7B
	.dc.b	$E6,$FD,$B3,$C6,$70,$AE,$95,$C6,$32,$FC,$E8,$C9,$E3,$8E,$95,$E9
	.dc.b	$69,$ED,$C3,$E4,$3D,$4F,$01,$E8,$AE,$8E,$1F,$A2,$F2,$38,$1F,$63
	.dc.b	$0F,$82,$EE,$FC,$4E,$B0,$F6,$BF,$30,$F4,$C7,$A1,$E0,$43,$A1,$DE
	.dc.b	$BE,$03,$E5,$8E,$0F,$33,$C7,$7E,$C9,$F4,$C7,$23,$C5,$87,$C3,$69
	.dc.b	$F0,$F0,$F4,$BC,$E7,$DE,$E8,$EB,$1A,$C3,$D7,$7D,$F0,$F1,$C6,$27
	.dc.b	$58,$7D,$EB,$E5,$87,$A8,$EA,$79,$AB,$84,$7B,$EF,$DB,$87,$58,$F8
	.dc.b	$93,$A7,$9C,$EB,$0E,$F1,$F3,$0D,$E1,$EF,$C3,$B3,$E3,$6C,$6C,$F8
	.dc.b	$15,$EA,$9D,$36,$3E,$31,$E5,$3E,$B6,$1E,$57,$89,$8B,$F1,$BF,$5C
	.dc.b	$F1,$9C,$8F,$D2,$1F,$E1,$8A,$FF,$0D,$91,$10,$FD,$61,$B3,$06,$08
	.dc.b	$79,$3E,$1F,$FC,$3E,$7D,$CA,$38,$8F,$51,$FC,$41,$C3,$F3,$0F,$E4
	.dc.b	$2E,$30,$FC,$5C,$F2,$9F,$E6,$53,$D4,$D6,$1F,$1E,$3F,$C7,$11,$87
	.dc.b	$E0,$98,$38,$9E,$29,$E8,$E4,$F1,$1E,$53,$66,$1D,$D7,$F2,$D9,$F1
	.dc.b	$4E,$A3,$CF,$1E,$18,$FA,$A4,$9E,$35,$FE,$42,$9E,$33,$E1,$AE,$2F
	.dc.b	$39,$F6,$47,$78,$C1,$EC,$70,$F1,$8D,$9D,$5D,$9E,$2C,$7F,$53,$5E
	.dc.b	$57,$F8,$29,$88,$DB,$8C,$3D,$58,$7A,$27,$C1,$7D,$5F,$01,$83,$EC
	.dc.b	$9C,$A7,$FA,$3D,$F7,$6C,$64,$FE,$49,$19,$69,$93,$F8,$6B,$96,$99
	.dc.b	$C3,$F7,$95,$E3,$7E,$19,$F2,$A3,$CA,$7C,$1A,$F5,$D7,$D0,$9D,$D7
	.dc.b	$C8,$AF,$49,$C3,$0F,$84,$FC,$AC,$3E,$13,$CC,$C5,$F3,$AF,$11,$CE
	.dc.b	$3F,$14,$73,$C3,$E0,$B4,$F7,$94,$F6,$97,$0F,$6B,$E4,$7C,$EF,$CF
	.dc.b	$3A,$B8,$7C,$87,$C0,$BF,$3E,$1E,$A8,$C6,$1E,$77,$A6,$1E,$7A,$F6
	.dc.b	$9C,$30,$F3,$E1,$D7,$0E,$4F,$79,$D5,$FA,$A3,$55,$C2,$FE,$1C,$FF
	.dc.b	$4D,$F8,$77,$C3,$7F,$0D,$B5,$7F,$0F,$9C,$D7,$F1,$29,$CB,$FC,$52
	.dc.b	$32,$FF,$A6,$BF,$0C,$FF,$50,$BB,$AE,$57,$FC,$5A,$29,$45,$72,$7F
	.dc.b	$8D,$B5,$D9,$FE,$A6,$F8,$8F,$F3,$46,$07,$FC,$D5,$1C,$AB,$D5,$7F
	.dc.b	$C9,$FE,$1D,$F0,$D8,$AF,$E5,$71,$87,$CC,$62,$35,$83,$E2,$3A,$D6
	.dc.b	$EB,$D4,$D7,$F9,$8B,$5F,$05,$7B,$54,$71,$AF,$92,$75,$F8,$8E,$F6
	.dc.b	$7C,$D5,$C2,$EF,$C3,$3E,$22,$92,$7D,$38,$3F,$0A,$E5,$1E,$C1,$5C
	.dc.b	$5F,$A0,$D7,$EA,$35,$FE,$96,$D7,$FA,$83,$B9,$F4,$FC,$4B,$D3,$F8
	.dc.b	$ED,$AF,$05,$6F,$5F,$EB,$1F,$12,$F8,$8F,$6B,$6B,$CC,$D7,$FD,$C7
	.dc.b	$F2,$BC,$D6,$5F,$90,$53,$B7,$FB,$0B,$5E,$F5,$F4,$63,$11,$FD,$A9
	.dc.b	$ED,$3F,$4B,$E2,$BF,$D8,$F8,$7E,$75,$B5,$BB,$FA,$7F,$13,$FE,$38
	.dc.b	$27,$3F,$E1,$BE,$25,$F4,$5A,$8F,$C0,$4F,$F9,$F6,$FF,$BA,$B5,$FE
	.dc.b	$F4,$D7,$FB,$EB,$FE,$60,$C9,$6F,$8B,$EC,$75,$AF,$B9,$F1,$4F,$53
	.dc.b	$F0,$ED,$8C,$C7,$E6,$BE,$2D,$F5,$DA,$8F,$CA,$56,$EB,$F6,$8D,$7F
	.dc.b	$C1,$6D,$7F,$C2,$0E,$F1,$FC,$CD,$DD,$FF,$82,$36,$BF,$C2,$F5,$57
	.dc.b	$C8,$3E,$47,$F4,$8D,$8F,$E0,$4D,$7F,$83,$35,$FF,$0E,$FC,$6F,$E2
	.dc.b	$7E,$3B,$F8,$43,$63,$F8,$62,$F6,$3F,$43,$1C,$38,$D5,$FC,$3F,$E3
	.dc.b	$5F,$1D,$F8,$BA,$FF,$16,$46,$35,$D4,$E6,$7F,$C6,$33,$97,$F8,$D7
	.dc.b	$C5,$3E,$33,$F7,$77,$FC,$97,$E3,$1E,$A6,$2F,$FC,$73,$7B,$AF,$DB
	.dc.b	$D6,$9F,$DB,$46,$67,$AD,$F8,$54,$65,$FF,$1F,$B5,$FF,$22,$2F,$87
	.dc.b	$F0,$71,$BD,$67,$C2,$F8,$EF,$9D,$6A,$73,$8F,$8A,$35,$87,$38,$F2
	.dc.b	$D6,$27,$96,$1C,$23,$F6,$B3,$FD,$27,$E3,$33,$8E,$9B,$1F,$C4,$EB
	.dc.b	$81,$F3,$7E,$33,$FC,$BB,$F1,$4F,$8A,$FE,$A5,$87,$9F,$1F,$F3,$0D
	.dc.b	$6A,$73,$85,$9F,$F3,$0C,$D7,$F8,$3D,$7F,$56,$BE,$91,$A0,$E1,$DD
	.dc.b	$71,$7B,$9A,$BE,$ED,$39,$4F,$1C,$3E,$13,$CE,$38,$B9,$BF,$15,$7F
	.dc.b	$61,$C3,$CD,$7C,$23,$17,$D4,$F1,$94,$E6,$F8,$AF,$B4,$2B,$9D,$F6
	.dc.b	$8F,$71,$FD,$B6,$3F,$A9,$3B,$3F,$AA,$3F,$DC,$7E,$29,$F1,$6F,$8C
	.dc.b	$C3,$6B,$FB,$C4,$EA,$72,$71,$F2,$7F,$9A,$CE,$9A,$FF,$A5,$27,$FA
	.dc.b	$36,$7B,$24,$7B,$F3,$FC,$C3,$F6,$5A,$FC,$39,$E2,$DA,$EF,$9F,$D9
	.dc.b	$AD,$B7,$6F,$F5,$33,$F7,$6F,$8B,$6B,$8D,$7F,$AA,$CE,$25,$AE,$2F
	.dc.b	$FA,$BE,$B8,$B6,$78,$DF,$FA,$D2,$38,$B6,$B8,$BF,$EB,$8A,$E2,$DA
	.dc.b	$E2,$FE,$7E,$76,$D3,$F3,$53,$C1,$C1,$FA,$E3,$B1,$FD,$F6,$BE,$3E
	.dc.b	$1F,$D9,$67,$84,$63,$5F,$65,$6E,$FF,$C1,$B5,$D7,$E3,$3F,$82,$CF
	.dc.b	$F8,$1D,$AF,$70,$E3,$0F,$EE,$4F,$F7,$56,$ED,$FE,$18,$8D,$61,$FE
	.dc.b	$D7,$6D,$7F,$B6,$9E,$11,$FE,$DB,$F8,$A7,$B3,$FE,$20,$6D,$7F,$86
	.dc.b	$5A,$FC,$1C,$3F,$B0,$77,$6E,$BF,$C5,$D5,$AA,$FF,$71,$B5,$ED,$1F
	.dc.b	$E3,$3F,$61,$F0,$FF,$E2,$E6,$CF,$F8,$9D,$AF,$C1,$F6,$5F,$17,$BA
	.dc.b	$93,$B7,$FC,$6F,$3A,$8C,$1F,$E5,$DA,$E6,$FF,$BD,$9A,$FF,$BE,$DA
	.dc.b	$FF,$BF,$B0,$FF,$7D,$E5,$FE,$1F,$6E,$7F,$1E,$7F,$CA,$BE,$2F,$F9
	.dc.b	$B9,$FA,$D3,$A7,$91,$F2,$CD,$FC,$5D,$7F,$CE,$11,$FC,$10,$8F,$F9
	.dc.b	$5F,$27,$C5,$7F,$94,$9B,$1F,$E4,$B6,$BF,$4E,$FF,$B8,$39,$9E,$04
	.dc.b	$7F,$CA,$B1,$A7,$F1,$93,$C9,$10,$FA,$B3,$E3,$23,$FE,$87,$F8,$8F
	.dc.b	$F1,$AA,$69,$FE,$38,$C9,$F0,$CF,$8A,$FF,$35,$CE,$BE,$2A,$9F,$E7
	.dc.b	$E9,$D5,$FF,$99,$E5,$3F,$E7,$96,$C7,$86,$30,$7F,$9F,$7E,$27,$FB
	.dc.b	$74,$7F,$9B,$DA,$FE,$A5,$B8,$7F,$A0,$2D,$F1,$5F,$E7,$E9,$CE,$FF
	.dc.b	$05,$B7,$F1,$1B,$7E,$CA,$E1,$E0,$F8,$49,$5F,$EA,$BA,$F0,$BF,$2C
	.dc.b	$7C,$4B,$E2,$BF,$D1,$AD,$8F,$F4,$33,$5F,$D7,$65,$FE,$1A,$73,$3C
	.dc.b	$08,$FF,$A4,$23,$4F,$F0,$79,$E4,$F8,$5A,$FD,$D6,$BF,$80,$6A,$F8
	.dc.b	$77,$C4,$7F,$A8,$9B,$1F,$E9,$B6,$BE,$6F,$89,$D4,$C8,$D3,$F7,$63
	.dc.b	$11,$FE,$D1,$9E,$2F,$C5,$6B,$F9,$16,$BF,$EF,$BE,$67,$AF,$FB,$EE
	.dc.b	$AF,$E1,$0C,$9F,$E1,$0E,$5A,$43,$F9,$54,$F2,$7F,$D6,$9D,$3F,$C7
	.dc.b	$38,$3F,$D3,$93,$FE,$FA,$E1,$FE,$09,$7C,$EF,$8B,$DD,$A7,$2C,$31
	.dc.b	$7C,$A6,$D7,$EA,$F0,$EC,$F4,$BF,$25,$6D,$E4,$62,$34,$F6,$3F,$91
	.dc.b	$5F,$47,$F6,$F8,$79,$DF,$DC,$DF,$67,$E4,$61,$E9,$C3,$B4,$73,$69
	.dc.b	$FE,$BB,$9E,$10,$AE,$8D,$AE,$58,$79,$2F,$8B,$4F,$76,$1E,$5A,$E0
	.dc.b	$7A,$CF,$2B,$9B,$EB,$87,$81,$EE,$7D,$59,$F7,$9C,$4E,$67,$A6,$BB
	.dc.b	$B9,$8E,$75,$D7,$0F,$15,$F5,$71,$5F,$38,$E4,$7C,$A7,$73,$A8,$C8
	.dc.b	$9E,$4F,$79,$F8,$6F,$CE,$3A,$61,$F1,$1E,$87,$A6,$78,$BD,$2F,$DD
	.dc.b	$3F,$46,$FC,$0F,$19,$F0,$46,$30,$F9,$D1,$F2,$8F,$11,$D6,$34,$F8
	.dc.b	$0F,$92,$77,$C3,$DA,$F0,$3D,$B8,$79,$2F,$C1,$87,$93,$0F,$9B,$5C
	.dc.b	$4E,$C7,$19,$F3,$57,$19,$F2,$CF,$D6,$9F,$63,$B8,$CC,$EE,$3C,$D3
	.dc.b	$CB,$0F,$4C,$71,$AF,$0C,$F9,$30,$F4,$CF,$8C,$F4,$D7,$0C,$3D,$11
	.dc.b	$F1,$4D,$57,$57,$C4,$78,$2F,$A9,$E8,$9F,$C4,$C6,$8F,$11,$CE,$7C
	.dc.b	$4F,$C2,$3A,$CF,$B2,$BC,$57,$F8,$49,$E2,$4B,$F2,$BF,$70,$EC,$E3
	.dc.b	$0F,$89,$1B,$9E,$53,$FA,$46,$9F,$1A,$77,$1E,$89,$F1,$9E,$D3,$D1
	.dc.b	$7E,$DA,$F9,$38,$7A,$A7,$E4,$9E,$28,$EA,$FC,$1B,$EB,$1C,$2B,$94
	.dc.b	$FC,$1A,$E4,$7C,$D7,$C2,$7C,$0A,$E2,$EE,$78,$BE,$03,$77,$F0,$70
	.dc.b	$F9,$31,$FA,$A9,$DD,$F5,$3B,$BF,$3E,$3D,$06,$CF,$D9,$1F,$A0,$8D
	.dc.b	$BF,$0C,$FA,$07,$83,$0F,$7B,$E8,$C3,$CE,$F9,$67,$8B,$EB,$BF,$2B
	.dc.b	$D6,$FC,$A7,$8D,$E8,$78,$CF,$B4,$7C,$13,$CB,$3F,$A4,$76,$73,$33
	.dc.b	$1F,$94,$69,$F6,$8A,$7A,$A3,$CB,$87,$C9,$8F,$86,$73,$3C,$27,$AE
	.dc.b	$7C,$EF,$C3,$BF,$E2,$6F,$F9,$8F,$0F,$F3,$A1,$1F,$F3,$43,$FE,$6C
	.dc.b	$F8,$89,$1F,$85,$6B,$F2,$5B,$87,$C8,$A4,$62,$FC,$35,$FC,$AE,$30
	.dc.b	$FD,$5C,$3F,$45,$18,$8F,$4D,$FF,$55,$8C,$4F,$F4,$86,$E1,$FC,$F9
	.dc.b	$B5,$FC,$BD,$AF,$F1,$E6,$D7,$F1,$86,$DF,$F1,$46,$CF,$C5,$6B,$F9
	.dc.b	$B6,$FC,$3B,$E1,$BF,$CC,$D6,$BF,$F3,$44,$E6,$FF,$CD,$73,$97,$FC
	.dc.b	$DB,$19,$7F,$D5,$5F,$11,$FE,$89,$9D,$42,$75,$18,$6D,$76,$BF,$0B
	.dc.b	$DA,$F9,$39,$BF,$F2,$81,$E0,$9A,$D2,$D7,$8D,$FF,$2B,$7C,$3B,$E1
	.dc.b	$7A,$C7,$FA,$12,$30,$FB,$2F,$FD,$0D,$18,$8F,$F6,$27,$C3,$BC,$FE
	.dc.b	$30,$9D,$67,$EE,$6D,$F2,$C3,$BC,$B0,$E5,$7D,$CF,$14,$FA,$9E,$8D
	.dc.b	$BF,$14,$7A,$A3,$5F,$06,$F8,$6F,$1B,$67,$A0,$70,$FF,$B7,$7E,$19
	.dc.b	$F0,$FC,$2A,$4F,$0A,$DE,$5F,$74,$5F,$09,$D5,$78,$88,$FD,$2A,$EE
	.dc.b	$FF,$A1,$7E,$15,$ED,$7C,$86,$C7,$99,$AF,$FB,$C7,$E1,$BF,$DF,$2F
	.dc.b	$FA,$E2,$33,$8F,$E2,$9B,$B7,$FA,$ED,$B2,$95,$72,$84,$F6,$7C,$97
	.dc.b	$F3,$E3,$11,$CE,$3E,$97,$C2,$7F,$5C,$27,$2F,$DA,$F8,$47,$C2,$FE
	.dc.b	$0A,$73,$7F,$41,$FE,$EF,$92,$B7,$39,$BE,$A4,$7F,$33,$5F,$E4,$57
	.dc.b	$FD,$59,$F0,$4F,$2B,$F1,$ED,$8F,$BA,$D7,$E0,$6B,$FB,$4F,$83,$7C
	.dc.b	$27,$0B,$95,$C1,$FC,$9C,$EA,$30,$7E,$AA,$FA,$3E,$29,$FC,$A4,$62
	.dc.b	$3F,$DD,$5F,$04,$F8,$8F,$CC,$36,$3F,$66,$D7,$F6,$ED,$7F,$D9,$1F
	.dc.b	$0A,$F4,$FF,$40,$D8,$FD,$DB,$5F,$DF,$B5,$FF,$68,$7C,$2F,$E3,$3E
	.dc.b	$17,$F8,$0B,$63,$F8,$2A,$BF,$CC,$47,$0E,$35,$7F,$08,$F8,$1F,$DE
	.dc.b	$6F,$FB,$3E,$37,$C1,$3E,$0F,$F4,$52,$C6,$5F,$07,$E8,$9F,$91,$AE
	.dc.b	$13,$FB,$29,$D5,$FE,$AE,$73,$5C,$04,$EA,$BF,$96,$CE,$AB,$F9,$4C
	.dc.b	$BD,$CF,$1D,$FE,$7F,$D0,$7B,$78,$8E,$73,$AB,$56,$B1,$F2,$4E,$F0
	.dc.b	$E9,$1E,$7B,$C4,$F3,$BE,$11,$E7,$5F,$07,$9C,$74,$DB,$FE,$23,$3C
	.dc.b	$0F,$E8,$35,$F2,$7E,$0F,$E5,$7C,$0B,$E0,$3F,$A7,$E3,$E9,$AD,$09
	.dc.b	$D3,$FD,$42,$3F,$A8,$46,$23,$EA,$0A,$EB,$5D,$30,$B3,$A3,$47,$58
	.dc.b	$D3,$4D,$4F,$07,$F8,$65,$F8,$27,$FA,$8C,$E7,$0F,$8B,$87,$8E,$FD
	.dc.b	$2F,$96,$33,$5C,$27,$B1,$D3,$0E,$CF,$08,$F1,$D7,$38,$F3,$BC,$F0
	.dc.b	$F1,$47,$98,$E7,$7D,$9C,$1F,$D9,$30,$FE,$98,$FC,$03,$07,$F6,$AF
	.dc.b	$80,$7C,$1B,$E0,$FF,$B5,$C6,$1B,$19,$9E,$0D,$7E,$64,$BE,$0F,$F9
	.dc.b	$94,$75,$6B,$A6,$BF,$62,$38,$61,$E0,$9E,$39,$7D,$DC,$FF,$41,$7F
	.dc.b	$4D,$AF,$CA,$BE,$2D,$AE,$33,$F8,$43,$75,$CF,$6F,$C3,$BF,$89,$BE
	.dc.b	$2D,$AE,$37,$FB,$FB,$E2,$73,$BE,$2F,$A6,$77,$2B,$E7,$3B,$8F,$45
	.dc.b	$70,$7F,$85,$4E,$9B,$5C,$AB,$BC,$78,$6B,$83,$D6,$75,$1F,$E1,$29
	.dc.b	$F8,$78,$7F,$8B,$9B,$AF,$96,$3C,$37,$FE,$04,$9C,$BD,$4F,$F0,$3E
	.dc.b	$1E,$35,$F0,$7F,$C3,$1F,$F0,$5B,$5F,$E3,$0D,$AF,$F1,$EB,$FD,$BC
	.dc.b	$C5,$FE,$D1,$B3,$FC,$A5,$AF,$6C,$7F,$C5,$7F,$48,$F8,$6F,$F1,$3F
	.dc.b	$1B,$E4,$E9,$A8,$97,$FC,$DE,$74,$D9,$CB,$FC,$B7,$E1,$1E,$BF,$F8
	.dc.b	$B5,$B1,$FE,$48,$6B,$FE,$4C,$6B,$FC,$1D,$CE,$9F,$E2,$3F,$05,$F1
	.dc.b	$DF,$E3,$C6,$D7,$F8,$C2,$B7,$B7,$F1,$46,$C7,$F8,$56,$76,$FF,$48
	.dc.b	$BF,$EF,$D7,$FE,$4C,$AD,$C9,$FC,$FB,$E3,$E1,$F5,$11,$F5,$3E,$3F
	.dc.b	$A5,$91,$E3,$8E,$86,$1F,$06,$09,$CC,$7F,$92,$3E,$19,$F2,$3F,$E5
	.dc.b	$36,$C7,$F9,$B9,$AF,$F9,$C9,$AF,$F2,$D7,$39,$7C,$7F,$A2,$F8,$C7
	.dc.b	$F9,$A2,$33,$A7,$F3,$86,$D7,$C5,$9D,$BE,$3B,$FF,$37,$FC,$33,$EE
	.dc.b	$BF,$CC,$CD,$9F,$F4,$0B,$5F,$F4,$2B,$5F,$E8,$0E,$71,$FF,$2E,$7C
	.dc.b	$1B,$E8,$3F,$CE,$AD,$AF,$F3,$4D,$6F,$3F,$E9,$AD,$8F,$F2,$74,$ED
	.dc.b	$FB,$4D,$BF,$F4,$FC,$EE,$5F,$0C,$FA,$6F,$F3,$F3,$63,$FD,$32,$D7
	.dc.b	$FD,$3A,$D7,$FA,$C3,$9A,$F4,$F2,$FA,$0F,$F4,$9C,$67,$1F,$EC,$6D
	.dc.b	$AF,$3C,$ED,$FF,$42,$C6,$57,$C3,$3E,$73,$FD,$24,$D8,$FF,$56,$B5
	.dc.b	$FF,$58,$B5,$FE,$D8,$E7,$0F,$9B,$EC,$7C,$C3,$FD,$4F,$19,$C7,$FB
	.dc.b	$BB,$6B,$E5,$CE,$DF,$F4,$CC,$65,$7C,$2F,$F8,$ED,$FF,$5F,$B5,$FF
	.dc.b	$61,$B5,$FE,$FC,$D7,$FC,$0A,$EE,$FF,$03,$D1,$F8,$03,$FD,$77,$19
	.dc.b	$C3,$7E,$5E,$6F,$FC,$89,$BF,$ED,$2B,$51,$FE,$43,$F8,$47,$2F,$FB
	.dc.b	$11,$B1,$FE,$B2,$6B,$FE,$3F,$F8,$57,$C7,$6E,$A4,$ED,$FF,$66,$CE
	.dc.b	$A3,$04,$3F,$C9,$15,$D3,$4F,$F7,$6C,$EE,$73,$5D,$1A,$7F,$B1,$23
	.dc.b	$53,$FE,$F4,$AE,$59,$FF,$8A,$9B,$5E,$DB,$F1,$E1,$CE,$F8,$45,$BF
	.dc.b	$F1,$8E,$1D,$1E,$51,$9B,$E4,$78,$6F,$0B,$0E,$ED,$9E,$EF,$BF,$0E
	.dc.b	$ED,$BE,$F7,$CA,$F8,$9C,$B0,$F1,$57,$80,$F8,$24,$BF,$8F,$5C,$70
	.dc.b	$F1,$D7,$0A,$F1,$1D,$CE,$55,$C1,$C1,$A3,$A6,$1E,$47,$38,$79,$2F
	.dc.b	$D7,$1E,$23,$FC,$A5,$1F,$84,$6C,$7F,$C0,$18,$C4,$7C,$A3,$EC,$78
	.dc.b	$7F,$84,$3D,$0E,$08,$78,$8F,$3D,$7D,$48,$E1,$1E,$4A,$AF,$E4,$E3
	.dc.b	$9B,$4E,$77,$F4,$1E,$83,$0F,$F2,$E5,$F5,$8E,$87,$07,$0F,$0C,$3E
	.dc.b	$A5,$F8,$2F,$A1,$2B,$6F,$E6,$A3,$C6,$DA,$E8,$7B,$08,$78,$E3,$A4
	.dc.b	$FC,$78,$F3,$DF,$99,$FF,$5C,$35,$FB,$15,$E1,$73,$5C,$9E,$B7,$E8
	.dc.b	$7C,$27,$0B,$ED,$5E,$58,$F9,$07,$BF,$0E,$4F,$94,$F8,$A6,$A3,$81
	.dc.b	$E1,$7A,$1E,$23,$BD,$7E,$25,$F0,$E1,$E9,$8C,$47,$3B,$F9,$27,$90
	.dc.b	$8F,$33,$47,$B2,$FC,$B5,$CD,$FD,$0E,$1E,$8B,$EB,$87,$9B,$0E,$B8
	.dc.b	$71,$3C,$75,$F5,$A3,$31,$EF,$3E,$11,$A3,$E0,$CF,$37,$C8,$FC,$1A
	.dc.b	$F9,$D1,$E6,$3E,$19,$30,$FB,$07,$9D,$FD,$8B,$8B,$F1,$47,$4B,$F1
	.dc.b	$D7,$A9,$F2,$D7,$19,$FC,$2D,$7B,$F0,$F8,$31,$F4,$1A,$F8,$EF,$93
	.dc.b	$D4,$87,$99,$F4,$E1,$ED,$8C,$4F,$CE,$AF,$2C,$7C,$59,$D5,$74,$8C
	.dc.b	$61,$E8,$BF,$0D,$69,$B1,$99,$F8,$37,$E9,$69,$E9,$29,$E9,$9E,$2F
	.dc.b	$A2,$FC,$18,$7C,$98,$FA,$B7,$F1,$AF,$CF,$39,$9F,$B6,$7D,$57,$E5
	.dc.b	$61,$F9,$97,$4F,$DB,$C3,$E0,$4F,$43,$E3,$CF,$0B,$FC,$AC,$7C,$03
	.dc.b	$C9,$3E,$67,$07,$CA,$3F,$14,$78,$AB,$E7,$57,$38,$FE,$06,$F8,$AB
	.dc.b	$AE,$1F,$51,$F9,$13,$F8,$03,$E3,$14,$F8,$C7,$C8,$AF,$A0,$D8,$E3
	.dc.b	$1E,$FC,$3E,$84,$F1,$8D,$B6,$76,$7B,$AF,$C9,$1F,$A9,$3E,$55,$7C
	.dc.b	$E3,$9D,$75,$8F,$2B,$5F,$29,$D4,$F8,$E7,$94,$A6,$4F,$93,$7F,$7A
	.dc.b	$7A,$DF,$8D,$B3,$F2,$1A,$FE,$BD,$B3,$F2,$63,$AD,$78,$CF,$79,$6B
	.dc.b	$DF,$1C,$CF,$4B,$DD,$A6,$4F,$5C,$71,$7A,$C6,$DC,$B6,$3B,$B4,$EF
	.dc.b	$1E,$73,$A8,$F9,$17,$C5,$22,$10,$D3,$5C,$C7,$07,$06,$8A,$E5,$FF
	.dc.b	$5C,$7A,$8E,$5E,$B8,$30,$F0,$29,$D5,$62,$7B,$E1,$C1,$B1,$91,$5C
	.dc.b	$6F,$51,$8B,$E6,$EA,$B8,$B4,$F8,$13,$A7,$E0,$35,$F8,$6D,$76,$FC
	.dc.b	$07,$D8,$F8,$29,$F0,$6E,$FC,$36,$BD,$07,$02,$D7,$9A,$BE,$24,$7B
	.dc.b	$27,$83,$FE,$E1,$6C,$6A,$39,$1E,$B1,$F3,$8F,$A5,$F3,$36,$3E,$BB
	.dc.b	$5F,$B8,$D7,$F0,$0E,$67,$C6,$3E,$51,$E6,$7B,$DB,$5E,$CA,$E1,$9F
	.dc.b	$E2,$9B,$1F,$5A,$76,$FF,$BE,$DB,$7E,$B9,$DC,$BE,$71,$E5,$7C,$F6
	.dc.b	$C7,$E4,$1A,$FE,$55,$AF,$E6,$5C,$D7,$AB,$81,$F3,$9F,$5D,$B5,$F5
	.dc.b	$2B,$79,$7E,$89,$B1,$F8,$E9,$DB,$F8,$9B,$FD,$55,$FE,$42,$3D,$18
	.dc.b	$5F,$7D,$F4,$3F,$87,$6C,$FE,$A1,$AF,$EB,$5A,$FE,$C5,$CD,$FD,$2F
	.dc.b	$81,$79,$9F,$8F,$6D,$7E,$32,$B7,$A7,$EE,$9B,$1F,$A6,$9D,$BE,$96
	.dc.b	$DF,$E2,$E7,$72,$FA,$E7,$D9,$7E,$7D,$B1,$FC,$01,$AF,$F0,$56,$BF
	.dc.b	$C2,$5C,$D7,$E2,$BE,$07,$FD,$4E,$9F,$F4,$DC,$6B,$1F,$E2,$2D,$AF
	.dc.b	$DE,$CF,$03,$77,$FE,$C9,$C3,$FD,$04,$49,$FE,$27,$D2,$8C,$47,$FB
	.dc.b	$2F,$E4,$3F,$D8,$E7,$17,$FD,$A1,$18,$7F,$D8,$11,$97,$F5,$9E,$FB
	.dc.b	$DF,$7B,$FF,$C8,$23,$0D,$F1,$1F,$15,$9A,$DD,$FF,$29,$F9,$87,$AD
	.dc.b	$FB,$98,$FE,$5F,$18,$E0,$FE,$56,$D8,$E1,$3B,$7F,$9B,$7B,$FF,$C3
	.dc.b	$6B,$F9,$17,$17,$F2,$DF,$7B,$F6,$F1,$FC,$EA,$75,$5F,$97,$F0,$15
	.dc.b	$A8,$E7,$E0,$D5,$5C,$30,$FE,$1C,$FF,$45,$C3,$F8,$CB,$97,$F9,$DD
	.dc.b	$F1,$97,$BF,$63,$0D,$9F,$E4,$33,$E2,$7F,$91,$DF,$F5,$1F,$7F,$FA
	.dc.b	$97,$BC,$F1,$FA,$D6,$A9,$5A,$9C,$C2,$B8,$3F,$D6,$67,$4F,$F4,$EC
	.dc.b	$7F,$96,$CE,$A3,$FA,$FC,$E9,$FE,$97,$3A,$AF,$EC,$13,$9C,$3D,$15
	.dc.b	$DE,$72,$D3,$B6,$1C,$DF,$61,$93,$AB,$97,$D6,$66,$7A,$E1,$E5,$8E
	.dc.b	$2F,$9E,$33,$5E,$08,$C6,$1E,$7A,$F8,$07,$37,$0F,$B8,$FE,$B1,$87
	.dc.b	$F5,$97,$C2,$7F,$7A,$7F,$BD,$FB,$DF,$CF,$BD,$FB,$E2,$9A,$4A,$FF
	.dc.b	$02,$4F,$F3,$6C,$B8,$F8,$1F,$DD,$27,$2D,$7F,$C1,$93,$B7,$33,$F4
	.dc.b	$73,$FE,$BF,$9F,$D5,$69,$FE,$14,$AE,$11,$F6,$5B,$1F,$E1,$83,$75
	.dc.b	$FE,$1B,$6A,$8D,$BF,$E1,$F6,$BF,$8F,$DF,$FC,$43,$5E,$16,$D7,$87
	.dc.b	$0F,$64,$F8,$5B,$1E,$19,$DD,$78,$5A,$F8,$67,$6F,$85,$AF,$84,$F8
	.dc.b	$51,$C2,$FC,$B5,$FE,$00,$D7,$B5,$7F,$82,$A3,$0F,$3C,$78,$B9,$BF
	.dc.b	$EA,$93,$A8,$F1,$4F,$E7,$FD,$FF,$E0,$71,$FE,$14,$6B,$FE,$0B,$6C
	.dc.b	$7F,$83,$DF,$EC,$70,$73,$7F,$E4,$86,$C7,$F9,$2D,$AE,$F1,$FE,$CD
	.dc.b	$EF,$FF,$22,$8F,$F2,$4B,$5F,$F2,$73,$5F,$F2,$93,$5F,$F2,$AB,$BB
	.dc.b	$FF,$0A,$F1,$7F,$5A,$C3,$FC,$96,$DA,$FF,$2E,$56,$E7,$FC,$BB,$EF
	.dc.b	$3D,$2F,$EF,$0D,$AF,$F1,$AB,$5F,$EF,$DE,$F3,$E0,$F7,$5F,$E4,$1A
	.dc.b	$D5,$7F,$99,$DA,$F6,$8D,$1F,$E4,$5A,$7C,$05,$39,$CF,$83,$7F,$19
	.dc.b	$1F,$E7,$B2,$FA,$1C,$47,$F9,$CF,$0F,$F2,$2B,$E1,$AF,$21,$1E,$3F
	.dc.b	$03,$E1,$38,$ED,$FE,$49,$F5,$1F,$01,$FE,$19,$6D,$7F,$95,$9A,$FC
	.dc.b	$2E,$ED,$4F,$F9,$86,$73,$3F,$E8,$76,$BF,$E8,$A6,$BF,$E8,$D7,$9D
	.dc.b	$7F,$93,$33,$7C,$0F,$C5,$3A,$D7,$F9,$9B,$DF,$7C,$0F,$F8,$B5,$B1
	.dc.b	$98,$F9,$9C,$59,$7F,$CE,$93,$A9,$FF,$4C,$35,$FF,$4E,$35,$FF,$4F
	.dc.b	$BB,$C3,$FC,$BF,$ED,$BE,$07,$FC,$F8,$DA,$F0,$1F,$E7,$AF,$80,$7C
	.dc.b	$0F,$F9,$05,$B1,$98,$E2,$64,$FF,$44,$46,$A3,$FD,$58,$D7,$FD,$60
	.dc.b	$D7,$FD,$66,$EE,$FC,$BD,$1E,$43,$A4,$FF,$A3,$3D,$F7,$C0,$FF,$94
	.dc.b	$DB,$19,$8F,$5F,$D1,$65,$FF,$4D,$4E,$A7,$FD,$7C,$D7,$FD,$84,$D7
	.dc.b	$FD,$8A,$EE,$FE,$3F,$C1,$7C,$63,$A4,$FF,$A7,$BE,$09,$E6,$7F,$99
	.dc.b	$5B,$19,$8F,$9B,$E7,$E5,$FF,$56,$4E,$A7,$FD,$A0,$D7,$FD,$A8,$D7
	.dc.b	$FD,$AE,$EE,$FE,$CF,$6F,$D8,$3A,$4F,$FA,$C3,$D1,$73,$FF,$9D,$DB
	.dc.b	$19,$8F,$F3,$D7,$C0,$BE,$0B,$51,$27,$47,$FA,$EA,$74,$E0,$FF,$59
	.dc.b	$57,$46,$CF,$4A,$E5,$18,$38,$61,$FE,$BB,$BE,$71,$92,$47,$FA,$FE
	.dc.b	$FA,$3B,$AE,$6E,$6E,$F3,$FF,$BE,$B0,$E7,$39,$C3,$BC,$FF,$BF,$B0
	.dc.b	$E6,$EF,$0E,$32,$13,$A7,$0D,$57,$DA,$BA,$61,$CA,$78,$BF,$05,$F0
	.dc.b	$9C,$9D,$4F,$18,$C1,$A3,$89,$F3,$EF,$A5,$F8,$4F,$14,$F1,$9D,$9F
	.dc.b	$22,$FC,$6F,$13,$A4,$F1,$BF,$24,$6C,$F0,$94,$F6,$DF,$82,$3C,$ED
	.dc.b	$8F,$35,$F8,$6B,$A4,$62,$74,$F3,$9D,$56,$30,$FB,$53,$E4,$8E,$AD
	.dc.b	$39,$14,$E5,$5C,$70,$F6,$C7,$FC,$01,$BE,$AE,$EF,$AB,$EB,$C3,$CA
	.dc.b	$EE,$F9,$3F,$02,$76,$7E,$1D,$E1,$87,$C6,$8C,$13,$0F,$0B,$C1,$A4
	.dc.b	$9E,$8F,$E3,$5F,$23,$97,$E6,$98,$36,$7E,$47,$0F,$5B,$E2,$3E,$D1
	.dc.b	$4F,$86,$71,$3F,$03,$3E,$37,$99,$89,$F0,$4F,$49,$F5,$4E,$A3,$13
	.dc.b	$E5,$C3,$F1,$B8,$7C,$6A,$E4,$79,$5C,$46,$A3,$C4,$D7,$CD,$7D,$9F
	.dc.b	$79,$8A,$F1,$D7,$2C,$3E,$74,$78,$5F,$C1,$CE,$CF,$34,$ED,$F8,$75
	.dc.b	$D6,$FE,$EB,$F0,$63,$CE,$48,$F2,$46,$EB,$AB,$67,$F1,$E7,$73,$8E
	.dc.b	$1F,$49,$FE,$08,$7E,$98,$A7,$EC,$0B,$5F,$C8,$5B,$5A,$7F,$2E,$60
	.dc.b	$FF,$7B,$B9,$7E,$97,$8B,$FC,$C8,$FA,$15,$FE,$A1,$C7,$F8,$E4,$60
	.dc.b	$D4,$FE,$06,$74,$FE,$26,$34,$FF,$37,$6B,$FA,$76,$E1,$FC,$C5,$B5
	.dc.b	$DA,$7A,$47,$67,$E2,$57,$67,$F9,$0C,$70,$7C,$07,$03,$C0,$53,$C0
	.dc.b	$5A,$F0,$36,$BC,$0F,$69,$F0,$4F,$D5,$8E,$24,$70,$72,$71,$3F,$89
	.dc.b	$7C,$26,$0F,$D1,$E1,$FC,$A2,$BA,$CE,$E7,$47,$B0,$F2,$1E,$C3,$C8
	.dc.b	$74,$AF,$0C,$7A,$1F,$11,$88,$E0,$E2,$3B,$36,$33,$18,$39,$11,$F3
	.dc.b	$E1,$E6,$76,$60,$F1,$46,$DF,$CF,$46,$DF,$51,$E8,$3D,$B7,$FA,$69
	.dc.b	$E1,$87,$DB,$8C,$4F,$63,$8D,$FA,$DB,$5D,$E7,$71,$83,$83,$8A,$FC
	.dc.b	$31,$D6,$FF,$AF,$E1,$E5,$9D,$36,$33,$5E,$C3,$DC,$7D,$4A,$E4,$FF
	.dc.b	$50,$9E,$0D,$38,$1F,$08,$B1,$C1,$FE,$A5,$1C,$1A,$70,$AF,$1B,$6B
	.dc.b	$C6,$FC,$86,$E1,$F6,$63,$B1,$E5,$AF,$D1,$3F,$0C,$CC,$6E,$FC,$4F
	.dc.b	$F2,$37,$E2,$CF,$F8,$1F,$DB,$7C,$17,$F0,$EA,$65,$FC,$42,$73,$87
	.dc.b	$F8,$56,$72,$FF,$85,$E3,$2A,$32,$3E,$4D,$F1,$1D,$6F,$A6,$1F,$C4
	.dc.b	$6D,$87,$27,$F8,$EB,$5D,$9F,$C5,$DE,$F8,$76,$F8,$47,$3F,$5A,$F1
	.dc.b	$46,$1F,$46,$1C,$9F,$33,$FC,$B2,$75,$18,$3F,$C6,$4E,$2B,$DF,$1D
	.dc.b	$A3,$B4,$3F,$0C,$78,$7E,$1A,$D1,$FC,$B6,$BA,$BF,$CE,$1A,$F8,$5F
	.dc.b	$84,$77,$BF,$9D,$F0,$AF,$17,$BB,$6B,$D6,$79,$AB,$C1,$F0,$6F,$85
	.dc.b	$A4,$9D,$E6,$F8,$F7,$76,$67,$FA,$35,$F1,$7F,$A2,$35,$CD,$FD,$78
	.dc.b	$C5,$FD,$EF,$86,$7A,$9F,$11,$B7,$E6,$15,$C3,$5F,$EB,$7F,$0C,$F4
	.dc.b	$7E,$43,$6B,$83,$5F,$A9,$8B,$F1,$FF,$EC,$2F,$D4,$9D,$4F,$F6,$46
	.dc.b	$BD,$AB,$F1,$D1,$88,$FF,$37,$3D,$67,$F2,$FC,$8F,$87,$FB,$AD,$AF
	.dc.b	$1D,$FD,$DF,$81,$7C,$3D,$24,$FC,$EF,$86,$7B,$B8,$8F,$EE,$55,$C1
	.dc.b	$FE,$E0,$D7,$35,$FA,$38,$C6,$5F,$AA,$B7,$C3,$FE,$41,$B7,$F3,$6B
	.dc.b	$7B,$7F,$81,$BE,$1D,$F0,$DF,$93,$6D,$7D,$36,$BF,$32,$DF,$5D,$BA
	.dc.b	$FD,$A5,$6A,$BF,$C1,$ED,$7B,$46,$8F,$D1,$FC,$47,$FA,$5F,$33,$FE
	.dc.b	$97,$AE,$F1,$FE,$AB,$CB,$FD,$84,$4B,$FF,$5E,$78,$F5,$FF,$0C,$78
	.dc.b	$C9,$FE,$0C,$D8,$FF,$0C,$35,$DC,$7F,$B2,$BC,$A9,$1F,$ED,$18,$FD
	.dc.b	$A7,$C5,$BD,$CF,$E2,$0D,$9F,$F1,$13,$5F,$F1,$43,$5F,$F1,$7F,$C4
	.dc.b	$BD,$A7,$C4,$FF,$8A,$63,$0D,$F3,$DF,$13,$9A,$DD,$FF,$8C,$FE,$1F
	.dc.b	$F7,$9C,$BF,$B4,$F1,$1F,$0F,$FC,$A2,$3F,$C7,$F3,$AA,$FD,$0D,$6A
	.dc.b	$7F,$59,$39,$9F,$F2,$1B,$5E,$3E,$3F,$F1,$86,$CF,$F8,$FF,$8F,$94
	.dc.b	$7F,$1F,$3F,$9C,$75,$F8,$5F,$E7,$D8,$D7,$35,$FB,$BA,$E2,$7F,$24
	.dc.b	$2B,$FE,$4F,$F8,$7F,$F6,$D7,$C3,$3D,$0F,$E9,$98,$73,$C3,$FD,$BD
	.dc.b	$5A,$9C,$DD,$9F,$F2,$EC,$D7,$FA,$65,$7F,$50,$BE,$91,$A0,$E1,$DD
	.dc.b	$71,$7B,$9A,$BE,$ED,$39,$4F,$1C,$3E,$2B,$CE,$38,$B9,$BF,$15,$7B
	.dc.b	$E7,$91,$E0,$29,$C1,$EF,$3C,$E1,$87,$B6,$FB,$D7,$3A,$ED,$87,$A6
	.dc.b	$F6,$7F,$67,$BF,$ED,$0E,$0F,$F3,$D3,$FE,$7B,$F8,$6F,$E6,$7F,$0F
	.dc.b	$7C,$83,$49,$5F,$E8,$59,$FE,$65,$97,$1F,$0B,$FB,$D4,$E5,$AF,$FA
	.dc.b	$42,$76,$E6,$7E,$B6,$7F,$D5,$63,$E1,$B7,$3F,$A9,$5C,$1B,$7D,$9F
	.dc.b	$B1,$3B,$3F,$D3,$E5,$A0,$AE,$51,$FE,$A4,$6C,$6A,$38,$3F,$91,$DB
	.dc.b	$FD,$53,$5E,$0C,$3F,$19,$7C,$2B,$47,$AA,$B7,$7D,$70,$F3,$36,$F5
	.dc.b	$50,$FC,$0D,$F3,$9F,$85,$3F,$E0,$86,$E9,$EC,$C3,$FC,$0D,$5B,$7F
	.dc.b	$C1,$13,$AB,$F5,$46,$B0,$F0,$C7,$C2,$F8,$67,$BD,$FE,$0F,$6C,$7F
	.dc.b	$6F,$6B,$FC,$23,$E3,$B5,$3F,$E2,$F9,$CC,$FF,$B2,$9A,$FF,$B3,$9A
	.dc.b	$FD,$0F,$88,$FE,$2A,$FF,$7B,$6B,$FC,$25,$BA,$7F,$89,$9F,$ED,$9A
	.dc.b	$7F,$B6,$1C,$B7,$87,$FC,$4D,$18,$AE,$D9,$62,$64,$77,$77,$18,$8F
	.dc.b	$04,$5B,$F0,$4D,$BE,$D5,$C4,$F4,$46,$5F,$65,$55,$7E,$87,$27,$13
	.dc.b	$B9,$C8,$C5,$FB,$E7,$94,$F1,$9E,$71,$83,$D2,$F0,$BF,$2B,$D0,$F9
	.dc.b	$57,$B7,$47,$C4,$29,$D0,$B5,$DC,$F9,$4E,$4E,$97,$F1,$1D,$DF,$9A
	.dc.b	$BB,$46,$AF,$EB,$46,$2B,$D9,$3E,$F3,$47,$C4,$29,$F2,$23,$84,$FB
	.dc.b	$27,$C1,$3A,$BF,$29,$2B,$71,$CE,$76,$78,$1E,$17,$F2,$1A,$7C,$09
	.dc.b	$E2,$FB,$E7,$8D,$F8,$49,$1D,$5F,$6B,$F0,$8E,$45,$3C,$D3,$D2,$FC
	.dc.b	$D1,$9A,$F2,$1D,$9E,$84,$3C,$55,$D9,$F3,$9D,$0E,$4F,$91,$F9,$06
	.dc.b	$0E,$53,$F6,$EF,$C5,$7E,$37,$E1,$5F,$8A,$F9,$57,$9A,$FB,$CE,$5F
	.dc.b	$C4,$18,$9D,$DF,$38,$FB,$93,$B7,$D2,$78,$A3,$4F,$E5,$0F,$31,$A3
	.dc.b	$D2,$75,$38,$1A,$AF,$24,$F5,$7F,$03,$7E,$AB,$F4,$BF,$93,$AE,$0E
	.dc.b	$5F,$86,$E5,$EC,$78,$67,$E3,$3D,$EF,$CF,$7F,$56,$35,$5C,$63,$C4
	.dc.b	$DB,$F0,$CF,$E0,$8E,$B7,$E8,$8C,$C7,$D9,$33,$3E,$57,$A9,$C6,$BA
	.dc.b	$B6,$78,$C7,$C1,$BF,$2C,$7E,$16,$37,$3D,$5A,$6D,$F8,$67,$D6,$8E
	.dc.b	$67,$DA,$72,$7D,$53,$CA,$6C,$F9,$27,$8C,$A7,$77,$9B,$F6,$0E,$27
	.dc.b	$38,$F1,$C7,$78,$CC,$7C,$DB,$FB,$71,$D4,$F8,$67,$53,$BC,$7A,$DF
	.dc.b	$E3,$C7,$F7,$A3,$F8,$51,$4F,$E3,$05,$8F,$E9,$AD,$BD,$3F,$CA,$0C
	.dc.b	$18,$38,$9F,$C0,$25,$F1,$7F,$ED,$E8,$FF,$70,$B5,$FC,$C4,$6B,$CF
	.dc.b	$FF,$7D,$CE,$63,$E0,$C9,$D3,$CA,$7F,$8C,$C6,$23,$FD,$AB,$3C,$70
	.dc.b	$F9,$B3,$90,$FF,$86,$9C,$3A,$8E,$13,$DD,$AF,$CC,$9E,$2E,$6B,$CE
	.dc.b	$78,$A7,$8C,$F4,$9F,$A3,$5D,$E7,$27,$4C,$3E,$CC,$FA,$2F,$9B,$FC
	.dc.b	$4E,$F9,$1C,$DF,$15,$F3,$9D,$C7,$CA,$BE,$B1,$E3,$33,$5F,$28,$E7
	.dc.b	$1E,$1C,$3F,$B9,$46,$27,$85,$7C,$D3,$C5,$5D,$1B,$1D,$2F,$E4,$36
	.dc.b	$3D,$E7,$C5,$39,$1E,$57,$4F,$A8,$D1,$F0,$4E,$A7,$0A,$F4,$57,$86
	.dc.b	$39,$19,$AF,$0B,$4F,$0C,$FA,$DE,$67,$53,$89,$5E,$B5,$E4,$C3,$E9
	.dc.b	$C7,$DE,$39,$4F,$07,$17,$F2,$63,$FC,$0B,$1A,$39,$BE,$33,$DC,$E6
	.dc.b	$FC,$F1,$E1,$6C,$74,$BF,$A3,$39,$3D,$75,$F0,$67,$AC,$64,$EB,$19
	.dc.b	$AE,$67,$E1,$63,$E0,$CF,$56,$C6,$5F,$63,$63,$A9,$B8,$C9,$F5,$9C
	.dc.b	$C6,$CE,$F1,$96,$C7,$E5,$EF,$E5,$C6,$E7,$47,$B2,$FE,$86,$1F,$92
	.dc.b	$C3,$D5,$39,$3E,$55,$F2,$78,$E1,$D6,$B8,$5F,$29,$FA,$77,$C9,$F8
	.dc.b	$B1,$97,$FC,$53,$5F,$4D,$EC,$7C,$BB,$ED,$19,$8E,$27,$8A,$FB,$B4
	.dc.b	$EF,$5D,$67,$FC,$00,$71,$AE,$77,$E3,$6C,$EF,$0F,$E4,$D3,$9C,$3A
	.dc.b	$E1,$E3,$6E,$1F,$4C,$FA,$31,$F2,$8F,$29,$B3,$C6,$79,$4F,$7C,$F5
	.dc.b	$7F,$90,$57,$F4,$1E,$B7,$43,$AF,$F9,$D4,$61,$BD,$6E,$BC,$D6,$F0
	.dc.b	$FE,$87,$D4,$EF,$FE,$4B,$1F,$D2,$E3,$19,$7F,$45,$6C,$70,$9D,$BF
	.dc.b	$D4,$F4,$6B,$FB,$1B,$C7,$5D,$AF,$E7,$13,$5F,$E8,$91,$6B,$8D,$7B
	.dc.b	$A7,$6D,$7F,$B0,$75,$FF,$BF,$EF,$94,$7F,$B7,$23,$13,$FE,$FF,$6C
	.dc.b	$3A,$5D,$1F,$D8,$A7,$31,$89,$FE,$D5,$18,$9D,$47,$17,$DD,$7F,$D2
	.dc.b	$AB,$84,$F9,$0D,$1E,$82,$99,$77,$87,$BA,$32,$7A,$4D,$B9,$8E,$B1
	.dc.b	$97,$8C,$F8,$6B,$CB,$18,$C3,$E0,$57,$F7,$B1,$85,$69,$97,$FC,$01
	.dc.b	$D2,$EB,$75,$E1,$B5,$FE,$08,$9D,$4E,$4E,$3C,$FF,$DF,$67,$4D,$7F
	.dc.b	$C2,$92,$8C,$47,$FC,$01,$37,$A7,$FC,$01,$CB,$E5,$E6,$EB,$FF,$0F
	.dc.b	$47,$CE,$C3,$95,$2F,$E6,$C7,$FC,$01,$79,$5F,$5A,$E1,$87,$FC,$00
	.dc.b	$9F,$C0,$E1,$D2,$BF,$C5,$47,$78,$E9,$5C,$30,$B3,$CC,$DD,$71,$AF
	.dc.b	$14,$71,$7C,$77,$DF,$0F,$24,$7E,$0A,$75,$87,$58,$C5,$FE,$A7,$4E
	.dc.b	$B1,$F9,$FA,$E9,$8F,$ED,$30,$E6,$FE,$DA,$33,$5C,$2B,$FC,$5C,$F9
	.dc.b	$DF,$8D,$D2,$EA,$ED,$18,$6A,$8D,$3F,$B1,$B7,$63,$B3,$75,$AA,$EE
	.dc.b	$73,$3F,$6C,$E6,$BF,$B9,$57,$19,$FF,$37,$C7,$F9,$87,$87,$FD,$6B
	.dc.b	$18,$C7,$FC,$F8,$7F,$9B,$1F,$F4,$B6,$DF,$E8,$07,$FD,$67,$5F,$E4
	.dc.b	$C7,$27,$13,$33,$C7,$B3,$FD,$25,$1F,$E1,$C3,$67,$F9,$CA,$3B,$B9
	.dc.b	$8D,$BF,$E1,$D3,$73,$DE,$32,$7F,$5B,$8C,$CF,$7B,$FF,$4E,$4E,$4D
	.dc.b	$BE,$E9,$F0,$1E,$63,$C0,$F0,$35,$5D,$DC,$47,$FA,$D4,$F7,$1F,$EC
	.dc.b	$88,$D3,$FE,$97,$3D,$13,$E0,$78,$1E,$03,$85,$79,$DE,$11,$E7,$AE
	.dc.b	$F1,$FE,$BA,$BF,$A0,$F0,$3C,$E7,$08,$F3,$D7,$38,$FF,$50,$36,$FE
	.dc.b	$F4,$78,$9F,$3B,$A3,$99,$4F,$4C,$F8,$A3,$E8,$57,$8A,$FF,$DA,$F5
	.dc.b	$E5,$3E,$1B,$89,$DC,$62,$78,$11,$C1,$0F,$89,$3C,$6F,$F1,$13,$B9
	.dc.b	$CB,$6B,$D2,$7B,$23,$C4,$68,$E8,$EA,$7F,$1B,$1E,$C3,$E4,$1F,$0D
	.dc.b	$F8,$C6,$AB,$E9,$46,$8E,$B3,$CD,$FC,$41,$CC,$F9,$11,$C1,$B7,$F5
	.dc.b	$1F,$9B,$3E,$73,$E2,$BE,$72,$3D,$5F,$BC,$75,$78,$57,$C4,$8F,$6B
	.dc.b	$E8,$7C,$75,$E1,$9D,$C7,$EA,$EF,$E4,$56,$E3,$AD,$6E,$3E,$05,$F9
	.dc.b	$E7,$C1,$18,$7E,$31,$EA,$70,$FC,$47,$A4,$FD,$98,$F0,$1E,$63,$81
	.dc.b	$E6,$9E,$D1,$E1,$8C,$1D,$8F,$09,$83,$B1,$C8,$F4,$C7,$22,$3C,$63
	.dc.b	$A3,$A3,$11,$A7,$99,$D8,$D1,$88,$E2,$7A,$DF,$DD,$D7,$CA,$7C,$51
	.dc.b	$93,$F8,$74,$66,$3A,$46,$67,$F9,$EB,$FC,$0E,$3B,$47,$8E,$FF,$06
	.dc.b	$7E,$06,$78,$4F,$F8,$BD,$E1,$1A,$BF,$9F,$7D,$E7,$CE,$DB,$E9,$3F
	.dc.b	$87,$BF,$C6,$BC,$1F,$A8,$79,$A7,$83,$F1,$88,$79,$8B,$7E,$17,$E0
	.dc.b	$C7,$07,$44,$8D,$3D,$4E,$F7,$E6,$6D,$F9,$9F,$C3,$9E,$63,$A9,$A3
	.dc.b	$C6,$7F,$83,$CF,$F0,$84,$7F,$97,$9C,$19,$70,$E0,$38,$77,$E0,$3D
	.dc.b	$9F,$F1,$B3,$63,$80,$70,$F3,$6B,$E0,$EC,$FF,$3F,$3D,$DA,$E7,$AF
	.dc.b	$FD,$00,$E7,$AB,$FD,$00,$FA,$32,$FF,$29,$1F,$E8,$01,$D1,$FE,$73
	.dc.b	$CF,$FC,$E8,$DD,$7F,$CE,$53,$C1,$FF,$28,$F1,$7F,$99,$1F,$F3,$B5
	.dc.b	$FF,$95,$E3,$0F,$F9,$78,$FF,$28,$F2,$FF,$95,$DC,$74,$7F,$9B,$DF
	.dc.b	$F3,$89,$FE,$62,$C7,$FC,$BC,$E2,$78,$1F,$E6,$47,$9D,$76,$8C,$57
	.dc.b	$39,$FF,$38,$4E,$4F,$3B,$FE,$75,$9D,$BD,$EF,$FC,$D2,$FD,$2A,$FF
	.dc.b	$34,$11,$FF,$31,$74,$7F,$99,$1F,$F2,$EB,$5F,$C3,$73,$7F,$99,$9D
	.dc.b	$F4,$7F,$99,$B3,$91,$FE,$60,$21,$FE,$66,$E3,$FF,$2B,$67,$FE,$46
	.dc.b	$3F,$C7,$99,$FF,$91,$C8,$E0,$93,$FE,$63,$C3,$FC,$CA,$FF,$94,$AF
	.dc.b	$FC,$C6,$FF,$95,$BA,$3F,$CB,$0A,$77,$7D,$23,$47,$03,$58,$7F,$96
	.dc.b	$CF,$2D,$72,$C3,$CD,$3E,$19,$D6,$1C,$67,$33,$CA,$3F,$CD,$6E,$DF
	.dc.b	$F3,$D7,$8D,$FE,$80,$BF,$D7,$36,$73,$5F,$E8,$78,$C4,$7F,$94,$0F
	.dc.b	$F3,$D9,$FE,$67,$9F,$F2,$DE,$3F,$E6,$6D,$7F,$CC,$98,$7F,$99,$E7
	.dc.b	$38,$FA,$AF,$81,$0E,$97,$C5,$F8,$67,$7C,$7F,$CD,$EF,$A4,$FF,$3C
	.dc.b	$9C,$6F,$B3,$FE,$D7,$9E,$4F,$98,$E5,$7E,$29,$F8,$71,$97,$8B,$A3
	.dc.b	$95,$F9,$DC,$12,$3F,$D0,$91,$97,$89,$30,$FF,$41,$4F,$57,$E1,$91
	.dc.b	$F5,$65,$F3,$27,$84,$7C,$07,$A9,$E6,$C7,$E0,$C7,$33,$55,$DA,$32
	.dc.b	$4C,$3D,$D3,$F1,$8F,$44,$F8,$5F,$C2,$19,$AF,$0B,$D4,$CD,$F9,$67
	.dc.b	$85,$F8,$B1,$F9,$AF,$7A,$F8,$F7,$E7,$C3,$FD,$3B,$39,$9F,$BB,$8F
	.dc.b	$C7,$C3,$9D,$79,$B1,$ED,$5E,$C2,$1E,$2B,$DB,$F7,$EB,$E0,$57,$62
	.dc.b	$47,$88,$F6,$47,$07,$C3,$87,$95,$E4,$76,$C7,$D7,$5C,$A3,$67,$23
	.dc.b	$C7,$7F,$07,$1F,$54,$62,$7D,$53,$DE,$FE,$5D,$F4,$AE,$E7,$AC,$E0
	.dc.b	$61,$D3,$F5,$8F,$C2,$CF,$3C,$3E,$C4,$EA,$BA,$E1,$E2,$BF,$B1,$7E
	.dc.b	$59,$F7,$47,$33,$EE,$57,$1B,$E7,$7C,$6F,$EB,$5F,$E4,$2F,$E0,$1D
	.dc.b	$30,$F6,$BE,$83,$A6,$1C,$DF,$95,$7C,$E7,$E5,$57,$3C,$7E,$ED,$73
	.dc.b	$3E,$B3,$8C,$3C,$38,$FD,$C9,$F8,$F1,$EB,$7C,$6E,$9F,$8A,$78,$EB
	.dc.b	$E1,$9F,$00,$FC,$DC,$71,$AE,$D7,$E4,$BE,$D7,$F1,$6B,$47,$CF,$7D
	.dc.b	$77,$CE,$71,$1F,$64,$F5,$1F,$0B,$0F,$80,$FD,$B6,$B9,$9E,$4F,$9E
	.dc.b	$7C,$31,$E6,$BF,$9B,$3D,$23,$D6,$FC,$08,$F4,$3F,$00,$D4,$FD,$A9
	.dc.b	$EB,$18,$35,$7F,$1D,$FC,$0D,$F9,$27,$AE,$1F,$12,$7D,$11,$CA,$B8
	.dc.b	$63,$F7,$67,$8C,$E7,$1F,$1C,$7C,$3A,$EC,$7C,$A7,$DC,$7C,$A3,$D4
	.dc.b	$76,$29,$F1,$4B,$3E,$48,$F9,$E7,$38,$F6,$1E,$23,$78,$FC,$16,$9D
	.dc.b	$F0,$F6,$3D,$5B,$5D,$63,$FA,$69,$FE,$4A,$A5,$FF,$90,$A3,$FA,$41
	.dc.b	$97,$11,$A8,$FF,$23,$B5,$FF,$28,$35,$E0,$6A,$3F,$91,$F6,$48,$FC
	.dc.b	$EC,$57,$FA,$F4,$7C,$C9,$FE,$76,$E6,$BF,$0D,$23,$65,$58,$7F,$99
	.dc.b	$EB,$FC,$A3,$3F,$9D,$BE,$53,$9B,$ED,$18,$9E,$95,$C2,$FB,$C7,$62
	.dc.b	$38,$3B,$BC,$2B,$93,$F5,$A7,$95,$F7,$8F,$79,$E7,$70,$F4,$7E,$FC
	.dc.b	$F2,$9F,$C5,$B6,$FE,$23,$6B,$F8,$2B,$5F,$DD,$36,$FF,$80,$B6,$FF
	.dc.b	$AC,$B6,$7F,$9D,$B5,$FE,$52,$DB,$FA,$31,$E7,$3D,$C7,$A8,$F4,$18
	.dc.b	$3E,$71,$B0,$FC,$27,$55,$F2,$6B,$E0,$BF,$2A,$37,$5F,$51,$E5,$5C
	.dc.b	$5C,$9E,$19,$E1,$19,$25,$FC,$69,$E3,$5C,$DF,$39,$C8,$F3,$98,$9E
	.dc.b	$2F,$CC,$38,$9D,$EF,$CE,$47,$F7,$A7,$3B,$F1,$D7,$76,$C7,$23,$BB
	.dc.b	$87,$C1,$7E,$76,$9E,$12,$9F,$70,$E5,$1F,$25,$F9,$75,$E3,$7F,$21
	.dc.b	$5F,$66,$BF,$14,$F8,$E7,$57,$EE,$AE,$87,$6B,$FC,$3B,$DA,$7F,$26
	.dc.b	$D3,$DF,$3F,$3E,$BA,$47,$88,$87,$88,$E0,$53,$B4,$FE,$49,$CC,$FE
	.dc.b	$79,$B7,$F0,$EF,$E3,$1E,$F8,$F9,$D5,$EA,$9D,$36,$3E,$44,$F0,$8F
	.dc.b	$4C,$78,$0F,$CF,$9D,$8D,$BC,$5F,$87,$1C,$5F,$7D,$FC,$B7,$CE,$FC
	.dc.b	$07,$C6,$7C,$0B,$F5,$3D,$CF,$D5,$9D,$CE,$B5,$DE,$BE,$7C,$6E,$7F
	.dc.b	$5E,$DB,$FB,$2F,$E4,$4E,$D7,$E1,$9F,$A5,$19,$3C,$64,$7C,$75,$EB
	.dc.b	$8F,$A4,$D8,$CD,$F8,$CF,$A9,$3E,$13,$DE,$75,$8E,$EF,$E6,$A3,$F1
	.dc.b	$75,$E3,$3F,$85,$71,$B8,$BF,$DF,$F5,$FC,$DC,$46,$7B,$71,$19,$87
	.dc.b	$27,$F3,$88,$C5,$72,$7F,$A0,$7B,$CF,$47,$F8,$44,$75,$8C,$65,$D2
	.dc.b	$37,$3A,$6C,$7F,$BA,$F6,$FF,$7D,$E9,$FE,$F4,$96,$AE,$5F,$F8,$03
	.dc.b	$1F,$EF,$A3,$FD,$F1,$3B,$9D,$5F,$FA,$AA,$7F,$AA,$72,$FF,$C0,$16
	.dc.b	$9F,$EF,$D1,$87,$FC,$00,$3C,$2F,$C2,$8E,$09,$A4,$7F,$BA,$3F,$DD
	.dc.b	$5D,$18,$3F,$DF,$E4,$3E,$6F,$0B,$95,$CB,$F1,$67,$FD,$E5,$18,$7E
	.dc.b	$44,$FF,$BE,$9E,$3C,$9F,$11,$CB,$5C,$E5,$CE,$7F,$DF,$79,$FF,$71
	.dc.b	$E4,$FF,$7B,$1F,$F0,$02,$33,$5F,$E3,$48,$CB,$FE,$D2,$8C,$BF,$E2
	.dc.b	$B8,$CB,$F2,$27,$FD,$97,$3A,$7F,$DF,$A7,$FB,$F9,$E2,$66,$78,$E7
	.dc.b	$FD,$4A,$BF,$C9,$13,$97,$FD,$0B,$19,$7F,$CB,$91,$97,$FD,$91,$19
	.dc.b	$8F,$F2,$15,$7F,$8C,$CF,$F1,$FD,$7F,$96,$1B,$5F,$E7,$C7,$FC,$73
	.dc.b	$19,$3F,$84,$4F,$27,$13,$FB,$98,$C3,$C4,$FE,$5F,$1A,$7F,$60,$7C
	.dc.b	$93,$86,$1F,$E5,$BD,$3F,$C8,$58,$7F,$8F,$9B,$87,$8A,$BA,$1F,$E1
	.dc.b	$5B,$E9,$8F,$F2,$DC,$7C,$33,$FE,$C0,$C3,$93,$FE,$4B,$BE,$4E,$5B
	.dc.b	$5F,$E9,$D6,$E1,$FE,$7C,$3F,$D0,$F7,$D2,$75,$5E,$C3,$C7,$5F,$21
	.dc.b	$E8,$64,$A7,$9C,$DE,$1F,$2B,$1F,$81,$1E,$53,$85,$76,$78,$1C,$AB
	.dc.b	$B4,$65,$B8,$79,$E7,$FB,$DE,$1E,$78,$C4,$7F,$9C,$23,$0F,$D1,$3E
	.dc.b	$99,$F9,$73,$E9,$D7,$FA,$15,$FF,$3F,$63,$F2,$9F,$F3,$C3,$6F,$F1
	.dc.b	$58,$FA,$EB,$85,$7F,$9D,$E7,$E5,$BD,$E3,$EB,$63,$F0,$23,$9C,$E9
	.dc.b	$F1,$1E,$4C,$3C,$EE,$2B,$89,$D8,$FF,$41,$3F,$19,$F8,$97,$E3,$8F
	.dc.b	$F4,$59,$D8,$F9,$B8,$FC,$4B,$FF,$52,$E1,$E3,$7D,$73,$F0,$8E,$2E
	.dc.b	$1F,$24,$6E,$BF,$23,$5F,$86,$8C,$4F,$DE,$39,$63,$F1,$EB,$71,$F5
	.dc.b	$A7,$6F,$DD,$8F,$44,$EE,$39,$1E,$0C,$3E,$05,$FE,$92,$3D,$67,$CD
	.dc.b	$7C,$51,$A8,$FC,$61,$F1,$63,$E0,$57,$FB,$03,$1F,$9A,$4C,$3C,$77
	.dc.b	$F2,$AF,$D3,$8F,$D5,$AD,$4F,$59,$F9,$87,$53,$C4,$7B,$5D,$4F,$C9
	.dc.b	$7E,$CC,$7C,$CB,$F8,$6F,$E5,$6F,$C3,$87,$9A,$3F,$23,$7E,$19,$CC
	.dc.b	$76,$38,$5F,$6B,$E7,$5F,$72,$31,$3F,$29,$F0,$E3,$F3,$63,$E8,$1E
	.dc.b	$E3,$E5,$9F,$8F,$C3,$E5,$BE,$73,$93,$88,$E5,$3F,$1E,$7E,$B1,$B3
	.dc.b	$A9,$88,$DC,$FB,$63,$F4,$C7,$83,$1F,$9A,$F1,$AE,$87,$A2,$FE,$13
	.dc.b	$EC,$8E,$AE,$5E,$58,$FC,$67,$D4,$7D,$33,$A5,$F8,$67,$F5,$35,$FC
	.dc.b	$2A,$7E,$63,$F2,$5D,$98,$C7,$F1,$17,$D7,$0F,$3E,$3E,$CA,$D1,$F6
	.dc.b	$0F,$1E,$1E,$73,$D7,$8F,$C9,$C7,$D7,$3D,$8F,$39,$F7,$0C,$5F,$68
	.dc.b	$DD,$F7,$9D,$C7,$9E,$31,$3B,$9F,$09,$CB,$0F,$41,$30,$F2,$61,$F0
	.dc.b	$CF,$D6,$4F,$E6,$CE,$B8,$FC,$09,$DC,$7B,$8F,$B2,$6E,$FC,$0B,$77
	.dc.b	$9B,$FE,$17,$8C,$36,$33,$1F,$E1,$B6,$BF,$E4,$1E,$3F,$F2,$73,$78
	.dc.b	$BF,$C9,$4F,$F9,$4B,$8F,$FC,$9C,$8F,$F2,$AC,$69,$FF,$1C,$EB,$FE
	.dc.b	$36,$7F,$C8,$47,$F9,$1B,$8B,$FC,$8A,$EB,$7F,$F2,$F6,$9F,$E4,$48
	.dc.b	$FF,$0C,$1F,$E3,$D3,$FC,$85,$5F,$E4,$19,$DD,$72,$9F,$F1,$DC,$F7
	.dc.b	$7A,$1F,$E5,$03,$93,$97,$73,$CB,$6F,$F2,$94,$60,$E0,$7F,$9C,$78
	.dc.b	$BF,$C9,$04,$7F,$C9,$07,$F9,$1B,$87,$FC,$84,$DE,$2F,$F1,$E4,$62
	.dc.b	$3F,$CF,$5C,$1F,$E4,$B9,$FF,$20,$71,$7F,$90,$E7,$2D,$7E,$64,$FF
	.dc.b	$92,$DC,$E9,$C0,$FF,$18,$E5,$FE,$47,$8F,$F2,$15,$7F,$93,$F3,$FF
	.dc.b	$18,$36,$75,$3F,$E3,$47,$FC,$87,$19,$8F,$B5,$1F,$E4,$4E,$2F,$F2
	.dc.b	$84,$E6,$7F,$B2,$D7,$F8,$FE,$BF,$C9,$13,$9C,$3F,$C8,$13,$9B,$E3
	.dc.b	$87,$90,$F3,$E1,$DA,$32,$FD,$E3,$AE,$1E,$6A,$F4,$D7,$19,$F4,$1C
	.dc.b	$4C,$46,$9C,$89,$F2,$CE,$9E,$84,$8F,$F2,$DE,$1E,$D8,$F8,$11,$B3
	.dc.b	$FC,$2D,$5E,$B7,$07,$98,$F1,$1E,$9C,$3F,$CD,$35,$CF,$0F,$F3,$4D
	.dc.b	$FC,$EC,$3E,$0D,$F9,$67,$E0,$E1,$E9,$3B,$47,$AB,$0E,$75,$C2,$FD
	.dc.b	$73,$DF,$0E,$CF,$4A,$DB,$F0,$A7,$77,$D5,$F8,$11,$E1,$3D,$73,$F6
	.dc.b	$23,$2F,$29,$F1,$92,$3D,$53,$D7,$0F,$4C,$7E,$52,$30,$74,$BF,$2B
	.dc.b	$EB,$7F,$0C,$62,$BA,$47,$D2,$C3,$DD,$7D,$23,$E8,$DF,$92,$F8,$D6
	.dc.b	$A7,$B3,$5F,$88,$FD,$69,$FB,$71,$82,$5F,$81,$F8,$46,$23,$D5,$19
	.dc.b	$7E,$49,$83,$E4,$98,$8F,$29,$F8,$28,$F9,$CE,$0E,$AE,$1F,$C4,$18
	.dc.b	$8F,$97,$1D,$EF,$E0,$4E,$A7,$B3,$63,$38,$7D,$78,$F0,$4F,$23,$CB
	.dc.b	$3A,$C3,$E7,$47,$C9,$C3,$E2,$4F,$E5,$2B,$E8,$1F,$60,$F8,$47,$A6
	.dc.b	$7C,$37,$DE,$3D,$07,$0C,$3D,$0F,$D2,$39,$3E,$F8,$E1,$3C,$B0,$F1
	.dc.b	$D7,$A0,$E5,$1E,$87,$95,$74,$7C,$67,$C1,$33,$1B,$73,$5D,$AF,$E0
	.dc.b	$61,$F8,$3B,$F1,$61,$E0,$BF,$49,$FE,$D6,$71,$87,$96,$B8,$3F,$2E
	.dc.b	$7C,$C7,$0C,$3F,$49,$1E,$1B,$EB,$5C,$2B,$D6,$70,$9E,$8F,$C4,$7A
	.dc.b	$3E,$27,$33,$EF,$39,$C6,$CE,$66,$F0,$F6,$B9,$3D,$55,$E1,$69,$9B
	.dc.b	$F3,$BE,$39,$E3,$5D,$EF,$C6,$7E,$99,$F0,$99,$9F,$0C,$64,$DC,$FE
	.dc.b	$26,$3B,$CE,$DB,$3F,$14,$DB,$FC,$57,$0F,$DC,$BE,$56,$CF,$1B,$F3
	.dc.b	$CF,$BC,$FB,$E7,$A4,$A6,$4F,$C0,$CF,$DA,$BF,$6D,$F8,$67,$D3,$1E
	.dc.b	$7A,$EB,$3E,$B9,$C9,$E5,$C3,$F4,$AD,$3F,$23,$1B,$7D,$A7,$8C,$F0
	.dc.b	$C7,$D1,$9E,$B5,$DE,$FD,$B7,$E3,$BE,$27,$A7,$0F,$9F,$19,$BE,$33
	.dc.b	$E1,$9F,$E1,$13,$E1,$9C,$CF,$F2,$BC,$3C,$ED,$3C,$F8,$78,$63,$37
	.dc.b	$FC,$7E,$B8,$B7,$0F,$2D,$FA,$E3,$C6,$DC,$3E,$CB,$D6,$3C,$78,$75
	.dc.b	$8E,$E7,$9E,$7F,$93,$3C,$4F,$9A,$71,$72,$FC,$17,$AC,$F8,$F0,$F3
	.dc.b	$C7,$F1,$79,$EA,$71,$3F,$8A,$57,$DE,$7B,$1F,$32,$3E,$35,$FB,$23
	.dc.b	$B1,$E2,$BE,$CF,$46,$9E,$C7,$D4,$76,$3F,$1F,$3D,$2F,$CB,$1F,$C1
	.dc.b	$30,$F9,$73,$E7,$AE,$53,$E7,$3F,$8C,$1F,$86,$3F,$41,$5C,$9F,$E1
	.dc.b	$97,$E7,$AE,$D1,$E5,$7A,$1F,$20,$CD,$76,$3D,$2E,$70,$FA,$91,$98
	.dc.b	$E3,$3E,$B3,$C5,$1F,$AF,$AF,$05,$70,$AE,$55,$F0,$8F,$4D,$7E,$11
	.dc.b	$F9,$67,$CF,$9E,$0F,$58,$D4,$F9,$8E,$71,$F0,$F0,$FE,$6C,$FC,$CA
	.dc.b	$E2,$F6,$9F,$C1,$9E,$87,$07,$D9,$71,$3C,$A3,$D7,$1F,$A9,$BF,$5C
	.dc.b	$7A,$67,$57,$E4,$BE,$F1,$8B,$EF,$5F,$6E,$F8,$BE,$77,$F8,$63,$E7
	.dc.b	$3A,$38,$7F,$AA,$F0,$7F,$A9,$9F,$F5,$11,$FE,$A3,$70,$E1,$70,$7F
	.dc.b	$AA,$23,$FD,$4F,$C9,$FE,$8B,$E6,$FF,$2C,$47,$FA,$AF,$4F,$F5,$14
	.dc.b	$62,$3F,$D2,$41,$C4,$6B,$1F,$0C,$FF,$AC,$F7,$FF,$5A,$56,$B8,$5C
	.dc.b	$38,$BD,$DA,$F0,$F8,$6B,$6F,$79,$DC,$FF,$AF,$F7,$FF,$55,$36,$3F
	.dc.b	$D5,$53,$FE,$7B,$3C,$47,$13,$A5,$F6,$7D,$78,$79,$63,$11,$C0,$F2
	.dc.b	$9E,$53,$FD,$5E,$78,$8F,$4E,$DF,$EA,$B6,$F0,$FF,$AB,$DF,$F5,$4B
	.dc.b	$54,$6D,$F8,$EF,$FA,$B5,$FF,$57,$EF,$FE,$B1,$76,$D8,$D4,$7E,$50
	.dc.b	$FD,$11,$F8,$4D,$3F,$A1,$4F,$E2,$B3,$FD,$AC,$FE,$97,$2F,$C7,$BF
	.dc.b	$C0,$2B,$F8,$14,$7F,$80,$8F,$CD,$38,$E0,$FE,$3D,$3F,$B0,$C3,$F8
	.dc.b	$7B,$F9,$79,$D3,$F9,$53,$F3,$0F,$88,$CB,$FA,$63,$F9,$3C,$72,$7C
	.dc.b	$47,$F0,$F3,$A1,$4C,$DF,$F0,$B7,$2D,$73,$7F,$BE,$8F,$7C,$7C,$59
	.dc.b	$ED,$3E,$D8,$F0,$E1,$F1,$63,$F7,$6F,$EA,$EB,$F6,$D3,$F2,$CF,$E4
	.dc.b	$E7,$8E,$79,$3F,$0E,$79,$3E,$48,$F1,$1D,$2F,$E1,$47,$1C,$3E,$5C
	.dc.b	$FF,$15,$8F,$A8,$7F,$0D,$39,$18,$3E,$E9,$F3,$0F,$1D,$FC,$18,$F1
	.dc.b	$9C,$E3,$E0,$5F,$91,$F8,$A7,$3C,$3E,$1D,$F8,$8F,$E2,$73,$F0,$6B
	.dc.b	$F0,$A7,$48,$CE,$1E,$AC,$3C,$11,$C7,$0F,$8D,$1F,$CE,$E7,$EF,$BF
	.dc.b	$2A,$3D,$0E,$5F,$84,$7E,$1D,$FA,$F8,$7F,$23,$8F,$84,$7A,$0F,$5C
	.dc.b	$F4,$BF,$2D,$F8,$67,$AC,$7E,$61,$EA,$7E,$3C,$FA,$D3,$C5,$EB,$1E
	.dc.b	$D7,$D2,$F1,$BF,$6C,$79,$63,$ED,$3F,$41,$FB,$AF,$D6,$3C,$A7,$82
	.dc.b	$3E,$01,$EC,$C3,$E6,$47,$82,$FD,$13,$9B,$F9,$97,$D2,$BE,$06,$1F
	.dc.b	$0D,$FC,$0D,$6D,$F1,$1F,$6F,$0F,$BD,$5E,$B8,$F8,$67,$E7,$AF,$E0
	.dc.b	$3C,$A3,$D7,$5F,$4C,$F7,$C7,$C7,$8F,$C5,$5F,$84,$FC,$3C,$7C,$33
	.dc.b	$E0,$39,$BF,$0E,$1E,$CA,$EA,$7C,$B3,$D8,$79,$6F,$D4,$78,$5C,$C7
	.dc.b	$1B,$EB,$5F,$21,$F4,$18,$C3,$E3,$BF,$05,$F6,$E1,$E6,$8F,$55,$7E
	.dc.b	$61,$F8,$E7,$D6,$C3,$DE,$7D,$6C,$3C,$F1,$F1,$CF,$B4,$FE,$39,$F8
	.dc.b	$07,$D4,$C3,$C9,$7D,$EB,$94,$F9,$4E,$F1,$F1,$70,$F2,$D7,$C6,$9E
	.dc.b	$77,$F0,$CF,$0B,$F9,$63,$C2,$7E,$1C,$DE,$1F,$15,$F1,$57,$C2,$7D
	.dc.b	$93,$F2,$9D,$90,$F8,$67,$E7,$8F,$76,$1F,$0A,$3D,$D1,$A3,$F1,$27
	.dc.b	$8E,$34,$F8,$CD,$3F,$6A,$3E,$F4,$7D,$33,$D2,$7D,$83,$CE,$7C,$4B
	.dc.b	$F9,$27,$C3,$79,$C7,$C2,$3F,$23,$1A,$AF,$92,$49,$EB,$82,$7F,$D7
	.dc.b	$79,$FF,$85,$6D,$8F,$F8,$8A,$30,$D8,$FF,$12,$D7,$1C,$3F,$C0,$5C
	.dc.b	$1F,$E2,$6C,$7F,$C5,$1C,$4E,$0F,$F0,$0C,$FF,$8A,$39,$AC,$E6,$BF
	.dc.b	$C0,$A7,$F8,$17,$3F,$F0,$53,$FE,$0F,$3F,$C1,$B9,$7F,$85,$23,$FC
	.dc.b	$55,$1F,$E1,$4D,$7F,$C3,$29,$A4,$9F,$F0,$DB,$C2,$3F,$C2,$47,$F8
	.dc.b	$53,$83,$C2,$E0,$FF,$0A,$1F,$E0,$CD,$FF,$C2,$1A,$FF,$83,$E3,$E0
	.dc.b	$C6,$0C,$EF,$FE,$0D,$26,$DF,$E2,$0A,$FF,$07,$9F,$9B,$AF,$A3,$3F
	.dc.b	$43,$1F,$F2,$79,$F4,$23,$FC,$58,$D7,$E1,$4F,$CF,$6C,$FD,$D7,$9D
	.dc.b	$7D,$DC,$7F,$1F,$1F,$89,$BF,$D3,$3F,$AD,$8F,$E1,$0F,$E3,$DD,$DF
	.dc.b	$EC,$E7,$F3,$71,$FB,$49,$FC,$8B,$C1,$FD,$39,$C6,$3F,$50,$FE,$48
	.dc.b	$EC,$FE,$71,$D9,$27,$E1,$4F,$82,$FF,$34,$F7,$AF,$3D,$78,$1E,$55
	.dc.b	$C1,$EF,$3B,$DF,$F7,$91,$C0,$8E,$09,$5F,$BB,$9E,$0D,$9F,$79,$1F
	.dc.b	$DB,$46,$9F,$61,$E3,$3F,$80,$C7,$88,$E0,$FC,$F9,$EE,$F0,$8F,$1B
	.dc.b	$DC,$EA,$F0,$AF,$A1,$18,$21,$CC,$E0,$53,$A4,$F6,$AF,$31,$2B,$A5
	.dc.b	$FC,$C6,$E1,$F8,$57,$D1,$7D,$B0,$F0,$BD,$6F,$93,$9B,$E4,$DB,$E4
	.dc.b	$FE,$04,$87,$DD,$8F,$CA,$4F,$DD,$9D,$CF,$F9,$C9,$FC,$EE,$1F,$C2
	.dc.b	$E0,$30,$FD,$06,$7F,$EA,$7C,$7F,$40,$DC,$58,$72,$B8,$7E,$91,$FF
	.dc.b	$4A,$57,$E1,$E3,$CA,$D7,$CD,$1F,$D0,$63,$04,$9F,$4D,$FC,$6A,$F8
	.dc.b	$8F,$CA,$7F,$5E,$4C,$7D,$CD,$8E,$15,$F3,$09,$5F,$2A,$79,$9F,$2C
	.dc.b	$E0,$79,$9F,$0B,$6F,$CD,$18,$8F,$49,$FD,$0E,$32,$47,$04,$BF,$DC
	.dc.b	$57,$B6,$75,$18,$AE,$93,$9A,$E1,$7F,$B0,$9C,$92,$F8,$C6,$5B,$3F
	.dc.b	$A4,$3D,$07,$23,$07,$A2,$FC,$11,$C2,$3F,$B3,$4F,$C8,$AF,$8D,$7F
	.dc.b	$22,$BF,$12,$F9,$6F,$8D,$F5,$38,$9E,$07,$24,$BF,$34,$6D,$FE,$CB
	.dc.b	$5E,$16,$C6,$71,$FF,$10,$57,$18,$F4,$DF,$86,$B9,$CE,$67,$C2,$7D
	.dc.b	$4C,$7F,$B3,$BF,$94,$7F,$31,$87,$E5,$31,$FC,$CC,$FF,$5D,$BF,$CF
	.dc.b	$9F,$D0,$B1,$FD,$24,$7E,$9C,$FE,$C5,$B7,$F6,$D4,$D3,$FB,$63,$FA
	.dc.b	$A2,$3B,$7F,$B3,$4F,$F6,$23,$FD,$E1,$8F,$F6,$58,$FF,$7D,$9F,$EF
	.dc.b	$F3,$FD,$EC,$FF,$6C,$7F,$B3,$1F,$EF,$83,$B1,$FD,$9C,$FF,$7A,$F1
	.dc.b	$FF,$5F,$D3,$FB,$23,$4F,$EC,$58,$FF,$64,$9F,$F7,$FB,$5F,$F7,$CB
	.dc.b	$F8,$D7,$E9,$1F,$E2,$F3,$E6,$E9,$F4,$5F,$A7,$1B,$BF,$A5,$87,$DA
	.dc.b	$6E,$5F,$6A,$3E,$B9,$2B,$8E,$7F,$81,$9D,$46,$0F,$09,$FE,$3A,$9E
	.dc.b	$87,$17,$C7,$7E,$18,$C1,$A8,$E5,$7E,$27,$D5,$19,$78,$DF,$85,$CD
	.dc.b	$F8,$9C,$9E,$2A,$FF,$23,$3D,$2B,$BC,$ED,$E8,$47,$04,$8E,$A4,$3D
	.dc.b	$8E,$5A,$66,$7B,$BE,$27,$D4,$E8,$8E,$A7,$C1,$1C,$5F,$F2,$94,$F5
	.dc.b	$9D,$36,$33,$3B,$3F,$C0,$E7,$B4,$87,$85,$F7,$BB,$3F,$B9,$19,$36
	.dc.b	$F8,$9F,$2B,$D9,$A7,$63,$81,$1E,$13,$E6,$8E,$D1,$FD,$E1,$FE,$DE
	.dc.b	$7F,$72,$C3,$F8,$A3,$FC,$E0,$FE,$39,$1E,$67,$B9,$E6,$7A,$9E,$39
	.dc.b	$ED,$1A,$8E,$D1,$FD,$C2,$3F,$A2,$92,$7B,$C7,$F4,$E2,$4F,$F2,$7B
	.dc.b	$FE,$4B,$87,$E4,$EF,$E1,$E1,$F3,$DF,$81,$5E,$26,$C6,$5F,$33,$63
	.dc.b	$C6,$7F,$5F,$BE,$F3,$FC,$B6,$3D,$D1,$EA,$23,$ED,$24,$76,$3E,$5C
	.dc.b	$FF,$85,$E3,$A1,$E9,$8F,$95,$87,$EF,$5F,$A9,$3E,$33,$83,$FD,$9E
	.dc.b	$BC,$6F,$D0,$3D,$E7,$63,$EB,$5F,$F4,$A2,$61,$F1,$8F,$D8,$47,$D1
	.dc.b	$AE,$AF,$CF,$9D,$BE,$47,$E6,$3E,$43,$B1,$F8,$42,$57,$D9,$BF,$AF
	.dc.b	$87,$B6,$3C,$F1,$ED,$8F,$A5,$3F,$40,$FF,$10,$CF,$F4,$87,$FD,$35
	.dc.b	$1F,$EA,$EA,$FE,$8F,$3F,$E2,$59,$FF,$56,$F0,$7F,$AA,$8F,$F5,$6F
	.dc.b	$1B,$87,$13,$FE,$AC,$E0,$FF,$16,$F0,$7F,$AB,$0F,$F5,$13,$8C,$BF
	.dc.b	$D5,$87,$FA,$82,$61,$FC,$FA,$3F,$D4,$A7,$F3,$B7,$FA,$08,$CF,$FD
	.dc.b	$48,$FF,$42,$32,$E1,$FE,$B4,$7F,$3E,$D7,$FA,$19,$1F,$E8,$6F,$D4
	.dc.b	$CF,$F1,$71,$F8,$93,$EB,$1F,$72,$3E,$D3,$FC,$96,$FF,$1E,$FC,$B3
	.dc.b	$67,$D4,$7F,$1E,$FE,$64,$FC,$39,$A9,$FB,$1A,$7E,$01,$FA,$EF,$E0
	.dc.b	$27,$EB,$9F,$A4,$9F,$C9,$3F,$74,$8F,$53,$F0,$C7,$57,$C2,$FE,$5C
	.dc.b	$CC,$FB,$9F,$43,$63,$B4,$7E,$4E,$BD,$F3,$FC,$C8,$D6,$BF,$98,$C3
	.dc.b	$F1,$F5,$E1,$BF,$3C,$E7,$0F,$64,$79,$A7,$C4,$E9,$F9,$87,$E7,$F0
	.dc.b	$FA,$51,$F3,$0F,$1D,$F6,$9E,$37,$FB,$57,$8E,$1F,$22,$F8,$C7,$D9
	.dc.b	$8F,$3B,$AB,$F1,$5F,$E8,$EF,$94,$FD,$73,$E6,$9E,$B7,$35,$F5,$AF
	.dc.b	$E1,$DF,$F1,$23,$CD,$3E,$F8,$CD,$7D,$EA,$F0,$BD,$4F,$98,$7B,$CF
	.dc.b	$C5,$1E,$7A,$E6,$F9,$CF,$11,$FC,$02,$3C,$4F,$17,$EA,$47,$08,$ED
	.dc.b	$7F,$1C,$F7,$3C,$63,$F0,$71,$88,$FC,$31,$F0,$0F,$9A,$6C,$C3,$B9
	.dc.b	$F5,$9E,$23,$E8,$47,$A8,$FC,$5C,$F8,$E3,$4F,$D0,$C3,$F3,$98,$7C
	.dc.b	$C8,$F1,$C6,$2B,$C0,$7C,$6A,$EA,$7B,$9F,$8C,$43,$B9,$67,$F0,$F3
	.dc.b	$F9,$08,$F9,$27,$B8,$E9,$1E,$3C,$3F,$0F,$1C,$89,$1C,$A3,$ED,$C7
	.dc.b	$EA,$23,$D8,$7A,$23,$FC,$DC,$FF,$9A,$63,$FC,$DD,$11,$F8,$90,$BF
	.dc.b	$F3,$35,$FF,$9D,$27,$F9,$79,$FE,$D2,$C3,$FD,$9F,$5C,$E3,$FC,$ED
	.dc.b	$5F,$E7,$49,$CC,$7F,$32,$3F,$91,$F0,$7F,$9C,$F5,$FE,$65,$18,$E1
	.dc.b	$6F,$FC,$AA,$3F,$CD,$47,$F9,$DF,$5C,$3F,$E6,$CC,$3F,$94,$13,$87
	.dc.b	$FC,$D7,$1F,$C9,$88,$7F,$2B,$29,$FD,$1F,$5F,$E5,$8E,$48,$64,$D9
	.dc.b	$EF,$DB,$FC,$E4,$E4,$D9,$98,$E6,$E4,$86,$4E,$67,$F3,$E3,$25,$34
	.dc.b	$74,$30,$4A,$4F,$8A,$36,$72,$8C,$CF,$8A,$32,$43,$94,$F1,$6C,$6D
	.dc.b	$A7,$9A,$79,$B6,$78,$46,$23,$CC,$D7,$C5,$1A,$24,$F5,$9D,$4F,$8D
	.dc.b	$AF,$80,$E7,$3C,$1F,$04,$F1,$8C,$46,$A3,$04,$76,$43,$45,$3C,$AE
	.dc.b	$DC,$90,$C9,$B3,$CD,$1E,$07,$26,$CC,$C7,$37,$24,$32,$73,$3D,$C6
	.dc.b	$4A,$68,$E8,$60,$93,$F1,$89,$3E,$38,$E0,$73,$8C,$CF,$8E,$32,$43
	.dc.b	$9C,$F1,$6C,$6D,$A7,$9E,$79,$B6,$78,$46,$23,$CE,$D7,$C5,$1A,$24
	.dc.b	$F5,$9D,$4F,$8D,$AF,$80,$E7,$3C,$1F,$04,$F1,$8C,$46,$A3,$04,$76
	.dc.b	$43,$45,$3C,$AE,$DC,$90,$C9,$B3,$CD,$1E,$07,$26,$CC,$C7,$37,$24
	.dc.b	$32,$73,$3D,$C6,$4A,$68,$E8,$60,$93,$F1,$89,$3E,$38,$E0,$73,$8C
	.dc.b	$CF,$8E,$32,$43,$9C,$F1,$6C,$6D,$A7,$9E,$79,$B6,$78,$46,$23,$CE
	.dc.b	$D7,$C5,$1A,$24,$F5,$9D,$4F,$8D,$AF,$80,$E7,$3C,$1F,$04,$F1,$8C
	.dc.b	$46,$A3,$04,$76,$43,$45,$3C,$AE,$DC,$90,$C9,$B3,$CD,$1E,$07,$26
	.dc.b	$CC,$C7,$37,$24,$32,$73,$3D,$C6,$4A,$68,$E8,$60,$93,$F1,$89,$3E
	.dc.b	$38,$E0,$73,$8C,$CF,$8E,$32,$43,$9C,$F1,$6C,$6D,$A7,$9E,$79,$B6
	.dc.b	$78,$46,$23,$CE,$D7,$C5,$1A,$24,$F5,$9D,$4F,$8D,$AF,$80,$E7,$3C
	.dc.b	$1F,$04,$F1,$8F,$0D,$38,$54,$F8,$47,$8B,$88,$FF,$58,$56,$DD,$46
	.dc.b	$71,$4E,$6B,$8D,$7F,$AD,$5A,$F2,$BF,$F5,$E3,$FE,$11,$9F,$F0,$7B
	.dc.b	$63,$2F,$58,$CB,$63,$2F,$FB,$04,$E0,$7F,$B0,$A3,$6E,$4E,$06,$63
	.dc.b	$8B,$E3,$30,$43,$99,$0E,$25,$3A,$10,$E4,$64,$A7,$02,$1B,$AF,$F1
	.dc.b	$74,$78,$9B,$1E,$13,$67,$23,$84,$FB,$1C,$12,$38,$BE,$16,$BC,$9A
	.dc.b	$F2,$9F,$04,$F3,$9C,$CF,$38,$C1,$27,$64,$8F,$31,$CC,$A7,$17,$4F
	.dc.b	$8E,$39,$3A,$9E,$E4,$8C,$C6,$E7,$8B,$E9,$8C,$B6,$32,$F8,$4E,$07
	.dc.b	$58,$DB,$93,$81,$98,$E2,$F9,$CC,$10,$E6,$43,$89,$4E,$84,$39,$19
	.dc.b	$29,$C0,$86,$E7,$E0,$47,$89,$B1,$E1,$36,$72,$38,$4F,$B1,$C1,$23
	.dc.b	$8B,$E1,$6B,$C9,$AF,$29,$F0,$4F,$39,$CC,$F3,$8C,$12,$76,$48,$F3
	.dc.b	$1C,$CA,$71,$74,$F8,$E3,$93,$A9,$EE,$48,$CC,$6E,$78,$BE,$98,$CB
	.dc.b	$63,$2F,$84,$E0,$75,$8D,$B9,$38,$19,$8E,$2F,$9C,$C1,$0E,$64,$38
	.dc.b	$94,$E8,$43,$91,$92,$9C,$08,$6E,$7E,$04,$78,$9B,$1E,$13,$67,$23
	.dc.b	$84,$FB,$1C,$12,$38,$BE,$16,$BC,$9A,$F2,$9F,$04,$F3,$9C,$CF,$38
	.dc.b	$C1,$27,$64,$8F,$31,$CC,$A7,$17,$4F,$8E,$39,$3A,$9E,$E4,$8C,$C6
	.dc.b	$E7,$8B,$E9,$8C,$B6,$32,$F8,$4E,$07,$58,$DB,$93,$81,$98,$E2,$F9
	.dc.b	$CC,$10,$E6,$43,$89,$4E,$84,$39,$19,$29,$C0,$86,$E7,$E0,$47,$89
	.dc.b	$B1,$E1,$36,$72,$38,$4F,$B1,$C1,$23,$8B,$E1,$6B,$C9,$AF,$29,$F0
	.dc.b	$4F,$39,$CC,$F3,$8C,$12,$76,$48,$F3,$1C,$CF,$47,$E5,$22,$00

;--------------------------------------------------------------------------------
;	FETOX.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fetox_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FETOX.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FETOX.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fetox_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fetox.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FETOX.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fetox.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FETOX.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fetox_data_compressed::
	.dc.b	$00,$02,$84,$2C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CF,$ED,$1F,$8A,$7F,$87,$A7,$FC,$D0,$5F,$8A,$E1
	.dc.b	$F0,$FC,$7D,$FE,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$1F
	.dc.b	$07,$9F,$DB,$F4,$7B,$7F,$DF,$7E,$6F,$FA,$B0,$F8,$8C,$7C,$46,$7E
	.dc.b	$23,$5F,$11,$BF,$88,$E1,$F1,$1C,$7E,$17,$97,$C0,$F3,$F8,$1E,$9F
	.dc.b	$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03
	.dc.b	$E3,$F8,$1F,$27,$C0,$F9,$7E,$07,$CD,$F0,$3E,$7F,$81,$FF,$0A,$7C
	.dc.b	$E7,$E1,$BE,$13,$FC,$40,$BE,$07,$3E,$66,$BE,$07,$7F,$03,$C3,$E0
	.dc.b	$78,$FC,$0F,$2F,$81,$E7,$F0,$3F,$63,$D2,$EB,$F0,$3D,$BE,$07,$BF
	.dc.b	$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93,$E0,$7C,$BF
	.dc.b	$03,$E6,$F8,$1F,$3F,$C0,$FA,$3E,$07,$FC,$59,$F3,$BF,$88,$E9,$FF
	.dc.b	$1C,$7C,$07,$F0,$A5,$1F,$EF,$63,$FD,$60,$7F,$9A,$0F,$EB,$47,$F1
	.dc.b	$23,$81,$FC,$41,$FF,$21,$2F,$7B,$BF,$BD,$E0,$F7,$BC,$3E,$F7,$8B
	.dc.b	$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD,$E8,$F7,$BF
	.dc.b	$3B,$EF,$FE,$7B,$5F,$CF,$EB,$FA,$0D,$7F,$43,$AF,$BB,$DE,$F7,$FB
	.dc.b	$DF,$E4,$BF,$6F,$FC,$48,$7E,$3F,$8B,$FC,$B0,$BD,$CD,$3B,$F7,$38
	.dc.b	$7B,$9C,$7D,$CE,$5E,$E7,$3F,$73,$A7,$B9,$FB,$0E,$8E,$DE,$E7,$7F
	.dc.b	$73,$C1,$EE,$7E,$D3,$AF,$F6,$BD,$7F,$B6,$EB,$FD,$BF,$5F,$EE,$3A
	.dc.b	$FF,$73,$D7,$FB,$AE,$6F,$DD,$F3,$7F,$98,$FF,$15,$FF,$1E,$1E,$9E
	.dc.b	$73,$E9,$7F,$CE,$4F,$FB,$38,$FF,$44,$1F,$DE,$0F,$E6,$07,$F2,$F7
	.dc.b	$8F,$0F,$5E,$1E,$DC,$1D,$F8,$3E,$47,$BB,$F2,$7D,$DF,$95,$EE,$FC
	.dc.b	$BF,$77,$E6,$7B,$BF,$37,$DD,$F9,$DE,$EF,$CF,$F7,$7E,$87,$BB,$FE
	.dc.b	$7B,$F9,$FF,$D4,$70,$1F,$D4,$3F,$E8,$95,$EE,$6F,$DC,$E1,$EE,$71
	.dc.b	$F7,$39,$7B,$9C,$FD,$CE,$9E,$E7,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B
	.dc.b	$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$F3,$7C,$9F,$9D,$D9,$F3
	.dc.b	$FB,$3E,$87,$67,$D1,$EC,$FA,$5D,$9F,$4F,$B3,$EA,$76,$7D,$5E,$CF
	.dc.b	$AD,$D9,$FE,$99,$F4,$FD,$E7,$BB,$B4,$FB,$9F,$F5,$42,$EB,$E1,$D7
	.dc.b	$C7,$AF,$97,$5F,$3E,$BE,$9D,$7D,$7A,$FE,$67,$B7,$F3,$7D,$BF,$9D
	.dc.b	$ED,$F8,$7B,$BE,$87,$B7,$FE,$AB,$FA,$0F,$E0,$1C,$05,$F0,$99,$F8
	.dc.b	$AD,$7C,$AE,$FE,$17,$87,$C0,$F1,$F8,$6E,$5F,$01,$CF,$E0,$3A,$74
	.dc.b	$75,$F8,$0F,$81,$E4,$FC,$1F,$27,$E1,$79,$3F,$0F,$C9,$F8,$9E,$4F
	.dc.b	$8F,$E0,$3C,$9E,$D7,$97,$DA,$F3,$7B,$5E,$7F,$6B,$D1,$ED,$7A,$7D
	.dc.b	$AF,$57,$B5,$EB,$F6,$BD,$9E,$D7,$FB,$1F,$E8,$7F,$D4,$07,$E9,$4F
	.dc.b	$D2,$76,$97,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1F,$67,$D9
	.dc.b	$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BB,$E1,$F7,$7C,$5E,$EF
	.dc.b	$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE,$F9,$FD,$DF,$47,$BB,$F8,$7F
	.dc.b	$57,$D5,$EE,$FF,$B8,$7C,$FF,$D3,$9F,$89,$3F,$11,$D2,$7F,$10,$FF
	.dc.b	$BC,$5F,$F7,$61,$FE,$A5,$3F,$CB,$AE,$FD,$9E,$9E,$CF,$5F,$67,$B7
	.dc.b	$AF,$DF,$CC,$F0,$79,$9E,$1F,$33,$C5,$E6,$78,$FC,$CF,$27,$67,$97
	.dc.b	$7F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$FB,$E7,$E8,$BF,$1E,$7E,$3B,$84
	.dc.b	$17,$B1,$BE,$3E,$1E,$C7,$1F,$63,$97,$B1,$CF,$D8,$E9,$EC,$75,$F6
	.dc.b	$3B,$7B,$1F,$AF,$DF,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E
	.dc.b	$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7A,$7D,$8F,$3F,$3B,$E8
	.dc.b	$FF,$DD,$87,$F3,$13,$F1,$A7,$E3,$1F,$6A,$F5,$F8,$BF,$EA,$83,$FC
	.dc.b	$C1,$C2,$EB,$CF,$AB,$AF,$57,$7E,$AF,$0F,$57,$8F,$AB,$CB,$D5,$E7
	.dc.b	$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E,$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F
	.dc.b	$57,$C7,$EA,$FB,$3E,$93,$F9,$99,$FC,$54,$FC,$69,$F8,$CE,$62,$F7
	.dc.b	$F8,$7B,$FC,$7D,$FE,$5E,$CF,$3F,$67,$A7,$B3,$D7,$D9,$ED,$EC,$F7
	.dc.b	$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9,$EC,$F9,$7C,$FF
	.dc.b	$37,$85,$E7,$F6,$7E,$9F,$C0,$7F,$1A,$3F,$19,$CB,$FC,$A7,$D9,$CF
	.dc.b	$B3,$AF,$67,$7E,$CF,$0F,$67,$8F,$B3,$CB,$D9,$E7,$E6,$74,$F6,$7A
	.dc.b	$FB,$3D,$BD,$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F,$67,$C7,$EC,$F9
	.dc.b	$3D,$9F,$2F,$B3,$E6,$F6,$7F,$13,$F4,$BF,$8A,$E6,$3F,$DE,$4F,$E2
	.dc.b	$A7,$E2,$5C,$FB,$5C,$3D,$DE,$3E,$EF,$2F,$6F,$9F,$B7,$D3,$DB,$EB
	.dc.b	$ED,$F6,$F6,$FB,$FB,$7E,$0F,$6F,$C3,$ED,$F8,$BD,$BF,$1F,$B7,$E4
	.dc.b	$F6,$FC,$BE,$DF,$9B,$DB,$F3,$FB,$7E,$8F,$6F,$F4,$3F,$4D,$FD,$E0
	.dc.b	$FE,$B4,$7F,$12,$3F,$0C,$7E,$17,$97,$F7,$AB,$DA,$E3,$ED,$72,$F6
	.dc.b	$B9,$FB,$5D,$3D,$AE,$BE,$D7,$6F,$6B,$BF,$B5,$E0,$F6,$BC,$3E,$D7
	.dc.b	$8B,$DA,$F1,$FB,$5E,$4F,$6B,$CB,$ED,$79,$BD,$AF,$3F,$B5,$FB,$8F
	.dc.b	$A7,$FE,$BE,$7E,$18,$FE,$1F,$C9,$FC,$49,$FE,$54,$DF,$63,$87,$B1
	.dc.b	$C7,$D8,$E5,$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1
	.dc.b	$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E
	.dc.b	$7F,$63,$F8,$7F,$AF,$F8,$73,$FC,$7C,$7F,$5A,$3F,$0C,$7F,$0D,$E4
	.dc.b	$FE,$56,$BD,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6,$F5
	.dc.b	$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC,$BE
	.dc.b	$BF,$9B,$D7,$F3,$FA,$FE,$8F,$5F,$F9,$3F,$D4,$7F,$8F,$8F,$F1,$01
	.dc.b	$FD,$60,$FE,$18,$7E,$1F,$94,$FE,$1D,$FE,$8A,$FF,$58,$1E,$AF,$4F
	.dc.b	$57,$AF,$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E
	.dc.b	$AF,$93,$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FA,$3D,$5F,$E7,$FF,$53
	.dc.b	$FE,$26,$3F,$86,$9F,$88,$E2,$3F,$88,$6F,$AB,$C3,$D5,$E3,$EA,$F2
	.dc.b	$F5,$79,$FA,$BD,$3D,$5E,$BE,$AF,$6F,$57,$BF,$AB,$E0,$F5,$7C,$3E
	.dc.b	$AF,$8B,$D5,$F1,$FA,$BE,$4F,$57,$CB,$EA,$F9,$BD,$5F,$3F,$AB,$FD
	.dc.b	$5B,$EA,$BF,$87,$1F,$E8,$53,$FC,$42,$7F,$0D,$3F,$11,$C8,$7F,$10
	.dc.b	$FF,$74,$5E,$A7,$3F,$53,$A7,$A9,$D7,$D4,$ED,$EA,$77,$F5,$3C,$1E
	.dc.b	$A7,$87,$D4,$F1,$7A,$9E,$3F,$53,$C9,$EA,$79,$7D,$4F,$37,$A9,$E7
	.dc.b	$F5,$3D,$1E,$A7,$A7,$D4,$FE,$DF,$E2,$7F,$A1,$8F,$F3,$41,$FE,$20
	.dc.b	$3F,$AA,$1F,$89,$3F,$11,$C8,$7F,$C3,$E7,$E2,$1F,$F0,$7A,$F4,$BA
	.dc.b	$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA,$5E,$2F,$4B,$C7,$E9,$79
	.dc.b	$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$A3,$D2,$FF,$03,$FD,$5F,$F9
	.dc.b	$B8,$FE,$AA,$7E,$24,$FC,$41,$E7,$E1,$FF,$11,$AF,$4B,$8F,$A5,$CB
	.dc.b	$D2,$E7,$E9,$74,$F4,$BA,$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA
	.dc.b	$5E,$2F,$4B,$C7,$E9,$79,$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$F8
	.dc.b	$7B,$EB,$3F,$AB,$1F,$EC,$D3,$FC,$D2,$7F,$55,$3F,$12,$7E,$23,$90
	.dc.b	$FE,$21,$FF,$1E,$BF,$EC,$D1,$E8,$75,$F4,$3B,$7A,$1D,$FD,$0F,$07
	.dc.b	$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F,$43,$CD,$E8,$79
	.dc.b	$FD,$0F,$47,$A1,$E9,$F4,$3F,$C6,$FF,$5B,$FE,$CE,$3F,$D6,$07,$F9
	.dc.b	$A0,$FF,$0E,$1F,$C4,$4F,$C4,$9F,$E6,$63,$F1,$1C,$47,$F1,$0F,$F9
	.dc.b	$61,$79,$FD,$BC,$FE,$FE,$7F,$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7
	.dc.b	$F9,$3C,$FF,$2F,$9F,$E6,$F3,$FC,$FE,$7F,$A3,$CF,$FF,$27,$FC,$17
	.dc.b	$FA,$D8,$FF,$0E,$9F,$C4,$4F,$E2,$07,$E2,$78,$3F,$C6,$9E,$7F,$0F
	.dc.b	$3F,$8F,$9F,$CB,$CF,$E7,$E7,$F4,$F3,$FA,$F9,$FD,$BC,$FE,$FE,$7F
	.dc.b	$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7,$F9,$3C,$FF,$2F,$9F,$E6,$F3
	.dc.b	$FF,$CC,$BF,$5D,$FE,$1D,$1F,$27,$8F,$8F,$CF,$C4,$EB,$DE,$DF,$BD
	.dc.b	$C3,$DA,$E3,$ED,$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$FE,$7C,$FA,$F7,$CB,$DF,$94,$C7,$C8,$67,$E2,$B5
	.dc.b	$E7,$6F,$E1,$B8,$7C,$37,$1F,$84,$E5,$F0,$3C,$FD,$EE,$9E,$F7,$5F
	.dc.b	$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27
	.dc.b	$BD,$E5,$F7,$BC,$DE,$F7,$FA,$4B,$EC,$3F,$DF,$27,$F9,$80,$FE,$A8
	.dc.b	$7F,$53,$3F,$08,$7F,$0B,$F7,$FF,$D5,$EB,$DE,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$FD,$53,$EC,$FF,$98
	.dc.b	$4F,$C2,$F1,$FF,$B0,$9F,$F7,$A9,$FE,$61,$3F,$AA,$9F,$C3,$4F,$E1
	.dc.b	$8F,$F8,$77,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8E,$FE,$C7,$83
	.dc.b	$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79,$3D,$8F,$2F,$B1,$E6,$F6,$3C
	.dc.b	$FE,$C7,$A3,$D8,$FF,$60,$FD,$8F,$E1,$B8,$BF,$DB,$2B,$D7,$CF,$AF
	.dc.b	$AF,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D
	.dc.b	$7F,$2F,$AF,$E6,$F5,$FC,$FE,$BF,$FB,$63,$EC,$9F,$35,$7E,$4F,$1F
	.dc.b	$1F,$9F,$8F,$D7,$C4,$EF,$E1,$B8,$7B,$BC,$7D,$DE,$5E,$EF,$3F,$77
	.dc.b	$A7,$BB,$D7,$DD,$ED,$EE,$F7,$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE
	.dc.b	$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$FE,$EE,$FB,$3F,$F5,$31,$F8
	.dc.b	$6F,$38,$BE,$17,$5F,$0B,$BF,$85,$E1,$F0,$3C,$7E,$17,$97,$C1,$73
	.dc.b	$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F,$82,$F0,$7C,$17,$87,$E0
	.dc.b	$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E,$0B,$CD,$F0,$5E,$7F,$82
	.dc.b	$F4,$7C,$13,$9E,$78,$FF,$87,$7F,$BE,$7C,$0E,$5F,$EF,$DF,$03,$67
	.dc.b	$FD,$DC,$7F,$98,$CF,$F0,$D8,$F7,$B5,$EF,$6F,$DE,$E1,$EF,$71,$F7
	.dc.b	$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE
	.dc.b	$1F,$7B,$C5,$EF,$78,$FD,$EF,$37,$DA,$7E,$23,$93,$D8,$FF,$BB,$8F
	.dc.b	$F3,$11,$FE,$61,$3F,$A9,$1F,$C3,$D3,$FC,$3D,$E3,$EC,$74,$F6,$3A
	.dc.b	$FB,$1D,$BD,$8E,$FE,$C7,$83,$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79
	.dc.b	$3D,$8F,$2F,$B1,$E6,$F6,$3C,$FE,$C7,$A3,$D8,$F9,$DF,$6B,$FE,$EF
	.dc.b	$3F,$14,$7E,$27,$87,$EB,$3F,$EE,$F3,$FC,$C6,$7F,$86,$CF,$E1,$E3
	.dc.b	$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD,$3E,$FE,$9F,$83,$D3,$F0,$FA
	.dc.b	$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F,$A7,$E6,$F4,$FC,$FE,$9F,$A3
	.dc.b	$D3,$FC,$37,$DB,$7E,$24,$FE,$1D,$C7,$F8,$D5,$E9,$EB,$E1,$37,$F0
	.dc.b	$9C,$3E,$13,$8F,$AB,$CB,$D5,$E7,$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E
	.dc.b	$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F,$57,$C7,$EA,$F9,$3D,$5F,$2F,$AB
	.dc.b	$E6,$F5,$7C,$FE,$AF,$E7,$78,$7F,$11,$CD,$FF,$00,$7E,$0B,$1F,$05
	.dc.b	$9F,$82,$D7,$C1,$6F,$DA,$E1,$F0,$5C,$7D,$AE,$5E,$D7,$3F,$6B,$A7
	.dc.b	$B5,$D7,$DA,$ED,$ED,$77,$F6,$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F
	.dc.b	$6B,$C9,$ED,$79,$7D,$AF,$37,$B5,$FB,$3F,$B7,$FD,$A7,$39,$FE,$20
	.dc.b	$7E,$1D,$C7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$D1,$EF,$7F,$0B,$FB
	.dc.b	$8F,$E1,$C7,$E1,$B9,$8F,$E1,$9F,$E2,$8B,$DD,$E1,$EE,$F1,$F3,$B9
	.dc.b	$7B,$BC,$FD,$DE,$9E,$EF,$5F,$77,$B7,$BB,$DF,$DD,$F0,$7B,$BE,$1F
	.dc.b	$77,$C5,$EE,$F8,$FD,$DF,$27,$BB,$E5,$F7,$7C,$DE,$EF,$9F,$DD,$F4
	.dc.b	$7B,$BF,$C8,$FE,$E7,$F0,$E3,$DE,$C7,$BD,$9F,$7B,$5E,$F6,$FD,$EE
	.dc.b	$1E,$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$75,$F7,$BB,$7B,$DD,$FD
	.dc.b	$EF,$07,$BD,$E1,$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD
	.dc.b	$EF,$7F,$39,$F6,$BF,$0E,$7F,$55,$3F,$0C,$3D,$DD,$7B,$BB,$F7,$78
	.dc.b	$4F,$1F,$77,$97,$BB,$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7B,$BD,$FD,$DF
	.dc.b	$07,$BB,$E1,$F7,$7C,$5E,$EF,$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE
	.dc.b	$F9,$FD,$DF,$EA,$1F,$75,$FD,$4C,$FC,$31,$F8,$5F,$78,$FE,$15,$FE
	.dc.b	$B4,$BD,$CE,$3E,$E7,$2F,$73,$9F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B
	.dc.b	$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7
	.dc.b	$9B,$DC,$F3,$FB,$9E,$8F,$73,$FB,$47,$DD,$FF,$0C,$3F,$0D,$CB,$FD
	.dc.b	$BD,$7B,$7A,$F6,$F7,$ED,$F0,$F6,$F8,$FB,$7C,$BD,$BE,$7E,$DF,$4F
	.dc.b	$6D,$FC,$03,$0B,$A7,$F7,$BF,$9F,$F8,$5F,$15,$FC,$1A,$31,$DF,$FC
	.dc.b	$79,$FE,$F9,$3A,$74,$D7,$F8,$6B,$5D,$C7,$17,$9C,$F5,$53,$B7,$27
	.dc.b	$49,$E0,$E9,$B1,$B8,$C4,$FF,$84,$FE,$D5,$B7,$82,$30,$D9,$4E,$9F
	.dc.b	$81,$E0,$7F,$85,$E3,$11,$D0,$DC,$F1,$8C,$DF,$C1,$77,$F4,$0F,$77
	.dc.b	$95,$6C,$97,$EC,$AE,$3C,$DA,$8C,$4E,$9F,$04,$F6,$C7,$51,$89,$D6
	.dc.b	$3E,$17,$FC,$5D,$8F,$86,$7D,$E7,$42,$1C,$4F,$BD,$87,$4A,$E5,$1A
	.dc.b	$6B,$F5,$2F,$59,$FF,$89,$71,$F8,$2D,$C7,$D1,$8F,$86,$FE,$39,$CE
	.dc.b	$74,$F5,$C7,$9E,$7E,$68,$F2,$13,$26,$5A,$8E,$66,$6F,$9A,$C3,$C1
	.dc.b	$97,$93,$0E,$98,$7A,$89,$8F,$39,$D5,$F4,$9D,$63,$9C,$BE,$6E,$5D
	.dc.b	$B2,$EB,$18,$CB,$AC,$F8,$AD,$97,$1C,$7E,$05,$78,$B1,$C4,$7A,$63
	.dc.b	$E0,$63,$D2,$36,$78,$F0,$E9,$7C,$30,$F5,$47,$7C,$30,$F8,$4F,$36
	.dc.b	$78,$C7,$CA,$7C,$2B,$C6,$1D,$63,$C8,$E8,$C1,$33,$FC,$DD,$70,$7E
	.dc.b	$39,$F9,$88,$E0,$E3,$3F,$C9,$E3,$B7,$F0,$75,$6B,$F0,$07,$08,$F9
	.dc.b	$4F,$D1,$9D,$3F,$82,$CF,$E2,$3F,$1C,$F2,$4F,$F1,$0C,$77,$9F,$D8
	.dc.b	$9E,$87,$CA,$C7,$D2,$E0,$E4,$FA,$5E,$66,$EB,$F8,$DE,$1D,$33,$F9
	.dc.b	$73,$C6,$FE,$8E,$7F,$2E,$31,$3F,$46,$3E,$5C,$74,$3B,$DF,$ED,$33
	.dc.b	$E9,$3C,$67,$57,$D1,$F8,$47,$48,$F4,$67,$F0,$A7,$59,$7C,$FA,$FD
	.dc.b	$C9,$B3,$9E,$5D,$9F,$D3,$D7,$1C,$BC,$D1,$E0,$CF,$57,$FC,$36,$39
	.dc.b	$3D,$5F,$D6,$65,$AC,$FE,$84,$69,$FB,$ED,$3F,$86,$E1,$DF,$3C,$47
	.dc.b	$F3,$CC,$39,$3F,$CB,$8C,$E7,$FC,$3A,$B8,$47,$C4,$CB,$F4,$2E,$9F
	.dc.b	$D0,$36,$FF,$53,$8F,$48,$EC,$64,$F9,$67,$38,$F9,$71,$BB,$FD,$BC
	.dc.b	$EF,$2F,$84,$EC,$FB,$58,$79,$23,$E2,$4F,$F0,$A9,$E1,$97,$E2,$1A
	.dc.b	$7F,$B8,$EB,$84,$7E,$7A,$BF,$20,$72,$C7,$E9,$BE,$03,$18,$F9,$AF
	.dc.b	$99,$EC,$3C,$91,$E2,$CF,$E0,$61,$DA,$7F,$34,$7E,$10,$DE,$1F,$59
	.dc.b	$A7,$D6,$CB,$E1,$67,$F0,$EF,$9C,$62,$7C,$39,$FF,$32,$7C,$78,$77
	.dc.b	$7F,$3A,$77,$C7,$D2,$FE,$48,$DD,$73,$7C,$27,$F3,$DB,$EE,$F6,$C3
	.dc.b	$C4,$F0,$30,$78,$A3,$6F,$2C,$3E,$93,$73,$FC,$2B,$73,$FC,$2B,$C8
	.dc.b	$C1,$F8,$4C,$7C,$E7,$E7,$F2,$F6,$5F,$EB,$B3,$FC,$FD,$70,$AF,$AD
	.dc.b	$1F,$6C,$FE,$3B,$1D,$63,$9C,$7C,$F6,$C7,$CF,$35,$1C,$5F,$ED,$D8
	.dc.b	$FB,$27,$71,$FD,$87,$2F,$86,$F8,$6B,$F8,$53,$C8,$FB,$E7,$CB,$3E
	.dc.b	$41,$B3,$EA,$3D,$DD,$19,$3F,$7A,$73,$3E,$D9,$A9,$F1,$DF,$ED,$73
	.dc.b	$FB,$51,$8C,$FE,$7B,$73,$F9,$F7,$D6,$FF,$5A,$DC,$BD,$D8,$F8,$E7
	.dc.b	$EC,$DF,$1A,$FA,$93,$9C,$3F,$98,$C7,$3C,$BE,$1D,$6C,$FE,$7B,$19
	.dc.b	$3F,$0C,$7C,$27,$55,$F6,$9B,$1F,$C4,$4F,$85,$9F,$E0,$2F,$F6,$75
	.dc.b	$B3,$C5,$97,$F3,$CB,$E3,$7F,$20,$F0,$1E,$D3,$C1,$3C,$1F,$F0,$66
	.dc.b	$1E,$88,$C4,$FA,$0E,$78,$FC,$66,$E5,$F1,$AF,$EE,$CF,$C6,$AE,$27
	.dc.b	$F0,$F8,$F4,$E7,$F5,$23,$31,$ED,$BF,$3E,$7E,$DC,$BF,$AD,$65,$E3
	.dc.b	$AE,$B9,$FC,$EB,$F1,$CE,$67,$F0,$79,$7E,$05,$FC,$33,$FA,$57,$F3
	.dc.b	$27,$C0,$39,$DF,$F5,$DA,$F5,$C6,$4E,$77,$FC,$F1,$F9,$07,$A8,$EA
	.dc.b	$FD,$27,$A9,$93,$E7,$C7,$47,$0F,$A1,$FC,$09,$8C,$7E,$1B,$FC,$A6
	.dc.b	$38,$1F,$13,$2F,$91,$1E,$83,$F2,$25,$8F,$AD,$3C,$B0,$FA,$CD,$CF
	.dc.b	$EB,$57,$B3,$1F,$81,$39,$8F,$81,$5F,$1D,$FD,$E1,$E6,$BF,$13,$F0
	.dc.b	$2B,$FB,$1D,$F4,$AE,$07,$E0,$67,$C9,$7F,$81,$BE,$31,$F1,$23,$C0
	.dc.b	$FC,$13,$67,$F0,$22,$9F,$E7,$B3,$CB,$8F,$CF,$7C,$A7,$D4,$9E,$A7
	.dc.b	$CD,$71,$3F,$33,$3F,$C6,$47,$F8,$2C,$F6,$46,$72,$F9,$B1,$F1,$4F
	.dc.b	$9A,$5A,$F9,$AF,$93,$0F,$9A,$DC,$7E,$16,$3E,$59,$F8,$53,$C7,$0F
	.dc.b	$84,$77,$7E,$51,$FD,$16,$B8,$E5,$FE,$0E,$CF,$F3,$97,$F3,$30,$FD
	.dc.b	$06,$1F,$76,$73,$1D,$CC,$E1,$E5,$8C,$DF,$F8,$73,$3F,$83,$97,$F8
	.dc.b	$12,$73,$5F,$D7,$A7,$38,$7F,$AD,$A7,$35,$FC,$FA,$73,$5F,$E6,$43
	.dc.b	$E6,$9F,$59,$CB,$F5,$23,$F3,$35,$EF,$8F,$C0,$1F,$1C,$FB,$77,$EF
	.dc.b	$7E,$99,$CF,$3F,$A7,$97,$C1,$9F,$E3,$95,$CE,$FE,$F1,$DA,$F8,$1F
	.dc.b	$31,$FE,$87,$9F,$CC,$8F,$EE,$59,$7B,$2B,$FA,$76,$5E,$8C,$7E,$B9
	.dc.b	$DE,$BF,$02,$DC,$3F,$03,$1F,$6C,$FC,$B1,$4F,$CB,$1F,$9B,$31,$1F
	.dc.b	$86,$6E,$1E,$39,$FE,$AF,$87,$8E,$7F,$B2,$DF,$8F,$3F,$E2,$15,$ED
	.dc.b	$9E,$D3,$9C,$BF,$73,$1F,$2C,$F0,$63,$E7,$CB,$FA,$24,$66,$FF,$4B
	.dc.b	$1F,$89,$8F,$9E,$74,$CF,$E7,$E3,$F8,$BC,$3E,$7C,$FC,$C3,$EC,$9E
	.dc.b	$A3,$EC,$9F,$30,$E7,$87,$85,$FF,$26,$57,$49,$FE,$0F,$1C,$30,$F8
	.dc.b	$2F,$0B,$F0,$3F,$E8,$2A,$E0,$FE,$94,$E1,$87,$F7,$3C,$BE,$44,$7F
	.dc.b	$80,$B2,$F8,$8F,$F5,$A3,$A6,$3F,$A1,$BE,$78,$FE,$12,$79,$9F,$D6
	.dc.b	$2F,$9E,$5F,$C8,$70,$E7,$5F,$E6,$2C,$7F,$8B,$BF,$2A,$FF,$5C,$76
	.dc.b	$3F,$55,$87,$69,$FF,$04,$67,$F7,$DF,$44,$EE,$BF,$59,$5B,$8F,$ED
	.dc.b	$35,$B7,$C8,$6E,$BF,$2F,$3B,$8F,$EB,$F3,$B7,$FB,$0E,$3E,$AC,$FF
	.dc.b	$93,$63,$EA,$C3,$F0,$75,$F6,$4F,$C3,$CF,$D0,$C7,$DD,$1F,$87,$9F
	.dc.b	$D7,$1F,$87,$7F,$7E,$7E,$1C,$FD,$73,$EE,$7E,$59,$E2,$3D,$07,$E7
	.dc.b	$31,$F4,$65,$FD,$1B,$2F,$43,$FD,$23,$2F,$46,$7F,$E8,$BB,$FC,$7C
	.dc.b	$FE,$2D,$FD,$6E,$3F,$45,$FD,$6D,$70,$7F,$D4,$38,$76,$CB,$F9,$96
	.dc.b	$3D,$9F,$E6,$95,$BB,$FF,$35,$56,$F1,$FF,$74,$5E,$DF,$F0,$8E,$1E
	.dc.b	$A3,$F1,$D3,$B7,$FD,$10,$7C,$A8,$FE,$34,$FC,$6C,$7D,$19,$FD,$18
	.dc.b	$FE,$0F,$1E,$B3,$FD,$2C,$F1,$3F,$54,$71,$CB,$EA,$67,$FE,$6A,$8F
	.dc.b	$87,$18,$7F,$C5,$11,$F5,$9F,$C9,$9C,$6F,$F4,$0F,$D6,$3F,$1A,$7D
	.dc.b	$6A,$FA,$0F,$B5,$FD,$06,$1E,$DC,$BF,$D7,$13,$BC,$3E,$4B,$B7,$F1
	.dc.b	$B3,$BC,$3F,$D1,$47,$8E,$7E,$A9,$BC,$3E,$E5,$FF,$38,$9F,$D5,$9B
	.dc.b	$C3,$E2,$67,$F5,$63,$FD,$2F,$7F,$02,$FF,$84,$BF,$1E,$FF,$84,$BE
	.dc.b	$F3,$F6,$27,$BC,$FD,$29,$E9,$8F,$B6,$EE,$3F,$62,$EE,$FF,$28,$EC
	.dc.b	$FF,$03,$1B,$AF,$D8,$BB,$C7,$F1,$AE,$EB,$FB,$63,$EF,$3F,$26,$6E
	.dc.b	$FF,$1A,$EE,$BF,$B7,$3B,$9F,$C9,$BB,$3F,$1A,$FE,$7E,$FE,$2C,$6E
	.dc.b	$7E,$2C,$FC,$B8,$FD,$C1,$FC,$DD,$C3,$FB,$F7,$95,$FE,$26,$3F,$C9
	.dc.b	$D3,$F4,$9E,$0F,$EF,$E7,$FA,$85,$7E,$6B,$1F,$BD,$5F,$E5,$E9,$E7
	.dc.b	$87,$E8,$1D,$E1,$D7,$2F,$D0,$57,$63,$CA,$7F,$09,$3E,$E1,$FC,$F4
	.dc.b	$FB,$98,$FF,$26,$7E,$81,$FC,$24,$F8,$95,$F9,$47,$67,$F2,$63,$67
	.dc.b	$ED,$8D,$C7,$E5,$1F,$69,$FA,$68,$F3,$DF,$E9,$A3,$C7,$1F,$D1,$5E
	.dc.b	$A7,$F1,$B3,$AE,$3F,$A6,$7A,$9F,$0E,$BF,$35,$87,$CD,$9F,$F0,$01
	.dc.b	$FB,$FC,$FF,$86,$57,$E3,$2F,$C2,$FE,$60,$F0,$DF,$F2,$B7,$C3,$5F
	.dc.b	$B6,$7C,$37,$FB,$07,$C2,$7F,$B8,$67,$C3,$87,$D3,$7E,$E4,$7C,$FB
	.dc.b	$FC,$1D,$F8,$E3,$F2,$47,$F8,$EE,$3C,$67,$E9,$1F,$19,$FC,$20,$F1
	.dc.b	$DF,$F0,$27,$C7,$7F,$BC,$BF,$C4,$1E,$0B,$FC,$44,$F9,$95,$7F,$BC
	.dc.b	$A7,$CF,$3F,$C8,$2F,$C3,$1F,$AE,$8F,$09,$FB,$A3,$C3,$7F,$BA,$9D
	.dc.b	$C7,$E1,$4D,$DF,$EE,$9D,$9F,$3E,$BC,$97,$F9,$5C,$BF,$37,$3F,$24
	.dc.b	$FC,$29,$FC,$60,$F6,$9F,$BA,$3A,$C7,$F0,$07,$AD,$FF,$00,$76,$FE
	.dc.b	$E9,$DE,$1F,$C0,$1D,$DF,$F0,$07,$6F,$EE,$8F,$92,$FE,$9B,$0F,$92
	.dc.b	$F2,$3F,$81,$C7,$C8,$AF,$E4,$27,$C2,$8F,$D0,$1B,$7F,$4C,$7E,$D9
	.dc.b	$F9,$D1,$F0,$AB,$F4,$13,$F8,$4B,$FC,$C6,$3F,$96,$8F,$61,$D7,$0F
	.dc.b	$D0,$61,$D7,$0F,$D6,$3F,$8D,$7D,$2F,$F0,$F7,$CC,$7E,$EE,$79,$5F
	.dc.b	$E6,$67,$91,$FA,$93,$E8,$D7,$F1,$E7,$E2,$80,$FF,$BF,$5E,$27,$D9
	.dc.b	$3C,$47,$EF,$6B,$DC,$FE,$B6,$BD,$D8,$7F,$4A,$7E,$21,$F2,$A7,$6F
	.dc.b	$E8,$4D,$CF,$F1,$57,$71,$EB,$36,$7E,$5D,$D9,$FC,$24,$DC,$7F,$15
	.dc.b	$7E,$0C,$FE,$9E,$3C,$C7,$DF,$BF,$CA,$3F,$A7,$7F,$4C,$FC,$47,$F8
	.dc.b	$66,$7F,$D9,$63,$E2,$61,$F0,$EF,$C3,$87,$F0,$97,$C3,$3F,$C6,$5F
	.dc.b	$0B,$F9,$B7,$6F,$EF,$5D,$CF,$F0,$F7,$67,$E6,$CD,$C7,$EC,$DD,$DF
	.dc.b	$F0,$37,$75,$F9,$B9,$F6,$9F,$83,$3D,$58,$7F,$15,$7E,$03,$F9,$28
	.dc.b	$FC,$4D,$FF,$15,$9F,$D3,$9F,$C5,$63,$E7,$BF,$BF,$AF,$1C,$FE,$FD
	.dc.b	$F1,$D7,$F0,$5A,$DD,$7E,$92,$77,$7F,$B9,$77,$1F,$C1,$4F,$19,$F8
	.dc.b	$79,$F1,$CF,$E9,$23,$F9,$39,$FE,$DD,$8E,$B7,$FA,$4A,$EB,$7F,$BF
	.dc.b	$7A,$9F,$AA,$9D,$DF,$E8,$63,$77,$FB,$C3,$E2,$9F,$B6,$3C,$6F,$E4
	.dc.b	$E7,$C7,$5F,$99,$7F,$5B,$3F,$8D,$7A,$C7,$E3,$63,$F8,$D4,$FD,$47
	.dc.b	$F0,$F4,$9F,$D6,$4F,$FB,$72,$FF,$60,$EE,$BF,$5F,$5B,$7F,$D6,$87
	.dc.b	$E5,$6B,$FD,$8A,$FE,$C0,$EF,$7E,$27,$F1,$CF,$A1,$FD,$53,$E8,$7F
	.dc.b	$28,$EE,$BF,$1C,$ED,$FD,$53,$BB,$FC,$A3,$B3,$F1,$D3,$E0,$3E,$3B
	.dc.b	$FB,$33,$91,$FA,$33,$95,$FE,$5D,$E5,$1F,$12,$FE,$1C,$7F,$9B,$EF
	.dc.b	$EA,$9F,$CA,$9E,$31,$F8,$D9,$EE,$7E,$4E,$3B,$BF,$AB,$76,$FE,$34
	.dc.b	$FD,$11,$F1,$EF,$E7,$C7,$8A,$3E,$7C,$F8,$9F,$F7,$49,$F3,$C5,$27
	.dc.b	$F9,$24,$7F,$A9,$1E,$F1,$F8,$F8,$F2,$9F,$A1,$8F,$2B,$FB,$97,$67
	.dc.b	$F5,$38,$FC,$05,$7F,$0B,$8D,$BF,$AD,$77,$3F,$B3,$3A,$C7,$E8,$63
	.dc.b	$77,$FA,$D7,$73,$FA,$73,$71,$FA,$18,$DD,$FE,$9D,$DD,$7F,$81,$A3
	.dc.b	$E1,$57,$F9,$24,$F9,$31,$FD,$92,$FE,$AD,$7F,$52,$8C,$3D,$27,$F5
	.dc.b	$87,$CF,$BF,$E5,$2F,$F9,$A4,$FC,$01,$F9,$C3,$FA,$5C,$F8,$5F,$9F
	.dc.b	$3F,$3A,$3F,$D6,$F5,$FB,$C8,$F9,$07,$D9,$8F,$91,$3F,$C8,$1F,$E3
	.dc.b	$53,$B8,$F9,$07,$AA,$BF,$9F,$BE,$07,$F9,$FB,$E7,$BF,$E4,$CE,$DF
	.dc.b	$E4,$CF,$58,$FE,$28,$FF,$26,$3E,$71,$B3,$F8,$D1,$B8,$FE,$10,$FF
	.dc.b	$0E,$9F,$EC,$0F,$D4,$8F,$E1,$15,$F9,$48,$C1,$FB,$A3,$91,$FB,$A3
	.dc.b	$95,$FF,$63,$77,$5F,$BA,$76,$FE,$E9,$D9,$FC,$5C,$DC,$7E,$EA,$7C
	.dc.b	$11,$E9,$9F,$E4,$0F,$29,$FD,$D3,$C8,$FE,$12,$72,$3E,$DC,$7F,$1B
	.dc.b	$BF,$BF,$1F,$E0,$A3,$F5,$A7,$DF,$8F,$C4,$BF,$E1,$13,$E8,$1F,$CC
	.dc.b	$4F,$14,$FF,$0B,$7C,$47,$EF,$4F,$15,$7F,$14,$7C,$47,$69,$F1,$5F
	.dc.b	$DA,$9F,$C1,$D7,$E7,$AF,$F9,$65,$7E,$EE,$FF,$BD,$E1,$FD,$F5,$E5
	.dc.b	$5F,$C4,$CF,$CE,$4F,$F1,$38,$F0,$D7,$F1,$B7,$C3,$7F,$92,$7F,$92
	.dc.b	$1E,$A3,$F8,$B1,$EA,$3F,$81,$1C,$B0,$FE,$38,$EC,$FE,$10,$6E,$3F
	.dc.b	$81,$3B,$3F,$8E,$3E,$C7,$15,$FC,$45,$E5,$1F,$C1,$5E,$55,$FC,$15
	.dc.b	$DD,$FE,$A9,$F4,$C7,$C4,$8F,$81,$1C,$2F,$F8,$33,$C0,$FD,$5D,$78
	.dc.b	$5F,$B3,$7F,$9C,$7C,$37,$FC,$3D,$E4,$7F,$04,$9F,$E0,$EE,$27,$E6
	.dc.b	$3C,$AF,$F3,$8F,$2B,$FD,$2D,$FE,$32,$BF,$93,$BE,$79,$FE,$6E,$7E
	.dc.b	$EE,$7F,$B2,$DF,$68,$FA,$6F,$9C,$C1,$FA,$03,$13,$FC,$12,$BF,$2B
	.dc.b	$3F,$E5,$D9,$DC,$F8,$6F,$F8,$83,$E1,$96,$7F,$EF,$0A,$FF,$68,$CF
	.dc.b	$F0,$6B,$FE,$25,$5F,$C1,$A7,$BE,$1F,$C3,$23,$CB,$5B,$C3,$F2,$D3
	.dc.b	$FC,$CA,$B9,$BF,$C2,$AB,$E3,$1F,$BB,$3E,$31,$FC,$1C,$E6,$EE,$3D
	.dc.b	$2F,$E8,$EB,$B4,$FD,$B3,$C8,$E1,$EA,$7E,$F8,$EB,$7F,$C3,$5E,$A7
	.dc.b	$F1,$33,$75,$FB,$37,$77,$FC,$36,$BE,$44,$7E,$3A,$FF,$A6,$E1,$F1
	.dc.b	$F1,$FD,$86,$1F,$A9,$3E,$A5,$F8,$DF,$4D,$FF,$05,$7E,$4B,$FC,$B8
	.dc.b	$F1,$E3,$FC,$7D,$F1,$CF,$ED,$5F,$19,$FC,$5B,$1F,$CE,$BC,$1F,$E3
	.dc.b	$AF,$0C,$7F,$67,$7D,$AF,$F9,$1C,$EC,$FE,$24,$6C,$FD,$9D,$F9,$CF
	.dc.b	$C5,$5F,$F0,$83,$FB,$33,$88,$FD,$04,$62,$FF,$05,$7F,$01,$FB,$15
	.dc.b	$BA,$F0,$1F,$C2,$0F,$05,$FE,$FE,$BC,$D7,$FC,$9E,$77,$1F,$C2,$1D
	.dc.b	$C7,$EF,$DF,$86,$7F,$1E,$3F,$53,$3F,$A0,$9F,$EE,$4F,$F2,$58,$FF
	.dc.b	$0F,$1F,$87,$8C,$CF,$F7,$5C,$3F,$C4,$75,$FC,$AF,$EE,$5F,$A4,$3D
	.dc.b	$8F,$F1,$3F,$DE,$3E,$4F,$FD,$99,$5A,$9C,$C6,$2B,$86,$7F,$ED,$1B
	.dc.b	$E0,$AF,$AF,$B5,$C2,$3C,$9A,$FE,$CF,$56,$9E,$4D,$7A,$7B,$5B,$C7
	.dc.b	$CA,$DC,$CD,$F6,$BA,$65,$F2,$0F,$F1,$85,$7B,$3D,$AE,$1E,$D4,$9F
	.dc.b	$36,$FE,$99,$E3,$A7,$E6,$38,$7F,$BF,$E1,$F1,$3D,$AF,$01,$30,$E2
	.dc.b	$7B,$74,$7B,$5C,$B8,$3F,$05,$8F,$F8,$17,$0F,$9B,$ED,$7A,$7D,$A9
	.dc.b	$8F,$29,$CB,$D0,$FF,$7A,$BF,$53,$DA,$F8,$04,$8E,$19,$3D,$AF,$89
	.dc.b	$ED,$7C,$5F,$6A,$56,$A3,$C9,$D5,$FE,$24,$C7,$EF,$FB,$5F,$2C,$95
	.dc.b	$D3,$DA,$F9,$DE,$D4,$8E,$EF,$D1,$F6,$B9,$C7,$F8,$A2,$F9,$BE,$4B
	.dc.b	$E6,$BB,$7F,$07,$3F,$8E,$C3,$ED,$7B,$52,$BD,$6F,$DF,$FD,$27,$F1
	.dc.b	$31,$C7,$A7,$FC,$99,$CB,$FB,$3C,$3B,$CE,$4F,$A5,$97,$83,$2E,$FC
	.dc.b	$2C,$FA,$7C,$7C,$BF,$A1,$9A,$3C,$1F,$19,$EC,$EA,$FF,$15,$E6,$A7
	.dc.b	$C5,$79,$75,$98,$F2,$8F,$47,$8B,$23,$AB,$EB,$F7,$7A,$78,$5F,$E2
	.dc.b	$5C,$7A,$5F,$2B,$F5,$47,$7C,$3E,$14,$FC,$3F,$76,$57,$A1,$F8,$DE
	.dc.b	$EF,$48,$FF,$0B,$DF,$47,$C8,$43,$AE,$6F,$77,$E7,$7B,$BF,$3F,$DD
	.dc.b	$95,$F1,$23,$A3,$FE,$1C,$BF,$85,$EE,$FD,$5F,$76,$4F,$CA,$8F,$B1
	.dc.b	$EE,$FD,$9F,$76,$4F,$CE,$8E,$7E,$2F,$F8,$77,$0E,$73,$C4,$87,$9B
	.dc.b	$DD,$FC,$17,$BB,$F8,$3F,$3A,$4F,$D8,$8E,$7C,$7F,$E1,$7B,$E7,$3E
	.dc.b	$EC,$FF,$14,$7E,$2F,$0F,$C6,$7B,$B2,$7F,$02,$FE,$3F,$DD,$E7,$E2
	.dc.b	$7F,$79,$C7,$FC,$0D,$E9,$BD,$4F,$E2,$98,$7F,$1B,$F1,$FF,$DF,$CF
	.dc.b	$D8,$6F,$85,$F9,$E6,$F8,$9F,$CA,$17,$C4,$FF,$1B,$5E,$ED,$25,$FF
	.dc.b	$03,$8F,$E7,$57,$FA,$2E,$3F,$F1,$8F,$0A,$7A,$D7,$8F,$5E,$87,$5C
	.dc.b	$7B,$7B,$BD,$FD,$D9,$5D,$63,$DE,$FF,$A5,$F0,$E9,$7C,$A3,$C4,$F3
	.dc.b	$7C,$9E,$EF,$A7,$DD,$91,$83,$F8,$C6,$7E,$6C,$DE,$EF,$BF,$DD,$95
	.dc.b	$ED,$8F,$85,$F1,$3F,$0F,$E2,$64,$FC,$18,$F2,$F9,$3F,$EA,$5D,$7E
	.dc.b	$05,$F2,$AF,$A1,$87,$CB,$8F,$99,$F1,$3F,$37,$96,$57,$CB,$8E,$7E
	.dc.b	$47,$F0,$AE,$0F,$F0,$DD,$7D,$CC,$7A,$DF,$94,$87,$DD,$F8,$9F,$B1
	.dc.b	$F1,$3F,$67,$E2,$64,$FD,$78,$EA,$FF,$85,$CF,$EE,$B8,$FE,$07,$E2
	.dc.b	$7F,$01,$F1,$32,$7E,$F4,$7E,$0F,$E2,$64,$7E,$05,$FC,$37,$C4,$F4
	.dc.b	$8F,$F0,$01,$FE,$58,$D3,$D1,$C8,$BE,$93,$F9,$3C,$3F,$1D,$1F,$DA
	.dc.b	$7F,$4A,$4D,$21,$FD,$45,$7C,$BF,$7E,$7F,$F7,$AF,$07,$E3,$EF,$AF
	.dc.b	$6F,$E7,$71,$EB,$8F,$1F,$92,$ED,$F1,$32,$30,$7E,$13,$C3,$FE,$71
	.dc.b	$A3,$E2,$7C,$84,$9F,$17,$C4,$F9,$FE,$27,$D1,$F1,$12,$7C,$D1,$E5
	.dc.b	$C3,$F6,$BC,$5F,$67,$1E,$95,$EB,$C7,$A3,$FC,$B2,$3F,$C0,$EF,$BF
	.dc.b	$E2,$7E,$27,$C4,$C8,$C1,$E7,$EB,$FE,$CD,$5D,$23,$95,$F4,$E0,$FE
	.dc.b	$B7,$3D,$27,$B7,$C4,$FC,$F2,$3F,$37,$E2,$7D,$18,$7E,$42,$3E,$9F
	.dc.b	$83,$FE,$31,$B7,$C4,$FD,$7F,$89,$95,$F5,$63,$ED,$7C,$4F,$DB,$EB
	.dc.b	$93,$F6,$63,$AB,$FE,$F7,$EE,$FF,$1B,$E1,$D2,$3E,$F6,$1D,$1F,$89
	.dc.b	$87,$E1,$7E,$26,$47,$E0,$DF,$C4,$7C,$37,$4C,$FF,$C9,$9E,$87,$F2
	.dc.b	$3A,$F4,$E5,$DB,$0E,$9C,$1F,$13,$2F,$C8,$47,$E4,$7E,$17,$F2,$5F
	.dc.b	$13,$30,$FC,$84,$75,$C3,$FD,$B3,$DE,$D9,$F1,$3F,$99,$F8,$99,$87
	.dc.b	$E5,$E3,$FC,$89,$FA,$5F,$F8,$1B,$5F,$F0,$6B,$7D,$5F,$F2,$BB,$6B
	.dc.b	$FC,$A4,$FF,$7F,$7E,$F1,$FE,$19,$C3,$BB,$FA,$3C,$3B,$BF,$89,$C3
	.dc.b	$BF,$C4,$78,$3E,$1A,$4F,$F8,$E7,$04,$FA,$E3,$F8,$09,$FE,$61,$C3
	.dc.b	$AB,$F9,$3C,$3A,$E5,$F8,$AC,$7D,$7D,$D2,$BA,$3E,$EF,$88,$E9,$1F
	.dc.b	$C0,$4F,$A2,$7B,$3E,$23,$E1,$FC,$44,$8F,$24,$7C,$6F,$88,$91,$E7
	.dc.b	$7E,$4F,$C3,$F4,$E7,$FB,$92,$E8,$FF,$42,$70,$30,$E9,$3F,$E1,$68
	.dc.b	$63,$F4,$A7,$E9,$FC,$37,$D4,$F8,$79,$7F,$16,$3B,$78,$BF,$CB,$F0
	.dc.b	$E9,$3F,$27,$0E,$BC,$1F,$6B,$1F,$BB,$F0,$5F,$7B,$E1,$E5,$FC,$F8
	.dc.b	$E6,$FC,$2E,$86,$2F,$86,$FC,$2F,$C3,$CA,$FA,$D1,$F8,$8F,$85,$FC
	.dc.b	$4F,$C3,$49,$FB,$51,$D1,$F8,$5D,$BF,$8A,$AE,$51,$F0,$6F,$F2,$31
	.dc.b	$F9,$2F,$87,$93,$F8,$37,$F2,$BF,$0F,$CF,$93,$E5,$D7,$0A,$F1,$9F
	.dc.b	$89,$F8,$7F,$F4,$F7,$E9,$9F,$DF,$36,$34,$9F,$0F,$CF,$4F,$F5,$9C
	.dc.b	$7F,$82,$AF,$FD,$14,$D7,$FC,$96,$DE,$3F,$F3,$8B,$FE,$0C,$F0,$B8
	.dc.b	$FC,$37,$80,$95,$CB,$E1,$FB,$3F,$EB,$9A,$E0,$DF,$88,$F3,$9E,$8C
	.dc.b	$7D,$3F,$11,$27,$C6,$FB,$3E,$23,$9E,$BE,$CA,$E3,$5E,$8B,$E4,$FA
	.dc.b	$BE,$17,$E1,$FC,$34,$8C,$1F,$B8,$E2,$FF,$3A,$E6,$F8,$7F,$93,$F1
	.dc.b	$12,$BE,$1C,$7C,$CF,$88,$91,$F1,$9F,$9F,$E0,$79,$2F,$94,$FC,$AB
	.dc.b	$E4,$FC,$AB,$FA,$BF,$11,$F5,$BE,$1E,$4F,$D0,$8E,$53,$C3,$97,$E7
	.dc.b	$DF,$DC,$8F,$BB,$F1,$12,$FE,$BB,$F8,$0F,$84,$F1,$7C,$27,$E0,$BE
	.dc.b	$22,$57,$DC,$8F,$C2,$FC,$44,$8F,$BE,$FF,$B9,$74,$FE,$FB,$1F,$17
	.dc.b	$93,$FD,$1D,$B2,$FF,$1A,$49,$FF,$68,$57,$F8,$01,$BF,$0D,$CC,$8E
	.dc.b	$BE,$1B,$CA,$F8,$C8,$7F,$A6,$27,$93,$7E,$07,$C9,$F0,$D2,$3B,$C7
	.dc.b	$E0,$A5,$F0,$BE,$9F,$86,$95,$E4,$8F,$67,$C3,$7B,$7E,$1A,$4F,$A2
	.dc.b	$39,$4F,$08,$F1,$57,$C3,$8F,$89,$F0,$D2,$7D,$EF,$C7,$F8,$6E,$31
	.dc.b	$F0,$AB,$97,$3F,$FB,$5B,$2F,$2C,$FC,$6F,$80,$F9,$DF,$03,$23,$07
	.dc.b	$CF,$3F,$D6,$33,$FE,$F3,$7E,$6F,$C2,$FD,$53,$EB,$7C,$2F,$D7,$F8
	.dc.b	$59,$3F,$49,$FB,$5F,$0B,$F1,$67,$84,$F2,$3E,$BB,$F7,$FE,$13,$F0
	.dc.b	$1F,$0B,$27,$ED,$C7,$17,$81,$F5,$70,$F0,$BF,$80,$F8,$1F,$C4,$7C
	.dc.b	$14,$8C,$1E,$B3,$FD,$81,$FA,$6F,$F6,$9F,$28,$D2,$7C,$5F,$E4,$38
	.dc.b	$7F,$DF,$B1,$FD,$8A,$BF,$C9,$F1,$8F,$86,$D7,$A3,$FE,$52,$C3,$9C
	.dc.b	$62,$91,$DB,$E1,$FF,$83,$E1,$DE,$F5,$3E,$26,$C7,$F7,$BA,$FF,$74
	.dc.b	$AF,$86,$CF,$C3,$49,$F1,$D7,$86,$B5,$1E,$FF,$89,$FF,$6A,$47,$78
	.dc.b	$D9,$E4,$7D,$EE,$30,$F8,$7F,$03,$E2,$24,$66,$7F,$B2,$BD,$71,$FF
	.dc.b	$6B,$43,$E1,$FD,$3F,$0F,$2B,$84,$7B,$3E,$17,$DB,$F0,$D2,$7A,$47
	.dc.b	$C4,$9E,$11,$F0,$2B,$E1,$C7,$C4,$F8,$4F,$8B,$F0,$B2,$BC,$91,$C5
	.dc.b	$F3,$57,$2E,$2F,$B3,$7F,$33,$DF,$F9,$BF,$05,$2F,$D9,$1F,$43,$E0
	.dc.b	$FE,$8F,$C2,$49,$F8,$11,$F5,$3E,$13,$EA,$FC,$24,$9F,$89,$1F,$63
	.dc.b	$E1,$24,$7C,$77,$EE,$FE,$9D,$FD,$84,$FF,$73,$C7,$CF,$7F,$AD,$73
	.dc.b	$FE,$25,$B1,$F8,$87,$6B,$54,$EF,$D3,$EA,$49,$E9,$F0,$9E,$0F,$84
	.dc.b	$91,$DD,$F1,$FC,$27,$93,$E1,$3C,$BF,$07,$2B,$C9,$1E,$8F,$83,$F4
	.dc.b	$FB,$D2,$7D,$11,$EC,$F7,$FD,$BF,$05,$27,$D9,$1F,$03,$E0,$BE,$0F
	.dc.b	$C1,$49,$F8,$11,$F1,$3D,$0F,$8B,$F0,$52,$7E,$24,$7C,$8F,$82,$F9
	.dc.b	$3E,$FC,$9F,$91,$1D,$5F,$95,$87,$57,$E3,$71,$FC,$BC,$7B,$47,$CF
	.dc.b	$C7,$B3,$A3,$E7,$63,$F5,$BE,$0B,$EB,$FC,$14,$9F,$AD,$1D,$DE,$1E
	.dc.b	$43,$A1,$9F,$87,$A3,$EA,$D7,$D9,$CF,$C2,$F0,$A6,$7E,$2A,$DE,$1C
	.dc.b	$F3,$FE,$07,$FA,$7F,$F6,$0C,$3B,$B6,$B6,$DD,$7F,$88,$3F,$D5,$A3
	.dc.b	$3F,$05,$C8,$95,$FB,$D9,$7C,$17,$6F,$81,$EF,$F0,$52,$B5,$1E,$2F
	.dc.b	$80,$91,$C5,$F2,$FB,$DD,$23,$CB,$8D,$CB,$CB,$8F,$59,$F0,$63,$D5
	.dc.b	$F5,$E3,$EE,$F8,$19,$1E,$57,$E0,$FC,$17,$C2,$F8,$2F,$87,$F0,$52
	.dc.b	$BD,$51,$F1,$BE,$0B,$E3,$FC,$14,$9F,$74,$7C,$AF,$80,$91,$F0,$5F
	.dc.b	$9B,$EF,$75,$8F,$89,$87,$57,$E7,$E1,$D5,$F9,$18,$75,$7E,$9E,$1F
	.dc.b	$5B,$E0,$BE,$BF,$C1,$49,$F9,$D1,$F6,$BE,$06,$47,$D1,$7E,$EF,$C1
	.dc.b	$7D,$EF,$82,$FB,$FF,$05,$2B,$EB,$47,$F4,$4F,$53,$FA,$B1,$FD,$0F
	.dc.b	$97,$F0,$ED,$EA,$FE,$A0,$FF,$5B,$EA,$65,$E0,$C3,$F1,$2F,$81,$EA
	.dc.b	$7E,$0F,$1F,$03,$F8,$EC,$7C,$1F,$03,$23,$FA,$26,$1F,$D8,$5F,$D0
	.dc.b	$AF,$81,$ED,$9F,$E5,$31,$B9,$FE,$53,$4F,$76,$1D,$C8,$77,$3A,$FB
	.dc.b	$DE,$F2,$3E,$0F,$81,$F8,$5F,$01,$23,$C6,$FC,$5F,$88,$F8,$DE,$DF
	.dc.b	$C7,$F7,$25,$7A,$23,$E5,$7B,$F2,$3D,$6F,$CD,$F8,$0E,$91,$EC,$C2
	.dc.b	$ED,$F3,$B1,$EB,$3F,$0B,$1E,$AF,$C2,$C7,$EB,$7B,$F2,$3E,$3B,$F6
	.dc.b	$7E,$03,$ED,$7C,$07,$DB,$E3,$95,$F3,$23,$EF,$7B,$F2,$3E,$7B,$F8
	.dc.b	$1F,$80,$FC,$17,$BF,$27,$E9,$BF,$85,$F8,$0E,$B1,$F5,$B0,$EA,$FD
	.dc.b	$6C,$3A,$3F,$63,$0B,$B7,$E2,$B1,$FC,$77,$C0,$4A,$FB,$EF,$F8,$53
	.dc.b	$F5,$0F,$F5,$C1,$FD,$4A,$BF,$8B,$B7,$9F,$FC,$40,$DA,$FF,$0F,$BF
	.dc.b	$D6,$BC,$F3,$FE,$AE,$6F,$C0,$75,$3B,$7C,$0F,$7F,$81,$97,$FE,$13
	.dc.b	$B4,$78,$FE,$07,$C9,$F1,$12,$B4,$F9,$FE,$23,$9C,$71,$C7,$B3,$CF
	.dc.b	$1F,$67,$C1,$7B,$7E,$0E,$4F,$78,$E8,$FF,$3E,$CF,$D9,$A7,$AB,$2E
	.dc.b	$D3,$CC,$EC,$43,$86,$1F,$27,$E0,$BE,$57,$C1,$C9,$F6,$47,$CD,$F8
	.dc.b	$2F,$9D,$F0,$72,$7E,$04,$7D,$1F,$7B,$E9,$7B,$F2,$7E,$24,$7D,$5F
	.dc.b	$77,$EB,$7B,$D2,$7E,$44,$75,$7E,$56,$1D,$5F,$66,$1D,$5F,$9D,$87
	.dc.b	$57,$E0,$61,$F8,$1F,$82,$FC,$17,$C1,$C9,$FA,$91,$F8,$5F,$82,$FC
	.dc.b	$37,$C1,$C9,$FB,$11,$FE,$5C,$FD,$47,$FA,$B4,$7F,$98,$9F,$EA,$DD
	.dc.b	$5F,$E6,$76,$C6,$4E,$04,$3F,$CA,$8F,$2E,$32,$F8,$1F,$14,$FE,$1F
	.dc.b	$0F,$13,$F6,$70,$E4,$FF,$78,$AE,$44,$F8,$1F,$37,$C0,$49,$EE,$FA
	.dc.b	$7E,$07,$D5,$F0,$7E,$BF,$83,$95,$E4,$8F,$77,$C1,$FB,$FE,$0E,$4F
	.dc.b	$A2,$3E,$17,$C0,$7C,$3F,$81,$93,$EC,$8E,$AF,$BB,$0E,$AF,$9F,$0E
	.dc.b	$AF,$C2,$C3,$AB,$EB,$C3,$E7,$7C,$04,$8F,$8E,$FD,$1F,$81,$EB,$1F
	.dc.b	$07,$0E,$BE,$3F,$F8,$43,$7F,$F6,$1C,$F3,$3B,$10,$C1,$F2,$7E,$0F
	.dc.b	$EE,$7C,$1C,$8F,$AB,$1F,$7F,$E0,$BF,$01,$F0,$72,$7E,$CC,$7E,$0F
	.dc.b	$DE,$FC,$27,$BD,$27,$EE,$C7,$E1,$FE,$06,$47,$E0,$1F,$C5,$7C,$17
	.dc.b	$58,$FC,$1E,$1D,$5F,$B9,$87,$57,$F0,$F8,$75,$7F,$01,$87,$FA,$BB
	.dc.b	$F5,$2F,$F6,$01,$3B,$71,$F1,$19,$F8,$4D,$7C,$24,$C3,$FD,$49,$7F
	.dc.b	$D8,$78,$7F,$C6,$EF,$FA,$BA,$3F,$C8,$6D,$8C,$FC,$27,$7F,$84,$F0
	.dc.b	$7C,$17,$87,$E0,$A6,$1D,$23,$C0,$F6,$C7,$C0,$F4,$CB,$D1,$84,$8F
	.dc.b	$1B,$D5,$F2,$FC,$0F,$B7,$E0,$A4,$60,$F4,$FC,$27,$C1,$23,$EC,$F8
	.dc.b	$4F,$89,$F0,$9F,$17,$E1,$24,$FC,$08,$F9,$1F,$03,$F2,$7E,$0A,$4F
	.dc.b	$C4,$8E,$6F,$C2,$C2,$F0,$FC,$6C,$7A,$CF,$CA,$C7,$AB,$F2,$31,$FA
	.dc.b	$9F,$05,$F5,$7E,$0A,$4F,$D0,$8E,$7C,$DF,$EE,$DE,$56,$78,$AF,$F7
	.dc.b	$6C,$32,$EF,$1F,$53,$2E,$F9,$FD,$8C,$FF,$05,$F0,$9F,$83,$F8,$49
	.dc.b	$7F,$7A,$3F,$0D,$F0,$9F,$EF,$4E,$C7,$C7,$53,$35,$FE,$E7,$8F,$E8
	.dc.b	$A7,$F5,$F2,$7B,$DC,$7D,$E9,$3B,$7A,$7C,$17,$28,$FF,$45,$D7,$67
	.dc.b	$A5,$E7,$4F,$F4,$2C,$F6,$F8,$3F,$21,$E5,$C7,$CD,$F0,$BE,$7F,$85
	.dc.b	$98,$F9,$63,$D5,$F0,$9E,$BF,$85,$93,$E9,$8F,$77,$C1,$FB,$FE,$12
	.dc.b	$4F,$B6,$3E,$17,$C0,$C8,$F8,$0F,$C5,$F8,$2E,$B1,$F0,$F0,$EA,$FB
	.dc.b	$30,$EA,$FC,$7C,$3A,$BF,$03,$0F,$9D,$F0,$3F,$3F,$E0,$A4,$FC,$D8
	.dc.b	$FA,$5F,$07,$23,$E8,$3F,$57,$E1,$3E,$B7,$C1,$FD,$7F,$84,$95,$F5
	.dc.b	$63,$ED,$7C,$0C,$8F,$B0,$FD,$DF,$82,$E9,$1F,$4B,$0B,$CD,$F4,$31
	.dc.b	$EB,$3F,$7F,$1E,$AF,$D8,$C4,$CA,$FF,$5D,$FE,$A7,$FD,$5C,$FE,$BB
	.dc.b	$F0,$A6,$9F,$D6,$FA,$19,$78,$3C,$4E,$74,$9C,$E1,$F8,$FC,$FC,$6F
	.dc.b	$E2,$33,$F1,$E5,$F8,$6C,$FC,$73,$FE,$D6,$CF,$FB,$33,$E5,$CB,$AC
	.dc.b	$7F,$B0,$A3,$03,$2F,$37,$C2,$72,$C3,$D7,$F0,$B7,$0E,$3F,$0B,$70
	.dc.b	$E1,$8B,$E1,$71,$97,$BD,$B0,$AB,$F0,$78,$F7,$E6,$1E,$2C,$3E,$34
	.dc.b	$F0,$F8,$29,$39,$C1,$1D,$3E,$0F,$AE,$3D,$B2,$97,$B8,$EE,$F1,$CB
	.dc.b	$BB,$F1,$71,$F2,$FC,$27,$9B,$E1,$64,$F7,$8F,$4F,$C2,$FA,$BE,$16
	.dc.b	$4F,$8E,$3D,$BF,$07,$EE,$F8,$49,$3E,$78,$F8,$3D,$12,$3D,$4F,$C4
	.dc.b	$F8,$2E,$B1,$ED,$C3,$AB,$F6,$70,$EA,$FC,$1C,$3A,$BF,$77,$0F,$9B
	.dc.b	$F0,$3F,$3B,$E0,$A4,$FC,$78,$FA,$3F,$09,$F4,$BE,$12,$4F,$CB,$8F
	.dc.b	$AF,$FA,$A3,$A2,$92,$7E,$A4,$7F,$B3,$DF,$C3,$3F,$D8,$E3,$F2,$ED
	.dc.b	$8C,$FC,$07,$4F,$81,$F0,$C7,$D9,$C3,$C2,$FE,$57,$0E,$CB,$B3,$C4
	.dc.b	$F6,$C7,$31,$8E,$B6,$5E,$36,$FC,$0F,$AB,$E1,$26,$1E,$88,$F0,$BF
	.dc.b	$A3,$CB,$C3,$87,$ED,$B2,$EB,$97,$A8,$B5,$EE,$F8,$5F,$8B,$F0,$92
	.dc.b	$3E,$1B,$F2,$3E,$17,$E4,$FC,$2C,$9F,$8E,$FC,$CF,$85,$F9,$BF,$03
	.dc.b	$27,$E5,$BF,$43,$E0,$3E,$8F,$BF,$F4,$BE,$02,$57,$D0,$8E,$8F,$C5
	.dc.b	$D3,$E7,$E3,$8D,$BE,$27,$1B,$3E,$D7,$B3,$EB,$E7,$8A,$DE,$7F,$80
	.dc.b	$F8,$4F,$C0,$FC,$2C,$BF,$BF,$1F,$84,$F8,$4F,$C2,$FC,$2C,$9F,$C1
	.dc.b	$C7,$E2,$3E,$0E,$47,$E1,$9F,$C5,$FC,$27,$E3,$3E,$12,$4F,$E2,$9F
	.dc.b	$E0,$BF,$AA,$FF,$5F,$8F,$37,$5F,$DC,$C3,$B4,$63,$81,$97,$8E,$3B
	.dc.b	$10,$D6,$5E,$07,$F8,$56,$1F,$EB,$83,$FD,$A4,$4F,$81,$F0,$FC,$2F
	.dc.b	$8B,$E1,$65,$70,$8F,$2F,$C1,$79,$BE,$0E,$4F,$48,$F4,$FC,$04,$8E
	.dc.b	$EF,$B3,$E0,$7A,$C7,$8B,$0E,$AF,$BB,$2F,$F7,$11,$FE,$DC,$7C,$F8
	.dc.b	$76,$75,$87,$C6,$F8,$0F,$8F,$F0,$32,$7D,$B1,$D9,$E9,$87,$6E,$7F
	.dc.b	$D3,$63,$DB,$0E,$91,$8C,$7D,$B7,$F1,$BE,$0F,$E9,$FC,$1C,$8C,$1F
	.dc.b	$2B,$E0,$BE,$BF,$C1,$C8,$C1,$F3,$BE,$07,$ED,$FC,$0C,$8C,$1F,$4B
	.dc.b	$E0,$3E,$F9,$1F,$AB,$F0,$3F,$82,$C3,$A4,$7C,$5C,$30,$F2,$3E,$45
	.dc.b	$F5,$7E,$56,$1F,$8A,$F8,$3F,$C5,$FC,$1C,$9F,$C0,$C7,$F5,$2A,$FF
	.dc.b	$64,$12,$31,$F2,$59,$F8,$39,$1F,$D4,$67,$FB,$3B,$F9,$47,$FB,$33
	.dc.b	$7D,$FE,$9F,$01,$23,$84,$77,$61,$CC,$EE,$75,$AE,$EF,$57,$CD,$F0
	.dc.b	$9E,$7F,$80,$F4,$7C,$0C,$AF,$24,$7A,$FE,$0E,$47,$9D,$F7,$7C,$1F
	.dc.b	$BF,$E0,$BE,$07,$C1,$CA,$F6,$47,$C3,$F8,$0F,$89,$F0,$32,$7E,$04
	.dc.b	$74,$7E,$06,$17,$AF,$F9,$87,$03,$2E,$D5,$F1,$F2,$EC,$FC,$7C,$BE
	.dc.b	$87,$C0,$48,$F9,$8F,$D3,$F8,$1E,$9E,$A7,$F8,$7B,$1F,$EF,$75,$EB
	.dc.b	$3E,$96,$3D,$9F,$A5,$8F,$6C,$BC,$79,$7D,$DF,$83,$FB,$DF,$07,$2F
	.dc.b	$ED,$47,$E0,$7E,$0B,$F0,$5F,$07,$27,$EF,$47,$E1,$7D,$EF,$C3,$7B
	.dc.b	$F2,$7F,$05,$1F,$E1,$AF,$D5,$BF,$A9,$C7,$F1,$67,$FA,$81,$37,$ED
	.dc.b	$1F,$89,$C3,$B3,$F8,$9C,$3B,$3F,$8D,$C3,$B3,$F8,$DC,$3B,$7C,$27
	.dc.b	$7F,$84,$93,$FE,$23,$CB,$FB,$67,$AA,$7F,$D9,$D9,$FF,$91,$9B,$3F
	.dc.b	$E3,$96,$CE,$FE,$13,$D5,$F0,$52,$BA,$3E,$DF,$82,$F7,$7C,$04,$9F
	.dc.b	$03,$F0,$7E,$03,$E1,$7C,$24,$9F,$23,$F1,$70,$E9,$1F,$17,$0B,$CB
	.dc.b	$FD,$DF,$36,$5F,$2F,$E0,$65,$FB,$5F,$9D,$F0,$3F,$3F,$E0,$BE,$87
	.dc.b	$C1,$CA,$F8,$51,$F4,$FE,$0A,$47,$C5,$7E,$B7,$C1,$FD,$7F,$81,$93
	.dc.b	$F2,$5F,$B5,$F0,$1D,$63,$E6,$61,$D5,$FB,$98,$74,$7E,$F6,$17,$5F
	.dc.b	$56,$3F,$83,$F8,$09,$5F,$55,$FC,$37,$C1,$F5,$8F,$C3,$63,$D6,$FE
	.dc.b	$16,$3C,$F9,$BE,$C6,$56,$FE,$C6,$5F,$8F,$F8,$3F,$F3,$4F,$EA,$FF
	.dc.b	$D9,$71,$FF,$10,$37,$0F,$C0,$4E,$BE,$0E,$61,$FE,$69,$A4,$72,$F8
	.dc.b	$7E,$7F,$09,$D3,$E0,$25,$EE,$3B,$FB,$F2,$39,$3E,$2F,$7F,$AC,$7F
	.dc.b	$76,$C7,$AB,$DF,$1E,$CF,$33,$C5,$87,$67,$86,$1E,$CF,$84,$F6,$FC
	.dc.b	$24,$9F,$44,$7C,$0F,$83,$F8,$3F,$09,$27,$D9,$1F,$13,$E0,$E4,$7B
	.dc.b	$DF,$8F,$F0,$7F,$23,$E0,$7E,$4F,$C0,$CA,$F8,$91,$D5,$F8,$D8,$75
	.dc.b	$7D,$98,$74,$7D,$D8,$5E,$4F,$A3,$8F,$D4,$F8,$1F,$AB,$E9,$CB,$FA
	.dc.b	$11,$F6,$3E,$0E,$47,$D3,$7E,$DF,$C1,$FD,$CF,$82,$FB,$BF,$07,$2B
	.dc.b	$EC,$47,$FA,$A3,$F5,$8F,$F5,$CC,$7F,$89,$1B,$9F,$F9,$85,$BD,$7F
	.dc.b	$EC,$67,$FD,$9B,$CE,$CB,$C3,$7F,$5D,$F0,$BD,$8F,$C3,$63,$E1,$7E
	.dc.b	$E6,$3E,$1F,$81,$F1,$7C,$0C,$9F,$F5,$A6,$3F,$D9,$33,$FC,$9E,$3D
	.dc.b	$9B,$3F,$91,$D3,$DC,$F6,$21,$D8,$E7,$F0,$5E,$FF,$83,$91,$83,$BF
	.dc.b	$C1,$7C,$32,$3E,$2F,$83,$F8,$DF,$03,$23,$CA,$FC,$9F,$7F,$E5,$7B
	.dc.b	$DF,$2F,$DF,$95,$EA,$8E,$6F,$CD,$C2,$E9,$F3,$71,$E9,$3F,$47,$1B
	.dc.b	$3F,$47,$1F,$AD,$F0,$5F,$5F,$E0,$E5,$FC,$68,$FB,$5F,$07,$F6,$FE
	.dc.b	$0E,$4F,$CA,$8F,$BD,$F0,$52,$3E,$6B,$F8,$1F,$83,$FC,$17,$C1,$49
	.dc.b	$FA,$2F,$E1,$7E,$0F,$AC,$7D,$4C,$3A,$BF,$87,$C3,$A3,$F8,$9C,$2E
	.dc.b	$9F,$89,$C7,$F1,$DF,$05,$2B,$EE,$BF,$91,$F8,$3F,$F7,$CF,$AF,$FD
	.dc.b	$82,$FF,$D6,$0D,$EF,$5B,$E0,$53,$98,$C3,$FE,$B5,$3F,$AC,$93,$E0
	.dc.b	$FA,$FC,$1F,$6F,$81,$EF,$F0,$32,$FA,$E5,$E0,$9F,$1E,$52,$72,$F8
	.dc.b	$B2,$F0,$BE,$8F,$87,$F4,$E5,$2B,$67,$91,$F0,$B8,$AF,$4F,$C1,$7C
	.dc.b	$0F,$81,$93,$9D,$53,$E1,$8F,$F4,$BC,$7B,$EB,$C3,$1B,$C3,$C0,$EB
	.dc.b	$0B,$9F,$CA,$D7,$E6,$FC,$27,$CE,$F8,$49,$7D,$E3,$E8,$FC,$1F,$D2
	.dc.b	$F8,$39,$3E,$38,$FA,$BE,$FC,$8F,$33,$F6,$3E,$03,$EC,$FC,$07,$DA
	.dc.b	$F8,$09,$5E,$B8,$EA,$FB,$70,$EA,$FB,$30,$E8,$FB,$B0,$BA,$7E,$0B
	.dc.b	$1F,$C2,$F0,$7E,$1B,$E0,$E5,$FC,$78,$FC,$4F,$C1,$FE,$2B,$E0,$E4
	.dc.b	$FC,$B8,$FC,$6F,$C1,$FE,$2A,$3F,$D8,$3F,$AC,$FF,$5D,$7F,$18,$DE
	.dc.b	$5F,$C6,$CE,$F3,$74,$7E,$06,$BF,$A9,$DF,$2F,$81,$D4,$E5,$B1,$D3
	.dc.b	$E1,$3C,$AF,$DD,$CF,$CA,$FD,$CC,$FC,$AF,$E1,$70,$F1,$1E,$3D,$19
	.dc.b	$F4,$D1,$2F,$81,$C3,$6F,$94,$63,$E0,$B3,$F0,$9A,$F8,$49,$5D,$E3
	.dc.b	$8F,$C1,$F2,$F8,$49,$3E,$38,$EB,$F0,$3D,$BE,$0A,$4F,$9E,$3A,$BE
	.dc.b	$9C,$3A,$BF,$3B,$0E,$AF,$B7,$0E,$AF,$D2,$C3,$D3,$F0,$52,$3E,$13
	.dc.b	$EC,$F8,$2E,$B1,$F6,$30,$EB,$CD,$F5,$23,$EC,$E3,$D9,$FB,$B8,$F6
	.dc.b	$9D,$6F,$F2,$32,$F8,$FF,$07,$2F,$E8,$3F,$2B,$E1,$3E,$5F,$C1,$C9
	.dc.b	$FA,$8F,$CE,$F8,$4F,$9F,$EF,$7D,$0F,$7A,$57,$D9,$8F,$A7,$F0,$5F
	.dc.b	$53,$E0,$E4,$FD,$D8,$E8,$FE,$4B,$0B,$AF,$9B,$9D,$97,$6A,$FC,$26
	.dc.b	$5D,$9F,$CC,$E5,$F8,$3F,$D6,$BF,$BA,$C7,$F2,$56,$EE,$68,$F8,$7E
	.dc.b	$1F,$0B,$C7,$E1,$65,$FF,$15,$AF,$EE,$1E,$17,$F1,$F5,$F0,$9D,$FE
	.dc.b	$12,$56,$E3,$C5,$F0,$5E,$3F,$82,$93,$CE,$3B,$BD,$71,$EE,$B8,$7E
	.dc.b	$AE,$3E,$08,$F5,$61,$D9,$D6,$5E,$3C,$BE,$04,$FC,$1F,$82,$95,$E9
	.dc.b	$7E,$27,$C1,$7C,$5F,$83,$93,$ED,$7E,$47,$C2,$7C,$9F,$83,$F9,$5F
	.dc.b	$09,$2B,$E1,$47,$CD,$F8,$1F,$9D,$F0,$52,$7E,$34,$74,$7C,$A7,$D1
	.dc.b	$C3,$1B,$FB,$31,$ED,$3F,$37,$1E,$CF,$0E,$5F,$AD,$97,$DB,$F8,$2F
	.dc.b	$B9,$F0,$52,$BE,$A4,$74,$E1,$FA,$99,$D9,$FC,$06,$5D,$9E,$9C,$4C
	.dc.b	$FB,$E9,$F3,$B3,$FC,$47,$C2,$7E,$27,$E1,$26,$1F,$81,$8F,$C6,$7C
	.dc.b	$24,$8F,$C2,$3F,$8F,$F8,$4F,$E2,$5F,$AD,$FF,$53,$26,$D8,$F8,$1C
	.dc.b	$FC,$0E,$BE,$06,$63,$FD,$86,$FF,$D9,$BB,$B2,$ED,$3E,$87,$B3,$BC
	.dc.b	$7B,$91,$E5,$8F,$8B,$E1,$3C,$7F,$0B,$27,$B4,$79,$BE,$12,$47,$85
	.dc.b	$F4,$FC,$2F,$AB,$E1,$3D,$7F,$0B,$2B,$CD,$1E,$EF,$82,$91,$E9,$7E
	.dc.b	$0F,$C1,$F4,$8F,$B5,$BF,$F7,$87,$B1,$EF,$C3,$B3,$F0,$70,$EC,$F0
	.dc.b	$C3,$E5,$FC,$1F,$CC,$F8,$39,$3F,$1E,$3E,$7F,$C2,$48,$F9,$4F,$D2
	.dc.b	$F8,$5F,$A7,$F0,$12,$7E,$73,$F5,$BE,$17,$EB,$FC,$14,$9F,$A4,$FD
	.dc.b	$AF,$83,$E9,$1E,$AC,$2F,$77,$F2,$8D,$D9,$75,$AF,$7E,$56,$B5,$97
	.dc.b	$E1,$3E,$0E,$57,$DF,$7F,$AC,$78,$9F,$DB,$71,$65,$DB,$A3,$FC,$59
	.dc.b	$18,$BF,$C3,$67,$D6,$73,$9D,$9F,$F2,$15,$32,$EB,$F0,$BD,$BE,$16
	.dc.b	$57,$F8,$BF,$3F,$EE,$FA,$FF,$42,$7F,$B3,$47,$EC,$57,$C2,$79,$8F
	.dc.b	$3F,$C2,$FA,$3E,$02,$57,$37,$D7,$F0,$3E,$CF,$80,$93,$DD,$F7,$FC
	.dc.b	$0F,$28,$F1,$61,$D5,$F2,$61,$D1,$F1,$61,$72,$F3,$63,$F2,$3E,$13
	.dc.b	$E4,$FC,$2C,$BF,$64,$7C,$CF,$85,$91,$EF,$7E,$7F,$C2,$FD,$0F,$85
	.dc.b	$93,$F0,$DF,$A7,$F0,$BF,$53,$E0,$FE,$AF,$C1,$4A,$F9,$11,$D1,$F8
	.dc.b	$D8,$5D,$3E,$56,$3D,$27,$E5,$63,$67,$E7,$63,$F8,$0F,$81,$FC,$0F
	.dc.b	$C1,$4B,$FA,$91,$F8,$4F,$84,$FC,$2F,$C2,$C9,$FB,$11,$FE,$1F,$FD
	.dc.b	$73,$FA,$D9,$FD,$66,$3F,$CD,$4D,$7F,$95,$3A,$F8,$4E,$1F,$05,$27
	.dc.b	$71,$DD,$FC,$25,$F7,$7F,$05,$7D,$DF,$07,$C2,$C7,$B9,$E3,$F8,$2F
	.dc.b	$27,$C1,$C8,$F1,$C7,$57,$F1,$97,$D7,$6F,$C8,$61,$D6,$7D,$38,$75
	.dc.b	$7D,$27,$BF,$E1,$23,$EE,$7E,$17,$C2,$FC,$3F,$85,$93,$F0,$9F,$8D
	.dc.b	$F0,$BF,$1F,$E0,$64,$FC,$67,$E5,$7C,$0F,$CB,$F8,$0F,$99,$F0,$32
	.dc.b	$BE,$5C,$74,$7E,$5E,$17,$5F,$F1,$4F,$1B,$2E,$B5,$F4,$32,$B5,$AC
	.dc.b	$BE,$C7,$C2,$7D,$9F,$85,$97,$F6,$23,$EE,$7C,$27,$DD,$F1,$E4,$FD
	.dc.b	$C8,$FC,$07,$C1,$FE,$07,$E1,$24,$FE,$02,$3F,$09,$F0,$1F,$85,$F8
	.dc.b	$49,$3F,$84,$8E,$AF,$E1,$B0,$EA,$FE,$1B,$0E,$AF,$E2,$B0,$EA,$FE
	.dc.b	$2B,$0F,$F3,$77,$EB,$BF,$EA,$F2,$7A,$58,$F8,$2C,$FC,$2E,$BE,$16
	.dc.b	$61,$FE,$BD,$B4,$F2,$F8,$3E,$7F,$09,$2B,$51,$DB,$E1,$24,$71,$7C
	.dc.b	$3F,$05,$DA,$3A,$63,$D9,$FF,$04,$63,$D9,$E6,$78,$70,$EC,$EB,$0F
	.dc.b	$5F,$C0,$C8,$F3,$3E,$EF,$82,$ED,$1D,$30,$ED,$CF,$FE,$68,$C7,$B4
	.dc.b	$F3,$3B,$10,$C1,$E4,$F8,$4F,$92,$47,$E2,$7C,$27,$CC,$F8,$3F,$9B
	.dc.b	$F0,$92,$7E,$44,$7D,$0F,$81,$91,$F2,$DF,$A7,$F0,$3F,$53,$E0,$7E
	.dc.b	$AF,$C0,$4A,$FA,$11,$D5,$FA,$58,$75,$7E,$3E,$1D,$1F,$93,$85,$D3
	.dc.b	$DD,$8F,$E0,$3E,$12,$57,$DB,$7F,$07,$F0,$9F,$84,$F8,$49,$3F,$7D
	.dc.b	$FF,$5A,$7E,$BD,$FE,$C9,$3F,$A7,$F5,$25,$F0,$E9,$7C,$27,$03,$8F
	.dc.b	$C2,$F2,$F8,$29,$7C,$75,$FF,$4C,$DF,$F5,$86,$FC,$1F,$8A,$7F,$07
	.dc.b	$9D,$9D,$FC,$7F,$49,$E5,$A7,$8F,$4F,$16,$9E,$19,$C9,$E6,$CB,$D2
	.dc.b	$DF,$84,$F6,$91,$DD,$7A,$FE,$13,$E0,$FC,$34,$9C,$DA,$3E,$2F,$C3
	.dc.b	$7C,$6F,$85,$95,$A7,$E4,$FC,$37,$CA,$F8,$4F,$97,$F0,$92,$B9,$C7
	.dc.b	$57,$FC,$D1,$8F,$57,$BE,$3D,$9E,$84,$37,$87,$67,$CB,$F0,$9F,$60
	.dc.b	$8F,$A3,$E1,$3B,$71,$7D,$0B,$BC,$3F,$27,$16,$BE,$AC,$3A,$9D,$EF
	.dc.b	$EB,$EB,$F8,$4F,$85,$FC,$2F,$C3,$4A,$F8,$B1,$F8,$8F,$85,$FC,$4F
	.dc.b	$C3,$49,$F9,$31,$F8,$CF,$81,$91,$F3,$1F,$C7,$FC,$17,$E3,$63,$FD
	.dc.b	$15,$3F,$EC,$03,$FB,$26,$5F,$EE,$06,$EB,$F8,$17,$FD,$8B,$A9,$F0
	.dc.b	$57,$CC,$C7,$C8,$F7,$21,$F4,$31,$F0,$BF,$82,$9F,$F5,$C7,$2F,$3F
	.dc.b	$85,$F2,$65,$E1,$6E,$5E,$1F,$86,$B8,$7E,$93,$0F,$F6,$6B,$CF,$E1
	.dc.b	$73,$8F,$95,$B2,$F8,$6B,$F0,$B8,$F8,$59,$87,$8E,$37,$F0,$7C,$3E
	.dc.b	$0A,$4F,$9E,$39,$E3,$23,$C0,$7A,$F0,$EA,$FB,$70,$EA,$FC,$EC,$3C
	.dc.b	$7F,$05,$E4,$F8,$29,$3F,$0E,$3C,$FF,$09,$E8,$F8,$59,$3F,$1E,$3D
	.dc.b	$7F,$07,$23,$E5,$3E,$EF,$84,$F7,$FC,$0C,$9F,$9C,$FC,$2F,$82,$EB
	.dc.b	$1F,$47,$0E,$AF,$E0,$B0,$E8,$FE,$13,$0B,$CD,$F8,$3C,$7E,$5F,$C0
	.dc.b	$CA,$FB,$4F,$CE,$F8,$2E,$B1,$F8,$CC,$7A,$DF,$E2,$F1,$E7,$CF,$F8
	.dc.b	$BC,$AD,$FD,$BC,$BE,$BF,$C1,$C8,$FC,$33,$F6,$BE,$13,$EE,$FE,$BE
	.dc.b	$C2,$FC,$2C,$BF,$E1,$B3,$96,$FC,$2F,$02,$3A,$F7,$F9,$FC,$14,$8E
	.dc.b	$2F,$6F,$83,$EF,$8F,$59,$FF,$78,$60,$C7,$31,$D8,$86,$B1,$F3,$E3
	.dc.b	$E8,$F8,$59,$3E,$67,$D7,$F0,$BE,$CF,$85,$F6,$FC,$2C,$AF,$5C,$7C
	.dc.b	$0F,$84,$91,$EE,$7E,$1F,$C2,$7C,$4F,$81,$93,$F0,$9F,$8F,$F0,$5D
	.dc.b	$63,$E2,$E1,$D5,$F5,$E1,$D1,$F6,$E1,$79,$FE,$57,$8C,$E0,$67,$F4
	.dc.b	$BE,$0A,$61,$F4,$1F,$AB,$F0,$5F,$5B,$E0,$FE,$BF,$C2,$4A,$FA,$B1
	.dc.b	$F6,$BE,$13,$ED,$FC,$24,$9F,$B3,$1F,$7B,$E0,$7F,$77,$FB,$03,$E0
	.dc.b	$A9,$9A,$E6,$7E,$03,$1E,$CF,$D6,$C7,$B3,$F8,$4C,$7B,$3F,$6B,$1E
	.dc.b	$BF,$0B,$23,$FA,$AE,$3F,$D6,$B7,$3F,$DC,$23,$B9,$F8,$2C,$7B,$F0
	.dc.b	$7C,$6E,$06,$7E,$0A,$FC,$3E,$7E,$0A,$D6,$7E,$CF,$85,$F6,$FC,$2C
	.dc.b	$BF,$0C,$7C,$0F,$84,$91,$E4,$7E,$1F,$C2,$7C,$4F,$7F,$E2,$FB,$F2
	.dc.b	$BD,$31,$F2,$3E,$0B,$E4,$FC,$04,$9F,$6C,$74,$5A,$7D,$EE,$16,$59
	.dc.b	$8C,$65,$DE,$BE,$2E,$5D,$DE,$3E,$1F,$F8,$17,$0F,$ED,$91,$83,$FC
	.dc.b	$32,$BE,$0F,$ED,$FC,$24,$9F,$9F,$1F,$7B,$E1,$24,$7D,$27,$F0,$3F
	.dc.b	$09,$F8,$2F,$83,$FC,$1F,$C2,$4A,$FA,$F1,$F8,$6F,$7A,$47,$DA,$7F
	.dc.b	$13,$F0,$5D,$23,$C9,$D7,$F7,$31,$C7,$07,$C1,$CB,$B5,$7E,$0B,$2E
	.dc.b	$CF,$0D,$59,$FE,$53,$E0,$A4,$FE,$21,$FC,$BF,$C2,$7F,$52,$FD,$87
	.dc.b	$FB,$74,$F8,$1B,$97,$EB,$1B,$C3,$FE,$2D,$7F,$B7,$CF,$F9,$01,$FE
	.dc.b	$CA,$4F,$84,$E9,$F0,$9D,$7E,$0A,$57,$F8,$B2,$BF,$B1,$BD,$FE,$0B
	.dc.b	$C4,$C3,$47,$97,$E2,$7C,$D8,$48,$E4,$F6,$7A,$7C,$17,$AC,$8F,$7F
	.dc.b	$83,$E8,$FB,$4B,$A7,$B6,$FA,$47,$C1,$BB,$3F,$07,$3F,$8D,$F0,$9F
	.dc.b	$1F,$E1,$65,$FA,$E3,$E5,$7C,$27,$CB,$F8,$59,$3E,$F8,$F9,$DF,$01
	.dc.b	$F3,$FE,$06,$4F,$C3,$8F,$A5,$F0,$32,$3E,$33,$F5,$7E,$07,$AC,$7C
	.dc.b	$9C,$3A,$BF,$5F,$0E,$AF,$CD,$C3,$AB,$F6,$F0,$FB,$DF,$0B,$F7,$FE
	.dc.b	$16,$4F,$D3,$8F,$C1,$7C,$2F,$E0,$FE,$16,$4F,$D7,$8F,$C3,$7C,$2C
	.dc.b	$8F,$B4,$FE,$27,$E1,$7F,$C3,$7E,$9F,$F5,$B9,$FC,$CB,$FD,$45,$FF
	.dc.b	$35,$B5,$CF,$5F,$81,$FC,$26,$1E,$07,$F2,$58,$77,$7F,$29,$85,$E4
	.dc.b	$FC,$A6,$3E,$0F,$82,$F0,$FC,$14,$BE,$F1,$E4,$F8,$5F,$2F,$C2,$C9
	.dc.b	$F1,$C7,$A3,$E1,$24,$79,$9F,$5F,$C2,$7B,$3E,$06,$4F,$A9,$F7,$FC
	.dc.b	$17,$58,$F6,$E1,$D5,$F8,$38,$74,$7E,$1E,$17,$4F,$EA,$FC,$2C,$BE
	.dc.b	$4F,$C1,$4B,$F8,$EF,$CC,$F8,$2E,$D1,$F3,$32,$ED,$87,$2C,$BA,$71
	.dc.b	$FF,$A8,$72,$67,$8C,$3F,$D6,$B1,$DD,$FA,$7F,$09,$F6,$08,$FD,$6F
	.dc.b	$84,$FB,$7F,$09,$23,$EC,$BF,$7B,$E1,$3E,$FF,$C0,$49,$FB,$AF,$E0
	.dc.b	$B0,$FC,$1F,$BF,$27,$F0,$2F,$E1,$BE,$03,$A4,$7E,$1B,$0B,$B7,$C4
	.dc.b	$C7,$A4,$FE,$2B,$1B,$3F,$23,$1F,$C7,$FC,$24,$AF,$C6,$BF,$E6,$CF
	.dc.b	$D8,$BF,$B9,$47,$FA,$E9,$FE,$CC,$4F,$84,$D7,$C4,$6F,$E0,$E5,$6D
	.dc.b	$E5,$F0,$9C,$FE,$07,$A7,$C0,$CA,$E9,$1D,$F0,$91,$DD,$EA,$F3,$21
	.dc.b	$FC,$F3,$1F,$F6,$F7,$2F,$FB,$5D,$61,$E9,$F8,$49,$5E,$97,$D9,$F0
	.dc.b	$5E,$DF,$83,$93,$ED,$7E,$07,$C2,$7C,$1F,$83,$93,$F0,$5F,$89,$F0
	.dc.b	$9F,$17,$E0,$64,$FC,$57,$E4,$7C,$0F,$48,$F3,$61,$79,$7F,$C5,$BA
	.dc.b	$B2,$EB,$5E,$BC,$AD,$6B,$2F,$A5,$EF,$FD,$3F,$81,$97,$F4,$E3,$A7
	.dc.b	$07,$C1,$CE,$CB,$E0,$9A,$77,$8F,$51,$0F,$01,$F7,$71,$FC,$07,$C2
	.dc.b	$7E,$07,$E1,$24,$FE,$06,$3F,$09,$F0,$7F,$85,$F8,$39,$3F,$85,$8F
	.dc.b	$F5,$9F,$EC,$7F,$D7,$A3,$C4,$FF,$5E,$C8,$D3,$4E,$3D,$EE,$3E,$F4
	.dc.b	$A1,$FD,$EA,$BF,$D9,$3A,$30,$F0,$CF,$E0,$F1,$F2,$BC,$38,$7F,$CB
	.dc.b	$77,$D8,$F0,$7B,$FE,$69,$F3,$FC,$24,$78,$70,$F8,$35,$F3,$CF,$B3
	.dc.b	$E0,$FD,$BF,$07,$2F,$67,$B9,$F8,$3F,$07,$F0,$BE,$0A,$4E,$B0,$F8
	.dc.b	$31,$F1,$BE,$0F,$E3,$FC,$07,$C8,$F8,$19,$7B,$D9,$3D,$E3,$E1,$65
	.dc.b	$77,$F6,$E7,$DA,$7E,$36,$56,$7E,$0E,$5F,$53,$E0,$3E,$AF,$C0,$CB
	.dc.b	$ED,$1F,$63,$E0,$FE,$CF,$C1,$C9,$F1,$47,$DC,$F8,$2F,$BB,$F0,$72
	.dc.b	$7C,$D1,$F8,$0F,$82,$FC,$0F,$C1,$C9,$F5,$47,$57,$D9,$87,$57,$EE
	.dc.b	$61,$D5,$F8,$18,$75,$7F,$01,$87,$E3,$3E,$0E,$47,$C5,$7F,$1F,$F0
	.dc.b	$7D,$63,$F0,$D8,$75,$DB,$F0,$38,$F5,$9F,$C5,$63,$D6,$7F,$0B,$8F
	.dc.b	$E3,$A7,$FC,$8F,$FB,$27,$F7,$19,$FF,$51,$B7,$A7,$F0,$8D,$9F,$D7
	.dc.b	$3F,$DC,$E7,$7F,$07,$CD,$B3,$F8,$19,$4E,$6B,$9F,$C2,$5A,$E7,$EF
	.dc.b	$5E,$A6,$0F,$7F,$0D,$AE,$95,$8F,$7E,$47,$77,$7F,$05,$C3,$39,$3E
	.dc.b	$37,$9E,$5D,$2F,$AE,$32,$BA,$9E,$9F,$84,$F0,$FC,$24,$8C,$1E,$DF
	.dc.b	$84,$F2,$FC,$24,$8C,$1F,$07,$E1,$3D,$3F,$09,$23,$07,$C5,$F8,$2F
	.dc.b	$69,$1F,$91,$F0,$5F,$03,$0E,$91,$F5,$B0,$C3,$C8,$FB,$37,$D5,$FB
	.dc.b	$78,$7C,$8F,$82,$F9,$3F,$05,$27,$EA,$47,$CC,$F8,$4F,$9B,$F0,$92
	.dc.b	$7E,$C4,$7D,$0F,$84,$91,$F6,$DF,$AD,$E0,$7F,$66,$8F,$F4,$7B,$FE
	.dc.b	$CB,$27,$C0,$6B,$E1,$3B,$4F,$E0,$F0,$EC,$FE,$43,$0E,$AF,$E4,$B0
	.dc.b	$BD,$9F,$EF,$4C,$7B,$37,$E1,$3C,$3F,$05,$2F,$F8,$75,$7F,$64,$7F
	.dc.b	$2B,$87,$6B,$FC,$DE,$3E,$29,$EC,$7A,$C8,$7A,$BE,$23,$DB,$F0,$9E
	.dc.b	$EF,$85,$93,$DA,$3E,$0F,$C2,$FC,$2F,$85,$93,$E2,$8F,$8B,$EF,$FC
	.dc.b	$6F,$83,$93,$E6,$8F,$93,$F0,$72,$3D,$2F,$CC,$F8,$1E,$B1,$EC,$C3
	.dc.b	$AB,$F3,$B0,$EA,$FC,$0C,$3A,$BF,$4B,$0F,$AB,$F0,$B2,$3E,$2B,$F6
	.dc.b	$3E,$17,$EC,$FC,$27,$DA,$F8,$59,$5F,$2A,$3E,$EF,$C2,$7D,$EF,$85
	.dc.b	$93,$F3,$A3,$F0,$3F,$05,$F8,$2F,$82,$93,$F4,$A3,$AB,$F5,$30,$EA
	.dc.b	$FE,$1B,$0E,$AF,$D8,$C3,$AB,$F8,$AC,$3F,$81,$7E,$CB,$FD,$98,$FE
	.dc.b	$D4,$48,$CF,$C3,$6B,$E1,$A4,$7F,$4F,$94,$F2,$F8,$49,$39,$7A,$FC
	.dc.b	$27,$6F,$81,$EF,$F0,$32,$B9,$47,$67,$A6,$3D,$9F,$F7,$6E,$3D,$9E
	.dc.b	$84,$33,$87,$67,$C9,$F0,$3E,$CF,$81,$91,$83,$D1,$9F,$5E,$E7,$B1
	.dc.b	$FD,$CF,$16,$5E,$3B,$EA,$77,$AD,$BF,$E2,$19,$FF,$72,$C6,$3D,$84
	.dc.b	$36,$F9,$BF,$05,$F3,$BE,$12,$61,$F2,$63,$E8,$FC,$17,$D2,$F8,$39
	.dc.b	$3F,$36,$3E,$AF,$BF,$23,$E8,$3F,$63,$E0,$3E,$CF,$BF,$27,$EA,$3F
	.dc.b	$73,$E0,$3A,$C7,$D7,$C3,$AB,$F3,$30,$E8,$FC,$EC,$2E,$DE,$BD,$59
	.dc.b	$7E,$1B,$CA,$97,$F8,$17,$F1,$3F,$07,$F8,$AF,$83,$93,$F8,$57,$F1
	.dc.b	$BF,$07,$FD,$3F,$F6,$6F,$EB,$9C,$66,$93,$E2,$F5,$F0,$5B,$F8,$29
	.dc.b	$8F,$F8,$76,$BF,$AE,$BD,$48,$7E,$1B,$1E,$EE,$FA,$3C,$04,$79,$61
	.dc.b	$E3,$F8,$2F,$27,$C1,$49,$ED,$1E,$7F,$84,$F4,$7C,$24,$9F,$14,$7A
	.dc.b	$FE,$13,$D9,$F0,$92,$7C,$D1,$EF,$F8,$2F,$81,$F0,$52,$7D,$51,$D5
	.dc.b	$F6,$61,$D5,$73,$FE,$2F,$7F,$F6,$2C,$74,$3E,$16,$3D,$DD,$ED,$FE
	.dc.b	$4B,$96,$5F,$3F,$E0,$7E,$87,$C1,$4B,$F9,$51,$E1,$78,$9F,$E7,$BA
	.dc.b	$FF,$68,$38,$E4,$FA,$D9,$EF,$3F,$26,$3E,$17,$CD,$9F,$C3,$BF,$AD
	.dc.b	$F0,$9F,$7C,$8F,$D9,$F8,$4F,$C1,$7C,$1F,$E0,$FE,$12,$4F,$DD,$8F
	.dc.b	$C3,$7B,$FF,$87,$F7,$E4,$FE,$06,$3F,$15,$F0,$52,$3F,$08,$FF,$86
	.dc.b	$FF,$67,$FE,$B3,$C7,$F0,$F0,$EB,$5E,$BE,$16,$5E,$28,$E8,$7E,$37
	.dc.b	$2E,$EE,$F2,$E9,$F0,$B2,$3F,$CD,$D9,$23,$C1,$F0,$BE,$1F,$85,$F1
	.dc.b	$7C,$2C,$BD,$C7,$97,$E0,$E4,$72,$7D,$1F,$09,$E9,$F8,$19,$3D,$9F
	.dc.b	$67,$C1,$75,$8F,$0E,$1D,$5F,$AD,$9F,$F7,$3E,$92,$C7,$BC,$E9,$EE
	.dc.b	$FA,$BE,$07,$E4,$7C,$14,$8C,$1E,$EF,$84,$F9,$84,$7E,$0F,$C2,$7C
	.dc.b	$FF,$84,$91,$F1,$1F,$A5,$F0,$9F,$4F,$E0,$64,$FC,$87,$EB,$7C,$17
	.dc.b	$38,$F4,$E1,$7B,$FF,$95,$6A,$CB,$AD,$7B,$B2,$B5,$AC,$BF,$01,$F0
	.dc.b	$3F,$81,$F8,$29,$7F,$5A,$3A,$76,$FF,$AB,$F2,$67,$8A,$FF,$6D,$47
	.dc.b	$77,$D4,$43,$C0,$7E,$03,$0F,$C7,$7C,$1C,$8F,$C1,$BF,$91,$F8,$4F
	.dc.b	$F3,$66,$5F,$ED,$03,$FD,$A8,$49,$CF,$BF,$1F,$F5,$C4,$7F,$5D,$7F
	.dc.b	$60,$FF,$52,$73,$F0,$5D,$3D,$F9,$18,$38,$9E,$0F,$88,$8E,$0E,$B7
	.dc.b	$DD,$EA,$F9,$BE,$07,$CF,$F0,$92,$7C,$6F,$AB,$E1,$7D,$7F,$0B,$EC
	.dc.b	$F8,$59,$5E,$88,$F7,$FC,$27,$C0,$F8,$59,$3E,$C8,$F8,$7F,$05,$F1
	.dc.b	$3E,$0E,$4F,$C0,$8E,$AF,$C2,$C3,$AB,$F0,$B0,$EA,$FC,$6C,$3A,$BF
	.dc.b	$1B,$0F,$9F,$F0,$72,$3E,$5B,$F4,$BE,$0B,$E9,$FC,$2C,$9F,$9E,$FD
	.dc.b	$6F,$85,$FA,$FF,$09,$F6,$3E,$12,$57,$D4,$8F,$B7,$F0,$32,$3E,$BB
	.dc.b	$F7,$BE,$13,$A4,$7D,$8C,$2F,$81,$FE,$03,$E9,$65,$DA,$BE,$FE,$5D
	.dc.b	$9F,$BF,$97,$FA,$CF,$F6,$8F,$EC,$A4,$F0,$31,$F0,$7D,$15,$3C,$A6
	.dc.b	$4D,$35,$39,$8F,$0B,$EB,$3C,$24,$3C,$27,$4D,$3F,$C3,$58,$FF,$6E
	.dc.b	$8F,$E8,$4B,$E1,$3C,$C7,$9F,$E1,$7D,$1F,$09,$E9,$F8,$59,$7C,$F4
	.dc.b	$F3,$E8,$C5,$7E,$FF,$85,$F8,$1F,$03,$30,$E0,$7B,$6B,$FB,$0B,$E2
	.dc.b	$F8,$1F,$8C,$47,$78,$75,$F8,$1F,$1C,$62,$BF,$DB,$4F,$AB,$3C,$BE
	.dc.b	$CC,$7C,$0F,$BB,$1F,$23,$C7,$E1,$3E,$A7,$C2,$C8,$C1,$D7,$E1,$3E
	.dc.b	$C1,$1F,$07,$C2,$FD,$BF,$85,$91,$E3,$7E,$F7,$C2,$FD,$FF,$84,$FC
	.dc.b	$07,$C1,$CA,$F4,$47,$37,$E8,$E1,$79,$BE,$AE,$3D,$27,$EA,$E3,$67
	.dc.b	$EC,$E3,$F8,$BF,$82,$FC,$67,$C1,$CB,$F8,$91,$F8,$FF,$85,$91,$F1
	.dc.b	$DF,$C6,$BF,$E0,$1F,$DA,$7F,$B5,$BF,$8F,$7F,$B0,$3F,$41,$AE,$FE
	.dc.b	$17,$8E,$3F,$E6,$39,$7C,$2E,$3A,$BF,$07,$1F,$D5,$C7,$C0,$E7,$0E
	.dc.b	$CD,$E9,$4F,$F5,$DE,$86,$7E,$8B,$FC,$AB,$B9,$F3,$D7,$0C,$F0,$4A
	.dc.b	$F4,$65,$BF,$81,$E1,$F0,$92,$7D,$91,$D7,$5F,$D1,$E5,$75,$F4,$FC
	.dc.b	$47,$3B,$F8,$07,$C3,$CB,$C5,$97,$8F,$E1,$64,$FC,$77,$CD,$F0,$BE
	.dc.b	$7F,$85,$F4,$7C,$2C,$AF,$99,$1E,$BF,$80,$91,$F3,$DF,$77,$C1,$FB
	.dc.b	$FE,$0E,$4F,$D3,$7E,$17,$C0,$F5,$8F,$AD,$87,$57,$EE,$61,$D1,$FB
	.dc.b	$D8,$5E,$0F,$B1,$8F,$CB,$F8,$5F,$99,$F0,$B2,$FF,$01,$1F,$3F,$E1
	.dc.b	$64,$7E,$0D,$FA,$5E,$67,$D3,$F8,$4F,$A9,$F0,$92,$BF,$11,$1F,$77
	.dc.b	$F6,$AF,$C7,$C7,$F9,$3D,$B9,$FF,$AA,$DB,$CD,$FC,$29,$7C,$1F,$82
	.dc.b	$BF,$2D,$95,$BF,$C8,$65,$E1,$9F,$CB,$65,$E1,$7F,$39,$97,$87,$E0
	.dc.b	$E4,$7F,$0A,$CB,$FB,$96,$1E,$1F,$83,$F3,$7C,$2C,$AD,$3E,$9F,$85
	.dc.b	$F5,$7C,$2F,$AF,$E1,$65,$73,$8F,$77,$C1,$7B,$FE,$0E,$4F,$78,$E8
	.dc.b	$FC,$1C,$2E,$5F,$07,$1E,$B3,$E5,$C7,$AB,$F1,$F1,$F9,$5F,$05,$F2
	.dc.b	$FE,$0E,$4F,$AE,$39,$F2,$7C,$9C,$AC,$FF,$49,$DB,$FD,$AC,$F9,$CE
	.dc.b	$C4,$3B,$1F,$17,$E1,$3E,$C7,$C2,$C8,$C1,$F2,$7E,$17,$EE,$7C,$2C
	.dc.b	$8C,$1F,$37,$DF,$FC,$01,$1F,$A1,$F0,$7F,$83,$F8,$2F,$C2,$7C,$1C
	.dc.b	$9F,$A9,$1F,$87,$C3,$A4,$7E,$1F,$0C,$3D,$48,$7E,$2F,$0F,$C7,$61
	.dc.b	$F8,$FF,$85,$93,$F7,$DF,$C9,$7C,$2F,$F0,$FF,$4F,$FD,$AA,$4E,$0C
	.dc.b	$7C,$2E,$7E,$13,$5F,$0B,$31,$FE,$A3,$7F,$D9,$FA,$58,$7F,$60,$8F
	.dc.b	$EB,$F3,$AF,$82,$EC,$77,$F8,$5F,$05,$49,$E6,$F8,$1E,$B8,$F7,$7F
	.dc.b	$D6,$97,$FD,$CD,$C9,$E2,$F8,$3F,$57,$C1,$C8,$C1,$E6,$F8,$5F,$71
	.dc.b	$1F,$4F,$C2,$FC,$1F,$85,$91,$EC,$7E,$27,$C2,$FC,$5F,$83,$F8,$DF
	.dc.b	$07,$2B,$E0,$C7,$37,$CB,$85,$E0,$F8,$71,$FE,$17,$D1,$97,$68,$F5
	.dc.b	$E5,$6B,$7E,$EA,$D9,$FD,$5F,$83,$98,$7D,$07,$EC,$7C,$1F,$7E,$1F
	.dc.b	$A1,$A5,$AF,$06,$9D,$67,$EA,$69,$67,$C9,$8F,$E0,$7E,$12,$47,$DC
	.dc.b	$7F,$09,$F0,$BF,$85,$D3,$F0,$DF,$0B,$2B,$F0,$31,$F8,$9F,$80,$91
	.dc.b	$F8,$47,$FA,$D7,$ED,$7F,$D5,$0F,$EC,$FC,$3E,$E6,$E9,$FE,$26,$6D
	.dc.b	$7F,$8F,$5F,$ED,$2F,$63,$F1,$F8,$78,$5F,$C1,$E1,$E0,$7F,$0B,$85
	.dc.b	$C3,$B7,$C0,$F8,$BE,$23,$C7,$F0,$F2,$FF,$C5,$39,$7F,$B2,$4F,$F6
	.dc.b	$43,$8F,$84,$F4,$9E,$AF,$85,$F5,$FC,$1F,$B3,$E0,$E5,$72,$8F,$7F
	.dc.b	$C0,$7C,$0F,$81,$93,$DA,$39,$BD,$48,$35,$F8,$98,$76,$8F,$2E,$1D
	.dc.b	$9D,$70,$FC,$7C,$7E,$6F,$C0,$FC,$EF,$81,$95,$EC,$8F,$A3,$F0,$72
	.dc.b	$3D,$EF,$D4,$F8,$3F,$AB,$F0,$5F,$5B,$E0,$E5,$7C,$48,$FB,$3F,$05
	.dc.b	$F6,$BE,$0A,$4F,$C8,$8E,$8F,$90,$FB,$18,$63,$7F,$BD,$8F,$69,$F9
	.dc.b	$F8,$F6,$78,$78,$4E,$46,$7F,$88,$F8,$1F,$0D,$77,$3E,$BB,$D7,$83
	.dc.b	$F1,$19,$D9,$FC,$46,$5D,$DE,$B8,$32,$F0,$67,$F9,$1C,$BF,$2B,$F0
	.dc.b	$5F,$E2,$FF,$DB,$3F,$B7,$E1,$FE,$8D,$6E,$BE,$59,$D7,$C2,$6F,$E1
	.dc.b	$25,$FF,$9C,$AF,$FB,$AB,$7E,$13,$A7,$BB,$23,$06,$FE,$0B,$C0,$47
	.dc.b	$97,$C0,$F6,$7A,$61,$D9,$F8,$F8,$76,$6E,$1D,$5C,$E1,$EB,$8F,$67
	.dc.b	$C1,$7B,$7E,$0E,$57,$9E,$3E,$07,$C2,$48,$F5,$3F,$0F,$E0,$FE,$27
	.dc.b	$C2,$7C,$5F,$83,$95,$EF,$8F,$91,$F0,$72,$3E,$13,$F2,$FE,$07,$A4
	.dc.b	$7A,$30,$BC,$BF,$C7,$B6,$65,$DA,$BE,$56,$5D,$9F,$7E,$5F,$57,$E0
	.dc.b	$64,$7C,$F7,$EC,$7C,$0F,$D9,$F8,$1F,$B5,$F0,$52,$BE,$A4,$7D,$DF
	.dc.b	$83,$FB,$DF,$07,$27,$EC,$47,$E0,$7E,$0F,$9C,$7D,$BC,$30,$F2,$3E
	.dc.b	$F6,$1D,$5F,$A7,$87,$57,$F0,$58,$75,$7E,$BE,$1F,$E7,$BF,$DB,$7F
	.dc.b	$A1,$C7,$E8,$1B,$BF,$F4,$06,$F8,$5F,$EC,$85,$F0,$9E,$DA,$F0,$90
	.dc.b	$F0,$9E,$A3,$C2,$43,$C2,$7F,$B1,$A3,$FB,$B4,$F5,$F8,$1F,$21,$E5
	.dc.b	$F8,$4F,$37,$C1,$CB,$E0,$FA,$7E,$13,$D5,$F0,$92,$7A,$3E,$DF,$84
	.dc.b	$F7,$7C,$07,$39,$F0,$61,$87,$99,$0F,$85,$87,$57,$99,$0F,$8D,$7D
	.dc.b	$5F,$57,$C0,$7C,$B2,$3E,$DF,$80,$F9,$DF,$01,$F3,$FE,$0E,$4F,$C1
	.dc.b	$8F,$A5,$F0,$9F,$4F,$E1,$24,$FC,$58,$FA,$DF,$09,$F5,$F0,$93,$F2
	.dc.b	$63,$AB,$F2,$F0,$EA,$FD,$BC,$3A,$BF,$3F,$0E,$AF,$DF,$C3,$F0,$5F
	.dc.b	$09,$F8,$3F,$84,$93,$F5,$63,$F0,$DF,$07,$F8,$7F,$84,$93,$F6,$63
	.dc.b	$F1,$5F,$09,$23,$EE,$3F,$EC,$7E,$AF,$F6,$09,$FD,$AB,$D1,$34,$9F
	.dc.b	$0D,$DA,$BF,$09,$87,$67,$F2,$98,$75,$7F,$2D,$85,$EC,$FF,$0A,$E1
	.dc.b	$DB,$E1,$3C,$04,$A1,$FC,$F2,$3F,$DB,$AF,$A4,$EE,$43,$B9,$EA,$21
	.dc.b	$E0,$39,$CF,$83,$E1,$7D,$DF,$09,$27,$3D,$BE,$0D,$7D,$F3,$F0,$FE
	.dc.b	$0B,$E2,$7C,$0F,$C5,$F8,$29,$87,$01,$7F,$1A,$7E,$57,$C1,$C9,$D4
	.dc.b	$FC,$A8,$F9,$DF,$09,$F3,$FE,$0E,$56,$B6,$47,$D3,$F8,$4F,$0C,$E7
	.dc.b	$2F,$0B,$F1,$72,$EC,$FC,$7C,$6E,$FF,$3B,$2F,$B9,$F0,$32,$BB,$3F
	.dc.b	$7F,$E0,$7F,$01,$EF,$C9,$F1,$3F,$83,$F8,$0F,$C2,$7C,$1F,$E1,$7E
	.dc.b	$12,$57,$9E,$39,$3E,$9B,$AF,$D9,$8E,$8F,$DB,$BB,$97,$E0,$B0,$EB
	.dc.b	$3F,$07,$0E,$AF,$13,$E2,$7C,$27,$E5,$7E,$0E,$46,$0F,$9B,$FB,$77
	.dc.b	$F6,$E3,$FA,$E6,$1F,$E6,$26,$F3,$FD,$D7,$5F,$03,$F8,$EA,$FF,$04
	.dc.b	$C7,$FB,$62,$75,$39,$AE,$5F,$01,$67,$97,$C0,$FE,$9F,$29,$8F,$8A
	.dc.b	$34,$FE,$13,$1F,$27,$C2,$63,$A5,$1E,$49,$D1,$97,$C4,$F1,$CF,$06
	.dc.b	$72,$D7,$C2,$6F,$19,$5F,$0A,$39,$3C,$F1,$8F,$C6,$39,$9F,$22,$EB
	.dc.b	$F0,$AF,$C5,$F0,$9E,$3F,$83,$93,$F3,$A3,$CD,$F0,$92,$3E,$8B,$E9
	.dc.b	$F8,$2F,$57,$C0,$C9,$FA,$AF,$B7,$E0,$3D,$DF,$03,$DA,$7A,$1F,$69
	.dc.b	$E8,$FD,$4B,$BC,$7F,$6F,$0E,$93,$F6,$30,$B3,$F7,$F0,$F9,$5F,$09
	.dc.b	$F2,$FE,$12,$5F,$E1,$A3,$E7,$7C,$1C,$8F,$C4,$BF,$47,$E0,$BE,$97
	.dc.b	$C0,$FD,$3F,$82,$95,$F8,$E8,$FA,$DE,$3C,$8F,$C8,$BF,$67,$E0,$BA
	.dc.b	$47,$E3,$30,$BA,$7E,$4F,$1E,$B3,$F9,$8C,$7A,$BF,$92,$C7,$F0,$5F
	.dc.b	$07,$F8,$7F,$DB,$FF,$BD,$CF,$F0,$E6,$F5,$FE,$2A,$75,$F0,$9B,$F8
	.dc.b	$39,$7F,$C4,$2F,$FC,$02,$DF,$80,$E9,$F0,$12,$30,$6F,$E0,$FC,$04
	.dc.b	$79,$7C,$0F,$67,$A6,$1D,$9F,$F3,$8E,$1D,$9B,$87,$57,$38,$7A,$E3
	.dc.b	$D9,$72,$7A,$9E,$7F,$83,$EC,$F4,$C3,$B7,$07,$FB,$73,$1E,$CD,$C3
	.dc.b	$B3,$9C,$7B,$5F,$C1,$F8,$3F,$95,$F0,$72,$30,$7C,$5F,$83,$F9,$C4
	.dc.b	$7E,$47,$C0,$FD,$1F,$82,$FA,$5F,$03,$27,$E6,$47,$D5,$C2,$47,$43
	.dc.b	$E1,$FC,$17,$37,$E2,$E1,$73,$F5,$63,$D6,$7E,$B6,$3D,$5F,$97,$8F
	.dc.b	$E0,$7E,$0E,$47,$DB,$7F,$09,$F0,$5F,$85,$F8,$39,$3F,$7D,$FC,$47
	.dc.b	$C0,$FE,$27,$E0,$A4,$FE,$0D,$FC,$67,$C1,$72,$8E,$84,$3E,$DD,$F4
	.dc.b	$7E,$ED,$DD,$3E,$96,$1D,$27,$F0,$38,$59,$FA,$D8,$7F,$1F,$FD,$C1
	.dc.b	$F0,$57,$E2,$A6,$1F,$D9,$27,$2D,$F8,$5E,$04,$75,$F0,$BC,$FE,$26
	.dc.b	$47,$17,$B7,$C3,$77,$C6,$4F,$57,$C5,$F0,$DC,$E3,$FD,$8F,$8B,$1C
	.dc.b	$68,$F6,$7F,$C0,$52,$CF,$BD,$F0,$CE,$DE,$E3,$FC,$03,$5F,$F0,$01
	.dc.b	$EB,$8F,$81,$F7,$3F,$0F,$E0,$FB,$E3,$DB,$2B,$5C,$B2,$EB,$3E,$2C
	.dc.b	$AC,$F6,$D3,$E6,$7C,$27,$CD,$F8,$49,$3F,$2E,$3E,$87,$C2,$7D,$1F
	.dc.b	$86,$93,$F3,$E3,$EA,$7C,$34,$8F,$A4,$FD,$7F,$86,$EB,$1C,$C8,$7C
	.dc.b	$8F,$86,$EA,$FD,$9B,$EA,$FC,$CB,$E8,$FC,$EB,$BC,$5F,$1F,$CB,$71
	.dc.b	$B2,$FC,$37,$C2,$CC,$3F,$08,$FE,$27,$E1,$7F,$15,$F0,$D2,$7F,$10
	.dc.b	$FE,$37,$E1,$BF,$1D,$F0,$BF,$8F,$F8,$69,$5F,$8D,$8E,$4F,$42,$1F
	.dc.b	$82,$BE,$8F,$E1,$2E,$E9,$E7,$CD,$8F,$6A,$FC,$C6,$3D,$9F,$C5,$E3
	.dc.b	$FD,$F4,$FF,$7F,$91,$C7,$C1,$E7,$E0,$F5,$F0,$72,$7F,$C8,$B5,$FE
	.dc.b	$00,$DC,$BE,$17,$A7,$C2,$CA,$DC,$74,$78,$9E,$13,$03,$5F,$A1,$5D
	.dc.b	$46,$1E,$19,$D7,$57,$F9,$86,$75,$7D,$E7,$27,$F9,$7A,$18,$7B,$E7
	.dc.b	$BE,$1D,$B3,$B8,$73,$F8,$4F,$81,$8F,$BB,$2F,$05,$F9,$34,$E5,$3F
	.dc.b	$0B,$2F,$04,$FA,$32,$F9,$9F,$0B,$2B,$E3,$BF,$3F,$E1,$3E,$87,$C2
	.dc.b	$49,$F9,$6F,$D3,$F8,$4F,$A9,$F0,$9F,$57,$E1,$65,$7D,$08,$E4,$F4
	.dc.b	$21,$F0,$38,$BE,$9E,$1D,$63,$1C,$4C,$71,$A7,$CE,$CB,$BD,$7D,$CC
	.dc.b	$BB,$BC,$32,$FC,$37,$C2,$7E,$1F,$E1,$64,$FE,$0A,$3F,$15,$F0,$7F
	.dc.b	$8B,$F8,$39,$3F,$86,$8F,$C7,$7C,$17,$E3,$FE,$0E,$4F,$E2,$A3,$AB
	.dc.b	$CC,$87,$A3,$E1,$7A,$BF,$90,$BE,$AF,$B2,$FA,$BF,$94,$BE,$AF,$C0
	.dc.b	$BF,$CF,$7C,$2F,$E7,$FE,$16,$4F,$E6,$A3,$FC,$BD,$FB,$8F,$F8,$58
	.dc.b	$FE,$EF,$D7,$FE,$AD,$6C,$FF,$AA,$1D,$7C,$17,$0F,$81,$93,$FE,$A8
	.dc.b	$94,$F8,$23,$B1,$0F,$AD,$0F,$81,$F0,$DF,$8B,$E1,$A5,$73,$8F,$2D
	.dc.b	$F6,$8E,$38,$61,$E9,$CB,$FA,$8C,$BC,$11,$DB,$1E,$AF,$97,$1E,$AF
	.dc.b	$8B,$1E,$4F,$43,$D5,$86,$30,$FE,$45,$93,$E1,$BE,$39,$2B,$E0,$7C
	.dc.b	$37,$CA,$F8,$6F,$97,$F0,$92,$7E,$24,$7C,$EF,$85,$F9,$FF,$09,$27
	.dc.b	$E4,$47,$D2,$BE,$91,$F1,$B0,$C3,$C8,$F9,$D8,$75,$7E,$5E,$1D,$5F
	.dc.b	$A5,$87,$57,$E7,$E1,$F7,$BE,$12,$47,$D7,$7F,$03,$F0,$7F,$82,$F8
	.dc.b	$49,$3F,$6D,$FC,$2F,$C2,$FE,$1B,$E1,$7F,$0F,$F0,$B2,$BF,$01,$1F
	.dc.b	$8A,$F8,$59,$1F,$83,$7F,$1B,$F0,$BF,$8E,$F8,$5F,$C7,$FC,$2C,$AF
	.dc.b	$C4,$47,$FA,$BF,$F7,$2F,$F0,$6C,$7E,$ED,$7C,$36,$4D,$7C,$36,$FE
	.dc.b	$12,$57,$FB,$E2,$BF,$C0,$CD,$F8,$4E,$9F,$09,$D7,$E1,$24,$EE,$3C
	.dc.b	$1F,$0D,$E1,$F8,$69,$3C,$E3,$AB,$E3,$C6,$F6,$FF,$65,$CB,$FC,$21
	.dc.b	$3C,$CF,$16,$3D,$9F,$56,$3D,$9F,$36,$3D,$9F,$76,$3F,$07,$E1,$7E
	.dc.b	$17,$C3,$49,$F6,$47,$C5,$F8,$4F,$8D,$F0,$B2,$7E,$04,$7C,$9F,$85
	.dc.b	$F9,$5F,$0B,$27,$E2,$47,$CD,$F8,$69,$1F,$1D,$FA,$1F,$0D,$F4,$7E
	.dc.b	$1A,$4F,$CB,$7E,$A7,$C3,$7D,$5F,$85,$FA,$DE,$DC,$AF,$A1,$1F,$67
	.dc.b	$E1,$7F,$DD,$1F,$B9,$FF,$BD,$E7,$F9,$A3,$79,$BE,$44,$EB,$E1,$A5
	.dc.b	$0D,$BC,$BE,$13,$9F,$C2,$F4,$F8,$49,$5B,$C3,$FC,$2D,$8F,$F6,$68
	.dc.b	$C7,$A3,$CF,$E0,$D3,$FD,$8A,$72,$23,$3F,$0D,$E8,$3D,$3F,$0D,$EA
	.dc.b	$F8,$6F,$5F,$C3,$4B,$E3,$7D,$27,$DF,$F0,$DF,$03,$E1,$A5,$6B,$2F
	.dc.b	$EA,$0D,$F8,$5F,$8B,$F0,$D2,$30,$6F,$E1,$7E,$4F,$C3,$48,$C1,$CF
	.dc.b	$E1,$3E,$6F,$C2,$C8,$C1,$DF,$E1,$3E,$8F,$C2,$C8,$C1,$E3,$F8,$6F
	.dc.b	$AB,$F0,$92,$30,$79,$FE,$1B,$E9,$3F,$DA,$FD,$2F,$F6,$C9,$FE,$DA
	.dc.b	$DD,$3A,$E8,$FF,$23,$4F,$F4,$6E,$43,$49,$F0,$DB,$E9,$FA,$90,$F8
	.dc.b	$4C,$65,$F4,$E3,$FA,$23,$F3,$9A,$F8,$E3,$77,$DB,$E0,$AD,$3E,$0E
	.dc.b	$98,$8C,$FC,$14,$C3,$A3,$C3,$E0,$F8,$FC,$17,$8B,$57,$E9,$9E,$5B
	.dc.b	$FC,$3C,$63,$4D,$C7,$97,$E1,$3C,$3F,$09,$23,$07,$A7,$E1,$3C,$BF
	.dc.b	$09,$23,$07,$B7,$E1,$3D,$24,$7E,$07,$C1,$FB,$2B,$DB,$F0,$92,$78
	.dc.b	$9F,$57,$E1,$3E,$09,$1F,$91,$F0,$9E,$BD,$9B,$D2,$4F,$B2,$1A,$FC
	.dc.b	$AE,$49,$3F,$45,$F9,$BF,$09,$F4,$3F,$74,$FE,$6B,$C5,$FE,$85,$7F
	.dc.b	$9A,$D7,$E9,$1A,$FF,$8C,$DD,$F4,$7E,$CD,$FE,$6B,$19,$F8,$2E,$87
	.dc.b	$5F,$83,$ED,$F0,$3D,$FE,$0A,$57,$58,$F1,$7C,$0C,$8F,$03,$E5,$F8
	.dc.b	$2F,$81,$CB,$27,$DC,$FC,$3E,$4F,$57,$C0,$C9,$F4,$3E,$DF,$82,$F7
	.dc.b	$7C,$0C,$9F,$63,$F0,$7E,$0B,$E1,$7C,$0C,$9F,$80,$FC,$5F,$82,$F8
	.dc.b	$DF,$05,$F1,$FE,$0A,$57,$C5,$8F,$95,$F0,$5F,$2F,$E0,$A4,$FC,$98
	.dc.b	$FC,$97,$E9,$DF,$22,$3F,$D0,$8D,$E8,$34,$F9,$EF,$D3,$F8,$2F,$AC
	.dc.b	$49,$FA,$9F,$07,$F6,$7E,$0E,$47,$D7,$7F,$57,$FB,$AF,$F3,$5E,$CF
	.dc.b	$E5,$0B,$E1,$32,$4A,$FE,$43,$1F,$EA,$13,$F9,$83,$FE,$71,$7F,$D4
	.dc.b	$4E,$BE,$0B,$AF,$C1,$F6,$F8,$2E,$FF,$07,$2B,$9C,$78,$BE,$0A,$47
	.dc.b	$67,$CB,$F0,$7E,$6F,$83,$93,$E2,$7D,$3F,$07,$EA,$F8,$3F,$5F,$C1
	.dc.b	$CA,$F3,$C7,$CC,$F0,$A4,$7E,$F3,$95,$1F,$43,$C2,$F8,$7F,$07,$F1
	.dc.b	$3E,$0E,$5F,$C0,$8F,$8F,$F0,$72,$3E,$1B,$F2,$BE,$0F,$E5,$FC,$1C
	.dc.b	$9F,$8E,$FC,$EF,$83,$F9,$FF,$07,$F4,$3E,$0E,$57,$CC,$8F,$A7,$F0
	.dc.b	$5F,$C7,$BF,$76,$76,$53,$35,$E9,$F1,$BF,$0E,$47,$D9,$E3,$69,$FF
	.dc.b	$52,$57,$F3,$53,$F9,$B7,$6F,$F3,$A8,$FE,$C7,$F9,$63,$EF,$3F,$DD
	.dc.b	$31,$FD,$9B,$E2,$5F,$99,$7F,$9F,$23,$0D,$7F,$D9,$F5,$C7,$C5,$F4
	.dc.b	$7C,$3C,$9F,$EC,$F9,$6F,$C7,$F5,$C7,$F7,$18,$FE,$5C,$F2,$CB,$C2
	.dc.b	$DF,$1F,$FA,$2E,$3E,$56,$CE,$AF,$9C,$BF,$09,$7C,$17,$F4,$CD,$5F
	.dc.b	$82,$7F,$4A,$BF,$0C,$63,$E4,$3A,$DF,$0C,$FF,$09,$B7,$EC,$88,$FF
	.dc.b	$82,$07,$E8,$DF,$C6,$B0,$F0,$D7,$19,$F8,$6B,$E0,$FA,$9D,$B3,$FB
	.dc.b	$98,$FD,$DF,$89,$97,$F1,$A3,$FB,$1B,$DB,$DC,$B9,$77,$E7,$FE,$EA
	.dc.b	$D9,$FF,$14,$7E,$27,$FC,$E9,$E4,$7D,$0F,$8B,$FC,$51,$E0,$7E,$9F
	.dc.b	$C5,$FE,$37,$E2,$E4,$60,$F7,$57,$35,$BF,$D8,$7C,$87,$C3,$C3,$C4
	.dc.b	$A7,$84,$FD,$EF,$8C,$73,$B1,$FC,$E7,$C3,$CC,$3C,$07,$E5,$BE,$1B
	.dc.b	$F4,$1C,$32,$3F,$13,$1F,$13,$2E,$BB,$7C,$2D,$FE,$29,$99,$FD,$41
	.dc.b	$FA,$9E,$29,$7C,$8F,$C9,$FC,$5F,$EB,$7E,$2E,$46,$0F,$83,$5D,$1F
	.dc.b	$75,$F5,$7F,$C0,$7F,$A8,$BD,$5F,$F0,$43,$63,$31,$8F,$53,$F7,$5F
	.dc.b	$07,$FB,$BF,$83,$95,$FA,$48,$E8,$FC,$ED,$7D,$8F,$80,$F1,$9F,$AA
	.dc.b	$F8,$CF,$E0,$84,$7F,$5B,$F1,$9F,$C1,$FE,$2F,$F8,$47,$C5,$C9,$FD
	.dc.b	$8C,$76,$E8,$FB,$5D,$3F,$60,$FB,$3C,$9F,$32,$BE,$FE,$EF,$63,$DA
	.dc.b	$48,$F8,$3E,$BF,$EB,$6B,$FA,$B7,$85,$F8,$1F,$91,$C1,$97,$C1,$BF
	.dc.b	$B2,$FD,$C7,$F9,$FF,$2F,$0F,$2F,$EF,$BE,$F1,$F1,$54,$95,$FC,$5A
	.dc.b	$1F,$17,$FB,$DF,$91,$FD,$E6,$3F,$E8,$3D,$DE,$EF,$D4,$35,$E2,$F8
	.dc.b	$B1,$F0,$75,$FD,$D7,$C6,$7E,$67,$4F,$B7,$EC,$C8,$FB,$EF,$DE,$F6
	.dc.b	$BF,$89,$F0,$BA,$69,$2B,$F9,$47,$E4,$FF,$C7,$39,$FF,$51,$CB,$F3
	.dc.b	$67,$F5,$27,$83,$93,$E0,$BF,$CC,$B0,$E8,$DC,$3A,$3F,$CB,$A1,$D5
	.dc.b	$F9,$0F,$6E,$4F,$E5,$1F,$C9,$FB,$7E,$F8,$EF,$97,$E0,$37,$F8,$4B
	.dc.b	$89,$7C,$AB,$EA,$CE,$63,$F3,$BE,$DF,$F9,$C3,$DA,$FF,$0A,$79,$46
	.dc.b	$93,$4F,$06,$DF,$8D,$CF,$E8,$56,$FA,$F9,$13,$3F,$F5,$5E,$1F,$C6
	.dc.b	$F3,$F2,$7B,$7F,$AC,$24,$FE,$C3,$DB,$EB,$8F,$D9,$AF,$9B,$CD,$AA
	.dc.b	$F4,$EE,$C3,$15,$FD,$40,$F8,$18,$FA,$FE,$3A,$5F,$89,$F7,$77,$76
	.dc.b	$8F,$D9,$1F,$86,$E5,$FD,$36,$DF,$06,$B8,$CF,$E7,$23,$0F,$B3,$B7
	.dc.b	$F8,$2F,$B9,$23,$F8,$4B,$FE,$8F,$D7,$F9,$66,$7E,$A9,$EE,$FD,$77
	.dc.b	$F8,$09,$EC,$9E,$4D,$3F,$D2,$B3,$FE,$B1,$CF,$FC,$37,$1F,$33,$DE
	.dc.b	$FB,$07,$D9,$F7,$BD,$91,$F9,$1E,$0F,$AB,$5C,$23,$D4,$DC,$7F,$9D
	.dc.b	$37,$BF,$FD,$D8,$BE,$3B,$D5,$E9,$CB,$F1,$BE,$DF,$47,$BE,$1E,$B9
	.dc.b	$F9,$5A,$6D,$FC,$7B,$FE,$B9,$78,$9E,$CF,$4F,$F2,$1F,$1D,$23,$F0
	.dc.b	$EF,$E4,$FE,$3B,$AE,$9F,$C3,$70,$F9,$55,$EA,$AE,$0D,$3F,$1F,$F1
	.dc.b	$DF,$47,$D5,$91,$F2,$9F,$A9,$97,$48,$FE,$0E,$7D,$2B,$FC,$54,$7F
	.dc.b	$AB,$23,$81,$0F,$5F,$AB,$FA,$7F,$82,$91,$F9,$E7,$F5,$5F,$05,$D6
	.dc.b	$3F,$6B,$9F,$FB,$C3,$A7,$F1,$BC,$4B,$E3,$5F,$51,$FD,$97,$C6,$FE
	.dc.b	$27,$D8,$93,$F8,$27,$F1,$9E,$C7,$5F,$0B,$F8,$6C,$FC,$E9,$DB,$4F
	.dc.b	$C4,$FB,$1F,$BD,$F8,$E9,$1F,$B4,$7F,$80,$7C,$6F,$3E,$AF,$C3,$C6
	.dc.b	$A7,$F0,$EF,$F0,$5F,$8E,$FC,$E7,$B1,$27,$F2,$AF,$E7,$FD,$9E,$51
	.dc.b	$F2,$F8,$4F,$F8,$3B,$91,$5C,$63,$F0,$AF,$E9,$FC,$EF,$E2,$5E,$FC
	.dc.b	$9F,$E1,$2F,$F1,$5F,$83,$EB,$1F,$8E,$32,$43,$F3,$B1,$FC,$6F,$E3
	.dc.b	$BF,$65,$E8,$C9,$FD,$5B,$FB,$5F,$6F,$9F,$5F,$E3,$0F,$C0,$C6,$E7
	.dc.b	$F3,$2F,$EE,$FD,$CF,$5C,$7F,$03,$FD,$DF,$FD,$58,$BC,$A4,$61,$B1
	.dc.b	$C2,$FF,$C5,$13,$FE,$B1,$8F,$8C,$7E,$48,$FD,$C1,$89,$E2,$FC,$A9
	.dc.b	$E2,$F3,$DF,$E6,$CE,$A3,$C9,$ED,$65,$FF,$1E,$FE,$A7,$FE,$CD,$27
	.dc.b	$A5,$9C,$11,$FE,$CE,$C3,$F2,$DC,$5F,$CC,$5F,$D8,$18,$BF,$13,$E3
	.dc.b	$9E,$71,$83,$BD,$F5,$8F,$F2,$D6,$3F,$13,$E3,$BC,$8D,$AE,$0D,$7E
	.dc.b	$67,$C7,$FD,$4C,$3C,$11,$F3,$EB,$8E,$9F,$ED,$4C,$FE,$5C,$63,$E3
	.dc.b	$FC,$1E,$8C,$BF,$25,$7B,$A3,$C9,$E9,$75,$9F,$F3,$6E,$DB,$7F,$B2
	.dc.b	$EA,$AC,$5F,$F4,$D8,$FF,$0A,$CF,$A3,$2F,$F2,$23,$F8,$3F,$82,$F2
	.dc.b	$9E,$6F,$83,$ED,$1F,$E1,$13,$C7,$E9,$BB,$3E,$37,$A9,$27,$31,$F5
	.dc.b	$67,$51,$82,$1F,$EB,$4C,$3F,$D2,$99,$7E,$67,$83,$EB,$CF,$E2,$E3
	.dc.b	$0F,$F9,$FA,$B9,$BF,$E0,$26,$C7,$F6,$96,$CF,$FA,$4B,$4F,$F4,$A9
	.dc.b	$2B,$FD,$08,$FD,$DF,$53,$DB,$E4,$FF,$88,$7C,$5F,$91,$5A,$21,$FB
	.dc.b	$73,$F5,$1F,$1F,$F6,$88,$FC,$BF,$8F,$E7,$C5,$FE,$0E,$C7,$C1,$3C
	.dc.b	$74,$FA,$5E,$B7,$E3,$49,$1F,$7B,$8F,$FA,$C3,$FA,$C3,$F9,$89,$F0
	.dc.b	$F5,$FC,$5D,$7C,$48,$FF,$5F,$DF,$68,$F8,$FD,$1F,$BA,$F8,$5F,$C7
	.dc.b	$92,$BF,$01,$F0,$BE,$67,$B6,$DF,$01,$FF,$20,$61,$FE,$32,$C9,$E9
	.dc.b	$FE,$8C,$93,$E7,$F5,$39,$3E,$DD,$3E,$83,$F2,$34,$F1,$F2,$7F,$98
	.dc.b	$23,$F2,$1F,$09,$FA,$23,$F4,$7F,$07,$CA,$BD,$9A,$78,$DF,$F4,$1E
	.dc.b	$9E,$3F,$5B,$F6,$84,$C3,$E4,$7A,$5C,$73,$F8,$A7,$8A,$3D,$67,$F1
	.dc.b	$3C,$DF,$07,$FB,$22,$47,$E8,$FE,$07,$93,$FA,$C3,$D5,$5C,$F2,$FC
	.dc.b	$99,$FC,$43,$D6,$FE,$0C,$49,$FB,$5E,$AF,$27,$F5,$07,$88,$93,$F6
	.dc.b	$7C,$D7,$9B,$FE,$EC,$6C,$FF,$BE,$DA,$FF,$A0,$BE,$93,$FD,$22,$FD
	.dc.b	$8F,$4B,$F8,$F3,$E7,$3E,$91,$FE,$99,$E0,$FC,$26,$7F,$12,$3F,$AD
	.dc.b	$FE,$97,$FA,$66,$F8,$5F,$EF,$C8,$C4,$7F,$B9,$FD,$AF,$57,$83,$FA
	.dc.b	$5C,$39,$CF,$7D,$7F,$D1,$35,$FE,$E1,$24,$7E,$66,$BF,$59,$97,$9A
	.dc.b	$3E,$1D,$7D,$EB,$E7,$18,$CB,$CC,$F8,$4F,$F7,$FE,$1F,$E8,$7C,$39
	.dc.b	$C7,$F9,$4A,$77,$1E,$58,$FA,$E7,$96,$33,$3E,$48,$F8,$A7,$73,$F5
	.dc.b	$84,$30,$4E,$5F,$C9,$4F,$E0,$4E,$FD,$FC,$F1,$F8,$35,$EF,$8F,$75
	.dc.b	$7F,$87,$27,$DD,$3E,$43,$AE,$1F,$EA,$F3,$FC,$6C,$FF,$9E,$70,$F7
	.dc.b	$5F,$2C,$AC,$F1,$C3,$C6,$FB,$48,$7C,$73,$91,$E3,$3E,$31,$ED,$BF
	.dc.b	$05,$7D,$09,$E7,$3E,$48,$F6,$63,$FE,$3F,$C3,$C3,$3F,$DF,$09,$8F
	.dc.b	$F1,$1C,$3E,$84,$7E,$72,$BE,$2B,$F6,$08,$7D,$08,$FB,$2F,$28,$DB
	.dc.b	$C4,$FC,$C1,$E9,$3D,$64,$9F,$05,$FC,$58,$D4,$7F,$06,$7E,$75,$7B
	.dc.b	$9F,$0C,$7D,$67,$C5,$3C,$5A,$7C,$DC,$BD,$F3,$B2,$5F,$99,$FF,$43
	.dc.b	$11,$C1,$F8,$59,$F4,$9D,$0F,$43,$F2,$DF,$3B,$FD,$79,$FE,$7A,$7D
	.dc.b	$13,$C4,$68,$FF,$28,$DF,$E1,$A7,$C7,$1F,$80,$8F,$97,$1E,$38,$C5
	.dc.b	$FD,$DB,$F5,$DF,$86,$7C,$66,$8F,$C2,$9E,$2B,$F9,$0F,$D3,$3F,$A6
	.dc.b	$9E,$33,$F1,$D7,$E2,$7D,$07,$DB,$3F,$1D,$7F,$D8,$5F,$A6,$FE,$E7
	.dc.b	$0F,$C6,$C7,$84,$E0,$78,$63,$83,$E1,$AF,$1D,$77,$7E,$15,$FD,$09
	.dc.b	$FD,$C6,$3F,$AF,$8F,$E1,$2F,$E7,$0F,$CE,$CF,$28,$F8,$55,$CE,$FF
	.dc.b	$8E,$9F,$1E,$FD,$D5,$D9,$FA,$38,$7C,$A7,$BE,$55,$F8,$A7,$9E,$BD
	.dc.b	$CF,$E4,$30,$F3,$3F,$7B,$1F,$33,$FB,$87,$CD,$97,$3C,$7C,$13,$F3
	.dc.b	$AB,$8D,$30,$FE,$55,$5E,$BA,$FC,$7E,$5F,$A1,$7F,$20,$FC,$7C,$BC
	.dc.b	$D7,$C6,$7E,$FD,$71,$7D,$35,$C4,$FD,$6B,$E9,$3E,$09,$F8,$19,$E4
	.dc.b	$FD,$89,$E5,$1E,$A8,$F9,$35,$F7,$9F,$C2,$D7,$E5,$5F,$BC,$7F,$84
	.dc.b	$31,$FE,$83,$7E,$33,$E0,$93,$2F,$AB,$3F,$A4,$3D,$B9,$7D,$A2,$3F
	.dc.b	$8B,$BE,$D8,$FC,$E8,$FE,$75,$87,$6B,$FA,$B1,$FA,$23,$CB,$7F,$B3
	.dc.b	$AE,$18,$7F,$1C,$C3,$E2,$4F,$DA,$8F,$2E,$3F,$33,$2F,$A7,$8F,$E6
	.dc.b	$2F,$78,$F8,$F0,$F2,$BF,$6C,$C1,$F8,$A3,$E6,$65,$E5,$7F,$B4,$61
	.dc.b	$F2,$71,$F9,$B8,$F6,$9F,$90,$7E,$B0,$D9,$F5,$CF,$59,$D4,$F8,$E7
	.dc.b	$E1,$AF,$F8,$B3,$A3,$C8,$49,$EE,$FE,$1B,$1F,$64,$F0,$9F,$AD,$7E
	.dc.b	$59,$FA,$B5,$FC,$1F,$2F,$D7,$C7,$CE,$CB,$F5,$D3,$FE,$17,$3A,$1E
	.dc.b	$A3,$F5,$07,$A8,$FA,$97,$E2,$7F,$1D,$3E,$27,$D4,$7C,$A3,$D1,$7F
	.dc.b	$4D,$EA,$7F,$7E,$BF,$CA,$3F,$08,$D9,$F9,$BC,$7E,$63,$BC,$BE,$93
	.dc.b	$FD,$5E,$BD,$CF,$A4,$F1,$9E,$9C,$7E,$FB,$F2,$F1,$FC,$2B,$D4,$F1
	.dc.b	$C7,$F8,$09,$F7,$9D,$4F,$B1,$1E,$F8,$FD,$33,$FA,$CA,$F1,$C7,$F6
	.dc.b	$FA,$F1,$D7,$B6,$FD,$33,$D4,$F1,$DF,$D6,$9E,$B8,$FD,$69,$F3,$E3
	.dc.b	$F5,$A7,$E8,$47,$D6,$7A,$1E,$8C,$BF,$95,$E5,$E2,$7F,$C5,$71,$D0
	.dc.b	$F8,$93,$E8,$BF,$AD,$1F,$C3,$27,$F5,$19,$7A,$2B,$FC,$FB,$97,$A2
	.dc.b	$7A,$D7,$E6,$CF,$88,$7E,$0A,$FD,$11,$F9,$8C,$7D,$13,$AB,$F4,$1D
	.dc.b	$1F,$41,$FC,$B7,$1F,$43,$FD,$9F,$1F,$15,$FF,$46,$C7,$C5,$3B,$3C
	.dc.b	$57,$F3,$DF,$71,$F1,$0F,$C0,$9F,$87,$37,$3F,$11,$F4,$1F,$11,$FE
	.dc.b	$FD,$97,$DB,$8D,$9E,$33,$AC,$FA,$5E,$AF,$DB,$9F,$D8,$65,$FA,$B7
	.dc.b	$F8,$5D,$7E,$1D,$EB,$1F,$43,$2F,$8B,$5E,$9B,$F1,$9E,$97,$EB,$3F
	.dc.b	$99,$AE,$AF,$E0,$9E,$B5,$F4,$23,$F5,$E7,$DB,$3F,$61,$7F,$9E,$7F
	.dc.b	$59,$7F,$D1,$2F,$AC,$7C,$53,$F3,$EF,$8C,$FA,$F7,$F4,$63,$EB,$9F
	.dc.b	$CD,$8F,$8A,$7F,$99,$B1,$FA,$EF,$F6,$7C,$7E,$BB,$B3,$AC,$FD,$1A
	.dc.b	$EB,$1B,$3A,$E1,$FC,$EE,$7D,$F1,$F8,$97,$DE,$FD,$73,$FC,$32,$7C
	.dc.b	$53,$E0,$C7,$EA,$DF,$97,$3F,$5C,$F7,$9F,$5D,$DE,$3F,$5D,$FE,$D2
	.dc.b	$7C,$6B,$F4,$D7,$8E,$36,$7D,$13,$75,$FD,$C9,$FF,$12,$E3,$FD,$52
	.dc.b	$FF,$C2,$D1,$FC,$6E,$BE,$95,$6C,$FE,$C3,$7F,$E0,$0C,$BF,$43,$5F
	.dc.b	$2C,$FC,$7E,$3F,$45,$FC,$7B,$FB,$27,$E5,$9F,$95,$3D,$EE,$EB,$DE
	.dc.b	$ED,$FE,$1F,$87,$BE,$7F,$A5,$3F,$89,$7F,$10,$4C,$3E,$54,$FD,$68
	.dc.b	$FF,$57,$61,$F1,$2B,$DB,$3E,$BB,$FC,$FE,$5F,$CF,$6F,$CE,$7E,$DC
	.dc.b	$F3,$9E,$D3,$CE,$7E,$04,$D9,$F8,$79,$F3,$BF,$3C,$F3,$9F,$81,$3F
	.dc.b	$C2,$A7,$E6,$5D,$BF,$87,$3E,$AC,$FF,$8D,$2F,$D1,$1F,$E1,$59,$F4
	.dc.b	$3A,$21,$FD,$BF,$2F,$D3,$57,$EE,$23,$F1,$C7,$C5,$C3,$E5,$63,$FE
	.dc.b	$21,$8F,$25,$7E,$05,$ED,$5F,$EA,$84,$41,$2A,$C4,$48,$C3,$83,$FD
	.dc.b	$1B,$15,$A5,$8D,$BB,$6B,$CE,$38,$3E,$06,$C6,$A3,$C8,$D7,$A9,$CE
	.dc.b	$36,$47,$C0,$D7,$46,$08,$60,$87,$8C,$C9,$4F,$D3,$7B,$5F,$E6,$D8
	.dc.b	$F8,$D8,$2F,$E4,$60,$F3,$35,$87,$CC,$AD,$3F,$3A,$5F,$05,$A8,$4E
	.dc.b	$9F,$A3,$8E,$63,$15,$D6,$7F,$D1,$B8,$F8,$B1,$F0,$E1,$B7,$E3,$E1
	.dc.b	$6B,$C1,$8D,$58,$78,$9E,$67,$9E,$B8,$B8,$3C,$AF,$BA,$79,$4F,$19
	.dc.b	$FE,$15,$87,$8D,$FF,$67,$DE,$AB,$78,$78,$72,$4E,$6B,$B6,$5E,$EA
	.dc.b	$D4,$E6,$7A,$9E,$53,$E5,$46,$9E,$39,$79,$5A,$68,$A7,$CC,$2E,$5F
	.dc.b	$2D,$EA,$7B,$8E,$06,$2F,$EA,$61,$E8,$8E,$38,$FB,$1E,$B1,$97,$85
	.dc.b	$F8,$DA,$74,$AF,$DE,$BF,$9F,$3D,$87,$94,$D9,$F5,$AB,$B3,$8A,$ED
	.dc.b	$3C,$9F,$D6,$BF,$86,$9D,$B9,$CB,$EA,$57,$58,$C3,$E0,$C3,$E1,$46
	.dc.b	$23,$E1,$1F,$86,$C7,$E7,$3E,$49,$DE,$3F,$E9,$BB,$EB,$1F,$83,$BE
	.dc.b	$47,$C8,$CB,$ED,$38,$AF,$CB,$DF,$C8,$9C,$C7,$38,$F2,$61,$E0,$AE
	.dc.b	$11,$F9,$A3,$EC,$C6,$67,$84,$7A,$2F,$CA,$FF,$83,$B0,$F2,$D6,$E3
	.dc.b	$07,$C1,$C7,$E3,$DF,$2C,$3D,$13,$F7,$8E,$E7,$9C,$C1,$F0,$EB,$A3
	.dc.b	$FE,$79,$BF,$87,$7D,$23,$B9,$F3,$32,$FB,$38,$77,$3F,$0D,$87,$BA
	.dc.b	$7E,$81,$DC,$FA,$15,$CD,$EC,$7E,$22,$B9,$BD,$A7,$F0,$4F,$A0,$FC
	.dc.b	$46,$5F,$39,$D5,$FD,$27,$27,$B2,$7C,$59,$7D,$B9,$F3,$4F,$EF,$0D
	.dc.b	$38,$7C,$E7,$23,$CE,$78,$67,$D9,$7E,$68,$F4,$9F,$5F,$0F,$64,$7F
	.dc.b	$BE,$71,$FC,$BB,$B3,$13,$F5,$CE,$97,$E1,$8D,$4F,$02,$4E,$CF,$AC
	.dc.b	$7C,$BB,$F6,$DF,$82,$77,$3F,$55,$F0,$E3,$F1,$E3,$CF,$5F,$C5,$5E
	.dc.b	$58,$7C,$1C,$7F,$1F,$1A,$8F,$99,$3F,$02,$B9,$63,$EA,$AF,$CB,$65
	.dc.b	$F8,$88,$C5,$7D,$89,$F3,$9F,$2D,$F3,$BF,$00,$FB,$27,$D3,$3E,$79
	.dc.b	$C1,$F0,$3F,$58,$D9,$F5,$9C,$BC,$1F,$BF,$3E,$3C,$7E,$1C,$7F,$20
	.dc.b	$7F,$37,$3E,$13,$DC,$78,$63,$CA,$F8,$70,$FA,$98,$7C,$0A,$E7,$1F
	.dc.b	$DE,$9F,$D9,$CF,$CA,$3C,$B8,$FC,$AC,$7E,$1D,$7C,$A3,$9C,$FA,$CF
	.dc.b	$09,$EB,$72,$FA,$DF,$A9,$8F,$AE,$3F,$4D,$7E,$27,$F8,$51,$E7,$3E
	.dc.b	$2D,$7E,$29,$D9,$0F,$CB,$1F,$05,$F2,$BF,$06,$BF,$05,$97,$E2,$E3
	.dc.b	$86,$5F,$25,$FD,$B6,$5F,$0D,$CD,$7C,$37,$F8,$F9,$F0,$CF,$6E,$5F
	.dc.b	$0F,$1F,$44,$7F,$47,$3F,$12,$6F,$1F,$1B,$FD,$1C,$F3,$47,$08,$FE
	.dc.b	$8B,$5F,$DF,$27,$E3,$1F,$43,$2F,$8C,$FF,$AC,$B2,$F5,$3F,$E2,$03
	.dc.b	$E8,$57,$6C,$3F,$82,$47,$C8,$C7,$D8,$FF,$AE,$AF,$9E,$5F,$D7,$AB
	.dc.b	$D5,$3F,$00,$FF,$1C,$C7,$EA,$DF,$CF,$63,$F2,$9F,$49,$B3,$E0,$47
	.dc.b	$F3,$47,$D2,$7C,$6A,$EF,$3F,$2E,$3F,$74,$7F,$01,$9F,$CE,$BF,$1F
	.dc.b	$2F,$88,$7E,$71,$FC,$11,$F8,$CC,$7F,$04,$FC,$33,$C6,$7C,$33,$C2
	.dc.b	$FC,$3B,$F0,$C7,$B2,$FC,$57,$FB,$B3,$C3,$1E,$E3,$F3,$F9,$7C,$E8
	.dc.b	$FE,$F5,$97,$CA,$7E,$45,$F4,$AF,$F4,$05,$6E,$FA,$92,$3F,$24,$74
	.dc.b	$3C,$F5,$D1,$FB,$38,$FC,$07,$F1,$E7,$8C,$FC,$7D,$7F,$32,$C3,$51
	.dc.b	$FD,$5D,$F2,$5F,$78,$FA,$6F,$ED,$5F,$09,$FE,$0D,$7F,$1D,$7F,$E5
	.dc.b	$BE,$17,$A5,$CA,$BE,$31,$0F,$DD,$1F,$E9,$28,$DB,$FE,$24,$6B,$FE
	.dc.b	$14,$6C,$FF,$99,$FE,$9D,$E2,$F5,$AF,$F5,$36,$9F,$E5,$3A,$7D,$C3
	.dc.b	$AB,$0D,$AF,$F2,$64,$EA,$73,$D5,$CF,$83,$F5,$35,$E7,$E1,$FD,$0F
	.dc.b	$8A,$EC,$7C,$37,$F9,$86,$30,$D8,$CD,$7F,$96,$67,$87,$C3,$75,$E5
	.dc.b	$FD,$EE,$1F,$AC,$D3,$CF,$E1,$79,$89,$87,$9B,$67,$85,$C6,$3F,$4D
	.dc.b	$7E,$4C,$DF,$09,$E1,$24,$F8,$61,$F0,$DC,$63,$F4,$E7,$F0,$C3,$FC
	.dc.b	$C5,$97,$C8,$F2,$7E,$39,$23,$C5,$E4,$F1,$9F,$E3,$D1,$C5,$FF,$4F
	.dc.b	$EF,$F1,$7E,$07,$E1,$12,$7C,$5E,$77,$1E,$BF,$C6,$6B,$F0,$77,$79
	.dc.b	$BF,$58,$97,$F0,$7A,$78,$3F,$4F,$A7,$E8,$7E,$F1,$FE,$A9,$D7,$C2
	.dc.b	$DC,$BF,$DB,$BA,$A3,$16,$8C,$7C,$E7,$48,$FC,$E9,$FE,$77,$AF,$F5
	.dc.b	$D6,$6F,$4B,$F1,$24,$9F,$A9,$E9,$FF,$81,$F6,$FE,$4B,$1F,$92,$3F
	.dc.b	$5F,$9F,$F1,$F2,$61,$FE,$F0,$F9,$07,$D0,$3D,$5E,$5B,$FF,$66,$C3
	.dc.b	$E6,$5F,$8B,$D8,$F7,$FC,$9C,$9F,$C6,$70,$A3,$E1,$FC,$D7,$29,$F0
	.dc.b	$DF,$ED,$F4,$FD,$1E,$1F,$A2,$24,$FE,$35,$FD,$2F,$B7,$FE,$47,$9D
	.dc.b	$CE,$48,$7E,$1B,$85,$ED,$FD,$2F,$8E,$93,$E4,$7E,$AF,$C8,$FF,$94
	.dc.b	$79,$BE,$94,$6A,$7F,$D6,$11,$87,$D7,$F1,$9F,$B6,$61,$FA,$13,$F7
	.dc.b	$3F,$31,$E5,$C5,$1A,$3C,$87,$EF,$BE,$03,$F0,$BF,$1B,$23,$E3,$BF
	.dc.b	$88,$F8,$DF,$27,$1A,$35,$3E,$48,$C3,$F3,$BD,$3F,$E1,$9F,$74,$F8
	.dc.b	$7A,$49,$FE,$17,$0F,$6B,$E5,$3B,$8C,$CE,$DF,$D7,$C7,$F1,$CE,$5F
	.dc.b	$CB,$E1,$FC,$43,$7F,$CC,$79,$7F,$9E,$3F,$3F,$EA,$7A,$1C,$61,$F8
	.dc.b	$27,$F4,$7F,$1B,$E6,$F8,$B9,$3E,$47,$D3,$F1,$9E,$47,$31,$E2,$7D
	.dc.b	$BE,$E7,$EB,$3E,$3A,$4F,$E3,$9F,$D7,$FC,$77,$C8,$73,$1F,$1D,$FD
	.dc.b	$9F,$C6,$FC,$6F,$83,$93,$F1,$1F,$93,$F0,$7F,$11,$CC,$7C,$27,$E6
	.dc.b	$FC,$1F,$EF,$3E,$3A,$4F,$E7,$9F,$DF,$FB,$DF,$61,$CC,$7D,$77,$F8
	.dc.b	$1F,$C6,$FD,$6F,$88,$93,$F5,$1F,$B3,$F1,$1F,$51,$CC,$7D,$27,$EE
	.dc.b	$FC,$47,$F0,$CF,$8E,$93,$FA,$E7,$F8,$7F,$C7,$7E,$21,$CC,$7E,$1D
	.dc.b	$FE,$85,$FB,$CF,$FB,$4B,$C8,$76,$3E,$3B,$F1,$24,$BF,$C3,$FC,$77
	.dc.b	$E1,$A3,$11,$F8,$47,$F1,$FF,$21,$D3,$D0,$93,$FD,$2F,$1F,$EA,$63
	.dc.b	$E2,$FF,$5C,$E2,$77,$3F,$AE,$87,$C6,$7F,$65,$E7,$6D,$49,$5F,$9A
	.dc.b	$F8,$CF,$E0,$53,$A8,$FE,$CE,$D7,$FA,$3B,$76,$78,$7F,$B4,$7E,$0E
	.dc.b	$5F,$13,$E4,$25,$FA,$5F,$8F,$EA,$FA,$1C,$C7,$9D,$F9,$7B,$F9,$FA
	.dc.b	$24,$F8,$9F,$57,$3F,$99,$CC,$79,$1F,$77,$27,$D4,$E3,$93,$F2,$5F
	.dc.b	$AF,$C7,$F2,$1C,$C7,$C7,$7E,$DF,$37,$C7,$ED,$93,$F0,$9F,$95,$C1
	.dc.b	$F1,$9C,$C7,$C4,$7E,$77,$4F,$F1,$68,$70,$7C,$65,$EB,$BF,$31,$FB
	.dc.b	$51,$F2,$78,$FF,$A3,$46,$23,$8C,$63,$A3,$A5,$F3,$E5,$7C,$9B,$D2
	.dc.b	$FF,$80,$19,$75,$BD,$E1,$C7,$1E,$C7,$8B,$2E,$D9,$79,$6D,$CE,$E2
	.dc.b	$FC,$47,$8B,$30,$FC,$1B,$D1,$6B,$EB,$F5,$7F,$13,$E4,$7E,$3C,$93
	.dc.b	$F8,$AF,$13,$85,$3F,$24,$E3,$3E,$48,$C1,$A8,$CC,$7C,$23,$BB,$FD
	.dc.b	$C3,$E3,$3F,$D2,$D3,$E3,$D3,$F4,$9E,$55,$E1,$FF,$1D,$C6,$30,$F8
	.dc.b	$2F,$EA,$72,$FD,$CF,$89,$27,$F5,$CF,$EF,$3C,$4F,$D1,$F9,$FF,$A9
	.dc.b	$24,$FE,$93,$1F,$E0,$3E,$3C,$8F,$DB,$3F,$C1,$3C,$8F,$DA,$F9,$1F
	.dc.b	$C1,$89,$3F,$BB,$F2,$BF,$84,$F9,$72,$3F,$7C,$FF,$0C,$F2,$FF,$67
	.dc.b	$EB,$7E,$E4,$93,$FB,$4F,$5B,$F8,$8F,$A1,$23,$F8,$33,$FC,$53,$D1
	.dc.b	$FE,$0B,$F2,$1F,$E5,$73,$F2,$93,$FD,$C9,$BD,$BF,$E3,$AF,$3F,$FA
	.dc.b	$A4,$FF,$89,$E5,$EB,$F1,$24,$7F,$97,$61,$EB,$7F,$7F,$77,$19,$9D
	.dc.b	$BD,$FD,$9F,$17,$74,$8E,$4F,$97,$D9,$FF,$5B,$65,$FE,$CB,$EC,$D5
	.dc.b	$79,$EA,$CF,$F9,$43,$1F,$F1,$B4,$FD,$CF,$69,$CD,$F8,$36,$C6,$63
	.dc.b	$F2,$3F,$8C,$3D,$FE,$FE,$3F,$ED,$FD,$FF,$C7,$E5,$FE,$BB,$8F,$E3
	.dc.b	$9F,$7F,$FE,$6E,$C7,$91,$0F,$CA,$78,$2C,$FB,$C6,$17,$67,$E5,$7F
	.dc.b	$12,$7A,$74,$95,$FA,$1F,$C3,$9E,$27,$FA,$6F,$2F,$04,$FF,$9D,$9B
	.dc.b	$A7,$F0,$B0,$E3,$4F,$F3,$0F,$0F,$B6,$BF,$3F,$97,$81,$B9,$78,$1F
	.dc.b	$4C,$3D,$5F,$C3,$7C,$34,$9F,$BC,$FE,$27,$E1,$BC,$9F,$1F,$27,$FD
	.dc.b	$57,$9A,$3D,$1F,$1F,$C6,$7F,$1D,$77,$5F,$B1,$F1,$9F,$93,$F8,$C9
	.dc.b	$5F,$8B,$7F,$2D,$87,$E5,$F0,$93,$F8,$F7,$F3,$5F,$27,$F9,$BF,$93
	.dc.b	$93,$F9,$37,$F3,$DF,$27,$F9,$FF,$93,$FD,$07,$C9,$CA,$FC,$C4,$7C
	.dc.b	$DF,$8E,$91,$EA,$7E,$87,$67,$E9,$FE,$4A,$4F,$E7,$DF,$D5,$7C,$97
	.dc.b	$EA,$FE,$4B,$FD,$85,$FB,$D7,$FA,$3B,$E3,$0D,$EA,$4E,$BE,$4E,$63
	.dc.b	$FE,$C5,$A4,$72,$F9,$2E,$7E,$0F,$4F,$81,$97,$B8,$EF,$E8,$C8,$E4
	.dc.b	$F8,$BD,$1F,$1F,$C8,$49,$EC,$F9,$BE,$47,$CF,$EC,$C9,$F1,$3E,$AF
	.dc.b	$67,$D7,$E3,$7B,$3E,$02,$57,$9E,$3F,$28,$C3,$F0,$27,$E5,$FD,$2F
	.dc.b	$87,$F0,$12,$3D,$CF,$C6,$F8,$6F,$8F,$F0,$F2,$7E,$13,$F2,$BE,$1F
	.dc.b	$E5,$F9,$72,$7E,$33,$F3,$BC,$AF,$9F,$F0,$9F,$43,$E3,$A5,$7C,$B8
	.dc.b	$FF,$7A,$7E,$F7,$FB,$AF,$3F,$FD,$86,$D9,$FF,$61,$37,$B4,$FF,$BC
	.dc.b	$FF,$04,$7C,$0F,$F5,$DF,$35,$18,$BF,$ED,$31,$89,$D4,$BE,$07,$C3
	.dc.b	$F0,$B2,$FF,$B7,$E3,$DA,$3C,$BF,$23,$E6,$F9,$0F,$3F,$C8,$CB,$DE
	.dc.b	$89,$F5,$FC,$8C,$9C,$BE,$EF,$90,$F7,$FC,$8F,$C0,$F9,$19,$5C,$A3
	.dc.b	$E1,$FC,$64,$8E,$AF,$C6,$F8,$CF,$4D,$49,$F6,$F0,$23,$DB,$5F,$33
	.dc.b	$E0,$A5,$79,$9F,$9F,$EF,$FD,$0F,$8C,$FA,$3C,$92,$BD,$71,$F5,$39
	.dc.b	$24,$7B,$9F,$AF,$F1,$9F,$63,$E3,$24,$FC,$27,$ED,$FC,$67,$DC,$F4
	.dc.b	$64,$FC,$67,$FC,$25,$D0,$FD,$F3,$FA,$DF,$5F,$F9,$4F,$87,$E5,$5F
	.dc.b	$B3,$9B,$EB,$9C,$2B,$FA,$BC,$7F,$B5,$1F,$EB,$6F,$D8,$6B,$9F,$8C
	.dc.b	$E1,$F1,$9C,$7E,$2B,$97,$C6,$4A,$E1,$1D,$7D,$A9,$1C,$DF,$07,$B5
	.dc.b	$E1,$F8,$49,$3D,$DF,$27,$C3,$79,$7E,$16,$4F,$8D,$F4,$7C,$2F,$A7
	.dc.b	$E1,$64,$F9,$DF,$67,$C2,$FB,$7E,$16,$4F,$AD,$F8,$1F,$0B,$F0,$7E
	.dc.b	$16,$4F,$BD,$F8,$9F,$09,$F1,$7E,$16,$4F,$C3,$7E,$47,$C2,$FC,$9F
	.dc.b	$85,$93,$F1,$DF,$99,$F0,$BF,$53,$D1,$FE,$48,$4E,$BC,$7C,$26,$7E
	.dc.b	$13,$5F,$07,$31,$FD,$3D,$A7,$97,$C2,$73,$F8,$49,$5A,$8E,$DF,$09
	.dc.b	$23,$8B,$E1,$F8,$4F,$17,$C2,$49,$EA,$F9,$7E,$13,$CD,$F0,$92,$7C
	.dc.b	$2F,$A7,$E0,$FD,$5F,$09,$27,$CA,$FB,$7E,$0F,$DD,$F0,$72,$7D,$2F
	.dc.b	$C1,$F8,$3F,$85,$F0,$7F,$0F,$E0,$E5,$7B,$A3,$E3,$7C,$1F,$C7,$F8
	.dc.b	$39,$3F,$0A,$3E,$57,$27,$E7,$BF,$7D,$FE,$30,$7F,$19,$7F,$D6,$5D
	.dc.b	$0A,$7A,$0A,$FF,$3D,$C7,$F1,$26,$BF,$E6,$B8,$D7,$C8,$F5,$F8,$4E
	.dc.b	$DF,$09,$DF,$D2,$97,$C6,$3C,$5F,$09,$23,$A3,$E5,$F8,$3F,$37,$C1
	.dc.b	$F9,$FE,$0E,$57,$86,$3A,$BE,$3C,$3A,$BE,$3D,$3F,$DA,$EF,$42,$19
	.dc.b	$C3,$B3,$CC,$F6,$5F,$67,$95,$FC,$7F,$82,$F9,$1F,$05,$27,$E1,$47
	.dc.b	$CB,$F8,$3F,$99,$F0,$72,$7E,$34,$7C,$FF,$83,$FA,$1F,$05,$27,$E5
	.dc.b	$47,$D3,$F8,$39,$1F,$35,$FA,$DF,$07,$F5,$FE,$0E,$4F,$D1,$7E,$D7
	.dc.b	$C1,$FD,$BF,$83,$FB,$9F,$07,$2B,$EB,$47,$F1,$CF,$DF,$BF,$8D,$C7
	.dc.b	$DC,$6F,$49,$A7,$F1,$AF,$86,$FF,$67,$46,$7E,$0F,$91,$2B,$FA,$9D
	.dc.b	$3E,$0F,$B7,$C1,$77,$F8,$39,$5A,$8F,$13,$0E,$27,$97,$E1,$FA,$3F
	.dc.b	$8B,$BF,$E4,$F9,$78,$9B,$E0,$7E,$57,$3F,$46,$5E,$0B,$F1,$E5,$E0
	.dc.b	$79,$65,$F0,$7E,$12,$47,$A1,$F8,$9F,$09,$F1,$7E,$13,$E3,$7C,$24
	.dc.b	$AF,$6C,$7C,$9F,$84,$91,$F0,$1F,$99,$F0,$9D,$23,$D1,$85,$DF,$D1
	.dc.b	$8F,$59,$F8,$F8,$F5,$7D,$D8,$F4,$7E,$06,$36,$FE,$06,$3E,$69,$F4
	.dc.b	$10,$F8,$D8,$75,$7E,$46,$1D,$6B,$E4,$61,$CF,$6F,$91,$BB,$2C,$61
	.dc.b	$F8,$3C,$BF,$0D,$F0,$92,$3E,$EB,$F8,$9F,$84,$FC,$57,$C2,$7E,$2F
	.dc.b	$E1,$25,$7E,$0A,$3F,$1D,$F0,$92,$3F,$0A,$FE,$47,$E1,$3A,$47,$DD
	.dc.b	$C2,$ED,$F7,$71,$EB,$3F,$8C,$C7,$AB,$F8,$3C,$7A,$3F,$85,$C6,$DF
	.dc.b	$E1,$71,$FC,$F7,$C2,$4A,$FC,$AB,$FA,$1F,$84,$FD,$17,$C2,$7E,$8F
	.dc.b	$E1,$25,$7E,$6A,$3F,$C0,$3F,$BF,$FF,$9C,$4F,$F3,$5C,$7F,$8C,$1A
	.dc.b	$FF,$1F,$7F,$CE,$DC,$3F,$EA,$78,$FD,$66,$1E,$17,$F3,$58,$78,$1F
	.dc.b	$D3,$E1,$E0,$7F,$3D,$87,$83,$8B,$C3,$F0,$B2,$7C,$11,$E4,$F8,$49
	.dc.b	$1E,$37,$CF,$F0,$BE,$8F,$84,$93,$E7,$7D,$7F,$0B,$EC,$F8,$5F,$6F
	.dc.b	$C2,$CA,$F6,$47,$27,$A1,$0F,$75,$F4,$7E,$05,$DE,$2F,$E0,$BC,$2C
	.dc.b	$7B,$57,$C7,$C7,$B3,$F1,$F1,$EC,$F4,$3E,$66,$3F,$43,$E1,$7A,$69
	.dc.b	$FB,$DC,$59,$E2,$FF,$A9,$C7,$77,$D2,$77,$21,$DC,$FB,$1F,$0B,$F7
	.dc.b	$08,$FD,$BF,$85,$FB,$FF,$0B,$F8,$0F,$85,$93,$F7,$E3,$F0,$7F,$0B
	.dc.b	$23,$F0,$4F,$E1,$BE,$17,$F0,$F7,$27,$A1,$F8,$7F,$85,$EA,$FE,$27
	.dc.b	$0E,$AF,$E2,$70,$E8,$FE,$2F,$0B,$C5,$F3,$31,$FC,$9F,$C2,$4A,$FC
	.dc.b	$93,$F9,$6F,$85,$FC,$BF,$C2,$49,$FC,$B3,$F9,$AF,$85,$FC,$DF,$C2
	.dc.b	$FF,$97,$3A,$FF,$D2,$C7,$FA,$5F,$87,$FB,$FD,$FC,$CC,$3A,$3F,$9F
	.dc.b	$78,$7C,$2F,$58,$FD,$06,$37,$4F,$C0,$E5,$DA,$7F,$D2,$B8,$F6,$F9
	.dc.b	$0C,$47,$F8,$16,$1F,$0D,$E3,$F8,$69,$3C,$23,$CD,$F0,$BE,$7F,$84
	.dc.b	$93,$D2,$3D,$5F,$0D,$EB,$F8,$49,$3E,$08,$F7,$5C,$8E,$87,$93,$E1
	.dc.b	$7A,$BE,$6C,$3A,$BE,$5C,$3A,$BE,$AC,$3A,$BE,$9C,$3E,$55,$F5,$8F
	.dc.b	$66,$18,$7C,$C7,$52,$1D,$4F,$31,$D4,$87,$53,$E4,$FC,$27,$D7,$23
	.dc.b	$F3,$3E,$17,$ED,$7C,$24,$8F,$9E,$FD,$DF,$85,$FB,$DF,$09,$F7,$FE
	.dc.b	$16,$57,$D4,$8F,$C1,$5F,$48,$FA,$98,$61,$E6,$43,$EC,$61,$D5,$E4
	.dc.b	$7D,$EB,$EA,$FD,$DB,$FC,$77,$C2,$FE,$3F,$E1,$64,$FE,$12,$3F,$25
	.dc.b	$F0,$BF,$93,$F8,$59,$3F,$88,$8F,$CB,$7C,$24,$8F,$C5,$BF,$99,$F8
	.dc.b	$5F,$F5,$AF,$F0,$07,$F8,$63,$9F,$FD,$D5,$5F,$8D,$C7,$2F,$63,$F2
	.dc.b	$78,$F6,$7F,$25,$8F,$57,$F2,$98,$DE,$AF,$F0,$FE,$0C,$BB,$FC,$2C
	.dc.b	$BF,$F7,$8E,$7F,$E2,$5F,$2B,$FC,$24,$FF,$85,$23,$FC,$CE,$FF,$88
	.dc.b	$DD,$FC,$27,$A0,$F4,$FC,$1F,$AB,$E1,$24,$F4,$7D,$BF,$09,$EE,$F8
	.dc.b	$6F,$04,$FF,$9B,$70,$F0,$3D,$A3,$A3,$D8,$86,$EF,$AB,$C6,$EE,$BF
	.dc.b	$E3,$CB,$61,$F3,$3E,$16,$5F,$B9,$F9,$FF,$09,$DA,$3C,$59,$DC,$39
	.dc.b	$65,$D3,$4F,$87,$9D,$BF,$87,$97,$D8,$F8,$59,$1F,$29,$FB,$7F,$0B
	.dc.b	$F7,$3E,$17,$EE,$FC,$2C,$AF,$9F,$1F,$80,$F8,$59,$1F,$49,$FC,$1F
	.dc.b	$C2,$7E,$13,$E1,$79,$CF,$D6,$C3,$0F,$23,$EC,$E1,$D5,$F9,$98,$74
	.dc.b	$7E,$76,$17,$6F,$8D,$8F,$E4,$3E,$16,$57,$E0,$9F,$C9,$FC,$2F,$E5
	.dc.b	$3E,$13,$F2,$BF,$09,$2B,$F0,$F1,$F9,$8F,$84,$91,$F8,$A7,$F3,$7D
	.dc.b	$7F,$EA,$F8,$FF,$55,$7F,$00,$B9,$3A,$C7,$E4,$31,$D6,$9F,$E7,$4A
	.dc.b	$D4,$F6,$3F,$2B,$97,$85,$FC,$46,$5E,$07,$F1,$59,$5B,$FC,$16,$5D
	.dc.b	$32,$FF,$36,$63,$FE,$1E,$9E,$DF,$09,$9C,$3F,$D8,$73,$FE,$EB,$7F
	.dc.b	$D8,$D3,$E3,$F8,$5D,$7C,$3A,$9F,$07,$86,$DF,$0C,$7B,$CE,$7E,$0B
	.dc.b	$5F,$0D,$BF,$85,$97,$B3,$C3,$1C,$FE,$1B,$A7,$C3,$49,$D6,$7E,$19
	.dc.b	$F0,$63,$27,$39,$F8,$9F,$9F,$97,$81,$CE,$5E,$07,$E9,$E5,$E9,$F8
	.dc.b	$69,$1C,$9F,$67,$C3,$75,$8F,$B3,$8D,$DF,$EB,$63,$D6,$7E,$EE,$3D
	.dc.b	$67,$ED,$63,$F1,$7E,$13,$E3,$7C,$2C,$BF,$3C,$7C,$9F,$85,$91,$EA
	.dc.b	$7E,$67,$C2,$FC,$DF,$86,$F9,$DF,$0D,$2B,$DF,$1F,$47,$E1,$3E,$97
	.dc.b	$C1,$C9,$F8,$71,$D1,$FC,$8E,$17,$4F,$C7,$63,$D6,$7E,$4E,$3D,$5F
	.dc.b	$CB,$E3,$F7,$7E,$0E,$47,$CE,$7F,$01,$F0,$DF,$84,$FC,$4D,$F0,$B7
	.dc.b	$E1,$A5,$FD,$89,$CB,$7E,$03,$87,$C3,$48,$C1,$F6,$B0,$EA,$D3,$97
	.dc.b	$C2,$F7,$C7,$AC,$36,$FF,$6D,$E2,$CB,$53,$D0,$D7,$77,$FB,$36,$59
	.dc.b	$78,$67,$23,$B5,$3F,$E1,$C8,$F5,$7C,$37,$BF,$E1,$64,$60,$F7,$7C
	.dc.b	$37,$C3,$23,$F0,$7E,$1B,$E3,$7C,$24,$8F,$88,$FC,$9F,$85,$F9,$5F
	.dc.b	$0B,$F2,$F1,$95,$F2,$63,$9B,$E6,$CF,$CB,$8E,$37,$F8,$39,$75,$AD
	.dc.b	$68,$CF,$15,$F1,$B3,$FB,$1F,$0D,$E0,$BE,$C7,$D7,$7A,$6F,$F1,$F2
	.dc.b	$B3,$F1,$71,$65,$D2,$BE,$66,$56,$B5,$8F,$E1,$7E,$17,$F0,$DF,$0D
	.dc.b	$27,$F0,$91,$F8,$9F,$85,$FC,$57,$C3,$49,$FC,$44,$7E,$37,$E1,$7F
	.dc.b	$1D,$F0,$D2,$7F,$19,$1F,$91,$F8,$69,$1F,$8F,$7F,$29,$F0,$DD,$63
	.dc.b	$F2,$58,$75,$7E,$2E,$1D,$5F,$CB,$61,$D5,$F9,$38,$7F,$25,$FC,$F3
	.dc.b	$F9,$D4,$7F,$17,$4F,$F7,$86,$FC,$3E,$E3,$7C,$9F,$CE,$A7,$7F,$0F
	.dc.b	$92,$4F,$33,$FC,$18,$7E,$4A,$3A,$FC,$1F,$3F,$83,$93,$B8,$ED,$F0
	.dc.b	$72,$39,$3E,$1F,$86,$EB,$1C,$CF,$F0,$D5,$F8,$9F,$07,$C2,$F9,$EF
	.dc.b	$AC,$6F,$0C,$3E,$5F,$86,$F6,$11,$F4,$7C,$37,$BF,$E1,$7E,$07,$C2
	.dc.b	$49,$F6,$47,$C3,$F8,$5F,$89,$F0,$B2,$7E,$04,$7C,$7F,$85,$F9,$18
	.dc.b	$49,$F8,$91,$D5,$F8,$D8,$75,$7D,$98,$75,$7E,$56,$1D,$5F,$81,$87
	.dc.b	$D3,$F8,$69,$1F,$3D,$FA,$DF,$0D,$D6,$3E,$36,$1D,$78,$FE,$C4,$31
	.dc.b	$EC,$FC,$BC,$7B,$4E,$B1,$FC,$07,$C3,$7E,$07,$E1,$65,$FD,$D8,$FC
	.dc.b	$27,$C3,$7E,$17,$E1,$A4,$FE,$06,$3F,$11,$F0,$D2,$3F,$08,$FE,$47
	.dc.b	$F3,$DF,$F0,$1C,$7E,$15,$FE,$F5,$E2,$FE,$4B,$9D,$18,$94,$63,$E1
	.dc.b	$B9,$7C,$3F,$8A,$7F,$21,$97,$89,$FC,$36,$5E,$17,$F1,$19,$5D,$BF
	.dc.b	$35,$97,$89,$BF,$05,$E5,$F8,$29,$7F,$A5,$AF,$EE,$55,$C3,$E0,$7D
	.dc.b	$44,$9D,$7C,$17,$B7,$E1,$BD,$DF,$0F,$27,$94,$7C,$1F,$86,$91,$D5
	.dc.b	$F8,$9F,$0D,$CE,$3E,$26,$17,$1F,$3E,$3D,$67,$C9,$8F,$57,$E5,$63
	.dc.b	$F3,$7E,$17,$E7,$7C,$34,$9F,$54,$7D,$1F,$84,$FA,$5F,$09,$27,$DD
	.dc.b	$1F,$57,$E0,$FE,$B7,$C2,$49,$F8,$51,$F6,$7E,$17,$ED,$7C,$34,$9F
	.dc.b	$8D,$1D,$5F,$91,$87,$57,$EF,$61,$D5,$F9,$98,$75,$7F,$05,$87,$E1
	.dc.b	$7E,$1A,$47,$D1,$7F,$11,$F0,$BD,$63,$F1,$18,$75,$E0,$FB,$78,$F5
	.dc.b	$9F,$C6,$63,$D6,$7E,$FE,$3F,$C8,$5F,$EF,$33,$FC,$55,$A7,$F1,$D2
	.dc.b	$F6,$FE,$4D,$FE,$E5,$1F,$C9,$DF,$ED,$4F,$1F,$86,$E8,$75,$F8,$3E
	.dc.b	$DF,$0D,$DF,$E0,$E5,$71,$8F,$17,$C2,$48,$E8,$F9,$7E,$07,$9C,$77
	.dc.b	$C2,$BE,$7C,$3A,$BE,$3C,$3A,$BE,$BC,$3D,$DF,$01,$EF,$F8,$09,$3E
	.dc.b	$98,$F8,$5F,$01,$F0,$FE,$12,$4F,$B6,$3E,$37,$C1,$C8,$F8,$0F,$C9
	.dc.b	$F8,$1F,$95,$F0,$52,$7E,$23,$F3,$7D,$FE,$B1,$F1,$F0,$EA,$FC,$FC
	.dc.b	$3A,$3F,$47,$0B,$C3,$F4,$71,$FA,$DF,$01,$F5,$FD,$F9,$7F,$46,$3E
	.dc.b	$D7,$C0,$48,$FA,$8F,$DD,$F8,$2F,$BD,$F0,$5F,$D7,$FE,$4F,$FA,$CF
	.dc.b	$47,$F0,$D6,$DF,$F0,$CC,$FF,$DD,$4F,$F6,$38,$FE,$C4,$4F,$83,$E5
	.dc.b	$F0,$7D,$E7,$F1,$19,$5B,$FC,$46,$5E,$09,$E4,$78,$4C,$1F,$D8,$F0
	.dc.b	$FE,$CB,$3F,$DA,$D7,$C0,$79,$CF,$47,$C0,$F5,$CF,$F2,$58,$DC,$3D
	.dc.b	$18,$F4,$C7,$94,$77,$3F,$2D,$A7,$C1,$F8,$19,$1E,$D7,$E2,$7C,$1F
	.dc.b	$C5,$F8,$3F,$8D,$F0,$72,$BE,$14,$7C,$9F,$83,$91,$F1,$5F,$99,$EF
	.dc.b	$FC,$DF,$80,$F9,$DE,$FC,$AF,$95,$1D,$5F,$99,$87,$57,$E1,$61,$D1
	.dc.b	$F8,$98,$5D,$3E,$26,$3F,$67,$E0,$3E,$D7,$C1,$CB,$FA,$D1,$F7,$7E
	.dc.b	$0E,$47,$D9,$7F,$01,$F0,$7F,$81,$F8,$2F,$C1,$7C,$1C,$AF,$BD,$1F
	.dc.b	$85,$F7,$E4,$7E,$05,$FC,$47,$C0,$74,$8F,$B1,$85,$D3,$EC,$63,$D6
	.dc.b	$7F,$11,$8F,$57,$EF,$63,$F9,$1F,$80,$91,$F8,$D7,$FC,$65,$F0,$FF
	.dc.b	$D4,$7A,$7F,$B0,$3F,$D7,$FE,$13,$FD,$23,$F0,$7A,$37,$F0,$9C,$3E
	.dc.b	$12,$63,$FE,$39,$BF,$EA,$B8,$72,$F8,$4E,$BF,$03,$2B,$4F,$83,$E0
	.dc.b	$79,$C7,$1C,$BB,$3C,$F2,$EA,$FF,$8E,$71,$EA,$FF,$72,$6E,$7F,$EA
	.dc.b	$6C,$7D,$7F,$01,$2B,$C8,$FB,$BE,$07,$BC,$73,$C7,$BB,$D7,$1E,$FC
	.dc.b	$5F,$E4,$DC,$BA,$93,$1F,$8F,$F0,$72,$FE,$03,$F2,$BE,$0F,$E5,$FC
	.dc.b	$17,$CC,$F8,$39,$5F,$16,$3E,$7F,$BD,$23,$E4,$3F,$4B,$DF,$FA,$7E
	.dc.b	$F4,$9F,$98,$FD,$6F,$7F,$AC,$7C,$FC,$3A,$BF,$17,$0E,$8F,$C7,$C2
	.dc.b	$EB,$EB,$C7,$EF,$FC,$17,$E0,$3E,$0E,$5F,$D9,$8F,$C1,$FC,$1F,$E1
	.dc.b	$3E,$0E,$4F,$DD,$8F,$F3,$C7,$C3,$7F,$54,$8F,$F1,$1B,$79,$FF,$D7
	.dc.b	$DC,$7A,$F8,$7D,$FC,$1C,$BF,$F3,$F5,$FF,$58,$D7,$FC,$F4,$FF,$B2
	.dc.b	$23,$B1,$F8,$BC,$BC,$2F,$E0,$72,$F0,$E3,$23,$9B,$DD,$EB,$F0,$3E
	.dc.b	$6F,$82,$91,$83,$C3,$F0,$BE,$A2,$3E,$4F,$85,$F6,$FC,$1C,$8F,$3B
	.dc.b	$F0,$3E,$13,$E0,$FC,$0C,$9F,$5B,$F1,$3E,$0B,$9C,$7F,$A3,$F6,$63
	.dc.b	$8D,$FD,$F9,$75,$AD,$65,$67,$E1,$E5,$F3,$BE,$0A,$57,$C8,$7E,$8F
	.dc.b	$C1,$73,$F1,$BF,$CA,$D9,$F6,$AF,$69,$EB,$3D,$24,$3B,$9F,$4A,$FE
	.dc.b	$DF,$C2,$48,$FA,$AF,$DE,$F8,$4F,$BF,$F0,$7F,$80,$F8,$49,$5F,$6A
	.dc.b	$3F,$07,$EF,$FE,$13,$DF,$93,$F7,$A3,$F0,$FE,$FC,$8F,$C0,$BF,$EC
	.dc.b	$1E,$CF,$EC,$23,$46,$1D,$AB,$E5,$63,$E2,$7A,$1F,$8A,$C7,$BB,$F3
	.dc.b	$F1,$E9,$F0,$9D,$7E,$12,$4F,$FB,$37,$2F,$F6,$61,$FD,$83,$7F,$F3
	.dc.b	$F3,$63,$FC,$FA,$D3,$FC,$FF,$F0,$7E,$7F,$82,$93,$CD,$F5,$7C,$17
	.dc.b	$AF,$DF,$93,$DD,$F7,$7C,$07,$BF,$E0,$E4,$F8,$DF,$85,$87,$48,$FF
	.dc.b	$6B,$62,$C7,$1D,$1F,$ED,$F8,$65,$F2,$7E,$06,$5F,$B5,$F9,$9F,$03
	.dc.b	$F3,$7E,$0B,$E7,$7C,$1C,$AF,$85,$1F,$47,$E0,$BE,$97,$C1,$C9,$F8
	.dc.b	$D1,$F5,$7E,$07,$EB,$7C,$0C,$9F,$95,$1D,$5F,$99,$87,$57,$D3,$87
	.dc.b	$57,$E8,$61,$D5,$F6,$E1,$F8,$1F,$80,$91,$F5,$5F,$C2,$7C,$0F,$58
	.dc.b	$F8,$78,$75,$DF,$E1,$63,$D6,$7E,$3E,$3D,$67,$E3,$63,$F8,$DF,$83
	.dc.b	$FC,$77,$C1,$CB,$FC,$14,$7E,$47,$E0,$E4,$7E,$15,$32,$DD,$1F,$EC
	.dc.b	$3F,$CF,$7F,$B3,$4F,$FB,$09,$FE,$CC,$3D,$FE,$30,$AE,$2E,$E3,$31
	.dc.b	$8E,$55,$72,$F8,$4C,$E1,$DD,$B8,$77,$6D,$70,$F7,$FD,$33,$F8,$8C
	.dc.b	$AE,$3D,$71,$F3,$DF,$6F,$80,$E3,$1E,$7B,$F7,$65,$ED,$CF,$35,$AF
	.dc.b	$84,$95,$A3,$E0,$3C,$BE,$13,$9F,$C2,$49,$D6,$5E,$C8,$EF,$F0,$9E
	.dc.b	$0F,$84,$F0,$FC,$24,$BD,$EA,$9F,$2F,$C0,$C9,$CB,$E8,$F8,$2E,$D1
	.dc.b	$F3,$77,$FA,$99,$63,$83,$E0,$67,$DE,$BA,$E7,$DD,$E0,$7D,$9C,$BE
	.dc.b	$1F,$C1,$48,$F1,$BF,$1B,$E0,$BE,$3F,$C2,$7C,$8F,$84,$95,$E8,$8F
	.dc.b	$97,$F0,$7F,$33,$E1,$24,$FB,$23,$E7,$FC,$0C,$8F,$7B,$F4,$BE,$0B
	.dc.b	$AC,$7C,$2C,$3A,$BE,$6E,$4F,$C2,$E3,$DA,$3E,$3E,$3D,$9D,$72,$32
	.dc.b	$FC,$2F,$17,$F7,$33,$FD,$8F,$3C,$DB,$C3,$F9,$17,$FB,$C5,$30,$F0
	.dc.b	$CF,$E2,$72,$F2,$DF,$6C,$3C,$13,$F8,$DC,$7C,$93,$D7,$E0,$FC,$24
	.dc.b	$7F,$07,$97,$F7,$D7,$C5,$F0,$5E,$6F,$82,$91,$A8,$F4,$FC,$07,$AB
	.dc.b	$E0,$24,$F2,$8F,$6F,$C0,$7B,$BD,$E9,$3D,$A3,$E0,$DF,$48,$F8,$18
	.dc.b	$61,$E4,$79,$30,$EA,$FC,$5C,$3E,$4F,$C1,$48,$F4,$BF,$33,$E0,$BE
	.dc.b	$6F,$C1,$49,$F6,$BF,$43,$E0,$BE,$8F,$C1,$49,$F8,$2F,$D4,$F8,$2F
	.dc.b	$AB,$F0,$1F,$5B,$E0,$25,$7C,$68,$E8,$FD,$7C,$2F,$27,$D7,$C7,$A4
	.dc.b	$FD,$BC,$6C,$FD,$BC,$7F,$03,$EF,$FE,$0B,$E0,$25,$FD,$28,$FC,$2F
	.dc.b	$C1,$48,$FA,$AF,$F0,$CF,$CF,$7F,$B9,$8F,$82,$C1,$27,$F7,$B0,$F8
	.dc.b	$2E,$1F,$01,$C7,$E0,$25,$6A,$3B,$3F,$91,$C2,$E9,$FC,$87,$0E,$F8
	.dc.b	$F5,$AF,$D0,$47,$85,$EF,$8F,$97,$E0,$3C,$DF,$01,$27,$C7,$1D,$3B
	.dc.b	$3F,$9C,$F8,$7F,$CE,$33,$C5,$21,$8F,$82,$3D,$79,$78,$73,$F2,$67
	.dc.b	$F0,$FE,$0B,$E2,$7C,$14,$BF,$83,$1F,$1F,$E0,$BE,$47,$C1,$49,$F8
	.dc.b	$B1,$F2,$FD,$D9,$1F,$21,$F9,$DE,$F7,$CF,$F7,$7E,$87,$BD,$2B,$E6
	.dc.b	$C7,$57,$E7,$E1,$D5,$F8,$B8,$74,$7E,$3E,$17,$4F,$99,$8F,$DB,$F8
	.dc.b	$19,$5F,$61,$FB,$DF,$05,$F7,$FE,$0B,$F0,$1F,$05,$2B,$EE,$C7,$E0
	.dc.b	$FE,$07,$F0,$9F,$05,$27,$F0,$31,$F8,$7F,$80,$FE,$A7,$F8,$9F,$F4
	.dc.b	$E3,$FA,$7F,$32,$3B,$57,$E2,$31,$EC,$FD,$EC,$7B,$3F,$8C,$C7,$B3
	.dc.b	$F8,$2C,$7A,$FC,$04,$8F,$E9,$B9,$7F,$B5,$4F,$EB,$4F,$F2,$77,$FD
	.dc.b	$86,$3E,$0B,$CA,$79,$BE,$0F,$CF,$F0,$5E,$8F,$82,$97,$D2,$3D,$7F
	.dc.b	$01,$EC,$F8,$09,$3E,$08,$E4,$F8,$B0,$EA,$F9,$30,$EA,$F9,$B0,$EA
	.dc.b	$F9,$70,$F8,$FE,$FC,$8F,$5B,$F2,$BE,$03,$AC,$7A,$F0,$EA,$FB,$70
	.dc.b	$EA,$FB,$F0,$EB,$CD,$FE,$0B,$D1,$97,$D4,$F8,$2F,$AB,$F0,$53,$0F
	.dc.b	$93,$1F,$63,$E0,$A4,$7C,$C7,$ED,$FC,$17,$DC,$F7,$FE,$EF,$BF,$2B
	.dc.b	$E8,$C7,$E0,$3D,$FF,$C0,$FB,$F2,$7E,$AC,$74,$7E,$A6,$17,$5F,$36
	.dc.b	$3D,$67,$ED,$E3,$D5,$FB,$58,$FE,$33,$E0,$A4,$7E,$01,$FC,$7F,$C1
	.dc.b	$7F,$88,$BE,$73,$FD,$A0,$7F,$5C,$F3,$7F,$80,$B6,$BF,$80,$3A,$F8
	.dc.b	$2E,$1F,$13,$27,$FC,$2F,$3F,$ED,$3A,$E3,$F0,$3D,$49,$3A,$F8,$1F
	.dc.b	$07,$C3,$F8,$7E,$22,$4F,$28,$F2,$65,$D6,$3F,$BD,$63,$FD,$87,$83
	.dc.b	$FC,$80,$F7,$8F,$2F,$C1,$7A,$C8,$F9,$3E,$0F,$DD,$F0,$92,$3C,$EF
	.dc.b	$C1,$F8,$5F,$85,$F0,$B2,$7D,$6F,$C5,$F8,$5F,$8D,$F0,$72,$7D,$EF
	.dc.b	$C9,$F8,$3E,$91,$E6,$C2,$E9,$E6,$C7,$A4,$FA,$B1,$B3,$EA,$C7,$E9
	.dc.b	$7C,$1C,$AF,$96,$FD,$5F,$83,$E7,$C9,$F5,$32,$B5,$F5,$32,$E7,$3F
	.dc.b	$63,$2B,$3F,$63,$1F,$BD,$F0,$9F,$7F,$E1,$64,$FD,$88,$FC,$17,$C2
	.dc.b	$48,$FB,$6F,$E1,$7E,$17,$F0,$DF,$03,$FE,$68,$F8,$0F,$F6,$79,$FE
	.dc.b	$D0,$E5,$46,$62,$5F,$F9,$69,$FE,$A4,$FF,$8C,$DF,$EC,$0F,$62,$1F
	.dc.b	$7A,$FC,$0F,$E3,$27,$C0,$C3,$07,$5F,$84,$F2,$11,$F0,$7C,$2F,$9F
	.dc.b	$E1,$24,$78,$DF,$57,$C2,$FA,$FE,$0B,$D9,$F0,$72,$BD,$11,$EF,$F8
	.dc.b	$0F,$81,$F0,$32,$7D,$91,$D1,$F6,$61,$7A,$7F,$D2,$9B,$7F,$B3,$A7
	.dc.b	$99,$F0,$F1,$EC,$FC,$3C,$7E,$67,$C0,$FC,$DF,$81,$93,$F2,$63,$E8
	.dc.b	$7C,$17,$D1,$F8,$39,$3F,$36,$3E,$A7,$C1,$C8,$FA,$0F,$D7,$F8,$3F
	.dc.b	$B1,$F0,$12,$7E,$A3,$F6,$FE,$07,$AC,$7D,$7C,$3A,$BF,$5F,$0E,$8F
	.dc.b	$D9,$C2,$ED,$EE,$C7,$F0,$9F,$01,$F8,$5F,$81,$97,$F8,$18,$EA,$FE
	.dc.b	$07,$1E,$B7,$F1,$B1,$E7,$DF,$FE,$10,$E1,$67,$8C,$3F,$C2,$10,$CF
	.dc.b	$F2,$5F,$05,$27,$F1,$CF,$FA,$F3,$F3,$DF,$F5,$B8,$F8,$3C,$12,$7F
	.dc.b	$D5,$B1,$FD,$83,$3F,$F6,$52,$F7,$B9,$7B,$F2,$B7,$1D,$7D,$DE,$DE
	.dc.b	$FC,$9E,$71,$D5,$FF,$48,$5F,$85,$EF,$8F,$97,$E2,$3C,$D8,$75,$9D
	.dc.b	$E3,$87,$CB,$F0,$7E,$BF,$84,$91,$83,$D3,$F0,$9E,$F2,$3E,$CF,$84
	.dc.b	$F8,$5F,$05,$23,$DE,$FC,$5F,$83,$F8,$DF,$03,$F1,$FE,$06,$57,$C4
	.dc.b	$8E,$6F,$AB,$0B,$C5,$F1,$B1,$E9,$3E,$EC,$6C,$FC,$AC,$7E,$97,$C0
	.dc.b	$7D,$3F,$81,$97,$F4,$23,$EB,$7C,$14,$8F,$A6,$FD,$9F,$83,$FB,$5F
	.dc.b	$07,$27,$EB,$BF,$77,$E0,$FF,$DF,$3F,$9E,$BE,$02,$FC,$0C,$BF,$F7
	.dc.b	$4C,$E5,$EA,$43,$EA,$E3,$DD,$EA,$43,$EC,$E1,$DD,$E7,$F0,$3E,$02
	.dc.b	$3D,$BE,$13,$D4,$F4,$3B,$6F,$F8,$CD,$3D,$51,$D0,$ED,$1F,$90,$D3
	.dc.b	$D7,$F0,$9E,$CF,$84,$95,$EA,$8F,$7F,$C2,$48,$F6,$BF,$0B,$E1,$3E
	.dc.b	$1F,$C0,$49,$F8,$2F,$C6,$F8,$0F,$8F,$F0,$1F,$23,$E0,$25,$7C,$68
	.dc.b	$E8,$FF,$B9,$F8,$98,$E3,$57,$95,$FE,$84,$96,$7D,$EF,$86,$76,$F7
	.dc.b	$1F,$E8,$7A,$67,$F6,$3E,$13,$EC,$FC,$24,$AF,$AF,$1F,$73,$E1,$24
	.dc.b	$7D,$A7,$EF,$FC,$27,$E0,$3E,$0E,$4F,$DE,$7F,$07,$F0,$9F,$84,$F8
	.dc.b	$2F,$C2,$FC,$24,$AF,$C1,$C7,$47,$DD,$85,$D7,$CF,$F1,$0D,$99,$F6
	.dc.b	$BF,$89,$9D,$BD,$D3,$3F,$C9,$FC,$24,$BF,$C8,$A6,$57,$6F,$CF,$71
	.dc.b	$18,$E3,$FF,$57,$C7,$F6,$69,$D7,$A9,$A6,$D6,$CF,$F5,$94,$FF,$6A
	.dc.b	$51,$D7,$E1,$78,$5F,$4F,$85,$B5,$C1,$BB,$30,$5E,$17,$C1,$52,$5F
	.dc.b	$86,$B8,$7C,$2E,$F3,$93,$9B,$F5,$65,$EC,$F8,$9E,$99,$75,$CE,$5E
	.dc.b	$E3,$C1,$F0,$7E,$1F,$84,$93,$CE,$3A,$BE,$EC,$3E,$3E,$58,$DB,$E6
	.dc.b	$65,$DA,$BC,$59,$76,$78,$1F,$23,$2F,$77,$C2,$FB,$FE,$1A,$4F,$A6
	.dc.b	$3E,$17,$C2,$FC,$3F,$86,$93,$ED,$8F,$8D,$F0,$32,$3E,$03,$F2,$7E
	.dc.b	$0B,$E5,$7C,$17,$CB,$F8,$29,$5F,$16,$3A,$BF,$1F,$0E,$AF,$9B,$97
	.dc.b	$EA,$63,$D6,$31,$C8,$CB,$1C,$1F,$8B,$CF,$EC,$7C,$34,$BF,$A4,$FD
	.dc.b	$BF,$86,$FB,$9F,$0B,$27,$EB,$3F,$7F,$E1,$BF,$01,$F0,$D2,$7E,$D3
	.dc.b	$F8,$6F,$53,$F7,$71,$E5,$7F,$AE,$37,$E0,$F4,$6F,$E0,$FB,$47,$A3
	.dc.b	$0B,$A7,$E8,$71,$ED,$3E,$CC,$6C,$FE,$CB,$1E,$ED,$F8,$3F,$17,$C1
	.dc.b	$CB,$FC,$95,$7F,$77,$6F,$C2,$F9,$FE,$1A,$46,$0D,$FC,$2F,$AC,$8F
	.dc.b	$2F,$85,$F7,$7C,$1C,$8E,$AF,$C1,$F8,$3F,$85,$87,$49,$F8,$58,$61
	.dc.b	$EA,$43,$E3,$61,$F2,$70,$F9,$5F,$07,$F2,$FE,$0E,$57,$AA,$39,$3D
	.dc.b	$0F,$6D,$E3,$6F,$89,$87,$A2,$7A,$9E,$93,$E1,$FC,$27,$D7,$F8,$59
	.dc.b	$18,$3E,$3F,$C2,$7D,$BF,$85,$91,$83,$E5,$FC,$27,$DF,$F8,$19,$18
	.dc.b	$3E,$7F,$C2,$7E,$0F,$E0,$A4,$60,$E4,$7E,$16,$FA,$BF,$87,$C3,$F1
	.dc.b	$58,$74,$8F,$C5,$61,$87,$F8,$67,$E7,$BF,$DD,$DF,$D1,$3F,$ED,$1F
	.dc.b	$0C,$D3,$5B,$7F,$38,$7F,$D8,$51,$AF,$84,$FE,$3B,$1F,$EC,$7F,$3B
	.dc.b	$F9,$94,$F2,$F8,$5C,$FC,$26,$BE,$16,$65,$E2,$8E,$3F,$07,$23,$CA
	.dc.b	$F4,$F8,$3E,$B7,$DB,$39,$5E,$A8,$EA,$FB,$30,$EA,$FB,$30,$F2,$FC
	.dc.b	$1F,$9B,$E0,$E4,$FC,$28,$F4,$FC,$27,$AB,$E1,$64,$FC,$68,$F6,$FC
	.dc.b	$24,$8F,$92,$FC,$0F,$85,$F8,$3F,$07,$F0,$BE,$0E,$57,$CE,$8E,$AF
	.dc.b	$D0,$C3,$AB,$F4,$30,$E8,$FD,$2C,$2F,$27,$CF,$C7,$E6,$FC,$1F,$CE
	.dc.b	$F8,$29,$7F,$6A,$3A,$BF,$6B,$1E,$B7,$F5,$F1,$E7,$CB,$F5,$39,$99
	.dc.b	$E3,$07,$37,$E0,$34,$FB,$7F,$09,$27,$F0,$EF,$DE,$F8,$5F,$BF,$F0
	.dc.b	$92,$7F,$16,$FF,$5E,$FA,$1F,$ED,$F1,$FC,$B5,$FE,$AA,$F6,$7F,$B5
	.dc.b	$6B,$FE,$02,$6E,$9F,$88,$77,$91,$FE,$CF,$5F,$D8,$D7,$BF,$DC,$F0
	.dc.b	$7B,$FE,$39,$FC,$DE,$37,$1E,$D1,$E5,$F8,$1F,$37,$C0,$C8,$E9,$1E
	.dc.b	$9F,$85,$91,$DD,$F6,$7C,$2F,$B7,$E1,$24,$F8,$DF,$81,$F0,$BF,$07
	.dc.b	$E1,$3E,$17,$C2,$CA,$F4,$47,$C5,$F8,$2F,$8D,$F0,$72,$7D,$91,$D1
	.dc.b	$F9,$18,$5D,$7E,$36,$3D,$67,$E1,$63,$D5,$F9,$D8,$FD,$1F,$82,$FA
	.dc.b	$5F,$05,$27,$E4,$47,$D5,$F8,$49,$1F,$2D,$FB,$1F,$0B,$F6,$7E,$12
	.dc.b	$4F,$CF,$7E,$E7,$C2,$7D,$DF,$82,$FB,$DF,$05,$2B,$EA,$47,$47,$F0
	.dc.b	$18,$5D,$BE,$F6,$3D,$27,$F0,$98,$D9,$FC,$16,$3F,$89,$F8,$29,$5F
	.dc.b	$7D,$FF,$14,$FE,$7B,$FD,$5B,$B7,$F0,$D9,$75,$6D,$7E,$17,$2E,$91
	.dc.b	$F8,$BC,$AC,$FE,$47,$1E,$7F,$0B,$23,$FC,$5D,$8F,$F6,$EC,$FF,$CA
	.dc.b	$AF,$FB,$5F,$0F,$E3,$6F,$F6,$19,$FF,$04,$2F,$83,$F2,$9E,$6F,$84
	.dc.b	$F3,$FC,$0F,$A3,$DE,$97,$DA,$3D,$7E,$F4,$8F,$0B,$EE,$F8,$1E,$51
	.dc.b	$E4,$C3,$AB,$E6,$C3,$AB,$E8,$C3,$AB,$E4,$C3,$E3,$FC,$1C,$8F,$6B
	.dc.b	$F2,$BE,$0F,$E5,$FC,$17,$CC,$F8,$39,$5F,$0A,$3E,$7F,$C1,$C8,$F8
	.dc.b	$AF,$D2,$F8,$3F,$A7,$F0,$1F,$53,$E0,$65,$7C,$A8,$EA,$FC,$CC,$3A
	.dc.b	$BF,$23,$0E,$8F,$CA,$C2,$F1,$FD,$0C,$7E,$FF,$C0,$FE,$03,$E0,$65
	.dc.b	$FD,$68,$FC,$1F,$C1,$FE,$13,$E0,$E4,$FD,$A8,$FF,$37,$F4,$7F,$58
	.dc.b	$E8,$FE,$A6,$FF,$5D,$27,$C4,$6B,$E0,$77,$F0,$53,$0F,$F3,$85,$7F
	.dc.b	$5E,$EB,$63,$DA,$7F,$03,$97,$89,$DE,$3E,$0F,$85,$91,$83,$9F,$C1
	.dc.b	$79,$08,$F6,$F8,$3E,$4F,$43,$F2,$77,$8E,$1F,$F5,$3E,$7E,$B9,$FE
	.dc.b	$6E,$FA,$F0,$F7,$E3,$F0,$3E,$16,$57,$A9,$F8,$7F,$0B,$F1,$3E,$12
	.dc.b	$4F,$B9,$F8,$FF,$0B,$F2,$3E,$03,$E4,$FC,$0C,$AF,$87,$1F,$33,$E0
	.dc.b	$3E,$6F,$C0,$49,$F8,$F1,$D1,$FF,$36,$6A,$C7,$1B,$7A,$B2,$ED,$5F
	.dc.b	$3B,$2E,$CF,$0C,$BE,$CF,$C2,$48,$FA,$6F,$DC,$F8,$5F,$BB,$F0,$92
	.dc.b	$7E,$BB,$F8,$0F,$85,$FC,$0F,$C2,$49,$FB,$6F,$E1,$3E,$13,$F0,$BF
	.dc.b	$01,$27,$EF,$BF,$EB,$6F,$CF,$7F,$AC,$ED,$F4,$31,$ED,$5E,$C7,$B3
	.dc.b	$D0,$87,$C0,$C3,$B3,$FE,$B9,$AF,$EB,$67,$F5,$E7,$FB,$7B,$7E,$14
	.dc.b	$EB,$E0,$FC,$5F,$09,$2B,$94,$79,$7E,$0A,$47,$57,$D1,$F0,$5E,$9F
	.dc.b	$80,$93,$E1,$7D,$9F,$01,$ED,$BE,$93,$FE,$A1,$C7,$FB,$5B,$D8,$E8
	.dc.b	$43,$58,$76,$7D,$9E,$FF,$C7,$F8,$09,$18,$39,$1D,$73,$EF,$BF,$FC
	.dc.b	$00,$C3,$FB,$DC,$F4,$3B,$46,$70,$ED,$1E,$5F,$81,$FA,$84,$7E,$67
	.dc.b	$C1,$7D,$7F,$81,$FB,$1F,$05,$27,$E8,$47,$DB,$F7,$E4,$7D,$37,$EF
	.dc.b	$7B,$FF,$7F,$DF,$FC,$07,$BF,$2B,$EC,$47,$57,$ED,$61,$D5,$F9,$18
	.dc.b	$74,$7E,$56,$17,$4F,$76,$3F,$8B,$F8,$1F,$C6,$7C,$14,$BF,$C2,$47
	.dc.b	$E3,$FE,$09,$1F,$EF,$FF,$CF,$7F,$AF,$3F,$E3,$46,$F4,$FE,$1A,$77
	.dc.b	$3F,$ED,$99,$FE,$B1,$E9,$FF,$9C,$1B,$1C,$7C,$6D,$E6,$AD,$7C,$46
	.dc.b	$1E,$1F,$03,$6F,$C0,$DC,$98,$FF,$61,$C3,$95,$79,$7E,$0B,$D4,$FA
	.dc.b	$70,$E1,$5E,$8E,$76,$98,$D2,$5F,$BD,$DE,$9C,$3E,$07,$8F,$C1,$4A
	.dc.b	$F8,$91,$D3,$E1,$24,$7C,$77,$BF,$C2,$78,$3E,$12,$4F,$CB,$7C,$7F
	.dc.b	$09,$E4,$F8,$1F,$2F,$C1,$4A,$FA,$11,$D1,$F9,$98,$5E,$2F,$7F,$1F
	.dc.b	$C2,$CB,$AD,$7D,$1C,$AD,$6B,$85,$9F,$C2,$F8,$2F,$87,$F0,$53,$0F
	.dc.b	$BD,$1D,$79,$3E,$C6,$96,$7F,$01,$9F,$49,$FB,$99,$D9,$FC,$26,$3F
	.dc.b	$3B,$E0,$FE,$7F,$C2,$49,$FC,$54,$7D,$2F,$83,$91,$F8,$D7,$EA,$FC
	.dc.b	$27,$D6,$F8,$1F,$AF,$EF,$CA,$FC,$94,$7E,$07,$F3,$DF,$EA,$1C,$66
	.dc.b	$93,$E0,$3A,$D7,$E4,$F1,$BC,$9F,$0F,$2E,$D3,$F9,$DC,$BB,$3F,$99
	.dc.b	$CB,$B7,$C2,$F7,$F8,$59,$3F,$A4,$C3,$E8,$E7,$FE,$D0,$1F,$09,$E5
	.dc.b	$24,$EF,$8B,$FD,$A3,$2F,$84,$F5,$7C,$1C,$AD,$1F,$EA,$09,$47,$BF
	.dc.b	$E0,$FE,$07,$C0,$FC,$1F,$81,$97,$C3,$8F,$E1,$4F,$82,$1A,$79,$B4
	.dc.b	$C3,$E8,$CF,$C1,$3A,$D3,$CB,$9E,$2B,$DB,$9F,$CF,$F8,$0F,$A1,$F0
	.dc.b	$33,$0F,$06,$C9,$FA,$9F,$05,$F5,$7E,$0E,$56,$A3,$EC,$7C,$1C,$8E
	.dc.b	$2F,$DB,$F8,$3F,$B9,$F0,$1F,$77,$E0,$E5,$76,$8E,$AF,$83,$0E,$AF
	.dc.b	$A4,$F9,$18,$75,$71,$C7,$F8,$5C,$71,$B7,$D8,$CB,$F1,$5F,$03,$F8
	.dc.b	$BF,$83,$98,$7B,$23,$BB,$CB,$66,$7E,$0C,$3F,$03,$9F,$4E,$2F,$BB
	.dc.b	$9D,$C3,$ED,$67,$F9,$7F,$83,$91,$F2,$1F,$E3,$5F,$0F,$FD,$F2,$FF
	.dc.b	$97,$EC,$8F,$F6,$79,$3E,$1F,$6F,$0F,$82,$98,$7E,$06,$3F,$AC,$E9
	.dc.b	$FC,$11,$FE,$D3,$1F,$09,$7B,$FE,$0F,$85,$93,$C6,$3C,$78,$48,$E8
	.dc.b	$F8,$1E,$D8,$78,$1E,$8F,$AB,$E0,$7D,$7F,$0B,$EC,$F8,$59,$5E,$58
	.dc.b	$F7,$FC,$2C,$8F,$43,$F0,$BE,$17,$E1,$FC,$2C,$9F,$63,$F1,$BE,$17
	.dc.b	$E3,$FC,$17,$C8,$F8,$39,$5F,$06,$3A,$3E,$FC,$2F,$0F,$C3,$C7,$A4
	.dc.b	$FC,$3C,$6C,$FC,$7C,$7E,$9F,$C1,$FD,$4F,$83,$97,$F3,$63,$EB,$FC
	.dc.b	$27,$D8,$F8,$59,$3F,$46,$3E,$DF,$C2,$7D,$CF,$85,$93,$F5,$63,$EF
	.dc.b	$FC,$17,$F5,$9F,$CF,$7F,$B4,$9F,$DA,$BA,$7F,$A9,$D7,$33,$EE,$63
	.dc.b	$D9,$FB,$58,$F6,$7F,$01,$8F,$67,$EF,$63,$D7,$E0,$E4,$7F,$05,$C7
	.dc.b	$FB,$8F,$59,$FF,$65,$42,$3F,$21,$87,$7D,$FF,$15,$97,$C0,$AF,$51
	.dc.b	$E0,$21,$E0,$3B,$7C,$37,$B7,$E1,$A4,$60,$F1,$7C,$1F,$C1,$23,$E5
	.dc.b	$F8,$4F,$89,$EF,$C8,$F4,$3F,$1F,$DF,$F9,$1E,$FF,$C9,$F7,$E5,$7B
	.dc.b	$63,$9B,$FC,$03,$76,$38,$D7,$FB,$3C,$32,$ED,$5B,$CA,$D6,$B2,$FA
	.dc.b	$DF,$07,$2B,$E5,$3F,$67,$E1,$3E,$D7,$C1,$7D,$BF,$83,$95,$F3,$E3
	.dc.b	$EF,$7C,$14,$8F,$A4,$FE,$07,$E0,$FF,$05,$F0,$52,$7E,$B3,$F8,$5F
	.dc.b	$82,$EB,$1F,$67,$0E,$AF,$B7,$0E,$8F,$BF,$0B,$A7,$BB,$1F,$C7,$7C
	.dc.b	$0F,$E3,$FE,$0A,$5F,$E0,$E3,$FC,$55,$F9,$EF,$F6,$A1,$F0,$98,$33
	.dc.b	$F1,$1A,$F8,$59,$7F,$D3,$E9,$1C,$BE,$17,$9F,$C1,$F4,$F8,$39,$7B
	.dc.b	$8E,$6F,$1C,$7B,$3C,$F1,$EA,$FF,$38,$C7,$AB,$DF,$1E,$CF,$87,$E0
	.dc.b	$FD,$24,$7C,$9F,$07,$D9,$E5,$87,$67,$A6,$1D,$B7,$FF,$14,$E3,$DA
	.dc.b	$7C,$18,$F6,$7C,$5F,$09,$F1,$BE,$16,$47,$C1,$8F,$93,$F0,$9F,$2B
	.dc.b	$E1,$64,$FC,$58,$F9,$BE,$FC,$8F,$90,$FD,$0F,$80,$FA,$3F,$01,$F4
	.dc.b	$BE,$02,$57,$CD,$8E,$AF,$CF,$C3,$AB,$F1,$70,$E8,$FC,$7C,$2E,$BE
	.dc.b	$CC,$7E,$EF,$C2,$7D,$EF,$85,$97,$F6,$63,$F0,$3F,$09,$F8,$2F,$85
	.dc.b	$93,$F7,$63,$F0,$BF,$09,$23,$F0,$0F,$F9,$C7,$F3,$DF,$EB,$11,$FC
	.dc.b	$25,$7C,$16,$7D,$59,$3F,$E1,$59,$CB,$D4,$FC,$56,$1E,$07,$F0,$18
	.dc.b	$77,$7F,$05,$85,$D3,$FC,$BF,$7E,$0F,$48,$FF,$60,$F8,$5F,$ED,$2F
	.dc.b	$F6,$09,$DF,$C1,$79,$FE,$0F,$D1,$F0,$9E,$9F,$85,$99,$F9,$63,$D9
	.dc.b	$F0,$9E,$DF,$85,$93,$E9,$8F,$81,$F0,$72,$3D,$8F,$C3,$F8,$3E,$B1
	.dc.b	$EF,$C3,$AB,$FE,$12,$D3,$FB,$8B,$CC,$F8,$98,$76,$75,$87,$CD,$F8
	.dc.b	$3F,$9D,$F0,$72,$7E,$54,$76,$7A,$61,$DB,$6F,$8B,$A7,$AA,$7A,$90
	.dc.b	$EA,$7D,$3F,$84,$FB,$44,$7E,$B7,$C2,$FD,$DF,$84,$91,$F6,$5F,$C0
	.dc.b	$7C,$2F,$E0,$7E,$06,$4F,$DD,$7F,$09,$F0,$3F,$85,$F8,$39,$3F,$81
	.dc.b	$7F,$11,$F0,$9D,$23,$E2,$61,$74,$FB,$B8,$F4,$9F,$91,$8D,$9F,$C0
	.dc.b	$E3,$F9,$1F,$85,$FF,$61,$79,$3F,$ED,$23,$FD,$A7,$C9,$FD,$AB,$0F
	.dc.b	$F2,$0C,$7F,$60,$3F,$AF,$F5,$7F,$18,$F8,$8E,$3F,$0D,$CB,$E1,$A5
	.dc.b	$FF,$9A,$EF,$FD,$99,$DB,$FE,$34,$5F,$0D,$E0,$F8,$49,$5B,$7C,$7F
	.dc.b	$07,$D2,$39,$65,$E6,$30,$F9,$31,$EE,$D3,$C1,$7E,$CF,$83,$F6,$FC
	.dc.b	$1C,$9F,$24,$7C,$0F,$85,$91,$E7,$7E,$1F,$C3,$7C,$4F,$85,$93,$EB
	.dc.b	$7E,$3F,$C2,$FC,$8F,$82,$F9,$3F,$07,$2B,$E0,$47,$47,$E4,$E1,$79
	.dc.b	$FF,$D9,$FC,$3F,$EC,$A9,$E8,$43,$E7,$E3,$D9,$F9,$5F,$07,$F5,$BE
	.dc.b	$0E,$46,$0F,$41,$0E,$E7,$D1,$D3,$D2,$F4,$3B,$61,$E2,$D3,$F0,$3F
	.dc.b	$0B,$27,$EC,$3F,$84,$F8,$5F,$C2,$FC,$24,$9F,$B8,$FE,$23,$E1,$51
	.dc.b	$FE,$A9,$FC,$F7,$FD,$96,$4F,$80,$CC,$7F,$A2,$A3,$FB,$2B,$BD,$F3
	.dc.b	$9F,$FB,$CA,$5F,$0D,$8E,$BF,$F3,$DC,$62,$F8,$1E,$22,$1F,$96,$CB
	.dc.b	$C8,$F8,$F1,$E4,$FF,$8D,$67,$C8,$F9,$FE,$0B,$3D,$C9,$FF,$68,$CF
	.dc.b	$AF,$E1,$73,$18,$8F,$57,$C3,$63,$E1,$A4,$7C,$17,$7F,$0D,$C3,$E1
	.dc.b	$24,$FC,$57,$9F,$C2,$F4,$F8,$2E,$BF,$07,$2B,$E5,$47,$47,$E4,$61
	.dc.b	$78,$BE,$36,$3D,$27,$E6,$63,$67,$E5,$63,$E8,$F8,$3F,$4F,$C1,$CB
	.dc.b	$FA,$D1,$EC,$F8,$4F,$6F,$C2,$C9,$FB,$51,$F0,$3E,$13,$E0,$FC,$2C
	.dc.b	$9F,$BD,$1F,$13,$E0,$A4,$7E,$05,$F8,$FF,$07,$D6,$3F,$09,$87,$57
	.dc.b	$F0,$58,$75,$7F,$11,$87,$57,$F0,$D8,$7D,$0F,$83,$FA,$BF,$9E,$FF
	.dc.b	$57,$3F,$AC,$72,$A3,$B5,$7E,$37,$1E,$DB,$FE,$2F,$2E,$D3,$F9,$1C
	.dc.b	$BB,$4F,$E3,$F2,$EB,$F0,$D2,$BF,$37,$2E,$CF,$F6,$5E,$9F,$DC,$27
	.dc.b	$FD,$C6,$FF,$B3,$5D,$C6,$BE,$13,$CE,$48,$C9,$FE,$85,$9F,$ED,$93
	.dc.b	$D3,$DD,$F6,$FB,$D2,$B7,$D8,$BF,$EB,$F1,$E0,$F8,$1F,$87,$F0,$52
	.dc.b	$37,$29,$F0,$47,$9B,$2B,$9F,$AE,$3C,$5F,$0B,$F3,$33,$F0,$47,$B3
	.dc.b	$3C,$3D,$3E,$17,$E8,$FC,$2C,$8C,$1E,$0F,$84,$FA,$BF,$0B,$23,$07
	.dc.b	$93,$E1,$7E,$CF,$C2,$C8,$C1,$E8,$F8,$3F,$BA,$47,$D7,$F0,$7F,$80
	.dc.b	$C3,$A4,$7D,$0C,$30,$F2,$3E,$9D,$F5,$7E,$AE,$1F,$88,$F8,$39,$1F
	.dc.b	$19,$FC,$5F,$C1,$FE,$33,$E1,$7F,$1B,$F0,$B2,$BE,$5C,$7E,$43,$E1
	.dc.b	$7F,$82,$FD,$57,$C9,$97,$27,$CB,$E9,$4E,$63,$1D,$1F,$DA,$5D,$47
	.dc.b	$FB,$20,$FE,$D2,$4F,$85,$F1,$5F,$DA,$CB,$C4,$FE,$3B,$2F,$13,$F8
	.dc.b	$4A,$ED,$D7,$FC,$C3,$2F,$0B,$A9,$F2,$B9,$9F,$EE,$31,$FE,$D5,$7F
	.dc.b	$B8,$BC,$FE,$07,$DB,$F0,$5E,$28,$FC,$EE,$3E,$2C,$FD,$79,$75,$7D
	.dc.b	$78,$FC,$43,$07,$7F,$85,$F8,$FF,$0B,$30,$F4,$47,$CA,$F8,$3F,$97
	.dc.b	$F0,$92,$7D,$91,$F3,$BE,$0B,$E7,$FB,$F2,$7E,$04,$7D,$2F,$82,$91
	.dc.b	$F0,$DF,$AB,$F0,$1D,$63,$E3,$61,$D5,$F8,$67,$C1,$C3,$B3,$F2,$F0
	.dc.b	$EC,$EB,$8B,$ED,$E3,$F8,$0F,$83,$FC,$0F,$C2,$4A,$FA,$51,$F8,$4F
	.dc.b	$84,$91,$F5,$5F,$C3,$FC,$27,$E2,$3E,$12,$4F,$D9,$7F,$17,$F0,$9F
	.dc.b	$8C,$F8,$1F,$C6,$FC,$24,$AF,$BD,$1D,$1F,$27,$53,$1C,$73,$7E,$33
	.dc.b	$2E,$B5,$AC,$AC,$FE,$43,$2F,$EB,$BF,$0A,$F8,$3B,$F0,$73,$0F,$E1
	.dc.b	$F3,$96,$FC,$2F,$02,$3A,$F8,$5E,$7F,$0B,$D3,$E1,$64,$F2,$8E,$FF
	.dc.b	$07,$E0,$F8,$39,$3D,$A3,$9B,$F3,$F1,$FF,$64,$6B,$FA,$B6,$F7,$FE
	.dc.b	$75,$FF,$00,$4E,$F0,$EE,$F1,$C7,$C6,$FA,$BE,$0F,$DE,$47,$DB,$F0
	.dc.b	$7E,$0C,$7B,$6B,$72,$ED,$AF,$2B,$E8,$7C,$4C,$71,$7E,$5F,$84,$F9
	.dc.b	$64,$9F,$93,$F0,$BF,$3B,$E1,$3E,$7F,$C2,$C9,$F9,$B1,$F4,$BE,$03
	.dc.b	$E9,$FC,$0C,$9F,$A3,$1F,$5B,$E0,$24,$7D,$47,$EC,$FC,$0F,$58,$FA
	.dc.b	$F8,$75,$7E,$56,$1D,$5F,$B7,$87,$57,$E7,$61,$F8,$2F,$85,$FC,$1F
	.dc.b	$C2,$C9,$FC,$0C,$7E,$1B,$E1,$3F,$C4,$7F,$9E,$FF,$B1,$4F,$F6,$37
	.dc.b	$3F,$F6,$FA,$FE,$AB,$1F,$D8,$F8,$3F,$91,$3F,$D8,$E3,$F5,$4D,$8C
	.dc.b	$FC,$2F,$4F,$83,$EB,$F0,$72,$F9,$C7,$85,$EA,$78,$8C,$1D,$27,$CB
	.dc.b	$F1,$12,$3C,$71,$E8,$F8,$49,$1E,$67,$D7,$F0,$9E,$CF,$85,$F6,$FC
	.dc.b	$34,$AF,$5C,$7C,$0F,$86,$91,$EE,$7E,$1F,$C3,$7C,$4F,$84,$F8,$BF
	.dc.b	$09,$2B,$E1,$C7,$57,$E2,$E1,$D5,$F8,$78,$74,$7E,$2E,$17,$7F,$93
	.dc.b	$8F,$D0,$F8,$3F,$A3,$F0,$92,$FE,$7C,$75,$7E,$6E,$3D,$6F,$E8,$E3
	.dc.b	$CF,$C1,$FE,$89,$C6,$CF,$18,$3D,$44,$B4,$FB,$FF,$0D,$2B,$EE,$BF
	.dc.b	$82,$F8,$6F,$C1,$FC,$37,$E1,$3E,$1A,$57,$E0,$A3,$F0,$FF,$01,$F8
	.dc.b	$8F,$81,$93,$F8,$68,$FC,$5F,$C3,$7F,$9B,$BF,$3D,$FE,$B3,$3E,$36
	.dc.b	$EB,$E4,$9E,$F3,$F8,$EC,$7B,$BF,$8C,$C7,$BB,$F9,$2C,$7B,$BF,$90
	.dc.b	$C7,$B7,$C3,$77,$EE,$93,$FE,$2C,$CD,$3E,$3F,$85,$93,$97,$CD,$F0
	.dc.b	$DE,$7F,$85,$93,$C5,$F5,$7C,$37,$AF,$E0,$E4,$F5,$7D,$DF,$09,$D2
	.dc.b	$3F,$A6,$65,$FD,$CB,$84,$D3,$FB,$9C,$6B,$0E,$ED,$3D,$18,$7C,$9F
	.dc.b	$83,$91,$EB,$7E,$67,$C2,$7C,$DF,$86,$93,$EF,$7E,$87,$C3,$7D,$1F
	.dc.b	$85,$FA,$5F,$0D,$2B,$E2,$47,$D5,$F8,$39,$1F,$1D,$FB,$1F,$07,$D2
	.dc.b	$3D,$78,$5F,$07,$FC,$45,$B3,$2E,$D5,$F3,$F2,$EC,$FC,$2C,$BF,$05
	.dc.b	$F0,$7F,$83,$F8,$39,$3F,$56,$3A,$78,$5F,$E1,$DC,$59,$E2,$BF,$11
	.dc.b	$97,$67,$E6,$E5,$DB,$3F,$1E,$7F,$8F,$F8,$59,$5F,$81,$7F,$25,$F0
	.dc.b	$DF,$EC,$0F,$CF,$3F,$B0,$4F,$F8,$05,$B7,$FE,$3D,$6E,$9F,$E2,$F7
	.dc.b	$FD,$B3,$3B,$F7,$F9,$11,$FF,$3C,$C7,$F5,$8A,$E5,$F0,$5D,$C9,$3A
	.dc.b	$F7,$FC,$5F,$0D,$23,$8B,$E5,$F8,$6E,$91,$DC,$EB,$8F,$A7,$E0,$BD
	.dc.b	$5F,$0B,$EB,$F8,$69,$5E,$28,$F7,$7C,$37,$BF,$E1,$A4,$F9,$A3,$E1
	.dc.b	$7C,$34,$8F,$4B,$F1,$7E,$1B,$E3,$7C,$1C,$9F,$6B,$F2,$7E,$13,$AC
	.dc.b	$7C,$0C,$3A,$BF,$2F,$0E,$8F,$CD,$C2,$F0,$FC,$DC,$7E,$97,$C1,$CA
	.dc.b	$F9,$2F,$D5,$F8,$4F,$AD,$F0,$DF,$5F,$E1,$A5,$7C,$E8,$FB,$5F,$0B
	.dc.b	$F6,$FE,$1A,$4F,$D2,$8F,$BD,$F0,$D2,$3E,$AB,$FF,$00,$3E,$15,$1D
	.dc.b	$8F,$B3,$87,$67,$F0,$98,$76,$7E,$EE,$1D,$9F,$C4,$61,$CF,$E1,$3A
	.dc.b	$7C,$2C,$9F,$F5,$DE,$3F,$D9,$FB,$7F,$15,$8F,$5D,$FF,$1D,$9F,$86
	.dc.b	$F9,$9D,$88,$60,$FF,$2B,$58,$F5,$11,$EF,$F0,$DE,$DF,$86,$91,$E2
	.dc.b	$7E,$07,$C3,$7C,$1F,$81,$F8,$5F,$05,$2B,$CF,$1F,$17,$E0,$BE,$37
	.dc.b	$C0,$C9,$F5,$C7,$47,$C7,$85,$D3,$FD,$49,$9B,$2E,$D5,$F0,$B2,$EC
	.dc.b	$FA,$B2,$FA,$5F,$0B,$23,$E3,$BF,$57,$E1,$BE,$B7,$C2,$C9,$F9,$6F
	.dc.b	$D9,$F8,$6F,$B5,$F0,$B2,$7E,$7B,$F7,$7E,$17,$EF,$7C,$1C,$9F,$A6
	.dc.b	$FE,$07,$E1,$3A,$47,$CC,$C2,$EF,$EA,$C7,$A4,$FD,$0C,$6C,$FB,$B1
	.dc.b	$FC,$57,$C1,$CA,$FB,$EF,$E3,$7E,$13,$F1,$DF,$0B,$27,$F0,$6F,$E4
	.dc.b	$1F,$F4,$87,$E4,$5F,$D8,$8F,$EA,$AF,$E5,$30,$FF,$67,$46,$23,$7F
	.dc.b	$0B,$C6,$73,$F0,$D7,$B3,$FD,$89,$7F,$D3,$C9,$F0,$D9,$C3,$FC,$05
	.dc.b	$1F,$D4,$2F,$EB,$E3,$E5,$9F,$C6,$BB,$F8,$35,$79,$72,$C3,$E3,$BF
	.dc.b	$2C,$63,$2C,$FC,$1E,$BE,$0E,$65,$EB,$D3,$CB,$3D,$F0,$F0,$CF,$C4
	.dc.b	$34,$79,$A7,$BE,$12,$33,$97,$92,$3C,$11,$8C,$9F,$0B,$E6,$24,$67
	.dc.b	$E1,$7D,$3F,$0B,$EA,$F8,$59,$3C,$63,$DB,$F0,$12,$3A,$3F,$03,$E0
	.dc.b	$3E,$0F,$BF,$27,$C0,$FC,$4F,$80,$EB,$1E,$3C,$3A,$BF,$7B,$0E,$8F
	.dc.b	$E0,$30,$BB,$FD,$EC,$7E,$6F,$C2,$CA,$F6,$3F,$43,$E1,$7E,$8F,$C2
	.dc.b	$49,$F8,$0F,$D4,$F8,$5F,$AB,$F0,$92,$7E,$23,$F6,$3E,$13,$EE,$7D
	.dc.b	$CF,$F5,$C9,$FF,$22,$37,$94,$D3,$FA,$FE,$9F,$92,$C7,$C1,$5F,$98
	.dc.b	$7C,$0F,$62,$1F,$9C,$C3,$C0,$FD,$8A,$FF,$5E,$BE,$0F,$80,$F1,$9E
	.dc.b	$4F,$80,$F2,$FC,$24,$9E,$0F,$A3,$E1,$3D,$3F,$03,$EA,$F8,$29,$5D
	.dc.b	$63,$DB,$EF,$C8,$F0,$3F,$03,$E0,$3A,$47,$FB,$FF,$36,$38,$D9,$DD
	.dc.b	$FD,$13,$83,$E1,$E9,$E0,$C3,$D5,$A7,$81,$E5,$A7,$CC,$F8,$09,$1E
	.dc.b	$F7,$E7,$FC,$07,$5C,$7A,$69,$E0,$AF,$69,$F1,$71,$EC,$F9,$71,$ED
	.dc.b	$9F,$92,$3E,$BE,$5F,$6B,$E0,$BE,$DF,$C1,$49,$F9,$D1,$F7,$BE,$0B
	.dc.b	$EF,$FC,$14,$9F,$A5,$1F,$82,$F7,$7F,$07,$EE,$C9,$FA,$D1,$F8,$6F
	.dc.b	$80,$FC,$3F,$C0,$C9,$FB,$51,$D5,$FB,$98,$75,$7E,$66,$1D,$5F,$C0
	.dc.b	$61,$D5,$FA,$18,$7E,$4B,$E0,$BF,$87,$7E,$77,$FC,$8E,$5B,$7B,$F9
	.dc.b	$53,$98,$C7,$17,$FB,$4E,$B7,$C7,$F5,$EB,$FB,$74,$7F,$A7,$DB,$7C
	.dc.b	$BE,$0F,$BF,$C1,$F8,$3D,$F9,$7F,$C3,$2F,$FB,$63,$7E,$0F,$CB,$F0
	.dc.b	$32,$34,$FA,$3E,$07,$D3,$8C,$9E,$4F,$67,$A6,$1D,$9F,$39,$E5,$7E
	.dc.b	$0D,$FC,$2F,$80,$93,$E3,$7E,$2F,$C0,$FC,$6F,$82,$93,$E7,$7E,$4F
	.dc.b	$C1,$FC,$AF,$83,$93,$EB,$7E,$6F,$C1,$FC,$EF,$81,$F9,$FF,$03,$2B
	.dc.b	$E0,$47,$47,$CB,$C5,$F4,$31,$C7,$2F,$D3,$E2,$FF,$64,$57,$52,$1C
	.dc.b	$16,$5E,$08,$F9,$9F,$03,$F7,$BE,$06,$46,$0F,$51,$0E,$58,$7B,$1E
	.dc.b	$B8,$78,$6F,$DB,$1D,$F4,$FC,$4F,$C1,$C8,$FB,$4F,$E3,$3E,$0F,$F1
	.dc.b	$BF,$07,$27,$EF,$3F,$90,$F8,$3F,$C8,$FC,$17,$F5,$FF,$3B,$FD,$A6
	.dc.b	$7F,$B5,$37,$46,$7D,$F9,$87,$F5,$2A,$FE,$C5,$DB,$FC,$1D,$7C,$17
	.dc.b	$79,$F8,$99,$5D,$FE,$56,$7E,$09,$E5,$87,$84,$C3,$F4,$B0,$FF,$64
	.dc.b	$1F,$DD,$E3,$B7,$C4,$79,$FE,$13,$D1,$F0,$72,$FC,$8F,$AF,$E1,$3D
	.dc.b	$9F,$07,$27,$D0,$FB,$FE,$0F,$E0,$7C,$07,$C1,$F8,$19,$5E,$D8,$E8
	.dc.b	$F8,$F0,$BB,$7E,$D7,$56,$5D,$6B,$D1,$95,$AD,$65,$F3,$7E,$06,$57
	.dc.b	$C9,$7E,$87,$C0,$FD,$1F,$83,$93,$F3,$5F,$A9,$F0,$7F,$57,$E0,$FE
	.dc.b	$B7,$C1,$CA,$FA,$51,$F6,$7E,$06,$47,$D5,$7E,$E7,$C1,$F4,$8F,$99
	.dc.b	$85,$D3,$D5,$8F,$59,$FB,$98,$F5,$7E,$96,$3F,$85,$F8,$29,$1F,$81
	.dc.b	$7F,$11,$F0,$7C,$F9,$FF,$05,$95,$A7,$8A,$E3,$67,$DE,$7E,$F6,$7D
	.dc.b	$F4,$F9,$9A,$7F,$8C,$7F,$3B,$FE,$E8,$4F,$87,$C7,$C3,$E7,$E1,$B5
	.dc.b	$9C,$CB,$FC,$4B,$7F,$DB,$B4,$3F,$D3,$E3,$F7,$AD,$CB,$F9,$73,$76
	.dc.b	$FE,$C0,$BE,$07,$C5,$F0,$12,$FA,$BE,$5F,$81,$F3,$7C,$34,$9F,$0B
	.dc.b	$E9,$F8,$6F,$57,$C2,$C9,$F2,$BE,$DF,$85,$F7,$7C,$27,$BF,$E1,$65
	.dc.b	$7A,$A3,$E1,$7C,$14,$8F,$6B,$F1,$7E,$0F,$E3,$7C,$14,$9F,$82,$FC
	.dc.b	$9F,$83,$F9,$5F,$09,$F2,$FE,$16,$57,$C6,$8F,$F3,$22,$FC,$A9,$BF
	.dc.b	$F5,$28,$7C,$B7,$FA,$6D,$FF,$63,$FE,$00,$E3,$FF,$3D,$4E,$65,$39
	.dc.b	$7F,$D9,$B1,$9B,$EB,$CF,$8F,$84,$F9,$FF,$11,$FE,$6F,$FC,$EF,$FA
	.dc.b	$F7,$57,$F8,$45,$B8,$FE,$2B,$AB,$FB,$06,$1F,$D9,$67,$FA,$E9,$3E
	.dc.b	$1B,$97,$C4,$7B,$33,$EF,$87,$AB,$3F,$F4,$86,$BE,$58,$C1,$EC,$DB
	.dc.b	$D7,$3F,$E9,$5A,$E3,$B7,$C1,$ED,$FC,$5E,$FE,$7C,$21,$F9,$07,$D5
	.dc.b	$F0,$F2,$3F,$CE,$FA,$FF,$6C,$7F,$A6,$AF,$87,$F3,$71,$7C,$FC,$7E
	.dc.b	$56,$BE,$56,$EB,$E4,$C3,$FD,$EF,$A7,$93,$5F,$93,$1D,$8F,$19,$F9
	.dc.b	$ED,$3E,$67,$C4,$48,$F1,$BF,$3E,$7C,$3C,$3F,$4E,$7E,$11,$4F,$B7
	.dc.b	$87,$85,$FC,$36,$7E,$7A,$FA,$DF,$11,$F6,$49,$3F,$03,$E2,$3E,$E7
	.dc.b	$C1,$C8,$F8,$6F,$DF,$F8,$8F,$0C,$7D,$23,$B9,$E1,$DF,$E1,$61,$F4
	.dc.b	$F1,$F0,$4F,$F0,$CE,$16,$5E,$08,$F9,$AF,$E2,$F2,$FC,$67,$C2,$49
	.dc.b	$FA,$6F,$E3,$FE,$13,$C3,$C5,$F3,$A7,$C1,$B7,$BF,$8B,$E7,$47,$7C
	.dc.b	$7B,$70,$7E,$45,$ED,$1F,$7B,$E1,$FF,$35,$F1,$12,$30,$7F,$AF,$0F
	.dc.b	$93,$5D,$98,$7C,$1B,$EF,$CE,$D9,$F1,$1F,$A5,$F8,$89,$7F,$8C,$8F
	.dc.b	$D4,$7C,$47,$EA,$7E,$22,$4F,$E4,$23,$DB,$C9,$FA,$0D,$FE,$D7,$2F
	.dc.b	$CC,$AF,$C5,$F2,$FC,$CD,$3C,$19,$76,$9F,$CB,$E7,$FB,$58,$FD,$B7
	.dc.b	$C2,$49,$FC,$EB,$FB,$9F,$84,$EF,$1F,$8A,$3E,$E6,$3D,$DF,$82,$7E
	.dc.b	$C3,$0E,$EF,$E9,$B0,$F0,$70,$7E,$F7,$1F,$E0,$9F,$0F,$2B,$F5,$6F
	.dc.b	$FB,$D7,$ED,$3F,$D0,$91,$FE,$4C,$3A,$F3,$7D,$B7,$C3,$B7,$DB,$DB
	.dc.b	$DD,$5F,$E1,$89,$FE,$AA,$FF,$86,$1F,$F7,$6B,$5F,$F2,$1B,$BC,$CB
	.dc.b	$E1,$7C,$44,$9E,$3F,$0D,$F0,$9F,$CD,$9E,$23,$F5,$9C,$BF,$7A,$BA
	.dc.b	$1D,$F5,$FF,$40,$37,$C5,$FF,$59,$3A,$F8,$4F,$80,$4B,$F3,$FC,$27
	.dc.b	$7E,$2F,$9C,$7C,$FC,$7B,$7C,$47,$C7,$25,$7B,$BE,$17,$E5,$7C,$2C
	.dc.b	$8F,$82,$FC,$DF,$86,$E9,$C5,$F8,$83,$E5,$6D,$FC,$1F,$87,$E1,$FC
	.dc.b	$3F,$D3,$26,$1F,$2B,$E1,$FE,$B7,$C3,$FD,$7F,$83,$93,$F3,$A3,$AB
	.dc.b	$FA,$2E,$7F,$E8,$DC,$FF,$5B,$0E,$73,$F4,$9F,$C0,$61,$F8,$1F,$83
	.dc.b	$FC,$17,$C1,$4A,$FB,$31,$D1,$FA,$07,$EB,$6F,$F0,$35,$D3,$4F,$9F
	.dc.b	$87,$4A,$FC,$0F,$C1,$FE,$37,$E1,$E4,$60,$FD,$D7,$27,$F0,$47,$EE
	.dc.b	$E9,$F8,$F3,$D3,$87,$C5,$E8,$7C,$3F,$E5,$FE,$1E,$5F,$E3,$A3,$F3
	.dc.b	$5F,$0E,$89,$7F,$EC,$1E,$E7,$C3,$E6,$7F,$D2,$B1,$87,$F2,$F5,$EF
	.dc.b	$CB,$C5,$E2,$FF,$27,$C3,$C5,$8F,$82,$FF,$37,$97,$EA,$63,$A4,$F7
	.dc.b	$F8,$7B,$E0,$7F,$C0,$1C,$D1,$8C,$3F,$C7,$71,$87,$CF,$3B,$F8,$7F
	.dc.b	$16,$5F,$C3,$E7,$F3,$FC,$9F,$B0,$CB,$C5,$97,$E7,$B3,$F1,$57,$23
	.dc.b	$AF,$C3,$F7,$EA,$FD,$73,$74,$FF,$25,$4E,$E7,$F6,$5F,$17,$E0,$8F
	.dc.b	$E2,$1B,$FD,$27,$F8,$75,$7F,$8E,$DE,$0F,$F8,$07,$1F,$9D,$D9,$D7
	.dc.b	$E2,$3B,$7C,$47,$7D,$A6,$1C,$23,$B3,$F8,$0D,$FE,$73,$DA,$BC,$D7
	.dc.b	$23,$BC,$7A,$7E,$23,$D5,$F1,$12,$7C,$71,$D3,$C5,$FD,$C6,$5F,$CC
	.dc.b	$F4,$F4,$57,$E0,$31,$E8,$FE,$0B,$1F,$89,$F1,$12,$3D,$AF,$C7,$F8
	.dc.b	$7E,$95,$FB,$89,$F8,$7F,$11,$F2,$C8,$FC,$4F,$88,$F9,$DF,$11,$23
	.dc.b	$E3,$BF,$47,$E2,$39,$C7,$DC,$E2,$FE,$05,$A3,$0E,$73,$F9,$67,$EC
	.dc.b	$61,$F6,$7E,$23,$ED,$7C,$3C,$AF,$A7,$1D,$1F,$DB,$61,$D2,$F9,$BF
	.dc.b	$A3,$79,$C7,$DA,$F8,$8F,$C2,$11,$FB,$BF,$07,$CF,$9F,$F0,$47,$F8
	.dc.b	$E7,$F3,$BF,$F2,$5F,$51,$A4,$F8,$8F,$C8,$2F,$88,$A4,$C3,$F2,$11
	.dc.b	$FE,$3B,$CF,$FA,$8F,$8D,$FC,$0F,$A3,$F9,$8F,$27,$CF,$CF,$C1,$87
	.dc.b	$42,$1C,$B2,$EF,$F0,$F2,$3A,$BE,$2F,$87,$F5,$76,$FE,$3A,$FD,$57
	.dc.b	$CA,$D9,$BC,$4F,$E1,$5B,$FB,$67,$D7,$5E,$CC,$24,$7A,$9F,$7F,$C1
	.dc.b	$76,$8F,$F0,$1E,$1F,$CF,$77,$7C,$2F,$C4,$F8,$79,$5F,$0A,$3E,$3F
	.dc.b	$C3,$FC,$8F,$87,$93,$F1,$A3,$A3,$FD,$D3,$2F,$2E,$1D,$23,$E5,$E1
	.dc.b	$D1,$F5,$EB,$AC,$7E,$9F,$C3,$FD,$4F,$87,$95,$F4,$A3,$AF,$2F,$F8
	.dc.b	$A3,$1E,$93,$C8,$FF,$68,$AC,$7B,$6B,$F1,$E7,$38,$FE,$03,$E1,$FF
	.dc.b	$03,$F0,$32,$7E,$FC,$76,$F1,$7F,$C4,$D8,$BE,$0B,$F0,$FF,$0B,$2F
	.dc.b	$F0,$B1,$F8,$AF,$86,$FC,$5F,$C3,$49,$FC,$4C,$73,$D3,$E7,$57,$19
	.dc.b	$F4,$73,$FE,$57,$0F,$C9,$CF,$E5,$3E,$1B,$F2,$BF,$0D,$2F,$F2,$71
	.dc.b	$CD,$FF,$16,$67,$E5,$7F,$2F,$7D,$71,$F8,$18,$FE,$87,$1F,$E5,$DF
	.dc.b	$9D,$FF,$8A,$A7,$2F,$F8,$7C,$9C,$BD,$7A,$FF,$D4,$51,$FE,$15,$C7
	.dc.b	$FD,$FA,$DC,$0D,$3F,$C4,$D1,$FC,$DD,$7C,$3F,$83,$E1,$E4,$E6,$3B
	.dc.b	$74,$FE,$6A,$B9,$7C,$5F,$98,$F3,$F9,$5E,$8F,$88,$F4,$FC,$3C,$C7
	.dc.b	$B4,$72,$7C,$B5,$C5,$F1,$3C,$9F,$27,$C3,$FC,$2F,$87,$91,$83,$FB
	.dc.b	$FF,$47,$FA,$41,$FE,$80,$7F,$4E,$CF,$F6,$87,$9F,$E0,$BE,$59,$23
	.dc.b	$E0,$7C,$1F,$CE,$F8,$69,$1F,$0D,$FA,$3F,$0F,$E7,$E2,$F9,$37,$C6
	.dc.b	$BF,$07,$C0,$8F,$AF,$87,$D8,$F8,$69,$5F,$35,$FB,$7F,$0F,$CE,$F8
	.dc.b	$4F,$D1,$BE,$9C,$FF,$7B,$0F,$C0,$FC,$3F,$E0,$BE,$1E,$5F,$D7,$8E
	.dc.b	$5F,$05,$F8,$6F,$82,$93,$F6,$E3,$F1,$3F,$0B,$FD,$DF,$F3,$BF,$F0
	.dc.b	$A4,$FE,$69,$B9,$7E,$32,$79,$CF,$12,$1B,$F8,$6E,$5F,$0D,$23,$FB
	.dc.b	$F5,$A3,$B7,$C3,$72,$9F,$8B,$8F,$E6,$8F,$15,$F8,$FE,$1B,$C9,$F0
	.dc.b	$72,$B9,$C7,$CF,$C7,$FC,$05,$E1,$7F,$28,$5F,$0D,$EB,$3D,$9E,$C7
	.dc.b	$B7,$AF,$DD,$F0,$B3,$1F,$2C,$73,$9D,$C7,$A1,$F8,$9F,$0D,$F1,$7E
	.dc.b	$1A,$4F,$B1,$F9,$1F,$0D,$C6,$38,$10,$F9,$53,$F3,$7E,$06,$47,$C4
	.dc.b	$7E,$87,$C2,$7A,$FE,$1B,$E9,$7C,$34,$AF,$93,$1F,$57,$E1,$64,$7C
	.dc.b	$C7,$EC,$7C,$37,$18,$F9,$F5,$C5,$FB,$55,$F7,$7E,$1B,$EF,$7C,$34
	.dc.b	$9F,$AB,$1C,$5D,$9E,$03,$FC,$6D,$F0,$9F,$DC,$07,$C1,$FF,$8F,$5F
	.dc.b	$ED,$D1,$AF,$86,$C4,$FE,$8E,$7F,$B7,$73,$7F,$5C,$6D,$3C,$74,$61
	.dc.b	$AF,$5F,$85,$ED,$E5,$CA,$F1,$C7,$5F,$8B,$F1,$12,$3C,$F5,$E5,$F8
	.dc.b	$69,$1E,$A7,$D1,$F0,$1C,$63,$DB,$5C,$5F,$E7,$35,$E4,$7E,$0F,$C0
	.dc.b	$FB,$C8,$FC,$4F,$85,$F7,$7C,$37,$C3,$24,$7C,$8F,$85,$F8,$DF,$0D
	.dc.b	$F1,$FE,$12,$4F,$CC,$8F,$8D,$3C,$23,$C3,$5F,$36,$3E,$77,$C2,$C9
	.dc.b	$FA,$6F,$D1,$F7,$F8,$C7,$D6,$AE,$5E,$67,$F4,$6D,$18,$7D,$7F,$80
	.dc.b	$97,$F7,$1F,$B5,$F0,$9F,$5F,$E1,$7E,$E1,$27,$F0,$3F,$09,$F7,$FE
	.dc.b	$16,$47,$E1,$1F,$F3,$F7,$E7,$7F,$D4,$7A,$7F,$95,$CF,$F2,$D2,$1F
	.dc.b	$E0,$93,$FC,$2A,$3E,$15,$D3,$FE,$A5,$3F,$17,$19,$71,$D9,$FE,$33
	.dc.b	$7F,$C4,$B7,$FE,$FB,$3F,$AC,$13,$E2,$FF,$A4,$57,$73,$AD,$7F,$AC
	.dc.b	$6F,$BC,$7F,$AD,$EF,$BC,$7F,$93,$9E,$BF,$09,$EC,$F8,$59,$5F,$E9
	.dc.b	$7C,$7F,$B1,$BF,$E1,$53,$B7,$C4,$FF,$92,$1F,$F3,$61,$FE,$62,$3D
	.dc.b	$47,$7C,$7C,$58,$7B,$23,$78,$F9,$3B,$D8,$3E,$1F,$E6,$92,$BC,$5F
	.dc.b	$0D,$EE,$F8,$9F,$F6,$D6,$3E,$A3,$A4,$F8,$71,$F6,$3F,$ED,$59,$FE
	.dc.b	$4A,$FF,$B5,$E3,$E1,$A8,$CB,$C6,$7E,$DB,$FD,$A0,$91,$FE,$E9,$3C
	.dc.b	$39,$7B,$23,$DC,$7B,$27,$8C,$F4,$AE,$ED,$8C,$FA,$1F,$DD,$74,$F9
	.dc.b	$57,$E1,$3F,$0F,$7E,$6A,$F7,$47,$63,$47,$DF,$BE,$8F,$BF,$D2,$FC
	.dc.b	$43,$CA,$7F,$DB,$8F,$D4,$3D,$78,$7F,$81,$6B,$D7,$1F,$16,$7E,$5D
	.dc.b	$FC,$A8,$ED,$7E,$F7,$E4,$3D,$9F,$C1,$FA,$39,$C7,$FD,$87,$D5,$F0
	.dc.b	$E7,$D1,$7E,$BC,$F8,$65,$F2,$F1,$E8,$FA,$AB,$D9,$9F,$C5,$CB,$A1
	.dc.b	$C1,$FF,$0E,$67,$F1,$27,$F3,$E7,$73,$CE,$7A,$CF,$88,$66,$BF,$D4
	.dc.b	$15,$D2,$31,$5D,$30,$F7,$5F,$5B,$FC,$04,$FC,$7C,$3C,$E7,$8F,$2E
	.dc.b	$D8,$7E,$1F,$3F,$82,$FD,$6C,$3C,$C7,$3B,$FA,$07,$38,$E3,$87,$C1
	.dc.b	$3B,$DF,$4B,$FC,$2E,$7F,$4E,$7E,$BD,$7D,$BC,$3A,$5F,$B7,$1F,$2C
	.dc.b	$FB,$23,$96,$7E,$F8,$FC,$4D,$F2,$C3,$EA,$46,$CF,$C3,$47,$8E,$FD
	.dc.b	$D1,$F8,$9C,$7C,$78,$FE,$6B,$1F,$82,$FB,$B3,$DE,$3F,$A0,$C3,$C7
	.dc.b	$87,$83,$3F,$B8,$F9,$0F,$04,$F8,$9F,$65,$72,$B4,$7D,$A3,$CB,$79
	.dc.b	$7F,$97,$E3,$F4,$2F,$F3,$15,$E3,$9E,$AF,$27,$F3,$27,$D0,$AE,$4F
	.dc.b	$D9,$CE,$CF,$18,$F4,$65,$B6,$18,$3E,$24,$FC,$DB,$FA,$D9,$7E,$2E
	.dc.b	$7E,$2E,$5D,$B3,$F9,$91,$E4,$3D,$37,$E0,$9F,$C8,$9B,$CB,$E4,$E1
	.dc.b	$D3,$1F,$91,$7E,$03,$91,$FA,$23,$E4,$98,$C3,$BB,$E7,$9F,$E4,$39
	.dc.b	$FD,$A8,$FE,$0C,$43,$94,$7D,$33,$C0,$FD,$03,$E0,$B8,$39,$E3,$F0
	.dc.b	$73,$FC,$24,$F8,$89,$3B,$8F,$A5,$7E,$48,$C4,$FC,$23,$D7,$3F,$46
	.dc.b	$33,$1F,$60,$F1,$1E,$B8,$FE,$3B,$1E,$28,$E3,$97,$CA,$7E,$16,$5F
	.dc.b	$2A,$BC,$B5,$B9,$F9,$47,$33,$F8,$BC,$7F,$22,$36,$7A,$27,$94,$7C
	.dc.b	$7B,$FE,$16,$FD,$4A,$F3,$D7,$A4,$E3,$1F,$99,$8F,$D2,$3F,$63,$1F
	.dc.b	$A4,$FF,$20,$CB,$F0,$EC,$3C,$87,$C1,$9E,$2E,$27,$8D,$FF,$08,$8F
	.dc.b	$9F,$7E,$97,$CD,$9F,$F0,$F7,$E4,$DF,$CC,$BE,$33,$FC,$53,$0F,$04
	.dc.b	$FE,$06,$FD,$37,$E1,$7E,$13,$E9,$CF,$F0,$0F,$6A,$E3,$3E,$18,$F6
	.dc.b	$1E,$07,$8B,$E9,$9E,$2E,$AB,$C0,$FD,$23,$A1,$C4,$F9,$53,$E4,$C3
	.dc.b	$D3,$7F,$D0,$63,$E8,$4F,$0A,$F8,$E4,$8E,$17,$E0,$C7,$E4,$4F,$82
	.dc.b	$BF,$43,$5F,$02,$7F,$2C,$F0,$8F,$BC,$FC,$43,$BB,$E9,$7D,$B3,$DE
	.dc.b	$3C,$51,$F9,$62,$38,$3D,$C7,$80,$C1,$F3,$CF,$79,$8C,$3D,$EF,$F4
	.dc.b	$DC,$FF,$3C,$FE,$1A,$FC,$EE,$F0,$F7,$65,$F7,$63,$F0,$A7,$82,$7E
	.dc.b	$0E,$1E,$82,$47,$13,$DE,$79,$1F,$7C,$70,$8F,$29,$E7,$3E,$11,$FB
	.dc.b	$B3,$F3,$31,$C2,$7C,$58,$FC,$59,$E1,$5C,$DF,$E6,$47,$EE,$8F,$C5
	.dc.b	$9F,$11,$F6,$BD,$8D,$9D,$B0,$F8,$73,$FC,$22,$76,$75,$C7,$E4,$C7
	.dc.b	$F4,$69,$DE,$1F,$7F,$3F,$DE,$E1,$D6,$F7,$3F,$E4,$43,$D5,$9F,$F6
	.dc.b	$F9,$FE,$47,$1F,$85,$7F,$85,$E1,$FC,$0B,$0E,$B1,$E9,$C3,$AE,$3E
	.dc.b	$3C,$FE,$AE,$3F,$99,$BD,$DF,$B6,$BD,$39,$FC,$8C,$BF,$13,$3F,$E3
	.dc.b	$AB,$E0,$FC,$63,$86,$3D,$AB,$84,$7D,$3C,$BC,$8E,$F0,$FA,$07,$6B
	.dc.b	$FF,$15,$67,$F0,$27,$FA,$26,$3F,$01,$FF,$04,$E3,$D6,$B7,$9F,$AA
	.dc.b	$7F,$28,$75,$CB,$C2,$FD,$EA,$FE,$DF,$87,$58,$FF,$4C,$61,$D7,$2F
	.dc.b	$F0,$F5,$75,$9F,$F2,$FD,$6F,$2F,$F0,$C5,$FF,$0D,$7D,$35,$BB,$FF
	.dc.b	$2C,$56,$F0,$FE,$E5,$87,$E8,$0F,$8A,$7D,$83,$F3,$C7,$D8,$C3,$F7
	.dc.b	$99,$FD,$2C,$7A,$CE,$CF,$C2,$65,$EF,$C7,$F1,$51,$EF,$CF,$F0,$8F
	.dc.b	$F1,$73,$67,$EA,$8F,$CC,$61,$EF,$3F,$D0,$F9,$FA,$5F,$F4,$8E,$7E
	.dc.b	$9A,$FB,$91,$FE,$3E,$AF,$A2,$78,$F1,$FC,$53,$D5,$FD,$54,$75,$8F
	.dc.b	$EA,$D7,$D5,$FE,$FD,$7D,$5F,$89,$3D,$70,$FA,$31,$D7,$3F,$B3,$5B
	.dc.b	$9F,$E4,$D5,$B3,$D3,$5B,$C7,$F0,$6F,$8E,$FF,$03,$3E,$38,$FF,$3A
	.dc.b	$9F,$17,$1F,$4B,$F4,$4F,$E1,$D9,$FD,$1A,$FC,$4C,$FE,$56,$7F,$07
	.dc.b	$9F,$D1,$BF,$DA,$5E,$F0,$FA,$38,$FF,$0D,$C3,$AB,$FD,$FC,$FD,$0E
	.dc.b	$1F,$83,$8F,$BA,$F8,$F1,$F7,$E3,$D6,$3F,$43,$87,$57,$E8,$9B,$C3
	.dc.b	$DE,$FA,$4F,$BB,$1B,$C3,$E5,$C6,$CF,$49,$F1,$67,$F7,$2F,$8E,$FD
	.dc.b	$F9,$FD,$78,$FF,$56,$9B,$3F,$5A,$6E,$7D,$F9,$7F,$38,$8F,$C7,$9E
	.dc.b	$3C,$BF,$07,$8F,$8E,$7E,$BC,$F8,$F1,$FD,$3E,$5E,$F8,$FD,$9D,$6D
	.dc.b	$FC,$49,$F7,$4F,$49,$D7,$2F,$BB,$1B,$8F,$BB,$5F,$46,$7E,$5C,$7A
	.dc.b	$63,$F4,$EF,$A4,$F9,$6E,$F1,$F9,$6F,$D7,$C3,$F6,$73,$FA,$D3,$6F
	.dc.b	$BF,$1F,$BA,$FF,$24,$7D,$F1,$B9,$FA,$27,$57,$D2,$F5,$7F,$1E,$6E
	.dc.b	$7E,$2C,$6C,$F8,$B1,$E9,$8F,$C7,$BB,$3E,$29,$F2,$DF,$DC,$E3,$F2
	.dc.b	$EB,$F8,$BC,$FF,$6F,$C3,$EE,$BF,$2E,$FA,$BF,$2F,$0F,$C1,$C7,$F8
	.dc.b	$4E,$BF,$2A,$F8,$F0,$EB,$5E,$33,$F2,$A7,$8C,$FA,$2F,$C5,$BF,$A3
	.dc.b	$1D,$4F,$C4,$CF,$58,$F1,$C7,$A4,$FE,$0B,$87,$A6,$3F,$94,$CE,$F0
	.dc.b	$FC,$7D,$7F,$80,$30,$FA,$2F,$FA,$DB,$3F,$F5,$96,$3F,$1A,$3E,$F5
	.dc.b	$FD,$28,$F2,$5F,$DE,$7E,$C6,$1F,$8A,$AF,$56,$08,$FF,$2C,$5F,$C1
	.dc.b	$9F,$29,$F0,$4F,$CB,$9C,$23,$75,$DA,$3E,$FB,$DA,$FE,$6B,$FD,$26
	.dc.b	$7E,$0D,$75,$AF,$D7,$9F,$1D,$FC,$B9,$F0,$5F,$C8,$9F,$06,$3E,$6C
	.dc.b	$7E,$2E,$3F,$80,$9E,$3A,$FE,$6D,$9F,$F6,$0C,$BE,$64,$7F,$3C,$CF
	.dc.b	$FD,$7D,$5F,$83,$C7,$EE,$BE,$17,$E5,$61,$E1,$3E,$54,$78,$6B,$E5
	.dc.b	$3F,$46,$FF,$D9,$4F,$84,$FE,$5D,$87,$85,$F9,$19,$7E,$51,$FF,$28
	.dc.b	$1F,$D9,$0F,$C1,$1F,$0A,$FD,$91,$FA,$C8,$FB,$78,$7E,$98,$F9,$D3
	.dc.b	$F9,$A3,$79,$7D,$B9,$D9,$FA,$68,$F9,$D5,$FE,$A8,$8F,$71,$F9,$A3
	.dc.b	$67,$E6,$8F,$71,$F6,$8F,$71,$FE,$45,$36,$FE,$1F,$1F,$9F,$1F,$E9
	.dc.b	$53,$DF,$5F,$9A,$BF,$C0,$4F,$E6,$A7,$FC,$A3,$3F,$6A,$7F,$0D,$1F
	.dc.b	$A6,$76,$FE,$00,$DC,$FD,$07,$FC,$64,$6C,$FD,$F3,$D6,$98,$7D,$19
	.dc.b	$F7,$9F,$98,$3D,$37,$F9,$87,$DF,$87,$CB,$C3,$C6,$FE,$49,$F1,$BF
	.dc.b	$82,$8F,$19,$F1,$4F,$1B,$B7,$C6,$7A,$6F,$C6,$FC,$53,$C6,$FE,$0E
	.dc.b	$7F,$BE,$47,$D7,$7E,$E9,$F9,$49,$FB,$B8,$7E,$AA,$7A,$C7,$8E,$BA
	.dc.b	$C7,$FA,$F4,$FA,$33,$F9,$F7,$FD,$E6,$7E,$7C,$FD,$D9,$E2,$BF,$AC
	.dc.b	$F4,$BF,$C7,$3F,$11,$F1,$1F,$10,$FD,$29,$F1,$0D,$BF,$12,$FE,$53
	.dc.b	$F8,$83,$FD,$0C,$7E,$95,$FF,$03,$61,$F5,$5F,$6C,$7D,$B3,$F4,$B8
	.dc.b	$7F,$32,$AF,$87,$1E,$89,$D9,$F3,$DD,$DF,$F0,$C9,$DC,$7E,$AD,$F4
	.dc.b	$3F,$9F,$9F,$DE,$47,$A2,$BE,$4C,$7F,$BD,$71,$FE,$0F,$5F,$E3,$6C
	.dc.b	$BF,$83,$BF,$B6,$AF,$D5,$57,$8A,$FE,$1C,$FC,$93,$ED,$1F,$24,$F4
	.dc.b	$3E,$27,$F2,$51,$E2,$76,$7C,$98,$F9,$EF,$C9,$8F,$F1,$49,$B7,$F6
	.dc.b	$C6,$CE,$A6,$E3,$FB,$CB,$F9,$EC,$BF,$6A,$FE,$14,$FC,$03,$F8,$57
	.dc.b	$FD,$B5,$97,$F8,$86,$FD,$B5,$E7,$7F,$2D,$87,$C2,$C3,$F8,$5E,$4C
	.dc.b	$7F,$CD,$D9,$FF,$52,$CB,$E7,$63,$F8,$57,$DB,$8F,$CE,$8F,$F5,$C6
	.dc.b	$1F,$26,$76,$7C,$93,$DB,$3B,$7F,$D0,$F1,$B7,$E4,$CF,$57,$79,$7E
	.dc.b	$2E,$3D,$18,$FE,$72,$7F,$0B,$8F,$F0,$5A,$D4,$7C,$73,$FC,$26,$7C
	.dc.b	$73,$F9,$C9,$FB,$7B,$F0,$61,$F0,$A3,$F5,$67,$E0,$EF,$E1,$3F,$E5
	.dc.b	$3C,$7E,$13,$E6,$3D,$93,$FC,$9D,$EA,$7D,$3A,$EB,$5E,$7C,$7F,$81
	.dc.b	$CF,$B0,$F1,$3E,$CC,$7E,$C5,$7F,$BB,$F1,$FB,$17,$E2,$8F,$EC,$98
	.dc.b	$F8,$B1,$F4,$63,$F7,$8F,$E0,$47,$E9,$4F,$84,$7F,$6A,$3E,$14,$FE
	.dc.b	$D9,$F0,$9A,$8F,$EA,$91,$F9,$37,$D8,$EE,$3D,$98,$F9,$DF,$D5,$1F
	.dc.b	$22,$3F,$13,$8F,$E2,$33,$FE,$D3,$3E,$0A,$FF,$27,$65,$F5,$A3,$F9
	.dc.b	$6C,$7E,$B9,$FB,$67,$F2,$D8,$D6,$3E,$FC,$BE,$56,$7F,$97,$BF,$E0
	.dc.b	$79,$FE,$73,$0F,$55,$7F,$67,$CF,$D4,$FF,$43,$8F,$D6,$1F,$14,$F0
	.dc.b	$9F,$96,$3C,$31,$FC,$C7,$1F,$0E,$5E,$A8,$E7,$9F,$EC,$1F,$E1,$95
	.dc.b	$FC,$D0,$F8,$67,$D3,$3F,$15,$1F,$0D,$FE,$DD,$1C,$9F,$C3,$BC,$9F
	.dc.b	$DC,$9C,$9F,$C1,$3F,$73,$0F,$73,$E8,$3F,$24,$7A,$31,$FE,$41,$87
	.dc.b	$A2,$3B,$C7,$EA,$E7,$ED,$BF,$4A,$3F,$02,$73,$8F,$C5,$D7,$82,$3D
	.dc.b	$27,$E8,$70,$F8,$55,$CF,$1F,$99,$7F,$BB,$AF,$F5,$3D,$FE,$83,$1F
	.dc.b	$CD,$DA,$65,$7F,$41,$7F,$95,$1F,$4F,$3F,$BD,$5C,$23,$78,$7F,$41
	.dc.b	$77,$5F,$0A,$B7,$1F,$1E,$FB,$3F,$E3,$1C,$3F,$26,$FA,$4F,$86,$7E
	.dc.b	$E2,$FD,$39,$7C,$78,$FE,$B6,$7E,$5C,$FF,$3A,$39,$9F,$E6,$F5,$F0
	.dc.b	$71,$FC,$13,$FC,$80,$FD,$24,$7F,$BF,$33,$FC,$B5,$FA,$30,$FD,$D5
	.dc.b	$79,$B1,$67,$FD,$53,$0E,$17,$FA,$BA,$FC,$ED,$F0,$8F,$DD,$DF,$C0
	.dc.b	$AD,$C7,$E3,$DF,$79,$FC,$1A,$FE,$BC,$7B,$F1,$F3,$BF,$CC,$30,$F3
	.dc.b	$9F,$20,$63,$FE,$D1,$CF,$F0,$F3,$F0,$8F,$C3,$E3,$E1,$AE,$15,$BC
	.dc.b	$BE,$5D,$6B,$1F,$0B,$F9,$53,$FA,$3E,$7F,$2E,$FF,$AD,$E7,$FA,$33
	.dc.b	$F8,$71,$FA,$33,$F1,$53,$F7,$4F,$C1,$E7,$FC,$FD,$EC,$7D,$C3,$B1
	.dc.b	$FC,$B4,$FE,$2E,$7F,$04,$7F,$C5,$CE,$B3,$F6,$A7,$F3,$44,$C7,$FC
	.dc.b	$B5,$1C,$4F,$F0,$D5,$FF,$3A,$9E,$27,$0C,$3D,$93,$FD,$43,$0F,$65
	.dc.b	$7E,$BE,$3C,$67,$EB,$EF,$F8,$DB,$F7,$CF,$C3,$61,$FC,$1D,$FC,$9D
	.dc.b	$72,$7E,$91,$CB,$3F,$A4,$FE,$B6,$FE,$CB,$FE,$33,$7C,$13,$F2,$DF
	.dc.b	$BF,$1B,$CF,$EC,$5F,$E1,$23,$F7,$33,$A7,$FD,$36,$FB,$F0,$FB,$B1
	.dc.b	$E1,$CF,$EE,$CE,$6B,$96,$7F,$ED,$AA,$F7,$47,$EE,$DF,$DA,$4F,$E7
	.dc.b	$23,$DD,$8F,$EB,$27,$F2,$C7,$F2,$49,$FD,$63,$F4,$63,$FC,$89,$1D
	.dc.b	$E3,$47,$E6,$CD,$3F,$D2,$CD,$47,$E3,$1E,$71,$E4,$3F,$18,$79,$0F
	.dc.b	$A2,$7F,$07,$3E,$8B,$FB,$F7,$B9,$F1,$8E,$E7,$F4,$97,$B9,$F4,$0E
	.dc.b	$F9,$7D,$68,$F9,$79,$7F,$04,$7F,$A9,$E1,$FC,$6F,$1F,$97,$39,$8F
	.dc.b	$C9,$63,$FE,$A5,$BE,$58,$7D,$EC,$3F,$8B,$1F,$99,$CF,$FC,$0F,$5F
	.dc.b	$32,$7F,$5A,$7A,$72,$F8,$55,$F1,$CF,$84,$FF,$0A,$3F,$A0,$1F,$C2
	.dc.b	$A3,$F8,$B6,$5F,$C5,$5F,$DF,$CF,$E3,$4F,$C9,$CF,$E3,$5F,$86,$7F
	.dc.b	$54,$8F,$86,$E9,$F8,$73,$EE,$8F,$D0,$1F,$0B,$3F,$E3,$CB,$1F,$E4
	.dc.b	$AF,$F6,$DB,$F3,$5F,$F3,$EC,$BF,$A5,$4F,$1C,$7E,$03,$FB,$19,$F8
	.dc.b	$07,$D3,$7E,$05,$FC,$AC,$FF,$43,$1F,$72,$FF,$31,$39,$8F,$71,$F3
	.dc.b	$4F,$EE,$58,$F9,$A3,$CB,$1F,$DE,$31,$F2,$BF,$C5,$8F,$F0,$D1,$FC
	.dc.b	$58,$FC,$59,$F9,$E7,$E9,$BF,$8B,$3E,$9E,$1F,$39,$FB,$B7,$FC,$35
	.dc.b	$E5,$3F,$A7,$7F,$8D,$5F,$ED,$23,$E9,$47,$23,$F7,$C7,$AA,$79,$3F
	.dc.b	$B9,$3F,$94,$57,$D6,$7D,$C7,$F8,$02,$7F,$A5,$3A,$C3,$F2,$99,$FF
	.dc.b	$1E,$8F,$F2,$64,$FF,$15,$78,$BF,$89,$7F,$AB,$E7,$F7,$DF,$EA,$D7
	.dc.b	$FC,$05,$FD,$74,$FA,$E3,$C5,$87,$9A,$3D,$79,$7D,$3A,$F8,$78,$7E
	.dc.b	$79,$E3,$7F,$9E,$7E,$63,$F7,$DF,$99,$3C,$6B,$CD,$7D,$6B,$F2,$75
	.dc.b	$ED,$3F,$27,$1F,$B0,$53,$FE,$E0,$7E,$61,$FB,$A3,$77,$F0,$E7,$FC
	.dc.b	$D3,$5F,$90,$9F,$D0,$1F,$90,$C3,$E1,$5F,$EB,$A7,$E1,$3F,$3A,$7F
	.dc.b	$0A,$7D,$CC,$FF,$95,$E7,$FD,$4A,$3C,$35,$FC,$06,$BD,$78,$FC,$37
	.dc.b	$91,$F2,$F3,$FA,$F5,$F8,$8A,$F8,$CF,$53,$F4,$17,$D7,$3F,$8D,$3E
	.dc.b	$99,$FB,$6F,$98,$FE,$06,$FE,$CD,$FC,$B5,$F8,$A3,$F4,$B8,$FE,$C9
	.dc.b	$FC,$2C,$F3,$CF,$E8,$47,$C3,$3E,$81,$F8,$43,$F9,$8D,$7A,$A7,$CE
	.dc.b	$7E,$13,$2F,$73,$E2,$9F,$D8,$3E,$23,$B4,$6A,$3F,$22,$76,$9F,$A0
	.dc.b	$F6,$9E,$51,$FE,$0E,$3F,$73,$1F,$D0,$CF,$E1,$87,$E3,$8F,$E1,$87
	.dc.b	$E3,$8F,$BA,$7E,$3B,$3F,$BB,$3F,$C5,$72,$FF,$2C,$4F,$2A,$F6,$3C
	.dc.b	$8F,$A4,$FF,$02,$C3,$F8,$8D,$F2,$9F,$E1,$53,$FE,$93,$7E,$D1,$C6
	.dc.b	$FE,$23,$F4,$0F,$31,$F4,$0F,$31,$F6,$8F,$80,$7C,$43,$F1,$EF,$27
	.dc.b	$F8,$03,$F1,$0F,$F0,$11,$F8,$F3,$E6,$BD,$E3,$D6,$F7,$3D,$71,$FB
	.dc.b	$1A,$F5,$C7,$D1,$8F,$89,$8F,$F8,$E6,$3F,$89,$1E,$2C,$FF,$2E,$E2
	.dc.b	$FF,$34,$FE,$B8,$F8,$27,$7A,$F4,$47,$F0,$F8,$FE,$91,$87,$F5,$77
	.dc.b	$F9,$09,$FC,$20,$F4,$4F,$B5,$FE,$AF,$87,$FA,$82,$3C,$B1,$FC,$E0
	.dc.b	$F1,$3F,$92,$BF,$2D,$FF,$14,$BF,$AD,$5F,$87,$BF,$E1,$13,$FC,$1F
	.dc.b	$3F,$D5,$CF,$D2,$9F,$D0,$BF,$87,$3E,$21,$F7,$B0,$FF,$5C,$BE,$E8
	.dc.b	$FE,$16,$F1,$CF,$F3,$8F,$B8,$FD,$91,$EE,$BF,$D9,$3F,$68,$FB,$CF
	.dc.b	$CA,$7F,$5C,$7E,$4E,$BE,$13,$FC,$33,$1F,$EC,$73,$F9,$3C,$FF,$82
	.dc.b	$65,$EE,$C7,$FB,$79,$EE,$C3,$E3,$D7,$E0,$A3,$CF,$87,$E0,$8F,$D2
	.dc.b	$BF,$82,$7E,$F1,$F6,$CF,$9A,$7E,$6F,$0F,$9B,$5F,$CD,$23,$67,$A2
	.dc.b	$7C,$0F,$C7,$3D,$B1,$F5,$5E,$95,$F5,$5F,$9A,$7D,$5C,$7F,$9B,$63
	.dc.b	$FE,$0E,$BF,$C0,$4F,$33,$FD,$11,$5C,$5F,$0B,$FC,$21,$F9,$07,$80
	.dc.b	$F4,$1E,$0B,$F4,$47,$CF,$38,$4F,$A2,$7F,$DB,$D8,$FC,$BB,$FE,$0B
	.dc.b	$39,$3F,$C8,$53,$E8,$7E,$7E,$1E,$8C,$FF,$D3,$38,$FE,$80,$FE,$69
	.dc.b	$5E,$33,$F4,$99,$7C,$A7,$F9,$71,$EA,$8F,$AE,$7E,$39,$FA,$27,$03
	.dc.b	$D4,$70,$CB,$D5,$1C,$23,$FD,$72,$7E,$57,$1F,$94,$7F,$A6,$9F,$AD
	.dc.b	$87,$9D,$FF,$53,$DF,$8D,$FD,$01,$E3,$3E,$86,$3F,$EA,$A9,$FC,$EE
	.dc.b	$7F,$E2,$CC,$7F,$9A,$63,$FC,$F6,$FF,$44,$7E,$E8,$F8,$17,$F9,$08
	.dc.b	$F8,$91,$FC,$78,$FD,$31,$FD,$18,$FD,$33,$FC,$69,$FD,$31,$A9,$EE
	.dc.b	$E8,$EE,$6A,$BC,$93,$FC,$93,$0F,$D3,$BF,$C7,$F1,$FD,$3D,$7D,$97
	.dc.b	$EE,$E3,$F0,$F0,$FF,$4C,$DF,$C3,$C7,$F4,$91,$F3,$AF,$E1,$DF,$F8
	.dc.b	$EF,$0F,$C5,$9F,$9D,$8F,$9A,$7D,$1C,$BF,$07,$87,$3A,$F0,$CF,$CD
	.dc.b	$3F,$40,$6C,$FA,$67,$DE,$3E,$9D,$F8,$5D,$9F,$84,$C3,$E6,$57,$5B
	.dc.b	$FF,$49,$67,$F9,$97,$F9,$D5,$FE,$9C,$E5,$5F,$98,$BF,$04,$7C,$07
	.dc.b	$C1,$1F,$03,$1F,$64,$FE,$9B,$2F,$8D,$1F,$EB,$9C,$7C,$17,$FE,$32
	.dc.b	$8F,$C9,$3F,$5D,$FA,$4F,$F9,$76,$7F,$13,$8F,$F4,$18,$FF,$1B,$1F
	.dc.b	$20,$F6,$65,$F9,$8C,$BF,$DA,$99,$79,$E7,$31,$F3,$1F,$3C,$FC,$CC
	.dc.b	$BD,$AF,$DC,$8F,$6E,$3E,$77,$75,$F1,$9F,$C1,$3F,$1A,$3E,$DC,$7F
	.dc.b	$08,$3C,$95,$F2,$63,$E1,$9E,$EB,$FD,$24,$FE,$7C,$F8,$07,$F5,$8B
	.dc.b	$FD,$85,$7C,$98,$F1,$9F,$ED,$3C,$BE,$4C,$7F,$53,$C7,$EC,$4F,$F5
	.dc.b	$D8,$F0,$3E,$37,$93,$F0,$8F,$C3,$9E,$6C,$3D,$2F,$F0,$73,$07,$F0
	.dc.b	$77,$D2,$FE,$6B,$1F,$F2,$C9,$F0,$DF,$CA,$63,$F3,$EB,$C3,$1C,$DF
	.dc.b	$09,$88,$FC,$8C,$7C,$C3,$F0,$47,$F3,$7C,$FF,$66,$FC,$3A,$FB,$79
	.dc.b	$7D,$58,$F5,$5F,$E4,$5F,$39,$FA,$0C,$3C,$EF,$E8,$23,$C1,$3F,$85
	.dc.b	$8E,$E7,$E2,$A3,$CB,$3F,$0B,$0F,$2C,$F4,$7F,$38,$FC,$B8,$FB,$55
	.dc.b	$FB,$B8,$F8,$C6,$4F,$8C,$7F,$9C,$AF,$E3,$5F,$E3,$9F,$C4,$47,$03
	.dc.b	$E3,$57,$48,$F9,$B1,$E6,$BF,$E9,$2F,$CE,$BF,$85,$7E,$CA,$FC,$E4
	.dc.b	$F9,$E7,$F3,$97,$F0,$AF,$EA,$3E,$D8,$F8,$57,$FE,$64,$9C,$D7,$8D
	.dc.b	$F8,$31,$F1,$23,$13,$C2,$BF,$9C,$47,$F8,$0C,$FF,$29,$1F,$DA,$8F
	.dc.b	$CD,$9F,$E1,$13,$F7,$47,$E7,$F0,$FB,$AE,$4F,$F1,$D1,$FC,$CF,$1F
	.dc.b	$DF,$4F,$F9,$56,$3F,$9B,$DF,$F9,$0E,$B8,$9F,$E3,$28,$F1,$61,$FE
	.dc.b	$32,$3E,$BB,$F3,$0F,$A7,$3F,$E2,$1C,$7E,$9D,$FC,$5C,$7E,$5B,$F1
	.dc.b	$63,$E5,$CF,$C1,$AF,$C6,$1F,$74,$FC,$64,$ED,$FD,$5C,$6E,$FC,$27
	.dc.b	$CA,$9F,$B9,$3F,$83,$3E,$E5,$AF,$FD,$59,$5B,$9F,$3D,$7F,$51,$9F
	.dc.b	$3C,$78,$67,$EB,$BE,$43,$DC,$79,$0F,$AF,$1E,$A9,$F0,$DF,$C1,$9F
	.dc.b	$0C,$F8,$2F,$EE,$47,$F6,$53,$E5,$E7,$FC,$C1,$FF,$29,$9F,$A7,$BE
	.dc.b	$67,$95,$E6,$7F,$A5,$B0,$F5,$3E,$33,$CC,$E0,$F4,$9F,$1A,$BC,$31
	.dc.b	$D4,$F8,$27,$F9,$39,$FC,$DB,$F0,$5F,$AE,$FB,$8F,$A3,$3E,$47,$85
	.dc.b	$FC,$87,$E2,$D7,$82,$B9,$C6,$1E,$6F,$99,$F1,$D7,$E4,$63,$E1,$9F
	.dc.b	$99,$CB,$F2,$B9,$FF,$81,$A3,$B9,$F0,$B2,$FF,$34,$5F,$29,$D4,$FE
	.dc.b	$9B,$0F,$F2,$24,$FE,$06,$3E,$43,$B7,$F4,$E7,$E8,$CF,$A3,$87,$8C
	.dc.b	$FF,$11,$BF,$AF,$CF,$F4,$78,$FF,$42,$C7,$FD,$99,$3F,$C9,$AB,$FA
	.dc.b	$D3,$FC,$17,$2F,$89,$8F,$D3,$C7,$E0,$BF,$4E,$3F,$11,$3F,$48,$F4
	.dc.b	$67,$FD,$E9,$FE,$A4,$7F,$39,$C3,$FA,$93,$FD,$90,$FE,$A5,$87,$E9
	.dc.b	$63,$F9,$B0,$7F,$47,$03,$FD,$6A,$53,$16,$C3,$E3,$46,$44,$67,$0C
	.dc.b	$47,$09,$DB,$D0,$86,$63,$6F,$28,$DB,$93,$9C,$F5,$6C,$69,$F0,$FA
	.dc.b	$4F,$A8,$E7,$48,$C4,$F2,$6B,$E9,$8D,$4F,$09,$CC,$FC,$19,$CB,$C6
	.dc.b	$74,$47,$04,$3A,$4F,$2F,$A9,$10

;--------------------------------------------------------------------------------
;	FETOXM1.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fetoxm1_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FETOXM1.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FETOXM1.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fetoxm1_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fetoxm1.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FETOXM1.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fetoxm1.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FETOXM1.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fetoxm1_data_compressed::
	.dc.b	$00,$02,$84,$2C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CF,$ED,$1F,$8A,$7F,$87,$A7,$FC,$D0,$5F,$8A,$E1
	.dc.b	$F0,$FC,$7D,$FE,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$1F
	.dc.b	$07,$9F,$DB,$F4,$7B,$7F,$DF,$7E,$6F,$FA,$B0,$F8,$8C,$7C,$46,$7E
	.dc.b	$23,$5F,$11,$BF,$88,$E1,$F1,$1C,$7E,$17,$97,$C0,$F3,$F8,$1E,$9F
	.dc.b	$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03
	.dc.b	$E3,$F8,$1F,$27,$C0,$F9,$7E,$07,$CD,$F0,$3E,$7F,$81,$FF,$0A,$7C
	.dc.b	$E7,$E1,$BE,$13,$FC,$40,$BE,$07,$3E,$66,$BE,$07,$7F,$03,$C3,$E0
	.dc.b	$78,$FC,$0F,$2F,$81,$E7,$F0,$3F,$63,$D2,$EB,$F0,$3D,$BE,$07,$BF
	.dc.b	$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93,$E0,$7C,$BF
	.dc.b	$03,$E6,$F8,$1F,$3F,$C0,$FA,$3E,$07,$FC,$59,$F3,$BF,$88,$E9,$FF
	.dc.b	$1C,$7C,$07,$F0,$A5,$1F,$EF,$63,$FD,$60,$7F,$9A,$0F,$EB,$47,$F1
	.dc.b	$23,$81,$FC,$41,$FF,$21,$2F,$7B,$BF,$BD,$E0,$F7,$BC,$3E,$F7,$8B
	.dc.b	$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD,$E8,$F7,$BF
	.dc.b	$3B,$EF,$FE,$7B,$5F,$CF,$EB,$FA,$0D,$7F,$43,$AF,$BB,$DE,$F7,$FB
	.dc.b	$DF,$E4,$BF,$6F,$FC,$48,$7E,$3F,$8B,$FC,$B0,$BD,$CD,$3B,$F7,$38
	.dc.b	$7B,$9C,$7D,$CE,$5E,$E7,$3F,$73,$A7,$B9,$FB,$0E,$8E,$DE,$E7,$7F
	.dc.b	$73,$C1,$EE,$7E,$D3,$AF,$F6,$BD,$7F,$B6,$EB,$FD,$BF,$5F,$EE,$3A
	.dc.b	$FF,$73,$D7,$FB,$AE,$6F,$DD,$F3,$7F,$98,$FF,$15,$FF,$1E,$1E,$9E
	.dc.b	$73,$E9,$7F,$CE,$4F,$FB,$38,$FF,$44,$1F,$DE,$0F,$E6,$07,$F2,$F7
	.dc.b	$8F,$0F,$5E,$1E,$DC,$1D,$F8,$3E,$47,$BB,$F2,$7D,$DF,$95,$EE,$FC
	.dc.b	$BF,$77,$E6,$7B,$BF,$37,$DD,$F9,$DE,$EF,$CF,$F7,$7E,$87,$BB,$FE
	.dc.b	$7B,$F9,$FF,$D4,$70,$1F,$D4,$3F,$E8,$95,$EE,$6F,$DC,$E1,$EE,$71
	.dc.b	$F7,$39,$7B,$9C,$FD,$CE,$9E,$E7,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B
	.dc.b	$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$F3,$7C,$9F,$9D,$D9,$F3
	.dc.b	$FB,$3E,$87,$67,$D1,$EC,$FA,$5D,$9F,$4F,$B3,$EA,$76,$7D,$5E,$CF
	.dc.b	$AD,$D9,$FE,$99,$F4,$FD,$E7,$BB,$B4,$FB,$9F,$F5,$42,$EB,$E1,$D7
	.dc.b	$C7,$AF,$97,$5F,$3E,$BE,$9D,$7D,$7A,$FE,$67,$B7,$F3,$7D,$BF,$9D
	.dc.b	$ED,$F8,$7B,$BE,$87,$B7,$FE,$AB,$FA,$0F,$E0,$1C,$05,$F0,$99,$F8
	.dc.b	$AD,$7C,$AE,$FE,$17,$87,$C0,$F1,$F8,$6E,$5F,$01,$CF,$E0,$3A,$74
	.dc.b	$75,$F8,$0F,$81,$E4,$FC,$1F,$27,$E1,$79,$3F,$0F,$C9,$F8,$9E,$4F
	.dc.b	$8F,$E0,$3C,$9E,$D7,$97,$DA,$F3,$7B,$5E,$7F,$6B,$D1,$ED,$7A,$7D
	.dc.b	$AF,$57,$B5,$EB,$F6,$BD,$9E,$D7,$FB,$1F,$E8,$7F,$D4,$07,$E9,$4F
	.dc.b	$D2,$76,$97,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1F,$67,$D9
	.dc.b	$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BB,$E1,$F7,$7C,$5E,$EF
	.dc.b	$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE,$F9,$FD,$DF,$47,$BB,$F8,$7F
	.dc.b	$57,$D5,$EE,$FF,$B8,$7C,$FF,$D3,$9F,$89,$3F,$11,$D2,$7F,$10,$FF
	.dc.b	$BC,$5F,$F7,$61,$FE,$A5,$3F,$CB,$AE,$FD,$9E,$9E,$CF,$5F,$67,$B7
	.dc.b	$AF,$DF,$CC,$F0,$79,$9E,$1F,$33,$C5,$E6,$78,$FC,$CF,$27,$67,$97
	.dc.b	$7F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$FB,$E7,$E8,$BF,$1E,$7E,$3B,$84
	.dc.b	$17,$B1,$BE,$3E,$1E,$C7,$1F,$63,$97,$B1,$CF,$D8,$E9,$EC,$75,$F6
	.dc.b	$3B,$7B,$1F,$AF,$DF,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E
	.dc.b	$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7A,$7D,$8F,$3F,$3B,$E8
	.dc.b	$FF,$DD,$87,$F3,$13,$F1,$A7,$E3,$1F,$6A,$F5,$F8,$BF,$EA,$83,$FC
	.dc.b	$C1,$C2,$EB,$CF,$AB,$AF,$57,$7E,$AF,$0F,$57,$8F,$AB,$CB,$D5,$E7
	.dc.b	$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E,$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F
	.dc.b	$57,$C7,$EA,$FB,$3E,$93,$F9,$99,$FC,$54,$FC,$69,$F8,$CE,$62,$F7
	.dc.b	$F8,$7B,$FC,$7D,$FE,$5E,$CF,$3F,$67,$A7,$B3,$D7,$D9,$ED,$EC,$F7
	.dc.b	$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9,$EC,$F9,$7C,$FF
	.dc.b	$37,$85,$E7,$F6,$7E,$9F,$C0,$7F,$1A,$3F,$19,$CB,$FC,$A7,$D9,$CF
	.dc.b	$B3,$AF,$67,$7E,$CF,$0F,$67,$8F,$B3,$CB,$D9,$E7,$E6,$74,$F6,$7A
	.dc.b	$FB,$3D,$BD,$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F,$67,$C7,$EC,$F9
	.dc.b	$3D,$9F,$2F,$B3,$E6,$F6,$7F,$13,$F4,$BF,$8A,$E6,$3F,$DE,$4F,$E2
	.dc.b	$A7,$E2,$5C,$FB,$5C,$3D,$DE,$3E,$EF,$2F,$6F,$9F,$B7,$D3,$DB,$EB
	.dc.b	$ED,$F6,$F6,$FB,$FB,$7E,$0F,$6F,$C3,$ED,$F8,$BD,$BF,$1F,$B7,$E4
	.dc.b	$F6,$FC,$BE,$DF,$9B,$DB,$F3,$FB,$7E,$8F,$6F,$F4,$3F,$4D,$FD,$E0
	.dc.b	$FE,$B4,$7F,$12,$3F,$0C,$7E,$17,$97,$F7,$AB,$DA,$E3,$ED,$72,$F6
	.dc.b	$B9,$FB,$5D,$3D,$AE,$BE,$D7,$6F,$6B,$BF,$B5,$E0,$F6,$BC,$3E,$D7
	.dc.b	$8B,$DA,$F1,$FB,$5E,$4F,$6B,$CB,$ED,$79,$BD,$AF,$3F,$B5,$FB,$8F
	.dc.b	$A7,$FE,$BE,$7E,$18,$FE,$1F,$C9,$FC,$49,$FE,$54,$DF,$63,$87,$B1
	.dc.b	$C7,$D8,$E5,$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1
	.dc.b	$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E
	.dc.b	$7F,$63,$F8,$7F,$AF,$F8,$73,$FC,$7C,$7F,$5A,$3F,$0C,$7F,$0D,$E4
	.dc.b	$FE,$56,$BD,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6,$F5
	.dc.b	$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC,$BE
	.dc.b	$BF,$9B,$D7,$F3,$FA,$FE,$8F,$5F,$F9,$3F,$D4,$7F,$8F,$8F,$F1,$01
	.dc.b	$FD,$60,$FE,$18,$7E,$1F,$94,$FE,$1D,$FE,$8A,$FF,$58,$1E,$AF,$4F
	.dc.b	$57,$AF,$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E
	.dc.b	$AF,$93,$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FA,$3D,$5F,$E7,$FF,$53
	.dc.b	$FE,$26,$3F,$86,$9F,$88,$E2,$3F,$88,$6F,$AB,$C3,$D5,$E3,$EA,$F2
	.dc.b	$F5,$79,$FA,$BD,$3D,$5E,$BE,$AF,$6F,$57,$BF,$AB,$E0,$F5,$7C,$3E
	.dc.b	$AF,$8B,$D5,$F1,$FA,$BE,$4F,$57,$CB,$EA,$F9,$BD,$5F,$3F,$AB,$FD
	.dc.b	$5B,$EA,$BF,$87,$1F,$E8,$53,$FC,$42,$7F,$0D,$3F,$11,$C8,$7F,$10
	.dc.b	$FF,$74,$5E,$A7,$3F,$53,$A7,$A9,$D7,$D4,$ED,$EA,$77,$F5,$3C,$1E
	.dc.b	$A7,$87,$D4,$F1,$7A,$9E,$3F,$53,$C9,$EA,$79,$7D,$4F,$37,$A9,$E7
	.dc.b	$F5,$3D,$1E,$A7,$A7,$D4,$FE,$DF,$E2,$7F,$A1,$8F,$F3,$41,$FE,$20
	.dc.b	$3F,$AA,$1F,$89,$3F,$11,$C8,$7F,$C3,$E7,$E2,$1F,$F0,$7A,$F4,$BA
	.dc.b	$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA,$5E,$2F,$4B,$C7,$E9,$79
	.dc.b	$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$A3,$D2,$FF,$03,$FD,$5F,$F9
	.dc.b	$B8,$FE,$AA,$7E,$24,$FC,$41,$E7,$E1,$FF,$11,$AF,$4B,$8F,$A5,$CB
	.dc.b	$D2,$E7,$E9,$74,$F4,$BA,$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA
	.dc.b	$5E,$2F,$4B,$C7,$E9,$79,$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$F8
	.dc.b	$7B,$EB,$3F,$AB,$1F,$EC,$D3,$FC,$D2,$7F,$55,$3F,$12,$7E,$23,$90
	.dc.b	$FE,$21,$FF,$1E,$BF,$EC,$D1,$E8,$75,$F4,$3B,$7A,$1D,$FD,$0F,$07
	.dc.b	$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F,$43,$CD,$E8,$79
	.dc.b	$FD,$0F,$47,$A1,$E9,$F4,$3F,$C6,$FF,$5B,$FE,$CE,$3F,$D6,$07,$F9
	.dc.b	$A0,$FF,$0E,$1F,$C4,$4F,$C4,$9F,$E6,$63,$F1,$1C,$47,$F1,$0F,$F9
	.dc.b	$61,$79,$FD,$BC,$FE,$FE,$7F,$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7
	.dc.b	$F9,$3C,$FF,$2F,$9F,$E6,$F3,$FC,$FE,$7F,$A3,$CF,$FF,$27,$FC,$17
	.dc.b	$FA,$D8,$FF,$0E,$9F,$C4,$4F,$E2,$07,$E2,$78,$3F,$C6,$9E,$7F,$0F
	.dc.b	$3F,$8F,$9F,$CB,$CF,$E7,$E7,$F4,$F3,$FA,$F9,$FD,$BC,$FE,$FE,$7F
	.dc.b	$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7,$F9,$3C,$FF,$2F,$9F,$E6,$F3
	.dc.b	$FF,$CC,$BF,$5D,$FE,$1D,$1F,$27,$8F,$8F,$CF,$C4,$EB,$DE,$DF,$BD
	.dc.b	$C3,$DA,$E3,$ED,$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$FE,$7C,$FA,$F7,$CB,$DF,$94,$C7,$C8,$67,$E2,$B5
	.dc.b	$E7,$6F,$E1,$B8,$7C,$37,$1F,$84,$E5,$F0,$3C,$FD,$EE,$9E,$F7,$5F
	.dc.b	$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27
	.dc.b	$BD,$E5,$F7,$BC,$DE,$F7,$FA,$4B,$EC,$3F,$DF,$27,$F9,$80,$FE,$A8
	.dc.b	$7F,$53,$3F,$08,$7F,$0B,$F7,$FF,$D5,$EB,$DE,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$FD,$53,$EC,$FF,$98
	.dc.b	$4F,$C2,$F1,$FF,$B0,$9F,$F7,$A9,$FE,$61,$3F,$AA,$9F,$C3,$4F,$E1
	.dc.b	$8F,$F8,$77,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8E,$FE,$C7,$83
	.dc.b	$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79,$3D,$8F,$2F,$B1,$E6,$F6,$3C
	.dc.b	$FE,$C7,$A3,$D8,$FF,$60,$FD,$8F,$E1,$B8,$BF,$DB,$2B,$D7,$CF,$AF
	.dc.b	$AF,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D
	.dc.b	$7F,$2F,$AF,$E6,$F5,$FC,$FE,$BF,$FB,$63,$EC,$9F,$35,$7E,$4F,$1F
	.dc.b	$1F,$9F,$8F,$D7,$C4,$EF,$E1,$B8,$7B,$BC,$7D,$DE,$5E,$EF,$3F,$77
	.dc.b	$A7,$BB,$D7,$DD,$ED,$EE,$F7,$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE
	.dc.b	$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$FE,$EE,$FB,$3F,$F5,$31,$F8
	.dc.b	$6F,$38,$BE,$17,$5F,$0B,$BF,$85,$E1,$F0,$3C,$7E,$17,$97,$C1,$73
	.dc.b	$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F,$82,$F0,$7C,$17,$87,$E0
	.dc.b	$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E,$0B,$CD,$F0,$5E,$7F,$82
	.dc.b	$F4,$7C,$13,$9E,$78,$FF,$87,$7F,$BE,$7C,$0E,$5F,$EF,$DF,$03,$67
	.dc.b	$FD,$DC,$7F,$98,$CF,$F0,$D8,$F7,$B5,$EF,$6F,$DE,$E1,$EF,$71,$F7
	.dc.b	$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE
	.dc.b	$1F,$7B,$C5,$EF,$78,$FD,$EF,$37,$DA,$7E,$23,$93,$D8,$FF,$BB,$8F
	.dc.b	$F3,$11,$FE,$61,$3F,$A9,$1F,$C3,$D3,$FC,$3D,$E3,$EC,$74,$F6,$3A
	.dc.b	$FB,$1D,$BD,$8E,$FE,$C7,$83,$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79
	.dc.b	$3D,$8F,$2F,$B1,$E6,$F6,$3C,$FE,$C7,$A3,$D8,$F9,$DF,$6B,$FE,$EF
	.dc.b	$3F,$14,$7E,$27,$87,$EB,$3F,$EE,$F3,$FC,$C6,$7F,$86,$CF,$E1,$E3
	.dc.b	$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD,$3E,$FE,$9F,$83,$D3,$F0,$FA
	.dc.b	$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F,$A7,$E6,$F4,$FC,$FE,$9F,$A3
	.dc.b	$D3,$FC,$37,$DB,$7E,$24,$FE,$1D,$C7,$F8,$D5,$E9,$EB,$E1,$37,$F0
	.dc.b	$9C,$3E,$13,$8F,$AB,$CB,$D5,$E7,$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E
	.dc.b	$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F,$57,$C7,$EA,$F9,$3D,$5F,$2F,$AB
	.dc.b	$E6,$F5,$7C,$FE,$AF,$E7,$78,$7F,$11,$CD,$FF,$00,$7E,$0B,$1F,$05
	.dc.b	$9F,$82,$D7,$C1,$6F,$DA,$E1,$F0,$5C,$7D,$AE,$5E,$D7,$3F,$6B,$A7
	.dc.b	$B5,$D7,$DA,$ED,$ED,$77,$F6,$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F
	.dc.b	$6B,$C9,$ED,$79,$7D,$AF,$37,$B5,$FB,$3F,$B7,$FD,$A7,$39,$FE,$20
	.dc.b	$7E,$1D,$C7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$D1,$EF,$7F,$0B,$FB
	.dc.b	$8F,$E1,$C7,$E1,$B9,$8F,$E1,$9F,$E2,$8B,$DD,$E1,$EE,$F1,$F3,$B9
	.dc.b	$7B,$BC,$FD,$DE,$9E,$EF,$5F,$77,$B7,$BB,$DF,$DD,$F0,$7B,$BE,$1F
	.dc.b	$77,$C5,$EE,$F8,$FD,$DF,$27,$BB,$E5,$F7,$7C,$DE,$EF,$9F,$DD,$F4
	.dc.b	$7B,$BF,$C8,$FE,$E7,$F0,$E3,$DE,$C7,$BD,$9F,$7B,$5E,$F6,$FD,$EE
	.dc.b	$1E,$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$75,$F7,$BB,$7B,$DD,$FD
	.dc.b	$EF,$07,$BD,$E1,$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD
	.dc.b	$EF,$7F,$39,$F6,$BF,$0E,$7F,$55,$3F,$0C,$3D,$DD,$7B,$BB,$F7,$78
	.dc.b	$4F,$1F,$77,$97,$BB,$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7B,$BD,$FD,$DF
	.dc.b	$07,$BB,$E1,$F7,$7C,$5E,$EF,$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE
	.dc.b	$F9,$FD,$DF,$EA,$1F,$75,$FD,$4C,$FC,$31,$F8,$5F,$78,$FE,$15,$FE
	.dc.b	$B4,$BD,$CE,$3E,$E7,$2F,$73,$9F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B
	.dc.b	$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7
	.dc.b	$9B,$DC,$F3,$FB,$9E,$8F,$73,$FB,$47,$DD,$FF,$0C,$3F,$0D,$CB,$FD
	.dc.b	$BD,$7B,$7A,$F6,$F7,$ED,$F0,$F6,$F8,$FB,$7C,$BD,$BE,$7E,$DF,$4F
	.dc.b	$6D,$FC,$03,$0B,$A7,$F7,$BF,$A6,$78,$5F,$DE,$E3,$FB,$D4,$63,$E3
	.dc.b	$BF,$84,$C6,$E7,$4D,$8C,$A9,$E0,$F0,$6B,$FC,$49,$AF,$69,$E6,$F1
	.dc.b	$6C,$6A,$3B,$1C,$E3,$27,$F9,$5F,$EE,$FF,$C2,$52,$F1,$31,$7D,$4F
	.dc.b	$F0,$C4,$EA,$33,$7D,$23,$97,$E2,$0E,$5C,$CF,$8E,$F8,$9C,$A7,$4E
	.dc.b	$23,$93,$DB,$F1,$17,$BB,$CB,$2F,$87,$8F,$9E,$73,$5C,$6F,$B0,$9F
	.dc.b	$34,$66,$B8,$BF,$53,$2E,$99,$FA,$A9,$97,$93,$0E,$57,$F0,$5A,$75
	.dc.b	$8D,$4F,$C8,$AE,$58,$F9,$2B,$67,$92,$7D,$19,$78,$30,$F2,$63,$6B
	.dc.b	$79,$7D,$19,$F4,$63,$E5,$CB,$BC,$78,$F1,$F9,$F9,$F1,$7A,$C7,$03
	.dc.b	$E6,$2C,$FC,$8F,$9E,$3C,$26,$F1,$F0,$E5,$CE,$3E,$B5,$D9,$F8,$0D
	.dc.b	$AE,$99,$DC,$3A,$DF,$94,$F9,$C7,$59,$FA,$77,$EE,$CB,$D9,$87,$3C
	.dc.b	$BD,$B3,$F8,$23,$ED,$61,$D1,$C5,$F8,$30,$E8,$76,$8E,$8F,$B1,$A7
	.dc.b	$D8,$8F,$BB,$7D,$CF,$46,$32,$B9,$D7,$EE,$F2,$F8,$0F,$EC,$DE,$11
	.dc.b	$AB,$E6,$DC,$3C,$D5,$F7,$E3,$F3,$57,$CB,$3F,$B9,$3E,$4B,$F8,$67
	.dc.b	$D3,$7C,$D9,$7C,$27,$91,$F4,$2F,$B5,$F9,$E3,$94,$7C,$1A,$E4,$74
	.dc.b	$7E,$94,$71,$36,$58,$DC,$79,$8E,$A7,$C2,$8F,$80,$DB,$F4,$CF,$09
	.dc.b	$E5,$3D,$A3,$E1,$C7,$67,$EB,$37,$0F,$8B,$5F,$B2,$8F,$9D,$5D,$71
	.dc.b	$F9,$B1,$C7,$1F,$85,$5B,$C3,$99,$E1,$2D,$FB,$AB,$CE,$7B,$8B,$1B
	.dc.b	$AE,$79,$FE,$42,$BD,$F5,$E5,$CB,$EB,$63,$27,$EF,$5F,$07,$F5,$6D
	.dc.b	$AF,$C6,$65,$EC,$C3,$F8,$0E,$1C,$67,$C5,$97,$C8,$C3,$E3,$61,$C7
	.dc.b	$2F,$DC,$C7,$89,$F6,$63,$CE,$3F,$35,$1F,$3E,$FD,$71,$CB,$3F,$8F
	.dc.b	$3C,$31,$FC,$23,$BC,$7C,$17,$FA,$B3,$89,$FA,$AC,$7C,$58,$7E,$F8
	.dc.b	$F1,$1F,$6F,$1F,$14,$FF,$04,$8F,$AC,$7C,$4C,$7C,$2E,$F1,$F0,$BF
	.dc.b	$4F,$0F,$0D,$7E,$68,$F0,$9F,$9C,$C7,$F0,$E7,$D5,$BE,$31,$FB,$EC
	.dc.b	$3C,$33,$F9,$B8,$F5,$57,$43,$F0,$98,$FA,$A3,$F2,$67,$1C,$79,$BE
	.dc.b	$19,$E1,$97,$F4,$BC,$FD,$33,$F0,$F2,$DD,$7E,$3F,$2F,$0E,$7F,$AD
	.dc.b	$3D,$97,$C2,$FD,$59,$78,$73,$FC,$9C,$78,$6B,$E3,$BF,$82,$AE,$18
	.dc.b	$7A,$31,$EF,$97,$E2,$2F,$78,$7D,$9B,$DC,$FE,$5F,$0F,$CB,$57,$D6
	.dc.b	$7E,$1C,$FE,$3E,$BE,$54,$FD,$53,$53,$F3,$F2,$F5,$BF,$B0,$3C,$91
	.dc.b	$B7,$F6,$18,$79,$23,$FA,$4C,$F0,$9F,$B3,$8F,$EE,$5B,$8F,$97,$0F
	.dc.b	$0C,$79,$88,$FA,$EF,$DA,$7E,$64,$F0,$3F,$99,$7C,$F5,$D4,$FA,$EF
	.dc.b	$E8,$8E,$13,$E6,$77,$5F,$26,$3D,$64,$AF,$19,$27,$F2,$C7,$03,$A4
	.dc.b	$7C,$A7,$C6,$76,$8F,$24,$7A,$A3,$F0,$27,$A9,$FC,$43,$C1,$FC,$01
	.dc.b	$B3,$A1,$E3,$7A,$11,$DC,$76,$3E,$33,$B3,$A9,$EF,$3A,$1F,$6F,$3F
	.dc.b	$C0,$4F,$92,$35,$3E,$F7,$AC,$7C,$07,$A1,$FA,$3C,$BF,$1B,$1F,$C7
	.dc.b	$32,$FC,$04,$FF,$1C,$BE,$AF,$AA,$FA,$3F,$A0,$9D,$C7,$F8,$42,$3A
	.dc.b	$E3,$F6,$70,$EB,$1F,$D7,$2B,$78,$7F,$4F,$8D,$E3,$FB,$38,$F9,$51
	.dc.b	$D5,$FA,$CF,$8D,$FA,$0F,$A5,$DD,$FE,$AD,$DE,$1E,$33,$E5,$91,$FA
	.dc.b	$13,$D4,$F9,$73,$FB,$FC,$7D,$F1,$D4,$F9,$73,$C2,$74,$F6,$3F,$82
	.dc.b	$47,$C5,$75,$7F,$A3,$7E,$85,$FD,$D7,$EB,$3E,$39,$FB,$AF,$CC,$AE
	.dc.b	$B1,$E4,$AF,$88,$FC,$6A,$FA,$C7,$C0,$8F,$D0,$57,$59,$F1,$10,$F7
	.dc.b	$3D,$63,$73,$E4,$3C,$64,$9F,$1B,$EA,$3F,$80,$DF,$D8,$8F,$E1,$15
	.dc.b	$E9,$7A,$CE,$DF,$AC,$7A,$5F,$D6,$BF,$2E,$3C,$51,$EA,$8D,$1C,$23
	.dc.b	$A5,$FE,$26,$7D,$C6,$E3,$DD,$3F,$A1,$8F,$C1,$CE,$E3,$EB,$9E,$A3
	.dc.b	$F4,$24,$3C,$51,$EF,$7D,$26,$CE,$87,$92,$7F,$A8,$3C,$0D,$4F,$E3
	.dc.b	$DD,$4F,$E6,$E7,$D3,$7F,$61,$F7,$CF,$E8,$67,$F1,$0F,$F9,$72,$FF
	.dc.b	$1F,$7E,$9A,$D9,$FE,$04,$9E,$A6,$EB,$DF,$7D,$5F,$A1,$3B,$3E,$2D
	.dc.b	$7E,$29,$F1,$C7,$D2,$8F,$44,$76,$8F,$41,$F8,$3A,$FC,$44,$F8,$E3
	.dc.b	$A9,$FE,$05,$8F,$1C,$FF,$5E,$7E,$21,$27,$AC,$F0,$8F,$41,$EA,$3F
	.dc.b	$A3,$BE,$3A,$FD,$34,$7C,$C7,$F2,$AF,$D7,$9F,$13,$E4,$8D,$BF,$60
	.dc.b	$FA,$F5,$F0,$23,$E2,$46,$EB,$F1,$07,$8E,$7D,$15,$F1,$63,$F0,$4F
	.dc.b	$69,$E8,$FA,$9F,$15,$79,$1F,$73,$D9,$DB,$E4,$8E,$91,$C2,$7D,$C4
	.dc.b	$7F,$04,$7E,$58,$F4,$BF,$45,$EA,$EE,$BE,$E3,$E9,$CB,$FB,$CC,$7A
	.dc.b	$AF,$C5,$7E,$93,$D1,$7E,$4B,$FD,$C9,$F9,$07,$E8,$9F,$E3,$B3,$C5
	.dc.b	$1C,$0F,$4B,$E4,$3E,$51,$BA,$FC,$43,$FE,$2E,$34,$48,$F1,$57,$90
	.dc.b	$F8,$84,$AF,$1C,$FE,$0E,$7F,$BE,$BD,$63,$EB,$67,$FE,$2B,$C3,$DE
	.dc.b	$7F,$84,$6B,$78,$7E,$D1,$F4,$61,$D7,$0F,$E1,$58,$7C,$B9,$F1,$BF
	.dc.b	$2C,$FC,$D5,$7A,$8F,$94,$F9,$1F,$D3,$CF,$8C,$FC,$A4,$7E,$29,$FB
	.dc.b	$87,$63,$E2,$C7,$8E,$BC,$4F,$D8,$8D,$BF,$D4,$E7,$F1,$F3,$F1,$5E
	.dc.b	$93,$F5,$CE,$87,$CB,$8D,$3E,$33,$53,$D5,$D4,$FA,$4E,$A7,$BD,$FA
	.dc.b	$D5,$F1,$A3,$51,$E4,$7F,$AC,$BF,$90,$3F,$2B,$3F,$16,$FE,$8E,$3F
	.dc.b	$C5,$B0,$F1,$CF,$87,$0E,$97,$E8,$7F,$12,$7D,$07,$D0,$EA,$38,$4F
	.dc.b	$EA,$E3,$6F,$4C,$3C,$71,$B7,$EB,$9F,$6E,$38,$1B,$AF,$EA,$71,$F8
	.dc.b	$8C,$BF,$AB,$5F,$DC,$9F,$0D,$6E,$3F,$D1,$F5,$FA,$BC,$3F,$AE,$57
	.dc.b	$E9,$27,$ED,$4F,$D4,$C3,$C5,$3F,$C6,$A7,$F8,$74,$FC,$D8,$E0,$F9
	.dc.b	$9F,$C3,$9F,$0E,$FC,$F1,$BA,$FC,$CB,$B9,$FB,$6F,$F3,$27,$EA,$1E
	.dc.b	$C3,$F1,$AE,$E3,$C7,$5E,$D3,$F1,$93,$A3,$81,$D0,$FC,$0D,$78,$9F
	.dc.b	$25,$78,$F3,$FF,$1A,$D7,$D4,$8F,$C2,$CF,$E3,$0F,$92,$FE,$2E,$3E
	.dc.b	$4C,$7B,$9F,$0B,$B7,$E4,$BC,$0F,$C5,$C7,$CF,$3F,$89,$5F,$C2,$8F
	.dc.b	$3B,$EC,$CF,$F8,$06,$3F,$52,$35,$3F,$38,$F5,$D7,$D5,$C7,$D9,$87
	.dc.b	$8A,$BF,$20,$F8,$8F,$65,$78,$8F,$9A,$FC,$9A,$FC,$7D,$7E,$17,$0F
	.dc.b	$F5,$D1,$E2,$8F,$3E,$3F,$A1,$7C,$04,$BF,$96,$4B,$F0,$47,$43,$4F
	.dc.b	$88,$F3,$5F,$A2,$35,$3F,$1E,$75,$1F,$7E,$75,$3F,$91,$8F,$DE,$BB
	.dc.b	$74,$76,$3F,$14,$FD,$37,$AC,$F9,$23,$E4,$9F,$0C,$D1,$D4,$FC,$51
	.dc.b	$F3,$DF,$BE,$FE,$2D,$F9,$0E,$CF,$E1,$0F,$53,$78,$7E,$5B,$3F,$E1
	.dc.b	$35,$FE,$18,$3F,$15,$8F,$CE,$CF,$F6,$F3,$FC,$F7,$2F,$9D,$8F,$97
	.dc.b	$0E,$59,$FE,$0E,$73,$97,$C6,$AF,$5B,$F3,$32,$FC,$CD,$FE,$09,$F8
	.dc.b	$27,$C6,$9E,$99,$FC,$CA,$F5,$C7,$E3,$4F,$81,$9F,$96,$BF,$C6,$F3
	.dc.b	$AB,$FC,$99,$F9,$97,$DF,$7F,$D0,$27,$E8,$CF,$F1,$C7,$F5,$C7,$F9
	.dc.b	$6A,$FC,$2F,$9A,$7E,$EC,$78,$67,$A5,$F8,$63,$FC,$5F,$97,$C5,$BF
	.dc.b	$95,$1C,$89,$87,$C1,$7F,$25,$1F,$01,$FD,$6E,$1E,$57,$E3,$C7,$9A
	.dc.b	$3C,$38,$7B,$31,$FE,$07,$3F,$1E,$FA,$11,$E9,$7F,$1B,$3D,$BF,$83
	.dc.b	$BF,$80,$FE,$54,$F0,$CF,$E5,$5F,$09,$BC,$3C,$38,$7D,$C9,$F0,$12
	.dc.b	$74,$FC,$73,$E5,$BF,$E1,$89,$F6,$61,$F9,$43,$EB,$57,$39,$EF,$1F
	.dc.b	$5A,$B9,$47,$C6,$3C,$A7,$9E,$34,$FC,$C8,$F8,$C7,$08,$FD,$54,$7C
	.dc.b	$18,$FB,$91,$F1,$63,$CE,$7C,$13,$47,$9E,$39,$1D,$1D,$5F,$E0,$30
	.dc.b	$F9,$75,$EB,$C7,$F0,$15,$FE,$C1,$3C,$87,$B1,$FD,$6B,$98,$FD,$96
	.dc.b	$5F,$89,$7F,$AB,$C7,$BC,$FC,$76,$3D,$A7,$F8,$F4,$7F,$8C,$E3,$E4
	.dc.b	$4F,$C2,$3F,$5C,$78,$F3,$F9,$93,$EC,$8F,$F0,$34,$7E,$D7,$2F,$CA
	.dc.b	$C7,$0B,$EA,$F8,$8F,$93,$97,$BE,$7E,$49,$F0,$E3,$E5,$5F,$9A,$32
	.dc.b	$FB,$EF,$B3,$EF,$8F,$C3,$4F,$9D,$E4,$48,$FB,$B1,$F1,$32,$F9,$B9
	.dc.b	$7C,$88,$FF,$1F,$99,$AE,$EF,$BD,$F5,$9F,$AB,$7B,$57,$8E,$3F,$A2
	.dc.b	$9A,$7F,$1F,$8F,$C0,$8F,$F0,$C1,$F9,$89,$FF,$28,$46,$67,$FB,$F9
	.dc.b	$EF,$3F,$0C,$7D,$E8,$F1,$3E,$F3,$D7,$3F,$02,$79,$47,$F3,$A3,$94
	.dc.b	$F6,$BF,$E5,$B8,$7D,$8C,$7E,$85,$76,$AF,$CC,$1F,$A8,$C3,$D4,$FB
	.dc.b	$89,$8F,$C7,$8F,$EA,$31,$F1,$E7,$94,$7C,$F8,$E5,$1E,$77,$96,$3E
	.dc.b	$6B,$FE,$0B,$1F,$82,$CB,$F2,$59,$79,$B0,$F9,$58,$7E,$35,$FF,$04
	.dc.b	$3F,$95,$7C,$B8,$7E,$75,$F9,$B1,$F9,$DC,$3F,$8D,$47,$DA,$9F,$06
	.dc.b	$7F,$C3,$A3,$8D,$FE,$0B,$3F,$C5,$E5,$F8,$29,$F9,$93,$E3,$C7,$E8
	.dc.b	$1E,$33,$FC,$41,$1E,$FC,$FF,$6B,$5C,$67,$F0,$35,$C6,$31,$1C,$4F
	.dc.b	$ED,$53,$F2,$DF,$ED,$F9,$7D,$49,$FD,$5D,$FE,$32,$7E,$56,$5E,$F7
	.dc.b	$FC,$33,$7E,$09,$FA,$67,$A0,$FE,$AE,$E6,$3F,$C8,$95,$F4,$B0,$F2
	.dc.b	$57,$09,$FC,$59,$E8,$78,$E3,$F8,$7C,$FE,$AC,$60,$FA,$B9,$79,$DF
	.dc.b	$AB,$87,$EF,$5F,$E8,$B1,$9B,$F8,$E7,$F2,$49,$FD,$EC,$7D,$68,$FC
	.dc.b	$49,$FA,$B3,$EF,$1F,$03,$0F,$AD,$5F,$0E,$BE,$71,$EB,$7C,$93,$F5
	.dc.b	$AB,$C0,$FC,$EC,$3E,$EC,$FB,$EB,$84,$FF,$6C,$9F,$74,$FE,$1E,$FE
	.dc.b	$2C,$FD,$4C,$3E,$0B,$B9,$F8,$31,$8B,$F8,$13,$ED,$9F,$83,$39,$7F
	.dc.b	$2B,$3F,$02,$3F,$96,$E5,$F6,$EB,$FC,$63,$3F,$A8,$3D,$B1,$9C,$7F
	.dc.b	$4F,$7F,$C6,$EB,$F7,$AF,$FA,$7C,$F8,$35,$F8,$E9,$F8,$27,$F8,$A2
	.dc.b	$3E,$41,$F0,$6B,$A6,$7F,$E2,$1C,$FF,$0D,$87,$F8,$EA,$BF,$14,$FB
	.dc.b	$CF,$99,$8F,$AD,$FC,$8E,$7F,$21,$FB,$39,$7F,$06,$8F,$9F,$1F,$90
	.dc.b	$C7,$BD,$FF,$60,$8F,$8C,$7E,$40,$FD,$26,$5E,$77,$F0,$37,$F0,$B3
	.dc.b	$FC,$0C,$62,$BD,$51,$89,$FA,$33,$F1,$0F,$C7,$BE,$39,$F8,$31,$E3
	.dc.b	$9E,$11,$F5,$0F,$AE,$7E,$04,$F9,$67,$E9,$63,$CB,$9F,$F8,$3B,$3F
	.dc.b	$DC,$47,$C7,$3F,$53,$7F,$8C,$3F,$06,$7D,$BA,$F3,$9F,$D3,$27,$8E
	.dc.b	$1F,$E5,$5A,$E3,$8F,$C6,$7F,$C2,$97,$C5,$F4,$5F,$1C,$BE,$6C,$78
	.dc.b	$67,$C1,$3E,$B8,$F9,$A7,$C3,$3F,$CF,$15,$FC,$8A,$3E,$94,$66,$FF
	.dc.b	$D6,$19,$7E,$82,$3D,$87,$C0,$9F,$F1,$C5,$FF,$52,$3F,$1D,$9F,$E0
	.dc.b	$E7,$F3,$39,$7E,$2B,$0F,$65,$79,$27,$F0,$F7,$FA,$83,$E7,$3F,$8D
	.dc.b	$8E,$71,$87,$E4,$D7,$3C,$FF,$D6,$98,$FD,$8B,$FE,$CF,$87,$F0,$9B
	.dc.b	$FE,$BD,$87,$D8,$9F,$DC,$3C,$CF,$89,$1F,$2E,$3E,$46,$5F,$3E,$3F
	.dc.b	$B0,$E3,$DE,$BF,$D2,$B8,$77,$BF,$ED,$98,$77,$9F,$D8,$47,$A5,$FC
	.dc.b	$34,$7E,$C3,$1F,$C3,$67,$F5,$A3,$E1,$E5,$DE,$BF,$A4,$4F,$EA,$72
	.dc.b	$F4,$E1,$CC,$F4,$9F,$AC,$AF,$B0,$FF,$5D,$CF,$F8,$4E,$5E,$5C,$7F
	.dc.b	$82,$CF,$EA,$DF,$16,$3F,$63,$3F,$C7,$E6,$8F,$F0,$91,$EA,$CB,$E7
	.dc.b	$E7,$E3,$8C,$1F,$22,$BE,$26,$1F,$E6,$5C,$FD,$4F,$EE,$AF,$E6,$63
	.dc.b	$EE,$CB,$C9,$97,$F6,$FA,$FE,$75,$8F,$E6,$4F,$D9,$BF,$4A,$77,$1F
	.dc.b	$94,$AF,$9D,$3F,$BA,$C3,$E0,$E1,$FC,$A6,$BF,$BC,$9C,$E7,$F8,$21
	.dc.b	$FA,$63,$F8,$9E,$3F,$02,$3D,$F1,$CF,$0F,$CF,$61,$FB,$18,$FC,$F1
	.dc.b	$FE,$2F,$AE,$0F,$F9,$D3,$3F,$CF,$4F,$CE,$C3,$DF,$97,$E5,$31,$F1
	.dc.b	$47,$F1,$29,$F4,$DF,$EC,$31,$EF,$7F,$46,$3A,$5F,$E5,$5F,$D2,$9F
	.dc.b	$CB,$9C,$3F,$AC,$CF,$EA,$3E,$98,$FC,$F3,$FB,$6B,$FB,$EF,$A6,$7F
	.dc.b	$21,$5E,$58,$F8,$71,$8A,$FC,$E9,$E8,$3F,$84,$47,$F0,$39,$FF,$55
	.dc.b	$E7,$F4,$27,$FA,$85,$FB,$0F,$E0,$EF,$89,$F9,$47,$8B,$1F,$F4,$3E
	.dc.b	$7F,$29,$FF,$1F,$4F,$B2,$7E,$54,$FE,$32,$7F,$C9,$47,$F8,$83,$0F
	.dc.b	$B6,$7E,$16,$BC,$CF,$F0,$0C,$FF,$43,$18,$8F,$DA,$C6,$1F,$7C,$7E
	.dc.b	$3A,$FF,$28,$ED,$FD,$3C,$7B,$E3,$F8,$EE,$1E,$F8,$FE,$53,$3E,$FC
	.dc.b	$7E,$1E,$7F,$6A,$FF,$17,$5F,$75,$F8,$53,$F7,$71,$F7,$CF,$EE,$6B
	.dc.b	$83,$F8,$03,$15,$C0,$E7,$7F,$32,$3F,$81,$1F,$77,$1F,$C8,$E1,$FD
	.dc.b	$DF,$3F,$C1,$CF,$84,$DD,$7D,$2A,$DD,$FF,$6F,$C7,$F2,$EF,$E0,$2B
	.dc.b	$E2,$9F,$3A,$BE,$2B,$89,$FC,$D5,$7C,$B7,$EA,$18,$AF,$D8,$CE,$DF
	.dc.b	$C4,$D7,$C7,$7E,$5E,$3F,$02,$BE,$19,$FC,$78,$FE,$1F,$1F,$D0,$EB
	.dc.b	$E1,$9E,$AC,$7E,$54,$7F,$74,$8C,$CF,$07,$F4,$C6,$23,$F8,$B6,$3F
	.dc.b	$13,$0F,$95,$5B,$3F,$A6,$D6,$CF,$D7,$B8,$3C,$95,$C8,$F2,$38,$8D
	.dc.b	$C7,$F3,$EC,$7E,$6C,$7F,$76,$CF,$F3,$53,$FE,$C5,$CF,$F9,$2D,$FF
	.dc.b	$67,$8E,$07,$E8,$72,$FB,$4F,$63,$F9,$7D,$F6,$3F,$A0,$CF,$85,$FE
	.dc.b	$05,$1B,$7F,$81,$1F,$1D,$FC,$4C,$FE,$59,$FC,$64,$7E,$3E,$79,$63
	.dc.b	$FB,$0C,$BE,$35,$EE,$B9,$1B,$BF,$1B,$E0,$3E,$7D,$F2,$9F,$4C,$62
	.dc.b	$77,$7D,$5F,$34,$7E,$56,$BE,$94,$7C,$28,$FA,$47,$58,$FA,$51,$CA
	.dc.b	$BA,$C6,$27,$DE,$F2,$7E,$2C,$EC,$FC,$F6,$3F,$4A,$3F,$C5,$79,$FF
	.dc.b	$4C,$7F,$B4,$4F,$E6,$2B,$F3,$6F,$E6,$33,$FC,$F3,$FE,$9A,$30,$6C
	.dc.b	$C4,$FF,$11,$70,$FF,$93,$AF,$F8,$F4,$7E,$5E,$BE,$33,$E4,$CB,$E7
	.dc.b	$BF,$A4,$3E,$AC,$7E,$91,$F5,$57,$E2,$E7,$6F,$FC,$01,$CF,$F8,$A5
	.dc.b	$7E,$EA,$3F,$B9,$63,$ED,$8F,$EC,$91,$FD,$B8,$FD,$A1,$E7,$30,$FC
	.dc.b	$07,$0F,$F9,$B6,$FF,$92,$65,$FD,$DE,$31,$3C,$8F,$E5,$B3,$F0,$1F
	.dc.b	$F5,$63,$FC,$B6,$7F,$A7,$4F,$59,$D9,$E0,$9F,$92,$E2,$3A,$9B,$CF
	.dc.b	$F9,$99,$FC,$26,$3E,$05,$FF,$09,$AF,$98,$7C,$6A,$FF,$1B,$E3,$F1
	.dc.b	$9F,$E6,$18,$FE,$C3,$2F,$A7,$87,$F8,$8F,$3F,$DB,$D7,$0C,$3F,$2B
	.dc.b	$97,$F0,$67,$F8,$05,$7E,$6C,$FC,$7B,$89,$FC,$94,$7E,$6E,$30,$7E
	.dc.b	$6E,$3E,$FC,$7D,$98,$EB,$9F,$EB,$67,$F3,$13,$FC,$EE,$7F,$81,$1F
	.dc.b	$85,$36,$7E,$15,$FD,$9C,$7E,$15,$F1,$9F,$C1,$4E,$47,$EA,$AB,$F6
	.dc.b	$0F,$A4,$F3,$1B,$7F,$A2,$4F,$82,$3C,$C6,$E3,$E1,$3B,$AF,$E6,$2F
	.dc.b	$C9,$9E,$15,$F2,$63,$84,$FB,$67,$E1,$C7,$ED,$0F,$39,$FB,$D7,$E3
	.dc.b	$1F,$64,$FF,$30,$9F,$01,$FB,$EF,$E9,$0F,$BE,$F2,$AF,$6D,$7D,$27
	.dc.b	$CF,$1D,$4F,$01,$F0,$0C,$1F,$1A,$3E,$A9,$E9,$39,$57,$A6,$3E,$05
	.dc.b	$7F,$2A,$3E,$94,$FD,$8A,$F9,$33,$F3,$EB,$F8,$AD,$7B,$23,$CD,$18
	.dc.b	$7C,$73,$E6,$3D,$33,$B8,$F3,$46,$1D,$CF,$8D,$F4,$C7,$82,$BD,$35
	.dc.b	$F2,$27,$E8,$BE,$63,$E8,$BE,$F7,$E7,$3F,$E7,$5C,$3F,$92,$3F,$E5
	.dc.b	$E9,$FF,$1A,$E5,$FE,$43,$8F,$E1,$75,$FC,$08,$FF,$39,$63,$FC,$92
	.dc.b	$3F,$8F,$4E,$EB,$FC,$6F,$5F,$BC,$C3,$F7,$37,$BB,$FE,$0A,$7F,$83
	.dc.b	$71,$FE,$0B,$5B,$7F,$87,$3F,$EE,$48,$FE,$1D,$5C,$A3,$F8,$72,$9F
	.dc.b	$F5,$5C,$7F,$0F,$3F,$C3,$57,$FC,$26,$39,$D7,$FA,$06,$31,$3E,$A3
	.dc.b	$9C,$F2,$9F,$24,$72,$3C,$86,$DF,$E3,$07,$C0,$7C,$0F,$C6,$3C,$F1
	.dc.b	$E0,$7F,$CA,$C6,$23,$94,$FF,$1E,$7A,$C7,$F5,$23,$A9,$F1,$A3,$FC
	.dc.b	$58,$E2,$BE,$04,$FF,$AE,$70,$FA,$0F,$F8,$37,$3F,$E8,$0F,$F9,$D3
	.dc.b	$1F,$A1,$5C,$2F,$FD,$05,$3E,$8C,$3E,$BC,$7A,$CF,$82,$7A,$DF,$8E
	.dc.b	$F9,$5F,$43,$B7,$CA,$61,$F4,$47,$F6,$67,$D1,$8F,$CD,$9D,$CF,$DD
	.dc.b	$7A,$9F,$D5,$8E,$55,$E7,$7F,$27,$1E,$73,$DC,$7E,$4E,$BC,$8F,$33
	.dc.b	$E5,$C7,$37,$11,$E7,$7C,$91,$DA,$BC,$91,$83,$73,$EA,$3F,$B1,$B8
	.dc.b	$AF,$22,$85,$33,$FE,$E7,$9F,$EB,$2F,$FD,$41,$7B,$AD,$65,$FD,$67
	.dc.b	$0F,$E5,$47,$F0,$99,$FE,$80,$FE,$E8,$FE,$42,$7F,$40,$7F,$90,$BF
	.dc.b	$C1,$B1,$FE,$18,$FE,$F4,$FE,$1F,$1C,$AB,$F8,$DB,$FC,$0E,$3F,$8D
	.dc.b	$9F,$E3,$B3,$67,$F4,$08,$DB,$E0,$7F,$8F,$4F,$F4,$08,$FF,$52,$19
	.dc.b	$9F,$E1,$B7,$FC,$4D,$FE,$1B,$5F,$E5,$D9,$FE,$33,$1E,$17,$F8,$CD
	.dc.b	$73,$AF,$EF,$95,$FD,$DA,$FF,$33,$5F,$EA,$EC,$BF,$33,$7F,$5A,$3E
	.dc.b	$99,$FB,$BC,$3E,$F9,$F9,$F3,$F7,$67,$E9,$4F,$C4,$3F,$E4,$07,$F1
	.dc.b	$17,$F8,$58,$E4,$7E,$14,$EA,$7E,$E1,$FC,$AC,$75,$8F,$CD,$BD,$5F
	.dc.b	$D3,$C7,$EA,$9F,$C4,$1F,$E1,$B8,$FC,$44,$FA,$63,$07,$80,$DE,$1F
	.dc.b	$C7,$5E,$A7,$ED,$8E,$B3,$FA,$07,$F8,$51,$FA,$03,$94,$7E,$51,$FC
	.dc.b	$9E,$5F,$92,$CF,$FC,$45,$5F,$91,$8F,$F1,$14,$7F,$08,$9F,$88,$74
	.dc.b	$C3,$E2,$3F,$06,$BE,$6B,$F4,$DF,$9A,$F9,$63,$F7,$6F,$8B,$0F,$76
	.dc.b	$1E,$B9,$FD,$75,$7C,$77,$75,$D4,$FA,$07,$C4,$30,$7C,$4A,$FE,$55
	.dc.b	$5F,$ED,$29,$FA,$73,$E3,$3E,$3B,$FC,$19,$F8,$E7,$B1,$D9,$FE,$65
	.dc.b	$3E,$B4,$7C,$77,$F6,$75,$FE,$0D,$9F,$D9,$C7,$F6,$C7,$FB,$35,$FC
	.dc.b	$A7,$E6,$E5,$F9,$69,$E9,$5F,$55,$EF,$9A,$FF,$C4,$FF,$76,$F9,$B7
	.dc.b	$89,$86,$E1,$FC,$14,$FE,$19,$3B,$F1,$35,$5F,$A7,$D7,$FC,$29,$3C
	.dc.b	$FC,$4E,$5E,$27,$36,$E1,$CF,$65,$78,$7C,$4F,$07,$89,$E1,$6D,$6F
	.dc.b	$EF,$5D,$CF,$7F,$FC,$7F,$18,$7F,$C8,$3C,$5F,$23,$2F,$E4,$37,$FC
	.dc.b	$8E,$B4,$F8,$72,$F8,$1E,$37,$BC,$F8,$1E,$35,$9D,$71,$3C,$6F,$89
	.dc.b	$E3,$7C,$56,$CE,$BC,$6F,$90,$7C,$9E,$8B,$1C,$BC,$6F,$98,$7C,$DF
	.dc.b	$1A,$C7,$6B,$E7,$B7,$A0,$F8,$B9,$78,$67,$C7,$97,$49,$D6,$BF,$67
	.dc.b	$3F,$05,$78,$CF,$01,$27,$76,$C7,$C3,$18,$C7,$C3,$BA,$FE,$FD,$F8
	.dc.b	$6B,$8E,$5E,$1A,$D4,$79,$B1,$F0,$BD,$9E,$B1,$AC,$BF,$1D,$9F,$C1
	.dc.b	$C7,$C9,$39,$F9,$C7,$C9,$7E,$51,$7C,$3D,$8C,$39,$F2,$65,$FD,$BA
	.dc.b	$79,$38,$D3,$E2,$C3,$C9,$E9,$E4,$F5,$6D,$F1,$CB,$AB,$4E,$99,$73
	.dc.b	$CF,$CD,$9F,$97,$3C,$57,$E9,$32,$EC,$F8,$F2,$EC,$D3,$CD,$E4,$FB
	.dc.b	$4F,$77,$93,$63,$D5,$E4,$FC,$13,$E1,$79,$36,$3D,$DF,$0F,$F1,$4F
	.dc.b	$8D,$F0,$F6,$3A,$93,$87,$E0,$E1,$D5,$F8,$78,$73,$7E,$2E,$17,$6F
	.dc.b	$A5,$8F,$39,$F9,$38,$D9,$FA,$DF,$0F,$F5,$7E,$1F,$EB,$37,$3F,$A5
	.dc.b	$E4,$7D,$9F,$87,$FB,$4D,$8F,$AD,$E0,$FD,$DF,$87,$FB,$CD,$8F,$B5
	.dc.b	$7C,$F3,$FB,$2D,$CB,$F0,$D8,$73,$9F,$BA,$D8,$FC,$55,$F3,$9F,$C0
	.dc.b	$B6,$3D,$37,$F8,$7C,$BD,$2D,$3D,$27,$E2,$F2,$F4,$B4,$F4,$9F,$8F
	.dc.b	$CB,$D2,$D3,$F2,$5F,$0F,$F9,$9F,$87,$FC,$D3,$63,$F2,$DF,$0F,$FC
	.dc.b	$C3,$E7,$3F,$C6,$D1,$83,$F8,$A3,$8D,$3F,$7A,$BE,$63,$81,$C7,$E4
	.dc.b	$ED,$78,$0F,$D1,$63,$E0,$69,$D6,$F9,$9F,$AD,$C3,$C4,$F2,$3B,$DF
	.dc.b	$59,$F1,$BE,$7F,$8A,$91,$E6,$F8,$AB,$5E,$6F,$8A,$F6,$7C,$57,$B5
	.dc.b	$B5,$EC,$F8,$AF,$80,$7C,$1F,$8A,$B1,$F0,$31,$E7,$C7,$FC,$1F,$76
	.dc.b	$78,$AF,$EA,$30,$C7,$AE,$7E,$19,$F9,$39,$F4,$8F,$93,$9F,$79,$F1
	.dc.b	$1F,$3B,$E2,$BE,$97,$C5,$7D,$36,$C7,$D2,$F8,$AF,$AC,$7D,$7F,$8A
	.dc.b	$B1,$F5,$BE,$2B,$ED,$7C,$57,$DB,$6C,$75,$3E,$DE,$1D,$5A,$75,$26
	.dc.b	$5F,$7A,$FA,$BF,$80,$AE,$6D,$9F,$C2,$DF,$38,$FC,$2D,$DD,$FE,$43
	.dc.b	$C9,$F9,$58,$FA,$6F,$BC,$7C,$DC,$3A,$B6,$3A,$9F,$92,$F8,$AF,$CB
	.dc.b	$1F,$97,$F8,$AB,$1F,$96,$F8,$AF,$CD,$1F,$9B,$F8,$AB,$1F,$9A,$F7
	.dc.b	$BF,$3C,$7E,$7F,$E1,$2C,$75,$26,$BF,$9E,$C3,$AB,$FA,$0C,$39,$BF
	.dc.b	$A2,$C2,$E5,$F8,$5C,$79,$CF,$E9,$B1,$B3,$F8,$9F,$8A,$FD,$71,$FE
	.dc.b	$1E,$F4,$3F,$AE,$67,$FA,$E8,$C7,$C9,$C8,$FF,$02,$4B,$E4,$31,$8F
	.dc.b	$F6,$E8,$7C,$77,$3F,$8C,$E8,$DB,$E8,$4A,$E0,$F8,$8E,$B8,$74,$CB
	.dc.b	$FB,$26,$1D,$E7,$27,$8A,$FA,$5F,$85,$B5,$C3,$E3,$3D,$87,$B7,$E3
	.dc.b	$2D,$FA,$BE,$1F,$E0,$7C,$47,$C1,$6C,$7B,$BE,$1F,$E2,$7C,$47,$C5
	.dc.b	$6C,$7C,$2C,$39,$E7,$F0,$71,$B9,$7A,$31,$E7,$3F,$15,$B1,$EC,$C3
	.dc.b	$9C,$FC,$96,$C7,$A7,$0E,$C7,$D0,$BE,$AF,$F9,$3B,$4F,$F2,$8D,$FD
	.dc.b	$1C,$FD,$53,$F5,$9F,$43,$F6,$33,$B3,$F5,$DF,$C0,$7C,$3F,$E0,$7E
	.dc.b	$22,$D7,$DE,$7F,$09,$F0,$D2,$3F,$03,$F0,$D7,$0F,$C0,$7C,$27,$E2
	.dc.b	$7E,$0F,$F1,$4D,$AF,$C3,$E3,$D5,$A7,$53,$F1,$78,$73,$DB,$EF,$E3
	.dc.b	$67,$DD,$7C,$F1,$EF,$1F,$92,$C7,$F3,$1F,$07,$F9,$9F,$86,$B3,$F9
	.dc.b	$67,$F3,$9F,$07,$23,$F3,$3F,$0D,$70,$FF,$22,$FC,$E7,$F8,$B2,$7F
	.dc.b	$CC,$0F,$F8,$B1,$BF,$0B,$FE,$82,$8C,$4F,$E7,$F1,$F0,$57,$E9,$1B
	.dc.b	$1F,$A3,$C3,$B6,$BF,$9A,$C6,$DF,$FA,$1B,$0F,$05,$F8,$BE,$17,$B6
	.dc.b	$5D,$DF,$07,$C2,$CB,$F2,$FC,$2D,$BF,$4F,$37,$97,$E1,$3C,$CD,$AF
	.dc.b	$5F,$C2,$7A,$7E,$13,$D4,$D8,$F7,$E3,$CF,$6F,$F2,$96,$BF,$E4,$69
	.dc.b	$C9,$F1,$30,$E9,$97,$82,$3E,$3E,$5D,$1F,$93,$95,$9F,$9A,$F5,$8F
	.dc.b	$81,$8F,$CB,$CF,$15,$FE,$9E,$C7,$FC,$8A,$FD,$4C,$3D,$77,$D8,$F5
	.dc.b	$9D,$8B,$1D,$9F,$B3,$F0,$9F,$5F,$E1,$3E,$C3,$63,$EE,$FC,$27,$DB
	.dc.b	$F8,$4F,$B8,$D8,$FC,$0F,$C2,$7D,$F3,$F0,$1F,$05,$63,$A9,$F8,$6C
	.dc.b	$3A,$B4,$EA,$7E,$2A,$FA,$B4,$EA,$4C,$7F,$23,$5D,$5F,$C9,$FC,$27
	.dc.b	$E3,$CF,$C8,$7C,$1D,$AF,$CA,$FC,$27,$E4,$CF,$CA,$7C,$25,$8F,$CC
	.dc.b	$FC,$27,$E5,$FE,$0F,$F3,$0D,$8F,$CE,$E1,$CF,$1F,$D0,$E3,$79,$BF
	.dc.b	$03,$97,$39,$FD,$2E,$56,$7F,$0B,$8F,$39,$FD,$4B,$63,$F1,$3F,$07
	.dc.b	$FA,$73,$FD,$D3,$F3,$9F,$E3,$78,$CE,$7F,$EC,$D8,$FF,$1C,$37,$E1
	.dc.b	$37,$F0,$BC,$1E,$3F,$09,$27,$87,$C2,$5C,$FF,$73,$97,$79,$E9,$14
	.dc.b	$FF,$60,$E3,$D9,$B8,$76,$9F,$01,$E3,$C3,$A6,$5F,$EB,$3B,$FF,$16
	.dc.b	$CE,$4F,$E0,$B8,$75,$9D,$61,$67,$FE,$00,$D7,$79,$E5,$87,$7A,$F0
	.dc.b	$D7,$37,$CB,$87,$C6,$C7,$C3,$3C,$BE,$0F,$E4,$FC,$1F,$CA,$6C,$7C
	.dc.b	$9F,$83,$F9,$A7,$CE,$F8,$5B,$1F,$37,$E1,$3E,$89,$F4,$BE,$0A,$C7
	.dc.b	$D1,$BE,$79,$FC,$66,$EB,$F0,$71,$E7,$3F,$2B,$1B,$3F,$17,$1E,$73
	.dc.b	$F3,$B1,$B3,$F2,$7E,$0B,$F0,$3F,$09,$F8,$26,$E7,$F8,$1F,$83,$FC
	.dc.b	$29,$F8,$6F,$84,$B1,$F8,$5F,$84,$FC,$49,$F8,$AF,$84,$B1,$F8,$9B
	.dc.b	$E7,$9F,$DE,$6E,$5F,$67,$1E,$73,$F8,$2C,$6C,$FD,$DC,$79,$CF,$E1
	.dc.b	$B1,$B3,$F8,$1F,$84,$FC,$C9,$F9,$AF,$84,$B9,$FE,$67,$E1,$3F,$3A
	.dc.b	$7E,$7B,$E1,$2C,$7E,$77,$E1,$3F,$E0,$0E,$1F,$EF,$DF,$9C,$FF,$16
	.dc.b	$E7,$F9,$49,$CD,$21,$F1,$0A,$DD,$F8,$EB,$F5,$19,$F8,$DA,$78,$C9
	.dc.b	$7F,$A0,$C7,$C6,$F7,$8E,$D5,$E3,$7F,$6B,$8F,$8A,$7C,$38,$78,$27
	.dc.b	$F3,$78,$7C,$09,$FD,$5E,$77,$0F,$D3,$67,$D9,$FD,$46,$57,$0F,$D1
	.dc.b	$E5,$D7,$3F,$64,$78,$0F,$19,$F0,$5C,$B8,$C7,$C7,$F0,$9C,$74,$FF
	.dc.b	$21,$43,$1F,$99,$39,$F8,$6D,$7C,$2E,$FE,$1A,$E3,$CF,$2F,$99,$1C
	.dc.b	$FE,$1B,$A7,$C2,$5A,$D6,$88,$F0,$65,$D6,$72,$DD,$3C,$51,$E9,$AF
	.dc.b	$F3,$0B,$CB,$0F,$0C,$E8,$F9,$AF,$77,$18,$FD,$FC,$31,$A7,$E1,$DF
	.dc.b	$81,$F0,$D3,$0F,$2F,$C2,$DC,$38,$E4,$F8,$6F,$8C,$7C,$7F,$86,$B7
	.dc.b	$EC,$F8,$5F,$95,$F0,$DF,$2D,$B1,$F0,$32,$E9,$A7,$E5,$F3,$B3,$ED
	.dc.b	$C7,$9E,$7E,$0C,$AD,$FE,$77,$1E,$73,$E4,$6C,$7E,$87,$E1,$BE,$D1
	.dc.b	$F6,$FE,$1A,$E3,$F4,$3E,$1B,$EF,$7C,$37,$DF,$6C,$7D,$4F,$85,$FC
	.dc.b	$17,$C3,$7E,$0D,$B1,$F6,$30,$E7,$9F,$C4,$C6,$E5,$FB,$1C,$79,$CF
	.dc.b	$C8,$6C,$7E,$D7,$0E,$73,$F3,$1B,$1E,$9C,$3B,$1F,$86,$BE,$AF,$E0
	.dc.b	$6F,$9B,$4F,$C5,$E5,$E7,$7F,$1B,$95,$C7,$EC,$BF,$9D,$F8,$69,$3F
	.dc.b	$92,$F8,$6B,$5F,$86,$E6,$FF,$23,$E6,$AF,$D4,$47,$F9,$52,$31,$3F
	.dc.b	$C2,$A3,$73,$AF,$85,$FD,$94,$3E,$0A,$FC,$16,$1B,$5F,$A2,$CB,$C3
	.dc.b	$8F,$F1,$9C,$AC,$F8,$3E,$1B,$9E,$5D,$1B,$1F,$AF,$CB,$B3,$72,$EC
	.dc.b	$D3,$5F,$0B,$E4,$F8,$3F,$2B,$63,$97,$C3,$7A,$3E,$1B,$D2,$D8,$ED
	.dc.b	$F0,$BE,$CF,$83,$F6,$B6,$3A,$9E,$3C,$3A,$B4,$EA,$79,$EF,$AB,$4E
	.dc.b	$A7,$AE,$BA,$B4,$F7,$7C,$2F,$CC,$3E,$6F,$C2,$58,$F8,$5F,$0B,$F4
	.dc.b	$3E,$17,$E8,$B6,$3E,$37,$C2,$7D,$43,$EA,$FC,$25,$8E,$A4,$E7,$FC
	.dc.b	$16,$1D,$5F,$C2,$5F,$36,$CF,$D1,$C3,$94,$7E,$27,$0E,$B3,$E0,$3A
	.dc.b	$72,$FE,$23,$C6,$71,$B4,$F5,$D6,$8F,$C7,$E5,$DE,$F8,$52,$31,$3F
	.dc.b	$94,$C3,$C3,$5A,$D2,$CF,$E6,$3E,$13,$F2,$67,$E5,$3E,$12,$C7,$E2
	.dc.b	$7E,$13,$F2,$FF,$0B,$F9,$86,$C7,$E3,$7E,$13,$FA,$67,$CE,$7F,$93
	.dc.b	$63,$F4,$71,$8E,$0E,$67,$EA,$71,$ED,$3F,$95,$7A,$3F,$97,$C2,$CF
	.dc.b	$EB,$DF,$D3,$D7,$39,$F0,$1F,$D2,$EF,$C1,$5F,$DE,$1F,$F2,$04,$7F
	.dc.b	$5B,$5F,$0D,$E6,$F8,$5F,$3B,$E8,$F8,$59,$5C,$FE,$16,$E1,$FD,$13
	.dc.b	$07,$C2,$7B,$BE,$17,$DE,$DB,$F1,$65,$D1,$A7,$97,$2E,$7C,$3F,$77
	.dc.b	$06,$7F,$23,$3C,$DF,$9F,$1E,$99,$F8,$5B,$87,$1F,$85,$F9,$FF,$0B
	.dc.b	$F4,$1B,$7F,$0F,$E1,$7E,$99,$F5,$3E,$16,$C7,$C7,$F8,$4F,$AE,$7D
	.dc.b	$8F,$85,$B1,$F2,$EF,$9E,$7F,$05,$B9,$7A,$71,$E7,$3F,$17,$1B,$3E
	.dc.b	$DC,$79,$CF,$C9,$C6,$CF,$A7,$1F,$B1,$97,$A5,$A7,$A4,$95,$D5,$B3
	.dc.b	$D5,$F5,$1F,$81,$AE,$AF,$DF,$F8,$5F,$C9,$9F,$94,$F8,$5B,$1F,$87
	.dc.b	$F8,$5F,$CB,$9F,$98,$F8,$5B,$1F,$8B,$F8,$5F,$CD,$9F,$9C,$F8,$5B
	.dc.b	$1F,$8F,$C3,$9E,$7F,$85,$C6,$E7,$F7,$72,$E7,$3F,$89,$CA,$CF,$E0
	.dc.b	$72,$E7,$3F,$8D,$CA,$CF,$E1,$7E,$17,$F5,$67,$EB,$3E,$16,$E9,$F9
	.dc.b	$FF,$85,$FF,$1D,$7C,$E7,$F9,$17,$1F,$D0,$46,$27,$3F,$07,$1F,$F1
	.dc.b	$85,$7F,$8E,$31,$FE,$1C,$A7,$AD,$FE,$7F,$1F,$05,$FE,$B5,$FD,$EE
	.dc.b	$1D,$63,$C4,$74,$3B,$4F,$F8,$AA,$FA,$47,$83,$E1,$B3,$8F,$83,$E1
	.dc.b	$BD,$3E,$3F,$A9,$B5,$E6,$F8,$6F,$6F,$C3,$7B,$9B,$1E,$AF,$84,$F8
	.dc.b	$3F,$09,$F0,$9B,$1E,$EB,$E7,$A7,$8D,$B9,$7C,$0C,$39,$CF,$9D,$B1
	.dc.b	$F1,$2F,$9C,$FA,$DB,$1E,$9B,$F9,$79,$7A,$5A,$7A,$4F,$9F,$97,$A5
	.dc.b	$A7,$A4,$FA,$79,$7A,$5A,$7D,$6F,$84,$FB,$BF,$09,$F7,$9B,$1F,$6B
	.dc.b	$E1,$3F,$02,$7E,$0B,$E1,$6C,$7D,$EF,$85,$FC,$29,$F8,$6F,$84,$B1
	.dc.b	$D4,$FC,$1E,$1D,$5A,$75,$26,$DF,$6E,$FA,$BF,$76,$F9,$BF,$7E,$EE
	.dc.b	$BF,$8C,$F8,$4F,$CA,$9F,$96,$F8,$4B,$8F,$E4,$BE,$17,$F3,$3F,$0B
	.dc.b	$F9,$A6,$C7,$E5,$BE,$13,$F3,$BF,$09,$F9,$E6,$C7,$E6,$B0,$E7,$9F
	.dc.b	$E3,$F1,$B9,$7E,$73,$1E,$73,$F9,$36,$C7,$E8,$30,$E7,$3F,$97,$6C
	.dc.b	$7F,$71,$F9,$CF,$F1,$BC,$FF,$B0,$9B,$ED,$1F,$F1,$F6,$5F,$EA,$77
	.dc.b	$27,$F8,$BA,$7F,$D1,$6D,$AE,$9F,$07,$FE,$B1,$8C,$57,$1F,$83,$C1
	.dc.b	$9F,$81,$B1,$FE,$A7,$9E,$BF,$09,$C7,$E1,$79,$3C,$FE,$1A,$56,$F1
	.dc.b	$B9,$F8,$B2,$E6,$F9,$32,$B9,$FB,$F1,$E7,$87,$7C,$AD,$FA,$B2,$E7
	.dc.b	$7E,$AA,$F2,$47,$27,$D4,$79,$F0,$C5,$7C,$9F,$86,$F8,$1F,$0B,$F0
	.dc.b	$5B,$5E,$DF,$86,$F8,$9F,$0D,$F1,$5B,$1F,$07,$E1,$3E,$41,$F2,$7E
	.dc.b	$1A,$C7,$C5,$BE,$78,$FD,$06,$E9,$F6,$F0,$E7,$3F,$51,$B1,$F7,$F0
	.dc.b	$E7,$3F,$63,$0B,$3F,$83,$F8,$5F,$B0,$7D,$9F,$86,$B9,$7D,$5F,$86
	.dc.b	$FB,$87,$DD,$F8,$6B,$1F,$67,$E1,$3F,$00,$7E,$07,$E1,$AC,$7D,$DC
	.dc.b	$39,$E7,$F8,$8C,$6E,$5F,$93,$CB,$9C,$FE,$33,$2B,$3F,$97,$CB,$9C
	.dc.b	$FE,$43,$2B,$3F,$9B,$F8,$6F,$C8,$7C,$37,$E4,$5B,$A7,$E3,$7E,$1B
	.dc.b	$F2,$87,$FA,$62,$BF,$CA,$5F,$01,$FE,$2C,$EC,$FD,$04,$E6,$99,$27
	.dc.b	$FC,$4B,$F0,$B9,$8C,$5F,$F8,$C2,$BF,$E0,$03,$FE,$34,$8C,$FC,$2F
	.dc.b	$8E,$7F,$3D,$9D,$9F,$D4,$BE,$48,$F3,$1C,$A3,$07,$FB,$E2,$BA,$61
	.dc.b	$FE,$36,$BF,$13,$E4,$AF,$0C,$75,$CB,$F7,$D8,$59,$FD,$D3,$D2,$3F
	.dc.b	$81,$65,$CB,$0F,$85,$DC,$8C,$67,$9C,$6E,$3D,$8F,$7F,$C2,$71,$39
	.dc.b	$7C,$25,$8F,$87,$8F,$5F,$86,$EC,$D8,$F8,$FF,$0B,$E1,$3C,$5F,$05
	.dc.b	$63,$A9,$38,$3E,$16,$1D,$5F,$89,$7C,$DB,$3F,$4B,$0E,$51,$F2,$70
	.dc.b	$EB,$3E,$03,$EB,$FC,$2F,$C1,$F8,$5F,$84,$D8,$FB,$7F,$0B,$F1,$7E
	.dc.b	$17,$E3,$36,$3E,$FF,$C2,$FC,$93,$E5,$7C,$2D,$8E,$A7,$E1,$30,$EA
	.dc.b	$D3,$A9,$F8,$8B,$EA,$D3,$A9,$32,$FC,$15,$75,$7F,$09,$F0,$BF,$67
	.dc.b	$E1,$7E,$D3,$6B,$F2,$7F,$0B,$F7,$7E,$17,$EF,$36,$3F,$2F,$F0,$7F
	.dc.b	$81,$3F,$05,$F0,$B6,$3F,$37,$7C,$F1,$EF,$1F,$9E,$C7,$9B,$FA,$0C
	.dc.b	$6C,$FE,$5D,$E4,$FE,$67,$0E,$B3,$E0,$3A,$9C,$CB,$1F,$9F,$AE,$7D
	.dc.b	$5F,$B6,$BB,$5F,$B4,$EC,$FF,$4B,$C7,$69,$FD,$DE,$1D,$B0,$F0,$5F
	.dc.b	$ED,$3E,$13,$F4,$FF,$39,$FE,$4B,$C7,$FD,$3F,$18,$9C,$FC,$36,$BE
	.dc.b	$16,$C7,$F0,$2C,$BF,$DB,$18,$7F,$93,$63,$C4,$BE,$17,$AF,$C1,$F6
	.dc.b	$6C,$F8,$4E,$47,$03,$F6,$18,$F8,$5C,$E7,$FC,$3E,$FF,$C9,$AB,$0F
	.dc.b	$0E,$3E,$48,$C5,$78,$A3,$C0,$F2,$CD,$7F,$EE,$E8,$C3,$E7,$6E,$5F
	.dc.b	$13,$E1,$7E,$2F,$C2,$59,$DB,$F2,$3E,$16,$47,$2F,$85,$B9,$F8,$8F
	.dc.b	$47,$C2,$FC,$EF,$84,$F9,$ED,$AF,$0E,$3D,$9A,$76,$3C,$B8,$74,$DB
	.dc.b	$E7,$E3,$67,$E7,$DF,$3C,$7C,$1A,$FB,$A3,$17,$F5,$BE,$17,$EF,$FC
	.dc.b	$2F,$E0,$1B,$87,$C2,$F8,$5F,$C1,$9F,$84,$F8,$5B,$1F,$1B,$E0,$FF
	.dc.b	$0E,$7E,$23,$E0,$EC,$7C,$AB,$E7,$A7,$8B,$8D,$18,$CF,$F0,$F9,$74
	.dc.b	$AD,$65,$67,$F1,$79,$74,$9E,$59,$59,$F5,$65,$F6,$33,$F5,$34,$F5
	.dc.b	$12,$BA,$B6,$7A,$BE,$B3,$F0,$35,$D5,$FC,$CF,$C1,$FE,$93,$E1,$7F
	.dc.b	$4A,$D8,$FC,$3F,$C2,$7E,$A0,$FD,$4F,$C2,$D8,$FC,$77,$CE,$7F,$92
	.dc.b	$09,$F0,$7F,$D5,$1F,$F2,$5F,$0B,$0E,$9A,$7C,$76,$E9,$FA,$CC,$BA
	.dc.b	$4F,$CB,$CA,$CB,$2F,$E0,$18,$F7,$F8,$4E,$D9,$F7,$7C,$1F,$09,$2B
	.dc.b	$CB,$F0,$D7,$2F,$C3,$E0,$F8,$6F,$37,$C3,$79,$DB,$7E,$CF,$84,$F5
	.dc.b	$1E,$BF,$86,$B1,$F0,$32,$E9,$AF,$97,$2F,$F0,$36,$4C,$F3,$7E,$8C
	.dc.b	$BA,$D7,$89,$B7,$C7,$A9,$83,$3E,$F7,$E8,$CE,$DE,$E9,$F0,$BF,$3C
	.dc.b	$FA,$1F,$0D,$73,$FA,$BF,$0D,$F4,$CF,$A9,$F0,$D6,$3E,$CF,$C3,$7D
	.dc.b	$73,$EC,$7C,$35,$8F,$BB,$87,$3D,$3E,$4E,$37,$3F,$16,$5C,$E7,$E6
	.dc.b	$E5,$67,$CD,$97,$39,$FA,$39,$59,$F5,$63,$CF,$2F,$BB,$95,$BF,$97
	.dc.b	$97,$39,$FC,$0E,$56,$7E,$7E,$5C,$E7,$F0,$B9,$59,$FA,$7F,$0D,$F9
	.dc.b	$33,$F2,$9F,$0D,$6B,$F3,$3F,$0D,$F9,$7F,$86,$FC,$C3,$63,$F3,$BF
	.dc.b	$0D,$F9,$BF,$86,$FC,$E3,$63,$F4,$38,$73,$C7,$F1,$B8,$DE,$0F,$AD
	.dc.b	$8F,$39,$FC,$8B,$63,$ED,$61,$CE,$7F,$2A,$D8,$FB,$DF,$0B,$FE,$3E
	.dc.b	$F9,$CF,$F2,$B7,$1F,$F8,$D7,$3F,$E7,$D1,$FE,$59,$3F,$C8,$A7,$F9
	.dc.b	$4F,$0F,$E4,$AE,$3E,$17,$97,$C2,$CA,$E3,$F0,$96,$FF,$4B,$8F,$77
	.dc.b	$F7,$4D,$9E,$FF,$0D,$E3,$C3,$C8,$D8,$F1,$E1,$CF,$3F,$CE,$64,$CB
	.dc.b	$15,$FD,$E6,$1F,$0B,$ED,$3D,$DF,$0B,$67,$DB,$F0,$BF,$04,$F8,$5F
	.dc.b	$0D,$63,$E0,$FC,$37,$C5,$3E,$37,$C2,$58,$F8,$B8,$73,$E0,$F8,$78
	.dc.b	$DD,$3C,$B9,$73,$9F,$8F,$95,$9F,$4E,$5C,$E7,$E5,$E5,$67,$DB,$8F
	.dc.b	$3C,$BD,$99,$5B,$FF,$44,$66,$CF,$A5,$7C,$1C,$ED,$6B,$3E,$93,$F1
	.dc.b	$73,$B3,$CB,$E1,$BF,$04,$7E,$0F,$E1,$6D,$7E,$0B,$E1,$BF,$0C,$7E
	.dc.b	$1F,$E1,$AC,$7E,$1B,$E1,$3F,$15,$F0,$9F,$8B,$6C,$7E,$2B,$0E,$78
	.dc.b	$FE,$23,$1B,$C3,$F6,$B2,$E7,$3F,$8C,$CA,$CF,$DE,$C7,$9C,$FE,$41
	.dc.b	$B1,$F8,$2F,$86,$FC,$D1,$F9,$BF,$85,$B9,$FE,$6B,$E1,$7F,$3C,$7E
	.dc.b	$7F,$E1,$AC,$7E,$7B,$E1,$3F,$44,$7F,$A8,$3C,$FF,$F1,$D7,$2A,$39
	.dc.b	$E9,$FA,$0C,$BB,$5F,$E9,$9E,$71,$FA,$4C,$7B,$4F,$84,$EC,$48,$FF
	.dc.b	$06,$E1,$DA,$FC,$1F,$0D,$E1,$F8,$5B,$7F,$E2,$FA,$FF,$22,$5F,$FB
	.dc.b	$1D,$FF,$17,$3C,$BE,$17,$D1,$F0,$DE,$97,$D5,$F0,$92,$F9,$7C,$35
	.dc.b	$BF,$0E,$5D,$E7,$B3,$63,$CB,$87,$39,$F1,$36,$3D,$37,$CF,$87,$FA
	.dc.b	$76,$4C,$71,$87,$F9,$AE,$3F,$C9,$F3,$AC,$FD,$B8,$7C,$27,$D3,$5E
	.dc.b	$FC,$EC,$FB,$63,$E1,$63,$D2,$3E,$16,$37,$D2,$64,$F8,$5F,$B4,$7D
	.dc.b	$BF,$85,$B8,$FC,$BF,$85,$FB,$C7,$DF,$F8,$6B,$1F,$3F,$E1,$BF,$05
	.dc.b	$F0,$7F,$83,$6C,$7D,$3C,$39,$E9,$F6,$B1,$BC,$1F,$0B,$2E,$73,$F7
	.dc.b	$B2,$B3,$F1,$B1,$E7,$3F,$82,$6C,$7C,$AF,$86,$FC,$97,$C3,$7E,$4D
	.dc.b	$B9,$FE,$0F,$E1,$BF,$2C,$7E,$5F,$E1,$AC,$7E,$1F,$E1,$3F,$35,$F0
	.dc.b	$DF,$9B,$6C,$7E,$2E,$F9,$E7,$F9,$26,$E5,$F6,$B1,$E7,$3F,$96,$C6
	.dc.b	$CF,$DE,$C3,$9C,$FE,$69,$B1,$F8,$2F,$85,$FD,$57,$C2,$FE,$AD,$B9
	.dc.b	$7F,$A2,$7E,$71,$F0,$BF,$E1,$DB,$FF,$0B,$46,$27,$5F,$0D,$97,$FD
	.dc.b	$69,$19,$CB,$FC,$EA,$BE,$1B,$50,$E8,$4F,$F8,$9E,$BC,$15,$A9,$CF
	.dc.b	$C3,$79,$67,$F5,$DA,$5C,$3F,$6A,$EF,$0F,$06,$5E,$6C,$F1,$9E,$5D
	.dc.b	$67,$BC,$AE,$5E,$F7,$A5,$FE,$9B,$1B,$3F,$A5,$79,$3F,$16,$FA,$B4
	.dc.b	$F4,$9C,$CB,$97,$C2,$BF,$37,$C3,$49,$F9,$DD,$56,$BE,$3F,$C3,$7A
	.dc.b	$FE,$1B,$D8,$DA,$FA,$9F,$0D,$EF,$F8,$6F,$80,$D8,$FB,$18,$F3,$D7
	.dc.b	$EB,$E5,$67,$EA,$E1,$CF,$2E,$F1,$F7,$F2,$E6,$FE,$07,$2B,$3F,$69
	.dc.b	$F9,$DF,$0D,$27,$F0,$DF,$0D,$6B,$EF,$FC,$37,$D3,$3E,$A7,$C3,$5A
	.dc.b	$FC,$67,$C2,$7D,$7F,$84,$FB,$0D,$8F,$C8,$63,$CF,$6F,$C7,$E5,$67
	.dc.b	$F1,$B8,$F3,$CB,$F1,$B9,$5B,$FC,$AE,$3C,$E7,$F2,$2D,$8F,$CC,$FC
	.dc.b	$27,$E1,$CF,$C4,$7C,$35,$C7,$F4,$1F,$0B,$F8,$B3,$F1,$9F,$0D,$63
	.dc.b	$F4,$9F,$0B,$F9,$5F,$9C,$67,$F9,$18,$C3,$CC,$FD,$0E,$3D,$AB,$F5
	.dc.b	$8F,$38,$FD,$2E,$1D,$A7,$C2,$76,$3F,$DF,$57,$D9,$C4,$FF,$B8,$23
	.dc.b	$0F,$68,$EA,$7E,$E7,$1F,$16,$5F,$B8,$BF,$03,$FB,$AC,$AD,$7E,$E3
	.dc.b	$2F,$4B,$DA,$7F,$D2,$98,$F6,$6F,$C2,$FC,$1F,$86,$F8,$4D,$AF,$3F
	.dc.b	$C3,$7C,$53,$E3,$7C,$35,$8F,$5F,$C3,$7C,$9F,$86,$F9,$4D,$8F,$7D
	.dc.b	$F3,$DF,$DC,$DC,$BC,$78,$F3,$9F,$85,$8D,$9F,$3E,$1C,$E7,$E3,$36
	.dc.b	$3D,$7F,$0D,$F6,$7E,$1B,$ED,$37,$2F,$9F,$F0,$DF,$77,$E1,$BE,$F3
	.dc.b	$63,$E9,$FC,$2F,$E0,$7E,$17,$F0,$4D,$8F,$AF,$7C,$F3,$FA,$CD,$CB
	.dc.b	$E5,$E1,$CE,$7E,$D3,$63,$E7,$DF,$39,$FB,$CD,$8F,$A7,$F0,$BF,$91
	.dc.b	$3F,$25,$F0,$B7,$1F,$C3,$FC,$2F,$E5,$4F,$CB,$7C,$35,$8F,$C5,$FC
	.dc.b	$1F,$E6,$4F,$CD,$7C,$35,$8F,$C7,$E1,$CF,$3F,$C7,$63,$72,$FC,$1E
	.dc.b	$5C,$E7,$F2,$59,$59,$FC,$3E,$5C,$E7,$F2,$D9,$59,$F4,$E5,$D8,$FC
	.dc.b	$EE,$3D,$5F,$D5,$F9,$EE,$86,$7D,$67,$30,$CF,$19,$7F,$4B,$86,$7D
	.dc.b	$EB,$79,$DA,$D7,$C2,$7E,$E7,$E1,$BF,$74,$DA,$FD,$7F,$C2,$FE,$F4
	.dc.b	$FE,$A5,$F3,$88,$C1,$FC,$03,$7F,$F7,$C4,$7F,$8A,$23,$D2,$FF,$90
	.dc.b	$B6,$39,$C7,$FB,$03,$9D,$18,$F8,$67,$F7,$8D,$C3,$F6,$38,$F8,$25
	.dc.b	$1E,$0C,$70,$F8,$BE,$1B,$CF,$F0,$BE,$86,$CF,$9C,$F2,$FC,$37,$B3
	.dc.b	$E1,$BD,$AD,$8D,$61,$E6,$F8,$6F,$82,$7C,$2F,$86,$B3,$AD,$FD,$BF
	.dc.b	$0B,$F1,$8F,$8F,$F0,$B6,$75,$C3,$FE,$62,$DB,$E3,$E9,$E0,$9C,$BD
	.dc.b	$33,$F6,$C7,$CA,$CF,$C0,$F2,$7A,$46,$8F,$97,$8F,$82,$7B,$BF,$63
	.dc.b	$E1,$64,$78,$BE,$16,$D7,$1D,$3E,$AF,$C2,$FD,$EF,$86,$FB,$ED,$BF
	.dc.b	$4F,$C2,$7E,$0B,$E1,$BF,$06,$D8,$F6,$E5,$D3,$6F,$C1,$E7,$67,$F0
	.dc.b	$78,$73,$CF,$C1,$BA,$31,$7F,$8A,$C7,$A5,$69,$B1,$F8,$EC,$7A,$4F
	.dc.b	$E2,$B1,$BA,$7A,$B2,$F9,$FA,$7A,$9A,$7A,$8F,$A7,$A7,$A9,$A7,$D6
	.dc.b	$F8,$5F,$CF,$FC,$2F,$E8,$1B,$1F,$6B,$E1,$7F,$46,$7E,$93,$E1,$AC
	.dc.b	$7D,$EF,$86,$FD,$39,$FA,$8F,$83,$B1,$D4,$FC,$1E,$1D,$5A,$75,$26
	.dc.b	$BF,$1A,$FA,$BF,$22,$F9,$BF,$2A,$EE,$FF,$B5,$8F,$F2,$67,$27,$F9
	.dc.b	$0B,$0F,$E1,$6D,$8C,$FC,$2F,$F9,$36,$31,$B9,$FF,$2B,$57,$E7,$5F
	.dc.b	$F2,$A5,$7E,$1D,$CB,$7E,$0F,$97,$C1,$73,$F8,$2B,$7C,$5E,$DF,$09
	.dc.b	$23,$A6,$57,$0E,$9F,$0B,$E3,$C7,$C8,$DA,$F1,$63,$D5,$A7,$53,$CD
	.dc.b	$F0,$9E,$B3,$D9,$F0,$96,$3D,$5F,$0B,$EF,$F8,$5F,$80,$D8,$F7,$7C
	.dc.b	$27,$C3,$F8,$4F,$88,$D8,$EA,$4D,$3E,$16,$1D,$5F,$89,$7C,$DB,$3F
	.dc.b	$27,$0E,$6F,$CB,$C2,$CF,$CB,$79,$71,$7D,$9E,$2F,$A5,$9F,$AA,$FC
	.dc.b	$13,$9E,$26,$5D,$E3,$0D,$9F,$BF,$A7,$A7,$2F,$B0,$DA,$E5,$F0,$9F
	.dc.b	$81,$3F,$05,$F0,$96,$FF,$01,$F0,$9F,$85,$F8,$5F,$C3,$36,$3F,$09
	.dc.b	$F0,$BF,$89,$F8,$5F,$C5,$36,$3F,$11,$87,$3C,$FF,$11,$8D,$CB,$D5
	.dc.b	$8F,$39,$FC,$63,$63,$DD,$87,$39,$FC,$83,$63,$E1,$7C,$2F,$E6,$7E
	.dc.b	$17,$F3,$4D,$CB,$F3,$1F,$0B,$F9,$D3,$FD,$55,$F3,$9F,$E3,$98,$FD
	.dc.b	$1C,$63,$E1,$24,$7F,$3B,$97,$C3,$63,$3F,$D0,$61,$D5,$B5,$FA,$5C
	.dc.b	$7B,$3E,$03,$94,$60,$FE,$5D,$6C,$3A,$65,$F7,$E3,$BD,$78,$67,$CF
	.dc.b	$F0,$B2,$3C,$9F,$0B,$70,$E7,$F0,$BE,$CF,$86,$F6,$B6,$BD,$5F,$07
	.dc.b	$F0,$3E,$1B,$E0,$B6,$3D,$D8,$F3,$D3,$C5,$95,$9F,$F0,$8E,$8C,$7A
	.dc.b	$67,$E0,$9F,$8D,$9F,$47,$E4,$67,$67,$DE,$F3,$7F,$DB,$D6,$CB,$E9
	.dc.b	$E3,$9E,$B4,$F7,$C3,$A9,$F4,$F2,$F0,$E7,$F1,$F4,$F5,$65,$F3,$1B
	.dc.b	$1F,$2F,$E1,$3E,$FF,$C3,$7E,$01,$B8,$7D,$DF,$85,$FC,$19,$F8,$4F
	.dc.b	$86,$B1,$F8,$1F,$86,$FC,$3F,$C3,$7E,$21,$B1,$F8,$5B,$E7,$9F,$DC
	.dc.b	$6E,$5F,$6F,$1E,$73,$F8,$0C,$6C,$FD,$FC,$39,$CF,$E1,$1B,$1F,$83
	.dc.b	$F8,$6F,$CB,$9F,$98,$F8,$6B,$97,$E5,$7E,$1B,$F3,$7F,$0D,$F9,$C6
	.dc.b	$C7,$E6,$7E,$13,$F3,$E7,$E8,$3E,$1A,$C7,$E7,$70,$E7,$9F,$E5,$31
	.dc.b	$B9,$7E,$4F,$1E,$73,$F9,$86,$C7,$E5,$F1,$E7,$3F,$9C,$C6,$CF,$E6
	.dc.b	$FE,$17,$F5,$E7,$F5,$37,$F8,$E7,$CE,$7F,$92,$73,$FD,$74,$E5,$FF
	.dc.b	$2F,$C7,$F9,$17,$98,$FF,$92,$4C,$17,$E1,$38,$62,$9D,$C6,$1E,$3F
	.dc.b	$05,$6B,$FC,$61,$7F,$9C,$C3,$CA,$FF,$8E,$63,$C9,$87,$ED,$A7,$87
	.dc.b	$C2,$F1,$F8,$6B,$7B,$7A,$65,$23,$96,$37,$0E,$51,$D7,$CC,$EF,$1F
	.dc.b	$E2,$FC,$BF,$B7,$4F,$87,$1E,$CD,$3B,$1E,$5C,$3A,$69,$F2,$B1,$B3
	.dc.b	$F2,$3E,$1B,$DC,$7B,$FE,$16,$C7,$B7,$E1,$BE,$11,$F0,$FE,$0E,$C7
	.dc.b	$C1,$F8,$3F,$8C,$7C,$7F,$85,$B1,$F1,$70,$E7,$B7,$A7,$1B,$9F,$DC
	.dc.b	$CB,$9C,$FB,$72,$B3,$F8,$0C,$B9,$CF,$C1,$CA,$CF,$E1,$3E,$13,$EB
	.dc.b	$7C,$2F,$D7,$6E,$9F,$57,$E1,$7E,$D1,$F6,$FE,$16,$C7,$D9,$F8,$4F
	.dc.b	$BC,$7D,$FF,$85,$B1,$F7,$6F,$9E,$7F,$45,$B9,$7E,$53,$1E,$73,$F5
	.dc.b	$71,$B3,$F9,$8C,$79,$CF,$D9,$C6,$CF,$E7,$3E,$17,$F1,$DF,$0B,$F8
	.dc.b	$F6,$E7,$F8,$DF,$85,$FC,$A7,$CE,$7F,$8B,$BA,$11,$8F,$85,$CF,$C2
	.dc.b	$EB,$E1,$2E,$1F,$E8,$DC,$3F,$C5,$F9,$F7,$8F,$CC,$67,$DD,$A7,$73
	.dc.b	$86,$1F,$A1,$C7,$B4,$73,$C7,$B3,$4E,$D8,$F3,$D7,$F2,$D8,$5C,$7F
	.dc.b	$72,$F3,$AE,$A7,$79,$FF,$45,$65,$FE,$4A,$18,$76,$8C,$B6,$FF,$C3
	.dc.b	$D3,$FE,$4C,$6F,$C2,$7C,$6F,$85,$F8,$EF,$C8,$F8,$6F,$93,$F0,$77
	.dc.b	$2F,$88,$FC,$CF,$86,$91,$F1,$FE,$12,$E9,$E4,$C3,$9C,$FC,$BC,$6C
	.dc.b	$FA,$31,$EB,$3F,$3D,$B1,$EC,$C3,$9E,$5E,$8C,$6D,$FF,$BD,$B3,$7C
	.dc.b	$37,$DE,$F8,$5F,$BE,$D9,$FB,$5F,$0D,$F8,$2F,$84,$FC,$1B,$63,$EF
	.dc.b	$7C,$27,$E1,$BE,$13,$F0,$ED,$8F,$C1,$61,$CF,$5F,$9F,$8D,$D3,$E5
	.dc.b	$E3,$CE,$7E,$9B,$63,$E7,$E1,$CE,$7E,$BB,$63,$E9,$FC,$37,$E5,$8F
	.dc.b	$CB,$FC,$2D,$CB,$F2,$5F,$0D,$F9,$AF,$84,$FC,$DB,$63,$F2,$DF,$09
	.dc.b	$F9,$EF,$86,$FC,$FB,$63,$F3,$58,$73,$CF,$F0,$F8,$DC,$BF,$07,$8F
	.dc.b	$39,$FC,$5B,$63,$F0,$F8,$73,$9F,$C7,$B6,$3D,$38,$76,$3F,$51,$7D
	.dc.b	$5F,$AD,$7C,$DA,$7E,$B7,$2F,$3B,$FA,$FC,$AE,$3F,$9C,$7F,$A6,$7C
	.dc.b	$27,$F9,$26,$3F,$A8,$3F,$E3,$B9,$47,$FB,$AE,$35,$E9,$1D,$46,$7E
	.dc.b	$1B,$9F,$C2,$74,$6D,$F2,$D3,$FD,$05,$6F,$84,$F0,$FC,$1F,$89,$B5
	.dc.b	$B3,$C3,$97,$7C,$BC,$FF,$0D,$E8,$6C,$ED,$F4,$E5,$E0,$7C,$91,$9A
	.dc.b	$F4,$67,$DE,$31,$97,$F8,$D6,$BD,$D8,$F8,$9F,$81,$1F,$18,$F8,$FF
	.dc.b	$09,$6B,$8F,$C3,$7C,$AF,$83,$F9,$6D,$8E,$BF,$0B,$F3,$8F,$9F,$F0
	.dc.b	$96,$3C,$38,$7D,$2F,$84,$E9,$97,$CF,$79,$CF,$9B,$1B,$3F,$4D,$E4
	.dc.b	$FD,$5C,$3A,$CF,$80,$F6,$FC,$27,$DE,$3E,$FF,$C2,$D8,$F8,$3F,$09
	.dc.b	$F8,$2F,$84,$FC,$1B,$63,$E2,$FC,$27,$E1,$8F,$C3,$FC,$2D,$8E,$A7
	.dc.b	$CA,$F8,$5E,$99,$FE,$1D,$E7,$1F,$3B,$0B,$3F,$8B,$79,$3F,$8D,$BE
	.dc.b	$B3,$E0,$3A,$C7,$36,$F4,$7E,$2A,$F9,$BF,$8C,$BB,$3F,$88,$DD,$1D
	.dc.b	$A7,$F2,$38,$76,$C3,$C1,$7F,$82,$F8,$5F,$D1,$9F,$A4,$F8,$5B,$1F
	.dc.b	$89,$F9,$CF,$F1,$5E,$1F,$ED,$B7,$FC,$78,$62,$3F,$2E,$E3,$E1,$BF
	.dc.b	$C7,$0F,$F8,$DB,$0F,$F1,$E2,$8C,$FC,$27,$69,$FD,$76,$3E,$1B,$F2
	.dc.b	$1E,$13,$9F,$C2,$F4,$69,$E5,$BE,$6D,$3D,$17,$E2,$F8,$59,$1E,$BF
	.dc.b	$86,$B8,$B3,$F6,$7C,$37,$A3,$E1,$BD,$2D,$BF,$85,$F0,$DE,$C3,$DB
	.dc.b	$F0,$D6,$3E,$36,$5D,$30,$F8,$D9,$59,$F8,$B8,$73,$CB,$C1,$18,$3E
	.dc.b	$76,$5C,$E3,$27,$C8,$CB,$BC,$F8,$8F,$A9,$F0,$DF,$3B,$E1,$BE,$7B
	.dc.b	$63,$EC,$7C,$2F,$D2,$F8,$6F,$A6,$D8,$FB,$9F,$0D,$F5,$8F,$AF,$F0
	.dc.b	$D6,$3F,$01,$87,$3C,$BC,$7E,$3B,$99,$9E,$78,$3F,$09,$9F,$5B,$DD
	.dc.b	$23,$13,$F8,$9C,$FB,$56,$B3,$B3,$F8,$DC,$BA,$67,$F8,$CC,$AD,$FE
	.dc.b	$2B,$1E,$73,$F9,$06,$C7,$E3,$B1,$E7,$3F,$94,$C6,$CF,$E4,$BE,$1B
	.dc.b	$F2,$A7,$E5,$BE,$1A,$D7,$E7,$FE,$1B,$F3,$3F,$05,$F9,$A6,$C7,$E8
	.dc.b	$FE,$13,$F3,$A7,$FA,$57,$E6,$1F,$0D,$89,$FD,$46,$3D,$32,$F8,$79
	.dc.b	$5E,$0F,$D5,$65,$D2,$7E,$3B,$63,$F5,$D9,$73,$9F,$A3,$97,$6C,$7C
	.dc.b	$47,$F3,$D7,$FC,$6A,$DF,$85,$F2,$7C,$37,$95,$F3,$7C,$24,$9E,$1F
	.dc.b	$0B,$70,$EB,$F0,$BE,$BF,$86,$F6,$36,$BA,$12,$7C,$78,$F5,$8F,$16
	.dc.b	$1C,$F4,$47,$F5,$AC,$D9,$66,$31,$8F,$5C,$7C,$31,$EB,$CB,$E6,$7C
	.dc.b	$34,$8F,$77,$C3,$5A,$F8,$7F,$05,$F4,$4F,$A5,$F0,$76,$BE,$27,$C1
	.dc.b	$FD,$5F,$85,$FA,$CD,$8F,$91,$8F,$3D,$BC,$F8,$FF,$B4,$76,$67,$9B
	.dc.b	$F5,$67,$D6,$BE,$9E,$77,$0E,$3D,$CC,$59,$F7,$BF,$B2,$DA,$DC,$B3
	.dc.b	$FC,$57,$C2,$77,$BF,$13,$D7,$3F,$09,$F7,$F2,$C4,$F9,$F1,$E9,$3F
	.dc.b	$3D,$B1,$EB,$F8,$5F,$CB,$7C,$2F,$E5,$DB,$5F,$89,$F8,$4F,$CD,$1F
	.dc.b	$9B,$F8,$4B,$1F,$8D,$F8,$4F,$CF,$7C,$2F,$E7,$DB,$1F,$91,$BE,$78
	.dc.b	$FE,$59,$BB,$FB,$B1,$E7,$3F,$9A,$C6,$CF,$C2,$C3,$9C,$FE,$79,$B1
	.dc.b	$F1,$BE,$17,$EA,$BF,$3B,$EF,$FF,$C6,$7C,$29,$CE,$7F,$E2,$89,$FF
	.dc.b	$1B,$9F,$E3,$6A,$E1,$F0,$9A,$C5,$C6,$AD,$5F,$1A,$D4,$E7,$E1,$2D
	.dc.b	$7F,$8E,$6B,$FC,$58,$A3,$3F,$09,$AC,$25,$7E,$D7,$2B,$1F,$CA,$AF
	.dc.b	$BB,$E2,$58,$F6,$F8,$5E,$EF,$83,$1E,$B8,$78,$1B,$7C,$FE,$1B,$CB
	.dc.b	$ED,$F9,$9B,$5E,$6F,$86,$F4,$9E,$AF,$86,$B1,$EA,$F8,$4F,$69,$EE
	.dc.b	$F8,$4B,$1E,$EB,$E7,$8F,$78,$F8,$38,$F2,$8F,$66,$3D,$67,$C0,$75
	.dc.b	$24,$7C,$1B,$EA,$FC,$3A,$EB,$C3,$F7,$70,$B7,$F6,$F0,$ED,$3F,$81
	.dc.b	$CA,$CF,$DF,$CB,$9C,$FE,$17,$2B,$3F,$83,$F8,$6F,$B2,$7D,$AF,$86
	.dc.b	$B1,$F6,$BE,$1B,$EE,$FC,$37,$DE,$6C,$7D,$EF,$86,$FC,$09,$F8,$2F
	.dc.b	$84,$B1,$F8,$2C,$3F,$0B,$F0,$9D,$38,$7E,$F3,$CE,$7F,$0F,$8D,$9F
	.dc.b	$C1,$3C,$9F,$C2,$61,$D6,$7C,$07,$E3,$BE,$1B,$F2,$3F,$0D,$F9,$26
	.dc.b	$C7,$E4,$BE,$17,$F2,$A7,$E5,$BE,$1A,$C7,$FA,$07,$CB,$FF,$1B,$63
	.dc.b	$FA,$06,$F6,$97,$C3,$7E,$81,$CF,$63,$3E,$F7,$F9,$9C,$EF,$27,$EC
	.dc.b	$F4,$EF,$3F,$9D,$D2,$CF,$F2,$4C,$7B,$93,$2E,$DA,$77,$F8,$6F,$07
	.dc.b	$C3,$58,$F4,$3E,$3F,$85,$91,$EB,$F8,$5B,$9F,$F8,$53,$17,$C2,$FA
	.dc.b	$4F,$57,$C2,$DB,$F8,$59,$76,$69,$D8,$F8,$D8,$F4,$D3,$D1,$95,$9F
	.dc.b	$F5,$9E,$3D,$9F,$99,$8F,$66,$9C,$F2,$F2,$67,$EB,$6C,$76,$23,$E9
	.dc.b	$BE,$CF,$D5,$AF,$43,$ED,$CE,$E3,$F2,$FE,$17,$EB,$7C,$37,$D7,$6E
	.dc.b	$3F,$7F,$E0,$FE,$D7,$C2,$7D,$B6,$C7,$E0,$FE,$13,$EF,$7C,$37,$DF
	.dc.b	$6C,$7E,$1F,$0E,$79,$FC,$DC,$6E,$5F,$6F,$1E,$73,$F4,$5B,$1F,$7F
	.dc.b	$0E,$73,$F5,$5B,$1F,$83,$F8,$5F,$C7,$1F,$8F,$F8,$5B,$97,$E5,$FE
	.dc.b	$0F,$F2,$5F,$07,$F9,$36,$C7,$E6,$FE,$0F,$F2,$DF,$0B,$F9,$76,$C7
	.dc.b	$E7,$F0,$FC,$D7,$C2,$74,$D3,$F0,$CF,$39,$FD,$26,$36,$7F,$14,$F3
	.dc.b	$8F,$D4,$61,$67,$F1,$CF,$F4,$4F,$C0,$3F,$C6,$51,$FE,$64,$7F,$C6
	.dc.b	$2D,$F8,$6D,$7C,$36,$FE,$1B,$87,$C3,$5C,$7F,$DE,$98,$23,$A7,$C3
	.dc.b	$49,$CF,$C3,$5C,$BB,$63,$D2,$78,$E5,$67,$C5,$97,$69,$EA,$D8,$F3
	.dc.b	$61,$CF,$8B,$FC,$2F,$9B,$2C,$60,$BF,$EC,$B3,$EC,$AD,$E5,$F0,$72
	.dc.b	$F0,$47,$2C,$BB,$37,$0F,$6E,$7D,$31,$E7,$9F,$C9,$CF,$1E,$33,$4F
	.dc.b	$E2,$58,$7C,$4F,$83,$F9,$DD,$FF,$3D,$FA,$1F,$0B,$F4,$7E,$16,$D7
	.dc.b	$CA,$7E,$A7,$C2,$7D,$5F,$84,$B3,$F3,$9E,$4F,$53,$E8,$E1,$8D,$7E
	.dc.b	$D6,$1D,$67,$EA,$B6,$3E,$F5,$F5,$9F,$B2,$D8,$FC,$17,$C3,$7E,$14
	.dc.b	$FC,$37,$C3,$5A,$FB,$FF,$09,$F8,$9F,$86,$FC,$53,$63,$F0,$7F,$0B
	.dc.b	$F8,$D3,$F1,$DF,$0B,$63,$F0,$F8,$7E,$47,$E1,$7A,$65,$F9,$17,$9C
	.dc.b	$FE,$33,$1B,$3F,$95,$79,$3F,$97,$C3,$AC,$F8,$0F,$C9,$FC,$37,$E7
	.dc.b	$4F,$CF,$7C,$25,$8F,$C2,$74,$7F,$97,$BF,$01,$FC,$34,$60,$FC,$47
	.dc.b	$32,$33,$8F,$FB,$21,$FF,$0F,$C6,$63,$FC,$5B,$1F,$E2,$09,$E5,$1F
	.dc.b	$E3,$48,$C3,$C7,$E0,$F7,$F0,$9E,$38,$E9,$F0,$D7,$1F,$23,$D2,$26
	.dc.b	$1D,$31,$F2,$7C,$36,$27,$C9,$87,$4A,$D6,$5F,$E3,$B6,$C7,$9B,$09
	.dc.b	$1F,$18,$DB,$CF,$1F,$89,$7E,$CC,$31,$5F,$13,$0F,$5D,$7C,$9C,$3E
	.dc.b	$0E,$5E,$1C,$3E,$37,$C2,$FC,$4F,$86,$F8,$AD,$C7,$CB,$F0,$DF,$20
	.dc.b	$F9,$3F,$0D,$63,$D3,$F0,$DF,$30,$F9,$BF,$0D,$63,$DB,$87,$3C,$7B
	.dc.b	$C6,$0E,$A4,$7E,$C5,$F5,$7E,$D5,$F3,$7E,$E5,$DD,$3F,$09,$F0,$7F
	.dc.b	$70,$FB,$BF,$0D,$71,$F9,$BF,$0D,$F8,$03,$F0,$3F,$09,$63,$E8,$FC
	.dc.b	$27,$E1,$3E,$1B,$F0,$AD,$8F,$AB,$87,$3C,$FF,$19,$8D,$CB,$F2,$99
	.dc.b	$73,$9F,$C8,$65,$67,$F3,$18,$F3,$9F,$CA,$36,$3F,$39,$F0,$DF,$94
	.dc.b	$F8,$4F,$CA,$B7,$3F,$C2,$FC,$2F,$E6,$0F,$CC,$FC,$2D,$8F,$C4,$FC
	.dc.b	$2F,$FA,$23,$F0,$2F,$F1,$6F,$4A,$31,$3C,$F4,$FD,$26,$5D,$A3,$F2
	.dc.b	$2F,$38,$FD,$46,$3D,$A7,$C2,$76,$3F,$39,$87,$6B,$EB,$1C,$8C,$FC
	.dc.b	$37,$7D,$3F,$DA,$B5,$FE,$3E,$9E,$78,$76,$7F,$43,$9F,$86,$BF,$7D
	.dc.b	$87,$4A,$FD,$D3,$63,$9F,$C3,$7B,$CF,$81,$F0,$B7,$0F,$2F,$C3,$7C
	.dc.b	$3F,$86,$F8,$8D,$8F,$4F,$C3,$7C,$73,$E4,$7C,$35,$8F,$6E,$1F,$2F
	.dc.b	$E0,$FA,$68,$CF,$D7,$1D,$2B,$E1,$E5,$6B,$51,$FC,$53,$64,$F5,$9C
	.dc.b	$9F,$EC,$28,$FF,$1E,$38,$F1,$10,$CB,$86,$3F,$2B,$E1,$3E,$F7,$C2
	.dc.b	$7D,$F6,$E5,$F5,$3E,$1B,$F0,$5F,$0B,$F8,$36,$C7,$D8,$F8,$6F,$C3
	.dc.b	$7C,$1F,$E1,$DB,$1F,$73,$0E,$7A,$F8,$F4,$65,$8C,$FE,$DE,$5D,$2B
	.dc.b	$4D,$8F,$BF,$8F,$49,$E4,$D8,$FC,$1F,$C3,$7E,$5C,$FC,$C7,$C3,$5C
	.dc.b	$FF,$1B,$F0,$9F,$9B,$3F,$39,$F0,$D6,$3F,$23,$F0,$BF,$9F,$3F,$41
	.dc.b	$F0,$D6,$3F,$2B,$87,$FB,$AB,$9F,$FC,$4F,$97,$82,$7F,$9A,$46,$1F
	.dc.b	$AD,$8F,$5D,$3E,$06,$56,$FF,$33,$97,$59,$F8,$99,$59,$61,$FD,$4B
	.dc.b	$1E,$95,$FE,$FE,$6F,$27,$DC,$CF,$C9,$F0,$BD,$6F,$C2,$7E,$B3,$3C
	.dc.b	$61,$F8,$5C,$FB,$4F,$E2,$F3,$B3,$D7,$E1,$BD,$DF,$0D,$EF,$6D,$7B
	.dc.b	$BE,$1B,$E1,$7C,$37,$C3,$6C,$7C,$2F,$86,$F8,$C7,$C7,$F8,$2B,$1F
	.dc.b	$1B,$0E,$78,$FE,$57,$83,$FC,$69,$3E,$F9,$E8,$FC,$CC,$6D,$6D,$E6
	.dc.b	$F1,$C3,$B4,$F8,$4F,$A7,$F0,$BF,$5F,$E0,$FE,$C3,$63,$EB,$FC,$2F
	.dc.b	$DB,$3E,$E7,$C2,$58,$FB,$7F,$0B,$F7,$FE,$0B,$F0,$0D,$8F,$BF,$87
	.dc.b	$E0,$FE,$17,$A6,$7E,$B7,$93,$ED,$C7,$AD,$78,$0E,$A7,$E2,$30,$EA
	.dc.b	$D3,$F1,$7F,$0B,$F8,$F3,$F2,$1F,$07,$63,$F1,$FF,$07,$F9,$33,$F2
	.dc.b	$9F,$0B,$63,$F2,$7F,$0B,$F9,$73,$F3,$1F,$0B,$63,$F2,$FF,$0B,$F9
	.dc.b	$B3,$A6,$9F,$4D,$E4,$FD,$5C,$3A,$CF,$80,$EA,$48,$FB,$55,$D5,$E5
	.dc.b	$1F,$E1,$7F,$C0,$FF,$C5,$3E,$07,$E7,$27,$F3,$58,$58,$C7,$80,$AD
	.dc.b	$DF,$E8,$71,$F1,$D6,$BA,$1A,$62,$BF,$4D,$A7,$86,$B5,$A5,$9F,$F1
	.dc.b	$DE,$3F,$E3,$83,$FC,$32,$E3,$E1,$7C,$4E,$BE,$16,$EC,$D7,$C9,$F0
	.dc.b	$B8,$7C,$DF,$0D,$71,$FF,$17,$57,$F2,$F5,$19,$F8,$6D,$7C,$2D,$CB
	.dc.b	$E2,$3C,$72,$91,$F1,$F3,$B8,$F3,$CF,$B6,$3D,$DB,$5F,$37,$1E,$AD
	.dc.b	$3A,$9F,$47,$E1,$BC,$C7,$9F,$E1,$AC,$7D,$5F,$86,$F5,$7C,$37,$AD
	.dc.b	$B1,$F6,$7E,$17,$DD,$F0,$BE,$F6,$C7,$52,$69,$F0,$70,$EA,$FC,$3B
	.dc.b	$E6,$D9,$FC,$26,$1C,$DF,$C3,$61,$67,$E5,$3F,$37,$E1,$7E,$77,$C2
	.dc.b	$DA,$FC,$63,$F4,$7E,$16,$47,$E3,$FE,$1A,$E1,$F4,$FE,$1B,$EB,$7C
	.dc.b	$2F,$D7,$6D,$72,$24,$FD,$9C,$7B,$C7,$E6,$70,$E7,$D3,$F9,$6E,$76
	.dc.b	$58,$AF,$C6,$E3,$D3,$1F,$04,$7E,$87,$2E,$71,$F8,$5F,$85,$B8,$F3
	.dc.b	$79,$47,$E3,$F1,$65,$F8,$DC,$B3,$D3,$FA,$5C,$FB,$63,$E3,$6E,$7F
	.dc.b	$9E,$F8,$4F,$CA,$1F,$95,$F8,$4B,$1F,$E8,$9F,$C1,$3F,$A6,$B4,$FC
	.dc.b	$E3,$FE,$2B,$D3,$FC,$10,$E9,$BF,$07,$AF,$85,$DB,$6B,$A7,$C2,$F2
	.dc.b	$3B,$D7,$27,$AC,$43,$C5,$F0,$D6,$7A,$C7,$96,$FA,$B4,$F4,$7C,$2F
	.dc.b	$9B,$C5,$F3,$B6,$3D,$9F,$0B,$EA,$F8,$6F,$5B,$63,$E0,$7C,$17,$B8
	.dc.b	$F7,$FC,$25,$8F,$89,$87,$C2,$F8,$3E,$99,$FC,$27,$9C,$FC,$9C,$6C
	.dc.b	$FC,$67,$93,$F2,$30,$EB,$3E,$03,$E8,$7C,$37,$CE,$F8,$4F,$9E,$D8
	.dc.b	$FA,$9F,$0D,$F4,$8F,$A7,$F0,$76,$3E,$C7,$C2,$FD,$63,$EB,$FC,$2D
	.dc.b	$8F,$B9,$F0,$9F,$68,$E9,$9F,$DA,$79,$3F,$73,$0E,$B3,$E0,$3A,$92
	.dc.b	$3F,$03,$5D,$5E,$51,$F8,$9F,$84,$EC,$D3,$94,$75,$7F,$1D,$8F,$83
	.dc.b	$AF,$F9,$EF,$43,$3E,$91,$8C,$6D,$FF,$BC,$29,$F0,$9F,$94,$3F,$98
	.dc.b	$7E,$0B,$FE,$20,$8C,$CF,$FB,$8E,$3F,$C1,$39,$7F,$37,$6F,$71,$7C
	.dc.b	$17,$2F,$85,$E7,$F0,$B3,$0E,$5F,$09,$71,$FF,$34,$60,$F8,$4F,$09
	.dc.b	$E3,$AF,$03,$CD,$F1,$61,$E6,$C7,$17,$CB,$1E,$9B,$7F,$5E,$C5,$96
	.dc.b	$31,$5F,$A7,$E1,$7D,$FF,$0B,$F0,$1B,$3E,$FF,$83,$F8,$7F,$05,$F1
	.dc.b	$1B,$1F,$0F,$E1,$7E,$39,$F2,$3E,$16,$C7,$C7,$C3,$E5,$FC,$27,$4C
	.dc.b	$BD,$EF,$39,$F9,$98,$D9,$F8,$6F,$27,$E2,$E1,$D6,$7C,$07,$D3,$F8
	.dc.b	$5F,$AF,$F0,$BF,$61,$B1,$F5,$FE,$17,$ED,$9F,$73,$E1,$2C,$7D,$BF
	.dc.b	$84,$FB,$E7,$E0,$3E,$12,$C7,$DF,$F8,$4F,$C1,$FC,$2F,$E1,$1B,$1D
	.dc.b	$49,$A7,$DA,$C3,$AB,$F7,$30,$E6,$FD,$EC,$2F,$2F,$D5,$C7,$9C,$FE
	.dc.b	$0B,$1B,$3F,$67,$2E,$6F,$DB,$C6,$E5,$F6,$BA,$BF,$21,$9F,$48,$C7
	.dc.b	$33,$4C,$67,$FE,$0D,$F7,$FF,$C3,$ED,$3F,$C1,$CB,$E0,$EC,$63,$E1
	.dc.b	$26,$1F,$88,$D3,$FC,$53,$7F,$E1,$26,$CF,$2F,$84,$E7,$F0,$76,$F8
	.dc.b	$1E,$7B,$47,$83,$E1,$64,$EB,$1F,$23,$76,$5A,$BC,$75,$FE,$CF,$8F
	.dc.b	$F1,$6B,$ED,$F8,$5C,$E7,$E6,$CB,$C7,$8F,$3B,$FF,$13,$B7,$2F,$83
	.dc.b	$12,$35,$F0,$97,$1E,$DF,$09,$F1,$BE,$13,$E3,$B6,$BA,$7C,$2F,$CA
	.dc.b	$F8,$5F,$96,$D8,$F0,$7C,$27,$CE,$F8,$5F,$9E,$D8,$F2,$7C,$17,$D2
	.dc.b	$F8,$2F,$A6,$D8,$F4,$7C,$17,$D6,$F8,$5F,$AE,$D8,$F6,$7C,$2F,$DA
	.dc.b	$3E,$DF,$C2,$58,$F8,$1F,$09,$F7,$8F,$F2,$67,$E0,$DF,$DB,$F9,$BE
	.dc.b	$DC,$63,$E0,$E5,$7D,$1A,$7C,$2E,$39,$3F,$DF,$D0,$F8,$4E,$67,$78
	.dc.b	$CB,$DE,$BC,$27,$33,$11,$CA,$FA,$63,$FE,$74,$BE,$F5,$E1,$9F,$3F
	.dc.b	$C2,$FA,$3E,$16,$CF,$95,$F5,$FC,$1C,$8F,$47,$C2,$DC,$BC,$FF,$07
	.dc.b	$F0,$0F,$83,$F0,$76,$BD,$DF,$0B,$F1,$3E,$17,$E2,$B6,$3E,$17,$C2
	.dc.b	$7C,$83,$E4,$FC,$2D,$8F,$8D,$F0,$9F,$33,$E0,$FE,$6B,$63,$E5,$7C
	.dc.b	$1F,$D0,$F8,$3F,$A2,$D8,$F9,$DF,$0B,$F5,$0F,$AB,$F0,$B6,$3E,$97
	.dc.b	$C2,$FD,$83,$FD,$09,$F8,$3F,$FB,$D3,$9F,$FA,$AF,$13,$A3,$FD,$3B
	.dc.b	$3F,$D5,$0F,$EA,$77,$FA,$B7,$1F,$07,$CB,$E0,$E4,$E7,$E0,$EE,$1C
	.dc.b	$E3,$F8,$75,$FF,$55,$79,$7C,$2F,$8B,$E1,$3C,$6F,$93,$E0,$A4,$F7
	.dc.b	$F8,$4B,$87,$33,$FD,$3D,$83,$E0,$BD,$7F,$05,$EC,$6D,$F9,$FE,$13
	.dc.b	$DF,$F0,$9F,$01,$B1,$EB,$F8,$3F,$86,$7C,$4F,$84,$B1,$EF,$F8,$2F
	.dc.b	$8F,$F0,$5F,$21,$B1,$F0,$FE,$0B,$E5,$FC,$0F,$CC,$6C,$7C,$7F,$83
	.dc.b	$F9,$E7,$D0,$F8,$4B,$1F,$C5,$7F,$08,$FE,$57,$F0,$47,$F9,$54,$67
	.dc.b	$8B,$FD,$75,$18,$7E,$14,$EF,$E0,$F2,$6B,$E0,$AD,$F5,$F8,$2E,$3F
	.dc.b	$05,$C9,$B1,$E1,$F8,$3E,$BF,$07,$D9,$B1,$E5,$F8,$3F,$0F,$C1,$78
	.dc.b	$9B,$1E,$9F,$83,$F2,$9E,$6F,$82,$B1,$ED,$F8,$3F,$49,$EA,$F8,$2B
	.dc.b	$1F,$07,$E0,$FD,$A7,$BB,$E0,$AC,$7C,$5F,$83,$F8,$27,$C2,$F8,$2B
	.dc.b	$1F,$27,$E0,$FE,$29,$F1,$BE,$0E,$C7,$CD,$F8,$3F,$93,$F0,$7F,$29
	.dc.b	$B1,$FD,$17,$C8,$FE,$34,$FF,$63,$6F,$4F,$FB,$AD,$FF,$47,$79,$DF
	.dc.b	$F0,$01,$CF,$51,$FE,$1F,$3F,$51,$B3,$BF,$82,$FF,$79,$46,$2F,$B7
	.dc.b	$C0,$E0,$CF,$C0,$D8,$F2,$7C,$0F,$03,$8F,$C0,$D8,$F4,$7C,$0F,$43
	.dc.b	$AF,$C0,$D8,$F6,$7C,$07,$83,$E0,$7C,$2D,$8F,$81,$F0,$3E,$43,$CB
	.dc.b	$F0,$36,$3E,$27,$C0,$7A,$0F,$4F,$C0,$D8,$E2,$4E,$2F,$81,$F0,$1C
	.dc.b	$5F,$85,$F0,$3F,$00,$F8,$3F,$03,$6B,$E8,$7C,$0F,$C4,$3E,$2F,$C0
	.dc.b	$D8,$FE,$F1,$F8,$4B,$E2,$91,$86,$E3,$FA,$29,$FF,$34,$9B,$9E,$3F
	.dc.b	$05,$AE,$AD,$F5,$AB,$F8,$DD,$F1,$F8,$2D,$61,$FE,$23,$9C,$DF,$07
	.dc.b	$FC,$C3,$1E,$4F,$82,$DD,$B0,$E9,$F0,$38,$7B,$4D,$B7,$C1,$53,$1D
	.dc.b	$D7,$1F,$7F,$C1,$6C,$E1,$F0,$16,$3E,$1F,$C0,$F3,$3A,$7C,$05,$8F
	.dc.b	$8F,$F0,$3D,$CF,$07,$C0,$D8,$F9,$7F,$03,$E3,$3C,$9F,$01,$63,$E7
	.dc.b	$FC,$0F,$9C,$F4,$7C,$05,$8F,$A7,$F0,$3E,$B3,$D9,$F0,$36,$3E,$BF
	.dc.b	$C0,$FB,$CF,$85,$F8,$57,$F1,$CE,$34,$63,$E1,$E5,$7F,$8D,$6B,$FC
	.dc.b	$63,$CC,$7F,$8F,$74,$FD,$87,$FC,$8D,$3F,$E5,$76,$F4,$FD,$17,$FC
	.dc.b	$66,$FF,$24,$73,$5C,$FE,$03,$C7,$EF,$79,$1B,$7E,$1F,$7F,$CE,$7A
	.dc.b	$3D,$EB,$1E,$5F,$7F,$D6,$7B,$3D,$FB,$1E,$9F,$7F,$DF,$EF,$FC,$06
	.dc.b	$C7,$B7,$DF,$F8,$67,$E2,$3F,$0A,$7E,$1D,$E8,$E1,$F7,$4F,$C8,$F8
	.dc.b	$19,$7F,$17,$E0,$6F,$1F,$B7,$E0,$7E,$6F,$C0,$FC,$E6,$D7,$F9,$87
	.dc.b	$3E,$BB,$F7,$D0,$FE,$63,$07,$DC,$7F,$98,$2B,$77,$F4,$FE,$0B,$EC
	.dc.b	$9F,$6B,$E0,$AC,$7D,$7F,$81,$FB,$BF,$05,$F7,$9B,$1F,$E7,$1C,$7A
	.dc.b	$FC,$27,$87,$83,$C1,$9F,$D8,$CF,$E0,$C7,$E9,$FF,$0C,$FE,$56,$4F
	.dc.b	$8A,$C6,$39,$F8,$BD,$7C,$35,$8F,$F3,$E5,$7F,$92,$5F,$06,$3F,$2D
	.dc.b	$FF,$6A,$E3,$F3,$B6,$F0,$FC,$3F,$C3,$7C,$1F,$0F,$2B,$9F,$C3,$DD
	.dc.b	$FF,$B6,$E4,$F8,$7F,$37,$C3,$F9,$DB,$7E,$0F,$07,$EE,$78,$1F,$CF
	.dc.b	$70,$FF,$72,$ED,$FE,$96,$AF,$A9,$39,$F0,$D9,$3E,$1F,$E1,$1F,$0F
	.dc.b	$E1,$ED,$FB,$BE,$1B,$E3,$7C,$3F,$C7,$6C,$7C,$0F,$1F,$C3,$B7,$CE
	.dc.b	$6C,$7C,$6F,$87,$F9,$C7,$CF,$F8,$7B,$1F,$2B,$E1,$FE,$97,$C3,$FD
	.dc.b	$36,$C7,$CC,$F2,$FD,$B9,$7D,$A6,$C7,$D2,$F8,$7F,$B4,$7D,$BF,$87
	.dc.b	$B1,$F5,$BE,$1F,$EF,$7C,$3F,$DF,$6C,$7D,$8E,$4F,$93,$97,$E1,$9B
	.dc.b	$1F,$7B,$E1,$FF,$0C,$7F,$2D,$E9,$FE,$7D,$D1,$F5,$A3,$F9,$9F,$C3
	.dc.b	$BF,$0D,$63,$8C,$7F,$0D,$7F,$E4,$9A,$D4,$E7,$C1,$FC,$AF,$93,$F6
	.dc.b	$B4,$BA,$7D,$87,$F9,$FF,$D4,$BE,$4B,$9F,$A3,$FD,$4A,$31,$86,$A5
	.dc.b	$E1,$58,$F2,$FD,$E7,$FB,$3A,$9F,$7B,$FC,$7B,$26,$DF,$D7,$EB,$53
	.dc.b	$98,$FC,$5F,$AD,$FC,$27,$F0,$F7,$C9,$FE,$B2,$30,$F9,$AB,$86,$3D
	.dc.b	$BD,$AC,$7C,$7E,$5F,$95,$53,$3F,$F7,$37,$3F,$F6,$07,$C7,$E3,$23
	.dc.b	$BC,$7E,$BB,$D9,$F1,$7C,$67,$EF,$1B,$1C,$AF,$A6,$DF,$E5,$6B,$7C
	.dc.b	$76,$39,$3F,$B9,$C7,$F6,$C9,$D6,$3F,$77,$E3,$BE,$F3,$FC,$23,$DD
	.dc.b	$F8,$1F,$1B,$73,$FE,$3D,$C7,$FE,$31,$7D,$35,$C1,$F5,$FC,$75,$C7
	.dc.b	$C5,$7F,$89,$3F,$15,$F1,$D6,$BF,$84,$FC,$07,$CD,$F8,$DF,$E2,$ED
	.dc.b	$8F,$8B,$5B,$BF,$67,$C7,$5D,$FD,$78,$7E,$54,$FC,$B7,$C7,$5C,$3F
	.dc.b	$8A,$FB,$7F,$67,$E3,$7F,$91,$B6,$3E,$8D,$6F,$1F,$91,$F1,$D7,$1F
	.dc.b	$8F,$87,$E8,$4F,$D1,$7C,$75,$C3,$F9,$0F,$C3,$7E,$17,$E3,$7F,$97
	.dc.b	$B6,$3E,$ED,$6F,$1F,$A9,$F1,$D7,$1F,$A7,$87,$EB,$4F,$D7,$7C,$75
	.dc.b	$C3,$F9,$6F,$C4,$7E,$47,$E3,$7F,$9D,$B6,$3F,$13,$5B,$C7,$F0,$1F
	.dc.b	$1D,$71,$FB,$F8,$7E,$E4,$FD,$D7,$C6,$DC,$3F,$9C,$FC,$67,$F6,$5F
	.dc.b	$C4,$3F,$41,$8F,$F5,$68,$C7,$87,$B8,$FC,$66,$1F,$E2,$FF,$47,$FD
	.dc.b	$53,$E0,$23,$15,$31,$FD,$6E,$7F,$E4,$2B,$FF,$07,$9D,$BE,$05,$F1
	.dc.b	$9E,$1F,$86,$FE,$AE,$DB,$E2,$7E,$92,$BF,$69,$E3,$FF,$0F,$C7,$C5
	.dc.b	$F1,$5D,$F5,$FF,$5A,$D7,$F5,$37,$FC,$8C,$DA,$FE,$2D,$F0,$BE,$9C
	.dc.b	$BF,$B2,$FC,$3F,$C1,$F7,$2E,$7F,$DE,$F8,$91,$B8,$F1,$E5,$D2,$FE
	.dc.b	$17,$C0,$E6,$3C,$35,$F1,$4F,$8D,$EC,$DB,$FE,$CF,$F1,$7F,$3F,$D2
	.dc.b	$FE,$EA,$D8,$EB,$AF,$F2,$BB,$E1,$87,$B3,$D8,$BC,$9F,$2F,$1F,$A4
	.dc.b	$7D,$3F,$42,$E3,$FD,$D3,$E3,$3E,$E7,$93,$FE,$02,$6C,$76,$E3,$5F
	.dc.b	$0C,$BE,$4F,$9D,$72,$FB,$18,$FD,$F3,$F0,$1E,$3D,$C7,$FC,$21,$F7
	.dc.b	$DF,$EB,$8F,$8B,$72,$BC,$2F,$F1,$04,$E6,$31,$5E,$1C,$39,$67,$F6
	.dc.b	$3C,$6B,$9F,$E1,$F1,$FC,$71,$F8,$FF,$02,$E1,$E0,$F8,$DF,$CC,$75
	.dc.b	$76,$6C,$7B,$6B,$78,$FE,$13,$BA,$E3,$F9,$3C,$3F,$34,$7E,$6F,$A2
	.dc.b	$E1,$EC,$F8,$DF,$F2,$47,$E2,$3F,$ED,$4C,$5C,$1F,$E0,$FE,$14,$FF
	.dc.b	$B5,$39,$59,$72,$BF,$C9,$F3,$5D,$3F,$D4,$7F,$88,$FF,$B6,$89,$C1
	.dc.b	$FE,$99,$7B,$67,$FD,$7B,$5F,$F1,$5E,$77,$1F,$EA,$F3,$74,$F0,$FB
	.dc.b	$9C,$6E,$1F,$8A,$C3,$FC,$40,$77,$8E,$C7,$7F,$76,$CF,$DA,$7D,$FF
	.dc.b	$15,$F6,$1B,$1C,$E7,$78,$B4,$F5,$FC,$7E,$1F,$6E,$1E,$B3,$D9,$E4
	.dc.b	$5C,$3F,$0F,$5F,$33,$E3,$7F,$0A,$D8,$F4,$56,$F0,$ED,$1F,$D9,$33
	.dc.b	$FF,$1C,$3A,$3D,$D7,$F2,$4F,$95,$F1,$36,$3F,$2B,$7F,$6B,$E3,$3F
	.dc.b	$26,$D8,$E2,$4C,$3B,$9F,$EC,$DC,$1F,$1B,$CE,$33,$E2,$25,$87,$D9
	.dc.b	$3E,$D7,$C6,$DC,$3F,$4B,$AF,$E2,$BE,$37,$F4,$6D,$8F,$AD,$5B,$C7
	.dc.b	$AE,$1F,$E2,$A9,$C9,$FC,$43,$E0,$FF,$C4,$5A,$7F,$87,$CF,$F1,$8F
	.dc.b	$4B,$2C,$57,$ED,$3E,$3B,$F3,$3F,$11,$FB,$26,$C7,$3F,$88,$FF,$37
	.dc.b	$37,$0F,$CF,$FC,$67,$2D,$3C,$78,$FE,$54,$FC,$B7,$C7,$5B,$FF,$50
	.dc.b	$4A,$BF,$D5,$CD,$8C,$FC,$6E,$27,$F3,$DF,$1D,$27,$FD,$51,$D7,$FE
	.dc.b	$4C,$E9,$FD,$2C,$FF,$AD,$FE,$0F,$FC,$49,$8A,$91,$8F,$8E,$FF,$28
	.dc.b	$56,$FE,$4F,$F8,$7F,$5F,$1F,$8B,$B9,$FF,$AD,$70,$FF,$15,$E5,$F7
	.dc.b	$B2,$E7,$87,$1A,$FE,$37,$85,$AF,$A7,$C7,$FB,$53,$F6,$DA,$DA,$FE
	.dc.b	$43,$D9,$E7,$F8,$8F,$E3,$ED,$8C,$9E,$1F,$8A,$CB,$4F,$DB,$70,$FF
	.dc.b	$01,$0F,$F4,$2F,$91,$F8,$8F,$E8,$63,$F9,$5F,$7F,$C5,$F8,$9B,$3E
	.dc.b	$D6,$E9,$ED,$E0,$F8,$F3,$FC,$DA,$BE,$64,$76,$9F,$CD,$E1,$F1,$39
	.dc.b	$31,$E0,$7F,$B9,$B8,$F5,$18,$9F,$A7,$39,$C7,$FA,$0F,$77,$D7,$F8
	.dc.b	$7F,$E7,$ED,$8C,$9F,$47,$E2,$32,$D3,$DD,$F1,$9D,$9F,$F3,$0E,$1F
	.dc.b	$E5,$87,$E1,$71,$67,$46,$5F,$EF,$AB,$FE,$9B,$E1,$FE,$1F,$E1,$EE
	.dc.b	$1F,$7D,$B7,$F0,$FE,$22,$ED,$F0,$7A,$BF,$17,$1D,$72,$68,$E5,$C3
	.dc.b	$72,$FA,$9F,$17,$89,$FE,$BB,$F2,$1F,$97,$F8,$8B,$5F,$8F,$6E,$9F
	.dc.b	$4F,$E2,$6E,$BE,$1E,$64,$76,$E8,$98,$7E,$06,$6D,$FF,$14,$3F,$A8
	.dc.b	$F8,$5F,$A5,$F8,$7F,$ED,$AD,$AF,$EE,$6F,$EA,$7E,$23,$31,$FA,$17
	.dc.b	$D7,$F1,$92,$3F,$1D,$D5,$70,$ED,$E8,$3C,$6F,$A1,$CC,$FB,$3F,$F2
	.dc.b	$2C,$E7,$0F,$F1,$1C,$E7,$E2,$7F,$BF,$DF,$0B,$7C,$7E,$2F,$71,$87
	.dc.b	$F3,$DD,$DF,$38,$F9,$FF,$1D,$63,$FC,$1B,$E4,$F3,$F8,$9F,$F0,$5B
	.dc.b	$63,$27,$4F,$8A,$CB,$4F,$F1,$BF,$E2,$2A,$FD,$5E,$E8,$C7,$9B,$F8
	.dc.b	$D7,$5E,$35,$AD,$D7,$F8,$03,$CB,$F6,$7C,$4F,$F8,$71,$B8,$FF,$88
	.dc.b	$9F,$81,$F1,$59,$8F,$7B,$FC,$DA,$24,$7C,$2F,$8A,$B7,$F4,$3C,$EF
	.dc.b	$91,$F1,$3F,$E2,$86,$D7,$F8,$B9,$F9,$9F,$15,$98,$F9,$6F,$CF,$F8
	.dc.b	$A9,$1F,$E3,$4F,$42,$E1,$F7,$1F,$A9,$F1,$1F,$E3,$46,$CE,$CF,$AB
	.dc.b	$F1,$59,$69,$FE,$60,$A7,$E0,$1F,$2B,$B3,$F0,$9F,$15,$FB,$59,$CF
	.dc.b	$2F,$E0,$63,$15,$BE,$AF,$F2,$0D,$7E,$6E,$31,$F1,$3A,$BF,$C4,$7C
	.dc.b	$55,$CB,$F0,$EF,$5F,$C0,$5E,$77,$E3,$7E,$2F,$0D,$AF,$C9,$F9,$8F
	.dc.b	$88,$FC,$CF,$C0,$FF,$95,$63,$13,$F9,$88,$FF,$2C,$0F,$88,$DF,$C5
	.dc.b	$70,$6C,$FC,$1D,$BF,$93,$2D,$FF,$AA,$70,$7F,$90,$23,$5D,$FF,$A1
	.dc.b	$F8,$AF,$82,$FC,$2F,$80,$B3,$CD,$B7,$FD,$4E,$DF,$17,$8D,$7F,$8B
	.dc.b	$46,$2F,$81,$F4,$FF,$39,$F1,$DF,$8F,$3F,$EC,$EA,$32,$FD,$8C,$7E
	.dc.b	$9F,$BB,$E3,$BF,$63,$67,$C5,$C1,$72,$F0,$BC,$2B,$F0,$5F,$B2,$FF
	.dc.b	$89,$1E,$26,$3E,$07,$57,$FC,$5F,$2F,$E7,$4F,$DF,$C7,$F0,$D1,$8C
	.dc.b	$7F,$A9,$4F,$E3,$27,$F8,$7F,$DB,$FF,$9B,$F4,$D6,$3F,$3F,$AD,$39
	.dc.b	$F8,$1F,$C8,$69,$FE,$1A,$C7,$F1,$91,$8F,$85,$FC,$8B,$F9,$2F,$76
	.dc.b	$E3,$F6,$DB,$5F,$E0,$3C,$3F,$05,$3F,$7E,$BC,$7C,$BC,$63,$F0,$5D
	.dc.b	$16,$FF,$34,$DD,$BD,$7A,$BD,$4F,$ED,$DF,$88,$B8,$3F,$11,$87,$EA
	.dc.b	$B5,$FD,$2D,$6A,$73,$7F,$D3,$EB,$F4,$DE,$DF,$5C,$F2,$FE,$B3,$3F
	.dc.b	$83,$F0,$96,$7E,$AB,$BA,$FC,$26,$3D,$5F,$C3,$65,$E0,$AF,$CD,$67
	.dc.b	$E5,$9C,$C7,$E7,$F1,$E5,$1F,$C0,$1F,$DD,$9F,$5B,$AE,$CF,$F4,$2E
	.dc.b	$04,$70,$D3,$E1,$4F,$6D,$7D,$B9,$78,$71,$65,$E1,$7B,$57,$3F,$8D
	.dc.b	$EB,$1F,$55,$FE,$17,$EE,$FE,$1B,$E3,$AD,$78,$5E,$97,$CF,$2F,$0E
	.dc.b	$1D,$B2,$F0,$DF,$EE,$F6,$F8,$93,$FC,$73,$1F,$27,$0F,$DA,$6D,$7D
	.dc.b	$4F,$6E,$D7,$EB,$DA,$FF,$24,$BE,$78,$F8,$32,$F1,$6B,$E4,$CF,$EE
	.dc.b	$D7,$0A,$FF,$1A,$DF,$F2,$38,$E7,$CC,$F8,$6A,$71,$9F,$DF,$7B,$1D
	.dc.b	$AB,$BB,$74,$FA,$5F,$1D,$E6,$EA,$FD,$3B,$FC,$CA,$7F,$5B,$D5,$6B
	.dc.b	$EC,$37,$0F,$8C,$7F,$9B,$F9,$5F,$1D,$DA,$BE,$D4,$E6,$30,$FE,$A7
	.dc.b	$E3,$39,$72,$7F,$8B,$6F,$FD,$6F,$3F,$CF,$FD,$BF,$DE,$FC,$5F,$F4
	.dc.b	$46,$E3,$FC,$B5,$F9,$3F,$15,$F2,$9B,$1E,$0A,$FE,$A3,$1F,$C1,$BE
	.dc.b	$2F,$94,$FE,$45,$FE,$9F,$5F,$C2,$BC,$2B,$3F,$95,$6D,$FB,$8F,$EB
	.dc.b	$7F,$17,$D9,$C7,$2A,$31,$A7,$AF,$BF,$8B,$70,$FE,$89,$3F,$C6,$39
	.dc.b	$7F,$AF,$B6,$32,$7C,$8E,$4F,$E3,$3F,$13,$F8,$69,$FD,$A4,$62,$7D
	.dc.b	$D8,$F8,$5F,$D6,$7C,$37,$46,$9F,$D4,$FE,$57,$F9,$5F,$8B,$FD,$B9
	.dc.b	$B1,$93,$9F,$C5,$7E,$52,$3F,$2A,$D8,$FD,$C7,$C2,$F1,$69,$FD,$7E
	.dc.b	$3F,$BC,$FC,$57,$F7,$A6,$C6,$4F,$A9,$F8,$1B,$C5,$7C,$47,$82,$73
	.dc.b	$18,$E3,$F9,$B7,$FE,$04,$F9,$6E,$95,$FD,$B5,$EB,$3F,$E5,$FF,$D9
	.dc.b	$9A,$E3,$93,$FA,$3C,$62,$FC,$55,$F6,$F3,$F1,$67,$E1,$8C,$67,$E1
	.dc.b	$F1,$7F,$82,$69,$E1,$C3,$8E,$5F,$CE,$34,$C6,$1F,$D0,$FE,$26,$CF
	.dc.b	$E6,$DF,$3C,$74,$CB,$B6,$3F,$C6,$73,$F0,$CF,$E2,$B0,$FE,$B7,$F0
	.dc.b	$9C,$B2,$FE,$B5,$DC,$8F,$F0,$E7,$99,$FD,$83,$E0,$AD,$7F,$2B,$7F
	.dc.b	$C5,$98,$F4,$CB,$B6,$FE,$BF,$23,$FA,$3F,$07,$F1,$AA,$D4,$7F,$67
	.dc.b	$F6,$DE,$5F,$F9,$E9,$B9,$7F,$9E,$1F,$E4,$BC,$5F,$EA,$37,$3C,$1F
	.dc.b	$E8,$FF,$6B,$F8,$AC,$FE,$EF,$DF,$FF,$19,$79,$DF,$DD,$9F,$2E,$BE
	.dc.b	$3B,$FF,$12,$69,$E2,$7F,$C2,$9A,$F9,$67,$5D,$AC,$DE,$2F,$97,$C4
	.dc.b	$F3,$37,$6F,$F2,$86,$DE,$4D,$BF,$D7,$1F,$8A,$FF,$9D,$23,$13,$FE
	.dc.b	$05,$8F,$F3,$BE,$BE,$1D,$4D,$F0,$1B,$78,$FA,$7F,$69,$AB,$BF,$E3
	.dc.b	$1F,$1E,$6E,$5E,$2F,$37,$FD,$47,$9F,$F8,$B3,$4F,$93,$3E,$1F,$67
	.dc.b	$E1,$E5,$FA,$03,$F0,$BC,$DF,$E0,$0D,$7E,$27,$17,$C1,$EC,$FA,$87
	.dc.b	$D5,$F8,$EB,$8F,$C0,$F3,$BB,$FB,$1F,$E6,$16,$C7,$F9,$CF,$0F,$71
	.dc.b	$ED,$DF,$D9,$C7,$FC,$EB,$2F,$F4,$F3,$FE,$BA,$F8,$AF,$F2,$C5,$FF
	.dc.b	$AE,$1F,$DB,$B6,$3F,$CE,$F9,$7B,$B4,$B3,$FE,$BE,$E0,$FF,$19,$D7
	.dc.b	$DE,$E8,$FC,$03,$F3,$9B,$CB,$F3,$AB,$E1,$4F,$F2,$CC,$3F,$D0,$7F
	.dc.b	$1D,$D2,$3F,$D0,$4E,$27,$FC,$B9,$EA,$59,$FF,$25,$3F,$31,$FD,$81
	.dc.b	$FC,$2A,$3A,$F7,$7F,$7F,$AF,$F4,$97,$CB,$FB,$AB,$AD,$7F,$A6,$63
	.dc.b	$1D,$5F,$E5,$06,$C7,$F8,$D3,$C5,$FE,$61,$5F,$EE,$F9,$CB,$FE,$7F
	.dc.b	$8C,$D7,$F9,$63,$8B,$E8,$EB,$F0,$38,$BE,$8E,$7F,$E1,$BC,$78,$E3
	.dc.b	$FE,$73,$F0,$BC,$AF,$F9,$89,$F3,$F3,$FF,$AA,$73,$FD,$7E,$9F,$E4
	.dc.b	$6B,$FA,$B7,$FC,$8F,$C0,$F8,$58,$FF,$02,$C3,$F1,$B3,$F4,$DF,$EC
	.dc.b	$D5,$C3,$5F,$77,$C7,$76,$C7,$E1,$3F,$18,$F8,$F3,$63,$E3,$3D,$B9
	.dc.b	$3F,$31,$B7,$B6,$7E,$7C,$79,$6B,$E7,$FC,$77,$F8,$AA,$FF,$DF,$99
	.dc.b	$7C,$98,$C4,$7B,$7B,$BB,$E1,$FE,$CB,$DB,$F6,$0F,$E6,$F2,$FC,$5F
	.dc.b	$0F,$E6,$E3,$BE,$3F,$EA,$FC,$3A,$E9,$F6,$72,$F0,$67,$E2,$BF,$83
	.dc.b	$F1,$D7,$0F,$B4,$DB,$FB,$19,$76,$3D,$93,$F8,$CD,$FE,$C6,$3E,$B3
	.dc.b	$AE,$1F,$89,$70,$FA,$78,$BA,$3F,$89,$7E,$45,$7D,$33,$F2,$C7,$5D
	.dc.b	$FE,$E5,$7C,$7F,$97,$E7,$3F,$E1,$3D,$51,$9C,$3F,$61,$E5,$7E,$72
	.dc.b	$3F,$88,$BD,$AB,$F6,$1C,$5F,$82,$69,$CE,$7F,$C1,$BD,$3F,$E8,$BD
	.dc.b	$3F,$29,$18,$E7,$FB,$1F,$1F,$FC,$73,$DD,$FD,$4B,$73,$FA,$99,$7D
	.dc.b	$2C,$BF,$E0,$0E,$FF,$9E,$DF,$F2,$D7,$F8,$1F,$8F,$E7,$87,$E4,$9C
	.dc.b	$3F,$C2,$7A,$7F,$6B,$1F,$CA,$9F,$57,$37,$DF,$C7,$E3,$1F,$7E,$7F
	.dc.b	$19,$F1,$FC,$A3,$E3,$38,$BF,$E2,$9D,$5F,$C0,$23,$F9,$AB,$D3,$97
	.dc.b	$EE,$F0,$FE,$9E,$F8,$DF,$F8,$CF,$8B,$FA,$FC,$FE,$DF,$0F,$8F,$E5
	.dc.b	$7F,$A1,$3C,$5F,$EA,$B7,$FC,$87,$C6,$3F,$D5,$3C,$8F,$E8,$71,$DF
	.dc.b	$8F,$E9,$6F,$FC,$FB,$C2,$7C,$6F,$F8,$F9,$B8,$7F,$90,$D7,$C6,$E6
	.dc.b	$7F,$55,$F1,$FF,$EE,$6E,$2F,$E3,$6D,$BF,$A7,$C3,$F1,$8F,$A6,$FD
	.dc.b	$5E,$C4,$FE,$CF,$E3,$F9,$CF,$F0,$47,$0F,$F2,$3C,$3F,$93,$C7,$F2
	.dc.b	$87,$D5,$7F,$98,$E2,$F9,$07,$F1,$B9,$FC,$04,$7E,$CD,$AB,$0F,$6A
	.dc.b	$EF,$FF,$7E,$FE,$CE,$E9,$ED,$92,$B4,$FF,$B9,$27,$4F,$BB,$1F,$0F
	.dc.b	$57,$F8,$7A,$3D,$D9,$78,$32,$3E,$2E,$47,$9D,$FD,$DB,$CA,$FA,$F9
	.dc.b	$FF,$86,$E3,$18,$75,$C7,$F8,$EF,$27,$D8,$EA,$FE,$08,$F6,$9F,$D0
	.dc.b	$63,$FD,$01,$FF,$07,$73,$F4,$E0,$EB,$92,$B4,$BC,$A7,$5F,$E2,$A3
	.dc.b	$DE,$7F,$05,$E0,$F6,$BF,$E2,$6C,$7F,$5D,$D1,$FD,$33,$3F,$B1,$1E
	.dc.b	$2F,$1F,$FC,$23,$BF,$86,$3E,$55,$F8,$67,$31,$ED,$8E,$DB,$FC,$97
	.dc.b	$F1,$18,$75,$7F,$88,$70,$FC,$9E,$FF,$F0,$EF,$AF,$D6,$7E,$B6,$FF
	.dc.b	$DE,$DF,$4E,$FE,$C9,$FC,$3E,$DF,$E6,$CC,$FD,$3C,$5F,$2A,$FA,$C7
	.dc.b	$CA,$F5,$7F,$C5,$3B,$75,$9F,$91,$18,$9F,$44,$7F,$1A,$BE,$8F,$A2
	.dc.b	$FA,$47,$E4,$E7,$EA,$F9,$DF,$E3,$4D,$3A,$68,$8C,$4F,$4D,$BF,$87
	.dc.b	$5F,$D0,$C7,$B6,$FF,$82,$7C,$E7,$EB,$70,$79,$9F,$E4,$0D,$3A,$B7
	.dc.b	$1E,$9A,$7F,$0E,$C3,$96,$3D,$72,$FF,$29,$73,$FE,$5A,$FF,$17,$E6
	.dc.b	$7F,$92,$B5,$F3,$B6,$39,$F2,$7E,$4C,$F9,$67,$E0,$B6,$FE,$01,$F8
	.dc.b	$3F,$F8,$53,$8F,$F1,$BB,$7C,$B8,$C7,$37,$A3,$0E,$77,$FB,$BD,$BC
	.dc.b	$A4,$7F,$0A,$7E,$EE,$3F,$9F,$57,$F8,$25,$FF,$36,$65,$D7,$2F,$F3
	.dc.b	$0B,$F8,$0B,$FC,$66,$DE,$FD,$FE,$AC,$7F,$86,$79,$7F,$82,$E3,$FD
	.dc.b	$E3,$1F,$E0,$1A,$BD,$FF,$F1,$2C,$E7,$4F,$D4,$47,$F8,$A3,$87,$F2
	.dc.b	$D9,$FE,$08,$ED,$AF,$ED,$6B,$E1,$F0,$FE,$5F,$1E,$93,$F0,$E3,$18
	.dc.b	$7A,$EB,$DE,$7E,$00,$FD,$81,$F8,$D9,$E9,$7F,$85,$C5,$E6,$72,$F3
	.dc.b	$7E,$73,$74,$F9,$B5,$E5,$DB,$E6,$63,$D4,$99,$FE,$05,$FB,$3E,$6F
	.dc.b	$92,$3E,$C3,$6F,$9D,$7C,$C3,$9E,$DF,$09,$F9,$91,$F5,$BC,$CF,$6E
	.dc.b	$5F,$81,$6C,$72,$C7,$DB,$97,$C8,$7F,$A4,$65,$E3,$8F,$77,$81,$F1
	.dc.b	$B3,$FC,$53,$6F,$95,$FF,$4E,$3E,$31,$CB,$87,$EE,$F9,$3F,$3B,$2F
	.dc.b	$C9,$36,$B8,$F9,$FF,$C2,$72,$F1,$1F,$B1,$CF,$C5,$AF,$C8,$E0,$79
	.dc.b	$5F,$63,$5F,$CE,$37,$1E,$7D,$9F,$D4,$38,$7E,$26,$FF,$99,$3F,$4B
	.dc.b	$E2,$FE,$03,$6F,$D2,$36,$F8,$9F,$AC,$9F,$53,$FA,$0E,$0F,$C3,$E3
	.dc.b	$FA,$B6,$C7,$09,$F8,$3B,$7A,$67,$FA,$97,$8F,$FE,$90,$F9,$1F,$F4
	.dc.b	$8B,$6F,$85,$7C,$0D,$FF,$0B,$C1,$F0,$31,$FC,$9A,$C3,$FD,$33,$F2
	.dc.b	$3C,$67,$FD,$24,$FE,$BD,$FB,$75,$F8,$4E,$AF,$D7,$C7,$C1,$C3,$E6
	.dc.b	$57,$F2,$DF,$89,$FE,$A1,$E2,$FF,$A7,$E3,$1E,$37,$3B,$FB,$97,$E5
	.dc.b	$7C,$7B,$7F,$6D,$F9,$1F,$F5,$5F,$C8,$FF,$AB,$1B,$BF,$B3,$5F,$17
	.dc.b	$6F,$EA,$1F,$6D,$7F,$AC,$BE,$47,$FD,$08,$DA,$FA,$19,$7E,$9F,$2F
	.dc.b	$D9,$F3,$7D,$A7,$FC,$15,$9F,$F7,$DC,$3F,$D1,$9F,$21,$FE,$8D,$6E
	.dc.b	$1C,$B2,$FD,$35,$F2,$E1,$F8,$6F,$29,$FF,$12,$5F,$F8,$0B,$25,$FF
	.dc.b	$68,$F1,$5F,$07,$D3,$05,$69,$FF,$0B,$4E,$B0,$F0,$47,$D2,$3C,$1B
	.dc.b	$7C,$17,$ED,$FC,$87,$FA,$8F,$E3,$FF,$D4,$8D,$AE,$33,$F4,$32,$F9
	.dc.b	$2F,$E4,$8F,$DA,$D7,$9E,$3F,$BE,$EF,$FE,$50,$AF,$F5,$EC,$E6,$BF
	.dc.b	$C5,$53,$97,$F2,$B5,$F1,$1F,$E2,$51,$F3,$79,$FF,$21,$18,$BE,$4F
	.dc.b	$F6,$7F,$85,$B5,$CF,$4F,$85,$39,$AF,$4F,$81,$FB,$BC,$7D,$35,$FA
	.dc.b	$39,$FF,$0C,$FC,$7F,$2F,$8F,$E6,$D8,$E3,$E0,$7E,$E7,$8F,$EB,$EB
	.dc.b	$E4,$D1,$1F,$EC,$2C,$3F,$7C,$6F,$83,$FD,$25,$A7,$87,$5F,$A2,$7C
	.dc.b	$9D,$3F,$8F,$3F,$E2,$43,$FD,$46,$5C,$3A,$D6,$63,$FD,$6C,$67,$04
	.dc.b	$66,$38,$9F,$E0,$53,$4E,$67,$9C,$73,$6B,$A3,$13,$D1,$D7,$C7,$BE
	.dc.b	$03,$93,$63,$31,$82,$38,$21,$EB,$8E,$0D,$8D,$B4,$D9,$F8,$AF,$8E
	.dc.b	$FF,$23,$E3,$F7,$30,$46,$38,$35,$5F,$E7,$5A,$D7,$C7,$66,$3F,$1F
	.dc.b	$8F,$9E,$79,$36,$BF,$2E,$D9,$FF,$3F,$7E,$0F,$FE,$74,$7A,$4E,$B2
	.dc.b	$B1,$D1,$AB,$25,$F3,$8C,$7C,$37,$8D,$D6,$56,$8D,$35,$CE,$F8,$DD
	.dc.b	$61,$CE,$BA,$E3,$E0,$6E,$1F,$C0,$B2,$FF,$36,$E7,$E9,$8C,$1E,$72
	.dc.b	$9E,$7B,$E8,$75,$CF,$C9,$F1,$F2,$7B,$3E,$7F,$3F,$C7,$97,$86,$F8
	.dc.b	$BF,$0B,$0F,$06,$DF,$3A,$34,$4C,$3C,$39,$B2,$E9,$1E,$BC,$6B,$AF
	.dc.b	$8F,$EE,$F0,$CB,$BB,$EA,$CB,$BE,$9F,$56,$35,$97,$7C,$FF,$11,$97
	.dc.b	$CF,$C6,$57,$C9,$7E,$A7,$ED,$0F,$83,$EF,$7E,$9A,$E6,$77,$F6,$F3
	.dc.b	$97,$79,$D1,$DA,$7E,$A7,$07,$DD,$23,$F5,$FE,$43,$8B,$F4,$E7,$8D
	.dc.b	$7D,$8B,$CC,$FD,$5D,$3D,$53,$FC,$B3,$4F,$C4,$61,$2B,$A1,$F8,$3F
	.dc.b	$03,$FD,$5B,$E4,$7F,$AF,$B8,$BF,$0B,$9F,$6A,$F0,$63,$DD,$EA,$60
	.dc.b	$87,$E6,$0E,$87,$FB,$86,$BA,$5F,$6E,$DF,$E7,$ED,$9F,$E7,$BD,$F5
	.dc.b	$F5,$1F,$2B,$0F,$12,$E2,$FB,$2F,$83,$2F,$3E,$5F,$34,$F8,$11,$F9
	.dc.b	$FB,$E9,$26,$9D,$FC,$97,$B3,$AE,$5F,$C3,$E5,$E0,$CB,$BE,$3E,$48
	.dc.b	$F1,$E3,$E0,$8F,$EA,$BB,$FE,$D7,$E4,$3B,$92,$3E,$16,$DF,$D7,$BE
	.dc.b	$43,$B7,$17,$CA,$9E,$9B,$FC,$D7,$E0,$3E,$B7,$DD,$87,$F0,$1A,$F5
	.dc.b	$92,$3A,$9F,$D9,$BE,$43,$AB,$C8,$FA,$BA,$7A,$B3,$F4,$47,$80,$EA
	.dc.b	$7F,$05,$9F,$8E,$47,$A8,$E7,$7C,$07,$6D,$BE,$1E,$FF,$C6,$71,$EC
	.dc.b	$FC,$AD,$3D,$58,$7E,$14,$FE,$17,$EA,$BE,$0B,$F0,$0D,$8E,$ED,$7F
	.dc.b	$DE,$7F,$2C,$E2,$F0,$5F,$4D,$BD,$79,$78,$38,$7E,$64,$7B,$8F,$E1
	.dc.b	$18,$7F,$30,$C5,$E9,$F8,$09,$1F,$2B,$DC,$66,$E2,$EF,$3F,$8F,$DB
	.dc.b	$DD,$D7,$FB,$0C,$BB,$47,$F2,$39,$ED,$EA,$7B,$08,$F5,$C1,$C9,$D2
	.dc.b	$3E,$91,$36,$F5,$3F,$64,$CE,$1E,$A7,$E4,$FA,$9F,$20,$8F,$5F,$91
	.dc.b	$E9,$3E,$4E,$0F,$E5,$DA,$79,$F4,$FA,$97,$F9,$83,$E2,$7A,$9F,$48
	.dc.b	$8F,$9F,$A3,$9C,$F0,$8F,$D9,$57,$3B,$F9,$F5,$FD,$5F,$89,$EA,$7D
	.dc.b	$C2,$5F,$C1,$F9,$0E,$77,$FA,$1D,$BD,$FD,$1F,$8A,$8F,$EA,$D8,$F3
	.dc.b	$F5,$3F,$0A,$4C,$3E,$57,$C8,$72,$7E,$36,$1C,$BA,$3F,$09,$B7,$D3
	.dc.b	$9F,$C7,$FA,$BF,$8E,$23,$F7,$FF,$17,$FF,$06,$F8,$9F,$C8,$6B,$DD
	.dc.b	$C7,$F8,$CE,$2F,$75,$FD,$38,$FA,$5B,$FF,$92,$38,$9E,$B7,$E6,$89
	.dc.b	$5D,$7E,$47,$CD,$5F,$C2,$F0,$E6,$FF,$0A,$E1,$F9,$1E,$BF,$E8,$89
	.dc.b	$7E,$5F,$91,$E3,$C1,$FA,$AE,$1F,$74,$FE,$4F,$93,$E3,$BF,$E0,$CF
	.dc.b	$8D,$78,$BF,$CA,$DB,$1F,$BA,$6A,$E2,$CF,$80,$E4,$FF,$31,$C7,$C3
	.dc.b	$C3,$F5,$7C,$9F,$A0,$BB,$1F,$80,$7C,$1E,$CF,$72,$3F,$37,$E4,$3D
	.dc.b	$2F,$F1,$2E,$2F,$83,$B7,$EE,$78,$7F,$C3,$DA,$BD,$8F,$41,$2F,$EB
	.dc.b	$7C,$87,$27,$F4,$D8,$72,$7F,$8C,$9F,$E1,$6A,$FF,$36,$BE,$FF,$5F
	.dc.b	$E1,$1F,$D3,$F8,$3D,$DE,$8F,$3E,$AF,$BD,$5E,$C3,$EF,$67,$E4,$E0
	.dc.b	$F9,$44,$8F,$C4,$7C,$87,$18,$F8,$91,$F8,$57,$F1,$33,$EC,$F6,$3E
	.dc.b	$91,$27,$F2,$1F,$1D,$C7,$8F,$F6,$B1,$E1,$EA,$FC,$23,$F6,$FD,$9F
	.dc.b	$B4,$47,$F3,$1B,$F1,$7F,$C1,$C7,$1E,$3F,$98,$FE,$0F,$D8,$FC,$11
	.dc.b	$1F,$D0,$7C,$87,$1C,$3F,$D4,$3D,$7F,$8C,$9F,$E1,$FB,$7E,$A6,$7F
	.dc.b	$D4,$B8,$BD,$9F,$C6,$12,$BF,$55,$E8,$F3,$CF,$FA,$D5,$FE,$EE,$7E
	.dc.b	$59,$F9,$5F,$67,$F2,$84,$7F,$65,$F0,$1C,$5F,$F3,$5F,$47,$DD,$7F
	.dc.b	$90,$F1,$FF,$98,$7C,$BF,$EF,$64,$8F,$EA,$BC,$DF,$C9,$7B,$BF,$31
	.dc.b	$1F,$4F,$93,$E6,$BF,$E4,$AB,$E9,$5F,$BA,$D3,$FC,$0F,$3F,$A3,$F0
	.dc.b	$2E,$1F,$E2,$3B,$FF,$0A,$7A,$8D,$3C,$D7,$FC,$88,$FD,$89,$D4,$BB
	.dc.b	$7F,$A4,$7C,$AF,$F3,$C1,$FE,$12,$24,$7F,$11,$DB,$9F,$27,$F7,$59
	.dc.b	$DC,$FF,$9D,$DF,$F4,$A7,$67,$ED,$7E,$3B,$F4,$FE,$DF,$EF,$49,$5F
	.dc.b	$C6,$FE,$3B,$9E,$3C,$3C,$57,$C0,$D2,$57,$FA,$53,$1F,$E0,$1E,$CF
	.dc.b	$FA,$37,$CE,$FF,$13,$92,$7F,$93,$7B,$DD,$1F,$F4,$54,$6D,$A7,$F9
	.dc.b	$EF,$9F,$FD,$31,$BF,$F2,$CE,$5F,$F4,$6C,$7F,$8F,$63,$12,$C7,$F8
	.dc.b	$87,$C6,$F5,$8F,$F3,$B5,$72,$23,$FE,$87,$F3,$7A,$9C,$C8,$FF,$3B
	.dc.b	$F8,$CE,$5C,$9F,$9C,$9D,$CF,$13,$E0,$65,$FE,$E9,$9F,$D0,$F9,$BE
	.dc.b	$52,$4F,$F8,$83,$E3,$DF,$17,$D3,$B3,$FD,$4F,$5E,$E2,$57,$FA,$6E
	.dc.b	$3E,$09,$EF,$23,$FE,$2A,$D9,$E7,$73,$8F,$F6,$4C,$F0,$79,$1F,$EA
	.dc.b	$5F,$2F,$E4,$4C,$8E,$4F,$E5,$39,$B9,$47,$F9,$96,$B9,$3F,$E9,$F7
	.dc.b	$E9,$D7,$D1,$F3,$64,$F8,$DF,$A9,$1C,$B0,$FF,$4E,$4F,$9E,$B9,$E9
	.dc.b	$FC,$1A,$B7,$3E,$B6,$BF,$A9,$07,$F9,$1F,$AB,$A4,$6B,$BF,$FA,$8E
	.dc.b	$1D,$AB,$81,$FE,$7D,$F3,$3F,$0C,$47,$E4,$3F,$CE,$78,$1F,$07,$E7
	.dc.b	$8D,$4F,$2A,$46,$FE,$37,$F2,$1E,$6C,$AF,$A4,$FF,$A3,$FF,$69,$7B
	.dc.b	$BF,$06,$35,$3C,$DF,$CC,$79,$72,$3E,$D3,$5F,$81,$D7,$FD,$C3,$E1
	.dc.b	$FE,$54,$E7,$6F,$7B,$FA,$1F,$3A,$5F,$E0,$DA,$F2,$DD,$F1,$DF,$4E
	.dc.b	$73,$7F,$21,$F8,$5D,$3F,$DF,$7E,$1E,$61,$FD,$F5,$F8,$DF,$1D,$F8
	.dc.b	$DD,$3F,$C0,$7F,$1D,$F2,$49,$5F,$E5,$FC,$DF,$1D,$FE,$07,$F8,$A9
	.dc.b	$3F,$E0,$77,$E8,$7C,$6F,$3A,$7C,$77,$D3,$25,$73,$F8,$EF,$F0,$87
	.dc.b	$C6,$C8,$FF,$08,$3F,$67,$E3,$39,$7C,$5F,$DB,$24,$F8,$F9,$FF,$C3
	.dc.b	$1F,$19,$23,$FC,$30,$FE,$07,$E2,$7C,$5F,$13,$F8,$3F,$89,$95,$EB
	.dc.b	$8F,$F1,$07,$C6,$48,$FF,$10,$3F,$89,$F8,$AF,$57,$C5,$7E,$2C,$93
	.dc.b	$F0,$FE,$27,$F8,$B7,$C0,$48,$FD,$C3,$F9,$1F,$89,$FE,$35,$E6,$7F
	.dc.b	$55,$F1,$BF,$D8,$A7,$FA,$92,$B6,$43,$F2,$91,$FC,$87,$DC,$91,$FC
	.dc.b	$0D,$E6,$FF,$05,$8E,$87,$F2,$A3,$C4,$FE,$76,$F6,$F4,$38,$10,$FD
	.dc.b	$5F,$BA,$F6,$F7,$E6,$FF,$7E,$BE,$95,$FB,$68,$FF,$00,$F9,$B8,$8E
	.dc.b	$84,$F1,$3F,$C8,$BF,$8C,$7F,$92,$1F,$DC,$CF,$F3,$B2,$4F,$FA,$A6
	.dc.b	$5F,$15,$88,$5F,$67,$C1,$E8,$7F,$06,$7F,$DF,$97,$FD,$7A,$FF,$C8
	.dc.b	$51,$A3,$F8,$5E,$C9,$E9,$8F,$8A,$7C,$1F,$13,$FD,$37,$49,$1F,$CB
	.dc.b	$DF,$F2,$06,$0F,$86,$CC,$F9,$32,$F1,$65,$F1,$E3,$C0,$1C,$3B,$3C
	.dc.b	$7E,$F7,$F5,$EC,$39,$E5,$FC,$71,$EF,$F1,$3F,$D7,$E6,$57,$F4,$17
	.dc.b	$C7,$F1,$38,$8E,$87,$78,$FE,$D2,$7F,$6A,$F8,$0B,$1C,$48,$7F,$10
	.dc.b	$D1,$F1,$3F,$DB,$66,$4F,$F4,$F7,$E0,$FC,$4E,$23,$BE,$89,$ED,$1F
	.dc.b	$E0,$3B,$7C,$4E,$23,$F9,$64,$F1,$D7,$F0,$7F,$0B,$FD,$EF,$E1,$25
	.dc.b	$7F,$5D,$7E,$E4,$FF,$A3,$BE,$4D,$EF,$EA,$BC,$35,$FE,$06,$3F,$C0
	.dc.b	$FF,$09,$63,$91,$0F,$77,$A7,$FE,$09,$89,$1F,$DA,$DF,$F0,$67,$C1
	.dc.b	$E2,$3D,$B1,$D1,$FB,$7F,$85,$BD,$AD,$67,$FD,$4A,$AF,$A1,$F9,$3A
	.dc.b	$FC,$A5,$FF,$4B,$C3,$AB,$FD,$E5,$FF,$21,$F2,$77,$EC,$FF,$34,$E1
	.dc.b	$DE,$BE,$DD,$F7,$7C,$C4,$3F,$AB,$47,$C6,$C3,$C1,$14,$FE,$D3,$3E
	.dc.b	$2C,$7F,$D6,$ED,$9D,$CF,$EB,$FD,$2F,$17,$57,$E6,$B0,$F1,$62,$8F
	.dc.b	$66,$9D,$6D,$8F,$37,$B9,$97,$1E,$AB,$95,$E8,$67,$DF,$97,$FD,$FB
	.dc.b	$6F,$2C,$7E,$BB,$CE,$E5,$97,$92,$BF,$73,$AF,$EF,$F5,$F2,$47,$BA
	.dc.b	$79,$BF,$E0,$2D,$3F,$BE,$BD,$77,$FA,$98,$F4,$8C,$1F,$1B,$DE,$F4
	.dc.b	$11,$C9,$EE,$BE,$0F,$D0,$3F,$7B,$D4,$CB,$C9,$C1,$F1,$A7,$8E,$5F
	.dc.b	$65,$F2,$63,$E6,$D7,$E9,$E3,$D1,$B3,$C2,$3F,$D0,$FD,$39,$D3,$A4
	.dc.b	$FB,$67,$E7,$7B,$79,$8F,$84,$71,$3F,$95,$79,$BA,$7C,$71,$5F,$E0
	.dc.b	$1F,$1D,$71,$FC,$2C,$DC,$3F,$C6,$3C,$9F,$67,$E3,$31,$87,$E3,$2A
	.dc.b	$CF,$DD,$F8,$BF,$C0,$7B,$B2,$BE,$0B,$FE,$25,$F4,$F1,$1F,$56,$2E
	.dc.b	$5F,$94,$F7,$6E,$7F,$A9,$9B,$5F,$E7,$DE,$97,$C5,$FE,$36,$E4,$FC
	.dc.b	$E7,$F9,$AF,$C5,$62,$3F,$05,$17,$1F,$D0,$FB,$F7,$2F,$DB,$4D,$AF
	.dc.b	$F4,$57,$57,$FA,$31,$FC,$ED,$7E,$6C,$8F,$E2,$36,$FE,$A7,$F1,$38
	.dc.b	$8F,$C8,$4D,$C5,$8F,$CA,$C3,$F8,$CD,$7E,$12,$B5,$3F,$C1,$DF,$DC
	.dc.b	$61,$FC,$EA,$F8,$D7,$F9,$8B,$E0,$9C,$7F,$B1,$6C,$7E,$59,$AF,$F9
	.dc.b	$2B,$1F,$F4,$2B,$E7,$5F,$0F,$D9,$FF,$34,$5F,$6C,$3F,$B6,$FC,$3E
	.dc.b	$6B,$5F,$0F,$BF,$86,$98,$F3,$8F,$DF,$7C,$2C,$8F,$CF,$BD,$73,$C4
	.dc.b	$7E,$CE,$2E,$5D,$F3,$7C,$36,$37,$FE,$51,$56,$FE,$9F,$B3,$E7,$F6
	.dc.b	$BF,$86,$7C,$34,$9F,$64,$57,$F8,$04,$5C,$BB,$7B,$57,$1F,$E6,$93
	.dc.b	$6B,$EF,$FB,$3F,$0F,$D7,$91,$F1,$DF,$E5,$5F,$14,$FA,$8F,$F3,$EC
	.dc.b	$FE,$F3,$E2,$3F,$8F,$11,$F9,$DE,$7D,$7F,$89,$45,$D3,$E0,$78,$DF
	.dc.b	$CD,$7E,$55,$F8,$33,$C6,$C3,$70,$FE,$2B,$3A,$9C,$F8,$BF,$CB,$7E
	.dc.b	$2A,$57,$F0,$E7,$EE,$79,$B8,$8F,$E4,$51,$72,$FA,$1E,$9F,$F4,$5F
	.dc.b	$92,$7C,$E3,$9B,$0D,$C3,$F9,$2C,$EA,$73,$C9,$FC,$F7,$E3,$25,$7F
	.dc.b	$1E,$7F,$33,$F8,$C7,$F3,$08,$FE,$65,$1A,$69,$FA,$98,$7A,$9E,$0E
	.dc.b	$79,$3E,$0C,$BF,$C4,$FA,$7E,$8B,$3F,$3F,$2F,$F4,$DF,$8D,$97,$FC
	.dc.b	$B9,$F0,$F9,$F8,$8F,$E8,$51,$6F,$F3,$D9,$FF,$BC,$A7,$B7,$37,$C0
	.dc.b	$23,$E1,$F6,$BF,$AD,$7C,$6C,$8F,$E7,$6F,$BB,$CA,$C4,$7F,$4F,$8B
	.dc.b	$87,$6F,$07,$E5,$F3,$4A,$F6,$3F,$3B,$97,$FB,$37,$C6,$C9,$FE,$8E
	.dc.b	$FC,$AE,$CC,$47,$F5,$B8,$B8,$7A,$BA,$3E,$D6,$3F,$DC,$78,$BF,$D6
	.dc.b	$37,$83,$3E,$AF,$F7,$3F,$8D,$91,$FD,$51,$FA,$FC,$58,$8F,$ED,$11
	.dc.b	$71,$F8,$FB,$76,$E7,$95,$F8,$4C,$BE,$CB,$7D,$6F,$F0,$07,$AB,$23
	.dc.b	$FB,$0B,$F8,$4C,$71,$1F,$DD,$62,$E3,$F5,$1F,$5F,$34,$9E,$CF,$BB
	.dc.b	$D5,$FF,$06,$7C,$6C,$9F,$ED,$AF,$E4,$3E,$37,$11,$FE,$01,$8B,$87
	.dc.b	$E0,$3E,$37,$E3,$FA,$B2,$BD,$4F,$CA,$E8,$FF,$0C,$61,$27,$FB,$CB
	.dc.b	$FC,$D3,$8B,$FD,$31,$E4,$FF,$8F,$63,$53,$FE,$0F,$75,$F1,$7F,$C2
	.dc.b	$3E,$1D,$85,$25,$7F,$84,$3C,$AF,$E1,$9A,$FF,$89,$F4,$93,$FE,$08
	.dc.b	$7C,$1F,$17,$E1,$F8,$BC,$CF,$F8,$76,$70,$FD,$3E,$5F,$C1,$F9,$B2
	.dc.b	$30,$7F,$57,$DD,$F1,$5E,$B2,$46,$75,$F7,$7C,$4F,$BF,$E2,$A4,$F1
	.dc.b	$8F,$85,$F0,$D2,$3A,$3F,$C9,$BE,$AB,$E2,$C7,$F9,$43,$E3,$BF,$C8
	.dc.b	$44,$7C,$5F,$13,$C5,$74,$7F,$51,$D1,$86,$63,$FB,$0F,$CE,$BD,$E6
	.dc.b	$1F,$A0,$F1,$A6,$3F,$93,$7F,$47,$E2,$FF,$95,$7E,$1F,$C7,$3F,$E5
	.dc.b	$9E,$CC,$1F,$6F,$E2,$78,$C7,$8A,$0B,$C7,$73,$31,$DE,$BF,$C4,$E1
	.dc.b	$E8,$FE,$CB,$C8,$98,$7E,$95,$FD,$AF,$91,$E3,$F0,$24,$FF,$93,$35
	.dc.b	$FF,$6E,$BF,$57,$D5,$E2,$F8,$CF,$16,$2C,$33,$97,$87,$CA,$FE,$03
	.dc.b	$E4,$4B,$FD,$A3,$FC,$13,$C8,$F8,$5F,$2B,$27,$C0,$FE,$6F,$E1,$78
	.dc.b	$47,$7E,$1F,$7D,$E3,$1F,$77,$89,$FC,$3B,$C7,$97,$FC,$05,$FE,$23
	.dc.b	$E1,$FC,$FF,$16,$4F,$B5,$FF,$0D,$FE,$31,$FE,$20,$8E,$DC,$0B,$E7
	.dc.b	$18,$7F,$8D,$F7,$FF,$55,$FA,$6F,$C5,$EA,$69,$39,$7E,$F7,$8F,$30
	.dc.b	$F9,$8F,$83,$E1,$39,$75,$C9,$FE,$B7,$7C,$B8,$FB,$6D,$27,$2F,$87
	.dc.b	$0F,$C5,$7C,$8C,$9F,$B0,$FB,$3E,$17,$84,$7C,$9C,$17,$8E,$3F,$EB
	.dc.b	$10,$BF,$67,$93,$2F,$C6,$FB,$FC,$9F,$CC,$32,$3F,$0C,$FC,$AF,$81
	.dc.b	$E1,$1D,$E2,$E1,$DB,$CA,$F9,$1E,$54,$AF,$7B,$F2,$FE,$47,$F4,$98
	.dc.b	$C9,$FC,$93,$F5,$BD,$CE,$11,$EB,$AB,$87,$AB,$C7,$FA,$9E,$54,$AF
	.dc.b	$96,$FD,$7F,$8F,$FD,$86,$72,$7F,$3C,$FE,$0B,$D9,$E1,$1F,$1E,$AE
	.dc.b	$1F,$1B,$C9,$FC,$07,$C6,$CA,$FA,$EF,$E0,$FE,$2B,$F7,$9A,$49,$FD
	.dc.b	$53,$F8,$EF,$67,$84,$7D,$3A,$B8,$7D,$2F,$23,$F1,$9F,$03,$2B,$F0
	.dc.b	$6F,$E3,$FE,$33,$F8,$47,$1C,$9F,$DB,$3F,$E7,$4F,$C6,$3F,$AF,$C7
	.dc.b	$DF,$AE,$4D,$3F,$1B,$E4,$7F,$8B,$7E,$99,$EB,$D2,$4F,$F8,$B2,$1F
	.dc.b	$1D,$FC,$5F,$9A,$4F,$F0,$57,$C1,$ED,$F0,$8F,$DB,$57,$47,$95,$F7
	.dc.b	$F9,$2F,$07,$A3,$27,$B3,$E3,$F9,$6F,$E4,$9C,$F2,$7F,$89,$3E,$DF
	.dc.b	$2F,$84,$76,$AB,$C7,$FE,$31,$CB,$FD,$13,$3E,$9F,$5F,$DC,$47,$D9
	.dc.b	$F1,$FF,$CC,$7B,$64,$7F,$1F,$7E,$57,$B1,$B8,$F5,$D5,$C3,$B7,$8D
	.dc.b	$F2,$7E,$3E,$57,$C7,$7E,$67,$C7,$7F,$3D,$EF,$93,$FC,$AD,$FA,$DE
	.dc.b	$C7,$08,$F8,$F5,$70,$F5,$78,$DF,$57,$E3,$A5,$7D,$37,$EC,$7C,$5F
	.dc.b	$F4,$9F,$02,$4F,$F3,$77,$F0,$5E,$C7,$08,$FA,$75,$70,$F8,$DE,$3F
	.dc.b	$E0,$7E,$2E,$57,$DF,$7F,$09,$F1,$5F,$D5,$7C,$79,$3F,$D0,$DF,$C7
	.dc.b	$7B,$1C,$23,$EF,$D5,$C3,$E9,$79,$3F,$8D,$F8,$99,$5F,$8B,$7F,$21
	.dc.b	$F1,$3F,$D8,$7C,$99,$3F,$D3,$DF,$F6,$6F,$E3,$1F,$D4,$23,$F1,$75
	.dc.b	$C9,$A7,$E5,$7C,$BF,$F4,$37,$D3,$FF,$93,$49,$18,$F8,$4F,$ED,$DE
	.dc.b	$54,$8F,$EB,$8F,$6F,$57,$8C,$43,$FD,$17,$5C,$FE,$2F,$FB,$7C,$7F
	.dc.b	$5F,$F8,$57,$C8,$FF,$80,$A3,$FB,$9F,$07,$F8,$0A,$31,$1B,$6C,$7F
	.dc.b	$B9,$BC,$AE,$3C,$3F,$E0,$1F,$46,$5F,$F6,$F7,$BD,$7F,$BB,$23,$FB
	.dc.b	$67,$C0,$E3,$DB,$FF,$00,$E5,$E2,$F2,$F1,$48,$C3,$E4,$9D,$E7,$E2
	.dc.b	$6F,$85,$FE,$C8,$FA,$77,$93,$49,$7F,$E4,$1F,$5D,$E4,$FF,$85,$FD
	.dc.b	$39,$3F,$E0,$57,$EB,$FA,$FE,$2D,$3F,$C4,$59,$5F,$2C,$D3,$1E,$2F
	.dc.b	$D8,$F8,$D9,$8F,$D0,$7E,$DF,$C6,$FB,$21,$F8,$7F,$F0,$B7,$FC,$20
	.dc.b	$63,$9F,$51,$9E,$57,$17,$F8,$9B,$E5,$1E,$07,$C1,$C3,$FC,$51,$D7
	.dc.b	$FE,$AD,$7F,$C1,$8D,$BF,$F5,$D3,$6F,$FC,$5F,$F0,$68,$7C,$7F,$E4
	.dc.b	$3E,$3E,$61,$F8,$87,$F2,$7E,$B7,$93,$DB,$93,$F4,$CF,$F1,$CE,$BF
	.dc.b	$CB,$1C,$6B,$E6,$F2,$BF,$37,$F1,$F2,$7F,$26,$FE,$7B,$E3,$BF,$3F
	.dc.b	$D5,$27,$F2,$EF,$E8,$BE,$3B,$E1,$FB,$92,$7C,$07,$CA,$7F,$5B,$D9
	.dc.b	$FA,$8F,$8E,$93,$FA,$07,$F5,$7E,$BF,$EB,$3E,$3A,$4F,$E9,$1F,$D7
	.dc.b	$FC,$77,$D0,$F7,$A4,$F6,$E0,$47,$E5,$38,$3F,$A4,$BF,$B6,$F5,$E5
	.dc.b	$7E,$B9,$FD,$CF,$91,$FB,$AF,$8D,$93,$FB,$27,$F7,$BE,$AF,$DE,$F8
	.dc.b	$09,$3D,$DE,$DC,$DF,$9A,$8F,$DC,$FC,$6F,$E9,$3B,$1E,$FF,$F8,$D8
	.dc.b	$FF,$1B,$F6,$A7,$51,$FC,$3F,$BF,$FD,$F4,$DE,$FF,$C0,$B7,$0F,$F3
	.dc.b	$82,$F8,$BF,$C8,$FC,$64,$9F,$84,$F9,$3D,$FE,$BF,$19,$DB,$D5,$95
	.dc.b	$F8,$3D,$7C,$B3,$E2,$F8,$B9,$39,$CB,$FC,$68,$DF,$8B,$F7,$FB,$FF
	.dc.b	$A0,$F3,$A4,$FE,$2B,$89,$3E,$CF,$47,$DB,$E8,$CA,$E9,$1F,$03,$E2
	.dc.b	$E4,$77,$7E,$1F,$C5,$FD,$9F,$C6,$3F,$90,$CF,$F9,$FF,$B9,$18,$E6
	.dc.b	$FD,$7C,$A7,$2F,$CC,$F8,$BF,$9B,$EC,$FC,$EF,$89,$93,$ED,$8F,$A3
	.dc.b	$EB,$48,$F8,$0F,$D4,$F5,$BE,$AF,$9F,$27,$E2,$3F,$63,$E0,$BE,$CF
	.dc.b	$BF,$F6,$BD,$D9,$5F,$26,$3E,$EF,$9B,$F8,$3F,$A6,$FE,$0D,$E0,$FE
	.dc.b	$F1,$B5,$F7,$27,$5C,$F2,$3E,$B4,$FE,$DD,$FF,$49,$3F,$E8,$7E,$1F
	.dc.b	$83,$DA,$8F,$D0,$CF,$0F,$13,$C0,$78,$7C,$9F,$17,$85,$27,$AB,$E5
	.dc.b	$EC,$F3,$7B,$92,$7C,$2F,$A7,$BF,$D5,$F1,$3E,$BF,$8A,$95,$E6,$8F
	.dc.b	$77,$8B,$23,$D2,$FC,$1F,$89,$F8,$5F,$15,$F0,$FE,$2A,$57,$BA,$3E
	.dc.b	$37,$C5,$48,$F8,$2F,$C9,$F1,$7E,$57,$C2,$FC,$BF,$89,$95,$F1,$A3
	.dc.b	$E7,$7C,$54,$8F,$92,$FD,$1F,$8A,$FD,$A7,$D2,$FE,$27,$AF,$F9,$6A
	.dc.b	$F8,$BC,$9A,$BD,$FC,$5C,$7F,$51,$3F,$E5,$EE,$BF,$A2,$BE,$23,$AF
	.dc.b	$C4,$4A,$DB,$E0,$F8,$DF,$0F,$C3,$F8,$BE,$1E,$57,$48,$F2,$FC,$3F
	.dc.b	$9B,$E1,$E4,$F8,$23,$D3,$F0,$F2,$3C,$6F,$B3,$E1,$FD,$BF,$0F,$27
	.dc.b	$CE,$FC,$0F,$86,$F8,$3F,$0D,$27,$D6,$FC,$4F,$86,$F8,$BF,$0D,$27
	.dc.b	$DE,$FC,$8F,$85,$E1,$1F,$EE,$3E,$25,$E3,$8F,$E1,$FC,$2F,$CE,$F8
	.dc.b	$5F,$9F,$F0,$B3,$0F,$93,$1F,$4B,$E1,$24,$7C,$C7,$F9,$47,$D2,$7F
	.dc.b	$B7,$0F,$E4,$1E,$79,$A4,$F8,$5D,$FC,$24,$BF,$E3,$93,$FE,$CE,$6F
	.dc.b	$C2,$F4,$F8,$59,$1A,$7B,$FC,$2F,$83,$E1,$24,$F2,$7C,$7F,$0B,$E4
	.dc.b	$F8,$5F,$2F,$C2,$CA,$EF,$1E,$8F,$83,$91,$E2,$7D,$7F,$05,$EC,$F8
	.dc.b	$39,$3E,$67,$DF,$F0,$7F,$03,$E0,$E4,$FA,$9F,$87,$F0,$7F,$13,$E0
	.dc.b	$E4,$FB,$9F,$8F,$F0,$5F,$23,$E0,$E4,$FC,$27,$E5,$FC,$1F,$F5,$3D
	.dc.b	$FF,$87,$F3,$9A,$4F,$83,$D7,$C4,$4C,$3F,$A3,$4F,$F0,$FF,$47,$FC
	.dc.b	$26,$FF,$0E,$87,$7F,$F1,$28,$7C,$1F,$72,$4F,$2F,$82,$F1,$7C,$14
	.dc.b	$8E,$AF,$97,$E0,$BC,$DF,$05,$27,$C2,$FA,$7E,$07,$D5,$F0,$52,$7C
	.dc.b	$AF,$B7,$E0,$BD,$DF,$05,$27,$D2,$FC,$1F,$81,$F8,$5F,$05,$27,$DA
	.dc.b	$FC,$5F,$82,$F8,$DF,$05,$27,$E0,$BF,$27,$E0,$7E,$57,$C1,$7C,$BF
	.dc.b	$82,$95,$F1,$A3,$E7,$7C,$17,$F8,$07,$F1,$9F,$E1,$D2,$EA,$46,$1F
	.dc.b	$A5,$3B,$F8,$39,$5F,$DD,$6B,$F8,$A3,$7E,$0F,$A7,$C1,$C8,$D3,$DF
	.dc.b	$E0,$FC,$1F,$07,$E1,$F8,$29,$5C,$E3,$C9,$F0,$7E,$5F,$82,$93,$DE
	.dc.b	$3D,$1F,$07,$E9,$F8,$39,$3E,$38,$F6,$7C,$1C,$8F,$33,$EF,$F8,$3F
	.dc.b	$81,$F0,$72,$7D,$4F,$C3,$F7,$FE,$27,$C1,$FC,$5F,$83,$95,$EF,$8F
	.dc.b	$91,$F0,$7F,$27,$E0,$A4,$FC,$38,$F9,$9F,$03,$F3,$7E,$0A,$4F,$C7
	.dc.b	$8F,$A1,$F0,$5F,$E3,$1F,$C6,$BF,$D3,$5E,$6F,$F6,$66,$F9,$DF,$E6
	.dc.b	$CE,$AF,$F1,$96,$1F,$E2,$39,$FE,$1E,$4F,$93,$E5,$F0,$9C,$FE,$13
	.dc.b	$A7,$C1,$CB,$E1,$1D,$FE,$0F,$C3,$1F,$67,$4C,$3D,$70,$F2,$11,$F0
	.dc.b	$61,$D5,$F1,$61,$D5,$FD,$6E,$7F,$E8,$2C,$BF,$8E,$46,$0F,$C6,$2F
	.dc.b	$83,$F8,$04,$8F,$67,$C1,$FC,$3F,$57,$E2,$7C,$14,$9F,$81,$1F,$1F
	.dc.b	$E0,$FE,$47,$C1,$C9,$F8,$91,$F2,$FE,$0F,$E6,$7C,$1C,$9F,$91,$1F
	.dc.b	$3F,$E0,$E4,$7C,$B7,$E9,$7C,$1F,$D3,$F8,$3F,$A9,$F0,$72,$BE,$84
	.dc.b	$7D,$7F,$83,$91,$F4,$DF,$B5,$F0,$7F,$6F,$E0,$E4,$FD,$77,$FC,$DD
	.dc.b	$F8,$DF,$F1,$C3,$FC,$4D,$D9,$FE,$24,$6F,$2F,$FA,$A9,$B5,$FB,$37
	.dc.b	$F8,$E3,$7E,$1F,$BB,$F8,$3C,$3A,$C7,$82,$3E,$87,$A6,$E7,$65,$9C
	.dc.b	$BC,$1F,$1D,$E4,$F8,$69,$87,$F9,$BF,$3F,$0E,$27,$C2,$DF,$85,$F5
	.dc.b	$92,$78,$7C,$2F,$BB,$E1,$7D,$FF,$0B,$27,$A4,$7C,$2F,$85,$F8,$7F
	.dc.b	$0B,$27,$C1,$1F,$1A,$FA,$47,$AB,$2F,$E6,$EF,$63,$99,$E7,$C3,$B3
	.dc.b	$BC,$3B,$3E,$BC,$3B,$3C,$F0,$FA,$7F,$0B,$23,$E0,$3F,$5B,$E1,$7D
	.dc.b	$51,$D4,$87,$53,$CE,$75,$21,$D4,$F9,$5F,$0B,$F8,$02,$3F,$37,$E1
	.dc.b	$7F,$07,$F0,$B2,$3E,$83,$F8,$6F,$85,$FC,$3F,$C2,$C9,$FA,$8F,$E2
	.dc.b	$BE,$17,$F1,$77,$27,$A1,$F6,$7E,$13,$A3,$F2,$70,$BC,$DF,$8F,$C7
	.dc.b	$A4,$FC,$DC,$6C,$FE,$4F,$1F,$CB,$FC,$2C,$AF,$C2,$3F,$9A,$F8,$5F
	.dc.b	$F6,$47,$B9,$FD,$14,$93,$8F,$85,$CF,$C2,$EB,$E1,$64,$FF,$AE,$EB
	.dc.b	$FC,$A9,$E0,$FE,$7F,$0E,$D3,$F7,$F1,$EC,$EC,$EE,$43,$96,$3D,$9E
	.dc.b	$98,$76,$7F,$43,$87,$67,$FA,$E2,$F8,$5F,$49,$23,$C9,$F0,$9E,$CF
	.dc.b	$85,$F6,$FC,$24,$9F,$44,$7C,$0F,$85,$F8,$38,$49,$F6,$47,$C4,$B9
	.dc.b	$1D,$0F,$81,$F0,$BD,$5F,$85,$87,$57,$CF,$87,$57,$E3,$61,$D5,$F5
	.dc.b	$E1,$F4,$3E,$17,$E8,$E1,$27,$E6,$47,$57,$E0,$E1,$D7,$8B,$FD,$63
	.dc.b	$83,$1E,$D5,$F1,$A3,$B4,$EB,$1F,$BB,$F0,$B2,$BE,$C3,$F8,$0F,$85
	.dc.b	$FC,$0F,$C2,$FE,$0B,$E1,$25,$7D,$D8,$FC,$2F,$C2,$C8,$FC,$03,$F8
	.dc.b	$8F,$84,$FC,$4F,$C2,$C9,$FC,$23,$F8,$CF,$84,$EB,$1F,$87,$C3,$AB
	.dc.b	$F7,$30,$E8,$FD,$EC,$2E,$9E,$EC,$7F,$2B,$F0,$B2,$BF,$20,$FE,$63
	.dc.b	$E0,$91,$2B,$FD,$AB,$F8,$E7,$F7,$08,$FF,$11,$3F,$E7,$C2,$75,$F0
	.dc.b	$8F,$E5,$B0,$F8,$2C,$57,$26,$D7,$6F,$39,$5B,$BF,$04,$E6,$3F,$B3
	.dc.b	$DF,$8B,$E0,$B8,$D7,$89,$B7,$C9,$B4,$F8,$4B,$86,$3E,$12,$61,$D8
	.dc.b	$F5,$C7,$0C,$A4,$67,$1F,$2C,$74,$F8,$5E,$BF,$0B,$DB,$E1,$65,$EF
	.dc.b	$3F,$EF,$4D,$F8,$4F,$19,$1D,$7C,$2F,$9B,$E1,$24,$71,$7D,$3F,$0B
	.dc.b	$EA,$F8,$4E,$F3,$D7,$2C,$3D,$48,$7D,$0C,$3E,$0E,$1D,$23,$E9,$E1
	.dc.b	$78,$BE,$DE,$3F,$1B,$E1,$79,$D7,$D8,$CB,$0F,$4A,$FC,$0E,$3E,$98
	.dc.b	$FB,$B8,$F3,$E0,$FA,$D9,$59,$FC,$2E,$5F,$4B,$E1,$7E,$9F,$C2,$C9
	.dc.b	$F8,$D1,$F5,$BE,$13,$EB,$FC,$2C,$9F,$95,$1F,$6B,$E1,$64,$7C,$D7
	.dc.b	$EE,$FC,$2F,$DE,$F8,$3E,$D3,$D0,$FA,$4F,$57,$EA,$5F,$57,$F2,$B7
	.dc.b	$D1,$FC,$BD,$DD,$FF,$3F,$87,$E3,$3E,$A7,$FD,$5E,$7F,$80,$09,$F0
	.dc.b	$99,$F8,$89,$87,$E1,$E7,$FC,$06,$7F,$AB,$6D,$C1,$F9,$F8,$FE,$EF
	.dc.b	$3F,$E1,$26,$CE,$A3,$FD,$1D,$18,$9E,$1F,$07,$E3,$F8,$19,$5D,$DF
	.dc.b	$37,$C1,$78,$A3,$A6,$37,$1E,$9D,$AC,$BC,$99,$78,$FE,$13,$DC,$4C
	.dc.b	$3D,$59,$FC,$1F,$82,$91,$ED,$7E,$27,$C1,$FC,$5F,$83,$F8,$DF,$07
	.dc.b	$2B,$E1,$47,$C9,$F8,$3F,$95,$F0,$72,$7E,$34,$72,$7E,$45,$D7,$E2
	.dc.b	$7C,$17,$57,$E6,$61,$D5,$F9,$18,$75,$7E,$86,$1D,$5F,$99,$87,$D9
	.dc.b	$F8,$29,$1F,$55,$FB,$9F,$05,$D6,$3E,$96,$1D,$78,$7E,$3F,$03,$2E
	.dc.b	$D5,$F5,$F2,$ED,$5A,$CB,$F0,$DF,$05,$F8,$7F,$83,$97,$F8,$38,$FC
	.dc.b	$57,$C0,$C8,$FC,$33,$F8,$DF,$82,$FE,$33,$F4,$FF,$EC,$B3,$FD,$8F
	.dc.b	$3F,$E5,$66,$E3,$F6,$9D,$7C,$27,$0F,$83,$95,$FC,$5E,$7F,$BE,$74
	.dc.b	$7D,$FC,$7C,$13,$F8,$F7,$C0,$EC,$F0,$FC,$24,$70,$73,$F8,$4F,$29
	.dc.b	$1E,$DF,$09,$E8,$F8,$1F,$4F,$C1,$49,$F1,$47,$B3,$E0,$64,$79,$5F
	.dc.b	$7F,$C1,$7C,$0F,$84,$F8,$38,$4A,$F5,$47,$57,$D9,$87,$57,$E2,$61
	.dc.b	$D1,$F8,$D8,$5E,$0F,$8D,$8F,$CC,$F8,$4F,$9B,$F0,$92,$FE,$34,$7D
	.dc.b	$0F,$84,$91,$F2,$5F,$A7,$F0,$3F,$53,$E0,$BE,$AF,$C1,$4A,$F9,$D1
	.dc.b	$F6,$3E,$0B,$EC,$FC,$24,$9F,$A5,$1D,$1F,$B5,$85,$D3,$ED,$63,$D6
	.dc.b	$7E,$C6,$3D,$5F,$C0,$63,$F8,$4F,$84,$91,$F7,$5F,$C3,$FC,$27,$3F
	.dc.b	$2B,$FB,$57,$4B,$3C,$5B,$A5,$2C,$FC,$13,$F8,$DC,$FC,$1A,$7E,$37
	.dc.b	$4F,$C9,$FC,$27,$F6,$CF,$23,$FC,$01,$91,$B9,$7F,$BC,$31,$FF,$55
	.dc.b	$C3,$E0,$B8,$12,$3F,$B5,$47,$F7,$D7,$7F,$03,$D7,$E1,$24,$6A,$3C
	.dc.b	$1F,$09,$E1,$F8,$19,$3C,$A3,$A3,$E0,$C2,$EB,$FC,$7B,$1E,$CD,$C3
	.dc.b	$AD,$79,$B1,$F6,$47,$B7,$E0,$E4,$F9,$9F,$81,$F0,$1F,$07,$E0,$3E
	.dc.b	$17,$C0,$4A,$F5,$C7,$C5,$F8,$09,$1E,$E7,$E4,$7C,$0F,$C9,$F8,$29
	.dc.b	$3F,$09,$F9,$9F,$03,$D6,$3E,$2E,$1D,$5F,$97,$87,$47,$E6,$E1,$75
	.dc.b	$F7,$63,$F5,$7E,$0A,$57,$CE,$7E,$C7,$C0,$7D,$9F,$81,$93,$F4,$9F
	.dc.b	$B9,$F0,$3F,$77,$E0,$7E,$F7,$C1,$4A,$FA,$F1,$FE,$2C,$FC,$75,$F0
	.dc.b	$77,$E1,$E5,$7F,$89,$67,$2F,$52,$1F,$84,$C7,$BB,$D0,$FC,$1E,$1D
	.dc.b	$DF,$C4,$E1,$DF,$E0,$BC,$1F,$13,$27,$B4,$7A,$9E,$87,$6D,$FF,$01
	.dc.b	$A7,$AE,$3A,$90,$EA,$79,$FE,$0F,$D9,$F0,$72,$30,$7A,$FE,$17,$E0
	.dc.b	$11,$F7,$7C,$2F,$C3,$F8,$5F,$89,$DD,$27,$E1,$47,$C7,$F8,$09,$1F
	.dc.b	$15,$F9,$5F,$05,$CE,$1A,$7F,$74,$E2,$65,$98,$C6,$5D,$EB,$E7,$65
	.dc.b	$DD,$E3,$E5,$FF,$9A,$70,$67,$F5,$FE,$03,$EC,$7C,$0C,$BF,$AD,$1F
	.dc.b	$6F,$E1,$64,$7D,$97,$EF,$7C,$2F,$DF,$F8,$5F,$C0,$7C,$1C,$AF,$BD
	.dc.b	$1F,$83,$F8,$49,$1F,$81,$7F,$0D,$F0,$3D,$23,$CA,$7F,$11,$D1,$8E
	.dc.b	$78,$3E,$06,$5D,$EB,$F1,$59,$77,$78,$F9,$CC,$19,$FE,$4F,$E1,$25
	.dc.b	$7E,$45,$FF,$40,$FE,$3D,$FE,$91,$8E,$2F,$F6,$8E,$1F,$F4,$73,$FD
	.dc.b	$90,$9F,$07,$C3,$E1,$38,$FC,$24,$BF,$F4,$05,$27,$AF,$C2,$76,$98
	.dc.b	$69,$F0,$DF,$8B,$E1,$64,$F2,$8E,$CF,$4C,$3B,$3E,$6B,$F4,$FC,$1C
	.dc.b	$8F,$0B,$EC,$F8,$4E,$B1,$E4,$C3,$AB,$EE,$C3,$AB,$F0,$30,$EB,$C5
	.dc.b	$FE,$40,$CF,$D5,$3E,$CF,$84,$F8,$FF,$09,$23,$07,$C0,$F8,$5F,$97
	.dc.b	$F0,$92,$30,$7C,$4F,$85,$F9,$E4,$7E,$3F,$C0,$FD,$2F,$81,$91,$F2
	.dc.b	$9F,$AB,$F0,$1F,$5B,$0E,$93,$F5,$B0,$C3,$D4,$87,$DA,$C3,$EE,$E1
	.dc.b	$F7,$BE,$07,$EF,$FC,$24,$AF,$AF,$1F,$82,$F8,$59,$1F,$69,$FC,$2F
	.dc.b	$C2,$FE,$1B,$E1,$7F,$0F,$F0,$B2,$BE,$FC,$7F,$B2,$BF,$1F,$FE,$BE
	.dc.b	$4E,$BC,$7C,$1F,$4A,$FC,$6E,$37,$97,$E7,$65,$D6,$7F,$13,$97,$57
	.dc.b	$F2,$79,$75,$F8,$9E,$DF,$07,$27,$FD,$9D,$A7,$F5,$8D,$3F,$DC,$8B
	.dc.b	$E1,$3C,$9F,$0B,$2B,$71,$E7,$F8,$6F,$47,$C3,$49,$E7,$1E,$BF,$86
	.dc.b	$F6,$7C,$1C,$9E,$F1,$EF,$BE,$91,$FB,$0C,$7F,$DD,$0F,$63,$99,$E6
	.dc.b	$C3,$B3,$BC,$3E,$47,$C2,$48,$F5,$BF,$2F,$E1,$3B,$47,$BB,$3E,$FC
	.dc.b	$5F,$E9,$4C,$7D,$D3,$D8,$F1,$63,$DA,$75,$8F,$D5,$F8,$49,$5F,$21
	.dc.b	$FB,$1F,$0D,$F6,$7E,$1B,$ED,$7C,$34,$AF,$9B,$1F,$77,$E1,$BE,$F7
	.dc.b	$C2,$49,$FA,$31,$F8,$1F,$80,$91,$F5,$1F,$C2,$7C,$1F,$58,$FA,$F8
	.dc.b	$75,$7E,$46,$1D,$5F,$B7,$87,$57,$E6,$61,$F8,$DF,$84,$44,$D7,$FD
	.dc.b	$6B,$CB,$FD,$AA,$3F,$90,$B6,$3E,$34,$F0,$9C,$FC,$2D,$9E,$3F,$09
	.dc.b	$7A,$3A,$FA,$AC,$53,$9C,$3F,$C8,$90,$C3,$15,$BF,$85,$B3,$FE,$C6
	.dc.b	$AD,$C6,$3E,$17,$3F,$0B,$27,$D5,$97,$82,$D3,$CB,$E0,$F9,$E5,$D3
	.dc.b	$39,$87,$03,$C7,$1E,$0C,$A4,$67,$3F,$23,$71,$FF,$72,$3E,$AF,$83
	.dc.b	$F3,$FC,$14,$8D,$47,$AB,$E0,$E4,$71,$7D,$BF,$0B,$EE,$F8,$5F,$7F
	.dc.b	$C2,$4A,$ED,$1F,$0B,$E1,$64,$78,$5F,$8B,$F0,$7D,$23,$EB,$E1,$79
	.dc.b	$3E,$DE,$3D,$67,$D1,$8F,$57,$EE,$E3,$F3,$BE,$0F,$E7,$FC,$0C,$9F
	.dc.b	$74,$73,$E5,$F9,$99,$59,$FB,$38,$F5,$7F,$0F,$8F,$5C,$BF,$17,$97
	.dc.b	$DA,$F8,$39,$5F,$25,$FB,$BF,$0B,$F7,$BE,$17,$EF,$FC,$24,$AF,$9D
	.dc.b	$1F,$84,$FC,$83,$FB,$66,$2D,$7F,$1B,$18,$6D,$6F,$E0,$38,$7C,$0F
	.dc.b	$1F,$84,$93,$F6,$EF,$FD,$AB,$CE,$CB,$C1,$3F,$A0,$CF,$C8,$EC,$F1
	.dc.b	$10,$E5,$8F,$97,$E0,$BC,$DF,$0B,$27,$B4,$7A,$7E,$16,$47,$85,$F6
	.dc.b	$7C,$2F,$B7,$E1,$3D,$DF,$09,$2B,$CD,$1F,$07,$E0,$A4,$7A,$5F,$89
	.dc.b	$F0,$5D,$23,$FC,$2F,$BB,$1C,$71,$FC,$79,$FF,$59,$4B,$2F,$04,$7C
	.dc.b	$3C,$BC,$0F,$2C,$BE,$87,$C1,$48,$F9,$0F,$D3,$F8,$2F,$A9,$F0,$9F
	.dc.b	$57,$E1,$25,$7C,$D8,$FB,$1F,$07,$23,$E8,$3F,$6F,$E1,$3E,$E7,$C1
	.dc.b	$C9,$FA,$8F,$DF,$F8,$3E,$B1,$F5,$F0,$EA,$FB,$B0,$E8,$FC,$0C,$2E
	.dc.b	$FE,$FF,$21,$C0,$CF,$F1,$5F,$05,$30,$FC,$13,$F8,$DF,$83,$EF,$1F
	.dc.b	$2B,$3E,$F8,$F3,$C3,$FD,$B1,$3E,$A2,$1E,$13,$F1,$98,$7F,$1A,$FC
	.dc.b	$87,$FD,$AE,$4E,$FC,$7C,$2E,$7E,$17,$5F,$0B,$30,$FE,$1D,$7F,$DC
	.dc.b	$5F,$F7,$43,$FD,$51,$FF,$29,$B7,$2F,$73,$70,$FE,$5E,$FF,$68,$9D
	.dc.b	$FB,$DE,$32,$3D,$BD,$FF,$13,$E0,$C3,$C4,$F6,$C7,$D5,$F0,$92,$3C
	.dc.b	$AF,$B7,$E0,$BD,$DF,$0B,$EF,$F1,$65,$7A,$A3,$E1,$7C,$2F,$C3,$F8
	.dc.b	$59,$3E,$E8,$F8,$DF,$09,$23,$E0,$BF,$27,$E1,$3E,$57,$C0,$C9,$F8
	.dc.b	$AF,$CD,$F8,$2E,$B1,$F2,$30,$EA,$FC,$6C,$3A,$3F,$23,$0B,$C7,$F2
	.dc.b	$31,$FA,$DF,$03,$F5,$FE,$0A,$5F,$D2,$8F,$B5,$F0,$92,$3E,$AB,$F7
	.dc.b	$7E,$13,$EF,$7C,$24,$9F,$B2,$FF,$5A,$FC,$8B,$FB,$47,$4F,$F6,$27
	.dc.b	$FA,$F4,$FF,$4B,$6B,$FE,$8E,$6E,$7F,$99,$7F,$B3,$C7,$80,$87,$E0
	.dc.b	$B0,$F1,$3E,$02,$1F,$86,$BF,$17,$C4,$F8,$FE,$06,$4F,$85,$F3,$7C
	.dc.b	$37,$B6,$3A,$1D,$B1,$FC,$6E,$7E,$D8,$E8,$76,$8F,$35,$FB,$FE,$1A
	.dc.b	$4F,$B5,$F8,$5F,$0D,$F0,$FE,$0F,$E2,$7C,$24,$AF,$85,$1F,$1F,$E0
	.dc.b	$3E,$47,$C0,$49,$F8,$D1,$F2,$FD,$F9,$1F,$25,$F9,$DF,$01,$D6,$3E
	.dc.b	$66,$1D,$5F,$99,$87,$57,$E8,$61,$D5,$FA,$18,$7D,$7F,$84,$FB,$1F
	.dc.b	$09,$27,$EB,$47,$DB,$F8,$39,$1F,$65,$FB,$DF,$09,$F7,$FE,$12,$4F
	.dc.b	$DD,$7F,$05,$F0,$9F,$83,$F8,$29,$3F,$81,$7F,$0D,$F0,$5D,$23,$F0
	.dc.b	$58,$5E,$4F,$C1,$63,$D2,$7F,$0D,$8D,$9F,$C3,$63,$FE,$28,$F7,$FF
	.dc.b	$B1,$53,$D4,$46,$1B,$1B,$F8,$2E,$1F,$09,$C7,$E1,$24,$FF,$87,$ED
	.dc.b	$3D,$7E,$0F,$B7,$C2,$4A,$D4,$78,$7E,$0F,$C5,$F0,$72,$79,$47,$67
	.dc.b	$A6,$3D,$9F,$F7,$6E,$3D,$9E,$67,$87,$0E,$CE,$B0,$F7,$7C,$14,$8F
	.dc.b	$33,$F0,$7E,$0F,$B4,$74,$C3,$B6,$FF,$D0,$F1,$ED,$3C,$CE,$C4,$30
	.dc.b	$79,$1F,$99,$F0,$92,$3E,$24,$7C,$FF,$84,$91,$F1,$DF,$A5,$F0,$9F
	.dc.b	$4F,$E0,$BE,$A7,$BF,$2B,$E6,$47,$D7,$F8,$29,$1F,$3D,$FB,$5E,$FF
	.dc.b	$48,$F8,$B8,$5C,$FD,$58,$F5,$9F,$AD,$8F,$57,$E5,$E3,$F8,$3F,$84
	.dc.b	$91,$F6,$DF,$C3,$7C,$27,$E1,$FE,$12,$4F,$DF,$7F,$15,$F0,$9F,$E7
	.dc.b	$2F,$C8,$FF,$C1,$12,$CF,$FB,$64,$3E,$13,$46,$FD,$3E,$1F,$01,$2F
	.dc.b	$FC,$D1,$5F,$D6,$FB,$4B,$E0,$7C,$33,$F8,$2C,$EE,$5F,$5F,$3E,$93
	.dc.b	$CB,$3E,$EF,$4C,$FC,$DF,$03,$23,$BB,$E9,$F8,$1F,$57,$C1,$FA,$FE
	.dc.b	$0E,$57,$92,$3D,$DF,$05,$EF,$F8,$39,$3E,$88,$F8,$5F,$01,$23,$D6
	.dc.b	$FC,$5F,$81,$EB,$1E,$EC,$3A,$BF,$1F,$3F,$F5,$0F,$0B,$2E,$F3,$F1
	.dc.b	$32,$EE,$EC,$FD,$C4,$B2,$FA,$5F,$03,$F4,$FE,$06,$57,$CC,$8F,$0B
	.dc.b	$C7,$93,$E6,$E7,$E2,$E2,$FA,$F9,$F8,$2B,$59,$7F,$AC,$DF,$11,$EF
	.dc.b	$3E,$BF,$C1,$FE,$04,$8F,$DA,$F8,$3F,$C2,$7C,$17,$E1,$7E,$0E,$4F
	.dc.b	$DE,$8F,$C4,$7C,$17,$E2,$7D,$F9,$3F,$82,$8F,$C6,$7C,$17,$FA,$FB
	.dc.b	$F2,$4F,$EC,$B3,$FC,$5D,$BB,$FE,$1A,$7B,$CF,$E2,$B1,$EE,$FA,$B1
	.dc.b	$EE,$FE,$3B,$1E,$EF,$BB,$1E,$DF,$0F,$23,$FD,$75,$97,$F6,$EE,$5F
	.dc.b	$F6,$AB,$FD,$BA,$3E,$53,$FE,$CF,$CC,$D2,$7C,$1F,$A7,$E0,$A5,$74
	.dc.b	$7D,$9F,$05,$ED,$F7,$FD,$DF,$01,$2B,$C3,$1F,$07,$E1,$24,$79,$1F
	.dc.b	$89,$87,$48,$F1,$E1,$79,$BF,$CB,$99,$32,$F9,$5F,$01,$F2,$FE,$06
	.dc.b	$61,$EE,$8F,$9D,$F0,$72,$3E,$0B,$F4,$7E,$0F,$E9,$7C,$14,$9F,$8A
	.dc.b	$FD,$5F,$83,$FA,$DF,$03,$F5,$FE,$06,$57,$CA,$8E,$8F,$C8,$C2,$EB
	.dc.b	$E6,$C7,$A4,$FC,$CC,$6C,$FA,$B1,$FC,$17,$C0,$FE,$0F,$E0,$65,$FD
	.dc.b	$68,$E7,$CD,$F5,$B2,$B3,$F8,$7C,$B9,$CF,$DA,$CA,$CF,$E2,$F1,$FC
	.dc.b	$77,$C1,$FE,$3F,$E0,$E4,$FE,$0A,$0F,$FB,$F7,$F2,$5F,$EC,$C3,$E1
	.dc.b	$33,$1F,$AD,$8C,$3B,$33,$B7,$FA,$3E,$3F,$A6,$B7,$E1,$33,$9F,$6E
	.dc.b	$B5,$6A,$30,$BD,$ED,$E1,$FD,$E6,$1F,$03,$8B,$E3,$F0,$56,$4D,$9C
	.dc.b	$56,$73,$8F,$9C,$F7,$65,$C4,$8F,$C1,$C7,$A7,$C1,$C8,$F8,$8F,$7F
	.dc.b	$84,$F0,$7C,$27,$87,$E1,$25,$7C,$98,$F2,$7C,$1F,$97,$E1,$24,$FC
	.dc.b	$D8,$F4,$7C,$17,$A7,$E0,$A4,$FD,$18,$EA,$FD,$3C,$3A,$BF,$47,$0E
	.dc.b	$AF,$D7,$C3,$AB,$F5,$70,$F8,$9F,$03,$23,$EE,$3F,$1F,$E0,$BE,$47
	.dc.b	$C2,$7C,$9F,$84,$95,$F8,$18,$F9,$9F,$09,$F3,$7E,$12,$4F,$E1,$63
	.dc.b	$E8,$7C,$17,$D1,$F8,$29,$3F,$89,$8E,$AF,$E2,$F0,$EA,$FE,$27,$0E
	.dc.b	$AF,$E3,$F0,$EA,$FE,$37,$0F,$BD,$E3,$FF,$70,$EC,$FF,$6D,$37,$B3
	.dc.b	$F1,$4D,$AF,$F1,$F3,$FD,$CD,$F5,$1E,$12,$1E,$13,$D4,$78,$48,$78
	.dc.b	$4F,$D1,$CF,$83,$E1,$3C,$7F,$09,$27,$37,$E2,$9F,$3F,$C1,$C9,$C9
	.dc.b	$E4,$7D,$7F,$07,$EC,$F7,$3D,$BE,$EC,$AD,$F8,$4D,$3F,$B3,$46,$1F
	.dc.b	$EC,$4D,$F8,$1F,$8A,$7C,$6F,$81,$F0,$47,$F6,$19,$FF,$71,$3E,$13
	.dc.b	$96,$3E,$17,$A6,$1E,$17,$86,$3F,$43,$E0,$E4,$78,$5F,$A7,$F0,$7F
	.dc.b	$53,$E0,$BE,$AF,$C1,$CA,$F3,$47,$D8,$F8,$2F,$B3,$F0,$52,$7D,$51
	.dc.b	$F7,$3E,$02,$47,$B5,$FB,$FF,$03,$D6,$3E,$06,$1D,$5F,$66,$1D,$5F
	.dc.b	$89,$87,$57,$E0,$61,$F8,$8F,$80,$91,$F2,$5F,$C5,$FC,$0F,$E3,$3E
	.dc.b	$0A,$4F,$CD,$7F,$86,$7E,$4D,$FD,$C0,$FE,$C9,$1F,$E8,$96,$F2,$9A
	.dc.b	$7F,$67,$EA,$74,$FF,$1D,$8F,$ED,$04,$F8,$0E,$BF,$01,$E0,$C3,$EB
	.dc.b	$63,$70,$E3,$1D,$FE,$1B,$C7,$F0,$92,$3F,$0B,$8F,$83,$E2,$BC,$FF
	.dc.b	$07,$27,$2F,$AB,$E0,$BA,$6B,$FA,$6C,$ED,$75,$39,$E3,$D5,$FF,$14
	.dc.b	$63,$D5,$EF,$F0,$9F,$13,$E1,$64,$60,$F1,$FC,$27,$C8,$23,$E6,$EF
	.dc.b	$F9,$7F,$01,$F3,$3E,$02,$4F,$AA,$3E,$7F,$C0,$48,$F6,$BF,$4B,$E0
	.dc.b	$39,$C7,$93,$0B,$C7,$F1,$B9,$BE,$96,$5D,$AB,$E2,$E5,$D9,$F5,$E5
	.dc.b	$F7,$3E,$0F,$EE,$FC,$24,$9F,$97,$1F,$80,$F8,$39,$1F,$39,$FC,$1F
	.dc.b	$C2,$7E,$13,$E0,$FF,$0B,$F0,$92,$BE,$9C,$7F,$53,$FC,$9F,$FD,$98
	.dc.b	$48,$C7,$C1,$74,$7E,$6E,$17,$7F,$57,$2B,$2E,$D5,$F7,$B2,$EC,$FD
	.dc.b	$4C,$BB,$7C,$17,$7F,$82,$93,$FC,$23,$4F,$ED,$4F,$FB,$7D,$FE,$CE
	.dc.b	$E7,$E1,$3C,$C7,$9F,$E1,$3D,$1F,$09,$27,$93,$EB,$F8,$4F,$67,$C0
	.dc.b	$C9,$EC,$FB,$FE,$0B,$94,$75,$21,$FE,$71,$C7,$C2,$DB,$EC,$D3,$D1
	.dc.b	$7F,$27,$E0,$7E,$57,$C1,$49,$F6,$47,$67,$A6,$7E,$A7,$E0,$67,$D3
	.dc.b	$8B,$F0,$99,$32,$C7,$85,$FE,$0F,$86,$7F,$5F,$E0,$E6,$3F,$25,$FB
	.dc.b	$5F,$09,$F6,$FE,$12,$4F,$CD,$7E,$F7,$C2,$7D,$FF,$81,$FC,$07,$C1
	.dc.b	$4A,$FA,$51,$F8,$3F,$81,$FC,$27,$C1,$49,$FA,$D1,$D1,$F8,$D8,$5D
	.dc.b	$FD,$38,$F5,$9F,$B9,$8F,$57,$E6,$63,$F8,$FF,$83,$FC,$87,$C2,$49
	.dc.b	$FC,$14,$7F,$89,$7F,$28,$FE,$C9,$E3,$1A,$4F,$89,$D7,$C2,$CB,$FE
	.dc.b	$A3,$3F,$ED,$53,$FD,$A8,$E3,$E1,$7A,$7C,$1C,$8C,$1C,$39,$BC,$04
	.dc.b	$79,$E1,$D5,$FF,$5A,$5F,$5C,$3C,$DF,$0D,$E7,$F8,$3F,$47,$C2,$4B
	.dc.b	$F1,$C7,$AF,$E1,$64,$79,$9F,$77,$C3,$7B,$FE,$1A,$4F,$A9,$F8,$5F
	.dc.b	$0D,$F0,$FE,$0F,$E2,$7C,$24,$AF,$7C,$74,$7C,$B8,$5E,$8F,$E1,$DB
	.dc.b	$7F,$B1,$67,$D3,$8F,$66,$9F,$23,$1F,$A1,$F0,$72,$3E,$5B,$F4,$FE
	.dc.b	$13,$A7,$17,$F9,$87,$0F,$EF,$75,$E0,$3E,$8E,$BD,$72,$F0,$E5,$6F
	.dc.b	$86,$5F,$77,$E1,$BE,$F7,$C3,$4C,$7E,$CC,$7E,$07,$E1,$A4,$7D,$C7
	.dc.b	$F0,$9F,$0D,$F8,$5F,$80,$93,$F8,$07,$FC,$D9,$97,$FB,$40,$FE,$BE
	.dc.b	$7F,$5D,$CB,$FC,$F0,$D9,$75,$23,$FB,$2E,$CC,$7C,$35,$F8,$07,$C2
	.dc.b	$F7,$21,$F8,$4C,$3C,$3F,$09,$E2,$F8,$6F,$1F,$C3,$4A,$FF,$12,$E3
	.dc.b	$FE,$D3,$1F,$0B,$E8,$24,$6B,$E1,$7D,$7F,$09,$23,$8B,$EE,$F8,$4F
	.dc.b	$7F,$C1,$7C,$0F,$82,$95,$DA,$39,$AE,$6F,$EA,$19,$78,$23,$C7,$97
	.dc.b	$C8,$CB,$B4,$6F,$8B,$E3,$E7,$8C,$38,$E7,$F3,$BE,$06,$5F,$B1,$FA
	.dc.b	$3F,$05,$F4,$BE,$12,$4F,$C0,$7E,$AF,$C2,$7D,$6F,$83,$FA,$FF,$09
	.dc.b	$2B,$E2,$C7,$DA,$F8,$39,$1F,$21,$FB,$BF,$05,$D2,$3C,$C7,$DA,$E1
	.dc.b	$63,$9D,$7D,$F9,$77,$AF,$A3,$97,$77,$8F,$57,$E0,$73,$FC,$4F,$C0
	.dc.b	$FE,$2B,$E0,$A5,$7D,$88,$E9,$C5,$F8,$AC,$EC,$F8,$72,$EC,$F4,$D1
	.dc.b	$9F,$7D,$3E,$8E,$7F,$EC,$1F,$CA,$7F,$DA,$77,$E2,$7F,$DA,$6D,$F8
	.dc.b	$5D,$3B,$F8,$4E,$1F,$09,$23,$FC,$DF,$5F,$DA,$38,$0F,$F5,$8C,$7F
	.dc.b	$C1,$CD,$DB,$FD,$B0,$DA,$FD,$0B,$FE,$D0,$79,$FB,$DE,$5F,$76,$46
	.dc.b	$0E,$F9,$78,$5F,$46,$5E,$57,$C7,$EF,$FB,$48,$F9,$BE,$03,$E0,$7C
	.dc.b	$24,$8F,$4B,$F0,$FE,$13,$E2,$7C,$1C,$9F,$6B,$F1,$FE,$13,$E4,$7C
	.dc.b	$1C,$9F,$82,$FC,$BF,$83,$F9,$9F,$03,$27,$E2,$BF,$3F,$E0,$7A,$47
	.dc.b	$CF,$C2,$F1,$FC,$6C,$7A,$4F,$D3,$C6,$CF,$CA,$C7,$EC,$7C,$07,$D9
	.dc.b	$F8,$19,$7F,$4A,$3E,$E7,$C1,$7D,$DF,$83,$93,$F5,$A3,$F0,$1F,$05
	.dc.b	$23,$EC,$BF,$83,$F8,$34,$7F,$AA,$BF,$2A,$FE,$C2,$4E,$AC,$C7,$E6
	.dc.b	$23,$FD,$8C,$FE,$26,$FB,$BF,$8B,$C3,$C6,$F6,$21,$F8,$FC,$3C,$0F
	.dc.b	$7F,$81,$EB,$AF,$F9,$57,$0F,$ED,$33,$DF,$E0,$73,$97,$9B,$B5,$4C
	.dc.b	$3B,$DF,$84,$FC,$AE,$58,$C7,$F2,$79,$7C,$19,$F1,$1F,$9A,$D3,$AB
	.dc.b	$FC,$D3,$0F,$F6,$0E,$3D,$BE,$0F,$2D,$9F,$8D,$F0,$98,$AF,$F6,$7C
	.dc.b	$7C,$5F,$83,$C1,$9F,$83,$D7,$C0,$EF,$DF,$95,$F3,$E3,$97,$8F,$CF
	.dc.b	$DF,$93,$F4,$E3,$93,$F5,$70,$EA,$FD,$7C,$3A,$BF,$67,$0E,$AF,$D7
	.dc.b	$C3,$CD,$F0,$5E,$7F,$83,$93,$F7,$E3,$D5,$F0,$7E,$BF,$83,$93,$F8
	.dc.b	$38,$F7,$7C,$17,$BF,$E0,$E4,$FE,$1E,$3E,$17,$C0,$C8,$FC,$53,$F1
	.dc.b	$7E,$0F,$AC,$7E,$37,$0E,$AF,$E2,$F0,$EA,$FE,$47,$0E,$AF,$E3,$F0
	.dc.b	$F9,$DF,$05,$F4,$BF,$2B,$7C,$1D,$33,$5F,$9A,$EA,$FF,$4F,$4F,$FB
	.dc.b	$40,$FE,$E1,$CE,$69,$9F,$83,$E8,$4A,$E5,$DD,$F8,$F9,$7C,$1F,$87
	.dc.b	$E0,$7C,$5F,$03,$2F,$67,$F7,$CA,$FE,$BC,$FF,$3E,$7C,$5F,$13,$E9
	.dc.b	$3D,$5F,$13,$DE,$75,$95,$D3,$D3,$97,$BF,$E0,$FE,$07,$C1,$CB,$EF
	.dc.b	$B7,$AE,$79,$C6,$33,$F6,$E7,$88,$EA,$67,$2E,$DB,$7B,$72,$F9,$9F
	.dc.b	$09,$2B,$93,$F3,$FE,$17,$E8,$7C,$2C,$9E,$CF,$D3,$F8,$5F,$A9,$F0
	.dc.b	$3F,$57,$E0,$65,$78,$E3,$EC,$7C,$07,$D9,$F8,$19,$3E,$78,$E8,$FC
	.dc.b	$7C,$2F,$1F,$DA,$C7,$AC,$FB,$71,$EA,$FC,$DC,$7F,$09,$F0,$B2,$3E
	.dc.b	$13,$F8,$7F,$85,$FC,$47,$C2,$49,$F8,$CF,$E2,$FE,$17,$F7,$BE,$C7
	.dc.b	$F6,$4C,$FF,$DC,$CF,$F6,$46,$FC,$2E,$AF,$7F,$05,$C3,$E0,$A4,$7D
	.dc.b	$4B,$FE,$C9,$D0,$CB,$C1,$3F,$83,$7C,$0E,$CF,$0E,$11,$C1,$CF,$E0
	.dc.b	$BC,$BF,$05,$23,$07,$7F,$84,$F4,$91,$F1,$7C,$27,$B3,$E1,$24,$79
	.dc.b	$5F,$7F,$C2,$7C,$0F,$82,$F8,$3A,$CA,$F5,$47,$37,$FC,$2F,$1F,$EC
	.dc.b	$57,$C0,$76,$3A,$10,$DD,$76,$7E,$27,$C1,$7C,$E2,$3F,$1F,$E0,$BB
	.dc.b	$71,$7D,$0C,$EF,$7F,$F3,$FC,$7F,$DA,$37,$D6,$BE,$AE,$3E,$C8,$F2
	.dc.b	$E7,$F6,$BE,$12,$47,$D3,$7E,$EF,$C2,$7D,$EF,$83,$FB,$FF,$09,$2B
	.dc.b	$EC,$47,$E0,$BE,$13,$F0,$7F,$05,$27,$EE,$47,$E1,$BE,$0A,$47,$DF
	.dc.b	$7F,$13,$F0,$5D,$63,$F0,$58,$75,$7E,$4E,$1D,$5F,$C3,$61,$D5,$F9
	.dc.b	$B8,$7F,$4E,$FC,$B3,$FD,$A2,$4E,$0C,$7C,$46,$7E,$17,$5F,$0D,$30
	.dc.b	$FE,$07,$7F,$D8,$DB,$F0,$BC,$C8,$EB,$E1,$BB,$7C,$17,$7F,$83,$93
	.dc.b	$CA,$3B,$3D,$30,$EC,$FF,$96,$30,$EC,$DC,$3A,$B4,$EC,$79,$3E,$0B
	.dc.b	$D9,$F0,$72,$30,$7A,$3E,$13,$E0,$7C,$2C,$8C,$1E,$CF,$84,$F8,$9F
	.dc.b	$0B,$23,$07,$C0,$F8,$2F,$91,$F0,$72,$30,$72,$3D,$17,$D5,$F5,$61
	.dc.b	$F3,$F0,$E9,$1E,$DC,$30,$FC,$CF,$82,$FA,$84,$7E,$7F,$C1,$FD,$7B
	.dc.b	$E9,$E0,$39,$BF,$C2,$DC,$0C,$F5,$87,$A0,$EC,$7D,$8C,$BB,$BF,$37
	.dc.b	$E1,$7F,$04,$47,$EE,$FC,$2F,$E1,$7E,$16,$47,$E0,$1F,$F0,$AF,$E5
	.dc.b	$BF,$D9,$23,$FA,$2A,$F8,$3C,$FC,$04,$9C,$C7,$0F,$80,$E3,$F0,$12
	.dc.b	$78,$C7,$67,$F0,$18,$5D,$FD,$D1,$FE,$DA,$C7,$C0,$FF,$A7,$1F,$03
	.dc.b	$E2,$C7,$C0,$E7,$1F,$3F,$C2,$C8,$F3,$BE,$AF,$85,$F5,$FC,$1C,$9F
	.dc.b	$5B,$EE,$F8,$5F,$7F,$C2,$49,$F7,$BF,$0B,$E1,$3E,$1F,$C1,$49,$F8
	.dc.b	$6F,$C6,$F8,$2E,$91,$E7,$C2,$ED,$EC,$F8,$56,$EC,$FB,$5F,$B7,$3B
	.dc.b	$7B,$A6,$7F,$4B,$E0,$BE,$9F,$C1,$CC,$3E,$9C,$7D,$6F,$84,$FA,$FF
	.dc.b	$0B,$27,$EB,$C7,$DA,$F8,$59,$1F,$69,$FB,$BF,$0B,$F7,$BE,$06,$4F
	.dc.b	$DE,$7F,$03,$F0,$BD,$63,$F0,$38,$75,$7E,$A6,$1D,$1F,$AD,$85,$D7
	.dc.b	$D5,$8F,$F9,$7B,$F2,$EF,$EC,$5E,$0F,$F9,$BD,$FE,$C4,$DF,$83,$D6
	.dc.b	$1D,$CF,$C0,$65,$DF,$0F,$89,$97,$5B,$FF,$3C,$70,$33,$C6,$1F,$89
	.dc.b	$CF,$C1,$F0,$D2,$3F,$C3,$BA,$7F,$B3,$47,$C3,$79,$7E,$17,$CD,$F0
	.dc.b	$B2,$B7,$1E,$9F,$81,$91,$C9,$F6,$7C,$2F,$B7,$E0,$7D,$DF,$03,$2B
	.dc.b	$BC,$75,$7C,$38,$75,$7F,$DE,$F9,$7F,$B6,$7A,$CF,$F6,$38,$CE,$3E
	.dc.b	$07,$78,$FC,$BF,$85,$F9,$9F,$07,$27,$DB,$1F,$3F,$E1,$7E,$87,$C2
	.dc.b	$C9,$F8,$31,$F4,$FE,$16,$47,$C4,$7E,$B7,$C2,$FD,$7F,$82,$93,$F2
	.dc.b	$1F,$B5,$F0,$7D,$63,$E5,$E1,$D5,$F5,$E1,$D1,$F6,$E1,$7A,$BF,$45
	.dc.b	$B3,$2F,$C2,$7C,$17,$E1,$7E,$0E,$61,$F5,$A3,$F1,$1F,$0B,$23,$EC
	.dc.b	$BF,$8B,$F8,$5F,$F5,$4F,$95,$FC,$22,$59,$7E,$BE,$30,$DE,$ED,$FC
	.dc.b	$2F,$0F,$85,$E3,$F0,$53,$0F,$F3,$0E,$1F,$D4,$B0,$F3,$E5,$E1,$9F
	.dc.b	$AE,$F8,$5D,$E3,$E2,$23,$CB,$1F,$2F,$C1,$79,$BE,$16,$4F,$68,$E9
	.dc.b	$87,$68,$F6,$9F,$96,$D3,$9B,$EC,$C3,$FD,$AD,$97,$B1,$A7,$B2,$3D
	.dc.b	$3F,$09,$F1,$3E,$16,$46,$0F,$6F,$C2,$FC,$82,$3F,$03,$E1,$7E,$5F
	.dc.b	$C0,$7C,$CF,$7E,$4F,$C4,$8F,$9F,$F0,$72,$3E,$3B,$F4,$BE,$03,$9C
	.dc.b	$79,$C7,$84,$D5,$96,$B7,$FA,$D9,$F8,$2F,$E8,$E7,$E0,$79,$61,$C7
	.dc.b	$3F,$BF,$F0,$BF,$80,$F8,$59,$5F,$62,$3F,$07,$F0,$B2,$3E,$DB,$F8
	.dc.b	$6F,$85,$FC,$3F,$C2,$49,$FB,$EF,$E2,$BE,$13,$F1,$7F,$05,$FE,$F4
	.dc.b	$FC,$BF,$FA,$8E,$FF,$CE,$5B,$CD,$F4,$A7,$B6,$1E,$6E,$26,$78,$C3
	.dc.b	$F2,$59,$F7,$AD,$67,$67,$F2,$D9,$F7,$F8,$2F,$07,$C1,$CB,$FF,$59
	.dc.b	$EB,$FD,$81,$F8,$6B,$E1,$3C,$DF,$0B,$27,$71,$E9,$F8,$59,$1C,$9F
	.dc.b	$67,$C2,$FB,$7E,$13,$DD,$F0,$92,$BB,$C7,$C1,$BE,$91,$E0,$C3,$0F
	.dc.b	$32,$1E,$4C,$3A,$BE,$8F,$84,$F9,$5F,$09,$23,$07,$98,$F6,$D7,$6D
	.dc.b	$FE,$76,$1E,$89,$E8,$76,$8F,$A5,$9F,$D5,$E3,$93,$F1,$9F,$B1,$F0
	.dc.b	$BF,$67,$E1,$24,$FC,$A7,$EE,$7C,$2F,$DD,$F8,$19,$3F,$39,$FC,$07
	.dc.b	$C1,$7E,$07,$E0,$7F,$05,$F0,$32,$BE,$9C,$74,$7E,$9E,$17,$4F,$C3
	.dc.b	$63,$D2,$7E,$BE,$36,$7F,$15,$8F,$E3,$7E,$17,$F1,$DF,$0B,$2F,$EF
	.dc.b	$C1,$FF,$79,$FE,$61,$FD,$A1,$FD,$5B,$7C,$33,$4F,$ED,$71,$AF,$84
	.dc.b	$E4,$FE,$AA,$11,$8C,$7F,$C2,$91,$89,$EF,$F0,$99,$F0,$BC,$58,$76
	.dc.b	$F8,$4C,$57,$8B,$E0,$AC,$F8,$9B,$87,$F5,$B3,$C9,$38,$33,$F0,$DA
	.dc.b	$CE,$47,$B9,$E3,$97,$2F,$83,$E7,$F0,$72,$BE,$1C,$76,$F8,$4E,$FF
	.dc.b	$0B,$27,$E3,$C7,$8B,$E1,$3C,$7F,$0B,$27,$E5,$C7,$9B,$E0,$BC,$FF
	.dc.b	$07,$27,$E7,$C7,$57,$E8,$E1,$D5,$F9,$58,$75,$7E,$AE,$1D,$5F,$9D
	.dc.b	$87,$C2,$F8,$39,$1F,$69,$F8,$BF,$05,$D6,$3E,$A6,$1D,$79,$7E,$9E
	.dc.b	$3D,$67,$EC,$63,$D6,$7E,$BE,$3F,$3B,$E1,$7E,$7F,$C2,$CB,$FC,$3C
	.dc.b	$7D,$2F,$83,$91,$F8,$A7,$EA,$FC,$27,$D6,$F8,$29,$3F,$8E,$7E,$CF
	.dc.b	$BF,$F7,$3F,$31,$FE,$9D,$3F,$CE,$97,$C1,$E4,$D7,$6F,$69,$FC,$76
	.dc.b	$37,$5F,$C6,$E5,$DA,$7F,$25,$95,$9F,$C8,$E5,$DF,$E1,$65,$7E,$86
	.dc.b	$7F,$CD,$19,$FF,$73,$CB,$F2,$6B,$E1,$7C,$DF,$09,$E7,$F8,$59,$87
	.dc.b	$1B,$DC,$FA,$FE,$16,$4E,$4F,$63,$EF,$F8,$5F,$81,$F0,$72,$75,$E1
	.dc.b	$B3,$FE,$B3,$18,$5F,$07,$E2,$8C,$67,$E2,$7F,$AC,$CB,$2F,$0C,$63
	.dc.b	$2B,$AF,$BB,$3F,$A1,$F0,$5F,$47,$E0,$E5,$F7,$8F,$A9,$F0,$BF,$57
	.dc.b	$E1,$64,$F8,$E3,$EC,$7C,$24,$8F,$33,$F6,$FE,$17,$EE,$7C,$27,$DD
	.dc.b	$F8,$49,$5E,$B8,$EA,$FB,$70,$EA,$FA,$B0,$E8,$FB,$30,$BA,$7D,$6C
	.dc.b	$7F,$11,$F0,$92,$BE,$33,$F8,$BF,$84,$EB,$1F,$13,$1E,$B7,$F7,$B1
	.dc.b	$E7,$CD,$F7,$B2,$B7,$F9,$1C,$BF,$29,$F0,$BF,$C4,$37,$FE,$E7,$3F
	.dc.b	$E0,$E6,$F3,$7D,$79,$D7,$C2,$EF,$E1,$65,$FD,$CB,$FE,$D9,$A7,$FA
	.dc.b	$D5,$FE,$D9,$19,$F7,$BB,$1D,$FD,$FF,$07,$C1,$78,$7E,$02,$57,$38
	.dc.b	$EE,$F5,$C7,$BB,$FE,$1A,$C3,$BE,$FF,$CB,$1E,$F1,$9C,$7B,$BE,$5C
	.dc.b	$7D,$DF,$07,$EF,$F8,$49,$3E,$98,$F8,$5F,$09,$F0,$FE,$12,$4F,$B6
	.dc.b	$3E,$37,$C2,$48,$F8,$0F,$C9,$F8,$4F,$95,$F0,$5F,$2F,$E0,$A5,$7C
	.dc.b	$58,$EA,$FC,$7C,$3A,$BE,$BC,$3A,$3E,$DC,$2F,$17,$CB,$C7,$EB,$7C
	.dc.b	$17,$D7,$F8,$29,$7F,$46,$3E,$D7,$C1,$FD,$BF,$84,$93,$F5,$63,$EF
	.dc.b	$7C,$1F,$DF,$F8,$49,$3F,$66,$3F,$A8,$FE,$65,$FD,$8C,$91,$8F,$82
	.dc.b	$EA,$FD,$FC,$3A,$BF,$4F,$0E,$AF,$E0,$F0,$EA,$FD,$7C,$3A,$FC,$17
	.dc.b	$6F,$82,$93,$FC,$13,$2F,$ED,$0F,$33,$B1,$0C,$1C,$CE,$C4,$30,$7F
	.dc.b	$97,$9F,$F6,$11,$3E,$17,$DB,$F0,$BE,$EF,$83,$95,$E1,$7E,$0F,$C2
	.dc.b	$7C,$2F,$7E,$4F,$95,$F8,$BF,$01,$F1,$BD,$F9,$3E,$97,$E4,$FC,$07
	.dc.b	$48,$F1,$61,$79,$9E,$0F,$F8,$0B,$56,$7D,$AF,$D1,$9D,$BD,$C7,$FA
	.dc.b	$02,$99,$FD,$6F,$83,$FA,$FF,$09,$2B,$E5,$47,$DA,$F8,$4F,$B7,$F0
	.dc.b	$92,$7E,$74,$7D,$EF,$83,$91,$F4,$5F,$C0,$FC,$27,$E0,$BE,$0E,$4F
	.dc.b	$D5,$7F,$0B,$F0,$7D,$63,$EC,$61,$D5,$FA,$18,$74,$7E,$96,$17,$5F
	.dc.b	$5F,$9E,$D9,$9F,$E4,$3E,$0F,$F2,$3F,$07,$31,$FC,$24,$7F,$89,$3F
	.dc.b	$33,$FE,$D3,$81,$A4,$F8,$5D,$7C,$2C,$BF,$E9,$F3,$FD,$7D,$FF,$54
	.dc.b	$BF,$D7,$DC,$FC,$07,$5F,$82,$91,$83,$8F,$47,$84,$8F,$4C,$3B,$3F
	.dc.b	$EA,$CB,$ED,$8F,$97,$1F,$47,$C2,$FA,$7E,$0B,$D5,$F0,$53,$0F,$2C
	.dc.b	$74,$79,$65,$78,$7E,$3E,$5F,$E0,$69,$F5,$E3,$E0,$9C,$BD,$A3,$DF
	.dc.b	$F0,$BF,$1C,$8F,$C2,$F8,$5F,$95,$F0,$92,$3E,$2B,$F3,$7D,$7F,$9D
	.dc.b	$EF,$C9,$F9,$2F,$D1,$F8,$0F,$A5,$F0,$1F,$4F,$E0,$25,$7C,$E8,$E8
	.dc.b	$FC,$3C,$2E,$9E,$AC,$7A,$4F,$C7,$C6,$CF,$BB,$1F,$BD,$F0,$72,$BE
	.dc.b	$CB,$F8,$1F,$84,$FC,$17,$C2,$49,$FB,$AF,$E1,$7E,$13,$F0,$DF,$09
	.dc.b	$27,$F0,$2F,$F9,$B3,$F3,$4F,$EB,$67,$F4,$88,$FF,$75,$AF,$82,$D1
	.dc.b	$BF,$85,$ED,$3F,$77,$0B,$A7,$D4,$C7,$B4,$FE,$07,$1B,$3F,$9E,$C7
	.dc.b	$BF,$C1,$F8,$7E,$0E,$5F,$F8,$8A,$7F,$AB,$D6,$FE,$13,$CD,$F0,$B2
	.dc.b	$B5,$1E,$9F,$84,$F5,$7C,$2C,$9E,$51,$ED,$F8,$29,$1D,$5F,$81,$F0
	.dc.b	$7F,$06,$FA,$4B,$1F,$36,$39,$8E,$67,$97,$1E,$CE,$FB,$BF,$D3,$39
	.dc.b	$B3,$F9,$9F,$07,$F3,$7E,$0E,$5F,$B6,$3A,$3C,$4F,$9F,$8F,$8B,$2F
	.dc.b	$81,$1E,$03,$5E,$23,$16,$7E,$2A,$F9,$99,$7D,$BF,$84,$98,$7C,$B7
	.dc.b	$EF,$7C,$2F,$DF,$F8,$49,$3F,$3D,$FC,$17,$C2,$FE,$0F,$E0,$FF,$09
	.dc.b	$F0,$92,$BE,$A4,$7E,$1F,$E0,$FF,$11,$F0,$92,$7E,$C4,$74,$7C,$FB
	.dc.b	$31,$C6,$BF,$81,$CB,$B5,$7D,$FC,$BB,$3C,$32,$FF,$5F,$FB,$3F,$ED
	.dc.b	$13,$FD,$9F,$C0,$69,$9F,$84,$DF,$C2,$4B,$FF,$35,$CF,$F6,$8C,$BF
	.dc.b	$96,$AF,$84,$EB,$F0,$92,$B7,$1E,$0F,$82,$91,$C9,$F1,$FC,$17,$93
	.dc.b	$2E,$D3,$F5,$30,$65,$98,$E6,$78,$B1,$EC,$EF,$1F,$6F,$C1,$48,$F3
	.dc.b	$BF,$03,$E0,$BE,$0F,$C2,$7C,$2F,$84,$95,$EC,$8F,$8B,$97,$C6,$F8
	.dc.b	$49,$3F,$02,$3E,$4F,$C0,$FC,$AF,$82,$93,$F1,$23,$AB,$F1,$B0,$EA
	.dc.b	$FA,$F0,$EA,$FC,$AC,$3A,$BE,$FC,$3E,$AF,$C1,$7D,$6F,$81,$93,$F4
	.dc.b	$23,$AB,$F1,$70,$EB,$C7,$F4,$B1,$EB,$3F,$27,$1E,$B3,$F5,$B1,$FC
	.dc.b	$0F,$C2,$7E,$0B,$E1,$25,$FD,$C8,$FC,$2F,$C1,$7E,$1B,$E0,$E4,$FE
	.dc.b	$02,$0F,$FA,$B3,$F3,$5F,$EC,$A7,$F4,$77,$F5,$6D,$C7,$FB,$8B,$AF
	.dc.b	$77,$8C,$E9,$B3,$A6,$F9,$3D,$BA,$D7,$FD,$A6,$FF,$09,$87,$91,$FC
	.dc.b	$36,$3E,$87,$C9,$7C,$95,$F2,$F8,$2C,$36,$36,$7A,$FE,$13,$38,$BE
	.dc.b	$12,$98,$8C,$FC,$14,$AF,$84,$F0,$F8,$3E,$3F,$05,$27,$E3,$3D,$3E
	.dc.b	$0B,$AF,$C0,$76,$F8,$09,$5F,$2E,$3A,$3F,$2B,$0B,$C5,$F2,$B1,$E9
	.dc.b	$3F,$3B,$1B,$3F,$3B,$1F,$4F,$C0,$7A,$BE,$02,$5F,$D7,$8F,$6F,$C0
	.dc.b	$C8,$FB,$4F,$C0,$F8,$2F,$83,$F0,$5F,$0B,$E0,$A5,$7D,$F8,$F8,$BF
	.dc.b	$01,$F1,$BE,$02,$4F,$E0,$E3,$A3,$F8,$2C,$2E,$9F,$82,$C7,$AC,$FE
	.dc.b	$27,$1E,$AF,$E2,$31,$FA,$3F,$01,$23,$F1,$CF,$D8,$FC,$DB,$FD,$8B
	.dc.b	$CC,$CB,$AF,$4F,$E1,$E3,$17,$F9,$5C,$BB,$BF,$94,$CB,$BE,$7F,$94
	.dc.b	$CF,$A7,$C1,$4A,$FC,$EC,$FF,$92,$33,$FE,$F1,$D2,$7F,$BC,$4E,$B3
	.dc.b	$FC,$B4,$7F,$6C,$6F,$C0,$F9,$CF,$47,$C1,$7A,$7E,$03,$D5,$EF,$4C
	.dc.b	$39,$1F,$96,$C1,$3E,$FF,$76,$4E,$B2,$F1,$47,$C3,$F7,$BA,$CE,$6D
	.dc.b	$97,$86,$31,$97,$85,$D6,$5E,$17,$FB,$0D,$78,$1E,$5F,$05,$F3,$FE
	.dc.b	$0A,$46,$0E,$DF,$03,$F4,$FE,$0A,$46,$0F,$17,$C1,$7D,$7F,$82,$91
	.dc.b	$83,$CD,$EF,$FD,$BF,$7E,$46,$0E,$47,$92,$FA,$BE,$6C,$3F,$05,$87
	.dc.b	$48,$F4,$E1,$87,$E1,$7B,$DF,$87,$23,$F1,$7D,$FF,$C5,$7C,$0F,$E2
	.dc.b	$FE,$0A,$4F,$C9,$8F,$E0,$BF,$9B,$FF,$16,$8F,$F0,$2B,$7A,$7F,$D1
	.dc.b	$6D,$E9,$FE,$2C,$FF,$6A,$3F,$B2,$38,$F7,$F9,$9D,$3E,$03,$C3,$7F
	.dc.b	$5F,$1F,$0B,$F4,$B1,$F0,$B0,$FC,$06,$1D,$DF,$E3,$6F,$7F,$7F,$CF
	.dc.b	$F0,$12,$77,$1D,$9F,$BB,$7D,$B3,$FC,$E6,$1D,$67,$DA,$58,$F6,$FC
	.dc.b	$47,$C2,$F8,$39,$3E,$27,$E2,$FC,$1F,$C6,$F8,$3F,$8F,$F0,$72,$BC
	.dc.b	$F1,$F2,$BD,$EF,$97,$EF,$C9,$F5,$C7,$CE,$F7,$BE,$7F,$BF,$27,$DF
	.dc.b	$1D,$5F,$83,$87,$57,$E9,$E1,$D5,$F8,$B8,$75,$7E,$BE,$1F,$6B,$E0
	.dc.b	$BE,$DF,$C1,$C9,$F9,$71,$F7,$BE,$0A,$47,$CE,$7F,$03,$F0,$7F,$82
	.dc.b	$F8,$29,$3F,$49,$FC,$2F,$C1,$FE,$1B,$E0,$A4,$FD,$67,$F1,$3F,$05
	.dc.b	$D2,$3F,$13,$85,$E4,$FC,$4E,$3D,$27,$F1,$B8,$D9,$FC,$6E,$3F,$D5
	.dc.b	$BC,$2F,$F6,$89,$3B,$71,$F0,$59,$F8,$39,$8F,$F0,$5A,$FF,$61,$9F
	.dc.b	$D9,$F0,$FF,$70,$AF,$82,$E9,$F0,$72,$B8,$47,$7F,$81,$91,$CD,$F1
	.dc.b	$7C,$0F,$68,$FF,$33,$DF,$6C,$7C,$DF,$0F,$E7,$C3,$A5,$6B,$1C,$3E
	.dc.b	$5F,$82,$F6,$7C,$14,$8C,$1E,$9C,$3B,$69,$E8,$7A,$C7,$98,$F7,$5F
	.dc.b	$69,$F6,$E1,$F1,$FE,$0E,$57,$C3,$7E,$57,$C2,$7C,$BF,$83,$93,$F1
	.dc.b	$DF,$9D,$F0,$9F,$3F,$E0,$64,$FC,$B7,$E9,$7C,$07,$D3,$F7,$E4,$FC
	.dc.b	$F7,$EB,$7C,$07,$48,$F8,$B8,$5D,$3E,$8E,$3D,$27,$E4,$E3,$67,$EA
	.dc.b	$E3,$F7,$FE,$0F,$F0,$1F,$09,$2F,$EE,$47,$E0,$FE,$13,$FC,$33,$F9
	.dc.b	$C3,$E4,$69,$9A,$FE,$8D,$1F,$D6,$09,$F0,$BC,$3E,$23,$8F,$C0,$CA
	.dc.b	$DB,$D3,$E0,$B9,$47,$FA,$92,$FB,$3D,$30,$CE,$AE,$FF,$F1,$B5,$76
	.dc.b	$AF,$0E,$5D,$DF,$1F,$C3,$7A,$08,$F9,$BE,$0F,$D7,$F0,$BE,$CF,$85
	.dc.b	$93,$EA,$8F,$7F,$C2,$7C,$0F,$84,$93,$EE,$8F,$87,$F0,$52,$3E,$0B
	.dc.b	$F1,$BC,$3E,$B1,$F1,$30,$EA,$FB,$70,$EA,$FC,$8C,$3A,$BF,$07,$0F
	.dc.b	$9F,$F0,$5F,$43,$E0,$A4,$FC,$E8,$EA,$FC,$7C,$3A,$F0,$7C,$58,$FF
	.dc.b	$28,$EF,$FE,$D9,$7D,$07,$72,$1D,$CF,$B3,$F0,$9F,$78,$8F,$DC,$F8
	.dc.b	$4F,$C0,$FC,$24,$8F,$BE,$FE,$13,$E1,$3F,$0B,$EF,$7E,$1B,$DF,$95
	.dc.b	$F8,$48,$FF,$30,$FE,$73,$FD,$2F,$E1,$CF,$FB,$08,$93,$D4,$FC,$16
	.dc.b	$17,$4F,$65,$34,$65,$E0,$BF,$C8,$E5,$E0,$7F,$15,$97,$83,$E1,$64
	.dc.b	$7F,$86,$F3,$FE,$CE,$FF,$9D,$9F,$F5,$FB,$9F,$83,$F3,$91,$E1,$F0
	.dc.b	$9E,$AF,$83,$91,$CD,$F6,$FC,$1F,$BB,$E0,$24,$F7,$7E,$0F,$C0,$F3
	.dc.b	$8F,$F8,$03,$8F,$FB,$65,$F0,$1D,$8E,$84,$37,$7D,$9F,$57,$C0,$7C
	.dc.b	$C2,$3E,$DF,$81,$F9,$FF,$05,$23,$E0,$3F,$4B,$E0,$FE,$9F,$C1,$49
	.dc.b	$F8,$8F,$D6,$F8,$3F,$AF,$F0,$32,$7E,$43,$F6,$BE,$07,$A4,$7A,$B0
	.dc.b	$BD,$1F,$5B,$56,$5D,$6B,$DF,$95,$AD,$65,$F8,$4F,$81,$FC,$2F,$C0
	.dc.b	$CB,$FA,$D1,$D3,$C9,$FF,$5A,$64,$CF,$15,$F6,$73,$E9,$5A,$CE,$CF
	.dc.b	$DD,$CB,$F2,$3F,$07,$23,$F0,$4F,$E5,$3E,$0F,$FD,$6B,$F9,$D7,$EA
	.dc.b	$65,$B7,$E4,$23,$FD,$99,$8F,$FB,$15,$B3,$FE,$BE,$78,$4E,$FD,$EE
	.dc.b	$64,$9F,$F3,$54,$FF,$65,$AE,$5E,$EF,$80,$93,$AF,$7B,$C7,$3E,$4F
	.dc.b	$84,$93,$CA,$3A,$3E,$67,$C9,$86,$31,$F4,$7C,$17,$B3,$E1,$25,$F8
	.dc.b	$9F,$7F,$C2,$7C,$0F,$84,$93,$E6,$7E,$1F,$C2,$7C,$4F,$84,$93,$EA
	.dc.b	$7E,$3F,$C2,$7C,$8F,$82,$93,$EE,$7E,$5F,$C1,$74,$8F,$31,$F1,$F0
	.dc.b	$C6,$9F,$37,$1E,$B3,$AF,$2D,$C6,$CF,$37,$F4,$F3,$FA,$FF,$05,$30
	.dc.b	$F9,$AF,$DA,$F8,$2F,$B7,$F0,$92,$7E,$8B,$F7,$BE,$13,$EF,$FC,$1F
	.dc.b	$E0,$3E,$12,$57,$D6,$8F,$C1,$FC,$24,$8F,$B2,$FF,$BE,$B8,$BF,$AA
	.dc.b	$C7,$9F,$66,$3D,$E3,$0F,$77,$A1,$F8,$2C,$3B,$BC,$70,$E9,$F0,$12
	.dc.b	$3F,$D6,$18,$7F,$B0,$8F,$F6,$0B,$D0,$F5,$90,$F1,$65,$EC,$7B,$10
	.dc.b	$EC,$74,$F8,$3F,$57,$C1,$C8,$C1,$E0,$F8,$2F,$77,$C1,$48,$C1,$E4
	.dc.b	$F8,$0F,$85,$EE,$48,$C1,$E8,$F7,$FE,$37,$C0,$48,$C1,$C8,$FB,$DA
	.dc.b	$30,$ED,$18,$C7,$E7,$63,$D6,$37,$8E,$1F,$8B,$F0,$5F,$4F,$E0,$A4
	.dc.b	$60,$F9,$3F,$03,$F5,$C8,$FC,$CF,$82,$FB,$5F,$05,$23,$E7,$BF,$77
	.dc.b	$E0,$BE,$F7,$C1,$49,$FA,$6F,$E0,$7E,$07,$9C,$7A,$B0,$BD,$DF,$D5
	.dc.b	$B8,$19,$75,$AF,$7E,$56,$B5,$97,$E2,$FE,$02,$57,$E0,$1F,$C7,$7C
	.dc.b	$0A,$3F,$DE,$5F,$9D,$FF,$69,$23,$93,$F5,$0B,$E1,$B1,$87,$F6,$88
	.dc.b	$FE,$BF,$E2,$1A,$4F,$88,$DF,$9B,$DE,$75,$F0,$78,$C5,$1C,$7E,$0F
	.dc.b	$0D,$8F,$2C,$E3,$0F,$2C,$7A,$0C,$65,$9F,$84,$91,$EC,$78,$65,$C7
	.dc.b	$E0,$79,$7C,$24,$AF,$83,$1D,$58,$75,$3E,$2D,$F4,$7E,$1D,$DD,$7E
	.dc.b	$16,$1E,$5F,$84,$95,$F3,$1F,$47,$C2,$7A,$7E,$12,$4F,$D0,$7D,$9F
	.dc.b	$09,$ED,$F7,$FD,$DE,$FC,$AF,$AB,$1F,$07,$DF,$91,$F6,$1F,$89,$EF
	.dc.b	$F4,$8F,$AF,$85,$D3,$EB,$63,$D6,$7E,$FE,$3D,$5F,$BB,$8F,$CD,$F8
	.dc.b	$4F,$9D,$F0,$92,$7F,$0B,$1F,$47,$E0,$E4,$7E,$21,$FA,$9F,$09,$F5
	.dc.b	$7E,$0E,$4F,$E3,$1F,$B1,$F0,$7F,$73,$F3,$CF,$EB,$7D,$9F,$83,$7F
	.dc.b	$AF,$F5,$9A,$6A,$FF,$0B,$D4,$CB,$C3,$1F,$94,$7C,$2F,$72,$1F,$98
	.dc.b	$C7,$C2,$FE,$7F,$83,$FC,$7F,$6F,$81,$F2,$12,$75,$97,$97,$E0,$7D
	.dc.b	$1F,$09,$27,$27,$E6,$2F,$FD,$A4,$E3,$4F,$57,$C2,$7B,$FE,$0F,$E0
	.dc.b	$7C,$1C,$AD,$5A,$7E,$27,$C1,$49,$CB,$F1,$FE,$0B,$C1,$1F,$D6,$6B
	.dc.b	$BB,$72,$F9,$98,$F7,$8E,$B8,$F7,$78,$63,$F4,$7E,$0A,$47,$89,$FA
	.dc.b	$9F,$05,$DB,$83,$E2,$E3,$77,$FA,$D9,$75,$C3,$C5,$97,$5B,$F9,$79
	.dc.b	$7D,$DF,$32,$57,$B9,$FC,$07,$C1,$FE,$F3,$EC,$1E,$DF,$E0,$FE,$12
	.dc.b	$5F,$C9,$9C,$BF,$1B,$E1,$3F,$10,$47,$E4,$FC,$17,$E2,$FD,$F9,$1F
	.dc.b	$31,$FC,$77,$C0,$F7,$F6,$A4,$F5,$7C,$5E,$FF,$E4,$FE,$0A,$4F,$D4
	.dc.b	$7F,$2D,$F0,$9E,$7F,$73,$D1,$ED,$4A,$F3,$47,$AF,$DA,$91,$E9,$7D
	.dc.b	$DE,$D7,$BF,$DA,$93,$ED,$7E,$17,$B9,$F0,$FD,$CF,$89,$EE,$4A,$F8
	.dc.b	$51,$FC,$95,$7E,$36,$DF,$FC,$93,$0F,$CF,$1E,$D7,$F3,$7F,$C7,$BF
	.dc.b	$B0,$4F,$F3,$D9,$CB,$FC,$FE,$32,$F1,$9F,$E7,$D7,$D7,$B7,$84,$7F
	.dc.b	$37,$E8,$7B,$FF,$CD,$3F,$35,$FF,$57,$1F,$EA,$F8,$CE,$1F,$A1,$8F
	.dc.b	$E8,$CD,$F7,$78,$1C,$7D,$DF,$66,$5D,$E7,$C4,$79,$7E,$13,$FA,$86
	.dc.b	$BE,$88,$6B,$E8,$8C,$57,$17,$DB,$E6,$33,$F1,$DF,$F8,$FF,$3F,$3B
	.dc.b	$F9,$1D,$3D,$5E,$F4,$8F,$43,$ED,$F8,$0F,$2F,$0F,$CE,$CB,$C5,$A7
	.dc.b	$87,$83,$E7,$6D,$E6,$7E,$4D,$6D,$F9,$19,$7C,$7C,$A1,$F9,$D7,$E5
	.dc.b	$7C,$0C,$8F,$90,$FC,$D9,$F0,$F1,$7D,$29,$F8,$85,$3E,$1E,$1E,$17
	.dc.b	$FB,$D7,$07,$C3,$AF,$ED,$DB,$3E,$07,$EC,$12,$BE,$B7,$C0,$FD,$BF
	.dc.b	$81,$91,$F6,$5F,$BD,$EF,$F8,$A3,$EA,$9E,$03,$C5,$BF,$C3,$7C,$3B
	.dc.b	$FC,$17,$FB,$DE,$3E,$17,$F0,$98,$F8,$5F,$A9,$8F,$E2,$FD,$F9,$1F
	.dc.b	$89,$7F,$1D,$EF,$78,$78,$BE,$8E,$FE,$FA,$F1,$1E,$4D,$FE,$03,$E0
	.dc.b	$C7,$B7,$07,$FB,$F3,$3F,$14,$F8,$0F,$CB,$7B,$FF,$9B,$F8,$19,$18
	.dc.b	$3F,$9E,$9F,$B0,$DB,$D8,$F8,$33,$F1,$BF,$4F,$2F,$0F,$43,$17,$C0
	.dc.b	$FE,$9F,$E0,$25,$FE,$96,$3F,$55,$F0,$3F,$AB,$F7,$A4,$FE,$A6,$3E
	.dc.b	$07,$3F,$E3,$A3,$E7,$F0,$FC,$1C,$FB,$E5,$DA,$7F,$61,$97,$67,$F4
	.dc.b	$19,$7E,$DF,$E0,$64,$7E,$D5,$FD,$D7,$BF,$DA,$3F,$1A,$7D,$EE,$7F
	.dc.b	$B8,$FC,$1E,$3F,$C9,$6F,$ED,$C3,$B4,$7C,$DC,$7F,$82,$47,$F0,$5F
	.dc.b	$81,$97,$FC,$0D,$FF,$1D,$FE,$6B,$FD,$EE,$3A,$1D,$F6,$FB,$99,$78
	.dc.b	$36,$F9,$75,$FE,$04,$3F,$BB,$57,$F6,$D6,$BF,$EB,$B7,$5F,$01,$D8
	.dc.b	$91,$BF,$6F,$DC,$FD,$9E,$1F,$98,$FE,$27,$97,$F3,$1B,$B0,$F0,$7C
	.dc.b	$0F,$A0,$F4,$FC,$57,$AB,$DC,$98,$78,$DF,$6F,$B7,$DB,$6F,$92,$7A
	.dc.b	$F2,$EB,$8F,$AF,$0E,$78,$7A,$FD,$CF,$8B,$F0,$52,$3D,$CF,$C8,$F8
	.dc.b	$2E,$9C,$5F,$18,$FC,$16,$3D,$3E,$07,$E6,$92,$FE,$37,$C0,$FD,$0F
	.dc.b	$80,$FA,$3F,$01,$27,$E5,$47,$47,$F1,$A7,$E6,$EF,$A3,$F3,$EF,$A3
	.dc.b	$F3,$EF,$ED,$FB,$FF,$73,$DF,$93,$F5,$63,$A3,$F7,$4E,$9D,$9F,$8A
	.dc.b	$E0,$5F,$48,$7C,$0F,$FA,$16,$7F,$DE,$95,$EA,$3F,$01,$F0,$1F,$8A
	.dc.b	$F8,$09,$18,$3F,$6F,$D9,$FC,$09,$FB,$C7,$E7,$3C,$5F,$E8,$FB,$7D
	.dc.b	$2F,$80,$FC,$9F,$BF,$2B,$F1,$91,$F9,$6F,$7F,$FD,$23,$E0,$7F,$BC
	.dc.b	$B6,$F2,$D7,$FA,$5A,$31,$B7,$9E,$BD,$F9,$7F,$04,$AF,$C6,$5F,$BF
	.dc.b	$0E,$D1,$F9,$7C,$7F,$47,$1D,$BE,$26,$4F,$F9,$27,$24,$78,$BE,$3B
	.dc.b	$AF,$83,$F9,$8C,$BD,$79,$75,$C3,$F4,$99,$76,$8F,$3F,$83,$F9,$BC
	.dc.b	$FD,$7F,$07,$2F,$B3,$EE,$F8,$AE,$B1,$F1,$F2,$FE,$B7,$87,$F0,$0D
	.dc.b	$5F,$13,$F1,$3B,$25,$F9,$E3,$E3,$FC,$4F,$C8,$E3,$93,$EB,$8E,$AF
	.dc.b	$D4,$3F,$80,$EF,$FD,$E3,$0E,$B1,$F0,$70,$EA,$B1,$F5,$CE,$A3,$1C
	.dc.b	$7F,$EC,$B8,$7C,$37,$D8,$F8,$A9,$5F,$2E,$3C,$3E,$27,$EC,$F0,$F6
	.dc.b	$6B,$D9,$EB,$97,$E0,$23,$F0,$3F,$15,$30,$FA,$8F,$E1,$3E,$1F,$9D
	.dc.b	$7E,$CA,$FE,$37,$C4,$7E,$20,$97,$F6,$FE,$23,$F1,$7F,$11,$23,$EF
	.dc.b	$3F,$8E,$F8,$8E,$71,$FA,$63,$FC,$47,$D0,$E0,$61,$D2,$7E,$09,$F9
	.dc.b	$37,$F2,$F8,$7E,$63,$E2,$3F,$33,$F0,$F2,$BF,$19,$1D,$9F,$CA,$9D
	.dc.b	$AF,$AB,$F9,$1B,$ED,$9F,$D0,$C3,$F4,$7F,$11,$2B,$F2,$CF,$E9,$BE
	.dc.b	$0F,$A7,$37,$E5,$27,$D3,$F9,$A9,$FF,$7D,$B5,$FF,$53,$35,$F5,$EA
	.dc.b	$8F,$F2,$4C,$61,$7C,$47,$22,$47,$2F,$87,$F0,$F6,$7E,$62,$3C,$27
	.dc.b	$80,$87,$83,$E2,$7C,$7F,$11,$23,$C6,$F9,$BE,$23,$9C,$7F,$87,$37
	.dc.b	$FE,$C3,$97,$91,$FD,$6B,$E9,$6C,$7B,$AF,$DF,$84,$8F,$7B,$F0,$BE
	.dc.b	$0F,$AC,$7E,$8A,$3E,$36,$9F,$07,$E1,$BE,$3F,$C4,$4A,$F8,$F1,$F2
	.dc.b	$BE,$23,$E5,$FC,$44,$9F,$97,$1D,$3C,$CF,$F0,$9E,$AB,$E5,$5F,$2F
	.dc.b	$0F,$A7,$1F,$53,$E2,$3E,$AF,$C4,$4A,$FA,$B1,$CF,$C7,$FE,$CB,$87
	.dc.b	$39,$FB,$38,$74,$E0,$F8,$78,$FD,$EF,$88,$FB,$FF,$05,$2F,$EF,$C7
	.dc.b	$3A,$F9,$7F,$07,$F8,$4F,$87,$95,$F8,$48,$FC,$3F,$C4,$7E,$23,$E1
	.dc.b	$E4,$FE,$22,$39,$3F,$8E,$9E,$2F,$E1,$EB,$F1,$F1,$F9,$0F,$87,$FC
	.dc.b	$8F,$C3,$CA,$FC,$8C,$72,$ED,$FF,$44,$EE,$BE,$79,$FD,$EC,$79,$E1
	.dc.b	$FE,$FD,$FC,$D7,$FA,$1B,$F9,$85,$F0,$9F,$9D,$AF,$E6,$31,$89,$E1
	.dc.b	$E8,$77,$EA,$FB,$F3,$FE,$1A,$AD,$FC,$3F,$1A,$CB,$6D,$D6,$BF,$E0
	.dc.b	$91,$E3,$F8,$8D,$4E,$7E,$22,$D1,$B3,$E1,$AE,$73,$E1,$C3,$BC,$7B
	.dc.b	$70,$E3,$E4,$72,$F8,$8E,$7F,$11,$2B,$E1,$47,$3C,$FC,$97,$E0,$24
	.dc.b	$FC,$7B,$F1,$FC,$47,$93,$E2,$24,$FC,$B8,$E5,$9F,$F8,$02,$3E,$67
	.dc.b	$C3,$7A,$8F,$5F,$C2,$7B,$3E,$1E,$4F,$D3,$7D,$FF,$11,$CA,$B8,$4F
	.dc.b	$D7,$78,$BF,$67,$E1,$FE,$29,$1F,$B9,$F1,$1C,$5E,$04,$3E,$F6,$1F
	.dc.b	$33,$E1,$FE,$6F,$C3,$C9,$FC,$24,$7A,$BE,$0F,$E8,$FC,$24,$9F,$C4
	.dc.b	$47,$D4,$F8,$7F,$AF,$F8,$27,$F8,$2A,$7E,$EB,$79,$3E,$F4,$F3,$9E
	.dc.b	$24,$3F,$93,$CF,$F8,$19,$FE,$5D,$D5,$FE,$00,$87,$C3,$F6,$24,$EB
	.dc.b	$A7,$F8,$7C,$BE,$1F,$B4,$FE,$6F,$2F,$C4,$9E,$6F,$88,$F3,$FC,$47
	.dc.b	$A3,$E1,$25,$73,$3D,$31,$F8,$3C,$1F,$05,$EE,$F8,$89,$5B,$DD,$5A
	.dc.b	$8C,$2F,$43,$E2,$FC,$3C,$9C,$C7,$5C,$3E,$49,$23,$8F,$C4,$7C,$CF
	.dc.b	$87,$91,$D1,$F9,$FF,$0F,$C6,$3B,$D7,$17,$E5,$D7,$D4,$F8,$39,$1E
	.dc.b	$47,$EB,$FC,$1F,$B7,$E1,$FE,$CF,$C3,$CA,$F4,$C7,$DC,$F8,$69,$1E
	.dc.b	$C7,$EF,$FC,$3F,$18,$F7,$D7,$17,$ED,$D7,$E1,$3E,$1F,$F0,$BF,$0F
	.dc.b	$27,$E2,$C7,$17,$67,$80,$F9,$9F,$0B,$FD,$4F,$AB,$FA,$5B,$FD,$5D
	.dc.b	$BF,$0F,$FC,$1A,$3F,$B5,$CF,$0F,$85,$FD,$D4,$3E,$1F,$44,$9E,$B1
	.dc.b	$C7,$E1,$F9,$7A,$92,$7C,$31,$CA,$B8,$47,$F7,$CA,$E8,$F2,$3C,$FF
	.dc.b	$0F,$E3,$23,$EA,$F8,$1E,$4F,$B2,$B9,$F8,$7F,$E5,$CB,$1E,$BF,$82
	.dc.b	$95,$F0,$9F,$77,$C3,$7B,$3E,$1F,$E0,$12,$7E,$3F,$C2,$FC,$3F,$86
	.dc.b	$F8,$9F,$0B,$27,$E5,$C7,$C4,$9E,$11,$EB,$AF,$95,$1F,$2F,$E1,$A4
	.dc.b	$FD,$27,$E7,$7C,$07,$18,$FA,$B5,$CB,$4F,$55,$FD,$3F,$80,$95,$F6
	.dc.b	$9F,$AD,$F0,$BF,$53,$E1,$BE,$C1,$27,$EF,$FC,$2F,$DB,$F8,$69,$1F
	.dc.b	$82,$7F,$A1,$FE,$67,$FE,$EF,$F8,$8F,$F5,$27,$C4,$FF,$B5,$3C,$6F
	.dc.b	$F2,$C5,$FF,$90,$BE,$25,$D7,$FE,$35,$3F,$0B,$08,$CB,$FE,$3B,$8C
	.dc.b	$C7,$F8,$FC,$CF,$9A,$C7,$FB,$05,$7F,$8F,$1F,$E9,$CF,$F5,$A6,$C6
	.dc.b	$FE,$23,$FA,$A4,$FF,$88,$CE,$DD,$FF,$E4,$7B,$F0,$C7,$9A,$3D,$31
	.dc.b	$E0,$BF,$EC,$73,$DF,$E1,$FE,$09,$F0,$BE,$13,$AE,$BF,$EC,$5C,$FD
	.dc.b	$27,$BF,$1E,$6F,$F8,$CB,$0F,$ED,$18,$7F,$17,$6B,$EB,$F8,$8F,$9B
	.dc.b	$F0,$FD,$3D,$0F,$F4,$B4,$FC,$7B,$E8,$FF,$98,$74,$F4,$C7,$82,$3F
	.dc.b	$C9,$87,$D1,$7F,$CC,$99,$3D,$3F,$81,$18,$7C,$B3,$C2,$BF,$D9,$C4
	.dc.b	$9F,$F3,$03,$F4,$DF,$0C,$78,$CF,$96,$7F,$A3,$63,$4F,$F9,$D6,$BC
	.dc.b	$0F,$9E,$7C,$0E,$DE,$A7,$FB,$44,$8F,$8D,$CB,$D8,$F0,$1F,$5C,$E4
	.dc.b	$6C,$F0,$57,$B9,$A6,$8F,$C1,$1F,$EC,$7E,$1F,$F5,$C3,$F1,$CE,$D7
	.dc.b	$F3,$34,$F8,$B1,$C3,$0F,$5B,$E2,$6C,$FC,$08,$EB,$5F,$A3,$7C,$13
	.dc.b	$F8,$5A,$F2,$47,$A4,$F0,$1D,$6F,$33,$D3,$0F,$83,$3D,$38,$78,$E1
	.dc.b	$E8,$C7,$C1,$5F,$81,$AE,$91,$E3,$BF,$EC,$1D,$1F,$46,$7E,$19,$E9
	.dc.b	$C7,$DB,$3F,$BA,$C7,$A9,$8C,$FF,$23,$87,$C3,$BE,$47,$6C,$7C,$B7
	.dc.b	$DA,$3A,$1A,$7D,$4F,$D5,$3D,$47,$2C,$BD,$53,$EE,$3D,$47,$AE,$7A
	.dc.b	$57,$AF,$1F,$8A,$79,$72,$E5,$97,$A3,$1F,$9F,$97,$A3,$1F,$CB,$E1
	.dc.b	$E9,$9F,$9F,$8F,$49,$FC,$9C,$7A,$0C,$D7,$17,$FD,$9D,$97,$DB,$73
	.dc.b	$8F,$C2,$AF,$3E,$7E,$39,$E5,$8F,$F0,$CC,$7E,$0C,$7E,$27,$1E,$8F
	.dc.b	$E4,$72,$E1,$5E,$52,$61,$F3,$2B,$77,$C8,$8E,$0E,$79,$79,$6B,$F0
	.dc.b	$11,$DA,$BC,$B8,$7C,$4C,$BE,$A6,$7F,$2A,$30,$7E,$B3,$0E,$55,$E1
	.dc.b	$AF,$76,$7F,$98,$8F,$A9,$87,$86,$79,$DC,$3E,$29,$CF,$3F,$A8,$E3
	.dc.b	$2F,$6B,$FA,$EB,$EC,$FC,$87,$E9,$C7,$AA,$BE,$EC,$C3,$94,$FD,$C2
	.dc.b	$3F,$8D,$CF,$EE,$3F,$9E,$AF,$A5,$9F,$CE,$AF,$E1,$B2,$CF,$F2,$2F
	.dc.b	$E5,$63,$E3,$BD,$8E,$4F,$90,$F3,$D7,$F2,$58,$FC,$2E,$3D,$5F,$61
	.dc.b	$0F,$C0,$63,$F1,$67,$C7,$1F,$41,$E4,$F8,$E7,$93,$FA,$D3,$96,$3E
	.dc.b	$C7,$F8,$31,$CE,$FC,$67,$B3,$0F,$5C,$7F,$1E,$8F,$5D,$FC,$B8,$F9
	.dc.b	$C7,$B0,$F9,$19,$7E,$2B,$0F,$AF,$3B,$3E,$BC,$F8,$CF,$59,$C8,$61
	.dc.b	$F9,$4C,$7F,$97,$63,$F8,$A9,$EB,$1F,$8A,$61,$F3,$4E,$39,$7E,$4D
	.dc.b	$C4,$F1,$7B,$DF,$8B,$2F,$C8,$DF,$F2,$CB,$E1,$3E,$79,$F8,$58,$FC
	.dc.b	$A7,$E0,$9F,$1A,$FE,$01,$E0,$3E,$93,$C1,$A7,$D9,$7B,$BF,$C2,$2B
	.dc.b	$BC,$73,$8F,$DD,$1C,$0F,$8C,$F0,$AF,$E3,$11,$F5,$CF,$86,$7E,$94
	.dc.b	$FC,$E3,$E9,$7C,$19,$FE,$0B,$2F,$E0,$17,$F8,$4C,$BE,$24,$7E,$0C
	.dc.b	$FE,$99,$5F,$32,$F9,$3E,$D3,$EB,$46,$F3,$FB,$D1,$E0,$C7,$D4,$7C
	.dc.b	$67,$CB,$5F,$96,$78,$CF,$EF,$DF,$41,$23,$BC,$F1,$8F,$AE,$FE,$9D
	.dc.b	$F9,$F5,$E1,$CF,$C1,$1F,$A8,$3E,$44,$F1,$9D,$47,$A3,$1F,$81,$9F
	.dc.b	$E5,$2B,$CF,$9F,$F1,$33,$E3,$4F,$D4,$AF,$4C,$FD,$B7,$F1,$CF,$5C
	.dc.b	$BD,$59,$7B,$49,$1F,$6F,$3F,$9B,$3A,$8F,$E1,$B5,$FA,$1C,$BC,$B7
	.dc.b	$C2,$3D,$B5,$B7,$D8,$7C,$5A,$F3,$BF,$55,$FC,$45,$7D,$C8,$FC,$45
	.dc.b	$F8,$33,$FF,$0B,$57,$9D,$FC,$F4,$FE,$24,$F9,$A7,$6C,$3F,$84,$DF
	.dc.b	$92,$BC,$E7,$E4,$4F,$EE,$B8,$FC,$5A,$E2,$7E,$60,$F8,$B9,$7C,$B7
	.dc.b	$F3,$55,$E5,$7F,$67,$9F,$D1,$7E,$69,$F9,$63,$75,$D6,$7C,$51,$E5
	.dc.b	$BF,$E5,$F1,$88,$FD,$2D,$FD,$87,$D9,$19,$CB,$E1,$47,$F8,$92,$BE
	.dc.b	$13,$FB,$13,$79,$FC,$F7,$D5,$97,$E4,$B0,$F5,$47,$F0,$DC,$3D,$53
	.dc.b	$FA,$4C,$3A,$BF,$E2,$8A,$DB,$FA,$89,$DE,$1F,$83,$CB,$E0,$61,$F2
	.dc.b	$23,$C7,$5F,$BC,$8F,$D2,$E7,$FC,$82,$B7,$7E,$FB,$D9,$F4,$AB,$D3
	.dc.b	$5B,$7F,$46,$75,$C3,$C7,$3F,$86,$CF,$E6,$67,$EF,$CF,$D3,$3F,$E2
	.dc.b	$7A,$DD,$FD,$43,$F8,$06,$7F,$95,$7E,$FD,$70,$7F,$AF,$4F,$07,$F0
	.dc.b	$79,$7C,$07,$94,$ED,$FE,$95,$3B,$9F,$B4,$6E,$3A,$E7,$F0,$31,$FA
	.dc.b	$98,$FC,$09,$FE,$E9,$97,$AB,$3F,$81,$1F,$A5,$3D,$13,$B7,$FA,$E9
	.dc.b	$F3,$0F,$F3,$2E,$5E,$3C,$7F,$8F,$C7,$FA,$0A,$7F,$3C,$F6,$3E,$69
	.dc.b	$C3,$1F,$25,$FE,$B3,$2F,$56,$5F,$89,$C3,$D4,$7E,$46,$B7,$7F,$91
	.dc.b	$AD,$DF,$EC,$A3,$C9,$9F,$F9,$FF,$2F,$26,$3F,$E2,$FB,$DD,$7E,$5C
	.dc.b	$DE,$5F,$E7,$AA,$DE,$1F,$C0,$E7,$FB,$ED,$F8,$F3,$F8,$D1,$E9,$3F
	.dc.b	$0B,$9F,$E2,$F1,$EB,$97,$E4,$4F,$BD,$1E,$99,$DD,$FD,$2A,$DE,$5F
	.dc.b	$91,$9F,$19,$FE,$73,$CB,$C6,$FC,$B9,$DB,$F2,$CD,$9F,$E7,$83,$EC
	.dc.b	$1F,$97,$3D,$47,$E2,$B1,$F4,$E3,$F1,$73,$F9,$6F,$EB,$E7,$79,$7E
	.dc.b	$2A,$B7,$8F,$C5,$BD,$BF,$9D,$AD,$E5,$F1,$5F,$BB,$87,$57,$DF,$9F
	.dc.b	$DD,$76,$FD,$78,$F8,$B3,$FA,$9B,$EB,$5F,$A2,$36,$7D,$79,$DD,$7E
	.dc.b	$59,$FE,$5B,$5F,$2E,$77,$1F,$46,$FF,$AF,$47,$E8,$8F,$19,$F8,$93
	.dc.b	$6F,$D1,$8E,$B1,$F7,$4F,$97,$8F,$DD,$76,$7D,$18,$D9,$EF,$77,$5E
	.dc.b	$37,$EC,$1F,$CC,$8F,$8B,$5F,$2D,$F1,$CF,$D1,$76,$7F,$25,$C7,$E8
	.dc.b	$BF,$83,$C7,$E2,$BF,$C0,$0F,$19,$FB,$49,$F1,$D7,$58,$F1,$E3,$F4
	.dc.b	$5E,$B3,$FC,$AE,$FA,$C7,$E2,$63,$E5,$CE,$DF,$F2,$6E,$3F,$83,$9F
	.dc.b	$49,$E3,$3D,$37,$F8,$37,$AB,$FC,$95,$D9,$D5,$FE,$9F,$8F,$C5,$AE
	.dc.b	$AF,$E2,$67,$F3,$75,$FE,$DF,$7A,$9F,$83,$3A,$9F,$CA,$DF,$E0,$B1
	.dc.b	$F8,$98,$D9,$F1,$4D,$9F,$A1,$3C,$71,$D5,$F8,$A7,$E0,$CD,$BE,$F7
	.dc.b	$C7,$1F,$E3,$47,$73,$F5,$CF,$4D,$FE,$6E,$FF,$1F,$7F,$BF,$AD,$C7
	.dc.b	$C5,$AD,$C7,$E6,$E3,$77,$E9,$8D,$9E,$F9,$D9,$F9,$B7,$FC,$37,$7F
	.dc.b	$A1,$AF,$96,$FD,$77,$FC,$47,$7E,$9A,$FC,$1D,$7D,$73,$A9,$F1,$5E
	.dc.b	$B3,$FA,$13,$C7,$5D,$67,$FC,$0C,$7F,$42,$CB,$F8,$8B,$F8,$F9,$DB
	.dc.b	$FA,$7A,$FC,$4B,$F2,$DE,$AF,$E9,$E3,$71,$FA,$73,$67,$CB,$7D,$38
	.dc.b	$7E,$0E,$3D,$33,$F8,$93,$C6,$7F,$1D,$3E,$BC,$6E,$7A,$9F,$A1,$9E
	.dc.b	$AF,$D1,$3A,$DF,$D7,$AF,$96,$7A,$63,$E5,$BE,$97,$6F,$D7,$37,$3E
	.dc.b	$98,$DB,$E9,$8F,$E8,$F3,$D5,$FE,$FF,$3F,$E1,$77,$F2,$B8,$7F,$66
	.dc.b	$AF,$8B,$1F,$2C,$EB,$3E,$93,$A9,$E9,$7A,$9E,$33,$A9,$FE,$04,$CB
	.dc.b	$F8,$2D,$75,$9F,$4D,$7E,$26,$B6,$7F,$23,$3F,$07,$3B,$7A,$C6,$E3
	.dc.b	$FD,$91,$1F,$14,$FA,$2F,$F2,$37,$E8,$DF,$BE,$37,$3F,$16,$77,$87
	.dc.b	$EF,$EF,$C6,$7E,$3E,$3D,$E7,$E5,$67,$EB,$BF,$A1,$8D,$BF,$89,$36
	.dc.b	$7E,$0C,$FF,$7A,$1F,$E9,$6C,$7F,$2B,$7F,$EB,$1C,$BF,$8E,$BF,$BF
	.dc.b	$BE,$B5,$FA,$1B,$EA,$FD,$13,$E2,$CF,$EE,$67,$6F,$C5,$8F,$F4,$76
	.dc.b	$1F,$8F,$7C,$71,$D6,$BE,$BC,$7A,$4F,$96,$FB,$DE,$B3,$FB,$FA,$EA
	.dc.b	$FF,$85,$63,$F9,$DD,$7A,$5F,$E2,$F1,$FC,$46,$77,$8F,$F7,$CB,$FF
	.dc.b	$16,$D7,$A6,$FE,$8C,$7E,$26,$3E,$8C,$FF,$05,$7C,$67,$CB,$3C,$6F
	.dc.b	$EF,$DF,$96,$7F,$A4,$0F,$1D,$A3,$FD,$07,$87,$FB,$02,$7F,$37,$3E
	.dc.b	$97,$F0,$67,$E5,$6F,$C6,$FD,$DB,$F8,$AF,$D1,$3C,$67,$A6,$BA,$3F
	.dc.b	$94,$AE,$98,$7E,$9A,$FE,$24,$FF,$34,$C3,$FC,$EB,$9B,$34,$FE,$0C
	.dc.b	$F4,$CF,$A1,$FF,$5D,$C7,$E9,$B0,$EB,$97,$E0,$B0,$EB,$8F,$F1,$07
	.dc.b	$EE,$5F,$CA,$3F,$82,$1E,$87,$73,$FC,$42,$3D,$C6,$E7,$F8,$23,$F8
	.dc.b	$8C,$3F,$4D,$97,$BA,$7E,$DC,$FF,$56,$AF,$B7,$7D,$23,$F1,$B8,$7E
	.dc.b	$06,$7E,$4C,$FE,$4D,$DC,$7D,$59,$FF,$7A,$1E,$E3,$EA,$E1,$F8,$7C
	.dc.b	$FF,$D7,$91,$F2,$4F,$F2,$3E,$7F,$25,$F9,$07,$DA,$C7,$F8,$16,$5E
	.dc.b	$7C,$7C,$38,$FC,$8A,$F8,$4F,$86,$BF,$CB,$F5,$B9,$F8,$55,$FC,$65
	.dc.b	$F8,$47,$E1,$63,$4F,$E4,$27,$FD,$8C,$FD,$83,$D6,$7F,$B7,$8F,$59
	.dc.b	$F1,$F0,$F5,$C7,$D7,$BF,$8D,$3E,$03,$F6,$47,$82,$3D,$6F,$C5,$3F
	.dc.b	$1D,$8F,$C5,$7F,$C6,$C7,$C5,$86,$1F,$8C,$8F,$F6,$DB,$72,$FE,$45
	.dc.b	$3F,$95,$BF,$85,$3F,$E0,$2C,$7D,$B9,$7C,$DC,$3F,$58,$7C,$D9,$FE
	.dc.b	$09,$87,$E9,$A3,$F8,$E4,$EF,$0F,$87,$1B,$3F,$71,$3B,$3F,$2A,$FC
	.dc.b	$9A,$F2,$61,$FC,$BE,$3E,$AE,$3F,$A1,$7E,$51,$F0,$0F,$94,$7C,$0B
	.dc.b	$F9,$51,$FA,$CC,$FE,$04,$78,$8F,$74,$73,$7F,$DF,$11,$FA,$B8,$F9
	.dc.b	$53,$FE,$59,$3F,$8C,$BC,$DF,$C6,$CF,$27,$EE,$BC,$9F,$BA,$FD,$07
	.dc.b	$F7,$77,$FE,$D1,$C3,$F4,$91,$FC,$60,$FE,$A8,$7C,$13,$FC,$2C,$7E
	.dc.b	$26,$7F,$26,$FD,$F7,$F5,$E7,$57,$EE,$E7,$FD,$23,$1F,$B3,$1F,$C2
	.dc.b	$AB,$FC,$9B,$87,$F2,$5C,$BE,$A6,$3F,$3A,$3F,$96,$57,$E3,$63,$ED
	.dc.b	$E7,$F8,$0C,$FE,$4C,$6E,$7F,$84,$9B,$3E,$DC,$ED,$F8,$B1,$FC,$4F
	.dc.b	$1F,$1C,$7F,$54,$9D,$3F,$8A,$CF,$E5,$1E,$E7,$EE,$BE,$EA,$FC,$C1
	.dc.b	$F2,$6B,$9E,$5E,$CA,$F9,$39,$7B,$1F,$92,$7E,$3C,$FE,$68,$7C,$9A
	.dc.b	$FE,$03,$97,$DE,$C3,$FD,$15,$3F,$78,$FE,$17,$97,$C0,$79,$C7,$9A
	.dc.b	$BF,$49,$1C,$CF,$DB,$1F,$D0,$CF,$88,$78,$0F,$5B,$E0,$7F,$2E,$7E
	.dc.b	$34,$FD,$B3,$FB,$5B,$FF,$73,$CF,$DD,$8F,$81,$7F,$97,$7F,$C6,$55
	.dc.b	$FA,$19,$D6,$7F,$81,$AD,$4F,$F8,$AC,$FC,$9D,$7C,$A9,$D4,$7E,$62
	.dc.b	$39,$D7,$CF,$8F,$CB,$CE,$A3,$EA,$1F,$9D,$8F,$04,$F3,$7F,$2C,$FC
	.dc.b	$47,$DE,$7C,$4C,$7F,$9F,$57,$F1,$E9,$FE,$9D,$3F,$C1,$B1,$F9,$CF
	.dc.b	$F9,$8E,$FE,$73,$F9,$9C,$FF,$C4,$39,$FF,$24,$8F,$D3,$1B,$BF,$C7
	.dc.b	$65,$F0,$23,$FC,$89,$7E,$B7,$F4,$A7,$98,$E8,$7D,$F7,$F5,$59,$FD
	.dc.b	$0A,$F6,$4F,$E2,$73,$FD,$B9,$FC,$5F,$0F,$15,$7E,$3E,$7C,$47,$F1
	.dc.b	$63,$E6,$1F,$44,$F8,$E7,$CC,$3C,$47,$E8,$CF,$C5,$1E,$C9,$FE,$67
	.dc.b	$5E,$C8,$E6,$7B,$1F,$C5,$5F,$E0,$23,$9C,$FE,$8E,$3F,$CB,$11,$CE
	.dc.b	$FF,$95,$3F,$94,$9F,$5E,$3F,$B2,$AF,$E4,$97,$DD,$FB,$EF,$E1,$0E
	.dc.b	$E7,$C4,$7E,$AD,$7A,$4F,$BE,$77,$C7,$F8,$5C,$FF,$7C,$7E,$A5,$FF
	.dc.b	$94,$E3,$9C,$FC,$33,$E7,$61,$FB,$CC,$3E,$FD,$FF,$2F,$9F,$E0,$D5
	.dc.b	$E5,$3F,$1D,$87,$E4,$CF,$BC,$FE,$9A,$3E,$3E,$1D,$DF,$C9,$61,$FD
	.dc.b	$FD,$F8,$4F,$77,$F0,$53,$FB,$1C,$BF,$A1,$47,$ED,$AF,$F8,$2B,$FB
	.dc.b	$63,$FD,$73,$97,$E0,$70,$FA,$F1,$F8,$19,$F9,$2F,$E0,$5F,$2E,$7F
	.dc.b	$AB,$C3,$D2,$7C,$D9,$F2,$E5,$FC,$4E,$3F,$97,$57,$F7,$E9,$F9,$B8
	.dc.b	$77,$CB,$ED,$D7,$CD,$9F,$7C,$7F,$9C,$E3,$E6,$C7,$E8,$0F,$9B,$7F
	.dc.b	$4A,$75,$1E,$03,$DD,$3E,$03,$FA,$0B,$F9,$4A,$F0,$57,$DC,$C3,$F7
	.dc.b	$53,$FE,$3C,$8F,$AD,$87,$E2,$A3,$E1,$5F,$B9,$F8,$4F,$90,$FA,$D1
	.dc.b	$F1,$A3,$F3,$87,$8D,$FC,$E3,$E8,$3D,$B3,$FD,$67,$2F,$E0,$F9,$7D
	.dc.b	$09,$FB,$73,$ED,$34,$FF,$52,$3C,$67,$F0,$03,$ED,$1F,$E2,$23,$F9
	.dc.b	$ED,$FE,$35,$D5,$FF,$82,$9F,$E8,$67,$33,$B9,$E4,$3F,$A8,$DF,$93
	.dc.b	$2F,$AB,$8F,$C6,$9F,$EF,$D5,$FB,$4C,$BE,$5C,$7F,$6A,$C3,$ED,$4F
	.dc.b	$C3,$CB,$F3,$13,$FA,$B3,$57,$CE,$3F,$89,$65,$F4,$2F,$EC,$57,$F9
	.dc.b	$19,$E6,$7D,$17,$E6,$9C,$EF,$FC,$F7,$1A,$8F,$7B,$FC,$D2,$BF,$C6
	.dc.b	$78,$79,$AB,$78,$78,$73,$FE,$11,$7E,$0C,$3F,$B0,$3C,$DF,$E0,$95
	.dc.b	$FB,$58,$FC,$05,$7E,$86,$7B,$9F,$89,$3E,$69,$FB,$E3,$EB,$9C,$A3
	.dc.b	$F4,$CF,$A5,$FF,$39,$1D,$F0,$FE,$40,$FF,$B1,$E3,$47,$C3,$9F,$43
	.dc.b	$F9,$BC,$FE,$B6,$5F,$9B,$8F,$F2,$45,$7C,$1C,$FF,$9E,$D6,$A7,$E8
	.dc.b	$9A,$3D,$18,$FF,$48,$8F,$E0,$EF,$E6,$0F,$E4,$B8,$F9,$27,$FD,$56
	.dc.b	$79,$0F,$8F,$8F,$91,$FD,$ED,$77,$AF,$9F,$1D,$CF,$C4,$E1,$ED,$CF
	.dc.b	$EA,$3F,$0A,$F9,$57,$CF,$AF,$D5,$4F,$F3,$73,$F3,$8F,$CB,$3F,$66
	.dc.b	$68,$FE,$65,$1E,$8C,$FF,$98,$C7,$D4,$8F,$DE,$3F,$A0,$7F,$2D,$97
	.dc.b	$F0,$A8,$F6,$65,$FD,$F9,$F4,$3F,$96,$79,$5F,$E1,$E3,$51,$FE,$21
	.dc.b	$AF,$D7,$1F,$2D,$FF,$0B,$9E,$49,$F4,$CF,$CB,$8F,$F0,$41,$FB,$38
	.dc.b	$FC,$15,$FA,$2B,$94,$FF,$70,$3F,$91,$3E,$D7,$F9,$F9,$FC,$EB,$2F
	.dc.b	$A1,$5F,$2E,$3D,$35,$F9,$E7,$F8,$29,$F9,$4B,$F1,$1F,$A6,$AE,$31
	.dc.b	$F8,$F3,$89,$F8,$F3,$CD,$8F,$D6,$BF,$CD,$E5,$F6,$A7,$81,$F0,$E3
	.dc.b	$F2,$59,$FE,$F6,$3C,$27,$D1,$8F,$DE,$61,$FC,$1A,$B9,$4F,$E3,$0F
	.dc.b	$09,$F8,$C3,$C2,$63,$2F,$C3,$3F,$63,$3F,$EC,$97,$F2,$A3,$E2,$9F
	.dc.b	$C9,$4F,$F0,$E9,$E1,$3F,$72,$7E,$50,$FC,$ED,$78,$AF,$EE,$DF,$F6
	.dc.b	$83,$EE,$D7,$F6,$DA,$FB,$A7,$D4,$7E,$EE,$7F,$65,$D4,$7D,$97,$FA
	.dc.b	$39,$F0,$30,$FC,$86,$1D,$DF,$4E,$5F,$36,$BF,$3A,$7B,$72,$FE,$D8
	.dc.b	$FF,$A8,$E7,$FD,$12,$7F,$80,$5F,$E8,$75,$FE,$41,$7F,$20,$7D,$FC
	.dc.b	$3C,$8F,$E1,$CF,$96,$7D,$73,$F8,$1E,$1E,$8C,$7F,$BF,$C7,$F8,$D2
	.dc.b	$78,$BF,$2F,$2F,$EB,$15,$F3,$E3,$C3,$9F,$F0,$A8,$FD,$C4,$78,$4F
	.dc.b	$A7,$87,$95,$F9,$06,$4E,$27,$CC,$8F,$8E,$FD,$6A,$FD,$96,$1F,$A4
	.dc.b	$3C,$B1,$FD,$B5,$FD,$DE,$5F,$DA,$23,$F9,$91,$F7,$CF,$E1,$07,$F4
	.dc.b	$AB,$FA,$11,$F3,$21,$4C,$BE,$6C,$FF,$73,$9F,$EC,$A7,$C8,$BF,$C2
	.dc.b	$61,$F9,$E8,$FE,$0A,$7E,$79,$F8,$91,$F8,$88,$FA,$07,$92,$32,$71
	.dc.b	$8F,$63,$FD,$3A,$3E,$81,$F2,$31,$F9,$8F,$DC,$3D,$98,$7E,$EB,$0F
	.dc.b	$E7,$F1,$E5,$AF,$E2,$D1,$F8,$5B,$F1,$4F,$D9,$8F,$EF,$AF,$EA,$E7
	.dc.b	$F7,$77,$FC,$0A,$3F,$64,$7E,$14,$FD,$31,$FC,$60,$F5,$67,$F9,$77
	.dc.b	$E6,$47,$1C,$3F,$53,$5E,$2C,$7E,$56,$3F,$E0,$9A,$D6,$3F,$A0,$C3
	.dc.b	$C4,$7E,$83,$2F,$AA,$FF,$8F,$0F,$D0,$59,$5F,$F9,$34,$FE,$07,$3C
	.dc.b	$1F,$EB,$58,$7E,$D2,$7E,$51,$F8,$DC,$BF,$BD,$3F,$84,$CB,$F8,$96
	.dc.b	$5E,$4C,$7F,$09,$1F,$83,$7C,$18,$77,$C3,$E3,$DF,$CF,$3C,$A7,$91
	.dc.b	$FF,$28,$BF,$C6,$4F,$8E,$7B,$8F,$8E,$71,$3F,$23,$97,$E4,$1F,$F1
	.dc.b	$A4,$FC,$77,$57,$E3,$7B,$1E,$08,$FD,$0B,$F9,$63,$B9,$F6,$63,$F5
	.dc.b	$35,$FB,$D3,$F1,$D1,$FA,$97,$F8,$A5,$FE,$A6,$BF,$B3,$C7,$E5,$A7
	.dc.b	$F8,$29,$F9,$A8,$CE,$3F,$81,$BF,$93,$3E,$6C,$3D,$55,$E4,$3D,$51
	.dc.b	$E4,$7F,$52,$F4,$3F,$29,$87,$F8,$6D,$F4,$67,$FD,$65,$FE,$73,$8F
	.dc.b	$8D,$FC,$9E,$1F,$D6,$5F,$05,$7D,$69,$FA,$37,$F8,$4C,$BE,$E5,$7E
	.dc.b	$2C,$F9,$F5,$B7,$E5,$57,$C2,$7E,$55,$7B,$9F,$B1,$7E,$E7,$D5,$3F
	.dc.b	$9D,$7E,$BB,$FE,$1D,$BE,$F5,$F9,$29,$FE,$3C,$FE,$CB,$3F,$E4,$55
	.dc.b	$FC,$7A,$7F,$25,$97,$F4,$4B,$E3,$3F,$CA,$A7,$F7,$F9,$7F,$32,$BE
	.dc.b	$39,$FE,$6B,$2F,$E0,$59,$FF,$6F,$BF,$8F,$3F,$57,$2E,$91,$F1,$8F
	.dc.b	$C4,$9F,$08,$F4,$9F,$D8,$B2,$F0,$BF,$C3,$F2,$FA,$18,$7E,$2A,$3F
	.dc.b	$05,$1F,$07,$3F,$74,$79,$8F,$D7,$9F,$D8,$70,$F3,$67,$E9,$C3,$F8
	.dc.b	$DC,$FF,$42,$38,$1F,$D0,$8E,$07,$F2,$23,$15,$FD,$00,$FF,$11,$38
	.dc.b	$BF,$CD,$C6,$0F,$9D,$3D,$A3,$E7,$61,$F1,$AB,$F9,$B9,$F1,$8F,$BB
	.dc.b	$8F,$C6,$7E,$E9,$F8,$42,$1F,$5B,$2F,$D3,$57,$C1,$3E,$B1,$F1,$B1
	.dc.b	$FC,$C5,$FF,$66,$CB,$E9,$57,$0B,$FE,$8D,$18,$7F,$1F,$7C,$DF,$A1
	.dc.b	$19,$CB,$F5,$27,$E4,$E3,$F4,$A7,$F8,$D4,$F3,$9F,$6D,$FE,$2C,$FF
	.dc.b	$2B,$CB,$ED,$BF,$3E,$BC,$CF,$19,$E8,$7E,$D1,$FD,$6E,$5F,$D3,$1E
	.dc.b	$B8,$FF,$11,$9E,$B1,$FC,$68,$CC,$FF,$00,$AF,$D5,$4F,$A1,$F6,$4F
	.dc.b	$87,$0F,$64,$7E,$9E,$3F,$29,$3D,$2F,$F5,$CF,$92,$78,$C7,$D9,$8F
	.dc.b	$EC,$F3,$FE,$81,$3A,$1F,$6B,$1F,$9C,$FD,$D3,$E3,$9F,$CC,$E3,$A3
	.dc.b	$F7,$27,$D9,$5E,$83,$C9,$1E,$CA,$FE,$0C,$FC,$D8,$D3,$FA,$73,$D2
	.dc.b	$FF,$10,$3F,$A2,$BC,$4F,$D7,$CF,$67,$FC,$39,$97,$E1,$B1,$FE,$E7
	.dc.b	$1E,$6C,$3F,$8C,$3D,$0F,$34,$FE,$A5,$FE,$B4,$7E,$96,$3F,$3E,$E0
	.dc.b	$FE,$78,$76,$3E,$F5,$FD,$23,$F4,$73,$CE,$3F,$5E,$70,$3F,$3C,$7E
	.dc.b	$50,$FC,$B9,$F3,$4C,$1F,$36,$7F,$11,$7F,$C1,$63,$F9,$D9,$FE,$D0
	.dc.b	$3D,$95,$F1,$27,$E8,$C7,$CB,$3D,$A7,$F0,$87,$F7,$D1,$FD,$E6,$3F
	.dc.b	$7C,$7D,$F7,$D2,$ED,$FE,$02,$6E,$3E,$1C,$7B,$1F,$E0,$91,$FE,$2F
	.dc.b	$3F,$B1,$47,$3C,$7F,$8A,$BC,$CC,$1F,$36,$38,$3F,$C8,$0F,$86,$7B
	.dc.b	$0F,$D7,$1F,$5C,$F6,$1F,$2A,$7F,$44,$F0,$7E,$3D,$FE,$07,$1F,$EC
	.dc.b	$11,$89,$FA,$E6,$EB,$F4,$35,$F5,$E7,$E2,$CF,$F4,$C9,$F2,$BF,$42
	.dc.b	$3F,$DE,$2F,$94,$FD,$B3,$F4,$30,$FC,$94,$79,$9F,$D6,$4F,$C4,$8F
	.dc.b	$F4,$81,$FD,$9E,$36,$4A,$F1,$4F,$D0,$9F,$11,$FD,$EF,$3F,$EA,$AF
	.dc.b	$C9,$31,$9F,$F3,$3C,$BE,$7C,$7D,$C3,$BE,$3F,$D2,$AF,$F8,$AC,$7E
	.dc.b	$DD,$FC,$C4,$F0,$9F,$BA,$7C,$29,$FE,$92,$E2,$7C,$91,$87,$E2,$DF
	.dc.b	$92,$3F,$3D,$8F,$C0,$7E,$E1,$F9,$B9,$F9,$EF,$E0,$CF,$2B,$FC,$FA
	.dc.b	$FF,$8C,$C7,$F0,$13,$F7,$D7,$FC,$67,$3F,$E9,$4F,$CF,$7F,$92,$47
	.dc.b	$EC,$CF,$61,$F7,$4F,$E2,$4F,$F9,$DE,$BC,$F8,$FC,$79,$F9,$47,$F3
	.dc.b	$4C,$FF,$BF,$63,$FD,$36,$3F,$69,$87,$E3,$9F,$51,$D2,$7F,$CE,$37
	.dc.b	$FC,$72,$7C,$91,$FC,$72,$BE,$E5,$7B,$F2,$FF,$35,$5F,$BE,$3A,$BC
	.dc.b	$4F,$A6,$F9,$27,$F8,$04,$FC,$DB,$FD,$7C,$FE,$00,$FE,$28,$E1,$F5
	.dc.b	$1F,$DC,$A7,$D4,$FE,$5D,$FE,$83,$5D,$A3,$EB,$E7,$F9,$07,$F6,$A6
	.dc.b	$0F,$46,$3F,$D7,$A5,$9F,$FA,$1E,$7F,$C3,$F1,$87,$E0,$67,$FE,$6F
	.dc.b	$74,$7F,$88,$CB,$87,$F9,$BD,$B0,$CF,$71,$FE,$80,$32,$E1,$E4,$78
	.dc.b	$67,$93,$67,$93,$CA,$32,$F3,$35,$3C,$E1,$E9,$3E,$DF,$0D,$AE,$4D
	.dc.b	$7D,$2D,$7C,$4D,$7F,$D5,$53,$D2,$79,$92,$39,$B5,$E6,$D5,$F9,$AC
	.dc.b	$00

;--------------------------------------------------------------------------------
;	FGETEXP.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fgetexp_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FGETEXP.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FGETEXP.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fgetexp_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fgetexp.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FGETEXP.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fgetexp.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FGETEXP.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fgetexp_data_compressed::
	.dc.b	$00,$02,$84,$2C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CF,$ED,$1F,$8A,$7F,$87,$A7,$FC,$D0,$5F,$8A,$E1
	.dc.b	$F0,$FC,$7D,$FE,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$1F
	.dc.b	$07,$9F,$DB,$F4,$7B,$7F,$DF,$7E,$6F,$FA,$B0,$F8,$8C,$7C,$46,$7E
	.dc.b	$23,$5F,$11,$BF,$88,$E1,$F1,$1C,$7E,$17,$97,$C0,$F3,$F8,$1E,$9F
	.dc.b	$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03
	.dc.b	$E3,$F8,$1F,$27,$C0,$F9,$7E,$07,$CD,$F0,$3E,$7F,$81,$FF,$0A,$7C
	.dc.b	$E7,$E1,$BE,$13,$FC,$40,$BE,$07,$3E,$66,$BE,$07,$7F,$03,$C3,$E0
	.dc.b	$78,$FC,$0F,$2F,$81,$E7,$F0,$3F,$63,$D2,$EB,$F0,$3D,$BE,$07,$BF
	.dc.b	$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93,$E0,$7C,$BF
	.dc.b	$03,$E6,$F8,$1F,$3F,$C0,$FA,$3E,$07,$FC,$59,$F3,$BF,$88,$E9,$FF
	.dc.b	$1C,$7C,$07,$F0,$A5,$1F,$EF,$63,$FD,$60,$7F,$9A,$0F,$EB,$47,$F1
	.dc.b	$23,$81,$FC,$41,$FF,$21,$2F,$7B,$BF,$BD,$E0,$F7,$BC,$3E,$F7,$8B
	.dc.b	$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD,$E8,$F7,$BF
	.dc.b	$3B,$EF,$FE,$7B,$5F,$CF,$EB,$FA,$0D,$7F,$43,$AF,$BB,$DE,$F7,$FB
	.dc.b	$DF,$E4,$BF,$6F,$FC,$48,$7E,$3F,$8B,$FC,$B0,$BD,$CD,$3B,$F7,$38
	.dc.b	$7B,$9C,$7D,$CE,$5E,$E7,$3F,$73,$A7,$B9,$FB,$0E,$8E,$DE,$E7,$7F
	.dc.b	$73,$C1,$EE,$7E,$D3,$AF,$F6,$BD,$7F,$B6,$EB,$FD,$BF,$5F,$EE,$3A
	.dc.b	$FF,$73,$D7,$FB,$AE,$6F,$DD,$F3,$7F,$98,$FF,$15,$FF,$1E,$1E,$9E
	.dc.b	$73,$E9,$7F,$CE,$4F,$FB,$38,$FF,$44,$1F,$DE,$0F,$E6,$07,$F2,$F7
	.dc.b	$8F,$0F,$5E,$1E,$DC,$1D,$F8,$3E,$47,$BB,$F2,$7D,$DF,$95,$EE,$FC
	.dc.b	$BF,$77,$E6,$7B,$BF,$37,$DD,$F9,$DE,$EF,$CF,$F7,$7E,$87,$BB,$FE
	.dc.b	$7B,$F9,$FF,$D4,$70,$1F,$D4,$3F,$E8,$95,$EE,$6F,$DC,$E1,$EE,$71
	.dc.b	$F7,$39,$7B,$9C,$FD,$CE,$9E,$E7,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B
	.dc.b	$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$F3,$7C,$9F,$9D,$D9,$F3
	.dc.b	$FB,$3E,$87,$67,$D1,$EC,$FA,$5D,$9F,$4F,$B3,$EA,$76,$7D,$5E,$CF
	.dc.b	$AD,$D9,$FE,$99,$F4,$FD,$E7,$BB,$B4,$FB,$9F,$F5,$42,$EB,$E1,$D7
	.dc.b	$C7,$AF,$97,$5F,$3E,$BE,$9D,$7D,$7A,$FE,$67,$B7,$F3,$7D,$BF,$9D
	.dc.b	$ED,$F8,$7B,$BE,$87,$B7,$FE,$AB,$FA,$0F,$E0,$1C,$05,$F0,$99,$F8
	.dc.b	$AD,$7C,$AE,$FE,$17,$87,$C0,$F1,$F8,$6E,$5F,$01,$CF,$E0,$3A,$74
	.dc.b	$75,$F8,$0F,$81,$E4,$FC,$1F,$27,$E1,$79,$3F,$0F,$C9,$F8,$9E,$4F
	.dc.b	$8F,$E0,$3C,$9E,$D7,$97,$DA,$F3,$7B,$5E,$7F,$6B,$D1,$ED,$7A,$7D
	.dc.b	$AF,$57,$B5,$EB,$F6,$BD,$9E,$D7,$FB,$1F,$E8,$7F,$D4,$07,$E9,$4F
	.dc.b	$D2,$76,$97,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1F,$67,$D9
	.dc.b	$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BB,$E1,$F7,$7C,$5E,$EF
	.dc.b	$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE,$F9,$FD,$DF,$47,$BB,$F8,$7F
	.dc.b	$57,$D5,$EE,$FF,$B8,$7C,$FF,$D3,$9F,$89,$3F,$11,$D2,$7F,$10,$FF
	.dc.b	$BC,$5F,$F7,$61,$FE,$A5,$3F,$CB,$AE,$FD,$9E,$9E,$CF,$5F,$67,$B7
	.dc.b	$AF,$DF,$CC,$F0,$79,$9E,$1F,$33,$C5,$E6,$78,$FC,$CF,$27,$67,$97
	.dc.b	$7F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$FB,$E7,$E8,$BF,$1E,$7E,$3B,$84
	.dc.b	$17,$B1,$BE,$3E,$1E,$C7,$1F,$63,$97,$B1,$CF,$D8,$E9,$EC,$75,$F6
	.dc.b	$3B,$7B,$1F,$AF,$DF,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E
	.dc.b	$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7A,$7D,$8F,$3F,$3B,$E8
	.dc.b	$FF,$DD,$87,$F3,$13,$F1,$A7,$E3,$1F,$6A,$F5,$F8,$BF,$EA,$83,$FC
	.dc.b	$C1,$C2,$EB,$CF,$AB,$AF,$57,$7E,$AF,$0F,$57,$8F,$AB,$CB,$D5,$E7
	.dc.b	$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E,$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F
	.dc.b	$57,$C7,$EA,$FB,$3E,$93,$F9,$99,$FC,$54,$FC,$69,$F8,$CE,$62,$F7
	.dc.b	$F8,$7B,$FC,$7D,$FE,$5E,$CF,$3F,$67,$A7,$B3,$D7,$D9,$ED,$EC,$F7
	.dc.b	$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9,$EC,$F9,$7C,$FF
	.dc.b	$37,$85,$E7,$F6,$7E,$9F,$C0,$7F,$1A,$3F,$19,$CB,$FC,$A7,$D9,$CF
	.dc.b	$B3,$AF,$67,$7E,$CF,$0F,$67,$8F,$B3,$CB,$D9,$E7,$E6,$74,$F6,$7A
	.dc.b	$FB,$3D,$BD,$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F,$67,$C7,$EC,$F9
	.dc.b	$3D,$9F,$2F,$B3,$E6,$F6,$7F,$13,$F4,$BF,$8A,$E6,$3F,$DE,$4F,$E2
	.dc.b	$A7,$E2,$5C,$FB,$5C,$3D,$DE,$3E,$EF,$2F,$6F,$9F,$B7,$D3,$DB,$EB
	.dc.b	$ED,$F6,$F6,$FB,$FB,$7E,$0F,$6F,$C3,$ED,$F8,$BD,$BF,$1F,$B7,$E4
	.dc.b	$F6,$FC,$BE,$DF,$9B,$DB,$F3,$FB,$7E,$8F,$6F,$F4,$3F,$4D,$FD,$E0
	.dc.b	$FE,$B4,$7F,$12,$3F,$0C,$7E,$17,$97,$F7,$AB,$DA,$E3,$ED,$72,$F6
	.dc.b	$B9,$FB,$5D,$3D,$AE,$BE,$D7,$6F,$6B,$BF,$B5,$E0,$F6,$BC,$3E,$D7
	.dc.b	$8B,$DA,$F1,$FB,$5E,$4F,$6B,$CB,$ED,$79,$BD,$AF,$3F,$B5,$FB,$8F
	.dc.b	$A7,$FE,$BE,$7E,$18,$FE,$1F,$C9,$FC,$49,$FE,$54,$DF,$63,$87,$B1
	.dc.b	$C7,$D8,$E5,$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1
	.dc.b	$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E
	.dc.b	$7F,$63,$F8,$7F,$AF,$F8,$73,$FC,$7C,$7F,$5A,$3F,$0C,$7F,$0D,$E4
	.dc.b	$FE,$56,$BD,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6,$F5
	.dc.b	$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC,$BE
	.dc.b	$BF,$9B,$D7,$F3,$FA,$FE,$8F,$5F,$F9,$3F,$D4,$7F,$8F,$8F,$F1,$01
	.dc.b	$FD,$60,$FE,$18,$7E,$1F,$94,$FE,$1D,$FE,$8A,$FF,$58,$1E,$AF,$4F
	.dc.b	$57,$AF,$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E
	.dc.b	$AF,$93,$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FA,$3D,$5F,$E7,$FF,$53
	.dc.b	$FE,$26,$3F,$86,$9F,$88,$E2,$3F,$88,$6F,$AB,$C3,$D5,$E3,$EA,$F2
	.dc.b	$F5,$79,$FA,$BD,$3D,$5E,$BE,$AF,$6F,$57,$BF,$AB,$E0,$F5,$7C,$3E
	.dc.b	$AF,$8B,$D5,$F1,$FA,$BE,$4F,$57,$CB,$EA,$F9,$BD,$5F,$3F,$AB,$FD
	.dc.b	$5B,$EA,$BF,$87,$1F,$E8,$53,$FC,$42,$7F,$0D,$3F,$11,$C8,$7F,$10
	.dc.b	$FF,$74,$5E,$A7,$3F,$53,$A7,$A9,$D7,$D4,$ED,$EA,$77,$F5,$3C,$1E
	.dc.b	$A7,$87,$D4,$F1,$7A,$9E,$3F,$53,$C9,$EA,$79,$7D,$4F,$37,$A9,$E7
	.dc.b	$F5,$3D,$1E,$A7,$A7,$D4,$FE,$DF,$E2,$7F,$A1,$8F,$F3,$41,$FE,$20
	.dc.b	$3F,$AA,$1F,$89,$3F,$11,$C8,$7F,$C3,$E7,$E2,$1F,$F0,$7A,$F4,$BA
	.dc.b	$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA,$5E,$2F,$4B,$C7,$E9,$79
	.dc.b	$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$A3,$D2,$FF,$03,$FD,$5F,$F9
	.dc.b	$B8,$FE,$AA,$7E,$24,$FC,$41,$E7,$E1,$FF,$11,$AF,$4B,$8F,$A5,$CB
	.dc.b	$D2,$E7,$E9,$74,$F4,$BA,$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA
	.dc.b	$5E,$2F,$4B,$C7,$E9,$79,$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$F8
	.dc.b	$7B,$EB,$3F,$AB,$1F,$EC,$D3,$FC,$D2,$7F,$55,$3F,$12,$7E,$23,$90
	.dc.b	$FE,$21,$FF,$1E,$BF,$EC,$D1,$E8,$75,$F4,$3B,$7A,$1D,$FD,$0F,$07
	.dc.b	$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F,$43,$CD,$E8,$79
	.dc.b	$FD,$0F,$47,$A1,$E9,$F4,$3F,$C6,$FF,$5B,$FE,$CE,$3F,$D6,$07,$F9
	.dc.b	$A0,$FF,$0E,$1F,$C4,$4F,$C4,$9F,$E6,$63,$F1,$1C,$47,$F1,$0F,$F9
	.dc.b	$61,$79,$FD,$BC,$FE,$FE,$7F,$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7
	.dc.b	$F9,$3C,$FF,$2F,$9F,$E6,$F3,$FC,$FE,$7F,$A3,$CF,$FF,$27,$FC,$17
	.dc.b	$FA,$D8,$FF,$0E,$9F,$C4,$4F,$E2,$07,$E2,$78,$3F,$C6,$9E,$7F,$0F
	.dc.b	$3F,$8F,$9F,$CB,$CF,$E7,$E7,$F4,$F3,$FA,$F9,$FD,$BC,$FE,$FE,$7F
	.dc.b	$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7,$F9,$3C,$FF,$2F,$9F,$E6,$F3
	.dc.b	$FF,$CC,$BF,$5D,$FE,$1D,$1F,$27,$8F,$8F,$CF,$C4,$EB,$DE,$DF,$BD
	.dc.b	$C3,$DA,$E3,$ED,$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$FE,$7C,$FA,$F7,$CB,$DF,$94,$C7,$C8,$67,$E2,$B5
	.dc.b	$E7,$6F,$E1,$B8,$7C,$37,$1F,$84,$E5,$F0,$3C,$FD,$EE,$9E,$F7,$5F
	.dc.b	$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27
	.dc.b	$BD,$E5,$F7,$BC,$DE,$F7,$FA,$4B,$EC,$3F,$DF,$27,$F9,$80,$FE,$A8
	.dc.b	$7F,$53,$3F,$08,$7F,$0B,$F7,$FF,$D5,$EB,$DE,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$FD,$53,$EC,$FF,$98
	.dc.b	$4F,$C2,$F1,$FF,$B0,$9F,$F7,$A9,$FE,$61,$3F,$AA,$9F,$C3,$4F,$E1
	.dc.b	$8F,$F8,$77,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8E,$FE,$C7,$83
	.dc.b	$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79,$3D,$8F,$2F,$B1,$E6,$F6,$3C
	.dc.b	$FE,$C7,$A3,$D8,$FF,$60,$FD,$8F,$E1,$B8,$BF,$DB,$2B,$D7,$CF,$AF
	.dc.b	$AF,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D
	.dc.b	$7F,$2F,$AF,$E6,$F5,$FC,$FE,$BF,$FB,$63,$EC,$9F,$35,$7E,$4F,$1F
	.dc.b	$1F,$9F,$8F,$D7,$C4,$EF,$E1,$B8,$7B,$BC,$7D,$DE,$5E,$EF,$3F,$77
	.dc.b	$A7,$BB,$D7,$DD,$ED,$EE,$F7,$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE
	.dc.b	$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$FE,$EE,$FB,$3F,$F5,$31,$F8
	.dc.b	$6F,$38,$BE,$17,$5F,$0B,$BF,$85,$E1,$F0,$3C,$7E,$17,$97,$C1,$73
	.dc.b	$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F,$82,$F0,$7C,$17,$87,$E0
	.dc.b	$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E,$0B,$CD,$F0,$5E,$7F,$82
	.dc.b	$F4,$7C,$13,$9E,$78,$FF,$87,$7F,$BE,$7C,$0E,$5F,$EF,$DF,$03,$67
	.dc.b	$FD,$DC,$7F,$98,$CF,$F0,$D8,$F7,$B5,$EF,$6F,$DE,$E1,$EF,$71,$F7
	.dc.b	$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE
	.dc.b	$1F,$7B,$C5,$EF,$78,$FD,$EF,$37,$DA,$7E,$23,$93,$D8,$FF,$BB,$8F
	.dc.b	$F3,$11,$FE,$61,$3F,$A9,$1F,$C3,$D3,$FC,$3D,$E3,$EC,$74,$F6,$3A
	.dc.b	$FB,$1D,$BD,$8E,$FE,$C7,$83,$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79
	.dc.b	$3D,$8F,$2F,$B1,$E6,$F6,$3C,$FE,$C7,$A3,$D8,$F9,$DF,$6B,$FE,$EF
	.dc.b	$3F,$14,$7E,$27,$87,$EB,$3F,$EE,$F3,$FC,$C6,$7F,$86,$CF,$E1,$E3
	.dc.b	$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD,$3E,$FE,$9F,$83,$D3,$F0,$FA
	.dc.b	$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F,$A7,$E6,$F4,$FC,$FE,$9F,$A3
	.dc.b	$D3,$FC,$37,$DB,$7E,$24,$FE,$1D,$C7,$F8,$D5,$E9,$EB,$E1,$37,$F0
	.dc.b	$9C,$3E,$13,$8F,$AB,$CB,$D5,$E7,$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E
	.dc.b	$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F,$57,$C7,$EA,$F9,$3D,$5F,$2F,$AB
	.dc.b	$E6,$F5,$7C,$FE,$AF,$E7,$78,$7F,$11,$CD,$FF,$00,$7E,$0B,$1F,$05
	.dc.b	$9F,$82,$D7,$C1,$6F,$DA,$E1,$F0,$5C,$7D,$AE,$5E,$D7,$3F,$6B,$A7
	.dc.b	$B5,$D7,$DA,$ED,$ED,$77,$F6,$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F
	.dc.b	$6B,$C9,$ED,$79,$7D,$AF,$37,$B5,$FB,$3F,$B7,$FD,$A7,$39,$FE,$20
	.dc.b	$7E,$1D,$C7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$D1,$EF,$7F,$0B,$FB
	.dc.b	$8F,$E1,$C7,$E1,$B9,$8F,$E1,$9F,$E2,$8B,$DD,$E1,$EE,$F1,$F3,$B9
	.dc.b	$7B,$BC,$FD,$DE,$9E,$EF,$5F,$77,$B7,$BB,$DF,$DD,$F0,$7B,$BE,$1F
	.dc.b	$77,$C5,$EE,$F8,$FD,$DF,$27,$BB,$E5,$F7,$7C,$DE,$EF,$9F,$DD,$F4
	.dc.b	$7B,$BF,$C8,$FE,$E7,$F0,$E3,$DE,$C7,$BD,$9F,$7B,$5E,$F6,$FD,$EE
	.dc.b	$1E,$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$75,$F7,$BB,$7B,$DD,$FD
	.dc.b	$EF,$07,$BD,$E1,$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD
	.dc.b	$EF,$7F,$39,$F6,$BF,$0E,$7F,$55,$3F,$0C,$3D,$DD,$7B,$BB,$F7,$78
	.dc.b	$4F,$1F,$77,$97,$BB,$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7B,$BD,$FD,$DF
	.dc.b	$07,$BB,$E1,$F7,$7C,$5E,$EF,$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE
	.dc.b	$F9,$FD,$DF,$EA,$1F,$75,$FD,$4C,$FC,$31,$F8,$5F,$78,$FE,$15,$FE
	.dc.b	$B4,$BD,$CE,$3E,$E7,$2F,$73,$9F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B
	.dc.b	$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7
	.dc.b	$9B,$DC,$F3,$FB,$9E,$8F,$73,$FB,$47,$DD,$FF,$0C,$3F,$0D,$CB,$FD
	.dc.b	$BD,$7B,$7A,$F6,$F7,$ED,$F0,$F6,$F8,$FB,$7C,$BD,$BE,$7E,$DF,$4F
	.dc.b	$6D,$FC,$03,$0B,$A7,$58,$7D,$E7,$F7,$DF,$83,$74,$F1,$8C,$36,$BF
	.dc.b	$9A,$B6,$77,$1C,$E7,$6D,$36,$58,$DA,$DD,$7E,$28,$D9,$97,$13,$C2
	.dc.b	$77,$5D,$AB,$4D,$8C,$C7,$83,$EE,$5F,$9B,$FF,$38,$8F,$F0,$B4,$63
	.dc.b	$E3,$3F,$92,$C7,$F8,$76,$30,$E1,$AF,$F2,$A6,$BC,$A7,$A3,$C5,$AA
	.dc.b	$36,$E6,$7B,$B9,$9D,$CE,$A3,$13,$A8,$C1,$27,$64,$8D,$36,$7D,$0D
	.dc.b	$7D,$0D,$7D,$11,$C0,$8F,$06,$BC,$1A,$F0,$34,$43,$45,$34,$78,$3E
	.dc.b	$0B,$CA,$47,$CA,$D7,$46,$08,$60,$87,$CF,$78,$34,$D9,$4D,$9F,$4D
	.dc.b	$DB,$4D,$94,$D9,$F4,$DD,$BF,$57,$8F,$E6,$91,$F9,$AD,$7E,$6B,$5F
	.dc.b	$C2,$35,$FC,$23,$5E,$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E
	.dc.b	$07,$E0,$38,$FF,$0A,$47,$F0,$AD,$74,$60,$86,$08,$7E,$7D,$E0,$D3
	.dc.b	$65,$36,$7E,$8D,$DB,$4D,$94,$D9,$FA,$37,$6F,$E9,$78,$FF,$3A,$47
	.dc.b	$F3,$AD,$7F,$3A,$D7,$F6,$8D,$7F,$68,$D7,$A1,$C8,$87,$22,$9C,$8A
	.dc.b	$72,$29,$C8,$E0,$43,$81,$FB,$0E,$3F,$DA,$91,$FD,$AB,$5D,$18,$21
	.dc.b	$82,$1F,$C3,$DE,$0D,$36,$53,$67,$F1,$37,$6D,$36,$53,$67,$F1,$37
	.dc.b	$6F,$F1,$5E,$3F,$E1,$A4,$7F,$86,$B5,$FE,$1A,$D7,$F9,$23,$5F,$E4
	.dc.b	$8D,$7A,$1C,$88,$72,$29,$C8,$A7,$22,$9C,$8E,$04,$38,$1F,$C8,$38
	.dc.b	$FF,$92,$91,$FE,$4A,$D7,$46,$08,$60,$87,$F3,$F7,$83,$4D,$94,$D9
	.dc.b	$FD,$0D,$DB,$4D,$94,$D9,$FD,$0D,$DB,$FD,$17,$8F,$F9,$E9,$1F,$E7
	.dc.b	$AD,$7F,$9E,$B5,$FE,$A8,$D7,$FA,$A3,$5E,$87,$22,$1C,$8A,$72,$29
	.dc.b	$C8,$A7,$23,$81,$0E,$07,$F5,$0E,$1F,$EA,$A4,$7F,$AA,$B5,$D1,$82
	.dc.b	$18,$21,$FD,$BD,$E0,$D3,$65,$36,$7F,$73,$76,$D3,$65,$36,$7F,$73
	.dc.b	$76,$FF,$75,$F8,$2F,$ED,$A4,$7F,$B6,$B5,$FE,$DA,$D7,$FC,$08,$D7
	.dc.b	$FC,$08,$D7,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81,$FE
	.dc.b	$00,$F8,$2F,$F0,$29,$1F,$F0,$2B,$5D,$18,$21,$82,$1F,$E1,$F7,$83
	.dc.b	$4D,$94,$D9,$FE,$23,$76,$D3,$65,$36,$7F,$88,$DD,$BF,$E2,$5F,$82
	.dc.b	$FF,$0E,$91,$FF,$0E,$B5,$FF,$0E,$B5,$FF,$1A,$35,$FF,$1A,$35,$E8
	.dc.b	$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$7F,$8C,$3E,$0B,$FC
	.dc.b	$6A,$47,$FC,$6A,$D7,$46,$08,$60,$87,$F9,$3D,$E0,$D3,$65,$36,$7F
	.dc.b	$94,$DD,$B4,$D9,$4D,$9F,$E5,$37,$6F,$F9,$57,$E0,$BF,$C9,$A4,$7F
	.dc.b	$C9,$AD,$7F,$C9,$AD,$7F,$CC,$8D,$7F,$CC,$8D,$7A,$1C,$88,$72,$29
	.dc.b	$C8,$A7,$22,$9C,$8E,$04,$38,$1F,$E6,$0F,$82,$FF,$32,$91,$FF,$32
	.dc.b	$B5,$D1,$82,$18,$21,$FE,$7F,$78,$34,$D9,$4D,$9F,$E8,$37,$6D,$36
	.dc.b	$53,$67,$FA,$0D,$DB,$FE,$85,$F8,$2F,$F3,$E9,$1F,$F3,$EB,$5F,$F3
	.dc.b	$EB,$5F,$F4,$A3,$5F,$F4,$A3,$5E,$87,$22,$1C,$8A,$72,$29,$C8,$A7
	.dc.b	$23,$81,$0E,$07,$FA,$43,$E0,$BF,$D2,$A4,$7F,$D2,$AD,$74,$60,$86
	.dc.b	$08,$7F,$AB,$DE,$0D,$36,$53,$67,$FA,$CD,$DB,$4D,$94,$D9,$FE,$B3
	.dc.b	$76,$FF,$AD,$7E,$0B,$FD,$5A,$47,$FD,$5A,$D7,$FD,$5A,$D7,$FD,$88
	.dc.b	$D7,$FD,$88,$D7,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81
	.dc.b	$FE,$C0,$F8,$2F,$F6,$29,$1F,$F6,$2B,$5D,$18,$21,$82,$1F,$ED,$F7
	.dc.b	$83,$4D,$94,$D9,$FE,$E3,$76,$D3,$65,$36,$7F,$B8,$DD,$BF,$EE,$5F
	.dc.b	$82,$FF,$6E,$91,$FF,$6E,$B5,$FF,$6E,$B5,$FF,$7A,$35,$FF,$7A,$35
	.dc.b	$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$08,$3A,$5E,$C6
	.dc.b	$22,$46,$1F,$F8,$01,$19,$78,$C6,$5C,$35,$E3,$1B,$7C,$11,$97,$8C
	.dc.b	$65,$F1,$36,$35,$1C,$23,$04,$7C,$64,$3C,$2E,$5A,$64,$F1,$94,$F0
	.dc.b	$94,$EC,$7B,$BE,$4A,$3D,$1A,$FB,$1A,$FA,$9A,$F9,$DA,$FB,$1B,$3F
	.dc.b	$39,$B2,$69,$27,$C0,$43,$C0,$53,$C0,$7C,$0F,$8C,$FA,$A4,$70,$43
	.dc.b	$E9,$B9,$69,$93,$E9,$B9,$69,$93,$E9,$39,$69,$93,$EC,$B9,$69,$9A
	.dc.b	$FC,$6C,$65,$B1,$9A,$4E,$5B,$19,$8C,$11,$C1,$0F,$59,$0F,$49,$4F
	.dc.b	$39,$4F,$29,$67,$C6,$D9,$7C,$27,$E8,$09,$1F,$8E,$6B,$F9,$E6,$BF
	.dc.b	$9C,$6B,$F9,$96,$BF,$9E,$6C,$FE,$C1,$B3,$E0,$9E,$C4,$3B,$14,$EC
	.dc.b	$53,$B1,$4F,$D5,$FC,$27,$EE,$08,$FE,$99,$AF,$ED,$9A,$FE,$D1,$AF
	.dc.b	$EC,$5A,$F4,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70,$3F,$84
	.dc.b	$FC,$27,$F1,$02,$3F,$BE,$6B,$A3,$04,$30,$43,$F8,$7B,$C1,$A6,$CA
	.dc.b	$6C,$FE,$20,$ED,$A6,$CA,$6C,$FE,$32,$ED,$FE,$37,$F0,$9F,$CA,$08
	.dc.b	$FF,$16,$6B,$FC,$99,$AF,$F2,$46,$BF,$C8,$5A,$F4,$39,$10,$E4,$53
	.dc.b	$91,$4E,$45,$39,$1C,$08,$70,$3F,$9C,$FC,$27,$F4,$02,$3F,$CB,$9A
	.dc.b	$E8,$C1,$0C,$10,$FE,$7E,$F0,$69,$B2,$9B,$3F,$A0,$3B,$69,$B2,$9B
	.dc.b	$3F,$A4,$BB,$7F,$A5,$FC,$27,$F5,$82,$3F,$D1,$9A,$FF,$56,$6B,$FD
	.dc.b	$51,$AF,$F5,$16,$BD,$0E,$44,$39,$14,$E4,$53,$91,$4E,$47,$02,$1C
	.dc.b	$0F,$ED,$DF,$28,$F4,$7F,$B3,$36,$3F,$B6,$35,$DC,$68,$8E,$9A,$E8
	.dc.b	$C1,$0C,$10,$FF,$01,$47,$56,$C7,$26,$9C,$4A,$71,$3F,$C1,$B1,$C5
	.dc.b	$B1,$C5,$A7,$12,$9C,$4F,$F0,$B4,$71,$6C,$71,$69,$C4,$A7,$11,$1C
	.dc.b	$5B,$1C,$5A,$71,$29,$C6,$38,$12,$38,$36,$38,$35,$E1,$1A,$23,$A6
	.dc.b	$BF,$E2,$4F,$85,$FF,$15,$91,$FF,$0C,$B5,$D1,$82,$18,$21,$FE,$2A
	.dc.b	$78,$34,$D9,$4D,$9F,$E2,$B7,$6D,$36,$53,$67,$F8,$D1,$DB,$FE,$36
	.dc.b	$F8,$5F,$F2,$19,$1F,$F1,$8B,$5F,$F2,$0B,$5F,$F1,$FB,$5F,$F1,$E3
	.dc.b	$5E,$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$07,$EB,$3C,$AF
	.dc.b	$D5,$91,$FD,$5B,$5D,$18,$21,$82,$1F,$A6,$78,$34,$D9,$4D,$9F,$A4
	.dc.b	$76,$D3,$65,$36,$7E,$89,$DB,$FE,$66,$F8,$5F,$F3,$99,$1F,$F3,$0B
	.dc.b	$5F,$F3,$8B,$5F,$F3,$7B,$5F,$F3,$63,$5E,$87,$22,$1C,$8A,$72,$29
	.dc.b	$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$13
	.dc.b	$83,$FD,$37,$1E,$D6,$C7,$99,$AF,$85,$AF,$66,$BD,$5A,$F5,$DD,$3D
	.dc.b	$5B,$1D,$5A,$F5,$6B,$D5,$AF,$56,$BF,$EA,$2F,$85,$FF,$57,$91,$FF
	.dc.b	$4E,$B5,$FF,$56,$B5,$FF,$55,$B5,$FF,$54,$35,$E8,$72,$21,$C8,$A7
	.dc.b	$22,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10
	.dc.b	$C1,$2B,$FD,$AF,$1E,$D6,$C7,$99,$AF,$85,$AF,$66,$BD,$5A,$FF,$B4
	.dc.b	$A7,$FD,$BE,$47,$FD,$9A,$D7,$FD,$BA,$D7,$FD,$B6,$D7,$FD,$B0,$D7
	.dc.b	$A1,$FE,$D6,$F8,$5F,$F7,$59,$1C,$10,$FF,$74,$B9,$69,$93,$FD,$D2
	.dc.b	$E5,$A7,$FB,$8B,$E1,$7F,$DE,$E4,$70,$43,$FD,$EA,$E5,$A6,$4F,$F7
	.dc.b	$AB,$96,$9F,$EE,$EF,$85,$FF,$7E,$5A,$D1,$86,$CE,$C8,$0F,$F8,$03
	.dc.b	$1B,$23,$A3,$25,$38,$4F,$78,$ED,$F0,$B2,$7F,$E0,$0B,$5E,$C7,$0A
	.dc.b	$46,$63,$35,$E9,$8C,$46,$DA,$A3,$4F,$63,$A9,$D7,$E1,$7E,$01,$1C
	.dc.b	$9C,$69,$19,$8E,$33,$F2,$63,$11,$E4,$9D,$BC,$A7,$4D,$8C,$BC,$9A
	.dc.b	$F1,$9F,$A5,$F0,$B2,$38,$35,$E0,$D7,$C0,$D7,$E4,$35,$F5,$36,$BF
	.dc.b	$01,$F0,$BF,$7C,$93,$E5,$6B,$E5,$6B,$ED,$6B,$F5,$5A,$FC,$77,$F1
	.dc.b	$5F,$0B,$23,$E0,$35,$F8,$0D,$7E,$43,$5F,$C0,$35,$FA,$4D,$9F,$CC
	.dc.b	$37,$03,$49,$3E,$02,$1E,$02,$9E,$02,$9E,$02,$9E,$02,$CF,$81,$B3
	.dc.b	$E0,$6C,$F8,$1E,$C4,$3B,$14,$EC,$53,$B1,$67,$B3,$67,$B3,$67,$B3
	.dc.b	$D0,$87,$42,$9D,$0B,$3D,$1B,$3D,$1B,$3D,$1E,$44,$39,$16,$79,$36
	.dc.b	$79,$36,$79,$3C,$09,$3C,$1B,$3C,$1B,$3C,$27,$44,$9D,$36,$35,$18
	.dc.b	$23,$82,$4F,$D4,$21,$F2,$4A,$7C,$02,$9E,$92,$9E,$42,$CF,$85,$B3
	.dc.b	$E0,$6C,$F8,$1E,$C4,$3B,$14,$EC,$53,$B1,$67,$B3,$67,$B3,$67,$B3
	.dc.b	$D0,$87,$42,$9F,$C8,$7F,$38,$7C,$9F,$F4,$06,$C7,$F1,$56,$BF,$C7
	.dc.b	$1A,$FF,$2E,$6B,$FC,$E9,$AF,$58,$E8,$47,$A3,$5E,$8D,$7A,$35,$E8
	.dc.b	$D7,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38
	.dc.b	$1A,$21,$A2,$9A,$30,$43,$04,$B4,$FC,$86,$CF,$BD,$B3,$E8,$6B,$E3
	.dc.b	$6B,$E0,$6B,$DD,$AF,$79,$EC,$49,$EC,$D8,$EC,$D7,$B3,$5E,$CD,$7B
	.dc.b	$35,$ED,$1D,$08,$F4,$6B,$D1,$AF,$46,$BD,$1A,$F4,$39,$10,$E4,$53
	.dc.b	$91,$4E,$45,$39,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08
	.dc.b	$60,$93,$F5,$09,$3F,$25,$B3,$F0,$1B,$3E,$96,$BE,$46,$BE,$16,$BE
	.dc.b	$06,$BE,$09,$EC,$49,$EC,$D8,$EC,$D7,$B3,$5E,$CD,$7B,$35,$ED,$1D
	.dc.b	$08,$F4,$6B,$D1,$AF,$46,$BD,$1A,$F4,$39,$10,$E4,$53,$91,$4E,$45
	.dc.b	$39,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$93,$F5
	.dc.b	$09,$3F,$25,$FE,$07,$F1,$FF,$CC,$C9,$3F,$CC,$DA,$FF,$33,$6B,$FC
	.dc.b	$CD,$AF,$F3,$36,$BF,$CF,$DB,$1D,$A3,$A1,$0E,$85,$3A,$14,$E8,$53
	.dc.b	$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8,$AF,$27,$81,$0E,$05,$38,$15
	.dc.b	$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$53,$F2,$1B,$3E,$F6,$BE,$86,$BE
	.dc.b	$36,$BE,$06,$BD,$DB,$3D,$E7,$B1,$27,$B3,$5E,$CD,$7B,$35,$EC,$D7
	.dc.b	$B3,$63,$B4,$74,$21,$D0,$A7,$42,$9D,$0A,$74,$2B,$D1,$E4,$43,$91
	.dc.b	$4E,$45,$39,$15,$E4,$F0,$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38
	.dc.b	$23,$82,$4F,$D4,$24,$FC,$96,$CF,$C0,$6B,$E9,$6B,$E4,$6B,$E1,$6B
	.dc.b	$E0,$6C,$F8,$3D,$7F,$F1,$91,$27,$FC,$64,$D7,$FC,$64,$D7,$FC,$64
	.dc.b	$D7,$FC,$64,$D7,$FC,$7C,$FF,$8F,$7E,$2A,$47,$F8,$F5,$AF,$F8,$F5
	.dc.b	$AF,$F8,$F5,$AF,$F8,$F5,$AF,$F9,$25,$B3,$FE,$1A,$F8,$8F,$F2,$51
	.dc.b	$27,$FC,$94,$D7,$FC,$94,$D7,$FC,$94,$D7,$FC,$94,$D7,$FC,$AC,$FF
	.dc.b	$95,$7E,$22,$47,$F9,$55,$AF,$F9,$55,$AF,$F9,$55,$AF,$F9,$55,$AF
	.dc.b	$F9,$85,$B3,$FE,$32,$F8,$8F,$F3,$11,$27,$FC,$C4,$D7,$FC,$C4,$D7
	.dc.b	$FC,$C4,$D7,$FC,$C4,$D7,$FC,$DC,$FF,$9B,$7E,$22,$47,$F9,$B5,$AF
	.dc.b	$F9,$B5,$AF,$F9,$B5,$AF,$F9,$B5,$AF,$F9,$E5,$B3,$FE,$4A,$F8,$8F
	.dc.b	$F3,$D1,$27,$FC,$F4,$D7,$FC,$F4,$D7,$FC,$F4,$D7,$FC,$F4,$D7,$FD
	.dc.b	$0C,$FF,$A1,$7E,$22,$47,$FA,$15,$AF,$FA,$15,$AF,$FA,$15,$AF,$FA
	.dc.b	$15,$AF,$FA,$45,$B3,$FE,$62,$EA,$FF,$49,$12,$7F,$D2,$4D,$7F,$D2
	.dc.b	$4D,$7F,$D2,$4D,$7F,$D2,$4D,$7F,$D3,$CF,$FA,$77,$E2,$24,$7F,$A7
	.dc.b	$5A,$FF,$A7,$5A,$FF,$A7,$5A,$FF,$A7,$5A,$FF,$AA,$5B,$3F,$E7,$AE
	.dc.b	$EF,$F5,$51,$27,$FD,$54,$D7,$FD,$54,$D7,$FD,$54,$D7,$FD,$54,$D7
	.dc.b	$FD,$6C,$FF,$AD,$7E,$22,$47,$FA,$D5,$AF,$FA,$D5,$AF,$FA,$D5,$AF
	.dc.b	$FA,$D5,$AF,$FB,$05,$B3,$FE,$92,$F8,$8F,$F6,$11,$27,$FD,$84,$D7
	.dc.b	$FD,$84,$D7,$FD,$84,$D7,$FD,$84,$D7,$FD,$9C,$FF,$B3,$7E,$22,$47
	.dc.b	$FB,$35,$AF,$FB,$35,$AF,$FB,$35,$AF,$FB,$35,$AF,$FB,$65,$B3,$FE
	.dc.b	$AA,$F8,$8F,$F6,$D1,$27,$FD,$B4,$D7,$FD,$B4,$D7,$FD,$B4,$D7,$FD
	.dc.b	$B4,$D7,$FD,$CC,$FF,$B9,$7E,$22,$47,$FB,$95,$AF,$FB,$95,$AF,$FB
	.dc.b	$95,$AF,$FB,$95,$AF,$FB,$C5,$B3,$FE,$C2,$F8,$8F,$F7,$91,$27,$FD
	.dc.b	$E4,$D7,$FD,$E4,$D7,$FD,$E4,$D7,$FD,$E4,$D7,$FD,$FC,$FF,$BF,$7E
	.dc.b	$22,$47,$FB,$F5,$AF,$FB,$F5,$AF,$FB,$F5,$AF,$FB,$F5,$AA,$7F,$DB
	.dc.b	$31,$88,$78,$CD,$D1,$08,$69,$AE,$63,$83,$83,$45,$72,$F0,$6B,$D1
	.dc.b	$AF,$38,$E4,$F2,$38,$38,$8D,$C7,$99,$AE,$CE,$91,$B2,$3E,$36,$BD
	.dc.b	$0C,$10,$C1,$0E,$66,$4A,$64,$F4,$19,$29,$93,$CC,$F8,$1E,$24,$36
	.dc.b	$53,$67,$32,$9C,$CF,$71,$C0,$86,$8E,$46,$8C,$1A,$30,$48,$F9,$24
	.dc.b	$8F,$34,$6E,$30,$F8,$CE,$11,$A7,$A9,$0C,$9B,$8E,$2F,$A5,$AF,$99
	.dc.b	$AF,$78,$E8,$F2,$38,$38,$8D,$C7,$76,$BB,$3A,$46,$C8,$F8,$DA,$F4
	.dc.b	$30,$43,$04,$39,$99,$29,$93,$D0,$64,$A6,$4F,$33,$E0,$78,$90,$D9
	.dc.b	$4D,$9C,$CA,$73,$3D,$C7,$02,$1A,$39,$1A,$30,$68,$C1,$23,$E5,$D3
	.dc.b	$9D,$F1,$1F,$BB,$8C,$36,$33,$5F,$C0,$A7,$2D,$8C,$BF,$C2,$23,$2D
	.dc.b	$8C,$BF,$C3,$63,$2D,$8C,$BF,$C4,$A3,$2D,$8C,$BF,$C6,$DB,$1A,$8C
	.dc.b	$4F,$9C,$8F,$95,$AF,$8D,$AF,$85,$AF,$76,$BD,$0E,$44,$39,$14,$E4
	.dc.b	$53,$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18
	.dc.b	$24,$7C,$72,$4F,$CA,$24,$FC,$08,$CB,$63,$33,$EB,$9C,$B6,$32,$FA
	.dc.b	$23,$2D,$8C,$BE,$78,$CB,$63,$2F,$A2,$32,$D8,$CB,$EA,$6C,$6A,$31
	.dc.b	$3E,$72,$3E,$56,$BE,$36,$BE,$16,$BD,$DA,$F4,$39,$10,$E4,$53,$91
	.dc.b	$4E,$45,$39,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60
	.dc.b	$91,$F1,$C9,$3F,$28,$93,$F0,$23,$2D,$8C,$CF,$AE,$72,$D8,$CB,$E8
	.dc.b	$8C,$B6,$32,$F9,$E3,$2D,$8C,$BE,$88,$CB,$63,$2F,$A9,$B1,$A8,$C4
	.dc.b	$F9,$C8,$F9,$5A,$F8,$DA,$F8,$5A,$F7,$6B,$D0,$E4,$43,$91,$4E,$45
	.dc.b	$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$FC,$AF,$F3,$97
	.dc.b	$C2,$FF,$39,$6C,$7F,$37,$6B,$FC,$D5,$AF,$18,$E0,$47,$83,$5E,$0D
	.dc.b	$78,$1A,$21,$A2,$9A,$30,$43,$04,$3F,$A0,$47,$91,$B1,$DD,$A7,$42
	.dc.b	$9C,$CA,$73,$3F,$CC,$11,$CD,$B1,$CD,$A7,$32,$9C,$CA,$73,$A4,$F3
	.dc.b	$6C,$F3,$6C,$F3,$6B,$CD,$AF,$39,$E4,$49,$E4,$D8,$E4,$D7,$93,$5E
	.dc.b	$51,$C0,$8F,$06,$BC,$1A,$F0,$34,$43,$45,$34,$60,$86,$08,$7D,$02
	.dc.b	$1F,$20,$B3,$F0,$1B,$3E,$96,$CF,$91,$A7,$84,$A7,$80,$A7,$80,$EC
	.dc.b	$49,$EC,$D9,$EC,$D9,$EC,$D3,$B1,$4E,$C5,$3B,$4F,$42,$4F,$46,$CF
	.dc.b	$46,$CF,$46,$BD,$1A,$F4,$9E,$44,$9E,$4F,$E1,$3E,$17,$F7,$44,$9F
	.dc.b	$DD,$35,$FD,$D3,$5F,$E0,$4D,$7F,$56,$D7,$F4,$6F,$F0,$5F,$85,$91
	.dc.b	$FC,$15,$AF,$F0,$56,$BF,$C3,$5A,$FE,$D9,$AF,$EB,$9B,$7F,$A6,$F8
	.dc.b	$5F,$E2,$44,$9F,$E2,$4D,$7F,$89,$35,$FE,$34,$D7,$F8,$1B,$5F,$DD
	.dc.b	$BF,$C7,$7E,$16,$47,$F1,$D6,$BF,$C7,$5A,$FF,$25,$6B,$FC,$39,$AF
	.dc.b	$F0,$66,$CF,$EF,$BE,$17,$F9,$51,$27,$F9,$53,$5F,$E5,$4D,$7F,$99
	.dc.b	$35,$FE,$36,$D7,$F8,$9B,$FC,$D7,$E1,$64,$7F,$35,$6B,$FC,$D5,$AF
	.dc.b	$F3,$D6,$BF,$C9,$9A,$FF,$1E,$6C,$FF,$16,$6E,$46,$92,$7C,$04,$3C
	.dc.b	$05,$3C,$05,$3C,$05,$3C,$05,$9F,$03,$67,$C0,$D9,$F0,$3D,$88,$76
	.dc.b	$29,$D8,$A7,$62,$CF,$66,$CF,$66,$CF,$67,$A1,$0E,$85,$3A,$16,$7A
	.dc.b	$36,$7A,$36,$7A,$3C,$88,$72,$2C,$F2,$6C,$F2,$6C,$F2,$78,$12,$78
	.dc.b	$36,$78,$36,$78,$4E,$89,$3A,$6C,$6A,$30,$47,$04,$9F,$A8,$43,$E4
	.dc.b	$94,$F8,$05,$3D,$25,$3C,$85,$9F,$0B,$67,$C0,$D9,$F0,$3D,$88,$76
	.dc.b	$29,$D8,$A7,$62,$CF,$66,$CF,$66,$CF,$67,$A1,$0E,$85,$3F,$C0,$BF
	.dc.b	$0B,$FE,$27,$23,$FE,$27,$6B,$FE,$27,$6B,$FE,$2F,$6B,$FE,$18,$6B
	.dc.b	$D0,$E4,$43,$91,$4E,$45,$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D
	.dc.b	$10,$D1,$4D,$18,$21,$82,$57,$F8,$82,$3D,$AD,$8F,$33,$5F,$0B,$5E
	.dc.b	$CD,$7A,$B5,$EB,$49,$EA,$D8,$EA,$D7,$AB,$5E,$AD,$7A,$B5,$EB,$1D
	.dc.b	$08,$F4,$6B,$D1,$AF,$46,$BD,$1A,$F4,$39,$10,$E4,$53,$91,$4E,$45
	.dc.b	$39,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$93,$F4
	.dc.b	$C9,$3F,$25,$B3,$F0,$1B,$3E,$96,$BE,$46,$BE,$16,$BE,$06,$BE,$09
	.dc.b	$EC,$49,$EC,$D8,$EC,$D7,$B3,$5E,$CD,$7B,$35,$ED,$1D,$08,$F4,$6B
	.dc.b	$D1,$AF,$46,$BD,$1A,$F4,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08
	.dc.b	$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$93,$F5,$09,$3F,$25
	.dc.b	$B3,$F0,$1B,$3E,$96,$BE,$46,$BF,$53,$E1,$7F,$DA,$44,$7F,$DA,$4D
	.dc.b	$7F,$DA,$4D,$7F,$DB,$4D,$7F,$D8,$6D,$7F,$D7,$6D,$8F,$F5,$9B,$68
	.dc.b	$D2,$4F,$80,$87,$80,$A7,$80,$A7,$80,$A7,$80,$B3,$E0,$6C,$F8,$1B
	.dc.b	$3E,$07,$B1,$0E,$C5,$3B,$14,$EC,$59,$EC,$D9,$EC,$D9,$EC,$F4,$21
	.dc.b	$D0,$A7,$42,$CF,$46,$CF,$46,$CF,$47,$91,$0E,$45,$9E,$4D,$9E,$4D
	.dc.b	$9E,$4F,$02,$4F,$06,$CF,$06,$CF,$09,$D1,$27,$4D,$8D,$46,$08,$E0
	.dc.b	$93,$F5,$08,$7C,$92,$9F,$00,$A7,$A4,$A7,$90,$B3,$E1,$6C,$F8,$1B
	.dc.b	$3E,$07,$B1,$0E,$C5,$3B,$14,$EC,$59,$EC,$D9,$EC,$D9,$EC,$F4,$21
	.dc.b	$D0,$FB,$BF,$0A,$61,$29,$49,$CC,$E6,$30,$E0,$4F,$09,$C4,$62,$76
	.dc.b	$F2,$9D,$3C,$5A,$E6,$36,$F7,$9E,$73,$A8,$C1,$C6,$76,$FA,$E3,$D3
	.dc.b	$F0,$BC,$24,$D2,$4F,$16,$BE,$16,$BE,$F6,$BD,$63,$9B,$C4,$F1,$7C
	.dc.b	$2F,$72,$38,$21,$ED,$72,$D3,$27,$CD,$72,$D3,$DD,$F0,$BE,$B2,$38
	.dc.b	$21,$F2,$5C,$B4,$C9,$F6,$5C,$B4,$C9,$ED,$72,$D3,$27,$DB,$72,$D3
	.dc.b	$37,$F8,$88,$CB,$63,$34,$9C,$B6,$33,$18,$23,$83,$EF,$7C,$2F,$DB
	.dc.b	$23,$82,$1F,$89,$72,$D3,$27,$E6,$5C,$B4,$FC,$BF,$2B,$C1,$FC,$CC
	.dc.b	$23,$0E,$A3,$31,$88,$FD,$13,$5F,$D3,$B5,$FD,$6B,$5F,$D8,$35,$ED
	.dc.b	$1D,$08,$F7,$21,$D9,$CB,$4C,$9E,$07,$2D,$32,$78,$9C,$B4,$C9,$E4
	.dc.b	$72,$D3,$27,$98,$B1,$E9,$24,$70,$8D,$1E,$52,$1E,$32,$9E,$12,$9D
	.dc.b	$CA,$72,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$23
	.dc.b	$DA,$FC,$08,$F7,$46,$23,$CE,$D7,$C4,$D7,$BB,$5E,$CD,$7B,$47,$42
	.dc.b	$3D,$C8,$76,$72,$D3,$27,$81,$CB,$4C,$9E,$27,$2D,$32,$79,$1C,$B4
	.dc.b	$C9,$E6,$2C,$7A,$49,$1C,$23,$47,$94,$87,$8C,$A7,$84,$A7,$72,$9C
	.dc.b	$8E,$04,$38,$14,$E0,$53,$81,$F8,$0C,$3F,$AD,$FC,$2F,$F4,$32,$47
	.dc.b	$F4,$96,$BF,$D4,$5A,$FF,$5F,$6B,$FD,$05,$AF,$F5,$D6,$C7,$68,$E8
	.dc.b	$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8,$87,$22,$9C,$8A,$72,$2B
	.dc.b	$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA,$70,$47,$04,$C5,$3F
	.dc.b	$21,$B3,$EF,$6B,$E8,$6B,$E3,$6B,$E0,$6B,$DD,$B3,$DE,$7B,$12,$7B
	.dc.b	$35,$EC,$D7,$B3,$5E,$CD,$7B,$36,$3B,$47,$42,$1D,$0A,$74,$29,$D0
	.dc.b	$A7,$42,$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A,$70
	.dc.b	$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$24,$FD,$42,$4F,$C9,$6C,$FC,$06
	.dc.b	$BE,$96,$BE,$46,$BE,$16,$BE,$06,$CF,$82,$7B,$12,$7B,$35,$EC,$D7
	.dc.b	$B3,$5E,$CD,$7B,$36,$3B,$47,$42,$1D,$0A,$74,$29,$D0,$A7,$42,$BD
	.dc.b	$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1,$D1
	.dc.b	$0D,$1F,$47,$E1,$7F,$CD,$C4,$7F,$CE,$8D,$7F,$CF,$8D,$7F,$D1,$4D
	.dc.b	$7F,$CD,$8D,$7F,$D1,$0D,$8F,$F4,$9B,$68,$D2,$4F,$80,$87,$80,$A7
	.dc.b	$80,$A7,$80,$A7,$80,$B3,$E0,$6C,$F8,$1B,$3E,$07,$B1,$0E,$C5,$3B
	.dc.b	$14,$EC,$59,$EC,$D9,$EC,$D9,$EC,$F4,$21,$D0,$A7,$42,$CF,$46,$CF
	.dc.b	$46,$CF,$47,$91,$0E,$45,$9E,$4D,$9E,$4D,$9E,$4F,$02,$4F,$06,$CF
	.dc.b	$06,$CF,$09,$D1,$27,$4D,$8D,$46,$08,$E0,$93,$F5,$08,$7C,$92,$9F
	.dc.b	$00,$A7,$A4,$A7,$90,$B3,$E1,$6C,$F8,$1B,$3E,$07,$B1,$0E,$C5,$3B
	.dc.b	$14,$EC,$59,$EC,$D9,$EC,$D9,$EC,$F4,$21,$D0,$FB,$BF,$0B,$FE,$DF
	.dc.b	$23,$FE,$E5,$6B,$FE,$ED,$6B,$FE,$FB,$6B,$FE,$DD,$6B,$FE,$F9,$7F
	.dc.b	$DF,$7F,$7C,$FC,$07,$FD,$71,$22,$08,$D6,$BC,$27,$73,$A9,$CB,$C0
	.dc.b	$E4,$E1,$C3,$5E,$11,$C9,$DB,$5F,$03,$63,$84,$6A,$7C,$47,$AB,$F0
	.dc.b	$67,$3D,$24,$F8,$CF,$54,$69,$C5,$23,$11,$E6,$6B,$ED,$6B,$EC,$6B
	.dc.b	$D2,$7B,$BC,$5A,$FF,$B2,$3C,$9F,$F6,$69,$1F,$F6,$AB,$5F,$F6,$7B
	.dc.b	$5E,$06,$88,$68,$A7,$CF,$F9,$AF,$F6,$EA,$69,$1F,$F7,$0B,$5F,$88
	.dc.b	$6C,$C9,$0C,$BB,$7E,$53,$DB,$A5,$E2,$FF,$BB,$1B,$3D,$1A,$FD,$16
	.dc.b	$BF,$85,$6B,$F8,$E6,$BF,$91,$6A,$69,$23,$C0,$48,$F0,$35,$F0,$35
	.dc.b	$F0,$35,$F0,$35,$F0,$35,$F0,$36,$3C,$11,$D8,$87,$62,$9D,$8A,$76
	.dc.b	$29,$D8,$A7,$62,$BD,$9E,$84,$3A,$14,$E8,$53,$A1,$4E,$85,$7A,$3C
	.dc.b	$88,$72,$29,$C8,$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A
	.dc.b	$2B,$A7,$04,$70,$48,$FA,$84,$8F,$92,$D7,$E0,$35,$F4,$B5,$F2,$35
	.dc.b	$F0,$B5,$F0,$36,$3C,$11,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$A7,$62
	.dc.b	$BD,$9E,$84,$3A,$14,$E8,$53,$A1,$4E,$85,$7A,$3C,$88,$7F,$3E,$F9
	.dc.b	$C7,$BF,$FC,$12,$11,$87,$F8,$3B,$5F,$E5,$8D,$7F,$A0,$35,$E9,$1C
	.dc.b	$E3,$11,$FD,$35,$AF,$68,$D1,$1E,$ED,$7B,$B5,$EE,$D7,$B9,$C0,$87
	.dc.b	$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$7B,$5F,$23,$4E,$E5
	.dc.b	$3A,$14,$E6,$53,$99,$F1,$4E,$25,$38,$94,$E2,$53,$8C,$7C,$E2,$47
	.dc.b	$92,$34,$77,$21,$82,$47,$07,$C2,$D7,$C2,$D7,$C2,$D7,$C3,$1D,$E3
	.dc.b	$11,$D9,$AF,$68,$D1,$1E,$ED,$7B,$B5,$EE,$D7,$B9,$C0,$87,$02,$9C
	.dc.b	$0A,$70,$34,$43,$45,$34,$60,$86,$08,$7B,$5F,$23,$4E,$E5,$3A,$14
	.dc.b	$E6,$53,$99,$F1,$4E,$25,$38,$94,$E2,$53,$8C,$7C,$E2,$47,$92,$34
	.dc.b	$77,$21,$82,$47,$07,$C2,$D7,$C2,$D7,$C2,$D7,$C3,$1D,$E3,$11,$D9
	.dc.b	$AF,$68,$D1,$1E,$ED,$7B,$B5,$EE,$D7,$B9,$C0,$87,$02,$9C,$0A,$70
	.dc.b	$34,$43,$45,$34,$60,$86,$08,$7B,$5F,$23,$4E,$E5,$3A,$14,$E6,$53
	.dc.b	$98,$F9,$D7,$7F,$F3,$56,$C7,$F8,$E9,$AF,$F9,$7D,$AF,$F9,$CD,$AF
	.dc.b	$F9,$F1,$AF,$F9,$FD,$AF,$78,$EC,$47,$B3,$5E,$CD,$7B,$35,$EC,$D7
	.dc.b	$B3,$5E,$C7,$42,$1D,$0A,$74,$29,$D0,$A7,$42,$9D,$0E,$44,$39,$14
	.dc.b	$E4,$53,$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82
	.dc.b	$18,$24,$7D,$42,$47,$C9,$6C,$7C,$06,$BE,$96,$BE,$46,$BE,$16,$BE
	.dc.b	$06,$BE,$08,$EC,$47,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$C7,$42
	.dc.b	$1D,$0A,$74,$29,$D0,$A7,$42,$9D,$0E,$44,$39,$14,$E4,$53,$91,$4E
	.dc.b	$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$7C,$23,$DB,$FF,$18,$36
	.dc.b	$3F,$D6,$AD,$7F,$DB,$4D,$7F,$DD,$CD,$7F,$DE,$ED,$7F,$DF,$4D,$7B
	.dc.b	$C7,$62,$3D,$9A,$FF,$B9,$FE,$7D,$DB,$FE,$43,$6C,$7F,$B2,$1A,$EE
	.dc.b	$34,$47,$4D,$74,$60,$86,$0F,$F7,$B7,$D0,$3D,$5F,$F2,$93,$63,$FD
	.dc.b	$AE,$D7,$71,$A2,$3A,$6B,$A3,$04,$30,$40,$41,$1D,$DC,$18,$8E,$AE
	.dc.b	$23,$4F,$16,$BD,$1A,$F1,$3C,$47,$18,$E6,$F1,$6B,$C5,$AF,$13,$D0
	.dc.b	$6C,$E6,$53,$27,$13,$25,$32,$73,$3C,$46,$8C,$1C,$CC,$9C,$23,$D6
	.dc.b	$48,$F2,$B6,$39,$35,$E1,$1A,$23,$A6,$BA,$30,$43,$04,$3C,$64,$3C
	.dc.b	$C4,$3D,$C6,$4A,$64,$F6,$98,$21,$82,$47,$A5,$E4,$62,$3C,$EE,$23
	.dc.b	$4F,$56,$BD,$1A,$F1,$3D,$87,$18,$E6,$F1,$6B,$C5,$AF,$13,$E1,$1B
	.dc.b	$39,$94,$C9,$C4,$C9,$4C,$9C,$CF,$11,$A3,$07,$33,$27,$08,$F5,$92
	.dc.b	$3C,$AD,$8E,$4D,$78,$46,$88,$E9,$AE,$8C,$10,$C1,$0F,$19,$0F,$31
	.dc.b	$0F,$71,$92,$99,$3D,$A6,$08,$60,$91,$E9,$79,$18,$8F,$3B,$88,$D3
	.dc.b	$D5,$AF,$46,$BC,$4F,$61,$C6,$39,$BC,$5A,$F1,$6B,$C4,$F8,$46,$CE
	.dc.b	$65,$32,$71,$32,$53,$27,$33,$C4,$68,$C1,$CC,$C9,$C2,$3D,$64,$8F
	.dc.b	$2B,$63,$93,$5E,$11,$A2,$3A,$6B,$A3,$04,$30,$43,$C6,$43,$CC,$43
	.dc.b	$DC,$64,$A6,$4F,$69,$82,$18,$24,$7A,$5E,$46,$23,$CE,$E2,$34,$F5
	.dc.b	$6B,$D1,$AF,$13,$D8,$71,$8E,$6F,$16,$BC,$5A,$F1,$3E,$11,$B3,$99
	.dc.b	$4C,$9C,$4C,$94,$C9,$CC,$F1,$1A,$30,$73,$32,$70,$8F,$59,$23,$CA
	.dc.b	$D8,$E4,$D7,$84,$68,$8E,$9A,$E8,$C1,$0C,$10,$F1,$90,$F3,$10,$F7
	.dc.b	$19,$29,$93,$DA,$60,$86,$09,$1E,$97,$F8,$0F,$AA,$F8,$DF,$F0,$1C
	.dc.b	$7F,$81,$E3,$0E,$03,$87,$6D,$7F,$C1,$6D,$79,$47,$17,$8B,$63,$51
	.dc.b	$FE,$16,$6B,$A3,$C0,$E0,$91,$C4,$91,$A6,$BE,$46,$BE,$46,$BE,$43
	.dc.b	$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$F8,$0F,$91
	.dc.b	$A7,$72,$9D,$0A,$73,$29,$CC,$F9,$06,$CA,$6C,$A6,$E3,$E5,$BD,$CC
	.dc.b	$9D,$08,$60,$8F,$07,$B3,$5E,$CD,$7B,$47,$B5,$E2,$D8,$D4,$76,$6B
	.dc.b	$A3,$C8,$E0,$91,$C4,$91,$A6,$BE,$46,$BE,$46,$BE,$43,$81,$0E,$05
	.dc.b	$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$F8,$0F,$91,$A7,$72,$9D
	.dc.b	$0A,$73,$29,$CC,$F9,$06,$CA,$6C,$A6,$E3,$E5,$BD,$CC,$9D,$08,$60
	.dc.b	$8F,$07,$B3,$5E,$CD,$7B,$47,$B5,$E2,$D8,$D4,$76,$6B,$A3,$C8,$E0
	.dc.b	$91,$C4,$91,$A6,$BE,$46,$BE,$46,$BE,$43,$81,$0E,$05,$38,$14,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$10,$F8,$0F,$91,$A7,$72,$9D,$0A,$73,$29
	.dc.b	$CC,$F9,$06,$CA,$6C,$A6,$E3,$E5,$BD,$CC,$9D,$08,$60,$8F,$07,$B3
	.dc.b	$5E,$CD,$7B,$47,$B5,$E2,$D8,$D4,$76,$6B,$A3,$C8,$E0,$91,$C4,$91
	.dc.b	$A6,$BE,$46,$BE,$46,$BE,$43,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A
	.dc.b	$68,$C1,$0C,$10,$F8,$0F,$91,$A7,$72,$9D,$0A,$73,$29,$CC,$F9,$06
	.dc.b	$CA,$6C,$A6,$E3,$E5,$BD,$CC,$9D,$08,$60,$8F,$07,$B3,$5E,$CD,$7B
	.dc.b	$47,$B5,$E2,$D8,$D4,$76,$6B,$A3,$C8,$E0,$91,$C4,$91,$A6,$BE,$46
	.dc.b	$BE,$46,$BE,$43,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C
	.dc.b	$10,$F8,$0F,$91,$A7,$72,$9D,$0A,$73,$29,$CC,$F9,$06,$CA,$6C,$A6
	.dc.b	$E3,$E5,$BD,$CC,$9D,$08,$60,$8F,$07,$B3,$5E,$CD,$7B,$47,$B5,$E2
	.dc.b	$D8,$D4,$76,$6B,$A3,$C8,$E0,$91,$C4,$91,$A6,$BE,$46,$BE,$46,$BE
	.dc.b	$43,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$F8,$0F
	.dc.b	$91,$A7,$72,$9D,$0A,$73,$29,$CC,$F9,$06,$CA,$6C,$A6,$E3,$E5,$BD
	.dc.b	$CC,$9D,$08,$60,$8F,$07,$B3,$5F,$F0,$77,$D1,$3D,$4F,$F4,$03,$63
	.dc.b	$FD,$14,$D7,$FD,$1E,$D4,$40,$E5,$C3,$5D,$B5,$DC,$78,$08,$F8,$1A
	.dc.b	$F8,$1A,$F7,$38,$9B,$21,$B2,$9B,$8F,$01,$23,$9C,$79,$9A,$F9,$9A
	.dc.b	$F6,$3A,$10,$E8,$EA,$3A,$47,$80,$E8,$53,$A1,$4E,$87,$23,$0F,$27
	.dc.b	$B1,$C8,$A7,$22,$9C,$8D,$1C,$CC,$BC,$DC,$10,$D9,$0E,$86,$4A,$64
	.dc.b	$EC,$60,$86,$08,$F7,$78,$35,$DB,$5D,$C7,$C9,$23,$F0,$9A,$FA,$9A
	.dc.b	$F9,$4E,$26,$C8,$6C,$A6,$E3,$C0,$48,$E7,$1E,$66,$BE,$66,$BD,$8E
	.dc.b	$84,$3A,$3A,$8E,$91,$E0,$3A,$14,$E8,$53,$A1,$C8,$C3,$C9,$EC,$72
	.dc.b	$29,$C8,$A7,$23,$47,$33,$2F,$37,$04,$36,$43,$A1,$92,$99,$3B,$18
	.dc.b	$21,$82,$3D,$DE,$0D,$76,$D7,$71,$F2,$48,$FC,$26,$BE,$A6,$BE,$53
	.dc.b	$89,$B2,$1B,$29,$B8,$F0,$12,$39,$C7,$99,$AF,$99,$AF,$63,$A1,$0E
	.dc.b	$8E,$A3,$A4,$78,$0E,$85,$3A,$14,$E8,$72,$30,$F2,$7B,$1C,$8A,$72
	.dc.b	$29,$C8,$D1,$CC,$CB,$CD,$C1,$0D,$90,$E8,$64,$A6,$4E,$C6,$08,$60
	.dc.b	$8F,$77,$83,$5D,$B5,$DC,$7C,$92,$3F,$09,$AF,$A9,$AF,$94,$E2,$6C
	.dc.b	$86,$CA,$6E,$3C,$04,$8E,$71,$E6,$6B,$E6,$6B,$D8,$E8,$43,$A3,$A8
	.dc.b	$E9,$1E,$03,$A1,$4E,$85,$3A,$1C,$8C,$3C,$9E,$C7,$22,$9C,$8A,$72
	.dc.b	$34,$73,$32,$F3,$70,$43,$64,$3A,$19,$29,$93,$B1,$82,$18,$23,$DD
	.dc.b	$E0,$D7,$6D,$77,$1F,$DE,$DF,$F0,$57,$D3,$BD,$87,$97,$99,$4E,$63
	.dc.b	$0D,$8E,$4D,$7F,$C0,$CD,$7A,$CF,$47,$6D,$76,$D7,$73,$DA,$31,$1F
	.dc.b	$E2,$09,$E8,$D8,$CB,$E6,$6C,$6A,$31,$3C,$C8,$F8,$9A,$F8,$9A,$F8
	.dc.b	$8D,$10,$D1,$4D,$18,$21,$82,$1E,$77,$AB,$4E,$45,$38,$94,$E2,$7A
	.dc.b	$8D,$94,$D9,$4D,$C7,$C1,$24,$FC,$42,$4F,$8A,$78,$1D,$C8,$60,$93
	.dc.b	$CA,$38,$36,$3C,$6D,$7C,$6D,$7C,$73,$E0,$76,$D7,$6D,$77,$3D,$A3
	.dc.b	$11,$E4,$9E,$8D,$8C,$BE,$66,$C6,$A3,$13,$CC,$8F,$89,$AF,$89,$AF
	.dc.b	$88,$D1,$0D,$14,$D1,$82,$18,$21,$E7,$7A,$B4,$E4,$53,$89,$4E,$27
	.dc.b	$A8,$D9,$4D,$94,$DC,$7C,$12,$4F,$C4,$24,$F8,$A7,$81,$DC,$86,$09
	.dc.b	$3C,$A3,$83,$63,$C6,$D7,$C6,$D7,$C7,$3E,$07,$6D,$76,$D7,$73,$DA
	.dc.b	$31,$1E,$49,$E8,$D8,$CB,$E6,$6C,$6A,$31,$3C,$C8,$F8,$9A,$F8,$9A
	.dc.b	$F8,$8D,$10,$D1,$4D,$18,$21,$82,$1E,$77,$AB,$4E,$45,$38,$94,$E2
	.dc.b	$7A,$8D,$94,$D9,$4D,$C7,$C1,$24,$FC,$6F,$AE,$7B,$EF,$73,$F9,$F4
	.dc.b	$61,$B1,$9B,$FE,$8D,$39,$6C,$65,$FF,$62,$46,$5B,$19,$7F,$DA,$D1
	.dc.b	$96,$C6,$5F,$F6,$DC,$65,$B1,$95,$19,$6C,$66,$30,$47,$04,$9F,$59
	.dc.b	$27,$D2,$D7,$CE,$D7,$CA,$D7,$C6,$D7,$C2,$D9,$F0,$36,$7C,$13,$D8
	.dc.b	$87,$62,$9D,$8A,$76,$29,$D8,$B3,$D9,$B1,$D9,$E8,$43,$A1,$4E,$85
	.dc.b	$3A,$16,$3A,$36,$3A,$3C,$88,$72,$29,$C8,$B1,$C9,$B1,$C9,$E0,$43
	.dc.b	$81,$63,$83,$63,$83,$A2,$46,$9B,$1A,$8C,$11,$C1,$27,$EA,$12,$7E
	.dc.b	$4B,$5F,$80,$D7,$D2,$D7,$C8,$D7,$C2,$D9,$F0,$36,$7C,$13,$D8,$87
	.dc.b	$62,$9D,$8A,$76,$29,$D8,$B3,$D9,$B1,$D9,$E8,$43,$A1,$4F,$A1,$F5
	.dc.b	$EE,$FF,$D6,$36,$FF,$C7,$CD,$9F,$F2,$23,$51,$03,$97,$0D,$76,$D7
	.dc.b	$73,$E0,$24,$F8,$1B,$1E,$06,$BD,$CE,$31,$B2,$3B,$6B,$B9,$F0,$12
	.dc.b	$79,$C7,$99,$AF,$99,$AF,$63,$A1,$0E,$8E,$A3,$A4,$F8,$27,$A3,$5E
	.dc.b	$8D,$7A,$1C,$9C,$47,$28,$EC,$72,$29,$C8,$A7,$23,$47,$33,$2F,$37
	.dc.b	$04,$36,$43,$A1,$92,$99,$3B,$18,$21,$82,$3D,$DE,$0D,$76,$D7,$73
	.dc.b	$F2,$49,$3F,$09,$FE,$7D,$E7,$BD,$CA,$49,$FF,$49,$46,$5B,$19,$AF
	.dc.b	$F4,$D4,$E5,$B1,$97,$F1,$51,$96,$C6,$5F,$CE,$46,$5B,$19,$7F,$43
	.dc.b	$19,$6C,$65,$46,$5B,$19,$8C,$11,$C1,$27,$D6,$49,$F4,$B5,$F3,$B5
	.dc.b	$F2,$B5,$F1,$B5,$F0,$B6,$7C,$0D,$9F,$04,$F6,$21,$FA,$FF,$B1,$7C
	.dc.b	$87,$F7,$16,$D7,$FB,$0D,$B3,$FE,$C9,$6B,$C6,$78,$12,$78,$36,$38
	.dc.b	$35,$E1,$1A,$23,$A6,$BA,$30,$43,$04,$3F,$8B,$4F,$91,$B3,$DD,$B3
	.dc.b	$D1,$A7,$32,$9C,$CF,$E4,$33,$CD,$B3,$CD,$B3,$CD,$A7,$32,$9C,$C4
	.dc.b	$F3,$6C,$F3,$6C,$F3,$69,$CC,$A7,$3F,$B2,$7B,$9F,$E2,$76,$CF,$FB
	.dc.b	$E5,$B3,$FE,$FF,$6B,$FC,$B5,$AF,$F3,$D6,$BF,$D0,$1A,$9A,$49,$F0
	.dc.b	$12,$7F,$9C,$4F,$F4,$8F,$90,$7C,$6F,$F8,$ED,$B3,$A9,$46,$1E,$2D
	.dc.b	$7F,$A1,$35,$FE,$A4,$D7,$B4,$F5,$24,$F5,$9D,$4F,$69,$CB,$E6,$FB
	.dc.b	$37,$1F,$F9,$3D,$B7,$E5,$6C,$EE,$74,$49,$D3,$63,$51,$82,$38,$3E
	.dc.b	$2F,$DA,$BE,$2B,$FC,$BC,$D9,$F8,$8D,$9D,$CE,$89,$3A,$6C,$6A,$30
	.dc.b	$47,$07,$D4,$F8,$57,$AF,$FE,$6D,$6C,$FD,$36,$CE,$E7,$44,$9D,$36
	.dc.b	$35,$18,$23,$83,$F0,$7F,$6C,$AF,$F3,$E3,$67,$F0,$4D,$9D,$CE,$89
	.dc.b	$3A,$6C,$6A,$30,$47,$07,$F8,$27,$CD,$FC,$8F,$C9,$7F,$A3,$DB,$3F
	.dc.b	$96,$6C,$EF,$55,$6D,$B5,$B6,$CE,$E7,$44,$9D,$36,$35,$1D,$FF,$03
	.dc.b	$7C,$A7,$8D,$B1,$98,$C1,$1C,$12,$7D,$93,$C1,$B3,$B6,$CE,$DF,$F1
	.dc.b	$C7,$AA,$F5,$5E,$17,$FA,$C6,$30,$D8,$CD,$7E,$EA,$72,$D8,$CD,$7F
	.dc.b	$8D,$67,$2D,$8C,$BE,$06,$BF,$E4,$A8,$DB,$9A,$F0,$12,$BB,$7C,$BF
	.dc.b	$8C,$8E,$09,$3D,$E3,$2D,$8C,$BE,$18,$CB,$63,$D5,$F2,$FE,$F2,$38
	.dc.b	$24,$7A,$E3,$2D,$8C,$BE,$D8,$CB,$63,$F8,$E7,$DC,$3E,$37,$FD,$B4
	.dc.b	$D8,$CC,$60,$8E,$09,$3F,$CA,$A7,$83,$67,$6D,$9D,$BF,$3F,$AB,$ED
	.dc.b	$12,$3E,$C3,$63,$EB,$35,$E1,$1A,$23,$A6,$BF,$8A,$FC,$EB,$FD,$07
	.dc.b	$1F,$85,$6C,$7E,$0D,$AE,$CF,$C7,$38,$7F,$04,$D5,$9A,$30,$D8,$E2
	.dc.b	$49,$F0,$34,$F0,$7C,$CF,$62,$47,$95,$B1,$A8,$C1,$1C,$10,$F7,$47
	.dc.b	$06,$C6,$DA,$6C,$F5,$C6,$DB,$1B,$69,$B3,$DF,$49,$E1,$18,$EA,$F8
	.dc.b	$E4,$AF,$9C,$D8,$F9,$CD,$7E,$A3,$5F,$9A,$D7,$A4,$7D,$7D,$FE,$99
	.dc.b	$1C,$12,$3E,$D4,$65,$B1,$97,$F0,$51,$96,$C7,$E0,$FC,$EF,$BE,$47
	.dc.b	$04,$8F,$C3,$46,$5B,$19,$7F,$1D,$19,$6C,$7E,$3F,$8B,$F1,$64,$70
	.dc.b	$48,$FC,$94,$65,$B1,$97,$F3,$51,$96,$C7,$E6,$FE,$67,$F2,$E4,$70
	.dc.b	$48,$FC,$F4,$65,$B1,$97,$F4,$D1,$96,$C7,$E9,$F9,$3F,$46,$47,$04
	.dc.b	$8F,$D5,$46,$5B,$19,$7F,$65,$19,$6C,$7E,$CF,$D2,$FD,$79,$1C,$12
	.dc.b	$3F,$6D,$19,$6C,$65,$FD,$F4,$65,$B1,$FB,$FF,$99,$FD,$D9,$1C,$12
	.dc.b	$3F,$81,$46,$5B,$19,$7F,$85,$46,$5B,$1F,$C2,$FE,$67,$F8,$39,$1C
	.dc.b	$12,$3F,$87,$46,$5B,$19,$7F,$8B,$46,$5B,$1F,$C5,$FE,$67,$F8,$99
	.dc.b	$1C,$12,$3F,$8D,$46,$5B,$19,$7F,$91,$46,$5B,$1F,$C8,$FE,$67,$F8
	.dc.b	$F9,$1C,$12,$3F,$93,$46,$5B,$19,$7F,$97,$46,$5B,$1F,$CB,$FE,$57
	.dc.b	$F9,$59,$1C,$12,$3F,$99,$46,$5B,$19,$7F,$9D,$46,$5B,$1F,$D6,$3E
	.dc.b	$5B,$FA,$B9,$1C,$12,$3F,$AA,$46,$5B,$19,$44,$1C,$C9,$E1,$19,$32
	.dc.b	$FF,$46,$F9,$99,$5F,$D3,$5B,$3F,$D3,$5A,$FF,$53,$6B,$FD,$31,$AF
	.dc.b	$F2,$E6,$F2,$FF,$4D,$F0,$7F,$A9,$92,$BF,$AD,$36,$BF,$AD,$35,$FE
	.dc.b	$C0,$D7,$FA,$CB,$5F,$E7,$6F,$F5,$EF,$5E,$4F,$F6,$56,$CF,$F6,$56
	.dc.b	$BF,$DA,$DA,$FF,$64,$6B,$FD,$19,$B5,$FD,$97,$E6,$7F,$B5,$92,$BF
	.dc.b	$B9,$36,$BF,$B9,$35,$FE,$F0,$D7,$FB,$8B,$5F,$EA,$6F,$F7,$6F,$99
	.dc.b	$93,$FD,$F5,$B3,$FD,$F5,$AF,$F8,$0D,$AF,$F7,$C6,$BF,$D7,$9B,$5F
	.dc.b	$DF,$7C,$FF,$F0,$19,$2B,$FC,$14,$DA,$FF,$05,$35,$FF,$08,$35,$FF
	.dc.b	$04,$B5,$FE,$D6,$FF,$83,$B3,$93,$FE,$15,$6C,$FF,$85,$5A,$FF,$86
	.dc.b	$DA,$FF,$85,$1A,$FF,$76,$6D,$7F,$85,$7B,$3F,$C3,$64,$AF,$F1,$13
	.dc.b	$6B,$FC,$44,$D7,$FC,$50,$D7,$FC,$42,$D7,$FC,$06,$FF,$9C,$FE,$25
	.dc.b	$34,$87,$F9,$CD,$CB,$4C,$CF,$F9,$CE,$32,$D8,$CB,$FE,$82,$8C,$B6
	.dc.b	$32,$FF,$9D,$63,$FD,$64,$EF,$C4,$F2,$12,$7C,$4D,$8F,$03,$5E,$11
	.dc.b	$A2,$3A,$7D,$5E,$C7,$BC,$8E,$09,$1E,$B8,$CB,$63,$2F,$AE,$32,$D8
	.dc.b	$F8,$DE,$BF,$CB,$23,$82,$47,$C7,$8C,$B6,$32,$FC,$78,$CB,$63,$E9
	.dc.b	$7A,$9F,$5C,$8E,$09,$1F,$4E,$32,$D8,$CB,$F4,$E3,$2D,$8F,$BD,$E8
	.dc.b	$7E,$0C,$8E,$09,$1F,$7E,$32,$D8,$CB,$F7,$E3,$2D,$8F,$C5,$7C,$0F
	.dc.b	$E3,$C8,$E0,$91,$F8,$B8,$CB,$63,$2F,$E2,$E3,$2D,$8F,$CB,$7C,$CF
	.dc.b	$E6,$C8,$E0,$91,$F9,$78,$CB,$63,$2F,$E5,$E3,$2D,$8F,$D1,$7C,$4F
	.dc.b	$E9,$C8,$E0,$91,$FA,$38,$CB,$63,$2F,$E8,$E3,$2D,$8F,$D7,$7C,$97
	.dc.b	$EC,$C8,$E0,$91,$FA,$F8,$CB,$63,$2F,$EB,$E3,$2D,$8F,$DD,$7B,$5F
	.dc.b	$BF,$23,$82,$47,$EE,$E3,$2D,$8C,$BF,$BB,$8C,$B6,$3F,$83,$7C,$CF
	.dc.b	$F0,$B2,$38,$24,$7F,$07,$8C,$B6,$32,$FF,$07,$8C,$B6,$32,$FF,$07
	.dc.b	$8C,$B6,$32,$FF,$AF,$37,$4E,$A3,$0D,$E1,$4E,$AB,$8C,$62,$7F,$92
	.dc.b	$7C,$CC,$9F,$E4,$4D,$BF,$E4,$0D,$7F,$8F,$35,$FE,$38,$D7,$C8,$DA
	.dc.b	$F0,$FA,$BF,$CC,$49,$5F,$CB,$DB,$5F,$CB,$5A,$FF,$2B,$6B,$FC,$A5
	.dc.b	$AF,$C1,$7F,$9E,$7C,$2C,$9F,$E7,$4D,$9F,$E7,$0D,$7F,$9B,$35,$FE
	.dc.b	$68,$D7,$E6,$36,$BE,$4F,$CC,$FF,$49,$25,$7F,$47,$6D,$7F,$45,$6B
	.dc.b	$FD,$0D,$AF,$F4,$16,$BF,$65,$FE,$A9,$F3,$32,$7F,$A9,$36,$7F,$A8
	.dc.b	$35,$FE,$9C,$D7,$FA,$63,$5F,$C2,$36,$BF,$03,$F3,$3F,$D8,$49,$5F
	.dc.b	$D7,$DB,$5F,$D7,$5A,$FF,$5B,$6B,$FD,$65,$AF,$E4,$5F,$ED,$9F,$33
	.dc.b	$27,$FB,$53,$67,$FB,$43,$5F,$EC,$CD,$7F,$B2,$35,$FC,$E3,$6B,$F3
	.dc.b	$3F,$33,$FD,$E4,$95,$FD,$DD,$B5,$FD,$D5,$AF,$F7,$36,$BF,$DC,$5A
	.dc.b	$FE,$A5,$FF,$02,$7C,$C4,$9F,$F0,$13,$67,$FC,$00,$D7,$FB,$F3,$5F
	.dc.b	$EF,$8D,$7F,$68,$DA,$7E,$76,$F7,$3F,$C2,$35,$FE,$75,$9C,$C6,$23
	.dc.b	$FC,$F9,$18,$BF,$F3,$D4,$23,$0F,$1A,$E9,$3A,$9C,$CF,$29,$CB,$BA
	.dc.b	$EC,$FF,$A2,$A7,$6D,$9D,$BC,$67,$6F,$86,$32,$EF,$E2,$73,$5E,$E6
	.dc.b	$C9,$A4,$BF,$34,$65,$B1,$9A,$F6,$35,$F5,$35,$FF,$50,$78,$AF,$03
	.dc.b	$CB,$1F,$EA,$26,$C6,$6B,$E4,$CE,$5B,$19,$8F,$9A,$D7,$FD,$50,$D7
	.dc.b	$BC,$F4,$24,$F2,$6C,$7D,$3F,$81,$F8,$24,$9F,$7B,$63,$51,$82,$38
	.dc.b	$21,$F6,$63,$83,$F6,$FE,$3B,$E6,$92,$7E,$5B,$63,$F0,$8D,$78,$46
	.dc.b	$88,$E9,$AE,$8C,$10,$C1,$0F,$C3,$47,$56,$C7,$26,$9C,$4A,$7E,$4F
	.dc.b	$E3,$7F,$0A,$48,$FC,$1B,$63,$F3,$8D,$78,$46,$88,$E9,$AE,$8C,$10
	.dc.b	$C1,$0F,$CF,$47,$56,$C7,$26,$9C,$4A,$71,$3F,$45,$1C,$5B,$1C,$5A
	.dc.b	$71,$29,$C7,$3F,$E0,$73,$C5,$B3,$F9,$4F,$17,$F3,$24,$70,$4A,$FC
	.dc.b	$BC,$65,$B1,$97,$F3,$31,$96,$C6,$5F,$D4,$C6,$5B,$19,$43,$24,$EE
	.dc.b	$33,$18,$9E,$11,$B9,$D3,$63,$33,$B8,$F3,$90,$F2,$94,$F1,$94,$EF
	.dc.b	$3D,$23,$94,$70,$21,$C0,$A7,$09,$F0,$CF,$12,$38,$24,$77,$3D,$05
	.dc.b	$9E,$EF,$52,$1D,$4D,$4F,$46,$CF,$47,$C6,$74,$2C,$F4,$79,$18,$9E
	.dc.b	$4D,$9E,$4F,$80,$E4,$59,$E4,$E0,$E6,$66,$79,$B6,$79,$BA,$21,$C4
	.dc.b	$92,$F4,$7F,$95,$92,$7F,$94,$B6,$7F,$94,$B5,$FE,$52,$D7,$F9,$BB
	.dc.b	$5F,$C4,$BF,$D1,$BE,$66,$4F,$F4,$46,$CF,$F5,$D6,$BF,$D6,$DA,$FF
	.dc.b	$57,$6B,$FD,$55,$B8,$7F,$70,$F9,$9F,$EA,$64,$AF,$EA,$2D,$AF,$ED
	.dc.b	$4D,$7F,$B4,$35,$FE,$C8,$D7,$FB,$13,$FD,$7B,$E6,$64,$FF,$5C,$6C
	.dc.b	$FF,$75,$6B,$FD,$CD,$AF,$F6,$F6,$BF,$DB,$5B,$5F,$E0,$86,$F0,$1A
	.dc.b	$4A,$F0,$12,$BC,$0D,$7C,$0F,$F7,$0F,$82,$FE,$EA,$47,$04,$8F,$F0
	.dc.b	$64,$65,$B1,$97,$FC,$19,$19,$6C,$7F,$7C,$F8,$DF,$F0,$09,$1C,$12
	.dc.b	$3F,$C3,$11,$96,$C6,$5F,$F0,$C4,$65,$B1,$FE,$08,$F9,$9F,$F0,$69
	.dc.b	$1C,$12,$3F,$C4,$91,$96,$C6,$5F,$F1,$24,$65,$B1,$FE,$14,$F9,$9F
	.dc.b	$F0,$C9,$1C,$12,$3F,$C6,$11,$96,$C6,$5F,$F1,$84,$65,$B1,$FE,$20
	.dc.b	$F8,$AF,$F1,$29,$1C,$12,$3F,$C7,$91,$96,$C6,$5F,$F1,$E4,$65,$B1
	.dc.b	$FE,$2C,$F0,$7F,$C6,$24,$70,$48,$FF,$24,$46,$5B,$19,$7F,$C9,$11
	.dc.b	$96,$C6,$5F,$F2,$34,$65,$B1,$97,$FC,$8D,$19,$6C,$67,$4F,$F3,$1C
	.dc.b	$E5,$B1,$FE,$44,$F9,$8F,$F2,$49,$1C,$12,$BF,$CC,$11,$96,$C6,$5F
	.dc.b	$F3,$04,$65,$B1,$97,$FC,$BD,$19,$6C,$65,$FF,$2F,$46,$5B,$19,$9F
	.dc.b	$F3,$DC,$E5,$B1,$9D,$53,$96,$C6,$63,$04,$70,$7F,$99,$3E,$63,$FC
	.dc.b	$D2,$47,$04,$BF,$F4,$04,$65,$B1,$97,$FD,$01,$19,$6C,$65,$FF,$3F
	.dc.b	$46,$5B,$19,$7F,$CF,$D1,$96,$C6,$6F,$FD,$37,$39,$6C,$66,$D3,$96
	.dc.b	$C6,$63,$04,$70,$4A,$F5,$90,$F4,$94,$F3,$94,$F2,$96,$BC,$6D,$AF
	.dc.b	$0B,$6B,$C0,$DA,$F0,$3D,$88,$76,$29,$D8,$B5,$D9,$B5,$D9,$B5,$D9
	.dc.b	$B5,$FE,$92,$E8,$79,$3F,$E9,$86,$C6,$63,$04,$70,$49,$FF,$61,$CF
	.dc.b	$06,$CE,$DB,$3B,$7F,$D8,$5F,$31,$FE,$B4,$24,$7F,$AC,$9B,$1F,$EC
	.dc.b	$B6,$BC,$23,$44,$74,$D7,$FD,$9D,$F0,$BF,$EC,$02,$47,$FA,$F9,$B1
	.dc.b	$FE,$D7,$6B,$C2,$34,$47,$4D,$7F,$DB,$5F,$0B,$FE,$CC,$24,$7F,$B2
	.dc.b	$9B,$1F,$EE,$36,$BC,$23,$44,$74,$D7,$FD,$CD,$F0,$BF,$ED,$82,$47
	.dc.b	$FB,$59,$B1,$FE,$EF,$6B,$C2,$34,$47,$4D,$7F,$DE,$5F,$0B,$FE,$E4
	.dc.b	$24,$7F,$B8,$9B,$1F,$EF,$B6,$BC,$23,$44,$74,$D7,$FC,$EE,$35,$7E
	.dc.b	$79,$FE,$7A,$9C,$CF,$FB,$FE,$30,$E2,$76,$FF,$BE,$A3,$2F,$48,$CC
	.dc.b	$F2,$8E,$D3,$A7,$A1,$C4,$91,$B9,$E3,$1D,$E7,$83,$C8,$F0,$7C,$2C
	.dc.b	$AE,$12,$8C,$4F,$16,$BE,$58,$D4,$F0,$9C,$CF,$BD,$AF,$A1,$F7,$7C
	.dc.b	$2C,$9E,$ED,$9F,$33,$5E,$CD,$7E,$53,$5F,$86,$DB,$FA,$3F,$0B,$F2
	.dc.b	$C9,$5E,$C6,$D7,$C3,$6B,$EB,$6B,$F5,$DA,$FD,$07,$EB,$7C,$2C,$9F
	.dc.b	$8E,$D9,$F9,$CD,$7E,$33,$5F,$C1,$35,$FB,$6D,$AF,$C4,$F6,$FE,$0C
	.dc.b	$95,$F5,$1B,$5F,$6D,$AF,$D3,$6B,$F8,$F6,$BF,$88,$7F,$1D,$F0,$B2
	.dc.b	$7E,$FB,$67,$F0,$CD,$7E,$F3,$5F,$CD,$35,$FC,$9B,$6B,$F4,$3F,$0B
	.dc.b	$F9,$B2,$57,$E3,$1B,$5F,$93,$6B,$F8,$B6,$BF,$A7,$6B,$FA,$07,$F4
	.dc.b	$DF,$0B,$27,$F2,$ED,$9F,$CF,$35,$FC,$B3,$5F,$D9,$35,$FD,$5B,$6B
	.dc.b	$F7,$3F,$0B,$FB,$32,$57,$E9,$1B,$5F,$AB,$6B,$FA,$36,$BF,$BF,$6B
	.dc.b	$FB,$87,$F7,$DF,$0B,$27,$F5,$ED,$9F,$DB,$35,$FD,$73,$5F,$E1,$4D
	.dc.b	$7F,$81,$B6,$BF,$88,$FC,$2F,$F0,$B2,$57,$EF,$1B,$5F,$C0,$DA,$FE
	.dc.b	$ED,$AF,$F1,$76,$BF,$C4,$1F,$E2,$DF,$0B,$27,$F8,$3B,$67,$F8,$73
	.dc.b	$5F,$E0,$CD,$7F,$91,$35,$FE,$36,$DA,$FE,$53,$F0,$BF,$C8,$C9,$5F
	.dc.b	$C5,$1B,$5F,$C6,$DA,$FF,$13,$6B,$FC,$BD,$AF,$F2,$87,$F9,$77,$C2
	.dc.b	$C9,$FE,$3E,$D9,$FE,$4C,$D7,$F8,$F3,$5F,$E7,$4D,$7F,$99,$B6,$BF
	.dc.b	$A0,$FC,$2F,$F3,$B2,$57,$F2,$C6,$D7,$F3,$36,$BF,$CA,$DA,$FF,$47
	.dc.b	$6B,$FD,$01,$FE,$8D,$F0,$B2,$7F,$9B,$B6,$7F,$9F,$35,$FE,$6C,$D7
	.dc.b	$FA,$93,$5F,$E9,$6D,$AF,$EB,$3E,$57,$F5,$32,$57,$F4,$46,$D7,$F4
	.dc.b	$B6,$BF,$D0,$DA,$FF,$5F,$6B,$FD,$61,$FE,$BD,$F0,$B2,$7F,$A7,$B6
	.dc.b	$7F,$AB,$35,$FE,$9C,$D7,$FB,$53,$5F,$EC,$6D,$AF,$EE,$3E,$7F,$F6
	.dc.b	$B2,$57,$F5,$C6,$D7,$F6,$36,$BF,$D6,$DA,$FF,$77,$6B,$FD,$C1,$B3
	.dc.b	$DA,$7A,$12,$7A,$35,$E8,$D7,$A3,$5E,$8D,$8F,$F0,$4F,$C2,$FF,$80
	.dc.b	$C9,$1F,$DD,$1B,$3F,$DE,$DA,$FF,$73,$6B,$FE,$0F,$6B,$FE,$08,$6C
	.dc.b	$F6,$9E,$84,$9E,$8D,$7A,$35,$E8,$D7,$A3,$63,$A4,$72,$21,$C8,$A7
	.dc.b	$22,$9C,$8A,$F2,$78,$10,$FF,$17,$F8,$0F,$84,$FF,$3D,$12,$BF,$CF
	.dc.b	$CD,$9F,$F3,$BB,$5F,$F1,$9C,$F1,$9D,$92,$36,$D7,$73,$C5,$F0,$B5
	.dc.b	$09,$E8,$F2,$21,$CA,$7A,$3C,$9A,$F2,$8E,$2E,$CE,$26,$CA,$6E,$7C
	.dc.b	$4F,$85,$B1,$CA,$38,$1C,$C8,$60,$87,$39,$F1,$BC,$DB,$1C,$E3,$47
	.dc.b	$88,$D1,$5D,$57,$F9,$8B,$E1,$7F,$CB,$84,$AF,$F2,$6B,$6B,$FC,$A8
	.dc.b	$D7,$FC,$98,$D7,$FC,$D0,$D7,$FC,$C2,$A3,$FC,$BD,$F9,$ED,$69,$99
	.dc.b	$FF,$37,$46,$5B,$19,$7F,$CD,$D1,$96,$C6,$5F,$BB,$49,$D4,$61,$AF
	.dc.b	$FA,$05,$B1,$B9,$E5,$39,$6C,$65,$F4,$E1,$8A,$F9,$24,$70,$48,$D9
	.dc.b	$2B,$DB,$19,$6C,$65,$FF,$48,$FC,$93,$E1,$BF,$CF,$91,$FE,$8B,$8C
	.dc.b	$77,$7F,$A3,$61,$18,$75,$3C,$1E,$6D,$7F,$D4,$31,$B7,$33,$CE,$31
	.dc.b	$1A,$9E,$0F,$EA,$23,$2D,$8C,$BE,$26,$C3,$E1,$9E,$BF,$56,$D7,$AD
	.dc.b	$AF,$C0,$6B,$E6,$6B,$E1,$6B,$F0,$1B,$3D,$E7,$B1,$27,$B3,$5F,$C4
	.dc.b	$E3,$F4,$5B,$8F,$D1,$7F,$0D,$19,$6C,$65,$FC,$44,$65,$B1,$9C,$FF
	.dc.b	$0D,$1F,$98,$8F,$CE,$91,$C1,$0F,$CC,$B9,$69,$93,$F3,$0F,$E3,$CF
	.dc.b	$D6,$47,$ED,$DF,$DA,$92,$3F,$6A,$D7,$F6,$AD,$7F,$6A,$D7,$F6,$AD
	.dc.b	$7F,$82,$9F,$C4,$0F,$E1,$25,$32,$7F,$12,$29,$A3,$1A,$FF,$11,$72
	.dc.b	$47,$04,$AF,$E4,$24,$9F,$E3,$2D,$9F,$E2,$CD,$7C,$53,$C6,$76,$48
	.dc.b	$DB,$5D,$CF,$17,$CA,$D7,$D0,$4A,$FE,$B4,$FF,$5F,$3F,$AD,$1F,$D7
	.dc.b	$CA,$E9,$C1,$B2,$18,$23,$B2,$78,$2D,$FF,$BE,$B6,$FF,$BA,$36,$BF
	.dc.b	$B6,$B5,$F0,$B5,$EF,$5D,$2B,$99,$2B,$9B,$67,$9B,$E0,$E0,$FF,$12
	.dc.b	$D9,$A4,$AF,$F0,$F4,$65,$B1,$9A,$FF,$14,$90,$FF,$14,$94,$FF,$14
	.dc.b	$94,$FF,$14,$96,$FF,$C6,$8D,$AE,$CF,$F9,$32,$BF,$C9,$E4,$70,$43
	.dc.b	$FC,$9E,$E5,$A6,$4F,$F2,$8B,$96,$9F,$E6,$37,$FC,$CC,$47,$04,$3F
	.dc.b	$CC,$CE,$5A,$64,$FF,$34,$39,$69,$FE,$7C,$7F,$CF,$E4,$70,$43,$FC
	.dc.b	$FE,$E5,$A6,$4F,$F4,$0B,$96,$9F,$E9,$37,$FD,$2C,$47,$04,$3F,$D2
	.dc.b	$CE,$5A,$64,$FF,$4C,$39,$69,$FE,$AC,$7F,$D5,$E4,$70,$43,$FD,$5E
	.dc.b	$E5,$A6,$4F,$F5,$8B,$96,$9F,$EC,$37,$FD,$8C,$47,$04,$3F,$D8,$CE
	.dc.b	$5A,$64,$FF,$64,$39,$69,$FE,$DC,$7F,$DB,$E4,$70,$43,$FD,$BE,$E5
	.dc.b	$A6,$4F,$F7,$0B,$96,$9F,$EF,$37,$FD,$EC,$47,$04,$3F,$DE,$CE,$5A
	.dc.b	$64,$FF,$7C,$39,$69,$FE,$56,$3F,$C3,$70,$E9,$FF,$09,$B6,$75,$3A
	.dc.b	$9C,$92,$33,$0A,$57,$18,$E3,$18,$AE,$11,$89,$D3,$E3,$9D,$57,$46
	.dc.b	$C6,$EB,$CB,$D5,$2B,$56,$8C,$57,$56,$BD,$63,$53,$C2,$73,$3F,$09
	.dc.b	$AF,$AE,$76,$F9,$FA,$FE,$29,$2B,$C2,$D8,$D4,$60,$8E,$08,$78,$E3
	.dc.b	$83,$E5,$EC,$FA,$24,$9F,$6B,$63,$E4,$35,$E1,$1A,$23,$A6,$BF,$07
	.dc.b	$B7,$EE,$92,$3E,$4B,$63,$EA,$35,$E1,$1A,$23,$A6,$BF,$37,$BB,$F1
	.dc.b	$24,$8F,$AA,$D8,$FC,$03,$5E,$11,$A2,$3A,$6B,$F6,$7B,$FF,$2A,$48
	.dc.b	$FC,$0B,$63,$F1,$8D,$78,$46,$88,$E9,$AF,$E1,$7C,$0F,$D0,$92,$3F
	.dc.b	$1A,$D8,$FC,$C3,$5E,$11,$A2,$3A,$6B,$F9,$1F,$07,$F5,$A4,$8F,$CC
	.dc.b	$B6,$3F,$48,$D7,$84,$68,$8E,$9A,$FE,$77,$9F,$F7,$24,$8F,$D2,$B6
	.dc.b	$3F,$60,$D7,$84,$68,$8E,$9A,$FE,$A4,$FE,$0A,$47,$F6,$2D,$3F,$78
	.dc.b	$53,$83,$A2,$3A,$6B,$FC,$07,$C9,$78,$7F,$EA,$46,$CF,$F0,$36,$CF
	.dc.b	$F0,$C6,$BC,$67,$81,$27,$83,$63,$83,$5E,$11,$D7,$C4,$F0,$11,$C1
	.dc.b	$23,$C1,$19,$6C,$65,$F5,$46,$5B,$1E,$9F,$17,$D8,$47,$04,$8F,$64
	.dc.b	$65,$B1,$97,$E3,$46,$5B,$1F,$17,$C6,$F9,$04,$70,$48,$F9,$11,$96
	.dc.b	$C6,$5F,$A5,$19,$6C,$7D,$1F,$1F,$EA,$11,$C1,$23,$EA,$46,$5B,$19
	.dc.b	$7E,$F4,$65,$B1,$FE,$CF,$E9,$79,$1F,$EE,$69,$46,$27,$53,$98,$C4
	.dc.b	$FF,$52,$6B,$FD,$61,$AF,$F5,$F6,$A3,$AC,$73,$24,$F6,$24,$7E,$3F
	.dc.b	$C9,$FC,$A1,$23,$F2,$8D,$8D,$46,$08,$E0,$87,$E7,$E3,$83,$63,$6D
	.dc.b	$36,$79,$FA,$1A,$AB,$8C,$E6,$31,$5C,$A3,$13,$A7,$FB,$B4,$E9,$EA
	.dc.b	$62,$35,$5D,$9B,$3B,$9F,$21,$0E,$B5,$DD,$E4,$DA,$E4,$D9,$E5,$3C
	.dc.b	$0F,$0C,$E9,$B1,$A7,$99,$8A,$F1,$92,$BA,$36,$35,$18,$23,$82,$1E
	.dc.b	$93,$A9,$6B,$C4,$DA,$F1,$46,$23,$D0,$7F,$17,$F2,$BF,$8B,$12,$BF
	.dc.b	$8B,$36,$3F,$8B,$35,$E1,$1A,$23,$A6,$BA,$30,$43,$04,$3F,$01,$1D
	.dc.b	$5B,$1C,$9A,$71,$29,$C4,$FC,$1C,$71,$6C,$71,$69,$C4,$A7,$19,$FC
	.dc.b	$9C,$F1,$6C,$FF,$37,$F2,$FF,$9D,$11,$C1,$2B,$F9,$D4,$65,$B1,$97
	.dc.b	$FA,$24,$65,$B1,$97,$F4,$11,$96,$C6,$5F,$CF,$C6,$5B,$19,$9F,$D5
	.dc.b	$CE,$5B,$1F,$D3,$FC,$CF,$EA,$44,$70,$4A,$FE,$A5,$19,$6C,$65,$FE
	.dc.b	$B9,$19,$6C,$7F,$5B,$F3,$7F,$AF,$11,$C1,$23,$FA,$F4,$65,$B1,$97
	.dc.b	$FB,$44,$65,$B1,$FD,$9F,$CE,$FE,$D4,$47,$04,$8F,$ED,$51,$96,$C6
	.dc.b	$5F,$EE,$91,$96,$C7,$F7,$3F,$3F,$FB,$B1,$1C,$12,$3F,$BB,$46,$5B
	.dc.b	$19,$7F,$C0,$11,$96,$C7,$F7,$FF,$43,$FC,$04,$47,$04,$8F,$F0,$14
	.dc.b	$65,$B1,$97,$FC,$19,$19,$6C,$7F,$82,$FD,$1F,$F0,$71,$1C,$12,$3F
	.dc.b	$C1,$D1,$96,$C6,$5F,$F0,$C4,$65,$B1,$97,$F9,$A4,$65,$B1,$97,$F9
	.dc.b	$9C,$65,$B1,$9C,$BF,$9F,$CE,$5B,$1F,$E2,$3F,$4B,$FC,$4C,$47,$04
	.dc.b	$AF,$F1,$34,$65,$B1,$97,$FC,$61,$19,$6C,$7F,$8B,$FD,$3F,$F1,$91
	.dc.b	$1C,$12,$3F,$C6,$51,$96,$C6,$5F,$F1,$E4,$65,$B1,$FE,$C8,$5E,$A3
	.dc.b	$B7,$FC,$00,$D9,$D3,$67,$FD,$9E,$D7,$FD,$AA,$D7,$FD,$B8,$D7,$FB
	.dc.b	$F3,$5E,$F3,$D8,$93,$D9,$B1,$DB,$CE,$F1,$92,$7C,$6D,$8D,$46,$08
	.dc.b	$E0,$87,$9A,$38,$36,$36,$D3,$67,$AA,$36,$D8,$DB,$4D,$9E,$E8,$DB
	.dc.b	$63,$6D,$36,$7D,$08,$DB,$63,$6D,$37,$49,$DB,$67,$6D,$9D,$CE,$89
	.dc.b	$3A,$6C,$6A,$30,$47,$04,$3E,$21,$0F,$80,$53,$D8,$53,$D0,$59,$F2
	.dc.b	$36,$7C,$2D,$9F,$03,$4F,$01,$D8,$87,$62,$9D,$8B,$3D,$9B,$3D,$9B
	.dc.b	$3D,$9E,$78,$FF,$8E,$BD,$5F,$F4,$F1,$27,$FD,$3C,$D9,$FF,$4F,$35
	.dc.b	$FF,$53,$35,$FF,$20,$35,$FF,$1E,$B6,$7B,$4F,$42,$4F,$46,$BD,$1A
	.dc.b	$F4,$6B,$D1,$B1,$D2,$39,$10,$E4,$53,$91,$4E,$45,$79,$3C,$08,$7F
	.dc.b	$96,$FD,$6F,$F6,$11,$27,$FD,$84,$D8,$FF,$61,$35,$FF,$65,$35,$E5
	.dc.b	$1C,$08,$F0,$6B,$C1,$AF,$03,$44,$34,$53,$46,$08,$7F,$9C,$FD,$7F
	.dc.b	$F6,$D1,$23,$FD,$B4,$D8,$FF,$6D,$35,$FF,$71,$35,$E5,$1C,$08,$FF
	.dc.b	$22,$F2,$3F,$93,$92,$3F,$93,$B6,$35,$18,$23,$82,$1F,$CA,$A3,$83
	.dc.b	$FE,$F3,$E9,$FF,$79,$12,$7F,$DE,$4D,$8F,$F7,$93,$5E,$11,$A2,$3A
	.dc.b	$6B,$FE,$FF,$F2,$FF,$DF,$C4,$8F,$F7,$F3,$63,$FD,$FC,$D7,$84,$68
	.dc.b	$8E,$9A,$FF,$AF,$46,$2E,$67,$B0,$AD,$57,$FA,$B6,$73,$18,$8E,$31
	.dc.b	$87,$75,$CE,$B9,$46,$27,$4F,$86,$75,$5C,$DB,$1B,$AE,$2E,$EB,$8B
	.dc.b	$E3,$8C,$B6,$B9,$4E,$A3,$CB,$EC,$E9,$B3,$A8,$C5,$FC,$16,$D1,$A4
	.dc.b	$9E,$04,$AF,$64,$65,$B1,$E2,$F6,$B9,$91,$C1,$23,$E2,$46,$5B,$19
	.dc.b	$7E,$04,$65,$B1,$EE,$F6,$FC,$E4,$70,$48,$FA,$11,$96,$C6,$5F,$99
	.dc.b	$19,$6C,$7C,$AF,$73,$E1,$91,$C1,$23,$EE,$46,$5B,$19,$7E,$C4,$65
	.dc.b	$B1,$F5,$BD,$DF,$9E,$47,$04,$8F,$C4,$46,$5B,$19,$7F,$09,$19,$6C
	.dc.b	$7E,$0B,$DE,$FB,$64,$70,$48,$FC,$A4,$65,$B1,$97,$F2,$11,$96,$C7
	.dc.b	$E3,$BD,$FF,$C3,$91,$C1,$23,$F4,$11,$96,$C6,$5F,$CE,$46,$5B,$1F
	.dc.b	$9A,$F8,$0F,$C9,$91,$C1,$23,$F5,$91,$96,$C6,$5F,$D4,$46,$5B,$1F
	.dc.b	$A6,$F8,$1F,$CF,$91,$C1,$23,$F7,$11,$96,$C6,$5F,$DA,$46,$5B,$19
	.dc.b	$7F,$84,$46,$5B,$19,$7F,$84,$C6,$5B,$19,$DB,$F8,$A4,$E5,$B1,$FB
	.dc.b	$EF,$82,$FD,$B9,$1C,$12,$BF,$88,$46,$5B,$19,$7F,$86,$46,$5B,$1F
	.dc.b	$C2,$BE,$0F,$F8,$19,$1C,$12,$3F,$8E,$46,$5B,$19,$7F,$8C,$46,$5B
	.dc.b	$1F,$C5,$BE,$13,$F8,$79,$1C,$12,$3F,$94,$46,$5B,$19,$7F,$92,$46
	.dc.b	$5B,$1F,$C8,$BE,$17,$F8,$D9,$1C,$12,$3F,$9A,$46,$5B,$19,$7F,$98
	.dc.b	$46,$5B,$1F,$CB,$BE,$1B,$F9,$39,$1C,$12,$3F,$A0,$46,$5B,$19,$7F
	.dc.b	$9E,$46,$5B,$1F,$CE,$BC,$9F,$E6,$64,$70,$48,$FE,$99,$19,$6C,$65
	.dc.b	$FE,$91,$19,$6C,$7F,$46,$F7,$3F,$9F,$91,$C1,$23,$FA,$C4,$65,$B1
	.dc.b	$97,$FA,$A4,$65,$B1,$FD,$4B,$DD,$FE,$96,$47,$04,$8F,$EC,$91,$96
	.dc.b	$C6,$5F,$EC,$11,$96,$C7,$F5,$EF,$87,$FE,$AE,$47,$04,$8F,$EE,$11
	.dc.b	$96,$C6,$5F,$ED,$91,$96,$C7,$F6,$AF,$77,$FB,$19,$1C,$12,$3F,$BE
	.dc.b	$46,$5B,$19,$7F,$BC,$46,$5B,$1F,$DD,$8F,$ED,$E4,$30,$47,$FC,$10
	.dc.b	$E5,$AE,$5F,$F0,$23,$96,$BF,$E0,$BF,$88,$7C,$0F,$F7,$E6,$C6,$63
	.dc.b	$04,$70,$49,$FF,$0D,$CF,$06,$CE,$DB,$3B,$7F,$C3,$92,$F7,$FC,$04
	.dc.b	$9F,$23,$67,$AB,$5E,$2D,$47,$F8,$DA,$7A,$4F,$22,$47,$9B,$BF,$DC
	.dc.b	$48,$F8,$4D,$8D,$46,$08,$E0,$87,$AA,$38,$3E,$CF,$82,$F9,$44,$9F
	.dc.b	$9C,$D8,$F8,$ED,$78,$46,$88,$E9,$AF,$C8,$F1,$3E,$B1,$23,$ED,$36
	.dc.b	$3E,$9B,$5E,$11,$A2,$3A,$6B,$F5,$3E,$0F,$F0,$44,$8F,$C3,$36,$3E
	.dc.b	$FB,$5E,$11,$A2,$3A,$6B,$F8,$0F,$84,$FC,$71,$23,$F2,$4D,$8F,$C5
	.dc.b	$B5,$E1,$1A,$23,$A6,$BF,$8C,$F8,$5F,$CD,$12,$3F,$3C,$D8,$FC,$BB
	.dc.b	$5E,$11,$A2,$3A,$6B,$FE,$7E,$F2,$3F,$1A,$FF,$9F,$5A,$E7,$4F,$F3
	.dc.b	$52,$8C,$BF,$E8,$B6,$BF,$98,$6B,$B8,$E3,$18,$23,$82,$4F,$EE,$7E
	.dc.b	$1B,$F7,$04,$70,$4A,$FD,$AC,$65,$B1,$97,$F6,$71,$96,$C7,$F0,$5F
	.dc.b	$87,$FE,$08,$47,$04,$8F,$E0,$31,$96,$C6,$5F,$DF,$C6,$5B,$1F,$C4
	.dc.b	$7E,$23,$F8,$81,$1C,$12,$3F,$86,$C6,$5B,$19,$7F,$85,$C6,$5B,$1F
	.dc.b	$C7,$7E,$1B,$F8,$E1,$1C,$12,$3F,$8C,$C6,$5B,$19,$7F,$8B,$C6,$5B
	.dc.b	$1F,$CA,$7E,$1B,$F9,$41,$1C,$12,$3F,$92,$C6,$5B,$19,$7F,$91,$C6
	.dc.b	$5B,$1F,$CD,$7E,$17,$F9,$A1,$1C,$12,$3F,$98,$C6,$5B,$19,$7F,$97
	.dc.b	$C6,$5B,$1F,$D0,$7E,$17,$FA,$01,$1C,$12,$3F,$9E,$C6,$5B,$19,$7F
	.dc.b	$9D,$C6,$5B,$1F,$D3,$7E,$17,$FA,$61,$1C,$12,$3F,$A4,$C6,$5B,$19
	.dc.b	$7F,$A3,$C6,$5B,$1F,$D6,$7D,$1F,$EB,$04,$70,$48,$FE,$AB,$19,$6C
	.dc.b	$65,$FE,$A7,$19,$6C,$7F,$65,$F8,$5F,$EC,$84,$70,$48,$FE,$C3,$19
	.dc.b	$6C,$65,$FE,$BF,$19,$6C,$7F,$71,$F8,$5F,$EE,$04,$70,$48,$FE,$DB
	.dc.b	$19,$6C,$65,$FE,$D7,$19,$6C,$7F,$7D,$F8,$5F,$EF,$84,$70,$48,$FE
	.dc.b	$F3,$19,$6C,$65,$FE,$EF,$19,$6C,$7F,$82,$7E,$17,$FC,$10,$47,$04
	.dc.b	$8F,$F0,$2C,$65,$B1,$97,$FC,$07,$19,$6C,$65,$FE,$FB,$E2,$BA,$95
	.dc.b	$B9,$CC,$62,$B8,$47,$F8,$0E,$74,$D8,$CD,$73,$9C,$B6,$3F,$C4,$93
	.dc.b	$FE,$23,$23,$82,$3F,$E2,$03,$25,$72,$FF,$87,$8C,$95,$FF,$0B,$B2
	.dc.b	$1F,$12,$F8,$65,$6A,$BE,$0C,$E6,$31,$0A,$E3,$5C,$23,$84,$62,$76
	.dc.b	$D8,$CD,$73,$6D,$71,$7C,$91,$96,$C6,$5F,$1B,$5F,$14,$6D,$CD,$79
	.dc.b	$9F,$EC,$5D,$BF,$D7,$09,$19,$E7,$FF,$24,$4F,$F6,$D9,$CB,$FD,$B5
	.dc.b	$AE,$67,$84,$62,$3F,$B6,$3E,$3C,$3E,$5F,$C3,$5C,$0D,$25,$FC,$48
	.dc.b	$CB,$63,$37,$F1,$9A,$FC,$46,$BF,$51,$ED,$F0,$D2,$7A,$36,$7E,$7B
	.dc.b	$5F,$9A,$D7,$EE,$B5,$F9,$6D,$BF,$C2,$FC,$37,$AC,$95,$E9,$6D,$7D
	.dc.b	$C6,$BF,$69,$AF,$E2,$9A,$FD,$87,$ED,$7C,$5B,$E2,$DC,$38,$6D,$FE
	.dc.b	$42,$75,$39,$24,$66,$77,$7F,$99,$AD,$CE,$6B,$73,$9A,$DB,$FA,$4A
	.dc.b	$DB,$99,$F2,$F2,$B9,$29,$30,$F2,$46,$5B,$19,$AF,$14,$E5,$B1,$97
	.dc.b	$C1,$19,$6C,$65,$FD,$BF,$C6,$3C,$4F,$F1,$EB,$6F,$F3,$CD,$9F,$DA
	.dc.b	$B5,$FD,$D3,$5E,$73,$C8,$93,$C9,$B1,$C9,$AF,$F0,$35,$F0,$3F,$E4
	.dc.b	$E6,$CF,$EB,$DB,$3F,$C1,$9A,$F1,$9E,$04,$9E,$0D,$8E,$0D,$78,$47
	.dc.b	$EE,$FE,$35,$E1,$7F,$97,$9B,$19,$8C,$11,$C1,$27,$F8,$CC,$F0,$6C
	.dc.b	$ED,$B3,$B7,$F9,$17,$C7,$3D,$CF,$F3,$6B,$67,$F8,$73,$67,$F9,$2B
	.dc.b	$5E,$33,$C0,$93,$C1,$B1,$C1,$AF,$08,$FE,$35,$D4,$D7,$FC,$FA,$D8
	.dc.b	$CC,$60,$8E,$09,$3F,$CE,$27,$83,$67,$6D,$9D,$BF,$D0,$7E,$3D,$EA
	.dc.b	$FF,$A3,$1B,$3F,$CB,$5B,$3F,$D1,$1A,$F1,$9E,$04,$9E,$0D,$8E,$0D
	.dc.b	$78,$47,$F3,$9F,$90,$FE,$79,$1F,$EE,$E6,$A6,$93,$AF,$FD,$42,$D9
	.dc.b	$E5,$18,$23,$82,$4F,$F5,$C9,$E0,$D9,$DB,$67,$6F,$F6,$59,$DB,$67
	.dc.b	$6D,$9D,$BF,$DC,$27,$6D,$9D,$B6,$76,$FB,$E3,$2D,$8C,$BF,$06,$32
	.dc.b	$D8,$CC,$F6,$8E,$07,$A4,$87,$94,$A7,$84,$A7,$03,$FC,$05,$F0,$0F
	.dc.b	$43,$FD,$9A,$DA,$FE,$DC,$D9,$FF,$02,$B5,$E3,$3C,$09,$3C,$1B,$1C
	.dc.b	$1A,$F0,$8D,$11,$D3,$5D,$18,$21,$82,$1F,$E2,$09,$F2,$36,$7B,$B6
	.dc.b	$7A,$34,$E6,$53,$98,$13,$D1,$C3,$5C,$CF,$67,$2D,$73,$3E,$07,$2D
	.dc.b	$72,$78,$8C,$94,$C9,$E4,$29,$E7,$3F,$C7,$FF,$22,$F8,$AF,$F1,$44
	.dc.b	$23,$13,$FE,$27,$6C,$7F,$8F,$9A,$FF,$90,$DA,$F4,$9E,$71,$88,$F8
	.dc.b	$8D,$7B,$46,$89,$3D,$DB,$1D,$DA,$F7,$6B,$DE,$38,$11,$E0,$D7,$83
	.dc.b	$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$6C,$79,$1B,$1D,$DA,$74,$29
	.dc.b	$CC,$A7,$33,$E2,$BC,$5A,$71,$29,$C4,$A7,$19,$F9,$C4,$9F,$24,$68
	.dc.b	$EE,$43,$04,$8E,$13,$E1,$6C,$78,$5A,$F8,$5A,$F8,$67,$BC,$62,$3B
	.dc.b	$35,$ED,$1A,$3F,$23,$F2,$4F,$42,$92,$FF,$CF,$11,$96,$C6,$6B,$FD
	.dc.b	$23,$39,$6C,$65,$FF,$4C,$46,$5B,$19,$7F,$D4,$31,$96,$C6,$5F,$DE
	.dc.b	$46,$5B,$19,$51,$96,$C6,$63,$04,$70,$49,$F5,$92,$7D,$2D,$7C,$ED
	.dc.b	$7C,$AD,$7C,$6D,$7C,$2D,$9F,$03,$67,$C1,$3D,$88,$76,$29,$D8,$A7
	.dc.b	$62,$9D,$8B,$3D,$9B,$1D,$9E,$84,$3A,$14,$E8,$53,$A1,$63,$A3,$63
	.dc.b	$A3,$C8,$87,$22,$9C,$8B,$1C,$9B,$1C,$9E,$04,$38,$16,$38,$36,$38
	.dc.b	$3A,$24,$69,$B1,$A8,$C1,$1C,$12,$7E,$A1,$27,$E4,$B5,$F8,$0D,$7D
	.dc.b	$2D,$7C,$8D,$7C,$2D,$9F,$03,$67,$C1,$3D,$88,$76,$29,$D8,$A7,$62
	.dc.b	$9D,$8F,$F8,$03,$F2,$2F,$53,$FC,$C9,$18,$6B,$FD,$94,$FE,$CD,$1B
	.dc.b	$E2,$FF,$7A,$36,$38,$46,$27,$8C,$71,$9D,$3C,$C3,$87,$51,$C1,$EE
	.dc.b	$D7,$6D,$8E,$11,$A9,$F0,$11,$F1,$90,$F0,$47,$07,$9C,$6D,$CC,$F5
	.dc.b	$24,$66,$3C,$AD,$7C,$07,$53,$25,$32,$74,$78,$3E,$03,$45,$34,$62
	.dc.b	$7C,$C4,$9F,$1C,$65,$AF,$02,$1E,$22,$9E,$23,$04,$30,$43,$C4,$64
	.dc.b	$A6,$63,$C5,$3E,$A6,$C7,$8E,$31,$3E,$48,$E5,$3A,$7C,$67,$42,$39
	.dc.b	$77,$1C,$5F,$03,$5D,$B6,$38,$46,$A7,$C2,$47,$C8,$43,$C3,$1C,$1E
	.dc.b	$71,$B7,$33,$D4,$91,$98,$F2,$B5,$F0,$1D,$4C,$94,$C9,$D1,$E0,$F8
	.dc.b	$0D,$14,$D1,$89,$F3,$12,$7C,$71,$96,$BC,$08,$78,$8A,$78,$8C,$10
	.dc.b	$C1,$0F,$11,$92,$99,$8F,$14,$FA,$9B,$1E,$38,$C4,$F9,$23,$94,$E9
	.dc.b	$F1,$9D,$08,$E5,$DC,$71,$7C,$0D,$76,$D8,$E1,$1A,$9F,$09,$1F,$21
	.dc.b	$0F,$0C,$70,$79,$C6,$DC,$CF,$52,$46,$63,$CA,$D7,$C0,$75,$32,$53
	.dc.b	$27,$47,$83,$E0,$34,$53,$46,$27,$CC,$49,$F1,$C6,$5A,$F0,$21,$E2
	.dc.b	$29,$FB,$DF,$93,$7B,$5F,$CC,$63,$F9,$E4,$61,$C0,$70,$ED,$AF,$F3
	.dc.b	$96,$BC,$A3,$8B,$C5,$B1,$A8,$FF,$57,$35,$D1,$E0,$70,$48,$E2,$48
	.dc.b	$D3,$5F,$23,$5F,$23,$5F,$21,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45
	.dc.b	$34,$60,$86,$08,$7C,$07,$C8,$D3,$B9,$4E,$85,$39,$94,$E6,$7C,$83
	.dc.b	$65,$36,$53,$71,$F2,$DE,$E6,$4E,$84,$30,$47,$83,$D9,$AF,$66,$BD
	.dc.b	$A3,$DA,$F1,$6C,$6A,$3B,$35,$D1,$E4,$70,$48,$E2,$48,$D3,$5F,$23
	.dc.b	$5F,$23,$5F,$21,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86
	.dc.b	$08,$7C,$07,$C8,$D3,$B9,$4E,$85,$39,$94,$E6,$7C,$83,$65,$36,$53
	.dc.b	$71,$F2,$DE,$E6,$4E,$84,$30,$47,$83,$D9,$AF,$66,$BD,$A3,$DA,$F1
	.dc.b	$6C,$6A,$3B,$35,$D1,$E4,$70,$48,$E2,$48,$D3,$5F,$23,$5F,$23,$5F
	.dc.b	$21,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$7C,$07
	.dc.b	$C8,$D3,$B9,$4E,$85,$39,$94,$E6,$7C,$83,$65,$36,$53,$71,$F2,$DE
	.dc.b	$E6,$4E,$84,$30,$47,$83,$D9,$AF,$66,$BD,$A3,$DA,$F1,$6C,$6A,$3B
	.dc.b	$35,$D1,$E4,$70,$48,$E2,$48,$D3,$5F,$23,$5F,$23,$5F,$21,$C0,$87
	.dc.b	$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$7C,$07,$C8,$D3,$B9
	.dc.b	$4E,$85,$39,$94,$E6,$7C,$83,$65,$36,$53,$71,$F2,$DE,$E6,$4E,$84
	.dc.b	$30,$47,$83,$D9,$AF,$66,$BD,$A3,$DA,$F1,$6C,$6A,$03,$D6,$78,$1F
	.dc.b	$E0,$86,$C7,$F8,$49,$AF,$F8,$5D,$AF,$FB,$25,$AF,$FB,$55,$AF,$66
	.dc.b	$BD,$E3,$B1,$1E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$1D,$08,$74
	.dc.b	$29,$D0,$A7,$42,$9D,$0A,$74,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C
	.dc.b	$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$91,$F5,$09,$1F
	.dc.b	$25,$B1,$F0,$1A,$FA,$5A,$F9,$1A,$F8,$5A,$F8,$1A,$F8,$23,$B1,$1E
	.dc.b	$CD,$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$1D,$08,$74,$29,$D0,$A7,$42
	.dc.b	$9D,$0A,$74,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70,$29,$C0
	.dc.b	$A7,$03,$44,$34,$53,$46,$08,$60,$91,$F5,$09,$1F,$25,$B1,$F0,$1A
	.dc.b	$FA,$5A,$F9,$1A,$F8,$5A,$F8,$1A,$F8,$23,$B1,$1E,$CD,$7B,$35,$EC
	.dc.b	$D7,$B3,$5E,$CD,$7B,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$74,$39
	.dc.b	$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34
	.dc.b	$53,$46,$08,$60,$91,$F5,$09,$1F,$25,$B1,$F0,$1A,$FA,$5A,$F9,$1A
	.dc.b	$F8,$5A,$F8,$1A,$F8,$23,$B1,$1E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$CD
	.dc.b	$7B,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$74,$39,$10,$E4,$53,$91
	.dc.b	$4E,$45,$39,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60
	.dc.b	$91,$F5,$09,$1F,$25,$B1,$F0,$1A,$FA,$5A,$F9,$1A,$F8,$5A,$F8,$1A
	.dc.b	$F8,$23,$B1,$1E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$1D,$08,$74
	.dc.b	$29,$D0,$A7,$42,$9D,$0A,$74,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C
	.dc.b	$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$91,$F5,$09,$1F
	.dc.b	$25,$B1,$F0,$1A,$FA,$5A,$F9,$1A,$F8,$5A,$F8,$1A,$F8,$23,$B1,$1E
	.dc.b	$CD,$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$1D,$08,$74,$29,$D0,$A7,$42
	.dc.b	$9D,$0A,$74,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70,$29,$C0
	.dc.b	$A7,$03,$44,$34,$53,$46,$08,$60,$91,$F5,$09,$1F,$25,$B1,$F0,$1A
	.dc.b	$FF,$59,$F9,$47,$AF,$FE,$8A,$88,$7F,$A3,$23,$2E,$0C,$45,$69,$63
	.dc.b	$6E,$DA,$F4,$6B,$DD,$AF,$23,$84,$75,$51,$A6,$C6,$63,$04,$70,$48
	.dc.b	$F2,$35,$EA,$60,$86,$08,$79,$DE,$0D,$36,$53,$67,$AD,$DB,$4D,$94
	.dc.b	$D9,$EF,$39,$10,$D1,$4D,$18,$21,$82,$1F,$10,$EE,$62,$3E,$23,$83
	.dc.b	$47,$88,$A7,$72,$9D,$8A,$73,$8F,$6B,$4E,$91,$D9,$E8,$D7,$A3,$5E
	.dc.b	$8D,$79,$1C,$23,$AB,$E0,$24,$6D,$B1,$A8,$C1,$1C,$12,$3C,$AD,$7B
	.dc.b	$18,$21,$82,$1E,$87,$83,$4D,$94,$D9,$EC,$76,$D3,$65,$36,$7C,$03
	.dc.b	$91,$0D,$14,$D1,$82,$18,$21,$F1,$0E,$E6,$23,$E2,$38,$34,$78,$8A
	.dc.b	$77,$29,$D8,$A7,$38,$F6,$B4,$E9,$1D,$9E,$8D,$7A,$35,$E8,$D7,$91
	.dc.b	$C2,$3A,$BE,$02,$46,$DB,$1A,$8C,$11,$C1,$23,$CA,$D7,$B1,$82,$18
	.dc.b	$21,$E8,$78,$34,$D9,$4D,$9E,$C7,$6D,$36,$53,$67,$C0,$39,$10,$D1
	.dc.b	$4D,$18,$21,$82,$1F,$10,$EE,$62,$3E,$23,$83,$47,$88,$A7,$72,$9D
	.dc.b	$8A,$73,$8F,$6B,$4E,$91,$D9,$E8,$D7,$A3,$5E,$8D,$79,$1C,$23,$AB
	.dc.b	$E0,$24,$6D,$B1,$A8,$C1,$1C,$12,$3C,$AD,$7B,$18,$21,$82,$1E,$87
	.dc.b	$83,$4D,$94,$D9,$EC,$76,$D3,$65,$36,$7C,$03,$91,$0D,$14,$D1,$82
	.dc.b	$18,$21,$F1,$0E,$E6,$23,$E2,$38,$34,$78,$8A,$77,$29,$D8,$A7,$38
	.dc.b	$F6,$B4,$E9,$1D,$9E,$8D,$7A,$35,$E8,$D7,$91,$C2,$3A,$BE,$02,$46
	.dc.b	$DB,$1F,$E0,$5F,$95,$7C,$6D,$24,$FF,$85,$63,$2D,$8C,$BF,$E1,$B8
	.dc.b	$CB,$63,$2F,$F8,$8A,$32,$D8,$CB,$FE,$27,$F3,$5E,$BF,$F8,$5D,$B1
	.dc.b	$FE,$25,$6B,$FE,$29,$6B,$FE,$2E,$6B,$CE,$39,$11,$E4,$D7,$93,$5F
	.dc.b	$F1,$BF,$CB,$3E,$27,$FC,$50,$D8,$FF,$1B,$35,$FF,$1D,$35,$E3,$1C
	.dc.b	$08,$F0,$6B,$C1,$AF,$03,$44,$34,$53,$46,$08,$60,$87,$F9,$46,$3C
	.dc.b	$8D,$8E,$ED,$3A,$14,$E6,$53,$98,$11,$D1,$C3,$5C,$C7,$67,$2D,$72
	.dc.b	$78,$0C,$94,$C9,$E2,$32,$53,$27,$90,$A7,$9C,$87,$03,$51,$E5,$24
	.dc.b	$78,$DB,$1E,$16,$BD,$DA,$F2,$8E,$04,$78,$35,$E0,$D7,$81,$A2,$1A
	.dc.b	$29,$A3,$04,$30,$43,$DA,$7C,$03,$DC,$62,$3C,$ED,$8F,$13,$4E,$E5
	.dc.b	$3B,$14,$EC,$74,$21,$DC,$91,$D9,$CB,$5C,$C7,$81,$CB,$5C,$9E,$23
	.dc.b	$25,$32,$79,$0C,$94,$C9,$E6,$29,$E9,$21,$C0,$D4,$79,$7E,$5D,$E9
	.dc.b	$FF,$A8,$DB,$3F,$EB,$16,$BF,$EB,$56,$BF,$EB,$A6,$BF,$EB,$D6,$BF
	.dc.b	$A5,$6A,$69,$23,$C0,$48,$F0,$35,$F0,$35,$F0,$35,$F0,$35,$F0,$35
	.dc.b	$F0,$36,$3C,$11,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$A7,$62,$BD,$9E
	.dc.b	$84,$3A,$14,$E8,$53,$A1,$4E,$85,$7A,$3C,$88,$72,$29,$C8,$A7,$22
	.dc.b	$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70,$48,$FA
	.dc.b	$84,$8F,$92,$D7,$E0,$35,$F4,$B5,$F2,$35,$F0,$B5,$F0,$36,$3C,$11
	.dc.b	$D8,$87,$62,$9D,$8A,$76,$29,$D8,$A7,$62,$BD,$9E,$84,$3A,$14,$E8
	.dc.b	$53,$A1,$4E,$85,$7A,$3C,$88,$72,$29,$C8,$A7,$22,$BC,$9E,$04,$38
	.dc.b	$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70,$48,$FA,$84,$8F,$92,$D7
	.dc.b	$E0,$35,$F4,$B5,$F2,$35,$F0,$B5,$F0,$36,$3C,$11,$D8,$87,$62,$9D
	.dc.b	$8A,$76,$29,$D8,$A7,$62,$BD,$9E,$84,$3A,$14,$E8,$53,$FC,$19,$F1
	.dc.b	$8F,$84,$FF,$07,$44,$32,$D7,$31,$B7,$01,$C3,$B6,$BC,$DA,$F2,$8E
	.dc.b	$2F,$16,$C6,$A3,$C2,$76,$70,$EC,$C4,$70,$23,$82,$1E,$37,$2D,$32
	.dc.b	$79,$5C,$B4,$C9,$E7,$29,$E4,$39,$10,$E0,$53,$81,$4E,$06,$88,$68
	.dc.b	$A6,$8C,$10,$C1,$0F,$84,$75,$29,$C8,$A7,$12,$9C,$63,$E5,$92,$3E
	.dc.b	$0C,$6E,$30,$FC,$03,$89,$1C,$BE,$16,$BE,$16,$BC,$E3,$AB,$E1,$6B
	.dc.b	$CD,$B1,$C2,$34,$78,$5C,$3B,$31,$1C,$08,$E0,$87,$95,$CB,$4C,$9E
	.dc.b	$77,$2D,$32,$7A,$4A,$79,$0E,$44,$38,$14,$E0,$53,$81,$A2,$1A,$29
	.dc.b	$A3,$04,$30,$43,$E1,$1D,$4A,$72,$29,$C4,$A7,$1F,$98,$FD,$8A,$69
	.dc.b	$1F,$DA,$B5,$FD,$FB,$5F,$E0,$8D,$7F,$84,$35,$FE,$17,$F3,$6F,$81
	.dc.b	$FD,$FB,$63,$F8,$4B,$5F,$E1,$CD,$7F,$89,$35,$E7,$1C,$88,$F2,$6B
	.dc.b	$C9,$AF,$F1,$CF,$39,$F3,$4F,$8F,$FE,$33,$18,$6C,$66,$7F,$90,$4E
	.dc.b	$5B,$19,$7F,$93,$46,$5B,$19,$7C,$0D,$7F,$D7,$51,$B7,$33,$E0,$23
	.dc.b	$E0,$F9,$B7,$81,$E6,$6C,$66,$30,$47,$04,$3C,$B1,$C1,$B1,$B6,$9B
	.dc.b	$3F,$A0,$FA,$6F,$59,$F3,$3F,$CF,$A3,$0D,$8C,$CF,$F4,$59,$CB,$63
	.dc.b	$2F,$F4,$B8,$CB,$63,$2F,$81,$AF,$FB,$7E,$36,$E6,$7C,$04,$7B,$7C
	.dc.b	$CF,$8C,$8E,$08,$77,$72,$D3,$27,$85,$CB,$4C,$9E,$3A,$46,$A3,$0D
	.dc.b	$55,$F2,$63,$0D,$8E,$66,$A7,$9B,$88,$D3,$E5,$21,$E3,$29,$E0,$38
	.dc.b	$47,$27,$27,$59,$F0,$C6,$63,$91,$0E,$47,$09,$F0,$B9,$39,$1D,$A7
	.dc.b	$99,$1C,$10,$E5,$3D,$DE,$A7,$39,$EA,$E0,$87,$63,$8C,$F5,$69,$D4
	.dc.b	$E5,$3D,$5E,$06,$23,$83,$4E,$07,$69,$F0,$BC,$0C,$4F,$38,$C9,$1C
	.dc.b	$12,$7A,$CF,$06,$BC,$C9,$3C,$63,$2D,$8F,$19,$A9,$E8,$E2,$34,$FC
	.dc.b	$02,$1E,$B2,$9E,$13,$84,$72,$72,$75,$9F,$0C,$66,$39,$10,$E4,$70
	.dc.b	$9F,$0B,$93,$91,$DA,$79,$91,$C1,$0E,$53,$DD,$F9,$BF,$31,$FC,$1C
	.dc.b	$8F,$F0,$56,$BF,$C0,$DA,$FF,$03,$6B,$FB,$16,$BF,$A9,$6D,$7F,$17
	.dc.b	$6C,$9A,$49,$F0,$10,$F0,$14,$F0,$14,$F0,$14,$F0,$16,$7C,$0D,$9F
	.dc.b	$03,$67,$C0,$F6,$21,$D8,$A7,$62,$9D,$8B,$3D,$9B,$3D,$9B,$3D,$9E
	.dc.b	$84,$3A,$14,$E8,$59,$E8,$D9,$E8,$D9,$E8,$F2,$21,$C8,$B3,$C9,$B3
	.dc.b	$C9,$B3,$C9,$E0,$49,$E0,$D9,$E0,$D9,$E1,$3A,$24,$E9,$B1,$A8,$C1
	.dc.b	$1C,$12,$7E,$A1,$0F,$92,$53,$E0,$14,$F4,$94,$F2,$16,$7C,$2D,$9F
	.dc.b	$03,$67,$C0,$F6,$21,$D8,$A7,$62,$9D,$8B,$3D,$9B,$3D,$9B,$3D,$9E
	.dc.b	$84,$3A,$14,$E8,$59,$E8,$D9,$E8,$D9,$E8,$F2,$21,$C8,$B3,$C9,$B3
	.dc.b	$C9,$B3,$C9,$E0,$49,$E0,$D9,$E0,$D9,$E1,$3A,$24,$E9,$B1,$A8,$C1
	.dc.b	$1C,$12,$7E,$A1,$0F,$92,$53,$E0,$14,$F4,$94,$F2,$16,$7C,$2D,$9F
	.dc.b	$03,$67,$C0,$F6,$21,$D8,$A7,$62,$9D,$8B,$3D,$9B,$3F,$A6,$5B,$3E
	.dc.b	$27,$FC,$DA,$48,$FF,$36,$B5,$FF,$36,$B5,$FF,$36,$B5,$FF,$3B,$B5
	.dc.b	$1D,$9C,$34,$EA,$43,$A9,$4E,$A5,$3A,$94,$EB,$1E,$97,$07,$53,$99
	.dc.b	$0E,$65,$39,$94,$E7,$1E,$68,$E9,$19,$3A,$1C,$88,$72,$29,$C8,$F1
	.dc.b	$BC,$9C,$9C,$8E,$04,$38,$1D,$DE,$0E,$4E,$06,$8E,$6E,$9C,$1B,$70
	.dc.b	$F0,$23,$82,$47,$BC,$91,$E8,$6B,$E3,$6B,$E0,$6B,$DD,$AF,$88,$87
	.dc.b	$81,$CB,$4E,$C4,$3B,$14,$EC,$53,$B1,$4E,$B1,$EA,$70,$75,$39,$90
	.dc.b	$E6,$53,$99,$4E,$71,$E6,$8E,$91,$93,$A1,$C8,$87,$22,$9C,$8F,$1B
	.dc.b	$C9,$C9,$C8,$E0,$43,$81,$DD,$E0,$E4,$E0,$7F,$A0,$7E,$4B,$FD,$54
	.dc.b	$47,$FD,$50,$D7,$FD,$4C,$D7,$FD,$4C,$D7,$FD,$28,$D7,$A1,$FE,$B1
	.dc.b	$CF,$FD,$6C,$47,$04,$3F,$D6,$4E,$5A,$64,$FF,$5A,$39,$69,$93,$FD
	.dc.b	$48,$E5,$A6,$4F,$F4,$F3,$96,$99,$C3,$FD,$9D,$19,$6C,$67,$0F,$F6
	.dc.b	$C3,$E0,$3C,$A4,$30,$43,$C2,$64,$A6,$4F,$09,$FE,$D1,$F9,$8F,$F6
	.dc.b	$D1,$1F,$F6,$C3,$5D,$18,$21,$82,$1F,$ED,$97,$83,$FE,$DB,$F9,$8F
	.dc.b	$F7,$31,$1F,$F7,$23,$5F,$F7,$13,$5E,$06,$88,$68,$A7,$FB,$9F,$B3
	.dc.b	$FD,$E4,$47,$FD,$E0,$D7,$FD,$DC,$D7,$81,$A2,$1A,$29,$FE,$F3,$F5
	.dc.b	$BF,$DF,$C4,$7F,$DF,$8D,$7F,$DF,$4D,$78,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$3F,$DE,$0F,$56,$9C,$8A,$71,$29,$C4,$FF,$77,$3C,$5A,$71,$29
	.dc.b	$C4,$A7,$1C,$18,$27,$94,$61,$B1,$99,$E9,$39,$6C,$65,$ED,$19,$6C
	.dc.b	$65,$F0,$35,$F5,$90,$F2,$96,$39,$3C,$27,$C6,$47,$C2,$D7,$BB,$5E
	.dc.b	$06,$88,$68,$A6,$8C,$10,$C1,$0F,$24,$FA,$5F,$39,$27,$CE,$EA,$7C
	.dc.b	$4D,$3B,$94,$EC,$53,$B1,$CE,$7B,$4F,$48,$C4,$F4,$69,$D0,$A7,$42
	.dc.b	$9D,$23,$91,$27,$A9,$27,$A4,$65,$B1,$99,$ED,$39,$6C,$65,$F0,$46
	.dc.b	$5B,$19,$7C,$4D,$7E,$01,$0F,$39,$63,$F2,$BF,$31,$FE,$36,$88,$23
	.dc.b	$2E,$0C,$4A,$71,$18,$6C,$F0,$8E,$6D,$7B,$35,$FF,$22,$B5,$E4,$77
	.dc.b	$70,$49,$F0,$46,$5A,$F6,$6B,$D9,$AF,$66,$BD,$8E,$13,$E8,$8C,$CF
	.dc.b	$78,$C1,$27,$C0,$D7,$89,$82,$18,$21,$E3,$78,$34,$D9,$4D,$9E,$77
	.dc.b	$6D,$36,$53,$67,$AC,$E4,$43,$45,$34,$60,$86,$09,$1F,$05,$EE,$62
	.dc.b	$7E,$29,$27,$CB,$19,$6C,$F1,$8F,$2B,$5F,$13,$5F,$0B,$5E,$47,$81
	.dc.b	$C1,$27,$C1,$19,$6B,$D9,$AF,$66,$BD,$9A,$F6,$38,$4F,$A2,$33,$3D
	.dc.b	$E3,$04,$9F,$03,$5E,$26,$08,$60,$87,$8D,$E0,$D3,$65,$36,$79,$DD
	.dc.b	$B4,$D9,$4D,$9E,$B3,$91,$0D,$14,$D1,$82,$18,$24,$7C,$17,$B9,$89
	.dc.b	$F8,$A4,$9F,$2C,$65,$B3,$C6,$3C,$AD,$7C,$4D,$7C,$2D,$79,$1E,$07
	.dc.b	$04,$9F,$04,$65,$AF,$66,$BD,$9A,$F6,$6B,$D8,$E1,$3E,$88,$CC,$F7
	.dc.b	$8C,$12,$7C,$0D,$78,$98,$21,$82,$1E,$37,$83,$4D,$94,$D9,$E7,$76
	.dc.b	$D3,$65,$36,$7A,$CE,$47,$C1,$F9,$8F,$E3,$84,$7F,$94,$35,$FE,$50
	.dc.b	$D7,$F9,$43,$5F,$E5,$0D,$7F,$A0,$B6,$7F,$C3,$AD,$A3,$49,$3E,$02
	.dc.b	$1E,$02,$9E,$02,$9E,$02,$9E,$02,$CF,$81,$B3,$E0,$6C,$F8,$1E,$C4
	.dc.b	$3B,$14,$EC,$53,$B1,$67,$B3,$67,$B3,$67,$B3,$D0,$87,$42,$9D,$0B
	.dc.b	$3D,$1B,$3D,$1B,$3D,$1E,$44,$39,$16,$79,$36,$79,$36,$79,$3C,$09
	.dc.b	$3C,$1B,$3C,$1B,$3C,$27,$44,$9D,$36,$35,$18,$23,$82,$4F,$D4,$21
	.dc.b	$F2,$4A,$7C,$02,$9E,$92,$9E,$42,$CF,$85,$B3,$E0,$6C,$F8,$1E,$C4
	.dc.b	$3B,$14,$EC,$53,$B1,$67,$B3,$67,$B3,$67,$B3,$D0,$87,$43,$EE,$FC
	.dc.b	$C7,$F8,$34,$8F,$F8,$65,$AF,$F8,$65,$AF,$F8,$65,$AF,$F8,$65,$AF
	.dc.b	$F8,$C9,$FF,$0C,$75,$48,$FF,$12,$35,$FF,$12,$35,$FF,$12,$35,$FF
	.dc.b	$12,$35,$FF,$1E,$B6,$FF,$D3,$3C,$BF,$E2,$52,$4F,$F8,$C5,$AF,$F8
	.dc.b	$C5,$AF,$F8,$C5,$AF,$F8,$C5,$AF,$F9,$29,$FF,$18,$7A,$12,$3F,$C7
	.dc.b	$8D,$7F,$C7,$8D,$7F,$C7,$8D,$7F,$C7,$8D,$7F,$CA,$AD,$9F,$F5,$8F
	.dc.b	$7F,$F8,$F4,$93,$FE,$49,$6B,$FE,$49,$6B,$FE,$49,$6B,$FE,$49,$6B
	.dc.b	$FE,$62,$7F,$C9,$1E,$5C,$8F,$F2,$A3,$5F,$F2,$A3,$5F,$F2,$A3,$5F
	.dc.b	$F2,$A3,$5F,$F3,$6B,$67,$FD,$92,$DC,$CD,$24,$F8,$08,$78,$0A,$78
	.dc.b	$0A,$78,$0A,$78,$0B,$3E,$06,$CF,$81,$B3,$E0,$7B,$10,$EC,$53,$B1
	.dc.b	$4E,$C5,$9E,$CD,$9E,$CD,$9E,$CF,$42,$1D,$0A,$74,$2C,$F4,$6C,$F4
	.dc.b	$6C,$F4,$79,$10,$E4,$59,$E4,$D9,$E4,$D9,$E4,$F0,$24,$F0,$6C,$F0
	.dc.b	$6C,$F0,$9D,$12,$74,$D8,$D4,$60,$8E,$09,$3F,$50,$87,$C9,$29,$F0
	.dc.b	$0A,$7A,$4A,$79,$0B,$3E,$16,$CF,$81,$B3,$E0,$7B,$10,$EC,$53,$B1
	.dc.b	$4E,$C5,$9E,$CD,$9E,$CD,$9E,$CF,$42,$1D,$0A,$7F,$B5,$FE,$71,$F3
	.dc.b	$3F,$ED,$26,$C7,$FB,$2D,$AF,$FB,$51,$AF,$FB,$51,$AF,$FB,$51,$AF
	.dc.b	$58,$E8,$47,$A3,$5E,$8D,$7A,$35,$E8,$D7,$A1,$C8,$87,$22,$9C,$8A
	.dc.b	$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04
	.dc.b	$B4,$FC,$86,$CF,$BD,$B3,$E8,$6B,$E3,$6B,$E0,$6B,$DD,$AF,$79,$EC
	.dc.b	$49,$EC,$D8,$EC,$D7,$B3,$5E,$CD,$7B,$35,$ED,$1D,$08,$F4,$6B,$D1
	.dc.b	$AF,$46,$BD,$1A,$F4,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70
	.dc.b	$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$93,$F5,$09,$3F,$25,$B3
	.dc.b	$F0,$1B,$3E,$96,$BE,$46,$BE,$16,$BE,$06,$BE,$09,$EC,$49,$EC,$D8
	.dc.b	$EC,$D7,$B3,$5E,$CD,$7B,$35,$ED,$1D,$08,$F4,$6B,$D1,$AF,$46,$BD
	.dc.b	$1A,$F4,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70,$29,$C0,$A7
	.dc.b	$03,$44,$34,$53,$46,$08,$60,$93,$F5,$09,$3F,$26,$1C,$AF,$92,$44
	.dc.b	$21,$A6,$B9,$8E,$0E,$0D,$15,$CB,$C1,$AF,$46,$BC,$E3,$93,$C8,$E0
	.dc.b	$E2,$37,$1E,$66,$BB,$3A,$46,$C8,$F8,$DA,$F4,$30,$43,$04,$39,$99
	.dc.b	$29,$93,$D0,$64,$A6,$4F,$33,$E0,$78,$90,$D9,$4D,$9C,$CA,$73,$3D
	.dc.b	$C7,$02,$1A,$39,$1A,$30,$68,$C1,$23,$E4,$92,$3C,$D1,$B8,$C3,$E3
	.dc.b	$38,$46,$9E,$A4,$32,$6E,$38,$BE,$96,$BE,$66,$BD,$E3,$A3,$C8,$E0
	.dc.b	$E2,$37,$1D,$DA,$EC,$E9,$1B,$23,$E3,$6B,$D0,$C1,$0C,$10,$E6,$64
	.dc.b	$A6,$4F,$41,$92,$99,$3C,$CF,$81,$E2,$43,$65,$36,$73,$29,$CC,$F7
	.dc.b	$1C,$08,$68,$E4,$68,$C1,$A3,$04,$8F,$97,$5F,$E0,$4F,$99,$FF,$1A
	.dc.b	$12,$7F,$C6,$8D,$7F,$C6,$8D,$7F,$C6,$8D,$7F,$C6,$8D,$7F,$C8,$0F
	.dc.b	$F8,$FF,$E6,$64,$7F,$8F,$DA,$FF,$8F,$DA,$FF,$8F,$DA,$FF,$8F,$DA
	.dc.b	$FF,$92,$DB,$3F,$E1,$CF,$99,$FF,$26,$12,$7F,$C9,$8D,$7F,$C9,$8D
	.dc.b	$7F,$C9,$8D,$7F,$C9,$8D,$7F,$CB,$0F,$F9,$5F,$E6,$64,$7F,$95,$DA
	.dc.b	$FF,$95,$DA,$FF,$95,$DA,$FF,$95,$DA,$FF,$98,$DB,$3F,$E3,$4F,$99
	.dc.b	$FF,$32,$12,$7F,$CC,$8D,$7F,$CC,$8D,$7F,$CC,$8D,$7F,$CC,$8D,$7F
	.dc.b	$CE,$0F,$F9,$BF,$C5,$91,$FE,$6F,$6B,$FE,$6F,$6B,$FE,$6F,$6B,$FE
	.dc.b	$6F,$6B,$FE,$7B,$6C,$FF,$93,$3D,$BF,$F3,$E1,$27,$FC,$F8,$D7,$FC
	.dc.b	$F8,$D7,$FC,$F8,$D7,$FC,$F8,$D7,$FD,$10,$FF,$A1,$FE,$66,$47,$FA
	.dc.b	$1D,$AF,$FA,$1D,$AF,$FA,$1D,$AF,$FA,$1D,$AF,$FA,$4D,$B3,$FE,$64
	.dc.b	$F9,$9F,$F4,$A1,$27,$FD,$28,$D7,$FD,$28,$D7,$FD,$28,$D7,$FD,$28
	.dc.b	$D7,$FD,$40,$FF,$A7,$FD,$49,$1F,$E9,$F6,$BF,$E9,$F6,$BF,$E9,$F6
	.dc.b	$BF,$E9,$F6,$BF,$EA,$B6,$CF,$F9,$F3,$7F,$F5,$61,$27,$FD,$58,$D7
	.dc.b	$FD,$58,$D7,$FD,$58,$D7,$FD,$58,$D7,$FD,$70,$FF,$AD,$FC,$19,$1F
	.dc.b	$EB,$76,$BF,$EB,$76,$BF,$EB,$76,$BF,$EB,$76,$BF,$EC,$36,$CF,$FA
	.dc.b	$53,$E6,$7F,$D8,$84,$9F,$F6,$23,$5F,$F6,$23,$5F,$F6,$23,$5F,$F6
	.dc.b	$23,$5F,$F6,$83,$FE,$CF,$F2,$64,$7F,$B3,$DA,$FF,$B3,$DA,$FF,$B3
	.dc.b	$DA,$FF,$B3,$DA,$FF,$B6,$DB,$3F,$EA,$CF,$73,$FD,$B8,$49,$FF,$6E
	.dc.b	$35,$FF,$6E,$35,$FF,$6E,$35,$FF,$6E,$35,$FF,$74,$3F,$EE,$7F,$5E
	.dc.b	$47,$FB,$9D,$AF,$FB,$9D,$AF,$FB,$9D,$AF,$FB,$9D,$AF,$FB,$CD,$B3
	.dc.b	$FE,$C4,$F5,$7F,$DE,$84,$9F,$F7,$A3,$5F,$F7,$A3,$5F,$F7,$A3,$5F
	.dc.b	$F7,$A3,$5F,$F8,$00,$FF,$BF,$FD,$49,$1F,$EF,$F6,$BF,$EF,$F6,$BF
	.dc.b	$EF,$F6,$BF,$EF,$F6,$A9,$FF,$6D,$C6,$1F,$F6,$EF,$C2,$FF,$94,$DB
	.dc.b	$3F,$E5,$86,$BF,$E5,$C6,$BF,$E5,$F6,$BF,$E6,$26,$BD,$FA,$13,$D5
	.dc.b	$B1,$D5,$AF,$56,$BD,$5A,$F5,$6B,$D6,$3A,$11,$E8,$D7,$A3,$5E,$8D
	.dc.b	$7A,$35,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81
	.dc.b	$4E,$06,$88,$68,$A6,$8C,$10,$C1,$27,$E9,$92,$7E,$4B,$67,$E0,$36
	.dc.b	$7D,$2D,$7C,$8D,$7C,$2D,$7C,$0D,$7C,$13,$D8,$93,$D9,$B1,$D9,$AF
	.dc.b	$66,$BD,$9A,$F6,$6B,$DA,$3A,$11,$E8,$D7,$A3,$5E,$8D,$7A,$35,$E8
	.dc.b	$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E,$06,$88
	.dc.b	$68,$A6,$8C,$10,$C1,$27,$EA,$12,$7E,$4B,$67,$E0,$36,$7D,$2D,$7C
	.dc.b	$8D,$7C,$2D,$7E,$A7,$B7,$FB,$D2,$3F,$BD,$6B,$FB,$D6,$BF,$BD,$6B
	.dc.b	$FB,$D6,$BF,$C2,$5B,$0A,$46,$1B,$3D,$C8,$77,$29,$DC,$A7,$72,$9D
	.dc.b	$CB,$3D,$E7,$B1,$23,$33,$E0,$7B,$10,$EC,$53,$B1,$4E,$C5,$9E,$CF
	.dc.b	$49,$EC,$E6,$7B,$3D,$08,$74,$29,$D0,$B3,$D1,$E5,$3D,$1C,$CF,$47
	.dc.b	$91,$0E,$45,$9E,$4F,$09,$E4,$E6,$79,$3C,$09,$3C,$1D,$4F,$07,$33
	.dc.b	$C2,$75,$18,$9D,$46,$09,$3B,$24,$69,$B3,$F2,$48,$7C,$02,$9E,$92
	.dc.b	$9E,$42,$9E,$12,$CF,$82,$7B,$12,$33,$3E,$07,$B1,$0E,$C5,$3B,$14
	.dc.b	$EC,$59,$EC,$F4,$9E,$CE,$67,$B3,$D0,$87,$42,$9D,$0B,$3D,$1E,$53
	.dc.b	$D1,$CC,$F4,$79,$10,$E4,$59,$E4,$F0,$9E,$4E,$67,$93,$C0,$93,$C1
	.dc.b	$D4,$F0,$73,$3C,$27,$51,$89,$D4,$60,$93,$B2,$46,$9B,$3F,$24,$87
	.dc.b	$C0,$29,$E9,$29,$E4,$29,$E1,$2C,$F8,$27,$B1,$23,$33,$E0,$7B,$10
	.dc.b	$EC,$53,$B1,$4E,$C5,$9E,$CF,$49,$EC,$E6,$7F,$4D,$F2,$AF,$8D,$FF
	.dc.b	$1B,$36,$7F,$C6,$CD,$7F,$C6,$CD,$7F,$C6,$CD,$7F,$C6,$CD,$7F,$C8
	.dc.b	$8D,$B3,$49,$3E,$02,$4F,$81,$AF,$81,$AF,$81,$AF,$81,$AF,$81,$B3
	.dc.b	$E0,$6C,$F8,$27,$B1,$0E,$C5,$3B,$14,$EC,$53,$B1,$67,$B3,$63,$B3
	.dc.b	$D0,$87,$42,$9D,$0A,$74,$2C,$74,$6C,$74,$79,$10,$E4,$53,$91,$63
	.dc.b	$93,$63,$93,$C0,$87,$02,$C7,$06,$C7,$07,$44,$8D,$36,$35,$18,$23
	.dc.b	$82,$4F,$D4,$24,$FC,$96,$BF,$01,$AF,$A5,$AF,$91,$AF,$85,$B3,$E0
	.dc.b	$6C,$F8,$27,$B1,$0E,$C5,$3B,$14,$EC,$53,$B1,$67,$B3,$63,$B3,$D0
	.dc.b	$87,$42,$9D,$0A,$74,$2C,$74,$6C,$74,$79,$10,$FD,$8F,$CB,$7E,$E8
	.dc.b	$8F,$EE,$9A,$FE,$E9,$AF,$EE,$9A,$F2,$38,$10,$FE,$03,$EF,$7F,$06
	.dc.b	$23,$FC,$19,$AE,$8C,$10,$C1,$0F,$E1,$2F,$07,$F8,$5F,$CC,$FF,$12
	.dc.b	$23,$FC,$49,$AF,$F1,$26,$BC,$0D,$10,$D1,$4F,$E2,$FF,$33,$FC,$78
	.dc.b	$8F,$F1,$E6,$BF,$C7,$9A,$F0,$34,$43,$45,$3F,$91,$FB,$5F,$CA,$88
	.dc.b	$FF,$2A,$6B,$FC,$A9,$AF,$03,$44,$34,$53,$F9,$7F,$C4,$FF,$36,$23
	.dc.b	$FC,$D9,$AF,$F3,$66,$BC,$0D,$10,$D1,$4F,$E7,$7F,$33,$FD,$08,$8F
	.dc.b	$F4,$26,$BF,$D0,$9A,$F0,$34,$43,$45,$3F,$A3,$E5,$FD,$38,$8F,$F4
	.dc.b	$E6,$BF,$D3,$9A,$F0,$34,$43,$45,$3F,$A9,$FC,$CF,$F5,$A2,$3F,$D6
	.dc.b	$9A,$FF,$5A,$6B,$C0,$D1,$0D,$14,$D1,$82,$18,$21,$FD,$89,$EA,$D3
	.dc.b	$91,$4E,$25,$38,$9F,$DA,$1E,$2D,$38,$94,$E2,$53,$8E,$9F,$DB,$A3
	.dc.b	$8B,$63,$8B,$5E,$2D,$78,$CA,$78,$B6,$38,$B5,$E2,$D7,$FC,$07,$F3
	.dc.b	$3F,$E0,$A2,$3F,$E0,$A6,$BF,$E0,$A6,$BC,$0D,$10,$D1,$4D,$18,$21
	.dc.b	$82,$1F,$E1,$27,$AB,$4E,$45,$38,$94,$E2,$7F,$86,$1E,$2D,$38,$94
	.dc.b	$E2,$53,$8D,$7F,$87,$A3,$8B,$63,$8B,$5E,$2D,$78,$CA,$78,$B6,$38
	.dc.b	$B5,$E2,$D7,$8C,$70,$23,$C1,$AF,$04,$C2,$0F,$11,$D7,$88,$D4,$61
	.dc.b	$C1,$B8,$AD,$33,$1B,$79,$35,$EE,$D7,$AC,$72,$8C,$47,$99,$AF,$53
	.dc.b	$9C,$6C,$8F,$81,$AE,$8C,$10,$C1,$0F,$1B,$C1,$A7,$F9,$8B,$E6,$7F
	.dc.b	$CC,$A4,$70,$43,$FC,$CA,$E5,$A6,$4F,$F3,$3B,$96,$99,$3F,$CD,$2E
	.dc.b	$5A,$64,$FF,$35,$B9,$69,$9C,$3F,$CD,$B1,$96,$C7,$F9,$EB,$E6,$7F
	.dc.b	$CF,$A4,$70,$43,$FC,$FA,$E5,$A6,$4F,$F3,$FB,$96,$99,$3F,$D0,$2E
	.dc.b	$5A,$64,$FF,$41,$B9,$69,$98,$FF,$42,$C6,$5B,$1F,$8A,$EF,$FC,$21
	.dc.b	$1C,$10,$FC,$6B,$96,$99,$3F,$1E,$E5,$A6,$4F,$CE,$B9,$69,$93,$F2
	.dc.b	$4E,$5A,$64,$FD,$43,$96,$99,$C1,$19,$6C,$66,$30,$47,$04,$3D,$64
	.dc.b	$3D,$25,$3C,$E5,$3C,$A5,$3C,$65,$8F,$0B,$63,$C0,$D3,$C0,$76,$21
	.dc.b	$D8,$FF,$60,$FB,$FF,$EC,$82,$3F,$EC,$86,$BF,$EC,$86,$BC,$0D,$10
	.dc.b	$D1,$4D,$18,$21,$82,$1F,$ED,$07,$AB,$4E,$45,$38,$94,$FF,$6C,$FC
	.dc.b	$67,$FB,$80,$8F,$FB,$81,$AF,$FB,$81,$AF,$03,$44,$34,$53,$46,$08
	.dc.b	$60,$87,$FB,$A1,$EA,$D3,$91,$4E,$25,$3F,$DE,$3F,$33,$FE,$F8,$23
	.dc.b	$FE,$F8,$6B,$FE,$F8,$6B,$C0,$D1,$0D,$14,$D1,$82,$18,$21,$FF,$00
	.dc.b	$1E,$AD,$39,$14,$E2,$53,$FA,$86,$0D,$5F,$25,$FD,$32,$30,$D8,$CC
	.dc.b	$FF,$53,$9C,$B6,$32,$FF,$5A,$8C,$B6,$32,$F8,$1A,$FF,$62,$8D,$B9
	.dc.b	$9F,$01,$1E,$CD,$7A,$35,$E0,$68,$86,$8A,$68,$C1,$0F,$27,$C2,$7B
	.dc.b	$C8,$FB,$5A,$FA,$DA,$F0,$34,$43,$45,$34,$60,$86,$08,$7C,$3A,$47
	.dc.b	$68,$C3,$55,$7D,$98,$C3,$63,$EA,$7C,$9F,$D9,$23,$82,$1F,$55,$CB
	.dc.b	$4C,$9F,$5D,$CB,$4C,$9E,$77,$2D,$32,$79,$A9,$1A,$8C,$36,$7F,$24
	.dc.b	$D9,$DC,$FE,$37,$E4,$3F,$24,$47,$04,$3F,$1C,$E5,$A6,$4F,$C8,$39
	.dc.b	$69,$93,$E8,$39,$69,$93,$D4,$E5,$A6,$6B,$F5,$51,$96,$C6,$65,$39
	.dc.b	$6C,$66,$30,$47,$04,$3D,$64,$3D,$25,$3C,$E5,$3C,$A5,$9F,$1B,$65
	.dc.b	$F3,$1F,$BD,$24,$7E,$ED,$AF,$EE,$5A,$FE,$E5,$AF,$E5,$DA,$FD,$F6
	.dc.b	$CF,$F0,$D6,$CF,$82,$7B,$10,$EC,$53,$B1,$4E,$C5,$3F,$18,$7F,$15
	.dc.b	$21,$FC,$4C,$A7,$F1,$12,$9F,$C4,$4A,$7E,$BC,$A7,$43,$91,$0E,$45
	.dc.b	$39,$14,$E4,$53,$91,$C0,$87,$03,$F9,$37,$CE,$BE,$47,$F9,$83,$63
	.dc.b	$F9,$73,$5D,$C6,$88,$E9,$AE,$8C,$10,$C1,$0F,$E8,$31,$D5,$B1,$C9
	.dc.b	$A7,$12,$9C,$4F,$E7,$11,$C5,$B1,$C5,$A7,$AB,$E6,$3D,$A4,$70,$43
	.dc.b	$E1,$B9,$69,$93,$CE,$E5,$A6,$4F,$1B,$96,$99,$3F,$93,$CA,$35,$18
	.dc.b	$6E,$09,$D4,$F1,$8C,$4F,$D6,$F9,$89,$3F,$59,$AF,$DC,$6B,$F3,$9A
	.dc.b	$FC,$86,$BE,$46,$CF,$85,$B4,$69,$27,$C0,$43,$C0,$53,$C0,$7E,$2B
	.dc.b	$E3,$BF,$1A,$47,$04,$3F,$26,$E5,$A6,$4F,$C3,$B9,$69,$93,$F0,$6E
	.dc.b	$5A,$64,$FA,$2E,$5A,$66,$BE,$84,$65,$B1,$FA,$2F,$86,$FD,$29,$1C
	.dc.b	$10,$FD,$5B,$96,$99,$3F,$3E,$E5,$A6,$4F,$CD,$B9,$69,$93,$F1,$2E
	.dc.b	$5A,$66,$3F,$11,$19,$6C,$7E,$EB,$E1,$BF,$7A,$47,$04,$3F,$81,$B9
	.dc.b	$69,$93,$F6,$EE,$5A,$64,$FD,$9B,$96,$99,$3F,$42,$E5,$A6,$63,$F4
	.dc.b	$11,$96,$C7,$F1,$2F,$86,$FE,$2A,$47,$04,$3F,$8D,$B9,$69,$93,$F8
	.dc.b	$7B,$96,$99,$3F,$85,$B9,$69,$93,$F7,$2E,$5A,$66,$3F,$71,$19,$6C
	.dc.b	$7F,$2A,$F8,$6F,$E5,$A4,$70,$43,$F9,$9B,$96,$99,$3F,$93,$B9,$69
	.dc.b	$93,$F9,$1B,$96,$99,$3F,$88,$B9,$69,$98,$FE,$21,$19,$6C,$7F,$42
	.dc.b	$F8,$6F,$E8,$A4,$70,$43,$FA,$5B,$96,$99,$3F,$9F,$B9,$69,$93,$F9
	.dc.b	$DB,$96,$99,$3F,$94,$B9,$69,$98,$FE,$51,$19,$6C,$7F,$5A,$F8,$6F
	.dc.b	$EB,$A4,$70,$43,$FB,$1B,$96,$99,$3F,$AB,$B9,$69,$93,$FA,$9B,$96
	.dc.b	$99,$3F,$A0,$B9,$69,$98,$FE,$81,$19,$6C,$7F,$72,$F8,$6F,$EE,$A4
	.dc.b	$70,$43,$FB,$DB,$96,$99,$3F,$B7,$B9,$69,$93,$FB,$5B,$96,$99,$3F
	.dc.b	$AC,$B9,$69,$98,$FE,$B1,$19,$6C,$7F,$82,$BE,$1B,$FC,$1A,$47,$04
	.dc.b	$3F,$C2,$6E,$5A,$64,$FF,$03,$B9,$69,$93,$FC,$06,$E5,$A6,$4F,$EE
	.dc.b	$2E,$5A,$66,$3F,$B8,$46,$5B,$1F,$E2,$2F,$86,$FF,$12,$91,$C1,$0F
	.dc.b	$F1,$5B,$96,$99,$3F,$C3,$EE,$5A,$64,$FF,$0D,$B9,$69,$93,$FC,$12
	.dc.b	$E5,$A6,$63,$FC,$11,$19,$6C,$7F,$8E,$BE,$1B,$FC,$7A,$47,$04,$3F
	.dc.b	$C8,$6E,$5A,$64,$FF,$1B,$B9,$69,$93,$FC,$66,$E5,$A6,$4F,$F1,$0B
	.dc.b	$96,$99,$8F,$F1,$04,$65,$B1,$FE,$52,$F8,$6F,$F2,$A9,$1C,$10,$FF
	.dc.b	$2D,$B9,$69,$93,$FC,$9E,$E5,$A6,$4F,$F2,$5B,$96,$99,$3F,$C7,$2E
	.dc.b	$5A,$66,$3F,$C7,$11,$96,$C7,$F9,$AB,$E1,$BF,$CD,$A4,$70,$43,$FC
	.dc.b	$E6,$E5,$A6,$4F,$F3,$3B,$96,$99,$3F,$CC,$6E,$5A,$64,$FF,$28,$B9
	.dc.b	$69,$98,$FF,$28,$46,$5B,$1F,$E8,$2F,$86,$FF,$42,$91,$C1,$0F,$F4
	.dc.b	$5B,$96,$99,$3F,$CF,$EE,$5A,$64,$FF,$3D,$B9,$69,$93,$FC,$D2,$E5
	.dc.b	$A6,$63,$FC,$D1,$19,$6C,$7F,$A6,$BE,$1B,$FD,$3A,$47,$04,$3F,$D4
	.dc.b	$6E,$5A,$64,$FF,$4B,$B9,$69,$93,$FD,$26,$E5,$A6,$4F,$F4,$0B,$96
	.dc.b	$99,$8F,$F4,$04,$65,$B1,$FE,$B2,$F8,$6F,$F5,$A9,$1C,$10,$FF,$5D
	.dc.b	$B9,$69,$93,$FD,$5E,$E5,$A6,$4F,$F5,$5B,$96,$99,$3F,$D3,$2E,$5A
	.dc.b	$66,$3F,$D3,$11,$96,$C7,$FB,$2B,$E1,$BF,$D9,$A4,$70,$43,$FD,$A6
	.dc.b	$E5,$A6,$4F,$F6,$3B,$96,$99,$3F,$D8,$6E,$5A,$64,$FF,$58,$B9,$69
	.dc.b	$98,$FF,$58,$46,$5B,$1F,$EE,$2F,$85,$FF,$72,$91,$C1,$0F,$F7,$5B
	.dc.b	$96,$99,$3F,$DB,$EE,$5A,$64,$FF,$6D,$B9,$69,$93,$FD,$92,$E5,$A6
	.dc.b	$63,$FD,$91,$19,$6C,$67,$9D,$39,$6C,$66,$30,$47,$00,$84,$F9,$E6
	.dc.b	$EF,$9F,$D7,$AE,$F2,$B5,$39,$B4,$E5,$E9,$7C,$AF,$8D,$F0,$6C,$EA
	.dc.b	$F9,$5F,$18,$5F,$17,$AD,$F1,$69,$C6,$F9,$BC,$2F,$9C,$EE,$FA,$5F
	.dc.b	$36,$CE,$AB,$91,$F0,$8A,$59,$F1,$36,$3E,$03,$6F,$83,$C0,$F3,$19
	.dc.b	$33,$3C,$9C,$90,$E1,$F4,$02,$00

;--------------------------------------------------------------------------------
;	FGETMAN.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fgetman_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FGETMAN.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FGETMAN.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fgetman_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fgetman.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FGETMAN.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fgetman.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FGETMAN.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fgetman_data_compressed::
	.dc.b	$00,$02,$84,$2C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CF,$ED,$1F,$8A,$7F,$87,$A7,$FC,$D0,$5F,$8A,$E1
	.dc.b	$F0,$FC,$7D,$FE,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$1F
	.dc.b	$07,$9F,$DB,$F4,$7B,$7F,$DF,$7E,$6F,$FA,$B0,$F8,$8C,$7C,$46,$7E
	.dc.b	$23,$5F,$11,$BF,$88,$E1,$F1,$1C,$7E,$17,$97,$C0,$F3,$F8,$1E,$9F
	.dc.b	$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03
	.dc.b	$E3,$F8,$1F,$27,$C0,$F9,$7E,$07,$CD,$F0,$3E,$7F,$81,$FF,$0A,$7C
	.dc.b	$E7,$E1,$BE,$13,$FC,$40,$BE,$07,$3E,$66,$BE,$07,$7F,$03,$C3,$E0
	.dc.b	$78,$FC,$0F,$2F,$81,$E7,$F0,$3F,$63,$D2,$EB,$F0,$3D,$BE,$07,$BF
	.dc.b	$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93,$E0,$7C,$BF
	.dc.b	$03,$E6,$F8,$1F,$3F,$C0,$FA,$3E,$07,$FC,$59,$F3,$BF,$88,$E9,$FF
	.dc.b	$1C,$7C,$07,$F0,$A5,$1F,$EF,$63,$FD,$60,$7F,$9A,$0F,$EB,$47,$F1
	.dc.b	$23,$81,$FC,$41,$FF,$21,$2F,$7B,$BF,$BD,$E0,$F7,$BC,$3E,$F7,$8B
	.dc.b	$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD,$E8,$F7,$BF
	.dc.b	$3B,$EF,$FE,$7B,$5F,$CF,$EB,$FA,$0D,$7F,$43,$AF,$BB,$DE,$F7,$FB
	.dc.b	$DF,$E4,$BF,$6F,$FC,$48,$7E,$3F,$8B,$FC,$B0,$BD,$CD,$3B,$F7,$38
	.dc.b	$7B,$9C,$7D,$CE,$5E,$E7,$3F,$73,$A7,$B9,$FB,$0E,$8E,$DE,$E7,$7F
	.dc.b	$73,$C1,$EE,$7E,$D3,$AF,$F6,$BD,$7F,$B6,$EB,$FD,$BF,$5F,$EE,$3A
	.dc.b	$FF,$73,$D7,$FB,$AE,$6F,$DD,$F3,$7F,$98,$FF,$15,$FF,$1E,$1E,$9E
	.dc.b	$73,$E9,$7F,$CE,$4F,$FB,$38,$FF,$44,$1F,$DE,$0F,$E6,$07,$F2,$F7
	.dc.b	$8F,$0F,$5E,$1E,$DC,$1D,$F8,$3E,$47,$BB,$F2,$7D,$DF,$95,$EE,$FC
	.dc.b	$BF,$77,$E6,$7B,$BF,$37,$DD,$F9,$DE,$EF,$CF,$F7,$7E,$87,$BB,$FE
	.dc.b	$7B,$F9,$FF,$D4,$70,$1F,$D4,$3F,$E8,$95,$EE,$6F,$DC,$E1,$EE,$71
	.dc.b	$F7,$39,$7B,$9C,$FD,$CE,$9E,$E7,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B
	.dc.b	$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$F3,$7C,$9F,$9D,$D9,$F3
	.dc.b	$FB,$3E,$87,$67,$D1,$EC,$FA,$5D,$9F,$4F,$B3,$EA,$76,$7D,$5E,$CF
	.dc.b	$AD,$D9,$FE,$99,$F4,$FD,$E7,$BB,$B4,$FB,$9F,$F5,$42,$EB,$E1,$D7
	.dc.b	$C7,$AF,$97,$5F,$3E,$BE,$9D,$7D,$7A,$FE,$67,$B7,$F3,$7D,$BF,$9D
	.dc.b	$ED,$F8,$7B,$BE,$87,$B7,$FE,$AB,$FA,$0F,$E0,$1C,$05,$F0,$99,$F8
	.dc.b	$AD,$7C,$AE,$FE,$17,$87,$C0,$F1,$F8,$6E,$5F,$01,$CF,$E0,$3A,$74
	.dc.b	$75,$F8,$0F,$81,$E4,$FC,$1F,$27,$E1,$79,$3F,$0F,$C9,$F8,$9E,$4F
	.dc.b	$8F,$E0,$3C,$9E,$D7,$97,$DA,$F3,$7B,$5E,$7F,$6B,$D1,$ED,$7A,$7D
	.dc.b	$AF,$57,$B5,$EB,$F6,$BD,$9E,$D7,$FB,$1F,$E8,$7F,$D4,$07,$E9,$4F
	.dc.b	$D2,$76,$97,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1F,$67,$D9
	.dc.b	$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BB,$E1,$F7,$7C,$5E,$EF
	.dc.b	$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE,$F9,$FD,$DF,$47,$BB,$F8,$7F
	.dc.b	$57,$D5,$EE,$FF,$B8,$7C,$FF,$D3,$9F,$89,$3F,$11,$D2,$7F,$10,$FF
	.dc.b	$BC,$5F,$F7,$61,$FE,$A5,$3F,$CB,$AE,$FD,$9E,$9E,$CF,$5F,$67,$B7
	.dc.b	$AF,$DF,$CC,$F0,$79,$9E,$1F,$33,$C5,$E6,$78,$FC,$CF,$27,$67,$97
	.dc.b	$7F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$FB,$E7,$E8,$BF,$1E,$7E,$3B,$84
	.dc.b	$17,$B1,$BE,$3E,$1E,$C7,$1F,$63,$97,$B1,$CF,$D8,$E9,$EC,$75,$F6
	.dc.b	$3B,$7B,$1F,$AF,$DF,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E
	.dc.b	$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7A,$7D,$8F,$3F,$3B,$E8
	.dc.b	$FF,$DD,$87,$F3,$13,$F1,$A7,$E3,$1F,$6A,$F5,$F8,$BF,$EA,$83,$FC
	.dc.b	$C1,$C2,$EB,$CF,$AB,$AF,$57,$7E,$AF,$0F,$57,$8F,$AB,$CB,$D5,$E7
	.dc.b	$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E,$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F
	.dc.b	$57,$C7,$EA,$FB,$3E,$93,$F9,$99,$FC,$54,$FC,$69,$F8,$CE,$62,$F7
	.dc.b	$F8,$7B,$FC,$7D,$FE,$5E,$CF,$3F,$67,$A7,$B3,$D7,$D9,$ED,$EC,$F7
	.dc.b	$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9,$EC,$F9,$7C,$FF
	.dc.b	$37,$85,$E7,$F6,$7E,$9F,$C0,$7F,$1A,$3F,$19,$CB,$FC,$A7,$D9,$CF
	.dc.b	$B3,$AF,$67,$7E,$CF,$0F,$67,$8F,$B3,$CB,$D9,$E7,$E6,$74,$F6,$7A
	.dc.b	$FB,$3D,$BD,$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F,$67,$C7,$EC,$F9
	.dc.b	$3D,$9F,$2F,$B3,$E6,$F6,$7F,$13,$F4,$BF,$8A,$E6,$3F,$DE,$4F,$E2
	.dc.b	$A7,$E2,$5C,$FB,$5C,$3D,$DE,$3E,$EF,$2F,$6F,$9F,$B7,$D3,$DB,$EB
	.dc.b	$ED,$F6,$F6,$FB,$FB,$7E,$0F,$6F,$C3,$ED,$F8,$BD,$BF,$1F,$B7,$E4
	.dc.b	$F6,$FC,$BE,$DF,$9B,$DB,$F3,$FB,$7E,$8F,$6F,$F4,$3F,$4D,$FD,$E0
	.dc.b	$FE,$B4,$7F,$12,$3F,$0C,$7E,$17,$97,$F7,$AB,$DA,$E3,$ED,$72,$F6
	.dc.b	$B9,$FB,$5D,$3D,$AE,$BE,$D7,$6F,$6B,$BF,$B5,$E0,$F6,$BC,$3E,$D7
	.dc.b	$8B,$DA,$F1,$FB,$5E,$4F,$6B,$CB,$ED,$79,$BD,$AF,$3F,$B5,$FB,$8F
	.dc.b	$A7,$FE,$BE,$7E,$18,$FE,$1F,$C9,$FC,$49,$FE,$54,$DF,$63,$87,$B1
	.dc.b	$C7,$D8,$E5,$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1
	.dc.b	$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E
	.dc.b	$7F,$63,$F8,$7F,$AF,$F8,$73,$FC,$7C,$7F,$5A,$3F,$0C,$7F,$0D,$E4
	.dc.b	$FE,$56,$BD,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6,$F5
	.dc.b	$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC,$BE
	.dc.b	$BF,$9B,$D7,$F3,$FA,$FE,$8F,$5F,$F9,$3F,$D4,$7F,$8F,$8F,$F1,$01
	.dc.b	$FD,$60,$FE,$18,$7E,$1F,$94,$FE,$1D,$FE,$8A,$FF,$58,$1E,$AF,$4F
	.dc.b	$57,$AF,$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E
	.dc.b	$AF,$93,$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FA,$3D,$5F,$E7,$FF,$53
	.dc.b	$FE,$26,$3F,$86,$9F,$88,$E2,$3F,$88,$6F,$AB,$C3,$D5,$E3,$EA,$F2
	.dc.b	$F5,$79,$FA,$BD,$3D,$5E,$BE,$AF,$6F,$57,$BF,$AB,$E0,$F5,$7C,$3E
	.dc.b	$AF,$8B,$D5,$F1,$FA,$BE,$4F,$57,$CB,$EA,$F9,$BD,$5F,$3F,$AB,$FD
	.dc.b	$5B,$EA,$BF,$87,$1F,$E8,$53,$FC,$42,$7F,$0D,$3F,$11,$C8,$7F,$10
	.dc.b	$FF,$74,$5E,$A7,$3F,$53,$A7,$A9,$D7,$D4,$ED,$EA,$77,$F5,$3C,$1E
	.dc.b	$A7,$87,$D4,$F1,$7A,$9E,$3F,$53,$C9,$EA,$79,$7D,$4F,$37,$A9,$E7
	.dc.b	$F5,$3D,$1E,$A7,$A7,$D4,$FE,$DF,$E2,$7F,$A1,$8F,$F3,$41,$FE,$20
	.dc.b	$3F,$AA,$1F,$89,$3F,$11,$C8,$7F,$C3,$E7,$E2,$1F,$F0,$7A,$F4,$BA
	.dc.b	$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA,$5E,$2F,$4B,$C7,$E9,$79
	.dc.b	$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$A3,$D2,$FF,$03,$FD,$5F,$F9
	.dc.b	$B8,$FE,$AA,$7E,$24,$FC,$41,$E7,$E1,$FF,$11,$AF,$4B,$8F,$A5,$CB
	.dc.b	$D2,$E7,$E9,$74,$F4,$BA,$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA
	.dc.b	$5E,$2F,$4B,$C7,$E9,$79,$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$F8
	.dc.b	$7B,$EB,$3F,$AB,$1F,$EC,$D3,$FC,$D2,$7F,$55,$3F,$12,$7E,$23,$90
	.dc.b	$FE,$21,$FF,$1E,$BF,$EC,$D1,$E8,$75,$F4,$3B,$7A,$1D,$FD,$0F,$07
	.dc.b	$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F,$43,$CD,$E8,$79
	.dc.b	$FD,$0F,$47,$A1,$E9,$F4,$3F,$C6,$FF,$5B,$FE,$CE,$3F,$D6,$07,$F9
	.dc.b	$A0,$FF,$0E,$1F,$C4,$4F,$C4,$9F,$E6,$63,$F1,$1C,$47,$F1,$0F,$F9
	.dc.b	$61,$79,$FD,$BC,$FE,$FE,$7F,$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7
	.dc.b	$F9,$3C,$FF,$2F,$9F,$E6,$F3,$FC,$FE,$7F,$A3,$CF,$FF,$27,$FC,$17
	.dc.b	$FA,$D8,$FF,$0E,$9F,$C4,$4F,$E2,$07,$E2,$78,$3F,$C6,$9E,$7F,$0F
	.dc.b	$3F,$8F,$9F,$CB,$CF,$E7,$E7,$F4,$F3,$FA,$F9,$FD,$BC,$FE,$FE,$7F
	.dc.b	$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7,$F9,$3C,$FF,$2F,$9F,$E6,$F3
	.dc.b	$FF,$CC,$BF,$5D,$FE,$1D,$1F,$27,$8F,$8F,$CF,$C4,$EB,$DE,$DF,$BD
	.dc.b	$C3,$DA,$E3,$ED,$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$FE,$7C,$FA,$F7,$CB,$DF,$94,$C7,$C8,$67,$E2,$B5
	.dc.b	$E7,$6F,$E1,$B8,$7C,$37,$1F,$84,$E5,$F0,$3C,$FD,$EE,$9E,$F7,$5F
	.dc.b	$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27
	.dc.b	$BD,$E5,$F7,$BC,$DE,$F7,$FA,$4B,$EC,$3F,$DF,$27,$F9,$80,$FE,$A8
	.dc.b	$7F,$53,$3F,$08,$7F,$0B,$F7,$FF,$D5,$EB,$DE,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$FD,$53,$EC,$FF,$98
	.dc.b	$4F,$C2,$F1,$FF,$B0,$9F,$F7,$A9,$FE,$61,$3F,$AA,$9F,$C3,$4F,$E1
	.dc.b	$8F,$F8,$77,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8E,$FE,$C7,$83
	.dc.b	$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79,$3D,$8F,$2F,$B1,$E6,$F6,$3C
	.dc.b	$FE,$C7,$A3,$D8,$FF,$60,$FD,$8F,$E1,$B8,$BF,$DB,$2B,$D7,$CF,$AF
	.dc.b	$AF,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D
	.dc.b	$7F,$2F,$AF,$E6,$F5,$FC,$FE,$BF,$FB,$63,$EC,$9F,$35,$7E,$4F,$1F
	.dc.b	$1F,$9F,$8F,$D7,$C4,$EF,$E1,$B8,$7B,$BC,$7D,$DE,$5E,$EF,$3F,$77
	.dc.b	$A7,$BB,$D7,$DD,$ED,$EE,$F7,$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE
	.dc.b	$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$FE,$EE,$FB,$3F,$F5,$31,$F8
	.dc.b	$6F,$38,$BE,$17,$5F,$0B,$BF,$85,$E1,$F0,$3C,$7E,$17,$97,$C1,$73
	.dc.b	$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F,$82,$F0,$7C,$17,$87,$E0
	.dc.b	$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E,$0B,$CD,$F0,$5E,$7F,$82
	.dc.b	$F4,$7C,$13,$9E,$78,$FF,$87,$7F,$BE,$7C,$0E,$5F,$EF,$DF,$03,$67
	.dc.b	$FD,$DC,$7F,$98,$CF,$F0,$D8,$F7,$B5,$EF,$6F,$DE,$E1,$EF,$71,$F7
	.dc.b	$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE
	.dc.b	$1F,$7B,$C5,$EF,$78,$FD,$EF,$37,$DA,$7E,$23,$93,$D8,$FF,$BB,$8F
	.dc.b	$F3,$11,$FE,$61,$3F,$A9,$1F,$C3,$D3,$FC,$3D,$E3,$EC,$74,$F6,$3A
	.dc.b	$FB,$1D,$BD,$8E,$FE,$C7,$83,$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79
	.dc.b	$3D,$8F,$2F,$B1,$E6,$F6,$3C,$FE,$C7,$A3,$D8,$F9,$DF,$6B,$FE,$EF
	.dc.b	$3F,$14,$7E,$27,$87,$EB,$3F,$EE,$F3,$FC,$C6,$7F,$86,$CF,$E1,$E3
	.dc.b	$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD,$3E,$FE,$9F,$83,$D3,$F0,$FA
	.dc.b	$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F,$A7,$E6,$F4,$FC,$FE,$9F,$A3
	.dc.b	$D3,$FC,$37,$DB,$7E,$24,$FE,$1D,$C7,$F8,$D5,$E9,$EB,$E1,$37,$F0
	.dc.b	$9C,$3E,$13,$8F,$AB,$CB,$D5,$E7,$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E
	.dc.b	$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F,$57,$C7,$EA,$F9,$3D,$5F,$2F,$AB
	.dc.b	$E6,$F5,$7C,$FE,$AF,$E7,$78,$7F,$11,$CD,$FF,$00,$7E,$0B,$1F,$05
	.dc.b	$9F,$82,$D7,$C1,$6F,$DA,$E1,$F0,$5C,$7D,$AE,$5E,$D7,$3F,$6B,$A7
	.dc.b	$B5,$D7,$DA,$ED,$ED,$77,$F6,$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F
	.dc.b	$6B,$C9,$ED,$79,$7D,$AF,$37,$B5,$FB,$3F,$B7,$FD,$A7,$39,$FE,$20
	.dc.b	$7E,$1D,$C7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$D1,$EF,$7F,$0B,$FB
	.dc.b	$8F,$E1,$C7,$E1,$B9,$8F,$E1,$9F,$E2,$8B,$DD,$E1,$EE,$F1,$F3,$B9
	.dc.b	$7B,$BC,$FD,$DE,$9E,$EF,$5F,$77,$B7,$BB,$DF,$DD,$F0,$7B,$BE,$1F
	.dc.b	$77,$C5,$EE,$F8,$FD,$DF,$27,$BB,$E5,$F7,$7C,$DE,$EF,$9F,$DD,$F4
	.dc.b	$7B,$BF,$C8,$FE,$E7,$F0,$E3,$DE,$C7,$BD,$9F,$7B,$5E,$F6,$FD,$EE
	.dc.b	$1E,$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$75,$F7,$BB,$7B,$DD,$FD
	.dc.b	$EF,$07,$BD,$E1,$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD
	.dc.b	$EF,$7F,$39,$F6,$BF,$0E,$7F,$55,$3F,$0C,$3D,$DD,$7B,$BB,$F7,$78
	.dc.b	$4F,$1F,$77,$97,$BB,$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7B,$BD,$FD,$DF
	.dc.b	$07,$BB,$E1,$F7,$7C,$5E,$EF,$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE
	.dc.b	$F9,$FD,$DF,$EA,$1F,$75,$FD,$4C,$FC,$31,$F8,$5F,$78,$FE,$15,$FE
	.dc.b	$B4,$BD,$CE,$3E,$E7,$2F,$73,$9F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B
	.dc.b	$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7
	.dc.b	$9B,$DC,$F3,$FB,$9E,$8F,$73,$FB,$47,$DD,$FF,$0C,$3F,$0D,$CB,$FD
	.dc.b	$BD,$7B,$7A,$F6,$F7,$ED,$F0,$F6,$F8,$FB,$7C,$BD,$BE,$7E,$DF,$4F
	.dc.b	$6D,$FC,$03,$0B,$A7,$58,$7D,$E7,$F7,$DF,$83,$74,$F1,$8C,$36,$BF
	.dc.b	$9A,$B6,$77,$1C,$E7,$6D,$36,$58,$DA,$DD,$7E,$28,$D9,$97,$13,$C2
	.dc.b	$77,$5D,$AB,$4D,$8C,$C7,$2F,$A8,$7B,$9F,$E1,$78,$FF,$0B,$46,$3E
	.dc.b	$3F,$F9,$2C,$7F,$87,$63,$0E,$1A,$FF,$2A,$6B,$CA,$7A,$3C,$5A,$A3
	.dc.b	$6E,$67,$BB,$99,$DC,$EA,$31,$3A,$FD,$65,$E4,$F4,$6C,$66,$30,$47
	.dc.b	$04,$9F,$6C,$F0,$6C,$ED,$B3,$B7,$FA,$3F,$D4,$BC,$9F,$5B,$67,$D2
	.dc.b	$D9,$F9,$2D,$78,$CF,$02,$4F,$06,$C7,$06,$BC,$23,$E2,$E1,$F0,$C9
	.dc.b	$8F,$C6,$24,$69,$B3,$F5,$DA,$FD,$76,$BF,$5E,$3D,$CF,$F9,$5F,$1F
	.dc.b	$88,$FF,$66,$6B,$9C,$7E,$33,$96,$B9,$9E,$6F,$F9,$96,$78,$B6,$76
	.dc.b	$BE,$E9,$F9,$F7,$F8,$7A,$FF,$23,$18,$DB,$F1,$6D,$8C,$CE,$EB,$FB
	.dc.b	$DC,$E5,$B1,$97,$FC,$E5,$19,$6C,$65,$FF,$3D,$46,$5B,$19,$7C,$CD
	.dc.b	$8D,$46,$3E,$1F,$CB,$3F,$DB,$EB,$CB,$18,$23,$82,$4F,$9A,$32,$D8
	.dc.b	$CB,$E6,$8E,$BE,$4C,$9F,$82,$D9,$F7,$B5,$F6,$B5,$F5,$B5,$F5,$B6
	.dc.b	$FE,$8F,$93,$EA,$25,$7C,$E6,$D7,$CC,$6B,$F2,$9A,$FC,$86,$BF,$21
	.dc.b	$F8,$BE,$4C,$9F,$B2,$D9,$FA,$ED,$7E,$AB,$5F,$A6,$D7,$E9,$B6,$BF
	.dc.b	$13,$E4,$FD,$22,$57,$E1,$9B,$5F,$84,$6B,$F8,$26,$BF,$80,$6B,$F8
	.dc.b	$07,$EE,$F9,$32,$7F,$22,$D9,$FC,$7B,$5F,$C6,$B5,$FC,$5B,$5F,$C5
	.dc.b	$B6,$BF,$43,$E4,$FE,$28,$95,$F9,$E6,$D7,$E7,$1A,$FE,$69,$AF,$E6
	.dc.b	$1A,$FE,$61,$FC,$AF,$3C,$9F,$D4,$B6,$7F,$4E,$D7,$F4,$AD,$7F,$46
	.dc.b	$D7,$F4,$6D,$AF,$DC,$F9,$3F,$A2,$25,$7E,$D9,$B5,$FB,$46,$BF,$B2
	.dc.b	$6B,$FB,$06,$BF,$B0,$7F,$5B,$E4,$C9,$FE,$02,$D9,$FD,$FB,$5F,$DE
	.dc.b	$B5,$FD,$DB,$5F,$DD,$B6,$BF,$88,$F9,$3F,$BA,$25,$7F,$0E,$6D,$7F
	.dc.b	$0C,$6B,$FC,$29,$AF,$F0,$86,$BF,$C2,$1F,$E0,$BE,$4C,$9F,$E3,$2D
	.dc.b	$9F,$E2,$ED,$7F,$8A,$B5,$FE,$26,$D7,$F8,$9B,$6B,$F9,$4F,$93,$FC
	.dc.b	$48,$95,$FC,$99,$B5,$FC,$91,$AF,$F2,$26,$BF,$C8,$1A,$FF,$20,$7F
	.dc.b	$8E,$F9,$32,$7F,$98,$B6,$7F,$97,$B5,$FE,$5A,$D7,$F9,$5B,$5F,$E5
	.dc.b	$6D,$AF,$E8,$3E,$4F,$F2,$A2,$57,$F3,$E6,$D7,$F3,$C6,$BF,$CE,$9A
	.dc.b	$FF,$38,$6B,$FC,$E1,$FE,$6B,$E4,$C9,$FE,$92,$D9,$FE,$8E,$D7,$FA
	.dc.b	$2B,$5F,$E8,$6D,$7F,$A1,$B6,$BF,$AC,$F9,$3F,$D0,$89,$5F,$D5,$9B
	.dc.b	$5F,$D5,$1A,$FF,$52,$6B,$FD,$41,$AF,$F5,$07,$FA,$6F,$93,$27,$FB
	.dc.b	$0B,$67,$FA,$FB,$5F,$EB,$AD,$7F,$AD,$B5,$FE,$B6,$DA,$68,$F9,$0F
	.dc.b	$ED,$ED,$BF,$EE,$6D,$BF,$ED,$ED,$7F,$B6,$B5,$FE,$D6,$D7,$FB,$F3
	.dc.b	$6B,$BD,$76,$25,$76,$6D,$76,$F8,$7F,$19,$27,$C6,$D8,$D4,$60,$8E
	.dc.b	$08,$79,$A3,$83,$E8,$F2,$FD,$E4,$9F,$7B,$63,$DE,$D7,$84,$68,$8E
	.dc.b	$9A,$FC,$4F,$2F,$E5,$92,$3E,$5B,$63,$E5,$B5,$E1,$1A,$23,$A6,$BF
	.dc.b	$43,$CB,$FA,$E4,$8F,$AE,$D8,$FA,$ED,$78,$46,$88,$E9,$AF,$DC,$F2
	.dc.b	$FF,$06,$48,$FC,$1B,$63,$F0,$6D,$78,$46,$88,$E9,$AF,$E2,$3C,$BF
	.dc.b	$C7,$92,$3F,$1E,$D8,$FC,$7B,$5E,$11,$A2,$3A,$6B,$F9,$4F,$2F,$F3
	.dc.b	$64,$8F,$CD,$B6,$3F,$36,$D7,$84,$68,$8E,$9A,$FE,$83,$CB,$FD,$39
	.dc.b	$23,$F4,$ED,$8F,$D3,$B5,$E1,$1A,$23,$A6,$BF,$AC,$F2,$BF,$66,$48
	.dc.b	$FD,$9B,$63,$F6,$6D,$78,$46,$88,$E9,$AF,$EE,$3E,$1F,$F7,$E4,$8F
	.dc.b	$DF,$B6,$3F,$7E,$D7,$84,$68,$8E,$9A,$FF,$04,$F8,$7F,$E1,$64,$8F
	.dc.b	$E1,$6D,$8F,$E1,$6D,$78,$46,$88,$E9,$AF,$F1,$0F,$87,$FE,$2E,$48
	.dc.b	$FE,$2E,$D8,$FE,$2E,$D7,$84,$68,$8E,$9A,$FF,$1C,$F2,$3F,$91,$92
	.dc.b	$3F,$91,$B6,$3F,$91,$B5,$E1,$1A,$23,$A6,$BF,$CA,$3E,$1F,$F9,$79
	.dc.b	$23,$F9,$7B,$63,$F9,$7B,$5E,$11,$A2,$3A,$6B,$FC,$D3,$E1,$FF,$9D
	.dc.b	$92,$3F,$9D,$B6,$3F,$9D,$B5,$E1,$1A,$23,$A6,$BF,$D0,$3E,$1F,$FA
	.dc.b	$39,$23,$FA,$3B,$63,$FA,$3B,$5E,$11,$A2,$3A,$6B,$FD,$33,$E1,$FF
	.dc.b	$A9,$92,$3F,$A9,$B6,$3F,$A9,$B5,$E1,$1A,$23,$A6,$BF,$D6,$3E,$1F
	.dc.b	$FA,$F9,$23,$FA,$FB,$63,$FA,$FB,$5E,$11,$A2,$3A,$6B,$FD,$93,$E1
	.dc.b	$FF,$B5,$92,$3F,$B5,$B6,$3F,$B5,$B5,$E1,$1A,$23,$A6,$BF,$DC,$3E
	.dc.b	$1F,$FB,$B9,$23,$FB,$BB,$63,$FB,$BB,$5E,$11,$A2,$3A,$6B,$FE,$E8
	.dc.b	$FC,$FC,$FF,$BC,$9B,$1F,$EF,$66,$C7,$FB,$C9,$AF,$13,$81,$23,$83
	.dc.b	$4E,$05,$38,$3D,$7E,$5B,$C0,$47,$04,$8F,$04,$65,$B1,$97,$D5,$19
	.dc.b	$6C,$7B,$19,$F7,$CF,$DC,$5C,$6A,$D5,$7C,$2A,$D4,$E6,$33,$18,$9D
	.dc.b	$C2,$30,$EA,$BA,$CE,$A7,$33,$E2,$9C,$BB,$AE,$CE,$5A,$F8,$E3,$6E
	.dc.b	$6B,$A4,$62,$3E,$CF,$E0,$0F,$D5,$1E,$47,$29,$4E,$63,$0D,$BF,$5B
	.dc.b	$67,$E1,$B5,$F9,$4D,$7B,$D7,$69,$E2,$D8,$E2,$D7,$8B,$5E,$35,$E2
	.dc.b	$8C,$47,$9B,$DD,$73,$79,$9B,$7E,$66,$CF,$99,$AF,$99,$AF,$C4,$6B
	.dc.b	$D6,$7A,$12,$7A,$36,$3A,$3F,$25,$7B,$2F,$67,$0D,$AF,$C1,$CE,$A7
	.dc.b	$24,$8C,$CE,$E7,$F1,$2D,$7F,$22,$D7,$F3,$0D,$7F,$6A,$D7,$B4,$74
	.dc.b	$9F,$95,$8C,$8F,$A6,$D9,$FA,$6D,$7E,$9B,$5F,$A7,$1C,$08,$F0,$7E
	.dc.b	$29,$1F,$8A,$D7,$E8,$5F,$C9,$6C,$65,$FA,$51,$96,$C6,$57,$86,$E0
	.dc.b	$7C,$AF,$F1,$32,$57,$F1,$36,$D7,$F1,$36,$BF,$C6,$DA,$FF,$1C,$AE
	.dc.b	$91,$88,$EF,$5D,$AB,$F8,$F7,$C7,$7F,$21,$23,$82,$4F,$F2,$18,$CB
	.dc.b	$63,$2F,$F2,$B8,$CB,$63,$F9,$57,$C7,$7F,$2D,$23,$82,$47,$F2,$D8
	.dc.b	$CB,$63,$2F,$F3,$78,$CB,$63,$F9,$B7,$C7,$7F,$39,$23,$82,$47,$F3
	.dc.b	$98,$CB,$63,$2F,$F4,$38,$CB,$63,$FA,$17,$C7,$7F,$45,$23,$82,$47
	.dc.b	$F4,$58,$CB,$63,$2F,$F4,$F8,$CB,$63,$FA,$77,$C7,$7F,$51,$23,$82
	.dc.b	$47,$F5,$18,$CB,$63,$2F,$F5,$B8,$CB,$63,$FA,$D7,$C7,$7F,$5D,$23
	.dc.b	$82,$47,$F5,$D8,$CB,$63,$2F,$F6,$78,$CB,$63,$FB,$37,$C7,$7F,$69
	.dc.b	$23,$82,$47,$F6,$98,$CB,$63,$2F,$F7,$38,$CB,$63,$FB,$97,$C7,$7F
	.dc.b	$75,$23,$82,$47,$F7,$58,$CB,$63,$2F,$F7,$F8,$CB,$63,$FB,$F7,$C7
	.dc.b	$7F,$80,$48,$E0,$91,$FE,$01,$8C,$B6,$32,$FF,$82,$E3,$2D,$8F,$F0
	.dc.b	$57,$C7,$7F,$83,$48,$E0,$91,$FE,$0D,$8C,$B6,$32,$FF,$85,$E3,$2D
	.dc.b	$8F,$F0,$B7,$C7,$7F,$86,$48,$E0,$91,$FE,$19,$8C,$B6,$32,$FF,$88
	.dc.b	$E3,$2D,$8F,$EF,$5F,$32,$F3,$3F,$B6,$76,$A3,$13,$A9,$CC,$62,$BF
	.dc.b	$C1,$2D,$7F,$C2,$6D,$7F,$C2,$0D,$7F,$C1,$8D,$AE,$D3,$D0,$95,$DC
	.dc.b	$FF,$1B,$7C,$7F,$F8,$E4,$8E,$09,$7F,$E3,$98,$CB,$63,$2F,$F9,$0E
	.dc.b	$32,$D8,$FE,$9C,$47,$FA,$CF,$47,$F5,$D6,$C6,$5F,$EB,$71,$96,$C6
	.dc.b	$5F,$EB,$B1,$97,$E0,$D7,$B1,$F8,$ED,$7E,$3B,$5D,$CF,$C1,$76,$D7
	.dc.b	$6D,$7F,$A1,$72,$2F,$DB,$18,$6B,$FD,$11,$BC,$DF,$D4,$DB,$3C,$6B
	.dc.b	$E3,$D7,$16,$BF,$D3,$3E,$3F,$FC,$D4,$48,$FF,$35,$36,$3F,$CD,$4D
	.dc.b	$7F,$CE,$4D,$7C,$ED,$7A,$47,$F9,$CF,$E3,$FF,$CE,$C4,$70,$48,$FF
	.dc.b	$3B,$46,$5B,$19,$7F,$D0,$11,$96,$C7,$F8,$5B,$E3,$FF,$C3,$A4,$70
	.dc.b	$48,$FF,$0C,$C6,$5B,$19,$7F,$C3,$71,$96,$C7,$E0,$5F,$F0,$DE,$7F
	.dc.b	$81,$6C,$65,$FF,$0E,$9F,$8C,$21,$F8,$A2,$9F,$8A,$78,$7C,$53,$CB
	.dc.b	$FF,$1C,$36,$3F,$C6,$AD,$9F,$F3,$CB,$5F,$F4,$0B,$5F,$F4,$43,$5F
	.dc.b	$F5,$2B,$53,$49,$3E,$02,$4F,$F9,$13,$E2,$9F,$15,$FE,$4B,$6C,$66
	.dc.b	$30,$47,$04,$BF,$F4,$7C,$F0,$6C,$ED,$B3,$B7,$FD,$37,$FA,$D3,$E0
	.dc.b	$7F,$CB,$4D,$9F,$F2,$BB,$67,$FD,$36,$D7,$8C,$F0,$24,$F0,$6C,$70
	.dc.b	$6B,$C2,$3F,$04,$FF,$96,$B3,$FC,$13,$63,$2F,$F9,$72,$32,$D8,$CB
	.dc.b	$FE,$70,$8C,$B6,$3F,$0D,$E1,$7F,$98,$C9,$A7,$E2,$E7,$F1,$27,$E1
	.dc.b	$C8,$7E,$1C,$FC,$49,$F9,$02,$9F,$73,$3F,$B2,$FE,$19,$AF,$E1,$9A
	.dc.b	$FE,$19,$AF,$E1,$9A,$F3,$AF,$C8,$C7,$D7,$8F,$F6,$7E,$7F,$6D,$B9
	.dc.b	$E5,$FF,$6C,$46,$5B,$19,$7F,$DE,$F1,$96,$C6,$54,$65,$B1,$FD,$63
	.dc.b	$F6,$3F,$F7,$CC,$7F,$B7,$DA,$FF,$B6,$DB,$C8,$69,$23,$91,$30,$FF
	.dc.b	$6F,$46,$5B,$19,$3E,$6E,$5F,$5A,$3E,$CE,$5F,$5E,$31,$3A,$8C,$12
	.dc.b	$76,$48,$D3,$5F,$BA,$7D,$B2,$9F,$82,$3F,$00,$62,$7B,$B6,$7B,$3D
	.dc.b	$08,$66,$7B,$3C,$88,$77,$29,$DC,$C1,$0C,$12,$7C,$2F,$06,$BB,$6B
	.dc.b	$B9,$F2,$BE,$33,$CA,$71,$72,$D9,$F3,$BC,$E3,$2D,$8C,$BD,$5B,$3E
	.dc.b	$02,$4F,$09,$D1,$E0,$39,$14,$C9,$E5,$20,$16,$FD,$D9,$E5,$FF,$8A
	.dc.b	$9B,$86,$EA,$56,$EB,$44,$9D,$4F,$59,$D4,$62,$3B,$D7,$47,$73,$A6
	.dc.b	$C6,$5E,$ED,$7A,$35,$E7,$5F,$E4,$0F,$25,$E2,$D2,$61,$E1,$C1,$3A
	.dc.b	$8C,$36,$FC,$6D,$7D,$ED,$7D,$8D,$7E,$4B,$67,$AD,$78,$27,$9A,$AF
	.dc.b	$85,$CF,$FE,$EB,$D7,$15,$F2,$23,$13,$AE,$6F,$E1,$51,$89,$D4,$60
	.dc.b	$95,$CA,$3E,$B4,$ED,$B3,$AA,$E9,$5B,$9F,$F1,$8B,$5F,$F7,$8F,$0D
	.dc.b	$BF,$F8,$00,$DB,$34,$97,$E7,$6D,$79,$9A,$F8,$9A,$FF,$90,$1E,$9A
	.dc.b	$C9,$F8,$0D,$9F,$7B,$5F,$5B,$5F,$F2,$4B,$55,$AD,$39,$8C,$7C,$97
	.dc.b	$AC,$97,$F3,$DB,$7F,$39,$AF,$CA,$6B,$FE,$58,$6B,$E0,$B4,$7C,$7F
	.dc.b	$22,$57,$DB,$6D,$7D,$A6,$BF,$59,$AF,$F9,$91,$AF,$56,$DB,$F4,$5F
	.dc.b	$C6,$57,$E3,$1B,$5F,$8B,$6B,$F8,$46,$BF,$C0,$67,$94,$FD,$F9,$CC
	.dc.b	$FC,$A6,$CE,$A7,$B1,$B2,$18,$3C,$5F,$35,$CC,$8E,$09,$5E,$58,$CB
	.dc.b	$63,$2F,$9E,$32,$D8,$F7,$7C,$87,$9C,$8E,$09,$1F,$06,$32,$D8,$CB
	.dc.b	$F0,$E3,$2D,$8F,$95,$F1,$FF,$0C,$8E,$09,$1F,$36,$32,$D8,$CB,$F3
	.dc.b	$E3,$2D,$8F,$AD,$F1,$FF,$3C,$8E,$09,$1F,$66,$32,$D8,$CB,$F6,$E3
	.dc.b	$2D,$8F,$F1,$AF,$2F,$ED,$9B,$BF,$F8,$DA,$31,$3F,$E8,$42,$1F,$E8
	.dc.b	$22,$9F,$E4,$32,$9F,$D3,$0F,$E9,$25,$3F,$AE,$1F,$D3,$DF,$EC,$BD
	.dc.b	$7F,$E7,$0A,$FF,$16,$92,$7F,$C7,$0D,$7F,$C7,$0D,$7F,$C7,$0D,$7F
	.dc.b	$C7,$0D,$CF,$F4,$1F,$1F,$F9,$72,$47,$E8,$5B,$3F,$A1,$6B,$FA,$16
	.dc.b	$BF,$A1,$D5,$18,$8F,$D0,$3F,$A3,$F8,$F9,$5F,$AD,$6C,$FE,$B5,$AF
	.dc.b	$EB,$5A,$F5,$6B,$FA,$B6,$DF,$EE,$7E,$3F,$F6,$04,$AF,$DD,$36,$BF
	.dc.b	$74,$D7,$F7,$4D,$7D,$4D,$7F,$70,$FE,$EF,$E3,$E4,$FF,$05,$6C,$FF
	.dc.b	$05,$6B,$FC,$15,$AF,$C5,$6B,$FC,$0D,$B5,$FC,$47,$E3,$FF,$84,$12
	.dc.b	$BF,$89,$36,$BF,$89,$35,$FE,$24,$D7,$E9,$35,$FE,$20,$FF,$13,$F8
	.dc.b	$F9,$3F,$C7,$5B,$3F,$C7,$5A,$FF,$1D,$6B,$F7,$5A,$FF,$1B,$6D,$7F
	.dc.b	$29,$F0,$3F,$90,$12,$BF,$95,$36,$BF,$95,$35,$FE,$54,$D7,$F1,$4D
	.dc.b	$7F,$94,$3F,$CA,$FD,$39,$3F,$CD,$5B,$3F,$CD,$5A,$FF,$35,$6B,$F9
	.dc.b	$56,$BF,$CC,$DB,$5F,$D0,$7E,$23,$F9,$C1,$2B,$FA,$13,$6B,$FA,$13
	.dc.b	$5F,$E8,$4D,$7F,$44,$D7,$FA,$03,$FE,$2B,$F5,$E4,$FF,$89,$9B,$3F
	.dc.b	$E2,$66,$BF,$E2,$66,$BF,$E2,$66,$BF,$E2,$66,$FA,$EF,$5D,$8F,$F8
	.dc.b	$46,$30,$D8,$CD,$FF,$92,$67,$2D,$8C,$DF,$F9,$42,$72,$D8,$CB,$FE
	.dc.b	$59,$8C,$B6,$32,$FF,$98,$E3,$2D,$8C,$A8,$CB,$FD,$97,$E3,$FF,$B4
	.dc.b	$12,$BF,$B9,$36,$3F,$B9,$35,$FE,$E4,$D7,$F8,$93,$5E,$91,$FE,$54
	.dc.b	$F8,$1F,$F2,$E1,$1C,$12,$3F,$CA,$F1,$96,$C6,$5F,$F2,$CC,$65,$B1
	.dc.b	$F2,$7E,$37,$E7,$11,$C1,$23,$E5,$46,$5B,$19,$7E,$54,$65,$B1,$FE
	.dc.b	$7F,$F8,$A7,$EF,$6E,$0F,$F2,$74,$61,$B1,$99,$E0,$DA,$FF,$45,$D6
	.dc.b	$E7,$35,$C0,$8E,$09,$3C,$2B,$93,$FE,$99,$AE,$4F,$F8,$97,$E4,$3F
	.dc.b	$C4,$04,$AF,$F1,$53,$63,$51,$82,$38,$21,$FE,$2D,$8E,$0F,$F8,$BF
	.dc.b	$E4,$3F,$C5,$84,$9F,$F1,$B3,$63,$FC,$6C,$D7,$84,$68,$8E,$9A,$FF
	.dc.b	$8E,$F1,$FF,$1C,$12,$3F,$C8,$4D,$8F,$F2,$13,$5E,$11,$A2,$3A,$6B
	.dc.b	$FE,$B0,$F2,$3F,$D7,$8F,$FA,$D1,$AE,$63,$FD,$59,$E4,$FF,$AC,$FC
	.dc.b	$8F,$F5,$D4,$E4,$E1,$3A,$AE,$31,$FE,$BA,$7F,$D4,$CF,$FA,$BB,$C8
	.dc.b	$AF,$FA,$B9,$AF,$FA,$B9,$B5,$FE,$AE,$AF,$F5,$CC,$63,$C9,$FF,$59
	.dc.b	$1F,$EB,$5E,$6F,$F5,$15,$FF,$A7,$AB,$FD,$42,$47,$FD,$57,$E4,$FF
	.dc.b	$A7,$49,$19,$8F,$F5,$4C,$61,$DC,$65,$FF,$3B,$3F,$E7,$D2,$3F,$E7
	.dc.b	$D6,$BF,$E7,$D6,$BC,$0D,$10,$D1,$4F,$F0,$B3,$FE,$6E,$7F,$CF,$8D
	.dc.b	$7F,$CF,$8D,$76,$FF,$9C,$DD,$B5,$DB,$5F,$F1,$17,$8B,$FE,$6E,$6C
	.dc.b	$7F,$9C,$9A,$7F,$9D,$48,$7F,$9C,$CA,$68,$C1,$0F,$F2,$51,$FE,$4D
	.dc.b	$21,$82,$01,$E0,$FF,$95,$9D,$99,$70,$F0,$37,$E1,$7F,$97,$4F,$F6
	.dc.b	$D3,$CB,$C9,$B3,$FE,$C2,$AE,$D3,$FE,$5A,$34,$E1,$FF,$2D,$99,$8E
	.dc.b	$0E,$4D,$95,$CB,$FE,$5A,$3F,$CB,$A4,$30,$43,$FC,$BA,$64,$A6,$4F
	.dc.b	$F2,$F9,$92,$9F,$E3,$58,$FF,$1D,$11,$FF,$1D,$35,$D1,$82,$18,$21
	.dc.b	$FE,$3D,$78,$3F,$E3,$27,$FC,$72,$47,$FC,$72,$D7,$FC,$72,$D7,$81
	.dc.b	$A2,$1A,$29,$FE,$33,$7F,$C7,$44,$7F,$C7,$4D,$7F,$C7,$4D,$78,$1A
	.dc.b	$21,$A2,$9F,$E0,$F7,$FC,$2C,$47,$FC,$2C,$D7,$FC,$2C,$D7,$81,$A2
	.dc.b	$1A,$29,$FE,$03,$E9,$FF,$0A,$12,$3F,$C2,$4D,$8F,$F0,$83,$5F,$F0
	.dc.b	$73,$5E,$51,$C0,$8F,$F8,$19,$FF,$04,$91,$C1,$0F,$F0,$4B,$96,$99
	.dc.b	$3F,$C1,$6E,$5A,$7E,$77,$27,$84,$F9,$0F,$F8,$00,$4A,$FC,$F4,$EA
	.dc.b	$31,$5C,$6B,$85,$69,$B1,$99,$DB,$FA,$EA,$E5,$5B,$9C,$CF,$FA,$C7
	.dc.b	$E4,$13,$CD,$B1,$DD,$AF,$3B,$46,$23,$94,$F1,$23,$C5,$AF,$18,$E0
	.dc.b	$AF,$E1,$C6,$1B,$3E,$DF,$90,$C5,$9A,$48,$D1,$2F,$E3,$46,$5B,$19
	.dc.b	$7D,$51,$97,$F8,$9F,$0B,$8D,$F0,$7E,$68,$C3,$63,$37,$FD,$C2,$30
	.dc.b	$47,$04,$AF,$E3,$95,$CA,$31,$1E,$0A,$E6,$E9,$AF,$F9,$9F,$CB,$72
	.dc.b	$7F,$9D,$5B,$5F,$E7,$36,$CF,$F9,$C5,$AF,$19,$E0,$49,$E0,$D8,$E0
	.dc.b	$D7,$84,$7E,$4B,$F4,$77,$57,$D5,$6C,$66,$30,$47,$04,$9F,$D0,$4F
	.dc.b	$06,$CE,$DB,$3B,$7F,$31,$2F,$91,$F0,$12,$7C,$8D,$9E,$AD,$78,$B5
	.dc.b	$FE,$B9,$B2,$BA,$57,$22,$4F,$9B,$E4,$7D,$C4,$8F,$84,$D8,$D4,$60
	.dc.b	$8E,$08,$7A,$A3,$83,$FD,$6F,$93,$FB,$19,$27,$FA,$E4,$7F,$5C,$34
	.dc.b	$E3,$93,$FA,$EC,$7F,$64,$21,$FD,$CE,$79,$3F,$D8,$DA,$F2,$31,$3C
	.dc.b	$E3,$31,$C8,$F0,$16,$7B,$BD,$67,$81,$1E,$64,$3B,$10,$E5,$3D,$9E
	.dc.b	$46,$CC,$C7,$37,$81,$A2,$9E,$43,$07,$03,$91,$82,$1B,$24,$FA,$9B
	.dc.b	$3E,$36,$CF,$85,$E8,$78,$27,$81,$27,$83,$67,$84,$F8,$23,$0F,$90
	.dc.b	$EA,$57,$2F,$59,$E6,$49,$E7,$1C,$0D,$38,$9E,$31,$E6,$21,$E1,$9E
	.dc.b	$4F,$76,$BC,$8C,$4F,$38,$CC,$72,$3C,$05,$9E,$EF,$59,$E0,$47,$99
	.dc.b	$0E,$C4,$39,$4F,$67,$91,$B3,$31,$CD,$E0,$68,$A7,$90,$C1,$C0,$E4
	.dc.b	$60,$86,$C9,$3E,$A6,$CF,$8D,$B3,$E1,$7A,$1F,$E5,$19,$FD,$0E,$4F
	.dc.b	$BB,$7C,$77,$FB,$A1,$B7,$FE,$ED,$6D,$FF,$B9,$9A,$FF,$B7,$DA,$FF
	.dc.b	$9B,$6B,$A5,$73,$25,$73,$6C,$F3,$6B,$FD,$5B,$F0,$07,$C0,$7F,$55
	.dc.b	$6D,$7F,$99,$78,$91,$8A,$FE,$A2,$D7,$95,$71,$25,$71,$9D,$4F,$FA
	.dc.b	$0E,$B9,$D6,$A3,$9B,$5F,$E6,$FF,$27,$FD,$14,$97,$FD,$25,$B1,$FD
	.dc.b	$09,$AF,$08,$D1,$1D,$35,$FE,$87,$EE,$FF,$51,$24,$7F,$55,$6C,$7F
	.dc.b	$4E,$6B,$C2,$34,$47,$4D,$7F,$A7,$FC,$9F,$F5,$D2,$47,$F6,$16,$C7
	.dc.b	$F5,$A6,$BC,$23,$44,$74,$D7,$FA,$DF,$C9,$FF,$69,$24,$7F,$6D,$6C
	.dc.b	$7F,$66,$6B,$C2,$34,$47,$4D,$7F,$B3,$FC,$9F,$F7,$52,$47,$F7,$96
	.dc.b	$C7,$F7,$26,$BC,$23,$44,$74,$D7,$FB,$9E,$9F,$E0,$12,$47,$F8,$15
	.dc.b	$B1,$FD,$F9,$AF,$08,$D1,$1D,$35,$FE,$FF,$E6,$7F,$83,$49,$1F,$E1
	.dc.b	$16,$C7,$F8,$29,$AF,$08,$D1,$1D,$35,$FF,$05,$F9,$3F,$E1,$92,$47
	.dc.b	$F8,$75,$B1,$FE,$16,$6B,$C2,$34,$47,$4D,$7F,$C2,$FC,$FF,$E2,$52
	.dc.b	$47,$F8,$A5,$B1,$FE,$22,$6B,$C2,$34,$47,$4D,$7F,$C4,$7F,$27,$FE
	.dc.b	$31,$24,$7F,$8D,$5B,$1F,$E2,$E6,$BC,$23,$44,$74,$D7,$FD,$FA,$7F
	.dc.b	$BE,$9F,$F0,$87,$83,$FE,$FB,$69,$23,$FD,$E1,$19,$6C,$65,$FE,$73
	.dc.b	$C9,$FC,$C4,$9C,$BF,$CC,$5F,$EE,$9E,$AB,$DD,$FF,$41,$D2,$31,$7F
	.dc.b	$E2,$46,$C7,$FB,$CC,$D4,$F0,$8C,$C7,$FB,$E4,$CC,$F1,$9D,$11,$C1
	.dc.b	$FE,$5D,$7F,$CB,$84,$70,$48,$FF,$2C,$C6,$5B,$19,$7F,$CA,$F1,$96
	.dc.b	$C7,$F9,$A7,$E4,$FF,$CD,$04,$70,$48,$FF,$32,$C6,$5B,$19,$7F,$CC
	.dc.b	$71,$96,$C7,$F9,$D7,$E4,$FF,$CE,$84,$70,$48,$FF,$38,$C6,$5B,$19
	.dc.b	$7F,$CD,$F1,$96,$C7,$FA,$07,$E4,$FF,$D0,$04,$70,$48,$FF,$3E,$C6
	.dc.b	$5B,$19,$7F,$CF,$71,$96,$C7,$FA,$37,$E4,$FF,$D1,$84,$70,$48,$FF
	.dc.b	$44,$C6,$5B,$19,$7F,$D0,$F1,$96,$C7,$FA,$67,$E4,$FF,$D3,$04,$70
	.dc.b	$48,$FF,$4A,$C6,$5B,$19,$7F,$D2,$71,$96,$C7,$FA,$97,$C3,$FF,$52
	.dc.b	$11,$C1,$23,$FD,$43,$19,$6C,$65,$FF,$4F,$C6,$5B,$1F,$EB,$1F,$33
	.dc.b	$FD,$60,$47,$04,$8F,$F5,$6C,$65,$B1,$97,$FD,$57,$19,$6C,$18,$7F
	.dc.b	$A9,$BE,$1B,$11,$88,$C1,$FE,$F5,$8C,$B9,$F0,$91,$88,$E7,$18,$78
	.dc.b	$12,$7A,$3F,$F0,$06,$78,$3D,$A3,$33,$C5,$B3,$A9,$E0,$23,$9F,$7F
	.dc.b	$FB,$1F,$B3,$13,$E9,$8C,$4E,$BC,$35,$6A,$BE,$0D,$69,$E1,$3A,$AE
	.dc.b	$15,$B9,$FF,$1D,$4E,$5D,$C6,$5F,$F7,$3C,$65,$EA,$D7,$8C,$6E,$BC
	.dc.b	$11,$FE,$3D,$F8,$EF,$F1,$E1,$2F,$FC,$68,$D8,$FF,$1D,$B5,$E1,$1A
	.dc.b	$23,$A6,$BF,$53,$07,$C9,$FD,$B2,$4F,$CE,$6C,$FD,$D9,$DC,$E4,$91
	.dc.b	$99,$DB,$D5,$AF,$C6,$EF,$F7,$64,$69,$2B,$EA,$36,$7E,$8B,$5F,$92
	.dc.b	$D7,$A4,$70,$3F,$CB,$1F,$1D,$FE,$63,$23,$82,$47,$F9,$8A,$32,$D8
	.dc.b	$CB,$FE,$6D,$8C,$B6,$3F,$23,$F2,$7F,$A1,$23,$82,$47,$C7,$8C,$B6
	.dc.b	$32,$FD,$18,$CB,$63,$F3,$BF,$27,$FA,$D2,$38,$24,$7D,$38,$CB,$63
	.dc.b	$2F,$DD,$8C,$B6,$3F,$53,$F2,$7F,$B9,$23,$82,$47,$DF,$8C,$B6,$32
	.dc.b	$FE,$26,$32,$D8,$FD,$AF,$C9,$FF,$05,$23,$82,$47,$E2,$E3,$2D,$8C
	.dc.b	$BF,$95,$8C,$B6,$3F,$D4,$84,$7F,$D4,$8D,$7F,$D5,$DE,$0F,$FA,$99
	.dc.b	$FF,$57,$79,$1F,$EB,$05,$C1,$FE,$B0,$AE,$11,$AA,$E0,$EA,$B8,$78
	.dc.b	$FF,$EA,$0A,$FF,$52,$78,$FF,$E9,$FA,$FF,$52,$79,$1F,$EA,$65,$5B
	.dc.b	$7F,$D4,$B5,$C0,$8E,$08,$7F,$AC,$9C,$BF,$EA,$33,$FD,$42,$43,$04
	.dc.b	$3F,$9C,$72,$BC,$AF,$E1,$90,$8C,$73,$7F,$AB,$A3,$73,$A9,$CD,$7F
	.dc.b	$AB,$67,$35,$FE,$84,$77,$E3,$D7,$A5,$A3,$17,$E1,$6C,$66,$FF,$D1
	.dc.b	$15,$C9,$AF,$29,$FF,$42,$BE,$4D,$2B,$D9,$B7,$ED,$6D,$FD,$B8,$C5
	.dc.b	$FF,$A2,$6B,$93,$5F,$3F,$8F,$FE,$86,$7E,$0B,$5C,$C7,$FA,$15,$CB
	.dc.b	$5C,$9F,$E8,$A2,$1F,$E8,$F3,$FC,$BE,$43,$FC,$D3,$E1,$3C,$4F,$F3
	.dc.b	$24,$61,$AF,$F6,$E8,$D4,$FF,$99,$E3,$1E,$27,$F9,$94,$91,$FD,$9A
	.dc.b	$13,$FE,$6A,$8C,$29,$FF,$34,$BF,$E6,$CF,$0F,$FC,$BD,$5F,$E6,$7F
	.dc.b	$13,$FC,$CA,$FF,$7E,$6B,$FE,$01,$6B,$FE,$68,$38,$47,$27,$FC,$D9
	.dc.b	$1C,$32,$FC,$87,$85,$FE,$3B,$9F,$C6,$B5,$FC,$3B,$63,$78,$FF,$90
	.dc.b	$E7,$6D,$76,$D8,$FC,$E7,$85,$FE,$3B,$9F,$CC,$B5,$FC,$9B,$63,$73
	.dc.b	$FE,$43,$9D,$B5,$DB,$63,$F5,$1E,$17,$F8,$EE,$7F,$4A,$D7,$F3,$ED
	.dc.b	$8D,$CF,$F9,$0E,$76,$D7,$6D,$E6,$FE,$F7,$1F,$DF,$A3,$FC,$12,$47
	.dc.b	$FC,$0E,$D7,$FC,$0B,$C5,$FD,$98,$91,$98,$E4,$70,$3F,$BE,$10,$FF
	.dc.b	$00,$94,$FE,$2B,$E0,$7F,$81,$5B,$1F,$E0,$26,$9B,$3F,$8C,$F2,$A7
	.dc.b	$84,$7F,$8D,$7B,$FF,$BD,$B6,$BF,$C0,$2F,$F7,$F6,$BF,$E4,$98,$DE
	.dc.b	$3F,$E4,$46,$F8,$1F,$E0,$A9,$FF,$04,$57,$F8,$F2,$FF,$C9,$84,$70
	.dc.b	$48,$FF,$09,$46,$5B,$19,$7F,$C3,$91,$96,$C7,$F9,$23,$E4,$FF,$CB
	.dc.b	$04,$70,$48,$FF,$0F,$46,$5B,$19,$7F,$C5,$11,$96,$C7,$F9,$53,$BF
	.dc.b	$FC,$C8,$47,$04,$8F,$F1,$54,$65,$B1,$97,$FC,$69,$19,$6C,$7F,$98
	.dc.b	$3E,$4F,$FC,$E0,$47,$04,$8F,$F1,$B4,$65,$B1,$97,$FC,$81,$19,$6C
	.dc.b	$7F,$9C,$BA,$9E,$A3,$C6,$FF,$2D,$46,$1B,$19,$9E,$0D,$AF,$F4,$35
	.dc.b	$6E,$73,$5C,$08,$E0,$93,$C2,$B9,$3F,$E9,$2A,$E4,$FF,$94,$7B,$7F
	.dc.b	$D2,$04,$AF,$F4,$1B,$63,$51,$82,$38,$21,$FE,$61,$8E,$0F,$FA,$77
	.dc.b	$9F,$F9,$63,$FE,$94,$6B,$9E,$8F,$E6,$6E,$5A,$E6,$BF,$99,$C6,$23
	.dc.b	$F9,$DC,$60,$8F,$16,$BC,$4C,$10,$C1,$0E,$66,$4A,$64,$FF,$5D,$4F
	.dc.b	$A1,$C9,$0C,$CF,$95,$B3,$E2,$74,$7F,$4E,$8D,$36,$34,$F0,$3C,$45
	.dc.b	$3C,$46,$CC,$94,$CC,$F9,$5E,$07,$81,$C3,$B3,$13,$FD,$70,$93,$FE
	.dc.b	$C2,$F9,$3F,$F6,$91,$1C,$12,$7F,$D3,$11,$96,$C6,$5F,$F5,$14,$65
	.dc.b	$B1,$FA,$3F,$05,$E1,$7E,$89,$B1,$98,$C1,$1C,$12,$7F,$47,$3C,$1B
	.dc.b	$3B,$6C,$ED,$FF,$5A,$7C,$9F,$FB,$B4,$91,$FE,$E0,$6C,$7F,$AC,$DA
	.dc.b	$F0,$8D,$11,$D3,$5F,$F6,$07,$C9,$FF,$BE,$49,$1F,$EE,$C6,$C7,$FA
	.dc.b	$FD,$AF,$08,$D1,$1D,$35,$FF,$66,$71,$FF,$B8,$79,$4C,$9F,$F7,$CB
	.dc.b	$93,$13,$FE,$CD,$70,$48,$D9,$23,$4D,$7B,$A7,$FD,$AD,$F2,$5D,$49
	.dc.b	$19,$8F,$F7,$04,$61,$DC,$65,$FF,$35,$CA,$75,$1C,$27,$4F,$F9,$B6
	.dc.b	$75,$18,$AF,$7D,$70,$AE,$7F,$25,$EF,$25,$6E,$D1,$89,$FF,$77,$35
	.dc.b	$FF,$7B,$35,$E7,$3C,$63,$11,$E3,$F6,$BE,$61,$27,$B3,$63,$FD,$F4
	.dc.b	$D7,$E2,$61,$F5,$A3,$11,$F0,$9A,$F5,$9F,$47,$83,$F6,$48,$E0,$91
	.dc.b	$F2,$EB,$F0,$33,$A8,$C6,$1F,$E6,$06,$C6,$5F,$1C,$65,$B1,$F8,$7D
	.dc.b	$E5,$FC,$96,$CF,$46,$BC,$9A,$FD,$C6,$BF,$65,$B7,$F9,$6F,$8E,$FC
	.dc.b	$A1,$2B,$EB,$36,$BD,$2D,$7C,$ED,$7F,$12,$D7,$F0,$CF,$E7,$FD,$59
	.dc.b	$3F,$81,$6C,$FC,$46,$BF,$09,$AF,$E5,$1A,$FE,$45,$AF,$FA,$13,$0F
	.dc.b	$F4,$6B,$FD,$09,$AE,$67,$FD,$1C,$E5,$AE,$4F,$F4,$D1,$FC,$C1,$FE
	.dc.b	$7E,$FF,$33,$3F,$9B,$90,$C1,$38,$BF,$9C,$9F,$E9,$18,$FF,$4C,$78
	.dc.b	$5F,$E8,$C2,$78,$7F,$E8,$BA,$FF,$49,$35,$FF,$49,$35,$FF,$49,$1F
	.dc.b	$E9,$07,$F8,$67,$E9,$0A,$A4,$3F,$64,$E5,$A6,$67,$F2,$91,$96,$C6
	.dc.b	$5F,$D1,$C6,$5B,$19,$7F,$DD,$F1,$FE,$26,$8E,$FD,$1E,$42,$38,$24
	.dc.b	$F8,$23,$2D,$8C,$BE,$08,$CB,$E2,$F8,$FF,$79,$27,$DA,$D8,$F5,$B5
	.dc.b	$E1,$1A,$23,$A6,$BE,$EF,$8F,$F9,$64,$8F,$92,$D8,$F8,$ED,$78,$46
	.dc.b	$88,$E9,$AF,$CA,$F8,$FF,$AE,$48,$FA,$AD,$8F,$A6,$D7,$84,$68,$8E
	.dc.b	$9A,$FD,$6F,$8F,$FC,$19,$23,$F0,$2D,$8F,$BE,$D7,$84,$68,$8E,$9A
	.dc.b	$FE,$0B,$E3,$FF,$1E,$48,$FC,$6B,$63,$F1,$6D,$78,$46,$88,$E9,$AF
	.dc.b	$E3,$BE,$3F,$F3,$64,$8F,$CC,$B6,$3F,$2E,$D7,$84,$68,$8E,$9A,$FE
	.dc.b	$6B,$E3,$FF,$4E,$48,$FD,$2B,$63,$F4,$6D,$78,$46,$88,$E9,$AF,$F4
	.dc.b	$8F,$90,$75,$7F,$7F,$6C,$FF,$63,$6C,$FF,$43,$6B,$C6,$78,$12,$78
	.dc.b	$36,$38,$35,$E1,$1F,$E9,$7F,$94,$7A,$BF,$E9,$66,$C6,$63,$04,$70
	.dc.b	$49,$FF,$4B,$CF,$06,$CE,$DB,$3B,$7F,$82,$FC,$8F,$F1,$22,$47,$F1
	.dc.b	$06,$C7,$F0,$E6,$BC,$23,$44,$74,$D7,$F8,$8F,$C6,$7F,$1E,$24,$7F
	.dc.b	$1C,$6C,$7F,$1A,$6B,$C2,$34,$47,$4D,$7F,$D5,$9F,$3E,$F8,$2F,$F5
	.dc.b	$EB,$67,$FD,$84,$D9,$FF,$5D,$B5,$E3,$3C,$09,$3C,$1B,$1C,$1A,$F0
	.dc.b	$8F,$E6,$BF,$25,$FC,$E8,$8E,$09,$1F,$CD,$A3,$2D,$8C,$BF,$CD,$A3
	.dc.b	$2D,$8F,$E8,$3F,$13,$FD,$18,$8E,$09,$1F,$D0,$A3,$2D,$8C,$BF,$D0
	.dc.b	$A3,$2D,$8F,$E9,$BF,$25,$FD,$48,$8E,$09,$1F,$D3,$A3,$2D,$8C,$BF
	.dc.b	$D3,$A3,$2D,$8F,$EB,$3F,$21,$FD,$78,$8E,$09,$1F,$D6,$A3,$2D,$8C
	.dc.b	$BF,$D6,$A3,$2D,$8F,$F5,$03,$FE,$5E,$F2,$3F,$D4,$0D,$8C,$BF,$E5
	.dc.b	$D8,$CB,$63,$3E,$07,$F9,$F2,$72,$D8,$FF,$4D,$4F,$F9,$CF,$CA,$79
	.dc.b	$3F,$EA,$18,$FF,$7E,$35,$D3,$54,$70,$F2,$7F,$D4,$D1,$C1,$AF,$06
	.dc.b	$BC,$27,$94,$F1,$AF,$F5,$04,$FF,$9F,$3C,$9F,$F5,$3B,$63,$2F,$F9
	.dc.b	$EE,$32,$D8,$CE,$1F,$E9,$A9,$CB,$63,$FC,$E4,$49,$FF,$3C,$B5,$FF
	.dc.b	$3C,$B5,$FF,$3C,$9A,$21,$A2,$9A,$3F,$CC,$7E,$3F,$F9,$CA,$7F,$CE
	.dc.b	$D3,$FE,$7C,$24,$FF,$9F,$1A,$F0,$6B,$FE,$7B,$6B,$FE,$7B,$3F,$CE
	.dc.b	$2F,$F9,$E1,$FF,$37,$B5,$FF,$51,$F8,$7F,$E7,$56,$C6,$5F,$F5,$14
	.dc.b	$65,$B1,$97,$FC,$98,$7F,$AB,$7C,$3F,$F3,$E3,$FE,$B3,$6B,$FE,$53
	.dc.b	$3F,$CA,$25,$38,$4F,$F9,$6A,$37,$1D,$0E,$51,$FE,$43,$7F,$D6,$B1
	.dc.b	$FE,$52,$F1,$B1,$1E,$16,$BF,$E5,$D3,$6E,$1F,$F2,$E4,$23,$11,$FE
	.dc.b	$5D,$7F,$CA,$9E,$27,$F9,$6A,$FF,$CB,$BE,$26,$08,$E0,$95,$FE,$C8
	.dc.b	$9E,$0F,$F8,$9C,$87,$F8,$B0,$A7,$F9,$3B,$C1,$FF,$18,$36,$7F,$C5
	.dc.b	$AD,$36,$32,$78,$8A,$E3,$39,$87,$85,$FE,$2F,$8F,$F2,$DB,$5D,$36
	.dc.b	$B9,$B6,$78,$DF,$F8,$CA,$B8,$B5,$ED,$E1,$7F,$8B,$DF,$F3,$3B,$5C
	.dc.b	$C7,$F8,$B9,$CB,$5C,$9F,$E3,$32,$1F,$E3,$83,$FC,$06,$FF,$81,$88
	.dc.b	$E0,$87,$F8,$18,$A0,$F8,$DD,$FF,$E1,$27,$18,$23,$11,$AE,$FF,$F0
	.dc.b	$83,$A8,$CF,$13,$BF,$FC,$11,$3D,$4F,$E3,$2E,$1F,$E4,$4D,$73,$DF
	.dc.b	$FE,$0D,$78,$4F,$38,$C4,$70,$6B,$C3,$1F,$F0,$93,$97,$FC,$10,$FF
	.dc.b	$A5,$3B,$FF,$C1,$ED,$8C,$BF,$81,$8C,$B6,$33,$97,$E2,$27,$2D,$8F
	.dc.b	$F5,$5F,$C7,$FF,$AC,$08,$E0,$95,$FE,$AC,$8C,$B6,$32,$FF,$AB,$23
	.dc.b	$2D,$8F,$F5,$BF,$C7,$FF,$AF,$08,$E0,$91,$FE,$B8,$8C,$B6,$32,$FF
	.dc.b	$AE,$23,$2D,$8F,$F6,$1F,$C7,$FF,$B2,$08,$E0,$91,$FE,$C4,$8C,$B6
	.dc.b	$32,$FF,$B1,$23,$2D,$8F,$F6,$7F,$C7,$7F,$B5,$08,$E0,$91,$FE,$D0
	.dc.b	$8C,$B6,$32,$FF,$B4,$23,$2D,$8F,$F3,$57,$97,$FE,$6D,$23,$82,$47
	.dc.b	$F9,$B6,$32,$D8,$CB,$FE,$77,$8C,$B6,$3F,$DC,$FF,$1D,$FE,$EC,$23
	.dc.b	$82,$47,$FB,$A2,$32,$D8,$CB,$FE,$E8,$8C,$B6,$3F,$48,$FF,$2F,$3F
	.dc.b	$99,$38,$7F,$97,$F3,$FF,$2E,$E8,$FE,$71,$39,$9F,$E6,$0D,$8C,$C7
	.dc.b	$F3,$88,$C4,$F1,$23,$82,$46,$C8,$78,$8A,$79,$48,$79,$1E,$2E,$C8
	.dc.b	$68,$F2,$9C,$8E,$05,$32,$70,$32,$5D,$3F,$9B,$FC,$6F,$F5,$16,$CF
	.dc.b	$C5,$6C,$FC,$A6,$BF,$C3,$DA,$FF,$1A,$E5,$57,$36,$D7,$36,$CF,$36
	.dc.b	$BC,$DA,$F3,$9E,$47,$97,$F4,$97,$B5,$49,$7E,$78,$CB,$63,$35,$E9
	.dc.b	$9C,$B6,$32,$FA,$E3,$2D,$8C,$BF,$3F,$2F,$C0,$BF,$2C,$F9,$A6,$72
	.dc.b	$FC,$0B,$F4,$63,$6E,$6F,$ED,$BF,$7B,$26,$0F,$85,$F7,$12,$BD,$8D
	.dc.b	$AF,$53,$5F,$43,$5F,$E9,$73,$CE,$78,$92,$3E,$17,$8F,$F2,$C9,$1F
	.dc.b	$25,$B1,$A8,$C1,$1C,$10,$F9,$11,$C1,$F9,$5E,$DF,$D7,$24,$FD,$56
	.dc.b	$C7,$D3,$6B,$C2,$34,$47,$4D,$7E,$B7,$2F,$E0,$C9,$1F,$81,$6C,$7D
	.dc.b	$F6,$BC,$23,$44,$74,$D7,$F0,$5F,$1B,$F8,$F2,$47,$E3,$5B,$1F,$8B
	.dc.b	$6B,$C2,$34,$47,$4D,$7F,$1D,$F1,$BF,$9B,$24,$7E,$65,$B1,$F9,$76
	.dc.b	$BC,$23,$44,$74,$D7,$F3,$5F,$09,$FA,$72,$47,$E9,$5B,$1F,$A3,$6B
	.dc.b	$C2,$34,$47,$4D,$7F,$4D,$F0,$9F,$B3,$24,$7E,$C5,$B1,$FA,$F6,$BC
	.dc.b	$23,$44,$74,$D7,$F6,$5F,$07,$FB,$F2,$47,$EF,$5B,$1F,$BB,$6B,$C2
	.dc.b	$34,$47,$4D,$7F,$7D,$F1,$BF,$C2,$C9,$1F,$C2,$5B,$1F,$C1,$DA,$F0
	.dc.b	$8D,$11,$D3,$5F,$E1,$5F,$1B,$FC,$5C,$91,$FC,$55,$B1,$FC,$4D,$AF
	.dc.b	$08,$D1,$1D,$35,$FE,$2D,$F1,$BF,$C8,$C9,$1F,$C8,$5B,$1F,$C7,$DA
	.dc.b	$F0,$8D,$11,$D3,$5F,$F4,$FB,$FD,$7F,$C7,$FF,$50,$11,$C1,$23,$FD
	.dc.b	$4E,$7F,$A8,$CF,$E9,$A7,$F1,$F3,$F9,$17,$27,$F2,$47,$F9,$47,$27
	.dc.b	$FA,$76,$7F,$91,$C7,$F9,$96,$30,$FF,$A5,$FC,$7F,$F4,$DB,$67,$FD
	.dc.b	$36,$FF,$7C,$F1,$FF,$D3,$6D,$9D,$B6,$76,$FF,$24,$8F,$E9,$44,$8F
	.dc.b	$E9,$0D,$8F,$E8,$CD,$7F,$A2,$35,$E5,$1C,$08,$FF,$4D,$F8,$DF,$EA
	.dc.b	$C4,$8F,$EA,$8D,$8D,$46,$08,$E0,$87,$F5,$38,$E0,$FF,$55,$F7,$BF
	.dc.b	$B1,$12,$7F,$B0,$36,$3F,$AF,$35,$E1,$1A,$23,$A6,$BF,$D8,$7E,$1F
	.dc.b	$FB,$71,$23,$FB,$63,$63,$FB,$53,$5E,$11,$A2,$3A,$6B,$FD,$B7,$E1
	.dc.b	$FF,$BD,$12,$3F,$BC,$36,$3F,$BB,$35,$E1,$1A,$23,$A6,$BF,$DE,$7E
	.dc.b	$1F,$FC,$0C,$48,$FF,$02,$36,$3F,$C0,$4D,$78,$46,$88,$E9,$AF,$F8
	.dc.b	$17,$E3,$7F,$C2,$44,$8F,$F0,$83,$63,$FC,$1C,$D7,$84,$68,$8E,$9A
	.dc.b	$FF,$84,$7E,$37,$FC,$3C,$48,$FF,$0E,$36,$3F,$C3,$4D,$78,$46,$88
	.dc.b	$E9,$AF,$F8,$DF,$F4,$AF,$F6,$94,$7F,$8F,$5B,$1F,$E3,$D6,$BB,$3F
	.dc.b	$C8,$8E,$1F,$F6,$BB,$5F,$F8,$00,$DE,$E3,$49,$1C,$89,$3E,$1F,$89
	.dc.b	$EA,$47,$04,$8F,$24,$65,$B1,$97,$D9,$19,$6C,$7B,$7E,$3B,$D2,$47
	.dc.b	$04,$8F,$81,$19,$6C,$65,$F9,$11,$96,$C7,$C9,$F5,$7E,$29,$1C,$12
	.dc.b	$3E,$64,$65,$B1,$97,$EA,$46,$5B,$1F,$57,$D5,$FA,$24,$70,$48,$FB
	.dc.b	$11,$96,$C6,$5F,$C0,$46,$5B,$1F,$81,$F8,$EF,$BA,$47,$04,$8F,$C2
	.dc.b	$46,$5B,$19,$7F,$19,$19,$6C,$7E,$37,$E2,$FF,$12,$47,$04,$8F,$C8
	.dc.b	$46,$5B,$19,$7F,$31,$19,$6C,$7E,$67,$E2,$FF,$2A,$47,$04,$8F,$CE
	.dc.b	$46,$5B,$19,$7F,$49,$19,$6C,$7E,$97,$E2,$FF,$42,$47,$04,$8F,$D4
	.dc.b	$46,$5B,$19,$7F,$61,$19,$6C,$7E,$C7,$E2,$FF,$5A,$47,$04,$8F,$DA
	.dc.b	$46,$5B,$19,$7F,$79,$19,$6C,$7F,$9F,$DF,$F4,$2B,$5F,$F4,$2F,$8F
	.dc.b	$FE,$87,$5B,$FF,$A4,$A3,$0D,$9E,$13,$AF,$1F,$FD,$18,$FF,$A4,$A3
	.dc.b	$33,$FE,$9A,$F1,$FF,$D0,$53,$FE,$92,$6B,$C5,$B1,$B9,$FF,$46,$4F
	.dc.b	$FA,$29,$AF,$16,$BE,$1F,$1F,$FD,$0E,$FF,$A7,$9A,$E6,$3F,$D0,$CE
	.dc.b	$5A,$E4,$FF,$45,$90,$FF,$48,$1F,$E5,$E2,$1F,$E6,$32,$9F,$EA,$CF
	.dc.b	$13,$FC,$CE,$D8,$FF,$32,$34,$D9,$F1,$E3,$6F,$C9,$F1,$3F,$CC,$6F
	.dc.b	$FA,$D9,$AE,$67,$FC,$C4,$E5,$AE,$4F,$F3,$39,$0F,$F3,$61,$FE,$5E
	.dc.b	$21,$FE,$63,$29,$FE,$C4,$F1,$3F,$CC,$ED,$8F,$F3,$23,$4D,$9F,$7E
	.dc.b	$36,$FF,$8D,$7C,$1F,$F2,$13,$FE,$40,$F0,$11,$88,$FF,$24,$99,$7F
	.dc.b	$C8,$0F,$F9,$0C,$FF,$20,$F8,$38,$7F,$C8,$E4,$13,$FE,$42,$7F,$C7
	.dc.b	$47,$F9,$04,$86,$08,$09,$E0,$FF,$91,$1C,$07,$0E,$BC,$1F,$F2,$3B
	.dc.b	$FE,$E8,$EF,$FF,$0E,$37,$2F,$F0,$CB,$4F,$F0,$E1,$0F,$F1,$09,$4D
	.dc.b	$18,$21,$FE,$F4,$EF,$FF,$0E,$36,$3F,$C3,$2D,$3F,$C3,$84,$3F,$C4
	.dc.b	$25,$34,$60,$87,$FC,$00,$EF,$FF,$0E,$36,$3F,$C3,$2D,$3F,$C3,$84
	.dc.b	$3F,$C4,$25,$34,$60,$87,$F8,$03,$B3,$FB,$73,$63,$FB,$93,$4F,$EE
	.dc.b	$A4,$3F,$B9,$94,$D1,$82,$1F,$E0,$CE,$CF,$ED,$CD,$8F,$EE,$4D,$3F
	.dc.b	$BA,$90,$FE,$E6,$53,$46,$08,$7F,$86,$3B,$3F,$B7,$36,$3F,$B9,$34
	.dc.b	$FE,$EA,$43,$FB,$99,$4D,$18,$21,$FE,$24,$F8,$BF,$F1,$39,$23,$FC
	.dc.b	$5C,$D8,$FF,$17,$35,$FF,$1A,$35,$E5,$1C,$08,$FF,$8D,$7D,$CF,$F1
	.dc.b	$99,$23,$FC,$74,$D8,$D4,$60,$8E,$08,$7F,$90,$E3,$83,$FE,$45,$F7
	.dc.b	$3F,$C7,$E4,$9F,$F2,$33,$63,$FC,$8C,$D7,$84,$68,$8E,$9A,$FF,$94
	.dc.b	$7D,$8F,$F2,$59,$23,$FC,$A4,$D8,$FF,$29,$35,$E1,$1A,$23,$A6,$BF
	.dc.b	$E5,$DF,$43,$FC,$AE,$48,$FF,$2F,$36,$3F,$CB,$CD,$78,$46,$88,$E9
	.dc.b	$AF,$F9,$A7,$E2,$FF,$CC,$64,$8F,$F3,$53,$63,$FC,$D4,$D7,$84,$68
	.dc.b	$8E,$9A,$FF,$24,$7F,$DC,$6F,$F2,$2E,$3F,$E4,$CD,$8C,$F1,$FF,$25
	.dc.b	$8F,$E5,$84,$3F,$95,$39,$6C,$F2,$8C,$47,$33,$FD,$EB,$3D,$27,$44
	.dc.b	$70,$48,$D9,$0F,$41,$4F,$11,$D4,$C9,$0C,$CF,$88,$93,$DE,$39,$1E
	.dc.b	$89,$F0,$3C,$4C,$1F,$CF,$0F,$E8,$13,$EF,$E9,$FF,$48,$FC,$5F,$FA
	.dc.b	$60,$95,$FE,$9F,$6D,$7F,$A7,$DA,$FF,$A9,$5A,$FF,$A7,$9A,$FF,$A7
	.dc.b	$9F,$F5,$1F,$BD,$27,$FD,$54,$D9,$FF,$55,$35,$FF,$58,$35,$FF,$54
	.dc.b	$B5,$FF,$54,$B6,$BF,$D5,$3C,$FF,$EB,$02,$57,$FA,$DD,$B5,$FE,$B7
	.dc.b	$6B,$FE,$BD,$6B,$FE,$B6,$6B,$FE,$B6,$7F,$D7,$78,$C9,$FF,$61,$36
	.dc.b	$7F,$D8,$4D,$7F,$D9,$0D,$7F,$D8,$2D,$7F,$D8,$2D,$AF,$F6,$0F,$C5
	.dc.b	$FF,$B2,$09,$5F,$EC,$F6,$D7,$FB,$3D,$AF,$FB,$55,$AF,$FB,$39,$AF
	.dc.b	$FB,$39,$FF,$69,$FC,$5C,$9F,$F6,$D3,$67,$FD,$B4,$D7,$FD,$C0,$D7
	.dc.b	$FD,$B2,$D7,$FD,$B2,$DA,$79,$6A,$95,$F8,$F9,$CC,$62,$3F,$C1,$B1
	.dc.b	$88,$FC,$A4,$7F,$9B,$27,$94,$62,$75,$3C,$E7,$2D,$9E,$8F,$F8,$6E
	.dc.b	$76,$F6,$9D,$B6,$76,$F7,$6C,$F0,$8E,$52,$F8,$BC,$B6,$BC,$8D,$9F
	.dc.b	$24,$E8,$93,$A6,$C6,$A3,$04,$70,$7B,$7E,$2F,$B9,$23,$BB,$63,$51
	.dc.b	$82,$38,$21,$F0,$63,$83,$F0,$FE,$0B,$D6,$49,$F5,$B6,$3E,$73,$5E
	.dc.b	$11,$A2,$3A,$6B,$F3,$FE,$0B,$E3,$92,$3E,$3B,$63,$ED,$35,$E1,$1A
	.dc.b	$23,$A6,$BF,$6F,$E0,$BE,$99,$23,$E9,$B6,$3F,$0C,$D7,$84,$68,$8E
	.dc.b	$9A,$FF,$9C,$88,$7F,$98,$B8,$7F,$CE,$F1,$F9,$68,$FF,$37,$EF,$FE
	.dc.b	$79,$74,$D7,$FD,$01,$1F,$CC,$9F,$E7,$3B,$FF,$37,$7F,$9A,$1F,$CF
	.dc.b	$CE,$CF,$F9,$B1,$FF,$37,$47,$7E,$FF,$F3,$8C,$7F,$9C,$FB,$F4,$D9
	.dc.b	$D4,$71,$DF,$F6,$7F,$17,$F9,$82,$4F,$E6,$1B,$5F,$AB,$6B,$FA,$76
	.dc.b	$BF,$A2,$6B,$FB,$57,$F4,$7F,$17,$27,$F4,$6D,$9F,$DB,$35,$FD,$93
	.dc.b	$5F,$D6,$B5,$FE,$00,$DA,$FE,$17,$F1,$7F,$B0,$25,$7E,$C1,$B5,$FC
	.dc.b	$0D,$AF,$EF,$DA,$FE,$E9,$AF,$F0,$D7,$F7,$7F,$05,$27,$F7,$6D,$9F
	.dc.b	$E1,$CD,$7F,$85,$35,$FE,$0A,$D7,$F8,$C3,$6B,$F9,$1F,$C5,$FF,$08
	.dc.b	$25,$7F,$08,$6D,$7F,$1B,$6B,$FC,$5D,$AF,$F1,$26,$BF,$C9,$5F,$E2
	.dc.b	$7F,$15,$27,$F8,$9B,$67,$F9,$33,$5F,$E4,$4D,$7F,$8E,$B5,$FE,$60
	.dc.b	$DA,$FE,$77,$F1,$5F,$C8,$09,$5F,$C8,$1B,$5F,$CC,$DA,$FF,$2F,$6B
	.dc.b	$FC,$A9,$AF,$F3,$D7,$F9,$5F,$C5,$49,$FE,$56,$D9,$FE,$7C,$D7,$F9
	.dc.b	$D3,$5F,$E6,$AD,$7F,$A4,$36,$BF,$A9,$FC,$57,$F3,$82,$57,$F3,$86
	.dc.b	$D7,$F4,$B6,$BF,$D1,$DA,$FF,$42,$6B,$FD,$55,$FE,$87,$F0,$F2,$7F
	.dc.b	$A1,$B6,$7F,$AB,$35,$FE,$A4,$D7,$FA,$6B,$5F,$EC,$0D,$AF,$ED,$7F
	.dc.b	$0F,$FD,$40,$95,$FD,$41,$B5,$FD,$8D,$AF,$F5,$F6,$BF,$D6,$9A,$FF
	.dc.b	$6D,$7F,$AD,$FC,$3C,$9F,$EB,$6D,$9F,$ED,$CD,$7F,$B5,$35,$FE,$CA
	.dc.b	$D7,$FB,$C3,$6B,$FC,$07,$F0,$FF,$DA,$09,$5F,$DA,$1B,$5F,$DE,$DA
	.dc.b	$FF,$77,$6B,$FD,$C9,$AF,$F8,$15,$FE,$E7,$F0,$F2,$7F,$B9,$B6,$7F
	.dc.b	$C0,$CD,$7F,$C0,$4D,$7F,$BE,$B5,$FF,$08,$36,$BF,$C3,$7F,$0D,$FE
	.dc.b	$00,$25,$7F,$80,$1B,$5F,$E1,$36,$BF,$E0,$F6,$BF,$E0,$A6,$BF,$E1
	.dc.b	$D7,$FC,$1F,$FA,$5B,$E1,$E9,$2B,$FC,$41,$19,$6C,$66,$BF,$C5,$93
	.dc.b	$96,$C6,$5F,$F1,$6C,$65,$B1,$97,$FC,$4D,$0F,$5F,$C6,$4A,$F0,$B6
	.dc.b	$7B,$B5,$EA,$D7,$9B,$56,$0A,$ED,$5E,$CF,$5B,$E0,$91,$C1,$2B,$DB
	.dc.b	$19,$6C,$65,$F6,$C6,$5B,$1F,$E9,$72,$3F,$E9,$76,$BF,$E9,$76,$BF
	.dc.b	$E9,$76,$BF,$E9,$76,$D7,$FA,$91,$BE,$F3,$C6,$FF,$48,$36,$3F,$D3
	.dc.b	$64,$AF,$F4,$DB,$5F,$F4,$DB,$5F,$F4,$DB,$5F,$F4,$DB,$FE,$97,$23
	.dc.b	$FE,$97,$6B,$FE,$97,$6B,$FE,$97,$6B,$FE,$97,$6D,$7F,$A9,$1B,$E2
	.dc.b	$7F,$98,$A7,$FC,$BB,$1F,$E5,$C2,$1F,$E5,$C2,$BF,$E6,$D7,$04,$70
	.dc.b	$43,$FC,$D8,$FF,$96,$DF,$F3,$87,$89,$FE,$60,$6C,$65,$FF,$36,$C6
	.dc.b	$5B,$19,$7F,$D0,$51,$96,$C7,$F9,$79,$FF,$2E,$35,$FF,$3B,$F8,$9F
	.dc.b	$E5,$E6,$C6,$5F,$F4,$64,$65,$B1,$9B,$FF,$4C,$CF,$F8,$E6,$7F,$C7
	.dc.b	$B3,$FE,$3F,$F0,$7F,$C8,$31,$FE,$96,$6B,$FE,$46,$3F,$C7,$E6,$7C
	.dc.b	$1E,$51,$FE,$F6,$9F,$F2,$14,$FF,$BC,$8F,$F2,$0C,$65,$E9,$E0,$FF
	.dc.b	$8F,$3C,$1F,$F1,$F9,$3C,$1C,$1F,$E3,$DD,$1D,$DF,$E0,$F6,$E3,$FE
	.dc.b	$1A,$26,$1F,$E1,$A6,$BF,$E1,$A6,$BF,$E1,$A6,$BF,$E1,$A7,$FC,$2C
	.dc.b	$47,$FC,$2C,$D7,$FC,$2C,$D7,$FC,$2C,$D7,$FC,$2C,$DD,$3F,$C4,$6D
	.dc.b	$97,$6F,$F8,$3D,$B1,$FE,$1A,$25,$7F,$86,$9A,$FF,$86,$9A,$FF,$86
	.dc.b	$9A,$FF,$86,$9F,$EE,$64,$7F,$B9,$B5,$FE,$E6,$D7,$FB,$9B,$5F,$EE
	.dc.b	$6D,$AF,$F6,$EF,$57,$F7,$5A,$75,$7F,$6C,$9F,$EE,$F5,$FD,$D4,$97
	.dc.b	$FD,$D5,$AF,$F7,$56,$BF,$DD,$5A,$FF,$75,$7F,$B9,$91,$FE,$E6,$D7
	.dc.b	$FB,$9B,$5F,$EE,$6D,$7F,$B9,$B6,$3F,$DE,$BD,$5F,$DD,$79,$5E,$BF
	.dc.b	$F5,$22,$57,$F5,$06,$DF,$F4,$E6,$BF,$D3,$1A,$FF,$4A,$6B,$FD,$02
	.dc.b	$7A,$CF,$F5,$AF,$87,$FE,$BE,$47,$04,$9F,$EB,$71,$96,$C6,$5F,$EB
	.dc.b	$71,$96,$C7,$F6,$6F,$87,$FE,$D6,$47,$04,$8F,$EC,$F1,$96,$C6,$5F
	.dc.b	$EC,$F1,$96,$C7,$F7,$2F,$87,$FE,$EE,$47,$04,$8F,$EE,$71,$96,$C6
	.dc.b	$5F,$EE,$71,$96,$C7,$F7,$EF,$6B,$FC,$06,$47,$04,$8F,$EF,$F1,$96
	.dc.b	$C6,$5F,$EF,$F1,$96,$C7,$F8,$2B,$DA,$FF,$07,$91,$C1,$23,$FC,$17
	.dc.b	$19,$6C,$65,$FF,$05,$C6,$5B,$1F,$C7,$1F,$F6,$D1,$FC,$74,$FF,$6F
	.dc.b	$B9,$E5,$FE,$3D,$1F,$C8,$39,$7F,$DC,$4E,$8F,$E4,$33,$B8,$CC,$73
	.dc.b	$21,$D8,$4B,$D2,$57,$F3,$76,$CF,$F3,$76,$A3,$FD,$F8,$E1,$B1,$C4
	.dc.b	$8F,$16,$BC,$4E,$53,$E1,$8C,$C7,$23,$FA,$37,$A5,$D5,$DB,$5F,$4B
	.dc.b	$64,$D2,$4F,$F5,$16,$CF,$48,$C1,$1C,$1F,$D5,$7D,$2E,$84,$8F,$EB
	.dc.b	$4D,$8D,$46,$08,$E0,$87,$BC,$F9,$F1,$C5,$FF,$26,$FB,$5F,$E5,$A2
	.dc.b	$4F,$F9,$61,$B1,$FE,$56,$6B,$C2,$34,$47,$4D,$7F,$CB,$3F,$0F,$FE
	.dc.b	$66,$24,$7F,$99,$1B,$1F,$E6,$26,$BC,$23,$44,$74,$D7,$FC,$CB,$F0
	.dc.b	$FF,$E7,$22,$47,$F9,$C1,$B1,$FE,$6E,$6B,$C2,$34,$47,$4D,$7F,$CE
	.dc.b	$3F,$0F,$FE,$7E,$24,$7F,$9F,$1B,$1F,$E7,$A6,$BC,$23,$44,$74,$D7
	.dc.b	$FC,$FB,$D7,$FE,$8A,$24,$7F,$A2,$1B,$1F,$E8,$66,$BC,$23,$44,$74
	.dc.b	$D7,$FD,$13,$D1,$FE,$96,$24,$7F,$A5,$1B,$1F,$E9,$26,$BC,$23,$44
	.dc.b	$74,$D7,$FD,$2B,$F0,$FF,$EA,$22,$47,$FA,$81,$B1,$FE,$9E,$6B,$C2
	.dc.b	$34,$47,$4D,$7F,$D4,$3E,$F7,$FA,$B8,$91,$FE,$AC,$6C,$7F,$AA,$9A
	.dc.b	$F0,$8D,$11,$D3,$5F,$F5,$6F,$BD,$FE,$BA,$24,$7F,$AE,$1B,$1F,$EB
	.dc.b	$66,$BC,$23,$44,$74,$D7,$FD,$73,$EF,$7F,$B1,$89,$1F,$EC,$46,$C7
	.dc.b	$FB,$09,$AF,$08,$D1,$1D,$35,$FF,$62,$FC,$3F,$FB,$48,$91,$FE,$D0
	.dc.b	$6C,$7F,$B3,$9A,$F0,$8D,$11,$D3,$5F,$F6,$8F,$C3,$FF,$B7,$89,$1F
	.dc.b	$ED,$C6,$C7,$FB,$69,$AF,$08,$D1,$1D,$35,$FE,$65,$DA,$FD,$33,$F9
	.dc.b	$E4,$FF,$17,$9F,$F6,$8C,$61,$FE,$35,$3B,$7F,$9A,$36,$78,$36,$37
	.dc.b	$3C,$5E,$ED,$7F,$A5,$46,$DC,$9D,$23,$4F,$88,$C1,$9F,$81,$95,$E0
	.dc.b	$94,$62,$7C,$2D,$7C,$2D,$7C,$31,$C2,$79,$4E,$67,$E0,$BE,$3F,$81
	.dc.b	$93,$DD,$B3,$EE,$6B,$EE,$6B,$D9,$AF,$CC,$6D,$FD,$1F,$81,$F8,$04
	.dc.b	$AF,$63,$6B,$E5,$B5,$F9,$6D,$7D,$6D,$7E,$CB,$FD,$97,$E0,$24,$FF
	.dc.b	$AC,$DB,$3F,$EB,$36,$BF,$D5,$FA,$11,$88,$FE,$F1,$18,$9E,$44,$8E
	.dc.b	$4F,$FA,$FF,$E0,$3F,$BA,$11,$C1,$23,$FD,$85,$19,$6C,$65,$F1,$46
	.dc.b	$5B,$1F,$7F,$E2,$3F,$0C,$47,$04,$8F,$C8,$C6,$5B,$19,$7F,$27,$19
	.dc.b	$6C,$7E,$2F,$E2,$3F,$24,$47,$04,$8F,$CE,$C6,$5B,$19,$7F,$3F,$19
	.dc.b	$6C,$7E,$5F,$E2,$3F,$3C,$47,$04,$8F,$D4,$C6,$5B,$19,$7F,$57,$19
	.dc.b	$6C,$7E,$8F,$E2,$3F,$54,$47,$04,$8F,$DA,$C6,$5B,$19,$7F,$6F,$19
	.dc.b	$6C,$7E,$BF,$E2,$3F,$6C,$47,$04,$8F,$E0,$31,$96,$C6,$5F,$E0,$71
	.dc.b	$96,$C7,$EE,$FE,$23,$F8,$11,$1C,$12,$3F,$86,$C6,$5B,$19,$7F,$87
	.dc.b	$C6,$5B,$1F,$C1,$FE,$1F,$F8,$71,$1C,$12,$3F,$8C,$C6,$5B,$19,$7F
	.dc.b	$8D,$C6,$5B,$1F,$CA,$7D,$77,$73,$C5,$FF,$2D,$46,$1B,$19,$9E,$0D
	.dc.b	$AF,$F2,$45,$6E,$73,$5C,$08,$E0,$93,$C2,$B9,$3F,$E6,$FA,$E4,$FF
	.dc.b	$94,$FE,$29,$93,$D2,$FF,$37,$46,$1B,$19,$C3,$FA,$24,$E5,$B1,$9A
	.dc.b	$E8,$D7,$FC,$FD,$3B,$73,$5D,$09,$5C,$9B,$1F,$03,$F3,$27,$57,$C8
	.dc.b	$6D,$7C,$D6,$CE,$E7,$44,$9D,$36,$35,$18,$23,$83,$FA,$BF,$8D,$FD
	.dc.b	$44,$91,$FD,$35,$B1,$A8,$C1,$1C,$10,$FE,$BF,$1C,$1F,$EC,$3E,$37
	.dc.b	$F5,$D2,$4F,$F5,$96,$C7,$F6,$86,$BC,$23,$44,$74,$D7,$FB,$6F,$8D
	.dc.b	$FD,$A4,$91,$FD,$95,$B1,$FD,$D1,$AF,$08,$D1,$1D,$35,$FE,$F3,$CF
	.dc.b	$FD,$D4,$91,$FD,$C5,$B1,$FE,$00,$6B,$C2,$34,$47,$4D,$7F,$C0,$BC
	.dc.b	$DF,$E0,$12,$47,$F7,$D6,$C7,$F8,$31,$AF,$08,$D1,$1D,$35,$FF,$08
	.dc.b	$FC,$4F,$F8,$34,$91,$FE,$09,$6C,$7F,$86,$1A,$F0,$8D,$11,$D3,$5F
	.dc.b	$F0,$EF,$B1,$FE,$19,$24,$7F,$85,$5B,$1F,$E2,$46,$BC,$23,$44,$74
	.dc.b	$D7,$FB,$A4,$FF,$B7,$C9,$1F,$E3,$78,$FF,$7D,$C6,$1F,$F1,$D7,$4A
	.dc.b	$31,$1C,$A7,$84,$E5,$B3,$A9,$FE,$FE,$FF,$BB,$6A,$4F,$06,$CF,$66
	.dc.b	$BF,$E0,$96,$BF,$E0,$86,$BF,$F0,$01,$BD,$7F,$E5,$1D,$FF,$C8,$04
	.dc.b	$AF,$F1,$E3,$6B,$FC,$96,$D7,$FC,$96,$D7,$FC,$76,$D7,$FC,$B2,$FF
	.dc.b	$92,$FC,$99,$3F,$E4,$76,$CF,$F9,$59,$AF,$F9,$59,$AF,$F9,$19,$AF
	.dc.b	$F9,$91,$B5,$FE,$69,$F8,$9F,$F2,$C1,$2B,$FC,$A8,$DA,$FF,$31,$B5
	.dc.b	$FF,$31,$B5,$FF,$29,$B5,$FF,$38,$BF,$E6,$3F,$89,$93,$FE,$5F,$6C
	.dc.b	$FF,$9B,$9A,$FF,$9B,$9A,$FF,$97,$9A,$FF,$9F,$1B,$4F,$D3,$55,$FE
	.dc.b	$7C,$6B,$FE,$7C,$6D,$7F,$A0,$9A,$FF,$A2,$9A,$FF,$9F,$1A,$A7,$B4
	.dc.b	$F5,$21,$D4,$B1,$D5,$F1,$7B,$51,$F1,$36,$3C,$4D,$7C,$4D,$7C,$4D
	.dc.b	$7C,$54,$8C,$57,$C1,$F6,$BE,$01,$2F,$E0,$36,$FE,$03,$5F,$80,$D7
	.dc.b	$E0,$35,$EE,$FC,$BF,$8A,$93,$F2,$DB,$3F,$2D,$AF,$CB,$6B,$F2,$DA
	.dc.b	$FA,$9B,$5F,$67,$E2,$BE,$C1,$2B,$EC,$36,$BE,$C3,$5F,$B0,$D7,$EC
	.dc.b	$35,$F8,$EF,$E0,$FE,$2A,$4F,$E0,$DB,$3F,$83,$6B,$F8,$36,$BF,$83
	.dc.b	$6B,$F4,$9B,$5F,$91,$F8,$AF,$C8,$12,$BF,$20,$DA,$FC,$83,$5F,$C8
	.dc.b	$35,$FC,$83,$5F,$BE,$FE,$6F,$E2,$A4,$FE,$6D,$B3,$F9,$B6,$BF,$9B
	.dc.b	$6B,$F9,$B6,$BF,$8A,$6D,$7E,$A7,$E2,$BF,$50,$4A,$FD,$43,$6B,$F5
	.dc.b	$0D,$7F,$50,$D7,$F5,$0D,$7F,$2E,$FE,$CF,$DF,$93,$FB,$36,$CF,$EC
	.dc.b	$DA,$FE,$CD,$AF,$EC,$DA,$FE,$89,$B5,$FC,$07,$E2,$BF,$80,$12,$BF
	.dc.b	$80,$36,$BF,$80,$35,$FE,$00,$D7,$F8,$03,$5F,$D7,$BF,$C2,$FE,$02
	.dc.b	$4F,$F0,$B6,$CF,$F0,$B6,$BF,$C2,$DA,$FF,$0B,$6B,$FB,$A6,$AE,$7F
	.dc.b	$F2,$83,$FE,$D6,$E5,$FE,$36,$FF,$62,$6B,$FE,$DD,$6B,$BA,$FE,$C7
	.dc.b	$18,$81,$D5,$C3,$63,$64,$7B,$35,$FF,$75,$72,$F0,$A3,$49,$3F,$DC
	.dc.b	$9B,$3F,$DE,$1A,$F2,$8D,$11,$D3,$FE,$01,$F8,$17,$AB,$E6,$6C,$66
	.dc.b	$30,$47,$04,$9F,$79,$27,$8B,$83,$E1,$B8,$70,$D7,$84,$F5,$7F,$A3
	.dc.b	$7C,$57,$F4,$B2,$57,$F4,$B6,$C6,$A3,$04,$70,$43,$FA,$74,$70,$7F
	.dc.b	$A8,$7C,$57,$F5,$72,$4F,$F5,$76,$C7,$F5,$76,$BC,$23,$44,$74,$D7
	.dc.b	$FA,$E7,$C5,$7F,$63,$24,$7F,$63,$6C,$7F,$63,$6B,$C2,$34,$47,$4D
	.dc.b	$7F,$B4,$7C,$27,$F6,$F2,$47,$F6,$F6,$C7,$F6,$F6,$BC,$23,$44,$74
	.dc.b	$D7,$FB,$A7,$C2,$7F,$7B,$24,$7F,$7B,$6C,$7F,$7B,$6B,$C2,$34,$47
	.dc.b	$4D,$7F,$C0,$1F,$09,$FE,$07,$24,$7F,$81,$DB,$1F,$E0,$76,$BC,$23
	.dc.b	$44,$74,$D7,$FC,$19,$F1,$5F,$E1,$32,$47,$F8,$4D,$B1,$FE,$13,$6B
	.dc.b	$C2,$34,$47,$4D,$7F,$33,$EE,$7E,$9C,$91,$FA,$B6,$C7,$E8,$46,$A8
	.dc.b	$CC,$72,$9E,$04,$78,$4E,$E3,$11,$FA,$8F,$73,$F6,$C4,$8F,$DD,$36
	.dc.b	$38,$B5,$E1,$1A,$23,$A6,$BF,$E3,$3F,$8A,$FF,$1D,$12,$3F,$C7,$4D
	.dc.b	$8F,$F1,$D3,$5E,$11,$A2,$3A,$6B,$FE,$3F,$F8,$AF,$F2,$31,$23,$FC
	.dc.b	$8C,$D8,$FF,$23,$35,$E1,$1A,$23,$A6,$BF,$C4,$B9,$9B,$3C,$EF,$E6
	.dc.b	$91,$86,$C6,$6B,$EB,$4E,$5B,$19,$AF,$E3,$53,$96,$C6,$5F,$03,$5F
	.dc.b	$E8,$31,$B7,$35,$E0,$25,$76,$F8,$AF,$19,$1C,$12,$7B,$C6,$5B,$19
	.dc.b	$7C,$31,$96,$C7,$AB,$E2,$BD,$E4,$70,$48,$F5,$C6,$5B,$19,$7D,$B1
	.dc.b	$96,$C7,$C6,$F8,$9F,$96,$47,$04,$8F,$8F,$19,$6C,$65,$F9,$31,$96
	.dc.b	$C7,$D2,$F8,$8F,$AE,$47,$04,$8F,$A7,$19,$6C,$65,$FA,$B1,$96,$C7
	.dc.b	$F6,$1F,$82,$FE,$C0,$47,$04,$8F,$D9,$C6,$5B,$19,$7F,$AD,$C6,$5B
	.dc.b	$1F,$DB,$7E,$0B,$FB,$61,$1C,$12,$3F,$7F,$19,$6C,$65,$FE,$CF,$19
	.dc.b	$6C,$7E,$5B,$E2,$3F,$36,$47,$04,$8F,$CB,$C6,$5B,$19,$7F,$33,$19
	.dc.b	$6C,$7E,$8B,$E2,$3F,$4E,$47,$04,$8F,$D1,$C6,$5B,$19,$7F,$4B,$19
	.dc.b	$6C,$7E,$BB,$D8,$FD,$99,$1C,$12,$3F,$5F,$19,$6C,$65,$FD,$8C,$65
	.dc.b	$B1,$FB,$AF,$88,$FD,$F9,$1C,$12,$3F,$77,$19,$6C,$65,$FD,$EC,$65
	.dc.b	$B1,$FC,$1B,$E2,$3F,$85,$91,$C1,$23,$F8,$3C,$65,$B1,$97,$F8,$4C
	.dc.b	$65,$B1,$FC,$4B,$DA,$FE,$2E,$47,$04,$8F,$E2,$71,$96,$C6,$5F,$E2
	.dc.b	$B1,$96,$C7,$F1,$EF,$88,$FE,$46,$47,$04,$8F,$E3,$F1,$96,$C6,$5F
	.dc.b	$E4,$31,$96,$C7,$F9,$37,$E1,$FF,$C9,$84,$70,$48,$FE,$BF,$19,$6C
	.dc.b	$65,$FF,$23,$C6,$5B,$1F,$E5,$9F,$67,$FC,$B0,$47,$04,$8F,$ED,$71
	.dc.b	$96,$C6,$5F,$F2,$9C,$65,$B1,$FE,$65,$F8,$7F,$F3,$21,$1C,$12,$3F
	.dc.b	$BB,$C6,$5B,$19,$7F,$CB,$F1,$96,$C7,$F4,$EF,$88,$FE,$A6,$47,$04
	.dc.b	$8F,$E9,$F1,$96,$C6,$5F,$EA,$31,$96,$C7,$F5,$AF,$0F,$FA,$F9,$1C
	.dc.b	$12,$3F,$AD,$C6,$5B,$19,$7F,$AE,$C6,$5B,$1F,$D9,$BB,$7F,$B5,$91
	.dc.b	$C1,$23,$FB,$3C,$65,$B1,$97,$FB,$4C,$65,$B1,$FD,$CB,$B3,$FB,$B9
	.dc.b	$1C,$12,$3F,$B9,$C6,$5B,$19,$7F,$BA,$C6,$5B,$1F,$DF,$B8,$FF,$C0
	.dc.b	$64,$70,$48,$FE,$FF,$19,$6C,$65,$FF,$00,$C6,$5B,$1F,$E0,$AE,$3F
	.dc.b	$F0,$79,$1C,$12,$3F,$C1,$71,$96,$C6,$5F,$F0,$6C,$65,$B1,$FE,$16
	.dc.b	$D3,$FC,$36,$47,$04,$8F,$F0,$BC,$65,$B1,$97,$FC,$33,$19,$6C,$7F
	.dc.b	$96,$5F,$F7,$31,$FE,$D3,$23,$9F,$17,$FD,$BE,$D8,$C9,$FE,$61,$8C
	.dc.b	$B4,$C9,$FE,$C5,$72,$D3,$27,$FB,$31,$7C,$3F,$FB,$8C,$93,$E1,$6C
	.dc.b	$F7,$6B,$D5,$AF,$36,$BF,$EF,$69,$EB,$3F,$E4,$0D,$BF,$C8,$A4,$70
	.dc.b	$49,$FF,$20,$C6,$5B,$19,$7F,$C8,$71,$96,$C7,$F9,$33,$6F,$F2,$89
	.dc.b	$1C,$12,$3F,$C9,$B1,$96,$C6,$5F,$F2,$7C,$65,$B1,$FE,$5A,$FD,$3B
	.dc.b	$E5,$C7,$F9,$49,$AB,$B1,$18,$6C,$7F,$99,$A3,$04,$70,$49,$EC,$D3
	.dc.b	$FC,$CB,$1C,$31,$E4,$49,$F2,$36,$76,$D7,$FC,$F1,$3F,$49,$CC,$F0
	.dc.b	$23,$83,$BF,$C4,$79,$C9,$1F,$05,$B1,$A8,$C1,$1C,$10,$F0,$C7,$07
	.dc.b	$C7,$F1,$1F,$0C,$93,$F3,$5B,$1F,$01,$AF,$08,$D1,$1D,$35,$F7,$FC
	.dc.b	$47,$CF,$24,$7D,$96,$C7,$CC,$6B,$C2,$34,$47,$4D,$7F,$59,$48,$FD
	.dc.b	$B3,$63,$F6,$8D,$77,$1A,$23,$A6,$BA,$30,$43,$07,$EE,$EB,$A3,$FB
	.dc.b	$96,$BF,$AF,$6A,$D1,$18,$6C,$FF,$0D,$8C,$11,$C1,$27,$F8,$74,$65
	.dc.b	$FC,$9F,$C4,$7E,$6C,$93,$FA,$56,$C6,$A3,$04,$70,$43,$F2,$B1,$C1
	.dc.b	$FC,$BF,$C4,$7E,$9C,$93,$FB,$16,$C7,$E9,$1A,$F0,$8D,$11,$D3,$5F
	.dc.b	$D1,$FC,$47,$EC,$C9,$1F,$BD,$6C,$7E,$C1,$AF,$08,$D1,$1D,$35,$FD
	.dc.b	$7F,$C3,$7E,$FC,$91,$FC,$25,$B1,$FB,$C6,$BC,$23,$44,$74,$D7,$F7
	.dc.b	$7F,$11,$FC,$2C,$91,$FC,$55,$B1,$FC,$21,$AF,$08,$D1,$1D,$35,$FE
	.dc.b	$0F,$F1,$1F,$C5,$C9,$1F,$C8,$5B,$1F,$C5,$1A,$F0,$8D,$11,$D3,$5F
	.dc.b	$E2,$7F,$0D,$FC,$8C,$91,$FC,$B5,$B1,$FC,$81,$AF,$08,$D1,$1D,$35
	.dc.b	$FE,$C1,$B7,$E7,$C9,$1F,$DA,$DB,$1F,$DA,$5A,$F0,$8D,$11,$D3,$5F
	.dc.b	$ED,$99,$FE,$AC,$91,$FD,$DD,$B1,$FD,$D5,$AF,$08,$D1,$1D,$35,$FE
	.dc.b	$F1,$9F,$ED,$C9,$1F,$E0,$36,$C7,$F8,$05,$AF,$08,$D1,$1D,$35,$FE
	.dc.b	$87,$F0,$DF,$D4,$C9,$1F,$D7,$5B,$1F,$D4,$1A,$F0,$8D,$11,$D3,$5F
	.dc.b	$E9,$FF,$0D,$FD,$7C,$91,$FD,$A5,$B1,$FD,$71,$AF,$08,$D1,$1D,$35
	.dc.b	$FE,$B7,$F0,$DF,$DA,$C9,$1F,$DD,$5B,$1F,$DA,$1A,$F0,$8D,$11,$D3
	.dc.b	$5F,$EC,$FF,$0D,$FD,$DC,$91,$FE,$01,$6C,$7F,$74,$6B,$C2,$34,$47
	.dc.b	$4D,$7F,$B9,$FC,$37,$F8,$0C,$91,$FE,$0D,$6C,$7F,$80,$1A,$F0,$8D
	.dc.b	$11,$D3,$5F,$EF,$FF,$0D,$FE,$0F,$24,$7F,$86,$5B,$1F,$E0,$C6,$BC
	.dc.b	$23,$44,$74,$D7,$FC,$17,$EC,$7F,$86,$C9,$1F,$E2,$56,$C7,$F8,$61
	.dc.b	$AF,$08,$D1,$1D,$35,$FF,$2C,$71,$7F,$57,$24,$7F,$98,$DB,$1F,$E6
	.dc.b	$16,$BC,$23,$44,$74,$D7,$FC,$C9,$C5,$FD,$8C,$91,$FE,$6F,$6C,$7F
	.dc.b	$9B,$5A,$F0,$8D,$11,$D3,$5F,$F1,$7F,$B1,$FE,$3F,$24,$7F,$92,$5B
	.dc.b	$1F,$E3,$C6,$BC,$23,$44,$74,$D7,$FC,$77,$F0,$DF,$E4,$B2,$47,$F9
	.dc.b	$55,$B1,$FE,$48,$6B,$C2,$34,$47,$4D,$44,$F3,$DF,$92,$3D,$FF,$F0
	.dc.b	$7C,$61,$B1,$9A,$E3,$52,$B5,$18,$8E,$93,$B7,$13,$A6,$C6,$5E,$4D
	.dc.b	$78,$D7,$80,$95,$D6,$15,$CD,$F1,$FC,$4B,$36,$DF,$E2,$88,$C3,$63
	.dc.b	$37,$CE,$72,$D8,$CD,$7A,$E7,$2D,$8C,$BE,$06,$BF,$32,$36,$E6,$BC
	.dc.b	$04,$AF,$55,$FD,$0F,$80,$F9,$A4,$8C,$E4,$8C,$5F,$C8,$6B,$F2,$1A
	.dc.b	$FD,$06,$BF,$1A,$39,$4F,$FA,$77,$E2,$3F,$D2,$C4,$70,$49,$FF,$49
	.dc.b	$C6,$5B,$19,$7F,$D1,$F1,$96,$C7,$D5,$F0,$1F,$A7,$FF,$9A,$63,$0D
	.dc.b	$9F,$C3,$4E,$A7,$FC,$BA,$D7,$66,$27,$84,$FE,$42,$74,$F2,$8C,$BF
	.dc.b	$9C,$8C,$BB,$8C,$BF,$EB,$96,$C6,$A3,$FD,$CD,$D9,$FE,$69,$25,$7F
	.dc.b	$BD,$5B,$1F,$EF,$36,$BF,$EF,$16,$BC,$A3,$81,$1F,$F7,$D7,$67,$F9
	.dc.b	$D4,$91,$FF,$00,$5B,$1A,$8C,$11,$C1,$0F,$F8,$01,$1C,$1F,$77,$C4
	.dc.b	$7C,$72,$4F,$D0,$6C,$7C,$76,$BC,$23,$44,$74,$D7,$E5,$7C,$47,$D3
	.dc.b	$24,$7D,$C6,$C7,$D3,$6B,$C2,$34,$47,$4D,$7E,$B7,$C4,$7D,$F2,$47
	.dc.b	$E2,$1B,$1F,$7D,$AF,$08,$D1,$1D,$35,$FC,$17,$C4,$7E,$2C,$91,$F9
	.dc.b	$46,$C7,$E2,$DA,$F0,$8D,$11,$D3,$5F,$C7,$7C,$47,$E5,$C9,$1F,$A0
	.dc.b	$6C,$7E,$5D,$AF,$08,$D1,$1D,$35,$FC,$D7,$C4,$7E,$8C,$91,$FA,$C6
	.dc.b	$C7,$E8,$DA,$F0,$8D,$11,$D3,$5F,$D3,$72,$FE,$BC,$91,$FB,$86,$C7
	.dc.b	$EB,$DA,$F0,$8D,$11,$D3,$5F,$E8,$9E,$23,$C3,$FF,$6B,$12,$7F,$9E
	.dc.b	$4F,$FB,$6E,$30,$FF,$3F,$9E,$13,$A6,$C6,$67,$6F,$66,$BF,$D5,$A3
	.dc.b	$6F,$8B,$C0,$FF,$73,$12,$BA,$36,$3C,$0D,$78,$46,$88,$E9,$AF,$BB
	.dc.b	$C0,$FF,$79,$12,$3D,$0D,$8F,$63,$5E,$11,$A2,$3A,$6B,$FC,$87,$E3
	.dc.b	$1E,$97,$F2,$86,$CF,$F3,$16,$CF,$F2,$E6,$BC,$67,$81,$27,$83,$63
	.dc.b	$83,$5E,$11,$D7,$D2,$F0,$11,$C1,$23,$C1,$19,$6C,$65,$F5,$46,$5B
	.dc.b	$1E,$9F,$3F,$D8,$47,$04,$8F,$64,$65,$B1,$97,$E3,$46,$5B,$1F,$17
	.dc.b	$CF,$F9,$04,$70,$48,$F9,$11,$96,$C6,$5F,$A5,$19,$6C,$7D,$1F,$37
	.dc.b	$EA,$11,$C1,$23,$EA,$46,$5B,$19,$7E,$F4,$65,$B1,$FD,$B7,$E3,$3F
	.dc.b	$C7,$2D,$8F,$ED,$31,$87,$FB,$6B,$5F,$ED,$0D,$7F,$B3,$35,$74,$23
	.dc.b	$0D,$8E,$A4,$9F,$01,$1E,$FE,$77,$88,$91,$E7,$6C,$6A,$30,$47,$04
	.dc.b	$3C,$F1,$C1,$FE,$13,$E4,$27,$F8,$13,$63,$F8,$33,$5D,$C6,$88,$E9
	.dc.b	$AE,$8C,$10,$C1,$FC,$63,$C8,$E8,$FF,$0B,$6B,$FE,$6A,$E1,$46,$21
	.dc.b	$18,$9F,$E3,$13,$98,$C4,$FF,$1D,$8C,$12,$76,$48,$D3,$F8,$1F,$0F
	.dc.b	$EF,$91,$C1,$23,$F0,$31,$96,$C6,$5F,$C1,$C6,$5B,$1F,$8D,$F1,$BF
	.dc.b	$16,$47,$04,$8F,$C6,$C6,$5B,$19,$7F,$1F,$19,$6C,$7F,$98,$3D,$47
	.dc.b	$AF,$FE,$8E,$6C,$66,$30,$47,$04,$9F,$F4,$5C,$F0,$6C,$ED,$B3,$B7
	.dc.b	$FD,$0B,$4F,$27,$FD,$08,$D9,$FF,$41,$36,$7F,$D0,$0D,$78,$CF,$02
	.dc.b	$4F,$06,$C7,$06,$BC,$23,$FC,$FF,$F1,$8F,$4B,$FD,$5A,$D8,$CC,$60
	.dc.b	$8E,$09,$3F,$EA,$A9,$E0,$D9,$DB,$67,$6F,$F0,$0E,$7F,$E0,$A4,$8F
	.dc.b	$E1,$4D,$8F,$E0,$ED,$78,$46,$88,$E9,$AF,$F0,$CD,$BF,$88,$92,$3F
	.dc.b	$8B,$36,$3F,$89,$B5,$E1,$1A,$23,$A6,$BF,$E0,$8F,$23,$F5,$04,$8F
	.dc.b	$F0,$13,$63,$FC,$0C,$D7,$84,$68,$8E,$9A,$FF,$85,$3C,$CF,$DA,$12
	.dc.b	$3F,$C1,$CD,$8F,$F0,$93,$5E,$11,$A2,$3A,$6B,$FC,$C3,$CB,$FE,$6A
	.dc.b	$48,$FE,$74,$D8,$FE,$6E,$D7,$84,$68,$8E,$9A,$FF,$3C,$F1,$FF,$A0
	.dc.b	$92,$3F,$A3,$36,$3F,$A1,$B5,$E1,$1A,$23,$A6,$BF,$EA,$FF,$8C,$FF
	.dc.b	$53,$92,$3F,$D6,$ED,$7F,$D8,$0D,$8F,$F5,$EB,$5E,$51,$C0,$87,$0F
	.dc.b	$0F,$99,$1C,$12,$39,$C6,$5B,$19,$7C,$D1,$96,$C7,$93,$C3,$F3,$91
	.dc.b	$C1,$23,$CF,$19,$6C,$65,$F8,$51,$96,$C7,$C0,$F2,$BE,$19,$1C,$12
	.dc.b	$3E,$1C,$65,$B1,$97,$E7,$46,$5B,$1F,$33,$E3,$3E,$79,$1C,$12,$3E
	.dc.b	$7C,$65,$B1,$97,$ED,$46,$5B,$1F,$E6,$BE,$1F,$EA,$84,$70,$48,$FF
	.dc.b	$39,$46,$5B,$19,$7F,$CF,$B1,$96,$C7,$F9,$DF,$CE,$FE,$C0,$47,$04
	.dc.b	$8F,$F3,$F4,$65,$B1,$97,$FD,$13,$19,$6C,$7F,$A2,$3E,$11,$EC,$FF
	.dc.b	$71,$6C,$66,$30,$47,$04,$91,$0B,$3C,$9C,$18,$9E,$4E,$04,$6D,$DB
	.dc.b	$5E,$BD,$8D,$5F,$05,$FA,$A2,$61,$FA,$A6,$D7,$EA,$9A,$FE,$CD,$AF
	.dc.b	$49,$E7,$3C,$49,$1F,$B2,$F8,$8F,$DB,$92,$3F,$6E,$D8,$D4,$60,$8E
	.dc.b	$08,$7F,$03,$8E,$0F,$BB,$DB,$FA,$04,$9F,$A8,$D8,$F8,$D8,$A3,$11
	.dc.b	$C6,$76,$47,$71,$A7,$EA,$4B,$DB,$FB,$C4,$AF,$C1,$36,$7B,$B5,$FA
	.dc.b	$6D,$79,$B5,$FB,$EF,$F8,$F3,$2F,$AB,$E9,$DC,$CD,$25,$FE,$06,$76
	.dc.b	$D9,$DB,$67,$77,$F8,$AA,$DB,$6B,$6D,$9D,$BF,$93,$9F,$C3,$FC,$1F
	.dc.b	$8C,$97,$DD,$B3,$CD,$AF,$1B,$46,$23,$F2,$71,$89,$E4,$48,$F2,$7C
	.dc.b	$1F,$C1,$24,$7B,$5B,$1A,$8C,$11,$C1,$0F,$14,$70,$7F,$4B,$E3,$FE
	.dc.b	$DC,$93,$FB,$B6,$C7,$E2,$1A,$F0,$8D,$11,$D3,$5F,$D8,$F8,$FF,$C0
	.dc.b	$C9,$1F,$C1,$DB,$1F,$94,$6B,$C2,$34,$47,$4D,$7E,$67,$BF,$F8,$52
	.dc.b	$47,$E0,$5B,$1F,$75,$AF,$08,$D1,$1D,$35,$FB,$1E,$FF,$E4,$49,$1F
	.dc.b	$8D,$6C,$7E,$25,$AF,$08,$D1,$1D,$35,$FE,$2B,$EF,$FF,$27,$24,$7F
	.dc.b	$2B,$6C,$7E,$E1,$AF,$08,$D1,$1D,$35,$FE,$43,$E0,$7F,$33,$24,$7F
	.dc.b	$37,$6C,$7F,$04,$6B,$C2,$34,$47,$4D,$7F,$96,$F9,$5F,$CF,$C9,$1F
	.dc.b	$D0,$DB,$1F,$C4,$1A,$F0,$8D,$11,$D3,$5F,$D4,$71,$7F,$01,$24,$7E
	.dc.b	$F5,$B1,$FB,$96,$BC,$23,$44,$74,$D7,$F6,$9C,$3F,$C3,$49,$1F,$C2
	.dc.b	$5B,$1F,$C1,$5A,$F0,$8D,$11,$D3,$5F,$EA,$FE,$47,$FB,$F5,$FF,$69
	.dc.b	$9F,$ED,$12,$47,$F5,$97,$2D,$8E,$04,$8D,$9F,$D7,$DC,$46,$E3,$FD
	.dc.b	$B9,$E4,$C8,$E4,$D8,$E0,$D7,$CA,$D7,$FB,$84,$A3,$11,$FE,$ED,$6F
	.dc.b	$13,$DF,$FE,$EF,$7F,$CC,$5B,$5F,$CB,$5A,$FF,$29,$6B,$FD,$C9,$AA
	.dc.b	$7A,$3F,$D0,$63,$31,$89,$ED,$39,$23,$83,$C5,$5C,$48,$E0,$91,$E5
	.dc.b	$8C,$B6,$32,$F9,$E3,$2D,$8F,$77,$BF,$E5,$23,$82,$47,$C1,$8C,$B6
	.dc.b	$32,$FC,$38,$CB,$63,$F0,$10,$69,$23,$F0,$11,$96,$C6,$5F,$C7,$C6
	.dc.b	$5B,$19,$7E,$EC,$65,$B1,$97,$FC,$63,$83,$DF,$D4,$61,$AA,$35,$3F
	.dc.b	$97,$9C,$C6,$27,$F2,$D1,$82,$4E,$C9,$1A,$6B,$F8,$7F,$23,$F0,$44
	.dc.b	$8F,$C7,$36,$3F,$1C,$D7,$84,$68,$8E,$9A,$FE,$4F,$D9,$FC,$71,$23
	.dc.b	$F3,$4D,$8F,$CD,$35,$E1,$1A,$23,$A6,$BF,$C0,$3D,$FF,$86,$48,$FD
	.dc.b	$E3,$63,$F7,$4D,$78,$46,$88,$E9,$AF,$F0,$CF,$7F,$E7,$92,$3F,$84
	.dc.b	$36,$3F,$83,$35,$E1,$1A,$23,$A6,$BF,$C6,$3D,$FF,$B6,$48,$FE,$28
	.dc.b	$D8,$FE,$24,$D7,$84,$68,$8E,$9A,$FF,$03,$F2,$BF,$7C,$48,$FE,$14
	.dc.b	$D8,$FE,$14,$D7,$84,$68,$8E,$9A,$FF,$0F,$F4,$3F,$85,$12,$3F,$8B
	.dc.b	$36,$3F,$8B,$35,$E1,$1A,$23,$A6,$BF,$CF,$3D,$0F,$CF,$92,$3F,$9C
	.dc.b	$36,$3F,$9B,$35,$E1,$1A,$23,$A6,$BF,$D2,$33,$FD,$59,$23,$FA,$23
	.dc.b	$63,$FA,$13,$5E,$11,$A2,$3A,$6B,$FE,$AD,$EE,$7B,$CF,$63,$F7,$F1
	.dc.b	$86,$C6,$6B,$FD,$5D,$39,$6C,$66,$BF,$D6,$B3,$96,$C6,$5F,$03,$5F
	.dc.b	$F5,$FC,$6D,$CD,$78,$09,$5F,$EB,$DF,$7D,$8B,$CF,$FE,$2B,$18,$6C
	.dc.b	$66,$78,$36,$FF,$D9,$35,$B9,$CD,$70,$23,$82,$4F,$0A,$E4,$FF,$B5
	.dc.b	$6B,$93,$EE,$F2,$BE,$39,$2B,$E2,$B6,$35,$18,$23,$82,$1F,$1A,$38
	.dc.b	$3F,$E0,$AE,$1F,$E5,$A4,$9F,$F0,$33,$63,$FC,$08,$D7,$84,$68,$8E
	.dc.b	$9A,$FF,$85,$BD,$EF,$E7,$24,$8F,$F0,$93,$63,$FC,$20,$D7,$84,$68
	.dc.b	$8E,$9A,$FF,$AA,$7D,$F7,$9B,$FE,$BD,$6C,$FF,$AE,$5B,$3F,$EB,$56
	.dc.b	$BC,$67,$81,$27,$83,$63,$83,$5E,$11,$D7,$DD,$F0,$11,$C1,$23,$C1
	.dc.b	$19,$6C,$65,$F5,$46,$5B,$1F,$E3,$7F,$73,$FB,$09,$1C,$12,$3F,$C7
	.dc.b	$11,$96,$C6,$5F,$F2,$3C,$65,$B1,$FE,$43,$F7,$3F,$B6,$91,$C1,$23
	.dc.b	$FC,$89,$19,$6C,$65,$FF,$29,$C6,$5B,$1F,$E4,$FF,$57,$FB,$C9,$1C
	.dc.b	$12,$3F,$CA,$11,$96,$C6,$5F,$F2,$FC,$65,$B1,$FE,$F5,$F7,$1E,$47
	.dc.b	$FB,$F5,$B1,$98,$C1,$1C,$12,$7F,$DF,$13,$C1,$B3,$B6,$CE,$DF,$F7
	.dc.b	$C4,$BC,$8F,$01,$27,$C8,$D9,$EA,$D7,$8B,$51,$FE,$57,$9E,$93,$C8
	.dc.b	$91,$FE,$75,$E8,$FF,$11,$92,$3F,$CF,$8D,$8D,$46,$08,$E0,$87,$FA
	.dc.b	$2A,$38,$3F,$E8,$CE,$8F,$F1,$79,$27,$FD,$10,$D8,$FF,$43,$35,$E1
	.dc.b	$1A,$23,$A6,$BF,$D1,$7D,$97,$B3,$FD,$11,$B3,$FD,$09,$B3,$FD,$01
	.dc.b	$AF,$19,$E0,$49,$E0,$D8,$E0,$D7,$84,$7F,$3F,$EA,$7A,$7F,$CF,$9B
	.dc.b	$19,$8C,$11,$C1,$27,$F9,$FC,$F0,$6C,$ED,$B3,$B7,$FB,$27,$B2,$F2
	.dc.b	$7F,$B1,$B6,$7F,$B0,$B6,$7F,$AF,$B5,$E3,$3C,$09,$3C,$1B,$1C,$1A
	.dc.b	$F0,$8F,$F6,$07,$57,$F9,$A8,$8E,$09,$1F,$EC,$18,$CB,$63,$2F,$FB
	.dc.b	$42,$32,$D8,$FF,$66,$7B,$3F,$E7,$62,$38,$24,$7F,$B3,$63,$2D,$8C
	.dc.b	$BF,$ED,$C8,$CB,$63,$F4,$AC,$F1,$9E,$9B,$D7,$FF,$45,$C6,$1B,$19
	.dc.b	$AE,$35,$FA,$EA,$D4,$62,$3F,$D8,$D3,$B7,$73,$A6,$C6,$5F,$13,$5E
	.dc.b	$8D,$79,$D7,$88,$95,$DE,$BD,$3D,$1E,$62,$46,$6F,$FA,$AD,$F4,$9C
	.dc.b	$BE,$26,$BE,$26,$BB,$9E,$51,$FE,$A8,$D2,$47,$FC,$01,$6C,$FF,$BF
	.dc.b	$DA,$FC,$6F,$1B,$F3,$8D,$8C,$BF,$EF,$A6,$BF,$EB,$C7,$FD,$6B,$EB
	.dc.b	$4A,$F9,$93,$FE,$BB,$8C,$3F,$3E,$91,$88,$F0,$35,$FB,$F1,$A9,$E6
	.dc.b	$D8,$E5,$3B,$7F,$94,$FA,$8F,$5B,$FD,$96,$DB,$F1,$B6,$7F,$00,$D7
	.dc.b	$F0,$AD,$7F,$1E,$D7,$AC,$F4,$24,$F4,$6C,$74,$7C,$3E,$74,$9F,$0B
	.dc.b	$67,$C2,$D7,$C2,$D7,$C2,$D7,$DA,$D5,$4F,$3B,$DE,$4A,$F7,$B6,$7D
	.dc.b	$ED,$7D,$ED,$7D,$ED,$7E,$5B,$F2,$BA,$A4,$FC,$A6,$CF,$CA,$6B,$F2
	.dc.b	$9A,$FC,$A6,$BF,$59,$AF,$AB,$AB,$EB,$92,$7E,$BB,$63,$EB,$B5,$FA
	.dc.b	$ED,$7E,$BB,$5F,$C1,$BF,$82,$F4,$24,$FE,$09,$B3,$F8,$26,$BF,$82
	.dc.b	$6B,$F8,$26,$BF,$8E,$6B,$F4,$B9,$3F,$1E,$49,$FC,$7B,$63,$F1,$ED
	.dc.b	$7F,$1E,$D7,$F1,$ED,$7F,$36,$FE,$6B,$D0,$93,$F9,$A6,$CF,$E6,$9A
	.dc.b	$FE,$69,$AF,$E6,$9A,$FE,$99,$AB,$6F,$F0,$8C,$FE,$BA,$0D,$23,$FB
	.dc.b	$93,$25,$72,$FF,$09,$32,$57,$2F,$EF,$0C,$95,$FD,$EF,$9F,$88,$FF
	.dc.b	$11,$CA,$31,$1E,$76,$C6,$67,$89,$27,$F8,$74,$E8,$91,$98,$FE,$3D
	.dc.b	$18,$7F,$8A,$79,$FD,$09,$3F,$C4,$1B,$1A,$8C,$11,$C1,$0F,$E5,$51
	.dc.b	$C1,$FE,$59,$E7,$FA,$09,$3F,$C7,$1B,$1F,$C9,$5A,$F0,$8D,$11,$D3
	.dc.b	$5F,$E7,$1E,$7F,$C4,$24,$7F,$28,$6C,$7F,$31,$6B,$C2,$34,$47,$4D
	.dc.b	$7F,$A2,$79,$FF,$40,$91,$FC,$D1,$B1,$FC,$F5,$AF,$08,$D1,$1D,$35
	.dc.b	$FE,$A1,$E7,$FD,$C2,$47,$F4,$06,$C7,$F4,$96,$BC,$23,$44,$74,$D7
	.dc.b	$FA,$E7,$99,$F8,$82,$47,$F4,$C6,$C7,$F5,$56,$BC,$23,$44,$74,$D7
	.dc.b	$FB,$47,$77,$E5,$09,$1F,$D6,$1B,$1F,$D8,$5A,$F0,$8D,$11,$D3,$5F
	.dc.b	$F7,$79,$FE,$ED,$21,$82,$1F,$DA,$3C,$AF,$F7,$B3,$63,$FD,$E6,$D3
	.dc.b	$89,$4F,$F7,$79,$FE,$ED,$21,$82,$1F,$DD,$3C,$AF,$F7,$B3,$63,$FD
	.dc.b	$E6,$D3,$89,$4F,$F6,$79,$FE,$D4,$21,$82,$1F,$E0,$0F,$23,$FD,$AC
	.dc.b	$D8,$FF,$6D,$34,$E2,$53,$FD,$9E,$7F,$B5,$08,$60,$87,$F8,$33,$C8
	.dc.b	$FF,$6B,$36,$3F,$DB,$4D,$38,$94,$FF,$67,$9F,$ED,$42,$18,$21,$FE
	.dc.b	$18,$F2,$3F,$DA,$CD,$8F,$F6,$D3,$4E,$25,$3F,$D6,$4F,$FA,$D9,$A7
	.dc.b	$FA,$D8,$FF,$11,$F8,$FF,$EB,$66,$C6,$DA,$6C,$FE,$53,$1B,$7F,$95
	.dc.b	$F8,$FF,$EB,$77,$FC,$5A,$D7,$33,$FE,$B6,$72,$D7,$27,$FA,$EC,$87
	.dc.b	$FB,$00,$FF,$43,$10,$FF,$42,$14,$FF,$1D,$F8,$9F,$E8,$86,$C7,$FA
	.dc.b	$31,$A6,$CF,$E8,$31,$B7,$FA,$1F,$89,$FE,$84,$7F,$C8,$AD,$73,$3F
	.dc.b	$E8,$A7,$2D,$72,$7F,$A3,$C8,$7F,$A3,$4F,$F3,$13,$FE,$65,$23,$82
	.dc.b	$1F,$E6,$37,$2D,$32,$7F,$99,$5C,$B4,$FF,$2B,$BF,$E5,$AE,$FF,$F2
	.dc.b	$C4,$7F,$99,$63,$04,$76,$43,$FC,$CA,$E5,$A6,$4F,$F2,$BF,$7F,$F9
	.dc.b	$49,$FF,$2D,$B5,$CC,$7F,$95,$C8,$FF,$95,$DA,$F0,$6B,$C0,$FF,$1F
	.dc.b	$3F,$E4,$12,$38,$21,$FE,$41,$72,$D3,$27,$F9,$0D,$CB,$4F,$F1,$EB
	.dc.b	$FE,$40,$23,$82,$1F,$E4,$07,$2D,$32,$7F,$90,$9C,$B4,$FF,$0B,$10
	.dc.b	$FF,$0A,$14,$FF,$09,$14,$FF,$08,$9F,$E0,$E2,$3F,$E1,$A7,$2D,$7F
	.dc.b	$C2,$8F,$F8,$35,$FF,$4C,$F3,$7F,$83,$5F,$F0,$7C,$65,$FF,$0E,$C6
	.dc.b	$5D,$C6,$5E,$67,$F8,$69,$FE,$E4,$FF,$A9,$39,$3F,$BC,$B6,$32,$FF
	.dc.b	$8A,$23,$2D,$8C,$BF,$E3,$F8,$CB,$63,$FB,$93,$FE,$B0,$E4,$FE,$F2
	.dc.b	$D8,$CB,$FE,$34,$8C,$B6,$32,$FF,$92,$E3,$2D,$8F,$EE,$4F,$FA,$F3
	.dc.b	$93,$FB,$CB,$63,$2F,$F9,$02,$32,$D8,$CB,$FE,$57,$8C,$B6,$3F,$AE
	.dc.b	$BF,$E4,$8E,$2F,$EC,$2D,$8C,$BF,$E5,$F8,$CB,$63,$3C,$E9,$CB,$63
	.dc.b	$FA,$EC,$FF,$95,$38,$FF,$B0,$B6,$32,$FF,$9A,$E3,$2D,$8C,$CA,$72
	.dc.b	$D8,$FE,$75,$3F,$CE,$1A,$FF,$9D,$78,$3F,$9D,$36,$33,$29,$CB,$63
	.dc.b	$31,$FC,$F6,$7F,$CF,$DC,$16,$8D,$25,$7F,$3F,$9D,$B6,$76,$FF,$1C
	.dc.b	$8F,$F8,$00,$D7,$FD,$15,$E0,$7F,$9D,$09,$19,$94,$62,$7F,$93,$9C
	.dc.b	$8A,$7F,$23,$3F,$8E,$1F,$E8,$2D,$FF,$93,$CF,$FC,$01,$8F,$E4,$24
	.dc.b	$06,$EB,$F9,$33,$83,$A1,$B3,$47,$F2,$12,$1F,$C8,$4A,$7F,$21,$30
	.dc.b	$43,$04,$3C,$3D,$EE,$0F,$E6,$11,$87,$F8,$4B,$FC,$34,$FF,$58,$18
	.dc.b	$3F,$89,$90,$FE,$26,$53,$F8,$79,$C6,$BF,$D6,$ED,$E0,$FE,$14,$FF
	.dc.b	$0C,$BF,$E1,$53,$FE,$A5,$8F,$E1,$44,$3F,$89,$94,$FE,$22,$53,$91
	.dc.b	$FE,$46,$3F,$54,$53,$F5,$05,$7F,$4C,$DD,$7F,$52,$49,$FD,$0E,$BF
	.dc.b	$A0,$23,$82,$4F,$F9,$20,$FD,$61,$F8,$4A,$FC,$4C,$7E,$2F,$3F,$C5
	.dc.b	$4F,$F7,$12,$1F,$DB,$CA,$7E,$34,$A7,$E5,$A3,$F0,$AF,$E0,$8F,$C2
	.dc.b	$1F,$86,$3F,$11,$3F,$8F,$6E,$7F,$92,$24,$F2,$8E,$07,$DF,$7F,$B4
	.dc.b	$B4,$FC,$59,$4F,$C5,$14,$FC,$50,$CF,$F1,$51,$CA,$31,$9F,$E3,$1F
	.dc.b	$A2,$7F,$A0,$0F,$AA,$53,$27,$F9,$F4,$86,$8A,$68,$C1,$0F,$E3,$07
	.dc.b	$D3,$29,$FE,$82,$3E,$A9,$4C,$9F,$E7,$F2,$1A,$29,$A3,$E8,$1F,$32
	.dc.b	$3E,$55,$FC,$F8,$C9,$0F,$9A,$47,$4D,$B4,$FD,$CF,$39,$EE,$3B,$33
	.dc.b	$7F,$C0,$C4,$EA,$33,$1F,$C1,$E3,$15,$B8,$DD,$73,$8C,$4E,$A7,$A3
	.dc.b	$5E,$53,$B7,$2D,$73,$5E,$07,$F8,$CC,$F1,$6C,$ED,$B3,$FA,$B7,$F5
	.dc.b	$D5,$FD,$63,$5C,$47,$E0,$71,$FD,$9B,$63,$F5,$EF,$DB,$8F,$EA,$0D
	.dc.b	$D7,$2F,$DB,$8C,$B6,$3F,$3E,$FF,$34,$3E,$85,$7D,$8C,$B3,$1F,$AA
	.dc.b	$AF,$AA,$48,$CC,$7D,$F3,$94,$66,$79,$12,$33,$3C,$1F,$C1,$4E,$DC
	.dc.b	$CE,$DF,$04,$ED,$F2,$9A,$21,$A8,$E0,$F8,$CC,$1A,$30,$43,$C3,$1D
	.dc.b	$67,$C5,$1C,$08,$6C,$F0,$98,$38,$CF,$81,$A7,$72,$9B,$38,$9C,$E3
	.dc.b	$BB,$67,$C4,$49,$E9,$1C,$0E,$66,$DC,$CF,$28,$C9,$0E,$04,$74,$D9
	.dc.b	$EC,$F9,$CC,$4F,$48,$CC,$7A,$67,$89,$23,$31,$E3,$39,$46,$67,$91
	.dc.b	$23,$33,$C1,$F3,$4E,$DC,$CE,$DF,$04,$EF,$C5,$79,$8F,$37,$F5,$D4
	.dc.b	$9C,$C6,$1B,$7F,$BF,$6C,$FF,$07,$6B,$FE,$0B,$6B,$FE,$10,$6B,$D6
	.dc.b	$BC,$13,$CD,$B1,$CD,$AF,$36,$BC,$DA,$FC,$0B,$F8,$4F,$D0,$3E,$09
	.dc.b	$0F,$83,$3F,$0D,$F9,$C6,$0D,$18,$21,$D5,$E0,$D3,$65,$39,$10,$E2
	.dc.b	$78,$CC,$1B,$33,$87,$D7,$72,$6C,$CC,$FB,$23,$51,$89,$F1,$BB,$3A
	.dc.b	$10,$C1,$23,$8B,$96,$B9,$9E,$C4,$9E,$B3,$A2,$46,$67,$E0,$9C,$67
	.dc.b	$71,$98,$F5,$CF,$09,$C9,$23,$8C,$65,$B3,$C5,$F7,$18,$9E,$0F,$48
	.dc.b	$D3,$EB,$6B,$E3,$38,$10,$E1,$1C,$9E,$A6,$0D,$18,$21,$D5,$E0,$D3
	.dc.b	$65,$39,$10,$E2,$78,$CC,$1B,$33,$3F,$05,$C9,$B3,$33,$EC,$8D,$46
	.dc.b	$27,$C6,$EC,$E8,$43,$04,$8E,$2E,$5A,$E6,$7B,$12,$7A,$CE,$89,$19
	.dc.b	$9F,$D0,$DF,$F0,$AF,$63,$F8,$36,$06,$93,$0F,$DF,$36,$BF,$7C,$D7
	.dc.b	$F7,$CD,$7F,$7C,$D7,$D1,$7E,$17,$E3,$E1,$E4,$AF,$3C,$E6,$7E,$31
	.dc.b	$F5,$8F,$74,$68,$8E,$67,$93,$67,$93,$E0,$9D,$BE,$08,$CB,$D0,$E1
	.dc.b	$1A,$7C,$26,$DC,$3C,$CD,$90,$D1,$4E,$24,$3F,$7B,$87,$F3,$8F,$8C
	.dc.b	$F9,$E4,$9F,$E5,$4D,$9F,$E5,$4D,$7F,$95,$35,$FE,$54,$D7,$F9,$A3
	.dc.b	$FD,$0B,$D0,$7C,$93,$D7,$C3,$6F,$FC,$81,$3A,$9C,$92,$33,$3B,$BF
	.dc.b	$F2,$65,$6E,$73,$5B,$9C,$D6,$DF,$F6,$A5,$6D,$CC,$FD,$0C,$BE,$BC
	.dc.b	$7F,$82,$65,$F2,$2F,$85,$F5,$37,$1F,$3B,$6B,$C6,$D7,$BB,$5F,$F7
	.dc.b	$3D,$74,$AE,$64,$AE,$6D,$9E,$6D,$7D,$38,$FA,$DE,$6D,$55,$ED,$77
	.dc.b	$8F,$C3,$6D,$FC,$62,$3F,$09,$AE,$8C,$1F,$89,$3F,$1C,$7F,$89,$F9
	.dc.b	$BF,$D4,$EF,$FA,$B8,$FF,$53,$C7,$FA,$A6,$33,$1F,$EB,$03,$26,$DC
	.dc.b	$BF,$EB,$26,$BF,$EB,$51,$F3,$4F,$9B,$7C,$7F,$F8,$32,$30,$D8,$CD
	.dc.b	$FF,$A2,$67,$2D,$8C,$D7,$FA,$4E,$72,$D8,$CA,$24,$69,$C1,$E4,$8D
	.dc.b	$BA,$6B,$9A,$F2,$1F,$C6,$BE,$42,$5F,$F8,$55,$B3,$FE,$15,$6B,$FE
	.dc.b	$15,$6B,$FE,$15,$6B,$FE,$1F,$6F,$03,$E4,$DF,$1D,$FE,$2D,$6D,$FF
	.dc.b	$8B,$5B,$7F,$E2,$D6,$BF,$E2,$D6,$BF,$E2,$D6,$BF,$E3,$96,$D7,$7A
	.dc.b	$EC,$4A,$EC,$DA,$FF,$1D,$7C,$A3,$C1,$FF,$20,$36,$BF,$C8,$0D,$9D
	.dc.b	$CE,$89,$3A,$6C,$6A,$30,$47,$07,$F8,$FF,$2F,$E9,$24,$8F,$F2,$23
	.dc.b	$63,$51,$82,$38,$21,$FE,$47,$8E,$0F,$F9,$27,$E4,$BF,$AA,$92,$7F
	.dc.b	$CA,$0D,$8F,$F2,$83,$5E,$11,$A2,$3A,$6B,$FC,$E0,$87,$F9,$23,$1F
	.dc.b	$F3,$6D,$78,$A3,$FC,$B0,$D7,$FC,$C4,$D8,$DC,$FB,$27,$D4,$D7,$8B
	.dc.b	$5F,$DE,$9F,$D1,$4F,$EA,$47,$88,$C9,$FB,$53,$F7,$67,$EF,$76,$FD
	.dc.b	$CB,$83,$F7,$9D,$FF,$E8,$27,$07,$18,$D9,$23,$F4,$F3,$FB,$7B,$B5
	.dc.b	$FB,$56,$BF,$B2,$DF,$F6,$8F,$EC,$63,$2D,$9E,$44,$79,$3F,$E8,$5F
	.dc.b	$82,$FF,$06,$91,$C1,$23,$FD,$11,$19,$6C,$65,$FF,$45,$46,$5B,$1F
	.dc.b	$E9,$1F,$84,$FF,$0C,$91,$C1,$23,$FD,$29,$19,$6C,$65,$FF,$4B,$46
	.dc.b	$5B,$1F,$E9,$DF,$92,$FF,$12,$91,$C1,$23,$FD,$41,$19,$6C,$65,$FF
	.dc.b	$51,$46,$5B,$1F,$EA,$9F,$92,$FF,$18,$91,$C1,$23,$FD,$59,$19,$6C
	.dc.b	$65,$FF,$57,$46,$5B,$1F,$EB,$5E,$0F,$F1,$E9,$1C,$12,$3F,$D7,$11
	.dc.b	$96,$C6,$5F,$F5,$D4,$65,$B1,$FC,$FE,$7F,$A2,$11,$C1,$23,$FA,$04
	.dc.b	$65,$B1,$97,$F1,$DC,$69,$D4,$61,$FF,$04,$75,$CA,$FC,$74,$F0,$CF
	.dc.b	$F2,$6F,$F4,$38,$CB,$F9,$53,$F2,$64,$38,$3A,$78,$19,$3F,$A7,$57
	.dc.b	$89,$B1,$E7,$69,$EA,$21,$E7,$74,$FF,$59,$8E,$4D,$8F,$F2,$5B,$FC
	.dc.b	$00,$FF,$67,$ED,$FB,$58,$FE,$F9,$B7,$EE,$E3,$13,$FD,$EE,$30,$49
	.dc.b	$D9,$FE,$3C,$F9,$87,$A9,$FE,$43,$6C,$66,$30,$47,$04,$AF,$F1,$FC
	.dc.b	$F0,$6C,$ED,$B3,$B7,$F8,$37,$C4,$3C,$5F,$F0,$43,$67,$FC,$0C,$D9
	.dc.b	$FF,$02,$35,$E3,$3C,$09,$3C,$1B,$1C,$1A,$F0,$8F,$F0,$3B,$FE,$0B
	.dc.b	$E3,$78,$2F,$93,$FF,$4E,$92,$B7,$4A,$DC,$E6,$39,$C6,$27,$6D,$8C
	.dc.b	$D7,$59,$CB,$63,$35,$CD,$FF,$0D,$D7,$17,$F9,$67,$EF,$EF,$7F,$FC
	.dc.b	$5C,$DC,$3F,$C5,$8D,$9F,$F1,$53,$5E,$33,$C0,$93,$C1,$B1,$C1,$AF
	.dc.b	$08,$FC,$8F,$85,$FE,$72,$8F,$EC,$18,$FA,$5F,$3B,$5F,$98,$D7,$E4
	.dc.b	$37,$0F,$64,$6B,$1F,$80,$DB,$F8,$B1,$88,$E2,$D9,$FA,$FF,$29,$FE
	.dc.b	$D2,$25,$7C,$C6,$D7,$D0,$6B,$E6,$6B,$E6,$6B,$E6,$7F,$84,$3F,$E6
	.dc.b	$D3,$F6,$07,$F6,$E1,$83,$8F,$FB,$53,$6B,$F0,$F3,$A9,$FD,$83,$FB
	.dc.b	$38,$FD,$8B,$48,$7F,$77,$E0,$DB,$F8,$2F,$D4,$5E,$5A,$BF,$1B,$18
	.dc.b	$6C,$66,$7F,$27,$39,$6C,$65,$E8,$D7,$F4,$51,$B7,$33,$D0,$8F,$26
	.dc.b	$BE,$5F,$8C,$E4,$FE,$96,$51,$8C,$D1,$8A,$F4,$CE,$63,$15,$E8,$8C
	.dc.b	$12,$76,$48,$D3,$5F,$8D,$E1,$73,$24,$7C,$66,$C7,$C4,$6B,$C2,$34
	.dc.b	$47,$4D,$7E,$97,$CA,$F9,$C9,$1F,$49,$B1,$F4,$1A,$F0,$8D,$11,$D3
	.dc.b	$5F,$BD,$E3,$7C,$32,$47,$DE,$6C,$7D,$C6,$BC,$23,$44,$74,$D7,$F1
	.dc.b	$5F,$2B,$F3,$C9,$1F,$8A,$6C,$7E,$21,$AF,$08,$D1,$1D,$35,$FC,$B7
	.dc.b	$CA,$FD,$B2,$47,$E5,$9B,$1F,$94,$6B,$C2,$34,$47,$4D,$7F,$45,$F0
	.dc.b	$FF,$87,$24,$7E,$89,$B1,$FA,$06,$BC,$23,$44,$74,$D7,$F5,$DF,$1F
	.dc.b	$F9,$32,$47,$EB,$9B,$1F,$AC,$6B,$C2,$34,$47,$4D,$7F,$75,$F1,$FF
	.dc.b	$9F,$24,$7E,$E9,$B1,$FB,$86,$BC,$23,$44,$74,$D7,$F8,$37,$C5,$7E
	.dc.b	$AC,$91,$FC,$19,$B1,$FC,$11,$AF,$08,$D1,$1D,$35,$FE,$25,$D3,$FB
	.dc.b	$72,$47,$F1,$26,$C7,$F1,$06,$BC,$23,$44,$74,$D7,$F8,$F7,$C9,$7F
	.dc.b	$03,$24,$7F,$1E,$6C,$7F,$1C,$6B,$C2,$34,$47,$4D,$7F,$95,$71,$FF
	.dc.b	$0F,$24,$7F,$2A,$6C,$7F,$28,$6B,$C2,$34,$47,$4D,$7F,$9B,$7C,$47
	.dc.b	$F1,$B2,$47,$F3,$66,$C7,$F3,$46,$BC,$23,$44,$74,$D7,$FA,$17,$C4
	.dc.b	$7F,$27,$24,$7F,$42,$6C,$7F,$40,$6B,$C2,$34,$47,$4D,$7F,$A7,$7C
	.dc.b	$47,$F3,$32,$47,$F4,$E6,$C7,$F4,$C6,$BC,$23,$44,$74,$D7,$FA,$D7
	.dc.b	$C4,$7F,$3F,$24,$7F,$5A,$6C,$7F,$58,$6B,$C2,$34,$47,$4D,$7F,$B3
	.dc.b	$7C,$47,$F4,$B2,$47,$F6,$66,$C7,$F6,$46,$BC,$23,$44,$74,$D7,$FB
	.dc.b	$97,$83,$FD,$5C,$91,$FD,$C9,$B1,$FD,$C1,$AF,$08,$D1,$1D,$35,$FF
	.dc.b	$29,$F5,$BD,$AF,$F0,$59,$27,$FC,$17,$19,$6C,$66,$7F,$CA,$B3,$96
	.dc.b	$C6,$5E,$AD,$44,$8E,$0F,$8B,$B3,$FC,$42,$DA,$F0,$C7,$F8,$78,$8E
	.dc.b	$0F,$F0,$DC,$7F,$86,$5A,$F6,$1E,$27,$FA,$51,$D1,$FE,$20,$35,$5E
	.dc.b	$26,$DF,$9D,$F3,$35,$FF,$18,$B6,$38,$46,$AB,$E2,$91,$F8,$6F,$F9
	.dc.b	$6A,$FD,$B1,$E5,$F9,$F7,$C0,$FF,$86,$A1,$18,$AF,$F1,$DB,$63,$FD
	.dc.b	$06,$D7,$FD,$16,$D7,$A4,$F3,$8C,$47,$D0,$6B,$DA,$34,$49,$EE,$FF
	.dc.b	$8B,$3F,$13,$7E,$06,$F7,$30,$DA,$FF,$4A,$CE,$A7,$24,$8C,$CE,$EF
	.dc.b	$FD,$43,$5B,$9C,$D6,$E7,$35,$B7,$ED,$56,$DC,$CF,$C7,$FA,$27,$AF
	.dc.b	$49,$87,$F9,$8A,$32,$D8,$CD,$7F,$AB,$27,$2D,$8C,$BF,$EB,$68,$CB
	.dc.b	$63,$2F,$E2,$33,$FB,$6F,$E6,$DF,$CA,$BF,$A4,$6B,$F9,$B3,$D3,$5F
	.dc.b	$00,$8E,$08,$7C,$0C,$7E,$13,$FA,$8C,$7E,$13,$9C,$7E,$43,$6B,$9B
	.dc.b	$4F,$F4,$56,$9F,$EA,$97,$FD,$14,$D7,$FD,$0D,$1F,$AB,$7D,$0D,$7F
	.dc.b	$DA,$FF,$80,$7F,$B0,$7E,$07,$FD,$1A,$D8,$FF,$6A,$B5,$DF,$94,$9D
	.dc.b	$B6,$36,$D7,$71,$A2,$3A,$6B,$A3,$BF,$C7,$7F,$A5,$1B,$E4,$79,$23
	.dc.b	$11,$E8,$6B,$FC,$0E,$D1,$88,$FE,$27,$18,$9E,$44,$8E,$A4,$3A,$39
	.dc.b	$7F,$DF,$7F,$78,$F7,$FF,$D6,$30,$8C,$3F,$EB,$46,$BF,$EF,$C6,$BC
	.dc.b	$A3,$8C,$62,$3F,$8D,$35,$FE,$52,$D7,$B4,$70,$23,$DD,$AF,$76,$BD
	.dc.b	$CD,$10,$D1,$4D,$18,$21,$83,$F2,$DF,$05,$FE,$D0,$DB,$F8,$3B,$77
	.dc.b	$34,$93,$FC,$11,$AF,$F0,$06,$BD,$0D,$10,$F0,$9F,$12,$BD,$6D,$3D
	.dc.b	$05,$36,$7C,$87,$6D,$36,$53,$FA,$41,$FD,$5B,$D2,$7C,$1F,$DC,$6C
	.dc.b	$7F,$BB,$9A,$F0,$0E,$1E,$6D,$7F,$AA,$35,$ED,$1D,$48,$F5,$6B,$D6
	.dc.b	$38,$61,$F3,$A3,$E2,$1F,$1C,$F9,$46,$0F,$E1,$18,$7D,$27,$02,$39
	.dc.b	$3C,$08,$73,$21,$FD,$AF,$F6,$9F,$C8,$13,$0F,$CF,$B6,$3E,$10,$70
	.dc.b	$FC,$56,$BF,$39,$AF,$58,$FA,$AF,$E7,$0F,$AA,$53,$ED,$94,$FC,$91
	.dc.b	$F7,$CA,$7D,$B2,$9B,$3F,$57,$C9,$FA,$E2,$3F,$B5,$6B,$FB,$56,$BF
	.dc.b	$BE,$6B,$C8,$E0,$43,$F0,$5C,$0E,$DF,$E0,$6D,$8F,$8A,$D7,$71,$A2
	.dc.b	$3A,$6B,$A3,$04,$30,$7C,$4D,$7F,$82,$12,$FE,$5B,$63,$E5,$B5,$E1
	.dc.b	$1E,$C7,$F8,$6D,$D7,$F8,$D3,$67,$FB,$7C,$7B,$CF,$88,$53,$E2,$1F
	.dc.b	$24,$C0,$F3,$9E,$77,$E8,$5B,$B7,$E2,$1B,$3F,$AB,$6B,$FB,$46,$BF
	.dc.b	$E3,$D6,$BF,$E4,$C6,$BD,$E7,$B1,$27,$B3,$63,$B7,$C2,$FF,$38,$A3
	.dc.b	$49,$1E,$56,$BC,$0C,$10,$C1,$0F,$43,$CF,$93,$91,$23,$C2,$D7,$6D
	.dc.b	$7D,$D1,$FD,$9D,$B1,$97,$E0,$46,$5B,$1E,$7F,$90,$91,$EE,$6B,$E4
	.dc.b	$6B,$D9,$AF,$46,$BF,$66,$51,$8A,$F8,$FF,$21,$F1,$49,$5F,$31,$AF
	.dc.b	$C2,$6B,$EC,$6B,$EA,$6B,$DD,$FA,$1F,$15,$23,$EB,$B5,$F9,$AD,$7E
	.dc.b	$3B,$5F,$8A,$D7,$D4,$D9,$FB,$FF,$15,$F7,$49,$3F,$84,$6B,$F6,$9A
	.dc.b	$FD,$46,$BF,$49,$AF,$C7,$7F,$11,$F1,$52,$3F,$1E,$D7,$F0,$AD,$7E
	.dc.b	$FB,$5F,$BA,$D7,$E9,$36,$7F,$2F,$F2,$1F,$95,$24,$FE,$71,$AF,$E4
	.dc.b	$9A,$FE,$31,$AF,$E2,$9A,$FD,$F7,$F4,$1F,$21,$23,$F4,$ED,$7F,$3A
	.dc.b	$D7,$F2,$ED,$7F,$2A,$D7,$F1,$4D,$9F,$D7,$F4,$7E,$B4,$93,$FB,$46
	.dc.b	$BF,$AA,$6B,$FA,$46,$BF,$A2,$6B,$F9,$77,$F7,$1D,$D2,$3F,$7E,$D7
	.dc.b	$F6,$AD,$7F,$5E,$D7,$F5,$AD,$7F,$44,$D9,$FE,$0F,$F2,$1F,$C1,$49
	.dc.b	$3F,$C3,$1A,$FF,$02,$6B,$FB,$C6,$BF,$BA,$6B,$FA,$F7,$F8,$87,$C8
	.dc.b	$48,$FE,$2E,$D7,$F8,$6B,$5F,$E0,$ED,$7F,$82,$B5,$FD,$D3,$67,$F8
	.dc.b	$FF,$C8,$7F,$1D,$24,$FF,$24,$6B,$FC,$69,$AF,$F1,$46,$BF,$C4,$9A
	.dc.b	$FF,$07,$7F,$94,$7C,$44,$8F,$E5,$ED,$7F,$92,$B5,$FE,$3E,$D7,$F8
	.dc.b	$EB,$5F,$E2,$4D,$9F,$E6,$FC,$3F,$CD,$49,$3F,$CF,$1A,$FF,$32,$6B
	.dc.b	$FC,$B1,$AF,$F2,$A6,$BF,$C7,$DF,$E8,$1F,$21,$23,$FA,$3B,$5F,$E7
	.dc.b	$AD,$7F,$9B,$B5,$FE,$6A,$D7,$F9,$53,$67,$FA,$7F,$8D,$FD,$34,$93
	.dc.b	$FD,$51,$AF,$F4,$A6,$BF,$D1,$1A,$FF,$42,$6B,$FC,$DD,$FE,$B1,$E4
	.dc.b	$48,$FE,$BE,$D7,$FA,$AB,$5F,$E9,$ED,$7F,$A6,$B5,$FE,$84,$D5,$D9
	.dc.b	$FE,$1B,$7F,$CD,$2D,$44,$70,$DE,$CF,$F0,$EB,$C1,$AF,$03,$AB,$87
	.dc.b	$81,$FE,$21,$38,$1F,$E1,$CE,$CF,$F0,$F9,$1C,$10,$FF,$0F,$94,$FF
	.dc.b	$22,$F5,$FF,$88,$DB,$1C,$5D,$9F,$E8,$AC,$7F,$C0,$BE,$37,$F8,$0C
	.dc.b	$91,$FE,$0A,$6B,$FE,$00,$6B,$FD,$E9,$AF,$F7,$86,$BF,$DB,$5F,$F0
	.dc.b	$77,$CC,$48,$FF,$0A,$B5,$FF,$05,$B5,$FF,$02,$B5,$FF,$01,$B5,$FE
	.dc.b	$F0,$D7,$FC,$C4,$7F,$99,$4F,$F3,$41,$0C,$10,$D9,$2F,$E3,$3B,$39
	.dc.b	$FC,$4B,$F5,$A7,$5F,$FA,$0A,$30,$D8,$CC,$FF,$A9,$E7,$2D,$8C,$BF
	.dc.b	$EB,$28,$CB,$63,$2F,$E1,$63,$2D,$8C,$BF,$90,$8C,$B6,$32,$A3,$2F
	.dc.b	$C6,$3E,$79,$0F,$9E,$53,$E7,$94,$F9,$E5,$39,$1C,$08,$79,$FE,$09
	.dc.b	$F1,$7F,$15,$B1,$F0,$DA,$FC,$16,$BC,$63,$81,$1E,$0D,$78,$35,$E0
	.dc.b	$7F,$B5,$7E,$39,$FB,$B7,$EE,$23,$0D,$7F,$DC,$31,$AE,$0F,$F5,$DB
	.dc.b	$63,$71,$89,$E1,$1F,$B7,$9D,$3C,$A3,$2F,$F0,$28,$CB,$B8,$CA,$8C
	.dc.b	$BB,$8C,$BF,$C0,$8F,$E0,$07,$F0,$59,$FE,$16,$E4,$FF,$51,$9F,$C1
	.dc.b	$0A,$6F,$83,$F8,$53,$B3,$44,$3F,$39,$A7,$E7,$9F,$D1,$35,$FD,$17
	.dc.b	$07,$F0,$57,$F8,$39,$FE,$E0,$FC,$B9,$F0,$7F,$EE,$28,$46,$1D,$46
	.dc.b	$63,$11,$FC,$7C,$FE,$4E,$E1,$C3,$5F,$E4,$8D,$77,$1C,$C8,$F6,$21
	.dc.b	$D4,$7C,$A3,$D4,$F3,$36,$3F,$DF,$2D,$78,$35,$F3,$35,$FE,$6C,$D7
	.dc.b	$F9,$FB,$5E,$F1,$D8,$8F,$66,$BF,$CB,$8F,$96,$53,$E9,$DF,$CF,$69
	.dc.b	$EC,$3E,$2D,$7C,$07,$DD,$7F,$15,$FA,$31,$97,$67,$13,$E5,$B9,$69
	.dc.b	$93,$FD,$41,$F1,$1F,$EC,$12,$3F,$EC,$86,$BF,$EB,$E6,$BF,$EB,$86
	.dc.b	$BF,$EB,$66,$BD,$0F,$F6,$F7,$EA,$AF,$4B,$FD,$C4,$D8,$CC,$60,$8E
	.dc.b	$08,$7F,$B9,$63,$83,$63,$6D,$36,$7F,$BA,$61,$F4,$1E,$02,$47,$91
	.dc.b	$AF,$56,$BC,$5A,$FF,$BC,$63,$9B,$C4,$87,$97,$E8,$3D,$A4,$7E,$0B
	.dc.b	$5D,$18,$21,$82,$1E,$97,$83,$EB,$DF,$E4,$91,$F9,$AD,$7E,$33,$5E
	.dc.b	$06,$88,$68,$A7,$E8,$34,$FC,$BB,$4F,$CD,$10,$FD,$3B,$A7,$F4,$AD
	.dc.b	$7F,$4A,$D7,$F4,$A7,$F1,$C3,$FC,$3A,$7F,$18,$29,$FC,$54,$A7,$F1
	.dc.b	$B3,$81,$FC,$6C,$C1,$C4,$C1,$FC,$6B,$4F,$D5,$34,$FF,$11,$1F,$AC
	.dc.b	$3F,$60,$64,$A7,$F2,$03,$66,$4F,$DB,$9F,$8B,$F5,$3F,$32,$47,$F3
	.dc.b	$AD,$7F,$2C,$D7,$81,$A2,$1A,$29,$D8,$F3,$14,$F1,$18,$21,$FD,$F7
	.dc.b	$F5,$A7,$67,$F8,$4D,$B1,$FE,$11,$6B,$FE,$0F,$6B,$FE,$0D,$6B,$FE
	.dc.b	$0B,$6B,$D6,$3A,$11,$E8,$D7,$A3,$EF,$7F,$4D,$35,$76,$33,$4E,$63
	.dc.b	$0D,$8C,$CF,$C4,$39,$94,$E6,$E9,$FE,$09,$F3,$9F,$C3,$08,$FF,$10
	.dc.b	$6B,$A3,$04,$30,$43,$F8,$43,$C1,$FE,$15,$F4,$3F,$C6,$08,$FF,$1C
	.dc.b	$6B,$FC,$55,$AF,$03,$44,$34,$53,$F8,$B7,$CF,$7F,$24,$23,$FC,$A1
	.dc.b	$AF,$F2,$16,$BC,$0D,$10,$D1,$4F,$DD,$1F,$B8,$3F,$74,$60,$87,$EE
	.dc.b	$8C,$94,$E0,$43,$67,$EF,$8F,$DA,$EB,$FE,$16,$27,$37,$F8,$49,$B1
	.dc.b	$A8,$C1,$FE,$16,$E6,$FF,$08,$13,$6F,$DC,$CF,$EC,$CF,$DA,$C2,$3F
	.dc.b	$C1,$9E,$BC,$9F,$F0,$A3,$5F,$EC,$AD,$7F,$C1,$AD,$7F,$C1,$AD,$7F
	.dc.b	$C2,$6D,$7F,$C1,$3E,$EF,$F8,$20,$8F,$F8,$21,$AF,$F8,$21,$AF,$F8
	.dc.b	$21,$AF,$F8,$21,$AF,$F8,$21,$F6,$D7,$BE,$3F,$87,$9E,$F3,$F8,$01
	.dc.b	$FE,$94,$E7,$57,$95,$B1,$F0,$08,$7C,$02,$9F,$00,$A7,$C0,$29,$F0
	.dc.b	$0F,$EE,$BD,$1F,$DD,$08,$E0,$87,$F7,$17,$2D,$32,$7F,$6F,$72,$D3
	.dc.b	$DB,$5E,$F7,$ED,$E1,$F0,$88,$E9,$AE,$8C,$10,$C1,$0F,$8F,$1D,$5B
	.dc.b	$1C,$9A,$71,$29,$C4,$FE,$56,$7D,$12,$C7,$36,$C7,$36,$9C,$CA,$73
	.dc.b	$3F,$9B,$90,$E8,$58,$E8,$D8,$E8,$D3,$A1,$4E,$87,$22,$47,$26,$C7
	.dc.b	$26,$9C,$8A,$72,$8E,$04,$8E,$0D,$8F,$0F,$07,$F3,$7F,$9B,$FF,$1B
	.dc.b	$92,$3F,$C7,$6D,$7F,$C6,$8D,$7F,$C6,$0D,$7F,$C5,$0D,$7F,$9C,$BE
	.dc.b	$0A,$91,$E7,$AF,$13,$E7,$8C,$B4,$D3,$87,$ED,$90,$FA,$65,$3E,$51
	.dc.b	$4F,$86,$58,$EE,$D8,$E8,$F2,$21,$C8,$A7,$22,$C7,$26,$C7,$27,$81
	.dc.b	$0E,$05,$8E,$0D,$8E,$0E,$89,$1A,$6C,$6A,$30,$47,$04,$8F,$8E,$43
	.dc.b	$E5,$10,$F8,$06,$4A,$64,$F5,$99,$29,$93,$D0,$64,$A6,$4F,$39,$92
	.dc.b	$99,$8F,$43,$96,$B9,$8F,$53,$4D,$38,$7C,$E4,$3C,$A5,$3C,$65,$3C
	.dc.b	$25,$8E,$ED,$8E,$8F,$22,$1C,$8A,$72,$2C,$72,$6C,$72,$78,$10,$E0
	.dc.b	$58,$E0,$D8,$FE,$07,$5F,$E2,$77,$FD,$4E,$48,$FF,$55,$B5,$FF,$52
	.dc.b	$35,$FF,$50,$35,$FF,$4C,$35,$FF,$14,$BE,$57,$C6,$49,$F1,$B6,$35
	.dc.b	$18,$23,$82,$47,$D4,$21,$F5,$C8,$7C,$D3,$25,$32,$7C,$83,$25,$32
	.dc.b	$7C,$53,$25,$32,$79,$CC,$94,$CC,$7A,$1C,$B5,$CC,$7A,$91,$0B,$0F
	.dc.b	$8F,$C4,$48,$C3,$83,$71,$5A,$58,$DB,$B6,$BC,$E3,$82,$8C,$BE,$36
	.dc.b	$BD,$0E,$51,$C2,$33,$1C,$7E,$3F,$04,$7C,$4D,$7C,$4D,$7D,$8D,$7C
	.dc.b	$0D,$7E,$1B,$ED,$CD,$FA,$13,$7C,$36,$BE,$24,$EA,$72,$48,$CC,$ED
	.dc.b	$F9,$B3,$B7,$33,$B7,$33,$B7,$EB,$4E,$DC,$CF,$DB,$7F,$1C,$7D,$7C
	.dc.b	$7E,$36,$32,$3E,$A3,$5F,$96,$60,$86,$08,$7D,$73,$25,$3E,$59,$F3
	.dc.b	$88,$60,$87,$CE,$32,$53,$27,$D0,$32,$53,$FD,$C1,$3F,$E6,$C6,$EF
	.dc.b	$FE,$6C,$7F,$CD,$C4,$3F,$DC,$D9,$7E,$7D,$F8,$67,$E2,$EF,$F1,$1E
	.dc.b	$C4,$8F,$CB,$35,$FC,$B3,$5F,$D0,$35,$FC,$A3,$5F,$D3,$B5,$FC,$D6
	.dc.b	$1F,$35,$F8,$C4,$3F,$38,$E5,$A7,$CC,$29,$F4,$0A,$71,$3F,$26,$7C
	.dc.b	$62,$9F,$28,$A7,$CD,$21,$F2,$CA,$68,$C1,$0F,$E0,$78,$3E,$2F,$F6
	.dc.b	$AD,$8F,$DF,$35,$FE,$04,$D7,$F8,$3B,$5F,$E1,$8D,$7A,$C7,$42,$3D
	.dc.b	$1A,$F4,$7C,$3F,$17,$23,$C2,$D7,$C2,$D7,$C2,$D7,$C2,$D7,$DA,$D7
	.dc.b	$F7,$59,$7E,$2D,$FC,$69,$F8,$B3,$D6,$E5,$A7,$C2,$7F,$60,$FF,$39
	.dc.b	$3F,$60,$7E,$30,$87,$E3,$0A,$7E,$30,$A7,$E3,$0A,$7E,$30,$A7,$F2
	.dc.b	$5C,$3F,$1D,$93,$E6,$3F,$91,$12,$7F,$98,$35,$FE,$60,$D7,$F9,$D3
	.dc.b	$5F,$E5,$CD,$7F,$A0,$BF,$CB,$7E,$62,$47,$F3,$B6,$BF,$CE,$DA,$FF
	.dc.b	$45,$6B,$FC,$E5,$AF,$F4,$C6,$BF,$A6,$3F,$40,$7F,$3D,$3F,$3C,$64
	.dc.b	$FD,$11,$93,$F4,$64,$30,$43,$81,$F9,$E3,$FA,$21,$FA,$72,$9F,$A7
	.dc.b	$29,$B9,$6F,$FD,$06,$32,$7E,$94,$A7,$E9,$BE,$4B,$F4,$A4,$7F,$4A
	.dc.b	$D7,$F4,$AD,$7F,$4A,$D7,$F4,$AD,$7A,$1F,$AE,$F9,$2F,$D8,$91,$C1
	.dc.b	$0F,$D8,$B9,$69,$93,$F6,$6E,$5A,$7E,$EB,$E4,$BF,$7A,$47,$04,$3F
	.dc.b	$7A,$E5,$A6,$4F,$DF,$B9,$69,$FB,$B2,$1F,$BB,$29,$FC,$1B,$5F,$DF
	.dc.b	$B4,$C9,$FC,$21,$CB,$4C,$9F,$BA,$3F,$81,$90,$FE,$06,$53,$F8,$19
	.dc.b	$4F,$E0,$65,$39,$1C,$08,$7F,$1A,$F8,$DF,$E4,$24,$7F,$90,$B5,$FE
	.dc.b	$42,$D7,$81,$A2,$1A,$29,$FC,$9B,$E3,$7F,$96,$91,$FE,$5A,$D7,$F9
	.dc.b	$6B,$5E,$06,$88,$68,$A7,$F1,$63,$F8,$D9,$0C,$10,$FE,$6B,$B7,$F1
	.dc.b	$D6,$9F,$C7,$4F,$F2,$4F,$17,$F1,$E7,$F8,$A1,$FC,$6C,$87,$F1,$B2
	.dc.b	$9F,$C6,$CA,$70,$34,$43,$45,$3F,$CA,$3F,$2B,$FE,$3D,$23,$FE,$4B
	.dc.b	$6B,$FE,$4B,$6B,$FE,$55,$6B,$FE,$49,$6B,$D0,$FF,$2A,$79,$5F,$E4
	.dc.b	$92,$38,$21,$FE,$57,$72,$D3,$27,$F9,$8D,$CB,$4F,$E6,$24,$3F,$98
	.dc.b	$94,$FE,$62,$53,$F9,$89,$4F,$E6,$25,$3F,$CC,$9C,$1F,$CD,$5F,$F3
	.dc.b	$37,$07,$F2,$AE,$03,$49,$1F,$CD,$C8,$7F,$37,$29,$FC,$DC,$FF,$3B
	.dc.b	$7C,$C7,$F9,$AC,$8E,$08,$7F,$9E,$5C,$B4,$C9,$FE,$85,$72,$D3,$FD
	.dc.b	$05,$F0,$FF,$E7,$72,$38,$21,$FE,$85,$72,$D3,$27,$FA,$45,$CB,$4F
	.dc.b	$F4,$77,$CC,$7F,$A0,$C8,$E0,$87,$FA,$45,$CB,$4C,$9F,$E9,$D7,$2D
	.dc.b	$3F,$AA,$9F,$D3,$0F,$EA,$05,$3F,$A8,$5F,$FA,$48,$9C,$FF,$E3,$D8
	.dc.b	$CB,$5F,$E9,$24,$8F,$E9,$EF,$F5,$56,$BF,$D5,$5A,$FF,$55,$E4,$FE
	.dc.b	$A9,$18,$8F,$EA,$9C,$EE,$8F,$F1,$B9,$23,$FC,$6E,$D7,$FC,$6E,$D7
	.dc.b	$FC,$6E,$D7,$FC,$6E,$D7,$FC,$86,$FF,$90,$BC,$B9,$1F,$E4,$26,$BF
	.dc.b	$E4,$26,$BF,$E4,$26,$BF,$E4,$26,$BF,$E4,$E6,$BF,$EA,$BE,$7F,$EC
	.dc.b	$4D,$3F,$B5,$90,$FE,$D6,$53,$FB,$59,$4F,$ED,$65,$3F,$B5,$9F,$D9
	.dc.b	$08,$7F,$62,$29,$FE,$56,$E7,$FE,$CC,$D3,$FB,$51,$4D,$9F,$E6,$67
	.dc.b	$6F,$F9,$A3,$9F,$FB,$13,$FE,$60,$6B,$93,$FB,$41,$92,$99,$3F,$B6
	.dc.b	$90,$FE,$D6,$7F,$9B,$BE,$57,$FC,$E6,$47,$FC,$E6,$D7,$46,$08,$60
	.dc.b	$87,$F9,$D9,$E0,$FF,$9E,$3E,$57,$FC,$FE,$47,$FC,$FE,$D7,$FC,$FE
	.dc.b	$D7,$81,$A2,$1A,$29,$FD,$F4,$FF,$01,$10,$C1,$0F,$EF,$C6,$4A,$64
	.dc.b	$FF,$44,$9F,$E0,$23,$FD,$E5,$D3,$FE,$7B,$CD,$D7,$49,$3F,$DF,$CF
	.dc.b	$F7,$C3,$87,$FB,$F9,$FE,$04,$3F,$D3,$7F,$2B,$FE,$A0,$23,$FE,$A0
	.dc.b	$6B,$A3,$04,$30,$43,$FD,$46,$F0,$7F,$D4,$BF,$2B,$FE,$AC,$23,$FE
	.dc.b	$AC,$6B,$FE,$AC,$6B,$C0,$D1,$0D,$14,$FF,$58,$FC,$AF,$FA,$E0,$8F
	.dc.b	$FA,$E1,$AF,$FA,$E1,$AF,$03,$44,$34,$53,$FC,$3A,$7F,$86,$C8,$60
	.dc.b	$87,$FB,$03,$B7,$FC,$48,$D3,$FC,$42,$53,$89,$4F,$F0,$E9,$FE,$1B
	.dc.b	$21,$82,$1F,$EC,$CE,$DF,$F1,$23,$4F,$F1,$09,$4E,$25,$3F,$DA,$BF
	.dc.b	$2B,$FE,$DC,$23,$FE,$DC,$6B,$FE,$DC,$6B,$C0,$D1,$0D,$14,$FF,$70
	.dc.b	$FC,$AF,$FB,$A0,$8F,$FB,$A1,$AF,$FB,$A1,$AF,$03,$44,$34,$53,$FC
	.dc.b	$70,$7F,$BD,$BB,$FF,$C7,$0D,$3F,$DB,$07,$F8,$E8,$FF,$20,$14,$E2
	.dc.b	$39,$5D,$FF,$E3,$D8,$C2,$61,$FE,$DC,$29,$98,$E2,$E0,$C4,$56,$9F
	.dc.b	$ED,$A8,$78,$1F,$ED,$A2,$46,$5A,$FF,$92,$0F,$F7,$01,$D7,$C2,$79
	.dc.b	$5C,$A3,$0D,$7C,$B1,$A9,$F1,$36,$37,$18,$9E,$11,$EE,$9D,$3C,$A3
	.dc.b	$2F,$C3,$8C,$BB,$8C,$A8,$CB,$B8,$CC,$7A,$3C,$2F,$01,$27,$D2,$D7
	.dc.b	$C0,$D7,$B3,$5E,$8D,$7C,$12,$8F,$F2,$FC,$7D,$D6,$DE,$63,$FC,$D4
	.dc.b	$43,$FC,$D4,$53,$FC,$D4,$53,$FC,$D4,$53,$F0,$1E,$17,$F9,$81,$FC
	.dc.b	$2C,$23,$11,$A3,$FC,$CE,$69,$CB,$F9,$16,$C7,$18,$D9,$FE,$1B,$F9
	.dc.b	$EF,$E7,$6F,$E3,$7B,$11,$88,$FC,$9C,$63,$93,$F2,$0D,$8E,$13,$B7
	.dc.b	$F3,$71,$97,$8C,$65,$FD,$04,$65,$DC,$65,$F2,$4F,$68,$C4,$EE,$7F
	.dc.b	$51,$EA,$BF,$2A,$E0,$DB,$FD,$23,$67,$F5,$0D,$7F,$5A,$D7,$F6,$2D
	.dc.b	$7A,$CF,$42,$4F,$46,$C7,$47,$CB,$F9,$CB,$87,$E2,$B6,$7E,$63,$5F
	.dc.b	$96,$D7,$E3,$B5,$F8,$AD,$7F,$CA,$0D,$C0,$D2,$4F,$80,$93,$FE,$9D
	.dc.b	$C9,$E3,$7F,$A7,$5B,$19,$8C,$11,$C1,$FE,$9A,$3F,$23,$E3,$7F,$A8
	.dc.b	$1F,$F4,$F1,$0F,$F4,$F1,$4D,$9C,$4F,$E3,$4F,$F1,$77,$FD,$5A,$7F
	.dc.b	$0C,$D7,$F8,$31,$1C,$10,$FE,$0C,$E5,$A6,$4F,$E0,$CE,$5A,$7F,$18
	.dc.b	$9F,$E2,$44,$70,$43,$F8,$93,$96,$99,$3F,$89,$39,$69,$FE,$B4,$3F
	.dc.b	$6B,$E3,$7F,$AD,$1A,$64,$FE,$74,$E5,$A6,$4F,$E8,$CE,$5A,$7F,$AD
	.dc.b	$0F,$E0,$3E,$37,$FA,$D1,$A6,$4F,$E8,$CE,$5A,$64,$FE,$A4,$E5,$A7
	.dc.b	$FA,$D0,$FE,$1B,$E3,$7F,$AD,$1A,$64,$FE,$A4,$E5,$A6,$4F,$EB,$CE
	.dc.b	$5A,$7F,$48,$C3,$FA,$11,$1C,$10,$FE,$84,$E5,$A6,$4F,$E8,$4E,$5A
	.dc.b	$7F,$54,$AF,$E9,$C4,$70,$43,$FA,$73,$96,$99,$3F,$A7,$39,$69,$FE
	.dc.b	$D4,$C7,$FD,$9E,$DF,$1B,$FD,$9D,$18,$8F,$EE,$8D,$7F,$BC,$B5,$FE
	.dc.b	$73,$E3,$7F,$B3,$08,$FF,$B4,$DA,$FF,$B4,$DF,$D3,$47,$FB,$46,$3F
	.dc.b	$DA,$1F,$96,$BD,$7F,$E6,$2D,$8F,$E7,$0D,$7F,$9B,$B5,$FE,$66,$D7
	.dc.b	$F9,$8B,$5E,$B1,$D0,$8F,$46,$BD,$1F,$E5,$BF,$91,$B4,$A4,$9F,$E5
	.dc.b	$F1,$96,$C6,$5F,$E6,$31,$96,$C6,$5F,$E6,$71,$96,$C6,$5F,$E9,$3F
	.dc.b	$97,$3D,$9F,$EA,$6D,$8F,$EB,$4D,$7F,$AC,$B5,$FE,$AA,$D7,$9C,$72
	.dc.b	$23,$C9,$AF,$26,$BF,$EF,$C3,$FE,$00,$90,$FF,$80,$25,$34,$60,$86
	.dc.b	$08,$7F,$90,$BC,$47,$27,$F2,$76,$C7,$F8,$FD,$79,$7F,$EF,$C6,$C6
	.dc.b	$63,$04,$70,$43,$FC,$95,$1C,$1B,$1F,$E3,$7C,$24,$7F,$90,$1A,$FF
	.dc.b	$8D,$DA,$FF,$8D,$5A,$FF,$8C,$DA,$FF,$8A,$1B,$D9,$FE,$4C,$E2,$FF
	.dc.b	$22,$12,$7F,$C9,$AD,$7F,$C8,$8D,$7F,$C8,$4D,$7F,$C8,$0D,$7F,$C6
	.dc.b	$AF,$C7,$F2,$E4,$7C,$76,$BF,$01,$AF,$F9,$BB,$04,$62,$3E,$93,$5F
	.dc.b	$F3,$B4,$69,$E8,$7D,$3F,$2F,$EB,$11,$C1,$0F,$9E,$E5,$A6,$4F,$13
	.dc.b	$96,$9F,$7F,$CB,$FC,$11,$1C,$10,$FB,$6E,$5A,$64,$F7,$39,$69,$FE
	.dc.b	$7C,$E3,$FF,$3B,$11,$C1,$0F,$F3,$B3,$96,$99,$3F,$CE,$CE,$5A,$7F
	.dc.b	$A2,$38,$FF,$D0,$44,$70,$43,$FD,$04,$E5,$A6,$4F,$F4,$13,$96,$9F
	.dc.b	$A3,$F2,$FF,$4C,$47,$04,$3F,$3E,$E5,$A6,$4F,$C1,$39,$69,$FA,$FF
	.dc.b	$2F,$F6,$44,$70,$43,$F5,$6E,$5A,$64,$FC,$73,$96,$9F,$EA,$FF,$D5
	.dc.b	$FF,$7A,$FF,$AB,$1A,$FF,$A9,$DA,$FF,$A9,$5A,$FF,$A8,$DB,$A9,$A4
	.dc.b	$E0,$FF,$5A,$B6,$7C,$11,$83,$3C,$12,$3C,$8D,$7C,$8D,$7C,$8D,$7C
	.dc.b	$8D,$7C,$8D,$AF,$26,$DE,$22,$4F,$C1,$6B,$F0,$5A,$FC,$16,$BF,$05
	.dc.b	$AF,$C1,$7F,$93,$79,$92,$3F,$93,$35,$FE,$3E,$D7,$F8,$03,$5F,$E6
	.dc.b	$0D,$7F,$7E,$D5,$BF,$FA,$55,$BD,$FF,$CD,$48,$FF,$35,$6B,$FC,$B9
	.dc.b	$AF,$F0,$F6,$BF,$CF,$DA,$FD,$BE,$0F,$AE,$47,$F0,$CD,$7F,$0C,$D7
	.dc.b	$F0,$CD,$7F,$0C,$D7,$A1,$F8,$FE,$0F,$C1,$91,$C1,$0F,$C9,$39,$69
	.dc.b	$93,$F2,$8E,$5A,$7F,$55,$F3,$3F,$AC,$11,$C1,$0F,$EA,$2E,$5A,$64
	.dc.b	$FE,$68,$E5,$A7,$F6,$1F,$33,$FB,$21,$1C,$10,$FE,$BA,$E5,$A6,$4F
	.dc.b	$E8,$0E,$5A,$7F,$6D,$F3,$3F,$B8,$11,$C1,$0F,$ED,$2E,$5A,$64,$FE
	.dc.b	$98,$E5,$A7,$EF,$F8,$7F,$66,$47,$04,$3F,$81,$39,$69,$93,$F8,$23
	.dc.b	$96,$9F,$C2,$F8,$7F,$7E,$47,$04,$3F,$87,$39,$69,$93,$F8,$83,$96
	.dc.b	$9F,$E1,$1F,$33,$FC,$28,$47,$04,$3F,$C1,$AE,$5A,$64,$FE,$E0,$E5
	.dc.b	$A7,$F8,$77,$CC,$FF,$10,$11,$C1,$0F,$F0,$CB,$96,$99,$3F,$BE,$39
	.dc.b	$69,$FC,$BF,$6F,$E4,$64,$70,$43,$F9,$93,$96,$99,$3F,$9A,$39,$69
	.dc.b	$FC,$EF,$5F,$E5,$E4,$70,$43,$F9,$F3,$96,$99,$3F,$A0,$39,$69,$FD
	.dc.b	$1F,$3F,$E7,$64,$70,$43,$FA,$53,$96,$99,$3F,$A6,$39,$69,$FE,$4D
	.dc.b	$F3,$3F,$CA,$04,$70,$43,$FC,$92,$E5,$A6,$4F,$F1,$63,$96,$9F,$E5
	.dc.b	$9F,$0F,$FC,$B8,$47,$04,$3F,$CA,$AE,$5A,$64,$FF,$1C,$39,$69,$FD
	.dc.b	$AF,$5F,$EB,$E4,$70,$43,$FB,$73,$96,$99,$3F,$B8,$39,$69,$FD,$DF
	.dc.b	$5F,$ED,$64,$70,$43,$FB,$D3,$96,$99,$3F,$BE,$39,$69,$FE,$7D,$F3
	.dc.b	$3F,$D0,$04,$70,$43,$FC,$F2,$E5,$A6,$4F,$F2,$E3,$96,$83,$FD,$AF
	.dc.b	$E6,$7F,$B3,$88,$E4,$FF,$6C,$B8,$76,$64,$D7,$82,$8D,$47,$03,$47
	.dc.b	$78,$D3,$80,$E1,$EA,$7F,$BA,$BC,$CF,$F7,$09,$1E,$10,$8C,$46,$CC
	.dc.b	$B8,$7F,$DE,$2D,$7C,$ED,$7F,$C5,$1B,$FF,$89,$08,$FF,$8C,$DA,$FF
	.dc.b	$8C,$DA,$FF,$8C,$DA,$FF,$8C,$DA,$F4,$3F,$C7,$9B,$FF,$8C,$08,$E0
	.dc.b	$87,$F8,$FD,$CB,$4C,$9F,$E4,$17,$2D,$3E,$6F,$33,$CC,$F8,$91,$86
	.dc.b	$BF,$5E,$35,$7F,$45,$B1,$B8,$C4,$F0,$8F,$C1,$4E,$9E,$51,$97,$F1
	.dc.b	$51,$97,$71,$95,$19,$77,$19,$8F,$47,$99,$E0,$24,$FA,$5A,$F8,$1A
	.dc.b	$F6,$6B,$D1,$AF,$82,$51,$FE,$60,$DE,$4F,$F9,$BD,$AF,$F9,$BD,$AF
	.dc.b	$F9,$BD,$AF,$F9,$BD,$AF,$F9,$BD,$B5,$FE,$6F,$DF,$FC,$DA,$49,$FF
	.dc.b	$3E,$35,$FF,$3E,$35,$FF,$3E,$35,$FF,$3E,$35,$FF,$3E,$3F,$E7,$8D
	.dc.b	$64,$7F,$A1,$DA,$FF,$A1,$DA,$FF,$A1,$DA,$FF,$A1,$DA,$FF,$A1,$DA
	.dc.b	$FE,$CA,$FF,$68,$D3,$F5,$EE,$34,$F9,$71,$82,$1F,$BE,$72,$D8,$E0
	.dc.b	$47,$67,$F0,$98,$DB,$7C,$8E,$C4,$7F,$7E,$D7,$B3,$5E,$06,$88,$68
	.dc.b	$A7,$FA,$A7,$E7,$3F,$D4,$24,$7F,$D6,$0D,$7F,$D6,$0D,$78,$1A,$21
	.dc.b	$A2,$9F,$EB,$5F,$4B,$FD,$5A,$47,$FD,$78,$D7,$FD,$78,$D7,$81,$A2
	.dc.b	$1A,$29,$FC,$CF,$B5,$E7,$39,$BF,$95,$46,$1B,$19,$9F,$E6,$33,$96
	.dc.b	$C6,$5F,$E7,$51,$96,$C6,$5F,$03,$5F,$E9,$11,$B7,$33,$E0,$23,$FD
	.dc.b	$03,$CE,$79,$DC,$61,$18,$9F,$E8,$4D,$8E,$13,$B8,$C4,$7F,$4F,$6B
	.dc.b	$FD,$55,$AF,$F5,$B6,$BD,$A3,$91,$27,$BB,$E3,$A7,$9C,$DF,$E5,$B6
	.dc.b	$FE,$4B,$5F,$8E,$D7,$EA,$35,$F8,$AD,$53,$DA,$7A,$92,$3A,$B5,$EA
	.dc.b	$FF,$BD,$BD,$99,$1F,$CF,$BC,$34,$62,$3F,$9E,$46,$1D,$46,$5F,$E7
	.dc.b	$31,$97,$71,$97,$F9,$CC,$65,$DC,$7F,$45,$87,$CA,$52,$4F,$8D,$AF
	.dc.b	$89,$AF,$81,$AF,$66,$BF,$D1,$23,$AB,$FD,$F3,$D0,$FE,$1C,$47,$04
	.dc.b	$3F,$87,$39,$69,$93,$F8,$63,$96,$9F,$E0,$8F,$43,$F8,$D1,$1C,$10
	.dc.b	$FE,$34,$E5,$A6,$4F,$E3,$0E,$5A,$7F,$72,$FD,$63,$F9,$4F,$31,$A4
	.dc.b	$8F,$EE,$0D,$78,$18,$21,$82,$1F,$DE,$5E,$7F,$29,$C8,$91,$E1,$6B
	.dc.b	$B6,$BF,$DE,$A1,$18,$8F,$EF,$31,$87,$91,$0F,$07,$CA,$7A,$08,$FB
	.dc.b	$DA,$E8,$C1,$0C,$10,$F1,$3C,$1F,$F0,$3F,$2F,$F8,$40,$8F,$F8,$65
	.dc.b	$AF,$F8,$41,$AF,$03,$44,$34,$53,$FC,$27,$8F,$F8,$70,$8F,$F8,$95
	.dc.b	$AF,$F8,$71,$AF,$03,$44,$34,$53,$E6,$7C,$27,$DC,$23,$F8,$36,$BF
	.dc.b	$65,$AF,$03,$44,$34,$53,$EC,$7C,$A7,$E2,$08,$FE,$3D,$AF,$E1,$5A
	.dc.b	$F0,$34,$43,$45,$3F,$B9,$7A,$BF,$DF,$48,$FF,$9B,$9A,$FF,$7D,$6B
	.dc.b	$C0,$D1,$0D,$14,$FE,$FD,$E8,$7F,$82,$48,$FF,$9E,$9A,$FF,$82,$5A
	.dc.b	$F0,$34,$43,$45,$3F,$C1,$5E,$87,$F8,$54,$8F,$FA,$19,$AF,$F8,$55
	.dc.b	$AF,$03,$44,$34,$53,$F5,$1F,$29,$FB,$82,$3F,$BF,$6B,$FB,$56,$BC
	.dc.b	$0D,$10,$D1,$4F,$DA,$7C,$A7,$F0,$42,$3F,$C2,$DA,$FF,$01,$6B,$C0
	.dc.b	$D1,$0D,$14,$FF,$39,$F0,$7F,$9F,$08,$FF,$A1,$5A,$FF,$9F,$1A,$F0
	.dc.b	$34,$43,$45,$3F,$CF,$F8,$7F,$A2,$08,$FF,$A4,$5A,$FF,$A2,$1A,$F0
	.dc.b	$34,$43,$45,$3F,$8C,$7C,$A7,$F2,$82,$3F,$CB,$DA,$FF,$25,$6B,$C0
	.dc.b	$D1,$0D,$14,$FE,$49,$F2,$9F,$CD,$08,$FF,$3B,$6B,$FC,$C5,$AF,$03
	.dc.b	$44,$34,$53,$F9,$87,$CA,$7F,$40,$23,$FD,$1D,$AF,$F3,$D6,$BC,$0D
	.dc.b	$10,$D1,$4F,$F3,$57,$AB,$FE,$75,$23,$FE,$E6,$6B,$FE,$75,$6B,$C0
	.dc.b	$D1,$0D,$14,$FF,$3B,$7A,$BF,$E8,$12,$3F,$EF,$26,$BF,$E8,$16,$BC
	.dc.b	$0D,$10,$D1,$4F,$EA,$9F,$29,$FD,$90,$8F,$F6,$B6,$BF,$D8,$5A,$F0
	.dc.b	$34,$43,$45,$3F,$B0,$7C,$A7,$F7,$02,$3F,$DD,$DA,$FF,$6D,$6B,$C0
	.dc.b	$D1,$0D,$14,$FF,$3A,$7B,$8F,$73,$FC,$E6,$D8,$FF,$38,$B5,$FF,$37
	.dc.b	$B5,$E3,$1C,$08,$F0,$6B,$C1,$AF,$03,$FC,$DD,$EE,$38,$BF,$CD,$6D
	.dc.b	$8C,$C6,$08,$E0,$87,$F9,$B6,$38,$36,$36,$D3,$67,$FA,$4F,$DC,$7B
	.dc.b	$9F,$E9,$26,$C7,$FA,$41,$AF,$FA,$39,$AF,$18,$E0,$47,$83,$5E,$0D
	.dc.b	$78,$1F,$E2,$FF,$94,$FF,$15,$91,$C1,$0F,$F1,$43,$96,$99,$3F,$C4
	.dc.b	$8E,$5A,$7F,$8E,$FE,$53,$FC,$6E,$47,$04,$3F,$C6,$8E,$5A,$64,$FF
	.dc.b	$18,$39,$68,$21,$3D,$CF,$F4,$3B,$83,$1E,$27,$FB,$D2,$33,$19,$0E
	.dc.b	$1E,$0D,$78,$9B,$37,$1C,$C8,$E6,$38,$3E,$56,$BC,$0D,$7B,$DE,$32
	.dc.b	$38,$21,$E5,$84,$6A,$30,$D7,$CE,$D7,$E1,$3F,$E6,$6F,$94,$91,$FE
	.dc.b	$6D,$6B,$FE,$63,$6B,$FE,$5D,$6B,$FE,$5C,$6A,$F0,$BF,$CE,$91,$87
	.dc.b	$FC,$ED,$F2,$9F,$E7,$42,$4F,$F9,$ED,$AF,$F9,$C5,$AF,$F9,$AD,$AF
	.dc.b	$F9,$A9,$AF,$7F,$09,$1F,$E8,$0F,$94,$93,$FE,$87,$6B,$FE,$7D,$6B
	.dc.b	$FE,$77,$6B,$FE,$76,$6B,$D5,$B4,$F5,$3F,$73,$3F,$8C,$8C,$70,$7E
	.dc.b	$05,$B1,$F8,$96,$BF,$8F,$6B,$C6,$79,$BF,$97,$8C,$B6,$32,$FC,$D6
	.dc.b	$C6,$A3,$13,$E1,$9C,$8E,$FC,$12,$7C,$91,$96,$C6,$5F,$34,$65,$B1
	.dc.b	$97,$D0,$D7,$CC,$FF,$AC,$3E,$52,$47,$FA,$DD,$AF,$FA,$B5,$AF,$FA
	.dc.b	$9D,$AF,$FA,$99,$AF,$DD,$6D,$FF,$AF,$FE,$2F,$FD,$7A,$49,$FF,$62
	.dc.b	$35,$FF,$5D,$35,$FF,$5A,$35,$FF,$59,$B5,$FC,$53,$FC,$07,$E0,$E4
	.dc.b	$7E,$DD,$AF,$EB,$DA,$FF,$02,$6B,$FC,$35,$AF,$EF,$5A,$FF,$0A,$F8
	.dc.b	$3F,$E1,$C4,$7F,$82,$35,$FD,$E3,$5F,$E1,$ED,$7F,$8D,$35,$FE,$14
	.dc.b	$FF,$19,$F8,$39,$1F,$C3,$DA,$FF,$07,$6B,$FC,$69,$AF,$F2,$56,$BF
	.dc.b	$C5,$5B,$7F,$EE,$FF,$94,$FF,$76,$92,$7F,$DE,$8D,$7F,$DD,$4D,$7F
	.dc.b	$DC,$8D,$7F,$DC,$6D,$7F,$74,$FF,$BE,$3E,$52,$47,$FB,$FD,$AF,$FB
	.dc.b	$D5,$AF,$FB,$BD,$AF,$FB,$B9,$AF,$F0,$56,$CF,$EE,$7B,$FF,$4C,$49
	.dc.b	$FD,$1B,$5F,$D0,$B5,$FD,$B3,$5F,$DA,$35,$FD,$A3,$FB,$1E,$F9,$1F
	.dc.b	$AE,$6B,$FA,$C6,$BF,$C0,$5A,$FE,$FD,$AF,$EF,$DB,$2E,$2F,$F0,$E4
	.dc.b	$FF,$07,$6B,$FC,$8F,$B3,$F8,$94,$62,$78,$46,$09,$3F,$C5,$67,$44
	.dc.b	$8C,$C7,$F1,$78,$E5,$E8,$7F,$83,$23,$F8,$B3,$5F,$E8,$0D,$A3,$49
	.dc.b	$D7,$FC,$95,$B3,$FC,$95,$AF,$F2,$56,$BF,$CB,$57,$C8,$71,$24,$71
	.dc.b	$6B,$C5,$AF,$16,$BE,$86,$BE,$87,$FC,$01,$F0,$D2,$3F,$BB,$35,$FE
	.dc.b	$DC,$D7,$FC,$02,$D7,$FC,$18,$D7,$FB,$E3,$5F,$F0,$4F,$C3,$7F,$83
	.dc.b	$48,$FF,$80,$DA,$FF,$7B,$6B,$FE,$0E,$6B,$FE,$19,$6B,$FE,$09,$7E
	.dc.b	$67,$C8,$48,$F9,$8D,$7E,$63,$5F,$98,$D7,$ED,$B5,$FB,$6D,$C7,$ED
	.dc.b	$F6,$7D,$92,$4F,$D9,$6B,$F6,$5A,$FD,$96,$BF,$88,$6B,$F8,$87,$FB
	.dc.b	$47,$A5,$FE,$24,$93,$49,$DD,$A2,$57,$F6,$D8,$CB,$63,$2F,$F6,$F8
	.dc.b	$CB,$7B,$BB,$91,$E4,$D7,$46,$08,$60,$87,$F8,$06,$3F,$C6,$2E,$8C
	.dc.b	$73,$7B,$08,$F9,$DA,$F9,$5A,$F0,$34,$43,$45,$3F,$49,$F1,$FF,$AB
	.dc.b	$23,$FA,$B6,$BF,$AB,$6B,$C0,$D1,$0D,$14,$FD,$87,$C7,$FE,$DC,$8F
	.dc.b	$ED,$DA,$FE,$DD,$AF,$03,$44,$34,$53,$FC,$D3,$F1,$3F,$E6,$F2,$3F
	.dc.b	$E6,$56,$BF,$E5,$D6,$BC,$0D,$10,$D1,$4F,$F3,$AF,$B5,$FE,$7B,$23
	.dc.b	$FE,$71,$6B,$FE,$69,$6B,$C0,$D1,$0D,$14,$FE,$29,$F1,$FF,$C6,$C8
	.dc.b	$FF,$1B,$6B,$FC,$6D,$AF,$03,$44,$34,$53,$F9,$07,$C7,$FF,$27,$23
	.dc.b	$FC,$9D,$AF,$F2,$76,$BC,$0D,$10,$D1,$4F,$E5,$9F,$1D,$FC,$CC,$8F
	.dc.b	$F3,$36,$BF,$CC,$DA,$F0,$34,$43,$45,$3F,$49,$DB,$FB,$C2,$3F,$B7
	.dc.b	$6B,$FB,$56,$BC,$0D,$10,$D1,$4F,$D8,$76,$FF,$08,$23,$FC,$0D,$AF
	.dc.b	$F0,$16,$BC,$0D,$10,$D1,$4F,$EA,$1F,$1D,$FD,$5C,$8F,$F5,$76,$BF
	.dc.b	$D5,$DA,$F0,$34,$43,$45,$3F,$AE,$7C,$57,$F6,$32,$3F,$D8,$DA,$FF
	.dc.b	$63,$6B,$C0,$D1,$0D,$14,$FF,$6A,$FC,$6F,$FB,$6C,$8F,$FB,$45,$AF
	.dc.b	$FB,$25,$AF,$03,$44,$34,$53,$FD,$C3,$F1,$BF,$EE,$72,$3F,$ED,$D6
	.dc.b	$BF,$ED,$56,$BC,$0D,$10,$D1,$4F,$F7,$6F,$C6,$FF,$BC,$C8,$FF,$BA
	.dc.b	$5A,$FF,$B8,$5A,$F0,$34,$43,$45,$3F,$C1,$9F,$17,$FE,$13,$23,$FE
	.dc.b	$13,$6B,$FE,$13,$6B,$C0,$D1,$0D,$14,$FF,$0C,$7C,$4F,$F8,$7C,$8F
	.dc.b	$F8,$7D,$AF,$F8,$7D,$AF,$03,$44,$34,$53,$FA,$87,$5F,$F6,$82,$3F
	.dc.b	$D8,$DA,$FF,$61,$6B,$C0,$D1,$0D,$14,$FE,$B9,$D7,$FD,$D0,$8F,$F6
	.dc.b	$F6,$BF,$DB,$5A,$F0,$34,$43,$45,$3F,$C7,$9E,$F7,$F9,$0C,$8F,$F9
	.dc.b	$0D,$AF,$F9,$0D,$AF,$03,$44,$34,$53,$FC,$91,$EF,$7F,$93,$C8,$FF
	.dc.b	$93,$DA,$FF,$93,$DA,$F0,$34,$43,$45,$3F,$CA,$9E,$D7,$F9,$6C,$8F
	.dc.b	$F9,$6D,$AF,$F9,$6D,$AF,$03,$44,$34,$50,$41,$DA,$F8,$DC,$44,$8C
	.dc.b	$38,$38,$B8,$70,$D7,$84,$6D,$DB,$5E,$ED,$8E,$11,$E3,$6B,$C0,$E4
	.dc.b	$E1,$E2,$63,$E3,$79,$11,$C1,$0F,$2B,$96,$99,$3D,$CE,$5A,$7F,$A2
	.dc.b	$7D,$CF,$F4,$61,$1C,$10,$FF,$46,$39,$69,$93,$FD,$1C,$E5,$A7,$FA
	.dc.b	$57,$CF,$FF,$4C,$11,$C1,$0F,$F4,$C3,$96,$99,$3F,$D3,$4E,$5A,$7F
	.dc.b	$96,$7A,$BF,$CC,$24,$70,$43,$FC,$B4,$E5,$A6,$4F,$F2,$93,$96,$9F
	.dc.b	$E6,$5E,$AF,$F3,$69,$1C,$10,$FF,$33,$39,$69,$93,$FC,$BC,$E5,$A7
	.dc.b	$F9,$C7,$8F,$FC,$F2,$47,$04,$3F,$CE,$4E,$5A,$64,$FF,$35,$39,$69
	.dc.b	$FE,$C5,$F3,$BF,$D9,$04,$70,$43,$FD,$90,$E5,$A6,$4F,$F6,$53,$96
	.dc.b	$9F,$ED,$1F,$2B,$FD,$A8,$47,$04,$3F,$DA,$8E,$5A,$64,$FF,$6B,$39
	.dc.b	$69,$FA,$4F,$8D,$FD,$51,$1C,$10,$FD,$AB,$96,$99,$3F,$7C,$E5,$A7
	.dc.b	$EC,$3C,$7F,$DB,$11,$C1,$0F,$E0,$2E,$5A,$64,$FE,$14,$E5,$A7,$FB
	.dc.b	$D7,$CB,$FF,$7C,$11,$C1,$0F,$F7,$C3,$96,$99,$3F,$DF,$4E,$5A,$7F
	.dc.b	$17,$F1,$9F,$AC,$FF,$B0,$49,$E3,$FF,$15,$8C,$B6,$33,$3C,$5A,$FF
	.dc.b	$24,$9D,$B9,$9E,$04,$78,$78,$FC,$C8,$E0,$87,$3A,$FF,$5E,$B6,$36
	.dc.b	$D3,$67,$F3,$7A,$F5,$C6,$A3,$FD,$8D,$D9,$FE,$DE,$23,$FE,$D9,$6B
	.dc.b	$FE,$D7,$6B,$C0,$D1,$0D,$14,$FF,$69,$76,$7F,$BA,$88,$FF,$B9,$5A
	.dc.b	$FF,$B8,$DA,$F0,$34,$43,$45,$3D,$96,$F1,$BE,$91,$23,$E9,$35,$F9
	.dc.b	$0D,$78,$B5,$FE,$BB,$87,$E0,$23,$10,$8C,$3F,$77,$C1,$FC,$01,$1C
	.dc.b	$10,$FA,$CE,$5A,$64,$F4,$B9,$69,$FD,$63,$E3,$BF,$AF,$11,$C1,$0F
	.dc.b	$ED,$2E,$5A,$64,$FE,$E4,$E5,$A7,$F6,$DF,$8E,$78,$3F,$DB,$5B,$19
	.dc.b	$8C,$11,$C1,$0F,$EF,$91,$C1,$B1,$B6,$9B,$3F,$C0,$B0,$F0,$7C,$04
	.dc.b	$8F,$23,$5E,$AD,$78,$B5,$FF,$09,$47,$37,$89,$0F,$D3,$78,$7F,$B5
	.dc.b	$23,$FB,$56,$BA,$30,$43,$04,$3F,$1E,$F0,$7F,$23,$E0,$FF,$01,$23
	.dc.b	$FC,$05,$AF,$ED,$DA,$F0,$34,$43,$45,$3F,$71,$E4,$FF,$87,$33,$34
	.dc.b	$9D,$DF,$C4,$DB,$3F,$C6,$5A,$FF,$11,$6B,$D2,$38,$1C,$7B,$64,$71
	.dc.b	$6B,$C5,$AF,$91,$AF,$EA,$25,$18,$8F,$D4,$46,$1E,$87,$F2,$5F,$0B
	.dc.b	$F9,$41,$1C,$10,$FE,$3C,$E5,$A6,$4F,$E0,$2E,$5A,$7F,$31,$EC,$FE
	.dc.b	$68,$47,$04,$3F,$95,$39,$69,$93,$F8,$6B,$96,$9F,$CF,$7B,$3F,$A0
	.dc.b	$11,$C1,$0F,$E6,$CE,$5A,$64,$FE,$32,$E5,$A7,$F5,$1F,$4F,$F3,$8F
	.dc.b	$F4,$96,$BF,$CE,$DA,$FF,$18,$6B,$FC,$5D,$B9,$1A,$4D,$BF,$AD,$36
	.dc.b	$7C,$11,$83,$3E,$24,$8F,$23,$5F,$23,$5F,$23,$5F,$23,$5F,$23,$6B
	.dc.b	$FB,$6F,$77,$F6,$92,$4F,$F6,$96,$BF,$D7,$DA,$FF,$40,$6B,$FC,$E3
	.dc.b	$95,$18,$8F,$E7,$11,$87,$FB,$A7,$3F,$F7,$72,$38,$21,$FD,$B5,$CB
	.dc.b	$4C,$9F,$D4,$1C,$B4,$FD,$87,$67,$EC,$C8,$E0,$87,$EC,$DC,$B4,$C9
	.dc.b	$FC,$01,$CB,$4F,$DE,$76,$7E,$FC,$8E,$08,$7E,$FD,$CB,$4C,$9F,$C3
	.dc.b	$1C,$B4,$FE,$11,$CD,$FC,$2C,$8E,$08,$7F,$0B,$72,$D3,$27,$F1,$87
	.dc.b	$2D,$3F,$C4,$9D,$3F,$E2,$72,$38,$21,$FE,$1D,$72,$D3,$27,$F8,$01
	.dc.b	$CB,$4F,$F1,$87,$1F,$F8,$CC,$8E,$08,$7F,$8A,$5C,$B4,$C9,$FE,$0C
	.dc.b	$72,$D3,$FB,$46,$9F,$D9,$C8,$E0,$87,$F6,$47,$2D,$32,$7F,$62,$72
	.dc.b	$D3,$FB,$A7,$89,$FD,$CC,$8E,$08,$7F,$70,$72,$D3,$27,$F6,$E7,$2D
	.dc.b	$3F,$CA,$9C,$9F,$E5,$72,$38,$21,$FE,$4D,$72,$D3,$27,$F8,$C1,$CB
	.dc.b	$4F,$F3,$06,$5F,$E6,$32,$38,$21,$FE,$59,$72,$D3,$27,$F8,$F1,$CB
	.dc.b	$4F,$F3,$66,$5F,$E6,$F2,$38,$21,$FE,$65,$72,$D3,$27,$F9,$21,$CB
	.dc.b	$4F,$ED,$1F,$1D,$FD,$AC,$8E,$08,$7F,$6B,$72,$D3,$27,$F7,$87,$2D
	.dc.b	$3F,$BA,$7C,$77,$F7,$72,$38,$21,$FD,$DD,$CB,$4C,$9F,$E0,$47,$2D
	.dc.b	$3F,$D2,$1A,$7F,$A4,$C8,$E0,$87,$FA,$25,$CB,$4C,$9F,$E6,$C7,$2D
	.dc.b	$3F,$D3,$93,$FE,$9F,$23,$82,$1F,$E9,$57,$2D,$32,$7F,$9E,$1C,$B4
	.dc.b	$10,$BD,$CF,$8E,$72,$E6,$75,$39,$8C,$47,$09,$DB,$89,$D3,$63,$2F
	.dc.b	$38,$CB,$63,$2F,$66,$BD,$67,$BA,$9E,$2E,$E3,$31,$88,$F6,$FC,$73
	.dc.b	$E3,$B9,$C2,$31,$5D,$1B,$1E,$C6,$BF,$05,$AF,$C7,$6B,$DA,$7A,$C6
	.dc.b	$23,$B4,$60,$93,$DD,$F1,$D3,$E3,$9A,$7A,$9B,$7F,$2D,$AF,$C9,$6B
	.dc.b	$F5,$1A,$FD,$A6,$BF,$1D,$B3,$DE,$7B,$12,$7B,$35,$FF,$70,$69,$FE
	.dc.b	$EF,$23,$FE,$EF,$6B,$A3,$04,$30,$43,$FD,$92,$F0,$7F,$D9,$7F,$1D
	.dc.b	$FE,$FB,$23,$FE,$FB,$6B,$FE,$F1,$6B,$C0,$D1,$0D,$14,$FF,$3D,$BF
	.dc.b	$E7,$72,$3F,$E7,$76,$BF,$E7,$76,$BC,$0D,$10,$D1,$4F,$F4,$3F,$C7
	.dc.b	$7F,$A0,$C8,$FF,$A0,$DA,$FF,$A0,$DA,$F0,$34,$43,$45,$3F,$DD,$EF
	.dc.b	$E8,$3A,$1C,$E8,$C3,$67,$F6,$73,$A9,$C9,$23,$33,$B7,$F7,$93,$B7
	.dc.b	$33,$B7,$33,$B7,$C6,$76,$FD,$6B,$F6,$6F,$FB,$DD,$B4,$69,$23,$D4
	.dc.b	$D7,$91,$82,$18,$21,$E6,$5F,$1F,$C8,$91,$E3,$6B,$B6,$BF,$C7,$67
	.dc.b	$C9,$18,$8F,$E4,$31,$87,$91,$0F,$F7,$7F,$9D,$FE,$F7,$23,$FE,$FB
	.dc.b	$6B,$A3,$04,$30,$43,$FD,$E6,$F0,$7F,$DE,$BE,$77,$CA,$F1,$0D,$24
	.dc.b	$7F,$20,$8C,$B6,$32,$FF,$32,$8C,$B6,$32,$F6,$38,$1F,$67,$E3,$FE
	.dc.b	$99,$1C,$10,$FA,$0E,$5A,$64,$F9,$98,$A3,$51,$87,$EF,$FC,$7C,$8F
	.dc.b	$C4,$35,$FB,$8D,$79,$B5,$FA,$B2,$8C,$47,$E3,$DB,$8B,$DA,$FC,$89
	.dc.b	$27,$F5,$0D,$7F,$4A,$D7,$F4,$2D,$7F,$0C,$D7,$F0,$AF,$4C,$F9,$92
	.dc.b	$39,$B5,$E6,$D7,$9B,$5E,$6D,$53,$E9,$8F,$3E,$12,$7C,$ED,$7C,$ED
	.dc.b	$7C,$ED,$7C,$ED,$7A,$D7,$D5,$79,$9F,$B9,$F8,$FF,$D9,$91,$C1,$0F
	.dc.b	$D8,$39,$69,$93,$F0,$2E,$5A,$7F,$05,$F8,$FF,$DF,$91,$C1,$0F,$DE
	.dc.b	$39,$69,$93,$F1,$AE,$5A,$7F,$BD,$FB,$7F,$DD,$64,$70,$43,$FD,$F4
	.dc.b	$E5,$A6,$4F,$F7,$E3,$96,$9F,$CA,$BC,$46,$FF,$EF,$82,$46,$5A,$FF
	.dc.b	$2A,$8D,$B9,$21,$98,$DB,$FC,$4E,$36,$E4,$FE,$59,$F1,$FF,$C9,$88
	.dc.b	$E0,$87,$F2,$37,$2D,$32,$7E,$D1,$CB,$4F,$E7,$1F,$1F,$FC,$C8,$8E
	.dc.b	$08,$7F,$2F,$72,$D3,$27,$F0,$07,$2D,$3F,$A2,$7C,$7F,$F3,$E2,$38
	.dc.b	$21,$FC,$ED,$CB,$4C,$9F,$C3,$1C,$B4,$FD,$CE,$FF,$BC,$23,$82,$1F
	.dc.b	$BC,$72,$D3,$27,$EF,$9C,$B4,$FE,$0B,$9F,$F0,$82,$38,$21,$FC,$21
	.dc.b	$CB,$4C,$9F,$C2,$9C,$B4,$FE,$D1,$F1,$FF,$D8,$88,$E0,$87,$F5,$F7
	.dc.b	$2D,$32,$7F,$30,$72,$D3,$FB,$A7,$C7,$FF,$6E,$23,$82,$1F,$DA,$DC
	.dc.b	$B4,$C9,$FC,$F1,$CB,$4F,$EA,$5E,$BF,$F5,$72,$38,$21,$FD,$4D,$CB
	.dc.b	$4C,$9F,$E8,$BE,$54,$6A,$30,$FF,$63,$F5,$E4,$7F,$61,$6B,$FD,$7D
	.dc.b	$AF,$36,$BF,$E9,$B9,$46,$23,$FD,$3F,$18,$7A,$1F,$DB,$7D,$7F,$EE
	.dc.b	$44,$70,$43,$FB,$73,$96,$99,$3D,$8E,$5A,$7F,$79,$F4,$BF,$BF,$11
	.dc.b	$C1,$0F,$EF,$4E,$5A,$64,$F9,$0E,$5A,$64,$F8,$0E,$5A,$64,$F8,$2E
	.dc.b	$5A,$64,$4A,$35,$18,$6C,$66,$38,$35,$F3,$90,$F2,$94,$F1,$94,$F0
	.dc.b	$97,$9B,$FC,$91,$F1,$FF,$E4,$62,$47,$F9,$3D,$AF,$F9,$0D,$AF,$F8
	.dc.b	$49,$AF,$F8,$21,$AF,$F9,$55,$FF,$28,$FC,$64,$8F,$F2,$D3,$5F,$F2
	.dc.b	$73,$5F,$F0,$EB,$5F,$F0,$9B,$5F,$F3,$03,$67,$FC,$C1,$F1,$9F,$E5
	.dc.b	$E2,$4F,$F9,$9D,$AF,$F9,$6D,$AF,$F8,$A9,$AF,$F8,$81,$AF,$F9,$B5
	.dc.b	$FF,$34,$FC,$64,$8F,$F3,$93,$5F,$F3,$33,$5F,$F1,$AB,$5F,$F1,$5B
	.dc.b	$5F,$F3,$C3,$67,$FC,$F1,$F1,$9F,$E7,$62,$4F,$F9,$FD,$AF,$F9,$CD
	.dc.b	$AF,$F9,$09,$AF,$F8,$E1,$AF,$FA,$15,$FF,$40,$FC,$64,$8F,$F4,$53
	.dc.b	$5F,$F3,$F3,$5F,$F2,$6B,$5F,$F2,$1B,$5F,$F4,$83,$67,$FD,$21,$F1
	.dc.b	$9F,$E8,$E2,$4F,$FA,$5D,$AF,$FA,$2D,$AF,$F9,$69,$AF,$F9,$41,$AF
	.dc.b	$FA,$75,$FF,$4C,$FC,$64,$8F,$F5,$13,$5F,$F4,$B3,$5F,$F3,$2B,$5F
	.dc.b	$F2,$DB,$5F,$F5,$43,$67,$FD,$51,$F1,$9F,$EA,$62,$4F,$FA,$BD,$AF
	.dc.b	$FA,$8D,$AF,$F9,$C9,$AF,$F9,$A1,$AF,$FA,$D5,$FF,$58,$FC,$64,$8F
	.dc.b	$F5,$D3,$5F,$F5,$73,$5F,$F3,$EB,$5F,$F3,$9B,$5F,$F6,$03,$67,$FD
	.dc.b	$81,$F1,$9F,$EB,$E2,$4F,$FB,$1D,$AF,$FA,$ED,$AF,$FA,$29,$AF,$FA
	.dc.b	$01,$AF,$FB,$35,$FF,$71,$7C,$93,$E3,$69,$27,$FD,$C3,$19,$6C,$65
	.dc.b	$FF,$6F,$C6,$5B,$19,$7F,$D8,$D1,$96,$C6,$51,$3E,$15,$E9,$FF,$B8
	.dc.b	$9B,$1A,$20,$72,$E0,$C4,$71,$23,$93,$25,$8E,$2F,$46,$BD,$DA,$F9
	.dc.b	$63,$9B,$FC,$BC,$FE,$74,$7F,$38,$E4,$FE,$76,$47,$04,$3F,$D9,$9E
	.dc.b	$DF,$FB,$54,$8F,$FB,$4D,$AE,$8C,$10,$C1,$0F,$F0,$03,$C1,$F9,$FF
	.dc.b	$1B,$F7,$08,$FD,$A6,$BF,$61,$AF,$03,$44,$34,$53,$ED,$FC,$6F,$E2
	.dc.b	$08,$FE,$19,$AF,$E1,$1A,$F0,$34,$43,$45,$3F,$0F,$87,$E5,$08,$FE
	.dc.b	$49,$AF,$E4,$1A,$F0,$34,$43,$45,$3F,$27,$F1,$BF,$A0,$23,$F9,$E6
	.dc.b	$BF,$9C,$6B,$C0,$D1,$0D,$14,$FC,$FF,$C6,$FE,$B0,$8F,$EA,$9A,$FE
	.dc.b	$A1,$AF,$03,$44,$34,$53,$F5,$7F,$1B,$FB,$82,$3F,$B6,$6B,$FB,$46
	.dc.b	$BC,$0D,$10,$D1,$4F,$DB,$FC,$6F,$F0,$42,$3F,$C0,$9A,$FF,$00,$6B
	.dc.b	$C0,$D1,$0D,$14,$FE,$07,$F1,$BF,$C4,$08,$FF,$0E,$6B,$FC,$31,$AF
	.dc.b	$03,$44,$34,$53,$F8,$7F,$C6,$FF,$1C,$23,$FC,$69,$AF,$F1,$86,$BC
	.dc.b	$0D,$10,$D1,$4F,$E3,$7A,$FF,$28,$23,$FC,$99,$AF,$F2,$46,$BC,$0D
	.dc.b	$10,$D1,$4F,$E4,$FA,$FF,$34,$23,$FC,$C9,$AF,$F3,$06,$BC,$0D,$10
	.dc.b	$D1,$4F,$E6,$7E,$FF,$F4,$02,$3F,$CF,$9A,$FF,$3C,$6B,$C0,$D1,$0D
	.dc.b	$14,$FF,$00,$1F,$E0,$52,$1F,$E0,$52,$9A,$30,$43,$04,$3F,$C1,$07
	.dc.b	$03,$FB,$F9,$FE,$06,$21,$FE,$06,$29,$FE,$06,$29,$C0,$D1,$0D,$14
	.dc.b	$FE,$FC,$7F,$81,$48,$7F,$81,$4A,$7F,$81,$4A,$70,$34,$43,$45,$3F
	.dc.b	$AF,$F1,$B8,$9E,$5F,$F7,$22,$4F,$F7,$06,$BF,$DB,$9A,$FF,$6C,$6B
	.dc.b	$D2,$39,$BC,$48,$7F,$76,$F2,$FF,$BF,$91,$FE,$FA,$D7,$46,$08,$60
	.dc.b	$87,$F7,$C7,$83,$FD,$FB,$D1,$FF,$05,$91,$FF,$04,$B5,$FF,$03,$B5
	.dc.b	$E0,$68,$86,$8A,$7F,$8A,$0F,$F1,$69,$0F,$F1,$69,$4D,$18,$21,$82
	.dc.b	$1F,$E3,$03,$81,$FE,$25,$3F,$C5,$84,$3F,$C5,$85,$3F,$C5,$85,$38
	.dc.b	$1A,$21,$A2,$9F,$57,$97,$FC,$40,$B0,$FC,$14,$61,$B1,$FE,$2E,$21
	.dc.b	$FE,$2C,$29,$FE,$2A,$29,$D4,$FF,$1B,$F8,$9F,$E3,$C2,$38,$21,$FE
	.dc.b	$38,$72,$D3,$27,$F8,$E1,$CB,$4F,$F2,$1F,$89,$FE,$48,$23,$82,$1F
	.dc.b	$E4,$47,$2D,$32,$7F,$91,$1C,$B4,$FF,$27,$F8,$9F,$E5,$42,$38,$21
	.dc.b	$FE,$50,$72,$D3,$27,$F9,$41,$CB,$4F,$F2,$07,$47,$F8,$E9,$FF,$20
	.dc.b	$35,$FF,$20,$35,$FF,$20,$1C,$08,$7F,$93,$F9,$BF,$C8,$AF,$F9,$30
	.dc.b	$C4,$7F,$90,$5F,$F2,$37,$89,$FE,$FB,$7F,$DF,$24,$3F,$DF,$25,$3F
	.dc.b	$DF,$25,$3F,$C9,$A7,$2E,$0F,$F1,$F3,$FE,$63,$84,$62,$3F,$CD,$71
	.dc.b	$87,$FC,$90,$7F,$91,$8A,$71,$8E,$0A,$38,$3F,$E4,$D3,$F8,$67,$C6
	.dc.b	$FF,$A1,$88,$FF,$A1,$1A,$FF,$A0,$9A,$FF,$A0,$1A,$F2,$38,$10,$FF
	.dc.b	$44,$FC,$6F,$FA,$48,$8F,$FA,$41,$AE,$8C,$10,$C1,$0F,$F4,$7B,$C1
	.dc.b	$FF,$44,$68,$D9,$EC,$FF,$A0,$E3,$0D,$8C,$CF,$FA,$2E,$72,$D8,$CB
	.dc.b	$FE,$92,$8C,$B6,$32,$F8,$1A,$FF,$40,$6B,$FC,$05,$B1,$CA,$38,$4F
	.dc.b	$F9,$D1,$FE,$79,$D9,$5F,$E1,$1C,$48,$C5,$7F,$4B,$6C,$66,$7F,$CF
	.dc.b	$07,$22,$9C,$A3,$FC,$EA,$FF,$49,$EC,$AF,$66,$D7,$F5,$76,$D1,$A4
	.dc.b	$9F,$F3,$71,$FD,$4F,$6F,$F3,$AB,$4C,$9E,$A7,$2D,$33,$5F,$D9,$63
	.dc.b	$2D,$8F,$A5,$F1,$BF,$5C,$8E,$08,$7D,$37,$2D,$32,$7D,$57,$2D,$3E
	.dc.b	$F7,$C6,$FE,$0C,$8E,$08,$7D,$F7,$2D,$32,$7E,$05,$CB,$4F,$C5,$7C
	.dc.b	$6F,$E3,$C8,$E0,$87,$E2,$DC,$B4,$C9,$F8,$D7,$2D,$3F,$D2,$47,$F8
	.dc.b	$0F,$BF,$FD,$32,$D3,$27,$E2,$9C,$B4,$CD,$FF,$85,$63,$2D,$8F,$F4
	.dc.b	$91,$FE,$0F,$EF,$FF,$4C,$B4,$C9,$F9,$67,$2D,$33,$1F,$E2,$18,$CB
	.dc.b	$63,$FD,$24,$7F,$86,$FB,$FF,$D3,$2D,$32,$7E,$89,$CB,$4C,$C7,$F8
	.dc.b	$B6,$32,$D8,$FF,$79,$1F,$CB,$78,$BF,$9A,$3F,$CC,$88,$7F,$32,$29
	.dc.b	$FC,$C8,$E1,$D9,$FC,$F5,$A7,$43,$71,$D1,$E4,$43,$91,$A8,$E4,$D3
	.dc.b	$91,$B8,$E4,$F0,$31,$1C,$1A,$70,$34,$48,$E2,$48,$E1,$19,$6C,$66
	.dc.b	$39,$35,$FE,$B3,$1A,$23,$A6,$BF,$C8,$7E,$37,$F9,$51,$1F,$E5,$0D
	.dc.b	$74,$60,$86,$08,$7F,$29,$78,$3F,$CA,$FE,$37,$F9,$B1,$1F,$E6,$8D
	.dc.b	$7F,$99,$35,$E0,$68,$86,$8A,$7F,$37,$F8,$CF,$E8,$44,$7F,$A0,$35
	.dc.b	$FE,$7C,$D7,$81,$A2,$1A,$29,$FD,$0F,$E3,$3F,$A7,$11,$FE,$98,$D7
	.dc.b	$FA,$53,$5E,$06,$88,$68,$A7,$F4,$FF,$83,$FE,$B4,$47,$FA,$C3,$5F
	.dc.b	$EA,$CD,$78,$1A,$21,$A2,$9F,$D6,$FE,$33,$FB,$31,$1F,$EC,$8D,$7F
	.dc.b	$B1,$35,$E0,$68,$86,$8A,$7F,$67,$F8,$CF,$EE,$44,$7F,$B8,$35,$FE
	.dc.b	$DC,$D7,$81,$A2,$1A,$29,$FD,$CF,$E3,$3F,$BF,$11,$FE,$F8,$D7,$FB
	.dc.b	$D3,$5E,$06,$88,$68,$A7,$F7,$FF,$8C,$FF,$05,$11,$FF,$04,$35,$FF
	.dc.b	$03,$35,$E0,$68,$86,$8A,$7F,$82,$FE,$33,$FC,$2C,$47,$FC,$28,$D7
	.dc.b	$FC,$24,$D7,$81,$A2,$1A,$29,$FE,$17,$F8,$CF,$F1,$11,$1F,$F1,$03
	.dc.b	$5F,$F0,$F3,$5E,$06,$88,$68,$A7,$F8,$8F,$E1,$FF,$C5,$C4,$7F,$C5
	.dc.b	$8D,$7F,$C5,$4D,$78,$1A,$21,$A2,$9F,$E7,$A3,$FC,$EA,$53,$FC,$E6
	.dc.b	$60,$FF,$3E,$94,$FF,$3E,$9A,$21,$C4,$87,$F3,$53,$F9,$19,$FC,$84
	.dc.b	$86,$0F,$F3,$41,$FE,$72,$F0,$3F,$CD,$BC,$3F,$E6,$C8,$FF,$39,$1A
	.dc.b	$3F,$CE,$11,$A2,$3F,$E7,$63,$FC,$8D,$E0,$FF,$95,$48,$FF,$94,$DA
	.dc.b	$FF,$94,$5A,$FF,$94,$5A,$FF,$94,$5A,$F4,$04,$2F,$CA,$BB,$71,$12
	.dc.b	$30,$E0,$DC,$56,$96,$36,$ED,$AF,$38,$E0,$A3,$2F,$8D,$AF,$43,$94
	.dc.b	$70,$8C,$C7,$F9,$AF,$E3,$7F,$CF,$44,$7F,$CF,$0D,$7F,$CE,$CD,$7F
	.dc.b	$CE,$CD,$7F,$CE,$CD,$7A,$1F,$E8,$1F,$89,$FF,$43,$11,$C1,$0F,$F4
	.dc.b	$13,$96,$99,$3F,$D0,$8E,$5A,$7C,$BF,$99,$7E,$F8,$F2,$3E,$64,$61
	.dc.b	$B1,$99,$E0,$D7,$EB,$CE,$DC,$CF,$02,$38,$21,$C2,$79,$3F,$83,$9E
	.dc.b	$4F,$FA,$77,$E2,$FF,$D4,$C4,$7F,$D4,$8D,$74,$60,$86,$08,$7F,$A9
	.dc.b	$5E,$0F,$FA,$9F,$E2,$FF,$D6,$44,$7F,$D6,$0D,$7F,$D5,$CD,$78,$1A
	.dc.b	$21,$A2,$9F,$EC,$2F,$D7,$3F,$1F,$DA,$69,$23,$FD,$88,$D7,$FD,$84
	.dc.b	$D7,$FD,$80,$D7,$A1,$C0,$E3,$F1,$51,$E2,$D7,$8B,$5F,$23,$5F,$D9
	.dc.b	$CF,$96,$31,$08,$C3,$D0,$F2,$FC,$57,$A0,$8E,$08,$7A,$1C,$B4,$C9
	.dc.b	$F1,$5C,$B4,$FD,$C7,$C5,$B9,$3F,$78,$D8,$CC,$60,$8E,$08,$7F,$09
	.dc.b	$8E,$0D,$8D,$B4,$D9,$F5,$FE,$17,$E9,$91,$FA,$6D,$7E,$9B,$5E,$06
	.dc.b	$88,$68,$A7,$F8,$20,$FF,$03,$10,$C1,$0F,$F7,$C7,$57,$F8,$3D,$A7
	.dc.b	$F8,$30,$A7,$12,$9F,$E0,$83,$FC,$0C,$43,$04,$3E,$FF,$98,$EB,$FF
	.dc.b	$08,$46,$1A,$FE,$06,$34,$FF,$84,$0D,$1F,$E0,$63,$9F,$5F,$F8,$35
	.dc.b	$A6,$4E,$4E,$5A,$67,$34,$A7,$51,$87,$F4,$BF,$0B,$2B,$F4,$AD,$7F
	.dc.b	$4A,$D7,$F6,$0D,$7F,$3B,$B2,$31,$1F,$9D,$8C,$3D,$0F,$D8,$FC,$57
	.dc.b	$ED,$08,$E0,$87,$ED,$1C,$B4,$C9,$FC,$05,$CB,$4F,$DE,$FC,$4F,$F0
	.dc.b	$02,$38,$21,$FC,$01,$CB,$4C,$9F,$C3,$5C,$B4,$FF,$16,$9F,$7B,$BB
	.dc.b	$FC,$68,$D3,$27,$DD,$72,$D3,$38,$FD,$48,$CB,$63,$FC,$5A,$7E,$2B
	.dc.b	$BB,$FC,$68,$D3,$27,$E2,$5C,$B4,$CC,$7E,$02,$32,$D8,$FF,$16,$9F
	.dc.b	$96,$DF,$FC,$68,$D3,$27,$E5,$5C,$B4,$CC,$7E,$32,$32,$D8,$FE,$5B
	.dc.b	$F1,$3F,$CC,$08,$E0,$87,$F3,$07,$2D,$32,$7F,$3D,$72,$D3,$F9,$CF
	.dc.b	$C4,$FF,$3C,$23,$82,$1F,$CF,$1C,$B4,$C9,$FD,$25,$CB,$4F,$E8,$BF
	.dc.b	$13,$FD,$20,$8E,$08,$7F,$48,$72,$D3,$27,$F5,$57,$2D,$3F,$CA,$C3
	.dc.b	$CF,$78,$1F,$E2,$86,$C6,$9A,$FF,$96,$8F,$F2,$F9,$4F,$F2,$F9,$FE
	.dc.b	$5C,$29,$A3,$FC,$7F,$E1,$7F,$C0,$16,$F9,$3F,$EF,$E6,$9F,$E6,$02
	.dc.b	$9F,$E4,$38,$47,$F9,$6A,$3F,$CB,$D1,$FE,$61,$7F,$C9,$CD,$7F,$CC
	.dc.b	$C7,$F9,$7C,$FF,$31,$95,$D3,$83,$FC,$DB,$97,$F7,$9F,$89,$FE,$EE
	.dc.b	$48,$FE,$EE,$D7,$FB,$BB,$5F,$F0,$13,$5F,$E2,$6D,$7F,$89,$3F,$E2
	.dc.b	$EF,$7A,$47,$F8,$B9,$AF,$F8,$9D,$AF,$F8,$9F,$BD,$18,$8F,$F3,$6F
	.dc.b	$7F,$F8,$F2,$31,$08,$C3,$D4,$FF,$07,$FC,$4F,$F8,$48,$8E,$08,$7F
	.dc.b	$84,$9C,$B4,$C9,$FE,$1F,$72,$D3,$FD,$10,$43,$FD,$10,$53,$FD,$10
	.dc.b	$53,$FD,$10,$53,$FD,$10,$70,$21,$C0,$FF,$3E,$1F,$E8,$1F,$17,$FC
	.dc.b	$EC,$47,$04,$3F,$CF,$65,$38,$98,$21,$83,$FC,$F2,$7F,$9F,$08,$60
	.dc.b	$87,$F9,$F0,$C9,$4C,$9F,$E7,$E3,$25,$3F,$C7,$BF,$13,$FE,$42,$23
	.dc.b	$FE,$42,$6B,$A3,$04,$30,$43,$FC,$A0,$F0,$7F,$CA,$5F,$13,$FE,$4E
	.dc.b	$23,$FE,$4E,$6B,$FE,$4E,$6B,$C0,$D1,$0D,$14,$FF,$2F,$7C,$4F,$F9
	.dc.b	$68,$8F,$F9,$69,$AF,$F9,$69,$AF,$03,$44,$34,$53,$FD,$2E,$7F,$A7
	.dc.b	$08,$60,$87,$F8,$73,$C6,$FF,$4F,$34,$FF,$51,$14,$E2,$53,$F5,$5E
	.dc.b	$37,$FA,$59,$FF,$4A,$35,$C9,$FE,$98,$29,$A3,$07,$EC,$34,$FF,$4D
	.dc.b	$B6,$39,$3F,$E3,$0F,$1B,$FD,$22,$D3,$FD,$2E,$7F,$A5,$4A,$64,$FF
	.dc.b	$4C,$94,$D1,$83,$FC,$6C,$FF,$27,$7F,$8F,$F0,$D7,$FD,$DF,$C3,$FC
	.dc.b	$B5,$C9,$FC,$AC,$A7,$F2,$B8,$E0,$A3,$83,$CA,$3A,$70,$FF,$A6,$FE
	.dc.b	$27,$FD,$32,$49,$FF,$4C,$B5,$FF,$4C,$B5,$FF,$52,$35,$FF,$1A,$B5
	.dc.b	$FF,$1A,$3F,$EA,$4F,$89,$91,$FE,$A4,$6B,$FE,$A4,$6B,$FE,$AF,$6B
	.dc.b	$FE,$40,$6B,$FE,$3F,$6B,$ED,$AF,$7D,$FB,$A3,$07,$C5,$3E,$01,$4C
	.dc.b	$9F,$1C,$C3,$C4,$91,$D0,$91,$A6,$BC,$48,$7F,$5C,$21,$F5,$0A,$78
	.dc.b	$8D,$9F,$E5,$CF,$89,$FF,$64,$91,$FF,$64,$B5,$FF,$64,$B5,$FF,$6A
	.dc.b	$35,$E4,$70,$20,$21,$FE,$C3,$F8,$A7,$AB,$88,$CE,$49,$CC,$64,$E1
	.dc.b	$35,$A6,$67,$84,$65,$B1,$99,$E6,$F7,$8C,$B6,$32,$F8,$9A,$FF,$B7
	.dc.b	$1A,$F8,$0E,$1E,$AF,$78,$34,$91,$E1,$6B,$C0,$C1,$0C,$10,$F1,$3F
	.dc.b	$EF,$CF,$8A,$91,$FE,$FC,$6B,$FE,$FC,$6B,$F2,$BC,$34,$62,$3F,$D3
	.dc.b	$2D,$7F,$D3,$0D,$B4,$BE,$2B,$E5,$36,$BE,$73,$5F,$94,$D7,$C1,$1C
	.dc.b	$08,$F0,$6B,$C1,$AF,$03,$AF,$C5,$78,$08,$E0,$87,$81,$CB,$4C,$9F
	.dc.b	$05,$CB,$4F,$4F,$C5,$7B,$08,$E0,$87,$B1,$CB,$4C,$9F,$35,$CB,$4F
	.dc.b	$CA,$7C,$53,$97,$F2,$D0,$8C,$3A,$8C,$C6,$23,$F3,$CD,$7F,$48,$D7
	.dc.b	$F5,$2D,$7F,$58,$D7,$B4,$74,$23,$DC,$FD,$37,$CC,$BE,$1F,$F4,$CD
	.dc.b	$8C,$C6,$08,$E0,$87,$EA,$63,$83,$63,$6D,$36,$7E,$F2,$1F,$05,$E0
	.dc.b	$24,$79,$1A,$F5,$6B,$C5,$AF,$F0,$BD,$13,$D2,$39,$10,$F3,$7C,$17
	.dc.b	$B8,$8F,$C2,$6B,$A3,$04,$30,$43,$D4,$F0,$7D,$9E,$17,$CA,$23,$F3
	.dc.b	$9A,$FC,$76,$BC,$0D,$10,$D1,$4F,$F4,$F1,$FE,$A9,$F0,$FF,$D5,$04
	.dc.b	$7F,$D5,$0D,$7F,$D4,$C6,$CC,$9F,$EB,$2E,$6F,$E9,$6F,$F4,$92,$1F
	.dc.b	$D0,$8F,$F4,$D9,$FE,$A0,$F6,$3F,$D4,$4F,$FA,$B1,$AF,$FA,$89,$AF
	.dc.b	$03,$EF,$7B,$FF,$8C,$23,$F9,$06,$BF,$89,$6B,$C0,$D1,$0D,$14,$FE
	.dc.b	$93,$EC,$BD,$16,$5F,$D0,$E3,$0D,$8C,$CF,$F4,$99,$CB,$63,$2F,$F5
	.dc.b	$18,$CB,$63,$2F,$81,$AF,$F5,$B8,$DB,$99,$F0,$11,$FD,$7F,$47,$EB
	.dc.b	$88,$E0,$87,$EA,$DC,$B4,$C9,$FA,$97,$2D,$3F,$77,$CF,$FB,$A2,$38
	.dc.b	$21,$FB,$77,$2D,$32,$7E,$D5,$CB,$4F,$EE,$1F,$21,$FE,$F9,$F4,$FF
	.dc.b	$B8,$36,$33,$B2,$72,$D8,$CC,$60,$8E,$08,$7F,$80,$67,$A3,$67,$93
	.dc.b	$67,$F8,$BF,$91,$23,$F8,$DB,$5F,$E2,$8D,$7F,$88,$35,$FE,$11,$C1
	.dc.b	$FE,$DE,$78,$90,$E2,$7F,$26,$E5,$FE,$4A,$47,$04,$3F,$91,$39,$69
	.dc.b	$93,$F9,$03,$96,$9F,$CC,$BE,$33,$F9,$89,$1C,$10,$FE,$5C,$E5,$A6
	.dc.b	$4F,$E5,$8E,$5A,$7F,$3E,$EE,$FE,$7A,$47,$04,$3F,$9D,$39,$69,$93
	.dc.b	$F9,$C3,$96,$9F,$D2,$B8,$7F,$A4,$91,$C1,$0F,$E8,$CE,$5A,$64,$FE
	.dc.b	$88,$E5,$A7,$F8,$F7,$E5,$DF,$0D,$FD,$A9,$B1,$98,$C1,$1C,$10,$FF
	.dc.b	$24,$47,$06,$C6,$DA,$6C,$FE,$AD,$F1,$BF,$D9,$08,$FF,$68,$6B,$FD
	.dc.b	$85,$AF,$03,$44,$34,$53,$FC,$0C,$7F,$82,$C8,$60,$87,$F5,$DE,$E7
	.dc.b	$67,$F8,$3A,$30,$D5,$D3,$FE,$38,$8C,$36,$3F,$C0,$07,$F8,$33,$B3
	.dc.b	$FC,$06,$D3,$27,$F8,$39,$CB,$4C,$9F,$E1,$07,$2D,$3F,$BE,$9F,$AC
	.dc.b	$EC,$FF,$00,$BF,$DF,$DA,$E4,$D9,$E8,$DF,$FC,$12,$EC,$CD,$7F,$7C
	.dc.b	$9E,$31,$BF,$8D,$FF,$0B,$91,$FF,$0D,$B5,$FF,$0A,$35,$FF,$08,$35
	.dc.b	$FE,$02,$D7,$A1,$FE,$24,$F8,$DF,$F1,$19,$1C,$10,$FF,$10,$39,$69
	.dc.b	$93,$FC,$3C,$E5,$A7,$F8,$C3,$AB,$FC,$5E,$47,$04,$3F,$C5,$8E,$5A
	.dc.b	$64,$FF,$15,$39,$69,$FE,$1F,$3F,$C4,$C5,$3F,$8F,$F7,$7F,$8A,$1F
	.dc.b	$A0,$DC,$3F,$C5,$2D,$C7,$FC,$55,$3F,$E3,$13,$91,$F5,$FA,$BF,$C4
	.dc.b	$0F,$F2,$A6,$B9,$3F,$C5,$07,$F8,$98,$A6,$CA,$6C,$FF,$16,$1F,$CD
	.dc.b	$39,$FF,$C4,$AD,$3F,$C4,$45,$3F,$C4,$A4,$3F,$C5,$45,$34,$60,$87
	.dc.b	$F4,$0F,$8D,$FF,$2F,$91,$FF,$31,$B5,$FF,$2E,$35,$FF,$2C,$35,$E4
	.dc.b	$70,$21,$FE,$63,$F2,$3F,$CD,$64,$7F,$CD,$ED,$74,$60,$86,$08,$7F
	.dc.b	$99,$DE,$0F,$F9,$A7,$E3,$7F,$CE,$E4,$7F,$CF,$6D,$7F,$CE,$8D,$78
	.dc.b	$1A,$21,$A2,$9F,$E4,$EA,$78,$3F,$E4,$B8,$C3,$5F,$F7,$CC,$69,$FF
	.dc.b	$26,$9F,$E5,$07,$0F,$F9,$4B,$C1,$FF,$27,$BF,$E4,$D0,$E1,$FF,$80
	.dc.b	$0D,$44,$5D,$1F,$EA,$A6,$C7,$FC,$01,$5D,$1F,$E5,$28,$EA,$D8,$CC
	.dc.b	$6E,$30,$73,$3F,$D4,$3F,$1B,$23,$FD,$4A,$D7,$FD,$3C,$D7,$FD,$34
	.dc.b	$D7,$FC,$16,$D7,$F8,$E3,$7D,$17,$C1,$7F,$B0,$9B,$3E,$E2,$07,$2E
	.dc.b	$0C,$47,$12,$39,$32,$58,$E2,$F4,$6B,$DD,$AF,$CE,$6B,$C8,$FF,$62
	.dc.b	$FC,$6F,$FB,$10,$8E,$08,$7F,$B0,$5C,$B4,$C9,$FE,$BF,$72,$D3,$FC
	.dc.b	$F4,$7F,$8C,$BC,$5F,$F4,$0B,$4C,$9F,$D1,$5C,$B4,$CF,$1F,$FB,$02
	.dc.b	$32,$D8,$FF,$3E,$70,$FF,$9C,$5F,$F3,$D3,$5F,$F3,$B1,$A2,$1A,$29
	.dc.b	$A3,$04,$3F,$CE,$07,$F9,$D4,$86,$08,$7F,$93,$BC,$4F,$F3,$BB,$4F
	.dc.b	$F3,$D9,$4E,$25,$3F,$09,$2F,$A4,$FF,$0C,$46,$1A,$FF,$B8,$5A,$FF
	.dc.b	$80,$9B,$1C,$7B,$BF,$0C,$D9,$FC,$63,$5F,$C8,$B5,$ED,$1C,$27,$99
	.dc.b	$1C,$7C,$74,$9E,$72,$8C,$47,$48,$C3,$A8,$CB,$EB,$6B,$EB,$6B,$EB
	.dc.b	$7C,$5F,$1D,$23,$BB,$5E,$CD,$7E,$33,$5F,$8C,$D7,$E3,$36,$DC,$EF
	.dc.b	$1F,$FD,$43,$5F,$E8,$66,$BF,$E1,$76,$CF,$06,$CF,$19,$FF,$55,$9C
	.dc.b	$4A,$3C,$7F,$F4,$EB,$FC,$05,$AF,$F2,$06,$BB,$E5,$FF,$50,$BB,$6B
	.dc.b	$B6,$BB,$F1,$FF,$D5,$AF,$FA,$9C,$FF,$56,$9C,$5C,$BF,$EA,$43,$FD
	.dc.b	$40,$53,$FD,$42,$43,$04,$3F,$D5,$3E,$3F,$FA,$AD,$FF,$1D,$37,$1F
	.dc.b	$96,$FF,$37,$E7,$93,$FC,$DD,$AF,$F4,$3E,$7F,$E8,$0F,$F3,$F8,$CB
	.dc.b	$FD,$11,$AF,$F2,$78,$DB,$FD,$2A,$32,$F1,$3B,$10,$EC,$E9,$E0,$64
	.dc.b	$EC,$53,$B3,$B8,$EE,$47,$04,$8E,$07,$84,$EE,$60,$EE,$60,$87,$80
	.dc.b	$A7,$80,$E1,$1D,$DB,$1D,$DD,$9D,$CE,$44,$39,$1A,$3C,$27,$F0,$9F
	.dc.b	$8E,$FD,$E1,$1F,$DC,$B5,$FD,$C3,$5E,$06,$88,$68,$A7,$E6,$F6,$72
	.dc.b	$7F,$55,$D3,$F8,$23,$63,$33,$FD,$76,$72,$D8,$CB,$FD,$92,$32,$D8
	.dc.b	$CB,$DE,$7A,$11,$E4,$D7,$F8,$97,$C7,$7F,$1B,$23,$FC,$61,$AE,$8C
	.dc.b	$10,$C1,$0F,$E6,$4F,$07,$F9,$A7,$C7,$7F,$27,$23,$FC,$91,$AF,$F2
	.dc.b	$36,$BC,$0D,$10,$D1,$4F,$E8,$1F,$1D,$FC,$CC,$8F,$F3,$06,$BF,$CB
	.dc.b	$DA,$F0,$34,$43,$45,$3F,$A6,$7C,$77,$F3,$F2,$3F,$CF,$1A,$FF,$3B
	.dc.b	$6B,$C0,$D1,$0D,$14,$FE,$B1,$F1,$DF,$D2,$C8,$FF,$48,$6B,$FD,$1D
	.dc.b	$AF,$03,$44,$34,$53,$F9,$57,$E4,$4D,$3F,$20,$D8,$FC,$73,$5F,$C6
	.dc.b	$35,$E3,$1C,$08,$F0,$6B,$C1,$AF,$03,$FC,$6D,$CE,$97,$C2,$FF,$8D
	.dc.b	$63,$0D,$8C,$CF,$06,$BF,$E4,$29,$DB,$99,$E0,$47,$04,$38,$4F,$27
	.dc.b	$FC,$99,$3C,$9F,$EE,$5F,$1F,$FD,$EC,$8F,$F7,$86,$BA,$30,$43,$04
	.dc.b	$3F,$C2,$4F,$07,$FC,$29,$F1,$5F,$E0,$72,$3F,$E0,$46,$BF,$E0,$36
	.dc.b	$BC,$0D,$10,$D1,$4F,$F1,$07,$2F,$F8,$4C,$8F,$F8,$41,$AF,$F8,$3D
	.dc.b	$AF,$03,$44,$34,$53,$FD,$3A,$7F,$A6,$8A,$7F,$A6,$FC,$2F,$F5,$03
	.dc.b	$FE,$94,$E7,$FF,$4E,$BA,$3F,$98,$1F,$CC,$CF,$E6,$26,$0F,$E6,$26
	.dc.b	$08,$6C,$86,$8B,$CB,$FE,$9E,$7F,$D0,$4F,$FA,$69,$BE,$17,$FA,$69
	.dc.b	$FF,$4E,$10,$C9,$4F,$F5,$6B,$FE,$9F,$70,$7F,$92,$7E,$2B,$FC,$7C
	.dc.b	$47,$FC,$76,$D7,$FC,$74,$D7,$FC,$A4,$D7,$91,$C0,$87,$FA,$73,$D3
	.dc.b	$79,$2F,$2B,$FD,$31,$18,$6C,$66,$7F,$D4,$13,$96,$C6,$5E,$8D,$7F
	.dc.b	$D5,$51,$B7,$33,$D0,$8F,$26,$BC,$A7,$C2,$47,$BB,$5D,$18,$21,$82
	.dc.b	$1E,$27,$DE,$E8,$FF,$3C,$7C,$47,$F9,$AC,$8F,$F9,$A1,$AF,$F9,$9D
	.dc.b	$AF,$03,$44,$34,$53,$F9,$0F,$81,$FC,$80,$8F,$F2,$06,$BF,$C8,$1A
	.dc.b	$F0,$34,$43,$45,$3F,$85,$FE,$3E,$F1,$7F,$8E,$36,$3F,$90,$B5,$FE
	.dc.b	$34,$D7,$8C,$70,$23,$C1,$AF,$06,$BC,$0F,$F4,$6F,$C4,$FF,$A4,$C8
	.dc.b	$E0,$87,$FA,$3D,$CB,$4C,$9F,$EA,$97,$2D,$3F,$D3,$3F,$13,$FE,$9F
	.dc.b	$23,$82,$1F,$E9,$B7,$2D,$32,$7F,$AD,$5C,$B4,$FF,$52,$F0,$7F,$AA
	.dc.b	$C8,$E0,$87,$FA,$9D,$CB,$4C,$9F,$EC,$17,$2D,$3F,$D6,$3F,$21,$FE
	.dc.b	$B7,$23,$82,$1F,$EB,$37,$2D,$32,$7F,$B3,$5C,$B4,$FD,$8E,$FF,$B6
	.dc.b	$23,$82,$1F,$B4,$72,$D3,$27,$E9,$5C,$B4,$FD,$EF,$AB,$FC,$08,$8E
	.dc.b	$08,$7F,$00,$72,$D3,$27,$EC,$5C,$B4,$FF,$03,$1F,$EE,$7E,$AF,$F0
	.dc.b	$6B,$4C,$9F,$EE,$97,$2D,$33,$E2,$7F,$BC,$E3,$2D,$8F,$F0,$31,$0F
	.dc.b	$F0,$31,$4F,$F0,$31,$82,$18,$21,$FE,$C1,$E9,$FF,$0A,$37,$AF,$FC
	.dc.b	$1E,$87,$57,$F7,$D7,$FB,$FC,$FF,$80,$5D,$46,$5B,$19,$78,$B6,$35
	.dc.b	$18,$EB,$FF,$07,$3F,$E0,$C6,$BD,$9A,$F8,$ED,$E3,$BD,$3F,$34,$FF
	.dc.b	$C0,$16,$B8,$9D,$C6,$23,$94,$F1,$71,$3A,$6C,$65,$E8,$62,$79,$29
	.dc.b	$E0,$E8,$87,$86,$79,$37,$E0,$38,$CF,$A9,$B2,$69,$27,$C6,$D7,$DB
	.dc.b	$3C,$5C,$90,$CC,$F3,$70,$78,$FC,$DE,$E4,$7D,$ED,$78,$B5,$E0,$68
	.dc.b	$86,$8A,$7F,$8D,$8F,$F1,$D1,$4F,$F1,$D1,$F4,$3B,$FF,$C7,$4D,$36
	.dc.b	$53,$67,$CD,$94,$70,$8C,$77,$FF,$8F,$1F,$B6,$D7,$27,$F8,$EC,$C9
	.dc.b	$4C,$9F,$E4,$02,$1F,$E4,$53,$FC,$6E,$43,$FC,$78,$53,$F1,$3D,$FF
	.dc.b	$E4,$06,$9F,$E3,$D2,$9B,$3D,$4E,$DF,$C4,$7B,$FF,$8A,$69,$F8,$47
	.dc.b	$18,$9A,$48,$D1,$0F,$CC,$B9,$69,$93,$F0,$EB,$E4,$3B,$12,$38,$B5
	.dc.b	$DB,$5F,$D5,$47,$17,$64,$36,$53,$7F,$21,$EB,$23,$E6,$6B,$E4,$6B
	.dc.b	$C0,$D1,$0D,$14,$FF,$31,$1F,$C0,$3C,$2F,$F3,$2B,$79,$FF,$CD,$2A
	.dc.b	$32,$FF,$9A,$E3,$31,$88,$D4,$60,$FA,$9E,$17,$F9,$8E,$3F,$CD,$8D
	.dc.b	$7F,$CD,$67,$F9,$A4,$8E,$5F,$F3,$5F,$95,$C8,$8F,$F0,$88,$D2,$7F
	.dc.b	$CD,$CF,$F9,$A0,$FF,$33,$9F,$CC,$5C,$BF,$C7,$9A,$E7,$A3,$FC,$E8
	.dc.b	$FF,$9C,$88,$73,$3F,$03,$F1,$FF,$8D,$23,$F8,$86,$BA,$30,$43,$04
	.dc.b	$3F,$06,$F0,$7F,$98,$FA,$2F,$DF,$1E,$3F,$F2,$68,$C3,$63,$33,$FC
	.dc.b	$CE,$72,$D8,$CB,$FC,$EA,$32,$D8,$CB,$E0,$6B,$FD,$5E,$36,$E6,$7C
	.dc.b	$04,$7B,$75,$F8,$C8,$E0,$87,$77,$2D,$32,$78,$5C,$B4,$FF,$49,$9F
	.dc.b	$E9,$82,$9F,$E9,$83,$FD,$3A,$60,$FE,$3D,$E3,$FF,$A7,$5A,$67,$87
	.dc.b	$FA,$FC,$67,$C7,$FF,$48,$4F,$FA,$07,$C7,$FF,$46,$B4,$FF,$48,$10
	.dc.b	$FF,$4A,$94,$D1,$82,$18,$3F,$D1,$9E,$3F,$FA,$4D,$A7,$FA,$34,$FF
	.dc.b	$02,$4F,$FA,$BA,$35,$18,$7F,$D2,$6F,$FA,$58,$C3,$A5,$1C,$5E,$87
	.dc.b	$F2,$4F,$90,$FE,$44,$47,$F8,$EB,$5F,$E3,$6D,$7F,$8A,$B5,$FF,$03
	.dc.b	$72,$23,$11,$D4,$FE,$4F,$F2,$1F,$CB,$C8,$E0,$87,$F2,$87,$2D,$32
	.dc.b	$7F,$24,$72,$D3,$F9,$9F,$77,$F3,$B2,$38,$21,$FC,$D1,$CB,$4C,$9F
	.dc.b	$CC,$1C,$B4,$FF,$59,$1F,$E1,$AF,$27,$FD,$72,$D3,$27,$F6,$97,$2D
	.dc.b	$33,$A7,$F8,$AA,$32,$D8,$FF,$59,$1F,$E4,$5F,$27,$FD,$72,$D3,$23
	.dc.b	$CF,$78,$69,$DC,$E6,$30,$F0,$9D,$CE,$9B,$1F,$EB,$52,$1F,$EB,$52
	.dc.b	$9F,$EB,$53,$04,$30,$43,$FC,$A7,$E5,$7F,$B0,$DB,$E5,$7F,$AF,$9F
	.dc.b	$E7,$3D,$1F,$CC,$89,$D3,$FC,$DA,$3F,$C8,$D1,$87,$F9,$C4,$65,$FE
	.dc.b	$86,$D7,$9B,$5F,$E8,$27,$02,$1D,$1D,$3F,$D2,$1B,$1D,$E3,$47,$03
	.dc.b	$B9,$4E,$E5,$3B,$9C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08
	.dc.b	$60,$87,$AC,$F2,$14,$EE,$53,$A1,$4E,$65,$39,$BF,$0D,$DB,$5D,$B5
	.dc.b	$DC,$7C,$A2,$3E,$28,$D3,$D4,$8E,$09,$1C,$23,$CF,$18,$79,$46,$5F
	.dc.b	$1B,$5F,$1B,$5F,$19,$C0,$87,$47,$4F,$89,$B1,$DE,$34,$70,$3B,$94
	.dc.b	$EE,$53,$B9,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08
	.dc.b	$7A,$CF,$21,$4E,$E5,$3A,$1F,$E2,$18,$FF,$30,$7C,$84,$8F,$F2,$DB
	.dc.b	$5F,$F2,$CB,$5F,$F2,$9B,$5F,$E7,$ED,$7F,$CD,$AD,$CF,$FC,$D5,$F2
	.dc.b	$1F,$E6,$D2,$4F,$F9,$A1,$AF,$F9,$99,$AF,$F9,$81,$AF,$F4,$C6,$BF
	.dc.b	$E7,$A7,$FC,$F1,$F2,$12,$3F,$CE,$6D,$7F,$CE,$2D,$7F,$CD,$6D,$7F
	.dc.b	$AB,$B5,$FF,$42,$B7,$8D,$F3,$AF,$7B,$FD,$A3,$18,$6C,$66,$BF,$DB
	.dc.b	$D3,$96,$C6,$5F,$F7,$2C,$65,$B1,$97,$FD,$D7,$19,$6C,$65,$FF,$00
	.dc.b	$B5,$FE,$F6,$D8,$E1,$03,$FA,$37,$CE,$FF,$9C,$DF,$F7,$96,$E8,$C4
	.dc.b	$7F,$BE,$23,$1E,$0F,$FB,$B9,$B1,$C2,$76,$FF,$C0,$08,$CB,$C6,$32
	.dc.b	$F8,$63,$2E,$E3,$2F,$92,$7B,$46,$27,$73,$FE,$0A,$F9,$2F,$F5,$A1
	.dc.b	$1F,$F5,$7B,$5F,$F5,$6B,$5F,$F5,$3B,$5E,$47,$02,$1F,$EB,$4F,$92
	.dc.b	$FF,$60,$11,$FF,$5D,$B5,$D1,$82,$18,$21,$FE,$B6,$78,$3F,$EB,$8F
	.dc.b	$91,$FF,$66,$11,$FF,$63,$B5,$FF,$62,$B5,$E0,$68,$86,$8A,$7F,$B1
	.dc.b	$3E,$47,$FD,$B0,$47,$FD,$A6,$D7,$FD,$A2,$D7,$81,$A2,$1A,$29,$F8
	.dc.b	$93,$F1,$5F,$12,$F3,$BF,$DD,$0D,$8D,$1B,$70,$E0,$38,$76,$D7,$84
	.dc.b	$74,$23,$C4,$E2,$E1,$E9,$1C,$5F,$2C,$7E,$5F,$F7,$47,$AF,$49,$3F
	.dc.b	$A0,$8C,$B6,$32,$FE,$86,$32,$D8,$CB,$F9,$F8,$CB,$63,$2F,$F9,$83
	.dc.b	$E4,$BD,$1C,$8F,$01,$18,$6C,$FC,$59,$D4,$E4,$91,$99,$DB,$F3,$A7
	.dc.b	$6E,$67,$6E,$67,$6F,$D9,$9D,$B9,$F9,$1F,$11,$1E,$CD,$74,$60,$86
	.dc.b	$08,$75,$94,$71,$8C,$47,$BC,$8F,$AD,$AF,$9D,$AF,$03,$44,$34,$53
	.dc.b	$F1,$FC,$2F,$2F,$F0,$D0,$8C,$3F,$89,$6B,$F9,$36,$BC,$A3,$8C,$62
	.dc.b	$3F,$40,$D7,$F9,$53,$5E,$D1,$C0,$8F,$77,$FC,$17,$D3,$23,$FB,$9F
	.dc.b	$57,$F8,$35,$FD,$3B,$FE,$61,$F2,$FD,$29,$A4,$8F,$D1,$46,$5B,$19
	.dc.b	$7F,$CC,$ED,$7A,$B5,$FD,$D4,$71,$7F,$0D,$F2,$3F,$93,$23,$F8,$F6
	.dc.b	$BA,$30,$43,$04,$3F,$04,$F0,$7F,$84,$79,$4F,$DF,$1F,$15,$FB,$B8
	.dc.b	$C3,$63,$33,$FC,$1E,$72,$D8,$CB,$FC,$3A,$32,$D8,$CB,$E0,$6B,$FD
	.dc.b	$86,$36,$E6,$7C,$04,$7B,$7C,$57,$8C,$8E,$08,$77,$72,$D3,$27,$85
	.dc.b	$CB,$4F,$DD,$7C,$8F,$F0,$12,$38,$21,$FB,$97,$2D,$32,$7E,$BD,$CB
	.dc.b	$4F,$F0,$7F,$C2,$3C,$8F,$F0,$AB,$63,$31,$82,$38,$21,$FE,$14,$8E
	.dc.b	$0D,$8D,$B4,$D9,$FE,$16,$87,$87,$E0,$24,$79,$1A,$F5,$6B,$C5,$AF
	.dc.b	$F8,$5F,$BD,$3D,$23,$91,$0F,$E3,$DF,$23,$FC,$AC,$8F,$F2,$76,$BA
	.dc.b	$30,$43,$04,$3F,$8D,$3C,$1F,$E3,$9F,$23,$FC,$DC,$8F,$F3,$36,$BF
	.dc.b	$CB,$DA,$F0,$34,$43,$45,$3F,$94,$7C,$77,$F4,$32,$3F,$CF,$DA,$FF
	.dc.b	$3B,$6B,$C0,$D1,$0D,$14,$FE,$F1,$EE,$BD,$C7,$BB,$FD,$AE,$30,$D8
	.dc.b	$CC,$FF,$77,$9C,$B6,$32,$FF,$7E,$8C,$B6,$32,$F8,$1A,$FF,$94,$63
	.dc.b	$6E,$67,$C0,$47,$FB,$F7,$C8,$3E,$1F,$F9,$DB,$63,$31,$82,$38,$21
	.dc.b	$FE,$0F,$8E,$0D,$8D,$B4,$D9,$FD,$EF,$E1,$7F,$C1,$04,$7F,$C0,$CD
	.dc.b	$7F,$C0,$8D,$78,$1A,$21,$A2,$9F,$E2,$E3,$FC,$58,$43,$04,$3F,$CE
	.dc.b	$DD,$BF,$E0,$7F,$0F,$FC,$6D,$18,$ED,$46,$21,$3B,$7F,$C6,$E6,$E3
	.dc.b	$FC,$4C,$FF,$8B,$BC,$2A,$FF,$8B,$8F,$F1,$67,$87,$FE,$29,$7F,$C5
	.dc.b	$A6,$23,$6F,$F8,$B0,$C1,$0F,$F1,$8B,$FE,$20,$5E,$3B,$C4,$FF,$12
	.dc.b	$C7,$FA,$05,$AE,$9B,$1B,$F1,$39,$36,$7F,$C6,$84,$3F,$C6,$45,$3F
	.dc.b	$C6,$07,$F8,$A7,$F6,$0F,$E7,$EF,$C4,$6E,$46,$93,$AF,$FC,$6A,$D9
	.dc.b	$E5,$18,$23,$82,$1F,$E3,$18,$75,$F2,$24,$78,$DA,$ED,$AF,$F8,$CE
	.dc.b	$38,$BB,$21,$B2,$9B,$F8,$CF,$39,$1F,$80,$D7,$C8,$D7,$81,$A2,$1A
	.dc.b	$29,$FE,$39,$F1,$BF,$C9,$2D,$3F,$C9,$84,$3F,$C9,$8E,$9F,$F2,$F1
	.dc.b	$93,$81,$A3,$07,$E6,$7C,$6F,$F2,$23,$4F,$F2,$59,$FE,$49,$72,$FE
	.dc.b	$73,$83,$FC,$96,$E8,$E0,$7F,$94,$0A,$71,$3F,$C8,$87,$E9,$DF,$F2
	.dc.b	$5B,$4C,$9D,$5C,$B4,$CE,$5F,$E7,$78,$CB,$63,$F1,$DE,$37,$E2,$08
	.dc.b	$E0,$87,$E1,$5C,$B4,$C9,$F8,$17,$2D,$3F,$35,$E7,$FE,$50,$8E,$08
	.dc.b	$7E,$45,$CB,$4C,$9F,$8D,$72,$D3,$F4,$DF,$19,$FA,$02,$38,$21,$F9
	.dc.b	$D7,$2D,$32,$7E,$65,$CB,$4F,$F3,$01,$0F,$F3,$77,$91,$FE,$64,$69
	.dc.b	$93,$FC,$E0,$53,$FC,$DE,$ED,$C9,$C0,$FF,$65,$78,$DF,$E6,$97,$FC
	.dc.b	$B7,$E4,$7F,$9B,$5C,$B5,$FF,$36,$BF,$E6,$14,$23,$FC,$C4,$FF,$99
	.dc.b	$8F,$F3,$59,$0F,$F3,$59,$4D,$18,$21,$82,$1F,$CE,$B6,$7C,$67,$F0
	.dc.b	$E2,$47,$F1,$76,$BF,$C2,$DA,$FF,$05,$6B,$CD,$AF,$F1,$D7,$FC,$7B
	.dc.b	$E8,$C8,$FF,$20,$B5,$FF,$1D,$35,$FF,$1B,$35,$FF,$17,$78,$48,$C4
	.dc.b	$3C,$3F,$F1,$8C,$61,$FF,$19,$FA,$3F,$E4,$C2,$4F,$F9,$41,$AF,$F9
	.dc.b	$25,$AF,$F9,$15,$AF,$81,$AF,$7A,$47,$FA,$26,$3F,$B0,$79,$55,$FE
	.dc.b	$08,$DE,$0F,$F7,$63,$78,$4D,$24,$FF,$A2,$79,$BF,$CF,$CF,$FA,$1D
	.dc.b	$AF,$FA,$0C,$D1,$0D,$14,$D1,$82,$1F,$E7,$D3,$FD,$04,$43,$04,$3F
	.dc.b	$D2,$5E,$57,$F4,$07,$46,$39,$FF,$D0,$ED,$3F,$D1,$65,$3F,$A8,$7C
	.dc.b	$67,$F5,$82,$3F,$D8,$5A,$E8,$C1,$0C,$10,$FE,$A4,$F0,$7F,$AA,$76
	.dc.b	$FF,$64,$23,$FD,$B5,$AF,$F6,$16,$BC,$0D,$10,$D1,$4F,$EC,$1F,$25
	.dc.b	$FD,$C0,$8F,$F7,$96,$BF,$DB,$5A,$F0,$34,$43,$45,$3F,$D4,$47,$FA
	.dc.b	$9C,$86,$08,$7F,$8A,$FC,$BF,$F5,$4B,$4F,$F5,$69,$4E,$25,$3F,$D3
	.dc.b	$DE,$5F,$FA,$89,$A7,$FA,$88,$FD,$5B,$96,$99,$3F,$3C,$E5,$A6,$78
	.dc.b	$3F,$D3,$FE,$5F,$FA,$8A,$3F,$48,$D8,$CC,$7F,$A8,$4F,$F5,$41,$4F
	.dc.b	$DD,$C6,$F8,$7F,$D5,$11,$FE,$AF,$3F,$90,$9F,$C9,$0A,$7F,$2A,$29
	.dc.b	$FC,$9B,$97,$F9,$73,$79,$BF,$96,$3B,$3F,$97,$9C,$0F,$E5,$8E,$9E
	.dc.b	$06,$63,$9B,$A3,$11,$B7,$07,$42,$1E,$72,$9E,$13,$04,$30,$43,$C2
	.dc.b	$70,$29,$B2,$9B,$3C,$26,$4A,$66,$3D,$24,$8F,$43,$63,$D6,$43,$C0
	.dc.b	$72,$3B,$90,$C1,$1E,$4F,$06,$BE,$36,$BE,$38,$F0,$91,$F0,$1B,$70
	.dc.b	$F3,$6B,$93,$67,$33,$F0,$3D,$CF,$DA,$FF,$13,$1F,$88,$8C,$61,$F7
	.dc.b	$DB,$1C,$27,$6F,$F6,$E8,$CB,$F8,$E6,$BC,$E3,$6F,$FA,$21,$B1,$A8
	.dc.b	$E5,$18,$24,$FE,$3E,$1F,$46,$BC,$36,$3F,$28,$D7,$F3,$AD,$7F,$42
	.dc.b	$D7,$F4,$AD,$7F,$C0,$ED,$7B,$C7,$62,$3D,$9A,$FF,$85,$CF,$F0,$C9
	.dc.b	$0F,$F0,$E9,$FE,$1F,$3F,$C2,$7D,$AE,$BF,$F1,$04,$68,$E4,$53,$91
	.dc.b	$C1,$FB,$6F,$D9,$3E,$F1,$F7,$0F,$C7,$65,$E9,$27,$9F,$FE,$23,$8F
	.dc.b	$F1,$13,$63,$FC,$47,$1F,$C1,$9A,$FF,$03,$8D,$11,$FF,$12,$1F,$C3
	.dc.b	$3E,$49,$E6,$7D,$56,$C7,$F0,$56,$BF,$C3,$9A,$FF,$12,$6B,$CE,$39
	.dc.b	$11,$E4,$D7,$93,$5F,$B7,$8F,$D2,$69,$F5,$48,$7E,$09,$D3,$F8,$06
	.dc.b	$BF,$80,$6B,$F8,$03,$EC,$0C,$7E,$C3,$86,$BF,$60,$FA,$C6,$71,$FB
	.dc.b	$AE,$0D,$1C,$DF,$B6,$FD,$89,$AF,$F0,$57,$E8,$3F,$E5,$03,$EB,$61
	.dc.b	$87,$FC,$A2,$7D,$73,$EA,$94,$E0,$53,$FD,$A5,$1F,$EE,$22,$3F,$EE
	.dc.b	$B6,$BF,$ED,$F6,$BC,$0D,$10,$D1,$4F,$F6,$F4,$7F,$BB,$88,$FF,$BD
	.dc.b	$DA,$FF,$BA,$DA,$F0,$34,$43,$45,$3F,$AD,$F5,$3E,$55,$E0,$7F,$53
	.dc.b	$8C,$36,$33,$3F,$D7,$A7,$2D,$8C,$BF,$D9,$23,$2D,$8C,$BE,$06,$BF
	.dc.b	$E8,$C8,$DB,$99,$F0,$11,$FE,$CD,$CC,$F3,$3F,$8D,$36,$33,$18,$23
	.dc.b	$82,$1F,$DD,$A3,$83,$63,$6D,$36,$7F,$3F,$3F,$82,$90,$C1,$0F,$E1
	.dc.b	$47,$A4,$F8,$93,$E9,$7E,$13,$5D,$18,$36,$43,$04,$36,$7F,$84,$27
	.dc.b	$D2,$D9,$F4,$B4,$F4,$9F,$E1,$88,$DB,$63,$6D,$36,$7F,$AF,$A3,$6D
	.dc.b	$8F,$C2,$71,$BA,$E9,$27,$F0,$91,$96,$C6,$5F,$C2,$46,$5B,$19,$7F
	.dc.b	$27,$19,$6C,$65,$FF,$62,$7C,$67,$E5,$7C,$23,$49,$1F,$95,$8C,$B6
	.dc.b	$32,$FE,$56,$32,$D8,$CB,$FA,$48,$CB,$63,$C1,$EC,$48,$EC,$D7,$A3
	.dc.b	$5E,$4D,$7F,$DB,$8D,$59,$A9,$39,$8C,$7C,$07,$BC,$95,$ED,$6B,$EB
	.dc.b	$6B,$E9,$6B,$FE,$EB,$6B,$E0,$A4,$7C,$BF,$80,$93,$F2,$5A,$FC,$76
	.dc.b	$BF,$15,$AF,$FB,$DD,$AF,$56,$D7,$B7,$E5,$3E,$C1,$27,$EB,$35,$FA
	.dc.b	$8D,$7E,$93,$5F,$E0,$D6,$8C,$47,$AD,$FC,$27,$2C,$8F,$C1,$35,$FC
	.dc.b	$03,$5F,$BC,$D7,$AB,$5F,$8C,$DA,$FA,$DF,$29,$F9,$12,$4F,$E3,$DA
	.dc.b	$FE,$35,$AF,$E2,$DA,$FA,$9A,$FD,$37,$F3,$9F,$29,$23,$F3,$4D,$7F
	.dc.b	$30,$D7,$F2,$CD,$7E,$2B,$5F,$BC,$D9,$FC,$77,$CA,$7E,$A4,$93,$FA
	.dc.b	$76,$BF,$A5,$6B,$FA,$36,$BF,$49,$AF,$E2,$DF,$E9,$1F,$B1,$3D,$6A
	.dc.b	$49,$FE,$91,$19,$6C,$65,$FE,$91,$19,$6C,$65,$FE,$AF,$19,$6C,$65
	.dc.b	$FE,$95,$0F,$57,$C6,$48,$F0,$B5,$EE,$D7,$AB,$5E,$6D,$7F,$6F,$1D
	.dc.b	$5F,$F5,$FF,$82,$F7,$FF,$73,$18,$6B,$FE,$C6,$8D,$70,$FF,$AF,$DB
	.dc.b	$1B,$8C,$4F,$08,$FF,$67,$CE,$9E,$51,$94,$23,$4F,$08,$D3,$96,$C6
	.dc.b	$A3,$FA,$27,$A9,$FD,$A8,$8F,$F6,$E6,$BF,$D9,$DA,$FF,$51,$6B,$C8
	.dc.b	$E0,$43,$FB,$31,$FD,$90,$87,$F8,$D4,$FF,$37,$1F,$E3,$53,$FC,$E1
	.dc.b	$1E,$57,$D1,$5F,$10,$8F,$C7,$21,$F1,$1C,$B4,$C9,$F0,$DC,$BE,$93
	.dc.b	$CC,$53,$DC,$60,$F6,$18,$21,$B2,$1A,$28,$F0,$DE,$A3,$D4,$FF,$60
	.dc.b	$36,$7F,$D8,$ED,$7F,$D7,$CD,$7F,$D9,$0D,$7F,$D8,$8D,$7B,$36,$7B
	.dc.b	$CF,$62,$4F,$66,$BF,$AE,$F8,$1F,$DB,$91,$FD,$AB,$5D,$18,$21,$82
	.dc.b	$1F,$B4,$78,$3F,$B6,$F9,$1F,$E0,$64,$7F,$80,$B5,$FD,$FB,$5E,$06
	.dc.b	$88,$68,$A7,$F0,$2F,$95,$FE,$1E,$47,$F8,$6B,$5F,$E1,$6D,$78,$1A
	.dc.b	$21,$A2,$9F,$C3,$BD,$9F,$E3,$64,$7F,$8C,$B5,$FE,$2E,$D7,$81,$A2
	.dc.b	$1A,$29,$FC,$6A,$FF,$93,$91,$FE,$4A,$D7,$F9,$1B,$5E,$06,$88,$68
	.dc.b	$A7,$F2,$6F,$8A,$FE,$66,$47,$F9,$8B,$5F,$E5,$ED,$78,$1A,$21,$A2
	.dc.b	$9F,$CC,$BE,$2B,$F9,$F9,$1F,$E7,$AD,$7F,$9D,$B5,$E0,$68,$86,$8A
	.dc.b	$7F,$3E,$F9,$1F,$E9,$64,$7F,$A4,$B5,$FE,$8E,$D7,$81,$A2,$1A,$29
	.dc.b	$FD,$2B,$E4,$7F,$AB,$91,$FE,$AA,$D7,$FA,$9B,$5E,$06,$88,$68,$A7
	.dc.b	$F5,$6F,$91,$FE,$C6,$47,$FB,$0B,$5F,$EB,$ED,$78,$1A,$21,$A2,$9F
	.dc.b	$D8,$BE,$37,$FB,$79,$1F,$ED,$AD,$7F,$B5,$B5,$E0,$68,$86,$8A,$7F
	.dc.b	$6E,$F8,$BF,$EF,$64,$7F,$BC,$B5,$FE,$EE,$D7,$81,$A2,$1A,$29,$FD
	.dc.b	$EB,$E5,$7F,$C0,$E4,$7F,$C0,$AD,$7F,$C0,$6D,$78,$1A,$21,$A2,$9F
	.dc.b	$E0,$6F,$95,$FF,$09,$91,$FF,$08,$B5,$FF,$07,$B5,$E0,$68,$86,$8A
	.dc.b	$7F,$84,$BC,$CF,$F0,$F9,$1F,$F0,$EB,$5F,$F0,$DB,$5E,$06,$88,$68
	.dc.b	$A7,$F8,$7B,$E5,$7F,$C5,$64,$7F,$C5,$2D,$7F,$C4,$ED,$78,$1A,$21
	.dc.b	$A2,$9F,$E2,$AF,$23,$FC,$6E,$47,$FC,$6A,$D7,$FC,$66,$D7,$81,$A2
	.dc.b	$1A,$29,$FE,$36,$E8,$FF,$21,$91,$FF,$20,$B5,$FF,$1F,$B5,$E0,$68
	.dc.b	$86,$8A,$7F,$90,$BC,$FF,$F2,$79,$1F,$F2,$6B,$5F,$F2,$5B,$5E,$06
	.dc.b	$88,$68,$A7,$F8,$B0,$FF,$5A,$FE,$C7,$59,$E0,$23,$11,$B8,$C3,$83
	.dc.b	$89,$5C,$BC,$1A,$F4,$8E,$0F,$03,$4E,$23,$71,$DE,$30,$4F,$95,$EC
	.dc.b	$9A,$48,$E6,$D7,$C0,$D7,$A3,$5F,$F2,$6C,$FB,$E3,$11,$F1,$3E,$57
	.dc.b	$B9,$1F,$3B,$5F,$63,$5F,$43,$5E,$B4,$8C,$47,$53,$E0,$FC,$AF,$B0
	.dc.b	$8E,$08,$7C,$97,$2D,$32,$7C,$77,$2D,$3E,$6F,$CA,$7C,$82,$38,$21
	.dc.b	$F5,$5C,$B4,$C9,$F4,$DC,$B4,$FB,$3F,$29,$F5,$08,$E0,$87,$E0,$5C
	.dc.b	$B4,$C9,$F7,$DC,$B4,$FC,$2F,$B5,$F8,$02,$38,$21,$F8,$D7,$2D,$32
	.dc.b	$7E,$2D,$CB,$4F,$C8,$FB,$3F,$8C,$23,$82,$1F,$99,$72,$D3,$27,$E5
	.dc.b	$DC,$B4,$FC,$EF,$C1,$FE,$60,$8E,$08,$7E,$95,$CB,$4C,$9F,$A3,$72
	.dc.b	$D3,$F5,$3F,$07,$FA,$42,$38,$21,$FB,$17,$2D,$32,$7E,$BD,$CB,$4F
	.dc.b	$DA,$FC,$8F,$EC,$08,$E0,$87,$EF,$5C,$B4,$C9,$FB,$B7,$2D,$3F,$80
	.dc.b	$FC,$8F,$EF,$08,$E0,$87,$F0,$97,$2D,$32,$7F,$07,$72,$D3,$F8,$6F
	.dc.b	$C8,$FF,$08,$23,$82,$1F,$C5,$5C,$B4,$C9,$FC,$4D,$CB,$4F,$E3,$3F
	.dc.b	$19,$FC,$50,$8E,$08,$7F,$21,$72,$D3,$27,$F1,$F7,$2D,$3F,$92,$FC
	.dc.b	$5F,$F2,$02,$38,$21,$FC,$B5,$CB,$4C,$9F,$CA,$DC,$B4,$FE,$63,$F1
	.dc.b	$7F,$CB,$08,$E0,$87,$F3,$97,$2D,$32,$7F,$37,$72,$D3,$F9,$EF,$C4
	.dc.b	$FF,$38,$23,$82,$1F,$D1,$5C,$B4,$C9,$FD,$0D,$CB,$4F,$E9,$3F,$13
	.dc.b	$FD,$10,$8E,$08,$7F,$51,$72,$D3,$27,$F4,$F7,$2D,$3F,$AA,$FC,$2F
	.dc.b	$F5,$02,$38,$21,$FD,$75,$CB,$4C,$9F,$D6,$DC,$B4,$FE,$C3,$F2,$1F
	.dc.b	$D7,$08,$E0,$87,$F6,$97,$2D,$32,$7F,$67,$72,$D3,$FB,$6F,$C8,$7F
	.dc.b	$68,$23,$82,$1F,$DD,$5C,$B4,$C9,$FD,$CD,$CB,$4F,$EF,$3F,$21,$FD
	.dc.b	$D0,$8E,$08,$7F,$80,$5C,$B4,$C9,$FD,$FD,$CB,$4F,$F0,$57,$73,$DE
	.dc.b	$FF,$07,$36,$33,$18,$23,$82,$1F,$E1,$C8,$E0,$D8,$DB,$4D,$9F,$E2
	.dc.b	$3F,$87,$7C,$2F,$F8,$61,$B1,$FE,$21,$6B,$FE,$26,$6B,$C6,$38,$11
	.dc.b	$E0,$D7,$83,$5E,$07,$F8,$B7,$E1,$97,$FE,$28,$6C,$66,$30,$47,$04
	.dc.b	$3F,$C7,$31,$C1,$B1,$B6,$9B,$3F,$C8,$5F,$0F,$EA,$7F,$C8,$CD,$41
	.dc.b	$C3,$6B,$FC,$7B,$18,$23,$82,$47,$F9,$26,$32,$D8,$F8,$4F,$F9,$56
	.dc.b	$FA,$92,$33,$1E,$28,$C3,$F3,$08,$7C,$C2,$9F,$30,$F7,$90,$FF,$30
	.dc.b	$5F,$C5,$6D,$FC,$56,$9F,$14,$FF,$34,$46,$DB,$1B,$69,$B3,$E3,$FE
	.dc.b	$02,$F2,$DF,$81,$7C,$AF,$91,$FC,$4C,$E6,$31,$97,$F9,$D6,$31,$6A
	.dc.b	$DD,$F2,$8C,$56,$E3,$13,$AB,$E7,$5E,$29,$CD,$6E,$73,$2B,$EF,$7C
	.dc.b	$63,$A9,$C6,$77,$3A,$AE,$53,$FE,$86,$8F,$3B,$FE,$8B,$6B,$A3,$17
	.dc.b	$E9,$6D,$7F,$A2,$E7,$71,$DC,$E9,$3B,$79,$10,$D1,$F9,$4F,$C0,$C4
	.dc.b	$00

;--------------------------------------------------------------------------------
;	FINT.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fint_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#59,-(sp)
	peamsg	'FINT.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FINT.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fint_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fint.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FINT.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fint.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FINT.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fint_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$BB,$F6,$7F,$79,$A5,$3F,$9A,$B5,$A5,$70,$9A,$47
	.dc.b	$44,$3A,$39,$7C,$1E,$AF,$F4,$57,$F1,$ED,$7F,$98,$B5,$FE,$8E,$D7
	.dc.b	$FC,$53,$F1,$BE,$58,$C1,$1C,$12,$3D,$91,$97,$DD,$EB,$73,$24,$FB
	.dc.b	$DB,$1A,$8C,$11,$C1,$0F,$95,$1C,$1F,$99,$EB,$79,$C9,$3F,$2D,$B1
	.dc.b	$F3,$DA,$F0,$8D,$11,$D3,$5F,$B1,$EB,$7C,$32,$47,$D7,$6C,$7D,$B6
	.dc.b	$BC,$23,$44,$74,$D7,$F0,$9E,$B7,$CF,$24,$7E,$0D,$B1,$F8,$76,$BC
	.dc.b	$23,$44,$74,$D7,$F2,$1E,$B7,$DB,$24,$7E,$3D,$B1,$F9,$36,$BC,$23
	.dc.b	$44,$74,$D7,$F3,$9E,$B7,$E1,$C9,$1F,$9B,$6C,$7E,$7D,$AF,$08,$D1
	.dc.b	$1D,$35,$FD,$47,$AD,$F9,$32,$47,$E9,$DB,$1F,$AB,$6B,$C2,$34,$47
	.dc.b	$4D,$7F,$69,$EB,$7E,$7C,$91,$FB,$36,$C7,$ED,$DA,$F0,$8D,$11,$D3
	.dc.b	$5F,$E0,$1E,$B7,$EA,$C9,$1F,$BF,$6C,$7F,$03,$6B,$C2,$34,$47,$4D
	.dc.b	$7F,$86,$7A,$DF,$B7,$24,$7F,$0B,$6C,$7F,$0F,$6B,$C2,$34,$47,$4D
	.dc.b	$7F,$8C,$7A,$DF,$C0,$C9,$1F,$C5,$DB,$1F,$C6,$DA,$F0,$8D,$11,$D3
	.dc.b	$5F,$E4,$9E,$B7,$F0,$F2,$47,$F2,$36,$C7,$F2,$76,$BC,$23,$44,$74
	.dc.b	$D7,$F9,$87,$AD,$FC,$6C,$91,$FC,$BD,$B1,$FC,$CD,$AF,$08,$D1,$1D
	.dc.b	$35,$FE,$79,$EB,$7F,$27,$24,$7F,$3B,$6C,$7F,$3F,$6B,$C2,$34,$47
	.dc.b	$4D,$7F,$A4,$7A,$DF,$CC,$C9,$1F,$D1,$DB,$1F,$D2,$DA,$F0,$8D,$11
	.dc.b	$D3,$5F,$EA,$9E,$B7,$F3,$F2,$47,$F5,$36,$C7,$F5,$76,$BC,$23,$44
	.dc.b	$74,$D7,$FB,$07,$AD,$FD,$2C,$91,$FD,$7D,$B1,$FD,$8D,$AF,$08,$D1
	.dc.b	$1D,$35,$FE,$D9,$EB,$7F,$57,$24,$7F,$6B,$6C,$7F,$6F,$6B,$C2,$34
	.dc.b	$47,$4D,$7F,$BC,$7A,$DF,$D8,$C9,$1F,$DD,$DB,$1F,$DE,$DA,$F0,$8D
	.dc.b	$11,$D3,$5F,$F0,$27,$AD,$FD,$BC,$91,$FE,$03,$6C,$7F,$81,$DA,$F0
	.dc.b	$8D,$11,$D3,$5F,$F2,$6F,$DD,$23,$FB,$F3,$5F,$F0,$93,$63,$FC,$2C
	.dc.b	$D7,$8C,$70,$21,$C0,$AF,$06,$BC,$1E,$BF,$19,$E0,$23,$82,$47,$82
	.dc.b	$32,$D8,$CB,$EA,$8C,$B6,$3D,$3F,$0B,$EC,$23,$82,$47,$B2,$32,$D8
	.dc.b	$CB,$F1,$A3,$2D,$8F,$8B,$EC,$7C,$82,$38,$24,$7C,$88,$CB,$63,$2F
	.dc.b	$D2,$8C,$B6,$3E,$8F,$B1,$F5,$08,$E0,$91,$F5,$23,$2D,$8C,$BF,$7A
	.dc.b	$32,$D8,$FB,$BE,$C7,$E0,$08,$E0,$91,$F8,$08,$CB,$63,$2F,$E2,$A3
	.dc.b	$2D,$8F,$C4,$FB,$1F,$8C,$23,$82,$47,$E3,$23,$2D,$8C,$BF,$96,$8C
	.dc.b	$B6,$3F,$2B,$EC,$7E,$60,$8E,$09,$1F,$98,$8C,$B6,$32,$FE,$8A,$32
	.dc.b	$D8,$FD,$0F,$B1,$FA,$42,$38,$24,$7E,$92,$32,$D8,$CB,$FA,$E8,$CB
	.dc.b	$63,$F5,$BE,$C7,$EC,$08,$E0,$91,$FB,$08,$CB,$63,$2F,$EE,$A3,$2D
	.dc.b	$8F,$DC,$FB,$1F,$BC,$23,$82,$47,$EF,$23,$2D,$8C,$BF,$C1,$A3,$2D
	.dc.b	$8F,$E0,$BE,$C7,$F0,$82,$38,$24,$7F,$08,$8C,$B6,$32,$FF,$12,$8C
	.dc.b	$B6,$3F,$88,$FB,$1F,$C5,$08,$E0,$91,$FC,$52,$32,$D8,$CB,$FC,$7A
	.dc.b	$32,$D8,$FE,$3B,$EC,$7F,$20,$23,$82,$47,$F2,$08,$CB,$63,$2F,$F2
	.dc.b	$A8,$CB,$63,$F9,$4F,$B1,$FC,$B0,$8E,$09,$1F,$CB,$23,$2D,$8C,$BF
	.dc.b	$CD,$A3,$2D,$8F,$E6,$BE,$C7,$F3,$82,$38,$24,$7F,$38,$8C,$B6,$32
	.dc.b	$FF,$42,$8C,$B6,$3F,$A0,$FB,$1F,$D1,$08,$E0,$91,$FD,$12,$32,$D8
	.dc.b	$CB,$FD,$3A,$32,$D8,$FE,$9B,$EC,$7F,$50,$23,$82,$47,$F5,$08,$CB
	.dc.b	$63,$2F,$F5,$A8,$CB,$63,$FA,$CF,$B1,$FD,$70,$8E,$09,$1F,$D7,$23
	.dc.b	$2D,$8C,$BF,$D9,$A3,$2D,$8F,$EC,$BE,$C7,$F6,$82,$38,$24,$7F,$68
	.dc.b	$8C,$B6,$32,$FF,$72,$8C,$B6,$3F,$B8,$FB,$1F,$DD,$08,$E0,$91,$FD
	.dc.b	$D2,$32,$D8,$CB,$FD,$FA,$32,$D8,$30,$9F,$74,$F8,$AF,$F8,$00,$D9
	.dc.b	$D4,$EA,$72,$48,$CC,$74,$9E,$0E,$1A,$F2,$6B,$CE,$38,$A8,$CB,$CE
	.dc.b	$7B,$3C,$FE,$2B,$CA,$4A,$E3,$48,$C4,$EE,$73,$18,$8F,$13,$5F,$82
	.dc.b	$D7,$D5,$ED,$7C,$22,$4F,$66,$C7,$B5,$AF,$B5,$AF,$CC,$6B,$D2,$3C
	.dc.b	$FE,$D7,$CE,$23,$82,$47,$C9,$8C,$B6,$32,$FC,$B8,$CB,$63,$E1,$FB
	.dc.b	$5F,$68,$8E,$09,$1F,$56,$32,$D8,$CB,$F5,$E3,$2D,$8F,$9F,$ED,$7E
	.dc.b	$18,$8E,$09,$1F,$81,$8C,$B6,$32,$FE,$0E,$32,$D8,$FB,$7E,$D7,$E4
	.dc.b	$88,$E0,$91,$F8,$D8,$CB,$63,$2F,$E3,$E3,$2D,$8F,$C3,$FB,$5F,$9E
	.dc.b	$23,$82,$47,$E6,$63,$2D,$8C,$BF,$9B,$8C,$B6,$3F,$27,$ED,$7E,$A8
	.dc.b	$8E,$09,$1F,$A5,$8C,$B6,$32,$FE,$9E,$32,$D8,$FC,$FF,$B5,$FB,$62
	.dc.b	$38,$24,$7E,$C6,$32,$D8,$CB,$FB,$38,$CB,$63,$F5,$7F,$0F,$FC,$08
	.dc.b	$8E,$09,$1F,$BD,$8C,$B6,$32,$FE,$FE,$32,$D8,$FD,$BF,$C3,$FF,$0E
	.dc.b	$23,$82,$47,$F0,$98,$CB,$63,$2F,$F0,$B8,$CB,$63,$F8,$1F,$C3,$FF
	.dc.b	$1A,$23,$82,$47,$F1,$58,$CB,$63,$2F,$F1,$78,$CB,$63,$F8,$7F,$C3
	.dc.b	$FF,$26,$23,$82,$47,$F2,$18,$CB,$63,$2F,$F2,$38,$CB,$63,$F8,$DF
	.dc.b	$C3,$FF,$32,$23,$82,$47,$F2,$D8,$CB,$63,$2F,$F2,$F8,$CB,$63,$F9
	.dc.b	$3F,$C3,$FF,$3E,$23,$82,$47,$F3,$98,$CB,$63,$2F,$F3,$B8,$CB,$63
	.dc.b	$F9,$9F,$C3,$FF,$4A,$23,$82,$47,$F4,$58,$CB,$63,$2F,$F4,$78,$CB
	.dc.b	$63,$F9,$FF,$C3,$FF,$56,$23,$82,$47,$F5,$18,$CB,$63,$2F,$F5,$38
	.dc.b	$CB,$63,$FA,$5F,$C3,$FF,$62,$23,$82,$47,$F5,$D8,$CB,$63,$2F,$F5
	.dc.b	$F8,$CB,$63,$FA,$BF,$C3,$FF,$6E,$23,$82,$47,$F6,$98,$CB,$63,$2F
	.dc.b	$F6,$B8,$CB,$63,$FB,$1F,$C3,$FF,$7A,$23,$82,$47,$F7,$58,$CB,$63
	.dc.b	$2F,$F7,$78,$CB,$63,$FB,$BF,$DD,$7F,$82,$9B,$06,$91,$FF,$02,$C6
	.dc.b	$DA,$6C,$B1,$B7,$FC,$1F,$1B,$69,$B2,$C6,$DF,$F0,$D4,$3E,$13,$51
	.dc.b	$86,$BF,$E2,$08,$D4,$FA,$A7,$32,$9C,$BE,$88,$CC,$62,$3C,$B1,$82
	.dc.b	$4F,$32,$47,$C0,$F8,$4E,$24,$8F,$92,$D8,$D4,$60,$8E,$08,$7C,$48
	.dc.b	$E0,$FC,$6F,$84,$F2,$92,$7E,$AB,$63,$E8,$B5,$E1,$1A,$23,$A6,$BF
	.dc.b	$4B,$E1,$3E,$09,$23,$F0,$2D,$8F,$BA,$D7,$84,$68,$8E,$9A,$FD,$EF
	.dc.b	$84,$F9,$A4,$8F,$C6,$B6,$3F,$12,$D7,$84,$68,$8E,$9A,$FE,$2B,$E1
	.dc.b	$3E,$C9,$23,$F3,$2D,$8F,$CA,$B5,$E1,$1A,$23,$A6,$BF,$96,$F8,$4F
	.dc.b	$C2,$92,$3F,$4A,$D8,$FD,$0B,$5E,$11,$A2,$3A,$6B,$FA,$2F,$84,$FC
	.dc.b	$89,$23,$F6,$2D,$8F,$D6,$B5,$E1,$1A,$23,$A6,$BF,$AE,$F8,$4F,$CE
	.dc.b	$92,$3F,$7A,$D8,$FD,$CB,$5E,$11,$A2,$3A,$6B,$FB,$AF,$84,$FD,$49
	.dc.b	$23,$F8,$4B,$63,$F8,$2B,$5E,$11,$A2,$3A,$6B,$FC,$1B,$E1,$3F,$6A
	.dc.b	$48,$FE,$2A,$D8,$FE,$22,$D7,$84,$68,$8E,$9A,$FF,$12,$F8,$4F,$E0
	.dc.b	$24,$8F,$E4,$2D,$8F,$E3,$AD,$78,$46,$88,$E9,$AF,$F1,$EF,$84,$FE
	.dc.b	$1A,$48,$FE,$5A,$D8,$FE,$52,$D7,$84,$68,$8E,$9A,$FF,$2A,$F8,$4F
	.dc.b	$E3,$24,$8F,$E7,$2D,$8F,$E6,$AD,$78,$46,$88,$E9,$AF,$F3,$6F,$84
	.dc.b	$FE,$4A,$48,$FE,$8A,$D8,$FE,$82,$D7,$84,$68,$8E,$9A,$E8,$FF,$62
	.dc.b	$3F,$CF,$BE,$0F,$FA,$47,$93,$FD,$0F,$95,$78,$AD,$1F,$ED,$33,$FA
	.dc.b	$2B,$9B,$FE,$85,$75,$FE,$81,$16,$93,$63,$0D,$AE,$95,$C9,$AF,$F4
	.dc.b	$98,$FE,$8E,$48,$FE,$CE,$D8,$FE,$C6,$D7,$FA,$FB,$5F,$49,$A8,$E2
	.dc.b	$7F,$B9,$BF,$3D,$7B,$7E,$96,$C7,$8B,$0F,$4B,$E4,$8C,$D7,$CA,$3F
	.dc.b	$E0,$0B,$8C,$7F,$B6,$36,$7C,$31,$8B,$EC,$48,$ED,$85,$BE,$87,$F6
	.dc.b	$9F,$83,$FE,$CE,$49,$FF,$03,$B6,$3F,$C0,$6D,$7F,$BF,$B5,$FE,$F6
	.dc.b	$D7,$FB,$C3,$FD,$CB,$E0,$E4,$FF,$84,$9B,$3F,$E0,$E6,$BF,$E0,$A6
	.dc.b	$BF,$E0,$66,$BF,$E0,$36,$BF,$E0,$1F,$83,$FE,$FE,$49,$FF,$0F,$B6
	.dc.b	$3F,$C3,$6D,$7F,$C2,$ED,$7F,$C2,$6D,$7F,$C2,$0F,$F8,$3B,$CF,$FF
	.dc.b	$13,$4A,$61,$23,$2E,$0D,$38,$70,$D7,$84,$75,$21,$C5,$E2,$E1,$C4
	.dc.b	$71,$73,$1B,$51,$96,$FC,$17,$A4,$93,$B9,$46,$27,$AB,$5F,$3B,$5F
	.dc.b	$73,$5F,$6B,$F1,$7E,$0A,$57,$56,$CF,$A1,$AF,$C2,$6B,$F2,$5A,$FC
	.dc.b	$86,$BE,$CF,$82,$FA,$44,$9F,$53,$63,$E2,$B5,$F9,$ED,$7E,$B3,$5F
	.dc.b	$AA,$FD,$DF,$82,$93,$F1,$5B,$3F,$41,$AF,$DA,$6B,$F8,$16,$BF,$80
	.dc.b	$6B,$F5,$35,$FC,$51,$27,$E9,$36,$3E,$EB,$5F,$C3,$B5,$FC,$73,$5F
	.dc.b	$C6,$BF,$95,$F8,$1F,$E4,$99,$1F,$E4,$DF,$9C,$BD,$EC,$C4,$C3,$F9
	.dc.b	$0D,$6D,$FE,$47,$5B,$73,$3B,$C3,$F8,$3D,$6D,$FE,$13,$5B,$73,$3B
	.dc.b	$7F,$3E,$F9,$6B,$F2,$0D,$8F,$CD,$B5,$DC,$68,$8E,$9A,$E8,$F1,$FB
	.dc.b	$1E,$5E,$2F,$F4,$76,$08,$C5,$7E,$D5,$B5,$FB,$46,$BF,$A8,$6C,$F1
	.dc.b	$9F,$82,$7C,$68,$C9,$1C,$CF,$89,$B5,$E2,$9D,$4F,$C4,$9E,$46,$8F
	.dc.b	$21,$5F,$23,$5F,$23,$6B,$C9,$1C,$CF,$01,$6B,$C1,$3B,$AF,$0C,$7C
	.dc.b	$4B,$FC,$2C,$62,$7F,$8C,$46,$1E,$04,$AE,$8F,$99,$A7,$98,$A7,$98
	.dc.b	$B7,$E6,$BF,$1C,$79,$9F,$14,$FC,$28,$C3,$D6,$33,$18,$3B,$3C,$8C
	.dc.b	$96,$38,$12,$3F,$97,$D7,$AD,$E4,$79,$67,$8C,$EE,$30,$DA,$FC,$E4
	.dc.b	$F5,$9F,$0C,$E6,$BA,$4E,$67,$89,$E0,$9F,$64,$66,$BE,$14,$7C,$E3
	.dc.b	$27,$7B,$FA,$55,$D8,$87,$F3,$AB,$EF,$3B,$3E,$A5,$F5,$AE,$35,$E2
	.dc.b	$25,$FA,$AB,$84,$7C,$28,$C3,$D9,$F2,$B6,$FB,$B9,$AE,$B3,$A8,$EB
	.dc.b	$3A,$7F,$00,$7F,$5C,$9F,$EC,$C4,$AF,$E9,$CD,$8F,$EB,$2D,$78,$46
	.dc.b	$88,$E9,$B6,$7F,$B7,$FD,$E3,$CF,$F1,$36,$3E,$6D,$7F,$75,$8C,$3F
	.dc.b	$82,$6D,$7D,$96,$CF,$E0,$DB,$3F,$49,$AF,$CD,$6E,$3F,$0D,$B8,$7F
	.dc.b	$94,$DB,$5F,$E4,$C6,$D7,$F8,$B9,$B7,$EB,$8F,$36,$1E,$A2,$57,$AA
	.dc.b	$15,$C2,$3E,$0C,$EA,$7D,$8D,$AF,$63,$67,$D8,$D9,$F6,$35,$F6,$37
	.dc.b	$0F,$63,$70,$F6,$36,$BD,$8D,$AF,$63,$CF,$0F,$8E,$79,$E3,$DB,$3C
	.dc.b	$08,$FA,$1B,$87,$C3,$8F,$7D,$73,$3C,$33,$F0,$EB,$AC,$62,$79,$1E
	.dc.b	$29,$E6,$7C,$02,$BE,$23,$75,$DE,$B9,$CF,$5A,$E9,$39,$7C,$51,$CA
	.dc.b	$BE,$3B,$6B,$DD,$1E,$88,$C4,$F0,$8F,$21,$E0,$8F,$B2,$73,$C3,$E7
	.dc.b	$BC,$0F,$11,$F1,$A3,$FC,$95,$5F,$9A,$6D,$7C,$77,$99,$FE,$47,$8D
	.dc.b	$38,$37,$3F,$E4,$58,$CC,$7C,$33,$27,$19,$E8,$E4,$EB,$7F,$57,$27
	.dc.b	$BF,$F1,$B0,$FF,$30,$C6,$1F,$8C,$DA,$F8,$35,$C6,$B8,$47,$F9,$A6
	.dc.b	$75,$3F,$3E,$BF,$C2,$35,$FA,$BF,$3A,$47,$FA,$DB,$F4,$17,$BF,$8C
	.dc.b	$7F,$59,$57,$0F,$C5,$56,$AF,$F6,$53,$97,$F8,$4B,$FE,$C9,$8D,$57
	.dc.b	$29,$D9,$F1,$E7,$E1,$C7,$EC,$9B,$99,$A4,$9E,$04,$9F,$A9,$19,$6C
	.dc.b	$65,$FB,$B1,$FE,$DB,$FB,$D7,$8F,$E7,$9F,$F4,$74,$62,$BC,$B1,$FE
	.dc.b	$B8,$8C,$3E,$93,$13,$C5,$46,$B1,$FD,$05,$F7,$AD,$4E,$6F,$E2,$9F
	.dc.b	$EE,$47,$17,$FC,$AE,$B8,$4F,$8A,$FF,$71,$5A,$AF,$8D,$3C,$6F,$A9
	.dc.b	$1E,$07,$FA,$A2,$78,$61,$F2,$E3,$81,$AA,$F8,$35,$C5,$F7,$5F,$82
	.dc.b	$30,$FB,$27,$8C,$E9,$ED,$3A,$7D,$B1,$F1,$9E,$CF,$C6,$8D,$3E,$F8
	.dc.b	$E7,$3D,$63,$81,$D4,$F2,$C7,$2B,$F5,$E1,$F1,$A7,$75,$E7,$AD,$5E
	.dc.b	$63,$13,$CC,$F1,$4F,$80,$F4,$9C,$0C,$3E,$03,$85,$FD,$12,$57,$AA
	.dc.b	$39,$9E,$29,$F8,$35,$E0,$23,$DA,$39,$CF,$96,$31,$3F,$04,$DC,$F8
	.dc.b	$5A,$FE,$00,$EC,$66,$38,$9E,$02,$B9,$21,$C6,$39,$D7,$AA,$7E,$04
	.dc.b	$F4,$3D,$73,$E3,$8E,$71,$86,$E1,$F8,$B8,$E1,$5C,$DF,$6C,$FC,$A7
	.dc.b	$81,$F7,$CD,$DF,$C1,$71,$7E,$48,$C4,$79,$E7,$C3,$5F,$B2,$8F,$F8
	.dc.b	$00,$D7,$E6,$18,$9F,$E1,$71,$98,$FB,$46,$67,$FD,$F5,$19,$8F,$B2
	.dc.b	$64,$14,$87,$8B,$E2,$7C,$B0,$69,$27,$E0,$46,$5B,$19,$9F,$8E,$DB
	.dc.b	$FA,$6D,$9E,$C7,$02,$4F,$06,$CF,$06,$9C,$23,$44,$8D,$36,$35,$18
	.dc.b	$23,$82,$47,$E2,$2B,$D4,$F1,$31,$5E,$76,$CF,$89,$A6,$CF,$E4,$B1
	.dc.b	$FC,$C4,$E0,$E2,$3F,$91,$46,$27,$BD,$7F,$20,$8C,$1C,$67,$B5,$71
	.dc.b	$3F,$9B,$46,$A3,$99,$AA,$E6,$FF,$23,$8D,$4F,$08,$D4,$75,$8F,$44
	.dc.b	$F0,$7F,$A2,$C7,$0A,$F2,$57,$03,$BC,$F0,$9E,$53,$F3,$23,$8C,$EA
	.dc.b	$31,$1E,$97,$9B,$C6,$3F,$2B,$3F,$9A,$23,$82,$47,$E5,$8F,$63,$87
	.dc.b	$F4,$B1,$C1,$D1,$E2,$29,$F2,$4D,$C7,$E8,$27,$FC,$8D,$E9,$39,$3D
	.dc.b	$73,$DE,$30,$FA,$CC,$46,$AB,$F5,$33,$A7,$E7,$0A,$EB,$3B,$84,$EE
	.dc.b	$7F,$CF,$19,$3E,$79,$F0,$5F,$E7,$B9,$FF,$3E,$66,$F4,$DD,$79,$D3
	.dc.b	$FD,$29,$87,$07,$D8,$7F,$7E,$9D,$B8,$CF,$FD,$03,$3B,$7C,$D1,$BA
	.dc.b	$E9,$1A,$C6,$E1,$E1,$9C,$E5,$E7,$9C,$E3,$DA,$F9,$9C,$70,$F0,$61
	.dc.b	$C5,$F1,$5F,$17,$35,$C7,$2F,$87,$97,$95,$F8,$31,$99,$F7,$E3,$E0
	.dc.b	$CB,$C7,$7B,$9F,$51,$D9,$C6,$5F,$E1,$8B,$E4,$F9,$4F,$3E,$5E,$57
	.dc.b	$FC,$45,$DB,$FC,$B0,$97,$FC,$A9,$B1,$F8,$C9,$46,$23,$F9,$BB,$5F
	.dc.b	$F0,$A3,$5F,$F1,$8B,$FC,$E3,$B6,$57,$F3,$66,$CF,$56,$BF,$D0,$9A
	.dc.b	$FF,$87,$DA,$FF,$8F,$1A,$BE,$F7,$FD,$BD,$3F,$A2,$8C,$3F,$E5,$28
	.dc.b	$CC,$7E,$99,$AF,$EA,$9A,$FE,$C5,$AF,$F4,$86,$BD,$23,$B4,$60,$8F
	.dc.b	$73,$C1,$94,$76,$D9,$F1,$35,$F1,$35,$F1,$35,$F1,$35,$F8,$1D,$DE
	.dc.b	$D2,$4F,$91,$B1,$EF,$6B,$EF,$6B,$EF,$6B,$EF,$7E,$47,$74,$9F,$7B
	.dc.b	$67,$E5,$35,$F9,$4D,$7E,$53,$5F,$94,$D7,$EC,$77,$7D,$52,$4F,$CC
	.dc.b	$6C,$7D,$76,$BF,$5D,$AF,$D7,$6B,$F5,$DF,$C0,$71,$49,$FA,$ED,$9F
	.dc.b	$C1,$35,$FC,$13,$5F,$C1,$35,$FC,$13,$5F,$C8,$7C,$4F,$E3,$49,$3F
	.dc.b	$84,$6C,$7E,$3D,$AF,$E3,$DA,$FE,$3D,$AF,$E3,$DF,$CC,$77,$49,$FC
	.dc.b	$7B,$67,$F3,$4D,$7F,$34,$D7,$F3,$4D,$7F,$34,$D7,$F5,$1F,$13,$FA
	.dc.b	$52,$4F,$E7,$1B,$1F,$A7,$6B,$FA,$76,$BF,$A7,$6B,$FA,$77,$F6,$1F
	.dc.b	$13,$27,$F4,$ED,$9F,$D9,$35,$FD,$93,$5F,$D9,$35,$FD,$93,$5F,$E0
	.dc.b	$1F,$13,$FB,$D2,$4F,$ED,$1B,$1F,$BF,$6B,$FB,$F6,$BF,$BF,$6B,$FB
	.dc.b	$F7,$F8,$47,$C4,$C9,$FD,$FB,$67,$F8,$53,$5F,$E1,$4D,$7F,$85,$35
	.dc.b	$FE,$14,$D7,$F8,$C7,$87,$FC,$54,$93,$FC,$31,$B1,$FC,$5D,$AF,$F1
	.dc.b	$76,$BF,$C5,$DA,$FF,$17,$7F,$90,$77,$C9,$FE,$2E,$D9,$FE,$44,$D7
	.dc.b	$F9,$13,$5F,$E4,$4D,$7F,$91,$35,$FE,$61,$F1,$3F,$CB,$49,$3F,$C9
	.dc.b	$1B,$1F,$CB,$DA,$FF,$2F,$6B,$FC,$BD,$AF,$F2,$F7,$F9,$C7,$81,$27
	.dc.b	$F9,$7B,$67,$F9,$D3,$5F,$E7,$4D,$7F,$9D,$35,$FE,$74,$D7,$FA,$47
	.dc.b	$4F,$F4,$52,$4F,$F3,$C6,$C7,$F4,$76,$BF,$D1,$DA,$FF,$47,$6B,$FD
	.dc.b	$1D,$FE,$A1,$F1,$1F,$E0,$E8,$FE,$AC,$DF,$0B,$FB,$BC,$62,$BF,$C3
	.dc.b	$0D,$3F,$C1,$A5,$EC,$47,$F9,$16,$78,$F3,$AF,$B3,$67,$9B,$6F,$9D
	.dc.b	$70,$AD,$9E,$6A,$FE,$B6,$D9,$FE,$CC,$D7,$FB,$33,$5F,$EC,$CD,$7F
	.dc.b	$B3,$35,$F4,$47,$F9,$6C,$7B,$FE,$07,$DE,$60,$F3,$CF,$C7,$71,$3F
	.dc.b	$0A,$FD,$58,$7C,$28,$FF,$5D,$CE,$A3,$11,$E0,$33,$7D,$DF,$97,$3B
	.dc.b	$3D,$27,$8A,$35,$3C,$A3,$53,$EC,$74,$79,$8A,$F9,$6F,$D1,$5C,$A7
	.dc.b	$E0,$36,$7C,$71,$C5,$F1,$C6,$5F,$BA,$D8,$F0,$C7,$AC,$CB,$C6,$BC
	.dc.b	$D3,$B2,$46,$4F,$09,$67,$C2,$79,$DF,$4C,$FC,$8B,$F9,$B1,$C2,$74
	.dc.b	$79,$9E,$4F,$88,$CC,$F8,$A3,$27,$A2,$79,$CF,$82,$7B,$5F,$C6,$AF
	.dc.b	$09,$DE,$32,$E2,$78,$39,$3E,$BB,$E3,$24,$78,$5F,$1C,$6D,$E2,$43
	.dc.b	$E5,$3F,$30,$97,$F2,$23,$9C,$71,$77,$1E,$77,$27,$E0,$CA,$7C,$39
	.dc.b	$F1,$1E,$A7,$33,$D8,$F8,$B3,$CE,$77,$1E,$C6,$CF,$9E,$3E,$55,$6D
	.dc.b	$F2,$18,$3E,$34,$78,$5F,$45,$FD,$08,$C1,$CE,$FD,$D7,$DA,$32,$7B
	.dc.b	$AB,$99,$D5,$F4,$CF,$85,$F8,$13,$C6,$7D,$91,$F4,$0E,$CF,$F9,$C4
	.dc.b	$FD,$CB,$FE,$D6,$6B,$F8,$D3,$13,$FB,$87,$04,$3F,$2E,$6A,$3A,$46
	.dc.b	$E3,$26,$CF,$5F,$DE,$FF,$04,$7E,$AB,$6B,$E0,$36,$BE,$F3,$67,$F0
	.dc.b	$CD,$79,$9F,$04,$39,$8F,$DA,$B6,$BF,$9F,$36,$3F,$9D,$B6,$7F,$9D
	.dc.b	$B6,$BF,$8B,$B6,$7F,$E0,$03,$64,$78,$0F,$13,$97,$11,$F8,$28,$C1
	.dc.b	$23,$64,$8D,$34,$F8,$A5,$9F,$8A,$D8,$F8,$AD,$3E,$29,$4F,$8A,$F4
	.dc.b	$23,$D1,$B1,$EE,$8F,$8A,$71,$9F,$9C,$79,$E7,$E3,$C6,$9E,$27,$C4
	.dc.b	$70,$75,$3E,$59,$89,$F4,$C7,$22,$46,$E3,$C5,$18,$77,$1C,$A3,$07
	.dc.b	$C0,$2B,$F4,$A7,$CC,$EC,$93,$C8,$EA,$E2,$79,$34,$E4,$7C,$13,$D0
	.dc.b	$EA,$38,$91,$C1,$F0,$4F,$1C,$7D,$03,$C4,$7E,$18,$EA,$FF,$B8,$67
	.dc.b	$F4,$51,$93,$F0,$E5,$29,$97,$6F,$73,$4E,$1E,$A6,$0F,$D9,$1C,$8E
	.dc.b	$07,$33,$FC,$3D,$F1,$5E,$E7,$F3,$79,$23,$10,$8C,$57,$B9,$B1,$F2
	.dc.b	$4D,$4F,$08,$CD,$FE,$AB,$D8,$FD,$89,$FE,$2F,$F3,$9E,$66,$31,$FD
	.dc.b	$9C,$DC,$7F,$C5,$B5,$A9,$C9,$23,$F6,$F3,$CE,$B8,$5F,$7A,$F9,$D3
	.dc.b	$F0,$B2,$34,$91,$A2,$5F,$CC,$8C,$B6,$32,$FD,$38,$CD,$7F,$06,$C3
	.dc.b	$C3,$5F,$C1,$9B,$5F,$95,$6B,$FC,$59,$B6,$69,$27,$DD,$5D,$5A,$75
	.dc.b	$2D,$75,$6B,$D5,$B5,$D5,$B5,$D6,$7E,$4C,$7F,$22,$8C,$57,$C9,$9F
	.dc.b	$E4,$71,$89,$FE,$43,$7F,$CA,$A3,$13,$FB,$18,$C4,$F6,$24,$76,$39
	.dc.b	$B8,$AE,$C7,$B6,$34,$D3,$31,$EF,$8C,$B6,$33,$1F,$04,$EA,$47,$2F
	.dc.b	$C3,$31,$1C,$4F,$59,$F0,$48,$E8,$F6,$38,$8F,$23,$C6,$32,$6C,$ED
	.dc.b	$1A,$72,$D7,$31,$C5,$FE,$AF,$3F,$21,$E5,$7F,$5D,$F7,$9E,$DA,$E5
	.dc.b	$19,$7D,$AD,$AE,$8D,$9F,$0C,$62,$FB,$57,$06,$C7,$0B,$B5,$DE,$30
	.dc.b	$7C,$13,$91,$F6,$4A,$F9,$8F,$69,$C4,$87,$B9,$E4,$E2,$FC,$CE,$4F
	.dc.b	$83,$5E,$27,$17,$F0,$23,$13,$FB,$AA,$E7,$5F,$74,$FE,$4D,$3F,$BF
	.dc.b	$24,$7F,$2F,$6C,$7F,$28,$6B,$C2,$34,$47,$4D,$AF,$F4,$CF,$C6,$BA
	.dc.b	$BC,$CD,$9F,$1B,$6B,$F0,$8D,$AF,$C1,$36,$7F,$04,$D9,$F8,$ED,$7E
	.dc.b	$58,$71,$87,$C8,$6E,$1F,$ED,$96,$D7,$FB,$59,$B5,$FE,$D6,$6D,$FA
	.dc.b	$B0,$F4,$92,$BD,$2D,$AF,$4B,$67,$D2,$D9,$F4,$B6,$7D,$2D,$7D,$35
	.dc.b	$E0,$9F,$EC,$4D,$8F,$F1,$4B,$5F,$74,$F1,$C3,$DC,$DC,$3D,$CD,$BF
	.dc.b	$73,$DE,$FE,$41,$CC,$AF,$08,$F6,$9C,$09,$7E,$E6,$E1,$F1,$5D,$C7
	.dc.b	$C9,$9E,$4E,$CC,$47,$03,$E1,$95,$F6,$C7,$94,$E3,$3F,$2A,$75,$3F
	.dc.b	$01,$EF,$1E,$18,$EA,$77,$70,$48,$EF,$1A,$8C,$3F,$4A,$3A,$D7,$C3
	.dc.b	$AF,$0C,$7D,$27,$C8,$F8,$B0,$F9,$EF,$C4,$AF,$F0,$A5,$7E,$6D,$B5
	.dc.b	$FE,$0C,$8C,$47,$39,$FF,$0C,$3E,$B3,$FC,$35,$1A,$7E,$1C,$6A,$31
	.dc.b	$1B,$3A,$C6,$67,$B0,$7F,$CD,$F7,$67,$FC,$E3,$FA,$63,$D8,$F6,$C6
	.dc.b	$1A,$A3,$55,$F4,$1B,$1F,$26,$3F,$C6,$D1,$8A,$F9,$73,$C6,$BF,$5D
	.dc.b	$D7,$F0,$49,$7E,$B6,$CF,$C0,$6B,$F2,$9A,$FE,$89,$AF,$E8,$9F,$99
	.dc.b	$D7,$27,$E3,$36,$7E,$5B,$5F,$AA,$D7,$F5,$AD,$7F,$5A,$D7,$F8,$57
	.dc.b	$5F,$D9,$24,$FD,$36,$C7,$D8,$6B,$F8,$26,$BF,$BA,$6B,$FB,$A7,$F0
	.dc.b	$9D,$72,$7E,$F3,$67,$F0,$6D,$7F,$1A,$D7,$F8,$2B,$5F,$E0,$AD,$7F
	.dc.b	$91,$75,$FE,$44,$93,$F8,$B6,$C7,$E4,$1A,$FE,$69,$AF,$F1,$26,$BF
	.dc.b	$C4,$9F,$CE,$75,$C9,$FC,$B3,$67,$F3,$6D,$7F,$4A,$D7,$F8,$EB,$5F
	.dc.b	$E3,$AD,$7F,$9D,$75,$7F,$37,$3F,$35,$A7,$E6,$E3,$FC,$DD,$F1,$99
	.dc.b	$BF,$E7,$55,$65,$5B,$E5,$5E,$63,$FA,$AD,$5B,$FF,$48,$56,$9E,$64
	.dc.b	$BF,$DA,$CF,$EE,$C9,$1F,$B5,$6C,$6A,$30,$47,$04,$3F,$83,$C7,$0A
	.dc.b	$F6,$7D,$F3,$E2,$7C,$95,$FE,$E8,$8C,$5F,$8E,$B7,$5A,$A5,$69,$FE
	.dc.b	$B5,$7F,$22,$FF,$D5,$93,$9B,$F8,$31,$FC,$18,$CD,$F3,$9D,$47,$F5
	.dc.b	$2A,$E5,$7E,$3A,$D1,$27,$C8,$72,$2C,$6D,$AF,$9A,$38,$3E,$68,$F1
	.dc.b	$3B,$30,$7D,$A7,$A3,$8B,$F8,$31,$D2,$FC,$F3,$FD,$9A,$FC,$EF,$79
	.dc.b	$D5,$F8,$2F,$E0,$CF,$53,$C0,$E0,$EA,$F7,$9F,$0C,$65,$A7,$C5,$3D
	.dc.b	$31,$DC,$E0,$74,$74,$F7,$35,$1C,$DD,$1E,$98,$E7,$7E,$C8,$DC,$78
	.dc.b	$1D,$1C,$E3,$44,$AF,$74,$EE,$78,$9C,$2B,$DA,$48,$D7,$1B,$0F,$93
	.dc.b	$87,$39,$F1,$E1,$D3,$0F,$39,$D0,$FD,$BC,$68,$8F,$96,$38,$3B,$9F
	.dc.b	$65,$F7,$AD,$E1,$E3,$AE,$B3,$97,$94,$E8,$F3,$E1,$E6,$C2,$46,$67
	.dc.b	$DC,$FC,$1C,$3D,$51,$F5,$0F,$14,$F0,$71,$1A,$C3,$C5,$1E,$C3,$C2
	.dc.b	$47,$38,$79,$5D,$CF,$1B,$FB,$F1,$8B,$EC,$FD,$FB,$F0,$CF,$9D,$ED
	.dc.b	$1B,$8F,$F0,$2D,$7E,$EE,$47,$E8,$8A,$7E,$9C,$A7,$E4,$4A,$F1,$78
	.dc.b	$10,$E0,$53,$81,$5F,$C5,$FC,$1F,$A2,$7F,$84,$B6,$7F,$8A,$35,$FE
	.dc.b	$22,$D9,$34,$93,$EF,$8C,$B6,$33,$3F,$36,$3B,$99,$23,$93,$C2,$E2
	.dc.b	$BE,$8C,$F2,$6C,$64,$FC,$14,$65,$A6,$67,$F3,$75,$E4,$7B,$1E,$C8
	.dc.b	$D3,$F9,$99,$E4,$F0,$6B,$98,$DC,$FC,$A6,$DF,$EB,$2B,$73,$DE,$FF
	.dc.b	$A0,$C6,$0D,$57,$26,$DF,$92,$B8,$4F,$C6,$8F,$E0,$53,$C6,$3F,$84
	.dc.b	$B6,$33,$1C,$CC,$95,$C9,$D6,$3B,$D7,$92,$74,$D9,$D4,$72,$78,$C7
	.dc.b	$D0,$6C,$79,$63,$07,$12,$3F,$4A,$36,$7D,$C3,$A9,$83,$E9,$18,$8F
	.dc.b	$13,$F9,$E9,$FF,$02,$3F,$0C,$C4,$FE,$AA,$33,$1D,$4C,$9F,$A8,$3A
	.dc.b	$98,$21,$CC,$87,$96,$7F,$C2,$5F,$9E,$3D,$2F,$8F,$7F,$B4,$8C,$3F
	.dc.b	$16,$77,$3A,$AF,$DD,$D6,$9F,$73,$67,$E2,$C6,$A7,$F9,$9F,$C1,$4B
	.dc.b	$FE,$9F,$FA,$8F,$F8,$5E,$BF,$A1,$4E,$5B,$3A,$DF,$18,$7F,$83,$AA
	.dc.b	$CF,$F0,$AB,$E5,$39,$8F,$EA,$33,$F2,$2F,$FC,$1A,$DC,$CD,$24,$FF
	.dc.b	$18,$6C,$FC,$D9,$DC,$E4,$91,$99,$DB,$FC,$61,$FE,$F3,$F0,$6F,$77
	.dc.b	$CF,$3F,$C8,$A3,$15,$E6,$3F,$B8,$B8,$8F,$82,$7F,$6F,$71,$87,$B6
	.dc.b	$31,$5D,$14,$EA,$BF,$C4,$F7,$E2,$AD,$4E,$6B,$FC,$69,$39,$AF,$14
	.dc.b	$74,$6D,$FF,$95,$EB,$84,$FC,$A9,$F3,$DF,$42,$3C,$0F,$E7,$B3,$C3
	.dc.b	$0F,$93,$5D,$4F,$6D,$F5,$BE,$31,$8B,$E3,$18,$9F,$80,$6E,$3B,$4F
	.dc.b	$CB,$9D,$47,$C0,$3C,$A6,$A3,$E2,$C6,$9F,$83,$1A,$7C,$07,$37,$85
	.dc.b	$FA,$70,$F8,$D3,$E9,$BF,$13,$C0,$EB,$5C,$E3,$27,$A1,$F5,$99,$2C
	.dc.b	$71,$AF,$9A,$4B,$F3,$47,$43,$C1,$3E,$D6,$9E,$72,$9F,$54,$B3,$CE
	.dc.b	$3D,$73,$EA,$AF,$2C,$73,$6D,$73,$6B,$CD,$B5,$CE,$3D,$15,$CE,$BD
	.dc.b	$75,$EF,$9E,$51,$82,$3F,$40,$F2,$CF,$17,$13,$DA,$31,$1E,$19,$E2
	.dc.b	$FC,$39,$F2,$E1,$F8,$C7,$84,$FC,$D9,$E6,$F0,$BF,$44,$F9,$67,$94
	.dc.b	$E6,$FE,$7D,$7E,$C6,$7F,$C2,$6D,$7E,$B9,$88,$FE,$19,$19,$8F,$A4
	.dc.b	$66,$7F,$9A,$3F,$64,$C1,$FC,$DE,$38,$39,$3A,$9E,$EF,$C0,$1E,$5F
	.dc.b	$C2,$6D,$7B,$65,$18,$7F,$02,$DC,$3E,$CB,$67,$A5,$73,$23,$CE,$35
	.dc.b	$5F,$60,$FF,$1B,$B8,$7E,$39,$89,$E2,$F8,$1B,$5E,$06,$D7,$80,$EC
	.dc.b	$E2,$7B,$1E,$A7,$0F,$63,$11,$C0,$8F,$76,$CF,$76,$9D,$E3,$44,$8D
	.dc.b	$36,$35,$18,$23,$FE,$54,$AF,$80,$7C,$1B,$FE,$42,$E2,$FF,$D6,$AE
	.dc.b	$70,$B1,$2B,$FD,$57,$39,$6C,$67,$0F,$F5,$54,$E5,$B1,$9C,$15,$E3
	.dc.b	$8F,$8B,$5E,$18,$F7,$D7,$D3,$7A,$98,$7E,$49,$D1,$C4,$71,$8F,$84
	.dc.b	$71,$34,$60,$E3,$5E,$27,$4F,$C6,$6C,$F9,$4E,$B5,$E9,$7E,$69,$98
	.dc.b	$EB,$3C,$5E,$B1,$98,$C5,$7E,$6E,$7F,$D0,$6D,$79,$18,$AF,$EF,$91
	.dc.b	$98,$F5,$19,$AF,$D8,$C6,$63,$E1,$99,$3F,$C5,$0E,$49,$3F,$EF,$0F
	.dc.b	$D4,$DC,$FE,$BA,$FF,$4C,$C6,$2F,$E7,$C7,$FB,$8E,$30,$FB,$A3,$FD
	.dc.b	$3B,$18,$C3,$E5,$4F,$C0,$39,$D7,$55,$3B,$AF,$C9,$79,$7E,$E2,$3E
	.dc.b	$C6,$CF,$A9,$AF,$C8,$6B,$CA,$38,$11,$F8,$3E,$5F,$CA,$24,$7C,$86
	.dc.b	$C6,$A3,$04,$70,$43,$EB,$C7,$07,$EC,$F7,$FD,$62,$4F,$D4,$6C,$7D
	.dc.b	$26,$BC,$23,$44,$74,$D7,$F0,$BE,$5F,$E0,$89,$1F,$80,$6C,$7D,$E6
	.dc.b	$BC,$23,$44,$74,$D7,$FD,$AB,$E7,$BD,$07,$9D,$FA,$3D,$7F,$49,$39
	.dc.b	$8C,$5F,$FB,$7A,$38,$67,$8C,$11,$B9,$FF,$71,$DF,$2C,$3A,$C6,$0E
	.dc.b	$4E,$29,$3B,$BE,$97,$CE,$73,$7C,$30,$CE,$1C,$70,$E1,$3D,$6B,$C7
	.dc.b	$39,$8E,$06,$6F,$A5,$70,$AE,$A7,$0B,$EA,$4A,$E1,$6C,$31,$5C,$30
	.dc.b	$E7,$5D,$A3,$15,$B8,$E3,$7F,$CB,$3B,$FF,$7A,$48,$FD,$DB,$6B,$F7
	.dc.b	$2D,$7F,$83,$B5,$FE,$74,$D7,$FA,$DB,$FC,$23,$69,$3F,$C1,$9B,$3F
	.dc.b	$C1,$1A,$FF,$12,$6B,$FD,$15,$AF,$F6,$66,$BF,$DD,$3B,$BF,$A2,$9F
	.dc.b	$4B,$6B,$5F,$D1,$22,$CF,$C2,$9B,$A7,$F0,$6E,$15,$62,$FF,$07,$56
	.dc.b	$FF,$BA,$CF,$F2,$22,$4F,$F2,$06,$C7,$F1,$E6,$BF,$CB,$1A,$FF,$5F
	.dc.b	$6B,$D2,$39,$1F,$87,$5F,$80,$BE,$43,$C5,$18,$6C,$66,$7D,$C7,$17
	.dc.b	$13,$FD,$82,$BF,$48,$F1,$9D,$BF,$8E,$30,$6D,$E5,$5D,$88,$F0,$C7
	.dc.b	$F9,$6D,$F6,$7E,$55,$F6,$8C,$4F,$BC,$E6,$71,$17,$23,$6F,$C3,$3C
	.dc.b	$8E,$1F,$BD,$1C,$9F,$31,$AB,$F2,$B8,$BF,$2C,$61,$F2,$CF,$16,$CF
	.dc.b	$18,$F2,$3E,$39,$E1,$1E,$98,$D1,$D5,$F4,$BF,$22,$79,$9E,$08,$E7
	.dc.b	$19,$79,$46,$63,$91,$DE,$37,$5E,$F2,$57,$8A,$74,$78,$AB,$BD,$78
	.dc.b	$AB,$6F,$B4,$EE,$66,$BE,$A1,$1E,$ED,$AE,$91,$C1,$EF,$5E,$72,$46
	.dc.b	$63,$D8,$69,$EC,$74,$3C,$4E,$4F,$B9,$1C,$9C,$1C,$8F,$BA,$78,$8C
	.dc.b	$D7,$E0,$71,$FD,$AD,$74,$9E,$13,$9C,$3D,$57,$D6,$7A,$5F,$07,$A1
	.dc.b	$C3,$0F,$98,$DC,$3D,$0E,$B0,$E9,$1D,$DF,$9B,$1F,$60,$E0,$74,$AE
	.dc.b	$D7,$E1,$9D,$BF,$5A,$7D,$77,$FE,$14,$7E,$39,$F8,$A3,$EE,$96,$7F
	.dc.b	$72,$D8,$E0,$E8,$E2,$66,$FF,$D4,$33,$C2,$32,$43,$81,$F3,$3E,$2B
	.dc.b	$E3,$CF,$E2,$5B,$1F,$B3,$6B,$FC,$49,$B2,$69,$27,$E3,$36,$7E,$53
	.dc.b	$6F,$E2,$36,$7E,$0C,$7F,$0A,$8C,$4F,$96,$79,$91,$E6,$D9,$E6,$D3
	.dc.b	$99,$EB,$31,$1C,$CE,$24,$8E,$2D,$8E,$2F,$C1,$78,$BF,$8E,$9F,$2B
	.dc.b	$63,$35,$FB,$79,$FE,$40,$48,$CB,$67,$B1,$D4,$FE,$3A,$F4,$9F,$99
	.dc.b	$3F,$CE,$63,$07,$1A,$F3,$9F,$1E,$34,$76,$2D,$76,$3E,$C3,$88,$E2
	.dc.b	$FC,$58,$F0,$BC,$5A,$71,$AF,$1B,$89,$E2,$FA,$5F,$13,$BA,$F3,$57
	.dc.b	$CA,$8F,$09,$C5,$F0,$4F,$28,$CB,$C8,$C5,$78,$23,$F4,$55,$FA,$07
	.dc.b	$99,$8A,$FC,$FC,$66,$78,$4F,$B4,$D1,$FA,$D3,$89,$92,$1F,$4C,$DD
	.dc.b	$7F,$91,$78,$BD,$71,$FE,$38,$6F,$07,$F9,$55,$BC,$06,$93,$1F,$EE
	.dc.b	$5F,$15,$FD,$8C,$FF,$22,$F1,$D8,$C9,$30,$FF,$1A,$C6,$5F,$EC,$33
	.dc.b	$71,$FF,$2C,$4D,$AF,$E8,$13,$F0,$48,$FF,$02,$6D,$7C,$1C,$91,$88
	.dc.b	$FE,$1C,$D7,$9C,$F1,$23,$C6,$3F,$C1,$3F,$80,$BE,$1B,$C5,$3F,$DE
	.dc.b	$A3,$15,$EB,$C7,$FD,$63,$18,$9F,$5C,$FF,$87,$63,$13,$FE,$0A,$BE
	.dc.b	$B7,$D2,$D5,$AB,$F2,$C7,$F8,$5E,$30,$FF,$82,$6B,$8C,$F8,$4F,$EE
	.dc.b	$B3,$A9,$FF,$15,$CE,$9F,$19,$CE,$BC,$13,$BC,$3E,$3D,$F9,$E7,$24
	.dc.b	$7B,$57,$C1,$8C,$3D,$30,$F8,$31,$87,$DD,$3C,$CE,$33,$EE,$9D,$47
	.dc.b	$BA,$78,$BA,$BF,$26,$1E,$99,$E7,$7E,$9A,$D5,$F2,$8F,$89,$1D,$DF
	.dc.b	$23,$F1,$23,$93,$4E,$26,$E7,$C3,$3C,$0F,$74,$64,$E0,$E4,$F5,$39
	.dc.b	$76,$66,$FE,$91,$2B,$CD,$1C,$0F,$51,$B7,$CA,$F2,$AF,$79,$E4,$25
	.dc.b	$78,$9C,$B4,$91,$C5,$F9,$86,$1F,$04,$72,$7E,$39,$E2,$70,$EA,$38
	.dc.b	$BC,$C8,$64,$E3,$1C,$2B,$DD,$3F,$32,$7A,$3E,$1C,$3F,$1B,$18,$39
	.dc.b	$CF,$DE,$9F,$95,$1D,$49,$3C,$2B,$B5,$FC,$69,$FC,$ED,$70,$BF,$0C
	.dc.b	$7C,$B8,$FC,$13,$FE,$0F,$8F,$DA,$36,$7F,$97,$46,$23,$EE,$99,$9F
	.dc.b	$E5,$4F,$D7,$30,$7F,$2D,$8E,$0F,$E0,$0C,$90,$FA,$0B,$D5,$F4,$36
	.dc.b	$7D,$11,$82,$38,$24,$FC,$73,$F8,$2C,$69,$C4,$F3,$6C,$66,$FE,$A3
	.dc.b	$6B,$C6,$D9,$FA,$2D,$7E,$85,$73,$AE,$35,$E2,$6C,$F6,$8C,$1D,$0A
	.dc.b	$F4,$6B,$D2,$BB,$57,$E9,$1B,$7F,$CB,$9B,$5F,$C9,$DD,$99,$79,$CF
	.dc.b	$95,$FE,$3F,$1A,$78,$1A,$78,$57,$9C,$FE,$54,$E2,$FD,$67,$F3,$87
	.dc.b	$13,$EB,$9D,$BD,$1E,$53,$B8,$C3,$C9,$F1,$98,$39,$3B,$8E,$67,$28
	.dc.b	$DB,$88,$E4,$FF,$4D,$9F,$1B,$F2,$CD,$4F,$09,$DC,$7E,$12,$3C,$0E
	.dc.b	$E3,$C0,$48,$F1,$9E,$08,$FD,$24,$FE,$88,$91,$FA,$0A,$FC,$64,$EA
	.dc.b	$31,$3A,$8F,$D1,$35,$D1,$88,$D9,$FE,$E0,$A3,$FE,$E2,$D5,$AB,$C8
	.dc.b	$F7,$C6,$1B,$19,$9F,$6C,$E5,$B1,$9B,$F8,$A2,$74,$FF,$BF,$FE,$4D
	.dc.b	$D5,$FE,$2D,$6D,$FE,$F9,$B3,$FE,$29,$6B,$FE,$97,$6B,$FE,$E2,$6B
	.dc.b	$FE,$FC,$6A,$69,$27,$C0,$49,$F0,$74,$78,$C8,$E0,$95,$E3,$8C,$B6
	.dc.b	$32,$F9,$63,$2D,$8F,$67,$4F,$BC,$8E,$09,$1E,$F8,$CB,$63,$2F,$C1
	.dc.b	$8C,$B6,$3E,$47,$3F,$CB,$23,$82,$47,$CB,$8C,$B6,$32,$FC,$D8,$CB
	.dc.b	$63,$EA,$71,$FD,$72,$38,$24,$7D,$78,$CB,$63,$2F,$D9,$8C,$B6,$3F
	.dc.b	$01,$C7,$F8,$32,$38,$24,$7E,$0E,$32,$D8,$CB,$F8,$58,$CB,$63,$F1
	.dc.b	$9E,$4F,$E8,$23,$F2,$4D,$DF,$1A,$FE,$8A,$31,$5F,$8D,$6D,$7E,$35
	.dc.b	$AF,$E3,$5C,$CF,$19,$FD,$71,$C4,$B3,$C5,$AF,$17,$31,$C6,$3C,$B1
	.dc.b	$AD,$55,$FA,$72,$47,$B0,$93,$FA,$78,$CB,$63,$2F,$EA,$63,$2D,$8E
	.dc.b	$37,$F1,$63,$F5,$1E,$0F,$BA,$FC,$35,$E3,$8C,$5E,$5B,$19,$3D,$11
	.dc.b	$96,$99,$AF,$69,$FC,$0E,$78,$B6,$36,$D3,$77,$E5,$B5,$70,$8C,$36
	.dc.b	$BF,$82,$36,$BF,$82,$35,$F6,$C7,$69,$E2,$D3,$89,$63,$8B,$5E,$27
	.dc.b	$91,$DB,$4D,$94,$DC,$7A,$6F,$E2,$46,$6F,$D7,$5D,$5B,$19,$BF,$44
	.dc.b	$62,$F9,$57,$CB,$6D,$7C,$76,$9E,$C7,$94,$6C,$F3,$C6,$9A,$48,$F4
	.dc.b	$C6,$5B,$19,$7D,$6F,$E7,$5F,$03,$99,$F3,$39,$9F,$11,$1D,$5F,$C2
	.dc.b	$8C,$57,$63,$F4,$C7,$43,$07,$80,$CB,$C1,$C1,$2B,$F2,$EF,$13,$17
	.dc.b	$F2,$8F,$05,$F8,$8F,$89,$1D,$A7,$D2,$F6,$8F,$64,$F0,$8C,$1B,$70
	.dc.b	$FE,$EC,$F0,$4F,$27,$C9,$7A,$8C,$11,$C4,$E5,$B5,$E4,$AE,$97,$EF
	.dc.b	$BE,$37,$EB,$AF,$D6,$57,$F6,$58,$FE,$D8,$48,$FE,$D8,$D8,$FE,$D8
	.dc.b	$D7,$84,$68,$8E,$9B,$5F,$C1,$7F,$01,$7C,$7F,$C0,$6C,$F9,$DB,$5F
	.dc.b	$45,$B5,$F5,$DB,$3F,$8B,$6C,$FD,$E6,$BF,$71,$B8,$7C,$B6,$E1,$FC
	.dc.b	$C9,$B5,$FC,$B1,$B4,$BF,$46,$1E,$72,$57,$9D,$B3,$E7,$6C,$F9,$DB
	.dc.b	$3E,$76,$CF,$9D,$AF,$9D,$B8,$79,$DB,$87,$9D,$B5,$E7,$6D,$79,$E0
	.dc.b	$CC,$3D,$78,$7A,$A7,$D0,$49,$E0,$6E,$BB,$46,$67,$84,$E8,$57,$49
	.dc.b	$EF,$18,$AE,$2E,$E7,$4F,$59,$F4,$11,$F9,$0D,$AE,$E7,$C9,$70,$76
	.dc.b	$9E,$73,$DE,$32,$F5,$39,$9E,$1A,$F1,$D7,$38,$C4,$F0,$7E,$83,$6B
	.dc.b	$A9,$F8,$A2,$57,$D6,$6C,$7C,$53,$C9,$1D,$9F,$BC,$73,$78,$19,$9F
	.dc.b	$C0,$47,$C1,$7E,$49,$EB,$29,$4C,$CF,$E9,$1B,$3F,$DB,$DF,$0C,$FF
	.dc.b	$B7,$A7,$51,$83,$9B,$87,$E3,$1C,$A3,$84,$F3,$8F,$95,$5F,$E1,$CE
	.dc.b	$17,$6F,$A7,$4F,$F3,$FB,$63,$38,$7D,$29,$CB,$63,$37,$F4,$0F,$C7
	.dc.b	$CE,$9F,$C3,$7B,$9F,$B3,$3F,$C4,$72,$EE,$C5,$FE,$D6,$6D,$72,$AD
	.dc.b	$4E,$49,$1D,$1F,$F1,$2C,$6A,$30,$43,$E6,$CF,$C1,$7F,$3A,$DC,$CD
	.dc.b	$25,$7C,$F6,$CF,$CE,$6B,$D2,$34,$47,$F5,$5F,$80,$BE,$37,$D7,$A7
	.dc.b	$F8,$A6,$31,$5E,$A8,$C4,$EA,$3F,$5F,$3A,$AF,$5D,$6A,$78,$1A,$AE
	.dc.b	$13,$FC,$12,$B8,$57,$F3,$3B,$FF,$29,$C6,$2F,$F5,$D1,$8B,$E3,$6A
	.dc.b	$D3,$FC,$36,$75,$5E,$08,$ED,$5C,$2B,$71,$F3,$27,$FC,$B3,$18,$C3
	.dc.b	$E6,$C6,$2B,$81,$1E,$95,$E1,$8C,$3F,$02,$3D,$C6,$AB,$A5,$F9,$27
	.dc.b	$47,$92,$75,$5C,$B0,$F8,$51,$E1,$32,$49,$E2,$78,$5C,$47,$CF,$8F
	.dc.b	$B7,$3C,$E7,$C7,$18,$76,$62,$78,$1E,$08,$D1,$1C,$C7,$D2,$3C,$91
	.dc.b	$97,$91,$97,$07,$B4,$A7,$CF,$2D,$FD,$23,$E2,$C7,$49,$EF,$3C,$5F
	.dc.b	$21,$87,$E0,$1D,$30,$FA,$EF,$4B,$F9,$D3,$C2,$BC,$B3,$CD,$AF,$8E
	.dc.b	$3A,$5F,$8E,$BE,$44,$EC,$86,$EB,$8C,$ED,$F9,$71,$A8,$F5,$BE,$38
	.dc.b	$F6,$1F,$16,$BB,$D7,$9A,$30,$74,$71,$3F,$9B,$8E,$2F,$CB,$3F,$9A
	.dc.b	$D7,$EB,$DB,$3F,$C1,$E3,$11,$F8,$03,$33,$FC,$E9,$F9,$26,$0F,$E7
	.dc.b	$B1,$C1,$EC,$71,$21,$ED,$F5,$3C,$AF,$EC,$5B,$06,$92,$7D,$33,$B6
	.dc.b	$CE,$DB,$3B,$9F,$C2,$B6,$FE,$AB,$67,$E8,$B6,$7C,$91,$C8,$8F,$26
	.dc.b	$CF,$26,$9C,$8B,$1C,$8E,$04,$8E,$0D,$8E,$0D,$38,$46,$89,$1A,$6C
	.dc.b	$7F,$1F,$9F,$E3,$A7,$A5,$F8,$87,$F3,$27,$15,$F1,$63,$15,$E6,$77
	.dc.b	$5F,$CD,$6B,$67,$49,$D5,$71,$9D,$46,$27,$55,$E9,$39,$C7,$F3,$A8
	.dc.b	$E6,$FF,$24,$6C,$78,$8E,$15,$D6,$78,$1D,$4E,$05,$8E,$15,$F0,$9E
	.dc.b	$E7,$C3,$8D,$47,$9A,$3A,$D7,$E6,$C9,$3F,$99,$AF,$89,$3F,$A2,$9C
	.dc.b	$BF,$9A,$6B,$99,$E1,$18,$8F,$39,$B8,$FF,$2A,$D1,$FF,$2B,$E8,$FD
	.dc.b	$01,$F2,$BF,$22,$B5,$6A,$D4,$E6,$F8,$DF,$07,$E1,$D7,$F8,$3A,$30
	.dc.b	$FC,$48,$DC,$F1,$8E,$93,$AB,$FE,$E5,$F0,$FF,$BD,$25,$FC,$36,$D7
	.dc.b	$EE,$9A,$FF,$06,$6B,$FD,$D9,$AF,$F7,$17,$F8,$47,$BF,$27,$E7,$36
	.dc.b	$7F,$82,$B5,$FE,$22,$D7,$FC,$02,$D7,$FB,$E3,$5F,$EE,$DD,$BF,$C5
	.dc.b	$49,$3F,$6D,$B1,$FC,$49,$AF,$F1,$E6,$BF,$E0,$E6,$BF,$E0,$97,$FC
	.dc.b	$4D,$F8,$13,$B7,$EA,$E3,$F5,$A9,$39,$BF,$BE,$7F,$A5,$3E,$29,$EB
	.dc.b	$7F,$81,$27,$55,$FE,$8D,$6D,$F3,$6D,$F1,$AE,$18,$78,$23,$FD,$23
	.dc.b	$3A,$BF,$F5,$0B,$70,$EF,$C3,$B9,$CD,$D7,$CE,$FB,$09,$7F,$CE,$6F
	.dc.b	$CB,$3F,$E7,$66,$BE,$09,$DB,$FE,$78,$8C,$BD,$A7,$94,$EA,$39,$3C
	.dc.b	$31,$F9,$0B,$E5,$3C,$97,$EB,$AF,$9D,$18,$9F,$41,$BB,$F9,$B3,$DC
	.dc.b	$FF,$5B,$46,$5F,$98,$F0,$9E,$53,$C2,$7A,$B6,$BC,$15,$C6,$FE,$93
	.dc.b	$C6,$74,$DA,$EB,$1C,$E3,$83,$F1,$4C,$5F,$82,$7C,$AD,$AF,$82,$D9
	.dc.b	$FF,$65,$D7,$2A,$F6,$56,$A7,$C2,$74,$2C,$74,$6C,$6E,$3C,$A6,$4A
	.dc.b	$E4,$F6,$C7,$38,$CB,$E2,$AF,$24,$7F,$BB,$2B,$C8,$FE,$22,$78,$57
	.dc.b	$28,$EC,$D3,$B4,$7C,$38,$CB,$63,$31,$EE,$8F,$01,$EA,$2B,$E4,$3C
	.dc.b	$55,$E4,$7E,$C9,$92,$CF,$86,$78,$B6,$38,$C7,$56,$9C,$6B,$DB,$7F
	.dc.b	$82,$BE,$D1,$83,$F0,$17,$CE,$BB,$DF,$AC,$EE,$57,$AC,$6E,$79,$4F
	.dc.b	$C8,$35,$3C,$8F,$84,$70,$73,$1C,$23,$E4,$9A,$38,$1F,$4C,$C9,$67
	.dc.b	$E6,$B6,$7C,$AF,$19,$E9,$3A,$BF,$81,$5F,$16,$BA,$C6,$1F,$8D,$7F
	.dc.b	$10,$93,$BB,$E6,$F6,$9F,$01,$EB,$3C,$13,$C1,$E7,$1C,$63,$17,$F1
	.dc.b	$DB,$1F,$09,$B3,$CA,$38,$11,$E6,$49,$E5,$19,$6C,$66,$3A,$19,$3D
	.dc.b	$15,$EA,$8F,$41,$F3,$9C,$36,$BF,$07,$3C,$A3,$E8,$CE,$8F,$D3,$92
	.dc.b	$7D,$31,$C6,$7C,$0F,$69,$E1,$3D,$23,$C3,$7F,$46,$FE,$DD,$6A,$FF
	.dc.b	$15,$5A,$8E,$0F,$53,$DC,$75,$BF,$0C,$78,$E7,$C3,$3B,$AF,$8A,$76
	.dc.b	$AD,$9D,$2F,$C5,$5F,$1A,$32,$F8,$AB,$EF,$9C,$DF,$C1,$4F,$08,$F1
	.dc.b	$47,$82,$FC,$51,$E6,$38,$1F,$74,$F9,$75,$D9,$FC,$ED,$73,$76,$78
	.dc.b	$27,$9D,$FE,$6D,$F8,$66,$2F,$EB,$47,$B5,$F1,$5F,$B6,$76,$D8,$F6
	.dc.b	$47,$1A,$F2,$57,$17,$D1,$19,$7C,$55,$CA,$BA,$9F,$16,$FD,$15,$B6
	.dc.b	$CF,$7A,$F7,$D6,$EB,$94,$7C,$06,$A3,$F9,$29,$4F,$E3,$65,$3F,$92
	.dc.b	$95,$FE,$C8,$D9,$FF,$0C,$3C,$88,$72,$29,$C8,$B1,$FD,$0B,$F0,$27
	.dc.b	$B7,$F9,$06,$DF,$E0,$DB,$5F,$C5,$1B,$8F,$F1,$46,$CF,$F0,$76,$CF
	.dc.b	$EC,$1A,$FF,$1E,$6D,$F8,$6F,$C0,$4A,$F0,$36,$BC,$0D,$BF,$03,$67
	.dc.b	$C0,$D9,$F0,$35,$F0,$36,$BC,$13,$D8,$93,$D9,$B5,$D9,$A7,$62,$C7
	.dc.b	$66,$BD,$9B,$3D,$A3,$A1,$27,$A3,$67,$A3,$4E,$85,$3A,$16,$7A,$4F
	.dc.b	$22,$4F,$26,$CF,$26,$BC,$9B,$3C,$A7,$81,$27,$83,$5E,$0D,$8E,$11
	.dc.b	$A2,$1A,$2B,$A7,$04,$70,$4C,$3F,$CB,$58,$7E,$11,$B8,$7D,$76,$DF
	.dc.b	$CE,$9F,$C5,$9B,$7F,$C8,$4E,$27,$77,$F2,$63,$13,$C6,$73,$87,$C4
	.dc.b	$9C,$D6,$E7,$35,$F0,$67,$35,$B9,$CD,$7C,$09,$CD,$6E,$73,$3F,$02
	.dc.b	$72,$EE,$33,$87,$C5,$9C,$CE,$E7,$38,$72,$9C,$D6,$E7,$37,$F0,$1D
	.dc.b	$1F,$9F,$9D,$D6,$27,$47,$C1,$2D,$FB,$DB,$87,$B5,$A7,$AC,$B1,$E9
	.dc.b	$6B,$E7,$6E,$1E,$56,$D7,$86,$55,$FE,$AD,$BF,$B1,$18,$AF,$E4,$11
	.dc.b	$F0,$8C,$CE,$F0,$E4,$F9,$6B,$81,$FC,$A6,$77,$5C,$AB,$6F,$A6,$76
	.dc.b	$27,$6F,$19,$DB,$ED,$37,$1A,$78,$46,$9F,$80,$69,$FA,$07,$D4,$71
	.dc.b	$3F,$1A,$79,$1E,$D9,$DC,$7C,$63,$4F,$1A,$F8,$11,$83,$EC,$57,$F4
	.dc.b	$33,$C6,$D6,$99,$AF,$E6,$EF,$F4,$96,$BF,$16,$38,$46,$23,$E1,$46
	.dc.b	$2B,$FA,$5C,$60,$E8,$E1,$EE,$75,$7F,$DA,$5F,$80,$BD,$3F,$D0,$B6
	.dc.b	$BF,$0E,$DA,$F8,$A7,$D4,$71,$87,$C0,$8C,$57,$22,$57,$26,$C7,$29
	.dc.b	$FF,$7C,$57,$2A,$F7,$4F,$27,$81,$1E,$0F,$F6,$F8,$E1,$1E,$F9,$E7
	.dc.b	$5D,$9F,$21,$8A,$E4,$E1,$E5,$1B,$7C,$31,$B9,$EE,$7F,$7F,$AE,$EF
	.dc.b	$D8,$8D,$1C,$DD,$57,$37,$C9,$5D,$8F,$09,$D8,$C3,$D8,$57,$6C,$3F
	.dc.b	$B8,$B7,$0F,$AF,$7D,$6B,$C5,$7C,$B0,$F1,$47,$29,$D6,$1C,$9B,$7C
	.dc.b	$AF,$C7,$5E,$63,$D2,$E1,$EA,$60,$F3,$61,$E3,$8F,$F7,$D3,$63,$EC
	.dc.b	$1A,$AE,$F3,$A6,$C6,$B0,$F3,$D7,$17,$E5,$D7,$C5,$9F,$04,$6A,$B8
	.dc.b	$BA,$8F,$41,$EF,$9F,$46,$1E,$F6,$E1,$EF,$9F,$24,$60,$F2,$15,$F2
	.dc.b	$3E,$17,$A1,$C0,$95,$E3,$8E,$0F,$8E,$3E,$39,$C1,$D3,$C2,$3A,$61
	.dc.b	$EF,$7E,$C1,$8C,$3D,$AE,$27,$DB,$3B,$7E,$1C,$6C,$FA,$D1,$F9,$43
	.dc.b	$51,$E1,$8E,$4D,$39,$CF,$C3,$AF,$46,$1F,$1E,$7E,$09,$D8,$DD,$7C
	.dc.b	$67,$27,$BA,$FC,$17,$CE,$7C,$93,$A9,$F1,$9A,$8E,$57,$F1,$0F,$91
	.dc.b	$1A,$AE,$55,$E1,$3B,$38,$7D,$46,$2B,$99,$EE,$BF,$55,$F4,$AF,$D3
	.dc.b	$5F,$86,$FE,$91,$D5,$F0,$5F,$28,$F0,$5F,$28,$F9,$33,$A9,$F6,$57
	.dc.b	$43,$C9,$5B,$8E,$91,$9A,$E9,$1E,$73,$8B,$F6,$23,$8E,$1E,$DB,$F3
	.dc.b	$61,$F1,$6B,$C3,$5C,$EF,$E8,$E1,$DD,$F8,$77,$E5,$BF,$81,$7D,$4F
	.dc.b	$0D,$6E,$7A,$C7,$D6,$9D,$D7,$C4,$39,$DF,$78,$FC,$2E,$1E,$58,$F3
	.dc.b	$D7,$B2,$F9,$DF,$E1,$8F,$69,$CD,$CB,$F0,$6F,$B9,$ED,$7E,$3C,$F9
	.dc.b	$DF,$EC,$D3,$FD,$3D,$FF,$55,$35,$FE,$4E,$62,$BF,$AE,$46,$63,$F8
	.dc.b	$B1,$99,$FE,$DD,$19,$8F,$DD,$99,$3B,$CF,$21,$3F,$EB,$DA,$B3,$FE
	.dc.b	$BF,$F9,$46,$FF,$93,$6D,$7D,$E6,$CF,$E9,$9B,$5F,$C1,$5B,$3F,$B4
	.dc.b	$6C,$FE,$7D,$AF,$E7,$9B,$87,$F5,$06,$E1,$E4,$BF,$2D,$27,$37,$E9
	.dc.b	$3F,$DA,$8E,$23,$FC,$56,$D9,$E2,$D9,$DE,$1E,$D2,$3E,$D6,$D7,$B5
	.dc.b	$B5,$ED,$6C,$FB,$5B,$3E,$D6,$BE,$D6,$E1,$ED,$6E,$1E,$DA,$F1,$9E
	.dc.b	$12,$57,$85,$B7,$E1,$69,$E1,$2C,$78,$5A,$F8,$5B,$7E,$16,$D7,$86
	.dc.b	$7F,$01,$7E,$49,$F9,$91,$87,$66,$1F,$A5,$1B,$7E,$7B,$6F,$E7,$B6
	.dc.b	$7D,$B3,$EB,$24,$FA,$DB,$1E,$B6,$9E,$B2,$9E,$B2,$DF,$AD,$B7,$EB
	.dc.b	$7C,$B3,$E8,$7C,$AD,$AF,$2B,$4F,$2B,$E3,$23,$E3,$77,$1C,$8F,$1E
	.dc.b	$1F,$9A,$3F,$18,$FC,$73,$51,$C0,$E6,$E5,$C4,$F5,$3C,$24,$8F,$81
	.dc.b	$1A,$78,$46,$DA,$43,$E2,$46,$DF,$82,$79,$B0,$F8,$95,$E6,$8C,$CE
	.dc.b	$E3,$07,$AE,$76,$7E,$69,$C5,$7E,$31,$B5,$D6,$3F,$09,$1C,$A7,$8C
	.dc.b	$61,$E4,$62,$78,$1F,$1A,$3D,$8F,$91,$F8,$4F,$B1,$B3,$D5,$E8,$48
	.dc.b	$E7,$5F,$07,$0F,$C2,$3E,$DB,$F0,$D6,$E3,$E9,$D7,$73,$E5,$57,$09
	.dc.b	$EF,$5C,$27,$CC,$7D,$48,$C9,$F0,$E3,$D9,$5D,$2F,$E2,$B6,$33,$3E
	.dc.b	$E9,$CB,$63,$35,$D4,$FB,$93,$EF,$7C,$27,$9C,$F8,$86,$A3,$CD,$5F
	.dc.b	$07,$0E,$73,$BA,$E5,$3F,$0E,$FD,$07,$C3,$3C,$45,$9F,$11,$C6,$FA
	.dc.b	$C7,$D4,$C3,$E3,$4F,$E7,$C1,$FB,$E2,$99,$30,$43,$04,$3F,$2E,$70
	.dc.b	$29,$B2,$9B,$BF,$E9,$3F,$80,$BE,$1B,$DE,$DB,$F5,$52,$30,$FC,$A6
	.dc.b	$E3,$FA,$76,$CF,$E4,$1B,$3F,$AD,$6B,$F4,$5B,$7E,$2B,$F0,$92,$BC
	.dc.b	$33,$AC,$7F,$8A,$36,$3F,$95,$36,$FF,$E0,$0B,$70,$FE,$BE,$D9,$F9
	.dc.b	$6D,$9F,$9E,$D7,$DD,$7E,$17,$DC,$DA,$F7,$36,$BD,$CD,$9F,$73,$5F
	.dc.b	$73,$70,$F7,$4F,$B2,$31,$87,$B1,$B1,$EC,$6D,$7B,$1B,$7E,$C6,$CF
	.dc.b	$B1,$B3,$EC,$6B,$EC,$8F,$01,$1F,$6B,$67,$DA,$D3,$DA,$58,$F6,$B5
	.dc.b	$F6,$B6,$FD,$B1,$D0,$93,$D1,$B3,$D1,$A7,$42,$9D,$0B,$3D,$27,$91
	.dc.b	$27,$93,$67,$93,$5E,$4D,$9E,$53,$C0,$93,$C1,$AF,$07,$E6,$1F,$4C
	.dc.b	$F8,$18,$7E,$68,$FC,$41,$6F,$ED,$46,$9B,$3E,$C9,$D4,$E4,$E9,$19
	.dc.b	$9D,$DF,$A6,$B7,$39,$AD,$CE,$6B,$71,$E8,$AD,$CE,$63,$66,$67,$75
	.dc.b	$EF,$3F,$0D,$1A,$70,$7E,$1C,$FC,$83,$E7,$69,$E3,$2C,$F7,$6C,$77
	.dc.b	$3A,$3F,$2C,$EE,$68,$EF,$5E,$88,$E1,$19,$38,$1E,$39,$FD,$04,$78
	.dc.b	$DB,$1E,$36,$9E,$32,$C7,$46,$C7,$82,$BE,$1B,$DA,$BC,$4E,$6B,$D3
	.dc.b	$7F,$88,$8F,$4D,$F2,$78,$5F,$29,$D4,$79,$EB,$F1,$2F,$95,$B5,$E5
	.dc.b	$69,$E5,$2C,$78,$4F,$85,$3D,$2F,$CB,$1E,$09,$F2,$E1,$EA,$9F,$E6
	.dc.b	$B3,$FD,$05,$AF,$88,$C6,$1F,$E8,$27,$04,$3E,$A9,$AA,$FE,$94,$E0
	.dc.b	$D1,$F8,$53,$46,$A7,$F9,$EF,$E0,$2F,$5B,$E6,$B6,$BF,$06,$DA,$FB
	.dc.b	$CD,$CB,$E6,$36,$7E,$1C,$A3,$0F,$D2,$8F,$EC,$B1,$88,$FA,$07,$F6
	.dc.b	$A7,$17,$E4,$BF,$19,$2B,$C6,$DA,$F1,$B6,$FC,$6D,$9F,$1D,$F4,$C3
	.dc.b	$FC,$CC,$D8,$FF,$3E,$36,$D7,$FE,$B0,$8C,$36,$FA,$4F,$AC,$F4,$38
	.dc.b	$7D,$67,$C1,$71,$87,$AE,$FD,$24,$9F,$4B,$6B,$D2,$D3,$D3,$5C,$B0
	.dc.b	$F4,$B6,$3D,$2D,$AF,$4C,$7C,$E8,$C5,$7A,$63,$A4,$FA,$4F,$33,$87
	.dc.b	$D3,$87,$C0,$8E,$46,$70,$F5,$4F,$A0,$95,$E8,$6C,$FA,$0E,$18,$7A
	.dc.b	$1A,$7A,$0B,$1E,$83,$E5,$B8,$9F,$44,$74,$3D,$17,$E5,$76,$64,$F4
	.dc.b	$E1,$F0,$2B,$D3,$19,$C3,$D3,$7E,$72,$4F,$9E,$75,$87,$9D,$E9,$87
	.dc.b	$E5,$CF,$03,$F2,$0F,$41,$6B,$CE,$74,$7C,$AE,$2B,$B3,$63,$31,$83
	.dc.b	$CF,$19,$8F,$5C,$7C,$77,$BC,$F8,$8F,$26,$1E,$7A,$F0,$4F,$8E,$B9
	.dc.b	$9D,$5E,$E4,$3A,$9D,$27,$94,$6E,$39,$1B,$C3,$E4,$C7,$48,$C6,$1C
	.dc.b	$1E,$98,$79,$6F,$11,$C8,$EF,$87,$B6,$75,$7E,$D8,$F8,$18,$76,$AE
	.dc.b	$77,$FA,$DB,$F7,$DF,$E5,$DB,$3F,$5A,$7B,$D7,$18,$F0,$4F,$67,$DD
	.dc.b	$5D,$CF,$74,$72,$BF,$92,$D8,$CD,$F4,$C3,$D7,$39,$BF,$C9,$CE,$0F
	.dc.b	$31,$BA,$ED,$1E,$A7,$D1,$19,$8F,$19,$2F,$E0,$DF,$3A,$E3,$3F,$21
	.dc.b	$FA,$0F,$CB,$C3,$DD,$87,$8E,$34,$77,$3E,$24,$F2,$9D,$E1,$E1,$C3
	.dc.b	$E9,$36,$7A,$DF,$3A,$DC,$71,$C3,$C9,$87,$9E,$3E,$29,$93,$CF,$5F
	.dc.b	$D1,$E3,$FD,$38,$D7,$F6,$66,$2F,$FA,$E3,$FB,$A3,$13,$FD,$69,$FD
	.dc.b	$71,$83,$FA,$EC,$F3,$78,$1D,$08,$7E,$77,$E6,$DE,$3F,$E4,$9B,$5F
	.dc.b	$31,$B3,$EB,$3F,$D7,$6E,$32,$FD,$93,$67,$E8,$B6,$7B,$57,$52,$4F
	.dc.b	$56,$C7,$51,$3C,$5F,$D2,$36,$FF,$B2,$37,$05,$FF,$B3,$B0,$B7,$E6
	.dc.b	$C3,$AC,$79,$9B,$3E,$66,$CF,$9A,$3C,$84,$7C,$97,$DE,$3B,$1F,$ED
	.dc.b	$A7,$11,$FC,$79,$B5,$E2,$8F,$5D,$76,$7D,$6D,$C3,$D6,$DC,$3D,$35
	.dc.b	$C6,$F9,$E1,$DE,$B4,$F8,$B0,$F7,$3E,$7A,$F7,$34,$F7,$16,$3D,$C7
	.dc.b	$B2,$79,$BE,$96,$CF,$85,$B3,$E1,$39,$4F,$A9,$B8,$7A,$9B,$87,$A9
	.dc.b	$E1,$3D,$DC,$9E,$97,$E2,$B8,$7C,$4D,$BF,$13,$4E,$6F,$BA,$3A,$11
	.dc.b	$C1,$23,$E0,$C6,$5B,$19,$8F,$86,$D7,$D8,$6C,$CC,$70,$23,$82,$1C
	.dc.b	$23,$96,$1F,$99,$3E,$CC,$7D,$63,$DA,$48,$DE,$1F,$1E,$35,$18,$AF
	.dc.b	$8D,$5C,$A7,$E3,$9C,$63,$E1,$9E,$47,$33,$F2,$1F,$51,$88,$DB,$E8
	.dc.b	$9F,$35,$7E,$55,$B3,$F7,$9B,$3C,$9B,$5E,$58,$F1,$1C,$EB,$D4,$F7
	.dc.b	$31,$5E,$A8,$EC,$6A,$BD,$87,$48,$F2,$CF,$CB,$6C,$6A,$78,$12,$3C
	.dc.b	$CD,$9F,$33,$6B,$CC,$7C,$32,$CF,$6C,$3F,$31,$5E,$A3,$D0,$78,$4F
	.dc.b	$A7,$1B,$C3,$A3,$E4,$BF,$5B,$ED,$3C,$65,$9F,$1B,$6F,$C3,$1D,$0F
	.dc.b	$8A,$78,$6F,$C7,$1E,$79,$F0,$1C,$4F,$01,$DA,$78,$5F,$91,$E8,$7A
	.dc.b	$AB,$9C,$F0,$8F,$19,$C1,$E8,$F0,$C3,$E7,$46,$23,$95,$75,$7E,$2C
	.dc.b	$E9,$FC,$D8,$3F,$7E,$53,$F4,$05,$3F,$32,$57,$FA,$6B,$70,$FE,$BE
	.dc.b	$E8,$E6,$43,$99,$4D,$1F,$03,$F0,$07,$CD,$79,$DB,$87,$AD,$B3,$F7
	.dc.b	$DB,$97,$CF,$8F,$54,$62,$7F,$26,$D8,$ED,$5D,$49,$3D,$5B,$1D,$5B
	.dc.b	$5D,$61,$5C,$DF,$C7,$B7,$0F,$EC,$0D,$C3,$F5,$6D,$C3,$F9,$43,$6B
	.dc.b	$FC,$DA,$DB,$FE,$A0,$D9,$FA,$4D,$9F,$6E,$1E,$49,$F6,$B6,$3D,$B1
	.dc.b	$EB,$23,$EB,$6D,$7A,$EB,$A4,$FB,$2B,$93,$EC,$6E,$1E,$C6,$E1,$EC
	.dc.b	$6E,$1E,$C6,$D7,$B1,$B8,$7B,$1B,$3E,$C6,$CF,$B2,$3C,$A7,$B0,$AF
	.dc.b	$B0,$F5,$12,$7D,$53,$CA,$3D,$53,$C1,$F5,$36,$FD,$4D,$BF,$53,$4F
	.dc.b	$51,$67,$D4,$DB,$F5,$34,$F5,$16,$3D,$47,$95,$F5,$34,$F5,$47,$A2
	.dc.b	$3B,$C7,$63,$47,$A8,$C4,$6D,$B1,$E0,$77,$87,$E4,$CF,$B6,$5C,$3D
	.dc.b	$B1,$A6,$D7,$33,$D7,$1D,$0F,$2D,$75,$8C,$9B,$70,$48,$D9,$F0,$23
	.dc.b	$71,$F0,$1B,$19,$8C,$1C,$5F,$74,$75,$74,$43,$EC,$9E,$53,$71,$C9
	.dc.b	$D5,$7C,$0A,$F6,$CF,$5C,$3E,$FD,$71,$AD,$4F,$D0,$BF,$2C,$F6,$6C
	.dc.b	$EA,$38,$5F,$E9,$23,$E6,$57,$E3,$5B,$3F,$36,$3A,$1E,$57,$CE,$7A
	.dc.b	$EF,$C7,$7D,$AF,$A1,$E8,$9F,$96,$70,$24,$F0,$AF,$91,$7E,$5B,$E5
	.dc.b	$1E,$3B,$E5,$7E,$B8,$DC,$76,$7C,$B1,$DA,$BD,$35,$F4,$63,$C1,$1D
	.dc.b	$67,$89,$D4,$E3,$1D,$4F,$A3,$5D,$5F,$AD,$7E,$77,$33,$C5,$CE,$1E
	.dc.b	$6B,$F0,$CF,$C7,$6C,$F9,$6F,$FA,$5C,$7F,$40,$8F,$EA,$4D,$7F,$44
	.dc.b	$62,$36,$49,$FF,$57,$4E,$E7,$F6,$A6,$E3,$33,$B8,$FE,$BC,$7D,$1F
	.dc.b	$C0,$1D,$7F,$31,$B5,$F0,$9B,$3F,$95,$6E,$5F,$8C,$6C,$FE,$7D,$B3
	.dc.b	$F9,$D6,$BD,$EB,$B1,$27,$B3,$63,$B3,$6B,$B3,$67,$B3,$67,$B3,$5E
	.dc.b	$D1,$D0,$8F,$46,$CF,$46,$9D,$0B,$1D,$1A,$F4,$39,$12,$39,$36,$39
	.dc.b	$34,$E4,$53,$94,$70,$24,$70,$6C,$70,$6B,$C2,$34,$47,$4D,$74,$60
	.dc.b	$86,$09,$87,$F7,$DB,$FA,$AD,$BF,$94,$DA,$F8,$2D,$AF,$53,$6F,$CA
	.dc.b	$D9,$F1,$36,$7C,$2D,$7C,$38,$7F,$9C,$B0,$F0,$B7,$0F,$0B,$6F,$C2
	.dc.b	$DA,$F0,$B7,$0F,$0B,$67,$C2,$D9,$FF,$20,$5F,$E4,$8E,$F3,$FE,$BD
	.dc.b	$71,$7B,$9F,$8D,$18,$9F,$C0,$B6,$3E,$D3,$6F,$EA,$B6,$7E,$93,$67
	.dc.b	$E8,$B5,$EB,$5D,$F0,$F8,$93,$96,$C6,$70,$F2,$4E,$5B,$19,$BF,$53
	.dc.b	$87,$CE,$E6,$D8,$7F,$BF,$EF,$C4,$7C,$28,$CD,$7F,$98,$E3,$35,$FA
	.dc.b	$B7,$C1,$1D,$89,$3C,$0F,$9B,$3B,$3C,$15,$DD,$F2,$47,$C8,$73,$3D
	.dc.b	$DC,$CF,$CA,$73,$3D,$27,$BC,$F3,$7D,$A6,$0F,$9F,$3C,$25,$5E,$A9
	.dc.b	$DC,$79,$CC,$CF,$9E,$39,$47,$D6,$8E,$53,$C6,$3D,$13,$A7,$EC,$B7
	.dc.b	$0F,$AE,$DC,$3D,$6F,$B4,$F0,$E1,$E9,$AE,$67,$8E,$FE,$94,$64,$93
	.dc.b	$C6,$76,$7F,$A5,$67,$F9,$9B,$70,$FF,$49,$C6,$23,$81,$99,$FE,$7E
	.dc.b	$FC,$06,$C6,$E3,$E7,$1F,$D1,$1C,$3B,$31,$3D,$8F,$F7,$4F,$DE,$3C
	.dc.b	$D7,$B5,$F1,$63,$0D,$8C,$D7,$BE,$B8,$C6,$31,$F7,$CF,$FA,$D6,$F8
	.dc.b	$BB,$C7,$FD,$71,$27,$FD,$71,$3A,$8E,$32,$8C,$4F,$C9,$9E,$13,$D6
	.dc.b	$BC,$13,$93,$9B,$88,$F8,$73,$C1,$E9,$19,$7E,$24,$66,$BB,$C7,$0A
	.dc.b	$E7,$87,$C1,$24,$7A,$CF,$C9,$4F,$07,$9C,$F0,$75,$5D,$6F,$C1,$8F
	.dc.b	$F6,$AB,$E0,$D8,$CE,$3F,$33,$1E,$F7,$F1,$2F,$73,$9C,$3A,$E3,$DE
	.dc.b	$B5,$87,$28,$F9,$E6,$4F,$8D,$3F,$27,$0E,$B1,$88,$F8,$B1,$87,$E1
	.dc.b	$1B,$3B,$96,$BB,$C6,$E3,$E0,$99,$9E,$98,$7A,$1F,$1C,$64,$EA,$78
	.dc.b	$0E,$A5,$AE,$AE,$27,$93,$F4,$2F,$DC,$DB,$F1,$63,$F2,$DB,$8F,$C7
	.dc.b	$C3,$99,$31,$F2,$E1,$CA,$3E,$B5,$70,$9F,$3E,$1E,$37,$71,$98,$F0
	.dc.b	$1E,$5A,$E5,$3F,$3E,$3D,$B8,$FA,$2F,$AC,$6E,$F9,$46,$0F,$24,$70
	.dc.b	$BE,$6F,$C7,$3E,$69,$E3,$35,$3C,$AF,$C1,$1A,$BF,$04,$63,$1F,$89
	.dc.b	$3E,$12,$61,$E0,$78,$61,$E0,$7B,$12,$F9,$BE,$37,$F4,$06,$1D,$47
	.dc.b	$EB,$63,$06,$9C,$1D,$5C,$61,$F3,$9D,$C6,$A3,$0E,$CC,$4F,$D0,$AF
	.dc.b	$45,$72,$8F,$5E,$1F,$0A,$39,$DF,$22,$4E,$8F,$55,$F5,$8C,$DF,$D3
	.dc.b	$BF,$BF,$5D,$DF,$95,$8F,$E2,$DE,$D7,$C6,$74,$7D,$7C,$7E,$7D,$F3
	.dc.b	$8C,$5F,$17,$C3,$87,$EA,$71,$F5,$57,$B3,$0F,$7B,$67,$CB,$8F,$92
	.dc.b	$3A,$5F,$D6,$C7,$BC,$F9,$2B,$A5,$F2,$8E,$06,$6B,$E1,$1F,$62,$76
	.dc.b	$76,$7F,$1A,$7B,$0E,$D8,$FA,$5E,$07,$2C,$7D,$35,$DA,$3D,$A6,$6F
	.dc.b	$C4,$F4,$C7,$D3,$5F,$5A,$3B,$36,$33,$3F,$35,$B8,$79,$AB,$C9,$8F
	.dc.b	$C0,$7B,$CF,$C7,$73,$8F,$9F,$0F,$2C,$78,$71,$F8,$55,$B3,$E5,$DF
	.dc.b	$39,$F6,$C6,$67,$F0,$27,$C5,$8C,$BE,$F6,$9D,$4F,$1B,$9A,$F0,$C7
	.dc.b	$EB,$B0,$E5,$1E,$F9,$F1,$9B,$C7,$E3,$D6,$CF,$8A,$FC,$38,$DC,$77
	.dc.b	$3C,$26,$EF,$E3,$BF,$46,$FD,$18,$7B,$70,$F1,$B7,$0E,$ED,$9D,$E3
	.dc.b	$F2,$67,$2D,$8C,$E3,$EA,$9C,$B6,$33,$87,$A1,$B8,$F9,$0F,$9F,$7C
	.dc.b	$E7,$EB,$D7,$16,$9C,$4B,$3C,$63,$E9,$C6,$63,$9C,$71,$6C,$66,$37
	.dc.b	$1C,$63,$F6,$71,$B3,$27,$CC,$7F,$3E,$E5,$D9,$CC,$B1,$CD,$D9,$92
	.dc.b	$B9,$C7,$EF,$47,$E1,$0F,$B0,$FD,$2A,$F8,$53,$E8,$38,$61,$F7,$EB
	.dc.b	$81,$F7,$6F,$CD,$8F,$D7,$8F,$8B,$1E,$9C,$3B,$C7,$E1,$27,$57,$D6
	.dc.b	$B5,$7D,$6B,$58,$78,$A7,$E1,$C7,$83,$0E,$67,$9E,$3E,$24,$7B,$9E
	.dc.b	$78,$FE,$D7,$0E,$D3,$C7,$0E,$D7,$C6,$3E,$33,$D2,$3F,$13,$18,$76
	.dc.b	$62,$7E,$1D,$75,$BF,$2D,$7E,$82,$F9,$4F,$E3,$A7,$C7,$3F,$A1,$3E
	.dc.b	$29,$E3,$8E,$4F,$DA,$8D,$D7,$EF,$E3,$06,$CE,$4F,$B1,$F0,$5F,$9A
	.dc.b	$BC,$11,$B9,$E4,$FC,$53,$15,$FB,$D9,$DC,$72,$7C,$6E,$CC,$1E,$08
	.dc.b	$F7,$DF,$F0,$A3,$DF,$1E,$C3,$75,$E0,$8C,$BE,$77,$85,$F9,$E7,$84
	.dc.b	$62,$FC,$2F,$0B,$EF,$5F,$C4,$A3,$15,$F1,$0E,$F3,$EB,$79,$CE,$CE
	.dc.b	$D7,$E1,$71,$7D,$9C,$D7,$68,$FC,$5E,$3F,$96,$8C,$5F,$EA,$E3,$0F
	.dc.b	$87,$0F,$43,$F1,$4C,$5F,$67,$07,$CB,$8D,$E1,$F1,$67,$C3,$3C,$30
	.dc.b	$F0,$E3,$FA,$18,$C1,$E1,$70,$F3,$9E,$D5,$CE,$BD,$84,$9F,$0D,$FC
	.dc.b	$87,$84,$76,$C7,$F5,$33,$92,$5F,$92,$B6,$78,$67,$71,$99,$DD,$77
	.dc.b	$BF,$15,$F8,$2B,$44,$9D,$5F,$D7,$AF,$A9,$1E,$3A,$FD,$83,$A7,$F8
	.dc.b	$AD,$7E,$C6,$BF,$4A,$71,$3E,$AD,$7F,$21,$C7,$EA,$D7,$A2,$31,$8F
	.dc.b	$C8,$C7,$CF,$3E,$5C,$7B,$3F,$C6,$CF,$E3,$B7,$CA,$77,$7E,$C2,$61
	.dc.b	$E1,$AF,$AB,$1D,$A3,$E2,$DF,$4B,$E5,$8F,$C1,$BF,$01,$2F,$9B,$EE
	.dc.b	$AF,$D9,$E1,$F8,$6C,$3B,$4F,$AD,$FE,$5F,$8F,$AF,$0E,$98,$72,$AF
	.dc.b	$51,$FC,$AA,$BC,$6F,$77,$9D,$FB,$1B,$5C,$9B,$87,$A7,$0E,$CD,$8E
	.dc.b	$78,$77,$C7,$DB,$39,$C7,$EC,$9D,$CF,$61,$FB,$47,$EF,$4F,$F2,$A8
	.dc.b	$D1,$E8,$C7,$F9,$A4,$64,$ED,$87,$8A,$77,$7E,$06,$C6,$67,$78,$79
	.dc.b	$8F,$5C,$E9,$C3,$E4,$8F,$C8,$BF,$9D,$3D,$2F,$26,$C7,$28,$C1,$E5
	.dc.b	$70,$48,$E8,$FE,$50,$FA,$E7,$81,$F7,$63,$F9,$B7,$F9,$99,$C0,$FC
	.dc.b	$5C,$70,$7C,$26,$4F,$C1,$1F,$96,$AF,$81,$1E,$18,$EB,$8F,$E2,$AB
	.dc.b	$F8,$2C,$E7,$0F,$91,$87,$AA,$3F,$57,$7D,$71,$FC,$23,$6F,$F0,$8D
	.dc.b	$C3,$EF,$5F,$C1,$8C,$1F,$34,$F3,$9E,$23,$F0,$07,$CD,$78,$39,$7E
	.dc.b	$89,$FA,$97,$8E,$1E,$6C,$3D,$18,$FB,$A7,$BD,$7A,$5F,$92,$73,$9F
	.dc.b	$A0,$FC,$57,$E8,$57,$38,$E3,$1D,$0C,$D7,$58,$E4,$7D,$23,$91,$E4
	.dc.b	$75,$3C,$E3,$67,$D4,$3F,$60,$6E,$7A,$91,$C1,$2B,$E1,$CF,$32,$57
	.dc.b	$62,$57,$19,$DC,$62,$B9,$4F,$DA,$7C,$78,$FD,$A6,$E3,$F6,$9B,$8F
	.dc.b	$DA,$AE,$18,$F8,$1E,$C6,$30,$ED,$8F,$82,$33,$7F,$42,$B8,$3E,$23
	.dc.b	$15,$C2,$79,$1C,$23,$FB,$B9,$F4,$88,$79,$0F,$C5,$57,$77,$A1,$C0
	.dc.b	$D3,$F8,$17,$15,$FE,$51,$44,$18,$B1,$63,$99,$D4,$E6,$31,$1C,$A3
	.dc.b	$0E,$1A,$F1,$6B,$CA,$7A,$BB,$6B,$B6,$BB,$9E,$EF,$9A,$32,$F8,$5A
	.dc.b	$FA,$BC,$E7,$37,$0B,$46,$27,$CC,$D7,$CC,$D7,$DE,$D7,$DC,$D7,$B4
	.dc.b	$F5,$8C,$4F,$68,$C1,$27,$BB,$71,$65,$DA,$3E,$9C,$62,$7B,$C6,$27
	.dc.b	$EB,$7D,$C3,$C9,$FB,$17,$BB,$F1,$87,$18,$78,$E3,$17,$DD,$A7,$E0
	.dc.b	$AF,$A3,$DC,$F0,$9A,$3A,$D7,$4C,$7E,$53,$E1,$9E,$33,$BB,$F0,$4E
	.dc.b	$F1,$E9,$3C,$B0,$F5,$E3,$EA,$7C,$95,$B9,$E5,$8F,$69,$DC,$72,$C3
	.dc.b	$AC,$70,$24,$7A,$EF,$A3,$EA,$BE,$84,$9E,$71,$DB,$0E,$98,$77,$8E
	.dc.b	$13,$AA,$E1,$18,$20,$C7,$E0,$C6,$A3,$DC,$6A,$33,$3E,$68,$C9,$23
	.dc.b	$2F,$4C,$7D,$AF,$1C,$7D,$6F,$08,$CB,$C6,$33,$18,$26,$1E,$B9,$F3
	.dc.b	$4E,$A3,$13,$AC,$7E,$AD,$F8,$AF,$B5,$68,$E5,$19,$AE,$13,$F2,$4F
	.dc.b	$C3,$46,$67,$85,$7C,$C3,$AD,$F8,$5C,$57,$1C,$7F,$15,$39,$8C,$63
	.dc.b	$F1,$F0,$E9,$7C,$2F,$E7,$46,$2B,$67,$C7,$35,$5F,$3F,$1F,$8F,$3B
	.dc.b	$AF,$AD,$18,$C3,$C1,$18,$C7,$DF,$18,$BE,$51,$8C,$3A,$9F,$5D,$C3
	.dc.b	$E2,$3E,$4E,$3F,$01,$D9,$83,$E3,$D7,$57,$94,$6D,$E5,$87,$CE,$C3
	.dc.b	$E2,$C7,$28,$C6,$1F,$95,$3C,$B5,$DB,$1F,$74,$62,$BB,$9C,$71,$FA
	.dc.b	$AE,$0E,$EE,$30,$EF,$18,$BE,$98,$F9,$E3,$D5,$3A,$3E,$44,$EA,$3B
	.dc.b	$35,$E4,$7C,$5C,$3C,$AF,$C0,$3C,$53,$C9,$D5,$77,$C7,$DF,$1F,$BD
	.dc.b	$3A,$9A,$2C,$FE,$12,$78,$61,$EB,$C7,$E3,$57,$7A,$D3,$C2,$3F,$1D
	.dc.b	$5C,$31,$F4,$57,$8C,$DD,$7A,$9A,$66,$B8,$C7,$33,$F2,$46,$DC,$B4
	.dc.b	$CC,$71,$6C,$7D,$78,$EA,$D3,$8B,$B8,$FE,$00,$77,$29,$D4,$AE,$F0
	.dc.b	$FC,$F4,$6D,$AE,$DB,$1B,$C3,$F3,$6D,$BF,$CA,$36,$BE,$BD,$F9,$5F
	.dc.b	$49,$F3,$5F,$39,$97,$C6,$DB,$EE,$DB,$EE,$DB,$EE,$DA,$EE,$75,$3B
	.dc.b	$9F,$92,$7B,$B9,$3B,$96,$BB,$B6,$BB,$B6,$BB,$B4,$EE,$75,$3B,$CF
	.dc.b	$E4,$9E,$A6,$4A,$E4,$EE,$E5,$A6,$67,$C3,$19,$6C,$66,$7C,$73,$96
	.dc.b	$C6,$63,$CA,$DB,$FD,$95,$FC,$76,$CF,$AD,$AF,$33,$8D,$FE,$95,$F6
	.dc.b	$B5,$F5,$B6,$BD,$2D,$3C,$E5,$BF,$2B,$67,$AC,$7E,$57,$1F,$DF,$C7
	.dc.b	$C7,$6D,$7C,$B6,$D7,$85,$B5,$FB,$2C,$3D,$87,$EE,$AF,$D5,$87,$B3
	.dc.b	$0F,$85,$39,$6C,$FF,$07,$3F,$3D,$87,$A2,$FC,$98,$FE,$FE,$7F,$85
	.dc.b	$1A,$8F,$59,$A7,$F5,$13,$C5,$FD,$76,$3E,$67,$8E,$3F,$A5,$3C,$93
	.dc.b	$F0,$DB,$5C,$E7,$B1,$F1,$8E,$77,$E3,$9F,$E0,$31,$F6,$71,$F6,$5F
	.dc.b	$36,$DF,$3A,$F0,$46,$2B,$99,$E8,$9F,$5B,$83,$C3,$3E,$47,$0F,$27
	.dc.b	$E4,$35,$FF,$04,$9C,$CF,$43,$FC,$4F,$0F,$84,$FA,$23,$EA,$D7,$8E
	.dc.b	$35,$87,$B6,$3F,$6F,$3A,$3B,$BF,$79,$A7,$C1,$29,$B3,$C6,$7A,$9E
	.dc.b	$C7,$F0,$FA,$F1,$BD,$B1,$FC,$1C,$F0,$BF,$65,$6A,$3F,$43,$87,$CB
	.dc.b	$BF,$14,$6D,$B3,$B8,$FD,$2C,$ED,$F3,$1B,$2C,$6E,$FF,$53,$3B,$9F
	.dc.b	$54,$E7,$1F,$C6,$61,$F1,$A7,$E7,$3F,$54,$C3,$EF,$3E,$1D,$FB,$2B
	.dc.b	$95,$FD,$EA,$D9,$EE,$C7,$E8,$5F,$B8,$FB,$F8,$FB,$23,$31,$EC,$32
	.dc.b	$7A,$9C,$E1,$EA,$8C,$9E,$A8,$CE,$3E,$AA,$FA,$51,$8C,$3C,$91,$8A
	.dc.b	$F6,$BF,$61,$FE,$71,$8F,$A9,$FC,$6C,$6D,$F8,$07,$CD,$3C,$07,$23
	.dc.b	$E1,$DF,$E9,$5B,$5F,$3F,$0F,$8D,$87,$8E,$7A,$9E,$07,$18,$FC,$FC
	.dc.b	$3C,$0F,$C9,$9E,$C7,$86,$B9,$B8,$C7,$D1,$3C,$63,$A1,$9B,$EB,$7E
	.dc.b	$F8,$FE,$45,$18,$7C,$E7,$CC,$BE,$AF,$C0,$BE,$B5,$C6,$7F,$94,$57
	.dc.b	$1A,$F2,$63,$F0,$5B,$8F,$C1,$8C,$61,$CA,$7B,$E3,$F8,$6C,$7F,$05
	.dc.b	$1E,$72,$1E,$1C,$3B,$57,$18,$EA,$79,$EF,$AC,$E8,$91,$CA,$FE,$53
	.dc.b	$CC,$C5,$F8,$1C,$12,$36,$48,$D3,$63,$F2,$87,$22,$9C,$63,$64,$73
	.dc.b	$1C,$5F,$C4,$9B,$33,$1B,$73,$1B,$C7,$E9,$B6,$FD,$F3,$E1,$37,$5C
	.dc.b	$6B,$6D,$9D,$CF,$73,$67,$80,$86,$09,$3C,$9C,$B5,$CC,$74,$AF,$26
	.dc.b	$3F,$7C,$93,$F5,$5B,$1F,$7E,$39,$CF,$1B,$F3,$56,$9F,$D0,$98,$9E
	.dc.b	$44,$AE,$4F,$8A,$78,$3F,$A0,$3D,$D7,$E7,$25,$F7,$8F,$0C,$F0,$7F
	.dc.b	$46,$7C,$B9,$DC,$65,$FC,$FC,$66,$79,$3E,$13,$C9,$7E,$87,$A9,$83
	.dc.b	$C9,$7F,$19,$C1,$E0,$3C,$F3,$D2,$BC,$31,$A7,$C2,$EA,$FC,$93,$FE
	.dc.b	$17,$C3,$F3,$F3,$93,$84,$65,$F4,$4F,$03,$FC,$38,$F9,$AF,$E5,$DF
	.dc.b	$81,$B5,$E3,$25,$71,$AF,$F0,$7E,$1F,$56,$B7,$39,$7F,$B9,$5F,$39
	.dc.b	$F8,$95,$F1,$E7,$6F,$9A,$BC,$6E,$67,$75,$C6,$3E,$4C,$77,$33,$7E
	.dc.b	$48,$E4,$79,$0F,$64,$6A,$BC,$8F,$F8,$03,$1F,$ED,$43,$15,$7F,$46
	.dc.b	$3F,$CF,$B1,$A8,$CC,$7F,$38,$3F,$CE,$31,$97,$81,$88,$FE,$B7,$1D
	.dc.b	$63,$38,$7F,$9D,$23,$84,$73,$8C,$4F,$03,$BE,$3F,$CE,$E7,$A1,$FD
	.dc.b	$E6,$B8,$C6,$6B,$8C,$E6,$B9,$3E,$0C,$7F,$CF,$F7,$C3,$0F,$E7,$51
	.dc.b	$8A,$DD,$78,$E3,$0F,$32,$1F,$DC,$2B,$93,$FC,$EA,$BE,$65,$72,$C3
	.dc.b	$C9,$7C,$AB,$84,$E6,$BC,$38,$FF,$A3,$B0,$E7,$5F,$CF,$30,$FF,$1D
	.dc.b	$46,$09,$7C,$67,$C0,$7D,$67,$0F,$F8,$D0,$FA,$E5,$3E,$CC,$7D,$37
	.dc.b	$26,$4F,$F3,$B1,$C1,$D3,$63,$6E,$8E,$6E,$1E,$44,$8E,$2F,$FA,$1D
	.dc.b	$F0,$4F,$FA,$8B,$27,$BF,$F7,$B1,$D6,$17,$0F,$F2,$7D,$69,$F8,$D1
	.dc.b	$97,$FC,$25,$19,$7F,$D8,$F8,$76,$AF,$F4,$B4,$E7,$0F,$F5,$8C,$E6
	.dc.b	$BA,$E1,$CE,$31,$1F,$E1,$D6,$D7,$1B,$F2,$CF,$FB,$12,$30,$FF,$99
	.dc.b	$27,$89,$C2,$BA,$61,$27,$84,$57,$CF,$1E,$BA,$E7,$1E,$18,$D6,$0C
	.dc.b	$3F,$63,$5A,$C3,$BD,$6A,$73,$85,$C3,$C9,$3D,$63,$11,$D4,$FD,$FE
	.dc.b	$1D,$5F,$E0,$38,$75,$AD,$E1,$EB,$9F,$A2,$7A,$8A,$7F,$BD,$A7,$83
	.dc.b	$8A,$E8,$A3,$2F,$C0,$3C,$38,$7A,$1E,$0F,$D8,$C3,$C3,$87,$A2,$B8
	.dc.b	$DF,$DA,$8F,$65,$F0,$3E,$D9,$4F,$C2,$D7,$27,$D0,$78,$4F,$43,$96
	.dc.b	$9C,$70,$F8,$0E,$EB,$F3,$13,$B6,$C6,$EB,$C6,$EF,$0F,$53,$93,$DE
	.dc.b	$7C,$EA,$E5,$1A,$8E,$07,$A8,$A7,$A8,$E4,$2B,$E6,$E1,$F1,$27,$F2
	.dc.b	$B3,$97,$E6,$3C,$B0,$F5,$D7,$0B,$F0,$CF,$8C,$F2,$9F,$5D,$CB,$4C
	.dc.b	$DF,$77,$CD,$3C,$DF,$A1,$1F,$9A,$34,$FA,$4F,$21,$DC,$F6,$BB,$9E
	.dc.b	$ED,$8E,$71,$F3,$27,$F0,$11,$E2,$8C,$5F,$C0,$9F,$14,$F6,$3D,$07
	.dc.b	$69,$DB,$FA,$43,$91,$0D,$16,$3B,$BC,$EB,$C8,$E8,$C4,$FA,$0E,$AF
	.dc.b	$EB,$4C,$9C,$EB,$CC,$7D,$4A,$F3,$9D,$8A,$E4,$DC,$7C,$78,$CB,$E9
	.dc.b	$3A,$91,$E5,$1C,$63,$83,$C9,$F8,$71,$C9,$F3,$36,$BE,$BD,$FE,$1A
	.dc.b	$B9,$56,$9B,$3A,$C3,$DE,$D9,$F7,$1E,$03,$CD,$7D,$70,$F8,$D1,$A3
	.dc.b	$A1,$B3,$CB,$7E,$7A,$F9,$F5,$EC,$9D,$47,$77,$E0,$4F,$9B,$0F,$E4
	.dc.b	$71,$D6,$75,$5E,$18,$FC,$59,$E1,$9D,$BC,$8F,$13,$88,$E1,$1F,$7D
	.dc.b	$F9,$11,$9A,$F8,$77,$EC,$AF,$A6,$7F,$70,$BF,$EA,$8D,$8F,$EC,$D1
	.dc.b	$88,$FD,$A9,$93,$FB,$6B,$97,$F4,$C6,$4F,$EE,$EE,$5F,$CE,$99,$35
	.dc.b	$7E,$83,$F9,$8F,$DB,$BD,$2F,$AF,$7F,$96,$8C,$57,$C7,$8C,$4E,$A3
	.dc.b	$EB,$4E,$9F,$8C,$7D,$63,$EC,$39,$AF,$06,$1F,$88,$8C,$3F,$1A,$79
	.dc.b	$D7,$57,$9C,$E9,$FB,$B7,$E2,$7C,$87,$02,$9C,$27,$95,$FE,$6E,$B9
	.dc.b	$4F,$94,$98,$7F,$6C,$8C,$BF,$8C,$C3,$CD,$7C,$9B,$5D,$AF,$A4,$E6
	.dc.b	$3D,$ED,$7B,$5F,$4C,$30,$D8,$E5,$7F,$E6,$88,$C4,$7C,$AC,$3C,$97
	.dc.b	$CD,$B1,$C0,$EE,$53,$BC,$6E,$FF,$C4,$55,$D3,$0F,$A3,$5F,$E7,$A8
	.dc.b	$C6,$1F,$46,$3F,$92,$46,$1F,$9E,$7F,$40,$70,$FC,$8A,$EB,$7D,$E3
	.dc.b	$4D,$79,$B6,$FF,$1E,$DB,$F8,$0D,$9F,$23,$55,$F8,$30,$FA,$91,$A3
	.dc.b	$E0,$3F,$11,$F9,$18,$7E,$15,$A7,$E1,$4A,$7E,$14,$E2,$FA,$9A,$7A
	.dc.b	$88,$70,$C3,$D2,$E1,$F9,$6F,$9A,$F9,$3B,$C3,$DA,$ED,$F9,$D3,$CA
	.dc.b	$3E,$29,$F6,$70,$FC,$A9,$EF,$BE,$73,$B3,$BC,$6A,$FF,$16,$6A,$FA
	.dc.b	$9E,$0B,$EA,$78,$CF,$90,$62,$B9,$CF,$18,$FB,$E6,$67,$94,$EA,$3D
	.dc.b	$46,$9F,$24,$68,$F0,$C7,$68,$CB,$B8,$D3,$97,$6F,$C9,$3A,$9F,$8D
	.dc.b	$2D,$7C,$36,$D7,$C0,$77,$87,$E1,$1E,$47,$A4,$F0,$1D,$C8,$75,$34
	.dc.b	$70,$BF,$73,$88,$E9,$5F,$3E,$3E,$B5,$7C,$77,$95,$7B,$9E,$57,$EC
	.dc.b	$AE,$11,$F4,$4C,$9E,$2B,$EB,$7C,$DB,$1E,$CC,$3F,$35,$87,$59,$CC
	.dc.b	$F9,$67,$38,$7C,$C8,$F1,$B7,$0F,$91,$1E,$E6,$DF,$9D,$B3,$BC,$3E
	.dc.b	$65,$7E,$20,$FF,$62,$CF,$F3,$D6,$C7,$FA,$EE,$31,$1F,$8F,$32,$7F
	.dc.b	$B0,$9C,$BF,$90,$32,$7F,$B3,$1C,$BF,$60,$C9,$AA,$F3,$4F,$FA,$EF
	.dc.b	$ED,$DF,$17,$F4,$6F,$E6,$D7,$FA,$2A,$73,$7F,$86,$3F,$16,$E1,$F8
	.dc.b	$57,$F6,$63,$13,$DA,$77,$7D,$6B,$4D,$8C,$9D,$4F,$C4,$D7,$2C,$3E
	.dc.b	$54,$68,$EE,$5B,$EC,$F0,$AE,$8D,$3A,$47,$E2,$EB,$A4,$7E,$63,$0F
	.dc.b	$7D,$A2,$4E,$9F,$C6,$92,$7F,$AF,$C6,$5E,$31,$FA,$28,$C6,$1E,$2B
	.dc.b	$F9,$58,$5C,$3F,$19,$5A,$BF,$2C,$78,$54,$65,$F7,$9F,$E6,$D7,$15
	.dc.b	$F0,$0E,$B1,$C8,$F1,$57,$76,$99,$3D,$73,$CA,$BE,$6C,$75,$9E,$0E
	.dc.b	$9B,$87,$C0,$6E,$1F,$89,$6C,$FF,$A6,$AF,$A3,$F3,$67,$4F,$86,$B9
	.dc.b	$36,$32,$FB,$B0,$FA,$D7,$D2,$3C,$86,$CE,$CE,$63,$E0,$C7,$1C,$3E
	.dc.b	$6C,$E5,$F9,$98,$79,$27,$ED,$5F,$82,$F8,$9F,$2E,$B7,$79,$7E,$2C
	.dc.b	$69,$CD,$7C,$BB,$FC,$31,$D7,$0F,$6C,$7A,$0F,$21,$F8,$CB,$E8,$FC
	.dc.b	$53,$D6,$43,$67,$E0,$9D,$BE,$83,$57,$E3,$79,$1D,$DC,$3B,$31,$7E
	.dc.b	$87,$AD,$FA,$2B,$EF,$B6,$3D,$91,$E0,$BF,$CE,$C7,$8C,$F0,$D7,$C4
	.dc.b	$3A,$5F,$82,$77,$7C,$EB,$BD,$F3,$C3,$F8,$0C,$72,$3D,$37,$E8,$AF
	.dc.b	$95,$3F,$01,$B3,$B3,$C4,$74,$31,$7E,$0B,$ED,$1F,$C1,$70,$F2,$9E
	.dc.b	$77,$E9,$1C,$2F,$C6,$FC,$63,$C7,$1C,$1D,$9D,$CD,$1F,$37,$0F,$45
	.dc.b	$7C,$A3,$C3,$1F,$C4,$4D,$DF,$95,$F0,$1E,$53,$AC,$FE,$45,$B5,$E1
	.dc.b	$7B,$37,$0F,$87,$1E,$26,$CF,$8A,$3E,$C9,$D8,$F1,$3F,$57,$0F,$7B
	.dc.b	$63,$C4,$7C,$C7,$C4,$DA,$FC,$43,$FC,$54,$FA,$25,$AF,$E8,$4D,$C3
	.dc.b	$4F,$02,$1F,$28,$D1,$A3,$99,$A2,$9B,$3E,$03,$FA,$2F,$93,$7C,$9F
	.dc.b	$E1,$63,$0D,$9F,$C3,$4E,$9F,$C4,$57,$E4,$F0,$61,$FC,$BE,$B5,$5F
	.dc.b	$2F,$0E,$6F,$06,$B9,$3A,$E1,$DD,$FC,$1C,$F2,$BF,$A5,$18,$9D,$47
	.dc.b	$ED,$27,$57,$EC,$AE,$67,$72,$9D,$CE,$46,$4A,$66,$7A,$57,$69,$FC
	.dc.b	$45,$76,$AF,$C5,$90,$FB,$87,$F5,$F7,$4E,$09,$87,$F6,$18,$CB,$F2
	.dc.b	$67,$AD,$77,$AE,$95,$AC,$31,$08,$C3,$E9,$C3,$E8,$BE,$1A,$DE,$1D
	.dc.b	$DB,$7F,$D8,$EB,$75,$CE,$72,$FF,$77,$8C,$DF,$68,$F8,$98,$7D,$DA
	.dc.b	$DC,$2B,$71,$86,$C7,$96,$7D,$AD,$3D,$B3,$B8,$E9,$19,$3E,$45,$78
	.dc.b	$6B,$F0,$27,$48,$F1,$9C,$8E,$18,$7B,$DB,$87,$BE,$FA,$C7,$03,$AE
	.dc.b	$1E,$17,$E4,$1C,$0F,$85,$1C,$2F,$C2,$FA,$A3,$95,$F7,$6D,$F7,$79
	.dc.b	$9F,$16,$79,$1F,$95,$70,$6A,$BD,$87,$67,$11,$C1,$FC,$75,$77,$9D
	.dc.b	$CF,$36,$CF,$03,$75,$E3,$7E,$F1,$EC,$34,$7E,$51,$D6,$1F,$98,$3E
	.dc.b	$85,$7B,$EB,$AD,$6C,$E4,$74,$3D,$65,$7D,$4D,$8F,$53,$C4,$DC,$F7
	.dc.b	$C3,$F3,$AD,$8F,$C1,$35,$FB,$0D,$AE,$AD,$AF,$84,$76,$7D,$33,$F8
	.dc.b	$53,$BD,$76,$3D,$E5,$7D,$71,$C1,$FC,$29,$88,$D3,$C8,$F2,$BC,$9A
	.dc.b	$6C,$A6,$CF,$4B,$B6,$9B,$29,$BA,$F6,$C6,$DB,$1B,$6B,$BA,$F8,$2F
	.dc.b	$DC,$AF,$56,$1F,$87,$9F,$A2,$FF,$0F,$9D,$5F,$B2,$FE,$EC,$E6,$7E
	.dc.b	$C1,$F8,$67,$17,$F3,$E3,$F1,$A7,$8E,$FA,$E1,$FC,$F2,$F9,$B6,$33
	.dc.b	$3F,$CC,$A7,$F7,$4F,$F6,$1A,$FE,$8E,$DC,$3F,$AD,$3F,$8D,$30,$7F
	.dc.b	$5D,$8D,$BF,$BC,$32,$7F,$66,$72,$FE,$C8,$C9,$0F,$C3,$47,$CD,$AF
	.dc.b	$F4,$37,$C9,$3D,$5C,$57,$C6,$3F,$3F,$19,$7E,$01,$93,$F4,$2E,$5F
	.dc.b	$A8,$67,$0F,$E5,$F1,$97,$F5,$78,$7C,$B9,$4E,$6F,$CB,$1D,$8C,$DF
	.dc.b	$91,$FA,$A4,$9E,$0F,$79,$FD,$7C,$62,$37,$5D,$E3,$0F,$84,$FE,$AE
	.dc.b	$E3,$0E,$D1,$8B,$B1,$E1,$9D,$57,$96,$B4,$D9,$D5,$70,$8F,$55,$70
	.dc.b	$7E,$29,$E9,$8D,$38,$8D,$61,$F1,$AB,$F8,$D4,$63,$0F,$8B,$5F,$42
	.dc.b	$F8,$DF,$F1,$6A,$D4,$FC,$33,$FD,$A4,$E1,$F8,$77,$D2,$B8,$DF,$C8
	.dc.b	$6C,$FF,$26,$AE,$18,$7C,$99,$CC,$FA,$4F,$80,$7B,$9C,$C7,$C4,$8F
	.dc.b	$7B,$C5,$D9,$EF,$8D,$DF,$8E,$33,$87,$AE,$73,$1F,$92,$6B,$F0,$4D
	.dc.b	$E1,$E9,$69,$E9,$8D,$CF,$E4,$9B,$3F,$A0,$6C,$FC,$67,$C4,$FA,$2A
	.dc.b	$D7,$18,$EA,$7E,$50,$F2,$18,$C3,$F2,$07,$C8,$9E,$37,$DE,$75,$78
	.dc.b	$7E,$B1,$CC,$F6,$1E,$13,$A9,$8B,$FC,$23,$87,$C8,$EC,$D4,$F7,$71
	.dc.b	$1A,$24,$7E,$29,$B1,$F0,$4F,$29,$D6,$34,$73,$8D,$C7,$E2,$8D,$1C
	.dc.b	$4E,$F1,$B2,$3E,$EB,$F8,$B5,$F8,$A6,$C7,$80,$F4,$9B,$8F,$8C,$F7
	.dc.b	$32,$7C,$7A,$F8,$6E,$27,$E8,$3E,$43,$55,$EA,$7C,$D3,$EA,$C3,$F2
	.dc.b	$91,$D1,$B1,$9C,$3F,$77,$3C,$E7,$35,$EC,$8F,$25,$7C,$4B,$F0,$C7
	.dc.b	$8F,$0F,$71,$EA,$BE,$31,$E2,$BF,$23,$E6,$30,$7C,$13,$C1,$87,$89
	.dc.b	$F3,$37,$0F,$34,$7E,$C4,$D9,$4D,$95,$DB,$FB,$13,$65,$36,$57,$67
	.dc.b	$EC,$DF,$C9,$9F,$EB,$9C,$3F,$A0,$37,$0F,$EB,$31,$88,$FC,$91,$93
	.dc.b	$FD,$6C,$E5,$FC,$41,$93,$FD,$7E,$E5,$FD,$21,$93,$53,$F3,$E7,$FC
	.dc.b	$7D,$F6,$CF,$9E,$FC,$3D,$FE,$32,$FF,$C6,$B3,$97,$F0,$91,$9B,$FF
	.dc.b	$41,$CE,$67,$EC,$1F,$89,$71,$1D,$E3,$15,$DA,$FF,$2B,$5A,$C7,$F0
	.dc.b	$98,$75,$57,$27,$4F,$DE,$3F,$00,$47,$2E,$27,$93,$C6,$FC,$0F,$88
	.dc.b	$F5,$DF,$8A,$33,$1F,$A6,$AE,$75,$CA,$3B,$4E,$B0,$FD,$DD,$6A,$F1
	.dc.b	$87,$A2,$D3,$97,$B9,$FE,$62,$70,$FC,$83,$E7,$DF,$27,$DD,$1A,$BF
	.dc.b	$95,$1A,$39,$9E,$79,$E0,$F9,$8C,$C7,$9A,$33,$87,$F7,$DA,$E9,$7F
	.dc.b	$43,$0F,$C1,$D6,$9B,$87,$29,$FD,$C5,$70,$8F,$A5,$7C,$EF,$84,$7B
	.dc.b	$5A,$F9,$CF,$76,$1E,$37,$A1,$EE,$8D,$C7,$E3,$1B,$7F,$2E,$FC,$15
	.dc.b	$F7,$70,$F2,$BD,$27,$CB,$5C,$23,$F0,$07,$63,$E2,$1B,$3C,$06,$70
	.dc.b	$F5,$36,$BA,$E1,$F4,$0F,$44,$71,$39,$BE,$77,$B6,$1F,$92,$3C,$F5
	.dc.b	$D2,$3E,$14,$F3,$C3,$D9,$87,$67,$E9,$4F,$C9,$21,$DE,$FA,$4F,$DE
	.dc.b	$6C,$78,$23,$93,$F5,$30,$AF,$E4,$A3,$81,$8C,$3F,$39,$5C,$A7,$B4
	.dc.b	$E6,$BF,$49,$1E,$2C,$3D,$0F,$CC,$3D,$78,$7A,$1A,$66,$36,$F8,$4E
	.dc.b	$51,$B8,$F9,$11,$C0,$FB,$86,$8E,$51,$B3,$E3,$B9,$36,$71,$BF,$9F
	.dc.b	$5F,$01,$B1,$B6,$BB,$BF,$86,$F3,$C3,$E1,$C7,$C1,$8F,$E0,$C7,$96
	.dc.b	$32,$EE,$BE,$2B,$96,$E1,$EC,$3E,$0D,$FC,$D6,$DF,$5A,$FD,$C1,$CC
	.dc.b	$EF,$1C,$70,$F8,$11,$F0,$88,$7C,$08,$D3,$8C,$3C,$AF,$8C,$E7,$3F
	.dc.b	$21,$FB,$A7,$73,$C2,$6C,$B3,$F5,$1F,$E3,$47,$D9,$29,$4C,$DF,$F4
	.dc.b	$16,$DF,$FA,$CA,$31,$1E,$13,$27,$23,$91,$92,$1F,$8D,$36,$7D,$73
	.dc.b	$FA,$1C,$AB,$F9,$77,$9C,$F8,$AF,$C0,$61,$FA,$08,$C5,$FE,$06,$31
	.dc.b	$3A,$8F,$C5,$4E,$9E,$C7,$E4,$DC,$5F,$CE,$8F,$84,$D8,$CD,$FC,$D8
	.dc.b	$D6,$1F,$D4,$A7,$35,$F6,$23,$F0,$55,$E8,$BF,$14,$F7,$6D,$F5,$6C
	.dc.b	$F5,$38,$4F,$AC,$F4,$C6,$9F,$C2,$DF,$81,$F3,$CF,$17,$33,$B7,$27
	.dc.b	$F6,$E9,$E1,$7E,$68,$D9,$DC,$F3,$16,$FC,$CD,$BE,$CA,$B9,$E1,$DD
	.dc.b	$B1,$B3,$C1,$87,$8F,$0F,$90,$DA,$FE,$D5,$87,$DF,$AD,$9C,$4F,$49
	.dc.b	$DE,$78,$3F,$86,$C3,$E3,$BF,$72,$7C,$ED,$8F,$3E,$1E,$48,$C1,$D8
	.dc.b	$A7,$67,$53,$F7,$AF,$CB,$5D,$A3,$15,$F5,$0F,$36,$1E,$C6,$D7,$07
	.dc.b	$B4,$F0,$7E,$69,$E0,$9E,$0D,$9E,$0F,$BC,$E6,$7B,$CE,$75,$F4,$A7
	.dc.b	$ED,$46,$CE,$46,$CC,$57,$CD,$6C,$F1,$AE,$F7,$F8,$B3,$E0,$C7,$C8
	.dc.b	$C3,$27,$DD,$2D,$7D,$59,$E1,$19,$8F,$04,$72,$7F,$1A,$2F,$C5,$19
	.dc.b	$8F,$A1,$3D,$1F,$A7,$1D,$1E,$D1,$99,$F9,$47,$27,$15,$EB,$9F,$0C
	.dc.b	$61,$FD,$A5,$D7,$E0,$98,$36,$43,$C3,$1B,$C3,$E6,$E1,$E3,$9C,$9B
	.dc.b	$C3,$C7,$87,$E1,$49,$1D,$4F,$80,$E1,$F0,$9F,$9C,$AF,$C0,$CF,$AE
	.dc.b	$3D,$E7,$59,$F6,$3C,$8F,$C0,$90,$E5,$87,$B6,$FE,$65,$F4,$AF,$03
	.dc.b	$EA,$C3,$D9,$5F,$C9,$AF,$B1,$E7,$C3,$C4,$F7,$3C,$F5,$C5,$D9,$E0
	.dc.b	$2E,$1E,$0A,$F1,$5F,$E0,$E7,$94,$E5,$DC,$66,$FB,$9F,$2A,$3D,$93
	.dc.b	$B7,$8C,$FA,$1F,$D8,$18,$3E,$39,$4F,$AA,$59,$E8,$FD,$09,$E9,$1E
	.dc.b	$0C,$3E,$2D,$FC,$3A,$E1,$1F,$42,$75,$1F,$3F,$0F,$84,$F7,$3F,$B4
	.dc.b	$DF,$F4,$D6,$E1,$FD,$85,$FD,$A1,$83,$FB,$24,$6D,$FE,$32,$64,$FE
	.dc.b	$E2,$E5,$FE,$2C,$64,$87,$E2,$DF,$CA,$57,$EF,$7E,$D7,$FB,$DB,$E8
	.dc.b	$9F,$81,$5F,$C3,$27,$38,$71,$8F,$D2,$46,$27,$E8,$C6,$1F,$B6,$7E
	.dc.b	$02,$34,$F4,$30,$68,$F9,$27,$FA,$F1,$DD,$F9,$AB,$F4,$51,$8A,$E1
	.dc.b	$7E,$9B,$B5,$A6,$9F,$9F,$AE,$0F,$E2,$8F,$E3,$CE,$1F,$09,$FA,$E7
	.dc.b	$13,$F0,$EF,$FB,$44,$62,$BE,$15,$72,$3B,$9C,$4B,$3C,$1B,$7F,$2D
	.dc.b	$A7,$13,$C8,$78,$AB,$B3,$F3,$0D,$1E,$69,$F8,$8F,$C7,$39,$9F,$B1
	.dc.b	$8E,$2F,$CF,$30,$6A,$FE,$43,$E0,$7F,$01,$1C,$27,$AC,$A3,$15,$F7
	.dc.b	$63,$85,$FC,$F6,$9F,$3C,$EC,$FF,$B7,$E7,$8C,$7E,$00,$D9,$F1,$CB
	.dc.b	$5F,$1D,$B5,$F1,$DA,$7C,$2B,$FA,$91,$E3,$8F,$88,$D8,$CC,$77,$32
	.dc.b	$57,$2F,$CF,$3C,$E7,$CB,$3B,$5F,$DE,$7E,$49,$83,$EA,$18,$24,$74
	.dc.b	$3E,$6B,$88,$E5,$18,$23,$83,$E3,$47,$BC,$F3,$9C,$4F,$21,$88,$E2
	.dc.b	$E1,$F0,$47,$28,$F3,$1F,$81,$8D,$BF,$6C,$FC,$BD,$FE,$D0,$FA,$AF
	.dc.b	$73,$47,$D6,$74,$F0,$3C,$D1,$E0,$7C,$53,$E1,$7C,$CD,$8F,$0D,$FE
	.dc.b	$32,$B9,$D7,$53,$9D,$7C,$37,$99,$0F,$53,$CC,$8F,$D2,$AF,$6C,$FA
	.dc.b	$9B,$1E,$73,$E1,$D7,$85,$EE,$71,$AE,$6F,$C0,$3D,$27,$86,$BA,$46
	.dc.b	$9B,$19,$3B,$1F,$26,$7B,$3E,$0B,$FC,$A5,$76,$9C,$CE,$E7,$2F,$BE
	.dc.b	$33,$7F,$22,$73,$3E,$58,$FA,$0D,$AE,$93,$AB,$F2,$CE,$4E,$91,$97
	.dc.b	$CF,$5F,$96,$7F,$D7,$73,$FD,$45,$F8,$46,$0F,$EA,$8E,$48,$E0,$87
	.dc.b	$FA,$E5,$CB,$4C,$9F,$8D,$7F,$A1,$FD,$AB,$DC,$FA,$73,$E7,$24,$65
	.dc.b	$FC,$DC,$FE,$21,$FF,$07,$46,$5E,$87,$27,$15,$F8,$28,$FD,$1C,$63
	.dc.b	$0F,$B5,$1B,$BF,$0D,$F5,$BF,$F3,$05,$EE,$BE,$F1,$CA,$B6,$F1,$3C
	.dc.b	$A5,$3C,$66,$CC,$94,$C9,$FA,$F8,$EC,$FC,$1A,$F1,$5A,$24,$EA,$73
	.dc.b	$1F,$A5,$8C,$3E,$43,$FC,$F4,$E1,$F0,$9F,$22,$B4,$F8,$5B,$7F,$C1
	.dc.b	$A7,$84,$F9,$A3,$E1,$CE,$AF,$F4,$F5,$A9,$F9,$D5,$D4,$EE,$F3,$AE
	.dc.b	$11,$E7,$6B,$F7,$9B,$7F,$84,$6C,$7F,$A7,$2B,$B5,$FD,$88,$F6,$CE
	.dc.b	$9A,$78,$23,$CB,$19,$8F,$21,$EB,$8C,$BB,$3C,$85,$3C,$85,$AF,$23
	.dc.b	$4F,$24,$F5,$BF,$8B,$19,$3C,$67,$73,$A9,$E2,$7E,$39,$C1,$C5,$7B
	.dc.b	$5A,$7E,$02,$7E,$3D,$FE,$18,$F8,$71,$D2,$B9,$1D,$2B,$F1,$10,$9E
	.dc.b	$15,$C8,$F4,$C7,$BE,$32,$D8,$CB,$E9,$39,$38,$AE,$D1,$E5,$3D,$EE
	.dc.b	$1D,$98,$AF,$64,$79,$9F,$2D,$7A,$A7,$E7,$9F,$00,$FD,$11,$DE,$37
	.dc.b	$5E,$98,$E8,$7E,$D0,$E1,$3C,$EF,$EB,$C7,$37,$F4,$EF,$90,$C1,$FB
	.dc.b	$7B,$F2,$3C,$DA,$ED,$F6,$4F,$C6,$8F,$88,$D8,$CD,$79,$E7,$95,$FA
	.dc.b	$1F,$39,$EC,$2D,$79,$DE,$07,$78,$F8,$E7,$17,$CE,$7F,$09,$AF,$2D
	.dc.b	$FB,$2B,$C5,$5C,$2B,$EA,$1C,$0F,$DF,$1E,$29,$E1,$7E,$06,$C6,$4F
	.dc.b	$89,$3E,$1B,$F6,$C6,$4F,$14,$64,$F0,$3E,$93,$C3,$1C,$1F,$13,$FC
	.dc.b	$0D,$E4,$79,$0C,$C7,$ED,$1A,$7E,$75,$F8,$4F,$E7,$0F,$B4,$5B,$FE
	.dc.b	$84,$DB,$FF,$55,$46,$23,$89,$0F,$8A,$6C,$E8,$60,$D1,$D8,$F7,$1E
	.dc.b	$C9,$BE,$7B,$F0,$97,$C2,$7D,$AB,$FA,$76,$AD,$4E,$70,$FC,$34,$62
	.dc.b	$7F,$9B,$FB,$DE,$82,$3E,$43,$F7,$4E,$1F,$3D,$A3,$11,$F0,$DA,$FE
	.dc.b	$B9,$AF,$F4,$A7,$E2,$FB,$92,$3B,$4A,$31,$1D,$9A,$FC,$F6,$BF,$BA
	.dc.b	$6B,$FD,$59,$AF,$F1,$1E,$CF,$A6,$47,$BB,$5F,$5B,$5F,$B8,$D7,$F8
	.dc.b	$3B,$5F,$EC,$6F,$DE,$F8,$6F,$E3,$FA,$B5,$FE,$D1,$EC,$3E,$7D,$F0
	.dc.b	$7A,$9C,$C6,$31,$FE,$CF,$87,$F4,$F3,$4A,$F9,$5E,$1B,$1C,$08,$30
	.dc.b	$F1,$3C,$6F,$B3,$9C,$3C,$37,$DE,$38,$9F,$9B,$C3,$D8,$FD,$46,$BF
	.dc.b	$80,$6B,$B3,$44,$34,$53,$47,$F8,$A1,$F0,$3E,$33,$FC,$1D,$E1,$7E
	.dc.b	$69,$F5,$E1,$FE,$22,$9D,$BE,$C8,$4E,$DE,$87,$0B,$FE,$7B,$5F,$2A
	.dc.b	$37,$C1,$CA,$73,$1B,$3C,$31,$97,$FC,$3F,$35,$E2,$6C,$FE,$B8,$E6
	.dc.b	$BC,$57,$E7,$9D,$6A,$AF,$15,$6B,$0F,$0D,$7A,$09,$7E,$D9,$D4,$78
	.dc.b	$8E,$75,$CB,$0F,$26,$1D,$67,$C9,$5C,$1F,$29,$C0,$FF,$1D,$5F,$5A
	.dc.b	$E3,$7E,$96,$CF,$26,$BE,$4C,$3F,$BD,$DF,$5C,$3C,$13,$93,$CC,$7D
	.dc.b	$B3,$E3,$96,$38,$5F,$78,$F1,$D7,$4B,$EF,$87,$96,$72,$7D,$43,$D4
	.dc.b	$73,$8F,$84,$E2,$38,$39,$31,$3D,$5E,$67,$92,$3E,$B3,$E9,$31,$1B
	.dc.b	$71,$87,$BA,$7A,$92,$7E,$CB,$C8,$C4,$F9,$CE,$37,$E6,$7C,$86,$0F
	.dc.b	$1E,$1F,$5C,$F8,$AF,$D1,$33,$87,$C4,$BF,$25,$F6,$9F,$BA,$7A,$6B
	.dc.b	$95,$6E,$3C,$53,$BC,$3D,$57,$E8,$39,$92,$7E,$CC,$FA,$E3,$11,$F1
	.dc.b	$0C,$BC,$5E,$AE,$A7,$AC,$68,$C9,$F0,$CF,$ED,$17,$FB,$76,$DF,$F4
	.dc.b	$D7,$F0,$E6,$0F,$EA,$11,$B7,$EA,$19,$3F,$B3,$B9,$7B,$1A,$21,$E2
	.dc.b	$3D,$44,$5F,$0D,$F5,$70,$F9,$91,$F4,$E3,$18,$7C,$28,$C5,$7D,$42
	.dc.b	$1F,$72,$BF,$71,$5C,$23,$C1,$3C,$1D,$57,$3B,$FE,$07,$5B,$9F,$24
	.dc.b	$61,$EB,$7E,$48,$C4,$72,$3A,$B8,$8D,$47,$0A,$FD,$BD,$75,$BF,$C2
	.dc.b	$CF,$E4,$23,$17,$96,$C7,$13,$F9,$9B,$87,$C0,$7F,$30,$71,$7F,$21
	.dc.b	$B7,$F1,$DB,$3F,$09,$AF,$D1,$7D,$11,$F0,$E3,$C8,$43,$BD,$FC,$9A
	.dc.b	$F9,$B1,$E5,$39,$1B,$7E,$0B,$FC,$3E,$BA,$BF,$18,$D5,$7C,$47,$95
	.dc.b	$75,$8C,$47,$82,$F8,$BE,$F3,$6E,$27,$F1,$31,$83,$A9,$F2,$0C,$1E
	.dc.b	$33,$27,$DE,$2B,$C9,$F0,$18,$AF,$8F,$19,$BF,$83,$3A,$3A,$11,$F8
	.dc.b	$26,$8D,$BF,$B0,$9F,$D7,$4F,$78,$C1,$FA,$E8,$C9,$1C,$10,$FD,$7B
	.dc.b	$96,$99,$9F,$05,$FC,$A7,$FC,$23,$A3,$E0,$1F,$21,$F2,$70,$FA,$B3
	.dc.b	$98,$C6,$1F,$2A,$30,$FF,$26,$F8,$7F,$F1,$05,$FF,$0E,$9E,$9A,$FF
	.dc.b	$6C,$99,$5F,$D2,$27,$2D,$AF,$E2,$AF,$F8,$22,$76,$FF,$84,$67,$6D
	.dc.b	$9D,$E6,$AD,$CF,$F1,$A7,$E5,$91,$FE,$2E,$D7,$F8,$AB,$5E,$06,$88
	.dc.b	$68,$B9,$7F,$80,$63,$CA,$FF,$87,$FE,$DB,$FC,$72,$E6,$7E,$21,$0F
	.dc.b	$EE,$11,$AD,$3F,$BB,$D7,$26,$CF,$F6,$29,$D3,$FE,$30,$9D,$34,$D6
	.dc.b	$F8,$7E,$41,$DE,$FC,$91,$9C,$3E,$34,$E7,$0F,$F8,$03,$87,$68,$E2
	.dc.b	$7F,$80,$F0,$E9,$19,$AE,$33,$93,$90,$11,$B9,$F0,$1F,$E3,$D7,$11
	.dc.b	$C0,$EF,$87,$99,$C5,$76,$8F,$95,$1F,$23,$0E,$6F,$1A,$E6,$F8,$C9
	.dc.b	$3C,$DB,$87,$CC,$69,$E7,$9E,$33,$D8,$93,$F1,$2B,$D0,$78,$2B,$C4
	.dc.b	$7D,$03,$CF,$86,$9C,$1D,$8F,$6C,$F3,$73,$5D,$9F,$B4,$74,$23,$97
	.dc.b	$89,$E7,$AE,$CF,$5C,$3E,$35,$75,$8E,$2F,$BE,$B9,$61,$F7,$4F,$7C
	.dc.b	$64,$E1,$1F,$1C,$F0,$57,$4A,$F0,$1E,$89,$E0,$FA,$27,$84,$72,$33
	.dc.b	$1E,$83,$F0,$77,$E8,$9E,$B8,$5A,$E1,$5C,$DE,$38,$79,$EF,$E8,$5F
	.dc.b	$6A,$F1,$9F,$06,$F9,$38,$8E,$AF,$9C,$C5,$F9,$30,$F6,$47,$90,$D1
	.dc.b	$E7,$73,$87,$86,$3F,$DD,$8F,$FA,$D1,$FC,$F1,$E2,$29,$CC,$A5,$32
	.dc.b	$F1,$7E,$B1,$8C,$3F,$C5,$91,$98,$E8,$68,$87,$A8,$87,$7B,$FE,$21
	.dc.b	$EE,$FC,$28,$C6,$1F,$31,$FD,$9C,$66,$BD,$91,$FC,$4A,$31,$79,$6C
	.dc.b	$7C,$49,$46,$23,$F1,$A7,$EA,$5C,$5F,$D5,$8E,$E7,$1A,$E8,$D3,$A4
	.dc.b	$F0,$BF,$35,$F6,$9D,$1D,$8E,$65,$7D,$37,$F0,$23,$6F,$86,$39,$46
	.dc.b	$EF,$C2,$D3,$C2,$70,$AF,$0D,$F8,$E3,$E6,$D7,$2B,$E9,$7C,$9F,$04
	.dc.b	$F3,$3E,$51,$FC,$DE,$B9,$3E,$88,$D9,$C9,$D1,$0E,$4E,$AF,$F9,$9C
	.dc.b	$F2,$BF,$14,$62,$BD,$07,$D1,$39,$57,$57,$C6,$47,$07,$57,$81,$E0
	.dc.b	$8E,$8E,$DF,$A7,$7F,$CD,$EE,$4F,$80,$F6,$4F,$0A,$EA,$D2,$47,$CA
	.dc.b	$AE,$2F,$53,$07,$A2,$36,$FE,$60,$FA,$35,$C1,$FD,$4B,$FA,$62,$3F
	.dc.b	$B4,$6B,$FB,$0B,$FC,$0B,$C0,$D1,$0D,$47,$07,$47,$94,$F6,$8C,$FF
	.dc.b	$C4,$9E,$9F,$AB,$1F,$95,$3A,$9C,$BE,$B3,$E6,$C7,$CF,$8C,$C7,$FB
	.dc.b	$BF,$ED,$9E,$E7,$FA,$AD,$B1,$FC,$20,$38,$7F,$85,$B5,$FE,$1E,$D7
	.dc.b	$F8,$9B,$5F,$EE,$4D,$79,$C7,$72,$3D,$E3,$44,$F1,$BC,$84,$8D,$35
	.dc.b	$F1,$B5,$F1,$B5,$F1,$B5,$F1,$BE,$FF,$26,$47,$85,$AF,$B9,$AF,$B9
	.dc.b	$AF,$B9,$AF,$B9,$AF,$C8,$DB,$E6,$11,$F7,$35,$F9,$6D,$7E,$5B,$5F
	.dc.b	$96,$D7,$E5,$BF,$5F,$D2,$91,$F2,$5A,$FD,$66,$BF,$59,$AF,$D6,$6B
	.dc.b	$F5,$9A,$FE,$03,$B3,$F0,$84,$7E,$B3,$5F,$C1,$B5,$FC,$1B,$5F,$C1
	.dc.b	$B5,$FC,$1B,$F8,$FE,$BF,$F3,$AF,$7C,$9F,$F1,$C6,$AE,$1F,$C5,$D7
	.dc.b	$F5,$E8,$C4,$FE,$32,$38,$3F,$E7,$1A,$D5,$70,$8E,$6D,$74,$7F,$9F
	.dc.b	$63,$6F,$33,$65,$3F,$3F,$3E,$77,$F2,$CD,$72,$60,$86,$08,$7E,$99
	.dc.b	$E0,$D3,$B9,$FA,$E7,$C6,$D3,$3A,$3E,$1B,$2A,$69,$27,$73,$5F,$DF
	.dc.b	$45,$C3,$E4,$C7,$EF,$27,$18,$7C,$59,$CD,$F5,$AE,$67,$8A,$F6,$FC
	.dc.b	$59,$DD,$E2,$3D,$07,$80,$E2,$52,$47,$69,$F5,$CF,$1A,$F0,$56,$AF
	.dc.b	$F8,$5C,$E6,$B4,$D9,$F2,$1D,$AF,$A3,$83,$91,$FC,$4A,$38,$38,$3E
	.dc.b	$11,$B8,$E2,$E6,$FE,$83,$9B,$F8,$C4,$7A,$19,$3E,$23,$97,$A9,$FC
	.dc.b	$BE,$BC,$35,$F2,$23,$D1,$3B,$AF,$83,$39,$9E,$07,$53,$E6,$B9,$24
	.dc.b	$F1,$BF,$97,$3D,$0F,$19,$0F,$1C,$F0,$71,$5E,$0A,$F9,$F1,$C4,$EE
	.dc.b	$F9,$6F,$AC,$7C,$5B,$B7,$CE,$36,$74,$9C,$D7,$96,$3A,$5F,$CC,$79
	.dc.b	$18,$9F,$8B,$3E,$28,$F8,$26,$8B,$5D,$EB,$D1,$1D,$CE,$B7,$F8,$66
	.dc.b	$9F,$26,$3C,$13,$C6,$75,$3D,$2B,$CC,$D3,$EE,$9F,$D8,$2F,$FB,$31
	.dc.b	$1F,$EB,$0D,$74,$60,$86,$08,$7F,$68,$78,$57,$AD,$FE,$03,$F6,$CF
	.dc.b	$87,$F9,$35,$E2,$AF,$9D,$39,$8F,$90,$7D,$77,$0F,$A4,$FB,$F7,$F2
	.dc.b	$E7,$F8,$0C,$E9,$F0,$4F,$1A,$EE,$F0,$3C,$15,$FE,$8E,$AE,$0F,$32
	.dc.b	$46,$8F,$F7,$84,$69,$C1,$CC,$8E,$5C,$63,$F5,$27,$9C,$27,$53,$98
	.dc.b	$E6,$D7,$55,$E8,$8C,$19,$9E,$38,$64,$F9,$B3,$6B,$E0,$1F,$28,$F9
	.dc.b	$13,$C1,$C6,$1F,$22,$78,$46,$9B,$1F,$51,$B8,$7A,$AB,$DC,$73,$AE
	.dc.b	$07,$F4,$07,$15,$C6,$31,$5C,$9F,$96,$FA,$1F,$1D,$7D,$38,$C4,$7D
	.dc.b	$FC,$3E,$04,$62,$BE,$EC,$9F,$53,$D8,$E6,$76,$30,$73,$C3,$95,$F7
	.dc.b	$8D,$9F,$40,$DC,$F6,$3A,$BD,$E3,$83,$CC,$DB,$88,$E0,$D2,$47,$DE
	.dc.b	$9F,$0B,$E5,$39,$1D,$4C,$4F,$80,$E6,$E2,$7C,$2F,$FB,$93,$0F,$D2
	.dc.b	$B6,$FF,$CB,$CF,$AC,$FF,$73,$C6,$9E,$24,$3C,$47,$13,$FC,$CC,$ED
	.dc.b	$C9,$B3,$A5,$79,$6B,$DB,$C2,$E7,$7C,$DF,$BA,$F7,$5A,$9C,$C7,$C8
	.dc.b	$3D,$84,$AE,$55,$FC,$23,$E2,$3F,$C3,$71,$D7,$E6,$ED,$9D,$5F,$E4
	.dc.b	$6B,$31,$23,$04,$3E,$75,$73,$7F,$C4,$AD,$7E,$91,$FB,$79,$EA,$FC
	.dc.b	$A7,$E6,$E4,$69,$23,$E4,$B5,$E0,$60,$86,$08,$7D,$B7,$BE,$5F,$C7
	.dc.b	$1F,$F1,$4F,$C2,$0C,$E7,$FC,$66,$2D,$0D,$4E,$1F,$51,$F0,$6B,$FC
	.dc.b	$4F,$1C,$E3,$15,$CD,$A7,$38,$D3,$F1,$2F,$E0,$3A,$3A,$9F,$22,$37
	.dc.b	$5E,$98,$D4,$F0,$75,$3E,$58,$C4,$EE,$72,$70,$BF,$55,$74,$BF,$34
	.dc.b	$F8,$E7,$33,$CD,$B1,$C4,$F3,$47,$17,$F9,$DC,$6D,$FF,$2B,$C6,$5F
	.dc.b	$5D,$78,$C9,$5F,$36,$3E,$69,$D4,$F1,$1C,$A7,$A3,$83,$D6,$49,$E6
	.dc.b	$F8,$8E,$04,$38,$3A,$7B,$9C,$67,$E0,$C6,$67,$D0,$7A,$6F,$E5,$CE
	.dc.b	$8F,$44,$68,$F2,$CF,$E0,$9F,$C6,$98,$8D,$B8,$AF,$74,$ED,$F3,$3E
	.dc.b	$5A,$EF,$1F,$87,$9E,$0F,$AA,$BC,$4E,$4E,$2E,$1F,$9E,$F8,$1C,$9F
	.dc.b	$14,$E4,$71,$3E,$11,$9A,$F8,$53,$E0,$7C,$97,$6B,$A5,$72,$3B,$9C
	.dc.b	$E7,$D8,$E6,$FD,$6F,$D0,$8C,$9D,$1F,$29,$99,$F5,$4F,$0B,$F9,$D3
	.dc.b	$F2,$27,$C3,$5D,$A3,$F1,$47,$DC,$2C,$FE,$B9,$AF,$F3,$A3,$0F,$A8
	.dc.b	$C9,$C8,$F8,$86,$0C,$3C,$1C,$9C,$4F,$1C,$C7,$E1,$3F,$86,$F8,$7F
	.dc.b	$91,$18,$21,$FC,$45,$CB,$F0,$4C,$CA,$32,$FC,$28,$FE,$09,$18,$8C
	.dc.b	$B6,$3E,$5C,$71,$6B,$A6,$9A,$8E,$0F,$E0,$8F,$21,$A3,$11,$CD,$AF
	.dc.b	$36,$9C,$CE,$15,$F7,$AB,$E1,$46,$1F,$8C,$F6,$38,$9D,$AB,$C7,$3C
	.dc.b	$63,$F9,$44,$62,$78,$D7,$8E,$B5,$5F,$24,$91,$99,$F0,$C7,$69,$FE
	.dc.b	$6F,$5F,$D0,$8E,$91,$D8,$FE,$32,$E1,$F3,$47,$07,$D4,$6C,$FA,$66
	.dc.b	$0D,$CF,$82,$3C,$B5,$FC,$32,$BC,$15,$B8,$F3,$3C,$E7,$71,$A8,$EA
	.dc.b	$7A,$6B,$A1,$1F,$C3,$C6,$E7,$CD,$1B,$3A,$4E,$DF,$8B,$3B,$8C,$3F
	.dc.b	$8D,$9E,$24,$8F,$9F,$1B,$9E,$67,$80,$F8,$E4,$3F,$4F,$5F,$AF,$23
	.dc.b	$FA,$D3,$F0,$07,$ED,$9C,$BC,$0D,$47,$E3,$DD,$18,$8D,$B8,$3E,$2D
	.dc.b	$79,$D6,$3F,$E2,$9F,$99,$F4,$47,$D4,$3D,$31,$97,$D1,$3C,$5D,$46
	.dc.b	$5F,$EE,$5E,$CF,$F0,$92,$3E,$66,$BF,$C1,$9A,$FF,$12,$6B,$FD,$21
	.dc.b	$AF,$F8,$0D,$FE,$29,$F0,$F2,$3E,$0B,$5F,$E2,$2D,$7F,$8E,$B5,$FE
	.dc.b	$A6,$D7,$FC,$1C,$D9,$FF,$25,$7C,$DF,$F8,$F8,$93,$FE,$36,$DD,$F4
	.dc.b	$18,$8F,$F1,$DC,$27,$18,$B5,$FF,$12,$61,$FE,$55,$C3,$31,$8B,$E1
	.dc.b	$87,$5C,$3F,$C9,$B5,$A8,$FA,$B3,$A7,$BC,$E9,$E0,$43,$27,$9E,$79
	.dc.b	$5F,$B2,$35,$87,$95,$EB,$3F,$E0,$D8,$C4,$3E,$D9,$DB,$FE,$7C,$6C
	.dc.b	$7E,$55,$AF,$F4,$36,$BF,$D3,$9A,$FF,$86,$5A,$FF,$91,$DA,$F7,$8E
	.dc.b	$C4,$7B,$35,$ED,$DB,$E3,$23,$E3,$6B,$A3,$04,$30,$43,$CC,$F0,$7D
	.dc.b	$1D,$BE,$F2,$3E,$F6,$BE,$F6,$BC,$0D,$10,$D1,$4F,$89,$DB,$F2,$C8
	.dc.b	$FC,$B6,$BF,$2D,$AF,$03,$44,$34,$53,$E8,$76,$7D,$BD,$BF,$C9,$1F
	.dc.b	$AA,$BE,$B3,$15,$FE,$77,$8C,$3F,$E7,$9D,$3F,$03,$A3,$0D,$CF,$E0
	.dc.b	$DB,$5C,$23,$FD,$2E,$DA,$E3,$3C,$2B,$95,$75,$C3,$C7,$87,$23,$55
	.dc.b	$27,$F1,$8F,$E3,$C8,$E0,$87,$E3,$DC,$B4,$C9,$F0,$8A,$7E,$45,$EC
	.dc.b	$FF,$AB,$AB,$CA,$DD,$F3,$87,$AA,$3D,$07,$E5,$EB,$D9,$3A,$C3,$A4
	.dc.b	$E8,$F8,$C5,$BF,$DC,$5F,$38,$F0,$5F,$17,$F4,$57,$5E,$26,$8E,$4E
	.dc.b	$6B,$CF,$5C,$CF,$AF,$19,$AD,$36,$BC,$51,$E8,$69,$0E,$C7,$C1,$31
	.dc.b	$3D,$6F,$DB,$19,$37,$19,$7E,$21,$E6,$38,$C7,$08,$D4,$7F,$0B,$AF
	.dc.b	$23,$CA,$7F,$82,$46,$27,$77,$F2,$27,$91,$2B,$B9,$E3,$9F,$82,$F8
	.dc.b	$0C,$B8,$75,$1C,$5F,$9B,$19,$24,$66,$7E,$6B,$E4,$9F,$61,$E7,$3A
	.dc.b	$B9,$7A,$98,$21,$EF,$38,$CF,$B2,$FE,$94,$70,$3C,$73,$E1,$BC,$D6
	.dc.b	$E7,$8C,$F7,$8C,$47,$1B,$F8,$CF,$81,$F4,$BF,$8C,$38,$99,$BE,$B5
	.dc.b	$CC,$F8,$0E,$23,$81,$EF,$3C,$A6,$2B,$E3,$BC,$CE,$A7,$BC,$E1,$5E
	.dc.b	$27,$11,$C1,$E6,$48,$F4,$BA,$AF,$1B,$D2,$32,$73,$25,$FC,$A8,$FE
	.dc.b	$9C,$FF,$55,$23,$FD,$55,$AF,$F5,$56,$BC,$0D,$10,$D1,$6B,$D1,$3E
	.dc.b	$6F,$B7,$7B,$1F,$1E,$BD,$07,$C5,$C3,$FD,$ED,$3A,$8F,$56,$1B,$8C
	.dc.b	$47,$D9,$8C,$57,$3B,$F7,$C7,$F3,$0A,$DE,$1E,$28,$DD,$F4,$3E,$8C
	.dc.b	$EB,$0F,$B9,$1E,$32,$08,$D3,$8B,$EE,$FF,$19,$8C,$BE,$C2,$47,$C1
	.dc.b	$C3,$CD,$7E,$29,$F5,$1E,$28,$CB,$4F,$31,$F9,$FA,$EB,$5E,$08,$C4
	.dc.b	$F0,$69,$C3,$0F,$C4,$57,$0B,$D4,$7F,$85,$A3,$CE,$66,$7C,$02,$35
	.dc.b	$5C,$30,$FA,$AF,$CD,$31,$23,$E3,$39,$7C,$51,$C1,$F8,$D1,$C1,$D9
	.dc.b	$83,$67,$3A,$F2,$C6,$DB,$1A,$38,$9E,$79,$EC,$FC,$C3,$07,$39,$F2
	.dc.b	$39,$AF,$6B,$E1,$31,$3C,$A3,$31,$83,$91,$B7,$F1,$8D,$8E,$07,$E2
	.dc.b	$9F,$C1,$57,$A1,$E4,$7A,$27,$AB,$F2,$4E,$A5,$29,$63,$F4,$2D,$AF
	.dc.b	$F4,$F3,$F0,$8C,$1F,$EA,$18,$DB,$F1,$0C,$9D,$4E,$06,$4F,$49,$E2
	.dc.b	$9F,$82,$7F,$97,$3F,$07,$72,$FC,$A8,$CE,$1C,$29,$5A,$7E,$21,$CE
	.dc.b	$35,$19,$8F,$AA,$7E,$B7,$E0,$BF,$C4,$F1,$FE,$D6,$8E,$7D,$1F,$BB
	.dc.b	$99,$1F,$E0,$79,$D1,$BC,$53,$A9,$EB,$3A,$AC,$61,$FA,$1A,$F9,$37
	.dc.b	$C2,$FE,$44,$60,$8E,$08,$7C,$98,$E0,$D8,$DB,$4D,$9F,$65,$F0,$E1
	.dc.b	$FC,$79,$FF,$22,$FD,$C3,$E8,$73,$5F,$A1,$9A,$A7,$4F,$F0,$F8,$CD
	.dc.b	$F8,$B0,$FF,$48,$D6,$AB,$E6,$CE,$70,$F0,$47,$13,$FC,$61,$3B,$AF
	.dc.b	$89,$19,$38,$DF,$9B,$0E,$EF,$92,$B8,$5F,$4A,$D9,$C6,$FC,$77,$C2
	.dc.b	$31,$3F,$9F,$AE,$31,$EF,$32,$53,$F0,$AE,$9F,$C3,$1A,$29,$A9,$EC
	.dc.b	$F8,$8C,$1B,$C3,$E1,$3D,$6F,$CA,$E7,$0F,$65,$72,$9F,$87,$3C,$8E
	.dc.b	$8E,$09,$87,$AE,$7E,$34,$F3,$3E,$6B,$87,$8C,$7C,$79,$F0,$C7,$57
	.dc.b	$9B,$E5,$8E,$8F,$23,$A1,$B7,$2F,$43,$81,$FC,$CF,$0F,$90,$7C,$9C
	.dc.b	$3E,$0C,$7C,$8A,$F0,$92,$BC,$84,$8E,$57,$E9,$9F,$2C,$71,$9F,$0C
	.dc.b	$71,$8F,$09,$93,$89,$E9,$3C,$C6,$4F,$CE,$4F,$59,$F0,$46,$6F,$D3
	.dc.b	$3F,$4E,$7B,$92,$7B,$DF,$B2,$B7,$1F,$1A,$FC,$8F,$63,$15,$C5,$C1
	.dc.b	$E4,$AE,$57,$E3,$8E,$B8,$57,$77,$F0,$AB,$71,$ED,$3F,$9C,$61,$FB
	.dc.b	$16,$DF,$F3,$68,$C4,$7D,$83,$27,$F3,$33,$C6,$60,$FE,$6C,$F0,$7E
	.dc.b	$59,$88,$E8,$47,$E7,$57,$C0,$97,$C6,$FA,$5F,$8B,$3F,$AC,$8C,$56
	.dc.b	$5B,$1C,$B1,$FD,$B4,$62,$3F,$20,$7E,$35,$C5,$F7,$6C,$FC,$DB,$F0
	.dc.b	$D7,$33,$D3,$5B,$70,$7C,$C9,$E1,$5C,$A3,$71,$E2,$77,$5E,$AA,$FE
	.dc.b	$3F,$5B,$9E,$47,$94,$DD,$FB,$5A,$72,$8E,$15,$FA,$66,$CF,$F3,$46
	.dc.b	$CF,$F2,$66,$BF,$24,$C1,$D0,$C1,$2B,$E9,$4F,$38,$D9,$F2,$8F,$39
	.dc.b	$FC,$F6,$74,$E6,$BC,$8F,$0A,$E6,$DB,$F8,$8E,$EF,$8C,$FC,$48,$F2
	.dc.b	$35,$F2,$4F,$D5,$9D,$BC,$A7,$AC,$E8,$F9,$47,$53,$C0,$61,$D9,$D4
	.dc.b	$91,$D6,$36,$7B,$4F,$1B,$DC,$8F,$DE,$21,$B3,$C4,$7C,$6A,$F8,$6F
	.dc.b	$CE,$3F,$5F,$7F,$AD,$23,$FA,$B8,$FC,$8B,$FA,$A8,$CB,$4D,$38,$79
	.dc.b	$98,$21,$E2,$8E,$74,$AF,$F1,$C7,$E0,$EF,$89,$F9,$77,$F1,$67,$DA
	.dc.b	$A7,$4F,$C4,$BE,$73,$A9,$CC,$FF,$60,$D3,$F8,$41,$1F,$63,$5F,$E0
	.dc.b	$AD,$7F,$88,$B5,$FE,$DC,$D7,$FA,$23,$FC,$4F,$8E,$47,$C7,$6B,$FC
	.dc.b	$41,$AF,$F1,$C6,$BF,$DE,$5A,$FF,$4F,$6B,$FE,$17,$DF,$F9,$01,$1F
	.dc.b	$A8,$D7,$F8,$EB,$5F,$E5,$2D,$7F,$C0,$CD,$7F,$AE,$3F,$CA,$FB,$24
	.dc.b	$7D,$F6,$BF,$CA,$1A,$FF,$34,$6B,$FE,$11,$6B,$FD,$9D,$AF,$F8,$BF
	.dc.b	$CA,$FE,$70,$47,$F1,$8D,$7F,$9A,$B5,$FE,$82,$D7,$FC,$3C,$D7,$FB
	.dc.b	$A3,$FE,$78,$F9,$EF,$C5,$47,$F4,$16,$BF,$E7,$56,$BF,$E8,$36,$BF
	.dc.b	$E8,$76,$BF,$EB,$16,$A6,$93,$2F,$01,$23,$C0,$F8,$FA,$BF,$D2,$5B
	.dc.b	$78,$BF,$51,$FF,$18,$1F,$E3,$1B,$4E,$AB,$C4,$DB,$FF,$61,$57,$0E
	.dc.b	$1C,$47,$8E,$6D,$F2,$8F,$F6,$46,$1D,$EB,$77,$DE,$F6,$79,$6B,$E3
	.dc.b	$35,$F8,$CD,$7E,$33,$5F,$29,$C0,$87,$02,$D7,$D0,$BF,$24,$7F,$A5
	.dc.b	$1B,$19,$F2,$BC,$73,$E8,$C7,$AB,$BB,$F6,$CE,$A3,$06,$CF,$AD,$1A
	.dc.b	$70,$7C,$52,$DF,$82,$BA,$DF,$86,$73,$5D,$63,$D0,$F3,$8E,$E4,$8C
	.dc.b	$BD,$2F,$85,$74,$24,$7E,$26,$32,$F9,$8E,$E7,$12,$9D,$63,$8B,$F8
	.dc.b	$F8,$E2,$FA,$CD,$1D,$4E,$33,$E4,$72,$71,$33,$3D,$2F,$E1,$47,$13
	.dc.b	$6E,$1E,$D3,$F1,$5C,$DF,$8A,$57,$63,$0F,$C7,$AF,$11,$2F,$F5,$91
	.dc.b	$97,$C5,$36,$FC,$55,$B7,$E3,$1F,$87,$AD,$3E,$4A,$F5,$46,$2F,$D0
	.dc.b	$47,$17,$E4,$AF,$9C,$7A,$8B,$7E,$D3,$C0,$F5,$3C,$77,$E8,$7E,$49
	.dc.b	$E4,$8E,$11,$98,$E8,$79,$CE,$AF,$B9,$A7,$43,$93,$EA,$72,$78,$49
	.dc.b	$7F,$39,$CB,$EE,$AF,$81,$7F,$46,$72,$D7,$C4,$4B,$F6,$D7,$C5,$8E
	.dc.b	$75,$F1,$5D,$9F,$C8,$EB,$F9,$51,$1F,$E5,$4D,$7F,$95,$35,$E0,$68
	.dc.b	$86,$8B,$5F,$21,$F8,$43,$EE,$BF,$96,$3E,$A3,$E1,$B8,$7D,$58,$7C
	.dc.b	$F9,$E5,$1C,$70,$F8,$6F,$06,$BC,$01,$B3,$B1,$93,$05,$7E,$83,$9A
	.dc.b	$F1,$4F,$82,$72,$F8,$A3,$35,$E5,$25,$7B,$A3,$2F,$C0,$AE,$33,$A8
	.dc.b	$F0,$D7,$5B,$FC,$FD,$F0,$9F,$DD,$1E,$58,$FE,$83,$19,$BE,$73,$D6
	.dc.b	$73,$1F,$0A,$BC,$CD,$BF,$05,$F1,$AC,$5F,$C6,$8C,$3F,$1A,$FF,$B4
	.dc.b	$C6,$23,$C4,$7F,$2E,$7C,$C6,$63,$BC,$66,$FA,$BE,$EB,$FA,$6D,$3D
	.dc.b	$06,$C8,$76,$13,$E2,$9F,$6C,$79,$67,$9D,$78,$A7,$BB,$67,$47,$B6
	.dc.b	$38,$BD,$6B,$A3,$E0,$7E,$93,$6B,$AC,$F4,$38,$CE,$A7,$CB,$3A,$78
	.dc.b	$1E,$C3,$27,$84,$C9,$F7,$23,$E3,$BA,$3E,$21,$98,$FA,$F3,$E0,$7F
	.dc.b	$10,$FE,$02,$37,$5F,$A7,$7E,$69,$F9,$C2,$B9,$7C,$C6,$8D,$18,$37
	.dc.b	$1F,$EE,$97,$27,$A4,$E1,$1C,$DC,$1C,$0F,$F7,$83,$F4,$5F,$63,$F9
	.dc.b	$BF,$1D,$E4,$7A,$E7,$A6,$1F,$49,$F8,$53,$A7,$E3,$D7,$39,$FF,$1D
	.dc.b	$CE,$67,$AF,$C9,$FF,$89,$3F,$42,$7D,$6D,$3F,$4F,$12,$7F,$C0,$B3
	.dc.b	$A6,$CE,$5E,$66,$08,$60,$87,$CC,$9F,$8D,$81,$A4,$8F,$86,$D7,$E4
	.dc.b	$B5,$E4,$68,$86,$8F,$D7,$3E,$5C,$3F,$67,$CF,$74,$FD,$0C,$DA,$FD
	.dc.b	$D4,$D7,$F9,$74,$5B,$FF,$51,$DF,$EF,$23,$07,$F0,$CA,$B5,$E0,$3F
	.dc.b	$7C,$E1,$F8,$53,$E2,$74,$62,$FC,$6F,$03,$31,$DD,$B1,$FB,$89,$E9
	.dc.b	$39,$9F,$C3,$C6,$27,$9C,$60,$E2,$7F,$17,$AF,$64,$EE,$30,$78,$4F
	.dc.b	$21,$82,$9F,$D4,$5E,$0F,$C6,$38,$57,$8C,$95,$E2,$7E,$04,$78,$CE
	.dc.b	$77,$F8,$17,$8C,$FB,$5F,$39,$D4,$E1,$3C,$8E,$AF,$95,$C4,$71,$8E
	.dc.b	$B7,$F5,$5E,$A6,$E3,$C8,$6E,$78,$C6,$48,$7B,$CE,$27,$8D,$F0,$B9
	.dc.b	$3C,$B1,$D1,$E3,$5F,$88,$8E,$C6,$63,$93,$87,$F2,$27,$C0,$8F,$88
	.dc.b	$F4,$32,$7B,$0F,$1B,$D5,$EF,$5E,$F7,$81,$D8,$F1,$DE,$6F,$D3,$5D
	.dc.b	$9F,$83,$3E,$B7,$35,$BB,$F3,$4F,$32,$57,$85,$E1,$1E,$D8,$E8,$F0
	.dc.b	$74,$62,$3C,$67,$E3,$CF,$E5,$F7,$FC,$10,$8F,$F2,$46,$BF,$CA,$9A
	.dc.b	$F0,$34,$43,$45,$BF,$6C,$F9,$FE,$ED,$F4,$1E,$DA,$F9,$B5,$F0,$A7
	.dc.b	$31,$F3,$AF,$4A,$32,$FC,$EB,$F6,$CF,$FB,$3A,$73,$5E,$08,$E3,$18
	.dc.b	$C3,$E2,$46,$30,$EB,$7C,$B0,$FF,$59,$DE,$DF,$2D,$F0,$73,$5C,$2F
	.dc.b	$94,$79,$CF,$CB,$D7,$17,$EA,$9C,$88,$7A,$E7,$93,$C1,$AF,$E2,$A3
	.dc.b	$95,$F9,$63,$13,$AC,$3E,$43,$70,$FC,$3D,$F0,$C3,$17,$FE,$4C,$8C
	.dc.b	$3E,$73,$FB,$D3,$E7,$32,$F6,$3D,$E6,$ED,$5F,$E2,$49,$E1,$7D,$AB
	.dc.b	$D8,$70,$3E,$B5,$FD,$58,$91,$CE,$B9,$46,$23,$67,$1B,$EE,$FA,$63
	.dc.b	$B5,$F8,$23,$91,$93,$85,$7D,$77,$DB,$3D,$6B,$F0,$91,$A3,$81,$E0
	.dc.b	$36,$64,$FB,$24,$AF,$03,$B8,$F2,$BF,$00,$E2,$66,$BD,$AF,$3C,$3E
	.dc.b	$D4,$F0,$AE,$73,$FE,$22,$7F,$3A,$DB,$FF,$0E,$46,$23,$99,$FE,$23
	.dc.b	$70,$FA,$8E,$07,$F8,$9C,$F2,$18,$38,$47,$42,$3E,$F8,$FC,$C6,$6F
	.dc.b	$6F,$E4,$63,$E0,$C3,$D1,$87,$F8,$66,$73,$87,$CA,$9C,$BF,$E1,$88
	.dc.b	$FD,$A1,$EA,$23,$E5,$C3,$B3,$63,$2F,$B5,$AF,$C5,$6B,$F9,$16,$BF
	.dc.b	$9D,$FA,$EF,$8E,$47,$B3,$5F,$90,$D7,$E8,$35,$FC,$E3,$5E,$87,$9F
	.dc.b	$B3,$E9,$91,$C1,$0F,$A8,$E5,$A6,$4F,$BC,$E5,$A7,$C3,$F8,$7F,$BE
	.dc.b	$47,$04,$3F,$00,$E5,$A6,$4F,$C5,$39,$69,$F3,$FA,$FF,$16,$47,$04
	.dc.b	$3F,$18,$E5,$A6,$4F,$CB,$39,$69,$F6,$FA,$FF,$2E,$47,$04,$3F,$30
	.dc.b	$E5,$A6,$4F,$D1,$39,$69,$F8,$7E,$BF,$D1,$91,$C1,$0F,$D2,$39,$69
	.dc.b	$93,$F5,$CE,$5A,$7E,$4F,$AF,$F5,$E4,$70,$43,$F6,$0E,$5A,$64,$FD
	.dc.b	$D3,$96,$9F,$9F,$EB,$FD,$D9,$1C,$10,$FD,$E3,$96,$99,$3F,$83,$39
	.dc.b	$69,$FB,$FF,$BB,$FE,$26,$FF,$AB,$1A,$FF,$B6,$9A,$83,$86,$F0,$7F
	.dc.b	$AA,$E3,$04,$70,$48,$F0,$35,$FF,$5B,$43,$C9,$E4,$48,$F3,$35,$DB
	.dc.b	$5F,$43,$5F,$43,$5F,$44,$75,$7E,$07,$5F,$9C,$8E,$08,$79,$1C,$B4
	.dc.b	$C9,$F1,$DC,$B4,$F9,$9E,$4F,$C3,$23,$82,$1F,$01,$CB,$4C,$9F,$4D
	.dc.b	$CB,$4F,$B1,$E4,$FC,$F2,$38,$21,$F3,$1C,$B4,$C9,$F7,$DC,$B4,$FC
	.dc.b	$27,$AB,$F6,$C8,$E0,$87,$D8,$72,$D3,$27,$E2,$DC,$B4,$FC,$87,$93
	.dc.b	$F8,$72,$38,$21,$F8,$47,$2D,$32,$7E,$5D,$CB,$4F,$CE,$7C,$F7,$E4
	.dc.b	$C8,$E0,$87,$E4,$1C,$B4,$C9,$FA,$37,$2D,$3F,$51,$EB,$7E,$7C,$8E
	.dc.b	$08,$7E,$71,$CB,$4C,$9F,$AF,$72,$D3,$F6,$9F,$03,$FA,$B2,$38,$21
	.dc.b	$FA,$87,$2D,$32,$7E,$ED,$CB,$4F,$EF,$DF,$69,$FE,$12,$E9,$FF,$13
	.dc.b	$FA,$0F,$45,$EB,$FE,$FE,$D5,$A9,$CC,$7F,$03,$8F,$F0,$DD,$F3,$E2
	.dc.b	$E0,$D9,$61,$D7,$0E,$97,$CD,$B7,$C0,$B5,$FC,$3E,$F9,$5F,$18,$F0
	.dc.b	$4F,$F1,$38,$C4,$79,$A3,$93,$A3,$93,$D2,$37,$7E,$19,$D0,$AF,$43
	.dc.b	$6B,$86,$1F,$02,$3B,$1E,$C7,$15,$F0,$A7,$D3,$3C,$4F,$15,$F8,$23
	.dc.b	$D1,$5C,$1F,$3C,$F4,$3F,$96,$FB,$3C,$BF,$4B,$7B,$F4,$EB,$C8,$9A
	.dc.b	$66,$BD,$75,$C1,$F6,$CE,$DB,$3B,$7C,$13,$B7,$C5,$3B,$6C,$FF,$3D
	.dc.b	$23,$FD,$25,$AF,$F3,$A6,$BF,$D2,$DA,$FF,$4B,$6B,$FC,$D1,$E3,$97
	.dc.b	$F9,$92,$7C,$0D,$8F,$97,$C2,$F8,$7F,$7D,$FA,$88,$7C,$98,$D3,$8A
	.dc.b	$E4,$DC,$BF,$39,$3A,$8F,$69,$93,$C8,$7E,$11,$D4,$F2,$8D,$91,$FF
	.dc.b	$3D,$C6,$A7,$DD,$3C,$CF,$B6,$73,$AF,$4B,$C6,$36,$FA,$63,$35,$DC
	.dc.b	$93,$E8,$8C,$BC,$23,$33,$FD,$D6,$BC,$4F,$A8,$EC,$58,$F3,$D7,$28
	.dc.b	$C5,$73,$3B,$3E,$9C,$3E,$9C,$66,$3B,$4F,$17,$DA,$76,$17,$E2,$73
	.dc.b	$7E,$66,$D7,$67,$85,$F6,$7C,$67,$63,$E4,$1E,$3C,$3F,$07,$84,$8C
	.dc.b	$1A,$3E,$51,$82,$92,$FC,$55,$DE,$3C,$F3,$CD,$F1,$D5,$AF,$8F,$1C
	.dc.b	$DA,$F3,$BF,$88,$FB,$0F,$41,$93,$D6,$64,$EE,$79,$E3,$C7,$3D,$E7
	.dc.b	$E2,$46,$27,$51,$83,$BD,$FA,$70,$F9,$C7,$CF,$7D,$86,$70,$E9,$5E
	.dc.b	$28,$C3,$D0,$C6,$1A,$7C,$A7,$C2,$34,$63,$0F,$23,$ED,$31,$87,$92
	.dc.b	$79,$61,$E3,$7F,$0C,$F9,$B0,$FF,$67,$57,$63,$FA,$71,$4F,$C2,$18
	.dc.b	$8F,$DB,$36,$38,$C6,$23,$64,$3E,$89,$B3,$FD,$B8,$ED,$FA,$B7,$F6
	.dc.b	$89,$7F,$C1,$33,$F4,$DF,$A9,$B5,$F0,$AF,$FB,$33,$67,$4D,$35,$3C
	.dc.b	$1F,$8D,$3C,$1B,$3C,$1A,$70,$39,$1C,$27,$95,$F9,$A3,$F7,$D1,$8B
	.dc.b	$FC,$24,$60,$E3,$3B,$7F,$67,$3B,$7E,$0C,$FB,$DD,$BC,$9D,$98,$36
	.dc.b	$75,$3C,$E7,$94,$D1,$F8,$23,$53,$E1,$BF,$C0,$47,$F2,$BA,$FE,$76
	.dc.b	$78,$DE,$E7,$F3,$97,$11,$F1,$E7,$F9,$6C,$60,$E3,$3E,$09,$D5,$7A
	.dc.b	$E3,$CB,$5C,$1F,$01,$BA,$E2,$EA,$78,$4E,$E3,$07,$AA,$32,$7C,$5A
	.dc.b	$FD,$7D,$72,$AF,$75,$DA,$DD,$FC,$83,$E2,$1E,$E9,$EC,$70,$7C,$A7
	.dc.b	$C1,$3C,$C6,$1D,$39,$23,$83,$9C,$7C,$18,$FD,$45,$7E,$98,$8F,$E8
	.dc.b	$9A,$FE,$C0,$FA,$27,$03,$44,$35,$7F,$99,$74,$62,$3E,$03,$F1,$CF
	.dc.b	$F6,$37,$E1,$4E,$DC,$C7,$C1,$BF,$8A,$A7,$4F,$BE,$33,$7F,$72,$39
	.dc.b	$D7,$F8,$0B,$ED,$7F,$C5,$B6,$F0,$29,$9A,$FE,$F1,$38,$7F,$C7,$B3
	.dc.b	$B7,$24,$3F,$C6,$0E,$5B,$3C,$DF,$F0,$64,$FF,$10,$23,$FC,$31,$AF
	.dc.b	$C6,$6B,$FC,$65,$AF,$F8,$41,$AF,$F7,$77,$8E,$5F,$E0,$A7,$C0,$B1
	.dc.b	$FF,$39,$7D,$EB,$ED,$73,$1C,$EF,$D5,$5B,$7F,$DE,$D5,$B7,$33,$B7
	.dc.b	$D6,$7F,$98,$23,$4E,$09,$3E,$37,$55,$E0,$7D,$07,$43,$BB,$97,$FC
	.dc.b	$03,$76,$BA,$D7,$39,$EB,$7E,$38,$F4,$42,$77,$39,$24,$78,$0F,$F3
	.dc.b	$34,$69,$C4,$F5,$24,$FC,$8B,$F2,$66,$C3,$C9,$18,$6D,$FA,$CF,$49
	.dc.b	$70,$EC,$E0,$F4,$9E,$0B,$EC,$F1,$BE,$CF,$1B,$EC,$D3,$E5,$9B,$BF
	.dc.b	$F4,$3D,$F6,$3B,$D7,$29,$EE,$7C,$5A,$D9,$D1,$F8,$6E,$1E,$87,$63
	.dc.b	$26,$0A,$4C,$3E,$3B,$EB,$3E,$D9,$C5,$CB,$F0,$5F,$AF,$5E,$79,$F1
	.dc.b	$46,$1F,$14,$FA,$2B,$D5,$39,$8E,$98,$7D,$C7,$A3,$F2,$9F,$23,$99
	.dc.b	$E2,$E6,$79,$4F,$5C,$3B,$C6,$0F,$3D,$70,$9F,$59,$D0,$F7,$1E,$03
	.dc.b	$31,$C6,$BC,$15,$D1,$F0,$18,$9F,$49,$C1,$C3,$D4,$FA,$C7,$CF,$9E
	.dc.b	$8F,$88,$FF,$0F,$E1,$FB,$06,$E1,$FE,$24,$8C,$47,$C3,$32,$7F,$37
	.dc.b	$3A,$9F,$CE,$5D,$BD,$4E,$04,$30,$79,$C8,$72,$9F,$D8,$FD,$EF,$FB
	.dc.b	$82,$7E,$E4,$FB,$84,$66,$BE,$24,$72,$8C,$3B,$9F,$D3,$C6,$23,$E2
	.dc.b	$46,$2F,$8C,$78,$5F,$9A,$8E,$EE,$A3,$C7,$1F,$EE,$59,$E9,$7F,$32
	.dc.b	$39,$1C,$63,$35,$D2,$32,$74,$3C,$D7,$E2,$9D,$C7,$C4,$BF,$33,$A3
	.dc.b	$E1,$46,$9E,$71,$A7,$DB,$3C,$EB,$D0,$D8,$E4,$7E,$16,$B9,$57,$B2
	.dc.b	$FE,$8D,$F0,$AE,$EF,$C3,$7E,$03,$96,$9F,$1E,$38,$BD,$9A,$E8,$F2
	.dc.b	$9B,$8E,$95,$FC,$C2,$76,$7C,$BA,$F6,$38,$9F,$95,$18,$78,$9E,$38
	.dc.b	$D5,$7D,$18,$C1,$A7,$C0,$64,$F3,$39,$7C,$D1,$E1,$BF,$80,$F4,$8F
	.dc.b	$B0,$7A,$8D,$CF,$18,$E1,$1D,$8D,$BC,$6B,$E7,$4E,$C9,$3F,$7E,$BB
	.dc.b	$C6,$23,$8B,$F0,$23,$31,$E9,$3C,$27,$13,$C1,$7F,$2E,$7D,$91,$FE
	.dc.b	$9C,$8F,$EB,$AD,$8F,$F5,$0C,$62,$36,$43,$B1,$B3,$F8,$03,$B7,$26
	.dc.b	$CF,$8A,$6C,$F2,$9E,$A3,$FC,$67,$E5,$FA,$9C,$12,$3D,$0E,$8F,$5C
	.dc.b	$EE,$33,$1C,$34,$56,$7B,$7D,$64,$9F,$3B,$5F,$F2,$75,$23,$11,$F1
	.dc.b	$5A,$FE,$51,$AF,$EE,$DF,$8F,$DB,$23,$E1,$B5,$EA,$D7,$E8,$35,$FC
	.dc.b	$FB,$5F,$E0,$CD,$7F,$8E,$F6,$FD,$42,$3F,$41,$AF,$A9,$AF,$DD,$6B
	.dc.b	$FA,$C6,$BF,$C4,$DF,$BF,$B4,$8F,$B6,$D7,$E2,$B5,$FC,$43,$5F,$DB
	.dc.b	$B5,$FE,$3C,$D7,$F9,$AF,$6F,$E3,$08,$FE,$21,$AF,$D2,$6B,$F9,$56
	.dc.b	$BF,$C1,$1A,$FF,$2B,$7F,$2F,$DB,$23,$F2,$6D,$7E,$EB,$5F,$D0,$35
	.dc.b	$FE,$1E,$D7,$F9,$B3,$5F,$E9,$BD,$BF,$A4,$23,$FA,$06,$BF,$8A,$6B
	.dc.b	$FA,$D6,$BF,$C7,$1A,$FF,$43,$7F,$5F,$DB,$23,$F5,$6D,$7F,$2A,$D7
	.dc.b	$F7,$0D,$7F,$93,$B5,$FE,$9C,$D7,$FB,$2F,$C4,$FE,$F0,$8F,$EE,$1A
	.dc.b	$FE,$89,$AF,$F0,$56,$BF,$CD,$1A,$FF,$5B,$7F,$C0,$0B,$F4,$57,$D0
	.dc.b	$7E,$ED,$B1,$FC,$85,$AF,$F6,$7F,$88,$D4,$C8,$D3,$FB,$D8,$C4,$39
	.dc.b	$7F,$CE,$75,$C9,$B5,$B9,$FE,$3F,$5F,$C6,$88,$FF,$16,$6B,$FC,$01
	.dc.b	$AF,$F2,$36,$BF,$D4,$9A,$FF,$79,$7F,$92,$FA,$7D,$E3,$D4,$7F,$2E
	.dc.b	$7F,$8A,$19,$AF,$F0,$CC,$E2,$74,$79,$88,$7F,$82,$67,$4F,$F1,$47
	.dc.b	$BB,$FC,$B9,$AE,$4C,$10,$C1,$0F,$E7,$CF,$07,$FA,$07,$A3,$FC,$F8
	.dc.b	$8F,$F3,$A6,$BF,$C9,$1A,$F0,$34,$43,$45,$3F,$A6,$7C,$27,$F4,$A2
	.dc.b	$3F,$D1,$9A,$FF,$30,$6B,$C0,$D1,$0D,$14,$FE,$B1,$F0,$3F,$D5,$88
	.dc.b	$FF,$52,$6B,$FC,$F1,$AF,$03,$44,$34,$53,$FB,$27,$BD,$FD,$88,$8F
	.dc.b	$F5,$E6,$BF,$D2,$1A,$F0,$34,$43,$45,$3F,$B8,$7C,$0F,$F6,$E2,$3F
	.dc.b	$DA,$9A,$FF,$54,$6B,$C0,$D1,$0D,$14,$FE,$F9,$E3,$FF,$7A,$23,$FD
	.dc.b	$D9,$AF,$F6,$06,$BC,$0D,$10,$D1,$4F,$F2,$00,$FB,$E7,$CC,$FF,$02
	.dc.b	$6C,$7F,$A0,$DA,$F0,$38,$46,$C8,$ED,$AE,$E3,$8B,$E0,$6B,$FE,$51
	.dc.b	$6B,$FE,$75,$6B,$E5,$F9,$4F,$21,$1F,$23,$5E,$CD,$7B,$35,$EC,$D7
	.dc.b	$B3,$EF,$F9,$E9,$1E,$F6,$BE,$96,$BE,$96,$BE,$96,$BE,$96,$BF,$37
	.dc.b	$E3,$BE,$61,$1F,$98,$D7,$E3,$35,$F8,$CD,$7E,$33,$5F,$8C,$FD,$7F
	.dc.b	$1A,$47,$D7,$6B,$F4,$5A,$FD,$16,$BF,$45,$AF,$D1,$6B,$F8,$5F,$73
	.dc.b	$F0,$84,$7F,$08,$D7,$EF,$35,$FB,$CD,$7E,$F3,$5F,$BC,$FE,$3F,$CD
	.dc.b	$91,$F8,$F6,$BF,$89,$6B,$F8,$96,$BF,$89,$6B,$F8,$96,$BF,$9D,$FB
	.dc.b	$AF,$CE,$11,$FC,$E3,$5F,$CB,$35,$FC,$B3,$5F,$CB,$35,$FC,$B3,$FA
	.dc.b	$7F,$72,$47,$E9,$DA,$FE,$85,$AF,$E8,$5A,$FE,$85,$AF,$E8,$5A,$FE
	.dc.b	$D7,$E1,$7F,$68,$47,$F6,$8D,$7F,$5C,$D7,$F5,$CD,$7F,$5C,$D7,$F5
	.dc.b	$CF,$EF,$FE,$E2,$47,$EF,$DA,$FE,$E5,$AF,$EE,$5A,$FE,$E5,$AF,$EE
	.dc.b	$5A,$FF,$0D,$F7,$FF,$86,$11,$FE,$18,$D7,$F8,$33,$5F,$E0,$CD,$7F
	.dc.b	$83,$35,$FE,$0C,$FF,$17,$F8,$59,$1F,$C5,$DA,$FF,$11,$6B,$FC,$45
	.dc.b	$AF,$F1,$16,$BF,$C4,$5A,$FF,$25,$FB,$7F,$E4,$84,$7F,$92,$35,$FE
	.dc.b	$3C,$D7,$F8,$F3,$5F,$E3,$CD,$7F,$8F,$3F,$CB,$FE,$32,$47,$F2,$F6
	.dc.b	$BF,$CA,$5A,$FF,$29,$6B,$FC,$A5,$AF,$F2,$96,$BF,$CF,$7E,$47,$F9
	.dc.b	$E1,$1F,$E7,$8D,$7F,$9B,$35,$FE,$6C,$D7,$F9,$B3,$5F,$E6,$CF,$F4
	.dc.b	$7F,$46,$47,$F4,$76,$BF,$D0,$5A,$FF,$41,$6B,$FD,$05,$AF,$F4,$16
	.dc.b	$BF,$D5,$7E,$73,$FA,$A1,$1F,$EA,$8D,$7F,$A7,$35,$FE,$9C,$D7,$FA
	.dc.b	$73,$5F,$E9,$CF,$F5,$FF,$92,$91,$FD,$7D,$AF,$F5,$96,$BF,$D6,$5A
	.dc.b	$FF,$59,$6B,$FD,$65,$AF,$F6,$DF,$47,$FB,$61,$1F,$ED,$8D,$7F,$B3
	.dc.b	$35,$FE,$CC,$D7,$FB,$33,$5F,$EC,$CF,$F7,$7F,$85,$91,$FD,$DD,$AF
	.dc.b	$F7,$16,$BF,$DC,$5A,$FF,$71,$6B,$FD,$C5,$AB,$EE,$BF,$D0,$AD,$3F
	.dc.b	$D0,$EE,$08,$E0,$87,$F8,$1C,$E0,$57,$6D,$36,$7F,$83,$57,$D0,$76
	.dc.b	$24,$71,$6B,$B6,$BF,$E1,$C8,$E2,$EC,$86,$CA,$6F,$C4,$F5,$91,$F3
	.dc.b	$35,$F2,$35,$E0,$68,$86,$8A,$78,$FE,$27,$E3,$91,$F8,$4D,$7E,$03
	.dc.b	$5E,$06,$88,$68,$A7,$BF,$E7,$3E,$99,$1F,$9C,$D7,$E6,$35,$E0,$68
	.dc.b	$86,$8A,$7C,$BF,$91,$FB,$E4,$7E,$D3,$5F,$B0,$D7,$81,$A2,$1A,$29
	.dc.b	$F5,$EF,$F1,$64,$7F,$0C,$D7,$F0,$8D,$78,$1A,$21,$A2,$9F,$83,$FA
	.dc.b	$5F,$CB,$91,$FC,$93,$5F,$C8,$35,$E0,$68,$86,$8A,$7E,$3F,$E7,$3F
	.dc.b	$46,$47,$F3,$CD,$7F,$38,$D7,$81,$A2,$1A,$29,$F9,$BF,$97,$FD,$79
	.dc.b	$1F,$D5,$35,$FD,$43,$5E,$06,$88,$68,$A7,$E9,$FD,$9F,$DD,$91,$FD
	.dc.b	$B3,$5F,$DA,$35,$E0,$68,$86,$8A,$7E,$CF,$E2,$FF,$83,$91,$FE,$04
	.dc.b	$D7,$F8,$03,$5E,$06,$88,$68,$A7,$EF,$FD,$8F,$E2,$64,$7F,$87,$35
	.dc.b	$FE,$18,$D7,$81,$A2,$1A,$29,$FC,$2F,$E2,$BF,$8F,$91,$FE,$34,$D7
	.dc.b	$F8,$C3,$5E,$06,$88,$68,$A7,$F1,$7F,$77,$F9,$59,$1F,$E4,$CD,$7F
	.dc.b	$92,$35,$E0,$68,$86,$8A,$7F,$23,$F8,$AF,$E6,$E4,$7F,$99,$35,$FE
	.dc.b	$60,$D7,$81,$A2,$1A,$29,$FC,$BF,$DF,$FE,$86,$47,$F9,$F3,$5F,$E7
	.dc.b	$8D,$78,$1A,$21,$A2,$9F,$CE,$FE,$2B,$FA,$79,$1F,$E9,$4D,$7F,$A4
	.dc.b	$35,$E0,$68,$86,$8A,$7F,$47,$F8,$AF,$EB,$64,$7F,$AB,$35,$FE,$A8
	.dc.b	$D7,$81,$A2,$1A,$29,$FD,$4F,$E2,$BF,$B3,$91,$FE,$C4,$D7,$FB,$03
	.dc.b	$5E,$06,$88,$68,$A7,$F5,$FF,$8A,$FE,$E6,$47,$FB,$73,$5F,$ED,$8D
	.dc.b	$78,$1A,$21,$A2,$9F,$DA,$FE,$2B,$FB,$F9,$1F,$EF,$4D,$7F,$BC,$35
	.dc.b	$E0,$68,$86,$8A,$08,$3E,$F9,$F0,$9F,$E0,$98,$C3,$63,$31,$FE,$0D
	.dc.b	$6B,$C8,$E8,$E1,$C3,$5E,$ED,$77,$1D,$48,$F5,$6B,$D4,$F3,$7C,$5F
	.dc.b	$8C,$8F,$8D,$AE,$E5,$18,$8F,$1B,$5F,$1B,$5F,$4B,$EF,$F8,$39,$1E
	.dc.b	$F6,$BD,$5A,$FB,$9A,$FB,$9A,$FC,$46,$BF,$3F,$E0,$FE,$61,$1F,$98
	.dc.b	$D7,$D4,$D7,$E5,$B5,$F9,$6D,$7E,$8B,$F5,$FE,$0E,$47,$D7,$6B,$F1
	.dc.b	$5A,$FD,$66,$BF,$59,$AF,$DC,$6B,$F8,$7F,$83,$FC,$21,$1F,$C2,$35
	.dc.b	$FA,$4D,$7F,$06,$D7,$F0,$6D,$7F,$12,$FE,$3F,$E0,$E4,$7E,$3D,$AF
	.dc.b	$DD,$6B,$F8,$E6,$BF,$8E,$6B,$F9,$46,$BF,$9F,$F8,$3F,$CE,$11,$FC
	.dc.b	$E3,$5F,$C5,$35,$FC,$DB,$5F,$CD,$B5,$FD,$0B,$FA,$7F,$83,$91,$FA
	.dc.b	$76,$BF,$95,$6B,$FA,$66,$BF,$A6,$6B,$FA,$C6,$BF,$B7,$F8,$3F,$DA
	.dc.b	$11,$FD,$A3,$5F,$D1,$35,$FD,$9B,$5F,$D9,$B5,$FD,$CB,$FB,$FF,$83
	.dc.b	$91,$FB,$F6,$BF,$AD,$6B,$FB,$E6,$BF,$BE,$6B,$FC,$11,$AF,$F0,$FF
	.dc.b	$8A,$FE,$18,$47,$F8,$63,$5F,$DD,$35,$FE,$16,$D7,$F8,$5B,$5F,$E2
	.dc.b	$2F,$F1,$7F,$82,$91,$FC,$5D,$AF,$F0,$56,$BF,$C5,$9A,$FF,$16,$6B
	.dc.b	$FC,$71,$AF,$F2,$7F,$88,$FE,$48,$47,$F9,$23,$5F,$E2,$4D,$7F,$91
	.dc.b	$B5,$FE,$46,$D7,$F9,$4B,$FC,$BF,$E0,$E4,$7F,$2F,$6B,$FC,$75,$AF
	.dc.b	$F2,$E6,$BF,$CB,$9A,$FF,$34,$6B,$FC,$FF,$E0,$FF,$9E,$11,$FE,$78
	.dc.b	$D7,$F9,$53,$5F,$E7,$6D,$7F,$9D,$B5,$FE,$82,$FF,$47,$F8,$39,$1F
	.dc.b	$D1,$DA,$FF,$35,$6B,$FD,$19,$AF,$F4,$66,$BF,$D3,$1A,$FF,$57,$F8
	.dc.b	$3F,$EA,$84,$7F,$AA,$35,$FE,$84,$D7,$FA,$9B,$5F,$EA,$6D,$7F,$AC
	.dc.b	$BF,$D7,$FE,$0E,$47,$F5,$F6,$BF,$D3,$5A,$FF,$5E,$6B,$FD,$79,$AF
	.dc.b	$F6,$46,$BF,$DB,$FE,$0F,$FB,$61,$1F,$ED,$8D,$7F,$AD,$35,$FE,$D6
	.dc.b	$D7,$FB,$5B,$5F,$EE,$2F,$F7,$7F,$83,$91,$FD,$DD,$AF,$F6,$56,$BF
	.dc.b	$DD,$9A,$FF,$76,$6B,$FD,$F1,$AB,$EF,$9F,$09,$FE,$07,$6C,$7F,$77
	.dc.b	$35,$19,$23,$98,$DB,$FD,$F5,$AF,$F8,$2D,$AF,$F8,$59,$AF,$F8,$65
	.dc.b	$AF,$63,$A4,$78,$FE,$0E,$47,$76,$BD,$DA,$F7,$6B,$DD,$AF,$76,$BF
	.dc.b	$07,$E0,$FE,$01,$1F,$63,$5F,$63,$5F,$63,$5F,$63,$5F,$63,$F2,$FE
	.dc.b	$0E,$47,$C7,$6B,$F1,$DA,$FC,$76,$BF,$1D,$AF,$C7,$6B,$F6,$7E,$0F
	.dc.b	$EC,$11,$FA,$8D,$7E,$A3,$5F,$A8,$D7,$EA,$35,$FA,$8F,$E0,$FE,$0E
	.dc.b	$47,$DF,$6B,$F7,$DA,$FD,$F6,$BF,$7D,$AF,$DF,$6B,$F9,$1F,$83,$FC
	.dc.b	$81,$1F,$C6,$35,$FC,$63,$5F,$C6,$35,$FC,$63,$5F,$C6,$3F,$9B,$F8
	.dc.b	$39,$1F,$97,$6B,$F9,$76,$BF,$97,$6B,$F9,$76,$BF,$97,$6B,$FA,$9F
	.dc.b	$83,$7C,$2F,$E8,$DB,$1F,$A3,$6B,$FA,$36,$BF,$A3,$6B,$FA,$36,$BD
	.dc.b	$63,$A1,$1E,$8D,$7A,$3F,$B3,$F8,$47,$DE,$D2,$4F,$ED,$63,$2D,$8C
	.dc.b	$BF,$B7,$8C,$B6,$32,$FE,$E6,$32,$D8,$CF,$DF,$08,$00

;--------------------------------------------------------------------------------
;	FINTRZ.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fintrz_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#59,-(sp)
	peamsg	'FINTRZ.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FINTRZ.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fintrz_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fintrz.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FINTRZ.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fintrz.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FINTRZ.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fintrz_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$BB,$F6,$7F,$79,$A5,$3F,$9A,$B5,$A5,$70,$9A,$47
	.dc.b	$44,$3A,$39,$7C,$1E,$AF,$F4,$57,$F1,$ED,$7F,$98,$B5,$FE,$8E,$D7
	.dc.b	$FC,$53,$F1,$BE,$58,$C1,$1C,$12,$3D,$91,$97,$DD,$EB,$73,$24,$FB
	.dc.b	$DB,$1A,$8C,$11,$C1,$0F,$95,$1C,$1F,$99,$EB,$79,$C9,$3F,$2D,$B1
	.dc.b	$F3,$DA,$F0,$8D,$11,$D3,$5F,$B1,$EB,$7C,$32,$47,$D7,$6C,$7D,$B6
	.dc.b	$BC,$23,$44,$74,$D7,$F0,$9E,$B7,$CF,$24,$7E,$0D,$B1,$F8,$76,$BC
	.dc.b	$23,$44,$74,$D7,$F2,$1E,$B7,$DB,$24,$7E,$3D,$B1,$F9,$36,$BC,$23
	.dc.b	$44,$74,$D7,$F3,$9E,$B7,$E1,$C9,$1F,$9B,$6C,$7E,$7D,$AF,$08,$D1
	.dc.b	$1D,$35,$FD,$47,$AD,$F9,$32,$47,$E9,$DB,$1F,$AB,$6B,$C2,$34,$47
	.dc.b	$4D,$7F,$69,$EB,$7E,$7C,$91,$FB,$36,$C7,$ED,$DA,$F0,$8D,$11,$D3
	.dc.b	$5F,$E0,$1E,$B7,$EA,$C9,$1F,$BF,$6C,$7F,$03,$6B,$C2,$34,$47,$4D
	.dc.b	$7F,$86,$7A,$DF,$B7,$24,$7F,$0B,$6C,$7F,$0F,$6B,$C2,$34,$47,$4D
	.dc.b	$7F,$8C,$7A,$DF,$C0,$C9,$1F,$C5,$DB,$1F,$C6,$DA,$F0,$8D,$11,$D3
	.dc.b	$5F,$E4,$9E,$B7,$F0,$F2,$47,$F2,$36,$C7,$F2,$76,$BC,$23,$44,$74
	.dc.b	$D7,$F9,$87,$AD,$FC,$6C,$91,$FC,$BD,$B1,$FC,$CD,$AF,$08,$D1,$1D
	.dc.b	$35,$FE,$79,$EB,$7F,$27,$24,$7F,$3B,$6C,$7F,$3F,$6B,$C2,$34,$47
	.dc.b	$4D,$7F,$A4,$7A,$DF,$CC,$C9,$1F,$D1,$DB,$1F,$D2,$DA,$F0,$8D,$11
	.dc.b	$D3,$5F,$EA,$9E,$B7,$F3,$F2,$47,$F5,$36,$C7,$F5,$76,$BC,$23,$44
	.dc.b	$74,$D7,$FB,$07,$AD,$FD,$2C,$91,$FD,$7D,$B1,$FD,$8D,$AF,$08,$D1
	.dc.b	$1D,$35,$FE,$D9,$EB,$7F,$57,$24,$7F,$6B,$6C,$7F,$6F,$6B,$C2,$34
	.dc.b	$47,$4D,$7F,$BC,$7A,$DF,$D8,$C9,$1F,$DD,$DB,$1F,$DE,$DA,$F0,$8D
	.dc.b	$11,$D3,$5F,$F0,$27,$AD,$FD,$BC,$91,$FE,$03,$6C,$7F,$81,$DA,$F0
	.dc.b	$8D,$11,$D3,$5F,$F2,$6F,$DD,$23,$FB,$F3,$5F,$F0,$93,$63,$FC,$2C
	.dc.b	$D7,$8C,$70,$21,$C0,$AF,$06,$BC,$1E,$BF,$19,$E0,$23,$82,$47,$82
	.dc.b	$32,$D8,$CB,$EA,$8C,$B6,$3D,$3F,$0B,$EC,$23,$82,$47,$B2,$32,$D8
	.dc.b	$CB,$F1,$A3,$2D,$8F,$8B,$EC,$7C,$82,$38,$24,$7C,$88,$CB,$63,$2F
	.dc.b	$D2,$8C,$B6,$3E,$8F,$B1,$F5,$08,$E0,$91,$F5,$23,$2D,$8C,$BF,$7A
	.dc.b	$32,$D8,$FB,$BE,$C7,$E0,$08,$E0,$91,$F8,$08,$CB,$63,$2F,$E2,$A3
	.dc.b	$2D,$8F,$C4,$FB,$1F,$8C,$23,$82,$47,$E3,$23,$2D,$8C,$BF,$96,$8C
	.dc.b	$B6,$3F,$2B,$EC,$7E,$60,$8E,$09,$1F,$98,$8C,$B6,$32,$FE,$8A,$32
	.dc.b	$D8,$FD,$0F,$B1,$FA,$42,$38,$24,$7E,$92,$32,$D8,$CB,$FA,$E8,$CB
	.dc.b	$63,$F5,$BE,$C7,$EC,$08,$E0,$91,$FB,$08,$CB,$63,$2F,$EE,$A3,$2D
	.dc.b	$8F,$DC,$FB,$1F,$BC,$23,$82,$47,$EF,$23,$2D,$8C,$BF,$C1,$A3,$2D
	.dc.b	$8F,$E0,$BE,$C7,$F0,$82,$38,$24,$7F,$08,$8C,$B6,$32,$FF,$12,$8C
	.dc.b	$B6,$3F,$88,$FB,$1F,$C5,$08,$E0,$91,$FC,$52,$32,$D8,$CB,$FC,$7A
	.dc.b	$32,$D8,$FE,$3B,$EC,$7F,$20,$23,$82,$47,$F2,$08,$CB,$63,$2F,$F2
	.dc.b	$A8,$CB,$63,$F9,$4F,$B1,$FC,$B0,$8E,$09,$1F,$CB,$23,$2D,$8C,$BF
	.dc.b	$CD,$A3,$2D,$8F,$E6,$BE,$C7,$F3,$82,$38,$24,$7F,$38,$8C,$B6,$32
	.dc.b	$FF,$42,$8C,$B6,$3F,$A0,$FB,$1F,$D1,$08,$E0,$91,$FD,$12,$32,$D8
	.dc.b	$CB,$FD,$3A,$32,$D8,$FE,$9B,$EC,$7F,$50,$23,$82,$47,$F5,$08,$CB
	.dc.b	$63,$2F,$F5,$A8,$CB,$63,$FA,$CF,$B1,$FD,$70,$8E,$09,$1F,$D7,$23
	.dc.b	$2D,$8C,$BF,$D9,$A3,$2D,$8F,$EC,$BE,$C7,$F6,$82,$38,$24,$7F,$68
	.dc.b	$8C,$B6,$32,$FF,$72,$8C,$B6,$3F,$B8,$FB,$1F,$DD,$08,$E0,$91,$FD
	.dc.b	$D2,$32,$D8,$CB,$FD,$FA,$32,$D8,$30,$9F,$74,$F8,$AF,$F8,$00,$D9
	.dc.b	$D4,$EA,$72,$48,$CC,$74,$9E,$0E,$1A,$F2,$6B,$CE,$38,$A8,$CB,$CE
	.dc.b	$7B,$3C,$FE,$2B,$CA,$4A,$E3,$48,$C4,$EE,$73,$18,$8F,$13,$5F,$82
	.dc.b	$D7,$D5,$ED,$7C,$22,$4F,$66,$C7,$B5,$AF,$B5,$AF,$CC,$6B,$D2,$3C
	.dc.b	$FE,$D7,$CE,$23,$82,$47,$C9,$8C,$B6,$32,$FC,$B8,$CB,$63,$E1,$FB
	.dc.b	$5F,$68,$8E,$09,$1F,$56,$32,$D8,$CB,$F5,$E3,$2D,$8F,$9F,$ED,$7E
	.dc.b	$18,$8E,$09,$1F,$81,$8C,$B6,$32,$FE,$0E,$32,$D8,$FB,$7E,$D7,$E4
	.dc.b	$88,$E0,$91,$F8,$D8,$CB,$63,$2F,$E3,$E3,$2D,$8F,$C3,$FB,$5F,$9E
	.dc.b	$23,$82,$47,$E6,$63,$2D,$8C,$BF,$9B,$8C,$B6,$3F,$27,$ED,$7E,$A8
	.dc.b	$8E,$09,$1F,$A5,$8C,$B6,$32,$FE,$9E,$32,$D8,$FC,$FF,$B5,$FB,$62
	.dc.b	$38,$24,$7E,$C6,$32,$D8,$CB,$FB,$38,$CB,$63,$F5,$7F,$0F,$FC,$08
	.dc.b	$8E,$09,$1F,$BD,$8C,$B6,$32,$FE,$FE,$32,$D8,$FD,$BF,$C3,$FF,$0E
	.dc.b	$23,$82,$47,$F0,$98,$CB,$63,$2F,$F0,$B8,$CB,$63,$F8,$1F,$C3,$FF
	.dc.b	$1A,$23,$82,$47,$F1,$58,$CB,$63,$2F,$F1,$78,$CB,$63,$F8,$7F,$C3
	.dc.b	$FF,$26,$23,$82,$47,$F2,$18,$CB,$63,$2F,$F2,$38,$CB,$63,$F8,$DF
	.dc.b	$C3,$FF,$32,$23,$82,$47,$F2,$D8,$CB,$63,$2F,$F2,$F8,$CB,$63,$F9
	.dc.b	$3F,$C3,$FF,$3E,$23,$82,$47,$F3,$98,$CB,$63,$2F,$F3,$B8,$CB,$63
	.dc.b	$F9,$9F,$C3,$FF,$4A,$23,$82,$47,$F4,$58,$CB,$63,$2F,$F4,$78,$CB
	.dc.b	$63,$F9,$FF,$C3,$FF,$56,$23,$82,$47,$F5,$18,$CB,$63,$2F,$F5,$38
	.dc.b	$CB,$63,$FA,$5F,$C3,$FF,$62,$23,$82,$47,$F5,$D8,$CB,$63,$2F,$F5
	.dc.b	$F8,$CB,$63,$FA,$BF,$C3,$FF,$6E,$23,$82,$47,$F6,$98,$CB,$63,$2F
	.dc.b	$F6,$B8,$CB,$63,$FB,$1F,$C3,$FF,$7A,$23,$82,$47,$F7,$58,$CB,$63
	.dc.b	$2F,$F7,$78,$CB,$63,$FB,$BF,$DD,$7F,$82,$9B,$06,$91,$FF,$02,$C6
	.dc.b	$DA,$6C,$B1,$B7,$FC,$1F,$1B,$69,$B2,$C6,$DF,$F0,$D4,$3E,$13,$51
	.dc.b	$86,$BF,$E2,$08,$D4,$FA,$A7,$32,$9C,$BE,$88,$CC,$62,$3C,$B1,$82
	.dc.b	$4F,$32,$47,$C0,$F8,$4E,$24,$8F,$92,$D8,$D4,$60,$8E,$08,$7C,$48
	.dc.b	$E0,$FC,$6F,$84,$F2,$92,$7E,$AB,$63,$E8,$B5,$E1,$1A,$23,$A6,$BF
	.dc.b	$4B,$E1,$3E,$09,$23,$F0,$2D,$8F,$BA,$D7,$84,$68,$8E,$9A,$FD,$EF
	.dc.b	$84,$F9,$A4,$8F,$C6,$B6,$3F,$12,$D7,$84,$68,$8E,$9A,$FE,$2B,$E1
	.dc.b	$3E,$C9,$23,$F3,$2D,$8F,$CA,$B5,$E1,$1A,$23,$A6,$BF,$96,$F8,$4F
	.dc.b	$C2,$92,$3F,$4A,$D8,$FD,$0B,$5E,$11,$A2,$3A,$6B,$FA,$2F,$84,$FC
	.dc.b	$89,$23,$F6,$2D,$8F,$D6,$B5,$E1,$1A,$23,$A6,$BF,$AE,$F8,$4F,$CE
	.dc.b	$92,$3F,$7A,$D8,$FD,$CB,$5E,$11,$A2,$3A,$6B,$FB,$AF,$84,$FD,$49
	.dc.b	$23,$F8,$4B,$63,$F8,$2B,$5E,$11,$A2,$3A,$6B,$FC,$1B,$E1,$3F,$6A
	.dc.b	$48,$FE,$2A,$D8,$FE,$22,$D7,$84,$68,$8E,$9A,$FF,$12,$F8,$4F,$E0
	.dc.b	$24,$8F,$E4,$2D,$8F,$E3,$AD,$78,$46,$88,$E9,$AF,$F1,$EF,$84,$FE
	.dc.b	$1A,$48,$FE,$5A,$D8,$FE,$52,$D7,$84,$68,$8E,$9A,$FF,$2A,$F8,$4F
	.dc.b	$E3,$24,$8F,$E7,$2D,$8F,$E6,$AD,$78,$46,$88,$E9,$AF,$F3,$6F,$84
	.dc.b	$FE,$4A,$48,$FE,$8A,$D8,$FE,$82,$D7,$84,$68,$8E,$9A,$FF,$42,$F8
	.dc.b	$3F,$E8,$DE,$4F,$F4,$1E,$54,$62,$BF,$A0,$B6,$BF,$9F,$35,$FE,$78
	.dc.b	$D9,$E3,$3F,$ED,$73,$89,$67,$8B,$5E,$2D,$8E,$31,$FE,$D7,$8E,$2D
	.dc.b	$38,$94,$E2,$7F,$62,$9F,$E9,$64,$8F,$ED,$6D,$8D,$46,$08,$E0,$87
	.dc.b	$C5,$7F,$DD,$3F,$9E,$BD,$BF,$80,$D8,$F5,$36,$7C,$8D,$C3,$FB,$63
	.dc.b	$65,$32,$30,$DA,$EA,$49,$EA,$D8,$EA,$FC,$FF,$83,$FE,$D2,$49,$FF
	.dc.b	$04,$B6,$7F,$C0,$AD,$7F,$C0,$2D,$7F,$BE,$B5,$FE,$F4,$FF,$74,$F8
	.dc.b	$39,$3F,$E1,$46,$CF,$F8,$41,$AF,$F8,$31,$AF,$F8,$21,$AF,$F8,$15
	.dc.b	$AF,$F8,$0F,$E0,$FF,$C0,$24,$9F,$F1,$0B,$63,$FC,$3A,$D7,$FC,$32
	.dc.b	$D7,$FC,$2A,$D7,$FC,$24,$FF,$84,$3C,$FF,$F1,$44,$FD,$09,$82,$32
	.dc.b	$E0,$C4,$71,$21,$97,$2D,$8E,$31,$D1,$AF,$76,$BF,$80,$6B,$C8,$F0
	.dc.b	$38,$1F,$05,$E9,$24,$F7,$6C,$F0,$6B,$E1,$6B,$E1,$6B,$E1,$7E,$27
	.dc.b	$C1,$49,$F5,$36,$7C,$6D,$7D,$8D,$7D,$8D,$7D,$8D,$7E,$67,$C1,$7D
	.dc.b	$12,$4F,$C7,$6C,$7C,$06,$BF,$25,$AF,$C9,$6B,$F2,$5F,$B9,$F0,$52
	.dc.b	$7E,$93,$67,$E5,$B5,$FA,$8D,$7E,$A3,$5F,$A8,$D7,$F0,$9A,$7E,$24
	.dc.b	$93,$F7,$DB,$1F,$61,$AF,$E0,$5A,$FE,$05,$AF,$E0,$5F,$CA,$7C,$0F
	.dc.b	$F2,$9C,$8D,$25,$7F,$25,$8C,$B6,$33,$5F,$C8,$67,$2D,$8C,$EF,$FE
	.dc.b	$7B,$9C,$B6,$38,$CF,$E5,$DF,$1D,$7E,$6D,$B1,$F9,$46,$BB,$8D,$11
	.dc.b	$D3,$5F,$D1,$7A,$DE,$D2,$47,$AD,$B1,$E9,$6D,$79,$DB,$3F,$A9,$6C
	.dc.b	$FE,$A5,$AF,$EE,$5B,$66,$92,$7C,$04,$9F,$03,$6B,$C0,$D3,$C0,$58
	.dc.b	$F0,$35,$F0,$36,$BC,$0D,$AF,$04,$F6,$25,$76,$6C,$F6,$69,$D8,$A7
	.dc.b	$62,$D7,$66,$D7,$6A,$E8,$4A,$E8,$D9,$E8,$D7,$A3,$6B,$A3,$6B,$A5
	.dc.b	$72,$25,$72,$6B,$C9,$B5,$C9,$B3,$CA,$78,$10,$E0,$59,$E0,$D8,$E0
	.dc.b	$E8,$91,$A6,$C6,$A3,$04,$70,$49,$FA,$84,$9F,$92,$DA,$F8,$0D,$3D
	.dc.b	$25,$8F,$23,$5F,$0B,$6B,$C0,$DA,$F0,$4F,$F3,$C7,$FA,$91,$23,$FA
	.dc.b	$43,$63,$51,$82,$38,$21,$FA,$C7,$FA,$97,$DD,$BC,$DF,$F1,$93,$63
	.dc.b	$FC,$58,$D9,$FF,$13,$37,$0F,$F7,$F3,$67,$FB,$23,$67,$FC,$2A,$D4
	.dc.b	$D2,$57,$80,$93,$E0,$6C,$F8,$1B,$5E,$06,$CF,$81,$B3,$E0,$6B,$E0
	.dc.b	$6D,$78,$27,$B1,$27,$B3,$6B,$B3,$4E,$C5,$8E,$CD,$7B,$36,$7B,$47
	.dc.b	$42,$4F,$46,$CF,$46,$9D,$0A,$74,$2C,$F4,$9E,$44,$9E,$4D,$9E,$4D
	.dc.b	$79,$36,$79,$4F,$02,$4F,$06,$BC,$1B,$1C,$23,$44,$34,$57,$4E,$08
	.dc.b	$E0,$95,$F5,$09,$3F,$25,$B3,$F0,$1B,$5E,$96,$CF,$91,$B3,$E1,$6B
	.dc.b	$E0,$6D,$78,$27,$B1,$27,$B3,$6B,$B3,$4E,$C5,$8E,$CF,$E2,$5F,$F2
	.dc.b	$1C,$FE,$9D,$B4,$9F,$D7,$B6,$3F,$C8,$D3,$B7,$F0,$47,$09,$D9,$1D
	.dc.b	$CF,$17,$46,$23,$FC,$99,$F9,$CB,$DE,$FF,$56,$36,$3F,$D4,$CD,$9F
	.dc.b	$F5,$03,$70,$E8,$D9,$FF,$2F,$36,$7F,$D1,$ED,$4D,$25,$78,$09,$3F
	.dc.b	$C2,$BC,$DF,$E1,$A4,$70,$4A,$FE,$1B,$19,$6C,$66,$FF,$87,$CE,$5B
	.dc.b	$1C,$67,$E8,$CF,$C5,$7F,$89,$B5,$FE,$26,$D7,$F8,$9B,$71,$34,$93
	.dc.b	$FC,$99,$B3,$F6,$1A,$FF,$A0,$BE,$F1,$E2,$7F,$B6,$DB,$5F,$ED,$56
	.dc.b	$CF,$FB,$3D,$B8,$7D,$B6,$CF,$FA,$35,$B3,$FE,$BE,$6A,$69,$2B,$C0
	.dc.b	$49,$F0,$36,$7C,$0D,$AF,$03,$67,$C0,$D9,$F0,$35,$F0,$36,$BC,$13
	.dc.b	$D8,$93,$D9,$B5,$D9,$A7,$62,$C7,$66,$BD,$9B,$3D,$A3,$A1,$27,$A3
	.dc.b	$67,$A3,$4E,$85,$3A,$16,$7A,$4F,$22,$4F,$26,$CF,$26,$BC,$9B,$3C
	.dc.b	$A7,$81,$27,$83,$5E,$0D,$8E,$11,$A2,$1A,$2B,$A7,$04,$70,$4A,$FA
	.dc.b	$84,$9F,$92,$D9,$F8,$0D,$AF,$4B,$67,$C8,$D9,$F0,$B5,$F0,$36,$BC
	.dc.b	$13,$D8,$93,$D9,$B5,$D9,$A7,$62,$C7,$67,$F1,$2F,$EA,$EB,$F6,$D5
	.dc.b	$FA,$88,$C4,$FF,$BA,$63,$11,$F8,$03,$27,$FB,$C5,$CB,$F8,$23,$24
	.dc.b	$38,$1F,$AF,$F8,$8F,$DC,$12,$7F,$70,$D8,$FD,$C3,$6F,$F7,$0D,$9E
	.dc.b	$47,$02,$4F,$06,$CF,$06,$9C,$23,$44,$8D,$36,$35,$18,$23,$82,$47
	.dc.b	$F0,$79,$F2,$36,$3B,$B4,$E8,$59,$E6,$D8,$E6,$FF,$0F,$9E,$6D,$8E
	.dc.b	$6D,$39,$96,$79,$B6,$39,$D2,$B9,$B6,$39,$B6,$79,$B6,$B9,$B6,$79
	.dc.b	$D7,$22,$3C,$9B,$5C,$9B,$5C,$9B,$3C,$8E,$04,$9E,$0D,$9E,$0D,$38
	.dc.b	$46,$89,$1A,$6C,$6A,$30,$47,$04,$8F,$A0,$43,$E4,$16,$BE,$03,$6B
	.dc.b	$D2,$D9,$F2,$34,$F0,$96,$BC,$0D,$8F,$03,$F6,$4F,$EB,$53,$FB,$02
	.dc.b	$47,$EB,$DB,$1F,$AD,$6B,$C2,$34,$47,$49,$84,$F3,$DC,$3F,$E2,$3B
	.dc.b	$FF,$0E,$46,$2B,$84,$62,$75,$1C,$5B,$5D,$6B,$82,$AE,$15,$C8,$E0
	.dc.b	$58,$E1,$5C,$95,$72,$AE,$31,$CE,$B8,$B4,$E2,$F3,$C1,$E7,$39,$BC
	.dc.b	$B1,$86,$C6,$67,$83,$6B,$A5,$6E,$73,$5C,$08,$E0,$93,$C2,$B9,$5F
	.dc.b	$F9,$36,$F9,$4E,$6F,$8B,$6F,$8D,$6E,$F8,$C7,$F8,$3D,$B3,$FE,$0F
	.dc.b	$6D,$F4,$AE,$17,$D1,$B7,$D0,$FF,$0D,$74,$7F,$3A,$23,$FC,$E5,$B8
	.dc.b	$7F,$37,$6B,$FE,$2B,$6B,$CA,$38,$11,$FE,$7D,$D1,$FD,$18,$91,$FD
	.dc.b	$15,$B1,$A8,$C1,$1C,$10,$FF,$1D,$47,$07,$F2,$E3,$EF,$3F,$CF,$53
	.dc.b	$88,$FC,$E4,$61,$C3,$5E,$2D,$7F,$42,$D7,$8C,$7F,$94,$E3,$11,$DE
	.dc.b	$30,$43,$C0,$E5,$B1,$C0,$FC,$54,$78,$30,$E8,$49,$DB,$63,$CE,$D7
	.dc.b	$84,$68,$8E,$9E,$FD,$3E,$72,$38,$24,$7C,$28,$CB,$63,$2F,$C4,$8C
	.dc.b	$B6,$3D,$7D,$3F,$0C,$8E,$09,$1F,$3A,$32,$D8,$CB,$F4,$23,$2D,$8F
	.dc.b	$8F,$D3,$F3,$C8,$E0,$91,$F6,$A3,$2D,$8C,$BF,$72,$32,$D8,$FA,$7B
	.dc.b	$FD,$B2,$38,$24,$7E,$1A,$32,$D8,$CB,$F8,$88,$CB,$63,$EF,$FC,$3F
	.dc.b	$E1,$C8,$E0,$91,$F9,$28,$CB,$63,$2F,$E5,$23,$2D,$8F,$C5,$F5,$7E
	.dc.b	$4C,$8E,$09,$1F,$9E,$8C,$B6,$32,$FE,$82,$32,$D8,$FC,$BF,$C3,$FE
	.dc.b	$7C,$8E,$09,$1F,$AA,$8C,$B6,$32,$FE,$B2,$32,$D8,$FD,$1F,$C3,$FE
	.dc.b	$AC,$8E,$09,$1F,$B6,$8C,$B6,$32,$FE,$E2,$32,$D8,$FD,$7F,$C3,$FE
	.dc.b	$DC,$8E,$09,$1F,$C0,$A3,$2D,$8C,$BF,$C1,$23,$2D,$8F,$DD,$FC,$3F
	.dc.b	$F0,$32,$38,$24,$7F,$0E,$8C,$B6,$32,$FF,$10,$8C,$B6,$3F,$83,$FC
	.dc.b	$3F,$F0,$F2,$38,$24,$7F,$1A,$8C,$B6,$32,$FF,$1C,$8C,$B6,$3F,$89
	.dc.b	$F5,$FF,$1B,$23,$82,$47,$F2,$68,$CB,$63,$2F,$F2,$88,$CB,$63,$F8
	.dc.b	$FF,$C3,$FF,$27,$23,$82,$47,$F3,$28,$CB,$63,$2F,$F3,$48,$CB,$63
	.dc.b	$F9,$5F,$C3,$FF,$33,$23,$82,$47,$F3,$E8,$CB,$63,$2F,$F4,$08,$CB
	.dc.b	$63,$F9,$BF,$C3,$FF,$3F,$23,$82,$47,$F4,$A8,$CB,$63,$2F,$F4,$C8
	.dc.b	$CB,$63,$FA,$1F,$C3,$7F,$86,$F9,$4D,$24,$7F,$88,$67,$FA,$E4,$61
	.dc.b	$FF,$0E,$18,$8D,$4F,$26,$CE,$FA,$11,$FD,$72,$30,$D3,$C6,$4A,$FE
	.dc.b	$CB,$1F,$D7,$C9,$19,$8F,$EC,$F1,$87,$71,$97,$A1,$FD,$C5,$7B,$BF
	.dc.b	$E3,$96,$9E,$D2,$C7,$B1,$B5,$E8,$8F,$F2,$A4,$60,$FF,$2A,$95,$CC
	.dc.b	$F5,$24,$75,$6C,$75,$6C,$F5,$8F,$8B,$18,$9E,$AD,$8E,$B1,$CC,$8F
	.dc.b	$36,$CF,$37,$C9,$1D,$1C,$9D,$0A,$F4,$39,$12,$39,$4F,$86,$39,$19
	.dc.b	$9E,$4D,$39,$4F,$08,$EB,$1C,$23,$0E,$CC,$47,$28,$C9,$1C,$12,$3A
	.dc.b	$B5,$E2,$60,$86,$08,$FC,$22,$4F,$AD,$A7,$98,$B1,$E5,$6C,$F8,$E3
	.dc.b	$AC,$60,$F1,$15,$CC,$F5,$24,$75,$6C,$75,$6C,$F5,$8F,$3C,$62,$7A
	.dc.b	$B6,$3A,$C7,$32,$3C,$DB,$3C,$DF,$24,$74,$72,$74,$2B,$D0,$E4,$48
	.dc.b	$E5,$3E,$18,$E4,$66,$7F,$02,$FF,$98,$1F,$D4,$B6,$3F,$38,$E8,$87
	.dc.b	$E3,$CD,$47,$F9,$58,$E2,$E8,$FC,$79,$A3,$F7,$51,$D1,$FF,$63,$7D
	.dc.b	$E7,$ED,$1F,$E2,$6D,$AF,$D6,$36,$BF,$D9,$6D,$9E,$27,$FA,$ED,$C2
	.dc.b	$69,$1D,$12,$3F,$92,$46,$5B,$19,$3C,$65,$9F,$1B,$63,$C6,$D3,$C6
	.dc.b	$F2,$23,$C9,$B1,$C9,$A7,$22,$C7,$23,$81,$23,$83,$63,$83,$4E,$11
	.dc.b	$A2,$46,$9B,$1A,$8C,$11,$C1,$0F,$94,$7B,$4A,$F9,$9A,$78,$4B,$1D
	.dc.b	$9B,$1D,$5A,$75,$9F,$A0,$F1,$6C,$F1,$6C,$F1,$6C,$F1,$3C,$8E,$1E
	.dc.b	$A4,$70,$47,$64,$8F,$54,$65,$B1,$93,$C8,$59,$F2,$36,$3C,$8D,$3C
	.dc.b	$8F,$22,$3C,$9B,$1C,$9A,$72,$2C,$72,$38,$12,$38,$3F,$84,$3F,$DA
	.dc.b	$F3,$FA,$96,$9F,$AD,$3E,$A9,$89,$FF,$65,$C6,$E3,$F1,$26,$4F,$F6
	.dc.b	$F3,$97,$F2,$46,$48,$09,$E1,$3E,$23,$FB,$75,$FF,$63,$8C,$5F,$F8
	.dc.b	$45,$B5,$FE,$10,$6C,$EE,$B8,$D7,$5A,$56,$A3,$0D,$8F,$0C,$F0,$6C
	.dc.b	$F7,$8C,$47,$26,$CF,$18,$FE,$0D,$EB,$7F,$09,$25,$7F,$04,$6C,$7F
	.dc.b	$14,$6D,$FF,$14,$6C,$FF,$14,$6C,$FF,$1C,$78,$BF,$59,$FA,$98,$1A
	.dc.b	$49,$FA,$AD,$9F,$A6,$D7,$94,$68,$8F,$D8,$F3,$3F,$91,$12,$3F,$8E
	.dc.b	$B6,$3F,$96,$B6,$FF,$96,$B6,$7F,$96,$B6,$7F,$9A,$B5,$FE,$58,$DB
	.dc.b	$34,$93,$E0,$24,$F8,$1B,$5E,$06,$9E,$02,$C7,$81,$AF,$81,$B5,$E0
	.dc.b	$6D,$78,$27,$B1,$2B,$B3,$67,$B3,$4E,$C5,$3B,$16,$BB,$36,$BB,$57
	.dc.b	$42,$57,$46,$CF,$46,$BD,$1B,$5D,$1B,$5D,$2B,$91,$2B,$93,$5E,$4D
	.dc.b	$AE,$4D,$9E,$53,$C0,$87,$02,$CF,$06,$C7,$07,$44,$8D,$36,$35,$18
	.dc.b	$23,$82,$4F,$D4,$24,$FC,$96,$D7,$C0,$69,$E9,$2C,$79,$1A,$F8,$5B
	.dc.b	$5E,$06,$D7,$82,$7F,$45,$1F,$C7,$A3,$F5,$E4,$70,$48,$D9,$0F,$E5
	.dc.b	$71,$B7,$07,$F1,$3F,$8B,$74,$FF,$0C,$6D,$7F,$AB,$5B,$3F,$EA,$C6
	.dc.b	$E1,$FE,$85,$6C,$F3,$AE,$44,$79,$36,$B9,$36,$B9,$36,$79,$1C,$09
	.dc.b	$3C,$1B,$3C,$1A,$70,$8D,$12,$34,$D8,$D4,$60,$8E,$09,$1F,$D0,$EB
	.dc.b	$DA,$DA,$F3,$36,$7C,$2D,$3B,$16,$BA,$B6,$3A,$AA,$EA,$DA,$EA,$D9
	.dc.b	$EA,$D3,$A9,$6B,$AB,$63,$AD,$74,$25,$74,$6C,$74,$6C,$F4,$6D,$74
	.dc.b	$6C,$F4,$AE,$44,$79,$36,$B9,$36,$B9,$36,$79,$1C,$09,$3C,$1B,$3C
	.dc.b	$1A,$70,$8D,$12,$34,$D8,$D4,$60,$8E,$09,$1F,$4C,$87,$C9,$2D,$7C
	.dc.b	$06,$D7,$A5,$B3,$F9,$47,$FB,$E1,$F5,$4B,$3F,$A8,$6C,$FF,$86,$63
	.dc.b	$11,$F9,$73,$33,$C9,$FC,$51,$83,$11,$C1,$C9,$C4,$FF,$06,$7E,$98
	.dc.b	$F5,$BF,$C0,$AD,$AF,$F0,$E6,$1F,$EF,$B8,$C5,$23,$15,$C0,$95,$C0
	.dc.b	$EA,$E2,$78,$46,$0D,$3F,$E2,$9A,$EB,$5D,$27,$32,$9C,$C7,$F1,$1E
	.dc.b	$AF,$88,$47,$E1,$4F,$E9,$63,$0F,$C2,$6B,$F3,$9A,$F3,$8E,$38,$27
	.dc.b	$33,$FB,$BE,$AF,$A4,$4A,$DB,$63,$E7,$B5,$FB,$6D,$7F,$94,$35,$E9
	.dc.b	$1E,$0E,$AF,$BC,$47,$04,$8F,$B7,$19,$6C,$65,$FC,$4C,$65,$B1,$EC
	.dc.b	$EA,$FC,$51,$1C,$12,$3F,$0F,$19,$6C,$65,$FC,$AC,$65,$B1,$F2,$3A
	.dc.b	$BF,$2C,$47,$04,$8F,$C9,$C6,$5B,$19,$7F,$43,$19,$6C,$7D,$4E,$AF
	.dc.b	$D1,$11,$C1,$23,$F3,$F1,$96,$C6,$5F,$D6,$C6,$5B,$1F,$80,$E9,$FE
	.dc.b	$B4,$47,$FA,$D3,$63,$FA,$D3,$75,$FE,$B4,$D9,$FE,$C8,$D9,$FE,$84
	.dc.b	$F1,$7F,$7A,$FE,$EC,$91,$E1,$24,$7E,$E2,$32,$D8,$CB,$C9,$AF,$FA
	.dc.b	$17,$EF,$5F,$0D,$FE,$7E,$6C,$FE,$A9,$B3,$FB,$23,$FD,$30,$E2,$D3
	.dc.b	$CE,$B9,$11,$E4,$DA,$E4,$2B,$94,$F1,$36,$48,$DC,$69,$F0,$CF,$3A
	.dc.b	$F0,$C6,$2B,$84,$62,$3B,$4E,$E7,$B4,$74,$8C,$3C,$4C,$47,$23,$73
	.dc.b	$CE,$BC,$53,$D2,$74,$F1,$9D,$57,$48,$D1,$D0,$91,$A9,$E3,$1A,$8C
	.dc.b	$3A,$8F,$09,$1E,$CD,$78,$36,$BC,$A4,$AF,$13,$63,$BB,$67,$B1,$D0
	.dc.b	$87,$42,$BB,$AF,$29,$27,$C1,$5D,$88,$F6,$6D,$76,$3A,$12,$BA,$4F
	.dc.b	$23,$81,$23,$84,$69,$F1,$CF,$4A,$F1,$46,$2B,$84,$62,$3B,$4E,$E7
	.dc.b	$B4,$74,$8C,$3C,$4C,$47,$23,$73,$CE,$BC,$53,$D2,$74,$F1,$9D,$57
	.dc.b	$48,$D1,$F8,$13,$FB,$CD,$7E,$B1,$B1,$F9,$77,$F2,$A6,$2B,$FA,$AC
	.dc.b	$EE,$32,$43,$F3,$06,$CF,$EF,$8E,$DC,$9F,$BF,$F8,$0F,$DF,$12,$7F
	.dc.b	$7C,$D8,$FD,$7B,$6F,$F4,$0D,$9E,$47,$02,$4F,$06,$CF,$06,$9C,$23
	.dc.b	$44,$8D,$36,$35,$18,$23,$82,$47,$F1,$29,$F2,$36,$3B,$B4,$E8,$59
	.dc.b	$E6,$D8,$E6,$FF,$13,$9E,$6D,$8E,$6D,$39,$96,$79,$B6,$39,$D2,$B9
	.dc.b	$B6,$39,$B6,$79,$B6,$B9,$B6,$79,$D7,$22,$3C,$9B,$5C,$9B,$5C,$9B
	.dc.b	$3C,$8E,$04,$9E,$0D,$9E,$0D,$38,$46,$89,$1A,$6C,$6A,$30,$47,$04
	.dc.b	$8F,$A0,$43,$E4,$16,$BE,$03,$6B,$D2,$D9,$F2,$34,$F0,$96,$BC,$0D
	.dc.b	$8F,$03,$F6,$4F,$D8,$CF,$EB,$DF,$C3,$18,$8F,$D7,$C6,$63,$81,$F9
	.dc.b	$93,$47,$EB,$9D,$57,$DB,$8C,$90,$13,$F3,$B7,$9F,$FD,$63,$0F,$EA
	.dc.b	$D1,$8C,$3F,$CB,$2D,$AF,$EC,$ED,$9D,$D7,$1A,$EB,$4A,$D4,$61,$B1
	.dc.b	$E1,$9E,$0D,$9E,$F1,$88,$E4,$D9,$E3,$1F,$DD,$FD,$FF,$F0,$31,$2B
	.dc.b	$FC,$0C,$D8,$FE,$EE,$DB,$FE,$D0,$D9,$FF,$08,$36,$7F,$C1,$6F,$17
	.dc.b	$F9,$2B,$F3,$09,$1F,$C8,$9B,$1F,$C7,$9A,$F0,$8D,$11,$D3,$FE,$7E
	.dc.b	$F7,$DE,$CF,$BD,$B1,$FE,$88,$6C,$FF,$85,$9B,$7F,$E0,$B6,$CF,$D0
	.dc.b	$6C,$FD,$E6,$A6,$92,$BC,$04,$9F,$03,$67,$C0,$DA,$F0,$36,$7C,$0D
	.dc.b	$9F,$03,$5F,$03,$6B,$C1,$3D,$89,$3D,$9B,$5D,$9A,$76,$2C,$76,$6B
	.dc.b	$D9,$B3,$DA,$3A,$12,$7A,$36,$7A,$34,$E8,$53,$A1,$67,$A4,$F2,$24
	.dc.b	$F2,$6C,$F2,$6B,$C9,$B3,$CA,$78,$12,$78,$35,$E0,$D8,$E1,$1A,$21
	.dc.b	$A2,$BA,$70,$47,$04,$AF,$A8,$49,$F9,$2D,$9F,$80,$DA,$F4,$B6,$7C
	.dc.b	$8D,$9F,$0B,$5F,$03,$6B,$C1,$3D,$89,$3D,$9B,$5D,$9A,$76,$2C,$76
	.dc.b	$7F,$12,$FF,$10,$3E,$B9,$67,$F5,$2D,$AF,$E5,$51,$88,$FC,$09,$99
	.dc.b	$FD,$A4,$E8,$8E,$09,$1D,$C7,$DF,$3C,$8F,$E3,$0D,$9F,$F8,$02,$D9
	.dc.b	$FF,$47,$B7,$0F,$F4,$23,$67,$F9,$4B,$67,$F9,$BB,$5E,$F5,$D8,$93
	.dc.b	$D9,$B1,$D9,$B5,$D9,$B3,$D9,$B3,$D9,$AF,$68,$E8,$47,$A3,$67,$A3
	.dc.b	$4E,$85,$8E,$8D,$7A,$1C,$89,$1C,$9B,$1C,$9A,$72,$29,$CA,$38,$12
	.dc.b	$38,$36,$38,$35,$E1,$1A,$23,$A6,$BA,$30,$43,$04,$AF,$A8,$4A,$F9
	.dc.b	$2D,$9F,$80,$D9,$F4,$B6,$BC,$8D,$9F,$0B,$67,$C0,$D7,$C1,$5D,$89
	.dc.b	$3D,$9B,$1D,$9B,$5D,$9B,$3D,$9B,$3D,$9A,$F6,$8E,$84,$7A,$36,$7A
	.dc.b	$34,$E8,$58,$E8,$D7,$A1,$C8,$91,$F8,$17,$F7,$55,$FA,$76,$D7,$F8
	.dc.b	$86,$31,$1F,$85,$33,$5F,$B3,$7F,$1A,$60,$FD,$B4,$70,$7F,$14,$64
	.dc.b	$87,$F8,$63,$F5,$37,$1F,$F8,$45,$B5,$FE,$28,$C3,$FD,$F5,$18,$A4
	.dc.b	$62,$B8,$12,$B8,$1D,$5C,$4F,$08,$C1,$A7,$FC,$6B,$5D,$6B,$A4,$E6
	.dc.b	$53,$98,$FE,$0F,$E4,$7C,$52,$3F,$0F,$04,$62,$BE,$1B,$5F,$9E,$D7
	.dc.b	$9C,$F1,$8C,$47,$F1,$2F,$23,$E9,$12,$7B,$36,$3E,$7B,$5F,$B6,$D7
	.dc.b	$F8,$EB,$5E,$91,$E7,$EC,$FB,$C4,$70,$48,$FB,$71,$96,$C6,$5F,$C4
	.dc.b	$C6,$5B,$1F,$0F,$C8,$FC,$51,$1C,$12,$3F,$0F,$19,$6C,$65,$FC,$AC
	.dc.b	$65,$B1,$FD,$0F,$CC,$79,$AF,$2F,$FC,$BB,$18,$6C,$66,$7F,$01,$39
	.dc.b	$6C,$66,$BF,$0C,$27,$4E,$1B,$AA,$BF,$CF,$B1,$86,$D7,$39,$D5,$73
	.dc.b	$8C,$4E,$AF,$CA,$47,$C6,$D9,$F0,$4F,$07,$94,$66,$BA,$DF,$86,$73
	.dc.b	$7C,$88,$F2,$AE,$11,$FD,$83,$AF,$F7,$C4,$7F,$34,$D9,$FD,$DB,$5F
	.dc.b	$E0,$ED,$79,$47,$02,$3F,$C0,$B7,$FE,$14,$48,$FD,$33,$63,$51,$82
	.dc.b	$38,$21,$FC,$5A,$38,$3F,$C6,$3A,$BF,$C0,$24,$9F,$F0,$0B,$63,$FC
	.dc.b	$02,$DC,$7F,$C0,$2D,$9F,$F0,$0B,$67,$FC,$02,$F1,$7B,$47,$EF,$63
	.dc.b	$F9,$21,$1C,$12,$3F,$90,$C6,$5B,$19,$79,$9F,$E1,$BF,$BF,$7C,$67
	.dc.b	$FA,$C5,$B3,$FC,$45,$B3,$F9,$40,$E2,$FF,$1B,$18,$AE,$44,$AE,$4D
	.dc.b	$8E,$55,$C6,$7F,$D9,$2D,$8F,$F6,$BB,$5E,$91,$C8,$8F,$29,$E9,$5E
	.dc.b	$2A,$E0,$48,$CD,$76,$6D,$76,$6B,$D8,$E9,$1D,$AB,$A4,$EA,$BA,$4E
	.dc.b	$4E,$85,$AE,$8D,$7A,$57,$28,$C4,$F2,$9E,$E4,$9D,$36,$7A,$B4,$EA
	.dc.b	$53,$AB,$A8,$E2,$47,$04,$8E,$31,$96,$C6,$5E,$66,$08,$60,$93,$E9
	.dc.b	$9E,$51,$8A,$F8,$C4,$AF,$7B,$63,$D1,$5C,$67,$C0,$D8,$EE,$D7,$A4
	.dc.b	$72,$23,$CA,$7A,$57,$8A,$B8,$12,$33,$5D,$9B,$5D,$9A,$F6,$3A,$47
	.dc.b	$6A,$E9,$3A,$AE,$93,$93,$A1,$6B,$EA,$9F,$E9,$58,$FD,$4B,$67,$F4
	.dc.b	$EF,$DC,$31,$3F,$E1,$B9,$DC,$64,$87,$42,$9F,$97,$38,$9F,$BB,$F8
	.dc.b	$6F,$DD,$12,$7F,$74,$D8,$FD,$8B,$6F,$F5,$0D,$9F,$D9,$B6,$7F,$66
	.dc.b	$D7,$F8,$0B,$68,$D2,$4F,$80,$93,$E0,$6D,$78,$1A,$78,$0B,$1E,$06
	.dc.b	$BE,$06,$D7,$81,$B5,$E0,$9E,$C4,$AE,$CD,$9E,$CD,$3B,$14,$EC,$5A
	.dc.b	$EC,$DA,$ED,$5D,$09,$5D,$1B,$3D,$1A,$F4,$6D,$74,$6D,$74,$AE,$44
	.dc.b	$AE,$4D,$79,$36,$B9,$36,$79,$4F,$02,$1C,$0B,$3C,$1B,$1C,$1D,$12
	.dc.b	$34,$D8,$D4,$60,$8E,$09,$3F,$50,$93,$F2,$5B,$5F,$01,$A7,$A4,$B1
	.dc.b	$E4,$6B,$E1,$6D,$78,$1B,$5E,$09,$FD,$14,$7E,$AE,$3F,$5F,$1F,$78
	.dc.b	$FD,$A3,$87,$EF,$98,$8E,$04,$79,$90,$D0,$21,$FE,$2D,$DF,$FC,$7E
	.dc.b	$48,$FF,$1E,$B6,$3F,$C5,$6D,$BF,$F1,$4E,$C8,$C5,$7F,$8E,$9B,$5D
	.dc.b	$63,$FB,$A7,$C3,$7F,$77,$23,$82,$47,$F6,$E8,$CB,$63,$37,$FD,$9E
	.dc.b	$72,$D8,$E3,$3F,$C6,$67,$E3,$62,$69,$27,$F8,$C3,$67,$FC,$18,$D7
	.dc.b	$94,$68,$8F,$C8,$3E,$57,$DF,$BE,$0B,$FC,$AF,$87,$F8,$32,$31,$5F
	.dc.b	$EA,$36,$D7,$F8,$4D,$B3,$BA,$E3,$5F,$52,$72,$D8,$CD,$78,$1A,$A7
	.dc.b	$6E,$67,$C1,$3A,$9C,$CE,$E3,$04,$9D,$92,$3C,$D1,$96,$C6,$6B,$D0
	.dc.b	$D3,$CC,$58,$F1,$1C,$09,$1C,$1B,$1C,$1A,$70,$8D,$12,$34,$D8,$D4
	.dc.b	$60,$8E,$08,$7B,$DF,$23,$4E,$E5,$8E,$8D,$8E,$6D,$39,$D7,$CA,$25
	.dc.b	$7C,$5A,$E3,$18,$AE,$ED,$AE,$ED,$9D,$D7,$1A,$EF,$39,$6C,$66,$BC
	.dc.b	$2D,$7C,$84,$9E,$0E,$A7,$C3,$3B,$9C,$CE,$E3,$04,$9D,$92,$3C,$F1
	.dc.b	$96,$C6,$6B,$D2,$D3,$CE,$58,$F1,$1C,$09,$1C,$1B,$1C,$1A,$70,$8D
	.dc.b	$1F,$87,$7F,$61,$5F,$AA,$6D,$7F,$23,$8C,$47,$E2,$CC,$D7,$ED,$DF
	.dc.b	$C9,$98,$31,$1C,$1C,$9C,$48,$7E,$FB,$D2,$FD,$51,$27,$F7,$8D,$8F
	.dc.b	$DE,$36,$FF,$81,$36,$7F,$78,$D9,$E9,$1C,$88,$F2,$6C,$F2,$69,$C8
	.dc.b	$B1,$C8,$E0,$48,$E0,$D8,$E0,$D3,$84,$68,$91,$A6,$C6,$A3,$04,$70
	.dc.b	$43,$F8,$94,$FB,$5B,$1E,$66,$9E,$12,$CF,$66,$C7,$56,$9D,$69,$5D
	.dc.b	$5B,$3D,$5B,$3D,$5B,$5D,$5B,$3D,$5B,$3D,$6B,$A1,$27,$A3,$63,$A3
	.dc.b	$6B,$A3,$67,$A3,$67,$A4,$72,$23,$C9,$B3,$C9,$A7,$22,$C7,$23,$81
	.dc.b	$23,$83,$63,$83,$4E,$11,$A2,$46,$9B,$1A,$8C,$11,$C1,$F8,$D3,$F6
	.dc.b	$33,$FA,$F2,$47,$EB,$63,$EE,$9A,$71,$5F,$52,$30,$7E,$C2,$78,$B6
	.dc.b	$77,$1F,$E2,$0E,$26,$8F,$13,$FB,$0C,$61,$B1,$9A,$FF,$0E,$63,$FE
	.dc.b	$14,$9C,$92,$32,$7F,$88,$6B,$93,$8C,$11,$85,$5C,$E7,$53,$98,$C4
	.dc.b	$EA,$7A,$5F,$8E,$30,$7F,$8B,$FE,$3D,$D5,$FE,$39,$6D,$7C,$96,$CF
	.dc.b	$F1,$76,$BF,$E0,$F6,$BF,$E3,$86,$BD,$67,$A1,$27,$A3,$63,$A3,$E1
	.dc.b	$E5,$93,$E1,$6C,$F8,$5A,$F8,$5A,$F8,$5A,$FB,$5A,$A9,$D1,$EF,$25
	.dc.b	$7B,$DB,$3E,$F6,$BE,$F6,$BE,$F6,$BF,$2D,$F9,$5D,$12,$7E,$53,$67
	.dc.b	$E5,$35,$F9,$4D,$7E,$53,$5F,$AC,$D7,$D5,$C5,$F5,$C9,$3F,$5D,$B1
	.dc.b	$F5,$DA,$FD,$76,$BF,$5D,$AF,$E0,$DF,$C1,$74,$49,$FC,$13,$67,$F0
	.dc.b	$4D,$7F,$04,$D7,$F0,$4D,$7F,$1C,$D7,$E9,$78,$BF,$E4,$6C,$4D,$25
	.dc.b	$7E,$86,$76,$D9,$DB,$67,$73,$F9,$C3,$65,$9D,$B6,$37,$AA,$BF,$31
	.dc.b	$1F,$A1,$23,$FA,$16,$CF,$E8,$5A,$FE,$85,$AF,$28,$E0,$47,$F5,$1F
	.dc.b	$21,$EF,$24,$7A,$DB,$1E,$76,$D7,$96,$3F,$6D,$18,$9F,$D7,$36,$33
	.dc.b	$1C,$C8,$F3,$6C,$F3,$8F,$D1,$4F,$16,$CE,$8E,$A5,$7A,$9C,$C9,$3C
	.dc.b	$ED,$5C,$DB,$5C,$6B,$A9,$C1,$EA,$D3,$AD,$73,$AE,$44,$9E,$55,$D2
	.dc.b	$B8,$1D,$2B,$84,$EA,$B8,$4E,$AB,$85,$6A,$31,$3A,$8C,$12,$BC,$AD
	.dc.b	$9F,$2C,$60,$8E,$08,$78,$DC,$B4,$CD,$78,$A3,$2D,$8C,$D7,$8E,$72
	.dc.b	$D8,$CC,$7B,$48,$FA,$9B,$5E,$78,$F0,$12,$BB,$36,$BA,$36,$B8,$9E
	.dc.b	$02,$BE,$03,$B1,$2B,$B5,$74,$25,$74,$6D,$72,$AE,$C7,$17,$B3,$4E
	.dc.b	$D5,$D2,$BF,$A9,$12,$7F,$A9,$36,$7F,$57,$18,$23,$82,$1F,$D5,$63
	.dc.b	$83,$FE,$E5,$FB,$F7,$C6,$7F,$B5,$9B,$5F,$C3,$DB,$3F,$C2,$DB,$7F
	.dc.b	$A9,$6C,$A6,$46,$1B,$5D,$49,$3D,$5B,$1D,$5B,$5D,$5B,$3D,$65,$5D
	.dc.b	$5B,$3A,$7B,$CF,$52,$3D,$5B,$5D,$5A,$75,$AE,$84,$AE,$95,$DA,$76
	.dc.b	$F6,$3A,$12,$BA,$36,$3A,$4F,$2A,$E9,$1C,$A7,$A1,$B3,$A5,$72,$24
	.dc.b	$72,$38,$4F,$29,$E0,$72,$9D,$BC,$A7,$83,$A9,$E1,$3A,$8E,$13,$A8
	.dc.b	$C4,$EA,$30,$49,$E8,$43,$89,$63,$83,$6B,$E0,$12,$7D,$2D,$8F,$23
	.dc.b	$6B,$C2,$D9,$F0,$4F,$62,$57,$66,$CE,$DF,$04,$F6,$23,$D9,$B5,$D9
	.dc.b	$A7,$6A,$E8,$4A,$E9,$5D,$A7,$6F,$63,$A1,$2B,$A3,$FE,$DB,$AF,$D4
	.dc.b	$BF,$AF,$AF,$EA,$0F,$D4,$31,$5F,$E7,$59,$DC,$64,$87,$E3,$CD,$9D
	.dc.b	$67,$9B,$93,$FC,$BB,$D9,$FE,$70,$25,$7F,$99,$38,$51,$89,$FF,$30
	.dc.b	$F1,$7F,$9E,$A3,$18,$FF,$6F,$8C,$5F,$19,$CB,$D8,$C1,$FC,$1F,$D8
	.dc.b	$FE,$14,$47,$04,$BF,$E1,$51,$96,$C6,$6F,$F8,$64,$E5,$B1,$C6,$7F
	.dc.b	$8C,$4F,$C0,$7F,$81,$B7,$23,$49,$3F,$35,$B3,$F5,$5A,$F4,$8D,$1F
	.dc.b	$EA,$7F,$BF,$7C,$57,$F8,$47,$0F,$F0,$44,$62,$7F,$C6,$8D,$AF,$F1
	.dc.b	$83,$67,$75,$C6,$BF,$8D,$4E,$5B,$19,$AF,$E5,$53,$96,$C6,$54,$65
	.dc.b	$B1,$98,$C1,$1C,$12,$7C,$ED,$9F,$3B,$6B,$CF,$3A,$24,$E9,$B1,$A8
	.dc.b	$C1,$1C,$12,$3D,$A7,$02,$C6,$DB,$1B,$8F,$7C,$6D,$A6,$CB,$1B,$7E
	.dc.b	$0C,$6D,$A6,$CB,$1B,$AF,$89,$3B,$6C,$ED,$B3,$BA,$F8,$95,$E6,$25
	.dc.b	$78,$9A,$F6,$6D,$72,$6C,$F2,$9E,$04,$38,$16,$78,$36,$38,$3A,$24
	.dc.b	$69,$B1,$A8,$C1,$1C,$12,$7E,$03,$67,$E0,$36,$BE,$04,$E8,$93,$A6
	.dc.b	$C6,$A3,$04,$70,$48,$F8,$27,$02,$C6,$DA,$FF,$7D,$3E,$C9,$6B,$F5
	.dc.b	$8D,$AF,$E0,$11,$88,$FC,$59,$99,$FD,$C4,$E8,$8E,$09,$1D,$CF,$D5
	.dc.b	$FC,$1F,$EA,$49,$1F,$BC,$6C,$7E,$F1,$B7,$FC,$09,$B3,$FB,$C6,$CF
	.dc.b	$EE,$9A,$F6,$8E,$84,$7A,$36,$7A,$34,$E8,$58,$E8,$D7,$A1,$C8,$91
	.dc.b	$C9,$B1,$C9,$A7,$22,$9C,$A3,$81,$23,$83,$63,$83,$5E,$11,$A2,$3A
	.dc.b	$6B,$A3,$04,$30,$4B,$57,$C8,$6C,$FB,$DB,$3E,$86,$D7,$8D,$B3,$E0
	.dc.b	$6C,$F7,$6B,$DE,$BB,$12,$7B,$36,$3B,$36,$BB,$36,$7B,$36,$7B,$35
	.dc.b	$ED,$1D,$08,$F4,$6C,$F4,$69,$D0,$B1,$D1,$AF,$43,$91,$23,$93,$63
	.dc.b	$93,$4E,$45,$39,$47,$07,$F6,$13,$FA,$E2,$4F,$F6,$36,$C7,$EA,$6B
	.dc.b	$E9,$BC,$27,$44,$75,$3C,$1F,$F3,$D6,$6F,$CF,$9F,$25,$FD,$86,$30
	.dc.b	$D8,$CD,$FF,$A3,$71,$FF,$0A,$46,$2F,$FC,$32,$D8,$CD,$71,$25,$71
	.dc.b	$3F,$D2,$D7,$D9,$C5,$A3,$0F,$49,$E2,$FF,$7C,$F8,$4F,$E2,$24,$AF
	.dc.b	$F0,$33,$63,$E2,$35,$FE,$34,$D7,$FC,$02,$D7,$FB,$FB,$FC,$73,$DE
	.dc.b	$93,$FE,$11,$6C,$FC,$F6,$BF,$C9,$5A,$FF,$83,$1A,$FF,$82,$9A,$FF
	.dc.b	$82,$3B,$3F,$94,$92,$7F,$C3,$CD,$8F,$B8,$D7,$F9,$93,$5F,$F0,$CB
	.dc.b	$5F,$F0,$BB,$FE,$33,$FC,$01,$D9,$FE,$17,$6D,$7C,$EA,$46,$16,$69
	.dc.b	$FF,$5F,$4E,$63,$13,$FE,$CD,$8C,$4E,$AB,$B1,$1E,$D1,$A9,$E1,$39
	.dc.b	$7B,$CF,$13,$71,$E9,$23,$FD,$16,$3C,$F1,$87,$FA,$44,$FA,$5F,$EB
	.dc.b	$0D,$7A,$1A,$8E,$0E,$9A,$FF,$B6,$D7,$C8,$FF,$8F,$5B,$5F,$1D,$B3
	.dc.b	$E8,$3F,$DD,$CE,$2F,$E2,$C7,$D0,$AE,$44,$AE,$4D,$8E,$46,$E3,$24
	.dc.b	$73,$5D,$63,$CD,$3C,$6B,$BC,$E6,$BC,$73,$9A,$D9,$CE,$B7,$19,$9D
	.dc.b	$CE,$4D,$BE,$08,$D3,$E5,$31,$1B,$78,$98,$24,$F6,$8D,$47,$18,$C3
	.dc.b	$A8,$EF,$18,$23,$82,$47,$68,$CB,$63,$2F,$56,$D7,$C3,$8C,$1B,$21
	.dc.b	$82,$3B,$25,$7C,$06,$CF,$A9,$B3,$E9,$39,$B8,$AF,$2C,$74,$AE,$44
	.dc.b	$AE,$4D,$8E,$46,$E3,$24,$73,$5D,$63,$DB,$3C,$6B,$BC,$E6,$BC,$73
	.dc.b	$9A,$D9,$CE,$B7,$19,$9D,$CE,$4D,$BE,$08,$D3,$E5,$31,$1B,$78,$98
	.dc.b	$24,$F6,$8D,$47,$18,$C3,$A8,$EF,$18,$23,$82,$47,$68,$CB,$63,$2F
	.dc.b	$56,$D7,$C3,$8C,$1B,$21,$82,$3B,$25,$7C,$06,$CF,$A9,$B3,$E9,$39
	.dc.b	$B8,$AF,$2C,$74,$AE,$44,$AE,$4D,$8E,$46,$E3,$24,$73,$5D,$63,$DB
	.dc.b	$3C,$6B,$BC,$E6,$BC,$73,$9A,$D9,$CE,$B7,$19,$9D,$CE,$4D,$BE,$08
	.dc.b	$D3,$E5,$31,$1B,$78,$98,$24,$F6,$8D,$47,$18,$C3,$A8,$EF,$18,$23
	.dc.b	$82,$47,$68,$FE,$12,$FF,$A5,$6B,$F9,$9B,$6B,$F9,$8B,$FC,$68,$C5
	.dc.b	$7F,$49,$9D,$C6,$48,$7F,$1A,$36,$7F,$A4,$DD,$B9,$04,$FC,$01,$EB
	.dc.b	$FF,$64,$C3,$FA,$0C,$62,$FF,$B7,$36,$BF,$9F,$B6,$77,$5C,$6B,$AD
	.dc.b	$2B,$51,$86,$C7,$86,$78,$36,$7B,$C6,$23,$93,$67,$8C,$62,$76,$47
	.dc.b	$CB,$3B,$24,$66,$BC,$ED,$3C,$91,$A9,$E3,$19,$7C,$86,$09,$3C,$9B
	.dc.b	$1C,$8E,$0F,$94,$C9,$CA,$38,$12,$38,$46,$8E,$11,$97,$84,$6A,$31
	.dc.b	$1A,$8C,$11,$D9,$0D,$16,$BE,$49,$2B,$E0,$57,$48,$C5,$7A,$9B,$5E
	.dc.b	$56,$CE,$EB,$8D,$78,$6B,$B9,$2B,$71,$96,$C7,$8E,$78,$B6,$7C,$11
	.dc.b	$88,$E4,$D9,$E3,$18,$9D,$91,$F3,$4E,$C9,$19,$AF,$43,$4F,$24,$6A
	.dc.b	$78,$C6,$5F,$21,$82,$4F,$26,$C7,$23,$83,$E5,$32,$72,$8E,$04,$8E
	.dc.b	$11,$A3,$84,$65,$E1,$1A,$8C,$46,$A3,$04,$76,$43,$45,$AF,$92,$4A
	.dc.b	$F8,$15,$D2,$31,$5E,$A6,$D7,$95,$B3,$BA,$E3,$5E,$1A,$EE,$4A,$DC
	.dc.b	$65,$B1,$E3,$9E,$2D,$9F,$04,$62,$39,$36,$78,$C6,$27,$64,$7C,$D3
	.dc.b	$B2,$46,$6B,$D0,$D3,$C9,$1A,$9E,$31,$97,$C8,$60,$93,$C9,$B1,$C8
	.dc.b	$E0,$F9,$4C,$9C,$A3,$81,$23,$84,$68,$E1,$19,$78,$46,$A3,$11,$FC
	.dc.b	$34,$FE,$5E,$53,$F8,$59,$8A,$FE,$70,$DA,$FE,$9A,$FF,$1C,$31,$5F
	.dc.b	$E9,$59,$DC,$72,$21,$B2,$8F,$BF,$7A,$1F,$CD,$DB,$5F,$EB,$56,$CF
	.dc.b	$FA,$19,$B8,$7F,$A3,$9B,$3F,$D2,$1B,$3F,$D5,$9A,$F7,$AE,$C4,$9E
	.dc.b	$CD,$8E,$CD,$AE,$CD,$9E,$CD,$9E,$CD,$7B,$47,$42,$3D,$1B,$3D,$1A
	.dc.b	$74,$2C,$74,$6B,$D0,$E4,$48,$E4,$D8,$E4,$D3,$91,$4E,$51,$C0,$91
	.dc.b	$C1,$B1,$C1,$AF,$08,$D1,$1D,$35,$D1,$82,$18,$25,$7D,$42,$57,$C9
	.dc.b	$6C,$FC,$06,$CF,$A5,$B5,$E4,$6C,$F8,$5B,$3E,$06,$BE,$0A,$EC,$49
	.dc.b	$EC,$D8,$EC,$DA,$EC,$D9,$EC,$D9,$EC,$D7,$B4,$74,$23,$D1,$B3,$D1
	.dc.b	$A7,$42,$C7,$46,$BD,$0E,$44,$8E,$4D,$8E,$4D,$39,$14,$E5,$1C,$09
	.dc.b	$1C,$1B,$1C,$1A,$F0,$8D,$11,$D3,$5D,$18,$21,$82,$57,$D4,$25,$7C
	.dc.b	$96,$CF,$C0,$6C,$FA,$5B,$5E,$46,$CF,$85,$B3,$E0,$6B,$E0,$AE,$C4
	.dc.b	$9E,$CD,$8E,$CD,$AE,$CD,$9E,$CD,$9E,$CD,$7B,$47,$42,$3D,$1B,$3D
	.dc.b	$1A,$74,$2C,$74,$6B,$D0,$E4,$48,$E4,$D8,$E4,$D3,$91,$4E,$51,$C0
	.dc.b	$91,$C1,$FE,$17,$1F,$D1,$2B,$F9,$8B,$6B,$FA,$0B,$FC,$4C,$C5,$7F
	.dc.b	$4A,$7F,$87,$98,$31,$3C,$9F,$E2,$A6,$48,$60,$FF,$40,$FC,$8B,$D9
	.dc.b	$FF,$3D,$36,$BF,$D2,$38,$1D,$CE,$88,$E8,$C9,$6B,$FD,$33,$5C,$D5
	.dc.b	$73,$73,$5C,$9C,$D7,$26,$9A,$7A,$2A,$E9,$5C,$E3,$15,$CE,$31,$3A
	.dc.b	$AE,$91,$89,$D4,$60,$95,$DC,$F0,$9D,$1F,$0D,$73,$8D,$57,$39,$CD
	.dc.b	$74,$9C,$C7,$47,$91,$D2,$B9,$38,$AE,$4E,$6B,$93,$9A,$E5,$5C,$09
	.dc.b	$3A,$6C,$78,$89,$3C,$DB,$5D,$88,$F4,$AE,$44,$76,$64,$B3,$EC,$21
	.dc.b	$E5,$2D,$78,$9F,$01,$2B,$BB,$AA,$EC,$E6,$BB,$34,$D3,$E0,$7B,$12
	.dc.b	$BB,$57,$48,$C5,$74,$8C,$4E,$AB,$AC,$62,$75,$18,$25,$78,$0F,$11
	.dc.b	$D1,$F1,$57,$38,$D5,$73,$9C,$D7,$49,$CC,$74,$79,$1D,$2B,$93,$8A
	.dc.b	$E4,$E6,$B9,$39,$AE,$55,$C0,$93,$A6,$C7,$88,$93,$CD,$B5,$D8,$8F
	.dc.b	$4A,$E4,$47,$66,$4B,$3E,$C2,$1E,$52,$D7,$89,$F0,$12,$BB,$BA,$AE
	.dc.b	$CE,$6B,$B3,$4D,$3E,$07,$B1,$2B,$B5,$74,$8C,$57,$48,$C4,$EA,$BA
	.dc.b	$C6,$27,$51,$82,$57,$80,$F1,$1D,$1F,$15,$73,$8D,$57,$39,$CD,$74
	.dc.b	$9C,$C7,$47,$91,$D2,$B9,$38,$AE,$4E,$6B,$93,$9A,$E5,$5C,$09,$3A
	.dc.b	$6C,$78,$89,$3F,$48,$FF,$48,$C7,$F3,$36,$CF,$F4,$E8,$C4,$7F,$31
	.dc.b	$33,$3F,$D3,$DF,$E3,$26,$0F,$EA,$51,$C1,$FE,$50,$64,$80,$9F,$7E
	.dc.b	$F8,$4F,$E8,$58,$7F,$39,$8C,$5F,$F3,$E6,$D7,$F3,$26,$CE,$EB,$8D
	.dc.b	$75,$A5,$6A,$30,$D8,$F0,$CF,$06,$CF,$78,$C4,$72,$6C,$F1,$8C,$4E
	.dc.b	$C8,$F9,$67,$64,$8C,$D7,$9D,$A7,$92,$35,$3C,$63,$2F,$90,$C1,$27
	.dc.b	$93,$63,$91,$C1,$F2,$99,$39,$47,$02,$47,$08,$D1,$C2,$32,$F0,$8D
	.dc.b	$46,$23,$51,$82,$3B,$21,$A2,$D7,$C9,$25,$7C,$0A,$E9,$18,$AF,$53
	.dc.b	$6B,$CA,$D9,$DD,$71,$AF,$0D,$77,$25,$6E,$32,$D8,$F1,$CF,$16,$CF
	.dc.b	$82,$31,$1C,$9B,$3C,$63,$13,$B2,$3E,$69,$D9,$23,$35,$E8,$69,$E4
	.dc.b	$8D,$4F,$18,$CB,$E4,$30,$49,$E4,$D8,$E4,$70,$7C,$A6,$4E,$51,$C0
	.dc.b	$91,$C2,$34,$70,$8C,$BC,$23,$51,$88,$D4,$60,$8E,$C8,$68,$B5,$F2
	.dc.b	$49,$5F,$02,$BA,$46,$2B,$D4,$DA,$F2,$B6,$77,$5C,$6B,$C3,$5D,$C9
	.dc.b	$5B,$8C,$B6,$3C,$73,$C5,$B3,$E0,$8C,$47,$26,$CF,$18,$C4,$EC,$8F
	.dc.b	$9A,$76,$48,$CD,$7A,$1A,$79,$23,$53,$C6,$32,$F9,$0C,$12,$79,$36
	.dc.b	$39,$1C,$1F,$29,$93,$94,$70,$24,$70,$8D,$1C,$23,$2F,$08,$D4,$62
	.dc.b	$3F,$86,$9F,$CB,$CA,$7F,$0B,$31,$5F,$CE,$1B,$5F,$E9,$A8,$C4,$6C
	.dc.b	$87,$F1,$F3,$75,$FD,$3E,$37,$19,$3F,$9F,$7D,$FB,$D4,$FE,$6E,$DA
	.dc.b	$FF,$42,$B6,$7F,$CF,$2D,$C3,$FD,$0C,$D9,$E7,$5C,$88,$F2,$6D,$72
	.dc.b	$6D,$72,$6C,$F2,$38,$12,$78,$36,$78,$34,$E1,$1A,$24,$69,$B1,$A8
	.dc.b	$C1,$1C,$12,$3F,$B7,$57,$B5,$B5,$E6,$6C,$F8,$5A,$76,$2D,$75,$6C
	.dc.b	$75,$55,$D5,$B5,$D5,$B3,$D5,$A7,$52,$D7,$56,$C7,$5A,$E8,$4A,$E8
	.dc.b	$D8,$E8,$D9,$E8,$DA,$E8,$D9,$E9,$5C,$88,$F2,$6D,$72,$6D,$72,$6C
	.dc.b	$F2,$38,$12,$78,$36,$78,$34,$E1,$1A,$24,$69,$B1,$A8,$C1,$1C,$12
	.dc.b	$3E,$99,$0F,$92,$5A,$F8,$0D,$AF,$4B,$67,$C8,$D3,$C2,$5A,$F0,$36
	.dc.b	$3C,$0F,$62,$57,$66,$D7,$66,$CF,$66,$9D,$8B,$5D,$9B,$1D,$AB,$A1
	.dc.b	$2B,$A3,$63,$A3,$67,$A3,$6B,$A3,$67,$A5,$72,$23,$C9,$B5,$C9,$B5
	.dc.b	$C9,$B3,$C8,$E0,$49,$E0,$D9,$E0,$D3,$84,$68,$91,$A6,$C6,$A3,$04
	.dc.b	$70,$48,$FA,$84,$3E,$49,$6B,$E0,$36,$BD,$2D,$9F,$23,$4F,$09,$6B
	.dc.b	$C0,$D8,$F0,$3D,$89,$5D,$9B,$5F,$C9,$9F,$E8,$91,$FC,$C5,$B0,$9F
	.dc.b	$F4,$54,$E6,$7F,$90,$9A,$8C,$CE,$E3,$FA,$71,$FC,$28,$D1,$89,$E4
	.dc.b	$FF,$1A,$32,$7F,$55,$F9,$87,$8B,$FE,$7C,$6D,$FF,$A4,$F0,$ED,$49
	.dc.b	$CA,$8C,$C6,$1B,$5C,$C9,$5C,$CF,$F4,$94,$ED,$B1,$AA,$E9,$18,$38
	.dc.b	$3F,$EA,$59,$E0,$D9,$E1,$5E,$0A,$DC,$F9,$E7,$31,$CE,$3A,$9C,$5D
	.dc.b	$D7,$48,$E1,$18,$9E,$31,$B3,$89,$A2,$CF,$12,$47,$08,$D5,$79,$5D
	.dc.b	$C6,$67,$75,$DC,$93,$A6,$D7,$12,$3E,$B6,$C7,$95,$AF,$82,$34,$47
	.dc.b	$4D,$74,$60,$86,$09,$3E,$67,$C8,$49,$EF,$19,$6D,$7B,$49,$5E,$A3
	.dc.b	$C1,$3B,$6C,$6A,$BA,$46,$0E,$0F,$92,$78,$36,$78,$57,$82,$B7,$3E
	.dc.b	$89,$CC,$73,$8E,$A7,$17,$75,$D2,$38,$46,$27,$8C,$6C,$E2,$68,$B3
	.dc.b	$C4,$91,$C2,$35,$5E,$57,$71,$99,$DD,$77,$24,$E9,$B5,$C4,$8F,$AD
	.dc.b	$B1,$E5,$6B,$E0,$8D,$11,$D3,$5D,$18,$21,$82,$4F,$99,$F2,$12,$7B
	.dc.b	$C6,$5B,$5E,$D2,$57,$A8,$F0,$4E,$DB,$1A,$AE,$91,$83,$83,$E4,$9E
	.dc.b	$0D,$9E,$15,$E0,$AD,$CF,$A2,$73,$1C,$E3,$A9,$C5,$DD,$74,$8E,$11
	.dc.b	$89,$E3,$1B,$38,$9A,$2C,$F1,$24,$70,$8D,$57,$95,$DC,$66,$77,$5D
	.dc.b	$C9,$3A,$6D,$71,$23,$EB,$6C,$79,$5A,$FE,$54,$FE,$7D,$5F,$CE,$9F
	.dc.b	$E4,$46,$27,$FA,$4B,$FC,$6C,$C1,$FD,$32,$BF,$9F,$B8,$21,$C8,$87
	.dc.b	$F2,$53,$FD,$41,$F7,$CF,$97,$FE,$AD,$87,$F4,$A8,$C5,$FF,$36,$6D
	.dc.b	$7F,$2C,$6C,$EE,$B8,$D2,$A5,$6A,$31,$08,$C3,$A9,$E2,$D9,$F0,$46
	.dc.b	$23,$83,$67,$8C,$62,$76,$47,$CD,$3B,$24,$66,$BD,$0D,$3C,$B5,$D0
	.dc.b	$93,$C2,$B9,$CE,$DF,$31,$82,$57,$46,$C7,$49,$E5,$39,$24,$64,$E9
	.dc.b	$3E,$B3,$89,$DA,$36,$D8,$E0,$E8,$E1,$1A,$6C,$6A,$31,$5E,$C8,$EE
	.dc.b	$D9,$E4,$D9,$DD,$7A,$2B,$91,$2B,$C4,$43,$B9,$6B,$94,$6E,$32,$48
	.dc.b	$CB,$B9,$EC,$D9,$F2,$C6,$23,$83,$67,$8C,$62,$76,$47,$D7,$3B,$24
	.dc.b	$66,$BD,$AD,$3C,$D5,$D0,$93,$C2,$B9,$CE,$DF,$31,$82,$57,$46,$C7
	.dc.b	$49,$E5,$39,$24,$64,$E9,$3E,$B3,$89,$DA,$36,$D8,$E0,$E8,$E1,$1A
	.dc.b	$6C,$6A,$31,$5E,$C8,$EE,$D9,$E4,$D9,$DD,$7A,$2B,$91,$2B,$C4,$43
	.dc.b	$B9,$6B,$94,$6E,$32,$48,$CB,$B9,$EC,$D9,$F2,$C6,$23,$83,$67,$8C
	.dc.b	$62,$76,$47,$D7,$3B,$24,$66,$BD,$AD,$3C,$D5,$D0,$93,$C2,$B9,$CE
	.dc.b	$DF,$31,$82,$57,$46,$C7,$49,$E5,$39,$24,$64,$E9,$3E,$B3,$89,$DA
	.dc.b	$36,$D8,$E0,$F8,$4F,$E7,$73,$FD,$25,$FE,$2A,$62,$3F,$A4,$D7,$F3
	.dc.b	$A8,$C1,$FE,$94,$70,$FF,$0C,$32,$43,$04,$3F,$9F,$7D,$F3,$A7,$F9
	.dc.b	$B3,$6B,$FC,$FA,$D9,$FF,$39,$B7,$0F,$F4,$BB,$67,$FA,$2B,$67,$AD
	.dc.b	$74,$24,$F4,$6C,$74,$6D,$74,$6C,$F4,$6C,$F4,$8E,$44,$79,$36,$79
	.dc.b	$34,$E4,$58,$E4,$70,$24,$70,$6C,$70,$69,$C2,$34,$48,$D3,$63,$51
	.dc.b	$82,$38,$20,$AF,$90,$DA,$F7,$B6,$7D,$0D,$3C,$65,$AF,$03,$63,$BB
	.dc.b	$4E,$F5,$D8,$95,$D9,$B3,$D9,$B3,$D9,$B5,$D9,$B3,$D9,$B3,$DA,$BA
	.dc.b	$12,$7A,$36,$3A,$36,$BA,$36,$7A,$36,$7A,$47,$22,$3C,$9B,$3C,$9A
	.dc.b	$72,$2C,$72,$38,$12,$38,$36,$38,$34,$E1,$1A,$24,$69,$B1,$A8,$C1
	.dc.b	$1C,$10,$FA,$84,$AF,$92,$DA,$F8,$0D,$9F,$4B,$4F,$21,$6B,$C2,$D8
	.dc.b	$F0,$34,$F0,$57,$62,$57,$66,$CF,$66,$CF,$66,$D7,$66,$CF,$66,$CF
	.dc.b	$6A,$E8,$49,$E8,$D8,$E8,$DA,$E8,$D9,$E8,$D9,$E9,$1C,$88,$F2,$6C
	.dc.b	$F2,$69,$C8,$B1,$C8,$E0,$48,$E0,$D8,$E0,$D3,$84,$68,$91,$A6,$C6
	.dc.b	$A3,$04,$70,$43,$EA,$12,$BE,$4B,$FB,$D3,$FC,$DE,$53,$F8,$91,$87
	.dc.b	$F9,$A3,$6B,$FA,$43,$B3,$F8,$E9,$99,$FF,$48,$CF,$08,$E6,$43,$67
	.dc.b	$F9,$FB,$EE,$9E,$5B,$D9,$FF,$42,$62,$C3,$53,$9B,$4E,$5F,$F4,$55
	.dc.b	$F2,$BE,$37,$C1,$B3,$AB,$E5,$7C,$61,$7C,$5E,$B7,$C5,$A7,$1B,$E6
	.dc.b	$F0,$BE,$73,$BB,$E9,$7C,$DB,$3A,$AE,$47,$43,$95,$F4,$AE,$17,$CC
	.dc.b	$93,$C2,$F9,$D6,$EF,$A4,$F2,$3A,$5F,$27,$85,$F2,$75,$7C,$9D,$D7
	.dc.b	$2A,$E1,$1A,$AE,$11,$82,$4E,$C9,$7E,$32,$57,$46,$DF,$72,$5F,$4A
	.dc.b	$E3,$5B,$24,$E9,$F4,$90,$F2,$16,$FC,$77,$DA,$FA,$36,$75,$7D,$AF
	.dc.b	$A4,$72,$25,$F2,$7C,$37,$C9,$A7,$2B,$E8,$F0,$BE,$93,$BB,$EB,$7C
	.dc.b	$DB,$3A,$AE,$47,$43,$95,$F4,$AE,$17,$CC,$93,$C2,$F9,$D6,$EF,$A4
	.dc.b	$F2,$3A,$5F,$27,$85,$F2,$75,$7C,$9D,$D7,$2A,$E1,$1A,$AE,$11,$82
	.dc.b	$4E,$C9,$7E,$32,$57,$46,$DF,$72,$5F,$4A,$E3,$5B,$24,$E9,$F4,$90
	.dc.b	$F2,$16,$FC,$77,$DA,$FA,$36,$75,$7D,$AF,$A4,$72,$25,$F2,$7C,$37
	.dc.b	$C9,$A7,$2B,$E8,$F0,$BE,$93,$BB,$EB,$7C,$DB,$3A,$AE,$47,$43,$95
	.dc.b	$F4,$AE,$17,$CC,$93,$C2,$F9,$D6,$EF,$A4,$F2,$3A,$5F,$27,$85,$F2
	.dc.b	$75,$7C,$9D,$D7,$2A,$E1,$1A,$AE,$11,$82,$4E,$C9,$7E,$32,$57,$46
	.dc.b	$DF,$72,$5F,$4A,$E3,$5B,$24,$E9,$F4,$90,$F2,$16,$FC,$77,$DA,$FA
	.dc.b	$36,$75,$7D,$AF,$A4,$72,$25,$F2,$7C,$37,$C9,$A7,$2B,$E8,$F0,$BE
	.dc.b	$93,$BB,$EB,$7C,$DB,$3A,$AE,$47,$43,$95,$F4,$AE,$17,$CC,$93,$C2
	.dc.b	$F9,$D6,$EF,$A4,$F2,$3A,$5F,$27,$85,$F2,$75,$7C,$9D,$D7,$2A,$E1
	.dc.b	$1A,$AE,$11,$82,$4E,$C9,$7E,$32,$57,$46,$DF,$72,$5F,$4A,$E3,$5B
	.dc.b	$24,$E9,$F4,$90,$F2,$16,$FC,$77,$DA,$FA,$36,$75,$7D,$AF,$A4,$72
	.dc.b	$25,$F2,$7C,$37,$C9,$A7,$2B,$E8,$F0,$BE,$93,$BB,$EB,$7C,$DB,$3A
	.dc.b	$AE,$47,$43,$95,$F4,$AE,$17,$CC,$93,$C2,$F9,$D6,$EF,$A4,$F2,$3A
	.dc.b	$5F,$27,$85,$F2,$75,$7C,$9D,$D7,$2A,$E1,$1A,$AE,$11,$82,$4E,$C9
	.dc.b	$7E,$32,$57,$46,$DF,$72,$5F,$4A,$E3,$5B,$24,$E9,$F4,$90,$F2,$16
	.dc.b	$FC,$77,$DA,$FA,$36,$75,$7D,$AF,$A4,$72,$25,$F2,$7C,$37,$C9,$A7
	.dc.b	$2B,$E8,$F0,$BE,$93,$BB,$EB,$7C,$DB,$3A,$AE,$47,$43,$95,$F4,$AE
	.dc.b	$17,$CC,$93,$C2,$F9,$D6,$EF,$A4,$F2,$3A,$5F,$27,$85,$F2,$75,$7C
	.dc.b	$9D,$D7,$2A,$E1,$1A,$AE,$11,$82,$4E,$C9,$7E,$32,$57,$46,$DF,$72
	.dc.b	$5F,$4A,$E3,$5B,$24,$E9,$F4,$90,$F2,$16,$FC,$77,$DA,$FA,$36,$75
	.dc.b	$7D,$AF,$A4,$72,$25,$F2,$7C,$37,$C9,$A7,$2B,$E8,$F0,$BE,$93,$BB
	.dc.b	$EB,$7C,$DB,$3A,$AE,$47,$43,$95,$F4,$AE,$17,$CC,$93,$C2,$F9,$D6
	.dc.b	$EF,$A4,$F2,$3A,$5F,$27,$85,$F2,$75,$7C,$9D,$D7,$2A,$E1,$1A,$AE
	.dc.b	$11,$82,$4E,$C9,$7E,$32,$57,$FB,$30,$FF,$4A,$1F,$EC,$E9,$FF,$57
	.dc.b	$36,$7F,$D7,$8D,$9E,$0F,$03,$FD,$64,$64,$E9,$3C,$9C,$90,$FF,$56
	.dc.b	$9D,$FC,$C7,$37,$49,$46,$27,$9B,$5E,$13,$B8,$C4,$7F,$B5,$DA,$FF
	.dc.b	$B4,$1A,$FF,$B2,$9B,$3D,$A7,$91,$27,$BB,$71,$FF,$63,$D7,$8D,$B1
	.dc.b	$FE,$C3,$8C,$63,$E1,$8C,$12,$BF,$D7,$F3,$A7,$C9,$19,$AE,$71,$A8
	.dc.b	$C5,$73,$8C,$12,$76,$48,$D3,$63,$A9,$B8,$C9,$1C,$C6,$DF,$89,$3E
	.dc.b	$07,$33,$B7,$27,$33,$24,$33,$3C,$EF,$E6,$46,$CC,$CF,$38,$CC,$EE
	.dc.b	$FE,$25,$68,$E9,$3A,$AE,$13,$A8,$C4,$EA,$30,$4B,$F5,$47,$32,$4F
	.dc.b	$06,$CE,$DF,$34,$64,$91,$9B,$F3,$57,$49,$CB,$67,$57,$E6,$8C,$12
	.dc.b	$FA,$CE,$9F,$44,$66,$FA,$C6,$A3,$15,$CE,$30,$49,$D9,$23,$4D,$8E
	.dc.b	$A6,$E3,$24,$73,$1B,$7E,$1C,$F8,$1C,$CE,$DC,$9C,$CC,$90,$CC,$F3
	.dc.b	$BF,$8B,$1B,$33,$3C,$E3,$33,$BB,$F8,$95,$A3,$A4,$EA,$B8,$4E,$A3
	.dc.b	$13,$A8,$C1,$2F,$D5,$1C,$C9,$3C,$1B,$3B,$7C,$D1,$92,$46,$6F,$CD
	.dc.b	$5D,$27,$2D,$9D,$5F,$9A,$30,$4B,$EB,$3A,$7D,$11,$9B,$EB,$1A,$8C
	.dc.b	$57,$38,$C1,$27,$64,$8D,$36,$3A,$9B,$8C,$91,$CC,$6D,$F8,$73,$E0
	.dc.b	$73,$3B,$72,$73,$32,$43,$33,$CE,$FE,$2C,$6C,$CC,$F3,$8C,$CE,$EF
	.dc.b	$E2,$56,$8E,$93,$AA,$E1,$3A,$8C,$4E,$A3,$04,$BF,$54,$73,$24,$F0
	.dc.b	$6C,$ED,$F3,$46,$49,$19,$BF,$35,$74,$9C,$B6,$75,$7E,$68,$C1,$2F
	.dc.b	$AC,$E9,$F4,$46,$6F,$AC,$6A,$31,$5C,$E3,$04,$9D,$92,$34,$D8,$EA
	.dc.b	$6E,$32,$47,$31,$B7,$E1,$CF,$81,$CC,$ED,$C9,$CC,$C9,$0C,$CF,$3B
	.dc.b	$F8,$B1,$B3,$33,$CE,$33,$3B,$BF,$89,$5A,$3A,$4E,$AB,$84,$EA,$31
	.dc.b	$3A,$8C,$12,$FD,$51,$CC,$93,$C1,$B3,$B7,$CD,$19,$24,$66,$FC,$D5
	.dc.b	$D2,$72,$D9,$D5,$F9,$A3,$04,$BE,$B3,$A7,$D1,$19,$BE,$B1,$A8,$C5
	.dc.b	$73,$8C,$12,$76,$48,$D3,$63,$A9,$B8,$C9,$1C,$C6,$DF,$87,$3E,$07
	.dc.b	$33,$B7,$27,$33,$24,$33,$3C,$EF,$E2,$C6,$CC,$CF,$38,$CC,$EE,$FE
	.dc.b	$25,$68,$E9,$3A,$AE,$13,$A8,$C4,$EA,$30,$4B,$F5,$47,$32,$4F,$06
	.dc.b	$CE,$DF,$34,$64,$91,$9B,$F3,$57,$49,$CB,$67,$57,$E6,$8C,$12,$FA
	.dc.b	$CE,$9F,$44,$66,$FA,$C6,$A3,$15,$CE,$30,$49,$D9,$23,$4D,$8E,$A6
	.dc.b	$E3,$24,$73,$1B,$7E,$1C,$F8,$1C,$CE,$DC,$9C,$CC,$90,$CC,$F3,$BF
	.dc.b	$8B,$1B,$33,$3C,$E3,$33,$BB,$F8,$95,$A3,$A4,$EA,$B8,$4E,$A3,$13
	.dc.b	$A8,$C1,$2F,$D5,$1C,$C9,$3C,$1B,$3B,$7C,$D1,$92,$46,$6F,$CD,$5D
	.dc.b	$27,$2D,$9D,$5F,$9A,$30,$4B,$EB,$3A,$7D,$11,$9B,$EB,$1A,$8C,$57
	.dc.b	$38,$C1,$27,$64,$8D,$36,$3A,$9B,$8C,$91,$CC,$6D,$F8,$73,$E0,$73
	.dc.b	$3B,$72,$73,$32,$43,$FC,$5D,$F6,$CF,$5F,$FD,$8B,$08,$C3,$FE,$C8
	.dc.b	$6B,$FE,$D1,$6A,$21,$C6,$39,$C6,$A3,$2D,$78,$35,$E0,$77,$8E,$24
	.dc.b	$70,$77,$B5,$74,$9F,$0C,$62,$7A,$C6,$27,$51,$C5,$AA,$78,$3E,$23
	.dc.b	$11,$A8,$CD,$7C,$26,$D7,$5A,$EE,$75,$29,$C8,$B1,$C1,$D1,$DD,$C3
	.dc.b	$B3,$15,$E1,$23,$C9,$AE,$8C,$10,$C1,$2B,$D1,$1C,$1B,$1B,$6B,$BA
	.dc.b	$F6,$34,$F3,$C6,$89,$1A,$6C,$6A,$30,$47,$04,$3D,$D1,$C1,$B1,$B6
	.dc.b	$9B,$AF,$8A,$4A,$F5,$4E,$E3,$13,$EB,$8C,$4E,$A3,$93,$5E,$A4,$9E
	.dc.b	$2F,$BC,$C4,$6E,$33,$5E,$76,$D7,$6A,$F0,$1D,$8A,$72,$2C,$70,$74
	.dc.b	$77,$70,$EC,$C5,$78,$48,$F2,$6B,$A3,$04,$30,$4A,$F4,$47,$06,$C6
	.dc.b	$DA,$EE,$BD,$8D,$3C,$F1,$A2,$46,$9B,$1A,$8C,$11,$C1,$0F,$74,$70
	.dc.b	$6C,$6D,$A6,$EB,$E2,$92,$BD,$53,$B8,$C4,$FA,$E3,$13,$A8,$E4,$D7
	.dc.b	$A9,$27,$8B,$EF,$31,$1B,$8C,$D7,$9D,$B5,$DA,$BC,$07,$62,$9C,$8B
	.dc.b	$1C,$1D,$1D,$DC,$3B,$31,$5E,$12,$3C,$9A,$E8,$C1,$0C,$12,$BD,$11
	.dc.b	$C1,$B1,$B6,$BB,$AF,$63,$4F,$3C,$68,$91,$A6,$C6,$A3,$04,$70,$43
	.dc.b	$DD,$9B,$DD,$FE,$A2,$D8,$FE,$96,$7F,$4C,$70,$E0,$38,$76,$D7,$84
	.dc.b	$74,$23,$C4,$E2,$E1,$E9,$1C,$5F,$EB,$8D,$76,$7F,$38,$7C,$78,$7F
	.dc.b	$48,$6E,$1E,$09,$D5,$7F,$2A,$9C,$C7,$48,$C1,$1D,$90,$FE,$85,$1C
	.dc.b	$1F,$09,$88,$E0,$D3,$85,$78,$E3,$2D,$8C,$D7,$95,$E0,$48,$D3,$63
	.dc.b	$51,$82,$38,$24,$7A,$CE,$A5,$39,$16,$38,$B6,$38,$BE,$C8,$CB,$63
	.dc.b	$35,$F0,$49,$5E,$C8,$CB,$67,$A1,$0E,$45,$9E,$2F,$98,$C9,$1C,$BE
	.dc.b	$83,$11,$D5,$A7,$18,$DD,$7A,$1B,$5E,$49,$D5,$71,$9C,$C7,$8E,$30
	.dc.b	$47,$64,$3C,$51,$C1,$F0,$98,$8E,$0D,$38,$57,$8E,$32,$D8,$CD,$79
	.dc.b	$5E,$04,$8D,$36,$35,$18,$23,$82,$47,$AC,$EA,$53,$91,$63,$8B,$63
	.dc.b	$8B,$EC,$8C,$B6,$33,$5F,$04,$95,$EC,$8C,$B6,$7A,$10,$E4,$59,$E2
	.dc.b	$F9,$8C,$91,$CB,$E8,$31,$1D,$5A,$71,$8D,$D7,$A1,$B5,$E4,$9D,$57
	.dc.b	$19,$CC,$78,$E3,$04,$76,$43,$C5,$1C,$1F,$09,$88,$E0,$D3,$85,$78
	.dc.b	$E3,$2D,$8C,$D7,$95,$E0,$48,$D3,$63,$51,$82,$38,$24,$7A,$CE,$A5
	.dc.b	$39,$16,$38,$B6,$38,$BE,$C8,$FC,$0F,$DA,$BD,$1F,$F4,$9B,$67,$F9
	.dc.b	$C3,$5F,$E8,$4D,$7F,$A3,$B5,$FF,$49,$B5,$FF,$4B,$35,$34,$91,$E0
	.dc.b	$24,$78,$1F,$F3,$1C,$78,$67,$FD,$0E,$DA,$EF,$3A,$25,$7F,$9F,$A7
	.dc.b	$4F,$8A,$32,$FF,$9F,$E3,$2F,$92,$33,$5D,$9B,$5D,$A7,$A4,$61,$E8
	.dc.b	$60,$87,$57,$4F,$53,$33,$D5,$B3,$D5,$E0,$60,$EA,$60,$91,$C5,$B1
	.dc.b	$C6,$30,$47,$04,$8E,$71,$96,$C6,$5F,$31,$D0,$C9,$0C,$C7,$16,$C7
	.dc.b	$1A,$F9,$64,$AF,$82,$D9,$F6,$1C,$0E,$8E,$5F,$01,$92,$39,$AF,$03
	.dc.b	$6B,$C1,$5D,$89,$5D,$A7,$4F,$68,$CB,$DA,$32,$F6,$8C,$D7,$66,$D7
	.dc.b	$69,$E9,$18,$7A,$18,$21,$D5,$D3,$D4,$CC,$F5,$6C,$F5,$78,$18,$3A
	.dc.b	$98,$24,$71,$6C,$71,$8C,$11,$C1,$23,$9C,$65,$B1,$97,$CC,$74,$32
	.dc.b	$43,$31,$C5,$B1,$C6,$BE,$59,$2B,$E0,$B6,$7D,$87,$03,$A3,$97,$C0
	.dc.b	$64,$8E,$6B,$C0,$DA,$F0,$57,$62,$57,$69,$D3,$DA,$32,$F6,$8C,$BD
	.dc.b	$A3,$35,$D9,$B5,$DA,$7A,$46,$1E,$86,$08,$75,$74,$F5,$33,$3D,$5B
	.dc.b	$3D,$5E,$06,$0E,$A6,$09,$1C,$5B,$1C,$63,$04,$70,$48,$E7,$19,$6C
	.dc.b	$7E,$07,$ED,$5F,$13,$FE,$77,$94,$61,$FF,$3D,$35,$FF,$42,$35,$FF
	.dc.b	$44,$B5,$11,$C3,$63,$B4,$62,$3B,$C6,$08,$F8,$1A,$A9,$5D,$A3,$89
	.dc.b	$E0,$8C,$BD,$CC,$46,$EB,$D5,$3A,$9E,$27,$59,$D5,$75,$9D,$57,$09
	.dc.b	$CD,$74,$3B,$1C,$4B,$3C,$2B,$94,$F0,$9D,$1E,$09,$D3,$CC,$C4,$6E
	.dc.b	$33,$1C,$2B,$C4,$F6,$9D,$11,$C1,$0E,$65,$8E,$2E,$08,$E0,$95,$E8
	.dc.b	$8E,$0D,$76,$D8,$DC,$7A,$5A,$F9,$E3,$44,$74,$D7,$46,$08,$60,$95
	.dc.b	$F0,$89,$5E,$F8,$F1,$9C,$23,$2F,$A0,$C4,$6E,$BC,$B3,$A9,$E2,$76
	.dc.b	$9D,$57,$96,$75,$5C,$27,$35,$D0,$EC,$71,$2C,$F0,$AE,$53,$C2,$74
	.dc.b	$78,$27,$4F,$33,$11,$B8,$CC,$70,$AF,$13,$DA,$74,$47,$04,$39,$96
	.dc.b	$38,$B8,$23,$82,$57,$A2,$38,$35,$DB,$63,$71,$E9,$6B,$E7,$8D,$11
	.dc.b	$D3,$5D,$18,$21,$82,$57,$C2,$25,$7B,$E3,$C6,$70,$8C,$BE,$83,$11
	.dc.b	$BA,$F2,$CE,$A7,$89,$DA,$75,$5E,$59,$D5,$70,$9C,$D7,$43,$B1,$C4
	.dc.b	$B3,$C2,$B9,$4F,$09,$D1,$E0,$9D,$3C,$CC,$46,$E3,$31,$C2,$BC,$4F
	.dc.b	$69,$D1,$1C,$10,$E6,$58,$E2,$E0,$8E,$09,$5E,$88,$E0,$D7,$6D,$8D
	.dc.b	$C7,$A5,$AF,$9E,$34,$47,$4D,$7E,$1F,$C7,$BE,$3F,$FA,$93,$63,$F9
	.dc.b	$F1,$FC,$FD,$C3,$80,$E1,$DB,$5E,$11,$D0,$8F,$13,$8B,$87,$A4,$71
	.dc.b	$7F,$AD,$B5,$D9,$FC,$DD,$F1,$E1,$FD,$2D,$B8,$78,$27,$55,$FC,$A6
	.dc.b	$73,$1D,$23,$04,$76,$43,$FA,$3C,$70,$7C,$26,$23,$83,$4E,$15,$E3
	.dc.b	$8C,$B6,$33,$5E,$57,$81,$23,$4D,$8D,$46,$08,$E0,$91,$EB,$3A,$94
	.dc.b	$E4,$58,$E2,$D8,$E2,$FB,$23,$2D,$8C,$D7,$C1,$25,$7B,$23,$2D,$9E
	.dc.b	$84,$39,$16,$78,$BE,$63,$24,$72,$FA,$0C,$47,$56,$9C,$63,$75,$E8
	.dc.b	$6D,$79,$27,$55,$C6,$73,$1E,$38,$C1,$1D,$90,$F1,$47,$07,$C2,$62
	.dc.b	$38,$34,$E1,$5E,$38,$CB,$63,$35,$E5,$78,$12,$34,$D8,$D4,$60,$8E
	.dc.b	$09,$1E,$B3,$A9,$4E,$45,$8E,$2D,$8E,$2F,$B2,$32,$D8,$CD,$7C,$12
	.dc.b	$57,$B2,$32,$D9,$E8,$43,$91,$67,$8B,$E6,$32,$47,$2F,$A0,$C4,$75
	.dc.b	$69,$C6,$37,$5E,$86,$D7,$92,$75,$5C,$67,$31,$E3,$8C,$11,$D9,$0F
	.dc.b	$14,$70,$7C,$26,$23,$83,$4E,$15,$E3,$8C,$B6,$33,$5E,$57,$81,$23
	.dc.b	$4D,$8D,$46,$08,$E0,$91,$EB,$3A,$94,$E4,$58,$E2,$D8,$E2,$FB,$23
	.dc.b	$F0,$3F,$68,$F5,$3F,$D2,$8D,$9F,$E7,$0D,$7F,$A1,$35,$FE,$8E,$D7
	.dc.b	$FD,$0C,$D7,$FD,$2A,$D4,$D2,$47,$80,$91,$E0,$7F,$CC,$51,$E1,$9F
	.dc.b	$F4,$53,$6B,$BC,$E8,$95,$FE,$7D,$9D,$3E,$28,$CB,$FE,$84,$8C,$BE
	.dc.b	$48,$CD,$76,$6D,$76,$9E,$91,$87,$A1,$82,$1D,$5D,$3D,$4C,$CF,$56
	.dc.b	$CF,$57,$81,$83,$A9,$82,$47,$16,$C7,$18,$C1,$1C,$12,$39,$C6,$5B
	.dc.b	$19,$7C,$C7,$43,$24,$33,$1C,$5B,$1C,$6B,$E5,$92,$BE,$0B,$67,$D8
	.dc.b	$70,$3A,$39,$7C,$06,$48,$E6,$BC,$0D,$AF,$05,$76,$25,$76,$9D,$3D
	.dc.b	$A3,$2F,$68,$CB,$DA,$33,$5D,$9B,$5D,$A7,$A4,$61,$E8,$60,$87,$57
	.dc.b	$4F,$53,$33,$D5,$B3,$D5,$E0,$60,$EA,$60,$91,$C5,$B1,$C6,$30,$47
	.dc.b	$04,$8E,$71,$96,$C6,$5F,$31,$D0,$C9,$0C,$C7,$16,$C7,$1A,$F9,$64
	.dc.b	$AF,$82,$D9,$F6,$1C,$0E,$8E,$5F,$01,$92,$39,$AF,$03,$6B,$C1,$5D
	.dc.b	$89,$5D,$A7,$4F,$68,$CB,$DA,$32,$F6,$8C,$D7,$66,$D7,$69,$E9,$18
	.dc.b	$7A,$18,$21,$D5,$D3,$D4,$CC,$F5,$6C,$F5,$78,$18,$3A,$98,$24,$71
	.dc.b	$6C,$71,$8C,$11,$C1,$23,$9C,$65,$B1,$F8,$1F,$B4,$7C,$E7,$F9,$DE
	.dc.b	$51,$87,$FC,$F4,$D7,$FD,$08,$D7,$FD,$12,$D4,$47,$0D,$8E,$D1,$88
	.dc.b	$EF,$18,$23,$E0,$6A,$A5,$76,$8E,$27,$82,$32,$F7,$31,$1B,$AF,$54
	.dc.b	$EA,$78,$9D,$67,$55,$D6,$75,$5C,$27,$35,$D0,$EC,$71,$2C,$F0,$AE
	.dc.b	$53,$C2,$74,$78,$27,$4F,$33,$11,$B8,$CC,$70,$AF,$13,$DA,$74,$47
	.dc.b	$04,$39,$96,$38,$B8,$23,$82,$57,$A2,$38,$35,$DB,$63,$71,$E9,$6B
	.dc.b	$E7,$8D,$11,$D3,$5D,$18,$21,$82,$57,$C2,$25,$7B,$E3,$C6,$70,$8C
	.dc.b	$BE,$83,$11,$BA,$F2,$CE,$A7,$89,$DA,$75,$5E,$59,$D5,$70,$9C,$D7
	.dc.b	$43,$B1,$C4,$B3,$C2,$B9,$4F,$09,$D1,$E0,$9D,$3C,$CC,$46,$E3,$31
	.dc.b	$C2,$BC,$4F,$69,$D1,$1C,$10,$E6,$58,$E2,$E0,$8E,$09,$5E,$88,$E0
	.dc.b	$D7,$6D,$8D,$C7,$A5,$AF,$9E,$34,$47,$4D,$74,$60,$86,$09,$5F,$08
	.dc.b	$95,$EF,$8F,$19,$C2,$32,$FA,$0C,$46,$EB,$CB,$3A,$9E,$27,$69,$D5
	.dc.b	$79,$67,$55,$C2,$73,$5D,$0E,$C7,$12,$CF,$0A,$E5,$3C,$27,$47,$82
	.dc.b	$74,$F3,$31,$1B,$8C,$C7,$0A,$F1,$3D,$A7,$44,$70,$43,$99,$63,$8B
	.dc.b	$82,$38,$25,$7A,$23,$83,$5D,$B6,$37,$1E,$96,$BE,$78,$D1,$1D,$35
	.dc.b	$F8,$7E,$5B,$E1,$FF,$A9,$36,$3F,$9F,$1F,$CF,$DC,$38,$0E,$1D,$B5
	.dc.b	$E1,$1D,$08,$F1,$38,$B8,$7A,$47,$17,$FA,$DB,$5D,$9F,$CD,$DF,$1E
	.dc.b	$1F,$D2,$DB,$87,$82,$75,$5F,$CA,$67,$31,$D2,$30,$47,$64,$3F,$A3
	.dc.b	$C7,$07,$C2,$62,$38,$34,$E1,$5E,$38,$CB,$63,$35,$E5,$78,$12,$34
	.dc.b	$D8,$D4,$60,$8E,$09,$1E,$B3,$A9,$4E,$45,$8E,$2D,$8E,$2F,$B2,$32
	.dc.b	$D8,$CD,$7C,$12,$57,$B2,$32,$D9,$E8,$43,$91,$67,$8B,$E6,$32,$47
	.dc.b	$2F,$A0,$C4,$75,$69,$C6,$37,$5E,$86,$D7,$92,$75,$5C,$67,$31,$E3
	.dc.b	$8C,$11,$D9,$0F,$14,$70,$7C,$26,$23,$83,$4E,$15,$E3,$8C,$B6,$33
	.dc.b	$5E,$57,$81,$23,$4D,$8D,$46,$08,$E0,$91,$EB,$3A,$94,$E4,$58,$E2
	.dc.b	$D8,$E2,$FB,$23,$2D,$8C,$D7,$C1,$25,$7B,$23,$2D,$9E,$84,$39,$16
	.dc.b	$78,$BE,$63,$24,$72,$FA,$0C,$47,$56,$9C,$63,$75,$E8,$6D,$79,$27
	.dc.b	$55,$C6,$73,$1E,$38,$C1,$1D,$90,$F1,$47,$07,$C2,$62,$38,$34,$E1
	.dc.b	$5E,$38,$CB,$63,$35,$E5,$78,$12,$34,$D8,$D4,$60,$8E,$09,$1E,$B3
	.dc.b	$A9,$4E,$45,$8E,$2D,$8E,$2F,$B2,$3F,$03,$F6,$7F,$E9,$38,$34,$8F
	.dc.b	$FA,$5D,$AF,$F3,$B6,$BF,$D1,$5A,$FF,$4A,$6B,$FE,$89,$7F,$A6,$C7
	.dc.b	$FA,$0D,$B3,$99,$FF,$30,$46,$1F,$F4,$EB,$5F,$29,$A2,$1F,$E7,$B7
	.dc.b	$44,$72,$7F,$A0,$9C,$3B,$33,$3E,$36,$CF,$8D,$D1,$0E,$46,$8F,$29
	.dc.b	$4F,$29,$B8,$E8,$D8,$E8,$F2,$30,$72,$29,$C8,$D1,$23,$9B,$63,$9C
	.dc.b	$60,$8E,$09,$1D,$4E,$05,$8D,$B6,$36,$F8,$63,$2D,$8C,$CF,$CA,$8E
	.dc.b	$66,$4E,$44,$30,$47,$84,$F4,$6C,$F4,$9F,$95,$18,$79,$35,$E4,$68
	.dc.b	$87,$C8,$74,$47,$27,$99,$C3,$B3,$33,$E3,$6C,$F8,$DD,$10,$E4,$68
	.dc.b	$F2,$94,$F2,$9B,$8E,$8D,$8E,$8F,$23,$07,$22,$9C,$8D,$12,$39,$B6
	.dc.b	$39,$C6,$08,$E0,$91,$D4,$E0,$58,$DB,$63,$6F,$86,$32,$D8,$CC,$FC
	.dc.b	$A8,$E6,$64,$E4,$43,$04,$78,$4F,$46,$CF,$49,$F9,$51,$87,$93,$5E
	.dc.b	$46,$88,$7C,$87,$44,$72,$79,$9C,$3B,$33,$3E,$36,$CF,$8D,$D1,$0E
	.dc.b	$46,$8F,$29,$4F,$29,$B8,$E8,$D8,$E8,$F2,$30,$72,$29,$C8,$D1,$23
	.dc.b	$9B,$63,$9C,$60,$8E,$0F,$DD,$FD,$9B,$DB,$FF,$39,$52,$30,$FF,$9D
	.dc.b	$5A,$FF,$9F,$DA,$FF,$A1,$1A,$88,$E1,$B1,$DA,$31,$1D,$E3,$04,$7C
	.dc.b	$0D,$54,$AE,$D1,$C4,$F0,$46,$5E,$E6,$23,$75,$EA,$9D,$4F,$13,$AC
	.dc.b	$EA,$BA,$CE,$AB,$84,$E6,$BA,$1D,$8E,$25,$9E,$15,$CA,$78,$4E,$8F
	.dc.b	$9D,$ED,$FC,$32,$3F,$3D,$AF,$CF,$6B,$F3,$DA,$FB,$1A,$FD,$07,$E7
	.dc.b	$7B,$12,$3E,$D3,$5F,$B4,$D7,$ED,$35,$F8,$ED,$7E,$DB,$5F,$C3,$74
	.dc.b	$7D,$B2,$3F,$87,$6B,$F8,$76,$BF,$87,$6B,$F5,$1A,$FE,$21,$FF,$62
	.dc.b	$7A,$4F,$9C,$7B,$FF,$2E,$93,$98,$C3,$63,$EA,$B5,$FA,$AD,$CF,$F0
	.dc.b	$CD,$CB,$F3,$0D,$9E,$B7,$E0,$79,$B5,$E6,$DA,$E6,$DA,$E6,$D9,$FD
	.dc.b	$1F,$C2,$7E,$84,$8F,$EB,$5A,$FE,$B5,$AF,$EB,$5A,$FE,$69,$AF,$43
	.dc.b	$91,$F9,$43,$F7,$4F,$C7,$69,$9D,$51,$96,$C6,$63,$04,$70,$43,$F3
	.dc.b	$51,$C1,$B1,$B6,$9B,$BF,$D5,$46,$5B,$19,$BF,$63,$4F,$61,$C8,$91
	.dc.b	$C1,$B1,$C1,$A7,$08,$D1,$23,$4D,$8D,$46,$08,$E0,$87,$C8,$3A,$96
	.dc.b	$39,$36,$38,$B4,$E3,$5F,$16,$76,$D9,$DB,$67,$75,$F1,$67,$2D,$8C
	.dc.b	$9F,$14,$EC,$49,$E4,$D8,$E0,$D3,$84,$68,$91,$A6,$C6,$A3,$04,$70
	.dc.b	$43,$E1,$C7,$06,$C6,$DA,$6E,$BE,$0C,$65,$B1,$9A,$F7,$B4,$F7,$9C
	.dc.b	$89,$1C,$1B,$1C,$1A,$70,$8D,$12,$34,$D8,$FE,$AD,$DE,$F8,$4F,$EC
	.dc.b	$2D,$8F,$E9,$A7,$F4,$E7,$0E,$03,$87,$6D,$78,$47,$42,$3C,$4E,$2E
	.dc.b	$1E,$91,$C5,$F2,$C6,$67,$FA,$BC,$F8,$9B,$1F,$BD,$8C,$57,$82,$BF
	.dc.b	$81,$46,$1E,$66,$27,$81,$1C,$10,$FE,$AD,$1C,$5F,$39,$9A,$F0,$B6
	.dc.b	$BC,$0F,$0A,$E8,$D3,$47,$88,$C1,$27,$9B,$67,$9C,$60,$8E,$09,$1C
	.dc.b	$E3,$2D,$8C,$BE,$83,$A1,$92,$19,$8E,$2D,$8E,$35,$F0,$49,$5E,$CA
	.dc.b	$DC,$62,$BD,$A7,$13,$AB,$97,$C2,$64,$8E,$63,$9C,$F5,$AF,$0B,$6B
	.dc.b	$9B,$5E,$2D,$AE,$53,$C2,$7C,$71,$87,$89,$89,$E0,$47,$04,$3C,$F1
	.dc.b	$C5,$EE,$66,$BC,$ED,$AF,$03,$C2,$BA,$34,$D1,$E2,$30,$49,$E6,$D9
	.dc.b	$E7,$18,$23,$83,$EA,$FC,$27,$EE,$09,$3F,$AD,$B4,$62,$3F,$5A,$D7
	.dc.b	$F6,$ED,$7F,$7C,$D7,$F8,$2B,$C6,$BF,$62,$FE,$BC,$91,$FA,$E6,$BF
	.dc.b	$A8,$6C,$70,$8D,$10,$D1,$FE,$16,$F7,$BF,$A1,$68,$69,$32,$FF,$0E
	.dc.b	$B6,$78,$46,$08,$E0,$87,$F8,$82,$3F,$A2,$E0,$F8,$3F,$E3,$E4,$9F
	.dc.b	$90,$D7,$F8,$9B,$5F,$E3,$AD,$7F,$92,$35,$FE,$56,$F1,$C3,$F8,$BB
	.dc.b	$FC,$54,$91,$FC,$51,$AF,$F0,$E6,$C7,$08,$D1,$0D,$15,$D3,$82,$38
	.dc.b	$26,$5E,$68,$E0,$D7,$6D,$8D,$C7,$A8,$D9,$4D,$95,$DB,$FD,$0A,$7D
	.dc.b	$33,$CC,$91,$B6,$BB,$3A,$94,$EA,$59,$EB,$5F,$5E,$36,$D7,$6D,$8D
	.dc.b	$D7,$8A,$7C,$04,$3A,$57,$8A,$B8,$36,$78,$4E,$8E,$13,$A6,$C6,$A3
	.dc.b	$11,$B2,$38,$24,$6C,$91,$EA,$6B,$EA,$6D,$7A,$A3,$44,$34,$57,$4E
	.dc.b	$08,$E0,$95,$ED,$8E,$0D,$76,$D8,$DC,$7B,$8D,$94,$D9,$5D,$BF,$02
	.dc.b	$7D,$33,$CC,$FF,$6A,$FD,$9F,$FA,$1E,$7F,$C1,$6D,$73,$BF,$FB,$C2
	.dc.b	$31,$1F,$EE,$96,$A6,$92,$38,$12,$3F,$C1,$91,$96,$C7,$FA,$06,$3F
	.dc.b	$D2,$51,$CE,$30,$47,$FD,$11,$1A,$24,$66,$3F,$DB,$11,$87,$71,$9B
	.dc.b	$FF,$75,$CE,$67,$73,$9A,$F7,$C7,$02,$1A,$2D,$72,$9E,$11,$F0,$DA
	.dc.b	$FA,$4D,$CF,$47,$91,$27,$93,$A3,$91,$4E,$46,$E3,$94,$70,$8C,$1C
	.dc.b	$0A,$70,$74,$48,$E2,$48,$E1,$19,$6C,$65,$E4,$DA,$F9,$13,$B9,$C9
	.dc.b	$C0,$86,$09,$1C,$27,$95,$7C,$68,$C4,$F0,$6B,$C2,$34,$47,$E2,$46
	.dc.b	$89,$19,$8F,$83,$18,$77,$19,$AF,$44,$E6,$77,$39,$AF,$2C,$70,$21
	.dc.b	$A2,$D7,$29,$E1,$1E,$96,$BE,$93,$73,$D1,$E4,$49,$E4,$FB,$3D,$9F
	.dc.b	$D7,$11,$FD,$3B,$5F,$F3,$E5,$A3,$11,$FB,$86,$BF,$BC,$6B,$FB,$77
	.dc.b	$94,$FE,$89,$FD,$81,$23,$F5,$AD,$7F,$56,$D8,$E1,$1A,$21,$A3,$FD
	.dc.b	$F9,$F6,$6F,$63,$FD,$15,$08,$C3,$FE,$8D,$6B,$FE,$97,$6B,$FE,$9C
	.dc.b	$6A,$23,$86,$C7,$68,$C4,$77,$8C,$11,$F0,$3B,$F0,$A4,$78,$DA,$F8
	.dc.b	$DA,$F8,$DA,$F8,$CF,$63,$87,$CA,$D7,$E0,$78,$BE,$52,$3F,$05,$AF
	.dc.b	$C1,$6B,$F0,$5A,$F7,$A4,$62,$3E,$1B,$F0,$75,$91,$F3,$5A,$FC,$D6
	.dc.b	$BF,$35,$AF,$56,$BF,$39,$AF,$D9,$F3,$FE,$71,$1F,$B4,$D7,$ED,$35
	.dc.b	$FB,$4D,$7D,$4D,$7E,$DB,$F6,$7A,$64,$7E,$15,$AF,$E1,$5A,$FE,$15
	.dc.b	$AF,$C5,$6B,$F8,$66,$F1,$7F,$95,$FA,$FF,$CA,$A4,$9F,$F2,$E3,$5F
	.dc.b	$F2,$93,$74,$6F,$F9,$1A,$E5,$5C,$08,$70,$2C,$F0,$7F,$2F,$D5,$F9
	.dc.b	$52,$3F,$A1,$6B,$FA,$16,$BF,$A1,$6B,$F8,$16,$BD,$0E,$47,$EC,$1F
	.dc.b	$CA,$47,$86,$3F,$61,$18,$AE,$51,$B2,$09,$E1,$A2,$B8,$4F,$73,$55
	.dc.b	$C9,$A6,$6B,$8C,$F0,$8F,$09,$DA,$32,$D9,$E5,$18,$8E,$67,$02,$D7
	.dc.b	$8D,$E0,$77,$9E,$24,$70,$43,$94,$F5,$24,$6E,$32,$D9,$ED,$1E,$0A
	.dc.b	$F5,$46,$2B,$8C,$F5,$8C,$57,$1A,$D9,$27,$CE,$76,$38,$39,$7B,$1A
	.dc.b	$9E,$6E,$8C,$46,$EB,$C3,$1C,$C9,$3C,$1B,$3B,$AF,$24,$72,$21,$E0
	.dc.b	$25,$74,$AE,$44,$AE,$53,$E3,$37,$5D,$1A,$66,$B9,$4F,$08,$F1,$9E
	.dc.b	$08,$CB,$67,$94,$62,$39,$9C,$0B,$5E,$37,$81,$DE,$78,$91,$C1,$0E
	.dc.b	$53,$D4,$91,$B8,$CB,$67,$B4,$78,$2B,$D5,$0F,$83,$FE,$9C,$49,$FE
	.dc.b	$B4,$D7,$FA,$D3,$5F,$EB,$4D,$7F,$9D,$35,$FE,$B6,$F1,$BF,$D8,$BF
	.dc.b	$90,$24,$7F,$6C,$6B,$FA,$46,$C7,$08,$D1,$0D,$15,$D3,$82,$38,$25
	.dc.b	$F9,$A3,$83,$5D,$B6,$37,$1E,$A3,$65,$36,$57,$6F,$F8,$0E,$7D,$33
	.dc.b	$CC,$91,$B6,$BB,$3A,$94,$EA,$59,$EB,$5F,$5E,$36,$D7,$6D,$8D,$D7
	.dc.b	$8A,$7C,$04,$3A,$57,$8A,$B8,$36,$78,$4E,$8E,$13,$A6,$C6,$A3,$11
	.dc.b	$B2,$38,$24,$6C,$91,$EA,$6B,$EA,$6D,$7A,$A3,$44,$34,$57,$4E,$08
	.dc.b	$E0,$95,$ED,$8E,$0D,$76,$D8,$DC,$7B,$8D,$94,$D9,$5D,$BF,$02,$7D
	.dc.b	$33,$CC,$FF,$16,$7D,$9B,$E1,$3F,$C9,$2D,$AF,$F1,$81,$A8,$C9,$1C
	.dc.b	$9F,$E4,$78,$E0,$E1,$AF,$26,$BC,$CE,$21,$CB,$CE,$3B,$3C,$DF,$EB
	.dc.b	$D1,$DA,$FF,$B8,$36,$BA,$57,$F5,$C8,$C3,$CC,$FF,$2E,$4F,$0A,$E6
	.dc.b	$48,$CD,$72,$9D,$46,$0E,$47,$F7,$2C,$9C,$CF,$97,$FE,$F3,$18,$6C
	.dc.b	$66,$79,$4E,$5B,$19,$7A,$36,$BF,$C3,$93,$B9,$CD,$74,$23,$C9,$AF
	.dc.b	$C1,$F8,$4F,$B6,$47,$E7,$E4,$8C,$47,$D7,$6B,$F5,$DA,$FE,$19,$AF
	.dc.b	$DF,$78,$DF,$D2,$7E,$A9,$23,$E4,$1F,$75,$C3,$E5,$33,$3E,$36,$C6
	.dc.b	$A7,$A4,$60,$87,$12,$3A,$6E,$3E,$77,$71,$92,$46,$63,$67,$AC,$DB
	.dc.b	$98,$DC,$64,$D9,$EB,$36,$65,$DB,$93,$75,$ED,$6D,$7B,$6B,$B1,$23
	.dc.b	$93,$5E,$55,$D2,$B9,$36,$B9,$4F,$02,$1C,$2B,$95,$70,$6C,$F0,$9D
	.dc.b	$1C,$27,$4D,$8D,$46,$23,$64,$70,$48,$D9,$23,$E2,$9B,$70,$FC,$23
	.dc.b	$31,$C5,$B1,$AA,$F7,$46,$08,$71,$23,$A6,$D7,$B9,$DC,$64,$91,$98
	.dc.b	$D9,$EF,$36,$E6,$37,$19,$36,$7A,$CD,$99,$7F,$03,$F0,$83,$FD,$E6
	.dc.b	$58,$FE,$3E,$D7,$F9,$4B,$5F,$E5,$ED,$7F,$DC,$0D,$7F,$DE,$8D,$7B
	.dc.b	$3D,$88,$F6,$6B,$D9,$FF,$43,$1E,$18,$FF,$4C,$36,$7B,$BA,$24,$FE
	.dc.b	$E9,$D3,$E2,$72,$FF,$A2,$9C,$BE,$43,$83,$A9,$EE,$D9,$EF,$3A,$21
	.dc.b	$CE,$34,$F7,$33,$3C,$9B,$3C,$9E,$06,$0E,$06,$09,$1C,$5B,$1C,$63
	.dc.b	$04,$70,$48,$E7,$19,$6C,$65,$F8,$07,$43,$24,$33,$1C,$5B,$1C,$67
	.dc.b	$E4,$90,$F8,$31,$A3,$93,$97,$DC,$64,$8E,$5F,$11,$87,$99,$99,$F1
	.dc.b	$36,$7C,$4E,$89,$3D,$9D,$3E,$27,$2F,$67,$2F,$73,$83,$A9,$EE,$D9
	.dc.b	$EF,$3A,$21,$CE,$34,$F7,$33,$3C,$9B,$3C,$9E,$06,$0E,$06,$09,$1C
	.dc.b	$5B,$1F,$47,$E0,$FF,$6E,$48,$FD,$BB,$5F,$DB,$B5,$FD,$BB,$5F,$DB
	.dc.b	$B5,$FD,$BB,$C6,$7F,$5E,$FE,$C8,$91,$FA,$D6,$BF,$AB,$6C,$70,$8D
	.dc.b	$10,$D1,$FB,$5F,$4F,$F8,$79,$1F,$E1,$ED,$7F,$87,$B5,$FE,$1E,$D7
	.dc.b	$F8,$7B,$5F,$E1,$EF,$F1,$AF,$82,$91,$FC,$69,$AF,$F1,$A6,$BF,$C6
	.dc.b	$9A,$FF,$1A,$6B,$FC,$69,$BA,$7F,$8A,$BE,$5B,$FC,$4E,$49,$FF,$12
	.dc.b	$B5,$FF,$12,$B6,$FF,$C4,$BB,$23,$19,$29,$7D,$5E,$64,$39,$96,$F9
	.dc.b	$D7,$28,$E3,$7D,$5D,$5F,$5B,$E9,$1C,$89,$7C,$9D,$9C,$AF,$C8,$A1
	.dc.b	$58,$FB,$37,$57,$F9,$92,$91,$87,$FC,$CE,$D7,$FC,$E2,$D7,$FC,$EE
	.dc.b	$D7,$A4,$73,$8C,$47,$C5,$6B,$DA,$34,$47,$BB,$C7,$AA,$47,$89,$AF
	.dc.b	$89,$AF,$89,$AF,$B5,$AF,$85,$AF,$BB,$AB,$CC,$47,$DE,$D7,$DE,$D7
	.dc.b	$DE,$D7,$E5,$35,$F7,$3F,$07,$AA,$47,$CA,$6B,$F2,$9A,$FC,$A6,$BF
	.dc.b	$55,$AF,$C9,$6A,$FD,$55,$F5,$3F,$7B,$3F,$BB,$3F,$93,$9C,$BF,$E6
	.dc.b	$46,$BF,$83,$CF,$F0,$B1,$89,$E3,$3D,$6B,$A5,$69,$B1,$D5,$AE,$CE
	.dc.b	$CE,$2B,$A4,$F1,$9C,$9F,$E7,$6C,$DD,$7F,$87,$24,$7E,$41,$AF,$E4
	.dc.b	$1A,$FE,$41,$AF,$E6,$DA,$FE,$3D,$E3,$87,$B2,$7B,$CF,$86,$7D,$31
	.dc.b	$88,$F8,$87,$13,$44,$32,$7E,$9A,$38,$91,$D4,$F7,$8C,$47,$09,$ED
	.dc.b	$39,$8D,$D7,$A5,$A7,$84,$E1,$5C,$A7,$84,$60,$D9,$E3,$2D,$76,$AE
	.dc.b	$91,$B2,$4F,$22,$46,$9B,$1E,$16,$BD,$4E,$31,$B2,$3B,$6B,$B8,$E3
	.dc.b	$3E,$38,$D1,$0D,$4F,$08,$D4,$60,$D3,$82,$57,$C3,$9E,$0D,$9D,$B5
	.dc.b	$E4,$49,$E2,$FC,$68,$D4,$62,$3D,$A7,$23,$44,$32,$73,$8E,$24,$75
	.dc.b	$5F,$02,$31,$1C,$27,$B4,$E6,$37,$5E,$A6,$9E,$13,$85,$72,$9E,$11
	.dc.b	$83,$67,$8C,$B5,$DA,$BA,$46,$C9,$3C,$89,$1A,$7F,$4D,$AF,$F3,$92
	.dc.b	$4F,$F4,$26,$BF,$D0,$9A,$FF,$42,$6B,$FD,$35,$AF,$F4,$17,$8C,$76
	.dc.b	$8F,$E0,$51,$FB,$22,$38,$21,$FA,$97,$2D,$33,$1C,$DC,$11,$C1,$27
	.dc.b	$C9,$1C,$1A,$ED,$B1,$B8,$F4,$1B,$29,$B2,$BB,$7F,$B4,$CF,$9A,$31
	.dc.b	$5F,$DE,$E7,$81,$A3,$A1,$4C,$BB,$8E,$33,$CE,$BC,$A4,$AF,$0B,$5E
	.dc.b	$53,$C1,$EB,$5C,$E7,$4D,$8E,$24,$38,$C6,$AB,$C1,$19,$9E,$0E,$8C
	.dc.b	$46,$9C,$12,$36,$48,$D3,$63,$CB,$18,$23,$82,$1F,$11,$CB,$4C,$D7
	.dc.b	$BD,$C1,$1C,$12,$BD,$91,$C1,$AE,$DB,$1B,$8F,$69,$B2,$9B,$2B,$B7
	.dc.b	$E0,$4F,$9A,$31,$5E,$F9,$FE,$93,$F6,$8F,$47,$FC,$60,$D8,$FF,$1B
	.dc.b	$9F,$E3,$87,$0E,$03,$87,$6D,$78,$47,$42,$3C,$4E,$2E,$1E,$91,$C5
	.dc.b	$FF,$2D,$35,$D9,$FE,$5D,$AF,$F0,$2C,$79,$1B,$3F,$E0,$28,$C6,$1E
	.dc.b	$2A,$E0,$D7,$E0,$CE,$C8,$E8,$F0,$B8,$76,$64,$7E,$08,$E5,$FF,$08
	.dc.b	$B6,$34,$D7,$FB,$BB,$70,$FF,$0E,$B6,$FF,$CC,$AD,$9F,$F3,$D3,$5E
	.dc.b	$F5,$D8,$8F,$66,$BF,$73,$DC,$FB,$64,$7E,$86,$28,$C4,$7D,$06,$BF
	.dc.b	$69,$AF,$E0,$5A,$FD,$47,$8E,$1F,$49,$FA,$04,$8F,$8D,$3D,$DB,$19
	.dc.b	$7C,$C6,$08,$60,$93,$D5,$CB,$5C,$E1,$E5,$72,$D7,$27,$9C,$C9,$4C
	.dc.b	$9F,$9A,$9F,$24,$F5,$24,$F3,$6C,$F1,$6B,$C4,$EA,$53,$AD,$7C,$16
	.dc.b	$D7,$6A,$E8,$48,$E8,$D7,$95,$76,$AE,$4D,$AE,$53,$C0,$87,$0A,$E5
	.dc.b	$5C,$1B,$3C,$27,$47,$09,$D3,$63,$51,$88,$D9,$1C,$12,$36,$48,$F8
	.dc.b	$71,$96,$C6,$5F,$86,$60,$86,$09,$5E,$F7,$2D,$73,$5E,$C7,$2D,$72
	.dc.b	$7A,$CC,$94,$C9,$EB,$9F,$24,$F5,$24,$F3,$6C,$F1,$6B,$F1,$3E,$D5
	.dc.b	$F3,$5F,$ED,$96,$C7,$F1,$F6,$BF,$CA,$5A,$FF,$2E,$6B,$FE,$E6,$6B
	.dc.b	$FE,$FB,$6A,$69,$23,$C0,$48,$F0,$3F,$EA,$48,$F0,$CF,$FA,$B1,$B5
	.dc.b	$DE,$74,$4A,$FE,$3D,$3A,$7C,$51,$97,$FD,$23,$19,$7C,$91,$9A,$EC
	.dc.b	$DA,$ED,$3D,$23,$0F,$43,$04,$3A,$BA,$7A,$99,$9E,$AD,$9E,$AF,$03
	.dc.b	$07,$53,$04,$8E,$2D,$8E,$31,$82,$38,$24,$73,$8C,$B6,$32,$F9,$8E
	.dc.b	$86,$48,$66,$38,$B6,$38,$D7,$CB,$25,$7C,$16,$CF,$B0,$E0,$74,$72
	.dc.b	$F8,$0C,$91,$CD,$78,$1B,$5E,$0A,$EC,$4A,$ED,$3A,$7B,$46,$5E,$D1
	.dc.b	$97,$B4,$66,$BB,$36,$BB,$4F,$48,$C3,$D0,$C1,$0E,$AE,$9E,$A6,$67
	.dc.b	$AB,$67,$AB,$C0,$C1,$D4,$C1,$23,$8B,$F9,$AF,$88,$FD,$C1,$1F,$DC
	.dc.b	$35,$FD,$C3,$5F,$DC,$35,$FD,$C3,$5F,$DC,$3C,$63,$F6,$0F,$EC,$C9
	.dc.b	$1F,$AE,$6B,$E3,$30,$43,$04,$8E,$6F,$ED,$B2,$FE,$20,$47,$F8,$83
	.dc.b	$5F,$E2,$0D,$7F,$88,$35,$FE,$20,$D7,$F8,$83,$FC,$6F,$86,$47,$F1
	.dc.b	$B6,$BF,$C6,$DA,$FF,$1B,$6B,$FC,$6D,$AF,$F1,$B6,$BF,$C6,$FA,$7F
	.dc.b	$94,$11,$FE,$50,$D7,$F9,$43,$5F,$E5,$0D,$7F,$94,$35,$FE,$50,$FF
	.dc.b	$33,$E9,$91,$FC,$CD,$AF,$F3,$36,$BF,$CC,$DA,$FF,$33,$6B,$FC,$CD
	.dc.b	$AF,$F3,$3F,$13,$FA,$01,$1F,$E8,$0D,$7F,$A0,$35,$FE,$80,$D7,$FA
	.dc.b	$03,$5F,$E8,$0F,$F9,$9D,$7C,$B3,$3C,$36,$3F,$CD,$8D,$7F,$CE,$CD
	.dc.b	$7F,$CF,$6D,$44,$70,$D8,$EA,$47,$AB,$5E,$AF,$97,$F5,$0F,$E8,$44
	.dc.b	$D7,$AC,$B4,$F2,$4E,$63,$FC,$D4,$D7,$55,$C2,$31,$3C,$2B,$9B,$6B
	.dc.b	$8C,$62,$BB,$DB,$A7,$E2,$92,$BE,$2B,$5F,$8A,$D7,$E2,$B5,$E6,$D7
	.dc.b	$E3,$3C,$6F,$E7,$4F,$79,$FF,$42,$35,$F5,$9C,$0C,$CF,$9D,$C1,$A3
	.dc.b	$EE,$B9,$69,$C0,$86,$E3,$B4,$F8,$A7,$84,$62,$BD,$35,$C2,$B4,$D8
	.dc.b	$CB,$CE,$B8,$91,$D1,$E6,$9E,$2E,$CC,$9E,$63,$04,$76,$49,$EE,$D8
	.dc.b	$E9,$18,$23,$82,$1D,$63,$83,$63,$6D,$37,$5F,$05,$B5,$EB,$7C,$93
	.dc.b	$C2,$75,$1C,$48,$E0,$93,$C6,$BC,$31,$89,$F8,$2D,$7A,$9E,$03,$33
	.dc.b	$C6,$BC,$0D,$8D,$1C,$0F,$0B,$96,$9C,$88,$6E,$7C,$15,$E4,$9E,$11
	.dc.b	$8A,$EF,$5C,$2B,$4D,$8C,$BC,$EB,$89,$1D,$1E,$69,$E2,$EC,$C9,$E6
	.dc.b	$30,$47,$64,$9E,$ED,$8F,$E2,$1E,$2F,$F1,$A2,$3F,$C6,$9A,$FF,$1A
	.dc.b	$6B,$FC,$69,$AF,$F0,$06,$BD,$0E,$47,$ED,$1F,$D2,$91,$C1,$0F,$D6
	.dc.b	$39,$69,$98,$FD,$FC,$65,$B1,$99,$F1,$34,$F1,$1C,$89,$1C,$1B,$1C
	.dc.b	$1A,$70,$8D,$12,$34,$D8,$D4,$60,$8E,$08,$7F,$48,$9F,$44,$6A,$30
	.dc.b	$79,$CA,$78,$8B,$3D,$DB,$1C,$6D,$5D,$27,$BC,$61,$E8,$D7,$A3,$6B
	.dc.b	$A3,$6B,$A5,$73,$AE,$E7,$47,$13,$B8,$C1,$DA,$32,$47,$04,$3C,$2E
	.dc.b	$5A,$66,$BC,$71,$96,$C6,$6B,$CA,$D3,$C2,$72,$24,$70,$6C,$70,$69
	.dc.b	$C2,$34,$48,$D3,$63,$51,$82,$38,$21,$F1,$89,$5F,$15,$C7,$DB,$7F
	.dc.b	$82,$1F,$EE,$67,$F7,$47,$0E,$03,$87,$6D,$78,$47,$F8,$4E,$31,$1F
	.dc.b	$E0,$A6,$BC,$A3,$44,$7B,$19,$2C,$70,$3C,$75,$FE,$19,$6D,$78,$27
	.dc.b	$FC,$2F,$39,$AF,$EC,$D5,$A9,$F1,$46,$08,$F0,$AE,$4F,$F6,$F9,$E0
	.dc.b	$F8,$CC,$4F,$F7,$1F,$D0,$1F,$53,$49,$3C,$CF,$F1,$4C,$69,$C0,$70
	.dc.b	$FF,$7D,$6E,$1F,$E3,$56,$E1,$FE,$26,$6C,$F4,$AE,$EF,$16,$D7,$C5
	.dc.b	$F8,$EF,$A4,$49,$F9,$6D,$7E,$E6,$28,$C4,$7D,$C6,$BF,$83,$6B,$F6
	.dc.b	$DE,$53,$DE,$3D,$B3,$CD,$B3,$F4,$88,$E0,$87,$CF,$72,$D3,$33,$D5
	.dc.b	$C1,$1C,$13,$0F,$34,$70,$6B,$B6,$C6,$E3,$D4,$6C,$A6,$CA,$ED,$FD
	.dc.b	$0C,$F9,$A3,$15,$EF,$9E,$06,$8E,$85,$32,$EE,$38,$CF,$3A,$F2,$92
	.dc.b	$BC,$2D,$79,$4F,$07,$AD,$73,$9D,$36,$38,$90,$E3,$1A,$AF,$04,$66
	.dc.b	$78,$3A,$31,$1A,$70,$48,$D9,$23,$4D,$8F,$2C,$60,$8E,$08,$7C,$47
	.dc.b	$2D,$33,$5E,$F7,$04,$70,$4A,$F6,$47,$06,$BB,$6C,$6E,$3D,$A6,$CA
	.dc.b	$6C,$AE,$DF,$81,$3E,$68,$C5,$7C,$2F,$88,$FE,$40,$4A,$FE,$32,$D7
	.dc.b	$F5,$ED,$7F,$94,$B5,$FE,$60,$D7,$F9,$43,$C6,$7F,$90,$3F,$C7,$09
	.dc.b	$1F,$B0,$6B,$FC,$65,$B1,$C2,$34,$43,$45,$74,$E0,$8E,$09,$7E,$68
	.dc.b	$E0,$D7,$6D,$8D,$C7,$A8,$D9,$4D,$95,$DB,$FD,$3A,$7D,$33,$CC,$91
	.dc.b	$B6,$BB,$3A,$94,$EA,$59,$EB,$4A,$EB,$3C,$C8,$6E,$BA,$CE,$E3,$AB
	.dc.b	$5E,$AD,$AE,$B5,$D2,$79,$1D,$27,$55,$D2,$75,$18,$9D,$46,$08,$F8
	.dc.b	$1A,$F8,$1B,$5E,$08,$D1,$0D,$15,$D3,$82,$38,$25,$79,$63,$83,$5D
	.dc.b	$B6,$37,$1E,$83,$65,$36,$57,$6F,$C4,$25,$7C,$16,$D7,$B2,$BB,$12
	.dc.b	$03,$ED,$DF,$33,$FE,$08,$8C,$36,$33,$1F,$E0,$C6,$BF,$E1,$B6,$BF
	.dc.b	$E2,$06,$BD,$8E,$67,$18,$E8,$47,$04,$3A,$3F,$EF,$58,$EF,$82,$BA
	.dc.b	$46,$1B,$3B,$25,$7F,$81,$A7,$4F,$C0,$6B,$E4,$15,$D5,$D4,$F2,$73
	.dc.b	$5D,$9E,$9E,$AF,$C5,$23,$F1,$5A,$FC,$56,$BF,$15,$AF,$A9,$AF,$95
	.dc.b	$FA,$1F,$55,$23,$E8,$35,$FA,$0D,$7E,$83,$5F,$8A,$D7,$E0,$35,$FC
	.dc.b	$07,$47,$DD,$23,$F7,$5A,$FD,$D6,$BF,$75,$AF,$D2,$6B,$F3,$5F,$C4
	.dc.b	$7B,$F2,$3F,$10,$D7,$F1,$0D,$7F,$10,$D7,$EE,$B5,$FB,$0D,$7F,$31
	.dc.b	$D3,$F9,$52,$3F,$95,$6B,$F9,$56,$BF,$95,$6B,$F8,$A6,$BF,$85,$7F
	.dc.b	$41,$D3,$23,$F4,$0D,$7F,$40,$D7,$F4,$0D,$7F,$2A,$D7,$F2,$0D,$7F
	.dc.b	$61,$D3,$FA,$D2,$3F,$AD,$6B,$FA,$D6,$BF,$AD,$6B,$FA,$26,$BF,$9D
	.dc.b	$7F,$71,$D3,$23,$F7,$0D,$7F,$70,$D7,$F7,$0D,$7F,$5A,$D7,$F5,$0D
	.dc.b	$7F,$84,$74,$FF,$05,$23,$FC,$15,$AF,$F0,$56,$BF,$C1,$5A,$FE,$E9
	.dc.b	$AF,$ED,$5F,$E2,$5F,$6F,$FB,$38,$FF,$54,$35,$FF,$56,$B5,$FE,$1C
	.dc.b	$D7,$F8,$EB,$53,$49,$AF,$62,$47,$FA,$A2,$32,$DF,$1B,$C6,$47,$04
	.dc.b	$3C,$AE,$5A,$64,$F3,$B9,$69,$DF,$A7,$DE,$47,$04,$3E,$0B,$96,$99
	.dc.b	$3E,$1B,$96,$9E,$BF,$1B,$E5,$91,$C1,$0F,$9A,$E5,$A6,$4F,$9E,$E5
	.dc.b	$A7,$C7,$F1,$BE,$B9,$1C,$10,$FB,$2E,$5A,$64,$FB,$6E,$5A,$7D,$3F
	.dc.b	$47,$F0,$64,$70,$43,$F0,$AE,$5A,$64,$FC,$3B,$96,$9F,$7F,$C6,$FC
	.dc.b	$79,$1C,$10,$FC,$8B,$96,$99,$3F,$26,$E5,$A7,$E2,$FE,$5F,$F3,$64
	.dc.b	$70,$43,$F3,$AE,$5A,$64,$FC,$FB,$96,$9F,$97,$F4,$BF,$4E,$47,$04
	.dc.b	$3F,$52,$E5,$A6,$4F,$D5,$B9,$69,$FA,$3F,$6F,$F6,$64,$70,$43,$F6
	.dc.b	$AE,$5A,$64,$FD,$BB,$96,$9F,$E0,$53,$FC,$27,$E5,$BC,$C7,$A5,$FE
	.dc.b	$09,$EB,$FE,$E9,$39,$8C,$70,$F0,$AD,$DF,$F8,$72,$F7,$18,$8F,$EF
	.dc.b	$22,$F9,$3B,$AE,$57,$D2,$35,$5C,$6F,$9C,$E6,$F8,$D7,$46,$CE,$A3
	.dc.b	$A1,$CA,$78,$57,$3B,$EF,$3A,$BE,$31,$88,$E9,$3B,$79,$1B,$71,$3C
	.dc.b	$23,$27,$0A,$EE,$47,$C6,$3F,$4A,$7C,$FF,$BE,$30,$D8,$CC,$70,$6B
	.dc.b	$F2,$8D,$99,$8E,$04,$70,$43,$84,$72,$C7,$F9,$E3,$71,$FE,$73,$F2
	.dc.b	$3F,$D0,$09,$5F,$CC,$DA,$FF,$42,$6B,$FD,$09,$AF,$23,$81,$0E,$07
	.dc.b	$C2,$E4,$47,$C1,$6C,$66,$30,$47,$04,$3E,$14,$70,$6C,$6D,$A6,$EF
	.dc.b	$FA,$D4,$ED,$B3,$B6,$CE,$EB,$D4,$F2,$24,$69,$B1,$A8,$C1,$1C,$12
	.dc.b	$3F,$B1,$CF,$78,$C4,$7F,$64,$8C,$3A,$3C,$45,$3B,$96,$7B,$36,$79
	.dc.b	$D7,$B5,$B5,$D2,$7B,$1D,$0A,$74,$2D,$74,$6C,$F2,$78,$57,$57,$C0
	.dc.b	$43,$65,$74,$E0,$8E,$09,$5E,$56,$C7,$68,$C1,$1C,$10,$F4,$47,$06
	.dc.b	$C6,$DA,$6E,$BD,$93,$B6,$CE,$DB,$3B,$AF,$80,$F2,$24,$69,$B1,$A8
	.dc.b	$C1,$1C,$12,$3E,$24,$F7,$8C,$47,$C4,$8C,$3F,$2F,$81,$F0,$1F,$EC
	.dc.b	$36,$C7,$F3,$73,$FD,$92,$E1,$C3,$5F,$E7,$8D,$7F,$D8,$4D,$78,$C7
	.dc.b	$52,$3D,$44,$6D,$B1,$A7,$BB,$FC,$66,$3C,$57,$FE,$47,$6D,$F8,$27
	.dc.b	$44,$AF,$E4,$73,$A7,$C6,$62,$34,$FF,$1F,$8D,$3E,$23,$44,$35,$1C
	.dc.b	$2B,$C4,$DA,$F1,$4E,$E7,$2D,$8C,$BC,$8C,$1E,$43,$04,$8E,$2D,$8E
	.dc.b	$31,$82,$38,$24,$73,$8C,$B6,$32,$FB,$4E,$86,$48,$66,$38,$B6,$38
	.dc.b	$D7,$C6,$25,$7C,$19,$E3,$18,$7D,$A7,$13,$31,$B7,$04,$8E,$0F,$90
	.dc.b	$C4,$6E,$33,$5E,$46,$D7,$92,$74,$4A,$F0,$CE,$9F,$09,$88,$D3,$DE
	.dc.b	$34,$F8,$0D,$10,$D4,$70,$AF,$13,$6B,$C5,$3B,$9C,$B6,$32,$F2,$30
	.dc.b	$79,$0C,$12,$38,$B6,$38,$C6,$0F,$D1,$6B,$FB,$92,$4F,$EB,$6F,$FC
	.dc.b	$5C,$D8,$CB,$FB,$A6,$BF,$BA,$6B,$FB,$A6,$BF,$C0,$9F,$D9,$11,$FD
	.dc.b	$6B,$5F,$D5,$36,$7C,$B1,$A2,$1A,$2B,$FE,$C1,$FC,$19,$DD,$FE,$17
	.dc.b	$6C,$FF,$87,$1A,$FF,$85,$5B,$3C,$67,$81,$27,$83,$5E,$0D,$8E,$11
	.dc.b	$FC,$1B,$EC,$3F,$91,$11,$F9,$4D,$7F,$90,$B5,$FE,$42,$D7,$F9,$0B
	.dc.b	$5F,$E5,$2F,$18,$FE,$30,$FF,$1B,$24,$78,$48,$7F,$13,$72,$D3,$31
	.dc.b	$C9,$B8,$F8,$1A,$78,$0A,$7F,$2E,$9E,$73,$C6,$3A,$11,$C1,$27,$A4
	.dc.b	$65,$B1,$98,$EC,$64,$AE,$5F,$01,$E2,$C5,$5E,$2A,$E6,$49,$E2,$D7
	.dc.b	$89,$D4,$B5,$D6,$B9,$C7,$5A,$E8,$F2,$21,$C2,$BA,$4F,$08,$E8,$DA
	.dc.b	$E9,$5C,$A7,$81,$CA,$75,$5C,$A7,$51,$89,$D4,$60,$8F,$94,$87,$81
	.dc.b	$CB,$4C,$D7,$89,$B5,$DD,$A7,$62,$9E,$62,$57,$8D,$B5,$DA,$BA,$47
	.dc.b	$72,$38,$25,$77,$8C,$B6,$33,$1E,$13,$25,$72,$F8,$0F,$2D,$78,$FE
	.dc.b	$E5,$F6,$1F,$E0,$19,$46,$1F,$F0,$23,$5F,$F0,$93,$5F,$F6,$53,$51
	.dc.b	$1C,$36,$3B,$46,$23,$BC,$60,$8F,$81,$AF,$FA,$86,$3C,$54,$BE,$F3
	.dc.b	$C8,$FF,$57,$CE,$A3,$2D,$8F,$11,$AA,$FF,$59,$46,$63,$91,$DE,$32
	.dc.b	$47,$81,$0D,$16,$BC,$8F,$63,$24,$72,$F6,$6C,$F8,$49,$3D,$E7,$85
	.dc.b	$7A,$A3,$15,$C1,$B1,$99,$F1,$35,$E2,$78,$E7,$83,$A3,$11,$B8,$CC
	.dc.b	$70,$3C,$D1,$92,$3C,$08,$68,$B5,$EA,$7B,$19,$23,$97,$B3,$67,$C2
	.dc.b	$49,$EF,$3C,$2B,$DE,$F6,$6B,$93,$BD,$73,$9E,$2D,$8E,$06,$A7,$8C
	.dc.b	$66,$39,$1E,$48,$C9,$1E,$04,$34,$5A,$F4,$3D,$8C,$91,$CB,$D9,$B3
	.dc.b	$E1,$24,$F7,$9E,$15,$EA,$8C,$57,$06,$C6,$67,$C4,$D7,$ED,$7D,$CF
	.dc.b	$E4,$1F,$D3,$87,$0F,$EA,$9A,$FE,$E9,$AE,$E3,$F8,$34,$60,$8E,$09
	.dc.b	$1F,$E8,$B8,$CB,$63,$F6,$F5,$FA,$29,$FD,$1B,$63,$27,$8C,$B3,$E3
	.dc.b	$6B,$E3,$3F,$87,$C7,$16,$C7,$16,$9C,$4A,$71,$11,$C5,$B1,$C5,$A7
	.dc.b	$C3,$EA,$F8,$24,$7D,$CD,$7F,$D6,$D4,$8C,$47,$D0,$8C,$3C,$08,$70
	.dc.b	$8D,$BF,$4F,$AB,$E7,$91,$F9,$8D,$7B,$B5,$EC,$D7,$91,$C0,$87,$9B
	.dc.b	$AB,$ED,$91,$FB,$0D,$74,$60,$86,$08,$7B,$5E,$0F,$BF,$3F,$C3,$91
	.dc.b	$FC,$23,$5F,$8E,$D7,$81,$A2,$1A,$29,$F2,$FA,$BF,$26,$47,$F2,$0D
	.dc.b	$7E,$9B,$5E,$06,$88,$68,$A7,$D7,$EA,$FC,$F9,$1F,$CE,$35,$FB,$ED
	.dc.b	$78,$1A,$21,$A2,$9F,$83,$EA,$FD,$59,$1F,$D4,$35,$FC,$5B,$5E,$06
	.dc.b	$88,$68,$A7,$E3,$FA,$BF,$6E,$47,$F6,$8D,$7F,$2E,$D7,$81,$A2,$1A
	.dc.b	$29,$F9,$BF,$84,$FE,$06,$47,$F8,$03,$5F,$D1,$B5,$E0,$68,$86,$8A
	.dc.b	$7F,$16,$FD,$0D,$F3,$1F,$C0,$1B,$1F,$BA,$6B,$FB,$60,$E3,$83,$FB
	.dc.b	$7B,$78,$3F,$C0,$ED,$9F,$F0,$BB,$5E,$35,$DC,$8F,$76,$BD,$F8,$1F
	.dc.b	$07,$FC,$90,$91,$FC,$85,$AF,$F0,$46,$BF,$C0,$DA,$FF,$31,$6B,$FC
	.dc.b	$C5,$E3,$87,$F2,$77,$F9,$0C,$F8,$27,$F8,$DB,$5F,$24,$68,$8E,$9A
	.dc.b	$E8,$C1,$0F,$67,$A1,$FC,$F8,$8F,$F3,$B6,$BF,$C7,$9A,$FF,$1D,$6B
	.dc.b	$FD,$2D,$AF,$F4,$B7,$FA,$4F,$D6,$48,$FE,$8C,$D7,$F9,$4B,$5F,$E5
	.dc.b	$0D,$7F,$AB,$35,$FE,$AC,$D7,$FA,$E7,$BF,$FD,$58,$8F,$F5,$36,$BF
	.dc.b	$CD,$9A,$FF,$35,$6B,$FD,$8D,$AF,$F6,$37,$FB,$0F,$D5,$C8,$FE,$BC
	.dc.b	$D7,$FA,$0B,$5F,$E8,$0D,$7F,$B7,$35,$FE,$DC,$D7,$FB,$A7,$B9,$FD
	.dc.b	$B8,$8F,$F6,$B6,$BF,$D3,$9A,$FF,$4D,$6B,$FD,$ED,$AF,$F7,$B7,$FB
	.dc.b	$CF,$D8,$48,$FE,$EC,$D7,$FA,$CB,$5F,$EB,$0D,$7F,$C0,$CD,$7F,$C0
	.dc.b	$CD,$7F,$C1,$9E,$1F,$F8,$18,$8F,$F8,$0D,$AF,$F6,$66,$BF,$D9,$5A
	.dc.b	$FF,$84,$DA,$FF,$84,$DF,$F4,$B7,$DD,$3E,$56,$92,$7F,$C3,$A7,$FA
	.dc.b	$4A,$34,$E0,$38,$76,$D7,$FC,$4A,$D7,$FD,$1A,$D7,$A4,$77,$78,$B5
	.dc.b	$E3,$F1,$FE,$62,$3E,$19,$46,$23,$C2,$D7,$89,$B7,$0F,$89,$AF,$B7
	.dc.b	$E6,$BE,$19,$1E,$CD,$7D,$AD,$7D,$AD,$7D,$AD,$7A,$1E,$7F,$8C,$F9
	.dc.b	$E4,$70,$43,$E4,$B9,$69,$93,$E5,$B9,$69,$F0,$FC,$4F,$B6,$47,$04
	.dc.b	$3E,$AB,$96,$99,$3E,$BB,$96,$9F,$3F,$DA,$FC,$39,$1C,$10,$FC,$0B
	.dc.b	$96,$99,$3F,$06,$E5,$A7,$DB,$FB,$6F,$C9,$91,$C1,$0F,$C6,$B9,$69
	.dc.b	$93,$F1,$EE,$5A,$7E,$1F,$ED,$BF,$3E,$47,$04,$3F,$32,$E5,$A6,$4F
	.dc.b	$CD,$B9,$69,$F9,$3F,$6B,$F5,$64,$70,$43,$F4,$AE,$5A,$64,$FD,$3B
	.dc.b	$96,$9F,$9F,$F8,$3F,$DB,$91,$C1,$0F,$D8,$B9,$69,$93,$F6,$6E,$5A
	.dc.b	$7E,$AF,$EC,$FF,$81,$91,$C1,$0F,$DE,$B9,$69,$93,$F7,$EE,$5A,$7E
	.dc.b	$DF,$E2,$BF,$87,$91,$C1,$0F,$E1,$2E,$5A,$64,$FE,$16,$E5,$A7,$F0
	.dc.b	$3F,$83,$FE,$36,$47,$04,$3F,$8A,$B9,$69,$93,$F8,$BB,$96,$9F,$C3
	.dc.b	$FE,$CB,$F9,$39,$1C,$10,$FE,$42,$E5,$A6,$4F,$E4,$6E,$5A,$7F,$1B
	.dc.b	$F8,$AF,$E6,$64,$70,$43,$F9,$6B,$96,$99,$3F,$97,$B9,$69,$FC,$9F
	.dc.b	$E4,$3F,$9F,$91,$C1,$0F,$E7,$2E,$5A,$64,$FE,$76,$E5,$A7,$F3,$3F
	.dc.b	$3F,$FA,$59,$1C,$10,$FE,$8A,$E5,$A6,$4F,$E8,$EE,$5A,$7F,$3F,$F9
	.dc.b	$9F,$EA,$E4,$70,$43,$FA,$8B,$96,$99,$3F,$A9,$B9,$69,$FD,$2F,$E4
	.dc.b	$7F,$B1,$91,$C1,$0F,$EB,$AE,$5A,$64,$FE,$BE,$E5,$A7,$F5,$7F,$3F
	.dc.b	$FB,$79,$1C,$10,$FE,$D2,$E5,$A6,$4F,$ED,$6E,$5A,$7F,$63,$F8,$2F
	.dc.b	$EF,$64,$70,$43,$FB,$AB,$96,$99,$3F,$BB,$B9,$69,$FD,$FF,$ED,$BF
	.dc.b	$C0,$08,$38,$69,$A7,$04,$70,$43,$A1,$4F,$F0,$7B,$B7,$C1,$F3,$BC
	.dc.b	$C8,$E0,$87,$07,$2D,$32,$79,$CF,$44,$23,$71,$93,$D6,$FA,$7C,$59
	.dc.b	$1F,$05,$AF,$99,$AF,$19,$FF,$54,$B6,$32,$F8,$63,$2D,$8C,$BE,$4F
	.dc.b	$88,$F9,$04,$7E,$83,$5F,$8A,$D7,$C0,$D7,$B3,$5E,$87,$D9,$F9,$9F
	.dc.b	$A8,$47,$04,$3E,$8B,$96,$99,$3D,$CE,$5A,$7E,$17,$E4,$3F,$00,$47
	.dc.b	$04,$3E,$EB,$96,$99,$3E,$53,$96,$9F,$91,$BF,$C6,$11,$C1,$0F,$C4
	.dc.b	$B9,$69,$93,$EB,$39,$69,$F9,$DF,$A2,$FC,$C1,$1C,$10,$FC,$AB,$96
	.dc.b	$99,$3F,$04,$E5,$A7,$EA,$7E,$63,$F4,$84,$70,$43,$F4,$2E,$5A,$64
	.dc.b	$FC,$73,$96,$9F,$B5,$F9,$6F,$D8,$11,$C1,$0F,$D6,$B9,$69,$93,$F3
	.dc.b	$4E,$5A,$7F,$01,$F5,$FF,$78,$47,$04,$3F,$72,$E5,$A6,$4F,$D3,$39
	.dc.b	$69,$FC,$37,$E2,$7F,$84,$11,$C1,$0F,$E0,$AE,$5A,$64,$FD,$93,$96
	.dc.b	$9F,$C6,$7E,$33,$F8,$A1,$1C,$10,$FE,$22,$E5,$A6,$4F,$DF,$39,$69
	.dc.b	$FC,$97,$E2,$BF,$90,$11,$C1,$0F,$E3,$AE,$5A,$64,$FE,$14,$E5,$A7
	.dc.b	$F3,$1F,$73,$F9,$61,$1C,$10,$FE,$52,$E5,$A6,$4F,$E2,$CE,$5A,$7F
	.dc.b	$3D,$F8,$AF,$E7,$04,$70,$43,$F9,$AB,$96,$99,$3F,$91,$39,$69,$FD
	.dc.b	$27,$E2,$BF,$A2,$11,$C1,$0F,$E8,$2E,$5A,$64,$FE,$5C,$E5,$A7,$F5
	.dc.b	$5F,$8A,$FE,$A0,$47,$04,$3F,$A6,$B9,$69,$93,$F9,$D3,$96,$9F,$D8
	.dc.b	$7E,$2B,$FA,$E1,$1C,$10,$FE,$B2,$E5,$A6,$4F,$E8,$CE,$5A,$7F,$6D
	.dc.b	$F8,$AF,$ED,$04,$70,$43,$FB,$2B,$96,$99,$3F,$A9,$39,$69,$FD,$E7
	.dc.b	$E2,$BF,$BA,$11,$C1,$0F,$EE,$2E,$5A,$64,$FE,$BC,$E5,$A7,$F8,$17
	.dc.b	$E2,$BF,$C0,$04,$70,$43,$FB,$EB,$96,$99,$3F,$B5,$39,$69,$FD,$D4
	.dc.b	$FF,$08,$7D,$D3,$DA,$FF,$09,$B6,$34,$1C,$3C,$5A,$FF,$7F,$6B,$FE
	.dc.b	$05,$6B,$DA,$3A,$91,$EB,$1A,$7B,$46,$5E,$DF,$17,$E4,$23,$C1,$AE
	.dc.b	$8C,$10,$C1,$0F,$33,$C1,$F4,$7C,$27,$C0,$23,$E4,$6B,$EF,$6B,$C0
	.dc.b	$D1,$0D,$14,$F8,$9F,$09,$F3,$08,$FC,$06,$BF,$2D,$AF,$03,$44,$34
	.dc.b	$53,$E8,$7C,$27,$D8,$23,$F3,$1A,$FD,$76,$BC,$0D,$10,$D1,$4F,$B9
	.dc.b	$F0,$9F,$84,$23,$F6,$1A,$FE,$0D,$AF,$03,$44,$34,$53,$F1,$1F,$09
	.dc.b	$F9,$02,$3F,$84,$6B,$F8,$F6,$BC,$0D,$10,$D1,$4F,$CA,$7C,$27,$E7
	.dc.b	$08,$FE,$41,$AF,$E6,$DA,$F0,$34,$43,$45,$3F,$41,$F0,$9F,$A8,$23
	.dc.b	$F9,$C6,$BF,$A7,$6B,$C0,$D1,$0D,$14,$FD,$67,$C2,$7E,$D0,$8F,$EA
	.dc.b	$1A,$FE,$CD,$AF,$03,$44,$34,$53,$F7,$1F,$09,$FC,$00,$8F,$ED,$1A
	.dc.b	$FE,$FD,$AF,$03,$44,$34,$53,$F8,$27,$C5,$7F,$0C,$23,$FC,$01,$AF
	.dc.b	$F0,$B6,$BC,$0D,$10,$D1,$4F,$E2,$1F,$05,$FC,$60,$8F,$F0,$C6,$BF
	.dc.b	$C5,$DA,$F0,$34,$43,$45,$3F,$8E,$7C,$47,$F2,$42,$3F,$C6,$1A,$FF
	.dc.b	$23,$6B,$C0,$D1,$0D,$14,$FE,$51,$F0,$7F,$CC,$08,$FF,$24,$6B,$FC
	.dc.b	$BD,$AF,$03,$44,$34,$53,$F9,$A7,$C1,$FF,$3C,$23,$FC,$C1,$AF,$F3
	.dc.b	$B6,$BC,$0D,$10,$D1,$4F,$E8,$1F,$07,$FD,$20,$8F,$F3,$C6,$BF,$D1
	.dc.b	$DA,$F0,$34,$43,$45,$3F,$A6,$7C,$1F,$F5,$42,$3F,$D2,$1A,$FF,$53
	.dc.b	$6B,$C0,$D1,$0D,$14,$FE,$B1,$F0,$7F,$D8,$08,$FF,$54,$6B,$FD,$7D
	.dc.b	$AF,$03,$44,$34,$53,$FB,$27,$C1,$FF,$6C,$23,$FD,$81,$AF,$F6,$B6
	.dc.b	$BC,$0D,$10,$D1,$4F,$EE,$1F,$07,$FD,$E0,$8F,$F6,$C6,$BF,$DD,$DA
	.dc.b	$F0,$34,$43,$45,$3F,$DF,$5F,$74,$F8,$4F,$F0,$1C,$23,$0F,$F8,$19
	.dc.b	$AF,$F8,$35,$AF,$28,$E3,$18,$81,$1C,$36,$36,$47,$B3,$5E,$DF,$07
	.dc.b	$D0,$8F,$8D,$AE,$8C,$10,$C1,$0F,$7B,$C1,$F8,$3F,$07,$E8,$23,$EF
	.dc.b	$6B,$EF,$6B,$C0,$D1,$0D,$14,$F9,$BF,$07,$F1,$08,$FC,$B6,$BF,$2D
	.dc.b	$AF,$03,$44,$34,$53,$EC,$FC,$1F,$D0,$23,$F5,$DA,$FD,$76,$BC,$0D
	.dc.b	$10,$D1,$4F,$C2,$FC,$1F,$DC,$23,$F8,$36,$BF,$83,$6B,$C0,$D1,$0D
	.dc.b	$14,$FC,$8F,$C1,$FE,$20,$8F,$E3,$DA,$FE,$3D,$AF,$03,$44,$34,$53
	.dc.b	$F3,$BF,$07,$F9,$42,$3F,$9B,$6B,$F9,$B6,$BC,$0D,$10,$D1,$4F,$D4
	.dc.b	$FC,$1B,$E1,$BF,$4C,$D8,$FD,$33,$5D,$C6,$88,$E9,$AE,$8C,$10,$C1
	.dc.b	$FB,$8F,$84,$7D,$CF,$ED,$5B,$1F,$B5,$6B,$B8,$D1,$1D,$35,$D1,$82
	.dc.b	$18,$3F,$77,$F7,$42,$00

;--------------------------------------------------------------------------------
;	FLOG10.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
flog10_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FLOG10.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FLOG10.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	flog10_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	flog10.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FLOG10.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	flog10.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FLOG10.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
flog10_data_compressed::
	.dc.b	$00,$02,$A1,$14,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$FD
	.dc.b	$4F,$E5,$1F,$97,$BE,$4B,$CF,$E9,$7A,$3D,$2D,$4B,$E1,$2F,$C9,$E3
	.dc.b	$E4,$73,$F1,$7A,$F8,$8D,$FB,$9C,$3D,$3E,$3E,$9F,$2F,$4F,$9F,$A7
	.dc.b	$D3,$D3,$EB,$E9,$F6,$F4,$FB,$FA,$7E,$0F,$4F,$C3,$E9,$F8,$BD,$3F
	.dc.b	$97,$EC,$7C,$CF,$63,$E6,$FB,$1F,$D8,$FF,$14,$FE,$40,$7E,$63,$C5
	.dc.b	$3F,$98,$7F,$B9,$3F,$D9,$9B,$BF,$1D,$F9,$6F,$CF,$7E,$9B,$F5,$DF
	.dc.b	$B6,$FD,$F7,$F0,$6F,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$FE,$ED,$F8,$AB
	.dc.b	$E6,$2F,$C7,$FE,$13,$DC,$FE,$F5,$F3,$9F,$47,$8F,$FC,$0D,$7F,$E7
	.dc.b	$B8,$FA,$11,$FE,$09,$7F,$DB,$87,$FA,$50,$FF,$03,$1F,$CF,$CF,$80
	.dc.b	$7B,$D3,$EF,$6F,$1F,$83,$8F,$C3,$C7,$E2,$E3,$F1,$F1,$F9,$38,$FC
	.dc.b	$BC,$7E,$6E,$3F,$3F,$1F,$A3,$8F,$D3,$C7,$EA,$E3,$FF,$0A,$77,$7B
	.dc.b	$F7,$FF,$10,$3F,$ED,$CE,$1F,$F2,$6B,$FE,$97,$2F,$C6,$F0,$F8,$AE
	.dc.b	$3F,$05,$CB,$DE,$E7,$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83
	.dc.b	$DE,$F0,$FB,$DE,$2F,$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$F5,$38,$3C
	.dc.b	$FE,$F7,$A3,$DE,$FF,$16,$7C,$EF,$F8,$14,$7C,$5E,$3E,$2F,$3F,$17
	.dc.b	$AF,$8B,$DF,$C5,$F0,$F8,$BE,$3F,$11,$CB,$E1,$79,$FC,$2F,$4F,$85
	.dc.b	$EB,$F0,$BD,$BE,$17,$BF,$C2,$F8,$3E,$17,$C3,$F0,$BE,$2F,$83,$F1
	.dc.b	$FC,$1F,$93,$DB,$F2,$FB,$7E,$6F,$6F,$CF,$EC,$7F,$90,$FE,$7B,$F1
	.dc.b	$5E,$61,$FC,$4B,$7C,$BD,$7C,$16,$FE,$0B,$87,$C1,$71,$F8,$2E,$5F
	.dc.b	$05,$CF,$E0,$BF,$33,$E8,$F5,$F7,$BB,$7B,$DD,$FD,$EF,$07,$BD,$E1
	.dc.b	$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD,$EF,$79,$FD,$EF
	.dc.b	$47,$BD,$FE,$5B,$F9,$FF,$C4,$73,$BD,$FF,$E7,$2A,$1F,$03,$BF,$8A
	.dc.b	$E1,$F2,$BC,$7D,$5E,$5C,$BC,$FE,$0B,$A7,$C1,$75,$F8,$2E,$DF,$05
	.dc.b	$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE,$0B,$C9,$F0,$5E
	.dc.b	$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$7B,$AF,$EF,$B5,$FD,$FE,$BF,$C0
	.dc.b	$35,$FE,$03,$AF,$B7,$E0,$BD,$DF,$05,$FE,$79,$F7,$FF,$CA,$47,$E3
	.dc.b	$BA,$3F,$D0,$CB,$E0,$34,$EF,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1C
	.dc.b	$FE,$03,$A7,$C0,$7F,$10,$E8,$ED,$F0,$1D,$FE,$03,$C1,$F0,$1F,$C5
	.dc.b	$3A,$FF,$8A,$F5,$FF,$16,$EB,$FE,$2F,$D7,$FC,$63,$AF,$F8,$CF,$5F
	.dc.b	$F1,$AE,$6F,$E3,$7C,$DF,$E9,$1F,$C5,$BF,$CD,$47,$A7,$9C,$FA,$5F
	.dc.b	$F5,$03,$FE,$FC,$3F,$D6,$C7,$F8,$98,$FE,$BC,$7F,$5D,$78,$F0,$F5
	.dc.b	$E1,$ED,$C1,$DF,$83,$E4,$7C,$0F,$C9,$F8,$1F,$95,$F0,$3F,$2F,$E0
	.dc.b	$7E,$67,$C0,$FC,$DF,$81,$F9,$DF,$03,$F3,$FE,$07,$E8,$7C,$0F,$FA
	.dc.b	$A7,$E8,$7F,$4E,$7E,$98,$7C,$0E,$7E,$07,$5F,$03,$BF,$81,$E1,$F0
	.dc.b	$3C,$7E,$07,$97,$C0,$F3,$F8,$1E,$9F,$03,$D7,$E0,$7B,$7C,$0F,$7F
	.dc.b	$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03,$F2,$FC,$9F,$99,$D7,$F3,$7A
	.dc.b	$FE,$77,$5F,$CF,$EB,$FA,$1D,$7F,$47,$AF,$E9,$75,$FD,$3E,$BF,$A9
	.dc.b	$D7,$FE,$C5,$F4,$FF,$81,$9F,$C0,$8F,$6F,$83,$FE,$D3,$5D,$7B,$EB
	.dc.b	$E1,$D7,$C7,$AF,$97,$5F,$3E,$BE,$9D,$7F,$27,$E0,$7E,$57,$C0,$FC
	.dc.b	$BF,$81,$F0,$77,$7C,$DF,$81,$FF,$6B,$7D,$17,$EF,$0F,$DC,$F0,$7F
	.dc.b	$B8,$D7,$BF,$AF,$7F,$7E,$FF,$0F,$7F,$8F,$BF,$CB,$DF,$E7,$EF,$F4
	.dc.b	$E7,$EB,$EF,$FB,$FC,$8F,$81,$E4,$7C,$1F,$23,$E1,$79,$1F,$0F,$C8
	.dc.b	$F1,$FB,$FE,$4F,$67,$CB,$EC,$F9,$BD,$9F,$3F,$B3,$E8,$F6,$7D,$3E
	.dc.b	$CF,$AB,$D9,$F5,$FB,$3E,$CF,$67,$FD,$E5,$F4,$7F,$EC,$B3,$F4,$47
	.dc.b	$E8,$7B,$0B,$DF,$DF,$BF,$C3,$DF,$E3,$EF,$F2,$F7,$FE,$C7,$B1,$D3
	.dc.b	$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F
	.dc.b	$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$F0,$DE,$A7
	.dc.b	$AB,$DC,$74,$4F,$3F,$FA,$59,$FD,$28,$FC,$4A,$7F,$12,$EF,$DB,$E2
	.dc.b	$E7,$DB,$B3,$FE,$33,$3F,$C2,$87,$F8,$4C,$7B,$1A,$F5,$B7,$E6,$70
	.dc.b	$F3,$38,$F9,$9C,$BC,$CE,$7E,$67,$4E,$CE,$BB,$F6,$F6,$3B,$FB,$1E
	.dc.b	$0F,$63,$C7,$F4,$9F,$D3,$0F,$C8,$F0,$9F,$C8,$BF,$01,$F6,$37,$D7
	.dc.b	$E3,$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$F6,$1A,$F8
	.dc.b	$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F,$5F
	.dc.b	$CF,$EB,$FA,$3D,$7F,$4F,$AF,$F3,$3E,$97,$E6,$F2,$9F,$E6,$07,$E3
	.dc.b	$CF,$C7,$3F,$69,$7A,$DC,$7D,$6E,$5E,$B7,$F0,$5E,$BE,$9E,$B7,$5F
	.dc.b	$5B,$B7,$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$78,$FD,$6F,$27
	.dc.b	$AD,$E5,$F5,$BC,$DE,$B7,$9F,$D6,$F4,$7A,$DE,$9F,$5B,$D5,$EB,$7E
	.dc.b	$1B,$E9,$BF,$C2,$E7,$F8,$58,$FC,$69,$F8,$CE,$72,$F5,$F8,$7A,$FC
	.dc.b	$7D,$EF,$E3,$DE,$77,$3F,$7B,$A7,$BD,$F9,$2F,$C5,$FF,$C1,$67,$E0
	.dc.b	$0F,$BF,$91,$FB,$ED,$F4,$38,$7A,$1C,$7D,$0E,$5E,$87,$3F,$43,$A7
	.dc.b	$A1,$D7,$D0,$ED,$E7,$F7,$F3,$FC,$1E,$7F,$87,$CF,$F1,$79,$FE,$3F
	.dc.b	$3F,$C9,$E7,$FC,$1F,$6B,$E1,$7B,$5F,$0F,$DA,$FF,$57,$FA,$7F,$EB
	.dc.b	$0F,$33,$FD,$63,$D3,$FE,$B7,$FC,$C3,$FC,$D9,$F2,$47,$F9,$5B,$7E
	.dc.b	$03,$5E,$F6,$FA,$78,$74,$F1,$E9,$E5,$D3,$CF,$A7,$A7,$4F,$5E,$9E
	.dc.b	$DD,$3D,$FA,$7C,$1D,$3F,$5F,$DE,$FB,$1E,$F7,$8F,$AF,$ED,$7B,$DF
	.dc.b	$6F,$DE,$FB,$9E,$F7,$DD,$F7,$BE,$F7,$4F,$DF,$E9,$FC,$07,$4F,$F0
	.dc.b	$6F,$82,$FF,$1F,$9F,$B4,$3E,$A9,$F5,$07,$36,$F9,$B8,$73,$71,$E6
	.dc.b	$E5,$CD,$CF,$9B,$A7,$37,$5E,$6E,$DC,$DD,$F9,$BC,$1C,$9E,$1F,$03
	.dc.b	$C5,$E0,$78,$FC,$0F,$27,$BD,$E5,$F7,$BC,$DE,$F7,$F1,$CF,$A9,$FD
	.dc.b	$89,$FE,$1F,$3F,$9F,$1F,$B0,$F1,$4F,$EC,$1F,$E5,$6B,$DD,$E5,$EE
	.dc.b	$F3,$F7,$7A,$7B,$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF
	.dc.b	$17,$BB,$E3,$F7,$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE
	.dc.b	$FA,$7D,$DF,$E6,$BF,$55,$FE,$86,$3F,$D0,$A7,$F3,$C3,$F6,$07,$EB
	.dc.b	$F9,$7F,$A1,$AF,$73,$8F,$B9,$CB,$DC,$E7,$EE,$74,$F6,$FA,$FB,$7D
	.dc.b	$BD,$BE,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F
	.dc.b	$2F,$AF,$E6,$F5,$FC,$FE,$BF,$F4,$CF,$AB,$FF,$42,$9F,$CE,$8F,$DD
	.dc.b	$9F,$BA,$3F,$0F,$C9,$FD,$69,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7
	.dc.b	$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F
	.dc.b	$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5,$BF,$B1,$FC,$07,$F3,$83
	.dc.b	$FC,$CC,$7F,$81,$4F,$C4,$1F,$87,$E4,$FE,$6F,$EB,$70,$F5,$B8,$FA
	.dc.b	$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD,$DB,$D6,$EF,$EB,$78,$3D,$6F
	.dc.b	$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6,$F2,$FA,$DE,$6F,$5B,$CF,$EB
	.dc.b	$7F,$7A,$FA,$CF,$F6,$59,$FE,$CA,$3F,$C0,$27,$F3,$53,$F0,$DC,$87
	.dc.b	$F0,$CF,$F8,$29,$7A,$BC,$FD,$5E,$9E,$AF,$5F,$57,$B7,$AB,$DF,$D5
	.dc.b	$F0,$7A,$BE,$1F,$57,$C5,$EA,$F8,$FD,$5F,$27,$AB,$E5,$F5,$7C,$DE
	.dc.b	$AF,$9F,$D5,$F4,$7A,$BF,$E1,$2F,$AD,$FF,$65,$9F,$E0,$03,$FA,$09
	.dc.b	$FD,$00,$FC,$3F,$19,$FC,$3B,$FE,$22,$7F,$DA,$23,$D3,$E9,$E9,$F5
	.dc.b	$F4,$FB,$7A,$7D,$FD,$3F,$07,$A7,$E1,$F4,$FC,$5E,$9F,$8F,$D3,$F2
	.dc.b	$7A,$7E,$5F,$4F,$CD,$E9,$F9,$FD,$3F,$47,$A7,$FE,$2A,$FA,$EF,$EF
	.dc.b	$E7,$FA,$B4,$FF,$26,$1F,$C3,$CF,$C4,$1F,$C3,$78,$BF,$C7,$2B,$D2
	.dc.b	$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9,$78
	.dc.b	$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5E,$7F,$4B
	.dc.b	$D1,$E9,$7F,$90,$BD,$B7,$CC,$5F,$97,$C7,$C5,$E7,$DF,$D7,$C2,$EF
	.dc.b	$DE,$E1,$EC,$71,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1
	.dc.b	$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6
	.dc.b	$3C,$DE,$C7,$F9,$63,$EB,$DF,$2F,$7E,$47,$1F,$0B,$9F,$85,$D7,$66
	.dc.b	$FE,$0B,$87,$C1,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7
	.dc.b	$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$E5
	.dc.b	$F7,$BC,$DE,$F7,$F9,$BB,$EC,$3F,$C8,$67,$FB,$C4,$FF,$48,$1F,$D4
	.dc.b	$CF,$E1,$87,$E1,$7D,$F3,$F8,$57,$FC,$FE,$FF,$BC,$47,$B9,$D7,$DC
	.dc.b	$ED,$EE,$77,$F7,$3C,$1E,$E7,$87,$DC,$F1,$7B,$9E,$3F,$73,$C9,$EE
	.dc.b	$79,$7D,$CF,$37,$B9,$E7,$F7,$3D,$1E,$E7,$A7,$DC,$FF,$43,$FD,$8F
	.dc.b	$E1,$B8,$BF,$D2,$2B,$DC,$CF,$B9,$AF,$73,$7E,$E7,$0F,$73,$8F,$B9
	.dc.b	$CB,$DC,$E7,$EE,$74,$F7,$3A,$FB,$9D,$BD,$CE,$FE,$E7,$83,$DC,$F0
	.dc.b	$FB,$9E,$2F,$73,$C7,$EE,$79,$3D,$CF,$2F,$B9,$E6,$F7,$3C,$FE,$E7
	.dc.b	$FA,$7B,$8B,$F0,$FC,$FF,$EA,$87,$FD,$1A,$7F,$93,$8F,$F2,$69,$FC
	.dc.b	$38,$FE,$1A,$FF,$96,$FD,$5E,$5E,$AF,$3F,$57,$A7,$AB,$D7,$D5,$ED
	.dc.b	$EA,$F7,$F5,$7C,$1E,$AF,$87,$D5,$F1,$7A,$BE,$3F,$57,$C9,$EA,$F9
	.dc.b	$7D,$5F,$37,$AB,$E7,$F5,$7F,$D6,$BF,$65,$FE,$8A,$1F,$31,$8F,$95
	.dc.b	$CF,$C8,$EB,$E2,$37,$F1,$1C,$3E,$17,$8F,$C2,$F2,$F7,$39,$FB,$9D
	.dc.b	$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC,$F1
	.dc.b	$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$FE,$CC,$FB,$3F,$C3
	.dc.b	$79,$07,$F0,$CF,$FB,$61,$FF,$6A,$9F,$E8,$93,$FA,$A9,$FD,$51,$E1
	.dc.b	$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D
	.dc.b	$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1
	.dc.b	$EC,$7F,$B9,$3E,$D3,$F0,$FC,$47,$FD,$A6,$7F,$A5,$0F,$F4,$99,$FD
	.dc.b	$50,$FC,$38,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1,$DF
	.dc.b	$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6,$3C
	.dc.b	$DE,$C7,$9F,$D8,$FF,$7D,$FA,$7F,$ED,$03,$F0,$FC,$63,$FC,$B9,$EB
	.dc.b	$6B,$D6,$DF,$AD,$C2,$B8,$FA,$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD
	.dc.b	$DB,$D6,$EF,$EB,$78,$3D,$6F,$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6
	.dc.b	$F2,$FA,$DE,$6F,$5B,$CF,$EB,$7A,$3D,$6F,$37,$33,$ED,$7F,$12,$7E
	.dc.b	$21,$C7,$BB,$A8,$FF,$68,$1F,$E1,$C3,$F1,$07,$F8,$6A,$3E,$02,$F5
	.dc.b	$B8,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77
	.dc.b	$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B,$C9,$EB,$7A,$BE,$DB
	.dc.b	$F1,$23,$E1,$B1,$F0,$D9,$F8,$6D,$7C,$36,$FE,$1B,$87,$B7,$C7,$DB
	.dc.b	$E5,$ED,$F3,$F6,$FA,$7B,$7D,$7D,$BE,$DE,$DF,$7F,$6F,$C1,$ED,$F8
	.dc.b	$7D,$BF,$17,$B7,$E3,$F6,$FC,$9E,$DF,$97,$DB,$F3,$7B,$7F,$43,$ED
	.dc.b	$FE,$8F,$8E,$7F,$12,$DF,$73,$5E,$E6,$FD,$CE,$1E,$E7,$1F,$73,$97
	.dc.b	$B9,$CF,$DC,$E9,$EE,$75,$F7,$3B,$7B,$9D,$FD,$CF,$07,$B9,$E1,$F7
	.dc.b	$3C,$5E,$E7,$8F,$DC,$F2,$7B,$9E,$5F,$73,$CD,$EE,$79,$FD,$CF,$47
	.dc.b	$B9,$F8,$8F,$1F,$F8,$79,$F8,$71,$EE,$67,$E1,$35,$F0,$7B,$F8,$4E
	.dc.b	$1F,$03,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D,$FE,$DE,$FF,$7F
	.dc.b	$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E,$FF,$97,$DF,$F3
	.dc.b	$7B,$FF,$9E,$FB,$8F,$E1,$A7,$E1,$7C,$BF,$DB,$2F,$7B,$5E,$F6,$FD
	.dc.b	$EE,$1D,$5C,$7D,$EE,$5E,$F7,$3F,$7B,$A7,$BD,$D7,$DE,$ED,$EF,$77
	.dc.b	$F7,$BC,$1E,$F7,$87,$DE,$F1,$7B,$DE,$3F,$7B,$C9,$EF,$79,$7D,$EF
	.dc.b	$37,$BD,$E7,$F7,$BF,$6B,$F7,$3F,$86,$3F,$0A,$3D,$EC,$FB,$DA,$F7
	.dc.b	$B7,$EF,$70,$F7,$B8,$FB,$DC,$BD,$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB
	.dc.b	$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD,$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2
	.dc.b	$FB,$DE,$6F,$7B,$CF,$EF,$7F,$0C,$FB,$AF,$EA,$67,$F0,$B3,$F0,$BE
	.dc.b	$67,$F2,$55,$EF,$6F,$DE,$E1,$EF,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E
	.dc.b	$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD
	.dc.b	$EF,$27,$BD,$E5,$F7,$BC,$DE,$F7,$9F,$DE,$FE,$47,$C3,$FD,$48,$FC
	.dc.b	$37,$2F,$FB,$F7,$DD,$CF,$BB,$AF,$77,$7E,$EF,$0F,$77,$8F,$BB,$CB
	.dc.b	$DD,$E7,$EE,$F4,$F7,$7A,$FB,$BD,$BD,$DE,$FE,$EF,$83,$DD,$F0,$FB
	.dc.b	$BE,$2F,$77,$C7,$EE,$F9,$3D,$DF,$2F,$BB,$E6,$F7,$7F,$9C,$FD,$DF
	.dc.b	$F0,$A3,$F8,$49,$F8,$5E,$6F,$EA,$4B,$DC,$DF,$B9,$C3,$DC,$E3,$EE
	.dc.b	$72,$F7,$39,$FB,$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C
	.dc.b	$3E,$E7,$8B,$DC,$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9
	.dc.b	$E8,$F7,$3F,$A8,$FD,$E7,$F8,$60,$FE,$A0,$7E,$1B,$9C,$FE,$19,$FE
	.dc.b	$D8,$FF,$BB,$0F,$F0,$C1,$FD,$45,$DF,$B3,$D3,$D9,$EB,$EC,$F6,$F6
	.dc.b	$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE
	.dc.b	$CF,$9B,$D9,$F3,$FB,$3E,$8F,$67,$FB,$4F,$DE,$FF,$86,$0F,$C3,$9F
	.dc.b	$C2,$F8,$7F,$C0,$4B,$D8,$DF,$B1,$C3,$D8,$E3,$EC,$72,$F6,$39,$FB
	.dc.b	$1D,$3D,$8E,$BE,$C7,$6F,$63,$BF,$B1,$E0,$F6,$3C,$3E,$C7,$8B,$D8
	.dc.b	$F1,$FB,$1E,$4F,$63,$CB,$EC,$79,$BD,$8F,$3F,$B1,$FE,$00,$F5,$3F
	.dc.b	$A8,$9F,$D4,$0F,$C4,$73,$7F,$B9,$7D,$7D,$7A,$FB,$F5,$F8,$57,$1F
	.dc.b	$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F,$07,$AF
	.dc.b	$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB,$F9,$FD
	.dc.b	$7F,$F0,$BF,$DF,$7F,$97,$8F,$F0,$B1,$F8,$73,$F0,$DC,$C5,$EE,$F0
	.dc.b	$F7,$38,$FB,$BC,$BD,$AE,$7E,$D7,$4F,$6B,$AF,$B5,$DB,$DA,$EF,$ED
	.dc.b	$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7,$93,$DA,$F2,$FB,$5E,$6F
	.dc.b	$6B,$CF,$ED,$7F,$8B,$BE,$FF,$FC,$BA,$7F,$0C,$3F,$0D,$CA,$7F,$0C
	.dc.b	$FF,$92,$57,$B3,$C7,$D9,$E5,$EC,$F3,$F6,$7A,$7B,$3D,$7D,$97,$F0
	.dc.b	$0C,$2E,$5C,$E1,$F8,$07,$F9,$57,$E1,$9C,$FC,$63,$0D,$AF,$EF,$6D
	.dc.b	$9D,$C7,$39,$DB,$4D,$96,$36,$FB,$6B,$C2,$E8,$C4,$F2,$69,$9D,$53
	.dc.b	$C2,$B9,$C6,$2B,$83,$6B,$85,$6E,$FD,$04,$8F,$23,$5E,$F7,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF
	.dc.b	$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63
	.dc.b	$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF
	.dc.b	$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88
	.dc.b	$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A
	.dc.b	$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB
	.dc.b	$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B
	.dc.b	$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D
	.dc.b	$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B
	.dc.b	$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60
	.dc.b	$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC
	.dc.b	$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE
	.dc.b	$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78
	.dc.b	$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12
	.dc.b	$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71
	.dc.b	$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C
	.dc.b	$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C
	.dc.b	$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6
	.dc.b	$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09
	.dc.b	$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC
	.dc.b	$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE
	.dc.b	$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A
	.dc.b	$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12
	.dc.b	$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D
	.dc.b	$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12
	.dc.b	$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2
	.dc.b	$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C
	.dc.b	$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09
	.dc.b	$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA
	.dc.b	$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB
	.dc.b	$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E
	.dc.b	$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD
	.dc.b	$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9
	.dc.b	$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A
	.dc.b	$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46
	.dc.b	$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31
	.dc.b	$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9
	.dc.b	$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26
	.dc.b	$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12
	.dc.b	$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3
	.dc.b	$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C
	.dc.b	$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F
	.dc.b	$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18
	.dc.b	$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A
	.dc.b	$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C
	.dc.b	$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67
	.dc.b	$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB
	.dc.b	$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D
	.dc.b	$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38
	.dc.b	$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E
	.dc.b	$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97
	.dc.b	$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D
	.dc.b	$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57
	.dc.b	$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA
	.dc.b	$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3
	.dc.b	$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF
	.dc.b	$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17
	.dc.b	$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF
	.dc.b	$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E
	.dc.b	$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D
	.dc.b	$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35
	.dc.b	$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1
	.dc.b	$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D
	.dc.b	$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4
	.dc.b	$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F
	.dc.b	$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D2,$12,$95,$2F,$10,$BE,$15,$BA,$D4,$F1,$9C,$BC,$AB
	.dc.b	$8C,$62,$77,$3C,$EF,$C2,$EC,$E4,$E2,$37,$1E,$5F,$BB,$F3,$7C,$9B
	.dc.b	$E2,$BC,$B3,$91,$19,$73,$3C,$27,$70,$8C,$3A,$38,$C6,$A3,$31,$F0
	.dc.b	$9A,$F7,$9E,$CE,$4E,$27,$37,$2E,$E7,$A3,$A3,$84,$69,$F2,$19,$38
	.dc.b	$1C,$E3,$8C,$6E,$7C,$73,$C2,$30,$F1,$31,$3C,$08,$F3,$1F,$7C,$FD
	.dc.b	$5B,$E4,$FC,$BF,$E0,$FF,$01,$3F,$85,$9D,$61,$8B,$55,$9E,$58,$66
	.dc.b	$31,$87,$4A,$DE,$1C,$E1,$F1,$1B,$8C,$36,$7C,$B2,$F8,$5C,$63,$E7
	.dc.b	$AB,$2E,$6A,$63,$9A,$E3,$EB,$3C,$1C,$DC,$39,$B8,$B6,$35,$9F,$83
	.dc.b	$9B,$AF,$17,$66,$CE,$B4,$F9,$1F,$0B,$E2,$3C,$7F,$0B,$67,$5A,$BE
	.dc.b	$17,$CF,$F0,$BE,$86,$CE,$BE,$17,$D7,$F0,$BE,$C6,$C7,$2F,$85,$F7
	.dc.b	$FC,$2F,$C0,$6C,$76,$BE,$7B,$FC,$E3,$E4,$B8,$E6,$FA,$78,$74,$9D
	.dc.b	$6D,$F8,$18,$C4,$FD,$8C,$3A,$D6,$B4,$46,$27,$EE,$E1,$D7,$1F,$B8
	.dc.b	$DB,$FB,$38,$74,$9F,$C0,$36,$3E,$ED,$F3,$9F,$C2,$36,$3F,$03,$F0
	.dc.b	$BF,$7B,$E1,$7E,$FB,$6B,$E6,$FC,$2F,$E0,$8F,$C2,$FE,$AF,$FC,$6E
	.dc.b	$32,$7D,$75,$F2,$9A,$37,$F1,$B6,$7E,$BD,$F5,$C7,$E0,$37,$5F,$CD
	.dc.b	$E3,$DE,$BA,$55,$3E,$5E,$3D,$A3,$C1,$8F,$66,$9D,$8F,$C2,$E5,$EC
	.dc.b	$69,$E8,$AF,$A1,$7E,$C7,$B6,$1F,$AF,$C3,$B4,$78,$30,$EC,$DB,$EC
	.dc.b	$FD,$9C,$3B,$36,$3B,$5F,$B1,$F5,$98,$3F,$80,$5F,$68,$F0,$5F,$66
	.dc.b	$CF,$CD,$F8,$8F,$A3,$F1,$1F,$49,$B1,$F4,$7E,$23,$EA,$FC,$47,$D6
	.dc.b	$6C,$7D,$5F,$88,$FB,$27,$DA,$F8,$8B,$1D,$4F,$B5,$87,$56,$9D,$4F
	.dc.b	$BD,$7D,$5A,$75,$26,$9F,$5A,$BA,$BF,$63,$1F,$3B,$67,$AB,$F7,$31
	.dc.b	$F3,$B6,$3A,$BF,$80,$C7,$9E,$5F,$57,$0B,$5F,$2B,$E2,$3F,$2A,$7E
	.dc.b	$5B,$E2,$2C,$7E,$57,$E2,$3F,$32,$7E,$6B,$E2,$2C,$7E,$67,$E2,$3F
	.dc.b	$3B,$F1,$1F,$9E,$6C,$7E,$77,$E2,$3F,$43,$F1,$1F,$A2,$6C,$7E,$87
	.dc.b	$E2,$3F,$4B,$F1,$1F,$A6,$6C,$7E,$97,$E2,$3F,$52,$7E,$AB,$E2,$2C
	.dc.b	$7E,$A7,$E2,$3F,$5A,$7F,$32,$FD,$63,$FA,$2C,$64,$FE,$48,$BE,$5F
	.dc.b	$5E,$AE,$DB,$3A,$F8,$EE,$5F,$15,$CD,$B1,$CB,$E2,$BB,$1D,$FE,$2A
	.dc.b	$C7,$6F,$8A,$F1,$7C,$57,$8D,$B1,$E2,$F8,$AF,$37,$BD,$E7,$6C,$79
	.dc.b	$BE,$2B,$D4,$7A,$FE,$2A,$C7,$AB,$E2,$BD,$DF,$15,$EF,$6C,$7B,$BE
	.dc.b	$2B,$E1,$1F,$0F,$E2,$AC,$7C,$2F,$8A,$F8,$C7,$C7,$F8,$AB,$19,$27
	.dc.b	$8D,$FC,$CB,$99,$F1,$5A,$8C,$5D,$F1,$BF,$B7,$53,$E2,$B1,$87,$F7
	.dc.b	$5C,$BF,$B7,$57,$F7,$36,$C6,$7E,$17,$FB,$4C,$63,$1F,$F0,$0B,$FD
	.dc.b	$CA,$7F,$AE,$3F,$DB,$DB,$F0,$9C,$3E,$17,$8F,$C2,$C9,$DF,$C2,$DC
	.dc.b	$3F,$A5,$E0,$F8,$4E,$FF,$0B,$E0,$6D,$F6,$F8,$4F,$19,$E4,$F8,$5B
	.dc.b	$1E,$2F,$85,$F3,$9E,$8F,$85,$B1,$E6,$F8,$0F,$59,$EC,$F8,$5B,$1E
	.dc.b	$AF,$81,$F7,$9F,$03,$E1,$6C,$7B,$BE,$0B,$E1,$FC,$2F,$C4,$6C,$7C
	.dc.b	$2F,$82,$F8,$E7,$C8,$F8,$1B,$1F,$1B,$E0,$BF,$C2,$7F,$AC,$FF,$4F
	.dc.b	$E4,$F8,$31,$87,$3F,$05,$2B,$FC,$11,$4F,$84,$C5,$7F,$72,$87,$C0
	.dc.b	$F4,$3A,$FC,$1D,$BE,$3F,$05,$E0,$3C,$3F,$07,$63,$AF,$C1,$F9,$0F
	.dc.b	$2F,$C1,$D8,$F0,$FC,$1F,$A3,$E0,$FD,$2D,$8F,$2F,$C1,$FB,$0F,$6F
	.dc.b	$C1,$D8,$F4,$FC,$17,$C0,$3E,$0F,$C1,$D8,$F6,$FC,$17,$C4,$F8,$3F
	.dc.b	$8A,$D8,$F8,$3F,$05,$F2,$3E,$0F,$E4,$B6,$3E,$2F,$C1,$FC,$C3,$E6
	.dc.b	$FC,$1D,$8F,$93,$F0,$7F,$E4,$2F,$D6,$BF,$88,$F4,$7C,$98,$C4,$E7
	.dc.b	$E0,$E3,$FE,$37,$AF,$F4,$EB,$63,$97,$C1,$73,$F8,$3B,$3B,$7F,$CA
	.dc.b	$58,$7F,$A3,$FE,$0F,$C0,$63,$D3,$38,$AF,$F2,$4C,$ED,$F0,$7C,$1E
	.dc.b	$E7,$D0,$D8,$FF,$27,$FA,$AF,$DE,$B9,$C7,$2C,$2E,$1F,$E3,$BA,$D6
	.dc.b	$1F,$07,$E0,$FE,$17,$C2,$5C,$3D,$CF,$C5,$F8,$3F,$8D,$F0,$96,$7E
	.dc.b	$13,$F2,$7E,$12,$47,$C5,$F8,$4B,$9F,$F9,$13,$89,$F0,$7F,$3C,$FA
	.dc.b	$1F,$09,$6F,$E6,$7C,$27,$D3,$F6,$FE,$A3,$63,$E8,$7C,$17,$D7,$F8
	.dc.b	$4F,$B0,$D8,$FA,$9F,$09,$F6,$CF,$B9,$F0,$76,$3F,$CD,$DF,$AE,$7F
	.dc.b	$17,$F8,$93,$FE,$C1,$8C,$FC,$57,$FA,$12,$31,$BF,$F9,$71,$FE,$38
	.dc.b	$DF,$85,$D7,$C3,$6D,$E1,$F0,$D3,$0E,$FF,$0D,$6F,$97,$C2,$F5,$3B
	.dc.b	$7C,$35,$AF,$2F,$C3,$78,$7E,$1B,$C4,$D8,$F4,$FC,$37,$97,$E1,$BC
	.dc.b	$CD,$8F,$6F,$C3,$7A,$4F,$57,$C3,$58,$F8,$3F,$0D,$ED,$3D,$DF,$0D
	.dc.b	$63,$E2,$FC,$2F,$C1,$F8,$6F,$84,$D8,$F9,$3F,$0D,$F1,$4F,$8D,$F0
	.dc.b	$D6,$3E,$6F,$C3,$7F,$AB,$3D,$4F,$E3,$5E,$5F,$FA,$EF,$A5,$39,$E5
	.dc.b	$EB,$7F,$53,$3B,$1F,$3D,$E8,$FD,$1C,$BB,$CF,$88,$EE,$7F,$AA,$71
	.dc.b	$EE,$FF,$B3,$70,$EE,$E6,$BC,$9F,$0D,$E5,$F8,$6B,$3C,$5F,$47,$C2
	.dc.b	$FA,$7E,$16,$CF,$57,$D9,$F0,$B2,$3C,$1F,$0B,$71,$FF,$50,$E6,$F8
	.dc.b	$5F,$85,$F0,$BF,$0D,$B7,$E7,$F8,$4F,$8C,$7C,$7F,$85,$B1,$EB,$F8
	.dc.b	$5F,$95,$F0,$BF,$2D,$B1,$EF,$F8,$5F,$9D,$F0,$BF,$3D,$B1,$F0,$FE
	.dc.b	$13,$E9,$1F,$4F,$E1,$6C,$7C,$7D,$7E,$B1,$F5,$FE,$16,$C7,$CB,$F8
	.dc.b	$5F,$B5,$F0,$BF,$6D,$B1,$D4,$9C,$3F,$03,$0E,$AF,$C2,$C3,$9B,$F1
	.dc.b	$30,$BD,$9F,$EF,$D9,$7C,$2F,$FB,$CB,$F5,$DF,$E6,$33,$98,$C6,$3D
	.dc.b	$23,$86,$9E,$B8,$F0,$C7,$3B,$EB,$0C,$5E,$82,$59,$6A,$73,$97,$F9
	.dc.b	$36,$FF,$DE,$4F,$F3,$C8,$FF,$6D,$B9,$6F,$C4,$F9,$FE,$1A,$57,$0F
	.dc.b	$86,$B5,$C3,$E0,$FD,$9F,$0D,$ED,$6D,$76,$F8,$4F,$80,$7C,$1F,$84
	.dc.b	$B1,$E2,$C7,$9E,$E8,$F6,$6E,$D5,$A6,$A7,$31,$FC,$FA,$B8,$69,$EE
	.dc.b	$D3,$C5,$5E,$E7,$BC,$7B,$B3,$F1,$4F,$94,$F1,$12,$3E,$1E,$1D,$DF
	.dc.b	$93,$A7,$B1,$F8,$13,$D4,$F9,$7F,$09,$F7,$7E,$13,$EF,$36,$3A,$65
	.dc.b	$E6,$D3,$D9,$AF,$E0,$F2,$ED,$1E,$FC,$BB,$56,$F5,$F7,$61,$F0,$35
	.dc.b	$EA,$FD,$AC,$EE,$9F,$69,$E7,$4C,$FB,$57,$88,$F9,$38,$F6,$69,$EB
	.dc.b	$BF,$C1,$FC,$27,$E5,$CF,$CC,$7C,$25,$9E,$53,$F8,$CB,$EA,$FE,$29
	.dc.b	$FA,$2F,$E3,$32,$B9,$FA,$DF,$D1,$7C,$1F,$E8,$FE,$0E,$E1,$F9,$47
	.dc.b	$F4,$DF,$05,$23,$F2,$FF,$07,$71,$FC,$BF,$C1,$7E,$AC,$FD,$67,$C1
	.dc.b	$DA,$FC,$EF,$C1,$FE,$BF,$E0,$FF,$60,$D8,$FD,$0E,$3C,$F5,$F9,$D9
	.dc.b	$59,$FC,$6E,$3C,$F2,$FD,$2E,$56,$FE,$D7,$EB,$DF,$DB,$49,$F0,$DF
	.dc.b	$92,$8C,$FC,$2F,$3B,$FD,$8E,$7F,$C0,$73,$C7,$3F,$E7,$73,$E9,$A7
	.dc.b	$E7,$34,$B8,$7E,$F3,$4F,$EE,$79,$F8,$3E,$13,$B5,$FE,$9A,$72,$D8
	.dc.b	$CD,$71,$F8,$5B,$92,$DF,$0B,$87,$8D,$77,$8C,$56,$7E,$16,$E3,$FC
	.dc.b	$3F,$2F,$83,$1C,$7E,$16,$4E,$4F,$27,$C2,$E3,$15,$FC,$8F,$84,$F0
	.dc.b	$1E,$1F,$85,$B7,$C6,$BE,$07,$C2,$79,$7E,$17,$CC,$D9,$D6,$FE,$BF
	.dc.b	$85,$F5,$1E,$BF,$85,$B3,$AE,$17,$C2,$FB,$FE,$17,$E0,$36,$75,$F0
	.dc.b	$BF,$0F,$E0,$FE,$23,$63,$97,$C1,$FC,$73,$E4,$7C,$2D,$8E,$DF,$0B
	.dc.b	$F3,$BF,$5F,$FE,$EB,$C3,$F6,$63,$0E,$7E,$16,$57,$AA,$BF,$D3,$8D
	.dc.b	$8E,$5F,$09,$23,$5F,$09,$6F,$E5,$C7,$F2,$6B,$7C,$27,$84,$F1,$7C
	.dc.b	$25,$9E,$DF,$07,$E5,$3C,$DF,$09,$63,$C5,$F0,$9E,$93,$D5,$F0,$96
	.dc.b	$3C,$DF,$07,$ED,$F8,$4F,$73,$63,$D5,$F0,$7F,$07,$E1,$3E,$13,$63
	.dc.b	$DD,$F0,$1F,$17,$E0,$BE,$33,$63,$A9,$F0,$F0,$EA,$D3,$A9,$F1,$EF
	.dc.b	$AB,$4E,$A7,$CB,$AE,$AD,$3A,$F0,$7C,$83,$FB,$3F,$02,$FB,$C6,$7B
	.dc.b	$93,$9A,$F0,$CE,$61,$87,$8A,$30,$FE,$07,$E0,$65,$7D,$DF,$81,$B5
	.dc.b	$C7,$E0,$FF,$0D,$F0,$9F,$87,$6D,$7E,$0F,$E1,$3F,$6F,$DD,$FE,$06
	.dc.b	$8F,$BD,$18,$73,$F1,$3A,$F8,$5B,$1F,$A4,$AF,$EF,$9E,$57,$F4,$F7
	.dc.b	$FA,$14,$67,$DC,$EB,$F0,$5D,$9E,$FF,$01,$27,$97,$BD,$76,$F8,$98
	.dc.b	$3E,$13,$CA,$79,$BE,$16,$DF,$52,$57,$0C,$BB,$47,$97,$E1,$7D,$BF
	.dc.b	$0B,$EE,$6C,$7A,$7E,$13,$E0,$9F,$0B,$E1,$2C,$74,$DF,$D9,$BF,$AA
	.dc.b	$30,$7F,$82,$35,$FF,$58,$BF,$0F,$E1,$3E,$5F,$C2,$7C,$C7,$E6,$FC
	.dc.b	$07,$CE,$F8,$0B,$5F,$25,$FA,$3F,$01,$23,$E6,$7C,$05,$E1,$F8,$7F
	.dc.b	$07,$F5,$8F,$AF,$F0,$96,$BF,$8A,$FE,$C1,$FE,$B1,$27,$C2,$FF,$32
	.dc.b	$7F,$D6,$9B,$32,$E9,$C1,$F5,$F3,$B8,$7F,$88,$AF,$F9,$AF,$86,$5F
	.dc.b	$09,$CF,$E0,$FA,$3D,$7E,$16,$5F,$8B,$4B,$BF,$DA,$D3,$AB,$FE,$48
	.dc.b	$D5,$A6,$34,$ED,$9F,$4C,$7D,$6D,$BF,$D3,$69,$FD,$2A,$7D,$FF,$09
	.dc.b	$ED,$F8,$5F,$73,$CF,$2F,$74,$FC,$A3,$F2,$D9,$FA,$5F,$8F,$B5,$C7
	.dc.b	$AB,$F1,$FE,$16,$4F,$CC,$F8,$5B,$5E,$B8,$E6,$75,$70,$79,$BE,$17
	.dc.b	$E8,$7C,$2F,$D1,$6C,$FD,$6F,$84,$FA,$87,$D5,$F8,$5B,$1D,$2F,$9F
	.dc.b	$2F,$D5,$8C,$3F,$E3,$BE,$06,$1D,$2B,$C1,$5F,$82,$C7,$9C,$7C,$8C
	.dc.b	$7B,$4F,$84,$E6,$7E,$67,$5F,$E0,$3C,$7F,$4F,$1E,$B3,$CC,$FC,$6E
	.dc.b	$3D,$F2,$F1,$7C,$2F,$E3,$7E,$17,$F1,$CD,$AE,$B9,$FD,$FC,$BA,$C6
	.dc.b	$39,$7F,$19,$F0,$7F,$95,$3F,$2D,$F0,$96,$7F,$37,$F0,$9F,$99,$E5
	.dc.b	$FC,$D3,$63,$F3,$FF,$09,$F9,$DF,$84,$FC,$F3,$63,$F4,$7F,$09,$FA
	.dc.b	$13,$FC,$27,$FB,$0F,$F6,$6C,$FF,$39,$18,$F8,$5C,$FC,$45,$BF,$EF
	.dc.b	$F7,$FD,$9B,$7F,$C8,$E3,$D5,$B3,$FB,$0C,$BA,$FC,$37,$6F,$86,$B3
	.dc.b	$C9,$F0,$FC,$37,$8B,$E1,$AC,$F6,$7D,$2F,$33,$DA,$5C,$3F,$A3,$65
	.dc.b	$FE,$E1,$8F,$2F,$C3,$7B,$3E,$1B,$DA,$FF,$8C,$A9,$F8,$23,$DC,$C3
	.dc.b	$6F,$DB,$F0,$F8,$99,$1F,$E2,$3A,$E3,$1F,$23,$E1,$F3,$1F,$D6,$CC
	.dc.b	$CE,$EF,$A4,$79,$BE,$1F,$7D,$BF,$E9,$1D,$3F,$AE,$4F,$AB,$E1,$FE
	.dc.b	$9F,$C2,$FD,$47,$EA,$FC,$27,$D6,$F8,$7B,$9F,$D2,$7E,$CF,$C3,$FD
	.dc.b	$AF,$86,$B3,$F5,$9F,$BB,$F0,$D2,$3E,$CF,$C3,$5C,$FE,$4F,$0F,$E0
	.dc.b	$8F,$C1,$FC,$35,$AF,$F1,$EF,$EC,$5F,$D8,$6B,$FD,$BA,$FF,$B5,$78
	.dc.b	$3F,$D5,$4E,$7C,$03,$9A,$FF,$64,$BF,$D8,$9B,$EE,$7F,$98,$A3,$17
	.dc.b	$D2,$31,$F0,$39,$F8,$5B,$7E,$27,$87,$C2,$F1,$F8,$5B,$3E,$67,$A7
	.dc.b	$C2,$C8,$F4,$FC,$25,$CB,$FC,$19,$9B,$E1,$7C,$47,$8F,$E1,$6D,$FC
	.dc.b	$0F,$85,$F3,$7C,$27,$9D,$B1,$F1,$3E,$17,$D4,$7A,$FE,$16,$C7,$C8
	.dc.b	$F8,$5F,$77,$A7,$EF,$6C,$7C,$CF,$63,$E1,$1F,$0F,$E1,$6C,$7D,$0F
	.dc.b	$85,$FF,$43,$DB,$E1,$7E,$44,$61,$CF,$C2,$CA,$FF,$39,$D7,$F5,$61
	.dc.b	$1C,$BE,$0A,$4E,$BE,$12,$F1,$FF,$8B,$ED,$F0,$9E,$03,$C3,$F0,$96
	.dc.b	$FA,$FC,$27,$90,$F2,$FC,$05,$8F,$0F,$C1,$FA,$0F,$4F,$C1,$D8,$F2
	.dc.b	$FC,$27,$B3,$E0,$BD,$AD,$8F,$4F,$C1,$FC,$0F,$83,$F8,$2D,$8F,$6F
	.dc.b	$C1,$FC,$43,$E2,$FC,$15,$8F,$83,$F0,$7F,$20,$F9,$3F,$01,$63,$E2
	.dc.b	$FC,$07,$CC,$3E,$6F,$C1,$D8,$FF,$4D,$FE,$C7,$FC,$8C,$9F,$09,$FE
	.dc.b	$B8,$71,$F0,$94,$C7,$C2,$5A,$E3,$F0,$9B,$F8,$4E,$0D,$8E,$BF,$09
	.dc.b	$CC,$E9,$F0,$76,$3C,$3F,$09,$DF,$E1,$3C,$0D,$8F,$2F,$C2,$78,$CF
	.dc.b	$27,$C1,$D8,$F4,$FC,$17,$9C,$F4,$7C,$15,$8F,$6F,$C1,$7A,$FE,$0B
	.dc.b	$D8,$D8,$F8,$3F,$05,$EF,$F8,$2F,$80,$D8,$F8,$BF,$05,$F0,$CF,$89
	.dc.b	$F0,$56,$3E,$4F,$C1,$7C,$7F,$82,$F9,$0D,$8F,$F6,$3F,$EC,$9F,$C1
	.dc.b	$C9,$F1,$3F,$EE,$47,$1F,$13,$4C,$7C,$1D,$AE,$3F,$07,$B3,$87,$C1
	.dc.b	$D8,$EB,$F0,$7C,$CE,$9F,$07,$63,$C3,$F0,$7D,$FE,$0F,$C0,$D8,$F2
	.dc.b	$FC,$17,$8C,$F2,$7C,$15,$8F,$4F,$C0,$F9,$CF,$47,$C1,$58,$F6,$FC
	.dc.b	$17,$AC,$F6,$7C,$15,$8F,$83,$F0,$5E,$F3,$E0,$7B,$56,$3E,$2F,$C1
	.dc.b	$7C,$33,$E2,$7C,$15,$8F,$93,$F0,$5F,$1C,$3F,$EF,$2F,$D9,$7F,$83
	.dc.b	$3A,$31,$F0,$5A,$F8,$0F,$F5,$64,$FF,$96,$8F,$E0,$B3,$C7,$E2,$B5
	.dc.b	$F0,$FB,$F8,$75,$7F,$97,$AB,$B7,$C2,$69,$60,$AD,$FC,$2E,$63,$17
	.dc.b	$FE,$55,$7C,$3F,$0B,$9A,$EA,$D9,$7C,$2D,$8C,$7C,$2E,$7E,$16,$E5
	.dc.b	$EF,$78,$7C,$2F,$1F,$85,$B3,$F0,$DE,$9F,$09,$23,$E3,$7C,$2D,$D7
	.dc.b	$E1,$FC,$27,$84,$F1,$7C,$2D,$AF,$99,$F0,$BE,$5F,$85,$F3,$36,$3E
	.dc.b	$87,$C2,$FA,$4F,$57,$C2,$D8,$FA,$9F,$0B,$ED,$3D,$DF,$0B,$63,$E4
	.dc.b	$F4,$7C,$FF,$4F,$FC,$64,$4F,$85,$F8,$CE,$3E,$12,$98,$F8,$4B,$7C
	.dc.b	$8F,$A7,$D4,$A7,$C1,$F2,$F8,$4E,$6D,$AE,$1C,$7F,$66,$9F,$09,$E0
	.dc.b	$3C,$3F,$09,$6B,$78,$F5,$F8,$3F,$2F,$C2,$79,$9B,$3A,$C1,$F0,$9E
	.dc.b	$AF,$84,$F5,$B6,$75,$F0,$9E,$E3,$DF,$F0,$96,$39,$7C,$1F,$C2,$F8
	.dc.b	$4F,$86,$D8,$ED,$F0,$7F,$18,$F8,$FF,$09,$63,$C5,$9F,$CA,$F8,$4F
	.dc.b	$96,$D8,$F3,$7C,$1F,$CE,$F8,$4F,$9E,$D8,$F7,$FE,$CD,$FE,$1D,$27
	.dc.b	$C2,$FE,$65,$C7,$C2,$5F,$84,$C3,$6B,$8F,$C2,$6F,$E1,$38,$36,$3A
	.dc.b	$FC,$27,$33,$A7,$C2,$58,$F0,$FC,$1F,$7F,$84,$F0,$36,$3C,$BF,$09
	.dc.b	$E3,$F8,$4F,$23,$63,$D3,$F0,$9E,$7F,$84,$F4,$36,$3D,$BF,$09,$EB
	.dc.b	$3D,$9F,$09,$63,$E0,$FC,$27,$BF,$E1,$3E,$03,$63,$E2,$FC,$27,$C3
	.dc.b	$F8,$4F,$88,$D8,$F9,$3F,$09,$F1,$CF,$E0,$1F,$B3,$FF,$16,$F2,$9E
	.dc.b	$87,$F0,$FE,$95,$6B,$E1,$77,$F0,$B7,$1F,$C8,$E3,$FE,$F2,$3F,$8A
	.dc.b	$63,$CA,$3F,$43,$18,$C7,$F8,$D3,$FC,$45,$BF,$09,$E2,$F8,$29,$5D
	.dc.b	$3E,$0E,$E1,$CC,$FE,$39,$83,$E0,$FD,$3F,$07,$EA,$6D,$F9,$3E,$0F
	.dc.b	$DA,$7B,$BE,$0E,$C7,$A3,$E0,$FE,$0F,$C1,$FC,$26,$C7,$B3,$E0,$FE
	.dc.b	$2F,$C1,$FC,$66,$C7,$C0,$F8,$2F,$93,$F0,$7F,$29,$B1,$F1,$3E,$0F
	.dc.b	$E6,$FC,$1F,$CE,$6C,$7C,$8F,$83,$FA,$3F,$07,$F4,$9B,$1F,$BC,$FD
	.dc.b	$A3,$FD,$84,$3E,$13,$F9,$B3,$86,$FC,$24,$9E,$1F,$09,$79,$BE,$47
	.dc.b	$C0,$30,$FE,$23,$5D,$7E,$13,$9F,$C2,$74,$7A,$FC,$24,$9F,$17,$C2
	.dc.b	$5C,$79,$E0,$F8,$3F,$19,$E4,$F8,$4B,$7E,$9F,$83,$F3,$FC,$27,$A1
	.dc.b	$B1,$ED,$F8,$4F,$59,$EC,$F8,$4B,$1F,$07,$E0,$FD,$E7,$C0,$F8,$4B
	.dc.b	$1F,$17,$E1,$3E,$1F,$C2,$7C,$46,$C7,$C9,$F8,$3F,$8F,$F0,$9F,$21
	.dc.b	$B1,$F3,$7E,$0F,$E5,$FC,$27,$CC,$6C,$7F,$2C,$EE,$FF,$55,$1F,$EA
	.dc.b	$9E,$3F,$ED,$8E,$3E,$13,$FB,$43,$FE,$A8,$E9,$2F,$84,$CF,$C2,$69
	.dc.b	$DF,$C1,$F0,$F8,$4B,$97,$77,$9F,$C2,$74,$F8,$4B,$3E,$37,$BF,$C1
	.dc.b	$C8,$F3,$7C,$25,$DB,$EF,$E9,$FE,$AB,$9F,$57,$C2,$79,$BE,$13,$CE
	.dc.b	$FA,$3E,$0E,$4F,$BF,$E1,$2E,$1C,$BE,$13,$DA,$7B,$BE,$12,$D7,$C5
	.dc.b	$F8,$3F,$82,$7C,$2F,$84,$B1,$F2,$7E,$0F,$E2,$9F,$1B,$E1,$2C,$7C
	.dc.b	$DF,$83,$F9,$27,$F8,$47,$F6,$9F,$E2,$5F,$05,$FE,$1F,$E1,$FE,$23
	.dc.b	$5F,$D7,$A7,$FC,$4B,$AF,$E7,$1B,$5F,$E2,$55,$F0,$7C,$FE,$13,$A7
	.dc.b	$C2,$4C,$39,$7C,$3D,$C7,$97,$C2,$F8,$7E,$17,$C4,$DA,$F0,$7C,$1F
	.dc.b	$94,$F3,$7C,$25,$8F,$27,$C2,$7A,$4F,$57,$C2,$58,$F4,$7C,$1F,$B7
	.dc.b	$E1,$3D,$CD,$8F,$67,$C1,$FC,$13,$E1,$7C,$25,$8F,$81,$F0,$7F,$14
	.dc.b	$F8,$DF,$09,$63,$E2,$7C,$27,$C9,$3E,$57,$3D,$8F,$91,$F0,$7F,$34
	.dc.b	$F9,$DF,$09,$63,$FC,$13,$FB,$57,$F9,$21,$BC,$9F,$E4,$38,$FE,$21
	.dc.b	$18,$37,$F0,$99,$F8,$4D,$36,$39,$FC,$27,$1F,$84,$E4,$D8,$EF,$F0
	.dc.b	$9D,$4E,$DF,$09,$63,$C7,$F0,$7E,$13,$C5,$F0,$76,$3C,$FF,$05,$E5
	.dc.b	$3C,$DF,$07,$63,$D7,$F0,$7E,$93,$D5,$F0,$76,$3D,$FF,$07,$ED,$F8
	.dc.b	$3F,$73,$63,$E1,$FC,$17,$C1,$3E,$17,$C1,$D8,$F8,$FF,$05,$F1,$7E
	.dc.b	$0F,$E3,$36,$3E,$5F,$C1,$7C,$93,$E5,$7C,$1D,$8F,$F1,$BF,$ED,$7F
	.dc.b	$E8,$66,$FC,$27,$F9,$C6,$31,$D5,$FC,$E1,$FE,$21,$F0,$87,$F8,$71
	.dc.b	$FC,$36,$B9,$FC,$27,$1F,$84,$E5,$F0,$9C,$FE,$12,$D7,$81,$ED,$F0
	.dc.b	$92,$3C,$7F,$09,$73,$F0,$7C,$27,$8F,$E1,$3C,$8D,$AF,$4F,$C2,$79
	.dc.b	$CF,$47,$C2,$58,$F6,$FC,$27,$AC,$F6,$7C,$25,$8F,$83,$F0,$9E,$FF
	.dc.b	$84,$F8,$0D,$8F,$8B,$F0,$9F,$0C,$F8,$9F,$09,$63,$E4,$FC,$27,$C7
	.dc.b	$3E,$47,$C2,$58,$F9,$BF,$09,$FE,$9B,$9F,$E1,$71,$FE,$72,$E7,$FE
	.dc.b	$17,$C3,$FD,$92,$72,$FF,$A9,$DF,$E1,$73,$FE,$78,$7F,$85,$BF,$E6
	.dc.b	$E7,$31,$BF,$82,$ED,$F0,$3D,$FE,$06,$DE,$DF,$17,$C0,$78,$FE,$06
	.dc.b	$CF,$37,$CD,$F0,$32,$3B,$7C,$0D,$CB,$CD,$F0,$3E,$B3,$D9,$F0,$36
	.dc.b	$BC,$9F,$01,$EF,$F8,$1F,$80,$D8,$F4,$7C,$0F,$C3,$F8,$1F,$88,$D8
	.dc.b	$F6,$7C,$0F,$C7,$F8,$1F,$90,$D8,$F8,$1F,$03,$F2,$FE,$07,$E6,$36
	.dc.b	$3E,$27,$C0,$FC,$F3,$E8,$7C,$0D,$8F,$F4,$AF,$ED,$9F,$C3,$78,$3F
	.dc.b	$C2,$6F,$F0,$C6,$9F,$ED,$17,$1F,$0B,$FE,$DF,$5F,$03,$93,$5F,$05
	.dc.b	$6F,$AF,$C0,$F1,$39,$7C,$15,$8F,$0F,$C1,$75,$3B,$7C,$15,$8F,$2F
	.dc.b	$C1,$78,$4F,$17,$C1,$58,$F4,$FC,$17,$97,$E0,$BC,$CD,$8F,$6F,$C1
	.dc.b	$7A,$4F,$57,$C1,$58,$F8,$3F,$03,$ED,$3D,$DF,$05,$63,$E2,$FC,$17
	.dc.b	$C1,$F8,$2F,$84,$D8,$F9,$3F,$03,$F1,$7E,$0B,$E3,$36,$3E,$6F,$C1
	.dc.b	$7C,$93,$E5,$7C,$15,$8F,$F6,$2F,$ED,$BF,$C3,$3A,$51,$FE,$D4,$8F
	.dc.b	$E1,$A6,$5C,$7C,$1E,$BA,$FF,$C9,$33,$FE,$1D,$27,$C0,$E6,$F5,$F0
	.dc.b	$69,$D7,$C1,$67,$1E,$DF,$05,$71,$5E,$EB,$0F,$19,$B6,$F8,$2B,$F0
	.dc.b	$58,$6E,$3E,$CF,$81,$D9,$C3,$E0,$AC,$7C,$0F,$82,$E7,$F0,$5D,$1B
	.dc.b	$1F,$13,$E0,$BB,$FC,$17,$81,$B1,$F2,$3E,$07,$C7,$F0,$5E,$46,$C7
	.dc.b	$CC,$F8,$2F,$3F,$C1,$7A,$1B,$1F,$43,$DF,$F5,$FC,$17,$B1,$B1,$F5
	.dc.b	$3E,$0B,$E0,$FE,$DD,$FC,$4B,$97,$E5,$46,$3E,$3F,$3F,$0F,$AF,$84
	.dc.b	$B8,$7F,$C0,$0C,$3F,$8A,$F5,$7F,$9F,$9F,$F5,$74,$23,$E7,$C6,$1B
	.dc.b	$F0,$7E,$0F,$84,$F0,$FC,$24,$BE,$9F,$09,$6F,$A7,$C2,$79,$8F,$3F
	.dc.b	$C2,$5A,$F1,$7C,$27,$AB,$E1,$3D,$6D,$8F,$37,$C1,$FB,$BE,$13,$DE
	.dc.b	$D8,$F5,$7C,$1F,$C2,$3E,$1F,$C2,$58,$F7,$7C,$1F,$C6,$F8,$4F,$8E
	.dc.b	$D8,$F8,$5F,$07,$F2,$8F,$97,$F0,$96,$3E,$37,$C2,$7C,$EF,$84,$F9
	.dc.b	$ED,$8F,$8D,$E6,$FE,$6B,$D3,$FE,$2C,$DF,$8A,$FC,$AC,$62,$7E,$67
	.dc.b	$72,$3F,$8B,$C6,$1F,$E2,$EF,$43,$1F,$07,$CB,$E1,$39,$BD,$3E,$13
	.dc.b	$AF,$C2,$5A,$E1,$D3,$FA,$6C,$13,$E2,$F8,$49,$5A,$CB,$C5,$F0,$98
	.dc.b	$C3,$FD,$7F,$87,$FA,$D6,$B4,$D8,$CF,$C1,$FB,$3E,$13,$DB,$F0,$7E
	.dc.b	$EF,$84,$B8,$70,$7A,$70,$FC,$5F,$84,$B3,$EF,$79,$3D,$AF,$E3,$E1
	.dc.b	$8C,$7C,$78,$F5,$AF,$1B,$63,$FD,$F8,$3F,$09,$F9,$93,$F0,$A3,$1E
	.dc.b	$FF,$F0,$97,$5A,$5A,$4D,$8F,$E1,$D9,$BF,$2D,$79,$B9,$C7,$E6,$CE
	.dc.b	$63,$18,$A3,$F8,$E7,$E5,$CE,$EC,$C6,$1F,$C1,$FC,$3F,$E1,$3E,$1E
	.dc.b	$E3,$F3,$9F,$C3,$FC,$3F,$E2,$3E,$1A,$CF,$D2,$7F,$17,$F0,$F2,$3E
	.dc.b	$AF,$C3,$DD,$3E,$CE,$FF,$90,$E0,$FC,$8B,$6B,$ED,$FC,$3F,$F2,$8F
	.dc.b	$DB,$FF,$9C,$5F,$E1,$E3,$13,$9F,$88,$D7,$C5,$D8,$FC,$45,$FF,$B5
	.dc.b	$F0,$F9,$31,$F8,$6C,$7B,$B4,$EE,$70,$3F,$47,$87,$67,$9E,$1D,$9A
	.dc.b	$76,$C3,$CD,$F1,$3E,$7F,$89,$B3,$E2,$7D,$5F,$13,$EB,$F8,$7B,$3E
	.dc.b	$67,$DD,$F1,$1E,$FF,$88,$B3,$EA,$79,$EF,$F6,$63,$F6,$BB,$B1,$CF
	.dc.b	$3F,$F5,$68,$C4,$FC,$2C,$BC,$0D,$3C,$07,$C6,$C7,$C0,$D3,$C0,$7A
	.dc.b	$4F,$97,$F1,$1E,$D6,$9C,$F1,$F2,$1F,$D8,$2F,$FA,$E4,$67,$0F,$26
	.dc.b	$7E,$CA,$DF,$86,$B6,$99,$BF,$53,$F8,$1F,$87,$FC,$17,$C4,$5C,$BE
	.dc.b	$EB,$F8,$5F,$87,$FC,$37,$C4,$59,$FC,$0B,$F8,$9F,$87,$91,$F8,$4F
	.dc.b	$88,$BA,$F8,$B4,$63,$F8,$FF,$87,$EF,$7F,$8A,$7A,$56,$F1,$BA,$FC
	.dc.b	$F7,$9B,$F4,$72,$FC,$C6,$58,$C3,$EB,$FC,$47,$E6,$CF,$CE,$78,$76
	.dc.b	$3F,$2F,$F1,$1F,$9F,$3F,$41,$F1,$16,$3F,$91,$7E,$E1,$FD,$A4,$FF
	.dc.b	$3A,$F1,$9C,$7C,$47,$F8,$39,$FE,$CF,$A7,$F8,$61,$B5,$E2,$39,$FC
	.dc.b	$47,$5D,$FE,$16,$5E,$09,$F1,$9E,$0F,$89,$E9,$8F,$56,$C7,$93,$E1
	.dc.b	$FC,$07,$87,$E2,$2C,$7A,$3E,$23,$C9,$F1,$1E,$56,$C7,$B3,$E2,$3D
	.dc.b	$1F,$11,$E9,$6C,$75,$3E,$0F,$C3,$F4,$D7,$F1,$3B,$A3,$A4,$FC,$6C
	.dc.b	$6D,$69,$E9,$1F,$2B,$0B,$3C,$9F,$91,$F1,$1F,$27,$E1,$ED,$7D,$07
	.dc.b	$9F,$17,$CB,$C2,$CF,$F6,$6C,$91,$D2,$7E,$86,$16,$B4,$FD,$5F,$88
	.dc.b	$FA,$DF,$11,$67,$EE,$BF,$67,$D3,$91,$F8,$0F,$87,$BA,$FC,$0F,$88
	.dc.b	$FB,$C7,$DF,$F8,$8B,$5F,$86,$F8,$8F,$C1,$7C,$3F,$E0,$DB,$1F,$8A
	.dc.b	$C7,$9E,$BF,$85,$CA,$CF,$AF,$1E,$79,$7C,$EC,$AD,$FE,$2F,$E2,$3F
	.dc.b	$CB,$7F,$B8,$FF,$81,$EB,$FC,$BF,$C2,$9C,$C6,$BE,$26,$65,$FD,$EE
	.dc.b	$FF,$C1,$33,$FC,$09,$FF,$06,$BF,$E6,$85,$F1,$1D,$8E,$FF,$11,$6B
	.dc.b	$97,$C4,$78,$8F,$1F,$C4,$58,$ED,$8F,$83,$1E,$DF,$13,$75,$FF,$3D
	.dc.b	$E7,$DA,$FC,$99,$F6,$69,$E7,$F8,$9F,$7F,$C3,$FC,$06,$C7,$AF,$E1
	.dc.b	$FE,$19,$F1,$3E,$1E,$C7,$BF,$E2,$7E,$39,$F2,$3E,$26,$C7,$C3,$F8
	.dc.b	$6F,$96,$7C,$CF,$86,$B1,$D4,$9B,$FC,$7C,$3A,$BF,$27,$0E,$6F,$CB
	.dc.b	$C2,$ED,$EE,$F8,$6F,$AE,$7D,$8F,$86,$B8,$F4,$9F,$87,$A7,$A6,$30
	.dc.b	$7F,$90,$B8,$19,$74,$8C,$63,$72,$FB,$1F,$0D,$F8,$43,$F0,$BF,$0D
	.dc.b	$6B,$F0,$1F,$0D,$F8,$8F,$86,$FC,$4B,$63,$F0,$9F,$0D,$FE,$A6,$F0
	.dc.b	$BF,$BF,$FA,$4D,$53,$9A,$D7,$C3,$47,$FC,$8D,$6F,$86,$C6,$BF,$8D
	.dc.b	$CB,$B3,$6B,$F2,$59,$FE,$6B,$1E,$91,$E2,$39,$9E,$3F,$86,$8F,$6F
	.dc.b	$86,$B7,$FD,$6B,$27,$C3,$7A,$7E,$13,$D4,$DB,$F2,$FC,$37,$B4,$F7
	.dc.b	$7C,$2D,$8F,$4F,$C3,$7C,$1F,$86,$F8,$4D,$8F,$6F,$C2,$7C,$5F,$86
	.dc.b	$F8,$CD,$8F,$83,$F0,$9F,$27,$E1,$3E,$53,$63,$E2,$FC,$27,$CD,$3E
	.dc.b	$77,$C2,$D8,$EA,$7C,$AC,$3A,$B4,$EA,$7C,$EB,$EA,$D3,$E8,$FC,$2F
	.dc.b	$D9,$F8,$5F,$B4,$D8,$FA,$BF,$07,$F7,$4F,$BD,$F0,$B6,$3F,$CF,$DF
	.dc.b	$B9,$7F,$B1,$49,$F0,$BF,$EE,$87,$1F,$0D,$7E,$17,$0D,$AE,$3F,$0B
	.dc.b	$BF,$85,$E0,$D8,$EB,$F0,$3C,$CE,$9F,$0B,$63,$C3,$F0,$5D,$FE,$17
	.dc.b	$C0,$D8,$F2,$FC,$2F,$8F,$E1,$7C,$8D,$8F,$4F,$C2,$F9,$CF,$47,$C2
	.dc.b	$D8,$F6,$FC,$27,$AF,$E1,$3D,$8D,$8F,$83,$F0,$9E,$F3,$E0,$7C,$25
	.dc.b	$8F,$8B,$F0,$BF,$0F,$C6,$F8,$8D,$8F,$93,$F0,$9F,$1C,$3F,$EA,$9F
	.dc.b	$DC,$FF,$9B,$7A,$4E,$8F,$E6,$D3,$9F,$82,$DF,$97,$FE,$34,$BF,$E6
	.dc.b	$76,$7F,$9B,$FB,$3A,$8C,$57,$08,$C4,$FF,$2B,$AE,$5F,$0B,$9C,$3F
	.dc.b	$D9,$73,$9A,$DC,$E6,$98,$77,$F8,$5C,$37,$0F,$57,$C2,$E3,$1F,$F6
	.dc.b	$3C,$7B,$7E,$0B,$06,$7E,$0E,$C7,$C1,$F7,$B8,$1C,$7E,$06,$C7,$C5
	.dc.b	$F7,$FA,$1D,$7D,$FB,$1F,$27,$E0,$FC,$1F,$07,$E1,$6C,$7C,$DF,$83
	.dc.b	$F2,$1E,$5F,$83,$B1,$F4,$7E,$0F,$D0,$7A,$7E,$0E,$C7,$D5,$F8,$0F
	.dc.b	$61,$ED,$F8,$0B,$1F,$67,$E0,$3E,$01,$F1,$3F,$74,$FE,$4B,$C7,$FC
	.dc.b	$92,$7F,$DC,$D0,$F8,$3D,$7C,$1E,$DB,$3A,$F8,$3E,$5F,$07,$CD,$B1
	.dc.b	$CB,$E0,$BB,$7C,$1F,$76,$C7,$6F,$83,$F1,$7C,$1F,$8D,$B1,$E2,$F8
	.dc.b	$2F,$31,$E7,$F8,$3B,$1E,$6F,$83,$F5,$7C,$07,$AD,$B1,$EA,$F8,$3F
	.dc.b	$77,$C0,$FB,$DB,$1E,$EF,$83,$F8,$47,$C3,$F8,$3B,$1F,$0B,$E0,$BE
	.dc.b	$31,$F1,$FE,$0E,$C7,$C6,$F8,$3F,$95,$F0,$7F,$2D,$B0,$F2,$FF,$D2
	.dc.b	$B0,$F8,$8F,$C9,$BF,$E7,$03,$13,$BF,$3B,$F4,$91,$9F,$82,$D9,$C3
	.dc.b	$E0,$ED,$6C,$ED,$E7,$29,$F0,$5D,$8E,$FF,$07,$6B,$86,$7F,$A6,$AF
	.dc.b	$F2,$7C,$62,$F9,$7C,$1F,$9B,$E0,$FC,$EF,$A3,$E0,$BD,$3F,$07,$6F
	.dc.b	$86,$1F,$E6,$96,$FC,$17,$BB,$E0,$FD,$ED,$8D,$FC,$1F,$C2,$3E,$1F
	.dc.b	$C1,$D8,$E7,$F0,$5F,$18,$F8,$FF,$07,$63,$BF,$C1,$FC,$AF,$83,$F9
	.dc.b	$6D,$8F,$1F,$C1,$7C,$E3,$E7,$FC,$1D,$8F,$3F,$C1,$FF,$04,$FD,$D7
	.dc.b	$F8,$DE,$FF,$C6,$B9,$7F,$D9,$71,$FC,$78,$FE,$32,$4F,$88,$E1,$F0
	.dc.b	$9C,$7E,$22,$63,$F2,$69,$F0,$B8,$C3,$9F,$C2,$F7,$F8,$5F,$03,$6F
	.dc.b	$87,$C2,$F8,$CF,$27,$C2,$D8,$E9,$F0,$BE,$7F,$85,$F4,$36,$3C,$1F
	.dc.b	$0B,$EB,$3D,$9F,$0B,$63,$C9,$F0,$9E,$F3,$E0,$7C,$2D,$8F,$47,$C2
	.dc.b	$FC,$3F,$85,$F8,$8D,$8F,$67,$C2,$FC,$73,$E4,$7C,$2D,$8F,$81,$F0
	.dc.b	$BF,$2F,$E1,$7E,$63,$63,$E2,$7C,$2F,$F3,$2F,$DD,$BF,$C6,$7C,$FF
	.dc.b	$C3,$F9,$BF,$DF,$31,$FE,$36,$3F,$88,$57,$E4,$DB,$1A,$F8,$4E,$5F
	.dc.b	$09,$FC,$FE,$1F,$85,$BE,$17,$F5,$D1,$FB,$12,$46,$7D,$5C,$46,$3E
	.dc.b	$15,$DF,$88,$FD,$C5,$6D,$C3,$5E,$2D,$79,$47,$17,$C7,$7E,$27,$6D
	.dc.b	$7C,$6D,$7B,$D7,$89,$EF,$19,$F4,$59,$AB,$D9,$EB,$7F,$69,$8C,$5F
	.dc.b	$C7,$9C,$EB,$FC,$12,$7E,$24,$FC,$BF,$84,$CB,$FD,$9A,$1F,$0B,$8C
	.dc.b	$7B,$B6,$3F,$CA,$73,$F4,$FE,$13,$04,$7E,$B7,$C2,$EF,$E1,$64,$7D
	.dc.b	$97,$97,$C2,$F3,$F8,$4E,$9F,$0B,$2B,$EF,$47,$7F,$84,$F0,$7C,$2C
	.dc.b	$9F,$C1,$47,$8F,$E1,$64,$7E,$15,$F3,$7C,$2F,$9F,$E1,$64,$FE,$25
	.dc.b	$F5,$7C,$2F,$AF,$D1,$93,$F8,$D7,$DD,$F0,$BF,$07,$F7,$7F,$B1,$D1
	.dc.b	$FE,$15,$5F,$15,$93,$5D,$BB,$F8,$79,$8F,$F2,$9A,$47,$3F,$87,$E9
	.dc.b	$F0,$FD,$7E,$1E,$5E,$E3,$C1,$F0,$DE,$1F,$87,$93,$CE,$3C,$9F,$0F
	.dc.b	$23,$B3,$E7,$F8,$7F,$47,$C3,$7A,$7E,$1E,$57,$8E,$3D,$9F,$0F,$23
	.dc.b	$CC,$FB,$FE,$1F,$E0,$7C,$3C,$9F,$53,$F0,$FE,$1F,$E2,$7C,$34,$9F
	.dc.b	$73,$F1,$FE,$1F,$E4,$7C,$3F,$C9,$F8,$79,$5F,$0E,$3E,$67,$C3,$FE
	.dc.b	$37,$D6,$FD,$8F,$57,$F9,$05,$BE,$6F,$F5,$AE,$9F,$F8,$01,$87,$F5
	.dc.b	$39,$FD,$22,$AE,$1E,$87,$33,$A7,$C3,$F5,$F8,$7E,$DF,$0F,$2F,$94
	.dc.b	$78,$7E,$1A,$47,$57,$C9,$F0,$FE,$5F,$87,$93,$E1,$7D,$1F,$0F,$E9
	.dc.b	$F8,$6F,$57,$C3,$CA,$F3,$47,$B7,$E1,$E4,$7A,$5F,$81,$F0,$FF,$07
	.dc.b	$E1,$E4,$FB,$5F,$89,$F0,$FF,$17,$E1,$BE,$37,$C3,$CA,$F8,$51,$F2
	.dc.b	$7E,$1F,$E5,$7C,$3C,$9F,$8D,$1F,$37,$E1,$E4,$7C,$97,$E8,$73,$FD
	.dc.b	$1F,$87,$93,$F3,$5F,$E0,$1F,$BC,$7F,$9E,$23,$E7,$75,$FF,$83,$23
	.dc.b	$FC,$85,$1F,$E0,$56,$BA,$3F,$CD,$EE,$3A,$BF,$CF,$8F,$FB,$A6,$38
	.dc.b	$7C,$1F,$62,$3C,$FE,$13,$C3,$F0,$92,$3B,$3E,$4F,$83,$F2,$FC,$17
	.dc.b	$9B,$E0,$A5,$78,$E3,$D3,$F0,$52,$3C,$CF,$B3,$E0,$BD,$BF,$03,$27
	.dc.b	$D4,$FC,$0F,$82,$F8,$3F,$03,$F0,$BE,$0A,$57,$BE,$3E,$2F,$C1,$7C
	.dc.b	$6F,$82,$93,$F0,$E3,$E4,$FC,$17,$CA,$F8,$29,$3F,$1E,$3E,$6F,$C1
	.dc.b	$7C,$EF,$82,$93,$F2,$E3,$F9,$87,$EF,$3F,$C4,$3A,$3F,$C7,$6D,$9F
	.dc.b	$F4,$DB,$7E,$14,$FF,$14,$8F,$F1,$C3,$FC,$41,$BF,$05,$D0,$EB,$F0
	.dc.b	$5D,$BE,$06,$4F,$07,$C3,$F0,$5E,$2F,$81,$93,$D1,$F2,$FC,$17,$9B
	.dc.b	$E0,$BC,$FF,$05,$2B,$C3,$1E,$AF,$81,$F5,$FC,$14,$9F,$2C,$7B,$BE
	.dc.b	$0A,$47,$A1,$F8,$3F,$05,$F0,$BE,$0B,$E1,$FC,$14,$AF,$6C,$7C,$6F
	.dc.b	$82,$91,$F0,$1F,$93,$F0,$5F,$2B,$E0,$7E,$5F,$C1,$4A,$F8,$B1,$F3
	.dc.b	$BE,$0A,$47,$C8,$7F,$B0,$7E,$F5,$FC,$4A,$3F,$D9,$2F,$F1,$0F,$37
	.dc.b	$FD,$AC,$BE,$07,$67,$0F,$82,$E3,$F0,$3C,$BE,$0A,$61,$FE,$6B,$B4
	.dc.b	$F6,$F8,$2E,$FF,$05,$2B,$51,$E2,$F8,$2F,$1F,$C1,$49,$E5,$1E,$6F
	.dc.b	$81,$F3,$FC,$14,$9E,$D1,$EA,$F8,$2F,$5F,$C1,$49,$F1,$47,$57,$C9
	.dc.b	$87,$57,$F7,$5E,$2B,$56,$5D,$E7,$D5,$97,$77,$72,$CB,$E4,$FC,$17
	.dc.b	$CA,$F8,$29,$5F,$06,$3E,$6F,$C0,$FC,$EF,$82,$93,$F1,$63,$E8,$FC
	.dc.b	$0C,$8F,$90,$FD,$4F,$82,$FA,$BF,$05,$F5,$BE,$0A,$57,$CD,$8F,$B3
	.dc.b	$F0,$52,$3E,$83,$FE,$1B,$EB,$FE,$39,$D6,$F3,$FF,$D4,$F1,$FC,$5D
	.dc.b	$BF,$11,$B3,$87,$C4,$F1,$F8,$3E,$5F,$07,$32,$FF,$53,$61,$FE,$30
	.dc.b	$7E,$1E,$5E,$16,$9A,$C7,$A3,$C3,$1A,$FD,$0A,$FE,$4C,$DF,$83,$F4
	.dc.b	$7C,$1C,$8E,$F1,$EB,$F8,$2F,$67,$C1,$C9,$F1,$C7,$BF,$E0,$FE,$07
	.dc.b	$C1,$C9,$F3,$C7,$C3,$F8,$39,$1E,$A7,$E3,$7C,$1F,$58,$F6,$E1,$D5
	.dc.b	$F3,$61,$D5,$F8,$38,$75,$7D,$58,$7C,$FF,$82,$FA,$1F,$07,$27,$E3
	.dc.b	$C7,$57,$E0,$61,$D7,$93,$EA,$43,$C1,$6E,$CF,$C1,$3F,$1F,$3F,$05
	.dc.b	$F0,$BD,$E7,$F7,$FE,$0B,$F0,$1F,$07,$30,$FA,$F1,$F8,$3F,$82,$91
	.dc.b	$F6,$9F,$C3,$7C,$1F,$E1,$FE,$0A,$4F,$DE,$7F,$15,$F0,$7F,$E6,$5F
	.dc.b	$A9,$FE,$64,$7F,$31,$9F,$D4,$B7,$4F,$23,$FC,$DF,$76,$3E,$0A,$FC
	.dc.b	$03,$E0,$7B,$10,$FC,$26,$1E,$07,$FD,$E9,$5F,$CF,$5E,$FF,$07,$E3
	.dc.b	$3C,$9F,$07,$E5,$F8,$3F,$37,$C1,$CA,$E3,$1E,$9F,$83,$F5,$7C,$1C
	.dc.b	$9E,$B1,$ED,$F8,$2F,$77,$C1,$C9,$F0,$C7,$57,$C7,$87,$57,$E0,$E1
	.dc.b	$D5,$F3,$E1,$D5,$F8,$B8,$7C,$9F,$82,$F9,$5F,$07,$27,$DB,$1F,$37
	.dc.b	$E0,$E4,$7C,$07,$E8,$7C,$1F,$D1,$F8,$29,$3F,$11,$FA,$9F,$07,$F5
	.dc.b	$7E,$0E,$4F,$C8,$7E,$C7,$C1,$F4,$8F,$AF,$85,$E7,$FA,$F8,$F4,$9F
	.dc.b	$B7,$8D,$9F,$B7,$8F,$E0,$7E,$0A,$57,$D4,$7F,$09,$F0,$7C,$FC,$A4
	.dc.b	$FE,$6F,$9D,$9E,$63,$19,$F5,$AD,$F8,$0B,$69,$9B,$E5,$9F,$E4,$7A
	.dc.b	$E4,$7E,$11,$FF,$5C,$7E,$F7,$FD,$67,$03,$FD,$67,$1E,$AD,$AF,$D7
	.dc.b	$36,$77,$F0,$9C,$BE,$0B,$9F,$C1,$CB,$F5,$F0,$FE,$BF,$0F,$EB,$71
	.dc.b	$8C,$7A,$FC,$17,$8B,$E0,$E4,$E4,$EF,$1E,$6F,$84,$91,$9D,$3C,$F1
	.dc.b	$EA,$D3,$C5,$39,$C5,$9F,$8E,$3F,$AB,$C7,$89,$BF,$03,$F0,$BE,$0A
	.dc.b	$47,$18,$F8,$BF,$01,$F1,$BE,$0A,$4F,$58,$F9,$3F,$07,$23,$C0,$FC
	.dc.b	$CF,$84,$F9,$BF,$05,$27,$C8,$FD,$0F,$83,$EB,$1E,$7C,$3A,$BE,$6C
	.dc.b	$3A,$3E,$8C,$2F,$1F,$CB,$C7,$EC,$FC,$14,$AF,$80,$FD,$CF,$83,$FB
	.dc.b	$BF,$07,$27,$E2,$3F,$80,$F8,$3F,$C0,$FC,$17,$E0,$BE,$0E,$57,$C9
	.dc.b	$8F,$C2,$FC,$04,$8F,$98,$FF,$BF,$FF,$7C,$FE,$CB,$D7,$F8,$6C,$7B
	.dc.b	$57,$CE,$7B,$3C,$CF,$A9,$87,$67,$E9,$E1,$CF,$E0,$64,$7D,$DC,$3F
	.dc.b	$B7,$3F,$CF,$5F,$EB,$FB,$7E,$2F,$0E,$ED,$9F,$C0,$67,$EC,$7B,$90
	.dc.b	$EE,$75,$F8,$3F,$51,$1F,$07,$C1,$FB,$7E,$0B,$DD,$F0,$72,$7C,$91
	.dc.b	$F0,$7D,$E9,$1E,$77,$E2,$7B,$FF,$16,$E4,$F4,$3D,$9E,$F7,$57,$DD
	.dc.b	$87,$57,$FC,$57,$AB,$1E,$B1,$8C,$6E,$9F,$D6,$69,$97,$D3,$F7,$FE
	.dc.b	$A7,$C1,$CC,$3E,$54,$7D,$7F,$83,$91,$F3,$5F,$B5,$F0,$5F,$6F,$E0
	.dc.b	$FE,$E7,$C1,$4A,$FA,$51,$F7,$FE,$0F,$B4,$74,$3E,$B3,$D1,$F4,$DD
	.dc.b	$D3,$CD,$87,$59,$FB,$98,$75,$7D,$F8,$7E,$2F,$E0,$7F,$19,$F0,$72
	.dc.b	$7F,$05,$1F,$8F,$F8,$2F,$C8,$7C,$1C,$9F,$C3,$47,$F2,$87,$FC,$A7
	.dc.b	$1F,$E9,$F3,$FB,$67,$67,$F2,$66,$C7,$F8,$1D,$D7,$C1,$F8,$63,$B9
	.dc.b	$FC,$36,$3F,$B4,$69,$F3,$EF,$C1,$3F,$5A,$7C,$8E,$FE,$27,$C5,$5F
	.dc.b	$E0,$D9,$EF,$E7,$1A,$79,$7E,$13,$79,$79,$5B,$87,$95,$B5,$C7,$E0
	.dc.b	$7B,$CF,$84,$FD,$35,$E3,$0F,$06,$3C,$23,$BF,$C5,$58,$C7,$C1,$CB
	.dc.b	$F7,$3B,$F8,$2E,$1F,$07,$27,$E1,$3C,$FE,$13,$A7,$C2,$C9,$F8,$CF
	.dc.b	$7F,$84,$F0,$7C,$2F,$39,$F9,$58,$61,$E6,$43,$EB,$E1,$D5,$E6,$43
	.dc.b	$ED,$DF,$57,$EA,$7C,$0F,$B4,$8F,$D7,$F8,$1F,$81,$F0,$12,$3E,$D3
	.dc.b	$F0,$FE,$03,$E2,$7C,$2F,$C5,$F8,$59,$5F,$7E,$3E,$47,$C2,$FC,$9C
	.dc.b	$24,$FE,$0E,$3A,$3F,$8E,$C2,$F2,$7D,$FC,$7A,$CF,$E2,$71,$EA,$FE
	.dc.b	$53,$1F,$A9,$F0,$BF,$57,$E1,$24,$FE,$3E,$3E,$FF,$EF,$BF,$DC,$C9
	.dc.b	$18,$F8,$2C,$FC,$14,$8F,$EC,$53,$FE,$CD,$6F,$C2,$F2,$F8,$39,$1A
	.dc.b	$7A,$FC,$27,$68,$87,$23,$B1,$D2,$FB,$3F,$EB,$EB,$EC,$DB,$F3,$FC
	.dc.b	$2F,$48,$F1,$DE,$1E,$84,$39,$1D,$8E,$97,$DB,$7F,$F3,$A6,$1D,$9B
	.dc.b	$87,$C4,$F8,$39,$7F,$01,$F8,$FF,$09,$F2,$3E,$06,$4F,$C4,$7E,$5F
	.dc.b	$C1,$7C,$CF,$82,$93,$F2,$1F,$9F,$F0,$5F,$43,$E1,$39,$CF,$CC,$C3
	.dc.b	$0F,$32,$1F,$0F,$0E,$AF,$32,$1F,$1E,$FA,$BF,$5B,$E1,$3E,$EF,$C2
	.dc.b	$48,$C1,$F6,$BE,$13,$F0,$24,$7E,$EF,$C0,$7E,$13,$E0,$24,$7E,$01
	.dc.b	$FC,$3F,$C1,$7E,$23,$E0,$E4,$FE,$11,$FC,$5F,$C2,$73,$8F,$B5,$85
	.dc.b	$E3,$F9,$F8,$F4,$9F,$BD,$8D,$9F,$A7,$8F,$F1,$2F,$DF,$BF,$BF,$D7
	.dc.b	$FA,$BD,$BE,$79,$A7,$F7,$87,$1F,$11,$C7,$E1,$65,$7F,$8E,$6B,$FB
	.dc.b	$DF,$A7,$FE,$5A,$5F,$0B,$DC,$93,$BF,$83,$F1,$7C,$27,$8F,$E1,$A4
	.dc.b	$F3,$8F,$36,$5D,$A1,$AF,$FA,$FF,$1F,$F7,$83,$67,$AC,$EB,$2F,$16
	.dc.b	$58,$AE,$19,$7C,$1F,$85,$97,$E9,$7E,$27,$C2,$78,$23,$D9,$AF,$86
	.dc.b	$FB,$47,$BF,$0E,$98,$F6,$8F,$61,$F0,$FE,$1B,$E7,$7C,$24,$8C,$1F
	.dc.b	$1F,$E1,$7E,$97,$C2,$48,$C1,$F2,$FE,$13,$EB,$7C,$1C,$8C,$1F,$3F
	.dc.b	$E1,$3E,$D1,$1F,$A5,$F0,$BF,$77,$0E,$91,$F0,$CF,$BD,$86,$5E,$66
	.dc.b	$79,$BF,$01,$87,$78,$C7,$23,$2F,$C4,$7C,$27,$E2,$7E,$16,$57,$E0
	.dc.b	$63,$F1,$9F,$09,$F8,$DF,$82,$93,$F8,$58,$FE,$C3,$FB,$FB,$E0,$AF
	.dc.b	$C1,$CA,$FF,$41,$4E,$5B,$F0,$DC,$08,$EB,$E1,$BB,$BF,$91,$C7,$BB
	.dc.b	$E8,$C7,$BE,$3D,$63,$DD,$87,$F7,$A3,$B3,$8F,$86,$F2,$91,$F1,$FC
	.dc.b	$37,$A3,$E0,$E4,$79,$9F,$5F,$C1,$FB,$3E,$1A,$4F,$A9,$F7,$FC,$27
	.dc.b	$C0,$B9,$3D,$0F,$7F,$C3,$74,$7C,$B8,$5E,$F7,$77,$FB,$13,$46,$7D
	.dc.b	$AF,$D7,$9D,$BD,$C3,$A1,$6D,$3E,$8F,$C3,$7D,$2D,$25,$FD,$08,$F0
	.dc.b	$67,$DF,$4B,$3C,$F0,$F0,$BE,$D2,$1E,$23,$EC,$E1,$F7,$BE,$1B,$EF
	.dc.b	$FC,$34,$9F,$BB,$1F,$82,$F8,$59,$1F,$80,$7F,$0B,$F0,$BF,$86,$F8
	.dc.b	$4F,$C3,$FC,$2C,$AF,$C2,$C7,$E2,$BE,$1B,$B4,$74,$3F,$12,$F4,$7E
	.dc.b	$FD,$DE,$1F,$8B,$C3,$F1,$31,$ED,$5F,$90,$C7,$B3,$F8,$6C,$7F,$2F
	.dc.b	$F0,$B2,$3F,$2A,$FE,$6B,$E1,$BF,$C7,$7E,$AF,$FB,$BC,$93,$8F,$86
	.dc.b	$CF,$C1,$C8,$FF,$6F,$4F,$F6,$3E,$4F,$F7,$FB,$FE,$D9,$D7,$CB,$C6
	.dc.b	$C7,$B5,$FE,$91,$EF,$F0,$DE,$08,$87,$43,$C0,$76,$C7,$C0,$FE,$F7
	.dc.b	$0F,$03,$F8,$57,$C1,$8F,$8B,$E1,$7D,$7F,$0D,$2B,$CF,$1E,$EF,$84
	.dc.b	$F7,$FC,$0C,$9F,$5C,$7C,$2F,$81,$91,$EE,$7E,$2F,$C2,$7C,$6F,$82
	.dc.b	$93,$F0,$9F,$93,$F0,$9D,$63,$E2,$E1,$D5,$F4,$61,$D1,$F5,$61,$78
	.dc.b	$7E,$DE,$AC,$B9,$D7,$CE,$C6,$BF,$07,$E1,$3B,$3F,$0F,$2E,$D8,$72
	.dc.b	$CB,$A6,$DF,$53,$3B,$7F,$1F,$2F,$BF,$F0,$92,$3E,$DB,$F8,$2F,$84
	.dc.b	$FC,$1F,$C1,$C9,$FB,$EF,$E1,$BE,$0F,$F0,$FF,$07,$27,$F0,$6F,$E2
	.dc.b	$BE,$0F,$94,$7E,$1A,$EB,$F6,$7E,$13,$A3,$F6,$F0,$BB,$7C,$2C,$7A
	.dc.b	$4F,$DF,$C6,$CF,$C6,$C7,$F2,$FC,$52,$BF,$26,$FF,$A5,$3F,$80,$3F
	.dc.b	$C0,$27,$F7,$F3,$FB,$DC,$7F,$0B,$6F,$80,$75,$EF,$F1,$F7,$E5,$0F
	.dc.b	$E9,$93,$FD,$DB,$8B,$17,$C5,$F1,$4F,$E5,$70,$F1,$47,$8F,$E1,$7C
	.dc.b	$9F,$09,$2B,$A6,$9F,$BA,$C7,$C3,$18,$BE,$B1,$FE,$10,$8C,$1E,$9C
	.dc.b	$BD,$D9,$FB,$FE,$12,$4E,$B4,$F6,$47,$C3,$F8,$5F,$89,$F0,$BF,$17
	.dc.b	$E1,$65,$EF,$24,$FC,$9F,$85,$93,$97,$E6,$7C,$2F,$38,$E1,$8F,$81
	.dc.b	$F8,$B8,$F6,$7A,$63,$D9,$F9,$38,$74,$7E,$5E,$17,$6F,$85,$8F,$59
	.dc.b	$E6,$79,$70,$FB,$9F,$07,$D6,$3E,$9E,$3D,$6F,$E5,$63,$CF,$7F,$93
	.dc.b	$95,$BF,$C0,$65,$F8,$5F,$83,$FC,$37,$C0,$C9,$F8,$51,$F8,$9F,$83
	.dc.b	$FC,$57,$C2,$C9,$F8,$D1,$F8,$DF,$85,$91,$F2,$5F,$C8,$7C,$1F,$58
	.dc.b	$E6,$43,$F1,$3F,$09,$D5,$FA,$17,$D5,$FC,$6D,$F4,$7F,$1F,$77,$6F
	.dc.b	$C3,$61,$F9,$DF,$80,$95,$F6,$5F,$D0,$7C,$0F,$FB,$D3,$F8,$07,$FC
	.dc.b	$07,$7F,$D1,$9B,$C1,$F9,$86,$DF,$F6,$D5,$F0,$BC,$7E,$17,$97,$C2
	.dc.b	$CA,$FC,$65,$27,$BC,$78,$48,$7E,$9B,$1F,$17,$C2,$C8,$E2,$F9,$70
	.dc.b	$ED,$1D,$30,$EC,$FF,$8E,$70,$EC,$F8,$3E,$17,$D9,$F0,$B2,$30,$79
	.dc.b	$3E,$07,$E0,$11,$F3,$FC,$2F,$C3,$F8,$5F,$89,$F0,$72,$7D,$71,$F1
	.dc.b	$F0,$91,$D0,$F1,$7C,$2F,$37,$C9,$85,$E1,$F9,$98,$F5,$9F,$8B,$8F
	.dc.b	$57,$D5,$8F,$D3,$C3,$AC,$7B,$71,$C3,$E9,$21,$D8,$F4,$9E,$63,$A9
	.dc.b	$0E,$A7,$D4,$F8,$5F,$C0,$11,$FA,$FF,$09,$F8,$3F,$85,$91,$F6,$9F
	.dc.b	$C3,$7C,$27,$E1,$FE,$12,$4F,$DE,$7F,$15,$F0,$7F,$8B,$BE,$93,$F5
	.dc.b	$70,$C3,$CC,$87,$D9,$C3,$AB,$CC,$87,$DD,$BE,$AF,$E3,$3E,$13,$F3
	.dc.b	$1F,$0B,$23,$07,$E4,$3E,$17,$F3,$9F,$07,$23,$07,$F1,$FF,$B8,$FE
	.dc.b	$6C,$DF,$86,$BC,$9F,$E2,$69,$46,$3B,$7F,$9C,$C7,$0A,$FC,$86,$5F
	.dc.b	$A9,$23,$D7,$E1,$3F,$59,$97,$86,$3F,$2D,$96,$1F,$11,$D4,$FF,$2C
	.dc.b	$5F,$81,$F2,$1A,$3D,$1F,$11,$BF,$41,$2F,$86,$C3,$6B,$67,$BF,$E1
	.dc.b	$73,$F0,$D2,$30,$7C,$3F,$82,$E2,$47,$E3,$7C,$1F,$4C,$A4,$74,$3F
	.dc.b	$D5,$18,$3E,$0F,$C3,$8F,$59,$CE,$38,$7B,$10,$E3,$8F,$A3,$1F,$4D
	.dc.b	$F6,$9E,$B8,$E1,$F3,$9D,$0F,$B5,$7D,$5E,$67,$5E,$2F,$B5,$87,$58
	.dc.b	$FC,$07,$C1,$FC,$62,$3F,$83,$F8,$6F,$93,$F0,$D2,$3F,$0C,$FC,$CF
	.dc.b	$86,$F9,$BF,$0D,$27,$F1,$4F,$D0,$F8,$6F,$A3,$7D,$27,$E2,$E1,$87
	.dc.b	$99,$0F,$93,$87,$57,$99,$0F,$9B,$7D,$5F,$CC,$7C,$2F,$DE,$F8,$59
	.dc.b	$18,$3F,$39,$F0,$BF,$82,$23,$F9,$FF,$85,$FC,$2F,$C3,$7E,$1B,$E1
	.dc.b	$A4,$FE,$8E,$3F,$13,$84,$8E,$87,$FB,$A7,$F2,$DF,$F0,$E7,$77,$48
	.dc.b	$FC,$0E,$37,$8B,$F5,$19,$76,$9F,$D7,$65,$D9,$FC,$3E,$5F,$9F,$C7
	.dc.b	$FC,$4D,$F0,$F6,$4D,$31,$F0,$F9,$F8,$6D,$7C,$3C,$AF,$F0,$ED,$FF
	.dc.b	$8B,$E7,$97,$C3,$73,$F8,$69,$5A,$11,$D1,$C6,$3D,$DD,$63,$DD,$E1
	.dc.b	$8F,$77,$E1,$1F,$E3,$B1,$8F,$F8,$FE,$FB,$C6,$B0,$F1,$BD,$08,$72
	.dc.b	$BF,$81,$F0,$9D,$E3,$AE,$1D,$F6,$FF,$6D,$E3,$ED,$9F,$56,$3D,$A7
	.dc.b	$30,$C7,$BB,$E6,$F8,$5F,$99,$F0,$B2,$3E,$14,$7C,$FF,$85,$FA,$1F
	.dc.b	$0D,$27,$E3,$47,$D3,$F8,$5F,$A9,$F0,$D2,$7E,$54,$7D,$7B,$91,$D0
	.dc.b	$F9,$DF,$09,$D5,$FA,$18,$75,$7E,$4E,$1D,$5F,$A9,$87,$57,$E6,$E1
	.dc.b	$F8,$3F,$85,$FC,$27,$C2,$C9,$FB,$51,$F8,$7F,$85,$FC,$47,$C3,$49
	.dc.b	$FB,$D1,$F8,$BF,$86,$91,$F8,$17,$F1,$DF,$0B,$F9,$1F,$B9,$FF,$0D
	.dc.b	$F2,$7F,$9D,$F1,$F0,$5F,$E3,$27,$2B,$0E,$CF,$E0,$B1,$F1,$3D,$48
	.dc.b	$7E,$1B,$1E,$EF,$3F,$85,$F0,$11,$ED,$F0,$FE,$3F,$86,$F2,$7B,$F2
	.dc.b	$7C,$51,$E7,$F8,$59,$1E,$57,$D5,$F0,$BE,$BB,$93,$D0,$F5,$7C,$2F
	.dc.b	$57,$D9,$87,$57,$F7,$9A,$B1,$EB,$18,$C6,$F5,$7E,$4A,$99,$7C,$9C
	.dc.b	$65,$F6,$3E,$47,$C2,$77,$7A,$E5,$DF,$0E,$59,$74,$F1,$BF,$D4,$DC
	.dc.b	$4C,$DA,$67,$1F,$D6,$C7,$81,$FA,$9F,$0B,$F6,$7E,$16,$46,$0F,$B1
	.dc.b	$F0,$5F,$77,$E0,$E4,$60,$FB,$9F,$07,$F8,$1F,$85,$91,$83,$F0,$1F
	.dc.b	$09,$F8,$5C,$64,$60,$E4,$7C,$8B,$EA,$FC,$AC,$3F,$19,$87,$48,$F9
	.dc.b	$B8,$61,$FC,$67,$C2,$FE,$44,$8F,$E3,$FE,$0F,$F2,$9F,$0B,$F9,$5F
	.dc.b	$83,$93,$F9,$38,$FC,$C7,$C1,$FF,$29,$E4,$FF,$09,$1F,$C1,$8B,$C7
	.dc.b	$F3,$38,$53,$F9,$FC,$66,$1D,$CF,$E5,$37,$FE,$0F,$7F,$07,$87,$81
	.dc.b	$FC,$2E,$3E,$47,$F4,$97,$E1,$61,$83,$9F,$C2,$79,$48,$F6,$F7,$BD
	.dc.b	$1F,$07,$23,$C2,$FA,$FE,$0B,$D9,$F0,$1E,$DF,$80,$95,$E6,$8F,$81
	.dc.b	$7D,$23,$FA,$A6,$8C,$73,$1D,$08,$6B,$1E,$CF,$33,$E0,$E1,$D9,$E7
	.dc.b	$87,$CD,$F8,$2E,$71,$DB,$1C,$3D,$31,$FA,$18,$7A,$E3,$C7,$97,$A9
	.dc.b	$EA,$43,$A9,$F4,$3E,$0B,$ED,$7B,$F2,$30,$7D,$4F,$80,$FB,$DF,$01
	.dc.b	$23,$07,$D8,$F8,$2F,$C1,$11,$FB,$7F,$03,$F8,$5F,$82,$FC,$36,$32
	.dc.b	$7E,$FC,$7E,$27,$0E,$91,$F2,$B0,$C3,$D4,$87,$CE,$C3,$F2,$18,$7E
	.dc.b	$47,$E0,$64,$FE,$29,$FC,$A7,$C1,$7E,$57,$E0,$A4,$FE,$39,$FC,$C7
	.dc.b	$C1,$7E,$67,$E0,$64,$FE,$49,$FC,$E7,$C1,$72,$8F,$CA,$DD,$7F,$C2
	.dc.b	$5F,$97,$3B,$BA,$C7,$E0,$B1,$BC,$9F,$A3,$CB,$AC,$FE,$1B,$2B,$3F
	.dc.b	$A7,$CB,$FC,$25,$8F,$F8,$42,$BF,$C1,$6B,$E0,$72,$6B,$E0,$B7,$F0
	.dc.b	$7C,$3E,$0A,$5F,$08,$E7,$F0,$5D,$3E,$0A,$4F,$48,$E4,$F6,$BA,$FB
	.dc.b	$3E,$0F,$AB,$E2,$C3,$AB,$F0,$30,$EA,$F9,$B0,$EA,$FC,$4C,$3D,$77
	.dc.b	$D3,$4F,$F0,$4E,$7E,$DA,$EC,$79,$CE,$C4,$3B,$1E,$73,$B1,$0E,$C7
	.dc.b	$C6,$F8,$2F,$93,$F0,$72,$30,$7C,$AF,$81,$F9,$A4,$7E,$6F,$C1,$7D
	.dc.b	$0F,$81,$91,$F4,$1F,$A7,$F0,$5F,$53,$09,$3D,$0F,$C0,$FC,$1F,$37
	.dc.b	$F0,$78,$5C,$FD,$F8,$F4,$9F,$C3,$E3,$67,$E1,$E3,$F8,$0F,$81,$95
	.dc.b	$F8,$07,$F0,$7F,$07,$F8,$4F,$82,$FC,$2F,$C1,$4A,$FC,$2C,$7E,$23
	.dc.b	$E0,$A4,$7E,$21,$FC,$5F,$C1,$F2,$8F,$C5,$DD,$7F,$CD,$5F,$11,$FE
	.dc.b	$10,$E2,$FF,$06,$47,$E4,$31,$EC,$FE,$7B,$1E,$AF,$E8,$31,$BA,$7D
	.dc.b	$CC,$BF,$CD,$58,$BE,$12,$92,$FF,$CC,$B0,$F8,$4D,$FC,$07,$0F,$81
	.dc.b	$95,$A8,$E7,$F0,$1D,$3E,$12,$4F,$28,$EF,$7D,$63,$D7,$8E,$1E,$47
	.dc.b	$83,$0E,$AF,$C0,$C3,$AB,$E4,$C3,$AB,$F1,$30,$E4,$FC,$6C,$2E,$7F
	.dc.b	$E6,$2C,$9F,$09,$EA,$AE,$C4,$3B,$1E,$73,$B1,$0E,$C7,$C4,$F8,$4F
	.dc.b	$93,$F0,$32,$30,$7C,$8F,$83,$F9,$A4,$7E,$5F,$C0,$FD,$0F,$82,$FA
	.dc.b	$3F,$05,$27,$E7,$C7,$D4,$F8,$2F,$AB,$8C,$9F,$A7,$1C,$DF,$C1,$E1
	.dc.b	$74,$F7,$63,$D6,$7E,$CE,$3D,$5F,$C4,$E3,$F8,$0F,$83,$91,$F7,$9F
	.dc.b	$C1,$FC,$1F,$E1,$3E,$0B,$F0,$BF,$05,$2B,$F0,$71,$F8,$8F,$82,$91
	.dc.b	$F8,$67,$F1,$7F,$05,$CA,$3A,$10,$FC,$CD,$F5,$7F,$1B,$7D,$5F,$CE
	.dc.b	$DF,$47,$F3,$F7,$74,$FB,$58,$7F,$B0,$7F,$11,$FF,$0B,$D7,$FA,$A9
	.dc.b	$BF,$0A,$69,$FE,$15,$AF,$F5,$1B,$FD,$FE,$7F,$D8,$2D,$7F,$DA,$CD
	.dc.b	$8D,$FC,$07,$7F,$80,$F0,$7C,$04,$BE,$31,$E3,$F8,$8F,$24,$43,$AB
	.dc.b	$E7,$F8,$59,$1E,$07,$D5,$7D,$63,$C7,$7D,$5F,$1D,$FB,$FE,$13,$B4
	.dc.b	$74,$3D,$2F,$47,$D3,$77,$1F,$4C,$79,$8E,$A4,$3A,$9F,$0B,$E1,$7E
	.dc.b	$67,$C2,$C8,$C1,$F1,$BE,$13,$E8,$7C,$2C,$8C,$1F,$2B,$E1,$3E,$A1
	.dc.b	$1F,$9B,$F0,$9F,$5F,$09,$1D,$0F,$9F,$F0,$9F,$6F,$0E,$91,$F4,$B0
	.dc.b	$C3,$D4,$87,$D6,$C3,$F0,$58,$7E,$0F,$E1,$24,$FD,$C7,$F0,$DF,$0B
	.dc.b	$F8,$7F,$85,$93,$F8,$07,$F1,$5F,$0B,$F8,$BF,$84,$93,$F8,$47,$F1
	.dc.b	$DF,$09,$CA,$3F,$0F,$75,$FC,$3F,$C2,$74,$7F,$13,$85,$E1,$FC,$7E
	.dc.b	$3D,$27,$F1,$B8,$D9,$FC,$9E,$3F,$9B,$F8,$49,$5F,$94,$4C,$9F,$F7
	.dc.b	$DF,$C4,$7F,$86,$63,$1F,$0D,$FE,$A2,$8F,$EE,$D3,$AF,$86,$E3,$48
	.dc.b	$C3,$C0,$76,$AB,$8D,$F3,$F8,$4E,$37,$E0,$F8,$5B,$3D,$DB,$87,$2B
	.dc.b	$5E,$30,$E9,$76,$7D,$15,$E3,$8F,$41,$9F,$84,$D7,$C3,$6F,$39,$5E
	.dc.b	$7C,$3C,$D3,$CF,$1E,$99,$4A,$D6,$5E,$26,$E3,$E1,$CA,$46,$0D,$E1
	.dc.b	$D9,$F2,$1E,$BC,$BC,$6F,$4C,$7B,$3A,$CB,$E4,$65,$E3,$8F,$0F,$C3
	.dc.b	$7B,$FE,$1A,$46,$0F,$2F,$C2,$FC,$3F,$86,$91,$83,$D3,$F0,$BF,$1C
	.dc.b	$8F,$B3,$E1,$7E,$56,$12,$3A,$1E,$3E,$07,$C2,$FC,$FC,$7A,$CE,$71
	.dc.b	$C3,$D8,$87,$1C,$7E,$B6,$3F,$5F,$E1,$7E,$C7,$C3,$4A,$F9,$B1,$F6
	.dc.b	$FE,$1B,$EE,$7C,$34,$9F,$A3,$1F,$7F,$E1,$3F,$01,$F0,$B2,$7E,$AC
	.dc.b	$75,$79,$90,$F5,$7C,$2F,$57,$ED,$DF,$57,$DD,$7D,$5F,$BF,$7D,$5F
	.dc.b	$85,$7F,$8F,$F8,$59,$1F,$84,$7F,$29,$C8,$F7,$EF,$C2,$E3,$E0,$65
	.dc.b	$FE,$3E,$BF,$C4,$8D,$F8,$4E,$24,$75,$F0,$BD,$30,$91,$C5,$EF,$F0
	.dc.b	$9D,$A3,$A6,$1D,$9F,$C5,$61,$D9,$B8,$75,$69,$D8,$F2,$57,$67,$8D
	.dc.b	$F6,$7D,$18,$F5,$F0,$3F,$83,$61,$FB,$EC,$7C,$55,$D0,$93,$A9,$63
	.dc.b	$F1,$B2,$F8,$FF,$0B,$2F,$E2,$3F,$2B,$E1,$7E,$5F,$C2,$49,$F9,$0F
	.dc.b	$CE,$F8,$5F,$9F,$F0,$9F,$43,$E1,$25,$7C,$D8,$E4,$FC,$FB,$AF,$C4
	.dc.b	$F8,$3E,$8F,$C6,$C2,$EB,$E7,$C7,$AC,$FD,$7C,$7A,$BF,$33,$1F,$BF
	.dc.b	$F0,$9F,$80,$F8,$59,$3F,$76,$3F,$07,$F0,$B2,$3F,$00,$FE,$1B,$E1
	.dc.b	$7F,$0F,$F0,$9F,$88,$F8,$59,$5F,$85,$8F,$C5,$FC,$2F,$68,$E8,$7E
	.dc.b	$25,$E8,$FD,$EB,$BA,$FC,$FC,$3A,$CF,$E3,$F0,$EA,$FE,$13,$0F,$CB
	.dc.b	$FC,$2F,$F2,$BF,$3B,$FC,$3A,$7F,$87,$B9,$91,$9F,$84,$D7,$C0,$4C
	.dc.b	$3F,$94,$5F,$F8,$84,$FE,$97,$7F,$EE,$1F,$84,$E9,$F0,$72,$B7,$1E
	.dc.b	$07,$F4,$58,$D7,$F9,$05,$BE,$13,$C1,$1D,$B1,$F0,$3A,$CB,$D1,$87
	.dc.b	$58,$E3,$8E,$1F,$2D,$F5,$DB,$F2,$59,$7B,$E3,$AC,$F8,$31,$EB,$5F
	.dc.b	$9C,$CB,$D9,$3F,$D6,$96,$3D,$E7,$58,$F7,$AF,$4F,$C2,$7C,$BF,$83
	.dc.b	$91,$F1,$A3,$E7,$7C,$24,$8F,$92,$FD,$1F,$83,$FA,$5F,$07,$F4,$FE
	.dc.b	$0E,$57,$CE,$8F,$AD,$72,$3A,$1F,$4B,$E0,$FA,$3F,$2B,$0B,$9F,$AF
	.dc.b	$1E,$B3,$F6,$31,$EA,$FD,$0C,$7F,$05,$F0,$92,$3E,$EB,$F8,$5F,$84
	.dc.b	$FC,$37,$C2,$49,$FC,$0B,$F8,$9F,$83,$FC,$57,$C1,$C9,$FC,$2B,$F8
	.dc.b	$DF,$83,$E5,$1D,$08,$7E,$02,$FA,$3F,$82,$BB,$A7,$D3,$C3,$A4,$FE
	.dc.b	$1B,$0B,$3F,$5F,$0F,$CD,$7C,$1C,$AF,$CA,$BF,$9D,$F8,$4F,$CF,$7C
	.dc.b	$27,$F8,$27,$F2,$EF,$F0,$3E,$1F,$C6,$9B,$CF,$F9,$A1,$3B,$F8,$49
	.dc.b	$8F,$F8,$1E,$BF,$C3,$6F,$F8,$79,$7C,$27,$78,$F0,$90,$FC,$EE,$3E
	.dc.b	$27,$96,$1E,$33,$0F,$F8,$06,$BA,$F7,$1A,$4B,$ED,$3D,$0F,$D9,$DF
	.dc.b	$B7,$E1,$39,$E7,$ED,$C7,$B5,$7E,$E3,$3F,$5B,$DF,$3B,$5F,$0F,$1F
	.dc.b	$8D,$F0,$92,$BE,$03,$F2,$7E,$17,$E5,$7C,$24,$9F,$88,$FC,$DF,$85
	.dc.b	$F9,$DF,$07,$27,$E4,$3F,$47,$E0,$F9,$47,$42,$1F,$06,$FA,$3F,$0E
	.dc.b	$EE,$5F,$5F,$0E,$93,$F1,$F0,$B3,$F6,$F0,$FB,$DF,$07,$F7,$FE,$0E
	.dc.b	$5F,$D9,$8F,$C1,$7C,$27,$E0,$FE,$0E,$4F,$DD,$8F,$C3,$7C,$27,$E1
	.dc.b	$FE,$0A,$4F,$E0,$63,$AB,$CC,$87,$D9,$F8,$3E,$AF,$E1,$EF,$AB,$F7
	.dc.b	$6F,$AB,$F8,$BB,$EA,$FE,$06,$FF,$2D,$F0,$5F,$97,$F8,$39,$3F,$91
	.dc.b	$8F,$CD,$7C,$27,$E6,$FE,$12,$4F,$E5,$63,$FC,$CD,$F7,$7F,$E1,$38
	.dc.b	$FF,$2F,$3F,$E0,$3E,$0F,$EF,$2D,$E9,$FE,$72,$EE,$7F,$25,$87,$89
	.dc.b	$F0,$1C,$BE,$0B,$C2,$FE,$8F,$1F,$0B,$F9,$9C,$7C,$3F,$09,$E2,$F8
	.dc.b	$49,$3E,$28,$F2,$DF,$58,$EE,$75,$3D,$47,$52,$1D,$4E,$47,$B2,$7A
	.dc.b	$BF,$E6,$9C,$7D,$59,$FB,$BE,$13,$E1,$FC,$24,$AF,$87,$1F,$1B,$E1
	.dc.b	$7E,$3F,$C2,$C9,$F8,$F1,$F2,$BE,$17,$E5,$FC,$24,$9F,$97,$1F,$3A
	.dc.b	$E4,$74,$3E,$7F,$C2,$75,$7E,$8E,$1D,$5F,$36,$1D,$5F,$AB,$87,$57
	.dc.b	$D5,$87,$DA,$F8,$39,$1F,$69,$FB,$BF,$09,$F7,$BE,$17,$EF,$FC,$24
	.dc.b	$AF,$BF,$1F,$82,$F8,$5F,$C1,$FC,$24,$9F,$C1,$C7,$27,$A1,$0F,$95
	.dc.b	$7D,$5F,$C4,$DF,$57,$E7,$5F,$57,$F1,$B7,$D5,$FA,$57,$F9,$2F,$84
	.dc.b	$91,$F9,$27,$F2,$BF,$07,$F9,$6F,$84,$FC,$BF,$C2,$CA,$FC,$BC,$7E
	.dc.b	$6B,$E1,$3F,$D8,$1F,$97,$7F,$B4,$27,$FD,$4C,$DE,$6F,$CD,$CF,$49
	.dc.b	$EC,$43,$F0,$98,$77,$7F,$47,$87,$77,$F1,$18,$77,$7F,$D8,$98,$75
	.dc.b	$EE,$FF,$0E,$B6,$3F,$DB,$BD,$1F,$E2,$67,$F5,$B8,$78,$5E,$E7,$2F
	.dc.b	$86,$F0,$3F,$95,$C7,$C1,$9F,$EC,$B2,$F0,$4F,$B0,$F7,$98,$8F,$F3
	.dc.b	$7E,$2F,$85,$F8,$7F,$0B,$2B,$CF,$1F,$1B,$E1,$7E,$3F,$C3,$49,$F5
	.dc.b	$C7,$CA,$F8,$59,$1E,$E7,$E6,$FC,$37,$CE,$B9,$3D,$0F,$87,$F0,$BD
	.dc.b	$5F,$8B,$87,$57,$E9,$E1,$D1,$FA,$B8,$5C,$FD,$38,$FD,$AF,$85,$95
	.dc.b	$F3,$9F,$BB,$F0,$9F,$7B,$E1,$7E,$FF,$C3,$4A,$FA,$71,$F8,$2F,$86
	.dc.b	$91,$F5,$9F,$C2,$FC,$37,$28,$E8,$43,$F0,$F7,$D5,$FB,$B7,$D5,$FC
	.dc.b	$5D,$F4,$7F,$1B,$77,$2F,$A3,$87,$E4,$BE,$1B,$F2,$7F,$0D,$2F,$F0
	.dc.b	$F1,$F9,$6F,$86,$91,$F8,$A7,$F3,$3F,$0D,$F9,$AF,$86,$93,$F8,$E7
	.dc.b	$F3,$BF,$09,$D6,$39,$90,$FC,$FF,$C1,$F4,$7F,$43,$77,$3F,$C2,$E1
	.dc.b	$D2,$7F,$4B,$85,$9F,$C4,$E1,$FA,$AC,$3B,$57,$EA,$B1,$C3,$E7,$21
	.dc.b	$D8,$FD,$1E,$1C,$FB,$BF,$47,$8D,$97,$C3,$3A,$99,$26,$67,$FE,$C8
	.dc.b	$ED,$FF,$17,$C7,$07,$FC,$3E,$4C,$78,$42,$78,$7C,$1E,$1B,$1D,$27
	.dc.b	$FD,$23,$5F,$E3,$28,$C1,$D7,$E1,$78,$4E,$BE,$12,$CF,$85,$B9,$AD
	.dc.b	$18,$5F,$09,$6E,$5F,$6B,$E1,$1B,$F8,$4E,$19,$CA,$D6,$1E,$57,$D9
	.dc.b	$5E,$8C,$BB,$65,$27,$6F,$87,$2F,$17,$C2,$78,$FE,$12,$57,$48,$F3
	.dc.b	$7C,$27,$9F,$E0,$E4,$F8,$23,$D5,$F0,$9E,$BF,$82,$93,$E4,$8E,$AF
	.dc.b	$32,$1E,$63,$EB,$7C,$1F,$67,$D7,$7D,$9D,$70,$FC,$9C,$3B,$C7,$C0
	.dc.b	$C3,$BB,$AD,$D8,$FC,$DF,$82,$93,$F1,$9F,$A1,$F0,$7F,$47,$E1,$3E
	.dc.b	$97,$C1,$CA,$F9,$71,$F5,$7E,$12,$47,$CE,$7E,$C7,$C1,$F2,$8E,$84
	.dc.b	$3C,$57,$D5,$FA,$B7,$D5,$F3,$5F,$47,$D1,$77,$A3,$F3,$58,$75,$9E
	.dc.b	$67,$DE,$C3,$F1,$1F,$09,$D6,$3E,$06,$3D,$6F,$F4,$58,$F3,$E7,$FC
	.dc.b	$CE,$56,$FF,$49,$97,$E5,$3F,$2E,$FF,$0D,$D7,$FB,$B9,$B3,$FB,$06
	.dc.b	$F0,$9F,$F1,$E4,$67,$E1,$39,$7C,$14,$8F,$CC,$4A,$7B,$7C,$1C,$9C
	.dc.b	$BE,$1F,$81,$F1,$7C,$27,$8F,$E1,$65,$72,$8F,$35,$F5,$8F,$C1,$E0
	.dc.b	$C7,$31,$D0,$86,$B1,$EC,$F9,$3E,$13,$E0,$7C,$1C,$8C,$1E,$8F,$84
	.dc.b	$F8,$84,$7D,$7F,$0B,$F1,$FE,$16,$47,$B9,$F9,$5F,$0B,$F2,$F0,$93
	.dc.b	$D0,$F3,$FC,$27,$37,$D3,$85,$E6,$71,$FD,$1C,$BA,$D7,$BB,$2B,$5A
	.dc.b	$8F,$A9,$97,$D9,$F8,$59,$1F,$41,$FB,$9F,$09,$F7,$7E,$17,$EF,$7C
	.dc.b	$2C,$AF,$AB,$1F,$81,$F8,$3F,$C1,$7C,$14,$9F,$B3,$1D,$1F,$9F,$85
	.dc.b	$CB,$C5,$D3,$F8,$3C,$BB,$57,$E0,$32,$EC,$FD,$6C,$BB,$3F,$84,$CB
	.dc.b	$B3,$F6,$B2,$F3,$BE,$92,$1F,$7A,$FA,$BF,$80,$BE,$B8,$F8,$B9,$58
	.dc.b	$F6,$AF,$C2,$E3,$DA,$B5,$8F,$E7,$FE,$0F,$F4,$1F,$0B,$2F,$F2,$F1
	.dc.b	$FC,$DB,$F2,$EF,$F1,$96,$A7,$F9,$C4,$FF,$33,$6B,$FC,$51,$D9,$FE
	.dc.b	$29,$75,$F0,$9C,$CE,$9F,$0B,$E1,$8E,$E4,$39,$7C,$1F,$85,$F1,$79
	.dc.b	$11,$F0,$1C,$CE,$A4,$3C,$53,$E9,$F8,$49,$1E,$77,$D9,$F0,$9E,$DF
	.dc.b	$84,$F7,$7C,$2C,$AF,$64,$7C,$1F,$85,$F8,$5F,$0B,$27,$E0,$47,$27
	.dc.b	$A1,$0F,$85,$7D,$5F,$8D,$7D,$5F,$8D,$7D,$5F,$95,$7D,$5F,$95,$7F
	.dc.b	$47,$E1,$3E,$97,$C2,$C9,$FA,$11,$F5,$7E,$17,$EB,$7C,$2C,$9F,$A9
	.dc.b	$1F,$67,$E0,$A4,$7D,$77,$EE,$7C,$2F,$58,$E6,$43,$ED,$7C,$2F,$57
	.dc.b	$EF,$5F,$57,$EF,$5F,$47,$F0,$17,$7A,$7F,$47,$CA,$C7,$F1,$58,$F7
	.dc.b	$BF,$C3,$65,$87,$CE,$76,$21,$D8,$F4,$10,$EE,$7E,$46,$7F,$2D,$F0
	.dc.b	$92,$3F,$28,$FE,$67,$E1,$3F,$35,$F0,$72,$7F,$30,$FE,$77,$E1,$7F
	.dc.b	$C2,$7F,$97,$7F,$89,$BA,$CD,$27,$C4,$73,$BE,$A4,$3F,$41,$87,$57
	.dc.b	$F4,$58,$5D,$BE,$1E,$3D,$BC,$99,$5F,$E1,$BC,$7A,$B7,$E1,$7C,$7F
	.dc.b	$0B,$23,$06,$FE,$17,$CE,$47,$97,$C2,$FA,$BE,$1B,$D7,$F0,$B2,$7B
	.dc.b	$47,$BB,$09,$1D,$0F,$F0,$5E,$2F,$86,$E9,$18,$C6,$EB,$FC,$7A,$BF
	.dc.b	$C6,$B3,$D0,$F5,$63,$DD,$E7,$8F,$CC,$F8,$49,$1E,$F7,$E7,$FC,$2F
	.dc.b	$D0,$F8,$4F,$A3,$F0,$B2,$BE,$24,$7D,$4F,$86,$91,$F1,$DF,$AF,$F0
	.dc.b	$DC,$A3,$A1,$0F,$5D,$F5,$7E,$75,$F5,$7D,$F7,$D1,$F8,$37,$74,$F7
	.dc.b	$61,$F8,$4F,$86,$ED,$5D,$0F,$B0,$F5,$7E,$4E,$1D,$6B,$E3,$61,$CF
	.dc.b	$8F,$FC,$E9,$D1,$F8,$CC,$B1,$87,$F9,$EB,$76,$7F,$92,$F8,$49,$3F
	.dc.b	$85,$7F,$2B,$F0,$BF,$96,$F8,$5F,$CB,$FC,$2C,$AF,$C5,$47,$E6,$BE
	.dc.b	$1A,$47,$E3,$5F,$F3,$7F,$C2,$7F,$86,$32,$61,$D2,$BF,$24,$EA,$7B
	.dc.b	$47,$E5,$70,$EC,$FE,$1B,$0E,$AF,$E2,$30,$BB,$FF,$62,$BF,$F1,$29
	.dc.b	$FE,$1F,$9F,$F3,$C3,$FE,$15,$6F,$C1,$F9,$3E,$13,$CB,$F0,$92,$76
	.dc.b	$FA,$3E,$1B,$D3,$F0,$92,$79,$BE,$CF,$85,$F6,$FC,$27,$39,$EF,$86
	.dc.b	$1E,$64,$3F,$CD,$39,$31,$ED,$1D,$08,$6B,$0E,$CF,$AB,$E1,$3E,$5F
	.dc.b	$C2,$48,$C1,$EE,$F8,$4F,$9F,$F0,$B2,$30,$7C,$2F,$85,$FA,$7F,$0B
	.dc.b	$23,$07,$C6,$F8,$3F,$AF,$8C,$8C,$1C,$8F,$4D,$F5,$7D,$78,$7D,$EC
	.dc.b	$3A,$47,$BB,$0C,$3F,$4B,$E1,$7F,$07,$8C,$8C,$1D,$78,$FE,$D6,$3D
	.dc.b	$5F,$B9,$8F,$29,$E8,$7D,$CC,$31,$3F,$81,$F8,$3F,$C7,$92,$7F,$03
	.dc.b	$F0,$9F,$92,$F8,$39,$1F,$84,$7F,$2B,$F0,$9F,$EC,$4F,$CB,$BF,$C1
	.dc.b	$E7,$F8,$33,$0F,$F1,$6B,$7C,$2F,$D5,$BF,$E1,$4E,$46,$3D,$F1,$FB
	.dc.b	$B9,$7E,$82,$3B,$C7,$E0,$32,$B8,$FE,$5F,$2F,$04,$F8,$7E,$0E,$3F
	.dc.b	$EC,$FC,$BC,$7F,$0B,$E5,$F8,$49,$39,$8F,$47,$C2,$FA,$7E,$16,$4F
	.dc.b	$18,$F6,$7C,$2F,$B7,$E1,$24,$F5,$8F,$81,$F0,$BF,$07,$09,$3E,$18
	.dc.b	$EA,$F8,$F0,$EA,$FF,$B0,$B2,$63,$DA,$3D,$18,$F6,$75,$8F,$CD,$F8
	.dc.b	$4F,$9D,$F0,$B2,$7D,$D1,$F4,$7D,$F9,$1F,$05,$FA,$9F,$0B,$F5,$7E
	.dc.b	$13,$EB,$7C,$2C,$AF,$8D,$1F,$67,$E1,$7B,$47,$43,$E5,$3D,$1F,$45
	.dc.b	$DE,$4F,$BB,$87,$59,$FA,$18,$75,$7D,$D8,$75,$79,$90,$F8,$5F,$07
	.dc.b	$CF,$AB,$F0,$B8,$D9,$7A,$4E,$16,$5D,$E7,$E5,$65,$DF,$3F,$C7,$E7
	.dc.b	$F9,$3F,$84,$FC,$A7,$C2,$4B,$FC,$24,$7E,$5F,$E0,$E4,$7E,$1D,$FC
	.dc.b	$D7,$C1,$FE,$6F,$E0,$E4,$FE,$2D,$33,$1F,$F4,$4F,$DB,$7F,$84,$47
	.dc.b	$C2,$F8,$67,$B9,$BE,$1F,$E9,$73,$A8,$F0,$10,$FC,$26,$3E,$27,$C0
	.dc.b	$43,$F1,$18,$78,$96,$1E,$2B,$C3,$E4,$F8,$3B,$1E,$06,$CF,$CC,$C1
	.dc.b	$18,$E1,$43,$E1,$70,$BE,$0E,$92,$FD,$87,$3F,$85,$DF,$C1,$C8,$CE
	.dc.b	$1D,$23,$9F,$C2,$74,$C6,$56,$B1,$F0,$37,$E1,$7B,$3E,$ED,$3E,$4E
	.dc.b	$78,$D7,$E8,$E9,$DE,$B4,$7B,$73,$C4,$FD,$6C,$FD,$7F,$0B,$EC,$F8
	.dc.b	$59,$7E,$18,$F7,$FC,$1F,$C0,$F8,$49,$3E,$58,$F8,$7F,$09,$F1,$3E
	.dc.b	$12,$4F,$A6,$3E,$3F,$C2,$F6,$8E,$87,$B5,$EA,$FB,$EF,$AB,$E8,$E3
	.dc.b	$FA,$38,$76,$8F,$89,$87,$67,$5C,$4C,$7B,$C7,$52,$1A,$F8,$5E,$EF
	.dc.b	$0C,$7B,$ED,$F9,$8C,$7B,$4F,$4C,$7B,$4F,$E7,$30,$FC,$0F,$C2,$7E
	.dc.b	$0B,$E0,$E5,$FD,$78,$FC,$2F,$C2,$48,$FB,$4F,$E2,$3E,$13,$F1,$9F
	.dc.b	$97,$3E,$26,$92,$BF,$0B,$0F,$5F,$7F,$05,$E0,$9E,$C6,$9E,$CF,$C2
	.dc.b	$C2,$E9,$FB,$2C,$7A,$4F,$CB,$C3,$C0,$F6,$C7,$C7,$8F,$93,$E1,$BC
	.dc.b	$BF,$07,$2B,$C7,$1E,$8F,$86,$91,$E6,$7D,$7F,$09,$EC,$F8,$6F,$6F
	.dc.b	$C3,$4A,$F5,$C7,$57,$99,$0F,$27,$C2,$F4,$7C,$D7,$76,$78,$3F,$E5
	.dc.b	$0D,$99,$77,$BF,$8F,$97,$77,$DD,$97,$CE,$F8,$69,$1F,$31,$FA,$3F
	.dc.b	$05,$F4,$BE,$0F,$E9,$FC,$34,$AF,$A3,$1F,$5B,$E1,$24,$7D,$47,$EC
	.dc.b	$FC,$2F,$28,$FA,$F7,$5F,$9D,$F0,$DD,$5F,$B7,$87,$57,$E9,$61,$D1
	.dc.b	$FA,$98,$5D,$FD,$F1,$FC,$FB,$66,$5F,$88,$C7,$B5,$6B,$C1,$53,$3D
	.dc.b	$61,$EA,$3C,$24,$3C,$27,$AC,$87,$88,$FC,$8E,$1F,$96,$F8,$6F,$CB
	.dc.b	$FC,$34,$9F,$CA,$C7,$E6,$BE,$1B,$F3,$7F,$0D,$27,$F3,$31,$FC,$A7
	.dc.b	$D2,$FE,$69,$1F,$BB,$6E,$9E,$96,$F2,$FF,$A3,$1F,$F1,$24,$FF,$5E
	.dc.b	$5F,$09,$E3,$8F,$09,$FC,$AA,$32,$F8,$0F,$D4,$E1,$E3,$7F,$41,$87
	.dc.b	$8D,$87,$13,$CD,$F0,$DE,$7F,$86,$91,$D5,$F5,$7C,$27,$AF,$E1,$64
	.dc.b	$F8,$5F,$77,$C2,$7B,$FE,$17,$E0,$7C,$34,$AF,$34,$72,$7A,$10,$F8
	.dc.b	$97,$D1,$F8,$D7,$77,$F8,$D8,$75,$9F,$81,$87,$57,$E6,$61,$F3,$FE
	.dc.b	$17,$E8,$7C,$34,$9F,$8D,$1F,$4F,$E1,$3E,$A7,$C1,$C9,$F9,$51,$F5
	.dc.b	$FE,$0A,$47,$CD,$7E,$D7,$C1,$F5,$8E,$64,$3E,$E7,$C3,$75,$7E,$A5
	.dc.b	$F5,$7F,$01,$7D,$1F,$C1,$5D,$D7,$F0,$58,$7E,$1F,$09,$5F,$75,$FC
	.dc.b	$57,$C3,$75,$8F,$C5,$63,$D6,$FF,$15,$8F,$3E,$7F,$C5,$65,$6F,$F1
	.dc.b	$59,$7E,$4F,$E1,$3F,$29,$F0,$B2,$7F,$15,$1F,$97,$F8,$4F,$CC,$7C
	.dc.b	$2C,$9F,$C7,$47,$E6,$FE,$17,$FC,$09,$F9,$77,$ED,$A7,$FA,$B3,$75
	.dc.b	$FC,$E4,$F7,$9E,$A4,$37,$D9,$DD,$FC,$D6,$1D,$DF,$D3,$61,$DD,$FF
	.dc.b	$03,$E1,$D7,$5F,$F7,$8B,$63,$F4,$D8,$79,$3E,$1A,$47,$07,$CF,$F0
	.dc.b	$DE,$8F,$85,$F4,$FC,$2C,$AE,$B1,$EC,$F8,$5F,$6F,$C2,$C9,$F0,$C7
	.dc.b	$27,$C7,$75,$F8,$3F,$0D,$D5,$F3,$E1,$D5,$F8,$B8,$75,$7D,$78,$75
	.dc.b	$7E,$4E,$1F,$33,$E1,$7E,$6F,$C3,$49,$F8,$31,$F4,$3E,$0A,$47,$C4
	.dc.b	$7E,$9F,$C1,$FD,$4F,$82,$93,$F2,$1F,$AF,$F0,$7F,$63,$E1,$BB,$4F
	.dc.b	$43,$E6,$BD,$1F,$B7,$77,$9F,$F9,$77,$1B,$1E,$B5,$F8,$0C,$6D,$6B
	.dc.b	$1E,$D3,$D0,$87,$E1,$BE,$17,$A7,$2F,$DA,$CA,$CF,$F9,$F3,$2E,$CF
	.dc.b	$A4,$87,$73,$F0,$77,$F9,$2F,$84,$91,$F8,$67,$F2,$BF,$0B,$F9,$6F
	.dc.b	$84,$FC,$BF,$C2,$4A,$FC,$5C,$7F,$97,$7F,$2E,$FF,$01,$C7,$F8,$BD
	.dc.b	$BA,$FF,$4D,$6E,$9F,$E7,$EF,$5D,$1F,$E1,$5A,$FC,$AE,$3E,$06,$9F
	.dc.b	$E5,$E8,$7C,$37,$8E,$3F,$37,$97,$8D,$FD,$4E,$5E,$3F,$86,$91,$C5
	.dc.b	$F3,$7C,$37,$9F,$E1,$64,$F5,$7D,$5F,$0D,$EB,$F8,$5F,$67,$C2,$CA
	.dc.b	$F1,$47,$BF,$E1,$3E,$07,$C2,$C9,$F3,$47,$27,$A1,$0F,$89,$7D,$5F
	.dc.b	$65,$F5,$7E,$45,$F5,$7E,$05,$F5,$7E,$65,$FC,$FF,$86,$91,$F1,$5F
	.dc.b	$A5,$F0,$DF,$4F,$E0,$64,$FC,$97,$EB,$7C,$2F,$D7,$F8,$69,$3F,$35
	.dc.b	$FB,$5F,$0D,$CA,$3A,$10,$FB,$97,$D1,$FB,$D7,$7A,$3E,$F6,$1D,$27
	.dc.b	$F0,$58,$59,$FC,$16,$1F,$87,$F8,$4E,$D5,$D0,$FB,$CF,$3E,$9F,$C3
	.dc.b	$73,$FF,$84,$E7,$D4,$7E,$0F,$1E,$93,$AC,$6C,$FF,$B7,$29,$87,$E5
	.dc.b	$7E,$1B,$F2,$DF,$09,$2B,$F1,$91,$F9,$9F,$85,$91,$F8,$F7,$F3,$9F
	.dc.b	$07,$F9,$DF,$84,$FF,$5D,$7E,$5D,$FE,$0E,$AF,$F7,$4B,$7B,$FF,$37
	.dc.b	$3D,$E7,$A9,$0F,$F4,$94,$FF,$85,$27,$F4,$37,$E0,$9F,$D4,$3E,$07
	.dc.b	$F4,$15,$E0,$61,$83,$FD,$83,$5F,$E1,$87,$AF,$C3,$79,$8F,$3F,$C2
	.dc.b	$FA,$3E,$17,$D3,$F0,$B2,$B8,$C7,$B3,$E1,$64,$74,$7D,$FF,$09,$CA
	.dc.b	$3B,$E1,$D5,$F0,$E1,$D5,$F1,$E1,$D5,$F7,$E1,$D1,$F8,$38,$5E,$6F
	.dc.b	$F2,$86,$4C,$BF,$DD,$34,$FC,$79,$E7,$E1,$B1,$FE,$E4,$35,$19,$F4
	.dc.b	$24,$67,$F4,$5F,$E0,$ED,$F2,$32,$FF,$B6,$8E,$D3,$E0,$97,$89,$4D
	.dc.b	$E7,$FD,$D7,$15,$63,$A9,$84,$9C,$BE,$56,$BD,$DB,$5C,$EB,$94,$F1
	.dc.b	$6B,$E4,$39,$38,$9E,$35,$E0,$9D,$1D,$67,$E3,$79,$BB,$79,$E3,$FC
	.dc.b	$C2,$7B,$1D,$08,$7C,$37,$F9,$93,$F9,$6F,$87,$FC,$FF,$C3,$48,$FC
	.dc.b	$7C,$7E,$8B,$E1,$FF,$47,$F0,$92,$7F,$27,$1F,$A6,$F8,$5F,$D3,$FC
	.dc.b	$2C,$9F,$CB,$C7,$EA,$BE,$17,$F5,$7F,$0F,$27,$F3,$71,$FA,$EF,$87
	.dc.b	$91,$F9,$E7,$F6,$3F,$0D,$FB,$2F,$87,$FD,$9F,$C2,$CA,$FD,$1C,$7E
	.dc.b	$DB,$E1,$BF,$6F,$F0,$D2,$7F,$4F,$1F,$27,$9B,$FD,$41,$E0,$7F,$88
	.dc.b	$7C,$3F,$F4,$2F,$C3,$E7,$A3,$FA,$B4,$A3,$18,$2B,$FD,$17,$58,$7F
	.dc.b	$D1,$BF,$0F,$6C,$D3,$1F,$0F,$9F,$85,$98,$7E,$2B,$2F,$1C,$71,$F8
	.dc.b	$6E,$5F,$0B,$2B,$59,$F9,$23,$B7,$C3,$77,$F8,$79,$5A,$3C,$DA,$23
	.dc.b	$C9,$E1,$F8,$25,$9F,$81,$E1,$B7,$FA,$9B,$4F,$14,$63,$06,$5E,$38
	.dc.b	$E3,$9F,$8D,$D6,$7E,$27,$27,$8B,$3C,$65,$F4,$33,$F1,$3E,$19,$DE
	.dc.b	$7E,$17,$C1,$A7,$79,$FA,$F9,$5C,$BE,$AE,$5D,$1F,$AF,$95,$C3,$E0
	.dc.b	$E5,$F4,$FB,$E4,$7A,$DF,$AD,$E0,$7D,$7F,$86,$FB,$1F,$0D,$2B,$E0
	.dc.b	$47,$DB,$F8,$49,$1F,$0D,$FB,$DF,$0F,$CA,$3A,$10,$F5,$F0,$B0,$ED
	.dc.b	$1F,$2F,$0E,$CE,$B0,$EA,$F0,$C2,$EF,$F9,$6C,$7B,$4F,$4C,$BD,$6F
	.dc.b	$6C,$7A,$BE,$0C,$7A,$DF,$E7,$B1,$E7,$AF,$E7,$32,$B7,$F8,$6C,$BF
	.dc.b	$31,$F0,$F2,$3F,$00,$FE,$6F,$E1,$BF,$39,$F0,$DF,$9D,$F8,$59,$5F
	.dc.b	$85,$8F,$D0,$7C,$34,$8F,$C4,$3F,$A5,$FC,$BF,$F8,$97,$3F,$FB,$CB
	.dc.b	$1E,$D5,$F1,$9E,$CF,$33,$F2,$18,$76,$7E,$5E,$1D,$5F,$9B,$85,$D9
	.dc.b	$AF,$F9,$37,$2E,$FC,$4C,$BB,$D7,$F0,$07,$BE,$1D,$0E,$E4,$30,$76
	.dc.b	$3D,$64,$3F,$2F,$5E,$C3,$F9,$4C,$F8,$A3,$F8,$34,$62,$FC,$9F,$0D
	.dc.b	$F0,$FC,$EF,$89,$E7,$7C,$5F,$86,$98,$7E,$CB,$0F,$EA,$6D,$F8,$5F
	.dc.b	$94,$47,$5F,$0D,$F3,$71,$91,$C5,$EC,$F3,$C3,$AB,$E8,$C3,$C2,$F5
	.dc.b	$21,$EC,$BF,$B1,$7D,$23,$DF,$77,$3F,$77,$0B,$2F,$ED,$F5,$F1,$32
	.dc.b	$EF,$5A,$96,$5E,$0A,$EC,$78,$27,$59,$78,$23,$E1,$7C,$2F,$E2,$48
	.dc.b	$FC,$5F,$86,$FC,$67,$C3,$48,$F9,$0F,$E3,$FE,$1B,$F2,$1F,$0D,$27
	.dc.b	$E6,$3F,$93,$F8,$5E,$71,$F6,$B0,$B9,$79,$F1,$E9,$3F,$7B,$1B,$3E
	.dc.b	$BC,$7A,$4F,$E0,$B1,$B3,$EF,$C7,$F4,$18,$74,$E2,$FC,$A6,$58,$C3
	.dc.b	$CE,$43,$B1,$F8,$1C,$79,$CF,$E6,$F1,$B3,$F8,$CC,$3F,$59,$F0,$BF
	.dc.b	$AD,$F8,$49,$3F,$89,$8F,$D8,$7C,$34,$8F,$C6,$3F,$B3,$F8,$4F,$DA
	.dc.b	$7C,$2C,$9F,$C8,$3F,$B7,$F8,$4E,$B1,$F9,$3C,$3A,$BF,$A0,$C3,$A3
	.dc.b	$FA,$2C,$2E,$FF,$7F,$1E,$93,$FA,$6C,$6C,$FE,$0F,$1E,$B3,$CC,$F4
	.dc.b	$1F,$A1,$F8,$5E,$7A,$FE,$EF,$1B,$3F,$AC,$C7,$9C,$FE,$FF,$1B,$3F
	.dc.b	$B0,$C3,$F8,$87,$C2,$FF,$75,$FC,$BF,$FC,$EC,$7F,$9D,$F9,$91,$9F
	.dc.b	$8A,$D7,$74,$C3,$FB,$1D,$FF,$9D,$C7,$C3,$F3,$F8,$69,$3A,$8E,$EF
	.dc.b	$EE,$70,$AF,$F5,$CA,$7C,$3F,$78,$EB,$87,$77,$58,$F9,$F0,$EB,$1C
	.dc.b	$71,$C3,$D4,$87,$F1,$2C,$7D,$D8,$75,$8F,$06,$1D,$78,$7F,$6F,$97
	.dc.b	$AF,$1E,$B1,$8C,$7A,$DF,$9F,$E1,$FE,$41,$1F,$87,$F0,$DF,$2F,$E0
	.dc.b	$E4,$7C,$67,$E7,$7C,$27,$CF,$F8,$3F,$A1,$F0,$92,$BE,$5C,$73,$7B
	.dc.b	$1F,$3B,$0F,$AD,$87,$48,$F9,$18,$5C,$BD,$38,$F5,$9F,$AB,$8F,$57
	.dc.b	$E7,$63,$D6,$BD,$F8,$DE,$4F,$F3,$BF,$86,$DB,$FD,$21,$87,$A4,$87
	.dc.b	$80,$FC,$0E,$5D,$DF,$B5,$97,$7C,$7E,$5E,$5F,$8D,$F8,$39,$5F,$88
	.dc.b	$7F,$21,$F0,$9F,$91,$F8,$3F,$C9,$7C,$24,$AF,$C6,$C7,$E5,$7E,$12
	.dc.b	$47,$E4,$1F,$CC,$7C,$27,$48,$FC,$56,$17,$2F,$B7,$8F,$59,$FC,$BE
	.dc.b	$3D,$5F,$C8,$63,$D1,$FC,$96,$36,$FF,$07,$8F,$9A,$7D,$04,$3F,$31
	.dc.b	$87,$57,$F3,$58,$75,$AF,$C5,$E1,$CF,$7F,$9B,$B3,$2C,$61,$FE,$67
	.dc.b	$86,$5F,$B3,$F8,$39,$3F,$AE,$7F,$6D,$F0,$9F,$B7,$F8,$39,$3F,$B2
	.dc.b	$7F,$75,$F0,$9F,$E6,$FF,$CB,$FF,$CF,$04,$CF,$1F,$0B,$D2,$FF,$61
	.dc.b	$8D,$DB,$F4,$39,$74,$9F,$DA,$65,$67,$F4,$B9,$75,$C6,$57,$F9,$93
	.dc.b	$2E,$F9,$79,$A3,$D0,$43,$F7,$D8,$F3,$AF,$87,$8D,$9F,$85,$87,$47
	.dc.b	$FC,$BF,$87,$A5,$ED,$87,$B7,$E1,$7D,$DF,$0B,$27,$C5,$1F,$07,$E1
	.dc.b	$64,$79,$5F,$89,$F0,$BF,$17,$E1,$7E,$37,$C2,$CA,$F5,$47,$57,$D9
	.dc.b	$87,$57,$C7,$87,$47,$CB,$85,$DF,$FB,$FE,$6C,$BB,$57,$C5,$CB,$B3
	.dc.b	$EC,$CB,$B3,$D0,$F4,$9F,$2F,$E1,$7A,$6F,$FB,$4C,$19,$62,$BF,$D5
	.dc.b	$D1,$DD,$F4,$9D,$C8,$77,$3E,$B7,$C2,$FE,$14,$8F,$D9,$F8,$5F,$C4
	.dc.b	$7C,$2C,$8F,$B8,$FE,$2F,$E1,$3F,$19,$F0,$BF,$8D,$F8,$59,$5F,$81
	.dc.b	$8E,$6F,$D8,$C2,$F0,$7C,$4C,$7A,$4F,$DC,$C6,$CF,$C8,$C7,$AC,$FE
	.dc.b	$2F,$1E,$AF,$E0,$B1,$E7,$A7,$C3,$CA,$CF,$C2,$CB,$9C,$FC,$7C,$AC
	.dc.b	$FC,$6C,$7A,$BF,$8C,$C7,$AE,$5F,$63,$2F,$D4,$7C,$2C,$AF,$CE,$3F
	.dc.b	$AB,$F8,$1F,$D6,$7C,$2F,$EB,$7E,$12,$57,$E8,$63,$F6,$1F,$0B,$23
	.dc.b	$F4,$8F,$EC,$FE,$17,$A4,$7E,$73,$0B,$97,$E2,$31,$EB,$3F,$AB,$C7
	.dc.b	$AB,$FA,$2C,$7A,$3F,$A4,$C6,$DF,$E4,$31,$E7,$B7,$E1,$F2,$B7,$F8
	.dc.b	$6C,$BA,$BF,$AC,$CB,$AE,$1F,$98,$CB,$9E,$1F,$8F,$CA,$DF,$E3,$B2
	.dc.b	$FF,$7D,$7C,$2F,$FA,$58,$9F,$15,$8F,$85,$CC,$6B,$E1,$65,$FF,$BA
	.dc.b	$23,$FC,$C3,$9F,$F9,$91,$FF,$2F,$C6,$7E,$13,$A9,$DB,$E1,$7A,$47
	.dc.b	$1C,$7C,$26,$1F,$F7,$1D,$74,$D3,$FA,$CB,$FE,$84,$8C,$E3,$E4,$7C
	.dc.b	$57,$D9,$F0,$18,$3C,$D9,$76,$7C,$06,$0F,$55,$7B,$1E,$A6,$0E,$47
	.dc.b	$93,$1E,$CF,$9A,$BE,$4F,$C3,$48,$F8,$8F,$CC,$F8,$4F,$9B,$F0,$BF
	.dc.b	$3B,$E1,$A5,$7C,$98,$FA,$3F,$0D,$F4,$BE,$1A,$4F,$CD,$8E,$4F,$42
	.dc.b	$1F,$1A,$FA,$BF,$4E,$FA,$BF,$2A,$FA,$BF,$5E,$FA,$BF,$3A,$FA,$BF
	.dc.b	$43,$0F,$53,$D4,$F3,$1D,$48,$75,$3C,$C7,$52,$1D,$4F,$C3,$7C,$37
	.dc.b	$E3,$BE,$16,$46,$0F,$C5,$7C,$37,$E4,$BE,$16,$46,$0F,$C7,$7C,$2F
	.dc.b	$E5,$88,$FE,$47,$E1,$7F,$33,$87,$48,$FC,$4E,$18,$79,$1F,$8C,$BE
	.dc.b	$AF,$E3,$B0,$E6,$FE,$43,$0B,$CD,$F9,$BC,$79,$4F,$E5,$31,$B3,$F9
	.dc.b	$FF,$85,$FD,$4E,$1D,$23,$F4,$58,$E2,$FD,$24,$3B,$1E,$93,$F5,$5F
	.dc.b	$0D,$FB,$22,$3F,$AD,$F8,$43,$2B,$FD,$A3,$F9,$7E,$23,$1D,$3F,$E3
	.dc.b	$E8,$C4,$27,$6E,$DE,$3F,$0D,$6B,$87,$C2,$DA,$7C,$03,$0F,$F4,$0C
	.dc.b	$76,$32,$E1,$ED,$5C,$9F,$59,$0F,$DD,$E5,$FC,$3B,$2E,$D7,$D6,$6D
	.dc.b	$FA,$32,$F3,$C7,$79,$F1,$65,$AC,$B7,$F0,$B2,$B5,$82,$7B,$47,$87
	.dc.b	$0F,$E2,$99,$62,$3C,$E7,$A8,$F0,$90,$E7,$87,$77,$AE,$1D,$DE,$38
	.dc.b	$FC,$5C,$3D,$3F,$0B,$27,$C6,$FB,$3E,$1B,$DB,$F0,$92,$7C,$EF,$C0
	.dc.b	$F8,$6F,$83,$F0,$DF,$0B,$E1,$A5,$7B,$23,$A3,$E0,$DD,$8E,$37,$FC
	.dc.b	$1E,$5D,$6B,$59,$59,$FC,$3E,$5D,$A7,$E3,$E5,$D9,$E9,$97,$3E,$1F
	.dc.b	$C5,$65,$67,$EF,$E5,$CE,$7F,$1D,$95,$9F,$C1,$E3,$D5,$F3,$63,$D7
	.dc.b	$2F,$CB,$E5,$F7,$BE,$17,$EF,$FC,$34,$BF,$B3,$1F,$82,$F8,$6F,$C1
	.dc.b	$FC,$34,$9F,$BB,$1F,$86,$F8,$5F,$C3,$FC,$34,$9F,$C0,$C7,$E2,$BE
	.dc.b	$1B,$F1,$7F,$0B,$27,$F0,$B1,$F8,$EF,$85,$91,$F8,$87,$F2,$3F,$0B
	.dc.b	$F9,$2F,$85,$FC,$9F,$C2,$CA,$FC,$6C,$7E,$63,$EC,$9C,$B7,$E1,$A5
	.dc.b	$7E,$5E,$72,$DF,$85,$E0,$47,$5F,$0D,$CF,$E0,$BA,$7C,$34,$9E,$51
	.dc.b	$DF,$E1,$7C,$1F,$0D,$27,$B4,$78,$FE,$1B,$C9,$F0,$B2,$7C,$51,$E7
	.dc.b	$F8,$59,$1E,$57,$D5,$F0,$BE,$BF,$85,$F6,$7C,$24,$AF,$54,$7B,$FE
	.dc.b	$16,$47,$B5,$F8,$5F,$0B,$F0,$FE,$16,$4F,$C1,$7E,$37,$C2,$7C,$7F
	.dc.b	$85,$93,$F1,$5F,$95,$F0,$BF,$2F,$E1,$64,$FC,$97,$F8,$2F,$E5,$AF
	.dc.b	$86,$BF,$0B,$2B,$F8,$24,$A3,$7F,$0B,$C3,$E1,$A5,$69,$E7,$F0,$BD
	.dc.b	$3E,$1A,$4F,$27,$BF,$C0,$F8,$3E,$1A,$4F,$67,$C7,$F0,$BD,$23,$F8
	.dc.b	$0E,$7F,$E1,$3F,$43,$EF,$FB,$08,$C7,$89,$FE,$F0,$7F,$C4,$75,$C3
	.dc.b	$2F,$0B,$4F,$5E,$5E,$07,$AE,$57,$1E,$B6,$CB,$C5,$18,$D2,$E5,$DF
	.dc.b	$4E,$B3,$C3,$3B,$3E,$3C,$FA,$3E,$5C,$AE,$7F,$D0,$33,$EC,$FC,$EF
	.dc.b	$86,$FA,$7F,$0B,$23,$07,$D2,$F8,$6F,$AE,$47,$EA,$FC,$1F,$DA,$F8
	.dc.b	$6F,$B7,$F0,$72,$7E,$CC,$7D,$EC,$3A,$47,$CC,$C3,$0F,$52,$1F,$43
	.dc.b	$0F,$C2,$E1,$D6,$3F,$07,$87,$57,$EA,$E1,$C9,$E8,$7E,$22,$F1,$B7
	.dc.b	$CC,$C3,$D1,$3D,$4F,$49,$E6,$3A,$90,$EA,$7E,$4B,$E0,$FF,$2E,$47
	.dc.b	$F2,$BF,$0D,$F9,$AF,$86,$FC,$DF,$C3,$49,$FC,$CC,$7E,$7B,$E1,$64
	.dc.b	$7E,$71,$FD,$0F,$C3,$73,$8F,$C7,$E1,$72,$FC,$1E,$3D,$67,$F4,$78
	.dc.b	$F5,$7F,$2B,$8F,$47,$F2,$F8,$DB,$FC,$5E,$3C,$F8,$BF,$2D,$95,$BF
	.dc.b	$C1,$65,$D5,$FC,$FE,$5D,$70,$FC,$9E,$5C,$F0,$FC,$F6,$56,$FF,$15
	.dc.b	$97,$F7,$EF,$3B,$FB,$E5,$7F,$86,$BA,$51,$8A,$FF,$07,$B6,$33,$F0
	.dc.b	$DC,$7E,$17,$97,$C3,$4B,$FF,$02,$5F,$F7,$F6,$FC,$2F,$7F,$85,$91
	.dc.b	$83,$67,$8C,$87,$2C,$3B,$3D,$2F,$B3,$FD,$FA,$FB,$36,$FA,$B9,$BF
	.dc.b	$74,$74,$8E,$05,$DB,$F8,$2D,$30,$EF,$3D,$70,$EF,$5A,$C7,$9C,$FB
	.dc.b	$B1,$EF,$3C,$B1,$F9,$9F,$0B,$F3,$7E,$16,$5F,$C4,$8F,$A1,$F0,$9F
	.dc.b	$47,$E1,$64,$FC,$88,$FA,$9F,$09,$23,$E5,$BF,$5F,$E1,$7A,$C7,$32
	.dc.b	$1F,$0F,$E1,$3A,$BF,$4A,$FA,$BF,$1E,$FA,$3F,$26,$EE,$7F,$0B,$0E
	.dc.b	$B3,$F3,$71,$B3,$F1,$B1,$EB,$3F,$47,$1E,$B3,$F2,$B1,$E7,$C5,$F8
	.dc.b	$BE,$16,$78,$C3,$F1,$79,$FE,$47,$E1,$64,$7E,$21,$FC,$A7,$C2,$7E
	.dc.b	$57,$E1,$7F,$2D,$F0,$92,$BF,$1B,$1F,$99,$F8,$59,$1F,$90,$7F,$39
	.dc.b	$F0,$BD,$23,$F0,$D8,$5D,$BE,$FE,$3D,$67,$F2,$F8,$F5,$7F,$19,$8F
	.dc.b	$47,$F1,$D8,$DB,$FC,$3E,$3C,$F5,$F7,$65,$6F,$F4,$F9,$75,$7F,$2D
	.dc.b	$97,$5C,$3F,$1F,$97,$3C,$3E,$36,$56,$FF,$5F,$97,$ED,$7E,$16,$47
	.dc.b	$EB,$1F,$DC,$7C,$27,$EE,$7E,$17,$FC,$EF,$F9,$6F,$F8,$72,$BF,$CC
	.dc.b	$2D,$CF,$F3,$B3,$AF,$86,$95,$FE,$79,$A4,$72,$E2,$EF,$3F,$AD,$C6
	.dc.b	$E1,$FA,$6C,$7C,$13,$C7,$4F,$09,$87,$C3,$87,$66,$9D,$F0,$E7,$3F
	.dc.b	$85,$C2,$CF,$F0,$EB,$E5,$87,$F7,$8C,$7D,$57,$FC,$55,$B5,$FC,$53
	.dc.b	$2E,$D5,$BC,$AD,$EF,$2F,$8B,$F0,$B3,$0F,$80,$FC,$8F,$84,$F9,$3F
	.dc.b	$0B,$F2,$BE,$12,$57,$C5,$8F,$9B,$F0,$BF,$3B,$E1,$24,$FC,$98,$E8
	.dc.b	$FD,$0C,$2E,$9E,$9C,$7A,$CF,$CF,$C7,$AB,$F5,$B1,$EA,$FD,$3C,$7A
	.dc.b	$BF,$6B,$1E,$79,$FC,$1C,$AC,$FF,$B0,$B6,$65,$DA,$3F,$03,$97,$6C
	.dc.b	$FE,$56,$7D,$A7,$F0,$B9,$F6,$9F,$9D,$9F,$E2,$BE,$17,$F1,$7F,$0B
	.dc.b	$2F,$F0,$D1,$F8,$EF,$85,$FC,$7F,$C2,$49,$FC,$54,$7E,$4B,$E1,$64
	.dc.b	$7E,$35,$FC,$AF,$C2,$F5,$8E,$64,$3F,$2F,$F0,$BD,$5F,$CA,$5F,$57
	.dc.b	$F3,$77,$D1,$FC,$ED,$DC,$FF,$0D,$87,$59,$FD,$0E,$36,$7F,$15,$8F
	.dc.b	$59,$FD,$2E,$3D,$67,$F1,$D8,$F3,$E2,$FC,$76,$56,$FE,$76,$5F,$AE
	.dc.b	$F8,$5F,$D7,$FC,$2C,$9F,$D5,$47,$EC,$BE,$17,$F6,$7F,$0B,$27,$F5
	.dc.b	$D1,$FB,$6F,$85,$91,$FB,$17,$FD,$DF,$F9,$6F,$F9,$AE,$39,$9F,$B5
	.dc.b	$C3,$B3,$FB,$CC,$3B,$3F,$B9,$C3,$B3,$FC,$03,$0E,$AF,$F0,$2C,$2E
	.dc.b	$BF,$E6,$2C,$7B,$4F,$33,$B1,$0C,$1C,$CE,$C4,$30,$75,$3D,$44,$3F
	.dc.b	$DE,$2F,$68,$F0,$7C,$2F,$B8,$92,$37,$F0,$BF,$0B,$E1,$24,$66,$BD
	.dc.b	$11,$F1,$BE,$13,$E3,$FC,$24,$AD,$70,$FF,$86,$72,$4F,$CC,$F8,$4E
	.dc.b	$F5,$E4,$CE,$E5,$F0,$F3,$EF,$3E,$8C,$EC,$FC,$7C,$FB,$4F,$B3,$1B
	.dc.b	$3F,$2F,$1E,$79,$F8,$72,$B6,$F5,$9A,$34,$EB,$7E,$7D,$2D,$EE,$BF
	.dc.b	$D2,$79,$7D,$7C,$FB,$DF,$B5,$F6,$57,$C0,$38,$7C,$27,$E2,$BE,$16
	.dc.b	$46,$73,$F0,$CF,$E3,$BE,$13,$F1,$FF,$03,$2B,$59,$A7,$F2,$7F,$03
	.dc.b	$F9,$4F,$84,$95,$A8,$FC,$BE,$1D,$A3,$EF,$E5,$87,$9D,$F5,$7F,$07
	.dc.b	$8F,$85,$E4,$78,$30,$EA,$FE,$23,$0E,$4F,$43,$CB,$78,$D7,$F3,$58
	.dc.b	$75,$7F,$39,$87,$5A,$E4,$75,$21,$83,$F2,$BF,$07,$FB,$02,$3F,$0B
	.dc.b	$E1,$3F,$67,$F0,$92,$3E,$2B,$FB,$6F,$84,$FD,$BF,$C1,$FE,$E3,$E1
	.dc.b	$25,$7C,$A8,$E6,$F6,$3E,$6E,$1F,$BE,$C3,$A4,$7E,$9B,$0B,$97,$ED
	.dc.b	$71,$EB,$3F,$53,$1E,$AF,$EB,$31,$EB,$5F,$BB,$C6,$F1,$FE,$87,$2E
	.dc.b	$6F,$E8,$F2,$B8,$7E,$37,$2E,$AF,$ED,$B2,$EB,$8F,$F0,$5C,$BF,$89
	.dc.b	$FC,$27,$F1,$4F,$84,$97,$F8,$68,$FD,$C2,$AF,$F7,$6F,$E5,$9F,$E8
	.dc.b	$53,$FC,$D5,$1F,$91,$6B,$BC,$FF,$D1,$D1,$BF,$83,$E7,$5C,$BE,$12
	.dc.b	$D6,$9B,$7F,$AB,$BF,$F3,$FD,$7F,$1B,$C3,$CB,$3F,$C4,$31,$F4,$BE
	.dc.b	$13,$8F,$3A,$FC,$77,$E8,$F8,$5C,$C7,$8E,$BD,$18,$E0,$CE,$7D,$23
	.dc.b	$B1,$FA,$8C,$71,$87,$C2,$8E,$78,$47,$C4,$7C,$3A,$E8,$F6,$30,$7C
	.dc.b	$7F,$84,$F1,$91,$F9,$5F,$0B,$E6,$F8,$4F,$3F,$C2,$C9,$F9,$D1,$EA
	.dc.b	$F8,$3F,$5F,$C2,$49,$FA,$51,$CD,$FA,$47,$CB,$C3,$19,$FD,$AC,$7B
	.dc.b	$4F,$D9,$C7,$B3,$C3,$7F,$95,$97,$78,$FB,$D9,$77,$75,$BB,$3F,$9B
	.dc.b	$8F,$6E,$1F,$C0,$E9,$8B,$F4,$9D,$C8,$72,$C7,$B3,$D3,$1E,$DA,$7E
	.dc.b	$43,$1F,$B3,$F0,$9F,$6B,$E1,$25,$FE,$46,$3E,$EF,$C2,$C8,$FC,$A3
	.dc.b	$F8,$0F,$85,$FC,$0F,$C2,$C9,$FC,$C3,$F8,$4F,$84,$EB,$1F,$9B,$C3
	.dc.b	$AB,$EE,$C3,$A3,$F0,$30,$B9,$7E,$93,$1E,$93,$F1,$31,$B3,$FA,$8C
	.dc.b	$7A,$CF,$C8,$C7,$AC,$FE,$B3,$1E,$7B,$7E,$A3,$2B,$7F,$A3,$CB,$9C
	.dc.b	$FE,$B3,$2B,$3F,$A7,$CB,$F3,$BF,$0B,$F9,$EF,$85,$93,$FB,$58,$FD
	.dc.b	$0F,$C2,$FE,$8B,$E1,$64,$FE,$E6,$3F,$4B,$F0,$B2,$3F,$78,$FE,$A3
	.dc.b	$E1,$7A,$C7,$EF,$F0,$EA,$FE,$0B,$0E,$AF,$F0,$3C,$3A,$BF,$86,$C3
	.dc.b	$A3,$F8,$8C,$2E,$FF,$C3,$31,$EB,$3F,$8C,$C7,$AC,$FF,$10,$C7,$AC
	.dc.b	$FE,$43,$1E,$B3,$FC,$53,$1E,$7A,$7F,$10,$CA,$DF,$F0,$BC,$BF,$82
	.dc.b	$F6,$7F,$A3,$3A,$0F,$FA,$92,$7F,$B5,$35,$FD,$13,$B1,$F0,$9C,$89
	.dc.b	$1C,$3E,$13,$AF,$C1,$C8,$E6,$F8,$3E,$0B,$C3,$9F,$8B,$E1,$65,$78
	.dc.b	$23,$9B,$E0,$C2,$EC,$9F,$F1,$56,$7E,$8C,$BB,$5F,$9B,$2B,$7B,$8F
	.dc.b	$F0,$A6,$6C,$FC,$31,$EE,$CF,$C2,$FB,$B2,$EE,$76,$AF,$7C,$7C,$4C
	.dc.b	$FE,$4E,$5D,$AB,$59,$5A,$EF,$8F,$CE,$F8,$49,$1F,$31,$FA,$3F,$07
	.dc.b	$F4,$BE,$12,$4F,$D0,$7E,$AF,$C1,$FD,$6F,$84,$FA,$FF,$05,$2B,$EA
	.dc.b	$C7,$DA,$F8,$3F,$B7,$F0,$72,$7E,$CC,$7D,$EF,$83,$FB,$FF,$09,$27
	.dc.b	$EE,$C7,$E0,$BE,$0F,$F0,$7F,$09,$27,$F0,$31,$F8,$6F,$83,$91,$F8
	.dc.b	$47,$F1,$3F,$09,$F8,$AF,$84,$93,$F8,$87,$FA,$8F,$D6,$FF,$98,$F9
	.dc.b	$7F,$CC,$1F,$3A,$FD,$AD,$E4,$FF,$5E,$8C,$79,$16,$4D,$9F,$0E,$3F
	.dc.b	$92,$CE,$F7,$7D,$DE,$FF,$CC,$EB,$E2,$AF,$CB,$EB,$6B,$5A,$F8,$E5
	.dc.b	$A7,$8F,$2F,$06,$72,$7F,$C9,$D3,$97,$FC,$99,$4F,$85,$F6,$92,$33
	.dc.b	$F0,$BF,$03,$E1,$64,$70,$7E,$1F,$C3,$FC,$4F,$86,$F8,$BF,$0F,$2B
	.dc.b	$AC,$7C,$8F,$86,$F9,$3F,$09,$27,$C3,$1F,$33,$E1,$24,$79,$1F,$9F
	.dc.b	$F0,$9F,$43,$E1,$7E,$8F,$C3,$4A,$F4,$C7,$D4,$F8,$69,$1E,$C7,$EB
	.dc.b	$FC,$37,$D8,$F8,$6F,$B3,$F0,$D2,$BE,$0C,$7D,$CF,$85,$FB,$BF,$0D
	.dc.b	$27,$E2,$C7,$E0,$3E,$1B,$F0,$3F,$0D,$27,$E4,$C7,$F8,$53,$F2,$CF
	.dc.b	$F1,$5C,$7F,$6D,$6F,$67,$DB,$6F,$85,$FB,$E7,$FC,$32,$DF,$86,$E5
	.dc.b	$EA,$4B,$FF,$40,$57,$F8,$9A,$B8,$7C,$1F,$7F,$84,$95,$A8,$F1,$7C
	.dc.b	$24,$8E,$2F,$97,$E1,$BC,$DF,$09,$27,$AB,$E9,$F8,$5F,$55,$F4,$9F
	.dc.b	$56,$18,$79,$90,$F7,$61,$D5,$E6,$43,$E1,$5F,$57,$CC,$7B,$27,$B6
	.dc.b	$BF,$16,$FD,$13,$D0,$ED,$1F,$27,$1F,$44,$74,$3B,$47,$CD,$C7,$E9
	.dc.b	$7C,$2F,$D3,$F1,$A5,$7C,$A8,$EB,$5F,$4F,$3B,$A7,$D3,$C7,$ED,$78
	.dc.b	$DF,$6F,$E1,$66,$3F,$4A,$3A,$BF,$53,$2E,$AF,$DF,$CB,$9B,$FE,$C7
	.dc.b	$DD,$96,$3C,$A4,$32,$EF,$5F,$7B,$2E,$EF,$E2,$72,$EE,$FE,$2F,$2E
	.dc.b	$FA,$7E,$2F,$3E,$98,$F6,$AF,$8F,$A7,$AD,$F4,$1D,$48,$75,$3F,$1B
	.dc.b	$F0,$BF,$99,$F8,$59,$18,$39,$9E,$BD,$BB,$EB,$F9,$FF,$84,$93,$F9
	.dc.b	$67,$F4,$5F,$0B,$FA,$3C,$39,$CF,$E8,$F0,$C3,$D3,$83,$E1,$E3,$D2
	.dc.b	$39,$10,$FD,$56,$1D,$1F,$21,$FA,$DB,$E6,$FE,$BF,$E1,$3E,$1D,$73
	.dc.b	$8E,$27,$3D,$BF,$65,$7C,$E3,$F6,$DF,$0B,$FB,$9F,$85,$91,$FB,$08
	.dc.b	$E6,$F9,$F7,$FB,$0F,$ED,$3E,$17,$FD,$27,$F9,$67,$F8,$3C,$FF,$08
	.dc.b	$4A,$3F,$82,$57,$4A,$FE,$09,$78,$7A,$65,$E7,$CB,$F8,$35,$F4,$E1
	.dc.b	$95,$DC,$F2,$1F,$C1,$2F,$A5,$7F,$0D,$C3,$CF,$3F,$A9,$3F,$89,$70
	.dc.b	$FF,$15,$C3,$9C,$7F,$A0,$73,$F3,$C7,$AB,$E1,$A4,$8D,$3E,$EF,$86
	.dc.b	$EB,$8F,$EA,$67,$F8,$AE,$1F,$0B,$E1,$A4,$71,$9F,$EB,$58,$A7,$E3
	.dc.b	$FC,$37,$6A,$EA,$6F,$5F,$F1,$0D,$7F,$8D,$6F,$8E,$3E,$09,$FE,$51
	.dc.b	$B7,$F0,$FB,$F5,$63,$E0,$69,$CC,$71,$B1,$65,$E1,$CB,$CB,$A5,$CB
	.dc.b	$C3,$9F,$7C,$7F,$40,$7B,$8E,$98,$F3,$CB,$C1,$87,$B9,$F6,$CE,$B2
	.dc.b	$FF,$1A,$37,$E1,$BF,$0E,$7E,$23,$E1,$BB,$61,$FA,$29,$EC,$7D,$5F
	.dc.b	$86,$FC,$6F,$C3,$48,$E9,$1C,$9E,$84,$3E,$E5,$74,$E4,$FE,$09,$1F
	.dc.b	$82,$38,$9E,$5A,$E8,$F7,$39,$67,$F6,$31,$ED,$1D,$0F,$CB,$9F,$A3
	.dc.b	$E1,$FD,$0E,$3C,$AB,$EF,$DE,$2B,$E1,$7C,$37,$E9,$BE,$16,$46,$0E
	.dc.b	$67,$E5,$EB,$F9,$F3,$F2,$7E,$1B,$F5,$BF,$0B,$23,$07,$13,$F3,$55
	.dc.b	$D1,$FC,$E5,$FC,$1E,$5F,$C3,$61,$F9,$2C,$39,$57,$E8,$70,$FD,$CC
	.dc.b	$7E,$EA,$F9,$57,$E6,$30,$C5,$F9,$A3,$F4,$0F,$CB,$3F,$55,$5D,$23
	.dc.b	$F3,$F7,$D2,$7F,$05,$F0,$DF,$C1,$BE,$16,$46,$0E,$9C,$BF,$9C,$7F
	.dc.b	$03,$7F,$A8,$F8,$6F,$E1,$BF,$0B,$27,$F1,$91,$FC,$43,$09,$1C,$CF
	.dc.b	$DE,$7C,$37,$C1,$79,$E1,$FA,$CA,$E5,$1F,$C0,$6F,$F8,$CC,$72,$8F
	.dc.b	$E0,$97,$70,$FD,$CE,$1C,$E7,$E0,$47,$95,$FC,$F5,$F4,$8C,$47,$F0
	.dc.b	$DF,$85,$FE,$4B,$F0,$B2,$3F,$49,$1C,$E7,$EF,$C7,$F1,$3F,$85,$FE
	.dc.b	$20,$A3,$FC,$C5,$F0,$3F,$E5,$DE,$93,$7A,$97,$FC,$CB,$0E,$D8,$7F
	.dc.b	$1D,$C7,$0F,$CE,$3F,$76,$7A,$B6,$FE,$1F,$5E,$0C,$BB,$FC,$65,$9F
	.dc.b	$F5,$9E,$5D,$EB,$9D,$F5,$3F,$7D,$86,$30,$FC,$29,$2F,$EE,$4F,$73
	.dc.b	$F8,$9E,$5E,$87,$F8,$5E,$1E,$D3,$B3,$8F,$84,$F1,$3F,$C6,$A3,$FC
	.dc.b	$EC,$E4,$E3,$F0,$9A,$D7,$E2,$CD,$1F,$09,$4C,$39,$F8,$39,$7E,$27
	.dc.b	$87,$C2,$71,$C3,$96,$32,$BB,$9E,$8B,$E8,$FA,$B0,$EA,$F3,$21,$F6
	.dc.b	$F3,$F9,$78,$F6,$8E,$66,$4F,$75,$F7,$71,$97,$E0,$31,$EB,$18,$C7
	.dc.b	$DA,$F8,$4F,$21,$E0,$D3,$F1,$38,$7B,$A3,$86,$7F,$17,$E0,$FE,$37
	.dc.b	$C2,$49,$FA,$71,$DA,$FF,$23,$AF,$BA,$7C,$18,$FC,$DF,$84,$F9,$DF
	.dc.b	$09,$27,$ED,$C7,$57,$EE,$E3,$D5,$F4,$63,$D7,$7F,$C6,$E3,$D7,$2F
	.dc.b	$CF,$E5,$D6,$FF,$0B,$97,$57,$E0,$65,$D5,$F8,$59,$75,$BF,$D3,$E5
	.dc.b	$D6,$7F,$11,$8F,$5C,$BF,$37,$97,$59,$F9,$59,$75,$C3,$F5,$F9,$7E
	.dc.b	$27,$E0,$E5,$7E,$59,$FC,$67,$C2,$75,$E0,$FD,$9E,$7D,$6B,$F2,$D9
	.dc.b	$7E,$47,$E1,$3F,$25,$F0,$92,$7F,$3F,$1D,$1F,$B5,$8D,$CF,$F7,$F9
	.dc.b	$75,$C3,$EE,$D7,$DF,$C7,$18,$73,$8F,$C1,$63,$DD,$F3,$10,$EA,$7E
	.dc.b	$BA,$FA,$57,$9F,$2F,$D3,$CF,$90,$FD,$9D,$75,$C7,$F8,$8E,$1F,$AB
	.dc.b	$F8,$3F,$D6,$7C,$24,$BF,$DC,$C7,$47,$F2,$47,$E1,$8F,$DE,$FC,$1F
	.dc.b	$EC,$C8,$FF,$00,$F8,$4E,$8F,$F0,$2A,$E8,$FE,$76,$BA,$3F,$04,$FC
	.dc.b	$E9,$FA,$1A,$E6,$D3,$A1,$E4,$3A,$10,$E8,$72,$C3,$F3,$AF,$A1,$FC
	.dc.b	$8B,$E8,$21,$D4,$FD,$85,$FF,$0D,$F8,$3F,$E1,$DF,$09,$27,$F8,$DC
	.dc.b	$7A,$2B,$AC,$7C,$33,$F7,$7F,$96,$7F,$B8,$9F,$F1,$AB,$75,$34,$FF
	.dc.b	$72,$C7,$F2,$1A,$EC,$FF,$01,$BF,$13,$EB,$E4,$FD,$5B,$EB,$7F,$3E
	.dc.b	$FF,$8E,$2B,$A3,$FC,$9D,$ED,$F1,$5C,$A3,$A1,$FC,$D2,$B1,$8F,$C4
	.dc.b	$8F,$A4,$47,$02,$FF,$CD,$18,$FA,$A7,$B1,$0F,$15,$F8,$7E,$1A,$47
	.dc.b	$95,$F2,$7C,$37,$C9,$BF,$5C,$FC,$9C,$3B,$3F,$23,$E1,$3D,$5F,$0B
	.dc.b	$23,$DD,$1C,$9E,$84,$3E,$8D,$F5,$7F,$82,$9F,$24,$E8,$7C,$79,$E8
	.dc.b	$FD,$89,$E8,$FD,$AC,$3D,$2F,$43,$96,$5F,$C1,$1F,$A4,$7F,$04,$3C
	.dc.b	$A7,$52,$1D,$4F,$AD,$E7,$7D,$5F,$46,$46,$0E,$BD,$1F,$C1,$33,$FC
	.dc.b	$69,$F4,$B4,$FA,$3B,$BE,$1B,$EF,$7C,$34,$C3,$F0,$51,$F8,$1C,$BA
	.dc.b	$C7,$E5,$32,$C3,$DB,$A7,$F8,$36,$7F,$E1,$08,$E2,$7E,$2F,$0E,$8F
	.dc.b	$E7,$30,$E2,$F3,$3F,$21,$78,$CF,$CD,$1F,$40,$FC,$56,$7F,$96,$79
	.dc.b	$1D,$48,$60,$FD,$4F,$C1,$FE,$67,$E1,$24,$7E,$76,$3E,$E6,$DF,$46
	.dc.b	$3F,$11,$5F,$13,$E1,$7F,$41,$F0,$D2,$7F,$49,$1F,$A3,$BE,$71,$FB
	.dc.b	$6C,$30,$FC,$33,$ED,$E1,$FC,$A1,$FD,$DD,$73,$7F,$7B,$7F,$AD,$8E
	.dc.b	$31,$CC,$FD,$95,$63,$1F,$87,$7E,$77,$ED,$E3,$F0,$DF,$CF,$C7,$F0
	.dc.b	$4B,$ED,$18,$E0,$63,$DF,$1F,$DF,$7C,$27,$EF,$48,$FF,$01,$F8,$5F
	.dc.b	$8B,$C1,$F8,$9C,$FB,$1F,$88,$F8,$4F,$E0,$9F,$0D,$2F,$F8,$44,$72
	.dc.b	$7F,$8C,$E1,$73,$F1,$63,$D3,$8F,$FB,$26,$8C,$39,$61,$FC,$83,$1F
	.dc.b	$E2,$11,$FC,$46,$F9,$65,$F9,$8C,$B1,$7F,$3C,$FC,$29,$FA,$93,$F9
	.dc.b	$4D,$74,$CB,$E1,$61,$D2,$7F,$8F,$7C,$37,$F1,$D2,$3F,$C8,$7E,$17
	.dc.b	$F9,$07,$C3,$48,$FE,$48,$FF,$23,$F8,$6F,$F2,$57,$E5,$9F,$C3,$A7
	.dc.b	$FC,$62,$DC,$7E,$9B,$74,$FE,$3E,$FF,$BD,$DB,$F0,$BC,$BE,$16,$5F
	.dc.b	$FB,$A2,$93,$DB,$E1,$E4,$E5,$F0,$FC,$27,$8B,$E1,$A4,$F1,$7C,$BF
	.dc.b	$0B,$E6,$F8,$5F,$3F,$C2,$CA,$ED,$1E,$AF,$85,$91,$E1,$7D,$BF,$0B
	.dc.b	$EE,$F8,$59,$3E,$57,$E0,$FC,$1F,$C2,$F8,$59,$3E,$97,$E2,$FC,$2F
	.dc.b	$C6,$F8,$49,$3E,$D7,$E4,$FC,$2F,$CA,$F8,$59,$3F,$05,$F9,$BF,$07
	.dc.b	$F3,$BE,$12,$4F,$C5,$7F,$CF,$5F,$2B,$FE,$9F,$3F,$CF,$F1,$FE,$62
	.dc.b	$6F,$4F,$E9,$1B,$D3,$FD,$41,$7C,$1F,$2F,$16,$58,$E4,$F6,$F1,$7B
	.dc.b	$FC,$1C,$9D,$63,$D6,$3C,$7F,$07,$E4,$F8,$2F,$2F,$C1,$CB,$DE,$9E
	.dc.b	$19,$F4,$FC,$14,$9C,$8F,$35,$B2,$7D,$FF,$07,$F0,$3E,$0E,$5F,$0B
	.dc.b	$DC,$7C,$4F,$83,$F8,$BF,$05,$F1,$BE,$0E,$5E,$F2,$4F,$CA,$F8,$29
	.dc.b	$39,$7E,$6F,$C1,$F6,$8F,$99,$8D,$E2,$F9,$99,$76,$9E,$99,$76,$7E
	.dc.b	$96,$3D,$1F,$A9,$8D,$BF,$A9,$8F,$3D,$FE,$A6,$56,$FE,$16,$6C,$FB
	.dc.b	$47,$DD,$CF,$B6,$3F,$77,$3E,$98,$FD,$DC,$ED,$F2,$CF,$F0,$FF,$07
	.dc.b	$F8,$8F,$83,$93,$F0,$E3,$AE,$3F,$87,$CB,$AE,$3F,$87,$CB,$F1,$FF
	.dc.b	$07,$30,$F9,$4F,$E4,$BE,$0F,$AC,$7C,$DC,$BA,$BF,$93,$CB,$AE,$1F
	.dc.b	$91,$C7,$AC,$FB,$B1,$E8,$FE,$67,$1B,$97,$E6,$72,$EB,$3F,$9D,$CB
	.dc.b	$AC,$FE,$77,$2E,$B3,$F2,$31,$EB,$C1,$F8,$AC,$B9,$E7,$FA,$1C,$ED
	.dc.b	$FC,$EC,$FF,$DC,$9E,$6F,$F9,$34,$91,$8F,$83,$EB,$8F,$E5,$32,$EB
	.dc.b	$C3,$F8,$AA,$FC,$95,$BE,$03,$99,$23,$3F,$07,$D9,$87,$03,$B1,$F9
	.dc.b	$5D,$7A,$E1,$E0,$BE,$93,$DA,$FA,$3F,$EE,$08,$5F,$28,$C1,$D8,$D6
	.dc.b	$5E,$AC,$58,$7A,$B1,$E7,$BF,$F6,$AC,$7B,$E1,$D7,$2F,$8B,$97,$C6
	.dc.b	$F8,$29,$3F,$01,$F9,$3F,$07,$D7,$2E,$93,$F1,$3E,$0F,$E6,$91,$F8
	.dc.b	$FF,$03,$C9,$F5,$5D,$CF,$C5,$87,$39,$FD,$69,$F1,$4E,$84,$3E,$15
	.dc.b	$FD,$AB,$E5,$C5,$FE,$6E,$D1,$8E,$2F,$F8,$74,$30,$ED,$1F,$AD,$3B
	.dc.b	$1D,$0E,$D8,$FC,$5C,$7B,$47,$E0,$3E,$0B,$F1,$3F,$07,$23,$07,$2A
	.dc.b	$FD,$63,$EA,$3F,$82,$9F,$88,$F8,$3F,$C8,$FC,$1C,$8C,$1C,$8F,$BD
	.dc.b	$5D,$5F,$C0,$5F,$A2,$FA,$C7,$43,$F0,$97,$CD,$FC,$35,$FE,$7A,$3F
	.dc.b	$3F,$5C,$EB,$F0,$17,$8B,$F4,$69,$FB,$77,$D1,$3F,$C3,$DF,$29,$D0
	.dc.b	$FD,$0E,$1D,$5F,$D1,$FC,$1F,$EB,$08,$FE,$9B,$E0,$FD,$17,$D6,$3D
	.dc.b	$17,$F9,$1F,$81,$FD,$9F,$C1,$49,$FD,$6C,$72,$7F,$41,$85,$AF,$CC
	.dc.b	$61,$D2,$7F,$81,$1E,$A3,$EF,$D7,$37,$F5,$17,$FC,$02,$39,$61,$F8
	.dc.b	$7C,$2D,$7E,$1A,$FA,$74,$7F,$01,$8E,$8B,$81,$87,$5C,$BF,$55,$97
	.dc.b	$5B,$FE,$0B,$F0,$5F,$C3,$88,$FF,$08,$F8,$2E,$BD,$3F,$B4,$8E,$B8
	.dc.b	$BE,$0B,$F8,$AF,$C1,$49,$FE,$1F,$1C,$9F,$E0,$58,$5C,$BF,$75,$8F
	.dc.b	$4E,$8F,$E1,$57,$C6,$FF,$84,$61,$FC,$82,$38,$70,$7E,$9F,$1B,$5F
	.dc.b	$A6,$C3,$93,$E4,$AE,$58,$FF,$07,$C3,$94,$FF,$22,$F8,$2F,$E5,$7F
	.dc.b	$05,$23,$07,$C8,$D7,$DC,$FE,$BC,$FC,$0E,$7F,$EE,$4F,$CA,$BF,$C9
	.dc.b	$11,$FE,$03,$7F,$C9,$0D,$F8,$9F,$F0,$B2,$CB,$FD,$03,$3C,$38,$BE
	.dc.b	$A6,$5E,$1A,$FE,$4D,$97,$F3,$E8,$F5,$C7,$80,$ED,$8F,$4C,$39,$E3
	.dc.b	$63,$C8,$D9,$D9,$EC,$3A,$9F,$D0,$AF,$B3,$EF,$DF,$F9,$1B,$E7,$AF
	.dc.b	$71,$CF,$16,$5E,$E6,$E1,$98,$F6,$1F,$E6,$4A,$FF,$10,$35,$E8,$7C
	.dc.b	$7F,$83,$E1,$C3,$EE,$CF,$F7,$13,$AC,$3B,$FC,$1F,$5F,$83,$93,$D2
	.dc.b	$3C,$18,$48,$EE,$F4,$7D,$87,$F0,$4C,$FF,$C0,$CF,$13,$CB,$5D,$1F
	.dc.b	$06,$4B,$E9,$18,$BB,$8F,$DA,$CB,$D1,$8F,$4A,$E3,$7C,$EB,$EF,$E3
	.dc.b	$F1,$63,$E3,$7C,$1C,$AF,$8A,$FC,$9F,$83,$F9,$5F,$05,$F2,$FE,$0E
	.dc.b	$57,$CA,$8F,$9D,$F0,$72,$3E,$6B,$F4,$7E,$0F,$A4,$7A,$B0,$B9,$FE
	.dc.b	$43,$1E,$B3,$F5,$31,$EA,$FC,$0C,$7A,$3F,$0B,$1B,$7F,$98,$C7,$9E
	.dc.b	$9F,$A1,$CA,$DF,$E8,$72,$EA,$FC,$AC,$BA,$E1,$FA,$0C,$B9,$E1,$FA
	.dc.b	$9C,$AD,$FE,$A7,$2F,$C5,$7C,$1F,$E2,$FE,$0E,$4F,$E2,$A3,$AE,$3F
	.dc.b	$AA,$CB,$AE,$3F,$B3,$CB,$F2,$5F,$07,$30,$FC,$8B,$F9,$5F,$83,$EB
	.dc.b	$1F,$94,$CB,$AB,$F8,$0C,$BA,$E1,$FB,$DC,$7A,$CF,$EF,$71,$E8,$FE
	.dc.b	$1B,$1B,$97,$EF,$32,$EB,$3F,$8A,$CB,$AC,$FF,$00,$CB,$AC,$FF,$07
	.dc.b	$C7,$AF,$27,$EB,$32,$E7,$9F,$F0,$DC,$ED,$FF,$0D,$CF,$F5,$DF,$07
	.dc.b	$FA,$FF,$83,$93,$FA,$E8,$EB,$8F,$ED,$B2,$EB,$CC,$CF,$F6,$DF,$07
	.dc.b	$30,$FD,$AB,$FB,$9F,$83,$EB,$1F,$B8,$CB,$AB,$FA,$4C,$BA,$E0,$C7
	.dc.b	$AE,$3C,$EB,$F5,$38,$DD,$FF,$8F,$E5,$D2,$7F,$5B,$97,$49,$FE,$47
	.dc.b	$97,$4C,$79,$D7,$90,$87,$43,$F8,$5D,$FF,$0F,$F8,$39,$1F,$C3,$9F
	.dc.b	$E2,$DF,$95,$3E,$27,$A4,$FE,$7C,$FD,$51,$FE,$B3,$8F,$F6,$7B,$FE
	.dc.b	$BB,$7F,$D9,$EE,$7E,$17,$9F,$C4,$76,$8F,$E0,$B7,$73,$71,$FF,$B4
	.dc.b	$F8,$58,$F8,$6F,$B9,$FA,$33,$C4,$7E,$BC,$F1,$61,$E8,$C6,$46,$0F
	.dc.b	$61,$0F,$09,$E9,$CB,$DA,$F5,$3A,$1F,$CA,$6B,$DB,$7E,$4A,$EB,$39
	.dc.b	$BF,$8F,$F0,$DF,$23,$E1,$A4,$FC,$58,$EB,$9F,$EB,$E3,$D6,$FC,$07
	.dc.b	$E4,$7C,$37,$D0,$F8,$69,$1F,$36,$3A,$BF,$3E,$FA,$BE,$5B,$EB,$AF
	.dc.b	$EB,$E3,$AB,$F2,$F4,$5F,$49,$F6,$61,$F7,$A3,$AC,$7B,$F0,$EB,$8F
	.dc.b	$97,$4F,$A7,$97,$6A,$FD,$81,$D8,$F4,$9F,$83,$C3,$B5,$71,$C7,$C3
	.dc.b	$3F,$87,$F8,$5F,$C6,$FC,$34,$8C,$1C,$F1,$FD,$83,$F7,$AF,$F9,$03
	.dc.b	$F8,$FF,$86,$FC,$AF,$C3,$48,$C1,$C8,$FB,$17,$D5,$FB,$58,$7A,$70
	.dc.b	$EB,$1F,$7B,$0F,$36,$1C,$E3,$F0,$18,$7E,$8A,$3F,$47,$7C,$EB,$D9
	.dc.b	$86,$2F,$D1,$D3,$FB,$B7,$D1,$BF,$F2,$B7,$CA,$74,$3F,$51,$8F,$57
	.dc.b	$F5,$5F,$0D,$FB,$02,$3F,$AD,$F8,$6F,$45,$F5,$8F,$45,$FD,$DF,$7F
	.dc.b	$F6,$FE,$54,$9F,$DA,$47,$27,$F3,$58,$5C,$FE,$A6,$3D,$3A,$7F,$8F
	.dc.b	$F0,$7F,$B7,$6F,$F4,$18,$F3,$7F,$45,$8F,$1C,$BF,$19,$95,$9F,$E0
	.dc.b	$18,$F3,$7E,$D6,$1C,$F2,$FC,$16,$3C,$DF,$C2,$63,$FC,$3B,$E1,$A5
	.dc.b	$FF,$0B,$7F,$88,$FC,$37,$38,$F7,$9F,$C3,$7E,$17,$F8,$AF,$C3,$48
	.dc.b	$FE,$25,$1C,$1E,$44,$3F,$75,$3C,$F0,$FC,$5C,$7E,$F6,$79,$3F,$BF
	.dc.b	$AF,$E4,$31,$CE,$3F,$81,$5F,$99,$E6,$7E,$03,$6F,$E5,$EF,$33,$F9
	.dc.b	$25,$74,$BF,$D1,$5F,$F2,$CF,$86,$FE,$5B,$F0,$D2,$FF,$95,$47,$97
	.dc.b	$83,$F1,$11,$FE,$39,$FC,$AB,$FC,$C2,$DF,$87,$FF,$25,$41,$B1,$D4
	.dc.b	$FE,$69,$7D,$5F,$E3,$57,$D5,$FE,$E4,$0E,$C7,$49,$8E,$71,$FE,$05
	.dc.b	$7E,$27,$F7,$95,$D5,$FD,$F5,$F8,$6B,$EB,$F4,$7E,$BA,$7F,$90,$46
	.dc.b	$2F,$96,$3F,$CF,$DF,$54,$77,$8C,$63,$EC,$8C,$0B,$53,$E1,$FE,$19
	.dc.b	$27,$73,$AF,$87,$F8,$FF,$0D,$27,$36,$8F,$97,$F0,$DF,$33,$E1,$BE
	.dc.b	$6F,$C3,$CB,$DC,$7D,$0F,$86,$91,$C9,$FA,$7F,$0D,$F5,$3E,$1E,$4F
	.dc.b	$67,$EB,$FC,$3F,$D8,$F8,$7F,$B3,$F0,$F2,$BC,$71,$F7,$3E,$1F,$EE
	.dc.b	$FC,$3C,$9F,$3C,$7E,$03,$E1,$E4,$7A,$9F,$C1,$FC,$3F,$E1,$3E,$1F
	.dc.b	$F0,$BF,$0F,$2B,$DF,$1F,$88,$F8,$6F,$C4,$FC,$34,$9F,$87,$1F,$E9
	.dc.b	$5E,$4F,$F5,$61,$FE,$95,$27,$56,$7E,$17,$5F,$0B,$2F,$E7,$D7,$FA
	.dc.b	$57,$B7,$FA,$43,$FE,$81,$8C,$FC,$27,$53,$B7,$C2,$F7,$F8,$5F,$07
	.dc.b	$C2,$CA,$E7,$1E,$3F,$7F,$C9,$E3,$C9,$EF,$1E,$7F,$85,$91,$E2,$7D
	.dc.b	$5F,$0B,$EB,$F8,$49,$3E,$67,$DD,$F0,$BE,$FF,$85,$F8,$1F,$0B,$2B
	.dc.b	$D7,$1F,$0F,$E0,$FE,$27,$C2,$49,$F7,$C7,$C7,$F8,$59,$1F,$09,$F9
	.dc.b	$5F,$0B,$F2,$FE,$17,$FD,$81,$F9,$57,$F8,$6E,$BF,$D6,$8D,$E6,$F8
	.dc.b	$F3,$AF,$84,$DF,$C2,$CB,$FC,$DD,$FF,$84,$87,$C1,$F4,$24,$6B,$E1
	.dc.b	$3B,$FC,$2F,$83,$E1,$64,$F2,$8F,$1F,$C2,$48,$EA,$F9,$BE,$13,$CF
	.dc.b	$F0,$BE,$8F,$85,$95,$E2,$8F,$5F,$C2,$7B,$3E,$12,$4F,$9A,$3D,$FF
	.dc.b	$05,$23,$D2,$FC,$2F,$83,$F8,$7F,$05,$F1,$3E,$0E,$57,$BA,$3E,$3F
	.dc.b	$C1,$7C,$8F,$83,$93,$F0,$A3,$E5,$FC,$1C,$8F,$8A,$FC,$EF,$83,$FF
	.dc.b	$77,$7E,$55,$FC,$3A,$7F,$84,$3F,$C3,$9F,$F7,$23,$7A,$BF,$C3,$CE
	.dc.b	$FE,$08,$FF,$0D,$9F,$E1,$6B,$E0,$BA,$9D,$BE,$0F,$BF,$9D,$E0,$F8
	.dc.b	$29,$7C,$23,$C7,$F0,$32,$39,$BE,$6F,$82,$F3,$FC,$17,$A3,$E0,$A5
	.dc.b	$78,$23,$D7,$F0,$5E,$CF,$82,$93,$E4,$8F,$7F,$C0,$C8,$F3,$BF,$0B
	.dc.b	$E0,$BE,$1F,$C0,$FC,$4F,$82,$95,$EC,$8F,$8F,$F0,$52,$3D,$EF,$CA
	.dc.b	$F8,$2F,$97,$F0,$32,$7E,$1B,$F3,$BE,$0B,$E7,$FC,$0C,$9F,$8E,$FD
	.dc.b	$2F,$82,$FE,$9D,$0E,$9F,$F2,$67,$E5,$5F,$E6,$18,$C7,$C1,$FF,$A0
	.dc.b	$63,$13,$A8,$E2,$F0,$F8,$9B,$1B,$F8,$4B,$7F,$D1,$6D,$F0,$98,$EE
	.dc.b	$46,$27,$FA,$2B,$FC,$5E,$FC,$1F,$09,$B8,$FE,$31,$3E,$43,$A4,$67
	.dc.b	$E1,$35,$F0,$7B,$F8,$49,$5D,$23,$97,$C1,$F3,$F8,$49,$3E,$08,$ED
	.dc.b	$F0,$92,$3C,$6F,$87,$E1,$3C,$5F,$09,$E3,$F8,$49,$5E,$88,$F3,$7C
	.dc.b	$1F,$9F,$E1,$24,$FB,$23,$D5,$F0,$72,$3D,$EF,$B7,$E1,$3D,$DF,$09
	.dc.b	$27,$E1,$BF,$07,$E1,$3E,$17,$C2,$49,$F8,$EF,$CB,$F8,$0F,$E3,$1C
	.dc.b	$DF,$EB,$77,$FD,$4A,$4F,$84,$D7,$C2,$6F,$E0,$E6,$1F,$DD,$2B,$F8
	.dc.b	$CB,$5E,$9F,$05,$23,$4F,$7F,$83,$F0,$7C,$1C,$9E,$4F,$8F,$E0,$FC
	.dc.b	$9F,$05,$27,$B3,$E7,$F8,$3F,$47,$C1,$49,$F1,$3E,$BF,$83,$F6,$7C
	.dc.b	$1F,$B7,$E0,$E5,$79,$E3,$A3,$F8,$3F,$11,$B3,$2C,$F9,$1F,$EA,$78
	.dc.b	$67,$DE,$F8,$43,$3C,$56,$F3,$F0,$D7,$C5,$CF,$C2,$F0,$CF,$E8,$7C
	.dc.b	$17,$D1,$F8,$39,$3F,$2E,$3E,$A7,$C1,$48,$F9,$CF,$D7,$F8,$3F,$B1
	.dc.b	$F0,$72,$7E,$93,$F6,$FE,$0F,$AC,$7D,$5C,$3A,$BE,$8C,$3A,$3E,$AC
	.dc.b	$2E,$9E,$FC,$7A,$4F,$BB,$1B,$3F,$0F,$1E,$B3,$F0,$B1,$EB,$3F,$1F
	.dc.b	$1E,$7D,$3F,$8B,$CA,$DF,$F8,$FF,$76,$7D,$2B,$F2,$19,$DA,$D6,$7F
	.dc.b	$BF,$FC,$AB,$F9,$D1,$23,$1F,$0F,$D3,$93,$EC,$E7,$73,$EB,$5F,$10
	.dc.b	$87,$AC,$FC,$C6,$1C,$9F,$F2,$6E,$1D,$5B,$87,$3B,$E9,$9F,$97,$1F
	.dc.b	$C3,$E5,$72,$65,$CF,$1F,$C6,$65,$CD,$B9,$73,$7E,$D9,$FA,$9C,$79
	.dc.b	$61,$CE,$FE,$07,$C1,$CA,$F2,$BF,$0F,$E1,$39,$47,$D9,$39,$1E,$ED
	.dc.b	$FF,$31,$7C,$E3,$89,$0F,$2D,$73,$7C,$67,$32,$4F,$E0,$DE,$9B,$7D
	.dc.b	$66,$9F,$58,$E9,$08,$FF,$49,$6A,$D3,$FB,$85,$78,$0F,$9F,$8F,$AA
	.dc.b	$FF,$5D,$B7,$E7,$67,$AD,$FC,$E6,$ED,$F3,$9F,$AF,$F0,$7F,$85,$23
	.dc.b	$F6,$BE,$13,$A6,$7F,$91,$8F,$8A,$7E,$04,$98,$FC,$67,$F0,$67,$42
	.dc.b	$1F,$6A,$BF,$25,$5D,$23,$F3,$58,$F9,$DE,$77,$F1,$5A,$7C,$53,$91
	.dc.b	$D2,$FE,$33,$FE,$A1,$CF,$FD,$30,$FA,$4F,$A7,$1F,$11,$B3,$FC,$D7
	.dc.b	$7F,$C3,$C7,$C4,$DF,$F4,$DF,$09,$FA,$7F,$84,$93,$F9,$C8,$F4,$6D
	.dc.b	$FA,$68,$F8,$78,$FC,$17,$CF,$1F,$3D,$FE,$3B,$7D,$1F,$D2,$DF,$47
	.dc.b	$F3,$B7,$D3,$66,$9F,$0F,$1E,$84,$CF,$E2,$3F,$C2,$6F,$C8,$74,$21
	.dc.b	$D0,$F7,$9F,$38,$B9,$7F,$36,$CF,$E0,$C7,$A3,$0F,$C2,$3F,$B0,$F8
	.dc.b	$3F,$E1,$1F,$09,$23,$F7,$B1,$E7,$CB,$CD,$97,$90,$93,$FC,$57,$8F
	.dc.b	$F9,$53,$C0,$FE,$0B,$7C,$DE,$C7,$93,$0F,$82,$DC,$7C,$B9,$FC,$18
	.dc.b	$FC,$13,$F8,$0C,$FF,$49,$9F,$38,$C6,$EC,$7F,$90,$4F,$9E,$FE,$BE
	.dc.b	$7E,$6C,$3E,$B6,$1D,$5F,$EB,$3B,$FF,$27,$F8,$39,$7F,$C7,$1F,$E5
	.dc.b	$5F,$09,$CE,$3F,$AC,$E9,$E6,$C7,$A6,$BF,$2B,$2F,$1D,$F2,$BF,$E3
	.dc.b	$78,$72,$7E,$06,$FF,$C9,$1F,$DA,$EF,$FC,$3D,$FE,$43,$5C,$6F,$DB
	.dc.b	$7C,$67,$F8,$6E,$3F,$B5,$69,$FD,$17,$2F,$23,$F4,$CF,$DA,$9F,$CB
	.dc.b	$3E,$13,$FA,$21,$1F,$E7,$3F,$09,$ED,$C3,$F8,$C4,$7E,$17,$2F,$24
	.dc.b	$7C,$EA,$E1,$3C,$CE,$27,$B0,$FB,$47,$B0,$FD,$71,$FD,$19,$E3,$1C
	.dc.b	$E3,$89,$EA,$DB,$F5,$CD,$D7,$FA,$9B,$E3,$39,$9F,$D1,$BE,$0F,$FA
	.dc.b	$F1,$1F,$EA,$7F,$09,$EB,$D3,$F8,$04,$78,$CF,$D0,$70,$6D,$E4,$7F
	.dc.b	$5A,$AF,$ED,$15,$C6,$3E,$61,$C7,$1F,$C4,$61,$F7,$F0,$FE,$6F,$C6
	.dc.b	$C3,$A5,$7E,$51,$E7,$7F,$14,$F6,$6D,$F5,$B3,$FC,$9E,$1F,$E4,$9C
	.dc.b	$FF,$BC,$7C,$24,$BF,$ED,$AF,$F7,$BF,$84,$FE,$F9,$F0,$9F,$DF,$7E
	.dc.b	$12,$57,$F7,$28,$FF,$00,$7C,$24,$8F,$EE,$AF,$F8,$0F,$E1,$3A,$47
	.dc.b	$F6,$EC,$2E,$DF,$AA,$C7,$AC,$FF,$7C,$C7,$AB,$FD,$D3,$1E,$8F,$F7
	.dc.b	$6C,$6D,$FE,$CB,$1E,$79,$FC,$4C,$AD,$FF,$37,$CB,$AB,$FD,$FB,$2E
	.dc.b	$B8,$7E,$EB,$2E,$78,$7C,$CC,$AD,$FF,$3F,$CB,$FD,$35,$F9,$57,$FA
	.dc.b	$92,$BF,$B3,$AF,$85,$C9,$A8,$EF,$8F,$F0,$6C,$BB,$C7,$D8,$CB,$AE
	.dc.b	$DF,$DB,$F3,$BC,$2C,$F9,$68,$3F,$CC,$1A,$77,$71,$9F,$F1,$58,$C4
	.dc.b	$6B,$2F,$09,$F8,$6C,$FC,$31,$FD,$67,$3E,$F9,$78,$31,$EF,$3C,$4E
	.dc.b	$B7,$EE,$3E,$0F,$C2,$F4,$8F,$EC,$58,$74,$D3,$FC,$11,$8F,$49,$FC
	.dc.b	$09,$FE,$36,$C3,$E4,$FC,$2F,$CA,$F8,$49,$3D,$85,$B4,$FF,$63,$47
	.dc.b	$E0,$4F,$C2,$10,$C1,$F8,$33,$B9,$4F,$F1,$DD,$F8,$1E,$D3,$D6,$BC
	.dc.b	$31,$F3,$2F,$D9,$3F,$E2,$1C,$7D,$9B,$32,$F6,$5F,$5D,$BF,$20,$F4
	.dc.b	$CF,$CF,$1F,$4F,$2E,$AF,$D5,$CB,$AE,$0C,$7A,$E3,$D2,$BF,$A5,$ED
	.dc.b	$FA,$7C,$BA,$7C,$27,$E4,$3E,$0E,$46,$0F,$D8,$1F,$0E,$BF,$1C,$E0
	.dc.b	$F3,$70,$7E,$6D,$F8,$B1,$E8,$7F,$17,$5D,$1F,$C6,$D7,$E7,$A3,$CF
	.dc.b	$8F,$9A,$78,$DF,$37,$A9,$FA,$53,$C8,$72,$3E,$25,$73,$7C,$97,$E6
	.dc.b	$7F,$64,$62,$7A,$47,$C2,$8E,$E7,$53,$F4,$3F,$09,$FB,$2F,$83,$91
	.dc.b	$F4,$63,$CE,$FC,$3A,$FD,$73,$CB,$83,$F5,$CD,$D7,$F5,$CF,$13,$ED
	.dc.b	$5F,$47,$B9,$E5,$3F,$3E,$75,$3F,$54,$73,$D3,$F5,$4F,$33,$F6,$F1
	.dc.b	$CF,$1F,$DB,$63,$E6,$9F,$82,$7E,$6A,$3E,$03,$F3,$30,$F3,$BF,$9A
	.dc.b	$3F,$1D,$F0,$9F,$C4,$BE,$12,$46,$0F,$34,$FE,$65,$F3,$3F,$B4,$7C
	.dc.b	$C7,$C1,$E3,$FE,$37,$5C,$A3,$F8,$6D,$61,$E7,$9F,$E5,$DE,$6F,$E8
	.dc.b	$DE,$67,$9D,$F9,$6F,$F1,$69,$E9,$87,$F1,$5A,$E9,$3E,$FD,$3F,$8A
	.dc.b	$3D,$38,$3F,$45,$1F,$82,$3F,$AE,$E3,$FC,$C3,$E1,$25,$7E,$BD,$FE
	.dc.b	$67,$F0,$7F,$2F,$1F,$C6,$1F,$CE,$E7,$CA,$FC,$C3,$F4,$07,$F2,$79
	.dc.b	$E7,$1F,$CB,$1E,$6F,$C0,$E1,$FD,$03,$ED,$E0,$FB,$AF,$F3,$2B,$E3
	.dc.b	$7C,$E3,$F8,$14,$78,$F1,$F6,$C7,$BC,$FF,$44,$57,$B9,$FC,$E5,$7F
	.dc.b	$39,$F8,$4F,$E9,$A4,$8F,$E1,$9F,$07,$CD,$FC,$89,$CF,$83,$F0,$B5
	.dc.b	$ED,$C3,$E4,$4F,$47,$FA,$AD,$F3,$8F,$E9,$2F,$F5,$AB,$E9,$7F,$87
	.dc.b	$9E,$9C,$0C,$7C,$D3,$FE,$C3,$D7,$96,$5E,$7B,$FE,$41,$8F,$37,$E4
	.dc.b	$1F,$6C,$FA,$39,$79,$2F,$94,$7F,$27,$F8,$4F,$ED,$64,$7F,$95,$7C
	.dc.b	$27,$93,$3F,$C8,$47,$93,$3F,$1C,$7E,$3B,$2F,$EC,$17,$C6,$BF,$B2
	.dc.b	$BF,$DD,$AF,$96,$DF,$CC,$30,$E3,$A6,$F0,$E3,$87,$8C,$F0,$9E,$A3
	.dc.b	$F6,$AF,$E2,$DF,$C1,$1F,$A9,$3F,$A0,$7C,$27,$F8,$13,$E1,$24,$60
	.dc.b	$FE,$89,$F0,$9F,$E0,$82,$3F,$D1,$FE,$13,$FB,$E2,$BF,$F4,$77,$E5
	.dc.b	$4F,$8C,$CC,$61,$BC,$3F,$EF,$69,$D5,$7F,$5C,$5F,$0D,$A8,$FE,$D7
	.dc.b	$0F,$86,$C7,$1A,$31,$3C,$67,$30,$F8,$6B,$E5,$63,$E1,$A5,$F7,$8D
	.dc.b	$FC,$34,$8F,$13,$CB,$E1,$B9,$FC,$34,$9F,$33,$DB,$E1,$BB,$FC,$34
	.dc.b	$9F,$53,$E2,$F8,$6F,$1F,$C3,$49,$F7,$3E,$6F,$85,$F3,$FC,$37,$A3
	.dc.b	$E1,$65,$7C,$38,$F5,$FC,$37,$B3,$DD,$93,$F1,$E3,$DF,$EB,$48,$F9
	.dc.b	$4F,$C2,$F6,$BE,$1F,$AF,$27,$E7,$3F,$2B,$D2,$FB,$31,$F6,$7A,$BF
	.dc.b	$C6,$DD,$29,$7C,$36,$C9,$5F,$E1,$28,$7C,$37,$3F,$86,$93,$97,$B7
	.dc.b	$C3,$77,$F8,$6F,$07,$C2,$CA,$E5,$1E,$3F,$86,$91,$D5,$F3,$7C,$2F
	.dc.b	$9F,$E1,$A4,$F8,$5F,$57,$C2,$FA,$FE,$1A,$4F,$95,$F7,$7C,$2F,$BF
	.dc.b	$E1,$BE,$07,$C2,$CA,$F5,$47,$C3,$F8,$69,$1E,$D7,$E3,$7C,$37,$C7
	.dc.b	$F8,$69,$3F,$05,$F9,$5E,$E7,$CB,$F5,$64,$FC,$57,$E7,$7A,$BF,$3F
	.dc.b	$D5,$FC,$CF,$E5,$5F,$E7,$1A,$FF,$23,$B7,$AB,$E8,$4E,$BE,$1F,$7F
	.dc.b	$0F,$2F,$FC,$83,$7F,$E6,$6F,$53,$FD,$F8,$FF,$8E,$3B,$BE,$CB,$F6
	.dc.b	$67,$7D,$BE,$0F,$86,$F0,$FC,$37,$8B,$E1,$A5,$74,$8F,$2F,$C2,$C8
	.dc.b	$EE,$FA,$3E,$1B,$D3,$F0,$D2,$7C,$6F,$B3,$E1,$BD,$BF,$0D,$EE,$F8
	.dc.b	$69,$5E,$88,$E8,$FE,$3C,$F8,$38,$63,$7F,$5E,$3D,$67,$5C,$6D,$D9
	.dc.b	$E6,$FE,$0E,$7E,$0B,$F8,$B9,$F8,$1E,$36,$CF,$B7,$07,$C6,$D2,$D7
	.dc.b	$D2,$D3,$A4,$FC,$AC,$EC,$FD,$6C,$BB,$3D,$32,$ED,$9F,$D2,$CB,$9E
	.dc.b	$9E,$7C,$EE,$8D,$39,$CF,$AF,$3B,$2C,$FA,$E3,$F7,$32,$EB,$A7,$DC
	.dc.b	$CF,$9E,$7D,$33,$E3,$97,$2C,$78,$DF,$E0,$B0,$E3,$A7,$E3,$B1,$E2
	.dc.b	$FD,$C3,$F1,$78,$72,$7E,$D9,$F8,$EB,$E6,$FE,$5A,$B9,$E9,$F4,$EF
	.dc.b	$F3,$7F,$0B,$F9,$CF,$86,$97,$F9,$68,$FC,$FF,$C2,$FE,$83,$E1,$64
	.dc.b	$FE,$6A,$3F,$47,$F0,$BF,$A4,$F8,$59,$3F,$9E,$8E,$AF,$E8,$30,$EA
	.dc.b	$FD,$CC,$3A,$BF,$A4,$C3,$AB,$F8,$0C,$3A,$BF,$A8,$C3,$AB,$F8,$4C
	.dc.b	$3A,$BF,$86,$C3,$AF,$0F,$EB,$31,$EB,$3F,$8A,$C7,$AC,$FE,$C3,$1E
	.dc.b	$B3,$F8,$EC,$7A,$CF,$ED,$31,$EA,$FE,$DB,$1E,$BB,$FE,$DB,$2E,$B3
	.dc.b	$FB,$AC,$BA,$CF,$EE,$B2,$EB,$3F,$BE,$C7,$AC,$FE,$FB,$1F,$EC,$D9
	.dc.b	$7F,$63,$E0,$FE,$1B,$F0,$B2,$3F,$DA,$86,$4B,$F0,$3F,$AA,$7F,$85
	.dc.b	$FC,$2F,$F1,$43,$9C,$66,$3F,$6C,$71,$3F,$67,$F9,$57,$F7,$86,$FC
	.dc.b	$37,$0C,$24,$76,$6B,$C3,$E1,$AE,$AD,$3F,$C1,$38,$F7,$8C,$50,$F0
	.dc.b	$38,$E3,$65,$FE,$00,$9F,$2D,$71,$BE,$78,$F6,$7F,$AD,$D7,$F8,$0E
	.dc.b	$BF,$DD,$2B,$E0,$FF,$C1,$6D,$8C,$C6,$3E,$13,$3F,$09,$27,$8B,$C3
	.dc.b	$E1,$38,$FC,$24,$9E,$AF,$4F,$84,$EB,$F0,$9D,$BE,$12,$57,$8A,$3A
	.dc.b	$3F,$33,$0B,$C1,$EF,$D1,$97,$9B,$E1,$3C,$FF,$09,$30,$F5,$C7,$AB
	.dc.b	$E0,$E4,$7B,$9F,$6F,$C2,$74,$E1,$F8,$79,$DA,$72,$7C,$CC,$BE,$1F
	.dc.b	$C1,$FC,$4F,$84,$95,$F2,$23,$E3,$FC,$24,$8F,$96,$FC,$AF,$84,$F9
	.dc.b	$7F,$09,$FE,$36,$FC,$AB,$FC,$1B,$C5,$F6,$5B,$AF,$D4,$9D,$7C,$2C
	.dc.b	$C7,$F0,$17,$B8,$E5,$F0,$BC,$FE,$17,$A7,$C2,$CB,$DE,$3F,$E1,$8C
	.dc.b	$95,$E2,$9C,$9E,$E8,$F0,$B4,$D1,$FD,$39,$7C,$27,$A7,$E1,$64,$F2
	.dc.b	$E2,$56,$A3,$0F,$81,$B8,$F9,$5E,$98,$76,$D3,$E1,$47,$C5,$31,$97
	.dc.b	$A7,$2E,$F1,$E3,$CB,$E5,$47,$CB,$F8,$5F,$99,$F0,$B2,$BC,$31,$C9
	.dc.b	$F6,$1E,$43,$1A,$7A,$2B,$AE,$3F,$4F,$0E,$B5,$EA,$F8,$5F,$B0,$47
	.dc.b	$DB,$F0,$BF,$6F,$E1,$64,$7C,$07,$EF,$7C,$2F,$DF,$F8,$49,$3F,$11
	.dc.b	$FC,$17,$C2,$F3,$8F,$9D,$85,$CB,$F0,$98,$F4,$9F,$A5,$8D,$9F,$C4
	.dc.b	$63,$D2,$7E,$B6,$36,$7F,$19,$8F,$E3,$FE,$16,$57,$D4,$7F,$25,$F0
	.dc.b	$BF,$93,$F8,$5F,$CA,$7C,$2C,$AF,$B3,$1F,$97,$F8,$59,$1F,$71,$FC
	.dc.b	$D7,$C2,$F4,$8F,$C3,$61,$72,$FC,$E6,$3D,$67,$F0,$78,$F5,$7F,$19
	.dc.b	$8F,$47,$F1,$D8,$DB,$FD,$26,$3F,$A7,$F8,$59,$5F,$8C,$7F,$D3,$FC
	.dc.b	$5F,$E4,$53,$FC,$81,$CD,$FE,$76,$5F,$09,$A3,$7F,$0B,$C3,$E1,$38
	.dc.b	$FC,$2C,$C7,$F3,$57,$FE,$47,$7F,$49,$8F,$81,$FD,$13,$E0,$C7,$9C
	.dc.b	$70,$C7,$B3,$CB,$1E,$CF,$4C,$3B,$3F,$E3,$8C,$3B,$74,$7E,$B3,$E1
	.dc.b	$3D,$64,$9F,$1F,$C2,$FB,$BE,$12,$47,$99,$F8,$3F,$09,$FE,$C3,$F9
	.dc.b	$67,$3B,$C5,$C7,$ED,$BF,$BC,$93,$67,$58,$7C,$0C,$BC,$0F,$95,$F9
	.dc.b	$B9,$78,$23,$D1,$95,$E4,$FE,$A1,$AF,$F8,$E6,$7F,$85,$E5,$E2,$CF
	.dc.b	$C3,$10,$CE,$5E,$07,$E1,$7B,$1F,$6B,$E1,$A4,$7D,$28,$FB,$A7,$DE
	.dc.b	$F8,$59,$19,$21,$F2,$FE,$1B,$F0,$67,$07,$39,$F9,$AB,$11,$F4,$7E
	.dc.b	$16,$CE,$63,$7C,$DF,$89,$F8,$6C,$BF,$8B,$BB,$3F,$D1,$36,$7C,$2F
	.dc.b	$E4,$BE,$1A,$4F,$E2,$9F,$CA,$FC,$2F,$E5,$BE,$1A,$4F,$E3,$9F,$CC
	.dc.b	$FC,$37,$E6,$BE,$1A,$4F,$E4,$9F,$F7,$DF,$E5,$7F,$E0,$BD,$BF,$BF
	.dc.b	$2F,$86,$EB,$5F,$98,$C7,$71,$DA,$3F,$1D,$8D,$BF,$AF,$8F,$69,$FC
	.dc.b	$96,$36,$7A,$FC,$3F,$83,$E2,$BC,$3F,$15,$2F,$F9,$F6,$9F,$E0,$DB
	.dc.b	$FC,$0E,$5D,$78,$7F,$60,$FB,$2B,$D4,$43,$B9,$CB,$1F,$6F,$C4,$48
	.dc.b	$EA,$FC,$0F,$88,$F8,$3E,$FF,$C2,$F7,$E5,$78,$A3,$E2,$FC,$47,$C6
	.dc.b	$F8,$89,$3E,$68,$F9,$3F,$0F,$CE,$3D,$38,$61,$E4,$7B,$30,$EA,$FF
	.dc.b	$AB,$36,$63,$DA,$3E,$0E,$3D,$9D,$63,$CD,$E1,$87,$77,$96,$3E,$A7
	.dc.b	$B1,$0E,$D7,$D5,$F0,$5F,$5D,$BE,$FE,$1F,$82,$F8,$8F,$C1,$FC,$34
	.dc.b	$BF,$A7,$1F,$86,$F8,$79,$1F,$59,$FC,$4F,$C3,$FE,$2B,$E1,$FF,$17
	.dc.b	$F0,$D2,$BE,$DC,$75,$7E,$EE,$1D,$5F,$85,$87,$47,$E2,$61,$72,$FC
	.dc.b	$9E,$3D,$67,$F0,$B8,$F5,$7E,$56,$3F,$9A,$F8,$7F,$CD,$FC,$3C,$9F
	.dc.b	$C5,$C7,$E7,$BE,$1E,$47,$E3,$9F,$D0,$F9,$1F,$E5,$F8,$FF,$6F,$FE
	.dc.b	$57,$FE,$16,$F4,$0D,$27,$C4,$77,$9F,$CB,$E3,$DD,$FB,$98,$F6,$7E
	.dc.b	$F6,$37,$5F,$D6,$E5,$DD,$FF,$36,$63,$D6,$B1,$D7,$FE,$71,$CF,$C3
	.dc.b	$18,$3F,$C8,$F2,$CF,$13,$E6,$F8,$8B,$26,$98,$F8,$6C,$FC,$3C,$C3
	.dc.b	$FC,$59,$48,$E3,$E3,$F2,$F8,$5E,$7F,$0D,$2F,$71,$D1,$E3,$8D,$7E
	.dc.b	$4E,$3D,$5F,$97,$85,$D7,$DF,$8F,$59,$F0,$E3,$D5,$FA,$38,$FA,$BE
	.dc.b	$1E,$47,$99,$F6,$FC,$3F,$BB,$E1,$FD,$FF,$0F,$2B,$D7,$1F,$0B,$E1
	.dc.b	$FE,$1F,$C3,$C9,$F7,$C7,$47,$EF,$E1,$74,$FA,$78,$F5,$9F,$8B,$8F
	.dc.b	$57,$F0,$B8,$F5,$7E,$4E,$3D,$5F,$C4,$E3,$F4,$BD,$69,$1F,$39,$FA
	.dc.b	$BF,$0F,$F5,$BE,$1F,$EB,$FC,$34,$AF,$A7,$1F,$6B,$E1,$FE,$DF,$C3
	.dc.b	$C9,$FA,$F1,$D1,$FC,$BE,$17,$6F,$C5,$E3,$D6,$7E,$EE,$3D,$5F,$CE
	.dc.b	$E3,$D5,$FC,$0E,$3D,$5F,$D0,$E3,$F8,$AF,$87,$FC,$5F,$C3,$49,$FC
	.dc.b	$3C,$7E,$43,$F2,$BF,$F6,$3C,$33,$FF,$63,$47,$F8,$9A,$3F,$D3,$AD
	.dc.b	$EA,$3C,$27,$7F,$0F,$CC,$95,$FE,$7F,$9F,$F0,$CF,$5B,$0F,$14,$FE
	.dc.b	$DB,$1F,$33,$BE,$DF,$27,$C3,$C8,$C1,$CF,$1E,$AF,$9E,$7F,$5D,$8F
	.dc.b	$F8,$EA,$3C,$1F,$11,$FE,$FC,$F8,$AC,$46,$CF,$5F,$C4,$FC,$13,$58
	.dc.b	$48,$F4,$7C,$3F,$C5,$30,$FF,$BE,$AB,$9E,$DF,$EE,$E1,$F7,$EF,$85
	.dc.b	$FD,$F4,$61,$E2,$FC,$E9,$E3,$A7,$F1,$76,$DF,$FB,$DD,$BF,$68,$FC
	.dc.b	$5D,$D7,$CA,$F8,$69,$70,$FA,$BF,$15,$5F,$F7,$EC,$67,$2F,$2F,$C4
	.dc.b	$7D,$FF,$8A,$FC,$07,$C6,$4B,$FA,$D1,$F8,$3F,$87,$91,$F6,$5F,$C3
	.dc.b	$6F,$F8,$7B,$FC,$47,$C5,$4A,$FB,$D1,$D5,$FC,$06,$1D,$5F,$A1,$D5
	.dc.b	$F8,$8C,$7A,$C6,$39,$99,$63,$3F,$C8,$E7,$DE,$BF,$15,$9F,$77,$86
	.dc.b	$7F,$99,$F8,$AF,$CD,$7C,$54,$9F,$C8,$47,$E7,$7E,$2B,$F3,$DF,$15
	.dc.b	$27,$F2,$91,$FA,$1F,$8A,$91,$F9,$77,$F4,$9F,$0B,$D6,$3F,$35,$87
	.dc.b	$57,$D1,$87,$57,$F3,$D8,$75,$7D,$98,$74,$7D,$D8,$5D,$3F,$5B,$8F
	.dc.b	$EC,$7E,$27,$F6,$5F,$13,$2F,$F5,$11,$FD,$13,$E0,$FF,$C7,$9C,$FF
	.dc.b	$D2,$DB,$2E,$44,$61,$BF,$15,$C7,$E2,$A5,$7F,$BC,$6B,$FC,$69,$7C
	.dc.b	$FE,$2B,$C5,$3F,$B7,$CB,$C4,$FD,$5C,$BC,$5F,$15,$23,$93,$E5,$CB
	.dc.b	$AC,$7D,$EC,$7A,$BE,$56,$E7,$E6,$C7,$D9,$F1,$52,$BC,$8F,$BE,$BE
	.dc.b	$07,$C3,$C9,$F4,$3F,$0F,$E2,$BE,$27,$C3,$C9,$F6,$3F,$1F,$E2,$3A
	.dc.b	$47,$8F,$0B,$C4,$E0,$F9,$59,$75,$AF,$46,$56,$B5,$1E,$9C,$BB,$3E
	.dc.b	$DC,$AC,$EB,$4F,$A3,$9F,$D7,$F8,$99,$7F,$3D,$FB,$5F,$11,$F6,$FE
	.dc.b	$27,$EE,$7C,$4C,$AF,$A9,$1F,$7F,$E2,$64,$7D,$77,$F0,$5F,$11,$D2
	.dc.b	$3E,$76,$17,$3F,$16,$EC,$BB,$57,$DF,$CB,$B3,$F5,$72,$EA,$FD,$7C
	.dc.b	$AE,$1C,$B2,$FC,$87,$C4,$CA,$FC,$43,$F9,$3F,$88,$FC,$A7,$C4,$C9
	.dc.b	$FC,$63,$F9,$7F,$89,$FC,$C7,$C4,$FF,$87,$3F,$2B,$FF,$19,$1F,$E3
	.dc.b	$3D,$FF,$D7,$99,$FA,$32,$ED,$8F,$E2,$5E,$CF,$42,$1F,$8D,$C7,$B3
	.dc.b	$F9,$BC,$3B,$30,$C0,$3F,$88,$E1,$FE,$2F,$71,$7F,$66,$3F,$C5,$D1
	.dc.b	$8E,$6F,$F1,$C4,$62,$38,$FC,$57,$AF,$E2,$64,$6C,$FD,$94,$A3,$E0
	.dc.b	$7C,$4F,$C1,$F8,$99,$5B,$E7,$5A,$7E,$37,$C4,$7C,$7C,$7E,$47,$C5
	.dc.b	$4C,$38,$5F,$A5,$F5,$E5,$E2,$69,$AC,$BB,$BF,$03,$1B,$A7,$BF,$2F
	.dc.b	$A9,$F1,$52,$BA,$BF,$5F,$E2,$BE,$C7,$C5,$7D,$9F,$89,$95,$E2,$8F
	.dc.b	$B9,$F1,$52,$3C,$AF,$DF,$F8,$AE,$91,$F4,$30,$B9,$7C,$FC,$7A,$CF
	.dc.b	$B3,$1E,$AF,$D6,$C7,$A3,$F6,$31,$B7,$F5,$F1,$FC,$67,$B3,$F8,$DF
	.dc.b	$8A,$97,$F1,$A3,$2F,$C8,$F8,$AF,$C9,$18,$7F,$07,$1F,$95,$F8,$99
	.dc.b	$1C,$48,$68,$B8,$7E,$1F,$E2,$B9,$3F,$89,$AB,$CF,$F8,$8F,$89,$B7
	.dc.b	$F8,$49,$BD,$1F,$80,$F8,$AF,$D2,$7C,$4C,$AF,$BA,$FE,$9F,$E2,$7F
	.dc.b	$D0,$7F,$95,$FF,$8B,$72,$FF,$45,$3F,$E3,$12,$3A,$F8,$19,$7F,$8D
	.dc.b	$9F,$F0,$5E,$25,$F1,$3D,$E7,$F4,$78,$DD,$7F,$45,$97,$83,$2F,$0F
	.dc.b	$C5,$4B,$E9,$1D,$9F,$EF,$78,$F6,$26,$3E,$8F,$8A,$93,$E3,$7D,$7F
	.dc.b	$15,$EC,$F8,$6F,$6F,$C2,$CA,$F4,$47,$C0,$F8,$6F,$83,$F0,$B2,$7D
	.dc.b	$91,$D1,$F0,$E1,$76,$FE,$65,$9B,$2E,$D5,$F0,$F2,$EC,$FA,$32,$EC
	.dc.b	$FC,$7C,$BB,$3E,$CC,$BE,$8F,$C2,$C8,$F9,$8F,$D4,$F8,$4F,$AB,$F0
	.dc.b	$BF,$5B,$E1,$25,$7D,$18,$FB,$3F,$0B,$F6,$BE,$12,$4F,$D5,$8E,$8F
	.dc.b	$CA,$C2,$EF,$EA,$C7,$AC,$FD,$BC,$7A,$BF,$43,$1E,$AF,$DF,$C7,$AB
	.dc.b	$F5,$31,$FC,$4F,$C2,$FE,$2B,$E1,$64,$FE,$16,$3F,$1B,$F0,$BF,$8E
	.dc.b	$F8,$39,$3F,$89,$8F,$C8,$FC,$24,$8F,$C6,$3F,$ED,$6F,$CA,$FF,$C1
	.dc.b	$DC,$AC,$3A,$D7,$E1,$31,$F0,$BC,$CF,$CA,$63,$D9,$FC,$4E,$3D,$5F
	.dc.b	$C5,$E3,$77,$FE,$D9,$97,$F8,$6C,$7C,$27,$84,$95,$FC,$AE,$3F,$C1
	.dc.b	$A3,$E1,$3C,$C4,$8D,$7C,$27,$A7,$D8,$F5,$7C,$24,$9E,$51,$D1,$FF
	.dc.b	$2B,$DF,$81,$ED,$87,$C1,$F8,$6E,$91,$AC,$6E,$5E,$8C,$BB,$4F,$97
	.dc.b	$2E,$CF,$4C,$BE,$57,$C2,$FC,$BF,$85,$93,$EB,$8F,$9D,$F0,$B2,$3D
	.dc.b	$CF,$D1,$F8,$5F,$A5,$F0,$B2,$7E,$13,$F5,$7E,$13,$AC,$7C,$5C,$3A
	.dc.b	$BE,$CC,$3A,$3E,$EC,$2E,$9F,$43,$1E,$93,$F0,$B1,$B3,$F5,$31,$EB
	.dc.b	$3F,$1B,$1E,$B3,$F6,$31,$E7,$E2,$7F,$71,$F2,$1C,$4D,$33,$8F,$F9
	.dc.b	$66,$1A,$76,$BE,$11,$FE,$1B,$9F,$69,$F7,$F3,$FC,$97,$C2,$FE,$4F
	.dc.b	$E1,$64,$FE,$0E,$3F,$2D,$F0,$BF,$97,$F8,$39,$3F,$87,$8F,$CD,$7C
	.dc.b	$2F,$E6,$FE,$12,$4F,$E2,$E3,$AB,$F8,$DC,$3A,$BF,$84,$C3,$AB,$F9
	.dc.b	$1C,$3A,$BF,$88,$C3,$AB,$F9,$5C,$3A,$BF,$8C,$C3,$FC,$28,$FF,$B4
	.dc.b	$C7,$E4,$EE,$2F,$F7,$94,$68,$FF,$0D,$46,$78,$B1,$18,$9D,$BB,$9D
	.dc.b	$3A,$6B,$98,$DB,$E2,$8C,$B6,$32,$F5,$9E,$EF,$9A,$32,$D8,$CC,$F2
	.dc.b	$8F,$3F,$03,$EE,$B2,$D8,$F2,$35,$F4,$35,$F2,$35,$E7,$1C,$88,$F2
	.dc.b	$6B,$C9,$AB,$EE,$E0,$00

;--------------------------------------------------------------------------------
;	FLOG2.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
flog2_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FLOG2.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FLOG2.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	flog2_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	flog2.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FLOG2.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	flog2.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FLOG2.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
flog2_data_compressed::
	.dc.b	$00,$02,$A1,$14,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$FD
	.dc.b	$4F,$E5,$1F,$97,$BE,$4B,$CF,$E9,$7A,$3D,$2D,$4B,$E1,$2F,$C9,$E3
	.dc.b	$E4,$73,$F1,$7A,$F8,$8D,$FB,$9C,$3D,$3E,$3E,$9F,$2F,$4F,$9F,$A7
	.dc.b	$D3,$D3,$EB,$E9,$F6,$F4,$FB,$FA,$7E,$0F,$4F,$C3,$E9,$F8,$BD,$3F
	.dc.b	$97,$EC,$7C,$CF,$63,$E6,$FB,$1F,$D8,$FF,$14,$FE,$40,$7E,$63,$C5
	.dc.b	$3F,$98,$7F,$B9,$3F,$D9,$9B,$BF,$1D,$F9,$6F,$CF,$7E,$9B,$F5,$DF
	.dc.b	$B6,$FD,$F7,$F0,$6F,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$FE,$ED,$F8,$AB
	.dc.b	$E6,$2F,$C7,$FE,$13,$DC,$FE,$F5,$F3,$9F,$47,$8F,$FC,$0D,$7F,$E7
	.dc.b	$B8,$FA,$11,$FE,$09,$7F,$DB,$87,$FA,$50,$FF,$03,$1F,$CF,$CF,$80
	.dc.b	$7B,$D3,$EF,$6F,$1F,$83,$8F,$C3,$C7,$E2,$E3,$F1,$F1,$F9,$38,$FC
	.dc.b	$BC,$7E,$6E,$3F,$3F,$1F,$A3,$8F,$D3,$C7,$EA,$E3,$FF,$0A,$77,$7B
	.dc.b	$F7,$FF,$10,$3F,$ED,$CE,$1F,$F2,$6B,$FE,$97,$2F,$C6,$F0,$F8,$AE
	.dc.b	$3F,$05,$CB,$DE,$E7,$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83
	.dc.b	$DE,$F0,$FB,$DE,$2F,$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$F5,$38,$3C
	.dc.b	$FE,$F7,$A3,$DE,$FF,$16,$7C,$EF,$F8,$14,$7C,$5E,$3E,$2F,$3F,$17
	.dc.b	$AF,$8B,$DF,$C5,$F0,$F8,$BE,$3F,$11,$CB,$E1,$79,$FC,$2F,$4F,$85
	.dc.b	$EB,$F0,$BD,$BE,$17,$BF,$C2,$F8,$3E,$17,$C3,$F0,$BE,$2F,$83,$F1
	.dc.b	$FC,$1F,$93,$DB,$F2,$FB,$7E,$6F,$6F,$CF,$EC,$7F,$90,$FE,$7B,$F1
	.dc.b	$5E,$61,$FC,$4B,$7C,$BD,$7C,$16,$FE,$0B,$87,$C1,$71,$F8,$2E,$5F
	.dc.b	$05,$CF,$E0,$BF,$33,$E8,$F5,$F7,$BB,$7B,$DD,$FD,$EF,$07,$BD,$E1
	.dc.b	$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD,$EF,$79,$FD,$EF
	.dc.b	$47,$BD,$FE,$5B,$F9,$FF,$C4,$73,$BD,$FF,$E7,$2A,$1F,$03,$BF,$8A
	.dc.b	$E1,$F2,$BC,$7D,$5E,$5C,$BC,$FE,$0B,$A7,$C1,$75,$F8,$2E,$DF,$05
	.dc.b	$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE,$0B,$C9,$F0,$5E
	.dc.b	$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$7B,$AF,$EF,$B5,$FD,$FE,$BF,$C0
	.dc.b	$35,$FE,$03,$AF,$B7,$E0,$BD,$DF,$05,$FE,$79,$F7,$FF,$CA,$47,$E3
	.dc.b	$BA,$3F,$D0,$CB,$E0,$34,$EF,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1C
	.dc.b	$FE,$03,$A7,$C0,$7F,$10,$E8,$ED,$F0,$1D,$FE,$03,$C1,$F0,$1F,$C5
	.dc.b	$3A,$FF,$8A,$F5,$FF,$16,$EB,$FE,$2F,$D7,$FC,$63,$AF,$F8,$CF,$5F
	.dc.b	$F1,$AE,$6F,$E3,$7C,$DF,$E9,$1F,$C5,$BF,$CD,$47,$A7,$9C,$FA,$5F
	.dc.b	$F5,$03,$FE,$FC,$3F,$D6,$C7,$F8,$98,$FE,$BC,$7F,$5D,$78,$F0,$F5
	.dc.b	$E1,$ED,$C1,$DF,$83,$E4,$7C,$0F,$C9,$F8,$1F,$95,$F0,$3F,$2F,$E0
	.dc.b	$7E,$67,$C0,$FC,$DF,$81,$F9,$DF,$03,$F3,$FE,$07,$E8,$7C,$0F,$FA
	.dc.b	$A7,$E8,$7F,$4E,$7E,$98,$7C,$0E,$7E,$07,$5F,$03,$BF,$81,$E1,$F0
	.dc.b	$3C,$7E,$07,$97,$C0,$F3,$F8,$1E,$9F,$03,$D7,$E0,$7B,$7C,$0F,$7F
	.dc.b	$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03,$F2,$FC,$9F,$99,$D7,$F3,$7A
	.dc.b	$FE,$77,$5F,$CF,$EB,$FA,$1D,$7F,$47,$AF,$E9,$75,$FD,$3E,$BF,$A9
	.dc.b	$D7,$FE,$C5,$F4,$FF,$81,$9F,$C0,$8F,$6F,$83,$FE,$D3,$5D,$7B,$EB
	.dc.b	$E1,$D7,$C7,$AF,$97,$5F,$3E,$BE,$9D,$7F,$27,$E0,$7E,$57,$C0,$FC
	.dc.b	$BF,$81,$F0,$77,$7C,$DF,$81,$FF,$6B,$7D,$17,$EF,$0F,$DC,$F0,$7F
	.dc.b	$B8,$D7,$BF,$AF,$7F,$7E,$FF,$0F,$7F,$8F,$BF,$CB,$DF,$E7,$EF,$F4
	.dc.b	$E7,$EB,$EF,$FB,$FC,$8F,$81,$E4,$7C,$1F,$23,$E1,$79,$1F,$0F,$C8
	.dc.b	$F1,$FB,$FE,$4F,$67,$CB,$EC,$F9,$BD,$9F,$3F,$B3,$E8,$F6,$7D,$3E
	.dc.b	$CF,$AB,$D9,$F5,$FB,$3E,$CF,$67,$FD,$E5,$F4,$7F,$EC,$B3,$F4,$47
	.dc.b	$E8,$7B,$0B,$DF,$DF,$BF,$C3,$DF,$E3,$EF,$F2,$F7,$FE,$C7,$B1,$D3
	.dc.b	$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F
	.dc.b	$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$F0,$DE,$A7
	.dc.b	$AB,$DC,$74,$4F,$3F,$FA,$59,$FD,$28,$FC,$4A,$7F,$12,$EF,$DB,$E2
	.dc.b	$E7,$DB,$B3,$FE,$33,$3F,$C2,$87,$F8,$4C,$7B,$1A,$F5,$B7,$E6,$70
	.dc.b	$F3,$38,$F9,$9C,$BC,$CE,$7E,$67,$4E,$CE,$BB,$F6,$F6,$3B,$FB,$1E
	.dc.b	$0F,$63,$C7,$F4,$9F,$D3,$0F,$C8,$F0,$9F,$C8,$BF,$01,$F6,$37,$D7
	.dc.b	$E3,$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$F6,$1A,$F8
	.dc.b	$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F,$5F
	.dc.b	$CF,$EB,$FA,$3D,$7F,$4F,$AF,$F3,$3E,$97,$E6,$F2,$9F,$E6,$07,$E3
	.dc.b	$CF,$C7,$3F,$69,$7A,$DC,$7D,$6E,$5E,$B7,$F0,$5E,$BE,$9E,$B7,$5F
	.dc.b	$5B,$B7,$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$78,$FD,$6F,$27
	.dc.b	$AD,$E5,$F5,$BC,$DE,$B7,$9F,$D6,$F4,$7A,$DE,$9F,$5B,$D5,$EB,$7E
	.dc.b	$1B,$E9,$BF,$C2,$E7,$F8,$58,$FC,$69,$F8,$CE,$72,$F5,$F8,$7A,$FC
	.dc.b	$7D,$EF,$E3,$DE,$77,$3F,$7B,$A7,$BD,$F9,$2F,$C5,$FF,$C1,$67,$E0
	.dc.b	$0F,$BF,$91,$FB,$ED,$F4,$38,$7A,$1C,$7D,$0E,$5E,$87,$3F,$43,$A7
	.dc.b	$A1,$D7,$D0,$ED,$E7,$F7,$F3,$FC,$1E,$7F,$87,$CF,$F1,$79,$FE,$3F
	.dc.b	$3F,$C9,$E7,$FC,$1F,$6B,$E1,$7B,$5F,$0F,$DA,$FF,$57,$FA,$7F,$EB
	.dc.b	$0F,$33,$FD,$63,$D3,$FE,$B7,$FC,$C3,$FC,$D9,$F2,$47,$F9,$5B,$7E
	.dc.b	$03,$5E,$F6,$FA,$78,$74,$F1,$E9,$E5,$D3,$CF,$A7,$A7,$4F,$5E,$9E
	.dc.b	$DD,$3D,$FA,$7C,$1D,$3F,$5F,$DE,$FB,$1E,$F7,$8F,$AF,$ED,$7B,$DF
	.dc.b	$6F,$DE,$FB,$9E,$F7,$DD,$F7,$BE,$F7,$4F,$DF,$E9,$FC,$07,$4F,$F0
	.dc.b	$6F,$82,$FF,$1F,$9F,$B4,$3E,$A9,$F5,$07,$36,$F9,$B8,$73,$71,$E6
	.dc.b	$E5,$CD,$CF,$9B,$A7,$37,$5E,$6E,$DC,$DD,$F9,$BC,$1C,$9E,$1F,$03
	.dc.b	$C5,$E0,$78,$FC,$0F,$27,$BD,$E5,$F7,$BC,$DE,$F7,$F1,$CF,$A9,$FD
	.dc.b	$89,$FE,$1F,$3F,$9F,$1F,$B0,$F1,$4F,$EC,$1F,$E5,$6B,$DD,$E5,$EE
	.dc.b	$F3,$F7,$7A,$7B,$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF
	.dc.b	$17,$BB,$E3,$F7,$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE
	.dc.b	$FA,$7D,$DF,$E6,$BF,$55,$FE,$86,$3F,$D0,$A7,$F3,$C3,$F6,$07,$EB
	.dc.b	$F9,$7F,$A1,$AF,$73,$8F,$B9,$CB,$DC,$E7,$EE,$74,$F6,$FA,$FB,$7D
	.dc.b	$BD,$BE,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F
	.dc.b	$2F,$AF,$E6,$F5,$FC,$FE,$BF,$F4,$CF,$AB,$FF,$42,$9F,$CE,$8F,$DD
	.dc.b	$9F,$BA,$3F,$0F,$C9,$FD,$69,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7
	.dc.b	$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F
	.dc.b	$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5,$BF,$B1,$FC,$07,$F3,$83
	.dc.b	$FC,$CC,$7F,$81,$4F,$C4,$1F,$87,$E4,$FE,$6F,$EB,$70,$F5,$B8,$FA
	.dc.b	$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD,$DB,$D6,$EF,$EB,$78,$3D,$6F
	.dc.b	$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6,$F2,$FA,$DE,$6F,$5B,$CF,$EB
	.dc.b	$7F,$7A,$FA,$CF,$F6,$59,$FE,$CA,$3F,$C0,$27,$F3,$53,$F0,$DC,$87
	.dc.b	$F0,$CF,$F8,$29,$7A,$BC,$FD,$5E,$9E,$AF,$5F,$57,$B7,$AB,$DF,$D5
	.dc.b	$F0,$7A,$BE,$1F,$57,$C5,$EA,$F8,$FD,$5F,$27,$AB,$E5,$F5,$7C,$DE
	.dc.b	$AF,$9F,$D5,$F4,$7A,$BF,$E1,$2F,$AD,$FF,$65,$9F,$E0,$03,$FA,$09
	.dc.b	$FD,$00,$FC,$3F,$19,$FC,$3B,$FE,$22,$7F,$DA,$23,$D3,$E9,$E9,$F5
	.dc.b	$F4,$FB,$7A,$7D,$FD,$3F,$07,$A7,$E1,$F4,$FC,$5E,$9F,$8F,$D3,$F2
	.dc.b	$7A,$7E,$5F,$4F,$CD,$E9,$F9,$FD,$3F,$47,$A7,$FE,$2A,$FA,$EF,$EF
	.dc.b	$E7,$FA,$B4,$FF,$26,$1F,$C3,$CF,$C4,$1F,$C3,$78,$BF,$C7,$2B,$D2
	.dc.b	$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9,$78
	.dc.b	$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5E,$7F,$4B
	.dc.b	$D1,$E9,$7F,$90,$BD,$B7,$CC,$5F,$97,$C7,$C5,$E7,$DF,$D7,$C2,$EF
	.dc.b	$DE,$E1,$EC,$71,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1
	.dc.b	$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6
	.dc.b	$3C,$DE,$C7,$F9,$63,$EB,$DF,$2F,$7E,$47,$1F,$0B,$9F,$85,$D7,$66
	.dc.b	$FE,$0B,$87,$C1,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7
	.dc.b	$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$E5
	.dc.b	$F7,$BC,$DE,$F7,$F9,$BB,$EC,$3F,$C8,$67,$FB,$C4,$FF,$48,$1F,$D4
	.dc.b	$CF,$E1,$87,$E1,$7D,$F3,$F8,$57,$FC,$FE,$FF,$BC,$47,$B9,$D7,$DC
	.dc.b	$ED,$EE,$77,$F7,$3C,$1E,$E7,$87,$DC,$F1,$7B,$9E,$3F,$73,$C9,$EE
	.dc.b	$79,$7D,$CF,$37,$B9,$E7,$F7,$3D,$1E,$E7,$A7,$DC,$FF,$43,$FD,$8F
	.dc.b	$E1,$B8,$BF,$D2,$2B,$DC,$CF,$B9,$AF,$73,$7E,$E7,$0F,$73,$8F,$B9
	.dc.b	$CB,$DC,$E7,$EE,$74,$F7,$3A,$FB,$9D,$BD,$CE,$FE,$E7,$83,$DC,$F0
	.dc.b	$FB,$9E,$2F,$73,$C7,$EE,$79,$3D,$CF,$2F,$B9,$E6,$F7,$3C,$FE,$E7
	.dc.b	$FA,$7B,$8B,$F0,$FC,$FF,$EA,$87,$FD,$1A,$7F,$93,$8F,$F2,$69,$FC
	.dc.b	$38,$FE,$1A,$FF,$96,$FD,$5E,$5E,$AF,$3F,$57,$A7,$AB,$D7,$D5,$ED
	.dc.b	$EA,$F7,$F5,$7C,$1E,$AF,$87,$D5,$F1,$7A,$BE,$3F,$57,$C9,$EA,$F9
	.dc.b	$7D,$5F,$37,$AB,$E7,$F5,$7F,$D6,$BF,$65,$FE,$8A,$1F,$31,$8F,$95
	.dc.b	$CF,$C8,$EB,$E2,$37,$F1,$1C,$3E,$17,$8F,$C2,$F2,$F7,$39,$FB,$9D
	.dc.b	$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC,$F1
	.dc.b	$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$FE,$CC,$FB,$3F,$C3
	.dc.b	$79,$07,$F0,$CF,$FB,$61,$FF,$6A,$9F,$E8,$93,$FA,$A9,$FD,$51,$E1
	.dc.b	$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D
	.dc.b	$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1
	.dc.b	$EC,$7F,$B9,$3E,$D3,$F0,$FC,$47,$FD,$A6,$7F,$A5,$0F,$F4,$99,$FD
	.dc.b	$50,$FC,$38,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1,$DF
	.dc.b	$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6,$3C
	.dc.b	$DE,$C7,$9F,$D8,$FF,$7D,$FA,$7F,$ED,$03,$F0,$FC,$63,$FC,$B9,$EB
	.dc.b	$6B,$D6,$DF,$AD,$C2,$B8,$FA,$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD
	.dc.b	$DB,$D6,$EF,$EB,$78,$3D,$6F,$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6
	.dc.b	$F2,$FA,$DE,$6F,$5B,$CF,$EB,$7A,$3D,$6F,$37,$33,$ED,$7F,$12,$7E
	.dc.b	$21,$C7,$BB,$A8,$FF,$68,$1F,$E1,$C3,$F1,$07,$F8,$6A,$3E,$02,$F5
	.dc.b	$B8,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77
	.dc.b	$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B,$C9,$EB,$7A,$BE,$DB
	.dc.b	$F1,$23,$E1,$B1,$F0,$D9,$F8,$6D,$7C,$36,$FE,$1B,$87,$B7,$C7,$DB
	.dc.b	$E5,$ED,$F3,$F6,$FA,$7B,$7D,$7D,$BE,$DE,$DF,$7F,$6F,$C1,$ED,$F8
	.dc.b	$7D,$BF,$17,$B7,$E3,$F6,$FC,$9E,$DF,$97,$DB,$F3,$7B,$7F,$43,$ED
	.dc.b	$FE,$8F,$8E,$7F,$12,$DF,$73,$5E,$E6,$FD,$CE,$1E,$E7,$1F,$73,$97
	.dc.b	$B9,$CF,$DC,$E9,$EE,$75,$F7,$3B,$7B,$9D,$FD,$CF,$07,$B9,$E1,$F7
	.dc.b	$3C,$5E,$E7,$8F,$DC,$F2,$7B,$9E,$5F,$73,$CD,$EE,$79,$FD,$CF,$47
	.dc.b	$B9,$F8,$8F,$1F,$F8,$79,$F8,$71,$EE,$67,$E1,$35,$F0,$7B,$F8,$4E
	.dc.b	$1F,$03,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D,$FE,$DE,$FF,$7F
	.dc.b	$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E,$FF,$97,$DF,$F3
	.dc.b	$7B,$FF,$9E,$FB,$8F,$E1,$A7,$E1,$7C,$BF,$DB,$2F,$7B,$5E,$F6,$FD
	.dc.b	$EE,$1D,$5C,$7D,$EE,$5E,$F7,$3F,$7B,$A7,$BD,$D7,$DE,$ED,$EF,$77
	.dc.b	$F7,$BC,$1E,$F7,$87,$DE,$F1,$7B,$DE,$3F,$7B,$C9,$EF,$79,$7D,$EF
	.dc.b	$37,$BD,$E7,$F7,$BF,$6B,$F7,$3F,$86,$3F,$0A,$3D,$EC,$FB,$DA,$F7
	.dc.b	$B7,$EF,$70,$F7,$B8,$FB,$DC,$BD,$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB
	.dc.b	$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD,$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2
	.dc.b	$FB,$DE,$6F,$7B,$CF,$EF,$7F,$0C,$FB,$AF,$EA,$67,$F0,$B3,$F0,$BE
	.dc.b	$67,$F2,$55,$EF,$6F,$DE,$E1,$EF,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E
	.dc.b	$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD
	.dc.b	$EF,$27,$BD,$E5,$F7,$BC,$DE,$F7,$9F,$DE,$FE,$47,$C3,$FD,$48,$FC
	.dc.b	$37,$2F,$FB,$F7,$DD,$CF,$BB,$AF,$77,$7E,$EF,$0F,$77,$8F,$BB,$CB
	.dc.b	$DD,$E7,$EE,$F4,$F7,$7A,$FB,$BD,$BD,$DE,$FE,$EF,$83,$DD,$F0,$FB
	.dc.b	$BE,$2F,$77,$C7,$EE,$F9,$3D,$DF,$2F,$BB,$E6,$F7,$7F,$9C,$FD,$DF
	.dc.b	$F0,$A3,$F8,$49,$F8,$5E,$6F,$EA,$4B,$DC,$DF,$B9,$C3,$DC,$E3,$EE
	.dc.b	$72,$F7,$39,$FB,$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C
	.dc.b	$3E,$E7,$8B,$DC,$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9
	.dc.b	$E8,$F7,$3F,$A8,$FD,$E7,$F8,$60,$FE,$A0,$7E,$1B,$9C,$FE,$19,$FE
	.dc.b	$D8,$FF,$BB,$0F,$F0,$C1,$FD,$45,$DF,$B3,$D3,$D9,$EB,$EC,$F6,$F6
	.dc.b	$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE
	.dc.b	$CF,$9B,$D9,$F3,$FB,$3E,$8F,$67,$FB,$4F,$DE,$FF,$86,$0F,$C3,$9F
	.dc.b	$C2,$F8,$7F,$C0,$4B,$D8,$DF,$B1,$C3,$D8,$E3,$EC,$72,$F6,$39,$FB
	.dc.b	$1D,$3D,$8E,$BE,$C7,$6F,$63,$BF,$B1,$E0,$F6,$3C,$3E,$C7,$8B,$D8
	.dc.b	$F1,$FB,$1E,$4F,$63,$CB,$EC,$79,$BD,$8F,$3F,$B1,$FE,$00,$F5,$3F
	.dc.b	$A8,$9F,$D4,$0F,$C4,$73,$7F,$B9,$7D,$7D,$7A,$FB,$F5,$F8,$57,$1F
	.dc.b	$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F,$07,$AF
	.dc.b	$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB,$F9,$FD
	.dc.b	$7F,$F0,$BF,$DF,$7F,$97,$8F,$F0,$B1,$F8,$73,$F0,$DC,$C5,$EE,$F0
	.dc.b	$F7,$38,$FB,$BC,$BD,$AE,$7E,$D7,$4F,$6B,$AF,$B5,$DB,$DA,$EF,$ED
	.dc.b	$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7,$93,$DA,$F2,$FB,$5E,$6F
	.dc.b	$6B,$CF,$ED,$7F,$8B,$BE,$FF,$FC,$BA,$7F,$0C,$3F,$0D,$CA,$7F,$0C
	.dc.b	$FF,$92,$57,$B3,$C7,$D9,$E5,$EC,$F3,$F6,$7A,$7B,$3D,$7D,$97,$F0
	.dc.b	$0C,$2E,$5C,$E1,$F8,$07,$F9,$57,$E1,$9C,$FC,$63,$0D,$AF,$EF,$6D
	.dc.b	$9D,$C7,$39,$DB,$4D,$96,$36,$FB,$6B,$C2,$E8,$C4,$F2,$69,$9D,$53
	.dc.b	$C2,$B9,$C6,$2B,$83,$6B,$85,$6E,$FD,$04,$8F,$23,$5E,$F7,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF
	.dc.b	$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63
	.dc.b	$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF
	.dc.b	$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88
	.dc.b	$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A
	.dc.b	$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB
	.dc.b	$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B
	.dc.b	$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D
	.dc.b	$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B
	.dc.b	$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60
	.dc.b	$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC
	.dc.b	$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE
	.dc.b	$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78
	.dc.b	$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12
	.dc.b	$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71
	.dc.b	$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C
	.dc.b	$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C
	.dc.b	$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6
	.dc.b	$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09
	.dc.b	$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC
	.dc.b	$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE
	.dc.b	$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A
	.dc.b	$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12
	.dc.b	$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D
	.dc.b	$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12
	.dc.b	$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2
	.dc.b	$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C
	.dc.b	$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09
	.dc.b	$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA
	.dc.b	$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB
	.dc.b	$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E
	.dc.b	$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD
	.dc.b	$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9
	.dc.b	$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A
	.dc.b	$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46
	.dc.b	$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31
	.dc.b	$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9
	.dc.b	$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26
	.dc.b	$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12
	.dc.b	$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3
	.dc.b	$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C
	.dc.b	$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F
	.dc.b	$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18
	.dc.b	$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A
	.dc.b	$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C
	.dc.b	$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67
	.dc.b	$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB
	.dc.b	$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D
	.dc.b	$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38
	.dc.b	$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E
	.dc.b	$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97
	.dc.b	$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D
	.dc.b	$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57
	.dc.b	$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA
	.dc.b	$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3
	.dc.b	$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF
	.dc.b	$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17
	.dc.b	$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF
	.dc.b	$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E
	.dc.b	$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D
	.dc.b	$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35
	.dc.b	$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1
	.dc.b	$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D
	.dc.b	$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4
	.dc.b	$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F
	.dc.b	$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D2,$12,$95,$2F,$10,$BE,$15,$BA,$D4,$F1,$9C,$BC,$AB
	.dc.b	$8C,$62,$77,$3C,$EF,$C2,$EC,$E4,$E2,$37,$1E,$5F,$BB,$F3,$7C,$9B
	.dc.b	$E2,$BC,$B3,$91,$19,$73,$3C,$27,$70,$8C,$3A,$38,$C6,$A3,$31,$F0
	.dc.b	$9A,$F7,$9E,$CE,$4E,$27,$37,$2E,$E7,$A3,$A3,$84,$69,$F2,$19,$38
	.dc.b	$1C,$E3,$8C,$6E,$7C,$73,$C2,$30,$F1,$31,$3C,$08,$F3,$1F,$7C,$FC
	.dc.b	$F5,$F2,$6F,$97,$FC,$27,$E0,$36,$F7,$5A,$9C,$9F,$89,$BE,$57,$BB
	.dc.b	$E1,$18,$9D,$DF,$2A,$D4,$71,$BE,$92,$BE,$31,$86,$D7,$09,$E5,$5D
	.dc.b	$2F,$B4,$71,$8E,$93,$B9,$CD,$74,$9C,$5F,$BD,$E4,$DB,$E4,$7C,$08
	.dc.b	$D9,$0D,$95,$DD,$7C,$B8,$C5,$77,$76,$7C,$29,$FC,$D5,$FA,$DB,$1F
	.dc.b	$1B,$F0,$07,$C6,$3D,$DD,$67,$F0,$EB,$F6,$AD,$BF,$DD,$36,$F8,$FB
	.dc.b	$BC,$31,$97,$E1,$AD,$79,$1F,$C0,$63,$13,$AC,$3E,$09,$E4,$2D,$77
	.dc.b	$71,$9F,$E1,$5B,$5E,$4A,$E5,$39,$AE,$04,$8C,$CE,$CF,$09,$F0,$6D
	.dc.b	$F2,$BA,$C1,$92,$73,$5C,$27,$EB,$CE,$7E,$3B,$83,$67,$59,$7E,$6E
	.dc.b	$39,$7B,$FA,$C0,$FE,$1F,$F3,$D7,$A1,$4F,$C8,$57,$87,$E0,$31,$8F
	.dc.b	$E4,$6B,$6D,$AD,$E0,$F8,$0D,$E9,$F8,$6B,$DB,$6B,$7F,$01,$AD,$3F
	.dc.b	$53,$5A,$6C,$F4,$CB,$59,$7E,$C6,$B4,$D9,$F0,$E3,$AC,$7F,$7F,$5A
	.dc.b	$6C,$F8,$2B,$D0,$48,$F2,$37,$0F,$0B,$4E,$C5,$8E,$6D,$E4,$5F,$76
	.dc.b	$DF,$8D,$B3,$F7,$FE,$03,$CD,$AF,$DD,$6E,$AF,$80,$A6,$3E,$02,$5F
	.dc.b	$D1,$F5,$FE,$C1,$F4,$3E,$02,$C6,$49,$AF,$E1,$B0,$7C,$06,$A3,$EA
	.dc.b	$5D,$DF,$F0,$D4,$F8,$0C,$61,$F3,$EA,$DF,$CC,$6E,$DF,$01,$BC,$06
	.dc.b	$92,$BC,$78,$FD,$E8,$FC,$3F,$C0,$74,$E0,$FC,$F3,$73,$7C,$05,$31
	.dc.b	$F0,$13,$1F,$D4,$7C,$07,$E4,$3E,$03,$F2,$2D,$8F,$D6,$7C,$07,$E5
	.dc.b	$0F,$CA,$E7,$63,$F6,$1C,$DF,$98,$3F,$33,$F0,$16,$3F,$8A,$7C,$6F
	.dc.b	$F2,$5D,$FF,$54,$FF,$0F,$31,$1F,$E0,$BD,$D1,$9F,$1F,$F5,$2B,$E3
	.dc.b	$EC,$61,$CF,$C0,$EB,$86,$E1,$E0,$78,$FC,$0C,$8F,$1F,$C0,$DE,$7F
	.dc.b	$E0,$39,$3E,$07,$BF,$C0,$F8,$1B,$7E,$AF,$43,$C6,$79,$3E,$06,$C7
	.dc.b	$BB,$E0,$7C,$E7,$A3,$E0,$6C,$7C,$2F,$81,$F5,$9E,$CF,$81,$B1,$F1
	.dc.b	$BE,$07,$DF,$F0,$3F,$01,$B1,$F2,$BE,$07,$E1,$FC,$0F,$C4,$6C,$7C
	.dc.b	$EF,$81,$FE,$1F,$F1,$DF,$E1,$C8,$C1,$FD,$15,$C7,$17,$F2,$07,$FC
	.dc.b	$49,$19,$F8,$FE,$3E,$A7,$27,$9F,$C7,$49,$E5,$F0,$97,$0E,$5F,$09
	.dc.b	$E0,$F8,$4F,$0B,$6B,$C1,$F0,$9E,$43,$CB,$F0,$96,$3C,$9F,$03,$E8
	.dc.b	$3D,$3F,$1D,$63,$D1,$F1,$DE,$CF,$8E,$F6,$B6,$3D,$9F,$1D,$F0,$0F
	.dc.b	$83,$F1,$D6,$3E,$07,$C7,$7C,$43,$E2,$FC,$75,$8F,$89,$F1,$DF,$23
	.dc.b	$E3,$BE,$4B,$63,$E4,$7C,$37,$CC,$3E,$6F,$C3,$58,$FE,$E1,$F1,$FF
	.dc.b	$D8,$F9,$3F,$A5,$BF,$D3,$9A,$7F,$42,$71,$E7,$FF,$4B,$5F,$0B,$9F
	.dc.b	$85,$D3,$6F,$AF,$C2,$71,$F8,$4E,$4D,$8F,$0F,$C2,$F5,$F8,$5E,$CD
	.dc.b	$8F,$2F,$C2,$F8,$7E,$17,$C4,$D8,$F4,$FC,$2F,$94,$F3,$7C,$2D,$8F
	.dc.b	$6F,$C2,$FA,$7E,$0F,$D4,$D8,$F8,$3F,$07,$ED,$F8,$3F,$73,$63,$E2
	.dc.b	$FC,$17,$C1,$3E,$17,$C1,$D8,$F9,$3F,$07,$F1,$7E,$0F,$E3,$36,$3E
	.dc.b	$6F,$C1,$FC,$9F,$83,$F9,$4D,$8F,$F1,$07,$0F,$F0,$AE,$9F,$EF,$EF
	.dc.b	$F8,$6C,$C7,$C1,$9C,$7C,$1F,$F7,$47,$FD,$33,$5C,$BE,$0F,$13,$9F
	.dc.b	$83,$93,$DB,$E0,$EE,$3F,$E0,$8C,$1F,$07,$CF,$E0,$FA,$36,$FC,$BF
	.dc.b	$03,$DC,$F0,$7C,$15,$8F,$4F,$C1,$78,$FE,$0B,$C8,$D8,$F6,$FC,$17
	.dc.b	$9C,$F4,$7C,$15,$8F,$83,$F0,$5E,$B3,$D9,$F0,$56,$3E,$2F,$C0,$FB
	.dc.b	$CF,$81,$F0,$56,$3E,$4F,$C0,$FC,$33,$E2,$7C,$15,$8F,$9B,$F0,$5F
	.dc.b	$1C,$FF,$0D,$FC,$7F,$F0,$AF,$3B,$FC,$49,$C8,$9C,$FC,$16,$BE,$0A
	.dc.b	$E1,$FE,$55,$C1,$1C,$BE,$0A,$4E,$7E,$0A,$F4,$FF,$8D,$2D,$F0,$3E
	.dc.b	$03,$C3,$F0,$56,$FA,$7C,$17,$93,$E0,$BC,$AD,$8F,$07,$C1,$7A,$3E
	.dc.b	$0B,$D2,$D8,$F2,$7C,$0F,$B3,$E0,$BD,$AD,$8F,$47,$C1,$7C,$0F,$82
	.dc.b	$F8,$2D,$8F,$67,$C1,$7C,$4F,$82,$F8,$AD,$8F,$81,$F0,$3F,$23,$E0
	.dc.b	$BE,$4B,$63,$E2,$7C,$17,$CC,$3E,$6F,$C1,$58,$FF,$3E,$7C,$6F,$F0
	.dc.b	$DE,$CF,$F2,$93,$FC,$2A,$3F,$B4,$39,$6F,$C4,$7F,$94,$23,$F8,$6C
	.dc.b	$72,$F8,$3C,$7C,$1E,$5D,$7C,$16,$FE,$0E,$D7,$81,$E5,$E0,$C8,$F1
	.dc.b	$FC,$1D,$ED,$FF,$4C,$FE,$7A,$E1,$FF,$54,$36,$BD,$3F,$07,$89,$91
	.dc.b	$FE,$B1,$9D,$57,$F9,$CB,$57,$C2,$6A,$3A,$DD,$E0,$F8,$3F,$09,$71
	.dc.b	$FF,$65,$63,$CA,$7F,$D9,$2D,$8C,$D6,$EF,$FD,$B3,$39,$6C,$67,$E2
	.dc.b	$53,$96,$C6,$63,$04,$E5,$F7,$92,$FD,$8D,$BF,$8A,$DC,$3C,$AD,$9F
	.dc.b	$4B,$67,$CA,$FD,$9F,$85,$93,$F8,$2F,$84,$B5,$F1,$FE,$17,$EF,$1F
	.dc.b	$7F,$E1,$2D,$7E,$23,$E1,$7F,$05,$F0,$BF,$83,$6C,$7E,$33,$E1,$7F
	.dc.b	$0D,$F0,$9F,$87,$6C,$7D,$5E,$EF,$B6,$7E,$2F,$E1,$6C,$64,$99,$FD
	.dc.b	$9F,$84,$CB,$F8,$4A,$BC,$1F,$51,$B5,$F6,$1B,$3F,$51,$BC,$7F,$49
	.dc.b	$BC,$47,$7F,$08,$E7,$7C,$2F,$E7,$CF,$F5,$7F,$E0,$9F,$D0,$23,$39
	.dc.b	$79,$B3,$FF,$5D,$4F,$FB,$87,$2E,$D3,$FE,$E9,$8C,$65,$DE,$31,$F0
	.dc.b	$BE,$09,$FE,$8F,$A7,$FB,$21,$FF,$2A,$4F,$1F,$85,$F0,$FC,$37,$89
	.dc.b	$FD,$1F,$93,$27,$B7,$C3,$5C,$3F,$53,$37,$4F,$2F,$C2,$DB,$FF,$32
	.dc.b	$64,$F8,$6F,$6F,$C3,$7B,$9B,$3F,$AB,$C3,$86,$3F,$EE,$0B,$FF,$70
	.dc.b	$4E,$49,$1E,$07,$E0,$61,$C3,$89,$5D,$E7,$E2,$63,$DD,$A7,$2C,$BF
	.dc.b	$5F,$8F,$D0,$CF,$C3,$C1,$FA,$A8,$E9,$87,$F6,$6B,$E0,$7E,$E2,$33
	.dc.b	$F0,$DC,$38,$51,$C3,$0C,$3C,$B4,$B7,$CB,$5F,$DD,$CE,$1F,$61,$33
	.dc.b	$F2,$CE,$F3,$F2,$37,$1E,$EF,$E0,$7E,$17,$F1,$27,$E2,$BE,$1A,$C7
	.dc.b	$7D,$FE,$2F,$C3,$7B,$0F,$6F,$C3,$49,$FC,$57,$C3,$7E,$4B,$E1,$BF
	.dc.b	$26,$D8,$FE,$41,$EA,$7F,$26,$3F,$2F,$F0,$D6,$32,$4D,$FF,$1D,$F0
	.dc.b	$D9,$7F,$97,$D5,$E0,$FE,$3E,$DA,$FE,$46,$D9,$FE,$3E,$DE,$0F,$E3
	.dc.b	$AD,$E0,$3B,$DB,$F0,$3F,$0D,$FA,$A3,$F5,$7F,$0D,$71,$F2,$65,$FA
	.dc.b	$AF,$86,$E1,$D1,$FE,$01,$9C,$9F,$EB,$BF,$80,$FF,$03,$3E,$41,$AA
	.dc.b	$D1,$BF,$8E,$B3,$98,$E5,$F0,$F8,$AC,$E8,$C3,$FD,$33,$7B,$11,$9F
	.dc.b	$87,$E1,$87,$23,$C3,$7C,$AD,$F0,$FA,$C3,$8C,$62,$B6,$7A,$FE,$1F
	.dc.b	$58,$F8,$6E,$D7,$F8,$5A,$B7,$7C,$A3,$3F,$0D,$A7,$7F,$0B,$2B,$78
	.dc.b	$7A,$FE,$1B,$18,$7A,$E3,$DD,$F0,$BD,$8E,$FF,$0D,$6F,$8E,$2F,$85
	.dc.b	$F1,$9E,$4F,$86,$B3,$AE,$7F,$39,$E8,$F0,$AC,$72,$F3,$7D,$7F,$0D
	.dc.b	$EC,$6C,$76,$F6,$7D,$E7,$C0,$F6,$AC,$78,$BE,$1B,$E1,$9F,$13,$E1
	.dc.b	$AC,$7B,$BF,$05,$FF,$0B,$71,$7D,$67,$FC,$34,$D3,$E6,$B8,$E2,$3F
	.dc.b	$E1,$A8,$F8,$8D,$E8,$FC,$1A,$F8,$5E,$3F,$0D,$CB,$E1,$B9,$FC,$35
	.dc.b	$CB,$C4,$F6,$F8,$6E,$FF,$0D,$67,$CC,$F8,$BE,$1A,$47,$A7,$E1,$AE
	.dc.b	$BE,$2F,$86,$F3,$FC,$37,$A1,$B5,$FE,$1D,$EB,$FF,$11,$9E,$CF,$86
	.dc.b	$B1,$92,$5F,$B3,$E1,$B2,$FF,$8B,$2A,$F1,$7B,$BE,$1A,$DF,$F8,$82
	.dc.b	$6D,$7C,$BC,$2F,$9D,$FE,$1A,$9C,$F2,$FC,$AC,$B3,$E7,$7E,$A7,$1F
	.dc.b	$0E,$7F,$E2,$8B,$C9,$23,$27,$72,$E5,$FE,$2A,$8E,$7E,$2F,$F9,$00
	.dc.b	$93,$FE,$3D,$6C,$FF,$91,$9B,$5F,$E3,$66,$CF,$F8,$E9,$B3,$FE,$36
	.dc.b	$7F,$07,$F0,$B2,$7F,$15,$F0,$D6,$BF,$80,$0F,$BE,$7C,$37,$EE,$E3
	.dc.b	$0F,$E3,$0C,$46,$AB,$F2,$9F,$0F,$F9,$03,$F2,$3F,$0F,$63,$F7,$77
	.dc.b	$C9,$B5,$C6,$3C,$2D,$CF,$F8,$CC,$6C,$CE,$1E,$02,$57,$26,$C6,$A3
	.dc.b	$04,$7F,$3F,$F0,$FF,$A4,$F8,$7F,$D2,$B6,$75,$7F,$C9,$3F,$06,$FF
	.dc.b	$71,$5F,$86,$31,$3F,$C6,$25,$F1,$F7,$E2,$30,$D9,$FE,$27,$3F,$E1
	.dc.b	$CB,$FE,$24,$DB,$ED,$F0,$FC,$FE,$23,$FC,$FF,$EA,$4C,$38,$FC,$2D
	.dc.b	$C7,$8F,$C2,$F8,$4F,$17,$C4,$5A,$EF,$F1,$1E,$53,$CD,$F1,$16,$3F
	.dc.b	$D0,$D8,$F3,$D3,$DD,$95,$9F,$E9,$7D,$8D,$9F,$11,$EF,$F8,$8F,$80
	.dc.b	$DA,$EE,$4D,$BD,$B9,$F7,$7F,$D3,$79,$F5,$7F,$A5,$E1,$FE,$2A,$73
	.dc.b	$5F,$1A,$FA,$47,$C6,$6E,$7F,$4F,$E2,$3E,$81,$F4,$7E,$22,$DF,$B0
	.dc.b	$97,$F3,$B5,$F0,$6B,$EB,$C7,$9D,$7B,$B2,$F6,$BF,$5B,$5F,$B9,$F0
	.dc.b	$F2,$3E,$CF,$C4,$5C,$BC,$E4,$D7,$E4,$CF,$81,$FF,$53,$ED,$E0,$C5
	.dc.b	$F1,$1F,$07,$C4,$FF,$51,$CE,$7E,$23,$1E,$22,$38,$70,$61,$E5,$77
	.dc.b	$2E,$C7,$FA,$DA,$31,$9F,$8E,$30,$69,$FF,$5A,$C6,$B3,$F2,$CF,$E5
	.dc.b	$3E,$23,$F3,$47,$E6,$FE,$22,$C7,$1E,$2F,$C1,$7C,$47,$AC,$FE,$EF
	.dc.b	$F8,$3F,$F8,$4B,$2F,$F0,$2B,$FE,$08,$C7,$FC,$16,$E6,$8E,$6F,$F1
	.dc.b	$AB,$E1,$B9,$FC,$37,$46,$CF,$FB,$6E,$3F,$DC,$27,$83,$4B,$1F,$DF
	.dc.b	$AB,$FC,$0D,$8F,$F7,$D7,$F9,$93,$7C,$9F,$3F,$95,$C2,$3F,$DD,$B8
	.dc.b	$5C,$FF,$DB,$AD,$C3,$FD,$C2,$D9,$FF,$6E,$B7,$8F,$FD,$B6,$DE,$2F
	.dc.b	$F0,$E4,$F8,$7E,$1B,$E2,$9F,$1B,$E1,$AE,$1E,$B3,$D1,$F0,$DD,$78
	.dc.b	$3E,$2B,$F3,$3E,$1B,$E6,$FC,$35,$AF,$73,$F4,$3E,$1A,$47,$C1,$F8
	.dc.b	$6B,$9F,$C1,$F8,$6F,$AA,$7D,$6F,$85,$B5,$F1,$FE,$1B,$EC,$9F,$6B
	.dc.b	$E1,$AC,$7F,$32,$F4,$BF,$9D,$91,$C1,$30,$FE,$6B,$19,$6C,$67,$93
	.dc.b	$F9,$BF,$A4,$AD,$46,$1F,$F2,$27,$2B,$93,$F0,$98,$23,$1F,$0F,$A9
	.dc.b	$CC,$63,$1F,$F2,$1B,$6B,$FC,$AC,$DC,$3F,$CB,$ED,$9F,$F3,$2B,$67
	.dc.b	$B5,$74,$25,$F7,$3C,$1F,$11,$2F,$BC,$FE,$7E,$FD,$13,$9B,$F0,$CE
	.dc.b	$49,$19,$C3,$C7,$18,$BE,$73,$F0,$B7,$FF,$3F,$5F,$A9,$B7,$EA,$6D
	.dc.b	$7A,$9B,$3F,$1D,$B8,$7F,$A4,$3B,$7A,$12,$BE,$3C,$E8,$91,$9A,$F0
	.dc.b	$77,$78,$C9,$5C,$1B,$1A,$8C,$11,$C1,$30,$F4,$4F,$07,$D5,$DF,$EF
	.dc.b	$25,$79,$1B,$5F,$05,$B3,$C2,$74,$49,$D3,$63,$E3,$78,$5F,$2C,$91
	.dc.b	$F0,$1B,$3F,$35,$B3,$C2,$74,$49,$D3,$63,$E9,$78,$3F,$5C,$91,$F3
	.dc.b	$1B,$3F,$65,$B3,$C2,$74,$49,$D3,$63,$EF,$78,$9F,$83,$24,$7D,$86
	.dc.b	$CF,$E1,$5B,$3C,$27,$44,$9D,$36,$3F,$15,$E2,$FE,$3C,$91,$F8,$46
	.dc.b	$CF,$E4,$5B,$3C,$27,$44,$9D,$36,$3F,$2D,$E4,$7E,$6C,$91,$F9,$06
	.dc.b	$CF,$E7,$5B,$3C,$27,$44,$9D,$36,$3F,$45,$E4,$FE,$9C,$91,$F9,$C6
	.dc.b	$CF,$EA,$5B,$3C,$27,$44,$9D,$36,$3F,$5D,$E4,$7E,$CC,$91,$FA,$86
	.dc.b	$CF,$ED,$5B,$3C,$27,$44,$9D,$36,$3F,$B1,$FC,$47,$F6,$8F,$88,$FE
	.dc.b	$D2,$D9,$FE,$B5,$F1,$1F,$DB,$0F,$F7,$1F,$A1,$FE,$BC,$F0,$BF,$B4
	.dc.b	$C6,$3E,$2F,$3F,$0F,$6F,$FD,$BF,$7F,$EB,$97,$F8,$37,$9F,$FC,$34
	.dc.b	$8F,$F0,$26,$E9,$FC,$3D,$BA,$7F,$0F,$6C,$F2,$38,$12,$7F,$8B,$7A
	.dc.b	$1F,$C6,$49,$1F,$C3,$9B,$1A,$8C,$11,$C1,$2B,$F8,$F4,$F0,$7F,$90
	.dc.b	$7A,$3F,$C9,$49,$5F,$C6,$9B,$5F,$C9,$DB,$3C,$27,$44,$9D,$36,$3F
	.dc.b	$96,$7A,$5F,$CC,$49,$1F,$C9,$9B,$3F,$CC,$DB,$3C,$27,$44,$9D,$36
	.dc.b	$3F,$9C,$63,$FC,$F4,$91,$FC,$C9,$B3,$FC,$FD,$B3,$C2,$74,$49,$D3
	.dc.b	$63,$FA,$27,$A3,$FD,$24,$91,$FC,$F9,$B3,$FD,$2D,$B3,$C2,$74,$49
	.dc.b	$D3,$63,$FA,$87,$A9,$FD,$54,$91,$FD,$29,$B3,$FD,$5D,$B3,$C2,$74
	.dc.b	$49,$D3,$63,$FA,$E7,$A9,$FD,$84,$91,$FD,$59,$B3,$FD,$8D,$B3,$C2
	.dc.b	$74,$49,$D3,$63,$FB,$47,$AD,$FD,$B4,$91,$FD,$89,$B3,$FD,$BD,$B3
	.dc.b	$C2,$74,$49,$D3,$63,$FB,$A7,$AD,$FD,$E4,$91,$FD,$B9,$B3,$FD,$ED
	.dc.b	$B3,$C2,$74,$49,$D3,$63,$FC,$00,$FF,$81,$48,$7F,$7A,$2C,$7F,$81
	.dc.b	$DB,$1C,$1D,$12,$34,$D8,$FF,$12,$FE,$7A,$F6,$7F,$C6,$6D,$9F,$F1
	.dc.b	$6B,$6B,$FC,$5A,$DA,$E3,$5C,$09,$5C,$1B,$5C,$1B,$3C,$27,$AF,$B3
	.dc.b	$E0,$23,$82,$57,$82,$32,$D8,$CC,$FA,$A7,$2D,$8F,$4F,$8F,$EC,$23
	.dc.b	$82,$57,$B2,$32,$D8,$CC,$FC,$69,$CB,$63,$E2,$FB,$7F,$20,$8E,$09
	.dc.b	$5F,$22,$32,$D8,$CC,$FD,$29,$CB,$63,$E8,$FB,$7F,$50,$8E,$09,$5F
	.dc.b	$52,$32,$D8,$CC,$FD,$E9,$CB,$63,$EE,$FB,$9F,$80,$23,$82,$57,$E0
	.dc.b	$23,$2D,$8C,$CF,$E2,$A7,$2D,$8F,$C4,$FB,$BF,$8C,$23,$82,$57,$E3
	.dc.b	$23,$2D,$8C,$CF,$E5,$A7,$2D,$8F,$CA,$FC,$07,$E6,$08,$E0,$95,$F9
	.dc.b	$88,$CB,$63,$33,$FA,$29,$CB,$61,$28,$FF,$4D,$E6,$9C,$C6,$1A,$70
	.dc.b	$97,$9A,$AE,$11,$CE,$75,$39,$78,$C7,$82,$30,$EC,$C4,$F2,$AF,$04
	.dc.b	$64,$EF,$19,$AE,$AD,$8E,$31,$D9,$DB,$6B,$BD,$7F,$04,$F7,$FF,$81
	.dc.b	$92,$7F,$81,$B6,$BF,$81,$B6,$78,$4E,$89,$3A,$6C,$7F,$10,$F8,$0F
	.dc.b	$E1,$E4,$8F,$E1,$ED,$9F,$E1,$ED,$9E,$13,$A2,$4E,$9B,$1F,$C7,$3E
	.dc.b	$0B,$F8,$D9,$23,$F8,$DB,$67,$F8,$DB,$67,$84,$E8,$93,$A6,$C7,$F2
	.dc.b	$8F,$82,$FE,$4E,$48,$FE,$4E,$D9,$FE,$4E,$D9,$E1,$3A,$24,$E9,$B1
	.dc.b	$FC,$D3,$E0,$FF,$99,$92,$3F,$99,$B6,$7F,$99,$B6,$78,$4E,$89,$3A
	.dc.b	$6C,$7F,$40,$F8,$4F,$E7,$E4,$8F,$E7,$ED,$9F,$E7,$ED,$9E,$13,$A2
	.dc.b	$4E,$9B,$1F,$D3,$3E,$17,$FA,$59,$23,$FA,$5B,$67,$FA,$5B,$67,$84
	.dc.b	$E8,$93,$A6,$C7,$F5,$8F,$86,$FE,$AE,$48,$FE,$AE,$D9,$FE,$AE,$D9
	.dc.b	$E1,$3A,$24,$E9,$B1,$FD,$93,$E1,$FF,$B1,$92,$3F,$B1,$B6,$7F,$B1
	.dc.b	$B6,$78,$4E,$89,$3A,$6C,$7F,$70,$F8,$8F,$ED,$E4,$8F,$ED,$ED,$9F
	.dc.b	$ED,$ED,$9E,$13,$A2,$4E,$9B,$1F,$DF,$3E,$27,$FB,$D9,$23,$FB,$DB
	.dc.b	$67,$FB,$DB,$67,$84,$E8,$93,$A6,$C7,$F8,$23,$C4,$FF,$03,$92,$3F
	.dc.b	$C0,$ED,$9F,$F0,$3B,$67,$84,$E8,$93,$A6,$C7,$F8,$53,$C5,$FF,$09
	.dc.b	$92,$3F,$C2,$6D,$9F,$F0,$9B,$67,$84,$E8,$93,$A6,$C7,$F3,$2F,$1D
	.dc.b	$E1,$31,$FF,$57,$9F,$EB,$0C,$AD,$FF,$AB,$33,$FF,$59,$9F,$EB,$4D
	.dc.b	$2C,$7F,$AC,$35,$FF,$5B,$9F,$EB,$8D,$AC,$7F,$8A,$B4,$FF,$16,$91
	.dc.b	$C1,$1F,$F1,$69,$92,$B9,$D3,$FC,$6F,$19,$6C,$7F,$48,$F5,$5F,$9E
	.dc.b	$BC,$9F,$E8,$11,$86,$C6,$67,$83,$6F,$FA,$75,$6E,$73,$5C,$08,$E0
	.dc.b	$93,$C2,$B9,$5F,$F5,$EB,$E5,$3F,$D5,$79,$FF,$C9,$17,$FD,$83,$BF
	.dc.b	$FA,$A4,$FF,$66,$6C,$EA,$3F,$B1,$46,$09,$3F,$D5,$DB,$1D,$5B,$7F
	.dc.b	$DB,$A3,$A4,$FB,$BC,$BF,$82,$47,$04,$8F,$54,$65,$B1,$99,$F6,$62
	.dc.b	$AD,$46,$1F,$9B,$E5,$CA,$F9,$AD,$BF,$8C,$DA,$E6,$D9,$FF,$07,$E5
	.dc.b	$F6,$A3,$14,$8C,$4F,$43,$EC,$FB,$5F,$70,$8E,$09,$3F,$56,$32,$D8
	.dc.b	$CD,$FB,$27,$2D,$8F,$F0,$9F,$E1,$2F,$8B,$FD,$34,$61,$B7,$FE,$15
	.dc.b	$9D,$72,$7F,$83,$DB,$5B,$8C,$57,$0B,$FF,$11,$D6,$AF,$94,$E6,$FF
	.dc.b	$C6,$33,$9A,$DC,$E6,$93,$9A,$DC,$E6,$1C,$2E,$3F,$23,$6B,$DA,$DB
	.dc.b	$F5,$37,$0F,$2B,$67,$C6,$D9,$F0,$B6,$FB,$DF,$62,$57,$66,$D7,$F8
	.dc.b	$DF,$F0,$37,$8F,$FA,$26,$D7,$F8,$D5,$B3,$B9,$D1,$27,$4D,$8D,$46
	.dc.b	$08,$E0,$FF,$23,$43,$E1,$7F,$62,$D9,$FF,$22,$36,$77,$3A,$24,$E9
	.dc.b	$B1,$A8,$C1,$1C,$1F,$E5,$5F,$C0,$DE,$27,$F0,$06,$CF,$F9,$4D,$B3
	.dc.b	$B9,$D1,$27,$4D,$8D,$46,$08,$E0,$FF,$32,$7E,$04,$F7,$BF,$87,$B6
	.dc.b	$7F,$CC,$4D,$9D,$CE,$89,$3A,$6C,$6A,$30,$47,$07,$F9,$CF,$BF,$F2
	.dc.b	$31,$FE,$6F,$6C,$67,$6F,$E4,$31,$8D,$FF,$8C,$36,$BF,$D1,$4D,$AF
	.dc.b	$F4,$9B,$53,$49,$5D,$09,$3F,$E7,$FF,$C0,$9E,$BF,$F3,$06,$C6,$63
	.dc.b	$04,$70,$4B,$FF,$47,$4F,$06,$CE,$DB,$3B,$C3,$F9,$97,$E0,$5F,$D1
	.dc.b	$23,$FD,$48,$D9,$FF,$56,$B5,$34,$9D,$9F,$D1,$9B,$3F,$E9,$96,$CF
	.dc.b	$FA,$85,$B5,$FD,$F7,$C0,$78,$7F,$EA,$1F,$0B,$FB,$FB,$61,$E2,$58
	.dc.b	$C7,$8B,$9F,$1A,$E7,$FE,$B4,$C3,$FC,$0F,$1C,$7C,$79,$39,$C5,$E4
	.dc.b	$E3,$5F,$F5,$C6,$1E,$0A,$D7,$2F,$76,$9C,$72,$F0,$34,$E9,$97,$3D
	.dc.b	$7D,$99,$FA,$72,$C5,$7F,$B5,$B1,$FF,$0D,$F9,$5E,$D3,$DD,$E5,$D9
	.dc.b	$F3,$79,$9F,$04,$F8,$5E,$6D,$8F,$57,$9B,$F1,$4F,$8D,$E4,$58,$FF
	.dc.b	$64,$7C,$83,$D3,$FE,$F2,$D8,$CC,$60,$8E,$09,$C5,$FE,$D7,$9E,$0D
	.dc.b	$9D,$B6,$77,$BF,$F8,$DF,$F0,$AF,$F7,$84,$7D,$36,$CF,$CE,$7F,$93
	.dc.b	$CE,$E7,$F0,$10,$F4,$AF,$A7,$86,$CF,$3F,$53,$67,$0C,$6C,$77,$39
	.dc.b	$F3,$F4,$69,$E3,$AE,$79,$78,$DB,$97,$A7,$0E,$73,$E7,$C2,$CF,$F8
	.dc.b	$5F,$F0,$27,$C4,$FF,$89,$1B,$8F,$FB,$DD,$7C,$56,$66,$4E,$5F,$7F
	.dc.b	$B1,$8C,$58,$75,$9C,$2A,$4E,$BF,$52,$7B,$BA,$9C,$CB,$2F,$1B,$D2
	.dc.b	$33,$18,$BE,$D1,$E3,$24,$EC,$EF,$3B,$8E,$38,$77,$9C,$B5,$F0,$36
	.dc.b	$5F,$0D,$FE,$48,$FD,$65,$D1,$F1,$E3,$0F,$3F,$3D,$F4,$EF,$95,$C3
	.dc.b	$73,$F2,$E9,$F0,$F4,$F8,$5F,$05,$E3,$86,$E8,$FC,$5F,$C0,$7E,$33
	.dc.b	$E0,$6E,$1F,$93,$7F,$1F,$F0,$52,$3F,$2D,$C5,$72,$FC,$B6,$1F,$94
	.dc.b	$3F,$2B,$E5,$DA,$FC,$E7,$C1,$FE,$60,$EC,$E0,$FD,$06,$1C,$F4,$FD
	.dc.b	$26,$36,$7F,$81,$76,$79,$DF,$D3,$63,$D9,$A7,$EA,$7E,$13,$F4,$BF
	.dc.b	$0B,$FA,$66,$C7,$EB,$7E,$1B,$F5,$3F,$0F,$FA,$A6,$C7,$EC,$7E,$23
	.dc.b	$F5,$BF,$13,$FA,$E6,$C7,$ED,$70,$FE,$61,$F8,$6F,$F6,$DD,$7F,$75
	.dc.b	$18,$F5,$FA,$57,$EE,$F2,$BC,$5F,$BB,$7A,$47,$EF,$F2,$B3,$FB,$F7
	.dc.b	$FD,$13,$F8,$2B,$CA,$BE,$56,$1B,$7F,$9E,$CF,$FD,$09,$4A,$CE,$1F
	.dc.b	$E5,$9F,$A6,$79,$38,$BD,$D5,$C3,$97,$53,$87,$18,$FE,$AB,$2E,$95
	.dc.b	$DC,$FD,$4B,$88,$FD,$10,$71,$87,$69,$E1,$3C,$F0,$EA,$ED,$B5,$C2
	.dc.b	$7A,$C6,$0F,$F5,$5F,$C1,$3C,$AA,$7F,$AB,$BF,$04,$FE,$5F,$8F,$FA
	.dc.b	$96,$75,$7E,$0C,$7B,$3E,$29,$7E,$36,$E1,$F7,$C6,$1B,$F8,$E3,$E5
	.dc.b	$29,$F8,$0F,$92,$98,$7F,$69,$E7,$F9,$AD,$8F,$C2,$FC,$C4,$8D,$78
	.dc.b	$97,$4F,$ED,$36,$F9,$AF,$C5,$FC,$D7,$E3,$1B,$7D,$7E,$6B,$F1,$FF
	.dc.b	$0D,$F9,$06,$C7,$87,$2E,$9B,$FF,$C0,$1C,$7F,$81,$65,$DA,$7A,$57
	.dc.b	$9F,$2B,$3E,$27,$F3,$7F,$35,$27,$D9,$F3,$56,$BC,$FE,$97,$E8,$0F
	.dc.b	$D0,$FA,$36,$BE,$17,$9F,$FA,$4E,$2F,$D2,$B6,$3E,$37,$CD,$7E,$A3
	.dc.b	$E6,$BF,$52,$D8,$F9,$58,$73,$C7,$E2,$E3,$77,$F6,$63,$CE,$7E,$4B
	.dc.b	$63,$E0,$74,$7E,$D3,$E4,$7F,$6A,$DC,$7E,$B7,$C9,$7E,$E3,$E1,$FF
	.dc.b	$72,$D8,$FF,$11,$7E,$21,$FD,$21,$B9,$7F,$8B,$61,$F3,$36,$31,$F1
	.dc.b	$99,$F9,$3B,$7C,$9E,$AF,$E1,$71,$B5,$F8,$07,$94,$76,$3C,$37,$89
	.dc.b	$FF,$16,$E0,$F9,$9F,$17,$C9,$F8,$DB,$7E,$C3,$B1,$6B,$B3,$D5,$E6
	.dc.b	$7A,$EF,$B6,$DF,$D1,$7E,$15,$8B,$E6,$7E,$01,$F0,$7E,$42,$E3,$F1
	.dc.b	$BE,$67,$E2,$1F,$17,$E3,$AC,$7C,$AF,$8F,$F9,$1F,$1D,$F2,$5B,$1F
	.dc.b	$3B,$E3,$BE,$67,$C7,$7C,$D6,$C7,$52,$6F,$F0,$B0,$EA,$FC,$4B,$E6
	.dc.b	$D9,$FB,$38,$7D,$7F,$8E,$FB,$1F,$1F,$67,$EE,$BF,$6F,$E3,$FE,$E7
	.dc.b	$C3,$59,$FC,$0B,$F7,$FD,$79,$1F,$84,$F4,$AE,$5F,$53,$E3,$FF,$09
	.dc.b	$F1,$FF,$85,$6D,$7F,$9A,$F8,$BF,$C7,$3C,$DF,$35,$FF,$1C,$C7,$F9
	.dc.b	$6D,$CB,$7E,$53,$FC,$F5,$1F,$E3,$B8,$E5,$EA,$E3,$E3,$F2,$EB,$E3
	.dc.b	$E4,$F7,$F8,$FB,$A7,$F9,$D3,$0E,$AB,$AD,$1E,$28,$C3,$A8,$CE,$5D
	.dc.b	$EF,$C2,$7A,$F2,$F0,$BE,$DF,$90,$F3,$9E,$8F,$90,$B1,$F0,$7E,$33
	.dc.b	$D7,$F1,$9E,$C6,$C7,$C5,$F8,$AF,$79,$F0,$3E,$2A,$C7,$C9,$F8,$9F
	.dc.b	$87,$F1,$9F,$11,$B1,$F3,$7E,$33,$E3,$FC,$5F,$C8,$6C,$7D,$1F,$8A
	.dc.b	$F9,$7F,$0B,$F3,$1B,$1F,$57,$E0,$BE,$7F,$C5,$FD,$06,$C7,$D9,$F8
	.dc.b	$AF,$A6,$7F,$AB,$7F,$11,$FE,$FD,$1F,$66,$31,$F0,$92,$3F,$D5,$53
	.dc.b	$FD,$0B,$A3,$FD,$4A,$A7,$97,$BF,$2B,$7F,$09,$6F,$7F,$09,$DC,$F0
	.dc.b	$7C,$0D,$AE,$BF,$07,$E3,$F8,$4F,$23,$63,$C3,$F0,$7E,$73,$D1,$F0
	.dc.b	$76,$3C,$BF,$01,$EB,$F8,$0F,$63,$63,$D3,$F0,$1E,$FF,$83,$F8,$0D
	.dc.b	$8F,$6F,$C1,$7C,$3F,$83,$F8,$8D,$8F,$83,$F0,$5F,$1F,$E0,$FE,$43
	.dc.b	$63,$E2,$FC,$1F,$CB,$3E,$67,$C0,$D8,$FF,$69,$7E,$25,$FC,$C7,$93
	.dc.b	$FD,$9A,$DF,$3F,$F8,$B3,$FC,$CA,$77,$3F,$EE,$18,$7C,$25,$F8,$4C
	.dc.b	$36,$BA,$FC,$1E,$FE,$13,$83,$63,$C3,$F0,$5C,$CE,$9E,$8D,$8F,$2F
	.dc.b	$C1,$77,$3C,$1F,$01,$63,$D3,$F0,$1E,$3F,$80,$F2,$36,$3D,$BF,$03
	.dc.b	$E7,$F8,$2F,$43,$63,$E0,$FC,$17,$AF,$E0,$BD,$8D,$8F,$8B,$F0,$5E
	.dc.b	$F3,$E0,$7C,$15,$8F,$93,$F0,$5F,$0C,$F8,$9F,$05,$63,$E6,$FC,$0F
	.dc.b	$C7,$3E,$47,$C1,$58,$FF,$1A,$FB,$EF,$C4,$FF,$82,$1F,$C1,$DC,$47
	.dc.b	$FA,$2F,$AF,$FA,$44,$E4,$FE,$31,$D7,$FE,$A3,$AE,$5E,$FE,$EF,$FD
	.dc.b	$F5,$5F,$E3,$4A,$D3,$C2,$F9,$FC,$16,$E7,$F8,$C4,$A3,$18,$79,$7E
	.dc.b	$0A,$CE,$70,$F4,$AF,$81,$CF,$C0,$E9,$B3,$AF,$80,$E2,$72,$F8,$1B
	.dc.b	$1C,$BE,$07,$AF,$C0,$F6,$6C,$76,$F8,$0F,$0F,$C0,$F8,$9B,$1E,$2F
	.dc.b	$80,$F2,$9E,$6F,$81,$B1,$E6,$F8,$0F,$49,$EA,$F8,$1B,$1E,$AF,$80
	.dc.b	$F6,$FC,$0F,$B9,$B1,$EE,$F8,$1F,$87,$ED,$7F,$14,$E7,$F8,$F1,$8F
	.dc.b	$8E,$CF,$C0,$CB,$FB,$B7,$FC,$57,$AB,$FC,$00,$FF,$15,$8C,$FC,$0F
	.dc.b	$4F,$81,$EA,$D8,$E3,$F0,$3E,$03,$C3,$F0,$36,$3A,$DF,$93,$E0,$7C
	.dc.b	$AD,$8F,$0F,$C0,$FA,$3E,$07,$D2,$D8,$F2,$FC,$07,$B3,$E0,$7D,$AD
	.dc.b	$8F,$4F,$C0,$7C,$03,$E0,$FC,$0D,$8F,$6F,$C0,$7C,$43,$E2,$FC,$0D
	.dc.b	$8F,$83,$F0,$1F,$23,$E0,$7E,$4B,$63,$E2,$FC,$0F,$CC,$F8,$1F,$9A
	.dc.b	$D8,$FD,$2F,$E2,$9F,$C4,$D7,$57,$F6,$D8,$FE,$28,$FE,$85,$D3,$9F
	.dc.b	$87,$FC,$CC,$7F,$16,$9E,$7F,$03,$8F,$82,$CB,$AF,$81,$93,$E0,$F8
	.dc.b	$2B,$8F,$5F,$82,$E6,$74,$F8,$2B,$5E,$6F,$82,$EE,$78,$3E,$0A,$C7
	.dc.b	$AB,$E0,$7C,$67,$93,$E0,$AC,$7B,$BE,$07,$CE,$7A,$3E,$0A,$C7,$FA
	.dc.b	$AF,$F0,$5F,$DA,$78,$FF,$0B,$58,$23,$2E,$0C,$7C,$3F,$F9,$A9,$C4
	.dc.b	$69,$E2,$D7,$A3,$5F,$9E,$FF,$B0,$79,$3A,$47,$7F,$6D,$78,$AE,$13
	.dc.b	$98,$C5,$F8,$63,$BD,$70,$7D,$3E,$DB,$99,$F0,$7E,$DF,$2B,$FC,$CD
	.dc.b	$18,$AF,$C1,$CE,$74,$FC,$1D,$F1,$F8,$4C,$57,$E1,$7E,$12,$CF,$56
	.dc.b	$E0,$6F,$C2,$63,$E1,$25,$FE,$41,$DF,$C2,$70,$F8,$49,$3F,$94,$79
	.dc.b	$FC,$27,$6F,$C5,$5A,$D2,$57,$F3,$E8,$7C,$46,$FE,$1A,$4E,$5E,$5C
	.dc.b	$7C,$F8,$FA,$7C,$34,$AE,$51,$DF,$E1,$24,$75,$7C,$5F,$0B,$E3,$F8
	.dc.b	$59,$3E,$17,$CD,$F0,$BE,$7F,$84,$F4,$7C,$2C,$AF,$34,$7A,$FE,$13
	.dc.b	$D9,$F0,$B2,$7D,$51,$EF,$F8,$5F,$81,$F0,$B2,$7D,$D1,$F0,$FE,$17
	.dc.b	$E2,$7C,$2C,$9F,$85,$1F,$1F,$E1,$24,$7C,$57,$E5,$7C,$2F,$CB,$F8
	.dc.b	$59,$3F,$25,$FB,$7F,$8B,$7E,$26,$3F,$B0,$3F,$E3,$F7,$FB,$5A,$F9
	.dc.b	$1D,$9C,$3E,$1F,$8F,$C3,$4A,$DB,$D3,$E1,$FA,$FC,$3C,$9E,$6F,$83
	.dc.b	$E1,$FC,$3F,$0D,$27,$BB,$E4,$F8,$7F,$2F,$C3,$49,$F1,$BE,$8F,$87
	.dc.b	$F4,$FC,$34,$9F,$3B,$EC,$F8,$7F,$6F,$C3,$C9,$F5,$BF,$03,$E1,$FE
	.dc.b	$0F,$C3,$C9,$F7,$BF,$13,$E1,$FE,$2F,$C3,$7C,$6F,$87,$95,$F1,$23
	.dc.b	$E4,$FC,$37,$CA,$F8,$79,$3F,$22,$3F,$4F,$E2,$FF,$0C,$F6,$3F,$DA
	.dc.b	$CD,$F3,$BF,$5E,$DA,$FF,$4F,$3F,$BD,$9D,$FC,$37,$2F,$87,$95,$FE
	.dc.b	$1E,$AF,$F2,$3B,$7E,$1F,$B9,$1D,$7C,$3F,$8B,$E1,$FC,$7F,$0F,$27
	.dc.b	$94,$79,$BE,$1F,$CF,$F0,$F2,$7B,$47,$AB,$E1,$FD,$7F,$0F,$27,$C5
	.dc.b	$1E,$EF,$87,$91,$E5,$7E,$0F,$C3,$FC,$2F,$84,$F8,$7F,$09,$2B,$D5
	.dc.b	$1F,$1B,$E0,$BE,$3F,$C1,$49,$F7,$47,$CA,$F8,$2F,$97,$F0,$52,$7E
	.dc.b	$14,$7C,$EF,$82,$91,$F1,$5F,$A3,$F0,$5F,$C6,$3F,$17,$FE,$17,$3F
	.dc.b	$E4,$66,$BF,$E3,$F6,$F4,$EB,$E1,$51,$FC,$3A,$78,$7C,$17,$3F,$82
	.dc.b	$91,$83,$A7,$C0,$F7,$23,$DF,$E0,$BC,$5F,$05,$E3,$ED,$93,$E3,$8F
	.dc.b	$37,$C1,$48,$F3,$3E,$9F,$82,$F5,$7C,$0F,$AF,$E0,$A5,$7A,$E3,$DD
	.dc.b	$F0,$32,$3D,$CF,$C1,$F8,$2F,$85,$F0,$32,$7E,$13,$F1,$7E,$0B,$E3
	.dc.b	$7C,$0F,$C7,$F8,$29,$5F,$1E,$3E,$57,$C1,$48,$F9,$4F,$CD,$F8,$2F
	.dc.b	$9D,$F0,$52,$7E,$73,$FD,$23,$F1,$8F,$E1,$DD,$3F,$35,$FE,$1D,$3F
	.dc.b	$E6,$F7,$F8,$73,$FE,$7A,$6B,$9F,$81,$E5,$F0,$3C,$FE,$06,$4E,$E3
	.dc.b	$B7,$C0,$C8,$E4,$F8,$7E,$07,$C5,$F0,$3E,$3F,$81,$95,$DE,$3C,$DF
	.dc.b	$03,$E7,$F8,$19,$3E,$38,$F5,$7C,$0C,$8F,$33,$ED,$F8,$1F,$77,$C0
	.dc.b	$C9,$F5,$3F,$07,$E0,$7E,$17,$C0,$FC,$3F,$81,$95,$EF,$8F,$8D,$F0
	.dc.b	$1F,$1F,$E0,$64,$FC,$38,$F9,$5F,$01,$F2,$FE,$06,$4F,$C7,$8F,$EE
	.dc.b	$BF,$8C,$FF,$0B,$F3,$4F,$F0,$BF,$82,$FE,$FA,$DA,$4E,$9F,$09,$F4
	.dc.b	$71,$F0,$BF,$57,$1F,$0B,$F5,$71,$F0,$BF,$67,$1F,$0F,$C1,$48,$FF
	.dc.b	$59,$E3,$FC,$59,$FF,$56,$2F,$82,$F3,$FC,$14,$9D,$BE,$AF,$82,$F5
	.dc.b	$FC,$0F,$B3,$E0,$A5,$74,$8F,$7F,$C0,$FC,$0F,$82,$93,$E0,$8F,$87
	.dc.b	$F0,$3F,$13,$E0,$A4,$F9,$23,$E3,$FC,$17,$C8,$F8,$29,$3E,$88,$F9
	.dc.b	$7F,$03,$F3,$3E,$0A,$4F,$B2,$3A,$BE,$EC,$3A,$BE,$EC,$3A,$BF,$0B
	.dc.b	$0E,$AF,$C2,$C3,$EB,$FC,$0F,$D8,$F8,$29,$3F,$22,$3E,$DF,$C0,$C8
	.dc.b	$F9,$6F,$DE,$F8,$2F,$BF,$F0,$5F,$80,$F8,$29,$5F,$42,$3F,$C6,$FB
	.dc.b	$FF,$1C,$ED,$FF,$1E,$B6,$7F,$02,$DE,$D3,$FC,$7E,$3E,$BE,$3E,$16
	.dc.b	$9F,$73,$1F,$0B,$F7,$B1,$F0,$BF,$7B,$1F,$0F,$C0,$F8,$BE,$06,$4F
	.dc.b	$FB,$DB,$14,$F5,$F4,$3F,$98,$75,$B3,$C6,$BF,$98,$D3,$BE,$3F,$8A
	.dc.b	$D3,$BE,$1C,$34,$F7,$FC,$0F,$C0,$F8,$19,$7D,$E3,$E1,$FC,$07,$C4
	.dc.b	$F8,$19,$3E,$38,$F8,$FF,$03,$23,$CC,$FC,$AF,$81,$F9,$7F,$01,$27
	.dc.b	$D4,$FC,$EF,$81,$EB,$1E,$DC,$3A,$AD,$3F,$A6,$71,$FF,$BF,$63,$A9
	.dc.b	$0D,$70,$7D,$6C,$BC,$11,$F1,$FE,$07,$ED,$91,$F9,$5F,$03,$F7,$BE
	.dc.b	$07,$EF,$FC,$0C,$9F,$9D,$1F,$82,$F8,$0F,$C1,$FC,$0C,$9F,$A5,$1F
	.dc.b	$86,$F8,$1F,$C3,$FC,$0C,$9F,$AD,$1D,$5F,$B1,$87,$57,$CE,$7E,$2B
	.dc.b	$0E,$CF,$DD,$C3,$B3,$AE,$DF,$C6,$E3,$F9,$4F,$81,$FF,$45,$FE,$35
	.dc.b	$FD,$13,$7F,$F4,$8B,$6F,$F2,$B3,$AF,$81,$91,$FE,$D8,$E2,$FE,$ED
	.dc.b	$5F,$D2,$9B,$F0,$9D,$3E,$07,$AF,$C1,$4B,$E1,$97,$69,$F0,$FC,$17
	.dc.b	$8B,$E0,$A5,$68,$FE,$4B,$92,$7C,$31,$EB,$8F,$05,$FA,$33,$F5,$E5
	.dc.b	$E1,$AE,$5B,$2D,$9F,$8E,$79,$67,$F0,$BE,$07,$E1,$FC,$14,$9E,$36
	.dc.b	$9F,$0E,$DF,$03,$3B,$A7,$C8,$D3,$C1,$87,$97,$3E,$D7,$F1,$B3,$F9
	.dc.b	$FF,$03,$2B,$AB,$F4,$BE,$0B,$E9,$FC,$14,$9F,$0B,$F5,$BE,$0B,$EB
	.dc.b	$FC,$17,$D8,$F8,$29,$5E,$68,$FB,$7F,$03,$23,$D2,$FD,$EF,$82,$E9
	.dc.b	$1F,$27,$0B,$A7,$DA,$C7,$AC,$FC,$0C,$7A,$BF,$3F,$1F,$C3,$FC,$0F
	.dc.b	$E2,$3E,$0A,$4F,$C6,$8F,$C5,$FC,$0F,$E3,$3E,$0A,$4F,$CA,$8F,$F6
	.dc.b	$CF,$E3,$7F,$D9,$4F,$EC,$5D,$26,$99,$F8,$3D,$FC,$1C,$BF,$B7,$3F
	.dc.b	$D6,$F6,$FC,$8E,$1D,$A7,$F0,$98,$F8,$9D,$9E,$0F,$8C,$8E,$0E,$7F
	.dc.b	$09,$E4,$F8,$49,$18,$3B,$FC,$1F,$A3,$E1,$24,$60,$F1,$F9,$BE,$C2
	.dc.b	$3E,$6F,$84,$F7,$FC,$0C,$8F,$4B,$F0,$BE,$0B,$E1,$E1,$D2,$7F,$6D
	.dc.b	$BB,$1C,$C7,$62,$1A,$C7,$E5,$E3,$F3,$3E,$07,$A4,$FC,$6C,$70,$F9
	.dc.b	$CE,$B7,$DF,$83,$FC,$DD,$8B,$2F,$4D,$75,$3B,$CE,$B1,$FB,$3F,$05
	.dc.b	$27,$E9,$BF,$73,$E0,$FE,$EF,$C2,$49,$FA,$EF,$E0,$3E,$0F,$F0,$3F
	.dc.b	$09,$F8,$2F,$81,$95,$F7,$23,$F0,$BF,$03,$CE,$3E,$FE,$18,$79,$90
	.dc.b	$F8,$F8,$75,$79,$1F,$86,$BE,$AF,$CC,$BF,$C8,$FC,$07,$E4,$BE,$06
	.dc.b	$4F,$E3,$20,$FF,$08,$FC,$73,$FB,$A1,$FD,$CF,$6F,$F4,$A3,$7C,$BD
	.dc.b	$F7,$23,$7F,$09,$CA,$FF,$98,$47,$F6,$E7,$A4,$66,$F9,$FC,$2D,$C3
	.dc.b	$9D,$BE,$0F,$18,$7F,$91,$A3,$0F,$96,$33,$8F,$87,$E1,$2D,$7A,$3E
	.dc.b	$26,$98,$C3,$39,$CC,$7C,$67,$BF,$2E,$EB,$F0,$97,$C2,$E3,$F0,$97
	.dc.b	$8F,$4D,$CE,$AB,$F9,$C1,$C7,$E1,$B9,$FE,$41,$FE,$8A,$E1,$FF,$5F
	.dc.b	$B7,$4F,$8D,$6A,$E6,$70,$F4,$E6,$1D,$27,$4D,$9D,$B4,$87,$F6,$43
	.dc.b	$C4,$63,$0F,$F6,$7C,$F2,$8D,$36,$33,$3D,$27,$2D,$8E,$3F,$0F,$F0
	.dc.b	$09,$3F,$03,$DF,$79,$3B,$CB,$DB,$85,$AF,$C6,$E7,$FD,$B5,$F2,$9D
	.dc.b	$8F,$34,$FE,$26,$3F,$AE,$C7,$E0,$7E,$23,$EE,$FC,$44,$8F,$CB,$47
	.dc.b	$E0,$3E,$22,$47,$E6,$5F,$C1,$F0,$FE,$13,$19,$3F,$9D,$7F,$0F,$A7
	.dc.b	$E2,$3E,$22,$4F,$E8,$5F,$C5,$FC,$47,$48,$FC,$DE,$17,$7F,$D1,$E3
	.dc.b	$D2,$7F,$3F,$8D,$9F,$D3,$E3,$F9,$4F,$85,$95,$FA,$D7,$F2,$FF,$0D
	.dc.b	$F9,$8F,$86,$93,$FB,$17,$F3,$7C,$5F,$9C,$F8,$79,$3F,$B5,$7F,$3F
	.dc.b	$F1,$1C,$A3,$F7,$17,$5F,$D8,$FC,$47,$47,$F6,$78,$5C,$FF,$77,$8F
	.dc.b	$49,$FD,$BE,$36,$7F,$7F,$8F,$EB,$BE,$1B,$FC,$49,$E0,$1A,$4F,$84
	.dc.b	$D7,$C2,$4C,$7F,$C2,$33,$F9,$F7,$FD,$7E,$BA,$FA,$7C,$27,$5F,$87
	.dc.b	$95,$C2,$3C,$1F,$11,$E1,$C6,$4F,$48,$EC,$FF,$8F,$70,$ED,$97,$97
	.dc.b	$2F,$46,$5D,$AB,$C7,$97,$67,$86,$5E,$DF,$87,$EF,$1D,$4F,$4B,$DB
	.dc.b	$57,$6F,$F2,$0C,$FB,$93,$1E,$D7,$E3,$C7,$B5,$F0,$8F,$83,$8F,$CB
	.dc.b	$B9,$1F,$19,$F9,$DF,$0F,$F3,$FE,$1F,$E8,$7C,$2C,$AF,$97,$1F,$4F
	.dc.b	$E1,$64,$7C,$E7,$EB,$7C,$27,$D7,$F8,$5E,$73,$F4,$B0,$C3,$C8,$FA
	.dc.b	$B8,$75,$7E,$56,$1D,$1F,$99,$85,$CF,$D5,$C3,$F8,$1C,$BF,$09,$F0
	.dc.b	$FF,$85,$F8,$79,$87,$E0,$23,$F1,$1E,$44,$8F,$C1,$BF,$8B,$F8,$7F
	.dc.b	$C6,$7C,$2F,$E3,$7C,$99,$5F,$88,$8F,$E7,$BF,$2F,$FD,$2E,$3E,$F6
	.dc.b	$1E,$07,$FC,$BB,$19,$D3,$CB,$C5,$FC,$DA,$7B,$90,$FC,$36,$3E,$07
	.dc.b	$F2,$B8,$F8,$1F,$C5,$63,$E0,$F8,$69,$1D,$9F,$1F,$C4,$79,$3E,$23
	.dc.b	$CB,$F1,$12,$BC,$71,$E8,$F8,$8F,$4F,$C1,$49,$F3,$C7,$B3,$E1,$39
	.dc.b	$C7,$AB,$0C,$3C,$8F,$6E,$1D,$5F,$F4,$D6,$EC,$7B,$47,$C2,$C7,$B3
	.dc.b	$AC,$7E,$4F,$C3,$F4,$8E,$39,$61,$EC,$74,$C7,$B7,$07,$DE,$CB,$B4
	.dc.b	$FA,$0E,$C4,$3B,$1F,$4B,$E1,$3E,$B9,$1F,$AB,$F1,$1F,$6B,$E1,$FE
	.dc.b	$DF,$C4,$49,$FB,$31,$F7,$BE,$1F,$EF,$FC,$2C,$9F,$BB,$1F,$82,$B9
	.dc.b	$1D,$0F,$C0,$FC,$2F,$57,$F0,$78,$75,$7E,$3E,$1D,$5F,$C3,$E1,$D5
	.dc.b	$F9,$78,$7E,$3B,$E1,$E4,$7E,$31,$FC,$8F,$C2,$FE,$4B,$E1,$BF,$27
	.dc.b	$F0,$F2,$BF,$23,$1F,$96,$F8,$6F,$CB,$FC,$3C,$9F,$CA,$C7,$F8,$5B
	.dc.b	$E5,$FF,$81,$47,$DA,$C3,$AB,$FE,$B6,$8C,$F1,$7D,$1C,$3B,$4F,$E0
	.dc.b	$31,$F1,$3D,$0F,$D0,$E3,$DD,$FC,$2E,$3D,$FE,$16,$47,$57,$C5,$F0
	.dc.b	$BE,$3F,$85,$F2,$7C,$44,$AF,$14,$79,$FE,$23,$D1,$F0,$72,$7C,$D1
	.dc.b	$C9,$F4,$5D,$7F,$DD,$DB,$3E,$13,$B4,$7B,$71,$EC,$EB,$1E,$CF,$C1
	.dc.b	$C7,$B3,$CB,$1F,$91,$F0,$5F,$27,$09,$3F,$1E,$3A,$BE,$0C,$3A,$F0
	.dc.b	$7F,$16,$C9,$97,$6A,$F2,$E5,$DA,$B5,$97,$D5,$F8,$29,$5F,$4D,$FB
	.dc.b	$1F,$05,$F6,$7E,$0A,$4F,$D7,$7E,$E7,$C2,$7D,$DF,$84,$FB,$DF,$09
	.dc.b	$2B,$EE,$47,$27,$EF,$5D,$7E,$3F,$C2,$74,$7E,$4E,$17,$3F,$56,$3D
	.dc.b	$67,$F0,$D8,$F5,$7E,$7E,$3F,$8D,$F8,$3F,$C7,$7C,$1C,$9F,$C6,$47
	.dc.b	$E4,$7E,$0E,$47,$E3,$DF,$CA,$7C,$1F,$E5,$7E,$13,$FC,$D7,$F8,$F7
	.dc.b	$F7,$E3,$FB,$FF,$22,$33,$F0,$BD,$F0,$EA,$7F,$A2,$36,$FE,$8B,$7F
	.dc.b	$E3,$8F,$65,$83,$2F,$0D,$FE,$19,$F0,$BF,$A1,$C7,$C3,$F2,$52,$30
	.dc.b	$75,$F8,$3F,$F4,$7E,$3F,$DF,$8F,$F0,$54,$64,$62,$8C,$C3,$2F,$2F
	.dc.b	$C2,$FB,$BE,$0A,$4F,$1B,$E1,$1F,$0B,$E0,$FE,$1F,$C3,$4A,$D6,$28
	.dc.b	$F8,$FF,$0D,$F2,$3D,$CF,$93,$94,$BD,$C7,$67,$C5,$A7,$AB,$2C,$69
	.dc.b	$F2,$33,$ED,$5A,$3D,$79,$F7,$7C,$39,$F8,$1F,$1F,$C2,$78,$5E,$E6
	.dc.b	$0F,$3E,$1D,$B1,$F0,$47,$AB,$0E,$F8,$FD,$8C,$2D,$7C,$ED,$3F,$07
	.dc.b	$F0,$7F,$84,$F8,$49,$8F,$C2,$8F,$C3,$FC,$24,$8F,$8A,$FE,$2B,$E1
	.dc.b	$3F,$17,$F0,$B2,$7E,$4B,$F8,$EF,$83,$FC,$7F,$C2,$F6,$9E,$87,$CE
	.dc.b	$7A,$3F,$0F,$93,$EE,$61,$8D,$BF,$1F,$8F,$5A,$D7,$13,$2C,$57,$E5
	.dc.b	$32,$FC,$EF,$C2,$CB,$FB,$6F,$E8,$3E,$17,$F4,$3F,$09,$FA,$2F,$85
	.dc.b	$95,$F8,$08,$FF,$67,$F9,$9F,$E0,$FE,$3F,$F7,$0A,$F8,$5C,$9A,$AE
	.dc.b	$87,$E2,$30,$AF,$AE,$5F,$0B,$E0,$8F,$C6,$E3,$E0,$7D,$78,$F3,$7F
	.dc.b	$2D,$5D,$DB,$86,$5E,$9F,$0B,$E6,$F8,$19,$1C,$23,$D3,$F0,$92,$39
	.dc.b	$BE,$CF,$85,$F6,$FC,$27,$BB,$E1,$65,$78,$23,$E0,$FC,$27,$C2,$C2
	.dc.b	$4F,$92,$3A,$3E,$DC,$2F,$1F,$F9,$23,$46,$5D,$AB,$D7,$97,$67,$E2
	.dc.b	$65,$F3,$BE,$13,$BC,$75,$3E,$0B,$D3,$8B,$E8,$E5,$67,$F8,$0E,$3D
	.dc.b	$9F,$41,$D8,$87,$63,$E5,$FC,$17,$DC,$23,$F3,$BE,$0F,$EF,$FC,$14
	.dc.b	$8F,$A2,$FE,$0B,$E0,$FF,$07,$F0,$92,$7E,$AB,$F8,$6F,$84,$FC,$3D
	.dc.b	$C9,$E8,$7D,$AF,$84,$E8,$FE,$0F,$0B,$C3,$F1,$F1,$E9,$3F,$87,$C6
	.dc.b	$CF,$CB,$C7,$F2,$7F,$09,$F9,$4F,$84,$97,$F8,$68,$FC,$BF,$C1,$C8
	.dc.b	$FC,$4B,$F9,$AF,$84,$47,$F2,$1F,$C7,$FF,$C2,$3C,$58,$F1,$51,$89
	.dc.b	$FE,$5A,$D8,$CF,$C2,$F2,$9F,$DB,$4B,$E1,$7C,$B1,$E3,$3B,$3E,$37
	.dc.b	$F4,$58,$DD,$7F,$0D,$97,$37,$FD,$1F,$7D,$C9,$96,$4F,$32,$C7,$18
	.dc.b	$79,$BE,$0E,$E7,$EA,$F8,$4B,$66,$CE,$3E,$0F,$3F,$09,$30,$F8,$91
	.dc.b	$C3,$E1,$78,$F4,$C9,$F9,$11,$D3,$E1,$79,$C7,$CB,$C3,$0F,$23,$E7
	.dc.b	$61,$D5,$F8,$F8,$75,$7E,$96,$1D,$5F,$97,$87,$27,$E6,$E1,$76,$F8
	.dc.b	$BF,$0B,$E9,$9E,$A4,$3A,$9E,$63,$A9,$0E,$A7,$E0,$7E,$13,$E2,$91
	.dc.b	$FC,$27,$C2,$7C,$8F,$83,$F9,$3F,$09,$27,$F1,$11,$F3,$3E,$17,$E6
	.dc.b	$FC,$24,$9F,$C6,$47,$D0,$BE,$91,$F8,$6C,$30,$F2,$3F,$25,$87,$57
	.dc.b	$F1,$78,$75,$7F,$2D,$87,$57,$F1,$F8,$7D,$CF,$85,$91,$F9,$B7,$EF
	.dc.b	$FC,$17,$E0,$3E,$17,$F0,$3F,$07,$2B,$F4,$11,$F8,$6F,$C8,$1F,$09
	.dc.b	$49,$3F,$E0,$58,$FF,$0E,$74,$7F,$86,$E3,$B1,$0F,$D0,$E1,$DD,$FD
	.dc.b	$66,$1D,$DF,$D2,$E1,$D1,$FE,$B9,$7E,$07,$B6,$1E,$3F,$86,$F2,$7C
	.dc.b	$2F,$97,$E2,$25,$78,$E3,$D1,$F0,$9E,$9F,$84,$93,$E7,$8F,$67,$C2
	.dc.b	$48,$F5,$3E,$FF,$84,$EB,$1C,$C8,$79,$3E,$1B,$AB,$F0,$6F,$AB,$E8
	.dc.b	$BE,$8F,$AA,$EF,$4B,$C3,$FE,$15,$B3,$2F,$9D,$94,$C3,$E6,$3F,$47
	.dc.b	$E1,$BB,$C7,$C4,$CF,$BE,$3C,$E3,$FC,$7B,$83,$3E,$F9,$76,$CE,$D6
	.dc.b	$B3,$FB,$9F,$0B,$F7,$7E,$1A,$4F,$DB,$8F,$C0,$7C,$2F,$E0,$7E,$16
	.dc.b	$4F,$DF,$8F,$C2,$7C,$27,$E1,$7E,$12,$4F,$E0,$E3,$F1,$1F,$0D,$DA
	.dc.b	$3A,$1F,$87,$7A,$BF,$89,$BE,$AF,$E0,$2F,$AB,$F8,$DB,$EA,$FE,$12
	.dc.b	$FF,$29,$F0,$BF,$95,$F8,$59,$3F,$93,$8F,$CC,$7C,$1C,$8F,$CB,$3F
	.dc.b	$9B,$F8,$4F,$E4,$3F,$13,$FE,$17,$9F,$F3,$2B,$7C,$F3,$4F,$F0,$64
	.dc.b	$7E,$52,$FC,$2F,$73,$FC,$CD,$1F,$E1,$87,$B9,$0F,$CD,$DF,$85,$D9
	.dc.b	$E2,$21,$CB,$E1,$7F,$9B,$BF,$EB,$6D,$AF,$9E,$69,$8C,$24,$FF,$9F
	.dc.b	$C7,$9A,$D3,$9B,$63,$56,$AE,$31,$FE,$6D,$38,$E3,$C0,$CF,$47,$FB
	.dc.b	$DA,$B8,$4F,$2A,$CF,$37,$F5,$F9,$CC,$F3,$6D,$71,$AD,$CF,$F6,$E9
	.dc.b	$CD,$71,$9C,$CF,$FB,$59,$B3,$E9,$8E,$91,$8C,$72,$F0,$C7,$E3,$4F
	.dc.b	$94,$87,$03,$E3,$3E,$07,$F0,$3F,$09,$F8,$42,$3F,$73,$E1,$7F,$0F
	.dc.b	$F0,$12,$3E,$FB,$F8,$AF,$84,$FC,$5F,$C1,$C9,$FC,$1B,$F8,$EF,$84
	.dc.b	$FC,$7D,$F4,$9F,$F3,$C7,$13,$1C,$C7,$42,$1A,$C7,$B3,$D0,$87,$2C
	.dc.b	$3B,$3F,$92,$F8,$5F,$CE,$91,$FC,$AF,$C2,$FE,$83,$E1,$64,$7E,$61
	.dc.b	$FD,$1F,$C2,$FE,$93,$E1,$3F,$4B,$F0,$92,$BF,$3B,$1F,$A8,$B9,$1D
	.dc.b	$0F,$D0,$FC,$27,$47,$DB,$85,$E6,$FE,$2D,$D5,$FA,$CC,$BB,$57,$EA
	.dc.b	$32,$EC,$FC,$4C,$BF,$6B,$F0,$9F,$B6,$F8,$59,$3F,$AE,$8F,$DC,$FC
	.dc.b	$27,$EE,$BE,$16,$4F,$EC,$A3,$F7,$BF,$09,$FB,$EF,$84,$93,$FB,$68
	.dc.b	$EA,$F3,$21,$FE,$AB,$E1,$FF,$1D,$72,$AF,$AC,$FD,$8C,$3C,$2F,$33
	.dc.b	$F8,$0E,$1D,$9F,$BB,$87,$F8,$BB,$0E,$78,$F8,$27,$B7,$6F,$27,$A1
	.dc.b	$FC,$2B,$0C,$70,$7C,$8C,$D8,$F5,$8C,$65,$EE,$BF,$F7,$B6,$5F,$E3
	.dc.b	$F6,$FC,$27,$84,$F1,$7C,$2F,$8F,$BF,$C9,$F0,$B2,$B8,$C7,$9F,$E1
	.dc.b	$3D,$1F,$0B,$27,$AC,$72,$7A,$90,$F9,$17,$D5,$F1,$DF,$57,$E6,$5F
	.dc.b	$57,$CF,$7D,$5F,$A1,$7F,$1F,$E1,$7E,$47,$83,$27,$DB,$1F,$2F,$E1
	.dc.b	$64,$7C,$07,$E7,$7C,$2F,$CF,$F8,$4F,$A1,$F0,$B2,$BE,$2C,$75,$79
	.dc.b	$90,$FC,$07,$C2,$74,$7F,$05,$77,$8B,$F3,$3C,$2C,$7B,$57,$D0,$C7
	.dc.b	$B3,$F8,$9C,$7F,$01,$F0,$B2,$3E,$AB,$F8,$3F,$84,$FC,$27,$C2,$C9
	.dc.b	$FB,$2F,$E1,$FE,$17,$F1,$1F,$09,$27,$EE,$BF,$8B,$F8,$3E,$51,$D0
	.dc.b	$87,$E6,$6F,$A3,$F9,$BB,$BA,$FA,$B0,$E9,$3F,$9F,$C2,$CF,$BB,$0E
	.dc.b	$B3,$CC,$FC,$6E,$1F,$9B,$F8,$4E,$7D,$BF,$EA,$3D,$D9,$E2,$DE,$22
	.dc.b	$5A,$76,$BE,$1A,$5B,$DD,$33,$FF,$4A,$FC,$DF,$F9,$06,$5A,$FF,$0A
	.dc.b	$8C,$37,$E1,$77,$F0,$BC,$3E,$13,$8F,$C2,$CC,$3F,$D7,$77,$C6,$7A
	.dc.b	$FC,$0C,$9C,$EB,$FE,$BD,$CB,$FC,$55,$18,$3D,$6B,$E0,$FB,$C6,$78
	.dc.b	$13,$E4,$AE,$5A,$6A,$BD,$31,$D7,$1F,$24,$60,$EB,$9F,$A5,$E9,$7E
	.dc.b	$27,$58,$76,$CB,$D1,$18,$9F,$23,$C3,$E0,$BE,$4F,$C1,$48,$C1,$D3
	.dc.b	$E1,$7E,$69,$1E,$FF,$0B,$F4,$3E,$0F,$E8,$FC,$24,$9F,$1C,$7D,$4C
	.dc.b	$24,$74,$3C,$5F,$07,$CD,$F2,$61,$78,$BE,$C6,$3D,$67,$DB,$8F,$57
	.dc.b	$D5,$8F,$E0,$3E,$12,$47,$C2,$7F,$07,$F0,$7F,$84,$F8,$49,$3F,$19
	.dc.b	$FC,$3F,$C1,$7E,$23,$E0,$3F,$13,$F0,$32,$BE,$5C,$72,$7A,$10,$F9
	.dc.b	$57,$D1,$F9,$97,$74,$FC,$86,$1D,$67,$EA,$E1,$D5,$FA,$58,$75,$79
	.dc.b	$9F,$6B,$0F,$CD,$FC,$17,$3E,$7F,$C7,$65,$6B,$F1,$B8,$F5,$7E,$E6
	.dc.b	$3D,$72,$FD,$06,$5F,$A4,$F8,$49,$5F,$86,$7F,$DF,$7F,$8A,$7F,$96
	.dc.b	$4F,$F2,$BD,$7F,$70,$7F,$CA,$93,$F8,$A6,$C6,$7E,$17,$8F,$C1,$4A
	.dc.b	$FC,$E5,$7F,$85,$5B,$F0,$7D,$9F,$09,$F9,$1C,$30,$F0,$BF,$1F,$C4
	.dc.b	$48,$C1,$D3,$0E,$CF,$8D,$F4,$DF,$AB,$E1,$3D,$7F,$0D,$2B,$C7,$1E
	.dc.b	$EF,$85,$F7,$FC,$34,$9F,$3C,$7C,$2F,$1A,$47,$A9,$F8,$BF,$0D,$D6
	.dc.b	$39,$90,$F8,$9F,$0B,$D5,$F8,$37,$D5,$F9,$17,$D1,$F9,$57,$78,$3E
	.dc.b	$76,$1F,$4B,$E1,$7E,$9F,$C2,$4B,$F9,$71,$F5,$BE,$16,$47,$CE,$7E
	.dc.b	$CF,$C3,$7D,$AF,$84,$93,$F4,$9F,$BB,$F0,$DD,$63,$99,$0F,$B9,$F0
	.dc.b	$DD,$1F,$BD,$77,$3F,$C1,$61,$D2,$7F,$05,$85,$9F,$C3,$61,$F8,$AC
	.dc.b	$3B,$57,$E1,$F1,$C3,$E7,$21,$D8,$FC,$3E,$1C,$FB,$11,$FD,$0B,$A1
	.dc.b	$96,$63,$18,$FE,$63,$E1,$BF,$33,$F0,$B2,$7F,$23,$1F,$9C,$F8,$5E
	.dc.b	$71,$FC,$8B,$F1,$3F,$F7,$03,$72,$D3,$FC,$CE,$3F,$C4,$4E,$FE,$17
	.dc.b	$8E,$FF,$E5,$29,$46,$2B,$CF,$E7,$27,$8D,$7E,$A7,$1F,$36,$1E,$43
	.dc.b	$B7,$C2,$78,$1F,$11,$E1,$C3,$15,$DF,$E1,$BB,$4F,$F6,$4C,$3C,$2F
	.dc.b	$9F,$2C,$E5,$E1,$AB,$4F,$86,$A4,$BF,$77,$C2,$EB,$E1,$B7,$F0,$B2
	.dc.b	$7E,$14,$72,$F8,$69,$1F,$15,$EB,$F0,$BD,$AF,$A4,$FC,$DC,$30,$F2
	.dc.b	$3E,$66,$1D,$5F,$A5,$87,$47,$EA,$61,$75,$F9,$D8,$FA,$BE,$16,$57
	.dc.b	$D5,$7D,$BF,$09,$EE,$F8,$5F,$7F,$C3,$4A,$FB,$51,$F0,$BE,$1A,$47
	.dc.b	$DD,$7E,$2F,$C3,$72,$8E,$84,$3F,$0D,$7D,$5F,$C2,$5F,$57,$F1,$57
	.dc.b	$D1,$FC,$65,$DC,$BF,$0D,$87,$59,$E6,$7E,$37,$0F,$AB,$F0,$BD,$63
	.dc.b	$F2,$98,$F5,$BF,$C7,$63,$CF,$97,$F1,$BC,$4C,$F1,$87,$E0,$F3,$FB
	.dc.b	$FF,$0B,$F8,$0F,$85,$93,$F9,$B8,$FC,$1F,$C2,$C8,$FC,$F3,$F8,$6F
	.dc.b	$86,$FC,$4F,$B7,$FE,$11,$9F,$F1,$8A,$F8,$6C,$9A,$F8,$5E,$F3,$D4
	.dc.b	$87,$F8,$3A,$5F,$0B,$DE,$3F,$61,$85,$E0,$FD,$56,$3C,$E7,$FC,$45
	.dc.b	$7D,$DB,$8E,$5E,$9F,$0B,$E6,$F8,$59,$1C,$23,$D3,$F0,$DE,$AF,$84
	.dc.b	$93,$D2,$3D,$BF,$0B,$23,$BB,$F0,$3E,$13,$E0,$DC,$9E,$87,$93,$2E
	.dc.b	$AF,$9B,$0E,$AF,$BF,$0E,$8F,$C1,$C2,$E9,$FE,$D8,$B6,$3F,$3B,$E0
	.dc.b	$FE,$7F,$C2,$4C,$3E,$0C,$7D,$2F,$85,$FA,$7F,$0B,$27,$E2,$C7,$D6
	.dc.b	$F8,$5F,$AF,$F0,$92,$7E,$4C,$75,$79,$90,$FA,$9F,$09,$D5,$F9,$F7
	.dc.b	$D5,$FB,$17,$D5,$FA,$77,$D5,$FB,$97,$F8,$6B,$ED,$1F,$7F,$0C,$3E
	.dc.b	$63,$A9,$0E,$A7,$98,$EA,$43,$A9,$F8,$2F,$83,$FC,$9F,$C2,$48,$C1
	.dc.b	$F8,$6F,$85,$FC,$BF,$C2,$C8,$C1,$FE,$6B,$F8,$EF,$E7,$EF,$F9,$81
	.dc.b	$FF,$0D,$F8,$3F,$CD,$9B,$1F,$E8,$E7,$7C,$5F,$5F,$1F,$D1,$92,$7A
	.dc.b	$7C,$2F,$E9,$B1,$F0,$47,$E7,$F1,$C3,$E1,$3F,$99,$E1,$DB,$CF,$34
	.dc.b	$EF,$97,$8F,$E1,$76,$4A,$F5,$7C,$2F,$2F,$84,$91,$ED,$7A,$FC,$2F
	.dc.b	$6F,$83,$93,$F0,$5F,$0F,$C2,$78,$AF,$A4,$FC,$3C,$30,$F3,$21,$F1
	.dc.b	$F0,$EA,$F3,$21,$F2,$EF,$AB,$F3,$FE,$13,$DF,$F0,$92,$30,$7D,$3F
	.dc.b	$85,$F8,$64,$7E,$B7,$C1,$7C,$6F,$85,$F8,$FF,$0B,$27,$ED,$47,$CA
	.dc.b	$C2,$47,$43,$EE,$7C,$27,$37,$EF,$61,$79,$DC,$3F,$7F,$2E,$D5,$F8
	.dc.b	$5C,$BB,$3F,$85,$CB,$EB,$E3,$DA,$3F,$11,$96,$1F,$51,$0E,$E7,$A8
	.dc.b	$F3,$9D,$48,$75,$3F,$29,$F0,$7F,$84,$F8,$39,$18,$3F,$31,$F0,$BF
	.dc.b	$88,$F8,$29,$18,$3F,$39,$F0,$9F,$8C,$F8,$49,$18,$39,$1F,$9F,$C3
	.dc.b	$F2,$27,$F8,$13,$F2,$1F,$F0,$FF,$07,$C4,$9F,$D0,$63,$D9,$FD,$36
	.dc.b	$3F,$98,$8F,$CC,$E3,$DA,$7F,$55,$8E,$1F,$F6,$C6,$1F,$E1,$EE,$0F
	.dc.b	$F1,$0B,$FE,$14,$8F,$F6,$7B,$71,$F0,$75,$23,$84,$FF,$BF,$97,$C2
	.dc.b	$F9,$C9,$1D,$BE,$0B,$D5,$F0,$72,$3C,$2F,$B7,$E0,$BD,$DF,$09,$27
	.dc.b	$CA,$FC,$1F,$84,$E7,$1E,$4C,$2E,$7E,$4C,$BF,$C4,$D3,$D0,$87,$AB
	.dc.b	$1E,$CF,$C0,$F8,$4F,$98,$47,$E1,$FC,$27,$CF,$F8,$4F,$A1,$F0,$92
	.dc.b	$7E,$3C,$7D,$3F,$82,$91,$F2,$9F,$AD,$F0,$7F,$5E,$FA,$4F,$CA,$C3
	.dc.b	$0F,$23,$E8,$E1,$D5,$F9,$F8,$74,$7E,$8E,$17,$3F,$76,$3C,$A7,$EA
	.dc.b	$E3,$67,$E1,$7C,$17,$A6,$7A,$90,$EA,$79,$C8,$76,$3F,$07,$5F,$8F
	.dc.b	$F8,$2F,$C8,$7C,$24,$9F,$C3,$C7,$E4,$FE,$0A,$47,$E2,$9F,$CB,$7C
	.dc.b	$1F,$E5,$FE,$07,$F3,$1F,$03,$2B,$F1,$F1,$D5,$E6,$41,$9F,$E3,$BF
	.dc.b	$22,$FF,$0D,$74,$FE,$3F,$1E,$F5,$F9,$47,$BB,$D0,$FC,$DE,$1D,$DF
	.dc.b	$CC,$E1,$FE,$6C,$BF,$F0,$A4,$76,$EA,$FF,$3F,$C7,$F8,$7A,$3F,$D2
	.dc.b	$F8,$F8,$7E,$0F,$87,$C1,$F1,$F8,$2E,$5F,$07,$2F,$78,$F2,$9E,$DF
	.dc.b	$05,$27,$23,$66,$69,$F2,$7C,$1F,$6A,$DD,$6B,$2C,$47,$9C,$E5,$6F
	.dc.b	$83,$F2,$47,$08,$67,$E5,$8E,$17,$E0,$CB,$C9,$18,$AF,$13,$C7,$2F
	.dc.b	$1B,$CF,$E0,$BE,$41,$1E,$DF,$07,$F2,$FD,$DF,$99,$F0,$72,$7C,$51
	.dc.b	$F3,$FE,$0B,$E8,$7C,$1C,$9F,$34,$7D,$3B,$E9,$1E,$3C,$30,$F2,$3D
	.dc.b	$98,$75,$7D,$18,$75,$7E,$06,$1D,$5F,$66,$1C,$9F,$76,$17,$8F,$F1
	.dc.b	$9F,$05,$E9,$9E,$A4,$3A,$9E,$63,$A9,$0E,$A7,$CF,$F8,$3F,$C6,$7C
	.dc.b	$1C,$8C,$1F,$4F,$E0,$BF,$21,$F0,$72,$30,$7D,$7F,$82,$FC,$A1,$1F
	.dc.b	$B5,$F0,$5F,$97,$C2,$47,$43,$ED,$F8,$7F,$E1,$07,$F0,$78,$75,$7E
	.dc.b	$E6,$1F,$9F,$8E,$91,$F7,$F0,$B9,$FE,$B7,$1F,$F6,$7E,$1F,$E1,$D1
	.dc.b	$F0,$78,$33,$F1,$1A,$F8,$09,$8F,$E5,$A9,$1C,$BE,$07,$9F,$C0,$CA
	.dc.b	$D3,$DB,$E0,$F9,$C7,$1C,$2B,$EC,$F8,$2E,$AF,$BB,$0B,$A7,$F5,$0C
	.dc.b	$99,$75,$AF,$87,$95,$AD,$65,$EC,$F8,$3F,$6F,$BB,$2F,$D1,$1F,$03
	.dc.b	$E0,$E4,$7A,$DF,$87,$F0,$7F,$13,$E0,$FE,$2F,$C1,$4A,$F8,$11,$D5
	.dc.b	$E6,$43,$EA,$FC,$1F,$47,$EB,$DD,$CF,$D5,$87,$59,$F9,$58,$75,$7E
	.dc.b	$EE,$1F,$53,$0E,$D1,$F8,$0C,$70,$F9,$C8,$76,$3E,$A5,$F5,$7F,$0F
	.dc.b	$7D,$70,$F9,$58,$7E,$03,$E0,$E5,$7D,$B7,$F0,$7F,$05,$F8,$4F,$83
	.dc.b	$93,$F7,$DF,$C3,$FC,$17,$E2,$3E,$0E,$4F,$E0,$DF,$C5,$FC,$1F,$28
	.dc.b	$E8,$43,$F3,$37,$D1,$FC,$DD,$DC,$BE,$F6,$1D,$27,$F3,$F8,$59,$FC
	.dc.b	$16,$1F,$98,$F8,$39,$5F,$93,$4F,$F2,$AF,$C8,$FF,$C4,$FE,$86,$3C
	.dc.b	$C4,$7F,$86,$1B,$F1,$1C,$2B,$F9,$3C,$FF,$87,$EB,$83,$B3,$FC,$3B
	.dc.b	$1D,$7E,$07,$86,$5F,$E2,$49,$CC,$F2,$F8,$5C,$DF,$23,$CF,$1F,$BA
	.dc.b	$CB,$2F,$F4,$5A,$F3,$B7,$E1,$33,$9F,$9A,$3D,$99,$E2,$BC,$99,$D2
	.dc.b	$4F,$C1,$CB,$5F,$05,$23,$E2,$3C,$7E,$17,$97,$C2,$F3,$F8,$49,$5F
	.dc.b	$26,$3B,$78,$1D,$FE,$0E,$4F,$CD,$8E,$4F,$42,$1F,$2A,$FA,$BF,$4E
	.dc.b	$FA,$BF,$3A,$FA,$BF,$5E,$FA,$BF,$4A,$FA,$BC,$CF,$B9,$7F,$07,$E1
	.dc.b	$3A,$C7,$D8,$C3,$AF,$17,$D3,$C7,$AC,$FD,$CC,$7A,$CF,$D7,$C7,$E5
	.dc.b	$7C,$24,$AF,$C4,$3F,$37,$E0,$FE,$77,$C2,$49,$FC,$63,$F4,$7E,$0F
	.dc.b	$E9,$7C,$27,$D3,$F8,$39,$5F,$91,$8E,$4F,$42,$1F,$8E,$BE,$8F,$E4
	.dc.b	$2E,$E5,$F8,$BC,$3A,$CF,$E6,$F0,$EA,$FE,$5B,$0F,$C1,$7C,$27,$E0
	.dc.b	$FE,$0E,$4F,$E8,$63,$F0,$DF,$09,$F8,$AF,$C9,$3F,$C5,$67,$F8,$B3
	.dc.b	$9D,$19,$F8,$39,$7F,$E1,$2A,$FF,$13,$E9,$FE,$80,$7F,$C4,$D1,$FA
	.dc.b	$6C,$3C,$2F,$73,$87,$C0,$F8,$1E,$5F,$11,$E1,$30,$FF,$85,$2B,$AF
	.dc.b	$7F,$ED,$5B,$1F,$E2,$76,$F6,$9F,$0E,$3E,$3F,$83,$F6,$12,$BD,$1F
	.dc.b	$05,$EF,$F8,$5F,$81,$F0,$B2,$7D,$91,$F0,$FE,$13,$E2,$7C,$04,$9F
	.dc.b	$81,$1F,$1E,$FA,$47,$9F,$0C,$3C,$8F,$8D,$87,$57,$D9,$87,$57,$E5
	.dc.b	$61,$D5,$F8,$18,$72,$7E,$16,$17,$1F,$9D,$F0,$7E,$99,$EA,$43,$A9
	.dc.b	$E6,$3A,$90,$EA,$7D,$9F,$84,$FC,$01,$1F,$B9,$F0,$7F,$83,$F8,$49
	.dc.b	$1F,$7D,$FC,$37,$C2,$7E,$1F,$E0,$FF,$11,$F0,$92,$BF,$09,$1F,$8B
	.dc.b	$BE,$91,$F6,$F0,$C3,$CC,$87,$DF,$C3,$AB,$C8,$FC,$75,$F5,$7F,$09
	.dc.b	$7F,$97,$F8,$39,$1F,$93,$7F,$35,$F0,$9F,$9B,$F8,$49,$3F,$97,$7F
	.dc.b	$3D,$F0,$9F,$CA,$3F,$25,$FF,$13,$73,$97,$C2,$64,$D5,$F6,$3F,$29
	.dc.b	$8D,$E1,$FD,$06,$5D,$A7,$F3,$19,$59,$FD,$26,$5C,$A7,$FC,$F3,$8F
	.dc.b	$66,$E3,$97,$FC,$E7,$1F,$E2,$E7,$7F,$0B,$E7,$3D,$1F,$09,$E9,$F8
	.dc.b	$49,$3D,$1F,$67,$C2,$7B,$7E,$13,$DD,$F0,$B2,$BC,$31,$C9,$EA,$43
	.dc.b	$C3,$7D,$1F,$1D,$DD,$3F,$DF,$B9,$31,$ED,$5E,$CC,$7B,$3E,$AC,$7B
	.dc.b	$3E,$CC,$BD,$4F,$C2,$CB,$A7,$7F,$F7,$1C,$7F,$C7,$33,$EB,$3E,$3E
	.dc.b	$3D,$9F,$8B,$8F,$6C,$BC,$79,$7D,$BF,$82,$95,$F3,$9F,$BD,$F0,$BF
	.dc.b	$7F,$E1,$64,$FD,$27,$F0,$5F,$09,$F8,$3F,$83,$FC,$27,$C2,$4A,$FA
	.dc.b	$F1,$D1,$FA,$B8,$5C,$BE,$1E,$3D,$27,$EC,$E3,$67,$E3,$E3,$D6,$7F
	.dc.b	$03,$8F,$57,$EF,$E3,$F9,$3F,$85,$91,$F8,$67,$F2,$DF,$0B,$F9,$7F
	.dc.b	$85,$93,$F8,$A7,$F3,$5F,$0B,$F9,$BF,$85,$93,$F8,$E7,$F3,$DF,$0B
	.dc.b	$CA,$3A,$10,$FC,$7D,$F4,$7F,$23,$77,$4F,$BF,$87,$49,$FC,$AE,$16
	.dc.b	$7F,$07,$87,$EA,$FE,$17,$F5,$9F,$0B,$2F,$F3,$F1,$FE,$16,$DF,$FC
	.dc.b	$33,$DE,$7F,$C5,$C4,$E7,$D7,$C2,$6F,$E1,$26,$1F,$EE,$0A,$FF,$18
	.dc.b	$3F,$EE,$7E,$96,$3E,$18,$EE,$6F,$E1,$3C,$18,$F8,$7E,$16,$4F,$38
	.dc.b	$EE,$F5,$C7,$BB,$FE,$E4,$C7,$BE,$7E,$7C,$3B,$4F,$ED,$B1,$F6,$47
	.dc.b	$68,$F3,$63,$D9,$E5,$8F,$67,$A6,$3D,$B6,$4F,$F9,$47,$2E,$E3,$E1
	.dc.b	$7E,$47,$C2,$CC,$3E,$14,$7C,$BF,$83,$F9,$9F,$09,$27,$E3,$47,$CF
	.dc.b	$F8,$3F,$A1,$F0,$92,$7E,$54,$7D,$3B,$91,$D0,$F9,$DF,$0B,$D5,$FA
	.dc.b	$18,$75,$7E,$26,$1D,$5F,$A9,$87,$57,$E4,$61,$F7,$FE,$17,$F0,$1F
	.dc.b	$0B,$27,$ED,$47,$E0,$FE,$16,$47,$DD,$7F,$0D,$F0,$BF,$87,$F8,$49
	.dc.b	$3F,$81,$7F,$15,$F0,$BD,$63,$99,$0F,$B1,$F0,$3D,$1F,$B5,$77,$8F
	.dc.b	$E7,$C3,$A3,$F2,$58,$F6,$8F,$C0,$63,$6B,$75,$F9,$7C,$7F,$35,$F0
	.dc.b	$BF,$9B,$F8,$49,$87,$E5,$63,$FC,$E7,$F9,$33,$E0,$E9,$27,$FD,$23
	.dc.b	$8F,$49,$FF,$0C,$9F,$E1,$8E,$8F,$F4,$92,$F8,$4E,$67,$4F,$88,$F0
	.dc.b	$DF,$72,$0B,$E9,$F0,$9D,$E3,$18,$74,$C7,$15,$E1,$F8,$6E,$95,$E4
	.dc.b	$AF,$0E,$3E,$08,$C1,$F1,$F2,$F3,$37,$1F,$1B,$FA,$7C,$FE,$0C,$76
	.dc.b	$BF,$04,$F2,$D9,$7F,$16,$30,$FC,$2D,$3C,$11,$E8,$3C,$18,$F9,$9C
	.dc.b	$57,$2F,$81,$F9,$E4,$8E,$7F,$09,$F4,$BE,$17,$E9,$FC,$24,$9E,$F1
	.dc.b	$F5,$BE,$16,$47,$89,$FB,$3F,$0B,$F6,$AF,$A4,$F8,$34,$63,$98,$E6
	.dc.b	$7A,$B1,$EC,$EF,$1E,$AF,$1C,$6F,$0F,$E0,$72,$FC,$3F,$C1,$CA,$F8
	.dc.b	$6F,$E2,$BE,$13,$F1,$7F,$09,$27,$E3,$BF,$8E,$F8,$5F,$C7,$FC,$27
	.dc.b	$E4,$3E,$16,$57,$CC,$8E,$8F,$A7,$0B,$97,$E4,$71,$E9,$3E,$DC,$6C
	.dc.b	$FE,$57,$1E,$B3,$F5,$B1,$EA,$FC,$3C,$7F,$DA,$AB,$F1,$2F,$E6,$AD
	.dc.b	$F2,$F3,$06,$F9,$F2,$7F,$1A,$3E,$4D,$EC,$F4,$6C,$6A,$FA,$9D,$A1
	.dc.b	$EC,$6C,$E9,$3F,$89,$3C,$97,$8D,$3F,$80,$C6,$0E,$91,$9E,$AF,$F7
	.dc.b	$F6,$1D,$A3,$12,$38,$1D,$63,$4F,$23,$CD,$E7,$F0,$77,$7E,$09,$FF
	.dc.b	$80,$2D,$AD,$47,$AF,$DA,$B1,$C7,$27,$B5,$FC,$33,$E1,$FF,$86,$FC
	.dc.b	$3C,$CB,$F5,$31,$FC,$43,$E1,$E4,$7E,$B1,$FE,$27,$F0,$FE,$98,$FD
	.dc.b	$4F,$E4,$FF,$E6,$02,$39,$D7,$F8,$FC,$7F,$95,$63,$F4,$17,$E1,$7B
	.dc.b	$9F,$BB,$8F,$F2,$B3,$FA,$5B,$F0,$B4,$FD,$ED,$F5,$DF,$E5,$78,$0B
	.dc.b	$F0,$D7,$33,$47,$F3,$57,$C5,$8F,$93,$E1,$73,$5C,$5B,$26,$FC,$3E
	.dc.b	$3E,$1A,$61,$E7,$77,$F0,$BC,$3E,$1A,$4F,$AD,$E7,$F0,$DC,$A3,$DD
	.dc.b	$87,$63,$0F,$CA,$BE,$AD,$3E,$25,$F4,$7E,$7D,$DC,$FE,$6E,$1E,$8F
	.dc.b	$86,$95,$F2,$DF,$5F,$BB,$EC,$E6,$93,$F3,$DF,$7F,$27,$C0,$EA,$93
	.dc.b	$F4,$DF,$87,$D5,$CA,$3E,$B5,$D7,$F0,$3F,$0D,$D1,$FC,$1E,$17,$3F
	.dc.b	$C0,$E3,$D2,$7F,$0F,$8D,$9F,$C2,$E3,$F4,$30,$E9,$1F,$88,$C7,$17
	.dc.b	$E7,$21,$D8,$FC,$46,$1C,$F8,$FF,$07,$C5,$FB,$E9,$EE,$7E,$37,$1F
	.dc.b	$BB,$F0,$D2,$3F,$20,$FE,$03,$E1,$BF,$03,$F0,$D2,$7F,$28,$FE,$13
	.dc.b	$E1,$BF,$0B,$F0,$9F,$86,$F8,$69,$5F,$99,$8E,$4F,$E6,$EE,$BF,$D3
	.dc.b	$BF,$28,$FF,$2B,$E6,$C7,$B4,$FE,$A9,$EC,$F3,$3F,$49,$87,$67,$F5
	.dc.b	$F8,$7E,$62,$FF,$91,$47,$F6,$36,$BD,$FC,$3D,$78,$3B,$F0,$E4,$ED
	.dc.b	$E5,$F0,$DC,$FE,$17,$A7,$C3,$4A,$E9,$1C,$9E,$D7,$5F,$67,$C2,$74
	.dc.b	$7D,$D8,$5E,$6F,$F0,$FE,$6C,$BB,$57,$9F,$2E,$CF,$C5,$CB,$D9,$F0
	.dc.b	$BE,$DF,$84,$93,$ED,$8F,$81,$F0,$BF,$07,$E1,$A4,$FC,$18,$F8,$9F
	.dc.b	$0B,$F1,$7E,$1A,$4F,$C5,$8E,$AF,$32,$1F,$57,$E1,$7A,$BF,$2E,$FA
	.dc.b	$BF,$66,$FA,$BF,$3E,$FA,$BF,$76,$FE,$A5,$F6,$8F,$C0,$61,$87,$CC
	.dc.b	$75,$21,$D4,$F3,$1D,$48,$75,$3E,$F7,$C2,$FE,$04,$8F,$E0,$7E,$17
	.dc.b	$F0,$9F,$0B,$23,$F0,$8F,$E1,$FE,$1B,$F1,$1F,$07,$27,$F1,$0F,$E2
	.dc.b	$FE,$1B,$F1,$97,$27,$A1,$FE,$2F,$F6,$FF,$C5,$CF,$E6,$F0,$EC,$D3
	.dc.b	$F2,$58,$75,$7F,$41,$85,$E5,$FC,$1E,$3F,$CC,$30,$7C,$2D,$25,$F7
	.dc.b	$F8,$4D,$7C,$26,$FE,$16,$4F,$8E,$39,$7C,$1F,$3F,$85,$93,$E7,$8E
	.dc.b	$D7,$D2,$3F,$C7,$F8,$F8,$1E,$C7,$33,$DD,$87,$67,$78,$76,$7E,$16
	.dc.b	$1D,$9E,$78,$7A,$FE,$0E,$47,$C7,$7D,$DF,$09,$EF,$F8,$49,$3F,$2D
	.dc.b	$F8,$59,$7C,$3F,$84,$93,$F3,$DF,$8D,$F0,$9C,$A3,$E9,$5D,$7D,$7F
	.dc.b	$09,$D1,$F6,$E1,$78,$7F,$D1,$F8,$8E,$B3,$F0,$B1,$B3,$AC,$7E,$A6
	.dc.b	$1D,$63,$8E,$38,$BF,$41,$0E,$E7,$E1,$70,$E7,$E7,$7F,$74,$58,$E2
	.dc.b	$7F,$D4,$D0,$C7,$F0,$5F,$09,$F8,$3F,$83,$95,$F8,$F8,$FC,$37,$C2
	.dc.b	$48,$FC,$93,$F8,$9F,$81,$FC,$57,$C1,$7E,$2F,$E0,$A5,$7E,$5E,$3A
	.dc.b	$BC,$C8,$7E,$0B,$E0,$BA,$3F,$84,$BB,$BF,$C8,$C3,$AC,$FE,$87,$0E
	.dc.b	$AF,$E2,$B0,$FF,$04,$FE,$53,$FE,$23,$8F,$F4,$5B,$6F,$FD,$7E,$DD
	.dc.b	$FF,$C2,$EF,$F8,$49,$BF,$13,$C8,$95,$FE,$9D,$9F,$F0,$93,$7E,$0F
	.dc.b	$BF,$C1,$48,$C1,$BF,$8A,$F1,$91,$E5,$F0,$7D,$1F,$2E,$17,$2F,$2F
	.dc.b	$C2,$7A,$B1,$E9,$5E,$AC,$70,$F8,$FE,$13,$DE,$47,$CD,$F0,$9F,$0B
	.dc.b	$E1,$24,$7A,$5F,$8B,$F0,$3F,$1B,$E1,$24,$FB,$5F,$93,$F0,$5F,$2A
	.dc.b	$FA,$4F,$CA,$C3,$0F,$32,$1F,$3B,$0E,$AF,$32,$1F,$4A,$FA,$BE,$62
	.dc.b	$1F,$5A,$BE,$CF,$C2,$7A,$23,$A1,$DB,$5F,$B3,$97,$A2,$3A,$1D,$A3
	.dc.b	$EE,$E5,$F8,$2F,$83,$FC,$1F,$C2,$4A,$FB,$51,$F8,$6F,$84,$91,$F7
	.dc.b	$5F,$C4,$FC,$1F,$E2,$BE,$0B,$F1,$7F,$09,$2B,$F0,$51,$D5,$E6,$43
	.dc.b	$F1,$FF,$05,$D1,$FC,$8D,$DC,$FF,$23,$87,$59,$FC,$66,$1D,$5F,$CB
	.dc.b	$E1,$F9,$AF,$83,$91,$F9,$17,$FC,$CB,$F9,$57,$FA,$7B,$07,$5F,$F9
	.dc.b	$C2,$3F,$C1,$F1,$8A,$7C,$32,$5F,$09,$C8,$E7,$F0,$BD,$3E,$12,$5F
	.dc.b	$FB,$EE,$BF,$D3,$78,$F5,$F8,$3F,$0B,$E4,$9F,$D6,$E7,$87,$86,$1D
	.dc.b	$5E,$58,$F6,$7A,$63,$D9,$FF,$7F,$E3,$D9,$F3,$BF,$E1,$57,$C5,$F0
	.dc.b	$5F,$00,$8F,$97,$E0,$FE,$1F,$C2,$C8,$F4,$3F,$1B,$DE,$F8,$FF,$09
	.dc.b	$27,$D8,$FC,$AF,$83,$F9,$77,$D2,$7C,$78,$61,$E6,$43,$CF,$87,$57
	.dc.b	$99,$0F,$5D,$F5,$7E,$57,$C2,$73,$7D,$F7,$F6,$A3,$D1,$1D,$0E,$DC
	.dc.b	$5F,$67,$8B,$EC,$67,$E9,$9E,$A7,$79,$D6,$8D,$3F,$0B,$F0,$7F,$86
	.dc.b	$F8,$49,$7F,$6E,$3F,$13,$F0,$92,$3E,$F3,$F8,$CF,$84,$FC,$6F,$C2
	.dc.b	$7E,$3B,$E1,$25,$7E,$0E,$3A,$BC,$C8,$7D,$AF,$84,$E8,$FD,$CB,$BA
	.dc.b	$79,$B0,$EB,$3F,$8D,$C3,$AB,$F8,$2C,$3F,$3B,$F0,$9F,$9E,$F8,$49
	.dc.b	$3F,$93,$8F,$F6,$2F,$E5,$7F,$E2,$0C,$7E,$33,$7A,$FF,$DA,$2D,$AF
	.dc.b	$D0,$BF,$E0,$D5,$3C,$3E,$13,$99,$20,$6A,$B9,$FC,$1F,$92,$3C,$44
	.dc.b	$34,$76,$F8,$4F,$2F,$C5,$C8,$E7,$87,$9A,$3D,$3F,$0D,$DE,$78,$F3
	.dc.b	$B3,$67,$9C,$3D,$B9,$7C,$0F,$85,$98,$74,$D3,$DF,$1F,$13,$E1,$BE
	.dc.b	$2F,$C3,$4A,$D1,$F0,$9F,$93,$F0,$9F,$2B,$E1,$BE,$5F,$C2,$4A,$DE
	.dc.b	$DE,$39,$E9,$18,$D9,$97,$8A,$3D,$30,$CB,$C1,$18,$CA,$EB,$F5,$B3
	.dc.b	$F0,$4F,$3C,$FC,$0F,$2C,$BB,$BD,$30,$B7,$F7,$BE,$17,$9E,$BF,$7B
	.dc.b	$2B,$7F,$7B,$2E,$AF,$93,$2E,$B8,$7E,$1B,$2F,$C4,$FC,$2C,$AF,$63
	.dc.b	$F8,$CF,$86,$FC,$6F,$C2,$FE,$3B,$E1,$A5,$7C,$18,$FC,$8F,$C2,$FE
	.dc.b	$4B,$E1,$64,$FC,$58,$E4,$F4,$21,$F1,$AF,$AB,$F2,$EF,$AB,$F2,$AF
	.dc.b	$AB,$F3,$EF,$AB,$F3,$AF,$F4,$3F,$09,$23,$EA,$3F,$A4,$F8,$5F,$D2
	.dc.b	$FC,$37,$E6,$A7,$F1,$BE,$2F,$F8,$7D,$FE,$86,$DE,$7F,$C8,$4E,$F7
	.dc.b	$FE,$21,$3F,$E2,$3A,$DF,$C2,$E7,$93,$F4,$B8,$7F,$0F,$8C,$4F,$4F
	.dc.b	$84,$F1,$3E,$42,$1F,$86,$CB,$CA,$FE,$3B,$0F,$11,$E3,$E8,$67,$D3
	.dc.b	$1E,$D5,$BC,$71,$7E,$59,$FE,$41,$8E,$3E,$13,$3F,$09,$30,$ED,$1C
	.dc.b	$3E,$13,$8F,$C2,$C9,$F1,$47,$4F,$85,$91,$E5,$7B,$FC,$1F,$58,$E6
	.dc.b	$43,$E6,$FC,$2F,$57,$D9,$7D,$5F,$A3,$7D,$1F,$A7,$77,$8B,$E7,$E1
	.dc.b	$EC,$C3,$B5,$7D,$8C,$70,$F9,$8E,$A4,$3A,$9E,$72,$1D,$8F,$9D,$3F
	.dc.b	$23,$E0,$FE,$4F,$C2,$C9,$FA,$51,$F3,$3E,$13,$E6,$FC,$2C,$9F,$AD
	.dc.b	$1F,$43,$E1,$24,$7D,$97,$E9,$FC,$2F,$58,$E6,$43,$F2,$3F,$09,$D5
	.dc.b	$FC,$05,$F5,$7F,$2B,$7D,$1F,$CB,$DD,$D7,$F2,$78,$7E,$03,$E1,$3F
	.dc.b	$03,$F0,$72,$FF,$15,$1F,$84,$F8,$49,$1F,$8D,$7F,$0F,$F0,$BF,$88
	.dc.b	$F8,$49,$3F,$91,$7F,$1D,$F9,$67,$F8,$AF,$BF,$F3,$BD,$6C,$7C,$17
	.dc.b	$D8,$FE,$4B,$5F,$E2,$EA,$DE,$3E,$09,$FD,$7B,$E0,$7B,$10,$FD,$9E
	.dc.b	$1E,$09,$F0,$FC,$27,$8B,$E1,$25,$76,$8F,$2F,$C3,$48,$F0,$BE,$8F
	.dc.b	$85,$F4,$FC,$1C,$9F,$2B,$EC,$F8,$5E,$B1,$CC,$87,$F5,$0D,$1F,$07
	.dc.b	$D6,$31,$85,$E3,$FF,$09,$53,$1E,$D5,$C7,$1B,$5A,$C7,$E5,$E3,$E0
	.dc.b	$AE,$D9,$61,$F4,$90,$EE,$7C,$CC,$39,$E9,$FC,$3B,$26,$58,$AF,$F2
	.dc.b	$DC,$31,$FE,$23,$0F,$C6,$1E,$7E,$1B,$1F,$D8,$8D,$46,$48,$E6,$36
	.dc.b	$FF,$7C,$36,$66,$36,$E6,$37,$A7,$F9,$BE,$36,$66,$7C,$BE,$13,$AA
	.dc.b	$92,$7C,$91,$96,$C6,$5F,$14,$65,$B1,$9A,$F0,$4E,$5B,$19,$CF,$FA
	.dc.b	$17,$E2,$FF,$E2,$19,$F3,$36,$33,$EC,$FC,$A8,$C6,$1F,$E2,$26,$A6
	.dc.b	$92,$38,$12,$3E,$74,$65,$FD,$3F,$C3,$FE,$B8,$8F,$EA,$B9,$FF,$63
	.dc.b	$F0,$F2,$3F,$5A,$FE,$D3,$E1,$7F,$6B,$F0,$D2,$7F,$62,$FE,$E3,$E1
	.dc.b	$7F,$73,$7D,$27,$F3,$98,$61,$E6,$43,$F4,$18,$75,$79,$90,$FD,$25
	.dc.b	$F5,$7C,$C7,$A8,$87,$EA,$6B,$D0,$F4,$3B,$69,$FA,$8C,$7D,$11,$D0
	.dc.b	$ED,$1F,$AC,$C7,$F8,$8F,$C2,$FF,$12,$F8,$49,$5F,$C3,$A3,$F8,$AF
	.dc.b	$C2,$C8,$FE,$22,$FF,$18,$F8,$6F,$E3,$3F,$0B,$FC,$6B,$E1,$A5,$7F
	.dc.b	$16,$8E,$AF,$F1,$8C,$3A,$BF,$C0,$70,$E8,$FF,$03,$C2,$E5,$FC,$03
	.dc.b	$1E,$B3,$FC,$83,$1E,$AF,$F0,$9C,$7A,$BF,$C2,$F1,$EB,$7F,$C2,$31
	.dc.b	$E7,$C5,$FB,$BC,$AD,$FE,$EB,$2E,$AF,$F1,$3C,$BA,$E1,$FC,$43,$2F
	.dc.b	$F0,$17,$E5,$BF,$E7,$BF,$48,$FE,$59,$BC,$9A,$F8,$6D,$FC,$34,$CB
	.dc.b	$FC,$B7,$7F,$E6,$C6,$FC,$37,$4F,$87,$91,$A7,$BF,$C3,$F3,$8E,$38
	.dc.b	$F8,$8C,$3F,$E7,$8B,$E8,$DC,$3C,$0E,$70,$EA,$EF,$0B,$A7,$FA,$F6
	.dc.b	$FF,$CD,$73,$E6,$CB,$B3,$E0,$30,$7A,$B0,$EB,$5C,$5F,$8B,$87,$4E
	.dc.b	$0F,$F1,$AE,$0C,$71,$83,$4F,$93,$A7,$CC,$F8,$7F,$9B,$F0,$D2,$BE
	.dc.b	$3C,$7D,$0F,$86,$FA,$3D,$F2,$7E,$5C,$7D,$4F,$86,$91,$F3,$9F,$AF
	.dc.b	$F0,$FD,$63,$E8,$E1,$D5,$F9,$18,$75,$7E,$AE,$1D,$5F,$99,$87,$47
	.dc.b	$E7,$61,$78,$3E,$46,$3D,$67,$99,$E8,$3E,$FF,$C3,$F5,$7E,$B6,$1D
	.dc.b	$6B,$E8,$61,$CF,$7F,$8F,$8D,$BF,$8D,$BB,$2F,$E9,$BF,$8C,$3D,$8B
	.dc.b	$ED,$CB,$FE,$99,$39,$6C,$FF,$54,$FA,$57,$B1,$4E,$71,$27,$FA,$A4
	.dc.b	$BD,$7E,$06,$63,$FA,$95,$F5,$6D,$67,$6F,$EA,$13,$96,$C6,$75,$FF
	.dc.b	$41,$4E,$5B,$19,$AF,$F6,$FC,$E5,$B1,$95,$19,$6C,$66,$1E,$E5,$8E
	.dc.b	$18,$BD,$BC,$5F,$C4,$3D,$FE,$D7,$EF,$3E,$26,$47,$ED,$9F,$DF,$FC
	.dc.b	$47,$F0,$0F,$89,$FE,$03,$F1,$12,$BF,$77,$1F,$C1,$3E,$27,$F8,$2F
	.dc.b	$C4,$C9,$FD,$FC,$72,$7A,$10,$FD,$C5,$F5,$7F,$82,$DF,$57,$F7,$97
	.dc.b	$D5,$FE,$13,$7D,$5F,$E0,$17,$D5,$E6,$7F,$0E,$BF,$E2,$FF,$13,$D6
	.dc.b	$3F,$83,$61,$D7,$87,$F8,$06,$3D,$67,$F8,$56,$3D,$67,$F8,$26,$3F
	.dc.b	$C8,$3E,$22,$57,$F1,$A7,$FD,$49,$F9,$77,$FA,$E4,$FF,$51,$0F,$88
	.dc.b	$C9,$AF,$89,$DF,$C3,$CB,$FF,$7E,$D2,$39,$FA,$FD,$27,$B9,$FC,$77
	.dc.b	$0C,$3C,$2F,$AB,$CB,$0E,$CF,$4B,$EC,$B2,$73,$FF,$B2,$32,$F0,$4F
	.dc.b	$8B,$2E,$AF,$73,$DC,$79,$8E,$A7,$39,$E5,$87,$8A,$30,$7F,$1D,$AF
	.dc.b	$87,$87,$8B,$0F,$0C,$E6,$3D,$EF,$C2,$F8,$8F,$97,$F1,$32,$30,$7C
	.dc.b	$6E,$CF,$9E,$47,$E4,$F4,$7D,$2E,$9F,$A7,$F1,$32,$7E,$6C,$7D,$6C
	.dc.b	$3A,$47,$9B,$C7,$68,$CB,$53,$DC,$86,$E1,$97,$E0,$32,$F0,$4F,$DA
	.dc.b	$CB,$C0,$EF,$2E,$6F,$1C,$AE,$9F,$0B,$3F,$0F,$57,$AE,$BB,$65,$6B
	.dc.b	$59,$FF,$03,$BF,$3E,$3D,$A7,$39,$78,$B0,$F3,$75,$7E,$53,$BA,$47
	.dc.b	$E3,$A3,$F2,$FD,$D2,$3F,$22,$FE,$6B,$E2,$7F,$37,$ED,$C9,$FC,$AB
	.dc.b	$F9,$EF,$87,$FC,$FD,$F4,$9F,$89,$86,$1E,$64,$3E,$46,$1D,$5E,$64
	.dc.b	$3E,$65,$F5,$7C,$C4,$3D,$67,$E9,$BE,$27,$D0,$F4,$3B,$67,$F1,$B2
	.dc.b	$F4,$47,$43,$B4,$7C,$AC,$BF,$6F,$F1,$1F,$B8,$F8,$99,$5F,$B2,$8F
	.dc.b	$DD,$FC,$4C,$8F,$DA,$A6,$5F,$D9,$E3,$FF,$6B,$47,$F7,$B7,$FD,$A4
	.dc.b	$DF,$8C,$E0,$FE,$3E,$33,$C5,$FC,$CF,$E1,$FC,$73,$FC,$0B,$2F,$1B
	.dc.b	$F8,$FC,$BC,$4F,$E4,$72,$B9,$78,$7A,$FA,$BE,$3C,$EC,$79,$F2,$FE
	.dc.b	$81,$18,$BF,$21,$EA,$9D,$66,$CF,$C7,$3F,$9F,$CF,$C7,$8F,$E4,$B3
	.dc.b	$F0,$E1,$FC,$27,$3B,$7F,$06,$3B,$67,$E0,$8E,$C7,$80,$F8,$59,$EB
	.dc.b	$3F,$04,$6A,$7F,$AB,$7C,$56,$49,$19,$97,$C5,$71,$F8,$99,$39,$7A
	.dc.b	$7C,$4F,$5F,$88,$ED,$F1,$52,$B9,$47,$86,$FA,$C7,$D5,$C7,$0F,$32
	.dc.b	$1F,$67,$0E,$AF,$23,$C9,$7D,$5F,$BD,$7C,$9E,$87,$A6,$B1,$BF,$E2
	.dc.b	$6F,$D3,$3D,$4F,$76,$3E,$97,$A9,$0E,$A7,$C3,$F8,$7F,$95,$F0,$B2
	.dc.b	$30,$7C,$7F,$87,$F9,$DF,$0D,$23,$07,$CB,$F8,$7F,$A4,$47,$E7,$7C
	.dc.b	$3F,$D5,$C3,$A4,$7E,$67,$0C,$3C,$8F,$CE,$5F,$57,$F3,$D8,$73,$7F
	.dc.b	$41,$85,$CF,$F5,$18,$F2,$9F,$D2,$63,$67,$F5,$9F,$0D,$F8,$5C,$3A
	.dc.b	$47,$EB,$F1,$C5,$FA,$48,$76,$3D,$27,$E1,$FE,$1F,$F1,$C4,$7F,$15
	.dc.b	$F0,$FF,$91,$F8,$7F,$C9,$7C,$34,$9F,$C7,$47,$E5,$7E,$1E,$47,$E4
	.dc.b	$5F,$CE,$FE,$5F,$FE,$A1,$E7,$63,$D6,$BF,$7C,$FE,$77,$1E,$D1,$F9
	.dc.b	$9C,$7B,$3F,$C0,$F1,$EA,$FF,$05,$C6,$E7,$FD,$17,$2E,$D8,$F5,$8F
	.dc.b	$E2,$0F,$5A,$F3,$9D,$48,$75,$3D,$04,$3B,$1F,$AF,$97,$C3,$FF,$05
	.dc.b	$8F,$F4,$B7,$05,$F8,$57,$15,$31,$12,$7F,$7A,$3E,$A1,$D1,$CD,$B1
	.dc.b	$AB,$57,$18,$FF,$03,$9C,$7A,$78,$19,$D3,$FC,$0F,$1E,$43,$94,$E7
	.dc.b	$B7,$F9,$6D,$F4,$9C,$90,$EA,$E5,$B3,$C5,$FE,$79,$19,$AE,$ED,$AE
	.dc.b	$B5,$BA,$F1,$1C,$8F,$67,$2E,$6B,$84,$74,$9F,$85,$D3,$97,$96,$4E
	.dc.b	$9F,$C5,$FC,$4F,$E3,$3E,$26,$61,$F5,$63,$F1,$FF,$13,$F9,$0C,$A4
	.dc.b	$FD,$98,$FC,$9F,$C4,$FE,$53,$E2,$24,$FD,$D8,$EA,$F3,$21,$FE,$21
	.dc.b	$E4,$7C,$4F,$68,$FC,$26,$1D,$9D,$61,$D9,$FC,$46,$1D,$9E,$58,$76
	.dc.b	$7A,$63,$78,$7F,$35,$8F,$59,$F0,$63,$D6,$7F,$3D,$8F,$59,$F2,$63
	.dc.b	$D6,$7F,$45,$8F,$EC,$3E,$1E,$57,$E6,$5F,$D9,$FC,$47,$ED,$3E,$23
	.dc.b	$F6,$BF,$0F,$2B,$F3,$D1,$FB,$8F,$89,$91,$FA,$17,$F7,$7F,$0D,$D2
	.dc.b	$3E,$26,$17,$2F,$D9,$63,$D6,$7F,$51,$8F,$57,$E5,$63,$D1,$F9,$98
	.dc.b	$DB,$FD,$D6,$3E,$69,$F4,$10,$FA,$58,$75,$7E,$A6,$1D,$6B,$F8,$16
	.dc.b	$1C,$F9,$FF,$BE,$70,$B2,$C6,$0F,$21,$2C,$FF,$8B,$FC,$37,$F1,$8F
	.dc.b	$84,$97,$FC,$06,$3F,$8D,$FC,$2C,$8F,$E0,$8F,$F1,$EF,$84,$FE,$3F
	.dc.b	$F0,$B2,$7F,$84,$3F,$C8,$BE,$1B,$AC,$7F,$0B,$C3,$AB,$F8,$DC,$3A
	.dc.b	$3F,$8F,$C2,$EF,$FC,$77,$1E,$93,$F9,$3C,$6C,$FF,$21,$C7,$AC,$F3
	.dc.b	$3D,$07,$F1,$9F,$86,$E7,$BF,$C4,$C6,$CF,$C3,$D5,$97,$4A,$F9,$39
	.dc.b	$5A,$D6,$3F,$E1,$8F,$CC,$3F,$C6,$F1,$FE,$85,$6E,$7F,$E7,$76,$DF
	.dc.b	$F8,$B5,$7C,$3F,$1F,$86,$E5,$F0,$D3,$0F,$F0,$E5,$7F,$B0,$DB,$F0
	.dc.b	$FD,$FE,$16,$46,$0D,$9E,$32,$1C,$B0,$EC,$F4,$BE,$CF,$9F,$0F,$55
	.dc.b	$F4,$8F,$55,$E1,$EA,$43,$DC,$7C,$1B,$EA,$FC,$1B,$EB,$97,$C1,$C3
	.dc.b	$94,$F4,$3D,$B7,$89,$FE,$6F,$F8,$D7,$C6,$9F,$F3,$43,$53,$48,$FF
	.dc.b	$51,$6B,$FD,$31,$AF,$F4,$66,$C7,$68,$FE,$B1,$F8,$D7,$F2,$B7,$FA
	.dc.b	$9B,$5F,$E9,$AD,$8F,$E9,$6D,$D3,$FC,$EA,$D9,$FF,$3F,$35,$34,$9E
	.dc.b	$1F,$80,$91,$E0,$7F,$CD,$DF,$72,$FB,$37,$83,$86,$D7,$F9,$92,$75
	.dc.b	$39,$24,$66,$76,$FF,$BB,$27,$6E,$67,$6E,$67,$78,$FF,$BB,$AB,$73
	.dc.b	$9A,$F2,$FE,$11,$FC,$26,$3C,$4D,$7B,$36,$B9,$37,$2F,$F5,$15,$A3
	.dc.b	$14,$8C,$76,$76,$24,$76,$6B,$D9,$B6,$BF,$F5,$77,$E0,$6F,$C1,$1D
	.dc.b	$3F,$E9,$98,$C3,$63,$38,$7F,$A3,$A7,$2D,$8C,$AB,$EC,$A7,$6E,$6F
	.dc.b	$C3,$5C,$2B,$75,$A6,$C6,$67,$77,$E3,$25,$F8,$7B,$3C,$C4,$70,$43
	.dc.b	$C0,$78,$E9,$1B,$8C,$E1,$F0,$E3,$18,$78,$9B,$87,$A1,$FE,$35,$F1
	.dc.b	$D2,$FF,$88,$BF,$C7,$7E,$37,$F8,$F7,$C7,$7F,$1F,$F8,$D9,$5F,$C5
	.dc.b	$A3,$F9,$17,$C6,$FF,$23,$F6,$E4,$FF,$1A,$8E,$8F,$F2,$5C,$2E,$7F
	.dc.b	$C9,$71,$EB,$3F,$C8,$31,$EA,$FF,$2B,$C7,$AB,$FC,$93,$1E,$AF,$F2
	.dc.b	$FC,$79,$E7,$FC,$B7,$2B,$3F,$D7,$7B,$19,$76,$8F,$E6,$D9,$76,$CF
	.dc.b	$F9,$B6,$7D,$A7,$F9,$D6,$7D,$A7,$F9,$D6,$7F,$CF,$FE,$37,$FA,$07
	.dc.b	$C6,$CB,$FE,$6F,$1F,$D0,$FE,$36,$47,$F3,$A7,$FA,$37,$C5,$FF,$47
	.dc.b	$F8,$DF,$E9,$1F,$19,$2B,$F9,$FC,$75,$7F,$A0,$E1,$D5,$FE,$99,$87
	.dc.b	$47,$FA,$76,$17,$2F,$E9,$D8,$F5,$9F,$E9,$38,$F5,$7F,$AA,$63,$D5
	.dc.b	$FE,$AD,$8F,$5B,$FE,$AD,$8F,$3E,$1F,$EA,$D9,$5B,$F9,$59,$75,$7F
	.dc.b	$AF,$65,$D7,$0F,$EB,$D9,$7F,$63,$F8,$DF,$EC,$9F,$17,$2F,$FA,$DC
	.dc.b	$7F,$67,$F8,$CF,$ED,$1F,$19,$27,$FA,$FC,$1F,$F7,$E7,$E6,$3F,$E7
	.dc.b	$AC,$7F,$63,$E7,$A3,$1B,$6E,$30,$F1,$AF,$D2,$C6,$FE,$37,$C9,$1F
	.dc.b	$DB,$33,$F2,$3F,$DD,$F3,$F2,$3F,$DC,$33,$F2,$3F,$DE,$F3,$E9,$9F
	.dc.b	$8B,$3F,$0C,$FF,$8A,$5B,$97,$8A,$3D,$39,$78,$1F,$F0,$26,$5E,$0B
	.dc.b	$FF,$02,$65,$E0,$9F,$F0,$46,$5E,$09,$FF,$04,$65,$E0,$CF,$DD,$9F
	.dc.b	$62,$39,$BF,$16,$57,$3F,$8D,$F1,$D6,$DF,$1B,$4C,$46,$7E,$3A,$61
	.dc.b	$F4,$1E,$1F,$01,$C7,$C6,$93,$F5,$1E,$9F,$07,$CA,$3E,$BE,$1D,$5F
	.dc.b	$B3,$87,$47,$E9,$E2,$C7,$1B,$7C,$BC,$BA,$D6,$B2,$B3,$F3,$F2,$ED
	.dc.b	$3C,$B2,$EC,$F4,$CB,$9E,$BF,$1B,$2B,$3F,$57,$2E,$73,$F2,$B2,$B3
	.dc.b	$F6,$71,$FE,$89,$EA,$7C,$7F,$83,$93,$FD,$0E,$32,$FE,$53,$E3,$7E
	.dc.b	$61,$87,$F9,$7F,$88,$9F,$A1,$F1,$D2,$79,$10,$DC,$71,$DF,$FA,$2B
	.dc.b	$73,$FE,$96,$D9,$FE,$7C,$DF,$17,$FA,$F3,$70,$F0,$4F,$5B,$E6,$DB
	.dc.b	$FC,$4F,$C7,$6E,$FA,$71,$FE,$AF,$E3,$B1,$7B,$3F,$39,$F1,$1F,$89
	.dc.b	$F8,$7F,$C5,$7C,$44,$9F,$DA,$C7,$E3,$7E,$2B,$F1,$DF,$1D,$27,$F7
	.dc.b	$31,$F9,$1F,$8D,$FC,$97,$C7,$49,$FD,$EC,$75,$79,$90,$FC,$3F,$C7
	.dc.b	$75,$7F,$81,$DF,$57,$F1,$77,$D5,$FE,$0F,$7D,$5F,$C7,$DF,$57,$F2
	.dc.b	$38,$5E,$0F,$DD,$63,$D6,$7F,$2B,$8F,$59,$FD,$F6,$3D,$67,$F3,$38
	.dc.b	$F5,$9F,$E0,$58,$FE,$BF,$E1,$1F,$1D,$7E,$32,$61,$FC,$6E,$FF,$8C
	.dc.b	$D6,$A3,$1C,$7F,$C6,$61,$F1,$BC,$C9,$3A,$3F,$D6,$D1,$FE,$36,$9E
	.dc.b	$1F,$19,$E1,$F8,$B9,$5B,$E4,$5A,$BC,$D9,$CA,$D4,$F8,$DF,$E4,$EB
	.dc.b	$E3,$3C,$31,$8C,$AE,$7E,$AF,$6F,$E0,$63,$E0,$AE,$39,$E1,$EB,$97
	.dc.b	$77,$AB,$F1,$F1,$E7,$A7,$C1,$CA,$D7,$BB,$1E,$4F,$43,$CF,$86,$32
	.dc.b	$F9,$7F,$1B,$F4,$49,$3E,$CF,$8D,$FA,$9F,$1B,$F5,$7E,$36,$4F,$C0
	.dc.b	$8F,$B1,$BC,$8F,$86,$FD,$BF,$77,$9C,$7C,$4C,$2E,$3F,$6B,$1E,$B3
	.dc.b	$F2,$B1,$EA,$FC,$AC,$7A,$3F,$33,$1B,$7F,$82,$C7,$F1,$18,$74,$D7
	.dc.b	$F0,$59,$63,$0F,$31,$D4,$87,$53,$CE,$43,$B1,$F7,$2B,$F2,$9F,$1B
	.dc.b	$23,$EF,$BF,$97,$ED,$FC,$C7,$BB,$27,$F0,$6F,$E6,$FD,$EF,$CE,$7C
	.dc.b	$5F,$E7,$7E,$2E,$57,$E2,$23,$A3,$F8,$8C,$2E,$BF,$9E,$C7,$A4,$FE
	.dc.b	$33,$1B,$3F,$A2,$C7,$AC,$FE,$4B,$1E,$AF,$E4,$B1,$F3,$3E,$82,$1F
	.dc.b	$96,$C3,$9E,$9F,$A9,$C6,$CF,$E9,$B0,$EA,$FE,$73,$0E,$B8,$FE,$CB
	.dc.b	$1F,$DC,$7C,$5F,$EE,$7E,$36,$5F,$E9,$23,$FB,$57,$E6,$5F,$ED,$3E
	.dc.b	$A3,$FE,$F8,$9F,$D7,$35,$FF,$07,$BB,$3F,$DF,$0E,$3E,$2B,$99,$D3
	.dc.b	$E2,$FC,$31,$FB,$2C,$3C,$2F,$EC,$B0,$F0,$BE,$2F,$8C,$8F,$80,$E6
	.dc.b	$78,$AE,$BE,$2B,$EA,$FE,$F2,$FA,$E7,$FC,$4B,$0E,$B3,$EA,$C3,$AB
	.dc.b	$EA,$39,$CF,$F8,$C3,$1F,$5B,$71,$F5,$4F,$C2,$F8,$DF,$8E,$47,$E2
	.dc.b	$FC,$6F,$CA,$F8,$8F,$97,$F0,$F2,$7E,$4C,$7C,$EF,$86,$91,$F3,$1F
	.dc.b	$A3,$F0,$FC,$E3,$E6,$E1,$72,$F2,$63,$D6,$7E,$9E,$3D,$5F,$A7,$8F
	.dc.b	$47,$EA,$E3,$6F,$D9,$8F,$3E,$5F,$F2,$56,$8C,$F1,$83,$87,$F0,$3A
	.dc.b	$77,$8F,$BF,$A7,$7C,$BE,$36,$9D,$72,$E7,$A5,$BE,$59,$B4,$FC,$6F
	.dc.b	$C3,$7E,$3B,$E1,$E5,$7E,$32,$3F,$23,$F0,$D2,$3F,$1E,$FE,$53,$E1
	.dc.b	$FF,$2B,$F0,$F2,$7F,$26,$FE,$63,$E1,$FA,$C7,$E5,$B0,$EA,$FE,$5B
	.dc.b	$0E,$8F,$E6,$30,$BB,$7D,$FC,$7A,$4F,$E7,$31,$B3,$F8,$3C,$7A,$CF
	.dc.b	$E8,$31,$EB,$3F,$87,$C7,$9E,$BF,$33,$2B,$7F,$0B,$2E,$73,$F4,$32
	.dc.b	$B3,$F1,$B2,$FD,$8F,$C3,$48,$FD,$7B,$FB,$4F,$87,$FD,$AF,$C3,$49
	.dc.b	$FD,$9B,$FB,$8F,$87,$FD,$CF,$C3,$49,$FD,$BB,$FB,$CF,$87,$E9,$1F
	.dc.b	$B8,$C2,$EF,$F9,$FC,$7A,$4F,$EF,$31,$B3,$FA,$3C,$7A,$4F,$F0,$0C
	.dc.b	$6C,$FE,$9F,$1E,$7A,$7E,$33,$2B,$7F,$82,$CB,$9C,$FE,$43,$2B,$3F
	.dc.b	$86,$CB,$9C,$FE,$53,$2B,$3F,$8A,$C7,$FD,$00,$BE,$31,$F0,$1F,$E8
	.dc.b	$1F,$CC,$FF,$D2,$67,$FA,$5A,$75,$1C,$27,$71,$FE,$07,$97,$C4,$F1
	.dc.b	$35,$52,$7F,$C2,$5F,$90,$BF,$42,$7B,$5F,$E9,$18,$C3,$B9,$70,$B5
	.dc.b	$FF,$5A,$67,$E0,$9C,$C6,$31,$E1,$5B,$9F,$F6,$15,$6E,$31,$1E,$A3
	.dc.b	$FC,$48,$53,$99,$C2,$79,$D7,$58,$DD,$73,$8C,$47,$03,$95,$76,$9D
	.dc.b	$47,$33,$8D,$7C,$DF,$89,$EB,$19,$39,$4F,$0E,$F6,$BF,$EB,$09,$E2
	.dc.b	$43,$77,$F2,$74,$7B,$3F,$67,$E2,$65,$7D,$57,$EE,$7C,$4F,$DD,$F8
	.dc.b	$9F,$BD,$F1,$32,$BE,$D4,$7E,$07,$19,$1F,$75,$FC,$27,$C4,$F4,$8F
	.dc.b	$F7,$2E,$7F,$EB,$37,$B1,$F8,$3C,$3B,$3F,$85,$C3,$B3,$C3,$0E,$AF
	.dc.b	$2C,$2E,$FF,$EF,$6C,$18,$F7,$8C,$65,$78,$3F,$DF,$14,$CF,$BE,$3E
	.dc.b	$4C,$FB,$E1,$CF,$3E,$B8,$70,$CE,$DF,$F3,$5C,$7B,$BF,$9B,$F8,$9F
	.dc.b	$D1,$11,$FC,$F7,$C3,$FE,$97,$E1,$FF,$4D,$F0,$F2,$7F,$45,$1F,$A9
	.dc.b	$F8,$6F,$D5,$7C,$3C,$9F,$D3,$47,$EB,$6F,$A4,$7C,$AC,$30,$F2,$3F
	.dc.b	$59,$87,$57,$E7,$E1,$D5,$FD,$86,$1D,$5F,$A7,$87,$27,$A1,$EB,$3D
	.dc.b	$27,$53,$EC,$D7,$5D,$7E,$85,$F5,$9F,$BB,$7D,$67,$EA,$5F,$F0,$5F
	.dc.b	$86,$95,$FC,$05,$FE,$11,$D1,$FC,$27,$93,$F8,$57,$C3,$4A,$FE,$0D
	.dc.b	$1F,$C3,$7E,$1E,$47,$F0,$94,$FF,$A4,$7F,$35,$FF,$46,$65,$F8,$3C
	.dc.b	$7B,$CF,$E4,$1E,$EF,$43,$F8,$96,$1D,$DF,$CA,$E1,$D9,$FC,$BE,$16
	.dc.b	$FF,$83,$74,$B1,$EF,$C7,$F8,$CC,$BB,$D7,$87,$0F,$E3,$D9,$FB,$5E
	.dc.b	$E4,$3B,$9E,$B2,$1E,$03,$D9,$F0,$DE,$E7,$F8,$7E,$3E,$1F,$88,$C5
	.dc.b	$F9,$32,$FF,$5D,$D7,$C3,$F8,$6C,$E7,$F1,$B0,$E1,$3F,$1F,$E1,$B2
	.dc.b	$FF,$83,$EF,$FD,$70,$FC,$BF,$88,$CD,$7B,$EA,$D3,$E1,$EF,$C4,$4C
	.dc.b	$3E,$94,$6B,$49,$1F,$55,$E3,$F0,$9C,$B0,$E7,$94,$AF,$21,$C8,$FB
	.dc.b	$98,$75,$7E,$0E,$3E,$17,$A1,$0C,$E1,$D9,$E6,$70,$AE,$CF,$2B,$F5
	.dc.b	$E3,$D7,$5F,$95,$8E,$2B,$FD,$C5,$F1,$4F,$8A,$7C,$2F,$C8,$8C,$36
	.dc.b	$33,$3C,$1A,$FF,$46,$9D,$B9,$9E,$04,$70,$43,$84,$F2,$BF,$EA,$D5
	.dc.b	$CA,$73,$F0,$DE,$12,$3E,$16,$BA,$30,$43,$04,$9E,$B4,$9E,$31,$F8
	.dc.b	$DF,$87,$FC,$09,$2B,$F6,$1F,$13,$F8,$4F,$8A,$91,$FB,$37,$F0,$FF
	.dc.b	$15,$F8,$8C,$24,$FE,$DD,$FC,$5E,$7C,$E3,$EE,$E1,$71,$FC,$E6,$3D
	.dc.b	$27,$F0,$38,$D9,$FD,$06,$3D,$27,$F0,$B8,$D9,$FD,$26,$3F,$56,$A5
	.dc.b	$7D,$56,$BF,$39,$B3,$F0,$9B,$C9,$FC,$3B,$4F,$C1,$46,$29,$18,$7A
	.dc.b	$1F,$81,$C7,$F0,$84,$70,$43,$EE,$B9,$69,$9B,$FA,$71,$96,$C7,$E3
	.dc.b	$72,$FC,$81,$1C,$10,$FC,$4B,$96,$99,$8F,$BF,$19,$6C,$7E,$67,$3F
	.dc.b	$CE,$11,$C1,$0F,$CA,$B9,$69,$98,$FC,$5C,$65,$B1,$FC,$DF,$E2,$BF
	.dc.b	$85,$91,$FE,$75,$F1,$5F,$AB,$E5,$FE,$1F,$F1,$52,$7F,$59,$1F,$C4
	.dc.b	$BE,$26,$47,$F4,$17,$F8,$AF,$C5,$73,$8F,$E0,$F8,$5C,$7F,$93,$63
	.dc.b	$D6,$7F,$A4,$63,$D5,$FD,$7E,$3D,$1F,$E1,$F8,$DB,$FE,$5D,$8F,$3D
	.dc.b	$FF,$96,$74,$33,$C6,$1F,$CA,$F3,$EC,$FF,$18,$CF,$B6,$3F,$AE,$CF
	.dc.b	$A6,$3C,$B3,$B7,$FC,$CF,$3F,$E5,$FF,$11,$23,$FA,$F3,$FC,$CB,$E2
	.dc.b	$7A,$E3,$FD,$07,$2E,$B8,$FE,$5F,$95,$9F,$F3,$AF,$23,$FA,$56,$5F
	.dc.b	$D2,$97,$C4,$E4,$D4,$77,$3F,$97,$E7,$72,$FE,$9D,$9F,$83,$2F,$83
	.dc.b	$9F,$82,$FF,$47,$9F,$81,$FE,$EB,$7F,$DA,$32,$F0,$B7,$2F,$04,$EB
	.dc.b	$0F,$EA,$D5,$A7,$E6,$E5,$E1,$6C,$FF,$7F,$CB,$C4,$FF,$59,$CB,$C5
	.dc.b	$C9,$FC,$DF,$4F,$5C,$F5,$3F,$DD,$98,$FC,$07,$1D,$A8,$C4,$7C,$53
	.dc.b	$E4,$7C,$4F,$C9,$F8,$A9,$1C,$25,$3E,$2A,$FF,$02,$63,$E2,$DF,$E1
	.dc.b	$5F,$CF,$F8,$AF,$A5,$F1,$52,$39,$47,$67,$A6,$5D,$9F,$AD,$8F,$5E
	.dc.b	$66,$3D,$74,$FB,$98,$73,$BF,$B9,$8E,$1F,$36,$1C,$F8,$BE,$96,$5D
	.dc.b	$E7,$F3,$E7,$E8,$88,$60,$F4,$1E,$FA,$ED,$3F,$6E,$FF,$19,$F1,$3F
	.dc.b	$8D,$F8,$A9,$7F,$16,$3D,$19,$F9,$E3,$E4,$7C,$57,$E4,$FE,$26,$46
	.dc.b	$0E,$07,$CD,$BE,$6F,$E5,$EF,$9B,$E6,$C7,$F5,$4F,$13,$F3,$95,$D1
	.dc.b	$FC,$F5,$F9,$9E,$84,$3A,$1F,$C3,$0E,$9C,$9F,$9C,$70,$72,$3F,$01
	.dc.b	$3D,$6B,$F3,$95,$F8,$1E,$27,$C4,$7E,$BC,$FC,$AF,$27,$E4,$F0,$EA
	.dc.b	$7E,$5C,$EB,$7E,$A7,$F6,$9F,$13,$FB,$73,$8F,$27,$E2,$A7,$44,$CF
	.dc.b	$F7,$DC,$7A,$25,$F4,$8C,$B4,$FD,$AC,$69,$FB,$D3,$FB,$5F,$89,$C3
	.dc.b	$D6,$3F,$4F,$F1,$38,$BE,$71,$FC,$27,$E2,$7F,$86,$7C,$4C,$8F,$D2
	.dc.b	$3F,$C3,$FE,$27,$84,$78,$0F,$D4,$5F,$F1,$3F,$89,$FE,$29,$F1,$32
	.dc.b	$7F,$59,$1C,$5E,$64,$3F,$8C,$57,$4D,$FE,$64,$7F,$1B,$BE,$8F,$ED
	.dc.b	$6F,$A3,$FC,$7E,$FA,$3F,$C8,$70,$F4,$3D,$0F,$8D,$7D,$33,$F8,$5D
	.dc.b	$9F,$C1,$71,$EB,$5F,$CA,$71,$EB,$5F,$C7,$71,$FE,$5D,$F1,$32,$BF
	.dc.b	$83,$3F,$CC,$7E,$27,$D1,$C3,$F6,$F7,$FE,$3F,$97,$F3,$5F,$89,$FF
	.dc.b	$37,$FE,$6C,$9A,$67,$1E,$64,$3F,$9D,$63,$D5,$F4,$E1,$D5,$FA,$07
	.dc.b	$C5,$3F,$8C,$5F,$47,$F9,$0D,$F7,$72,$7F,$1D,$C7,$D4,$F6,$30,$76
	.dc.b	$E3,$FD,$06,$5E,$58,$FF,$4A,$D7,$36,$E5,$9D,$3F,$D5,$32,$F8,$AF
	.dc.b	$7F,$C4,$CB,$F1,$C7,$C6,$C7,$FA,$C6,$3F,$13,$E3,$3E,$2F,$C3,$CB
	.dc.b	$F4,$C7,$C8,$BE,$51,$E2,$C3,$0F,$9C,$FD,$17,$17,$DB,$78,$9F,$0A
	.dc.b	$FA,$3E,$AB,$E2,$F3,$3E,$3D,$63,$3F,$D3,$71,$FE,$1F,$5F,$D2,$C7
	.dc.b	$F5,$FC,$FD,$17,$D4,$87,$53,$E8,$FC,$3F,$E0,$08,$FD,$4F,$86,$F8
	.dc.b	$4F,$EA,$0F,$AF,$F0,$FF,$86,$F8,$79,$18,$39,$90,$FA,$75,$F8,$CA
	.dc.b	$E7,$1F,$83,$3E,$BD,$72,$7E,$CD,$7E,$4A,$3C,$71,$C8,$E9,$87,$DA
	.dc.b	$AE,$6F,$BE,$3C,$47,$23,$F1,$B5,$CD,$FC,$7F,$C4,$7E,$7B,$E1,$E4
	.dc.b	$60,$FF,$54,$7B,$1F,$A2,$F0,$64,$60,$FC,$BF,$C3,$7E,$9B,$E1,$E4
	.dc.b	$60,$E4,$7E,$4A,$FA,$BF,$94,$C3,$FD,$6D,$87,$48,$FC,$BE,$18,$79
	.dc.b	$1F,$9A,$BE,$AF,$E7,$30,$FD,$AD,$F4,$CB,$F3,$78,$E2,$BC,$C7,$43
	.dc.b	$F5,$B8,$75,$79,$1F,$B0,$C3,$AD,$7E,$93,$0F,$E0,$3F,$0D,$FC,$0B
	.dc.b	$E1,$E5,$FE,$DA,$39,$57,$EA,$B0,$F5,$3F,$EB,$8C,$FF,$84,$FC,$27
	.dc.b	$F0,$AF,$84,$98,$7E,$FA,$3A,$BF,$C0,$31,$EA,$FE,$E3,$1E,$BC,$68
	.dc.b	$74,$3A,$FF,$C4,$D1,$EC,$3D,$F9,$FE,$F3,$1E,$F1,$FC,$0B,$2F,$1B
	.dc.b	$E9,$3C,$04,$3C,$07,$2C,$3C,$18,$77,$D3,$DB,$1F,$E5,$BD,$39,$D7
	.dc.b	$53,$F8,$CD,$E3,$1F,$E1,$BF,$0D,$FC,$AB,$E1,$E5,$7F,$1E,$8F,$6B
	.dc.b	$EC,$D3,$D6,$FD,$73,$FD,$DF,$87,$F3,$2F,$87,$F8,$50,$C3,$F9,$36
	.dc.b	$8F,$CD,$DF,$11,$A9,$CC,$62,$BF,$96,$DF,$F3,$CC,$BC,$71,$FC,$9B
	.dc.b	$3C,$3F,$04,$F2,$6D,$F0,$1F,$F3,$5E,$7E,$37,$FC,$AF,$87,$57,$C9
	.dc.b	$B7,$C8,$8C,$65,$FC,$BB,$2E,$B5,$DC,$FE,$7F,$8E,$27,$F1,$11,$F6
	.dc.b	$CF,$84,$E1,$FE,$77,$3E,$DB,$ED,$9D,$9F,$C2,$E1,$E0,$9F,$B3,$AA
	.dc.b	$F5,$AF,$C7,$F8,$7B,$66,$98,$F8,$8F,$AB,$3F,$81,$C7,$C9,$9F,$8C
	.dc.b	$F9,$BF,$0F,$C8,$91,$F1,$3E,$23,$9B,$F1,$B0,$AF,$C6,$3C,$F8,$76
	.dc.b	$CF,$F0,$38,$7E,$AE,$7E,$96,$3F,$77,$1E,$B1,$C6,$30,$F7,$3B,$10
	.dc.b	$D6,$5E,$D7,$EB,$E1,$DF,$83,$EE,$77,$BC,$77,$17,$6C,$FD,$B8,$7D
	.dc.b	$EC,$FE,$3E,$DF,$13,$37,$C4,$7C,$A3,$F3,$79,$73,$C7,$F1,$1A,$FC
	.dc.b	$9C,$7F,$36,$7B,$2F,$BB,$F8,$A9,$E3,$9E,$23,$64,$CF,$E8,$70,$FE
	.dc.b	$C6,$3F,$2D,$8F,$76,$D7,$1A,$DC,$7E,$60,$CD,$F6,$3F,$03,$E3,$58
	.dc.b	$F2,$7C,$1D,$7C,$18,$FE,$6F,$E1,$3F,$0F,$F0,$B2,$7F,$42,$FE,$2B
	.dc.b	$E1,$B8,$65,$E0,$DB,$F6,$98,$FE,$3B,$E2,$3F,$1F,$F1,$13,$1F,$D4
	.dc.b	$C7,$37,$EB,$E3,$70,$FD,$CE,$3D,$33,$F8,$15,$F7,$31,$E8,$FE,$CB
	.dc.b	$1E,$8F,$E0,$31,$E5,$5F,$B3,$C6,$CF,$E9,$B0,$E8,$FC,$53,$A1,$E4
	.dc.b	$3A,$10,$E8,$7F,$00,$F8,$8F,$D4,$91,$FE,$07,$F1,$1D,$38,$7E,$A4
	.dc.b	$74,$3F,$27,$EF,$7E,$C3,$E2,$24,$7F,$0A,$8F,$D9,$D7,$38,$FC,$CD
	.dc.b	$E1,$FB,$27,$D6,$38,$1F,$C4,$AB,$9B,$D8,$F1,$9E,$72,$1F,$A4,$9F
	.dc.b	$37,$1F,$E1,$F3,$FE,$31,$B7,$F0,$4C,$79,$EF,$FC,$AB,$1C,$63,$FC
	.dc.b	$83,$E2,$3F,$84,$91,$FE,$47,$F0,$FD,$39,$7F,$86,$E3,$FC,$3B,$E1
	.dc.b	$E5,$7F,$28,$7F,$89,$FE,$71,$FC,$F2,$3F,$6F,$87,$36,$9D,$0F,$82
	.dc.b	$7E,$50,$FE,$3F,$5D,$1F,$DF,$D5,$CB,$FA,$2D,$F5,$7F,$A1,$DF,$56
	.dc.b	$D7,$53,$A7,$07,$D8,$8F,$CD,$74,$31,$E9,$B7,$F3,$DC,$AE,$1F,$DD
	.dc.b	$32,$ED,$3F,$DA,$17,$C4,$FB,$89,$3F,$DE,$F1,$F2,$FC,$47,$C7,$8F
	.dc.b	$E9,$78,$FC,$4F,$88,$F8,$BF,$0F,$27,$86,$5D,$67,$9C,$7F,$23,$CB
	.dc.b	$BB,$A3,$C9,$97,$CE,$C7,$B4,$7E,$4F,$2F,$8B,$7D,$63,$96,$3E,$17
	.dc.b	$97,$53,$16,$1E,$0C,$3E,$6E,$3D,$6F,$EA,$C7,$C8,$CF,$94,$74,$39
	.dc.b	$DB,$1C,$D7,$FA,$7F,$6F,$E6,$5D,$BF,$D6,$DB,$7E,$B6,$BF,$D2,$5B
	.dc.b	$3C,$79,$7F,$99,$CE,$DA,$ED,$B1,$BA,$ED,$18,$DB,$FC,$5D,$7D,$1B
	.dc.b	$19,$B4,$E5,$B0,$EE,$E6,$48,$F1,$35,$F1,$37,$0F,$43,$6F,$BB,$67
	.dc.b	$AB,$E6,$EF,$91,$ED,$6B,$ED,$6D,$7C,$36,$D7,$A9,$B3,$E8,$6E,$1F
	.dc.b	$D8,$7C,$1F,$86,$49,$F9,$4D,$7E,$53,$6F,$E8,$36,$FE,$3B,$67,$E2
	.dc.b	$BF,$3B,$C3,$91,$F5,$5A,$FD,$56,$D7,$DB,$6D,$7D,$26,$CF,$D0,$6D
	.dc.b	$FF,$79,$F1,$7E,$D9,$27,$F0,$4D,$7F,$04,$DB,$FC,$43,$6F,$EF,$B6
	.dc.b	$7E,$EB,$FC,$5B,$E2,$3F,$8B,$FC,$3C,$9F,$CD,$C7,$E5,$FF,$40,$FF
	.dc.b	$17,$47,$E7,$BE,$27,$F8,$E9,$9F,$27,$F8,$FF,$C4,$49,$FD,$1B,$FC
	.dc.b	$8B,$E2,$7B,$47,$E9,$B1,$EC,$FE,$F7,$1E,$4F,$52,$1F,$A3,$C7,$C2
	.dc.b	$FF,$03,$BE,$CD,$3F,$63,$7D,$9F,$E1,$17,$DB,$4F,$E4,$F8,$76,$9F
	.dc.b	$E1,$99,$7A,$9F,$D9,$63,$DA,$BF,$96,$E3,$D9,$B3,$D8,$FD,$FF,$C4
	.dc.b	$FF,$3E,$23,$FC,$0B,$E2,$7B,$4F,$F3,$6C,$3B,$4F,$AF,$4F,$E8,$BF
	.dc.b	$13,$23,$F8,$4B,$FD,$23,$DD,$E7,$1F,$C8,$71,$B8,$7F,$40,$C7,$AF
	.dc.b	$07,$F3,$3C,$BA,$CF,$F0,$FD,$3D,$53,$FC,$DF,$2E,$D8,$FF,$2B,$7B
	.dc.b	$3E,$A2,$1D,$CF,$51,$F1,$33,$EB,$3D,$E3,$07,$A0,$87,$53,$F9,$1E
	.dc.b	$7F,$D8,$BE,$26,$47,$F2,$67,$FB,$2F,$C4,$7A,$A3,$FA,$76,$7D,$9F
	.dc.b	$E2,$79,$7A,$9F,$EA,$5F,$13,$FD,$B3,$E2,$24,$7F,$30,$8E,$AF,$F4
	.dc.b	$7C,$2E,$BF,$D9,$B1,$ED,$E0,$FF,$3F,$CB,$B7,$33,$3E,$D3,$FD,$0F
	.dc.b	$2E,$B9,$7F,$51,$CF,$FB,$DC,$74,$D7,$FB,$16,$96,$BF,$B3,$67,$DB
	.dc.b	$B9,$A7,$6C,$FA,$C3,$1E,$BA,$FF,$7A,$CF,$AE,$3F,$D2,$BE,$27,$FC
	.dc.b	$13,$F1,$32,$30,$7D,$03,$F3,$27,$E9,$AB,$F0,$1C,$3F,$E1,$0F,$88
	.dc.b	$FF,$08,$FC,$4C,$AF,$EA,$B1,$FE,$14,$C3,$A4,$7F,$6D,$C3,$0F,$5C
	.dc.b	$3F,$30,$F5,$E0,$FE,$93,$AF,$AB,$0E,$93,$FD,$D7,$1F,$03,$E6,$3A
	.dc.b	$1F,$D9,$30,$EA,$FE,$84,$E8,$7E,$00,$F0,$5F,$F8,$36,$BA,$E1,$FE
	.dc.b	$1E,$BE,$B5,$C4,$FF,$53,$72,$2F,$FD,$82,$D8,$E6,$7E,$2B,$3F,$C6
	.dc.b	$71,$7E,$06,$33,$C7,$F8,$19,$E3,$F1,$32,$73,$AF,$E4,$6B,$51,$8A
	.dc.b	$FC,$C4,$62,$3C,$99,$FD,$27,$44,$33,$97,$CA,$AF,$C9,$E4,$F8,$7E
	.dc.b	$78,$F7,$3F,$27,$0F,$88,$CD,$70,$F8,$9F,$7F,$C4,$CC,$BF,$C3,$76
	.dc.b	$8F,$87,$F1,$1C,$27,$E1,$EE,$AE,$91,$C3,$E2,$7E,$49,$1E,$7F,$13
	.dc.b	$D5,$F9,$F8,$5D,$FE,$76,$1C,$F1,$ED,$5E,$E1,$1D,$F4,$E9,$1F,$63
	.dc.b	$1B,$7C,$B2,$65,$D7,$75,$F2,$9F,$63,$EE,$CF,$DA,$FC,$DB,$FA,$0F
	.dc.b	$E1,$31,$EB,$7C,$B2,$F8,$19,$62,$BA,$65,$F8,$8F,$89,$91,$F2,$9F
	.dc.b	$C5,$FC,$4F,$B3,$2E,$A7,$9F,$2F,$C7,$FC,$04,$BF,$A2,$FE,$4B,$E2
	.dc.b	$38,$C7,$32,$1F,$98,$BE,$9A,$7E,$0E,$3E,$CD,$F3,$7F,$3B,$77,$0F
	.dc.b	$8D,$87,$49,$FD,$0E,$3E,$87,$A1,$F0,$F0,$E8,$FC,$33,$CA,$43,$A9
	.dc.b	$F8,$89,$FD,$4F,$C4,$FE,$AB,$E2,$24,$FE,$32,$3C,$EF,$EA,$CF,$39
	.dc.b	$F9,$0F,$89,$FD,$8F,$C4,$C8,$C1,$C4,$FC,$AD,$74,$7F,$73,$5D,$3A
	.dc.b	$3F,$55,$9F,$CD,$9E,$27,$EF,$AF,$A3,$FC,$03,$0F,$3B,$D0,$87,$43
	.dc.b	$E1,$1E,$73,$E1,$1C,$4F,$D3,$C7,$49,$FC,$94,$FF,$0A,$F8,$9F,$E1
	.dc.b	$7F,$13,$2F,$F5,$B1,$E7,$E9,$FD,$AF,$07,$F1,$0F,$89,$FE,$23,$D5
	.dc.b	$2F,$F6,$71,$CD,$E2,$41,$97,$EA,$BF,$3A,$79,$FD,$6F,$F8,$DE,$1D
	.dc.b	$6F,$F7,$B1,$D5,$FE,$43,$87,$57,$F9,$1E,$57,$1E,$98,$73,$9F,$AE
	.dc.b	$70,$39,$90,$C1,$E4,$E2,$FD,$46,$4F,$43,$F6,$73,$93,$F7,$B1,$97
	.dc.b	$FD,$56,$D7,$5D,$28,$7A,$3C,$B1,$FE,$6D,$4F,$8E,$73,$E7,$2F,$DC
	.dc.b	$7E,$4B,$FD,$D0,$FF,$01,$56,$E7,$4F,$8B,$D4,$B9,$F7,$F5,$6E,$9F
	.dc.b	$E9,$9A,$F2,$D6,$35,$91,$E8,$77,$AF,$48,$EC,$7F,$B1,$B2,$F1,$70
	.dc.b	$E7,$5F,$3C,$F4,$3E,$0E,$5E,$03,$C3,$8F,$68,$F8,$79,$7C,$5D,$BF
	.dc.b	$01,$9F,$82,$7F,$DB,$87,$80,$F9,$39,$78,$1C,$E3,$E0,$79,$E1,$DD
	.dc.b	$F0,$9E,$33,$BF,$AF,$F0,$88,$FD,$6F,$80,$F5,$6F,$F2,$30,$ED,$3F
	.dc.b	$74,$F1,$65,$F9,$5F,$8B,$F9,$7E,$DC,$9F,$BF,$1D,$1E,$E7,$E0,$B1
	.dc.b	$FA,$38,$F6,$8F,$95,$BF,$73,$0F,$C4,$AE,$8F,$C6,$BF,$B3,$1D,$70
	.dc.b	$FD,$36,$17,$8B,$F2,$B8,$F2,$9F,$EA,$79,$7A,$9F,$C6,$EB,$F1,$DF
	.dc.b	$56,$5F,$96,$CB,$BC,$62,$3E,$AF,$C5,$FE,$23,$E2,$64,$7E,$72,$3B
	.dc.b	$65,$F9,$4D,$7D,$53,$F5,$36,$FC,$7F,$C4,$FE,$43,$E2,$A4,$FE,$92
	.dc.b	$3F,$27,$87,$48,$FC,$1E,$38,$79,$1F,$18,$87,$D3,$D3,$D5,$97,$ED
	.dc.b	$2F,$AC,$FE,$33,$0F,$0B,$E7,$3F,$67,$7D,$F3,$FE,$1D,$8F,$AA,$7F
	.dc.b	$BA,$EB,$F1,$F1,$FD,$EE,$5D,$A7,$F7,$B9,$76,$C3,$F8,$B6,$5D,$A7
	.dc.b	$F3,$7F,$17,$FA,$E2,$3F,$C0,$BE,$2F,$B7,$43,$3E,$D3,$FD,$F3,$2E
	.dc.b	$D3,$FC,$17,$E2,$FF,$6F,$F1,$72,$BF,$85,$C7,$47,$B9,$FC,$3B,$1F
	.dc.b	$DE,$E3,$DB,$1E,$B3,$F9,$8D,$7F,$8C,$DF,$38,$FD,$8E,$1F,$C1,$23
	.dc.b	$A6,$9F,$CC,$71,$B7,$FC,$83,$1E,$9C,$3F,$7B,$1F,$A5,$C3,$C5,$FE
	.dc.b	$3F,$97,$5C,$7F,$9D,$65,$E3,$ED,$FD,$61,$F3,$EB,$C8,$7D,$0C,$3A
	.dc.b	$57,$E5,$B0,$F8,$FF,$92,$FF,$42,$AF,$F1,$0C,$FF,$8B,$A3,$11,$E2
	.dc.b	$3F,$0E,$52,$47,$37,$2D,$73,$9F,$DE,$AF,$26,$FF,$7E,$7C,$A4,$70
	.dc.b	$48,$FE,$5B,$7F,$C6,$6B,$95,$76,$6B,$FD,$02,$3F,$9F,$B9,$3F,$01
	.dc.b	$3E,$EE,$8F,$E6,$BF,$19,$27,$FA,$4B,$FC,$E3,$E2,$FC,$B5,$F1,$3B
	.dc.b	$BF,$9D,$FC,$67,$F3,$CF,$8B,$98,$7F,$50,$8E,$4F,$F2,$EC,$2E,$BF
	.dc.b	$DB,$71,$E7,$3E,$3E,$56,$9E,$68,$FE,$6B,$87,$47,$F9,$BE,$3E,$07
	.dc.b	$CA,$43,$B1,$FD,$82,$FA,$BF,$38,$FF,$30,$D7,$5C,$3F,$BF,$5F,$57
	.dc.b	$FC,$01,$7D,$6B,$FB,$3F,$E6,$5F,$DB,$9B,$F1,$5F,$D7,$20,$D8,$EF
	.dc.b	$5F,$74,$FC,$36,$9B,$F8,$B9,$3F,$8A,$E0,$FF,$31,$60,$9E,$9F,$19
	.dc.b	$CE,$B4,$7F,$9E,$27,$BB,$8D,$91,$88,$F3,$9E,$2B,$FC,$0E,$3E,$28
	.dc.b	$E1,$3D,$DB,$87,$91,$F7,$E1,$E1,$7D,$E7,$0C,$B1,$7E,$DC,$3A,$92
	.dc.b	$7E,$6E,$5C,$E7,$FD,$27,$8F,$79,$F9,$78,$FB,$9F,$1F,$C6,$7C,$BF
	.dc.b	$89,$91,$83,$F2,$C7,$E8,$78,$BF,$3F,$8F,$D1,$F8,$AF,$A5,$F0,$92
	.dc.b	$BD,$B1,$F5,$70,$E7,$1F,$6B,$0C,$3D,$39,$BF,$54,$FC,$BE,$3F,$C2
	.dc.b	$6D,$F5,$30,$E9,$3F,$82,$C7,$C0,$F9,$4F,$C2,$E3,$D1,$FC,$3F,$C5
	.dc.b	$7D,$17,$F5,$A7,$5C,$F8,$BC,$CF,$A7,$58,$DF,$E3,$FC,$5F,$E3,$49
	.dc.b	$3F,$63,$E2,$FA,$3F,$C6,$39,$3F,$4F,$87,$E4,$FE,$2E,$4F,$DE,$7F
	.dc.b	$2D,$EE,$F2,$8F,$CD,$E1,$71,$FA,$B8,$F4,$E0,$FE,$83,$C1,$F0,$AF
	.dc.b	$F2,$18,$F4,$7F,$46,$F4,$7F,$17,$87,$3E,$0F,$CD,$BF,$A5,$C3,$A7
	.dc.b	$17,$D8,$AF,$21,$0E,$67,$E5,$33,$FF,$24,$FA,$AF,$49,$DF,$86,$D7
	.dc.b	$E9,$27,$53,$92,$46,$67,$6F,$F9,$96,$76,$E6,$76,$E6,$77,$B3,$3F
	.dc.b	$F6,$FD,$F0,$AD,$5F,$8F,$D3,$91,$DD,$AF,$36,$DF,$38,$57,$B6,$73
	.dc.b	$7D,$B0,$46,$2B,$71,$87,$B1,$ED,$F4,$BE,$21,$1C,$10,$F6,$39,$69
	.dc.b	$9C,$3C,$98,$7C,$F9,$D4,$62,$91,$F4,$7D,$59,$7F,$35,$AF,$C9,$6D
	.dc.b	$71,$6E,$3E,$D6,$CF,$B1,$B8,$7D,$FF,$57,$EF,$12,$7E,$D3,$5F,$AC
	.dc.b	$DB,$F3,$36,$FE,$53,$67,$E4,$BF,$89,$F5,$E4,$7E,$15,$AF,$E0,$5B
	.dc.b	$5F,$05,$B5,$F5,$5B,$3F,$51,$B7,$F9,$7F,$5F,$F2,$C4,$9F,$C9,$35
	.dc.b	$FC,$73,$6F,$E7,$36,$FF,$04,$D9,$FC,$0B,$FA,$1F,$62,$47,$E7,$5A
	.dc.b	$FE,$65,$B5,$F6,$5B,$5F,$8D,$6C,$FE,$31,$B7,$FA,$FF,$67,$F5,$C4
	.dc.b	$9F,$D5,$35,$FD,$33,$6F,$F0,$CD,$BF,$CD,$36,$7F,$32,$FE,$E7,$DA
	.dc.b	$91,$FB,$56,$BF,$B1,$6D,$7E,$45,$B5,$FA,$56,$CF,$E9,$1B,$7F,$C1
	.dc.b	$F8,$BF,$83,$12,$7F,$81,$35,$FD,$F3,$6F,$F3,$CD,$BF,$D9,$36,$7F
	.dc.b	$62,$FF,$11,$F6,$E4,$7F,$0D,$6B,$FC,$25,$B5,$FA,$96,$D7,$EF,$5B
	.dc.b	$3F,$BC,$6D,$FF,$1F,$F7,$3F,$8F,$12,$7F,$8D,$35,$FE,$2C,$DB,$FD
	.dc.b	$B3,$6F,$F8,$53,$67,$F8,$4B,$FC,$A7,$DE,$91,$FC,$95,$AF,$F2,$16
	.dc.b	$D7,$F0,$16,$D7,$F1,$56,$CF,$F1,$46,$DF,$F3,$7F,$7B,$F9,$B1,$27
	.dc.b	$F9,$93,$5F,$E5,$CD,$BF,$E1,$CD,$BF,$E4,$4D,$9F,$E4,$2F,$F4,$1F
	.dc.b	$7E,$47,$F3,$D6,$BF,$CE,$5B,$5F,$C6,$5B,$5F,$CB,$5B,$3F,$CB,$1B
	.dc.b	$7F,$D3,$FD,$FF,$E9,$C4,$9F,$E9,$4D,$7F,$A3,$36,$FF,$93,$36,$FF
	.dc.b	$9D,$36,$7F,$9C,$BF,$D6,$7E,$02,$47,$F5,$56,$BF,$D4,$5B,$5F,$CC
	.dc.b	$5B,$5F,$D1,$5B,$3F,$D1,$1B,$7F,$D9,$FE,$07,$FB,$31,$27,$FB,$13
	.dc.b	$5F,$EB,$CD,$BF,$E7,$CD,$BF,$EA,$4D,$9F,$EA,$2F,$F9,$7F,$E2,$FF
	.dc.b	$CC,$1F,$17,$27,$FC,$97,$1F,$E6,$3F,$8B,$91,$FE,$4E,$7F,$CC,$DF
	.dc.b	$15,$FE,$67,$F8,$AF,$F3,$47,$C5,$4A,$FF,$29,$C7,$57,$FC,$AB,$87
	.dc.b	$57,$FC,$E3,$87,$47,$FC,$E7,$85,$CF,$FC,$9B,$8F,$59,$FF,$2E,$E3
	.dc.b	$D5,$FF,$3C,$E3,$FE,$77,$F8,$AF,$F3,$C7,$C5,$49,$FF,$31,$C7,$F8
	.dc.b	$37,$5F,$F3,$E7,$C5,$49,$FF,$05,$46,$5F,$F3,$87,$E5,$5F,$DE,$D6
	.dc.b	$7F,$DF,$7C,$B4,$E6,$3F,$B5,$57,$86,$BF,$CE,$38,$F8,$5F,$F4,$6E
	.dc.b	$3D,$5F,$F4,$7E,$3E,$17,$FC,$29,$59,$BE,$D8,$75,$84,$B1,$F0,$C6
	.dc.b	$CF,$F4,$6E,$1D,$EF,$FC,$E7,$8F,$8E,$FA,$9F,$E8,$4C,$7C,$13,$E8
	.dc.b	$8F,$82,$7C,$0B,$EE,$FC,$07,$C3,$87,$82,$36,$64,$F2,$62,$E0,$E3
	.dc.b	$7E,$AE,$2B,$87,$2A,$46,$2F,$8D,$FA,$DF,$F5,$16,$3E,$6B,$FF,$14
	.dc.b	$19,$C7,$C5,$3E,$16,$E3,$EB,$8C,$63,$E3,$7E,$DE,$3D,$EB,$EC,$B6
	.dc.b	$75,$3F,$18,$FA,$A7,$C7,$3D,$F5,$D4,$FA,$B7,$C4,$F2,$4F,$28,$F8
	.dc.b	$13,$E0,$3F,$17,$8F,$A7,$D2,$65,$ED,$9F,$F2,$0E,$1C,$8E,$8F,$81
	.dc.b	$F9,$1C,$9E,$7E,$59,$3F,$96,$8E,$95,$CF,$97,$FC,$D9,$7C,$EB,$EE
	.dc.b	$1C,$CF,$AF,$5C,$9F,$8F,$7F,$0E,$39,$C7,$C1,$3E,$26,$5F,$06,$BE
	.dc.b	$D1,$E6,$8C,$BE,$67,$F5,$35,$D6,$7F,$37,$5D,$27,$E6,$9C,$0F,$82
	.dc.b	$79,$8F,$B3,$5F,$33,$9B,$EB,$73,$C9,$FD,$AC,$73,$BF,$F6,$8E,$5F
	.dc.b	$89,$CB,$E0,$D7,$39,$FA,$EF,$C1,$DF,$F7,$37,$CE,$3F,$0F,$87,$E0
	.dc.b	$E3,$A6,$3F,$EF,$8C,$7A,$78,$7F,$B3,$BF,$CF,$CF,$C1,$C7,$E2,$47
	.dc.b	$E8,$AF,$A4,$7E,$EB,$0E,$93,$E1,$D7,$F0,$2F,$CD,$9E,$39,$FA,$23
	.dc.b	$E0,$9E,$9C,$3F,$42,$FA,$77,$74,$FE,$64,$F2,$CF,$E2,$F5,$63,$DA
	.dc.b	$FE,$76,$1D,$71,$F8,$71,$D8,$CE,$9E,$D9,$D6,$1E,$09,$CB,$D9,$FC
	.dc.b	$54,$69,$FB,$26,$8C,$CF,$33,$E4,$61,$FC,$A3,$AB,$94,$FE,$06,$3F
	.dc.b	$95,$F5,$E3,$1F,$97,$1F,$CB,$BB,$3F,$6D,$DB,$2F,$F9,$AB,$FB,$9E
	.dc.b	$EE,$1A,$79,$F0,$FE,$53,$8E,$2F,$CB,$1F,$CB,$30,$E5,$1F,$C1,$B0
	.dc.b	$B3,$FC,$C3,$0E,$1A,$7B,$72,$F2,$DF,$B3,$2F,$DF,$E3,$C1,$FE,$39
	.dc.b	$8D,$9F,$E4,$78,$72,$D3,$E0,$EE,$D3,$18,$7F,$11,$C7,$F8,$97,$7C
	.dc.b	$C7,$FA,$8B,$FC,$57,$C0,$DC,$7F,$54,$BA,$FF,$1E,$AE,$0F,$F2,$0A
	.dc.b	$BB,$7F,$4A,$BD,$BF,$12,$7D,$47,$E2,$EB,$78,$7F,$35,$BB,$3F,$CF
	.dc.b	$2F,$8C,$FF,$39,$C2,$CF,$F4,$0B,$FE,$55,$E0,$C8,$FE,$D2,$FF,$2D
	.dc.b	$F0,$B7,$BF,$CA,$9F,$C3,$1F,$81,$D1,$7C,$63,$F9,$CD,$DD,$3F,$B1
	.dc.b	$E1,$CA,$FE,$55,$7A,$8E,$57,$BC,$BF,$A7,$61,$67,$FA,$AD,$F1,$CB
	.dc.b	$95,$FF,$41,$F0,$E5,$7F,$80,$1F,$E8,$9E,$26,$E3,$B6,$7F,$94,$CF
	.dc.b	$CF,$1B,$21,$FD,$42,$B8,$BC,$33,$FC,$A1,$F2,$5C,$1F,$94,$7F,$B2
	.dc.b	$61,$E2,$8E,$47,$4B,$FF,$01,$E1,$E3,$8F,$93,$1E,$39,$FE,$B5,$E2
	.dc.b	$FF,$5B,$F6,$A5,$7F,$86,$A3,$FB,$77,$23,$C5,$FE,$09,$87,$F8,$6F
	.dc.b	$E0,$1F,$17,$FD,$95,$B8,$7F,$84,$25,$A7,$F8,$4B,$0E,$5D,$BD,$31
	.dc.b	$E7,$3F,$E0,$6C,$1F,$13,$CC,$D5,$79,$B1,$FF,$16,$E9,$E6,$7F,$BB
	.dc.b	$E9,$A7,$C1,$87,$3D,$3C,$86,$63,$FC,$21,$9F,$89,$D9,$FE,$25,$CB
	.dc.b	$C2,$E3,$91,$18,$8E,$3A,$7A,$1F,$F0,$CE,$3E,$3D,$7F,$C6,$39,$F9
	.dc.b	$E7,$7A,$78,$9E,$98,$7C,$67,$27,$8B,$2F,$23,$FE,$15,$8F,$05,$FC
	.dc.b	$FD,$7D,$4F,$D1,$F4,$3E,$81,$2B,$AF,$BB,$AD,$FE,$14,$78,$37,$C2
	.dc.b	$CB,$BE,$1D,$63,$CB,$8E,$B7,$F8,$39,$75,$9F,$23,$F7,$F2,$D6,$7E
	.dc.b	$BC,$3F,$09,$97,$6C,$BA,$E7,$F7,$73,$FB,$99,$7B,$F7,$F9,$5A,$76
	.dc.b	$8F,$8B,$A7,$7C,$BF,$1D,$A7,$6C,$FE,$55,$7C,$82,$4F,$A5,$F9,$18
	.dc.b	$F9,$3E,$2F,$F2,$9F,$13,$27,$E8,$C7,$37,$E2,$1E,$3D,$FD,$75,$E6
	.dc.b	$8E,$BA,$FE,$56,$B9,$47,$D5,$BF,$D0,$47,$38,$FB,$18,$FC,$1D,$7D
	.dc.b	$BB,$FD,$AC,$3E,$15,$77,$3C,$67,$E1,$30,$EB,$9F,$EB,$71,$E8,$FC
	.dc.b	$C9,$E0,$79,$0F,$CC,$65,$E7,$8F,$D7,$C7,$DD,$C7,$C3,$F1,$3F,$B5
	.dc.b	$24,$7E,$4B,$E2,$BA,$E5,$F8,$68,$E9,$C3,$F6,$1E,$A7,$E1,$0F,$0E
	.dc.b	$1F,$C0,$6F,$9C,$7E,$5D,$FE,$05,$7D,$23,$F5,$9A,$FC,$3D,$BC,$D8
	.dc.b	$FC,$29,$F5,$70,$FE,$BB,$0E,$78,$F5,$9F,$31,$E0,$AF,$34,$7C,$38
	.dc.b	$F3,$E7,$F3,$B4,$EB,$1F,$07,$1F,$CA,$FC,$57,$F1,$53,$C5,$1F,$80
	.dc.b	$3E,$B9,$D4,$E0,$7B,$CF,$C5,$99,$D7,$E2,$12,$FF,$81,$36,$BA,$CF
	.dc.b	$91,$B8,$7D,$49,$D4,$F3,$AF,$C2,$FC,$4E,$B2,$FE,$4D,$1F,$88,$F8
	.dc.b	$AC,$57,$90,$FE,$09,$F1,$5F,$CB,$F8,$E4,$7F,$0C,$7F,$99,$7C,$4F
	.dc.b	$0C,$BE,$D7,$07,$F1,$AC,$31,$8F,$E5,$31,$DE,$3C,$48,$7F,$11,$C3
	.dc.b	$7A,$FF,$3D,$BB,$3F,$C6,$AB,$8B,$FC,$5F,$2F,$1B,$C4,$F7,$57,$F0
	.dc.b	$6A,$E3,$1E,$A3,$F9,$1F,$C4,$FF,$49,$23,$FC,$9B,$E2,$37,$BF,$F4
	.dc.b	$FA,$BB,$7F,$47,$CB,$77,$FC,$9F,$0B,$7F,$D5,$70,$DC,$FF,$3D,$C2
	.dc.b	$EB,$FC,$F3,$0D,$E3,$FD,$47,$1B,$7F,$D0,$F0,$DE,$6C,$6E,$1F,$D7
	.dc.b	$FE,$27,$FB,$07,$C4,$CC,$3F,$9E,$C7,$0A,$DC,$60,$FE,$83,$3B,$7F
	.dc.b	$9F,$4E,$DF,$8E,$7E,$4F,$3F,$33,$C4,$FE,$89,$3C,$74,$FE,$E5,$5C
	.dc.b	$67,$F2,$07,$1C,$58,$FF,$75,$F8,$99,$3F,$D4,$9F,$EF,$1F,$13,$C7
	.dc.b	$0F,$CB,$70,$FE,$70,$9C,$1F,$DF,$70,$E3,$87,$F5,$5C,$2D,$FF,$80
	.dc.b	$F0,$E5,$C1,$F9,$AC,$3F,$33,$AF,$AB,$0E,$07,$84,$87,$F5,$FA,$E2
	.dc.b	$FE,$7E,$FC,$75,$FD,$3B,$1F,$F0,$47,$A9,$2B,$FA,$53,$5F,$EE,$AD
	.dc.b	$AF,$5B,$71,$FD,$6B,$67,$F9,$33,$78,$FF,$C3,$DE,$AF,$F8,$54,$93
	.dc.b	$FD,$5D,$AF,$F8,$09,$B7,$F2,$1B,$7F,$BA,$6C,$FF,$33,$7F,$C4,$1E
	.dc.b	$BC,$8F,$EC,$4D,$7F,$C1,$AD,$AF,$A6,$DA,$FE,$0A,$D9,$FE,$7C,$DB
	.dc.b	$FF,$1B,$7A,$FF,$E2,$D2,$4F,$F6,$F6,$BF,$E1,$A6,$DF,$E0,$1B,$7F
	.dc.b	$C4,$9B,$3F,$D2,$DF,$F1,$C7,$B1,$23,$FB,$D3,$5F,$F1,$2B,$6B,$F1
	.dc.b	$6D,$AF,$E3,$AD,$9F,$EA,$CD,$BF,$F2,$77,$B3,$FE,$45,$24,$FF,$81
	.dc.b	$DA,$FF,$8C,$9B,$7F,$98,$6D,$FF,$2A,$6C,$FF,$63,$7F,$CA,$1E,$D4
	.dc.b	$8F,$F0,$93,$5F,$F1,$EB,$6B,$F4,$6D,$AF,$E6,$AD,$9F,$ED,$CD,$BF
	.dc.b	$F3,$37,$17,$F9,$74,$93,$FE,$1F,$6B,$FE,$4A,$6D,$FE,$C1,$B7,$FD
	.dc.b	$09,$B3,$FD,$ED,$FF,$34,$7B,$72,$3F,$C5,$4D,$7F,$CA,$AD,$AF,$DD
	.dc.b	$B6,$BF,$A6,$B6,$7F,$C0,$CD,$BF,$F3,$F7,$BD,$FE,$7B,$F8,$99,$5F
	.dc.b	$E3,$78,$FF,$3C,$07,$FC,$CD,$F9,$DD,$69,$A3,$81,$AF,$89,$D6,$5F
	.dc.b	$E1,$6E,$65,$27,$85,$7F,$9B,$AF,$C2,$FF,$98,$F0,$CE,$1F,$C5,$A7
	.dc.b	$9E,$7E,$3C,$79,$DE,$03,$87,$C6,$75,$3C,$84,$30,$4A,$FF,$4F,$5F
	.dc.b	$89,$F0,$47,$27,$D5,$8F,$C1,$84,$61,$F3,$DF,$F7,$CC,$3C,$98,$7C
	.dc.b	$0C,$6E,$1F,$06,$FA,$5E,$DA,$7F,$8E,$EF,$9B,$4C,$E5,$F6,$1B,$A2
	.dc.b	$77,$39,$9F,$65,$75,$69,$CC,$F4,$BF,$2E,$DF,$13,$CB,$0E,$95,$E4
	.dc.b	$CB,$EC,$FC,$56,$72,$F2,$C7,$DC,$F8,$8C,$56,$F1,$E6,$F9,$AB,$70
	.dc.b	$BD,$CF,$92,$F9,$3F,$EB,$61,$F0,$DF,$EE,$A7,$CE,$7D,$EA,$7C,$57
	.dc.b	$CC,$3F,$05,$1C,$3D,$0B,$5C,$78,$3E,$CF,$C5,$7D,$23,$33,$2B,$F3
	.dc.b	$BF,$13,$F5,$8C,$3E,$03,$A1,$FA,$2C,$3C,$0F,$9F,$0E,$DC,$1F,$A5
	.dc.b	$6C,$79,$F0,$F0,$3B,$C3,$AE,$3F,$8C,$C7,$C1,$39,$21,$E1,$BE,$6F
	.dc.b	$88,$C5,$72,$C3,$C0,$D8,$FD,$C4,$F0,$24,$66,$39,$5F,$FA,$29,$B5
	.dc.b	$C9,$F0,$C7,$E8,$3E,$2B,$85,$F2,$9F,$E0,$7F,$13,$8C,$3C,$1A,$FE
	.dc.b	$FB,$E2,$7F,$3F,$C1,$2F,$F8,$4B,$FA,$2E,$7F,$D1,$F1,$C9,$FE,$1A
	.dc.b	$FE,$9B,$A3,$F4,$FD,$32,$7F,$88,$BF,$AA,$EA,$E9,$1F,$84,$CB,$F8
	.dc.b	$6E,$38,$DF,$F6,$F9,$75,$AD,$72,$33,$C5,$7E,$F3,$3E,$D5,$AC,$EC
	.dc.b	$FF,$00,$CF,$F7,$3D,$7F,$BA,$EC,$97,$FC,$92,$3F,$1F,$EF,$FE,$4B
	.dc.b	$F0,$57,$95,$4C,$D7,$F2,$AF,$03,$F8,$19,$1F,$E5,$BE,$0F,$F0,$6F
	.dc.b	$0E,$47,$F3,$07,$F8,$5F,$E7,$9F,$5F,$5F,$B1,$D0,$CF,$B5,$FE,$A7
	.dc.b	$3F,$0B,$F9,$58,$E3,$F1,$52,$3F,$9E,$5F,$F3,$5C,$13,$DB,$E2,$FC
	.dc.b	$F5,$F9,$1B,$B7,$E0,$6F,$67,$E7,$DB,$3F,$90,$6B,$FB,$A6,$E5,$E7
	.dc.b	$8F,$0C,$FF,$3B,$C5,$F1,$5E,$EF,$8B,$95,$A3,$F9,$EC,$E9,$FE,$0D
	.dc.b	$0F,$67,$F8,$0F,$E0,$EF,$76,$9F,$C1,$27,$55,$CF,$B5,$83,$E3,$3E
	.dc.b	$69,$27,$50,$F8,$CF,$A3,$F1,$92,$72,$FD,$4F,$8B,$ED,$A7,$82,$7F
	.dc.b	$8D,$69,$E3,$7C,$F1,$F6,$BE,$16,$47,$67,$EE,$F8,$9D,$63,$F3,$99
	.dc.b	$31,$C7,$1F,$D7,$CB,$BD,$76,$DF,$C0,$FF,$17,$DF,$DE,$F2,$CB,$BB
	.dc.b	$4F,$6E,$5F,$8C,$F8,$B9,$1F,$01,$FC,$7F,$27,$68,$FE,$45,$B7,$AD
	.dc.b	$B8,$E7,$3F,$C5,$63,$D9,$B3,$D8,$F9,$7F,$15,$F9,$AF,$8A,$91,$83
	.dc.b	$E7,$FC,$57,$E7,$BE,$2A,$46,$0F,$A7,$F1,$3F,$A2,$23,$F5,$BE,$27
	.dc.b	$F4,$B8,$74,$8F,$93,$86,$1E,$47,$CC,$BE,$AF,$CE,$C3,$9B,$F4,$30
	.dc.b	$B8,$FE,$97,$1F,$D8,$E7,$2B,$F0,$AF,$ED,$3E,$27,$FB,$FF,$E7,$AF
	.dc.b	$86,$BF,$0B,$2F,$F2,$13,$96,$FC,$57,$02,$3A,$E9,$EC,$FE,$0B,$1B
	.dc.b	$9F,$ED,$F2,$EF,$97,$5A,$FE,$45,$8B,$2C,$C7,$62,$1A,$CB,$CF,$97
	.dc.b	$A3,$E2,$64,$F9,$9F,$5F,$C5,$7B,$3E,$27,$DB,$F1,$52,$BD,$71,$F0
	.dc.b	$3E,$2A,$47,$B9,$F8,$7F,$15,$D2,$3C,$B8,$5D,$FF,$C8,$9A,$B2,$ED
	.dc.b	$5F,$1B,$2E,$CF,$B3,$2E,$AF,$BB,$2B,$87,$2C,$BE,$8F,$C4,$4A,$F9
	.dc.b	$EF,$D4,$F8,$3F,$AB,$F0,$D2,$7E,$9B,$F6,$3E,$2B,$EC,$FA,$12,$7E
	.dc.b	$BB,$F7,$3E,$2B,$A4,$7C,$EC,$2E,$5E,$AC,$7A,$4F,$D2,$C6,$CF,$BB
	.dc.b	$1E,$93,$F5,$B1,$B3,$F0,$B1,$FC,$4F,$C4,$CA,$FC,$3B,$F8,$CF,$8A
	.dc.b	$FC,$6F,$C5,$49,$FC,$5B,$F9,$0F,$8A,$FC,$8F,$C5,$7F,$94,$3C,$4F
	.dc.b	$F2,$B6,$5F,$E7,$16,$F9,$FF,$E6,$9E,$65,$77,$C7,$F1,$39,$DC,$3E
	.dc.b	$AE,$7D,$E7,$F1,$B9,$D9,$FB,$39,$F8,$27,$F9,$66,$7D,$9F,$F3,$9B
	.dc.b	$7C,$E6,$7E,$09,$DF,$C5,$79,$C8,$F2,$F8,$9F,$F3,$E7,$C5,$7F,$77
	.dc.b	$7D,$9F,$13,$27,$FA,$24,$E5,$BF,$01,$FE,$50,$7F,$C0,$99,$3E,$2B
	.dc.b	$E2,$1D,$27,$5A,$FC,$5F,$8A,$F9,$04,$BF,$5F,$C5,$6F,$3F,$04,$74
	.dc.b	$CF,$0F,$82,$3A,$BF,$0B,$1C,$BF,$E5,$8C,$FC,$CD,$CB,$C3,$C9,$FE
	.dc.b	$AD,$9C,$E6,$8C,$4E,$A9,$5A,$9C,$E1,$DA,$F8,$56,$AF,$85,$EE,$31
	.dc.b	$5B,$BE,$31,$F6,$3E,$2B,$F1,$04,$7E,$DF,$C5,$7F,$9B,$3E,$39,$EE
	.dc.b	$7F,$A7,$3F,$28,$FE,$D1,$5F,$DA,$DA,$FF,$99,$E7,$6F,$F9,$ED,$AF
	.dc.b	$F9,$EB,$47,$EF,$1F,$E6,$66,$FA,$DD,$F0,$FC,$D7,$C6,$CB,$FC,$6B
	.dc.b	$FE,$BF,$F9,$E7,$ED,$4F,$8B,$E1,$5E,$3F,$80,$F3,$18,$7C,$98,$7F
	.dc.b	$B1,$F2,$EF,$1F,$E0,$D6,$BA,$3B,$90,$FC,$5E,$3E,$1C,$3C,$11,$ED
	.dc.b	$AE,$EF,$84,$C1,$0F,$81,$9F,$BC,$FF,$3D,$E9,$F1,$30,$FD,$B7,$C7
	.dc.b	$FE,$E4,$95,$FA,$CF,$8F,$FD,$E7,$C7,$C8,$FD,$7B,$FB,$FE,$6F,$E0
	.dc.b	$1B,$7F,$01,$F8,$E9,$5F,$B4,$8E,$6F,$E9,$30,$B8,$FF,$05,$C7,$A4
	.dc.b	$FE,$A3,$1B,$3F,$C2,$71,$EB,$3F,$BE,$C7,$AB,$FA,$EC,$7F,$88,$7C
	.dc.b	$77,$F1,$1F,$8D,$93,$FC,$12,$3F,$8A,$7C,$77,$F1,$5F,$8E,$93,$FC
	.dc.b	$22,$3F,$8C,$7C,$74,$8F,$E1,$6F,$F1,$BF,$8E,$EB,$1F,$C3,$B0,$EA
	.dc.b	$FE,$FB,$0E,$AF,$F1,$2C,$3A,$BF,$C0,$B0,$E8,$FF,$04,$C2,$E9,$FC
	.dc.b	$97,$1F,$F7,$BF,$E8,$1F,$E6,$DE,$C5,$BC,$64,$FF,$99,$A3,$FC,$94
	.dc.b	$D7,$FD,$24,$F1,$78,$3F,$76,$79,$CF,$2E,$5E,$E7,$83,$E3,$3C,$3F
	.dc.b	$11,$E2,$EA,$98,$F6,$8F,$2E,$1E,$6F,$8E,$93,$E2,$8E,$6F,$93,$0A
	.dc.b	$F9,$30,$EA,$FA,$30,$EA,$FA,$30,$F8,$3F,$1B,$F0,$BE,$32,$4F,$BA
	.dc.b	$3E,$2F,$C6,$7C,$6F,$8A,$93,$F0,$A3,$E4,$F8,$3F,$2B,$E3,$24,$FC
	.dc.b	$68,$EA,$FC,$8C,$3A,$BF,$23,$0E,$AF,$CC,$C3,$AB,$F3,$30,$EA,$FD
	.dc.b	$0C,$3A,$BF,$43,$0F,$B3,$E6,$48,$FA,$AF,$DC,$F8,$BF,$BB,$F1,$7F
	.dc.b	$7B,$E2,$A5,$7D,$A8,$FC,$0F,$C5,$FE,$0B,$E2,$E4,$FD,$E8,$E8,$FD
	.dc.b	$EC,$2F,$47,$F3,$0E,$6F,$C2,$E5,$DA,$BF,$0B,$97,$67,$F0,$B9,$76
	.dc.b	$7F,$13,$97,$67,$F1,$39,$7E,$4B,$E2,$E4,$7E,$39,$FC,$AF,$C5,$FE
	.dc.b	$5B,$E2,$FF,$2B,$3F,$EE,$1F,$D0,$7F,$C9,$8D,$F8,$CD,$3F,$E9,$F8
	.dc.b	$FF,$22,$37,$E3,$38,$EF,$E9,$9C,$CF,$3F,$8C,$F0,$BF,$9C,$CA,$D7
	.dc.b	$BF,$BB,$FD,$04,$DC,$B9,$F6,$FE,$7B,$1F,$1C,$F5,$CF,$C7,$8F,$9F
	.dc.b	$3D,$BF,$AD,$AF,$0D,$78,$E7,$3C,$49,$69,$87,$FD,$3B,$5E,$9F,$8C
	.dc.b	$C1,$9F,$8C,$D7,$C6,$4C,$BE,$13,$C7,$E2,$F9,$7C,$64,$9F,$8C,$F5
	.dc.b	$F8,$CE,$71,$F2,$70,$AF,$CF,$C3,$A3,$F4,$70,$B8,$FC,$BC,$7A,$4F
	.dc.b	$D5,$C6,$CF,$CF,$C7,$D5,$F1,$92,$BE,$B3,$ED,$F8,$CF,$77,$C6,$7B
	.dc.b	$FE,$32,$57,$DB,$8F,$85,$F1,$92,$3E,$F3,$F1,$7E,$33,$A4,$7E,$17
	.dc.b	$0B,$97,$DF,$C7,$AC,$FE,$17,$1E,$AF,$E2,$F1,$E8,$FE,$37,$1B,$7F
	.dc.b	$87,$C7,$E9,$7C,$64,$AF,$C7,$3F,$57,$E2,$FE,$B7,$C5,$C9,$FC,$93
	.dc.b	$F6,$7E,$2F,$ED,$7C,$5F,$DB,$F8,$C9,$5F,$97,$8E,$8F,$E7,$70,$B9
	.dc.b	$7E,$5F,$1E,$93,$FA,$1C,$6C,$FE,$6F,$1E,$B3,$FA,$1C,$7A,$BF,$A7
	.dc.b	$C7,$F1,$BF,$A1,$7F,$95,$47,$C6,$60,$CC,$EB,$E1,$F7,$F1,$52,$3F
	.dc.b	$A2,$E5,$FC,$06,$FF,$C6,$DD,$87,$FC,$A1,$3A,$EF,$FE,$83,$0F,$8C
	.dc.b	$F0,$9E,$2F,$8B,$EF,$5B,$3F,$A8,$4F,$91,$C5,$F5,$CF,$70,$C7,$C5
	.dc.b	$9F,$96,$31,$97,$83,$3F,$54,$72,$C7,$CA,$E8,$EF,$9F,$91,$D6,$7F
	.dc.b	$1B,$E3,$3E,$3F,$C5,$C9,$E5,$1F,$2B,$E2,$FE,$5F,$C4,$C9,$ED,$1F
	.dc.b	$3B,$E1,$A4,$78,$5F,$A3,$F0,$DD,$63,$C9,$87,$57,$C9,$C3,$F2,$F1
	.dc.b	$ED,$1E,$9C,$7B,$3A,$D1,$97,$68,$C6,$57,$83,$EB,$67,$F8,$0F,$86
	.dc.b	$FC,$0F,$C3,$4B,$F8,$91,$F8,$4F,$85,$FC,$2F,$C3,$49,$F9,$11,$F8
	.dc.b	$8F,$85,$91,$F2,$DF,$C5,$FC,$37,$58,$F9,$D8,$75,$7C,$D8,$75,$7E
	.dc.b	$96,$1D,$5F,$56,$1D,$1F,$66,$17,$2F,$C7,$63,$F9,$8F,$85,$FC,$CF
	.dc.b	$C3,$4B,$FB,$91,$F9,$CF,$86,$FC,$EF,$C3,$49,$FC,$04,$7F,$2C,$D7
	.dc.b	$FC,$2D,$C8,$7F,$C3,$59,$7F,$5B,$6F,$4F,$F8,$61,$DF,$1B,$2F,$06
	.dc.b	$3F,$47,$3F,$23,$D8,$FC,$76,$7E,$17,$EB,$67,$E1,$C7,$C5,$F0,$D2
	.dc.b	$7F,$3B,$97,$90,$FF,$14,$3F,$E3,$16,$E9,$E7,$F8,$6F,$57,$C3,$7A
	.dc.b	$FE,$17,$D9,$F0,$D3,$0E,$71,$EF,$F8,$59,$1D,$9F,$85,$F0,$DF,$0F
	.dc.b	$09,$3E,$27,$AB,$C8,$F2,$DF,$57,$F2,$B9,$30,$EB,$18,$C2,$F1,$7C
	.dc.b	$BC,$7E,$87,$C2,$CA,$F7,$BF,$4F,$E1,$BE,$A7,$C2,$FD,$5F,$86,$95
	.dc.b	$F1,$23,$EC,$7C,$34,$8F,$8E,$FD,$BF,$86,$E9,$1E,$4C,$2E,$5F,$6F
	.dc.b	$1E,$B3,$F3,$B1,$EA,$FA,$B1,$E8,$FB,$31,$B7,$F8,$3C,$7F,$11,$F0
	.dc.b	$B2,$BE,$BB,$F8,$BF,$86,$FC,$67,$C2,$FE,$37,$E1,$A5,$7D,$C8,$FC
	.dc.b	$87,$C2,$C8,$FB,$EF,$E4,$FE,$1B,$A4,$7C,$CC,$2E,$5F,$93,$C7,$AC
	.dc.b	$FE,$1B,$1E,$AF,$D2,$C7,$A3,$F5,$31,$B7,$F9,$BC,$7F,$41,$F0,$BF
	.dc.b	$E0,$CF,$94,$FF,$0A,$F5,$FF,$9D,$1B,$87,$E3,$E7,$5F,$0B,$BF,$85
	.dc.b	$93,$FC,$CA,$BF,$C2,$F7,$C7,$E1,$3A,$7C,$24,$AD,$47,$81,$E1,$F0
	.dc.b	$DE,$13,$0F,$F1,$EB,$E6,$FF,$A3,$DF,$F0,$FB,$DF,$0F,$1B,$D0,$F1
	.dc.b	$5F,$77,$95,$FB,$7E,$16,$47,$9D,$F8,$1F,$0B,$F0,$7E,$16,$4F,$AD
	.dc.b	$F8,$9F,$0B,$F1,$7E,$17,$E3,$7C,$2C,$AF,$81,$1D,$1F,$46,$17,$8B
	.dc.b	$FC,$C5,$AB,$2E,$B5,$ED,$CA,$D6,$B2,$ED,$3F,$2F,$2E,$CF,$C3,$CB
	.dc.b	$EB,$7C,$24,$8F,$A0,$FD,$9F,$85,$FB,$5F,$0B,$F6,$FE,$16,$57,$D5
	.dc.b	$8F,$BD,$F0,$B2,$3E,$C3,$F8,$1F,$85,$E9,$1F,$47,$0B,$A7,$AB,$1E
	.dc.b	$B3,$F7,$F1,$EA,$FD,$7C,$7A,$3F,$67,$1B,$7F,$0B,$1F,$C7,$7C,$2F
	.dc.b	$E3,$FE,$16,$5F,$E2,$63,$F2,$5F,$09,$F9,$3F,$85,$93,$F8,$D8,$FF
	.dc.b	$30,$FC,$CF,$F9,$FA,$3F,$B1,$BF,$E0,$36,$FC,$36,$9E,$E7,$E5,$70
	.dc.b	$EE,$FE,$37,$0E,$EF,$E6,$70,$EE,$FE,$47,$0E,$4F,$F8,$3B,$0E,$D9
	.dc.b	$7F,$A7,$5B,$D7,$FE,$DA,$7C,$91,$9F,$86,$F3,$FC,$34,$BE,$51,$EA
	.dc.b	$F8,$59,$1D,$5F,$6F,$C3,$7B,$BE,$16,$4F,$85,$F8,$3F,$0D,$F0,$AF
	.dc.b	$A4,$FC,$2C,$30,$F3,$21,$F1,$B0,$EA,$F3,$21,$F2,$AF,$AB,$FE,$2F
	.dc.b	$F8,$EF,$F1,$C7,$85,$F4,$3E,$1A,$4E,$7D,$FF,$E0,$B5,$A8,$F8,$DF
	.dc.b	$0F,$F5,$8F,$F4,$DC,$66,$3F,$C5,$8D,$DB,$E5,$7C,$3F,$DC,$37,$7D
	.dc.b	$0F,$E2,$3E,$E3,$87,$FC,$37,$87,$D3,$CB,$79,$C8,$F0,$90,$EF,$7C
	.dc.b	$E3,$B4,$7D,$BC,$38,$3A,$8B,$8B,$89,$2C,$FD,$11,$86,$C6,$70,$EC
	.dc.b	$FC,$5B,$E7,$5B,$9C,$CF,$73,$47,$F9,$74,$7C,$53,$D8,$FC,$FF,$C4
	.dc.b	$49,$D4,$E5,$B3,$FA,$4F,$88,$91,$82,$1F,$A6,$F7,$3F,$52,$7F,$B3
	.dc.b	$9E,$F3,$6B,$A4,$7E,$B2,$FC,$0F,$F9,$8C,$D9,$D3,$4F,$D1,$61,$E0
	.dc.b	$8D,$57,$29,$F9,$51,$F2,$A3,$19,$7E,$9F,$1F,$1C,$F3,$8F,$F7,$CF
	.dc.b	$E4,$4F,$1B,$FC,$F5,$3F,$BF,$11,$9C,$31,$18,$9D,$BF,$C1,$67,$4F
	.dc.b	$16,$B9,$8D,$BE,$06,$BF,$80,$8D,$B9,$9E,$F1,$88,$D4,$F0,$7C,$53
	.dc.b	$AF,$15,$F9,$5D,$24,$F8,$E3,$2D,$8C,$BE,$98,$CB,$63,$2F,$C6,$8C
	.dc.b	$B6,$33,$F9,$60,$80

;--------------------------------------------------------------------------------
;	FLOGN.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
flogn_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FLOGN.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FLOGN.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	flogn_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	flogn.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FLOGN.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	flogn.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FLOGN.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
flogn_data_compressed::
	.dc.b	$00,$02,$A1,$14,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$FD
	.dc.b	$4F,$E5,$1F,$97,$BE,$4B,$CF,$E9,$7A,$3D,$2D,$4B,$E1,$2F,$C9,$E3
	.dc.b	$E4,$73,$F1,$7A,$F8,$8D,$FB,$9C,$3D,$3E,$3E,$9F,$2F,$4F,$9F,$A7
	.dc.b	$D3,$D3,$EB,$E9,$F6,$F4,$FB,$FA,$7E,$0F,$4F,$C3,$E9,$F8,$BD,$3F
	.dc.b	$97,$EC,$7C,$CF,$63,$E6,$FB,$1F,$D8,$FF,$14,$FE,$40,$7E,$63,$C5
	.dc.b	$3F,$98,$7F,$B9,$3F,$D9,$9B,$BF,$1D,$F9,$6F,$CF,$7E,$9B,$F5,$DF
	.dc.b	$B6,$FD,$F7,$F0,$6F,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$FE,$ED,$F8,$AB
	.dc.b	$E6,$2F,$C7,$FE,$13,$DC,$FE,$F5,$F3,$9F,$47,$8F,$FC,$0D,$7F,$E7
	.dc.b	$B8,$FA,$11,$FE,$09,$7F,$DB,$87,$FA,$50,$FF,$03,$1F,$CF,$CF,$80
	.dc.b	$7B,$D3,$EF,$6F,$1F,$83,$8F,$C3,$C7,$E2,$E3,$F1,$F1,$F9,$38,$FC
	.dc.b	$BC,$7E,$6E,$3F,$3F,$1F,$A3,$8F,$D3,$C7,$EA,$E3,$FF,$0A,$77,$7B
	.dc.b	$F7,$FF,$10,$3F,$ED,$CE,$1F,$F2,$6B,$FE,$97,$2F,$C6,$F0,$F8,$AE
	.dc.b	$3F,$05,$CB,$DE,$E7,$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83
	.dc.b	$DE,$F0,$FB,$DE,$2F,$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$F5,$38,$3C
	.dc.b	$FE,$F7,$A3,$DE,$FF,$16,$7C,$EF,$F8,$14,$7C,$5E,$3E,$2F,$3F,$17
	.dc.b	$AF,$8B,$DF,$C5,$F0,$F8,$BE,$3F,$11,$CB,$E1,$79,$FC,$2F,$4F,$85
	.dc.b	$EB,$F0,$BD,$BE,$17,$BF,$C2,$F8,$3E,$17,$C3,$F0,$BE,$2F,$83,$F1
	.dc.b	$FC,$1F,$93,$DB,$F2,$FB,$7E,$6F,$6F,$CF,$EC,$7F,$90,$FE,$7B,$F1
	.dc.b	$5E,$61,$FC,$4B,$7C,$BD,$7C,$16,$FE,$0B,$87,$C1,$71,$F8,$2E,$5F
	.dc.b	$05,$CF,$E0,$BF,$33,$E8,$F5,$F7,$BB,$7B,$DD,$FD,$EF,$07,$BD,$E1
	.dc.b	$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD,$EF,$79,$FD,$EF
	.dc.b	$47,$BD,$FE,$5B,$F9,$FF,$C4,$73,$BD,$FF,$E7,$2A,$1F,$03,$BF,$8A
	.dc.b	$E1,$F2,$BC,$7D,$5E,$5C,$BC,$FE,$0B,$A7,$C1,$75,$F8,$2E,$DF,$05
	.dc.b	$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE,$0B,$C9,$F0,$5E
	.dc.b	$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$7B,$AF,$EF,$B5,$FD,$FE,$BF,$C0
	.dc.b	$35,$FE,$03,$AF,$B7,$E0,$BD,$DF,$05,$FE,$79,$F7,$FF,$CA,$47,$E3
	.dc.b	$BA,$3F,$D0,$CB,$E0,$34,$EF,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1C
	.dc.b	$FE,$03,$A7,$C0,$7F,$10,$E8,$ED,$F0,$1D,$FE,$03,$C1,$F0,$1F,$C5
	.dc.b	$3A,$FF,$8A,$F5,$FF,$16,$EB,$FE,$2F,$D7,$FC,$63,$AF,$F8,$CF,$5F
	.dc.b	$F1,$AE,$6F,$E3,$7C,$DF,$E9,$1F,$C5,$BF,$CD,$47,$A7,$9C,$FA,$5F
	.dc.b	$F5,$03,$FE,$FC,$3F,$D6,$C7,$F8,$98,$FE,$BC,$7F,$5D,$78,$F0,$F5
	.dc.b	$E1,$ED,$C1,$DF,$83,$E4,$7C,$0F,$C9,$F8,$1F,$95,$F0,$3F,$2F,$E0
	.dc.b	$7E,$67,$C0,$FC,$DF,$81,$F9,$DF,$03,$F3,$FE,$07,$E8,$7C,$0F,$FA
	.dc.b	$A7,$E8,$7F,$4E,$7E,$98,$7C,$0E,$7E,$07,$5F,$03,$BF,$81,$E1,$F0
	.dc.b	$3C,$7E,$07,$97,$C0,$F3,$F8,$1E,$9F,$03,$D7,$E0,$7B,$7C,$0F,$7F
	.dc.b	$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03,$F2,$FC,$9F,$99,$D7,$F3,$7A
	.dc.b	$FE,$77,$5F,$CF,$EB,$FA,$1D,$7F,$47,$AF,$E9,$75,$FD,$3E,$BF,$A9
	.dc.b	$D7,$FE,$C5,$F4,$FF,$81,$9F,$C0,$8F,$6F,$83,$FE,$D3,$5D,$7B,$EB
	.dc.b	$E1,$D7,$C7,$AF,$97,$5F,$3E,$BE,$9D,$7F,$27,$E0,$7E,$57,$C0,$FC
	.dc.b	$BF,$81,$F0,$77,$7C,$DF,$81,$FF,$6B,$7D,$17,$EF,$0F,$DC,$F0,$7F
	.dc.b	$B8,$D7,$BF,$AF,$7F,$7E,$FF,$0F,$7F,$8F,$BF,$CB,$DF,$E7,$EF,$F4
	.dc.b	$E7,$EB,$EF,$FB,$FC,$8F,$81,$E4,$7C,$1F,$23,$E1,$79,$1F,$0F,$C8
	.dc.b	$F1,$FB,$FE,$4F,$67,$CB,$EC,$F9,$BD,$9F,$3F,$B3,$E8,$F6,$7D,$3E
	.dc.b	$CF,$AB,$D9,$F5,$FB,$3E,$CF,$67,$FD,$E5,$F4,$7F,$EC,$B3,$F4,$47
	.dc.b	$E8,$7B,$0B,$DF,$DF,$BF,$C3,$DF,$E3,$EF,$F2,$F7,$FE,$C7,$B1,$D3
	.dc.b	$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F
	.dc.b	$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$F0,$DE,$A7
	.dc.b	$AB,$DC,$74,$4F,$3F,$FA,$59,$FD,$28,$FC,$4A,$7F,$12,$EF,$DB,$E2
	.dc.b	$E7,$DB,$B3,$FE,$33,$3F,$C2,$87,$F8,$4C,$7B,$1A,$F5,$B7,$E6,$70
	.dc.b	$F3,$38,$F9,$9C,$BC,$CE,$7E,$67,$4E,$CE,$BB,$F6,$F6,$3B,$FB,$1E
	.dc.b	$0F,$63,$C7,$F4,$9F,$D3,$0F,$C8,$F0,$9F,$C8,$BF,$01,$F6,$37,$D7
	.dc.b	$E3,$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$F6,$1A,$F8
	.dc.b	$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F,$5F
	.dc.b	$CF,$EB,$FA,$3D,$7F,$4F,$AF,$F3,$3E,$97,$E6,$F2,$9F,$E6,$07,$E3
	.dc.b	$CF,$C7,$3F,$69,$7A,$DC,$7D,$6E,$5E,$B7,$F0,$5E,$BE,$9E,$B7,$5F
	.dc.b	$5B,$B7,$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$78,$FD,$6F,$27
	.dc.b	$AD,$E5,$F5,$BC,$DE,$B7,$9F,$D6,$F4,$7A,$DE,$9F,$5B,$D5,$EB,$7E
	.dc.b	$1B,$E9,$BF,$C2,$E7,$F8,$58,$FC,$69,$F8,$CE,$72,$F5,$F8,$7A,$FC
	.dc.b	$7D,$EF,$E3,$DE,$77,$3F,$7B,$A7,$BD,$F9,$2F,$C5,$FF,$C1,$67,$E0
	.dc.b	$0F,$BF,$91,$FB,$ED,$F4,$38,$7A,$1C,$7D,$0E,$5E,$87,$3F,$43,$A7
	.dc.b	$A1,$D7,$D0,$ED,$E7,$F7,$F3,$FC,$1E,$7F,$87,$CF,$F1,$79,$FE,$3F
	.dc.b	$3F,$C9,$E7,$FC,$1F,$6B,$E1,$7B,$5F,$0F,$DA,$FF,$57,$FA,$7F,$EB
	.dc.b	$0F,$33,$FD,$63,$D3,$FE,$B7,$FC,$C3,$FC,$D9,$F2,$47,$F9,$5B,$7E
	.dc.b	$03,$5E,$F6,$FA,$78,$74,$F1,$E9,$E5,$D3,$CF,$A7,$A7,$4F,$5E,$9E
	.dc.b	$DD,$3D,$FA,$7C,$1D,$3F,$5F,$DE,$FB,$1E,$F7,$8F,$AF,$ED,$7B,$DF
	.dc.b	$6F,$DE,$FB,$9E,$F7,$DD,$F7,$BE,$F7,$4F,$DF,$E9,$FC,$07,$4F,$F0
	.dc.b	$6F,$82,$FF,$1F,$9F,$B4,$3E,$A9,$F5,$07,$36,$F9,$B8,$73,$71,$E6
	.dc.b	$E5,$CD,$CF,$9B,$A7,$37,$5E,$6E,$DC,$DD,$F9,$BC,$1C,$9E,$1F,$03
	.dc.b	$C5,$E0,$78,$FC,$0F,$27,$BD,$E5,$F7,$BC,$DE,$F7,$F1,$CF,$A9,$FD
	.dc.b	$89,$FE,$1F,$3F,$9F,$1F,$B0,$F1,$4F,$EC,$1F,$E5,$6B,$DD,$E5,$EE
	.dc.b	$F3,$F7,$7A,$7B,$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF
	.dc.b	$17,$BB,$E3,$F7,$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE
	.dc.b	$FA,$7D,$DF,$E6,$BF,$55,$FE,$86,$3F,$D0,$A7,$F3,$C3,$F6,$07,$EB
	.dc.b	$F9,$7F,$A1,$AF,$73,$8F,$B9,$CB,$DC,$E7,$EE,$74,$F6,$FA,$FB,$7D
	.dc.b	$BD,$BE,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F
	.dc.b	$2F,$AF,$E6,$F5,$FC,$FE,$BF,$F4,$CF,$AB,$FF,$42,$9F,$CE,$8F,$DD
	.dc.b	$9F,$BA,$3F,$0F,$C9,$FD,$69,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7
	.dc.b	$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F
	.dc.b	$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5,$BF,$B1,$FC,$07,$F3,$83
	.dc.b	$FC,$CC,$7F,$81,$4F,$C4,$1F,$87,$E4,$FE,$6F,$EB,$70,$F5,$B8,$FA
	.dc.b	$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD,$DB,$D6,$EF,$EB,$78,$3D,$6F
	.dc.b	$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6,$F2,$FA,$DE,$6F,$5B,$CF,$EB
	.dc.b	$7F,$7A,$FA,$CF,$F6,$59,$FE,$CA,$3F,$C0,$27,$F3,$53,$F0,$DC,$87
	.dc.b	$F0,$CF,$F8,$29,$7A,$BC,$FD,$5E,$9E,$AF,$5F,$57,$B7,$AB,$DF,$D5
	.dc.b	$F0,$7A,$BE,$1F,$57,$C5,$EA,$F8,$FD,$5F,$27,$AB,$E5,$F5,$7C,$DE
	.dc.b	$AF,$9F,$D5,$F4,$7A,$BF,$E1,$2F,$AD,$FF,$65,$9F,$E0,$03,$FA,$09
	.dc.b	$FD,$00,$FC,$3F,$19,$FC,$3B,$FE,$22,$7F,$DA,$23,$D3,$E9,$E9,$F5
	.dc.b	$F4,$FB,$7A,$7D,$FD,$3F,$07,$A7,$E1,$F4,$FC,$5E,$9F,$8F,$D3,$F2
	.dc.b	$7A,$7E,$5F,$4F,$CD,$E9,$F9,$FD,$3F,$47,$A7,$FE,$2A,$FA,$EF,$EF
	.dc.b	$E7,$FA,$B4,$FF,$26,$1F,$C3,$CF,$C4,$1F,$C3,$78,$BF,$C7,$2B,$D2
	.dc.b	$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9,$78
	.dc.b	$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5E,$7F,$4B
	.dc.b	$D1,$E9,$7F,$90,$BD,$B7,$CC,$5F,$97,$C7,$C5,$E7,$DF,$D7,$C2,$EF
	.dc.b	$DE,$E1,$EC,$71,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1
	.dc.b	$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6
	.dc.b	$3C,$DE,$C7,$F9,$63,$EB,$DF,$2F,$7E,$47,$1F,$0B,$9F,$85,$D7,$66
	.dc.b	$FE,$0B,$87,$C1,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7
	.dc.b	$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$E5
	.dc.b	$F7,$BC,$DE,$F7,$F9,$BB,$EC,$3F,$C8,$67,$FB,$C4,$FF,$48,$1F,$D4
	.dc.b	$CF,$E1,$87,$E1,$7D,$F3,$F8,$57,$FC,$FE,$FF,$BC,$47,$B9,$D7,$DC
	.dc.b	$ED,$EE,$77,$F7,$3C,$1E,$E7,$87,$DC,$F1,$7B,$9E,$3F,$73,$C9,$EE
	.dc.b	$79,$7D,$CF,$37,$B9,$E7,$F7,$3D,$1E,$E7,$A7,$DC,$FF,$43,$FD,$8F
	.dc.b	$E1,$B8,$BF,$D2,$2B,$DC,$CF,$B9,$AF,$73,$7E,$E7,$0F,$73,$8F,$B9
	.dc.b	$CB,$DC,$E7,$EE,$74,$F7,$3A,$FB,$9D,$BD,$CE,$FE,$E7,$83,$DC,$F0
	.dc.b	$FB,$9E,$2F,$73,$C7,$EE,$79,$3D,$CF,$2F,$B9,$E6,$F7,$3C,$FE,$E7
	.dc.b	$FA,$7B,$8B,$F0,$FC,$FF,$EA,$87,$FD,$1A,$7F,$93,$8F,$F2,$69,$FC
	.dc.b	$38,$FE,$1A,$FF,$96,$FD,$5E,$5E,$AF,$3F,$57,$A7,$AB,$D7,$D5,$ED
	.dc.b	$EA,$F7,$F5,$7C,$1E,$AF,$87,$D5,$F1,$7A,$BE,$3F,$57,$C9,$EA,$F9
	.dc.b	$7D,$5F,$37,$AB,$E7,$F5,$7F,$D6,$BF,$65,$FE,$8A,$1F,$31,$8F,$95
	.dc.b	$CF,$C8,$EB,$E2,$37,$F1,$1C,$3E,$17,$8F,$C2,$F2,$F7,$39,$FB,$9D
	.dc.b	$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC,$F1
	.dc.b	$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$FE,$CC,$FB,$3F,$C3
	.dc.b	$79,$07,$F0,$CF,$FB,$61,$FF,$6A,$9F,$E8,$93,$FA,$A9,$FD,$51,$E1
	.dc.b	$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D
	.dc.b	$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1
	.dc.b	$EC,$7F,$B9,$3E,$D3,$F0,$FC,$47,$FD,$A6,$7F,$A5,$0F,$F4,$99,$FD
	.dc.b	$50,$FC,$38,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1,$DF
	.dc.b	$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6,$3C
	.dc.b	$DE,$C7,$9F,$D8,$FF,$7D,$FA,$7F,$ED,$03,$F0,$FC,$63,$FC,$B9,$EB
	.dc.b	$6B,$D6,$DF,$AD,$C2,$B8,$FA,$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD
	.dc.b	$DB,$D6,$EF,$EB,$78,$3D,$6F,$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6
	.dc.b	$F2,$FA,$DE,$6F,$5B,$CF,$EB,$7A,$3D,$6F,$37,$33,$ED,$7F,$12,$7E
	.dc.b	$21,$C7,$BB,$A8,$FF,$68,$1F,$E1,$C3,$F1,$07,$F8,$6A,$3E,$02,$F5
	.dc.b	$B8,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77
	.dc.b	$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B,$C9,$EB,$7A,$BE,$DB
	.dc.b	$F1,$23,$E1,$B1,$F0,$D9,$F8,$6D,$7C,$36,$FE,$1B,$87,$B7,$C7,$DB
	.dc.b	$E5,$ED,$F3,$F6,$FA,$7B,$7D,$7D,$BE,$DE,$DF,$7F,$6F,$C1,$ED,$F8
	.dc.b	$7D,$BF,$17,$B7,$E3,$F6,$FC,$9E,$DF,$97,$DB,$F3,$7B,$7F,$43,$ED
	.dc.b	$FE,$8F,$8E,$7F,$12,$DF,$73,$5E,$E6,$FD,$CE,$1E,$E7,$1F,$73,$97
	.dc.b	$B9,$CF,$DC,$E9,$EE,$75,$F7,$3B,$7B,$9D,$FD,$CF,$07,$B9,$E1,$F7
	.dc.b	$3C,$5E,$E7,$8F,$DC,$F2,$7B,$9E,$5F,$73,$CD,$EE,$79,$FD,$CF,$47
	.dc.b	$B9,$F8,$8F,$1F,$F8,$79,$F8,$71,$EE,$67,$E1,$35,$F0,$7B,$F8,$4E
	.dc.b	$1F,$03,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D,$FE,$DE,$FF,$7F
	.dc.b	$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E,$FF,$97,$DF,$F3
	.dc.b	$7B,$FF,$9E,$FB,$8F,$E1,$A7,$E1,$7C,$BF,$DB,$2F,$7B,$5E,$F6,$FD
	.dc.b	$EE,$1D,$5C,$7D,$EE,$5E,$F7,$3F,$7B,$A7,$BD,$D7,$DE,$ED,$EF,$77
	.dc.b	$F7,$BC,$1E,$F7,$87,$DE,$F1,$7B,$DE,$3F,$7B,$C9,$EF,$79,$7D,$EF
	.dc.b	$37,$BD,$E7,$F7,$BF,$6B,$F7,$3F,$86,$3F,$0A,$3D,$EC,$FB,$DA,$F7
	.dc.b	$B7,$EF,$70,$F7,$B8,$FB,$DC,$BD,$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB
	.dc.b	$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD,$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2
	.dc.b	$FB,$DE,$6F,$7B,$CF,$EF,$7F,$0C,$FB,$AF,$EA,$67,$F0,$B3,$F0,$BE
	.dc.b	$67,$F2,$55,$EF,$6F,$DE,$E1,$EF,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E
	.dc.b	$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD
	.dc.b	$EF,$27,$BD,$E5,$F7,$BC,$DE,$F7,$9F,$DE,$FE,$47,$C3,$FD,$48,$FC
	.dc.b	$37,$2F,$FB,$F7,$DD,$CF,$BB,$AF,$77,$7E,$EF,$0F,$77,$8F,$BB,$CB
	.dc.b	$DD,$E7,$EE,$F4,$F7,$7A,$FB,$BD,$BD,$DE,$FE,$EF,$83,$DD,$F0,$FB
	.dc.b	$BE,$2F,$77,$C7,$EE,$F9,$3D,$DF,$2F,$BB,$E6,$F7,$7F,$9C,$FD,$DF
	.dc.b	$F0,$A3,$F8,$49,$F8,$5E,$6F,$EA,$4B,$DC,$DF,$B9,$C3,$DC,$E3,$EE
	.dc.b	$72,$F7,$39,$FB,$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C
	.dc.b	$3E,$E7,$8B,$DC,$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9
	.dc.b	$E8,$F7,$3F,$A8,$FD,$E7,$F8,$60,$FE,$A0,$7E,$1B,$9C,$FE,$19,$FE
	.dc.b	$D8,$FF,$BB,$0F,$F0,$C1,$FD,$45,$DF,$B3,$D3,$D9,$EB,$EC,$F6,$F6
	.dc.b	$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE
	.dc.b	$CF,$9B,$D9,$F3,$FB,$3E,$8F,$67,$FB,$4F,$DE,$FF,$86,$0F,$C3,$9F
	.dc.b	$C2,$F8,$7F,$C0,$4B,$D8,$DF,$B1,$C3,$D8,$E3,$EC,$72,$F6,$39,$FB
	.dc.b	$1D,$3D,$8E,$BE,$C7,$6F,$63,$BF,$B1,$E0,$F6,$3C,$3E,$C7,$8B,$D8
	.dc.b	$F1,$FB,$1E,$4F,$63,$CB,$EC,$79,$BD,$8F,$3F,$B1,$FE,$00,$F5,$3F
	.dc.b	$A8,$9F,$D4,$0F,$C4,$73,$7F,$B9,$7D,$7D,$7A,$FB,$F5,$F8,$57,$1F
	.dc.b	$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F,$07,$AF
	.dc.b	$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB,$F9,$FD
	.dc.b	$7F,$F0,$BF,$DF,$7F,$97,$8F,$F0,$B1,$F8,$73,$F0,$DC,$C5,$EE,$F0
	.dc.b	$F7,$38,$FB,$BC,$BD,$AE,$7E,$D7,$4F,$6B,$AF,$B5,$DB,$DA,$EF,$ED
	.dc.b	$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7,$93,$DA,$F2,$FB,$5E,$6F
	.dc.b	$6B,$CF,$ED,$7F,$8B,$BE,$FF,$FC,$BA,$7F,$0C,$3F,$0D,$CA,$7F,$0C
	.dc.b	$FF,$92,$57,$B3,$C7,$D9,$E5,$EC,$F3,$F6,$7A,$7B,$3D,$7D,$97,$F0
	.dc.b	$0C,$2E,$5C,$E1,$F8,$07,$F9,$57,$E1,$9C,$FC,$63,$0D,$AF,$EF,$6D
	.dc.b	$9D,$C7,$39,$DB,$4D,$96,$36,$FB,$6B,$C2,$E8,$C4,$F2,$69,$9D,$53
	.dc.b	$C2,$B9,$C6,$2B,$83,$6B,$85,$6E,$FD,$04,$8F,$23,$5E,$F7,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF
	.dc.b	$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63
	.dc.b	$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF
	.dc.b	$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88
	.dc.b	$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A
	.dc.b	$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB
	.dc.b	$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B
	.dc.b	$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D
	.dc.b	$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B
	.dc.b	$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60
	.dc.b	$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC
	.dc.b	$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE
	.dc.b	$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78
	.dc.b	$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12
	.dc.b	$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71
	.dc.b	$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C
	.dc.b	$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C
	.dc.b	$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6
	.dc.b	$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09
	.dc.b	$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC
	.dc.b	$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE
	.dc.b	$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A
	.dc.b	$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12
	.dc.b	$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D
	.dc.b	$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12
	.dc.b	$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2
	.dc.b	$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C
	.dc.b	$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09
	.dc.b	$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA
	.dc.b	$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB
	.dc.b	$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E
	.dc.b	$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD
	.dc.b	$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9
	.dc.b	$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A
	.dc.b	$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46
	.dc.b	$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31
	.dc.b	$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9
	.dc.b	$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26
	.dc.b	$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12
	.dc.b	$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3
	.dc.b	$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C
	.dc.b	$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F
	.dc.b	$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18
	.dc.b	$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A
	.dc.b	$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C
	.dc.b	$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67
	.dc.b	$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB
	.dc.b	$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D
	.dc.b	$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38
	.dc.b	$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E
	.dc.b	$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97
	.dc.b	$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D
	.dc.b	$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57
	.dc.b	$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA
	.dc.b	$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3
	.dc.b	$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF
	.dc.b	$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17
	.dc.b	$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF
	.dc.b	$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E
	.dc.b	$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D
	.dc.b	$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35
	.dc.b	$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1
	.dc.b	$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D
	.dc.b	$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4
	.dc.b	$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F
	.dc.b	$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D2,$12,$95,$2F,$10,$BE,$15,$BA,$D4,$F1,$9C,$BC,$AB
	.dc.b	$8C,$62,$77,$3C,$EF,$C2,$EC,$E4,$E2,$37,$1E,$5F,$BB,$F3,$7C,$9B
	.dc.b	$E2,$BC,$B3,$91,$19,$73,$3C,$27,$70,$8C,$3A,$38,$C6,$A3,$31,$F0
	.dc.b	$9A,$F7,$9E,$CE,$4E,$27,$37,$2E,$E7,$A3,$A3,$84,$69,$F2,$19,$38
	.dc.b	$1C,$E3,$8C,$6E,$7C,$73,$C2,$30,$F1,$31,$3C,$08,$F3,$1F,$7A,$FC
	.dc.b	$E5,$E5,$BE,$5F,$F0,$9F,$80,$DB,$DD,$3C,$DC,$4E,$70,$E5,$86,$30
	.dc.b	$FC,$64,$75,$AC,$61,$CB,$CD,$B3,$E1,$6E,$2F,$36,$C6,$3C,$D9,$8F
	.dc.b	$90,$EB,$4F,$37,$38,$B4,$F2,$4B,$CD,$EB,$E6,$F6,$6E,$1C,$EF,$E1
	.dc.b	$FE,$75,$F0,$63,$1A,$A1,$F2,$DE,$87,$55,$DE,$BD,$F1,$A7,$8B,$FA
	.dc.b	$C6,$F2,$71,$F8,$BE,$4F,$3F,$3F,$A7,$9F,$6B,$8B,$DF,$CF,$F0,$79
	.dc.b	$F6,$7A,$BE,$3F,$3E,$47,$83,$CF,$BA,$79,$4F,$6E,$3D,$A7,$CA,$D8
	.dc.b	$D6,$DE,$FC,$7B,$D7,$A9,$B3,$AD,$58,$F6,$E2,$FC,$14,$32,$C6,$1F
	.dc.b	$13,$2E,$D5,$A6,$C7,$C8,$C7,$AC,$F2,$6C,$7C,$CC,$39,$E3,$F5,$B0
	.dc.b	$B9,$7D,$5E,$EF,$A9,$E7,$FD,$56,$E1,$F4,$BC,$FF,$B0,$7D,$9F,$3E
	.dc.b	$C7,$D6,$F3,$FE,$E1,$F7,$7C,$FB,$1F,$6A,$F9,$E7,$F0,$5B,$B7,$C0
	.dc.b	$C7,$9C,$FC,$5C,$6C,$FC,$4C,$79,$CF,$C9,$C6,$CF,$C8,$C3,$9E,$3F
	.dc.b	$8A,$6D,$7E,$53,$1E,$73,$F8,$EC,$6C,$FE,$63,$1E,$73,$F9,$2C,$6C
	.dc.b	$FE,$73,$0E,$79,$FC,$7A,$FC,$AE,$7C,$A3,$F3,$D9,$7C,$59,$F9,$07
	.dc.b	$C5,$25,$FA,$5B,$9F,$E0,$3C,$FF,$D2,$79,$FF,$A5,$6C,$FE,$8B,$CF
	.dc.b	$FD,$41,$FA,$9F,$3E,$C7,$E9,$BC,$FF,$D6,$1F,$C1,$3E,$0F,$F9,$DC
	.dc.b	$64,$FD,$5D,$F4,$CB,$F2,$4D,$CB,$F2,$38,$F4,$9F,$CB,$63,$67,$F2
	.dc.b	$B8,$F6,$C3,$F7,$B8,$F6,$BF,$CE,$4F,$A6,$7F,$73,$95,$9F,$D0,$3C
	.dc.b	$E3,$F7,$F8,$FA,$DF,$71,$E7,$9E,$6F,$6F,$1B,$C1,$1F,$AE,$C3,$FA
	.dc.b	$D4,$F8,$BC,$6F,$81,$F1,$7F,$05,$F8,$5F,$11,$F0,$FD,$2B,$5E,$87
	.dc.b	$E3,$7A,$5F,$1F,$D2,$B3,$EC,$7E,$57,$A5,$23,$DF,$E9,$5D,$3C,$5E
	.dc.b	$7F,$CF,$F8,$8F,$A0,$DA,$F8,$BF,$11,$F4,$CF,$A9,$F1,$16,$3E,$4F
	.dc.b	$07,$D7,$F8,$8F,$B0,$D8,$F9,$BF,$11,$F6,$FE,$23,$EE,$36,$3E,$8F
	.dc.b	$C4,$7D,$F3,$F0,$1F,$11,$63,$EA,$FC,$47,$E0,$FE,$23,$F0,$8D,$8F
	.dc.b	$E6,$BF,$9D,$FF,$86,$5B,$C1,$FD,$32,$1F,$2D,$63,$1F,$1F,$2B,$8F
	.dc.b	$C5,$D9,$FE,$43,$6F,$8B,$E4,$73,$F8,$BB,$7E,$0F,$8B,$ED,$F1,$7D
	.dc.b	$DB,$1E,$4F,$8B,$F1,$7C,$5F,$8D,$B1,$E8,$F8,$BF,$31,$E7,$F8,$BB
	.dc.b	$1E,$CF,$8B,$F5,$7C,$5F,$AD,$B1,$F0,$3E,$1B,$DD,$F0,$DE,$F6,$C7
	.dc.b	$C4,$F8,$5F,$84,$7C,$3F,$86,$B1,$F2,$3E,$1B,$E3,$7C,$37,$C7,$6C
	.dc.b	$7C,$CF,$84,$F9,$47,$CB,$F8,$5B,$1F,$D9,$3F,$3C,$FE,$C3,$E9,$FF
	.dc.b	$80,$5B,$1F,$E0,$07,$FB,$17,$47,$F4,$C7,$2F,$F8,$41,$FE,$C8,$E7
	.dc.b	$E1,$3F,$C0,$70,$F8,$5B,$18,$F8,$2C,$FC,$2D,$C3,$C4,$F0,$F8,$39
	.dc.b	$1E,$5F,$85,$B9,$78,$7E,$13,$A9,$DB,$E1,$2D,$7A,$FE,$07,$C3,$F0
	.dc.b	$3E,$26,$C7,$BF,$E0,$3C,$BF,$09,$E6,$6C,$7C,$3F,$84,$F4,$9E,$AF
	.dc.b	$84,$B1,$F1,$FE,$0F,$DA,$7B,$BE,$06,$C7,$CB,$F8,$1F,$82,$7C,$2F
	.dc.b	$83,$B1,$F3,$FE,$0F,$E2,$9F,$E2,$8F,$CF,$7F,$8C,$74,$FD,$18,$C7
	.dc.b	$C1,$4A,$FF,$0C,$D3,$E0,$F1,$C9,$FD,$FA,$1F,$07,$CF,$E0,$FA,$36
	.dc.b	$F8,$FC,$1F,$7F,$83,$F0,$36,$3A,$FC,$1F,$8C,$F2,$7C,$1D,$8F,$0F
	.dc.b	$C1,$79,$CF,$47,$C1,$D8,$F2,$FC,$1F,$AF,$E0,$FD,$8D,$8F,$4F,$C1
	.dc.b	$7B,$FE,$0F,$E0,$36,$3D,$BF,$05,$F0,$FE,$0F,$E2,$36,$3E,$0F,$C1
	.dc.b	$FC,$7F,$83,$F9,$0D,$8F,$8B,$F0,$5F,$2F,$E0,$FE,$63,$63,$FC,$83
	.dc.b	$E1,$7F,$0D,$E7,$FF,$12,$BF,$C3,$23,$31,$FE,$5A,$8C,$3A,$F8,$2C
	.dc.b	$7C,$16,$5B,$5D,$3E,$07,$87,$C0,$F1,$6C,$78,$3E,$03,$A1,$D7,$E0
	.dc.b	$6C,$79,$3E,$03,$C1,$F0,$3E,$16,$C7,$A3,$E0,$7C,$9F,$03,$E5,$6C
	.dc.b	$7B,$3E,$03,$D1,$F0,$3E,$96,$C7,$C0,$F8,$0F,$67,$C0,$FB,$5B,$1F
	.dc.b	$13,$E0,$7E,$07,$C0,$FC,$16,$C7,$C8,$F8,$0F,$88,$7C,$5F,$81,$B1
	.dc.b	$F3,$3E,$07,$E4,$7C,$0F,$C9,$6C,$7F,$9C,$3F,$3E,$FE,$0F,$F0,$C6
	.dc.b	$9F,$C3,$3D,$1F,$F4,$83,$8F,$81,$FF,$45,$BF,$C2,$6B,$94,$63,$E0
	.dc.b	$B3,$F0,$52,$7B,$7C,$15,$EE,$FF,$30,$E0,$F8,$2E,$7F,$05,$D1,$B7
	.dc.b	$E5,$F8,$2E,$E7,$83,$E0,$AC,$7A,$7E,$07,$C6,$79,$3E,$0A,$C7,$B7
	.dc.b	$E0,$BC,$FE,$37,$A1,$B1,$F0,$7E,$0B,$D7,$F0,$5E,$C6,$C7,$C5,$F8
	.dc.b	$2F,$79,$F0,$3E,$0A,$C7,$C9,$F8,$2F,$86,$7C,$4F,$82,$B1,$D4,$F9
	.dc.b	$D8,$75,$69,$D4,$9C,$5F,$12,$FA,$BF,$1A,$F9,$BF,$22,$EF,$3F,$D4
	.dc.b	$F8,$2F,$A6,$7F,$AF,$BF,$3F,$FE,$2F,$8F,$FB,$13,$95,$39,$F8,$3D
	.dc.b	$7C,$1D,$AF,$F5,$8D,$FF,$19,$6F,$C0,$F3,$F8,$2E,$8D,$8D,$FC,$0F
	.dc.b	$73,$C1,$F0,$56,$39,$FC,$17,$8F,$E0,$BC,$8D,$8E,$FF,$05,$E7,$F8
	.dc.b	$2F,$43,$63,$C7,$F0,$3E,$B3,$D9,$F0,$56,$3C,$FF,$03,$EF,$3E,$07
	.dc.b	$C1,$58,$F5,$FC,$0F,$C3,$F8,$2F,$88,$D8,$F7,$FC,$0F,$C7,$3E,$47
	.dc.b	$C1,$58,$EA,$4E,$AF,$F5,$BF,$23,$1E,$D1,$8C,$3A,$36,$BE,$5E,$3F
	.dc.b	$4F,$E0,$64,$7C,$EF,$82,$B5,$D3,$4E,$7D,$7F,$EF,$DC,$FB,$5F,$86
	.dc.b	$31,$97,$66,$C7,$63,$FD,$D3,$FA,$07,$F2,$98,$43,$07,$9E,$AF,$E5
	.dc.b	$35,$A9,$CC,$71,$AE,$15,$BF,$87,$D5,$EE,$39,$7C,$1E,$2B,$76,$F8
	.dc.b	$4C,$63,$D2,$AD,$F8,$66,$D7,$F9,$06,$FD,$3F,$07,$8F,$84,$CB,$AF
	.dc.b	$84,$DF,$C2,$5A,$F7,$BC,$B2,$E7,$8D,$9F,$86,$F2,$7A,$9F,$1B,$0C
	.dc.b	$67,$FE,$F2,$C9,$8F,$6A,$F9,$6D,$9D,$61,$DA,$7E,$7B,$63,$96,$5D
	.dc.b	$34,$F8,$D8,$31,$C6,$1F,$1B,$E0,$FE,$0F,$C2,$7C,$26,$C7,$43,$D2
	.dc.b	$5C,$7C,$DA,$72,$7E,$BE,$5F,$1E,$7D,$B1,$EB,$AE,$6D,$8F,$C2,$DF
	.dc.b	$C5,$8E,$71,$D7,$2F,$3E,$1D,$76,$FC,$36,$3D,$6F,$DB,$7F,$8E,$F8
	.dc.b	$3F,$B3,$F0,$9F,$69,$B1,$CC,$FC,$1E,$1D,$F1,$EB,$1E,$67,$E2,$E1
	.dc.b	$D5,$B1,$F9,$AF,$83,$FC,$2F,$C2,$7E,$19,$B1,$F9,$EF,$83,$FC,$49
	.dc.b	$F8,$AF,$84,$B1,$FA,$2F,$83,$FC,$6F,$C2,$7E,$39,$B1,$FA,$6F,$84
	.dc.b	$FC,$8F,$C2,$7E,$49,$B1,$D4,$9C,$1F,$85,$C3,$AB,$F8,$7B,$E6,$D9
	.dc.b	$FD,$8E,$1F,$9C,$F8,$3F,$CE,$FC,$25,$9F,$DA,$BC,$B6,$FC,$2E,$9E
	.dc.b	$99,$EF,$1F,$88,$C7,$AB,$63,$A9,$FB,$B7,$F6,$53,$FD,$E0,$FE,$ED
	.dc.b	$1F,$B8,$71,$F1,$5F,$B0,$7F,$D0,$33,$C3,$E1,$31,$F0,$B9,$75,$F0
	.dc.b	$B3,$0E,$C7,$3F,$85,$C6,$CC,$3C,$3F,$09,$D7,$E1,$7B,$36,$F8,$D7
	.dc.b	$F1,$0C,$DF,$0B,$E3,$F8,$5F,$23,$6B,$7D,$8A,$7C,$2F,$A4,$F5,$7C
	.dc.b	$2D,$AD,$CB,$E1,$3D,$DF,$0B,$EF,$6C,$EB,$E1,$7E,$17,$C2,$7C,$36
	.dc.b	$C7,$2F,$85,$F8,$C7,$C7,$F8,$5B,$1D,$BE,$17,$E5,$1F,$2F,$E1,$6C
	.dc.b	$78,$BE,$13,$F8,$EF,$E8,$3F,$DC,$38,$FE,$04,$63,$E2,$B3,$F0,$D2
	.dc.b	$FD,$F7,$FD,$BE,$FE,$64,$7F,$30,$84,$F4,$F8,$49,$3C,$3E,$16,$DF
	.dc.b	$03,$F9,$ED,$BE,$17,$C6,$79,$3E,$16,$DF,$83,$E1,$3C,$E7,$A3,$E1
	.dc.b	$6C,$79,$3E,$13,$D7,$F0,$BE,$C6,$C7,$A3,$E1,$7D,$FF,$0B,$F0,$1B
	.dc.b	$1E,$CF,$7F,$E1,$9F,$13,$E0,$EC,$7C,$0F,$7F,$E3,$FC,$1F,$C8,$6C
	.dc.b	$75,$3E,$2E,$1D,$5A,$75,$27,$07,$C1,$F0,$5C,$3F,$DA,$23,$0B,$2E
	.dc.b	$D5,$F4,$31,$B5,$A7,$B3,$C3,$0B,$E3,$7F,$5F,$C1,$1E,$A9,$EB,$87
	.dc.b	$B7,$0F,$1D,$EF,$E1,$7F,$07,$F0,$BF,$84,$6C,$FE,$03,$E1,$7F,$0F
	.dc.b	$F0,$BF,$88,$6C,$7E,$8B,$F4,$2F,$EE,$59,$7F,$78,$7F,$BB,$C7,$F7
	.dc.b	$47,$3B,$1C,$CE,$FE,$0B,$FB,$7C,$1B,$EE,$61,$CF,$B9,$AF,$81,$B8
	.dc.b	$F8,$5E,$3F,$07,$23,$C9,$F0,$97,$2F,$A1,$93,$E0,$FB,$9E,$0F,$83
	.dc.b	$B7,$EB,$CB,$B3,$4E,$C7,$BF,$E0,$BC,$E7,$A3,$E0,$EC,$7C,$3F,$82
	.dc.b	$F5,$9E,$CF,$83,$B1,$CF,$3F,$1E,$DE,$A6,$C7,$CB,$F4,$7E,$19,$F1
	.dc.b	$3E,$06,$C7,$CF,$F7,$BE,$39,$F2,$3E,$06,$C7,$D3,$F8,$2F,$97,$F0
	.dc.b	$7F,$31,$B1,$F5,$FE,$0F,$E7,$9F,$43,$E0,$EC,$7F,$1C,$FD,$0F,$FD
	.dc.b	$5A,$4F,$84,$FF,$11,$BF,$EB,$0E,$66,$1D,$36,$F9,$AD,$F0,$3F,$C6
	.dc.b	$15,$FC,$EB,$0F,$F1,$AB,$FE,$B1,$AF,$D1,$39,$7F,$C7,$EE,$67,$8C
	.dc.b	$F8,$0F,$0E,$15,$F7,$1F,$91,$CF,$DC,$D3,$D5,$F0,$9E,$8F,$84,$F4
	.dc.b	$B6,$3D,$DF,$09,$EC,$3D,$BF,$09,$63,$AE,$7F,$88,$C3,$AB,$63,$A9
	.dc.b	$F1,$F0,$EA,$D3,$E5,$7C,$1F,$C8,$3E,$4F,$C2,$58,$EB,$A7,$F9,$57
	.dc.b	$76,$3D,$A3,$0F,$D0,$F8,$39,$5F,$57,$E1,$2E,$7F,$2B,$E1,$3E,$A9
	.dc.b	$F5,$BE,$12,$D7,$63,$F6,$58,$75,$69,$F7,$F0,$E7,$8F,$D8,$C6,$CF
	.dc.b	$D8,$C3,$9E,$3F,$63,$1B,$7E,$9E,$3F,$C0,$5F,$EF,$DB,$5F,$73,$1E
	.dc.b	$6F,$DE,$C6,$F8,$CE,$4F,$EC,$98,$7E,$47,$E0,$FF,$1F,$F0,$9F,$90
	.dc.b	$7D,$18,$7F,$0D,$DA,$E1,$FC,$2D,$FC,$AF,$C1,$FE,$5B,$E1,$2E,$3F
	.dc.b	$9B,$7F,$33,$F0,$9F,$9A,$F8,$4B,$3F,$9F,$7F,$3B,$F0,$92,$3F,$45
	.dc.b	$F0,$97,$3F,$CB,$FC,$27,$FA,$13,$D9,$FE,$D1,$AF,$A6,$31,$F0,$99
	.dc.b	$F8,$4D,$7C,$45,$CB,$FC,$0F,$87,$F6,$9E,$7F,$D2,$E9,$D5,$FD,$3E
	.dc.b	$96,$BF,$D3,$74,$F8,$5F,$07,$C2,$F8,$5B,$3D,$3E,$17,$C9,$F0,$BE
	.dc.b	$56,$C7,$EE,$F2,$E9,$C1,$FA,$EC,$AF,$07,$FB,$17,$E1,$7D,$87,$B7
	.dc.b	$E1,$6E,$1E,$8F,$84,$F8,$1F,$0B,$F0,$5B,$1E,$CF,$85,$F8,$9F,$05
	.dc.b	$F1,$5B,$1F,$03,$E0,$BE,$47,$17,$C9,$6C,$7C,$4F,$85,$F9,$9F,$0B
	.dc.b	$F3,$5B,$1F,$23,$E0,$FE,$87,$C2,$7D,$16,$C7,$CC,$F8,$3F,$A8,$7D
	.dc.b	$5F,$84,$B1,$FE,$36,$FD,$13,$FA,$89,$3E,$23,$FD,$7A,$E3,$E1,$6F
	.dc.b	$C3,$61,$B5,$C7,$E0,$B6,$70,$F7,$AC,$75,$F8,$6E,$67,$4F,$86,$B1
	.dc.b	$E1,$F8,$6E,$E7,$83,$E1,$AC,$79,$7E,$1B,$C7,$F0,$DE,$46,$C7,$A7
	.dc.b	$E1,$BC,$FF,$0D,$E8,$6C,$7B,$7E,$1B,$D7,$F0,$DE,$C6,$C7,$C1,$F8
	.dc.b	$6F,$7F,$C3,$7C,$06,$C7,$C5,$F6,$BE,$19,$F1,$3D,$5B,$1F,$27,$E1
	.dc.b	$BE,$39,$F2,$3E,$1A,$C7,$F9,$8B,$F4,$5F,$F7,$CB,$4F,$F7,$4B,$FE
	.dc.b	$7E,$F2,$BF,$DF,$6E,$9F,$F1,$13,$FD,$07,$C9,$34,$FE,$85,$3F,$EF
	.dc.b	$C7,$1F,$0B,$D0,$EB,$F0,$B6,$7C,$5F,$09,$E0,$F8,$4F,$0B,$63,$CD
	.dc.b	$F0,$3E,$4F,$81,$F2,$B6,$3D,$5F,$07,$E8,$F8,$3F,$4B,$63,$DD,$F0
	.dc.b	$3E,$CF,$82,$F6,$B6,$3E,$17,$C1,$7C,$0F,$82,$F8,$2D,$8F,$8D,$F0
	.dc.b	$5F,$10,$F8,$BF,$05,$63,$E5,$7B,$3F,$20,$F9,$3E,$CD,$8F,$9D,$F0
	.dc.b	$5F,$30,$3F,$E9,$2F,$D1,$BF,$89,$F2,$A7,$3C,$FF,$E4,$F9,$FF,$3A
	.dc.b	$D6,$FE,$07,$3F,$03,$AE,$95,$6A,$F8,$56,$AD,$0A,$DF,$C0,$66,$31
	.dc.b	$7F,$C5,$EB,$C9,$F0,$39,$BD,$C7,$9F,$E0,$31,$4F,$81,$B1,$87,$3F
	.dc.b	$03,$AF,$81,$B7,$EF,$78,$FC,$07,$2F,$81,$B3,$F0,$DE,$BF,$01,$DB
	.dc.b	$E0,$6C,$FC,$77,$C3,$F0,$12,$3E,$57,$C0,$DE,$0F,$91,$F0,$1E,$6F
	.dc.b	$81,$F3,$B6,$BE,$87,$C0,$7A,$8F,$5F,$C0,$D8,$FA,$9F,$03,$EE,$F8
	.dc.b	$1F,$7B,$63,$FD,$93,$E0,$7F,$99,$09,$F0,$3F,$1D,$C7,$BF,$7E,$03
	.dc.b	$0D,$AE,$3F,$01,$B3,$87,$C0,$58,$EB,$F0,$1C,$CE,$9F,$01,$63,$C3
	.dc.b	$F0,$1D,$FE,$03,$C0,$D8,$F2,$FB,$FE,$33,$C9,$F0,$16,$3D,$3E,$FF
	.dc.b	$9F,$E0,$3D,$0D,$8F,$6F,$BF,$EB,$F8,$0F,$63,$63,$E0,$FB,$FE,$F3
	.dc.b	$E0,$7C,$05,$8F,$8B,$F0,$1F,$0F,$E0,$3E,$23,$63,$E4,$FC,$07,$C7
	.dc.b	$3E,$47,$C0,$58,$FF,$7A,$FE,$8F,$FC,$3F,$D0,$FC,$F3,$FC,$4F,$CD
	.dc.b	$39,$9F,$CE,$C6,$3E,$27,$F3,$13,$FC,$54,$FE,$27,$E1,$FE,$85,$C7
	.dc.b	$C4,$6B,$E0,$F7,$F0,$77,$1F,$0B,$CB,$E0,$A4,$79,$3E,$0E,$E3,$FE
	.dc.b	$C3,$C5,$F0,$5E,$03,$C3,$F0,$76,$FD,$7F,$05,$E4,$F8,$3F,$2B,$63
	.dc.b	$DF,$F0,$5E,$8F,$83,$F4,$B6,$3E,$1F,$8F,$EC,$3D,$BF,$07,$63,$E3
	.dc.b	$FC,$1F,$C0,$3E,$0F,$C1,$D8,$F9,$7F,$07,$F1,$3E,$0F,$E2,$B6,$3E
	.dc.b	$7F,$C1,$7E,$FF,$F4,$8F,$F1,$2F,$0F,$F1,$39,$FE,$13,$D5,$FC,$3E
	.dc.b	$7F,$C5,$27,$F0,$F7,$1F,$07,$C7,$E1,$B9,$3C,$FE,$12,$5F,$23,$EB
	.dc.b	$DB,$E1,$33,$CF,$FE,$ED,$87,$C1,$F8,$FE,$0F,$C8,$DB,$E5,$5F,$C6
	.dc.b	$E9,$F0,$7E,$9F,$83,$F5,$36,$B7,$E1,$29,$F0,$7E,$FF,$83,$F8,$0D
	.dc.b	$AD,$CB,$E0,$FE,$21,$F1,$7E,$0E,$CE,$BE,$0F,$E4,$1F,$27,$E0,$EC
	.dc.b	$72,$F8,$2F,$99,$F0,$7F,$35,$B1,$DB,$E0,$FE,$87,$C1,$FD,$16,$C7
	.dc.b	$8B,$E0,$BE,$A7,$C1,$FD,$56,$C7,$B3,$F4,$9F,$F7,$A9,$3E,$13,$F9
	.dc.b	$C3,$8F,$83,$A6,$3E,$12,$D7,$1F,$84,$D9,$C3,$E1,$2C,$75,$F8,$4E
	.dc.b	$67,$4F,$84,$B1,$E1,$F8,$3E,$FF,$09,$E0,$6C,$79,$7E,$13,$C6,$79
	.dc.b	$3E,$12,$C7,$A7,$E0,$FC,$FF,$09,$E8,$6C,$7B,$7E,$0F,$D6,$7B,$3E
	.dc.b	$12,$C7,$C1,$F8,$4F,$7F,$C2,$7C,$06,$C7,$C5,$F8,$3F,$87,$F0,$9F
	.dc.b	$11,$B1,$F2,$7E,$13,$E3,$9F,$23,$E1,$2C,$7E,$23,$1F,$E1,$DE,$4F
	.dc.b	$F6,$96,$F8,$E7,$F8,$A1,$83,$FC,$48,$4F,$83,$FE,$D4,$E3,$E0,$EF
	.dc.b	$C1,$E1,$B8,$77,$F8,$3D,$FC,$1F,$06,$C7,$8F,$E0,$F9,$9D,$3E,$0E
	.dc.b	$C7,$9F,$E0,$BB,$9E,$0F,$83,$B1,$EB,$F8,$2F,$1F,$C1,$F9,$1B,$1E
	.dc.b	$FF,$83,$F3,$FC,$1F,$A1,$B1,$F0,$FE,$0F,$D7,$F0,$7E,$C6,$C7,$C7
	.dc.b	$F8,$2F,$79,$F0,$3E,$0E,$C7,$CB,$F8,$3F,$F0,$67,$E9,$5F,$EB,$2E
	.dc.b	$5F,$E1,$DD,$0F,$83,$FF,$0E,$4F,$FA,$D8,$FE,$1D,$7F,$CF,$9B,$3F
	.dc.b	$E1,$B7,$FD,$69,$19,$F8,$3E,$BF,$05,$DB,$E0,$AC,$FF,$02,$AF,$E2
	.dc.b	$0D,$F8,$1F,$19,$E4,$F8,$2B,$1B,$F8,$1F,$3F,$C1,$7A,$1B,$1C,$FE
	.dc.b	$07,$D7,$F0,$5E,$C6,$C7,$7F,$81,$F7,$9F,$03,$E0,$6C,$78,$FE,$07
	.dc.b	$E1,$9F,$13,$E0,$6C,$79,$FE,$03,$E3,$9F,$23,$E0,$6C,$7A,$FE,$03
	.dc.b	$E5,$FC,$0F,$CC,$6C,$7B,$FE,$03,$E7,$FC,$0F,$D0,$6C,$7F,$2A,$F9
	.dc.b	$87,$C7,$35,$FF,$22,$46,$35,$B3,$9A,$E0,$A3,$3A,$EB,$5B,$5A,$78
	.dc.b	$EB,$23,$8E,$97,$A7,$F0,$11,$FD,$EB,$F4,$BF,$F2,$BB,$FC,$E1,$B1
	.dc.b	$A3,$FC,$B4,$BE,$13,$FC,$AC,$EB,$B3,$FC,$C8,$BE,$0F,$26,$BE,$0E
	.dc.b	$E1,$D7,$E0,$F8,$FC,$1F,$26,$C7,$87,$E0,$FA,$9D,$BE,$0E,$C7,$97
	.dc.b	$E0,$BC,$27,$8B,$E0,$EC,$7A,$7E,$0F,$CA,$79,$BE,$0E,$C7,$B7,$E0
	.dc.b	$BD,$3F,$07,$EA,$6C,$7C,$1F,$83,$F6,$F9,$5E,$E6,$C7,$C5,$F8,$2F
	.dc.b	$83,$F0,$7F,$09,$B1,$F2,$7E,$0B,$E2,$9F,$1B,$E0,$EC,$7C,$DF,$83
	.dc.b	$F9,$3F,$07,$F2,$9B,$1F,$DA,$3F,$4C,$FE,$3E,$4F,$89,$FF,$42,$38
	.dc.b	$F9,$0B,$F0,$B8,$6D,$71,$F8,$4D,$9C,$3E,$16,$C7,$5F,$84,$E7,$F0
	.dc.b	$BD,$1B,$1E,$1F,$85,$EF,$F0,$BE,$06,$C7,$97,$E1,$7C,$67,$93,$E1
	.dc.b	$6C,$7A,$7E,$13,$CE,$7A,$3E,$16,$C7,$B7,$E1,$3D,$67,$B3,$E1,$6C
	.dc.b	$7C,$1F,$84,$F7,$FC,$2F,$C0,$6C,$7C,$5F,$85,$F8,$7F,$0B,$F1,$1B
	.dc.b	$1F,$27,$E1,$7F,$D5,$5E,$93,$E3,$BE,$4C,$61,$CF,$C3,$6B,$E2,$AD
	.dc.b	$FF,$86,$2F,$F8,$EF,$A7,$FE,$B9,$7F,$D6,$CF,$E8,$FE,$21,$18,$F4
	.dc.b	$FF,$D8,$8F,$F1,$E9,$D3,$9A,$FF,$33,$BF,$E2,$E7,$B7,$C2,$79,$BE
	.dc.b	$17,$CE,$FA,$3E,$12,$4F,$8F,$E1,$6E,$1D,$3E,$13,$DB,$F0,$BE,$E6
	.dc.b	$D7,$A7,$E1,$7E,$09,$F0,$BE,$0A,$C7,$B7,$E0,$3E,$2F,$C0,$7C,$66
	.dc.b	$C7,$C1,$DB,$E4,$FC,$07,$CA,$6C,$7C,$5F,$80,$F9,$A7,$CE,$F8,$5B
	.dc.b	$1F,$27,$E1,$3E,$89,$F4,$BE,$16,$C7,$F9,$13,$F4,$DF,$E1,$0F,$FA
	.dc.b	$29,$FE,$18,$E7,$83,$FD,$C5,$18,$7E,$83,$FC,$36,$39,$7C,$0E,$BE
	.dc.b	$0B,$6F,$0F,$81,$E3,$F0,$56,$BC,$0F,$4F,$82,$91,$E3,$F8,$2B,$97
	.dc.b	$6F,$82,$F0,$FC,$17,$89,$B5,$E9,$F8,$2F,$29,$E6,$F8,$2B,$1E,$DF
	.dc.b	$81,$F4,$9E,$AF,$82,$B1,$F0,$7E,$07,$DB,$F0,$5E,$E6,$C7,$C5,$F8
	.dc.b	$1F,$83,$F0,$5F,$09,$B1,$F2,$7E,$0B,$E2,$9F,$1B,$E0,$AC,$7C,$DF
	.dc.b	$81,$F9,$27,$CA,$F8,$2B,$1F,$E7,$4F,$D3,$BF,$C5,$67,$FA,$F4,$BF
	.dc.b	$06,$8D,$7C,$0D,$E4,$FD,$05,$23,$1E,$66,$7C,$D4,$FF,$B0,$1B,$3A
	.dc.b	$BF,$F6,$3C,$BE,$07,$18,$2F,$8F,$C0,$E1,$E7,$36,$93,$7E,$02,$61
	.dc.b	$EB,$F8,$1B,$8F,$9B,$E0,$38,$1C,$7E,$06,$D7,$C1,$F8,$1E,$9F,$03
	.dc.b	$D5,$B1,$F1,$7E,$07,$C1,$F0,$3E,$16,$C7,$C9,$F8,$1F,$21,$E5,$F8
	.dc.b	$1B,$1F,$37,$E0,$3D,$07,$A7,$E0,$6C,$7D,$1F,$80,$F6,$1E,$DF,$81
	.dc.b	$B1,$F5,$7E,$07,$E0,$1F,$0F,$F4,$FF,$E3,$51,$93,$FD,$5A,$FF,$8D
	.dc.b	$39,$CD,$3F,$8C,$B8,$F8,$1E,$5F,$17,$CF,$E2,$ED,$F2,$7B,$7C,$14
	.dc.b	$8E,$BF,$05,$79,$FF,$90,$65,$FE,$C0,$9F,$0F,$C0,$F9,$7E,$0B,$CC
	.dc.b	$F9,$FE,$06,$4F,$9B,$E0,$AE,$1C,$BE,$0B,$D8,$7B,$7E,$0A,$D7,$B3
	.dc.b	$E0,$7E,$07,$C1,$7C,$16,$C7,$C0,$F8,$1F,$89,$F0,$5F,$15,$B1,$F1
	.dc.b	$3E,$0B,$E4,$1F,$27,$E0,$AC,$7C,$8F,$82,$F9,$87,$CD,$F8,$2B,$1F
	.dc.b	$ED,$8E,$EF,$E2,$BB,$7E,$55,$FE,$25,$E6,$7E,$7D,$CF,$51,$CD,$2B
	.dc.b	$87,$C0,$FE,$6E,$5F,$03,$63,$0E,$7E,$07,$5F,$03,$70,$F1,$3C,$7E
	.dc.b	$07,$97,$C0,$D9,$F3,$3D,$7E,$02,$47,$A7,$E0,$6E,$9F,$D5,$F4,$7C
	.dc.b	$0F,$8C,$F2,$7C,$0D,$BF,$81,$F0,$3E,$7F,$81,$F4,$36,$3E,$27,$C0
	.dc.b	$7A,$CF,$67,$C0,$D8,$F9,$1F,$01,$EF,$F8,$1F,$80,$D8,$F9,$9F,$03
	.dc.b	$F0,$FE,$07,$E2,$36,$3E,$87,$C0,$FC,$73,$E4,$7C,$0D,$8F,$99,$D5
	.dc.b	$FC,$3F,$F5,$0F,$F6,$19,$3E,$0B,$F8,$43,$8F,$82,$BF,$0D,$86,$DF
	.dc.b	$23,$EB,$77,$2B,$F8,$A4,$67,$9B,$F8,$A4,$61,$FF,$00,$3F,$C4,$A3
	.dc.b	$53,$C7,$E0,$FC,$3F,$03,$E2,$F8,$1B,$5A,$B4,$79,$BE,$07,$CF,$F0
	.dc.b	$36,$B4,$FA,$BE,$02,$47,$1F,$81,$BA,$78,$BE,$03,$DF,$F0,$3F,$01
	.dc.b	$B5,$DF,$E0,$7E,$1F,$81,$F1,$1B,$1E,$3F,$81,$F8,$E7,$C8,$F8,$1B
	.dc.b	$1E,$7F,$81,$F9,$7F,$03,$F3,$1B,$1E,$BF,$81,$FE,$6D,$FA,$8F,$F1
	.dc.b	$4D,$7E,$2C,$63,$E1,$F3,$F0,$72,$FE,$3D,$FF,$14,$3F,$8C,$1F,$E2
	.dc.b	$5B,$FE,$84,$FF,$8A,$7B,$3F,$A6,$39,$E9,$39,$AF,$B2,$FF,$8A,$EB
	.dc.b	$FC,$68,$BE,$03,$CB,$F0,$3E,$67,$CF,$F0,$1E,$8F,$80,$B8,$79,$1F
	.dc.b	$5F,$C1,$48,$F3,$FC,$1D,$CB,$AE,$3C,$E7,$D6,$D8,$F0,$E1,$D6,$7D
	.dc.b	$ED,$8F,$2F,$C1,$7C,$8F,$82,$F9,$2D,$AF,$8B,$F0,$5F,$33,$E0,$BE
	.dc.b	$6B,$63,$E4,$FC,$0F,$D0,$F8,$2F,$A2,$D8,$F9,$BF,$03,$F5,$3E,$0B
	.dc.b	$EA,$B6,$3E,$8F,$C1,$7D,$8F,$82,$FB,$2D,$8F,$AB,$F0,$5F,$70,$FB
	.dc.b	$BF,$05,$63,$F3,$FF,$A9,$7F,$20,$3F,$8F,$F0,$7F,$79,$71,$F1,$7F
	.dc.b	$DE,$5F,$F1,$B4,$7F,$99,$DF,$E4,$33,$CB,$E1,$35,$F0,$9B,$F8,$49
	.dc.b	$3D,$BE,$12,$E1,$C7,$1E,$9F,$09,$D5,$B5,$CC,$F2,$E1,$E1,$C7,$AC
	.dc.b	$7A,$31,$E7,$DF,$FE,$17,$CD,$9E,$2B,$F8,$1C,$3E,$13,$D4,$7A,$FE
	.dc.b	$12,$CF,$C2,$F8,$4F,$77,$C2,$7B,$DB,$1F,$1B,$E1,$3E,$11,$F0,$FE
	.dc.b	$0E,$C7,$CA,$C3,$E3,$7C,$27,$4E,$2F,$8C,$F3,$9F,$9F,$8D,$9F,$94
	.dc.b	$F2,$7E,$66,$1D,$67,$C0,$7D,$6A,$FA,$5F,$09,$DA,$3E,$CB,$CF,$6F
	.dc.b	$85,$85,$9F,$82,$F5,$8F,$8D,$7D,$70,$EF,$5F,$82,$F8,$4F,$BD,$F0
	.dc.b	$7F,$7D,$B1,$F8,$6F,$84,$FC,$17,$C1,$FE,$0D,$B1,$F8,$AF,$84,$FC
	.dc.b	$37,$C2,$7E,$1D,$B1,$F8,$EC,$39,$E3,$F8,$9C,$6E,$7F,$67,$1E,$73
	.dc.b	$F8,$D6,$C7,$DD,$C3,$9C,$FE,$45,$B1,$F8,$1F,$84,$FF,$27,$61,$FE
	.dc.b	$5F,$C7,$F3,$B1,$87,$3F,$09,$1F,$E7,$F3,$FE,$61,$6C,$72,$F8,$4E
	.dc.b	$7F,$07,$67,$6F,$37,$8F,$C3,$5C,$FF,$D9,$58,$7F,$4B,$6E,$1D,$DC
	.dc.b	$FC,$2D,$CF,$FC,$DD,$5F,$E6,$D3,$FA,$A3,$8C,$BC,$38,$F9,$BE,$17
	.dc.b	$DB,$F0,$9E,$E6,$C7,$AB,$E1,$3E,$0F,$C2,$7C,$26,$C7,$BB,$E1,$3E
	.dc.b	$29,$F1,$BE,$12,$C7,$C2,$C3,$E4,$FC,$1F,$4D,$BD,$6F,$39,$F8,$F8
	.dc.b	$D9,$F7,$BC,$9F,$83,$87,$59,$F0,$1F,$3B,$E1,$3E,$AF,$C1,$FD,$66
	.dc.b	$C7,$A4,$EA,$5A,$EA,$F2,$7A,$BF,$63,$2F,$06,$DF,$2B,$E1,$3F,$03
	.dc.b	$F0,$9F,$82,$6D,$7D,$EF,$84,$FC,$2F,$C1,$FE,$19,$B1,$F8,$2F,$84
	.dc.b	$FC,$4F,$C2,$7E,$29,$B1,$FE,$04,$FD,$4F,$FB,$26,$5F,$EA,$56,$C7
	.dc.b	$F8,$43,$99,$9F,$FA,$7E,$73,$AF,$FA,$8E,$59,$78,$23,$F0,$ED,$C3
	.dc.b	$F0,$78,$F8,$27,$F1,$6D,$8F,$07,$C2,$75,$F8,$AE,$CD,$CF,$CD,$F0
	.dc.b	$1E,$13,$C5,$F0,$36,$3D,$5F,$03,$E5,$F8,$3F,$33,$63,$DD,$F0,$5E
	.dc.b	$9F,$83,$F5,$36,$3A,$93,$56,$9F,$0F,$1E,$D1,$89,$FF,$19,$F0,$7F
	.dc.b	$7B,$AF,$91,$97,$82,$7E,$63,$F1,$FE,$0B,$E4,$7C,$1D,$9F,$A0,$FC
	.dc.b	$BF,$82,$F9,$9F,$07,$67,$EA,$3F,$3F,$E0,$A4,$7D,$7F,$83,$BA,$78
	.dc.b	$CF,$F6,$A6,$8F,$82,$FA,$DF,$07,$F5,$DB,$7F,$7F,$2E,$ED,$3B,$9F
	.dc.b	$83,$C7,$AE,$DF,$83,$CA,$CF,$C0,$F8,$2F,$C1,$1F,$83,$F8,$3B,$1C
	.dc.b	$F4,$F2,$7C,$2B,$56,$DE,$D9,$CC,$62,$BC,$04,$8C,$E3,$E0,$7D,$BF
	.dc.b	$09,$F9,$1F,$84,$FC,$93,$6B,$FC,$BB,$FA,$A7,$FA,$78,$FE,$FF,$91
	.dc.b	$C7,$C2,$7F,$BD,$DF,$EF,$FB,$2B,$FC,$DB,$18,$6F,$C2,$6F,$E1,$38
	.dc.b	$7C,$24,$9E,$FF,$09,$75,$F7,$E1,$DA,$7C,$71,$DF,$2E,$51,$E6,$C3
	.dc.b	$C5,$8E,$2B,$8F,$C2,$F9,$7E,$07,$CC,$DB,$F6,$FC,$27,$A7,$E1,$3D
	.dc.b	$4D,$8F,$83,$F0,$3E,$DF,$81,$F7,$36,$3E,$2F,$C2,$7C,$13,$E1,$7C
	.dc.b	$25,$8F,$93,$F0,$9F,$14,$F8,$DF,$05,$63,$E6,$F9,$BF,$27,$E0,$BE
	.dc.b	$53,$63,$E8,$FC,$27,$CD,$3E,$77,$C1,$D8,$EA,$4E,$1F,$8F,$87,$57
	.dc.b	$E4,$DF,$36,$CF,$DC,$C3,$EC,$7C,$24,$8F,$BF,$F0,$96,$BF,$1F,$FA
	.dc.b	$AF,$F6,$56,$FC,$1F,$DD,$73,$57,$B3,$1D,$AB,$5E,$6B,$FD,$5F,$1F
	.dc.b	$DA,$6F,$AF,$C1,$65,$5F,$1F,$83,$C3,$CE,$6D,$7F,$60,$BE,$7F,$0B
	.dc.b	$8F,$85,$CB,$AF,$85,$93,$FE,$B6,$AF,$F5,$8B,$63,$9F,$C2,$C8,$D7
	.dc.b	$C1,$DD,$3E,$0F,$C1,$F8,$0F,$0F,$C0,$DA,$E9,$F0,$9E,$43,$CB,$F0
	.dc.b	$76,$3C,$1F,$09,$E8,$3D,$3F,$09,$63,$C9,$F0,$5E,$CF,$82,$F6,$B6
	.dc.b	$3D,$1F,$09,$F0,$0F,$83,$F0,$76,$3D,$9F,$09,$F1,$BD,$AF,$E9,$3C
	.dc.b	$FF,$46,$31,$F1,$19,$F8,$4D,$7C,$2D,$C3,$FD,$B3,$82,$39,$7C,$2C
	.dc.b	$9C,$FC,$15,$E5,$FE,$23,$7F,$CD,$67,$FD,$D2,$D8,$FB,$2F,$FA,$C6
	.dc.b	$3A,$FC,$2F,$97,$E1,$3C,$DF,$0B,$67,$C2,$FE,$07,$C2,$7E,$54,$F1
	.dc.b	$B0,$DA,$F3,$FC,$2E,$27,$E0,$7C,$2C,$9F,$5F,$C2,$5D,$7C,$FF,$03
	.dc.b	$F1,$4F,$8D,$F0,$96,$BF,$DF,$D9,$F8,$3D,$3E,$E7,$83,$D4,$B3,$F1
	.dc.b	$BE,$07,$E7,$1F,$3F,$E1,$6C,$7C,$AE,$3F,$A4,$7D,$3F,$85,$B1,$DB
	.dc.b	$85,$F0,$3E,$CF,$81,$F6,$B6,$7E,$9F,$C2,$FD,$B3,$EE,$7C,$1D,$8F
	.dc.b	$AF,$F0,$9F,$7C,$FD,$2F,$EA,$CE,$3C,$4F,$B3,$F2,$B5,$FC,$D5,$76
	.dc.b	$30,$F2,$F8,$5E,$7F,$0B,$6F,$A7,$27,$EF,$32,$FE,$75,$97,$ED,$9F
	.dc.b	$F5,$DB,$FD,$51,$B1,$BD,$3C,$BF,$0B,$E1,$E3,$F1,$73,$5C,$7C,$CF
	.dc.b	$AB,$82,$47,$80,$F6,$78,$AC,$9F,$09,$AD,$3C,$1F,$0B,$F0,$BE,$13
	.dc.b	$E1,$B7,$0F,$BB,$DF,$FE,$D6,$DC,$D3,$D6,$E3,$87,$E5,$FC,$2C,$AF
	.dc.b	$16,$3E,$FE,$CC,$3E,$FE,$CF,$A3,$F0,$BF,$49,$B5,$C3,$7F,$47,$C2
	.dc.b	$FD,$63,$EB,$FC,$2D,$9D,$70,$BE,$0F,$ED,$FC,$1F,$DC,$6C,$EB,$C5
	.dc.b	$FB,$FF,$0B,$F8,$06,$C7,$2F,$85,$FC,$19,$F8,$4F,$85,$B1,$DB,$E1
	.dc.b	$7F,$0E,$7F,$2C,$FD,$5F,$F9,$E7,$2F,$DD,$8C,$7C,$36,$7E,$1A,$DF
	.dc.b	$AE,$FF,$9C,$5F,$F3,$E5,$F0,$DC,$CE,$9F,$0B,$6B,$87,$C3,$77,$F8
	.dc.b	$5F,$03,$63,$A7,$C5,$78,$CF,$27,$C3,$58,$F0,$7C,$37,$9C,$F4,$7C
	.dc.b	$35,$8C,$93,$B3,$FC,$4B,$A3,$E1,$B5,$18,$F8,$6F,$81,$F0,$DF,$05
	.dc.b	$B7,$ED,$F8,$6F,$88,$7C,$5F,$86,$B1,$92,$5F,$6F,$85,$CB,$E0,$F8
	.dc.b	$6F,$98,$7C,$DF,$86,$B5,$F2,$7E,$1B,$E8,$1F,$47,$E1,$AC,$64,$97
	.dc.b	$FE,$3E,$FD,$63,$F9,$DE,$EB,$FE,$CD,$F2,$BF,$E6,$AF,$13,$FB,$45
	.dc.b	$BF,$0C,$65,$AE,$B5,$6A,$73,$9F,$E7,$2F,$85,$E4,$E1,$E1,$D9,$D5
	.dc.b	$70,$AF,$F3,$21,$C2,$3F,$4B,$8F,$F9,$F3,$C4,$E1,$3A,$FC,$2F,$FD
	.dc.b	$36,$9F,$F4,$EC,$FF,$A0,$BC,$4F,$4C,$68,$9F,$15,$E9,$F1,$B2,$CC
	.dc.b	$7D,$13,$CB,$67,$8D,$9C,$BC,$BF,$85,$3C,$26,$3F,$C9,$66,$09,$D4
	.dc.b	$62,$3F,$95,$47,$E7,$3C,$FC,$46,$EB,$9B,$B9,$E2,$D8,$CB,$DA,$39
	.dc.b	$39,$AE,$71,$C2,$35,$3C,$1F,$34,$70,$73,$3E,$5D,$5F,$9C,$BD,$5F
	.dc.b	$66,$7F,$E3,$39,$CC,$63,$5F,$E8,$91,$8C,$3F,$01,$F8,$53,$D2,$D4
	.dc.b	$7F,$49,$8C,$3F,$73,$1F,$E5,$ED,$F5,$B6,$A7,$6F,$E1,$1B,$97,$8E
	.dc.b	$9E,$AF,$6C,$7D,$2F,$2A,$DE,$5F,$CF,$BA,$DF,$1B,$E4,$F4,$91,$8C
	.dc.b	$FF,$96,$35,$FD,$37,$C6,$EE,$1F,$1B,$7E,$36,$57,$08,$D1,$BF,$8D
	.dc.b	$91,$93,$AF,$C8,$F3,$23,$E0,$F8,$EE,$DF,$1D,$DF,$DD,$93,$E4,$8C
	.dc.b	$BE,$6F,$8E,$F2,$18,$7F,$79,$F1,$DE,$7F,$8E,$91,$EC,$8F,$57,$C1
	.dc.b	$7A,$FE,$0E,$4F,$C0,$8C,$BF,$0B,$E3,$BE,$01,$87,$F8,$47,$C7,$7C
	.dc.b	$32,$3F,$23,$E3,$BE,$37,$C6,$FC,$7F,$8C,$93,$F3,$23,$E5,$7C,$64
	.dc.b	$8F,$9E,$FD,$3F,$4F,$F5,$3D,$FF,$E7,$E5,$E3,$64,$95,$FD,$86,$1F
	.dc.b	$19,$C7,$E2,$F9,$7C,$64,$AD,$47,$5F,$1B,$B7,$8B,$27,$94,$78,$7C
	.dc.b	$3F,$17,$C6,$49,$ED,$1E,$5F,$1F,$CD,$F1,$92,$7C,$51,$E9,$F8,$B9
	.dc.b	$1E,$57,$D9,$F1,$9E,$DF,$8C,$93,$E9,$7E,$07,$AF,$FA,$6E,$27,$C1
	.dc.b	$3E,$17,$0B,$E1,$A9,$88,$F9,$1E,$CF,$C9,$F6,$26,$5F,$16,$3E,$67
	.dc.b	$B1,$F3,$7D,$89,$3F,$26,$3C,$1F,$95,$FF,$1E,$8F,$F2,$BB,$56,$7F
	.dc.b	$EB,$18,$C3,$6B,$EC,$7B,$1F,$67,$E3,$25,$FD,$58,$FB,$9F,$19,$FA
	.dc.b	$FF,$D6,$7F,$D7,$93,$D3,$D5,$FF,$18,$C6,$31,$FC,$05,$79,$38,$FC
	.dc.b	$BC,$73,$0F,$2C,$73,$F8,$D9,$1F,$E2,$3C,$3F,$D8,$6D,$F8,$DF,$07
	.dc.b	$C6,$F8,$7D,$E9,$3B,$8F,$83,$DD,$F0,$BB,$24,$FC,$28,$F4,$73,$7A
	.dc.b	$79,$E4,$F7,$8F,$67,$C6,$FB,$7E,$36,$4F,$8E,$3E,$6B,$0F,$9A,$7D
	.dc.b	$0F,$17,$E2,$75,$C8,$F5,$3F,$1F,$E3,$7E,$47,$C6,$FC,$9F,$8C,$95
	.dc.b	$EF,$8F,$C4,$6A,$F3,$BE,$77,$C6,$CA,$FF,$48,$CE,$5F,$8D,$F1,$BF
	.dc.b	$4C,$8F,$C9,$F8,$DF,$AD,$E6,$7D,$7F,$8D,$93,$F3,$63,$ED,$7C,$6F
	.dc.b	$DB,$F8,$C9,$3F,$46,$3E,$F7,$C6,$48,$FA,$8F,$F2,$FF,$D6,$BF,$AC
	.dc.b	$F8,$1F,$EF,$A7,$F9,$73,$7E,$13,$46,$FE,$1B,$87,$C1,$CC,$3F,$CC
	.dc.b	$17,$FC,$C9,$BF,$09,$D7,$E1,$24,$69,$F0,$7C,$27,$87,$E1,$24,$F2
	.dc.b	$7C,$9F,$07,$E5,$F8,$49,$3D,$9F,$47,$C2,$7A,$7E,$0F,$D5,$F0,$92
	.dc.b	$BC,$71,$ED,$F8,$39,$1E,$67,$E0,$7C,$27,$C1,$EC,$F8,$5F,$09,$2B
	.dc.b	$D7,$1F,$17,$E1,$24,$7B,$9F,$91,$F0,$7F,$27,$E1,$3E,$57,$C2,$4A
	.dc.b	$F8,$71,$FD,$6F,$F5,$BF,$E5,$3E,$AF,$FA,$71,$FE,$53,$3F,$E8,$76
	.dc.b	$BF,$B1,$1F,$0E,$8E,$31,$87,$F9,$6B,$C7,$E1,$BB,$1D,$FE,$17,$C1
	.dc.b	$F0,$DE,$1F,$86,$95,$DE,$3C,$9E,$94,$8F,$13,$E7,$F2,$3D,$1F,$0D
	.dc.b	$27,$CC,$FA,$FE,$17,$D9,$F0,$DE,$DF,$86,$95,$EB,$8E,$8F,$AF,$6F
	.dc.b	$56,$38,$F1,$5C,$1F,$0F,$3E,$D7,$BE,$06,$98,$BD,$CF,$C1,$D3,$E6
	.dc.b	$7B,$9F,$37,$DD,$93,$F2,$E3,$E8,$7C,$34,$8F,$9C,$FD,$3F,$85,$FA
	.dc.b	$9F,$0D,$27,$E9,$3F,$5F,$E1,$BE,$C7,$C2,$FD,$9F,$84,$95,$F5,$E3
	.dc.b	$EE,$7C,$2C,$8F,$B4,$FD,$FF,$85,$FF,$0D,$5F,$F2,$9E,$2F,$F6,$DB
	.dc.b	$6B,$EE,$B7,$90,$FF,$26,$8F,$90,$BE,$0F,$9F,$C1,$49,$FF,$67,$4E
	.dc.b	$5F,$09,$0F,$95,$8F,$8D,$DE,$1E,$42,$3C,$BE,$17,$CF,$F0,$9E,$8F
	.dc.b	$83,$93,$DA,$3D,$7F,$09,$23,$C2,$FB,$BE,$0F,$DF,$F0,$9F,$03,$E0
	.dc.b	$E5,$79,$A3,$E1,$FC,$1C,$8F,$4B,$F1,$BE,$0B,$A4,$7C,$0C,$2F,$5F
	.dc.b	$E0,$38,$BE,$2E,$5D,$AB,$E0,$E5,$D9,$F8,$F9,$7D,$0F,$83,$FA,$3F
	.dc.b	$07,$27,$E3,$C7,$4E,$0F,$A7,$9D,$9F,$A7,$8F,$57,$E8,$E3,$D7,$3F
	.dc.b	$17,$13,$3F,$BB,$F0,$7F,$7B,$E0,$E6,$1F,$52,$3F,$03,$F0,$5F,$82
	.dc.b	$F8,$39,$3F,$62,$3F,$0B,$F0,$52,$3E,$DB,$F8,$8F,$83,$FC,$4F,$C1
	.dc.b	$FF,$98,$FF,$5C,$FE,$6B,$C2,$BF,$E0,$B1,$87,$F1,$73,$E0,$7F,$0B
	.dc.b	$8F,$81,$FC,$5E,$3D,$DF,$C6,$E3,$76,$F6,$E5,$DF,$E0,$E5,$77,$9F
	.dc.b	$A5,$97,$F4,$4C,$FC,$5C,$A8,$7C,$27,$9F,$E1,$26,$1C,$6F,$75,$FD
	.dc.b	$16,$3F,$82,$2F,$83,$F7,$7C,$1C,$9E,$07,$B9,$F8,$5F,$07,$F0,$FE
	.dc.b	$0E,$4E,$B3,$F1,$47,$C7,$F8,$3F,$0C,$F6,$CE,$EB,$EA,$CB,$B4,$F8
	.dc.b	$B2,$B3,$EE,$CB,$E8,$7C,$1C,$AE,$DA,$23,$EA,$7C,$1F,$D5,$F8,$29
	.dc.b	$5A,$7E,$C7,$C1,$FD,$9F,$83,$FB,$5F,$07,$2B,$9C,$7D,$DF,$82,$FB
	.dc.b	$DF,$07,$27,$BC,$74,$7E,$3E,$17,$5F,$A7,$8F,$59,$F2,$E3,$D5,$F9
	.dc.b	$B8,$FE,$27,$E0,$E4,$7A,$9F,$C6,$7C,$1F,$3E,$6F,$AB,$95,$AF,$C6
	.dc.b	$E3,$D5,$FA,$F8,$F5,$CB,$F0,$79,$7E,$57,$E0,$BF,$2D,$F0,$72,$FE
	.dc.b	$3C,$7F,$AE,$FF,$5D,$FE,$B3,$D2,$69,$3E,$1B,$5F,$0B,$2F,$EE,$4F
	.dc.b	$F5,$E7,$FA,$52,$F8,$5E,$9F,$09,$D7,$E1,$65,$70,$8F,$03,$0E,$67
	.dc.b	$8F,$0E,$CF,$4C,$2E,$BF,$C3,$F1,$ED,$3E,$2F,$83,$F5,$7C,$1C,$8C
	.dc.b	$1E,$6F,$82,$F7,$7C,$14,$8C,$1E,$AF,$84,$F8,$5F,$0B,$23,$07,$BB
	.dc.b	$E0,$FE,$31,$1F,$83,$F0,$9F,$27,$0E,$91,$F0,$B0,$C3,$C8,$F8,$B7
	.dc.b	$D5,$F8,$F8,$7D,$1F,$84,$91,$F3,$1F,$A9,$F0,$9F,$57,$E1,$3E,$B7
	.dc.b	$A9,$2B,$E8,$C7,$D9,$F8,$39,$1F,$51,$FB,$9F,$09,$CA,$3E,$BD,$D7
	.dc.b	$EB,$FC,$0F,$57,$ED,$E1,$D5,$FB,$78,$74,$7E,$EE,$17,$8F,$E9,$63
	.dc.b	$F8,$9F,$84,$FC,$56,$32,$FF,$0B,$1D,$5F,$C2,$E3,$D6,$FE,$D6,$3C
	.dc.b	$FC,$1F,$E2,$FC,$AC,$F1,$87,$F8,$BA,$3F,$D5,$33,$FC,$73,$F5,$EF
	.dc.b	$EE,$84,$7F,$98,$4F,$F8,$6E,$3F,$BB,$61,$FD,$55,$FE,$E5,$19,$F8
	.dc.b	$5D,$CE,$5B,$5D,$7E,$16,$D7,$F8,$FA,$DF,$07,$8E,$F4,$62,$7F,$8F
	.dc.b	$35,$F1,$7C,$26,$E3,$C7,$87,$A7,$E1,$70,$67,$E0,$75,$9C,$C3,$9B
	.dc.b	$C7,$3E,$59,$C9,$EE,$F5,$CB,$B7,$C0,$C9,$F1,$BE,$1F,$80,$F1,$7C
	.dc.b	$0C,$9F,$3B,$E5,$F8,$5F,$37,$C2,$F9,$FE,$12,$57,$B2,$39,$3E,$EB
	.dc.b	$AF,$CF,$F8,$5E,$8F,$D1,$C2,$F1,$FD,$1C,$7A,$CF,$C6,$C7,$AB,$F5
	.dc.b	$F1,$F8,$DF,$09,$F1,$FE,$16,$4F,$CC,$8F,$95,$F0,$7F,$2F,$E1,$24
	.dc.b	$FD,$08,$F9,$DF,$09,$23,$E9,$BF,$47,$E1,$3E,$97,$C2,$F6,$9E,$87
	.dc.b	$D8,$7A,$BF,$6A,$FA,$BF,$8B,$BE,$8F,$E3,$6E,$EB,$F8,$DC,$3F,$09
	.dc.b	$EC,$7F,$7F,$EB,$3F,$DC,$EF,$FD,$0E,$D8,$CF,$C7,$78,$27,$F3,$19
	.dc.b	$78,$30,$FC,$C6,$5D,$BA,$FF,$41,$9D,$BF,$CA,$67,$E0,$F8,$5F,$0F
	.dc.b	$C1,$C9,$FE,$E1,$AF,$F8,$01,$BF,$07,$E6,$23,$AF,$83,$F4,$FC,$1C
	.dc.b	$8E,$2F,$B3,$E1,$7D,$BF,$0B,$EE,$F8,$49,$5D,$A3,$9B,$FE,$8B,$C7
	.dc.b	$FC,$10,$F8,$0E,$C7,$42,$1B,$BE,$CF,$A3,$E1,$3E,$59,$1F,$5F,$C0
	.dc.b	$FC,$EF,$81,$F9,$FF,$01,$27,$DF,$1F,$4B,$E0,$7E,$9F,$C1,$C9,$F8
	.dc.b	$71,$F5,$BE,$13,$EB,$E1,$27,$E3,$C7,$57,$E4,$E1,$D5,$F6,$61,$D5
	.dc.b	$F9,$B8,$75,$7E,$06,$1F,$82,$CF,$F0,$7F,$09,$27,$E9,$C7,$E1,$BE
	.dc.b	$0E,$47,$D6,$7F,$13,$F0,$9F,$C3,$7F,$5F,$FE,$E6,$7F,$B4,$3A,$FF
	.dc.b	$CD,$CD,$EA,$FF,$1E,$BA,$F8,$1E,$1F,$0B,$31,$FF,$18,$4F,$F7,$5B
	.dc.b	$FF,$3B,$64,$CB,$C3,$5F,$5D,$F0,$BC,$0F,$17,$C1,$C7,$07,$4F,$84
	.dc.b	$F3,$11,$EF,$F0,$BD,$30,$EF,$77,$4F,$F2,$9E,$2D,$7D,$D8,$75,$3F
	.dc.b	$DA,$59,$FB,$5F,$F7,$6B,$F0,$BE,$17,$E2,$7C,$2C,$9F,$6C,$7C,$7F
	.dc.b	$83,$91,$F0,$1F,$95,$F0,$7F,$2F,$E0,$A4,$FC,$47,$E7,$7C,$1F,$CF
	.dc.b	$F8,$59,$3F,$21,$FA,$5F,$0B,$D2,$3C,$B8,$5E,$0F,$A7,$8F,$49,$F4
	.dc.b	$E3,$67,$EB,$E3,$F6,$FE,$17,$EE,$7C,$2C,$BF,$AB,$1F,$7F,$E0,$7F
	.dc.b	$01,$F0,$52,$7E,$CC,$7E,$0F,$E0,$A4,$7D,$C7,$F0,$DF,$0B,$F8,$7F
	.dc.b	$85,$93,$F8,$07,$F1,$5F,$0B,$D6,$3F,$07,$87,$57,$E7,$E1,$D1,$FA
	.dc.b	$38,$5D,$3F,$1F,$8F,$F5,$FF,$D8,$3F,$DC,$44,$E8,$C7,$C2,$67,$E0
	.dc.b	$F5,$F0,$73,$2F,$F4,$05,$FF,$70,$C7,$FD,$8C,$FF,$65,$6F,$C1,$75
	.dc.b	$3B,$7C,$1F,$7F,$84,$97,$C9,$F1,$7C,$27,$78,$EB,$8F,$75,$87,$6C
	.dc.b	$B9,$C7,$8B,$1E,$CF,$93,$1F,$67,$C2,$74,$8F,$3E,$38,$7D,$37,$D7
	.dc.b	$0F,$4E,$3E,$19,$EF,$AF,$FA,$BB,$4F,$63,$73,$EF,$3A,$1E,$E2,$B7
	.dc.b	$18,$5F,$07,$F3,$C9,$1F,$2B,$E0,$FE,$97,$C1,$48,$F9,$AF,$D5,$F8
	.dc.b	$3F,$AD,$F0,$92,$7E,$8B,$F6,$7E,$0F,$ED,$7C,$27,$DB,$C2,$57,$D6
	.dc.b	$8E,$8F,$9F,$3F,$36,$38,$DF,$EC,$E5,$D6,$B5,$A3,$3C,$57,$DE,$CF
	.dc.b	$F1,$1F,$03,$F8,$9F,$82,$98,$7E,$12,$3F,$19,$F0,$52,$3F,$0E,$FE
	.dc.b	$3F,$E0,$BF,$21,$F0,$92,$7F,$16,$FF,$8C,$7F,$61,$FE,$DC,$7F,$6B
	.dc.b	$27,$C4,$E7,$E1,$7A,$57,$9B,$0B,$A7,$E4,$B1,$E9,$3E,$AC,$6C,$B9
	.dc.b	$BF,$DD,$39,$76,$F8,$5F,$01,$2F,$FD,$B9,$1F,$D7,$9F,$90,$BE,$13
	.dc.b	$CD,$F0,$32,$77,$1E,$9F,$84,$91,$C9,$F6,$7C,$1F,$B7,$E1,$7B,$4F
	.dc.b	$4C,$70,$F2,$3C,$38,$75,$7C,$38,$74,$7C,$78,$5C,$FD,$91,$F9,$2C
	.dc.b	$99,$7C,$BF,$85,$EB,$3E,$AC,$F0,$F7,$3D,$B9,$77,$C3,$9F,$4B,$1F
	.dc.b	$81,$5E,$B2,$1E,$23,$E3,$E3,$F6,$7E,$16,$47,$CA,$7E,$E7,$C2,$FD
	.dc.b	$DF,$83,$93,$F3,$9F,$C0,$7C,$27,$E0,$7E,$0E,$4F,$D2,$7F,$09,$F0
	.dc.b	$9C,$A3,$EA,$DD,$7E,$AF,$C2,$F4,$7E,$BE,$17,$5F,$5E,$8C,$BA,$D7
	.dc.b	$DC,$CA,$D6,$B2,$FC,$97,$BD,$F9,$3F,$85,$97,$F8,$48,$FC,$B7,$C0
	.dc.b	$48,$FC,$3B,$F9,$9F,$80,$FF,$45,$77,$FF,$BC,$0F,$EC,$1C,$46,$99
	.dc.b	$F8,$5E,$F7,$D4,$1F,$B7,$8F,$EF,$DD,$5F,$98,$C3,$C1,$3F,$94,$7C
	.dc.b	$0F,$0D,$3E,$CD,$78,$7E,$13,$C6,$79,$27,$CB,$F0,$BE,$6F,$85,$91
	.dc.b	$C7,$1F,$0E,$7E,$7A,$F5,$FC,$2F,$B3,$E0,$E5,$EC,$FE,$43,$92,$7E
	.dc.b	$0F,$C2,$7C,$2F,$82,$95,$BC,$7E,$14,$F8,$A3,$17,$FF,$00,$1F,$81
	.dc.b	$8F,$CA,$F8,$2F,$0C,$6B,$2F,$0B,$F1,$72,$EE,$F2,$CB,$BB,$F2,$71
	.dc.b	$E4,$F6,$C3,$C0,$FC,$DC,$3A,$BF,$3F,$0E,$BB,$7C,$AC,$7A,$CF,$D3
	.dc.b	$C7,$AC,$FC,$EC,$7E,$FF,$C1,$4A,$F5,$BF,$82,$F8,$2F,$C1,$FC,$24
	.dc.b	$9F,$7B,$F8,$6F,$84,$FC,$3F,$C1,$FE,$23,$E1,$25,$7C,$48,$E4,$F4
	.dc.b	$21,$F8,$5B,$E8,$FE,$1E,$EE,$5F,$82,$C3,$AC,$FC,$EC,$3A,$BF,$8D
	.dc.b	$C3,$F2,$FF,$03,$23,$E9,$BF,$9A,$F8,$2F,$CD,$FC,$17,$E7,$3E,$0A
	.dc.b	$57,$D8,$8F,$CF,$FC,$27,$E8,$3E,$12,$4F,$DC,$8E,$4F,$DE,$BA,$FF
	.dc.b	$53,$FD,$8B,$F8,$CB,$F9,$FB,$EA,$FE,$87,$0F,$0B,$CC,$FC,$3E,$1D
	.dc.b	$9F,$D3,$61,$FE,$F9,$BF,$F7,$9C,$7F,$74,$6F,$41,$A7,$F8,$67,$A7
	.dc.b	$F3,$EB,$DF,$E5,$F0,$72,$FF,$31,$39,$6F,$C0,$77,$F8,$49,$18,$37
	.dc.b	$F0,$5E,$3F,$85,$91,$83,$9E,$3D,$5F,$85,$8F,$A6,$3D,$58,$74,$9F
	.dc.b	$8D,$86,$1E,$47,$C9,$C3,$E0,$1F,$07,$E1,$39,$47,$A3,$0E,$DC,$DF
	.dc.b	$CF,$33,$F5,$4F,$D1,$C7,$B4,$E7,$1E,$D3,$F5,$7E,$17,$E6,$11,$F8
	.dc.b	$7F,$0B,$F3,$FE,$17,$E8,$7C,$24,$9F,$8F,$1F,$4F,$E1,$24,$7C,$A7
	.dc.b	$EB,$7C,$1F,$D7,$BE,$93,$DC,$EA,$72,$3E,$8D,$F5,$7F,$11,$7D,$1F
	.dc.b	$C5,$5D,$CB,$DD,$87,$E0,$FE,$0E,$57,$DA,$7F,$0D,$F0,$BF,$87,$F8
	.dc.b	$4F,$C4,$79,$32,$BE,$FC,$7E,$2F,$E0,$BF,$19,$F0,$52,$7F,$07,$1C
	.dc.b	$9E,$84,$3F,$39,$7D,$5F,$C4,$DF,$57,$F4,$17,$D5,$FC,$6D,$F5,$7F
	.dc.b	$49,$7F,$9B,$F8,$29,$1F,$92,$7F,$A2,$BF,$E3,$BF,$D8,$DF,$11,$8D
	.dc.b	$7F,$BC,$C3,$E1,$B1,$D9,$FC,$F2,$31,$7F,$CC,$DA,$F3,$F8,$6D,$BF
	.dc.b	$E5,$C8,$7C,$36,$3B,$D1,$FE,$25,$C7,$BF,$C1,$E6,$37,$7D,$73,$F4
	.dc.b	$BF,$BF,$D3,$51,$BD,$24,$F8,$DE,$59,$F3,$CF,$A6,$32,$BD,$11,$C9
	.dc.b	$EC,$43,$E2,$7C,$37,$47,$E3,$5D,$CF,$E2,$E1,$D6,$7E,$16,$1D,$5F
	.dc.b	$99,$87,$AF,$E1,$3D,$9F,$0D,$27,$E4,$47,$BF,$E1,$7E,$07,$C3,$49
	.dc.b	$F9,$91,$F0,$FE,$1A,$47,$CF,$7E,$37,$C3,$75,$8E,$64,$3E,$E7,$C2
	.dc.b	$F5,$7E,$B5,$F5,$7F,$01,$7D,$1F,$C1,$5D,$D7,$F0,$38,$7D,$3F,$84
	.dc.b	$95,$F7,$DF,$AD,$F0,$9F,$5F,$E1,$3E,$C7,$C3,$4A,$FC,$24,$7D,$BF
	.dc.b	$84,$FB,$9F,$0B,$27,$F1,$11,$C9,$E8,$43,$F2,$97,$D5,$FC,$75,$F5
	.dc.b	$7F,$31,$7D,$5F,$C9,$5F,$57,$F3,$97,$F8,$BF,$83,$FC,$67,$C2,$C9
	.dc.b	$FC,$C4,$7E,$4F,$F6,$4F,$D1,$C7,$FA,$75,$FF,$17,$91,$D7,$C3,$6F
	.dc.b	$E1,$24,$7F,$84,$E7,$FC,$5A,$FE,$7E,$BC,$0F,$63,$5F,$0B,$DD,$F0
	.dc.b	$7C,$34,$79,$1D,$8E,$97,$D9,$F2,$61,$E7,$A9,$1D,$0F,$17,$C2,$75
	.dc.b	$7C,$97,$D5,$F6,$1D,$4F,$75,$F5,$E2,$FF,$3B,$E5,$EA,$F0,$4B,$E1
	.dc.b	$7E,$37,$C2,$4C,$3E,$04,$7C,$9F,$85,$91,$F0,$DF,$99,$F0,$DF,$37
	.dc.b	$E1,$BE,$77,$C3,$4A,$F9,$11,$F4,$6E,$47,$43,$E6,$7C,$37,$47,$EA
	.dc.b	$61,$74,$F3,$E3,$D6,$7E,$96,$3D,$5F,$B5,$8F,$DD,$F8,$69,$1F,$5D
	.dc.b	$FC,$07,$C2,$FE,$07,$E1,$A4,$FD,$B7,$F0,$9F,$0D,$F8,$5F,$86,$FC
	.dc.b	$37,$C3,$4A,$FC,$04,$72,$7A,$10,$FC,$55,$F4,$7F,$19,$77,$4F,$9F
	.dc.b	$87,$59,$FC,$56,$1D,$5F,$C9,$61,$F9,$5F,$86,$91,$F8,$F7,$F3,$1F
	.dc.b	$0B,$F9,$9F,$86,$FE,$59,$7F,$E2,$2C,$7F,$D6,$2D,$F4,$BF,$CE,$7C
	.dc.b	$AA,$DF,$C3,$4C,$BF,$CD,$97,$FE,$1E,$3F,$C3,$8B,$9F,$F0,$79,$78
	.dc.b	$A3,$F4,$D9,$F8,$9F,$D1,$61,$E2,$CA,$46,$0E,$B8,$F7,$59,$F9,$A3
	.dc.b	$D3,$F0,$BE,$AF,$86,$E9,$5F,$B3,$CB,$0F,$33,$8E,$1F,$E8,$ED,$3D
	.dc.b	$D1,$E7,$6E,$9F,$9A,$CF,$C1,$9F,$7A,$F1,$9F,$03,$2F,$93,$F0,$B2
	.dc.b	$3E,$1B,$F3,$3E,$13,$E6,$FC,$24,$9F,$8E,$FD,$0F,$86,$FA,$3F,$0B
	.dc.b	$F4,$BE,$16,$57,$CC,$8E,$4F,$CE,$BA,$FA,$8F,$47,$C2,$75,$71,$AB
	.dc.b	$66,$59,$D7,$EE,$E7,$E0,$BF,$B3,$9F,$81,$E3,$6C,$FF,$0B,$F0,$B2
	.dc.b	$7E,$FB,$F8,$8F,$85,$FC,$4F,$C2,$FE,$2B,$E1,$65,$7E,$12,$3F,$1B
	.dc.b	$F0,$BF,$8E,$F8,$49,$3F,$88,$8E,$4F,$E2,$AE,$BE,$2F,$85,$EA,$FE
	.dc.b	$3B,$0E,$AF,$9B,$0E,$AF,$E4,$B0,$EA,$FA,$B0,$FC,$EF,$C2,$48,$FC
	.dc.b	$BB,$FA,$0F,$85,$FD,$0F,$C2,$49,$FC,$DB,$FE,$0F,$FD,$97,$FC,$3B
	.dc.b	$1F,$EC,$97,$FC,$2F,$C4,$EA,$FF,$6B,$C6,$1B,$F0,$7C,$7E,$0F,$B5
	.dc.b	$FE,$9B,$1A,$FD,$6B,$FF,$0B,$BF,$43,$1F,$13,$4F,$07,$C2,$F8,$F0
	.dc.b	$91,$E3,$7B,$3E,$5F,$83,$F4,$11,$F4,$7C,$27,$4D,$3F,$7B,$85,$D3
	.dc.b	$D3,$8F,$4B,$F2,$63,$6F,$DD,$1F,$BB,$C9,$A7,$C5,$F8,$59,$3F,$15
	.dc.b	$F9,$1F,$09,$F2,$7E,$17,$E5,$7C,$1C,$AF,$95,$1F,$37,$E1,$24,$7C
	.dc.b	$D7,$E8,$7C,$27,$28,$FA,$17,$5F,$91,$F0,$9D,$5F,$A9,$87,$57,$E6
	.dc.b	$61,$D1,$F9,$D8,$5D,$3D,$7E,$0B,$56,$7F,$7F,$E1,$7F,$01,$F0,$B3
	.dc.b	$1F,$C0,$47,$E0,$FE,$17,$F0,$9F,$09,$27,$F0,$91,$F8,$7F,$85,$FC
	.dc.b	$47,$C2,$49,$FC,$44,$75,$79,$90,$FC,$17,$C2,$F5,$7F,$1D,$7D,$5F
	.dc.b	$C3,$5F,$57,$F2,$57,$D5,$FC,$55,$FE,$5F,$E1,$3F,$31,$F0,$B2,$7F
	.dc.b	$31,$1F,$E6,$9F,$D9,$BF,$C2,$FC,$1F,$09,$7C,$1E,$4D,$56,$FE,$1B
	.dc.b	$87,$C2,$48,$1F,$C3,$AB,$FC,$00,$3E,$1B,$C5,$1E,$03,$78,$77,$8E
	.dc.b	$B1,$8E,$EE,$7C,$AB,$F1,$CE,$4F,$E9,$4E,$1F,$01,$E5,$35,$1C,$70
	.dc.b	$F3,$47,$97,$1F,$81,$F0,$DF,$07,$19,$3C,$30,$4F,$92,$3D,$99,$FC
	.dc.b	$37,$47,$73,$E0,$61,$DB,$3F,$95,$8F,$CD,$F8,$59,$5D,$5F,$A1,$F0
	.dc.b	$DF,$47,$E1,$64,$F8,$5F,$A9,$F0,$BF,$57,$E1,$24,$F9,$5F,$B1,$F0
	.dc.b	$BC,$A3,$A1,$0F,$A5,$7D,$1F,$A9,$77,$2F,$BD,$87,$49,$FB,$18,$59
	.dc.b	$FC,$16,$1F,$85,$F8,$6F,$C3,$7C,$34,$BF,$8D,$1F,$89,$F8,$6F,$C5
	.dc.b	$7C,$34,$9F,$95,$1F,$8D,$F8,$69,$1F,$35,$FC,$87,$C3,$7E,$47,$D9
	.dc.b	$ED,$3D,$0F,$A4,$F5,$7E,$A5,$F5,$7F,$1D,$7D,$1F,$C8,$5D,$CF,$F3
	.dc.b	$58,$7E,$77,$E1,$65,$7D,$D7,$F4,$1F,$07,$FA,$1F,$83,$93,$F8,$17
	.dc.b	$FD,$99,$FB,$3F,$F8,$00,$7C,$36,$3E,$16,$57,$E3,$A7,$2F,$E2,$6F
	.dc.b	$AB,$4D,$7C,$37,$67,$F4,$D8,$5C,$FF,$88,$DF,$7C,$7A,$D7,$F7,$88
	.dc.b	$FF,$14,$F7,$1E,$F1,$9C,$3C,$E4,$7C,$9F,$0D,$D7,$C0,$FD,$5F,$7F
	.dc.b	$F8,$DE,$7D,$C7,$A6,$3D,$F8,$F8,$32,$F1,$EB,$8A,$F7,$7C,$37,$C4
	.dc.b	$F8,$59,$18,$3E,$17,$C2,$FC,$8F,$84,$91,$83,$E3,$7C,$2F,$CC,$23
	.dc.b	$F2,$7E,$1B,$E7,$E1,$23,$A1,$F0,$7E,$1B,$E9,$E1,$D2,$3E,$26,$18
	.dc.b	$7A,$90,$F9,$18,$7D,$AC,$3E,$DF,$C3,$49,$FB,$0F,$DE,$F8,$4F,$BF
	.dc.b	$F0,$B2,$7E,$E3,$F8,$2F,$84,$FC,$1F,$C2,$C9,$FC,$03,$F8,$6F,$86
	.dc.b	$E5,$1D,$08,$7D,$9B,$E8,$FD,$BB,$BC,$3F,$80,$C3,$A4,$FD,$FC,$2C
	.dc.b	$FE,$13,$0F,$C9,$FC,$34,$AF,$C8,$3F,$96,$F8,$6F,$CB,$FC,$27,$F3
	.dc.b	$29,$FF,$1F,$FA,$FF,$E1,$66,$FC,$26,$9F,$E4,$B1,$FE,$15,$7F,$CE
	.dc.b	$8D,$ED,$34,$FE,$FB,$1F,$9F,$BF,$13,$F9,$5C,$3C,$4F,$E8,$F0,$F1
	.dc.b	$65,$FD,$B2,$FC,$4D,$C7,$C5,$5E,$6F,$84,$DF,$57,$F4,$0B,$F0,$46
	.dc.b	$27,$CB,$F0,$97,$4F,$2F,$C3,$78,$27,$F4,$DA,$E7,$0F,$F1,$4E,$7F
	.dc.b	$AC,$D3,$1A,$4B,$F3,$1F,$0B,$3E,$04,$7E,$2E,$7C,$FE,$1A,$47,$C8
	.dc.b	$7B,$7C,$37,$7F,$86,$F0,$7C,$24,$AF,$9B,$1E,$3F,$85,$F2,$7C,$2C
	.dc.b	$9F,$A3,$1C,$9F,$A7,$75,$FB,$3F,$09,$D5,$FA,$F8,$75,$7E,$EE,$1D
	.dc.b	$5F,$B7,$87,$57,$F0,$38,$7C,$3F,$86,$F8,$9F,$09,$27,$F0,$31,$F1
	.dc.b	$FE,$17,$E4,$7C,$24,$9F,$C2,$C7,$CB,$F8,$5F,$99,$F0,$92,$7F,$13
	.dc.b	$1F,$3F,$E1,$7B,$47,$43,$F1,$AF,$57,$F1,$F7,$D5,$FC,$AD,$F5,$7F
	.dc.b	$27,$7D,$5F,$CC,$DF,$DB,$E6,$91,$F9,$87,$EF,$7C,$27,$DF,$F8,$59
	.dc.b	$3F,$9C,$7F,$0D,$FB,$47,$F8,$73,$B7,$FC,$16,$FF,$81,$FA,$BF,$C7
	.dc.b	$EB,$E0,$B6,$70,$F8,$4E,$B8,$7E,$9B,$1A,$FF,$AA,$A9,$F0,$9E,$08
	.dc.b	$FD,$7E,$57,$1E,$B1,$CC,$FF,$09,$D7,$76,$E3,$97,$A5,$F6,$AF,$D4
	.dc.b	$61,$EA,$8E,$99,$7E,$97,$1B,$7F,$EE,$BC,$7D,$CD,$CB,$D6,$F6,$3C
	.dc.b	$2F,$C4,$F8,$6F,$8B,$F0,$D2,$7C,$B1,$F2,$3E,$1A,$47,$A1,$F9,$7F
	.dc.b	$0D,$F3,$3E,$1B,$E6,$FC,$34,$AF,$6C,$7D,$0F,$86,$ED,$1D,$0F,$82
	.dc.b	$F4,$7E,$55,$DE,$0F,$A7,$87,$59,$F8,$F8,$75,$7E,$86,$1F,$73,$E1
	.dc.b	$BE,$EF,$C2,$C9,$F9,$B1,$F8,$0F,$86,$FC,$0F,$C2,$49,$FA,$31,$F8
	.dc.b	$4F,$85,$FC,$2F,$C2,$C9,$FA,$B1,$F8,$8F,$85,$ED,$1D,$0F,$B2,$F5
	.dc.b	$7E,$DD,$F5,$7F,$09,$7D,$5F,$BF,$7D,$5F,$C4,$5F,$E5,$3E,$1A,$47
	.dc.b	$E1,$1F,$CB,$FC,$37,$E6,$3E,$1A,$4F,$E2,$1F,$CD,$FC,$37,$F2,$1F
	.dc.b	$DA,$7F,$D6,$33,$FE,$FC,$7F,$C0,$7C,$5F,$E9,$A6,$F5,$7F,$9A,$9D
	.dc.b	$C7,$E8,$30,$F0,$D7,$E6,$F1,$F0,$BD,$C8,$7E,$7F,$1F,$0B,$FE,$6C
	.dc.b	$BE,$78,$75,$F8,$3F,$21,$E5,$7C,$DF,$07,$D2,$BF,$51,$8E,$1E,$B9
	.dc.b	$7E,$CB,$1F,$7C,$7E,$B7,$3E,$4F,$7C,$7A,$BE,$1C,$7E,$17,$C3,$48
	.dc.b	$F2,$3F,$17,$E1,$BE,$37,$C2,$49,$F4,$3F,$27,$E1,$7E,$57,$C2,$7C
	.dc.b	$BF,$84,$95,$ED,$8E,$4F,$BE,$EB,$F3,$FE,$13,$A3,$F4,$70,$BB,$7D
	.dc.b	$1C,$7A,$CF,$C7,$C7,$AB,$F5,$F1,$FB,$5F,$09,$23,$E6,$3F,$77,$E1
	.dc.b	$7E,$F7,$C2,$C9,$FA,$0F,$E0,$7E,$17,$F0,$5F,$0B,$27,$EA,$3F,$85
	.dc.b	$F8,$5E,$51,$F5,$EE,$BF,$87,$F8,$5E,$8F,$E2,$70,$BA,$FE,$27,$1E
	.dc.b	$93,$F8,$DC,$6C,$FE,$37,$1F,$C9,$7C,$2F,$E4,$FE,$16,$5F,$E1,$63
	.dc.b	$F2,$DF,$07,$F9,$7F,$83,$93,$F8,$98,$FE,$F5,$FB,$57,$F8,$13,$9F
	.dc.b	$FD,$7E,$FF,$84,$49,$F0,$9A,$F8,$3E,$F7,$D4,$FF,$5E,$CF,$F8,$51
	.dc.b	$FD,$15,$F6,$7F,$49,$87,$89,$DD,$F8,$3E,$12,$46,$0E,$78,$77,$69
	.dc.b	$DB,$E1,$79,$3D,$0F,$D1,$5E,$36,$67,$EA,$C7,$AD,$27,$2B,$E1,$9A
	.dc.b	$7B,$A3,$D7,$F0,$DF,$13,$E1,$A4,$60,$F7,$FC,$27,$C8,$23,$F0,$BE
	.dc.b	$17,$E5,$FC,$27,$CC,$F8,$59,$3F,$1A,$3E,$7E,$12,$3A,$1E,$5D,$7E
	.dc.b	$8F,$C3,$74,$8C,$1E,$7C,$D9,$67,$5F,$5E,$BE,$AD,$3C,$37,$F5,$34
	.dc.b	$F0,$BC,$A7,$DF,$A7,$E0,$3E,$17,$F0,$3F,$0D,$2B,$ED,$C7,$E1,$3E
	.dc.b	$13,$F0,$BF,$0B,$27,$EF,$C7,$E2,$3E,$12,$47,$E0,$9F,$C5,$FC,$2F
	.dc.b	$58,$E6,$43,$C5,$BB,$E1,$BB,$47,$E2,$B0,$EC,$EB,$0E,$AF,$0C,$2E
	.dc.b	$FF,$03,$06,$3F,$9A,$F8,$6F,$CD,$FC,$2C,$C3,$F2,$B1,$F9,$EF,$84
	.dc.b	$FF,$2C,$77,$FF,$76,$E1,$37,$14,$EB,$E0,$E6,$03,$E5,$57,$F8,$67
	.dc.b	$9F,$18,$72,$F8,$6F,$15,$78,$08,$70,$F8,$6F,$13,$CF,$6F,$B3,$87
	.dc.b	$8C,$F1,$65,$E0,$9E,$F9,$5A,$DE,$1E,$AC,$3B,$D7,$CC,$D3,$0F,$ED
	.dc.b	$30,$F7,$B0,$C1,$E1,$CF,$D3,$97,$AE,$3F,$48,$DC,$FE,$2E,$7F,$23
	.dc.b	$2F,$93,$F0,$92,$F9,$1C,$1F,$9B,$F0,$BF,$3B,$E1,$3E,$7F,$C2,$4A
	.dc.b	$DE,$3F,$42,$7E,$9F,$C1,$C9,$CE,$68,$FA,$FF,$09,$D2,$73,$8D,$7E
	.dc.b	$9F,$C2,$F7,$78,$E5,$DD,$FA,$F8,$F5,$7E,$CE,$17,$8B,$E8,$E3,$F8
	.dc.b	$4F,$84,$95,$E2,$7F,$0F,$F0,$BF,$88,$F8,$39,$3E,$67,$F1,$7F,$09
	.dc.b	$F8,$CF,$83,$93,$EA,$7F,$1F,$F0,$9C,$A3,$A1,$0F,$C5,$DF,$47,$F1
	.dc.b	$B7,$72,$FC,$2E,$1D,$27,$F2,$38,$59,$FC,$4E,$1F,$9C,$F8,$59,$5F
	.dc.b	$29,$FC,$FF,$C2,$FE,$83,$E0,$FF,$43,$F0,$72,$BE,$7C,$7E,$93,$E0
	.dc.b	$FF,$D7,$9F,$B5,$FF,$83,$27,$F9,$33,$79,$FF,$4F,$3D,$27,$EB,$E1
	.dc.b	$5F,$E1,$D5,$FE,$20,$7F,$17,$7D,$DF,$D4,$E1,$E3,$7E,$FD,$F8,$3A
	.dc.b	$E4,$60,$FC,$3D,$F3,$C7,$F6,$F8,$F7,$D3,$F9,$0C,$E7,$A1,$9F,$79
	.dc.b	$FD,$6E,$7D,$F1,$FD,$CE,$7D,$9F,$3E,$7F,$05,$EF,$F0,$BF,$08,$8F
	.dc.b	$8B,$E1,$3E,$2F,$C2,$7C,$6F,$83,$93,$E6,$8F,$93,$F0,$72,$3D,$2F
	.dc.b	$CC,$F8,$2F,$9B,$7D,$27,$E3,$E1,$87,$91,$F0,$30,$EA,$FC,$CC,$3A
	.dc.b	$3F,$3B,$0B,$A7,$C2,$C7,$EC,$FC,$1F,$DA,$F8,$49,$7F,$2A,$3E,$EF
	.dc.b	$C2,$48,$F9,$AF,$E0,$3E,$13,$F0,$3F,$07,$F8,$2F,$83,$95,$F4,$A3
	.dc.b	$F0,$BF,$07,$DA,$3A,$1F,$59,$E8,$FE,$0A,$EE,$7F,$5B,$0E,$B3,$F7
	.dc.b	$30,$EA,$FE,$23,$0F,$C8,$FC,$1F,$E4,$BE,$12,$4F,$E0,$A3,$F2,$BF
	.dc.b	$09,$F9,$6E,$89,$3F,$86,$8F,$CC,$FC,$24,$8F,$C4,$BF,$9C,$F8,$1F
	.dc.b	$E7,$51,$FD,$8F,$F6,$CF,$F1,$03,$F9,$EC,$3C,$2F,$F1,$68,$CF,$73
	.dc.b	$B5,$8F,$82,$BF,$3F,$97,$91,$EE,$43,$F4,$79,$78,$5E,$9F,$05,$E2
	.dc.b	$E0,$FE,$F3,$97,$62,$7C,$2F,$3A,$EA,$7E,$8B,$0C,$57,$E8,$30,$F7
	.dc.b	$4F,$63,$F4,$FA,$F5,$EC,$FE,$AB,$8F,$F8,$D6,$FB,$FC,$26,$6F,$FC
	.dc.b	$45,$28,$C6,$C8,$E1,$7C,$BE,$0F,$30,$69,$8F,$84,$CF,$C1,$CC,$7E
	.dc.b	$53,$C3,$E0,$F8,$FC,$2C,$9F,$9C,$F4,$F8,$3E,$71,$F4,$70,$AF,$D7
	.dc.b	$C3,$A3,$F6,$70,$BA,$FC,$AC,$7C,$BF,$09,$2B,$ED,$3E,$8F,$85,$F4
	.dc.b	$FC,$27,$AB,$E1,$65,$7D,$F8,$F6,$FC,$1C,$8F,$C1,$3F,$03,$E1,$39
	.dc.b	$47,$42,$1F,$8B,$BE,$AF,$E2,$6F,$AB,$F8,$FB,$E8,$FE,$46,$EE,$5F
	.dc.b	$82,$C3,$E6,$FC,$27,$CE,$F8,$49,$7F,$93,$8F,$A3,$F0,$BF,$4B,$E0
	.dc.b	$A4,$FE,$5E,$3E,$AF,$C2,$FD,$6F,$83,$93,$F9,$B8,$EA,$F3,$21,$FE
	.dc.b	$45,$FD,$B7,$FC,$49,$CF,$FE,$35,$8F,$D1,$61,$D9,$FD,$46,$1D,$9F
	.dc.b	$D3,$61,$D9,$FD,$66,$1F,$8A,$C3,$9E,$3E,$09,$EC,$7A,$8E,$A4,$3A
	.dc.b	$9E,$73,$A9,$0E,$A7,$F8,$2A,$3F,$C2,$C4,$F8,$4F,$07,$C2,$78,$7E
	.dc.b	$17,$C5,$F0,$B2,$F8,$47,$97,$E1,$64,$73,$7D,$1F,$07,$CA,$3B,$61
	.dc.b	$EB,$30,$FC,$7A,$E8,$FC,$9B,$F0,$3C,$C8,$7C,$DB,$EA,$FA,$7E,$13
	.dc.b	$E3,$7C,$24,$8C,$1E,$DF,$84,$F9,$5F,$0B,$23,$07,$C1,$F8,$5F,$9D
	.dc.b	$F0,$B2,$30,$7C,$5F,$85,$FA,$58,$C8,$C1,$C8,$FC,$05,$F5,$7F,$05
	.dc.b	$87,$D9,$C3,$A4,$7E,$17,$0C,$3F,$47,$E1,$7E,$F7,$C2,$C8,$C1,$F5
	.dc.b	$7E,$17,$F0,$44,$7E,$C7,$C2,$FE,$17,$E1,$64,$7D,$B7,$F1,$1F,$0B
	.dc.b	$F8,$9C,$24,$F4,$3F,$2D,$F0,$BC,$DF,$CC,$61,$7C,$0F,$F5,$1F,$43
	.dc.b	$2E,$B5,$F9,$DC,$AD,$6B,$2F,$CB,$63,$2B,$F1,$8F,$E6,$7E,$17,$A7
	.dc.b	$9B,$FE,$7E,$C9,$9E,$2F,$FD,$95,$1D,$DF,$51,$0F,$01,$F9,$7C,$3F
	.dc.b	$96,$9F,$E2,$B2,$56,$3E,$1F,$3F,$0B,$2B,$FC,$CF,$48,$E3,$F0,$DC
	.dc.b	$BE,$16,$56,$9E,$BF,$0D,$CE,$38,$E3,$E0,$30,$FF,$B0,$F0,$E8,$F6
	.dc.b	$C3,$C0,$E7,$0E,$AE,$F0,$BB,$7F,$B4,$2D,$8F,$B3,$E1,$A5,$FA,$1F
	.dc.b	$7F,$C3,$7C,$0F,$85,$F8,$3F,$0D,$2B,$DB,$1F,$13,$E1,$7E,$2F,$C2
	.dc.b	$C9,$F8,$31,$C9,$E8,$43,$D5,$7D,$5F,$8F,$7D,$5F,$75,$F5,$7E,$5D
	.dc.b	$F5,$7E,$15,$FD,$4F,$84,$91,$F4,$1F,$AF,$F0,$BF,$63,$E1,$BE,$CF
	.dc.b	$C2,$CA,$FA,$B1,$F7,$3E,$1A,$47,$D8,$7E,$FF,$C2,$F2,$8E,$84,$3E
	.dc.b	$95,$F5,$7E,$FD,$F5,$7E,$B5,$F4,$7E,$C5,$DE,$1F,$A5,$87,$59,$E6
	.dc.b	$7E,$23,$0F,$C7,$FC,$1F,$58,$FC,$06,$3D,$6F,$ED,$63,$CF,$9B,$FD
	.dc.b	$B7,$C6,$CF,$18,$3E,$03,$FC,$DB,$A7,$E6,$BE,$0F,$F3,$7F,$07,$2B
	.dc.b	$F2,$F1,$FE,$0A,$FD,$BB,$FC,$52,$3E,$17,$06,$7E,$1B,$5F,$0B,$2F
	.dc.b	$FD,$8B,$5F,$E1,$07,$C0,$FF,$85,$5C,$FC,$27,$85,$EE,$60,$E3,$5D
	.dc.b	$9E,$77,$DD,$EB,$7D,$D7,$03,$7F,$F7,$DE,$3E,$1D,$4F,$86,$B7,$3A
	.dc.b	$F8,$4F,$77,$C2,$49,$F4,$C7,$C1,$F8,$4F,$85,$F0,$B2,$7D,$B1,$F1
	.dc.b	$7E,$17,$E3,$7C,$2C,$9F,$83,$1F,$26,$E4,$74,$3E,$2F,$C2,$75,$7E
	.dc.b	$3E,$1D,$5F,$29,$F2,$30,$EC,$FC,$CC,$3B,$3A,$D3,$DB,$8F,$D7,$F8
	.dc.b	$59,$3F,$49,$FB,$5F,$0B,$F6,$FE,$17,$EE,$7B,$F2,$BE,$BC,$7D,$FF
	.dc.b	$85,$91,$F6,$9F,$C1,$7C,$1F,$28,$E8,$43,$C5,$C4,$C3,$B4,$7E,$0B
	.dc.b	$0E,$CE,$B0,$EA,$F0,$C2,$F1,$FD,$98,$FC,$3E,$3D,$DE,$B9,$7B,$1E
	.dc.b	$F8,$F6,$7C,$38,$F6,$BE,$5A,$7E,$03,$2E,$BC,$7F,$84,$CE,$E1,$D8
	.dc.b	$FC,$BF,$C2,$7E,$7C,$8F,$E6,$BE,$0F,$F4,$5F,$09,$23,$F3,$AF,$F9
	.dc.b	$BB,$F6,$FF,$F4,$9C,$7D,$35,$F0,$79,$F8,$59,$23,$51,$D9,$F9,$38
	.dc.b	$5D,$7D,$3C,$4C,$BC,$15,$F3,$B2,$B5,$EE,$CB,$C2,$DC,$7B,$D7,$8F
	.dc.b	$2F,$34,$79,$FE,$13,$D1,$F0,$92,$BC,$97,$CE,$7D,$9F,$09,$27,$3B
	.dc.b	$FA,$78,$59,$AB,$E1,$7C,$2F,$C3,$F8,$3F,$89,$F0,$B3,$1E,$27,$F1
	.dc.b	$1B,$4F,$8E,$33,$87,$4C,$B1,$1F,$2F,$E0,$FC,$4F,$CD,$CA,$E9,$F0
	.dc.b	$73,$F0,$4F,$2C,$99,$F8,$63,$EA,$67,$F5,$FE,$13,$EC,$7C,$24,$9E
	.dc.b	$11,$F6,$FE,$12,$47,$37,$EF,$7C,$2F,$DF,$F8,$49,$3D,$DF,$C1,$7C
	.dc.b	$2F,$58,$F1,$61,$D5,$FC,$26,$1D,$1F,$C3,$61,$75,$FB,$58,$F4,$9F
	.dc.b	$C5,$63,$67,$EF,$63,$D6,$79,$9E,$83,$E0,$7C,$0F,$3E,$3F,$AD,$B3
	.dc.b	$2C,$57,$D1,$CB,$A5,$6B,$2B,$3F,$57,$1F,$CE,$7C,$2F,$E7,$7E,$16
	.dc.b	$4F,$CD,$8F,$D0,$7A,$5F,$A1,$F8,$59,$3F,$46,$3F,$49,$F0,$B2,$3E
	.dc.b	$A3,$FE,$D4,$FD,$C3,$F7,$D1,$FA,$7B,$EA,$FE,$A7,$0F,$0B,$CC,$FB
	.dc.b	$98,$76,$7F,$5D,$87,$57,$F6,$18,$5E,$0F,$E0,$38,$F6,$6F,$C2,$F8
	.dc.b	$49,$5F,$90,$9F,$F1,$4B,$FC,$AD,$7C,$27,$9C,$91,$BF,$84,$F5,$7C
	.dc.b	$24,$8E,$4F,$B7,$E1,$3D,$DF,$0B,$DA,$7A,$63,$87,$C1,$87,$47,$C0
	.dc.b	$FC,$5C,$3A,$47,$8F,$0B,$97,$B3,$1F,$95,$F0,$BF,$2F,$E1,$25,$FA
	.dc.b	$E3,$E7,$7C,$2F,$CF,$F8,$49,$3E,$F8,$FA,$5F,$0B,$23,$E1,$3F,$57
	.dc.b	$E1,$7A,$C7,$C5,$C3,$AB,$F1,$70,$EA,$FC,$9C,$3A,$BF,$27,$0E,$8F
	.dc.b	$CB,$C2,$E5,$F5,$31,$EB,$3C,$CF,$41,$F4,$FE,$17,$AB,$F4,$F0,$EB
	.dc.b	$5F,$73,$0E,$7D,$3F,$C4,$F8,$99,$63,$0F,$F0,$F4,$32,$FC,$87,$C1
	.dc.b	$C9,$FC,$03,$F9,$3F,$84,$FC,$A7,$C2,$7E,$57,$E1,$25,$7E,$16,$3F
	.dc.b	$31,$F0,$B2,$3F,$10,$FE,$6F,$E1,$39,$47,$42,$1F,$8B,$BE,$AF,$E3
	.dc.b	$EF,$AB,$F8,$FB,$E8,$FE,$46,$EE,$DF,$9A,$C3,$F5,$1F,$0B,$FD,$26
	.dc.b	$BF,$C0,$1E,$93,$E1,$31,$7F,$D8,$23,$13,$F9,$DA,$CF,$07,$F3,$99
	.dc.b	$FF,$1D,$4F,$F3,$07,$FC,$4C,$4F,$84,$DC,$7F,$91,$E3,$FC,$51,$7D
	.dc.b	$7E,$23,$3E,$8F,$F5,$56,$CB,$E1,$BC,$F7,$E1,$21,$E8,$F8,$4D,$E3
	.dc.b	$C3,$E1,$A4,$F8,$E3,$9E,$12,$3C,$CF,$6B,$EF,$F0,$D2,$7D,$4F,$8B
	.dc.b	$E1,$3C,$7F,$0B,$E4,$F8,$49,$5E,$F8,$F3,$FC,$2C,$8F,$84,$FA,$BE
	.dc.b	$1B,$94,$74,$21,$F4,$2F,$AB,$F2,$6F,$AB,$F5,$2F,$A3,$F5,$AE,$F0
	.dc.b	$7D,$5C,$3A,$CF,$33,$E9,$61,$F2,$BE,$17,$AC,$7D,$EC,$7A,$DF,$DD
	.dc.b	$C7,$9F,$17,$DB,$CA,$DF,$D7,$CB,$E9,$FC,$2C,$8F,$BC,$FD,$6F,$84
	.dc.b	$FA,$FF,$0B,$F6,$3E,$16,$57,$E0,$E3,$ED,$FC,$34,$8F,$C3,$3F,$7B
	.dc.b	$E1,$B9,$47,$42,$1F,$94,$BE,$AF,$E3,$6F,$AB,$F9,$8B,$E8,$FE,$6A
	.dc.b	$EE,$BF,$99,$C3,$F1,$7C,$FF,$8C,$F8,$59,$7F,$97,$8F,$C8,$FE,$E3
	.dc.b	$FE,$20,$DF,$FC,$28,$FF,$8C,$27,$FC,$16,$D7,$FC,$CC,$ED,$FF,$51
	.dc.b	$2F,$86,$E7,$F0,$92,$78,$C7,$89,$F0,$10,$E7,$F0,$BE,$27,$C7,$F0
	.dc.b	$D1,$F0,$9E,$69,$E8,$F8,$AB,$0F,$9B,$E1,$7D,$7F,$0B,$23,$07,$AB
	.dc.b	$E1,$3D,$E4,$7D,$BF,$0B,$F0,$BE,$1A,$47,$C0,$7E,$2F,$C2,$FC,$6C
	.dc.b	$24,$F4,$3E,$1F,$C3,$73,$7E,$2E,$17,$8F,$FC,$D3,$BF,$F8,$AA,$7A
	.dc.b	$10,$F9,$78,$F6,$7A,$4F,$0C,$3B,$C7,$63,$D4,$43,$B9,$EA,$3A,$47
	.dc.b	$82,$7D,$71,$F6,$31,$FB,$FF,$09,$23,$EE,$BF,$82,$F8,$4F,$C1,$FC
	.dc.b	$24,$9F,$C0,$BF,$86,$F8,$4F,$C3,$FC,$37,$E2,$3E,$16,$57,$E1,$A3
	.dc.b	$93,$D0,$87,$E2,$2F,$A3,$F8,$AB,$B9,$FC,$6C,$3A,$CF,$E4,$30,$EA
	.dc.b	$FE,$43,$0F,$CB,$FC,$2C,$8F,$CA,$BF,$9A,$F8,$4F,$CD,$FC,$2C,$9F
	.dc.b	$CC,$BF,$9E,$F8,$5F,$E5,$FF,$B9,$7F,$86,$FA,$FF,$CE,$4B,$E1,$72
	.dc.b	$6A,$3A,$5F,$62,$1F,$A2,$C3,$B3,$FA,$4C,$2E,$BF,$80,$C7,$94,$FF
	.dc.b	$9D,$31,$EC,$DC,$72,$FF,$B4,$63,$FC,$53,$5F,$EB,$06,$E3,$E6,$F8
	.dc.b	$6F,$4F,$C3,$7A,$BE,$1B,$D7,$F0,$D3,$1E,$D1,$EE,$F8,$6F,$7F,$C2
	.dc.b	$C9,$F1,$47,$27,$C9,$87,$C4,$30,$F8,$EB,$A3,$E5,$BF,$03,$C8,$F6
	.dc.b	$5F,$57,$D5,$7C,$B3,$F5,$63,$72,$F9,$FF,$0D,$CA,$FA,$1F,$17,$1C
	.dc.b	$5F,$C1,$C3,$A3,$F0,$F1,$F6,$57,$CB,$F8,$4F,$B6,$47,$E7,$7C,$2F
	.dc.b	$DE,$F8,$59,$1F,$45,$FC,$0F,$C2,$7E,$0B,$E1,$BF,$07,$F0,$B2,$BE
	.dc.b	$B4,$7E,$1A,$FA,$47,$D5,$C3,$0F,$32,$1F,$67,$0E,$AF,$23,$F0,$17
	.dc.b	$D5,$FB,$D7,$F9,$2F,$85,$FC,$9F,$C2,$49,$FC,$34,$7E,$5B,$E1,$7F
	.dc.b	$2F,$F0,$B2,$7F,$15,$1F,$9A,$F8,$5F,$CD,$FC,$24,$9F,$C7,$47,$57
	.dc.b	$F2,$18,$75,$7F,$1D,$87,$57,$F2,$98,$75,$7F,$25,$87,$57,$F3,$18
	.dc.b	$75,$7F,$2D,$87,$F8,$63,$F7,$3F,$F1,$2C,$38,$5C,$BF,$ED,$D9,$FF
	.dc.b	$12,$CE,$63,$1F,$05,$C7,$E1,$B9,$7C,$27,$3F,$84,$97,$FE,$DA,$C1
	.dc.b	$3D,$FE,$13,$C1,$F0,$B2,$B5,$1C,$DE,$18,$57,$C0,$3E,$19,$83,$E1
	.dc.b	$BC,$33,$DF,$2F,$0B,$C2,$B7,$97,$86,$3C,$99,$78,$5D,$61,$FE,$37
	.dc.b	$7D,$18,$78,$75,$F8,$71,$E1,$9F,$8F,$18,$C3,$C7,$5E,$2A,$E7,$8F
	.dc.b	$8A,$B5,$7F,$0A,$7D,$E7,$88,$87,$8B,$3F,$A9,$F0,$DF,$57,$E1,$24
	.dc.b	$FC,$B8,$FB,$1F,$0B,$23,$E7,$3F,$6F,$E1,$7E,$E7,$C2,$FD,$DF,$86
	.dc.b	$95,$F4,$E3,$AB,$CC,$87,$C2,$F8,$6E,$8F,$C4,$BB,$A7,$AB,$0E,$B3
	.dc.b	$F7,$70,$EA,$FC,$AC,$3F,$19,$F0,$DF,$8D,$F8,$59,$3F,$83,$8F,$C8
	.dc.b	$7C,$37,$E4,$7E,$1A,$4F,$E1,$E3,$F2,$9F,$0D,$23,$F1,$4F,$E5,$FE
	.dc.b	$17,$AC,$73,$21,$F6,$BE,$1B,$AB,$F9,$1B,$EA,$FD,$EB,$E8,$FE,$02
	.dc.b	$EE,$BF,$5B,$0F,$F3,$DF,$67,$F8,$97,$9C,$D2,$7C,$2E,$BE,$16,$62
	.dc.b	$3F,$21,$3F,$E2,$B7,$18,$7E,$1E,$1F,$0B,$DB,$E1,$BB,$FC,$0C,$BD
	.dc.b	$F8,$3C,$F3,$F0,$D7,$93,$3F,$2D,$43,$67,$E9,$71,$FF,$1F,$65,$DF
	.dc.b	$3F,$2C,$F9,$1F,$6E,$7E,$18,$E3,$8F,$BB,$3F,$14,$7A,$B2,$F0,$3A
	.dc.b	$C7,$AE,$7E,$56,$C7,$87,$E1,$BB,$E7,$E1,$8E,$19,$F8,$1E,$58,$76
	.dc.b	$9F,$89,$8F,$8E,$7E,$4F,$C2,$7D,$2F,$86,$91,$E1,$8F,$AB,$F0,$9F
	.dc.b	$5B,$E1,$24,$F9,$63,$EC,$FC,$37,$DA,$F8,$59,$3E,$98,$FB,$B7,$23
	.dc.b	$A1,$ED,$F8,$5E,$AF,$BF,$0E,$AF,$DC,$C3,$AB,$F0,$F0,$EA,$FE,$03
	.dc.b	$0F,$C4,$FC,$2F,$E2,$BE,$16,$4F,$C9,$8F,$C6,$FC,$27,$E3,$BE,$16
	.dc.b	$4F,$CD,$8F,$C8,$FC,$2F,$E4,$BE,$16,$4F,$D1,$8E,$AF,$32,$1F,$90
	.dc.b	$F8,$5E,$AF,$D7,$BE,$AF,$E5,$2F,$AB,$F6,$EF,$AB,$F9,$8B,$FD,$0F
	.dc.b	$C2,$FF,$B7,$3F,$74,$FF,$17,$4F,$F3,$16,$F5,$7E,$92,$75,$F0,$D2
	.dc.b	$BF,$17,$5F,$E2,$67,$F9,$93,$FE,$22,$79,$7C,$1F,$5F,$84,$91,$C2
	.dc.b	$3C,$2F,$E3,$EE,$BC,$BE,$0F,$B3,$D3,$0B,$D3,$FC,$CF,$1E,$D9,$16
	.dc.b	$3D,$AF,$C7,$97,$B6,$3A,$DF,$ED,$72,$BD,$DF,$B1,$C1,$9F,$68,$C6
	.dc.b	$77,$2F,$89,$F0,$BC,$DF,$83,$9F,$7C,$BC,$99,$FC,$BF,$84,$F9,$9F
	.dc.b	$09,$2F,$E3,$C7,$CF,$F8,$5F,$A1,$F0,$72,$7E,$5C,$7D,$3F,$84,$FA
	.dc.b	$9F,$09,$27,$E7,$C7,$57,$99,$0F,$A3,$F0,$BD,$5F,$AB,$7D,$5F,$AB
	.dc.b	$7D,$5F,$B3,$7D,$5F,$B3,$7F,$83,$F8,$49,$1F,$79,$FC,$37,$C2,$FE
	.dc.b	$1F,$E1,$24,$FE,$09,$FC,$57,$C2,$FE,$2F,$E1,$24,$FE,$19,$FC,$77
	.dc.b	$C2,$F2,$8E,$84,$3F,$13,$7D,$1F,$C5,$DD,$E1,$FC,$06,$1D,$27,$F1
	.dc.b	$F8,$59,$FC,$26,$1F,$9B,$F3,$3F,$39,$F0,$B2,$FF,$2F,$1F,$CE,$5F
	.dc.b	$F0,$47,$EE,$BF,$DF,$BA,$3F,$99,$36,$3F,$C2,$4D,$EC,$3F,$E2,$6C
	.dc.b	$7F,$C0,$8F,$F7,$F9,$E7,$F0,$3D,$67,$2D,$BE,$BF,$09,$E5,$AF,$D4
	.dc.b	$E5,$E5,$7F,$53,$96,$DF,$0F,$C2,$DC,$FC,$2D,$CF,$C8,$78,$F1,$EB
	.dc.b	$5F,$CF,$A1,$96,$CE,$19,$78,$67,$B9,$F1,$0E,$1F,$05,$D7,$09,$1C
	.dc.b	$DF,$07,$C3,$72,$8E,$D7,$D5,$F0,$5F,$97,$E1,$BC,$DF,$0D,$27,$C9
	.dc.b	$1E,$9F,$84,$F5,$7C,$2C,$9F,$44,$7B,$7E,$12,$47,$AD,$F8,$1F,$09
	.dc.b	$D6,$3D,$D8,$75,$7E,$DE,$1D,$5F,$85,$87,$57,$EF,$E1,$D1,$FC,$0E
	.dc.b	$17,$8F,$ED,$E3,$F3,$7E,$0F,$E7,$7C,$24,$BF,$99,$1F,$47,$E1,$7E
	.dc.b	$97,$C2,$49,$FA,$11,$F5,$7E,$17,$EB,$7C,$1C,$9F,$A9,$1D,$5E,$64
	.dc.b	$3F,$27,$F0,$9D,$5F,$B5,$7D,$5F,$CB,$DF,$57,$EF,$5F,$57,$F3,$77
	.dc.b	$F8,$5F,$84,$FC,$37,$C2,$49,$FC,$24,$7E,$27,$E1,$7F,$19,$FB,$B7
	.dc.b	$F8,$84,$FF,$11,$75,$23,$3F,$0D,$AF,$84,$98,$7F,$82,$AF,$FC,$07
	.dc.b	$DD,$FA,$4C,$7B,$4F,$EB,$32,$F1,$3F,$93,$C7,$BB,$0C,$1C,$F1,$EA
	.dc.b	$FF,$9E,$F0,$FF,$16,$36,$39,$BF,$B8,$BB,$C7,$FE,$75,$AF,$F1,$7C
	.dc.b	$F4,$39,$1D,$CF,$DF,$DF,$B1,$F5,$65,$E0,$AE,$58,$7C,$4F,$84,$F8
	.dc.b	$BF,$09,$2F,$E0,$47,$C8,$F8,$5F,$93,$F0,$92,$7E,$24,$7C,$CF,$85
	.dc.b	$91,$F1,$DF,$9F,$F0,$7F,$43,$E1,$3E,$8F,$C1,$CA,$F9,$91,$F5,$3E
	.dc.b	$13,$EA,$FC,$2C,$9F,$A1,$1F,$63,$E1,$24,$7D,$37,$ED,$FC,$2F,$DC
	.dc.b	$F8,$3F,$BB,$F0,$92,$BE,$C4,$7E,$03,$E0,$FF,$03,$F0,$92,$7E,$E4
	.dc.b	$7E,$13,$E1,$64,$7D,$F7,$F0,$FF,$0B,$FC,$4D,$FF,$02,$9F,$E0,$1E
	.dc.b	$6F,$F5,$33,$7D,$0F,$F2,$E3,$AF,$82,$E1,$F0,$33,$0F,$F2,$9C,$FF
	.dc.b	$7A,$1F,$05,$D4,$91,$AF,$84,$F0,$61,$E1,$F8,$49,$3C,$A3,$C9,$8F
	.dc.b	$58,$FF,$62,$61,$FE,$04,$7B,$C7,$47,$38,$5C,$BD,$38,$F6,$9F,$2E
	.dc.b	$3D,$9E,$78,$F6,$7D,$39,$77,$BF,$83,$8F,$3A,$F8,$38,$DB,$FF,$6E
	.dc.b	$65,$FE,$0E,$7E,$0F,$C1,$FC,$B2,$3F,$13,$E1,$3E,$77,$C1,$7C,$FF
	.dc.b	$83,$93,$F2,$23,$E9,$7C,$14,$8F,$96,$FD,$5F,$83,$E7,$1F,$0B,$0B
	.dc.b	$AF,$D6,$C7,$AC,$FD,$2C,$7A,$BF,$23,$1E,$8F,$CA,$C6,$DF,$DE,$C7
	.dc.b	$9E,$9F,$7B,$2B,$7E,$EC,$BA,$BF,$4B,$2E,$B8,$7E,$1B,$2E,$78,$7E
	.dc.b	$1B,$2B,$7F,$1B,$2F,$C7,$7C,$24,$8F,$C3,$BF,$91,$F8,$4F,$C9,$7C
	.dc.b	$27,$E4,$FE,$12,$57,$E3,$23,$F2,$DF,$09,$F9,$7F,$84,$93,$F9,$08
	.dc.b	$E8,$FE,$1B,$0B,$AF,$E6,$B1,$EB,$3F,$96,$C7,$AB,$F8,$CC,$7A,$BF
	.dc.b	$9A,$C7,$AB,$F9,$0C,$79,$E9,$FA,$2C,$AC,$FE,$0B,$1E,$AF,$E5,$B1
	.dc.b	$EB,$97,$EA,$B2,$EB,$3F,$9A,$CB,$AC,$FE,$BB,$2F,$D9,$7C,$24,$AF
	.dc.b	$D5,$BF,$B5,$F8,$3F,$DB,$7C,$27,$F8,$27,$F7,$7F,$F0,$D1,$FE,$1B
	.dc.b	$E5,$46,$7E,$0E,$61,$FE,$C6,$AF,$F0,$62,$F4,$D1,$FE,$1F,$CF,$F7
	.dc.b	$F9,$78,$6B,$F5,$EF,$85,$FD,$EE,$1E,$1F,$84,$91,$83,$B6,$3E,$05
	.dc.b	$6E,$2F,$F7,$26,$3E,$2C,$BC,$3F,$07,$D6,$FB,$9F,$00,$FE,$17,$AF
	.dc.b	$C0,$7D,$59,$78,$1F,$E0,$79,$F7,$CB,$B4,$7B,$71,$EC,$4C,$3E,$47
	.dc.b	$C2,$49,$F8,$8F,$CB,$F8,$5F,$99,$F0,$9F,$37,$E1,$65,$7C,$98,$FA
	.dc.b	$1F,$09,$23,$E6,$3F,$4F,$E1,$7A,$47,$C0,$E0,$F7,$E3,$8D,$3C,$BC
	.dc.b	$1F,$43,$3E,$F7,$F5,$73,$EE,$F1,$3D,$59,$F7,$71,$5F,$63,$3C,$61
	.dc.b	$D3,$56,$9E,$BB,$F6,$90,$E3,$0C,$BC,$31,$8C,$BC,$37,$CF,$2E,$BA
	.dc.b	$7D,$8C,$AD,$FC,$DC,$BF,$25,$F0,$B2,$3F,$1C,$FE,$57,$E1,$7F,$2D
	.dc.b	$F0,$9F,$97,$F8,$59,$5F,$93,$8F,$CD,$7C,$1C,$8F,$CB,$3F,$9D,$F8
	.dc.b	$4E,$51,$D0,$87,$AA,$FA,$BF,$9D,$BE,$AF,$BA,$FA,$3F,$02,$EE,$BF
	.dc.b	$1F,$0E,$B3,$F1,$31,$F5,$3D,$4F,$31,$D4,$87,$53,$CE,$43,$B1,$FA
	.dc.b	$F8,$FD,$B7,$C1,$C8,$FD,$93,$FB,$9F,$84,$FD,$D7,$C2,$C9,$FD,$B3
	.dc.b	$FB,$DF,$83,$FD,$F7,$C2,$7E,$FF,$E0,$E5,$7E,$EE,$39,$3D,$08,$7D
	.dc.b	$EB,$E8,$FE,$02,$EE,$7F,$87,$C3,$AC,$FF,$05,$C3,$AB,$F8,$6C,$3A
	.dc.b	$BF,$88,$C7,$D4,$F5,$3C,$E4,$3B,$1F,$C3,$EB,$AB,$F9,$0A,$EB,$7F
	.dc.b	$97,$BF,$F5,$17,$EF,$1F,$EB,$9A,$FC,$EB,$67,$F4,$ED,$E3,$4F,$F1
	.dc.b	$98,$FF,$2B,$9F,$E4,$C9,$DF,$C2,$74,$3A,$FC,$47,$6F,$83,$EF,$F0
	.dc.b	$93,$0E,$77,$DA,$7C,$7F,$0B,$E4,$98,$6B,$2F,$0F,$6B,$3F,$0E,$7D
	.dc.b	$EB,$47,$95,$F0,$B8,$9E,$18,$78,$63,$DF,$97,$95,$D5,$FB,$1F,$09
	.dc.b	$F1,$31,$EC,$FC,$6C,$2E,$5F,$1B,$3E,$53,$D0,$E7,$86,$27,$E5,$FC
	.dc.b	$27,$D0,$24,$F8,$3E,$17,$E9,$FC,$2C,$8F,$1B,$F5,$BE,$17,$EB,$FC
	.dc.b	$24,$9F,$3B,$F6,$BE,$17,$9C,$7D,$AC,$2E,$3F,$6B,$1E,$93,$F7,$B1
	.dc.b	$B3,$F7,$B1,$E9,$3F,$82,$C6,$CF,$E0,$B1,$FC,$3E,$1D,$38,$7E,$9E
	.dc.b	$EC,$F3,$8F,$A0,$87,$73,$E6,$E5,$D2,$B8,$E5,$67,$F1,$98,$FE,$53
	.dc.b	$E1,$3F,$2B,$F0,$B2,$7E,$AC,$7E,$63,$E1,$7F,$33,$F0,$B2,$7E,$CC
	.dc.b	$7E,$73,$E1,$24,$7D,$C7,$F3,$FF,$0B,$D6,$3E,$FE,$1D,$5F,$D0,$E1
	.dc.b	$D5,$FC,$1E,$1D,$5F,$D2,$E1,$D1,$FD,$3E,$17,$83,$F4,$F8,$F5,$9F
	.dc.b	$D5,$E3,$D6,$7F,$57,$8F,$59,$FD,$7E,$3D,$67,$F5,$F8,$F3,$D3,$E6
	.dc.b	$E5,$6F,$F6,$19,$7E,$E3,$E1,$24,$7E,$61,$FD,$DF,$C2,$FE,$F3,$D3
	.dc.b	$FD,$EF,$C2,$CA,$FC,$EC,$7E,$E9,$FE,$B7,$FB,$CF,$F7,$86,$E5,$9D
	.dc.b	$3F,$B3,$CE,$5B,$F0,$BE,$0A,$FE,$0F,$95,$9F,$E0,$F9,$78,$67,$F5
	.dc.b	$79,$78,$5F,$E1,$B9,$75,$CB,$F7,$79,$75,$69,$99,$FF,$13,$46,$5B
	.dc.b	$97,$B9,$F8,$04,$3F,$8C,$E1,$E1,$7F,$8D,$E1,$E1,$C7,$F8,$DE,$3D
	.dc.b	$7B,$58,$F5,$6E,$59,$8C,$1F,$EB,$1C,$3C,$7F,$09,$9F,$85,$91,$FC
	.dc.b	$36,$53,$C7,$E1,$79,$7C,$34,$AD,$47,$5F,$86,$91,$C5,$F0,$7C,$37
	.dc.b	$86,$FA,$CF,$D6,$C7,$0F,$23,$C1,$87,$57,$ED,$E1,$D1,$FB,$B8,$5D
	.dc.b	$7E,$DE,$3C,$A7,$F0,$38,$D9,$FB,$FF,$0B,$E9,$9E,$A4,$3A,$9E,$72
	.dc.b	$1D,$8F,$85,$5F,$27,$E1,$A4,$7C,$57,$E6,$7C,$37,$CD,$F8,$6F,$9D
	.dc.b	$F0,$D2,$BE,$54,$7D,$1F,$86,$FA,$5F,$0B,$27,$E7,$47,$47,$F3,$38
	.dc.b	$5C,$FF,$2F,$8F,$59,$FA,$98,$F5,$7F,$3F,$8F,$57,$EC,$63,$D5,$FD
	.dc.b	$1E,$3C,$F8,$BF,$2B,$95,$9F,$C9,$E3,$D5,$FD,$4E,$3D,$72,$FD,$3E
	.dc.b	$5D,$67,$F5,$B9,$75,$9F,$D5,$E5,$F8,$DF,$86,$FC,$77,$C2,$CB,$FC
	.dc.b	$54,$7E,$47,$E1,$BF,$25,$F0,$D2,$7F,$1D,$1F,$95,$F8,$59,$1F,$91
	.dc.b	$7F,$31,$F0,$9D,$63,$F2,$98,$75,$7F,$7F,$87,$57,$F3,$18,$75,$7F
	.dc.b	$81,$E1,$D1,$FE,$0B,$85,$CB,$F8,$1E,$3D,$67,$99,$E8,$3F,$B7,$7A
	.dc.b	$4F,$85,$ED,$1F,$C3,$B1,$ED,$7F,$C3,$31,$E9,$C5,$FC,$0B,$2B,$7F
	.dc.b	$C0,$32,$FD,$B6,$5F,$E9,$72,$7C,$36,$3E,$1B,$3F,$0D,$AF,$85,$98
	.dc.b	$78,$A3,$8F,$C2,$72,$F8,$59,$3E,$68,$E4,$FA,$30,$EE,$61,$FF,$3F
	.dc.b	$E2,$BE,$B1,$8C,$3C,$2F,$33,$E0,$E1,$D9,$E3,$87,$3D,$BF,$D2,$38
	.dc.b	$7A,$A7,$E3,$E5,$C9,$E8,$7C,$AC,$31,$5C,$2F,$A3,$CB,$0F,$6D,$7D
	.dc.b	$0F,$85,$F8,$FF,$0B,$23,$07,$D4,$F8,$1F,$97,$F0,$B2,$30,$7D,$8F
	.dc.b	$84,$F9,$E4,$7E,$DF,$C2,$FD,$2C,$24,$74,$3E,$2F,$C2,$FD,$6C,$3A
	.dc.b	$47,$C8,$C3,$0F,$52,$1F,$33,$0F,$BB,$87,$48,$F9,$F8,$5C,$FE,$3E
	.dc.b	$3C,$A7,$A1,$F8,$EC,$31,$3F,$33,$0F,$44,$F5,$3D,$27,$E5,$3E,$17
	.dc.b	$F1,$7F,$07,$23,$07,$E6,$3E,$17,$F1,$FF,$0B,$23,$07,$E7,$3E,$17
	.dc.b	$F2,$7F,$0B,$23,$07,$23,$F0,$F8,$7E,$60,$FC,$CF,$2F,$E6,$B0,$E9
	.dc.b	$3F,$8C,$C3,$0F,$23,$F1,$F7,$D5,$FC,$8E,$1C,$DE,$D9,$FE,$2B,$1E
	.dc.b	$B1,$E6,$3A,$1F,$B2,$BE,$AF,$23,$F6,$B7,$D6,$BF,$27,$7F,$D8,$FF
	.dc.b	$7A,$FF,$3F,$53,$CF,$43,$E1,$74,$6E,$78,$7C,$2C,$7F,$CD,$D3,$FE
	.dc.b	$83,$EE,$FF,$3B,$AF,$85,$ED,$F0,$DD,$FE,$1A,$5F,$08,$E8,$F2,$29
	.dc.b	$FE,$F0,$AF,$F2,$73,$DB,$0F,$03,$97,$D5,$7D,$A3,$C7,$7D,$9E,$57
	.dc.b	$D3,$46,$3E,$69,$C3,$E8,$F8,$3E,$B3,$A9,$F8,$99,$7B,$5B,$97,$81
	.dc.b	$F5,$1E,$02,$1E,$03,$E2,$FC,$2F,$CF,$23,$F2,$3E,$1B,$E9,$7C,$2F
	.dc.b	$D3,$F8,$59,$3F,$32,$3E,$B7,$C2,$C8,$F9,$EF,$D9,$F8,$4F,$B5,$F0
	.dc.b	$BF,$6F,$E1,$25,$7D,$48,$FB,$DF,$0B,$F7,$FE,$16,$4F,$D8,$8F,$C1
	.dc.b	$7C,$2C,$8F,$B6,$FE,$17,$E1,$7F,$0D,$F0,$92,$7E,$FB,$F8,$9F,$83
	.dc.b	$FC,$57,$C2,$7F,$8B,$7F,$7B,$FF,$16,$1F,$E2,$DE,$44,$67,$E1,$B5
	.dc.b	$F0,$33,$1F,$F5,$B5,$FF,$8B,$FB,$8B,$E1,$7A,$7C,$24,$AD,$C7,$7F
	.dc.b	$85,$F0,$7C,$2C,$9E,$71,$E3,$F8,$49,$1D,$9F,$37,$C2,$F9,$FE,$17
	.dc.b	$D1,$F0,$B2,$BC,$71,$EB,$F8,$49,$1E,$67,$DD,$F0,$BE,$FF,$84,$93
	.dc.b	$EA,$7E,$17,$C2,$FC,$3F,$84,$93,$EE,$7E,$37,$C2,$FC,$7F,$83,$F9
	.dc.b	$1F,$09,$2B,$E1,$C7,$CB,$F8,$49,$1F,$19,$F9,$DF,$09,$FE,$5B,$FD
	.dc.b	$F3,$FB,$84,$FF,$B7,$DA,$BC,$BF,$CD,$74,$27,$31,$85,$F0,$BC,$FE
	.dc.b	$12,$5F,$37,$B7,$C2,$F7,$F8,$5F,$07,$C2,$CA,$F0,$47,$8F,$E1,$A4
	.dc.b	$78,$DF,$36,$1D,$23,$C9,$E9,$FF,$BF,$32,$65,$9C,$FC,$99,$F7,$BF
	.dc.b	$66,$7D,$DE,$27,$F1,$7B,$67,$E0,$8C,$E9,$EF,$D3,$18,$FB,$F4,$ED
	.dc.b	$A7,$87,$4B,$87,$3D,$3C,$0F,$2C,$DA,$78,$71,$F9,$1A,$7D,$1F,$86
	.dc.b	$FA,$5F,$07,$30,$FA,$51,$F5,$7E,$1A,$47,$D5,$7E,$C7,$C3,$7D,$9F
	.dc.b	$84,$93,$F6,$5F,$B9,$F0,$9D,$63,$EE,$61,$D5,$F2,$61,$D1,$F3,$61
	.dc.b	$72,$FB,$98,$F4,$9F,$56,$36,$7F,$01,$8F,$59,$E6,$7A,$0F,$C5,$7C
	.dc.b	$37,$3E,$0F,$AD,$8D,$9F,$A1,$8F,$39,$FB,$58,$D9,$FA,$98,$7E,$57
	.dc.b	$E1,$BF,$2D,$F0,$72,$7F,$2D,$1F,$99,$F8,$69,$1F,$99,$7F,$39,$F0
	.dc.b	$BF,$9D,$F8,$5F,$CF,$7C,$24,$AF,$CF,$47,$57,$F4,$18,$75,$7E,$C6
	.dc.b	$1D,$1F,$B5,$85,$DF,$F4,$18,$F5,$9F,$D4,$63,$D5,$FC,$06,$3D,$5F
	.dc.b	$C1,$63,$D6,$FF,$51,$8F,$3D,$3F,$35,$95,$BF,$CA,$65,$D5,$FC,$56
	.dc.b	$5D,$70,$FD,$86,$5F,$B9,$F8,$59,$5F,$B9,$7F,$79,$F0,$DF,$ED,$7E
	.dc.b	$CF,$EE,$BD,$86,$93,$E1,$B5,$F0,$53,$01,$FA,$29,$FE,$B1,$5C,$7E
	.dc.b	$1B,$BC,$FE,$77,$1B,$9F,$F0,$3C,$BC,$13,$0E,$71,$FC,$2B,$0F,$04
	.dc.b	$63,$7E,$F8,$F9,$72,$F1,$4F,$85,$F5,$65,$DA,$BF,$80,$65,$6B,$F7
	.dc.b	$59,$77,$7D,$05,$CB,$C1,$9F,$47,$C5,$8D,$C3,$C3,$9F,$F8,$3E,$FC
	.dc.b	$87,$F8,$6D,$7C,$2F,$CB,$F8,$69,$3A,$D7,$E6,$4F,$CF,$F8,$69,$39
	.dc.b	$BF,$F0,$8B,$F2,$FE,$1B,$EA,$7C,$34,$8D,$3F,$5F,$E1,$7E,$C6,$12
	.dc.b	$79,$3F,$6F,$E1,$BA,$C7,$D5,$C6,$ED,$F3,$71,$E9,$3F,$67,$1B,$3F
	.dc.b	$47,$1E,$93,$F7,$71,$B3,$F5,$71,$E7,$AF,$C4,$CA,$DF,$C3,$CB,$9C
	.dc.b	$FC,$8C,$AC,$FC,$7C,$BF,$21,$F0,$DF,$91,$F8,$69,$3F,$0E,$3F,$29
	.dc.b	$F0,$B2,$3E,$33,$F9,$7F,$86,$FC,$C7,$C2,$C9,$F9,$4F,$E6,$FE,$1B
	.dc.b	$AC,$7C,$DC,$3A,$BF,$97,$C3,$A3,$F9,$9C,$2E,$FF,$91,$C7,$A4,$FE
	.dc.b	$77,$1B,$3F,$95,$C7,$AC,$FE,$87,$1E,$B3,$F9,$9C,$79,$E9,$F8,$8C
	.dc.b	$AD,$FE,$1F,$2E,$73,$F8,$CC,$AC,$FE,$2F,$2F,$DA,$7C,$34,$8F,$C3
	.dc.b	$3F,$B7,$F8,$6F,$DC,$7C,$34,$9F,$C5,$3F,$BB,$F8,$6F,$DE,$7C,$2F
	.dc.b	$ED,$A7,$FA,$BF,$EF,$BF,$E4,$F6,$E7,$A7,$A9,$0F,$DE,$63,$DD,$EA
	.dc.b	$43,$F8,$06,$1D,$DF,$CD,$5F,$7D,$FF,$C0,$78,$77,$6E,$1D,$EB,$A9
	.dc.b	$DE,$3F,$80,$61,$DE,$3A,$9D,$E3,$F8,$26,$1D,$E3,$F8,$86,$1D,$EF
	.dc.b	$F8,$2E,$1D,$DA,$77,$BF,$1F,$C3,$6E,$FF,$D2,$54,$F8,$6C,$75,$23
	.dc.b	$13,$FB,$97,$FC,$90,$7F,$91,$DC,$7C,$27,$08,$C4,$7C,$EF,$85,$C7
	.dc.b	$C2,$67,$E1,$64,$F3,$8E,$19,$71,$CA,$4F,$78,$E9,$8C,$8E,$E7,$8F
	.dc.b	$0E,$AF,$97,$0E,$AF,$DE,$C3,$AB,$CC,$87,$E0,$BE,$17,$A3,$F8,$4B
	.dc.b	$BA,$7D,$8C,$3A,$CF,$E2,$30,$EB,$3F,$73,$0F,$81,$F0,$B2,$BE,$33
	.dc.b	$F0,$FE,$13,$E2,$7C,$2C,$9F,$94,$FC,$7F,$85,$F9,$1F,$0B,$F2,$7E
	.dc.b	$16,$57,$CF,$8E,$8F,$E6,$30,$B9,$7E,$33,$1E,$93,$F9,$CC,$6C,$FE
	.dc.b	$43,$1E,$B3,$F6,$71,$EA,$FE,$8B,$1F,$33,$E8,$21,$FA,$6C,$39,$F1
	.dc.b	$FE,$33,$1B,$3F,$88,$C3,$AB,$FA,$CC,$3A,$E3,$FA,$0C,$7F,$09,$F0
	.dc.b	$B2,$BF,$14,$FE,$1F,$E1,$7F,$11,$F0,$BF,$89,$F8,$49,$5F,$8F,$8F
	.dc.b	$C6,$7C,$2F,$E3,$7E,$16,$4F,$E4,$E3,$A3,$FB,$CC,$2E,$5F,$B0,$C7
	.dc.b	$AC,$FE,$67,$1E,$AF,$F0,$2C,$7A,$BF,$9D,$C7,$AB,$FC,$1B,$1F,$33
	.dc.b	$E8,$21,$FC,$2A,$FA,$BF,$C3,$2F,$AE,$1F,$C0,$30,$EB,$3F,$C4,$30
	.dc.b	$EB,$3F,$C1,$30,$FD,$47,$97,$FB,$0F,$DF,$BF,$D0,$FC,$1F,$DF,$9B
	.dc.b	$87,$ED,$A7,$5F,$0B,$BF,$85,$93,$FD,$B6,$BF,$D1,$AD,$F8,$5E,$9F
	.dc.b	$07,$23,$07,$EE,$F0,$F0,$7C,$34,$8C,$1C,$F1,$EA,$FF,$9F,$B0,$FF
	.dc.b	$44,$B9,$39,$99,$BE,$CE,$F0,$F6,$BE,$5B,$EC,$FF,$32,$C3,$D6,$FA
	.dc.b	$71,$EF,$A7,$FA,$AB,$1F,$5C,$F3,$3D,$D8,$F6,$9D,$63,$F2,$7E,$16
	.dc.b	$57,$C3,$7E,$67,$C2,$7C,$DF,$85,$F9,$DF,$09,$2B,$E4,$47,$D1,$F8
	.dc.b	$49,$1F,$2D,$FA,$9F,$09,$D2,$3E,$16,$17,$2F,$36,$3D,$67,$E9,$63
	.dc.b	$D5,$F9,$18,$F4,$7E,$56,$36,$FD,$D8,$F3,$E1,$FC,$06,$56,$DE,$33
	.dc.b	$66,$9D,$E7,$EA,$69,$DF,$2F,$91,$A7,$5C,$BF,$15,$A5,$BE,$78,$B4
	.dc.b	$FC,$87,$C2,$49,$FC,$53,$F9,$3F,$85,$FC,$A7,$C2,$7E,$57,$E1,$65
	.dc.b	$7E,$3E,$3F,$31,$F0,$B2,$3F,$24,$FE,$6F,$E1,$7A,$47,$E2,$70,$BB
	.dc.b	$7D,$DC,$7A,$CF,$E6,$71,$EA,$FE,$3F,$1E,$8F,$E4,$71,$B7,$F8,$5C
	.dc.b	$7C,$D3,$E8,$21,$F9,$7C,$3A,$BF,$99,$C3,$AD,$7E,$37,$0E,$7A,$FE
	.dc.b	$BF,$1B,$7F,$1B,$1F,$DA,$7C,$27,$ED,$7E,$16,$4F,$EB,$E3,$F7,$1F
	.dc.b	$09,$FB,$9F,$84,$93,$FB,$38,$FE,$D7,$FB,$FF,$FA,$30,$91,$8F,$85
	.dc.b	$EA,$FE,$E7,$0E,$AF,$EB,$F0,$EA,$FE,$F7,$0E,$AF,$EC,$F0,$EA,$C3
	.dc.b	$FC,$DD,$87,$7C,$3A,$C7,$EE,$70,$EB,$AF,$EB,$71,$EB,$3F,$BD,$C7
	.dc.b	$AC,$FE,$C7,$1E,$93,$FE,$FE,$C7,$D2,$F6,$C7,$DB,$F0,$BE,$EF,$86
	.dc.b	$93,$E2,$8F,$83,$F0,$BF,$0B,$E1,$A4,$F9,$A3,$E2,$FC,$2C,$8F,$4B
	.dc.b	$F2,$3E,$17,$AC,$73,$21,$E3,$F8,$5E,$AF,$C0,$BE,$AF,$9E,$FA,$3E
	.dc.b	$9B,$BA,$B8,$FF,$D5,$7A,$B2,$EF,$7E,$FC,$ED,$EE,$3F,$67,$4C,$FC
	.dc.b	$31,$F1,$73,$F0,$D6,$FB,$97,$FE,$A4,$AF,$69,$0F,$21,$F6,$31,$FC
	.dc.b	$3F,$C2,$C8,$FB,$6F,$E2,$BE,$1B,$F1,$7F,$0B,$F8,$CF,$86,$95,$F8
	.dc.b	$08,$FC,$7F,$C3,$48,$FC,$1B,$F9,$2F,$86,$E9,$1F,$6F,$0B,$A7,$9F
	.dc.b	$36,$5D,$AB,$F1,$79,$76,$7F,$05,$97,$57,$F0,$99,$5C,$39,$65,$E7
	.dc.b	$9F,$49,$0F,$C5,$61,$D5,$FC,$66,$1D,$6B,$C5,$87,$3E,$0F,$AB,$8D
	.dc.b	$BF,$A3,$8F,$EB,$3E,$1B,$F5,$BF,$0D,$27,$F4,$31,$FB,$0F,$84,$FD
	.dc.b	$8F,$C3,$49,$FD,$2C,$7E,$D3,$E1,$64,$7E,$A1,$FD,$BF,$C3,$75,$8F
	.dc.b	$D5,$E1,$D5,$FD,$16,$1D,$5F,$D7,$E1,$D5,$FD,$36,$1D,$1F,$D4,$61
	.dc.b	$75,$F9,$D8,$F5,$9F,$D6,$63,$D6,$7E,$96,$3D,$67,$F6,$18,$F5,$9F
	.dc.b	$AD,$8F,$3D,$3F,$33,$95,$BF,$CA,$E5,$FC,$43,$E1,$BF,$CF,$67,$FA
	.dc.b	$0E,$3F,$A5,$B7,$93,$F2,$AE,$BE,$13,$7F,$0B,$2B,$FD,$F1,$29,$E7
	.dc.b	$F0,$9D,$3E,$12,$56,$A3,$C0,$FF,$12,$AA,$B5,$FD,$24,$3E,$0F,$C1
	.dc.b	$3D,$B1,$F0,$3B,$D4,$FA,$32,$F0,$BE,$4C,$BC,$2E,$CF,$6E,$5E,$0C
	.dc.b	$BF,$39,$95,$C3,$9E,$68,$F1,$5F,$BB,$1F,$0D,$F9,$B2,$F2,$CF,$5B
	.dc.b	$E1,$8F,$82,$BD,$79,$79,$2B,$58,$BE,$13,$E8,$FC,$2C,$9F,$97,$1F
	.dc.b	$53,$E1,$64,$7C,$E7,$EB,$FC,$2F,$D8,$F8,$59,$3F,$49,$FB,$7F,$0B
	.dc.b	$F7,$3E,$12,$4F,$D6,$7E,$FF,$C2,$FE,$03,$E1,$7F,$03,$F0,$B2,$BE
	.dc.b	$DC,$7E,$13,$E1,$24,$7D,$E7,$F0,$FF,$0B,$F8,$8F,$84,$FC,$4F,$C2
	.dc.b	$CA,$FC,$1C,$7E,$33,$E0,$FF,$D8,$FF,$C0,$1F,$E5,$6E,$73,$79,$13
	.dc.b	$AF,$86,$98,$7D,$CE,$7F,$E4,$D7,$FE,$3E,$6F,$C2,$F8,$67,$F2,$39
	.dc.b	$F8,$5F,$9B,$9F,$85,$E5,$D9,$F7,$30,$F1,$1E,$1C,$BB,$CF,$D9,$8E
	.dc.b	$D3,$BC,$BD,$38,$FA,$BE,$13,$D7,$F0,$B3,$0E,$A7,$F8,$CB,$04,$FC
	.dc.b	$0F,$85,$93,$AD,$FE,$0C,$7C,$4F,$85,$F8,$BF,$09,$2B,$58,$23,$E4
	.dc.b	$FC,$2F,$CA,$F8,$5F,$97,$F0,$92,$F7,$1F,$3B,$E1,$3E,$7F,$C2,$49
	.dc.b	$E7,$1F,$4B,$E0,$E4,$76,$7E,$AF,$C2,$7D,$6F,$84,$93,$E2,$7E,$CF
	.dc.b	$C2,$7D,$AF,$83,$93,$E6,$7E,$EF,$C2,$7D,$EF,$83,$FB,$FF,$09,$2B
	.dc.b	$D7,$1F,$B6,$7F,$4D,$FC,$03,$5A,$67,$BB,$FB,$24,$65,$BF,$0B,$9E
	.dc.b	$CF,$DA,$CB,$E0,$B1,$7F,$E6,$28,$7C,$2E,$3A,$91,$89,$FA,$4F,$16
	.dc.b	$FC,$2E,$E1,$F0,$74,$C4,$67,$E1,$75,$F0,$92,$F9,$C7,$1F,$85,$91
	.dc.b	$D9,$E9,$F0,$BD,$7E,$13,$B7,$C1,$CA,$F1,$C7,$57,$CB,$87,$57,$E3
	.dc.b	$E1,$D1,$F9,$38,$5E,$1F,$8F,$8F,$59,$F6,$E3,$D5,$F9,$F8,$F5,$79
	.dc.b	$9E,$83,$E1,$FC,$27,$3E,$1F,$9B,$8D,$9F,$99,$87,$57,$EC,$E1,$D7
	.dc.b	$1F,$AF,$8F,$C9,$F8,$4F,$95,$F0,$92,$FE,$7C,$73,$E1,$FA,$79,$5E
	.dc.b	$16,$7F,$47,$E0,$FE,$97,$C2,$4C,$7E,$BC,$75,$7E,$CE,$5D,$5F,$C5
	.dc.b	$E5,$CF,$1E,$98,$73,$9F,$BF,$87,$37,$F2,$38,$73,$78,$9E,$63,$F0
	.dc.b	$BF,$07,$CD,$FB,$47,$E2,$2B,$A3,$F9,$AA,$E9,$AF,$E6,$2F,$F1,$5F
	.dc.b	$09,$F8,$BF,$84,$97,$F9,$08,$E8,$FD,$A3,$A1,$F9,$4F,$84,$FC,$97
	.dc.b	$A7,$23,$07,$13,$F3,$35,$D1,$FD,$5D,$74,$C3,$F5,$3A,$7E,$42,$78
	.dc.b	$9F,$B0,$BE,$8F,$EC,$B0,$F3,$BD,$4E,$67,$F1,$7F,$C6,$3F,$C0,$5C
	.dc.b	$1F,$E5,$B8,$F8,$67,$A4,$FD,$DE,$BF,$A9,$AE,$B8,$FE,$EB,$0C,$57
	.dc.b	$3F,$84,$FD,$A6,$2F,$84,$A6,$23,$D5,$CD,$F9,$2A,$F8,$A7,$F6,$DA
	.dc.b	$7C,$27,$2F,$84,$93,$E7,$8E,$4F,$A7,$0E,$AF,$84,$F4,$E5,$F1,$9B
	.dc.b	$87,$F0,$CD,$3C,$B7,$D2,$7C,$18,$61,$E4,$78,$F2,$EE,$F9,$7E,$17
	.dc.b	$D2,$FF,$13,$DF,$E9,$3E,$9B,$E9,$97,$FA,$07,$46,$3D,$6B,$DD,$F0
	.dc.b	$9F,$1B,$E1,$64,$7D,$38,$EB,$D3,$FA,$A8,$F8,$B8,$73,$F8,$5F,$9B
	.dc.b	$F0,$B2,$7E,$D4,$7D,$0B,$E7,$1F,$2F,$0C,$3F,$08,$F4,$17,$0F,$E5
	.dc.b	$3B,$FD,$8B,$E9,$3F,$4F,$0C,$3E,$52,$1E,$A3,$F1,$5F,$0B,$E8,$E5
	.dc.b	$FC,$54,$7E,$20,$F2,$9F,$8F,$9E,$B9,$FA,$E3,$F2,$5F,$0B,$F8,$8F
	.dc.b	$85,$91,$83,$CD,$7D,$2B,$EE,$47,$53,$F1,$1F,$0B,$F9,$0F,$84,$91
	.dc.b	$F9,$D8,$FC,$9E,$1D,$23,$F1,$D8,$61,$EB,$C7,$FC,$31,$E0,$7E,$96
	.dc.b	$F9,$BD,$8F,$21,$F9,$7A,$E4,$FE,$67,$E1,$7E,$15,$7E,$13,$1F,$C1
	.dc.b	$63,$C9,$FC,$26,$1F,$A6,$8F,$D3,$FC,$2F,$EA,$3E,$16,$63,$FB,$58
	.dc.b	$E5,$D3,$FA,$78,$FD,$A6,$1F,$D1,$1F,$DD,$7C,$27,$EC,$3E,$16,$46
	.dc.b	$0E,$47,$EF,$F0,$EA,$FE,$BF,$0E,$AF,$E1,$0E,$B8,$7E,$9D,$E4,$7E
	.dc.b	$D6,$BA,$BF,$B7,$BF,$4B,$D8,$E8,$7F,$0D,$F8,$5E,$BC,$BF,$9B,$8E
	.dc.b	$A7,$F5,$03,$AE,$5F,$A7,$BE,$AF,$EA,$6F,$FC,$03,$87,$FA,$0B,$E2
	.dc.b	$2D,$FF,$8E,$8F,$F7,$A7,$C3,$62,$3B,$F1,$FE,$1E,$BE,$46,$1F,$1E
	.dc.b	$FF,$D0,$D1,$FE,$8C,$F8,$6E,$7F,$0B,$27,$FC,$7F,$39,$68,$C3,$C3
	.dc.b	$8C,$8C,$C7,$83,$8B,$F1,$51,$D8,$61,$D5,$C6,$1E,$A8,$F4,$47,$22
	.dc.b	$1C,$7E,$1B,$A3,$FA,$2E,$AE,$0F,$92,$BA,$66,$CF,$FD,$4E,$7A,$67
	.dc.b	$C5,$F0,$DF,$23,$E1,$A4,$7B,$23,$AB,$F5,$38,$7F,$81,$BF,$03,$E1
	.dc.b	$7E,$7F,$C2,$C8,$C1,$C4,$F8,$B7,$D1,$EE,$7F,$18,$C3,$E8,$91,$FC
	.dc.b	$01,$F1,$6B,$A3,$F1,$EB,$EE,$C7,$DE,$C2,$4F,$73,$E6,$7C,$27,$A3
	.dc.b	$2F,$E0,$67,$DD,$3E,$14,$4A,$EB,$18,$C3,$FD,$97,$87,$6A,$FB,$BF
	.dc.b	$0B,$F8,$CF,$84,$91,$83,$F0,$3F,$0B,$F9,$0F,$85,$91,$83,$F0,$BF
	.dc.b	$0B,$F9,$4F,$85,$91,$83,$91,$F8,$3B,$EA,$FE,$17,$0F,$CD,$E1,$D2
	.dc.b	$3F,$11,$86,$1E,$47,$E2,$EF,$AB,$F8,$DC,$3F,$47,$7D,$33,$F7,$4B
	.dc.b	$1C,$BE,$73,$A9,$F9,$CC,$3B,$3C,$CF,$CF,$E1,$DA,$B9,$E1,$FB,$0F
	.dc.b	$85,$FD,$8F,$C2,$CB,$FD,$2C,$72,$AF,$0E,$1E,$B7,$C7,$9F,$EE,$3E
	.dc.b	$16,$5F,$EB,$1F,$DD,$FC,$27,$58,$FD,$7E,$3D,$5F,$D5,$63,$D7,$7F
	.dc.b	$CE,$C3,$2E,$D1,$8C,$7A,$CF,$EC,$71,$B9,$7C,$4C,$BB,$4F,$ED,$72
	.dc.b	$ED,$3F,$23,$2E,$D3,$F8,$AC,$BA,$BF,$8C,$C7,$9E,$5E,$2C,$AC,$FE
	.dc.b	$43,$2F,$E2,$BF,$0B,$FC,$63,$EC,$7F,$D3,$D7,$FB,$46,$D7,$E5,$A7
	.dc.b	$BC,$A3,$D8,$7D,$53,$EC,$67,$F9,$9A,$7C,$37,$52,$4F,$F1,$88,$FF
	.dc.b	$50,$3F,$C4,$6F,$C4,$C3,$F8,$DB,$E0,$7B,$9F,$84,$3F,$8C,$65,$D1
	.dc.b	$FE,$4F,$3E,$07,$B5,$7B,$7E,$1B,$DF,$1E,$13,$B1,$E1,$3D,$38,$FD
	.dc.b	$8D,$3E,$5B,$EF,$C3,$B6,$5E,$DC,$2C,$FC,$7C,$3E,$57,$C3,$7C,$BF
	.dc.b	$85,$99,$7B,$E3,$D5,$97,$48,$65,$F4,$7E,$1A,$47,$C5,$7E,$A7,$C2
	.dc.b	$F4,$8F,$91,$87,$47,$DF,$87,$47,$E9,$1F,$10,$E4,$43,$E2,$D7,$47
	.dc.b	$C8,$74,$21,$D0,$F8,$67,$E3,$4E,$BA,$7E,$99,$E8,$FD,$F8,$F4,$C7
	.dc.b	$D2,$BF,$C6,$7C,$37,$E3,$7C,$29,$3F,$81,$8F,$43,$F6,$EF,$A3,$F6
	.dc.b	$BC,$0F,$CA,$11,$FC,$47,$C2,$74,$7F,$15,$7D,$1F,$C0,$DF,$4B,$FD
	.dc.b	$EC,$7A,$B7,$70,$FF,$AF,$A3,$F2,$58,$F5,$7F,$13,$8F,$57,$F1,$78
	.dc.b	$F5,$C3,$F3,$98,$F5,$C7,$F0,$D5,$D7,$1E,$8F,$A4,$EA,$7A,$0F,$D0
	.dc.b	$7C,$2F,$EB,$FE,$1A,$46,$0F,$31,$F9,$EC,$3D,$2F,$DB,$3F,$3B,$F0
	.dc.b	$BF,$B7,$23,$FA,$BF,$86,$FD,$D5,$F4,$8F,$D1,$DE,$1E,$BC,$7F,$91
	.dc.b	$E3,$E3,$1D,$0F,$DA,$61,$FC,$0F,$0E,$11,$C8,$FD,$C5,$E3,$1F,$AE
	.dc.b	$79,$89,$9F,$F1,$D9,$E6,$43,$B1,$FB,$6A,$FE,$21,$F0,$9F,$C4,$7E
	.dc.b	$16,$4F,$F0,$58,$F9,$CF,$E4,$4F,$85,$7F,$C3,$7E,$1B,$F8,$C1,$23
	.dc.b	$F8,$67,$C2,$F4,$7F,$87,$5F,$47,$F8,$2D,$F4,$C7,$F8,$D6,$BF,$C7
	.dc.b	$37,$63,$CE,$FF,$85,$BF,$C9,$71,$E9,$1F,$C3,$B1,$E9,$97,$F1,$FC
	.dc.b	$BA,$4F,$D9,$CB,$E1,$3C,$8F,$E4,$18,$75,$AF,$E4,$D8,$7F,$31,$F8
	.dc.b	$6F,$E6,$5F,$0D,$2F,$F9,$34,$74,$D3,$CD,$F0,$DF,$CD,$C9,$3F,$E4
	.dc.b	$2C,$3F,$C1,$6F,$F2,$2A,$E4,$FF,$24,$BE,$CF,$E0,$8F,$DD,$E5,$E6
	.dc.b	$AF,$E7,$17,$D2,$7F,$95,$BF,$E5,$C8,$EA,$7F,$3D,$BE,$AF,$F3,$FB
	.dc.b	$EB,$5F,$5A,$FF,$4A,$FE,$F5,$EC,$7F,$4D,$BE,$D5,$E0,$9F,$54,$F7
	.dc.b	$F8,$5F,$07,$C3,$49,$F7,$47,$F4,$3E,$3F,$EA,$77,$FE,$F2,$9F,$85
	.dc.b	$F0,$7E,$7F,$86,$91,$83,$89,$F1,$F0,$E8,$FF,$9B,$A9,$87,$58,$FE
	.dc.b	$11,$1D,$5D,$5F,$47,$85,$FC,$58,$F4,$47,$52,$1D,$6B,$A3,$F1,$4F
	.dc.b	$F3,$15,$75,$3F,$C8,$07,$FB,$C9,$FB,$95,$DE,$33,$5F,$4F,$E1,$25
	.dc.b	$7E,$05,$FA,$DF,$09,$F5,$FE,$13,$EC,$7C,$1C,$AF,$C3,$47,$DB,$F8
	.dc.b	$4F,$B9,$F0,$72,$7F,$15,$1F,$7F,$E1,$3F,$01,$F0,$52,$7F,$1D,$1F
	.dc.b	$83,$F8,$39,$1F,$91,$7F,$0D,$F0,$9F,$87,$F8,$39,$3F,$95,$7F,$15
	.dc.b	$F0,$9F,$8B,$F8,$2F,$C6,$7C,$1C,$AF,$CD,$47,$E3,$FE,$12,$47,$E7
	.dc.b	$5F,$F5,$6F,$E9,$BF,$E9,$D8,$FF,$50,$BF,$EA,$DE,$93,$FE,$9A,$8C
	.dc.b	$FC,$17,$0F,$84,$E3,$F0,$52,$78,$BD,$3E,$0F,$AF,$C0,$F6,$F8,$29
	.dc.b	$5D,$A3,$C3,$F0,$32,$3C,$2F,$93,$E0,$7C,$BF,$05,$E6,$F8,$29,$5E
	.dc.b	$68,$F4,$FC,$17,$AB,$C4,$93,$EA,$8F,$6F,$91,$EE,$F8,$19,$3E,$E8
	.dc.b	$F8,$3F,$05,$23,$E0,$BF,$13,$E0,$7E,$2F,$C1,$49,$F8,$AF,$C8,$F8
	.dc.b	$2F,$93,$F0,$52,$7E,$4B,$F3,$3E,$0B,$FD,$B3,$F0,$9F,$E4,$0E,$9F
	.dc.b	$9D,$5F,$EC,$E8,$C1,$FE,$2D,$7F,$DA,$AF,$F8,$EB,$CA,$FF,$4C,$36
	.dc.b	$7F,$D7,$AD,$9F,$F6,$02,$F8,$0E,$FF,$03,$E0,$F8,$19,$5D,$E3,$C6
	.dc.b	$C3,$C4,$79,$BE,$0F,$CF,$F0,$72,$3C,$CF,$AA,$FA,$C7,$A6,$FA,$BE
	.dc.b	$8B,$E8,$FA,$8B,$97,$AB,$0E,$91,$EE,$C2,$CF,$BB,$0E,$B3,$F0,$B0
	.dc.b	$EB,$3F,$0B,$0F,$97,$F0,$72,$BE,$53,$F3,$BE,$03,$9E,$9F,$1B,$2B
	.dc.b	$5F,$1B,$2F,$A7,$F0,$3F,$53,$E0,$E4,$FD,$38,$E8,$FD,$2C,$6E,$7F
	.dc.b	$4B,$2E,$79,$36,$FA,$4D,$CF,$E3,$61,$CC,$FB,$D9,$76,$7E,$EE,$5D
	.dc.b	$2B,$ED,$E5,$67,$ED,$E3,$DB,$1E,$A4,$CF,$D3,$3D,$08,$74,$3F,$19
	.dc.b	$F0,$7F,$90,$23,$F8,$FF,$83,$E9,$A7,$5A,$FC,$AF,$C1,$7E,$5B,$E0
	.dc.b	$E5,$FE,$56,$38,$BF,$94,$C2,$EF,$F9,$4C,$79,$CF,$E1,$8F,$CE,$61
	.dc.b	$D1,$FC,$F6,$1D,$1F,$CE,$E1,$CB,$5F,$CD,$E3,$67,$F3,$78,$74,$7F
	.dc.b	$0C,$74,$3C,$87,$42,$1D,$0F,$D6,$7C,$1F,$EC,$08,$FE,$BF,$E0,$7A
	.dc.b	$67,$FA,$CD,$3E,$BC,$FE,$C3,$E0,$BF,$6F,$F0,$52,$3F,$6D,$1C,$9F
	.dc.b	$DA,$E1,$6F,$F6,$B8,$74,$CF,$F2,$15,$E7,$3A,$1F,$BF,$C3,$A3,$83
	.dc.b	$99,$E8,$21,$FC,$0A,$7A,$3F,$BE,$C7,$CF,$7F,$C1,$31,$E6,$FF,$06
	.dc.b	$BF,$E1,$F1,$FC,$43,$E0,$A5,$FF,$0F,$7F,$89,$FC,$17,$4C,$BF,$31
	.dc.b	$3E,$7C,$3F,$87,$FC,$17,$F1,$8F,$82,$93,$FC,$5E,$39,$BF,$C5,$B0
	.dc.b	$B5,$FC,$5B,$0E,$9C,$7F,$97,$AE,$07,$F1,$FC,$39,$BF,$C7,$B0,$E3
	.dc.b	$8F,$F1,$CC,$6C,$FF,$1C,$C3,$9E,$9F,$02,$BF,$90,$DF,$2C,$BF,$92
	.dc.b	$E3,$FC,$BA,$3F,$DD,$58,$25,$5F,$EE,$D1,$F9,$87,$F8,$C1,$C3,$67
	.dc.b	$6D,$8D,$FC,$17,$8A,$7E,$34,$70,$F8,$3E,$F7,$DA,$78,$FC,$46,$2B
	.dc.b	$AB,$63,$AD,$59,$F1,$97,$0F,$01,$30,$F0,$57,$F4,$4E,$7D,$5E,$F0
	.dc.b	$92,$3E,$06,$1E,$57,$18,$F9,$63,$BC,$61,$F8,$27,$84,$F1,$61,$E0
	.dc.b	$9F,$E6,$93,$B3,$D2,$59,$F9,$75,$E2,$8F,$09,$A2,$C7,$D0,$F8,$3F
	.dc.b	$71,$23,$96,$4F,$84,$F0,$57,$F5,$EC,$FE,$27,$C2,$7C,$5F,$84,$98
	.dc.b	$70,$8E,$4F,$6C,$3F,$C7,$CD,$C7,$E6,$63,$D2,$3F,$A1,$1C,$4F,$0D
	.dc.b	$F4,$79,$5F,$4C,$FE,$46,$1D,$1B,$87,$4A,$FE,$40,$7F,$16,$24,$7C
	.dc.b	$87,$99,$D8,$86,$0F,$47,$C1,$FE,$04,$8F,$C4,$F8,$4F,$46,$9F,$D3
	.dc.b	$F4,$FC,$37,$C1,$FE,$1F,$E1,$25,$FC,$A8,$E2,$F4,$3E,$6E,$1F,$8D
	.dc.b	$C3,$9F,$3F,$F8,$1E,$78,$1F,$4A,$F9,$BF,$2E,$F9,$E5,$F0,$B0,$B9
	.dc.b	$7E,$A3,$1E,$79,$79,$30,$E4,$43,$91,$F7,$BE,$0F,$F3,$C4,$7F,$03
	.dc.b	$F0,$9C,$9F,$C1,$FC,$27,$E8,$FE,$12,$46,$0E,$44,$3F,$01,$5F,$A9
	.dc.b	$AE,$51,$F8,$3A,$E2,$FE,$16,$BF,$5D,$1E,$18,$E2,$73,$CF,$EF,$D7
	.dc.b	$27,$F0,$33,$C9,$FC,$1D,$72,$AF,$CC,$FC,$27,$EE,$7E,$12,$46,$0F
	.dc.b	$D4,$7C,$2B,$EC,$9E,$9F,$F0,$38,$C7,$A7,$64,$D9,$FE,$09,$F0,$B2
	.dc.b	$FF,$4E,$FF,$07,$F8,$6F,$E1,$19,$78,$27,$F4,$59,$61,$F0,$E3,$E0
	.dc.b	$6E,$3D,$CE,$CF,$42,$1F,$AB,$C3,$B3,$EB,$3A,$9F,$B8,$AE,$CF,$F0
	.dc.b	$0C,$BD,$D9,$75,$8E,$E7,$89,$F5,$3D,$4E,$FA,$7E,$BB,$3F,$E4,$5F
	.dc.b	$0D,$FC,$97,$F4,$DF,$F3,$3C,$FF,$19,$6F,$27,$F3,$89,$F4,$DF,$87
	.dc.b	$5F,$92,$F8,$0C,$63,$CF,$E1,$64,$7F,$19,$C3,$FD,$3C,$7F,$A7,$5F
	.dc.b	$E1,$57,$E1,$E9,$91,$FC,$5A,$3B,$E5,$FC,$FF,$4F,$B3,$3D,$E3,$19
	.dc.b	$7C,$F8,$FE,$43,$87,$6C,$7C,$31,$8F,$86,$F7,$BE,$3C,$BB,$BE,$23
	.dc.b	$E4,$1D,$74,$FE,$7D,$97,$82,$3D,$A7,$AE,$FA,$E5,$F1,$DF,$95,$87
	.dc.b	$CB,$F8,$6F,$99,$F0,$D2,$FE,$14,$7B,$31,$EB,$A7,$CA,$9F,$99,$C5
	.dc.b	$97,$FB,$5E,$AF,$85,$FA,$C4,$8F,$99,$F0,$DD,$DF,$9D,$97,$77,$E1
	.dc.b	$E5,$CF,$6F,$56,$DF,$23,$2E,$DA,$33,$ED,$87,$D7,$CF,$B3,$F3,$33
	.dc.b	$EC,$FC,$EC,$FB,$63,$F8,$7C,$FB,$69,$EB,$8E,$D5,$F3,$0F,$AB,$87
	.dc.b	$48,$FC,$73,$F9,$1C,$3F,$25,$F0,$DF,$93,$F8,$69,$7F,$89,$8E,$BA
	.dc.b	$FE,$7E,$3E,$99,$F8,$0F,$85,$FC,$D1,$1F,$C8,$7C,$37,$47,$F2,$57
	.dc.b	$D1,$FC,$3D,$F4,$7F,$00,$38,$3F,$5A,$FE,$66,$FA,$3D,$CE,$A7,$94
	.dc.b	$EA,$43,$A9,$F0,$73,$F8,$6F,$DE,$31,$C1,$FE,$27,$E0,$57,$6A,$FD
	.dc.b	$93,$FB,$5C,$BF,$8C,$7C,$27,$F1,$9F,$83,$98,$FF,$10,$8F,$53,$D4
	.dc.b	$F8,$B7,$FA,$6F,$83,$FE,$40,$48,$FE,$2F,$F0,$7D,$1F,$E3,$37,$D1
	.dc.b	$FD,$F5,$F4,$BF,$49,$FB,$0E,$9E,$53,$FC,$0F,$0B,$3F,$B4,$C3,$A4
	.dc.b	$FF,$07,$C3,$A4,$FE,$E3,$0E,$9A,$7E,$8B,$83,$F1,$38,$F2,$8F,$D2
	.dc.b	$BF,$CD,$F1,$FE,$2D,$F0,$BF,$C5,$FE,$1A,$67,$FC,$36,$39,$BF,$67
	.dc.b	$97,$F2,$58,$7E,$BB,$5F,$E3,$DF,$0D,$2B,$F8,$A3,$FC,$87,$4E,$B1
	.dc.b	$FC,$5F,$1E,$AF,$F0,$CC,$7A,$F1,$7E,$AB,$1E,$93,$C8,$87,$F1,$2C
	.dc.b	$3A,$3E,$53,$A1,$0E,$87,$E0,$8F,$46,$BF,$9D,$DB,$E7,$47,$47,$F9
	.dc.b	$A4,$FA,$63,$F8,$FE,$5F,$CE,$77,$FE,$75,$A4,$9F,$E6,$11,$E8,$D7
	.dc.b	$E1,$C7,$C5,$9F,$E8,$18,$C8,$FE,$6C,$FF,$43,$DF,$A4,$7F,$39,$9E
	.dc.b	$8F,$F2,$EB,$E9,$5F,$C2,$B8,$BA,$E9,$FC,$2E,$3A,$57,$F3,$4C,$39
	.dc.b	$BF,$CD,$B0,$E7,$A7,$F4,$FC,$79,$CF,$C3,$C3,$93,$E4,$21,$CC,$FE
	.dc.b	$99,$C1,$FD,$67,$E1,$A4,$60,$F9,$9B,$FD,$77,$E6,$9D,$7E,$23,$FD
	.dc.b	$32,$FF,$9A,$1B,$8F,$F9,$E5,$B8,$7F,$A6,$D7,$C3,$FF,$67,$C7,$BD
	.dc.b	$FF,$50,$C7,$0F,$83,$7F,$9F,$97,$84,$FE,$17,$1D,$CF,$F4,$A6,$1D
	.dc.b	$9F,$F5,$4B,$71,$3E,$08,$F6,$1D,$CF,$ED,$97,$E0,$7F,$7F,$3F,$AB
	.dc.b	$7F,$8D,$D7,$83,$2F,$64,$78,$08,$7F,$A3,$6F,$B3,$F0,$FD,$7F,$2F
	.dc.b	$93,$23,$51,$F5,$9E,$E7,$A3,$E1,$BD,$64,$7A,$7C,$3F,$BA,$79,$47
	.dc.b	$CF,$AC,$3F,$1F,$8B,$F4,$19,$FE,$B2,$3E,$A6,$1C,$DA,$74,$3C,$67
	.dc.b	$A6,$BA,$E5,$F6,$70,$F3,$3F,$AA,$9E,$24,$7E,$F5,$7D,$2C,$7E,$9F
	.dc.b	$C2,$CA,$F8,$8F,$D6,$F8,$5F,$AF,$F0,$DF,$63,$E1,$E5,$7C,$98,$FB
	.dc.b	$7F,$0B,$F7,$3E,$16,$4F,$CD,$8F,$BF,$F0,$DF,$80,$F8,$79,$3F,$46
	.dc.b	$3F,$07,$F0,$F2,$3E,$A3,$F8,$6F,$87,$FC,$3F,$C3,$49,$FB,$0F,$E2
	.dc.b	$BE,$1F,$F1,$7F,$0F,$F8,$CF,$87,$95,$F7,$63,$F1,$FF,$0D,$23,$F0
	.dc.b	$0F,$E4,$BE,$1F,$FD,$CD,$D1,$FE,$F3,$1F,$0D,$83,$2E,$BE,$1E,$5F
	.dc.b	$FB,$56,$11,$CB,$E1,$F9,$FC,$1F,$4F,$87,$97,$B8,$EF,$F0,$D2,$39
	.dc.b	$3E,$2F,$87,$F1,$FC,$27,$93,$E1,$25,$77,$8F,$3F,$C2,$FA,$3E,$1A
	.dc.b	$4F,$8E,$3D,$7F,$0D,$EC,$F8,$69,$3E,$78,$F7,$FC,$37,$C0,$F8,$69
	.dc.b	$3E,$B8,$F8,$7F,$0D,$23,$DC,$FC,$6F,$85,$F8,$FF,$0D,$27,$E1,$3F
	.dc.b	$2B,$E1,$BE,$5F,$C3,$49,$F8,$C9,$FF,$7E,$7E,$9B,$FE,$6E,$5C,$E8
	.dc.b	$7C,$36,$E7,$FD,$FB,$18,$78,$C6,$7C,$CE,$31,$CA,$7F,$CD,$D1,$8A
	.dc.b	$E9,$F0,$9B,$7C,$47,$7F,$85,$C5,$F1,$6C,$70,$8F,$3F,$C2,$E3,$07
	.dc.b	$C2,$D3,$11,$9F,$84,$D7,$C2,$CB,$F7,$C7,$1F,$85,$91,$F0,$9E,$9F
	.dc.b	$0B,$D7,$E1,$24,$FC,$67,$C1,$F0,$BE,$1F,$85,$F1,$7C,$2C,$AF,$97
	.dc.b	$1E,$5F,$82,$91,$F3,$9F,$47,$C1,$FA,$7D,$D9,$3F,$49,$F6,$7C,$27
	.dc.b	$B7,$E0,$E4,$FD,$67,$E0,$7C,$17,$C4,$FC,$6F,$F9,$B1,$FC,$A3,$A3
	.dc.b	$E7,$37,$9B,$E6,$BA,$F8,$0E,$1F,$03,$31,$FC,$15,$F0,$9E,$9F,$03
	.dc.b	$27,$38,$72,$8F,$07,$C0,$F8,$7E,$06,$56,$8F,$C4,$E4,$8F,$37,$C0
	.dc.b	$F9,$FE,$02,$56,$F1,$5A,$7D,$9F,$03,$ED,$F8,$0F,$77,$C0,$CC,$38
	.dc.b	$5A,$7E,$17,$C0,$7C,$3F,$81,$95,$A8,$F8,$DF,$03,$F1,$FE,$06,$4F
	.dc.b	$28,$F9,$5F,$03,$F2,$FE,$06,$4F,$68,$F9,$DF,$01,$23,$C2,$FD,$1F
	.dc.b	$81,$FA,$5F,$03,$F4,$FE,$06,$57,$9A,$3F,$33,$FA,$6F,$E9,$21,$D3
	.dc.b	$FA,$38,$C3,$79,$37,$F0,$5C,$3E,$0B,$8F,$C1,$4C,$3E,$46,$1F,$C7
	.dc.b	$9B,$F0,$3D,$88,$EB,$E0,$BC,$3F,$05,$E2,$F8,$29,$3C,$A3,$CB,$F0
	.dc.b	$3E,$6F,$82,$93,$DA,$3D,$3F,$03,$EA,$F8,$29,$3E,$28,$F6,$FC,$0F
	.dc.b	$BB,$E0,$A4,$F9,$A3,$E0,$FC,$0C,$8F,$4B,$F1,$3E,$0B,$E2,$FC,$17
	.dc.b	$C6,$F8,$29,$5E,$E8,$F9,$3F,$05,$23,$E0,$BF,$33,$E0,$BE,$6F,$C1
	.dc.b	$49,$F8,$AF,$F0,$8F,$86,$FE,$39,$D3,$FE,$28,$F8,$3C,$12,$BF,$25
	.dc.b	$3F,$C5,$A7,$F8,$9A,$F8,$3E,$5F,$07,$2B,$71,$D7,$E0,$7B,$7C,$14
	.dc.b	$9E,$71,$E1,$F8,$1F,$17,$C1,$49,$EF,$1E,$5F,$82,$91,$E2,$7D,$1F
	.dc.b	$05,$E9,$F8,$2F,$57,$C1,$4A,$F3,$C7,$B7,$E0,$7D,$DF,$05,$27,$D7
	.dc.b	$1F,$07,$E0,$A4,$7B,$9F,$89,$F0,$5F,$17,$E0,$BE,$37,$C1,$4A,$F8
	.dc.b	$71,$D5,$F8,$B8,$75,$7F,$33,$C9,$F1,$B1,$EB,$18,$3F,$A4,$F2,$32
	.dc.b	$CF,$97,$FD,$06,$19,$F8,$6B,$E8,$E7,$E1,$79,$63,$FE,$5E,$CF,$ED
	.dc.b	$7C,$17,$DB,$F8,$29,$5F,$62,$3F,$A2,$FE,$9B,$FE,$57,$24,$63,$2C
	.dc.b	$FC,$14,$8F,$E2,$12,$8E,$3F,$07,$DA,$7F,$07,$87,$67,$C7,$6C,$3B
	.dc.b	$C4,$39,$9D,$CE,$B7,$DD,$F5,$E1,$DD,$F0,$E1,$DD,$EB,$87,$7D,$7E
	.dc.b	$36,$1D,$67,$89,$63,$FB,$55,$FF,$3D,$8F,$EE,$EF,$57,$BE,$1E,$19
	.dc.b	$DE,$1F,$1B,$E0,$FE,$3F,$C1,$CB,$F8,$71,$D7,$5F,$EE,$78,$FC,$A9
	.dc.b	$F8,$FA,$73,$9C,$E5,$67,$C3,$8F,$57,$E6,$E3,$D5,$F8,$38,$F5,$7C
	.dc.b	$E5,$B6,$3E,$B6,$E1,$D5,$B3,$F2,$0F,$39,$D8,$87,$63,$E9,$90,$F9
	.dc.b	$27,$A8,$FA,$A4,$3E,$39,$D4,$FA,$FF,$07,$F8,$CF,$83,$91,$F8,$88
	.dc.b	$F4,$5F,$C5,$8F,$BB,$9F,$9E,$3E,$96,$9F,$5B,$1F,$C6,$D7,$13,$F2
	.dc.b	$38,$74,$7B,$9E,$79,$FB,$AF,$6B,$E9,$5D,$74,$FC,$FE,$3C,$6F,$F0
	.dc.b	$78,$63,$0E,$59,$7D,$A7,$91,$EF,$9F,$23,$63,$EF,$DF,$EA,$BE,$0F
	.dc.b	$F5,$7F,$07,$2B,$F4,$F1,$F0,$1E,$67,$C8,$25,$F4,$7F,$3F,$5C,$DF
	.dc.b	$D7,$D7,$37,$F4,$75,$CD,$F3,$9F,$9A,$CF,$DE,$F9,$4F,$C6,$9E,$23
	.dc.b	$99,$0E,$67,$2C,$3F,$16,$F9,$B0,$F2,$12,$3F,$33,$1F,$7E,$3F,$57
	.dc.b	$F0,$3F,$C2,$BE,$0E,$4F,$F0,$68,$F3,$3F,$8A,$3E,$F9,$E5,$AF,$CA
	.dc.b	$BF,$09,$F9,$6F,$13,$F8,$7C,$F4,$7F,$81,$CF,$4A,$FB,$F1,$F9,$53
	.dc.b	$91,$FB,$CD,$7F,$00,$F3,$3F,$85,$D7,$39,$F4,$C7,$C1,$DB,$F2,$91
	.dc.b	$FC,$1B,$1F,$97,$3D,$31,$FD,$7E,$9F,$CA,$3E,$07,$F9,$4F,$C1,$4A
	.dc.b	$FE,$4B,$1E,$5A,$FC,$54,$79,$70,$F9,$4F,$94,$F8,$58,$B0,$E9,$1F
	.dc.b	$CB,$B0,$E8,$FF,$22,$C3,$A3,$FC,$08,$E9,$87,$D4,$7C,$D7,$FA,$A8
	.dc.b	$F9,$71,$C4,$E8,$43,$07,$D5,$C3,$E5,$B6,$3F,$8D,$E3,$F0,$A7,$D3
	.dc.b	$1F,$0A,$7F,$A1,$FC,$0F,$F4,$AF,$82,$91,$83,$E6,$1F,$8B,$9F,$35
	.dc.b	$72,$C3,$CD,$3F,$96,$3F,$9F,$DF,$27,$FA,$0D,$FF,$56,$8F,$83,$7F
	.dc.b	$96,$9F,$82,$7C,$B3,$E0,$EB,$FC,$21,$E6,$7F,$58,$AE,$98,$7F,$42
	.dc.b	$BE,$78,$7F,$61,$C7,$DD,$8F,$98,$F9,$35,$FD,$23,$E0,$7F,$B3,$FC
	.dc.b	$14,$CB,$FB,$1C,$72,$7F,$AD,$1E,$23,$EE,$ED,$FD,$BA,$38,$47,$F5
	.dc.b	$C9,$C3,$F1,$8F,$C9,$61,$E2,$7D,$97,$FD,$49,$EE,$7F,$72,$9E,$71
	.dc.b	$E5,$8E,$47,$D9,$3F,$82,$1F,$8E,$D3,$F4,$6F,$F6,$CF,$82,$FF,$00
	.dc.b	$FC,$14,$8F,$EF,$B1,$EB,$7E,$6E,$7F,$B6,$BF,$B1,$7F,$35,$FE,$D7
	.dc.b	$5C,$6B,$FB,$BD,$FF,$83,$23,$94,$7F,$00,$3E,$21,$F9,$6C,$F4,$FF
	.dc.b	$83,$27,$95,$74,$8F,$5B,$F9,$43,$F6,$8F,$AC,$F9,$A7,$F8,$53,$E0
	.dc.b	$7F,$C3,$64,$7F,$C2,$FF,$05,$FE,$E2,$FD,$37,$FD,$0F,$9F,$FB,$A1
	.dc.b	$7C,$1E,$4D,$46,$FE,$0E,$5F,$FA,$BE,$BF,$D1,$0F,$F8,$B9,$7C,$1F
	.dc.b	$78,$FF,$0B,$E3,$72,$FF,$07,$E5,$D2,$79,$65,$DD,$E9,$97,$67,$A6
	.dc.b	$56,$FF,$E0,$06,$5D,$A7,$DC,$43,$B9,$EE,$39,$9D,$88,$79,$EB,$A1
	.dc.b	$F0,$6A,$C2,$13,$DE,$3D,$DF,$05,$F2,$48,$FC,$3F,$82,$F6,$3F,$0F
	.dc.b	$1F,$63,$DC,$F9,$18,$76,$C3,$AB,$F4,$F4,$E7,$3D,$8F,$9D,$87,$D7
	.dc.b	$C3,$AE,$5F,$0B,$1E,$AF,$C4,$C7,$A6,$5F,$52,$3E,$3E,$5D,$71,$E6
	.dc.b	$43,$B1,$F6,$70,$EB,$7F,$83,$C3,$AB,$F8,$53,$AD,$76,$CB,$F1,$5F
	.dc.b	$03,$30,$FC,$23,$F8,$DF,$82,$E7,$1F,$2F,$3F,$C6,$BF,$4B,$3F,$43
	.dc.b	$F1,$08,$7E,$00,$FC,$75,$F5,$7F,$1D,$76,$7F,$07,$7D,$B1,$FA,$F9
	.dc.b	$7A,$2F,$C1,$1E,$87,$AD,$7E,$2B,$0E,$53,$FA,$1C,$2D,$79,$F0,$E9
	.dc.b	$8F,$A2,$3E,$E1,$F8,$C3,$CC,$4C,$3E,$7B,$E8,$3F,$4D,$F0,$3F,$AF
	.dc.b	$23,$FA,$9F,$82,$EB,$9F,$9F,$2F,$C7,$9E,$79,$E8,$FE,$D7,$3E,$31
	.dc.b	$FA,$FB,$B8,$7E,$93,$0E,$73,$F3,$6F,$EC,$3F,$01,$FC,$F3,$4F,$A9
	.dc.b	$7C,$A3,$F8,$15,$D9,$FB,$15,$D1,$F8,$27,$E6,$CF,$29,$F7,$74,$FC
	.dc.b	$23,$FC,$17,$E0,$7F,$87,$11,$FE,$11,$F0,$5C,$F2,$FC,$D4,$7C,$B7
	.dc.b	$E4,$BD,$31,$FD,$2B,$F2,$CF,$E1,$D5,$CD,$A7,$43,$94,$7E,$65,$F8
	.dc.b	$47,$E0,$8F,$80,$F4,$7B,$69,$FB,$DA,$FE,$1B,$5D,$5B,$3D,$4F,$80
	.dc.b	$FA,$1F,$86,$FE,$69,$E7,$87,$C2,$7F,$92,$7C,$0F,$F2,$F2,$3F,$CA
	.dc.b	$7E,$0B,$E1,$63,$E7,$8E,$AF,$D6,$79,$9F,$06,$FF,$90,$D7,$28,$EA
	.dc.b	$73,$3E,$09,$FA,$FD,$BF,$19,$97,$F3,$E8,$FA,$C7,$C2,$7F,$83,$5F
	.dc.b	$E3,$63,$F9,$E5,$75,$BE,$F1,$D4,$F2,$9D,$4F,$9B,$3F,$98,$7A,$63
	.dc.b	$F0,$67,$F9,$BF,$C1,$7F,$4F,$24,$FF,$49,$F8,$2F,$82,$F3,$3A,$DF
	.dc.b	$D5,$AF,$C1,$91,$F3,$47,$F4,$FA,$E6,$D3,$A1,$F0,$A3,$F8,$1B,$E6
	.dc.b	$C7,$E1,$47,$53,$E7,$57,$59,$FD,$19,$FD,$76,$BA,$C7,$AA,$3A,$9E
	.dc.b	$6A,$FE,$09,$1F,$83,$3F,$04,$73,$AE,$D3,$F4,$0F,$EC,$DC,$1F,$DC
	.dc.b	$7E,$06,$47,$F6,$B7,$FB,$A7,$C1,$7C,$4A,$F4,$CF,$7C,$FF,$82,$61
	.dc.b	$FC,$66,$38,$10,$FE,$E5,$7C,$9F,$85,$8F,$9A,$FF,$1D,$5E,$67,$E0
	.dc.b	$92,$7C,$D1,$D5,$E6,$75,$7F,$B8,$D7,$58,$F8,$5A,$FE,$3A,$BC,$A7
	.dc.b	$33,$FB,$4E,$5F,$E0,$FF,$82,$95,$FE,$09,$7F,$C2,$5F,$05,$EF,$CB
	.dc.b	$CD,$3F,$B7,$3F,$42,$7E,$DA,$B8,$BF,$E0,$FA,$B9,$7F,$81,$2F,$96
	.dc.b	$3F,$CE,$33,$F1,$DF,$DC,$38,$1E,$12,$1F,$E1,$99,$E2,$FC,$88,$E3
	.dc.b	$97,$AA,$3F,$C4,$1F,$04,$66,$1F,$E0,$77,$FC,$11,$FA,$6B,$E3,$35
	.dc.b	$09,$D5,$AB,$53,$D0,$E7,$1D,$4D,$7C,$27,$2A,$E3,$48,$C4,$EE,$31
	.dc.b	$5C,$8F,$1F,$C6,$69,$F2,$FC,$25,$AF,$17,$C2,$DB,$36,$71,$F0,$99
	.dc.b	$F8,$59,$87,$5C,$3D,$53,$C7,$E1,$39,$7C,$2C,$AD,$64,$9E,$DF,$0B
	.dc.b	$DF,$E1,$65,$6A,$3C,$5F,$09,$23,$8B,$E5,$F8,$5F,$37,$C2,$79,$FE
	.dc.b	$16,$57,$68,$F5,$7C,$2C,$8F,$0B,$ED,$F8,$5F,$77,$C2,$7B,$FE,$16
	.dc.b	$57,$9A,$3E,$17,$C2,$48,$F4,$BF,$17,$E1,$7E,$47,$47,$DC,$9F,$B9
	.dc.b	$C9,$F5,$E3,$1D,$3F,$79,$B1,$9F,$5F,$8F,$AD,$CB,$D5,$98,$7C,$BB
	.dc.b	$FF,$60,$DF,$DD,$F8,$4E,$FF,$0B,$2B,$51,$E2,$F8,$5F,$1F,$C2,$C9
	.dc.b	$E5,$1E,$6F,$84,$91,$D5,$F4,$FC,$2F,$AB,$E1,$64,$F8,$5F,$6F,$C2
	.dc.b	$FB,$BE,$13,$DF,$F0,$B2,$BC,$D1,$F0,$BE,$12,$47,$A5,$F8,$BF,$0B
	.dc.b	$F1,$BE,$12,$4F,$B5,$F9,$3F,$09,$F2,$BE,$12,$4F,$C1,$7E,$6F,$C2
	.dc.b	$7C,$EF,$84,$F9,$FF,$09,$2B,$E3,$47,$E6,$7F,$4D,$FE,$76,$7F,$08
	.dc.b	$E6,$34,$CF,$A5,$BF,$5A,$5F,$E6,$67,$FB,$0B,$FE,$7F,$5F,$0B,$D4
	.dc.b	$91,$BF,$85,$F0,$7C,$2C,$8E,$4F,$8F,$E1,$7C,$9F,$0B,$E5,$F8,$59
	.dc.b	$5D,$E3,$D1,$F0,$92,$3C,$4F,$AF,$E1,$7D,$9F,$09,$ED,$F8,$59,$5E
	.dc.b	$78,$F8,$1F,$09,$23,$D4,$FC,$3F,$85,$E9,$1F,$94,$3D,$98,$63,$C9
	.dc.b	$FE,$23,$C0,$CB,$BD,$7C,$3C,$BB,$BC,$78,$FE,$5E,$7D,$E3,$16,$CF
	.dc.b	$18,$F4,$CF,$B6,$1F,$C6,$69,$9E,$30,$F9,$D9,$F8,$1E,$79,$F8,$31
	.dc.b	$F2,$E7,$D3,$1E,$59,$DB,$FA,$D9,$F3,$9F,$BF,$95,$CF,$D9,$5F,$1C
	.dc.b	$F6,$10,$F4,$E1,$CE,$BF,$0F,$85,$A6,$1C,$F0,$E9,$9F,$9F,$2F,$93
	.dc.b	$9F,$9F,$1F,$76,$5C,$72,$E5,$8F,$1C,$FE,$E9,$F8,$FC,$79,$5F,$C5
	.dc.b	$C3,$94,$FE,$47,$0E,$4F,$DC,$3F,$2D,$7F,$9F,$F8,$4F,$D0,$7C,$2C
	.dc.b	$9F,$CD,$47,$E8,$FE,$13,$F4,$9F,$07,$27,$F3,$D1,$FA,$7F,$83,$91
	.dc.b	$FA,$17,$FA,$17,$E9,$BF,$D4,$A3,$EE,$DF,$57,$EF,$E1,$E1,$79,$9F
	.dc.b	$A9,$C3,$B3,$F8,$4C,$3A,$BF,$86,$C2,$EF,$FD,$83,$1E,$CF,$E4,$B1
	.dc.b	$EC,$FE,$53,$1E,$D7,$CC,$EC,$43,$07,$53,$D4,$43,$F3,$15,$DB,$5F
	.dc.b	$C9,$5F,$67,$F2,$95,$DA,$FF,$47,$8F,$A9,$FD,$2E,$5F,$2F,$1E,$91
	.dc.b	$FA,$8C,$BD,$78,$7F,$75,$CB,$B5,$21,$1F,$DC,$A3,$C7,$F0,$7F,$38
	.dc.b	$F9,$FE,$6F,$D0,$F8,$2F,$A3,$F0,$72,$BF,$B6,$5A,$7E,$AF,$C1,$7D
	.dc.b	$6F,$83,$95,$A8,$E6,$F0,$C7,$B3,$CB,$1E,$CF,$4C,$7B,$3F,$1B,$0E
	.dc.b	$AF,$83,$0E,$AF,$CA,$C3,$F0,$BF,$03,$23,$CA,$FE,$23,$E0,$FF,$13
	.dc.b	$F0,$3F,$8A,$F8,$29,$5E,$A8,$FC,$6F,$C1,$7E,$3B,$E0,$A4,$FB,$A3
	.dc.b	$F2,$3F,$03,$23,$E0,$BF,$94,$F8,$2E,$B1,$F1,$30,$EA,$FD,$EC,$3F
	.dc.b	$33,$F0,$52,$3E,$4B,$F9,$CF,$82,$FC,$EF,$C0,$C9,$F9,$AF,$E8,$3E
	.dc.b	$0B,$AC,$7E,$23,$0E,$BD,$1F,$97,$C7,$F4,$BF,$05,$2B,$EA,$BF,$A8
	.dc.b	$F8,$2F,$F1,$57,$E9,$BF,$DF,$A7,$F0,$4B,$E0,$F2,$6A,$37,$F0,$7C
	.dc.b	$3E,$0E,$47,$F8,$AE,$BF,$C0,$23,$E0,$BA,$FC,$1C,$9D,$47,$83,$E0
	.dc.b	$BC,$3F,$07,$27,$94,$76,$7A,$61,$D9,$FE,$79,$87,$6F,$83,$F4,$92
	.dc.b	$3C,$5F,$07,$D5,$F2,$61,$D5,$E3,$87,$57,$9E,$1D,$78,$3F,$CB,$B8
	.dc.b	$F5,$9F,$66,$3D,$5F,$0E,$3F,$23,$E0,$FE,$4F,$C1,$C9,$F8,$51,$D7
	.dc.b	$0E,$78,$F5,$F1,$7F,$CC,$99,$75,$BF,$5E,$5D,$6F,$C3,$97,$D4,$F8
	.dc.b	$29,$5F,$35,$FA,$FF,$07,$F6,$3E,$0E,$4F,$D1,$7E,$DF,$C1,$FD,$CF
	.dc.b	$83,$93,$F5,$5F,$BF,$F0,$7D,$23,$E5,$E1,$72,$F8,$38,$F4,$9F,$9F
	.dc.b	$8D,$9F,$8B,$8F,$49,$FA,$78,$D9,$F9,$38,$FE,$33,$E0,$E5,$7E,$15
	.dc.b	$FC,$7F,$C1,$FE,$43,$E0,$BF,$23,$F0,$72,$BF,$15,$1F,$E7,$DF,$86
	.dc.b	$FF,$6F,$93,$8B,$0F,$49,$FC,$1E,$17,$2F,$B3,$8F,$59,$FC,$A6,$3D
	.dc.b	$5F,$C4,$E3,$D7,$19,$1F,$E8,$3C,$7B,$9E,$1F,$84,$F1,$7C,$24,$9D
	.dc.b	$47,$97,$E1,$3C,$DF,$09,$27,$94,$7A,$7E,$12,$47,$57,$D9,$F0,$9D
	.dc.b	$63,$C1,$87,$57,$FC,$31,$8F,$F8,$F3,$23,$FE,$3C,$8F,$36,$3D,$DD
	.dc.b	$E3,$D9,$E3,$8D,$DB,$FD,$6F,$6C,$BE,$67,$C1,$FC,$DF,$84,$98,$7C
	.dc.b	$18,$FA,$1F,$07,$F4,$7E,$0A,$4F,$C5,$8F,$A9,$F0,$32,$3E,$43,$F5
	.dc.b	$FE,$0B,$AC,$7C,$BC,$3A,$BE,$CC,$3A,$BF,$3F,$0E,$AF,$C0,$C3,$A3
	.dc.b	$F0,$B0,$B9,$FA,$B1,$FC,$27,$C1,$FE,$17,$E0,$E5,$FD,$98,$FC,$47
	.dc.b	$C1,$FE,$27,$E0,$E4,$FD,$D8,$FC,$67,$C1,$C8,$FC,$03,$F8,$FF,$83
	.dc.b	$EB,$1F,$83,$C3,$AB,$F5,$30,$EA,$FE,$1F,$0E,$AF,$D8,$C3,$A3,$F6
	.dc.b	$B0,$B9,$7D,$2C,$7F,$39,$F0,$72,$BF,$20,$FF,$B6,$FF,$4D,$FE,$E7
	.dc.b	$CE,$7F,$B9,$5F,$FB,$C1,$B1,$FE,$ED,$6D,$7F,$79,$5F,$07,$CF,$E0
	.dc.b	$7A,$7C,$14,$AF,$F7,$05,$7F,$75,$6F,$C0,$F8,$48,$EB,$E0,$BC,$9F
	.dc.b	$07,$23,$8B,$E7,$C7,$B4,$74,$C3,$B3,$FE,$71,$C3,$B3,$7E,$0F,$DC
	.dc.b	$47,$C7,$F0,$7D,$5E,$18,$5D,$7E,$06,$3D,$A7,$A6,$3D,$A7,$E2,$69
	.dc.b	$F2,$7D,$AF,$95,$ED,$CB,$F7,$C7,$CD,$F8,$2F,$9D,$F0,$72,$7E,$1C
	.dc.b	$7D,$1F,$83,$91,$F1,$9F,$A9,$F0,$7F,$57,$E0,$FE,$B7,$C1,$CA,$F9
	.dc.b	$71,$D5,$F9,$B8,$75,$7E,$16,$1D,$1F,$89,$85,$D3,$EE,$63,$D6,$7E
	.dc.b	$AE,$3D,$5F,$95,$8F,$57,$E6,$63,$D6,$FF,$09,$8F,$3E,$5F,$C2,$65
	.dc.b	$6F,$F0,$99,$7E,$37,$E0,$E4,$7E,$09,$FC,$87,$C1,$FE,$47,$E0,$FF
	.dc.b	$25,$F0,$72,$BF,$0F,$06,$67,$FE,$47,$7F,$C9,$1F,$A6,$FF,$82,$73
	.dc.b	$3F,$E0,$98,$DC,$71,$33,$F0,$9C,$E3,$FD,$6D,$3C,$7A,$59,$F9,$A7
	.dc.b	$F1,$4F,$99,$F1,$9F,$9A,$CB,$CC,$FE,$3F,$2E,$99,$79,$72,$F1,$CE
	.dc.b	$1F,$16,$3E,$58,$F0,$E5,$97,$D6,$59,$F5,$FC,$25,$A3,$4C,$7B,$D9
	.dc.b	$F8,$39,$7F,$00,$F8,$AF,$1F,$84,$E5,$F0,$7C,$FE,$12,$56,$F2,$F9
	.dc.b	$33,$D6,$31,$97,$81,$B9,$78,$1C,$E3,$E0,$7E,$5B,$DD,$E3,$5D,$9F
	.dc.b	$9F,$87,$89,$EB,$F0,$3E,$DF,$81,$91,$83,$C3,$F0,$7F,$04,$8F,$93
	.dc.b	$E1,$3E,$27,$C1,$FC,$5F,$83,$93,$E8,$8F,$91,$87,$48,$FB,$F8,$61
	.dc.b	$EA,$43,$F0,$78,$7C,$FC,$3A,$C7,$C2,$C3,$AB,$F8,$8C,$3E,$A7,$C1
	.dc.b	$C8,$F8,$EF,$D7,$F8,$3F,$B1,$F0,$7F,$67,$E0,$E5,$7C,$C8,$FB,$9F
	.dc.b	$07,$23,$E7,$BF,$7F,$E0,$FA,$47,$E5,$B0,$BD,$1F,$8D,$E4,$65,$DA
	.dc.b	$BE,$BE,$5D,$9F,$CE,$E5,$D5,$FC,$FE,$57,$0E,$59,$7E,$37,$E0,$BF
	.dc.b	$1D,$F0,$72,$FF,$03,$1F,$93,$FD,$37,$F6,$10,$BF,$CF,$47,$F8,$1A
	.dc.b	$3F,$14,$D7,$5D,$1F,$B1,$8E,$33,$C3,$E1,$3A,$7C,$24,$BE,$31,$E3
	.dc.b	$7F,$23,$8F,$8D,$FD,$BE,$3E,$3C,$7C,$9F,$03,$27,$C3,$1E,$7B,$E9
	.dc.b	$1E,$23,$FD,$9D,$93,$1D,$47,$A3,$E0,$FD,$C4,$7D,$7F,$09,$F0,$7E
	.dc.b	$13,$E1,$7C,$24,$9F,$7C,$7C,$5F,$83,$F8,$DF,$09,$27,$E1,$C7,$37
	.dc.b	$C5,$A3,$3F,$F1,$64,$61,$BC,$1F,$33,$2F,$04,$FC,$CC,$BC,$0E,$F2
	.dc.b	$F0,$3F,$43,$2F,$03,$CF,$2F,$B1,$F0,$7F,$67,$E1,$24,$FD,$68,$FB
	.dc.b	$9F,$09,$F7,$7E,$12,$4F,$DA,$8F,$C0,$7C,$1F,$E0,$7E,$12,$4F,$DE
	.dc.b	$8E,$AF,$E0,$30,$EA,$FA,$F0,$EA,$FE,$13,$0E,$AF,$BF,$0E,$AF,$E2
	.dc.b	$30,$EA,$FC,$3C,$3F,$21,$F0,$7F,$91,$F8,$49,$3F,$8E,$8F,$CA,$7C
	.dc.b	$27,$E5,$7E,$12,$4F,$E4,$A3,$F3,$1F,$09,$FC,$AF,$CC,$FF,$0E,$9F
	.dc.b	$E1,$EE,$54,$76,$AF,$CC,$E3,$D9,$FA,$98,$F6,$7F,$3B,$8F,$67,$EC
	.dc.b	$63,$D9,$FE,$55,$8F,$6F,$89,$C4,$7E,$02,$3F,$C4,$FC,$FF,$CE,$BE
	.dc.b	$07,$C8,$79,$7E,$07,$CD,$F0,$1E,$7F,$81,$98,$74,$8F,$57,$C2,$FA
	.dc.b	$FE,$16,$4F,$82,$39,$BE,$2C,$2B,$E1,$C3,$AB,$E6,$C3,$AB,$E5,$C3
	.dc.b	$AB,$EA,$C3,$AB,$E9,$C3,$E5,$7C,$2F,$CB,$F8,$59,$3F,$02,$3E,$77
	.dc.b	$C2,$7C,$FF,$85,$93,$F1,$23,$E9,$7C,$14,$8F,$8E,$FD,$5F,$83,$EB
	.dc.b	$1F,$2B,$0E,$AF,$C9,$C3,$AB,$F3,$B0,$EA,$FC,$DC,$3A,$3F,$3F,$0B
	.dc.b	$E1,$FF,$73,$E1,$65,$F8,$3F,$83,$97,$F6,$1F,$C3,$7C,$1F,$E1,$FE
	.dc.b	$12,$4F,$DC,$7F,$15,$F0,$BF,$8B,$F8,$39,$3F,$80,$7F,$1D,$F0,$BD
	.dc.b	$23,$F0,$18,$5C,$FC,$58,$F4,$9F,$C2,$63,$67,$CD,$8F,$49,$FC,$46
	.dc.b	$36,$7D,$58,$FE,$6F,$E0,$FF,$39,$F0,$92,$FF,$23,$1F,$E0,$9F,$D3
	.dc.b	$7F,$C4,$84,$8C,$7C,$2E,$77,$91,$FE,$01,$9F,$EF,$CF,$F1,$65,$F0
	.dc.b	$BD,$E3,$F3,$B8,$77,$7F,$39,$87,$76,$1C,$CE,$E7,$5B,$EE,$F5,$7C
	.dc.b	$D5,$E7,$F8,$49,$3E,$37,$D5,$F0,$BE,$BF,$85,$93,$E7,$7D,$DF,$0B
	.dc.b	$EF,$F8,$5F,$81,$F0,$B2,$BD,$91,$D1,$F6,$61,$78,$3F,$C6,$DB,$32
	.dc.b	$EB,$5F,$07,$2B,$5A,$CB,$B4,$FC,$7C,$BB,$3F,$1F,$2F,$A1,$F0,$B2
	.dc.b	$3E,$63,$F4,$FE,$17,$EA,$7C,$24,$9F,$A0,$FD,$7F,$85,$FB,$1F,$0B
	.dc.b	$F6,$7E,$12,$57,$D5,$8E,$8F,$D5,$C2,$E9,$EA,$C7,$A4,$FD,$9C,$6C
	.dc.b	$FB,$B1,$EB,$3F,$7F,$1E,$AF,$DF,$C7,$F1,$1F,$07,$23,$F0,$8F,$E2
	.dc.b	$FE,$13,$F1,$9F,$09,$F8,$DF,$84,$95,$F8,$98,$FC,$87,$C1,$C8,$FC
	.dc.b	$63,$F9,$3F,$84,$E9,$1F,$8D,$C2,$E9,$F4,$B1,$EB,$3F,$93,$C7,$AB
	.dc.b	$F9,$3C,$7A,$3F,$95,$C6,$DF,$DA,$C7,$FC,$D1,$FA,$6F,$F8,$3B,$CF
	.dc.b	$FF,$40,$36,$D0,$8C,$37,$E1,$78,$FC,$2F,$2F,$85,$97,$FE,$65,$B4
	.dc.b	$F6,$F8,$5E,$FF,$0B,$2B,$51,$E2,$C6,$47,$17,$CB,$8F,$68,$E9,$8F
	.dc.b	$67,$FC,$0B,$8F,$67,$99,$E1,$C3,$B3,$AC,$3D,$FF,$09,$23,$CC,$FC
	.dc.b	$2F,$85,$F8,$7F,$09,$F1,$3E,$16,$57,$AE,$3E,$3F,$C2,$7C,$8F,$85
	.dc.b	$93,$EF,$8E,$8F,$93,$0B,$C3,$FE,$9D,$D1,$97,$6A,$F8,$D9,$76,$7D
	.dc.b	$79,$76,$7E,$56,$5D,$9F,$7E,$5F,$63,$E1,$3E,$CF,$C2,$C9,$FA,$11
	.dc.b	$F7,$3E,$12,$47,$D3,$7E,$FF,$C2,$FE,$03,$E0,$A4,$FD,$77,$F0,$7F
	.dc.b	$07,$D6,$3E,$D6,$1D,$5F,$9F,$87,$47,$E8,$E1,$77,$F7,$63,$D2,$7E
	.dc.b	$AE,$36,$7E,$16,$3F,$90,$F8,$2F,$C8,$FC,$1C,$BF,$C4,$47,$E5,$3E
	.dc.b	$13,$FD,$79,$FA,$6B,$E1,$69,$9A,$FF,$59,$47,$F8,$18,$9F,$0B,$C3
	.dc.b	$E1,$7B,$4F,$E4,$F1,$EC,$FE,$23,$1E,$4F,$F9,$3A,$FB,$3D,$30,$CE
	.dc.b	$7F,$ED,$59,$EB,$3E,$0C,$7B,$3E,$2F,$84,$F3,$91,$F2,$FC,$2F,$AB
	.dc.b	$E1,$3D,$7F,$0B,$27,$D3,$1E,$EF,$84,$91,$EC,$7E,$0F,$C2,$F3,$8F
	.dc.b	$36,$17,$3F,$26,$3D,$67,$E1,$E3,$D5,$F6,$63,$D1,$F7,$63,$6F,$D9
	.dc.b	$8F,$CE,$F8,$49,$5F,$31,$FA,$3F,$0B,$F4,$BE,$17,$E9,$FC,$2C,$AF
	.dc.b	$A3,$1F,$5B,$E1,$7E,$BF,$C2,$C9,$FA,$B1,$D1,$F9,$D8,$5C,$BE,$66
	.dc.b	$3D,$67,$ED,$E3,$D5,$FA,$98,$F5,$7E,$FE,$3D,$5F,$B1,$8F,$E1,$BE
	.dc.b	$17,$F0,$FF,$0B,$27,$F0,$B1,$F8,$AF,$85,$91,$F8,$87,$F1,$BF,$0B
	.dc.b	$F8,$EF,$83,$93,$F8,$C7,$F2,$3F,$09,$D6,$3F,$1F,$87,$57,$F1,$18
	.dc.b	$74,$7F,$15,$85,$DB,$F1,$18,$F4,$9F,$C7,$63,$67,$F1,$98,$F5,$9F
	.dc.b	$C9,$63,$D6,$7F,$21,$8F,$3F,$39,$5F,$CD,$3A,$19,$E6,$31,$9F,$5A
	.dc.b	$D8,$A6,$79,$AE,$59,$F4,$7F,$D5,$FD,$95,$BF,$0B,$9D,$3F,$A1,$C7
	.dc.b	$F4,$49,$4E,$71,$EF,$1C,$6B,$70,$8F,$F1,$3C,$61,$FF,$13,$D7,$B2
	.dc.b	$31,$F0,$9C,$E3,$B7,$C2,$DB,$E5,$F0,$B7,$03,$4C,$7C,$26,$6F,$59
	.dc.b	$47,$B5,$F9,$9E,$58,$73,$C6,$46,$8F,$09,$67,$C2,$DC,$7C,$2D,$C7
	.dc.b	$C2,$F1,$1F,$94,$BA,$5A,$7C,$D9,$FB,$73,$98,$C4,$70,$9D,$BB,$9D
	.dc.b	$36,$32,$FD,$78,$CB,$63,$2F,$DD,$8C,$B6,$32,$F8,$9A,$F8,$67,$C7
	.dc.b	$3C,$23,$CF,$D2,$F9,$4C,$B6,$3C,$CD,$7C,$8D,$7C,$4D,$79,$C7,$22
	.dc.b	$3C,$9A,$F2,$6A,$F9,$A8,$00

;--------------------------------------------------------------------------------
;	FLOGNP1.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
flognp1_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FLOGNP1.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FLOGNP1.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	flognp1_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1687+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	flognp1.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FLOGNP1.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1687+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	flognp1.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FLOGNP1.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
flognp1_data_compressed::
	.dc.b	$00,$02,$93,$04,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$FA,$77,$E2,$7F,$C2
	.dc.b	$CF,$C1,$F0,$9F,$C1,$BF,$D7,$5F,$EA,$2D,$F3,$38,$F9,$9C,$BC,$CE
	.dc.b	$7E,$67,$4F,$33,$AF,$99,$DB,$CC,$EF,$E6,$78,$3C,$CF,$0F,$99,$E2
	.dc.b	$F3,$3C,$7E,$67,$F6,$1F,$C5,$1F,$2F,$7E,$2B,$E7,$FA,$7F,$D9,$7E
	.dc.b	$6D,$F2,$9B,$F5,$71,$F0,$F9,$F9,$5D,$7C,$9E,$FE,$3B,$87,$C5,$F1
	.dc.b	$F7,$F9,$7B,$7C,$FD,$BE,$9D,$9D,$7B,$3B,$76,$77,$EC,$F0,$76,$78
	.dc.b	$7B,$3C,$5D,$9E,$3E,$CF,$27,$67,$97,$B3,$CD,$D9,$E7,$EC,$FE,$F9
	.dc.b	$C9,$EC,$ED,$FF,$04,$3F,$EC,$4E,$DF,$F1,$AB,$FE,$7F,$2F,$C5,$F0
	.dc.b	$F8,$8E,$3F,$01,$CB,$DD,$E7,$EE,$F4,$F7,$7A,$FB,$BD,$BD,$DE,$FE
	.dc.b	$EF,$83,$DD,$F0,$FB,$BE,$2F,$77,$C7,$EE,$F9,$3D,$DF,$2F,$BB,$F4
	.dc.b	$78,$3C,$FE,$EF,$A3,$DD,$FF,$0A,$7C,$E7,$F6,$D1,$F1,$38,$F8,$9C
	.dc.b	$FC,$4E,$BE,$27,$7F,$13,$C3,$E2,$78,$FC,$37,$2F,$84,$E7,$F0,$9D
	.dc.b	$3E,$13,$AF,$C2,$76,$F8,$4E,$FF,$09,$E0,$F8,$4F,$0F,$C2,$78,$BE
	.dc.b	$13,$C7,$F0,$9E,$4F,$84,$F2,$FC,$27,$9B,$E1,$3C,$FF,$03,$FE,$2B
	.dc.b	$F9,$DF,$C3,$F9,$9F,$E3,$75,$EF,$E7,$CC,$D7,$BF,$BF,$7F,$87,$BF
	.dc.b	$C7,$DF,$E5,$EF,$F3,$F7,$FF,$2D,$E9,$75,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$47,$BF,$FE,$43,$F9,$EF,$C4,$74,$BE,$03,$F9,$0A,$87
	.dc.b	$C2,$6F,$E0,$B8,$7C,$9F,$1F,$5B,$97,$2F,$3F,$83,$E9,$F0,$7D,$7E
	.dc.b	$0F,$B7,$C1,$F7,$F8,$3F,$07,$C1,$F8,$7E,$0F,$C5,$F0,$7E,$3F,$83
	.dc.b	$F2,$7C,$1F,$97,$E0,$FC,$DF,$07,$E7,$F8,$3F,$DD,$6B,$FB,$BD,$7F
	.dc.b	$79,$AF,$EF,$75,$FD,$F6,$BE,$DF,$83,$F7,$7C,$1F,$F9,$87,$E0,$3F
	.dc.b	$C7,$47,$E3,$BA,$3F,$CD,$CB,$E0,$74,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$FF,$0D,$E8,$ED,$F0,$3D,$FE,$07,$C1
	.dc.b	$F0,$3F,$C4,$7A,$FF,$89,$75,$FF,$13,$EB,$FE,$29,$D7,$FC,$57,$AF
	.dc.b	$F8,$B7,$5F,$F1,$7E,$6F,$E3,$1C,$DF,$E7,$9F,$C5,$7F,$CA,$47,$A7
	.dc.b	$9C,$FA,$5F,$F4,$43,$FE,$E4,$3F,$D3,$C7,$F8,$38,$FE,$8C,$7F,$45
	.dc.b	$78,$F0,$F5,$E1,$ED,$C1,$DF,$83,$E4,$7C,$17,$C9,$F8,$2F,$95,$F0
	.dc.b	$5F,$2F,$E0,$BE,$67,$C1,$7C,$DF,$82,$F9,$DF,$05,$F3,$FE,$0B,$E8
	.dc.b	$7C,$17,$FA,$47,$E8,$3F,$4F,$C0,$7F,$4E,$FF,$A7,$D7,$C0,$EF,$E0
	.dc.b	$78,$7C,$0F,$1F,$81,$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF
	.dc.b	$03,$DF,$E0,$7C,$1F,$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0
	.dc.b	$3F,$37,$C9,$F9,$DD,$9F,$3F,$B3,$E8,$76,$7D,$1E,$CF,$A5,$D9,$F4
	.dc.b	$FB,$3E,$A7,$67,$D5,$EC,$FA,$DD,$9F,$EA,$FF,$4F,$DE,$7B,$BB,$4F
	.dc.b	$B9,$FF,$5F,$2E,$BE,$1D,$7C,$7A,$F9,$75,$F3,$EB,$E9,$D7,$D7,$AF
	.dc.b	$E6,$7C,$0F,$CD,$F8,$1F,$9D,$F0,$3E,$1E,$EF,$A1,$F0,$3F,$EC,$1F
	.dc.b	$A1,$FD,$F9,$FB,$CE,$0F,$F6,$72,$F7,$F5,$EF,$EF,$DF,$E1,$EF,$F1
	.dc.b	$F7,$F9,$7B,$FC,$FD,$FE,$9C,$FD,$7D,$FF,$81,$E4,$7C,$1F,$23,$E1
	.dc.b	$79,$1F,$0F,$C8,$F8,$9E,$47,$8F,$DF,$F2,$7B,$3E,$5F,$67,$CD,$EC
	.dc.b	$F9,$FD,$9F,$47,$B3,$E9,$F6,$7D,$5E,$CF,$AF,$D9,$F6,$7B,$3F,$ED
	.dc.b	$DF,$A2,$FF,$5B,$1F,$A5,$3F,$49,$D8,$5E,$FE,$FD,$FE,$1E,$FF,$1F
	.dc.b	$7F,$97,$BF,$F6,$7D,$8E,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B
	.dc.b	$9E,$1F,$73,$C5,$EE,$78,$FD,$CF,$27,$B9,$E5,$F7,$3C,$DE,$E7,$9F
	.dc.b	$DC,$F4,$7B,$9F,$87,$F5,$3D,$5E,$E7,$FB,$DF,$CF,$FD,$39,$F8,$93
	.dc.b	$F1,$1D,$27,$F1,$08,$C7,$B7,$C7,$DB,$E5,$ED,$F3,$F6,$FA,$7B,$7D
	.dc.b	$7D,$BE,$DE,$CF,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC
	.dc.b	$9D,$FE,$5E,$2F,$37,$B7,$E7,$F6,$FD,$1E,$DF,$8F,$8D,$F4,$7F,$D3
	.dc.b	$8F,$C8,$38,$F6,$F5,$1E,$B8,$F5,$3F,$EB,$D3,$FC,$EA,$7F,$54,$3F
	.dc.b	$CE,$6E,$BD,$7E,$3E,$BF,$2F,$5F,$F6,$19,$74,$F5,$FA,$FA,$FD,$BD
	.dc.b	$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F
	.dc.b	$AF,$EB,$FA,$4F,$67,$09,$FE,$64,$7E,$38,$FC,$6B,$F4,$D7,$AD,$C7
	.dc.b	$D6,$E5,$EB,$7F,$05,$E8,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07
	.dc.b	$AD,$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$79
	.dc.b	$FD,$6F,$47,$AD,$E9,$F5,$BD,$5E,$B7,$D5,$FA,$5F,$E6,$67,$F1,$53
	.dc.b	$F1,$A7,$E3,$39,$CB,$D6,$E1,$EB,$71,$F5,$BF,$8F,$79,$7C,$FD,$6E
	.dc.b	$9E,$B7,$E0,$7F,$16,$FF,$06,$9F,$80,$3F,$87,$65,$F9,$77,$F0,$09
	.dc.b	$FF,$7E,$1F,$EB,$43,$FC,$26,$7E,$01,$D7,$99,$D7,$CC,$ED,$E6,$77
	.dc.b	$F2,$FC,$1E,$5F,$87,$CB,$F1,$79,$7E,$3F,$2F,$C9,$E5,$F9,$7C,$BF
	.dc.b	$85,$EB,$FC,$3F,$5F,$E2,$7A,$FF,$17,$D7,$F8,$DE,$BF,$C7,$F5,$FE
	.dc.b	$47,$AF,$F2,$7D,$7F,$95,$EB,$FC,$BF,$57,$E6,$79,$DF,$37,$AF,$F4
	.dc.b	$5F,$01,$FE,$2E,$3F,$42,$7C,$2E,$9F,$E8,$1E,$CE,$BD,$FD,$FB,$FC
	.dc.b	$3D,$FE,$3E,$FF,$2E,$EE,$7D,$DD,$3B,$BA,$F7,$76,$EE,$EF,$DD,$E0
	.dc.b	$EE,$F0,$F7,$78,$BB,$BC,$7D,$DE,$4F,$6F,$CB,$ED,$F9,$BD,$BF,$DC
	.dc.b	$7D,$3F,$E7,$8F,$F0,$71,$FC,$A8,$FC,$EF,$29,$FC,$EB,$FC,$45,$7B
	.dc.b	$5C,$BD,$AE,$7E,$D7,$4F,$6B,$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF
	.dc.b	$0F,$B5,$E2,$F6,$BC,$7E,$D7,$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED
	.dc.b	$7A,$3D,$AF,$4F,$B5,$FC,$47,$EA,$3F,$C1,$E7,$F7,$13,$F9,$49,$F9
	.dc.b	$D3,$F3,$9C,$BF,$CA,$57,$B3,$C7,$D9,$E5,$EC,$F3,$F6,$7A,$7B,$1D
	.dc.b	$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C
	.dc.b	$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$F9,$47,$D4,$FF,$9D,$0F,$E4
	.dc.b	$E7,$E1,$CF,$D3,$8F,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB
	.dc.b	$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F
	.dc.b	$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6,$F5,$FF,$9F,$7B,$BF,$C9,$0F,$F2
	.dc.b	$A1,$FD,$C0,$FC,$39,$F8,$6F,$1B,$F9,$1F,$AF,$C3,$D7,$E3,$EB,$F2
	.dc.b	$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5,$FC,$3E
	.dc.b	$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F,$AF,$FD
	.dc.b	$57,$EA,$BF,$CA,$67,$F8,$D4,$FE,$DA,$7F,$21,$3F,$0B,$C8,$7F,$0A
	.dc.b	$FF,$6F,$5E,$B7,$3F,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E
	.dc.b	$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7
	.dc.b	$F5,$BD,$1E,$B7,$F6,$DF,$AB,$FF,$5C,$9F,$DB,$0F,$E1,$87,$F2,$C3
	.dc.b	$F0,$DC,$7F,$E0,$55,$EA,$F1,$F5,$79,$7A,$BC,$FD,$5E,$9E,$AF,$5F
	.dc.b	$57,$B7,$AB,$DF,$D5,$F0,$7A,$BE,$1F,$57,$C5,$EA,$F8,$FD,$5F,$27
	.dc.b	$AB,$E5,$F5,$7C,$DE,$AF,$9F,$D5,$FF,$02,$7D,$67,$F6,$A3,$FD,$34
	.dc.b	$7F,$8D,$8F,$E1,$A7,$E1,$CF,$E1,$7C,$9F,$E1,$D5,$EA,$72,$F5,$39
	.dc.b	$FA,$9D,$3D,$4E,$BE,$A7,$6F,$53,$BF,$A9,$E0,$F5,$3C,$3E,$A7,$8B
	.dc.b	$D4,$F1,$FA,$9E,$4F,$53,$CB,$EA,$79,$BD,$4F,$3F,$A9,$E8,$F5,$3F
	.dc.b	$C3,$9E,$BF,$FA,$68,$FF,$3E,$1F,$E3,$43,$FB,$41,$F8,$83,$F0,$FC
	.dc.b	$9F,$B7,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7,$4F,$53,$AF,$A9,$DB,$D4
	.dc.b	$EF,$EA,$78,$3D,$4F,$0F,$A9,$E2,$F5,$3C,$7E,$A7,$93,$D4,$F2,$FA
	.dc.b	$9E,$6F,$53,$FC,$65,$F5,$BF,$EF,$E3,$FC,$62,$7F,$51,$3F,$B8,$9F
	.dc.b	$86,$E2,$3F,$86,$7F,$C9,$8B,$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD
	.dc.b	$3E,$FE,$9F,$83,$D3,$F0,$FA,$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F
	.dc.b	$A7,$E6,$F4,$FC,$FE,$9F,$A3,$D3,$FF,$25,$7D,$77,$F8,$C0,$FF,$70
	.dc.b	$1F,$E7,$C3,$FA,$99,$FC,$34,$FC,$3F,$19,$FC,$3B,$FE,$65,$7F,$DC
	.dc.b	$03,$D1,$EB,$E8,$F6,$F4,$7B,$FA,$3E,$0F,$47,$C3,$E8,$F8,$BD,$1F
	.dc.b	$1F,$A3,$E4,$F4,$7C,$BE,$8F,$9B,$D1,$F3,$FA,$3E,$8F,$47,$D3,$E8
	.dc.b	$FF,$98,$FE,$BF,$FD,$C2,$7F,$B0,$CF,$F3,$D9,$FE,$2F,$3F,$87,$9F
	.dc.b	$88,$3F,$86,$F1,$7F,$9F,$57,$A1,$CF,$D0,$E9,$E8,$75,$F4,$3B,$7A
	.dc.b	$1D,$FD,$0F,$07,$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F
	.dc.b	$43,$CD,$E8,$79,$FD,$0F,$F3,$D7,$C1,$3E,$5E,$FC,$7E,$3E,$23,$3F
	.dc.b	$07,$AF,$5F,$7E,$DF,$0F,$5B,$8F,$AD,$CB,$D6,$E7,$EB,$74,$F5,$BA
	.dc.b	$FA,$DD,$BD,$6E,$FE,$B7,$83,$D6,$F0,$FA,$DE,$2F,$5B,$C7,$EB,$79
	.dc.b	$3D,$6F,$2F,$AD,$E6,$F5,$BF,$D2,$1F,$61,$FE,$76,$1F,$2B,$8F,$91
	.dc.b	$CF,$C6,$6B,$E0,$F7,$F0,$7C,$3E,$03,$8F,$C0,$72,$F7,$79,$FB,$BD
	.dc.b	$3D,$DE,$BE,$EF,$6F,$77,$BF,$BB,$E0,$F7,$7C,$3E,$EF,$8B,$DD,$F1
	.dc.b	$FB,$BE,$4F,$77,$CB,$EE,$F9,$BD,$DF,$3F,$BB,$FE,$A7,$FB,$17,$CC
	.dc.b	$5F,$96,$C7,$C9,$67,$E3,$35,$E9,$6F,$E2,$B8,$7C,$37,$1F,$82,$E5
	.dc.b	$F0,$5C,$FE,$0B,$A7,$C1,$75,$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05
	.dc.b	$E1,$F8,$2F,$17,$C1,$78,$FE,$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17
	.dc.b	$FA,$F7,$EC,$BF,$09,$F0,$BF,$ED,$17,$FD,$7A,$7F,$97,$4F,$F3,$E9
	.dc.b	$FC,$34,$FE,$18,$9F,$E1,$8F,$1F,$77,$A7,$BB,$D7,$DD,$ED,$EE,$F7
	.dc.b	$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE,$3F,$77,$C9,$EE,$F9,$7D,$DF
	.dc.b	$37,$BB,$E7,$F7,$7D,$1E,$EF,$FB,$57,$D8,$FF,$5D,$9F,$86,$E1,$FF
	.dc.b	$75,$2F,$6F,$5E,$DE,$FD,$BE,$13,$C7,$DB,$E5,$ED,$F3,$F6,$FA,$7B
	.dc.b	$7D,$7D,$BE,$DE,$DF,$7F,$6F,$C1,$ED,$F8,$7D,$BF,$17,$B7,$E3,$F6
	.dc.b	$FC,$9E,$DF,$97,$DB,$F3,$7B,$7E,$7F,$6F,$D1,$ED,$FF,$BB,$7E,$CF
	.dc.b	$F0,$FC,$FF,$F0,$05,$7B,$59,$F6,$B5,$ED,$6F,$DA,$E1,$ED,$71,$F6
	.dc.b	$B9,$7B,$5C,$FD,$AE,$9E,$D7,$5F,$6B,$B7,$B5,$DF,$DA,$F0,$7B,$5E
	.dc.b	$1F,$6B,$C5,$ED,$78,$FD,$AF,$27,$B5,$E5,$F6,$BC,$DE,$D7,$9F,$DA
	.dc.b	$73,$CF,$B4,$FC,$4A,$5F,$0B,$A8,$C7,$C2,$D9,$FF,$63,$9F,$D5,$CF
	.dc.b	$C5,$9F,$8A,$8E,$5E,$E6,$FD,$CE,$1E,$E7,$1F,$73,$97,$B9,$CF,$DC
	.dc.b	$E9,$EE,$75,$F7,$3B,$7B,$9D,$FD,$CF,$07,$B9,$E1,$F7,$3C,$5E,$E7
	.dc.b	$8F,$DC,$F3,$7D,$AF,$9F,$84,$FE,$2D,$BE,$EE,$BD,$DD,$FB,$BC,$3D
	.dc.b	$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B,$FB,$BE
	.dc.b	$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF,$9B,$DD
	.dc.b	$F3,$FB,$BE,$8F,$77,$E7,$7C,$27,$E2,$B9,$FF,$C8,$BE,$EE,$3D,$DC
	.dc.b	$FB,$BA,$F7,$77,$EE,$F0,$F7,$78,$FB,$BC,$BD,$DE,$7E,$EF,$4F,$77
	.dc.b	$AF,$BB,$DB,$DD,$EF,$EE,$F8,$3D,$DF,$0F,$BB,$E2,$F7,$7C,$7E,$EF
	.dc.b	$93,$DD,$F2,$FB,$BE,$6F,$77,$F0,$9F,$6D,$F8,$83,$F0,$FC,$FF,$8C
	.dc.b	$7F,$DD,$83,$E1,$F7,$F0,$FC,$3A,$B8,$FC,$3F,$2F,$7B,$9F,$BD,$D3
	.dc.b	$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$7B,$C3,$EF,$78,$BD,$EF,$1F
	.dc.b	$BD,$E4,$F7,$BC,$BE,$F7,$9B,$DE,$F3,$FB,$DF,$9B,$FB,$7F,$C3,$1F
	.dc.b	$85,$3F,$87,$73,$7E,$91,$FF,$76,$9F,$EA,$73,$FC,$36,$7F,$86,$92
	.dc.b	$F6,$7A,$7B,$3D,$7D,$9E,$DE,$CF,$7F,$67,$C1,$EC,$F8,$7D,$9F,$17
	.dc.b	$B3,$E3,$F6,$7C,$9E,$CF,$97,$D9,$F3,$7B,$3E,$7F,$67,$D1,$EC,$FA
	.dc.b	$7D,$9F,$DA,$7D,$C7,$E1,$CF,$E1,$FC,$7F,$BC,$5E,$CE,$BD,$9D,$FB
	.dc.b	$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B
	.dc.b	$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF
	.dc.b	$9B,$D9,$F3,$FB,$3F,$C2,$F8,$7F,$87,$9F,$88,$E6,$FF,$7F,$FC,$16
	.dc.b	$7D,$9D,$7B,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3
	.dc.b	$D7,$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67
	.dc.b	$C9,$EC,$F9,$7D,$9F,$37,$B3,$FC,$87,$EE,$7F,$86,$1F,$C2,$CF,$C3
	.dc.b	$F3,$17,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE,$FF,$39,$FB,$AF
	.dc.b	$E1,$87,$E1,$B9,$CF,$E1,$9F,$E8,$6B,$DC,$E1,$EE,$71,$F3,$79,$7B
	.dc.b	$9C,$FD,$CE,$9E,$E7,$5F,$73,$B7,$B9,$DF,$DC,$F0,$7B,$9E,$1F,$73
	.dc.b	$C5,$EE,$78,$FD,$CF,$27,$B9,$E5,$F7,$3C,$DE,$E7,$9F,$DC,$F4,$7B
	.dc.b	$9F,$D4,$7E,$EF,$FA,$A9,$F8,$71,$EE,$E7,$DD,$D7,$BB,$BF,$77,$87
	.dc.b	$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B,$BD,$7D,$DE,$DE,$EF,$7F,$77
	.dc.b	$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7,$7C,$9E,$EF,$97,$DD,$F3,$7B
	.dc.b	$BF,$D9,$FD,$9F,$EA,$47,$F5,$13,$F0,$E7,$E1,$BC,$AF,$EE,$2B,$DB
	.dc.b	$E1,$ED,$F1,$8E,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$37
	.dc.b	$B7,$E7,$F6,$FD,$1E,$DF,$F7,$FF,$BC,$FE,$A4,$7E,$1C,$FC,$37,$29
	.dc.b	$FC,$33,$FE,$0A,$5E,$D7,$1F,$6B,$97,$B5,$CF,$DA,$E9,$ED,$75,$F6
	.dc.b	$BB,$7B,$5D,$FD,$AF,$07,$B5,$E1,$F6,$BC,$5E,$D7,$8F,$DA,$F2,$7B
	.dc.b	$5E,$5F,$6B,$CD,$ED,$79,$FD,$AF,$47,$B5,$FE,$17,$FB,$DF,$F0,$D9
	.dc.b	$FC,$34,$FC,$3F,$2F,$F8,$81,$FE,$A8,$DF,$63,$87,$B1,$C7,$D8,$E5
	.dc.b	$EC,$73,$F6,$3A,$7B,$0F,$E0,$18,$5C,$7A,$43,$EF,$BF,$C5,$BF,$04
	.dc.b	$E7,$E3,$18,$6D,$7F,$36,$6C,$EE,$39,$CE,$DA,$6C,$B1,$B5,$AA,$FC
	.dc.b	$51,$B3,$2E,$27,$84,$EE,$BB,$56,$9B,$19,$8E,$4D,$9E,$35,$CE,$FD
	.dc.b	$24,$AF,$2B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD,$F8,$A7,$73,$9A,$E9
	.dc.b	$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57,$72,$47,$9D,$B7,$E4
	.dc.b	$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C,$73,$B6,$9B,$2C,$6D
	.dc.b	$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E,$EB,$C1,$5A,$6C,$66
	.dc.b	$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C,$A3,$83,$A2,$1A,$9E
	.dc.b	$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F,$46,$BC,$4C,$46,$C9
	.dc.b	$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C,$F0,$6C,$ED,$B3,$B8
	.dc.b	$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD,$71,$C4,$D3,$89,$E5
	.dc.b	$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF,$D6,$4A,$F3,$B5,$E8
	.dc.b	$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE,$97,$E3,$6C,$EA,$31
	.dc.b	$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E,$48,$C1,$1C,$12,$FC
	.dc.b	$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6,$DF,$51,$2F,$DA,$4B
	.dc.b	$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6,$63,$93,$67,$8D,$73
	.dc.b	$BF,$59,$2B,$CE,$D7,$A1,$CA,$38,$3A,$21,$A9,$EB,$7E,$29,$DC,$E6
	.dc.b	$BA,$5F,$8D,$B3,$A8,$C4,$F4,$6B,$C4,$C4,$6C,$95,$DC,$91,$E7,$6D
	.dc.b	$F9,$23,$04,$70,$4B,$F1,$CF,$06,$CE,$DB,$3B,$8F,$1C,$ED,$A6,$CB
	.dc.b	$1B,$7D,$44,$BF,$69,$2F,$D7,$1C,$4D,$38,$9E,$53,$BA,$F0,$56,$9B
	.dc.b	$19,$8E,$4D,$9E,$35,$CE,$FD,$64,$AF,$3B,$5E,$87,$28,$E0,$E8,$86
	.dc.b	$A7,$AD,$F8,$A7,$73,$9A,$E9,$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11
	.dc.b	$B2,$57,$72,$47,$9D,$B7,$E4,$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C
	.dc.b	$EE,$3C,$73,$B6,$9B,$2C,$6D,$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2
	.dc.b	$79,$4E,$EB,$C1,$5A,$6C,$66,$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED
	.dc.b	$7A,$1C,$A3,$83,$A2,$1A,$9E,$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A
	.dc.b	$8C,$4F,$46,$BC,$4C,$46,$C9,$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04
	.dc.b	$BF,$1C,$F0,$6C,$ED,$B3,$B8,$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6
	.dc.b	$92,$FD,$71,$C4,$D3,$89,$E5,$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3
	.dc.b	$5C,$EF,$D6,$4A,$F3,$B5,$E8,$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77
	.dc.b	$39,$AE,$97,$E3,$6C,$EA,$31,$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79
	.dc.b	$DB,$7E,$48,$C1,$1C,$12,$FC,$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69
	.dc.b	$B2,$C6,$DF,$51,$2F,$DA,$4B,$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15
	.dc.b	$A6,$C6,$63,$93,$67,$8D,$73,$BF,$59,$2B,$CE,$D7,$A1,$CA,$38,$3A
	.dc.b	$21,$A9,$EB,$7E,$29,$DC,$E6,$BA,$5F,$8D,$B3,$A8,$C4,$F4,$6B,$C4
	.dc.b	$C4,$6C,$95,$DC,$91,$E7,$6D,$F9,$23,$04,$70,$4B,$F1,$CF,$06,$CE
	.dc.b	$DB,$3B,$8F,$1C,$ED,$A6,$CB,$1B,$7D,$44,$BF,$69,$2F,$D7,$1C,$4D
	.dc.b	$38,$9E,$53,$BA,$F0,$56,$9B,$19,$8E,$4D,$9E,$35,$CE,$FD,$64,$AF
	.dc.b	$3B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD,$F8,$A7,$73,$9A,$E9,$7E,$36
	.dc.b	$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57,$72,$47,$9D,$B7,$E4,$8C,$11
	.dc.b	$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C,$73,$B6,$9B,$2C,$6D,$F5,$12
	.dc.b	$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E,$EB,$C1,$5A,$6C,$66,$39,$36
	.dc.b	$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C,$A3,$83,$A2,$1A,$9E,$B7,$E2
	.dc.b	$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F,$46,$BC,$4C,$46,$C9,$5D,$C9
	.dc.b	$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C,$F0,$6C,$ED,$B3,$B8,$F1,$CE
	.dc.b	$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD,$71,$C4,$D3,$89,$E5,$3B,$AF
	.dc.b	$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF,$D6,$4A,$F3,$B5,$E8,$72,$8E
	.dc.b	$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE,$97,$E3,$6C,$EA,$31,$3D,$1A
	.dc.b	$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E,$48,$C1,$1C,$12,$FC,$73,$C1
	.dc.b	$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6,$DF,$51,$2F,$DA,$4B,$F5,$C7
	.dc.b	$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6,$63,$93,$67,$8D,$73,$BF,$59
	.dc.b	$2B,$CE,$D7,$A1,$CA,$38,$3A,$21,$A9,$EB,$7E,$29,$DC,$E6,$BA,$5F
	.dc.b	$8D,$B3,$A8,$C4,$F4,$6B,$C4,$C4,$6C,$95,$DC,$91,$E7,$6D,$F9,$23
	.dc.b	$04,$70,$4B,$F1,$CF,$06,$CE,$DB,$3B,$8F,$1C,$ED,$A6,$CB,$1B,$7D
	.dc.b	$44,$BF,$69,$2F,$D7,$1C,$4D,$38,$9E,$53,$BA,$F0,$56,$9B,$19,$8E
	.dc.b	$4D,$9E,$35,$CE,$FD,$64,$AF,$3B,$5E,$87,$28,$E0,$E8,$86,$A7,$AD
	.dc.b	$F8,$A7,$73,$9A,$E9,$7E,$36,$CE,$A3,$13,$D1,$AF,$13,$11,$B2,$57
	.dc.b	$72,$47,$9D,$B7,$E4,$8C,$11,$C1,$2F,$C7,$3C,$1B,$3B,$6C,$EE,$3C
	.dc.b	$73,$B6,$9B,$2C,$6D,$F5,$12,$FD,$A4,$BF,$5C,$71,$34,$E2,$79,$4E
	.dc.b	$EB,$C1,$5A,$6C,$66,$39,$36,$78,$D7,$3B,$F5,$92,$BC,$ED,$7A,$1C
	.dc.b	$A3,$83,$A2,$1A,$9E,$B7,$E2,$9D,$CE,$6B,$A5,$F8,$DB,$3A,$8C,$4F
	.dc.b	$46,$BC,$4C,$46,$C9,$5D,$C9,$1E,$76,$DF,$92,$30,$47,$04,$BF,$1C
	.dc.b	$F0,$6C,$ED,$B3,$B8,$F1,$CE,$DA,$6C,$B1,$B7,$D4,$4B,$F6,$92,$FD
	.dc.b	$71,$C4,$D3,$89,$E5,$3B,$AF,$05,$69,$B1,$98,$E4,$D9,$E3,$5C,$EF
	.dc.b	$D6,$4A,$F3,$B5,$E8,$72,$8E,$0E,$88,$6A,$7A,$DF,$8A,$77,$39,$AE
	.dc.b	$97,$E3,$6C,$EA,$31,$3D,$1A,$F1,$31,$1B,$25,$77,$24,$79,$DB,$7E
	.dc.b	$48,$C1,$1C,$12,$FC,$73,$C1,$B3,$B6,$CE,$E3,$C7,$3B,$69,$B2,$C6
	.dc.b	$DF,$51,$2F,$DA,$4B,$F5,$C7,$13,$4E,$27,$94,$EE,$BC,$15,$A6,$C6
	.dc.b	$63,$93,$67,$8D,$73,$BF,$59,$28,$29,$2B,$CD,$AF,$77,$BA,$D2,$9D
	.dc.b	$3A,$1F,$71,$DB,$E4,$9F,$19,$99,$C8,$8C,$BC,$67,$84,$ED,$B1,$93
	.dc.b	$84,$65,$A6,$67,$9B,$E8,$9E,$2D,$36,$58,$DB,$F1,$23,$6D,$3A,$FD
	.dc.b	$D3,$DB,$7B,$6F,$90,$F9,$7F,$7D,$6D,$7B,$F7,$71,$39,$8E,$57,$8B
	.dc.b	$FA,$71,$B7,$99,$8C,$1E,$EE,$F1,$EB,$0F,$77,$17,$E5,$AB,$39,$BE
	.dc.b	$77,$C3,$DD,$C9,$AF,$76,$CE,$B0,$7E,$89,$F8,$68,$C3,$6D,$A7,$AE
	.dc.b	$1F,$29,$E2,$75,$19,$F8,$FC,$7C,$56,$5B,$87,$4F,$81,$E0,$71,$F8
	.dc.b	$1B,$1E,$81,$F0,$EF,$82,$F8,$51,$86,$C3,$E0,$AC,$63,$E0,$B3,$F0
	.dc.b	$56,$FA,$5A,$38,$FC,$14,$9C,$FC,$15,$E1,$F7,$C7,$C3,$B7,$C1,$78
	.dc.b	$3E,$0B,$C2,$D9,$EB,$F0,$5E,$43,$CB,$F0,$56,$3C,$3F,$05,$E8,$3D
	.dc.b	$3F,$05,$63,$CB,$F0,$5E,$CF,$82,$F6,$B6,$3D,$3F,$05,$F0,$0F,$83
	.dc.b	$F0,$56,$3E,$47,$D7,$7E,$55,$A7,$C8,$5F,$21,$63,$1E,$4E,$7E,$2E
	.dc.b	$C7,$27,$87,$C2,$71,$F8,$4B,$3D,$9E,$9F,$09,$D7,$E1,$2C,$F8,$9F
	.dc.b	$07,$C2,$48,$F2,$FC,$25,$E8,$F9,$A7,$EA,$F5,$FE,$21,$3E,$AF,$84
	.dc.b	$F4,$7A,$7E,$97,$D5,$F0,$9E,$BF,$84,$B5,$F0,$1F,$77,$C2,$48,$F8
	.dc.b	$7F,$09,$72,$EF,$0C,$9F,$09,$F1,$7E,$13,$E3,$36,$7E,$5F,$C2,$7C
	.dc.b	$93,$E5,$7C,$25,$8F,$9F,$F0,$9F,$94,$FA,$EF,$E3,$91,$83,$F3,$4E
	.dc.b	$32,$F0,$BF,$C9,$23,$3E,$E7,$1F,$17,$93,$CF,$E2,$E4,$F2,$F8,$6B
	.dc.b	$87,$2F,$86,$F0,$1E,$1F,$86,$B5,$E0,$F8,$6F,$21,$E5,$F8,$6B,$1E
	.dc.b	$4F,$83,$F4,$1E,$9F,$8B,$B1,$E8,$F8,$BF,$61,$ED,$F8,$BB,$1E,$CF
	.dc.b	$8B,$F8,$1F,$03,$F0,$5B,$1F,$03,$E1,$3E,$27,$C2,$FC,$56,$C7,$C4
	.dc.b	$F8,$4F,$91,$F0,$BF,$25,$B1,$F2,$3E,$17,$E6,$1F,$37,$E1,$6C,$7F
	.dc.b	$05,$FA,$DF,$E1,$EF,$F1,$76,$F2,$7F,$35,$7F,$8A,$C6,$E3,$F8,$4C
	.dc.b	$3E,$06,$98,$F8,$2B,$5D,$7E,$0B,$7F,$05,$C1,$B1,$E1,$F8,$1E,$67
	.dc.b	$4F,$82,$B1,$E5,$F8,$2E,$E7,$83,$E0,$AC,$7A,$7E,$0B,$C6,$79,$3E
	.dc.b	$0A,$C7,$B7,$E0,$7C,$E7,$A3,$E0,$AC,$7C,$1F,$82,$F5,$9E,$CF,$82
	.dc.b	$B1,$F1,$7E,$0B,$DE,$7C,$0F,$82,$B1,$F2,$7E,$07,$E1,$FC,$17,$C4
	.dc.b	$6C,$7C,$DF,$82,$F8,$E7,$C8,$F8,$2B,$1F,$CD,$7E,$07,$F8,$77,$4F
	.dc.b	$EB,$1F,$E1,$B1,$93,$F9,$BB,$88,$D3,$83,$3F,$03,$67,$A7,$C0,$F0
	.dc.b	$F8,$1E,$2D,$8F,$07,$C0,$F4,$F8,$1E,$AD,$8F,$27,$C0,$F8,$3E,$07
	.dc.b	$C2,$D8,$F4,$7C,$0F,$93,$E0,$7C,$AD,$8F,$67,$C0,$FA,$0F,$4F,$C0
	.dc.b	$D8,$F8,$1F,$03,$EC,$3D,$BF,$03,$63,$E2,$7B,$FF,$00,$F8,$3F,$01
	.dc.b	$63,$E4,$7C,$07,$C4,$F8,$0F,$8A,$D8,$F9,$9F,$01,$FD,$7B,$EB,$7F
	.dc.b	$84,$7C,$27,$F8,$36,$7F,$C6,$13,$9F,$89,$D7,$C0,$FF,$80,$BF,$51
	.dc.b	$7C,$16,$2F,$FB,$E4,$61,$BF,$03,$BF,$81,$E0,$D8,$DF,$C0,$F3,$F8
	.dc.b	$1E,$8D,$8F,$EE,$9A,$FF,$74,$6E,$3E,$7F,$83,$F4,$3D,$23,$C3,$8D
	.dc.b	$F0,$7F,$9F,$FC,$4B,$44,$F6,$AF,$3E,$76,$F7,$48,$F8,$9F,$09,$F1
	.dc.b	$75,$B7,$E4,$7E,$46,$1F,$27,$5B,$3E,$87,$E6,$5F,$CD,$F8,$4B,$3E
	.dc.b	$C7,$E8,$7C,$24,$8F,$7F,$C2,$5D,$3D,$3F,$A9,$7F,$09,$9F,$F1,$42
	.dc.b	$F6,$BF,$C6,$6E,$67,$71,$8F,$66,$57,$0F,$6A,$D7,$4F,$6B,$89,$CB
	.dc.b	$E2,$AD,$76,$F7,$3A,$E3,$D9,$B1,$E2,$C7,$9F,$5F,$F9,$3F,$36,$78
	.dc.b	$AF,$F1,$E7,$43,$E1,$BC,$FE,$E7,$E3,$DB,$3E,$CF,$73,$D7,$EE,$7B
	.dc.b	$1B,$1F,$03,$DC,$F7,$9F,$03,$DC,$B1,$F1,$3D,$DF,$87,$EE,$FC,$46
	.dc.b	$C7,$53,$E4,$E1,$D5,$A7,$52,$72,$7D,$0B,$EA,$FD,$2F,$77,$F4,$DF
	.dc.b	$0B,$F4,$1B,$5D,$67,$EB,$DF,$56,$C7,$4E,$1F,$8D,$9F,$A6,$30,$7F
	.dc.b	$9D,$38,$5E,$F7,$DC,$F7,$BE,$EB,$67,$F0,$3E,$F7,$E0,$3D,$DF,$C0
	.dc.b	$B6,$3F,$0B,$EF,$7E,$13,$DE,$FC,$2B,$63,$F5,$7F,$0D,$FB,$F3,$A6
	.dc.b	$BF,$E1,$6E,$7F,$F4,$84,$F6,$3F,$1F,$8F,$66,$9D,$8F,$C9,$E1,$D9
	.dc.b	$A7,$EE,$3E,$17,$F2,$BF,$01,$FC,$35,$B1,$FE,$81,$D7,$F9,$B7,$07
	.dc.b	$FA,$5D,$B1,$FE,$96,$7F,$D8,$0E,$FE,$27,$FD,$23,$0F,$80,$BE,$FE
	.dc.b	$1B,$5D,$4D,$FC,$07,$06,$9C,$C9,$B7,$F0,$6B,$EE,$DC,$3A,$4F,$9F
	.dc.b	$0B,$3B,$7F,$94,$7C,$3F,$93,$E0,$2D,$7B,$5F,$3F,$BF,$E8,$F8,$0B
	.dc.b	$3F,$05,$F5,$FB,$F2,$3E,$27,$C0,$5C,$BC,$9F,$03,$F0,$3E,$07,$E0
	.dc.b	$B6,$BE,$56,$3D,$5A,$75,$3E,$76,$1C,$F8,$7E,$76,$36,$7E,$76,$1F
	.dc.b	$33,$E0,$7A,$65,$F0,$DF,$3B,$F1,$73,$B8,$7D,$77,$94,$7F,$B4,$B8
	.dc.b	$3F,$CD,$6F,$CB,$C7,$3E,$B2,$5F,$01,$F6,$BE,$03,$ED,$B7,$4F,$C2
	.dc.b	$7C,$07,$DE,$F7,$FE,$FB,$63,$F1,$1F,$01,$F8,$2F,$80,$FC,$1B,$63
	.dc.b	$F1,$98,$7F,$B5,$BF,$53,$FE,$B7,$1F,$8B,$8C,$7C,$0F,$48,$FC,$96
	.dc.b	$37,$6F,$C1,$3D,$23,$F2,$D8,$D9,$FC,$33,$FD,$9F,$E1,$E4,$FF,$B6
	.dc.b	$F1,$FE,$AD,$5F,$AB,$9F,$17,$C0,$4A,$D7,$BF,$6B,$FA,$AD,$BD,$FF
	.dc.b	$47,$BF,$E9,$6D,$F5,$F7,$FD,$9E,$FF,$B5,$B1,$E1,$CB,$A7,$3F,$FC
	.dc.b	$01,$C5,$9E,$2B,$FE,$00,$C7,$77,$D1,$97,$76,$9F,$DC,$3E,$23,$FC
	.dc.b	$08,$7C,$BF,$7E,$C7,$BF,$DF,$F9,$DE,$C7,$CF,$6C,$7C,$3F,$53,$E9
	.dc.b	$1F,$4F,$CD,$B1,$F1,$FC,$9F,$AD,$E2,$FD,$76,$C7,$53,$E6,$61,$D5
	.dc.b	$A7,$52,$71,$FC,$AB,$EA,$FC,$CB,$E6,$FC,$EB,$BB,$FC,$5C,$BC,$F3
	.dc.b	$F4,$9B,$1F,$27,$2E,$6F,$CB,$C2,$E3,$F2,$BE,$1F,$FC,$51,$F1,$1F
	.dc.b	$8F,$6C,$31,$FF,$08,$FE,$A9,$FE,$86,$F0,$0F,$FA,$1A,$32,$27,$86
	.dc.b	$36,$15,$C3,$E0,$71,$39,$8E,$7F,$01,$8B,$F0,$46,$1E,$66,$30,$FF
	.dc.b	$04,$38,$C3,$AC,$7F,$82,$6F,$FC,$81,$E9,$32,$FF,$22,$1E,$CA,$FE
	.dc.b	$8C,$F7,$78,$97,$0F,$36,$1D,$5F,$7C,$67,$0F,$81,$87,$A5,$F0,$57
	.dc.b	$97,$1D,$7C,$0E,$DE,$D0,$FB,$1F,$F5,$7B,$74,$CC,$60,$EB,$1F,$44
	.dc.b	$E1,$3A,$37,$56,$3E,$11,$D7,$E0,$B9,$FC,$17,$46,$C6,$AD,$F0,$7E
	.dc.b	$03,$C3,$F0,$16,$75,$F0,$7E,$43,$CB,$F0,$76,$39,$7C,$07,$A0,$F4
	.dc.b	$FC,$05,$8E,$DF,$05,$EC,$F8,$3F,$6B,$63,$E6,$70,$BE,$0F,$E6,$1F
	.dc.b	$37,$E0,$AC,$EB,$D1,$FA,$1F,$05,$F4,$5B,$1E,$BC,$7E,$4D,$F5,$E7
	.dc.b	$73,$FF,$A0,$78,$BB,$5F,$82,$BA,$D7,$E7,$32,$B7,$F6,$F1,$EB,$3F
	.dc.b	$A0,$6C,$73,$9F,$AE,$7D,$88,$B7,$F3,$72,$FB,$7F,$03,$F7,$1B,$1F
	.dc.b	$47,$E0,$BE,$FF,$C1,$7E,$01,$B1,$F5,$7E,$0B,$F0,$FF,$61,$FE,$C6
	.dc.b	$CF,$D5,$18,$9C,$FC,$0C,$7F,$03,$5F,$E0,$0A,$FF,$50,$3F,$E1,$2D
	.dc.b	$7F,$68,$FF,$83,$E7,$7F,$07,$DB,$E0,$BB,$BE,$0F,$83,$F0,$FC,$1D
	.dc.b	$AE,$AF,$93,$E0,$FC,$BF,$07,$67,$C2,$FA,$3E,$02,$47,$93,$E0,$EE
	.dc.b	$9F,$52,$3F,$39,$A3,$E0,$7D,$E7,$C0,$F8,$3B,$3E,$CF,$83,$F8,$7F
	.dc.b	$13,$F1,$1B,$1F,$03,$E1,$3E,$39,$F2,$3E,$0A,$C7,$C4,$F8,$4F,$97
	.dc.b	$F0,$5F,$31,$B1,$F2,$3E,$13,$E7,$FC,$1F,$D0,$6C,$7E,$B3,$CD,$FE
	.dc.b	$E3,$BF,$F1,$D6,$C7,$FA,$67,$E2,$7F,$7B,$18,$3F,$B4,$F4,$FF,$B4
	.dc.b	$5B,$3F,$BB,$7F,$D9,$95,$F0,$CE,$9F,$13,$CB,$E0,$39,$FC,$25,$BF
	.dc.b	$13,$DB,$E0,$3B,$FC,$1D,$9F,$33,$E2,$F8,$4F,$1F,$C4,$59,$F5,$3E
	.dc.b	$6F,$82,$F3,$FB,$F6,$7D,$CF,$AB,$E0,$A4,$7C,$1F,$80,$BB,$79,$E3
	.dc.b	$FD,$71,$F0,$1F,$03,$E0,$7E,$0B,$63,$E4,$7C,$17,$C4,$F8,$6F,$8A
	.dc.b	$D8,$F9,$9F,$09,$F2,$3E,$13,$E4,$B6,$3E,$87,$C1,$FF,$28,$FA,$FF
	.dc.b	$E4,$FC,$FF,$CE,$E7,$F8,$FC,$7F,$2A,$3F,$93,$E9,$FD,$45,$B0,$7F
	.dc.b	$9A,$C6,$7E,$0F,$A5,$75,$F8,$1B,$7D,$1F,$07,$C1,$48,$EF,$F0,$57
	.dc.b	$0F,$85,$8B,$E0,$BC,$C7,$9F,$E0,$AD,$F9,$BE,$0B,$D5,$F0,$5E,$B6
	.dc.b	$C7,$AB,$E0,$BD,$C7,$BF,$E0,$AC,$7B,$BE,$07,$E1,$1F,$0F,$E0,$2C
	.dc.b	$7C,$2F,$80,$F8,$C7,$C7,$F8,$0B,$1F,$1B,$DF,$F9,$47,$CB,$F8,$0B
	.dc.b	$1F,$2B,$DF,$F9,$C7,$CF,$F8,$0B,$1F,$3B,$E0,$3E,$91,$FD,$6B,$EB
	.dc.b	$FF,$8A,$47,$F5,$5E,$34,$E7,$E0,$75,$F1,$96,$BF,$AA,$DF,$F1,$57
	.dc.b	$FA,$E3,$FC,$4F,$9B,$F1,$8D,$F3,$BF,$C1,$12,$87,$C1,$78,$7E,$0B
	.dc.b	$C5,$F0,$52,$BA,$FC,$15,$C7,$9F,$C1,$79,$FE,$0B,$D0,$DA,$F1,$FC
	.dc.b	$17,$AC,$F6,$7C,$15,$8F,$3F,$C0,$FB,$FE,$07,$E0,$36,$3D,$7F,$01
	.dc.b	$F0,$FE,$07,$E2,$36,$3D,$FF,$03,$F1,$CF,$91,$F0,$36,$3E,$1F,$C0
	.dc.b	$7C,$B3,$E6,$7C,$0D,$8F,$8F,$F0,$1F,$3F,$E0,$7E,$83,$63,$FB,$FF
	.dc.b	$D7,$FF,$12,$1F,$05,$FE,$0E,$70,$DF,$82,$C7,$C5,$5A,$E2,$EF,$E0
	.dc.b	$F8,$7C,$1D,$9E,$AF,$3F,$83,$91,$E0,$F8,$3B,$D3,$FC,$0F,$37,$C1
	.dc.b	$F8,$4F,$17,$C1,$DB,$F3,$FC,$1F,$97,$E0,$FC,$CD,$8F,$5F,$C1,$FA
	.dc.b	$4F,$57,$C1,$D8,$F7,$FC,$17,$B4,$F7,$7C,$1D,$8F,$87,$F0,$5F,$07
	.dc.b	$E0,$FE,$13,$63,$E3,$FC,$17,$C5,$3E,$37,$C1,$D8,$F9,$7F,$07,$F2
	.dc.b	$4F,$F1,$BF,$37,$F8,$DB,$87,$F8,$87,$17,$F2,$18,$FE,$1A,$7F,$0C
	.dc.b	$A3,$8F,$87,$FF,$26,$4F,$E9,$3E,$0A,$FA,$27,$AF,$C1,$E1,$EE,$4C
	.dc.b	$3F,$C7,$78,$3E,$0F,$18,$78,$7E,$0F,$CB,$F0,$7E,$66,$DF,$F9,$8E
	.dc.b	$7F,$67,$F0,$3E,$2F,$89,$FE,$43,$18,$9D,$47,$A7,$17,$C2,$7C,$03
	.dc.b	$E0,$DD,$AF,$1F,$C2,$7C,$4F,$84,$F8,$AD,$8E,$7B,$77,$3D,$FF,$05
	.dc.b	$63,$D7,$F0,$9F,$30,$F9,$BF,$09,$63,$DF,$3F,$43,$E1,$3E,$8B,$63
	.dc.b	$CF,$C5,$EB,$3E,$5F,$C0,$D8,$F8,$FF,$09,$F6,$0F,$B3,$F0,$96,$3E
	.dc.b	$5F,$C2,$7D,$CA,$FB,$AD,$8F,$87,$CD,$F1,$CF,$B3,$FC,$01,$FE,$29
	.dc.b	$E5,$72,$FD,$3A,$FA,$BF,$0B,$F8,$63,$F0,$FD,$36,$3F,$CE,$7F,$61
	.dc.b	$FC,$9A,$FF,$D4,$8F,$F8,$CE,$0E,$5B,$F0,$BF,$5F,$AA,$47,$DB,$C9
	.dc.b	$F0,$18,$C3,$FB,$96,$1F,$CA,$DE,$BF,$0B,$FE,$8E,$8F,$E5,$57,$FE
	.dc.b	$89,$74,$F8,$3E,$0F,$57,$B6,$C7,$E1,$3B,$BF,$10,$7C,$2F,$7E,$C7
	.dc.b	$B3,$E0,$BB,$1D,$FE,$06,$C7,$C0,$F8,$2F,$15,$78,$DB,$1F,$90,$F0
	.dc.b	$7F,$28,$7C,$EF,$7A,$C7,$C8,$F8,$2F,$51,$EB,$F8,$1B,$1F,$EB,$FF
	.dc.b	$8D,$7E,$23,$FE,$2C,$E2,$43,$F1,$27,$C2,$FF,$AF,$63,$7E,$ED,$9C
	.dc.b	$E1,$FA,$2F,$27,$F4,$C7,$E0,$3E,$0A,$C7,$FB,$37,$F0,$67,$E1,$5F
	.dc.b	$E6,$B6,$F6,$7F,$B2,$5C,$F5,$59,$FF,$38,$67,$FC,$D3,$D8,$D7,$E0
	.dc.b	$2F,$13,$C9,$F0,$49,$CC,$7E,$DF,$CE,$9A,$7E,$07,$E1,$2F,$1F,$F9
	.dc.b	$1F,$56,$5C,$EB,$CC,$D9,$D2,$F1,$3A,$CF,$E5,$39,$3F,$0C,$7E,$1F
	.dc.b	$A2,$C7,$F0,$4F,$43,$F8,$41,$FC,$6B,$F8,$07,$E7,$63,$FA,$28,$DB
	.dc.b	$F9,$2B,$FC,$87,$C4,$E3,$E3,$3F,$BC,$F8,$AF,$8A,$FC,$A1,$F9,$5F
	.dc.b	$8A,$B9,$FF,$12,$F5,$3F,$8B,$1D,$FE,$1E,$C7,$EA,$AB,$79,$FE,$83
	.dc.b	$AE,$E5,$DB,$E2,$BF,$40,$7E,$87,$E2,$AE,$1F,$C7,$BD,$7F,$E4,$47
	.dc.b	$AF,$E1,$EC,$7E,$DA,$B7,$8F,$EB,$39,$EE,$3F,$E1,$2F,$B6,$FF,$3F
	.dc.b	$93,$B3,$FD,$F6,$FF,$A1,$2B,$F8,$2F,$B1,$FC,$BC,$F9,$1F,$0F,$72
	.dc.b	$FF,$7A,$D7,$2E,$FE,$3E,$2F,$27,$9F,$B9,$2B,$7F,$17,$72,$FF,$43
	.dc.b	$1F,$C4,$BD,$CF,$E7,$87,$D5,$F8,$6B,$5D,$AF,$86,$DE,$9E,$6B,$2C
	.dc.b	$0B,$47,$C5,$FB,$0F,$6F,$C5,$D9,$FE,$8F,$EF,$7F,$4B,$3F,$09,$F0
	.dc.b	$F6,$3D,$B5,$BC,$3F,$D0,$D8,$7F,$81,$27,$27,$C2,$F8,$BF,$92,$7C
	.dc.b	$AF,$8B,$B1,$FD,$53,$E0,$7F,$AC,$1F,$91,$F8,$6B,$1C,$49,$87,$6E
	.dc.b	$5E,$2F,$83,$CC,$FA,$A7,$D6,$F8,$BB,$5F,$D8,$3E,$0F,$FB,$21,$FD
	.dc.b	$DB,$E7,$BF,$B2,$FC,$1A,$37,$87,$3A,$FB,$3C,$D8,$E5,$FC,$27,$C6
	.dc.b	$7E,$08,$FC,$1F,$C6,$5A,$FE,$DB,$E9,$7F,$71,$3B,$7C,$35,$8F,$C1
	.dc.b	$D6,$F1,$F8,$FC,$D7,$1F,$C8,$7C,$67,$E3,$8F,$C7,$FC,$65,$C3,$FB
	.dc.b	$CF,$A5,$FD,$F4,$F5,$7B,$D6,$3F,$1F,$5B,$C7,$E9,$F2,$DC,$7F,$CD
	.dc.b	$BF,$6D,$F8,$3B,$FC,$24,$62,$BF,$0A,$7F,$AF,$E7,$FD,$56,$6D,$FC
	.dc.b	$9F,$17,$FA,$02,$BF,$C1,$7F,$13,$F2,$FD,$3F,$F0,$83,$CB,$0E,$93
	.dc.b	$D8,$F0,$61,$CE,$3C,$7F,$19,$B9,$CE,$34,$FC,$BD,$3E,$33,$12,$8C
	.dc.b	$3D,$67,$7F,$19,$A9,$CC,$7E,$AB,$E4,$FF,$BA,$26,$F4,$BC,$AF,$F1
	.dc.b	$CC,$7F,$75,$AE,$39,$FC,$8C,$BA,$73,$75,$7B,$7C,$6C,$D3,$C4,$79
	.dc.b	$7E,$2B,$1A,$78,$7E,$07,$A9,$F9,$1F,$06,$D7,$0C,$D8,$70,$CF,$EB
	.dc.b	$F3,$5A,$F9,$3F,$05,$ED,$3D,$DF,$01,$63,$FC,$85,$BF,$F8,$17,$4F
	.dc.b	$77,$C1,$7E,$7F,$AB,$DA,$EE,$3E,$F6,$1F,$1F,$9F,$19,$FE,$1B,$D9
	.dc.b	$F9,$47,$CB,$F5,$AE,$7D,$BE,$3F,$E3,$9F,$5F,$E7,$BF,$87,$71,$E2
	.dc.b	$7D,$97,$C3,$3F,$C6,$73,$DC,$BF,$27,$E8,$FD,$73,$EC,$79,$D7,$0F
	.dc.b	$5F,$C6,$FD,$43,$B6,$1F,$EC,$48,$C9,$F2,$6B,$84,$7E,$67,$9E,$E1
	.dc.b	$FA,$0F,$23,$F0,$87,$E1,$7C,$6B,$87,$C8,$F8,$EF,$C0,$9D,$BE,$36
	.dc.b	$C7,$D5,$AD,$E3,$FA,$5E,$6B,$8F,$EB,$3C,$2F,$C8,$1F,$96,$E3,$FC
	.dc.b	$53,$F8,$CF,$8E,$98,$FE,$5F,$93,$F9,$A7,$C6,$63,$2F,$D8,$5F,$2A
	.dc.b	$FC,$5E,$1F,$E4,$9E,$EF,$D8,$53,$2F,$F7,$3C,$FF,$8F,$88,$78,$7A
	.dc.b	$BF,$37,$B4,$C3,$C1,$6F,$86,$C3,$CE,$3F,$82,$5F,$3B,$F5,$72,$73
	.dc.b	$9E,$3F,$01,$F4,$8F,$F1,$77,$C5,$BD,$7C,$4F,$E4,$BE,$37,$F5,$67
	.dc.b	$D8,$F8,$4B,$1C,$89,$9F,$F8,$77,$E2,$FF,$CB,$F1,$8D,$FD,$BC,$FF
	.dc.b	$2F,$97,$E6,$36,$7D,$7F,$09,$F8,$43,$F0,$BF,$05,$63,$E7,$73,$FF
	.dc.b	$B9,$9B,$F0,$1F,$97,$8C,$67,$F0,$43,$23,$FC,$91,$A4,$AF,$F2,$96
	.dc.b	$EF,$67,$18,$7F,$0D,$F8,$EF,$E1,$27,$5F,$80,$B7,$E2,$F8,$1B,$5D
	.dc.b	$F3,$EF,$18,$E0,$FB,$7D,$3F,$89,$3F,$15,$C7,$70,$FD,$EF,$67,$F1
	.dc.b	$63,$D5,$F0,$16,$32,$4C,$BB,$65,$F5,$63,$18,$FE,$23,$4F,$CB,$1F
	.dc.b	$97,$C2,$CF,$F0,$AF,$8E,$FE,$46,$7C,$7F,$80,$B1,$92,$63,$DB,$E0
	.dc.b	$32,$FC,$B8,$76,$AF,$F4,$87,$E9,$7E,$2A,$CF,$F1,$87,$F9,$89,$F5
	.dc.b	$7E,$06,$C6,$49,$7E,$CF,$81,$CB,$DF,$0F,$CE,$4F,$EA,$3C,$1F,$D9
	.dc.b	$7C,$57,$EC,$DF,$E7,$A4,$8F,$C2,$FC,$05,$BF,$E8,$D3,$78,$3F,$B1
	.dc.b	$F0,$FE,$13,$E0,$71,$87,$E9,$BE,$27,$F7,$C7,$EF,$FA,$EC,$7F,$32
	.dc.b	$CF,$FA,$59,$F8,$FF,$80,$B1,$92,$67,$F5,$BE,$07,$2F,$AF,$C5,$FE
	.dc.b	$14,$7F,$0B,$F8,$8B,$5F,$CF,$B2,$FE,$AE,$7F,$31,$D7,$FB,$64,$7F
	.dc.b	$8F,$EF,$FA,$F4,$EA,$38,$1F,$C4,$65,$F2,$79,$AF,$95,$E1,$FF,$19
	.dc.b	$3F,$8D,$78,$76,$FF,$A6,$ED,$FD,$98,$F0,$7B,$B6,$32,$4B,$FC,$F6
	.dc.b	$2F,$80,$D4,$7E,$17,$ED,$7F,$D7,$35,$FD,$A1,$B1,$9F,$8A,$FE,$FF
	.dc.b	$1F,$E6,$FF,$71,$C1,$FD,$D8,$F8,$1E,$FD,$C3,$FB,$0C,$5C,$3C,$3F
	.dc.b	$01,$73,$FF,$0E,$FE,$10,$FB,$57,$0F,$F8,$3E,$30,$D8,$70,$D8,$FF
	.dc.b	$02,$F2,$4C,$FE,$57,$C0,$5C,$FF,$C1,$D3,$72,$FC,$48,$F8,$1C,$5F
	.dc.b	$F9,$CF,$AF,$F9,$D7,$C0,$FD,$93,$F9,$5F,$75,$9F,$EF,$79,$FF,$85
	.dc.b	$0F,$BD,$EF,$D8,$C9,$33,$F9,$FF,$03,$97,$E8,$FC,$07,$F8,$73,$7F
	.dc.b	$C3,$36,$BF,$C0,$F1,$67,$EB,$FB,$F7,$6F,$F2,$1C,$7F,$89,$E7,$15
	.dc.b	$E9,$97,$C0,$62,$BF,$D1,$DF,$2F,$FE,$A5,$3E,$DC,$FC,$2F,$81,$FF
	.dc.b	$1E,$4A,$75,$EB,$61,$DE,$1D,$7E,$03,$13,$23,$FC,$7D,$5C,$63,$F4
	.dc.b	$1F,$03,$98,$FF,$53,$FD,$CF,$FA,$72,$B6,$D8,$FC,$9F,$C1,$78,$A3
	.dc.b	$19,$FD,$6F,$82,$FF,$0D,$7C,$9F,$FA,$9B,$0F,$D3,$46,$1C,$FC,$26
	.dc.b	$BE,$0E,$C7,$F7,$BA,$FF,$18,$6D,$FD,$33,$0F,$92,$D9,$FC,$04,$FF
	.dc.b	$24,$7D,$71,$9C,$FE,$4F,$89,$F0,$CF,$1F,$C1,$DA,$DC,$59,$FE,$95
	.dc.b	$7F,$DA,$1F,$F6,$3D,$F3,$7E,$77,$C7,$FC,$D3,$DB,$F0,$57,$0F,$0C
	.dc.b	$59,$E3,$85,$CB,$D3,$36,$BD,$3E,$ED,$BF,$BF,$EF,$FD,$93,$EE,$FE
	.dc.b	$CF,$E3,$C3,$E0,$C6,$6F,$DE,$F1,$AF,$C1,$39,$7D,$2F,$D2,$F7,$F1
	.dc.b	$5F,$2B,$E0,$3F,$0C,$7D,$6F,$7E,$E1,$F2,$A2,$E1,$F1,$3D,$FB,$7C
	.dc.b	$6B,$47,$D7,$F7,$72,$FF,$27,$E0,$7C,$87,$E4,$CF,$5E,$D6,$7E,$BC
	.dc.b	$5C,$7E,$7E,$57,$1F,$BF,$36,$BE,$FF,$BD,$6F,$F4,$BF,$03,$F9,$F3
	.dc.b	$0F,$E8,$DB,$93,$E0,$3F,$D6,$7F,$46,$E1,$F3,$65,$FA,$79,$CC,$63
	.dc.b	$2F,$C5,$FC,$3F,$23,$F5,$84,$7F,$2D,$EF,$DC,$FF,$0F,$87,$39,$43
	.dc.b	$1F,$03,$FB,$C9,$D3,$9C,$3E,$46,$1C,$F5,$F8,$38,$78,$8F,$CE,$FC
	.dc.b	$97,$EF,$4F,$D6,$FB,$F6,$3F,$43,$A7,$AB,$1F,$E1,$2D,$7F,$83,$65
	.dc.b	$E0,$CF,$C7,$1F,$C2,$70,$E4,$FE,$8B,$DD,$B9,$7E,$C3,$E1,$3F,$86
	.dc.b	$1F,$94,$F7,$6E,$7E,$03,$DF,$8F,$4E,$0F,$A1,$18,$DF,$E6,$65,$FA
	.dc.b	$CD,$7C,$95,$A8,$C5,$7B,$8E,$8E,$09,$97,$8E,$32,$FF,$04,$AE,$67
	.dc.b	$4D,$7D,$D8,$7E,$4E,$3E,$7C,$FF,$0A,$D6,$DF,$E5,$FE,$17,$F9,$41
	.dc.b	$FB,$0F,$66,$C7,$F0,$C8,$F3,$9D,$9F,$66,$5F,$CD,$A3,$31,$B3,$F8
	.dc.b	$D6,$BC,$A7,$F8,$5F,$C2,$FF,$35,$3F,$92,$DD,$AF,$E3,$31,$F1,$0F
	.dc.b	$B5,$EA,$3C,$17,$AD,$9A,$F0,$E1,$FD,$06,$F9,$3F,$C6,$BD,$3B,$AF
	.dc.b	$F2,$8F,$85,$FE,$88,$7F,$34,$F4,$2E,$1F,$CA,$23,$E9,$9D,$6F,$F2
	.dc.b	$87,$F5,$27,$31,$B3,$FD,$61,$E6,$FF,$55,$8F,$EA,$1F,$23,$27,$3E
	.dc.b	$35,$BF,$E5,$73,$F8,$03,$9B,$D8,$FE,$B4,$E6,$7F,$9D,$BF,$EE,$3B
	.dc.b	$78,$DC,$F2,$FE,$87,$EB,$7F,$5F,$35,$CD,$70,$FE,$87,$1F,$8E,$3E
	.dc.b	$5D,$BC,$1C,$4F,$CA,$E9,$57,$1C,$FF,$D1,$97,$FC,$6A,$B9,$9F,$D3
	.dc.b	$BE,$17,$FB,$71,$FD,$CB,$F6,$8F,$EE,$CE,$23,$A3,$DA,$BD,$D8,$23
	.dc.b	$F4,$26,$E3,$BF,$85,$74,$FE,$BF,$F2,$5F,$DE,$CC,$61,$FC,$B5,$E1
	.dc.b	$DD,$FB,$AF,$9B,$FF,$42,$FC,$37,$82,$BF,$0B,$18,$C7,$C3,$97,$F8
	.dc.b	$17,$E2,$39,$C7,$F8,$39,$C5,$7F,$6A,$7E,$87,$8B,$89,$FE,$C7,$5F
	.dc.b	$C2,$EF,$A6,$3F,$BF,$8C,$7C,$4F,$F9,$8A,$BD,$F8,$71,$D3,$FB,$CD
	.dc.b	$5C,$BF,$CD,$9E,$A7,$FA,$AE,$FF,$C1,$F9,$FD,$FF,$0B,$FC,$18,$FF
	.dc.b	$87,$B4,$EF,$CB,$FE,$73,$6D,$7F,$89,$B1,$F0,$ED,$FB,$4E,$F7,$23
	.dc.b	$C5,$FC,$51,$F8,$BF,$12,$ED,$FE,$1C,$D7,$FC,$34,$7E,$3F,$C4,$B1
	.dc.b	$FE,$26,$D3,$BE,$86,$F3,$7D,$B9,$F9,$FC,$7F,$07,$93,$F9,$FE,$BE
	.dc.b	$2F,$03,$F3,$5D,$DF,$9B,$6D,$FF,$8B,$36,$FF,$15,$1F,$9F,$ED,$B1
	.dc.b	$E1,$C7,$A7,$1F,$C1,$6E,$BF,$06,$7C,$3C,$5F,$07,$8B,$F3,$71,$FB
	.dc.b	$6E,$BF,$D5,$72,$7E,$AD,$FF,$1C,$72,$49,$FD,$DF,$4D,$C7,$FD,$F4
	.dc.b	$7B,$F2,$EE,$DC,$BF,$1C,$FE,$E7,$8F,$E9,$F0,$FD,$0E,$9F,$DC,$1F
	.dc.b	$B9,$F8,$9B,$97,$F9,$2B,$9B,$FC,$90,$7E,$F7,$E4,$EC,$7B,$F2,$F5
	.dc.b	$E5,$F9,$88,$BC,$DF,$33,$E1,$3F,$D6,$B2,$CD,$18,$6D,$AF,$FB,$9F
	.dc.b	$67,$FB,$4A,$B8,$DF,$F9,$A7,$3F,$F3,$41,$B5,$3B,$F9,$2C,$DF,$96
	.dc.b	$7E,$06,$FF,$13,$AB,$F0,$51,$8C,$BF,$86,$63,$C2,$39,$DF,$FA,$5F
	.dc.b	$27,$C5,$E7,$3F,$F2,$EE,$DF,$E7,$93,$C4,$F9,$FE,$2F,$17,$E1,$DF
	.dc.b	$F8,$06,$BE,$1E,$6F,$D0,$E7,$FC,$3B,$1D,$BF,$E5,$AD,$2E,$5F,$E9
	.dc.b	$2C,$FE,$09,$F0,$BC,$1F,$D9,$46,$6B,$F6,$18,$FE,$03,$93,$F3,$F9
	.dc.b	$76,$CF,$FC,$13,$8F,$2B,$EB,$A7,$FB,$9E,$3A,$EF,$FE,$96,$7F,$DC
	.dc.b	$B1,$F4,$8F,$85,$F2,$39,$9E,$E7,$DE,$EA,$FF,$25,$63,$DD,$E9,$1B
	.dc.b	$AE,$58,$7E,$07,$85,$F1,$59,$8F,$CC,$E4,$E4,$F1,$1F,$CD,$FB,$6E
	.dc.b	$1F,$2F,$2F,$D6,$65,$F9,$CD,$7E,$6B,$FD,$BB,$0E,$55,$FE,$A8,$6E
	.dc.b	$3D,$A7,$9B,$E0,$8F,$80,$63,$83,$E8,$3E,$E8,$FC,$31,$F9,$D2,$DF
	.dc.b	$CC,$CB,$E4,$6D,$FE,$39,$AE,$F7,$F6,$63,$1C,$9F,$34,$FE,$A5,$C7
	.dc.b	$73,$F6,$65,$FE,$52,$E1,$FB,$31,$F3,$63,$FC,$AB,$8B,$0E,$C7,$7C
	.dc.b	$BC,$35,$F0,$62,$D7,$CC,$C6,$E5,$FE,$DC,$C7,$F9,$29,$DA,$7D,$7A
	.dc.b	$FE,$0E,$FA,$E1,$F2,$F1,$E7,$A7,$E2,$B7,$FC,$31,$FD,$AF,$92,$D7
	.dc.b	$AB,$1F,$E2,$07,$C7,$AF,$AA,$FF,$82,$75,$FC,$0C,$FD,$7B,$EF,$39
	.dc.b	$7E,$57,$C5,$7E,$4C,$FD,$6C,$FE,$CD,$C3,$DF,$4F,$56,$1D,$8F,$CF
	.dc.b	$CF,$29,$FE,$9D,$C4,$F9,$1C,$63,$E3,$7F,$40,$76,$7F,$2D,$F1,$58
	.dc.b	$AE,$DC,$DF,$B9,$7E,$A4,$FF,$04,$AF,$C0,$1D,$31,$FD,$7B,$F8,$3B
	.dc.b	$FD,$61,$D7,$A7,$F9,$0F,$C5,$63,$0E,$BC,$DF,$9B,$E3,$FC,$EE,$5F
	.dc.b	$1F,$0E,$55,$F9,$DC,$6C,$FF,$8C,$FC,$67,$87,$FE,$FD,$3B,$6C,$F8
	.dc.b	$AC,$61,$DA,$FE,$A6,$5D,$B6,$FE,$6D,$F1,$5F,$E5,$AC,$0D,$F8,$AC
	.dc.b	$3F,$9D,$DF,$F8,$31,$DB,$E2,$AE,$5D,$70,$FF,$07,$3F,$1B,$99,$5C
	.dc.b	$63,$E9,$FC,$55,$D3,$F0,$9A,$FF,$11,$3E,$92,$FD,$A7,$F9,$8D,$7E
	.dc.b	$EE,$7F,$69,$8F,$DF,$DB,$F8,$8C,$F7,$9F,$D9,$1D,$67,$F9,$0B,$FB
	.dc.b	$5D,$3F,$8F,$1D,$E7,$F8,$33,$D2,$BE,$21,$D3,$0F,$D2,$9F,$B2,$AF
	.dc.b	$E3,$CD,$AF,$C1,$F0,$3C,$AF,$4B,$63,$F9,$0F,$2F,$F6,$0F,$8B,$CD
	.dc.b	$77,$BF,$E0,$1C,$FF,$8D,$7B,$9F,$AA,$D3,$DC,$F7,$33,$8B,$CB,$EE
	.dc.b	$7E,$03,$E2,$EE,$7F,$4F,$5F,$56,$1C,$B0,$E9,$1E,$E7,$A4,$67,$0F
	.dc.b	$F0,$9C,$7F,$83,$1B,$E0,$7F,$73,$F8,$BF,$6F,$99,$E9,$AE,$67,$F8
	.dc.b	$74,$F4,$F7,$7F,$0E,$D7,$FC,$27,$CD,$F9,$3D,$3E,$3F,$8D,$F1,$8F
	.dc.b	$CB,$FC,$5D,$8F,$46,$1C,$F5,$F4,$63,$CF,$BF,$F9,$F7,$2F,$D2,$A6
	.dc.b	$EF,$33,$E8,$9F,$A2,$F8,$BB,$97,$4A,$79,$7F,$E0,$2C,$FA,$EF,$FD
	.dc.b	$A4,$F3,$E5,$D7,$27,$9B,$F7,$0F,$D6,$FC,$5D,$CB,$A6,$3C,$F9,$7E
	.dc.b	$66,$7F,$10,$9C,$5F,$05,$FB,$3E,$5F,$E1,$8F,$E0,$9F,$85,$FF,$88
	.dc.b	$70,$FE,$CF,$3E,$C9,$F9,$DC,$5F,$C7,$B0,$EA,$FC,$39,$FD,$4D,$7E
	.dc.b	$4B,$C9,$FC,$89,$D3,$E3,$2C,$79,$B9,$BE,$45,$F1,$E2,$FD,$2C,$FF
	.dc.b	$1C,$FB,$DF,$F1,$79,$FE,$35,$D1,$F1,$BF,$B4,$8F,$F1,$9D,$F1,$D3
	.dc.b	$F1,$A7,$93,$93,$E5,$73,$FD,$6C,$7A,$3C,$4F,$7C,$58,$F2,$1F,$CB
	.dc.b	$2B,$A6,$DF,$94,$BF,$97,$87,$F6,$CA,$4C,$8C,$3F,$E4,$DC,$9F,$1B
	.dc.b	$E7,$F3,$BE,$4E,$5F,$AD,$D7,$EF,$E9,$FC,$1F,$2F,$81,$C5,$ED,$3E
	.dc.b	$9F,$C9,$58,$F2,$E7,$F9,$CC,$3D,$AF,$D0,$3F,$6F,$A3,$D1,$F8,$C7
	.dc.b	$DC,$F8,$EB,$7C,$B3,$F0,$DF,$EF,$B2,$FC,$24,$7E,$D7,$0F,$EF,$D5
	.dc.b	$F8,$26,$D7,$F9,$50,$F7,$7C,$4F,$D3,$F4,$FD,$BA,$FF,$49,$9F,$E1
	.dc.b	$6F,$CB,$D3,$F3,$78,$7D,$DF,$4F,$ED,$1F,$90,$F8,$DB,$1E,$5E,$3F
	.dc.b	$B5,$7F,$17,$0F,$6F,$A9,$F8,$13,$F2,$FF,$1B,$6F,$84,$F8,$0F,$8D
	.dc.b	$5F,$81,$7E,$2F,$A9,$F8,$93,$F3,$FF,$1B,$6B,$8C,$FE,$9F,$A3,$F0
	.dc.b	$2F,$F9,$C3,$C2,$FD,$F7,$0B,$D6,$FC,$91,$FA,$8F,$8D,$B8,$72,$C3
	.dc.b	$89,$FE,$C2,$E5,$F9,$BB,$FE,$2B,$D3,$FC,$D1,$FB,$0F,$8D,$B7,$C7
	.dc.b	$93,$EE,$F3,$FF,$27,$C3,$8E,$DE,$D8,$63,$F9,$28,$6B,$FB,$DF,$33
	.dc.b	$FD,$29,$18,$DF,$F7,$BF,$1B,$FA,$97,$EB,$C7,$F7,$23,$A5,$FD,$07
	.dc.b	$F1,$98,$78,$7D,$7F,$F3,$A1,$FD,$93,$D5,$7A,$78,$DB,$D1,$CD,$F7
	.dc.b	$CF,$A1,$7E,$9F,$5F,$FC,$F8,$6F,$1F,$F3,$7B,$E2,$8F,$EE,$F1,$F6
	.dc.b	$0F,$CA,$67,$FC,$B7,$77,$AF,$FE,$85,$39,$E3,$FE,$6D,$79,$6B,$FC
	.dc.b	$0A,$FD,$64,$3F,$90,$DF,$3C,$FF,$D1,$C7,$F7,$EF,$D0,$FF,$AD,$61
	.dc.b	$FE,$77,$8E,$75,$FD,$06,$FE,$26,$5F,$4F,$0F,$27,$A9,$FE,$97,$39
	.dc.b	$E1,$FE,$7F,$78,$DF,$F3,$EF,$23,$F8,$FF,$27,$DC,$C7,$FD,$41,$87
	.dc.b	$B7,$D4,$FF,$3C,$1F,$E7,$9C,$3F,$D4,$4F,$0E,$0F,$D6,$63,$F9,$5C
	.dc.b	$39,$3F,$C5,$4F,$76,$EF,$4F,$FD,$04,$74,$D3,$FD,$28,$F3,$9F,$C0
	.dc.b	$1C,$EF,$F5,$DA,$FE,$3F,$D2,$FF,$45,$1F,$ED,$D7,$F1,$F3,$FC,$83
	.dc.b	$D1,$76,$6B,$1F,$C0,$4F,$4E,$CF,$D0,$E9,$FC,$02,$3F,$87,$7A,$3F
	.dc.b	$E9,$53,$83,$FE,$73,$7C,$98,$FE,$08,$FE,$2A,$7C,$FB,$FE,$68,$78
	.dc.b	$F8,$7F,$41,$87,$F5,$7B,$E5,$3F,$CD,$2B,$CF,$09,$D4,$74,$AF,$C3
	.dc.b	$4F,$C1,$3F,$1F,$5F,$D0,$7C,$EE,$D1,$FE,$A3,$6B,$FD,$CA,$2D,$A7
	.dc.b	$D3,$5D,$23,$E0,$EB,$FA,$28,$FE,$CF,$E7,$70,$38,$D5,$BF,$93,$8F
	.dc.b	$C1,$AF,$C7,$9F,$DE,$B8,$3F,$80,$E5,$E7,$3F,$80,$1F,$E9,$E2,$E4
	.dc.b	$D1,$9E,$A7,$F9,$21,$A8,$CB,$99,$E1,$3B,$8E,$87,$FA,$08,$E0,$E6
	.dc.b	$37,$1C,$67,$9B,$A3,$A7,$9A,$F7,$F8,$36,$33,$18,$23,$82,$1E,$C8
	.dc.b	$E0,$D8,$DB,$4D,$9F,$84,$F3,$7F,$92,$67,$F8,$2C,$3F,$B9,$D7,$E1
	.dc.b	$73,$FC,$4E,$1E,$3A,$D5,$F0,$9C,$D6,$F0,$FF,$72,$5E,$E7,$30,$BF
	.dc.b	$C7,$FC,$FF,$F9,$C6,$38,$1F,$91,$AE,$0F,$2B,$CE,$3E,$48,$E5,$87
	.dc.b	$E7,$3E,$27,$C5,$1F,$EF,$3F,$3A,$BF,$9B,$CE,$D9,$B3,$8C,$7B,$D7
	.dc.b	$87,$1E,$F1,$8C,$BC,$D3,$E0,$3C,$85,$38,$17,$0F,$36,$7F,$2F,$2F
	.dc.b	$0F,$9D,$E4,$FD,$AB,$EA,$CF,$EC,$1B,$AF,$ED,$A7,$C9,$9F,$88,$98
	.dc.b	$F7,$CB,$C7,$87,$D7,$CF,$D3,$9F,$83,$1F,$9F,$3E,$1B,$F9,$3E,$6F
	.dc.b	$89,$F9,$79,$7C,$48,$EF,$1E,$9C,$39,$9F,$07,$DA,$CD,$F7,$CF,$C6
	.dc.b	$F5,$AF,$81,$E6,$F3,$7B,$9F,$61,$E3,$97,$90,$91,$FE,$44,$D3,$D7
	.dc.b	$18,$8F,$97,$5D,$DF,$99,$B7,$DA,$F3,$3E,$57,$BF,$27,$F0,$B1,$D9
	.dc.b	$FA,$B5,$C4,$FC,$7E,$79,$8E,$B8,$77,$DD,$8F,$E1,$BA,$BB,$57,$43
	.dc.b	$F2,$2F,$19,$F9,$D1,$DA,$BF,$C1,$99,$FA,$A3,$04,$3B,$9F,$E6,$29
	.dc.b	$EE,$FE,$73,$CB,$FF,$0A,$79,$5F,$E1,$29,$FE,$8F,$3E,$03,$47,$FA
	.dc.b	$39,$F0,$BD,$09,$5E,$17,$11,$A2,$63,$FA,$2F,$2B,$BF,$57,$F4,$4C
	.dc.b	$6D,$F9,$5A,$FF,$43,$9B,$7C,$F8,$BF,$0B,$C5,$FB,$8C,$BC,$13,$98
	.dc.b	$FC,$2E,$9F,$B5,$C3,$F7,$3F,$21,$D3,$C7,$7A,$F4,$EB,$1D,$B2,$EB
	.dc.b	$5F,$93,$C3,$B7,$0F,$D1,$9C,$F1,$7E,$44,$FE,$31,$F2,$1C,$AF,$E3
	.dc.b	$61,$E8,$8E,$98,$79,$27,$EA,$6B,$E8,$C3,$9E,$9F,$83,$E1,$FE,$0F
	.dc.b	$F1,$FC,$6F,$E8,$63,$F0,$A3,$9C,$79,$49,$9F,$E3,$9E,$C7,$D2,$C3
	.dc.b	$F6,$B7,$FC,$53,$CF,$7C,$07,$D0,$F8,$07,$5B,$D2,$C3,$71,$FF,$46
	.dc.b	$57,$83,$2F,$CE,$6B,$FA,$7D,$BD,$5B,$F8,$39,$3E,$FE,$1F,$CB,$34
	.dc.b	$7A,$5D,$3C,$D9,$3F,$E9,$78,$E9,$8F,$E8,$F8,$3F,$80,$4F,$37,$E2
	.dc.b	$BF,$99,$E0,$F5,$7A,$3E,$6D,$A4,$FF,$A8,$63,$9B,$FA,$1C,$FF,$2F
	.dc.b	$AF,$9A,$3E,$3E,$3F,$3D,$F3,$7A,$3F,$03,$39,$1F,$EA,$B8,$E6,$FC
	.dc.b	$BD,$BF,$06,$F9,$5F,$81,$9F,$CD,$F4,$7E,$4E,$72,$BF,$D6,$91,$CB
	.dc.b	$3F,$CB,$47,$F2,$BE,$7F,$C6,$61,$CA,$3E,$47,$A1,$F4,$B2,$91,$FE
	.dc.b	$BD,$8E,$5B,$FC,$48,$6D,$ED,$E2,$FA,$B1,$F0,$75,$FC,$07,$A1,$F7
	.dc.b	$35,$93,$FE,$C7,$8E,$79,$FE,$23,$0E,$5C,$1F,$4E,$3F,$B0,$E1,$F8
	.dc.b	$9F,$43,$F0,$BA,$4C,$7F,$DA,$11,$C9,$F9,$A7,$F2,$DD,$BD,$AE,$0F
	.dc.b	$F1,$F5,$7E,$4F,$D0,$FC,$7E,$72,$7F,$DB,$51,$CF,$5F,$5C,$7B,$AF
	.dc.b	$9C,$63,$83,$F8,$DC,$FF,$63,$E5,$79,$FF,$9A,$CE,$4F,$FB,$96,$3A
	.dc.b	$6D,$F9,$9C,$79,$BF,$83,$9F,$ED,$9E,$2F,$FA,$D3,$C3,$78,$69,$76
	.dc.b	$7E,$A3,$6F,$D3,$9F,$C6,$2F,$85,$7C,$9A,$F8,$1C,$7F,$31,$FC,$A7
	.dc.b	$2F,$ED,$1F,$F0,$8D,$7F,$95,$5E,$D8,$F8,$1E,$FD,$7E,$C7,$38,$78
	.dc.b	$F2,$FD,$CE,$BE,$1F,$3B,$DB,$5F,$C3,$B6,$F6,$BF,$83,$3B,$F9,$FF
	.dc.b	$C0,$7F,$21,$F3,$C7,$B1,$AB,$44,$F4,$8F,$E7,$F5,$D1,$FE,$7F,$CD
	.dc.b	$FE,$20,$AF,$F2,$E4,$79,$8E,$98,$77,$3E,$37,$9D,$CF,$97,$F6,$B8
	.dc.b	$73,$7F,$84,$D7,$D2,$E0,$F3,$65,$23,$A1,$E5,$8E,$47,$F0,$87,$F1
	.dc.b	$D3,$FA,$EE,$6F,$E6,$DC,$9F,$EF,$1C,$FF,$C6,$55,$E5,$DD,$B7,$A2
	.dc.b	$1C,$5F,$46,$73,$AF,$DD,$F3,$BB,$F5,$7F,$02,$BE,$FA,$FE,$CF,$5F
	.dc.b	$C9,$EB,$FE,$9C,$D1,$E7,$7C,$FA,$97,$EA,$9C,$BF,$DB,$63,$91,$EB
	.dc.b	$AF,$CF,$3D,$78,$BE,$77,$9B,$F6,$73,$97,$CE,$3C,$AF,$F0,$D3,$F8
	.dc.b	$B1,$F1,$27,$EB,$73,$7E,$07,$29,$3E,$38,$E3,$8F,$F4,$B8,$F6,$E9
	.dc.b	$FC,$55,$FD,$2F,$9B,$F8,$9C,$64,$7A,$E3,$8F,$17,$F8,$53,$93,$EE
	.dc.b	$6B,$F9,$FA,$FF,$2B,$7C,$87,$F7,$EF,$4B,$F4,$11,$F4,$3A,$7F,$5C
	.dc.b	$A3,$3C,$BF,$E1,$5A,$EB,$3F,$C3,$4F,$E2,$A7,$BB,$5F,$BE,$FF,$47
	.dc.b	$3F,$C3,$11,$FE,$0A,$3F,$C1,$7C,$1C,$A3,$F8,$61,$C8,$F7,$9F,$E4
	.dc.b	$03,$FD,$03,$F1,$FF,$E1,$4F,$D1,$5E,$ED,$3F,$C2,$B1,$CF,$B7,$FC
	.dc.b	$26,$FF,$A0,$2E,$5F,$23,$F7,$FE,$97,$EC,$79,$24,$7E,$0E,$39,$70
	.dc.b	$7F,$A3,$3A,$9F,$48,$F5,$F0,$DB,$FF,$50,$4E,$A7,$3F,$07,$FE,$90
	.dc.b	$F9,$1F,$DF,$73,$CB,$FF,$1D,$47,$67,$FD,$3A,$F7,$EF,$67,$4E,$53
	.dc.b	$FE,$2F,$EA,$FE,$11,$EB,$48,$FD,$9F,$0F,$FB,$03,$9B,$FC,$95,$D3
	.dc.b	$F9,$AB,$FD,$37,$3F,$FA,$C1,$E7,$7F,$E9,$7C,$FF,$D4,$F8,$7F,$A4
	.dc.b	$A3,$FC,$E9,$F2,$3F,$C5,$7D,$79,$1D,$A3,$D0,$FF,$A4,$DE,$37,$23
	.dc.b	$FC,$89,$1D,$3E,$47,$F8,$F7,$B1,$27,$CF,$1C,$B2,$FF,$61,$BF,$07
	.dc.b	$D6,$93,$FE,$4D,$8F,$2F,$7F,$F9,$53,$DC,$7C,$05,$3F,$CA,$FA,$A7
	.dc.b	$A1,$FE,$74,$3C,$57,$CC,$FF,$2C,$3F,$03,$E4,$78,$B2,$3E,$5C,$72
	.dc.b	$7F,$D2,$D1,$E2,$78,$9F,$E6,$1F,$91,$F0,$FC,$9F,$EA,$FC,$B9,$3F
	.dc.b	$52,$39,$3F,$E0,$69,$EC,$FF,$AE,$EB,$E8,$C7,$A8,$E7,$1C,$4F,$BA
	.dc.b	$F2,$C0,$F2,$D7,$FD,$D7,$5C,$AB,$F0,$7F,$25,$FE,$7A,$FD,$AF,$FC
	.dc.b	$9F,$1F,$CE,$1A,$FE,$FC,$FF,$59,$47,$53,$C4,$F8,$0E,$67,$E3,$BE
	.dc.b	$4F,$A7,$B3,$23,$07,$F2,$9E,$4F,$E0,$58,$7F,$11,$8F,$F4,$35,$7E
	.dc.b	$E7,$E0,$E4,$70,$3C,$47,$F8,$AB,$CE,$FD,$F7,$C3,$C8,$FC,$EC,$7C
	.dc.b	$13,$EF,$D4,$8C,$9F,$A2,$F9,$3F,$ED,$9F,$03,$23,$07,$94,$FE,$1B
	.dc.b	$B7,$F2,$D3,$F1,$0F,$FB,$4E,$7F,$17,$72,$38,$9E,$67,$FD,$6A,$FC
	.dc.b	$9F,$13,$FA,$CF,$C0,$FF,$7B,$3F,$C2,$5E,$07,$E2,$63,$FC,$33,$F1
	.dc.b	$1F,$E7,$FF,$40,$D3,$1F,$11,$FE,$19,$F0,$3F,$C0,$9E,$84,$CB,$FA
	.dc.b	$B4,$73,$BF,$F4,$24,$78,$5B,$E2,$7F,$67,$69,$FE,$15,$1F,$05,$FE
	.dc.b	$25,$F0,$BF,$C2,$1E,$74,$BF,$EC,$51,$D7,$A3,$CC,$48,$FE,$ED,$6F
	.dc.b	$81,$FF,$17,$78,$9F,$E1,$9F,$42,$57,$F6,$B8,$F3,$7C,$07,$BF,$E0
	.dc.b	$24,$F2,$8F,$F1,$B7,$47,$F8,$87,$C9,$93,$FD,$CE,$3D,$DE,$F7,$C7
	.dc.b	$24,$78,$BD,$EF,$DF,$FC,$9F,$FA,$A7,$F2,$4F,$E4,$1B,$7F,$8C,$DB
	.dc.b	$5F,$E4,$7F,$19,$5F,$C1,$BD,$5F,$F1,$44,$7F,$B9,$CC,$CF,$F8,$4E
	.dc.b	$9E,$77,$53,$BC,$7F,$2E,$F9,$3F,$F7,$2F,$CA,$1A,$7F,$12,$3B,$47
	.dc.b	$E8,$8F,$EC,$C7,$F8,$0C,$FC,$FD,$F8,$1B,$8F,$07,$38,$F3,$7C,$07
	.dc.b	$F3,$E7,$5E,$17,$F8,$53,$C7,$FE,$7B,$E3,$FE,$BE,$B9,$9F,$D1,$2F
	.dc.b	$A7,$5F,$F9,$6E,$3F,$D2,$FF,$27,$FD,$AF,$CF,$91,$FD,$2A,$3F,$95
	.dc.b	$9F,$D1,$32,$91,$FE,$FF,$8F,$E8,$E7,$D1,$F4,$64,$71,$3F,$A9,$FC
	.dc.b	$9F,$D5,$CA,$46,$08,$34,$FD,$8F,$88,$E2,$7A,$3F,$DC,$5B,$5F,$E8
	.dc.b	$1A,$FD,$BB,$FE,$B0,$C3,$FA,$C9,$F8,$3F,$42,$47,$6D,$BE,$07,$0F
	.dc.b	$F7,$78,$FD,$C6,$1F,$CA,$9F,$C0,$7A,$1F,$D2,$BD,$89,$1D,$F5,$FF
	.dc.b	$11,$DB,$2E,$F3,$93,$F1,$D5,$FD,$A8,$FC,$A7,$A1,$23,$FB,$AC,$7F
	.dc.b	$56,$F6,$3F,$31,$E7,$C9,$FE,$F3,$19,$7A,$1F,$EC,$C9,$E8,$E3,$04
	.dc.b	$FF,$78,$3F,$45,$E7,$49,$E6,$7F,$81,$3D,$DF,$D3,$7A,$12,$30,$43
	.dc.b	$E9,$69,$FA,$9E,$8F,$A0,$7F,$81,$2A,$4F,$E9,$7A,$BF,$60,$7F,$70
	.dc.b	$8F,$EE,$3F,$15,$FA,$4B,$95,$FE,$14,$8C,$BE,$99,$E6,$F8,$0F,$DB
	.dc.b	$47,$F8,$48,$FD,$E7,$4C,$8E,$47,$F8,$7B,$E4,$79,$BF,$AA,$9F,$F0
	.dc.b	$0C,$63,$9B,$F0,$BB,$27,$A5,$7C,$CF,$A5,$7A,$BC,$B9,$FF,$0B,$57
	.dc.b	$C1,$FF,$30,$D7,$F9,$8A,$FF,$AA,$B6,$BF,$D2,$9D,$0E,$DF,$E2,$47
	.dc.b	$F9,$A2,$3F,$5D,$3B,$BF,$C4,$E1,$E2,$C3,$E0,$0C,$BC,$CF,$97,$2E
	.dc.b	$55,$E2,$8F,$21,$F1,$8F,$F5,$8D,$F5,$7F,$DC,$8B,$E3,$FF,$92,$74
	.dc.b	$C9,$DC,$74,$77,$7F,$1F,$87,$CB,$1E,$13,$67,$F2,$3E,$7F,$E5,$C7
	.dc.b	$15,$87,$77,$C7,$F0,$1F,$C9,$38,$A4,$60,$F3,$E1,$DD,$F5,$1F,$0A
	.dc.b	$BE,$9B,$A3,$C2,$7D,$0C,$7C,$11,$A8,$FA,$73,$8B,$91,$C4,$C9,$F8
	.dc.b	$83,$E7,$E9,$F3,$9F,$04,$F4,$3A,$DC,$8E,$5B,$7D,$AF,$91,$FF,$15
	.dc.b	$7C,$64,$9F,$A3,$1E,$43,$93,$C3,$7F,$C4,$C6,$63,$FA,$67,$57,$F5
	.dc.b	$73,$0F,$E3,$3D,$9F,$F1,$B7,$97,$23,$EE,$C7,$C0,$3F,$D4,$1E,$BC
	.dc.b	$8C,$90,$FE,$37,$D3,$FD,$8C,$E0,$FE,$1B,$C5,$FF,$54,$7C,$8C,$8C
	.dc.b	$10,$F1,$1C,$F5,$FC,$CC,$FD,$03,$FB,$77,$3C,$9E,$13,$F9,$EC,$3F
	.dc.b	$B9,$75,$C9,$FC,$AC,$7D,$83,$93,$C3,$6F,$D6,$C6,$60,$7F,$8A,$FF
	.dc.b	$25,$FC,$E7,$8C,$7F,$3D,$1F,$90,$9F,$C9,$78,$2F,$6F,$F3,$97,$BE
	.dc.b	$4B,$87,$E5,$70,$FF,$1C,$56,$BB,$DE,$35,$F1,$A5,$78,$B6,$6B,$E4
	.dc.b	$CF,$FA,$B5,$F1,$F1,$B9,$78,$D3,$0E,$17,$BC,$95,$F9,$8F,$6B,$C7
	.dc.b	$DB,$2F,$F7,$31,$97,$88,$CB,$E0,$76,$7A,$8D,$47,$F1,$29,$F6,$F5
	.dc.b	$C8,$E6,$72,$AF,$91,$D3,$F0,$89,$1F,$C2,$3A,$3F,$56,$7C,$6E,$89
	.dc.b	$19,$21,$F4,$F9,$BE,$59,$C1,$F3,$F9,$4F,$86,$61,$FC,$36,$30,$D8
	.dc.b	$CC,$F9,$79,$9A,$7E,$F3,$E4,$3E,$BE,$72,$BF,$8E,$C6,$5E,$0F,$DC
	.dc.b	$30,$FF,$A6,$7E,$59,$F3,$2F,$4F,$F8,$F4,$61,$B1,$99,$F9,$F3,$C3
	.dc.b	$AB,$F8,$6F,$B9,$F8,$8F,$92,$95,$FC,$BA,$32,$FA,$7D,$BF,$C6,$98
	.dc.b	$7F,$D5,$5D,$2F,$9B,$7A,$7F,$CB,$E3,$0D,$8C,$CF,$E0,$27,$87,$99
	.dc.b	$FC,$7B,$1F,$CC,$F9,$12,$BF,$9F,$C6,$5F,$95,$E3,$FE,$78,$C3,$E2
	.dc.b	$C3,$FC,$4E,$DF,$4F,$C0,$78,$7C,$1F,$E5,$7A,$7E,$9B,$E4,$E5,$7F
	.dc.b	$4C,$8C,$BF,$5F,$D8,$FD,$59,$87,$B7,$8B,$EB,$23,$EB,$C7,$F4,$2F
	.dc.b	$F5,$6F,$77,$F6,$7F,$25,$23,$04,$3F,$D5,$5D,$1F,$E5,$5F,$1F,$F6
	.dc.b	$31,$FA,$38,$C7,$17,$F7,$4B,$F9,$5E,$3C,$AF,$03,$F3,$7C,$9F,$E8
	.dc.b	$FC,$7F,$C1,$30,$7C,$2D,$33,$7F,$DA,$A7,$5F,$23,$23,$07,$AF,$D9
	.dc.b	$FB,$04,$7D,$DE,$CF,$F5,$4E,$6E,$DF,$21,$27,$FB,$AC,$65,$FD,$A6
	.dc.b	$4F,$90,$F2,$19,$8F,$B1,$ED,$FE,$14,$8F,$CB,$F5,$FF,$B0,$FC,$07
	.dc.b	$AF,$E3,$E4,$FF,$80,$A3,$2F,$6F,$8E,$F8,$06,$1F,$C2,$7B,$BF,$91
	.dc.b	$23,$F5,$FD,$DF,$ED,$BD,$BF,$1F,$E2,$64,$FF,$83,$A3,$2F,$AB,$E2
	.dc.b	$3E,$61,$87,$F2,$1E,$FF,$E7,$48,$FE,$0F,$DD,$FE,$F3,$E0,$7D,$3F
	.dc.b	$85,$93,$FE,$1A,$8C,$BF,$1B,$CC,$FB,$06,$1F,$C2,$FC,$47,$DB,$23
	.dc.b	$FC,$4B,$D1,$7A,$BF,$E0,$6F,$85,$FC,$07,$C4,$CA,$FF,$13,$C7,$07
	.dc.b	$99,$F5,$2F,$83,$F9,$9F,$81,$FD,$B1,$1F,$CE,$7C,$0F,$F4,$EF,$80
	.dc.b	$FB,$51,$FC,$BD,$AB,$C7,$FC,$CF,$B7,$FE,$33,$EC,$E1,$AF,$F0,$1F
	.dc.b	$81,$97,$FA,$77,$A7,$5F,$F8,$D7,$E6,$5E,$6B,$E3,$B0,$DB,$FE,$8F
	.dc.b	$3A,$9F,$E5,$DE,$D7,$F8,$9B,$CE,$FE,$C7,$F8,$2F,$FB,$8B,$53,$FE
	.dc.b	$E4,$C3,$FC,$73,$3D,$2E,$47,$4C,$7B,$3E,$BF,$3F,$FB,$15,$FF,$B9
	.dc.b	$23,$F8,$C7,$E0,$9F,$E5,$19,$34,$9C,$3F,$E4,$7F,$5B,$F7,$35,$FD
	.dc.b	$76,$FF,$D7,$3C,$5F,$19,$B8,$7F,$80,$A5,$5C,$23,$AC,$7F,$2B,$C3
	.dc.b	$B3,$7E,$0F,$1F,$07,$23,$CA,$EF,$E0,$FC,$5F,$2B,$F1,$35,$95,$FD
	.dc.b	$C0,$EF,$1C,$5F,$E7,$78,$7F,$93,$F4,$FE,$27,$97,$BB,$1F,$17,$C1
	.dc.b	$78,$FB,$66,$1F,$11,$F3,$7C,$27,$C1,$F5,$3E,$9F,$AB,$2B,$C1,$C1
	.dc.b	$FC,$51,$E0,$43,$C3,$7E,$FE,$6F,$81,$E7,$C9,$FA,$2F,$C3,$F8,$2F
	.dc.b	$9D,$EA,$FE,$03,$D1,$95,$DA,$38,$3E,$CA,$BC,$1E,$CE,$17,$BF,$F3
	.dc.b	$7D,$E9,$7F,$79,$FA,$1E,$E7,$DB,$F9,$1F,$C6,$FA,$32,$BD,$71,$C1
	.dc.b	$F9,$35,$71,$ED,$EE,$7D,$9F,$6E,$57,$E2,$9F,$B9,$EC,$7E,$1F,$D9
	.dc.b	$FC,$CF,$95,$2B,$E3,$C7,$07,$EA,$D5,$C3,$D5,$EC,$7E,$17,$D5,$95
	.dc.b	$F9,$67,$F1,$1E,$8F,$E4,$FC,$CF,$F2,$3F,$CD,$FF,$93,$8F,$F2,$6F
	.dc.b	$57,$EE,$B1,$F8,$97,$CB,$0F,$C1,$BF,$91,$F4,$3F,$25,$E4,$C9,$FD
	.dc.b	$1B,$FD,$FF,$E1,$3F,$3F,$1F,$9A,$F3,$FB,$F8,$72,$3F,$03,$1C,$5F
	.dc.b	$C7,$D5,$C7,$F1,$F6,$F1,$7A,$78,$52,$FF,$C1,$13,$F8,$97,$FC,$22
	.dc.b	$BB,$7F,$61,$E7,$7B,$BC,$29,$5F,$90,$8E,$4F,$0B,$B8,$F8,$3A,$BD
	.dc.b	$5C,$D2,$BB,$BE,$DE,$3F,$DE,$7B,$5F,$2B,$C0,$95,$F9,$C8,$E0,$F9
	.dc.b	$2A,$E1,$EC,$E0,$F8,$D9,$CA,$F5,$BF,$26,$FF,$84,$79,$BF,$5B,$B6
	.dc.b	$57,$EA,$23,$83,$F0,$2A,$E1,$F2,$27,$E9,$7B,$32,$BE,$3B,$F5,$7D
	.dc.b	$9F,$E2,$8F,$E0,$BB,$24,$FE,$D2,$38,$3F,$31,$B7,$F5,$3E,$43,$EF
	.dc.b	$7B,$12,$BE,$9B,$F8,$1F,$63,$F9,$06,$3F,$8E,$EB,$95,$FC,$02,$38
	.dc.b	$3F,$62,$AE,$1F,$80,$F8,$4F,$C5,$7B,$32,$BE,$FB,$F8,$DF,$67,$F9
	.dc.b	$66,$7F,$9A,$EC,$95,$FC,$32,$38,$3F,$84,$AB,$87,$E3,$3E,$43,$F2
	.dc.b	$DF,$21,$2B,$F1,$6F,$F9,$5F,$D4,$FE,$8D,$E1,$35,$FF,$56,$FC,$BF
	.dc.b	$E3,$F1,$FE,$BA,$DF,$53,$F9,$C5,$6F,$1E,$64,$3F,$2B,$8F,$5F,$91
	.dc.b	$ED,$F1,$12,$7F,$CC,$59,$FF,$99,$1F,$F3,$5B,$FD,$11,$FF,$62,$DF
	.dc.b	$F9,$E5,$FE,$A8,$7F,$A9,$F1,$76,$79,$8F,$EA,$1C,$BD,$67,$9E,$17
	.dc.b	$4F,$E7,$78,$FF,$99,$9B,$F0,$FF,$04,$95,$E4,$F6,$3F,$AD,$72,$FC
	.dc.b	$2E,$B9,$3F,$CE,$A3,$6F,$A2,$AD,$F6,$F8,$CF,$99,$F0,$F2,$BD,$EF
	.dc.b	$CF,$F6,$3F,$B3,$73,$7C,$EE,$C9,$5F,$D1,$A3,$83,$F1,$2A,$E1,$EA
	.dc.b	$F8,$7F,$B1,$F0,$F2,$BE,$5B,$F6,$FD,$8F,$EE,$5D,$BF,$6B,$AA,$57
	.dc.b	$F5,$28,$E0,$FD,$0A,$B8,$7C,$6F,$87,$FF,$5E,$7E,$12,$EF,$A4,$BF
	.dc.b	$F6,$1C,$3D,$0F,$F0,$07,$7F,$E2,$32,$95,$FD,$82,$3F,$C3,$DF,$85
	.dc.b	$BD,$6A,$49,$FF,$0E,$C3,$C4,$FF,$66,$7E,$36,$F3,$69,$2B,$FD,$A7
	.dc.b	$1F,$C8,$5F,$ED,$BE,$1F,$F8,$2F,$7F,$F1,$9F,$3F,$F8,$66,$7F,$13
	.dc.b	$F0,$88,$C4,$FF,$A4,$27,$FC,$23,$5F,$E3,$0F,$1E,$47,$F8,$95,$FF
	.dc.b	$19,$F8,$FF,$A7,$F8,$A9,$3F,$9A,$7F,$55,$ED,$F8,$7C,$9F,$F0,$E7
	.dc.b	$D3,$FF,$8C,$F3,$78,$19,$C3,$FE,$00,$46,$1F,$F0,$A4,$BE,$43,$F6
	.dc.b	$9F,$19,$27,$F4,$EF,$ED,$FE,$33,$F7,$1E,$F4,$9F,$D5,$BF,$BB,$F8
	.dc.b	$CE,$D0,$C3,$DF,$9A,$3C,$39,$78,$F8,$BE,$23,$C2,$B7,$2A,$79,$2F
	.dc.b	$8C,$FE,$10,$49,$FF,$6E,$7E,$0F,$FE,$45,$1F,$19,$FE,$E3,$6C,$6A
	.dc.b	$31,$F1,$7D,$FC,$EF,$16,$6F,$0E,$99,$BE,$BF,$01,$CC,$8F,$83,$3E
	.dc.b	$DF,$17,$23,$C6,$F8,$7D,$EF,$67,$AB,$DB,$92,$57,$D0,$8F,$37,$C5
	.dc.b	$F9,$FE,$2E,$4F,$B2,$3D,$5E,$F4,$8F,$7B,$ED,$F8,$BF,$91,$F1,$FF
	.dc.b	$2F,$F0,$8F,$F2,$E5,$7E,$41,$BE,$1F,$E3,$FA,$3F,$89,$74,$7D,$31
	.dc.b	$CF,$FE,$34,$BF,$97,$EF,$FC,$93,$E5,$7C,$67,$CB,$F8,$C9,$3F,$41
	.dc.b	$F9,$DF,$19,$F6,$7C,$2F,$17,$2C,$AF,$C6,$C7,$D3,$F8,$1F,$A9,$E6
	.dc.b	$C9,$FB,$31,$F5,$FE,$32,$47,$DC,$7E,$D7,$9B,$ED,$E3,$93,$F8,$BE
	.dc.b	$1F,$F5,$0F,$12,$7F,$15,$EF,$7E,$07,$E3,$3F,$05,$E9,$CC,$3F,$0F
	.dc.b	$1F,$85,$F5,$64,$7E,$29,$FD,$37,$2B,$F0,$8F,$F4,$21,$FE,$25,$9F
	.dc.b	$04,$7C,$98,$C4,$7F,$46,$8E,$1F,$19,$F4,$F7,$E7,$5F,$D6,$6B,$F3
	.dc.b	$8F,$F0,$97,$83,$75,$FF,$40,$BB,$BF,$17,$A7,$CB,$E0,$E5,$F8,$8E
	.dc.b	$71,$DB,$E2,$64,$66,$BD,$51,$E2,$F8,$9F,$1F,$C2,$79,$3E,$26,$5E
	.dc.b	$F2,$4F,$A3,$E2,$64,$E5,$F5,$FC,$27,$B3,$D8,$93,$C5,$F7,$FB,$1F
	.dc.b	$03,$E0,$3E,$0F,$C3,$CA,$ED,$1F,$13,$E1,$E4,$78,$5F,$8F,$F0,$DF
	.dc.b	$23,$E0,$64,$F9,$5F,$97,$F0,$3F,$33,$DF,$FA,$FF,$84,$7F,$19,$AF
	.dc.b	$80,$D7,$E6,$C6,$BE,$2A,$46,$9E,$3E,$BF,$2F,$46,$4F,$27,$AF,$9D
	.dc.b	$DB,$C7,$EF,$E1,$4A,$EF,$1E,$2F,$89,$91,$E2,$7C,$BF,$13,$E6,$F6
	.dc.b	$64,$F9,$9F,$4F,$C4,$FA,$BE,$23,$D7,$F1,$32,$BD,$71,$EE,$F7,$24
	.dc.b	$7B,$9F,$83,$F0,$DF,$0B,$E1,$A4,$FC,$27,$E2,$FC,$2F,$C6,$F7,$3E
	.dc.b	$3F,$C2,$CA,$F8,$F1,$F2,$BE,$17,$E5,$FC,$24,$9F,$97,$1F,$A4,$E3
	.dc.b	$FF,$79,$12,$31,$F0,$99,$F8,$4D,$7C,$24,$9F,$C9,$52,$79,$7C,$27
	.dc.b	$3F,$6A,$56,$A3,$B7,$C2,$48,$E2,$F8,$7E,$0F,$C5,$F0,$9E,$3F,$84
	.dc.b	$95,$DA,$3C,$DE,$D7,$9F,$E0,$64,$F8,$A3,$D5,$F0,$52,$3C,$AF,$B7
	.dc.b	$E0,$BD,$DF,$05,$EF,$F6,$A5,$7A,$A3,$E1,$7C,$14,$8F,$6B,$F1,$7E
	.dc.b	$0B,$E3,$7C,$14,$9F,$82,$FC,$9F,$80,$F9,$5F,$03,$27,$E2,$BF,$37
	.dc.b	$E0,$7F,$88,$FE,$11,$FE,$05,$F8,$1F,$EC,$55,$FD,$8E,$3F,$87,$C6
	.dc.b	$3D,$F4,$61,$FE,$78,$F1,$6F,$C6,$74,$25,$7F,$05,$9F,$E2,$13,$C3
	.dc.b	$E0,$7C,$24,$9D,$7C,$17,$93,$E0,$A4,$71,$7C,$FF,$05,$E8,$F8,$09
	.dc.b	$3D,$5F,$5F,$C1,$7B,$3E,$06,$4F,$85,$F7,$FC,$0F,$C0,$F8,$19,$3E
	.dc.b	$57,$E1,$FC,$0F,$C4,$F8,$09,$3E,$97,$E3,$FC,$0F,$C8,$F8,$1F,$93
	.dc.b	$F0,$32,$BD,$D1,$F3,$3E,$06,$47,$C1,$7E,$7F,$C0,$FD,$0F,$80,$93
	.dc.b	$F1,$5F,$E8,$7F,$84,$7F,$0B,$E8,$FA,$0F,$F8,$5B,$98,$D2,$7C,$07
	.dc.b	$0F,$80,$98,$7F,$33,$9F,$F4,$75,$70,$F3,$BA,$FB,$F2,$B4,$F8,$3E
	.dc.b	$03,$C3,$F0,$12,$79,$3E,$4F,$80,$F2,$FC,$07,$9B,$E0,$25,$77,$8F
	.dc.b	$4F,$BF,$23,$C4,$FB,$3E,$03,$DB,$F0,$1E,$EF,$80,$95,$E7,$8F,$83
	.dc.b	$F0,$1F,$0B,$E0,$24,$FA,$E3,$E2,$FB,$FF,$1B,$E0,$24,$FB,$E3,$E4
	.dc.b	$FB,$FF,$2B,$E0,$24,$FC,$38,$FE,$DF,$F8,$47,$F8,$5B,$97,$FC,$28
	.dc.b	$FF,$8E,$27,$FB,$63,$7A,$7F,$40,$EE,$3F,$C8,$CF,$F0,$D7,$FB,$03
	.dc.b	$5E,$3E,$FF,$6F,$80,$EF,$EF,$C9,$DB,$E2,$F8,$0F,$1F,$C0,$79,$3E
	.dc.b	$02,$57,$48,$F3,$FB,$F2,$3B,$BE,$AF,$80,$F5,$FB,$F2,$7C,$6F,$BB
	.dc.b	$E0,$3D,$FF,$01,$27,$CE,$FC,$2F,$80,$FF,$3B,$7D,$03,$F6,$26,$78
	.dc.b	$6D,$FF,$80,$67,$53,$9D,$35,$A4,$AD,$5A,$39,$69,$DE,$7E,$3E,$35
	.dc.b	$F9,$FF,$03,$DF,$19,$1C,$9E,$CF,$4B,$EC,$FF,$B9,$B2,$F4,$47,$68
	.dc.b	$FA,$7F,$05,$F7,$88,$FD,$6F,$82,$FC,$0F,$C1,$7E,$0B,$E0,$A4,$FD
	.dc.b	$A8,$FC,$2F,$C1,$7E,$1B,$E0,$A4,$FD,$E8,$FC,$4F,$C1,$48,$FC,$0B
	.dc.b	$FE,$3A,$F8,$0F,$E4,$51,$FE,$8B,$7F,$93,$37,$E1,$F4,$6F,$E0,$F8
	.dc.b	$7C,$8F,$1A,$95,$FE,$2B,$7F,$93,$BC,$FE,$0F,$B7,$C1,$C8,$D4,$7D
	.dc.b	$7C,$E4,$7D,$27,$ED,$65,$F7,$B4,$F2,$F0,$9A,$4D,$3D,$58,$4C,$3A
	.dc.b	$9E,$3F,$83,$EA,$FD,$1E,$7F,$F7,$23,$DC,$FC,$0E,$3D,$FC,$29,$1F
	.dc.b	$89,$C7,$B3,$ED,$F8,$3F,$90,$47,$5F,$05,$E7,$DA,$47,$17,$D5,$97
	.dc.b	$AE,$BD,$9F,$07,$2B,$B4,$7B,$EE,$47,$43,$C5,$F0,$7D,$1F,$C5,$62
	.dc.b	$C7,$1E,$37,$F6,$2E,$0F,$8D,$9F,$7B,$F5,$67,$DD,$E3,$9F,$E0,$63
	.dc.b	$E7,$65,$27,$ED,$47,$4D,$3F,$DA,$98,$7D,$1C,$B1,$48,$EE,$FA,$8E
	.dc.b	$E4,$3B,$9F,$85,$D3,$F2,$04,$7F,$6D,$FB,$1F,$F4,$9B,$EF,$AB,$DF
	.dc.b	$FC,$17,$C0,$4B,$FA,$91,$F8,$5F,$80,$91,$F5,$DF,$C4,$7C,$07,$F0
	.dc.b	$6C,$7A,$4F,$9D,$B9,$9B,$8B,$E0,$3F,$21,$8F,$7B,$F9,$F9,$61,$EC
	.dc.b	$43,$E9,$E5,$E0,$7D,$7F,$01,$E1,$23,$FA,$5E,$0F,$27,$C0,$48,$F8
	.dc.b	$2F,$9F,$DF,$F4,$7B,$F2,$7E,$2B,$EB,$F7,$FF,$71,$7D,$27,$F8,$66
	.dc.b	$CC,$73,$1D,$08,$6B,$1E,$CF,$42,$1C,$B0,$EC,$FD,$4F,$5F,$E5,$11
	.dc.b	$FA,$FE,$BF,$CD,$F6,$E4,$7D,$A7,$E8,$7B,$7F,$47,$DA,$FA,$5E,$DC
	.dc.b	$AF,$BF,$1F,$56,$FA,$47,$AB,$0C,$3C,$C8,$7B,$B0,$EA,$F2,$3F,$13
	.dc.b	$7D,$5F,$87,$7F,$C7,$7E,$12,$47,$F1,$57,$F0,$9E,$DF,$A2,$3A,$1D
	.dc.b	$BB,$FF,$55,$BB,$4F,$54,$F6,$21,$D8,$FE,$41,$F0,$9F,$CA,$C8,$FF
	.dc.b	$51,$FD,$91,$EB,$FF,$58,$7F,$BA,$75,$9A,$4F,$87,$D7,$B7,$2F,$93
	.dc.b	$C7,$DB,$E5,$ED,$F3,$C6,$57,$78,$EC,$C3,$C4,$78,$6F,$A3,$FD,$4F
	.dc.b	$0F,$F0,$16,$BF,$C2,$5B,$E0,$FF,$00,$58,$FA,$7D,$4F,$57,$B1,$30
	.dc.b	$F7,$47,$B7,$D5,$F7,$7B,$12,$7E,$14,$7C,$1F,$6F,$E1,$7B,$72,$7E
	.dc.b	$34,$75,$79,$90,$F3,$7B,$7D,$5F,$99,$7D,$5F,$55,$F5,$7E,$85,$F5
	.dc.b	$7D,$D7,$F4,$7D,$9F,$A5,$ED,$49,$FA,$D1,$F5,$7D,$AF,$AD,$E5,$49
	.dc.b	$FB,$51,$F6,$7D,$A9,$1F,$75,$FB,$9E,$D7,$58,$E6,$43,$E7,$7A,$FD
	.dc.b	$5F,$C2,$5F,$57,$E9,$5F,$47,$EA,$5D,$E4,$FA,$98,$7E,$27,$0E,$D5
	.dc.b	$F6,$71,$C3,$E6,$3A,$90,$EA,$79,$C8,$76,$3F,$2D,$3F,$E2,$CF,$D9
	.dc.b	$7F,$C0,$3D,$47,$FB,$89,$3D,$8D,$7B,$72,$FF,$C5,$D2,$8E,$5E,$DF
	.dc.b	$3F,$6E,$56,$9E,$DE,$DF,$7C,$24,$F2,$7C,$5E,$B7,$58,$FF,$14,$63
	.dc.b	$D5,$EF,$8F,$67,$A1,$0D,$E1,$D9,$FF,$0A,$7C,$3F,$F8,$7F,$E1,$E4
	.dc.b	$60,$FF,$0C,$7A,$BF,$E2,$3F,$56,$46,$0F,$F0,$E7,$AB,$FE,$27,$23
	.dc.b	$FE,$1F,$F5,$FE,$47,$B7,$23,$E1,$BF,$2F,$DB,$F9,$9E,$DC,$9F,$8E
	.dc.b	$FC,$FF,$6F,$E8,$7B,$1F,$47,$D0,$95,$F3,$23,$93,$D0,$87,$C4,$BE
	.dc.b	$8F,$C6,$BB,$D5,$FE,$03,$E2,$63,$DA,$BE,$BE,$3D,$9F,$9B,$8F,$F9
	.dc.b	$0F,$DB,$91,$FE,$3B,$4C,$C3,$FD,$D9,$F8,$67,$F6,$CF,$28,$FF,$AF
	.dc.b	$CF,$EA,$53,$B8,$E2,$6B,$DA,$E7,$0A,$E7,$F1,$58,$7A,$FB,$76,$F8
	.dc.b	$E0,$F7,$71,$82,$7C,$31,$87,$C7,$56,$9E,$E5,$F7,$66,$1E,$53,$94
	.dc.b	$E9,$C5,$FB,$3E,$23,$91,$23,$32,$F8,$8E,$DF,$0F,$27,$2F,$87,$E2
	.dc.b	$3C,$5E,$FF,$8F,$DB,$95,$CA,$3C,$DE,$F4,$8E,$AF,$A7,$DF,$F5,$7C
	.dc.b	$0C,$9F,$0B,$ED,$F8,$1F,$77,$C0,$FB,$FE,$22,$57,$9A,$3E,$17,$C3
	.dc.b	$FC,$3F,$88,$93,$EA,$8F,$91,$F8,$67,$FB,$1F,$8C,$FF,$B2,$DF,$B2
	.dc.b	$D9,$FA,$0E,$CF,$E8,$EE,$3D,$EE,$67,$4F,$83,$EB,$F0,$5D,$BE,$0E
	.dc.b	$57,$CE,$B4,$F8,$BE,$03,$C7,$F0,$32,$B5,$1E,$6F,$7E,$47,$17,$D3
	.dc.b	$F0,$1F,$C6,$7C,$D9,$3F,$C5,$1F,$E3,$9E,$77,$BB,$DF,$F7,$FC,$04
	.dc.b	$AF,$14,$7C,$2F,$81,$F8,$7F,$11,$27,$CD,$1C,$9E,$84,$3F,$8A,$DF
	.dc.b	$CA,$F8,$2F,$97,$EF,$49,$F7,$47,$CE,$F8,$19,$1F,$05,$FA,$3F,$05
	.dc.b	$D6,$3F,$8F,$E9,$EA,$7A,$9F,$1F,$E0,$BE,$B9,$1F,$95,$F0,$3F,$6B
	.dc.b	$E0,$A4,$7C,$D7,$EE,$FC,$3F,$5F,$3B,$F8,$26,$1D,$5F,$E0,$DA,$FE
	.dc.b	$A7,$F0,$CF,$F6,$1F,$61,$FF,$61,$B7,$E0,$74,$EF,$E0,$38,$7C,$1C
	.dc.b	$D3,$F5,$37,$FC,$DD,$F4,$9E,$C3,$E4,$63,$E0,$77,$F0,$3E,$12,$3C
	.dc.b	$BE,$0F,$C9,$F0,$7E,$5F,$7E,$4F,$68,$F6,$67,$FE,$96,$F7,$5B,$FB
	.dc.b	$63,$1B,$FB,$7D,$FF,$77,$BF,$31,$F3,$C7,$C1,$F7,$E4,$7A,$9F,$89
	.dc.b	$EF,$76,$8F,$F5,$06,$5D,$9F,$5E,$3D,$9F,$83,$EF,$FC,$BF,$87,$91
	.dc.b	$83,$E2,$FC,$27,$CF,$23,$F2,$3E,$13,$B7,$C2,$FF,$71,$DF,$D5,$3E
	.dc.b	$0D,$FE,$B1,$1F,$9E,$FD,$98,$FB,$5F,$0B,$F6,$FE,$16,$57,$D4,$8E
	.dc.b	$BA,$FA,$63,$EA,$1F,$5B,$9B,$F9,$63,$FC,$B7,$9F,$FD,$48,$FF,$24
	.dc.b	$F3,$7F,$A0,$1F,$D0,$63,$31,$FC,$EF,$83,$F9,$FC,$BE,$0F,$F9,$A3
	.dc.b	$FD,$12,$74,$7F,$39,$71,$7E,$0A,$FF,$33,$D7,$9B,$3F,$83,$AF,$F8
	.dc.b	$4F,$5F,$F0,$B6,$9D,$3E,$0A,$67,$FE,$12,$8E,$FF,$07,$E0,$F8,$39
	.dc.b	$3E,$98,$E8,$FD,$83,$E6,$1C,$48,$7F,$84,$27,$FC,$49,$C1,$EA,$F8
	.dc.b	$39,$3F,$0D,$F6,$FC,$1F,$48,$F9,$7B,$FE,$75,$E4,$74,$7F,$BA,$5F
	.dc.b	$48,$FF,$14,$F2,$7F,$8C,$79,$24,$60,$FF,$16,$EB,$F0,$B9,$3E,$C6
	.dc.b	$1D,$4E,$73,$D7,$8B,$F4,$58,$F5,$8F,$F1,$4E,$3F,$E3,$D2,$3F,$E3
	.dc.b	$8F,$85,$F4,$5F,$F0,$48,$EE,$73,$3F,$3F,$5F,$41,$FF,$1F,$EB,$FE
	.dc.b	$47,$AE,$71,$D8,$E8,$72,$3E,$1E,$1F,$B5,$E2,$FC,$8D,$F3,$9F,$8A
	.dc.b	$7F,$8F,$AF,$FB,$EF,$13,$D9,$FC,$54,$F4,$3F,$C8,$78,$EA,$33,$1F
	.dc.b	$99,$C7,$B3,$F7,$6B,$F1,$0F,$AD,$FF,$2E,$61,$E1,$CB,$FC,$4D,$8F
	.dc.b	$3F,$85,$E9,$C9,$30,$FF,$03,$61,$E3,$9F,$07,$1C,$9C,$E7,$FE,$15
	.dc.b	$6F,$27,$97,$92,$46,$9F,$47,$27,$A7,$E0,$FC,$13,$CB,$1C,$3D,$31
	.dc.b	$EC,$FE,$A3,$2F,$89,$1F,$07,$1E,$B3,$9C,$99,$63,$5F,$DD,$E1,$F2
	.dc.b	$F3,$F9,$1F,$03,$30,$F4,$BF,$2F,$DE,$F9,$9F,$05,$F3,$7E,$0A,$57
	.dc.b	$BA,$3E,$87,$C1,$7D,$1F,$82,$93,$F0,$A3,$93,$D0,$87,$96,$FA,$BF
	.dc.b	$22,$FA,$BE,$9B,$EA,$FC,$CB,$EA,$FB,$6F,$F0,$BF,$B3,$7F,$B5,$23
	.dc.b	$E8,$ED,$F8,$33,$3F,$01,$F8,$5F,$81,$93,$F5,$9F,$C4,$7C,$17,$28
	.dc.b	$E7,$F1,$11,$FB,$F8,$7F,$BD,$8F,$F0,$EF,$01,$F0,$47,$D3,$6F,$0F
	.dc.b	$CE,$77,$93,$1F,$0D,$7D,$57,$C2,$F6,$3C,$07,$3C,$3C,$8F,$8B,$E0
	.dc.b	$7D,$FF,$07,$23,$07,$AC,$87,$53,$F2,$79,$F2,$7A,$1E,$CA,$C6,$5E
	.dc.b	$7F,$81,$F9,$7F,$05,$2B,$E0,$C7,$CE,$F8,$29,$1F,$11,$FA,$3F,$05
	.dc.b	$F4,$BE,$06,$4F,$C8,$7E,$AF,$C1,$7D,$6F,$81,$FA,$F8,$4A,$F9,$B1
	.dc.b	$D1,$F9,$58,$5C,$BE,$56,$3D,$27,$E7,$63,$67,$E7,$63,$F8,$2F,$80
	.dc.b	$95,$F6,$1F,$C2,$FB,$9F,$86,$F8,$0F,$C3,$FC,$04,$AF,$BB,$1F,$8A
	.dc.b	$F8,$09,$1F,$80,$7F,$1B,$F0,$1C,$A3,$F0,$77,$5F,$C0,$7C,$07,$57
	.dc.b	$F0,$F8,$75,$7F,$09,$87,$47,$F0,$D8,$5C,$FF,$0D,$8F,$F1,$EF,$D9
	.dc.b	$FF,$DD,$E4,$F8,$1C,$7B,$99,$F8,$1D,$7C,$0C,$CB,$F8,$9D,$FF,$8B
	.dc.b	$71,$FD,$EA,$F8,$1E,$9F,$03,$2B,$71,$D1,$E3,$8D,$7F,$86,$53,$E0
	.dc.b	$7C,$11,$DB,$1F,$03,$AC,$BD,$18,$75,$8E,$38,$E1,$F2,$DF,$5C,$BC
	.dc.b	$B9,$7B,$E3,$AC,$F8,$31,$EB,$5E,$AC,$79,$4F,$43,$DF,$86,$27,$DF
	.dc.b	$F0,$3F,$24,$93,$F1,$3E,$03,$E6,$7C,$04,$8F,$8E,$FC,$FF,$80,$FA
	.dc.b	$1F,$01,$27,$E5,$BF,$4F,$DA,$FA,$97,$D2,$7E,$2E,$18,$79,$90,$F9
	.dc.b	$38,$75,$79,$90,$F9,$B7,$D5,$FB,$3E,$E7,$E0,$48,$FF,$4E,$FD,$A3
	.dc.b	$FA,$DE,$7F,$73,$D9,$FC,$31,$27,$F0,$3E,$CF,$E2,$7D,$B9,$1F,$84
	.dc.b	$7F,$B4,$DB,$E5,$AF,$C0,$4A,$EC,$EB,$DA,$E9,$1F,$82,$C6,$E5,$F8
	.dc.b	$CC,$BA,$CF,$E1,$B2,$B3,$FD,$33,$2E,$D3,$E9,$F6,$BC,$3E,$E4,$8C
	.dc.b	$1E,$DF,$6F,$CB,$ED,$C8,$C1,$F0,$7D,$BF,$49,$1F,$89,$F1,$3E,$CC
	.dc.b	$3A,$47,$F6,$3C,$58,$E6,$39,$99,$C3,$B3,$BC,$7A,$3C,$71,$B9,$7A
	.dc.b	$F2,$E5,$3D,$72,$B3,$EF,$F7,$7E,$6E,$1D,$23,$E1,$63,$8B,$F5,$10
	.dc.b	$EC,$7A,$8F,$BD,$F1,$3F,$58,$8F,$E0,$7D,$EF,$B3,$EE,$48,$FC,$23
	.dc.b	$F7,$3D,$DF,$BB,$EE,$49,$FC,$43,$F8,$0F,$77,$F0,$37,$D2,$7E,$36
	.dc.b	$18,$79,$90,$F9,$58,$75,$79,$90,$F9,$D7,$D5,$FC,$B7,$B5,$F8,$E2
	.dc.b	$3F,$E2,$6F,$80,$FE,$CF,$C7,$FD,$86,$3F,$39,$ED,$FE,$50,$FC,$AF
	.dc.b	$B3,$FE,$3E,$A4,$DF,$66,$3C,$CD,$7B,$5D,$9F,$C0,$63,$87,$A9,$0F
	.dc.b	$C2,$61,$D9,$E4,$79,$6B,$AB,$FE,$2B,$BF,$0B,$E9,$F7,$BC,$9E,$E4
	.dc.b	$8C,$1E,$DF,$83,$F4,$7B,$92,$30,$7C,$1F,$77,$D9,$EE,$48,$C1,$C8
	.dc.b	$F0,$61,$F0,$4F,$85,$EE,$7C,$3C,$3A,$4F,$97,$0C,$3C,$8F,$45,$F5
	.dc.b	$7D,$58,$72,$7B,$74,$7F,$90,$33,$FF,$0A,$FC,$0F,$CE,$F8,$0F,$F7
	.dc.b	$2E,$7D,$2F,$B1,$F6,$72,$C6,$1C,$71,$EC,$F3,$CB,$EC,$47,$D9,$F7
	.dc.b	$A5,$FE,$05,$FB,$9E,$F7,$DD,$F7,$E4,$FE,$15,$FC,$07,$C3,$FE,$07
	.dc.b	$E0,$3F,$05,$F0,$F2,$BF,$15,$1C,$9E,$84,$3E,$C5,$F4,$7E,$D5,$DC
	.dc.b	$FE,$06,$1D,$67,$F2,$98,$75,$7F,$01,$87,$E4,$7D,$E9,$1F,$99,$7F
	.dc.b	$29,$EF,$FE,$57,$DF,$FC,$B7,$BF,$2B,$F3,$D1,$F9,$9F,$81,$FC,$D7
	.dc.b	$C0,$C9,$FD,$14,$72,$7F,$49,$75,$FF,$3D,$7E,$D3,$FE,$EE,$7F,$25
	.dc.b	$7D,$5F,$CA,$61,$E1,$79,$9F,$AD,$C3,$B3,$F9,$9C,$3F,$D4,$F6,$F8
	.dc.b	$2A,$49,$FF,$4C,$C7,$F8,$1F,$9F,$F0,$4B,$DE,$E3,$F0,$32,$B7,$1D
	.dc.b	$3D,$EE,$BF,$03,$27,$9C,$78,$3E,$03,$B4,$7B,$32,$C3,$C8,$F0,$E1
	.dc.b	$D5,$F8,$38,$75,$7C,$B8,$75,$7E,$2E,$1C,$9F,$8F,$85,$E4,$FF,$4D
	.dc.b	$E5,$FE,$17,$9E,$C7,$CC,$C7,$B4,$EB,$1E,$D3,$F4,$31,$ED,$3C,$B1
	.dc.b	$F9,$3F,$11,$2B,$E3,$BF,$33,$DE,$F9,$BE,$FF,$CE,$F7,$E5,$7C,$C8
	.dc.b	$FA,$3F,$03,$23,$E7,$BF,$53,$E0,$39,$47,$42,$1F,$82,$BE,$AF,$D6
	.dc.b	$BE,$AF,$E1,$AF,$A3,$F8,$8B,$B9,$7C,$2C,$3F,$03,$F0,$32,$BE,$FB
	.dc.b	$F8,$4F,$87,$FC,$2F,$C1,$7E,$1B,$C6,$95,$F8,$48,$FC,$4F,$BF,$F8
	.dc.b	$AF,$7E,$4F,$E2,$23,$93,$D0,$87,$E6,$AF,$AB,$F8,$EB,$EA,$FE,$7A
	.dc.b	$FA,$BF,$92,$BE,$AF,$E8,$AF,$F3,$3E,$FC,$8F,$CB,$BF,$CD,$33,$FF
	.dc.b	$6C,$FE,$D5,$FE,$0E,$9F,$F7,$13,$FE,$26,$E9,$FE,$38,$D8,$79,$48
	.dc.b	$C3,$7E,$07,$A5,$70,$F8,$1B,$7C,$2F,$FD,$F4,$F6,$F7,$F3,$87,$3A
	.dc.b	$B5,$E3,$C7,$F2,$79,$7A,$70,$F2,$FC,$1E,$BE,$0E,$47,$F5,$EA,$F5
	.dc.b	$C7,$2D,$39,$E3,$D3,$39,$7B,$D3,$CC,$D3,$CD,$8F,$88,$8E,$74,$F8
	.dc.b	$1F,$01,$E0,$8F,$8F,$95,$CF,$E5,$E7,$E0,$9E,$1A,$FC,$09,$F9,$F9
	.dc.b	$E6,$7E,$7F,$C0,$7C,$0F,$83,$91,$C2,$3E,$1F,$C1,$FC,$4F,$83,$93
	.dc.b	$D2,$3E,$3F,$C0,$48,$EE,$FC,$AF,$81,$F9,$77,$27,$A1,$E4,$F8,$1E
	.dc.b	$AF,$9B,$0E,$AF,$E0,$B0,$E8,$FE,$13,$0B,$9F,$E2,$31,$FA,$FF,$07
	.dc.b	$2B,$DE,$FD,$AF,$82,$FB,$7F,$07,$F7,$3E,$1E,$57,$C4,$8F,$BF,$F0
	.dc.b	$3F,$80,$F8,$B9,$3F,$22,$39,$3D,$08,$7E,$5A,$FA,$BF,$3A,$FA,$BF
	.dc.b	$9A,$BE,$AF,$D2,$BE,$AF,$E7,$AF,$F1,$FF,$05,$F9,$0F,$84,$93,$F6
	.dc.b	$23,$F2,$BF,$B5,$FF,$86,$A1,$AB,$AF,$FC,$01,$2F,$84,$D9,$C2,$B8
	.dc.b	$FC,$2F,$2F,$84,$91,$F8,$BA,$4F,$8A,$3C,$04,$3B,$DF,$6F,$85,$E9
	.dc.b	$1C,$70,$EE,$F3,$3B,$9D,$70,$EE,$F2,$3F,$73,$87,$81,$F1,$57,$B8
	.dc.b	$8F,$97,$E1,$7B,$BC,$7B,$D5,$EF,$8F,$11,$E1,$3D,$99,$78,$33,$F8
	.dc.b	$D1,$E0,$C7,$CA,$F3,$A7,$C2,$7C,$EF,$83,$93,$F1,$63,$E8,$FC,$24
	.dc.b	$8F,$90,$FD,$4F,$85,$FA,$BF,$0B,$F5,$BE,$16,$57,$CD,$8F,$B3,$72
	.dc.b	$3A,$1F,$47,$E1,$7A,$3E,$3C,$2E,$9E,$8D,$BD,$F9,$76,$AF,$B1,$97
	.dc.b	$67,$D5,$97,$E1,$BE,$16,$47,$DD,$7F,$13,$F0,$9F,$8A,$F8,$59,$3F
	.dc.b	$81,$7F,$1B,$F0,$BF,$8E,$F8,$5F,$C7,$FC,$2C,$AF,$C3,$47,$27,$A1
	.dc.b	$0F,$95,$7D,$1F,$99,$77,$5F,$57,$13,$1E,$D5,$F9,$1C,$7B,$3F,$4F
	.dc.b	$1F,$CF,$FC,$2C,$8F,$CB,$3F,$A2,$F8,$4F,$D1,$FC,$2F,$F3,$8B,$FF
	.dc.b	$17,$65,$F1,$DB,$D7,$FA,$49,$D7,$C2,$CC,$7F,$95,$57,$F8,$AD,$BF
	.dc.b	$0B,$D5,$EE,$43,$F0,$D8,$F8,$1E,$38,$78,$4C,$3F,$95,$AE,$8D,$BF
	.dc.b	$03,$9B,$EC,$F4,$3F,$65,$7E,$BF,$83,$E7,$97,$AF,$1E,$D5,$FB,$7C
	.dc.b	$B9,$CF,$BF,$1B,$5F,$E8,$1C,$BF,$C8,$57,$FD,$99,$7C,$17,$C7,$F8
	.dc.b	$49,$3F,$0A,$3E,$57,$C2,$48,$F8,$AF,$CD,$F8,$7F,$9D,$F0,$BF,$3F
	.dc.b	$E1,$65,$7C,$A8,$FA,$57,$23,$A1,$F3,$BE,$13,$A3,$F1,$70,$BA,$FD
	.dc.b	$7C,$7A,$CF,$D4,$C7,$AB,$F2,$F1,$FB,$DF,$0B,$23,$EC,$BF,$81,$F8
	.dc.b	$5F,$C1,$7C,$2F,$E0,$FE,$16,$57,$DE,$8F,$C3,$7C,$2F,$E1,$FE,$12
	.dc.b	$4F,$E0,$A3,$93,$D0,$87,$DB,$BE,$AF,$E2,$2F,$AB,$F7,$EF,$AB,$F8
	.dc.b	$CB,$EA,$FE,$0E,$FF,$2D,$F0,$9F,$97,$F8,$49,$3F,$92,$8F,$CD,$7C
	.dc.b	$24,$8F,$CA,$BF,$E1,$4F,$B4,$FF,$09,$73,$7F,$AF,$5F,$F0,$F9,$3E
	.dc.b	$0B,$5F,$05,$DA,$FA,$10,$FE,$51,$2F,$85,$ED,$1F,$97,$C2,$F2,$7F
	.dc.b	$0F,$AE,$F3,$FE,$2E,$7B,$3F,$E1,$17,$C3,$7E,$5B,$F0,$4F,$E8,$72
	.dc.b	$C3,$EA,$21,$DC,$E5,$F0,$BE,$DC,$A4,$75,$7D,$6F,$7F,$81,$F8,$5F
	.dc.b	$0B,$23,$07,$8F,$E0,$FE,$37,$C0,$C8,$C1,$E7,$F8,$2F,$94,$47,$D5
	.dc.b	$F0,$5F,$37,$09,$1D,$0F,$8F,$F0,$5F,$47,$0E,$91,$F2,$B0,$C3,$D4
	.dc.b	$87,$CE,$C3,$EC,$61,$F6,$7E,$0F,$ED,$7C,$1C,$AF,$95,$1F,$77,$E0
	.dc.b	$FE,$F7,$C1,$49,$F9,$D1,$F8,$1F,$83,$FC,$17,$C1,$49,$FA,$51,$D5
	.dc.b	$FA,$98,$75,$7E,$FE,$1D,$5F,$B1,$87,$57,$F0,$78,$75,$7E,$E6,$1D
	.dc.b	$5F,$C3,$E1,$F9,$1F,$81,$91,$F8,$17,$FC,$CF,$F5,$6F,$81,$B3,$8F
	.dc.b	$82,$8F,$F9,$5E,$7F,$C3,$1D,$5F,$D6,$DF,$F0,$04,$7F,$7B,$5F,$01
	.dc.b	$D0,$EB,$F0,$3D,$B0,$F0,$D7,$E6,$B1,$C3,$D4,$EE,$73,$C7,$C6,$F8
	.dc.b	$6B,$A3,$E3,$BE,$AF,$96,$F9,$BE,$37,$DB,$F0,$5C,$A3,$A1,$EB,$BC
	.dc.b	$65,$EB,$C3,$AC,$FA,$F0,$EA,$FB,$7E,$07,$E3,$91,$F8,$9F,$05,$F2
	.dc.b	$BE,$06,$47,$C7,$7E,$6F,$C0,$7C,$EF,$7E,$4F,$CB,$7E,$8F,$C0,$7D
	.dc.b	$2B,$E9,$3D,$CE,$A7,$32,$1F,$3E,$FA,$BC,$C8,$7D,$3A,$EA,$FD,$9F
	.dc.b	$82,$FB,$FF,$03,$23,$07,$DD,$F8,$1F,$C1,$FC,$0C,$8C,$1F,$81,$F8
	.dc.b	$0F,$C3,$91,$FC,$27,$C0,$7E,$2B,$09,$1D,$0F,$C2,$7B,$5F,$8E,$C3
	.dc.b	$A4,$7E,$1F,$0C,$3D,$48,$7E,$2F,$0F,$CA,$E1,$F9,$6F,$7B,$F2,$FF
	.dc.b	$01,$2B,$F2,$91,$F9,$AF,$7A,$47,$E5,$DF,$F6,$1F,$D6,$FF,$82,$A3
	.dc.b	$FD,$9E,$FF,$7C,$E6,$FF,$1A,$3F,$DB,$BE,$0C,$D3,$5E,$8A,$BF,$DA
	.dc.b	$FD,$AC,$7C,$91,$E2,$3F,$D7,$94,$F8,$1F,$1C,$7E,$97,$2B,$97,$83
	.dc.b	$1F,$2F,$C0,$F9,$BE,$0A,$5F,$28,$F4,$E1,$DA,$3F,$5D,$96,$1F,$71
	.dc.b	$0E,$C7,$8A,$FA,$BE,$4B,$EA,$FF,$9C,$72,$F6,$BF,$13,$E0,$7E,$39
	.dc.b	$1F,$5F,$B9,$F2,$BD,$FF,$97,$EF,$49,$F7,$C7,$CE,$F7,$E4,$7C,$27
	.dc.b	$E8,$FC,$0F,$38,$F0,$61,$77,$FA,$38,$F5,$9F,$93,$8F,$57,$CD,$8F
	.dc.b	$47,$D1,$8D,$BF,$B3,$8F,$DE,$F8,$09,$5F,$49,$FC,$0F,$BF,$F8,$2F
	.dc.b	$80,$93,$F5,$9F,$C2,$FB,$FF,$86,$F8,$09,$3F,$69,$FC,$4F,$C4,$72
	.dc.b	$8E,$84,$3E,$55,$F4,$7E,$65,$DC,$BF,$1B,$87,$49,$FA,$18,$59,$FC
	.dc.b	$8E,$1F,$96,$F7,$3F,$2F,$F0,$32,$FF,$17,$1F,$9A,$F8,$1F,$C9,$47
	.dc.b	$FA,$A7,$E1,$BF,$BE,$F6,$1B,$CA,$AD,$E5,$FE,$91,$97,$BD,$8E,$8F
	.dc.b	$CE,$61,$C6,$31,$8A,$30,$FE,$86,$F9,$7C,$0F,$9E,$3F,$3F,$A7,$9D
	.dc.b	$FC,$66,$9E,$77,$86,$3D,$B2,$F3,$9E,$6C,$FB,$E5,$DA,$BD,$99,$E2
	.dc.b	$BC,$F1,$C8,$F2,$2C,$B5,$F0,$3E,$39,$F0,$9B,$BF,$89,$1E,$38,$F9
	.dc.b	$1F,$03,$D4,$C6,$BF,$17,$1F,$01,$27,$96,$3D,$DB,$F0,$FE,$52,$3A
	.dc.b	$F8,$2F,$47,$C1,$7A,$7E,$06,$4F,$28,$F6,$7C,$0F,$B7,$E0,$64,$F6
	.dc.b	$8F,$81,$72,$3A,$1E,$2F,$81,$EA,$F9,$30,$EA,$FE,$0F,$0E,$AF,$A3
	.dc.b	$0E,$AF,$E1,$F0,$F9,$9F,$05,$F3,$7E,$06,$4F,$BA,$3E,$87,$C1,$7D
	.dc.b	$1F,$81,$93,$F0,$A3,$EA,$7C,$17,$D5,$F8,$19,$3F,$1A,$3E,$C7,$C1
	.dc.b	$76,$8E,$87,$CA,$7A,$BF,$32,$FA,$BF,$9B,$BE,$AF,$D0,$BE,$AF,$E7
	.dc.b	$EF,$F0,$9C,$32,$3E,$AB,$F8,$7F,$82,$FC,$47,$C3,$C9,$FB,$2F,$E3
	.dc.b	$BE,$CB,$FC,$33,$D9,$FA,$05,$F0,$59,$F8,$29,$7F,$89,$9C,$BF,$82
	.dc.b	$C3,$B3,$4D,$7C,$17,$77,$F6,$98,$DE,$AF,$F8,$01,$7E,$09,$FD,$63
	.dc.b	$D1,$46,$B0,$F0,$C6,$32,$F3,$65,$E9,$C7,$C3,$3F,$C0,$72,$C3,$EE
	.dc.b	$21,$E2,$3D,$47,$C1,$9E,$B8,$77,$8F,$74,$7C,$5F,$84,$F8,$DF,$09
	.dc.b	$27,$E1,$47,$C9,$F8,$49,$1F,$15,$F9,$9F,$05,$F3,$7E,$0B,$E7,$7C
	.dc.b	$24,$AF,$95,$1F,$47,$E0,$FB,$47,$43,$E7,$3D,$1F,$80,$7C,$6B,$C6
	.dc.b	$7F,$2B,$0E,$D3,$F5,$70,$EC,$F0,$F1,$1C,$2C,$BF,$01,$F0,$9F,$81
	.dc.b	$F8,$39,$7F,$76,$3F,$09,$F0,$9F,$85,$F8,$39,$3F,$81,$8F,$C4,$7C
	.dc.b	$1F,$E2,$7E,$0E,$4F,$E1,$63,$F1,$9F,$07,$DA,$3A,$1F,$89,$7A,$BF
	.dc.b	$8B,$BE,$AF,$9F,$36,$1D,$A3,$F2,$18,$76,$75,$87,$E6,$7E,$12,$47
	.dc.b	$E5,$5F,$CE,$7C,$27,$E7,$7E,$02,$4F,$E6,$5F,$D0,$7C,$1F,$F2,$EF
	.dc.b	$AE,$FF,$11,$70,$9F,$F0,$0B,$7E,$13,$5D,$BD,$30,$EC,$43,$D1,$8F
	.dc.b	$67,$D5,$8D,$E6,$FC,$FE,$5C,$A7,$F9,$96,$5D,$89,$87,$8F,$E0,$FB
	.dc.b	$4F,$43,$6F,$38,$FD,$B6,$16,$7F,$53,$7D,$5F,$C0,$5F,$5C,$B1,$8B
	.dc.b	$0F,$EB,$B3,$FE,$33,$AF,$1F,$C2,$7C,$33,$E2,$7C,$47,$C5,$F8,$59
	.dc.b	$3E,$87,$E4,$7C,$47,$C9,$F8,$4F,$95,$F0,$B2,$BD,$B1,$C9,$F7,$DD
	.dc.b	$7C,$BF,$0B,$D1,$F3,$E1,$74,$F3,$47,$F6,$FC,$D9,$77,$8F,$93,$97
	.dc.b	$77,$DF,$97,$DB,$F8,$59,$1F,$39,$FB,$DF,$0D,$F7,$FE,$1A,$4F,$D2
	.dc.b	$7F,$05,$F0,$DF,$83,$F8,$69,$3F,$59,$FC,$37,$C3,$72,$8F,$B3,$75
	.dc.b	$F9,$FF,$0D,$D1,$FA,$38,$5D,$BD,$7D,$3F,$8D,$CB,$AD,$7D,$6C,$AD
	.dc.b	$6B,$76,$7F,$95,$F8,$6F,$CB,$7C,$2C,$C3,$F1,$31,$F9,$9F,$85,$FC
	.dc.b	$D7,$C2,$C9,$FC,$6C,$7F,$82,$BF,$6C,$FF,$11,$9F,$E0,$88,$FE,$08
	.dc.b	$DC,$BC,$AE,$BE,$1B,$C1,$3D,$8F,$F0,$44,$7F,$7F,$7F,$1B,$5D,$DF
	.dc.b	$C7,$DF,$8D,$DF,$C2,$F8,$7E,$16,$46,$0E,$78,$79,$48,$F6,$F8,$6E
	.dc.b	$4F,$43,$F4,$F7,$8C,$BC,$F8,$74,$9F,$5E,$1E,$F8,$F8,$1F,$0D,$F0
	.dc.b	$7E,$1A,$57,$AA,$3E,$27,$C2,$48,$F6,$BF,$1F,$E1,$7E,$47,$C2,$7C
	.dc.b	$9F,$85,$95,$F0,$A3,$AB,$CC,$87,$CC,$F8,$6E,$8F,$CE,$BB,$A7,$CE
	.dc.b	$C3,$AC,$FC,$CC,$3A,$BF,$53,$0F,$B1,$F0,$BF,$67,$E1,$A4,$FD,$28
	.dc.b	$FB,$9F,$09,$F7,$7E,$16,$4F,$D6,$8F,$C0,$7C,$24,$8F,$B2,$FE,$0F
	.dc.b	$E1,$7A,$C7,$32,$1F,$84,$F8,$6E,$AF,$E0,$2F,$AB,$F8,$8B,$E8,$FE
	.dc.b	$2A,$EE,$BF,$8A,$C3,$F2,$1F,$0D,$F9,$1F,$83,$97,$F8,$A8,$FC,$A7
	.dc.b	$C3,$7F,$97,$BB,$7F,$C0,$5E,$11,$BD,$A9,$D7,$C2,$4C,$3F,$CB,$55
	.dc.b	$FE,$07,$CF,$F5,$AB,$E1,$BC,$33,$DC,$87,$F6,$38,$7C,$37,$8A,$39
	.dc.b	$9E,$33,$06,$EF,$C0,$D3,$C3,$5E,$99,$EF,$1F,$AC,$CF,$0F,$EA,$31
	.dc.b	$E8,$F6,$30,$7A,$AF,$AE,$3E,$A7,$E2,$61,$F1,$7E,$16,$57,$C0,$7E
	.dc.b	$47,$C3,$7C,$9F,$85,$F9,$5F,$0B,$2B,$E2,$C7,$CD,$F8,$49,$1F,$21
	.dc.b	$FA,$1F,$0B,$CA,$3E,$5D,$D7,$E3,$7C,$37,$57,$E7,$E1,$D5,$F9,$58
	.dc.b	$74,$7E,$66,$17,$4F,$A7,$8F,$DD,$F8,$59,$5F,$61,$FC,$07,$C3,$7E
	.dc.b	$07,$E0,$E4,$FD,$C7,$F0,$9F,$0B,$F8,$5F,$83,$93,$F8,$07,$F1,$1F
	.dc.b	$09,$CA,$3A,$10,$FB,$D7,$D1,$FC,$05,$DC,$BF,$0F,$87,$49,$FC,$26
	.dc.b	$16,$7F,$17,$87,$E5,$7E,$1A,$57,$E4,$1F,$CC,$7C,$2F,$E6,$7E,$17
	.dc.b	$F3,$5F,$09,$2B,$F2,$B1,$F9,$DF,$84,$FF,$5D,$FE,$DB,$FE,$0E,$3F
	.dc.b	$C1,$FD,$88,$E7,$5F,$9C,$C2,$BF,$E3,$9A,$FF,$09,$D2,$FB,$4F,$E6
	.dc.b	$F0,$F1,$3F,$A4,$C3,$BF,$C2,$48,$C1,$FE,$C1,$C3,$9E,$3D,$B4,$FE
	.dc.b	$AB,$1D,$B1,$EB,$1F,$AC,$C7,$AD,$FE,$A7,$1E,$8F,$97,$1F,$73,$D7
	.dc.b	$E1,$BD,$E4,$7C,$1F,$0B,$F0,$BE,$17,$E1,$FC,$24,$9F,$24,$7C,$6F
	.dc.b	$84,$91,$E7,$7E,$4F,$C1,$FC,$AB,$E9,$3F,$17,$0C,$3C,$8F,$76,$1D
	.dc.b	$5F,$95,$87,$47,$E6,$61,$74,$F8,$38,$FD,$6F,$84,$FA,$FF,$0B,$2F
	.dc.b	$E4,$47,$DA,$F8,$59,$1F,$2D,$FB,$BF,$0B,$F7,$BE,$13,$EF,$FC,$24
	.dc.b	$AF,$A1,$1F,$82,$F8,$4E,$D1,$D0,$FA,$8F,$47,$F0,$17,$73,$FA,$B8
	.dc.b	$75,$9F,$B5,$87,$57,$F0,$D8,$7E,$3B,$E0,$FF,$1F,$F0,$B2,$7F,$01
	.dc.b	$1F,$92,$F8,$4F,$C9,$F5,$49,$FC,$24,$7E,$5B,$E1,$24,$7E,$1D,$FC
	.dc.b	$CF,$C1,$22,$69,$FE,$1C,$FD,$B9,$F0,$5E,$19,$EE,$6B,$A7,$87,$A6
	.dc.b	$AD,$CF,$80,$FC,$A6,$7E,$37,$F4,$19,$F8,$9F,$D1,$67,$72,$FC,$BE
	.dc.b	$7E,$39,$FF,$4F,$61,$D1,$B9,$E6,$BC,$CD,$CF,$C7,$5F,$E2,$17,$CB
	.dc.b	$DC,$CB,$C5,$7F,$A4,$CF,$E4,$63,$F0,$C8,$79,$0F,$D6,$65,$D5,$69
	.dc.b	$EF,$F8,$5C,$3F,$07,$E1,$6E,$86,$CE,$3E,$0E,$5F,$C9,$39,$D2,$38
	.dc.b	$FC,$27,$2F,$83,$95,$BC,$7E,$74,$76,$F8,$3E,$B3,$9C,$3F,$C7,$5A
	.dc.b	$FD,$9C,$BC,$7F,$05,$E0,$AF,$97,$9D,$AF,$BD,$9F,$69,$F9,$F9,$59
	.dc.b	$FC,$16,$5E,$CF,$83,$95,$DD,$F7,$FC,$27,$C0,$F8,$3F,$83,$F0,$92
	.dc.b	$BC,$91,$F1,$3E,$0A,$47,$9D,$F8,$FF,$07,$CA,$3A,$10,$FC,$0D,$F5
	.dc.b	$7D,$D7,$D5,$FC,$2D,$F4,$7F,$0F,$77,$2F,$CD,$61,$F5,$3E,$0F,$EA
	.dc.b	$FC,$1C,$BF,$91,$1F,$63,$E1,$3E,$CF,$C0,$C9,$F9,$91,$F7,$3E,$13
	.dc.b	$EE,$FC,$14,$9F,$A1,$1D,$5E,$64,$3F,$DE,$DF,$B7,$FF,$88,$B9,$D7
	.dc.b	$D6,$7F,$3B,$87,$85,$E6,$7D,$BC,$3B,$3F,$A2,$C3,$F1,$F8,$73,$C7
	.dc.b	$C1,$3D,$8E,$47,$43,$F0,$97,$8D,$3F,$79,$7D,$1F,$DF,$61,$ED,$AF
	.dc.b	$C7,$61,$FE,$31,$EE,$FD,$EB,$63,$F4,$2D,$E9,$3F,$E3,$58,$E3,$F0
	.dc.b	$9E,$6F,$82,$91,$83,$AF,$C1,$7A,$88,$F8,$3E,$03,$DB,$F0,$7E,$EF
	.dc.b	$84,$93,$E4,$8F,$83,$87,$48,$F9,$B8,$61,$EA,$43,$E8,$E1,$F2,$30
	.dc.b	$F9,$3F,$07,$F2,$BE,$0E,$57,$C0,$8F,$9B,$F0,$7F,$3B,$E1,$24,$FC
	.dc.b	$48,$FA,$3F,$09,$F4,$BE,$12,$4F,$C8,$8F,$AB,$F0,$9D,$A3,$A1,$F3
	.dc.b	$1E,$AF,$CE,$BE,$AF,$E1,$AF,$AB,$F4,$AF,$AB,$F8,$AB,$FC,$0F,$C2
	.dc.b	$7E,$0B,$E1,$24,$FD,$88,$FC,$2F,$C2,$48,$FB,$6F,$E2,$3E,$13,$F1
	.dc.b	$3F,$09,$27,$EF,$BF,$8C,$F8,$4E,$B1,$CC,$87,$E6,$BE,$13,$A3,$F9
	.dc.b	$CB,$BD,$7F,$EE,$2E,$86,$3D,$6B,$F4,$38,$DA,$D6,$3F,$9A,$C6,$57
	.dc.b	$E4,$1F,$CE,$FC,$27,$4E,$BF,$C8,$E7,$6B,$F2,$39,$73,$9F,$CA,$E5
	.dc.b	$67,$F2,$B8,$FF,$34,$13,$49,$1F,$CA,$E1,$F1,$1B,$F8,$49,$39,$79
	.dc.b	$7C,$2F,$3F,$84,$93,$C5,$ED,$F0,$BD,$EF,$AC,$B7,$F0,$BF,$D2,$31
	.dc.b	$AE,$72,$FE,$D3,$1E,$28,$F0,$10,$D1,$C7,$2F,$1B,$E1,$21,$AF,$03
	.dc.b	$FC,$95,$97,$92,$3D,$DF,$0B,$F0,$C8,$FC,$1F,$85,$F8,$DF,$09,$F1
	.dc.b	$FE,$16,$4F,$C5,$8F,$95,$F0,$9F,$2F,$E1,$24,$FC,$98,$F9,$D7,$D2
	.dc.b	$3C,$5A,$31,$CC,$73,$3E,$86,$3D,$9D,$E3,$D9,$FA,$98,$F6,$79,$E3
	.dc.b	$F6,$FE,$0E,$47,$D9,$7E,$F7,$C2,$7D,$FF,$85,$FC,$07,$C2,$4A,$FB
	.dc.b	$D1,$F8,$3F,$85,$91,$F8,$17,$F0,$DF,$05,$CA,$3A,$10,$F5,$DF,$57
	.dc.b	$F1,$17,$D5,$F7,$DF,$47,$E0,$DD,$E4,$FC,$0F,$17,$F9,$F7,$1E,$D5
	.dc.b	$D0,$FC,$96,$3F,$97,$F8,$2E,$D1,$F2,$B2,$ED,$87,$2E,$16,$7D,$79
	.dc.b	$7F,$13,$A5,$C3,$F1,$D9,$FE,$87,$E1,$7F,$45,$F0,$72,$7F,$41,$1F
	.dc.b	$E1,$8F,$DC,$3F,$C5,$99,$F9,$9F,$F1,$44,$FF,$87,$9B,$B9,$DE,$5F
	.dc.b	$E4,$07,$FC,$1E,$DF,$83,$E8,$75,$F8,$4E,$DF,$07,$E3,$C3,$C2,$7F
	.dc.b	$86,$EF,$FC,$40,$DF,$86,$F2,$7C,$44,$8C,$1B,$C3,$AB,$4E,$58,$7A
	.dc.b	$FE,$17,$D9,$F0,$B2,$7B,$47,$BF,$E1,$7E,$07,$C3,$49,$F1,$47,$C3
	.dc.b	$F8,$6F,$89,$F0,$D2,$7C,$D1,$D5,$E6,$43,$F6,$1A,$BE,$17,$B4,$7B
	.dc.b	$70,$EC,$EB,$0E,$CF,$C1,$C3,$B3,$CB,$0F,$A9,$F0,$D2,$3E,$33,$F5
	.dc.b	$FE,$1B,$EC,$7C,$37,$D9,$F8,$39,$5F,$2E,$3E,$E7,$C3,$48,$F9,$CF
	.dc.b	$DF,$F8,$5E,$51,$D0,$87,$AA,$FA,$BF,$56,$FA,$BE,$EB,$E8,$FC,$0B
	.dc.b	$BD,$1F,$89,$C3,$AC,$FC,$4C,$7D,$4F,$53,$CC,$75,$21,$D4,$F3,$90
	.dc.b	$EC,$7E,$1E,$3F,$31,$F0,$B2,$3F,$14,$FE,$6F,$E1,$7F,$39,$F0,$B2
	.dc.b	$7F,$1C,$FF,$9D,$BF,$71,$FF,$46,$47,$F2,$77,$FC,$34,$7F,$86,$5C
	.dc.b	$7C,$2E,$FE,$1B,$B4,$7D,$EC,$2E,$7F,$A7,$C7,$B4,$FE,$0B,$1B,$3F
	.dc.b	$C0,$31,$EE,$DC,$3C,$58,$CB,$FF,$3E,$57,$8B,$E1,$BC,$DF,$0D,$27
	.dc.b	$31,$E9,$F8,$69,$1C,$1F,$67,$C3,$7B,$7E,$17,$DD,$F0,$D2,$BA,$C7
	.dc.b	$C1,$B9,$1D,$0F,$0F,$C2,$74,$7E,$26,$17,$5F,$56,$3D,$67,$CF,$8F
	.dc.b	$57,$E5,$63,$F3,$7E,$17,$E7,$7C,$24,$9F,$6C,$7D,$1F,$85,$91,$F0
	.dc.b	$1F,$A9,$F0,$DF,$57,$E1,$64,$FC,$47,$EC,$7C,$37,$58,$F8,$F8,$75
	.dc.b	$7E,$D6,$1D,$1F,$B9,$85,$D3,$E9,$63,$D2,$7F,$01,$8D,$9F,$AD,$8F
	.dc.b	$59,$E6,$7A,$0F,$AB,$F0,$7C,$FC,$BF,$F6,$FF,$1B,$2C,$57,$FB,$12
	.dc.b	$19,$75,$AD,$E5,$6B,$58,$FE,$4F,$E0,$FF,$29,$F0,$72,$7F,$07,$1F
	.dc.b	$97,$F5,$3F,$31,$F0,$B2,$7F,$0F,$1F,$9B,$F8,$59,$1F,$8A,$7F,$DA
	.dc.b	$7F,$B9,$7F,$83,$B8,$3F,$25,$87,$5A,$FD,$06,$3E,$17,$99,$F9,$2C
	.dc.b	$7B,$3F,$A5,$C7,$AB,$FA,$7C,$6E,$1F,$D7,$31,$ED,$A2,$BF,$DC,$B1
	.dc.b	$FE,$1B,$7F,$C2,$8D,$9E,$1F,$09,$E6,$F8,$4F,$3F,$C3,$4B,$FF,$6F
	.dc.b	$5A,$3D,$7F,$0B,$EC,$F8,$69,$5A,$7D,$FF,$0B,$CE,$38,$E3,$F0,$8C
	.dc.b	$3E,$2B,$EC,$D3,$BD,$F4,$7C,$F7,$72,$F7,$61,$F3,$3E,$1B,$E6,$FC
	.dc.b	$2C,$BF,$3C,$7D,$0F,$86,$FA,$3F,$0B,$27,$D7,$1F,$53,$E1,$A4,$7B
	.dc.b	$9F,$AF,$F0,$DD,$63,$E0,$E1,$D5,$F9,$38,$75,$7E,$2E,$1D,$5F,$9B
	.dc.b	$87,$47,$E7,$E1,$72,$FA,$D8,$F5,$9E,$67,$A0,$F9,$FF,$0D,$D5,$FA
	.dc.b	$F8,$75,$AF,$BD,$87,$3E,$2F,$BD,$EF,$B8,$99,$E7,$1F,$C0,$E7,$F9
	.dc.b	$2F,$84,$91,$F7,$1F,$CA,$FC,$2F,$E5,$BE,$17,$F2,$FF,$0B,$2B,$F0
	.dc.b	$31,$F9,$AF,$86,$91,$F8,$47,$F3,$BF,$0B,$D2,$3F,$1B,$85,$DF,$F2
	.dc.b	$98,$F5,$9F,$C5,$E3,$D5,$FC,$9E,$3D,$1F,$CA,$E3,$6F,$F3,$98,$FE
	.dc.b	$AB,$E1,$BF,$39,$5F,$E6,$DF,$4B,$FC,$42,$B7,$46,$1F,$89,$3C,$2F
	.dc.b	$FC,$83,$3F,$E2,$A3,$FC,$4F,$5C,$5B,$5C,$BE,$13,$87,$0F,$CE,$97
	.dc.b	$C1,$E2,$F8,$36,$FC,$2D,$9F,$F1,$5C,$FE,$C3,$1F,$53,$FA,$C7,$5F
	.dc.b	$05,$BC,$78,$7C,$24,$AF,$DD,$67,$E8,$9E,$98,$C9,$CE,$7E,$58,$F0
	.dc.b	$7C,$37,$87,$E1,$65,$68,$F7,$3E,$5F,$82,$F3,$7C,$1F,$9F,$E0,$A5
	.dc.b	$6F,$83,$E5,$CF,$AF,$E0,$E4,$E7,$85,$1E,$FF,$84,$E9,$39,$CB,$C2
	.dc.b	$FD,$7C,$BB,$BC,$72,$EE,$FD,$BC,$7A,$BF,$77,$0B,$BF,$D8,$C7,$AC
	.dc.b	$FE,$07,$1E,$B3,$F7,$31,$EB,$3F,$85,$C7,$AC,$FE,$03,$1E,$7B,$7D
	.dc.b	$BC,$AD,$FE,$07,$2F,$B1,$F0,$92,$3D,$CF,$DB,$F8,$3F,$B9,$F0,$9F
	.dc.b	$77,$E1,$25,$7C,$38,$FC,$07,$C1,$C8,$F8,$CF,$E0,$FE,$0F,$A4,$7E
	.dc.b	$67,$0B,$AF,$E5,$31,$EB,$3F,$37,$1E,$AF,$E7,$F1,$E8,$FE,$87,$1B
	.dc.b	$7F,$9C,$C7,$F2,$1C,$3F,$91,$F8,$39,$7F,$5E,$3F,$2D,$FB,$9F,$F8
	.dc.b	$8B,$7F,$E1,$2B,$E1,$32,$6A,$B7,$F0,$9C,$3E,$0E,$47,$E1,$6B,$FC
	.dc.b	$6D,$9F,$EB,$31,$F0,$CF,$73,$5F,$09,$CD,$E1,$8F,$67,$91,$D8,$E9
	.dc.b	$87,$67,$FD,$D5,$87,$6C,$BC,$FF,$07,$EA,$F8,$39,$5E,$38,$F6,$FC
	.dc.b	$14,$8F,$33,$F0,$3E,$0F,$E0,$FC,$24,$9F,$53,$F1,$3E,$0F,$E2,$DC
	.dc.b	$9E,$87,$BF,$E1,$3A,$3E,$4C,$2F,$07,$FB,$33,$46,$5D,$6B,$D3,$95
	.dc.b	$AD,$65,$D6,$7D,$B9,$59,$E5,$97,$4D,$3E,$AE,$76,$FF,$8F,$67,$D9
	.dc.b	$F4,$90,$EE,$7D,$5C,$7E,$FF,$C1,$C8,$FB,$0F,$E0,$BE,$0B,$F0,$7F
	.dc.b	$05,$27,$EE,$3F,$86,$F8,$2F,$C3,$FC,$27,$E2,$3E,$0E,$57,$E0,$63
	.dc.b	$93,$D0,$87,$D8,$BE,$8F,$DA,$BB,$B7,$C7,$C3,$AC,$FE,$2F,$0E,$AF
	.dc.b	$E0,$30,$FC,$BF,$C1,$C8,$FC,$83,$F9,$AF,$82,$FC,$DF,$C1,$C9,$FC
	.dc.b	$A3,$F9,$EF,$83,$FE,$69,$FB,$A7,$F8,$92,$7F,$A3,$3F,$E2,$1E,$2F
	.dc.b	$ED,$2D,$8C,$C7,$57,$B9,$0F,$C9,$DF,$77,$F2,$B7,$74,$FC,$16,$1E
	.dc.b	$0B,$F0,$FC,$1C,$BF,$E5,$98,$F8,$FC,$B3,$FE,$2D,$8C,$FC,$27,$A0
	.dc.b	$F4,$FC,$27,$AB,$E1,$3D,$7F,$09,$2B,$9C,$7B,$BE,$13,$DF,$F0,$72
	.dc.b	$7B,$C7,$27,$A9,$06,$CD,$7E,$1E,$3D,$E7,$CF,$8F,$77,$73,$F2,$31
	.dc.b	$F0,$47,$B3,$1F,$03,$A3,$E7,$E3,$E1,$71,$A7,$C0,$C7,$19,$78,$67
	.dc.b	$EA,$FC,$27,$85,$DE,$AD,$3C,$55,$BC,$3E,$4E,$7E,$2B,$DE,$7E,$2A
	.dc.b	$D6,$0C,$FF,$07,$F0,$52,$FE,$9B,$F8,$6F,$83,$FC,$3F,$C1,$49,$FA
	.dc.b	$EF,$E2,$BE,$0B,$F1,$7F,$09,$F8,$CF,$83,$95,$F7,$23,$A3,$E8,$C2
	.dc.b	$E7,$E2,$C7,$A4,$FB,$31,$B3,$E6,$C7,$AC,$FE,$1B,$1E,$AF,$C2,$C7
	.dc.b	$F3,$7F,$07,$F9,$CF,$82,$93,$F8,$C8,$FC,$FF,$C1,$FE,$83,$E0,$E4
	.dc.b	$FE,$42,$3F,$47,$F0,$7F,$A4,$F8,$39,$3F,$94,$8E,$AF,$E5,$B0,$EA
	.dc.b	$FD,$2C,$3A,$BF,$9A,$C3,$AB,$F5,$B0,$EA,$FE,$7B,$0E,$AF,$DA,$C3
	.dc.b	$FC,$4D,$FB,$AF,$F8,$D4,$9F,$13,$8F,$80,$CF,$C2,$6B,$E0,$E6,$1F
	.dc.b	$E1,$3B,$FF,$1A,$9F,$E2,$CE,$1F,$C1,$2F,$85,$EA,$76,$F8,$4E,$F9
	.dc.b	$4B,$E4,$F7,$7A,$63,$D9,$FC,$AD,$FF,$8F,$9C,$9E,$1A,$EA,$F8,$EF
	.dc.b	$B3,$E5,$CB,$B3,$CB,$0E,$CF,$4C,$3B,$63,$E8,$CB,$DF,$3D,$4E,$84
	.dc.b	$3C,$78,$FC,$8F,$84,$F9,$3F,$07,$27,$E2,$47,$CC,$F8,$49,$1F,$1D
	.dc.b	$F9,$FF,$09,$F4,$3E,$13,$E8,$FC,$44,$AF,$99,$1D,$5E,$64,$3E,$1F
	.dc.b	$C2,$F4,$7E,$2D,$DD,$3E,$8E,$1D,$67,$EB,$61,$D5,$F9,$78,$7E,$03
	.dc.b	$E1,$7F,$03,$F0,$92,$7E,$E4,$7E,$13,$E1,$7F,$0B,$F0,$B2,$7F,$01
	.dc.b	$1F,$88,$F8,$59,$1F,$83,$7F,$17,$F0,$7D,$63,$99,$0F,$B7,$F0,$BD
	.dc.b	$5F,$C5,$5F,$57,$EF,$DF,$47,$F0,$37,$75,$FC,$6E,$1F,$E7,$DE,$9F
	.dc.b	$F1,$45,$7F,$A3,$DB,$E4,$9A,$7F,$8A,$E3,$FC,$D4,$FF,$86,$5E,$1F
	.dc.b	$0B,$CC,$E9,$F0,$7D,$7E,$17,$B7,$C0,$CA,$E3,$1E,$1F,$85,$F1,$44
	.dc.b	$3A,$BC,$DE,$F7,$5F,$E5,$F8,$30,$ED,$1E,$4C,$3B,$3A,$C3,$B3,$D0
	.dc.b	$87,$2F,$85,$E9,$A7,$F3,$1C,$3D,$99,$FC,$1C,$BD,$77,$E8,$3B,$10
	.dc.b	$EC,$7C,$4F,$82,$F9,$9F,$0B,$23,$07,$C8,$F8,$5F,$A1,$F0,$72,$30
	.dc.b	$7C,$CF,$85,$FA,$9F,$09,$23,$07,$23,$E0,$E1,$F6,$4F,$B5,$F0,$9F
	.dc.b	$6F,$0E,$93,$F1,$B0,$C3,$C8,$F9,$37,$D5,$F9,$78,$7E,$0F,$E1,$3F
	.dc.b	$09,$F0,$92,$7E,$FC,$7E,$1F,$E0,$FF,$11,$F0,$92,$7F,$07,$1F,$8B
	.dc.b	$F8,$4F,$C6,$7C,$24,$9F,$C3,$C7,$57,$99,$0F,$B7,$F0,$9D,$5F,$C6
	.dc.b	$DF,$57,$EF,$DF,$57,$F2,$37,$D5,$FC,$1D,$FE,$6F,$E1,$3F,$DB,$1F
	.dc.b	$BB,$7F,$84,$CF,$F0,$A7,$2A,$33,$F0,$B2,$FF,$D8,$95,$FE,$27,$E6
	.dc.b	$2F,$85,$E7,$F0,$9D,$3E,$0A,$5F,$08,$F0,$3F,$A4,$C6,$BF,$E0,$4B
	.dc.b	$FE,$FC,$F6,$C7,$C0,$E5,$F3,$E1,$DA,$3C,$78,$76,$79,$61,$D3,$AB
	.dc.b	$F2,$D8,$DF,$85,$53,$3F,$4F,$C2,$76,$BF,$F0,$66,$9E,$D2,$67,$DE
	.dc.b	$FC,$D9,$F7,$AD,$E7,$F2,$BE,$0B,$E5,$FC,$24,$BF,$8F,$1F,$3B,$E1
	.dc.b	$7E,$7F,$C2,$49,$F9,$71,$F4,$BE,$13,$E9,$FC,$24,$9F,$9F,$1D,$5E
	.dc.b	$64,$3E,$47,$C2,$75,$7E,$AD,$F5,$7E,$65,$F5,$7E,$CD,$F5,$7E,$85
	.dc.b	$FE,$0B,$E0,$E4,$7D,$E7,$F0,$BF,$09,$F8,$6F,$83,$93,$F8,$27,$F1
	.dc.b	$3F,$09,$F8,$AF,$83,$93,$F8,$67,$F1,$BF,$09,$CA,$3A,$10,$FC,$05
	.dc.b	$F4,$7F,$05,$77,$87,$E9,$E1,$D2,$7F,$0D,$85,$9F,$AF,$87,$E6,$BA
	.dc.b	$BF,$37,$F0,$92,$FF,$2F,$1F,$80,$7F,$CA,$3F,$BB,$BE,$0B,$18,$7F
	.dc.b	$92,$E3,$13,$F8,$68,$C4,$EB,$AF,$F1,$3D,$6B,$7C,$46,$70,$FC,$64
	.dc.b	$3E,$0F,$15,$BF,$83,$B3,$FE,$0D,$AF,$1D,$7A,$5F,$D6,$E9,$E9,$7F
	.dc.b	$49,$A7,$0C,$E4,$79,$73,$F1,$E1,$E8,$8F,$6E,$7D,$B1,$EF,$94,$9E
	.dc.b	$07,$9A,$3C,$0F,$63,$E3,$9A,$CF,$DD,$F0,$7E,$08,$F9,$39,$5D,$7E
	.dc.b	$46,$3D,$E7,$86,$5E,$D9,$F9,$99,$66,$7E,$8F,$C2,$FC,$2F,$85,$91
	.dc.b	$C2,$3E,$2F,$C2,$7C,$6F,$85,$93,$D2,$3E,$4F,$C2,$48,$EE,$FC,$CF
	.dc.b	$84,$F9,$B7,$D2,$7E,$FE,$18,$79,$1E,$6C,$3A,$BF,$84,$C3,$A3,$F8
	.dc.b	$6C,$2E,$5F,$85,$C7,$EC,$FC,$1F,$DA,$F8,$49,$7F,$02,$3E,$EF,$C2
	.dc.b	$FD,$EF,$84,$93,$F1,$23,$F0,$3F,$0B,$F8,$2F,$83,$93,$F2,$23,$AB
	.dc.b	$CC,$87,$E6,$3E,$13,$AB,$F3,$AF,$AB,$F9,$CB,$EA,$FD,$2B,$EA,$FE
	.dc.b	$82,$FF,$23,$F0,$9F,$92,$F8,$49,$3F,$62,$3F,$2B,$F0,$BF,$98,$FD
	.dc.b	$E3,$FC,$3A,$7F,$87,$B9,$91,$9F,$86,$D7,$C1,$4C,$3F,$11,$7F,$E2
	.dc.b	$17,$F5,$F8,$76,$7F,$63,$8F,$89,$FC,$3E,$1D,$D8,$60,$E7,$87,$57
	.dc.b	$FD,$93,$7F,$E1,$03,$B7,$17,$EE,$70,$E9,$3E,$03,$C7,$F0,$DE,$C7
	.dc.b	$CA,$76,$3F,$81,$E1,$EB,$7D,$39,$77,$BF,$F6,$76,$1D,$BB,$FF,$A4
	.dc.b	$8F,$86,$74,$7F,$01,$8C,$36,$33,$87,$EC,$27,$87,$3E,$79,$F5,$CD
	.dc.b	$2B,$56,$9E,$5C,$FC,$F9,$65,$6A,$3B,$65,$23,$8B,$E1,$C7,$B4,$74
	.dc.b	$C7,$B3,$FE,$F5,$CB,$D5,$8F,$68,$F0,$E3,$D9,$D6,$4C,$BD,$9C,$D2
	.dc.b	$7C,$EF,$BF,$97,$E0,$73,$7C,$1E,$79,$5E,$C8,$F8,$9C,$9F,$17,$9A
	.dc.b	$4F,$C0,$8E,$4F,$42,$1E,$2B,$EA,$FC,$6B,$EA,$F9,$AF,$AB,$F2,$AF
	.dc.b	$AB,$EA,$BF,$A9,$C9,$23,$E7,$BF,$67,$F1,$FF,$F3,$EE,$E7,$F0,$12
	.dc.b	$8C,$73,$EC,$E1,$CD,$C7,$9E,$57,$DF,$BF,$F3,$83,$7E,$1F,$B3,$E1
	.dc.b	$3E,$6E,$38,$78,$61,$D5,$A7,$3C,$3A,$BF,$66,$FA,$B7,$0F,$56,$1C
	.dc.b	$E3,$A9,$EE,$3C,$87,$53,$95,$DE,$8F,$74,$7B,$EF,$FD,$31,$97,$C3
	.dc.b	$C3,$C3,$5A,$8C,$57,$B4,$F8,$1C,$5F,$2F,$E1,$E4,$60,$F8,$9F,$0D
	.dc.b	$F3,$C8,$FC,$7F,$87,$FA,$5F,$0D,$F4,$FE,$1A,$4F,$CB,$8F,$AD,$84
	.dc.b	$8E,$87,$C3,$F8,$5E,$6F,$C5,$C2,$E7,$E7,$D5,$97,$6A,$FA,$D9,$76
	.dc.b	$7E,$66,$5F,$83,$F8,$69,$1F,$6D,$FC,$37,$C3,$FE,$1F,$E1,$BF,$11
	.dc.b	$F0,$F2,$BF,$01,$1F,$8B,$F8,$59,$1F,$83,$7F,$1D,$F0,$DD,$23,$ED
	.dc.b	$61,$75,$F3,$63,$D6,$7F,$15,$8F,$57,$F0,$18,$F4,$7F,$05,$8D,$BF
	.dc.b	$76,$3F,$9B,$F8,$59,$5F,$93,$7F,$3D,$F0,$DF,$9F,$F8,$7F,$D0,$7C
	.dc.b	$34,$AF,$CC,$47,$F1,$BF,$C7,$FF,$07,$1F,$C1,$DB,$8F,$E0,$5B,$CF
	.dc.b	$FC,$91,$7C,$47,$82,$BF,$2D,$95,$BF,$A5,$97,$86,$7F,$4D,$97,$85
	.dc.b	$FC,$E6,$5E,$1C,$A4,$7F,$29,$CB,$BB,$FB,$37,$BB,$FA,$FB,$ED,$3F
	.dc.b	$CD,$B9,$19,$7C,$0A,$F6,$9E,$02,$1E,$03,$B4,$F8,$1F,$80,$FC,$3C
	.dc.b	$FE,$27,$C3,$7C,$5F,$85,$95,$E5,$8F,$91,$F0,$D2,$3D,$0F,$CB,$F8
	.dc.b	$8F,$99,$F0,$D2,$7D,$8F,$CF,$F8,$7E,$B1,$CC,$87,$D1,$F8,$5E,$8F
	.dc.b	$D3,$BB,$C5,$F4,$F0,$E9,$3F,$5F,$0B,$3F,$5F,$0F,$B9,$F0,$DF,$77
	.dc.b	$E1,$65,$FC,$D8,$FC,$07,$C3,$48,$FA,$0F,$E0,$FE,$1F,$F0,$9F,$0F
	.dc.b	$F8,$5F,$87,$95,$F5,$63,$AB,$CC,$87,$E2,$7E,$1F,$A3,$F8,$BB,$B9
	.dc.b	$FE,$2F,$0E,$B3,$F7,$F0,$EA,$FE,$47,$0F,$CA,$7C,$3F,$E5,$7E,$1E
	.dc.b	$4F,$E1,$63,$F3,$1F,$0F,$F9,$9F,$86,$93,$F8,$98,$FC,$E7,$C3,$C8
	.dc.b	$FC,$63,$FD,$E7,$F1,$FF,$F3,$56,$9F,$9D,$C3,$BD,$75,$3F,$BD,$4F
	.dc.b	$FA,$01,$E8,$7E,$57,$0E,$EF,$E9,$F0,$EC,$FE,$A7,$0B,$87,$F2,$9B
	.dc.b	$EF,$D3,$FA,$9C,$7B,$DF,$EB,$DF,$27,$C2,$7B,$23,$B1,$0E,$C7,$A8
	.dc.b	$87,$73,$CF,$F0,$BE,$DF,$88,$91,$EA,$7E,$07,$C3,$7C,$1F,$87,$93
	.dc.b	$EE,$7E,$27,$C3,$FC,$5F,$87,$F8,$DF,$0F,$2B,$E1,$C7,$27,$A1,$0F
	.dc.b	$F0,$76,$BF,$E8,$3D,$7F,$C5,$2B,$0E,$F3,$A7,$BB,$D0,$F9,$F7,$DD
	.dc.b	$E7,$7F,$5F,$E1,$64,$7D,$47,$ED,$7C,$2F,$DB,$F8,$4F,$B9,$F0,$B2
	.dc.b	$BE,$CC,$7D,$FF,$84,$FC,$07,$C2,$C9,$FB,$B1,$C9,$E8,$43,$D7,$7D
	.dc.b	$5F,$C1,$DF,$57,$DF,$7D,$5F,$C3,$DF,$57,$E1,$DF,$E3,$FE,$17,$F2
	.dc.b	$1F,$0B,$27,$F1,$B1,$F9,$3F,$84,$FC,$A7,$C2,$C9,$FC,$8C,$7E,$5F
	.dc.b	$E1,$3F,$31,$F0,$B2,$7F,$2B,$1D,$5F,$CB,$E1,$D5,$FA,$78,$75,$7F
	.dc.b	$37,$87,$57,$EB,$E1,$D5,$FC,$FE,$1D,$5F,$B7,$87,$57,$99,$E8,$3F
	.dc.b	$4B,$F0,$BD,$5F,$C0,$DF,$5E,$BF,$D0,$61,$D6,$7F,$0B,$87,$59,$FD
	.dc.b	$26,$1F,$E6,$BE,$6F,$F1,$99,$FE,$32,$E6,$34,$CF,$C2,$EC,$98,$7F
	.dc.b	$9B,$E3,$FC,$4B,$7F,$E1,$75,$F0,$BD,$7E,$0A,$56,$E3,$C1,$F0,$7E
	.dc.b	$1F,$82,$93,$CE,$3C,$9F,$05,$E5,$F8,$19,$3D,$E3,$D1,$F0,$52,$3C
	.dc.b	$4F,$AF,$E0,$FD,$9F,$07,$27,$CC,$FB,$FE,$0F,$E0,$7C,$17,$C1,$F8
	.dc.b	$39,$5E,$B8,$F8,$9F,$07,$23,$DC,$FC,$7F,$83,$47,$FB,$E7,$F7,$9F
	.dc.b	$D8,$2F,$89,$46,$1B,$E4,$70,$97,$2F,$F9,$0A,$77,$E3,$E6,$31,$7F
	.dc.b	$4A,$BB,$78,$D9,$E9,$F0,$E1,$DA,$31,$3E,$1A,$E0,$F6,$AD,$C3,$C7
	.dc.b	$DD,$F9,$9B,$19,$AF,$67,$8F,$8C,$E4,$7B,$DD,$E7,$C3,$AB,$8E,$72
	.dc.b	$BE,$24,$74,$C6,$47,$C7,$7B,$E1,$D2,$3C,$C7,$A7,$0C,$69,$F0,$B1
	.dc.b	$ED,$3F,$08,$87,$0C,$FE,$7E,$3D,$A3,$19,$FA,$DB,$97,$81,$FB,$1E
	.dc.b	$3F,$C0,$23,$F6,$FC,$6F,$87,$E3,$FC,$4F,$1A,$4F,$DF,$8F,$8F,$E3
	.dc.b	$C8,$FC,$13,$F2,$BC,$6F,$97,$7D,$27,$C7,$86,$1E,$47,$E2,$70,$EA
	.dc.b	$FA,$30,$E8,$FA,$B0,$BB,$7E,$13,$1E,$53,$D0,$F5,$9E,$93,$A9,$F0
	.dc.b	$AF,$AC,$FE,$32,$F9,$F1,$7E,$13,$0B,$7F,$89,$C3,$F0,$7E,$3C,$8F
	.dc.b	$CF,$3F,$86,$F1,$7F,$0F,$E3,$49,$FD,$13,$F8,$AF,$17,$F1,$7E,$37
	.dc.b	$E3,$3C,$79,$5F,$A7,$8E,$4F,$42,$1F,$62,$FA,$3F,$6A,$EE,$BF,$A4
	.dc.b	$C3,$AC,$FE,$C7,$0E,$AF,$E0,$30,$EA,$FE,$0B,$1F,$53,$D4,$F3,$90
	.dc.b	$EC,$7E,$EE,$BA,$BF,$8A,$AE,$B7,$FB,$4B,$FD,$4F,$EF,$5F,$BB,$AF
	.dc.b	$C9,$36,$7F,$78,$DD,$7F,$6E,$FF,$B7,$B9,$4B,$C8,$E6,$4B,$FE,$1B
	.dc.b	$7C,$BE,$1F,$BF,$91,$E0,$F2,$65,$6B,$2E,$B3,$E4,$F8,$8F,$2C,$C3
	.dc.b	$47,$F1,$5C,$BC,$47,$85,$CE,$7E,$4C,$FC,$51,$EC,$CB,$C2,$EA,$99
	.dc.b	$78,$A3,$E0,$E3,$E1,$74,$78,$4F,$8B,$8F,$4C,$FE,$2E,$37,$3F,$8B
	.dc.b	$97,$6B,$F9,$78,$F5,$BF,$97,$8F,$D0,$F2,$65,$78,$DF,$A7,$E5,$7D
	.dc.b	$4F,$26,$4F,$9D,$FA,$FE,$57,$D8,$F2,$A4,$FA,$DF,$B7,$E5,$74,$8F
	.dc.b	$B7,$85,$CB,$ED,$E3,$D2,$7E,$FE,$36,$7E,$FE,$3D,$27,$F0,$78,$D9
	.dc.b	$FC,$1E,$3E,$69,$F4,$10,$FC,$4E,$1C,$F5,$FC,$3E,$36,$7E,$C6,$3C
	.dc.b	$E7,$F1,$78,$D9,$FB,$98,$7F,$7B,$F8,$3F,$EF,$9F,$0B,$27,$FB,$B4
	.dc.b	$7F,$7F,$F8,$7F,$F0,$07,$C4,$49,$FE,$F5,$1F,$D2,$B1,$7E,$43,$FE
	.dc.b	$C2,$69,$97,$07,$FB,$12,$7F,$9C,$C7,$F8,$22,$FC,$51,$E0,$3F,$23
	.dc.b	$1D,$3E,$17,$C5,$1F,$E0,$9C,$3C,$4F,$2F,$19,$8F,$7B,$93,$CB,$1F
	.dc.b	$0F,$C3,$F6,$8E,$19,$5E,$15,$FF,$76,$3C,$13,$EE,$3C,$04,$3C,$18
	.dc.b	$76,$7D,$78,$7B,$A7,$E0,$61,$F0,$7E,$1F,$E1,$7C,$34,$BF,$24,$7C
	.dc.b	$5F,$86,$91,$E7,$7E,$47,$C3,$7C,$9F,$86,$F9,$5F,$0F,$2B,$D9,$1D
	.dc.b	$5F,$76,$1D,$5F,$1E,$1D,$1F,$2E,$17,$5F,$7C,$7C,$0D,$19,$77,$8F
	.dc.b	$91,$97,$77,$DB,$97,$77,$A9,$EA,$3E,$77,$C3,$75,$D7,$E5,$60,$CB
	.dc.b	$15,$DF,$1E,$CF,$C8,$C7,$B6,$5E,$4C,$3F,$17,$97,$E2,$7E,$1E,$5F
	.dc.b	$DB,$7F,$19,$F0,$DF,$8D,$F8,$6F,$C7,$7C,$2C,$AF,$C0,$47,$E4,$7E
	.dc.b	$1A,$47,$E0,$DF,$CA,$7C,$3F,$48,$FB,$38,$5C,$FC,$5B,$B2,$ED,$5F
	.dc.b	$8B,$CB,$B3,$F8,$0C,$BA,$BF,$82,$CA,$E1,$CB,$2F,$3C,$FA,$48,$7E
	.dc.b	$23,$0E,$AF,$E2,$B0,$EB,$5E,$2C,$39,$E9,$F3,$31,$B7,$F6,$71,$FD
	.dc.b	$77,$C1,$C8,$FD,$03,$FB,$1F,$87,$FD,$97,$C3,$7E,$CF,$E1,$E5,$7E
	.dc.b	$96,$3F,$6D,$F0,$F2,$3F,$50,$FE,$F7,$F2,$1F,$F7,$1C,$7E,$1E,$FB
	.dc.b	$3F,$A0,$C3,$B3,$CC,$FD,$86,$1D,$9F,$D2,$E1,$D5,$FD,$3E,$17,$5F
	.dc.b	$C0,$63,$DB,$4F,$E1,$79,$76,$9F,$A7,$97,$6C,$39,$9D,$88,$60,$EA
	.dc.b	$7A,$88,$7F,$10,$BF,$5F,$98,$FD,$ED,$AF,$F9,$79,$5D,$31,$19,$C7
	.dc.b	$5A,$4C,$7F,$B1,$DA,$79,$6B,$27,$2F,$5D,$79,$C7,$0C,$7B,$3C,$B1
	.dc.b	$EC,$F4,$C7,$B3,$FE,$59,$C7,$B3,$D0,$86,$70,$EC,$F3,$3B,$10,$E5
	.dc.b	$5D,$9E,$95,$DB,$4F,$F4,$75,$F6,$9E,$87,$68,$CD,$F6,$8F,$F3,$67
	.dc.b	$9E,$F8,$7F,$F4,$3A,$F8,$9A,$62,$33,$EF,$EB,$DF,$97,$FE,$0C,$BF
	.dc.b	$F5,$11,$FE,$94,$C7,$FD,$40,$BE,$07,$A9,$DB,$C4,$E9,$5C,$72,$EE
	.dc.b	$F3,$CB,$BB,$D7,$1E,$EF,$A3,$2E,$4F,$87,$0E,$AF,$8F,$0E,$AF,$BB
	.dc.b	$0E,$AF,$C0,$C3,$AB,$F0,$B0,$EB,$97,$A3,$1E,$B3,$F1,$B1,$EA,$FC
	.dc.b	$8C,$7E,$3F,$BF,$F2,$3E,$06,$4F,$C3,$8F,$97,$EF,$FC,$CF,$80,$93
	.dc.b	$F1,$E3,$E7,$FB,$F2,$3E,$53,$F4,$BE,$03,$AC,$73,$21,$F6,$3E,$03
	.dc.b	$AB,$F4,$6F,$AB,$F7,$2F,$A3,$F7,$AE,$EB,$F5,$30,$EB,$3F,$82,$C6
	.dc.b	$CF,$D8,$C7,$AC,$FE,$1B,$1E,$B3,$F7,$31,$E7,$D7,$FE,$D6,$E2,$67
	.dc.b	$8C,$3F,$DC,$D0,$CF,$F1,$BF,$01,$27,$F1,$0F,$E4,$27,$F2,$3F,$01
	.dc.b	$F9,$2F,$80,$95,$F8,$D8,$FC,$AF,$BF,$F9,$6F,$7E,$4F,$E4,$63,$A3
	.dc.b	$F9,$CC,$2E,$BF,$92,$C7,$AC,$FE,$5F,$1E,$AF,$E8,$B1,$EA,$FE,$6F
	.dc.b	$1E,$AF,$E9,$B1,$E7,$A7,$BF,$2B,$3E,$EC,$7A,$BF,$AC,$C7,$AE,$5F
	.dc.b	$A2,$CB,$AC,$FE,$C3,$2E,$B3,$FA,$6C,$BF,$DE,$FF,$BE,$7F,$A2,$E9
	.dc.b	$E0,$23,$FD,$18,$48,$DF,$BD,$1F,$F3,$EC,$7F,$9F,$DF,$F7,$42,$F7
	.dc.b	$BA,$FB,$BD,$BD,$E9,$5C,$23,$C2,$C3,$99,$E4,$F7,$BB,$3D,$30,$B9
	.dc.b	$FF,$85,$32,$F3,$63,$DA,$BC,$78,$F6,$7C,$78,$F5,$7C,$A5,$BE,$59
	.dc.b	$32,$EB,$C2,$AF,$87,$3E,$DA,$DE,$7F,$23,$2F,$05,$7C,$0C,$BC,$15
	.dc.b	$D3,$2F,$EC,$9D,$D2,$BF,$CA,$FE,$37,$F1,$F6,$F7,$FF,$69,$EE,$91
	.dc.b	$A7,$FB,$67,$7F,$F6,$DE,$E9,$3C,$9F,$EE,$1D,$FD,$63,$FA,$C6,$37
	.dc.b	$3F,$EC,$F8,$F4,$9F,$EB,$98,$D9,$FE,$D7,$8F,$49,$FE,$C1,$8D,$9F
	.dc.b	$ED,$F8,$FF,$7D,$C3,$A7,$07,$F9,$87,$3F,$56,$1F,$D9,$E7,$A4,$E7
	.dc.b	$2B,$3F,$E6,$8A,$65,$D6,$B9,$65,$6B,$58,$FF,$82,$FB,$BF,$C1,$9D
	.dc.b	$F2,$7E,$4C,$7F,$83,$FB,$64,$7C,$C7,$FC,$25,$DD,$FE,$13,$EE,$FF
	.dc.b	$0A,$77,$4A,$FA,$31,$D5,$E6,$43,$FC,$01,$DF,$D1,$FF,$01,$5D,$E1
	.dc.b	$FF,$08,$E1,$D6,$7E,$DE,$1D,$5F,$F0,$46,$1D,$5F,$F0,$56,$36,$FF
	.dc.b	$C3,$38,$F3,$D7,$E5,$65,$6F,$E3,$65,$D5,$FF,$09,$65,$D7,$0F,$F1
	.dc.b	$0E,$5F,$E2,$FE,$F9,$5F,$8C,$7F,$C6,$5F,$03,$FE,$33,$EE,$93,$F9
	.dc.b	$07,$FC,$6D,$DF,$FE,$37,$EE,$FF,$1C,$77,$CA,$FC,$AC,$72,$7A,$10
	.dc.b	$FF,$12,$5F,$47,$FC,$4D,$77,$2F,$F1,$AE,$1D,$67,$F3,$F8,$75,$7F
	.dc.b	$C5,$98,$75,$7F,$C5,$D8,$FA,$9E,$A7,$9C,$87,$63,$F5,$35,$D5,$FF
	.dc.b	$1B,$57,$5B,$FF,$22,$DF,$F9,$4F,$C0,$FF,$68,$7E,$47,$FC,$77,$80
	.dc.b	$DE,$74,$EB,$E0,$66,$5F,$EC,$CA,$47,$2E,$FE,$7D,$F2,$B4,$F6,$EE
	.dc.b	$F0,$47,$13,$C2,$60,$FF,$6D,$DF,$47,$B5,$F8,$1C,$DF,$57,$77,$74
	.dc.b	$FF,$50,$5B,$0E,$91,$FE,$64,$C6,$E7,$FE,$D5,$A7,$7F,$5A,$E3,$9D
	.dc.b	$AD,$67,$D5,$E1,$95,$CF,$FD,$BF,$8F,$77,$E2,$78,$3F,$30,$8F,$C7
	.dc.b	$F8,$1F,$9F,$F0,$3F,$43,$E0,$64,$FC,$B8,$FA,$7F,$03,$23,$E7,$3F
	.dc.b	$5B,$E0,$7E,$BF,$C4,$FD,$8F,$89,$95,$F4,$E3,$ED,$FC,$47,$DC,$F8
	.dc.b	$99,$3F,$5E,$3E,$FF,$C4,$FE,$03,$E2,$64,$FD,$B8,$FC,$1F,$C2,$48
	.dc.b	$FB,$CF,$E1,$BE,$1F,$F0,$FF,$0B,$F8,$8F,$86,$95,$F8,$38,$FE,$88
	.dc.b	$BA,$3F,$DA,$5F,$05,$FE,$E6,$8C,$73,$FF,$49,$3F,$DB,$9F,$0D,$C2
	.dc.b	$B6,$D8,$D5,$71,$9E,$7F,$0F,$9B,$57,$FB,$96,$30,$D8,$FE,$AD,$3E
	.dc.b	$2F,$86,$CD,$BE,$1E,$B2,$B7,$7E,$09,$DF,$C3,$70,$F8,$59,$5A,$3C
	.dc.b	$37,$FF,$00,$1C,$5F,$FB,$DF,$E1,$BB,$92,$33,$2F,$87,$F1,$FC,$3C
	.dc.b	$9C,$BE,$6F,$86,$F3,$FC,$2F,$A3,$E1,$A5,$72,$8F,$5F,$C3,$7B,$3E
	.dc.b	$0E,$4F,$68,$F7,$FC,$27,$C0,$F8,$69,$3E,$28,$F8,$7F,$0B,$F1,$3E
	.dc.b	$12,$4F,$9A,$3E,$77,$C2,$3E,$1B,$F3,$D3,$F6,$8F,$F1,$1B,$FA,$A7
	.dc.b	$E2,$46,$FC,$EE,$33,$C3,$A3,$FC,$28,$FF,$82,$EB,$6D,$7A,$57,$7F
	.dc.b	$2B,$8F,$95,$CB,$C9,$95,$C1,$EB,$E5,$76,$F4,$24,$F4,$7C,$3E,$5F
	.dc.b	$8B,$09,$3E,$07,$AB,$CC,$87,$8B,$26,$1D,$A3,$A1,$0D,$5F,$67,$A1
	.dc.b	$0E,$55,$D9,$E9,$3E,$A6,$9F,$0B,$0F,$8F,$E5,$FC,$8F,$2A,$4F,$C6
	.dc.b	$8F,$97,$E5,$C8,$F9,$2F,$CE,$F2,$BE,$7F,$97,$27,$E6,$BF,$4B,$CA
	.dc.b	$EB,$1F,$43,$0E,$AF,$BF,$0E,$8F,$C1,$C2,$F0,$FD,$8C,$7A,$4F,$C5
	.dc.b	$C6,$CF,$DC,$C7,$AC,$FC,$9C,$7A,$CF,$E0,$31,$E7,$E0,$7F,$92,$78
	.dc.b	$BF,$19,$9E,$30,$FC,$5E,$7D,$2B,$5C,$0D,$31,$5F,$90,$D3,$F1,$BE
	.dc.b	$5C,$8F,$C5,$BF,$90,$F2,$FF,$23,$E5,$C9,$FC,$7B,$F9,$4F,$2B,$F2
	.dc.b	$BE,$5F,$E7,$BF,$7F,$FE,$1D,$A7,$95,$B7,$F9,$E9,$EC,$FE,$17,$1B
	.dc.b	$C3,$F9,$CC,$BB,$4F,$E2,$72,$B3,$FA,$0C,$BB,$CF,$E9,$B1,$EB,$C0
	.dc.b	$6E,$CC,$FA,$E9,$F0,$B4,$B8,$7E,$C3,$4E,$B3,$F1,$B4,$B3,$FB,$4C
	.dc.b	$FB,$BF,$AC,$CB,$B9,$DB,$1F,$81,$E6,$7C,$1F,$32,$57,$96,$3E,$27
	.dc.b	$99,$23,$D0,$FC,$7F,$33,$E4,$79,$3F,$27,$CC,$95,$ED,$8E,$AF,$BF
	.dc.b	$0E,$AF,$8F,$0E,$8F,$97,$0B,$C5,$FE,$99,$CF,$F9,$B4,$F3,$3E,$46
	.dc.b	$3D,$9F,$66,$3D,$9E,$87,$A4,$F9,$DE,$67,$4D,$FE,$E6,$36,$7F,$D7
	.dc.b	$78,$76,$79,$9D,$88,$60,$F9,$5E,$67,$E1,$C8,$FD,$BF,$33,$F1,$5E
	.dc.b	$64,$8F,$BC,$FE,$37,$CC,$FC,$77,$97,$F8,$FF,$32,$57,$E0,$E3,$9B
	.dc.b	$D8,$FC,$36,$1F,$95,$C3,$A4,$7D,$BC,$2E,$FF,$1F,$1E,$B3,$F8,$DC
	.dc.b	$7A,$BF,$81,$C7,$93,$D0,$87,$E1,$70,$E7,$A7,$E8,$71,$B3,$F1,$B0
	.dc.b	$EA,$FE,$2F,$0E,$B8,$FD,$7C,$7F,$B6,$F4,$7F,$72,$FC,$91,$F1,$B4
	.dc.b	$CE,$1F,$D9,$A3,$FD,$4B,$7F,$DE,$A3,$FB,$DC,$63,$3F,$F0,$33,$63
	.dc.b	$3F,$0D,$D7,$E1,$BB,$7C,$34,$9E,$91,$D9,$FE,$AD,$6B,$F1,$C7,$87
	.dc.b	$1D,$63,$E1,$AD,$7C,$3F,$A4,$95,$E7,$CB,$C0,$FA,$72,$F0,$3D,$B1
	.dc.b	$65,$D3,$0F,$4E,$5F,$06,$BF,$C0,$39,$F4,$78,$65,$73,$F1,$62,$CF
	.dc.b	$BD,$75,$CF,$BD,$6B,$3F,$9D,$F0,$FF,$3F,$E1,$E5,$FC,$D8,$FA,$5F
	.dc.b	$0D,$23,$E8,$3F,$57,$E1,$FE,$B7,$C3,$49,$FA,$8F,$D9,$F8,$6E,$B1
	.dc.b	$F5,$F0,$EA,$FB,$F0,$E8,$FC,$1C,$2E,$5E,$AC,$7A,$4F,$C5,$C6,$CF
	.dc.b	$BB,$1E,$B3,$F2,$71,$EB,$3F,$0B,$1E,$7C,$3F,$82,$CA,$DF,$E1,$B2
	.dc.b	$E7,$3F,$86,$CA,$CF,$E2,$B2,$FC,$97,$C2,$C8,$FC,$83,$F9,$5F,$86
	.dc.b	$FC,$B7,$C3,$C9,$FC,$A3,$F9,$9F,$85,$FC,$D7,$C3,$FE,$6F,$E1,$E5
	.dc.b	$7E,$66,$3A,$3F,$81,$C2,$EF,$F6,$B1,$E9,$3F,$85,$C6,$CF,$DE,$C7
	.dc.b	$AC,$FE,$8F,$1E,$AF,$E2,$F1,$E7,$A7,$E8,$B2,$B3,$FA,$6C,$B9,$CF
	.dc.b	$E9,$B2,$B3,$FA,$AC,$7A,$BF,$95,$C7,$AE,$5F,$8E,$CB,$F6,$DF,$0F
	.dc.b	$2B,$F6,$8F,$EE,$7E,$1B,$F7,$5F,$0F,$27,$F7,$0F,$EF,$7E,$1B,$FC
	.dc.b	$DF,$F9,$27,$FB,$24,$FF,$5C,$4F,$EC,$1B,$8F,$EB,$5F,$F6,$77,$4B
	.dc.b	$1F,$05,$7E,$A5,$F0,$3D,$88,$7E,$B7,$0F,$03,$FE,$6E,$AF,$07,$C3
	.dc.b	$F8,$8F,$1C,$72,$3A,$1E,$C3,$F8,$86,$5E,$B7,$A1,$DB,$1F,$D8,$E5
	.dc.b	$CA,$3B,$57,$57,$C0,$7F,$99,$3F,$80,$3D,$4F,$F9,$81,$5E,$4D,$65
	.dc.b	$BC,$B8,$61,$2F,$84,$7F,$9B,$BF,$80,$7F,$D5,$31,$FE,$62,$6E,$5F
	.dc.b	$C9,$7D,$34,$63,$8B,$FC,$54,$FC,$E7,$FC,$59,$9F,$92,$FF,$CD,$6F
	.dc.b	$91,$F3,$67,$E4,$7C,$27,$A7,$2F,$23,$C3,$2F,$23,$0F,$F3,$EE,$1E
	.dc.b	$15,$8F,$93,$4E,$91,$E0,$C3,$E1,$3F,$17,$4F,$84,$F7,$21,$DC,$EA
	.dc.b	$77,$3E,$11,$F1,$67,$E3,$7C,$4C,$8F,$1B,$F2,$7D,$CF,$95,$EF,$7C
	.dc.b	$BF,$81,$95,$E8,$8F,$9D,$EF,$C8,$F5,$BF,$47,$E0,$7A,$47,$D1,$C2
	.dc.b	$F1,$7D,$1D,$BF,$D0,$73,$CC,$F8,$78,$F6,$7E,$C6,$3D,$5F,$B5,$8D
	.dc.b	$BE,$58,$F4,$E2,$FB,$39,$5B,$F9,$79,$76,$79,$9D,$88,$60,$EA,$7A
	.dc.b	$88,$7E,$26,$BF,$19,$F0,$12,$3E,$CB,$F8,$FF,$81,$FC,$87,$C0,$7E
	.dc.b	$47,$E0,$65,$7D,$E8,$FC,$A7,$C0,$C8,$FC,$0B,$F9,$7F,$81,$E9,$1F
	.dc.b	$97,$C2,$EB,$F1,$F1,$EB,$3F,$88,$C7,$AB,$F9,$DC,$7A,$3F,$9F,$C6
	.dc.b	$DF,$CF,$C7,$CD,$3E,$82,$1F,$A5,$C3,$AB,$FA,$7C,$3A,$D7,$D7,$C3
	.dc.b	$9E,$BF,$A6,$C6,$DF,$CE,$C7,$F6,$1F,$03,$FB,$1F,$81,$93,$F9,$E8
	.dc.b	$FD,$A7,$C0,$FE,$D7,$E0,$64,$FE,$8A,$3F,$71,$F0,$32,$3F,$4A,$FE
	.dc.b	$EF,$E0,$7A,$C7,$EA,$30,$EA,$FE,$F7,$0E,$AF,$EB,$30,$EA,$FF,$01
	.dc.b	$C3,$A3,$FC,$0F,$0B,$AF,$E4,$F1,$EB,$3F,$C1,$F1,$EB,$3F,$97,$C7
	.dc.b	$AC,$FF,$0B,$C7,$AC,$FE,$6F,$1E,$7A,$7F,$0A,$CA,$DF,$E4,$B2,$FE
	.dc.b	$29,$F0,$32,$3F,$80,$A7,$FD,$FE,$BD,$DC,$47,$FB,$EA,$1E,$FE,$32
	.dc.b	$FE,$35,$0F,$80,$C6,$5F,$EC,$38,$C5,$F4,$BC,$BF,$C3,$A1,$EF,$63
	.dc.b	$C0,$FF,$53,$7C,$0D,$9F,$F3,$D6,$5D,$AB,$D1,$7F,$CA,$E6,$CF,$8B
	.dc.b	$3E,$18,$F1,$CE,$5F,$84,$F6,$E1,$D7,$19,$18,$3E,$19,$0E,$C7,$C4
	.dc.b	$BE,$AF,$C6,$AE,$AF,$C5,$BF,$45,$F4,$8F,$91,$78,$7F,$AF,$75,$7F
	.dc.b	$6E,$23,$FD,$87,$E2,$7F,$B8,$F6,$FF,$72,$EE,$93,$FD,$96,$3F,$BA
	.dc.b	$F7,$7F,$76,$ED,$93,$FD,$A6,$39,$BF,$DD,$F0,$BB,$7F,$77,$C7,$AC
	.dc.b	$FF,$6F,$C7,$AB,$FD,$F7,$1E,$AF,$F7,$3C,$7A,$BF,$E0,$1C,$7F,$C0
	.dc.b	$B7,$D3,$87,$F1,$03,$C6,$67,$AB,$F4,$1D,$C8,$77,$3D,$07,$72,$1D
	.dc.b	$CF,$F0,$27,$81,$FE,$14,$EF,$91,$83,$FC,$11,$DF,$FE,$18,$EE,$91
	.dc.b	$83,$FC,$19,$DF,$FE,$1C,$23,$FE,$0F,$EF,$FF,$0F,$E1,$D2,$3F,$C3
	.dc.b	$D8,$61,$E4,$7F,$88,$6F,$AB,$FE,$23,$C3,$9B,$FE,$25,$C2,$F0,$FF
	.dc.b	$89,$71,$E5,$3D,$0F,$F0,$F6,$18,$9F,$F1,$55,$F4,$7F,$C5,$98,$7F
	.dc.b	$8C,$A3,$9E,$9F,$3B,$46,$58,$C5,$3F,$47,$2F,$F1,$CF,$7F,$F8,$EB
	.dc.b	$BA,$57,$F8,$B6,$3F,$C7,$BD,$FF,$E3,$EE,$E9,$3F,$E3,$18,$FF,$20
	.dc.b	$F7,$FF,$90,$BB,$E4,$FF,$8D,$63,$AB,$FE,$37,$C3,$AB,$FE,$45,$C3
	.dc.b	$AB,$FE,$3B,$C3,$AB,$FE,$49,$C3,$AB,$FE,$3F,$C3,$AB,$FE,$4D,$C3
	.dc.b	$AB,$CC,$F4,$1F,$E4,$5E,$FE,$AF,$F9,$4E,$FA,$F0,$7F,$94,$F0,$EB
	.dc.b	$3F,$E5,$7C,$3A,$CF,$F9,$5F,$0F,$F2,$EF,$74,$AF,$F2,$83,$FC,$53
	.dc.b	$57,$E4,$DF,$ED,$99,$FF,$67,$BF,$C8,$1F,$E4,$D7,$FE,$CD,$8F,$E3
	.dc.b	$B1,$AE,$EE,$51,$B8,$E3,$DB,$8E,$D9,$1C,$B0,$FD,$4D,$7F,$22,$8F
	.dc.b	$19,$0F,$F3,$86,$3D,$27,$AF,$77,$6C,$65,$73,$3F,$07,$7E,$1C,$BC
	.dc.b	$D9,$78,$27,$D7,$97,$9A,$3E,$14,$76,$3C,$13,$ED,$C7,$BC,$75,$7E
	.dc.b	$2E,$1E,$08,$F8,$B8,$F3,$75,$8F,$F3,$8A,$F6,$E3,$99,$F9,$7D,$FF
	.dc.b	$18,$8F,$0F,$83,$F9,$3F,$07,$23,$9B,$F3,$3E,$0B,$E6,$FC,$14,$9E
	.dc.b	$EF,$D0,$F8,$2F,$A3,$F0,$5F,$4B,$E0,$65,$79,$23,$EA,$FC,$17,$D6
	.dc.b	$F8,$89,$3E,$88,$FB,$3F,$13,$23,$D6,$FD,$CF,$88,$FB,$BF,$13,$F7
	.dc.b	$BE,$22,$57,$C0,$8F,$C0,$FC,$4F,$E0,$BE,$1E,$4F,$C4,$8F,$C7,$79
	.dc.b	$6F,$88,$BF,$11,$2B,$E7,$4E,$5E,$84,$3F,$27,$87,$2F,$88,$EF,$1F
	.dc.b	$43,$1E,$EF,$E6,$31,$EF,$8F,$58,$76,$7F,$08,$7F,$84,$E2,$CB,$71
	.dc.b	$E2,$F8,$7F,$3F,$C4,$48,$C1,$E6,$F8,$7F,$59,$1F,$4F,$C4,$7B,$BE
	.dc.b	$22,$47,$B1,$F8,$3F,$11,$F0,$BE,$22,$4F,$C0,$7E,$2F,$C4,$7C,$6F
	.dc.b	$88,$FF,$02,$7E,$38,$69,$9A,$F8,$FF,$13,$F3,$08,$FC,$AF,$89,$F9
	.dc.b	$FF,$13,$F4,$3E,$22,$4F,$CE,$8E,$BF,$09,$DB,$E1,$24,$FE,$AF,$E0
	.dc.b	$99,$FF,$AB,$E3,$E4,$F9,$5F,$D8,$39,$9F,$13,$F7,$30,$95,$F6,$A3
	.dc.b	$EF,$E3,$23,$EE,$BF,$82,$F8,$6F,$F1,$07,$E3,$5F,$DA,$27,$C0,$D7
	.dc.b	$B9,$FE,$B4,$71,$D5,$F6,$72,$64,$F8,$8F,$E3,$5F,$93,$7F,$AE,$E4
	.dc.b	$FF,$B2,$F0,$FE,$8E,$D8,$CC,$7F,$8C,$7F,$19,$79,$D7,$CE,$93,$F0
	.dc.b	$E7,$2F,$9B,$E0,$FD,$24,$7D,$3F,$07,$CF,$D0,$E9,$E8,$49,$E5,$1D
	.dc.b	$F1,$EB,$1E,$EE,$56,$2C,$B5,$3D,$C8,$6F,$83,$FD,$59,$C1,$FE,$EB
	.dc.b	$59,$F8,$67,$D1,$9F,$85,$ED,$83,$3E,$B1,$8C,$EE,$3F,$EF,$5A,$7A
	.dc.b	$1E,$DA,$F0,$6B,$6B,$C1,$97,$79,$EB,$97,$79,$E5,$97,$CD,$F3,$FE
	.dc.b	$77,$A1,$2F,$E6,$47,$3E,$0F,$F3,$97,$03,$7F,$6C,$61,$B5,$F8,$3C
	.dc.b	$FE,$C7,$9F,$F6,$7D,$09,$3F,$5E,$3C,$0F,$D9,$CB,$C0,$FC,$2C,$BC
	.dc.b	$0F,$E3,$76,$B9,$F8,$71,$EB,$3F,$81,$C7,$AB,$F2,$B1,$EA,$F3,$3D
	.dc.b	$67,$E1,$FD,$0E,$79,$FA,$30,$F6,$63,$FA,$0C,$7A,$DF,$D4,$C7,$AE
	.dc.b	$DF,$47,$2F,$CA,$7A,$1F,$95,$F3,$E5,$FE,$4E,$39,$3F,$84,$AF,$C1
	.dc.b	$63,$88,$FB,$FE,$87,$E7,$7C,$E9,$1F,$9B,$8E,$AF,$E7,$70,$EA,$FE
	.dc.b	$1B,$0E,$58,$74,$8E,$DC,$3F,$4F,$4F,$D4,$61,$D2,$7F,$1B,$8E,$1F
	.dc.b	$31,$0E,$87,$EA,$FC,$FE,$B7,$F7,$73,$F8,$F8,$71,$AE,$67,$EC,$6F
	.dc.b	$18,$FF,$01,$8F,$E8,$C7,$F0,$6B,$FF,$5D,$79,$DF,$E1,$B8,$FE,$78
	.dc.b	$7F,$79,$37,$3F,$98,$7D,$67,$E0,$4F,$DC,$9B,$F4,$39,$79,$F2,$30
	.dc.b	$73,$F4,$3B,$5C,$8C,$1C,$8F,$D3,$9D,$B8,$BF,$18,$F2,$3D,$02,$BB
	.dc.b	$38,$BF,$55,$C8,$EA,$70,$F4,$3E,$4B,$E9,$3B,$9F,$C1,$74,$F8,$55
	.dc.b	$D3,$2F,$91,$86,$2B,$E2,$7A,$1F,$1F,$D0,$91,$83,$D5,$5F,$81,$7A
	.dc.b	$BF,$25,$F4,$FA,$1F,$3F,$D0,$91,$F3,$E3,$E9,$5F,$48,$F7,$DE,$1E
	.dc.b	$BB,$FE,$05,$F8,$DD,$1F,$C6,$9E,$47,$D9,$C3,$AB,$E0,$3C,$C7,$CA
	.dc.b	$BE,$8F,$CC,$F4,$3E,$2F,$07,$EC,$23,$AE,$DF,$85,$CF,$9E,$9F,$8B
	.dc.b	$C7,$15,$FE,$3F,$F0,$FF,$A7,$E9,$FE,$BE,$5F,$0B,$FD,$38,$FE,$9F
	.dc.b	$F0,$BD,$B6,$FD,$F1,$33,$FE,$A7,$F0,$FF,$D5,$3E,$22,$65,$D6,$38
	.dc.b	$BD,$F0,$E8,$F8,$70,$E9,$D1,$FA,$48,$E0,$79,$6F,$9B,$FA,$1C,$98
	.dc.b	$74,$8C,$5D,$D7,$FB,$45,$7F,$68,$9F,$A2,$7E,$26,$3E,$F1,$F7,$4D
	.dc.b	$D7,$67,$8D,$FB,$27,$E3,$7C,$47,$F7,$0C,$64,$60,$EB,$C1,$F4,$5F
	.dc.b	$C4,$ED,$FC,$25,$F9,$DA,$FF,$78,$F8,$79,$18,$39,$1E,$DB,$EA,$FB
	.dc.b	$F0,$F8,$AF,$DC,$38,$1F,$0E,$79,$BF,$16,$BF,$C0,$B3,$23,$AE,$5F
	.dc.b	$13,$E2,$39,$C7,$98,$FE,$9C,$74,$8F,$93,$7D,$1F,$97,$7F,$E1,$0F
	.dc.b	$87,$FF,$08,$FC,$3C,$BF,$C5,$47,$4A,$F9,$B9,$FC,$19,$FB,$1F,$0D
	.dc.b	$FE,$18,$F8,$79,$1F,$91,$8E,$8F,$E4,$F0,$E8,$FD,$FC,$3A,$5F,$EA
	.dc.b	$63,$F8,$01,$C4,$FC,$D5,$F4,$7B,$9E,$42,$1C,$CF,$D0,$7C,$2F,$47
	.dc.b	$F0,$93,$E7,$C7,$F1,$38,$F3,$7F,$17,$7F,$E3,$08,$FF,$18,$FC,$37
	.dc.b	$F8,$CB,$E1,$66,$1F,$AB,$8E,$9C,$3F,$83,$C3,$F1,$D8,$FE,$B7,$E1
	.dc.b	$BF,$C7,$3F,$0F,$30,$FD,$94,$74,$7F,$69,$8F,$47,$F4,$18,$F4,$E9
	.dc.b	$FD,$84,$70,$3F,$75,$7C,$DF,$D3,$5F,$37,$89,$E5,$3F,$80,$7C,$3F
	.dc.b	$3E,$5F,$E5,$B9,$7F,$92,$AB,$96,$7F,$AA,$C3,$FC,$9F,$1F,$EE,$9F
	.dc.b	$BF,$73,$5C,$26,$5F,$EC,$99,$CB,$FD,$4F,$E0,$3F,$AB,$7C,$04,$8C
	.dc.b	$1B,$E6,$E8,$47,$97,$3F,$7C,$7A,$C7,$F6,$1C,$3F,$C4,$BA,$9E,$F1
	.dc.b	$FD,$3B,$1E,$CF,$F5,$0C,$7F,$B3,$47,$58,$E5,$8D,$D7,$FC,$DD,$6C
	.dc.b	$BC,$15,$EB,$CF,$C3,$5A,$CF,$9C,$FF,$6F,$CB,$BC,$FF,$77,$CB,$A6
	.dc.b	$6A,$3E,$7C,$F3,$18,$CF,$C0,$FC,$9E,$8F,$9D,$CF,$23,$07,$C8,$3D
	.dc.b	$87,$84,$C6,$3E,$AE,$8F,$AC,$48,$FA,$9C,$DF,$66,$FA,$47,$CB,$BC
	.dc.b	$3C,$B2,$F4,$1E,$9D,$3D,$AE,$21,$A7,$57,$EA,$63,$72,$F9,$79,$74
	.dc.b	$9E,$C7,$E1,$71,$C4,$FC,$78,$F5,$1E,$C3,$ED,$1F,$8B,$AE,$98,$79
	.dc.b	$33,$47,$E4,$F0,$FC,$A7,$3F,$E5,$7A,$26,$1F,$93,$8F,$AE,$F5,$C7
	.dc.b	$FC,$59,$A6,$1F,$C0,$F3,$7E,$74,$93,$F9,$BE,$7E,$AF,$E7,$70,$EA
	.dc.b	$FE,$4F,$0E,$5A,$7C,$5C,$7E,$3C,$FF,$BC,$76,$E6,$FE,$67,$1B,$97
	.dc.b	$E4,$32,$EB,$3F,$9D,$CB,$AC,$FE,$53,$2E,$B3,$FE,$FB,$C7,$AE,$5F
	.dc.b	$8D,$CB,$9B,$F1,$72,$B5,$F8,$DC,$BF,$71,$C9,$FB,$9E,$59,$3F,$B7
	.dc.b	$8E,$B7,$F7,$E3,$E9,$E7,$F5,$5F,$D8,$73,$7F,$01,$23,$F9,$D8,$FF
	.dc.b	$6B,$1F,$9E,$7A,$1F,$C0,$EF,$BB,$FB,$9B,$EE,$B3,$F5,$BF,$8C,$3F
	.dc.b	$85,$5F,$67,$C2,$77,$C3,$F2,$B8,$77,$8E,$87,$72,$18,$3F,$16,$7E
	.dc.b	$06,$3E,$B3,$F9,$F3,$F6,$53,$DE,$7F,$81,$BF,$B9,$8F,$F3,$84,$77
	.dc.b	$3D,$FF,$0D,$97,$E0,$FC,$3D,$BF,$2B,$6B,$F8,$0C,$AB,$E0,$EB,$F4
	.dc.b	$21,$F1,$1B,$C1,$0F,$87,$C2,$F8,$7A,$62,$33,$86,$B2,$99,$7C,$73
	.dc.b	$C3,$1E,$07,$E0,$61,$DB,$1E,$51,$B8,$F0,$E1,$DA,$31,$8F,$8D,$B7
	.dc.b	$D9,$F8,$B8,$78,$9E,$A7,$DB,$3E,$8C,$FE,$1B,$0E,$F1,$F3,$2F,$B3
	.dc.b	$F3,$B1,$F8,$5F,$93,$7C,$FC,$BF,$89,$BF,$26,$7C,$CD,$7F,$D4,$8E
	.dc.b	$DF,$D8,$1E,$2C,$FE,$4E,$1D,$32,$EA,$C9,$E7,$1E,$0B,$F0,$E7,$27
	.dc.b	$BC,$7C,$97,$D6,$FE,$4A,$BA,$47,$91,$E8,$F9,$2B,$A3,$E8,$AE,$8F
	.dc.b	$A2,$BA,$67,$FA,$F3,$E3,$47,$90,$E8,$43,$A1,$F0,$B2,$F8,$F9,$C8
	.dc.b	$C1,$F0,$4F,$BF,$3F,$8E,$7E,$4F,$B9,$F3,$88,$FC,$CF,$77,$E8,$DF
	.dc.b	$38,$F9,$B7,$87,$A1,$F0,$8F,$CE,$E7,$C9,$FA,$77,$70,$FE,$09,$8B
	.dc.b	$56,$5C,$EF,$A9,$F7,$31,$C5,$FC,$3B,$FE,$A1,$AF,$E9,$F2,$E9,$1F
	.dc.b	$84,$CA,$E1,$FD,$BB,$2E,$AF,$E1,$BD,$AF,$C5,$FB,$72,$30,$7A,$B2
	.dc.b	$EC,$FA,$8F,$C6,$FB,$9F,$92,$F6,$E4,$60,$E2,$7E,$42,$BA,$3F,$92
	.dc.b	$BF,$B7,$97,$A2,$3F,$53,$9F,$E7,$30,$E7,$1F,$98,$C3,$0F,$4C,$7E
	.dc.b	$3E,$4C,$BA,$C6,$32,$EB,$5F,$9E,$3F,$8A,$E3,$C9,$E8,$7E,$9B,$0C
	.dc.b	$57,$4F,$73,$F5,$BE,$D4,$AF,$D5,$C7,$4D,$3F,$A6,$47,$A0,$FD,$7F
	.dc.b	$B7,$FB,$42,$3F,$B2,$F7,$3F,$6F,$5C,$E3,$F6,$77,$87,$E7,$5F,$C3
	.dc.b	$7C,$E7,$42,$1F,$BA,$BF,$E0,$37,$C6,$39,$9F,$C0,$2B,$18,$7D,$8B
	.dc.b	$F3,$CF,$E4,$DF,$E3,$8F,$48,$F9,$77,$E8,$8F,$E0,$F7,$FC,$3B,$DB
	.dc.b	$91,$FC,$31,$FE,$23,$EE,$7F,$30,$D7,$F9,$8E,$D2,$BF,$91,$C7,$F1
	.dc.b	$6F,$73,$F8,$BF,$B7,$27,$F8,$AC,$7F,$1A,$F7,$3F,$9E,$7E,$64,$F7
	.dc.b	$29,$9A,$FE,$5B,$AF,$F3,$E2,$3F,$CC,$35,$E3,$EE,$C8,$FE,$6D,$68
	.dc.b	$EB,$EE,$76,$F7,$3B,$FB,$B2,$F7,$1F,$D2,$B3,$91,$FC,$F1,$FE,$A1
	.dc.b	$F8,$57,$96,$3A,$FB,$BE,$8F,$7A,$46,$0F,$0F,$BB,$EC,$23,$E4,$F7
	.dc.b	$79,$D7,$4A,$93,$FD,$36,$D3,$F1,$3D,$E9,$3E,$C7,$E3,$FC,$57,$C8
	.dc.b	$F8,$AF,$93,$E1,$CA,$F8,$31,$E8,$3D,$31,$23,$B4,$7D,$0F,$2B,$E8
	.dc.b	$FC,$4C,$9F,$93,$1F,$53,$E2,$BE,$AF,$C4,$49,$F9,$B1,$F1,$3E,$26
	.dc.b	$47,$A5,$F8,$FF,$13,$F7,$19,$1F,$51,$FB,$F1,$FD,$EF,$A9,$E9,$52
	.dc.b	$57,$F6,$F8,$7A,$3B,$F8,$49,$39,$79,$7C,$2F,$3F,$53,$A7,$A9,$2B
	.dc.b	$94,$74,$7F,$C1,$3E,$47,$F9,$EB,$1F,$0C,$77,$CB,$C9,$97,$68,$D9
	.dc.b	$FE,$96,$96,$59,$F2,$3F,$9C,$C3,$3F,$15,$7A,$73,$F1,$3C,$B2,$FC
	.dc.b	$F6,$7E,$3C,$38,$E7,$67,$C9,$9F,$57,$CD,$9D,$C3,$C5,$97,$67,$A6
	.dc.b	$6C,$FB,$E5,$E4,$CF,$FC,$51,$F8,$67,$F8,$07,$23,$49,$EA,$75,$8F
	.dc.b	$D9,$67,$71,$F6,$65,$FE,$21,$B7,$AB,$D3,$CF,$93,$98,$EF,$84,$8E
	.dc.b	$0F,$8B,$1E,$98,$F5,$D7,$CB,$7C,$AF,$E3,$63,$FD,$CD,$B1,$CD,$CE
	.dc.b	$17,$4F,$ED,$35,$FE,$1F,$7E,$9E,$1D,$AB,$F1,$58,$76,$7D,$59,$77
	.dc.b	$AE,$38,$FC,$6F,$56,$57,$BD,$F9,$3E,$A7,$38,$F9,$F8,$FE,$41,$F8
	.dc.b	$9E,$A7,$CE,$F5,$64,$60,$E4,$43,$DB,$7D,$1F,$CD,$10,$EA,$7A,$0E
	.dc.b	$47,$C4,$8E,$AF,$C6,$9F,$3B,$D0,$ED,$8F,$C4,$C3,$D0,$FC,$48,$E8
	.dc.b	$71,$3E,$E5,$74,$8F,$97,$5F,$88,$F3,$65,$7E,$05,$FC,$5F,$9D,$D3
	.dc.b	$7F,$CA,$CF,$4B,$FA,$7E,$AF,$E4,$3D,$49,$3F,$89,$8E,$6F,$DE,$C2
	.dc.b	$D7,$DA,$C3,$A4,$FE,$BF,$5F,$9C,$FE,$12,$F9,$BF,$86,$C3,$F3,$B1
	.dc.b	$CB,$1F,$F7,$AF,$0B,$2C,$5A,$7F,$43,$8F,$67,$F3,$F1,$E9,$78,$DF
	.dc.b	$6C,$BF,$1B,$8F,$6A,$FD,$17,$A9,$FA,$C2,$3F,$A5,$F5,$3A,$72,$FF
	.dc.b	$06,$8E,$87,$87,$47,$A1,$FB,$4F,$42,$4F,$EB,$63,$9B,$FA,$2C,$2E
	.dc.b	$5F,$9E,$C7,$AE,$FF,$12,$3F,$43,$7C,$AF,$F5,$38,$7F,$00,$8E,$3A
	.dc.b	$FC,$4C,$6D,$78,$F1,$E6,$F9,$A7,$F8,$2D,$73,$C3,$F5,$B8,$73,$AF
	.dc.b	$ED,$BF,$15,$FD,$E0,$8F,$F7,$0F,$89,$E6,$FF,$14,$E8,$FF,$5C,$79
	.dc.b	$5F,$EC,$78,$FE,$44,$BE,$3F,$26,$A3,$99,$FC,$9F,$D3,$61,$8C,$BF
	.dc.b	$CC,$F0,$C7,$BC,$7F,$32,$C7,$95,$FF,$AD,$F1,$EC,$DC,$72,$F5,$C7
	.dc.b	$A6,$88,$FC,$36,$5E,$07,$D8,$7E,$D7,$1F,$D5,$3F,$B6,$3D,$C7,$73
	.dc.b	$D6,$77,$30,$78,$F7,$F8,$64,$7C,$D9,$7A,$AF,$FA,$A3,$F2,$3E,$23
	.dc.b	$E4,$FC,$44,$AF,$64,$71,$7E,$5D,$DC,$3F,$D1,$7A,$7F,$38,$9F,$34
	.dc.b	$7A,$09,$8F,$68,$E6,$7C,$7C,$3B,$3F,$5B,$0E,$93,$F1,$45,$FE,$76
	.dc.b	$7C,$27,$A8,$F8,$D1,$FC,$11,$F0,$BD,$DE,$87,$72,$18,$3F,$0B,$F1
	.dc.b	$7F,$88,$23,$F6,$FE,$17,$E3,$3F,$D8,$8F,$BD,$F0,$DF,$8E,$F8,$69
	.dc.b	$18,$3F,$05,$5C,$9E,$A7,$33,$CE,$7F,$0F,$2D,$FF,$94,$F5,$FC,$CD
	.dc.b	$F4,$9F,$CC,$DE,$1E,$A4,$3A,$9F,$9E,$C3,$CF,$7C,$E3,$C8,$72,$3F
	.dc.b	$2D,$3C,$DF,$D2,$FC,$3F,$EA,$08,$FE,$6F,$E1,$F9,$F4,$FD,$DD,$FE
	.dc.b	$FF,$C3,$FE,$BB,$E1,$A5,$FE,$86,$38,$BD,$0F,$D2,$61,$FB,$4C,$39
	.dc.b	$EB,$FB,$6A,$E3,$5F,$B6,$BC,$3C,$08,$73,$3F,$77,$5C,$9F,$01,$FB
	.dc.b	$18,$E7,$87,$9A,$3F,$5C,$FF,$33,$5F,$0E,$FD,$F5,$E9,$7E,$CE,$7F
	.dc.b	$C6,$33,$9F,$43,$17,$C6,$31,$3D,$37,$FD,$D4,$73,$BD,$C6,$30,$7A
	.dc.b	$38,$BF,$0C,$3D,$1C,$61,$E3,$6C,$70,$9E,$C7,$0F,$47,$24,$73,$5E
	.dc.b	$BE,$BE,$39,$72,$D2,$56,$B0,$F7,$B7,$2E,$F9,$48,$C1,$BC,$BB,$3F
	.dc.b	$12,$91,$E6,$C7,$BC,$F5,$C3,$BB,$BC,$7D,$78,$F5,$8E,$58,$E1,$EC
	.dc.b	$43,$B6,$1D,$5F,$83,$9B,$F1,$77,$A9,$F2,$E3,$1E,$AD,$93,$67,$B5
	.dc.b	$78,$0F,$85,$86,$37,$FC,$26,$1D,$7D,$5E,$44,$AF,$0E,$3F,$21,$FF
	.dc.b	$36,$3E,$27,$38,$F8,$4F,$03,$F2,$8F,$01,$2F,$F1,$5B,$FD,$D8,$E7
	.dc.b	$7D,$DA,$78,$35,$E6,$F8,$2F,$B3,$E4,$CF,$D5,$97,$E3,$F3,$F9,$D9
	.dc.b	$76,$D7,$A3,$EE,$D1,$9E,$2F,$F9,$74,$7B,$67,$5B,$7C,$BB,$99,$7C
	.dc.b	$28,$EE,$FE,$6B,$1E,$F8,$7E,$7E,$BE,$94,$49,$F9,$2F,$D5,$DB,$EB
	.dc.b	$5F,$49,$F9,$75,$87,$E4,$61,$F1,$AF,$AC,$7E,$A2,$7A,$3F,$4A,$2C
	.dc.b	$FB,$6F,$94,$F4,$3E,$CD,$62,$7E,$24,$7E,$08,$87,$E7,$63,$9E,$1F
	.dc.b	$12,$EC,$FF,$8E,$F1,$FD,$0B,$F8,$4D,$FF,$21,$E0,$C8,$C1,$EA,$BF
	.dc.b	$A8,$F6,$C7,$C2,$FF,$AD,$7E,$2F,$FC,$2D,$1F,$92,$6C,$FF,$61,$6C
	.dc.b	$FF,$92,$DF,$F0,$6B,$F9,$2B,$F0,$3F,$94,$BF,$D1,$47,$D6,$E2,$FE
	.dc.b	$19,$3E,$1B,$FA,$CF,$EB,$2F,$AC,$1A,$4F,$0B,$9C,$FE,$8B,$0E,$F8
	.dc.b	$FE,$2F,$1E,$DC,$7F,$C6,$EB,$B1,$F5,$CE,$C7,$ED,$F0,$EA,$DC,$31
	.dc.b	$E0,$E7,$A2,$4F,$96,$3A,$ED,$F8,$2D,$7F,$9C,$6F,$CF,$C0,$97,$EB
	.dc.b	$7B,$78,$3D,$23,$DD,$8F,$47,$E8,$63,$D1,$F3,$E1,$D1,$E4,$43,$EB
	.dc.b	$5F,$47,$C8,$74,$21,$D0,$FC,$01,$FA,$53,$AE,$9F,$BF,$DB,$EF,$4F
	.dc.b	$A2,$7A,$9F,$4F,$1F,$F2,$CF,$C5,$7F,$96,$BE,$26,$4F,$F9,$36,$3D
	.dc.b	$0F,$F0,$BE,$2F,$B6,$FF,$97,$BE,$2B,$FC,$C2,$47,$FC,$A9,$F1,$5D
	.dc.b	$1F,$F2,$B5,$F4,$7F,$CC,$B7,$D3,$2F,$D8,$47,$AB,$0F,$F1,$8E,$1C
	.dc.b	$E3,$B1,$D0,$F2,$1D,$08,$74,$3E,$1E,$9F,$92,$7D,$07,$E1,$4E,$64
	.dc.b	$AF,$F2,$5C,$FF,$9E,$B1,$FF,$3D,$FB,$5F,$E7,$CF,$6A,$5F,$F9,$BE
	.dc.b	$3D,$18,$FE,$BA,$3F,$74,$7F,$9C,$FD,$9F,$F4,$19,$1F,$F3,$B7,$B5
	.dc.b	$D1,$FF,$3C,$5F,$47,$FD,$0F,$7D,$1F,$B4,$74,$3F,$D1,$75,$CD,$A7
	.dc.b	$43,$C8,$74,$21,$D0,$F8,$7C,$1F,$A7,$7E,$1E,$9F,$91,$79,$93,$4F
	.dc.b	$F3,$6D,$7F,$A7,$31,$FE,$C1,$F9,$A7,$F8,$6E,$95,$FD,$6A,$1E,$DE
	.dc.b	$8D,$E9,$EB,$C7,$E9,$57,$F1,$23,$87,$B5,$D0,$98,$7F,$70,$C1,$ED
	.dc.b	$F2,$8C,$63,$D5,$D6,$3D,$74,$F6,$63,$D6,$3F,$CF,$58,$31,$EB,$7E
	.dc.b	$89,$CC,$7B,$31,$ED,$39,$3F,$DD,$55,$EC,$7C,$27,$80,$F9,$7D,$1F
	.dc.b	$C1,$71,$FD,$76,$1D,$27,$D5,$87,$83,$1F,$8D,$8F,$CA,$F6,$FE,$5F
	.dc.b	$B7,$2F,$E0,$47,$AF,$2F,$EB,$F8,$B3,$7B,$5F,$48,$97,$F1,$FD,$BE
	.dc.b	$8F,$C9,$C7,$A3,$EB,$BE,$58,$F5,$8F,$D7,$61,$D1,$FA,$18,$74,$78
	.dc.b	$43,$0E,$B1,$8C,$3A,$E7,$F8,$1C,$3A,$4F,$E5,$E3,$F0,$98,$73,$6C
	.dc.b	$79,$CF,$BD,$ED,$FE,$30,$8F,$E0,$7D,$BF,$E7,$1E,$8F,$F3,$9F,$42
	.dc.b	$4F,$F4,$98,$FC,$9F,$B7,$F9,$4F,$6E,$4F,$E2,$63,$F2,$FE,$D7,$E6
	.dc.b	$3D,$A9,$3F,$8D,$8F,$E8,$9D,$D2,$3F,$AA,$3F,$D1,$FB,$FF,$3F,$ED
	.dc.b	$49,$FC,$A3,$FA,$2F,$6B,$F4,$7F,$15,$FA,$4F,$8A,$95,$F9,$98,$FE
	.dc.b	$A1,$DF,$23,$FB,$03,$FD,$4F,$BF,$F5,$7F,$13,$FA,$CF,$8A,$95,$FA
	.dc.b	$18,$FD,$7F,$C3,$48,$FD,$23,$FE,$51,$FC,$D5,$EE,$7F,$94,$FB,$9D
	.dc.b	$77,$B2,$61,$FE,$5A,$9C,$BF,$E5,$CB,$7C,$5F,$32,$46,$7E,$2F,$B7
	.dc.b	$C5,$77,$F8,$B9,$3C,$63,$C1,$D3,$E1,$EA,$93,$DE,$3C,$DF,$11,$E7
	.dc.b	$F8,$49,$3E,$18,$F5,$7C,$5F,$AF,$E2,$E4,$F9,$63,$D9,$CF,$23,$D4
	.dc.b	$FB,$FA,$3E,$17,$C5,$49,$F6,$3F,$17,$E2,$BE,$37,$C0,$49,$F8,$0F
	.dc.b	$C9,$F8,$1F,$F3,$D7,$E6,$BF,$0A,$7F,$CF,$ED,$7F,$50,$DF,$23,$FC
	.dc.b	$EE,$B8,$FE,$97,$C1,$7D,$3F,$83,$97,$F3,$63,$EB,$7C,$1F,$D7,$F7
	.dc.b	$64,$FD,$18,$F0,$F0,$78,$B8,$64,$F8,$A3,$EF,$7A,$12,$3E,$C3,$F8
	.dc.b	$1F,$83,$FC,$17,$C0,$C9,$FB,$8F,$E1,$7E,$0B,$DB,$9F,$BB,$49,$5E
	.dc.b	$E8,$FF,$4D,$7E,$3D,$FE,$32,$24,$63,$E0,$B3,$F0,$32,$3F,$D4,$72
	.dc.b	$8E,$3F,$05,$F2,$B2,$95,$F2,$9F,$9B,$97,$6F,$82,$EF,$F0,$52,$B9
	.dc.b	$C7,$8B,$E2,$BC,$7F,$15,$27,$BC,$7F,$AB,$FF,$14,$55,$24,$FF,$AD
	.dc.b	$A1,$7E,$CF,$5A,$4F,$A1,$F7,$FA,$FF,$03,$E2,$E4,$FB,$1F,$87,$F1
	.dc.b	$5D,$DF,$03,$23,$B4,$7C,$8F,$8B,$F9,$3F,$17,$27,$E2,$C7,$CC,$F6
	.dc.b	$7E,$6F,$B3,$27,$E4,$C7,$AF,$E2,$7D,$9F,$13,$27,$D5,$1F,$50,$8F
	.dc.b	$D0,$7E,$BC,$7D,$8F,$88,$FB,$3F,$13,$2B,$EA,$C7,$FB,$57,$F3,$5F
	.dc.b	$EE,$DE,$2F,$FB,$39,$BE,$A2,$8A,$33,$18,$7F,$B0,$3F,$D2,$7A,$DE
	.dc.b	$E7,$E2,$3D,$C9,$3F,$84,$8F,$F7,$27,$C1,$F2,$C3,$FA,$B9,$83,$3F
	.dc.b	$13,$E5,$8F,$37,$C2,$CB,$FF,$77,$E5,$FD,$4D,$FE,$B1,$3F,$DE,$5C
	.dc.b	$FC,$2F,$B4,$EF,$F1,$1E,$0F,$88,$F0,$FB,$D2,$BF,$DF,$37,$FE,$0A
	.dc.b	$7A,$FC,$2F,$C5,$F8,$69,$18,$3C,$3F,$0B,$F2,$7E,$1A,$46,$0E,$7F
	.dc.b	$05,$EE,$C2,$46,$0F,$4F,$C2,$FD,$12,$3E,$CF,$86,$FA,$9F,$0D,$23
	.dc.b	$DE,$FD,$37,$FB,$B7,$9B,$FC,$F1,$F4,$F1,$7C,$C2,$3E,$CE,$1E,$1D
	.dc.b	$1F,$6A,$3E,$DD,$3E,$0B,$3C,$DF,$7C,$43,$E0,$F3,$5C,$1B,$3A,$AB
	.dc.b	$4F,$80,$A4,$BF,$9B,$F0,$FF,$80,$F8,$3F,$C0,$FC,$24,$9F,$A3,$1C
	.dc.b	$BE,$06,$47,$E4,$74,$F4,$67,$E4,$9E,$FF,$03,$E0,$F8,$1F,$0F,$C0
	.dc.b	$CC,$38,$5F,$AF,$55,$7E,$43,$AB,$F2,$3D,$52,$FF,$03,$1E,$AF,$7A
	.dc.b	$47,$23,$E3,$BE,$EF,$7F,$DF,$EE,$FC,$0F,$76,$56,$F3,$F9,$33,$DE
	.dc.b	$3F,$1B,$8F,$77,$F2,$F8,$F6,$7E,$DE,$37,$7F,$7E,$9F,$E4,$69,$FA
	.dc.b	$F8,$76,$7F,$03,$97,$89,$F0,$7E,$6A,$F4,$3E,$AA,$F4,$69,$88,$CC
	.dc.b	$C3,$89,$C3,$0E,$3E,$E4,$8E,$AF,$4F,$73,$94,$78,$2F,$AB,$E2,$BE
	.dc.b	$8F,$F9,$2F,$0F,$F7,$C9,$30,$E9,$19,$C2,$E7,$F7,$A9,$8F,$6A,$E5
	.dc.b	$8F,$67,$A6,$3D,$AF,$84,$5D,$FF,$35,$87,$E6,$A3,$D2,$43,$C0,$7C
	.dc.b	$8C,$3E,$3F,$A5,$F2,$3D,$39,$3F,$32,$3A,$3F,$C8,$AB,$D6,$FC,$73
	.dc.b	$A6,$1E,$9C,$FD,$78,$F8,$B1,$EB,$5F,$5B,$1E,$AF,$C5,$C7,$AD,$31
	.dc.b	$EB,$3F,$73,$4E,$6F,$87,$1B,$5B,$C7,$AC,$FC,$FC,$7A,$E5,$F2,$F2
	.dc.b	$EB,$97,$4A,$FA,$99,$72,$8F,$EE,$D1,$EA,$C7,$D0,$FE,$33,$D2,$FC
	.dc.b	$5F,$A7,$23,$07,$E2,$08,$7D,$F3,$E2,$1F,$38,$98,$FC,$57,$E8,$1C
	.dc.b	$8F,$C3,$47,$57,$F1,$13,$E8,$7D,$38,$79,$F4,$F4,$69,$F9,$EA,$E1
	.dc.b	$87,$E2,$70,$C6,$1E,$48,$FC,$2B,$71,$FC,$2E,$3C,$E3,$E0,$1F,$77
	.dc.b	$0F,$D4,$FA,$7F,$AA,$F4,$E4,$FE,$BA,$3C,$8F,$33,$F0,$95,$CD,$F7
	.dc.b	$DF,$CA,$7F,$45,$3C,$9F,$D2,$57,$67,$94,$FE,$05,$F7,$9E,$47,$F0
	.dc.b	$99,$FC,$08,$F1,$1F,$C0,$27,$A6,$3F,$A9,$BF,$2C,$FD,$33,$F1,$47
	.dc.b	$3D,$3E,$FB,$E6,$3E,$01,$FC,$2B,$D2,$FE,$13,$E9,$C8,$C1,$E6,$9F
	.dc.b	$C5,$5F,$ED,$F3,$F8,$0F,$DD,$9F,$E0,$1A,$7F,$12,$C3,$94,$7F,$03
	.dc.b	$C3,$0F,$33,$DE,$78,$F4,$F7,$BE,$7A,$DB,$C4,$FE,$3F,$38,$BF,$BB
	.dc.b	$1E,$F2,$3E,$57,$E0,$69,$F2,$5F,$F0,$3F,$95,$FE,$12,$F2,$E4,$7F
	.dc.b	$80,$A3,$E4,$D7,$43,$94,$79,$74,$F9,$31,$EF,$3F,$8C,$63,$FE,$1C
	.dc.b	$AE,$75,$FE,$14,$BC,$3E,$F3,$E5,$69,$F2,$33,$FE,$2F,$97,$F2,$D9
	.dc.b	$F1,$1F,$E1,$4A,$E7,$8F,$F6,$2F,$29,$8F,$9B,$0F,$E3,$F7,$FB,$0A
	.dc.b	$F2,$47,$48,$F9,$47,$F8,$7B,$E1,$3F,$C6,$3F,$09,$23,$07,$94,$F9
	.dc.b	$33,$FA,$67,$E4,$9C,$8F,$F1,$7C,$F1,$7F,$C6,$33,$FE,$3C,$8F,$24
	.dc.b	$7F,$27,$3C,$93,$E3,$7E,$56,$5E,$F7,$99,$FE,$32,$AE,$97,$EF,$BE
	.dc.b	$77,$FA,$5A,$FC,$D9,$C8,$F2,$4F,$C5,$F3,$3F,$C9,$BE,$64,$9F,$F1
	.dc.b	$EC,$72,$DF,$FA,$94,$7B,$0F,$B9,$BF,$F9,$52,$78,$47,$F9,$32,$B0
	.dc.b	$FC,$5B,$FC,$26,$1F,$16,$3C,$47,$23,$B9,$FE,$4C,$9E,$73,$E5,$8E
	.dc.b	$53,$F6,$1F,$5C,$7D,$87,$F9,$F1,$EC,$8F,$F1,$A6,$AF,$33,$FC,$63
	.dc.b	$F9,$B7,$F9,$06,$97,$FE,$26,$8C,$3F,$87,$9F,$81,$3F,$48,$EA,$7A
	.dc.b	$23,$F3,$2F,$F9,$8A,$3A,$FC,$4C,$8C,$1E,$83,$EE,$1E,$8C,$3E,$E5
	.dc.b	$7A,$5F,$DC,$C7,$F0,$23,$89,$E6,$7F,$C9,$D7,$E6,$6D,$7F,$A1,$2F
	.dc.b	$CF,$5D,$4F,$3D,$F5,$D3,$F6,$37,$FC,$B6,$3F,$C8,$BF,$9B,$3C,$8F
	.dc.b	$F2,$AB,$6F,$FC,$68,$D7,$C4,$7F,$A7,$17,$93,$C2,$F8,$FA,$B2,$B8
	.dc.b	$C7,$4F,$52,$47,$47,$BF,$A9,$E0,$C6,$4F,$81,$F1,$E1,$D6,$3C,$78
	.dc.b	$75,$7C,$78,$31,$FF,$52,$1F,$E9,$B8,$CB,$72,$F7,$E5,$E0,$9E,$C4
	.dc.b	$37,$8F,$81,$F4,$9E,$02,$1D,$2F,$D6,$F6,$9B,$5F,$33,$2F,$97,$E3
	.dc.b	$4A,$F9,$6F,$CE,$F1,$FD,$91,$E6,$CF,$D9,$7F,$53,$1E,$67,$FA,$37
	.dc.b	$3F,$53,$72,$F5,$5F,$BB,$1E,$CD,$3E,$CE,$3D,$B0,$FB,$F8,$F6,$C3
	.dc.b	$A1,$DA,$BE,$E6,$16,$3E,$66,$5D,$2F,$F0,$F8,$D9,$F2,$63,$D5,$F3
	.dc.b	$63,$D7,$1F,$C3,$47,$3C,$BD,$31,$D8,$FC,$6E,$3F,$90,$E8,$91,$F9
	.dc.b	$07,$F2,$7D,$3D,$63,$F2,$18,$75,$C3,$F0,$07,$5D,$3B,$61,$CA,$BF
	.dc.b	$03,$85,$D3,$F3,$F8,$F3,$C3,$F1,$18,$F3,$CB,$F1,$39,$73,$C7,$CA
	.dc.b	$78,$C8,$74,$3F,$4B,$7C,$E7,$FD,$F7,$87,$CE,$9F,$C4,$9F,$85,$3F
	.dc.b	$40,$43,$F2,$B9,$7E,$C3,$9E,$47,$EC,$1F,$D9,$F4,$75,$CB,$CD,$9F
	.dc.b	$96,$BA,$1C,$8E,$04,$3F,$43,$5C,$9F,$1E,$1E,$57,$F2,$06,$27,$F1
	.dc.b	$8F,$96,$3F,$4E,$FC,$23,$99,$D5,$FE,$13,$3D,$63,$F4,$87,$E9,$A3
	.dc.b	$B1,$F9,$B7,$A4,$F6,$7D,$31,$FA,$77,$F8,$7F,$3F,$F1,$22,$3F,$C4
	.dc.b	$B9,$FE,$13,$D6,$BF,$32,$F4,$3F,$18,$74,$7F,$51,$97,$28,$FE,$07
	.dc.b	$56,$FF,$90,$DF,$49,$F8,$67,$A8,$E4,$74,$8F,$86,$FC,$F3,$99,$E8
	.dc.b	$21,$FC,$42,$3A,$47,$C2,$BF,$3C,$F4,$7F,$8C,$6B,$F9,$88,$FF,$71
	.dc.b	$73,$3E,$27,$FA,$B9,$23,$3F,$0F,$D2,$FF,$5F,$1F,$64,$F3,$3F,$37
	.dc.b	$0F,$39,$F3,$63,$F9,$3D,$F5,$7B,$DF,$57,$37,$D5,$FD,$79,$D6,$FC
	.dc.b	$EF,$CF,$6B,$F0,$EF,$FB,$55,$73,$6C,$7A,$0E,$5B,$FE,$9A,$BD,$31
	.dc.b	$F3,$E7,$CF,$1E,$92,$47,$F4,$FD,$3F,$BC,$FA,$32,$3E,$3B,$FD,$F3
	.dc.b	$D2,$F4,$57,$EF,$27,$BE,$9F,$A7,$C3,$E1,$1F,$DF,$EF,$94,$FC,$47
	.dc.b	$93,$E6,$E0,$FD,$7B,$EF,$CB,$F8,$CE,$1F,$02,$F9,$BF,$DD,$2F,$9C
	.dc.b	$75,$AF,$B7,$8F,$92,$FE,$04,$7F,$24,$3A,$63,$F8,$17,$EA,$FA,$5F
	.dc.b	$E1,$62,$3F,$84,$F8,$9E,$59,$FF,$0F,$AF,$66,$7F,$9E,$AF,$C4,$DF
	.dc.b	$07,$99,$C4,$F6,$1F,$A9,$3C,$47,$E4,$F0,$F6,$93,$4E,$59,$FF,$89
	.dc.b	$6A,$DF,$F8,$4B,$2F,$1E,$1F,$38,$F5,$9F,$C1,$EB,$FC,$59,$F0,$FF
	.dc.b	$E3,$02,$47,$E7,$BE,$23,$8B,$F8,$1B,$FB,$E7,$CD,$2E,$4C,$B8,$CF
	.dc.b	$E6,$AE,$E1,$FE,$39,$C3,$9B,$FC,$3E,$7D,$99,$FC,$D8,$F1,$C7,$EB
	.dc.b	$AB,$8E,$3F,$E2,$CB,$E2,$FE,$DA,$BF,$27,$5F,$AC,$24,$7F,$64,$7F
	.dc.b	$6F,$EA,$7F,$93,$3D,$49,$18,$07,$F8,$5B,$F2,$6F,$D8,$37,$8B,$2D
	.dc.b	$37,$83,$0E,$22,$9C,$7C,$27,$55,$B9,$71,$E6,$F8,$C3,$8F,$15,$B6
	.dc.b	$CF,$5A,$FF,$32,$FA,$9F,$E6,$6F,$56,$5F,$F1,$48,$DE,$0D,$EE,$F2
	.dc.b	$BD,$38,$F8,$27,$7B,$F0,$DE,$56,$8F,$44,$7F,$9D,$3E,$22,$47,$F2
	.dc.b	$27,$FC,$EF,$F1,$3E,$07,$C2,$C8,$E5,$87,$C5,$9F,$2F,$AD,$27,$39
	.dc.b	$A3,$D3,$EB,$FF,$A0,$9F,$F4,$1B,$27,$F9,$8C,$7B,$BE,$26,$47,$27
	.dc.b	$E0,$FC,$4F,$C8,$E5,$FB,$33,$FE,$52,$6F,$27,$D4,$CF,$EF,$CE,$7C
	.dc.b	$DF,$F4,$8F,$B3,$C7,$CE,$93,$FD,$02,$3A,$7C,$3F,$5F,$86,$93,$EE
	.dc.b	$C9,$3E,$1F,$88,$F1,$7C,$4C,$AD,$47,$97,$E1,$FC,$DF,$11,$27,$94
	.dc.b	$7A,$7D,$D9,$1D,$5F,$67,$BD,$ED,$F7,$7D,$DE,$F4,$AF,$14,$7C,$1F
	.dc.b	$7A,$47,$95,$F8,$9E,$F7,$C5,$F7,$64,$FA,$5F,$91,$EF,$7C,$9F,$7A
	.dc.b	$4F,$B5,$F9,$9E,$F7,$CD,$F8,$6F,$9D,$F0,$F2,$BE,$14,$7E,$63,$F3
	.dc.b	$6F,$D0,$C7,$FA,$09,$FE,$5B,$D4,$69,$AE,$BF,$A8,$FF,$3A,$3F,$86
	.dc.b	$4F,$E2,$9A,$F3,$F3,$3B,$7C,$3C,$8D,$3E,$1F,$87,$F1,$7C,$14,$9E
	.dc.b	$4F,$97,$E1,$FC,$DF,$0F,$E7,$F8,$79,$5D,$E3,$D5,$F0,$32,$3C,$4F
	.dc.b	$B7,$E0,$BD,$DF,$05,$EF,$F8,$29,$5E,$78,$F8,$5F,$05,$23,$D4,$FC
	.dc.b	$5F,$82,$E9,$1F,$17,$0B,$C3,$F0,$B1,$EB,$3F,$07,$1E,$AF,$CB,$C7
	.dc.b	$A3,$F3,$71,$B7,$F2,$B1,$E7,$C5,$FC,$5B,$8B,$FD,$8D,$3F,$3B,$3F
	.dc.b	$56,$1E,$73,$B1,$0E,$C7,$A0,$87,$73,$EA,$5F,$67,$F9,$25,$76,$BF
	.dc.b	$05,$76,$9E,$67,$E0,$F2,$F5,$E1,$F8,$0C,$BD,$37,$E7,$CA,$DF,$4C
	.dc.b	$BB,$4F,$E2,$71,$B9,$7F,$B7,$AB,$D4,$75,$3D,$58,$F5,$7D,$2F,$E3
	.dc.b	$B0,$EB,$7F,$93,$C3,$9E,$1D,$33,$FA,$38,$73,$26,$1C,$DF,$84,$47
	.dc.b	$A3,$F1,$0F,$CC,$7C,$0F,$E8,$FE,$0A,$46,$0F,$CE,$7C,$17,$E9,$FE
	.dc.b	$0A,$46,$0F,$D0,$7C,$27,$EA,$C8,$FF,$2C,$FC,$D9,$F0,$DF,$AF,$C7
	.dc.b	$AC,$FE,$BF,$1C,$3C,$CF,$D9,$61,$D9,$FD,$A6,$3D,$1F,$DB,$63,$72
	.dc.b	$FE,$4D,$97,$69,$F4,$1D,$4F,$DA,$61,$D9,$E6,$7E,$DF,$0E,$D5,$FB
	.dc.b	$CC,$3A,$6D,$F9,$3C,$6D,$FF,$6B,$C7,$B3,$FC,$1F,$2F,$5E,$1F,$C1
	.dc.b	$32,$EC,$F3,$9E,$C4,$3C,$27,$A8,$F9,$24,$0F,$B3,$3F,$ED,$B7,$FE
	.dc.b	$00,$8F,$7F,$BF,$83,$3E,$66,$BD,$ED,$FB,$F3,$1F,$A5,$1C,$BE,$0F
	.dc.b	$9F,$C1,$C9,$FA,$D1,$C9,$FB,$18,$75,$7E,$D6,$1D,$5F,$B9,$87,$57
	.dc.b	$CF,$BF,$F8,$3B,$3F,$F0,$4C,$74,$3F,$05,$8F,$77,$67,$CB,$C7,$DB
	.dc.b	$EF,$48,$FC,$43,$F0,$3D,$FF,$83,$EF,$7C,$2F,$7E,$57,$E3,$63,$E2
	.dc.b	$FC,$07,$C6,$F8,$19,$3F,$91,$8F,$93,$F0,$12,$3F,$28,$FC,$CF,$81
	.dc.b	$EB,$1F,$97,$C3,$AB,$E6,$D9,$B3,$2F,$A7,$EF,$CA,$FC,$F3,$F5,$BE
	.dc.b	$03,$EB,$FB,$92,$7F,$44,$FD,$AF,$88,$EF,$1D,$70,$65,$E0,$E7,$FC
	.dc.b	$DE,$5F,$80,$F8,$19,$7F,$AB,$7F,$07,$F0,$3F,$E3,$0F,$CD,$BF,$C0
	.dc.b	$FB,$7F,$8F,$97,$C1,$64,$D5,$EF,$E0,$B8,$7C,$4C,$8F,$F1,$0D,$27
	.dc.b	$A7,$C1,$75,$F8,$39,$5A,$8F,$07,$C1,$78,$7E,$0E,$4F,$28,$EC,$F4
	.dc.b	$C7,$B3,$F2,$F1,$EC,$F8,$3E,$0F,$D4,$47,$C7,$F0,$7D,$9F,$2E,$1D
	.dc.b	$9E,$58,$76,$7A,$61,$DB,$87,$FC,$51,$8F,$69,$F0,$63,$D5,$A7,$63
	.dc.b	$E0,$FC,$1F,$CB,$F8,$39,$18,$3E,$2E,$DD,$78,$3E,$76,$5E,$1C,$BB
	.dc.b	$E3,$D5,$B7,$D4,$F8,$1F,$05,$F5,$C8,$FD,$0F,$83,$FB,$5F,$07,$23
	.dc.b	$E9,$BF,$77,$E0,$FE,$F7,$C1,$C9,$FA,$EF,$E0,$7E,$0F,$9C,$76,$3E
	.dc.b	$DE,$1F,$85,$C3,$A4,$7D,$1C,$2E,$3F,$1B,$1E,$93,$F5,$71,$B3,$F2
	.dc.b	$B1,$FC,$77,$C1,$CA,$FC,$3B,$F9,$1F,$83,$FC,$97,$C1,$7E,$4F,$E0
	.dc.b	$E5,$7E,$32,$3F,$D0,$BF,$0F,$FE,$29,$E1,$FF,$49,$B6,$7F,$D2,$CD
	.dc.b	$83,$FE,$32,$E5,$63,$E1,$AF,$C5,$3E,$17,$B1,$F9,$AC,$3C,$2F,$E3
	.dc.b	$F0,$F0,$E1,$23,$FC,$FF,$87,$73,$FC,$7E,$7F,$BD,$0F,$F1,$CE,$7F
	.dc.b	$E7,$16,$CF,$A3,$E1,$3D,$7F,$09,$EC,$F8,$49,$87,$48,$F7,$FC,$24
	.dc.b	$8E,$EF,$C2,$F8,$4F,$87,$7F,$13,$E1,$A5,$79,$23,$AB,$E6,$C3,$AB
	.dc.b	$FC,$4F,$26,$3D,$63,$18,$DC,$FD,$19,$7D,$0F,$84,$FA,$3F,$0B,$2F
	.dc.b	$E0,$C7,$D4,$F8,$0F,$AB,$F0,$12,$7E,$2C,$7D,$8F,$84,$91,$F2,$1F
	.dc.b	$B7,$F0,$BD,$63,$E5,$E1,$D5,$F3,$61,$D5,$F9,$F8,$75,$7D,$58,$74
	.dc.b	$7D,$98,$5C,$BE,$86,$3F,$88,$F8,$4F,$C4,$FC,$24,$BF,$B3,$1F,$8C
	.dc.b	$F8,$1F,$C6,$FC,$0C,$9F,$BB,$1F,$90,$F8,$39,$1F,$80,$7F,$27,$F0
	.dc.b	$9D,$63,$F0,$78,$75,$7E,$76,$1D,$5F,$C3,$E1,$D5,$FA,$58,$74,$7E
	.dc.b	$A6,$17,$2F,$C4,$63,$FA,$0F,$84,$95,$F9,$07,$FD,$CF,$F8,$A7,$F8
	.dc.b	$2A,$BF,$D7,$CF,$F7,$DE,$9F,$F7,$BB,$6B,$FD,$F4,$D8,$3F,$EF,$A8
	.dc.b	$FF,$6C,$BF,$DC,$1F,$EB,$0F,$FB,$DD,$CF,$C0,$F8,$3E,$13,$C3,$F0
	.dc.b	$32,$78,$BE,$4F,$82,$F2,$FC,$1C,$9E,$AF,$A3,$E1,$39,$47,$42,$1E
	.dc.b	$0B,$F6,$FC,$24,$8F,$2B,$F0,$3E,$13,$A4,$79,$AE,$EB,$F0,$B0,$EB
	.dc.b	$3E,$AC,$FE,$39,$89,$F8,$FE,$D7,$CA,$F6,$E5,$7C,$28,$F9,$BF,$07
	.dc.b	$F3,$BE,$12,$4F,$C6,$8F,$A3,$F0,$92,$3E,$4B,$F5,$3E,$13,$EA,$FC
	.dc.b	$27,$D6,$F8,$49,$5F,$3A,$3A,$BF,$43,$0E,$AF,$D0,$C3,$A3,$F4,$B0
	.dc.b	$B9,$FD,$EC,$7A,$CF,$D8,$C7,$AB,$F6,$31,$EA,$FD,$AC,$7A,$DF,$E1
	.dc.b	$B1,$E7,$CB,$F8,$6C,$AD,$FE,$1B,$2F,$C6,$FC,$24,$8F,$C2,$BF,$90
	.dc.b	$F8,$4F,$C8,$FC,$1F,$E4,$BE,$0E,$57,$E2,$A0,$CC,$D9,$60,$7E,$6D
	.dc.b	$FE,$10,$9C,$BF,$E1,$07,$84,$6E,$31,$1F,$E2,$B7,$84,$7E,$6B,$1F
	.dc.b	$34,$FE,$61,$F3,$3E,$33,$F3,$B8,$79,$9F,$CE,$E1,$E6,$7A,$17,$0F
	.dc.b	$3B,$E4,$F8,$4C,$61,$E3,$6C,$7A,$0B,$1E,$CF,$76,$E5,$FE,$09,$BE
	.dc.b	$73,$9A,$F6,$9F,$03,$DD,$DF,$C1,$C8,$CC,$FC,$68,$E7,$F0,$9D,$3E
	.dc.b	$0B,$AF,$C1,$CB,$DE,$49,$F0,$E7,$27,$2F,$93,$2E,$D1,$F2,$F1,$B9
	.dc.b	$F9,$B0,$65,$DE,$BA,$E5,$DD,$FA,$59,$7B,$BE,$07,$DF,$F0,$92,$7C
	.dc.b	$71,$F0,$BE,$12,$47,$99,$F8,$BF,$09,$F1,$BE,$13,$E3,$FC,$24,$AF
	.dc.b	$5C,$75,$7D,$B8,$75,$7F,$05,$87,$47,$F0,$98,$5D,$7D,$58,$F5,$9F
	.dc.b	$8B,$8F,$57,$F1,$58,$FD,$6F,$84,$91,$F2,$9F,$B3,$F0,$9F,$6B,$E1
	.dc.b	$3E,$DF,$C2,$4A,$F9,$F1,$F7,$BE,$0A,$47,$D2,$7F,$03,$F0,$9D,$23
	.dc.b	$F3,$18,$5D,$3E,$76,$3D,$67,$EC,$E3,$D5,$FC,$F6,$3D,$1F,$D0,$63
	.dc.b	$6F,$EB,$63,$F8,$EF,$82,$FC,$7F,$C1,$CB,$FC,$1C,$7E,$53,$F1,$6F
	.dc.b	$F1,$21,$FE,$24,$8F,$C6,$B5,$FF,$4B,$3A,$F8,$3E,$1F,$09,$27,$71
	.dc.b	$DD,$FC,$7D,$F7,$7F,$69,$7D,$FB,$3C,$1F,$07,$27,$BC,$78,$EF,$A4
	.dc.b	$78,$30,$C3,$E4,$F8,$3F,$41,$1F,$3F,$C2,$7A,$FE,$13,$D9,$F0,$92
	.dc.b	$7D,$71,$EF,$F8,$3F,$81,$F0,$92,$7D,$F1,$CD,$F7,$E1,$78,$7F,$6B
	.dc.b	$C0,$CB,$B5,$7C,$6C,$BB,$3F,$1B,$2E,$CF,$CA,$CB,$B3,$F2,$B2,$FA
	.dc.b	$1F,$07,$F4,$7E,$12,$4F,$D0,$8F,$A9,$F0,$9F,$57,$E1,$24,$FD,$48
	.dc.b	$FB,$1F,$07,$F6,$7E,$12,$4F,$D8,$8E,$AF,$DA,$C3,$AB,$F6,$B0,$EA
	.dc.b	$FD,$EC,$3A,$BF,$7B,$0E,$AF,$E0,$B0,$EA,$FE,$0B,$0F,$C4,$7C,$1F
	.dc.b	$E2,$7E,$12,$4F,$E2,$23,$F1,$9F,$09,$F8,$DF,$84,$93,$F8,$C8,$FC
	.dc.b	$87,$C2,$7F,$24,$F3,$BF,$C2,$DD,$3F,$D0,$5B,$5F,$D1,$39,$55,$E0
	.dc.b	$8F,$CA,$E5,$E0,$7F,$2B,$97,$81,$FC,$CE,$5E,$07,$F3,$39,$78,$1F
	.dc.b	$E5,$38,$F6,$CB,$FD,$A9,$F0,$F9,$9F,$F7,$A4,$FF,$88,$C9,$EF,$F9
	.dc.b	$FD,$FF,$47,$C1,$7A,$7E,$02,$5F,$48,$F6,$7C,$27,$B7,$E1,$24,$F8
	.dc.b	$23,$E0,$61,$23,$C6,$F5,$79,$1E,$6B,$EA,$F9,$2F,$AB,$EA,$BE,$AF
	.dc.b	$A2,$FE,$67,$C2,$7C,$DF,$84,$93,$F0,$23,$E8,$7C,$1F,$D1,$F8,$49
	.dc.b	$3F,$12,$3E,$A7,$C0,$C8,$F8,$EF,$D7,$F8,$2E,$B1,$F2,$B0,$EA,$FC
	.dc.b	$8C,$3A,$BF,$3B,$0E,$AF,$CC,$C3,$A3,$F3,$B0,$BE,$1F,$F6,$2E,$06
	.dc.b	$5F,$85,$F8,$29,$7F,$61,$FC,$47,$C1,$7E,$27,$E0,$E4,$FD,$C7,$F1
	.dc.b	$9F,$09,$F8,$DF,$83,$93,$F8,$07,$F2,$1F,$09,$D2,$3E,$FE,$17,$3F
	.dc.b	$16,$3D,$27,$F0,$78,$D9,$F3,$63,$D2,$7F,$0F,$8D,$9F,$56,$3F,$9D
	.dc.b	$F8,$3F,$CF,$7C,$24,$BF,$C8,$C7,$F8,$13,$F3,$6F,$F0,$B9,$23,$1F
	.dc.b	$0B,$9D,$E4,$7F,$80,$27,$FC,$11,$D9,$FE,$27,$5F,$0B,$DE,$7F,$3B
	.dc.b	$8F,$77,$F3,$78,$F7,$61,$CC,$EE,$75,$C3,$BA,$CF,$FC,$2B,$87,$82
	.dc.b	$3F,$C6,$EF,$81,$F2,$7C,$2F,$AC,$8F,$9F,$E1,$3D,$DF,$07,$23,$D4
	.dc.b	$FC,$1F,$84,$F8,$5F,$07,$F0,$FE,$12,$57,$BE,$39,$BE,$3D,$FE,$06
	.dc.b	$38,$CF,$CB,$97,$5A,$D1,$E6,$CB,$13,$EA,$CB,$BC,$FC,$BC,$BB,$BC
	.dc.b	$31,$66,$CF,$EB,$7C,$24,$AF,$A4,$FD,$9F,$84,$FB,$5F,$07,$27,$EB
	.dc.b	$3F,$77,$E1,$3E,$F7,$C2,$7D,$FF,$84,$95,$F6,$E3,$A3,$E1,$C9,$8E
	.dc.b	$36,$FA,$39,$75,$AD,$65,$67,$EA,$E5,$DA,$7F,$0D,$97,$67,$A6,$5F
	.dc.b	$8F,$F8,$39,$1F,$8B,$7F,$25,$F0,$9F,$93,$F8,$4F,$CA,$7C,$24,$AF
	.dc.b	$C8,$47,$E5,$FE,$0E,$47,$E4,$DF,$CD,$7C,$27,$48,$F5,$61,$74,$FC
	.dc.b	$4E,$3D,$67,$F3,$58,$F5,$7E,$06,$3D,$1F,$85,$8D,$BF,$C8,$E3,$FE
	.dc.b	$6A,$FC,$DB,$FC,$E3,$5F,$5D,$B3,$FE,$6A,$6F,$1F,$FA,$2D,$FF,$0D
	.dc.b	$46,$7E,$13,$97,$C2,$4B,$FF,$35,$D2,$7B,$7C,$27,$7F,$84,$95,$A8
	.dc.b	$F1,$63,$23,$8B,$E5,$F8,$5F,$35,$F5,$9F,$36,$38,$79,$1E,$0C,$3A
	.dc.b	$BE,$BC,$3D,$DF,$09,$23,$CA,$FC,$1F,$85,$F8,$5F,$09,$F0,$FE,$16
	.dc.b	$57,$AA,$3E,$37,$C2,$7C,$7F,$85,$93,$EE,$8E,$8F,$C9,$C2,$F0,$7C
	.dc.b	$9C,$7A,$CF,$C4,$C7,$AB,$F3,$F1,$EA,$FC,$8C,$7A,$BF,$4F,$1F,$AD
	.dc.b	$F0,$9F,$5F,$E1,$64,$FC,$E8,$FB,$5F,$09,$23,$E8,$BF,$77,$E1,$7E
	.dc.b	$F7,$C1,$49,$FA,$AF,$E0,$7E,$0F,$AC,$7D,$8C,$3A,$BF,$83,$C3,$A3
	.dc.b	$F8,$5C,$2E,$DF,$85,$C7,$A4,$FE,$27,$1B,$3F,$89,$C7,$F1,$DF,$05
	.dc.b	$F8,$FF,$83,$97,$F8,$68,$FC,$97,$C2,$7F,$AF,$3F,$36,$FF,$05,$1F
	.dc.b	$E0,$BE,$AF,$F0,$BD,$7F,$AD,$A1,$F0,$7B,$24,$67,$E1,$3B,$BF,$92
	.dc.b	$C7,$BB,$F9,$DC,$7B,$BC,$FE,$17,$C0,$61,$FE,$31,$87,$67,$C1,$87
	.dc.b	$67,$38,$79,$FE,$12,$47,$95,$F5,$7C,$2F,$AF,$E1,$3D,$9F,$0B,$2B
	.dc.b	$D5,$1E,$FF,$84,$91,$ED,$7E,$17,$C2,$F4,$8F,$1E,$17,$8B,$FD,$BB
	.dc.b	$A3,$2E,$D5,$F1,$72,$EC,$FA,$B2,$EA,$FB,$32,$B8,$72,$CB,$E8,$7C
	.dc.b	$24,$AF,$9C,$FD,$3F,$85,$FA,$9F,$0B,$F5,$7E,$16,$57,$D3,$8F,$B1
	.dc.b	$F0,$BF,$67,$E1,$64,$FD,$78,$E8,$FC,$CC,$2E,$5E,$AC,$7A,$CF,$DD
	.dc.b	$C7,$AB,$F4,$B1,$EA,$FE,$07,$1E,$AF,$D6,$C7,$F1,$1F,$0B,$F8,$9F
	.dc.b	$85,$93,$F8,$78,$FC,$67,$C2,$C8,$FC,$53,$F8,$FF,$85,$FC,$87,$C1
	.dc.b	$49,$FC,$73,$F9,$3F,$83,$EB,$1F,$91,$C3,$AB,$F8,$6C,$3A,$3F,$88
	.dc.b	$C2,$ED,$F5,$B1,$E9,$3F,$8C,$C6,$CF,$DA,$C7,$AC,$FE,$43,$1E,$B3
	.dc.b	$F7,$B1,$E7,$E7,$A8,$B9,$99,$EA,$73,$0C,$FB,$D7,$13,$86,$78,$AD
	.dc.b	$E7,$D5,$FE,$F1,$D1,$5B,$F0,$B9,$D3,$C3,$1E,$29,$4E,$70,$ED,$1C
	.dc.b	$6B,$70,$8F,$F1,$03,$7E,$17,$85,$F3,$6C,$6F,$E1,$6D,$3E,$16,$92
	.dc.b	$FE,$26,$1D,$A5,$F0,$BC,$27,$8E,$51,$DE,$3E,$67,$AE,$3D,$B2,$91
	.dc.b	$A3,$DD,$1D,$DF,$8B,$95,$CB,$C1,$8F,$AA,$5F,$3A,$E9,$FF,$20,$47
	.dc.b	$CB,$32,$E1,$F9,$D1,$BA,$D4,$69,$B1,$98,$E4,$D7,$8C,$F3,$7E,$D4
	.dc.b	$65,$B1,$97,$F1,$51,$96,$C6,$5F,$01,$DB,$A5,$F3,$1E,$26,$C6,$63
	.dc.b	$04,$70,$43,$DB,$1C,$1B,$1B,$69,$B1,$F4,$10,$00

;--------------------------------------------------------------------------------
;	FMOD.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmod_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FMOD.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOD.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fmod_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fmod.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOD.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fmod.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOD.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmod_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3D,$61,$F3,$7F,$DA,$BD,$E7,$99,$C6,$30,$DA,$FE
	.dc.b	$50,$D9,$DC,$7F,$72,$F8,$17,$97,$FD,$A2,$FF,$B9,$CA,$73,$F2,$7E
	.dc.b	$08,$FE,$EF,$18,$9D,$B6,$33,$5C,$DB,$5C,$5F,$E2,$6D,$7C,$51,$B7
	.dc.b	$35,$D2,$31,$1A,$9E,$0F,$96,$32,$D8,$CD,$78,$E3,$99,$E2,$AE,$93
	.dc.b	$AA,$E3,$5B,$AE,$D5,$A2,$3C,$48,$74,$21,$C0,$C9,$4C,$9E,$57,$9B
	.dc.b	$88,$F0,$39,$38,$94,$D4,$F9,$89,$5E,$92,$57,$49,$F0,$10,$E4,$5A
	.dc.b	$EE,$4A,$E7,$3C,$67,$64,$9D,$57,$68,$F2,$C6,$27,$83,$63,$35,$D1
	.dc.b	$B5,$C5,$F5,$B5,$F2,$46,$DC,$D7,$48,$C4,$6A,$78,$3E,$58,$CB,$63
	.dc.b	$35,$E3,$8E,$67,$8A,$BA,$4E,$AB,$8D,$6E,$BB,$56,$88,$F1,$21,$D0
	.dc.b	$87,$03,$25,$32,$79,$5E,$6E,$23,$C0,$E4,$E2,$53,$53,$E6,$25,$7A
	.dc.b	$49,$5D,$27,$C0,$43,$91,$6B,$B9,$2B,$9C,$F1,$9D,$92,$75,$5D,$A3
	.dc.b	$CB,$18,$9E,$0D,$8C,$D7,$46,$D7,$17,$D6,$D7,$C9,$1B,$73,$5D,$23
	.dc.b	$11,$A9,$E0,$F9,$63,$2D,$8C,$D7,$8E,$39,$9E,$2A,$E9,$3A,$AE,$35
	.dc.b	$BA,$ED,$5A,$23,$C4,$87,$42,$1C,$0C,$94,$C9,$E5,$79,$B8,$8F,$03
	.dc.b	$93,$89,$4D,$4F,$98,$95,$E9,$25,$74,$9F,$01,$0E,$45,$D3,$FD,$25
	.dc.b	$1F,$D1,$A3,$FA,$63,$5F,$EA,$B9,$7F,$3F,$C7,$FA,$3C,$62,$75,$5C
	.dc.b	$A3,$FA,$69,$C8,$A6,$CA,$EE,$79,$E2,$F8,$0F,$F3,$95,$FF,$4A,$6D
	.dc.b	$FF,$4E,$6B,$FE,$59,$F8,$EF,$E7,$B1,$88,$FE,$B4,$D7,$FA,$C3,$6B
	.dc.b	$B4,$F0,$25,$77,$6D,$77,$6B,$DE,$34,$47,$4D,$74,$60,$86,$08,$79
	.dc.b	$A3,$AB,$63,$93,$4E,$25,$38,$9E,$C8,$E2,$D8,$E2,$D3,$89,$4E,$33
	.dc.b	$F0,$A3,$6D,$8D,$B5,$DD,$7C,$92,$57,$82,$75,$3E,$52,$1D,$CA,$E9
	.dc.b	$C1,$1C,$12,$7A,$D7,$91,$B3,$E4,$6B,$E4,$AF,$14,$62,$3B,$B5,$EC
	.dc.b	$DA,$ED,$3C,$09,$5D,$DB,$5D,$DA,$F7,$8D,$11,$D3,$5D,$18,$21,$82
	.dc.b	$1E,$68,$EA,$D8,$E4,$D3,$89,$4E,$27,$B2,$38,$B6,$38,$B4,$E2,$53
	.dc.b	$8C,$FC,$28,$DB,$63,$6D,$77,$5F,$24,$95,$E0,$9D,$4F,$94,$87,$72
	.dc.b	$BA,$70,$47,$04,$9E,$B5,$E4,$6C,$F9,$1A,$F9,$2B,$C5,$18,$8E,$ED
	.dc.b	$7B,$36,$BB,$4F,$02,$57,$76,$D7,$76,$BD,$E3,$44,$74,$D7,$46,$08
	.dc.b	$60,$87,$9A,$3A,$B6,$39,$34,$E2,$53,$89,$EC,$8E,$2D,$8E,$2D,$38
	.dc.b	$94,$E3,$3F,$0A,$36,$D8,$DB,$5D,$D7,$C9,$3F,$96,$61,$FC,$E7,$4F
	.dc.b	$F4,$51,$27,$FA,$06,$3F,$D1,$A7,$FA,$24,$0F,$F4,$91,$4F,$E8,$C5
	.dc.b	$3F,$A3,$15,$EC,$F4,$9E,$CF,$F4,$B9,$ED,$97,$F5,$59,$CD,$70,$AD
	.dc.b	$46,$27,$51,$CC,$86,$8B,$3C,$49,$1C,$23,$44,$8C,$E3,$FD,$8A,$75
	.dc.b	$18,$9F,$EB,$F1,$82,$4E,$C9,$1A,$6D,$FB,$1B,$7E,$96,$CF,$9A,$B9
	.dc.b	$BC,$49,$5C,$5B,$5C,$5A,$71,$AE,$6F,$AE,$B9,$D7,$12,$57,$16,$CF
	.dc.b	$13,$99,$C6,$B9,$DF,$C1,$9D,$CE,$4D,$C6,$6B,$A1,$23,$A4,$F0,$9D
	.dc.b	$46,$27,$4F,$39,$D5,$FA,$27,$35,$C2,$B5,$18,$9D,$47,$32,$1A,$2C
	.dc.b	$F1,$24,$70,$8D,$12,$33,$7F,$02,$75,$18,$9F,$7C,$60,$93,$B2,$46
	.dc.b	$9B,$7E,$C6,$DF,$A5,$B3,$E6,$AE,$6F,$12,$57,$16,$D7,$16,$9C,$6B
	.dc.b	$9B,$EB,$AE,$75,$C4,$95,$C5,$B3,$C4,$E6,$71,$AE,$77,$F0,$67,$73
	.dc.b	$93,$71,$9A,$E8,$48,$E9,$3C,$27,$51,$89,$D3,$CE,$75,$7E,$89,$CD
	.dc.b	$70,$AD,$46,$27,$51,$CC,$86,$8B,$3C,$49,$1C,$23,$44,$8C,$DF,$C0
	.dc.b	$9D,$46,$27,$DF,$18,$24,$EC,$91,$A6,$DF,$B1,$B7,$E9,$6C,$F9,$AB
	.dc.b	$9B,$C4,$95,$C5,$B5,$C5,$A7,$1A,$E6,$FA,$EB,$9D,$71,$25,$71,$6C
	.dc.b	$F1,$39,$9C,$6B,$9D,$FC,$19,$DC,$E4,$DC,$66,$BA,$12,$3A,$4F,$09
	.dc.b	$D4,$62,$74,$F3,$9D,$5F,$A2,$73,$5C,$2B,$51,$89,$D4,$73,$21,$A2
	.dc.b	$CF,$12,$47,$08,$D1,$23,$37,$F0,$27,$51,$89,$F7,$C6,$09,$3B,$24
	.dc.b	$69,$B7,$EC,$6D,$FA,$5B,$3E,$6A,$E6,$F1,$25,$71,$6D,$71,$69,$C6
	.dc.b	$B9,$BE,$BA,$E7,$5C,$49,$5C,$5B,$3C,$4E,$67,$1A,$E7,$7F,$06,$77
	.dc.b	$39,$37,$19,$AE,$84,$8F,$E9,$87,$F8,$06,$D7,$FE,$05,$8F,$F0,$39
	.dc.b	$FE,$0A,$3F,$C1,$19,$BE,$0F,$51,$FE,$0F,$8C,$38,$6B,$CD,$AF,$FB
	.dc.b	$B1,$EE,$FF,$84,$FC,$D6,$5D,$6F,$C5,$18,$BD,$D7,$F8,$6E,$31,$3A
	.dc.b	$AE,$13,$E9,$F8,$9A,$FF,$87,$DA,$FF,$88,$DA,$FF,$89,$DB,$AB,$76
	.dc.b	$4F,$9B,$F8,$93,$98,$C5,$FF,$8A,$E3,$13,$AA,$E2,$F9,$73,$DD,$7A
	.dc.b	$71,$F9,$53,$92,$47,$F8,$DE,$5F,$78,$F3,$FE,$87,$E0,$7F,$E3,$99
	.dc.b	$CC,$62,$B8,$56,$E5,$5B,$8C,$37,$C2,$FA,$75,$F6,$E5,$18,$AF,$0B
	.dc.b	$63,$31,$F8,$53,$CB,$E4,$FC,$C6,$9F,$5A,$BA,$46,$64,$EF,$EF,$BF
	.dc.b	$CA,$D4,$FC,$83,$D9,$18,$6D,$FE,$3D,$B3,$B8,$FA,$1A,$FD,$67,$E8
	.dc.b	$6F,$F9,$4D,$FE,$B4,$60,$87,$08,$D4,$F1,$8C,$1F,$E6,$5F,$9D,$7E
	.dc.b	$70,$ED,$FD,$07,$C4,$E6,$31,$92,$31,$3A,$BF,$F3,$85,$F2,$8C,$47
	.dc.b	$F9,$CF,$2F,$2B,$E0,$21,$9C,$7C,$8F,$F9,$E7,$E9,$DE,$8F,$EC,$72
	.dc.b	$FD,$B4,$61,$FB,$6D,$73,$1B,$9F,$C6,$6D,$F9,$C9,$FB,$4D,$8F,$A3
	.dc.b	$AF,$E0,$63,$11,$E4,$3F,$D1,$98,$FA,$1F,$34,$FC,$A3,$9F,$CA,$7D
	.dc.b	$78,$EE,$D7,$EF,$D2,$31,$1F,$C2,$E3,$17,$F7,$63,$7F,$29,$FC,$3F
	.dc.b	$E9,$3F,$0A,$D8,$F5,$B5,$DC,$A7,$EF,$C7,$98,$A7,$AB,$4F,$9A,$F8
	.dc.b	$7E,$57,$F8,$E7,$D2,$7E,$49,$B1,$F2,$1A,$FA,$9A,$F3,$CF,$E1,$3D
	.dc.b	$86,$9F,$1D,$D1,$97,$0E,$AF,$C6,$D9,$E6,$FD,$13,$07,$E9,$BA,$FE
	.dc.b	$B1,$32,$FB,$6F,$EF,$23,$67,$1B,$EE,$78,$E5,$A2,$B9,$57,$59,$E6
	.dc.b	$7F,$3E,$FA,$47,$E9,$9F,$CA,$1B,$1F,$BF,$94,$63,$BB,$FA,$53,$63
	.dc.b	$C5,$9F,$BB,$2F,$1B,$F0,$27,$9F,$3F,$0C,$7F,$0B,$7D,$1B,$3A,$8F
	.dc.b	$CD,$F2,$7E,$5C,$99,$7E,$71,$F8,$58,$F6,$D7,$E5,$D7,$E0,$CF,$D1
	.dc.b	$FE,$07,$F9,$2B,$F8,$6D,$D8,$D3,$1F,$1B,$FD,$87,$E8,$7F,$8B,$B6
	.dc.b	$FF,$5A,$D7,$F3,$ED,$7F,$02,$7C,$2C,$3C,$6E,$0F,$21,$0C,$1F,$A5
	.dc.b	$F8,$DF,$EE,$1F,$47,$FC,$99,$B1,$FB,$C6,$BF,$AE,$6B,$FA,$DF,$57
	.dc.b	$FB,$C7,$B9,$FC,$B9,$B1,$FC,$11,$AF,$ED,$9A,$FE,$D7,$E6,$FF,$C0
	.dc.b	$1F,$3B,$FC,$D9,$B1,$FC,$31,$AF,$EF,$9A,$FE,$F7,$E7,$3F,$C1,$1F
	.dc.b	$5D,$FC,$F9,$B1,$FC,$51,$AF,$F0,$66,$BF,$C1,$7E,$4B,$FC,$21,$F0
	.dc.b	$3F,$D1,$9B,$1F,$C7,$1A,$FF,$0E,$6B,$F2,$8F,$D1,$9F,$34,$A7,$F1
	.dc.b	$1F,$92,$FF,$0E,$7E,$1D,$FD,$49,$B1,$FC,$A1,$AF,$F1,$A6,$BF,$C6
	.dc.b	$7C,$2F,$7F,$2F,$EB,$1B,$CB,$FC,$D7,$5F,$EB,$B1,$F1,$2B,$C2,$4E
	.dc.b	$6F,$CB,$3F,$D8,$BE,$7B,$FC,$5D,$F5,$37,$13,$49,$3F,$CF,$3E,$57
	.dc.b	$FC,$69,$F8,$87,$F6,$A6,$C6,$69,$39,$6C,$7F,$44,$F9,$6F,$F1,$E7
	.dc.b	$D7,$7F,$72,$6C,$66,$53,$96,$C7,$E6,$CF,$E3,$9E,$0F,$EF,$49,$D5
	.dc.b	$F9,$E8,$F9,$F3,$FC,$43,$E5,$FF,$C9,$5E,$FF,$F8,$01,$B3,$FD,$59
	.dc.b	$AF,$F4,$86,$BF,$D1,$FE,$87,$FC,$A5,$F7,$9F,$E0,$86,$C7,$F5,$E6
	.dc.b	$BF,$D4,$1A,$FF,$4F,$FA,$2E,$78,$7F,$22,$6F,$21,$A7,$82,$7E,$B6
	.dc.b	$7E,$C7,$FA,$5D,$3F,$05,$FE,$E9,$F7,$1F,$E3,$56,$CF,$F8,$ED,$AF
	.dc.b	$F8,$81,$AF,$E5,$79,$3F,$02,$FC,$B2,$1F,$C0,$EB,$B7,$D2,$7F,$50
	.dc.b	$68,$F3,$7C,$11,$9B,$3F,$D5,$AB,$D3,$C1,$F4,$F0,$FE,$85,$5F,$E6
	.dc.b	$DF,$A6,$F4,$7B,$9F,$E7,$D6,$E4,$69,$F7,$AF,$96,$9F,$0F,$1F,$39
	.dc.b	$EB,$70,$FE,$BF,$1E,$F1,$FE,$89,$FC,$83,$E3,$36,$7E,$33,$5F,$8C
	.dc.b	$D7,$FD,$21,$93,$E0,$BF,$D2,$5F,$91,$7C,$F6,$CF,$CF,$6B,$F3,$DA
	.dc.b	$F9,$71,$F4,$91,$F6,$10,$F5,$9F,$20,$C1,$F6,$7E,$4B,$FD,$45,$EC
	.dc.b	$7D,$F6,$C7,$DF,$6B,$BD,$53,$F8,$5F,$98,$FF,$37,$64,$69,$27,$C9
	.dc.b	$5E,$2D,$7F,$0A,$FC,$BD,$3F,$C3,$33,$F9,$1F,$A7,$FF,$5A,$7E,$1F
	.dc.b	$F9,$16,$D7,$E4,$5A,$FE,$45,$AF,$C3,$3D,$A5,$3E,$DC,$FE,$77,$DC
	.dc.b	$FF,$60,$7C,$C7,$FA,$F2,$30,$1C,$67,$FE,$63,$8C,$1F,$A1,$9E,$2F
	.dc.b	$AF,$1F,$C2,$DF,$DC,$FA,$1F,$F6,$67,$C6,$F2,$6D,$7E,$A5,$B3,$FA
	.dc.b	$96,$BF,$84,$F8,$9F,$F6,$8C,$FE,$BD,$B1,$FA,$F6,$BB,$8F,$F4,$8B
	.dc.b	$FB,$3F,$A1,$FF,$47,$47,$ED,$DB,$19,$8F,$F4,$B1,$92,$BF,$4F,$8B
	.dc.b	$EB,$BD,$E3,$AB,$DC,$F8,$B5,$E2,$8F,$E1,$0F,$92,$3F,$1C,$7C,$59
	.dc.b	$E6,$FF,$59,$F8,$BF,$F7,$D2,$0B,$E2,$FF,$DD,$7D,$2B,$D5,$6A,$73
	.dc.b	$18,$AE,$1E,$EF,$3B,$E1,$7F,$E5,$BA,$FF,$55,$91,$FF,$54,$35,$FF
	.dc.b	$40,$B5,$E0,$7D,$1C,$7C,$45,$8F,$82,$E2,$3E,$34,$23,$AC,$43,$A9
	.dc.b	$4F,$E0,$39,$6F,$E3,$7F,$1E,$D9,$FA,$91,$84,$F8,$2A,$47,$8C,$95
	.dc.b	$F8,$B5,$E5,$7E,$45,$A7,$E4,$4B,$1B,$8F,$43,$E6,$21,$F0,$8B,$1F
	.dc.b	$A6,$FD,$4B,$F4,$BC,$3E,$F6,$C1,$A7,$49,$E4,$43,$FD,$09,$7F,$3D
	.dc.b	$A7,$C5,$29,$E0,$E2,$72,$7C,$DF,$DA,$BF,$82,$60,$69,$2B,$97,$3B
	.dc.b	$95,$C7,$F8,$49,$F6,$9D,$8F,$24,$EE,$35,$18,$9E,$39,$7F,$1F,$C3
	.dc.b	$27,$F6,$7E,$8F,$C0,$46,$49,$5F,$CA,$65,$F9,$4F,$F2,$D6,$C6,$63
	.dc.b	$07,$C3,$C7,$C5,$57,$0F,$0D,$7B,$1A,$FA,$A7,$6E,$71,$F4,$BE,$D2
	.dc.b	$46,$9E,$13,$C9,$FA,$B8,$7F,$80,$7B,$3F,$35,$FC,$03,$FD,$4E,$0D
	.dc.b	$24,$FF,$5E,$87,$F0,$07,$F6,$26,$C6,$63,$07,$AF,$1F,$2E,$3E,$39
	.dc.b	$F9,$D1,$8C,$7C,$91,$E8,$6C,$EE,$31,$3B,$3E,$9E,$EE,$5F,$D9,$7C
	.dc.b	$57,$F8,$00,$95,$FE,$19,$8C,$BF,$1F,$5F,$21,$FE,$10,$CD,$F9,$95
	.dc.b	$24,$FF,$88,$7C,$4F,$DF,$47,$08,$E4,$43,$E6,$E5,$E4,$AF,$E0,$9F
	.dc.b	$A5,$7F,$8A,$DB,$3F,$E2,$B6,$BF,$E2,$B6,$BF,$13,$0E,$CF,$F2,$78
	.dc.b	$FF,$1F,$0F,$D6,$FF,$92,$C7,$F9,$0E,$30,$FF,$28,$77,$1A,$6C,$7F
	.dc.b	$93,$57,$EA,$3F,$E5,$16,$C6,$63,$07,$F9,$76,$3A,$C7,$F9,$6D,$AF
	.dc.b	$F2,$1F,$2D,$E7,$7F,$23,$BE,$07,$AA,$7D,$87,$C7,$29,$ED,$97,$A3
	.dc.b	$FC,$AF,$F3,$BF,$F3,$B1,$27,$FC,$F4,$FF,$A0,$7F,$71,$91,$FE,$85
	.dc.b	$6B,$CF,$BB,$F9,$A7,$E7,$1F,$E9,$12,$47,$C5,$C3,$F9,$2C,$79,$63
	.dc.b	$F1,$BA,$FF,$A7,$B4,$7E,$89,$49,$3C,$7D,$9F,$E8,$11,$FE,$AF,$97
	.dc.b	$EE,$14,$95,$FE,$B1,$8C,$BF,$04,$EE,$7F,$B0,$57,$EB,$7F,$EC,$56
	.dc.b	$C6,$63,$07,$FB,$46,$1F,$A1,$52,$4F,$FB,$4E,$32,$FF,$B7,$5F,$5B
	.dc.b	$6F,$FD,$C5,$18,$8F,$F7,$03,$5F,$F7,$82,$FD,$6F,$FD,$E0,$D8,$FF
	.dc.b	$78,$35,$FF,$78,$35,$FF,$80,$0B,$F3,$9F,$F8,$00,$D8,$FF,$80,$0D
	.dc.b	$7F,$E0,$03,$5F,$C2,$90,$F8,$E7,$EB,$B3,$FC,$E6,$AF,$8B,$FE,$E5
	.dc.b	$F9,$DE,$1B,$5F,$7D,$AF,$E3,$DA,$FE,$67,$E3,$BF,$BA,$3F,$85,$6B
	.dc.b	$F9,$56,$BB,$C3,$FB,$AB,$F6,$FB,$FF,$8B,$1F,$C7,$61,$83,$E3,$FF
	.dc.b	$A3,$D7,$E4,$5A,$FE,$85,$AF,$EA,$DA,$BE,$41,$F2,$3E,$28,$C3,$63
	.dc.b	$37,$E4,$9F,$F1,$4C,$63,$CD,$FC,$F6,$3E,$6F,$13,$F9,$7C,$F0,$FC
	.dc.b	$DB,$F8,$5D,$F4,$C1,$18,$9D,$4F,$F1,$06,$BF,$B9,$F9,$4F,$4D,$7C
	.dc.b	$06,$E8,$69,$EC,$C7,$C5,$E0,$FF,$3B,$E1,$FF,$19,$5F,$13,$FB,$7D
	.dc.b	$F1,$6D,$F2,$78,$61,$F4,$73,$FE,$8E,$4C,$33,$1F,$87,$34,$F4,$F9
	.dc.b	$5F,$E7,$71,$FC,$0D,$AF,$F1,$36,$BF,$C7,$5A,$FE,$48,$FC,$AD,$7D
	.dc.b	$EF,$97,$FF,$21,$7C,$77,$E5,$1B,$3F,$C3,$DA,$FF,$1F,$6B,$FC,$A7
	.dc.b	$E6,$3F,$1B,$5F,$88,$6E,$66,$9F,$81,$D7,$E4,$EB,$FC,$B6,$BF,$1F
	.dc.b	$9F,$D0,$23,$F2,$9F,$C5,$FC,$CB,$B7,$FC,$B1,$F7,$FF,$80,$8C,$52
	.dc.b	$31,$3A,$CF,$FA,$7D,$6A,$7F,$17,$BF,$CF,$C7,$FA,$9F,$BC,$F3,$3F
	.dc.b	$41,$93,$E7,$73,$18,$6F,$47,$F9,$A5,$B3,$BC,$FE,$BE,$1D,$B5,$FB
	.dc.b	$71,$FE,$18,$E0,$FE,$C9,$CF,$FC,$0A,$3E,$39,$2F,$F4,$4F,$F9,$D7
	.dc.b	$E7,$5F,$3B,$FE,$77,$E9,$FE,$C9,$18,$AF,$CF,$D7,$F6,$98,$C4,$EA
	.dc.b	$B8,$4E,$73,$FC,$FD,$7B,$6F,$FC,$8A,$DC,$FA,$CE,$BE,$7F,$FD,$0D
	.dc.b	$FB,$57,$81,$B8,$77,$6B,$FD,$FE,$D1,$88,$FF,$54,$7D,$17,$F8,$7F
	.dc.b	$17,$E7,$1F,$B8,$8C,$37,$D3,$FF,$4A,$36,$77,$5F,$E0,$8E,$8F,$F3
	.dc.b	$04,$FF,$00,$6C,$67,$97,$FC,$D3,$39,$6C,$7F,$06,$ED,$F8,$6F,$F8
	.dc.b	$49,$AF,$FA,$E1,$AF,$FA,$FD,$B8,$FA,$78,$3C,$84,$9F,$F0,$DC,$65
	.dc.b	$F1,$71,$FD,$47,$FD,$87,$8F,$F9,$DE,$BC,$93,$99,$FF,$66,$4E,$5F
	.dc.b	$E3,$71,$FC,$A9,$FF,$41,$E5,$FE,$66,$6F,$07,$F2,$06,$9F,$E2,$CF
	.dc.b	$80,$FF,$66,$BF,$C9,$1A,$E7,$E0,$5F,$03,$C6,$73,$3F,$E3,$4E,$DF
	.dc.b	$F5,$13,$70,$EC,$FE,$ED,$B1,$A8,$FF,$6F,$B5,$FF,$73,$35,$E0,$79
	.dc.b	$DD,$65,$FE,$99,$BF,$86,$49,$F9,$2F,$E1,$7E,$1B,$FD,$AC,$DB,$F4
	.dc.b	$CF,$D0,$9C,$B6,$32,$FF,$BC,$B5,$7B,$7A,$8C,$35,$FC,$AB,$54,$70
	.dc.b	$9E,$71,$F5,$B8,$3F,$D6,$BF,$47,$F6,$89,$1F,$80,$7F,$CB,$1E,$D3
	.dc.b	$EA,$FF,$DD,$DF,$69,$86,$D7,$F9,$7E,$75,$82,$B4,$D9,$D7,$D8,$FC
	.dc.b	$F8,$52,$30,$D8,$FE,$C6,$6A,$33,$F5,$BF,$EF,$FF,$C1,$FC,$AD,$BF
	.dc.b	$13,$5F,$F3,$C5,$A3,$10,$BF,$0F,$D7,$5F,$A7,$F7,$B6,$7D,$8D,$79
	.dc.b	$B5,$F9,$5C,$FF,$99,$E3,$FF,$6D,$47,$E9,$08,$7F,$A1,$B2,$79,$5E
	.dc.b	$1F,$8F,$FE,$FF,$18,$C9,$18,$9D,$57,$16,$A2,$7E,$10,$F9,$6F,$67
	.dc.b	$A5,$FE,$0A,$8C,$61,$FE,$35,$AF,$F4,$BC,$62,$75,$5C,$27,$F7,$E7
	.dc.b	$A0,$FD,$FF,$27,$FB,$E9,$C1,$F0,$0A,$33,$F8,$2F,$23,$F0,$9F,$80
	.dc.b	$FD,$C2,$63,$F7,$38,$51,$FA,$AB,$ED,$5C,$4F,$0E,$DF,$36,$33,$1F
	.dc.b	$73,$F0,$1F,$A5,$87,$E2,$9B,$81,$A6,$3E,$C7,$F0,$1F,$59,$F9,$C6
	.dc.b	$DF,$E6,$5A,$FE,$21,$AF,$BC,$F2,$14,$F2,$9F,$A6,$FB,$1F,$C6,$7C
	.dc.b	$CD,$C4,$D2,$4F,$FB,$0F,$C1,$FE,$1D,$26,$91,$F8,$1A,$49,$F9,$FB
	.dc.b	$7C,$E7,$F0,$43,$F4,$5F,$81,$5F,$8A,$8F,$F6,$AC,$63,$93,$FC,$9A
	.dc.b	$D8,$CE,$BF,$8C,$7F,$19,$93,$EB,$3F,$91,$37,$0F,$E6,$CD,$9F,$F2
	.dc.b	$BB,$5F,$8A,$7E,$0F,$83,$EB,$BF,$1C,$87,$C4,$AE,$A4,$AE,$AF,$F3
	.dc.b	$3F,$49,$F6,$AF,$0B,$F7,$1E,$07,$F8,$7A,$31,$68,$C4,$EB,$0F,$8B
	.dc.b	$7C,$A3,$11,$F6,$0F,$3E,$DF,$2D,$F3,$4E,$EF,$E5,$47,$EB,$E3,$1B
	.dc.b	$76,$9C,$CA,$72,$FF,$23,$FB,$DF,$E3,$64,$9F,$D0,$3F,$9F,$E0,$FB
	.dc.b	$93,$E2,$D7,$E3,$D3,$DC,$7D,$97,$E7,$9B,$0E,$FF,$C0,$DF,$F1,$B9
	.dc.b	$CC,$62,$79,$57,$1E,$7F,$CA,$DF,$E2,$27,$B4,$62,$3F,$86,$46,$0F
	.dc.b	$CC,$BB,$B4,$EE,$3E,$DE,$BF,$25,$FC,$06,$5E,$C7,$F2,$BF,$97,$7D
	.dc.b	$6C,$BF,$D3,$AD,$8F,$4B,$5D,$EF,$EF,$FB,$EF,$E6,$DF,$9A,$FA,$F3
	.dc.b	$46,$2B,$C8,$D7,$D3,$1A,$D1,$5C,$7E,$FF,$FA,$06,$76,$7F,$85,$9F
	.dc.b	$3E,$33,$7E,$F6,$BD,$3E,$EF,$FA,$47,$E2,$5E,$36,$C7,$C5,$6B,$BA
	.dc.b	$4F,$A7,$E8,$3F,$A8,$7E,$3B,$64,$D2,$57,$CF,$FC,$17,$9B,$F7,$5B
	.dc.b	$3F,$85,$6E,$87,$72,$FA,$67,$D1,$7F,$5F,$FC,$8B,$FD,$9B,$18,$C1
	.dc.b	$18,$9D,$57,$E3,$2B,$53,$F4,$7E,$F7,$FB,$47,$A5,$70,$FF,$09,$DF
	.dc.b	$E1,$67,$2D,$8C,$BF,$53,$EF,$7F,$B8,$7E,$3D,$F6,$DB,$1F,$8D,$6B
	.dc.b	$B9,$4F,$FB,$93,$F5,$F7,$DD,$FF,$71,$A4,$62,$BF,$49,$1F,$A4,$8C
	.dc.b	$74,$F1,$AF,$D5,$78,$FF,$E0,$1F,$CD,$FF,$18,$DB,$FC,$F3,$5F,$C5
	.dc.b	$B5,$FD,$77,$E1,$7F,$E0,$9F,$C2,$7F,$28,$D8,$FD,$33,$5F,$C9,$B5
	.dc.b	$FD,$B7,$C5,$FF,$84,$78,$7F,$38,$D8,$FD,$73,$5F,$CD,$B5,$FD,$7F
	.dc.b	$85,$FC,$B1,$FD,$79,$FD,$CB,$BF,$F9,$1F,$8B,$FA,$EF,$03,$F9,$E4
	.dc.b	$A7,$2F,$EC,$23,$FA,$07,$07,$F0,$FF,$C1,$BF,$C4,$9E,$7F,$EB,$DB
	.dc.b	$5F,$C0,$5A,$FE,$B9,$AF,$F1,$1F,$6F,$FC,$59,$F8,$97,$ED,$DB,$1F
	.dc.b	$C2,$5A,$FE,$D9,$AF,$F1,$9F,$D4,$3E,$76,$1F,$C8,$5B,$DA,$69,$F4
	.dc.b	$E7,$F0,$3C,$9E,$EE,$6F,$C1,$6D,$F7,$F0,$FE,$6B,$5A,$7F,$9B,$9F
	.dc.b	$CB,$7E,$53,$FC,$85,$F6,$1F,$C3,$5B,$3F,$C7,$DA,$FF,$0C,$6B,$F4
	.dc.b	$BC,$5F,$E3,$CF,$F5,$DF,$29,$FA,$57,$F9,$3F,$07,$E7,$9F,$CD,$A7
	.dc.b	$3E,$F7,$F9,$46,$31,$5B,$6C,$66,$FF,$98,$7A,$3F,$E5,$7F,$CD,$BF
	.dc.b	$91,$36,$33,$8A,$72,$D8,$FD,$16,$FF,$E5,$FE,$BF,$88,$48,$CC,$A3
	.dc.b	$F9,$77,$E1,$BF,$E6,$5F,$B5,$B2,$69,$2B,$FA,$27,$C6,$FF,$9A,$FF
	.dc.b	$3B,$FE,$6C,$D8,$CD,$27,$2D,$8F,$EC,$1C,$1F,$29,$FC,$D9,$FE,$1C
	.dc.b	$E0,$7E,$D4,$F8,$DF,$F3,$84,$61,$B1,$9B,$FE,$03,$5F,$E6,$5A,$D4
	.dc.b	$62,$3F,$B3,$7C,$97,$FA,$07,$F3,$4F,$E9,$8D,$8F,$EB,$CD,$7F,$A5
	.dc.b	$B5,$FE,$DD,$F2,$9F,$2D,$FE,$C8,$D7,$FA,$8B,$5D,$D7,$CD,$7D,$E7
	.dc.b	$F8,$EB,$C2,$FE,$FF,$E6,$FF,$70,$F2,$7F,$BC,$36,$7F,$BB,$35,$FE
	.dc.b	$EA,$D7,$F2,$47,$3D,$7F,$C0,$6F,$F8,$23,$F7,$E7,$99,$F2,$1B,$8F
	.dc.b	$C3,$D5,$18,$C7,$F2,$CD,$BF,$0E,$1F,$93,$7C,$2D,$7F,$A6,$36,$FF
	.dc.b	$A5,$B7,$1F,$F7,$3C,$FA,$BE,$3B,$FD,$81,$18,$8F,$5C,$7F,$88,$DF
	.dc.b	$51,$FE,$8E,$AF,$EA,$BF,$39,$FE,$7E,$A3,$49,$1F,$74,$FE,$B7,$F3
	.dc.b	$BB,$23,$99,$47,$F4,$9F,$6F,$FC,$30,$4D,$BF,$96,$3F,$E1,$BE,$57
	.dc.b	$DA,$7F,$4F,$6E,$1F,$8B,$C9,$18,$DB,$E8,$B6,$3E,$CF,$2F,$FB,$36
	.dc.b	$A9,$FE,$39,$2D,$7F,$B4,$E7,$EE,$CE,$DF,$F2,$B1,$0F,$BE,$7E,$28
	.dc.b	$87,$FB,$0F,$F1,$A7,$E3,$DF,$E6,$8B,$FF,$1D,$CE,$63,$0F,$71,$3C
	.dc.b	$3B,$BC,$55,$FE,$8C,$8C,$3C,$A7,$FC,$15,$F3,$BE,$56,$C7,$AD,$FC
	.dc.b	$AC,$66,$3F,$C1,$AE,$E3,$F7,$A7,$AC,$A7,$FA,$16,$7B,$BE,$56,$CF
	.dc.b	$8E,$FD,$2D,$BF,$3D,$73,$D3,$ED,$71,$7F,$0B,$6C,$7E,$30,$FF,$33
	.dc.b	$FC,$AF,$FA,$BF,$E5,$7F,$DD,$5F,$32,$F9,$7F,$F5,$5C,$E6,$93,$97
	.dc.b	$FD,$E1,$19,$8C,$42,$31,$3C,$BD,$65,$F6,$8C,$36,$B9,$CE,$9E,$F3
	.dc.b	$A6,$9F,$EF,$FC,$1C,$6C,$BC,$33,$98,$C5,$7F,$02,$AD,$D6,$AB,$FC
	.dc.b	$EB,$39,$AD,$C7,$FA,$2A,$B7,$39,$AE,$11,$89,$FA,$1A,$7C,$72,$D7
	.dc.b	$D7,$D7,$F2,$0D,$D5,$3B,$AF,$F7,$DC,$F5,$23,$E1,$C7,$C6,$FE,$2A
	.dc.b	$5F,$39,$F9,$77,$F1,$66,$0F,$E4,$BC,$1D,$A3,$5C,$C8,$C4,$F9,$89
	.dc.b	$5F,$A3,$C7,$CB,$8F,$9E,$7C,$D3,$F2,$F4,$F8,$39,$79,$23,$15,$E8
	.dc.b	$7D,$67,$A6,$76,$48,$FE,$77,$C5,$E2,$9F,$DE,$DF,$A5,$B9,$FA,$27
	.dc.b	$50,$3F,$86,$F2,$FB,$5F,$F5,$FF,$2F,$F1,$2A,$FE,$1F,$39,$9E,$07
	.dc.b	$0B,$F9,$2F,$97,$0F,$43,$EA,$30,$7F,$4E,$E6,$E4,$FF,$A1,$DB,$3F
	.dc.b	$CB,$0E,$71,$FD,$4A,$33,$B7,$F3,$49,$CC,$FF,$A3,$F9,$F2,$48,$CC
	.dc.b	$7E,$48,$F0,$9F,$28,$87,$FA,$5F,$A3,$EA,$BF,$D2,$63,$F3,$E7,$FA
	.dc.b	$62,$33,$1F,$D2,$A3,$35,$FA,$55,$FC,$03,$FD,$B3,$6F,$C9,$B6,$33
	.dc.b	$5F,$D7,$9A,$FF,$82,$7A,$BF,$B3,$EB,$FD,$93,$B3,$E2,$F6,$7F,$68
	.dc.b	$6C,$EA,$BD,$67,$CD,$3D,$67,$BC,$FE,$F2,$7C,$82,$5F,$C8,$3F,$D6
	.dc.b	$6F,$D2,$38,$1C,$C8,$61,$79,$EE,$EF,$D7,$FD,$FF,$FB,$9D,$B5,$FE
	.dc.b	$16,$A4,$E5,$B1,$FE,$20,$5F,$BA,$7F,$89,$1B,$19,$8C,$1F,$E5,$BF
	.dc.b	$0F,$FC,$61,$A5,$E6,$F8,$B1,$8A,$FF,$1D,$35,$F8,$7D,$7F,$C2,$DB
	.dc.b	$D1,$FA,$DC,$7D,$7D,$1F,$C2,$3F,$80,$3F,$C8,$E4,$AF,$F2,$5B,$FE
	.dc.b	$54,$FB,$B9,$1F,$A0,$AF,$87,$E4,$7F,$10,$9C,$CF,$E9,$3D,$2F,$F6
	.dc.b	$36,$FF,$E8,$52,$47,$F9,$89,$FF,$34,$FE,$F7,$23,$FC,$DA,$D7,$FC
	.dc.b	$99,$E8,$7F,$9F,$79,$FF,$CE,$F9,$7E,$22,$5C,$FF,$E7,$E8,$C3,$67
	.dc.b	$F6,$33,$F7,$09,$BF,$C3,$7F,$93,$F2,$BD,$1F,$E5,$3F,$7D,$FE,$95
	.dc.b	$C4,$D2,$57,$F9,$EB,$EF,$3F,$D4,$04,$70,$7C,$5B,$FF,$55,$57,$F9
	.dc.b	$02,$31,$EB,$7F,$36,$EA,$FE,$13,$1F,$EB,$39,$D3,$FC,$4D,$FF,$5A
	.dc.b	$18,$7F,$D8,$0B,$F4,$AF,$F6,$03,$63,$FD,$80,$D7,$FD,$80,$D7,$DF
	.dc.b	$ED,$7F,$46,$FD,$B7,$FD,$A1,$26,$92,$7F,$DB,$10,$CB,$FD,$BC,$D8
	.dc.b	$CC,$60,$F9,$FA,$FC,$D8,$F9,$FD,$1F,$8F,$D1,$EF,$FF,$56,$FA,$8F
	.dc.b	$F7,$9E,$26,$92,$FF,$DF,$B0,$FD,$5B,$FE,$00,$B6,$33,$18,$3E,$5C
	.dc.b	$7E,$DB,$CC,$F9,$D3,$F4,$DA,$FC,$DC,$11,$88,$FC,$1B,$5F,$F6,$D7
	.dc.b	$7F,$FB,$74,$8F,$FA,$E9,$EF,$3F,$E7,$EF,$D8,$6F,$87,$FD,$C9,$B8
	.dc.b	$1A,$70,$F8,$7F,$EE,$FC,$D6,$7F,$25,$8F,$E2,$67,$2D,$8C,$BF,$D4
	.dc.b	$3E,$11,$F1,$3F,$E0,$0F,$DB,$BF,$D2,$71,$8A,$46,$27,$55,$F9,$0A
	.dc.b	$D4,$FA,$BE,$2F,$FC,$13,$FA,$C5,$C0,$D2,$5F,$E7,$7E,$3B,$FC,$1F
	.dc.b	$E1,$7C,$26,$C6,$69,$39,$6C,$7E,$97,$E4,$3B,$B6,$3B,$B6,$FE,$6C
	.dc.b	$EA,$3E,$C7,$C8,$FF,$87,$B4,$FA,$2D,$8F,$D8,$35,$DD,$FC,$67,$F6
	.dc.b	$9F,$27,$FE,$26,$FE,$00,$FB,$2D,$8C,$CA,$72,$D8,$FD,$1F,$27,$E8
	.dc.b	$88,$E0,$FE,$A6,$7E,$83,$CF,$FF,$0A,$3F,$C0,$FE,$57,$FC,$6B,$F8
	.dc.b	$0F,$E2,$1B,$19,$B4,$E5,$B1,$FC,$2F,$C6,$FF,$02,$F4,$7F,$20,$9F
	.dc.b	$EB,$4D,$C7,$F2,$73,$FD,$6F,$BB,$F6,$44,$EE,$FF,$16,$61,$FD,$6F
	.dc.b	$E3,$5F,$2F,$FD,$F2,$3F,$C9,$31,$8C,$BF,$C0,$2D,$F6,$3F,$5A,$D9
	.dc.b	$FF,$10,$FC,$C3,$DB,$7C,$87,$F9,$4B,$C5,$FE,$3B,$39,$8C,$57,$0A
	.dc.b	$DE,$3F,$C2,$E3,$FC,$65,$5B,$8C,$47,$F7,$2D,$3D,$5C,$5F,$32,$37
	.dc.b	$7C,$BE,$AF,$BB,$63,$B6,$3F,$E6,$49,$CB,$63,$F9,$8F,$CE,$3E,$73
	.dc.b	$E1,$42,$31,$97,$F9,$ED,$BB,$7F,$C0,$06,$CF,$E0,$33,$F2,$CF,$F5
	.dc.b	$09,$CC,$FF,$91,$9A,$FF,$93,$DA,$F7,$BF,$0F,$CF,$3E,$7B,$E3,$CF
	.dc.b	$F9,$3A,$31,$8F,$C4,$84,$63,$7F,$E9,$71,$8A,$E2,$FC,$9E,$2F,$C0
	.dc.b	$E7,$E6,$8F,$E4,$FE,$3F,$C0,$C7,$F9,$19,$FE,$99,$2E,$9D,$1B,$3F
	.dc.b	$E0,$1E,$27,$CF,$FE,$6A,$3F,$AC,$46,$27,$F3,$31,$89,$D3,$F8,$E6
	.dc.b	$BF,$A7,$AF,$E9,$38,$FF,$85,$6D,$F4,$1F,$E9,$CF,$A5,$FC,$8B,$6F
	.dc.b	$C4,$D7,$C2,$D7,$F3,$7A,$7F,$9B,$2B,$E4,$F4,$7E,$6A,$30,$7E,$9F
	.dc.b	$0F,$8A,$4E,$5F,$C7,$5F,$E9,$3E,$8B,$F4,$DA,$B1,$46,$1B,$1F,$A8
	.dc.b	$8F,$C5,$EF,$F3,$EB,$FC,$63,$7D,$1F,$80,$7F,$08,$FA,$57,$D2,$7F
	.dc.b	$AE,$B5,$FB,$71,$8C,$7E,$1D,$7F,$82,$23,$13,$AA,$E1,$3F,$E0,$CF
	.dc.b	$A6,$6F,$F8,$28,$46,$30,$FC,$53,$78,$FF,$12,$D9,$FC,$E6,$DF,$0F
	.dc.b	$87,$FD,$37,$7F,$07,$E9,$FE,$61,$1F,$F3,$67,$02,$3E,$87,$CC,$7F
	.dc.b	$B5,$3F,$4A,$FE,$12,$DC,$7F,$14,$D7,$F1,$2D,$7F,$89,$FD,$4F,$CD
	.dc.b	$26,$1F,$C5,$1F,$A1,$F5,$7F,$EE,$0F,$D9,$3F,$8A,$B6,$BF,$2C,$D7
	.dc.b	$F2,$AD,$7F,$DA,$1D,$DF,$EF,$12,$3F,$EC,$D7,$F5,$7F,$40,$FA,$6F
	.dc.b	$F7,$6F,$D6,$E1,$B5,$FE,$3A,$9D,$64,$AD,$36,$75,$E0,$F0,$7F,$D9
	.dc.b	$35,$FC,$8A,$BB,$CE,$67,$FD,$A9,$39,$7F,$DC,$13,$FB,$EE,$97,$13
	.dc.b	$A7,$B5,$7F,$B5,$27,$31,$88,$FF,$77,$57,$F0,$65,$F1,$DF,$C8,$1B
	.dc.b	$93,$93,$53,$92,$46,$67,$F8,$7C,$62,$77,$39,$9F,$F7,$2F,$CA,$FF
	.dc.b	$8B,$BE,$57,$C0,$D8,$FE,$2E,$D7,$FC,$C6,$D7,$FC,$CB,$F2,$2A,$F4
	.dc.b	$47,$FB,$84,$F5,$46,$5B,$1A,$8C,$47,$F2,$46,$BD,$23,$44,$78,$90
	.dc.b	$E0,$FC,$7C,$FE,$5D,$FD,$1D,$3E,$83,$74,$34,$FC,$4F,$E1,$1E,$96
	.dc.b	$DE,$B0,$FC,$44,$E5,$B1,$97,$F0,$BF,$2F,$96,$C6,$5F,$C2,$C6,$5B
	.dc.b	$19,$1F,$27,$FE,$D8,$5F,$9A,$FF,$99,$A3,$0D,$F3,$BE,$73,$67,$75
	.dc.b	$FE,$BF,$A7,$D7,$7C,$A8,$FA,$91,$89,$FE,$8C,$D8,$ED,$3B,$9F,$1F
	.dc.b	$95,$F6,$FE,$93,$C4,$DA,$EC,$D7,$FD,$45,$7F,$DC,$A7,$D1,$F4,$5C
	.dc.b	$E0,$D2,$5F,$D7,$DF,$F4,$97,$F5,$48,$7F,$AC,$38,$3F,$4A,$FF,$20
	.dc.b	$FB,$4F,$63,$FC,$99,$BE,$11,$A6,$BE,$D3,$18,$78,$AB,$FD,$77,$18
	.dc.b	$C7,$FB,$9B,$63,$35,$F9,$B9,$F1,$C7,$ED,$F1,$F3,$BE,$56,$B9,$E3
	.dc.b	$FD,$C5,$F8,$30,$EB,$87,$E9,$5A,$7D,$58,$D5,$7D,$9F,$C6,$7E,$63
	.dc.b	$F6,$1B,$3F,$86,$6E,$87,$72,$D3,$E3,$BE,$5C,$7E,$71,$1F,$EA,$FF
	.dc.b	$82,$7E,$DF,$DF,$B8,$9A,$49,$FC,$67,$91,$FB,$DF,$CA,$7F,$28,$D8
	.dc.b	$CD,$27,$2D,$8F,$11,$F2,$81,$E7,$DB,$E4,$B8,$CB,$D2,$E8,$6B,$F1
	.dc.b	$9C,$35,$E6,$D7,$7B,$7C,$F7,$FD,$F3,$F8,$E3,$AB,$F8,$B7,$D1,$FF
	.dc.b	$8C,$A3,$19,$23,$13,$AA,$E2,$D7,$EF,$1E,$B3,$D5,$BF,$CD,$7B,$E1
	.dc.b	$F9,$6C,$3F,$23,$1D,$E3,$07,$E6,$BD,$5F,$F2,$CC,$7F,$95,$F9,$B0
	.dc.b	$4C,$BA,$47,$80,$F9,$64,$3F,$CB,$11,$B7,$94,$7F,$49,$9E,$B1,$FE
	.dc.b	$6A,$9D,$36,$75,$1F,$AC,$D3,$E8,$CF,$D3,$D7,$EA,$BE,$E3,$E0,$5F
	.dc.b	$47,$C3,$7C,$88,$FF,$89,$3E,$73,$F6,$FE,$81,$A4,$9F,$6C,$F7,$D2
	.dc.b	$BD,$CF,$F1,$54,$ED,$E9,$3F,$D9,$6B,$A3,$FE,$6C,$FA,$37,$E2,$DF
	.dc.b	$E1,$D8,$FF,$15,$D7,$E2,$E7,$3E,$77,$F5,$7E,$AF,$C7,$4A,$73,$5C
	.dc.b	$23,$F2,$F1,$89,$FC,$06,$BF,$1A,$7C,$C7,$D5,$3F,$E0,$07,$C1,$7F
	.dc.b	$16,$F8,$8F,$F7,$F4,$7F,$BF,$70,$FE,$F9,$18,$8F,$E7,$1F,$8C,$FF
	.dc.b	$69,$FC,$E3,$FA,$03,$63,$34,$9C,$B6,$3F,$A0,$7E,$33,$FD,$C7,$EB
	.dc.b	$3F,$A4,$36,$33,$29,$CB,$63,$FA,$47,$2F,$F7,$9F,$CE,$7F,$A8,$36
	.dc.b	$33,$29,$CB,$63,$FA,$87,$E0,$FF,$E0,$1F,$B9,$FE,$B0,$D8,$CC,$A7
	.dc.b	$2D,$8F,$EB,$1F,$8C,$FF,$82,$7F,$46,$FE,$C0,$D8,$CC,$A7,$2D,$8F
	.dc.b	$B3,$B3,$7F,$EB,$F1,$86,$BF,$52,$7F,$B7,$B9,$6B,$FD,$6B,$F0,$CF
	.dc.b	$F0,$D7,$E3,$1F,$DC,$5B,$1F,$DB,$5A,$EE,$53,$FE,$BC,$F9,$17,$D2
	.dc.b	$7F,$88,$F9,$B0,$DB,$FC,$7C,$EA,$FF,$21,$18,$9D,$3E,$73,$F1,$04
	.dc.b	$3D,$07,$F8,$17,$C6,$FF,$18,$78,$76,$BF,$93,$CF,$F8,$0E,$72,$D8
	.dc.b	$CB,$FD,$E3,$E0,$3F,$C7,$1F,$8D,$FF,$83,$DB,$1F,$E0,$B6,$BB,$94
	.dc.b	$FF,$85,$70,$FE,$85,$3F,$E0,$26,$BF,$E1,$06,$CF,$F8,$79,$CC,$7E
	.dc.b	$A0,$FE,$00,$7E,$B3,$C1,$FE,$1F,$BF,$E9,$63,$E1,$1F,$00,$87,$F8
	.dc.b	$43,$97,$FC,$A1,$F9,$4F,$E5,$09,$1F,$98,$7F,$7B,$83,$F3,$BF,$EE
	.dc.b	$5F,$77,$86,$D7,$F4,$F9,$D1,$FE,$5B,$9D,$36,$35,$2F,$C5,$3F,$CC
	.dc.b	$1F,$91,$73,$6C,$E9,$AF,$F6,$0A,$F1,$46,$23,$FB,$1E,$5E,$F6,$F2
	.dc.b	$FF,$09,$25,$7F,$BF,$70,$7A,$BF,$E6,$FF,$1F,$D4,$DB,$F2,$B5,$F0
	.dc.b	$4F,$0A,$57,$CF,$F1,$FF,$97,$F8,$5F,$C6,$FE,$BD,$F3,$9F,$E5,$7C
	.dc.b	$51,$8C,$3F,$CB,$91,$89,$D4,$E7,$D9,$F9,$9C,$7F,$E8,$3F,$C7,$6E
	.dc.b	$06,$93,$0F,$E6,$3F,$18,$F9,$1F,$03,$6B,$BB,$70,$FE,$6F,$48,$C3
	.dc.b	$F5,$3E,$0F,$FD,$25,$F8,$E7,$DD,$6C,$66,$D3,$96,$C7,$F4,$2C,$FD
	.dc.b	$DE,$5F,$66,$C6,$75,$FE,$90,$DA,$FE,$8C,$FE,$68,$FE,$E8,$53,$FA
	.dc.b	$21,$FE,$A5,$E1,$7C,$27,$F9,$BD,$B3,$F7,$63,$06,$7E,$D7,$FD,$59
	.dc.b	$F6,$DF,$97,$6D,$FE,$49,$AF,$DE,$6B,$DF,$F1,$AE,$C4,$BF,$C5,$3F
	.dc.b	$DF,$F8,$BF,$BE,$F1,$FF,$5B,$23,$FE,$34,$9F,$D4,$7C,$17,$9C,$8F
	.dc.b	$E4,$DE,$5C,$BF,$8A,$9E,$67,$F8,$F3,$7F,$F7,$77,$7F,$FA,$4D,$B7
	.dc.b	$F6,$9B,$97,$F7,$26,$BF,$AE,$FC,$6F,$F9,$7B,$F2,$5B,$3F,$BB,$6E
	.dc.b	$E6,$FC,$2F,$CD,$24,$FE,$A5,$F3,$D7,$E8,$4F,$F2,$A4,$7F,$05,$EB
	.dc.b	$FA,$84,$7F,$60,$FD,$7F,$8A,$FF,$6F,$76,$FF,$09,$6D,$7F,$04,$6B
	.dc.b	$FB,$06,$BF,$83,$21,$FC,$84,$FE,$45,$F3,$2F,$A4,$7D,$D7,$FB,$9F
	.dc.b	$1F,$F3,$14,$E6,$31,$5F,$E6,$A8,$C4,$FF,$0D,$8C,$57,$28,$C4,$7E
	.dc.b	$47,$E3,$FF,$D7,$91,$FC,$69,$B1,$FC,$19,$AE,$E1,$F0,$DF,$EF,$4E
	.dc.b	$4F,$C2,$F1,$FD,$F3,$F6,$72,$F9,$2F,$A4,$FF,$A2,$CF,$E5,$8E,$3E
	.dc.b	$13,$FD,$FF,$1B,$E1,$FF,$49,$3A,$6E,$3F,$E3,$0F,$BA,$FE,$04,$A7
	.dc.b	$FC,$75,$18,$6F,$85,$F1,$27,$FC,$75,$39,$7F,$90,$FC,$97,$FA,$02
	.dc.b	$BF,$9C,$B6,$33,$1F,$E8,$13,$25,$7F,$72,$7D,$7C,$FE,$0B,$FE,$48
	.dc.b	$F4,$57,$DE,$3C,$38,$78,$E7,$2D,$9D,$4F,$3E,$9F,$D3,$EB,$F2,$E3
	.dc.b	$F2,$87,$F0,$A1,$F1,$AF,$99,$78,$DF,$EA,$AF,$25,$D5,$FE,$A7,$AD
	.dc.b	$4E,$63,$15,$C2,$FF,$A4,$C6,$27,$51,$FE,$B4,$AE,$2E,$8C,$57,$1B
	.dc.b	$EF,$18,$AE,$2F,$9A,$75,$7F,$EE,$33,$F9,$BE,$AC,$99,$72,$BD,$1F
	.dc.b	$ED,$7A,$DC,$E9,$B3,$B9,$FF,$43,$46,$0F,$A5,$9F,$F4,$77,$F0,$D1
	.dc.b	$F8,$9D,$3F,$11,$18,$8F,$AB,$B7,$C8,$3F,$12,$57,$C6,$43,$B9,$E5
	.dc.b	$3F,$0B,$F3,$1F,$9A,$EC,$70,$FE,$56,$30,$DA,$D9,$DA,$FC,$5C,$3F
	.dc.b	$84,$8F,$06,$1E,$49,$CE,$5E,$B9,$FD,$20,$F6,$BF,$53,$CB,$FA,$E6
	.dc.b	$F2,$7E,$95,$B3,$B3,$D1,$87,$AC,$FF,$6B,$E1,$F2,$5F,$E0,$50,$FB
	.dc.b	$0F,$67,$47,$EB,$DB,$19,$36,$FC,$BE,$AF,$E2,$37,$7A,$BF,$88,$37
	.dc.b	$AF,$E8,$DF,$F1,$A9,$CB,$E4,$C7,$E6,$9E,$5E,$5F,$C3,$72,$7D,$F9
	.dc.b	$FF,$1C,$76,$7F,$01,$FA,$AF,$E6,$10,$69,$2F,$FC,$6B,$DF,$F8,$06
	.dc.b	$CF,$3E,$6F,$F4,$33,$67,$53,$E1,$E8,$FC,$83,$E5,$3E,$39,$FE,$40
	.dc.b	$D7,$6F,$F7,$B6,$D7,$F9,$35,$B7,$F8,$D7,$FC,$9D,$C5,$FC,$67,$F1
	.dc.b	$7B,$46,$92,$C4,$F0,$7F,$B3,$DF,$F6,$5F,$94,$FF,$44,$37,$C3,$F1
	.dc.b	$CE,$AB,$D1,$C1,$F2,$AB,$EF,$47,$E6,$F1,$FE,$F6,$7C,$3F,$13,$CE
	.dc.b	$47,$F5,$D1,$F0,$0F,$89,$C3,$8B,$E9,$1F,$E6,$2F,$17,$DE,$4A,$F9
	.dc.b	$CF,$C2,$F1,$BF,$9C,$7C,$DF,$F8,$66,$4D,$25,$7F,$9B,$3C,$9F,$92
	.dc.b	$49,$FB,$0F,$D0,$8F,$21,$F8,$82,$18,$3F,$CA,$90,$F9,$9F,$F1,$D7
	.dc.b	$97,$FD,$25,$B1,$99,$FF,$1D,$47,$F9,$0F,$F1,$99,$1F,$E4,$76,$BF
	.dc.b	$50,$FF,$28,$78,$9F,$D4,$F8,$BF,$CA,$7E,$87,$F9,$5D,$B1,$99,$FF
	.dc.b	$2E,$B5,$FF,$39,$F3,$FF,$5B,$FC,$63,$FC,$CA,$48,$FF,$46,$C6,$5F
	.dc.b	$F2,$75,$7A,$08,$E0,$FB,$F1,$F9,$87,$EB,$F1,$FF,$A1,$33,$7D,$15
	.dc.b	$24,$FF,$A1,$A3,$2F,$FB,$8E,$7F,$CD,$BE,$77,$E9,$9B,$D7,$F4,$63
	.dc.b	$13,$FE,$9D,$A7,$E2,$1F,$EA,$16,$C6,$63,$07,$FA,$AA,$1F,$BD,$D2
	.dc.b	$4F,$FA,$B2,$3F,$BF,$47,$FA,$E3,$E5,$9E,$E7,$F8,$07,$F5,$7C,$36
	.dc.b	$FF,$D7,$13,$AB,$56,$9B,$3E,$D2,$1F,$6F,$1F,$5B,$72,$FC,$77,$4F
	.dc.b	$E7,$A3,$E5,$9F,$C5,$EB,$FD,$B4,$3F,$4F,$FF,$71,$36,$3C,$7D,$DF
	.dc.b	$E1,$79,$C9,$FE,$ED,$97,$E3,$54,$95,$FE,$EF,$8C,$BF,$EF,$7F,$D5
	.dc.b	$3F,$DF,$84,$70,$7A,$8F,$F6,$3C,$7E,$E3,$2F,$5C,$63,$83,$FD,$4F
	.dc.b	$F0,$AC,$BF,$C5,$5F,$A8,$61,$B5,$F9,$D1,$3B,$C1,$5B,$6C,$FD,$0F
	.dc.b	$88,$FE,$38,$D8,$FE,$38,$DF,$13,$ED,$4E,$A3,$E7,$4B,$EC,$7F,$1A
	.dc.b	$D8,$F1,$67,$FC,$DA,$72,$7E,$0F,$E2,$FF,$C8,$5F,$BE,$5B,$34,$97
	.dc.b	$F8,$4F,$47,$FC,$91,$F5,$BF,$8B,$6C,$66,$93,$96,$C7,$E3,$39,$FF
	.dc.b	$CA,$1F,$B7,$FE,$2C,$91,$99,$47,$E5,$FE,$4B,$FC,$AD,$C3,$67,$E8
	.dc.b	$57,$E5,$A7,$2D,$8C,$BF,$D0,$17,$4B,$E5,$3F,$CC,$1F,$B7,$FE,$8D
	.dc.b	$B5,$FA,$06,$BC,$67,$85,$2B,$E6,$F2,$CA,$FC,$40,$E1,$FC,$3B,$FC
	.dc.b	$A7,$A7,$F4,$0F,$F2,$C6,$FB,$2F,$97,$7C,$BF,$F2,$BA,$FE,$49,$39
	.dc.b	$8C,$47,$F5,$A8,$C6,$5F,$CB,$E3,$15,$CA,$31,$1F,$D8,$3E,$61,$F3
	.dc.b	$3F,$CF,$DB,$3F,$D5,$1B,$9F,$E5,$E9,$18,$7F,$91,$FA,$1F,$D5,$FC
	.dc.b	$5F,$E4,$91,$FD,$1A,$74,$FF,$3B,$6B,$E1,$C3,$C0,$7F,$4F,$F8,$B7
	.dc.b	$CD,$52,$65,$FD,$9A,$32,$F7,$C6,$65,$FD,$C3,$0F,$EA,$B5,$A8,$F4
	.dc.b	$36,$3F,$60,$D7,$FA,$73,$5F,$F0,$1B,$53,$49,$7F,$C6,$32,$F7,$CF
	.dc.b	$DB,$6E,$DF,$72,$38,$CF,$AA,$FC,$6D,$8F,$E0,$5F,$37,$FC,$AE,$7E
	.dc.b	$73,$63,$3C,$FF,$DC,$E3,$11,$FC,$98,$87,$F8,$69,$7C,$E7,$F9,$96
	.dc.b	$3A,$B5,$FE,$D9,$1A,$D1,$F0,$7E,$0A,$D5,$7F,$89,$3E,$5B,$F9,$F7
	.dc.b	$DC,$7E,$05,$B1,$FC,$65,$AF,$F7,$76,$BF,$9B,$D7,$E1,$93,$7F,$F2
	.dc.b	$77,$3B,$E8,$3F,$9C,$57,$F4,$C6,$BF,$AD,$6B,$FE,$D3,$8E,$19,$FF
	.dc.b	$41,$8F,$C6,$7C,$C7,$F9,$CD,$FE,$59,$C8,$8C,$46,$A7,$FC,$F8,$EA
	.dc.b	$32,$F9,$F0,$FC,$5B,$FD,$46,$DF,$43,$FE,$75,$8F,$EC,$2D,$7F,$81
	.dc.b	$35,$F3,$35,$67,$F1,$1B,$C1,$FE,$ED,$D7,$E6,$36,$3E,$63,$5F,$BA
	.dc.b	$7E,$37,$0F,$F2,$3D,$7F,$71,$FA,$FE,$39,$9A,$4A,$FE,$9D,$EE,$FE
	.dc.b	$21,$FF,$2F,$61,$FE,$3D,$8C,$41,$BC,$BF,$99,$27,$07,$D9,$E4,$FF
	.dc.b	$03,$7D,$1F,$F9,$CB,$73,$49,$9F,$DE,$E6,$F8,$8F,$F2,$E6,$CF,$EE
	.dc.b	$BE,$1B,$FC,$21,$18,$FA,$4E,$84,$AF,$C3,$BD,$FE,$99,$0F,$DA,$3F
	.dc.b	$C3,$11,$86,$C6,$76,$FF,$0B,$CE,$5F,$EE,$1F,$52,$F9,$1F,$F2,$34
	.dc.b	$7F,$A0,$A3,$14,$F6,$7F,$90,$46,$1B,$5C,$29,$2F,$84,$C3,$6F,$94
	.dc.b	$F8,$27,$27,$1E,$1F,$F5,$8E,$1F,$EF,$02,$4F,$F8,$16,$32,$FF,$31
	.dc.b	$EC,$FF,$3A,$3F,$B9,$E1,$FE,$68,$D3,$58,$7F,$A3,$1C,$1F,$E0,$F9
	.dc.b	$E2,$48,$D3,$67,$9B,$67,$FC,$2D,$F2,$4F,$91,$7C,$97,$82,$30,$D8
	.dc.b	$CC,$FF,$47,$9C,$BE,$6F,$C2,$3E,$CB,$71,$F4,$A8,$CB,$FD,$3D,$AF
	.dc.b	$F8,$9D,$AF,$FA,$15,$AF,$FA,$E9,$AF,$38,$EB,$18,$23,$82,$47,$8C
	.dc.b	$87,$85,$F9,$B9,$7F,$B9,$B0,$FE,$C4,$D9,$FA,$DC,$CE,$BF,$83,$1F
	.dc.b	$E9,$99,$D3,$FE,$9B,$DF,$F0,$97,$F8,$CD,$FF,$36,$DC,$78,$D7,$E4
	.dc.b	$E7,$2F,$FA,$8E,$1F,$15,$FD,$EF,$2F,$F2,$54,$63,$C8,$FD,$2C,$62
	.dc.b	$75,$5C,$5A,$FF,$88,$5F,$F6,$2D,$7F,$7D,$3A,$FC,$3F,$CE,$69,$98
	.dc.b	$DC,$F5,$F8,$3F,$D9,$FC,$A5,$A3,$49,$3E,$1F,$AE,$FD,$CF,$D4,$79
	.dc.b	$5B,$19,$A4,$E5,$B1,$F3,$B8,$3E,$9E,$7F,$3E,$3E,$9B,$5D,$18,$8D
	.dc.b	$91,$C1,$FC,$93,$EA,$3F,$B1,$7C,$75,$E6,$FE,$1A,$DC,$0D,$38,$D7
	.dc.b	$62,$1F,$7F,$0F,$13,$FE,$D9,$FC,$E5,$F8,$4F,$F1,$8F,$95,$FF,$15
	.dc.b	$46,$30,$46,$27,$55,$FE,$85,$AD,$4F,$DB,$FB,$1F,$9F,$3F,$6D,$AF
	.dc.b	$DB,$6B,$FC,$9B,$85,$18,$87,$3B,$2F,$3C,$0C,$B8,$6E,$DF,$25,$FC
	.dc.b	$57,$8F,$FC,$CB,$E6,$FF,$1C,$D8,$CE,$A9,$CB,$63,$F3,$5F,$A2,$FC
	.dc.b	$08,$FD,$E8,$70,$FE,$71,$B1,$C3,$4F,$E8,$33,$F4,$A3,$E8,$61,$E5
	.dc.b	$69,$C7,$ED,$7F,$A3,$FD,$75,$A3,$49,$3F,$68,$F8,$5A,$7C,$5B,$E9
	.dc.b	$18,$D3,$E2,$4E,$63,$C4,$66,$76,$FC,$83,$25,$3B,$5F,$BF,$ED,$BF
	.dc.b	$5D,$99,$A4,$8E,$77,$DE,$BC,$C7,$33,$FD,$35,$5C,$67,$F7,$4D,$8F
	.dc.b	$95,$5C,$DF,$99,$F6,$FE,$02,$57,$07,$F9,$5D,$F7,$9F,$96,$78,$1C
	.dc.b	$3F,$DC,$AD,$F7,$1F,$BA,$8F,$1B,$5D,$36,$FE,$F3,$67,$F8,$3C,$7A
	.dc.b	$3A,$7E,$19,$1F,$83,$9F,$D0,$26,$7E,$EF,$BA,$E4,$49,$FE,$0F,$8F
	.dc.b	$C4,$8E,$DF,$2F,$FE,$08,$F7,$3F,$18,$D9,$FC,$9B,$5F,$96,$D7,$EE
	.dc.b	$1F,$6B,$4F,$82,$FD,$13,$27,$C5,$C7,$CA,$FE,$67,$EF,$3F,$C2,$FF
	.dc.b	$33,$72,$34,$93,$FA,$4F,$1F,$FC,$3B,$FA,$57,$E8,$5B,$19,$A4,$E5
	.dc.b	$B1,$FA,$CF,$B4,$FF,$12,$FE,$35,$FA,$96,$C6,$65,$39,$6C,$7E,$D3
	.dc.b	$D8,$FF,$16,$FE,$8D,$FB,$16,$C6,$65,$39,$6C,$7E,$F3,$F0,$2F,$3B
	.dc.b	$F9,$46,$BF,$87,$6C,$7E,$FD,$CC,$7C,$B2,$6D,$F3,$8F,$C7,$CF,$F0
	.dc.b	$5F,$2F,$FC,$83,$FA,$07,$F0,$56,$D7,$F0,$C6,$BF,$A4,$6B,$F9,$7F
	.dc.b	$C1,$BF,$C9,$3F,$69,$FC,$35,$B1,$FC,$51,$AF,$EB,$1A,$FD,$E3,$F4
	.dc.b	$64,$30,$7E,$8F,$F0,$2F,$F2,$AF,$D6,$FF,$19,$6C,$7F,$20,$6B,$FB
	.dc.b	$86,$BF,$AB,$FA,$1F,$F2,$EF,$EA,$5F,$C8,$5B,$1F,$CA,$1A,$FF,$00
	.dc.b	$6B,$FB,$3F,$C3,$3F,$CC,$BF,$6B,$FC,$A5,$B1,$FC,$C1,$AF,$F0,$86
	.dc.b	$BE,$A3,$E8,$F4,$7E,$69,$FC,$E7,$67,$EC,$8F,$E1,$BF,$87,$7F,$9C
	.dc.b	$BE,$37,$F9,$B3,$67,$F9,$EB,$5F,$E2,$AD,$7F,$84,$7D,$6F,$F9,$EB
	.dc.b	$F3,$0F,$E7,$CD,$8F,$E8,$AD,$7F,$8E,$B5,$FE,$21,$E7,$73,$E4,$FE
	.dc.b	$90,$DF,$0F,$FB,$66,$3F,$D2,$63,$C1,$5E,$CC,$BD,$4F,$F8,$A3,$C0
	.dc.b	$FE,$C1,$83,$ED,$B1,$3F,$E8,$E6,$C6,$63,$FC,$65,$3C,$27,$F8,$26
	.dc.b	$BF,$1E,$7F,$08,$77,$33,$7F,$E2,$DA,$FF,$23,$5B,$EE,$FC,$53,$FE
	.dc.b	$9C,$6C,$66,$BB,$CE,$5F,$A6,$74,$3D,$5F,$80,$7F,$A9,$3E,$46,$D1
	.dc.b	$A4,$9E,$24,$3C,$FD,$7F,$AF,$CB,$D5,$1F,$83,$8C,$61,$FD,$8A,$D9
	.dc.b	$3D,$4F,$F5,$9F,$E3,$DF,$2D,$B8,$7C,$66,$BF,$3A,$78,$52,$BE,$97
	.dc.b	$BD,$FD,$16,$7F,$CF,$1E,$C6,$2B,$F9,$DD,$FF,$B0,$27,$2F,$A4,$F3
	.dc.b	$67,$E8,$7E,$81,$92,$9E,$E7,$59,$FB,$23,$F0,$5F,$27,$FE,$CC,$F8
	.dc.b	$BB,$87,$FB,$8E,$7E,$FC,$E5,$B1,$97,$D9,$DB,$FE,$67,$C4,$D2,$4F
	.dc.b	$E4,$27,$C9,$87,$85,$F5,$47,$CC,$F8,$5F,$F6,$EF,$DC,$7E,$55,$B3
	.dc.b	$F9,$06,$BF,$2D,$AF,$C2,$E0,$EC,$FB,$DA,$FC,$9F,$88,$FF,$74,$FE
	.dc.b	$29,$FE,$AE,$C3,$ED,$C6,$1F,$CE,$B6,$7E,$C3,$5E,$C7,$E4,$4F,$D1
	.dc.b	$FC,$57,$FB,$D3,$E2,$7A,$6A,$8C,$5E,$8F,$F7,$6D,$7E,$0E,$32,$FE
	.dc.b	$17,$E3,$7F,$D8,$B3,$FA,$E6,$C6,$63,$F6,$ED,$3B,$FC,$87,$FA,$32
	.dc.b	$3F,$6A,$D8,$FC,$83,$5F,$C5,$35,$FC,$29,$EC,$3C,$11,$D9,$FC,$84
	.dc.b	$F8,$9B,$19,$3F,$47,$19,$8E,$64,$DB,$EF,$3F,$4C,$FF,$72,$7A,$1F
	.dc.b	$EE,$6C,$9F,$15,$FE,$71,$97,$C5,$7F,$BA,$27,$31,$FE,$75,$8C,$4E
	.dc.b	$DB,$19,$AE,$93,$96,$C6,$7E,$2B,$FC,$FD,$7D,$23,$BC,$E9,$FF,$3F
	.dc.b	$D7,$FA,$A4,$8F,$FA,$C5,$AF,$F9,$C1,$AF,$F9,$85,$FE,$59,$8B,$2F
	.dc.b	$69,$F3,$23,$F0,$B1,$96,$C6,$A7,$A1,$DC,$86,$0F,$B6,$BE,$53,$F0
	.dc.b	$2D,$3F,$02,$58,$DC,$76,$99,$3E,$32,$4F,$E4,$D7,$9B,$F9,$AD,$BD
	.dc.b	$4D,$8C,$9F,$9B,$7D,$D3,$E5,$3E,$11,$E8,$29,$FA,$A8,$7E,$21,$FA
	.dc.b	$6E,$1F,$8C,$D8,$C9,$FB,$22,$CF,$53,$EA,$9D,$4A,$67,$7F,$F4,$C6
	.dc.b	$3F,$52,$32,$7F,$29,$E4,$FA,$9F,$5F,$FC,$36,$CD,$24,$F5,$BE,$6D
	.dc.b	$EA,$73,$7E,$26,$3D,$A4,$3B,$9E,$0B,$FE,$8F,$D1,$F7,$E3,$FB,$4B
	.dc.b	$5F,$E5,$31,$CC,$D9,$79,$9D,$3F,$8B,$FB,$0F,$E6,$A4,$AF,$E6,$6F
	.dc.b	$BF,$2F,$3C,$7C,$3C,$BC,$FA,$E5,$B1,$EB,$CF,$D4,$FA,$2B,$8B,$EB
	.dc.b	$6B,$99,$E2,$F5,$9F,$37,$6F,$E7,$F8,$FF,$AC,$12,$7F,$C0,$F1,$97
	.dc.b	$FB,$26,$8F,$C1,$7F,$B4,$36,$33,$18,$3D,$17,$E6,$C7,$C6,$4C,$7C
	.dc.b	$75,$E7,$6B,$E7,$9E,$11,$89,$E1,$1F,$37,$C2,$FD,$B7,$D8,$7F,$82
	.dc.b	$09,$7F,$E1,$E8,$CB,$F1,$F4,$F1,$9F,$E1,$6F,$A7,$91,$FE,$16,$CD
	.dc.b	$18,$8F,$8D,$3C,$1E,$44,$3F,$C0,$87,$F0,$7F,$B1,$FF,$17,$91,$FF
	.dc.b	$17,$99,$38,$1E,$FC,$7F,$11,$E5,$3C,$9F,$E2,$1F,$92,$7F,$90,$72
	.dc.b	$34,$97,$FE,$4E,$87,$E4,$BF,$E5,$26,$C6,$63,$07,$F9,$7A,$39,$C7
	.dc.b	$F9,$79,$AF,$F9,$79,$AF,$FB,$97,$CF,$F1,$BE,$83,$EA,$67,$F0,$DF
	.dc.b	$8A,$7F,$9D,$65,$F6,$7F,$E7,$96,$C6,$63,$07,$FA,$16,$1F,$BF,$D2
	.dc.b	$4F,$FA,$1E,$32,$FF,$A4,$57,$CA,$7F,$A5,$5B,$19,$8C,$1E,$57,$E7
	.dc.b	$5F,$FA,$9E,$5F,$61,$49,$3F,$EA,$98,$CB,$FE,$B3,$5F,$B2,$7F,$AD
	.dc.b	$DB,$19,$8C,$1D,$DF,$8A,$7F,$B2,$E1,$FB,$C7,$F6,$58,$FF,$65,$46
	.dc.b	$1D,$46,$63,$FD,$98,$66,$3F,$CF,$BB,$BE,$03,$FA,$A7,$EE,$3F,$ED
	.dc.b	$92,$4F,$FB,$75,$FF,$72,$7E,$E5,$23,$FD,$D0,$D7,$9F,$91,$FD,$77
	.dc.b	$1F,$F7,$91,$23,$FD,$EC,$FF,$BF,$FD,$B9,$1F,$DF,$72,$64,$F8,$7F
	.dc.b	$EC,$D5,$AA,$FC,$19,$0F,$A2,$7E,$82,$B8,$F8,$5F,$DB,$7E,$CB,$E0
	.dc.b	$61,$FE,$74,$94,$62,$75,$5E,$1A,$D4,$F0,$F8,$AF,$E2,$93,$F1,$9A
	.dc.b	$FF,$79,$94,$62,$FD,$2E,$A7,$EB,$9E,$78,$F6,$FC,$5F,$F3,$D8,$E4
	.dc.b	$D8,$C9,$FB,$87,$14,$F8,$CE,$CD,$9F,$EF,$D5,$E9,$9C,$9E,$1F,$03
	.dc.b	$F8,$DF,$27,$E5,$EB,$E6,$FC,$6F,$42,$3F,$C1,$9E,$FF,$1D,$23,$F8
	.dc.b	$4B,$67,$F2,$55,$D3,$0F,$91,$39,$6D,$70,$3F,$96,$F7,$4B,$E3,$E1
	.dc.b	$FF,$2F,$24,$66,$BF,$81,$61,$D5,$FC,$AF,$C8,$7F,$89,$FE,$D7,$ED
	.dc.b	$B6,$3E,$C3,$5D,$E2,$9E,$F7,$D6,$7F,$3F,$F2,$5F,$E3,$1E,$9F,$C2
	.dc.b	$B6,$3F,$04,$D7,$77,$FA,$57,$F2,$DF,$2C,$F9,$5F,$F1,$D7,$D0,$7E
	.dc.b	$B6,$31,$48,$C4,$EA,$B8,$B5,$FC,$D1,$F8,$63,$FC,$89,$4F,$97,$FF
	.dc.b	$23,$7D,$7F,$56,$CF,$89,$AF,$FB,$DA,$78,$60,$AF,$E0,$92,$F9,$9F
	.dc.b	$E8,$52,$F7,$73,$18,$6F,$93,$FE,$55,$6C,$EE,$BE,$AF,$17,$D8,$E3
	.dc.b	$FE,$21,$F0,$0F,$9B,$FF,$6C,$E3,$86,$DF,$F9,$E6,$75,$F0,$9F,$E6
	.dc.b	$1A,$D3,$67,$AF,$1F,$D7,$7F,$7A,$7E,$A7,$67,$CE,$7E,$BD,$BA,$7E
	.dc.b	$8A,$7E,$14,$E4,$D7,$CE,$FC,$68,$F9,$AD,$95,$23,$0D,$8F,$95,$3F
	.dc.b	$80,$F2,$7F,$6D,$3F,$D1,$A7,$FC,$68,$DD,$7F,$05,$3A,$F9,$EF,$F3
	.dc.b	$F7,$C9,$7E,$15,$B8,$7E,$29,$AF,$D9,$6A,$3F,$CD,$FF,$3F,$FE,$87
	.dc.b	$FA,$0F,$B7,$F0,$28,$C4,$FF,$41,$8C,$66,$AD,$D6,$AB,$FD,$55,$DC
	.dc.b	$E4,$FF,$2B,$D7,$E1,$A3,$1C,$1C,$6B,$FA,$64,$62,$75,$5C,$27,$F1
	.dc.b	$BE,$CF,$F1,$B7,$D0,$D4,$D2,$63,$FE,$4C,$8C,$B6,$33,$B7,$E6,$27
	.dc.b	$2D,$8C,$C7,$82,$78,$12,$74,$D8,$FE,$C9,$F0,$CF,$31,$F0,$DD,$63
	.dc.b	$0D,$8C,$CF,$FB,$4A,$3C,$4F,$F9,$BA,$AF,$BD,$FE,$9C,$8F,$D6,$B5
	.dc.b	$FE,$E4,$D7,$FD,$B0,$D7,$EE,$35,$E8,$7B,$DC,$12,$3B,$92,$34,$D7
	.dc.b	$C2,$D7,$C2,$D7,$C2,$D5,$A7,$DE,$AF,$B7,$AF,$C0,$9F,$C2,$E9,$FE
	.dc.b	$9A,$9C,$9F,$E0,$9F,$2D,$F4,$FF,$ED,$9F,$C1,$B0,$DB,$FE,$F5,$3A
	.dc.b	$C9,$5A,$6C,$EB,$EB,$3E,$34,$29,$18,$6C,$7F,$1A,$35,$19,$FA,$DF
	.dc.b	$F7,$57,$E7,$5E,$56,$DF,$89,$AF,$EA,$EB,$B4,$62,$3F,$DE,$56,$FA
	.dc.b	$8F,$F7,$A7,$C7,$FB,$DB,$3E,$C6,$BC,$E7,$85,$AB,$E5,$CC,$3F,$53
	.dc.b	$AF,$FB,$11,$C1,$FC,$8F,$EC,$F9,$BF,$BF,$F0,$51,$88,$46,$2F,$B4
	.dc.b	$7F,$30,$9D,$7D,$63,$EC,$15,$F9,$FA,$FF,$22,$4E,$63,$11,$FC,$34
	.dc.b	$46,$79,$7F,$A3,$B6,$3D,$3C,$7F,$B3,$E4,$FE,$0B,$1E,$A3,$F8,$B3
	.dc.b	$87,$66,$4F,$C7,$E7,$F0,$5C,$9F,$86,$F1,$FC,$FE,$9F,$E3,$5B,$19
	.dc.b	$C9,$39,$6C,$72,$3F,$2F,$87,$96,$FB,$46,$0F,$CC,$69,$F1,$3E,$A2
	.dc.b	$D9,$A4,$9F,$CC,$FD,$AF,$CA,$9F,$D0,$36,$33,$1F,$BF,$3F,$31,$19
	.dc.b	$8F,$01,$E5,$3E,$2D,$F9,$1B,$87,$E6,$F8,$5F,$5D,$F6,$3E,$AB,$F5
	.dc.b	$CD,$AF,$D5,$B5,$FC,$83,$5F,$E8,$DF,$5F,$F9,$27,$FB,$FB,$5F,$F4
	.dc.b	$23,$5F,$C0,$35,$F1,$9F,$5F,$7F,$B4,$FA,$8F,$C3,$7E,$82,$FA,$BF
	.dc.b	$CA,$37,$5F,$C9,$CE,$A7,$27,$E6,$7D,$67,$E0,$FF,$C8,$5B,$5F,$CF
	.dc.b	$5B,$5F,$E1,$66,$BF,$24,$FC,$4E,$BF,$74,$8E,$EB,$A1,$2B,$A3,$FC
	.dc.b	$BF,$E7,$3F,$41,$E0,$7F,$88,$9B,$5F,$EA,$76,$BF,$A3,$6B,$F1,$0F
	.dc.b	$16,$7F,$11,$C1,$BE,$8F,$DB,$4F,$EB,$7A,$BB,$46,$0F,$DE,$E3,$FC
	.dc.b	$77,$F0,$9F,$E3,$84,$7F,$04,$FF,$45,$C3,$B6,$1F,$00,$FA,$9C,$7F
	.dc.b	$89,$AF,$E4,$D7,$D3,$8F,$C1,$5F,$CC,$23,$07,$83,$F3,$9F,$CD,$72
	.dc.b	$7F,$B0,$5B,$3F,$EA,$F6,$BF,$C9,$5A,$FB,$88,$78,$CF,$A9,$E0,$BE
	.dc.b	$53,$F8,$B3,$71,$FC,$1C,$FF,$86,$63,$1A,$FF,$94,$9B,$1E,$1E,$0F
	.dc.b	$E2,$FF,$92,$F9,$5B,$19,$B4,$E5,$B1,$E9,$F9,$CF,$E3,$FF,$9D,$7B
	.dc.b	$5B,$19,$94,$E5,$B1,$F0,$77,$FE,$4F,$F3,$3F,$15,$B1,$99,$4E,$5B
	.dc.b	$1F,$27,$F0,$BF,$E5,$FE,$B7,$CD,$6C,$66,$53,$96,$C7,$D1,$E9,$EE
	.dc.b	$FD,$16,$BF,$CE,$B8,$3F,$5D,$87,$D9,$9C,$CF,$F6,$CE,$CF,$E7,$FF
	.dc.b	$92,$FC,$D6,$C7,$F3,$D3,$58,$27,$4D,$8F,$C1,$7E,$07,$FD,$23,$E0
	.dc.b	$FF,$0C,$D8,$CC,$A7,$2D,$8F,$C5,$7E,$09,$FD,$43,$F4,$0F,$C7,$36
	.dc.b	$33,$29,$CB,$63,$CD,$F8,$0F,$E1,$1F,$EA,$8D,$7F,$CC,$4D,$4E,$61
	.dc.b	$AF,$F9,$83,$F3,$6F,$CF,$7A,$DF,$D8,$9B,$26,$9F,$A3,$FB,$9F,$EC
	.dc.b	$FF,$77,$64,$D2,$63,$FA,$7F,$5B,$FB,$6F,$85,$FA,$B6,$C6,$69,$39
	.dc.b	$6C,$7F,$0A,$E6,$FE,$84,$FE,$AC,$FE,$5C,$7F,$9F,$FE,$BD,$C3,$FE
	.dc.b	$01,$6D,$7E,$B2,$51,$8E,$DF,$F5,$EB,$63,$F7,$9F,$82,$7F,$80,$BF
	.dc.b	$2A,$FE,$00,$D8,$CD,$A7,$2D,$8F,$E0,$9F,$84,$FF,$82,$BF,$2A,$FE
	.dc.b	$10,$D8,$CC,$A7,$2D,$8F,$E1,$9F,$86,$7E,$E9,$FE,$18,$D7,$F3,$8D
	.dc.b	$8F,$B2,$E6,$3E,$EF,$B5,$FE,$21,$E5,$FF,$25,$69,$FE,$DA,$6C,$FE
	.dc.b	$22,$7F,$8F,$7E,$1F,$FE,$20,$FC,$0F,$F8,$F3,$67,$F8,$F3,$5F,$E3
	.dc.b	$CD,$7F,$8E,$F9,$5F,$CC,$09,$C3,$F5,$B8,$3E,$8F,$E5,$1F,$3B,$8B
	.dc.b	$FC,$5A,$DD,$8D,$3F,$97,$7C,$1F,$F8,$CB,$F2,$AB,$26,$93,$2F,$54
	.dc.b	$72,$6D,$7B,$F5,$EF,$18,$3F,$9C,$7C,$37,$F8,$FB,$F2,$8F,$E7,$8D
	.dc.b	$8C,$DA,$72,$D8,$EF,$F8,$17,$58,$34,$91,$F9,$7F,$4F,$FB,$7B,$FE
	.dc.b	$1B,$9F,$D7,$FC,$57,$F4,$D2,$3F,$D0,$0C,$97,$C0,$FF,$7A,$FE,$91
	.dc.b	$FD,$47,$E3,$3F,$CA,$AD,$93,$4F,$EA,$DF,$23,$F0,$30,$FE,$A8,$D7
	.dc.b	$F9,$23,$67,$F8,$0B,$98,$FD,$61,$F7,$38,$FF,$5F,$C9,$FD,$E6,$BF
	.dc.b	$83,$FC,$97,$F6,$52,$3E,$77,$FC,$E1,$1F,$82,$D3,$E0,$DF,$ED,$A9
	.dc.b	$FA,$6B,$E5,$7F,$BE,$B6,$3E,$13,$7A,$FF,$99,$47,$09,$DF,$B7,$FE
	.dc.b	$76,$FC,$FF,$36,$8C,$61,$FE,$75,$BE,$F1,$88,$FF,$4D,$4E,$E5,$5C
	.dc.b	$BC,$BF,$F5,$5C,$FF,$37,$9F,$F5,$43,$6B,$51,$EC,$7E,$36,$56,$3D
	.dc.b	$BF,$33,$EC,$23,$EC,$7D,$FF,$35,$FE,$6B,$AF,$35,$A3,$11,$E6,$6B
	.dc.b	$FE,$94,$8D,$57,$F9,$C6,$3F,$1C,$7F,$87,$AF,$CF,$F0,$9F,$30,$93
	.dc.b	$F3,$1F,$9F,$F3,$9F,$EB,$68,$F0,$B6,$BE,$2B,$5F,$03,$5F,$C2,$93
	.dc.b	$7F,$BC,$F8,$48,$7F,$A9,$7F,$07,$7E,$81,$FD,$EA,$FF,$D9,$52,$13
	.dc.b	$AF,$5B,$FD,$63,$10,$CD,$71,$71,$3B,$AE,$71,$F1,$7E,$03,$D1,$1F
	.dc.b	$8E,$6E,$9F,$EE,$58,$FF,$67,$1F,$42,$3D,$98,$FA,$AB,$E5,$9E,$9C
	.dc.b	$3C,$15,$71,$F2,$B4,$F0,$E3,$E3,$AF,$41,$F1,$79,$7F,$52,$4D,$7F
	.dc.b	$11,$19,$6F,$C8,$BE,$2D,$F0,$9F,$17,$DD,$7A,$3F,$EC,$FA,$D4,$E6
	.dc.b	$31,$5C,$30,$5F,$07,$E7,$5F,$07,$35,$C2,$F9,$46,$AF,$95,$F1,$85
	.dc.b	$F1,$8C,$36,$FF,$D9,$27,$F8,$9E,$7F,$D1,$ED,$D9,$93,$2E,$95,$FE
	.dc.b	$BA,$AD,$4E,$5B,$3A,$9E,$67,$FB,$AA,$3F,$C6,$99,$FE,$05,$FC,$1E
	.dc.b	$DF,$8D,$6E,$DF,$EF,$A6,$ED,$F2,$8E,$57,$0F,$49,$4F,$C5,$2F,$B1
	.dc.b	$FC,$93,$F8,$B3,$07,$E2,$4B,$1F,$C3,$38,$11,$88,$FC,$CD,$F8,$E3
	.dc.b	$F1,$17,$E4,$23,$E1,$9F,$79,$E4,$21,$F1,$CD,$69,$F3,$33,$F8,$CD
	.dc.b	$AF,$E4,$7C,$5E,$2C,$BD,$3C,$48,$C6,$3F,$C0,$0F,$31,$FC,$D1,$CE
	.dc.b	$3F,$C9,$B8,$F9,$3F,$C5,$9A,$E6,$7E,$33,$96,$BE,$83,$D6,$7A,$4F
	.dc.b	$5B,$FD,$0B,$93,$E3,$FC,$E7,$F3,$A6,$CF,$F2,$46,$BF,$C8,$E5,$3F
	.dc.b	$D0,$39,$BC,$73,$FC,$95,$B3,$FC,$C8,$F1,$C6,$5A,$F9,$09,$A7,$C9
	.dc.b	$7F,$05,$1F,$D5,$B9,$FF,$15,$1F,$CD,$9A,$FF,$4A,$6B,$FC,$D5,$B1
	.dc.b	$FD,$40,$7E,$DB,$FD,$5B,$AB,$F0,$AD,$8C,$D7,$F4,$29,$CB,$FD,$13
	.dc.b	$AF,$D7,$B7,$F6,$5D,$FF,$AF,$C6,$0F,$83,$3E,$AE,$6F,$D5,$3F,$86
	.dc.b	$6B,$FE,$D5,$ED,$F8,$BE,$09,$A4,$CB,$E2,$EB,$F0,$EB,$ED,$35,$FB
	.dc.b	$47,$F8,$2A,$1F,$8F,$FF,$84,$1B,$1F,$EE,$5F,$03,$F4,$73,$93,$FC
	.dc.b	$31,$FB,$54,$9F,$F7,$57,$87,$FA,$88,$C4,$7F,$BA,$FC,$4F,$39,$30
	.dc.b	$FB,$67,$F8,$9B,$6F,$E0,$FD,$12,$BE,$07,$4F,$F0,$9F,$41,$E4,$FE
	.dc.b	$D3,$E2,$3F,$C7,$04,$AF,$F2,$4B,$FE,$3E,$FC,$AA,$47,$F9,$09,$AF
	.dc.b	$FA,$B2,$7D,$6F,$F9,$43,$6F,$F2,$11,$1F,$F0,$87,$13,$CE,$FE,$07
	.dc.b	$F9,$47,$F9,$68,$93,$A8,$CB,$FE,$CF,$F4,$7F,$D2,$F7,$FE,$65,$24
	.dc.b	$67,$D1,$F0,$BF,$53,$9F,$EE,$BF,$50,$FD,$7F,$0F,$FA,$F3,$D2,$FE
	.dc.b	$27,$F8,$C7,$F9,$F2,$0D,$24,$FF,$9A,$B8,$FF,$CB,$84,$70,$7F,$A5
	.dc.b	$25,$F8,$CF,$F8,$D3,$67,$AD,$FC,$82,$30,$DA,$E1,$1F,$E9,$EE,$39
	.dc.b	$1F,$EA,$26,$BF,$E0,$3B,$FF,$69,$FB,$5F,$CA,$BF,$25,$FF,$56,$92
	.dc.b	$7F,$D6,$2F,$FA,$27,$29,$1F,$16,$BF,$7B,$C1,$FE,$C0,$9F,$C2,$4F
	.dc.b	$C7,$E6,$FE,$4F,$BF,$C8,$BF,$F7,$0F,$3F,$F3,$8F,$B8,$FF,$68,$12
	.dc.b	$FF,$DA,$8F,$FB,$77,$F1,$79,$1F,$7B,$E0,$3F,$9F,$C6,$23,$D5,$C2
	.dc.b	$F8,$2F,$E8,$7F,$9D,$FF,$BA,$89,$3F,$86,$8C,$BF,$EF,$AF,$8F,$FD
	.dc.b	$DF,$11,$A4,$D3,$F4,$38,$3E,$1B,$FA,$7F,$E4,$BF,$4D,$B8,$7E,$49
	.dc.b	$AF,$91,$AF,$F5,$69,$FF,$72,$FC,$F7,$DB,$EA,$FE,$B2,$DC,$0D,$3F
	.dc.b	$BF,$7C,$8B,$E2,$FF,$AF,$FE,$4F,$86,$E3,$FE,$3D,$3F,$C7,$D3,$BB
	.dc.b	$56,$DB,$3F,$A8,$F8,$DF,$ED,$3E,$2F,$E3,$5B,$19,$94,$E5,$B1,$FB
	.dc.b	$0F,$8F,$FE,$E3,$DD,$F9,$56,$C6,$67,$F2,$ED,$9F,$CD,$74,$76,$8F
	.dc.b	$B2,$D7,$E3,$36,$5B,$7F,$87,$1F,$0C,$FE,$03,$DB,$FF,$00,$FE,$E1
	.dc.b	$FA,$56,$C7,$F0,$46,$BB,$94,$FE,$AF,$E4,$FF,$51,$3F,$8A,$6B,$F5
	.dc.b	$DB,$3E,$27,$30,$DB,$F1,$13,$FD,$79,$BD,$BF,$63,$97,$FA,$3C,$62
	.dc.b	$BF,$93,$F6,$7E,$99,$FE,$7B,$7E,$E3,$FC,$37,$F5,$9F,$BF,$6C,$7F
	.dc.b	$1E,$6B,$FA,$06,$BF,$E2,$2F,$6D,$E5,$FF,$74,$6C,$FF,$3A,$6E,$5F
	.dc.b	$EB,$E6,$D7,$90,$87,$F0,$55,$F2,$DF,$DB,$9B,$5F,$E4,$C0,$E3,$1E
	.dc.b	$93,$FD,$DA,$9F,$1D,$D9,$FC,$B3,$5F,$F4,$A3,$5D,$E3,$FD,$BB,$D1
	.dc.b	$FE,$C5,$1F,$DB,$FC,$3F,$E1,$67,$F9,$5E,$BD,$BF,$49,$FE,$63,$6C
	.dc.b	$FE,$8C,$E7,$9A,$75,$87,$F9,$21,$FF,$35,$7A,$CE,$1F,$43,$6B,$FC
	.dc.b	$C8,$DD,$37,$3E,$DC,$FC,$A4,$E5,$FE,$8B,$97,$F8,$7E,$F8,$3E,$36
	.dc.b	$BF,$B7,$6B,$FE,$B5,$6A,$F9,$87,$57,$F8,$41,$B9,$7F,$8B,$1B,$C3
	.dc.b	$FE,$83,$6D,$7D,$A3,$FA,$B6,$3F,$2B,$E6,$5F,$41,$F5,$5B,$87,$FA
	.dc.b	$05,$BA,$FF,$A3,$5B,$7F,$EF,$DE,$3F,$E1,$9C,$9F,$89,$EE,$FC,$8E
	.dc.b	$1F,$82,$6C,$EA,$BC,$47,$D9,$C3,$F2,$1F,$37,$FE,$78,$FA,$CB,$81
	.dc.b	$A4,$AF,$07,$CF,$7F,$95,$1F,$CF,$36,$7F,$22,$DC,$0E,$E7,$C0,$7C
	.dc.b	$39,$FC,$BF,$CF,$FF,$A2,$0F,$D1,$16,$3D,$CD,$7F,$44,$D7,$E5,$F1
	.dc.b	$FD,$ED,$FF,$01,$D1,$F9,$4D,$3F,$79,$5E,$03,$F9,$A7,$D0,$7E,$CB
	.dc.b	$AB,$FD,$70,$D8,$FC,$B3,$5D,$F4,$FF,$A7,$27,$FD,$7D,$F4,$3E,$CD
	.dc.b	$3F,$CD,$6D,$E6,$CE,$DF,$E5,$48,$C6,$1F,$43,$4F,$2E,$7F,$BF,$E5
	.dc.b	$FE,$08,$4C,$7F,$82,$3F,$C1,$FE,$8B,$FD,$57,$FB,$7F,$F0,$56,$E1
	.dc.b	$F8,$16,$BF,$C1,$5A,$FF,$11,$F1,$FF,$83,$12,$FF,$83,$71,$FF,$5C
	.dc.b	$6F,$B5,$FE,$B8,$FC,$7B,$F8,$93,$6F,$F1,$CD,$7F,$89,$35,$0D,$3F
	.dc.b	$D6,$9D,$FF,$ED,$A2,$47,$F0,$18,$CB,$FE,$E7,$FA,$6F,$E2,$89,$A4
	.dc.b	$C3,$FD,$C5,$19,$6C,$67,$6F,$F7,$84,$E5,$B1,$98,$ED,$3C,$09,$3A
	.dc.b	$6C,$6A,$30,$47,$04,$3F,$BE,$7C,$73,$E3,$DF,$21,$E2,$8C,$36,$33
	.dc.b	$3F,$C8,$44,$E9,$C7,$E6,$8F,$8D,$FF,$74,$C6,$1A,$FF,$29,$8D,$63
	.dc.b	$F0,$1B,$1B,$8C,$4F,$08,$FF,$09,$4E,$9E,$51,$97,$EC,$C6,$5D,$C6
	.dc.b	$54,$65,$DC,$66,$31,$1A,$8C,$12,$7C,$64,$8F,$31,$0F,$1B,$96,$99
	.dc.b	$3F,$C5,$59,$FF,$BF,$F1,$F9,$4D,$8F,$C6,$75,$BA,$FE,$7C,$70,$8C
	.dc.b	$57,$D1,$FA,$9F,$F3,$BE,$1A,$AF,$CC,$C6,$27,$5E,$6A,$B5,$4A,$D3
	.dc.b	$FE,$3D,$87,$C4,$BE,$CF,$87,$E8,$1F,$D7,$E7,$31,$8A,$E1,$5B,$95
	.dc.b	$6E,$30,$DE,$0F,$17,$E1,$16,$8D,$25,$FC,$2D,$7F,$47,$B3,$6F,$89
	.dc.b	$3F,$21,$AE,$CF,$07,$CC,$FB,$F7,$B8,$FC,$53,$DB,$19,$9F,$4B,$5F
	.dc.b	$F4,$C6,$9F,$D1,$E7,$F3,$C7,$C4,$F1,$3F,$1C,$D3,$31,$B9,$F2,$65
	.dc.b	$E8,$C7,$CB,$1F,$66,$30,$70,$FC,$5B,$EA,$7C,$CF,$C7,$6C,$FC,$C6
	.dc.b	$BE,$46,$BF,$6E,$FE,$D6,$0F,$AB,$FA,$1C,$5F,$7D,$B1,$9A,$FA,$F3
	.dc.b	$97,$E4,$10,$F3,$1E,$A3,$FA,$5F,$93,$F8,$9E,$5B,$5F,$BD,$9F,$E9
	.dc.b	$B3,$96,$C6,$5F,$97,$F5,$DE,$88,$FC,$6B,$71,$34,$F6,$CF,$F3,$1D
	.dc.b	$7E,$4B,$FC,$D3,$87,$F0,$5B,$7C,$A8,$F3,$1F,$92,$FC,$53,$EA,$E0
	.dc.b	$69,$31,$FA,$FF,$6F,$FA,$3F,$CD,$3F,$2C,$DB,$FC,$DB,$5F,$C1,$B5
	.dc.b	$F1,$65,$E4,$7C,$D8,$F8,$DF,$94,$7D,$5F,$86,$FC,$63,$70,$FC,$96
	.dc.b	$49,$CC,$7E,$C1,$B0,$FC,$03,$F7,$1D,$9F,$AB,$6D,$FE,$C1,$AF,$E6
	.dc.b	$1A,$F9,$B0,$F2,$3F,$70,$F2,$E1,$FB,$EF,$C4,$B6,$D3,$F9,$2C,$7E
	.dc.b	$F2,$33,$81,$BF,$6B,$F6,$49,$4E,$6F,$E0,$11,$F6,$B8,$3E,$BE,$1F
	.dc.b	$58,$F1,$10,$FC,$D4,$7C,$02,$57,$C6,$7E,$2F,$EA,$5F,$DD,$21,$F5
	.dc.b	$1F,$C4,$E3,$0D,$AF,$E3,$CD,$9D,$C7,$D4,$3E,$B9,$E3,$3E,$31,$CF
	.dc.b	$3F,$E4,$DF,$86,$59,$34,$93,$FC,$9E,$DF,$84,$7E,$75,$B5,$F0,$E7
	.dc.b	$53,$93,$E2,$70,$FE,$27,$3F,$8B,$3E,$56,$BF,$E1,$2F,$CB,$5F,$B3
	.dc.b	$BE,$3B,$FA,$F4,$61,$B1,$9C,$7F,$BA,$4E,$5F,$E3,$39,$3E,$93,$FA
	.dc.b	$17,$C4,$FC,$96,$E3,$EB,$6B,$F3,$E7,$85,$2B,$E8,$7D,$CF,$F4,$BF
	.dc.b	$A7,$B2,$69,$2F,$E3,$79,$9F,$D4,$7C,$BF,$B0,$D8,$CD,$7F,$82,$9B
	.dc.b	$3F,$77,$E1,$7F,$AC,$7E,$6D,$68,$D2,$57,$D3,$FB,$6F,$EB,$DF,$96
	.dc.b	$FE,$15,$B1,$9A,$4E,$5B,$1F,$47,$8F,$EA,$13,$9B,$E4,$F3,$7F,$69
	.dc.b	$FA,$FF,$C8,$36,$BF,$04,$D7,$E4,$35,$FE,$E1,$C2,$FC,$47,$FA,$4B
	.dc.b	$71,$FA,$B1,$83,$3E,$6F,$E8,$09,$1F,$D6,$67,$FB,$3B,$7E,$FF,$F3
	.dc.b	$F7,$F9,$36,$BF,$65,$AF,$89,$B0,$97,$AB,$FE,$05,$FD,$4F,$F5,$8D
	.dc.b	$BF,$CF,$35,$FC,$23,$5F,$6F,$B3,$FA,$12,$5F,$E8,$63,$F1,$51,$FA
	.dc.b	$0D,$7F,$89,$C7,$EC,$0F,$E4,$FE,$EB,$E0,$3F,$7D,$E9,$FF,$BB,$63
	.dc.b	$13,$FC,$A9,$B8,$7F,$87,$DB,$3F,$C0,$FD,$CF,$F0,$F7,$E4,$56,$8D
	.dc.b	$25,$FE,$E7,$DF,$FF,$12,$7E,$39,$FC,$2D,$B1,$9A,$4E,$5B,$1F,$C0
	.dc.b	$78,$BE,$C4,$7D,$A2,$47,$F1,$5E,$2F,$E5,$BC,$3F,$63,$CB,$C5,$7F
	.dc.b	$2E,$AF,$D7,$56,$BE,$17,$F8,$51,$27,$E5,$4F,$E2,$A3,$E5,$3F,$C6
	.dc.b	$4F,$E3,$67,$F3,$CB,$FF,$5D,$6D,$FC,$97,$E0,$BF,$C8,$0D,$B3,$4F
	.dc.b	$E5,$59,$7A,$AE,$CF,$A9,$FE,$3D,$F0,$DF,$BF,$6C,$7E,$FD,$BE,$07
	.dc.b	$F3,$39,$D4,$7F,$09,$AE,$8F,$C7,$AE,$8F,$F0,$DF,$53,$FC,$A9,$F8
	.dc.b	$E7,$F3,$B6,$C7,$F2,$D6,$BB,$94,$FC,$1F,$27,$F9,$3C,$FE,$2C,$FE
	.dc.b	$99,$3F,$9D,$E5,$FD,$79,$1F,$E4,$6F,$EC,$BE,$1B,$FC,$C9,$FA,$47
	.dc.b	$F4,$B6,$D7,$F3,$D6,$BF,$C8,$DA,$FF,$9A,$FE,$5F,$FC,$D9,$6F,$93
	.dc.b	$FF,$33,$46,$1B,$1F,$E7,$16,$BB,$9F,$E2,$1D,$5F,$A4,$7E,$D9,$FD
	.dc.b	$7E,$9F,$21,$FE,$29,$6D,$7E,$10,$38,$F4,$BC,$33,$F8,$6A,$EA,$7F
	.dc.b	$5F,$AD,$E1,$FE,$5A,$BF,$44,$7F,$97,$9B,$3E,$A9,$D7,$87,$FE,$89
	.dc.b	$FC,$EF,$DA,$D9,$FF,$44,$57,$F9,$D6,$D5,$A6,$CE,$BE,$5B,$FC,$8C
	.dc.b	$FB,$5A,$FC,$B6,$C7,$B1,$CC,$7C,$43,$EF,$10,$C4,$7C,$2F,$91,$F8
	.dc.b	$C4,$7E,$35,$A3,$E5,$79,$D2,$72,$DD,$3F,$19,$97,$F4,$1C,$3E,$46
	.dc.b	$1F,$3A,$FF,$C2,$51,$F5,$63,$FD,$BB,$1F,$09,$B1,$98,$DB,$F8,$7F
	.dc.b	$9B,$FF,$3F,$C7,$C8,$6C,$7D,$F6,$BB,$8F,$F1,$61,$F6,$98,$7E,$17
	.dc.b	$D3,$DC,$7E,$09,$B5,$FE,$03,$C3,$FD,$DF,$39,$39,$D7,$E3,$7E,$27
	.dc.b	$64,$8C,$C7,$E7,$63,$9F,$D0,$49,$C9,$F9,$07,$15,$FE,$F5,$E5,$FD
	.dc.b	$9D,$FD,$72,$3F,$B6,$AF,$CB,$FD,$D7,$98,$8F,$4C,$51,$F0,$AB,$8E
	.dc.b	$9F,$4D,$F8,$27,$E0,$6F,$B9,$E0,$F4,$5F,$10,$F2,$15,$FB,$5D,$6A
	.dc.b	$AD,$CE,$B9,$15,$A9,$CD,$71,$39,$B8,$9D,$C6,$2F,$C0,$EC,$EC,$E2
	.dc.b	$37,$18,$BF,$05,$F7,$AD,$3F,$EC,$73,$FD,$96,$49,$FF,$72,$35,$D1
	.dc.b	$82,$1F,$C2,$72,$FF,$80,$1F,$21,$F8,$2D,$3E,$6E,$A8,$C4,$EA,$3F
	.dc.b	$76,$DA,$EA,$43,$FD,$19,$7E,$17,$F9,$16,$FB,$6D,$75,$3F,$82,$C6
	.dc.b	$5F,$C4,$19,$8F,$C9,$35,$FC,$83,$4F,$E3,$B3,$C7,$0F,$AB,$19,$7F
	.dc.b	$CC,$18,$7B,$78,$34,$49,$E9,$7F,$EB,$6B,$F4,$34,$F8,$E7,$EB,$EB
	.dc.b	$E3,$70,$F7,$25,$77,$C0,$F8,$AB,$F7,$D7,$F3,$67,$2D,$9D,$4F,$C8
	.dc.b	$3B,$9E,$13,$E6,$3F,$CE,$B8,$BE,$BC,$7F,$2C,$6B,$FD,$0D,$AF,$F5
	.dc.b	$16,$C7,$F2,$13,$EE,$47,$F3,$4E,$3C,$DF,$F2,$19,$CC,$FF,$4D,$9E
	.dc.b	$6F,$8C,$F3,$1F,$7C,$C4,$7F,$33,$1E,$1F,$F5,$EE,$5F,$86,$D8,$CC
	.dc.b	$FF,$3F,$6B,$FD,$6F,$93,$E4,$FE,$E7,$FD,$14,$91,$FD,$2A,$32,$FF
	.dc.b	$4C,$E3,$F9,$6D,$AF,$D4,$BF,$DC,$63,$2F,$F6,$76,$BE,$A2,$1E,$D3
	.dc.b	$D7,$97,$F6,$AD,$BF,$B3,$D7,$C4,$9F,$EF,$9D,$1F,$20,$F8,$BC,$CC
	.dc.b	$5D,$3E,$2A,$E9,$39,$8C,$37,$83,$FB,$FF,$C3,$4A,$FE,$CF,$8F,$E2
	.dc.b	$A3,$11,$FE,$0D,$87,$ED,$3F,$E1,$36,$C7,$F8,$4D,$AE,$E3,$FD,$C5
	.dc.b	$DD,$F0,$27,$FC,$40,$7C,$62,$DE,$9F,$2E,$5E,$A8,$FD,$1E,$3E,$7C
	.dc.b	$3F,$47,$5F,$E8,$5F,$03,$F3,$FF,$88,$FF,$8D,$60,$D2,$4F,$F8,$F6
	.dc.b	$FF,$C9,$24,$70,$7F,$92,$3F,$0B,$FF,$23,$51,$A4,$8D,$9F,$E4,$98
	.dc.b	$F5,$79,$1F,$EA,$7A,$E9,$3F,$ED,$5F,$23,$F6,$7F,$8B,$FF,$97,$49
	.dc.b	$1F,$EF,$58,$CB,$FB,$0C,$FD,$CF,$E7,$5A,$FF,$9B,$70,$FC,$11,$FE
	.dc.b	$A5,$E6,$FD,$6E,$3F,$A0,$8C,$B7,$8F,$FC,$F7,$F9,$A4,$BF,$E9,$B5
	.dc.b	$FE,$AD,$F2,$FF,$83,$CE,$67,$D9,$A7,$A9,$FF,$62,$F9,$1F,$67,$1F
	.dc.b	$74,$7F,$A4,$7F,$13,$FF,$48,$11,$C1,$DA,$1E,$8F,$F1,$4F,$CB,$3F
	.dc.b	$D4,$04,$9D,$46,$5F,$EC,$3E,$77,$FA,$AB,$E5,$7F,$D5,$DC,$FF,$D9
	.dc.b	$B6,$46,$2B,$F8,$FB,$63,$31,$F2,$F9,$FF,$14,$DE,$AF,$AF,$EA,$7C
	.dc.b	$82,$57,$C8,$7E,$5F,$AB,$FC,$AB,$F7,$3F,$F6,$7C,$9A,$4B,$FF,$6E
	.dc.b	$43,$EB,$3F,$BC,$C7,$FB,$6E,$30,$EA,$33,$1F,$ED,$DF,$67,$F9,$BC
	.dc.b	$7F,$AC,$33,$95,$F2,$79,$7D,$AD,$BF,$6B,$F5,$8F,$C9,$9F,$59,$FC
	.dc.b	$9F,$1F,$D7,$F6,$BF,$A1,$FD,$77,$83,$91,$18,$BF,$AC,$D7,$E7,$46
	.dc.b	$AD,$5F,$B2,$E0,$F9,$75,$F9,$56,$BF,$D4,$29,$EF,$7F,$52,$FC,$AF
	.dc.b	$C0,$D9,$FC,$23,$5E,$ED,$79,$7C,$0F,$F5,$AF,$CE,$BD,$0D,$8F,$C6
	.dc.b	$35,$F3,$B5,$F1,$7C,$1F,$F6,$2F,$DC,$FE,$03,$63,$F2,$8D,$7D,$ED
	.dc.b	$7D,$5F,$0B,$FD,$AB,$C4,$F9,$0D,$8F,$CE,$35,$F8,$ED,$7E,$17,$C3
	.dc.b	$7F,$72,$F8,$FF,$A0,$D8,$FD,$23,$5F,$9E,$D5,$F1,$0F,$89,$7A,$BF
	.dc.b	$D2,$23,$0D,$8C,$F1,$FF,$50,$9C,$BF,$CA,$BB,$67,$87,$F6,$5F,$E2
	.dc.b	$FF,$15,$FC,$20,$93,$FC,$21,$FE,$17,$1F,$E0,$AF,$B6,$FC,$43,$72
	.dc.b	$FD,$E3,$5F,$C3,$B5,$FE,$B7,$E4,$FF,$18,$7F,$A8,$F3,$FE,$1B,$07
	.dc.b	$3F,$EE,$FB,$7F,$29,$5F,$A2,$3F,$84,$14,$FC,$99,$4F,$F0,$EF,$C7
	.dc.b	$BE,$3B,$F4,$B1,$FC,$16,$31,$7F,$A6,$8D,$CE,$9B,$19,$7C,$B9,$7C
	.dc.b	$67,$FA,$6D,$72,$F2,$BF,$6E,$D8,$FE,$52,$D7,$C1,$7F,$E1,$E8,$7C
	.dc.b	$7F,$F2,$C9,$DB,$5F,$33,$5F,$5B,$5F,$1E,$FF,$5C,$F0,$FC,$86,$F2
	.dc.b	$5A,$30,$D8,$79,$9C,$1B,$7E,$8E,$6F,$85,$1C,$9B,$46,$9F,$1E,$7C
	.dc.b	$67,$F2,$9C,$3E,$1F,$C8,$FF,$93,$BF,$3A,$B8,$1A,$4A,$F8,$FF,$29
	.dc.b	$FE,$01,$C4,$D2,$57,$D1,$9F,$E1,$9B,$D7,$EF,$FC,$AF,$CD,$24,$FF
	.dc.b	$62,$B4,$7F,$91,$BE,$85,$EB,$7F,$90,$2F,$F9,$9C,$62,$FF,$C7,$F1
	.dc.b	$FE,$B0,$8C,$7A,$7F,$CE,$21,$18,$7E,$01,$0F,$C7,$7B,$1F,$E6,$EF
	.dc.b	$BE,$FC,$2B,$63,$F0,$AD,$77,$69,$FC,$5F,$CD,$7E,$1C,$91,$EB,$7F
	.dc.b	$C7,$7D,$3F,$CB,$2B,$E8,$9F,$E7,$2D,$1F,$37,$F2,$C9,$7F,$E2,$B8
	.dc.b	$CB,$7E,$73,$EE,$91,$FD,$21,$F8,$A7,$EF,$76,$7E,$DC,$98,$78,$5F
	.dc.b	$F0,$8C,$FE,$43,$E1,$FF,$D1,$BF,$75,$FA,$46,$E5,$FA,$46,$BF,$B3
	.dc.b	$6B,$FB,$8F,$9F,$FF,$4A,$FE,$8B,$FA,$C6,$C7,$EB,$1A,$FE,$ED,$AF
	.dc.b	$FB,$9B,$77,$D0,$FE,$55,$B8,$FE,$ED,$BE,$A7,$DB,$6D,$7F,$74,$CF
	.dc.b	$FC,$C5,$3F,$7D,$B1,$9D,$3F,$CD,$33,$96,$C7,$F7,$CF,$95,$7C,$FF
	.dc.b	$F9,$EA,$11,$8C,$7F,$CE,$0D,$EC,$FC,$63,$67,$95,$71,$8C,$47,$FC
	.dc.b	$00,$6A,$3A,$C6,$C9,$5D,$9B,$5D,$9B,$3D,$A7,$44,$9D,$36,$35,$18
	.dc.b	$23,$82,$1F,$F0,$07,$E3,$BF,$4C,$BE,$3D,$C0,$9C,$C6,$1B,$19,$9F
	.dc.b	$87,$F8,$0F,$FB,$59,$B8,$74,$C9,$39,$9F,$A0,$D7,$E3,$35,$F8,$D1
	.dc.b	$C2,$79,$4E,$67,$EA,$B6,$75,$1E,$02,$1D,$A7,$89,$1C,$9E,$07,$0E
	.dc.b	$CC,$9E,$29,$EC,$48,$DC,$71,$36,$79,$1C,$BC,$5F,$9B,$AF,$FA,$DF
	.dc.b	$0F,$C9,$76,$7C,$9A,$FC,$20,$9D,$3F,$99,$8D,$2A,$7C,$47,$E9,$BF
	.dc.b	$02,$FF,$28,$B6,$7F,$D0,$6D,$7F,$CD,$CD,$79,$2F,$B8,$E9,$EE,$72
	.dc.b	$6C,$F2,$6B,$C9,$AF,$E6,$F1,$FF,$69,$3E,$C2,$1E,$93,$F5,$DF,$7F
	.dc.b	$FC,$DD,$BE,$07,$DF,$D9,$39,$9F,$D4,$35,$F4,$FD,$D7,$E1,$74,$F8
	.dc.b	$2D,$CC,$D3,$F0,$D8,$F9,$2B,$C4,$43,$E7,$B9,$7E,$2F,$E0,$DF,$37
	.dc.b	$E5,$BE,$5B,$63,$E5,$B5,$DE,$09,$FA,$3F,$67,$99,$F9,$AD,$9F,$AE
	.dc.b	$DC,$4E,$E7,$F0,$1C,$FF,$EF,$E7,$E4,$9E,$67,$0F,$A4,$FA,$0E,$1D
	.dc.b	$70,$FE,$3A,$31,$1A,$8F,$EC,$57,$F8,$8F,$47,$C4,$47,$D5,$3F,$88
	.dc.b	$8F,$26,$7F,$16,$39,$1E,$42,$9F,$EA,$4F,$C4,$3F,$15,$5F,$C2,$9B
	.dc.b	$1F,$DC,$5A,$EF,$87,$EC,$7E,$01,$F9,$BF,$A4,$FC,$CB,$67,$F3,$2D
	.dc.b	$7F,$32,$D7,$C2,$7C,$03,$C8,$7C,$AC,$BD,$FA,$BC,$1F,$C3,$13,$2C
	.dc.b	$BF,$C3,$FD,$67,$ED,$5F,$E1,$09,$FE,$83,$87,$ED,$27,$3E,$5F,$EC
	.dc.b	$79,$BF,$D8,$71,$8A,$DD,$7F,$3D,$8C,$4F,$93,$7F,$95,$3E,$53,$E5
	.dc.b	$53,$F5,$97,$E0,$7F,$BD,$EC,$FF,$65,$C6,$27,$F8,$09,$B9,$D5,$7F
	.dc.b	$C0,$09,$D5,$70,$FB,$4F,$D3,$41,$A4,$BF,$C3,$6F,$2F,$CD,$CB,$F8
	.dc.b	$89,$F2,$56,$E7,$A4,$7E,$2A,$FA,$37,$F0,$4F,$E2,$9F,$31,$E8,$6D
	.dc.b	$79,$5A,$FF,$BF,$E9,$18,$8F,$B6,$43,$E3,$1F,$9F,$AF,$8D,$1F,$C8
	.dc.b	$7E,$27,$B4,$7F,$B8,$27,$B3,$5E,$D3,$B7,$E3,$7E,$0B,$FB,$3B,$34
	.dc.b	$95,$FA,$29,$FC,$61,$FA,$F3,$C6,$60,$FE,$AB,$F6,$1F,$87,$7F,$C1
	.dc.b	$4D,$7F,$C0,$0D,$77,$C5,$F5,$3F,$06,$FE,$75,$F5,$BF,$79,$B3,$F6
	.dc.b	$DA,$FC,$76,$BF,$D0,$71,$FF,$0A,$7E,$51,$F8,$6F,$2B,$FA,$23,$72
	.dc.b	$34,$FC,$5F,$8F,$FD,$27,$E0,$AC,$9A,$4C,$3F,$19,$F6,$9D,$DB,$3F
	.dc.b	$97,$6E,$3F,$96,$9D,$47,$DD,$F4,$3E,$E5,$7E,$75,$B4,$69,$F8,$19
	.dc.b	$7D,$8B,$E9,$7F,$AF,$7E,$1D,$FE,$79,$8C,$60,$8C,$4E,$AB,$FB,$49
	.dc.b	$C9,$C3,$F8,$FF,$3B,$FB,$47,$EB,$1F,$AC,$6C,$7E,$A5,$AE,$ED,$3F
	.dc.b	$B2,$F5,$BF,$56,$48,$FD,$13,$FD,$67,$93,$EF,$57,$F9,$21,$BD,$E6
	.dc.b	$9A,$AF,$E0,$1E,$C7,$F7,$BF,$8B,$FD,$FB,$6F,$F7,$8D,$7F,$50,$D7
	.dc.b	$F0,$5E,$D7,$F8,$0F,$EE,$7F,$83,$B6,$3F,$82,$35,$FD,$83,$5F,$C5
	.dc.b	$7B,$1F,$A3,$26,$1F,$A1,$8C,$B5,$FD,$09,$FB,$5F,$13,$F8,$D9,$FD
	.dc.b	$17,$3F,$77,$5F,$C3,$25,$7F,$99,$B4,$F8,$1C,$9F,$E3,$BD,$7F,$22
	.dc.b	$4C,$7F,$8D,$73,$FB,$6F,$F9,$26,$1F,$E1,$FF,$C5,$3F,$91,$B6,$BF
	.dc.b	$90,$35,$FE,$18,$D7,$F4,$DC,$1F,$5D,$FB,$55,$F6,$0F,$E3,$5F,$01
	.dc.b	$FC,$AA,$3F,$8A,$35,$FD,$73,$63,$F9,$B3,$72,$7C,$17,$F0,$B2,$4F
	.dc.b	$C9,$3F,$9E,$3F,$DA,$BC,$4F,$EE,$77,$F2,$89,$BF,$F6,$DA,$FE,$85
	.dc.b	$F0,$7D,$48,$FF,$40,$C3,$FD,$F4,$DC,$BD,$97,$FD,$05,$AF,$AA,$3F
	.dc.b	$A9,$BE,$A6,$9E,$A2,$CF,$FB,$E1,$B3,$CA,$7E,$4C,$7C,$77,$FA,$9F
	.dc.b	$C2,$FF,$94,$7E,$C2,$C9,$A4,$9F,$E5,$18,$FE,$1B,$2F,$82,$FB,$F1
	.dc.b	$7C,$3F,$F5,$62,$61,$FD,$35,$FE,$B3,$F1,$1F,$E5,$FF,$AF,$FE,$C6
	.dc.b	$DC,$7F,$B0,$35,$FE,$98,$D7,$FC,$CD,$FA,$D7,$FA,$BE,$FF,$C4,$2D
	.dc.b	$8F,$E0,$F0,$8C,$47,$F9,$CA,$30,$FF,$92,$B2,$F4,$93,$CC,$FE,$D1
	.dc.b	$1B,$A4,$7F,$8F,$DB,$5F,$83,$6F,$B3,$E0,$6D,$7C,$6B,$E8,$FF,$81
	.dc.b	$23,$F9,$A5,$7A,$3C,$4F,$F3,$04,$62,$79,$46,$3E,$43,$FA,$0C,$78
	.dc.b	$DA,$FC,$19,$46,$23,$E2,$B6,$BE,$17,$47,$62,$3F,$21,$F8,$59,$7C
	.dc.b	$0A,$FE,$C4,$7F,$78,$9F,$95,$F2,$5F,$04,$8F,$C1,$7E,$27,$CA,$48
	.dc.b	$F8,$8D,$D3,$FC,$23,$7D,$6F,$A6,$4F,$96,$FF,$66,$61,$F3,$DB,$3E
	.dc.b	$D6,$BF,$80,$6C,$7D,$DF,$3F,$FC,$F8,$FD,$76,$CF,$C5,$6B,$F8,$86
	.dc.b	$BF,$B6,$DB,$EA,$BF,$E1,$3B,$7C,$CF,$0B,$FC,$53,$63,$31,$FE,$85
	.dc.b	$3F,$0D,$19,$F8,$39,$19,$AF,$AD,$18,$CF,$FD,$CD,$3F,$95,$F9,$AE
	.dc.b	$84,$78,$57,$E7,$A3,$CF,$CD,$FA,$FD,$7E,$54,$FF,$38,$AF,$CF,$7C
	.dc.b	$E7,$94,$8F,$23,$27,$48,$F8,$13,$CA,$BA,$47,$D6,$3F,$CD,$FF,$0E
	.dc.b	$F2,$5E,$BA,$5E,$B3,$E2,$3F,$DB,$B7,$BA,$D4,$E6,$B8,$4E,$5E,$35
	.dc.b	$C2,$FA,$C6,$0D,$B8,$AE,$50,$AE,$51,$86,$D7,$3B,$F0,$BF,$EA,$83
	.dc.b	$FD,$E6,$4A,$FF,$78,$35,$D1,$82,$1F,$AD,$CF,$FD,$E9,$F2,$5F,$63
	.dc.b	$85,$18,$D7,$F8,$33,$6B,$84,$2B,$FD,$F9,$1D,$88,$7F,$AE,$6F,$C4
	.dc.b	$FD,$F9,$E0,$7F,$11,$DB,$A4,$66,$BC,$51,$89,$FE,$3D,$C1,$71,$34
	.dc.b	$91,$D2,$FF,$84,$70,$EC,$91,$98,$FE,$5B,$1E,$07,$FD,$61,$8F,$99
	.dc.b	$A7,$C6,$3F,$59,$3C,$4F,$DA,$16,$DC,$7E,$53,$CD,$C8,$9F,$C5,$1F
	.dc.b	$C0,$4B,$5F,$C0,$5A,$FF,$01,$6B,$E2,$3E,$59,$4F,$98,$7F,$39,$E5
	.dc.b	$FB,$2D,$8F,$A8,$DD,$3F,$9C,$4E,$A3,$F9,$67,$35,$7F,$A7,$B6,$BF
	.dc.b	$91,$9C,$A3,$6D,$7C,$27,$93,$8D,$CD,$FD,$2B,$4F,$BF,$5F,$CD,$0F
	.dc.b	$CE,$D7,$F5,$CE,$7F,$89,$3F,$D2,$DB,$1F,$D4,$DA,$FF,$41,$E8,$F9
	.dc.b	$5F,$91,$7F,$51,$6C,$7F,$47,$B4,$E5,$B1,$FD,$77,$AB,$9B,$FD,$B1
	.dc.b	$B1,$FD,$A5,$BC,$1F,$D6,$31,$FC,$16,$7E,$CE,$9F,$4E,$1F,$44,$FA
	.dc.b	$58,$7F,$71,$8F,$91,$7F,$DA,$37,$FE,$CD,$18,$3E,$75,$71,$BE,$98
	.dc.b	$76,$D7,$E4,$4F,$F7,$33,$C6,$4C,$7E,$49,$FD,$F0,$BD,$CE,$CF,$C7
	.dc.b	$7D,$37,$F8,$3A,$4D,$25,$7F,$BB,$FB,$FE,$0D,$FF,$86,$AB,$9C,$60
	.dc.b	$F8,$F3,$E7,$C3,$CA,$FD,$F6,$BF,$EB,$BF,$07,$F3,$DF,$33,$FE,$2C
	.dc.b	$24,$7F,$BF,$63,$2F,$F8,$FE,$FD,$3E,$1F,$88,$8F,$B3,$C6,$FD,$47
	.dc.b	$CA,$FF,$87,$B3,$FF,$1F,$36,$33,$7F,$E1,$F9,$CB,$D7,$6F,$F3,$39
	.dc.b	$1C,$1B,$F1,$BF,$69,$F1,$DD,$5B,$19,$C5,$39,$6C,$7F,$03,$3E,$27
	.dc.b	$17,$E1,$38,$BE,$DC,$78,$FC,$8F,$D5,$65,$FE,$6E,$BF,$D4,$1F,$E6
	.dc.b	$C2,$B9,$E4,$F6,$47,$6A,$FF,$31,$63,$82,$3E,$A3,$FC,$89,$4D,$BE
	.dc.b	$4E,$9E,$F9,$FF,$45,$DB,$F0,$9F,$F4,$8B,$63,$FD,$22,$D7,$71,$FE
	.dc.b	$EC,$F8,$79,$1F,$E9,$C9,$46,$23,$FB,$37,$9F,$FE,$A8,$9F,$DE,$B6
	.dc.b	$FF,$DB,$6D,$9F,$ED,$87,$FB,$37,$CC,$FC,$6E,$DF,$22,$3F,$19,$BF
	.dc.b	$CB,$E5,$FC,$6E,$2D,$5E,$87,$CC,$24,$7C,$A7,$E7,$FA,$3F,$CB,$FF
	.dc.b	$46,$FF,$65,$C9,$A4,$C7,$FD,$AB,$1E,$C6,$ED,$FE,$DA,$8C,$47,$FB
	.dc.b	$61,$AF,$A5,$F7,$FA,$4D,$7F,$86,$ED,$EF,$AF,$8B,$39,$AF,$CF,$93
	.dc.b	$7F,$AD,$3F,$E9,$67,$FD,$F9,$0F,$DC,$69,$27,$FD,$FD,$19,$7E,$58
	.dc.b	$F6,$5E,$C7,$F4,$FF,$C2,$73,$39,$8C,$3C,$27,$74,$AD,$C6,$1F,$D8
	.dc.b	$7B,$7F,$D6,$78,$2D,$7D,$2B,$FD,$84,$E5,$B1,$97,$F2,$BE,$E7,$F6
	.dc.b	$1F,$DD,$FF,$74,$D8,$FD,$C3,$5D,$CA,$7F,$80,$7B,$3F,$DA,$7E,$CE
	.dc.b	$C9,$A4,$AF,$E0,$5F,$01,$FD,$11,$FD,$2B,$5F,$EE,$5C,$C8,$C5,$7E
	.dc.b	$35,$B0,$F8,$1F,$EF,$1F,$45,$FC,$45,$B7,$FC,$3D,$AF,$EC,$9A,$FE
	.dc.b	$17,$8B,$F4,$3C,$7F,$81,$6F,$2F,$E1,$A3,$13,$FC,$2F,$AF,$F4,$4F
	.dc.b	$F6,$C2,$67,$FA,$1F,$84,$FE,$16,$47,$F8,$4B,$FC,$3B,$E1,$7F,$C1
	.dc.b	$DF,$1F,$FC,$9D,$B5,$FC,$95,$AF,$F0,$86,$BD,$CF,$E3,$79,$7F,$78
	.dc.b	$D5,$3F,$76,$78,$61,$F7,$5F,$97,$19,$6C,$64,$FD,$C7,$C3,$FF,$4E
	.dc.b	$23,$FD,$61,$FE,$03,$DD,$FD,$2A,$7C,$27,$F8,$BE,$78,$9B,$26,$1F
	.dc.b	$D0,$EB,$FA,$CF,$C4,$7F,$38,$AF,$E4,$6D,$7E,$FB,$6B,$F8,$53,$98
	.dc.b	$EE,$7F,$10,$AF,$D5,$FC,$4B,$E2,$BF,$71,$E5,$61,$B5,$FC,$6A,$75
	.dc.b	$C3,$FD,$4A,$3F,$D1,$11,$8C,$57,$FE,$F6,$F8,$C7,$C6,$7F,$84,$1B
	.dc.b	$3F,$DF,$1B,$96,$E7,$E2,$F8,$BF,$C8,$C9,$1F,$E2,$C9,$7C,$6F,$F9
	.dc.b	$37,$E8,$FB,$37,$3F,$E5,$F8,$AB,$6A,$B6,$D9,$FE,$17,$F2,$1F,$CF
	.dc.b	$A3,$FC,$3E,$DC,$3F,$85,$B6,$77,$E3,$79,$A7,$FC,$35,$97,$9A,$7C
	.dc.b	$38,$B9,$7F,$C9,$8D,$C7,$FB,$1B,$7C,$CF,$8D,$3F,$E2,$6E,$8F,$D1
	.dc.b	$DF,$D6,$3F,$2C,$E0,$9C,$BF,$89,$7F,$53,$87,$BF,$E4,$7B,$91,$EE
	.dc.b	$F8,$BE,$4A,$47,$89,$B8,$FA,$B4,$98,$FA,$BE,$4F,$EC,$12,$FE,$C3
	.dc.b	$FD,$03,$E5,$1F,$2B,$49,$3F,$D9,$23,$2F,$C5,$AF,$C8,$1F,$9F,$DB
	.dc.b	$E8,$72,$7E,$1F,$87,$FD,$10,$DA,$FC,$4D,$7C,$C9,$CB,$F3,$BE,$5F
	.dc.b	$FD,$1D,$FB,$77,$E3,$DB,$19,$A4,$E5,$B1,$FD,$2B,$83,$FC,$BD,$E9
	.dc.b	$FF,$48,$6C,$67,$89,$C9,$FE,$15,$8F,$F7,$34,$E9,$FE,$FB,$E7,$BC
	.dc.b	$7F,$F2,$4B,$71,$FF,$34,$B7,$8B,$FA,$AB,$6B,$FC,$0F,$5C,$09,$5C
	.dc.b	$1B,$5C,$1B,$3C,$27,$44,$9D,$36,$35,$18,$23,$82,$1E,$A9,$EA,$D9
	.dc.b	$E4,$D9,$E2,$D3,$FB,$9F,$C5,$7F,$C0,$16,$FC,$67,$FC,$00,$6F,$C5
	.dc.b	$FF,$74,$7F,$DF,$D1,$97,$FC,$4B,$F5,$1F,$3B,$0D,$36,$3F,$BD,$B5
	.dc.b	$FF,$00,$B5,$E3,$A7,$FB,$C9,$CB,$5C,$CF,$48,$C4,$04,$F3,$70,$D7
	.dc.b	$31,$C1,$B3,$CC,$8F,$9A,$35,$3C,$48,$E0,$93,$C2,$7C,$ED,$7C,$E7
	.dc.b	$37,$0F,$92,$38,$B9,$21,$C8,$A7,$F9,$43,$4F,$F5,$D6,$1F,$9D,$D7
	.dc.b	$F3,$6D,$AF,$A5,$4A,$DB,$F3,$BE,$2B,$FD,$F5,$5F,$49,$B1,$9C,$38
	.dc.b	$E2,$9C,$CF,$87,$1E,$EF,$EA,$F1,$EF,$8F,$68,$D4,$7E,$0E,$DF,$85
	.dc.b	$3E,$DF,$9D,$7D,$68,$C6,$5B,$6E,$7F,$84,$6C,$FF,$42,$F8,$37,$D4
	.dc.b	$AD,$F3,$7F,$E4,$A5,$3A,$73,$5C,$6B,$85,$6E,$B4,$FC,$9E,$39,$5E
	.dc.b	$D6,$BF,$8D,$BE,$BF,$57,$E1,$F9,$5E,$4D,$9F,$33,$5E,$18,$23,$11
	.dc.b	$E9,$87,$DD,$FE,$6A,$FD,$2D,$8C,$C7,$F9,$88,$C9,$FB,$1E,$0F,$F7
	.dc.b	$E5,$7E,$04,$87,$D5,$3D,$DF,$47,$F1,$FE,$5A,$D1,$A4,$9F,$9D,$F5
	.dc.b	$FF,$5A,$7E,$5B,$70,$34,$FB,$73,$F8,$FE,$4F,$E5,$AF,$88,$FC,$19
	.dc.b	$4F,$81,$F7,$5E,$37,$EE,$35,$F8,$ED,$77,$7E,$77,$FC,$31,$F0,$AF
	.dc.b	$B8,$FB,$71,$86,$BF,$E1,$A8,$D4,$FE,$0A,$31,$3A,$7F,$A0,$DF,$A1
	.dc.b	$F9,$04,$3D,$27,$E4,$7C,$4F,$84,$4A,$FE,$40,$7E,$41,$F8,$DF,$5F
	.dc.b	$F9,$5F,$97,$FC,$73,$6B,$F3,$4D,$7F,$0A,$D7,$C7,$D5,$F7,$70,$F2
	.dc.b	$47,$88,$87,$E2,$7A,$7F,$45,$FA,$3F,$E6,$DB,$1F,$A7,$6B,$F9,$26
	.dc.b	$BF,$A9,$FC,$B3,$F5,$59,$FE,$35,$B1,$FA,$B6,$BF,$E3,$26,$BE,$23
	.dc.b	$F0,$18,$79,$C8,$FF,$AE,$3F,$49,$7B,$5F,$CB,$DB,$7F,$D2,$5B,$3B
	.dc.b	$9F,$D1,$5B,$ED,$E9,$2F,$CC,$FF,$39,$E6,$FD,$0D,$FD,$73,$E4,$9F
	.dc.b	$8F,$95,$3E,$7B,$A9,$23,$E0,$3F,$D2,$BD,$67,$ED,$DF,$B7,$9F,$F6
	.dc.b	$8C,$63,$AF,$53,$9C,$3F,$CD,$ED,$7C,$47,$C4,$3F,$40,$53,$FC,$E1
	.dc.b	$F8,$8B,$EE,$FC,$0D,$C3,$BB,$67,$FC,$F6,$D7,$E9,$71,$7F,$21,$F9
	.dc.b	$AF,$AC,$D8,$CD,$27,$2D,$8F,$B5,$7F,$CA,$7F,$0D,$FB,$CD,$8C,$CA
	.dc.b	$72,$D8,$FC,$17,$DE,$FF,$31,$F8,$3F,$C3,$36,$33,$29,$CB,$63,$F1
	.dc.b	$5F,$7D,$FC,$E7,$F1,$6F,$C7,$36,$33,$29,$CB,$63,$F2,$5E,$AF,$F4
	.dc.b	$1F,$C7,$7F,$2C,$D8,$CC,$A7,$2D,$8F,$F0,$31,$FE,$07,$FC,$3B,$FC
	.dc.b	$B7,$38,$8C,$77,$73,$77,$3A,$74,$D7,$FA,$27,$2F,$F0,$EE,$0F,$D4
	.dc.b	$77,$FF,$55,$6E,$26,$9F,$AD,$FB,$9F,$E3,$D1,$F9,$66,$D1,$A6,$3D
	.dc.b	$2F,$E6,$90,$69,$27,$F8,$9C,$3E,$4F,$F5,$64,$8F,$DF,$3F,$AE,$F8
	.dc.b	$DF,$ED,$5D,$7F,$BF,$6E,$7F,$BF,$6B,$EE,$6B,$FC,$27,$D8,$F1,$92
	.dc.b	$FC,$2F,$F4,$4A,$FD,$AC,$FF,$07,$6B,$FB,$63,$F8,$67,$B3,$EC,$25
	.dc.b	$7A,$9F,$7F,$2F,$F7,$FF,$5F,$F1,$64,$AF,$C7,$BF,$C5,$FD,$CF,$F0
	.dc.b	$27,$E9,$37,$03,$49,$5F,$C6,$FD,$FF,$9E,$49,$FE,$9D,$1F,$72,$3E
	.dc.b	$84,$3D,$2F,$E4,$9C,$1F,$E1,$16,$C6,$6B,$F9,$34,$E5,$FE,$51,$F0
	.dc.b	$1F,$E1,$7F,$63,$F9,$63,$63,$35,$FE,$58,$6C,$FF,$08,$DB,$E3,$C7
	.dc.b	$F0,$93,$FA,$0C,$FF,$1C,$F8,$1E,$51,$F9,$26,$BF,$CF,$5B,$1B,$3B
	.dc.b	$47,$E6,$3E,$13,$F1,$E4,$8F,$83,$1F,$D1,$63,$F2,$7F,$01,$FE,$18
	.dc.b	$9F,$E5,$6D,$E5,$FF,$80,$3E,$BF,$F4,$78,$C5,$7F,$6D,$AF,$E6,$2D
	.dc.b	$AF,$F7,$CF,$B1,$FD,$36,$35,$F4,$EF,$89,$FF,$1E,$7D,$57,$F8,$EE
	.dc.b	$31,$82,$31,$3A,$CF,$F7,$72,$AD,$D7,$C9,$EC,$FE,$D3,$5F,$8F,$3E
	.dc.b	$4C,$BE,$2F,$FC,$60,$DB,$FF,$17,$36,$7C,$53,$AF,$8C,$95,$A6,$CF
	.dc.b	$9F,$D0,$FF,$2A,$7E,$41,$68,$D2,$4F,$F4,$6E,$2F,$BA,$4C,$BC,$F5
	.dc.b	$EB,$F8,$FF,$E0,$B9,$9A,$48,$FA,$F7,$F8,$53,$DC,$43,$0F,$C5,$F9
	.dc.b	$1F,$79,$1F,$7B,$F0,$BB,$3F,$CD,$9F,$63,$F4,$5B,$5F,$3B,$0F,$F7
	.dc.b	$14,$E6,$CF,$D7,$F9,$4F,$F3,$A6,$16,$4D,$25,$FC,$1C,$BE,$AB,$E9
	.dc.b	$3C,$F8,$3E,$57,$EA,$92,$FE,$AB,$F6,$3E,$23,$FD,$05,$F1,$DF,$87
	.dc.b	$6D,$F9,$DA,$FF,$9E,$6D,$18,$8F,$8D,$B7,$D1,$F6,$BF,$C3,$31,$EF
	.dc.b	$F9,$DF,$F4,$73,$FE,$D3,$CB,$F1,$D1,$87,$47,$FA,$16,$3B,$46,$5F
	.dc.b	$07,$D7,$FF,$A5,$FF,$10,$E5,$92,$31,$7F,$35,$B5,$C2,$15,$FE,$E1
	.dc.b	$C7,$CD,$3F,$9D,$C7,$F2,$FF,$3D,$FE,$A3,$FC,$C3,$AB,$70,$FB,$0D
	.dc.b	$AF,$84,$D7,$FD,$55,$93,$E6,$FF,$D0,$D5,$F7,$DB,$3F,$25,$AF,$06
	.dc.b	$C7,$EC,$7E,$5E,$D1,$A4,$8F,$F4,$81,$F8,$1C,$7E,$7D,$FE,$D3,$E8
	.dc.b	$B8,$92,$77,$1F,$66,$3C,$2F,$E1,$F3,$F8,$6F,$33,$E2,$1F,$E3,$0F
	.dc.b	$92,$FB,$FF,$22,$8F,$F2,$17,$91,$FE,$02,$9C,$C6,$0D,$38,$F8,$2F
	.dc.b	$ED,$D1,$8E,$35,$F3,$9C,$CA,$B8,$5F,$5B,$DC,$F5,$34,$E2,$B8,$D7
	.dc.b	$09,$FF,$40,$D7,$FB,$EC,$8F,$FB,$BD,$AF,$FB,$B9,$AF,$03,$F6,$D8
	.dc.b	$FF,$B9,$7E,$5B,$F1,$ED,$CF,$22,$32,$D3,$B4,$C3,$B1,$4F,$C3,$EB
	.dc.b	$CD,$FD,$63,$5C,$C7,$37,$2D,$7F,$10,$D3,$F9,$C6,$FC,$5A,$67,$7F
	.dc.b	$CA,$37,$83,$FE,$00,$7E,$5F,$F9,$E8,$34,$8F,$9E,$FF,$D2,$B9,$FC
	.dc.b	$66,$9F,$AF,$87,$E2,$FF,$B4,$E5,$46,$0D,$46,$5C,$4F,$EF,$9F,$F1
	.dc.b	$91,$FC,$0A,$BF,$7B,$CD,$C9,$B1,$A9,$F9,$64,$3E,$C1,$5F,$EB,$5C
	.dc.b	$3E,$2F,$C4,$BF,$8E,$41,$A4,$9F,$E6,$9D,$3D,$DB,$3F,$94,$7F,$7B
	.dc.b	$19,$7F,$91,$35,$F0,$90,$F8,$A7,$8F,$57,$67,$C4,$F9,$2F,$E7,$84
	.dc.b	$9F,$E8,$91,$97,$FA,$37,$7F,$CA,$FA,$CF,$E8,$C4,$8C,$E2,$8F,$EF
	.dc.b	$DD,$FE,$38,$FE,$AC,$7A,$4B,$AE,$9F,$65,$79,$8F,$47,$57,$F9,$B7
	.dc.b	$AF,$2D,$AF,$F0,$57,$83,$FA,$3A,$FE,$B4,$D7,$FA,$23,$5F,$E4,$6D
	.dc.b	$7B,$12,$7A,$3F,$A3,$BF,$F0,$87,$87,$F8,$8F,$C4,$BF,$C0,$04,$9F
	.dc.b	$ED,$11,$97,$FB,$57,$8B,$EF,$E0,$34,$97,$F0,$63,$FC,$29,$97,$F8
	.dc.b	$4A,$31,$E3,$71,$3E,$D5,$F9,$7C,$46,$BF,$09,$B7,$FC,$1A,$FF,$C4
	.dc.b	$DE,$47,$E7,$BE,$AF,$FC,$57,$06,$93,$0F,$F8,$01,$E5,$7E,$93,$E7
	.dc.b	$7F,$C7,$14,$69,$27,$FC,$67,$E6,$BC,$EF,$D5,$FD,$57,$F9,$16,$CD
	.dc.b	$25,$76,$8F,$E2,$8D,$F1,$3B,$91,$EB,$E8,$FE,$DB,$F0,$7F,$F5,$C5
	.dc.b	$7F,$97,$5B,$8A,$75,$5C,$29,$5F,$C0,$F9,$BE,$DC,$FE,$CC,$FC,$64
	.dc.b	$AD,$EA,$72,$9F,$F5,$FB,$5F,$E0,$F1,$A9,$EB,$1F,$E7,$CB,$FF,$42
	.dc.b	$FD,$04,$8F,$F4,$2C,$A3,$11,$FE,$8F,$8F,$82,$DC,$BD,$7E,$BF,$E5
	.dc.b	$67,$C8,$DC,$1E,$CF,$F1,$8F,$D2,$7F,$D4,$04,$C3,$E9,$C6,$5F,$F7
	.dc.b	$AF,$CC,$7F,$BD,$08,$E0,$FF,$59,$E0,$F9,$FF,$97,$EF,$7F,$24,$6C
	.dc.b	$66,$77,$1F,$2F,$A3,$F1,$F1,$FC,$C7,$83,$E6,$BF,$EC,$7A,$E0,$FF
	.dc.b	$BA,$3D,$FE,$8F,$E2,$1A,$E6,$36,$FC,$EF,$80,$FE,$6B,$F1,$FF,$ED
	.dc.b	$79,$34,$93,$F9,$2F,$82,$FE,$19,$BF,$FB,$8E,$BF,$86,$37,$B3,$F0
	.dc.b	$F3,$A7,$C3,$5F,$71,$B5,$F8,$36,$D7,$F4,$0D,$3F,$DF,$53,$F6,$2B
	.dc.b	$F0,$7F,$07,$F5,$49,$19,$3F,$19,$F0,$9F,$D3,$BD,$4F,$CB,$B6,$BF
	.dc.b	$52,$D7,$FC,$31,$3C,$0F,$F1,$63,$FE,$1B,$3F,$C3,$9F,$0A,$F8,$6F
	.dc.b	$EB,$3F,$8B,$66,$73,$18,$71,$3B,$A5,$6E,$53,$98,$FD,$5F,$C4,$F1
	.dc.b	$9F,$1E,$88,$C4,$7E,$89,$B7,$FA,$96,$CB,$E2,$FF,$B6,$FE,$37,$FB
	.dc.b	$66,$DF,$F0,$86,$BD,$9A,$FE,$BB,$E2,$BF,$BA,$FD,$F7,$EF,$9B,$1F
	.dc.b	$C4,$1A,$F9,$9A,$FE,$DB,$E4,$3F,$BE,$FE,$95,$FC,$19,$B1,$FC,$61
	.dc.b	$AF,$B9,$AF,$EF,$BA,$7F,$C0,$BD,$1F,$C3,$9B,$1F,$C8,$1A,$FC,$66
	.dc.b	$AF,$95,$7C,$A2,$FF,$B9,$46,$1B,$19,$DF,$FB,$D4,$E5,$FE,$79,$E6
	.dc.b	$4E,$8F,$D8,$3A,$EA,$7C,$AF,$46,$D6,$63,$07,$E9,$3E,$12,$47,$E9
	.dc.b	$1B,$7D,$09,$9F,$F3,$CF,$96,$FD,$59,$2B,$F5,$6F,$F5,$2F,$97,$7A
	.dc.b	$94,$93,$FD,$E2,$32,$FA,$6B,$F8,$36,$FF,$C2,$1F,$F1,$17,$37,$F8
	.dc.b	$CA,$7F,$6E,$7E,$E8,$87,$F8,$17,$E7,$9F,$33,$FE,$38,$FD,$27,$FA
	.dc.b	$1C,$62,$51,$89,$D5,$FE,$A6,$55,$BA,$F7,$CF,$F4,$3F,$13,$F8,$CD
	.dc.b	$6B,$E7,$3F,$C8,$BE,$9F,$66,$D7,$3C,$11,$88,$FD,$CC,$62,$95,$E6
	.dc.b	$F0,$3F,$86,$CF,$EC,$FC,$4C,$57,$F2,$03,$CB,$3F,$C5,$2F,$FB,$F7
	.dc.b	$2F,$C9,$7F,$C9,$5C,$CF,$9E,$78,$7F,$DD,$E7,$F8,$34,$E6,$31,$1F
	.dc.b	$EA,$38,$C6,$1F,$DE,$E3,$75,$CA,$31,$1E,$AF,$9F,$FE,$A5,$99,$A4
	.dc.b	$9F,$E4,$53,$E6,$E6,$F9,$6F,$F9,$B2,$9E,$57,$C0,$25,$7B,$E3,$2D
	.dc.b	$F8,$1E,$E4,$7F,$95,$42,$3F,$C6,$DC,$3F,$D3,$AF,$C0,$7F,$8D,$65
	.dc.b	$F4,$3F,$E4,$D6,$DF,$E2,$63,$06,$7A,$FF,$C2,$F5,$F6,$DA,$FD,$A6
	.dc.b	$BE,$B6,$AB,$FF,$29,$7A,$3F,$CA,$DB,$1F,$CC,$DA,$FF,$39,$6B,$F1
	.dc.b	$7E,$8B,$EA,$B7,$0F,$A2,$DF,$63,$F3,$B3,$AF,$98,$7D,$1F,$E7,$F9
	.dc.b	$BF,$A0,$46,$3A,$BF,$AD,$B7,$1F,$ED,$6D,$9F,$37,$67,$F9,$21,$BC
	.dc.b	$FF,$DC,$33,$F6,$4E,$5F,$53,$5F,$EA,$54,$FA,$4F,$CB,$37,$0E,$4D
	.dc.b	$C3,$FA,$73,$6B,$9E,$1D,$F5,$FF,$3D,$57,$88,$93,$EC,$9F,$F2,$FD
	.dc.b	$7F,$97,$AF,$FD,$06,$DE,$4E,$B3,$A2,$4E,$9B,$1A,$8C,$11,$C1,$2B
	.dc.b	$C1,$19,$6C,$65,$F8,$33,$D1,$C9,$0C,$CF,$16,$CF,$14,$2B,$D4,$F1
	.dc.b	$32,$E1,$E2,$64,$E1,$5E,$17,$FC,$09,$F2,$2F,$8A,$7C,$97,$F7,$D9
	.dc.b	$4E,$63,$0D,$8F,$F0,$13,$5F,$C5,$FD,$A7,$FB,$91,$FB,$2D,$73,$97
	.dc.b	$5C,$53,$99,$FC,$2B,$5F,$5B,$5F,$80,$D8,$E3,$3C,$E3,$04,$70,$49
	.dc.b	$F1,$B5,$F1,$98,$21,$82,$1E,$47,$83,$4D,$94,$D9,$E8,$76,$D3,$65
	.dc.b	$36,$7B,$1D,$B4,$D9,$FA,$CC,$7E,$AB,$FA,$2C,$3F,$62,$47,$F4,$A7
	.dc.b	$ED,$08,$60,$FF,$21,$7C,$E3,$F7,$DF,$E3,$B1,$FE,$41,$8C,$73,$FD
	.dc.b	$16,$E1,$FE,$3F,$6C,$FF,$AA,$BE,$17,$FA,$54,$BE,$3F,$1D,$C8,$C5
	.dc.b	$7F,$69,$AD,$D6,$A7,$33,$FC,$77,$93,$FD,$A6,$FA,$0C,$69,$F9,$3E
	.dc.b	$3F,$F7,$C4,$78,$7E,$CB,$B7,$05,$C3,$F4,$93,$DE,$72,$D8,$CB,$FE
	.dc.b	$66,$C9,$F6,$7F,$8F,$94,$61,$FF,$2F,$18,$8D,$46,$67,$CC,$7E,$D3
	.dc.b	$0F,$2B,$F3,$CF,$49,$F1,$FE,$D3,$E2,$7D,$4D,$C0,$D2,$4F,$F9,$CF
	.dc.b	$E7,$D9,$7C,$BE,$3F,$F3,$84,$62,$FF,$07,$5F,$E7,$D8,$C4,$EA,$B8
	.dc.b	$4F,$8C,$F3,$6D,$F3,$9C,$1F,$8D,$38,$5F,$91,$F1,$4F,$A1,$B1,$97
	.dc.b	$C1,$19,$7F,$D1,$DF,$17,$FE,$03,$C7,$FD,$0C,$D8,$FF,$47,$35,$DF
	.dc.b	$8D,$C5,$F2,$F0,$7D,$FC,$7C,$D1,$F2,$8F,$56,$3F,$5B,$ED,$7E,$99
	.dc.b	$1F,$AF,$BA,$3E,$E7,$E7,$BF,$EA,$4E,$AF,$F8,$02,$DB,$FF,$1E,$35
	.dc.b	$FF,$54,$B5,$F1,$9F,$78,$F8,$C7,$89,$FE,$99,$F3,$DD,$32,$FD,$03
	.dc.b	$71,$34,$F0,$F5,$7B,$E1,$5F,$C6,$B6,$CC,$7C,$DC,$3F,$55,$39,$7E
	.dc.b	$BF,$1F,$E3,$9F,$88,$43,$CE,$6B,$5F,$92,$E0,$F6,$FD,$4F,$86,$8D
	.dc.b	$25,$F8,$67,$DE,$75,$29,$CC,$FE,$A7,$C0,$BF,$E6,$CD,$9F,$E9,$4D
	.dc.b	$9F,$C1,$CF,$D4,$FB,$7F,$CA,$C7,$D0,$6B,$F0,$1B,$5F,$19,$CC,$78
	.dc.b	$89,$8F,$C3,$7C,$9D,$4F,$D8,$5F,$1B,$FC,$6D,$B8,$7F,$1B,$6D,$7F
	.dc.b	$21,$6B,$F8,$4F,$B9,$FE,$34,$7E,$20,$AE,$65,$19,$6C,$7E,$33,$EE
	.dc.b	$BF,$91,$71,$FE,$41,$B1,$99,$FE,$E0,$D9,$FC,$AF,$DD,$FF,$29,$FC
	.dc.b	$F6,$D1,$A4,$AF,$CC,$F3,$FF,$2F,$FD,$6B,$F3,$AD,$8C,$D2,$72,$D8
	.dc.b	$FD,$0F,$D9,$7C,$EE,$43,$49,$3F,$C3,$A7,$F0,$07,$D0,$E2,$FC,$6B
	.dc.b	$F8,$48,$FC,$DF,$91,$CC,$8F,$13,$25,$B7,$DF,$7E,$D8,$91,$FC,$C2
	.dc.b	$32,$FD,$7E,$1F,$B4,$47,$06,$FE,$FF,$FA,$87,$DA,$7E,$F9,$B8,$FE
	.dc.b	$F9,$AF,$EF,$9A,$FE,$BB,$CE,$FE,$B1,$FA,$9F,$F0,$66,$C7,$F0,$66
	.dc.b	$BF,$C1,$9A,$FE,$DB,$D2,$FE,$9E,$FF,$0A,$6B,$9C,$3F,$A8,$B7,$C7
	.dc.b	$FE,$19,$D3,$F9,$98,$F9,$9E,$07,$B1,$FC,$A6,$9E,$F6,$DF,$C5,$E5
	.dc.b	$FE,$DD,$F9,$77,$E1,$89,$3F,$C7,$23,$2F,$F1,$EF,$2F,$FB,$A7,$DA
	.dc.b	$7F,$22,$6C,$66,$D3,$96,$C7,$F2,$6E,$2F,$AA,$FF,$2A,$6B,$FC,$AB
	.dc.b	$D6,$FC,$51,$27,$EF,$3F,$8D,$F5,$FF,$C0,$9F,$5D,$FC,$C9,$B5,$FC
	.dc.b	$C9,$AF,$F3,$26,$BF,$90,$DB,$E3,$13,$9F,$F8,$EE,$1F,$CA,$BD,$9F
	.dc.b	$CD,$11,$FC,$73,$F9,$DC,$BC,$F7,$FC,$F9,$AF,$F2,$EC,$BF,$C3,$1E
	.dc.b	$0F,$F4,$66,$C7,$B6,$D3,$96,$C7,$F4,$AF,$6F,$F5,$44,$8F,$CF,$3F
	.dc.b	$AD,$F7,$7F,$C4,$9F,$65,$FD,$49,$B5,$FD,$49,$AF,$F5,$26,$BF,$07
	.dc.b	$3F,$D7,$E9,$F0,$79,$3F,$4D,$EC,$7F,$5E,$26,$58,$3F,$C3,$3F,$B0
	.dc.b	$FF,$8C,$D7,$D9,$7F,$A2,$B1,$FF,$45,$C2,$75,$39,$9F,$F1,$D4,$61
	.dc.b	$E1,$3B,$3F,$C7,$D3,$B8,$7E,$7B,$FE,$44,$F4,$B0,$D8,$E8,$D7,$C4
	.dc.b	$D7,$DD,$8F,$F8,$04,$8F,$8E,$FC,$DF,$0D,$FD,$A7,$13,$49,$1F,$D6
	.dc.b	$73,$F8,$D5,$FA,$ED,$3F,$0F,$6F,$87,$FF,$20,$36,$FE,$3C,$60,$CC
	.dc.b	$FF,$96,$FF,$58,$F1,$57,$F5,$D6,$C6,$6F,$E2,$CE,$5B,$F1,$7F,$E6
	.dc.b	$3F,$C8,$38,$B6,$33,$8A,$72,$D8,$FE,$F5,$A7,$C2,$7F,$C4,$36,$F8
	.dc.b	$FF,$49,$2B,$EF,$BE,$CF,$91,$FF,$09,$D7,$91,$AF,$E0,$29,$18,$8F
	.dc.b	$BE,$D7,$F8,$99,$F2,$6F,$F1,$7F,$27,$FE,$BD,$C3,$DE,$D9,$E8,$D7
	.dc.b	$F1,$2D,$8F,$C7,$7B,$3F,$E8,$5F,$C0,$BF,$D8,$91,$80,$E3,$14,$EE
	.dc.b	$7E,$6B,$6F,$FC,$29,$C1,$F7,$9F,$8B,$F3,$19,$7F,$32,$DA,$FC,$F3
	.dc.b	$6C,$DF,$84,$FF,$3A,$CF,$DA,$6D,$7C,$76,$BF,$A4,$6C,$7E,$A7,$E6
	.dc.b	$6D,$9A,$48,$FF,$3F,$9F,$72,$BE,$61,$F8,$CF,$9A,$E2,$48,$DC,$7D
	.dc.b	$78,$F0,$BF,$50,$E4,$43,$E1,$F1,$7F,$74,$F8,$6F,$F7,$B7,$2B,$2F
	.dc.b	$F7,$FF,$AB,$FE,$CA,$8C,$58,$FF,$7F,$C6,$67,$7D,$39,$AD,$D6,$A3
	.dc.b	$13,$A8,$FD,$63,$67,$77,$E2,$8F,$F5,$D9,$FC,$1F,$6F,$F2,$47,$17
	.dc.b	$4B,$FF,$40,$CE,$72,$4E,$67,$FD,$C5,$39,$8C,$57,$28,$C1,$F2,$F4
	.dc.b	$7C,$FF,$E2,$F5,$FF,$6A,$36,$32,$7E,$E2,$32,$FF,$21,$AE,$98,$5C
	.dc.b	$7C,$78,$76,$8C,$63,$E9,$BF,$C1,$1D,$23,$F5,$DB,$74,$6C,$65,$FE
	.dc.b	$5B,$19,$7F,$19,$19,$3F,$9A,$56,$5A,$67,$7F,$B8,$DE,$0F,$F7,$87
	.dc.b	$E2,$1F,$A9,$6C,$7F,$14,$6D,$FF,$01,$6B,$F4,$0F,$5D,$FB,$5F,$AE
	.dc.b	$7E,$C2,$55,$FB,$5E,$2E,$31,$FB,$98,$C3,$A3,$F8,$29,$E6,$9F,$E0
	.dc.b	$AD,$7F,$82,$B4,$FE,$A5,$C8,$75,$1F,$80,$3C,$A4,$3C,$A7,$F1,$BE
	.dc.b	$5E,$1F,$61,$FC,$D9,$B1,$FC,$86,$53,$96,$C7,$F3,$8E,$7F,$06,$5F
	.dc.b	$CF,$9B,$19,$9F,$BD,$A7,$F3,$17,$F2,$3B,$FE,$12,$7C,$87,$CB,$3C
	.dc.b	$A7,$F6,$AE,$9F,$67,$E1,$DF,$CF,$89,$1F,$D0,$9F,$E9,$74,$FD,$D6
	.dc.b	$92,$7F,$A6,$DF,$C5,$8F,$EA,$7C,$FE,$69,$FE,$78,$D7,$FB,$CB,$6B
	.dc.b	$FB,$F3,$98,$F9,$87,$98,$A7,$A4,$FF,$00,$F6,$FE,$A1,$FE,$BA,$D7
	.dc.b	$FA,$53,$5D,$CF,$EC,$1F,$86,$74,$27,$2F,$EB,$1F,$F0,$0F,$C5,$7D
	.dc.b	$FF,$8C,$FF,$03,$B6,$3F,$C0,$ED,$77,$49,$FE,$E6,$FE,$1F,$EA,$EA
	.dc.b	$69,$27,$FD,$FF,$E1,$FC,$3F,$60,$D2,$57,$DC,$EA,$9C,$7F,$A3,$BF
	.dc.b	$89,$3F,$EF,$8F,$17,$F2,$DF,$4F,$FE,$3B,$83,$49,$3F,$E4,$58,$F4
	.dc.b	$37,$3F,$F0,$E6,$4F,$23,$DD,$3F,$F0,$01,$B4,$97,$95,$FA,$5F,$91
	.dc.b	$FF,$29,$12,$FF,$CA,$CF,$F9,$D6,$BF,$8A,$57,$FA,$AB,$3C,$12,$BF
	.dc.b	$C7,$70,$F3,$7B,$B6,$BC,$F3,$FE,$46,$8C,$65,$E8,$AF,$3C,$FD,$CE
	.dc.b	$5F,$D1,$3F,$8B,$E2,$FB,$CF,$EE,$ED,$E7,$7E,$F7,$EC,$BF,$D0,$59
	.dc.b	$1A,$4A,$FF,$38,$65,$FE,$6F,$23,$83,$FD,$29,$3E,$EF,$3F,$EE,$11
	.dc.b	$F8,$3E,$97,$F0,$BF,$9F,$FF,$4E,$12,$BF,$D4,$0F,$FB,$BB,$F7,$09
	.dc.b	$1F,$EA,$9A,$46,$23,$FB,$37,$A4,$F8,$57,$B1,$FA,$F8,$C3,$63,$37
	.dc.b	$FD,$76,$72,$FE,$32,$BF,$4B,$BF,$C6,$7F,$63,$1F,$7F,$DA,$FE,$45
	.dc.b	$F5,$1F,$86,$6C,$FD,$86,$BF,$EC,$F9,$4F,$FB,$65,$7E,$AD,$49,$3F
	.dc.b	$ED,$B8,$CB,$FE,$E5,$5F,$35,$FE,$E9,$6C,$66,$30,$7F,$B4,$B9,$A4
	.dc.b	$7F,$A2,$F2,$F6,$C6,$23,$FD,$5D,$5F,$54,$91,$F9,$18,$FC,$67,$C1
	.dc.b	$7E,$40,$93,$F9,$08,$FC,$44,$7E,$3B,$3F,$8C,$4E,$6F,$C7,$CF,$ED
	.dc.b	$39,$9F,$07,$F9,$A2,$61,$CC,$FD,$53,$F9,$EF,$85,$FE,$9D,$FB,$87
	.dc.b	$EC,$9B,$7F,$A4,$6B,$FA,$16,$BF,$9B,$F7,$FF,$AB,$7E,$73,$FB,$A6
	.dc.b	$C7,$E2,$63,$18,$1F,$DF,$FC,$4F,$F5,$DF,$0A,$C9,$A4,$AF,$DA,$FC
	.dc.b	$57,$F6,$3F,$A9,$FE,$0E,$D8,$CD,$27,$2D,$8F,$A9,$CF,$F9,$2E,$2F
	.dc.b	$A9,$1F,$ED,$A8,$C3,$AE,$8F,$E8,$9C,$5F,$37,$C5,$FC,$B4,$FF,$5A
	.dc.b	$C3,$F8,$97,$C6,$7F,$06,$23,$F3,$78,$D1,$FC,$13,$E3,$5C,$BF,$D3
	.dc.b	$DB,$5F,$E3,$B6,$E9,$FE,$C9,$6E,$1F,$A4,$AF,$BA,$FF,$0D,$C7,$F8
	.dc.b	$DF,$C7,$7F,$4F,$C3,$F9,$23,$72,$34,$FE,$A9,$5C,$71,$FE,$74,$D3
	.dc.b	$B5,$FB,$9F,$E2,$DF,$21,$CC,$98,$7F,$5B,$8F,$E4,$51,$DE,$7F,$C1
	.dc.b	$B1,$FC,$95,$AF,$F1,$F6,$CF,$F3,$97,$31,$E1,$3F,$93,$63,$F9,$A9
	.dc.b	$FE,$7D,$1F,$85,$3F,$9A,$94,$FE,$64,$53,$F9,$61,$57,$2B,$E4,$BF
	.dc.b	$C0,$0D,$8F,$F2,$B3,$71,$FF,$76,$36,$BE,$1C,$7C,$6D,$BF,$8A,$6B
	.dc.b	$FE,$B2,$BF,$85,$5F,$E8,$78,$C7,$BB,$F6,$7D,$AF,$9F,$18,$9F,$F4
	.dc.b	$61,$FE,$8D,$F9,$47,$CA,$7D,$46,$E1,$D2,$51,$8C,$38,$C7,$1A,$FB
	.dc.b	$78,$FE,$57,$2F,$36,$BF,$86,$C3,$9F,$8F,$C3,$0E,$78,$7F,$56,$9C
	.dc.b	$C7,$46,$BF,$EA,$1C,$BE,$4F,$CA,$FE,$37,$63,$49,$1E,$7C,$FC,$55
	.dc.b	$F2,$0F,$7C,$7D,$2F,$96,$FA,$24,$7E,$8B,$F5,$31,$FE,$73,$3F,$E5
	.dc.b	$FC,$3E,$1F,$89,$F5,$27,$37,$F7,$1D,$D7,$E2,$4F,$F1,$B1,$FD,$22
	.dc.b	$9E,$9F,$F9,$A7,$D2,$FB,$AD,$9F,$2C,$EA,$FF,$C4,$31,$89,$D3,$FD
	.dc.b	$7B,$E6,$1E,$5F,$F5,$D6,$D7,$F5,$B6,$F0,$FD,$6F,$19,$18,$BF,$88
	.dc.b	$7E,$7B,$9F,$FC,$59,$C5,$FE,$B9,$24,$7F,$AA,$9F,$B1,$F1,$EE,$9F
	.dc.b	$F2,$B3,$72,$FF,$61,$37,$2D,$CE,$89,$3A,$6C,$6A,$30,$47,$04,$BF
	.dc.b	$C7,$D7,$56,$D7,$26,$D7,$16,$D7,$17,$F3,$F5,$C5,$B5,$C5,$B5,$C5
	.dc.b	$B5,$C5,$FF,$79,$57,$16,$D7,$16,$D7,$16,$D7,$15,$5C,$5B,$5F,$E7
	.dc.b	$FF,$84,$7C,$33,$E1,$BF,$4D,$5F,$EF,$B3,$F5,$13,$A8,$CB,$63,$EA
	.dc.b	$F9,$5F,$EB,$56,$E5,$CB,$14,$E6,$7E,$5B,$5F,$8A,$D7,$DE,$D7,$D2
	.dc.b	$D8,$E7,$3D,$63,$04,$70,$49,$F1,$90,$F0,$B9,$69,$93,$C6,$E5,$A6
	.dc.b	$4F,$2B,$96,$99,$3C,$EE,$5A,$66,$7D,$2D,$9F,$43,$67,$84,$EA,$3C
	.dc.b	$A7,$F2,$0E,$6F,$ED,$3A,$FD,$78,$C7,$4A,$31,$1F,$CB,$A7,$79,$7F
	.dc.b	$21,$6C,$7F,$2A,$FA,$7F,$F4,$76,$3C,$20,$23,$35,$FC,$A5,$B1,$C7
	.dc.b	$D7,$FF,$7A,$4F,$ED,$BD,$4B,$3F,$A8,$6B,$FA,$26,$BF,$9C,$6A,$C3
	.dc.b	$DB,$C7,$E7,$CF,$D8,$DC,$CD,$3F,$9E,$7E,$66,$FB,$A6,$BA,$C7,$FA
	.dc.b	$3C,$E6,$31,$1F,$D4,$A3,$15,$CA,$B8,$BF,$D3,$A9,$F0,$2F,$A6,$EB
	.dc.b	$F3,$59,$8C,$4A,$31,$3A,$C7,$FA,$0D,$6A,$7D,$51,$0F,$B3,$97,$EC
	.dc.b	$34,$FB,$B1,$FE,$21,$F8,$3F,$D1,$78,$86,$93,$2F,$EB,$87,$9B,$EA
	.dc.b	$FC,$15,$FD,$67,$24,$62,$3E,$2C,$62,$7D,$11,$E6,$9D,$3F,$4F,$7F
	.dc.b	$C6,$3E,$73,$F1,$8E,$1F,$49,$FE,$37,$7C,$E4,$9F,$E1,$78,$7F,$8F
	.dc.b	$A7,$F1,$5C,$5D,$B8,$FF,$23,$AE,$2F,$C7,$5F,$E0,$69,$CB,$D8,$87
	.dc.b	$8B,$5F,$91,$8F,$8E,$38,$61,$F8,$CA,$FC,$BC,$62,$FF,$3C,$FE,$76
	.dc.b	$7F,$23,$F2,$BF,$E2,$3F,$9D,$FF,$03,$36,$7F,$C1,$ED,$7F,$C3,$2D
	.dc.b	$7F,$4F,$BF,$D8,$7C,$A6,$31,$F2,$37,$2F,$E5,$B5,$FE,$31,$FB,$1E
	.dc.b	$FE,$57,$E6,$1B,$99,$A7,$F8,$67,$E9,$5F,$4D,$FC,$95,$B3,$FC,$E1
	.dc.b	$B8,$EE,$7F,$95,$67,$E5,$8F,$90,$7D,$B3,$C8,$6E,$B9,$4F,$52,$3F
	.dc.b	$D1,$3F,$5D,$7C,$87,$F3,$56,$C6,$63,$07,$83,$5F,$83,$7F,$8B,$D7
	.dc.b	$EB,$91,$FF,$1A,$5F,$26,$D7,$2B,$E8,$7F,$54,$FD,$7D,$F4,$FF,$96
	.dc.b	$6C,$7E,$22,$BF,$8C,$C6,$38,$FF,$C8,$0D,$8F,$AA,$78,$CF,$84,$53
	.dc.b	$F8,$E4,$FF,$62,$E3,$FE,$4B,$1F,$95,$21,$FB,$CF,$5B,$F8,$7F,$A3
	.dc.b	$FC,$0D,$B1,$9A,$4E,$5B,$1F,$C1,$3E,$C3,$F8,$BF,$EC,$BF,$C2,$DB
	.dc.b	$19,$94,$E5,$B1,$FC,$33,$EC,$BF,$8F,$FD,$27,$F1,$36,$C6,$65,$39
	.dc.b	$6C,$7F,$9C,$B9,$7F,$94,$7D,$DF,$08,$FE,$83,$18,$9D,$C6,$3C,$15
	.dc.b	$6E,$3F,$2D,$18,$9D,$C7,$F4,$AF,$D0,$5E,$5F,$F3,$4F,$CD,$BF,$CE
	.dc.b	$51,$8A,$46,$27,$55,$FE,$7E,$AD,$4F,$F0,$5F,$C1,$A4,$FD,$66,$BF
	.dc.b	$E8,$CF,$9D,$7C,$9F,$F4,$2F,$C6,$3F,$D0,$11,$8C,$91,$89,$D5,$7F
	.dc.b	$A3,$2B,$53,$9A,$FC,$64,$A7,$97,$36,$49,$1F,$88,$3F,$0A,$7F,$0F
	.dc.b	$29,$FB,$BF,$CB,$3F,$D1,$9F,$13,$FE,$9A,$E0,$46,$27,$EA,$35,$FA
	.dc.b	$6D,$7F,$9C,$FC,$17,$F5,$EF,$D4,$FF,$A9,$36,$3F,$A7,$B5,$FE,$84
	.dc.b	$D7,$F4,$BD,$7F,$D9,$BF,$08,$FD,$2B,$63,$FD,$5B,$D9,$FA,$C9,$CE
	.dc.b	$A7,$F4,$FE,$17,$F0,$BA,$FE,$12,$7E,$B3,$E9,$3F,$8C,$57,$F2,$E6
	.dc.b	$FA,$3F,$EB,$16,$CF,$FA,$E9,$AF,$F5,$7E,$AF,$EF,$3F,$B1,$FE,$5C
	.dc.b	$91,$F9,$B7,$FB,$86,$7F,$DF,$BE,$A6,$E2,$69,$2B,$FD,$8F,$7F,$E9
	.dc.b	$DE,$2F,$F6,$13,$63,$3E,$E7,$F6,$08,$C4,$7F,$B5,$BE,$C1,$C9,$FE
	.dc.b	$16,$6E,$3F,$D2,$9B,$AF,$FB,$35,$B5,$F2,$B3,$FF,$09,$F9,$9F,$28
	.dc.b	$91,$9B,$47,$CC,$F2,$3F,$93,$F0,$7C,$DA,$FE,$28,$7F,$BA,$38,$3E
	.dc.b	$6F,$A7,$F2,$F9,$FF,$C2,$F3,$F2,$DA,$ED,$AF,$F8,$F3,$6F,$F5,$F7
	.dc.b	$1F,$C2,$E0,$FC,$04,$FF,$BD,$27,$2F,$C4,$D3,$F0,$0F,$57,$FD,$16
	.dc.b	$DF,$5B,$C5,$18,$8F,$03,$5F,$F2,$5F,$CB,$FC,$8F,$CB,$BF,$DE,$4D
	.dc.b	$8F,$F7,$DB,$5F,$F2,$88,$70,$FF,$95,$34,$FC,$6F,$DA,$78,$09,$1E
	.dc.b	$57,$E5,$79,$BF,$D3,$63,$FB,$9E,$1F,$E3,$AF,$BD,$FF,$47,$34,$FF
	.dc.b	$14,$FA,$C8,$CB,$63,$FC,$5B,$FA,$EF,$E7,$3E,$AB,$FC,$88,$DB,$FF
	.dc.b	$7E,$B5,$FF,$34,$CA,$31,$1F,$E5,$45,$EB,$FF,$95,$BF,$02,$E6,$D9
	.dc.b	$E6,$D7,$9C,$F0,$A5,$7B,$A6,$1E,$2F,$C0,$7D,$B4,$69,$2B,$E2,$4F
	.dc.b	$F8,$08,$FE,$C7,$CD,$FB,$BE,$17,$E5,$9F,$E4,$59,$FF,$70,$B5,$FF
	.dc.b	$6D,$B5,$FF,$6A,$B5,$F7,$7A,$DF,$E7,$2F,$CA,$FE,$3B,$63,$E3,$B5
	.dc.b	$F6,$B5,$F8,$DF,$82,$7A,$89,$8F,$D2,$7F,$C1,$9C,$3E,$A9,$FE,$22
	.dc.b	$7A,$CF,$8F,$AF,$C2,$25,$7D,$A7,$E2,$FE,$0D,$FE,$86,$F8,$3F,$F6
	.dc.b	$A4,$FD,$78,$C3,$F7,$1B,$5F,$49,$AF,$80,$87,$ED,$8B,$F2,$4F,$8C
	.dc.b	$FF,$72,$57,$E7,$25,$39,$8F,$C3,$C7,$E8,$23,$0F,$FB,$95,$B1,$E3
	.dc.b	$3F,$DE,$E5,$F8,$7F,$4F,$13,$0F,$F5,$0C,$61,$B3,$F7,$BC,$0F,$E1
	.dc.b	$A4,$D3,$F3,$B3,$FE,$A8,$E4,$E1,$5F,$8F,$CB,$E0,$35,$FC,$D4,$EE
	.dc.b	$3F,$43,$F8,$3F,$BA,$7F,$2E,$D7,$F4,$0D,$8F,$D2,$37,$23,$49,$1F
	.dc.b	$12,$78,$B5,$F9,$C4,$E4,$FE,$4C,$7F,$53,$8E,$A7,$12,$C7,$C6,$9E
	.dc.b	$AF,$FB,$18,$FC,$41,$F6,$A7,$9E,$5F,$16,$3F,$5A,$68,$7C,$43,$4F
	.dc.b	$F7,$97,$D4,$7F,$87,$A7,$FD,$D3,$39,$F1,$FF,$DD,$73,$99,$E3,$5C
	.dc.b	$3D,$6F,$E3,$F7,$B9,$56,$E3,$0D,$AE,$13,$D9,$AF,$4B,$EF,$7F,$EA
	.dc.b	$53,$FD,$D0,$48,$FF,$67,$B5,$D1,$82,$1F,$6B,$3F,$F1,$B3,$63,$E5
	.dc.b	$46,$23,$EA,$C7,$CB,$7F,$D3,$CF,$56,$9D,$0D,$C7,$E2,$0E,$47,$D8
	.dc.b	$8E,$A4,$74,$D8,$F1,$47,$E8,$FE,$E3,$F2,$E4,$70,$77,$8F,$F4,$84
	.dc.b	$F8,$DA,$7E,$AF,$E6,$BF,$3E,$49,$C1,$FB,$7E,$2F,$DB,$57,$EC,$F8
	.dc.b	$D3,$F0,$6F,$F7,$71,$89,$D4,$E7,$66,$CE,$3F,$C1,$61,$F1,$0F,$A2
	.dc.b	$43,$F5,$B5,$FD,$45,$FD,$8F,$77,$ED,$09,$1F,$D4,$39,$BC,$13,$63
	.dc.b	$A9,$FD,$46,$33,$1F,$C9,$23,$C2,$FC,$23,$C0,$53,$C8,$7F,$32,$79
	.dc.b	$F1,$B9,$F9,$CE,$73,$FE,$C7,$18,$AF,$EC,$1D,$1F,$1B,$F2,$BF,$E8
	.dc.b	$30,$69,$2F,$FA,$5F,$C4,$74,$ED,$FE,$A2,$DE,$D3,$4F,$0C,$F9,$C9
	.dc.b	$C1,$F3,$B7,$F6,$B7,$4F,$EE,$FD,$DE,$5D,$BF,$B2,$69,$EC,$8C,$1D
	.dc.b	$2B,$86,$1F,$89,$6B,$F5,$67,$FB,$E7,$67,$E0,$BE,$A3,$FB,$CC,$1A
	.dc.b	$49,$FF,$02,$F9,$3F,$88,$FA,$CF,$F0,$4B,$63,$34,$9C,$B6,$3F,$C4
	.dc.b	$6F,$F8,$5F,$3F,$E9,$4D,$8C,$F8,$7F,$E1,$79,$CB,$63,$D5,$CD,$F9
	.dc.b	$19,$FD,$16,$7E,$C9,$FF,$16,$63,$F9,$CF,$D7,$7F,$C7,$64,$8F,$F2
	.dc.b	$03,$FE,$38,$F4,$3F,$43,$F9,$25,$CC,$D2,$57,$F9,$1F,$E5,$FF,$C8
	.dc.b	$A4,$70,$7F,$93,$7E,$FF,$FC,$A5,$46,$92,$3F,$CA,$1F,$59,$FE,$4E
	.dc.b	$23,$83,$DB,$C9,$F8,$0D,$BD,$D1,$FC,$07,$D8,$7A,$9F,$B8,$EF,$FF
	.dc.b	$4D,$92,$7F,$CE,$4F,$F9,$E7,$5F,$DE,$FC,$AD,$D0,$D2,$57,$F9,$97
	.dc.b	$E7,$7D,$24,$70,$7D,$4C,$3F,$83,$FD,$57,$FA,$58,$91,$9B,$47,$FB
	.dc.b	$D7,$F7,$39,$3F,$EF,$96,$BF,$E6,$CC,$3F,$CD,$E4,$7F,$CE,$8F,$DC
	.dc.b	$D3,$E0,$C7,$EF,$39,$3E,$F2,$D7,$E1,$E3,$F3,$E3,$FD,$25,$C1,$FE
	.dc.b	$BE,$23,$FE,$CA,$87,$EA,$5F,$EC,$E6,$C6,$63,$07,$FB,$6F,$E0,$3F
	.dc.b	$CC,$3A,$D9,$FE,$4E,$DA,$FA,$71,$FE,$DE,$32,$7F,$89,$7E,$39,$F2
	.dc.b	$9F,$BE,$6C,$FF,$18,$6F,$4F,$F8,$AE,$91,$87,$FD,$F5,$F0,$BF,$CD
	.dc.b	$B8,$BF,$DF,$44,$8C,$DA,$3F,$DF,$DF,$8B,$49,$CB,$5F,$F4,$9D,$F8
	.dc.b	$BF,$09,$F0,$36,$3D,$19,$7F,$44,$9C,$B6,$3F,$C8,$1F,$14,$CF,$FA
	.dc.b	$4F,$E0,$FE,$98,$C6,$68,$C4,$EA,$B8,$B5,$0E,$E7,$C6,$3E,$37,$FA
	.dc.b	$AF,$D5,$FF,$93,$E7,$31,$8A,$FF,$56,$C6,$31,$56,$EF,$94,$62,$BF
	.dc.b	$86,$FC,$07,$17,$FD,$ED,$19,$6D,$7F,$10,$9D,$47,$F6,$6F,$0D,$F2
	.dc.b	$3F,$DA,$3F,$7C,$FE,$2A,$D9,$FE,$32,$D7,$FD,$53,$3C,$30,$57,$F1
	.dc.b	$CF,$94,$FE,$E3,$F4,$96,$4D,$25,$FD,$48,$FC,$47,$0F,$E2,$C8,$FF
	.dc.b	$07,$F0,$E7,$07,$EA,$5F,$EC,$BD,$9F,$A6,$E7,$FC,$73,$70,$FF,$7C
	.dc.b	$46,$3A,$7F,$1D,$1F,$E7,$DF,$97,$7C,$BF,$F5,$66,$E5,$FD,$C1,$B9
	.dc.b	$7E,$65,$B5,$FC,$6A,$3C,$46,$6B,$91,$2B,$93,$FD,$8B,$E6,$1E,$A5
	.dc.b	$24,$FF,$7E,$8C,$BE,$EB,$EC,$7F,$89,$63,$37,$FA,$B6,$C7,$FA,$5D
	.dc.b	$AF,$FB,$D1,$AF,$E4,$BE,$69,$F3,$5F,$E2,$4F,$B1,$FD,$A4,$63,$04
	.dc.b	$62,$75,$5F,$EA,$3A,$D4,$FC,$08,$F2,$9F,$9E,$11,$CF,$CA,$E5,$7F
	.dc.b	$EA,$D9,$E5,$7F,$CF,$5B,$19,$7F,$73,$C3,$FC,$F1,$FC,$27,$8F,$FB
	.dc.b	$69,$FE,$7B,$18,$3C,$27,$F8,$88,$FF,$07,$78,$1F,$C4,$B8,$78,$12
	.dc.b	$3F,$1C,$FF,$13,$F4,$7F,$C7,$F5,$FE,$2E,$6E,$D9,$F2,$3F,$C3,$31
	.dc.b	$8B,$EE,$7F,$86,$A3,$F9,$CB,$FA,$9F,$53,$F1,$AF,$E9,$1A,$E6,$FF
	.dc.b	$8D,$3F,$E0,$28,$CA,$A7,$CE,$78,$09,$5E,$57,$FC,$41,$7E,$0A,$F1
	.dc.b	$73,$FE,$1D,$FD,$9F,$CE,$F8,$2D,$D8,$8C,$36,$3C,$73,$C1,$AF,$F1
	.dc.b	$8F,$9E,$FE,$2F,$0F,$43,$F9,$D4,$61,$BE,$17,$F9,$B9,$B3,$BA,$FF
	.dc.b	$75,$79,$BF,$E3,$F9,$FE,$84,$D8,$CE,$7F,$E6,$F9,$CB,$63,$FD,$E5
	.dc.b	$EC,$BE,$7F,$F0,$91,$F8,$F8,$C6,$1F,$84,$8C,$4E,$9F,$E6,$33,$AE
	.dc.b	$4F,$C5,$12,$3F,$14,$7E,$42,$B9,$3F,$CE,$31,$FF,$2B,$4F,$93,$6C
	.dc.b	$CA,$72,$FA,$A3,$36,$F8,$8F,$F3,$0B,$70,$FF,$3D,$B7,$8F,$FB,$23
	.dc.b	$6F,$DF,$5C,$09,$5C,$1B,$5C,$1B,$3C,$27,$44,$9D,$36,$35,$18,$23
	.dc.b	$82,$1F,$36,$7A,$B6,$79,$36,$78,$B4,$E2,$7D,$09,$E2,$D9,$E2,$D9
	.dc.b	$E2,$D3,$88,$AF,$88,$F5,$21,$FE,$EB,$F8,$3F,$F0,$63,$FE,$EB,$3F
	.dc.b	$C1,$DF,$0D,$FE,$F5,$7F,$8A,$19,$3F,$17,$D9,$FE,$E1,$7F,$14,$D7
	.dc.b	$F0,$ED,$7C,$4D,$AF,$E4,$B8,$78,$1B,$3F,$6A,$74,$49,$D3,$63,$51
	.dc.b	$82,$38,$21,$E6,$72,$D3,$27,$A1,$CB,$4C,$9E,$A7,$2D,$32,$7B,$1C
	.dc.b	$B4,$CC,$FB,$1F,$01,$0E,$C5,$3A,$14,$E4,$57,$93,$C0,$87,$EB,$76
	.dc.b	$FF,$77,$61,$FA,$D2,$47,$EA,$DF,$F2,$47,$E9,$0F,$B7,$FF,$7D,$C6
	.dc.b	$1B,$3F,$7A,$75,$82,$B4,$D9,$FE,$17,$C1,$C1,$FE,$0F,$19,$6C,$FE
	.dc.b	$D2,$75,$1F,$E3,$AE,$AF,$F2,$EF,$C9,$FF,$53,$24,$7F,$86,$5F,$F5
	.dc.b	$FF,$BF,$FE,$34,$F9,$6B,$E7,$B2,$46,$1B,$7F,$EC,$8A,$E0,$FE,$63
	.dc.b	$F0,$07,$CB,$71,$BF,$F3,$54,$62,$F8,$D7,$F7,$18,$C4,$EA,$B8,$4F
	.dc.b	$5F,$8B,$FC,$F5,$B0,$46,$1B,$1F,$E7,$73,$51,$9A,$E7,$5F,$E5,$C9
	.dc.b	$E5,$18,$8F,$75,$78,$9F,$64,$65,$7C,$8F,$FB,$C6,$BD,$31,$FE,$66
	.dc.b	$8C,$61,$FE,$87,$6C,$7E,$E9,$AF,$86,$DF,$88,$7A,$49,$7E,$98,$CB
	.dc.b	$FC,$1B,$A5,$C6,$E9,$F0,$E1,$FC,$EE,$31,$97,$F3,$F2,$47,$F0,$AF
	.dc.b	$C8,$FE,$04,$7F,$A6,$DB,$19,$3D,$CE,$1F,$F1,$25,$F8,$2F,$F1,$D1
	.dc.b	$FA,$16,$BE,$16,$C6,$F8,$FF,$D5,$1F,$0D,$F9,$CD,$AF,$37,$E2,$DB
	.dc.b	$A7,$E1,$B1,$FE,$3F,$5B,$7F,$90,$FE,$6B,$FA,$4D,$3F,$D6,$4D,$8C
	.dc.b	$E1,$F5,$63,$11,$FE,$3B,$AF,$D5,$FA,$FD,$89,$1D,$AB,$F3,$91,$FE
	.dc.b	$53,$FC,$57,$F9,$47,$9A,$69,$32,$FE,$6F,$F5,$AF,$AC,$FC,$23,$63
	.dc.b	$E9,$B7,$4F,$F6,$43,$72,$FF,$2F,$5F,$37,$C1,$5F,$D0,$B0,$F0,$3F
	.dc.b	$81,$F1,$3F,$38,$47,$07,$F4,$8F,$49,$EF,$FE,$55,$B1,$F8,$46,$E5
	.dc.b	$FE,$81,$9F,$CF,$4F,$EA,$6B,$A6,$1F,$C0,$1A,$FE,$CF,$E4,$C7,$F0
	.dc.b	$1F,$9C,$FD,$E3,$8A,$46,$23,$53,$C5,$AF,$F8,$57,$0F,$F3,$11,$1F
	.dc.b	$F5,$43,$FE,$15,$61,$FE,$13,$29,$FA,$7F,$C5,$7F,$89,$7E,$31,$FC
	.dc.b	$59,$B1,$FC,$45,$AE,$F0,$4F,$F1,$CF,$C4,$BF,$8D,$7D,$5D,$93,$49
	.dc.b	$5F,$C6,$F7,$FE,$41,$F5,$9F,$C9,$1B,$19,$A4,$E5,$B1,$FE,$E0,$FD
	.dc.b	$1B,$F9,$DF,$D6,$7F,$B9,$5B,$19,$E6,$FE,$59,$39,$6C,$7F,$80,$3E
	.dc.b	$09,$D1,$FC,$C7,$EB,$3F,$C0,$73,$FE,$DE,$8C,$3B,$6B,$FE,$9B,$9E
	.dc.b	$38,$AB,$FC,$85,$F5,$92,$BF,$C8,$0D,$7F,$D4,$5E,$4F,$F8,$FE,$3B
	.dc.b	$B6,$3F,$C2,$6D,$77,$E8,$7F,$45,$9D,$FD,$DF,$E0,$27,$FA,$83,$55
	.dc.b	$23,$0D,$C3,$F1,$73,$F8,$12,$1F,$C4,$BC,$4F,$E4,$0F,$E1,$0F,$E6
	.dc.b	$FF,$4D,$FD,$57,$F6,$9F,$EB,$2D,$8F,$EA,$8D,$7F,$98,$35,$FE,$7F
	.dc.b	$DF,$FD,$77,$F5,$9F,$EC,$2D,$8F,$EB,$8D,$7F,$9C,$35,$0D,$BF,$DC
	.dc.b	$3F,$51,$F5,$9B,$1F,$6F,$E0,$FF,$B5,$4E,$5B,$1F,$E7,$1E,$BF,$E6
	.dc.b	$EF,$BF,$C8,$FE,$7B,$E4,$7F,$31,$9F,$81,$3F,$CB,$CB,$1F,$CD,$5F
	.dc.b	$EF,$3F,$69,$FD,$DF,$F5,$1B,$C0,$69,$27,$F2,$50,$FA,$DF,$C9,$36
	.dc.b	$33,$18,$3F,$23,$F6,$0F,$6B,$D3,$3F,$EE,$98,$C5,$35,$DD,$6A,$BE
	.dc.b	$3D,$6A,$7D,$3C,$7F,$72,$7E,$F9,$FD,$34,$F8,$D6,$FD,$B6,$92,$3E
	.dc.b	$64,$65,$FF,$4D,$78,$BF,$E9,$F2,$38,$3A,$9F,$24,$BA,$FC,$97,$FC
	.dc.b	$DF,$FB,$BB,$F0,$7F,$F1,$07,$E8,$BF,$B0,$8C,$60,$8C,$4E,$AB,$8B
	.dc.b	$5F,$B4,$7B,$B8,$BF,$BD,$67,$EF,$8F,$F0,$17,$C7,$FF,$8B,$7F,$44
	.dc.b	$FF,$18,$B6,$3F,$C5,$8D,$77,$8A,$7A,$44,$3D,$75,$D5,$F9,$46,$2F
	.dc.b	$C2,$47,$F5,$55,$FE,$2B,$F8,$BF,$F1,$F7,$D7,$FF,$90,$9B,$3F,$E3
	.dc.b	$D6,$BF,$E1,$96,$BF,$E4,$5F,$73,$FC,$59,$3F,$E9,$8A,$46,$27,$F4
	.dc.b	$E6,$CE,$31,$A7,$99,$F5,$4F,$92,$7A,$4F,$71,$82,$1E,$E2,$9C,$4C
	.dc.b	$10,$C1,$36,$FC,$03,$FE,$43,$9F,$F1,$64,$FF,$B8,$63,$CE,$FA,$B0
	.dc.b	$FF,$31,$78,$3E,$DA,$DB,$63,$35,$C2,$7F,$A2,$D7,$F9,$4D,$B1,$C7
	.dc.b	$C7,$FF,$2D,$C6,$23,$DC,$43,$D7,$87,$99,$F5,$36,$BE,$6F,$D5,$7F
	.dc.b	$9D,$7E,$43,$FC,$ED,$5E,$B8,$C4,$F8,$DB,$19,$84,$FF,$B2,$EB,$B4
	.dc.b	$FC,$5D,$7F,$14,$DC,$BF,$10,$F9,$4F,$C0,$61,$FD,$B3,$0E,$C7,$F6
	.dc.b	$C2,$F9,$7E,$49,$DC,$65,$F6,$DF,$6C,$3C,$27,$B7,$27,$DE,$7F,$A4
	.dc.b	$6F,$91,$F8,$C7,$0F,$CF,$6D,$FD,$B6,$BF,$95,$FB,$EF,$8B,$3F,$55
	.dc.b	$B1,$F8,$16,$BB,$8F,$F4,$33,$F8,$5F,$BF,$CC,$7E,$71,$B6,$69,$CA
	.dc.b	$7E,$41,$F9,$83,$A1,$EA,$85,$72,$25,$78,$27,$EF,$E3,$F1,$49,$A7
	.dc.b	$D2,$8F,$92,$D7,$FA,$9F,$C3,$FF,$A9,$3D,$CF,$EA,$BE,$A7,$F3,$79
	.dc.b	$CF,$C0,$FF,$B2,$E3,$10,$AE,$0F,$F9,$03,$BF,$F6,$F1,$89,$DB,$63
	.dc.b	$35,$CD,$B7,$E0,$6D,$F3,$3F,$D1,$0F,$F8,$6B,$7F,$F4,$63,$72,$FF
	.dc.b	$4E,$37,$2F,$2D,$70,$3F,$84,$9F,$E8,$F8,$D1,$1F,$B9,$9F,$F5,$27
	.dc.b	$EE,$35,$FB,$91,$F4,$74,$FF,$6E,$3F,$A1,$BF,$0C,$F6,$C3,$CA,$D8
	.dc.b	$CB,$FD,$D6,$7F,$81,$6D,$CB,$0F,$E0,$0D,$9F,$C4,$47,$84,$FC,$41
	.dc.b	$76,$FC,$0B,$71,$E1,$3E,$68,$F5,$DF,$E1,$F7,$FF,$79,$7C,$05,$9F
	.dc.b	$F6,$2D,$7E,$7D,$AF,$F3,$39,$DC,$7A,$74,$F9,$73,$F2,$4C,$5F,$C9
	.dc.b	$7F,$82,$74,$FE,$57,$8F,$FE,$00,$36,$CD,$3E,$45,$F4,$DB,$E3,$DF
	.dc.b	$C5,$3A,$90,$C1,$0F,$B8,$7F,$13,$E8,$EC,$4B,$FE,$16,$F6,$C7,$CD
	.dc.b	$3C,$08,$7A,$CE,$85,$35,$3F,$D4,$B9,$BE,$FC,$7F,$35,$6B,$FD,$05
	.dc.b	$AF,$F5,$B6,$C7,$F4,$11,$F5,$BF,$CF,$DB,$3F,$CF,$F1,$46,$23,$FA
	.dc.b	$64,$62,$7D,$31,$E5,$26,$3F,$21,$F5,$B5,$CF,$0F,$E4,$1C,$B5,$CD
	.dc.b	$7D,$18,$C4,$F1,$23,$C6,$FF,$B7,$0F,$A8,$FE,$B5,$7E,$86,$C6,$63
	.dc.b	$FA,$D1,$93,$FC,$04,$E5,$FF,$01,$35,$F8,$3F,$04,$D3,$84,$6A,$BE
	.dc.b	$0E,$1E,$4C,$7D,$04,$D3,$39,$7A,$89,$19,$BF,$C5,$1F,$73,$1F,$C9
	.dc.b	$E4,$FD,$07,$FC,$52,$D9,$FF,$14,$B5,$FF,$14,$C7,$6A,$E9,$1F,$88
	.dc.b	$F0,$3F,$C7,$BF,$A0,$FF,$8F,$2C,$D2,$4F,$88,$FA,$39,$73,$FA,$4F
	.dc.b	$E3,$04,$8F,$D8,$3F,$E5,$AF,$E0,$19,$1F,$E5,$86,$BF,$98,$D7,$E2
	.dc.b	$57,$E1,$BC,$7F,$F3,$5F,$D5,$7F,$9A,$B2,$34,$97,$FE,$7A,$FA,$7F
	.dc.b	$F3,$C9,$1C,$1F,$E1,$AD,$DE,$67,$E5,$BD,$BF,$F4,$29,$27,$51,$97
	.dc.b	$DD,$9F,$91,$BC,$4E,$17,$CC,$7F,$37,$24,$FF,$0F,$7F,$CF,$39,$C8
	.dc.b	$FF,$3B,$B5,$FF,$55,$E8,$FD,$67,$FD,$62,$D8,$FF,$58,$B5,$DC,$7F
	.dc.b	$AF,$A1,$F7,$F4,$93,$FE,$C0,$8C,$BE,$8C,$3E,$FB,$FC,$23,$D8,$FF
	.dc.b	$68,$71,$FF,$B6,$28,$D2,$4F,$FB,$83,$C8,$FD,$EF,$D3,$7F,$B9,$1B
	.dc.b	$19,$A4,$E5,$B1,$FC,$C6,$3F,$AE,$11,$C1,$FE,$F9,$87,$D2,$FF,$B2
	.dc.b	$7D,$EF,$E1,$6D,$8C,$CE,$E3,$FD,$6F,$F4,$B2,$3F,$D7,$6D,$7F,$DA
	.dc.b	$B2,$E4,$7C,$17,$F1,$6F,$C4,$7F,$CE,$D1,$89,$46,$27,$55,$F8,$96
	.dc.b	$BD,$29,$F0,$9F,$D1,$A3,$F1,$CD,$74,$D7,$F8,$2B,$6F,$C9,$F0,$DF
	.dc.b	$CA,$3F,$1F,$F1,$B6,$BF,$30,$D7,$C0,$D7,$F9,$1F,$0F,$DC,$7F,$96
	.dc.b	$70,$BE,$83,$F9,$9F,$37,$F8,$22,$53,$9C,$7F,$33,$1F,$E0,$C8,$C4
	.dc.b	$ED,$B1,$FE,$98,$F9,$37,$A3,$2F,$F0,$F3,$98,$C7,$2F,$F8,$5A,$B7
	.dc.b	$1F,$D1,$23,$0E,$A7,$F6,$9F,$15,$FD,$3B,$73,$49,$87,$F7,$8F,$8C
	.dc.b	$7C,$67,$B2,$FF,$C4,$71,$89,$F5,$C7,$FA,$BA,$31,$A7,$C1,$6C,$78
	.dc.b	$27,$FC,$5A,$7E,$16,$FB,$F8,$5F,$84,$7F,$B1,$CB,$E3,$7F,$0D,$3F
	.dc.b	$2D,$B1,$FE,$34,$6B,$FE,$BE,$6A,$1F,$1C,$F8,$5F,$54,$7F,$B1,$23
	.dc.b	$18,$7A,$A3,$84,$63,$2F,$B4,$D8,$F0,$73,$79,$B0,$EF,$1E,$78,$C1
	.dc.b	$EA,$EA,$7C,$67,$F7,$6F,$98,$FC,$24,$63,$24,$62,$75,$7F,$E5,$0A
	.dc.b	$D4,$FE,$2F,$E4,$78,$67,$6F,$D7,$7E,$F9,$CB,$FE,$58,$6B,$FE,$E2
	.dc.b	$BF,$07,$D0,$F7,$C4,$D2,$4F,$F8,$E7,$E6,$BF,$CC,$84,$70,$7F,$99
	.dc.b	$3E,$57,$F7,$F9,$7E,$7D,$B1,$93,$CA,$1C,$BF,$EF,$4F,$1D,$E3,$FF
	.dc.b	$3D,$AF,$E7,$73,$98,$C3,$FC,$F4,$FE,$99,$EB,$FE,$64,$99,$7E,$33
	.dc.b	$9D,$1F,$8A,$F9,$67,$D8,$78,$9B,$5F,$7E,$D1,$8C,$7F,$70,$D8,$FE
	.dc.b	$B5,$F2,$FF,$3A,$BF,$8E,$37,$13,$4C,$7B,$FF,$AD,$24,$F9,$5F,$CE
	.dc.b	$F7,$70,$8F,$13,$71,$FE,$10,$DC,$3F,$79,$7D,$FD,$D9,$7F,$97,$3F
	.dc.b	$97,$38,$BF,$E1,$5E,$0F,$F9,$13,$E7,$2D,$1A,$4C,$3F,$D3,$BF,$38
	.dc.b	$F9,$BF,$F2,$5F,$27,$F1,$98,$C5,$A3,$13,$AA,$E2,$D7,$F6,$03,$E7
	.dc.b	$3F,$3D,$3F,$C8,$DB,$1F,$EA,$C6,$BC,$1A,$FE,$E7,$93,$F0,$BE,$37
	.dc.b	$86,$3F,$22,$7E,$15,$C3,$6B,$D1,$D7,$FC,$2B,$2F,$D6,$E5,$FE,$92
	.dc.b	$6F,$81,$A8,$C4,$EC,$8E,$09,$1B,$25,$FC,$89,$DC,$62,$3E,$0D,$76
	.dc.b	$75,$5C,$9B,$3B,$7F,$8B,$4E,$DE,$35,$D1,$D9,$0C,$D7,$36,$D7,$35
	.dc.b	$5C,$DE,$B5,$CD,$D9,$CC,$CD,$73,$6D,$73,$AE,$51,$D2,$B9,$3A,$39
	.dc.b	$19,$AE,$4F,$FB,$F7,$89,$F0,$7F,$D8,$E3,$0D,$F8,$4F,$F8,$03,$1F
	.dc.b	$EE,$E6,$BF,$2B,$E5,$BF,$D3,$AF,$E4,$DA,$FE,$61,$AE,$EF,$B6,$29
	.dc.b	$CC,$FC,$E6,$BF,$19,$B1,$B9,$E3,$18,$23,$82,$4F,$8D,$AF,$8D,$AF
	.dc.b	$8C,$D1,$0D,$14,$D1,$82,$18,$21,$E9,$7A,$B4,$E4,$53,$89,$4E,$27
	.dc.b	$BD,$E2,$D3,$89,$4E,$25,$38,$CF,$C5,$76,$D7,$6D,$7F,$69,$7F,$E2
	.dc.b	$1D,$7F,$72,$DE,$AF,$F3,$D3,$76,$FC,$2D,$7F,$88,$F8,$BF,$DB,$73
	.dc.b	$FC,$60,$91,$9C,$76,$FF,$0E,$F9,$DF,$F0,$CC,$62,$3F,$C0,$11,$89
	.dc.b	$FF,$19,$36,$3F,$8A,$87,$0F,$F9,$63,$95,$F8,$3F,$FB,$B3,$E4,$F7
	.dc.b	$82,$72,$A3,$31,$86,$CF,$1A,$F4,$1F,$CC,$88,$60,$FE,$4F,$5D,$7F
	.dc.b	$04,$FC,$BB,$4F,$F2,$51,$63,$BC,$6A,$DF,$61,$DE,$7D,$4D,$55,$E9
	.dc.b	$8C,$34,$F5,$47,$8C,$FE,$73,$E7,$7F,$95,$DF,$39,$8D,$3C,$1E,$B4
	.dc.b	$AF,$2E,$08,$C5,$7F,$98,$7F,$04,$FE,$93,$A5,$EA,$F2,$4A,$72,$FD
	.dc.b	$46,$BD,$F0,$FE,$11,$97,$8E,$3C,$C7,$F8,$F5,$C3,$FD,$43,$C3,$7C
	.dc.b	$47,$EF,$9B,$5F,$C4,$5B,$3F,$E1,$79,$FE,$07,$E3,$7F,$8C,$FE,$4B
	.dc.b	$FC,$48,$49,$FF,$19,$79,$3F,$2A,$3E,$1C,$78,$77,$FE,$17,$97,$8A
	.dc.b	$31,$97,$F1,$FC,$3F,$C7,$DF,$43,$FE,$2F,$24,$F4,$94,$7F,$97,$7D
	.dc.b	$EF,$E2,$BE,$6F,$FB,$A2,$7F,$94,$36,$B5,$1F,$13,$1F,$09,$1C,$1C
	.dc.b	$2B,$99,$1C,$1F,$63,$BF,$F1,$44,$70,$72,$AF,$79,$FE,$9A,$AE,$CF
	.dc.b	$E8,$27,$83,$67,$8B,$FB,$D2,$1F,$B0,$EB,$FE,$F6,$E0,$FB,$C7,$6C
	.dc.b	$FE,$53,$DE,$7F,$03,$F1,$9F,$38,$8E,$0F,$C8,$FD,$CF,$E7,$BE,$43
	.dc.b	$F6,$6D,$AF,$D6,$F3,$A3,$11,$FB,$F6,$BF,$A6,$87,$E2,$DF,$A8,$FD
	.dc.b	$BF,$F7,$AD,$9E,$6D,$7F,$84,$35,$E5,$E1,$7E,$C3,$F1,$AF,$E0,$AD
	.dc.b	$8F,$1B,$5F,$E2,$0D,$7C,$5D,$17,$3F,$E2,$0D,$C7,$F5,$10,$BF,$E2
	.dc.b	$33,$C6,$9F,$0F,$FE,$B5,$FC,$CF,$FD,$A4,$D8,$FE,$42,$D7,$FC,$6E
	.dc.b	$D7,$FC,$61,$E2,$3C,$47,$E9,$5F,$CD,$C9,$3F,$6E,$75,$1F,$ED,$FF
	.dc.b	$C1,$3F,$92,$B7,$C4,$FE,$22,$DD,$39,$DF,$F3,$2A,$DB,$E4,$FA,$BF
	.dc.b	$EE,$B9,$1A,$4A,$FF,$05,$57,$F8,$F3,$C7,$F0,$74,$7F,$48,$C3,$A3
	.dc.b	$5F,$F2,$9C,$ED,$FC,$1F,$A9,$FC,$93,$F1,$BF,$E7,$2D,$8C,$E0,$9C
	.dc.b	$B6,$3F,$17,$E2,$6D,$FE,$90,$D7,$F1,$4D,$8F,$E9,$0E,$61,$D7,$FA
	.dc.b	$27,$E5,$70,$7F,$89,$63,$11,$FE,$5D,$3F,$B3,$67,$1D,$B5,$FB,$3E
	.dc.b	$6F,$BD,$FF,$70,$6C,$8C,$47,$E1,$1B,$19,$C7,$FB,$41,$FD,$77,$9B
	.dc.b	$FA,$3F,$D7,$FE,$62,$0D,$25,$FE,$B3,$F0,$5F,$E9,$BF,$AA,$FF,$62
	.dc.b	$6C,$66,$93,$96,$C7,$FA,$1B,$EB,$5F,$A7,$FF,$49,$9F,$EF,$51,$8F
	.dc.b	$33,$D8,$D8,$CC,$EF,$5F,$C0,$C7,$F2,$AB,$FC,$0B,$5F,$C0,$F2,$7F
	.dc.b	$62,$F1,$B6,$7F,$11,$8F,$CF,$F0,$FD,$1E,$9F,$D0,$46,$4F,$AE,$FE
	.dc.b	$8E,$7F,$C3,$04,$8F,$B2,$FF,$83,$17,$EB,$AF,$A7,$FE,$E3,$E9,$66
	.dc.b	$31,$97,$F9,$DA,$BF,$C2,$71,$89,$D5,$70,$9F,$8D,$A7,$C5,$C3,$FC
	.dc.b	$33,$AF,$C3,$6C,$66,$77,$B7,$BD,$F8,$75,$F1,$88,$E0,$DE,$48,$E7
	.dc.b	$3F,$E0,$86,$C6,$67,$71,$E6,$8F,$F7,$E4,$7A,$B0,$FF,$07,$F9,$D9
	.dc.b	$24,$7E,$78,$F9,$9C,$DF,$9E,$7E,$89,$FE,$2A,$F2,$3F,$C3,$84,$BF
	.dc.b	$F0,$6B,$FE,$1F,$FC,$13,$FC,$37,$F0,$FF,$E3,$06,$D7,$F3,$E6,$BF
	.dc.b	$E3,$B6,$BF,$CF,$7E,$02,$DF,$F8,$ED,$BE,$29,$A7,$09,$FC,$0F,$1F
	.dc.b	$E1,$48,$FE,$26,$9F,$87,$FF,$8B,$9F,$F1,$BB,$63,$FC,$92,$D7,$FC
	.dc.b	$7E,$D7,$F7,$C7,$C5,$21,$83,$FD,$8B,$F6,$4F,$9C,$FF,$4D,$F2,$23
	.dc.b	$17,$FE,$F7,$6D,$7F,$BF,$1A,$F8,$3E,$2B,$FC,$87,$FB,$37,$91,$B1
	.dc.b	$9A,$4E,$5B,$1E,$8D,$BF,$CA,$B9,$1A,$49,$F8,$B3,$EA,$3E,$E9,$5F
	.dc.b	$8F,$D1,$C8,$8F,$C8,$7F,$D6,$BC,$3F,$85,$9F,$A2,$7E,$22,$3F,$6A
	.dc.b	$F5,$32,$58,$FC,$67,$23,$D4,$7D,$B7,$FB,$2B,$0F,$E2,$11,$85,$7F
	.dc.b	$EC,$CA,$E1,$5B,$8C,$C6,$27,$67,$7C,$3C,$73,$F5,$CF,$79,$4F,$29
	.dc.b	$70,$FD,$04,$BE,$11,$F8,$47,$F9,$DB,$ED,$FF,$DB,$B3,$98,$C3,$FE
	.dc.b	$86,$9D,$A9,$DC,$61,$E3,$CA,$F5,$3D,$58,$FF,$87,$5B,$7F,$E9,$B6
	.dc.b	$CE,$E3,$DB,$3E,$ED,$FF,$B9,$46,$23,$FD,$17,$18,$7C,$ED,$8F,$CE
	.dc.b	$F9,$7E,$09,$F8,$27,$C4,$21,$88,$F1,$1D,$4A,$FA,$AF,$CA,$73,$CF
	.dc.b	$F0,$84,$D3,$E8,$CF,$48,$C7,$37,$E4,$A3,$E9,$7B,$EF,$80,$FF,$67
	.dc.b	$F1,$FF,$65,$9F,$C0,$CE,$7C,$3F,$C1,$CE,$65,$39,$7F,$41,$E4,$FE
	.dc.b	$21,$CD,$F4,$8C,$4E,$EB,$93,$63,$33,$BA,$F0,$B5,$FE,$1F,$1F,$EE
	.dc.b	$52,$3F,$E0,$66,$BE,$1C,$59,$7F,$8D,$E7,$33,$FD,$71,$FF,$6A,$E7
	.dc.b	$FE,$C2,$7E,$66,$9F,$E0,$19,$FF,$70,$4E,$63,$FC,$CE,$D8,$ED,$5D
	.dc.b	$5F,$11,$0C,$E9,$F1,$DF,$C2,$EB,$C5,$B7,$DD,$B8,$7E,$22,$75,$3F
	.dc.b	$82,$8F,$29,$1F,$19,$C4,$FD,$21,$FE,$7B,$3F,$49,$C5,$FE,$DE,$F5
	.dc.b	$7F,$46,$D8,$C9,$BB,$F3,$D7,$C6,$7E,$79,$EB,$3E,$41,$4C,$9C,$0F
	.dc.b	$E1,$3C,$BE,$DE,$5F,$E0,$93,$FC,$27,$9B,$F8,$63,$67,$E1,$57,$23
	.dc.b	$A9,$CC,$86,$08,$6C,$E1,$7F,$CA,$0F,$9C,$59,$FE,$9B,$A7,$DD,$9F
	.dc.b	$E9,$AD,$7A,$9C,$F6,$FC,$3E,$BF,$5E,$3A,$F1,$FF,$3B,$C3,$AB,$70
	.dc.b	$FE,$B3,$C5,$CE,$BF,$80,$B6,$BF,$B0,$F4,$17,$EC,$7F,$D2,$49,$5F
	.dc.b	$D3,$5F,$EA,$9D,$9C,$FB,$3F,$AD,$37,$B3,$FA,$DE,$5A,$C3,$CD,$AF
	.dc.b	$A2,$3E,$C3,$71,$FB,$13,$FD,$C8,$F2,$69,$FD,$CE,$BF,$B8,$C6,$0F
	.dc.b	$54,$FC,$3D,$BF,$38,$FC,$6B,$E8,$FF,$3E,$E9,$FF,$02,$53,$E4,$7F
	.dc.b	$80,$4F,$F8,$2E,$30,$FF,$B3,$BB,$BE,$14,$EA,$7F,$C1,$C6,$A7,$FC
	.dc.b	$28,$D7,$FD,$A1,$DF,$C1,$FF,$0D,$46,$45,$A7,$53,$E5,$F0,$3E,$94
	.dc.b	$9A,$4A,$F9,$39,$7B,$AB,$F1,$B9,$7B,$9F,$C7,$F8,$DF,$E5,$4F,$CD
	.dc.b	$BF,$C7,$84,$8F,$F1,$E5,$A3,$FC,$87,$E4,$49,$FF,$24,$B5,$FF,$37
	.dc.b	$CA,$3F,$CE,$A4,$8F,$F3,$AB,$FE,$4C,$F8,$69,$1F,$E5,$26,$BF,$CB
	.dc.b	$6F,$FC,$CF,$F8,$27,$F9,$3C,$91,$E4,$7F,$CE,$18,$FE,$72,$AE,$5F
	.dc.b	$ED,$96,$F4,$1A,$7E,$2A,$3E,$25,$FE,$0E,$7F,$49,$E9,$FA,$73,$E1
	.dc.b	$3F,$E8,$CB,$FF,$47,$91,$FF,$34,$FC,$37,$F9,$98,$8E,$0F,$F5,$17
	.dc.b	$EA,$9F,$EA,$12,$38,$3F,$A5,$57,$FA,$C4,$7E,$2F,$FE,$B1,$24,$6B
	.dc.b	$85,$E9,$7E,$BE,$7F,$DF,$BC,$DF,$B1,$C6,$E0,$69,$30,$FF,$67,$BF
	.dc.b	$10,$8E,$0F,$E4,$1B,$7E,$15,$F8,$95,$F9,$BC,$5E,$A7,$EF,$FF,$17
	.dc.b	$C3,$6B,$FD,$CB,$8A,$72,$D8,$FF,$76,$C7,$FB,$B0,$8E,$0F,$F5,$CE
	.dc.b	$3F,$EB,$C2,$38,$3F,$63,$EA,$FF,$85,$31,$FF,$58,$36,$33,$E3,$7E
	.dc.b	$FA,$31,$1F,$DD,$F8,$1E,$CF,$F1,$7F,$A4,$FF,$5C,$46,$33,$46,$27
	.dc.b	$55,$FC,$02,$B5,$3F,$8C,$24,$7F,$BE,$9A,$FF,$81,$6D,$E1,$7F,$25
	.dc.b	$FC,$DB,$CE,$D9,$F5,$B5,$E0,$D7,$F9,$66,$EF,$77,$F9,$77,$E2,$FF
	.dc.b	$4D,$B3,$F8,$56,$BF,$C6,$5A,$F2,$E4,$BA,$FE,$3E,$BF,$DF,$C4,$9F
	.dc.b	$C7,$7C,$07,$F3,$BF,$E0,$1F,$C0,$36,$33,$49,$CB,$63,$F2,$DE,$6F
	.dc.b	$F4,$3F,$1B,$F1,$0D,$8C,$CF,$FA,$6D,$B3,$FE,$E9,$F8,$4F,$F2,$97
	.dc.b	$25,$F3,$FF,$9B,$C6,$2B,$F9,$8B,$5F,$F1,$93,$FC,$86,$FF,$AE,$36
	.dc.b	$BA,$78,$7C,$8F,$E8,$3D,$5F,$9D,$27,$7F,$F0,$27,$F9,$25,$FF,$30
	.dc.b	$F0,$BF,$96,$91,$FE,$E3,$C5,$FC,$17,$0E,$B1,$E0,$C3,$B3,$4D,$D7
	.dc.b	$19,$FD,$53,$72,$FE,$97,$F0,$CF,$86,$FE,$AA,$DB,$FE,$EC,$DB,$FE
	.dc.b	$A3,$48,$C3,$F1,$30,$F0,$CF,$90,$91,$DA,$FA,$92,$FA,$BF,$D9,$3E
	.dc.b	$1D,$F1,$1F,$DE,$FC,$3E,$F1,$8A,$FE,$47,$7F,$E6,$18,$C4,$EA,$B8
	.dc.b	$4F,$8E,$BE,$16,$1E,$28,$FA,$F8,$F7,$AF,$F0,$47,$CD,$71,$B3,$49
	.dc.b	$5F,$DC,$3D,$B7,$C5,$7F,$7A,$6D,$7F,$86,$DB,$8F,$C6,$6D,$78,$E3
	.dc.b	$C4,$FF,$B0,$38,$BF,$0C,$E4,$FE,$FB,$D5,$FC,$7A,$DF,$21,$F4,$23
	.dc.b	$0D,$F1,$7F,$C4,$0D,$9D,$D7,$F3,$7F,$7F,$FB,$A9,$23,$F1,$2F,$D6
	.dc.b	$8F,$01,$F1,$23,$E3,$BF,$CF,$FE,$37,$AB,$63,$AB,$FE,$15,$8C,$B6
	.dc.b	$32,$FF,$08,$F8,$EF,$F0,$29,$23,$F3,$2F,$86,$3C,$19,$FC,$45,$9F
	.dc.b	$C4,$7F,$C4,$5E,$2F,$67,$F8,$9B,$6F,$FC,$2A,$DE,$7F,$E5,$59,$FE
	.dc.b	$5B,$E3,$DF,$33,$FD,$57,$0F,$F1,$4C,$63,$0F,$EA,$D1,$89,$D3,$FE
	.dc.b	$AD,$9D,$79,$3F,$AE,$24,$7E,$35,$AF,$FA,$49,$FE,$B6,$D3,$FA,$E1
	.dc.b	$FE,$B7,$F5,$BF,$B4,$DF,$F9,$85,$BA,$7E,$B1,$B3,$FE,$3E,$6B,$C7
	.dc.b	$C4,$FF,$24,$C6,$20,$1C,$36,$36,$4A,$F0,$36,$BC,$0D,$9F,$03,$5F
	.dc.b	$04,$F0,$24,$F0,$6C,$70,$6B,$C2,$34,$47,$4D,$74,$60,$86,$08,$7B
	.dc.b	$AB,$DA,$4A,$F3,$46,$5A,$FC,$1A,$F1,$B9,$6B,$9A,$F1,$39,$6B,$93
	.dc.b	$FC,$DF,$EF,$7F,$BC,$5B,$EF,$3D,$FF,$E4,$71,$FE,$F2,$6C,$EE,$3F
	.dc.b	$01,$CB,$FE,$95,$7E,$F3,$5F,$BC,$D7,$78,$74,$CB,$FB,$83,$67,$53
	.dc.b	$F0,$DA,$FE,$05,$B3,$C2,$3B,$CF,$12,$38,$21,$E2,$72,$D3,$27,$90
	.dc.b	$A7,$85,$DB,$93,$81,$0C,$10,$E0,$72,$3D,$0F,$27,$27,$12,$9C,$4D
	.dc.b	$9C,$63,$DC,$EC,$C9,$5C,$BC,$5A,$F4,$21,$C0,$D4,$FC,$A2,$4F,$96
	.dc.b	$1E,$07,$FB,$6B,$0F,$F3,$E9,$23,$FD,$6B,$19,$7F,$A9,$47,$F7,$62
	.dc.b	$38,$3F,$CE,$3E,$C7,$F5,$DF,$90,$FF,$5C,$36,$33,$D9,$FE,$DD,$9C
	.dc.b	$B6,$3B,$F8,$6F,$4F,$F2,$78,$FF,$BB,$A7,$FD,$22,$43,$FD,$93,$F0
	.dc.b	$3F,$DA,$DF,$EF,$6D,$73,$7F,$8C,$8C,$47,$F0,$EC,$D9,$3E,$CB,$C5
	.dc.b	$39,$8C,$4B,$3F,$5D,$6A,$BF,$DE,$D5,$A9,$F2,$47,$F1,$A3,$E6,$E7
	.dc.b	$EB,$72,$7F,$1E,$CB,$D8,$FA,$BD,$2F,$31,$31,$FD,$9E,$48,$FD,$2F
	.dc.b	$7F,$F7,$6C,$7F,$BA,$78,$1E,$3C,$13,$97,$FB,$F4,$67,$1E,$F1,$E5
	.dc.b	$6B,$E3,$CB,$CC,$FD,$2D,$BF,$81,$DF,$E3,$A3,$AF,$07,$88,$9A,$7F
	.dc.b	$08,$E7,$FE,$A5,$F0,$1F,$D5,$FA,$52,$8C,$36,$FF,$AB,$35,$F2,$65
	.dc.b	$E2,$6E,$5F,$82,$8C,$47,$9E,$3C,$17,$DB,$E1,$FA,$92,$7A,$BF,$E3
	.dc.b	$EF,$C9,$9E,$0F,$F0,$E6,$E3,$FC,$81,$B5,$B9,$FE,$27,$B7,$C6,$8F
	.dc.b	$B7,$8F,$8C,$8E,$EF,$95,$75,$23,$FC,$B3,$E7,$1D,$7F,$C7,$9B,$19
	.dc.b	$8C,$1E,$0D,$3D,$F7,$E0,$3F,$12,$53,$F1,$C2,$B9,$BF,$E5,$C6,$BA
	.dc.b	$6B,$F9,$56,$BF,$DC,$3F,$07,$7E,$45,$F8,$B6,$E1,$FA,$19,$EF,$18
	.dc.b	$E4,$FF,$34,$B6,$3F,$10,$79,$B4,$F9,$07,$C9,$F5,$B9,$C7,$FB,$3B
	.dc.b	$0F,$E7,$F1,$88,$FF,$69,$C7,$F4,$4C,$3F,$6B,$F9,$0F,$E2,$BE,$5B
	.dc.b	$F8,$5B,$6B,$F8,$7B,$5F,$DC,$35,$FD,$AF,$DA,$5A,$FD,$FB,$7A,$CD
	.dc.b	$31,$7F,$98,$F8,$AF,$E3,$2D,$BF,$E3,$AD,$7F,$83,$35,$FE,$07,$C1
	.dc.b	$70,$FE,$1A,$DC,$3E,$FD,$FF,$24,$8E,$14,$D7,$C7,$7F,$EE,$C8,$F9
	.dc.b	$F8,$FE,$AB,$C7,$F9,$ED,$8C,$CE,$DF,$F7,$36,$9F,$B0,$F9,$5F,$9D
	.dc.b	$26,$92,$7F,$D8,$3D,$5B,$6C,$ED,$B3,$FE,$CB,$9D,$47,$FB,$1B,$E5
	.dc.b	$3F,$88,$79,$46,$92,$FF,$11,$3E,$47,$FC,$1B,$C3,$F7,$E3,$07,$98
	.dc.b	$FC,$DF,$C6,$3F,$79,$FF,$03,$FE,$79,$86,$CF,$F2,$D9,$D6,$5F,$C3
	.dc.b	$AB,$4D,$9F,$EA,$7A,$7F,$12,$F9,$6F,$EA,$6D,$8C,$E4,$9C,$B6,$3F
	.dc.b	$5B,$F8,$43,$EC,$FF,$97,$36,$7D,$B3,$FE,$58,$8C,$74,$FF,$B3,$1B
	.dc.b	$1F,$A4,$21,$FC,$70,$FD,$DE,$BF,$0D,$FF,$30,$B5,$34,$D5,$7F,$58
	.dc.b	$FA,$7F,$E5,$BF,$82,$FF,$6E,$6D,$7F,$72,$6B,$FD,$6D,$AB,$6F,$F4
	.dc.b	$5F,$D3,$61,$B6,$BF,$A7,$E1,$AB,$4D,$9D,$5F,$F1,$2C,$5F,$83,$FF
	.dc.b	$53,$6C,$FC,$69,$F5,$CE,$4D,$6F,$FC,$08,$93,$F8,$67,$FC,$87,$A7
	.dc.b	$CF,$F1,$7E,$66,$FF,$D3,$1B,$5F,$D2,$70,$FA,$53,$97,$E7,$9F,$D1
	.dc.b	$CF,$E2,$9E,$0F,$F2,$CF,$8D,$7E,$2B,$FD,$5B,$CF,$FE,$57,$49,$CE
	.dc.b	$1F,$E6,$58,$FE,$5F,$18,$9D,$B6,$3F,$29,$F0,$CF,$91,$FF,$03,$71
	.dc.b	$A3,$18,$7E,$22,$31,$3A,$9C,$F2,$F8,$BF,$0C,$7E,$0B,$E5,$C7,$F9
	.dc.b	$FC,$63,$1F,$24,$7F,$83,$E3,$19,$F1,$AE,$13,$FE,$AC,$CB,$BD,$FC
	.dc.b	$76,$EF,$FB,$0C,$BC,$AD,$CF,$E5,$5F,$F8,$6F,$4F,$87,$EA,$7C,$0D
	.dc.b	$BF,$BE,$B7,$2F,$F5,$C5,$7C,$69,$CB,$F4,$8F,$EB,$07,$E8,$4F,$AB
	.dc.b	$0F,$A0,$FF,$1A,$C7,$F8,$21,$B1,$9F,$57,$FA,$24,$63,$C7,$FC,$6F
	.dc.b	$3F,$F9,$3B,$9F,$FC,$1E,$DB,$FD,$4E,$1F,$E5,$49,$CB,$FE,$56,$F9
	.dc.b	$9F,$F0,$AF,$CC,$FF,$95,$9B,$19,$A4,$E5,$B1,$EB,$F3,$3F,$AE,$F3
	.dc.b	$FF,$76,$7F,$C5,$73,$EB,$F1,$BC,$E4,$7F,$1A,$64,$FF,$04,$D7,$3A
	.dc.b	$E4,$7F,$87,$B7,$7B,$6F,$97,$FF,$53,$C6,$1B,$19,$BF,$F7,$83,$6F
	.dc.b	$FC,$E3,$F8,$5F,$F8,$CB,$F1,$CB,$46,$92,$7F,$CF,$3F,$83,$7F,$71
	.dc.b	$E1,$34,$95,$E8,$9F,$B4,$43,$F5,$C7,$F9,$85,$FB,$9E,$07,$D8,$23
	.dc.b	$F0,$B8,$FF,$8D,$FC,$0B,$EF,$BF,$C8,$FF,$51,$FE,$CB,$A4,$E5,$46
	.dc.b	$63,$0D,$9E,$33,$F2,$FD,$7F,$36,$3E,$43,$F4,$5F,$9D,$BE,$FF,$FD
	.dc.b	$B5,$5F,$D2,$E3,$0F,$FA,$16,$77,$3A,$6C,$65,$FF,$2E,$E3,$E5,$F8
	.dc.b	$7F,$14,$9A,$47,$C9,$3F,$C4,$B0,$FC,$1D,$7B,$67,$89,$0F,$94,$E5
	.dc.b	$FF,$56,$F8,$9E,$3A,$FF,$48,$B6,$3F,$D3,$6D,$DC,$D2,$47,$A4,$D9
	.dc.b	$63,$6F,$B6,$7F,$6E,$7C,$92,$18,$78,$BF,$5E,$38,$B6,$7B,$BF,$E8
	.dc.b	$23,$F0,$C7,$42,$73,$FE,$AB,$0F,$AF,$18,$E6,$FF,$30,$FC,$1B,$E1
	.dc.b	$3F,$99,$75,$7F,$9A,$A8,$27,$5E,$36,$27,$53,$9A,$E3,$EB,$FF,$AD
	.dc.b	$6F,$73,$FB,$E8,$C4,$EA,$3A,$CE,$AB,$9B,$6B,$77,$DD,$FF,$4B,$9F
	.dc.b	$D2,$B2,$63,$FE,$DB,$8C,$37,$0F,$F5,$74,$63,$2F,$3C,$7F,$A7,$63
	.dc.b	$04,$7F,$03,$AF,$F9,$F0,$FB,$31,$F3,$DC,$6B,$F6,$A3,$E8,$4F,$33
	.dc.b	$B9,$0F,$39,$D1,$E4,$49,$C1,$23,$FD,$C9,$AF,$FA,$F2,$75,$18,$9F
	.dc.b	$26,$DF,$B7,$FA,$DF,$D1,$92,$3F,$37,$C7,$FE,$C4,$8F,$2D,$7F,$AB
	.dc.b	$2F,$DC,$F9,$08,$7E,$BA,$7F,$D9,$7E,$57,$EB,$89,$7D,$0F,$F1,$A4
	.dc.b	$FE,$DA,$BE,$0F,$90,$E5,$FD,$EC,$61,$B5,$FC,$07,$97,$D8,$FC,$5B
	.dc.b	$F9,$07,$88,$F3,$1D,$FB,$7F,$7E,$48,$C1,$FC,$87,$A3,$FD,$53,$87
	.dc.b	$FA,$DA,$3F,$91,$9F,$78,$AE,$63,$75,$E1,$3B,$9E,$1E,$2F,$C5,$F0
	.dc.b	$F1,$E1,$FE,$6B,$85,$EA,$FA,$97,$FC,$EA,$30,$6F,$B3,$FD,$E7,$F8
	.dc.b	$77,$F4,$22,$4F,$F3,$7B,$4E,$A3,$F5,$A3,$BF,$FA,$85,$FE,$59,$BA
	.dc.b	$7F,$B4,$9B,$3C,$1A,$FA,$79,$BF,$40,$DC,$FE,$16,$BF,$EA,$0B,$F8
	.dc.b	$5E,$07,$AA,$BF,$7B,$8F,$F6,$79,$CB,$FA,$68,$F8,$A7,$D6,$21,$85
	.dc.b	$9B,$C1,$47,$F8,$D6,$7F,$BF,$78,$FF,$DF,$48,$FF,$81,$E1,$F9,$2F
	.dc.b	$F8,$2C,$91,$97,$EC,$78,$DE,$78,$FF,$0C,$36,$7F,$C3,$07,$B9,$D4
	.dc.b	$7A,$09,$C5,$F7,$32,$FD,$0C,$7F,$1C,$F2,$3F,$CA,$3F,$3F,$FE,$2D
	.dc.b	$24,$7F,$8B,$70,$47,$F8,$DD,$7C,$6D,$24,$FF,$95,$33,$F8,$31,$FE
	.dc.b	$45,$9F,$F0,$6C,$FF,$93,$3C,$DF,$8B,$E4,$64,$91,$C6,$5F,$17,$DC
	.dc.b	$91,$97,$FC,$85,$32,$3C,$CD,$7F,$01,$9F,$B9,$F8,$5E,$87,$FA,$AF
	.dc.b	$1F,$F3,$D9,$23,$FD,$63,$DD,$E8,$7F,$D1,$DE,$4F,$CA,$F3,$3F,$4F
	.dc.b	$1F,$EC,$DC,$3F,$DB,$E7,$F9,$EA,$3F,$D2,$39,$5F,$33,$F0,$2D,$D3
	.dc.b	$D7,$7F,$E9,$99,$CB,$FE,$CD,$F4,$7F,$D3,$DF,$6B,$FE,$A2,$F4,$FF
	.dc.b	$12,$DB,$34,$FF,$59,$38,$AF,$F5,$93,$5F,$F5,$93,$5F,$C9,$F4,$3D
	.dc.b	$5E,$B1,$EE,$9F,$CF,$67,$EE,$8F,$F6,$2E,$BF,$04,$FC,$5F,$37,$E2
	.dc.b	$63,$F8,$43,$5F,$E3,$1C,$3F,$ED,$7D,$BF,$D4,$E4,$8F,$9B,$87,$C2
	.dc.b	$7F,$5F,$93,$0F,$D3,$FC,$F7,$FB,$90,$93,$FE,$E8,$7F,$DD,$F4,$FC
	.dc.b	$EE,$92,$7F,$DE,$DE,$DB,$DA,$FD,$8C,$FD,$69,$F7,$71,$D7,$ED,$1F
	.dc.b	$C4,$07,$BA,$F7,$7E,$AC,$7F,$14,$8C,$63,$E9,$8F,$F0,$1C,$63,$5F
	.dc.b	$F5,$DB,$63,$A3,$E3,$97,$CE,$EB,$2F,$E0,$AD,$8C,$D7,$09,$CB,$F5
	.dc.b	$67,$F1,$DA,$E2,$3F,$93,$F0,$FC,$9B,$E3,$18,$9D,$3F,$E1,$8F,$81
	.dc.b	$78,$5F,$C5,$3D,$1F,$F6,$8C,$62,$FF,$C7,$35,$F0,$E3,$13,$AA,$E1
	.dc.b	$3F,$CE,$76,$FE,$3B,$F2,$3F,$D3,$F8,$D1,$89,$D6,$2A,$D4,$E6,$7F
	.dc.b	$03,$F0,$95,$FC,$AB,$6B,$F9,$5B,$6C,$DC,$3F,$96,$FC,$CF,$85,$B7
	.dc.b	$FD,$35,$AF,$E3,$9A,$FE,$8B,$E1,$AE,$1F,$CE,$1B,$81,$A7,$09,$6D
	.dc.b	$FA,$56,$ED,$FA,$5B,$FD,$4B,$FA,$6F,$2B,$FA,$8C,$FF,$46,$E9,$FD
	.dc.b	$33,$FD,$23,$83,$EC,$72,$7D,$2B,$FC,$7C,$FD,$76,$A9,$DB,$F9,$7F
	.dc.b	$87,$F9,$F8,$FF,$55,$6F,$07,$E6,$E5,$18,$78,$47,$97,$CB,$FE,$B9
	.dc.b	$97,$92,$7F,$A2,$E7,$E8,$78,$D7,$F4,$AB,$E5,$1F,$DD,$9A,$FF,$99
	.dc.b	$1A,$BE,$25,$F1,$3F,$D4,$9B,$97,$F6,$B6,$DF,$9D,$B5,$FC,$33,$0F
	.dc.b	$16,$3E,$53,$CD,$3A,$9F,$F0,$3B,$5F,$F3,$AB,$5F,$16,$1F,$D8,$BE
	.dc.b	$2D,$E1,$7F,$74,$F9,$4F,$8D,$18,$CD,$18,$9D,$5F,$F0,$5A,$D4,$FF
	.dc.b	$4B,$C7,$C9,$97,$98,$F9,$13,$CF,$C6,$E5,$8F,$F6,$6C,$39,$63,$FD
	.dc.b	$35,$FF,$0F,$D6,$DF,$E2,$5F,$1B,$FC,$1F,$5F,$B4,$D8,$CF,$67,$F7
	.dc.b	$C8,$C4,$7F,$5E,$3E,$47,$07,$F8,$5E,$1F,$1D,$F8,$A6,$E5,$FC,$66
	.dc.b	$91,$8E,$4E,$F5,$F9,$0F,$90,$7C,$87,$F8,$29,$B8,$7F,$88,$DB,$9F
	.dc.b	$E3,$DB,$8F,$8A,$3E,$36,$FF,$A2,$AF,$F1,$2F,$C8,$FE,$79,$A7,$F2
	.dc.b	$B2,$E5,$FC,$2E,$3C,$5E,$8F,$84,$93,$E1,$7F,$35,$7D,$B5,$FC,$DE
	.dc.b	$1D,$FA,$D5,$DD,$B7,$DD,$B8,$FF,$19,$9F,$D7,$F8,$4F,$94,$FF,$0A
	.dc.b	$E9,$FC,$E9,$AF,$F4,$36,$BF,$D0,$9A,$B9,$7F,$CD,$3E,$77,$FA,$BC
	.dc.b	$91,$FD,$3E,$32,$FF,$96,$F8,$FF,$C6,$FE,$B7,$DC,$8F,$F6,$7B,$5F
	.dc.b	$F3,$24,$68,$9A,$7F,$95,$67,$44,$8C,$EF,$FC,$0E,$31,$3B,$9C,$D7
	.dc.b	$86,$30,$4A,$E1,$3A,$9F,$F3,$73,$5F,$2B,$5F,$2C,$F1,$7A,$4F,$28
	.dc.b	$C4,$72,$6B,$C9,$AF,$28,$E0,$47,$99,$0E,$4E,$5A,$64,$E8,$E5,$A6
	.dc.b	$4E,$C5,$AF,$8B,$1B,$8C,$9D,$0F,$F7,$89,$FC,$B7,$E0,$BF,$DF,$6F
	.dc.b	$F8,$25,$AE,$8F,$F4,$B7,$89,$FE,$D0,$7E,$DB,$5F,$F4,$F3,$6B,$FB
	.dc.b	$35,$FF,$39,$9C,$B6,$33,$1F,$6E,$78,$12,$74,$D8,$D4,$60,$8E,$08
	.dc.b	$79,$9C,$B4,$C9,$FE,$AD,$8E,$8E,$48,$66,$38,$B4,$E2,$7C,$07,$2D
	.dc.b	$32,$7B,$8B,$3E,$E7,$A9,$B2,$9B,$2B,$B7,$C0,$62,$7C,$EE,$08,$73
	.dc.b	$29,$CC,$B1,$CD,$D1,$0D,$17,$1F,$F7,$2F,$C1,$7F,$6F,$6C,$7E,$F7
	.dc.b	$85,$C4,$AD,$46,$1F,$F1,$47,$E3,$BF,$EC,$6F,$8C,$B9,$9A,$4C,$3F
	.dc.b	$CA,$DF,$62,$F8,$9E,$11,$86,$BF,$EE,$A8,$D5,$73,$8D,$35,$0F,$B2
	.dc.b	$7E,$70,$FB,$5F,$CC,$46,$1B,$19,$AF,$F0,$74,$E5,$FF,$08,$1F,$EE
	.dc.b	$2F,$C4,$BC,$77,$FE,$90,$6C,$64,$F1,$46,$5A,$F8,$FD,$0E,$2D,$8C
	.dc.b	$D7,$F9,$CB,$3F,$0D,$74,$FA,$C7,$C6,$FE,$BE,$D1,$87,$CA,$DC,$3F
	.dc.b	$D0,$11,$A9,$FD,$A7,$85,$FE,$77,$C3,$FB,$8F,$47,$8A,$31,$7F,$DC
	.dc.b	$AF,$CF,$1F,$29,$B1,$B2,$3E,$CF,$B6,$FE,$A3,$1F,$E3,$56,$BF,$E9
	.dc.b	$16,$BF,$50,$38,$7F,$A7,$43,$E4,$BF,$D1,$FD,$88,$C4,$FF,$57,$6B
	.dc.b	$FD,$5D,$AF,$E0,$2B,$FB,$1E,$3E,$3E,$EF,$9F,$18,$9F,$24,$7D,$2E
	.dc.b	$EF,$EC,$96,$F9,$AF,$0B,$6B,$C1,$18,$3F,$C7,$9F,$13,$FC,$7F,$CC
	.dc.b	$34,$95,$FC,$7F,$ED,$DE,$F7,$E0,$1B,$1F,$84,$6E,$9F,$E5,$D6,$E3
	.dc.b	$FE,$B9,$C3,$AD,$7F,$29,$6C,$78,$A3,$18,$7E,$17,$C5,$7C,$05,$24
	.dc.b	$FE,$6A,$32,$FF,$33,$F4,$1F,$4D,$F9,$46,$C7,$E6,$1B,$9F,$F9,$C5
	.dc.b	$B8,$FD,$B9,$5F,$48,$FF,$54,$61,$E0,$6C,$FF,$41,$F7,$9D,$5F,$A2
	.dc.b	$6C,$7E,$99,$B8,$FF,$9F,$DB,$5F,$81,$8F,$AA,$47,$07,$FB,$53,$F0
	.dc.b	$6F,$B8,$4A,$FD,$8F,$0A,$3F,$6F,$F9,$B3,$E4,$3F,$1D,$F8,$CF,$EA
	.dc.b	$23,$16,$8C,$4E,$B0,$F8,$15,$A9,$E1,$E7,$FE,$5F,$DD,$B8,$7F,$AF
	.dc.b	$2F,$73,$96,$C6,$5F,$E2,$F8,$3E,$4D,$F2,$3F,$A1,$FD,$2B,$FD,$F5
	.dc.b	$39,$8C,$57,$EA,$E3,$16,$AD,$DF,$28,$C5,$7A,$3E,$8F,$F5,$9F,$43
	.dc.b	$EF,$6C,$66,$53,$96,$C7,$EE,$47,$CA,$7C,$6F,$A1,$FF,$5A,$36,$34
	.dc.b	$D7,$F8,$03,$5F,$DF,$C7,$EB,$88,$FE,$F5,$FF,$5F,$7D,$63,$EA,$7F
	.dc.b	$81,$7D,$1E,$1B,$5E,$39,$D6,$0A,$D3,$67,$EB,$FD,$A7,$E0,$09,$1F
	.dc.b	$CD,$87,$1A,$3F,$9A,$DF,$E1,$E7,$F9,$CB,$79,$BC,$BC,$BA,$3F,$17
	.dc.b	$F8,$E7,$F1,$7F,$DB,$7F,$14,$DC,$BF,$0E,$D7,$F9,$66,$48,$C4,$7E
	.dc.b	$4F,$E4,$DC,$3F,$C8,$7E,$9B,$F0,$31,$8D,$11,$89,$D5,$7E,$0E,$B5
	.dc.b	$3F,$A9,$EC,$FE,$55,$E4,$DC,$3F,$93,$DB,$B7,$F6,$33,$93,$FB,$DF
	.dc.b	$AF,$FD,$63,$D8,$7B,$1F,$E0,$0E,$1C,$93,$0F,$C2,$79,$0F,$7F,$FA
	.dc.b	$EB,$73,$FE,$C2,$DA,$FF,$7D,$36,$BF,$57,$F8,$F7,$F3,$FF,$CE,$3F
	.dc.b	$64,$D8,$CD,$A7,$2D,$8F,$CC,$7E,$1B,$F9,$72,$38,$3F,$03,$1F,$D1
	.dc.b	$8F,$BF,$D1,$F1,$23,$3B,$7F,$01,$8C,$47,$F5,$7E,$8F,$C6,$E7,$F8
	.dc.b	$C2,$47,$E4,$7A,$BF,$8D,$71,$7F,$57,$F6,$BE,$71,$27,$F5,$1C,$7F
	.dc.b	$C7,$A3,$E6,$E5,$27,$FC,$59,$C7,$F5,$38,$FE,$7B,$F5,$B8,$7E,$55
	.dc.b	$7C,$8E,$2F,$2D,$FF,$87,$A7,$2F,$D2,$8C,$EF,$EE,$3F,$C5,$91,$F0
	.dc.b	$B1,$F7,$46,$23,$F2,$7B,$FA,$5F,$64,$7E,$1B,$7F,$D6,$E3,$FA,$C2
	.dc.b	$4F,$CC,$FA,$4E,$64,$70,$72,$C3,$F6,$64,$70,$7C,$03,$F1,$E4,$30
	.dc.b	$7F,$4D,$FC,$E5,$F2,$5F,$E2,$E6,$E3,$FB,$96,$F1,$7F,$88,$1B,$5F
	.dc.b	$CD,$BC,$1F,$F0,$57,$C7,$FF,$3D,$6C,$66,$D3,$96,$C7,$F3,$EF,$A7
	.dc.b	$DB,$FB,$86,$BF,$E1,$5E,$E4,$62,$7F,$85,$BA,$9F,$F7,$47,$57,$F9
	.dc.b	$A1,$F7,$1F,$E2,$3E,$4F,$BE,$E0,$FF,$10,$65,$FD,$DB,$3F,$67,$99
	.dc.b	$F0,$49,$8F,$ED,$70,$E9,$7F,$20,$E8,$E1,$BC,$BF,$E5,$AF,$D0,$3F
	.dc.b	$AE,$E1,$FE,$46,$6C,$FF,$7B,$6B,$FE,$41,$6B,$EF,$FA,$6F,$F1,$A7
	.dc.b	$A7,$FD,$9D,$B1,$FD,$91,$AF,$F1,$E6,$BF,$1F,$DB,$FF,$2D,$6F,$FE
	.dc.b	$57,$8F,$ED,$71,$83,$FD,$95,$B7,$C4,$8F,$61,$FE,$7C,$70,$FB,$0F
	.dc.b	$57,$3F,$F9,$CB,$1F,$E7,$5A,$FF,$3D,$6E,$7A,$8F,$E8,$1E,$E7,$F5
	.dc.b	$E5,$CD,$FE,$4F,$C3,$FC,$C1,$39,$8C,$47,$F9,$4E,$31,$3F,$E2,$16
	.dc.b	$C7,$8C,$F3,$9F,$80,$DB,$FD,$C1,$97,$A2,$3F,$CB,$98,$BC,$0F,$41
	.dc.b	$30,$F4,$3F,$E8,$1E,$5F,$F3,$16,$BE,$29,$FF,$40,$C6,$1F,$13,$6F
	.dc.b	$D2,$D7,$E4,$70,$FE,$0D,$B9,$79,$23,$11,$E7,$3F,$24,$53,$D7,$97
	.dc.b	$A6,$FF,$C7,$FF,$15,$F1,$DF,$F2,$17,$A6,$8C,$57,$FA,$26,$31,$3F
	.dc.b	$E8,$B8,$C4,$E9,$FF,$3F,$71,$3E,$C7,$14,$8C,$3C,$67,$73,$A9,$CC
	.dc.b	$7D,$63,$D0,$7D,$32,$1F,$00,$FF,$52,$DF,$91,$F9,$8D,$AF,$5D,$F8
	.dc.b	$A3,$13,$C8,$EA,$73,$3E,$2E,$3F,$2B,$1F,$BD,$1A,$27,$C4,$7F,$AE
	.dc.b	$BE,$1B,$FD,$AB,$F5,$0F,$8A,$FF,$52,$4E,$69,$39,$7F,$DB,$71,$98
	.dc.b	$C4,$23,$13,$CB,$83,$FD,$E9,$3C,$A3,$17,$E0,$6C,$64,$EB,$3C,$2F
	.dc.b	$AE,$1F,$D4,$AB,$FC,$06,$4A,$FE,$AE,$D7,$F8,$6B,$5F,$C6,$BF,$3B
	.dc.b	$3F,$F5,$D3,$F3,$9B,$A7,$CE,$3E,$11,$63,$ED,$57,$86,$BB,$90,$EE
	.dc.b	$7F,$A4,$75,$FC,$2C,$FE,$95,$B5,$9D,$BB,$D7,$FB,$6E,$72,$FE,$4A
	.dc.b	$7C,$2D,$AF,$34,$62,$3F,$DA,$5B,$7E,$6A,$7C,$BB,$37,$FA,$75,$F0
	.dc.b	$0F,$31,$4F,$39,$0C,$12,$FF,$76,$3E,$E3,$F7,$DC,$9F,$C5,$5B,$C5
	.dc.b	$FE,$E3,$6C,$EC,$FE,$71,$C7,$FE,$E7,$F8,$DF,$E7,$8D,$C3,$32,$9C
	.dc.b	$B6,$3C,$E7,$F1,$7C,$FE,$D3,$4F,$E1,$B1,$FE,$BD,$7F,$91,$61,$FE
	.dc.b	$F5,$EB,$F1,$92,$7F,$A5,$74,$F8,$6B,$FA,$53,$5F,$E4,$2D,$87,$3F
	.dc.b	$F3,$8C,$7C,$3C,$5F,$81,$BF,$90,$78,$CF,$F3,$4F,$2F,$E3,$1F,$E9
	.dc.b	$37,$F4,$8F,$F3,$DC,$66,$37,$19,$EA,$FE,$18,$BF,$2D,$FE,$B4,$D9
	.dc.b	$FE,$B4,$D7,$FA,$D3,$5F,$E7,$3E,$B7,$A7,$2F,$F6,$3E,$5F,$DA,$E3
	.dc.b	$07,$BE,$7D,$07,$E0,$0A,$62,$9D,$DF,$EF,$4A,$FE,$FF,$1F,$56,$47
	.dc.b	$F8,$13,$5F,$91,$5B,$8F,$B3,$A7,$C7,$7B,$36,$BF,$C2,$63,$E9,$3F
	.dc.b	$DB,$DE,$02,$31,$5F,$E1,$F6,$BF,$E1,$F6,$BF,$E1,$FE,$0E,$15,$FE
	.dc.b	$27,$24,$7F,$8A,$DF,$F1,$97,$85,$E6,$25,$7A,$E7,$FC,$3B,$1F,$17
	.dc.b	$0F,$44,$7C,$46,$98,$F1,$7F,$D7,$BE,$FF,$FA,$F0,$93,$FE,$56,$7F
	.dc.b	$CB,$7A,$48,$FF,$3F,$B5,$FF,$3D,$79,$7E,$FF,$93,$FF,$62,$92,$3F
	.dc.b	$CB,$EF,$F9,$92,$A4,$7F,$A2,$1A,$FD,$21,$E5,$7F,$A7,$67,$FC,$E9
	.dc.b	$99,$A4,$9F,$F3,$EB,$FE,$AC,$23,$83,$FD,$0B,$E6,$7D,$3B,$FF,$5B
	.dc.b	$12,$33,$48,$FF,$6E,$F0,$C9,$FF,$3B,$B5,$FF,$3D,$CF,$8D,$FF,$50
	.dc.b	$DF,$FA,$00,$8F,$FB,$FF,$E7,$FF,$DF,$C4,$70,$7F,$AC,$3F,$38,$FF
	.dc.b	$57,$91,$C1,$FE,$8F,$AF,$B8,$FC,$4F,$37,$F4,$19,$FF,$B0,$71,$F8
	.dc.b	$A7,$FB,$30,$93,$FE,$A0,$6B,$F3,$7C,$E6,$9F,$5E,$3E,$0E,$7F,$ED
	.dc.b	$92,$3F,$ED,$FF,$4F,$F3,$5F,$3D,$FE,$E3,$CC,$D2,$4F,$FB,$CA,$1F
	.dc.b	$BC,$FF,$BD,$9B,$19,$8C,$1F,$EF,$E8,$F6,$F1,$3D,$8F,$DE,$36,$33
	.dc.b	$1F,$C6,$FB,$7F,$8D,$5F,$FB,$43,$5F,$F5,$79,$1F,$F5,$6F,$CF,$FE
	.dc.b	$2D,$FD,$EE,$0E,$4F,$DA,$4E,$65,$39,$C5,$39,$8C,$3B,$F8,$37,$C5
	.dc.b	$7E,$F7,$A7,$0D,$C3,$C1,$3A,$C3,$FD,$5B,$18,$9D,$3F,$9F,$F7,$9D
	.dc.b	$F8,$8F,$E2,$71,$8A,$E3,$1B,$9D,$29,$D6,$1E,$FA,$6A,$E1,$FE,$25
	.dc.b	$F2,$99,$8C,$4A,$31,$3A,$C3,$F6,$53,$CA,$31,$1F,$C9,$B1,$7C,$03
	.dc.b	$E0,$BF,$8F,$7E,$8F,$F3,$E7,$31,$8A,$FD,$E4,$63,$05,$6E,$F9,$46
	.dc.b	$2B,$C3,$93,$E1,$3F,$95,$7D,$07,$A5,$B3,$A6,$BF,$CE,$AB,$E0,$C6
	.dc.b	$23,$FD,$95,$8F,$F3,$3F,$D2,$7E,$13,$63,$BB,$5E,$58,$7D,$19,$FC
	.dc.b	$07,$37,$F4,$FE,$2F,$CA,$73,$63,$2F,$DA,$C6,$30,$FB,$D1,$8E,$3E
	.dc.b	$04,$7F,$77,$D7,$FB,$26,$F5,$FF,$5B,$8C,$61,$F9,$56,$DF,$FA,$0B
	.dc.b	$C1,$FE,$E0,$FF,$8A,$1B,$6F,$45,$18,$6C,$7F,$8A,$A5,$F1,$2F,$8A
	.dc.b	$FE,$D4,$D9,$FE,$1E,$DC,$7F,$4C,$DA,$F1,$4F,$F8,$0E,$FA,$F9,$DF
	.dc.b	$E8,$0C,$7C,$07,$9A,$9E,$E7,$F2,$DB,$F1,$36,$BF,$64,$D9,$FD,$C3
	.dc.b	$5F,$E8,$9F,$17,$FD,$12,$BF,$91,$B7,$2F,$DD,$36,$7F,$80,$35,$FE
	.dc.b	$99,$5C,$DF,$EF,$ED,$7C,$BE,$3F,$9E,$31,$C1,$FB,$9F,$9C,$FD,$EE
	.dc.b	$26,$92,$FF,$87,$7C,$7B,$E3,$FF,$C2,$8D,$AF,$E7,$2D,$CF,$F8,$83
	.dc.b	$6B,$F0,$FE,$37,$62,$38,$3A,$DF,$CB,$9E,$9A,$F2,$8C,$1F,$3F,$E4
	.dc.b	$3F,$45,$87,$F1,$D6,$DF,$F2,$36,$BF,$DB,$DA,$FF,$88,$97,$C8,$FF
	.dc.b	$8B,$5B,$97,$F5,$26,$D7,$59,$FF,$27,$7B,$0F,$57,$F2,$6D,$BF,$F2
	.dc.b	$43,$72,$FE,$2C,$DC,$7F,$9B,$7C,$9B,$E4,$FF,$C7,$8D,$BF,$EC,$2D
	.dc.b	$CB,$F9,$C3,$6B,$FD,$CB,$D3,$FE,$5F,$E6,$FD,$3C,$7F,$B9,$63,$07
	.dc.b	$F9,$96,$5F,$28,$F9,$5F,$D5,$C6,$1B,$19,$C7,$FD,$23,$39,$7F,$55
	.dc.b	$F2,$CF,$89,$FF,$29,$36,$3F,$BB,$B7,$3F,$E9,$CD,$CB,$FD,$F7,$F2
	.dc.b	$FE,$72,$57,$F8,$E4,$FD,$D3,$FE,$39,$F8,$1F,$F5,$17,$CC,$FF,$57
	.dc.b	$F6,$11,$8B,$FE,$BD,$18,$F8,$25,$6E,$B5,$5F,$E2,$DF,$99,$FE,$81
	.dc.b	$8F,$F0,$36,$F2,$7F,$63,$6C,$FF,$69,$6B,$C7,$CF,$FD,$E4,$62,$3F
	.dc.b	$CD,$2D,$4D,$24,$70,$25,$78,$5B,$5E,$16,$D7,$85,$AF,$86,$78,$12
	.dc.b	$78,$36,$38,$35,$E1,$1A,$23,$A6,$BA,$30,$43,$04,$3D,$F3,$E4,$6C
	.dc.b	$F7,$6C,$F4,$69,$CC,$A7,$33,$E3,$CF,$36,$CF,$36,$CF,$36,$9C,$CA
	.dc.b	$73,$AF,$A5,$F1,$3F,$B3,$F8,$9F,$F0,$F4,$A3,$0F,$F8,$85,$AF,$F8
	.dc.b	$9D,$AF,$F7,$8F,$A0,$FF,$4D,$3F,$E2,$7F,$31,$18,$8F,$F0,$03,$5E
	.dc.b	$18,$F8,$27,$6D,$9D,$4F,$F8,$2D,$AF,$F8,$89,$B3,$C2,$3C,$11,$82
	.dc.b	$4F,$32,$46,$9A,$F9,$1A,$F9,$0D,$10,$D1,$4D,$18,$21,$82,$1E,$97
	.dc.b	$AB,$4E,$45,$38,$94,$E2,$7B,$DE,$2D,$38,$94,$E2,$53,$8C,$FC,$57
	.dc.b	$2D,$73,$3F,$19,$E7,$19,$7C,$44,$3A,$15,$D3,$82,$38,$DB,$F8,$76
	.dc.b	$1F,$C3,$76,$34,$95,$FE,$22,$7F,$B4,$91,$C1,$FE,$75,$F9,$C7,$CE
	.dc.b	$FF,$B9,$E7,$FC,$E9,$18,$CB,$AD,$6E,$B4,$D8,$CB,$FE,$C9,$E5,$7D
	.dc.b	$76,$A3,$0D,$7F,$DF,$F1,$A9,$E3,$18,$9D,$2F,$13,$FC,$F5,$F3,$BF
	.dc.b	$E0,$22,$47,$F8,$F9,$FF,$45,$7A,$DF,$DF,$DB,$E1,$FF,$C0,$16,$E0
	.dc.b	$69,$FE,$73,$EA,$43,$E4,$30,$DC,$3F,$CC,$73,$A9,$C9,$FE,$14,$88
	.dc.b	$7F,$94,$0A,$7F,$96,$BD,$5F,$D7,$D7,$FA,$11,$BE,$9F,$F9,$EA,$3F
	.dc.b	$D4,$46,$4A,$FD,$1A,$FF,$44,$6D,$E1,$BF,$F3,$AB,$5F,$F3,$3B,$51
	.dc.b	$F2,$7D,$1F,$DE,$EB,$FE,$A3,$8F,$29,$A7,$0F,$FA,$03,$E9,$FC,$A4
	.dc.b	$C7,$C6,$FF,$08,$E0,$7C,$23,$EC,$FE,$4F,$DA,$7F,$82,$A3,$16,$8C
	.dc.b	$4E,$B0,$FC,$35,$6A,$7C,$71,$E7,$AF,$E4,$F8,$F9,$23,$D6,$64,$E7
	.dc.b	$0F,$92,$A4,$8E,$B8,$7D,$C8,$FF,$67,$7E,$BA,$FC,$2B,$F8,$FB,$6F
	.dc.b	$F9,$AB,$6F,$73,$F2,$35,$95,$E4,$6C,$EA,$38,$CF,$42,$3F,$69,$FB
	.dc.b	$E4,$70,$73,$24,$7C,$2C,$7E,$03,$F9,$4E,$0F,$F1,$13,$FE,$C9,$F1
	.dc.b	$FF,$12,$4A,$FC,$3F,$CD,$FE,$6B,$E8,$7F,$D9,$CD,$AF,$F5,$DB,$5F
	.dc.b	$F5,$6B,$5E,$DF,$11,$6B,$FD,$B2,$DC,$3F,$D0,$17,$FE,$D9,$8C,$7D
	.dc.b	$BF,$EA,$3F,$2A,$FF,$6E,$37,$0F,$F6,$4B,$5F,$F5,$DB,$5F,$67,$C1
	.dc.b	$DC,$7F,$DC,$ED,$C7,$FD,$A7,$7F,$EE,$78,$C7,$C1,$7E,$DB,$E8,$7F
	.dc.b	$DD,$4D,$C3,$FD,$AE,$D7,$FD,$92,$D7,$F4,$7E,$1F,$F8,$D7,$EC,$7F
	.dc.b	$94,$12,$3F,$C7,$3E,$5B,$C5,$4F,$FB,$EF,$F2,$2F,$F6,$E3,$6B,$F8
	.dc.b	$3B,$75,$DC,$7F,$BE,$27,$6D,$AF,$E0,$31,$D6,$7F,$7E,$D9,$ED,$3F
	.dc.b	$6D,$FB,$D3,$FC,$0B,$8B,$11,$D7,$E7,$FF,$8B,$F9,$3E,$56,$C6,$6B
	.dc.b	$FC,$0A,$D9,$F5,$7D,$3E,$A3,$FD,$ED,$8A,$31,$1F,$C8,$A3,$15,$CA
	.dc.b	$37,$3A,$87,$0F,$E9,$E7,$F6,$B1,$8E,$7F,$CF,$47,$EE,$0F,$CD,$3F
	.dc.b	$E0,$ED,$BF,$AC,$C7,$D0,$D3,$FA,$0A,$8C,$BF,$C0,$1A,$FF,$97,$3E
	.dc.b	$D1,$C9,$FD,$2F,$CD,$46,$31,$FE,$97,$18,$9D,$4E,$7C,$BF,$B9,$E5
	.dc.b	$7B,$F0,$FB,$0D,$7E,$2B,$5F,$88,$DC,$1B,$3E,$77,$0D,$8C,$D7,$E0
	.dc.b	$E7,$26,$BE,$1F,$FB,$14,$F2,$6D,$1A,$63,$E9,$A4,$FF,$66,$3C,$2E
	.dc.b	$31,$FC,$B7,$87,$C8,$93,$BB,$FE,$64,$FC,$ED,$BF,$D2,$DE,$5F,$F8
	.dc.b	$24,$99,$FD,$37,$FD,$21,$AF,$CB,$6E,$DF,$DA,$E7,$FD,$35,$39,$6C
	.dc.b	$65,$FC,$6E,$95,$FC,$63,$70,$F8,$07,$F7,$38,$FD,$44,$6F,$7F,$B1
	.dc.b	$3F,$E9,$A9,$EA,$D7,$FD,$3F,$7F,$7E,$7C,$B3,$99,$FC,$D7,$D4,$FF
	.dc.b	$7C,$F5,$BF,$74,$D8,$FD,$B3,$5F,$D1,$B5,$FC,$76,$3F,$E0,$4F,$23
	.dc.b	$A9,$23,$C6,$FF,$8C,$58,$7E,$6C,$A7,$EB,$BE,$87,$F8,$09,$30,$FE
	.dc.b	$02,$FF,$04,$FB,$BF,$F0,$8F,$EC,$BF,$C3,$DB,$5F,$C2,$DA,$FE,$E1
	.dc.b	$AF,$ED,$FE,$53,$F8,$69,$2F,$F8,$6B,$FA,$FE,$FF,$E3,$93,$E4,$3F
	.dc.b	$91,$B8,$7D,$07,$F1,$9F,$BE,$FE,$2E,$4A,$FE,$2E,$F3,$E9,$FF,$4D
	.dc.b	$F1,$FF,$B7,$EB,$A1,$FB,$9B,$E8,$FF,$0E,$F3,$3F,$90,$93,$0F,$E4
	.dc.b	$2F,$F2,$4F,$83,$FF,$02,$4F,$F1,$46,$BF,$C4,$DB,$3F,$B0,$6F,$33
	.dc.b	$E0,$7F,$96,$92,$7F,$96,$BE,$7A,$F2,$E3,$E7,$7D,$E7,$F2,$2F,$6F
	.dc.b	$F9,$A9,$2B,$F9,$AB,$FC,$E3,$F0,$8F,$F2,$2F,$37,$F4,$36,$D7,$F3
	.dc.b	$F6,$BF,$CB,$1A,$FC,$ED,$7E,$3E,$BF,$12,$30,$7D,$3C,$3F,$85,$F9
	.dc.b	$DF,$EB,$E7,$FA,$53,$63,$F9,$BB,$5F,$E6,$CD,$8F,$E2,$7F,$47,$FE
	.dc.b	$5A,$FC,$DB,$FD,$89,$18,$0E,$36,$4E,$E7,$FA,$C3,$6F,$F9,$86,$1E
	.dc.b	$77,$EE,$63,$EA,$23,$FD,$97,$E3,$3F,$CD,$37,$D5,$AF,$F6,$36,$C7
	.dc.b	$F5,$06,$BF,$E7,$0F,$B1,$E3,$F4,$BE,$86,$D7,$F4,$C6,$CF,$F9,$72
	.dc.b	$11,$88,$E5,$F2,$5F,$E7,$59,$E0,$D8,$E1,$28,$C4,$70,$9F,$F4,$01
	.dc.b	$F8,$47,$EA,$94,$F8,$27,$F5,$B7,$C2,$7D,$19,$F8,$31,$E9,$8C,$1F
	.dc.b	$5D,$E2,$73,$E1,$FC,$93,$F3,$B1,$F9,$6E,$72,$FA,$13,$A8,$FF,$3D
	.dc.b	$FC,$0B,$E2,$BF,$D6,$7E,$C3,$E0,$FF,$CC,$B3,$9F,$1B,$FD,$03,$39
	.dc.b	$BE,$37,$C2,$F6,$D8,$CD,$70,$9C,$B6,$32,$7F,$9B,$E3,$2D,$3F,$3C
	.dc.b	$FF,$B4,$C8,$FF,$7D,$6B,$EF,$6B,$C0,$FB,$D9,$FF,$AC,$A3,$E5,$3E
	.dc.b	$D3,$25,$8F,$B6,$EC,$EC,$C3,$B1,$4F,$BD,$F0,$DF,$8D,$DB,$FD,$7E
	.dc.b	$D9,$34,$F0,$F4,$48,$F0,$B6,$BF,$35,$1F,$EB,$96,$FB,$FF,$A1,$8C
	.dc.b	$47,$F8,$AF,$8B,$FD,$9B,$3F,$EC,$E8,$F2,$90,$F8,$45,$9F,$DC,$7E
	.dc.b	$63,$FB,$7E,$4F,$F6,$AB,$60,$D3,$F7,$9C,$FF,$0E,$BF,$83,$CF,$C7
	.dc.b	$7F,$83,$B4,$FE,$1E,$65,$FD,$EE,$9F,$51,$A7,$C9,$8F,$F0,$FF,$47
	.dc.b	$FB,$9B,$ED,$FF,$91,$41,$A4,$9F,$F3,$17,$5F,$DB,$6C,$FE,$4D,$FE
	.dc.b	$60,$D7,$FC,$A1,$1B,$7E,$21,$E2,$2F,$23,$97,$FC,$ED,$8F,$F9,$D7
	.dc.b	$7E,$15,$FE,$59,$6E,$1F,$E3,$66,$F4,$FF,$47,$8F,$01,$FD,$2C,$7C
	.dc.b	$47,$F9,$FF,$B5,$18,$BF,$F3,$E4,$62,$75,$39,$9F,$EB,$CF,$4A,$FF
	.dc.b	$70,$E1,$FE,$89,$EE,$F0,$9E,$6A,$F6,$1F,$80,$21,$8A,$FF,$26,$F7
	.dc.b	$FA,$08,$F5,$79,$69,$F0,$FA,$FE,$55,$74,$24,$67,$0F,$F4,$D7,$81
	.dc.b	$F0,$89,$87,$9D,$F8,$BE,$0F,$C0,$FD,$83,$FC,$25,$26,$93,$0F,$F4
	.dc.b	$F7,$89,$F3,$09,$3F,$05,$F2,$F0,$FD,$69,$FD,$91,$E4,$70,$FF,$8E
	.dc.b	$BC,$6F,$9F,$F9,$5F,$F8,$E8,$91,$FE,$40,$7F,$C7,$36,$FC,$CE,$92
	.dc.b	$7F,$C7,$71,$97,$FC,$8F,$CD,$FE,$44,$23,$83,$FC,$9F,$0F,$93,$FF
	.dc.b	$5F,$79,$BF,$80,$6C,$66,$77,$1F,$E4,$E2,$3F,$E5,$46,$BF,$CE,$DF
	.dc.b	$F5,$03,$FD,$A1,$BC,$1F,$5D,$B9,$FF,$B3,$DF,$F3,$CD,$3E,$32,$92
	.dc.b	$FF,$D9,$9E,$87,$E5,$23,$FD,$A9,$E8,$D9,$FF,$68,$35,$FF,$5E,$B5
	.dc.b	$FF,$5E,$36,$BF,$CE,$33,$EE,$24,$FC,$47,$FD,$C1,$E9,$7E,$8B,$F1
	.dc.b	$2B,$46,$92,$BF,$D5,$31,$96,$C7,$FA,$BA,$31,$1F,$EA,$C6,$BF,$E6
	.dc.b	$E7,$E0,$3F,$E7,$EA,$FF,$3D,$91,$FF,$63,$3F,$23,$9B,$71,$FD,$A6
	.dc.b	$30,$7F,$75,$F5,$9E,$B7,$F0,$26,$DF,$F0,$06,$F4,$6E,$7F,$82,$FA
	.dc.b	$FF,$C0,$39,$7F,$DC,$F7,$FE,$CE,$C7,$FD,$F9,$39,$6C,$65,$FF,$75
	.dc.b	$FB,$39,$6C,$65,$FF,$76,$46,$5B,$19,$3F,$DD,$7E,$D7,$9C,$8F,$9D
	.dc.b	$FE,$33,$A4,$8F,$E1,$CD,$CB,$FC,$29,$DD,$FE,$25,$FC,$1A,$E1,$FC
	.dc.b	$69,$B6,$69,$F2,$C9,$1F,$EF,$36,$BF,$E2,$FD,$5E,$EF,$F2,$4F,$A0
	.dc.b	$F9,$CD,$9F,$F8,$03,$3A,$B5,$69,$B3,$AF,$80,$FE,$5B,$F9,$67,$D7
	.dc.b	$6C,$66,$53,$96,$C7,$6E,$0D,$BE,$56,$BF,$CE,$30,$46,$25,$18,$79
	.dc.b	$CF,$B7,$E0,$FF,$9F,$E5,$F8,$96,$C7,$B1,$AE,$E7,$E1,$BF,$03,$E1
	.dc.b	$72,$F8,$5B,$1E,$06,$D9,$BC,$1F,$7B,$7F,$CC,$36,$C4,$1F,$0C,$F3
	.dc.b	$3F,$A5,$37,$1F,$E8,$2D,$BF,$E2,$2D,$AF,$AB,$CB,$FA,$32,$57,$E8
	.dc.b	$DF,$EA,$9F,$0F,$FA,$49,$F0,$07,$0F,$8D,$AF,$99,$AF,$9B,$0F,$29
	.dc.b	$2B,$CA,$F6,$A9,$1F,$D6,$2B,$89,$D3,$DD,$7C,$43,$CA,$FE,$E7,$F1
	.dc.b	$3E,$49,$CC,$63,$2F,$8D,$18,$C1,$5B,$BE,$51,$8A,$F6,$E7,$F1,$5F
	.dc.b	$9C,$D7,$E8,$B5,$F6,$47,$F8,$5E,$FB,$79,$9D,$71,$FF,$02,$57,$8B
	.dc.b	$0E,$F3,$9B,$FF,$03,$4F,$E5,$71,$FE,$28,$48,$FE,$28,$FC,$7D,$7E
	.dc.b	$CE,$B6,$BC,$EF,$F1,$BF,$8A,$FF,$2C,$C7,$EC,$1A,$FE,$DD,$AF,$E8
	.dc.b	$DB,$7F,$A2,$F5,$3F,$92,$11,$FE,$48,$F7,$9F,$87,$55,$F8,$6F,$F2
	.dc.b	$BF,$8C,$E7,$C6,$69,$2F,$FC,$D7,$5F,$8A,$F8,$E7,$C7,$7F,$6D,$8C
	.dc.b	$35,$FC,$F4,$6A,$7F,$B8,$46,$27,$4F,$8A,$32,$DF,$1B,$E3,$C6,$23
	.dc.b	$EB,$C7,$F7,$2E,$17,$C4,$7F,$63,$94,$62,$BF,$A5,$C7,$EC,$23,$1E
	.dc.b	$1F,$1A,$E1,$F1,$ED,$E9,$30,$E3,$82,$75,$18,$E5,$FD,$B6,$1E,$62
	.dc.b	$57,$DC,$9E,$3E,$57,$F8,$EE,$7F,$AD,$CF,$F6,$02,$1F,$C0,$FE,$41
	.dc.b	$F1,$7F,$E4,$B6,$DF,$F9,$09,$BC,$9F,$E1,$26,$D7,$EF,$1A,$F3,$AE
	.dc.b	$44,$AE,$4D,$AE,$4D,$9E,$4D,$79,$4F,$02,$4F,$06,$C7,$06,$BC,$23
	.dc.b	$44,$74,$D7,$46,$08,$60,$87,$F2,$8A,$F6,$B6,$BC,$CD,$AF,$0B,$6B
	.dc.b	$B3,$4E,$A5,$3A,$8A,$EA,$DA,$EA,$DA,$EA,$DA,$EA,$D3,$A9,$4E,$B5
	.dc.b	$D0,$95,$D1,$B5,$D3,$E0,$BF,$DF,$D5,$FE,$4E,$6B,$FC,$E1,$AF,$FB
	.dc.b	$D8,$FC,$07,$C9,$7F,$B4,$5B,$3F,$E5,$67,$D9,$19,$7F,$1C,$D7,$E6
	.dc.b	$35,$DC,$71,$A4,$E5,$B1,$98,$C1,$1C,$12,$7F,$2D,$19,$6C,$65,$F3
	.dc.b	$98,$21,$82,$1E,$87,$83,$4D,$94,$D9,$EC,$76,$D3,$65,$36,$7C,$03
	.dc.b	$25,$33,$3F,$01,$CB,$5C,$CF,$B5,$B3,$E0,$79,$1E,$52,$1E,$12,$9A
	.dc.b	$30,$43,$04,$9F,$03,$C1,$AE,$DA,$EE,$7E,$E6,$5F,$EC,$BF,$81,$FF
	.dc.b	$37,$12,$7F,$C0,$71,$97,$F8,$C4,$7F,$30,$23,$83,$FB,$B7,$B5,$FE
	.dc.b	$EA,$8F,$E6,$4D,$8C,$E0,$9C,$B6,$3B,$FA,$0F,$29,$F1,$9F,$E8,$12
	.dc.b	$4E,$A3,$2F,$F8,$63,$87,$F9,$8C,$FF,$AC,$9B,$19,$E2,$FF,$7E,$CE
	.dc.b	$5B,$1F,$E2,$FC,$1F,$0A,$F8,$54,$3E,$47,$FC,$4A,$EA,$33,$18,$9E
	.dc.b	$13,$BA,$E9,$18,$9F,$F4,$D4,$43,$D4,$53,$FC,$81,$F3,$6F,$AF,$FD
	.dc.b	$83,$6F,$F7,$0D,$9F,$F6,$7B,$5E,$FA,$F9,$1B,$3E,$37,$FC,$9A,$FB
	.dc.b	$A3,$51,$8C,$11,$88,$FF,$29,$F0,$63,$6F,$E1,$0D,$8F,$41,$A9,$FB
	.dc.b	$0F,$91,$F4,$92,$FD,$2F,$F0,$BF,$D1,$DF,$0B,$F3,$BD,$8F,$F7,$4C
	.dc.b	$62,$91,$89,$D5,$FF,$95,$AB,$53,$E2,$86,$1E,$57,$E6,$90,$FF,$36
	.dc.b	$4F,$5D,$BA,$61,$F7,$E7,$A6,$1F,$80,$9C,$9E,$4F,$5F,$D0,$4A,$FE
	.dc.b	$55,$68,$FD,$5E,$5F,$0B,$0F,$01,$F1,$73,$F9,$31,$98,$F1,$CF,$37
	.dc.b	$B5,$F4,$8F,$E6,$4F,$E5,$36,$FA,$8D,$3E,$E4,$FE,$5B,$5F,$C3,$6F
	.dc.b	$F9,$3A,$FB,$53,$F1,$30,$F0,$3E,$33,$E4,$6B,$F1,$DF,$F4,$86,$8F
	.dc.b	$13,$FA,$03,$72,$FE,$90,$D9,$FF,$43,$9F,$A7,$9F,$3E,$5F,$E3,$BF
	.dc.b	$94,$FD,$27,$83,$FA,$A6,$E0,$69,$FA,$9F,$A7,$D6,$1F,$A8,$6B,$F9
	.dc.b	$56,$BE,$F6,$CF,$EE,$3B,$3F,$6F,$F2,$BF,$B3,$6D,$7E,$C9,$AF,$E7
	.dc.b	$DA,$FC,$9C,$6D,$7F,$02,$69,$FC,$11,$57,$EF,$E3,$38,$7F,$06,$F9
	.dc.b	$5F,$E0,$4D,$BF,$E0,$2D,$7F,$5C,$D7,$FD,$7F,$F6,$2F,$BA,$C4,$23
	.dc.b	$18,$AF,$E7,$C6,$1B,$5C,$E7,$F8,$CF,$83,$71,$34,$91,$CF,$0F,$E2
	.dc.b	$B5,$F8,$CA,$EF,$C7,$FC,$6E,$30,$7E,$2B,$83,$F0,$15,$FE,$C6,$6B
	.dc.b	$F8,$8E,$0F,$C3,$9F,$82,$FA,$BF,$E5,$3F,$A4,$7F,$25,$6C,$FF,$24
	.dc.b	$6B,$FC,$45,$AF,$E6,$FE,$FE,$E7,$FC,$D1,$B9,$9A,$63,$F0,$1F,$51
	.dc.b	$27,$F4,$A7,$F3,$C7,$F4,$7C,$BF,$9F,$BF,$D1,$1F,$B6,$3F,$4D,$D3
	.dc.b	$F9,$EC,$FD,$77,$49,$3F,$8D,$5E,$6D,$24,$7E,$3E,$32,$FF,$42,$CB
	.dc.b	$FA,$5C,$FE,$26,$73,$3F,$CC,$67,$2E,$E3,$2F,$EF,$7F,$02,$FA,$C4
	.dc.b	$8F,$6C,$A3,$EB,$5F,$A2,$BF,$6B,$B7,$D6,$47,$ED,$EF,$F7,$27,$F6
	.dc.b	$0F,$AB,$FC,$11,$1E,$AF,$C8,$EA,$FE,$13,$3F,$18,$FE,$12,$E2,$BB
	.dc.b	$47,$53,$C3,$7E,$37,$FB,$1F,$D5,$FE,$3C,$98,$7A,$9F,$4B,$F3,$A3
	.dc.b	$F1,$D3,$C5,$FE,$AD,$E9,$FE,$5C,$93,$F0,$9E,$31,$E0,$8E,$4F,$E5
	.dc.b	$5F,$E6,$3A,$C8,$F4,$BF,$0B,$4F,$F1,$3C,$FF,$80,$BC,$8F,$E3,$64
	.dc.b	$75,$F4,$FF,$E0,$4F,$9B,$FF,$00,$37,$0F,$EF,$ED,$7F,$B6,$35,$FF
	.dc.b	$05,$FC,$FF,$FA,$E1,$FF,$05,$B5,$FF,$09,$35,$FE,$EE,$D7,$F2,$E7
	.dc.b	$DA,$3E,$91,$F9,$9E,$2F,$CE,$5B,$F0,$4F,$F4,$C3,$F7,$9B,$3F,$E1
	.dc.b	$A6,$BF,$E2,$16,$BD,$67,$F4,$CF,$63,$1C,$1F,$82,$8E,$F2,$FA,$BF
	.dc.b	$EA,$ED,$BF,$04,$60,$CF,$AB,$FE,$2C,$F5,$3C,$4D,$C7,$C4,$D7,$FC
	.dc.b	$2F,$68,$C4,$7F,$8D,$21,$F6,$DE,$02,$61,$EF,$7F,$B9,$F0,$7D,$FA
	.dc.b	$F8,$D9,$FF,$62,$AF,$8D,$F7,$7E,$82,$3F,$1D,$F5,$FC,$F7,$F7,$0C
	.dc.b	$4D,$24,$FF,$72,$BF,$8C,$7D,$62,$18,$7E,$A7,$D9,$FF,$93,$23,$FD
	.dc.b	$69,$18,$3F,$C6,$AF,$C0,$8D,$47,$D5,$6C,$7C,$4F,$B4,$FF,$28,$CF
	.dc.b	$D9,$6C,$66,$3F,$CB,$06,$4A,$FD,$AC,$BE,$23,$4F,$F1,$DC,$7E,$27
	.dc.b	$EE,$3B,$C7,$CF,$6C,$64,$F3,$C6,$5A,$7E,$47,$E8,$7F,$CD,$38,$FB
	.dc.b	$B2,$46,$27,$F2,$0D,$AF,$BE,$D7,$EF,$3E,$09,$E3,$3D,$9B,$3D,$67
	.dc.b	$CE,$FD,$22,$70,$FE,$68,$FC,$BC,$73,$25,$73,$78,$C6,$F5,$FC,$F1
	.dc.b	$32,$F8,$F7,$F7,$27,$51,$FE,$EB,$F6,$3F,$DE,$BD,$7F,$DC,$2B,$F9
	.dc.b	$7C,$63,$C5,$FD,$C3,$63,$38,$FF,$A8,$6B,$84,$27,$4D,$9D,$57,$26
	.dc.b	$C7,$2F,$5B,$F9,$ED,$F2,$C7,$FC,$F1,$7F,$E3,$66,$D7,$FA,$A6,$31
	.dc.b	$97,$FA,$BD,$B9,$78,$A7,$FC,$B1,$18,$23,$F6,$33,$FF,$06,$96,$3E
	.dc.b	$93,$88,$FB,$6D,$79,$9E,$12,$1F,$93,$8F,$E2,$99,$7F,$AC,$1E,$64
	.dc.b	$9E,$C7,$87,$D7,$EE,$47,$07,$F1,$BD,$F9,$34,$F1,$1F,$99,$9F,$CA
	.dc.b	$46,$AF,$C4,$79,$0F,$39,$83,$F5,$9F,$41,$F9,$BE,$1F,$F6,$2B,$6C
	.dc.b	$D3,$F4,$1C,$5D,$6B,$F8,$C3,$67,$F8,$06,$1A,$7E,$34,$FF,$84,$A3
	.dc.b	$E7,$BE,$13,$C6,$7F,$21,$E2,$F0,$BF,$CE,$5A,$FF,$28,$6B,$B9,$F2
	.dc.b	$BF,$CF,$F8,$FE,$73,$63,$F0,$07,$F2,$42,$F1,$A7,$73,$E3,$3C,$C7
	.dc.b	$C7,$29,$FC,$E3,$93,$6F,$F3,$76,$D7,$F5,$56,$D7,$F3,$A8,$EF,$CB
	.dc.b	$FE,$F0,$F9,$6F,$E8,$B0,$69,$2F,$FA,$C7,$47,$22,$4F,$A1,$F0,$C7
	.dc.b	$90,$F9,$25,$31,$1F,$D9,$FD,$AF,$19,$1F,$82,$F2,$E5,$FD,$4C,$FC
	.dc.b	$16,$BF,$52,$5D,$4F,$3F,$DE,$D8,$F9,$AD,$DF,$FC,$03,$3F,$13,$BD
	.dc.b	$DA,$B7,$CC,$7F,$83,$9B,$8F,$F8,$39,$AF,$F8,$39,$AF,$FB,$0F,$C0
	.dc.b	$E8,$48,$FB,$B6,$8F,$67,$5F,$F0,$1A,$FB,$5A,$7B,$1F,$8D,$D9,$FE
	.dc.b	$C1,$D3,$FC,$58,$48,$FF,$1A,$E3,$D3,$C1,$FC,$7A,$C9,$18,$6C,$7E
	.dc.b	$4A,$7F,$C7,$6D,$7F,$8C,$F8,$5F,$63,$0F,$04,$76,$C3,$C0,$FF,$96
	.dc.b	$AB,$B7,$87,$D4,$8F,$33,$27,$58,$FF,$30,$4F,$27,$BC,$7F,$1F,$72
	.dc.b	$7C,$2C,$FE,$23,$F7,$DA,$FF,$BB,$FC,$4F,$61,$27,$C1,$92,$3F,$63
	.dc.b	$3D,$0F,$F2,$95,$F8,$5C,$1F,$E7,$DA,$E3,$E2,$FC,$72,$3D,$2F,$F0
	.dc.b	$9F,$9F,$FF,$A2,$49,$5F,$E8,$77,$FD,$28,$47,$FD,$24,$D7,$C2,$7C
	.dc.b	$B2,$18,$3D,$DE,$17,$C3,$DF,$F0,$47,$E7,$AF,$FD,$54,$7B,$BC,$8F
	.dc.b	$B8,$47,$E0,$F9,$3F,$99,$FD,$EF,$FC,$AD,$87,$FA,$EE,$51,$87,$FD
	.dc.b	$78,$62,$3F,$CB,$8E,$23,$FD,$90,$E6,$3F,$D8,$CD,$7F,$4D,$93,$CB
	.dc.b	$F0,$47,$FB,$36,$74,$FC,$E2,$70,$FE,$31,$F8,$07,$2F,$37,$F2,$C4
	.dc.b	$9F,$B6,$FE,$67,$7F,$DA,$7C,$87,$A7,$1F,$EC,$2D,$8C,$DF,$AA,$72
	.dc.b	$FC,$2F,$47,$FD,$F7,$5F,$EF,$8C,$0D,$24,$F1,$F4,$B9,$BF,$9F,$B4
	.dc.b	$62,$3E,$0B,$6B,$F6,$0D,$9F,$D9,$43,$E3,$FF,$DA,$04,$AF,$F6,$73
	.dc.b	$FE,$DB,$23,$FE,$D9,$6B,$E1,$F5,$7F,$8A,$7C,$77,$C1,$6C,$7B,$DA
	.dc.b	$EE,$93,$F1,$FD,$7C,$CF,$D8,$6C,$FF,$07,$6E,$07,$73,$F6,$3A,$BF
	.dc.b	$82,$12,$BF,$2D,$D3,$FB,$49,$FE,$79,$CB,$F9,$88,$FD,$BC,$FF,$13
	.dc.b	$F6,$3F,$89,$11,$FD,$BB,$FB,$6F,$69,$ED,$7F,$48,$6C,$FF,$05,$6F
	.dc.b	$17,$E7,$31,$46,$1F,$E1,$1D,$3E,$42,$4E,$E7,$53,$FE,$B9,$6B,$FE
	.dc.b	$F4,$6B,$D2,$FC,$5E,$DF,$F4,$3F,$87,$E4,$D9,$FD,$6D,$23,$11,$D5
	.dc.b	$AF,$F6,$0D,$3E,$26,$3E,$48,$E3,$3F,$D3,$FC,$0E,$B9,$FF,$38,$9C
	.dc.b	$B5,$E8,$D4,$3D,$E6,$DF,$DC,$1B,$3F,$CB,$5B,$AF,$C9,$6D,$7C,$88
	.dc.b	$FF,$0F,$BF,$DA,$F8,$78,$FA,$1F,$AF,$C3,$FA,$51,$30,$FE,$79,$83
	.dc.b	$83,$C0,$48,$F0,$3F,$D7,$B4,$F8,$18,$FF,$1E,$CB,$F7,$5D,$5F,$D5
	.dc.b	$49,$7F,$CE,$9F,$E7,$3F,$02,$F8,$0F,$EE,$9F,$0B,$F8,$18,$C5,$23
	.dc.b	$13,$AC,$3F,$90,$D6,$A7,$FC,$DA,$48,$FE,$4B,$C3,$F2,$32,$E5,$18
	.dc.b	$D5,$18,$8F,$F8,$01,$ED,$3D,$7F,$F0,$37,$C1,$FE,$4E,$31,$A2,$31
	.dc.b	$3A,$AF,$E6,$75,$A9,$FC,$D7,$C2,$3E,$1F,$55,$F7,$04,$67,$0E,$91
	.dc.b	$FC,$F2,$31,$5C,$AB,$8C,$FE,$13,$3F,$64,$FF,$A3,$09,$1E,$DC,$B9
	.dc.b	$BD,$FD,$0F,$F0,$A6,$3F,$DC,$DB,$C3,$FA,$F6,$CF,$F4,$C6,$BC,$7C
	.dc.b	$CF,$F1,$FC,$62,$3F,$2A,$D4,$D2,$47,$02,$57,$85,$B5,$E1,$6D,$78
	.dc.b	$5A,$F8,$67,$81,$27,$83,$63,$83,$5E,$11,$A2,$3A,$6B,$A3,$04,$30
	.dc.b	$43,$DF,$3E,$46,$CF,$76,$CF,$46,$9C,$CA,$73,$3E,$3C,$F3,$6C,$F3
	.dc.b	$6C,$F3,$69,$CC,$A7,$3A,$F9,$F3,$C5,$B3,$C5,$B3,$C7,$DD,$7B,$8F
	.dc.b	$77,$FD,$DC,$D9,$FE,$65,$3A,$9C,$9F,$CD,$61,$F4,$EF,$88,$FF,$36
	.dc.b	$47,$F3,$98,$C6,$5B,$6C,$66,$76,$FF,$86,$A3,$2D,$8C,$BF,$A7,$8C
	.dc.b	$B6,$32,$FF,$26,$8C,$B6,$32,$F9,$9B,$1A,$8C,$4F,$96,$72,$47,$04
	.dc.b	$8F,$43,$5F,$33,$5F,$23,$5F,$13,$5E,$E7,$22,$1C,$8A,$72,$29,$C8
	.dc.b	$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$7E
	.dc.b	$51,$27,$E4,$47,$9A,$31,$3F,$05,$B1,$99,$DB,$E5,$8C,$BF,$C6,$6B
	.dc.b	$FC,$05,$B7,$F3,$96,$F1,$FF,$38,$6E,$DF,$97,$AF,$E5,$9F,$19,$FE
	.dc.b	$1D,$CB,$11,$FD,$AA,$31,$5F,$E0,$D6,$C7,$F9,$A9,$AF,$F4,$2E,$4F
	.dc.b	$F6,$34,$FF,$3E,$AE,$72,$E4,$FE,$87,$39,$3F,$9B,$4B,$E2,$29,$2B
	.dc.b	$F9,$E7,$27,$FB,$46,$3F,$CE,$DE,$DF,$E9,$BE,$16,$F0,$7F,$B5,$5B
	.dc.b	$81,$A7,$F8,$97,$EB,$3F,$D6,$CD,$C7,$81,$FE,$84,$8C,$B4,$C9,$FE
	.dc.b	$BD,$B7,$8D,$FA,$AB,$46,$1F,$F1,$91,$88,$D4,$66,$7D,$3D,$9D,$1F
	.dc.b	$F7,$64,$7F,$78,$C3,$37,$E9,$9F,$F4,$8C,$E5,$FF,$4C,$FD,$6F,$D7
	.dc.b	$AF,$F2,$03,$63,$3E,$47,$55,$19,$7D,$5E,$3F,$90,$98,$7A,$30,$47
	.dc.b	$F0,$6E,$07,$C6,$3E,$81,$6F,$8C,$FE,$F3,$39,$8C,$63,$FE,$04,$8C
	.dc.b	$4E,$AB,$8B,$FE,$57,$FB,$07,$C9,$F9,$DB,$87,$C1,$C3,$FC,$01,$18
	.dc.b	$DB,$85,$7C,$9C,$BC,$B8,$F9,$3C,$5E,$D8,$9A,$4C,$7F,$CA,$BF,$47
	.dc.b	$FC,$7B,$CD,$34,$95,$FC,$DF,$C9,$7D,$3F,$CB,$6C,$79,$6F,$FC,$9F
	.dc.b	$18,$D3,$FC,$95,$5F,$47,$0E,$B7,$FD,$05,$B7,$FE,$71,$C3,$BB,$F9
	.dc.b	$CE,$CF,$C2,$57,$E3,$1B,$C4,$69,$F7,$7C,$7F,$F1,$16,$5F,$AA,$25
	.dc.b	$FE,$A2,$FC,$C7,$E5,$BE,$53,$99,$1C,$E0,$8E,$FF,$2B,$F7,$BF,$00
	.dc.b	$FD,$7B,$6F,$F2,$6D,$7F,$48,$D7,$F3,$7C,$57,$0F,$D2,$B7,$0F,$BB
	.dc.b	$7F,$BA,$8C,$7C,$BF,$E3,$3F,$28,$FD,$EB,$70,$FD,$0B,$5F,$D9,$35
	.dc.b	$FD,$4F,$CC,$DC,$7F,$66,$DC,$4D,$31,$EB,$7E,$6B,$E7,$FF,$85,$B6
	.dc.b	$FF,$5E,$D7,$F8,$03,$5F,$F4,$6E,$FF,$E5,$19,$FE,$32,$48,$FA,$0F
	.dc.b	$F9,$5D,$FD,$3F,$EE,$3F,$E9,$4C,$53,$95,$19,$8C,$36,$7F,$CC,$13
	.dc.b	$A7,$C4,$F0,$F1,$FF,$0F,$3F,$E7,$48,$FB,$ED,$E1,$E7,$3A,$F9,$E9
	.dc.b	$5A,$6C,$F8,$3E,$7F,$F7,$BF,$79,$68,$D2,$63,$E4,$C5,$F3,$9C,$23
	.dc.b	$0D,$7E,$04,$6A,$B9,$C6,$9A,$FF,$2D,$E6,$FE,$1B,$F7,$DF,$A3,$93
	.dc.b	$49,$3F,$E7,$7C,$BF,$AD,$91,$C1,$FE,$87,$F2,$5F,$69,$FE,$7A,$9F
	.dc.b	$A9,$18,$F5,$7F,$99,$37,$0F,$DE,$36,$7E,$9F,$BD,$FC,$83,$EC,$ED
	.dc.b	$1A,$4C,$7E,$27,$D2,$ED,$B3,$B6,$CF,$DE,$9D,$47,$C6,$E2,$F8,$2F
	.dc.b	$CD,$6B,$9A,$F8,$6E,$5A,$FC,$EE,$2D,$BF,$90,$6C,$7E,$39,$BB,$9B
	.dc.b	$C7,$FC,$F3,$F2,$9F,$C8,$B6,$FE,$F3,$5F,$B6,$D7,$FD,$59,$E9,$FF
	.dc.b	$B1,$38,$7F,$C1,$A4,$8F,$F5,$63,$F4,$38,$7F,$D5,$6D,$E1,$F3,$F0
	.dc.b	$FD,$18,$C6,$9E,$CB,$FD,$4E,$BF,$D4,$3D,$0F,$CD,$CF,$F9,$A6,$7E
	.dc.b	$14,$60,$E3,$A4,$8F,$C1,$65,$FE,$BA,$8C,$47,$E5,$3C,$9F,$EB,$7F
	.dc.b	$A9,$7E,$C1,$B1,$F9,$D6,$BB,$C1,$3E,$6F,$CA,$A4,$F6,$D7,$FB,$2C
	.dc.b	$62,$3F,$DB,$1E,$B7,$FB,$AD,$FF,$18,$90,$FF,$68,$1F,$C0,$3E,$D3
	.dc.b	$F7,$64,$C3,$F7,$6F,$EF,$BE,$B3,$FB,$8F,$EE,$3F,$C1,$9B,$5F,$B3
	.dc.b	$6B,$FB,$06,$BF,$C4,$BE,$BB,$FC,$31,$CC,$69,$30,$FF,$4B,$63,$E7
	.dc.b	$AF,$29,$0F,$5F,$23,$8B,$FC,$3F,$83,$EB,$FF,$89,$93,$0F,$E2,$6F
	.dc.b	$F8,$4F,$0E,$98,$72,$BE,$F8,$73,$8F,$88,$7F,$08,$F3,$FF,$8E,$93
	.dc.b	$0F,$E3,$AF,$F2,$0F,$A9,$FF,$09,$FE,$E1,$FC,$A1,$B5,$FC,$55,$AF
	.dc.b	$F1,$26,$BF,$CD,$3E,$C7,$E1,$92,$FD,$6F,$FA,$8B,$2F,$3C,$FE,$7B
	.dc.b	$2F,$39,$1F,$E6,$FE,$F7,$F3,$22,$5F,$F3,$27,$F9,$AF,$9F,$FE,$28
	.dc.b	$FC,$B3,$F9,$EB,$6B,$F9,$53,$5F,$E4,$ED,$7C,$07,$C4,$29,$F4,$CF
	.dc.b	$E9,$BF,$67,$FD,$08,$97,$FD,$09,$FE,$8B,$EF,$FF,$8E,$3F,$2C,$FF
	.dc.b	$64,$CF,$F5,$08,$C3,$FC,$ED,$B5,$FC,$E1,$AF,$80,$FB,$07,$8C,$67
	.dc.b	$F1,$A7,$FA,$DF,$C7,$7F,$91,$B9,$7B,$F2,$23,$13,$FD,$25,$B7,$FD
	.dc.b	$19,$AF,$F6,$8F,$B6,$FF,$27,$FE,$59,$C9,$B5,$FD,$45,$B3,$FD,$39
	.dc.b	$AF,$F9,$5B,$E3,$FF,$CA,$F3,$FD,$C6,$3F,$B9,$46,$1F,$86,$D8,$FE
	.dc.b	$DE,$D8,$F7,$9F,$3C,$F7,$F3,$7E,$5D,$FA,$52,$FC,$BF,$BF,$95,$F3
	.dc.b	$9B,$1F,$DE,$9B,$3F,$E0,$36,$BF,$34,$87,$6E,$4F,$C7,$3E,$1D,$BF
	.dc.b	$26,$E4,$FA,$5F,$10,$F8,$97,$C4,$27,$FC,$C5,$F1,$75,$56,$E7,$5F
	.dc.b	$00,$AD,$4E,$6B,$89,$CD,$C4,$EE,$31,$7E,$07,$67,$67,$11,$B8,$C5
	.dc.b	$F8,$2F,$BD,$69,$FE,$08,$7F,$B5,$89,$3F,$E2,$BC,$1A,$FF,$11,$9C
	.dc.b	$CF,$03,$51,$96,$E5,$E1,$7F,$D3,$39,$FE,$13,$E4,$3E,$B6,$BF,$E9
	.dc.b	$C6,$A6,$9F,$EA,$09,$EC,$43,$F8,$26,$1E,$27,$F2,$13,$FE,$A6,$8F
	.dc.b	$C6,$B4,$FC,$69,$63,$F2,$6D,$7F,$40,$BF,$2D,$FD,$03,$4F,$D0,$16
	.dc.b	$3F,$40,$D7,$FC,$9F,$C7,$FE,$B6,$FA,$2F,$D8,$B6,$7F,$60,$23,$2D
	.dc.b	$3E,$3D,$FA,$EB,$F2,$4F,$B6,$3F,$84,$0F,$CB,$7F,$7B,$D1,$FE,$C4
	.dc.b	$6C,$64,$FE,$18,$5A,$FF,$21,$F5,$7C,$F7,$FC,$94,$D7,$31,$F4,$5C
	.dc.b	$B5,$FA,$07,$90,$F3,$90,$FE,$52,$BF,$2D,$FE,$5E,$D8,$FE,$5E,$D7
	.dc.b	$71,$FC,$EA,$3B,$47,$F9,$E3,$BB,$C8,$D8,$C9,$FE,$78,$8C,$C7,$8C
	.dc.b	$F2,$F0,$FE,$5B,$83,$F2,$11,$93,$FA,$85,$7E,$4C,$FF,$40,$F7,$FE
	.dc.b	$1E,$3F,$CD,$8D,$7F,$CD,$2D,$7F,$D1,$7E,$07,$FB,$CB,$F7,$9F,$EC
	.dc.b	$14,$69,$27,$FC,$ED,$E1,$FA,$A7,$FD,$1E,$DE,$EF,$F7,$C7,$27,$FA
	.dc.b	$42,$3D,$79,$FA,$78,$7F,$04,$DC,$3E,$71,$FE,$9C,$F1,$39,$13,$1F
	.dc.b	$8B,$8F,$F8,$0D,$F9,$73,$F0,$49,$5C,$DF,$D5,$47,$FA,$8B,$C5,$F7
	.dc.b	$34,$F8,$D5,$FE,$17,$6C,$6A,$1E,$0A,$5F,$88,$FF,$89,$5B,$7F,$E2
	.dc.b	$56,$BF,$E2,$56,$BF,$EB,$6F,$23,$E2,$12,$3D,$6F,$C4,$EB,$FE,$19
	.dc.b	$3F,$79,$AF,$F0,$23,$FC,$85,$2F,$CB,$BF,$CB,$64,$8C,$BF,$E4,$F8
	.dc.b	$7E,$A3,$49,$3F,$E5,$18,$CB,$D3,$F2,$DF,$F3,$29,$1C,$1F,$E8,$3F
	.dc.b	$17,$FC,$CA,$47,$07,$93,$D1,$FF,$36,$91,$C1,$FC,$99,$FF,$6C,$3F
	.dc.b	$D4,$5B,$BF,$E6,$DB,$97,$F9,$05,$FF,$43,$E0,$C2,$92,$FF,$D1,$31
	.dc.b	$97,$FD,$91,$EB,$7D,$7A,$FF,$6F,$B6,$33,$3F,$EB,$86,$CF,$FA,$00
	.dc.b	$8F,$FA,$85,$AF,$FB,$3B,$D7,$FC,$17,$EF,$DF,$EA,$DA,$34,$93,$FE
	.dc.b	$D9,$F6,$B6,$D9,$DB,$67,$FD,$7E,$7F,$A5,$39,$E4,$7D,$E6,$BF,$E5
	.dc.b	$B7,$E2,$F5,$FE,$83,$5F,$E0,$51,$83,$FD,$B9,$94,$AF,$F7,$EF,$C1
	.dc.b	$3D,$CF,$85,$1F,$EE,$3C,$BE,$13,$FE,$EA,$F7,$7F,$3D,$FB,$E7,$F7
	.dc.b	$B8,$FD,$07,$47,$FB,$B6,$31,$48,$C4,$FF,$BB,$E7,$FD,$F8,$47,$FD
	.dc.b	$F5,$F0,$1D,$1B,$49,$FA,$AF,$EF,$38,$7F,$84,$11,$FD,$FE,$DF,$C2
	.dc.b	$C8,$E0,$FE,$07,$5F,$B5,$FA,$AF,$2E,$3F,$E1,$59,$46,$27,$47,$F8
	.dc.b	$5E,$3D,$4E,$0F,$F0,$D7,$C1,$3C,$7C,$47,$F8,$7E,$31,$3D,$23,$E3
	.dc.b	$46,$2B,$F3,$AD,$8F,$A3,$F0,$7C,$A9,$BA,$30,$D8,$E9,$3E,$98,$C5
	.dc.b	$FD,$05,$F0,$9F,$C4,$FF,$31,$C3,$6B,$EF,$B5,$E4,$D7,$F8,$D4,$BE
	.dc.b	$1B,$18,$79,$1B,$19,$85,$F8,$67,$3F,$11,$FC,$8F,$F0,$BF,$23,$70
	.dc.b	$FC,$7B,$5F,$53,$5F,$0F,$C5,$DC,$3E,$1B,$71,$34,$C7,$C6,$7F,$4D
	.dc.b	$83,$49,$3F,$D0,$B6,$FA,$BD,$1F,$4B,$83,$FB,$4C,$63,$1F,$AF,$1C
	.dc.b	$F9,$E5,$FF,$70,$C7,$F4,$CD,$8F,$EE,$3F,$1A,$F7,$7F,$B3,$60,$8C
	.dc.b	$57,$F7,$26,$F9,$3F,$C7,$9B,$3F,$AB,$F8,$F7,$C7,$FF,$7C,$6D,$FF
	.dc.b	$5A,$6E,$BF,$C9,$9B,$5F,$B3,$BE,$6F,$F9,$BD,$AF,$F9,$CD,$AF,$97
	.dc.b	$1F,$EF,$9F,$20,$F9,$27,$9D,$FD,$6F,$E4,$FF,$7D,$39,$8C,$5F,$F9
	.dc.b	$EE,$31,$9A,$B7,$7C,$A3,$15,$FC,$1F,$E4,$DC,$BF,$E1,$86,$CF,$F7
	.dc.b	$66,$EB,$FD,$09,$B5,$DE,$FA,$7C,$FF,$6C,$FF,$B8,$37,$03,$4F,$D2
	.dc.b	$65,$FD,$F3,$8B,$CD,$3F,$E2,$16,$F0,$EA,$3F,$C4,$FF,$2A,$F9,$6F
	.dc.b	$F1,$6B,$63,$FC,$16,$DD,$7F,$AB,$B7,$4F,$E5,$7F,$2F,$F9,$32,$57
	.dc.b	$E2,$5F,$CB,$7C,$C7,$8C,$91,$E3,$7F,$13,$F3,$3F,$C5,$70,$FE,$C6
	.dc.b	$DB,$FE,$60,$D7,$FD,$58,$D7,$F7,$3F,$25,$FA,$32,$5F,$E6,$5F,$CF
	.dc.b	$F4,$FF,$85,$FE,$77,$F4,$52,$69,$2B,$F0,$3E,$E7,$F8,$77,$E7,$7F
	.dc.b	$BA,$F3,$A3,$13,$AB,$56,$A7,$33,$FE,$C1,$EC,$F9,$53,$FD,$F9,$B1
	.dc.b	$FD,$2D,$AE,$F8,$38,$BF,$EC,$BF,$37,$FC,$B9,$5F,$D7,$89,$1F,$EA
	.dc.b	$E7,$FC,$13,$F3,$8D,$7F,$66,$DC,$BF,$CC,$6D,$E3,$DC,$E8,$93,$A6
	.dc.b	$C6,$A3,$04,$70,$4B,$FE,$C9,$5D,$5B,$5C,$9B,$5C,$5B,$5C,$5F,$F7
	.dc.b	$35,$71,$6D,$71,$6D,$71,$6D,$71,$7F,$DE,$15,$C5,$B5,$C5,$B5,$C5
	.dc.b	$B5,$C5,$57,$16,$D7,$16,$D7,$16,$D7,$1A,$E0,$4A,$E0,$DA,$E0,$D9
	.dc.b	$E1,$3A,$24,$E9,$B1,$A8,$C1,$1C,$1F,$E4,$8F,$89,$7C,$3B,$E2,$3F
	.dc.b	$C9,$78,$FF,$C0,$03,$51,$96,$C6,$B1,$FE,$0B,$1F,$8A,$E8,$FF,$78
	.dc.b	$3F,$71,$AF,$D6,$6B,$F4,$1A,$FC,$A6,$C7,$48,$E5,$8F,$79,$C9,$1C
	.dc.b	$12,$3C,$4D,$78,$98,$21,$82,$1E,$57,$83,$4D,$94,$D9,$E9,$76,$D3
	.dc.b	$65,$36,$7B,$5D,$B4,$D9,$4D,$CF,$C1,$7C,$04,$3A,$14,$E0,$57,$83
	.dc.b	$A2,$1A,$2B,$A7,$04,$70,$49,$F9,$D1,$E6,$8C,$4F,$A9,$B1,$99,$DB
	.dc.b	$EA,$6B,$E5,$6B,$FC,$57,$D6,$79,$BF,$8D,$81,$97,$0D,$C3,$FC,$EE
	.dc.b	$7F,$32,$23,$FE,$12,$C3,$6B,$C4,$FF,$26,$CF,$FA,$9B,$E7,$9D,$5F
	.dc.b	$ED,$18,$C3,$6B,$FC,$ED,$3A,$BE,$6D,$AF,$F3,$BF,$C4,$FF,$83,$BE
	.dc.b	$2E,$F8,$FF,$ED,$B6,$D9,$A7,$F4,$8C,$21,$E7,$29,$FE,$6D,$A7,$C5
	.dc.b	$FE,$9A,$91,$87,$FD,$7E,$62,$35,$19,$9F,$2E,$BE,$37,$FC,$E9,$08
	.dc.b	$C4,$77,$6B,$FE,$F3,$8D,$57,$A6,$3C,$C4,$3F,$CF,$7D,$7F,$B8,$D7
	.dc.b	$FC,$F3,$1F,$E9,$AF,$BA,$77,$FA,$5B,$7E,$86,$CE,$E7,$F8,$1F,$63
	.dc.b	$F3,$7F,$F3,$C7,$2F,$F8,$A9,$B5,$FE,$29,$6B,$F8,$69,$E1,$E9,$31
	.dc.b	$FE,$37,$3E,$73,$FD,$1D,$9F,$9C,$91,$FD,$D2,$FF,$D0,$DF,$65,$DE
	.dc.b	$F9,$36,$4D,$3F,$03,$F4,$5D,$FE,$DE,$C9,$A4,$C7,$C4,$7E,$4F,$1F
	.dc.b	$0B,$F8,$03,$E8,$57,$37,$EB,$B5,$FC,$23,$5D,$D7,$77,$EE,$F7,$FE
	.dc.b	$08,$91,$F8,$6E,$0F,$F1,$94,$F7,$3C,$51,$FC,$FB,$83,$EC,$C7,$C7
	.dc.b	$8F,$C7,$FD,$27,$E3,$08,$F3,$7F,$1F,$F6,$5F,$3F,$EE,$FF,$46,$DA
	.dc.b	$FC,$7B,$5F,$CE,$35,$FC,$77,$CA,$5B,$FC,$EB,$71,$34,$C7,$D4,$7D
	.dc.b	$DF,$CA,$BF,$62,$DB,$FC,$EB,$5F,$D5,$35,$FC,$E7,$85,$70,$FD,$5B
	.dc.b	$70,$34,$C7,$D6,$7E,$2F,$F3,$DF,$DF,$B6,$FF,$56,$D7,$F7,$0D,$7F
	.dc.b	$8C,$FD,$6A,$31,$08,$C6,$0A,$FE,$3F,$32,$73,$18,$8E,$75,$D6,$BF
	.dc.b	$81,$FD,$75,$C4,$D2,$47,$5A,$FE,$03,$F6,$1F,$C1,$89,$1F,$85,$7E
	.dc.b	$EC,$7F,$1C,$E0,$AF,$F0,$87,$F8,$6F,$D8,$FE,$B7,$F3,$DF,$E4,$2D
	.dc.b	$9F,$E1,$AD,$7F,$8D,$35,$FE,$19,$F6,$97,$0F,$E3,$6D,$C8,$D3,$1C
	.dc.b	$BF,$9A,$AF,$E5,$ED,$7E,$27,$9B,$FC,$6F,$C8,$FE,$7E,$DF,$1B,$FA
	.dc.b	$2C,$EB,$C3,$FE,$9B,$1F,$2F,$F0,$A0,$F9,$2F,$E8,$15,$FD,$42,$33
	.dc.b	$18,$8D,$C6,$F4,$FE,$A6,$E9,$FE,$AA,$7F,$36,$E6,$FE,$25,$F7,$9F
	.dc.b	$96,$83,$49,$5F,$CD,$39,$F6,$D9,$DB,$67,$FA,$7C,$EA,$3F,$A3,$7D
	.dc.b	$BF,$96,$BF,$A2,$B7,$80,$D3,$1F,$0F,$27,$CF,$28,$C6,$9F,$D7,$3D
	.dc.b	$AE,$44,$AD,$BF,$0F,$06,$FF,$47,$EB,$7F,$2D,$18,$E0,$FE,$F7,$18
	.dc.b	$9D,$4F,$D3,$E1,$4F,$D3,$E0,$F9,$8D,$E1,$FE,$74,$DC,$B7,$7F,$A6
	.dc.b	$AF,$A1,$3F,$DC,$B6,$DC,$FF,$5E,$6B,$F1,$AB,$8B,$67,$A5,$48,$E8
	.dc.b	$FD,$D3,$0F,$F7,$3F,$A9,$FE,$99,$E9,$7F,$80,$1B,$1F,$DB,$1A,$EE
	.dc.b	$FF,$D0,$6F,$4C,$7F,$AA,$79,$3E,$B2,$46,$65,$1F,$1F,$5F,$F1,$1C
	.dc.b	$DD,$3F,$CE,$7A,$FE,$7E,$30,$7F,$84,$BE,$87,$FC,$02,$4C,$3F,$A0
	.dc.b	$3F,$E0,$9C,$5F,$67,$FD,$97,$F4,$DF,$E1,$95,$FA,$18,$C5,$FF,$12
	.dc.b	$8C,$09,$DD,$72,$8C,$4E,$E3,$FB,$2C,$7F,$85,$1B,$1C,$3C,$AF,$11
	.dc.b	$FE,$7E,$C7,$D5,$1F,$83,$3D,$65,$3A,$7D,$E7,$A4,$97,$D5,$E5,$A7
	.dc.b	$C6,$9E,$07,$F2,$AA,$EA,$D8,$D3,$FE,$4C,$AF,$E0,$1F,$7B,$F1,$48
	.dc.b	$FA,$5F,$91,$D3,$FE,$0B,$F1,$FE,$0B,$6B,$E0,$35,$E9,$9F,$EF,$AD
	.dc.b	$5F,$3B,$E0,$58,$7F,$8E,$DB,$3F,$E3,$A6,$DE,$E7,$DB,$C1,$F8,$0C
	.dc.b	$BD,$33,$FE,$63,$3E,$39,$E0,$F3,$7E,$F1,$30,$FA,$4F,$E0,$7E,$FB
	.dc.b	$FC,$49,$CD,$F6,$9B,$5F,$65,$AF,$A9,$AF,$73,$E6,$94,$F9,$98,$7D
	.dc.b	$FF,$BF,$FF,$18,$47,$FB,$03,$0F,$E5,$91,$87,$F0,$CD,$8F,$8E,$D7
	.dc.b	$FC,$71,$AB,$F0,$0F,$F1,$D7,$E1,$7C,$E3,$16,$8C,$4E,$A3,$F2,$4D
	.dc.b	$AF,$29,$F1,$8F,$41,$23,$FA,$07,$A9,$FE,$47,$7B,$35,$FC,$E3,$63
	.dc.b	$EE,$B5,$F3,$FB,$DF,$E4,$FF,$63,$CC,$DA,$FD,$23,$67,$F0,$AD,$7D
	.dc.b	$FE,$BF,$F9,$5F,$EB,$3D,$CD,$AF,$D6,$36,$7F,$1A,$D7,$D6,$7F,$2F
	.dc.b	$D7,$F0,$C4,$7F,$6F,$7F,$E5,$5A,$F1,$9F,$E5,$52,$F0,$7A,$1C,$4F
	.dc.b	$C8,$3E,$F9,$D4,$F9,$45,$33,$CD,$FA,$E7,$C7,$EF,$B4,$78,$7F,$EB
	.dc.b	$AD,$7F,$C5,$B5,$FE,$BC,$AD,$4E,$7C,$6F,$F1,$0C,$62,$7F,$D4,$31
	.dc.b	$8B,$E7,$18,$39,$38,$BE,$B0,$AE,$51,$86,$D7,$3B,$F0,$57,$FB,$36
	.dc.b	$3F,$97,$11,$FE,$AC,$D7,$F3,$ED,$78,$1F,$33,$E0,$BE,$51,$32,$C1
	.dc.b	$CA,$21,$C8,$A7,$E0,$97,$D6,$7E,$09,$A7,$E0,$8B,$1F,$82,$6B,$F9
	.dc.b	$25,$F0,$3F,$6C,$8F,$E0,$1F,$CE,$BF,$EB,$08,$FD,$7E,$7F,$EA,$7E
	.dc.b	$0C,$92,$7C,$B3,$F8,$DC,$7E,$03,$4F,$D7,$C3,$F0,$BF,$D9,$B4,$CB
	.dc.b	$83,$F7,$5C,$DF,$EB,$5A,$B3,$FE,$70,$3F,$82,$96,$BF,$82,$B5,$F9
	.dc.b	$E7,$C9,$21,$83,$F8,$4F,$3F,$FB,$07,$F0,$AF,$E3,$D2,$69,$27,$FA
	.dc.b	$C7,$17,$8D,$B3,$F8,$A7,$F8,$9C,$65,$FE,$5D,$19,$7E,$11,$E2,$29
	.dc.b	$E3,$5D,$9F,$CF,$B0,$FE,$79,$1D,$4C,$F6,$75,$AF,$F6,$C5,$6D,$FE
	.dc.b	$9B,$2F,$94,$FE,$A2,$D8,$CC,$60,$FE,$EB,$5C,$E7,$FA,$E9,$DF,$AF
	.dc.b	$A5,$7F,$56,$9F,$03,$E8,$27,$37,$F9,$AF,$1F,$CF,$5F,$F9,$97,$C0
	.dc.b	$E9,$81,$A4,$8F,$89,$5C,$6F,$F1,$2D,$7E,$BC,$7F,$9B,$FC,$1E,$64
	.dc.b	$7C,$CF,$6F,$0F,$FD,$F5,$FA,$7F,$F8,$41,$B5,$FE,$0F,$6B,$FD,$C1
	.dc.b	$AF,$FC,$01,$F3,$5E,$37,$C3,$6E,$1F,$45,$BA,$7F,$8A,$63,$D3,$E1
	.dc.b	$BB,$E9,$27,$F8,$07,$87,$FE,$7F,$FC,$13,$FC,$A8,$4C,$BF,$C7,$4F
	.dc.b	$F9,$17,$07,$E1,$94,$93,$FE,$47,$8C,$BF,$E4,$1E,$6F,$F2,$F1,$1C
	.dc.b	$1F,$E5,$B8,$7C,$57,$F9,$7D,$B1,$98,$C1,$FE,$67,$23,$FE,$4D,$6B
	.dc.b	$FD,$E1,$FF,$26,$11,$FF,$3A,$3F,$E8,$18,$7C,$4D,$24,$FF,$91,$75
	.dc.b	$7A,$0A,$DC,$E7,$D1,$C5,$FF,$93,$AF,$FD,$1D,$5A,$AF,$F1,$94,$62
	.dc.b	$75,$5C,$27,$E0,$47,$BC,$8E,$0F,$F5,$37,$D6,$FF,$A9,$2C,$D2,$47
	.dc.b	$9F,$D3,$F3,$47,$FA,$CC,$91,$99,$E2,$FF,$9C,$EB,$FD,$20,$47,$07
	.dc.b	$FA,$65,$FF,$3A,$11,$C1,$F1,$7D,$4F,$C8,$12,$7F,$20,$FE,$4F,$D5
	.dc.b	$F9,$1F,$9E,$7C,$56,$D7,$CD,$6B,$F1,$9A,$FE,$77,$D7,$B6,$D1,$18
	.dc.b	$6C,$2F,$FD,$EE,$DA,$FF,$7E,$3F,$A3,$F6,$7F,$4C,$47,$F4,$DE,$D7
	.dc.b	$EA,$48,$E0,$FD,$57,$B7,$F8,$1F,$CF,$3E,$F3,$6F,$F0,$CD,$7E,$FB
	.dc.b	$50,$F7,$DE,$F6,$63,$0D,$8C,$DA,$FF,$D9,$F1,$87,$F1,$D3,$D6,$7E
	.dc.b	$23,$73,$D5,$FE,$66,$31,$D3,$F9,$9F,$CA,$7F,$14,$49,$FF,$79,$3F
	.dc.b	$EF,$92,$3F,$87,$6B,$F9,$AF,$82,$FD,$17,$E0,$FF,$9F,$6C,$7E,$9D
	.dc.b	$AE,$F1,$4F,$E9,$7D,$9C,$CF,$F1,$96,$CF,$DD,$6E,$5F,$E8,$C8,$E3
	.dc.b	$3F,$6B,$1F,$D7,$6F,$F6,$5B,$3F,$D5,$FE,$15,$F0,$FF,$D0,$1B,$19
	.dc.b	$8C,$1F,$97,$9F,$DC,$72,$DC,$3F,$39,$5F,$CA,$7E,$1F,$F8,$07,$AB
	.dc.b	$FB,$D6,$CF,$F0,$56,$BF,$BE,$6B,$F8,$0C,$7C,$38,$7E,$0A,$24,$61
	.dc.b	$C0,$70,$EB,$1F,$E7,$95,$D3,$0F,$D1,$37,$33,$4F,$0F,$BE,$F8,$97
	.dc.b	$C4,$FF,$19,$FA,$FE,$F3,$98,$C6,$5E,$58,$C6,$2A,$DD,$F2,$8C,$57
	.dc.b	$A3,$3F,$3B,$EF,$CF,$E6,$CF,$CF,$3F,$C1,$D8,$74,$FB,$5E,$79,$7F
	.dc.b	$2B,$6C,$9A,$7F,$BE,$3E,$31,$EC,$7F,$7F,$6C,$FF,$83,$5B,$BE,$E7
	.dc.b	$C2,$FC,$CC,$BE,$66,$58,$7D,$0F,$F6,$7F,$2B,$F9,$E7,$C9,$FF,$39
	.dc.b	$6C,$FF,$41,$6B,$FC,$E9,$AF,$F7,$1F,$7F,$FB,$81,$32,$FE,$E0,$F7
	.dc.b	$8F,$B5,$1F,$0F,$0F,$26,$DF,$22,$3F,$8E,$7C,$7F,$46,$8C,$7F,$A6
	.dc.b	$46,$78,$CF,$8B,$07,$C8,$3E,$47,$FA,$64,$7E,$3E,$31,$87,$F5,$38
	.dc.b	$FE,$27,$18,$F1,$7E,$33,$63,$E2,$CE,$DF,$92,$DC,$BF,$57,$5E,$AD
	.dc.b	$BB,$D7,$F1,$EF,$51,$BF,$F8,$36,$BF,$A4,$4F,$FB,$A9,$AF,$F2,$4F
	.dc.b	$8A,$FF,$3D,$E3,$FE,$4F,$6E,$FF,$65,$B3,$FA,$66,$BF,$AB,$6B,$CF
	.dc.b	$E0,$3F,$C7,$F1,$88,$34,$91,$A2,$57,$85,$B5,$E1,$6D,$78,$5A,$F8
	.dc.b	$5A,$F8,$67,$91,$27,$93,$63,$93,$5E,$4D,$79,$47,$02,$3C,$1A,$F0
	.dc.b	$6B,$C0,$D1,$0D,$14,$D1,$82,$18,$21,$F2,$27,$DA,$D9,$F3,$36,$7C
	.dc.b	$2D,$3B,$14,$EA,$53,$AD,$7D,$29,$E6,$D9,$E6,$D9,$E6,$D7,$9B,$5F
	.dc.b	$F7,$2F,$C1,$7E,$FD,$FF,$72,$7C,$0F,$FB,$A9,$FE,$5B,$F0,$7F,$EE
	.dc.b	$F6,$C6,$5F,$E4,$1F,$4B,$FE,$3D,$7F,$84,$75,$23,$11,$FC,$A1,$AF
	.dc.b	$0C,$BC,$13,$B6,$CE,$A7,$FC,$03,$39,$7A,$C6,$54,$65,$ED,$3C,$88
	.dc.b	$F0,$6B,$A3,$04,$30,$43,$CC,$64,$A6,$4F,$2C,$F8,$E7,$C5,$1A,$21
	.dc.b	$A9,$F0,$B8,$3D,$33,$C4,$91,$A8,$E0,$78,$4A,$78,$4E,$2E,$DA,$6C
	.dc.b	$A6,$E7,$CA,$F2,$3C,$4E,$1D,$98,$9F,$1B,$CE,$34,$48,$CC,$75,$6B
	.dc.b	$C6,$7C,$93,$B6,$CE,$A7,$C4,$43,$C2,$59,$E0,$F8,$63,$F0,$1B,$7F
	.dc.b	$26,$F5,$BF,$25,$28,$C6,$5F,$D6,$5B,$1F,$E0,$66,$BF,$D3,$23,$F9
	.dc.b	$96,$1C,$B8,$7F,$C3,$F1,$83,$FC,$1B,$0F,$73,$FA,$1F,$C9,$61,$B7
	.dc.b	$FE,$15,$9D,$72,$FF,$A7,$EB,$4D,$9F,$17,$1F,$FB,$25,$F3,$35,$CE
	.dc.b	$7F,$B3,$CF,$DD,$1F,$DC,$FE,$71,$E1,$FE,$85,$B9,$7E,$E5,$B3,$B9
	.dc.b	$FD,$54,$FA,$79,$3C,$53,$FE,$6C,$8C,$0C,$FD,$93,$A9,$FF,$5C,$69
	.dc.b	$FD,$3A,$7C,$F9,$79,$73,$F4,$57,$F8,$D7,$3F,$3B,$EA,$3D,$1E,$17
	.dc.b	$9C,$99,$79,$DF,$DD,$7C,$E3,$EA,$FF,$D8,$DC,$9F,$E1,$18,$C5,$23
	.dc.b	$13,$AB,$FF,$20,$D6,$A7,$CB,$10,$F1,$10,$FF,$25,$47,$4F,$7F,$9D
	.dc.b	$9A,$49,$EB,$D5,$FE,$D9,$F8,$4F,$23,$63,$34,$9C,$B6,$3C,$07,$F7
	.dc.b	$DD,$3E,$3B,$75,$F9,$47,$F3,$1C,$7F,$CD,$F7,$C8,$93,$D9,$FC,$87
	.dc.b	$CB,$7D,$5C,$3F,$28,$DE,$03,$4F,$B5,$5F,$A5,$C7,$C2,$DC,$7C,$2F
	.dc.b	$C7,$3F,$CD,$DE,$0F,$F8,$81,$FF,$21,$35,$FF,$37,$B5,$DF,$2F,$E3
	.dc.b	$B0,$7C,$F7,$FB,$FF,$E5,$FF,$4C,$DA,$FD,$34,$EB,$35,$69,$B3,$AF
	.dc.b	$9F,$E0,$FE,$5E,$91,$88,$FC,$AB,$67,$F6,$CD,$97,$CF,$25,$F9,$37
	.dc.b	$EE,$63,$0D,$8C,$DF,$9A,$72,$FA,$3E,$26,$BE,$36,$BF,$A6,$6C,$FF
	.dc.b	$07,$6D,$BE,$8F,$C5,$F9,$DF,$F0,$96,$DF,$C4,$6B,$EF,$6B,$FE,$85
	.dc.b	$B4,$FF,$18,$24,$7F,$94,$1F,$E2,$9E,$6F,$09,$FE,$30,$DB,$FF,$49
	.dc.b	$4E,$A7,$9D,$7F,$19,$FC,$EF,$F8,$27,$67,$F2,$06,$C6,$4F,$C8,$38
	.dc.b	$7F,$03,$3F,$E0,$9E,$1F,$C2,$47,$FB,$E1,$A9,$CC,$25,$C5,$F5,$5F
	.dc.b	$03,$6B,$C8,$D7,$F8,$C4,$70,$9E,$51,$E9,$FA,$7D,$3C,$DB,$1D,$DB
	.dc.b	$26,$FD,$4F,$B0,$93,$FC,$8E,$BE,$24,$7E,$AB,$BB,$F8,$EC,$7F,$1F
	.dc.b	$E6,$FD,$5C,$62,$3F,$78,$7E,$A3,$9B,$FC,$2D,$9F,$C2,$23,$FC,$D3
	.dc.b	$2F,$F1,$07,$E1,$DF,$3F,$C5,$FC,$B3,$73,$34,$FA,$9F,$5B,$AC,$7E
	.dc.b	$C3,$5F,$8A,$D7,$E6,$36,$7E,$87,$D7,$7D,$A2,$3E,$F9,$F9,$B1,$FC
	.dc.b	$23,$2F,$3C,$FD,$DC,$BD,$67,$A7,$F0,$3F,$C1,$92,$39,$5F,$DB,$8F
	.dc.b	$91,$AF,$C6,$27,$07,$F1,$09,$FE,$17,$7D,$2F,$B1,$23,$F2,$5F,$80
	.dc.b	$FE,$40,$91,$D8,$FC,$53,$E7,$9E,$53,$F4,$67,$93,$F9,$DE,$4F,$CC
	.dc.b	$92,$39,$6F,$FD,$F5,$F4,$CF,$E5,$EB,$9B,$E6,$8E,$6F,$E0,$70,$F5
	.dc.b	$34,$FE,$54,$7E,$6F,$ED,$7E,$F1,$27,$FD,$D8,$7E,$85,$FC,$17,$DB
	.dc.b	$7F,$15,$FA,$0F,$D8,$36,$BF,$68,$D7,$F4,$2D,$7F,$51,$ED,$7E,$2C
	.dc.b	$97,$DB,$24,$7F,$8A,$35,$FA,$35,$E6,$3F,$A1,$1F,$BF,$FA,$9F,$CA
	.dc.b	$12,$F9,$3C,$73,$F7,$CF,$E7,$8F,$EB,$B3,$FB,$BF,$B8,$FC,$E1,$1F
	.dc.b	$13,$F9,$FF,$8B,$FE,$65,$F2,$7F,$C3,$5B,$5F,$C4,$5A,$FE,$F9,$AF
	.dc.b	$F3,$9F,$B0,$47,$F8,$71,$B7,$F3,$1B,$3F,$E3,$86,$BF,$DC,$8F,$79
	.dc.b	$7A,$7F,$71,$3F,$C4,$FE,$03,$F6,$84,$7E,$53,$FB,$7F,$AE,$FE,$42
	.dc.b	$48,$FC,$4B,$F9,$EC,$FE,$2D,$F6,$3F,$08,$7F,$2D,$FB,$AF,$F5,$35
	.dc.b	$53,$FC,$65,$1F,$C7,$23,$2D,$8C,$BF,$C9,$7C,$CF,$F2,$64,$7F,$37
	.dc.b	$6C,$7F,$24,$6B,$B8,$FF,$1E,$1E,$FD,$FF,$03,$1F,$63,$85,$F0,$3E
	.dc.b	$08,$FE,$64,$DA,$FE,$6C,$D9,$DC,$79,$07,$ED,$4F,$91,$FE,$DD,$D9
	.dc.b	$FE,$C0,$9C,$C6,$1F,$F3,$34,$ED,$FE,$88,$62,$35,$87,$F9,$56,$75
	.dc.b	$E7,$73,$6D,$6C,$FF,$18,$16,$3F,$D9,$BC,$1F,$CC,$EB,$E9,$1F,$97
	.dc.b	$3E,$99,$D6,$FF,$C7,$95,$C4,$91,$97,$AB,$C0,$E4,$43,$A9,$27,$C0
	.dc.b	$4C,$FF,$C5,$FE,$D7,$F9,$67,$D8,$FE,$C9,$EE,$2F,$FD,$E2,$7F,$65
	.dc.b	$70,$FF,$BC,$A3,$6E,$B4,$4E,$AB,$A4,$F1,$35,$19,$F5,$17,$CA,$7A
	.dc.b	$1B,$3F,$C8,$53,$FE,$7C,$23,$FE,$92,$6B,$FC,$7D,$AF,$F9,$65,$FF
	.dc.b	$52,$EB,$F6,$F9,$3E,$B6,$BF,$E2,$F6,$EA,$69,$F5,$B4,$FF,$00,$DF
	.dc.b	$4C,$3C,$67,$84,$FB,$D3,$C2,$A9,$23,$EB,$7A,$FD,$89,$1F,$85,$7F
	.dc.b	$25,$C1,$F9,$8B,$BC,$3F,$E4,$36,$D7,$CC,$3F,$C4,$51,$97,$C5,$7F
	.dc.b	$04,$A7,$C1,$3F,$64,$BE,$07,$BF,$1F,$F9,$3D,$B1,$93,$6F,$93,$93
	.dc.b	$B5,$7E,$FF,$0C,$57,$8E,$30,$F8,$1F,$8A,$78,$C8,$7F,$8A,$79,$BC
	.dc.b	$0F,$F8,$ED,$AF,$F8,$A1,$AE,$E3,$D0,$FF,$90,$74,$F8,$8D,$8F,$C7
	.dc.b	$B6,$BF,$C8,$93,$A8,$F0,$9E,$43,$E2,$1E,$49,$73,$FF,$36,$AF,$E6
	.dc.b	$9D,$1C,$6B,$F9,$E3,$6B,$6D,$D7,$F9,$EF,$CA,$FF,$47,$93,$49,$5F
	.dc.b	$E5,$9E,$BE,$04,$9F,$4C,$7F,$92,$23,$C0,$F9,$C9,$D0,$E2,$FC,$9C
	.dc.b	$7D,$EA,$FF,$2B,$76,$78,$C8,$F4,$8F,$F2,$C4,$74,$AE,$5D,$79,$9E
	.dc.b	$B7,$D1,$F8,$99,$7F,$74,$C3,$91,$FE,$63,$ED,$F8,$04,$73,$0E,$EF
	.dc.b	$F4,$D7,$CD,$7F,$81,$A4,$D2,$5F,$F9,$EF,$C1,$F9,$44,$9E,$6F,$9F
	.dc.b	$1F,$24,$FD,$82,$1F,$17,$B9,$E1,$7C,$86,$DF,$B5,$BC,$5F,$E2,$F3
	.dc.b	$FD,$11,$E1,$FE,$3A,$7F,$CF,$CD,$D3,$FC,$5D,$D1,$A9,$FC,$8E,$3F
	.dc.b	$E8,$AB,$E4,$FF,$8B,$DA,$FF,$A0,$3C,$4E,$44,$9E,$31,$FE,$89,$8F
	.dc.b	$CE,$BF,$E3,$98,$F3,$65,$E9,$23,$F5,$78,$7E,$BD,$FF,$11,$6D,$7E
	.dc.b	$07,$17,$8B,$E9,$25,$F8,$0F,$F4,$D3,$E0,$D3,$E1,$D7,$A4,$87,$B4
	.dc.b	$FF,$57,$78,$DF,$0C,$95,$CE,$D1,$F1,$FC,$7F,$F6,$C7,$E1,$3F,$E8
	.dc.b	$0A,$34,$97,$FE,$B9,$F2,$B6,$D9,$DB,$67,$FD,$18,$7C,$C7,$FD,$6D
	.dc.b	$AF,$C4,$D7,$11,$F9,$8E,$6F,$8D,$C5,$F8,$68,$FD,$F6,$1F,$EB,$9F
	.dc.b	$2F,$EE,$11,$F5,$3F,$7F,$CC,$FF,$78,$7E,$75,$FE,$CB,$6D,$7F,$AD
	.dc.b	$E9,$39,$6C,$7F,$B0,$1A,$FF,$B0,$30,$E2,$D8,$CB,$FE,$C9,$8C,$BE
	.dc.b	$C8,$FC,$09,$31,$F8,$0F,$FB,$20,$FA,$11,$F9,$52,$3F,$77,$95,$E8
	.dc.b	$21,$F9,$CF,$FB,$A4,$95,$FD,$E2,$3F,$DD,$E6,$9F,$EF,$63,$E0,$1E
	.dc.b	$2F,$3A,$CC,$E6,$30,$F0,$9D,$D7,$38,$FF,$02,$C6,$2B,$AC,$B0,$7A
	.dc.b	$7C,$AF,$FC,$1D,$39,$8C,$47,$AE,$77,$3D,$63,$13,$A8,$E9,$EA,$7C
	.dc.b	$1F,$CA,$3B,$DA,$31,$7F,$06,$31,$4A,$DD,$6A,$BC,$74,$F5,$B1,$1E
	.dc.b	$36,$DF,$AD,$B3,$B8,$E9,$3D,$FD,$8F,$A9,$F1,$BE,$46,$C6,$65,$39
	.dc.b	$6C,$79,$FD,$BD,$BE,$46,$BF,$15,$B1,$F3,$5C,$C7,$F0,$DE,$0F,$D3
	.dc.b	$F6,$7F,$00,$8F,$DF,$1F,$D9,$3B,$BF,$85,$B7,$9F,$F9,$7E,$1F,$5F
	.dc.b	$DC,$FE,$4D,$A1,$A4,$8F,$E3,$B4,$F7,$7F,$07,$E6,$7E,$21,$B1,$F8
	.dc.b	$E6,$BF,$E4,$A6,$BF,$CF,$31,$F1,$B7,$1F,$DE,$4E,$A3,$B7,$2F,$F8
	.dc.b	$E5,$E8,$D7,$A3,$5F,$F2,$D4,$3D,$FF,$D3,$79,$1E,$56,$CF,$95,$AF
	.dc.b	$95,$AF,$8F,$A7,$F8,$79,$37,$E5,$A7,$FA,$5A,$79,$35,$E4,$D7,$C7
	.dc.b	$5F,$B7,$F4,$3F,$5B,$F0,$4F,$81,$FF,$00,$B6,$FF,$BB,$37,$0D,$CF
	.dc.b	$F6,$5C,$7C,$58,$FF,$8C,$76,$F9,$38,$7E,$BF,$D3,$FB,$24,$7F,$84
	.dc.b	$3F,$D4,$B1,$7C,$17,$80,$93,$E0,$7B,$65,$FD,$47,$83,$F9,$76,$1D
	.dc.b	$B8,$FF,$B7,$F9,$5F,$B0,$F0,$FF,$0E,$4C,$7F,$0B,$DF,$FC,$1B,$E1
	.dc.b	$7F,$85,$B7,$6F,$E1,$5C,$BF,$C6,$67,$2D,$8F,$E2,$B9,$7F,$86,$27
	.dc.b	$F8,$9B,$63,$39,$B8,$15,$A8,$C3,$FC,$6B,$E1,$1E,$3F,$F8,$B9,$B9
	.dc.b	$FF,$88,$DB,$BE,$E7,$44,$9D,$36,$35,$18,$23,$82,$63,$FC,$BA,$BA
	.dc.b	$B6,$B9,$36,$B8,$B6,$B8,$BF,$EB,$0A,$E2,$DA,$E2,$DA,$E2,$DA,$E2
	.dc.b	$FF,$AE,$EB,$8B,$6B,$8B,$6B,$8B,$6B,$8A,$AE,$2D,$AE,$2D,$AE,$2D
	.dc.b	$AE,$35,$C0,$95,$C1,$B5,$C1,$B3,$C2,$74,$49,$D3,$63,$51,$82,$38
	.dc.b	$25,$7C,$B2,$1F,$14,$A7,$BC,$A7,$FB,$48,$FE,$CB,$ED,$FF,$B6,$9B
	.dc.b	$ED,$FF,$B6,$1F,$EC,$B1,$97,$EB,$75,$7F,$73,$7F,$20,$D7,$F0,$4D
	.dc.b	$77,$7F,$DB,$9F,$9D,$19,$7E,$C3,$5C,$C6,$F1,$FF,$6C,$36,$75,$18
	.dc.b	$9F,$0C,$E4,$8E,$08,$79,$1C,$B4,$C9,$E6,$29,$E4,$76,$E4,$E0,$43
	.dc.b	$04,$38,$1C,$8F,$59,$63,$D0,$F1,$36,$53,$65,$76,$F5,$32,$57,$2F
	.dc.b	$23,$13,$F1,$89,$3E,$A8,$D1,$0E,$65,$9E,$6E,$08,$E0,$93,$D2,$3D
	.dc.b	$8D,$7D,$0D,$7C,$73,$E1,$79,$C6,$5F,$03,$5C,$C6,$E7,$BB,$F8,$9D
	.dc.b	$7F,$D7,$BE,$DF,$F2,$1D,$BF,$2D,$62,$4E,$9F,$E5,$31,$A7,$FC,$95
	.dc.b	$E6,$FF,$67,$B7,$C0,$7F,$8E,$A3,$0D,$EB,$FF,$28,$36,$77,$5F,$E4
	.dc.b	$CF,$B1,$FF,$1E,$CF,$F9,$29,$B1,$93,$89,$67,$8B,$FE,$44,$FA,$FF
	.dc.b	$F2,$E4,$7F,$94,$83,$88,$FF,$26,$B5,$FF,$2F,$35,$E5,$F1,$3D,$9B
	.dc.b	$3D,$9F,$F3,$44,$A7,$51,$86,$AF,$B2,$FE,$CF,$5C,$63,$1D,$7A,$9C
	.dc.b	$E3,$FE,$68,$6C,$FF,$9C,$7E,$91,$F1,$BF,$BD,$C1,$18,$7D,$CD,$9F
	.dc.b	$F3,$4B,$5F,$E8,$5F,$1F,$FE,$64,$E5,$FE,$98,$D8,$CF,$9B,$FE,$83
	.dc.b	$9C,$B6,$3F,$CF,$FF,$69,$FE,$57,$CF,$FC,$FA,$D8,$CF,$99,$F0,$23
	.dc.b	$11,$FD,$27,$0F,$F3,$9F,$C6,$74,$CC,$D2,$4F,$F5,$7F,$C6,$33,$2F
	.dc.b	$87,$FE,$C5,$18,$6D,$7F,$A6,$5B,$3B,$AF,$F4,$AF,$8B,$F5,$63,$E8
	.dc.b	$37,$3F,$F1,$4F,$17,$D6,$8C,$57,$87,$0E,$8D,$BE,$8F,$F6,$3C,$7C
	.dc.b	$4E,$71,$FC,$5F,$99,$F8,$2A,$FC,$7B,$79,$0D,$3F,$0B,$E1,$FF,$A8
	.dc.b	$B4,$FF,$0C,$D7,$E1,$63,$18,$7F,$87,$A3,$F2,$07,$E6,$BC,$2E,$84
	.dc.b	$73,$82,$3C,$1E,$47,$FA,$F7,$F0,$5F,$CF,$B6,$FF,$3E,$D7,$F5,$0D
	.dc.b	$7F,$3B,$D1,$70,$FD,$4B,$70,$FC,$DD,$FE,$AA,$31,$F2,$7F,$EC,$BF
	.dc.b	$5F,$F5,$AD,$C3,$F5,$AD,$7F,$6C,$D7,$F5,$7F,$29,$FE,$D3,$F8,$FF
	.dc.b	$DA,$B6,$3F,$6A,$D7,$F7,$CD,$7F,$67,$F2,$BF,$ED,$BF,$CF,$7F,$7A
	.dc.b	$D8,$FD,$EB,$5F,$E0,$CD,$7F,$8E,$ED,$FE,$43,$F9,$1F,$E4,$44,$8F
	.dc.b	$E3,$EF,$F0,$8F,$8D,$FF,$74,$37,$73,$49,$3F,$C3,$3C,$AF,$C1,$12
	.dc.b	$7E,$BB,$F8,$5F,$98,$91,$F6,$9B,$C9,$FC,$5B,$E6,$FF,$DB,$97,$FC
	.dc.b	$59,$AF,$F1,$F6,$BF,$C5,$1B,$5F,$C8,$3E,$72,$F2,$1A,$48,$FF,$77
	.dc.b	$57,$F2,$2E,$74,$3F,$0A,$FC,$EB,$67,$4D,$7F,$98,$4F,$09,$FE,$95
	.dc.b	$8C,$9F,$E8,$ED,$7F,$A6,$7C,$9F,$F2,$5E,$EF,$E5,$2D,$C3,$F9,$73
	.dc.b	$6B,$FA,$95,$70,$F1,$BF,$6E,$FF,$3F,$F9,$FF,$3F,$E7,$BF,$CF,$1B
	.dc.b	$1F,$CF,$1A,$EF,$04,$FE,$39,$7E,$1D,$49,$3F,$90,$8C,$BF,$6D,$F8
	.dc.b	$0F,$F4,$2E,$1F,$80,$E7,$4F,$E9,$B1,$88,$FE,$A9,$C3,$5F,$E9,$2D
	.dc.b	$AF,$EA,$AD,$E0,$37,$E8,$BE,$9F,$E1,$9F,$D6,$1B,$7F,$D6,$1A,$FF
	.dc.b	$61,$6B,$F9,$CF,$3F,$CE,$BD,$0F,$E0,$D1,$86,$E7,$FE,$01,$3E,$7C
	.dc.b	$B8,$7E,$A4,$61,$B1,$F5,$1A,$FD,$5E,$2F,$56,$BF,$8A,$FB,$1F,$87
	.dc.b	$B1,$A4,$AE,$5C,$5F,$CD,$F5,$EC,$D8,$CE,$1E,$CC,$7F,$0D,$5F,$DE
	.dc.b	$BE,$E3,$F2,$FF,$86,$5C,$8D,$26,$1C,$A1,$F8,$67,$6C,$7F,$3E,$D8
	.dc.b	$CC,$EE,$3F,$C4,$8C,$3F,$C4,$25,$3F,$C2,$1F,$51,$FC,$FC,$93,$FC
	.dc.b	$D1,$FE,$85,$F5,$3F,$AC,$F0,$7F,$C2,$2D,$AF,$F0,$8B,$5F,$F0,$D3
	.dc.b	$5F,$F0,$7F,$DC,$FF,$80,$79,$8D,$26,$1F,$E3,$BF,$03,$FD,$B9,$5E
	.dc.b	$53,$F0,$45,$8F,$F1,$1F,$D7,$7F,$55,$23,$FD,$29,$FF,$0C,$F2,$FE
	.dc.b	$1A,$79,$1F,$89,$AE,$AF,$C2,$3F,$6B,$4F,$D7,$5F,$61,$FD,$8D,$B1
	.dc.b	$FD,$69,$BD,$DF,$8E,$8E,$13,$BF,$3F,$F9,$A4,$73,$6D,$7F,$81,$3C
	.dc.b	$6E,$0D,$C3,$C4,$7F,$06,$CB,$D0,$E0,$F8,$25,$3D,$5D,$7F,$1C,$8F
	.dc.b	$4F,$B1,$F4,$13,$1F,$43,$EB,$FB,$2F,$E4,$BF,$86,$78,$9B,$5F,$0B
	.dc.b	$14,$62,$3F,$CD,$57,$F1,$63,$11,$E4,$3D,$1B,$7C,$C6,$E3,$E4,$7A
	.dc.b	$7D,$DF,$CB,$25,$FC,$B7,$E7,$7A,$5F,$CE,$3F,$0C,$FF,$64,$4F,$D3
	.dc.b	$8C,$3E,$56,$D7,$93,$2F,$F1,$74,$AB,$FD,$9D,$AF,$CA,$AF,$19,$F8
	.dc.b	$27,$0F,$F4,$6A,$7D,$B7,$F4,$8F,$9E,$F0,$C6,$30,$46,$27,$51,$F1
	.dc.b	$DB,$5E,$3F,$B7,$FE,$A3,$F6,$7C,$9B,$5F,$35,$B3,$B8,$4F,$C0,$F8
	.dc.b	$DF,$E1,$D1,$EC,$9F,$A3,$18,$D3,$5E,$27,$F9,$26,$BF,$33,$39,$7F
	.dc.b	$0D,$A7,$BA,$FF,$0E,$7E,$A7,$47,$DD,$FF,$66,$FB,$BF,$7B,$6F,$F0
	.dc.b	$4D,$AF,$9C,$D7,$F0,$04,$3A,$9E,$0E,$2F,$C1,$39,$C3,$C4,$78,$FE
	.dc.b	$29,$F0,$4F,$88,$FF,$5B,$7A,$2F,$2F,$FB,$3D,$6A,$73,$18,$AE,$18
	.dc.b	$7F,$95,$23,$13,$A8,$FF,$60,$57,$17,$46,$2B,$8D,$F7,$8C,$57,$15
	.dc.b	$5C,$67,$51,$FB,$0B,$FF,$67,$11,$FF,$65,$B5,$FD,$0B,$5F,$EB,$4F
	.dc.b	$F0,$0C,$BF,$C2,$BF,$77,$F4,$34,$34,$8F,$4A,$87,$42,$9F,$E1,$ED
	.dc.b	$FF,$0F,$F1,$7F,$84,$93,$48,$FE,$44,$7E,$01,$F9,$56,$99,$70,$7E
	.dc.b	$7D,$78,$54,$8F,$E8,$4C,$9E,$4B,$F8,$C5,$DF,$FB,$BE,$7F,$C8,$39
	.dc.b	$3F,$C6,$BF,$86,$7E,$B9,$B7,$FB,$D1,$19,$69,$FC,$28,$7C,$97,$F2
	.dc.b	$AE,$CF,$F1,$FB,$63,$38,$EE,$3E,$93,$E4,$3E,$51,$0F,$E0,$5D,$DF
	.dc.b	$E4,$8F,$C3,$3F,$92,$51,$A4,$9F,$E4,$DE,$0F,$81,$B3,$E0,$7F,$8A
	.dc.b	$46,$5B,$19,$7F,$06,$7C,$5E,$77,$81,$F8,$E8,$C9,$E5,$3D,$06,$09
	.dc.b	$3F,$D3,$47,$E7,$9F,$D3,$3C,$4F,$F3,$0B,$63,$33,$FC,$F6,$72,$FF
	.dc.b	$3F,$F1,$7C,$CF,$F2,$36,$CF,$F5,$C3,$D0,$5D,$3E,$5F,$0F,$DC,$AF
	.dc.b	$56,$3E,$9A,$FE,$EA,$70,$F2,$3F,$0B,$7F,$D9,$DB,$5F,$DC,$1A,$FB
	.dc.b	$EF,$A3,$73,$F8,$BC,$5C,$7C,$8F,$F3,$EB,$71,$F0,$61,$FE,$0C,$6C
	.dc.b	$EA,$FF,$C2,$B0,$F9,$DF,$F0,$CB,$63,$31,$83,$CD,$E6,$7C,$2B,$34
	.dc.b	$95,$F6,$EB,$EA,$93,$2F,$4B,$F5,$FD,$2F,$F4,$27,$E1,$7F,$E3,$52
	.dc.b	$47,$F8,$ED,$FF,$20,$7E,$17,$23,$E2,$7A,$7F,$E9,$A8,$C4,$7F,$95
	.dc.b	$70,$7E,$17,$FE,$3F,$24,$7F,$91,$5F,$F2,$BF,$C9,$48,$FF,$40,$35
	.dc.b	$E5,$F8,$57,$F9,$1C,$8F,$F9,$35,$FF,$34,$BF,$E0,$E6,$E7,$FE,$70
	.dc.b	$6F,$0F,$E6,$1B,$C6,$7F,$CF,$D8,$F9,$AF,$FC,$FC,$D7,$FC,$FC,$D7
	.dc.b	$FD,$6F,$BB,$DA,$FF,$5C,$E7,$F8,$7F,$6F,$FD,$77,$85,$C3,$B6,$1F
	.dc.b	$89,$9C,$B6,$32,$FF,$86,$30,$FE,$99,$18,$8F,$F5,$03,$44,$6A,$BF
	.dc.b	$CF,$E7,$FA,$B7,$F0,$99,$5F,$EA,$B6,$BF,$EB,$3F,$CF,$BF,$D7,$16
	.dc.b	$69,$23,$FC,$F1,$CF,$FE,$75,$23,$83,$FD,$2F,$B7,$FA,$72,$3E,$59
	.dc.b	$23,$3B,$7F,$06,$8F,$F6,$DE,$32,$7F,$DB,$6D,$7E,$07,$C0,$FF,$B8
	.dc.b	$B9,$FF,$60,$D8,$FD,$83,$5D,$E5,$FE,$74,$7F,$DE,$CD,$7F,$DE,$CD
	.dc.b	$7F,$DE,$CD,$3E,$4F,$C1,$73,$8F,$F6,$2C,$7F,$01,$DB,$F5,$44,$7F
	.dc.b	$82,$74,$FE,$B4,$8E,$0F,$E1,$1F,$09,$FE,$FA,$FA,$5F,$E1,$0D,$BF
	.dc.b	$E1,$0D,$7E,$F3,$5F,$F6,$44,$7E,$50,$8F,$CA,$E4,$47,$F1,$3F,$86
	.dc.b	$FE,$47,$3F,$87,$6B,$F9,$46,$DF,$E1,$9B,$C6,$F7,$12,$FA,$5F,$F7
	.dc.b	$A1,$2F,$FD,$E4,$FF,$21,$F8,$9D,$ED,$68,$D2,$57,$F2,$5F,$8B,$EB
	.dc.b	$F8,$6F,$F2,$96,$C6,$69,$39,$6C,$7F,$2D,$F8,$ED,$BF,$A3,$6B,$FB
	.dc.b	$06,$C7,$E8,$9B,$9B,$9B,$F9,$11,$27,$F9,$E3,$FD,$22,$9F,$1E,$F9
	.dc.b	$DC,$36,$BF,$93,$37,$A3,$F8,$6B,$6B,$FA,$17,$A9,$FC,$BC,$95,$FD
	.dc.b	$25,$FE,$AB,$F3,$6F,$92,$79,$3F,$43,$D3,$FF,$7F,$CE,$63,$18,$09
	.dc.b	$19,$C3,$FC,$4B,$19,$11,$9A,$FE,$43,$F2,$6F,$88,$F7,$36,$FF,$A0
	.dc.b	$37,$4D,$CF,$9B,$2F,$2D,$76,$C7,$C3,$18,$C3,$CE,$D8,$F1,$63,$E1
	.dc.b	$6F,$CA,$3E,$5F,$F1,$BF,$43,$EE,$8C,$62,$8C,$4E,$AB,$8B,$5F,$77
	.dc.b	$CB,$BE,$63,$EC,$36,$FF,$AD,$37,$5F,$E8,$2D,$AE,$72,$FA,$1E,$ED
	.dc.b	$8C,$C6,$0F,$F7,$27,$CD,$3E,$6F,$FB,$94,$FF,$B2,$A3,$15,$FE,$07
	.dc.b	$6F,$87,$C2,$B7,$5F,$67,$E4,$DD,$9F,$8F,$6E,$1F,$DC,$5B,$9F,$F5
	.dc.b	$86,$D7,$F8,$37,$8F,$FB,$D1,$2B,$FC,$28,$FF,$7D,$7B,$9F,$D0,$8B
	.dc.b	$D5,$FB,$C7,$FC,$1A,$79,$25,$F3,$9F,$EE,$4C,$3F,$B3,$B6,$3F,$18
	.dc.b	$68,$FD,$33,$97,$FD,$D7,$F3,$AF,$43,$FC,$5E,$DB,$FF,$17,$B6,$F9
	.dc.b	$C2,$31,$5F,$C6,$FE,$7F,$FB,$89,$1C,$1F,$A4,$E8,$F4,$73,$FF,$A1
	.dc.b	$2F,$FC,$9D,$3F,$A0,$9C,$92,$33,$3B,$E0,$FF,$01,$B6,$BA,$4E,$AB
	.dc.b	$85,$69,$B3,$A8,$C4,$EC,$8E,$09,$1B,$21,$FA,$49,$E6,$D9,$E2,$E8
	.dc.b	$FD,$EC,$69,$B1,$A7,$84,$F6,$69,$C4,$F8,$51,$B7,$8C,$6D,$A6,$EB
	.dc.b	$E1,$3C,$88,$68,$A7,$48,$E4,$D3,$95,$79,$DC,$1C,$27,$AB,$CC,$86
	.dc.b	$8A,$E9,$C1,$1C,$12,$3B,$36,$BD,$D3,$B9,$C9,$23,$33,$BA,$F4,$B6
	.dc.b	$BE,$47,$C2,$FF,$BB,$5F,$CF,$B5,$FF,$79,$9F,$E4,$A7,$0F,$F3,$3F
	.dc.b	$C1,$DF,$29,$FD,$A2,$3F,$C9,$F1,$8C,$3F,$CA,$30,$8C,$3A,$9E,$0F
	.dc.b	$EB,$A3,$2D,$8C,$BF,$CF,$63,$2D,$8C,$BE,$36,$BD,$DB,$1C,$23,$53
	.dc.b	$E4,$8C,$46,$E3,$13,$C0,$8E,$09,$1E,$96,$BE,$76,$BE,$56,$BE,$03
	.dc.b	$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$F7,$1D,$4A
	.dc.b	$72,$29,$C4,$A7,$19,$F8,$E4,$9F,$8B,$1C,$E3,$13,$F0,$23,$89,$23
	.dc.b	$2E,$E7,$8B,$E1,$8C,$B6,$32,$F8,$E3,$2D,$8C,$BE,$56,$BD,$DB,$1F
	.dc.b	$7B,$D0,$FF,$11,$79,$6F,$88,$FF,$4F,$92,$FF,$C9,$8F,$F8,$73,$C3
	.dc.b	$FF,$21,$B7,$1E,$13,$FD,$13,$C3,$FF,$12,$4E,$4F,$F5,$7F,$DF,$3E
	.dc.b	$8B,$8D,$7F,$4A,$8C,$5F,$28,$FF,$0C,$46,$2B,$8D,$70,$9F,$F0,$E7
	.dc.b	$2B,$ED,$B1,$1F,$E2,$38,$C6,$1C,$E3,$13,$A7,$FA,$EB,$71,$FF,$1D
	.dc.b	$7E,$11,$EB,$24,$FA,$F3,$FF,$30,$C7,$FB,$23,$E0,$BF,$BA,$B7,$B3
	.dc.b	$35,$FD,$92,$72,$D8,$CB,$FE,$37,$C5,$F1,$EF,$8E,$FF,$36,$FC,$C7
	.dc.b	$F6,$28,$C4,$A3,$13,$AA,$FF,$1E,$4F,$28,$C4,$7F,$8F,$F3,$FF,$2A
	.dc.b	$3F,$54,$87,$AE,$7F,$88,$F9,$4F,$4F,$F8,$2B,$63,$31,$83,$F8,$5B
	.dc.b	$E8,$F9,$1E,$F5,$FE,$84,$C9,$18,$9D,$1E,$99,$F3,$CE,$A3,$11,$E7
	.dc.b	$DB,$F8,$AE,$3E,$88,$F3,$1E,$83,$E4,$57,$F8,$B1,$F8,$DE,$1F,$92
	.dc.b	$31,$1D,$A3,$F9,$0F,$C1,$3F,$0A,$FC,$6C,$FF,$86,$23,$1C,$9A,$9C
	.dc.b	$CF,$F9,$9E,$73,$97,$F3,$6C,$7C,$93,$E7,$8C,$65,$E8,$3B,$FE,$3E
	.dc.b	$F9,$7A,$49,$FF,$13,$D3,$05,$6E,$72,$7E,$13,$2F,$F5,$7F,$E0,$1F
	.dc.b	$83,$6D,$FC,$A6,$BF,$11,$AF,$D3,$FA,$8B,$5F,$8E,$6E,$C6,$98,$FA
	.dc.b	$AF,$F5,$EE,$DF,$91,$6D,$FD,$86,$BF,$49,$AF,$E0,$7E,$AE,$E1,$F9
	.dc.b	$C6,$E0,$69,$8F,$AD,$FF,$65,$FE,$17,$F9,$F6,$DF,$E1,$9A,$FE,$01
	.dc.b	$AF,$FA,$7F,$E6,$DF,$05,$88,$46,$31,$5F,$D0,$8C,$36,$B9,$CF,$EC
	.dc.b	$BD,$3B,$97,$FA,$1E,$BF,$65,$18,$33,$F1,$DF,$8F,$25,$7D,$8A,$47
	.dc.b	$DF,$C7,$F1,$DC,$3F,$75,$BC,$79,$7F,$47,$EA,$FF,$BB,$3E,$07,$F8
	.dc.b	$03,$63,$38,$FF,$BC,$5B,$3F,$C1,$7E,$BF,$51,$FA,$46,$BF,$B3,$6C
	.dc.b	$FF,$0E,$73,$CD,$F9,$C9,$FD,$17,$1F,$E8,$1F,$F6,$54,$65,$FD,$14
	.dc.b	$65,$FF,$67,$78,$FF,$D1,$BC,$5F,$E8,$91,$FE,$CD,$8C,$3F,$D1,$8F
	.dc.b	$D5,$77,$7F,$A3,$67,$F6,$0D,$8C,$E3,$FC,$E3,$0F,$F6,$5B,$FC,$DC
	.dc.b	$FE,$4B,$CC,$5F,$29,$F9,$56,$D7,$E5,$5A,$FE,$55,$AF,$E5,$5A,$7E
	.dc.b	$54,$BC,$3F,$95,$6F,$3F,$7A,$FF,$49,$F8,$1F,$CD,$F3,$B8,$7F,$4E
	.dc.b	$BF,$E7,$0D,$9D,$3F,$CD,$17,$91,$92,$4F,$B6,$32,$DF,$AB,$E4,$47
	.dc.b	$07,$F5,$9A,$6D,$FD,$DB,$C6,$FE,$16,$D8,$FF,$59,$B5,$FA,$9E,$FF
	.dc.b	$8A,$7E,$8D,$7F,$52,$9F,$E2,$5C,$1F,$EC,$19,$CC,$9A,$7C,$07,$13
	.dc.b	$ED,$39,$3F,$08,$CB,$FD,$77,$1F,$EB,$D3,$D3,$82,$4F,$E0,$2B,$F9
	.dc.b	$13,$63,$E7,$FD,$8F,$DF,$FC,$0A,$D1,$A4,$AE,$5E,$87,$A0,$8E,$0F
	.dc.b	$87,$9F,$D2,$7E,$3B,$5C,$ED,$FE,$72,$D7,$FC,$43,$1F,$80,$FA,$FF
	.dc.b	$EB,$FA,$9A,$4C,$BF,$9B,$C3,$ED,$3F,$2D,$F3,$DF,$88,$6D,$FE,$0D
	.dc.b	$AF,$C5,$6B,$F9,$D9,$7D,$AF,$80,$97,$D9,$FF,$3C,$E3,$E6,$AF,$49
	.dc.b	$EC,$CB,$D5,$9F,$B6,$3E,$97,$DB,$7A,$89,$87,$A1,$E7,$3B,$9F,$DE
	.dc.b	$B5,$FC,$7B,$6F,$C1,$1F,$77,$ED,$FE,$29,$2F,$E1,$BF,$23,$4F,$DC
	.dc.b	$7B,$FF,$AA,$6D,$7E,$95,$AF,$E1,$DA,$FC,$3E,$FF,$A2,$4B,$F9,$EF
	.dc.b	$A3,$3F,$79,$27,$E1,$1F,$B8,$FB,$9F,$B2,$4A,$FA,$EF,$DC,$F8,$4F
	.dc.b	$E1,$9C,$BF,$BE,$6D,$7E,$E5,$AF,$E6,$DA,$FC,$6B,$F0,$3F,$48,$87
	.dc.b	$E8,$7E,$EB,$FD,$81,$87,$F0,$36,$C6,$63,$FC,$84,$D3,$FD,$7E,$FF
	.dc.b	$90,$FE,$EF,$F9,$07,$B3,$8E,$34,$62,$7F,$87,$36,$BF,$60,$D7,$CC
	.dc.b	$7C,$43,$CA,$60,$FE,$2D,$EB,$FF,$2D,$9E,$CD,$72,$66,$3F,$79,$19
	.dc.b	$7F,$7D,$EE,$FB,$1F,$CD,$36,$7F,$CB,$8D,$B3,$7E,$FB,$FC,$BF,$3F
	.dc.b	$C9,$DB,$5F,$C2,$5A,$FE,$95,$AF,$E5,$33,$FA,$2F,$F1,$E6,$BE,$29
	.dc.b	$F9,$AD,$3E,$55,$78,$CF,$43,$CC,$F8,$C7,$32,$9F,$1C,$C1,$F2,$8E
	.dc.b	$00,$83,$DD,$FF,$3E,$78,$19,$F8,$B7,$BF,$A9,$CD,$27,$2F,$38,$CC
	.dc.b	$62,$3A,$67,$FE,$AE,$8C,$4E,$AB,$A4,$E5,$B7,$E1,$6D,$F7,$7E,$41
	.dc.b	$FC,$24,$8F,$F2,$46,$BA,$30,$43,$F0,$38,$FB,$0F,$EC,$18,$FF,$77
	.dc.b	$9F,$9F,$19,$9D,$C7,$F7,$B7,$A1,$0F,$D6,$3E,$07,$F0,$10,$F8,$2F
	.dc.b	$D1,$69,$FE,$03,$6C,$64,$FC,$51,$67,$BB,$E4,$3C,$4F,$E1,$6B,$CA
	.dc.b	$7E,$82,$3F,$D1,$27,$FA,$CF,$5F,$F0,$87,$B1,$92,$47,$C1,$AF,$39
	.dc.b	$4F,$CA,$57,$F8,$23,$7F,$F0,$C7,$E1,$7F,$B9,$6D,$7E,$E0,$46,$5A
	.dc.b	$7F,$07,$3C,$E7,$FB,$17,$8B,$E6,$39,$AF,$E0,$91,$8B,$F9,$47,$88
	.dc.b	$F2,$75,$7F,$BF,$27,$F8,$F5,$3E,$FA,$92,$BF,$90,$46,$5F,$F0,$D7
	.dc.b	$3F,$DB,$6C,$7E,$4D,$FE,$B1,$19,$7F,$C0,$0D,$7E,$29,$E4,$BF,$13
	.dc.b	$F2,$67,$FD,$C1,$D1,$B6,$9D,$4B,$87,$F8,$CA,$35,$C4,$E9,$FF,$22
	.dc.b	$7E,$EF,$FD,$36,$CD,$26,$1F,$E3,$3E,$CE,$24,$9E,$D6,$8F,$44,$FC
	.dc.b	$93,$D1,$DB,$FE,$3A,$CB,$F2,$97,$FE,$F7,$EE,$F1,$91,$E6,$F2,$D3
	.dc.b	$DF,$7C,$BA,$BE,$43,$D4,$FF,$00,$77,$FB,$09,$7E,$47,$DF,$E0,$7F
	.dc.b	$98,$7E,$5F,$FC,$19,$26,$92,$BF,$CA,$1E,$17,$C6,$24,$FB,$9F,$26
	.dc.b	$9E,$B9,$FC,$2B,$5F,$D9,$78,$3F,$E7,$2F,$C3,$3F,$C8,$64,$8F,$F1
	.dc.b	$C3,$FE,$34,$FC,$32,$47,$F9,$7F,$C7,$FF,$3B,$46,$23,$FD,$17,$E4
	.dc.b	$FF,$4D,$E8,$FF,$26,$92,$3F,$C8,$4F,$F9,$37,$F0,$C9,$1F,$E4,$DD
	.dc.b	$11,$88,$E7,$1F,$E8,$38,$FF,$2B,$92,$3F,$C9,$8F,$F8,$8C,$8F,$F8
	.dc.b	$F9,$AF,$F9,$57,$CB,$FF,$46,$7C,$47,$F9,$F9,$B1,$FE,$79,$6B,$BA
	.dc.b	$4F,$FA,$13,$8B,$33,$FE,$39,$6C,$B9,$FF,$CF,$B5,$B8,$E3,$5E,$D8
	.dc.b	$FF,$36,$11,$C1,$FD,$2A,$3F,$D2,$DE,$07,$F4,$C8,$FF,$4D,$36,$33
	.dc.b	$1B,$F3,$FF,$D3,$53,$FE,$A5,$24,$7F,$A9,$5A,$FC,$A9,$F1,$B6,$BF
	.dc.b	$D0,$24,$7F,$CC,$AF,$F9,$34,$8E,$0F,$F6,$2F,$BB,$FB,$FE,$17,$A4
	.dc.b	$8C,$36,$FF,$84,$4E,$A7,$F8,$06,$2F,$4F,$FD,$7D,$EB,$FF,$B7,$6F
	.dc.b	$FD,$D7,$8F,$B6,$72,$D8,$CB,$FD,$C5,$53,$D6,$DC,$E6,$31,$3C,$23
	.dc.b	$FD,$99,$18,$BF,$F7,$8C,$E7,$17,$AE,$F0,$BD,$CD,$9F,$5B,$7B,$FF
	.dc.b	$DF,$B1,$FC,$67,$06,$1E,$E2,$46,$5F,$83,$EC,$7F,$B7,$7F,$0B,$FB
	.dc.b	$94,$8C,$67,$F7,$5A,$FA,$DA,$FF,$84,$7D,$C7,$B9,$88,$46,$31,$61
	.dc.b	$F8,$B9,$CD,$6E,$B5,$3E,$2F,$77,$AD,$FC,$C6,$E2,$69,$8F,$6B,$FD
	.dc.b	$ED,$F8,$5F,$A9,$B9,$7E,$55,$AF,$D3,$6B,$F4,$BD,$CB,$87,$FC,$00
	.dc.b	$15,$FE,$FF,$24,$FE,$7F,$DF,$4B,$F0,$1F,$93,$18,$6C,$66,$78,$35
	.dc.b	$FC,$27,$36,$DF,$D5,$35,$CC,$6D,$FC,$3C,$65,$72,$7E,$1B,$93,$F0
	.dc.b	$A4,$AF,$C2,$BF,$D3,$3E,$0D,$F0,$74,$93,$FD,$12,$32,$FE,$9B,$A7
	.dc.b	$F5,$1C,$DF,$C0,$70,$FD,$44,$E7,$2F,$D6,$9F,$AA,$F4,$E4,$7E,$45
	.dc.b	$B5,$F7,$F0,$EA,$DC,$7C,$2D,$AF,$CC,$7C,$27,$F8,$B7,$4F,$E1,$8D
	.dc.b	$7F,$54,$D7,$F5,$2D,$5E,$1B,$E1,$7F,$B5,$B6,$BF,$B0,$B6,$FF,$24
	.dc.b	$DA,$FE,$3D,$8F,$99,$B8,$F9,$9B,$7F,$AD,$F0,$3C,$4F,$F0,$06,$F3
	.dc.b	$7F,$31,$8F,$F2,$44,$FF,$25,$9F,$37,$C4,$3E,$23,$FA,$54,$FF,$2D
	.dc.b	$8C,$57,$F5,$08,$C4,$E9,$FF,$3E,$CE,$BC,$5F,$90,$48,$F2,$B5,$FF
	.dc.b	$41,$FC,$4B,$C4,$FE,$8F,$08,$C6,$1F,$CC,$A3,$FD,$15,$18,$F1,$3F
	.dc.b	$6C,$D8,$FE,$85,$F1,$5F,$CA,$E7,$F7,$4D,$8C,$F2,$76,$8C,$47,$FA
	.dc.b	$53,$7F,$C6,$D7,$DF,$24,$7C,$C7,$F9,$EB,$FE,$AB,$6B,$FE,$DF,$7F
	.dc.b	$84,$7C,$03,$E2,$FF,$C4,$ED,$CB,$FC,$3A,$DE,$3D,$CE,$89,$3A,$6C
	.dc.b	$6A,$30,$47,$04,$BF,$EC,$55,$D5,$B5,$C9,$B5,$C5,$B5,$C5,$FF,$5A
	.dc.b	$D7,$16,$D7,$16,$D7,$16,$D7,$17,$FD,$7D,$5C,$5B,$5C,$5B,$5C,$5B
	.dc.b	$5C,$55,$71,$6D,$71,$6D,$71,$6D,$71,$AE,$04,$AE,$0D,$AE,$0D,$9E
	.dc.b	$13,$A2,$4E,$9B,$1A,$8C,$11,$C1,$2B,$E5,$90,$F8,$A5,$3D,$E5,$3D
	.dc.b	$25,$AF,$23,$6B,$C2,$FF,$87,$7E,$03,$FD,$C6,$DE,$17,$C0,$7F,$57
	.dc.b	$8F,$F7,$0B,$67,$71,$FE,$27,$9F,$F6,$63,$FE,$25,$6B,$9C,$79,$1F
	.dc.b	$C1,$5F,$EC,$4D,$78,$3A,$8E,$31,$FE,$0D,$6B,$FD,$41,$B1,$C6,$3B
	.dc.b	$1D,$23,$24,$8E,$84,$8D,$35,$F0,$98,$21,$82,$1D,$8A,$79,$5E,$4E
	.dc.b	$48,$64,$E6,$60,$F5,$3C,$1D,$1D,$0C,$9C,$0C,$9C,$5F,$80,$E0,$87
	.dc.b	$03,$93,$A7,$07,$18,$F8,$8F,$53,$27,$28,$E0,$48,$DC,$74,$6B,$CE
	.dc.b	$3D,$4F,$22,$3C,$4C,$95,$F0,$35,$E8,$EA,$3A,$C7,$8D,$AF,$46,$C7
	.dc.b	$18,$F1,$1F,$E2,$DF,$55,$E8,$BC,$CF,$CD,$CF,$F9,$EE,$73,$18,$81
	.dc.b	$23,$38,$F3,$AE,$4F,$F7,$72,$4F,$F6,$46,$BC,$87,$C1,$FF,$5C,$8F
	.dc.b	$F2,$03,$63,$FD,$18,$D7,$83,$5F,$14,$FA,$7E,$1B,$C5,$81,$A4,$9F
	.dc.b	$EA,$38,$E5,$B1,$FD,$5B,$8B,$FC,$6E,$D9,$D4,$FE,$B7,$F0,$6F,$F1
	.dc.b	$DF,$C6,$7E,$E5,$B1,$FB,$D6,$BB,$94,$F9,$22,$1F,$37,$1F,$A0,$E0
	.dc.b	$F8,$9D,$8F,$AE,$FD,$DB,$6F,$F8,$73,$67,$FC,$2B,$3F,$C0,$A7,$FD
	.dc.b	$79,$DB,$E2,$9F,$E0,$4D,$8C,$9E,$48,$FE,$29,$1B,$5B,$FE,$01,$F2
	.dc.b	$1F,$3B,$5F,$E3,$D1,$98,$FF,$60,$60,$F9,$BF,$39,$30,$F3,$BF,$C5
	.dc.b	$BE,$A7,$FC,$63,$DB,$E5,$6C,$F8,$EF,$F1,$B1,$8D,$7F,$D9,$8D,$8F
	.dc.b	$91,$97,$8B,$5F,$93,$18,$3C,$E3,$F6,$C7,$C1,$7F,$24,$6E,$3F,$CD
	.dc.b	$DB,$3F,$E8,$F6,$BF,$CD,$FE,$5B,$FC,$F3,$F2,$3F,$CD,$DB,$19,$A4
	.dc.b	$E5,$B1,$FC,$FF,$E4,$F6,$FF,$35,$6B,$FC,$D5,$B1,$FE,$F3,$6F,$EE
	.dc.b	$DF,$E8,$88,$7E,$05,$FE,$EB,$8C,$36,$3F,$D1,$AD,$77,$5F,$EF,$59
	.dc.b	$7C,$BE,$A3,$0D,$7F,$DA,$91,$A9,$E3,$18,$9D,$3F,$E9,$A9,$7C,$7F
	.dc.b	$FA,$73,$F0,$2F,$23,$67,$C0,$D7,$AB,$5F,$E1,$3E,$1F,$FA,$2A,$7F
	.dc.b	$CF,$84,$8F,$F3,$4B,$EF,$DF,$86,$2C,$3F,$8B,$A9,$D3,$6B,$E0,$57
	.dc.b	$0A,$7C,$F7,$E8,$5B,$5F,$8F,$AF,$64,$E4,$FD,$16,$FF,$82,$8F,$E2
	.dc.b	$CD,$7F,$90,$B5,$FC,$77,$07,$EC,$6F,$FC,$EB,$F8,$DF,$D7,$F2,$BF
	.dc.b	$D8,$2D,$C4,$D3,$EE,$73,$EB,$0F,$B0,$D7,$EA,$B5,$FA,$4D,$E5,$73
	.dc.b	$FF,$B7,$EF,$D9,$18,$F5,$FF,$5D,$39,$E8,$FD,$5B,$70,$FD,$41,$0F
	.dc.b	$E6,$9D,$5F,$BD,$6F,$07,$EB,$9F,$87,$B7,$EA,$C9,$97,$EA,$DF,$85
	.dc.b	$F2,$FF,$EE,$0F,$C3,$3F,$30,$DA,$FC,$A3,$5F,$C8,$B5,$FC,$77,$9F
	.dc.b	$6F,$F0,$0D,$E1,$FF,$5B,$5F,$E8,$E3,$1F,$4B,$F7,$49,$5F,$74,$FC
	.dc.b	$43,$F7,$31,$F3,$E1,$FB,$F6,$BF,$A7,$F2,$BF,$0C,$4A,$E5,$8A,$3F
	.dc.b	$93,$69,$F1,$6B,$F9,$21,$FC,$41,$C3,$F8,$4A,$F8,$84,$7C,$0F,$F8
	.dc.b	$F2,$9E,$EF,$E5,$49,$1E,$07,$D0,$F1,$9F,$ED,$07,$53,$99,$FC,$17
	.dc.b	$E8,$FF,$3C,$48,$F4,$BC,$A3,$67,$2E,$6F,$CD,$1F,$9B,$C3,$D2,$4D
	.dc.b	$FF,$01,$3E,$E3,$F8,$67,$D3,$7E,$7C,$98,$7A,$DF,$D1,$7E,$08,$87
	.dc.b	$C9,$7F,$19,$8C,$36,$33,$5F,$C5,$E7,$2F,$F8,$16,$1F,$8D,$FF,$B8
	.dc.b	$23,$E0,$C6,$3A,$BF,$B0,$4E,$67,$83,$5F,$CC,$F8,$9F,$D1,$C8,$FF
	.dc.b	$5C,$AF,$E1,$98,$FA,$A3,$F9,$73,$5F,$D2,$1F,$04,$E6,$71,$DD,$F5
	.dc.b	$BF,$C0,$49,$5F,$75,$FE,$09,$F5,$FF,$0F,$F0,$5F,$E7,$4D,$BF,$E6
	.dc.b	$CD,$7F,$99,$B5,$FE,$61,$F6,$5F,$C3,$49,$7F,$89,$7F,$CF,$FC,$1F
	.dc.b	$5F,$3F,$6D,$7A,$2B,$F9,$47,$A7,$FC,$54,$8F,$E4,$5F,$E3,$1F,$25
	.dc.b	$FC,$94,$91,$FA,$07,$F3,$FC,$5F,$82,$AF,$F0,$2B,$70,$F0,$47,$E5
	.dc.b	$3E,$81,$F4,$4F,$92,$FC,$14,$FF,$A8,$27,$31,$8C,$7F,$C8,$51,$87
	.dc.b	$FC,$9A,$62,$B9,$38,$AF,$EA,$FE,$87,$F8,$CA,$3F,$AF,$36,$33,$1F
	.dc.b	$E3,$03,$25,$7E,$C9,$F7,$0F,$5E,$9F,$21,$F0,$FE,$EE,$F1,$3C,$D1
	.dc.b	$86,$BF,$9D,$8D,$57,$96,$31,$3A,$84,$BD,$0C,$36,$38,$1A,$8C,$9A
	.dc.b	$F0,$FD,$97,$DD,$B3,$D1,$B5,$FD,$0A,$3D,$D3,$F9,$72,$1E,$E3,$F4
	.dc.b	$11,$E1,$7F,$DA,$71,$E1,$3C,$71,$97,$F6,$4D,$8F,$92,$43,$AE,$BF
	.dc.b	$82,$7E,$63,$5F,$F3,$F7,$B0,$F6,$3F,$D1,$BE,$6F,$F2,$B9,$FF,$45
	.dc.b	$CE,$7D,$8F,$F4,$6C,$E6,$7F,$DD,$67,$F8,$43,$C2,$E7,$7C,$9B,$5C
	.dc.b	$1B,$3B,$3A,$CE,$DB,$3B,$7F,$CB,$93,$FE,$88,$23,$FE,$4E,$6B,$FC
	.dc.b	$29,$AF,$03,$E2,$E9,$FC,$7E,$3E,$5E,$5F,$75,$B1,$9C,$F6,$7F,$4B
	.dc.b	$3B,$54,$3B,$14,$FC,$02,$F9,$EE,$BA,$FF,$2D,$6C,$64,$FA,$F1,$97
	.dc.b	$BD,$F8,$9F,$05,$7E,$7A,$FC,$8D,$8C,$BF,$9A,$69,$F9,$AD,$B7,$19
	.dc.b	$AF,$7C,$62,$3F,$57,$1F,$19,$A7,$AE,$7F,$C7,$5C,$1E,$58,$FD,$BE
	.dc.b	$3F,$92,$EB,$FF,$4F,$36,$75,$2E,$3F,$E0,$D8,$7F,$0B,$E3,$FE,$0A
	.dc.b	$DE,$4F,$2C,$EB,$2F,$95,$7E,$23,$E6,$90,$FF,$2E,$73,$7C,$B7,$FC
	.dc.b	$D0,$D7,$FC,$CA,$D7,$71,$F4,$DF,$F3,$77,$3F,$E0,$5B,$1F,$99,$7F
	.dc.b	$CD,$11,$97,$FC,$DE,$D7,$F1,$47,$8C,$F3,$18,$3F,$A0,$C7,$33,$FC
	.dc.b	$C5,$D1,$C5,$A6,$75,$46,$57,$4F,$F7,$3F,$AF,$FE,$A3,$46,$92,$FF
	.dc.b	$D1,$1D,$9C,$49,$3D,$5F,$44,$79,$89,$D2,$EB,$75,$FC,$19,$F5,$DF
	.dc.b	$F6,$E3,$CD,$DB,$E5,$23,$BB,$E5,$56,$F9,$F0,$FE,$2E,$72,$FF,$80
	.dc.b	$BB,$9D,$FF,$0D,$B8,$7C,$76,$CE,$E7,$E2,$78,$1F,$E1,$5E,$2F,$F0
	.dc.b	$91,$34,$FF,$09,$60,$8F,$F0,$1B,$E8,$25,$E0,$F2,$65,$E6,$AF,$C7
	.dc.b	$E5,$E6,$7F,$C5,$DE,$2F,$C8,$F0,$FF,$DA,$37,$FE,$42,$24,$FF,$AD
	.dc.b	$FC,$4F,$C7,$63,$FE,$9F,$6F,$19,$A7,$E4,$E7,$FC,$9D,$5C,$9F,$F3
	.dc.b	$AD,$7D,$1B,$EF,$1F,$EC,$3F,$17,$99,$27,$F3,$51,$FE,$C2,$8E,$93
	.dc.b	$C2,$39,$D7,$57,$FD,$43,$8F,$E0,$B2,$F7,$47,$E6,$5A,$FE,$67,$0F
	.dc.b	$F5,$C5,$7A,$88,$F8,$23,$FD,$99,$1E,$0A,$F8,$C4,$9F,$53,$F9,$BF
	.dc.b	$1B,$E1,$92,$39,$1F,$E7,$E2,$E6,$F1,$FF,$CA,$5F,$85,$7F,$A1,$E4
	.dc.b	$D2,$63,$FE,$94,$8C,$B6,$3F,$D3,$31,$88,$FF,$4B,$B5,$F8,$6F,$98
	.dc.b	$98,$FC,$78,$FD,$2E,$DF,$85,$7E,$33,$5F,$F0,$77,$83,$FE,$B9,$EF
	.dc.b	$FF,$5A,$F9,$9F,$C3,$FC,$1F,$F5,$D6,$3F,$1D,$B3,$FE,$C8,$87,$E3
	.dc.b	$74,$98,$7F,$B2,$A3,$2F,$F9,$AF,$5F,$D2,$CF,$FB,$4C,$91,$9E,$6F
	.dc.b	$F3,$C4,$7A,$78,$BF,$05,$3E,$ED,$7F,$05,$87,$FB,$CF,$D0,$FC,$A9
	.dc.b	$1F,$A2,$FE,$63,$6F,$F4,$17,$D8,$7F,$BD,$89,$7F,$EF,$6A,$47,$FA
	.dc.b	$72,$1E,$9B,$D9,$E3,$1F,$EA,$08,$C5,$70,$8C,$4E,$91,$27,$75,$FE
	.dc.b	$88,$D5,$E5,$EE,$BF,$D2,$D1,$8C,$3A,$47,$F9,$4A,$31,$5C,$6B,$85
	.dc.b	$25,$EA,$FF,$A7,$FE,$C3,$B3,$70,$FA,$0D,$7A,$35,$F0,$D5,$3A,$16
	.dc.b	$BF,$D5,$4A,$BD,$4E,$FC,$D7,$B1,$FE,$AF,$F8,$BF,$F5,$2C,$62,$51
	.dc.b	$89,$D6,$1F,$EA,$DA,$D4,$F0,$F6,$F3,$3F,$E7,$29,$46,$23,$FD,$79
	.dc.b	$18,$BE,$B1,$CE,$74,$FF,$14,$3F,$90,$C7,$E9,$1B,$C3,$FC,$A3,$C3
	.dc.b	$FE,$7C,$47,$FD,$BB,$CF,$FE,$67,$E9,$FD,$17,$9D,$FE,$5D,$E9,$CB
	.dc.b	$70,$FB,$E7,$F9,$DD,$C3,$85,$EB,$FF,$58,$6E,$5F,$D5,$5B,$8F,$E4
	.dc.b	$5B,$5F,$CE,$B9,$BF,$6C,$FE,$EB,$3F,$EE,$5A,$FE,$4B,$2F,$9F,$CD
	.dc.b	$FC,$E2,$31,$5F,$36,$3F,$9B,$60,$F3,$FC,$44,$CF,$C4,$FF,$2B,$CB
	.dc.b	$BF,$83,$FC,$AB,$0F,$06,$9F,$73,$DC,$EC,$4C,$3F,$90,$E6,$8F,$E8
	.dc.b	$9F,$04,$F7,$FF,$DE,$5F,$0B,$FA,$D8,$C5,$A3,$13,$AC,$7E,$58,$AD
	.dc.b	$CF,$ED,$72,$FE,$EA,$DC,$FF,$BF,$B7,$83,$A5,$F2,$AD,$BD,$3E,$07
	.dc.b	$FB,$AF,$C1,$FE,$FD,$B1,$9F,$09,$A2,$B5,$18,$7F,$83,$77,$FF,$02
	.dc.b	$6F,$1E,$6F,$F8,$1D,$F1,$6C,$F9,$48,$F9,$74,$FF,$0B,$C6,$23,$F4
	.dc.b	$93,$FE,$19,$3F,$D3,$05,$E0,$FF,$10,$B8,$E1,$FE,$30,$D9,$FD,$8B
	.dc.b	$B2,$46,$A7,$89,$23,$33,$C2,$75,$18,$8F,$1C,$60,$8E,$C8,$68,$A7
	.dc.b	$EE,$63,$93,$63,$8B,$B2,$19,$38,$9F,$BE,$8E,$2D,$8E,$2E,$CE,$26
	.dc.b	$4E,$31,$F0,$DC,$B5,$E0,$43,$81,$A9,$F8,$93,$DA,$32,$6E,$79,$38
	.dc.b	$23,$82,$4F,$28,$F1,$B5,$ED,$1E,$C3,$65,$8D,$47,$76,$BD,$9B,$1C
	.dc.b	$27,$CC,$D9,$EE,$F3,$24,$6A,$79,$FB,$1F,$ED,$FA,$FE,$C0,$D7,$F6
	.dc.b	$2D,$7F,$DB,$67,$ED,$3E,$13,$FA,$ED,$FF,$67,$84,$61,$FE,$D4,$D7
	.dc.b	$81,$FA,$C3,$25,$33,$1C,$E3,$84,$7F,$2C,$6B,$FD,$B1,$AF,$80,$E4
	.dc.b	$6C,$A6,$CA,$6E,$3B,$38,$8E,$64,$70,$48,$D9,$1F,$4B,$5F,$49,$82
	.dc.b	$18,$21,$E5,$78,$34,$D9,$4D,$9E,$97,$6D,$36,$53,$67,$B4,$C9,$4C
	.dc.b	$C7,$B6,$39,$C6,$4F,$01,$0E,$85,$34,$60,$86,$09,$1D,$5F,$0B,$5F
	.dc.b	$09,$F2,$CC,$94,$CC,$79,$23,$84,$78,$5A,$F8,$1A,$F8,$0E,$46,$CA
	.dc.b	$6C,$A6,$E3,$B3,$F8,$4D,$BF,$D7,$FE,$CF,$F4,$92,$47,$F3,$5E,$04
	.dc.b	$7F,$91,$7F,$31,$7C,$FF,$F4,$68,$C3,$6B,$FA,$CC,$EB,$0E,$31,$89
	.dc.b	$D3,$FE,$08,$E2,$FF,$63,$DF,$F4,$EA,$DC,$E7,$8B,$6F,$F9,$32,$B6
	.dc.b	$FF,$4E,$F8,$5F,$EA,$34,$69,$23,$FC,$2B,$0F,$71,$F7,$3F,$D7,$BE
	.dc.b	$0B,$FB,$7C,$E6,$31,$5C,$2B,$72,$AD,$C6,$1F,$F3,$47,$D7,$3E,$5B
	.dc.b	$8D,$FF,$89,$23,$17,$C6,$3F,$BE,$C6,$2B,$8D,$70,$9F,$F3,$4F,$E3
	.dc.b	$EF,$8E,$FD,$CB,$70,$FE,$20,$D9,$FF,$3C,$35,$F4,$7D,$67,$39,$F8
	.dc.b	$0D,$C8,$D3,$C1,$2F,$B2,$FF,$35,$D7,$9E,$3F,$C7,$B1,$8C,$BF,$D0
	.dc.b	$CD,$8F,$F0,$BB,$5F,$E4,$B6,$FB,$CF,$E2,$77,$F4,$5B,$19,$8F,$EF
	.dc.b	$2D,$9F,$F2,$8F,$E0,$BF,$CB,$B0,$B1,$F1,$7C,$84,$E6,$3F,$CA,$31
	.dc.b	$87,$D7,$DF,$FE,$96,$AF,$C2,$92,$3F,$C2,$AF,$F8,$BB,$EB,$DF,$AD
	.dc.b	$FD,$B9,$FF,$2D,$C6,$38,$BE,$A3,$73,$F8,$92,$8C,$5F,$E1,$3F,$04
	.dc.b	$FF,$14,$FE,$19,$6C,$D2,$5F,$E3,$7F,$05,$DB,$67,$6D,$9F,$C7,$CE
	.dc.b	$A3,$E6,$FE,$09,$FE,$36,$FA,$6F,$C9,$B6,$3F,$30,$D7,$76,$9F,$CC
	.dc.b	$F8,$5F,$E3,$EF,$94,$B2,$69,$2B,$F4,$5F,$33,$FE,$44,$E2,$FD,$0B
	.dc.b	$63,$35,$F4,$5F,$C0,$C6,$5F,$F4,$2F,$84,$F1,$DE,$1B,$E5,$7F,$CB
	.dc.b	$04,$AF,$E2,$0F,$F9,$8B,$E6,$9F,$6E,$FB,$5F,$F2,$AC,$E6,$31,$9F
	.dc.b	$3B,$FF,$49,$DF,$0B,$DC,$62,$B7,$7F,$BA,$EF,$FD,$21,$30,$FB,$A7
	.dc.b	$F0,$77,$F5,$1E,$7C,$8C,$F6,$7F,$9E,$B9,$FF,$81,$CF,$EB,$7E,$7F
	.dc.b	$FC,$D3,$F8,$E7,$F0,$66,$E7,$FC,$2D,$AF,$E9,$9A,$FF,$9C,$2D,$F4
	.dc.b	$58,$C7,$F8,$C3,$63,$30,$BF,$E2,$73,$CB,$BF,$F4,$52,$FC,$33,$F5
	.dc.b	$2D,$C3,$F4,$FC,$BF,$E1,$A9,$C9,$FE,$89,$D2,$4F,$F4,$16,$BF,$EF
	.dc.b	$CE,$17,$E7,$EF,$AF,$FE,$31,$18,$6C,$67,$1F,$D3,$CE,$5F,$93,$F4
	.dc.b	$DF,$E8,$EF,$9B,$FE,$56,$DC,$3F,$98,$35,$FE,$16,$D7,$E3,$75,$FF
	.dc.b	$A5,$BF,$0E,$FC,$C9,$23,$F9,$C4,$65,$FE,$75,$C9,$FE,$9C,$D7,$F0
	.dc.b	$18,$FC,$39,$FE,$2F,$39,$77,$19,$7E,$A7,$26,$5F,$E3,$AD,$73,$1B
	.dc.b	$3F,$A7,$46,$5F,$EA,$1F,$15,$FE,$AC,$FC,$3B,$FA,$4B,$6B,$FA,$73
	.dc.b	$5F,$E5,$2D,$43,$B3,$F5,$5D,$7E,$58,$C7,$67,$EB,$33,$FF,$61,$12
	.dc.b	$3F,$47,$C3,$F4,$E3,$F5,$8D,$8C,$F1,$7A,$F8,$BE,$B5,$7C,$6E,$D9
	.dc.b	$5F,$11,$AF,$E0,$EF,$EF,$69,$D3,$0F,$46,$1F,$7E,$31,$A9,$FE,$D3
	.dc.b	$E6,$7F,$B3,$BF,$0E,$B2,$69,$31,$F2,$C3,$F0,$EE,$64,$8C,$BF,$EE
	.dc.b	$16,$1F,$E1,$92,$9F,$D5,$3E,$B7,$FB,$11,$27,$F7,$6F,$F6,$5F,$AF
	.dc.b	$FF,$70,$FE,$2B,$FD,$F9,$B5,$FE,$03,$6B,$FD,$79,$AF,$EE,$7E,$C7
	.dc.b	$FA,$E9,$2F,$FA,$EB,$FE,$2F,$E1,$FE,$43,$3E,$43,$EE,$F1,$FD,$C7
	.dc.b	$14,$FC,$7B,$FD,$02,$D8,$FE,$D7,$18,$3F,$D0,$73,$C6,$3E,$01,$0F
	.dc.b	$E6,$35,$E0,$7F,$C4,$73,$FC,$CA,$BF,$8F,$B7,$B3,$FC,$46,$D9,$DF
	.dc.b	$2F,$F8,$2E,$3F,$A5,$53,$F6,$5D,$F9,$08,$C5,$F6,$8C,$4E,$A7,$33
	.dc.b	$FE,$AA,$A7,$D2,$75,$24,$F9,$1F,$F0,$AE,$7E,$DA,$F5,$E3,$FE,$B1
	.dc.b	$8C,$47,$CA,$CF,$DC,$F5,$E5,$F4,$13,$0F,$7B,$EB,$FB,$7F,$47,$E2
	.dc.b	$3F,$01,$B5,$F1,$9A,$F9,$DA,$F7,$C7,$C5,$87,$82,$30,$7D,$AC,$DF
	.dc.b	$73,$F2,$49,$97,$D0,$7E,$67,$BD,$F2,$7F,$11,$FF,$62,$4F,$E0,$63
	.dc.b	$0F,$D8,$6D,$FC,$96,$BE,$03,$F0,$05,$3E,$39,$F7,$3E,$EF,$EB,$7D
	.dc.b	$A7,$6C,$51,$8B,$FC,$2B,$6B,$EB,$35,$FB,$D2,$FA,$CF,$F6,$E5,$7E
	.dc.b	$31,$B3,$F7,$5A,$F0,$6C,$7E,$76,$FF,$CE,$4F,$E5,$1B,$3F,$85,$6B
	.dc.b	$E0,$6B,$FC,$83,$4F,$A2,$47,$E8,$57,$6B,$F3,$5F,$FB,$CC,$FD,$64
	.dc.b	$64,$F2,$CF,$33,$F6,$85,$3E,$59,$4F,$96,$63,$83,$F2,$2B,$E1,$5E
	.dc.b	$2F,$FA,$83,$E5,$DF,$0F,$FC,$EA,$B5,$39,$8C,$57,$0F,$85,$FE,$7F
	.dc.b	$7C,$2F,$FA,$0C,$E5,$46,$5E,$73,$C2,$B9,$DF,$59,$E2,$D7,$67,$28
	.dc.b	$FD,$59,$E5,$7F,$38,$D7,$E7,$37,$0F,$EA,$2D,$E2,$FF,$1D,$B7,$5F
	.dc.b	$25,$72,$3F,$D9,$51,$FC,$33,$3F,$E0,$CF,$E4,$5B,$A7,$E2,$4F,$8E
	.dc.b	$58,$FE,$19,$3E,$63,$99,$0C,$1F,$87,$87,$C7,$FE,$77,$7F,$E2,$8D
	.dc.b	$8C,$9F,$8F,$2C,$FE,$5D,$F0,$B6,$BC,$D1,$88,$FC,$D3,$4F,$E6,$DC
	.dc.b	$1F,$C8,$7E,$3B,$F5,$8D,$8F,$D5,$88,$CB,$4F,$54,$FC,$73,$D6,$53
	.dc.b	$F5,$FC,$5F,$CB,$7F,$0E,$FD,$C3,$67,$F8,$20,$8C,$B4,$FE,$7F,$CB
	.dc.b	$F0,$AA,$C7,$C6,$6E,$5F,$D0,$27,$51,$E2,$3F,$02,$79,$8A,$7F,$23
	.dc.b	$97,$E1,$DF,$CA,$5B,$1F,$CA,$5A,$EE,$3F,$A6,$F4,$F6,$8F,$E1,$6D
	.dc.b	$9F,$EC,$0D,$E6,$FF,$25,$61,$F8,$DD,$65,$F8,$F8,$BF,$26,$FF,$5C
	.dc.b	$EA,$E0,$4C,$7D,$D1,$FD,$3E,$3A,$76,$7F,$57,$FC,$D7,$FA,$A1,$27
	.dc.b	$FA,$C5,$23,$FB,$1F,$27,$8C,$91,$D2,$7F,$91,$C7,$AD,$F8,$A7,$D9
	.dc.b	$29,$88,$FE,$F4,$7E,$98,$B8,$FF,$70,$8F,$D2,$9F,$DD,$8A,$7D,$3C
	.dc.b	$7D,$ED,$E5,$FD,$8B,$FC,$36,$9E,$AB,$C0,$E3,$3F,$E1,$59,$F0,$4E
	.dc.b	$5B,$3A,$78,$F8,$3F,$DE,$FF,$44,$FF,$11,$41,$A4,$BF,$05,$BC,$4F
	.dc.b	$B4,$DA,$F8,$C7,$92,$32,$7C,$7C,$BD,$75,$F1,$DA,$FC,$AF,$27,$FD
	.dc.b	$11,$F8,$7F,$F9,$28,$91,$FE,$44,$7F,$C9,$3F,$87,$C8,$F1,$F9,$5F
	.dc.b	$E1,$38,$C4,$7F,$95,$F9,$BF,$C2,$DF,$87,$FF,$95,$09,$1F,$E4,$E7
	.dc.b	$FD,$0B,$F3,$72,$3F,$D0,$6D,$79,$68,$FC,$3B,$FC,$BA,$48,$FF,$2B
	.dc.b	$BF,$C8,$C8,$FF,$2D,$D3,$FC,$F3,$18,$8F,$F3,$FC,$AF,$FD,$0A,$D8
	.dc.b	$FF,$42,$B5,$DC,$7F,$B4,$23,$C5,$1F,$4B,$D3,$FF,$19,$56,$49,$1F
	.dc.b	$E9,$C7,$FD,$42,$47,$07,$FA,$A2,$5F,$87,$FF,$43,$7F,$17,$E2,$7F
	.dc.b	$8F,$63,$1C,$3F,$D7,$A7,$75,$C6,$7F,$D7,$70,$FB,$0A,$49,$FF,$5E
	.dc.b	$C6,$5F,$F3,$EF,$67,$FA,$18,$8E,$0F,$91,$CE,$E9,$FD,$84,$7F,$80
	.dc.b	$67,$FA,$81,$1F,$F6,$E9,$F0,$8B,$5F,$19,$FF,$55,$FB,$7F,$EE,$6D
	.dc.b	$3F,$DC,$A7,$FB,$AA,$5F,$8E,$D2,$57,$E7,$BD,$DF,$F2,$F4,$7F,$BF
	.dc.b	$67,$73,$FE,$FD,$6B,$FE,$FD,$6B,$FE,$FD,$F7,$F8,$BF,$5E,$3F,$4B
	.dc.b	$B3,$83,$F5,$24,$8C,$FB,$5F,$AC,$23,$83,$F8,$97,$4B,$E0,$5F,$05
	.dc.b	$FE,$73,$FA,$3F,$EF,$93,$98,$C6,$5F,$E1,$38,$C5,$AB,$77,$CA,$31
	.dc.b	$5F,$4B,$86,$BF,$E7,$2A,$FF,$00,$11,$FE,$07,$ED,$FF,$A0,$FF,$11
	.dc.b	$FC,$3B,$63,$36,$9C,$B6,$3F,$25,$F4,$5F,$8E,$E8,$FF,$46,$36,$4D
	.dc.b	$3F,$27,$F0,$FF,$E9,$1F,$AD,$B2,$69,$2F,$F8,$BF,$C4,$FF,$A5,$FF
	.dc.b	$11,$FC,$DB,$63,$34,$9C,$B6,$3F,$8F,$FC,$66,$DF,$F7,$AC,$7F,$30
	.dc.b	$8C,$47,$FA,$8F,$91,$18,$9D,$1C,$E7,$FD,$FB,$F1,$AF,$8F,$7C,$7F
	.dc.b	$F6,$58,$C3,$63,$35,$FD,$46,$72,$FE,$93,$E4,$1F,$39,$FD,$4D,$B1
	.dc.b	$98,$C1,$FD,$C7,$B9,$F1,$BE,$AF,$03,$FC,$85,$18,$CF,$53,$97,$F5
	.dc.b	$73,$FC,$0B,$E4,$BF,$A4,$6C,$69,$31,$FF,$21,$D3,$E4,$DF,$4D,$FD
	.dc.b	$9D,$B3,$FB,$C6,$E1,$FE,$51,$6D,$7F,$4E,$F9,$5F,$E9,$A4,$AF,$D4
	.dc.b	$BF,$DC,$7E,$59,$E0,$7F,$60,$A4,$62,$BC,$6D,$F5,$3F,$CB,$CD,$9F
	.dc.b	$F3,$17,$CB,$BD,$0F,$EF,$8D,$BF,$E2,$2D,$CF,$FC,$CC,$DA,$FE,$C7
	.dc.b	$CB,$FD,$88,$95,$FB,$E7,$E8,$AF,$9E,$FF,$37,$72,$7D,$66,$F2,$7F
	.dc.b	$BA,$DB,$3B,$BF,$F3,$B7,$CD,$3E,$6F,$FD,$E1,$F3,$DF,$E7,$78,$C5
	.dc.b	$23,$13,$AA,$E2,$D7,$FD,$29,$F3,$AF,$82,$C4,$04,$66,$FA,$47,$FA
	.dc.b	$1E,$31,$7F,$E8,$76,$C7,$F0,$78,$FF,$42,$11,$C1,$FE,$8F,$F9,$EF
	.dc.b	$D4,$F1,$7F,$A3,$DB,$19,$3E,$D3,$87,$FD,$54,$D7,$CC,$D4,$D2,$3C
	.dc.b	$89,$7F,$94,$8C,$B6,$33,$3E,$36,$9E,$38,$D1,$23,$4D,$8D,$46,$08
	.dc.b	$E0,$87,$A8,$EA,$58,$E4,$D8,$E2,$D3,$89,$F0,$0E,$25,$8E,$2D,$8E
	.dc.b	$2D,$38,$9F,$18,$E2,$58,$E2,$D8,$E2,$D3,$8C,$FC,$88,$CB,$63,$27
	.dc.b	$47,$0F,$A4,$87,$8C,$A7,$52,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$3E
	.dc.b	$32,$4F,$BA,$32,$D8,$FE,$11,$F1,$5F,$EE,$F7,$FD,$C8,$D7,$FD,$DB
	.dc.b	$F1,$3F,$EE,$57,$FD,$D6,$7E,$6F,$E0,$7F,$DA,$ED,$AD,$5A,$73,$3F
	.dc.b	$9A,$6B,$F9,$A6,$BF,$9A,$6C,$71,$9E,$71,$82,$38,$24,$FE,$AE,$32
	.dc.b	$D8,$CB,$E5,$30,$43,$04,$3C,$CF,$06,$9B,$29,$B3,$D4,$ED,$A6,$CA
	.dc.b	$6C,$F7,$3B,$69,$B2,$9B,$9F,$84,$E5,$AE,$67,$DE,$D9,$E9,$3A,$8F
	.dc.b	$39,$0F,$19,$4E,$E5,$78,$3A,$21,$A2,$BA,$70,$47,$04,$9F,$2C,$75
	.dc.b	$6B,$C9,$AF,$16,$C7,$19,$F5,$C6,$08,$E0,$93,$F1,$63,$2D,$8C,$BF
	.dc.b	$E3,$BA,$FE,$7C,$CF,$5D,$EA,$7F,$9B,$E7,$33,$FD,$0E,$73,$E6,$27
	.dc.b	$32,$9C,$C7,$F5,$82,$47,$2E,$3F,$F0,$9C,$62,$3C,$3F,$3F,$FE,$61
	.dc.b	$C7,$A4,$78,$23,$13,$D9,$B1,$9F,$45,$5C,$23,$FC,$67,$F0,$6F,$AD
	.dc.b	$CF,$C8,$FF,$A6,$E3,$14,$8C,$4E,$AB,$FD,$D5,$5A,$9F,$EE,$31,$0F
	.dc.b	$57,$47,$ED,$B3,$FF,$16,$C7,$F8,$2B,$D7,$FD,$CF,$71,$A4,$CB,$E3
	.dc.b	$8C,$7F,$CD,$D8,$F9,$7D,$5F,$16,$1E,$A6,$BE,$E9,$DB,$E2,$AF,$E2
	.dc.b	$58,$F8,$A3,$07,$9C,$FB,$BF,$44,$FA,$CF,$E1,$AD,$AF,$E3,$4D,$9F
	.dc.b	$B6,$D7,$FD,$0D,$E2,$FF,$BE,$DF,$F4,$5B,$5C,$FC,$17,$F1,$AF,$63
	.dc.b	$FC,$7A,$FC,$5D,$E6,$1E,$66,$D7,$E3,$3C,$CF,$F3,$07,$A5,$FC,$89
	.dc.b	$B5,$F8,$C6,$CE,$FC,$E5,$7F,$28,$AF,$95,$F8,$4F,$E0,$BF,$1A,$FB
	.dc.b	$CD,$BF,$B6,$D7,$F0,$AD,$7E,$67,$E1,$96,$BE,$A3,$78,$4D,$31,$F8
	.dc.b	$0F,$E4,$3F,$42,$FC,$63,$6F,$F1,$2D,$7F,$26,$D7,$ED,$7E,$1F,$70
	.dc.b	$FC,$13,$70,$34,$C7,$E1,$FF,$9E,$FC,$97,$F3,$4D,$BF,$CB,$B5,$FD
	.dc.b	$0B,$5F,$DC,$78,$3F,$EE,$9F,$B9,$FD,$BF,$83,$FB,$A6,$C6,$67,$FD
	.dc.b	$7D,$1F,$EC,$8F,$C3,$9F,$1F,$FA,$E8,$C3,$67,$FD,$95,$3A,$D1,$5A
	.dc.b	$6C,$FE,$B3,$EB,$7E,$C6,$A6,$92,$7F,$01,$C9,$F7,$9B,$C9,$F8,$58
	.dc.b	$C5,$FE,$0E,$39,$F3,$7F,$02,$FC,$4B,$F7,$CD,$9F,$DD,$B5,$FE,$0A
	.dc.b	$D7,$F5,$9F,$63,$70,$FD,$A3,$72,$34,$E1,$D5,$F9,$E9,$FE,$42,$DA
	.dc.b	$FD,$2F,$57,$E8,$FA,$FF,$4D,$1F,$CB,$78,$7F,$96,$3F,$CC,$48,$67
	.dc.b	$C7,$FE,$34,$AD,$F6,$AF,$9A,$FE,$3F,$E4,$FF,$BF,$23,$19,$F0,$AD
	.dc.b	$D6,$AF,$FD,$FD,$5A,$BF,$CB,$DB,$F0,$9A,$4A,$FC,$CC,$65,$FE,$47
	.dc.b	$D5,$FC,$BA,$7F,$95,$36,$33,$3B,$7F,$98,$C6,$5F,$E6,$7F,$21,$E3
	.dc.b	$6C,$FB,$8F,$E4,$51,$98,$FE,$3C,$D8,$7D,$CF,$F3,$E7,$83,$70,$FE
	.dc.b	$AB,$1A,$38,$46,$6B,$FA,$EC,$72,$DB,$24,$79,$BE,$18,$FE,$B5,$B3
	.dc.b	$67,$D9,$7E,$99,$B3,$FD,$93,$AF,$FB,$1C,$E4,$FA,$9B,$FF,$54,$6E
	.dc.b	$3C,$2B,$FD,$59,$C1,$F5,$A3,$11,$F7,$F8,$3E,$DD,$7F,$57,$DB,$94
	.dc.b	$FF,$54,$6B,$F2,$A3,$55,$CA,$FA,$1E,$02,$4F,$E0,$B0,$FE,$CF,$1F
	.dc.b	$23,$57,$D0,$7F,$6A,$FC,$0B,$0D,$BF,$93,$3A,$C1,$5A,$6C,$F5,$5E
	.dc.b	$2F,$C0,$24,$65,$F9,$4C,$3E,$56,$FF,$3D,$C1,$FE,$D1,$9F,$C0,$C7
	.dc.b	$F3,$FB,$FF,$0A,$C7,$F6,$A6,$CF,$AE,$75,$F7,$1F,$E0,$4F,$CD,$3D
	.dc.b	$CD,$9F,$5B,$5E,$D4,$8C,$47,$F8,$32,$3F,$DD,$6F,$C7,$F6,$7C,$44
	.dc.b	$C3,$FD,$05,$C3,$FE,$24,$9F,$49,$FC,$C3,$8B,$39,$7B,$63,$E5,$FD
	.dc.b	$B7,$AC,$97,$EB,$7F,$D0,$37,$CE,$7E,$2D,$F2,$23,$F6,$7F,$06,$F8
	.dc.b	$84,$BF,$88,$FC,$7F,$33,$FC,$4B,$F6,$3F,$7D,$B5,$F7,$1A,$FC,$26
	.dc.b	$BF,$07,$4F,$A0,$4B,$FA,$0F,$9B,$3F,$64,$FF,$19,$3E,$09,$FA,$AC
	.dc.b	$7E,$B7,$B3,$F6,$49,$78,$3E,$D7,$DE,$FE,$52,$BF,$1E,$D7,$EA,$B5
	.dc.b	$FA,$8D,$7B,$1E,$0D,$3E,$43,$E3,$3F,$33,$F1,$DF,$E4,$67,$FD,$6B
	.dc.b	$5F,$C3,$23,$0E,$8F,$C8,$47,$E1,$23,$2F,$E1,$BF,$00,$FF,$27,$FE
	.dc.b	$13,$CB,$44,$62,$FF,$4E,$DA,$E1,$0A,$FF,$60,$E1,$E6,$9F,$C3,$1E
	.dc.b	$9D,$3E,$5B,$F9,$1F,$C0,$7F,$90,$5F,$EC,$9B,$1F,$98,$6B,$B8,$FE
	.dc.b	$4C,$FE,$73,$F0,$2F,$4B,$F9,$B6,$C7,$F2,$C6,$E4,$6F,$E0,$7F,$E6
	.dc.b	$89,$FE,$02,$DA,$FD,$3B,$5F,$D3,$35,$FA,$67,$E6,$0F,$66,$FF,$B7
	.dc.b	$AE,$E4,$AF,$9F,$3E,$56,$BC,$8F,$B4,$43,$E3,$9D,$48,$64,$E0,$07
	.dc.b	$C0,$3E,$57,$FD,$1B,$D6,$E2,$FF,$7D,$D7,$F2,$26,$C6,$63,$95,$71
	.dc.b	$F0,$BA,$4E,$67,$AD,$F4,$BE,$24,$8E,$31,$97,$F8,$24,$7F,$4B,$23
	.dc.b	$FB,$E6,$BF,$E0,$36,$BC,$0F,$E6,$78,$FF,$A8,$23,$F2,$4D,$C7,$27
	.dc.b	$B4,$B1,$F9,$0D,$BF,$AE,$3D,$2B,$F6,$CF,$69,$FE,$DF,$A7,$FA,$B3
	.dc.b	$E2,$7F,$3C,$D8,$FC,$18,$8C,$B4,$EE,$4A,$FC,$93,$E3,$6D,$7E,$A7
	.dc.b	$7F,$F5,$CF,$C5,$DA,$34,$8F,$C5,$A8,$7C,$52,$9F,$CD,$F8,$FF,$D8
	.dc.b	$7F,$89,$7E,$D5,$B3,$FB,$41,$19,$69,$FC,$07,$0F,$DE,$72,$FC,$BB
	.dc.b	$FE,$05,$3E,$97,$FA,$24,$65,$FA,$87,$8C,$86,$0F,$E2,$F0,$FC,$43
	.dc.b	$FB,$87,$3F,$FB,$61,$B1,$99,$FE,$42,$D7,$FA,$AD,$FD,$F7,$FB,$9B
	.dc.b	$5C,$C7,$E4,$DF,$EB,$11,$97,$C4,$7D,$E3,$E3,$94,$FE,$C3,$CB,$B9
	.dc.b	$FF,$17,$B4,$FF,$16,$4F,$F8,$1D,$C7,$47,$FB,$C3,$E8,$3F,$A6,$51
	.dc.b	$A4,$BF,$F0,$4F,$5F,$32,$4F,$9C,$FF,$1E,$BE,$78,$F2,$1E,$8E,$97
	.dc.b	$1F,$E6,$23,$EF,$D7,$F9,$13,$B3,$CC,$47,$A6,$48,$E9,$7C,$BA,$BE
	.dc.b	$35,$F6,$24,$67,$4F,$EF,$D8,$72,$ED,$F8,$44,$7A,$F7,$17,$E6,$1F
	.dc.b	$E0,$E6,$E1,$FE,$0E,$6B,$FE,$0E,$9E,$13,$FE,$17,$9F,$3E,$9F,$3C
	.dc.b	$8F,$9B,$1F,$25,$FD,$86,$BF,$8E,$F0,$FF,$C9,$5E,$A7,$F9,$38,$91
	.dc.b	$FE,$33,$7F,$C6,$DF,$89,$48,$FF,$1D,$F7,$7A,$63,$11,$FE,$54,$F2
	.dc.b	$7F,$D1,$7C,$BF,$E5,$82,$47,$F8,$F5,$FF,$27,$FC,$6C,$8F,$F2,$A6
	.dc.b	$68,$C4,$73,$F2,$3E,$34,$7F,$98,$89,$1F,$E4,$77,$FC,$D8,$47,$FC
	.dc.b	$3E,$D7,$FC,$6B,$E5,$FC,$FF,$8A,$FF,$2C,$36,$3F,$CA,$AD,$77,$69
	.dc.b	$FF,$2F,$39,$8F,$F1,$E3,$4F,$F4,$41,$B7,$FC,$BE,$E7,$D1,$79,$B4
	.dc.b	$91,$B9,$FF,$4D,$D7,$FA,$92,$1F,$87,$D2,$4F,$FA,$99,$FF,$58,$91
	.dc.b	$FF,$57,$B5,$F9,$11,$D5,$FF,$42,$65,$E9,$E0,$FF,$44,$5F,$FA,$30
	.dc.b	$FF,$4A,$12,$7F,$C8,$ED,$7F,$D1,$F9,$3D,$1F,$E0,$95,$FE,$D9,$E7
	.dc.b	$FF,$6B,$E3,$FC,$26,$31,$7F,$EA,$8D,$3F,$DC,$32,$FC,$83,$FD,$5F
	.dc.b	$18,$6F,$A7,$FA,$36,$CE,$EB,$FD,$E2,$D7,$FD,$E2,$D7,$F5,$54,$F5
	.dc.b	$39,$47,$0A,$FF,$7F,$57,$F1,$DF,$41,$EB,$BD,$7F,$E3,$D1,$86,$C6
	.dc.b	$67,$F8,$76,$2F,$63,$F8,$F3,$63,$31,$83,$E2,$F4,$BD,$AF,$DF,$7E
	.dc.b	$51,$FE,$0B,$D3,$FD,$73,$39,$B4,$E6,$7F,$DF,$13,$98,$C6,$E8,$C4
	.dc.b	$FF,$9F,$27,$FD,$98,$47,$07,$D1,$F7,$3F,$87,$7E,$4F,$F4,$1B,$19
	.dc.b	$C1,$39,$6C,$7F,$BD,$3F,$39,$FC,$A1,$1C,$1F,$EF,$E7,$ED,$FC,$07
	.dc.b	$F1,$B6,$C9,$A7,$E0,$3E,$0B,$F9,$07,$E4,$16,$4D,$25,$FE,$27,$E1
	.dc.b	$36,$D9,$DB,$67,$F1,$73,$A8,$FF,$1C,$71,$FD,$BA,$FC,$07,$17,$DB
	.dc.b	$E7,$FC,$19,$2B,$E8,$7C,$3F,$E2,$6B,$F9,$EB,$67,$EA,$4A,$30,$F0
	.dc.b	$E7,$FD,$1F,$57,$E9,$6F,$F4,$4D,$BF,$D2,$92,$BE,$F7,$C3,$BC,$FF
	.dc.b	$EB,$0D,$BF,$E9,$AD,$AF,$0B,$6B,$C1,$3E,$6C,$3F,$B0,$63,$E4,$9F
	.dc.b	$C7,$35,$FD,$2F,$C4,$7F,$8C,$DF,$DA,$35,$FF,$73,$74,$A3,$11,$C2
	.dc.b	$FF,$C6,$CF,$E5,$7E,$2D,$F1,$7F,$DB,$DB,$3F,$D9,$1B,$9F,$C8,$6D
	.dc.b	$7A,$23,$E5,$E3,$E5,$8F,$99,$97,$A8,$FD,$6F,$C6,$78,$B0,$66,$8C
	.dc.b	$36,$3C,$B5,$EA,$9F,$4A,$F8,$E7,$C7,$7F,$39,$8C,$36,$33,$5F,$C9
	.dc.b	$E7,$2D,$C7,$F0,$9E,$1F,$A3,$2E,$51,$8A,$FA,$73,$FC,$9D,$FF,$7C
	.dc.b	$35,$FF,$6D,$3F,$8A,$F9,$07,$C8,$7F,$85,$DB,$87,$F8,$31,$BA,$EE
	.dc.b	$74,$49,$D3,$63,$51,$82,$38,$25,$FF,$0D,$AE,$AD,$AE,$4D,$AE,$2D
	.dc.b	$AE,$2F,$F1,$9A,$E2,$DA,$E2,$DA,$E2,$DA,$E2,$FF,$AA,$EB,$8B,$6B
	.dc.b	$8B,$6B,$8B,$6B,$8A,$AE,$2D,$AE,$2D,$AE,$2D,$AE,$35,$C0,$95,$C1
	.dc.b	$B5,$C1,$B3,$C2,$74,$49,$D3,$63,$51,$82,$38,$25,$7C,$B2,$1F,$14
	.dc.b	$A7,$BC,$A7,$A4,$B5,$E4,$6D,$78,$5B,$5E,$06,$D7,$81,$EC,$43,$B1
	.dc.b	$4F,$F6,$8F,$C1,$7F,$1E,$69,$FC,$83,$E0,$FF,$DB,$0D,$F8,$34,$EF
	.dc.b	$17,$6F,$F9,$6E,$BF,$3C,$D7,$F2,$0D,$7F,$14,$D7,$F0,$8D,$7E,$F3
	.dc.b	$63,$AE,$3D,$E3,$04,$70,$49,$F0,$B5,$F0,$B5,$F0,$B5,$F0,$B5,$F0
	.dc.b	$B6,$7C,$31,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$AF,$47,$91,$0E,$45
	.dc.b	$39,$14,$E4,$57,$93,$C0,$87,$02,$9C,$0A,$F0,$74,$43,$45,$74,$E0
	.dc.b	$8E,$09,$3F,$4E,$3D,$AD,$7C,$CD,$7C,$2D,$7B,$35,$EA,$D8,$EB,$3F
	.dc.b	$4E,$30,$47,$04,$9F,$0B,$5F,$0B,$5F,$0B,$50,$F4,$7F,$D7,$D7,$FE
	.dc.b	$00,$6C,$7F,$8D,$5A,$FF,$8E,$DA,$F2,$D7,$FC,$15,$3F,$D1,$EB,$F4
	.dc.b	$F3,$FD,$E2,$30,$7F,$8F,$E1,$EC,$3F,$0E,$FF,$62,$7C,$8E,$63,$12
	.dc.b	$8C,$4E,$AF,$FC,$99,$3C,$A3,$11,$E5,$F5,$5E,$CF,$EF,$73,$FF,$6D
	.dc.b	$4F,$B6,$72,$D9,$FE,$05,$F8,$FB,$EE,$BF,$68,$DC,$73,$18,$3F,$74
	.dc.b	$FA,$BF,$10,$F0,$DF,$FB,$77,$44,$62,$74,$7A,$E7,$D3,$3A,$8C,$47
	.dc.b	$A7,$0F,$46,$3E,$68,$FA,$4D,$74,$7F,$89,$BE,$E3,$F8,$57,$79,$A4
	.dc.b	$BF,$F1,$D9,$F4,$3F,$10,$F2,$62,$69,$23,$C7,$7F,$86,$D7,$E4,$BE
	.dc.b	$2C,$BC,$A4,$7F,$C8,$FE,$CF,$F8,$AE,$BF,$96,$36,$33,$E0,$FF,$1F
	.dc.b	$8C,$47,$C2,$C9,$4F,$96,$FC,$14,$61,$B1,$99,$E0,$D7,$F3,$6B,$F4
	.dc.b	$37,$D1,$F1,$9F,$C3,$C6,$27,$84,$78,$23,$15,$F9,$56,$C7,$0D,$BC
	.dc.b	$BF,$49,$96,$C6,$6D,$39,$6C,$78,$3F,$0D,$DB,$E6,$B4,$62,$3C,$4D
	.dc.b	$9F,$53,$65,$F3,$DF,$1B,$E6,$BC,$CD,$BF,$6B,$5E,$CD,$7F,$7F,$F8
	.dc.b	$C3,$E2,$B1,$08,$C6,$2B,$FF,$40,$C6,$1B,$5C,$E7,$E7,$7D,$15,$C4
	.dc.b	$D2,$47,$3C,$3E,$8F,$D2,$7E,$8E,$8D,$24,$FE,$42,$5C,$3F,$86,$7F
	.dc.b	$D9,$4D,$7F,$8C,$B5,$DF,$17,$ED,$2F,$E9,$D3,$F1,$BF,$B9,$BF,$E3
	.dc.b	$9B,$19,$BF,$C2,$CE,$5F,$C3,$FD,$4D,$7E,$C3,$5F,$C4,$36,$7F,$20
	.dc.b	$DB,$FF,$75,$77,$7E,$2F,$DE,$F8,$04,$AF,$77,$47,$EB,$30,$FD,$D9
	.dc.b	$FA,$D3,$F8,$A1,$FA,$88,$FD,$5E,$7F,$A9,$23,$FB,$CC,$1E,$27,$EA
	.dc.b	$BF,$0E,$FC,$D3,$6F,$F4,$2D,$7F,$24,$D7,$F4,$5F,$5B,$6F,$F5,$4D
	.dc.b	$B3,$4C,$7C,$A7,$DB,$24,$FF,$08,$3F,$64,$FF,$24,$AF,$AB,$1E,$A2
	.dc.b	$3E,$D3,$F6,$1E,$77,$E1,$49,$1C,$EF,$FD,$2F,$AF,$C4,$AF,$8C,$43
	.dc.b	$E4,$9F,$86,$CF,$E2,$BD,$4F,$EB,$E7,$E2,$1F,$C3,$7D,$87,$E4,$C9
	.dc.b	$7D,$DF,$47,$17,$E3,$49,$3D,$2B,$F8,$5F,$A7,$F9,$B2,$5F,$9D,$E3
	.dc.b	$3B,$9F,$24,$FC,$07,$D0,$7F,$2D,$F0,$BF,$9E,$37,$C0,$FE,$58,$FF
	.dc.b	$18,$FC,$2F,$F3,$84,$AF,$C0,$3F,$9F,$FC,$2F,$F9,$4F,$CF,$7F,$19
	.dc.b	$6D,$7F,$20,$6B,$FC,$45,$AF,$F2,$1F,$AA,$FD,$41,$2F,$F1,$8F,$F9
	.dc.b	$57,$97,$F2,$D3,$E5,$3E,$91,$F4,$0F,$E5,$58,$79,$5F,$E5,$4D,$7D
	.dc.b	$E7,$F8,$3F,$2F,$7B,$72,$F7,$BF,$CE,$BF,$4F,$7D,$C7,$F6,$E6,$E5
	.dc.b	$FD,$69,$B3,$FE,$0B,$6B,$FC,$FB,$E2,$FF,$A5,$F5,$7F,$3B,$6C,$66
	.dc.b	$93,$96,$C7,$F4,$6F,$BC,$FE,$14,$48,$FD,$73,$F0,$F2,$F4,$E9,$F1
	.dc.b	$2B,$D4,$7C,$43,$FA,$67,$C4,$FF,$13,$25,$FE,$DD,$FF,$04,$FE,$DF
	.dc.b	$FD,$63,$9B,$FC,$48,$DA,$FF,$0E,$B5,$FE,$BA,$D7,$E9,$1F,$32,$FC
	.dc.b	$47,$D6,$B4,$FF,$88,$9F,$EC,$B3,$CB,$6F,$F4,$C4,$76,$6C,$66,$36
	.dc.b	$FF,$74,$97,$DF,$FF,$8D,$A7,$CA,$D9,$F2,$B5,$E0,$D7,$E2,$9E,$7C
	.dc.b	$BE,$49,$FA,$09,$F3,$FB,$1E,$29,$F0,$36,$32,$79,$0B,$1F,$EA,$F7
	.dc.b	$DF,$C1,$BC,$3F,$0B,$1F,$2A,$30,$67,$D0,$F3,$47,$C6,$6C,$FF,$8F
	.dc.b	$F4,$E8,$D9,$F4,$67,$F9,$6C,$3E,$89,$F5,$30,$EB,$3E,$5A,$F0,$CE
	.dc.b	$63,$E5,$BF,$08,$FD,$41,$4E,$64,$30,$4D,$3F,$1D,$EB,$FF,$B5,$3E
	.dc.b	$55,$EC,$FF,$1C,$8C,$37,$C1,$FF,$38,$CE,$AF,$FC,$FF,$5A,$6C,$EA
	.dc.b	$B8,$7A,$DF,$E8,$4B,$E1,$5F,$EE,$46,$DF,$18,$FD,$C5,$7F,$9A,$48
	.dc.b	$FF,$91,$5A,$FF,$96,$DA,$FF,$1D,$7F,$3F,$9F,$F9,$01,$F9,$ED,$CB
	.dc.b	$EC,$1E,$E2,$C7,$FB,$AF,$27,$07,$F9,$FA,$3A,$DF,$90,$F1,$1F,$AA
	.dc.b	$DF,$FC,$9B,$F4,$1F,$71,$B8,$64,$46,$5A,$78,$09,$3F,$6E,$BC,$F1
	.dc.b	$E5,$8F,$DA,$F0,$FA,$5B,$19,$7F,$36,$D3,$F3,$73,$EB,$C3,$D5,$3E
	.dc.b	$83,$E4,$14,$FD,$5C,$F9,$1F,$D7,$B5,$CD,$FE,$63,$8B,$DB,$1F,$BB
	.dc.b	$BF,$CD,$F0,$F9,$63,$F7,$8D,$BF,$E0,$EE,$63,$E5,$1E,$22,$1F,$64
	.dc.b	$FE,$2A,$7C,$C2,$C7,$CA,$6D,$FF,$1C,$74,$7F,$98,$78,$FF,$00,$FF
	.dc.b	$9D,$1A,$E6,$7F,$22,$FF,$2A,$6B,$E4,$3C,$47,$90,$C3,$FE,$81,$E4
	.dc.b	$E7,$1F,$E7,$D6,$C6,$4E,$A5,$C3,$FA,$17,$C7,$4A,$FE,$8C,$D7,$FA
	.dc.b	$A6,$FC,$BA,$3F,$A9,$D7,$F5,$18,$C1,$E0,$9F,$31,$38,$3F,$2F,$A7
	.dc.b	$A9,$B5,$FD,$9B,$1F,$1F,$82,$EA,$F3,$B6,$33,$1D,$6B,$8E,$1F,$85
	.dc.b	$3B,$70,$7E,$29,$FF,$47,$75,$F2,$25,$F6,$7A,$F6,$7F,$AC,$3E,$5F
	.dc.b	$FC,$11,$26,$92,$BF,$C2,$51,$E2,$26,$7E,$93,$F6,$38,$F9,$27,$EA
	.dc.b	$35,$FF,$18,$4F,$DA,$F0,$7F,$D8,$D5,$FE,$5D,$24,$FF,$8B,$27,$F1
	.dc.b	$4F,$FA,$9F,$C1,$FC,$4B,$FE,$B7,$6C,$7F,$AF,$DC,$C7,$F9,$3A,$79
	.dc.b	$BF,$E3,$06,$BF,$EB,$FF,$0B,$8E,$C6,$92,$FC,$33,$E8,$AE,$2F,$53
	.dc.b	$FC,$96,$53,$DF,$AF,$D1,$7F,$2A,$49,$FF,$6A,$78,$7E,$02,$3D,$8C
	.dc.b	$9E,$08,$E8,$78,$4E,$F9,$7C,$86,$F8,$7F,$86,$F1,$3D,$64,$AE,$99
	.dc.b	$A3,$DF,$E2,$FF,$BB,$BE,$D3,$FD,$BC,$DC,$3F,$DC,$AD,$7F,$DA,$CD
	.dc.b	$7F,$DC,$DE,$0D,$C3,$FD,$DC,$DC,$3F,$D2,$30,$BF,$F4,$8B,$6B,$CE
	.dc.b	$FE,$07,$1F,$D3,$65,$EB,$9F,$AC,$7E,$6F,$83,$F1,$0F,$E7,$39,$7F
	.dc.b	$4B,$E4,$7D,$82,$61,$EE,$7E,$DF,$8A,$87,$E4,$7F,$EF,$B8,$C3,$63
	.dc.b	$37,$FE,$BE,$6B,$FE,$F5,$F2,$B6,$FF,$B1,$23,$2F,$5D,$7F,$D9,$B5
	.dc.b	$FE,$CA,$3F,$0B,$9F,$AE,$7E,$C1,$EC,$3F,$E0,$07,$99,$F9,$02,$5F
	.dc.b	$D6,$7F,$27,$E6,$FA,$7E,$37,$F1,$B3,$FE,$EE,$0E,$31,$FF,$77,$C6
	.dc.b	$2F,$FD,$3A,$EE,$7F,$DE,$EE,$63,$FD,$43,$E8,$FF,$BF,$A7,$24,$8E
	.dc.b	$34,$F4,$BC,$51,$86,$A3,$E1,$38,$6D,$F7,$7E,$37,$A7,$F4,$FE,$3B
	.dc.b	$D0,$D8,$EE,$D7,$A4,$AD,$39,$9F,$A1,$EA,$DA,$D3,$68,$D3,$1E,$BF
	.dc.b	$E0,$FE,$33,$E4,$36,$FD,$ED,$7C,$34,$8C,$47,$E2,$FD,$B7,$B5,$8C
	.dc.b	$51,$85,$87,$D8,$8C,$36,$79,$D7,$EF,$09,$C3,$F8,$77,$1C,$FF,$A0
	.dc.b	$27,$77,$F0,$B8,$FE,$BB,$83,$DB,$FE,$42,$DC,$BF,$85,$37,$CC,$F1
	.dc.b	$4F,$F2,$4D,$BC,$DE,$0F,$F1,$8C,$3E,$3E,$9F,$D5,$E7,$31,$FC,$71
	.dc.b	$AF,$F1,$CE,$BF,$EB,$19,$F8,$F4,$FE,$79,$A7,$C2,$F7,$3C,$64,$7C
	.dc.b	$6F,$F3,$4F,$75,$E8,$FF,$65,$6C,$FF,$7C,$6E,$9E,$56,$E5,$FC,$0F
	.dc.b	$1F,$01,$32,$FE,$F3,$BB,$DE,$F8,$64,$C7,$E1,$BF,$D1,$3D,$F7,$C1
	.dc.b	$FF,$05,$F8,$3F,$85,$18,$A4,$62,$75,$97,$F3,$EA,$D4,$FC,$9D,$FF
	.dc.b	$BD,$B7,$36,$7C,$AF,$FA,$2D,$69,$B3,$AA,$FA,$5F,$06,$F8,$3F,$F0
	.dc.b	$43,$73,$FF,$0F,$B7,$7D,$CE,$89,$3A,$6C,$6A,$30,$47,$04,$C3,$FA
	.dc.b	$CD,$75,$6D,$72,$6D,$71,$6D,$71,$7F,$81,$0A,$E4,$E1,$AF,$F0,$76
	.dc.b	$BB,$AE,$CE,$DA,$ED,$AE,$EB,$C4,$ED,$AE,$DA,$EE,$BC,$CE,$5A,$E4
	.dc.b	$F5,$90,$E8,$EA,$BC,$E4,$AF,$1B,$6B,$BB,$67,$84,$E8,$93,$A6,$C6
	.dc.b	$A3,$04,$70,$4A,$F2,$1E,$92,$9E,$93,$CE,$62,$BC,$4D,$AE,$ED,$AE
	.dc.b	$CD,$AE,$CF,$22,$1D,$CA,$77,$30,$41,$EC,$3C,$87,$B5,$FC,$36,$73
	.dc.b	$18,$7F,$88,$4E,$E7,$4D,$8F,$E2,$9F,$60,$F9,$CF,$EF,$51,$86,$BF
	.dc.b	$E3,$08,$D6,$3D,$DB,$1B,$8C,$4F,$08,$FC,$BC,$E9,$E5,$19,$7F,$AE
	.dc.b	$46,$5D,$C6,$54,$65,$DC,$66,$31,$1A,$8C,$12,$7C,$64,$8F,$31,$0F
	.dc.b	$1B,$96,$99,$3C,$6E,$5A,$64,$F1,$B9,$69,$93,$C8,$E5,$A6,$67,$D2
	.dc.b	$49,$F4,$46,$5A,$F9,$08,$78,$8A,$78,$0A,$76,$2C,$75,$9F,$13,$63
	.dc.b	$B4,$62,$7B,$C7,$52,$1D,$4A,$75,$2C,$F3,$7B,$C7,$27,$81,$0E,$05
	.dc.b	$8E,$0F,$28,$E0,$E8,$91,$A7,$84,$6A,$31,$1A,$8C,$12,$7D,$84,$8F
	.dc.b	$79,$0F,$E8,$BB,$FE,$9D,$FE,$5B,$BF,$FB,$23,$1F,$E9,$D3,$99,$FF
	.dc.b	$19,$4E,$5F,$F2,$0F,$BA,$FC,$0B,$FC,$C5,$EF,$FF,$92,$A3,$18,$A3
	.dc.b	$13,$AA,$E2,$D7,$F9,$EF,$D9,$3D,$7C,$47,$F9,$06,$31,$7C,$E3,$FD
	.dc.b	$47,$18,$BF,$F0,$73,$63,$FD,$4F,$D2,$F1,$7F,$58,$DB,$FE,$10,$D9
	.dc.b	$FF,$01,$4F,$ED,$65,$93,$EC,$BC,$71,$86,$C6,$63,$C9,$5F,$E4,$C8
	.dc.b	$C6,$3F,$E2,$1F,$B3,$7C,$7F,$B2,$91,$8B,$E3,$1F,$E5,$F8,$C5,$71
	.dc.b	$AE,$1F,$1D,$FE,$AD,$BF,$E4,$0D,$BF,$F1,$73,$5F,$F6,$3B,$5F,$27
	.dc.b	$CD,$7A,$09,$3E,$03,$DE,$F9,$7B,$3F,$C2,$D8,$7F,$85,$67,$CB,$49
	.dc.b	$CB,$FE,$17,$ED,$FA,$FF,$24,$F9,$2F,$F5,$AE,$1F,$ED,$88,$C7,$56
	.dc.b	$A7,$37,$FD,$F2,$73,$8F,$8F,$67,$E0,$BF,$EA,$DF,$A3,$FC,$63,$6F
	.dc.b	$F2,$2D,$7E,$83,$5F,$F5,$A4,$BF,$05,$C5,$FD,$C6,$C6,$61,$7F,$99
	.dc.b	$9C,$FC,$9F,$FA,$F7,$D7,$FC,$E3,$70,$FD,$0B,$5F,$C0,$35,$F0,$FB
	.dc.b	$57,$0F,$C5,$B7,$13,$4C,$7C,$B7,$FB,$2F,$9B,$F5,$4D,$BF,$D7,$B5
	.dc.b	$FC,$73,$5F,$F5,$37,$17,$F9,$36,$7E,$C1,$23,$EB,$61,$F9,$B7,$F7
	.dc.b	$3F,$2F,$C6,$3F,$3E,$DC,$4D,$31,$E2,$7E,$80,$93,$F5,$DF,$D1,$FD
	.dc.b	$84,$8F,$B4,$DE,$4F,$E0,$9D,$5F,$EE,$7E,$2B,$5F,$7B,$2F,$E1,$73
	.dc.b	$96,$C6,$5F,$D6,$7C,$4E,$5B,$19,$7E,$FC,$65,$B1,$93,$F3,$A3,$F0
	.dc.b	$EF,$E2,$DD,$3F,$EF,$96,$C6,$67,$F8,$EC,$E5,$FE,$4F,$8F,$F2,$42
	.dc.b	$38,$3F,$CA,$BF,$3A,$FC,$45,$F4,$1F,$C6,$A3,$0D,$8C,$E1,$FC,$02
	.dc.b	$72,$FE,$87,$94,$BE,$43,$F2,$2D,$C3,$F2,$2D,$7F,$22,$D7,$F2,$2D
	.dc.b	$3F,$22,$5E,$2F,$C8,$B7,$9B,$BD,$7F,$39,$E2,$F8,$17,$FC,$6D,$AF
	.dc.b	$EB,$1B,$5F,$C9,$9B,$6E,$2A,$49,$F9,$95,$FC,$96,$37,$DB,$EE,$FC
	.dc.b	$3B,$FA,$43,$67,$FA,$6B,$5F,$E5,$8D,$7F,$A9,$F9,$FF,$EA,$2E,$2F
	.dc.b	$EA,$E4,$8F,$EB,$AF,$FA,$21,$6D,$F3,$A3,$0D,$8F,$B3,$BF,$D4,$9C
	.dc.b	$B6,$3F,$AF,$E9,$F5,$3E,$1B,$E3,$68,$69,$27,$9F,$0F,$9F,$1F,$BB
	.dc.b	$A7,$48,$C6,$3F,$82,$8F,$ED,$11,$F8,$0F,$C3,$AE,$26,$93,$0E,$B0
	.dc.b	$FC,$3B,$A3,$63,$31,$83,$FB,$DB,$0F,$F0,$09,$4F,$EB,$1F,$4B,$FD
	.dc.b	$70,$93,$FC,$CD,$FE,$C3,$F8,$5F,$E5,$21,$F9,$CF,$CD,$51,$98,$C3
	.dc.b	$6B,$FD,$73,$3A,$3F,$3B,$3A,$F1,$FF,$B8,$57,$F6,$3A,$C5,$7D,$97
	.dc.b	$F4,$71,$97,$D2,$43,$FC,$33,$C3,$FE,$E4,$7D,$87,$0F,$98,$EA,$4A
	.dc.b	$E8,$7A,$1F,$F3,$87,$9C,$AE,$55,$D4,$FC,$7F,$9F,$FE,$8F,$9D,$3C
	.dc.b	$B0,$E1,$87,$F6,$B6,$F5,$FC,$2C,$3E,$49,$F2,$BE,$AE,$56,$DB,$3F
	.dc.b	$23,$D2,$F9,$E4,$9E,$8F,$F8,$27,$1F,$0E,$5E,$E0,$E1,$F4,$4F,$D4
	.dc.b	$FA,$CF,$B2,$47,$CC,$FD,$CF,$27,$F8,$7F,$E1,$BF,$61,$B5,$F6,$34
	.dc.b	$FE,$CB,$6A,$D3,$67,$E3,$1E,$2D,$3E,$4B,$F5,$35,$F9,$11,$88,$FB
	.dc.b	$7F,$4B,$F8,$C2,$57,$C9,$7F,$1F,$EE,$7E,$02,$3C,$4D,$91,$F8,$47
	.dc.b	$0D,$8F,$C7,$3E,$13,$E3,$69,$F1,$5C,$1E,$DF,$67,$F9,$7F,$9B,$FE
	.dc.b	$C0,$AF,$CD,$C6,$1D,$1F,$E0,$79,$F0,$BA,$30,$FE,$5B,$E8,$7F,$D9
	.dc.b	$71,$F2,$5B,$1C,$35,$46,$23,$84,$F4,$75,$F5,$FF,$E6,$C8,$FA,$4D
	.dc.b	$8C,$C7,$F9,$B4,$C9,$5F,$69,$0F,$CC,$9F,$A4,$7F,$6F,$3E,$16,$9E
	.dc.b	$18,$F1,$9F,$0E,$BC,$F1,$98,$FB,$DA,$FD,$C2,$3F,$3B,$5F,$BC,$E8
	.dc.b	$C4,$7F,$BC,$BD,$2F,$F7,$B7,$2B,$3F,$B7,$5F,$EF,$AF,$67,$EE,$CE
	.dc.b	$63,$EF,$FB,$7F,$EF,$98,$C4,$6C,$38,$8D,$57,$46,$CF,$2B,$F0,$CE
	.dc.b	$5B,$1F,$BE,$8F,$F5,$09,$1F,$F5,$3B,$5F,$CE,$35,$E0,$7D,$FC,$BD
	.dc.b	$85,$8F,$95,$A7,$F7,$C6,$C6,$A7,$99,$D8,$86,$0F,$BF,$F0,$9F,$4F
	.dc.b	$6F,$F0,$2B,$64,$D3,$AD,$48,$F1,$12,$BF,$25,$1F,$E0,$EF,$83,$FB
	.dc.b	$CD,$9F,$CF,$C6,$13,$FE,$14,$9F,$21,$F0,$0A,$7E,$12,$7F,$A0,$F1
	.dc.b	$7F,$86,$BF,$0D,$FC,$53,$6F,$F7,$62,$32,$D3,$F8,$87,$27,$BA,$EC
	.dc.b	$7B,$9B,$8F,$F4,$E9,$D4,$78,$B9,$3F,$9A,$E3,$F5,$63,$C8,$7F,$1B
	.dc.b	$E6,$FF,$16,$FE,$1B,$FC,$68,$91,$9C,$51,$FD,$6B,$A3,$AC,$FF,$52
	.dc.b	$6C,$FF,$2D,$3C,$91,$FB,$D8,$DB,$E0,$3C,$9C,$AE,$5F,$CC,$C7,$C9
	.dc.b	$3F,$AE,$F4,$F4,$E7,$34,$98,$76,$9F,$EC,$7D,$7B,$24,$64,$FE,$E6
	.dc.b	$F3,$C2,$46,$6F,$FA,$BE,$FF,$D5,$63,$A5,$F9,$CF,$21,$4C,$3F,$D9
	.dc.b	$CF,$49,$75,$FE,$FF,$CF,$F9,$D9,$FE,$90,$D7,$E0,$E7,$F0,$1E,$9B
	.dc.b	$FD,$67,$FA,$77,$66,$DF,$EF,$ED,$C3,$FC,$1A,$DC,$BF,$DD,$18,$FF
	.dc.b	$83,$A7,$FC,$CF,$F9,$AD,$4D,$25,$FF,$84,$BC,$B7,$81,$F6,$9B,$5F
	.dc.b	$25,$BA,$FF,$87,$5B,$5F,$B4,$3E,$2E,$7E,$97,$F4,$7E,$37,$F9,$EB
	.dc.b	$F0,$CF,$F1,$99,$23,$FC,$6E,$FF,$8F,$3F,$0C,$91,$E3,$F1,$BF,$CF
	.dc.b	$F1,$88,$FE,$BB,$5F,$E8,$4F,$C3,$3F,$C7,$A4,$8F,$F2,$0B,$FE,$52
	.dc.b	$D2,$47,$F9,$5B,$4F,$6B,$63,$FC,$8B,$F8,$5C,$8F,$F2,$4B,$5F,$EE
	.dc.b	$CF,$F8,$BF,$3F,$F0,$EB,$79,$3F,$30,$DE,$33,$FE,$78,$C1,$8D,$26
	.dc.b	$1F,$E7,$A8,$CB,$F4,$7C,$FF,$F4,$ED,$7C,$D6,$C6,$67,$AB,$67,$FC
	.dc.b	$DE,$47,$DC,$D7,$FB,$9E,$FF,$E9,$CE,$FF,$B1,$E8,$FF,$AA,$9B,$66
	.dc.b	$9F,$EA,$C2,$57,$FA,$B1,$AF,$D2,$F3,$3F,$D6,$FC,$7F,$E6,$E2,$47
	.dc.b	$F9,$C9,$F8,$84,$7F,$D2,$1B,$7F,$91,$A3,$11,$E6,$F4,$FD,$ED,$C3
	.dc.b	$DE,$DC,$FF,$DA,$67,$E4,$7D,$4F,$F6,$1F,$C5,$DA,$34,$95,$FE,$E8
	.dc.b	$7F,$DD,$A4,$70,$7C,$DC,$3E,$84,$FE,$7F,$6F,$D1,$1F,$A3,$F4,$64
	.dc.b	$7E,$8D,$B3,$FA,$3F,$63,$FD,$AB,$F1,$16,$8D,$26,$1F,$EB,$1F,$6D
	.dc.b	$EB,$70,$8C,$35,$FF,$00,$C6,$AB,$9C,$69,$AF,$DE,$F7,$38,$3F,$EE
	.dc.b	$6E,$44,$62,$51,$87,$8C,$EE,$75,$3F,$EE,$2F,$6B,$F0,$A4,$9F,$E0
	.dc.b	$DC,$BF,$EF,$28,$FF,$78,$92,$3F,$DE,$2D,$7F,$23,$E8,$7F,$BE,$3E
	.dc.b	$9B,$F8,$43,$63,$F8,$83,$5D,$E2,$9F,$E1,$DE,$E6,$67,$D8,$D9,$F5
	.dc.b	$B7,$13,$BF,$7D,$F0,$0F,$4B,$FA,$BC,$61,$B1,$9B,$FE,$95,$39,$7F
	.dc.b	$A0,$ED,$F4,$78,$BF,$D6,$0D,$CB,$E9,$46,$38,$3F,$94,$74,$4B,$FE
	.dc.b	$50,$D9,$FF,$5C,$7C,$0B,$E6,$FF,$9E,$60,$8C,$5F,$F4,$36,$F8,$FC
	.dc.b	$2B,$7F,$06,$F2,$BF,$B5,$B6,$FF,$AB,$B7,$1F,$F6,$33,$71,$FE,$7D
	.dc.b	$F0,$5F,$CF,$49,$5F,$CF,$5F,$EE,$BF,$08,$E6,$F2,$B6,$7F,$AB,$37
	.dc.b	$3F,$F6,$93,$6F,$FC,$AE,$3E,$15,$F0,$DF,$E0,$06,$DF,$F6,$C6,$E7
	.dc.b	$FE,$DB,$6D,$7F,$53,$F8,$7F,$EA,$44,$AF,$EA,$4F,$F8,$2B,$C3,$FF
	.dc.b	$67,$FC,$57,$F3,$12,$57,$F0,$EF,$27,$FC,$19,$1F,$EE,$7F,$89,$FF
	.dc.b	$04,$E1,$FD,$BD,$BA,$E7,$D4,$FA,$71,$8C,$3D,$6D,$7F,$CE,$6D,$7F
	.dc.b	$D5,$ED,$4D,$24,$74,$25,$78,$5B,$5E,$18,$C1,$1C,$12,$BC,$73,$C1
	.dc.b	$B3,$B6,$CE,$DF,$3C,$ED,$B3,$B6,$CE,$DF,$5C,$ED,$B3,$B6,$CE,$DF
	.dc.b	$7C,$ED,$B3,$B6,$CE,$EB,$E1,$CE,$5B,$19,$AE,$51,$8A,$F5,$90,$F3
	.dc.b	$94,$F1,$94,$EE,$57,$93,$C0,$87,$02,$9C,$0A,$F0,$74,$43,$45,$74
	.dc.b	$E0,$8E,$09,$1E,$C2,$57,$C1,$6D,$7C,$18,$C1,$1C,$12,$BD,$93,$FC
	.dc.b	$23,$DD,$FE,$39,$EF,$7F,$B2,$49,$3F,$B2,$3F,$D8,$47,$E6,$BE,$E7
	.dc.b	$FD,$2C,$DC,$34,$FE,$62,$32,$FE,$79,$AF,$E7,$9A,$EE,$38,$E2,$9C
	.dc.b	$B6,$33,$18,$23,$82,$4F,$ED,$63,$2D,$8C,$BE,$73,$04,$30,$43,$D0
	.dc.b	$F0,$69,$B2,$9B,$3D,$8E,$DA,$6C,$A6,$CF,$80,$64,$A6,$67,$E0,$39
	.dc.b	$6B,$99,$F6,$B6,$7C,$0F,$23,$CA,$43,$C2,$53,$46,$08,$60,$93,$E0
	.dc.b	$78,$35,$DB,$5D,$CF,$91,$DB,$5D,$B5,$DC,$FA,$27,$91,$27,$4D,$8D
	.dc.b	$46,$08,$E0,$93,$F1,$E3,$2D,$8C,$BF,$1C,$C1,$0C,$10,$13,$46,$9F
	.dc.b	$EC,$DC,$3F,$CE,$C4,$9F,$E8,$DE,$72,$3F,$BD,$5D,$9F,$F1,$74,$60
	.dc.b	$8F,$F5,$EF,$AD,$7C,$F2,$B8,$4F,$88,$4E,$A3,$31,$E6,$8C,$57,$ED
	.dc.b	$DB,$1F,$DA,$AD,$F1,$7D,$6B,$F7,$91,$8B,$E7,$5F,$EE,$9A,$D3,$C5
	.dc.b	$AF,$F9,$2B,$B7,$CD,$F7,$FE,$E2,$47,$F9,$2D,$FE,$11,$F2,$DF,$D9
	.dc.b	$9B,$C3,$C2,$7F,$BE,$CE,$5B,$19,$7E,$8E,$0F,$69,$F7,$5F,$C2,$A3
	.dc.b	$0D,$8C,$C7,$F1,$73,$27,$F1,$4F,$0B,$FC,$C1,$5F,$50,$87,$F9,$84
	.dc.b	$FF,$07,$C7,$7E,$2E,$84,$7A,$57,$E8,$5F,$27,$DE,$7D,$7F,$C2,$B8
	.dc.b	$CA,$31,$7F,$66,$B6,$2B,$73,$A9,$F7,$D3,$EF,$71,$1F,$83,$8C,$5F
	.dc.b	$18,$E9,$18,$BF,$0D,$70,$AF,$91,$E7,$FE,$2F,$F0,$1F,$2B,$70,$E2
	.dc.b	$D7,$C2,$D7,$C1,$F2,$D6,$BE,$93,$71,$34,$C7,$AB,$F9,$9F,$C6,$7E
	.dc.b	$1B,$6F,$CE,$D7,$DE,$D7,$F5,$FF,$3E,$F8,$CC,$42,$31,$8A,$FF,$AA
	.dc.b	$46,$1B,$5C,$E7,$F1,$1F,$19,$72,$34,$91,$CF,$0F,$AD,$EA,$3E,$73
	.dc.b	$F1,$CD,$9F,$C6,$36,$7F,$0E,$D7,$E9,$F1,$7E,$1D,$FC,$1B,$5E,$BF
	.dc.b	$35,$FB,$DF,$CF,$78,$B6,$33,$69,$CB,$63,$CB,$D1,$B7,$F4,$B8,$23
	.dc.b	$11,$FC,$26,$31,$3C,$63,$73,$A8,$FD,$2C,$7E,$63,$9F,$F5,$44,$8F
	.dc.b	$EF,$7E,$0F,$F2,$0E,$DF,$E0,$AD,$EE,$FE,$41,$7F,$C5,$63,$0F,$F1
	.dc.b	$7F,$83,$7E,$29,$FC,$A3,$0F,$E2,$F1,$8F,$0F,$F8,$5D,$23,$13,$AA
	.dc.b	$E1,$87,$E6,$21,$F4,$54,$97,$F9,$A8,$CB,$FA,$4E,$8F,$E5,$53,$FB
	.dc.b	$46,$C6,$67,$6F,$EC,$23,$2F,$CC,$FA,$1F,$2C,$B7,$46,$1B,$1F,$CC
	.dc.b	$8D,$46,$6E,$46,$67,$D2,$D8,$E5,$F4,$72,$34,$DC,$7F,$C6,$1A,$FD
	.dc.b	$1F,$0A,$EC,$DD,$18,$6E,$9F,$D4,$A3,$FD,$53,$BF,$CC,$6E,$FF,$4D
	.dc.b	$B7,$FC,$BE,$BE,$94,$E5,$FD,$1E,$D9,$D3,$DA,$76,$8F,$8C,$D7,$F2
	.dc.b	$11,$CF,$0F,$B6,$76,$9F,$BD,$3F,$85,$AF,$C2,$4E,$4F,$D5,$FC,$EF
	.dc.b	$F6,$1F,$C1,$2D,$9A,$4B,$E5,$8F,$F6,$7F,$67,$B1,$23,$34,$8F,$8D
	.dc.b	$10,$FE,$EE,$53,$F5,$7F,$4D,$FC,$00,$93,$FC,$01,$FE,$07,$F4,$FF
	.dc.b	$81,$24,$7D,$D7,$F0,$9E,$EF,$F0,$82,$47,$F0,$87,$FC,$BF,$C1,$FA
	.dc.b	$BB,$F0,$1F,$E0,$DC,$7E,$E3,$8A,$FC,$87,$CF,$7F,$12,$23,$FC,$49
	.dc.b	$E7,$3C,$27,$D6,$43,$F6,$8E,$5F,$E0,$FE,$DF,$F1,$C2,$57,$F1,$C7
	.dc.b	$F8,$FF,$9D,$FE,$11,$FD,$3B,$F8,$EB,$6B,$F9,$2B,$5F,$E2,$2D,$7F
	.dc.b	$88,$7D,$57,$F8,$67,$9F,$F9,$2B,$63,$F9,$6B,$5F,$E3,$2D,$7E,$17
	.dc.b	$37,$F1,$0C,$BD,$11,$83,$E2,$74,$BE,$AF,$F9,$A9,$34,$FE,$6A,$FF
	.dc.b	$38,$FA,$EF,$F1,$57,$37,$F3,$66,$DF,$F3,$E6,$BF,$CA,$9A,$F7,$3E
	.dc.b	$51,$4F,$97,$8F,$F4,$2F,$AE,$FF,$1B,$47,$FA,$8B,$0F,$E0,$D1,$87
	.dc.b	$FA,$5B,$63,$F9,$BB,$5F,$E6,$DE,$DF,$F8,$FF,$F0,$4E,$59,$A3,$17
	.dc.b	$FD,$51,$B5,$FD,$01,$AF,$94,$83,$8F,$F1,$CF,$AA,$7F,$8D,$7C,$3F
	.dc.b	$F9,$35,$EE,$D7,$FB,$0B,$63,$FA,$6B,$5F,$D9,$79,$DF,$CE,$DF,$F6
	.dc.b	$0F,$52,$31,$5F,$E5,$78,$C7,$07,$BD,$B1,$FE,$3C,$AF,$EB,$3F,$65
	.dc.b	$FE,$5F,$FA,$1F,$82,$D9,$FE,$E6,$D9,$DC,$27,$FD,$9B,$97,$BA,$7E
	.dc.b	$A9,$FA,$C3,$E1,$61,$F8,$AC,$3C,$31,$E8,$C7,$CE,$FF,$9F,$A3,$38
	.dc.b	$7A,$67,$33,$F3,$2F,$B3,$6F,$F2,$AF,$6D,$FF,$30,$47,$C5,$ED,$7F
	.dc.b	$BB,$17,$CC,$BD,$CF,$F3,$E4,$E6,$31,$3F,$E9,$08,$C4,$EB,$B3,$FD
	.dc.b	$15,$5A,$A5,$69,$E1,$3A,$AE,$1E,$27,$FA,$6A,$F8,$5E,$EF,$D1,$5F
	.dc.b	$D2,$48,$FF,$A6,$9A,$FE,$81,$AF,$ED,$9F,$F4,$A3,$F8,$4C,$7F,$D3
	.dc.b	$11,$F3,$5B,$A7,$CB,$6C,$FC,$FD,$3F,$C0,$95,$F6,$A7,$FA,$E3,$DA
	.dc.b	$7C,$A7,$8C,$FC,$1D,$3E,$03,$F0,$ED,$32,$E0,$F1,$3F,$3C,$F1,$16
	.dc.b	$3C,$55,$E4,$7D,$51,$97,$FC,$55,$C3,$FE,$B3,$ED,$FD,$3C,$9A,$47
	.dc.b	$CD,$3F,$1C,$A7,$C7,$3F,$74,$BF,$04,$FD,$FF,$27,$FA,$FD,$B1,$93
	.dc.b	$6F,$F0,$4E,$7F,$34,$7F,$0E,$3C,$C5,$CF,$4F,$A6,$FE,$49,$E3,$21
	.dc.b	$E3,$3F,$90,$2F,$C0,$BF,$93,$36,$3F,$93,$35,$DC,$7F,$98,$3A,$FE
	.dc.b	$FC,$7F,$9A,$1A,$FF,$95,$1A,$FF,$94,$DA,$F9,$0F,$11,$0C,$4F,$F8
	.dc.b	$BF,$B3,$FD,$C5,$F2,$3F,$D2,$A0,$D2,$4F,$F5,$48,$7E,$9D,$FD,$61
	.dc.b	$B1,$98,$C1,$FE,$78,$D7,$CB,$3F,$E3,$D6,$F5,$FD,$D8,$FD,$75,$7F
	.dc.b	$9A,$E7,$D3,$D1,$FE,$63,$D3,$F2,$CD,$9F,$BF,$5F,$DE,$0E,$25,$9F
	.dc.b	$F4,$3F,$85,$D8,$E7,$C5,$FA,$0A,$FC,$39,$F1,$4F,$87,$8B,$A3,$CA
	.dc.b	$4C,$3A,$C6,$5B,$E1,$97,$C1,$7F,$A5,$A3,$0D,$8C,$DF,$FA,$82,$72
	.dc.b	$FF,$A8,$BC,$5F,$86,$49,$EE,$FA,$3B,$3E,$2C,$F9,$F5,$F8,$4F,$E3
	.dc.b	$BC,$2F,$F3,$DF,$E0,$1F,$E4,$12,$47,$F8,$D1,$FF,$1E,$FE,$09,$23
	.dc.b	$FD,$63,$E4,$FA,$63,$10,$3F,$CA,$BB,$FB,$79,$FF,$C9,$84,$8F,$F1
	.dc.b	$FB,$FE,$B4,$F2,$FE,$17,$E0,$97,$63,$49,$5F,$E4,$67,$FC,$AC,$47
	.dc.b	$07,$F1,$ED,$3F,$D4,$51,$FE,$6F,$24,$67,$0F,$F3,$A4,$7F,$B1,$3C
	.dc.b	$8F,$97,$3F,$EB,$66,$BF,$EB,$56,$DF,$FA,$35,$BC,$5F,$E8,$6E,$2F
	.dc.b	$F6,$6F,$9D,$82,$57,$D8,$BF,$F3,$8C,$FB,$48,$FF,$9E,$23,$4F,$FB
	.dc.b	$E9,$7C,$AF,$FA,$8D,$B1,$FE,$A3,$8C,$1F,$EA,$F2,$3F,$EA,$D6,$BF
	.dc.b	$4F,$87,$FD,$6F,$5F,$E7,$D2,$47,$F9,$C1,$FF,$4A,$11,$F8,$DC,$5E
	.dc.b	$47,$FB,$44,$7F,$B4,$6F,$F4,$FE,$C3,$D2,$FD,$AB,$63,$31,$FB,$6F
	.dc.b	$4F,$F3,$DF,$85,$7F,$BC,$9B,$87,$FB,$9A,$93,$96,$C7,$FB,$BA,$3F
	.dc.b	$DD,$A4,$70,$7F,$BE,$A3,$C7,$EB,$F8,$88,$FE,$C3,$D8,$FE,$0E,$47
	.dc.b	$07,$AB,$D9,$FD,$97,$E4,$3F,$D9,$A1,$9A,$73,$87,$E7,$4F,$DC,$46
	.dc.b	$65,$5C,$63,$0F,$F7,$09,$7B,$AF,$73,$9C,$73,$8C,$5F,$28,$FE,$EF
	.dc.b	$18,$AE,$4D,$8F,$14,$F9,$BD,$EE,$71,$E4,$6C,$64,$E9,$1F,$B5,$71
	.dc.b	$5F,$8D,$FC,$1A,$4F,$FB,$BD,$AE,$F1,$7C,$07,$F1,$79,$FF,$7F,$12
	.dc.b	$7F,$DF,$8F,$C1,$F8,$1F,$E3,$9F,$9A,$5C,$4D,$25,$7A,$B0,$7C,$27
	.dc.b	$08,$C3,$5F,$13,$5F,$A2,$DA,$FA,$FC,$1F,$80,$F1,$FE,$93,$6B,$FC
	.dc.b	$43,$F0,$AF,$87,$FE,$BA,$DB,$FE,$4A,$DB,$FB,$0D,$AF,$37,$37,$E7
	.dc.b	$89,$5F,$9E,$3F,$96,$F3,$3E,$1F,$F8,$DD,$7F,$8C,$23,$13,$FC,$72
	.dc.b	$31,$3A,$7F,$19,$3A,$EF,$F0,$E5,$99,$FE,$59,$8F,$9E,$73,$87,$39
	.dc.b	$FC,$A4,$E5,$FF,$1F,$F5,$3E,$23,$F8,$DC,$23,$15,$FB,$D6,$F5,$FE
	.dc.b	$5D,$B3,$FE,$48,$F8,$97,$C0,$72,$6D,$FF,$05,$6D,$FE,$7D,$B5,$F4
	.dc.b	$33,$F0,$E5,$F3,$EB,$D2,$49,$FE,$1E,$FF,$11,$7F,$D7,$2D,$7F,$A6
	.dc.b	$7C,$53,$87,$FC,$2A,$DC,$7F,$AE,$B7,$87,$F6,$0D,$AF,$F3,$0B,$5F
	.dc.b	$DD,$35,$EB,$5D,$09,$5D,$1B,$5D,$1B,$3D,$1A,$F4,$6B,$D2,$79,$12
	.dc.b	$79,$36,$39,$35,$E4,$D7,$94,$70,$23,$C1,$AF,$06,$BC,$0D,$10,$D1
	.dc.b	$4D,$18,$21,$82,$0A,$F9,$0D,$AF,$7B,$6B,$D0,$DA,$F1,$B4,$F0,$14
	.dc.b	$EE,$53,$BD,$76,$25,$76,$6D,$76,$6D,$76,$6D,$76,$6B,$D9,$AF,$6A
	.dc.b	$E8,$4A,$E8,$DA,$E8,$D9,$E8,$D7,$A3,$5E,$93,$C8,$93,$C9,$B1,$C9
	.dc.b	$FF,$1A,$FB,$5F,$ED,$18,$FF,$5B,$B5,$FF,$62,$9F,$D2,$DC,$3F,$E2
	.dc.b	$FF,$8C,$78,$BF,$E4,$B8,$FE,$A1,$18,$AF,$EA,$50,$8C,$3A,$9E,$0F
	.dc.b	$FB,$1A,$32,$D8,$CB,$FD,$86,$32,$D8,$CB,$E3,$6B,$DD,$B1,$C2,$35
	.dc.b	$3E,$48,$C4,$6E,$31,$3C,$08,$E0,$91,$E9,$6B,$E7,$6B,$E5,$6B,$E0
	.dc.b	$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$71,$D4
	.dc.b	$A7,$22,$9C,$4A,$71,$9F,$8E,$49,$F8,$B1,$CE,$31,$3F,$02,$38,$92
	.dc.b	$32,$EE,$78,$BE,$18,$CB,$63,$2F,$8E,$32,$D8,$CB,$E5,$6B,$DD,$B1
	.dc.b	$C2,$35,$3E,$58,$C4,$6E,$31,$3C,$08,$E0,$91,$E9,$6B,$E7,$6B,$E5
	.dc.b	$7F,$43,$E3,$BC,$7F,$EC,$5E,$EF,$F8,$D6,$31,$C4,$8C,$4E,$AF,$FC
	.dc.b	$C7,$1F,$DC,$08,$FF,$5D,$BE,$0D,$8F,$F2,$07,$C8,$3E,$47,$FD,$7D
	.dc.b	$18,$6C,$FF,$92,$C4,$EE,$B9,$C7,$F9,$3E,$31,$5E,$4C,$1F,$11,$49
	.dc.b	$5F,$E7,$87,$FB,$97,$2F,$FB,$35,$B8,$70,$9F,$EF,$93,$96,$C6,$5F
	.dc.b	$F2,$37,$96,$FB,$DF,$F6,$B7,$C1,$FF,$99,$63,$14,$8C,$4E,$AB,$FC
	.dc.b	$AB,$5A,$9F,$44,$43,$FD,$1A,$53,$FC,$D9,$F6,$BF,$E3,$37,$FC,$D0
	.dc.b	$D7,$E3,$35,$DF,$47,$F1,$0C,$1F,$61,$FE,$5D,$FA,$BF,$F3,$D3,$6B
	.dc.b	$FC,$CE,$D7,$FC,$C4,$D7,$FD,$E4,$FF,$89,$65,$3C,$B5,$FF,$7B,$36
	.dc.b	$33,$5D,$27,$2F,$6F,$B3,$FF,$7E,$7E,$01,$E2,$6C,$66,$93,$96,$C7
	.dc.b	$9B,$EF,$F6,$F9,$9A,$F9,$9B,$1F,$01,$B8,$BE,$F9,$2F,$A4,$F8,$71
	.dc.b	$86,$C6,$6F,$E2,$CE,$5F,$CE,$4F,$F9,$38,$8E,$0F,$97,$F7,$1B,$7E
	.dc.b	$5B,$5F,$96,$DA,$FA,$AD,$B7,$CC,$7E,$1E,$BE,$9B,$5F,$A6,$DA,$FB
	.dc.b	$AE,$61,$BF,$E0,$D1,$8D,$FF,$A8,$C6,$63,$15,$FE,$B3,$3F,$0F,$F3
	.dc.b	$7F,$1B,$F0,$0F,$C3,$B6,$7F,$0E,$D7,$F0,$ED,$7F,$0F,$F3,$D6,$BF
	.dc.b	$22,$DC,$8D,$31,$F4,$1E,$B8,$34,$93,$FA,$BA,$FD,$29,$0F,$F6,$87
	.dc.b	$6F,$F1,$67,$F5,$07,$C0,$CB,$DC,$4C,$FE,$09,$FA,$5C,$1C,$DF,$84
	.dc.b	$FC,$7F,$F4,$AD,$BF,$D2,$B5,$FD,$2B,$5F,$D2,$F9,$17,$2F,$D7,$B6
	.dc.b	$FF,$D1,$77,$FB,$08,$C7,$D1,$FB,$C9,$5E,$DB,$FD,$8C,$7D,$BC,$BD
	.dc.b	$13,$FE,$16,$6C,$FE,$EB,$E9,$3E,$41,$27,$94,$7E,$E2,$3F,$09,$A7
	.dc.b	$C5,$8F,$90,$7E,$11,$C3,$F2,$BC,$1F,$F5,$13,$FC,$3B,$5F,$8F,$7E
	.dc.b	$16,$C7,$F0,$AF,$A5,$FB,$04,$9F,$09,$FC,$41,$F5,$4F,$09,$FE,$2E
	.dc.b	$D7,$F8,$7F,$BD,$F8,$12,$57,$2D,$91,$CA,$3C,$44,$9F,$33,$FD,$6B
	.dc.b	$0F,$4C,$7B,$86,$3E,$E7,$E0,$46,$9F,$E3,$FF,$17,$F8,$32,$57,$E8
	.dc.b	$F0,$47,$E1,$FE,$8B,$FA,$AD,$7F,$26,$6B,$FC,$99,$B5,$FC,$C1,$B4
	.dc.b	$FA,$6F,$C8,$12,$7B,$3F,$E4,$4C,$7C,$D7,$E8,$C7,$CC,$FA,$3B,$BD
	.dc.b	$C7,$97,$2F,$3C,$7F,$3C,$6B,$F0,$6B,$99,$2B,$9E,$1F,$D1,$34,$FD
	.dc.b	$11,$1F,$80,$FE,$97,$EB,$3F,$93,$FE,$31,$FD,$21,$B8,$7F,$48,$6B
	.dc.b	$FD,$21,$AF,$F4,$8F,$AC,$FE,$5F,$E5,$7F,$50,$6C,$7F,$50,$6B,$FD
	.dc.b	$41,$AF,$BC,$FD,$D9,$F0,$F4,$F8,$2F,$F8,$17,$17,$DC,$7F,$3C,$7F
	.dc.b	$8D,$36,$34,$D7,$F9,$19,$FD,$09,$C3,$FC,$9A,$39,$F9,$7F,$CB,$67
	.dc.b	$FC,$21,$E3,$72,$8C,$1F,$AC,$D7,$EA,$DF,$94,$F3,$ED,$F8,$FC,$3C
	.dc.b	$BF,$35,$FE,$91,$7C,$26,$0F,$04,$F8,$71,$4E,$E3,$0D,$7D,$17,$FE
	.dc.b	$10,$AE,$75,$FE,$39,$7F,$C2,$06,$23,$53,$D5,$AF,$CD,$3F,$0C,$FA
	.dc.b	$AF,$D1,$4F,$B1,$F0,$5F,$CE,$6C,$66,$3D,$37,$F2,$E7,$CD,$F3,$8F
	.dc.b	$9C,$7B,$58,$6D,$F0,$9D,$4E,$4E,$A7,$9A,$3E,$EB,$70,$E3,$AF,$FA
	.dc.b	$D2,$72,$7C,$0C,$7F,$8D,$6D,$F8,$67,$DA,$FF,$92,$D6,$1E,$09,$F1
	.dc.b	$63,$E4,$9F,$F5,$FC,$E5,$B1,$99,$E7,$1F,$7B,$4F,$90,$4B,$F0,$47
	.dc.b	$88,$FF,$6A,$7C,$97,$F9,$8B,$67,$A9,$FE,$D8,$6C,$FF,$76,$F5,$15
	.dc.b	$BA,$D1,$27,$51,$96,$FA,$5F,$DF,$AF,$9C,$2F,$9C,$E6,$FE,$29,$FE
	.dc.b	$67,$24,$FF,$9B,$F8,$7F,$CF,$F1,$88,$D9,$9C,$99,$76,$8F,$E2,$93
	.dc.b	$FE,$77,$D7,$E6,$3F,$23,$83,$FC,$5E,$DD,$7E,$09,$FB,$29,$DB,$D6
	.dc.b	$FA,$5F,$EF,$EB,$C4,$F6,$3F,$3F,$7F,$E3,$AF,$9C,$FC,$17,$0F,$E3
	.dc.b	$1B,$19,$C3,$62,$3C,$57,$23,$C5,$C1,$F9,$48,$C6,$1F,$B3,$E2,$F4
	.dc.b	$D7,$EB,$9B,$19,$8F,$5B,$67,$FC,$9A,$FC,$82,$1E,$92,$C7,$F1,$8D
	.dc.b	$3C,$D1,$FB,$58,$F3,$1F,$B9,$2B,$9C,$3D,$FC,$1E,$28,$FE,$34,$D9
	.dc.b	$FE,$19,$7A,$7C,$B3,$E1,$21,$F3,$4F,$A1,$1F,$C7,$63,$E3,$47,$F0
	.dc.b	$CE,$3F,$9C,$D8,$CB,$FC,$5E,$32,$FF,$19,$8F,$BE,$FF,$14,$6B,$97
	.dc.b	$F1,$07,$F3,$83,$2F,$93,$7F,$C8,$BE,$42,$1F,$D1,$63,$FB,$9B,$FD
	.dc.b	$25,$AE,$7A,$7F,$A7,$46,$23,$FD,$33,$CF,$FE,$80,$FC,$5B,$FA,$D4
	.dc.b	$9A,$49,$FF,$51,$F5,$79,$27,$FB,$31,$E8,$EC,$46,$5B,$1F,$98,$C7
	.dc.b	$D3,$5F,$80,$3E,$71,$F8,$09,$FF,$05,$CF,$3C,$79,$11,$E5,$B7,$EB
	.dc.b	$B4,$FD,$0D,$F2,$3E,$D4,$FF,$AF,$3B,$3C,$44,$7C,$CF,$97,$B7,$FD
	.dc.b	$3F,$EC,$FF,$B0,$1B,$5F,$EC,$06,$BF,$E1,$BA,$4F,$F8,$2D,$F3,$12
	.dc.b	$F0,$79,$33,$F7,$D7,$D9,$C7,$D0,$FF,$8F,$AB,$F7,$7E,$17,$C6,$BF
	.dc.b	$F1,$B9,$27,$FD,$A3,$E0,$7E,$29,$FF,$67,$B6,$BF,$DB,$2D,$D7,$FC
	.dc.b	$87,$1F,$92,$9F,$F3,$F6,$DF,$8A,$7F,$D3,$77,$D5,$FF,$71,$CF,$D9
	.dc.b	$F0,$79,$11,$E4,$7F,$95,$CE,$91,$C8,$F5,$4F,$37,$BC,$7D,$6E,$99
	.dc.b	$3F,$0F,$3F,$83,$5F,$EE,$AF,$0B,$D0,$4B,$EF,$92,$3E,$04,$F7,$3C
	.dc.b	$F7,$DD,$C1,$FE,$F2,$F0,$FE,$11,$2F,$9B,$F1,$7C,$4F,$F6,$EF,$E3
	.dc.b	$FF,$E8,$F9,$34,$95,$FE,$9C,$8C,$B6,$3F,$D4,$31,$88,$FF,$4F,$B5
	.dc.b	$F8,$37,$E5,$DF,$EE,$47,$C6,$3F,$6E,$7E,$0F,$4F,$8A,$FF,$88,$33
	.dc.b	$FF,$5E,$DF,$C2,$F2,$3E,$E1,$1F,$8B,$E4,$FF,$BD,$BF,$03,$FF,$1F
	.dc.b	$57,$FB,$22,$51,$89,$D5,$F5,$6B,$FE,$D4,$9D,$BF,$ED,$86,$B9,$9F
	.dc.b	$F6,$EC,$62,$3D,$A7,$AC,$FA,$65,$3F,$DE,$2F,$DA,$F3,$7F,$28,$47
	.dc.b	$EE,$F9,$C5,$F8,$AF,$AA,$30,$D8,$CD,$7C,$59,$CB,$FE,$56,$1E,$EF
	.dc.b	$F0,$EA,$FF,$0D,$36,$34,$D7,$FC,$BA,$D7,$C5,$93,$D2,$EB,$1F,$E5
	.dc.b	$F8,$C6,$1E,$28,$C4,$E9,$E2,$D8,$4B,$D3,$F7,$7E,$09,$F5,$9B,$7F
	.dc.b	$7D,$AF,$76,$BE,$4F,$5A,$DF,$46,$D9,$A6,$38,$3E,$67,$D1,$7E,$11
	.dc.b	$B7,$F8,$D6,$BE,$D6,$BF,$08,$B7,$EA,$6D,$FF,$A0,$6B,$F1,$F1,$C0
	.dc.b	$FD,$89,$38,$BF,$3E,$FE,$10,$FE,$7D,$D9,$FC,$89,$FE,$6C,$7F,$37
	.dc.b	$B7,$A6,$F6,$7F,$8E,$36,$3F,$88,$B7,$C7,$FF,$4E,$36,$BF,$05,$F0
	.dc.b	$0F,$6F,$F1,$9F,$03,$FE,$5B,$8C,$60,$8C,$4E,$AB,$8B,$5F,$F3,$2C
	.dc.b	$7F,$98,$F2,$77,$FF,$8F,$23,$0D,$9F,$F1,$A4,$7F,$9A,$7D,$E7,$95
	.dc.b	$FD,$22,$42,$33,$8F,$F4,$A8,$FC,$D4,$63,$C5,$E5,$5C,$67,$F3,$BE
	.dc.b	$25,$E8,$FE,$D7,$5F,$9E,$8C,$1B,$CF,$CF,$C9,$D6,$FF,$CF,$F5,$E8
	.dc.b	$8C,$47,$FB,$9E,$3F,$D0,$1E,$FB,$D7,$FE,$8C,$DD,$3F,$C9,$ED,$F0
	.dc.b	$77,$3A,$24,$E9,$B1,$A8,$C1,$1C,$12,$FF,$75,$5D,$5B,$5C,$9B,$5C
	.dc.b	$5B,$5C,$5F,$F6,$BD,$71,$6D,$71,$6D,$71,$6D,$71,$7E,$AD,$71,$6D
	.dc.b	$71,$6D,$71,$6D,$71,$55,$C5,$B5,$C5,$B5,$C5,$B5,$C6,$B8,$12,$B8
	.dc.b	$36,$B8,$36,$78,$4E,$89,$3A,$6C,$6A,$30,$47,$04,$AF,$96,$43,$E2
	.dc.b	$94,$F7,$94,$F4,$96,$BC,$8D,$AF,$0B,$6B,$C0,$DA,$F0,$3D,$88,$76
	.dc.b	$29,$D8,$B5,$D9,$B5,$D9,$B5,$D9,$B5,$D9,$FF,$60,$9F,$EC,$EF,$4F
	.dc.b	$FD,$BC,$FF,$B0,$7D,$54,$63,$D5,$FF,$61,$CF,$E9,$3E,$87,$FD,$2E
	.dc.b	$FE,$59,$AF,$E4,$1A,$FE,$29,$AF,$E1,$1B,$91,$A4,$C7,$C7,$19,$6C
	.dc.b	$66,$7C,$51,$82,$38,$21,$E5,$72,$D3,$27,$9D,$CB,$4C,$9E,$97,$2D
	.dc.b	$32,$7A,$DC,$B4,$CC,$FA,$5B,$3E,$97,$C4,$43,$C0,$53,$B1,$4E,$85
	.dc.b	$8E,$8D,$8E,$8F,$22,$1C,$8A,$72,$2C,$72,$6C,$72,$78,$10,$E0,$58
	.dc.b	$E0,$D8,$E0,$E8,$91,$A6,$C6,$A3,$04,$70,$49,$F9,$71,$96,$C6,$67
	.dc.b	$E5,$C6,$08,$E0,$87,$C5,$72,$D3,$27,$C0,$72,$D3,$27,$B5,$CB,$FD
	.dc.b	$63,$5F,$E9,$78,$7F,$49,$24,$7F,$8C,$63,$2F,$F8,$CF,$F4,$E6,$7F
	.dc.b	$E3,$DA,$FF,$1B,$4A,$73,$7F,$B0,$8E,$15,$B6,$C6,$5F,$ED,$27,$F6
	.dc.b	$9E,$1D,$38,$E2,$FE,$D7,$87,$F6,$8A,$E1,$3F,$DC,$27,$2F,$F9,$0B
	.dc.b	$87,$FC,$A1,$F0,$3F,$DA,$68,$D2,$4F,$82,$1E,$BB,$EE,$3F,$CB,$1E
	.dc.b	$D7,$F9,$46,$73,$18,$AE,$15,$B9,$56,$E3,$0F,$84,$93,$E5,$9D,$B6
	.dc.b	$38,$7D,$D7,$58,$FF,$05,$CF,$92,$31,$1F,$E6,$BA,$46,$27,$47,$49
	.dc.b	$E3,$2F,$83,$FF,$39,$1E,$52,$C7,$A9,$AF,$06,$BD,$FE,$F2,$D7,$C6
	.dc.b	$6D,$1A,$63,$7F,$F4,$07,$89,$F0,$DB,$7F,$21,$AF,$99,$AF,$B7,$E4
	.dc.b	$BF,$D1,$1F,$0B,$F2,$DB,$1F,$41,$AF,$B9,$AF,$C5,$FB,$6F,$F4,$87
	.dc.b	$E2,$9F,$4D,$B1,$F6,$1A,$FC,$66,$BF,$A3,$F0,$1D,$CF,$87,$FF,$16
	.dc.b	$12,$7F,$49,$3A,$8F,$C0,$7C,$5F,$FA,$7D,$BB,$F1,$AF,$C3,$CE,$5B
	.dc.b	$19,$7E,$BF,$C4,$FD,$BE,$03,$49,$3F,$76,$7F,$05,$F1,$9B,$24,$64
	.dc.b	$FC,$DB,$CF,$E3,$FF,$D6,$5F,$7F,$F9,$56,$D7,$E6,$9A,$FE,$21,$AF
	.dc.b	$E3,$FE,$4B,$FD,$75,$F1,$BF,$9D,$6C,$7E,$89,$AF,$E4,$1A,$FE,$5F
	.dc.b	$A7,$FD,$85,$F8,$0F,$E9,$5B,$1F,$AA,$6B,$F9,$86,$BF,$4F,$83,$E0
	.dc.b	$2F,$07,$FC,$C3,$D7,$FE,$5F,$24,$7F,$0A,$EF,$7E,$8A,$F8,$FF,$D4
	.dc.b	$C6,$1B,$19,$E0,$F9,$93,$97,$E5,$71,$3E,$C7,$FD,$B5,$F6,$3F,$BD
	.dc.b	$6E,$3F,$C0,$9A,$FE,$C1,$AF,$E6,$E9,$F8,$8F,$E6,$DB,$1F,$9B,$6B
	.dc.b	$F9,$B6,$BF,$BF,$E5,$F8,$71,$FC,$59,$B8,$7E,$1A,$1C,$1F,$C4,$67
	.dc.b	$3C,$B2,$73,$7F,$C6,$23,$15,$FC,$5F,$E6,$7F,$DE,$DF,$81,$DB,$34
	.dc.b	$97,$FA,$0E,$2E,$6F,$F3,$8E,$2F,$F4,$9E,$1F,$E9,$12,$47,$D3,$E0
	.dc.b	$F9,$D1,$FE,$FB,$E0,$F9,$CD,$8C,$CE,$F8,$7D,$BA,$B3,$7D,$47,$B5
	.dc.b	$B5,$FC,$CA,$75,$39,$3A,$57,$B2,$7A,$57,$B1,$BA,$6A,$31,$1F,$CA
	.dc.b	$3E,$77,$FD,$CC,$D8,$FF,$73,$37,$DA,$FE,$7D,$3A,$8F,$F2,$6E,$5F
	.dc.b	$8B,$E1,$E8,$48,$F0,$BF,$EB,$56,$1F,$2C,$A7,$F4,$0F,$C0,$DF,$3F
	.dc.b	$FC,$F5,$B8,$FE,$F5,$BC,$FF,$D5,$5B,$5F,$D6,$D7,$DB,$7F,$6A,$8F
	.dc.b	$95,$18,$AF,$ED,$21,$C7,$0F,$3A,$E1,$39,$F9,$BF,$F3,$0B,$4F,$EF
	.dc.b	$E5,$BF,$14,$7F,$99,$38,$FF,$DA,$B1,$EC,$3F,$74,$59,$F0,$7D,$27
	.dc.b	$4C,$4D,$24,$7F,$9E,$AB,$9D,$7C,$72,$1F,$91,$F0,$BF,$9B,$BF,$0F
	.dc.b	$E9,$3E,$41,$2B,$E4,$3F,$2F,$D6,$FC,$67,$DC,$FC,$96,$D7,$C9,$6B
	.dc.b	$F4,$DA,$FE,$4F,$85,$F2,$FF,$95,$FC,$AF,$FC,$0D,$18,$C5,$18,$9D
	.dc.b	$57,$F8,$3E,$B5,$3F,$16,$3F,$8E,$65,$EA,$7D,$A7,$C8,$3E,$E7,$D4
	.dc.b	$FE,$0C,$98,$7E,$0D,$FC,$37,$D5,$FD,$E9,$FC,$13,$5F,$C6,$35,$FC
	.dc.b	$1B,$5F,$0F,$6F,$CF,$BF,$14,$7C,$D8,$C1,$F8,$0E,$07,$D6,$7E,$54
	.dc.b	$98,$FE,$55,$FC,$C7,$B9,$FB,$9F,$A3,$FF,$56,$CF,$E3,$E3,$0F,$E5
	.dc.b	$DB,$87,$E8,$5A,$FC,$13,$C4,$7C,$7E,$2F,$C1,$BF,$A7,$EE,$FE,$0D
	.dc.b	$F0,$7D,$B2,$FE,$0F,$82,$31,$3A,$8F,$D3,$36,$83,$E7,$DF,$3D,$FE
	.dc.b	$2E,$AF,$F1,$9C,$EA,$73,$EC,$7F,$15,$8F,$F6,$4C,$62,$77,$1C,$EB
	.dc.b	$93,$E5,$FA,$DF,$F2,$6C,$FE,$DD,$B1,$FC,$05,$AF,$EE,$1A,$FD,$63
	.dc.b	$E1,$70,$7D,$67,$E3,$1D,$F0,$F2,$57,$F9,$B3,$0F,$24,$F9,$B0,$F1
	.dc.b	$CE,$5B,$1F,$38,$FC,$89,$DC,$BA,$7D,$07,$B9,$93,$81,$FE,$7A,$F8
	.dc.b	$1F,$F7,$6F,$CF,$FF,$A1,$A3,$FD,$EA,$D7,$FD,$E9,$DD,$F4,$23,$13
	.dc.b	$C3,$D8,$E5,$C3,$FE,$89,$9C,$B6,$33,$5C,$5B,$6A,$ED,$83,$5F,$F1
	.dc.b	$DD,$7F,$A5,$A3,$1C,$9E,$3B,$FF,$07,$4E,$6B,$71,$FD,$7E,$31,$3B
	.dc.b	$8F,$D3,$C6,$0F,$F4,$CB,$F4,$74,$FF,$50,$BF,$87,$D3,$F1,$53,$F4
	.dc.b	$76,$FE,$B8,$DA,$DD,$78,$8F,$01,$0F,$D9,$47,$F3,$FD,$FF,$B2,$7C
	.dc.b	$47,$E2,$E8,$D2,$3F,$76,$FB,$B4,$EF,$C1,$F9,$7B,$F1,$1F,$A6,$87
	.dc.b	$89,$48,$FE,$A0,$C9,$E5,$C7,$D0,$79,$8C,$3F,$BF,$1F,$82,$7E,$D7
	.dc.b	$87,$FB,$E3,$63,$27,$F0,$32,$CF,$F6,$3E,$7F,$7B,$FD,$75,$AE,$63
	.dc.b	$E0,$B9,$6B,$E3,$21,$E6,$23,$FD,$87,$A3,$FC,$1D,$F8,$27,$F6,$B6
	.dc.b	$CF,$F2,$A9,$4E,$5B,$1F,$CD,$1E,$EF,$F6,$EE,$AF,$2B,$63,$27,$F6
	.dc.b	$F8,$CC,$7E,$28,$F9,$45,$3C,$F1,$FE,$5A,$ED,$FF,$11,$7E,$57,$FD
	.dc.b	$2E,$0D,$24,$FF,$59,$87,$EF,$5F,$D7,$5B,$19,$8C,$1F,$D9,$E3,$CB
	.dc.b	$D8,$F0,$BD,$31,$FE,$05,$6C,$66,$3D,$F5,$E7,$3E,$61,$4C,$3F,$DF
	.dc.b	$8F,$D7,$17,$83,$FC,$31,$E1,$FE,$8A,$7F,$C3,$0D,$7E,$B9,$CF,$A7
	.dc.b	$F6,$F3,$FA,$F8,$FF,$0F,$78,$9B,$7F,$C3,$AD,$BF,$F0,$B4,$FF,$85
	.dc.b	$1F,$0E,$1F,$E2,$1F,$88,$FF,$0F,$41,$A4,$BF,$F1,$57,$91,$F0,$09
	.dc.b	$3F,$19,$FB,$38,$F9,$27,$EC,$B5,$FC,$7F,$77,$FA,$2F,$F0,$5F,$F2
	.dc.b	$71,$23,$FC,$A6,$FF,$92,$AD,$F8,$2D,$24,$FF,$54,$3F,$E0,$07,$47
	.dc.b	$F9,$8E,$3F,$CB,$1F,$82,$C9,$FF,$2E,$B5,$FF,$65,$D3,$EA,$3F,$CD
	.dc.b	$04,$8F,$F3,$44,$65,$FF,$32,$FE,$0B,$FE,$63,$23,$83,$FD,$00,$FF
	.dc.b	$22,$23,$83,$CD,$0F,$39,$EA,$7F,$9E,$70,$FF,$46,$12,$B7,$3A,$A4
	.dc.b	$FF,$2A,$7F,$DB,$22,$F1,$2F,$57,$FD,$09,$18,$6D,$FF,$A7,$6B,$83
	.dc.b	$EE,$9F,$F4,$41,$1C,$1F,$EB,$1F,$C5,$3F,$D6,$84,$70,$7F,$AF,$65
	.dc.b	$F8,$E7,$FB,$05,$B1,$98,$C1,$FE,$8B,$F0,$E4,$7F,$A4,$1A,$FF,$8A
	.dc.b	$DF,$E9,$64,$7F,$D5,$8F,$FB,$96,$7D,$31,$FE,$E3,$6B,$FE,$EB,$69
	.dc.b	$F3,$FD,$BF,$83,$1F,$EF,$18,$7C,$EF,$FB,$DD,$B1,$FE,$F7,$6B,$B8
	.dc.b	$F9,$BE,$FF,$FA,$BF,$F0,$8B,$46,$92,$BE,$76,$1F,$46,$7F,$8E,$EB
	.dc.b	$F5,$5B,$AF,$F0,$A8,$FF,$6B,$4F,$FA,$E4,$95,$F9,$27,$EE,$FC,$17
	.dc.b	$FA,$FF,$E7,$2E,$06,$92,$BF,$DA,$4F,$F7,$42,$38,$3F,$1B,$F0,$BF
	.dc.b	$EC,$9F,$C6,$FF,$12,$D8,$CD,$27,$2D,$8F,$F7,$A7,$CE,$7F,$BD,$C8
	.dc.b	$E0,$FC,$CF,$77,$FB,$56,$FF,$2A,$D8,$CD,$27,$2D,$8F,$D0,$FC,$67
	.dc.b	$FB,$77,$E6,$FF,$3A,$D8,$CC,$A7,$2D,$8F,$D4,$FC,$76,$DF,$F2,$CC
	.dc.b	$7E,$C6,$31,$1F,$EE,$9D,$91,$89,$D1,$CE,$78,$CF,$F9,$7F,$6F,$D7
	.dc.b	$F9,$FF,$64,$95,$FD,$03,$E5,$9F,$23,$FD,$69,$B1,$98,$C1,$F8,$1E
	.dc.b	$69,$1F,$86,$6D,$FF,$8A,$31,$5F,$2F,$97,$C3,$63,$FC,$5A,$D7,$F9
	.dc.b	$23,$5F,$F3,$CF,$C9,$BE,$51,$8F,$F3,$D9,$FF,$1A,$CE,$63,$11,$FC
	.dc.b	$B6,$31,$E3,$78,$DB,$1F,$E3,$CF,$95,$71,$F4,$AF,$44,$62,$F9,$D7
	.dc.b	$FA,$2A,$31,$3A,$AE,$13,$FE,$46,$F1,$BF,$CA,$33,$FE,$4C,$24,$67
	.dc.b	$D9,$7B,$3D,$21,$F2,$EF,$97,$FE,$C1,$18,$6C,$67,$3F,$F2,$B3,$72
	.dc.b	$F9,$CD,$AF,$F2,$BB,$5F,$E9,$CD,$7F,$D4,$6D,$7B,$4F,$4A,$EE,$47
	.dc.b	$04,$AE,$F1,$96,$C6,$67,$C3,$39,$6C,$65,$F1,$C6,$5B,$19,$7C,$B1
	.dc.b	$96,$C6,$5F,$3C,$F8,$08,$F6,$6B,$D1,$AF,$26,$BC,$8E,$04,$38,$14
	.dc.b	$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$4A,$FA,$44,$AF,$81,$19,$6C
	.dc.b	$66,$BE,$13,$6B,$D6,$DA,$F3,$35,$F1,$B5,$F1,$35,$ED,$3D,$2B,$B9
	.dc.b	$1C,$12,$BB,$C6,$5B,$19,$9F,$0C,$E5,$B1,$97,$C7,$19,$6C,$65,$FF
	.dc.b	$7B,$FB,$AF,$65,$E6,$7F,$0B,$9F,$F6,$69,$FC,$36,$75,$19,$6C,$7E
	.dc.b	$93,$E6,$BF,$DF,$4F,$E9,$1A,$FE,$91,$AF,$E9,$1A,$FE,$D1,$B9,$1A
	.dc.b	$4C,$3C,$91,$96,$C6,$67,$C5,$18,$23,$82,$1E,$57,$2D,$32,$79,$DC
	.dc.b	$B4,$C9,$E9,$72,$D3,$27,$AD,$CB,$4C,$CF,$A5,$B3,$E9,$7C,$44,$3C
	.dc.b	$05,$3B,$14,$E8,$58,$E8,$D8,$E8,$F2,$21,$C8,$A7,$22,$C7,$26,$C7
	.dc.b	$27,$81,$0E,$05,$8E,$0D,$8E,$0E,$89,$1A,$6C,$6A,$30,$47,$04,$9F
	.dc.b	$97,$19,$6C,$66,$7E,$5C,$60,$8E,$08,$7C,$57,$2D,$32,$7C,$07,$2D
	.dc.b	$32,$7B,$5C,$B4,$C8,$D9,$AF,$F5,$2C,$7F,$BB,$46,$35,$FD,$7D,$7F
	.dc.b	$57,$6B,$FE,$97,$8D,$4F,$F7,$AE,$6F,$EE,$6D,$C3,$74,$3F,$C0,$33
	.dc.b	$A7,$FB,$FB,$5F,$F5,$17,$CD,$3E,$6B,$11,$C6,$31,$5D,$21,$18,$BF
	.dc.b	$F0,$73,$63,$FD,$2F,$E8,$F0,$7F,$D2,$0D,$73,$7F,$C0,$2D,$39,$9F
	.dc.b	$F0,$8E,$1F,$C0,$B9,$BF,$C4,$31,$ED,$21,$DB,$EE,$F9,$51,$A4,$BF
	.dc.b	$F1,$11,$DB,$EF,$3F,$C3,$3E,$07,$91,$B7,$E8,$6B,$FE,$40,$B4,$62
	.dc.b	$3F,$C4,$30,$FC,$07,$17,$EA,$6C,$66,$17,$F0,$A7,$3F,$7D,$FE,$2A
	.dc.b	$FC,$67,$E2,$B7,$0F,$92,$D7,$C4,$D7,$C3,$CF,$70,$F9,$2D,$C7,$FC
	.dc.b	$2D,$7F,$42,$31,$F8,$0F,$F8,$E3,$F1,$2F,$A6,$DC,$3E,$BB,$5F,$80
	.dc.b	$D7,$F1,$5C,$7F,$95,$FC,$67,$F2,$FC,$DF,$E2,$A6,$F3,$1A,$7E,$03
	.dc.b	$C6,$FF,$23,$B7,$5D,$61,$F8,$69,$CB,$63,$2F,$D1,$E8,$FB,$A4,$8F
	.dc.b	$B7,$B7,$F0,$C7,$F1,$3C,$7F,$7E,$BF,$1E,$D7,$FD,$D3,$F1,$3F,$E7
	.dc.b	$AF,$17,$FD,$C4,$D8,$C9,$FE,$5C,$8C,$B4,$E3,$9D,$7F,$D0,$73,$FE
	.dc.b	$4A,$B5,$74,$AD,$57,$F9,$9A,$B5,$3F,$A2,$E8,$FC,$B9,$2F,$D6,$FF
	.dc.b	$60,$DB,$F7,$B7,$FC,$07,$AB,$E6,$91,$FD,$E1,$FB,$DF,$11,$F5,$1F
	.dc.b	$B8,$84,$63,$0F,$DC,$36,$BF,$D0,$8D,$9F,$C7,$CB,$E2,$7F,$26,$D8
	.dc.b	$CC,$60,$FC,$64,$66,$3F,$4D,$CB,$FE,$84,$9C,$92,$3E,$8F,$D1,$7F
	.dc.b	$0C,$A6,$A8,$C3,$63,$F8,$C9,$A8,$CF,$D2,$C8,$CC,$A3,$19,$7D,$CF
	.dc.b	$A6,$E6,$4A,$DB,$F1,$75,$FE,$5D,$F4,$BF,$98,$26,$5F,$CB,$DF,$8B
	.dc.b	$B7,$FA,$79,$B6,$69,$27,$DB,$19,$6C,$7E,$3B,$3E,$51,$83,$EC,$9F
	.dc.b	$5A,$BF,$D5,$7F,$7D,$D8,$91,$AD,$FE,$D4,$7E,$63,$EA,$3F,$D6,$3F
	.dc.b	$1D,$72,$34,$97,$C7,$0F,$F5,$BF,$E0,$DE,$A2,$46,$69,$1F,$BA,$88
	.dc.b	$7F,$51,$29,$FA,$0F,$AA,$FE,$71,$C2,$69,$2B,$F9,$7C,$3E,$77,$64
	.dc.b	$8C,$BF,$D0,$7E,$71,$F6,$74,$93,$FC,$E2,$32,$FF,$94,$78,$7F,$B3
	.dc.b	$63,$E4,$3F,$8A,$F7,$FF,$AF,$1C,$4F,$E3,$B4,$F3,$91,$F2,$BF,$AF
	.dc.b	$9E,$13,$EF,$21,$FD,$BD,$CB,$FC,$03,$EB,$3E,$09,$2B,$DE,$FC,$4F
	.dc.b	$97,$FF,$72,$FE,$51,$FC,$39,$B5,$FC,$49,$AF,$F0,$A6,$BF,$C1,$FE
	.dc.b	$BF,$FD,$DB,$F3,$3F,$C5,$9B,$1F,$C6,$9A,$FF,$12,$6B,$F0,$BB,$FE
	.dc.b	$36,$5E,$88,$C1,$F1,$38,$1F,$61,$F6,$89,$A7,$D8,$7E,$EF,$D8,$FE
	.dc.b	$52,$BF,$94,$B5,$FE,$42,$D7,$F8,$F3,$5F,$F0,$F5,$F8,$5B,$87,$90
	.dc.b	$8F,$F3,$2F,$B2,$61,$FE,$17,$3F,$9B,$B8,$8F,$F4,$4C,$7F,$08,$8C
	.dc.b	$3A,$38,$47,$F3,$4F,$B3,$E7,$F8,$37,$0C,$D1,$8A,$FE,$86,$DA,$E1
	.dc.b	$0A,$FF,$4A,$F1,$7E,$3A,$7F,$06,$7C,$A2,$CF,$F1,$AF,$B6,$F4,$BD
	.dc.b	$DA,$FF,$50,$6C,$7F,$48,$6B,$FB,$8F,$B0,$FF,$0C,$4F,$AA,$31,$E3
	.dc.b	$FB,$9F,$F1,$D7,$5A,$B7,$1F,$EA,$C8,$C4,$EE,$3F,$A8,$7D,$B7,$CC
	.dc.b	$FA,$8F,$84,$DC,$3F,$B1,$B6,$7F,$AD,$B5,$F7,$1F,$A1,$DF,$EB,$BF
	.dc.b	$0B,$0E,$F5,$E7,$6C,$78,$F1,$F2,$CF,$F5,$69,$CB,$F2,$A7,$F0,$67
	.dc.b	$EE,$4A,$67,$3F,$A0,$F8,$4D,$7B,$3F,$EC,$CE,$57,$B7,$FD,$8A,$30
	.dc.b	$DF,$6F,$FC,$D3,$3A,$AF,$F6,$DC,$62,$75,$E3,$FF,$99,$6B,$55,$FE
	.dc.b	$DC,$6D,$F4,$8E,$33,$BA,$F8,$87,$F8,$80,$91,$FE,$77,$6B,$A3,$04
	.dc.b	$3E,$DE,$7F,$80,$8F,$91,$BF,$E8,$9B,$9F,$B8,$F9,$53,$B7,$95,$7E
	.dc.b	$FE,$7B,$B4,$FB,$B1,$FE,$02,$2F,$89,$F8,$27,$11,$F8,$06,$BF,$90
	.dc.b	$7C,$0F,$9A,$FC,$91,$E2,$6C,$66,$36,$FF,$87,$F8,$7F,$77,$F4,$5F
	.dc.b	$A3,$93,$48,$FA,$A7,$F8,$2E,$1E,$A6,$9F,$B6,$87,$E0,$DF,$BB,$E5
	.dc.b	$FE,$0E,$D8,$C9,$B7,$F8,$07,$3F,$9A,$BF,$C6,$4D,$9F,$E1,$47,$A0
	.dc.b	$B7,$FC,$03,$A1,$C9,$F8,$4B,$F9,$87,$DB,$21,$FC,$82,$9F,$81,$FF
	.dc.b	$93,$3A,$BF,$8D,$B6,$33,$3F,$CA,$9A,$FF,$92,$BA,$FF,$00,$FF,$95
	.dc.b	$9A,$E6,$3C,$47,$F8,$A2,$33,$1E,$42,$1F,$20,$F2,$C7,$F8,$BB,$B3
	.dc.b	$F9,$6F,$07,$F4,$B8,$FF,$53,$4F,$F4,$B6,$BF,$D2,$DA,$FF,$96,$7B
	.dc.b	$7F,$9A,$FE,$7D,$FD,$6E,$4D,$24,$FF,$9B,$BC,$0F,$2C,$FF,$8F,$9B
	.dc.b	$DE,$69,$EA,$9F,$41,$EA,$EB,$FF,$71,$70,$7E,$72,$FF,$BD,$1F,$0A
	.dc.b	$FF,$CB,$91,$FE,$7A,$F0,$7A,$9C,$EB,$8D,$FE,$35,$AF,$C4,$F0,$BF
	.dc.b	$C1,$75,$FE,$32,$C1,$18,$9F,$F3,$6C,$63,$C3,$F5,$CE,$67,$FD,$03
	.dc.b	$E2,$7B,$09,$5E,$57,$DF,$E2,$C8,$F4,$37,$0F,$CF,$63,$E4,$BF,$81
	.dc.b	$CB,$F9,$09,$FF,$27,$FE,$0D,$27,$FC,$6C,$D7,$FD,$23,$E4,$7F,$68
	.dc.b	$FC,$1B,$FC,$84,$48,$FF,$4F,$C6,$5F,$F1,$F7,$37,$F9,$68,$8E,$0F
	.dc.b	$F2,$CF,$C3,$7F,$95,$F4,$34,$91,$FE,$47,$7F,$CC,$64,$70,$7F,$97
	.dc.b	$33,$67,$FA,$A9,$FF,$62,$92,$75,$19,$7F,$D6,$BE,$77,$F8,$0B,$E0
	.dc.b	$FF,$CF,$58,$1A,$49,$FF,$5D,$FA,$1B,$6C,$ED,$B3,$FE,$BE,$9D,$47
	.dc.b	$FA,$E3,$CF,$AF,$FA,$D5,$B5,$FE,$8F,$39,$46,$DA,$FF,$BC,$17,$E0
	.dc.b	$7F,$EA,$22,$47,$E2,$3D,$67,$A9,$FE,$1C,$9F,$F5,$59,$27,$FD,$5E
	.dc.b	$D7,$FD,$77,$E3,$7F,$AD,$F2,$FF,$47,$92,$3F,$CE,$CF,$FB,$B8,$8F
	.dc.b	$CA,$F0,$7C,$AF,$F8,$BA,$3F,$D9,$D8,$FF,$08,$C5,$EA,$FE,$D1,$B1
	.dc.b	$98,$FD,$AD,$FF,$8C,$BD,$AF,$F6,$D4,$9A,$4C,$3F,$DC,$F1,$96,$C7
	.dc.b	$FB,$A2,$31,$1F,$EE,$63,$A1,$F1,$7C,$0F,$11,$23,$CC,$F9,$7D,$89
	.dc.b	$1E,$96,$F4,$7E,$DF,$27,$B3,$FE,$48,$FC,$67,$0D,$C7,$E9,$35,$FF
	.dc.b	$26,$4F,$0C,$15,$FE,$B2,$9F,$F2,$AB,$79,$BF,$33,$A6,$A7,$E7,$4E
	.dc.b	$79,$7F,$A2,$BE,$3D,$7F,$CB,$3F,$89,$DC,$4D,$25,$FE,$17,$8F,$FC
	.dc.b	$BF,$F8,$CF,$FB,$A8,$91,$9A,$47,$FB,$C8,$91,$FE,$F6,$6B,$EC,$F8
	.dc.b	$1F,$F3,$4F,$07,$C5,$6C,$7E,$4D,$AE,$E9,$3F,$2F,$E0,$F3,$3F,$E7
	.dc.b	$2D,$D1,$8A,$46,$1E,$33,$F9,$F9,$D4,$FF,$86,$E3,$EF,$4F,$DC,$27
	.dc.b	$0F,$CF,$F8,$47,$C2,$FF,$41,$6C,$FF,$6F,$A4,$62,$B8,$57,$E6,$79
	.dc.b	$65,$FE,$64,$93,$F6,$FE,$1B,$F3,$BC,$F6,$BF,$7F,$8A,$72,$FF,$BC
	.dc.b	$9A,$89,$8F,$92,$BC,$38,$7E,$BA,$31,$3F,$EF,$A6,$BF,$E6,$79,$FD
	.dc.b	$AB,$4F,$DB,$3F,$8F,$F8,$7F,$A3,$5E,$F6,$E3,$9D,$7F,$B2,$46,$2F
	.dc.b	$C8,$79,$5C,$38,$6B,$98,$E2,$4A,$EC,$DA,$ED,$18,$23,$82,$4F,$80
	.dc.b	$57,$66,$D7,$37,$51,$92,$46,$63,$CC,$D7,$CD,$5E,$07,$8C,$65,$EB
	.dc.b	$19,$78,$B5,$E3,$5E,$67,$2D,$72,$74,$8E,$0F,$48,$D3,$4D,$38,$AF
	.dc.b	$4C,$F0,$6C,$66,$BC,$D1,$8A,$E2,$43,$C6,$53,$C6,$60,$86,$08,$78
	.dc.b	$CC,$94,$CB,$DD,$CB,$5C,$C7,$90,$95,$E9,$6D,$7A,$63,$04,$70,$49
	.dc.b	$F4,$1E,$12,$57,$76,$D7,$47,$71,$92,$46,$63,$CE,$D7,$CF,$5E,$17
	.dc.b	$8C,$7F,$90,$FD,$67,$B0,$F5,$FF,$9B,$52,$73,$18,$6C,$7F,$3B,$6B
	.dc.b	$F9,$9E,$87,$9E,$C7,$9C,$61,$B1,$99,$FE,$89,$39,$6C,$65,$E8,$D7
	.dc.b	$F6,$B1,$B7,$33,$D0,$8F,$26,$BA,$30,$43,$04,$3A,$CF,$77,$F8,$B4
	.dc.b	$F7,$73,$3D,$1A,$73,$29,$CC,$E3,$3C,$DF,$F4,$FB,$67,$A4,$6A,$7A
	.dc.b	$34,$E8,$53,$A1,$C9,$DB,$4D,$94,$DC,$76,$70,$79,$88,$75,$29,$A3
	.dc.b	$04,$30,$47,$A1,$27,$D8,$49,$F2,$C6,$5B,$19,$9F,$3C,$E5,$B1,$97
	.dc.b	$D2,$D7,$CB,$1B,$73,$3D,$08,$F2,$6B,$A3,$04,$30,$43,$AC,$F7,$7E
	.dc.b	$04,$F7,$73,$3D,$1A,$73,$29,$CC,$E3,$3C,$DF,$90,$D9,$E9,$1A,$9E
	.dc.b	$8F,$F6,$5E,$AF,$F1,$2E,$1F,$E2,$3A,$FD,$B4,$FF,$94,$E3,$07,$F6
	.dc.b	$82,$3F,$D9,$F8,$7F,$65,$18,$8F,$EE,$16,$F3,$1E,$5F,$ED,$FD,$BF
	.dc.b	$F5,$B4,$62,$51,$89,$D5,$7F,$AE,$2B,$53,$E5,$88,$7F,$99,$CA,$7F
	.dc.b	$6C,$FB,$E7,$E0,$3F,$A3,$6E,$1F,$C4,$1B,$3F,$E4,$76,$BD,$F2,$FE
	.dc.b	$19,$F1,$3E,$66,$C6,$69,$39,$6C,$79,$EF,$6F,$C2,$C7,$C3,$5F,$02
	.dc.b	$73,$3F,$C6,$67,$2F,$BF,$E7,$BF,$8E,$7D,$4F,$C6,$6C,$67,$04,$E5
	.dc.b	$B1,$FE,$DE,$FB,$4F,$E4,$91,$88,$FA,$71,$89,$FF,$73,$36,$38,$4F
	.dc.b	$67,$7F,$1F,$FC,$B7,$E2,$F6,$D8,$EC,$D7,$72,$9F,$F3,$97,$E1,$0F
	.dc.b	$BF,$D5,$7F,$8C,$E3,$15,$B8,$FF,$0A,$46,$2B,$8D,$70,$A6,$0F,$83
	.dc.b	$C3,$63,$96,$1E,$E9,$C9,$AF,$23,$F1,$71,$F0,$9A,$FF,$49,$AF,$F6
	.dc.b	$ED,$7E,$4A,$3E,$54,$E9,$70,$FE,$19,$FC,$76,$FF,$8D,$25,$7F,$A4
	.dc.b	$4F,$3F,$D6,$7F,$55,$FC,$9F,$E7,$B6,$BE,$B3,$5F,$AA,$D7,$C9,$F8
	.dc.b	$25,$AF,$82,$DC,$8D,$31,$F2,$BE,$C9,$F8,$ED,$93,$4F,$67,$57,$FA
	.dc.b	$92,$FF,$52,$7F,$03,$D1,$D5,$FD,$DA,$75,$3F,$E4,$9D,$7E,$13,$FE
	.dc.b	$6D,$6B,$F1,$B5,$F8,$6F,$E3,$BE,$43,$FB,$B7,$E0,$57,$43,$49,$3F
	.dc.b	$99,$F9,$8D,$B6,$76,$D9,$FC,$CC,$EA,$3F,$3B,$D3,$EF,$25,$7D,$B8
	.dc.b	$CB,$5F,$46,$1E,$67,$D8,$D7,$F4,$FE,$1F,$C9,$24,$F1,$3F,$26,$FC
	.dc.b	$AD,$7E,$3C,$FC,$A1,$A7,$C8,$7F,$CF,$ED,$7C,$35,$FC,$4E,$FC,$11
	.dc.b	$83,$F6,$BC,$9F,$64,$93,$E1,$E0,$FD,$4F,$67,$F1,$9B,$E3,$7F,$46
	.dc.b	$BA,$1F,$93,$A7,$CC,$FE,$0C,$91,$CD,$FC,$F7,$07,$DA,$25,$F4,$CB
	.dc.b	$D1,$87,$B3,$3F,$7B,$FC,$C4,$87,$F0,$EF,$9D,$FF,$0E,$63,$FB,$86
	.dc.b	$E4,$69,$8F,$BA,$93,$E7,$6D,$7F,$A5,$E3,$FD,$AF,$F5,$9D,$3C,$C3
	.dc.b	$49,$1F,$97,$C7,$CB,$5F,$41,$AF,$E7,$79,$74,$7C,$5C,$7C,$75,$49
	.dc.b	$3F,$6F,$3F,$82,$4B,$E7,$3F,$C5,$3E,$83,$DA,$47,$E8,$3F,$03,$E8
	.dc.b	$BF,$CA,$DF,$80,$FF,$30,$6E,$1F,$CD,$DA,$FF,$36,$6B,$FC,$57,$F0
	.dc.b	$0F,$F2,$F7,$77,$F3,$86,$C7,$F3,$F6,$BF,$CF,$9A,$FB,$CF,$56,$3F
	.dc.b	$09,$FB,$5A,$7E,$17,$EF,$FE,$A9,$33,$C1,$F8,$9F,$AB,$FE,$14,$47
	.dc.b	$F2,$4F,$E8,$67,$E5,$D7,$67,$F2,$1F,$5B,$FE,$71,$AF,$EA,$8D,$9F
	.dc.b	$EA,$6D,$7F,$99,$36,$1F,$7C,$FB,$0F,$F0,$44,$7F,$BA,$0C,$B8,$AF
	.dc.b	$F0,$7C,$EE,$7F,$C1,$66,$DC,$74,$FF,$9F,$78,$7E,$D6,$3E,$8B,$FC
	.dc.b	$C9,$FE,$8B,$A7,$DF,$78,$DB,$87,$F8,$72,$7B,$CE,$4F,$57,$CC,$FF
	.dc.b	$46,$F4,$6F,$5F,$FA,$82,$F8,$5A,$B5,$18,$6C,$7F,$A4,$E3,$E0,$F3
	.dc.b	$FB,$63,$C8,$D8,$C9,$EE,$8F,$83,$1B,$7E,$E9,$F4,$CF,$E7,$B3,$DA
	.dc.b	$51,$23,$0D,$AF,$43,$FE,$A5,$8C,$BF,$8B,$3E,$39,$D8,$B9,$FC,$C7
	.dc.b	$C0,$76,$3B,$FA,$3F,$EB,$CE,$47,$7E,$63,$0D,$CF,$FC,$A3,$3A,$AE
	.dc.b	$31,$89,$D7,$A1,$EA,$AD,$57,$FB,$2E,$B4,$F0,$9D,$57,$08,$FF,$05
	.dc.b	$3F,$C1,$77,$E3,$1F,$C1,$63,$FC,$C4,$47,$D4,$7E,$A1,$C3,$C5,$AF
	.dc.b	$DD,$CF,$F0,$11,$F2,$5B,$A7,$C3,$D3,$F0,$33,$9C,$7F,$DB,$4F,$E5
	.dc.b	$79,$7F,$B8,$BD,$30,$F0,$1E,$2D,$53,$A8,$FC,$F6,$DF,$EE,$3F,$81
	.dc.b	$FC,$57,$0F,$F2,$6C,$D1,$89,$D4,$70,$32,$79,$74,$F3,$DF,$E0,$F3
	.dc.b	$F2,$CE,$5B,$19,$8F,$B6,$7E,$BF,$87,$31,$FA,$B6,$C7,$EF,$1B,$8F
	.dc.b	$EB,$34,$F8,$1B,$7D,$2B,$F7,$9F,$34,$F8,$1A,$7B,$DF,$DF,$F1,$FA
	.dc.b	$9B,$8F,$A9,$BA,$FC,$29,$D4,$7F,$0E,$9F,$3B,$FC,$49,$AE,$70,$FD
	.dc.b	$97,$73,$97,$D7,$3E,$43,$F0,$C7,$93,$2F,$47,$07,$D4,$7F,$17,$87
	.dc.b	$F9,$63,$9B,$EA,$57,$F9,$D5,$AF,$F9,$D1,$AF,$F9,$4D,$B1,$FE,$60
	.dc.b	$E7,$FC,$2C,$FF,$9F,$5A,$FF,$9F,$1A,$FF,$96,$DA,$F9,$C9,$7E,$5C
	.dc.b	$BF,$2F,$5F,$E6,$CE,$82,$F8,$1F,$E9,$CD,$BF,$E9,$CD,$7F,$A7,$35
	.dc.b	$FF,$35,$75,$72,$FA,$AF,$EB,$D2,$69,$27,$FD,$2D,$DB,$E8,$9F,$ED
	.dc.b	$47,$C8,$DD,$19,$6C,$7C,$EE,$2F,$C1,$57,$D2,$3F,$0C,$43,$FC,$F9
	.dc.b	$DD,$C4,$93,$D5,$F8,$BA,$FC,$39,$FB,$5A,$FE,$BE,$31,$1F,$ED,$AE
	.dc.b	$FF,$D5,$4F,$F8,$4D,$B1,$FD,$4D,$AE,$FA,$7D,$1E,$04,$C7,$D8,$DB
	.dc.b	$FF,$58,$E3,$E4,$6C,$F9,$4F,$77,$1C,$8F,$26,$7F,$03,$1F,$F1,$F4
	.dc.b	$7F,$90,$6B,$F5,$1D,$5F,$ED,$E8,$C1,$FE,$BE,$F0,$BF,$17,$B7,$F9
	.dc.b	$0A,$BF,$81,$DF,$FB,$1E,$72,$7E,$49,$EA,$7F,$99,$23,$AF,$85,$F8
	.dc.b	$CC,$3B,$4F,$FB,$3F,$C3,$EA,$4A,$EB,$1F,$ED,$08,$E9,$3C,$23,$FD
	.dc.b	$53,$3D,$09,$8F,$BB,$3F,$81,$3F,$50,$FF,$6E,$47,$A8,$93,$DA,$3F
	.dc.b	$D7,$51,$DB,$0F,$31,$27,$AC,$FF,$B9,$7C,$4F,$86,$48,$E4,$7F,$B3
	.dc.b	$9F,$8F,$E2,$FE,$93,$EA,$3F,$D1,$F2,$69,$2B,$FD,$E1,$E4,$6D,$B3
	.dc.b	$B6,$CF,$FA,$84,$FA,$AF,$CD,$C6,$BF,$11,$F8,$44,$7F,$DF,$9E,$4F
	.dc.b	$DA,$25,$7A,$37,$FF,$5E,$B7,$CA,$FD,$DF,$E4,$FF,$EF,$E6,$DF,$F8
	.dc.b	$A6,$BF,$D9,$34,$AD,$36,$7F,$DA,$AD,$7F,$DA,$AD,$7F,$DD,$1E,$67
	.dc.b	$47,$33,$EA,$7F,$D3,$B8,$D7,$F1,$67,$FB,$C6,$BD,$1C,$7F,$94,$23
	.dc.b	$EB,$F2,$3F,$89,$FC,$0F,$FA,$07,$2F,$F7,$F3,$63,$37,$EB,$6B,$FE
	.dc.b	$0F,$F4,$1C,$B8,$8F,$F0,$FC,$63,$05,$FF,$7C,$94,$E6,$30,$D8,$F9
	.dc.b	$1E,$8F,$3A,$62,$8C,$36,$3A,$CF,$06,$BF,$E3,$07,$89,$FC,$C3,$E7
	.dc.b	$7D,$CD,$8F,$AC,$D7,$84,$FF,$36,$84,$FF,$A9,$7D,$57,$AB,$BB,$FE
	.dc.b	$77,$18,$AD,$C7,$F9,$A6,$31,$5C,$6B,$85,$25,$CF,$FD,$13,$D9,$EC
	.dc.b	$DC,$39,$35,$F6,$4F,$0A,$57,$97,$D7,$D5,$79,$1B,$3D,$DB,$47,$73
	.dc.b	$F8,$83,$F1,$67,$E9,$C8,$7F,$3F,$3F,$AE,$F0,$7F,$25,$F0,$7F,$9C
	.dc.b	$C7,$C2,$F6,$3F,$B5,$D9,$A4,$C3,$F8,$54,$BD,$CF,$14,$3D,$CF,$C3
	.dc.b	$46,$1B,$97,$F6,$86,$CE,$EB,$FC,$09,$1F,$8D,$C7,$FC,$08,$DC,$9D
	.dc.b	$DB,$AF,$F2,$94,$FE,$4E,$1E,$DF,$F7,$A6,$E1,$FE,$25,$6F,$3E,$E7
	.dc.b	$44,$9D,$36,$35,$18,$23,$82,$5F,$F9,$7A,$BA,$B6,$B9,$36,$B8,$B6
	.dc.b	$B8,$BF,$E7,$FA,$E2,$DA,$E2,$DA,$E2,$DA,$E2,$FF,$A5,$2B,$8B,$6B
	.dc.b	$8B,$6B,$8B,$6B,$8B,$F3,$27,$6D,$9D,$B6,$77,$5D,$A3,$07,$AC,$87
	.dc.b	$94,$A7,$72,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$3E,$32,$57,$A9,$B5
	.dc.b	$EA,$6D,$7A,$A7,$44,$9D,$36,$35,$18,$23,$82,$57,$B2,$BA,$B6,$B9
	.dc.b	$36,$B8,$B6,$B8,$BE,$FA,$E2,$DA,$E2,$DA,$E2,$DA,$E2,$FC,$4A,$E2
	.dc.b	$DA,$E2,$FF,$AE,$CF,$E7,$1E,$AB,$D3,$7A,$BF,$CE,$E7,$FD,$84,$D9
	.dc.b	$D3,$9F,$A8,$FE,$6D,$3F,$D0,$5A,$FF,$AF,$9A,$FF,$B6,$5A,$FF,$BA
	.dc.b	$1A,$9A,$4E,$7F,$F3,$5C,$75,$6B,$D5,$AF,$56,$BD,$5A,$F5,$6C,$75
	.dc.b	$CF,$D5,$18,$23,$82,$4F,$85,$AF,$85,$AF,$85,$AF,$85,$AF,$85,$B3
	.dc.b	$E1,$8E,$84,$3A,$14,$E8,$53,$A1,$4E,$85,$7A,$3C,$88,$72,$29,$C8
	.dc.b	$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70
	.dc.b	$49,$FA,$71,$ED,$6B,$E6,$6B,$E1,$6B,$D9,$AF,$56,$C7,$59,$FA,$71
	.dc.b	$82,$38,$24,$F8,$5A,$F8,$5A,$F8,$5A,$F8,$5A,$F8,$5B,$3E,$18,$FB
	.dc.b	$5B,$7F,$B7,$EF,$FA,$E6,$7F,$B2,$9F,$F7,$24,$60,$30,$FF,$67,$FA
	.dc.b	$8F,$B3,$FF,$68,$FB,$19,$9C,$C6,$1E,$19,$2B,$82,$AE,$0E,$6B,$85
	.dc.b	$77,$E3,$FE,$F7,$5F,$DD,$89,$1D,$E3,$2F,$6B,$7D,$A7,$28,$F3,$B7
	.dc.b	$0F,$13,$67,$71,$E3,$9E,$5F,$7D,$FE,$EC,$FC,$6F,$CB,$48,$C4,$EA
	.dc.b	$95,$A9,$CC,$FF,$91,$E9,$F0,$18,$8F,$92,$DB,$F8,$8D,$9D,$C7,$49
	.dc.b	$F6,$7D,$D7,$FB,$FA,$BC,$6D,$8C,$CF,$66,$CF,$A3,$EE,$F5,$1E,$36
	.dc.b	$BF,$69,$B3,$F5,$5B,$7F,$CD,$61,$F2,$FF,$0B,$95,$18,$BF,$A1,$18
	.dc.b	$9D,$4E,$67,$ED,$75,$65,$FB,$0D,$73,$1B,$7F,$1F,$B7,$F8,$EA,$BE
	.dc.b	$8F,$66,$E7,$F2,$2D,$AF,$C6,$37,$1F,$F3,$14,$73,$3F,$21,$F4,$5F
	.dc.b	$7C,$8F,$DF,$AF,$AD,$1F,$83,$F8,$79,$39,$9F,$CF,$C6,$38,$BF,$07
	.dc.b	$EC,$FC,$2F,$C9,$2D,$9A,$4B,$FC,$A7,$C4,$ED,$B3,$B6,$CF,$E3,$A7
	.dc.b	$51,$F8,$9E,$6F,$D7,$BF,$9D,$6B,$9A,$FD,$63,$96,$BF,$B7,$E9,$78
	.dc.b	$1F,$E5,$8C,$BF,$7E,$49,$FD,$C4,$EB,$AB,$E0,$7D,$4F,$E6,$3C,$BF
	.dc.b	$D5,$37,$1C,$F3,$A7,$2D,$8F,$90,$7E,$FF,$A3,$F0,$FF,$25,$F9,$88
	.dc.b	$34,$99,$7E,$E7,$A7,$6D,$9D,$B6,$7F,$39,$18,$8F,$B2,$FC,$07,$F8
	.dc.b	$AF,$1F,$C6,$6C,$64,$FC,$19,$79,$1C,$9B,$8F,$E2,$CD,$8C,$C6,$E3
	.dc.b	$1F,$19,$FA,$2F,$2B,$F8,$2B,$6F,$F8,$93,$5F,$E0,$4D,$7F,$95,$F0
	.dc.b	$7C,$FC,$3F,$98,$92,$3E,$9F,$17,$FA,$3A,$3F,$96,$70,$7D,$16,$EF
	.dc.b	$F3,$DB,$C1,$EF,$9D,$5F,$F2,$3D,$7F,$D9,$6D,$D7,$FD,$8C,$DE,$7F
	.dc.b	$E3,$B3,$A8,$FE,$7F,$5F,$7F,$5E,$B7,$FE,$E8,$9F,$E8,$53,$96,$E5
	.dc.b	$F8,$1A,$FE,$4F,$F5,$DF,$C1,$BE,$72,$E0,$69,$2F,$8E,$7F,$C3,$3E
	.dc.b	$C3,$B9,$23,$34,$8F,$EA,$F1,$0F,$8F,$CB,$FE,$49,$E0,$FC,$D4,$7F
	.dc.b	$38,$F9,$1F,$E7,$C4,$BF,$CA,$F0,$23,$FA,$17,$C9,$7F,$8E,$A7,$FA
	.dc.b	$B3,$64,$D3,$FC,$7B,$7D,$AF,$C2,$78,$CF,$D8,$71,$7F,$B1,$9C,$1F
	.dc.b	$80,$70,$F0,$21,$FD,$1F,$E5,$3C,$04,$C3,$FB,$64,$7F,$70,$8F,$0C
	.dc.b	$F2,$8E,$F9,$7C,$18,$C4,$77,$8E,$35,$FD,$C7,$CD,$FE,$C0,$47,$E1
	.dc.b	$BF,$D9,$7A,$BF,$BF,$C3,$D9,$FE,$EB,$18,$6D,$FF,$42,$6C,$EE,$3F
	.dc.b	$BF,$CF,$FB,$9A,$1E,$97,$2C,$3F,$A4,$B6,$33,$5C,$27,$2F,$C4,$3D
	.dc.b	$47,$CA,$19,$FD,$17,$CF,$7F,$C3,$2B,$F8,$8B,$4F,$4F,$57,$B4,$F7
	.dc.b	$7C,$B7,$F5,$AF,$82,$F6,$B6,$FD,$2D,$7A,$B5,$F1,$9E,$53,$C7,$AF
	.dc.b	$D0,$6E,$5F,$43,$07,$CB,$FC,$D2,$65,$E6,$7E,$87,$CD,$7F,$6C,$FB
	.dc.b	$EF,$F4,$7C,$FF,$89,$63,$0F,$CC,$6E,$1F,$01,$AF,$90,$FC,$19,$F1
	.dc.b	$F5,$F8,$CF,$F8,$DB,$E7,$BF,$BC,$FD,$17,$6C,$91,$8C,$3E,$CB,$6B
	.dc.b	$E6,$B5,$7C,$93,$EA,$5E,$17,$FA,$A2,$FF,$C8,$D3,$98,$C5,$7F,$92
	.dc.b	$A3,$13,$F8,$E8,$C5,$72,$8C,$47,$F9,$33,$CA,$FF,$26,$C7,$E3,$1B
	.dc.b	$1F,$80,$6B,$B8,$FF,$26,$1F,$57,$3F,$83,$1F,$50,$FD,$9D,$F8,$9F
	.dc.b	$F5,$EB,$67,$D3,$7E,$39,$CB,$63,$A7,$07,$DF,$BE,$A4,$70,$4E,$0F
	.dc.b	$DA,$1F,$ED,$BF,$37,$FC,$DD,$DA,$F5,$BF,$CD,$9E,$AF,$FB,$7B,$3F
	.dc.b	$ED,$91,$8A,$DB,$7D,$3F,$F3,$9D,$F1,$AF,$F7,$63,$5F,$F7,$49,$C8
	.dc.b	$CC,$F1,$70,$7E,$59,$FF,$40,$91,$FE,$56,$D7,$F8,$6B,$5F,$E1,$8F
	.dc.b	$F9,$FF,$5F,$B6,$FC,$6D,$FF,$C8,$AD,$D7,$DC,$7C,$C8,$CB,$DA,$A1
	.dc.b	$D8,$A7,$DB,$F9,$2F,$CB,$70,$7F,$94,$1B,$26,$9D,$EE,$47,$73,$C3
	.dc.b	$5F,$8E,$BF,$0B,$F9,$C8,$7C,$8F,$E8,$F8,$BF,$CB,$AD,$8C,$9F,$A4
	.dc.b	$2C,$F9,$C8,$7C,$43,$D2,$FE,$E4,$7D,$F7,$EB,$F9,$BF,$CD,$2D,$8C
	.dc.b	$9F,$BD,$2C,$FF,$1C,$E8,$F8,$0F,$F2,$C6,$B9,$8F,$84,$E5,$AF,$90
	.dc.b	$FB,$C7,$CC,$29,$FC,$71,$7D,$F7,$F2,$26,$C7,$F2,$26,$BB,$8F,$E5
	.dc.b	$71,$DA,$3F,$D8,$3D,$9E,$46,$C6,$4F,$F5,$F4,$66,$3C,$27,$C7,$3C
	.dc.b	$DC,$DF,$E5,$1D,$7F,$A2,$7E,$5B,$27,$FA,$25,$A3,$11,$FD,$00,$FF
	.dc.b	$49,$F8,$5F,$D4,$88,$FF,$52,$32,$7F,$B1,$7C,$1F,$35,$7F,$63,$3C
	.dc.b	$DE,$1A,$32,$D8,$F7,$4F,$A0,$98,$F9,$DF,$BF,$2F,$0F,$F4,$31,$FE
	.dc.b	$D3,$6B,$FD,$F6,$BF,$6C,$FF,$7B,$6B,$D3,$AD,$D7,$8C,$3B,$4F,$C6
	.dc.b	$9C,$E9,$FB,$38,$E1,$5D,$2F,$FC,$32,$5C,$7F,$C3,$2D,$CF,$C7,$5F
	.dc.b	$EF,$A6,$C7,$FB,$D7,$C4,$F8,$24,$7E,$2B,$F6,$72,$F1,$D7,$D9,$21
	.dc.b	$F7,$3C,$AF,$F4,$37,$DE,$FF,$92,$09,$1F,$E3,$A7,$FC,$87,$F7,$B2
	.dc.b	$3F,$DF,$1E,$57,$FB,$32,$31,$1F,$E5,$2F,$33,$FD,$19,$F7,$BF,$E5
	.dc.b	$22,$47,$F9,$11,$FF,$2C,$E6,$F8,$2A,$49,$FF,$30,$3F,$E5,$AF,$BD
	.dc.b	$91,$FE,$4C,$6B,$FE,$2A,$7F,$C5,$FA,$7F,$A8,$DB,$C5,$F5,$1B,$C2
	.dc.b	$7F,$D0,$38,$31,$A4,$C3,$FD,$07,$19,$47,$F8,$BB,$87,$FD,$1B,$58
	.dc.b	$8C,$7A,$5F,$EE,$E8,$C4,$7F,$B4,$AB,$C0,$ED,$FF,$4B,$98,$7F,$D4
	.dc.b	$71,$F1,$08,$FF,$A0,$5F,$F5,$4F,$C7,$C8,$FF,$57,$35,$F2,$7A,$BF
	.dc.b	$EF,$E8,$FF,$5C,$12,$3C,$18,$27,$51,$FE,$76,$F0,$7F,$D0,$E4,$70
	.dc.b	$7F,$9D,$9F,$92,$47,$07,$FB,$66,$FD,$F8,$2D,$EB,$7C,$3A,$D1,$0F
	.dc.b	$95,$EB,$A9,$F9,$17,$FB,$AD,$B3,$FE,$F0,$8C,$1F,$EF,$82,$3E,$CF
	.dc.b	$6B,$94,$62,$3E,$9C,$FF,$06,$D7,$F8,$29,$1F,$E1,$3E,$BF,$F0,$82
	.dc.b	$38,$3F,$86,$FB,$1E,$5F,$97,$F6,$E8,$8C,$61,$F3,$1A,$FE,$15,$AF
	.dc.b	$E2,$3D,$C7,$BB,$88,$46,$30,$61,$FE,$88,$8C,$36,$B9,$CF,$CD,$F3
	.dc.b	$EE,$26,$92,$39,$E1,$FE,$F3,$F8,$7F,$F7,$89,$1C,$1F,$84,$E3,$FA
	.dc.b	$51,$FE,$FF,$24,$66,$51,$F0,$BA,$BE,$C7,$E6,$36,$4D,$25,$7E,$43
	.dc.b	$E0,$36,$D9,$DB,$67,$E6,$CE,$A3,$F5,$DF,$05,$FC,$F0,$95,$FD,$01
	.dc.b	$FE,$A9,$B7,$C9,$E7,$F8,$D3,$FD,$02,$31,$5F,$B2,$6B,$FB,$9E,$6F
	.dc.b	$E8,$E4,$9F,$E9,$6F,$F2,$7E,$BF,$E7,$7F,$0D,$FC,$20,$95,$FE,$D8
	.dc.b	$7F,$D7,$BF,$08,$F8,$5F,$E6,$B2,$8C,$61,$FD,$19,$BE,$3F,$0A,$DC
	.dc.b	$62,$BF,$83,$B5,$FE,$18,$D7,$F8,$8B,$5E,$D1,$C8,$95,$DD,$B5,$DE
	.dc.b	$30,$47,$04,$9F,$0C,$F0,$6C,$ED,$B3,$B7,$CB,$3B,$6C,$ED,$B3,$B7
	.dc.b	$D3,$3B,$6C,$ED,$B3,$B7,$DB,$19,$6C,$66,$BE,$11,$2B,$A4,$E8,$F3
	.dc.b	$90,$F1,$94,$EE,$53,$81,$A2,$1A,$29,$A3,$04,$30,$48,$F2,$57,$A5
	.dc.b	$B5,$E9,$AF,$3C,$62,$BC,$4D,$7B,$B5,$EC,$D7,$B4,$72,$25,$77,$6D
	.dc.b	$77,$8C,$11,$C1,$27,$C3,$3C,$1B,$3B,$6C,$ED,$F2,$CE,$DB,$3B,$6C
	.dc.b	$ED,$F4,$CE,$DF,$ED,$7E,$BF,$FA,$A5,$FE,$CE,$D7,$FD,$40,$63,$8D
	.dc.b	$EB,$FF,$6C,$8F,$EE,$7F,$44,$EA,$FE,$95,$19,$8C,$65,$B8,$46,$1D
	.dc.b	$4F,$07,$FB,$BC,$65,$B1,$97,$FB,$F4,$65,$B1,$97,$C6,$D7,$BB,$63
	.dc.b	$84,$6A,$7C,$91,$88,$DC,$62,$78,$11,$C1,$23,$D2,$D7,$CE,$D7,$CA
	.dc.b	$D7,$C0,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$1E
	.dc.b	$E3,$A9,$4E,$45,$38,$94,$E3,$3F,$1C,$93,$F1,$63,$9C,$62,$7E,$04
	.dc.b	$71,$24,$65,$DC,$F1,$7C,$31,$96,$C6,$5F,$1C,$65,$B1,$97,$CA,$D7
	.dc.b	$BB,$63,$84,$6A,$7C,$B1,$88,$DC,$62,$78,$11,$C1,$23,$D2,$D7,$CE
	.dc.b	$D7,$CA,$D7,$C0,$70,$21,$C0,$A7,$02,$9C,$0D,$1F,$E5,$CF,$0D,$B7
	.dc.b	$FA,$83,$0F,$F3,$69,$27,$51,$97,$FC,$C7,$F9,$D3,$E7,$3F,$C8,$15
	.dc.b	$FB,$B8,$C6,$1F,$BE,$AD,$D6,$9B,$19,$7F,$CE,$1F,$37,$FE,$4E,$8C
	.dc.b	$47,$F9,$DA,$31,$7F,$C1,$DB,$1F,$E7,$46,$BC,$BE,$67,$B3,$6B,$9A
	.dc.b	$94,$EA,$3B,$35,$D7,$CF,$59,$34,$95,$D2,$1F,$3D,$B2,$46,$5E,$9F
	.dc.b	$A0,$BD,$AA,$49,$F2,$C6,$5B,$9F,$02,$38,$37,$F3,$FD,$08,$E0,$E7
	.dc.b	$E1,$F8,$08,$E0,$EF,$8F,$90,$8E,$0F,$1F,$CF,$7A,$08,$E0,$F3,$FC
	.dc.b	$F7,$B0,$8E,$0F,$5F,$CF,$7C,$02,$38,$3E,$9F,$1F,$E4,$3C,$F7,$DC
	.dc.b	$7F,$AF,$09,$C3,$F9,$08,$CB,$F8,$88,$FC,$76,$BF,$E8,$F6,$DE,$A7
	.dc.b	$FD,$75,$C9,$F9,$29,$FC,$47,$76,$E7,$F1,$0D,$7F,$22,$D7,$F2,$0D
	.dc.b	$7F,$11,$C5,$76,$FB,$F8,$77,$9F,$CB,$4E,$4E,$1C,$3F,$6F,$0F,$EB
	.dc.b	$CD,$9F,$F6,$5B,$5F,$F6,$83,$5F,$F6,$AF,$03,$D7,$FF,$56,$7A,$DF
	.dc.b	$D9,$E3,$18,$A3,$13,$AA,$FF,$6B,$D6,$A7,$E0,$D6,$E7,$E0,$B5,$F8
	.dc.b	$2D,$7E,$0B,$56,$FF,$5F,$2F,$F7,$89,$23,$EB,$3F,$48,$8F,$EF,$9A
	.dc.b	$F6,$7C,$64,$7C,$6F,$AA,$5F,$71,$49,$3E,$CC,$7F,$D9,$D1,$EC,$24
	.dc.b	$7B,$1A,$FF,$BE,$F1,$F8,$FC,$1D,$49,$1C,$ED,$1F,$18,$91,$F1,$9A
	.dc.b	$89,$8F,$E6,$5F,$F0,$D3,$5F,$CC,$18,$E9,$EA,$F1,$39,$74,$FD,$9E
	.dc.b	$7F,$E2,$4D,$C3,$F9,$0B,$4E,$13,$FE,$28,$D7,$F8,$34,$7F,$BD,$BE
	.dc.b	$1D,$F5,$DF,$EF,$17,$0D,$AF,$05,$A7,$6D,$8C,$C6,$0E,$9F,$5F,$FE
	.dc.b	$23,$9E,$0D,$83,$4F,$F1,$36,$7F,$06,$BD,$CD,$7E,$3E,$7F,$09,$F8
	.dc.b	$A7,$83,$E2,$BA,$13,$0F,$F2,$21,$E7,$7A,$CF,$19,$EA,$D7,$AE,$BF
	.dc.b	$E4,$5B,$F0,$66,$F8,$8F,$3E,$3F,$31,$B1,$F3,$0D,$65,$F4,$A3,$38
	.dc.b	$7C,$4F,$96,$5B,$F2,$3F,$A1,$18,$6C,$66,$78,$35,$FC,$43,$F4,$7E
	.dc.b	$CB,$E5,$11,$F7,$EC,$8F,$7E,$9E,$BD,$7D,$D7,$FB,$26,$EB,$F3,$70
	.dc.b	$FB,$9F,$67,$F5,$48,$F4,$79,$D7,$E4,$EF,$A1,$1C,$10,$FA,$A7,$CE
	.dc.b	$7F,$CE,$1F,$20,$FB,$7F,$EC,$CD,$BF,$ED,$AD,$E0,$FF,$3E,$36,$F9
	.dc.b	$9F,$58,$EC,$75,$29,$93,$67,$BE,$FF,$07,$EA,$7D,$AF,$8F,$FF,$3F
	.dc.b	$60,$8C,$3F,$9E,$6D,$7D,$06,$BF,$6B,$E6,$1F,$1B,$FE,$8E,$AF,$D2
	.dc.b	$C6,$1F,$AF,$3F,$62,$30,$F8,$1B,$1F,$A8,$FB,$5F,$D8,$34,$FD,$71
	.dc.b	$78,$7C,$E2,$37,$5F,$04,$F7,$EF,$FA,$77,$E0,$1F,$C0,$29,$7E,$08
	.dc.b	$F2,$61,$E9,$6C,$FA,$1F,$74,$65,$F9,$87,$72,$18,$21,$E0,$3C,$7C
	.dc.b	$1F,$90,$F3,$5E,$6F,$F8,$EF,$57,$9F,$FC,$6E,$73,$C3,$FC,$96,$73
	.dc.b	$87,$FB,$12,$F8,$5E,$DB,$19,$AE,$13,$96,$C6,$45,$74,$57,$E0,$73
	.dc.b	$5F,$97,$3F,$3C,$47,$FC,$14,$D7,$46,$08,$7C,$A7,$EF,$63,$FC,$01
	.dc.b	$B9,$7C,$03,$F0,$F1,$97,$8C,$FF,$12,$9E,$CD,$3F,$09,$E9,$FE,$0B
	.dc.b	$6F,$E1,$6D,$A3,$4E,$F5,$23,$B9,$E1,$AF,$33,$F9,$FE,$9F,$E2,$BE
	.dc.b	$F7,$E9,$38,$BF,$8F,$B6,$33,$3B,$11,$FE,$E3,$E1,$FF,$19,$61,$E8
	.dc.b	$C3,$F6,$0D,$8C,$C7,$B8,$C9,$79,$3F,$81,$7A,$7F,$C0,$39,$BF,$93
	.dc.b	$B7,$13,$4E,$9C,$7F,$DC,$31,$BC,$DF,$77,$1F,$90,$7D,$62,$18,$3F
	.dc.b	$AF,$73,$FD,$57,$FA,$CB,$5F,$E5,$FC,$E8,$C4,$70,$C7,$B3,$EA,$3A
	.dc.b	$94,$EC,$FF,$37,$3C,$85,$BF,$EE,$1D,$1D,$23,$13,$FD,$03,$F2,$3F
	.dc.b	$E8,$64,$70,$7F,$80,$79,$FF,$A7,$CF,$C4,$8C,$76,$7F,$56,$D7,$FA
	.dc.b	$A3,$63,$3C,$5F,$31,$FA,$5C,$3F,$51,$F4,$B5,$FF,$01,$76,$F9,$76
	.dc.b	$34,$98,$FC,$2A,$E1,$5F,$04,$8F,$38,$FF,$00,$9F,$0A,$BF,$B5,$EB
	.dc.b	$FE,$06,$8C,$61,$DF,$BB,$F3,$B8,$7F,$83,$1A,$FD,$46,$BF,$E1,$76
	.dc.b	$E9,$FE,$21,$D3,$FB,$A7,$7F,$FA,$8D,$BE,$07,$C1,$BF,$F1,$4C,$E5
	.dc.b	$F8,$7C,$7F,$8B,$6F,$1F,$DE,$F1,$FF,$D9,$9F,$41,$FE,$62,$24,$FF
	.dc.b	$98,$9F,$F2,$81,$1F,$F2,$23,$5F,$F3,$2F,$DA,$FF,$9A,$88,$FF,$9A
	.dc.b	$9F,$F2,$C1,$1F,$F2,$63,$5F,$F3,$6F,$87,$FE,$72,$23,$FE,$72,$7F
	.dc.b	$D3,$24,$7F,$DC,$AD,$7F,$B9,$3F,$DE,$48,$FF,$A9,$5F,$F3,$EF,$DA
	.dc.b	$48,$FF,$3E,$B5,$FF,$31,$78,$FF,$E1,$A8,$FF,$37,$12,$3F,$CC,$4F
	.dc.b	$FA,$5B,$4F,$F0,$FF,$DA,$5E,$23,$49,$5F,$EF,$1F,$27,$6D,$9D,$B6
	.dc.b	$7F,$DF,$13,$A8,$FF,$47,$C7,$FA,$50,$8F,$FA,$51,$FF,$5D,$91,$FF
	.dc.b	$30,$F9,$AF,$3B,$94,$7F,$9C,$E3,$17,$E0,$8F,$F0,$C5,$7F,$87,$C8
	.dc.b	$FF,$B2,$DF,$F5,$9F,$9F,$F8,$B8,$FF,$3E,$79,$A8,$C4,$23,$15,$FB
	.dc.b	$78,$FF,$5B,$CE,$A5,$E8,$52,$56,$EB,$AC,$6F,$D1,$93,$99,$ED,$18
	.dc.b	$E6,$FF,$63,$F1,$F5,$25,$6F,$04,$78,$7C,$E9,$39,$6D,$7F,$B4,$3D
	.dc.b	$2F,$31,$27,$93,$E9,$DA,$47,$56,$CF,$E0,$49,$97,$FB,$63,$83,$E0
	.dc.b	$1F,$1B,$1F,$F7,$1C,$FD,$EB,$F6,$46,$2B,$F0,$93,$FE,$9A,$7D,$3E
	.dc.b	$9F,$12,$57,$07,$A7,$A9,$23,$9B,$6B,$FD,$B6,$4A,$FF,$6D,$B5,$FF
	.dc.b	$80,$38,$FD,$AE,$1F,$C2,$CF,$85,$B3,$F8,$C9,$FB,$FE,$03,$AB,$FD
	.dc.b	$87,$7F,$E6,$72,$57,$FA,$CA,$75,$D5,$F6,$7D,$67,$AD,$FD,$3A,$91
	.dc.b	$8D,$3F,$A7,$B7,$C1,$F8,$4D,$9F,$F5,$EB,$5F,$F6,$33,$5F,$F6,$83
	.dc.b	$5E,$B5,$DE,$30,$47,$04,$AF,$0B,$6B,$C2,$DA,$F0,$B5,$F0,$B5,$F0
	.dc.b	$B5,$F0,$CF,$42,$4F,$46,$C7,$46,$BD,$1A,$F4,$6B,$D2,$39,$11,$E4
	.dc.b	$D7,$93,$5E,$4D,$79,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46
	.dc.b	$08,$60,$95,$F4,$E7,$DA,$D9,$F3,$36,$7C,$2D,$7B,$35,$EA,$D7,$AD
	.dc.b	$7D,$38,$C1,$1C,$12,$BC,$2D,$AF,$0B,$6B,$C2,$D7,$C2,$D7,$C2,$D7
	.dc.b	$C3,$3D,$09,$3D,$1B,$1D,$1A,$F4,$6B,$D1,$AF,$C9,$F3,$FF,$CF,$88
	.dc.b	$FF,$42,$1F,$C2,$5C,$B8,$3F,$53,$F0,$5F,$EF,$67,$F5,$AD,$73,$3F
	.dc.b	$C4,$5F,$DA,$C6,$5F,$D7,$B5,$FD,$7B,$5D,$C7,$1C,$13,$96,$C6,$63
	.dc.b	$04,$70,$49,$F2,$B5,$F2,$98,$21,$82,$1E,$67,$83,$4D,$94,$D9,$EA
	.dc.b	$76,$D3,$65,$36,$7B,$8C,$94,$CC,$FB,$9C,$B5,$CC,$FB,$DC,$B5,$CC
	.dc.b	$F8,$9E,$87,$9C,$87,$8C,$A6,$8C,$10,$C1,$27,$C4,$F0,$6B,$B6,$BB
	.dc.b	$9F,$23,$B6,$BB,$6B,$B9,$F4,$4F,$22,$4E,$9B,$1A,$8C,$11,$C1,$27
	.dc.b	$E2,$B5,$F8,$A6,$08,$60,$87,$9D,$E0,$D3,$65,$36,$7A,$DD,$B4,$D9
	.dc.b	$4D,$9E,$F3,$25,$3F,$B0,$68,$C9,$E0,$FF,$7B,$D3,$FB,$1B,$67,$4D
	.dc.b	$9D,$D7,$F8,$EB,$8D,$86,$23,$F7,$C3,$81,$3A,$BF,$EF,$0D,$9E,$55
	.dc.b	$C5,$FE,$0D,$F9,$EB,$C9,$D4,$74,$8C,$4E,$E1,$18,$BF,$F0,$A3,$63
	.dc.b	$C6,$6B,$C0,$E6,$F6,$6B,$9A,$E9,$29,$CB,$FE,$B0,$8C,$F8,$F8,$23
	.dc.b	$C3,$C6,$D9,$1C,$1D,$3C,$6E,$64,$70,$7A,$FE,$6D,$E3,$79,$1B,$19
	.dc.b	$8C,$1E,$8E,$EF,$F3,$3E,$9F,$E3,$36,$E9,$F0,$1A,$FD,$46,$BF,$CC
	.dc.b	$ED,$F3,$9D,$C9,$87,$8D,$F8,$3F,$33,$23,$E2,$B7,$0F,$9F,$F3,$1F
	.dc.b	$18,$93,$F2,$9F,$57,$BB,$23,$DC,$D9,$FF,$25,$7E,$46,$FA,$8F,$A6
	.dc.b	$DA,$FB,$CD,$9D,$CF,$D8,$E8,$FF,$1C,$FB,$DF,$95,$26,$BF,$9E,$E8
	.dc.b	$47,$E4,$31,$B3,$F9,$08,$C1,$1F,$C3,$F2,$7D,$F8,$FC,$5F,$27,$E7
	.dc.b	$9B,$5C,$AB,$F2,$93,$97,$8F,$25,$7F,$30,$D7,$F3,$AD,$7F,$32,$DE
	.dc.b	$1F,$A7,$5D,$67,$EB,$70,$E4,$91,$98,$FD,$77,$3F,$EC,$BC,$EF,$71
	.dc.b	$23,$F5,$AF,$E8,$BD,$06,$CF,$8F,$E1,$5F,$B8,$9C,$C6,$23,$FB,$C4
	.dc.b	$63,$05,$F3,$BE,$51,$8A,$FA,$1E,$6F,$F0,$72,$38,$3F,$7E,$FD,$32
	.dc.b	$38,$3F,$6F,$F9,$3B,$C6,$FE,$1B,$8F,$FA,$3E,$31,$68,$C4,$EA,$B8
	.dc.b	$B5,$FE,$27,$DD,$FC,$4A,$9F,$01,$F8,$32,$4F,$F1,$27,$F8,$3C,$3D
	.dc.b	$37,$DD,$FF,$1E,$FB,$5C,$C6,$25,$18,$9D,$5F,$F2,$29,$E5,$18,$8F
	.dc.b	$1F,$07,$F2,$8E,$7F,$CA,$12,$3F,$91,$BF,$CB,$D8,$7E,$6C,$A7,$F2
	.dc.b	$EF,$D3,$5E,$DF,$F8,$75,$B8,$7F,$89,$5B,$3F,$EB,$76,$BE,$5C,$9E
	.dc.b	$D7,$F8,$9E,$BE,$03,$63,$31,$FE,$BF,$32,$5F,$A4,$7D,$47,$16,$C6
	.dc.b	$63,$07,$F8,$C7,$93,$FD,$5F,$87,$A8,$FF,$22,$38,$7D,$C7,$97,$E3
	.dc.b	$7C,$44,$AF,$16,$08,$E9,$E4,$FF,$5B,$AE,$87,$E7,$9C,$3D,$0F,$EA
	.dc.b	$BE,$F3,$C8,$FF,$22,$37,$0F,$F2,$63,$67,$FD,$BC,$D7,$E4,$65,$FD
	.dc.b	$83,$F1,$4F,$C1,$B6,$33,$49,$CB,$63,$F0,$9F,$8E,$BE,$27,$E8,$36
	.dc.b	$7E,$73,$67,$F2,$6D,$7D,$47,$C8,$CB,$DA,$FF,$70,$D7,$E5,$36,$32
	.dc.b	$FF,$76,$D5,$F2,$3F,$74,$98,$7C,$C7,$F0,$1F,$57,$FD,$F3,$EF,$FF
	.dc.b	$3E,$DB,$FC,$43,$5F,$F3,$E5,$7E,$9E,$31,$1E,$5E,$5F,$F3,$B6,$5E
	.dc.b	$48,$C1,$F5,$B7,$FB,$57,$F4,$38,$7E,$99,$1F,$BF,$E0,$7F,$83,$BF
	.dc.b	$00,$FF,$41,$47,$ED,$63,$0F,$E6,$DB,$7E,$4D,$7F,$8A,$D7,$FA,$17
	.dc.b	$1F,$14,$FE,$17,$2F,$2E,$DF,$1A,$30,$7F,$02,$FC,$F5,$F2,$9F,$E1
	.dc.b	$EF,$6F,$C3,$18,$D1,$18,$9D,$47,$F0,$19,$D4,$FA,$3F,$3A,$7B,$FF
	.dc.b	$E9,$DA,$FE,$0D,$18,$7E,$5E,$1F,$32,$30,$F1,$9E,$0F,$D0,$F6,$BF
	.dc.b	$C6,$35,$FC,$5C,$FE,$20,$E3,$6F,$8E,$D8,$C9,$F8,$33,$F1,$73,$F0
	.dc.b	$0B,$AF,$D7,$86,$1E,$27,$CF,$87,$8A,$BF,$C8,$33,$97,$D1,$19,$7A
	.dc.b	$1F,$93,$D7,$E7,$91,$FD,$BD,$FD,$42,$6B,$F4,$FD,$A7,$C7,$7F,$94
	.dc.b	$3E,$39,$A7,$FB,$29,$B8,$7F,$B3,$6B,$7E,$97,$FB,$3E,$73,$29,$CA
	.dc.b	$8C,$C6,$1B,$7D,$EF,$B1,$2B,$F3,$8F,$F9,$8C,$8F,$F8,$3D,$AF,$BD
	.dc.b	$AF,$03,$E4,$EB,$EA,$2C,$7D,$D7,$11,$F2,$9A,$F1,$3A,$1D,$C8,$7E
	.dc.b	$0D,$FF,$39,$FC,$67,$E0,$1A,$7E,$00,$39,$69,$E2,$24,$78,$67,$C8
	.dc.b	$7E,$76,$3F,$CE,$31,$FC,$03,$87,$F4,$5E,$DE,$49,$3E,$49,$F4,$10
	.dc.b	$FD,$85,$7F,$6A,$E5,$FF,$44,$FE,$03,$FB,$CE,$3F,$F0,$5B,$63,$37
	.dc.b	$FB,$D1,$1F,$EF,$FE,$8F,$E1,$37,$FC,$DB,$A3,$FD,$26,$DA,$FC,$0E
	.dc.b	$21,$5C,$1F,$31,$E4,$2F,$13,$8B,$FD,$4D,$7F,$C8,$AD,$F8,$15,$25
	.dc.b	$FF,$24,$8C,$BF,$DB,$7B,$3C,$6D,$8F,$19,$FE,$0D,$8C,$C7,$F1,$76
	.dc.b	$BF,$1F,$0F,$17,$2F,$AA,$3C,$C7,$E0,$8F,$E6,$3D,$BF,$D2,$71,$FE
	.dc.b	$8F,$DD,$CB,$14,$E6,$BF,$A8,$CE,$78,$3F,$BB,$FE,$2B,$FD,$58,$93
	.dc.b	$FD,$69,$FE,$C3,$E1,$79,$23,$E4,$B7,$7F,$ED,$19,$6B,$2F,$36,$9E
	.dc.b	$88,$FD,$0D,$FA,$0F,$C1,$CF,$F8,$B7,$83,$C4,$47,$E2,$E5,$F5,$67
	.dc.b	$DC,$7C,$3E,$0F,$A6,$FE,$36,$5E,$1F,$F8,$43,$0F,$F1,$CB,$67,$FC
	.dc.b	$64,$D7,$FA,$BB,$5F,$EC,$9E,$27,$C3,$25,$78,$9F,$81,$E2,$FF,$8B
	.dc.b	$2F,$FC,$55,$E3,$60,$95,$F0,$31,$F2,$E1,$E8,$3E,$01,$3C,$BF,$F2
	.dc.b	$97,$BB,$FE,$93,$24,$7F,$B8,$9F,$F2,$55,$3F,$55,$A4,$9F,$F1,$C7
	.dc.b	$97,$FE,$F0,$8F,$F4,$E7,$C6,$C9,$FF,$75,$B5,$FF,$79,$FA,$3F,$E6
	.dc.b	$8F,$C1,$3F,$CC,$F8,$1A,$49,$FF,$79,$FC,$0F,$FA,$9C,$8E,$0F,$F5
	.dc.b	$63,$FE,$EC,$7F,$B5,$37,$93,$F2,$ED,$D5,$D5,$FE,$EB,$AF,$F4,$75
	.dc.b	$7F,$A2,$3E,$37,$FD,$15,$DE,$C9,$18,$6D,$70,$9D,$4F,$FA,$EC,$91
	.dc.b	$FE,$FE,$E0,$FF,$3F,$46,$23,$F8,$00,$F6,$3F,$D5,$BF,$1B,$E9,$6C
	.dc.b	$68,$D7,$B3,$E3,$9D,$36,$3F,$D7,$B8,$A7,$E1,$7B,$AF,$6F,$55,$FE
	.dc.b	$C0,$AD,$3C,$2B,$FD,$92,$47,$FD,$9F,$EE,$70,$7F,$8B,$5A,$31,$1F
	.dc.b	$EB,$EA,$E6,$FF,$A2,$9A,$FF,$1D,$CF,$FD,$3D,$7F,$ED,$F3,$E9,$6D
	.dc.b	$A9,$FF,$10,$46,$08,$FC,$1F,$7F,$FD,$E1,$F0,$FF,$EE,$EC,$0D,$24
	.dc.b	$FC,$7D,$5F,$03,$F6,$67,$E6,$36,$74,$D7,$E9,$CF,$0A,$EB,$1F,$EF
	.dc.b	$0F,$82,$D3,$F0,$9A,$B5,$46,$1B,$3C,$A7,$EF,$C3,$E0,$F4,$49,$D4
	.dc.b	$FC,$E8,$EB,$7F,$EF,$0C,$7E,$FD,$6A,$BF,$06,$D9,$D4,$7E,$1F,$E1
	.dc.b	$1F,$11,$E6,$6D,$71,$6D,$FE,$9D,$B5,$F2,$7B,$7C,$24,$AE,$CF,$B7
	.dc.b	$C6,$91,$E3,$CF,$FD,$2D,$87,$E4,$27,$E1,$71,$CA,$D3,$67,$F2,$BC
	.dc.b	$7F,$1C,$93,$CD,$F9,$5D,$32,$3B,$36,$7F,$3B,$E6,$FC,$F2,$4F,$8D
	.dc.b	$FA,$BC,$BF,$EE,$D9,$FE,$08,$DD,$3F,$69,$EE,$E8,$FB,$7D,$5F,$D7
	.dc.b	$2B,$F2,$A4,$9F,$D9,$46,$5F,$DA,$7C,$3B,$E1,$FF,$A9,$37,$2F,$EC
	.dc.b	$8D,$F1,$3F,$8A,$36,$B8,$D7,$02,$57,$06,$D7,$06,$CF,$09,$D1,$27
	.dc.b	$4D,$8D,$46,$08,$E0,$87,$F1,$2A,$F2,$36,$BB,$B6,$BA,$36,$B9,$B4
	.dc.b	$E6,$7F,$86,$EB,$9B,$6B,$9B,$6B,$9B,$6B,$9B,$4E,$62,$B9,$B6,$B9
	.dc.b	$B6,$B9,$B6,$B9,$B4,$E7,$5C,$89,$5C,$9B,$5C,$9B,$5C,$9B,$5C,$AB
	.dc.b	$81,$2B,$83,$6B,$83,$67,$84,$E8,$93,$A6,$C6,$A3,$04,$70,$43,$E8
	.dc.b	$10,$F9,$05,$3E,$01,$6B,$D2,$DA,$F2,$36,$BC,$2D,$AF,$03,$4F,$01
	.dc.b	$D8,$87,$62,$D7,$66,$D7,$66,$D7,$66,$D7,$67,$FD,$4D,$E9,$BD,$6F
	.dc.b	$EF,$51,$FE,$A5,$8C,$3F,$DF,$7D,$6F,$F5,$3B,$FD,$95,$AF,$E6,$FC
	.dc.b	$7F,$EF,$0F,$EC,$1A,$FE,$E1,$AE,$F0,$EF,$92,$73,$3F,$9B,$6B,$F9
	.dc.b	$36,$C6,$E7,$8C,$60,$8E,$09,$3E,$36,$BE,$36,$BE,$33,$44,$34,$53
	.dc.b	$46,$08,$60,$87,$A5,$EA,$D3,$91,$4E,$25,$38,$9E,$F7,$8B,$4E,$25
	.dc.b	$38,$94,$E3,$3F,$15,$DB,$5D,$B5,$DC,$FC,$87,$6D,$76,$D7,$73,$E4
	.dc.b	$9E,$51,$E9,$21,$E3,$2B,$A7,$04,$70,$49,$ED,$1C,$1A,$ED,$B1,$B9
	.dc.b	$F1,$46,$08,$E0,$93,$EC,$6B,$EC,$6B,$EC,$34,$43,$45,$34,$60,$86
	.dc.b	$08,$7A,$5E,$AD,$39,$14,$E2,$53,$89,$EF,$78,$B4,$FF,$1C,$E1,$1F
	.dc.b	$EC,$BE,$17,$F9,$B2,$31,$1F,$E5,$EF,$D7,$5F,$73,$FE,$6B,$9F,$F3
	.dc.b	$B4,$62,$FF,$7B,$5B,$AD,$3F,$E6,$19,$D5,$7F,$80,$2B,$73,$FE,$54
	.dc.b	$E2,$DB,$FE,$2E,$6B,$98,$DA,$A7,$C4,$EA,$30,$D8,$CC,$F0,$6D,$FF
	.dc.b	$86,$23,$9F,$47,$F9,$8D,$BC,$DC,$E3,$13,$EE,$FB,$9E,$84,$AE,$8E
	.dc.b	$FA,$64,$74,$6C,$FF,$A0,$FA,$BC,$DD,$A6,$92,$BC,$DB,$FA,$48,$E0
	.dc.b	$F8,$32,$F1,$39,$92,$32,$FC,$1F,$6A,$47,$C1,$6E,$1F,$E8,$CF,$07
	.dc.b	$DC,$49,$F8,$6F,$C9,$F0,$A4,$7C,$96,$CF,$F9,$E4,$9C,$3F,$E7,$DE
	.dc.b	$1F,$A5,$E1,$FC,$E8,$FC,$5F,$8A,$F8,$AF,$F4,$39,$23,$2F,$E2,$AB
	.dc.b	$FD,$6C,$DC,$78,$4F,$E2,$67,$2D,$8C,$BF,$E7,$4E,$3C,$B6,$32,$FF
	.dc.b	$A8,$23,$2D,$8C,$BF,$8B,$DF,$2D,$8C,$F2,$27,$2D,$8E,$3D,$B9,$C9
	.dc.b	$F1,$5F,$EA,$42,$61,$F0,$9F,$F6,$15,$3B,$7F,$D2,$D1,$86,$CF,$FB
	.dc.b	$32,$75,$BF,$FB,$66,$B4,$D9,$FF,$66,$73,$B9,$B1,$1F,$E9,$66,$BF
	.dc.b	$EC,$06,$BF,$E9,$F6,$E5,$E9,$F8,$AF,$F5,$D9,$27,$EC,$BE,$5E,$4F
	.dc.b	$F7,$57,$C5,$5C,$CD,$25,$7E,$03,$E2,$7F,$D8,$A4,$70,$7B,$75,$F9
	.dc.b	$5F,$13,$F0,$78,$7F,$DE,$CD,$C0,$D3,$FD,$9F,$C1,$2F,$F1,$AD,$7F
	.dc.b	$DE,$7A,$7D,$37,$FD,$AA,$43,$F2,$65,$37,$69,$FF,$0B,$CF,$19,$D9
	.dc.b	$E7,$EB,$FF,$05,$E1,$F6,$5B,$89,$A6,$39,$12,$F5,$BE,$F4,$61,$B1
	.dc.b	$9B,$F8,$93,$97,$C2,$79,$B1,$F2,$E5,$E7,$8F,$50,$19,$7A,$5E,$06
	.dc.b	$8C,$C7,$17,$A9,$C6,$38,$36,$34,$F2,$36,$72,$9F,$19,$1F,$CD,$77
	.dc.b	$7C,$0F,$C0,$AE,$26,$92,$7F,$05,$D9,$F2,$89,$3E,$B8,$CB,$D5,$F0
	.dc.b	$E3,$E6,$7C,$47,$B5,$C1,$31,$F4,$C7,$D2,$D5,$F7,$9F,$8E,$8F,$C9
	.dc.b	$35,$F9,$57,$FB,$58,$C4,$71,$6B,$E0,$3C,$75,$D9,$C1,$F1,$C8,$7C
	.dc.b	$BC,$FD,$4F,$6B,$7D,$EF,$E1,$BF,$43,$FF,$2E,$E3,$FC,$16,$30,$FE
	.dc.b	$9D,$B5,$C5,$AF,$53,$E1,$61,$E3,$CB,$ED,$D3,$E6,$DF,$65,$F9,$48
	.dc.b	$EF,$18,$71,$7F,$0A,$31,$1A,$9E,$0F,$EE,$72,$FF,$3A,$B4,$F9,$35
	.dc.b	$ED,$8C,$B6,$32,$FD,$3F,$0F,$F4,$51,$FC,$0D,$B1,$F1,$5A,$EE,$3F
	.dc.b	$54,$7D,$0D,$3E,$5C,$7C,$13,$C4,$7C,$AA,$FE,$5D,$3E,$03,$F9,$83
	.dc.b	$E4,$8F,$C2,$35,$FB,$27,$33,$EC,$98,$26,$3F,$8D,$AF,$0C,$7F,$A7
	.dc.b	$7C,$1F,$D9,$74,$7F,$3D,$8F,$F5,$1C,$62,$FF,$39,$E3,$BC,$8F,$F5
	.dc.b	$1D,$EE,$F9,$07,$13,$BB,$E5,$3D,$0D,$5F,$4B,$E6,$D9,$DC,$FE,$DE
	.dc.b	$3F,$58,$FE,$4D,$AE,$4F,$EE,$EE,$31,$FE,$B8,$DD,$BF,$93,$37,$2F
	.dc.b	$45,$73,$3F,$05,$1F,$44,$FD,$86,$78,$24,$7F,$1B,$8E,$44,$7C,$27
	.dc.b	$E0,$E1,$F0,$DF,$6F,$5F,$E5,$6D,$8C,$9B,$7C,$0F,$E0,$27,$F2,$15
	.dc.b	$82,$47,$F6,$DD,$FF,$41,$F2,$3F,$9D,$83,$48,$F9,$2B,$D0,$53,$E0
	.dc.b	$1F,$B3,$5F,$A1,$7E,$47,$8B,$FA,$0B,$63,$27,$EE,$CB,$3D,$08,$74
	.dc.b	$21,$F5,$4C,$94,$D3,$FC,$5C,$7B,$7F,$E0,$BE,$5F,$EA,$2D,$8C,$CF
	.dc.b	$F2,$06,$BD,$37,$FB,$EF,$B4,$EC,$47,$2E,$F7,$FC,$1B,$FC,$DB,$4F
	.dc.b	$F0,$D7,$37,$F2,$F8,$C1,$0F,$E4,$1D,$1F,$D9,$3E,$F7,$FA,$29,$23
	.dc.b	$FC,$09,$19,$7C,$58,$78,$1F,$86,$D7,$C9,$87,$89,$F2,$71,$7E,$51
	.dc.b	$CB,$67,$81,$27,$84,$F4,$6B,$F3,$72,$FF,$1B,$75,$7E,$42,$3F,$BA
	.dc.b	$C7,$E5,$1B,$C5,$FD,$E8,$FF,$1D,$76,$7C,$18,$FE,$7C,$D7,$FA,$93
	.dc.b	$67,$FC,$14,$79,$6F,$C3,$87,$97,$1C,$13,$5F,$D3,$61,$EE,$7F,$65
	.dc.b	$C9,$23,$FC,$DE,$D7,$FA,$75,$FE,$09,$FF,$18,$35,$CE,$1F,$82,$7F
	.dc.b	$B3,$46,$5F,$EB,$1E,$E7,$EF,$C8,$FF,$9F,$1F,$F2,$17,$B7,$23,$FC
	.dc.b	$79,$D5,$FE,$20,$8C,$47,$F8,$67,$D4,$FE,$10,$47,$FD,$0A,$FF,$12
	.dc.b	$F1,$7F,$CB,$58,$BD,$AC,$78,$DF,$E2,$A6,$C6,$6F,$FC,$C3,$39,$7F
	.dc.b	$C9,$9E,$3D,$7F,$B7,$35,$FF,$01,$36,$7F,$CD,$A7,$38,$FF,$17,$57
	.dc.b	$F1,$D2,$3F,$E9,$A7,$FD,$BF,$A7,$F8,$F3,$DA,$B4,$69,$2B,$FC,$EB
	.dc.b	$E5,$ED,$B3,$B6,$CF,$FB,$4C,$FF,$50,$11,$FF,$4F,$35,$FF,$21,$3F
	.dc.b	$CC,$48,$FF,$AB,$9F,$E7,$7E,$6F,$FB,$83,$D4,$BC,$FF,$E5,$26,$D9
	.dc.b	$A7,$FB,$EB,$3D,$61,$FE,$F7,$6B,$FF,$00,$5A,$FF,$93,$FC,$D4,$62
	.dc.b	$1B,$3C,$FC,$37,$0F,$04,$FF,$A6,$27,$27,$1B,$FE,$C1,$3A,$9F,$F2
	.dc.b	$14,$63,$0F,$EA,$ED,$8F,$EB,$6D,$7F,$C9,$1E,$87,$FB,$96,$B9,$47
	.dc.b	$F9,$96,$31,$5F,$D8,$1B,$1F,$D9,$1A,$F8,$36,$F2,$12,$B9,$C7,$FB
	.dc.b	$16,$BF,$DF,$91,$8A,$FF,$7E,$37,$97,$E2,$46,$27,$F3,$24,$AF,$CC
	.dc.b	$D7,$4C,$BF,$43,$3E,$EE,$9F,$88,$49,$F0,$E5,$FE,$8F,$78,$70,$7C
	.dc.b	$E2,$38,$37,$B7,$CD,$23,$83,$C1,$E8,$FD,$62,$38,$3B,$F6,$7D,$52
	.dc.b	$38,$3E,$1F,$5F,$FB,$87,$2F,$F0,$21,$34,$FF,$80,$2F,$F8,$3B,$D3
	.dc.b	$7A,$7F,$C8,$E5,$18,$BF,$E7,$0D,$EC,$E1,$5B,$8C,$57,$F8,$69,$AF
	.dc.b	$FA,$1D,$AF,$FA,$8D,$AF,$68,$E4,$4A,$EE,$DA,$EF,$18,$23,$82,$4F
	.dc.b	$86,$78,$36,$76,$D9,$DB,$E5,$9D,$B6,$76,$D9,$DB,$E9,$9D,$B6,$76
	.dc.b	$D9,$DB,$ED,$8C,$B6,$33,$5F,$08,$95,$D2,$74,$79,$C8,$78,$CA,$77
	.dc.b	$29,$C0,$D1,$0D,$14,$D1,$82,$18,$24,$79,$2B,$D2,$DA,$F4,$D7,$9E
	.dc.b	$31,$5E,$26,$BD,$DA,$F6,$6B,$DA,$39,$12,$BB,$B6,$BB,$C6,$08,$E0
	.dc.b	$93,$E1,$9E,$0D,$9D,$B6,$76,$F9,$67,$6D,$9D,$B6,$76,$FA,$67,$6D
	.dc.b	$9D,$B6,$76,$FB,$63,$2F,$FA,$28,$FF,$3F,$F9,$3F,$E8,$97,$FC,$FE
	.dc.b	$D7,$3E,$4B,$C9,$FF,$41,$C7,$ED,$3E,$B5,$78,$C9,$18,$9F,$DE,$35
	.dc.b	$FD,$A3,$5F,$DA,$35,$FD,$A3,$5E,$D3,$D6,$31,$3D,$A3,$04,$9E,$ED
	.dc.b	$7B,$B5,$EE,$D7,$BB,$5E,$E7,$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81
	.dc.b	$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$3E,$33,$E4,$6B
	.dc.b	$DD,$AF,$46,$BC,$DA,$F3,$9F,$A2,$49,$F9,$13,$A2,$46,$67,$C0,$D7
	.dc.b	$C0,$D7,$C0,$D7,$C0,$D7,$C1,$3D,$A3,$13,$DA,$30,$49,$EE,$D7,$BB
	.dc.b	$5E,$ED,$7B,$B5,$EE,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0
	.dc.b	$53,$81,$4E,$06,$88,$68,$A3,$47,$33,$C1,$FD,$B4,$FF,$BE,$A7,$31
	.dc.b	$88,$12,$33,$8F,$FB,$DC,$FF,$7F,$FA,$EF,$5F,$F7,$B1,$86,$D7,$23
	.dc.b	$9C,$EF,$1E,$D1,$89,$DB,$BE,$77,$1E,$23,$C7,$18,$AE,$31,$FE,$46
	.dc.b	$9D,$36,$33,$2F,$03,$2D,$8C,$C6,$0C,$F3,$48,$E2,$DC,$BF,$D1,$9E
	.dc.b	$C7,$22,$4F,$67,$C3,$F2,$AF,$5A,$92,$7C,$F1,$97,$D3,$D9,$FE,$3E
	.dc.b	$BF,$F2,$09,$33,$FF,$25,$76,$FB,$3D,$A7,$B7,$49,$87,$C2,$8C,$BD
	.dc.b	$3C,$BF,$87,$8F,$F9,$67,$1F,$F2,$B4,$60,$F8,$7E,$E4,$AF,$8E,$D9
	.dc.b	$FF,$55,$78,$5F,$20,$93,$F3,$1E,$BE,$1C,$8F,$5B,$67,$FC,$DB,$C9
	.dc.b	$F5,$3C,$2F,$C0,$46,$2B,$EC,$C6,$36,$FC,$31,$F6,$7D,$BF,$F5,$D7
	.dc.b	$37,$F4,$86,$D1,$A7,$FA,$FF,$97,$55,$CB,$9B,$EF,$4E,$6B,$84,$E6
	.dc.b	$B6,$FF,$9F,$B8,$AB,$FE,$7D,$6A,$E0,$FF,$39,$DF,$F9,$D8,$F3,$C3
	.dc.b	$87,$E8,$C6,$1B,$1E,$C6,$BF,$A3,$6B,$FD,$93,$BD,$F5,$7F,$D9,$BD
	.dc.b	$CF,$87,$9A,$72,$A3,$31,$86,$CF,$1A,$F8,$5D,$15,$FD,$63,$5F,$D5
	.dc.b	$B5,$FF,$57,$B9,$BF,$07,$CD,$5C,$0D,$24,$78,$EB,$EC,$F9,$BF,$9D
	.dc.b	$F7,$7E,$2B,$63,$3D,$7F,$E0,$29,$CB,$63,$F0,$3F,$A7,$3E,$2F,$11
	.dc.b	$FC,$26,$31,$82,$FE,$8C,$61,$B5,$C2,$BF,$8C,$E1,$FA,$FF,$C7,$7B
	.dc.b	$E0,$69,$27,$F8,$B4,$7C,$FF,$9C,$FF,$0B,$FC,$07,$CE,$6C,$EB,$54
	.dc.b	$62,$3F,$93,$D7,$11,$5F,$86,$F0,$FF,$96,$CD,$F1,$7F,$99,$37,$C5
	.dc.b	$FF,$13,$CE,$A3,$B7,$CE,$7F,$8A,$3E,$03,$EF,$B6,$33,$8A,$72,$D8
	.dc.b	$F3,$78,$BF,$E2,$78,$FC,$53,$74,$34,$C7,$C9,$7F,$8D,$BE,$E7,$F1
	.dc.b	$AD,$BF,$80,$D7,$DF,$48,$C4,$7F,$47,$D3,$DA,$FF,$4D,$6B,$FD,$0F
	.dc.b	$3F,$73,$63,$2F,$E8,$FE,$77,$9E,$5F,$A1,$6E,$06,$9E,$09,$FE,$BD
	.dc.b	$5C,$9B,$AF,$CC,$8C,$61,$E0,$9F,$D8,$79,$1F,$2D,$F5,$B5,$CE,$3F
	.dc.b	$2D,$CB,$5F,$DC,$78,$9F,$E5,$CF,$DD,$7F,$66,$D8,$FC,$33,$5F,$8E
	.dc.b	$D7,$FC,$C7,$E8,$FF,$43,$E0,$FD,$FB,$63,$F2,$6D,$7F,$1A,$D7,$E7
	.dc.b	$67,$ED,$23,$F8,$B3,$9F,$CB,$FA,$F1,$52,$30,$D8,$F6,$1A,$8C,$FC
	.dc.b	$F7,$F9,$E3,$F0,$7F,$F3,$A4,$F2,$8C,$3E,$66,$DF,$99,$AF,$98,$F2
	.dc.b	$1F,$83,$C3,$CE,$FB,$BE,$53,$C6,$4A,$F1,$E2,$8F,$1C,$3F,$1B,$F1
	.dc.b	$CB,$0F,$F4,$6C,$61,$B3,$E5,$38,$1E,$1B,$F1,$9F,$91,$C3,$C6,$FC
	.dc.b	$EF,$AA,$FF,$4B,$FE,$7B,$D9,$B5,$F5,$1B,$3B,$84,$F9,$3E,$B3,$F5
	.dc.b	$51,$F6,$9B,$3F,$69,$AF,$E0,$9B,$1F,$5B,$E7,$BF,$80,$3F,$82,$6C
	.dc.b	$FE,$09,$AF,$E2,$9A,$FE,$80,$F6,$9F,$44,$F8,$07,$E3,$3E,$9F,$8C
	.dc.b	$FE,$11,$BA,$1A,$74,$E4,$FD,$45,$75,$39,$9F,$24,$EE,$64,$C1,$A3
	.dc.b	$FC,$97,$ED,$7F,$B2,$7C,$0F,$EB,$91,$8F,$6B,$FC,$A7,$18,$84,$EF
	.dc.b	$AB,$FC,$94,$D8,$E1,$E9,$7F,$3E,$9C,$DF,$26,$C6,$67,$62,$FB,$61
	.dc.b	$FE,$60,$AF,$E6,$44,$9F,$CB,$35,$FC,$33,$5F,$D8,$BF,$B8,$C7,$FD
	.dc.b	$BB,$F0,$DF,$94,$6E,$9F,$3C,$46,$5A,$76,$A8,$76,$29,$F7,$5D,$BF
	.dc.b	$81,$6B,$98,$FD,$1E,$BE,$08,$FC,$74,$BF,$12,$FC,$A3,$4F,$CA,$16
	.dc.b	$37,$1F,$9F,$87,$E8,$34,$8F,$E8,$4C,$9E,$69,$FE,$5F,$8F,$C5,$69
	.dc.b	$FB,$1F,$C5,$3F,$69,$26,$91,$FE,$04,$7C,$68,$B8,$FF,$24,$E7,$F9
	.dc.b	$B1,$83,$C1,$A7,$D9,$6F,$77,$FA,$77,$1F,$E5,$7C,$EC,$5F,$83,$F8
	.dc.b	$88,$FF,$05,$7F,$85,$F5,$F6,$C3,$F9,$7B,$5F,$E6,$ED,$9F,$E4,$C7
	.dc.b	$9E,$3E,$C1,$0F,$C2,$9F,$69,$FE,$63,$E5,$F1,$E6,$FF,$41,$6B,$FD
	.dc.b	$02,$30,$75,$CB,$FD,$4D,$FB,$A7,$F4,$CB,$34,$98,$7F,$84,$FC,$1E
	.dc.b	$04,$9F,$24,$7F,$5A,$C3,$EA,$6D,$F1,$E3,$CC,$7D,$63,$FB,$DF,$85
	.dc.b	$F1,$F8,$FF,$C9,$AD,$8F,$F1,$03,$5D,$DF,$E3,$5F,$A3,$DB,$FB,$97
	.dc.b	$E0,$9F,$0E,$3F,$CB,$5E,$1F,$18,$FF,$30,$36,$32,$73,$8F,$F0,$6C
	.dc.b	$67,$4F,$F0,$BC,$F5,$3F,$C3,$65,$72,$7F,$99,$3C,$5F,$F3,$65,$7F
	.dc.b	$89,$B3,$FF,$12,$91,$C1,$E8,$9F,$8B,$C7,$F8,$98,$FF,$1F,$79,$1F
	.dc.b	$89,$F8,$CF,$F2,$16,$26,$92,$7F,$CB,$B0,$FC,$EF,$FC,$9E,$48,$CB
	.dc.b	$FE,$56,$87,$C5,$D2,$4F,$F9,$8B,$CD,$FC,$C4,$7F,$96,$09,$1D,$9A
	.dc.b	$FF,$8C,$BD,$0F,$CF,$7C,$57,$FA,$11,$B1,$FE,$57,$6B,$BB,$4F,$FB
	.dc.b	$71,$B1,$FE,$88,$6B,$FE,$88,$6B,$FE,$88,$69,$FE,$95,$F4,$FA,$C7
	.dc.b	$FA,$6D,$B1,$93,$7A,$FF,$A6,$AB,$FD,$4D,$CB,$FB,$5F,$8A,$B4,$69
	.dc.b	$30,$FF,$71,$C6,$5B,$1F,$EB,$2F,$61,$EC,$F3,$9C,$9F,$EB,$FA,$E0
	.dc.b	$89,$ED,$BD,$BE,$11,$91,$19,$9E,$31,$C2,$76,$F2,$F7,$2B,$C1,$B4
	.dc.b	$69,$AF,$75,$E9,$7F,$90,$AB,$FD,$C3,$18,$C3,$FC,$6F,$18,$9D,$3E
	.dc.b	$59,$D7,$81,$F0,$7D,$EC,$CF,$B7,$44,$62,$3F,$DB,$4D,$8C,$E1,$DA
	.dc.b	$7E,$06,$4F,$7F,$83,$67,$53,$F2,$27,$26,$BE,$02,$4E,$A7,$E4,$C6
	.dc.b	$31,$FA,$DF,$03,$E1,$25,$6F,$14,$79,$FE,$01,$C5,$49,$5D,$63,$2F
	.dc.b	$1E,$1F,$61,$1C,$1E,$3E,$E7,$8B,$F0,$DB,$19,$8C,$1F,$03,$DF,$91
	.dc.b	$EB,$DD,$18,$D3,$FE,$00,$C3,$E0,$BE,$39,$2B,$83,$F2,$B9,$64,$73
	.dc.b	$6D,$72,$E5,$F9,$E4,$9F,$03,$F6,$7E,$0F,$EF,$F1,$7E,$54,$38,$7F
	.dc.b	$30,$D7,$F3,$CD,$7F,$D4,$75,$FE,$CF,$23,$FE,$73,$7A,$2F,$84,$7C
	.dc.b	$2F,$F2,$A8,$FD,$24,$63,$0F,$E7,$D1,$B9,$D3,$FA,$D9,$D7,$5F,$46
	.dc.b	$C7,$A5,$AF,$49,$E2,$F3,$AE,$B3,$A8,$E9,$3A,$8C,$46,$EB,$B4,$E4
	.dc.b	$95,$C9,$B1,$A8,$C1,$1C,$12,$7C,$D3,$C1,$B3,$B6,$CE,$DF,$54,$65
	.dc.b	$B1,$97,$D3,$19,$6C,$65,$F5,$1D,$08,$72,$29,$C0,$A7,$03,$44,$34
	.dc.b	$53,$46,$08,$60,$95,$F3,$09,$5F,$06,$39,$46,$2B,$D5,$1C,$27,$4F
	.dc.b	$49,$D5,$7B,$5B,$1E,$86,$BD,$27,$8B,$CE,$BA,$CE,$A3,$A4,$EA,$31
	.dc.b	$1B,$AE,$D3,$92,$57,$26,$C6,$A3,$04,$70,$49,$F3,$4F,$06,$CE,$DB
	.dc.b	$3B,$7D,$51,$96,$C6,$5F,$4C,$65,$B1,$97,$D4,$74,$21,$C8,$FE,$01
	.dc.b	$EB,$3D,$67,$AD,$FB,$89,$FF,$53,$C6,$1F,$DD,$CE,$E7,$4D,$8F,$DA
	.dc.b	$7D,$E7,$F9,$95,$B8,$73,$C9,$39,$9F,$DC,$35,$FE,$06,$D7,$F8,$03
	.dc.b	$5F,$E1,$4D,$8E,$73,$D6,$30,$47,$04,$9F,$19,$0F,$0B,$96,$99,$3C
	.dc.b	$6E,$5A,$64,$F2,$B9,$69,$93,$CE,$E5,$A6,$67,$D2,$D9,$F4,$36,$78
	.dc.b	$4E,$A3,$CA,$43,$C6,$53,$C2,$53,$B9,$5E,$4F,$02,$1C,$0A,$70,$2B
	.dc.b	$C1,$D1,$0D,$15,$D3,$82,$38,$24,$FB,$A3,$C8,$D7,$BB,$5E,$8D,$79
	.dc.b	$B6,$39,$CF,$C6,$8C,$11,$C1,$27,$E6,$10,$F0,$B9,$69,$93,$C6,$E5
	.dc.b	$A6,$4F,$2B,$96,$99,$3C,$EE,$5A,$66,$7D,$2D,$9F,$43,$67,$84,$EA
	.dc.b	$3C,$A4,$3C,$65,$3F,$CA,$5E,$6B,$C8,$79,$7F,$B9,$9F,$F2,$94,$E6
	.dc.b	$31,$1C,$27,$79,$7F,$78,$AD,$3F,$DE,$BB,$DF,$11,$FE,$E1,$8C,$36
	.dc.b	$FF,$C9,$ED,$7F,$C9,$8D,$7F,$CB,$99,$78,$1F,$F2,$E3,$79,$7F,$CB
	.dc.b	$8D,$E3,$E1,$2C,$1F,$13,$86,$D7,$36,$E3,$E7,$9E,$38,$7F,$8A,$EB
	.dc.b	$B7,$36,$08,$F8,$79,$F6,$47,$07,$97,$E3,$39,$91,$C1,$EB,$EF,$FF
	.dc.b	$43,$B7,$B3,$D3,$18,$CB,$FD,$19,$1E,$9E,$09,$5E,$96,$CF,$F9,$72
	.dc.b	$3A,$77,$F2,$23,$EF,$F6,$BE,$11,$1C,$1F,$2F,$C2,$F8,$24,$70,$79
	.dc.b	$E3,$CA,$47,$07,$CB,$E2,$2F,$89,$FB,$ED,$E2,$FB,$ED,$7E,$FB,$5F
	.dc.b	$BF,$1F,$E5,$38,$C4,$7E,$02,$30,$7F,$98,$35,$EE,$D9,$DF,$02,$73
	.dc.b	$5F,$E5,$78,$C4,$F4,$DB,$E8,$DF,$CE,$8C,$6D,$C2,$FF,$D6,$55,$A7
	.dc.b	$FD,$69,$2E,$C7,$C9,$FC,$0F,$8A,$FF,$32,$4E,$63,$15,$C2,$B7,$2A
	.dc.b	$DC,$61,$FF,$3D,$F1,$F0,$AF,$F5,$1B,$5F,$F6,$3B,$5F,$F6,$2B,$70
	.dc.b	$7C,$35,$26,$1D,$EB,$FC,$F1,$3A,$8C,$47,$FB,$4D,$F6,$76,$FD,$AF
	.dc.b	$89,$F2,$B6,$7D,$4D,$78,$65,$FE,$03,$A5,$7F,$9E,$BD,$4F,$F3,$B3
	.dc.b	$7C,$1E,$17,$FE,$E1,$9C,$BF,$E8,$B6,$BD,$30,$FC,$5F,$C4,$F8,$B0
	.dc.b	$34,$93,$FE,$93,$F9,$A7,$C0,$FE,$4B,$E2,$BF,$D1,$71,$8B,$46,$27
	.dc.b	$55,$C5,$AF,$F2,$0E,$BF,$E4,$6F,$F1,$D6,$BF,$EA,$1F,$03,$F9,$23
	.dc.b	$9F,$07,$FD,$33,$87,$E1,$7E,$DB,$F4,$7F,$13,$F7,$9B,$5F,$83,$6B
	.dc.b	$F2,$9A,$FE,$AB,$85,$F2,$BF,$E1,$DB,$FC,$83,$63,$31,$E0,$6C,$FE
	.dc.b	$33,$E6,$BF,$67,$F1,$B6,$8D,$24,$FA,$F4,$F6,$BF,$07,$3F,$F1,$E6
	.dc.b	$9E,$E8,$D4,$77,$F8,$FE,$B1,$F9,$D6,$BF,$84,$6B,$EB,$6A,$9E,$8F
	.dc.b	$73,$E5,$10,$FD,$47,$DC,$7C,$4C,$7F,$60,$DC,$CD,$31,$F8,$1F,$F1
	.dc.b	$2F,$8C,$FD,$6B,$6F,$F6,$4D,$7F,$32,$D7,$E8,$73,$F5,$26,$0E,$1F
	.dc.b	$DB,$C7,$D6,$9F,$56,$BF,$21,$FE,$C9,$AF,$C8,$8C,$46,$A3,$EB,$79
	.dc.b	$9E,$92,$61,$D6,$7F,$83,$47,$B7,$EB,$FF,$95,$FE,$AD,$FE,$61,$AF
	.dc.b	$E2,$11,$87,$F8,$63,$6B,$F6,$6D,$7C,$67,$CB,$3E,$71,$F6,$8F,$F7
	.dc.b	$D7,$A6,$F9,$87,$D6,$FE,$62,$D5,$2B,$53,$98,$C5,$70,$AF,$8C,$70
	.dc.b	$85,$70,$8C,$C7,$99,$7E,$CD,$ED,$E2,$63,$FD,$2A,$30,$D9,$F7,$9C
	.dc.b	$0F,$55,$43,$DA,$43,$F5,$5F,$80,$7F,$54,$FC,$CF,$B3,$67,$F9,$A3
	.dc.b	$67,$F8,$DB,$5F,$D9,$7E,$05,$FA,$A9,$FE,$74,$D8,$FE,$42,$D7,$71
	.dc.b	$FA,$E3,$FA,$3F,$C0,$7F,$2D,$E3,$FB,$4D,$8C,$CF,$CB,$3F,$B4,$C6
	.dc.b	$5F,$51,$F8,$93,$ED,$1F,$70,$EB,$C3,$C6,$53,$FE,$3A,$7E,$7C,$F4
	.dc.b	$9D,$4E,$6B,$FD,$55,$CF,$FB,$2A,$E8,$78,$1C,$3E,$13,$E7,$1A,$21
	.dc.b	$FE,$C0,$F4,$7F,$0F,$E7,$BD,$8F,$C1,$7B,$1F,$EB,$F8,$C5,$7E,$32
	.dc.b	$31,$3A,$F1,$7F,$D9,$15,$AA,$E8,$1C,$57,$08,$E5,$3C,$2B,$95,$F8
	.dc.b	$9F,$E3,$47,$F3,$52,$4F,$F1,$A6,$BA,$30,$43,$FB,$26,$5F,$2F,$E2
	.dc.b	$7E,$CE,$9F,$E1,$E6,$EA,$69,$F3,$3A,$7F,$CD,$78,$76,$C3,$F8,$2B
	.dc.b	$6B,$EB,$C7,$37,$F0,$67,$32,$D7,$E2,$23,$47,$E4,$21,$F1,$7F,$A9
	.dc.b	$E0,$FF,$18,$B6,$32,$7E,$60,$B3,$F9,$E5,$FB,$EF,$E9,$1A,$7E,$90
	.dc.b	$B1,$B8,$F6,$65,$F5,$63,$D6,$7C,$72,$9F,$BD,$FD,$77,$F7,$34,$69
	.dc.b	$1F,$E0,$DD,$1F,$C1,$72,$FF,$7F,$36,$33,$D3,$FC,$36,$72,$D8,$F0
	.dc.b	$5F,$F2,$DD,$3E,$73,$4F,$E0,$FF,$B7,$78,$89,$38,$3F,$97,$9F,$66
	.dc.b	$2D,$FF,$8D,$3B,$7E,$F4,$60,$EF,$CB,$FE,$A3,$C3,$BB,$63,$FC,$7D
	.dc.b	$DD,$C1,$FF,$12,$B5,$FE,$9B,$87,$77,$2D,$7F,$C0,$1D,$FF,$E6,$AF
	.dc.b	$95,$FE,$B3,$26,$92,$7F,$B2,$C7,$13,$FB,$29,$7B,$BF,$B2,$B7,$C2
	.dc.b	$F1,$4F,$C5,$3F,$40,$43,$EC,$9F,$DD,$E3,$F1,$AF,$F7,$D6,$BF,$DF
	.dc.b	$5A,$EF,$8F,$E1,$F2,$FE,$82,$3E,$3B,$5F,$F2,$3F,$8B,$F9,$67,$FC
	.dc.b	$C8,$D7,$35,$D0,$FF,$37,$46,$63,$FC,$49,$E3,$7F,$A3,$B6,$FF,$34
	.dc.b	$B6,$3F,$CD,$CD,$7F,$CA,$AD,$7F,$C5,$9E,$3F,$D3,$BF,$F1,$B1,$F1
	.dc.b	$DC,$66,$7E,$3E,$FF,$76,$BE,$31,$F2,$4F,$8E,$3C,$AF,$F2,$77,$C6
	.dc.b	$FF,$93,$70,$34,$93,$FE,$6A,$87,$C6,$FF,$97,$49,$19,$7F,$CD,$10
	.dc.b	$F8,$CA,$49,$FF,$34,$71,$FF,$AC,$23,$FC,$CC,$48,$EC,$D7,$FD,$37
	.dc.b	$E9,$7F,$AD,$BE,$2B,$FB,$B4,$7F,$7E,$F4,$FF,$D0,$91,$88,$46,$1F
	.dc.b	$F4,$4C,$F1,$7F,$D5,$9E,$A5,$7F,$D2,$F8,$AF,$FD,$36,$D8,$DB,$5F
	.dc.b	$F4,$4F,$AD,$C5,$FF,$30,$36,$3F,$D5,$46,$E3,$2D,$7C,$BE,$BF,$F6
	.dc.b	$87,$FD,$3A,$D7,$FD,$72,$D7,$DC,$D5,$EC,$FF,$B1,$3C,$4F,$F6,$51
	.dc.b	$23,$FD,$91,$ED,$6E,$5F,$03,$49,$87,$F3,$AA,$FF,$10,$4E,$A3,$11
	.dc.b	$FE,$2A,$F8,$17,$BB,$E1,$6C,$EA,$51,$8C,$7F,$DD,$27,$D3,$DF,$FB
	.dc.b	$15,$7F,$B4,$6B,$FD,$77,$E6,$7F,$A1,$DB,$19,$7F,$DB,$31,$97,$CB
	.dc.b	$EF,$57,$FC,$7B,$92,$31,$1F,$11,$B1,$9A,$E9,$3E,$BA,$DE,$3E,$4E
	.dc.b	$9F,$C4,$36,$7F,$1A,$D7,$FD,$D7,$EF,$F6,$C8,$D2,$61,$DA,$7E,$1F
	.dc.b	$77,$62,$46,$6D,$1C,$F8,$24,$F1,$C1,$18,$C7,$F3,$99,$B8,$3D,$A4
	.dc.b	$BF,$04,$65,$BF,$01,$F0,$88,$E0,$DF,$27,$C6,$23,$83,$CD,$F0,$3F
	.dc.b	$18,$8E,$0E,$FC,$7F,$38,$8E,$0F,$77,$1F,$CE,$23,$83,$E3,$76,$7C
	.dc.b	$2F,$2E,$EE,$69,$35,$FD,$E7,$C1,$3C,$7F,$E4,$9C,$68,$C6,$3F,$DF
	.dc.b	$63,$13,$A9,$CF,$85,$CA,$30,$4A,$E1,$3A,$AF,$E1,$AD,$7F,$D0,$8D
	.dc.b	$7F,$D1,$AD,$7B,$C7,$29,$E9,$5D,$E3,$04,$70,$49,$EF,$3C,$1B,$3B
	.dc.b	$6C,$ED,$F1,$CE,$DB,$3B,$6C,$ED,$F3,$CE,$DB,$3B,$6C,$ED,$F5,$B6
	.dc.b	$BD,$C4,$AE,$53,$A3,$CA,$43,$C2,$53,$A9,$4E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$C1,$23,$C9,$5E,$B8,$C4,$F9,$A7,$35,$EB,$8C,$12,$B8,$4E,$AB
	.dc.b	$CC,$D7,$C6,$D7,$C4,$D7,$BC,$72,$9E,$95,$DE,$30,$47,$04,$9E,$F3
	.dc.b	$C1,$B3,$B6,$CE,$DF,$1C,$ED,$B3,$B6,$CE,$DF,$3C,$ED,$B3,$B6,$CE
	.dc.b	$DF,$5B,$6B,$DC,$4A,$FC,$0F,$A2,$F4,$5E,$8F,$F7,$1C,$3F,$D4,$27
	.dc.b	$F7,$69,$D4,$65,$B1,$FD,$DF,$EF,$BF,$DE,$F9,$71,$6C,$64,$46,$5A
	.dc.b	$65,$C1,$1C,$10,$FF,$03,$C7,$46,$9C,$8B,$1C,$5A,$71,$04,$8E,$4E
	.dc.b	$03,$87,$47,$53,$4E,$5D,$99,$37,$1E,$17,$66,$4D,$B9,$76,$79,$4A
	.dc.b	$7A,$08,$72,$78,$3C,$8E,$05,$36,$72,$38,$1A,$8F,$41,$1F,$23,$63
	.dc.b	$AB,$63,$81,$A2,$46,$9A,$69,$C1,$1C,$10,$F2,$9E,$73,$CA,$62,$3C
	.dc.b	$0D,$3A,$96,$3A,$34,$E8,$72,$21,$F0,$48,$77,$2C,$76,$76,$64,$8E
	.dc.b	$5D,$9E,$23,$6E,$5D,$99,$37,$1E,$67,$66,$4D,$B9,$76,$7A,$8A,$7B
	.dc.b	$48,$72,$78,$3C,$8E,$05,$36,$72,$38,$1A,$8F,$41,$1F,$23,$63,$F1
	.dc.b	$BB,$7F,$7B,$BF,$F3,$93,$6B,$FC,$EE,$D7,$F5,$0D,$7F,$5B,$F0,$8F
	.dc.b	$8B,$FD,$FD,$7F,$9E,$E3,$17,$FC,$3A,$BF,$D0,$31,$89,$D5,$70,$9F
	.dc.b	$F4,$2F,$37,$F9,$6E,$7F,$C5,$3C,$79,$9F,$F2,$FB,$5F,$F2,$5B,$5F
	.dc.b	$F3,$6F,$C3,$5A,$34,$98,$F8,$E7,$F8,$17,$C3,$BE,$53,$93,$67,$53
	.dc.b	$FC,$2A,$31,$8F,$FA,$5D,$B1,$E8,$E9,$F0,$12,$B8,$C7,$FA,$02,$3A
	.dc.b	$FA,$92,$73,$48,$C5,$FF,$9C,$7B,$3D,$9D,$BF,$E4,$CB,$E6,$D9,$D4
	.dc.b	$76,$F8,$97,$C5,$7C,$36,$C6,$63,$07,$4F,$06,$47,$46,$D7,$F9,$63
	.dc.b	$E2,$FE,$51,$27,$D7,$C0,$8F,$A1,$F5,$AF,$81,$A4,$AF,$83,$19,$7E
	.dc.b	$1F,$37,$F9,$27,$E3,$3F,$1F,$E0,$7F,$92,$5B,$E0,$1A,$7E,$1A,$B1
	.dc.b	$C9,$F8,$AE,$8F,$E9,$B3,$92,$47,$D3,$9D,$BF,$EA,$6E,$5D,$B9,$9F
	.dc.b	$F3,$AC,$62,$3F,$07,$C9,$5F,$C0,$35,$72,$FD,$7B,$EB,$1A,$6A,$F6
	.dc.b	$3F,$B2,$FC,$67,$FB,$75,$B1,$FE,$E2,$6B,$FD,$4D,$AF,$FA,$9B,$A7
	.dc.b	$FB,$4C,$FF,$A2,$9B,$1F,$A6,$6B,$FE,$C2,$6B,$FD,$C3,$CA,$7E,$0F
	.dc.b	$8C,$11,$85,$7F,$EF,$47,$33,$B9,$D4,$FF,$BD,$7D,$37,$C4,$FF,$7C
	.dc.b	$F8,$DC,$37,$1F,$F7,$EC,$EB,$15,$69,$B3,$FD,$C4,$7C,$6B,$E1,$B9
	.dc.b	$47,$F7,$58,$C4,$F0,$8C,$4E,$9F,$EF,$B3,$AA,$ED,$F1,$B2,$7C,$78
	.dc.b	$FF,$84,$A3,$11,$CB,$04,$78,$E3,$C7,$86,$DF,$36,$1E,$29,$D3,$63
	.dc.b	$F9,$2F,$2F,$FA,$85,$FC,$C9,$0C,$F2,$F5,$C3,$E8,$FD,$7F,$F8,$93
	.dc.b	$F2,$FF,$A1,$BA,$31,$7F,$E1,$5A,$D8,$AD,$CE,$A7,$FC,$63,$4F,$A8
	.dc.b	$FF,$0A,$5F,$DE,$6C,$66,$3C,$4D,$9F,$0F,$DF,$7F,$8D,$BE,$26,$D1
	.dc.b	$A4,$9F,$59,$ED,$3F,$A1,$E7,$F0,$5C,$9E,$9B,$FF,$7D,$4F,$BB,$D6
	.dc.b	$ED,$39,$6B,$EA,$6B,$FE,$38,$EC,$FF,$67,$3D,$CC,$5F,$77,$C3,$AF
	.dc.b	$B3,$BF,$E2,$47,$E8,$5B,$97,$E8,$1B,$97,$F1,$98,$F9,$93,$E7,$F5
	.dc.b	$3F,$CA,$FF,$A2,$7D,$46,$C6,$65,$39,$6C,$7B,$FB,$7F,$CB,$FF,$23
	.dc.b	$F7,$1B,$19,$94,$E5,$B1,$F2,$0F,$59,$EF,$1A,$7C,$27,$E6,$7A,$DD
	.dc.b	$1F,$C0,$B5,$FE,$00,$D9,$FD,$FB,$72,$FE,$C7,$0F,$DE,$7F,$23,$DB
	.dc.b	$FE,$76,$6C,$66,$FE,$D4,$E5,$F3,$1F,$4B,$6F,$F4,$46,$3E,$58,$F4
	.dc.b	$9F,$E6,$DF,$01,$F5,$7F,$E8,$6E,$7F,$F3,$C4,$62,$7F,$45,$8E,$EB
	.dc.b	$4D,$8C,$9F,$9E,$F3,$BF,$D1,$7F,$9C,$FF,$A0,$67,$F8,$DC,$61,$FC
	.dc.b	$83,$67,$84,$27,$FC,$F9,$8F,$9E,$7D,$47,$CF,$71,$3F,$C9,$BF,$00
	.dc.b	$FF,$4C,$FE,$E1,$DF,$74,$62,$7F,$3C,$DB,$FD,$9B,$5F,$E6,$5F,$81
	.dc.b	$7E,$B6,$7F,$48,$D8,$FD,$CB,$5D,$C7,$EC,$4F,$D5,$7D,$A7,$E3,$1A
	.dc.b	$7E,$28,$BC,$7E,$58,$D4,$FD,$22,$1E,$E3,$EC,$CF,$F4,$2A,$DC,$78
	.dc.b	$2B,$DD,$18,$9D,$1D,$E7,$99,$F9,$4E,$8F,$D9,$3D,$4C,$E7,$F3,$5D
	.dc.b	$70,$FE,$1A,$77,$18,$3F,$98,$7A,$DF,$EC,$CF,$7D,$ED,$FF,$AD,$63
	.dc.b	$0D,$F6,$7F,$C9,$53,$AA,$FF,$5E,$D6,$9B,$3A,$AE,$1E,$8F,$FB,$0A
	.dc.b	$F8,$52,$F8,$36,$F8,$3F,$9D,$BF,$F0,$99,$1F,$E7,$6D,$7F,$88,$B5
	.dc.b	$E0,$7D,$EC,$DF,$15,$F3,$1A,$65,$C1,$F9,$4C,$99,$F9,$AB,$99,$D8
	.dc.b	$F0,$9F,$82,$9F,$C0,$57,$48,$FE,$87,$BF,$FB,$C2,$72,$78,$7C,$C9
	.dc.b	$5E,$56,$D7,$E6,$21,$FB,$87,$E7,$F8,$BF,$DF,$2D,$8C,$9F,$A0,$2C
	.dc.b	$F9,$48,$7A,$08,$FE,$D8,$7E,$73,$FA,$BE,$74,$60,$FD,$C9,$67,$F7
	.dc.b	$33,$FC,$23,$A3,$FD,$F7,$2F,$CC,$71,$5C,$9B,$19,$3F,$85,$16,$3C
	.dc.b	$A7,$E5,$37,$FB,$0D,$3F,$8F,$C3,$F7,$DF,$E4,$64,$8C,$BF,$DE,$3B
	.dc.b	$FE,$74,$7F,$34,$9F,$BE,$DD,$B5,$18,$8F,$0F,$33,$8B,$EF,$E3,$E2
	.dc.b	$8C,$71,$7E,$06,$FF,$C6,$FE,$07,$6C,$3F,$BF,$B5,$FF,$1F,$36,$36
	.dc.b	$77,$8F,$F2,$17,$85,$F3,$BF,$76,$FE,$B0,$48,$FE,$B8,$FF,$84,$BC
	.dc.b	$4F,$95,$1F,$DE,$5A,$FF,$83,$9A,$FF,$6D,$AF,$A4,$FE,$50,$F6,$E7
	.dc.b	$EC,$70,$7F,$94,$FC,$5F,$BF,$CD,$FE,$03,$AF,$F3,$34,$FF,$80,$DA
	.dc.b	$FF,$80,$DA,$FE,$2C,$FA,$C7,$6C,$FE,$37,$73,$C7,$EB,$3F,$E1,$16
	.dc.b	$BF,$E2,$66,$BF,$E6,$56,$D7,$F8,$9C,$E7,$E4,$7E,$53,$83,$FC,$6A
	.dc.b	$DA,$FF,$0E,$35,$FF,$18,$F9,$3F,$12,$BF,$C6,$FC,$BF,$E3,$67,$E2
	.dc.b	$CE,$9A,$FA,$F8,$BE,$FE,$1F,$81,$DB,$E4,$4F,$F9,$57,$E2,$FF,$C9
	.dc.b	$FE,$6F,$E9,$9B,$81,$A7,$F9,$B4,$93,$FE,$60,$8D,$36,$38,$FC,$54
	.dc.b	$8F,$F2,$ED,$7E,$CA,$31,$02,$7A,$3F,$E7,$B8,$FF,$3D,$12,$3B,$3F
	.dc.b	$E5,$AF,$4F,$F7,$7F,$13,$71,$34,$95,$FE,$99,$8C,$B6,$3F,$D3,$D1
	.dc.b	$88,$FF,$4E,$35,$F2,$71,$FF,$02,$9F,$F5,$64,$78,$0F,$F5,$61,$4F
	.dc.b	$F5,$61,$4F,$F5,$B7,$17,$FB,$E7,$CA,$FF,$53,$47,$FA,$4F,$D5,$FF
	.dc.b	$5E,$4E,$70,$FF,$53,$CF,$FB,$2E,$3F,$D8,$F1,$88,$FF,$66,$35,$F9
	.dc.b	$1E,$D3,$AF,$AC,$7C,$B1,$F0,$2F,$67,$31,$F3,$E3,$17,$CE,$37,$3A
	.dc.b	$6C,$67,$93,$F8,$3E,$3F,$EE,$82,$4F,$FA,$41,$FF,$4E,$7B,$5F,$EF
	.dc.b	$1F,$85,$BE,$E7,$F3,$06,$E0,$69,$FF,$00,$5C,$5F,$FA,$DF,$4F,$F3
	.dc.b	$44,$E4,$91,$FE,$BB,$F8,$07,$BD,$D6,$30,$D7,$E1,$46,$AF,$AC,$62
	.dc.b	$74,$FE,$47,$E0,$1F,$01,$88,$46,$2B,$8C,$6E,$75,$39,$C3,$FD,$79
	.dc.b	$83,$E0,$79,$37,$0E,$38,$23,$15,$C2,$BC,$1C,$5F,$EF,$F8,$FC,$21
	.dc.b	$2B,$F1,$1C,$7E,$5D,$3F,$01,$8F,$46,$D7,$E2,$09,$5F,$8F,$F8,$2E
	.dc.b	$A4,$AF,$1E,$08,$E9,$EC,$C9,$E8,$DB,$FF,$6C,$60,$F3,$7C,$44,$AF
	.dc.b	$13,$F2,$79,$7F,$1B,$5F,$ED,$F6,$E5,$FE,$F3,$6B,$F9,$76,$BF,$E7
	.dc.b	$F8,$FF,$73,$72,$7F,$56,$6F,$77,$F8,$61,$BE,$27,$CC,$F8,$3F,$EE
	.dc.b	$FE,$0F,$FB,$B5,$B8,$7F,$C0,$16,$BF,$A5,$6B,$FC,$45,$A9,$A4,$7F
	.dc.b	$99,$1B,$29,$B2,$BB,$CF,$FC,$59,$19,$6C,$66,$7C,$D3,$C0,$93,$A6
	.dc.b	$C6,$A3,$04,$70,$48,$F7,$1C,$0B,$1B,$6C,$6D,$F8,$46,$CB,$1B,$6C
	.dc.b	$6D,$F8,$11,$96,$C6,$5F,$73,$67,$DD,$1E,$02,$1D,$0A,$72,$29,$C8
	.dc.b	$AF,$27,$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9,$C1,$1C,$11,$F8
	.dc.b	$C7,$02,$9B,$2B,$B9,$F8,$D1,$96,$C6,$67,$E3,$4F,$02,$4E,$9B,$1A
	.dc.b	$8C,$11,$C1,$23,$E3,$9C,$0B,$1B,$6C,$6D,$F8,$C6,$CB,$1B,$6C,$6D
	.dc.b	$F8,$11,$96,$C6,$5F,$73,$67,$DD,$1F,$D4,$CF,$F5,$07,$AB,$5F,$F5
	.dc.b	$4B,$5C,$9F,$EB,$6F,$4D,$19,$9F,$F4,$BD,$DC,$3F,$CF,$77,$FB,$F6
	.dc.b	$BF,$BF,$6B,$FB,$F6,$BF,$BF,$6B,$FB,$F6,$CF,$68,$E9,$3D,$C8,$E0
	.dc.b	$87,$77,$2D,$32,$78,$5C,$B4,$C9,$E3,$72,$D3,$27,$95,$CB,$4C,$C7
	.dc.b	$9D,$F0,$10,$EC,$53,$A1,$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF,$07
	.dc.b	$44,$34,$57,$4E,$08,$E0,$93,$F4,$89,$3F,$02,$32,$D8,$CC,$FC,$26
	.dc.b	$BE,$B6,$BE,$66,$BE,$36,$BE,$26,$CF,$68,$E9,$3D,$C8,$E0,$87,$77
	.dc.b	$2D,$32,$78,$5C,$B4,$C9,$E3,$72,$D3,$27,$95,$CB,$4C,$C7,$9D,$F0
	.dc.b	$10,$EC,$53,$A1,$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF,$DC,$CF,$FD
	.dc.b	$97,$7F,$DD,$09,$3F,$C0,$63,$2F,$F0,$3F,$88,$7C,$3F,$FB,$D2,$BF
	.dc.b	$82,$46,$30,$FE,$03,$5B,$AD,$36,$32,$FF,$97,$F8,$3F,$C2,$91,$88
	.dc.b	$FF,$32,$F0,$70,$74,$D7,$31,$B5,$2E,$1D,$46,$1B,$19,$9E,$0D,$C7
	.dc.b	$FC,$4B,$E7,$E8,$93,$A7,$8F,$7C,$8E,$2D,$9F,$F1,$6F,$9D,$D8,$93
	.dc.b	$D9,$F1,$F4,$FF,$92,$E7,$FC,$91,$3E,$98,$C6,$5F,$E6,$5E,$9F,$4E
	.dc.b	$DF,$F0,$06,$BD,$CD,$BF,$9F,$CB,$A3,$AF,$64,$AE,$8D,$9F,$F2,$EF
	.dc.b	$07,$88,$93,$E1,$7E,$2F,$99,$23,$E2,$B6,$7F,$C6,$FF,$0B,$33,$FC
	.dc.b	$26,$2E,$54,$E6,$7E,$E7,$03,$1C,$C7,$FA,$16,$31,$3F,$42,$73,$3C
	.dc.b	$1A,$FF,$A9,$B8,$36,$FF,$A3,$5A,$FF,$AB,$27,$FD,$5E,$7F,$57,$97
	.dc.b	$82,$FB,$BF,$17,$C3,$66,$31,$28,$C4,$EA,$FF,$D6,$53,$CA,$31,$1F
	.dc.b	$92,$8E,$0F,$FA,$7F,$9B,$2F,$E5,$2B,$FC,$DF,$18,$87,$E4,$1F,$A5
	.dc.b	$9F,$ED,$4D,$8F,$ED,$CD,$7F,$D8,$4D,$7F,$D8,$F9,$FC,$BF,$03,$E4
	.dc.b	$12,$3E,$66,$08,$FE,$ED,$F7,$96,$1F,$11,$89,$FA,$8D,$8C,$D7,$F7
	.dc.b	$FA,$E1,$3F,$1F,$C6,$FB,$5F,$9D,$DB,$34,$97,$C7,$0F,$C0,$7C,$47
	.dc.b	$52,$46,$69,$1F,$E8,$28,$87,$E3,$38,$BF,$C1,$9D,$CE,$4F,$C7,$4F
	.dc.b	$D9,$FC,$23,$F1,$9F,$05,$71,$34,$93,$F6,$B9,$1F,$4F,$FE,$0A,$AF
	.dc.b	$C2,$B6,$33,$1F,$E1,$D3,$25,$5E,$5F,$E6,$FE,$27,$F1,$4D,$BE,$CD
	.dc.b	$47,$E8,$67,$8D,$2B,$DF,$3F,$EF,$FC,$FD,$DA,$7C,$0D,$3D,$F3,$F0
	.dc.b	$CF,$CC,$7D,$F7,$57,$A3,$70,$FC,$FB,$71,$3B,$96,$1D,$9F,$09,$E6
	.dc.b	$70,$F7,$3F,$43,$E7,$FE,$E6,$FE,$6C,$61,$FA,$2D,$72,$FC,$05,$1C
	.dc.b	$75,$FD,$A2,$FA,$6F,$E0,$9F,$9D,$E1,$B7,$E0,$6B,$CE,$78,$5A,$B9
	.dc.b	$FC,$37,$8A,$B4,$D7,$F8,$85,$23,$15,$E4,$75,$3F,$13,$4F,$9E,$FD
	.dc.b	$36,$BF,$15,$AE,$F5,$FC,$35,$FF,$35,$CB,$DB,$5D,$A3,$15,$F2,$A3
	.dc.b	$1F,$71,$FC,$8F,$F3,$0F,$86,$D9,$F9,$ED,$7E,$03,$5F,$16,$1F,$85
	.dc.b	$23,$F6,$B9,$5E,$07,$F7,$1C,$BF,$CF,$11,$8F,$0F,$E2,$37,$5F,$AF
	.dc.b	$7F,$C1,$EF,$83,$F8,$0F,$93,$FE,$71,$F8,$3F,$F9,$D2,$BE,$DC,$61
	.dc.b	$D0,$8D,$6C,$9D,$57,$CB,$CF,$CE,$FE,$34,$C1,$F8,$4C,$BF,$A3,$7E
	.dc.b	$43,$C9,$B8,$FE,$49,$B3,$F8,$36,$BF,$7B,$C7,$FD,$A4,$FE,$61,$B1
	.dc.b	$F8,$96,$BB,$8F,$D9,$1F,$9E,$FA,$1F,$E1,$6D,$3F,$85,$97,$6F,$6C
	.dc.b	$6A,$7D,$27,$AC,$FA,$A7,$EB,$27,$F4,$35,$B8,$F0,$57,$C6,$8C,$4E
	.dc.b	$8E,$F3,$D0,$87,$43,$B1,$F9,$13,$F7,$25,$33,$F1,$8F,$47,$FC,$63
	.dc.b	$C8,$F4,$BF,$89,$D7,$FA,$DA,$30,$FF,$AE,$8F,$D3,$46,$AB,$97,$95
	.dc.b	$FC,$6A,$B7,$5C,$6B,$57,$D9,$B5,$B6,$CF,$43,$F3,$2F,$FB,$27,$A7
	.dc.b	$F8,$FB,$79,$FF,$C1,$4D,$E0,$EF,$5C,$0F,$E8,$91,$92,$3F,$7F,$37
	.dc.b	$C3,$7C,$D6,$99,$70,$7F,$B3,$30,$F0,$CF,$42,$1F,$97,$AF,$E2,$FB
	.dc.b	$FF,$86,$6B,$F0,$37,$D2,$BF,$8A,$4E,$63,$F1,$13,$E1,$F6,$E4,$79
	.dc.b	$5B,$5F,$99,$87,$E4,$3F,$9F,$69,$F9,$F2,$C6,$E3,$D1,$9F,$C6,$8F
	.dc.b	$39,$F1,$4F,$41,$FB,$4F,$D2,$BF,$67,$CB,$FE,$31,$6D,$9A,$7F,$BC
	.dc.b	$3A,$7F,$82,$65,$FC,$23,$A7,$FC,$F4,$DE,$6F,$85,$8F,$E2,$AB,$6F
	.dc.b	$88,$FE,$49,$B7,$D4,$7C,$67,$F1,$AC,$5F,$96,$7F,$20,$6C,$7F,$20
	.dc.b	$6B,$B8,$FE,$57,$1F,$72,$3F,$95,$35,$FE,$89,$A7,$91,$B1,$97,$E2
	.dc.b	$10,$F2,$9F,$21,$FE,$84,$6C,$B8,$FF,$4C,$71,$D3,$FD,$25,$BD,$DF
	.dc.b	$E5,$BF,$DD,$BF,$AC,$41,$A4,$BF,$EC,$97,$FD,$82,$7C,$D1,$FD,$51
	.dc.b	$A7,$F6,$63,$2F,$98,$F9,$27,$A7,$8F,$F2,$EF,$F6,$DF,$07,$FB,$EC
	.dc.b	$B8,$FF,$C0,$AD,$9F,$F1,$07,$83,$FE,$71,$9C,$9F,$56,$3A,$69,$F8
	.dc.b	$B7,$BD,$7F,$9A,$3C,$3E,$11,$FE,$1C,$8E,$8F,$86,$32,$D8,$CB,$FE
	.dc.b	$1F,$F1,$3F,$D0,$1C,$1F,$E6,$F6,$C7,$F8,$CD,$AF,$AD,$AF,$F8,$9F
	.dc.b	$C5,$F8,$B9,$9A,$4B,$FB,$D3,$F1,$CF,$5F,$4F,$F0,$87,$03,$CA,$FF
	.dc.b	$31,$7C,$27,$F9,$7C,$91,$FE,$53,$7F,$CB,$9F,$AB,$C8,$FF,$24,$79
	.dc.b	$7F,$E9,$98,$C4,$7F,$2A,$AF,$F3,$67,$C1,$FF,$9A,$49,$1F,$E5,$D7
	.dc.b	$FD,$39,$ED,$48,$FF,$3A,$B5,$FF,$3E,$F5,$7F,$85,$9F,$F4,$0B,$5C
	.dc.b	$E7,$FE,$2A,$72,$D7,$FD,$25,$D1,$FE,$8E,$8C,$47,$FB,$92,$30,$7F
	.dc.b	$A7,$23,$27,$FA,$6C,$BB,$7F,$A8,$A3,$47,$F4,$69,$FF,$58,$4F,$FA
	.dc.b	$A9,$FF,$58,$35,$FF,$7C,$75,$FF,$97,$5C,$D7,$FA,$EC,$8F,$FA,$51
	.dc.b	$AF,$F8,$0D,$FE,$7C,$47,$FD,$00,$F8,$5F,$F7,$DF,$AB,$FE,$C9,$F7
	.dc.b	$31,$ED,$A3,$13,$E7,$8C,$60,$AD,$D6,$AB,$FB,$0F,$B4,$CF,$11,$FE
	.dc.b	$76,$8C,$5F,$18,$F8,$51,$8B,$FF,$74,$CE,$67,$9F,$91,$C9,$52,$30
	.dc.b	$D9,$E1,$3D,$E7,$87,$B3,$FE,$FD,$BF,$F7,$DD,$6A,$93,$97,$EA,$35
	.dc.b	$F5,$FB,$5D,$89,$3C,$70,$47,$97,$37,$B7,$49,$5D,$E3,$2D,$DB,$81
	.dc.b	$1C,$1B,$F5,$BA,$11,$C1,$CF,$87,$C0,$47,$07,$7F,$67,$C8,$47,$07
	.dc.b	$C1,$D3,$E8,$69,$F0,$37,$F8,$E4,$9F,$8D,$E0,$7F,$AA,$37,$FF,$04
	.dc.b	$12,$BF,$D9,$4F,$E2,$FD,$C7,$57,$F0,$E9,$46,$2F,$F8,$43,$7B,$78
	.dc.b	$56,$E3,$15,$FA,$56,$BF,$E2,$C6,$BF,$EC,$06,$BD,$A3,$91,$2B,$BB
	.dc.b	$6B,$BC,$60,$8E,$09,$3E,$19,$E0,$D9,$DB,$67,$6F,$96,$76,$D9,$DB
	.dc.b	$67,$6F,$A6,$76,$D9,$DB,$67,$6F,$B6,$32,$D8,$CD,$7C,$22,$57,$49
	.dc.b	$D1,$E7,$21,$E3,$29,$DC,$A7,$03,$44,$34,$53,$46,$08,$60,$91,$E4
	.dc.b	$AF,$4B,$6B,$D3,$5E,$78,$C5,$78,$9A,$F7,$6B,$D9,$AF,$68,$E4,$4A
	.dc.b	$EE,$DA,$EF,$18,$23,$82,$4F,$86,$78,$36,$76,$D9,$DB,$E5,$9D,$B6
	.dc.b	$76,$D9,$DB,$E9,$9D,$B6,$76,$D9,$DB,$ED,$8C,$B6,$33,$5F,$08,$95
	.dc.b	$D2,$74,$79,$CF,$BB,$E7,$7F,$A4,$1F,$E8,$4D,$7F,$AA,$35,$FF,$46
	.dc.b	$9F,$B3,$F8,$86,$1F,$D2,$63,$F7,$B1,$87,$F8,$13,$5E,$11,$BB,$4E
	.dc.b	$67,$F7,$8D,$7F,$78,$D7,$AC,$F1,$8C,$4F,$28,$C1,$27,$C2,$D7,$C2
	.dc.b	$60,$86,$08,$78,$9E,$0D,$36,$53,$67,$99,$DB,$4D,$94,$D9,$EA,$32
	.dc.b	$53,$31,$EB,$72,$D7,$33,$F0,$09,$3E,$08,$E4,$79,$48,$78,$4A,$68
	.dc.b	$C1,$0C,$12,$7C,$0F,$06,$BB,$6B,$B9,$F1,$BE,$86,$BE,$89,$F3,$4F
	.dc.b	$02,$4E,$A7,$B3,$5E,$CD,$7B,$4F,$28,$C4,$F2,$8C,$12,$7C,$4D,$7C
	.dc.b	$46,$08,$60,$87,$89,$E0,$D3,$65,$36,$79,$9D,$B4,$D9,$4D,$9E,$A3
	.dc.b	$25,$33,$1E,$B7,$2D,$73,$3F,$00,$93,$E0,$8E,$47,$94,$87,$84,$A7
	.dc.b	$DE,$C9,$E0,$FF,$76,$F3,$7F,$C8,$51,$8C,$91,$89,$D5,$7F,$95,$EB
	.dc.b	$53,$FE,$49,$F7,$9F,$07,$FC,$12,$30,$DA,$FF,$2F,$4E,$B1,$EB,$18
	.dc.b	$9D,$22,$4B,$E0,$F3,$18,$6A,$8D,$57,$48,$E4,$66,$77,$CC,$E4,$A4
	.dc.b	$AE,$15,$E1,$9D,$46,$21,$F0,$4F,$82,$F0,$B6,$74,$DC,$FF,$CD,$8D
	.dc.b	$AF,$F3,$DC,$7F,$9B,$3A,$3C,$64,$AE,$3E,$77,$42,$38,$3C,$5D,$9F
	.dc.b	$E4,$E6,$F6,$7C,$28,$C5,$FF,$9E,$DB,$87,$FA,$37,$E0,$FE,$2C,$FF
	.dc.b	$9F,$DB,$19,$DB,$E5,$C6,$23,$E4,$FE,$EC,$FA,$BF,$9E,$DC,$BE,$23
	.dc.b	$70,$FF,$45,$36,$BF,$D0,$DF,$19,$F4,$09,$5F,$2F,$81,$1F,$6B,$C8
	.dc.b	$79,$94,$95,$F4,$63,$2F,$1E,$3F,$EA,$3F,$0D,$FE,$96,$6E,$3F,$E9
	.dc.b	$56,$BF,$82,$C7,$ED,$CF,$D4,$E4,$CC,$FE,$3F,$93,$ED,$4E,$70,$34
	.dc.b	$F9,$9C,$5C,$2B,$FD,$22,$DE,$2F,$A9,$3A,$AE,$35,$F9,$D5,$F0,$BF
	.dc.b	$A4,$6C,$66,$30,$7F,$B0,$E1,$E4,$3E,$3F,$FB,$7C,$E6,$31,$3C,$AB
	.dc.b	$F5,$0F,$0A,$DC,$7E,$8A,$30,$6D,$FC,$F6,$BF,$DE,$B9,$BE,$54,$9A
	.dc.b	$49,$FF,$6B,$69,$F9,$F8,$F8,$C4,$8C,$EB,$CD,$F0,$62,$F0,$3F,$C0
	.dc.b	$FF,$09,$E9,$6C,$E9,$AF,$AA,$78,$62,$AF,$81,$9E,$AB,$FC,$23,$5F
	.dc.b	$0A,$73,$5B,$9F,$0D,$6D,$76,$7F,$1D,$7F,$8E,$77,$7F,$86,$63,$11
	.dc.b	$FC,$77,$BB,$FD,$32,$FC,$FE,$2F,$F0,$A6,$1F,$51,$B9,$7E,$CE,$75
	.dc.b	$3F,$EE,$B3,$EC,$FD,$97,$F8,$9B,$F0,$7F,$B0,$D9,$F8,$AD,$7E,$03
	.dc.b	$5F,$6F,$DA,$76,$25,$75,$B4,$7F,$89,$72,$F4,$D7,$F2,$06,$BF,$CA
	.dc.b	$DA,$FC,$2C,$BD,$AF,$D6,$FB,$5F,$59,$30,$EC,$F5,$9D,$CF,$56,$BD
	.dc.b	$5A,$F5,$D3,$CD,$F4,$7F,$18,$97,$83,$D5,$C7,$FE,$4A,$FC,$3B,$F3
	.dc.b	$8D,$BF,$C4,$B5,$FC,$23,$5F,$C0,$FD,$07,$D2,$25,$7C,$57,$D1,$87
	.dc.b	$DD,$C7,$CB,$9F,$C2,$3E,$2D,$7E,$BB,$ED,$FE,$D9,$1F,$9C,$FF,$94
	.dc.b	$7F,$4B,$7D,$5F,$F5,$06,$CF,$F4,$56,$F8,$5F,$D4,$33,$FE,$B5,$18
	.dc.b	$7F,$19,$97,$8C,$8F,$E4,$8F,$DE,$78,$BF,$8C,$25,$FD,$F7,$FC,$C5
	.dc.b	$F2,$8E,$4F,$E0,$73,$DE,$30,$FF,$0E,$31,$1A,$AF,$EC,$73,$AE,$DF
	.dc.b	$F3,$86,$BF,$0F,$6F,$8A,$4C,$BF,$1D,$E4,$FF,$9F,$BE,$03,$FC,$F5
	.dc.b	$9F,$F1,$48,$C3,$FB,$E6,$E5,$FB,$96,$BF,$E8,$6F,$8A,$71,$7F,$A1
	.dc.b	$6B,$F8,$A4,$61,$FE,$33,$3F,$C6,$A3,$0F,$81,$B1,$FC,$6F,$A5,$F3
	.dc.b	$DF,$AD,$8C,$36,$BF,$91,$CE,$A4,$FA,$8C,$1F,$0B,$3F,$81,$3F,$4C
	.dc.b	$FA,$E7,$F3,$3B,$F0,$BE,$5C,$3D,$4D,$BF,$4B,$EF,$8C,$BF,$A5,$CF
	.dc.b	$EF,$BD,$08,$7E,$AD,$CB,$4F,$11,$FE,$1A,$F3,$FF,$D6,$DF,$13,$FC
	.dc.b	$AE,$3F,$C4,$D1,$8F,$03,$5F,$13,$FE,$2C,$AD,$52,$B4,$F0,$9D,$57
	.dc.b	$0E,$6F,$F1,$85,$F0,$BD,$DF,$1B,$DB,$EB,$BF,$E5,$24,$7F,$3A,$D7
	.dc.b	$D8,$D7,$81,$F8,$0C,$7F,$D9,$31,$F7,$5B,$96,$4F,$E8,$33,$F3,$67
	.dc.b	$EB,$65,$EC,$8E,$67,$63,$C2,$7E,$0A,$5F,$83,$FD,$FD,$FF,$DA,$ED
	.dc.b	$8C,$9B,$7C,$4F,$82,$7E,$D3,$46,$BF,$97,$C3,$67,$E9,$3B,$FF,$CB
	.dc.b	$9D,$37,$8B,$FD,$CA,$DB,$FE,$0F,$87,$E9,$5A,$FA,$88,$7F,$29,$BF
	.dc.b	$90,$B9,$BF,$7B,$87,$FB,$9D,$B5,$FB,$E3,$FC,$0E,$58,$F2,$74,$7C
	.dc.b	$EA,$FE,$E6,$D7,$FB,$73,$5F,$ED,$6D,$7C,$47,$CC,$3E,$71,$0F,$E3
	.dc.b	$B8,$29,$E9,$FF,$21,$97,$57,$F2,$68,$C3,$63,$F9,$33,$5F,$F0,$07
	.dc.b	$5F,$8E,$BF,$9A,$1F,$7C,$B9,$FF,$88,$63,$CD,$D3,$FE,$97,$8F,$8E
	.dc.b	$DA,$F4,$1F,$D1,$AB,$96,$7F,$66,$3F,$C2,$BC,$5D,$E3,$11,$FD,$3A
	.dc.b	$5F,$BC,$7F,$54,$6C,$7F,$54,$6B,$B8,$FF,$1A,$77,$F3,$24,$F8,$1F
	.dc.b	$3F,$4B,$97,$F3,$57,$F3,$8F,$98,$53,$FC,$3B,$E0,$79,$49,$7E,$A7
	.dc.b	$9C,$F2,$CF,$E0,$57,$26,$BF,$E0,$AA,$F8,$3E,$0F,$BC,$8F,$C4,$EF
	.dc.b	$FD,$2C,$FF,$85,$23,$F5,$2D,$E7,$FF,$0C,$C6,$23,$FC,$6D,$E1,$7C
	.dc.b	$92,$47,$CE,$7C,$BC,$1F,$53,$4F,$80,$FF,$03,$9F,$F2,$9F,$BF,$FE
	.dc.b	$38,$25,$FF,$8E,$21,$1F,$E4,$76,$C7,$F9,$1D,$AF,$F9,$1E,$7F,$25
	.dc.b	$3B,$3F,$CF,$5E,$44,$9F,$F2,$AD,$7F,$B0,$5B,$1F,$E6,$02,$3F,$E8
	.dc.b	$98,$D3,$FE,$89,$8F,$CF,$1F,$E4,$EE,$19,$1E,$76,$BF,$ED,$0F,$4F
	.dc.b	$FD,$BD,$1F,$E7,$F2,$47,$FA,$0D,$FF,$43,$7B,$B2,$3F,$C1,$1E,$67
	.dc.b	$FA,$F6,$31,$1F,$ED,$AF,$1B,$FD,$9D,$1F,$E9,$52,$47,$B1,$FF,$3C
	.dc.b	$91,$F5,$35,$FF,$71,$F9,$DF,$EA,$AE,$AF,$DB,$36,$3F,$C3,$0D,$77
	.dc.b	$E7,$FE,$F6,$7F,$79,$E7,$FF,$AE,$35,$FD,$FD,$7F,$AF,$9A,$FF,$AB
	.dc.b	$FD,$0E,$0E,$6B,$FD,$8B,$E9,$62,$3F,$D9,$AD,$7F,$E0,$07,$A5,$FE
	.dc.b	$CD,$9D,$DF,$49,$0F,$4F,$FD,$FF,$3F,$ED,$BF,$0B,$94,$E5,$B3,$A8
	.dc.b	$F9,$BC,$DA,$E0,$FF,$75,$5F,$FA,$DC,$93,$FE,$9E,$F4,$B9,$F0,$FF
	.dc.b	$A8,$F2,$F1,$4A,$31,$5F,$EF,$EA,$F0,$EC,$F5,$29,$30,$E3,$19,$79
	.dc.b	$DB,$D3,$EE,$48,$CB,$D3,$9A,$47,$26,$E3,$FC,$BF,$9B,$DD,$9F,$E4
	.dc.b	$77,$F8,$71,$88,$F9,$2D,$4B,$C7,$D9,$30,$F7,$C6,$5F,$83,$DC,$EB
	.dc.b	$FE,$35,$8F,$FA,$A8,$98,$7F,$97,$9F,$E8,$5E,$B3,$D6,$FE,$98,$DC
	.dc.b	$3F,$90,$B7,$B7,$73,$A2,$4E,$9B,$1A,$8C,$11,$C1,$2F,$FC,$E5,$5D
	.dc.b	$5B,$5C,$9B,$5C,$5B,$5C,$5F,$C7,$8A,$E4,$E1,$AF,$E5,$1A,$EE,$BB
	.dc.b	$3B,$6B,$B6,$BB,$AF,$13,$B6,$BB,$6B,$BA,$F3,$39,$6B,$93,$D6,$43
	.dc.b	$A3,$AA,$F3,$92,$BC,$6D,$AE,$ED,$9E,$13,$A2,$4E,$9B,$1A,$8C,$11
	.dc.b	$C1,$2B,$C8,$7A,$4A,$7A,$4F,$39,$8A,$F1,$36,$BB,$B6,$BB,$36,$BB
	.dc.b	$3C,$88,$77,$29,$DC,$C1,$0C,$12,$BC,$2F,$06,$BB,$6B,$BA,$F2,$BB
	.dc.b	$6B,$B6,$BB,$AF,$4B,$B6,$BB,$6B,$BA,$F6,$B9,$6B,$93,$E1,$10,$E8
	.dc.b	$EA,$BC,$E4,$AF,$1B,$F9,$AF,$35,$E6,$BC,$FF,$F1,$E6,$1F,$E8,$36
	.dc.b	$CE,$9B,$3B,$9F,$E0,$3F,$35,$FE,$45,$6E,$5C,$B0,$4E,$67,$F6,$ED
	.dc.b	$7F,$5C,$D7,$F5,$CD,$7F,$68,$D8,$E7,$3D,$63,$04,$70,$49,$F1,$90
	.dc.b	$F0,$B9,$69,$93,$C6,$E5,$A6,$4F,$2B,$96,$99,$3C,$EE,$5A,$66,$7D
	.dc.b	$2D,$9F,$43,$67,$84,$EA,$3C,$A4,$3C,$65,$3C,$25,$3B,$95,$E4,$F0
	.dc.b	$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$4F,$BA,$3C,$8D
	.dc.b	$7B,$B5,$E8,$D7,$9B,$63,$9C,$FC,$68,$C1,$1C,$12,$7E,$61,$0F,$0B
	.dc.b	$96,$99,$3C,$6E,$5A,$64,$F2,$B9,$69,$93,$CE,$E5,$A6,$67,$D2,$D9
	.dc.b	$F4,$36,$78,$4E,$A3,$CA,$43,$C6,$53,$C2,$53,$B9,$5E,$4F,$02,$1C
	.dc.b	$0F,$EC,$3A,$FF,$7F,$F1,$7F,$79,$18,$DB,$FD,$35,$97,$FA,$FE,$B4
	.dc.b	$D8,$CC,$FF,$6A,$3F,$C1,$FC,$1F,$EC,$79,$FF,$0C,$36,$35,$C1,$FE
	.dc.b	$0F,$9C,$E5,$E0,$9F,$F2,$0F,$2B,$DA,$E1,$1F,$DD,$A3,$13,$B8,$46
	.dc.b	$2B,$FC,$B3,$5C,$27,$9F,$25,$76,$DA,$34,$C7,$B7,$30,$ED,$28,$C6
	.dc.b	$1F,$E6,$6E,$8E,$44,$AD,$BD,$7D,$C9,$1C,$9B,$3F,$EE,$9E,$B7,$67
	.dc.b	$95,$B5,$ED,$6C,$EE,$75,$CF,$2B,$4D,$9E,$9E,$EF,$32,$4F,$37,$B7
	.dc.b	$B9,$23,$B3,$67,$E4,$C3,$DC,$A4,$DF,$EF,$6D,$F4,$A3,$FD,$0F,$BA
	.dc.b	$B8,$57,$FA,$B6,$73,$18,$8F,$F5,$73,$5F,$BF,$AF,$19,$FF,$48,$B6
	.dc.b	$3B,$B5,$FC,$13,$5F,$F4,$3F,$27,$E6,$7D,$9F,$CA,$73,$1A,$49,$FC
	.dc.b	$B7,$CC,$3E,$FB,$D1,$3F,$99,$94,$E7,$2E,$51,$C2,$B6,$D8,$CB,$F9
	.dc.b	$5E,$5F,$85,$E1,$7C,$A2,$47,$C8,$8C,$BF,$ED,$18,$F8,$A4,$70,$6E
	.dc.b	$7F,$B2,$6D,$FA,$07,$EA,$92,$3C,$5E,$F7,$A7,$1F,$43,$71,$FA,$8D
	.dc.b	$A3,$F1,$AA,$1F,$B9,$D3,$F8,$13,$83,$FD,$67,$2F,$AA,$7C,$37,$F7
	.dc.b	$78,$C3,$63,$31,$F2,$CC,$97,$87,$FD,$5D,$5F,$C1,$E3,$F6,$3A,$73
	.dc.b	$9C,$CF,$F9,$D2,$72,$71,$FB,$DD,$11,$C1,$FE,$18,$C7,$D3,$3E,$BE
	.dc.b	$D7,$0E,$5B,$1E,$D3,$F3,$1E,$DF,$E6,$6F,$D0,$D8,$F3,$35,$F2,$35
	.dc.b	$FE,$4D,$5D,$48,$F5,$E2,$7E,$6A,$FB,$7F,$F0,$F4,$61,$B1,$9C,$3F
	.dc.b	$25,$39,$7E,$47,$91,$FA,$DF,$C3,$3E,$46,$DF,$EC,$B9,$46,$27,$58
	.dc.b	$FC,$9A,$D4,$F0,$E3,$F9,$C4,$BF,$A8,$FF,$8B,$F6,$F9,$37,$EC,$3E
	.dc.b	$30,$72,$FC,$8D,$BF,$05,$B7,$FB,$1E,$FE,$4C,$60,$FB,$9F,$82,$7F
	.dc.b	$05,$F9,$CF,$4B,$6F,$CC,$D7,$F1,$F9,$A3,$11,$FB,$EA,$FB,$2F,$DD
	.dc.b	$E0,$F9,$ED,$3F,$A9,$71,$3F,$19,$7B,$5F,$E5,$5C,$FF,$D7,$F3,$98
	.dc.b	$C5,$04,$66,$7F,$43,$19,$8C,$3F,$E5,$B8,$FE,$93,$F8,$2F,$F2,$1F
	.dc.b	$C0,$71,$A2,$31,$3F,$4D,$B5,$EE,$6B,$ED,$CB,$D6,$FD,$13,$D8,$7F
	.dc.b	$B5,$B5,$FF,$77,$7D,$BF,$F2,$F9,$FF,$36,$46,$2F,$F3,$E6,$FC,$E5
	.dc.b	$6E,$30,$DB,$FD,$EF,$88,$3E,$1B,$FC,$ED,$C4,$FC,$28,$C4,$6A,$38
	.dc.b	$37,$27,$E0,$BF,$D0,$FE,$23,$E1,$B6,$BF,$28,$DB,$FB,$ED,$7E,$11
	.dc.b	$F9,$0E,$4F,$C7,$BF,$91,$E2,$FC,$0B,$E8,$C7,$C4,$79,$9F,$A8,$DB
	.dc.b	$FE,$14,$D8,$F9,$84,$3B,$9F,$38,$C1,$DF,$5F,$A6,$E8,$FC,$31,$FE
	.dc.b	$A5,$1F,$11,$F2,$5F,$C5,$7A,$3F,$EA,$77,$35,$C2,$31,$3B,$9E,$4D
	.dc.b	$8C,$CE,$FD,$1F,$F5,$74,$E5,$B1,$9F,$1F,$F3,$13,$CF,$3F,$F1,$D4
	.dc.b	$FF,$62,$26,$1F,$A7,$6B,$FC,$61,$AF,$E4,$5F,$E6,$87,$E0,$0F,$96
	.dc.b	$7E,$0F,$2F,$96,$D8,$FA,$51,$F4,$5E,$47,$72,$1D,$CF,$C0,$AF,$7B
	.dc.b	$F0,$AD,$32,$E0,$F0,$BD,$4F,$C7,$96,$7C,$CF,$90,$F1,$71,$7E,$8F
	.dc.b	$CB,$FD,$2F,$1F,$E8,$A7,$FC,$5B,$87,$EF,$EB,$51,$FA,$43,$31,$EA
	.dc.b	$3D,$04,$30,$7F,$87,$F8,$BF,$7D,$0F,$B6,$FE,$03,$CB,$FE,$23,$6C
	.dc.b	$64,$FE,$04,$5A,$FE,$19,$D3,$FC,$5A,$3E,$57,$47,$F1,$1D,$3F,$DD
	.dc.b	$0D,$8F,$E3,$33,$B9,$FF,$75,$6D,$F8,$0B,$F9,$E7,$98,$FC,$D4,$7F
	.dc.b	$30,$A7,$C7,$D2,$4F,$F3,$28,$CB,$FD,$7B,$97,$F8,$13,$63,$F3,$0F
	.dc.b	$F4,$16,$8E,$CF,$E8,$57,$F8,$C9,$F0,$9E,$42,$9F,$CF,$BB,$7A,$B6
	.dc.b	$B9,$ED,$FE,$31,$AF,$EB,$11,$8E,$FF,$F2,$77,$E1,$7F,$D7,$E0,$D2
	.dc.b	$5F,$ED,$BC,$1F,$ED,$57,$FE,$95,$6C,$67,$C2,$FE,$E1,$39,$6C,$79
	.dc.b	$F6,$F9,$8F,$E0,$4F,$49,$FE,$08,$C3,$D2,$D7,$E1,$E9,$FE,$0B,$BF
	.dc.b	$F0,$4C,$6A,$78,$92,$BA,$BF,$8C,$8F,$F4,$9F,$89,$D1,$A7,$C4,$71
	.dc.b	$A1,$BC,$32,$7E,$4B,$70,$FF,$17,$78,$BF,$E3,$46,$F8,$DD,$63,$18
	.dc.b	$FF,$8B,$DA,$FB,$F2,$F2,$BE,$7E,$1F,$F1,$5D,$FF,$92,$FE,$06,$4F
	.dc.b	$F9,$8D,$AF,$F9,$8F,$27,$DC,$FF,$A3,$09,$1F,$E5,$D7,$FD,$0B,$AC
	.dc.b	$8F,$F4,$6B,$FA,$D6,$CF,$FA,$D9,$FF,$32,$47,$FA,$58,$8E,$0F,$F3
	.dc.b	$A7,$C1,$7F,$97,$C8,$E0,$FF,$04,$3F,$E8,$62,$38,$3F,$D1,$F2,$F8
	.dc.b	$2F,$F4,$9B,$63,$31,$83,$FD,$3A,$47,$CE,$D7,$E1,$3F,$EA,$C2,$3F
	.dc.b	$E7,$D7,$FD,$85,$E8,$FF,$AA,$7D,$AB,$46,$92,$BF,$DD,$DE,$A3,$D4
	.dc.b	$E1,$18,$6B,$FD,$5A,$35,$5C,$E3,$4D,$75,$AF,$07,$FC,$45,$3F,$E0
	.dc.b	$0B,$FF,$6B,$4E,$65,$39,$7C,$7E,$06,$63,$FD,$B3,$5F,$ED,$7B,$ED
	.dc.b	$5D,$2B,$97,$A2,$F4,$E9,$2B,$75,$FE,$FC,$8D,$FA,$B2,$73,$48,$C6
	.dc.b	$1F,$EB,$98,$FF,$7B,$4E,$BD,$6E,$E4,$78,$F0,$78,$88,$E0,$EB,$EA
	.dc.b	$F9,$88,$E0,$F2,$FA,$EF,$63,$E2,$36,$33,$18,$3E,$47,$3B,$9F,$FD
	.dc.b	$27,$9F,$FB,$58,$9B,$7F,$96,$DF,$F6,$97,$B2,$F6,$BF,$84,$37,$1F
	.dc.b	$E6,$34,$8C,$71,$F0,$AD,$92,$B7,$3A,$AF,$F0,$C9,$C6,$75,$5C,$6B
	.dc.b	$71,$89,$DD,$71,$7F,$C4,$2D,$7F,$CD,$0D,$7A,$4F,$22,$57,$52,$47
	.dc.b	$4A,$EC,$F4,$6B,$D1,$AF,$48,$E2,$EE,$7A,$BC,$DA,$E6,$36,$F3,$8C
	.dc.b	$B6,$32,$F5,$39,$38,$AF,$61,$2B,$C6,$DA,$E7,$5C,$4F,$01,$C8,$86
	.dc.b	$08,$78,$4C,$94,$C9,$CC,$C1,$0C,$12,$3C,$55,$E7,$AF,$29,$2B,$CB
	.dc.b	$3A,$AE,$E7,$19,$D5,$71,$AD,$C6,$27,$75,$C5,$F2,$B5,$F2,$35,$E9
	.dc.b	$3C,$89,$5D,$49,$1D,$2B,$B3,$D1,$AF,$46,$BD,$23,$8B,$B9,$EA,$F3
	.dc.b	$6B,$98,$DB,$CE,$32,$D8,$CB,$D4,$E4,$E2,$BD,$84,$AF,$1B,$6B,$9D
	.dc.b	$71,$3C,$07,$23,$F5,$FE,$77,$FA,$0D,$BE,$73,$CE,$FD,$9C,$7F,$A2
	.dc.b	$23,$0E,$27,$83,$FB,$0F,$91,$5F,$08,$FE,$05,$18,$7F,$84,$B5,$FD
	.dc.b	$CB,$5E,$51,$C7,$04,$E6,$7F,$76,$D7,$AC,$EE,$31,$3C,$23,$04,$9F
	.dc.b	$0B,$5F,$0B,$5F,$09,$A2,$1A,$29,$A3,$04,$30,$43,$CE,$F5,$69,$C8
	.dc.b	$A7,$12,$9C,$4F,$69,$B2,$9B,$29,$B8,$F8,$0E,$DA,$ED,$AE,$E7,$E3
	.dc.b	$92,$7C,$D1,$D0,$F1,$90,$C1,$27,$9B,$96,$B9,$9E,$8F,$8D,$AF,$8D
	.dc.b	$AF,$8E,$7C,$33,$C8,$93,$A9,$EC,$D7,$B4,$F0,$8C,$4F,$08,$C1,$27
	.dc.b	$C4,$D7,$C4,$D7,$C4,$68,$86,$8A,$68,$C1,$0C,$10,$F3,$BD,$5A,$72
	.dc.b	$29,$C4,$A7,$13,$DA,$6C,$A6,$CA,$6E,$3E,$03,$B6,$BB,$6B,$B9,$F8
	.dc.b	$E4,$9F,$34,$74,$3C,$67,$E0,$75,$FF,$03,$7A,$1F,$C0,$A0,$48,$CE
	.dc.b	$5F,$C2,$1B,$1F,$DF,$1A,$F0,$1F,$22,$EB,$FE,$13,$5F,$C4,$23,$18
	.dc.b	$77,$AE,$F1,$89,$D5,$70,$9F,$F0,$4F,$33,$93,$11,$FE,$7F,$AF,$F1
	.dc.b	$94,$62,$3D,$33,$BC,$3B,$47,$93,$E6,$F4,$D8,$CE,$29,$CB,$63,$8D
	.dc.b	$7F,$8F,$B0,$FF,$20,$73,$F2,$8C,$61,$E6,$8E,$3F,$41,$2F,$AB,$6F
	.dc.b	$FC,$FA,$4A,$FF,$40,$B5,$F9,$5F,$6D,$F0,$DF,$F1,$4B,$5C,$E9,$EA
	.dc.b	$8C,$47,$CA,$E6,$F4,$93,$0F,$6B,$BD,$24,$7C,$4A,$F7,$61,$FD,$11
	.dc.b	$B7,$FE,$38,$D3,$EB,$FB,$FF,$52,$CD,$24,$FD,$6D,$BF,$A3,$C7,$D4
	.dc.b	$9D,$CE,$5A,$FF,$AD,$E3,$F0,$9A,$70,$8F,$03,$5F,$C6,$5A,$31,$1F
	.dc.b	$88,$6B,$FE,$79,$AE,$AF,$E0,$BB,$33,$BF,$FA,$06,$73,$3B,$C3,$E0
	.dc.b	$6D,$CB,$3F,$7C,$FD,$08,$C1,$DB,$2F,$CC,$F8,$DF,$0C,$93,$F1,$E7
	.dc.b	$FB,$3C,$7C,$12,$47,$C1,$6B,$EC,$8F,$ED,$B8,$7C,$D2,$47,$37,$FB
	.dc.b	$36,$3F,$9B,$7E,$61,$23,$E6,$3F,$B5,$DD,$C5,$FD,$DF,$8B,$F7,$B5
	.dc.b	$98,$C4,$EA,$3F,$CE,$B4,$FD,$51,$F1,$BF,$E0,$0F,$67,$31,$89,$46
	.dc.b	$27,$55,$FC,$06,$79,$46,$23,$F8,$5C,$3B,$DF,$6F,$FE,$01,$8C,$36
	.dc.b	$33,$1F,$E1,$63,$25,$5E,$4F,$5A,$FF,$0A,$C6,$27,$CB,$1E,$38,$C6
	.dc.b	$1E,$E6,$C7,$F1,$9C,$BE,$13,$73,$F8,$A4,$7E,$0F,$81,$DF,$0F,$8C
	.dc.b	$DD,$3F,$D6,$58,$EA,$7C,$35,$FC,$8F,$5F,$A2,$F5,$3E,$91,$0F,$91
	.dc.b	$F9,$3B,$87,$B3,$6B,$9C,$FF,$8C,$A3,$17,$F6,$1B,$1E,$FE,$EF,$F1
	.dc.b	$BF,$E7,$3E,$F6,$C6,$6D,$39,$6C,$7C,$7F,$BC,$FA,$44,$8F,$A0,$FF
	.dc.b	$8D,$72,$F5,$CF,$B7,$6F,$92,$DD,$BE,$43,$EC,$F5,$3D,$64,$C3,$D6
	.dc.b	$33,$47,$CD,$E0,$79,$9F,$E5,$0F,$84,$FB,$0D,$C3,$EE,$35,$FC,$6C
	.dc.b	$F0,$C1,$5F,$36,$7C,$85,$3F,$AE,$65,$E8,$7E,$67,$B3,$F4,$09,$5E
	.dc.b	$6C,$91,$E5,$F9,$4F,$F3,$1F,$EC,$5F,$E5,$FA,$FC,$E4,$61,$FC,$83
	.dc.b	$6F,$CA,$D7,$C3,$AF,$C6,$7C,$A7,$E1,$4F,$F7,$3F,$CA,$BE,$FF,$F8
	.dc.b	$5D,$E1,$B3,$FD,$BE,$75,$3F,$87,$8C,$4E,$9F,$F3,$CF,$AD,$FC,$A6
	.dc.b	$7F,$67,$1F,$B4,$8C,$3E,$B6,$C7,$EE,$DB,$1F,$BA,$F2,$FF,$8B,$F3
	.dc.b	$7C,$16,$D7,$F0,$16,$CE,$E3,$F8,$51,$ED,$7D,$E4,$3E,$19,$FC,$4A
	.dc.b	$BB,$BE,$8F,$0F,$FD,$1D,$3E,$36,$C7,$A5,$AF,$23,$B1,$4E,$C3,$93
	.dc.b	$F3,$2E,$48,$67,$83,$F0,$8B,$CA,$74,$7F,$A9,$BE,$05,$9F,$F8,$7E
	.dc.b	$73,$E3,$FF,$8A,$27,$37,$FE,$AD,$8C,$56,$EF,$93,$63,$33,$B1,$5C
	.dc.b	$9F,$15,$70,$75,$19,$C7,$FC,$73,$7F,$CF,$49,$3F,$C6,$DA,$FE,$D5
	.dc.b	$AF,$E4,$9F,$A1,$97,$FA,$F9,$FC,$13,$73,$FA,$07,$C0,$2C,$7F,$AF
	.dc.b	$6B,$C3,$5D,$C8,$77,$3F,$09,$3F,$EC,$AF,$81,$FC,$43,$4C,$87,$2D
	.dc.b	$3B,$92,$3F,$00,$F8,$DB,$5F,$EC,$FE,$0F,$D1,$F3,$DE,$1F,$E8,$B7
	.dc.b	$FA,$93,$FA,$91,$63,$D5,$BF,$F6,$8B,$F8,$B8,$7E,$0F,$8B,$E9,$3F
	.dc.b	$1C,$FD,$ED,$BF,$61,$FE,$05,$83,$93,$FD,$CF,$18,$6C,$F0,$8F,$E1
	.dc.b	$1C,$FF,$C3,$23,$E0,$5F,$F4,$2E,$7F,$8C,$D8,$CB,$FD,$AA,$78,$6D
	.dc.b	$E6,$DF,$E8,$CF,$CF,$21,$FD,$93,$A7,$FD,$F3,$EB,$FF,$2A,$C0,$D2
	.dc.b	$4F,$F3,$78,$EE,$FF,$4F,$EB,$F1,$BE,$16,$B9,$8D,$BF,$1F,$7F,$27
	.dc.b	$07,$CB,$8F,$C1,$9F,$DD,$BB,$3F,$1C,$FF,$4F,$AF,$9A,$7F,$81,$63
	.dc.b	$31,$FC,$F1,$AF,$F7,$EE,$64,$BF,$35,$FE,$B4,$49,$FE,$BC,$FF,$65
	.dc.b	$EF,$F3,$CF,$F4,$F6,$EF,$FD,$B7,$7D,$4F,$AB,$7F,$97,$87,$DF,$3E
	.dc.b	$59,$ED,$3F,$C3,$DE,$03,$C1,$EA,$DA,$F9,$27,$F8,$18,$AE,$6F,$E2
	.dc.b	$F3,$7E,$32,$FA,$1F,$62,$BF,$C4,$DE,$17,$A3,$9F,$FC,$39,$7F,$E1
	.dc.b	$98,$C1,$9F,$0F,$E2,$D7,$F8,$9B,$8B,$FC,$3D,$18,$BF,$F1,$5C,$7F
	.dc.b	$8D,$E7,$C7,$E2,$7A,$08,$FC,$1E,$3F,$B1,$97,$C0,$DB,$E1,$BE,$63
	.dc.b	$FD,$5B,$E5,$FD,$2E,$7F,$F3,$F1,$23,$FC,$FC,$FF,$A1,$C8,$FF,$96
	.dc.b	$DA,$FF,$B1,$FC,$6F,$F5,$B6,$7F,$E8,$72,$47,$FA,$1D,$FF,$48,$11
	.dc.b	$FF,$32,$35,$FF,$51,$E2,$F5,$FF,$CF,$6D,$8F,$F3,$D3,$5D,$C7,$FB
	.dc.b	$5D,$FF,$43,$46,$5B,$19,$7F,$D1,$8D,$AF,$F4,$53,$FE,$98,$9E,$13
	.dc.b	$FE,$E0,$EF,$FC,$AC,$E4,$91,$FE,$A1,$7F,$D4,$84,$70,$7F,$9F,$DF
	.dc.b	$F5,$71,$1C,$1F,$EC,$2F,$37,$F4,$3E,$97,$FA,$DD,$B1,$9B,$4E,$5B
	.dc.b	$1F,$EB,$BD,$76,$FF,$AB,$1A,$FF,$B2,$63,$6E,$5A,$87,$99,$FE,$FA
	.dc.b	$AF,$9E,$48,$F9,$EF,$C7,$F3,$BA,$C7,$FA,$D6,$31,$5D,$23,$07,$FB
	.dc.b	$01,$C7,$9B,$FE,$17,$9F,$9C,$49,$FF,$77,$4F,$1D,$1E,$7F,$66,$C6
	.dc.b	$63,$C3,$7F,$EF,$BB,$FB,$8D,$BF,$F6,$F6,$7A,$39,$70,$4C,$3B,$36
	.dc.b	$FF,$DC,$44,$AF,$F5,$B3,$5F,$F3,$13,$FE,$AF,$23,$FE,$8D,$7F,$19
	.dc.b	$E8,$BD,$1F,$E3,$B2,$8C,$5F,$F0,$56,$F2,$F0,$AD,$C6,$2B,$F3,$4D
	.dc.b	$7E,$B1,$CE,$36,$4A,$E8,$DA,$E9,$18,$23,$82,$4F,$69,$E0,$D9,$DB
	.dc.b	$67,$6F,$8A,$BC,$8F,$E2,$5A,$F9,$67,$8B,$BA,$EA,$D8,$CB,$C6,$32
	.dc.b	$D8,$CB,$CE,$BC,$2E,$5A,$E4,$F3,$10,$EC,$EA,$BC,$A4,$AE,$F3,$C4
	.dc.b	$D9,$C4,$EC,$53,$46,$0D,$90,$C1,$0D,$92,$3C,$15,$EC,$6D,$79,$6B
	.dc.b	$C7,$18,$AF,$03,$5E,$C7,$38,$D9,$2B,$A3,$6B,$A4,$60,$8E,$09,$3D
	.dc.b	$A7,$83,$67,$6D,$9D,$BE,$2A,$F2,$3E,$F6,$BE,$59,$E2,$EE,$BA,$B6
	.dc.b	$32,$F1,$8C,$B6,$32,$F3,$AF,$0B,$96,$B9,$3C,$C4,$3B,$3A,$AF,$29
	.dc.b	$2B,$BC,$F1,$36,$71,$3B,$14,$FB,$7E,$57,$FA,$11,$FE,$94,$D7,$FA
	.dc.b	$83,$5F,$F4,$11,$FC,$13,$D4,$FF,$0E,$B7,$0F,$EA,$4F,$F0,$58,$CB
	.dc.b	$FC,$01,$AF,$EB,$DA,$FE,$BD,$AF,$18,$E7,$69,$CB,$63,$31,$82,$38
	.dc.b	$24,$F9,$48,$78,$DC,$B4,$C9,$E5,$72,$D3,$27,$9D,$CB,$4C,$9E,$92
	.dc.b	$CF,$9D,$B3,$E6,$6C,$F2,$78,$1E,$32,$1E,$12,$9D,$CA,$70,$34,$43
	.dc.b	$45,$34,$60,$86,$09,$3E,$77,$AB,$5E,$4D,$78,$B5,$E3,$3E,$D7,$8B
	.dc.b	$5E,$2D,$78,$B5,$E3,$3F,$0E,$7A,$12,$74,$D8,$D4,$60,$8E,$09,$3F
	.dc.b	$38,$87,$91,$CB,$4C,$9E,$67,$2D,$32,$7A,$1C,$B4,$C9,$EA,$2C,$F9
	.dc.b	$DB,$3E,$66,$CF,$27,$81,$E3,$21,$E1,$29,$DC,$A7,$03,$44,$34,$53
	.dc.b	$46,$08,$60,$93,$E7,$7F,$CA,$1C,$CE,$27,$5B,$CB,$FF,$28,$4E,$63
	.dc.b	$15,$C2,$B7,$5A,$6C,$7F,$96,$FD,$4F,$F4,$94,$FF,$96,$5B,$19,$EF
	.dc.b	$FF,$54,$CE,$5B,$1F,$E6,$3E,$47,$0E,$23,$FC,$EC,$D7,$FD,$03,$A7
	.dc.b	$91,$CB,$5F,$F2,$34,$FF,$BB,$AF,$FC,$D1,$D3,$C6,$53,$97,$C1,$19
	.dc.b	$8F,$F1,$CC,$FF,$9B,$F9,$70,$4B,$ED,$7C,$B7,$E0,$48,$CD,$FF,$1B
	.dc.b	$BF,$F3,$D7,$3F,$C2,$BF,$F1,$2B,$5F,$F2,$E3,$5F,$F2,$EE,$7E,$7F
	.dc.b	$59,$EB,$52,$63,$E2,$8C,$B7,$97,$81,$1C,$1E,$8E,$2F,$B9,$EA,$FD
	.dc.b	$9E,$33,$49,$97,$D7,$AC,$B6,$3E,$C4,$62,$3F,$C7,$FC,$5F,$ED,$19
	.dc.b	$F9,$13,$99,$FA,$7C,$1F,$EE,$89,$C9,$23,$F0,$1C,$19,$6D,$71,$7E
	.dc.b	$73,$78,$3F,$A3,$57,$CC,$DB,$85,$7C,$C6,$BF,$31,$AF,$CC,$AE,$99
	.dc.b	$FE,$77,$C4,$F8,$44,$9F,$CF,$68,$8F,$79,$23,$E2,$35,$FF,$2F,$CF
	.dc.b	$E9,$EB,$E6,$12,$39,$BF,$20,$8F,$CD,$6B,$CB,$93,$F6,$6F,$D5,$21
	.dc.b	$E4,$3F,$D9,$10,$E3,$7E,$12,$87,$D7,$6A,$73,$18,$AF,$F3,$AB,$C6
	.dc.b	$B8,$56,$DB,$3A,$A7,$C3,$FF,$7C,$BF,$03,$63,$31,$FE,$D7,$32,$55
	.dc.b	$E8,$79,$7E,$AF,$C8,$DB,$EC,$D7,$FD,$7D,$5E,$D8,$C4,$7F,$B0,$73
	.dc.b	$FE,$28,$FC,$43,$E4,$6B,$FE,$16,$8C,$C7,$A7,$F0,$1E,$37,$F1,$DB
	.dc.b	$89,$A7,$69,$57,$27,$A9,$DC,$F1,$6B,$F5,$9F,$8D,$D1,$F2,$1F,$4D
	.dc.b	$FF,$B5,$B0,$FA,$11,$89,$D3,$F5,$E7,$5B,$7D,$7F,$9D,$FB,$5F,$BD
	.dc.b	$5C,$0D,$25,$FD,$5F,$96,$E7,$3F,$83,$6E,$06,$9E,$09,$F6,$E5,$E9
	.dc.b	$7D,$A7,$E3,$48,$7A,$BE,$47,$99,$2F,$D7,$1E,$98,$F0,$FC,$47,$D8
	.dc.b	$F3,$BF,$C8,$B4,$8C,$3F,$90,$6D,$7D,$E6,$BF,$33,$83,$ED,$61,$E5
	.dc.b	$8C,$1F,$47,$95,$F9,$4B,$E6,$3E,$5C,$61,$B1,$9C,$7C,$F3,$97,$D2
	.dc.b	$79,$6B,$F5,$1F,$61,$E8,$CB,$F2,$2D,$8C,$CF,$A5,$A7,$C8,$21,$E6
	.dc.b	$3E,$49,$FB,$39,$FE,$DB,$0F,$88,$ED,$9F,$ED,$DB,$19,$3E,$34,$65
	.dc.b	$FD,$BF,$90,$DB,$F5,$51,$86,$E1,$FA,$89,$D4,$BA,$3E,$9C,$61,$F7
	.dc.b	$7D,$AF,$F0,$09,$FA,$8D,$AF,$A8,$D7,$D4,$D7,$F1,$38,$7D,$62,$3F
	.dc.b	$A5,$97,$9B,$F0,$39,$BF,$06,$D8,$FE,$22,$D9,$FD,$FB,$5F,$94,$7C
	.dc.b	$C3,$B9,$FB,$12,$9F,$38,$A6,$FC,$9F,$F5,$07,$BC,$F1,$7F,$D3,$AD
	.dc.b	$90,$9D,$CE,$89,$1A,$8C,$B7,$D7,$FF,$11,$D7,$38,$E5,$18,$AF,$E2
	.dc.b	$67,$F6,$D2,$47,$F4,$C6,$BA,$30,$43,$E1,$E5,$FC,$EF,$EA,$7E,$4B
	.dc.b	$72,$F9,$2D,$9D,$C2,$7F,$D6,$7D,$BF,$E3,$9B,$EF,$5F,$AD,$6D,$7E
	.dc.b	$D7,$5E,$53,$F7,$A3,$91,$F8,$42,$BF,$84,$6B,$FD,$3F,$7F,$C7,$FD
	.dc.b	$5F,$E3,$B8,$7F,$A9,$36,$8D,$3F,$3F,$FB,$F4,$8F,$CF,$87,$17,$F1
	.dc.b	$48,$7A,$4F,$8C,$7E,$B4,$FE,$C1,$F5,$DF,$B7,$93,$48,$FE,$02,$3F
	.dc.b	$BE,$74,$7F,$05,$D7,$11,$8E,$9F,$A0,$EF,$A7,$F8,$44,$E9,$B1,$E2
	.dc.b	$3F,$47,$B7,$CE,$69,$FC,$1F,$07,$EC,$DE,$52,$39,$7F,$93,$F7,$7E
	.dc.b	$E3,$0F,$EF,$1B,$78,$AF,$F9,$95,$FD,$D7,$FB,$A4,$65,$F8,$87,$C6
	.dc.b	$3C,$27,$C8,$8F,$E8,$87,$17,$FA,$AF,$77,$18,$CD,$7F,$81,$E3,$16
	.dc.b	$F0,$3F,$C1,$9F,$BF,$7F,$53,$26,$1F,$E0,$58,$CB,$FD,$8B,$0E,$CF
	.dc.b	$F5,$DF,$0B,$99,$23,$31,$E6,$3D,$24,$30,$FF,$77,$3F,$12,$FF,$7B
	.dc.b	$6B,$FD,$BB,$C3,$FC,$7B,$9E,$3F,$A5,$CD,$FA,$18,$FA,$CD,$7F,$BA
	.dc.b	$F8,$9C,$1B,$7D,$1F,$EE,$91,$96,$C6,$4F,$F0,$EF,$07,$F8,$B7,$C5
	.dc.b	$FF,$18,$36,$3F,$82,$4F,$F7,$D9,$CB,$FE,$2B,$F1,$BE,$69,$27,$E3
	.dc.b	$9F,$E3,$33,$E2,$E1,$E7,$9F,$88,$D7,$E4,$0F,$31,$E4,$7F,$90,$7D
	.dc.b	$7F,$F3,$A9,$27,$FC,$F4,$FF,$9F,$C8,$FF,$97,$9A,$FF,$9E,$A3,$FC
	.dc.b	$28,$FF,$A0,$78,$7F,$CF,$E4,$7F,$CC,$AF,$FA,$24,$8E,$0F,$D5,$EF
	.dc.b	$E4,$C7,$FA,$1C,$62,$BF,$CA,$DE,$0F,$F9,$E6,$73,$3F,$E7,$96,$BF
	.dc.b	$EB,$AF,$33,$6F,$FA,$12,$7F,$6E,$FF,$A1,$1B,$E6,$27,$73,$FB,$F8
	.dc.b	$79,$B8,$8F,$F4,$9D,$F6,$7F,$D8,$71,$96,$C6,$74,$F5,$61,$FE,$A0
	.dc.b	$24,$FF,$A9,$5F,$F5,$51,$1F,$F4,$4B,$5F,$E2,$1E,$77,$F5,$3A,$78
	.dc.b	$1E,$A8,$C3,$78,$BF,$96,$57,$F1,$49,$CB,$7C,$FC,$BD,$1A,$FF,$5C
	.dc.b	$8F,$F6,$1D,$75,$9C,$CF,$FA,$86,$3E,$81,$1F,$A9,$7E,$47,$FD,$63
	.dc.b	$1F,$33,$4E,$3A,$7C,$DA,$D5,$7C,$D9,$CE,$92,$73,$3F,$3A,$FF,$D8
	.dc.b	$67,$FB,$CA,$33,$1F,$EF,$2D,$FA,$92,$33,$83,$D1,$F0,$92,$32,$FB
	.dc.b	$36,$FC,$76,$1F,$ED,$42,$61,$FE,$D9,$7F,$DB,$C4,$7F,$D7,$AD,$7F
	.dc.b	$A2,$7A,$5E,$DD,$BF,$1A,$D8,$FE,$8A,$D7,$7A,$7E,$F1,$B8,$FF,$00
	.dc.b	$53,$B7,$FC,$1E,$D7,$33,$C5,$CB,$5C,$CF,$39,$E0,$49,$D3,$63,$51
	.dc.b	$82,$38,$25,$7B,$23,$2D,$8C,$D7,$BA,$72,$D8,$CB,$EE,$9E,$04,$74
	.dc.b	$D7,$46,$08,$60,$87,$C0,$8E,$AD,$8E,$4D,$38,$94,$E2,$7C,$67,$6D
	.dc.b	$36,$53,$75,$F1,$E3,$6D,$8D,$B5,$DD,$7C,$26,$D7,$99,$E8,$78,$C8
	.dc.b	$60,$93,$CD,$CB,$5C,$CF,$57,$2D,$73,$3D,$E7,$81,$27,$4D,$8D,$46
	.dc.b	$08,$E0,$95,$F0,$23,$2D,$8C,$D7,$C2,$9C,$B6,$32,$FC,$29,$E0,$47
	.dc.b	$4D,$74,$60,$86,$08,$7C,$08,$EA,$D8,$E4,$D3,$89,$4E,$27,$C6,$76
	.dc.b	$D3,$65,$37,$5F,$1E,$3F,$0F,$E4,$3C,$77,$A5,$FD,$32,$FF,$CF,$6D
	.dc.b	$9D,$36,$77,$3F,$C1,$B7,$63,$C7,$1F,$E9,$ED,$8C,$CE,$E3,$04,$70
	.dc.b	$43,$F8,$7C,$70,$6C,$6D,$A6,$CF,$E3,$31,$B6,$C6,$DA,$6C,$FE,$27
	.dc.b	$1B,$6C,$6D,$A6,$CF,$E3,$71,$B6,$C6,$DA,$6E,$7E,$0B,$67,$84,$62
	.dc.b	$3D,$24,$3C,$A5,$3C,$25,$3A,$95,$E4,$F0,$21,$C0,$A7,$02,$BC,$1D
	.dc.b	$10,$D1,$5D,$38,$23,$82,$47,$B0,$93,$F0,$E3,$2D,$8C,$CF,$C3,$8C
	.dc.b	$11,$C1,$0F,$74,$70,$6C,$6D,$A6,$CF,$6C,$6D,$B1,$B6,$9B,$3D,$D1
	.dc.b	$B6,$C6,$DA,$6C,$F8,$31,$B6,$C6,$DA,$6E,$7E,$0B,$67,$84,$62,$3D
	.dc.b	$24,$3C,$A5,$3C,$25,$3A,$95,$E4,$F0,$21,$C0,$A7,$02,$BC,$1D,$10
	.dc.b	$D1,$5D,$38,$23,$8E,$67,$9E,$F2,$7F,$81,$D7,$FB,$36,$73,$18,$80
	.dc.b	$8C,$E5,$FE,$5A,$6C,$7F,$B4,$BD,$57,$C5,$7F,$0C,$AF,$F2,$FC,$63
	.dc.b	$3E,$94,$8C,$4E,$AB,$84,$FF,$B7,$38,$FF,$C5,$57,$C1,$B5,$FE,$68
	.dc.b	$E3,$DD,$7F,$99,$DA,$FA,$30,$7A,$D8,$8F,$03,$63,$38,$79,$1B,$7F
	.dc.b	$E7,$4D,$B8,$63,$FC,$92,$31,$5E,$39,$E9,$CE,$F9,$EA,$49,$F7,$C6
	.dc.b	$5F,$57,$B1,$D0,$8E,$0F,$8B,$D9,$F1,$34,$FF,$49,$ED,$E9,$8C,$47
	.dc.b	$D4,$0E,$1F,$AE,$FA,$9F,$F4,$DE,$BC,$49,$5F,$4B,$7F,$BF,$EB,$FD
	.dc.b	$BE,$2F,$C2,$37,$88,$D3,$ED,$DF,$F4,$86,$DE,$B0,$FC,$14,$E5,$B1
	.dc.b	$97,$FD,$39,$B6,$5B,$19,$7F,$D3,$B5,$F3,$8F,$F4,$76,$DB,$8F,$F4
	.dc.b	$33,$56,$DF,$36,$77,$5C,$C7,$A8,$FB,$AF,$EB,$FF,$25,$F2,$70,$FE
	.dc.b	$C3,$39,$A4,$E5,$E3,$5C,$2B,$E8,$46,$0D,$BC,$F6,$FD,$4F,$9B,$F2
	.dc.b	$49,$19,$DF,$F6,$11,$D7,$17,$CC,$7F,$72,$FC,$7F,$0D,$BE,$F3,$AC
	.dc.b	$55,$A6,$CF,$5E,$1F,$EC,$99,$7E,$EA,$7E,$D1,$27,$47,$DD,$21,$83
	.dc.b	$BF,$C8,$7F,$81,$3D,$7F,$33,$63,$32,$9C,$B6,$3C,$FE,$EF,$F7,$58
	.dc.b	$FA,$2D,$C8,$D3,$1E,$9F,$F8,$4B,$D9,$F8,$8D,$BF,$82,$D7,$F8,$9D
	.dc.b	$FD,$C8,$C4,$7F,$16,$CB,$F8,$CB,$EC,$3F,$8E,$6F,$FC,$7D,$C0,$DD
	.dc.b	$F4,$0F,$C0,$39,$C6,$1B,$19,$C5,$97,$E2,$6B,$51,$8B,$F9,$47,$83
	.dc.b	$7F,$F1,$A6,$5E,$48,$C9,$F9,$0E,$BF,$98,$FB,$36,$46,$23,$57,$F3
	.dc.b	$5D,$46,$63,$F3,$BE,$27,$F8,$EB,$F4,$FF,$C6,$36,$BF,$12,$D7,$B3
	.dc.b	$5F,$D0,$F5,$73,$C3,$FC,$83,$C5,$F9,$58,$C1,$E3,$9F,$E8,$5C,$1F
	.dc.b	$2D,$F5,$98,$D3,$DF,$19,$8F,$67,$91,$D4,$99,$FA,$E3,$F6,$31,$E1
	.dc.b	$F0,$BF,$CA,$7F,$82,$7F,$92,$6B,$F7,$31,$87,$F4,$6D,$AF,$90,$D7
	.dc.b	$F0,$57,$E1,$6E,$1E,$42,$3F,$67,$E7,$1E,$7F,$E7,$F0,$FE,$59,$C4
	.dc.b	$9C,$C7,$E7,$CC,$B8,$D7,$FC,$BD,$3B,$AE,$57,$E4,$B7,$EA,$5E,$BD
	.dc.b	$FF,$AD,$B7,$4F,$F3,$8B,$67,$67,$CC,$E3,$FC,$45,$FE,$2C,$FC,$CC
	.dc.b	$7F,$1C,$E6,$FF,$3C,$FC,$4F,$56,$C7,$F0,$86,$CF,$E3,$5A,$FF,$20
	.dc.b	$FA,$EF,$D8,$4F,$F0,$E6,$C7,$E4,$DA,$EE,$3F,$42,$7F,$22,$F3,$5F
	.dc.b	$41,$F0,$27,$FD,$17,$18,$9F,$91,$1F,$B8,$8C,$6F,$F7,$5B,$3E,$FE
	.dc.b	$2F,$B1,$5F,$8B,$3D,$FB,$3F,$04,$ED,$5F,$CF,$1B,$19,$8E,$F3,$FC
	.dc.b	$F2,$7C,$8F,$33,$B7,$1F,$E0,$9A,$7F,$04,$34,$7F,$6D,$F3,$BF,$D5
	.dc.b	$2B,$E0,$DE,$86,$7D,$0F,$F5,$5C,$62,$7F,$C4,$F5,$C2,$31,$1F,$EB
	.dc.b	$43,$95,$71,$F3,$7F,$C5,$B3,$AA,$E1,$39,$AD,$9D,$2F,$BC,$7F,$0D
	.dc.b	$3F,$BB,$12,$3F,$A1,$35,$D1,$82,$1F,$C3,$F1,$FF,$60,$FB,$BF,$95
	.dc.b	$6E,$7F,$D4,$1B,$3B,$84,$FF,$AD,$F9,$3F,$C8,$F7,$DE,$BF,$5E,$DA
	.dc.b	$6F,$F8,$3B,$FC,$07,$0F,$4B,$FC,$3B,$6B,$6D,$DD,$C5,$FE,$D6,$F6
	.dc.b	$7F,$54,$DA,$FC,$C0,$8C,$B4,$FD,$10,$FC,$9F,$F4,$CD,$32,$E0,$F6
	.dc.b	$71,$FF,$99,$74,$F6,$65,$FC,$75,$F6,$63,$FE,$54,$FD,$47,$F7,$04
	.dc.b	$C3,$F8,$A8,$7F,$82,$F6,$7F,$0A,$D6,$F5,$FC,$1E,$CF,$E1,$17,$F3
	.dc.b	$0F,$E1,$B1,$F1,$72,$F0,$1F,$60,$F9,$A5,$8F,$E4,$A3,$F0,$5F,$E1
	.dc.b	$44,$7C,$6F,$F1,$1E,$FF,$BF,$3F,$E0,$C6,$BF,$CD,$A7,$EE,$B9,$6B
	.dc.b	$F8,$A2,$1E,$22,$BF,$E2,$EF,$03,$9C,$7F,$4C,$8E,$2D,$C7,$FA,$64
	.dc.b	$60,$FE,$B2,$3F,$4F,$FE,$B2,$D9,$FE,$B2,$D7,$FA,$CB,$5F,$EC,$B3
	.dc.b	$FE,$2B,$F1,$3C,$33,$FE,$40,$6E,$DF,$DB,$23,$C7,$5F,$30,$F5,$94
	.dc.b	$F6,$1F,$E0,$5F,$1B,$FC,$01,$C7,$FD,$FD,$B1,$FD,$19,$AF,$FA,$9B
	.dc.b	$6F,$F0,$1C,$FE,$23,$9F,$F4,$B3,$F6,$4E,$D7,$FE,$19,$3F,$30,$FF
	.dc.b	$87,$9A,$FF,$80,$FC,$7F,$D0,$39,$AF,$F2,$47,$90,$AF,$E1,$B3,$FE
	.dc.b	$5A,$8C,$72,$6A,$72,$7F,$8C,$8B,$9B,$C9,$F8,$A4,$8F,$A9,$19,$7E
	.dc.b	$3E,$5E,$AC,$3E,$3B,$5F,$8E,$7F,$80,$7D,$BF,$F4,$41,$1F,$F2,$A3
	.dc.b	$FE,$57,$FC,$8A,$47,$F9,$73,$CD,$F5,$46,$23,$FC,$51,$C9,$FE,$91
	.dc.b	$23,$FE,$5D,$7F,$CD,$1E,$77,$C1,$FA,$0B,$91,$A4,$AF,$F3,$F7,$A5
	.dc.b	$FE,$B1,$F7,$3F,$D0,$A4,$8C,$FA,$9F,$2E,$3F,$D7,$A4,$8F,$F6,$97
	.dc.b	$4E,$5B,$82,$7F,$D4,$24,$8F,$F4,$FB,$5F,$F2,$13,$FE,$AB,$23,$FE
	.dc.b	$77,$7F,$D2,$4F,$FA,$EB,$D5,$FE,$D2,$FF,$AE,$27,$F9,$4B,$7C,$3D
	.dc.b	$FA,$DF,$ED,$86,$FA,$FC,$27,$FD,$89,$3F,$EE,$F9,$CD,$72,$9C,$D6
	.dc.b	$DF,$F7,$EF,$B1,$FA,$5A,$F1,$B6,$3F,$DE,$5D,$9B,$51,$97,$FD,$A5
	.dc.b	$AF,$FB,$76,$BF,$D6,$24,$8F,$9B,$87,$85,$EF,$A7,$3B,$34,$93,$B9
	.dc.b	$FF,$7B,$B7,$D9,$F8,$11,$FE,$CC,$BF,$F6,$89,$E8,$F4,$BA,$12,$B9
	.dc.b	$60,$8F,$55,$FC,$5B,$FF,$61,$35,$FF,$76,$35,$07,$0F,$F9,$7E,$FF
	.dc.b	$DC,$C4,$7F,$D6,$4F,$DC,$F6,$BD,$B8,$7E,$DA,$51,$8E,$AD,$65,$FB
	.dc.b	$89,$FF,$6A,$4E,$5B,$19,$7F,$E0,$04,$65,$B1,$97,$CF,$19,$6C,$65
	.dc.b	$F0,$B6,$35,$18,$AF,$34,$60,$93,$B2,$46,$9B,$5E,$A8,$C1,$1C,$12
	.dc.b	$BD,$71,$96,$C6,$5F,$5C,$65,$B1,$97,$D7,$19,$6C,$65,$F5,$36,$BD
	.dc.b	$33,$B9,$C9,$E0,$21,$D8,$A7,$42,$9C,$0D,$10,$D1,$4D,$18,$21,$82
	.dc.b	$47,$92,$7C,$D5,$F2,$A7,$BC,$E5,$B1,$97,$B4,$65,$B1,$97,$BC,$65
	.dc.b	$B1,$97,$C2,$D8,$D4,$62,$BE,$34,$60,$93,$B2,$46,$9B,$5E,$A8,$C1
	.dc.b	$1C,$12,$BD,$71,$96,$C6,$5F,$5C,$65,$B1,$97,$D7,$19,$6C,$65,$F5
	.dc.b	$36,$BD,$33,$B9,$C9,$E0,$21,$D8,$A7,$42,$9C,$0D,$10,$D1,$4D,$18
	.dc.b	$21,$F8,$1F,$33,$FD,$16,$FF,$0C,$3F,$CF,$9E,$77,$F4,$27,$FD,$0C
	.dc.b	$64,$A7,$F0,$0F,$2B,$FC,$00,$FF,$0D,$6B,$FC,$21,$AE,$EB,$A5,$FF
	.dc.b	$9A,$9B,$3A,$9F,$E0,$ED,$7F,$8A,$B6,$78,$47,$79,$E2,$47,$04,$3C
	.dc.b	$4E,$5A,$64,$F2,$14,$F0,$BB,$72,$70,$21,$82,$1C,$0E,$47,$A1,$E4
	.dc.b	$E4,$E2,$53,$89,$B3,$8C,$7B,$9D,$99,$2B,$97,$8B,$5E,$84,$38,$1A
	.dc.b	$9F,$94,$49,$F2,$C6,$8E,$13,$CD,$C1,$1C,$12,$79,$47,$A1,$AF,$99
	.dc.b	$AF,$86,$7B,$CF,$16,$CE,$A7,$C2,$D7,$C0,$D9,$E1,$1D,$E7,$89,$1C
	.dc.b	$10,$F1,$39,$69,$93,$C8,$53,$C2,$ED,$C9,$C0,$86,$08,$70,$39,$1E
	.dc.b	$87,$93,$93,$89,$4E,$26,$CE,$31,$EE,$76,$64,$AE,$5E,$2D,$7A,$10
	.dc.b	$E0,$6A,$7E,$51,$27,$CB,$1A,$38,$4F,$37,$04,$71,$D6,$F2,$5E,$7B
	.dc.b	$1F,$F2,$4C,$E6,$31,$7F,$C2,$63,$13,$AA,$E2,$FF,$94,$3D,$B7,$BB
	.dc.b	$FE,$8B,$8C,$36,$FF,$89,$4E,$B2,$EB,$1F,$C6,$63,$15,$FE,$31,$9C
	.dc.b	$CF,$F9,$33,$87,$93,$FC,$7F,$8B,$FC,$65,$3A,$8F,$F2,$5C,$E9,$FF
	.dc.b	$29,$F6,$57,$FC,$9E,$DA,$34,$D7,$E3,$8F,$A2,$A4,$C3,$B4,$65,$F0
	.dc.b	$53,$D9,$E8,$48,$CB,$E1,$EA,$F6,$E3,$FC,$CD,$AF,$F9,$BB,$5D,$1E
	.dc.b	$7E,$87,$2D,$26,$1E,$F8,$CB,$F0,$7E,$B5,$E5,$7B,$DB,$19,$8C,$1F
	.dc.b	$07,$7F,$F6,$07,$BF,$F3,$C9,$97,$F4,$7D,$7E,$6C,$7F,$9E,$B8,$37
	.dc.b	$39,$6B,$F4,$DA,$FD,$0C,$13,$FE,$94,$DF,$E4,$37,$83,$74,$9C,$BF
	.dc.b	$D5,$DA,$FF,$5D,$FC,$11,$DD,$FE,$D7,$F8,$0F,$F4,$FC,$63,$14,$62
	.dc.b	$75,$5F,$D8,$6B,$53,$E1,$AD,$CF,$7C,$36,$FF,$6B,$C3,$EF,$56,$CF
	.dc.b	$ED,$54,$FB,$87,$C4,$7F,$BA,$7E,$0B,$31,$89,$46,$27,$58,$7F,$77
	.dc.b	$9E,$51,$88,$FF,$42,$C7,$E0,$3E,$B3,$11,$FE,$F2,$6C,$64,$EB,$1C
	.dc.b	$F9,$13,$C2,$1D,$7F,$BB,$7F,$71,$D9,$FE,$05,$8C,$47,$EF,$7B,$3F
	.dc.b	$81,$BE,$EF,$31,$F8,$2B,$27,$E1,$79,$8C,$36,$33,$5E,$DA,$E3,$18
	.dc.b	$8F,$6C,$3E,$D3,$A7,$D2,$F7,$6C,$F1,$6B,$C1,$AF,$8E,$5F,$7D,$E5
	.dc.b	$F8,$2F,$43,$67,$C4,$D7,$C2,$D7,$FD,$67,$A7,$C2,$7F,$C5,$4D,$73
	.dc.b	$A7,$C4,$7E,$36,$BF,$EE,$EC,$D8,$FF,$03,$C7,$E4,$72,$F5,$9C,$B6
	.dc.b	$75,$7C,$8F,$A6,$43,$07,$C9,$FB,$5F,$9A,$DA,$F9,$4F,$C1,$8C,$BF
	.dc.b	$31,$AF,$C5,$FB,$FF,$B1,$FB,$E7,$DB,$6C,$7D,$56,$BF,$51,$AF,$DF
	.dc.b	$FA,$FF,$C0,$5B,$E3,$3F,$43,$18,$6C,$7E,$15,$AE,$E7,$D6,$7C,$CD
	.dc.b	$7E,$53,$75,$FC,$94,$62,$3C,$1F,$7B,$CD,$FD,$36,$08,$C4,$7E,$9A
	.dc.b	$31,$7D,$E3,$13,$A7,$D5,$F1,$1F,$99,$BF,$4B,$63,$AB,$5E,$8D,$7F
	.dc.b	$03,$87,$98,$8F,$D7,$3F,$DF,$7E,$F3,$F1,$F7,$D8,$FE,$93,$2F,$F3
	.dc.b	$A4,$E6,$31,$5C,$2B,$75,$FA,$C3,$6E,$39,$FF,$CE,$70,$D3,$FB,$C2
	.dc.b	$F9,$EC,$C6,$1B,$A7,$ED,$DB,$3B,$3E,$9E,$1E,$C3,$D4,$53,$E7,$FD
	.dc.b	$07,$EF,$FE,$47,$AB,$67,$22,$32,$D3,$F8,$E7,$B6,$FB,$1F,$DA,$46
	.dc.b	$1B,$7F,$C7,$A7,$52,$E8,$F7,$C6,$1F,$6F,$B5,$F8,$D9,$FC,$1B,$6B
	.dc.b	$F0,$4D,$7F,$30,$D7,$F9,$0E,$7F,$8D,$6F,$0F,$DA,$7E,$1D,$30,$F1
	.dc.b	$F2,$78,$B2,$E9,$39,$6C,$EA,$3E,$5D,$74,$3E,$59,$4C,$F5,$7F,$0B
	.dc.b	$70,$7E,$84,$FF,$5A,$0F,$59,$F1,$DF,$E3,$0E,$6F,$F0,$0C,$FF,$AF
	.dc.b	$03,$87,$84,$6F,$C2,$FF,$61,$57,$2F,$51,$5C,$27,$35,$B8,$E9,$7C
	.dc.b	$EF,$84,$6C,$CD,$7E,$DC,$FE,$92,$47,$F9,$1B,$5D,$18,$21,$F7,$F3
	.dc.b	$FE,$95,$F0,$9F,$37,$33,$48,$F0,$9F,$D8,$D7,$56,$9F,$82,$8E,$25
	.dc.b	$9F,$C3,$B8,$8F,$C2,$B5,$FE,$B5,$A7,$E4,$3E,$1B,$C4,$49,$FC,$B0
	.dc.b	$7F,$3B,$F0,$F2,$3F,$42,$DA,$F7,$90,$F2,$96,$3F,$71,$C3,$FD,$A7
	.dc.b	$E6,$BF,$66,$48,$FE,$3B,$19,$7F,$80,$F5,$FF,$6F,$FC,$27,$FD,$E7
	.dc.b	$DB,$FC,$26,$51,$8E,$DF,$E1,$2D,$AE,$13,$F2,$71,$FC,$FE,$DF,$25
	.dc.b	$A7,$F2,$08,$7C,$B7,$F2,$46,$C6,$63,$07,$F4,$0E,$FF,$B3,$1F,$CD
	.dc.b	$1A,$FF,$30,$9F,$B6,$E5,$AF,$C5,$3C,$25,$E6,$FF,$47,$5F,$F4,$8E
	.dc.b	$19,$3F,$D0,$2B,$AB,$FD,$4D,$A7,$F7,$0F,$03,$FC,$29,$FB,$77,$F5
	.dc.b	$D8,$34,$93,$FD,$A7,$4F,$F0,$E7,$87,$F3,$63,$04,$3C,$9C,$1F,$EA
	.dc.b	$4C,$BE,$46,$7E,$69,$FF,$0E,$78,$BF,$E2,$3E,$4F,$EF,$F9,$FE,$1E
	.dc.b	$72,$D7,$FC,$3F,$3B,$7E,$A5,$73,$7B,$F6,$7E,$ED,$FF,$0E,$78,$DB
	.dc.b	$7F,$C2,$F7,$CD,$CB,$4F,$F0,$8C,$6E,$3F,$C4,$9E,$3F,$F8,$DF,$E9
	.dc.b	$FF,$C5,$94,$69,$27,$FC,$6D,$E5,$7C,$37,$FC,$6A,$D7,$FC,$5C,$D9
	.dc.b	$FF,$0F,$B9,$8F,$B6,$7E,$80,$86,$23,$FC,$AD,$F0,$FF,$E6,$8D,$7F
	.dc.b	$CA,$B0,$8C,$79,$9F,$E4,$D6,$C6,$6B,$FC,$CD,$E7,$FF,$9A,$E7,$FC
	.dc.b	$D5,$39,$9F,$F3,$3C,$E5,$DF,$9D,$CD,$FF,$3E,$4F,$F3,$D6,$F7,$7F
	.dc.b	$A0,$48,$FF,$A1,$7D,$3F,$E0,$AD,$9E,$0F,$FA,$46,$32,$D8,$CB,$FC
	.dc.b	$BB,$D2,$FF,$62,$7C,$2F,$FA,$34,$91,$E2,$7F,$D5,$52,$69,$23,$FD
	.dc.b	$59,$19,$7F,$D6,$5E,$C7,$F9,$C6,$7F,$DC,$84,$8C,$CE,$DF,$E7,$E3
	.dc.b	$DA,$FF,$60,$FC,$1F,$F4,$E6,$C6,$8D,$7B,$7F,$E7,$D9,$D3,$63,$5E
	.dc.b	$E7,$FB,$3A,$7F,$AB,$B6,$33,$C3,$B7,$C3,$19,$7F,$DB,$98,$2B,$FA
	.dc.b	$FF,$BD,$FF,$00,$5B,$EF,$71,$AF,$EA,$F3,$97,$FD,$A5,$EF,$E5,$7B
	.dc.b	$FF,$EE,$FC,$3A,$46,$9A,$FC,$28,$D6,$1F,$EB,$F2,$47,$E1,$1A,$FF
	.dc.b	$B6,$67,$8B,$FE,$D9,$8F,$AE,$47,$E7,$BF,$DE,$25,$B7,$95,$B5,$CA
	.dc.b	$BC,$F3,$E9,$F8,$0F,$64,$7D,$66,$BF,$DD,$9A,$9B,$D8,$EF,$76,$F5
	.dc.b	$8C,$36,$33,$97,$FA,$F8,$FF,$06,$FC,$0B,$E0,$BF,$D7,$5C,$38,$6E
	.dc.b	$5F,$83,$9D,$69,$FC,$4E,$11,$8A,$E1,$39,$8C,$4F,$1A,$E6,$FE,$4A
	.dc.b	$32,$D8,$CB,$FE,$14,$8C,$B6,$32,$F8,$1B,$1A,$8C,$4F,$86,$72,$47
	.dc.b	$04,$AF,$39,$27,$CD,$19,$6C,$65,$F3,$46,$5B,$19,$7C,$D1,$96,$C6
	.dc.b	$5F,$2B,$6B,$D0,$4A,$F3,$46,$5B,$3E,$22,$1E,$02,$9D,$8A,$76,$2C
	.dc.b	$F5,$AF,$24,$64,$E4,$43,$91,$4D,$18,$21,$82,$47,$5A,$F4,$4F,$68
	.dc.b	$C4,$F8,$AB,$D6,$F5,$8C,$B6,$32,$F6,$8C,$B6,$32,$F8,$1B,$1A,$8C
	.dc.b	$4F,$86,$72,$47,$04,$AF,$39,$27,$CD,$19,$6C,$65,$F3,$46,$5B,$19
	.dc.b	$7C,$D1,$96,$C6,$5F,$2B,$6B,$D0,$4A,$F3,$46,$5B,$3E,$22,$1E,$02
	.dc.b	$9D,$8A,$76,$2C,$F5,$AF,$24,$64,$E4,$43,$91,$4D,$1F,$80,$F3,$FF
	.dc.b	$D0,$2F,$F4,$56,$BF,$E8,$8F,$4B,$FD,$2E,$FF,$A4,$1B,$4F,$93,$5E
	.dc.b	$1B,$5F,$C4,$9A,$FF,$0F,$6B,$FC,$1D,$AF,$F0,$56,$BF,$C4,$9B,$3D
	.dc.b	$E7,$B1,$27,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B3,$63,$B4,$74,$21,$D0
	.dc.b	$A7,$42,$9D,$0A,$74,$2B,$D1,$E4,$43,$91,$4E,$45,$39,$15,$E4,$F0
	.dc.b	$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$4F,$D4,$24,$FC
	.dc.b	$96,$CF,$C0,$6B,$E9,$6B,$E4,$6B,$E1,$6B,$E0,$6C,$F8,$27,$B1,$27
	.dc.b	$B3,$5E,$CD,$7B,$35,$EC,$D7,$B3,$63,$B4,$74,$21,$D0,$A7,$42,$9D
	.dc.b	$0A,$74,$2B,$D1,$E4,$43,$91,$4E,$45,$39,$15,$E4,$F0,$21,$C0,$A7
	.dc.b	$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$4F,$D4,$24,$FE,$AF,$5F,$F0
	.dc.b	$AD,$FF,$93,$ED,$18,$7F,$CA,$4D,$7F,$CB,$2D,$43,$E1,$5D,$5F,$C3
	.dc.b	$AB,$A4,$62,$D1,$89,$D5,$7F,$97,$EB,$53,$FE,$39,$E1,$CC,$FF,$91
	.dc.b	$DA,$F8,$38,$BF,$C7,$73,$BC,$3A,$CF,$F9,$A7,$E9,$5F,$37,$C2,$11
	.dc.b	$89,$DC,$7F,$9D,$23,$15,$F0,$DB,$1F,$E7,$3B,$7C,$3F,$36,$DF,$29
	.dc.b	$F2,$4E,$4F,$07,$4B,$E2,$69,$2B,$85,$27,$51,$8E,$9F,$87,$3F,$E8
	.dc.b	$36,$BF,$E8,$A6,$BF,$E7,$4D,$BB,$FC,$D4,$AE,$B6,$8C,$69,$F4,$65
	.dc.b	$F1,$14,$98,$7D,$3C,$3F,$A3,$C7,$CD,$24,$7F,$A7,$1A,$FC,$9D,$B6
	.dc.b	$FD,$16,$B9,$AE,$EE,$5A,$FE,$29,$7C,$83,$E7,$3F,$AB,$FC,$4F,$FA
	.dc.b	$8E,$73,$18,$AF,$F5,$5C,$60,$4E,$EB,$94,$62,$7C,$13,$F7,$6F,$94
	.dc.b	$79,$63,$13,$B8,$F2,$57,$2F,$B9,$E9,$87,$F6,$A6,$D1,$A7,$EA,$FE
	.dc.b	$E9,$F5,$1F,$DC,$3E,$33,$0D,$C7,$FD,$7D,$3A,$B5,$69,$B3,$FB,$2E
	.dc.b	$FF,$F4,$6B,$FB,$7E,$9F,$F4,$6B,$63,$2E,$FC,$0F,$DE,$53,$E2,$3F
	.dc.b	$65,$F0,$7F,$2D,$B5,$F2,$9A,$FD,$36,$BF,$E0,$85,$F0,$3F,$E0,$2A
	.dc.b	$E8,$D8,$CC,$78,$1B,$3D,$FF,$07,$FF,$08,$FC,$35,$A3,$49,$3E,$93
	.dc.b	$D9,$8F,$F8,$73,$3F,$7C,$64,$F4,$7E,$05,$D1,$B7,$D1,$B7,$EF,$9D
	.dc.b	$47,$F1,$AD,$3E,$3B,$70,$ED,$18,$8F,$E3,$B8,$78,$1F,$89,$E2,$FF
	.dc.b	$8A,$B5,$F9,$4D,$8F,$94,$D7,$79,$27,$E8,$74,$FF,$8C,$BF,$75,$B2
	.dc.b	$69,$2B,$EA,$7D,$FF,$37,$E9,$36,$7E,$E3,$77,$3B,$9F,$E7,$79,$FA
	.dc.b	$DB,$A7,$B8,$8F,$E0,$FF,$00,$E4,$4C,$3D,$11,$F8,$58,$ED,$EA,$FF
	.dc.b	$92,$BE,$5B,$F1,$4D,$9F,$C5,$35,$FC,$53,$5F,$97,$87,$8F,$7F,$99
	.dc.b	$18,$C7,$C6,$FE,$4F,$F0,$1F,$61,$32,$F1,$C7,$E5,$63,$E0,$78,$5F
	.dc.b	$E5,$CF,$DA,$3F,$CB,$35,$F9,$F8,$C3,$F9,$D6,$D7,$E7,$5A,$F8,$4F
	.dc.b	$C7,$90,$FC,$71,$FA,$3F,$1B,$FC,$D3,$FA,$3F,$6E,$0F,$E9,$76,$8C
	.dc.b	$4E,$A3,$F5,$8D,$AF,$D6,$7E,$0B,$FA,$29,$FD,$83,$63,$31,$FA,$43
	.dc.b	$25,$7F,$93,$7A,$BF,$E6,$DE,$5F,$7B,$67,$F9,$1B,$67,$7E,$67,$E5
	.dc.b	$E7,$D7,$1E,$F2,$9F,$84,$3F,$82,$5E,$DF,$1D,$FB,$E3,$11,$FC,$0D
	.dc.b	$AF,$13,$9E,$FF,$71,$C1,$FA,$D3,$E6,$14,$F0,$F9,$CF,$8E,$FE,$F3
	.dc.b	$F1,$CE,$AF,$F4,$E5,$23,$0F,$F8,$42,$B8,$79,$3F,$E1,$29,$CC,$FF
	.dc.b	$A8,$4E,$15,$CA,$33,$5C,$27,$27,$3B,$F0,$D6,$CF,$E0,$87,$F1,$22
	.dc.b	$47,$F2,$D6,$BA,$30,$43,$ED,$FC,$3F,$C9,$D0,$D2,$3B,$8F,$D1,$4F
	.dc.b	$46,$9F,$73,$5F,$F5,$CD,$7E,$01,$B5,$93,$A9,$63,$C1,$F0,$72,$3F
	.dc.b	$16,$DA,$FC,$E4,$3F,$43,$FC,$CB,$4F,$CC,$96,$37,$1F,$EC,$4D,$5D
	.dc.b	$0E,$8F,$A3,$5E,$83,$E2,$10,$FE,$95,$CD,$FE,$D2,$FD,$E3,$F5,$4D
	.dc.b	$C7,$F6,$C2,$32,$D3,$F8,$21,$F2,$A6,$E7,$FD,$6F,$A7,$E2,$C6,$0F
	.dc.b	$98,$7F,$20,$DB,$E6,$34,$FE,$2F,$4E,$AF,$E3,$6D,$8C,$C6,$0F,$E4
	.dc.b	$D1,$E1,$8F,$EC,$DD,$9E,$06,$C6,$5F,$ED,$11,$97,$E2,$10,$F1,$1F
	.dc.b	$16,$3F,$9D,$9C,$4B,$87,$F7,$7E,$DE,$51,$89,$FE,$F1,$DF,$FE,$FC
	.dc.b	$F4,$7F,$A5,$92,$33,$28,$FE,$AE,$F5,$ED,$FD,$73,$73,$EB,$18,$C7
	.dc.b	$E3,$F0,$FD,$EE,$0F,$BB,$8F,$A0,$F5,$E1,$FE,$0D,$F0,$4B,$93,$FB
	.dc.b	$83,$67,$FB,$AC,$60,$EB,$C7,$F8,$9C,$EE,$1F,$8A,$BF,$F0,$C7,$87
	.dc.b	$D2,$FF,$C1,$04,$C3,$FC,$31,$19,$7F,$C3,$5E,$27,$8B,$0F,$F0,$F3
	.dc.b	$63,$37,$D7,$5F,$F0,$BD,$7F,$88,$FC,$EF,$82,$49,$F8,$CF,$D9,$AF
	.dc.b	$80,$4D,$7E,$25,$7E,$33,$C7,$FF,$1D,$7C,$07,$F8,$E4,$95,$FE,$3E
	.dc.b	$F2,$7E,$2C,$7F,$92,$32,$93,$FE,$5A,$F3,$3F,$CA,$F1,$88,$FF,$34
	.dc.b	$7B,$DF,$E5,$42,$3F,$E6,$97,$FC,$A2,$47,$97,$9B,$F4,$A3,$11,$FE
	.dc.b	$71,$F1,$7F,$CB,$A4,$7F,$CE,$4F,$FA,$A4,$8F,$FA,$7F,$D4,$FB,$71
	.dc.b	$88,$FF,$5F,$9F,$EB,$FD,$B2,$FF,$A1,$F7,$4F,$FA,$34,$91,$C0,$FF
	.dc.b	$4F,$74,$7E,$1B,$D8,$FF,$68,$7A,$7F,$E9,$5B,$46,$27,$55,$C6,$B5
	.dc.b	$3F,$ED,$7F,$53,$FD,$98,$DC,$77,$5F,$EA,$56,$BF,$EA,$5C,$3A,$BC
	.dc.b	$FC,$1C,$BF,$EB,$16,$E1,$FE,$B2,$BF,$2C,$6B,$B3,$FD,$7F,$1F,$EE
	.dc.b	$76,$D7,$FB,$17,$D5,$D1,$D3,$CB,$FF,$67,$57,$FA,$30,$93,$FE,$A1
	.dc.b	$7F,$DB,$04,$7E,$86,$3E,$67,$FD,$C0,$D7,$FD,$DF,$EB,$7C,$07,$E8
	.dc.b	$4F,$FA,$EC,$8F,$FA,$05,$FF,$34,$91,$C1,$F4,$FD,$77,$8F,$FA,$58
	.dc.b	$FB,$B1,$8B,$FD,$FB,$7A,$7E,$D4,$A3,$0F,$3A,$E4,$8D,$38,$6D,$74
	.dc.b	$8C,$4F,$29,$F0,$C6,$09,$3B,$25,$78,$AB,$85,$77,$6B,$DA,$35,$5C
	.dc.b	$9B,$5C,$A7,$D0,$D7,$C2,$D7,$A4,$F1,$77,$3C,$88,$E0,$87,$27,$2D
	.dc.b	$32,$74,$9E,$CF,$9C,$C9,$4E,$04,$38,$47,$46,$9C,$2B,$E0,$92,$BD
	.dc.b	$11,$C4,$E8,$70,$31,$1B,$AF,$1B,$6B,$B4,$F0,$24,$65,$F0,$57,$67
	.dc.b	$A1,$0D,$B9,$6D,$76,$8C,$4F,$39,$F1,$C6,$09,$3B,$25,$79,$2B,$85
	.dc.b	$77,$6B,$DA,$35,$5C,$9B,$5C,$A7,$D2,$D7,$C2,$D7,$A4,$F1,$77,$3C
	.dc.b	$88,$E0,$87,$27,$2D,$32,$74,$9E,$CF,$9C,$C9,$4E,$04,$38,$47,$46
	.dc.b	$9C,$2B,$E0,$92,$BD,$11,$C4,$E8,$70,$31,$1B,$AF,$1B,$6B,$B4,$FE
	.dc.b	$07,$C9,$FF,$34,$3F,$CE,$DA,$FF,$9D,$8C,$79,$2F,$2F,$F8,$3C,$7F
	.dc.b	$07,$F8,$4F,$F0,$2B,$FB,$96,$BF,$C4,$DA,$FF,$12,$6B,$FB,$D6,$BC
	.dc.b	$F1,$EB,$8A,$72,$D8,$CC,$60,$8E,$09,$3E,$36,$BE,$36,$BE,$36,$BE
	.dc.b	$36,$BE,$33,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A
	.dc.b	$70,$34,$43,$45,$34,$60,$86,$09,$3F,$25,$F2,$35,$EE,$D7,$A3,$5E
	.dc.b	$6D,$79,$CF,$D0,$79,$B5,$E6,$D7,$9B,$5E,$6D,$79,$CF,$D0,$9E,$C4
	.dc.b	$9D,$36,$35,$18,$23,$82,$4F,$91,$AF,$91,$AF,$91,$AF,$91,$AF,$90
	.dc.b	$E4,$43,$91,$4E,$45,$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10
	.dc.b	$D1,$4D,$18,$21,$82,$4F,$C9,$7C,$8D,$7B,$B5,$E8,$D7,$9B,$5E,$73
	.dc.b	$F4,$1E,$6D,$7D,$3D,$DF,$E8,$DB,$FF,$06,$B6,$3F,$C9,$0D,$7F,$CA
	.dc.b	$2D,$7F,$C1,$DE,$D3,$DC,$FF,$48,$D7,$F8,$5A,$31,$8F,$F1,$3A,$FF
	.dc.b	$29,$46,$27,$55,$C2,$7F,$CA,$DC,$0E,$BC,$47,$F9,$82,$31,$87,$38
	.dc.b	$C4,$E9,$FF,$10,$4E,$AF,$FC,$8D,$F3,$59,$9F,$F5,$93,$79,$0D,$35
	.dc.b	$F0,$0F,$86,$A4,$CB,$C1,$7F,$E2,$D9,$D4,$62,$1D,$5E,$D7,$FC,$E0
	.dc.b	$D7,$FC,$F2,$D7,$FC,$6C,$D5,$D0,$F8,$FF,$43,$6B,$B3,$73,$FF,$21
	.dc.b	$B6,$BE,$BF,$92,$F2,$5E,$F7,$DC,$8C,$36,$33,$3C,$1A,$FF,$A3,$B8
	.dc.b	$BE,$CC,$FD,$6C,$73,$3F,$80,$E2,$FF,$67,$CE,$4F,$F2,$AF,$0F,$D6
	.dc.b	$6F,$16,$AB,$EE,$B5,$FB,$37,$D1,$78,$EF,$B7,$FF,$6D,$FB,$DF,$E6
	.dc.b	$08,$C5,$23,$13,$AA,$FF,$4F,$56,$A7,$C3,$5B,$9F,$0B,$5F,$0C,$68
	.dc.b	$E7,$0F,$7A,$92,$7A,$DF,$FB,$BA,$3A,$92,$3A,$B5,$FD,$3E,$DF,$4D
	.dc.b	$FA,$04,$3E,$81,$FE,$B5,$B7,$EF,$2F,$8E,$FE,$E3,$18,$6C,$67,$1F
	.dc.b	$F4,$1C,$E5,$FF,$42,$8F,$25,$F0,$88,$7C,$DF,$FB,$0D,$D4,$66,$31
	.dc.b	$3C,$27,$78,$FF,$B4,$1D,$98,$7F,$D1,$CB,$E1,$5F,$23,$FE,$08,$8C
	.dc.b	$36,$33,$1E,$66,$CF,$C3,$3E,$3F,$2F,$F8,$63,$B9,$B7,$C9,$9F,$8E
	.dc.b	$78,$F1,$FE,$1A,$FA,$39,$7C,$11,$88,$D4,$7F,$86,$2B,$9D,$F5,$BE
	.dc.b	$95,$E0,$3C,$47,$E3,$7F,$6C,$FD,$56,$5F,$39,$B1,$F5,$DA,$FF,$BD
	.dc.b	$36,$46,$23,$E2,$FA,$1F,$B9,$FD,$47,$E8,$36,$3E,$63,$5D,$F6,$7D
	.dc.b	$79,$DE,$FF,$2E,$BE,$3D,$70,$8C,$1E,$DC,$BD,$71,$FE,$C1,$D3,$DE
	.dc.b	$FB,$03,$87,$C3,$F8,$27,$52,$69,$F8,$AC,$11,$D7,$5F,$90,$4A,$E9
	.dc.b	$7C,$DF,$94,$7D,$62,$9F,$ED,$33,$F9,$E7,$CC,$7F,$92,$BA,$4D,$26
	.dc.b	$3D,$67,$91,$EB,$CB,$DA,$FA,$9B,$7E,$18,$D6,$1C,$5F,$09,$FA,$1F
	.dc.b	$8D,$7D,$57,$E8,$F5,$FF,$30,$E2,$9C,$BF,$38,$C4,$ED,$FD,$14,$6A
	.dc.b	$78,$F9,$43,$FC,$DB,$3F,$A7,$70,$FE,$9A,$FF,$4F,$18,$78,$BC,$1F
	.dc.b	$C3,$F5,$FE,$CE,$7F,$0A,$D8,$CC,$7E,$E8,$C9,$5F,$79,$F1,$71,$F8
	.dc.b	$4F,$C4,$3F,$80,$57,$86,$7C,$CF,$86,$3F,$83,$39,$7C,$EE,$5E,$87
	.dc.b	$E0,$33,$FA,$24,$AE,$F1,$A3,$EC,$8F,$29,$E5,$7F,$80,$FE,$0B,$E4
	.dc.b	$48,$D4,$67,$0F,$C4,$57,$0F,$13,$57,$C2,$F7,$18,$AD,$DF,$19,$E9
	.dc.b	$7C,$55,$F1,$8C,$36,$78,$4F,$C9,$AF,$EC,$C4,$7F,$B3,$35,$FE,$2E
	.dc.b	$D7,$F2,$2F,$CE,$3E,$5E,$5F,$38,$FC,$04,$7C,$E8,$CB,$F3,$A3,$2F
	.dc.b	$84,$EE,$43,$B9,$F8,$0F,$90,$FC,$CF,$27,$F6,$46,$EB,$FC,$FD,$B4
	.dc.b	$7F,$1E,$F8,$8F,$CC,$5F,$8A,$FC,$91,$98,$FC,$C9,$FA,$6D,$FF,$4B
	.dc.b	$EA,$DE,$1F,$E9,$8D,$A3,$4F,$F6,$5F,$17,$F3,$0B,$F8,$D8,$7B,$4F
	.dc.b	$59,$0F,$92,$7B,$CA,$7F,$09,$E4,$FE,$0F,$7F,$81,$8F,$C4,$46,$27
	.dc.b	$F8,$74,$63,$93,$F3,$6E,$27,$4F,$53,$C4,$76,$21,$82,$5F,$A5,$F2
	.dc.b	$5B,$5F,$E4,$AF,$8C,$C7,$2F,$82,$7F,$A0,$B5,$FE,$E5,$3C,$31,$F8
	.dc.b	$77,$D6,$7E,$11,$E4,$BE,$DC,$BC,$5A,$77,$BF,$ED,$77,$FD,$1A,$31
	.dc.b	$CF,$FD,$FF,$F4,$FF,$E9,$ED,$C3,$FA,$7B,$5F,$E9,$F2,$9F,$EB,$9C
	.dc.b	$DE,$7E,$AF,$EB,$6D,$EA,$FE,$C7,$8E,$AF,$E9,$61,$F1,$BA,$7F,$CA
	.dc.b	$13,$EC,$6C,$7F,$72,$9F,$8D,$D9,$FE,$84,$AF,$EF,$11,$8E,$CF,$8D
	.dc.b	$5C,$F1,$EB,$BF,$C5,$8F,$AD,$5F,$E0,$C3,$E8,$76,$FC,$28,$FF,$0B
	.dc.b	$35,$FF,$0A,$B5,$FF,$11,$63,$FE,$1D,$E7,$FF,$0F,$F7,$FE,$9E,$7F
	.dc.b	$C1,$8D,$7F,$C3,$B3,$B7,$FB,$F7,$81,$FE,$31,$D3,$FC,$70,$D8,$CF
	.dc.b	$83,$F2,$63,$11,$F7,$33,$F5,$3E,$DA,$FF,$2A,$DC,$8F,$F2,$AB,$5F
	.dc.b	$ED,$1E,$1F,$F9,$82,$3F,$CA,$A4,$8F,$F2,$9E,$9F,$E3,$F7,$FC,$BB
	.dc.b	$EB,$FF,$98,$08,$E0,$DB,$FE,$5E,$23,$83,$FC,$CB,$EB,$FF,$9A,$08
	.dc.b	$E0,$FF,$5C,$F0,$7F,$92,$E3,$FD,$08,$48,$CD,$23,$FD,$1C,$48,$FF
	.dc.b	$47,$35,$FF,$02,$79,$3F,$E9,$DF,$4B,$FD,$37,$CF,$FE,$57,$6D,$9A
	.dc.b	$7E,$FA,$BF,$D5,$F8,$6D,$FF,$57,$B5,$FF,$57,$B5,$FF,$57,$B5,$79
	.dc.b	$9F,$EB,$A9,$E8,$D8,$FE,$16,$D7,$FC,$34,$D7,$FA,$C7,$9F,$FE,$CA
	.dc.b	$BF,$34,$7F,$12,$8C,$57,$FB,$4A,$73,$3F,$E2,$39,$CB,$FE,$94,$9F
	.dc.b	$F4,$B9,$1C,$1F,$31,$FF,$59,$11,$C1,$FE,$B9,$7F,$D7,$84,$70,$7F
	.dc.b	$24,$F4,$5E,$8F,$EB,$9B,$7F,$BA,$6F,$07,$CD,$6D,$71,$AE,$04,$AE
	.dc.b	$0D,$AE,$0D,$9E,$13,$A2,$4E,$9B,$1A,$8C,$11,$C1,$0F,$F2,$55,$79
	.dc.b	$1B,$5D,$DB,$5D,$1B,$5C,$DA,$73,$3F,$C2,$15,$CD,$B5,$CD,$B5,$CD
	.dc.b	$B5,$CD,$A7,$31,$5C,$DB,$5C,$DB,$5C,$DB,$5C,$DA,$73,$AE,$44,$AE
	.dc.b	$4D,$AE,$4D,$AE,$4D,$AE,$55,$C0,$95,$C1,$B5,$C1,$B3,$C2,$74,$49
	.dc.b	$D3,$63,$51,$82,$38,$21,$F4,$08,$7C,$82,$9F,$00,$B5,$E9,$6D,$79
	.dc.b	$1B,$5E,$16,$D7,$81,$A7,$80,$EC,$43,$B1,$6B,$B3,$6B,$B3,$6B,$B3
	.dc.b	$6B,$B3,$4E,$C7,$42,$57,$46,$D7,$46,$D7,$46,$D7,$46,$9D,$2B,$91
	.dc.b	$2B,$93,$6B,$93,$6B,$93,$6B,$95,$70,$3C,$5E,$3F,$F9,$8E,$3F,$C4
	.dc.b	$ED,$7F,$CD,$3E,$47,$F2,$D7,$FC,$DA,$7F,$03,$F4,$18,$E7,$04,$62
	.dc.b	$75,$39,$8C,$47,$F1,$16,$BF,$BA,$6B,$FB,$26,$BF,$AA,$6C,$F6,$9E
	.dc.b	$84,$9E,$E4,$3B,$39,$69,$93,$C0,$E5,$A6,$4F,$13,$96,$99,$3C,$8E
	.dc.b	$5A,$66,$3C,$CD,$9F,$49,$27,$84,$EA,$7C,$A4,$3C,$65,$3C,$25,$3B
	.dc.b	$95,$E4,$F0,$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$4F
	.dc.b	$B6,$3E,$04,$FB,$A3,$11,$E7,$6B,$E2,$6B,$DD,$AF,$66,$CF,$69,$E8
	.dc.b	$49,$EE,$43,$B3,$96,$99,$3C,$0E,$5A,$64,$F1,$39,$69,$93,$C8,$E5
	.dc.b	$A6,$63,$CC,$D9,$F4,$92,$78,$4E,$A7,$CA,$43,$C6,$53,$C2,$53,$B9
	.dc.b	$5E,$4F,$02,$1C,$0A,$70,$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$24,$FB
	.dc.b	$63,$E0,$4F,$E9,$79,$1D,$6E,$B6,$3F,$E3,$99,$CC,$62,$FF,$D7,$31
	.dc.b	$89,$D5,$71,$7F,$C8,$1F,$86,$38,$BF,$89,$C6,$1B,$7F,$EC,$29,$D6
	.dc.b	$3D,$60,$48,$CD,$7F,$92,$F8,$3F,$DB,$35,$BA,$FF,$1E,$F8,$BD,$1F
	.dc.b	$F1,$C3,$5C,$C6,$D7,$9E,$F8,$ED,$46,$1B,$19,$9E,$0D,$C3,$FC,$B9
	.dc.b	$EB,$B9,$F8,$B6,$BB,$36,$77,$3D,$B2,$78,$14,$97,$E5,$8C,$BE,$7F
	.dc.b	$4D,$EC,$79,$9B,$19,$8C,$1E,$AE,$1F,$AF,$E9,$FD,$4E,$33,$49,$BF
	.dc.b	$D1,$E2,$FE,$8F,$1F,$57,$0D,$CE,$5A,$FD,$41,$C5,$F6,$27,$DB,$3A
	.dc.b	$9F,$B3,$BF,$C3,$9C,$92,$3A,$78,$BD,$C8,$E0,$EC,$7F,$98,$BE,$8F
	.dc.b	$FA,$DF,$E1,$1F,$E8,$BE,$A4,$62,$77,$18,$C5,$5B,$AD,$57,$A7,$2F
	.dc.b	$C7,$FA,$BE,$F2,$47,$BD,$F9,$0C,$3F,$3F,$D1,$FA,$57,$07,$E5,$B0
	.dc.b	$7D,$CB,$C0,$FE,$CF,$18,$6C,$66,$3F,$D4,$86,$4B,$E5,$BF,$0A,$E2
	.dc.b	$D8,$CC,$60,$EB,$F9,$E3,$CF,$A4,$9E,$B1,$97,$FC,$0D,$CB,$FB,$AC
	.dc.b	$BD,$27,$EB,$9C,$3E,$EE,$5F,$2F,$A7,$C6,$3E,$9D,$23,$11,$F8,$38
	.dc.b	$C6,$1F,$02,$73,$3F,$87,$BE,$AD,$D7,$E5,$3D,$CF,$1F,$1B,$F3,$F7
	.dc.b	$C4,$FF,$84,$23,$0D,$8C,$E5,$FE,$CD,$9C,$BF,$2B,$F0,$3F,$F1,$0F
	.dc.b	$D2,$FA,$38,$11,$8C,$7D,$11,$8A,$56,$EB,$55,$F7,$7F,$5C,$7C,$07
	.dc.b	$DC,$6D,$FD,$A6,$CF,$C6,$6B,$ED,$21,$F0,$78,$3E,$9E,$BF,$0A,$3F
	.dc.b	$15,$BF,$8D,$F9,$ED,$7E,$73,$5D,$E3,$FE,$34,$E8,$7D,$43,$E3,$BF
	.dc.b	$C8,$3B,$FE,$AA,$73,$18,$BF,$DD,$46,$34,$56,$EF,$94,$62,$BD,$27
	.dc.b	$E1,$F2,$F4,$37,$3F,$59,$FE,$53,$FA,$B7,$D1,$7F,$93,$EF,$F7,$71
	.dc.b	$87,$F5,$15,$FC,$2E,$30,$FF,$94,$1B,$1F,$5B,$DD,$FF,$2D,$FC,$D7
	.dc.b	$F9,$5A,$1C,$69,$CC,$6C,$4E,$E7,$51,$83,$D7,$6E,$6F,$CA,$BE,$B3
	.dc.b	$F4,$B0,$FC,$05,$F3,$3F,$E6,$C8,$F1,$46,$1C,$4F,$F0,$38,$C4,$6A
	.dc.b	$78,$3F,$A7,$FC,$21,$F5,$5F,$16,$FF,$85,$46,$27,$E2,$C7,$F9,$EA
	.dc.b	$30,$FC,$16,$C7,$EC,$7E,$1D,$F3,$DF,$E8,$07,$E3,$18,$83,$88,$D4
	.dc.b	$71,$6D,$7E,$54,$FE,$0A,$7C,$5E,$1F,$BE,$DE,$1F,$E3,$38,$78,$A3
	.dc.b	$D3,$8F,$AD,$F4,$37,$04,$EE,$7E,$71,$F4,$B4,$FA,$2F,$53,$C2,$E0
	.dc.b	$98,$78,$BD,$47,$C3,$7F,$84,$FE,$1D,$EB,$FF,$A9,$30,$46,$1F,$F0
	.dc.b	$5D,$70,$F4,$BF,$C1,$B3,$99,$FF,$54,$B6,$75,$1C,$AB,$A4,$E6,$FC
	.dc.b	$11,$83,$F7,$87,$F6,$F2,$4F,$F2,$B6,$BA,$30,$43,$E4,$E7,$EA,$3F
	.dc.b	$0D,$A7,$FA,$EA,$7F,$00,$D8,$C9,$B9,$E8,$43,$F6,$93,$E0,$7F,$09
	.dc.b	$B7,$FB,$13,$E0,$7F,$05,$46,$91,$EB,$5E,$07,$B5,$7D,$C6,$9F,$72
	.dc.b	$FC,$91,$FA,$1E,$1F,$F6,$8F,$D7,$7E,$86,$8D,$23,$F0,$67,$CE,$53
	.dc.b	$F1,$D7,$FC,$E7,$9B,$FD,$B7,$CB,$FB,$B6,$D7,$EF,$E3,$0F,$F0,$43
	.dc.b	$F8,$7F,$47,$92,$FF,$C9,$7D,$1F,$C4,$70,$FE,$1D,$39,$94,$E6,$BC
	.dc.b	$5C,$7F,$E6,$0B,$F9,$55,$E4,$3F,$94,$F4,$FF,$26,$F7,$7F,$94,$76
	.dc.b	$FF,$BC,$1B,$81,$A7,$F3,$CE,$EF,$B9,$7F,$CF,$1A,$FF,$3C,$9F,$CA
	.dc.b	$3F,$EF,$AE,$FF,$EA,$D8,$F8,$B2,$F2,$4F,$9B,$A3,$FC,$F1,$3F,$D4
	.dc.b	$7C,$1E,$72,$C1,$E0,$FF,$58,$AD,$47,$63,$35,$C1,$FE,$5B,$97,$1F
	.dc.b	$A9,$FE,$C9,$06,$93,$0F,$F2,$27,$83,$FD,$CB,$0F,$F1,$CB,$63,$3E
	.dc.b	$27,$F7,$89,$CB,$63,$D5,$DF,$F0,$1F,$9C,$7E,$6F,$0F,$F0,$69,$E7
	.dc.b	$DB,$FC,$39,$7F,$D9,$BC,$6F,$44,$62,$B9,$CF,$C7,$DB,$E2,$C7,$F8
	.dc.b	$5F,$C7,$D3,$FE,$39,$6B,$F1,$DB,$81,$BC,$DF,$E3,$3B,$FF,$06,$79
	.dc.b	$1F,$1E,$31,$7F,$E0,$C8,$F9,$5E,$4F,$F8,$FE,$BF,$C5,$2D,$9F,$F1
	.dc.b	$43,$5F,$F2,$53,$5F,$D1,$9F,$BB,$3E,$19,$83,$FD,$49,$E8,$FD,$DF
	.dc.b	$7F,$FC,$F0,$48,$FF,$3C,$3F,$E6,$1C,$9F,$5B,$49,$3F,$E6,$8F,$43
	.dc.b	$F0,$D1,$FE,$81,$E8,$93,$FE,$81,$6B,$FE,$81,$F4,$FF,$D3,$DF,$01
	.dc.b	$FE,$80,$24,$7F,$9E,$23,$2F,$FA,$26,$3F,$D1,$84,$70,$7F,$93,$DF
	.dc.b	$F4,$B1,$1C,$1F,$E8,$2F,$5B,$FD,$3F,$EF,$FF,$A1,$21,$18,$75,$EB
	.dc.b	$FE,$7E,$75,$39,$9F,$F4,$C7,$B1,$FE,$AC,$9F,$F6,$57,$B3,$B9,$FF
	.dc.b	$47,$CE,$5B,$19,$7E,$07,$B5,$5F,$F4,$EB,$56,$5F,$EC,$1B,$FF,$51
	.dc.b	$9E,$7F,$6F,$FD,$94,$DF,$73,$84,$62,$BF,$D3,$ED,$7F,$DB,$3C,$0A
	.dc.b	$FF,$70,$7B,$BD,$F3,$FF,$70,$C6,$27,$55,$FE,$B7,$8F,$F5,$D1,$1C
	.dc.b	$1F,$E7,$D7,$FD,$88,$47,$07,$FA,$C3,$DF,$7B,$FF,$C2,$A1,$18,$AF
	.dc.b	$DB,$37,$9F,$EA,$36,$79,$57,$18,$C4,$7F,$B2,$5A,$FF,$BF,$9A,$F6
	.dc.b	$8E,$04,$AE,$ED,$AE,$ED,$9E,$F3,$A2,$4E,$9B,$1A,$8C,$11,$C1,$0F
	.dc.b	$34,$F5,$6C,$F2,$6C,$F1,$69,$C4,$F6,$4F,$16,$CF,$16,$CF,$16,$9C
	.dc.b	$4F,$85,$1B,$6C,$6D,$A6,$EB,$E4,$92,$BC,$11,$A3,$CA,$43,$B9,$4D
	.dc.b	$18,$21,$82,$47,$5A,$F2,$36,$BC,$8D,$9F,$25,$78,$A3,$11,$DD,$AF
	.dc.b	$66,$BD,$A3,$81,$2B,$BB,$6B,$BB,$67,$BC,$E8,$93,$A6,$C6,$A3,$04
	.dc.b	$70,$43,$CD,$3D,$5B,$3C,$9B,$3C,$5A,$71,$3D,$93,$C5,$B3,$C5,$B3
	.dc.b	$C5,$A7,$13,$E1,$46,$DB,$1B,$69,$BA,$F9,$24,$AF,$04,$68,$F2,$90
	.dc.b	$EE,$53,$46,$08,$60,$91,$D6,$BC,$8F,$D8,$F4,$BF,$CC,$F2,$3F,$CE
	.dc.b	$45,$3F,$CE,$3E,$8B,$D3,$FE,$75,$09,$E2,$FF,$05,$FA,$6F,$F0,$C3
	.dc.b	$FC,$31,$AF,$F1,$56,$BB,$C3,$9E,$09,$CC,$FF,$0D,$6B,$FC,$69,$B1
	.dc.b	$B9,$E3,$18,$23,$82,$4F,$8D,$AF,$8D,$AF,$8C,$D1,$0D,$14,$D1,$82
	.dc.b	$18,$21,$E9,$7A,$B4,$E4,$53,$89,$4E,$27,$BD,$E2,$D3,$89,$4E,$25
	.dc.b	$38,$CF,$C5,$76,$D7,$6D,$77,$3F,$21,$DB,$5D,$B5,$DC,$F9,$27,$94
	.dc.b	$7A,$48,$78,$CA,$E9,$C1,$1C,$12,$7B,$47,$06,$BB,$6C,$6E,$7C,$51
	.dc.b	$82,$38,$24,$FB,$1A,$FB,$1A,$FB,$0D,$10,$D1,$4D,$18,$21,$82,$1E
	.dc.b	$97,$AB,$4E,$45,$38,$94,$E2,$7B,$DE,$2D,$38,$94,$E2,$53,$8C,$FC
	.dc.b	$57,$6D,$76,$D7,$73,$F2,$1D,$B5,$DB,$5D,$CF,$92,$79,$47,$A4,$87
	.dc.b	$8C,$AE,$9C,$13,$C9,$78,$0F,$0F,$F8,$74,$FF,$11,$13,$A8,$CC,$7F
	.dc.b	$15,$8C,$65,$FE,$4D,$6C,$71,$F9,$6F,$E3,$98,$F0,$8F,$E3,$D1,$8A
	.dc.b	$FF,$2A,$36,$38,$1F,$EB,$18,$FF,$1B,$11,$FF,$27,$F0,$E5,$B5,$D5
	.dc.b	$70,$BE,$47,$51,$FE,$5A,$6B,$FC,$BE,$35,$3C,$5B,$7F,$E6,$7E,$57
	.dc.b	$37,$0B,$FF,$2B,$B6,$33,$7E,$16,$D7,$F9,$BE,$DD,$BC,$5B,$5D,$E3
	.dc.b	$07,$5E,$97,$55,$24,$FB,$63,$2F,$BF,$6F,$A9,$F0,$5F,$47,$87,$ED
	.dc.b	$37,$84,$D3,$C9,$BF,$FB,$75,$B7,$AD,$FF,$CF,$73,$96,$C6,$5F,$C0
	.dc.b	$5F,$E1,$FC,$8E,$C4,$8F,$C4,$63,$F8,$78,$FC,$55,$3B,$5F,$4D,$FE
	.dc.b	$ED,$F7,$73,$18,$94,$62,$75,$87,$F6,$39,$E5,$18,$8F,$1E,$8F,$81
	.dc.b	$F8,$04,$8F,$1E,$7F,$9D,$8F,$89,$10,$CC,$FF,$6C,$CF,$F3,$D1,$93
	.dc.b	$A9,$4F,$EE,$CE,$9F,$27,$8E,$F9,$5F,$ED,$39,$23,$0F,$FA,$C4,$C4
	.dc.b	$6A,$33,$3F,$19,$7B,$CA,$7B,$D9,$8C,$36,$33,$5E,$29,$CB,$EC,$3D
	.dc.b	$07,$EE,$B3,$FD,$F3,$8C,$7D,$4F,$B8,$E0,$43,$04,$3E,$0C,$72,$74
	.dc.b	$63,$2F,$95,$19,$8E,$07,$86,$3E,$16,$2E,$EF,$77,$CF,$FC,$16,$D7
	.dc.b	$CC,$6B,$FC,$2A,$78,$64,$AF,$91,$F8,$27,$CB,$24,$76,$7F,$C3,$39
	.dc.b	$7B,$AF,$CC,$7C,$1C,$BC,$CE,$0F,$59,$E7,$3F,$01,$4F,$94,$79,$BF
	.dc.b	$91,$9F,$AD,$18,$9F,$CC,$C7,$E5,$E3,$1C,$5F,$A2,$6C,$76,$C3,$C8
	.dc.b	$F7,$31,$87,$91,$C1,$F2,$A3,$D5,$7E,$3F,$4F,$F0,$DF,$86,$FF,$8F
	.dc.b	$A3,$01,$C6,$09,$DC,$F9,$5B,$7D,$70,$F0,$61,$E4,$3C,$BF,$84,$BE
	.dc.b	$F7,$F2,$9E,$87,$F9,$2A,$31,$8A,$31,$3A,$8F,$E9,$53,$A9,$F3,$70
	.dc.b	$3D,$EF,$F2,$BD,$7E,$A9,$B9,$7D,$F6,$BF,$A3,$9E,$13,$FD,$32,$3F
	.dc.b	$09,$F8,$4F,$E9,$23,$F6,$6D,$8C,$C7,$EA,$8C,$95,$F8,$1C,$9F,$98
	.dc.b	$23,$FB,$3B,$F0,$CF,$EE,$DB,$7E,$29,$F5,$C7,$99,$FA,$E7,$E2,$CE
	.dc.b	$A7,$E3,$EB,$AB,$F8,$EF,$27,$F4,$5E,$1F,$FA,$07,$E1,$1E,$67,$FA
	.dc.b	$06,$73,$49,$CB,$FE,$84,$AE,$15,$BE,$8F,$EF,$35,$AA,$FF,$43,$46
	.dc.b	$27,$55,$C2,$79,$1D,$63,$2D,$3F,$78,$FF,$4B,$23,$FC,$A1,$AF,$F3
	.dc.b	$26,$BC,$0F,$B9,$9F,$B0,$B1,$F2,$1C,$47,$DF,$6B,$C4,$E8,$77,$21
	.dc.b	$F7,$3E,$0F,$F1,$7A,$FF,$27,$6C,$9A,$77,$A9,$1E,$22,$57,$FA,$BF
	.dc.b	$7F,$CD,$F2,$7E,$6A,$7F,$80,$1F,$9A,$E1,$FD,$FB,$4C,$D7,$98,$87
	.dc.b	$A4,$8F,$ED,$3E,$E3,$F6,$7C,$BF,$CF,$5B,$06,$9F,$EB,$6E,$4F,$F1
	.dc.b	$86,$5F,$32,$FF,$95,$36,$33,$1F,$10,$F0,$1A,$3F,$89,$74,$7F,$4B
	.dc.b	$F5,$3F,$89,$36,$33,$82,$72,$D8,$E8,$73,$3D,$A5,$3B,$D7,$89,$CB
	.dc.b	$5F,$DC,$75,$7F,$5B,$C3,$F9,$BD,$7E,$56,$72,$D3,$FB,$44,$EE,$3F
	.dc.b	$A1,$43,$E8,$3F,$A3,$36,$33,$18,$3E,$27,$4F,$F9,$6E,$EF,$37,$7A
	.dc.b	$F4,$13,$1F,$33,$6B,$83,$F0,$23,$24,$8C,$C7,$B9,$AF,$DE,$E8,$FE
	.dc.b	$F3,$8F,$E4,$63,$FB,$A3,$5F,$EB,$5D,$BF,$97,$73,$C7,$FE,$01,$EF
	.dc.b	$F8,$11,$FD,$7D,$AF,$F2,$16,$D7,$F8,$28,$F6,$63,$E5,$CA,$E5,$E4
	.dc.b	$BF,$64,$60,$99,$FC,$4F,$5B,$FC,$8C,$49,$FF,$11,$BF,$E3,$1C,$3F
	.dc.b	$05,$1F,$E7,$AF,$0F,$FC,$6D,$C1,$92,$47,$F8,$AB,$D6,$FF,$28,$11
	.dc.b	$C1,$FE,$4A,$9F,$F2,$79,$1C,$1F,$E3,$27,$FC,$B0,$47,$07,$FA,$4F
	.dc.b	$C8,$FF,$30,$FA,$7F,$E9,$58,$46,$1D,$79,$5F,$E2,$99,$D4,$E6,$47
	.dc.b	$FA,$07,$CB,$D3,$FE,$9F,$F3,$3F,$CE,$B5,$FE,$79,$9C,$CF,$03,$FD
	.dc.b	$1A,$E1,$E7,$E6,$FF,$A0,$A7,$FD,$11,$E7,$66,$7F,$D0,$FA,$A7,$2D
	.dc.b	$8F,$F3,$6C,$7F,$9C,$88,$E0,$FF,$4E,$BF,$E8,$BF,$3F,$A6,$26,$9D
	.dc.b	$FD,$0F,$F5,$7C,$FF,$AB,$2B,$04,$BF,$03,$FE,$68,$8F,$F4,$31,$1F
	.dc.b	$F3,$1B,$F3,$7D,$17,$A5,$F9,$29,$46,$2B,$F5,$8D,$E3,$E1,$5B,$8C
	.dc.b	$57,$FB,$2D,$AF,$CB,$6B,$F5,$1A,$F6,$8E,$44,$AE,$ED,$AE,$F1,$82
	.dc.b	$38,$24,$F8,$67,$83,$67,$6D,$9D,$BE,$59,$DB,$67,$6D,$9D,$BE,$99
	.dc.b	$DB,$67,$6D,$9D,$BE,$D8,$CB,$63,$35,$F0,$89,$5D,$27,$47,$9C,$87
	.dc.b	$8C,$A7,$72,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$47,$92,$BD,$2D,$AF
	.dc.b	$4D,$79,$E3,$15,$E2,$6B,$DD,$AF,$66,$BD,$A3,$91,$2B,$BB,$6B,$BC
	.dc.b	$60,$8E,$09,$3E,$19,$E0,$D9,$DB,$67,$6F,$96,$76,$D9,$DB,$67,$6F
	.dc.b	$A6,$76,$D9,$DB,$67,$6F,$B6,$32,$D8,$CD,$7C,$22,$57,$49,$D1,$E7
	.dc.b	$21,$E3,$29,$DC,$A7,$03,$44,$34,$53,$46,$08,$60,$91,$E4,$AF,$4B
	.dc.b	$6B,$D3,$5F,$C1,$FC,$7F,$F2,$CD,$FF,$29,$3F,$CA,$7E,$2F,$F3,$96
	.dc.b	$9F,$CE,$E3,$6F,$F2,$FD,$9F,$23,$FE,$E3,$8F,$E8,$11,$8C,$B9,$36
	.dc.b	$33,$3B,$7C,$11,$96,$C6,$5E,$CD,$7F,$9D,$B5,$47,$17,$73,$E0,$9D
	.dc.b	$4E,$67,$8C,$E4,$8F,$91,$AF,$89,$AF,$43,$44,$34,$53,$46,$08,$60
	.dc.b	$87,$88,$E0,$53,$65,$36,$7B,$A7,$D6,$48,$F2,$C6,$C8,$6D,$CB,$4E
	.dc.b	$E5,$3B,$94,$EF,$3F,$1D,$C1,$D1,$C1,$27,$C5,$1D,$23,$13,$E0,$6C
	.dc.b	$66,$76,$F8,$23,$2D,$8C,$BE,$26,$BE,$16,$BE,$42,$47,$27,$84,$F8
	.dc.b	$67,$73,$99,$E5,$39,$23,$E5,$6B,$E3,$6B,$D0,$D1,$0D,$14,$D1,$82
	.dc.b	$18,$21,$E2,$38,$14,$D9,$4D,$9E,$E9,$F5,$92,$3C,$B1,$B2,$1B,$72
	.dc.b	$D3,$B9,$4E,$E5,$3B,$CF,$C7,$70,$74,$70,$49,$F1,$47,$48,$C4,$F8
	.dc.b	$1B,$19,$9D,$BE,$08,$CB,$63,$2F,$F8,$EF,$A5,$D2,$E9,$65,$FE,$4D
	.dc.b	$9C,$C6,$2B,$F8,$D4,$62,$75,$5C,$5F,$F2,$A7,$C6,$BC,$B6,$7C,$70
	.dc.b	$FF,$25,$4E,$63,$11,$FE,$54,$8C,$57,$2A,$E2,$FF,$97,$FD,$67,$AF
	.dc.b	$8B,$FE,$59,$18,$C3,$9C,$23,$15,$C6,$B8,$4F,$F9,$6F,$97,$FD,$BB
	.dc.b	$3F,$E6,$B6,$C7,$F9,$8F,$C8,$DB,$FE,$68,$8C,$BF,$CD,$79,$F2,$FF
	.dc.b	$97,$AB,$FC,$BA,$47,$FC,$D9,$EC,$6A,$8D,$26,$5E,$8E,$77,$AD,$D1
	.dc.b	$B1,$98,$C1,$E0,$D3,$E0,$7A,$FE,$D9,$34,$98,$FA,$E3,$2D,$8F,$81
	.dc.b	$A7,$D6,$6C,$EA,$7F,$D0,$1F,$1D,$E8,$87,$A7,$FE,$91,$8C,$37,$3F
	.dc.b	$BE,$D9,$DD,$7F,$61,$F9,$07,$CC,$6A,$30,$D7,$FD,$1D,$1A,$9E,$31
	.dc.b	$89,$D3,$FE,$9A,$E5,$FC,$AB,$F9,$2E,$6F,$ED,$91,$88,$FC,$BF,$37
	.dc.b	$E7,$5F,$EE,$56,$F2,$5F,$71,$F9,$6F,$AA,$CC,$62,$51,$89,$D5,$7F
	.dc.b	$A8,$E7,$94,$62,$3F,$D5,$B1,$D7,$EC,$3F,$B6,$34,$FF,$5C,$16,$3B
	.dc.b	$C6,$AD,$E9,$7E,$A3,$D9,$F1,$B6,$BC,$13,$AB,$56,$9B,$3F,$03,$4F
	.dc.b	$87,$9F,$C1,$8C,$1F,$1C,$F7,$F5,$75,$7D,$38,$7E,$AF,$0C,$E1,$D2
	.dc.b	$73,$5B,$7E,$95,$F5,$7F,$86,$35,$CE,$1E,$23,$E0,$E1,$FE,$D8,$BF
	.dc.b	$9B,$3E,$CF,$1B,$E5,$3F,$C2,$27,$5A,$FD,$49,$CD,$7D,$9F,$9A,$FE
	.dc.b	$1B,$F9,$87,$D8,$6E,$1F,$19,$AF,$B1,$AF,$5F,$AE,$E7,$89,$A4,$BF
	.dc.b	$2C,$FB,$32,$F5,$3F,$24,$F9,$67,$D4,$FC,$07,$99,$27,$8C,$7A,$E3
	.dc.b	$D3,$EE,$7E,$1B,$F5,$8F,$F1,$B5,$23,$0F,$E3,$5B,$5F,$8C,$6B,$F3
	.dc.b	$0F,$11,$E5,$29,$FC,$C1,$FE,$79,$1E,$23,$B5,$F8,$88,$E6,$3F,$33
	.dc.b	$F8,$EF,$91,$B1,$E3,$AF,$E7,$73,$96,$C7,$C2,$7C,$45,$3C,$C7,$E9
	.dc.b	$FC,$BF,$9D,$0F,$C9,$3E,$0C,$61,$B9,$FF,$48,$6C,$EC,$FD,$87,$A4
	.dc.b	$F9,$5F,$D3,$46,$1B,$87,$EC,$A7,$53,$FA,$63,$45,$8F,$99,$F8,$47
	.dc.b	$F0,$18,$FA,$0D,$8C,$C7,$F0,$23,$25,$7D,$E4,$3D,$E5,$52,$F8,$EE
	.dc.b	$93,$FD,$92,$31,$3C,$E3,$13,$A8,$E2,$F0,$87,$5F,$F1,$D7,$C0,$63
	.dc.b	$9B,$F6,$0D,$DF,$F2,$EF,$D3,$30,$7E,$18,$FF,$48,$FC,$13,$D0,$44
	.dc.b	$F3,$DF,$05,$FE,$01,$AD,$4E,$63,$15,$C2,$39,$B6,$B8,$F9,$7D,$2F
	.dc.b	$8D,$68,$E5,$19,$AE,$17,$D2,$7F,$84,$C7,$F3,$D2,$3F,$C7,$5A,$FE
	.dc.b	$D5,$AF,$03,$F0,$F9,$FF,$84,$BC,$AF,$D7,$ED,$F4,$1B,$19,$DB,$62
	.dc.b	$3F,$D4,$DB,$7F,$88,$B0,$EB,$87,$CD,$7B,$D7,$EE,$77,$E5,$3F,$88
	.dc.b	$BF,$D5,$46,$23,$F6,$8D,$9F,$C2,$F0,$7F,$89,$BD,$CF,$2B,$63,$F2
	.dc.b	$F1,$84,$FE,$93,$F2,$49,$1F,$AD,$E4,$FF,$17,$C6,$30,$F8,$87,$C7
	.dc.b	$21,$83,$F8,$61,$FE,$36,$FC,$77,$F7,$94,$69,$1F,$E2,$9C,$7F,$2E
	.dc.b	$6C,$7C,$76,$E7,$FD,$BB,$A3,$F8,$8B,$F2,$36,$F9,$53,$E2,$3E,$B1
	.dc.b	$E2,$A7,$8D,$E2,$23,$FC,$44,$C9,$FC,$BC,$F0,$47,$F5,$8E,$BF,$23
	.dc.b	$63,$2F,$F3,$16,$BF,$64,$F1,$14,$F8,$87,$F4,$57,$89,$FE,$07,$D3
	.dc.b	$AC,$67,$0F,$F0,$3C,$62,$7F,$AA,$0F,$D4,$7F,$AA,$36,$BF,$AA,$35
	.dc.b	$FE,$A8,$D7,$FB,$07,$81,$CF,$BB,$FB,$36,$1F,$D8,$6B,$C7,$18,$AF
	.dc.b	$4E,$7F,$31,$F4,$9E,$B3,$FB,$57,$85,$F8,$D7,$FC,$12,$D7,$FB,$DD
	.dc.b	$7E,$31,$CB,$5E,$73,$D5,$FA,$E4,$8F,$F1,$57,$89,$CE,$3F,$BC,$36
	.dc.b	$32,$73,$8C,$B4,$FF,$12,$72,$FF,$88,$9F,$E0,$B1,$FE,$01,$F1,$7F
	.dc.b	$D1,$33,$93,$FC,$65,$3F,$0B,$C6,$F8,$C4,$7E,$EE,$1E,$7B,$F8,$44
	.dc.b	$71,$C1,$FE,$4D,$E9,$FE,$31,$3F,$C8,$23,$0F,$F9,$52,$BF,$9C,$B6
	.dc.b	$7F,$CB,$84,$78,$F9,$5F,$E6,$58,$C4,$7F,$A3,$7C,$CF,$F5,$0F,$B5
	.dc.b	$FE,$6C,$24,$7F,$9C,$DF,$F3,$C1,$1F,$F4,$DE,$DC,$1B,$82,$7C,$F1
	.dc.b	$FC,$57,$CF,$FF,$56,$F8,$98,$E2,$FF,$43,$B6,$33,$3E,$E9,$CB,$FE
	.dc.b	$94,$B5,$7F,$A8,$7C,$DD,$CF,$FA,$95,$AF,$FA,$85,$AB,$D5,$7A,$58
	.dc.b	$6D,$F4,$9F,$9B,$18,$BF,$4B,$63,$FD,$29,$1F,$E9,$02,$38,$3E,$0B
	.dc.b	$FE,$9D,$23,$83,$EC,$FA,$9F,$EC,$B8,$FF,$64,$57,$86,$7F,$D9,$FE
	.dc.b	$A7,$A2,$B5,$1F,$E6,$5A,$FF,$44,$11,$C1,$E8,$F5,$7F,$22,$DB,$FC
	.dc.b	$AB,$F0,$63,$2F,$F6,$1C,$11,$88,$FE,$D7,$18,$9E,$2D,$E5,$F8,$75
	.dc.b	$BA,$E9,$5A,$8C,$56,$E7,$C5,$5C,$2B,$44,$70,$48,$D9,$0F,$35,$74
	.dc.b	$24,$F1,$8D,$1E,$88,$CC,$F3,$8E,$33,$B6,$C6,$5E,$11,$96,$C6,$5F
	.dc.b	$6B,$6B,$D9,$1C,$8F,$09,$A2,$BA,$79,$98,$AF,$2B,$6B,$C0,$F1,$3A
	.dc.b	$10,$E2,$53,$46,$08,$60,$93,$E1,$6C,$F8,$1E,$51,$92,$46,$63,$9C
	.dc.b	$62,$7C,$AD,$AF,$85,$5B,$AE,$B5,$A8,$C5,$6E,$7D,$D5,$C2,$B4,$47
	.dc.b	$04,$8D,$90,$F3,$D7,$42,$4F,$18,$D1,$E8,$8C,$CF,$38,$E3,$3B,$6C
	.dc.b	$65,$E1,$19,$6C,$65,$F6,$B6,$BD,$91,$C8,$F0,$9A,$2B,$A7,$99,$8A
	.dc.b	$F2,$B6,$BC,$0F,$13,$A1,$0E,$25,$34,$60,$86,$09,$3E,$16,$CF,$81
	.dc.b	$E5,$19,$24,$66,$39,$C7,$1F,$23,$FC,$C7,$1F,$CE,$9A,$FF,$52,$6B
	.dc.b	$FE,$62,$F1,$FF,$9F,$BF,$CB,$FC,$86,$18,$B4,$62,$7F,$96,$B5,$FE
	.dc.b	$62,$D7,$F8,$B3,$5F,$E2,$AD,$7B,$4F,$58,$C4,$F6,$8C,$12,$7B,$B5
	.dc.b	$EE,$D7,$BB,$5E,$ED,$7B,$9C,$88,$72,$29,$C8,$A7,$22,$9C,$8E,$04
	.dc.b	$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$48,$F8,$CF,$91,$AF
	.dc.b	$76,$BD,$1A,$F3,$6B,$CE,$7E,$89,$27,$E4,$4E,$89,$19,$9F,$03,$5F
	.dc.b	$03,$5F,$03,$5F,$03,$5F,$04,$F6,$8C,$4F,$68,$C1,$27,$BB,$5E,$ED
	.dc.b	$7B,$B5,$EE,$D7,$B9,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81
	.dc.b	$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$8F,$8C,$F9,$1A,$F7,$6B
	.dc.b	$D1,$AF,$36,$BC,$E7,$E8,$92,$7E,$44,$E8,$91,$99,$F0,$35,$F0,$35
	.dc.b	$F0,$35,$FC,$D7,$3B,$C9,$78,$4C,$7F,$C9,$F3,$98,$C5,$7F,$9A,$23
	.dc.b	$13,$AA,$E2,$FF,$95,$BD,$57,$B1,$FC,$62,$30,$DB,$FF,$38,$CE,$B3
	.dc.b	$EB,$02,$46,$6B,$FC,$C3,$EB,$3A,$39,$56,$C4,$66,$F9,$C7,$38,$C5
	.dc.b	$72,$AE,$33,$DB,$9E,$AC,$11,$86,$E1,$D2,$75,$3C,$36,$FF,$30,$E1
	.dc.b	$FE,$5E,$E5,$C1,$2F,$84,$BE,$4F,$81,$23,$2F,$7D,$3F,$9F,$FA,$FF
	.dc.b	$22,$4D,$26,$5F,$E8,$67,$D9,$1A,$8C,$10,$F8,$D9,$FC,$FE,$7E,$84
	.dc.b	$8F,$A7,$7F,$D3,$A3,$FD,$1F,$DC,$F7,$BF,$A8,$FB,$18,$6D,$FF,$AA
	.dc.b	$A7,$58,$AB,$4D,$9F,$C4,$71,$7E,$39,$FC,$47,$27,$F5,$EE,$4F,$C7
	.dc.b	$46,$0F,$F5,$AD,$7F,$A5,$BE,$07,$FB,$0F,$AF,$8E,$B4,$62,$BF,$05
	.dc.b	$5B,$15,$B9,$D4,$FE,$36,$5E,$5B,$E5,$7F,$B2,$46,$1B,$19,$8F,$F5
	.dc.b	$89,$92,$AF,$98,$FE,$EB,$ED,$7A,$23,$12,$8C,$4E,$AF,$CF,$5A,$9F
	.dc.b	$81,$1F,$1B,$4F,$89,$AF,$C7,$8C,$1F,$07,$C7,$E4,$FA,$F5,$46,$23
	.dc.b	$FD,$8A,$DC,$BF,$D7,$EE,$61,$B7,$D0,$7A,$90,$FD,$63,$97,$B7,$91
	.dc.b	$F2,$9F,$F6,$8B,$6B,$FD,$96,$DB,$39,$CD,$F7,$FF,$E1,$AF,$DC,$3E
	.dc.b	$33,$70,$F8,$6D,$7E,$B4,$F0,$C1,$5F,$37,$F0,$1E,$84,$8F,$BF,$5F
	.dc.b	$32,$3C,$CF,$BB,$4F,$94,$FB,$1A,$FF,$0F,$CF,$DC,$D8,$CD,$7F,$B9
	.dc.b	$BC,$EF,$61,$1E,$F3,$F4,$63,$E1,$79,$7F,$8D,$F9,$FF,$F1,$95,$23
	.dc.b	$0F,$E5,$9B,$5F,$93,$6B,$E1,$E0,$FC,$09,$1F,$C2,$9F,$C9,$7A,$5F
	.dc.b	$5E,$F8,$1F,$D2,$E1,$FC,$8E,$73,$18,$AE,$15,$BA,$FD,$29,$B7,$1D
	.dc.b	$1E,$8C,$7F,$4D,$F8,$6D,$BF,$F2,$8B,$6C,$D3,$FC,$9D,$B7,$CA,$9F
	.dc.b	$A1,$9F,$A5,$FC,$61,$FE,$5A,$B7,$E0,$DF,$E5,$CF,$C1,$FC,$0D,$C7
	.dc.b	$E3,$36,$BE,$33,$5F,$F3,$1F,$1B,$EC,$7F,$63,$5F,$AC,$6C,$FE,$29
	.dc.b	$AF,$06,$C7,$F0,$2F,$87,$F8,$D3,$D1,$B1,$93,$E3,$C7,$E5,$23,$6F
	.dc.b	$B8,$FA,$A7,$D9,$2C,$7F,$0E,$F4,$39,$D9,$A4,$8E,$B3,$C4,$F8,$E7
	.dc.b	$52,$E9,$F3,$49,$C9,$F9,$A8,$D1,$FE,$89,$3F,$D0,$BE,$7D,$F3,$4D
	.dc.b	$24,$7F,$6F,$F3,$BF,$D2,$B3,$A8,$C7,$C0,$7F,$80,$23,$10,$9D,$D7
	.dc.b	$28,$FB,$EF,$F3,$C2,$3F,$C6,$DA,$FE,$C7,$87,$9B,$F9,$F8,$CB,$F7
	.dc.b	$33,$FF,$4F,$FA,$5F,$AA,$D7,$E4,$6A,$8C,$4E,$A3,$F9,$5B,$6B,$B5
	.dc.b	$F4,$7C,$A7,$8C,$FE,$63,$B6,$DF,$C0,$57,$53,$25,$7F,$00,$D7,$F1
	.dc.b	$4B,$DD,$F3,$70,$7F,$AD,$9B,$19,$3F,$28,$59,$FC,$DA,$FC,$1F,$F4
	.dc.b	$2D,$3F,$42,$58,$DC,$7E,$BB,$83,$FC,$6F,$8F,$B7,$87,$FC,$6C,$FC
	.dc.b	$AB,$FE,$21,$7F,$EC,$8F,$C3,$3F,$70,$48,$FE,$25,$19,$7F,$81,$74
	.dc.b	$FC,$3E,$AF,$E1,$31,$EF,$6E,$FF,$EC,$D6,$C7,$CB,$AF,$9E,$53,$EE
	.dc.b	$1D,$97,$CF,$FF,$09,$23,$97,$F9,$3F,$84,$EE,$FC,$0D,$7F,$7C,$8C
	.dc.b	$57,$E0,$63,$13,$A7,$F9,$94,$65,$F1,$90,$F1,$95,$FF,$00,$F7,$F1
	.dc.b	$8F,$E1,$6D,$8C,$9C,$E3,$2D,$3F,$C1,$3E,$0F,$FB,$D3,$F6,$4F,$EB
	.dc.b	$70,$69,$27,$D5,$E1,$F5,$C7,$FB,$2D,$73,$6E,$9F,$E1,$D9,$F2,$57
	.dc.b	$E6,$0F,$51,$4F,$59,$FD,$FA,$BF,$24,$FC,$7F,$13,$F2,$4D,$8C,$BF
	.dc.b	$E0,$78,$CB,$D3,$87,$F1,$2F,$D9,$3B,$4F,$F8,$9B,$C6,$DC,$7F,$30
	.dc.b	$6C,$64,$EB,$19,$69,$FE,$27,$E2,$FF,$12,$78,$E8,$FE,$02,$66,$7F
	.dc.b	$C4,$71,$99,$FF,$1C,$4F,$C5,$F2,$3E,$41,$1F,$A9,$5F,$0C,$98,$7A
	.dc.b	$9F,$8F,$E5,$7F,$95,$3C,$3F,$F2,$99,$23,$FC,$B1,$E5,$FF,$53,$7F
	.dc.b	$CC,$36,$F6,$7F,$CE,$04,$8C,$BF,$E6,$EF,$66,$47,$FA,$44,$FF,$48
	.dc.b	$FA,$DE,$D8,$CC,$7F,$9C,$3C,$FC,$3D,$A3,$FC,$F8,$47,$97,$A7,$FE
	.dc.b	$85,$C1,$C9,$D6,$53,$98,$C3,$63,$8F,$A7,$F3,$67,$FD,$25,$5C,$A7
	.dc.b	$AF,$A9,$96,$C6,$5F,$F5,$55,$A7,$51,$FE,$88,$F5,$AB,$FE,$A7,$AE
	.dc.b	$0D,$BF,$F6,$04,$EA,$3E,$1C,$E9,$B3,$FE,$96,$23,$F4,$1F,$F4,$F1
	.dc.b	$1C,$1F,$E6,$37,$FD,$0E,$47,$07,$C8,$F2,$7F,$10,$D9,$FE,$9A,$FC
	.dc.b	$88,$CB,$EE,$6B,$FE,$CF,$C1,$18,$8E,$13,$CA,$72,$D9,$D4,$63,$8F
	.dc.b	$E9,$C6,$09,$3B,$24,$69,$B5,$E7,$8C,$11,$C1,$0F,$44,$70,$6C,$6D
	.dc.b	$A6,$CF,$64,$F8,$A3,$51,$89,$E4,$D8,$CB,$F0,$1B,$5E,$D8,$E8,$64
	.dc.b	$AE,$5F,$09,$0E,$86,$AB,$C4,$DA,$F0,$BC,$CE,$05,$38,$98,$9E,$CE
	.dc.b	$08,$6C,$8E,$9B,$3E,$07,$93,$5E,$31,$B2,$46,$63,$8C,$F9,$A7,$2D
	.dc.b	$9D,$46,$2B,$E3,$C6,$09,$3B,$24,$69,$B5,$F0,$E3,$04,$70,$43,$D3
	.dc.b	$1C,$1B,$1B,$69,$B3,$DB,$3E,$28,$D4,$62,$79,$36,$32,$FC,$06,$D7
	.dc.b	$B6,$3A,$19,$2B,$97,$C2,$43,$A1,$AA,$F1,$36,$BC,$2F,$33,$81,$4E
	.dc.b	$26,$27,$B3,$82,$1B,$23,$A6,$CF,$81,$E4,$D7,$8C,$6C,$91,$98,$FC
	.dc.b	$67,$92,$F2,$9E,$B7,$F3,$39,$FF,$31,$46,$1F,$E6,$D3,$B9,$D3,$63
	.dc.b	$F9,$BF,$C6,$7F,$61,$6E,$3C,$F1,$4E,$67,$FA,$1B,$5F,$E6,$AD,$7F
	.dc.b	$8C,$35,$FE,$8A,$D8,$E7,$3D,$63,$04,$70,$49,$F1,$90,$F0,$B9,$69
	.dc.b	$93,$C6,$E5,$A6,$4F,$2B,$96,$99,$3C,$EE,$5A,$66,$7D,$2D,$9F,$43
	.dc.b	$67,$84,$EA,$3C,$A4,$3C,$65,$3C,$25,$3B,$95,$E4,$F0,$21,$C0,$A7
	.dc.b	$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$4F,$BA,$3C,$8D,$7B,$B5,$E8
	.dc.b	$D7,$9B,$63,$9C,$FC,$68,$C1,$1C,$12,$7E,$61,$0F,$0B,$96,$99,$3C
	.dc.b	$6E,$5A,$64,$F2,$B9,$69,$93,$CE,$E5,$A6,$67,$D2,$D9,$F4,$36,$78
	.dc.b	$4E,$A3,$CA,$43,$C6,$53,$C2,$53,$B9,$5E,$4F,$02,$1C,$0A,$70,$2B
	.dc.b	$C1,$D1,$0D,$15,$D3,$82,$38,$24,$FB,$A3,$C8,$D7,$BB,$5E,$8D,$5C
	.dc.b	$CF,$15,$E6,$7F,$15,$9F,$F7,$86,$5F,$E4,$BA,$D4,$E6,$47,$2A,$E3
	.dc.b	$5F,$E5,$1F,$69,$E2,$FF,$9E,$63,$0D,$AF,$F3,$14,$EB,$3E,$90,$8C
	.dc.b	$57,$F9,$6F,$D7,$73,$71,$AF,$F3,$35,$F0,$9C,$B6,$75,$3E,$6F,$65
	.dc.b	$D1,$88,$FF,$27,$E1,$FE,$67,$9C,$DF,$89,$B1,$FE,$6B,$FC,$1F,$F9
	.dc.b	$2B,$77,$E5,$18,$BF,$F3,$E3,$5F,$8A,$7C,$5D,$9E,$D7,$CB,$DF,$E7
	.dc.b	$B6,$33,$3F,$E8,$5E,$0F,$9F,$3F,$23,$3C,$CF,$4D,$FE,$44,$E4,$91
	.dc.b	$C7,$0F,$B7,$C1,$D4,$91,$9C,$7E,$AB,$F4,$A3,$FD,$5C,$77,$23,$9C
	.dc.b	$36,$FB,$31,$FC,$1B,$E9,$21,$D4,$FF,$4F,$43,$A9,$EE,$7F,$AE,$FD
	.dc.b	$AC,$C6,$25,$18,$9D,$57,$CA,$9E,$51,$88,$FC,$8C,$3D,$37,$5F,$F5
	.dc.b	$F8,$C3,$63,$31,$FE,$A5,$32,$57,$FD,$74,$BD,$7F,$4C,$61,$BD,$FF
	.dc.b	$ED,$36,$CE,$EF,$E0,$10,$F8,$DB,$7E,$B1,$F8,$A7,$A3,$37,$D2,$73
	.dc.b	$8F,$74,$62,$BA,$46,$E7,$4D,$8C,$BF,$B2,$D7,$E9,$BE,$13,$17,$E1
	.dc.b	$71,$2F,$8F,$FF,$76,$3F,$42,$3E,$74,$63,$6F,$B2,$D8,$CC,$FD,$17
	.dc.b	$FD,$E9,$A7,$EC,$3F,$64,$F8,$FD,$3F,$EF,$76,$D1,$A7,$CC,$F6,$39
	.dc.b	$47,$D9,$6D,$1A,$77,$BF,$D5,$1F,$28,$F6,$F1,$FF,$8B,$74,$F8,$31
	.dc.b	$F8,$4C,$BF,$8E,$61,$ED,$8C,$69,$E0,$9C,$E9,$EE,$8F,$1F,$C8,$FF
	.dc.b	$AE,$1F,$CC,$B6,$BF,$24,$D7,$F2,$2D,$7C,$67,$94,$F3,$9F,$4F,$91
	.dc.b	$EC,$FF,$8B,$AF,$F8,$2C,$FF,$26,$8C,$3F,$C2,$27,$7D,$3F,$3D,$B9
	.dc.b	$7A,$E4,$B0,$F7,$FB,$BF,$E4,$38,$C6,$5A,$8C,$4E,$A7,$D7,$39,$7F
	.dc.b	$11,$5E,$83,$F3,$3A,$FC,$D9,$FC,$54,$7C,$0E,$CF,$0F,$DF,$74,$6D
	.dc.b	$FC,$66,$CE,$E1,$3E,$5E,$17,$C4,$FE,$AE,$7F,$7D,$18,$7F,$64,$62
	.dc.b	$35,$5F,$C0,$27,$5C,$FC,$FE,$2F,$F8,$65,$F2,$6D,$6D,$AF,$E6,$1A
	.dc.b	$FB,$CF,$B2,$7C,$2E,$3F,$C5,$BF,$9E,$C3,$8D,$7F,$0F,$6E,$1E,$7E
	.dc.b	$2F,$64,$74,$BE,$37,$D4,$8E,$09,$A7,$CC,$8F,$09,$F4,$0F,$F4,$2F
	.dc.b	$89,$FE,$87,$F1,$5E,$37,$F4,$CF,$1B,$FD,$0D,$18,$AF,$EE,$71,$89
	.dc.b	$D5,$71,$6F,$87,$FD,$E2,$F8,$5F,$FA,$52,$31,$3A,$84,$EA,$FA,$47
	.dc.b	$EC,$27,$FA,$C1,$1F,$E9,$8D,$7F,$12,$D7,$81,$F6,$F5,$7B,$7F,$21
	.dc.b	$A6,$5C,$1F,$E9,$BC,$FF,$C2,$35,$CE,$BF,$B6,$3D,$6B,$FA,$16,$FF
	.dc.b	$A4,$BF,$C2,$47,$4A,$FC,$13,$5F,$C1,$34,$FB,$83,$D3,$F2,$11,$F4
	.dc.b	$3F,$9F,$5F,$7B,$48,$FE,$84,$C9,$F1,$26,$1F,$10,$A7,$EE,$B9,$BF
	.dc.b	$8B,$FB,$3F,$B6,$E0,$FD,$EB,$68,$D3,$FD,$81,$D3,$FC,$13,$2F,$E1
	.dc.b	$6D,$C3,$F8,$57,$5F,$FA,$F2,$72,$D8,$F0,$9F,$28,$BB,$7C,$E7,$F8
	.dc.b	$FF,$07,$F0,$EF,$57,$F9,$1B,$63,$39,$27,$2D,$8F,$E9,$79,$FD,$36
	.dc.b	$C7,$E0,$1B,$8F,$F3,$18,$C4,$7C,$52,$1F,$88,$2B,$FD,$1C,$DB,$FD
	.dc.b	$4B,$BB,$B3,$FC,$15,$AE,$63,$67,$F8,$1F,$C0,$FE,$53,$FB,$27,$F5
	.dc.b	$62,$4F,$F1,$28,$CB,$FD,$82,$78,$F8,$1F,$DA,$30,$FE,$CB,$1D,$8F
	.dc.b	$ED,$87,$82,$7E,$61,$EA,$29,$EB,$3F,$9C,$F8,$7F,$DF,$F8,$FF,$C4
	.dc.b	$7E,$1F,$F3,$C6,$D7,$E4,$EB,$FC,$0B,$39,$7E,$BE,$9F,$E7,$8E,$6F
	.dc.b	$D2,$70,$7D,$99,$FE,$4D,$E2,$6D,$B8,$73,$6E,$3F,$E1,$A3,$FC,$3F
	.dc.b	$E2,$FF,$4C,$C2,$D7,$F0,$6A,$FE,$FF,$39,$7F,$9B,$B5,$FF,$0E,$78
	.dc.b	$DF,$0F,$4F,$F1,$E6,$BF,$E3,$BB,$FA,$31,$8A,$F4,$F8,$7D,$5F,$8A
	.dc.b	$4F,$23,$FD,$09,$A7,$F9,$84,$91,$FE,$53,$7F,$CD,$04,$7F,$CB,$8D
	.dc.b	$7F,$CC,$DC,$FF,$39,$FF,$2D,$5F,$F9,$AC,$8F,$F9,$8D,$FF,$39,$11
	.dc.b	$C1,$FE,$7C,$F1,$BF,$B7,$F9,$BF,$E8,$1C,$8D,$24,$FF,$7C,$F1,$BF
	.dc.b	$D1,$71,$FE,$8D,$F2,$F8,$CE,$5A,$FF,$A1,$A7,$6F,$F8,$13,$CC,$FF
	.dc.b	$4D,$D7,$FA,$87,$9F,$73,$FE,$B0,$9C,$BB,$8C,$BF,$EA,$9C,$EE,$BF
	.dc.b	$EA,$DB,$FF,$45,$12,$7F,$D0,$4F,$FA,$44,$8E,$0F,$F0,$AF,$9A,$D3
	.dc.b	$EF,$47,$FB,$02,$31,$A7,$D9,$6F,$1F,$F8,$72,$51,$87,$9D,$72,$72
	.dc.b	$D8,$D4,$7F,$AA,$1A,$F7,$8D,$12,$BC,$0D,$AF,$05,$70,$AE,$B5,$A2
	.dc.b	$4E,$AB,$84,$F9,$27,$2D,$9D,$46,$23,$C9,$3C,$1B,$19,$7C,$CD,$7C
	.dc.b	$33,$C8,$8F,$52,$1C,$9C,$B4,$C9,$D1,$CB,$4C,$9D,$8B,$5F,$04,$95
	.dc.b	$E7,$8E,$07,$52,$18,$21,$CC,$C9,$4C,$C7,$4A,$F7,$36,$BC,$B3,$B2
	.dc.b	$46,$5F,$2D,$78,$DC,$B6,$35,$1E,$06,$BE,$08,$D1,$2B,$C2,$DA,$F0
	.dc.b	$D7,$0A,$EB,$5A,$24,$EA,$B8,$4F,$92,$72,$D9,$D4,$62,$3C,$93,$C1
	.dc.b	$B1,$97,$CC,$D7,$C3,$3C,$88,$F5,$21,$C9,$CB,$4C,$9D,$1C,$B4,$C9
	.dc.b	$D8,$B5,$F0,$49,$5E,$78,$E0,$75,$21,$82,$1C,$CC,$94,$CC,$74,$AF
	.dc.b	$73,$6B,$CB,$3B,$24,$65,$F2,$D7,$8D,$CB,$63,$EC,$78,$4F,$11,$E2
	.dc.b	$7F,$0B,$BF,$F2,$B1,$FE,$64,$71,$3B,$75,$4F,$BA,$FF,$52,$D7,$F0
	.dc.b	$46,$BF,$C7,$1A,$FF,$0F,$6B,$FC,$75,$AF,$F1,$A6,$C7,$5C,$BB,$C6
	.dc.b	$08,$E0,$93,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$6C,$F8,$63,$A1
	.dc.b	$0E,$85,$3A,$14,$E8,$53,$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8,$AF
	.dc.b	$27,$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$7E,$9C
	.dc.b	$7B,$5A,$F9,$9A,$F8,$5A,$F6,$6B,$D5,$B1,$D6,$7E,$9C,$60,$8E,$09
	.dc.b	$3E,$16,$BE,$16,$BE,$16,$BE,$16,$BE,$16,$CF,$86,$3A,$10,$E8,$53
	.dc.b	$A1,$4E,$85,$3A,$15,$E8,$F2,$21,$C8,$A7,$22,$9C,$8A,$F2,$78,$10
	.dc.b	$E0,$53,$81,$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$27,$E9,$C7,$B5,$AF
	.dc.b	$99,$AF,$85,$AF,$66,$BD,$5B,$1E,$1E,$97,$53,$9B,$F8,$BD,$88,$27
	.dc.b	$71,$A8,$CC,$63,$1F,$F1,$43,$63,$89,$FE,$81,$C7,$83,$D7,$AB,$B4
	.dc.b	$62,$35,$5F,$E3,$19,$CB,$FE,$49,$9D,$71,$B8,$D1,$86,$CE,$CF,$F1
	.dc.b	$D7,$80,$E4,$CE,$3E,$48,$C5,$F8,$5B,$5C,$2B,$75,$F0,$BD,$37,$D3
	.dc.b	$62,$3E,$34,$63,$0F,$53,$6F,$FC,$94,$D9,$F8,$7A,$7C,$AF,$3F,$E3
	.dc.b	$EF,$F4,$5B,$B9,$A7,$CA,$C5,$F3,$EF,$27,$E0,$FD,$16,$63,$12,$8C
	.dc.b	$4E,$B2,$FB,$73,$CA,$31,$1F,$6B,$1F,$C1,$6D,$E2,$24,$7A,$23,$2D
	.dc.b	$CB,$FA,$E6,$7F,$84,$8F,$49,$0F,$C4,$D3,$47,$C1,$FD,$8F,$37,$FD
	.dc.b	$FF,$18,$CD,$18,$9D,$5F,$2A,$D4,$FC,$3F,$39,$F8,$1F,$F6,$0C,$11
	.dc.b	$87,$FC,$F4,$62,$35,$19,$BE,$5F,$47,$F8,$AF,$46,$D9,$A4,$9F,$80
	.dc.b	$7B,$4F,$91,$AF,$C2,$DB,$F5,$51,$E5,$F1,$BA,$3F,$43,$14,$62,$3F
	.dc.b	$43,$18,$C3,$C1,$1B,$9D,$3F,$B3,$DF,$E7,$3E,$22,$1E,$23,$F5,$FB
	.dc.b	$BF,$44,$F8,$77,$F3,$27,$E9,$C6,$36,$DB,$63,$38,$FD,$07,$CB,$F4
	.dc.b	$BF,$AE,$E7,$FA,$0D,$8F,$63,$5D,$E1,$FA,$67,$E0,$7C,$27,$37,$F6
	.dc.b	$F6,$8C,$4F,$D8,$6C,$66,$3D,$13,$ED,$3D,$E7,$C3,$D7,$E7,$B8,$F8
	.dc.b	$C7,$E7,$2F,$B8,$F0,$46,$1B,$19,$BE,$B4,$AD,$46,$07,$94,$F2,$FF
	.dc.b	$87,$FC,$CF,$EC,$E3,$12,$8C,$4E,$AF,$F5,$F5,$A9,$FC,$44,$7A,$79
	.dc.b	$3F,$10,$47,$EC,$FE,$73,$FC,$93,$8A,$C7,$C9,$E3,$4E,$63,$F5,$AD
	.dc.b	$7F,$95,$E1,$FC,$97,$27,$E2,$1F,$E4,$87,$D7,$E1,$7F,$28,$8C,$4E
	.dc.b	$DB,$19,$9F,$9F,$5F,$D0,$78,$3E,$7B,$F4,$C8,$7D,$CF,$47,$F9,$9F
	.dc.b	$EC,$5D,$F8,$D1,$8B,$FD,$13,$6B,$F4,$2D,$7F,$9D,$FE,$2F,$FC,$F2
	.dc.b	$7F,$64,$7F,$18,$71,$D7,$EB,$6C,$64,$FD,$D1,$FC,$7B,$DA,$7D,$EF
	.dc.b	$F0,$F8,$C3,$6F,$FA,$8C,$EA,$7F,$88,$9A,$2C,$7B,$8F,$85,$C5,$F5
	.dc.b	$DF,$C9,$C2,$FD,$1C,$5E,$58,$ED,$3E,$C6,$9B,$32,$F4,$3F,$0A,$53
	.dc.b	$E8,$F6,$7F,$00,$7C,$4D,$7F,$14,$7F,$A2,$C7,$8E,$F3,$FF,$BC,$F9
	.dc.b	$BF,$E8,$D9,$CC,$63,$CE,$FE,$FF,$5B,$9F,$F0,$05,$6E,$31,$1F,$E9
	.dc.b	$86,$C6,$6B,$97,$95,$FE,$07,$9C,$DF,$49,$FE,$4A,$7F,$5C,$25,$FF
	.dc.b	$49,$6B,$A3,$04,$3F,$C0,$B9,$FE,$0F,$D4,$F9,$99,$9A,$47,$84,$FE
	.dc.b	$B6,$BA,$B4,$FB,$93,$FA,$3D,$BF,$00,$DD,$FF,$C1,$6D,$BE,$C7,$E3
	.dc.b	$47,$9B,$E2,$23,$97,$F4,$1C,$5F,$E1,$4F,$D8,$AC,$9A,$47,$CF,$8F
	.dc.b	$E3,$32,$F3,$B4,$FE,$13,$CF,$FE,$1E,$FC,$47,$F6,$B4,$69,$1F,$E0
	.dc.b	$73,$EA,$2B,$F1,$DB,$97,$F0,$C8,$C4,$7D,$12,$1F,$1C,$AF,$F2,$2E
	.dc.b	$EF,$F1,$7F,$CE,$FF,$1E,$83,$49,$3F,$E4,$8F,$03,$F9,$5D,$FF,$90
	.dc.b	$9B,$19,$F0,$7F,$99,$4E,$5B,$1F,$13,$9B,$FC,$9F,$A7,$C4,$6C,$7F
	.dc.b	$4C,$C7,$D2,$FF,$4C,$F0,$FE,$03,$FD,$B1,$AE,$63,$67,$F8,$1B,$C4
	.dc.b	$FF,$24,$FB,$7F,$D7,$20,$D2,$4F,$F6,$48,$EF,$E1,$FF,$6D,$EE,$EA
	.dc.b	$D8,$CF,$8F,$FD,$C3,$5F,$47,$83,$FE,$6D,$D6,$F7,$FE,$B3,$2F,$EC
	.dc.b	$7E,$47,$F9,$6B,$A3,$FC,$11,$97,$E3,$A7,$FC,$04,$D7,$FC,$0F,$3B
	.dc.b	$7E,$36,$1D,$5B,$7D,$6B,$FC,$47,$E4,$EE,$7F,$C1,$ED,$7F,$B9,$B6
	.dc.b	$36,$77,$8F,$EE,$FE,$57,$F9,$BB,$2F,$F1,$83,$63,$FC,$38,$D7,$FC
	.dc.b	$36,$D7,$FC,$E1,$E5,$FF,$8F,$74,$FF,$1E,$5F,$CA,$8C,$4F,$F8,$F7
	.dc.b	$CC,$FA,$EF,$E8,$4F,$C9,$94,$FC,$A1,$FD,$A7,$A3,$FC,$CE,$47,$FC
	.dc.b	$A8,$FF,$98,$29,$ED,$52,$4F,$F9,$8B,$D0,$FF,$33,$C7,$F9,$C3,$CA
	.dc.b	$93,$FE,$5D,$6B,$FE,$81,$F3,$BF,$D1,$FE,$B7,$F9,$F1,$B1,$FE,$9E
	.dc.b	$CF,$FD,$05,$39,$C4,$FF,$A2,$DB,$3F,$E9,$1E,$5E,$2F,$FA,$BB,$C6
	.dc.b	$FF,$49,$D6,$CF,$F5,$01,$23,$FD,$57,$E9,$FF,$7D,$8C,$47,$FA,$DB
	.dc.b	$9B,$FD,$57,$1F,$EA,$98,$EA,$7F,$AB,$CA,$7F,$AC,$DF,$F4,$79,$1F
	.dc.b	$F3,$E3,$5F,$F6,$0F,$A8,$F5,$7E,$E4,$23,$19,$7E,$6E,$31,$3A,$9C
	.dc.b	$F2,$7F,$B0,$A7,$35,$B9,$CC,$7F,$B1,$9A,$F8,$2B,$91,$27,$83,$6B
	.dc.b	$A5,$70,$AF,$F6,$C3,$5E,$F1,$A9,$E1,$1D,$DA,$F7,$AE,$A4,$9E,$0D
	.dc.b	$8E,$0D,$78,$46,$88,$E9,$AE,$8C,$10,$C1,$0F,$4C,$F9,$1B,$3D,$DB
	.dc.b	$3D,$1A,$73,$29,$CC,$F8,$26,$4A,$66,$BE,$31,$2B,$C1,$1A,$39,$90
	.dc.b	$C1,$23,$85,$79,$63,$13,$CA,$73,$5E,$79,$CD,$6E,$73,$1E,$56,$BE
	.dc.b	$5A,$E4,$49,$E0,$DA,$E9,$5C,$2B,$CC,$D7,$BC,$6A,$78,$47,$76,$BD
	.dc.b	$EB,$A9,$27,$83,$63,$83,$5E,$11,$A2,$3A,$6B,$A3,$04,$30,$43,$D3
	.dc.b	$3E,$46,$CF,$76,$CF,$46,$9C,$CA,$73,$3E,$09,$92,$99,$AF,$8C,$4A
	.dc.b	$F0,$46,$8E,$64,$30,$48,$E1,$5E,$58,$C4,$F2,$9C,$D7,$9E,$73,$5B
	.dc.b	$9C,$C7,$95,$AF,$96,$BF,$87,$F9,$7F,$E5,$98,$1F,$E5,$03,$FC,$C0
	.dc.b	$60,$FF,$2C,$F9,$4F,$2F,$F8,$84,$7F,$17,$C5,$8E,$31,$FF,$33,$36
	.dc.b	$33,$3F,$C3,$1A,$FF,$15,$6B,$FC,$91,$AF,$F1,$06,$BD,$E3,$A1,$1E
	.dc.b	$8D,$7A,$35,$E8,$D7,$A3,$5E,$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23
	.dc.b	$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$7E,$8B,$E4
	.dc.b	$6B,$DD,$AF,$46,$BC,$DA,$F3,$9F,$A3,$18,$23,$82,$4F,$CB,$8C,$B6
	.dc.b	$33,$3E,$36,$BE,$36,$BE,$36,$BE,$36,$BE,$38,$E8,$47,$A3,$5E,$8D
	.dc.b	$7A,$35,$E8,$D7,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81
	.dc.b	$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$9F,$A2,$F9,$1A,$F7,$6B
	.dc.b	$D1,$AF,$36,$BC,$E7,$E8,$C6,$08,$E0,$93,$F2,$E3,$2D,$8C,$CF,$8D
	.dc.b	$AF,$8D,$AF,$8D,$AF,$8D,$AF,$8E,$3F,$88,$F8,$2E,$B7,$5F,$F1,$5A
	.dc.b	$FE,$30,$27,$51,$98,$FE,$39,$18,$C7,$FC,$54,$D8,$E3,$EB,$B8,$BF
	.dc.b	$8E,$D7,$F8,$BE,$31,$8F,$3A,$FE,$53,$18,$9D,$57,$09,$FF,$2D,$7D
	.dc.b	$25,$7F,$96,$72,$23,$11,$FC,$CA,$31,$87,$58,$46,$2F,$FC,$81,$F0
	.dc.b	$AF,$67,$8D,$23,$17,$C5,$B5,$C2,$B7,$CA,$EB,$A4,$BF,$04,$65,$F7
	.dc.b	$0F,$B1,$7C,$F7,$F4,$0F,$6B,$E7,$CE,$63,$19,$FF,$49,$8C,$09,$DD
	.dc.b	$72,$8C,$4F,$C6,$6B,$F1,$B5,$E6,$D8,$CB,$F1,$E3,$2F,$F5,$AE,$0F
	.dc.b	$BE,$FD,$CE,$2F,$C2,$B7,$8B,$F1,$31,$88,$F3,$7D,$43,$C8,$FE,$A3
	.dc.b	$8A,$30,$FF,$9B,$CC,$46,$A3,$33,$F0,$BD,$37,$CE,$7F,$64,$F0,$B0
	.dc.b	$D9,$F8,$93,$AC,$15,$A6,$CE,$BB,$FC,$74,$69,$27,$D7,$9C,$AF,$71
	.dc.b	$0F,$A5,$F6,$FF,$DD,$A7,$D4,$D8,$FB,$2D,$77,$1F,$DC,$35,$F9,$71
	.dc.b	$FA,$C6,$BF,$AA,$D7,$E5,$36,$FF,$D2,$9C,$0F,$B1,$7D,$17,$F7,$C8
	.dc.b	$C3,$63,$38,$F0,$6D,$FC,$BC,$3F,$C1,$BF,$9B,$5A,$34,$93,$F3,$3E
	.dc.b	$E3,$E3,$92,$7E,$C6,$C8,$F6,$5F,$A2,$BF,$10,$D7,$F8,$43,$74,$F7
	.dc.b	$C7,$F8,$7B,$5F,$E2,$18,$7C,$AA,$FC,$3C,$62,$7E,$5C,$76,$D7,$EA
	.dc.b	$47,$C9,$8E,$84,$8C,$C7,$C9,$3E,$C1,$FE,$34,$EC,$7E,$3C,$FB,$BF
	.dc.b	$F1,$5E,$3F,$E3,$59,$CC,$62,$B8,$56,$EB,$FC,$7A,$6D,$C4,$7F,$8C
	.dc.b	$32,$F6,$BF,$03,$0F,$1B,$F1,$1A,$E4,$F2,$99,$3F,$80,$F2,$3E,$5F
	.dc.b	$F1,$D8,$7E,$B1,$B3,$E2,$9D,$4F,$E3,$CF,$D7,$4E,$A7,$FC,$97,$F4
	.dc.b	$AF,$4B,$D9,$18,$6D,$27,$4F,$BA,$7D,$F1,$87,$FC,$9F,$0E,$07,$B9
	.dc.b	$FB,$58,$FD,$9B,$67,$E3,$CE,$A7,$F6,$A6,$8B,$1F,$5C,$FB,$99,$7C
	.dc.b	$6C,$FF,$0F,$3E,$5C,$7C,$F8,$F8,$E3,$F9,$23,$6F,$CF,$3E,$E7,$E7
	.dc.b	$1D,$0F,$9E,$60,$F0,$10,$FA,$46,$CF,$EC,$3E,$3B,$C7,$FE,$C9,$D0
	.dc.b	$F1,$BF,$CF,$CD,$AF,$F3,$EF,$87,$FE,$82,$8C,$42,$F9,$4E,$6B,$89
	.dc.b	$27,$89,$FC,$55,$C3,$F8,$B7,$F9,$81,$1F,$8E,$D7,$F8,$43,$5F,$C8
	.dc.b	$3F,$D3,$31,$FB,$AF,$DA,$6E,$5F,$2B,$2F,$F4,$8C,$E5,$B1,$DA,$A1
	.dc.b	$D8,$A7,$FA,$63,$5F,$BD,$F5,$DF,$86,$6C,$FD,$F8,$C6,$FF,$E9,$E7
	.dc.b	$B4,$7F,$9B,$0F,$BC,$5C,$7C,$B7,$F7,$23,$5C,$3F,$A0,$C7,$F2,$6D
	.dc.b	$8F,$D3,$1F,$97,$9F,$F5,$5C,$65,$F4,$10,$F8,$A7,$A8,$F9,$C5,$DF
	.dc.b	$FC,$3F,$9F,$CB,$E4,$FE,$0B,$7F,$E8,$26,$C7,$FA,$C9,$B1,$FC,$17
	.dc.b	$93,$E4,$DF,$73,$A9,$6F,$AB,$DF,$97,$FC,$63,$87,$D9,$8D,$9F,$C9
	.dc.b	$E3,$EA,$BF,$D4,$9A,$FF,$14,$E6,$FE,$57,$3F,$CA,$30,$FC,$8C,$FB
	.dc.b	$8F,$05,$76,$7E,$CC,$7E,$EF,$9F,$9C,$7D,$B6,$C6,$4E,$65,$AF,$E9
	.dc.b	$5F,$9A,$4A,$FE,$E7,$7F,$ED,$A8,$C4,$7F,$B9,$3B,$3F,$AB,$DF,$91
	.dc.b	$B7,$FE,$4E,$6C,$FF,$B9,$DA,$FA,$39,$FD,$6F,$C4,$3E,$2D,$32,$FE
	.dc.b	$E9,$7F,$DC,$A3,$D3,$18,$9F,$EE,$71,$CB,$3F,$A3,$1D,$6B,$F5,$13
	.dc.b	$FC,$77,$9B,$E0,$F2,$7F,$81,$E3,$E0,$B6,$7F,$C2,$D3,$AE,$EF,$CE
	.dc.b	$5F,$F8,$51,$AF,$F2,$26,$BF,$A7,$6D,$FF,$88,$8F,$9B,$C3,$FE,$26
	.dc.b	$BF,$E8,$7D,$FF,$4A,$31,$7F,$87,$D3,$D5,$BF,$DB,$8F,$C7,$78,$FF
	.dc.b	$E6,$0F,$3F,$FC,$F3,$1F,$E6,$0F,$09,$9A,$73,$5B,$AF,$F2,$44,$7F
	.dc.b	$93,$3C,$4D,$B6,$B8,$4F,$F9,$4F,$C4,$FF,$28,$4E,$4E,$FE,$2F,$F9
	.dc.b	$7F,$CC,$BE,$3F,$81,$B8,$1A,$7F,$9A,$DC,$61,$FE,$90,$F2,$7F,$CE
	.dc.b	$33,$92,$47,$F9,$75,$F4,$E9,$D6,$30,$40,$7F,$A3,$3A,$F4,$E0,$79
	.dc.b	$5F,$E9,$2A,$FF,$47,$E1,$E2,$9D,$4E,$5F,$EA,$B3,$FE,$61,$23,$83
	.dc.b	$FD,$4F,$E5,$FF,$45,$AF,$CB,$B7,$93,$E9,$B6,$77,$1F,$B0,$51,$97
	.dc.b	$CC,$D7,$D6,$D7,$71,$C6,$30,$47,$04,$9F,$1B,$4F,$19,$67,$C7,$1A
	.dc.b	$24,$69,$B1,$A8,$C1,$1C,$10,$F4,$BD,$5B,$1C,$9B,$1C,$5A,$71,$3D
	.dc.b	$EF,$16,$C7,$16,$C7,$16,$9C,$4F,$8B,$1B,$6C,$6D,$A6,$E7,$E4,$4E
	.dc.b	$DB,$3B,$6C,$EE,$3C,$8F,$23,$D2,$43,$C6,$53,$46,$08,$60,$91,$D9
	.dc.b	$E0,$D7,$6D,$77,$1E,$28,$C1,$1C,$12,$7D,$8D,$3D,$85,$9F,$64,$68
	.dc.b	$91,$A6,$C6,$A3,$04,$70,$43,$D2,$F5,$6C,$72,$6C,$71,$69,$C4,$F7
	.dc.b	$BC,$5B,$1C,$5B,$1C,$5A,$71,$3E,$2C,$6D,$B1,$B6,$9B,$9F,$91,$3B
	.dc.b	$6C,$ED,$B3,$B8,$F2,$3C,$8F,$49,$0F,$19,$4D,$18,$21,$82,$47,$67
	.dc.b	$83,$5D,$B5,$DC,$7F,$09,$F0,$FF,$C9,$B1,$FD,$01,$AF,$F9,$6B,$C4
	.dc.b	$78,$9F,$D1,$23,$FC,$A9,$18,$7F,$8A,$FE,$0D,$FD,$19,$B8,$EF,$14
	.dc.b	$E6,$7F,$8F,$B5,$FE,$2E,$D7,$F8,$BB,$5F,$E3,$AD,$8E,$73,$D6,$30
	.dc.b	$47,$04,$9F,$19,$0F,$0B,$96,$99,$3C,$6E,$5A,$64,$F2,$B9,$69,$93
	.dc.b	$CE,$E5,$A6,$67,$D2,$D9,$F4,$36,$78,$4E,$A3,$CA,$43,$C6,$53,$C2
	.dc.b	$53,$B9,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1,$D1,$0D,$15,$D3,$82,$38
	.dc.b	$24,$FB,$A3,$C8,$D7,$BB,$5E,$8D,$79,$B6,$39,$CF,$C6,$8C,$11,$C1
	.dc.b	$27,$E6,$10,$F0,$B9,$69,$93,$C6,$E5,$A6,$4F,$2B,$96,$99,$3C,$EE
	.dc.b	$5A,$66,$7D,$2D,$9F,$43,$67,$84,$EA,$3C,$A4,$3C,$65,$3C,$25,$3B
	.dc.b	$95,$E4,$F0,$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$4F
	.dc.b	$BA,$3C,$8D,$7B,$B5,$E8,$D7,$9B,$63,$9C,$FC,$68,$FC,$37,$63,$99
	.dc.b	$CF,$FC,$57,$0F,$F1,$C4,$E6,$31,$1F,$E3,$E8,$C6,$5F,$E6,$06,$C7
	.dc.b	$F8,$FC,$73,$BE,$63,$FC,$C7,$5F,$E6,$26,$CE,$E7,$59,$76,$8E,$33
	.dc.b	$A7,$FC,$95,$D8,$E2,$C4,$7F,$C0,$16,$A8,$D5,$73,$6E,$1F,$E7,$6F
	.dc.b	$8B,$7C,$67,$06,$D7,$85,$B3,$B9,$EB,$5D,$3E,$33,$FC,$FE,$DC,$3E
	.dc.b	$03,$5F,$3E,$3F,$E8,$29,$F8,$18,$BC,$2A,$4A,$F8,$59,$FF,$A2,$23
	.dc.b	$E6,$6F,$F5,$26,$EF,$FD,$2F,$7F,$A9,$18,$3E,$86,$0F,$D3,$DE,$57
	.dc.b	$F4,$58,$C3,$63,$31,$FE,$98,$32,$5F,$0D,$E8,$F1,$6C,$66,$30,$75
	.dc.b	$FA,$77,$E0,$54,$93,$D6,$32,$FA,$74,$F6,$65,$F8,$7C,$FD,$6D,$8C
	.dc.b	$BF,$92,$6B,$FE,$AC,$C9,$7E,$B2,$5F,$AE,$32,$F7,$D2,$57,$C7,$3B
	.dc.b	$3F,$EB,$8D,$7F,$D4,$DF,$0D,$FD,$AD,$B5,$9E,$C4,$E5,$B1,$F1,$FD
	.dc.b	$5F,$F5,$FE,$7F,$4B,$74,$63,$2F,$F5,$15,$27,$2F,$EC,$63,$31,$8A
	.dc.b	$E5,$A7,$D1,$25,$7D,$57,$FB,$F6,$BF,$22,$FE,$36,$9E,$E7,$DA,$7E
	.dc.b	$13,$5F,$EA,$97,$F8,$5E,$9F,$C1,$11,$FF,$0B,$7D,$B3,$ED,$FF,$56
	.dc.b	$D8,$FD,$A3,$7A,$7F,$DB,$ED,$C3,$B9,$0F,$19,$F3,$CF,$F6,$2F,$3B
	.dc.b	$E8,$9E,$47,$F8,$87,$0F,$F6,$34,$E6,$31,$5C,$2B,$75,$FE,$CE,$36
	.dc.b	$E2,$3F,$C4,$9F,$64,$FC,$27,$F5,$55,$FC,$4E,$30,$FE,$C4,$C4,$6A
	.dc.b	$7F,$DA,$D3,$AE,$6F,$F1,$77,$07,$CD,$9F,$E4,$1C,$5F,$92,$51,$96
	.dc.b	$C6,$74,$7D,$73,$2F,$E2,$75,$FE,$E3,$9C,$C6,$23,$FD,$D3,$18,$D7
	.dc.b	$FC,$7C,$D9,$FC,$2F,$E1,$3F,$12,$BA,$72,$A3,$13,$F8,$D6,$C6,$63
	.dc.b	$E4,$4F,$C9,$F5,$5E,$47,$97,$DC,$FF,$27,$52,$72,$A3,$31,$86,$CF
	.dc.b	$19,$FC,$05,$7C,$6E,$7F,$C9,$3F,$82,$3D,$31,$E2,$7D,$38,$F9,$89
	.dc.b	$19,$8F,$97,$3C,$9E,$87,$E2,$CA,$78,$0D,$99,$3F,$CD,$FC,$87,$FA
	.dc.b	$27,$91,$5C,$98,$35,$E4,$7F,$9F,$3C,$34,$EE,$30,$DF,$23,$FC,$F5
	.dc.b	$3A,$AE,$31,$89,$D5,$F8,$6B,$55,$E0,$6F,$1F,$F6,$DA,$FE,$7A,$49
	.dc.b	$FE,$2A,$D7,$F7,$AD,$7E,$73,$FC,$3F,$3F,$C7,$3F,$6B,$7F,$C0,$36
	.dc.b	$33,$A7,$D2,$8F,$CE,$F1,$FF,$80,$1E,$97,$F4,$1E,$D3,$F8,$3D,$BF
	.dc.b	$35,$EB,$FE,$02,$0D,$23,$D4,$95,$FB,$7C,$7C,$71,$88,$FC,$FF,$27
	.dc.b	$E9,$7C,$CF,$CF,$B6,$7F,$C1,$5C,$9F,$A6,$9C,$C1,$F4,$F1,$7C,$CB
	.dc.b	$F8,$67,$C8,$21,$FB,$CE,$8F,$D9,$79,$DF,$C1,$9B,$7F,$C0,$84,$65
	.dc.b	$A7,$F0,$EE,$BF,$93,$5F,$C4,$DA,$FF,$84,$FA,$FF,$8A,$EF,$FE,$B7
	.dc.b	$E9,$FE,$35,$7F,$8A,$3C,$75,$E6,$3E,$D9,$81,$A7,$F2,$FF,$07,$F9
	.dc.b	$6F,$6F,$F0,$96,$D1,$A7,$F3,$DE,$EF,$C0,$DF,$F4,$EE,$FF,$F5,$FE
	.dc.b	$59,$BF,$8C,$7F,$41,$3E,$44,$FD,$F3,$C5,$CF,$F9,$37,$CF,$1F,$D5
	.dc.b	$FC,$0E,$B1,$F6,$DA,$FF,$12,$6C,$6C,$EF,$C9,$FD,$7B,$E6,$7F,$B2
	.dc.b	$78,$5F,$C8,$DB,$26,$9F,$E4,$9F,$07,$FB,$8E,$57,$5F,$ED,$ED,$F0
	.dc.b	$FC,$33,$AC,$FD,$3D,$7F,$9B,$6F,$3F,$C0,$DF,$FC,$07,$87,$F6,$3D
	.dc.b	$BC,$F1,$8A,$F9,$11,$EC,$7E,$21,$F5,$88,$7D,$93,$F9,$E3,$FE,$1C
	.dc.b	$9F,$F1,$8B,$63,$FB,$FB,$5F,$EB,$0D,$7F,$9C,$F8,$BF,$23,$7F,$F1
	.dc.b	$6D,$7F,$82,$63,$07,$9B,$C6,$9B,$FA,$DB,$3F,$B8,$E1,$FB,$18,$79
	.dc.b	$8F,$49,$FC,$87,$CB,$FF,$36,$F4,$FF,$94,$89,$1F,$E5,$0C,$3F,$42
	.dc.b	$FF,$96,$B9,$FF,$AC,$7C,$07,$F9,$79,$B1,$99,$4E,$5B,$1F,$E8,$1F
	.dc.b	$33,$FC,$D7,$C9,$FE,$69,$9F,$CC,$4E,$7C,$CF,$CE,$36,$32,$FF,$9A
	.dc.b	$1A,$FF,$9F,$5A,$FF,$9F,$70,$E2,$EE,$BF,$CD,$A4,$7D,$8D,$7F,$C4
	.dc.b	$AF,$F9,$A8,$8F,$F9,$09,$FF,$4C,$2F,$37,$F5,$2F,$FA,$6D,$AF,$F4
	.dc.b	$96,$BF,$D3,$5A,$F1,$E9,$FC,$54,$60,$8E,$09,$7F,$B2,$8C,$B6,$33
	.dc.b	$B7,$92,$72,$D8,$CD,$79,$5A,$F9,$5A,$F9,$5A,$F9,$6B,$B1,$27,$A3
	.dc.b	$63,$A3,$5E,$8D,$7A,$35,$E9,$1C,$88,$F2,$6B,$C9,$AF,$26,$BC,$8E
	.dc.b	$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$4A,$F9,$EF,$56
	.dc.b	$BC,$9A,$F1,$6B,$C6,$BE,$7C,$60,$8E,$09,$5F,$26,$32,$D8,$CD,$7C
	.dc.b	$49,$CB,$63,$35,$E5,$6B,$E5,$6B,$E5,$6B,$E5,$AE,$C4,$9E,$8D,$8E
	.dc.b	$8D,$7A,$35,$E8,$D7,$A4,$72,$23,$C9,$AF,$26,$BC,$9A,$F2,$38,$10
	.dc.b	$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$2B,$E7,$BD,$5A,$F2
	.dc.b	$6B,$C5,$AF,$1A,$F9,$F1,$82,$38,$3F,$67,$E3,$BC,$77,$8F,$FC,$4A
	.dc.b	$FF,$CA,$A7,$F9,$5C,$A7,$F9,$30,$FE,$09,$E0,$FF,$A3,$5B,$97,$1C
	.dc.b	$7F,$D3,$F1,$82,$1F,$C6,$E7,$83,$67,$84,$62,$3F,$8F,$35,$FE,$3C
	.dc.b	$D7,$F9,$33,$67,$BC,$73,$3B,$9C,$A7,$A9,$1C,$9D,$1C,$3C,$48,$71
	.dc.b	$36,$78,$1D,$B9,$36,$73,$37,$1E,$47,$44,$34,$70,$3A,$1C,$27,$D8
	.dc.b	$49,$EF,$1C,$0E,$47,$03,$94,$F7,$74,$43,$47,$07,$4E,$0D,$38,$24
	.dc.b	$F8,$63,$C9,$3E,$27,$81,$27,$71,$96,$C7,$02,$1E,$69,$E8,$D9,$E1
	.dc.b	$18,$8F,$33,$5F,$2B,$5F,$03,$67,$BC,$73,$3B,$9C,$A7,$A9,$1C,$9D
	.dc.b	$1C,$3C,$48,$71,$36,$78,$1D,$B9,$36,$73,$37,$1E,$47,$44,$34,$70
	.dc.b	$3A,$1C,$27,$D8,$49,$EF,$1C,$0E,$47,$03,$94,$F7,$74,$43,$47,$07
	.dc.b	$4E,$0D,$38,$24,$F8,$63,$C9,$3E,$27,$81,$27,$71,$96,$C7,$02,$1E
	.dc.b	$69,$E8,$D9,$E1,$18,$8F,$33,$5F,$2B,$5F,$F7,$AD,$7F,$8A,$38,$9D
	.dc.b	$2E,$9F,$E3,$75,$FC,$66,$73,$18,$8F,$F7,$94,$62,$BF,$C6,$3E,$83
	.dc.b	$E3,$59,$6E,$FF,$DF,$53,$98,$C4,$23,$15,$CA,$B8,$BF,$E5,$52,$4F
	.dc.b	$F9,$65,$A8,$EB,$1E,$0F,$3F,$95,$9A,$49,$F6,$65,$EF,$ED,$F8,$4D
	.dc.b	$8C,$E6,$9C,$B6,$32,$60,$86,$08,$7C,$69,$F9,$8E,$8C,$47,$F9,$62
	.dc.b	$DD,$7F,$EC,$F9,$FE,$91,$18,$E2,$FE,$74,$D3,$84,$EE,$7E,$11,$F4
	.dc.b	$17,$17,$D2,$F2,$F9,$C6,$29,$18,$9D,$57,$29,$F1,$C6,$23,$CE,$7A
	.dc.b	$4F,$A1,$BF,$E1,$1C,$07,$0F,$C2,$37,$97,$C2,$70,$6D,$C3,$D8,$CC
	.dc.b	$72,$70,$43,$9E,$5F,$DB,$E3,$51,$F0,$78,$3F,$23,$FB,$A7,$C0,$6C
	.dc.b	$7B,$73,$46,$23,$F1,$59,$7D,$28,$C4,$7C,$BF,$C2,$FE,$8D,$1A,$4A
	.dc.b	$F1,$65,$EA,$AF,$29,$F1,$31,$F2,$BE,$FC,$B6,$47,$E4,$F8,$AF,$A8
	.dc.b	$FC,$D4,$7F,$82,$63,$18,$7E,$6E,$31,$3A,$7F,$03,$3A,$E1,$F0,$63
	.dc.b	$E3,$9F,$01,$8C,$3C,$0F,$CE,$8F,$46,$1F,$92,$F7,$3F,$73,$F1,$DF
	.dc.b	$E0,$E8,$C0,$71,$AA,$77,$3F,$35,$B7,$F5,$CF,$09,$DF,$0F,$B1,$ED
	.dc.b	$7E,$DE,$BF,$87,$36,$7F,$86,$B5,$FC,$7B,$63,$F4,$FF,$4B,$FE,$29
	.dc.b	$7E,$E3,$67,$ED,$E0,$8C,$47,$E9,$5B,$3F,$B0,$FA,$5F,$D3,$3F,$85
	.dc.b	$6C,$F2,$6B,$FA,$D6,$BE,$C3,$DD,$BF,$DE,$3F,$27,$3E,$63,$F1,$B1
	.dc.b	$E4,$AF,$0D,$F9,$67,$33,$E5,$6B,$D4,$FC,$59,$F2,$4F,$B4,$43,$FC
	.dc.b	$9D,$D9,$FE,$53,$E9,$FE,$49,$E9,$BA,$3F,$CA,$0D,$AF,$E5,$15,$B5
	.dc.b	$19,$51,$98,$C3,$67,$99,$27,$9F,$6F,$F4,$8A,$E5,$9F,$F4,$8A,$FE
	.dc.b	$52,$4A,$FE,$5A,$D7,$F9,$1B,$5F,$DE,$BF,$C6,$33,$FE,$96,$FE,$41
	.dc.b	$B9,$7E,$00,$FE,$44,$7C,$F8,$FF,$31,$57,$86,$BB,$90,$EE,$7D,$F9
	.dc.b	$7A,$3F,$9C,$D7,$FA,$CB,$63,$26,$DF,$13,$E0,$D7,$ED,$61,$82,$47
	.dc.b	$E7,$07,$87,$E4,$DF,$FB,$2B,$63,$26,$DF,$4F,$1F,$F7,$3A,$F8,$99
	.dc.b	$7F,$09,$7E,$35,$7E,$D3,$E9,$BF,$67,$C7,$FD,$C1,$B0,$69,$D3,$8F
	.dc.b	$EF,$5F,$CE,$3E,$21,$4F,$8A,$43,$04,$1C,$DF,$C5,$32,$FF,$4F,$46
	.dc.b	$39,$FD,$AF,$F1,$6C,$3E,$0B,$67,$D7,$C1,$F5,$63,$B1,$ED,$3C,$65
	.dc.b	$32,$ED,$FE,$69,$3E,$A3,$FB,$4F,$47,$E1,$A3,$F9,$F3,$5F,$E0,$6D
	.dc.b	$7F,$A3,$43,$F7,$3F,$E8,$CD,$8F,$E8,$DD,$5F,$E1,$A9,$C9,$E8,$DF
	.dc.b	$EA,$4F,$9B,$0F,$23,$E7,$3E,$33,$82,$47,$08,$F5,$B5,$DB,$5E,$44
	.dc.b	$8F,$81,$7F,$DD,$0F,$CB,$C7,$F5,$BE,$CF,$CB,$B6,$32,$FF,$22,$8C
	.dc.b	$BF,$C9,$38,$3E,$14,$FF,$82,$DF,$F0,$29,$8E,$2F,$E1,$11,$F8,$69
	.dc.b	$F1,$61,$E5,$C3,$D7,$BE,$5B,$87,$C5,$8C,$4F,$F9,$2F,$C6,$FF,$14
	.dc.b	$92,$FF,$C5,$AF,$F8,$DB,$69,$1F,$EE,$7F,$07,$FC,$8D,$18,$8F,$F0
	.dc.b	$EF,$43,$C4,$FF,$26,$F8,$9F,$90,$8C,$6A,$8C,$4E,$AB,$FD,$E5,$5A
	.dc.b	$9F,$F2,$29,$23,$C1,$E1,$ED,$FF,$15,$E3,$FE,$57,$9F,$F2,$09,$23
	.dc.b	$FC,$8A,$D7,$FD,$CD,$AF,$FB,$E8,$8F,$FB,$D5,$AE,$8C,$10,$C1,$0F
	.dc.b	$F7,$63,$C1,$A6,$CA,$6C,$FF,$76,$3B,$69,$B2,$9B,$F1,$3F,$C4,$39
	.dc.b	$7D,$D6,$CF,$16,$CF,$16,$BC,$78,$FF,$54,$F4,$3C,$64,$30,$48,$E2
	.dc.b	$E5,$AE,$63,$9C,$60,$8E,$09,$5E,$76,$D7,$9D,$B5,$E7,$6B,$E7,$9E
	.dc.b	$04,$9E,$0D,$8E,$0D,$78,$46,$88,$E9,$AE,$8C,$10,$C1,$0F,$81,$3E
	.dc.b	$46,$CF,$76,$CF,$46,$9C,$CA,$73,$3E,$44,$71,$6C,$71,$69,$C4,$A7
	.dc.b	$1A,$F9,$B3,$C5,$B3,$C5,$B3,$C5,$AF,$1A,$F5,$3D,$0F,$29,$0C,$12
	.dc.b	$38,$B9,$6B,$98,$E7,$18,$23,$82,$57,$9D,$B5,$E7,$6D,$79,$DA,$F9
	.dc.b	$E7,$81,$27,$83,$63,$83,$5E,$11,$A2,$3A,$6B,$A3,$04,$30,$43,$F4
	.dc.b	$9D,$DF,$E2,$F7,$FC,$6F,$DD,$FE,$E2,$7F,$C5,$ED,$8D,$47,$FA,$F3
	.dc.b	$0F,$F2,$33,$FD,$B9,$AF,$FB,$00,$38,$78,$77,$7F,$BC,$9E,$0D,$78
	.dc.b	$46,$DF,$F7,$53,$5E,$71,$B7,$9E,$1F,$DF,$9C,$B6,$78,$12,$78,$4F
	.dc.b	$46,$BC,$0F,$1C,$F1,$8C,$CF,$18,$C1,$27,$64,$3D,$33,$C0,$91,$B8
	.dc.b	$CB,$5F,$49,$C4,$A6,$8C,$1E,$77,$83,$4C,$9E,$93,$81,$0E,$64,$38
	.dc.b	$19,$29,$93,$91,$67,$E0,$3D,$9A,$F1,$36,$72,$21,$82,$3C,$1E,$93
	.dc.b	$F1,$1E,$8D,$7A,$46,$DE,$6D,$79,$C6,$DE,$73,$F1,$1C,$B6,$78,$12
	.dc.b	$78,$4F,$46,$BC,$0F,$1C,$F1,$8C,$CF,$18,$C1,$27,$64,$3D,$33,$C0
	.dc.b	$91,$B8,$CB,$5F,$49,$C4,$A6,$8C,$1E,$77,$83,$4C,$9E,$93,$81,$0E
	.dc.b	$64,$38,$19,$29,$93,$91,$67,$E0,$3D,$9A,$F1,$36,$72,$21,$82,$3C
	.dc.b	$1E,$93,$F1,$1E,$8D,$7A,$46,$DE,$6D,$79,$C6,$DE,$73,$F1,$1C,$B6
	.dc.b	$78,$12,$7E,$FF,$1B,$89,$C6,$EF,$FF,$43,$CE,$63,$18,$FF,$1B,$8C
	.dc.b	$4E,$AB,$8B,$FE,$8E,$F1,$DE,$CB,$3E,$38,$09,$3A,$8C,$C6,$27,$85
	.dc.b	$73,$6C,$78,$6F,$CF,$E1,$F8,$72,$34,$93,$FC,$C7,$1E,$D1,$CF,$1D
	.dc.b	$CE,$70,$F3,$4E,$4F,$F7,$C4,$79,$CF,$F7,$F1,$FE,$FF,$9F,$F2,$D7
	.dc.b	$D4,$7B,$7E,$67,$F9,$F4,$60,$38,$E3,$4E,$EB,$A5,$6E,$B8,$F5,$3C
	.dc.b	$AF,$E6,$B2,$8C,$3F,$EB,$A3,$11,$A8,$CD,$72,$FC,$3B,$E8,$7C,$D5
	.dc.b	$B3,$49,$3F,$2F,$6F,$E9,$1C,$3F,$7E,$3E,$09,$FE,$7A,$F0,$79,$BE
	.dc.b	$8D,$51,$88,$FB,$98,$7F,$9F,$63,$11,$C6,$3C,$2F,$73,$A9,$F2,$E5
	.dc.b	$E9,$7C,$58,$E7,$18,$9F,$97,$1E,$19,$D3,$FE,$8E,$6C,$7C,$7F,$77
	.dc.b	$F2,$18,$7C,$C6,$D7,$D0,$D1,$18,$8F,$4C,$62,$95,$F5,$3F,$06,$EB
	.dc.b	$26,$92,$7C,$B5,$F8,$7A,$F9,$4F,$BC,$86,$79,$3F,$5E,$FE,$17,$EB
	.dc.b	$BD,$39,$9A,$4B,$EB,$2F,$31,$F8,$63,$E5,$7F,$5D,$EB,$7F,$82,$67
	.dc.b	$31,$8B,$FC,$D4,$61,$4E,$EB,$94,$60,$F3,$DF,$ED,$F3,$F5,$BE,$9E
	.dc.b	$1F,$C3,$E9,$EA,$8F,$D3,$78,$8F,$59,$F3,$1F,$A5,$C7,$31,$87,$F4
	.dc.b	$A6,$23,$55,$FC,$16,$B9,$46,$23,$E0,$65,$E6,$C3,$D1,$AF,$F0,$D6
	.dc.b	$E1,$FC,$6C,$F7,$CE,$DF,$5D,$7E,$EB,$3F,$5B,$ED,$AF,$53,$F9,$AF
	.dc.b	$C9,$7F,$DA,$FD,$9F,$E3,$3B,$46,$1F,$DC,$B6,$78,$67,$FC,$7E,$B8
	.dc.b	$4B,$E8,$FF,$67,$7C,$1B,$7D,$DA,$FF,$27,$9E,$13,$FB,$83,$BF,$CA
	.dc.b	$7C,$37,$C6,$D8,$34,$F9,$13,$EF,$C7,$DA,$FB,$F7,$FB,$27,$F0,$BA
	.dc.b	$7C,$17,$32,$61,$F5,$AB,$E2,$47,$4D,$3E,$4C,$75,$BE,$6F,$EF,$23
	.dc.b	$26,$BC,$47,$8B,$FD,$2B,$CA,$78,$FF,$CB,$67,$3E,$2F,$F4,$59,$CE
	.dc.b	$1F,$D1,$E7,$2A,$32,$FF,$9D,$27,$85,$73,$BE,$CD,$8C,$CF,$39,$CB
	.dc.b	$63,$2F,$F0,$0D,$7F,$B1,$D7,$F4,$D2,$47,$F2,$76,$BF,$A3,$6B,$F3
	.dc.b	$5F,$E8,$7A,$FC,$CD,$5F,$29,$F8,$4D,$BF,$B2,$36,$32,$7E,$32,$32
	.dc.b	$F7,$B8,$77,$29,$F8,$7D,$F6,$FE,$65,$AE,$63,$AB,$96,$BF,$8F,$E0
	.dc.b	$FE,$ED,$F2,$DF,$98,$6C,$FE,$5C,$46,$5A,$7E,$8B,$8F,$FB,$F7,$2F
	.dc.b	$E8,$1B,$19,$3F,$86,$DF,$E9,$9A,$FC,$73,$D6,$7D,$1E,$3F,$BA,$E4
	.dc.b	$FF,$04,$74,$FF,$19,$EF,$FD,$F9,$2B,$F7,$F8,$7F,$83,$E3,$FD,$55
	.dc.b	$87,$F0,$A9,$F7,$B7,$5C,$F4,$FF,$13,$6F,$4F,$C8,$CB,$C5,$C3,$F7
	.dc.b	$1A,$7D,$0C,$FF,$95,$7C,$4F,$F1,$02,$65,$E2,$0F,$F0,$9E,$DF,$E7
	.dc.b	$91,$F8,$2A,$FE,$D5,$DB,$F8,$26,$C6,$5F,$EC,$31,$97,$F0,$04,$3E
	.dc.b	$21,$23,$FC,$0D,$DD,$B8,$FF,$18,$77,$FF,$57,$9E,$91,$89,$D4,$70
	.dc.b	$32,$7F,$69,$1F,$85,$FF,$69,$6C,$FF,$69,$F0,$7F,$C7,$93,$93,$FC
	.dc.b	$3D,$C3,$D3,$2F,$EF,$1B,$E3,$0E,$B1,$8C,$3E,$FF,$77,$F9,$37,$7F
	.dc.b	$9C,$D8,$FF,$07,$78,$5F,$94,$6D,$FE,$49,$BC,$DF,$E2,$D9,$D4,$7C
	.dc.b	$53,$EB,$96,$BA,$3F,$DF,$BC,$4D,$BF,$19,$B7,$FE,$2F,$6E,$E6,$F8
	.dc.b	$BF,$E5,$BC,$BF,$C6,$ED,$BF,$F0,$63,$5F,$F0,$5B,$5F,$F3,$16,$7F
	.dc.b	$E4,$2C,$3E,$1F,$8F,$F1,$C8,$FD,$69,$F4,$9F,$78,$A4,$F2,$7F,$CD
	.dc.b	$1E,$7F,$F9,$6C,$91,$FE,$5B,$94,$7F,$99,$49,$1F,$E3,$F6,$BF,$E5
	.dc.b	$D9,$FF,$39,$5F,$F9,$94,$91,$CD,$FF,$37,$11,$FF,$24,$35,$FF,$33
	.dc.b	$79,$8F,$33,$E0,$42,$31,$87,$EF,$23,$13,$A9,$CE,$7F,$E6,$F9,$FF
	.dc.b	$7F,$35,$DD,$71,$9D,$C6,$DA,$F8,$AB,$A1,$27,$83,$6B,$A5,$70,$AE
	.dc.b	$E7,$39,$D3,$6B,$89,$2B,$8C,$EA,$7C,$13,$97,$84,$EA,$31,$1A,$8C
	.dc.b	$11,$D9,$0D,$14,$F2,$98,$21,$82,$1F,$19,$E8,$E8,$C1,$CC,$A6,$6B
	.dc.b	$E1,$38,$39,$B8,$25,$7B,$E3,$2D,$78,$12,$37,$5F,$02,$36,$D7,$53
	.dc.b	$C6,$77,$5E,$89,$DB,$5D,$D7,$19,$DC,$7A,$DA,$FA,$EB,$A1,$27,$83
	.dc.b	$6B,$A5,$70,$AE,$E7,$39,$D3,$6B,$89,$2B,$8C,$EA,$7C,$13,$97,$84
	.dc.b	$EA,$31,$1A,$8C,$11,$D9,$0D,$14,$F2,$98,$21,$82,$1F,$19,$E8,$E8
	.dc.b	$C1,$CC,$A6,$6B,$E1,$38,$39,$B8,$25,$7B,$E3,$2D,$78,$12,$37,$5F
	.dc.b	$02,$36,$D7,$53,$C6,$77,$5E,$89,$DB,$5D,$D7,$19,$DC,$7A,$DA,$FA
	.dc.b	$EB,$A1,$27,$F2,$1E,$37,$F9,$2A,$7F,$87,$35,$FF,$23,$1F,$C3,$DC
	.dc.b	$23,$F8,$D7,$B2,$9F,$E2,$91,$86,$DA,$75,$3F,$C3,$9A,$FF,$17,$6B
	.dc.b	$FC,$A1,$AF,$78,$E5,$3D,$27,$33,$DC,$E9,$19,$23,$82,$1D,$DC,$B4
	.dc.b	$C9,$E1,$72,$D3,$27,$8E,$3C,$33,$E4,$8C,$9E,$33,$A1,$0E,$45,$39
	.dc.b	$4F,$66,$9C,$A3,$A3,$B3,$A1,$C8,$87,$03,$4F,$07,$27,$03,$46,$1D
	.dc.b	$38,$21,$B2,$3A,$6C,$FC,$02,$4F,$A6,$32,$D9,$E0,$49,$DC,$FB,$5A
	.dc.b	$FA,$1A,$F9,$5A,$F9,$23,$94,$F5,$9C,$CF,$80,$E9,$19,$23,$82,$1D
	.dc.b	$DC,$B4,$C9,$E1,$72,$D3,$27,$8E,$3C,$33,$E4,$8C,$9E,$33,$A1,$0E
	.dc.b	$45,$39,$4F,$66,$9C,$A3,$A3,$B3,$A1,$C8,$87,$03,$4F,$07,$27,$03
	.dc.b	$46,$1D,$38,$21,$B2,$3A,$6C,$FC,$02,$4F,$A6,$32,$D9,$E0,$49,$DC
	.dc.b	$FB,$5A,$FA,$1A,$F9,$5A,$F9,$23,$94,$F5,$9C,$CF,$80,$E9,$19,$23
	.dc.b	$82,$1F,$E2,$BE,$07,$88,$F0,$1E,$2F,$F8,$B6,$73,$18,$AE,$15,$BA
	.dc.b	$D3,$63,$FC,$6D,$E0,$BE,$5F,$F9,$25,$7F,$8D,$E3,$18,$F3,$AF,$E5
	.dc.b	$42,$33,$3B,$AE,$33,$E0,$C1,$E7,$78,$1B,$19,$8C,$19,$61,$C4,$A7
	.dc.b	$F3,$9C,$1E,$BB,$EC,$3F,$9D,$7A,$1F,$E4,$F9,$CC,$62,$BC,$F1,$81
	.dc.b	$3B,$AE,$51,$89,$DF,$94,$FB,$FF,$E7,$72,$8C,$3F,$E5,$B3,$11,$A8
	.dc.b	$CC,$F8,$D7,$E0,$3F,$D3,$FD,$2C,$36,$7B,$B5,$F7,$4F,$0B,$57,$C2
	.dc.b	$9F,$EB,$5C,$3F,$79,$C1,$FE,$EF,$3F,$CE,$1E,$53,$C4,$7C,$F7,$59
	.dc.b	$4E,$63,$0D,$C3,$BC,$FD,$E8,$C6,$1D,$B0,$F0,$4F,$D5,$21,$F5,$74
	.dc.b	$FC,$3E,$BF,$37,$87,$E9,$4E,$5A,$FF,$71,$9D,$BF,$43,$F0,$8F,$EE
	.dc.b	$9F,$B3,$FC,$96,$C6,$73,$4E,$5B,$1F,$63,$D3,$EB,$1F,$5D,$BB,$9A
	.dc.b	$78,$A7,$E4,$1E,$A3,$E6,$F3,$FE,$B5,$FA,$15,$F9,$0F,$8C,$EB,$91
	.dc.b	$A4,$8F,$C5,$4F,$E5,$FE,$63,$F3,$1F,$80,$7F,$82,$25,$18,$7F,$2A
	.dc.b	$D9,$E0,$7F,$85,$67,$FC,$1B,$BF,$D0,$8F,$D8,$B5,$F3,$F4,$7F,$86
	.dc.b	$A3,$1C,$4F,$C0,$DE,$47,$E9,$B2,$FB,$91,$87,$F4,$66,$23,$58,$7F
	.dc.b	$89,$EB,$94,$62,$3F,$C5,$19,$7E,$CF,$EA,$FD,$8D,$BF,$5C,$FF,$8B
	.dc.b	$67,$2D,$8F,$9C,$FD,$03,$D7,$B7,$CF,$7E,$96,$1E,$9B,$FF,$1B,$7B
	.dc.b	$5F,$E3,$08,$C5,$A3,$19,$7C,$C8,$C1,$EE,$AE,$2F,$EF,$3E,$5F,$F6
	.dc.b	$4D,$3F,$5A,$5E,$7F,$AD,$C6,$9D,$E5,$FB,$4F,$67,$E1,$5F,$F0,$36
	.dc.b	$D1,$A7,$C7,$9F,$71,$F5,$8F,$D5,$69,$EF,$7F,$7F,$E4,$F2,$25,$FD
	.dc.b	$48,$CB,$5E,$47,$CA,$E0,$FC,$13,$D8,$8E,$09,$5D,$E3,$FC,$C5,$E2
	.dc.b	$79,$FC,$1F,$F3,$27,$B4,$F1,$FF,$CD,$2D,$AD,$E7,$FD,$2E,$31,$0B
	.dc.b	$E4,$A3,$31,$86,$CF,$1A,$EA,$4A,$FC,$CB,$FD,$14,$8F,$F4,$96,$BF
	.dc.b	$86,$E5,$F0,$BC,$0D,$1F,$CE,$34,$FF,$3E,$79,$FF,$69,$BA,$E4,$46
	.dc.b	$5A,$72,$9E,$D7,$E0,$7B,$1F,$81,$8D,$B6,$7F,$0F,$87,$68,$C4,$FE
	.dc.b	$37,$CF,$FC,$5F,$07,$FA,$3D,$B0,$69,$F9,$AE,$2F,$F4,$A7,$A9,$64
	.dc.b	$D2,$3E,$7B,$F2,$90,$70,$FF,$80,$B3,$FE,$C9,$7F,$DA,$7E,$97,$F6
	.dc.b	$76,$69,$1F,$E0,$47,$BA,$FF,$A3,$F2,$FF,$07,$D7,$F5,$FD,$3E,$F9
	.dc.b	$CB,$5F,$0F,$27,$F8,$57,$2F,$95,$18,$E1,$FB,$2D,$8F,$2B,$FE,$B3
	.dc.b	$F3,$7C,$A4,$CB,$F8,$58,$7F,$95,$76,$FF,$31,$8F,$B7,$3F,$DF,$A3
	.dc.b	$19,$7F,$7D,$ED,$FC,$1C,$E5,$AF,$DE,$3E,$51,$E3,$3E,$65,$7F,$49
	.dc.b	$3A,$BF,$DD,$3B,$B8,$BF,$EC,$9E,$9F,$E9,$93,$A8,$C4,$EA,$3F,$C1
	.dc.b	$1E,$07,$FB,$3F,$F7,$EF,$EB,$04,$AF,$EB,$8F,$F6,$79,$E4,$7F,$87
	.dc.b	$BC,$2E,$59,$FF,$8C,$B3,$D1,$F3,$39,$3E,$F6,$1E,$A6,$BF,$CE,$7C
	.dc.b	$5F,$F0,$0E,$1F,$8F,$8F,$EF,$91,$8E,$7F,$F0,$3B,$63,$3E,$2F,$E5
	.dc.b	$DE,$9C,$7F,$6D,$BC,$7F,$B2,$C3,$FC,$49,$E3,$6E,$7F,$C3,$2D,$7E
	.dc.b	$33,$63,$67,$8E,$3F,$C3,$FE,$3F,$FB,$CF,$97,$FC,$50,$D8,$FF,$19
	.dc.b	$B5,$DC,$F8,$76,$FF,$1C,$ED,$FE,$3B,$F2,$3E,$39,$27,$EA,$57,$DD
	.dc.b	$3F,$00,$43,$F5,$1B,$FF,$9A,$3C,$FF,$F2,$B1,$23,$FC,$B4,$FF,$97
	.dc.b	$65,$E7,$D2,$4F,$F9,$AB,$89,$3A,$8F,$F3,$14,$FF,$98,$48,$E0,$FF
	.dc.b	$7E,$69,$FE,$6C,$7F,$DD,$AD,$5E,$73,$CE,$F8,$51,$A3,$89,$0F,$F3
	.dc.b	$AC,$6E,$FF,$5E,$DA,$E8,$D9,$E0,$73,$38,$D7,$67,$C2,$D8,$E1,$1A
	.dc.b	$AE,$B3,$92,$38,$25,$78,$23,$2D,$8C,$CF,$88,$EE,$E1,$ED,$1B,$23
	.dc.b	$BA,$F1,$BE,$92,$1C,$A3,$6F,$23,$85,$78,$9B,$5D,$DE,$35,$D1,$E4
	.dc.b	$6A,$B9,$BA,$AE,$4F,$52,$19,$AE,$8F,$23,$25,$35,$3D,$1C,$12,$BC
	.dc.b	$44,$AE,$F1,$C0,$F2,$10,$EB,$1B,$AF,$0B,$6B,$B3,$67,$81,$CC,$E3
	.dc.b	$5D,$9F,$13,$63,$84,$6A,$BA,$CE,$48,$E0,$95,$E0,$8C,$B6,$33,$3E
	.dc.b	$23,$BB,$87,$B4,$6C,$8E,$EB,$C6,$FA,$48,$72,$8D,$BC,$8E,$15,$E2
	.dc.b	$6D,$77,$78,$D7,$47,$91,$AA,$E6,$EA,$B9,$3D,$48,$66,$BA,$3C,$8C
	.dc.b	$94,$D4,$F4,$70,$4A,$F1,$12,$BB,$C7,$03,$C8,$43,$AC,$6E,$BC,$2D
	.dc.b	$AE,$CD,$9E,$07,$33,$8D,$7F,$01,$F1,$3F,$C7,$B1,$FC,$55,$A8,$FF
	.dc.b	$22,$F8,$BF,$C5,$DF,$F2,$27,$86,$8C,$5B,$D5,$FE,$4B,$5F,$C4,$9A
	.dc.b	$FF,$21,$8D,$65,$C2,$73,$3F,$C7,$1A,$F8,$67,$6F,$39,$E0,$47,$04
	.dc.b	$3B,$4F,$27,$F9,$6B,$67,$9C,$6A,$79,$34,$E4,$78,$DC,$B4,$C9,$E5
	.dc.b	$9E,$84,$8D,$C6,$5A,$F8,$9B,$3E,$28,$E8,$78,$48,$60,$87,$33,$C2
	.dc.b	$73,$2B,$CD,$D1,$C8,$A7,$23,$81,$92,$99,$79,$38,$38,$90,$C1,$1D
	.dc.b	$92,$7E,$14,$79,$1A,$F7,$8D,$4F,$C2,$9C,$CF,$85,$AF,$79,$DB,$CE
	.dc.b	$78,$11,$C1,$0E,$D3,$C9,$F3,$36,$79,$C6,$A7,$93,$4E,$47,$8D,$CB
	.dc.b	$4C,$9E,$59,$E8,$48,$DC,$65,$AF,$89,$B3,$E2,$8E,$87,$84,$86,$08
	.dc.b	$73,$3C,$27,$32,$BC,$DD,$1C,$8A,$72,$38,$19,$29,$97,$93,$83,$89
	.dc.b	$0C,$11,$D9,$27,$E1,$47,$91,$AF,$78,$D4,$FC,$29,$CC,$F8,$5A,$F7
	.dc.b	$9D,$BC,$E7,$81,$1C,$10,$ED,$3C,$9F,$33,$67,$9C,$7C,$1D,$9C,$AE
	.dc.b	$5F,$E3,$93,$FE,$37,$9C,$C6,$23,$FD,$AD,$8F,$F8,$E6,$72,$FF,$8E
	.dc.b	$FC,$F7,$D8,$7F,$91,$A3,$0D,$AF,$F2,$14,$EB,$3E,$91,$FC,$BE,$31
	.dc.b	$5D,$71,$FF,$7E,$4F,$FC,$00,$6E,$1E,$48,$FF,$26,$5B,$B9,$E2,$7F
	.dc.b	$95,$BC,$9C,$C6,$25,$18,$9D,$57,$F3,$F3,$FD,$40,$E1,$FE,$85,$3D
	.dc.b	$7E,$4B,$F9,$AB,$4F,$F2,$D1,$63,$BC,$6A,$DE,$27,$F9,$A3,$D0,$F1
	.dc.b	$B6,$BC,$17,$E9,$8C,$47,$82,$B8,$CA,$BE,$26,$FF,$D4,$F3,$F8,$57
	.dc.b	$F5,$F4,$F8,$A7,$DB,$7E,$0F,$AD,$D1,$FF,$5E,$37,$0F,$90,$DC,$4E
	.dc.b	$E7,$E8,$61,$DE,$FA,$BF,$02,$9F,$2B,$F1,$EB,$FD,$90,$DC,$3E,$A3
	.dc.b	$67,$71,$F4,$A7,$E0,$C3,$E3,$FF,$D1,$BF,$B7,$FD,$86,$CE,$9A,$FC
	.dc.b	$56,$BF,$ED,$7F,$6F,$B1,$2F,$B4,$7C,$B8,$F3,$3E,$C3,$E5,$90,$F1
	.dc.b	$FC,$CF,$9C,$93,$CA,$3F,$DC,$91,$EF,$F4,$DF,$A0,$FE,$1B,$F1,$CF
	.dc.b	$EF,$F1,$8A,$FF,$53,$C6,$27,$57,$F8,$FA,$D2,$9F,$F0,$1F,$07,$DC
	.dc.b	$BF,$CA,$35,$FF,$0A,$35,$39,$9F,$D1,$7C,$CF,$CF,$23,$E9,$32,$7A
	.dc.b	$EF,$C3,$EB,$79,$6F,$FD,$7A,$D9,$34,$F2,$F0,$FE,$0A,$7E,$41,$E6
	.dc.b	$7F,$20,$FB,$7E,$F1,$F4,$FF,$EC,$89,$F8,$B1,$8B,$FE,$22,$6E,$74
	.dc.b	$D3,$35,$FA,$38,$FD,$54,$7E,$21,$B1,$98,$FD,$31,$92,$BF,$BA,$F4
	.dc.b	$FF,$77,$D5,$FE,$34,$94,$61,$FD,$9B,$67,$81,$FC,$2E,$3F,$5D,$CB
	.dc.b	$FA,$32,$47,$C2,$3F,$DC,$BC,$9E,$DB,$EC,$DA,$F2,$46,$30,$E9,$5F
	.dc.b	$26,$F9,$BD,$8E,$87,$E5,$CC,$10,$FE,$31,$E0,$BC,$1F,$E7,$3E,$C3
	.dc.b	$C4,$CC,$61,$BD,$FF,$E5,$D9,$D5,$FF,$97,$83,$87,$FC,$BF,$1B,$9E
	.dc.b	$31,$A6,$CE,$EF,$C5,$5B,$AF,$CB,$EF,$FD,$36,$FF,$8C,$92,$3E,$5B
	.dc.b	$5F,$84,$D7,$8E,$3E,$37,$E6,$E4,$D1,$E7,$FC,$B6,$9F,$81,$70,$7E
	.dc.b	$62,$B9,$57,$EA,$30,$F2,$3E,$22,$1F,$80,$AE,$6D,$AF,$C4,$6F,$FD
	.dc.b	$11,$B1,$E3,$7C,$5E,$54,$AF,$C9,$36,$7F,$D1,$9C,$6E,$2F,$D1,$76
	.dc.b	$FE,$93,$4F,$63,$70,$FE,$66,$7E,$8D,$F5,$6B,$F1,$70,$F5,$6B,$EC
	.dc.b	$E6,$FE,$EF,$3F,$D6,$E5,$97,$F0,$EF,$5B,$F7,$84,$C7,$F8,$97,$4F
	.dc.b	$F5,$E8,$FC,$BF,$47,$AE,$FF,$9A,$B5,$FE,$A8,$D7,$F3,$ED,$7C,$78
	.dc.b	$79,$71,$F1,$C7,$CC,$3C,$77,$FD,$B3,$F1,$8F,$E1,$C4,$C7,$C7,$3F
	.dc.b	$76,$3F,$53,$D7,$DE,$7F,$BF,$37,$A4,$D3,$C5,$87,$82,$FF,$16,$78
	.dc.b	$4F,$29,$FD,$43,$B3,$91,$2F,$F1,$71,$96,$BF,$D3,$D7,$E2,$1F,$D5
	.dc.b	$5B,$1F,$D5,$5A,$EE,$3F,$C2,$3E,$07,$9C,$93,$D2,$7F,$B3,$B5,$F2
	.dc.b	$9E,$82,$1F,$A2,$D3,$FB,$DC,$39,$29,$27,$FB,$5F,$17,$E9,$63,$E9
	.dc.b	$73,$FE,$7A,$79,$9D,$8F,$E2,$FE,$16,$DF,$EE,$6D,$7F,$C2,$4D,$77
	.dc.b	$7E,$17,$FB,$CF,$87,$FB,$3C,$7F,$C4,$33,$FB,$73,$FC,$4C,$E0,$FF
	.dc.b	$18,$3F,$04,$FF,$1C,$78,$9F,$6C,$8F,$C3,$DB,$ED,$5F,$C1,$3E,$D3
	.dc.b	$85,$E2,$FF,$92,$FC,$AF,$F2,$7D,$1A,$49,$FE,$72,$FF,$8A,$88,$E0
	.dc.b	$FF,$72,$61,$FE,$EF,$23,$82,$1F,$EE,$57,$2D,$32,$7F,$B8,$5C,$B4
	.dc.b	$C9,$FE,$E3,$72,$D3,$3E,$43,$C9,$F9,$91,$86,$BD,$C8,$75,$29,$CC
	.dc.b	$B1,$CE,$FF,$90,$B6,$BC,$2D,$9E,$D3,$AA,$EF,$1D,$48,$75,$2C,$F1
	.dc.b	$7B,$C7,$07,$44,$8D,$3C,$23,$51,$88,$D4,$60,$93,$E3,$25,$7A,$23
	.dc.b	$2D,$8C,$CF,$AA,$30,$47,$04,$3C,$F1,$C1,$B1,$B6,$9B,$3D,$11,$B6
	.dc.b	$C6,$DA,$6C,$F5,$C6,$DB,$1B,$69,$BA,$F8,$44,$AF,$6C,$65,$AF,$90
	.dc.b	$87,$80,$A7,$42,$C7,$3A,$F1,$36,$BC,$4D,$9E,$D3,$AA,$EF,$1D,$48
	.dc.b	$75,$2C,$F1,$7B,$C7,$07,$44,$8D,$3C,$23,$51,$88,$D4,$60,$93,$E3
	.dc.b	$25,$7A,$23,$2D,$8C,$CF,$AA,$30,$47,$04,$3C,$F1,$C1,$B1,$B6,$9B
	.dc.b	$3D,$11,$B6,$C6,$DA,$6C,$F5,$C7,$E0,$7C,$0F,$F1,$2C,$7F,$75,$3F
	.dc.b	$C5,$A5,$72,$FF,$79,$6B,$FE,$D5,$F0,$7F,$BD,$CF,$18,$FF,$1B,$F5
	.dc.b	$FF,$88,$DE,$93,$A6,$C6,$5F,$F0,$24,$65,$B1,$97,$B3,$5F,$F2,$33
	.dc.b	$5E,$F3,$E1,$9E,$51,$C2,$31,$3B,$9C,$CF,$29,$C9,$1F,$2B,$5F,$1B
	.dc.b	$5E,$86,$88,$68,$A6,$8C,$10,$C1,$0F,$11,$C0,$A6,$CA,$6C,$F7,$9B
	.dc.b	$29,$B2,$9B,$8F,$7B,$83,$99,$0C,$11,$D9,$27,$E4,$12,$7E,$14,$F1
	.dc.b	$9D,$46,$23,$94,$ED,$F3,$4E,$9B,$19,$7D,$11,$96,$C6,$5F,$53,$5F
	.dc.b	$03,$5E,$F3,$E1,$9E,$51,$C2,$31,$3B,$9C,$CF,$29,$C9,$1F,$2B,$5F
	.dc.b	$1B,$5E,$86,$88,$68,$A6,$8C,$10,$C1,$0F,$11,$C0,$A6,$CA,$6C,$F7
	.dc.b	$9B,$29,$B2,$9B,$8F,$7B,$83,$99,$0C,$11,$D9,$27,$E4,$12,$7E,$14
	.dc.b	$F1,$9D,$46,$23,$94,$ED,$F3,$4E,$9B,$19,$7D,$11,$96,$C6,$5F,$53
	.dc.b	$5F,$03,$5E,$F3,$E1,$9E,$51,$C2,$31,$3B,$9C,$CF,$29,$C9,$F9,$9D
	.dc.b	$1C,$0E,$C7,$83,$FC,$96,$73,$18,$BF,$F7,$8C,$E6,$B8,$3F,$EE,$6A
	.dc.b	$DC,$D9,$E8,$FF,$A8,$F1,$7A,$EF,$33,$FD,$95,$E2,$E6,$31,$9A,$31
	.dc.b	$3A,$B1,$FE,$BA,$AD,$CF,$FA,$BE,$9E,$27,$AD,$FE,$7C,$63,$4F,$E5
	.dc.b	$8D,$33,$1B,$BE,$4B,$F0,$6F,$F6,$F7,$97,$86,$CF,$88,$FF,$71,$4E
	.dc.b	$E9,$5B,$6C,$FF,$B0,$F5,$F8,$AD,$D7,$E2,$B7,$1F,$E9,$93,$A8,$F7
	.dc.b	$D3,$9B,$FC,$DE,$F4,$8C,$C6,$30,$F8,$4D,$8F,$34,$F1,$E6,$F4,$C7
	.dc.b	$A6,$FB,$3F,$EF,$E3,$EC,$61,$FD,$97,$77,$D6,$F8,$70,$FA,$95,$F0
	.dc.b	$67,$38,$FD,$58,$FF,$76,$D6,$F6,$FC,$03,$63,$C7,$AB,$EF,$98,$BF
	.dc.b	$10,$FA,$31,$86,$C6,$6B,$8C,$E5,$E7,$F5,$5E,$17,$E9,$B6,$7F,$06
	.dc.b	$DB,$3B,$9F,$F7,$C9,$F0,$0F,$CE,$69,$FE,$05,$F8,$97,$D8,$3E,$37
	.dc.b	$B4,$61,$B1,$9C,$7E,$2D,$FE,$5A,$B5,$18,$86,$EF,$B6,$FB,$BE,$8F
	.dc.b	$E9,$5B,$3F,$35,$AF,$E8,$27,$83,$F2,$27,$D0,$FE,$24,$F4,$9C,$CF
	.dc.b	$89,$1C,$2B,$C4,$DE,$0D,$47,$97,$07,$C2,$3E,$DF,$EA,$FC,$37,$F8
	.dc.b	$66,$73,$18,$AF,$D1,$46,$14,$EE,$B9,$46,$0F,$36,$30,$FC,$97,$47
	.dc.b	$E8,$9B,$BF,$E4,$7D,$17,$9F,$EF,$BF,$C8,$46,$16,$DF,$E2,$DD,$93
	.dc.b	$98,$C3,$67,$F2,$3E,$D7,$EF,$1A,$7E,$E0,$BD,$7F,$E3,$77,$F8,$5C
	.dc.b	$6F,$2F,$D9,$F2,$BF,$00,$FE,$33,$5F,$9C,$6C,$E9,$AF,$F0,$66,$BF
	.dc.b	$90,$3E,$11,$FB,$12,$C7,$80,$FE,$44,$79,$B8,$BC,$94,$BE,$B3,$B9
	.dc.b	$FF,$25,$75,$FE,$F2,$FB,$1F,$30,$FC,$39,$4F,$09,$A3,$F9,$6F,$85
	.dc.b	$FE,$66,$F0,$9E,$1F,$F1,$F8,$C3,$7D,$5F,$E8,$93,$AA,$E3,$18,$9D
	.dc.b	$78,$9F,$CE,$2B,$55,$FE,$70,$6D,$F4,$8E,$33,$BA,$FC,$19,$FB,$42
	.dc.b	$47,$F2,$76,$BA,$30,$43,$EB,$E7,$FC,$3B,$D5,$F9,$2D,$D3,$F0,$31
	.dc.b	$84,$FC,$AE,$5F,$EC,$97,$DA,$FF,$A8,$E9,$FD,$92,$31,$7F,$D4,$37
	.dc.b	$E9,$5F,$81,$8E,$67,$E1,$4A,$E7,$0F,$07,$97,$27,$F2,$6D,$9F,$E0
	.dc.b	$1C,$5F,$9F,$F4,$BF,$37,$CD,$FC,$A1,$B6,$69,$E9,$E7,$F9,$F7,$E8
	.dc.b	$D7,$D6,$FC,$53,$11,$AE,$AF,$DD,$78,$DF,$B9,$EC,$FE,$6A,$DC,$7F
	.dc.b	$BC,$9F,$C0,$3A,$FF,$7B,$1F,$B1,$ED,$FE,$AF,$BF,$F0,$C2,$65,$FC
	.dc.b	$6C,$FA,$47,$84,$71,$7C,$B7,$0D,$7F,$90,$4F,$F6,$2D,$7F,$8B,$92
	.dc.b	$3A,$63,$F0,$5F,$E2,$7D,$DF,$CC,$F0,$FE,$ED,$0D,$3F,$99,$C6,$1B
	.dc.b	$1C,$3B,$FC,$AF,$E2,$C8,$7C,$6E,$4F,$BF,$D7,$FD,$2B,$1F,$C3,$CF
	.dc.b	$F5,$26,$BF,$C1,$BC,$0E,$6F,$F5,$48,$CB,$FE,$08,$F0,$7F,$B0,$FD
	.dc.b	$47,$F5,$E2,$47,$F5,$E9,$47,$F6,$AF,$17,$9F,$85,$FD,$C2,$79,$B7
	.dc.b	$5D,$47,$6B,$F9,$BC,$DF,$80,$7D,$66,$39,$BF,$0A,$DD,$7F,$C1,$3E
	.dc.b	$37,$F7,$4E,$AB,$81,$A4,$AF,$AF,$87,$57,$E3,$1F,$8D,$9F,$F1,$9F
	.dc.b	$8F,$CE,$3F,$C0,$2D,$7F,$95,$B5,$FF,$15,$B6,$3F,$C6,$D8,$7F,$8D
	.dc.b	$31,$FF,$03,$B6,$3F,$3A,$6A,$BF,$C2,$9C,$FF,$E4,$AC,$FF,$C9,$33
	.dc.b	$F1,$A3,$13,$AF,$23,$EA,$BF,$04,$F5,$E1,$F1,$3A,$3F,$25,$1F,$A6
	.dc.b	$F2,$FF,$C0,$A4,$7F,$C7,$0F,$FB,$AB,$29,$1F,$EE,$A6,$BF,$EF,$56
	.dc.b	$BF,$EF,$36,$BF,$EF,$C6,$F8,$5F,$E0,$6C,$BE,$AB,$63,$3C,$1F,$0A
	.dc.b	$BB,$10,$EC,$53,$B1,$4E,$C5,$9E,$8D,$8E,$8F,$22,$1C,$8A,$72,$2C
	.dc.b	$72,$6C,$72,$78,$10,$E0,$58,$E0,$D8,$E0,$E8,$91,$A6,$C6,$A3,$04
	.dc.b	$70,$4A,$F9,$2D,$AF,$93,$18,$23,$82,$57,$C3,$8C,$B6,$32,$FB,$A3
	.dc.b	$2D,$8C,$BE,$B8,$CB,$63,$2F,$AA,$32,$D8,$CD,$7A,$E7,$2D,$8C,$D7
	.dc.b	$AE,$BC,$84,$3C,$45,$3C,$05,$3B,$16,$7A,$36,$3A,$3C,$88,$72,$29
	.dc.b	$C8,$B1,$C9,$B1,$C9,$E0,$43,$81,$63,$83,$63,$83,$A2,$46,$9B,$1A
	.dc.b	$8C,$11,$C1,$2B,$E4,$B6,$BE,$4C,$60,$8E,$09,$5F,$0E,$32,$D8,$CB
	.dc.b	$EE,$8C,$B6,$32,$FA,$E3,$2D,$8C,$BE,$A8,$CB,$63,$35,$EB,$9C,$B6
	.dc.b	$3F,$AD,$CB,$C2,$78,$7F,$EF,$58,$FE,$BF,$18,$F0,$F7,$1B,$9D,$3F
	.dc.b	$E3,$43,$9F,$D4,$BE,$03,$FD,$67,$18,$6B,$FD,$AA,$35,$97,$46,$C6
	.dc.b	$E3,$13,$C2,$3F,$DC,$D3,$A7,$94,$65,$FF,$7D,$C6,$5D,$C6,$54,$65
	.dc.b	$DC,$66,$31,$1A,$8C,$12,$7C,$64,$8F,$31,$0F,$1B,$96,$99,$3C,$6E
	.dc.b	$5A,$64,$F1,$B9,$69,$93,$C8,$E5,$A6,$67,$D2,$49,$F4,$46,$5A,$F9
	.dc.b	$08,$78,$8A,$78,$0A,$76,$2C,$75,$9F,$13,$63,$B4,$62,$7B,$C7,$52
	.dc.b	$1D,$4A,$75,$2C,$F3,$7B,$C7,$27,$81,$0E,$05,$8E,$0F,$28,$E0,$E8
	.dc.b	$91,$A7,$84,$6A,$31,$1A,$8C,$12,$7D,$84,$8F,$79,$0F,$3B,$96,$99
	.dc.b	$3C,$8E,$5A,$64,$F1,$B9,$69,$93,$C8,$E5,$A6,$67,$D2,$49,$F4,$46
	.dc.b	$5A,$F9,$08,$78,$8A,$78,$0A,$76,$2C,$75,$9F,$13,$63,$B4,$62,$7B
	.dc.b	$C7,$52,$1D,$4A,$75,$2C,$F3,$7B,$C7,$27,$81,$0E,$05,$8E,$0F,$28
	.dc.b	$E0,$E8,$91,$A7,$84,$6A,$31,$1A,$8C,$12,$7D,$87,$EC,$75,$71,$B7
	.dc.b	$FE,$47,$7F,$E6,$3C,$BF,$96,$D6,$A7,$2D,$8F,$E6,$3F,$76,$FA,$A6
	.dc.b	$3C,$A3,$0D,$8C,$CF,$06,$E1,$FE,$94,$A7,$CA,$62,$04,$8C,$DF,$59
	.dc.b	$D4,$E5,$AF,$F2,$6F,$76,$EA,$69,$2B,$D7,$A7,$C1,$AF,$7B,$5F,$F5
	.dc.b	$3E,$7F,$05,$CE,$9F,$D2,$B2,$F6,$C7,$AE,$F1,$9F,$F3,$79,$CC,$F2
	.dc.b	$C3,$A3,$EC,$3F,$D2,$07,$42,$1F,$EB,$6F,$C0,$5D,$7E,$C6,$E1,$F5
	.dc.b	$9B,$B7,$F6,$26,$D7,$59,$77,$BF,$02,$FA,$BC,$7F,$EC,$29,$CC,$62
	.dc.b	$B8,$56,$E7,$FB,$64,$62,$74,$FD,$48,$EB,$D1,$FD,$65,$A7,$F7,$42
	.dc.b	$C7,$7A,$4F,$F5,$BD,$FE,$64,$FC,$5D,$FF,$BA,$C6,$23,$E3,$35,$FC
	.dc.b	$1E,$5F,$09,$F8,$1A,$7F,$65,$8C,$47,$C9,$8F,$ED,$9E,$13,$F1,$07
	.dc.b	$C9,$7E,$5B,$EA,$3F,$DC,$11,$8D,$51,$89,$D6,$1F,$5E,$F9,$46,$23
	.dc.b	$CB,$C3,$F7,$1F,$CB,$E7,$E7,$7E,$E9,$ED,$3F,$57,$D5,$FA,$2C,$7F
	.dc.b	$1A,$D8,$F7,$B5,$E6,$D8,$FF,$70,$FC,$67,$EB,$BD,$4F,$F0,$9C,$60
	.dc.b	$38,$F1,$13,$B9,$FC,$24,$61,$D9,$E5,$B8,$7B,$38,$3E,$9E,$5E,$D9
	.dc.b	$7C,$87,$FB,$FD,$F0,$B5,$FC,$83,$6B,$F0,$CD,$7F,$82,$F1,$BE,$1B
	.dc.b	$FC,$5B,$5F,$90,$6C,$FE,$36,$75,$3F,$8F,$8C,$4E,$9F,$E1,$FF,$2C
	.dc.b	$FA,$9F,$E2,$3F,$53,$F1,$23,$16,$8C,$4E,$A3,$F4,$D3,$A9,$FF,$1B
	.dc.b	$70,$7D,$92,$67,$F7,$70,$FB,$67,$8A,$FC,$8F,$B3,$1F,$24,$7E,$C1
	.dc.b	$B9,$7A,$A7,$AD,$73,$BE,$BA,$7C,$E2,$4F,$F0,$23,$E9,$14,$FE,$47
	.dc.b	$D3,$FE,$5B,$5C,$2F,$0B,$FC,$B4,$D8,$DF,$84,$EC,$FE,$57,$39,$9F
	.dc.b	$F3,$14,$62,$B8,$92,$38,$C6,$5A,$78,$23,$F0,$C7,$F2,$42,$3F,$BB
	.dc.b	$6B,$A3,$04,$3E,$53,$F9,$AC,$FE,$FB,$74,$FE,$96,$DD,$7D,$C7,$1A
	.dc.b	$FD,$7E,$BF,$D6,$67,$BC,$63,$4F,$73,$F8,$6E,$0F,$EA,$BF,$51,$F8
	.dc.b	$2C,$0D,$23,$F9,$33,$BC,$F9,$9C,$5F,$8C,$FD,$0F,$99,$FD,$83,$AE
	.dc.b	$CF,$F1,$23,$F4,$F1,$97,$F4,$AD,$AF,$88,$43,$D1,$5E,$A9,$FD,$E7
	.dc.b	$2F,$EE,$F9,$BF,$B6,$B6,$0D,$3F,$33,$C1,$FD,$CB,$C6,$B2,$69,$1F
	.dc.b	$95,$83,$93,$FB,$F6,$3F,$2A,$31,$BF,$CD,$7F,$91,$76,$7F,$27,$FA
	.dc.b	$2F,$E4,$D8,$1A,$49,$FE,$65,$1F,$49,$77,$7F,$5A,$C3,$F9,$CC,$E6
	.dc.b	$31,$3E,$4C,$7C,$6D,$C7,$C4,$FF,$05,$EF,$FC,$2B,$FD,$61,$AE,$6F
	.dc.b	$F0,$27,$F5,$4C,$3F,$A9,$78,$1F,$E1,$6F,$C5,$2C,$9A,$4A,$F4,$EB
	.dc.b	$E6,$23,$83,$C9,$CB,$ED,$CF,$CF,$1F,$6C,$78,$5F,$E2,$2C,$BF,$C0
	.dc.b	$C4,$8F,$F1,$07,$87,$E8,$7F,$02,$74,$8F,$A5,$E2,$FF,$86,$73,$FB
	.dc.b	$53,$FE,$D1,$F1,$7E,$F3,$63,$2F,$FB,$3A,$32,$FF,$83,$F0,$FF,$14
	.dc.b	$E7,$FE,$38,$F1,$B3,$3F,$E2,$9B,$E2,$D9,$FF,$0E,$11,$F3,$C6,$9F
	.dc.b	$6E,$5C,$9F,$C3,$E9,$9D,$3F,$C9,$17,$FE,$28,$24,$FF,$B3,$B4,$FF
	.dc.b	$73,$11,$C1,$0F,$F6,$7B,$96,$99,$3F,$D9,$EE,$5A,$64,$FF,$67,$B9
	.dc.b	$69,$9D,$3F,$DF,$31,$96,$C6,$7C,$7F,$A8,$F0,$3C,$24,$3B,$94,$D1
	.dc.b	$82,$18,$24,$76,$78,$35,$DB,$5D,$C7,$8A,$30,$47,$04,$CB,$F5,$D1
	.dc.b	$96,$C6,$6B,$DC,$DA,$F7,$4E,$89,$3A,$6C,$6A,$30,$47,$04,$3D,$73
	.dc.b	$D5,$B3,$C9,$B3,$C5,$A7,$13,$E0,$CF,$16,$CF,$16,$CF,$16,$9C,$4F
	.dc.b	$8F,$1B,$6C,$6D,$A6,$EB,$E3,$46,$5B,$19,$AF,$13,$C0,$F3,$90,$F0
	.dc.b	$94,$D1,$82,$18,$24,$76,$78,$35,$DB,$5D,$C7,$8A,$30,$47,$04,$AF
	.dc.b	$81,$19,$6C,$66,$BD,$CD,$AF,$74,$E8,$93,$A6,$C6,$A3,$04,$70,$43
	.dc.b	$D7,$3D,$5B,$3C,$9B,$3C,$5A,$71,$3E,$0C,$F1,$6C,$F1,$6C,$F1,$69
	.dc.b	$C4,$FF,$0D,$9F,$EA,$D3,$FC,$2E,$5F,$03,$FC,$50,$DF,$03,$FD,$BE
	.dc.b	$B0,$FF,$32,$3C,$4C,$74,$FF,$B7,$1C,$46,$9E,$CD,$7F,$D7,$8D,$7F
	.dc.b	$DB,$2D,$79,$E1,$D5,$A7,$49,$EC,$F4,$6B,$D1,$AF,$46,$BC,$8E,$11
	.dc.b	$D6,$7C,$04,$9D,$B6,$35,$18,$23,$82,$4F,$95,$AF,$63,$04,$30,$43
	.dc.b	$D0,$F0,$69,$B2,$9B,$3D,$8E,$DA,$6C,$A6,$CF,$80,$72,$21,$A2,$9A
	.dc.b	$30,$43,$04,$9F,$88,$F7,$31,$3F,$11,$C4,$69,$F1,$35,$EE,$D7,$B3
	.dc.b	$5E,$73,$ED,$69,$D2,$7B,$3D,$1A,$F4,$6B,$D1,$AF,$23,$84,$75,$9F
	.dc.b	$01,$27,$6D,$8D,$46,$08,$E0,$93,$E5,$6B,$D8,$C1,$0C,$10,$F4,$3C
	.dc.b	$1A,$6C,$A6,$CF,$63,$B6,$9B,$29,$B3,$E0,$1C,$88,$68,$A6,$8C,$10
	.dc.b	$C1,$27,$E2,$3D,$CC,$4F,$C4,$71,$1A,$7C,$4D,$7B,$B5,$EC,$D7,$9C
	.dc.b	$FB,$5A,$74,$9E,$CF,$46,$BD,$1A,$F4,$6B,$C8,$E1,$1D,$67,$C0,$49
	.dc.b	$DB,$E3,$CD,$C8,$D7,$FD,$8D,$87,$F2,$99,$CC,$62,$07,$FA,$B7,$C0
	.dc.b	$FE,$4D,$3A,$7F,$94,$70,$BE,$A1,$9F,$1B,$E3,$39,$8C,$47,$F9,$9A
	.dc.b	$77,$5C,$9B,$1F,$CE,$E9,$C5,$88,$FF,$2E,$35,$FF,$2D,$57,$F9,$76
	.dc.b	$FA,$D7,$05,$E1,$64,$95,$D6,$FC,$F1,$FE,$EF,$8F,$F7,$F9,$FE,$CB
	.dc.b	$31,$8F,$FB,$E1,$B9,$FF,$BE,$C9,$9F,$FB,$EA,$3E,$54,$7C,$D7,$FA
	.dc.b	$A6,$0F,$09,$F7,$DF,$1B,$C9,$CC,$62,$51,$89,$D5,$7D,$23,$E6,$D6
	.dc.b	$E7,$8F,$CB,$FA,$A7,$57,$DE,$31,$1F,$6A,$30,$7C,$47,$74,$9D,$C7
	.dc.b	$C1,$3D,$78,$FB,$5F,$16,$5F,$0D,$C9,$F1,$23,$FB,$55,$F3,$F9,$3F
	.dc.b	$C1,$F9,$7F,$0B,$04,$62,$FC,$F1,$8A,$56,$EB,$55,$E0,$D3,$E7,$BD
	.dc.b	$C8,$78,$88,$7B,$ED,$CF,$F9,$4F,$C3,$7F,$B7,$E3,$EF,$8C,$3E,$26
	.dc.b	$D7,$16,$BF,$0C,$F1,$65,$F0,$DC,$1F,$A6,$FA,$A7,$C4,$FE,$8B,$CF
	.dc.b	$FE,$F7,$18,$C1,$18,$9D,$47,$F8,$66,$75,$3E,$6D,$DF,$4F,$FE,$13
	.dc.b	$AF,$AE,$DC,$3E,$E4,$EA,$7E,$F4,$62,$74,$FE,$23,$E5,$FF,$3F,$3F
	.dc.b	$80,$6C,$72,$6B,$B8,$FD,$49,$F3,$1F,$79,$FA,$63,$F4,$EF,$E7,$67
	.dc.b	$BB,$4F,$1C,$7E,$62,$33,$5E,$78,$FF,$0F,$F0,$FE,$2D,$F9,$A7,$DC
	.dc.b	$29,$E1,$3C,$1D,$8E,$CF,$E3,$3E,$3B,$BB,$F4,$D3,$9E,$BF,$E3,$33
	.dc.b	$9B,$FE,$37,$39,$51,$97,$FC,$96,$D8,$D3,$6F,$BD,$F6,$6D,$74,$6B
	.dc.b	$FE,$E4,$8F,$E4,$FC,$9F,$CB,$DB,$C3,$DE,$7F,$E0,$09,$F8,$C8,$C9
	.dc.b	$1F,$D8,$BF,$C1,$B4,$FE,$5A,$FE,$15,$BA,$FE,$10,$F8,$45,$8F,$9B
	.dc.b	$7E,$1A,$E6,$43,$C0,$7E,$0E,$5E,$5F,$E4,$77,$FE,$76,$D8,$C9,$B7
	.dc.b	$C4,$F8,$27,$F2,$0D,$7F,$00,$D3,$F0,$97,$FA,$17,$F8,$81,$FA,$1E
	.dc.b	$2F,$E9,$3C,$B9,$24,$79,$EB,$E3,$90,$F9,$C7,$EE,$5F,$E5,$C5,$E1
	.dc.b	$FD,$FC,$FF,$02,$8C,$CF,$EF,$5A,$F3,$E1,$FE,$C3,$87,$D1,$8F,$8A
	.dc.b	$43,$59,$FC,$B7,$15,$FC,$7C,$7C,$BF,$F1,$96,$C7,$F1,$EE,$7F,$EC
	.dc.b	$B3,$93,$B4,$74,$DF,$F0,$8F,$B0,$F0,$14,$D6,$1F,$D0,$2F,$EB,$C7
	.dc.b	$F3,$AC,$BE,$BB,$63,$2F,$F0,$68,$CB,$FC,$23,$AB,$FB,$BF,$91,$FD
	.dc.b	$3E,$8D,$24,$FD,$13,$C4,$4D,$3E,$3B,$96,$BE,$86,$BE,$83,$5A,$7C
	.dc.b	$97,$FC,$0B,$DB,$E4,$6D,$FA,$1B,$C1,$FD,$F4,$FE,$FD,$E1,$48,$FE
	.dc.b	$F5,$28,$C4,$7F,$67,$EF,$FF,$08,$47,$F8,$36,$75,$3F,$E0,$DF,$03
	.dc.b	$FC,$2F,$39,$3F,$6A,$48,$FE,$C2,$D7,$FD,$3B,$B7,$FA,$94,$8F,$FA
	.dc.b	$C9,$AE,$8C,$10,$C1,$0F,$F5,$0B,$C1,$A6,$CA,$6C,$FF,$65,$BB,$69
	.dc.b	$B2,$9B,$F0,$5E,$17,$F0,$88,$D4,$63,$0F,$C7,$B6,$BA,$36,$77,$5C
	.dc.b	$5E,$8D,$7C,$A4,$3C,$25,$8E,$4F,$0A,$EF,$3C,$63,$64,$9D,$BC,$63
	.dc.b	$51,$88,$D4,$60,$93,$D0,$95,$EA,$8C,$B6,$33,$3E,$C6,$BE,$58,$D1
	.dc.b	$1D,$35,$D1,$82,$18,$21,$E9,$8E,$AD,$8E,$4D,$38,$94,$E2,$7B,$A3
	.dc.b	$8B,$63,$8B,$4E,$25,$38,$D7,$C7,$25,$7C,$28,$DC,$62,$BA,$B6,$BA
	.dc.b	$B6,$77,$5C,$5E,$AD,$7D,$04,$3C,$45,$8E,$4F,$0A,$EF,$3C,$63,$64
	.dc.b	$9D,$BC,$63,$51,$88,$D4,$60,$93,$D0,$95,$EA,$8C,$B6,$33,$3E,$C6
	.dc.b	$BE,$58,$D1,$1D,$35,$D1,$82,$18,$21,$E9,$8E,$AD,$8E,$4D,$38,$94
	.dc.b	$E2,$7B,$A3,$8B,$F1,$BA,$BF,$C3,$0F,$FA,$0C,$38,$7F,$C2,$9D,$79
	.dc.b	$7F,$C2,$8D,$58,$7F,$8A,$6B,$0D,$8E,$0D,$7B,$35,$FF,$49,$35,$FF
	.dc.b	$62,$B5,$EB,$D5,$FE,$C2,$79,$B5,$E6,$D7,$9B,$5E,$6D,$79,$CF,$9A
	.dc.b	$31,$3D,$C8,$E0,$91,$B2,$4F,$89,$AF,$89,$AF,$89,$AF,$89,$AF,$88
	.dc.b	$E4,$43,$91,$4E,$45,$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10
	.dc.b	$D1,$4D,$18,$21,$82,$4F,$C8,$8F,$6B,$63,$CC,$D7,$C2,$D7,$B3,$5E
	.dc.b	$AD,$7A,$CF,$D2,$79,$B5,$E6,$D7,$9B,$5E,$6D,$79,$CF,$9A,$31,$3D
	.dc.b	$C8,$E0,$91,$B2,$4F,$89,$AF,$89,$AF,$89,$AF,$89,$AF,$88,$E4,$43
	.dc.b	$91,$4E,$45,$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D
	.dc.b	$18,$21,$82,$4F,$C8,$8F,$6B,$63,$CC,$D7,$C2,$D7,$B3,$5E,$AD,$7A
	.dc.b	$CF,$D2,$79,$B5,$E6,$D7,$9B,$5E,$6D,$79,$CF,$9A,$31,$3D,$C8,$E0
	.dc.b	$91,$B2,$4F,$FA,$C2,$A3,$FC,$A7,$81,$C0,$E4,$56,$AB,$FC,$95,$E1
	.dc.b	$B0,$CC,$7F,$31,$E0,$4E,$6B,$F9,$84,$62,$B7,$5C,$9A,$84,$6F,$C3
	.dc.b	$E3,$7D,$DB,$46,$9F,$D0,$57,$4B,$F0,$2E,$18,$75,$3D,$11,$99,$E5
	.dc.b	$19,$78,$CF,$08,$F2,$46,$0D,$BF,$EF,$22,$1E,$9D,$3F,$DF,$0F,$6F
	.dc.b	$7F,$E1,$3D,$6D,$18,$8E,$B1,$8A,$F8,$67,$AA,$75,$3E,$8F,$BA,$FA
	.dc.b	$1F,$93,$FB,$ED,$18,$9E,$CD,$7A,$B5,$F8,$3F,$79,$FD,$37,$0F,$8E
	.dc.b	$DA,$34,$FE,$A1,$8F,$B2,$8F,$BA,$BE,$7F,$B9,$C4,$8F,$F6,$F3,$25
	.dc.b	$B7,$DB,$7B,$23,$EB,$B6,$33,$1F,$09,$BC,$1F,$DE,$6F,$C4,$78,$61
	.dc.b	$F4,$FC,$49,$5C,$5E,$9F,$67,$F9,$7F,$33,$FB,$84,$FE,$1A,$30,$FD
	.dc.b	$56,$E7,$F5,$1A,$FD,$DE,$5F,$D2,$3E,$36,$BF,$A9,$6B,$F8,$49,$7D
	.dc.b	$77,$E9,$D7,$99,$E2,$8C,$36,$3F,$58,$D7,$67,$E5,$61,$F6,$3F,$A3
	.dc.b	$8C,$36,$BF,$2F,$3A,$9F,$D3,$1A,$2C,$6B,$CA,$E0,$FE,$EE,$51,$89
	.dc.b	$F4,$B6,$33,$1C,$A7,$E2,$74,$FF,$8A,$F6,$FA,$B9,$7C,$29,$E8,$F1
	.dc.b	$25,$71,$3E,$2B,$E2,$8F,$C6,$BC,$5B,$5D,$63,$15,$DA,$30,$49,$FF
	.dc.b	$1F,$76,$3A,$FF,$8A,$E4,$ED,$FF,$21,$D7,$B9,$B1,$98,$FF,$22,$99
	.dc.b	$2B,$97,$FC,$9B,$3D,$27,$89,$23,$FD,$F8,$7F,$BF,$F1,$FF,$7E,$3E
	.dc.b	$A8,$FC,$46,$1C,$E3,$67,$C6,$C9,$E0,$FE,$16,$FF,$97,$36,$32,$7E
	.dc.b	$0E,$32,$FE,$46,$BA,$57,$73,$17,$E0,$3F,$93,$69,$F6,$B0,$FC,$96
	.dc.b	$5C,$AB,$F0,$98,$76,$6C,$FE,$5F,$5D,$47,$E1,$DB,$3F,$9E,$6D,$FE
	.dc.b	$46,$BF,$35,$C5,$2B,$F3,$47,$F3,$22,$9F,$CE,$0F,$3D,$C3,$E5,$65
	.dc.b	$F1,$E3,$F7,$03,$C3,$FD,$DD,$38,$7F,$AA,$46,$1B,$3F,$BD,$38,$1F
	.dc.b	$CC,$F8,$FF,$AD,$78,$3F,$CA,$DB,$8E,$65,$39,$6C,$79,$4F,$B2,$7C
	.dc.b	$F3,$F9,$DF,$37,$26,$C7,$06,$CF,$F3,$99,$D4,$7F,$54,$E7,$FE,$D9
	.dc.b	$C1,$FC,$F5,$B1,$FD,$25,$AF,$F4,$86,$BE,$2E,$4F,$EF,$3A,$FE,$09
	.dc.b	$B5,$FD,$1F,$A3,$79,$7F,$40,$3E,$66,$DF,$D0,$1B,$1A,$8F,$E9,$B0
	.dc.b	$FC,$93,$FA,$87,$57,$F8,$05,$B1,$9C,$BF,$A8,$B5,$FE,$BB,$D9,$EE
	.dc.b	$8F,$EB,$87,$C3,$8F,$EC,$CD,$33,$B7,$E5,$E3,$E7,$1F,$34,$FE,$1F
	.dc.b	$97,$F6,$1E,$DF,$D2,$C7,$F6,$D6,$BF,$DB,$1A,$FF,$80,$1A,$FF,$83
	.dc.b	$BB,$EF,$81,$FE,$15,$E5,$FF,$06,$92,$7F,$C3,$93,$F5,$9E,$18,$FA
	.dc.b	$5B,$8F,$F8,$50,$FD,$E9,$1F,$DA,$35,$FF,$56,$63,$FE,$B1,$23,$FE
	.dc.b	$B1,$6B,$A3,$04,$30,$43,$FD,$68,$F0,$69,$B2,$9B,$3F,$D9,$2E,$DA
	.dc.b	$6C,$A6,$FC,$1F,$F7,$24,$6D,$B1,$B6,$BB,$EB,$FC,$9B,$C8,$F0,$90
	.dc.b	$C1,$23,$8B,$96,$B9,$8E,$71,$82,$38,$25,$FF,$20,$8C,$B6,$33,$5E
	.dc.b	$96,$D7,$A5,$AF,$A6,$78,$12,$78,$36,$38,$35,$E1,$1A,$23,$A6,$BA
	.dc.b	$30,$43,$04,$3E,$14,$F9,$1B,$3D,$DB,$3D,$1A,$73,$29,$CC,$F9,$51
	.dc.b	$C5,$B1,$C5,$A7,$12,$9C,$6B,$E5,$C6,$DB,$1B,$6B,$BA,$F4,$3C,$8F
	.dc.b	$19,$0C,$12,$38,$B9,$6B,$98,$E7,$18,$23,$82,$57,$B6,$32,$D8,$CD
	.dc.b	$7A,$5B,$5E,$96,$BE,$99,$E0,$49,$E0,$D8,$E0,$D7,$84,$68,$8E,$9A
	.dc.b	$E8,$C1,$0C,$10,$F8,$53,$E4,$6C,$F7,$6C,$F4,$69,$CC,$A7,$E0,$7B
	.dc.b	$3F,$C4,$5D,$FF,$EA,$18,$FF,$10,$96,$32,$FF,$8A,$2F,$FD,$66,$FF
	.dc.b	$AC,$F0,$7D,$37,$FA,$FE,$39,$08,$CE,$38,$8C,$4E,$DF,$04,$E9,$E2
	.dc.b	$D7,$31,$B7,$FD,$6D,$19,$6C,$65,$F2,$35,$EF,$3E,$19,$E1,$1A,$8C
	.dc.b	$4E,$E7,$33,$C0,$F1,$C6,$49,$3C,$09,$1A,$6B,$E8,$6B,$E6,$6B,$E1
	.dc.b	$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$71,$C0
	.dc.b	$A6,$CA,$6E,$7E,$29,$27,$E1,$C7,$94,$D9,$23,$53,$F0,$23,$13,$C1
	.dc.b	$F0,$4E,$9E,$4D,$73,$1B,$7C,$91,$96,$C6,$5F,$33,$5E,$F3,$E1,$9E
	.dc.b	$11,$A8,$C4,$EE,$73,$3C,$0F,$1C,$64,$93,$C0,$91,$A6,$BE,$86,$BE
	.dc.b	$66,$BE,$13,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$10
	.dc.b	$F7,$1C,$0A,$6C,$A6,$E7,$E2,$92,$7E,$1C,$79,$4D,$92,$35,$3F,$02
	.dc.b	$31,$3C,$1F,$04,$E9,$E4,$D7,$31,$B7,$C9,$19,$6C,$65,$F3,$35,$EF
	.dc.b	$3E,$19,$E1,$1A,$8C,$4E,$E7,$33,$C0,$F1,$C6,$49,$3C,$09,$1A,$7F
	.dc.b	$3D,$C9,$FE,$6E,$C7,$FD,$76,$49,$FF,$59,$3F,$EB,$EF,$0A,$47,$FA
	.dc.b	$D5,$A8,$78,$2F,$D4,$DE,$DB,$15,$FF,$39,$AD,$4E,$63,$15,$C2,$B7
	.dc.b	$7D,$4E,$B1,$99,$FF,$68,$C4,$3F,$D9,$05,$3C,$92,$EF,$7D,$DF,$F3
	.dc.b	$18,$C3,$63,$31,$FD,$10,$C9,$5F,$27,$DF,$7F,$B4,$31,$F6,$36,$4D
	.dc.b	$3F,$DA,$BF,$6F,$ED,$C3,$D5,$3E,$38,$C4,$79,$AD,$18,$8F,$91,$18
	.dc.b	$9F,$87,$1F,$0F,$5F,$F5,$2B,$EF,$3E,$93,$87,$E1,$9E,$09,$7D,$C7
	.dc.b	$BC,$95,$ED,$7E,$17,$97,$F5,$F6,$FF,$76,$B6,$F9,$35,$F4,$35,$F0
	.dc.b	$F8,$EF,$03,$FB,$56,$5E,$39,$F4,$4E,$5F,$46,$3F,$57,$A3,$E9,$11
	.dc.b	$FA,$0F,$D5,$E7,$FC,$57,$C1,$FF,$73,$9F,$B9,$18,$7D,$CD,$AF,$98
	.dc.b	$D7,$CB,$8F,$93,$0F,$14,$7C,$63,$CB,$1F,$7F,$5E,$17,$F8,$96,$9F
	.dc.b	$8B,$58,$78,$E3,$0D,$FB,$4F,$CE,$4F,$D1,$6D,$FD,$F6,$BF,$79,$B1
	.dc.b	$F9,$7F,$B7,$B6,$69,$23,$F5,$67,$A6,$BD,$A7,$E0,$C6,$9F,$45,$FC
	.dc.b	$5F,$9D,$D0,$93,$CA,$3F,$25,$1E,$47,$8E,$7F,$19,$F8,$A7,$5A,$E8
	.dc.b	$EB,$3F,$98,$47,$C5,$D7,$FE,$34,$F2,$DB,$7E,$EE,$30,$DE,$C7,$57
	.dc.b	$F0,$BA,$DD,$FF,$0D,$BD,$B6,$B6,$A7,$6A,$76,$D9,$DC,$EA,$73,$8B
	.dc.b	$2F,$CC,$63,$FE,$F7,$C3,$F0,$6D,$7F,$DE,$E7,$FB,$EA,$78,$65,$E0
	.dc.b	$8C,$9F,$2B,$47,$77,$E1,$75,$FE,$44,$D8,$C9,$FB,$58,$CB,$D7,$1F
	.dc.b	$1D,$5B,$ED,$3F,$85,$38,$C7,$F0,$9D,$B8,$B6,$32,$FF,$0A,$8C,$BF
	.dc.b	$8F,$8C,$9F,$C9,$78,$32,$D3,$3C,$08,$CB,$FA,$1F,$80,$FD,$03,$FC
	.dc.b	$F5,$B2,$69,$FE,$63,$D3,$E3,$EB,$F2,$F0,$FC,$43,$F3,$6B,$F7,$67
	.dc.b	$13,$F7,$65,$8F,$DD,$B5,$FE,$4B,$1E,$18,$FE,$3D,$C7,$FD,$3E,$3F
	.dc.b	$D0,$D1,$FC,$34,$F9,$38,$27,$55,$C2,$7E,$D3,$E4,$3E,$85,$79,$69
	.dc.b	$C9,$FD,$6F,$CB,$FE,$4B,$66,$92,$FF,$98,$73,$F6,$7F,$9F,$35,$FE
	.dc.b	$7A,$D9,$FE,$AC,$E6,$3C,$47,$E4,$8F,$19,$FC,$EF,$A3,$F1,$6D,$9F
	.dc.b	$C7,$B7,$8B,$FA,$8C,$62,$3F,$A7,$75,$7F,$73,$FC,$93,$FA,$C1,$23
	.dc.b	$FA,$34,$65,$FE,$91,$D9,$E5,$D3,$FB,$2E,$BF,$31,$B3,$AC,$3E,$4E
	.dc.b	$FF,$7A,$3C,$E7,$A4,$FF,$03,$75,$FF,$80,$27,$C9,$1F,$DF,$31,$F5
	.dc.b	$BA,$EE,$FF,$01,$69,$FE,$04,$9C,$CF,$F3,$A8,$FF,$5C,$E1,$FB,$26
	.dc.b	$F3,$7F,$51,$6F,$37,$6A,$FE,$2D,$E0,$E0,$8F,$FB,$2E,$FF,$77,$1F
	.dc.b	$EA,$5D,$64,$FF,$A9,$5A,$FF,$A9,$5A,$FF,$AD,$5A,$FF,$AE,$9B,$E0
	.dc.b	$FF,$B3,$1B,$E0,$FD,$D8,$C4,$F5,$21,$D4,$A7,$52,$9D,$4A,$F2,$78
	.dc.b	$10,$E0,$53,$81,$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$23,$D8,$4D,$7F
	.dc.b	$57,$19,$6C,$66,$BE,$1C,$60,$8E,$09,$5E,$E9,$E0,$D9,$DB,$67,$6F
	.dc.b	$B6,$76,$D9,$DB,$67,$6F,$BA,$76,$D9,$DB,$67,$6F,$C1,$9D,$B6,$76
	.dc.b	$D9,$DD,$7C,$16,$D7,$08,$C4,$FA,$48,$79,$4A,$78,$4A,$75,$2B,$C9
	.dc.b	$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA,$70,$47,$04,$8F,$61,$2B
	.dc.b	$E1,$C6,$5B,$19,$AF,$87,$18,$23,$82,$57,$BA,$78,$36,$76,$D9,$DB
	.dc.b	$ED,$9D,$B6,$76,$D9,$DB,$EE,$9D,$B6,$76,$D9,$DB,$F0,$67,$6D,$9D
	.dc.b	$B6,$77,$5F,$05,$B5,$C2,$31,$3F,$73,$AD,$DB,$FE,$1C,$8F,$F5,$01
	.dc.b	$FE,$19,$EE,$AF,$F8,$85,$AE,$4C,$D3,$DC,$FF,$58,$46,$1A,$F4,$8D
	.dc.b	$65,$C9,$B1,$B8,$C4,$F0,$8F,$F5,$94,$E9,$E5,$19,$7F,$D8,$51,$97
	.dc.b	$71,$95,$19,$77,$19,$8C,$46,$A3,$04,$9F,$19,$23,$CC,$43,$C6,$E5
	.dc.b	$A6,$4F,$1B,$96,$99,$3C,$6E,$5A,$64,$F2,$39,$69,$99,$F4,$92,$7D
	.dc.b	$11,$96,$BE,$42,$1E,$22,$9E,$02,$9D,$8B,$1D,$67,$C4,$D8,$ED,$18
	.dc.b	$9E,$F1,$D4,$87,$52,$9D,$4B,$3C,$DE,$F1,$C9,$E0,$43,$81,$63,$83
	.dc.b	$CA,$38,$3A,$24,$69,$E1,$1A,$8C,$46,$A3,$04,$9F,$61,$23,$DE,$43
	.dc.b	$CE,$E5,$A6,$4F,$23,$96,$99,$3C,$6E,$5A,$64,$F2,$39,$69,$99,$F4
	.dc.b	$92,$7D,$11,$96,$BE,$42,$1E,$22,$9E,$02,$9D,$8B,$1D,$67,$C4,$D8
	.dc.b	$ED,$18,$9E,$F1,$D4,$87,$52,$9D,$4B,$3C,$DE,$F1,$C9,$E0,$43,$81
	.dc.b	$63,$83,$CA,$38,$3A,$24,$69,$E1,$1A,$8C,$46,$A3,$04,$9F,$61,$23
	.dc.b	$DE,$43,$CE,$E5,$F9,$3A,$7F,$AD,$AF,$FC,$86,$48,$FF,$20,$6E,$DB
	.dc.b	$F9,$84,$07,$C0,$3E,$35,$8E,$A7,$31,$8C,$3F,$9A,$D6,$EB,$4D,$8F
	.dc.b	$F7,$46,$1F,$EE,$BB,$FF,$7B,$9F,$EF,$9E,$47,$D9,$BD,$BF,$E5,$91
	.dc.b	$86,$C6,$71,$FE,$87,$39,$7F,$A2,$8E,$77,$C4,$FA,$3B,$BF,$CF,$0D
	.dc.b	$C7,$DC,$7B,$27,$75,$FE,$9E,$8C,$BC,$E7,$84,$7C,$58,$ED,$F5,$5F
	.dc.b	$D0,$5A,$78,$A9,$19,$6C,$7F,$58,$E0,$FF,$74,$D7,$C4,$E0,$C6,$BF
	.dc.b	$16,$BD,$5F,$65,$C4,$8F,$57,$A7,$D9,$FF,$5E,$8F,$3E,$28,$C6,$1E
	.dc.b	$78,$C4,$FA,$63,$13,$A7,$E9,$9F,$13,$1F,$2B,$F8,$6C,$7C,$A7,$F7
	.dc.b	$2F,$83,$7E,$A4,$F5,$F9,$46,$1B,$19,$C7,$C1,$6A,$D4,$60,$FB,$18
	.dc.b	$F8,$DA,$67,$2F,$4D,$7F,$7A,$8C,$BF,$EE,$DD,$3F,$18,$B3,$F4,$BE
	.dc.b	$B8,$CC,$62,$3F,$01,$7D,$EF,$AD,$FC,$A8,$C1,$9F,$AF,$91,$E3,$6D
	.dc.b	$78,$E9,$18,$9F,$A6,$D9,$F1,$69,$F2,$9F,$C0,$1F,$EC,$B8,$D3,$F9
	.dc.b	$36,$BE,$BC,$7D,$24,$9E,$51,$F5,$4E,$65,$3F,$C5,$3D,$5F,$E2,$CE
	.dc.b	$5F,$DE,$79,$6E,$DF,$D0,$CE,$69,$39,$7F,$C5,$C6,$27,$7D,$8B,$E7
	.dc.b	$18,$6C,$F0,$AE,$4F,$F8,$D5,$B7,$DD,$CE,$DF,$EE,$79,$FF,$73,$1F
	.dc.b	$C4,$5C,$37,$2F,$E2,$D8,$7F,$B7,$09,$1F,$EE,$E3,$E4,$69,$ED,$3E
	.dc.b	$89,$5F,$D1,$69,$FC,$6A,$7E,$C3,$67,$94,$FC,$FD,$3F,$DF,$77,$E0
	.dc.b	$8E,$77,$C2,$BA,$47,$82,$30,$48,$D9,$23,$C2,$F0,$3F,$3B,$1F,$89
	.dc.b	$DB,$C5,$3F,$B4,$9D,$6D,$FB,$8A,$FC,$F6,$1E,$42,$3F,$0A,$B8,$3C
	.dc.b	$EB,$8D,$F4,$9F,$CF,$C7,$A3,$7F,$D8,$D7,$F9,$91,$B1,$FC,$91,$B3
	.dc.b	$FD,$0D,$AF,$D3,$BF,$0B,$F4,$67,$D6,$7E,$A4,$F0,$92,$BF,$4D,$C5
	.dc.b	$F4,$67,$F8,$BB,$5F,$E2,$CD,$7F,$9A,$35,$FA,$91,$EF,$7C,$4D,$9F
	.dc.b	$E6,$2F,$FA,$21,$B1,$FC,$93,$8F,$A4,$62,$3F,$92,$72,$7F,$5C,$F9
	.dc.b	$8F,$E8,$B2,$69,$27,$FA,$3F,$3F,$F4,$FC,$3F,$D2,$AD,$8C,$F4,$7F
	.dc.b	$55,$9C,$B6,$3F,$1B,$C3,$F9,$C7,$D6,$7F,$A7,$BA,$BF,$B4,$77,$FF
	.dc.b	$6A,$EC,$FE,$DE,$DC,$CD,$3F,$D4,$76,$E3,$FE,$F5,$18,$6E,$1F,$CE
	.dc.b	$63,$1C,$1C,$A3,$FD,$0F,$9F,$FA,$7C,$8E,$08,$7F,$A2,$1C,$B4,$C9
	.dc.b	$FE,$88,$72,$D3,$27,$FA,$21,$CB,$4C,$F7,$7F,$AC,$A3,$2D,$8C,$F7
	.dc.b	$7E,$06,$31,$1E,$12,$1D,$CA,$75,$2B,$C1,$D1,$0D,$15,$D3,$82,$38
	.dc.b	$24,$78,$C9,$97,$F1,$E9,$E5,$18,$AF,$5B,$63,$D7,$3B,$7A,$C6,$5F
	.dc.b	$3B,$5E,$71,$BA,$F0,$B6,$75,$3C,$A3,$04,$AF,$0B,$63,$C3,$18,$23
	.dc.b	$82,$1E,$28,$E0,$D8,$DB,$4D,$9E,$67,$2D,$32,$79,$5C,$B4,$CD,$7B
	.dc.b	$23,$2D,$8C,$D7,$BA,$31,$1E,$12,$1D,$CA,$75,$2B,$C1,$D1,$0D,$15
	.dc.b	$D3,$82,$38,$24,$78,$C9,$5F,$1A,$79,$46,$2B,$D6,$D8,$F5,$CE,$DE
	.dc.b	$B1,$97,$CE,$D7,$9C,$6E,$BC,$2D,$9D,$4F,$28,$C1,$2B,$C2,$D8,$F0
	.dc.b	$C6,$08,$E0,$87,$8A,$38,$36,$36,$D3,$67,$99,$CB,$4C,$9E,$57,$2D
	.dc.b	$33,$5F,$B0,$E8,$FF,$09,$47,$F8,$63,$AB,$FD,$1C,$A3,$2F,$F8,$3C
	.dc.b	$FF,$3C,$9F,$E9,$79,$7D,$3F,$FA,$5A,$32,$23,$38,$62,$31,$3B,$79
	.dc.b	$CE,$9E,$2D,$73,$1B,$7F,$D1,$31,$96,$C6,$5F,$23,$5E,$F3,$E1,$9E
	.dc.b	$11,$A8,$C4,$EE,$73,$3C,$0F,$1C,$64,$93,$C0,$91,$A6,$BE,$86,$BE
	.dc.b	$66,$BE,$13,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$10
	.dc.b	$F7,$1C,$0A,$6C,$A6,$E7,$E2,$92,$7E,$1C,$79,$4D,$92,$35,$3F,$02
	.dc.b	$31,$3C,$1F,$04,$E9,$E4,$D7,$31,$B7,$C9,$19,$6C,$65,$F3,$35,$EF
	.dc.b	$3E,$19,$E1,$1A,$8C,$4E,$E7,$33,$C0,$F1,$C6,$49,$3C,$09,$1A,$6B
	.dc.b	$E8,$6B,$E6,$6B,$E1,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$C1,$0F,$71,$C0,$A6,$CA,$6E,$7E,$29,$27,$E1,$C7,$94,$D9,$23
	.dc.b	$53,$F0,$23,$13,$C1,$F0,$4E,$9E,$4D,$73,$1B,$7C,$91,$96,$C6,$5F
	.dc.b	$33,$5E,$F3,$E1,$9E,$11,$A8,$C4,$EE,$73,$3C,$0F,$1C,$64,$93,$C0
	.dc.b	$91,$A6,$BE,$86,$BF,$ED,$2E,$BF,$F2,$19,$1F,$F7,$F3,$FE,$CA,$7F
	.dc.b	$DA,$E5,$3F,$D0,$05,$87,$4B,$E8,$90,$F7,$3F,$CE,$53,$98,$C5,$08
	.dc.b	$F2,$AE,$35,$C1,$B3,$B3,$AF,$EF,$CF,$87,$FE,$5D,$68,$C3,$FD,$00
	.dc.b	$C4,$6A,$33,$3E,$B3,$FD,$7C,$5E,$3F,$F3,$57,$0F,$FB,$B2,$3E,$09
	.dc.b	$E1,$BE,$DC,$9F,$D3,$5C,$5A,$31,$3E,$C8,$C4,$FB,$A3,$13,$A7,$E5
	.dc.b	$9F,$EF,$73,$B9,$0C,$94,$F3,$9B,$3E,$1B,$CE,$3C,$A7,$78,$CB,$4C
	.dc.b	$BF,$EC,$A3,$E0,$72,$2F,$E4,$5B,$ED,$BA,$36,$F9,$36,$FF,$B4,$47
	.dc.b	$D1,$9D,$FD,$BC,$8D,$B6,$77,$4F,$87,$EA,$4A,$EA,$FC,$59,$F7,$9F
	.dc.b	$07,$3F,$AA,$FF,$A6,$8F,$F5,$F5,$2F,$B5,$FD,$46,$CF,$1C,$BD,$CF
	.dc.b	$C0,$CB,$EF,$3F,$38,$87,$F8,$3F,$91,$C3,$F9,$7F,$09,$CD,$FE,$11
	.dc.b	$6D,$7E,$6F,$99,$7C,$21,$3A,$7A,$5F,$3B,$E4,$48,$E5,$5C,$6B,$85
	.dc.b	$F4,$BE,$15,$FE,$E0,$3F,$D9,$78,$FD,$9C,$3F,$DC,$91,$FE,$CC,$3F
	.dc.b	$51,$AF,$FB,$3C,$FC,$EC,$7C,$8C,$FF,$6E,$FD,$46,$E9,$F2,$4F,$79
	.dc.b	$63,$FC,$41,$3C,$EF,$C1,$7F,$F0,$07,$5F,$B9,$7C,$23,$F4,$5B,$74
	.dc.b	$6C,$65,$FD,$DC,$65,$FD,$EE,$36,$7F,$78,$DA,$FB,$D1,$C4,$C9,$5F
	.dc.b	$CB,$63,$F8,$ED,$F8,$12,$7C,$17,$F1,$4F,$94,$7C,$6C,$7D,$6D,$DF
	.dc.b	$FD,$79,$3F,$AC,$BF,$CC,$F0,$78,$C9,$1D,$34,$F8,$95,$FE,$BD,$6D
	.dc.b	$7D,$46,$DF,$F9,$61,$FA,$6F,$F0,$AB,$E7,$3D,$AF,$D9,$1F,$16,$7D
	.dc.b	$04,$AE,$6F,$9E,$BB,$61,$EA,$8F,$69,$FC,$A3,$87,$EF,$BF,$C6,$1A
	.dc.b	$FF,$17,$6B,$BC,$3E,$0F,$27,$F3,$9F,$A7,$FE,$69,$06,$92,$7F,$27
	.dc.b	$9F,$C1,$C3,$C5,$1E,$76,$E5,$E8,$8F,$E8,$4F,$F3,$0E,$9F,$F3,$BB
	.dc.b	$63,$E3,$4F,$F1,$59,$CB,$63,$FC,$B9,$9F,$F9,$7C,$8E,$08,$7F,$97
	.dc.b	$DC,$B4,$C9,$FE,$71,$72,$D3,$27,$F9,$E1,$CB,$4C,$F4,$7F,$A3,$23
	.dc.b	$2D,$8C,$F4,$7C,$C7,$81,$E1,$21,$DC,$A6,$8C,$10,$C1,$23,$B3,$C1
	.dc.b	$AE,$DA,$EE,$3C,$51,$82,$38,$26,$7F,$C5,$E7,$94,$62,$BD,$AD,$8F
	.dc.b	$6C,$ED,$EC,$DA,$EA,$D9,$E1,$3E,$76,$BD,$A3,$44,$AE,$ED,$8E,$F1
	.dc.b	$82,$38,$21,$D6,$38,$36,$36,$D3,$67,$85,$CB,$4C,$9E,$C7,$2D,$33
	.dc.b	$5E,$E8,$CB,$63,$35,$EE,$78,$1E,$12,$1D,$CA,$68,$C1,$0C,$12,$3B
	.dc.b	$3C,$1A,$ED,$AE,$E3,$C5,$18,$23,$82,$57,$C8,$9E,$51,$8A,$F6,$B6
	.dc.b	$3D,$B3,$B7,$B3,$6B,$AB,$67,$84,$F9,$DA,$F6,$8D,$12,$BB,$B6,$3B
	.dc.b	$C6,$08,$E0,$87,$58,$E0,$D8,$DB,$4D,$9E,$17,$2D,$32,$7B,$1C,$B4
	.dc.b	$CD,$7B,$A3,$2F,$EC,$39,$3F,$BC,$4F,$F9,$64,$FE,$FB,$C9,$5A,$58
	.dc.b	$DB,$BF,$9A,$7C,$CF,$F9,$B2,$38,$46,$31,$E5,$08,$C3,$A9,$E0,$F3
	.dc.b	$6B,$FE,$6B,$8D,$B9,$9E,$71,$88,$D4,$F0,$7F,$CF,$11,$96,$C6,$5F
	.dc.b	$13,$63,$51,$89,$EB,$39,$23,$82,$47,$A5,$AF,$94,$C1,$0C,$10,$F3
	.dc.b	$BC,$1A,$6C,$A6,$CF,$4B,$B6,$9B,$29,$B3,$D4,$64,$A6,$67,$DE,$49
	.dc.b	$F6,$C6,$E3,$13,$DE,$38,$9E,$42,$1E,$02,$9A,$30,$43,$04,$8E,$AF
	.dc.b	$79,$F0,$BE,$56,$BD,$23,$6E,$67,$C9,$18,$8D,$4F,$07,$C1,$19,$6C
	.dc.b	$65,$F1,$36,$35,$18,$9E,$B3,$92,$38,$24,$7A,$5A,$F9,$4C,$10,$C1
	.dc.b	$0F,$3B,$C1,$A6,$CA,$6C,$F4,$BB,$69,$B2,$9B,$3D,$46,$4A,$66,$7D
	.dc.b	$E4,$9F,$6C,$6E,$31,$3D,$E3,$89,$E4,$21,$E0,$29,$A3,$04,$30,$48
	.dc.b	$EA,$F7,$9F,$0B,$E5,$6B,$D2,$36,$E6,$7C,$91,$88,$D4,$F0,$7C,$11
	.dc.b	$96,$C6,$5F,$13,$63,$51,$89,$EB,$39,$23,$82,$47,$A5,$AF,$94,$C1
	.dc.b	$0C,$1F,$B5,$D7,$FD,$F3,$87,$F9,$1B,$6F,$F5,$94,$FF,$9E,$E3,$07
	.dc.b	$F3,$1E,$C7,$CA,$30,$DD,$04,$EA,$33,$18,$9E,$15,$CD,$B1,$FC,$FE
	.dc.b	$9E,$5E,$23,$A1,$D1,$C3,$DE,$37,$7D,$9B,$7F,$D0,$3E,$2B,$75,$FC
	.dc.b	$FD,$BA,$1A,$6B,$3F,$F7,$3C,$FF,$9A,$B3,$FF,$6F,$BF,$EE,$C3,$C1
	.dc.b	$9F,$F3,$9C,$7D,$D1,$EB,$CB,$FA,$6C,$62,$3D,$30,$E8,$7C,$7F,$F5
	.dc.b	$58,$CC,$61,$F9,$46,$23,$53,$89,$D3,$FE,$CC,$E3,$FF,$7A,$BE,$73
	.dc.b	$FD,$FC,$53,$FD,$51,$D9,$F4,$2E,$9F,$EF,$62,$ED,$FE,$D6,$8D,$4F
	.dc.b	$7C,$3C,$2F,$DF,$C3,$C3,$7E,$53,$CF,$1B,$6B,$FE,$96,$C3,$CA,$FF
	.dc.b	$B1,$4F,$29,$4E,$47,$F7,$08,$F8,$D5,$E0,$69,$D0,$B1,$BF,$D9,$1D
	.dc.b	$5B,$6C,$ED,$B3,$B9,$D1,$27,$4D,$8F,$99,$AF,$E0,$36,$FC,$1C,$79
	.dc.b	$CF,$AB,$7D,$2F,$EE,$B6,$33,$7C,$E7,$E9,$47,$13,$E2,$5F,$5C,$7E
	.dc.b	$2C,$60,$F8,$A5,$3F,$C4,$5C,$7F,$E2,$4D,$1C,$BF,$A4,$8C,$37,$B9
	.dc.b	$BF,$EC,$EB,$77,$FE,$28,$6A,$BE,$2E,$6F,$95,$71,$6C,$2B,$8E,$BF
	.dc.b	$ED,$38,$FF,$6B,$65,$FE,$D9,$25,$7F,$BB,$1F,$F5,$26,$6D,$7E,$AC
	.dc.b	$FD,$38,$F5,$9F,$2B,$5F,$E1,$91,$93,$E9,$9C,$CF,$E2,$C7,$F9,$16
	.dc.b	$3A,$47,$07,$91,$D2,$FE,$9B,$5F,$E3,$15,$E2,$8D,$D7,$76,$92,$7F
	.dc.b	$39,$1F,$95,$DF,$F6,$CD,$DF,$C1,$3A,$AF,$F7,$DC,$F9,$A3,$F4,$CE
	.dc.b	$E3,$C4,$7E,$71,$C1,$FB,$15,$D1,$48,$FE,$E1,$F8,$73,$D8,$EB,$11
	.dc.b	$F1,$10,$FE,$2E,$FF,$B3,$F8,$7F,$9A,$57,$F1,$72,$4F,$33,$F9,$7E
	.dc.b	$8E,$16,$BF,$ED,$7E,$4F,$E5,$18,$F9,$AF,$FC,$E1,$C9,$FD,$0F,$3F
	.dc.b	$E6,$13,$99,$FF,$34,$4E,$5F,$E7,$DC,$B5,$FE,$7B,$A7,$F3,$AB,$FE
	.dc.b	$9D,$39,$9F,$E7,$CD,$7F,$A5,$5F,$57,$DB,$3F,$CC,$48,$FF,$39,$7F
	.dc.b	$AE,$11,$C1,$FE,$79,$D3,$FC,$D0,$47,$04,$3F,$CC,$AE,$5A,$64,$FF
	.dc.b	$31,$B9,$69,$93,$FC,$F4,$E5,$A6,$71,$FF,$4A,$46,$5B,$19,$E6,$FD
	.dc.b	$1B,$6F,$E9,$46,$23,$C6,$43,$C2,$53,$B9,$67,$AB,$63,$93,$C0,$87
	.dc.b	$02,$C7,$06,$C7,$07,$44,$8D,$36,$35,$18,$23,$82,$4F,$B0,$95,$F0
	.dc.b	$E3,$2D,$8C,$D7,$C3,$8C,$11,$C1,$0F,$74,$70,$6C,$6D,$A6,$CF,$6C
	.dc.b	$6D,$B1,$B6,$9B,$3D,$D1,$B6,$C6,$DA,$6E,$BE,$0C,$ED,$B3,$B6,$CE
	.dc.b	$EB,$E0,$B6,$B8,$46,$23,$D2,$43,$CA,$53,$C2,$59,$EA,$D8,$E4,$F0
	.dc.b	$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D,$8D,$46,$08,$E0,$93,$EC,$25
	.dc.b	$7C,$38,$CB,$63,$35,$F0,$E3,$04,$70,$43,$DD,$1C,$1B,$1B,$69,$B3
	.dc.b	$DB,$1B,$6C,$6D,$A6,$CF,$74,$6D,$B1,$B6,$9B,$AF,$83,$3B,$6C,$ED
	.dc.b	$B3,$F8,$DE,$47,$2F,$F9,$AA,$3F,$BC,$36,$33,$CB,$FE,$66,$9F,$EE
	.dc.b	$E5,$A1,$FE,$77,$74,$61,$C3,$B6,$BF,$E7,$07,$8E,$1C,$23,$83,$63
	.dc.b	$31,$D1,$AF,$38,$D4,$74,$8F,$F4,$33,$A7,$B3,$97,$A9,$CD,$C1,$B7
	.dc.b	$8B,$82,$47,$62,$47,$16,$BE,$33,$04,$30,$43,$C6,$F0,$69,$B2,$9B
	.dc.b	$3C,$6E,$DA,$6C,$A6,$CF,$69,$63,$D8,$E0,$EA,$43,$89,$4D,$18,$21
	.dc.b	$82,$3D,$08,$7C,$67,$C8,$69,$C3,$E3,$6B,$DD,$E3,$1E,$D8,$E0,$D8
	.dc.b	$CC,$74,$6B,$CE,$35,$1D,$23,$CE,$E9,$EC,$E5,$EA,$73,$70,$6D,$E2
	.dc.b	$E0,$91,$D8,$91,$C5,$AF,$8C,$C1,$0C,$10,$F1,$BC,$1A,$6C,$A6,$CF
	.dc.b	$1B,$B6,$9B,$29,$B3,$DA,$58,$F6,$38,$3A,$90,$E2,$53,$46,$08,$60
	.dc.b	$8F,$42,$1F,$19,$F2,$1A,$70,$F8,$DA,$F7,$78,$C7,$B6,$38,$36,$33
	.dc.b	$1D,$1A,$F3,$8D,$47,$48,$F3,$BA,$7B,$39,$7A,$9C,$DC,$1B,$78,$B8
	.dc.b	$24,$76,$24,$71,$6B,$E3,$30,$43,$04,$3E,$77,$07,$F9,$87,$7F,$F5
	.dc.b	$5C,$7F,$2D,$8C,$5F,$F2,$A6,$C6,$78,$3F,$CC,$F3,$FC,$BB,$77,$43
	.dc.b	$1E,$77,$FE,$60,$8C,$56,$DB,$1F,$CE,$EB,$8D,$70,$8F,$F3,$1B,$54
	.dc.b	$38,$38,$C7,$F9,$B1,$AF,$0B,$FF,$36,$D3,$DE,$A4,$9F,$F3,$AC,$65
	.dc.b	$FF,$6C,$9F,$ED,$33,$FD,$B4,$32,$FF,$72,$BE,$DE,$47,$77,$FB,$E5
	.dc.b	$B5,$FE,$F9,$6C,$FF,$AA,$E7,$FA,$A4,$F6,$8E,$9A,$7F,$B9,$1B,$AF
	.dc.b	$FB,$9C,$8F,$FB,$9A,$57,$D1,$44,$73,$7E,$32,$39,$8F,$5E,$1E,$A7
	.dc.b	$B9,$D7,$F1,$37,$C6,$79,$A3,$A4,$62,$B9,$47,$DD,$8C,$3F,$DB,$21
	.dc.b	$18,$7C,$B1,$F1,$E7,$E3,$12,$3E,$33,$F2,$7E,$DA,$47,$C9,$6C,$FF
	.dc.b	$76,$BF,$41,$33,$F9,$0F,$D4,$AF,$8B,$7D,$6B,$FD,$B4,$D8,$FA,$13
	.dc.b	$C7,$3F,$9B,$58,$BE,$67,$D3,$E2,$FF,$07,$C3,$AD,$CB,$EC,$9C,$C6
	.dc.b	$27,$E9,$C6,$27,$5C,$BF,$AC,$AD,$57,$F8,$73,$87,$EC,$C6,$2F,$84
	.dc.b	$F2,$AE,$17,$CA,$F6,$FF,$B6,$30,$FF,$6B,$BF,$EF,$1A,$C6,$3D,$9F
	.dc.b	$C2,$E5,$FB,$D9,$D4,$62,$B8,$BF,$BB,$AE,$0F,$E4,$A7,$FC,$5D,$9F
	.dc.b	$CC,$3F,$DF,$7C,$1F,$83,$8F,$9B,$18,$D7,$F8,$04,$7F,$BE,$2B,$C6
	.dc.b	$F8,$1E,$87,$E1,$27,$F3,$7B,$70,$71,$68,$C4,$FE,$76,$31,$3E,$29
	.dc.b	$CC,$FE,$36,$3C,$44,$AF,$29,$F9,$5D,$FD,$11,$E1,$6C,$FE,$7B,$0E
	.dc.b	$73,$F9,$A7,$AC,$7A,$0F,$59,$F1,$EB,$E5,$CF,$EC,$0F,$8B,$C1,$CB
	.dc.b	$1F,$E0,$73,$FC,$06,$7F,$12,$71,$A8,$74,$21,$F5,$72,$FA,$7A,$7D
	.dc.b	$98,$D9,$9A,$FA,$B7,$DA,$FE,$4E,$1E,$12,$3E,$0A,$EA,$7A,$CE,$47
	.dc.b	$43,$F9,$6F,$17,$F3,$9F,$4F,$F9,$49,$23,$F9,$3B,$FC,$B3,$A2,$47
	.dc.b	$F2,$A6,$BF,$CF,$E9,$1F,$CF,$DB,$1F,$CF,$F9,$BF,$A3,$4E,$4F,$E9
	.dc.b	$DC,$0D,$77,$5F,$9C,$9C,$C6,$23,$E3,$47,$F9,$87,$49,$1F,$E7,$36
	.dc.b	$BF,$E5,$E6,$BF,$E5,$A6,$BF,$E7,$E6,$F3,$FF,$A0,$9A,$FF,$BF,$09
	.dc.b	$8F,$E9,$9D,$4F,$72,$1D,$CA,$77,$29,$DC,$AF,$27,$81,$0E,$05,$38
	.dc.b	$15,$E0,$E8,$86,$8A,$E9,$C1,$1C,$10,$F6,$CF,$C0,$9F,$74,$62,$7C
	.dc.b	$ED,$7C,$4D,$7B,$B5,$EC,$D3,$B4,$74,$24,$F7,$24,$F6,$8C,$B6,$32
	.dc.b	$F8,$23,$2D,$8C,$BE,$28,$CB,$63,$2F,$92,$32,$D8,$CC,$F9,$9A,$FA
	.dc.b	$49,$3C,$1D,$4F,$94,$87,$8C,$A7,$84,$A7,$72,$BC,$9E,$04,$38,$14
	.dc.b	$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70,$43,$DB,$3F,$02,$7D,$D1,$89
	.dc.b	$F3,$B5,$F1,$35,$EE,$D7,$B3,$4E,$D1,$D0,$93,$DC,$93,$DA,$32,$D8
	.dc.b	$CB,$E0,$8C,$B6,$32,$F8,$A3,$2D,$8C,$BE,$48,$CB,$63,$33,$E6,$6B
	.dc.b	$E9,$24,$F0,$75,$3E,$52,$1E,$33,$F7,$FC,$9F,$E0,$A7,$FC,$46,$7F
	.dc.b	$80,$4A,$53,$2E,$DF,$F1,$4C,$3E,$5F,$FC,$D9,$1C,$23,$15,$CA,$11
	.dc.b	$87,$53,$C1,$F0,$C6,$5B,$19,$7B,$B5,$FF,$26,$C6,$DC,$CF,$78,$C4
	.dc.b	$6A,$78,$3E,$26,$C6,$A3,$13,$C6,$72,$47,$04,$8F,$4B,$5F,$3B,$5F
	.dc.b	$19,$A2,$1A,$29,$A3,$04,$30,$43,$D4,$F5,$69,$C8,$A7,$12,$9C,$4F
	.dc.b	$69,$B2,$9B,$29,$B9,$F8,$84,$9F,$85,$1C,$63,$13,$E3,$8E,$27,$80
	.dc.b	$86,$09,$1C,$1E,$53,$D1,$F1,$46,$5B,$19,$7C,$8D,$7A,$C6,$DC,$CF
	.dc.b	$8E,$31,$1A,$9E,$0F,$89,$B1,$A8,$C4,$F1,$9C,$91,$C1,$23,$D2,$D7
	.dc.b	$CE,$D7,$C6,$68,$86,$8A,$68,$C1,$0C,$10,$F5,$3D,$5A,$72,$29,$C4
	.dc.b	$A7,$13,$DA,$6C,$A6,$CA,$6E,$7E,$21,$27,$E1,$47,$18,$C4,$F8,$E3
	.dc.b	$89,$E0,$21,$82,$47,$07,$94,$F4,$7C,$51,$96,$C6,$5F,$23,$5E,$B1
	.dc.b	$B7,$33,$E3,$8C,$46,$A7,$83,$E2,$6C,$6A,$31,$3C,$67,$24,$70,$48
	.dc.b	$F4,$B5,$F3,$B5,$F1,$9A,$21,$A2,$9F,$EA,$4E,$4F,$F2,$49,$1F,$F5
	.dc.b	$03,$FC,$CB,$A1,$DC,$A4,$E6,$30,$D8,$CD,$7F,$3C,$AE,$31,$88,$FF
	.dc.b	$70,$C6,$3A,$78,$12,$38,$5F,$78,$FF,$3D,$FC,$D5,$9F,$F5,$1B,$5F
	.dc.b	$F5,$13,$5F,$37,$22,$31,$82,$1D,$3F,$CE,$E3,$0D,$8C,$C7,$F5,$23
	.dc.b	$27,$FB,$92,$BF,$D9,$67,$F9,$C0,$FF,$47,$F2,$7C,$27,$FD,$6E,$D7
	.dc.b	$FD,$6C,$D7,$CA,$D8,$F2,$63,$F2,$9F,$F5,$FB,$67,$FD,$7C,$D7,$DA
	.dc.b	$D5,$CD,$FE,$F9,$7C,$87,$FB,$64,$FF,$51,$F8,$3F,$2E,$BF,$D9,$6D
	.dc.b	$8F,$F6,$53,$5F,$92,$D8,$F9,$1F,$57,$71,$34,$90,$D3,$FB,$63,$F3
	.dc.b	$FE,$B3,$FB,$AB,$FE,$DA,$6E,$1F,$ED,$96,$BF,$61,$AF,$B3,$9B,$FD
	.dc.b	$40,$FB,$0F,$A2,$7B,$4F,$F7,$2F,$CA,$79,$49,$5E,$33,$25,$8E,$24
	.dc.b	$3F,$D9,$87,$52,$9D,$4D,$1E,$3E,$7F,$F0,$67,$2F,$E2,$B8,$D2,$EE
	.dc.b	$73,$E2,$3F,$C3,$91,$87,$75,$C5,$E1,$3B,$8C,$4E,$E7,$8C,$E7,$8F
	.dc.b	$F5,$B3,$9A,$DC,$E6,$FC,$73,$C7,$87,$FD,$CD,$87,$FB,$79,$AF,$FB
	.dc.b	$78,$C6,$9F,$EC,$DB,$FF,$65,$9F,$EF,$AE,$06,$BF,$EE,$1D,$7E,$5F
	.dc.b	$57,$DE,$9C,$C2,$31,$3B,$C3,$AE,$DF,$EF,$8C,$7B,$47,$FB,$E2,$FF
	.dc.b	$0C,$7F,$BC,$DF,$C1,$34,$CE,$3D,$F7,$FE,$24,$F8,$78,$7B,$57,$F0
	.dc.b	$06,$D7,$E6,$0F,$21,$F4,$CC,$65,$E9,$E2,$DE,$86,$92,$78,$DF,$AB
	.dc.b	$0F,$8A,$FA,$6B,$F8,$EF,$4F,$C3,$E4,$FF,$7C,$DF,$EE,$8F,$F7,$F9
	.dc.b	$63,$E4,$72,$FF,$BF,$AF,$F8,$3C,$FF,$BE,$4F,$E2,$D1,$98,$FE,$13
	.dc.b	$1F,$EF,$D7,$F2,$58,$78,$9F,$21,$ED,$C3,$F8,$CC,$7F,$9B,$27,$F8
	.dc.b	$CB,$63,$33,$BE,$4F,$96,$FF,$2C,$AF,$B4,$DC,$FF,$98,$C6,$23,$F9
	.dc.b	$7B,$5F,$D0,$F3,$FF,$3E,$9F,$E8,$44,$8F,$E7,$73,$B5,$3F,$21,$CE
	.dc.b	$5F,$D2,$48,$FF,$A1,$38,$9C,$0E,$1F,$85,$18,$6C,$66,$BF,$D1,$33
	.dc.b	$97,$FC,$CF,$A4,$BF,$F3,$3B,$5F,$F3,$CB,$5F,$F2,$FB,$5F,$F3,$AB
	.dc.b	$7A,$7F,$D1,$AD,$E9,$F9,$33,$C8,$EA,$43,$04,$3A,$99,$29,$93,$B9
	.dc.b	$92,$99,$9F,$0B,$96,$B9,$8F,$1B,$D0,$87,$22,$C7,$06,$C7,$07,$44
	.dc.b	$8D,$36,$35,$18,$23,$82,$67,$FC,$66,$7A,$B5,$E4,$D9,$E2,$D9,$E3
	.dc.b	$5F,$36,$30,$47,$04,$AF,$85,$3C,$1B,$3B,$6C,$ED,$F8,$D1,$96,$C6
	.dc.b	$5F,$3B,$5F,$3B,$6B,$CE,$DA,$F0,$4F,$23,$B9,$0C,$10,$EA,$64,$A6
	.dc.b	$4E,$E6,$4A,$66,$7C,$2E,$5A,$E6,$3C,$6F,$42,$1C,$8B,$1C,$1B,$1C
	.dc.b	$1D,$12,$34,$D8,$D4,$60,$8E,$09,$5F,$36,$7A,$B5,$E4,$D9,$E2,$D9
	.dc.b	$E3,$5F,$36,$30,$47,$04,$AF,$85,$3C,$1B,$3B,$6C,$ED,$F8,$D1,$96
	.dc.b	$C6,$5F,$3B,$5F,$3B,$6B,$CE,$DA,$F0,$4F,$23,$B9,$0C,$10,$EA,$64
	.dc.b	$A6,$4E,$E6,$4F,$A7,$CF,$FE,$07,$7F,$C1,$81,$C3,$FE,$01,$32,$67
	.dc.b	$6F,$F3,$83,$FE,$13,$B7,$D3,$7F,$9E,$E3,$71,$8C,$78,$C2,$30,$EA
	.dc.b	$78,$3E,$28,$CB,$63,$2F,$76,$BF,$E2,$68,$DB,$99,$EF,$18,$8D,$4F
	.dc.b	$07,$C4,$D8,$D4,$62,$78,$CE,$48,$E0,$91,$E9,$6B,$E7,$6B,$E3,$34
	.dc.b	$43,$45,$34,$60,$86,$08,$7A,$9E,$AD,$39,$14,$E2,$53,$89,$ED,$36
	.dc.b	$53,$65,$37,$3F,$10,$93,$F0,$A3,$8C,$62,$7C,$71,$C4,$F0,$10,$C1
	.dc.b	$23,$83,$CA,$7A,$3E,$28,$CB,$63,$2F,$91,$AF,$58,$DB,$99,$F1,$C6
	.dc.b	$23,$53,$C1,$F1,$36,$35,$18,$9E,$33,$92,$38,$24,$7A,$5A,$F9,$DA
	.dc.b	$F8,$CD,$10,$D1,$4D,$18,$21,$82,$1E,$A7,$AB,$4E,$45,$38,$94,$E2
	.dc.b	$7B,$4D,$94,$D9,$4D,$CF,$C4,$24,$FC,$28,$E3,$18,$9F,$1C,$71,$3C
	.dc.b	$04,$30,$48,$E0,$F2,$9E,$8F,$8A,$32,$D8,$CB,$E4,$6B,$D6,$36,$E6
	.dc.b	$7C,$71,$88,$D4,$F0,$7C,$4D,$8D,$46,$27,$8C,$E4,$8E,$09,$1E,$96
	.dc.b	$BE,$76,$BE,$33,$44,$34,$53,$47,$EC,$B3,$FF,$29,$74,$7F,$94,$35
	.dc.b	$FF,$2B,$37,$6F,$E6,$35,$FE,$C9,$8C,$74,$C8,$FF,$2F,$57,$F9,$76
	.dc.b	$31,$1F,$EC,$CF,$4D,$8B,$D2,$B5,$FD,$0A,$73,$18,$AF,$F6,$84,$62
	.dc.b	$BF,$DE,$91,$87,$91,$87,$FD,$B2,$B9,$1E,$77,$F3,$88,$C3,$63,$31
	.dc.b	$FD,$38,$C9,$FD,$23,$4F,$F3,$A5,$7F,$BA,$CF,$F7,$41,$FE,$ED,$8E
	.dc.b	$DD,$DC,$C8,$F3,$A4,$77,$FA,$EF,$83,$AF,$FB,$3D,$B5,$FE,$9F,$6D
	.dc.b	$FA,$AD,$18,$8F,$F7,$D9,$FE,$FA,$C3,$CA,$FF,$BF,$4F,$9B,$3C,$29
	.dc.b	$A7,$76,$9D,$CB,$3D,$EF,$E9,$C7,$FA,$C7,$D2,$B3,$E3,$6B,$D9,$B1
	.dc.b	$F3,$DB,$6F,$B2,$FE,$E9,$3F,$EB,$B6,$D7,$B5,$AF,$A1,$AF,$CE,$3D
	.dc.b	$85,$07,$C2,$3F,$17,$F6,$DF,$80,$C7,$FB,$DD,$23,$0F,$FB,$9D,$B3
	.dc.b	$F8,$66,$BC,$CE,$C5,$CB,$E6,$BF,$3F,$83,$FD,$70,$FE,$0E,$32,$F9
	.dc.b	$39,$5C,$AD,$BC,$FE,$0B,$A3,$FC,$38,$DB,$E1,$5B,$51,$97,$FC,$41
	.dc.b	$5C,$AB,$89,$23,$8C,$65,$A7,$28,$FF,$75,$B0,$F9,$38,$7F,$B3,$E3
	.dc.b	$FD,$E1,$18,$C7,$FD,$A9,$18,$CF,$E5,$4F,$EE,$3A,$FE,$6E,$9F,$AE
	.dc.b	$6C,$66,$7E,$73,$4F,$01,$FE,$FE,$3F,$82,$F3,$7F,$AA,$E3,$F6,$DB
	.dc.b	$FF,$04,$C3,$EC,$F0,$7E,$26,$7E,$41,$F8,$92,$DF,$E2,$5A,$FE,$4D
	.dc.b	$61,$CB,$87,$C5,$39,$8C,$4E,$8F,$D1,$3F,$A5,$23,$83,$CF,$7F,$1A
	.dc.b	$7D,$47,$E4,$8A,$7E,$C3,$8B,$F8,$EF,$57,$EB,$C9,$5F,$AF,$7D,$BC
	.dc.b	$9F,$ED,$C8,$FE,$12,$7C,$3C,$BF,$83,$B6,$3F,$DB,$5B,$FD,$7C,$3C
	.dc.b	$64,$3C,$4E,$5F,$E1,$1C,$BD,$6F,$F8,$EE,$5F,$E6,$EC,$3F,$8E,$B6
	.dc.b	$33,$5F,$CA,$F9,$FE,$F3,$6B,$E5,$B7,$4F,$E6,$D1,$FC,$C1,$FE,$88
	.dc.b	$48,$FB,$AD,$7C,$BC,$2E,$06,$BF,$D2,$C9,$3F,$84,$7F,$CB,$99,$48
	.dc.b	$FF,$2E,$35,$FF,$37,$35,$FF,$2A,$35,$FF,$35,$37,$A7,$FD,$04,$DE
	.dc.b	$9F,$F7,$F9,$32,$FF,$7C,$36,$35,$18,$23,$82,$57,$90,$87,$90,$A7
	.dc.b	$90,$A7,$90,$B5,$E4,$6D,$78,$5B,$5E,$06,$D7,$81,$EC,$43,$B1,$4E
	.dc.b	$C5,$AE,$CD,$AE,$CD,$AE,$CD,$AE,$CF,$42,$1D,$0B,$5D,$1B,$5D,$1B
	.dc.b	$5D,$1B,$5D,$1E,$44,$AE,$4D,$AE,$4D,$AE,$4D,$AE,$55,$C0,$95,$C1
	.dc.b	$B5,$C1,$B3,$C2,$74,$49,$D3,$63,$51,$82,$38,$25,$7D,$42,$1F,$24
	.dc.b	$A7,$C0,$29,$E9,$2D,$79,$1B,$5E,$16,$D7,$81,$B5,$E0,$7B,$10,$EC
	.dc.b	$53,$B1,$6B,$B3,$6B,$B3,$6B,$B3,$6B,$B3,$D0,$87,$42,$D7,$46,$D7
	.dc.b	$46,$D7,$46,$D7,$47,$91,$2B,$93,$6B,$93,$6B,$93,$6B,$95,$70,$25
	.dc.b	$70,$6D,$70,$6C,$F0,$9D,$12,$74,$D8,$D4,$60,$8E,$0F,$E1,$DC,$FF
	.dc.b	$DF,$B9,$7F,$A1,$61,$FD,$0E,$30,$E0,$FE,$F2,$57,$2F,$0C,$1E,$47
	.dc.b	$F9,$92,$38,$46,$30,$E5,$08,$C3,$A9,$E0,$F3,$6B,$FD,$56,$36,$E6
	.dc.b	$79,$C6,$23,$53,$C1,$FF,$16,$C6,$5B,$19,$7C,$4D,$8D,$46,$27,$AC
	.dc.b	$E4,$8E,$09,$1E,$96,$BE,$53,$04,$30,$43,$CE,$F0,$69,$B2,$9B,$3D
	.dc.b	$2E,$DA,$6C,$A6,$CF,$51,$92,$99,$9F,$79,$27,$DB,$1B,$8C,$4F,$78
	.dc.b	$E2,$79,$08,$78,$0A,$68,$C1,$0C,$12,$3A,$BD,$E7,$C2,$F9,$5A,$F4
	.dc.b	$8D,$B9,$9F,$24,$62,$35,$3C,$1F,$04,$65,$B1,$97,$C4,$D8,$D4,$62
	.dc.b	$7A,$CE,$48,$E0,$91,$E9,$6B,$E5,$30,$43,$04,$3C,$EF,$06,$9B,$29
	.dc.b	$B3,$D2,$ED,$A6,$CA,$6C,$F5,$19,$29,$99,$F7,$92,$7D,$B1,$B8,$C4
	.dc.b	$F7,$8E,$27,$90,$87,$80,$A6,$8C,$10,$C1,$23,$AB,$DE,$7C,$2F,$95
	.dc.b	$AF,$48,$DB,$99,$F2,$46,$23,$53,$C1,$F0,$46,$5B,$19,$7C,$4D,$8D
	.dc.b	$46,$27,$AC,$E4,$8E,$09,$1E,$96,$BE,$53,$04,$30,$43,$CE,$F0,$69
	.dc.b	$B3,$E5,$F0,$B5,$FF,$28,$63,$FC,$CE,$31,$AF,$FA,$BE,$BF,$CA,$AD
	.dc.b	$7F,$D4,$4D,$5F,$80,$BB,$3F,$DF,$31,$FD,$02,$31,$8E,$DB,$87,$F4
	.dc.b	$16,$CF,$F4,$6F,$9B,$78,$6B,$FF,$7A,$CE,$63,$1D,$BF,$D2,$6B,$75
	.dc.b	$AA,$FF,$48,$46,$27,$FD,$A5,$A7,$FB,$52,$7F,$DC,$FA,$7F,$B5,$3B
	.dc.b	$9F,$AB,$3D,$BF,$F7,$7C,$61,$B1,$9C,$BF,$DD,$A7,$FA,$76,$74,$E2
	.dc.b	$FF,$D1,$FD,$FE,$68,$FE,$C3,$87,$FB,$61,$65,$FD,$4A,$30,$D9,$E1
	.dc.b	$5F,$EE,$BE,$0F,$EB,$B5,$FE,$EA,$8F,$F7,$88,$7A,$6F,$C7,$0F,$F7
	.dc.b	$61,$E0,$BF,$95,$19,$3F,$DA,$B7,$FD,$AE,$7F,$E0,$03,$6F,$FE,$00
	.dc.b	$35,$FA,$5D,$FF,$E9,$59,$F7,$46,$3C,$CF,$3B,$63,$DA,$7F,$B0,$44
	.dc.b	$A7,$71,$87,$8C,$FB,$31,$F7,$13,$6F,$B1,$1E,$4A,$F0,$9D,$30,$F3
	.dc.b	$34,$CE,$1E,$79,$F9,$D5,$C9,$FC,$C9,$F8,$4A,$F9,$B1,$98,$F1,$36
	.dc.b	$FE,$87,$2F,$F8,$63,$37,$37,$F8,$6F,$C1,$6A,$C5,$5A,$71,$7C,$6F
	.dc.b	$85,$6E,$B5,$5F,$E2,$26,$D6,$E3,$8D,$F6,$7A,$DF,$26,$9F,$B4,$7F
	.dc.b	$DE,$58,$7F,$B7,$E3,$07,$FB,$8C,$A7,$EF,$23,$F8,$06,$3F,$ED,$17
	.dc.b	$FD,$C5,$9F,$DF,$7F,$0F,$A7,$C0,$3F,$0E,$5A,$FA,$91,$F3,$B4,$F5
	.dc.b	$C7,$62,$1F,$97,$BE,$CF,$E1,$6F,$F0,$96,$EA,$FF,$1B,$BF,$85,$30
	.dc.b	$6F,$6F,$E0,$91,$A9,$E2,$F9,$B1,$95,$E6,$6D,$FE,$D2,$3F,$DD,$A7
	.dc.b	$E9,$4A,$E7,$AB,$F5,$47,$AE,$7E,$5E,$3E,$89,$F9,$07,$1B,$FD,$81
	.dc.b	$2F,$07,$F0,$5E,$2F,$E1,$17,$FE,$52,$E2,$FE,$0E,$D8,$CC,$EF,$8F
	.dc.b	$F9,$04,$F8,$70,$FB,$07,$C9,$D7,$E6,$B6,$3F,$91,$72,$FC,$48,$FE
	.dc.b	$4D,$7F,$2D,$B9,$6A,$31,$1F,$97,$3F,$9C,$F3,$FF,$2F,$D7,$F9,$A3
	.dc.b	$63,$F9,$D3,$5E,$1C,$FF,$CF,$27,$F1,$39,$1A,$49,$FE,$43,$B2,$75
	.dc.b	$1F,$E6,$FC,$FF,$CB,$C4,$70,$43,$FC,$B6,$E5,$A6,$4F,$F2,$CB,$96
	.dc.b	$99,$3F,$CE,$0E,$5A,$67,$A7,$FD,$15,$19,$6C,$67,$A7,$E9,$B7,$3F
	.dc.b	$A3,$3A,$8F,$19,$0F,$09,$4E,$E5,$78,$3A,$21,$A2,$BA,$70,$47,$04
	.dc.b	$9F,$2C,$75,$6B,$C9,$AF,$16,$C7,$19,$F5,$C6,$08,$E0,$95,$F1,$63
	.dc.b	$2D,$8C,$D7,$C5,$8C,$11,$C1,$0F,$34,$70,$6C,$6D,$A6,$CF,$54,$6D
	.dc.b	$B1,$B6,$9B,$3D,$D1,$B6,$C6,$DA,$6E,$BE,$14,$65,$B1,$9A,$F7,$B6
	.dc.b	$BA,$4E,$A3,$CE,$43,$C6,$53,$B9,$5E,$0E,$88,$68,$AE,$9C,$11,$C1
	.dc.b	$27,$CB,$1D,$5A,$F2,$6B,$C5,$B1,$C6,$7D,$71,$82,$38,$25,$7C,$58
	.dc.b	$CB,$63,$35,$F1,$63,$04,$70,$43,$CD,$1C,$1B,$1B,$69,$B3,$D5,$1B
	.dc.b	$6C,$6D,$A6,$CF,$74,$6D,$B1,$B6,$9B,$AF,$85,$19,$6C,$66,$BD,$ED
	.dc.b	$AE,$93,$A8,$F3,$9F,$7F,$95,$CB,$FE,$65,$84,$63,$9F,$F9,$84,$7F
	.dc.b	$79,$2C,$65,$C8,$A7,$BF,$FE,$63,$8C,$C6,$30,$DC,$6E,$74,$F6,$9D
	.dc.b	$3C,$23,$2F,$F4,$18,$CB,$B8,$CB,$FE,$1B,$8C,$BB,$8C,$BD,$E7,$C3
	.dc.b	$3A,$8E,$13,$A8,$C4,$6C,$F1,$4F,$09,$D1,$1C,$12,$36,$43,$CC,$E5
	.dc.b	$A6,$4F,$33,$96,$99,$3C,$AE,$5A,$64,$F3,$1D,$08,$72,$29,$C0,$A7
	.dc.b	$03,$44,$34,$53,$46,$08,$60,$93,$F2,$09,$3F,$16,$3C,$D1,$89,$F5
	.dc.b	$47,$09,$D3,$E4,$9D,$3C,$23,$2F,$8E,$32,$EE,$32,$F9,$23,$2E,$E3
	.dc.b	$2F,$79,$F0,$CE,$A3,$84,$EA,$31,$1B,$3C,$53,$C2,$74,$47,$04,$8D
	.dc.b	$90,$F3,$39,$69,$93,$CC,$E5,$A6,$4F,$2B,$96,$99,$3C,$C7,$42,$1C
	.dc.b	$8A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18,$24,$FC,$82,$4F,$C5,$8F
	.dc.b	$34,$62,$7D,$51,$C2,$74,$F9,$27,$4F,$08,$CB,$E3,$8C,$BB,$8C,$BE
	.dc.b	$48,$CB,$B8,$CB,$DE,$7C,$33,$A8,$E1,$3A,$8C,$46,$CF,$14,$F0,$9D
	.dc.b	$11,$C1,$23,$64,$3C,$CE,$5A,$64,$F3,$39,$69,$93,$CA,$E5,$FF,$5C
	.dc.b	$73,$B5,$FF,$2B,$63,$FE,$5C,$24,$FF,$96,$B5,$FF,$59,$3F,$E6,$9F
	.dc.b	$01,$E9,$7F,$B6,$9B,$8F,$FB,$8A,$75,$39,$3F,$DB,$F8,$FF,$B7,$6F
	.dc.b	$FD,$A2,$49,$FF,$6A,$10,$FE,$8F,$7D,$3F,$03,$FE,$72,$DC,$33,$81
	.dc.b	$4E,$A3,$0F,$FA,$D6,$BA,$1F,$E8,$1C,$3E,$06,$1F,$CE,$BD,$FF,$F8
	.dc.b	$01,$3F,$D4,$5B,$A9,$A7,$FB,$C3,$AB,$FA,$B6,$5F,$EF,$09,$C1,$1F
	.dc.b	$F7,$9F,$2F,$FC,$00,$93,$49,$1F,$1E,$FF,$DE,$A7,$CF,$3C,$E7,$B6
	.dc.b	$7B,$BF,$EC,$1A,$EC,$D9,$FF,$79,$1C,$58,$7B,$F6,$FF,$7E,$13,$1F
	.dc.b	$A1,$1A,$3C,$27,$FB,$B7,$B3,$FD,$E7,$C5,$FE,$F1,$6F,$07,$FB,$E7
	.dc.b	$93,$44,$AF,$F7,$BC,$65,$A7,$FB,$D8,$FF,$7F,$47,$FB,$29,$87,$FB
	.dc.b	$9B,$6F,$F7,$C3,$FE,$BD,$3F,$D8,$2E,$0F,$F7,$E6,$4D,$3E,$8D,$7C
	.dc.b	$12,$47,$FB,$B7,$4F,$F6,$5B,$CB,$4F,$F6,$D6,$1D,$DB,$1C,$F5,$FF
	.dc.b	$76,$BF,$EF,$56,$CF,$CC,$C3,$EC,$CF,$FB,$D2,$31,$3A,$7E,$B5,$F8
	.dc.b	$1E,$0D,$73,$2C,$3C,$F1,$FE,$13,$DB,$4F,$E4,$67,$47,$F8,$6E,$33
	.dc.b	$5F,$93,$8C,$CF,$CA,$3E,$41,$72,$F9,$7C,$3F,$B4,$FD,$33,$F4,$94
	.dc.b	$69,$1F,$8D,$C1,$FE,$EE,$AF,$A1,$AF,$C8,$C3,$A9,$F2,$67,$51,$9C
	.dc.b	$3F,$80,$F4,$76,$26,$1F,$B6,$3E,$45,$FF,$10,$AF,$F7,$27,$3F,$FB
	.dc.b	$9E,$3F,$2D,$87,$FB,$99,$B1,$FE,$28,$D6,$47,$F8,$89,$AF,$F8,$81
	.dc.b	$AF,$F8,$79,$AF,$F8,$DD,$BA,$7F,$92,$5B,$CB,$FE,$C4,$26,$3F,$ED
	.dc.b	$E6,$C6,$A3,$04,$70,$4A,$F2,$10,$F2,$14,$F2,$14,$F2,$16,$BC,$8D
	.dc.b	$AF,$0B,$6B,$C0,$DA,$F0,$3D,$88,$76,$29,$D8,$B5,$D9,$B5,$D9,$B5
	.dc.b	$D9,$B5,$D9,$E8,$43,$A1,$6B,$A3,$6B,$A3,$6B,$A3,$6B,$A3,$C8,$95
	.dc.b	$C9,$B5,$C9,$B5,$C9,$B5,$CA,$B8,$12,$B8,$36,$B8,$36,$78,$4E,$89
	.dc.b	$3A,$6C,$6A,$30,$47,$04,$AF,$A8,$43,$E4,$94,$F8,$05,$3D,$25,$AF
	.dc.b	$23,$6B,$C2,$DA,$F0,$36,$BC,$0F,$62,$1D,$8A,$76,$2D,$76,$6D,$76
	.dc.b	$6D,$76,$6D,$76,$7A,$10,$E8,$5A,$E8,$DA,$E8,$DA,$E8,$DA,$E8,$F2
	.dc.b	$25,$72,$6D,$72,$6D,$72,$6D,$72,$AE,$04,$AE,$0D,$AE,$0D,$9E,$13
	.dc.b	$A2,$4E,$9B,$1A,$8C,$11,$C1,$2B,$EA,$1F,$77,$9B,$FD,$B8,$FF,$B7
	.dc.b	$B9,$3F,$DC,$91,$88,$FF,$71,$1F,$ED,$3F,$9E,$FF,$64,$12,$FF,$D9
	.dc.b	$AD,$8F,$F7,$0B,$5F,$F6,$F3,$5E,$51,$C0,$8F,$06,$BC,$1A,$F0,$34
	.dc.b	$43,$45,$34,$60,$86,$08,$7F,$BA,$63,$C8,$D8,$EE,$D3,$A1,$4E,$65
	.dc.b	$39,$8C,$53,$D2,$30,$D8,$CC,$F6,$9C,$B6,$33,$1E,$03,$25,$72,$F8
	.dc.b	$8C,$95,$CB,$E4,$6B,$E7,$21,$C0,$D4,$F9,$49,$3E,$36,$C7,$85,$AF
	.dc.b	$76,$BC,$A3,$81,$1E,$0D,$78,$35,$E0,$68,$86,$8A,$68,$C1,$0C,$10
	.dc.b	$F6,$CF,$C0,$7D,$C6,$27,$CE,$D9,$F1,$34,$EE,$53,$B1,$4E,$C7,$42
	.dc.b	$4F,$72,$4F,$68,$CB,$63,$33,$E0,$9C,$B6,$33,$1E,$23,$25,$72,$F9
	.dc.b	$0C,$95,$CB,$E6,$6B,$E9,$21,$C0,$D4,$F9,$49,$3E,$36,$C7,$85,$AF
	.dc.b	$76,$BC,$A3,$81,$1E,$0D,$78,$35,$E0,$68,$86,$8A,$68,$C1,$0C,$10
	.dc.b	$F6,$CF,$C0,$7D,$C6,$27,$CE,$D9,$F1,$34,$EE,$53,$B1,$4F,$DF,$EF
	.dc.b	$FE,$B4,$E4,$FF,$68,$C7,$F9,$11,$AF,$FB,$63,$DB,$FF,$60,$47,$FB
	.dc.b	$06,$3F,$D7,$91,$8A,$46,$14,$EE,$39,$46,$0F,$F6,$79,$C5,$FF,$6F
	.dc.b	$35,$E7,$C3,$FE,$CF,$8E,$11,$A6,$BF,$E5,$DD,$DF,$7B,$E2,$8F,$F2
	.dc.b	$EB,$5D,$35,$63,$FE,$D4,$8C,$2C,$7F,$DB,$D1,$E9,$9D,$C6,$2F,$A5
	.dc.b	$6D,$F0,$98,$8D,$4F,$8F,$C2,$79,$F8,$8F,$8F,$18,$AE,$31,$B9,$D2
	.dc.b	$9D,$61,$F2,$FC,$1F,$F6,$EF,$91,$CD,$B8,$72,$B4,$62,$B8,$74,$AF
	.dc.b	$AE,$7F,$37,$DE,$F7,$37,$1F,$69,$FE,$FF,$8C,$B4,$F7,$0F,$6B,$8E
	.dc.b	$A8,$C1,$A8,$CB,$8C,$FF,$0F,$9F,$A2,$7C,$B8,$F1,$6D,$6E,$BD,$79
	.dc.b	$7A,$63,$11,$F6,$BA,$1B,$FE,$42,$3E,$AC,$63,$87,$F3,$71,$B9,$D3
	.dc.b	$F8,$63,$6F,$E2,$9B,$B7,$E1,$2B,$AB,$F8,$9E,$47,$2F,$FB,$DF,$9B
	.dc.b	$FD,$E9,$7F,$EF,$03,$FD,$E7,$DB,$FE,$FA,$7F,$DE,$95,$C3,$2F,$F7
	.dc.b	$C4,$6C,$9C,$4C,$BD,$7F,$77,$E1,$D0,$D2,$57,$A3,$AB,$FD,$78,$FE
	.dc.b	$8D,$BC,$5F,$84,$F0,$7E,$59,$38,$3F,$4D,$19,$7E,$DC,$7D,$62,$38
	.dc.b	$3F,$06,$3E,$2F,$F0,$6D,$32,$E0,$FB,$38,$48,$EF,$C5,$F6,$63,$18
	.dc.b	$FE,$7E,$7D,$A7,$ED,$F9,$FF,$49,$3F,$B1,$6C,$67,$A3,$F1,$D3,$F8
	.dc.b	$C2,$47,$E5,$1B,$5F,$ED,$E7,$E3,$B4,$FB,$EF,$F0,$3E,$9F,$E1,$DA
	.dc.b	$7E,$FD,$B5,$FC,$58,$46,$5F,$CF,$3F,$9C,$23,$FA,$46,$D7,$FB,$B0
	.dc.b	$FA,$C4,$30,$7F,$92,$B8,$FF,$C9,$04,$7F,$C9,$0D,$74,$60,$86,$08
	.dc.b	$7F,$90,$5E,$0D,$36,$53,$67,$F8,$E1,$DB,$4D,$94,$DF,$5F,$F9,$02
	.dc.b	$36,$D8,$DB,$5D,$F5,$FF,$03,$9D,$B6,$36,$D7,$76,$9D,$B6,$36,$D7
	.dc.b	$71,$A2,$3A,$6B,$A3,$04,$30,$43,$E2,$10,$F8,$05,$AF,$63,$6B,$D0
	.dc.b	$DA,$F2,$36,$BC,$2D,$3C,$05,$3C,$07,$62,$57,$66,$D7,$66,$D7,$66
	.dc.b	$D7,$66,$9D,$8A,$76,$AE,$84,$AE,$8D,$AE,$8D,$AE,$8D,$AE,$8D,$7A
	.dc.b	$57,$22,$57,$26,$D7,$26,$CF,$26,$BC,$A7,$81,$27,$83,$63,$83,$5E
	.dc.b	$11,$A2,$3A,$6B,$A3,$04,$30,$43,$EA,$10,$F9,$25,$AF,$80,$DA,$F4
	.dc.b	$B6,$BC,$8D,$AF,$0B,$4F,$01,$4F,$01,$D8,$95,$D9,$B5,$D9,$B5,$D9
	.dc.b	$B5,$D9,$A7,$62,$9D,$AB,$A1,$2B,$A3,$6B,$A3,$6B,$A3,$6B,$A3,$5E
	.dc.b	$95,$C8,$95,$C9,$B5,$C9,$B3,$C9,$AF,$C2,$E4,$FF,$7B,$3F,$E4,$A6
	.dc.b	$A6,$98,$B7,$CA,$7F,$8D,$67,$FC,$A3,$18,$C3,$FC,$7D,$1F,$ED,$B9
	.dc.b	$D3,$63,$33,$B8,$FF,$22,$35,$FF,$24,$35,$FF,$26,$35,$EF,$3D,$23
	.dc.b	$11,$D4,$E7,$19,$24,$F8,$1B,$1A,$8C,$11,$C1,$0F,$1C,$70,$6C,$6D
	.dc.b	$A6,$CF,$3C,$6D,$B1,$B6,$9B,$3D,$6E,$5A,$64,$F6,$1D,$88,$72,$29
	.dc.b	$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$93,$F3,$89,$3E,$D8,$DC,$62
	.dc.b	$7D,$51,$C2,$74,$D8,$CC,$EE,$3D,$CD,$7D,$2D,$7C,$CD,$7B,$CF,$48
	.dc.b	$C4,$75,$39,$C6,$49,$3E,$06,$C6,$A3,$04,$70,$43,$C7,$1C,$1B,$1B
	.dc.b	$69,$B3,$CF,$1B,$6C,$6D,$A6,$CF,$5B,$96,$99,$3D,$87,$62,$1C,$8A
	.dc.b	$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18,$24,$FC,$E2,$4F,$B6,$37,$18
	.dc.b	$9F,$54,$70,$9D,$36,$33,$3B,$8F,$73,$5F,$4B,$5F,$33,$5E,$F3,$D2
	.dc.b	$31,$1D,$4E,$71,$92,$4F,$81,$B1,$A8,$C1,$1C,$10,$F1,$C7,$06,$C6
	.dc.b	$DA,$6C,$F3,$C6,$DB,$1B,$69,$B3,$D6,$FC,$8E,$4F,$F4,$97,$1F,$FA
	.dc.b	$9A,$7F,$DD,$DC,$BF,$EA,$A9,$CB,$FE,$E8,$E6,$FF,$7C,$4F,$FB,$B1
	.dc.b	$B2,$69,$AC,$5F,$35,$49,$8F,$FB,$F2,$D3,$A8,$C5,$FF,$BD,$E7,$FD
	.dc.b	$6A,$43,$FD,$74,$7F,$C0,$17,$FD,$E9,$D3,$FE,$D0,$24,$FF,$B6,$1F
	.dc.b	$F4,$67,$17,$FB,$BB,$8F,$FD,$DB,$3F,$E8,$D3,$FD,$D5,$7F,$EF,$19
	.dc.b	$FF,$79,$F1,$3E,$13,$FD,$C5,$22,$46,$69,$DD,$FE,$8B,$AD,$D6,$9A
	.dc.b	$FF,$B4,$FB,$3F,$D7,$C4,$7F,$DA,$6F,$57,$FD,$4C,$FA,$2B,$FD,$C3
	.dc.b	$8B,$1F,$F7,$44,$7A,$67,$FD,$F7,$D7,$FE,$E4,$BF,$F6,$51,$0F,$F6
	.dc.b	$F9,$4F,$90,$ED,$C9,$0F,$F6,$EF,$1F,$FB,$65,$FF,$6D,$F6,$7F,$B8
	.dc.b	$5F,$F6,$FF,$36,$A3,$11,$FE,$DD,$9E,$1C,$5F,$E7,$9E,$0F,$C6,$DF
	.dc.b	$FB,$74,$95,$FE,$F2,$6B,$E4,$E4,$FF,$30,$4F,$08,$D1,$F5,$FA,$FE
	.dc.b	$21,$34,$C1,$FE,$FC,$7F,$DD,$44,$8C,$1F,$34,$70,$FD,$9E,$BF,$B1
	.dc.b	$5F,$ED,$D9,$FA,$53,$98,$E0,$D8,$FF,$6D,$75,$7D,$52,$4F,$E2,$DF
	.dc.b	$F6,$9C,$7E,$1F,$07,$77,$FB,$6A,$BF,$DB,$04,$9F,$F6,$B1,$0E,$1B
	.dc.b	$B3,$61,$FE,$DC,$9F,$F1,$0F,$3F,$FB,$52,$FF,$C3,$2D,$8F,$F0,$CB
	.dc.b	$5D,$C7,$FB,$70,$F8,$58,$FE,$DD,$C4,$7E,$AF,$6E,$0F,$EF,$63,$2F
	.dc.b	$AE,$33,$3F,$C2,$3A,$FE,$64,$A3,$0F,$FB,$84,$F5,$E3,$FE,$EC,$7F
	.dc.b	$C3,$7B,$FF,$88,$C8,$E0,$87,$F8,$69,$CB,$4C,$9F,$E1,$A7,$2D,$32
	.dc.b	$7F,$90,$5C,$B4,$CF,$47,$F9,$26,$32,$D8,$CF,$57,$FA,$A0,$99,$7F
	.dc.b	$AA,$1B,$5C,$1B,$3C,$27,$44,$9D,$36,$35,$18,$23,$82,$1E,$F2,$1E
	.dc.b	$D2,$9E,$B2,$D7,$A5,$B5,$E4,$6D,$78,$5B,$5E,$06,$9E,$03,$B1,$0E
	.dc.b	$C5,$AE,$CD,$AE,$CD,$AE,$CD,$AE,$CD,$3B,$1D,$09,$5D,$1B,$5D,$1B
	.dc.b	$5D,$1B,$5D,$1A,$74,$AE,$44,$AE,$4D,$AE,$4D,$AE,$4D,$AE,$55,$C0
	.dc.b	$95,$C1,$B5,$C1,$B3,$C2,$74,$49,$D3,$63,$51,$82,$38,$21,$F5,$08
	.dc.b	$7C,$92,$9F,$00,$B5,$E9,$6D,$79,$1B,$5E,$16,$D7,$81,$A7,$80,$EC
	.dc.b	$43,$B1,$6B,$B3,$6B,$B3,$6B,$B3,$6B,$B3,$4E,$C7,$42,$57,$46,$D7
	.dc.b	$46,$D7,$46,$D7,$46,$9D,$2B,$91,$2B,$93,$6B,$93,$6B,$93,$6B,$95
	.dc.b	$70,$25,$70,$6D,$70,$50,$8F,$E4,$7C,$5F,$ED,$AE,$3F,$F1,$65,$FF
	.dc.b	$B7,$DA,$FF,$B7,$DB,$83,$E5,$BF,$D9,$C4,$C3,$FD,$B2,$DA,$FF,$6C
	.dc.b	$B5,$FF,$6C,$B5,$FF,$6C,$B5,$FF,$73,$B5,$FF,$17,$B6,$7C,$13,$D8
	.dc.b	$93,$D9,$AF,$66,$BD,$9A,$F6,$6B,$D9,$B1,$DA,$3A,$10,$E8,$53,$A1
	.dc.b	$4E,$85,$3A,$15,$E8,$F2,$21,$C8,$A7,$22,$9C,$8A,$F2,$78,$10,$E0
	.dc.b	$53,$81,$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$27,$EA,$12,$7E,$4B,$67
	.dc.b	$E0,$35,$F4,$B5,$F2,$35,$F0,$B5,$F0,$36,$7C,$13,$D8,$93,$D9,$AF
	.dc.b	$66,$BD,$9A,$F6,$6B,$D9,$B1,$DA,$3A,$10,$E8,$53,$A1,$4E,$85,$3A
	.dc.b	$15,$E8,$F2,$21,$C8,$A7,$22,$9C,$8A,$F2,$78,$10,$E0,$53,$81,$5E
	.dc.b	$0E,$88,$68,$AE,$9C,$11,$C1,$27,$EA,$12,$7E,$4B,$67,$E0,$35,$F4
	.dc.b	$B5,$F2,$35,$F0,$B5,$F0,$36,$7C,$13,$D8,$93,$D9,$AF,$66,$BD,$9A
	.dc.b	$F6,$6B,$D9,$B1,$DA,$3A,$10,$E8,$53,$A1,$4E,$87,$FB,$4F,$83,$FC
	.dc.b	$C9,$D7,$74,$FE,$5B,$4F,$9B,$FF,$68,$13,$0F,$F6,$74,$65,$BC,$3F
	.dc.b	$ED,$2B,$FE,$73,$8F,$F9,$A2,$72,$7F,$B4,$B8,$7F,$9E,$F6,$6A,$7F
	.dc.b	$CC,$53,$FE,$C8,$E1,$FF,$35,$4E,$67,$FC,$D0,$DE,$5F,$E8,$B5,$FE
	.dc.b	$CE,$6C,$66,$3F,$D9,$E7,$63,$FD,$84,$3C,$57,$9D,$FE,$CC,$8C,$36
	.dc.b	$33,$1F,$ED,$23,$8B,$8E,$EF,$26,$3E,$26,$CF,$FB,$52,$30,$FF,$A1
	.dc.b	$0F,$F6,$BE,$1E,$8E,$BF,$34,$1A,$47,$FD,$17,$9F,$F6,$2F,$3B,$BB
	.dc.b	$63,$22,$32,$D3,$FD,$27,$C5,$FE,$9F,$8F,$F4,$5E,$08,$C6,$BA,$E2
	.dc.b	$E4,$DA,$FE,$DF,$C9,$FE,$D2,$E5,$FF,$68,$F3,$7F,$B5,$78,$BF,$DF
	.dc.b	$1A,$7F,$75,$C3,$FD,$B2,$43,$FD,$B1,$7F,$ED,$AC,$FC,$19,$3C,$5F
	.dc.b	$1E,$9F,$EB,$36,$C6,$6F,$C9,$1E,$8C,$3F,$C0,$6F,$B8,$9A,$FC,$27
	.dc.b	$E5,$F6,$C8,$FF,$66,$9F,$E0,$A7,$14,$3F,$D9,$51,$FE,$0E,$ED,$FB
	.dc.b	$2F,$FB,$5E,$32,$D7,$FD,$64,$EC,$FC,$1F,$25,$8E,$91,$F8,$A3,$91
	.dc.b	$67,$F3,$CD,$3C,$A7,$FB,$48,$BD,$9F,$84,$23,$F7,$9F,$BA,$47,$07
	.dc.b	$FB,$58,$FF,$65,$CB,$A1,$D1,$FE,$D0,$6C,$E9,$BD,$9F,$AD,$7F,$74
	.dc.b	$DE,$AE,$04,$9E,$BC,$1F,$ED,$2C,$3E,$51,$FE,$D5,$E0,$FF,$76,$5F
	.dc.b	$DA,$24,$7F,$87,$78,$FF,$C4,$04,$70,$43,$FC,$40,$E5,$A6,$4F,$F1
	.dc.b	$13,$96,$99,$3F,$C4,$8E,$5A,$67,$1F,$F1,$9C,$65,$B1,$9E,$BF,$F2
	.dc.b	$7C,$E5,$B1,$9C,$7F,$3C,$F9,$08,$78,$8A,$78,$0B,$5D,$9B,$3D,$1B
	.dc.b	$3D,$1E,$44,$39,$16,$79,$36,$79,$36,$79,$3C,$09,$3C,$1B,$3C,$1B
	.dc.b	$3C,$27,$44,$9D,$36,$35,$18,$23,$82,$57,$C9,$6D,$7C,$98,$C1,$1C
	.dc.b	$10,$F8,$6E,$5A,$64,$F7,$39,$69,$93,$D6,$E5,$A6,$6B,$D5,$19,$6C
	.dc.b	$66,$BD,$73,$96,$C6,$6B,$D6,$F9,$08,$78,$8A,$78,$0B,$5D,$9B,$3D
	.dc.b	$1B,$3D,$1E,$44,$39,$16,$79,$36,$79,$36,$79,$3C,$09,$3C,$1B,$3C
	.dc.b	$1B,$3C,$27,$44,$9D,$36,$35,$18,$23,$82,$57,$C9,$6D,$7C,$98,$C1
	.dc.b	$1C,$10,$F8,$6E,$5A,$64,$F7,$39,$69,$93,$D6,$E5,$A6,$6B,$D5,$19
	.dc.b	$6C,$66,$BD,$73,$96,$C6,$6B,$D6,$F9,$08,$78,$8A,$78,$0B,$5D,$9B
	.dc.b	$3F,$47,$97,$FD,$B5,$CB,$FC,$AA,$7F,$95,$C6,$1C,$1F,$CE,$FE,$67
	.dc.b	$FC,$44,$FF,$59,$B4,$62,$35,$87,$06,$9C,$A7,$71,$96,$C6,$67,$94
	.dc.b	$62,$3C,$8D,$7F,$9F,$B5,$F1,$47,$02,$4F,$56,$C7,$58,$C1,$1C,$10
	.dc.b	$EF,$1C,$1B,$1B,$69,$B3,$C7,$1B,$6C,$6D,$A6,$CF,$3B,$96,$99,$3E
	.dc.b	$39,$D8,$87,$22,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$09,$3F,$35
	.dc.b	$B3,$ED,$77,$19,$24,$66,$37,$3E,$D6,$9E,$D9,$DC,$65,$B1,$99,$E5
	.dc.b	$18,$8F,$6B,$5F,$53,$5F,$1C,$70,$24,$F5,$6C,$75,$8C,$11,$C1,$0E
	.dc.b	$F1,$C1,$B1,$B6,$9B,$3C,$71,$B6,$C6,$DA,$6C,$F3,$B9,$69,$93,$E3
	.dc.b	$9D,$88,$72,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$93,$F3,$5B
	.dc.b	$3E,$D7,$71,$92,$46,$63,$73,$ED,$69,$ED,$9D,$C6,$5B,$19,$9E,$51
	.dc.b	$88,$F6,$B5,$F5,$35,$F1,$C7,$02,$4F,$56,$C7,$58,$C1,$1C,$10,$EF
	.dc.b	$1C,$1B,$1B,$69,$B3,$C7,$1B,$6C,$6D,$A6,$CF,$3B,$96,$99,$3E,$39
	.dc.b	$FB,$CD,$3F,$D2,$F8,$7F,$B8,$CF,$F4,$A9,$FE,$93,$9F,$F6,$D1,$1F
	.dc.b	$F5,$89,$FE,$C4,$EE,$42,$30,$D3,$FA,$04,$A7,$73,$A3,$FD,$8A,$49
	.dc.b	$FF,$43,$B6,$BF,$99,$F9,$1C,$27,$F9,$BB,$63,$31,$FE,$F6,$8C,$47
	.dc.b	$FB,$BB,$BB,$FD,$8E,$49,$FF,$65,$BF,$ED,$42,$3F,$E9,$66,$C7,$FA
	.dc.b	$03,$07,$3F,$FB,$FA,$BF,$DA,$73,$FE,$D2,$23,$FE,$CD,$C3,$E5,$45
	.dc.b	$3F,$B0,$69,$FE,$CC,$7F,$D9,$0D,$7F,$D9,$31,$FE,$CE,$DB,$FD,$A1
	.dc.b	$1E,$07,$FB,$3F,$4F,$FB,$37,$87,$FD,$23,$3F,$ED,$BE,$1F,$F4,$8B
	.dc.b	$E7,$8C,$BE,$9C,$BF,$CF,$F9,$7F,$BA,$B0,$FF,$74,$91,$FF,$59,$92
	.dc.b	$7F,$D3,$FD,$9F,$72,$3F,$D7,$11,$87,$EF,$F2,$FF,$AD,$63,$B4,$69
	.dc.b	$FF,$68,$6B,$F2,$23,$10,$8C,$4F,$FA,$CF,$7F,$F5,$C1,$2B,$FD,$70
	.dc.b	$FE,$33,$3F,$C6,$DF,$86,$7F,$D8,$38,$79,$67,$33,$FE,$DF,$8C,$4F
	.dc.b	$FB,$42,$7F,$26,$47,$F2,$7C,$1D,$5A,$7F,$B0,$33,$C5,$FB,$2B,$AB
	.dc.b	$E9,$8C,$BE,$E6,$BF,$EC,$5E,$A4,$FD,$48,$F8,$7A,$FD,$46,$CE,$9F
	.dc.b	$F0,$0E,$BF,$E0,$92,$38,$21,$FE,$00,$72,$D3,$27,$F8,$01,$CB,$4C
	.dc.b	$9F,$E2,$87,$2D,$33,$C7,$FE,$28,$8C,$B6,$33,$C7,$FE,$A6,$26,$5F
	.dc.b	$EA,$B6,$D7,$06,$CF,$09,$D1,$27,$4D,$8D,$46,$08,$E0,$87,$BC,$87
	.dc.b	$B4,$A7,$AC,$B5,$E9,$6D,$79,$1B,$5E,$16,$D7,$81,$A7,$80,$EC,$43
	.dc.b	$B1,$6B,$B3,$6B,$B3,$6B,$B3,$6B,$B3,$4E,$C7,$42,$57,$46,$D7,$46
	.dc.b	$D7,$46,$D7,$46,$9D,$2B,$91,$2B,$93,$6B,$93,$6B,$93,$6B,$95,$70
	.dc.b	$25,$70,$6D,$70,$6C,$F0,$9D,$12,$74,$D8,$D4,$60,$8E,$08,$7D,$42
	.dc.b	$1F,$24,$A7,$C0,$2D,$7A,$5B,$5E,$46,$D7,$85,$B5,$E0,$69,$E0,$3B
	.dc.b	$10,$EC,$5A,$EC,$DA,$EC,$DA,$EC,$DA,$EC,$D3,$B1,$D0,$95,$D1,$B5
	.dc.b	$D1,$B5,$D1,$B5,$D1,$A7,$4A,$E4,$4A,$E4,$DA,$E4,$DA,$E4,$DA,$E5
	.dc.b	$5C,$09,$5C,$1B,$5C,$1B,$3C,$27,$F2,$DC,$5F,$EB,$F8,$FE,$C7,$C3
	.dc.b	$FE,$07,$7F,$D8,$30,$8C,$47,$FA,$D3,$E6,$FF,$D6,$E4,$AF,$F5,$E3
	.dc.b	$63,$FD,$80,$D7,$84,$68,$8E,$9A,$E8,$C1,$0C,$10,$FF,$61,$C7,$56
	.dc.b	$C7,$26,$9C,$4A,$71,$3F,$DB,$51,$C5,$B1,$C5,$A7,$12,$9C,$46,$29
	.dc.b	$E5,$18,$6C,$66,$7A,$4E,$5B,$19,$8E,$C6,$4A,$E5,$F0,$35,$F5,$90
	.dc.b	$F2,$94,$E4,$70,$9F,$19,$27,$C2,$D8,$EE,$D7,$84,$68,$8E,$9A,$E8
	.dc.b	$C1,$0C,$10,$F2,$4F,$A5,$F3,$90,$F3,$9A,$9F,$13,$67,$BB,$4E,$C5
	.dc.b	$3B,$1C,$E7,$B3,$D0,$C4,$F4,$6C,$F4,$69,$D0,$A7,$43,$91,$27,$A9
	.dc.b	$27,$A4,$65,$B1,$99,$ED,$39,$6C,$66,$3C,$06,$4A,$E5,$F1,$35,$F8
	.dc.b	$04,$3C,$E5,$39,$1C,$27,$C6,$49,$F0,$B6,$3B,$B5,$E1,$1A,$23,$A6
	.dc.b	$BA,$30,$43,$04,$3C,$93,$E9,$7C,$E4,$3C,$E6,$A7,$C4,$D9,$EE,$D3
	.dc.b	$B1,$4E,$C7,$39,$EC,$F4,$31,$3D,$1B,$3D,$1A,$74,$29,$D0,$E4,$49
	.dc.b	$EA,$49,$E9,$19,$6C,$66,$7F,$03,$C5,$FE,$BC,$AF,$F4,$91,$FE,$85
	.dc.b	$CF,$F9,$6B,$FE,$52,$E3,$FF,$5F,$1F,$EB,$CA,$FF,$5B,$EF,$FE,$AA
	.dc.b	$9F,$F3,$04,$7F,$AE,$DF,$F5,$27,$06,$70,$FF,$5E,$4E,$67,$FD,$03
	.dc.b	$39,$35,$10,$D4,$7F,$A0,$DC,$5F,$F9,$BA,$9B,$B1,$72,$30,$FF,$65
	.dc.b	$CE,$63,$11,$ED,$6B,$FE,$6E,$6B,$FE,$91,$C7,$FD,$AF,$1F,$ED,$62
	.dc.b	$63,$FE,$82,$7C,$55,$E1,$AE,$38,$79,$30,$E7,$1F,$E7,$F3,$33,$B7
	.dc.b	$CB,$1A,$9E,$11,$A8,$C4,$EE,$3F,$D7,$F9,$38,$29,$23,$FD,$8F,$B7
	.dc.b	$FB,$06,$F5,$1D,$A7,$A1,$1C,$10,$FC,$03,$97,$EC,$E5,$FE,$B4,$C7
	.dc.b	$EC,$F2,$7F,$BE,$09,$1E,$A7,$D0,$49,$C1,$FE,$FC,$8F,$F5,$0C,$FF
	.dc.b	$AC,$23,$FD,$D1,$18,$33,$A4,$7F,$D4,$A7,$FA,$C3,$8B,$FD,$69,$3F
	.dc.b	$EA,$F8,$FC,$89,$FE,$B3,$24,$7D,$4C,$7F,$D5,$13,$F9,$B8,$FF,$57
	.dc.b	$3F,$EB,$17,$FB,$F6,$32,$3F,$C0,$6D,$7F,$BC,$B5,$FE,$E6,$D7,$FB
	.dc.b	$43,$77,$FE,$D8,$DE,$06,$1F,$4A,$B9,$9D,$88,$60,$87,$63,$25,$32
	.dc.b	$78,$0C,$94,$CD,$78,$9C,$B5,$CC,$F9,$2B,$CD,$5E,$59,$EC,$43,$A1
	.dc.b	$6B,$93,$6B,$93,$D9,$AF,$69,$E8,$F2,$24,$F2,$6C,$F0,$7A,$4F,$09
	.dc.b	$D1,$27,$53,$C2,$75,$18,$9D,$46,$09,$5E,$E2,$57,$9E,$32,$D8,$CB
	.dc.b	$E9,$6B,$E4,$6B,$E1,$6C,$F8,$1B,$5E,$32,$57,$6A,$E8,$77,$21,$82
	.dc.b	$1D,$CC,$94,$C9,$E1,$32,$53,$35,$E3,$72,$D7,$33,$E4,$AF,$3D,$79
	.dc.b	$67,$B1,$0E,$85,$AE,$4D,$AE,$4F,$66,$BD,$A7,$A3,$C8,$93,$C9,$B3
	.dc.b	$C1,$E9,$3C,$27,$44,$9D,$4F,$09,$D4,$62,$75,$18,$25,$7B,$89,$5E
	.dc.b	$78,$CB,$63,$2F,$A5,$AF,$91,$AF,$85,$B3,$E0,$6D,$78,$C9,$5D,$AB
	.dc.b	$A1,$DC,$86,$08,$77,$32,$53,$27,$84,$C9,$4C,$D7,$8D,$CB,$5C,$CF
	.dc.b	$92,$BC,$F5,$E5,$9E,$C4,$3A,$16,$BF,$71,$C3,$FE,$B5,$8F,$ED,$6D
	.dc.b	$7F,$B8,$43,$9F,$F9,$0C,$43,$2E,$2B,$FB,$8C,$ED,$FE,$F0,$62,$35
	.dc.b	$3F,$DE,$27,$4F,$33,$FD,$3F,$1A,$7A,$CF,$27,$84,$65,$DC,$65,$D9
	.dc.b	$CE,$7A,$C7,$84,$D4,$70,$35,$3C,$E7,$71,$DA,$77,$18,$8E,$B1,$CD
	.dc.b	$E2,$EA,$79,$11,$C1,$23,$64,$79,$11,$F0,$19,$2B,$97,$C2,$60,$86
	.dc.b	$08,$7B,$5E,$0D,$36,$53,$73,$ED,$70,$47,$04,$9F,$64,$79,$4E,$2E
	.dc.b	$27,$BC,$ED,$F8,$06,$23,$53,$E2,$9D,$3C,$CF,$04,$69,$EB,$3C,$9E
	.dc.b	$11,$97,$71,$97,$67,$39,$EB,$1E,$13,$51,$C0,$D4,$F3,$9D,$C7,$69
	.dc.b	$DC,$62,$3A,$C7,$37,$8B,$A9,$E4,$47,$04,$8D,$91,$E4,$47,$C0,$64
	.dc.b	$AE,$5F,$09,$82,$18,$21,$ED,$78,$34,$D9,$4D,$CF,$B5,$C1,$1C,$12
	.dc.b	$7D,$91,$E5,$38,$B8,$9E,$F3,$B7,$E0,$18,$8D,$4F,$8A,$74,$F3,$3C
	.dc.b	$11,$A7,$AC,$F2,$78,$46,$5D,$C6,$5D,$9C,$E7,$AC,$78,$4D,$47,$03
	.dc.b	$53,$CE,$77,$1D,$A7,$71,$88,$EB,$1C,$DE,$2E,$A7,$91,$1C,$12,$36
	.dc.b	$47,$91,$1F,$01,$92,$B9,$7C,$26,$08,$60,$87,$B5,$E0,$B4,$FF,$62
	.dc.b	$3F,$EC,$3E,$2F,$F4,$A1,$1F,$F3,$8B,$FE,$C3,$7F,$D5,$B0,$E0,$FF
	.dc.b	$60,$1F,$EC,$5E,$47,$CA,$D2,$61,$FE,$BE,$8C,$BF,$EC,$0A,$FF,$34
	.dc.b	$B4,$FF,$60,$9F,$E8,$B3,$55,$FE,$B6,$7F,$D0,$7C,$9F,$EC,$58,$C6
	.dc.b	$1E,$46,$BF,$EC,$5A,$F3,$53,$B7,$83,$4F,$F4,$6B,$83,$FD,$18,$7F
	.dc.b	$AD,$49,$3F,$EB,$A7,$FD,$93,$AB,$91,$A7,$FB,$4F,$1F,$F4,$14,$E6
	.dc.b	$31,$BF,$FA,$E7,$04,$E5,$F9,$51,$98,$C5,$7F,$A2,$F2,$FF,$67,$76
	.dc.b	$FF,$B1,$C9,$7F,$EC,$C7,$FD,$A7,$3F,$E9,$27,$DA,$4C,$FF,$D5,$2F
	.dc.b	$C0,$E5,$91,$FE,$AC,$6D,$7F,$AD,$A3,$FD,$66,$DE,$3F,$F7,$94,$7F
	.dc.b	$AE,$4C,$C7,$FB,$C7,$7F,$B7,$3F,$DD,$DB,$3F,$EF,$06,$CF,$FB,$D1
	.dc.b	$AF,$D5,$C3,$FD,$8E,$49,$FF,$5E,$9F,$EE,$8A,$FF,$5D,$E3,$FE,$EF
	.dc.b	$7F,$D7,$BC,$5F,$E0,$BD,$3F,$B7,$92,$BF,$BE,$37,$0F,$EF,$0D,$7F
	.dc.b	$C1,$2D,$7F,$C0,$ED,$9E,$93,$C8,$93,$C9,$AF,$26,$BC,$9B,$1C,$A3
	.dc.b	$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9,$C1,$1C,$13,$8B,$FD,$85
	.dc.b	$3E,$46,$CF,$76,$BD,$1A,$F3,$6C,$F3,$95,$7C,$79,$C9,$23,$27,$72
	.dc.b	$D7,$76,$BD,$DA,$F7,$6D,$77,$AE,$D1,$8A,$ED,$18,$24,$77,$6D,$77
	.dc.b	$6B,$DD,$AF,$76,$D7,$79,$E4,$49,$E4,$D7,$93,$5E,$4D,$8E,$51,$C0
	.dc.b	$87,$02,$9C,$0A,$F0,$74,$43,$45,$74,$E0,$8E,$09,$5F,$1A,$7C,$8D
	.dc.b	$9E,$ED,$7A,$35,$E6,$D9,$E7,$3F,$44,$95,$F2,$27,$44,$8C,$9E,$02
	.dc.b	$D7,$81,$AF,$81,$AF,$81,$B5,$E0,$AE,$D1,$8A,$ED,$18,$24,$77,$6D
	.dc.b	$77,$6B,$DD,$AF,$76,$D7,$79,$E4,$49,$E4,$D7,$93,$5E,$4D,$8E,$51
	.dc.b	$C0,$87,$02,$9C,$0A,$F0,$74,$43,$45,$74,$E0,$8E,$09,$5F,$1A,$7C
	.dc.b	$8D,$9F,$F5,$DE,$FF,$D6,$1F,$EB,$6D,$46,$B7,$AC,$3F,$AC,$C6,$DF
	.dc.b	$F5,$C7,$57,$FA,$9C,$97,$FE,$A6,$6C,$7F,$AC,$5B,$1F,$EC,$46,$C7
	.dc.b	$28,$E0,$47,$83,$4E,$05,$78,$1A,$23,$A6,$9A,$70,$47,$04,$3F,$DB
	.dc.b	$B1,$E4,$6C,$77,$69,$D0,$AF,$36,$9C,$C6,$09,$E9,$18,$6C,$66,$7B
	.dc.b	$4E,$5B,$19,$8F,$01,$92,$B9,$8F,$14,$65,$B1,$98,$F2,$35,$F3,$90
	.dc.b	$E0,$6A,$7C,$A4,$9F,$1B,$63,$C2,$D8,$EE,$D8,$E5,$1C,$08,$F0,$69
	.dc.b	$C0,$AF,$03,$44,$74,$D3,$4E,$08,$E0,$87,$B6,$7E,$03,$EE,$31,$3E
	.dc.b	$76,$CF,$89,$A7,$72,$BD,$9A,$76,$3A,$12,$7B,$92,$7B,$46,$5B,$19
	.dc.b	$9F,$04,$E5,$B1,$98,$F1,$19,$2B,$98,$F2,$46,$5B,$19,$8F,$33,$5F
	.dc.b	$49,$0E,$06,$A7,$CA,$49,$F1,$B6,$3C,$2D,$8E,$ED,$8E,$51,$C0,$8F
	.dc.b	$06,$9C,$0A,$F0,$34,$47,$4D,$34,$E0,$8E,$08,$7B,$67,$E0,$3E,$E3
	.dc.b	$13,$E7,$6C,$F8,$9A,$77,$2B,$D9,$A7,$63,$A1,$27,$B9,$27,$B4,$65
	.dc.b	$B1,$99,$FE,$03,$9F,$FA,$D2,$BF,$D3,$5C,$EE,$6F,$E6,$11,$FE,$B9
	.dc.b	$8C,$65,$FE,$B7,$8F,$F3,$77,$89,$FE,$B3,$26,$1F,$EB,$17,$FD,$67
	.dc.b	$3F,$E7,$7E,$6B,$3F,$E7,$74,$FF,$AD,$6F,$C1,$CD,$FD,$07,$0F,$06
	.dc.b	$28,$C6,$4A,$DD,$7F,$47,$8C,$56,$E3,$FD,$54,$43,$FD,$97,$3E,$DE
	.dc.b	$6F,$F5,$7C,$62,$5E,$5F,$6A,$F8,$4D,$8C,$CF,$79,$CB,$FD,$5F,$9F
	.dc.b	$FD,$57,$C5,$FE,$AD,$E1,$FE,$C1,$C9,$FE,$B6,$6C,$67,$7F,$F5,$BD
	.dc.b	$FD,$0C,$5E,$5F,$72,$47,$78,$CB,$FE,$B3,$21,$E9,$E8,$FF,$5B,$BE
	.dc.b	$BE,$8F,$41,$1E,$4F,$CD,$E8,$FF,$52,$C1,$A4,$E3,$FA,$BD,$1F,$EA
	.dc.b	$FE,$5B,$CB,$FE,$AF,$6D,$FE,$0A,$3F,$D6,$EF,$38,$CB,$ED,$9F,$F5
	.dc.b	$6F,$0F,$E3,$5F,$F5,$6F,$3F,$FA,$00,$97,$FE,$74,$6E,$FF,$DD,$09
	.dc.b	$1F,$DD,$1B,$7F,$DE,$DA,$FF,$6A,$6B,$FD,$75,$BC,$FF,$D8,$5B,$5D
	.dc.b	$AB,$A1,$2B,$A3,$5E,$8D,$7A,$36,$BA,$36,$7A,$4F,$22,$1C,$8A,$72
	.dc.b	$2C,$F2,$6C,$72,$78,$10,$E0,$58,$E0,$D8,$E0,$E8,$91,$A6,$C6,$A3
	.dc.b	$04,$70,$49,$FA,$84,$AF,$92,$D3,$E0,$16,$BD,$2D,$7C,$8D,$7C,$2D
	.dc.b	$AF,$03,$6B,$C1,$5D,$89,$1D,$9B,$5D,$9A,$F6,$6B,$D9,$B5,$D9,$B5
	.dc.b	$DA,$BA,$12,$BA,$35,$E8,$D7,$A3,$6B,$A3,$67,$A4,$F2,$21,$C8,$A7
	.dc.b	$22,$CF,$26,$C7,$27,$81,$0E,$05,$8E,$0D,$8E,$0E,$89,$1A,$6C,$6A
	.dc.b	$30,$47,$04,$9F,$A8,$4A,$F9,$2D,$3E,$01,$6B,$D2,$D7,$C8,$D7,$C2
	.dc.b	$DA,$F0,$36,$BC,$15,$D8,$91,$D9,$B5,$D9,$AF,$66,$BD,$9B,$5D,$9B
	.dc.b	$5D,$AB,$A1,$2B,$A3,$5E,$8D,$7A,$36,$BA,$36,$7A,$4F,$22,$1C,$8A
	.dc.b	$7E,$D7,$8D,$C5,$FD,$C2,$3F,$D4,$3C,$5F,$D8,$DA,$7F,$AA,$9C,$60
	.dc.b	$F9,$CF,$ED,$73,$0F,$F5,$6C,$67,$1F,$F5,$79,$FE,$AD,$71,$1B,$3F
	.dc.b	$D1,$33,$CA,$74,$D8,$FF,$5A,$CF,$09,$C9,$D2,$33,$3B,$7C,$2D,$7A
	.dc.b	$C6,$DC,$B5,$C9,$E3,$9E,$CF,$48,$D4,$62,$38,$9B,$9E,$D1,$98,$E7
	.dc.b	$3E,$67,$4F,$29,$D9,$1C,$12,$36,$47,$91,$1F,$21,$92,$B9,$7C,$AD
	.dc.b	$7C,$66,$88,$68,$A6,$8C,$10,$C1,$27,$DC,$49,$F5,$47,$8C,$E1,$19
	.dc.b	$9F,$61,$E3,$71,$1B,$3A,$4F,$29,$D3,$63,$CB,$3C,$27,$27,$48,$CC
	.dc.b	$ED,$F0,$B5,$EB,$1B,$72,$D7,$27,$8E,$7B,$3D,$23,$51,$88,$E2,$6E
	.dc.b	$7B,$46,$63,$9C,$F9,$9D,$3C,$A7,$64,$70,$48,$D9,$1E,$44,$7C,$86
	.dc.b	$4A,$E5,$F2,$B5,$F1,$9A,$21,$A2,$9A,$30,$43,$04,$9F,$71,$27,$D5
	.dc.b	$1E,$33,$84,$66,$7D,$87,$8D,$C4,$6C,$E9,$3C,$A7,$4D,$8F,$2C,$F0
	.dc.b	$9C,$9D,$23,$33,$B7,$C2,$D7,$AC,$6D,$CB,$5C,$9E,$39,$EC,$F4,$8D
	.dc.b	$46,$23,$89,$B9,$ED,$19,$8E,$73,$E6,$74,$F2,$9D,$91,$C1,$23,$64
	.dc.b	$79,$11,$F2,$19,$2B,$97,$CA,$D7,$C6,$68,$86,$8A,$68,$FC,$07,$3F
	.dc.b	$FA,$D5,$FF,$51,$DF,$FA,$C5,$FF,$44,$3F,$EA,$4E,$5F,$F5,$5B,$FC
	.dc.b	$EE,$51,$8A,$FF,$53,$36,$84,$9D,$CA,$B9,$75,$EE,$B4,$D5,$48,$C3
	.dc.b	$6B,$91,$E0,$3F,$9F,$3D,$A7,$CD,$7F,$D1,$63,$13,$FE,$9F,$BF,$75
	.dc.b	$7C,$03,$07,$FA,$BF,$2F,$F5,$5B,$EA,$CF,$FD,$62,$D8,$CB,$FD,$A3
	.dc.b	$0F,$F5,$95,$7F,$AC,$23,$1A,$FC,$26,$9C,$63,$58,$FF,$B3,$36,$FF
	.dc.b	$4C,$F2,$9F,$F6,$21,$F4,$73,$61,$EC,$69,$F0,$5C,$1F,$31,$FF,$53
	.dc.b	$D7,$C7,$9C,$12,$3E,$43,$CC,$8E,$0F,$F6,$23,$E4,$3C,$78,$7F,$74
	.dc.b	$AF,$EB,$E4,$9C,$12,$3F,$B6,$46,$5B,$19,$7F,$BA,$C6,$5B,$19,$7F
	.dc.b	$BD,$C6,$5B,$19,$DB,$FC,$21,$39,$6C,$67,$6F,$F8,$03,$39,$6C,$64
	.dc.b	$F9,$35,$E4,$21,$E2,$29,$E0,$2D,$76,$6C,$F4,$6C,$F4,$79,$10,$E4
	.dc.b	$59,$E4,$D9,$E4,$D9,$E4,$F0,$24,$F0,$6C,$F0,$6C,$F0,$9D,$12,$74
	.dc.b	$D8,$D4,$60,$8E,$09,$5F,$25,$A7,$C9,$70,$47,$04,$AF,$87,$19,$6C
	.dc.b	$65,$F7,$46,$5B,$19,$7D,$71,$96,$C6,$6B,$D5,$39,$6C,$66,$BD,$73
	.dc.b	$96,$C6,$4F,$5D,$79,$08,$78,$8A,$78,$0B,$5D,$9B,$3D,$1B,$3D,$1E
	.dc.b	$44,$39,$16,$79,$36,$79,$36,$79,$3C,$09,$3C,$1B,$3C,$1B,$3C,$27
	.dc.b	$44,$9D,$36,$35,$18,$23,$82,$57,$C9,$69,$F2,$5C,$11,$C1,$2B,$E1
	.dc.b	$C6,$5B,$19,$7D,$D1,$96,$C6,$5F,$5C,$65,$B1,$9A,$F5,$4E,$5B,$19
	.dc.b	$AF,$5C,$E5,$B1,$93,$D7,$5E,$42,$1E,$22,$9E,$02,$D7,$66,$CF,$46
	.dc.b	$CF,$47,$91,$0E,$45,$9F,$DD,$6B,$FE,$B4,$CF,$FD,$64,$D9,$FF,$59
	.dc.b	$10,$FF,$56,$76,$FF,$A4,$89,$7F,$E9,$86,$C7,$FA,$75,$B1,$C0,$D1
	.dc.b	$23,$4D,$34,$E0,$8E,$08,$FF,$AD,$A3,$AB,$63,$93,$4E,$25,$78,$BF
	.dc.b	$EC,$88,$E2,$D8,$E2,$D3,$89,$5E,$2A,$D3,$CA,$30,$D8,$CC,$F4,$9C
	.dc.b	$B6,$33,$1D,$8C,$95,$CC,$78,$1B,$1E,$B2,$1E,$52,$9C,$A3,$84,$F8
	.dc.b	$C8,$F8,$5B,$1D,$DB,$1C,$0D,$12,$34,$D3,$4E,$08,$E0,$8F,$92,$7D
	.dc.b	$2F,$9C,$87,$9C,$D4,$F8,$9B,$3D,$DA,$76,$2B,$D9,$E7,$3D,$9E,$86
	.dc.b	$27,$A3,$67,$A3,$4E,$85,$7A,$3C,$89,$3D,$49,$3D,$23,$2D,$8C,$CF
	.dc.b	$69,$CB,$63,$31,$E0,$32,$57,$31,$E2,$6C,$7C,$02,$1E,$72,$9C,$A3
	.dc.b	$84,$F8,$C8,$F8,$5B,$1D,$DB,$1C,$0D,$12,$34,$D3,$4E,$08,$E0,$8F
	.dc.b	$92,$7D,$2F,$9C,$87,$9C,$D4,$F8,$9B,$3D,$DA,$76,$2B,$D9,$E7,$3D
	.dc.b	$9E,$86,$27,$A3,$67,$A3,$4E,$85,$7A,$3C,$89,$3D,$49,$3D,$23,$2D
	.dc.b	$8C,$CF,$69,$CB,$63,$31,$E0,$32,$7F,$A7,$78,$5A,$FF,$A5,$E3,$FD
	.dc.b	$0A,$7F,$9E,$F5,$FF,$54,$13,$6F,$F4,$A9,$30,$FF,$4A,$CE,$E3,$FD
	.dc.b	$33,$7D,$23,$FD,$08,$7F,$A9,$5C,$38,$A4,$62,$3C,$0D,$78,$5F,$26
	.dc.b	$96,$78,$C6,$63,$FD,$57,$3C,$5F,$F4,$99,$32,$FF,$4E,$B7,$7F,$F4
	.dc.b	$EC,$60,$8F,$FA,$82,$FA,$ED,$DE,$FE,$3B,$70,$FF,$4B,$E1,$EB,$9C
	.dc.b	$B6,$32,$FB,$1A,$FB,$27,$91,$FE,$A1,$DF,$FD,$2B,$1F,$2F,$3F,$F4
	.dc.b	$EE,$67,$FD,$3F,$87,$D8,$AF,$F4,$C5,$7F,$A6,$8F,$EC,$57,$FD,$80
	.dc.b	$95,$82,$47,$F5,$B8,$CB,$63,$2F,$F5,$78,$CB,$63,$2F,$F4,$E8,$CB
	.dc.b	$63,$3A,$7F,$54,$9C,$B6,$33,$AF,$F9,$D8,$8F,$F9,$ED,$B3,$C1,$B1
	.dc.b	$C0,$D1,$23,$4D,$34,$E0,$8E,$09,$5E,$F2,$1E,$D2,$9E,$B2,$D7,$A5
	.dc.b	$B5,$E4,$6C,$F8,$5B,$5E,$06,$C7,$81,$EC,$43,$B1,$6B,$B3,$6B,$B3
	.dc.b	$67,$B3,$6B,$B3,$63,$B3,$D0,$95,$D1,$B5,$D1,$B3,$D1,$B5,$D1,$B1
	.dc.b	$D2,$B9,$12,$B9,$36,$39,$36,$B9,$36,$39,$4F,$02,$3C,$1B,$3C,$1B
	.dc.b	$1C,$0D,$12,$34,$D3,$4E,$08,$E0,$95,$F5,$08,$7C,$92,$9F,$00,$B5
	.dc.b	$E9,$6D,$79,$1B,$3E,$16,$D7,$81,$B1,$E0,$7B,$10,$EC,$5A,$EC,$DA
	.dc.b	$EC,$D9,$EC,$DA,$EC,$D8,$EC,$F4,$25,$74,$6D,$74,$6C,$F4,$6D,$74
	.dc.b	$6C,$74,$AE,$44,$AE,$4D,$8E,$4D,$AE,$4D,$8E,$53,$C0,$8F,$06,$CF
	.dc.b	$06,$C7,$03,$44,$8D,$34,$D3,$83,$F9,$45,$09,$AF,$FA,$45,$FE,$AA
	.dc.b	$D7,$41,$C3,$C3,$5F,$EC,$CF,$F5,$9A,$EA,$53,$B1,$5F,$09,$E1,$D7
	.dc.b	$F9,$D4,$6A,$3F,$B7,$3C,$9A,$E5,$E0,$E5,$AE,$54,$74,$7B,$98,$8E
	.dc.b	$0D,$78,$46,$DF,$31,$23,$C2,$72,$8D,$38,$8D,$3E,$23,$4F,$68,$DC
	.dc.b	$66,$36,$F8,$4D,$BE,$23,$81,$0C,$11,$D9,$1E,$71,$DA,$3A,$46,$0E
	.dc.b	$4E,$CF,$41,$0F,$21,$4E,$25,$8E,$84,$8E,$67,$53,$C8,$58,$EC,$74
	.dc.b	$21,$D0,$B1,$CE,$38,$C7,$77,$AB,$5C,$BC,$1C,$B5,$CB,$D4,$91,$D5
	.dc.b	$F1,$98,$8E,$2D,$78,$C6,$DF,$51,$23,$C4,$72,$8D,$38,$8D,$3E,$33
	.dc.b	$4F,$68,$DC,$66,$36,$F8,$4D,$BE,$23,$81,$0C,$11,$D9,$1E,$71,$DA
	.dc.b	$3A,$46,$0E,$4E,$CF,$41,$0F,$21,$4E,$25,$8E,$84,$8E,$67,$53,$C8
	.dc.b	$58,$EC,$74,$21,$D0,$B1,$CE,$38,$C7,$77,$AB,$5C,$BC,$1C,$B5,$CB
	.dc.b	$D4,$91,$D5,$F1,$98,$8E,$2D,$78,$C6,$DF,$51,$23,$C4,$72,$8D,$38
	.dc.b	$8D,$3E,$33,$4F,$68,$DC,$66,$36,$F8,$4D,$BE,$23,$81,$0C,$11,$D9
	.dc.b	$1E,$71,$DA,$3A,$46,$0E,$4E,$CF,$41,$0F,$21,$4E,$25,$8F,$E0,$39
	.dc.b	$FF,$A9,$36,$FF,$3F,$12,$7F,$D4,$A7,$F3,$BD,$7F,$CE,$9D,$3F,$C9
	.dc.b	$DB,$87,$FA,$75,$B3,$FE,$7B,$9F,$F5,$86,$E7,$FD,$14,$48,$FF,$44
	.dc.b	$3F,$E9,$FC,$FF,$D2,$78,$7F,$59,$BF,$F4,$5C,$62,$91,$8C,$FF,$D3
	.dc.b	$64,$AE,$1A,$7F,$5C,$BF,$F5,$0E,$9F,$EA,$8E,$7F,$F4,$D1,$31,$FF
	.dc.b	$4C,$3F,$EA,$DC,$7F,$D2,$0F,$B7,$43,$48,$FF,$A4,$67,$FD,$3D,$C1
	.dc.b	$FE,$94,$7F,$AD,$E3,$FD,$00,$93,$FD,$9D,$B1,$A8,$C1,$1C,$10,$FE
	.dc.b	$C3,$1C,$1B,$1B,$69,$B3,$FB,$04,$6D,$B1,$B6,$9B,$C3,$FB,$AC,$ED
	.dc.b	$B3,$B6,$CE,$E3,$FD,$69,$3B,$69,$B2,$C6,$D4,$6D,$A6,$CB,$1B,$8D
	.dc.b	$12,$34,$D8,$D4,$60,$8E,$08,$7C,$42,$1F,$00,$B3,$EC,$6C,$7A,$1A
	.dc.b	$F9,$1B,$3E,$16,$C7,$81,$A7,$80,$EC,$49,$EC,$D8,$EC,$D7,$B3,$67
	.dc.b	$B3,$63,$B3,$4E,$D3,$D0,$91,$D1,$AF,$46,$CF,$46,$C7,$46,$CF,$48
	.dc.b	$E4,$43,$91,$67,$93,$4E,$45,$8E,$4F,$02,$47,$06,$9C,$0B,$1C,$23
	.dc.b	$44,$8D,$36,$35,$18,$23,$82,$1F,$50,$87,$C9,$2C,$FC,$06,$C7,$A5
	.dc.b	$AF,$91,$B3,$E1,$6C,$78,$1A,$78,$0E,$C4,$9E,$CD,$8E,$CD,$7B,$36
	.dc.b	$7B,$36,$3B,$34,$ED,$3D,$09,$1D,$1A,$F4,$6C,$F4,$6C,$74,$6C,$F4
	.dc.b	$8E,$44,$39,$16,$79,$34,$E4,$58,$E4,$F0,$24,$70,$69,$C0,$B1,$C2
	.dc.b	$34,$48,$D3,$F1,$31,$FF,$4A,$CF,$FA,$5F,$3F,$F4,$AB,$FC,$FD,$AF
	.dc.b	$FA,$67,$3F,$F4,$EB,$FE,$9E,$C1,$C5,$FE,$79,$26,$3F,$E8,$66,$CF
	.dc.b	$FA,$0D,$B1,$CA,$78,$BB,$24,$ED,$A6,$E7,$8C,$78,$5A,$FF,$A9,$5A
	.dc.b	$FF,$A7,$1B,$3D,$A3,$A1,$27,$A1,$D9,$E8,$D7,$A3,$5E,$8D,$8E,$8F
	.dc.b	$28,$E8,$72,$29,$C8,$A7,$22,$C7,$27,$84,$73,$23,$82,$1C,$DC,$B4
	.dc.b	$C9,$D5,$CB,$4C,$C7,$77,$81,$0D,$15,$D3,$82,$38,$24,$FC,$E2,$4F
	.dc.b	$C5,$6C,$FB,$9B,$1E,$49,$E2,$EC,$93,$B6,$9B,$9E,$31,$EA,$6B,$E7
	.dc.b	$6B,$E6,$6C,$F6,$8E,$84,$9E,$87,$67,$A3,$5E,$8D,$7A,$36,$3A,$3C
	.dc.b	$A3,$A1,$C8,$A7,$22,$9C,$8B,$1C,$9E,$11,$CC,$8E,$08,$73,$72,$D3
	.dc.b	$27,$57,$2D,$33,$1D,$DE,$04,$34,$57,$4E,$08,$E0,$93,$F3,$89,$3F
	.dc.b	$15,$B3,$EE,$6C,$79,$27,$8B,$B2,$4E,$DA,$6E,$78,$C7,$A9,$AF,$9D
	.dc.b	$AF,$99,$B3,$DA,$3A,$12,$7A,$1D,$9E,$8D,$7A,$35,$E8,$D8,$E8,$F2
	.dc.b	$8E,$87,$22,$9C,$8A,$72,$2C,$72,$78,$47,$32,$38,$21,$F5,$77,$FF
	.dc.b	$47,$3F,$E8,$C1,$28,$CC,$7F,$37,$6A,$36,$B7,$75,$7F,$3E,$9C,$C6
	.dc.b	$1F,$E7,$13,$B9,$D5,$77,$AD,$57,$F4,$9C,$7F,$D1,$2F,$F4,$93,$FD
	.dc.b	$1F,$A6,$5A,$67,$7F,$F4,$94,$F0,$8F,$EA,$8D,$D7,$FD,$31,$3F,$E6
	.dc.b	$29,$F5,$61,$E1,$E8,$F1,$E7,$EB,$6E,$46,$9F,$E8,$DA,$E3,$3F,$07
	.dc.b	$7D,$9E,$CD,$BF,$C8,$44,$DB,$FC,$84,$DD,$7F,$AB,$12,$3F,$A5,$B6
	.dc.b	$FF,$A5,$B5,$FE,$96,$D7,$FA,$5B,$72,$FE,$C6,$DA,$ED,$5D,$09,$5D
	.dc.b	$1A,$F4,$6B,$D1,$B5,$D1,$B3,$D2,$79,$10,$E4,$53,$91,$67,$93,$63
	.dc.b	$93,$C0,$87,$02,$C7,$06,$C7,$07,$44,$8D,$36,$35,$18,$23,$82,$4F
	.dc.b	$D4,$25,$7C,$96,$9F,$00,$B5,$E9,$6B,$E4,$6B,$E1,$6D,$78,$1B,$5E
	.dc.b	$0A,$EC,$48,$EC,$DA,$EC,$D7,$B3,$5E,$CD,$AE,$CD,$AE,$D5,$D0,$95
	.dc.b	$D1,$AF,$46,$BD,$1B,$5D,$1B,$3D,$27,$91,$0E,$45,$39,$16,$79,$36
	.dc.b	$39,$3C,$08,$70,$2C,$70,$6C,$70,$74,$48,$D3,$63,$51,$82,$38,$24
	.dc.b	$FD,$42,$57,$C9,$69,$F0,$0B,$5E,$96,$BE,$46,$BE,$16,$D7,$81,$B5
	.dc.b	$E0,$AE,$C4,$8E,$CD,$AE,$CD,$7B,$35,$EC,$DA,$EC,$DA,$ED,$5D,$09
	.dc.b	$5D,$1A,$F4,$6B,$D1,$B5,$D1,$B3,$D2,$79,$10,$E4,$53,$91,$67,$93
	.dc.b	$63,$93,$FC,$03,$4F,$F4,$59,$37,$FF,$44,$C7,$F3,$A3,$FA,$3E,$AC
	.dc.b	$3F,$A9,$1F,$E8,$37,$11,$0F,$E7,$AE,$23,$47,$08,$FF,$49,$46,$5B
	.dc.b	$19,$78,$B5,$FE,$CB,$1B,$72,$77,$33,$1B,$C1,$3B,$9C,$9B,$8C,$CE
	.dc.b	$E7,$51,$89,$D4,$60,$93,$B2,$46,$9A,$7B,$9C,$11,$C1,$1F,$79,$92
	.dc.b	$B9,$7D,$AD,$7D,$2D,$9F,$33,$67,$CB,$3C,$CE,$27,$42,$18,$21,$D0
	.dc.b	$C9,$4C,$CF,$67,$2D,$73,$1E,$09,$F1,$1F,$0A,$37,$19,$6C,$65,$E2
	.dc.b	$D7,$C0,$49,$EC,$D9,$E4,$F0,$3D,$C6,$67,$94,$F0,$24,$F0,$9D,$1C
	.dc.b	$23,$33,$C2,$75,$18,$9D,$46,$09,$3B,$24,$69,$A7,$C2,$70,$47,$04
	.dc.b	$7E,$19,$92,$B9,$7D,$CD,$7D,$2D,$9F,$33,$67,$CB,$3C,$CE,$27,$42
	.dc.b	$18,$21,$D0,$C9,$4C,$CF,$67,$2D,$73,$1E,$09,$F1,$1F,$0A,$37,$19
	.dc.b	$6C,$65,$E2,$D7,$C0,$49,$EC,$D9,$E4,$F0,$3D,$C6,$67,$94,$F0,$24
	.dc.b	$F0,$9D,$1C,$23,$33,$C2,$75,$18,$9D,$46,$09,$3B,$24,$69,$A7,$C2
	.dc.b	$70,$47,$04,$7E,$19,$92,$B9,$7D,$CD,$7D,$2D,$9F,$33,$67,$CB,$3C
	.dc.b	$CE,$27,$42,$18,$21,$D0,$C9,$4C,$CF,$EA,$F3,$FF,$48,$C7,$FA,$33
	.dc.b	$6F,$F3,$A9,$1F,$F4,$A1,$FE,$83,$CF,$FD,$0D,$7F,$E8,$48,$FF,$43
	.dc.b	$9F,$D2,$DC,$12,$36,$48,$FF,$3E,$EC,$D7,$FD,$01,$1F,$E9,$C7,$FA
	.dc.b	$9E,$4A,$FE,$64,$47,$FA,$7B,$67,$FA,$6B,$5F,$E9,$AD,$7F,$A4,$B7
	.dc.b	$2F,$E8,$AD,$C3,$FB,$03,$6B,$C0,$76,$25,$76,$6B,$D9,$AF,$66,$D7
	.dc.b	$66,$D7,$66,$9D,$A7,$A1,$0E,$85,$3A,$16,$7A,$36,$7A,$36,$7A,$3C
	.dc.b	$88,$72,$2C,$F2,$6C,$F2,$6C,$F2,$78,$12,$78,$36,$78,$36,$78,$4E
	.dc.b	$89,$3A,$6C,$6A,$30,$47,$04,$AF,$A8,$47,$E4,$B6,$BE,$03,$5F,$4B
	.dc.b	$5F,$23,$6B,$C2,$DA,$F0,$36,$BC,$07,$62,$57,$66,$BD,$9A,$F6,$6D
	.dc.b	$76,$6D,$76,$69,$DA,$7A,$10,$E8,$53,$A1,$67,$A3,$67,$A3,$67,$A3
	.dc.b	$C8,$87,$22,$CF,$26,$CF,$26,$CF,$27,$81,$27,$83,$67,$83,$67,$84
	.dc.b	$E8,$93,$A6,$C6,$A3,$04,$70,$4A,$FA,$84,$7E,$4B,$6B,$E0,$35,$F4
	.dc.b	$B5,$F2,$36,$BC,$2D,$AF,$03,$6B,$C0,$76,$25,$76,$6B,$D9,$AF,$66
	.dc.b	$D7,$66,$D7,$66,$9D,$A7,$A1,$0E,$85,$3A,$16,$7A,$36,$7A,$36,$7A
	.dc.b	$3C,$88,$72,$2C,$F2,$6C,$F2,$77,$A7,$FA,$1F,$2F,$E8,$31,$FD,$0A
	.dc.b	$30,$E0,$53,$36,$79,$9C,$C6,$1E,$13,$B9,$D3,$63,$33,$B7,$C3,$19
	.dc.b	$6C,$65,$FE,$AB,$19,$6C,$65,$FE,$9E,$74,$9E,$F1,$98,$F2,$C6,$08
	.dc.b	$EC,$87,$81,$CB,$4C,$9E,$07,$2D,$32,$78,$8E,$04,$34,$53,$46,$08
	.dc.b	$60,$87,$A0,$EA,$53,$91,$4E,$25,$38,$CF,$C7,$6C,$FC,$38,$C4,$6C
	.dc.b	$8E,$09,$1B,$24,$FC,$09,$E6,$D7,$8C,$69,$F3,$CF,$29,$D3,$63,$33
	.dc.b	$B7,$D7,$19,$6C,$65,$F6,$C6,$5B,$19,$7C,$67,$49,$EF,$19,$8F,$2C
	.dc.b	$60,$8E,$C8,$78,$1C,$B4,$C9,$E0,$72,$D3,$27,$88,$E0,$43,$45,$34
	.dc.b	$60,$86,$08,$7A,$0E,$A5,$39,$14,$E2,$53,$8C,$FC,$76,$CF,$C3,$8C
	.dc.b	$46,$C8,$E0,$91,$B2,$4F,$C0,$9E,$6D,$78,$C6,$9F,$3C,$F2,$9D,$36
	.dc.b	$33,$3B,$7D,$71,$96,$C6,$5F,$6C,$65,$B1,$97,$C6,$74,$9E,$F1,$98
	.dc.b	$F2,$C6,$08,$EC,$87,$81,$CB,$4C,$9E,$07,$2D,$32,$78,$8E,$04,$34
	.dc.b	$53,$46,$08,$60,$87,$A0,$EA,$53,$91,$4E,$25,$38,$CF,$C7,$6C,$FC
	.dc.b	$38,$C4,$6C,$FC,$15,$34,$60,$9F,$F4,$0C,$E6,$31,$8F,$FA,$06,$7F
	.dc.b	$D1,$04,$3F,$D0,$D9,$FF,$A0,$DA,$7F,$A1,$1F,$E6,$C4,$7F,$98,$35
	.dc.b	$FE,$5C,$D7,$81,$A2,$1A,$29,$A3,$04,$30,$43,$FA,$1B,$D5,$A7,$22
	.dc.b	$9C,$4A,$71,$C3,$FA,$E4,$71,$6C,$71,$6B,$C5,$AF,$1C,$7F,$B4,$CF
	.dc.b	$16,$C7,$16,$BC,$5A,$F1,$CB,$FD,$50,$ED,$AE,$DA,$EE,$7B,$46,$23
	.dc.b	$D6,$4A,$F2,$B6,$BB,$B6,$78,$4E,$89,$3A,$6C,$6A,$30,$47,$04,$AF
	.dc.b	$19,$2B,$D4,$D7,$D4,$D7,$D4,$68,$86,$8A,$68,$C1,$0C,$10,$F6,$3D
	.dc.b	$5A,$72,$29,$C4,$A7,$1A,$F7,$C7,$16,$C7,$16,$BC,$5A,$F1,$AF,$89
	.dc.b	$3C,$5B,$1C,$5A,$F1,$6B,$C6,$BE,$53,$B6,$BB,$6B,$B9,$ED,$18,$8F
	.dc.b	$59,$2B,$CA,$DA,$EE,$D9,$E1,$3A,$24,$E9,$B1,$A8,$C1,$1C,$12,$BC
	.dc.b	$64,$AF,$53,$5F,$53,$5F,$51,$A2,$1A,$29,$A3,$04,$30,$43,$D8,$F5
	.dc.b	$69,$C8,$A7,$12,$9C,$6B,$DF,$1C,$5B,$1C,$5A,$F1,$6B,$C6,$BE,$24
	.dc.b	$F1,$6C,$71,$6B,$C5,$AF,$1A,$F9,$4E,$DA,$ED,$AE,$E7,$B4,$7F,$01
	.dc.b	$C7,$FC,$FB,$3F,$E8,$06,$BF,$E6,$EF,$90,$78,$7F,$C0,$1B,$7F,$E7
	.dc.b	$C6,$CF,$F9,$E1,$AF,$19,$E0,$49,$E0,$D8,$E0,$D7,$84,$68,$8E,$9A
	.dc.b	$E8,$C1,$0C,$10,$FE,$B1,$8A,$BC,$B1,$86,$BF,$E9,$C6,$C6,$EB,$C9
	.dc.b	$3B,$6B,$B6,$C6,$EB,$CB,$3B,$6B,$B6,$C6,$E3,$D0,$6C,$A6,$CA,$ED
	.dc.b	$F6,$19,$2B,$97,$91,$8A,$F5,$92,$BC,$ED,$AF,$1B,$67,$BB,$5E,$53
	.dc.b	$C0,$93,$C1,$B1,$C1,$AF,$08,$D1,$1D,$35,$D1,$82,$18,$21,$EC,$21
	.dc.b	$F0,$4B,$5F,$05,$C1,$1C,$12,$BD,$91,$C1,$AE,$DB,$1B,$AF,$5C,$ED
	.dc.b	$AE,$DB,$1B,$AF,$64,$ED,$AE,$DB,$1B,$8F,$79,$B2,$9B,$2B,$B7,$E1
	.dc.b	$99,$2B,$97,$91,$8A,$F5,$92,$BC,$ED,$AF,$1B,$67,$BB,$5E,$53,$C0
	.dc.b	$93,$C1,$B1,$C1,$AF,$08,$D1,$1D,$35,$D1,$82,$18,$21,$EC,$21,$F0
	.dc.b	$4B,$5F,$05,$C1,$1C,$12,$BD,$91,$C1,$AE,$DB,$1B,$AF,$5C,$ED,$AE
	.dc.b	$DB,$1B,$AF,$64,$ED,$AE,$DB,$1B,$8F,$79,$B2,$9B,$2B,$B7,$E1,$99
	.dc.b	$2B,$97,$91,$8A,$FE,$3B,$7F,$E7,$78,$FF,$3C,$E9,$FE,$7C,$6E,$9F
	.dc.b	$E7,$AC,$3F,$CE,$D5,$BD,$3F,$CF,$95,$FE,$78,$26,$3F,$C9,$5F,$E7
	.dc.b	$64,$3F,$9D,$94,$FE,$84,$53,$F9,$51,$4F,$E5,$45,$7A,$3C,$88,$72
	.dc.b	$29,$C8,$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$2B,$A7
	.dc.b	$04,$70,$4C,$7F,$B4,$C7,$B5,$AF,$99,$AF,$85,$AF,$66,$BD,$5B,$1D
	.dc.b	$72,$FF,$5A,$4F,$36,$BC,$DA,$F3,$6B,$CD,$B1,$CE,$3C,$D1,$89,$EE
	.dc.b	$47,$04,$8D,$90,$F1,$14,$F1,$14,$F1,$14,$F1,$16,$7C,$4F,$22,$1C
	.dc.b	$8A,$72,$29,$C8,$AF,$27,$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9
	.dc.b	$C1,$1C,$12,$7E,$44,$7B,$5A,$F9,$9A,$F8,$5A,$F6,$6B,$D5,$B1,$D6
	.dc.b	$7E,$94,$F3,$6B,$CD,$AF,$36,$BC,$DB,$1C,$E3,$CD,$18,$9E,$E4,$70
	.dc.b	$48,$D9,$0F,$11,$4F,$11,$4F,$11,$4F,$11,$67,$C4,$F2,$21,$C8,$A7
	.dc.b	$22,$9C,$8A,$F2,$78,$10,$E0,$53,$81,$5E,$0E,$88,$68,$AE,$9C,$11
	.dc.b	$C1,$27,$E4,$47,$B5,$AF,$99,$AF,$85,$AF,$66,$BD,$5B,$1F,$91,$BF
	.dc.b	$F3,$BC,$83,$85,$0F,$8F,$D6,$2D,$BF,$83,$CE,$69,$39,$7F,$86,$C6
	.dc.b	$63,$11,$FE,$6C,$AE,$75,$C9,$B1,$C1,$AE,$CE,$B1,$B6,$9B,$29,$B9
	.dc.b	$EE,$FA,$A7,$9B,$4D,$96,$37,$5E,$7A,$ED,$18,$AE,$33,$9A,$F0,$4E
	.dc.b	$5E,$71,$97,$C5,$5C,$88,$F0,$6B,$A3,$04,$30,$43,$CC,$F0,$69,$B2
	.dc.b	$9B,$8F,$13,$82,$38,$25,$7B,$09,$1E,$A6,$D7,$B4,$95,$E1,$AE,$35
	.dc.b	$B2,$4E,$9F,$24,$6A,$31,$1E,$5A,$EB,$5C,$DB,$1C,$1A,$EC,$EB,$1B
	.dc.b	$69,$B2,$9B,$9E,$EF,$C0,$9E,$6D,$36,$58,$DD,$79,$EB,$B4,$62,$B8
	.dc.b	$CE,$6B,$C1,$39,$79,$C6,$5F,$15,$72,$23,$C1,$AE,$8C,$10,$C1,$0F
	.dc.b	$33,$C1,$A6,$CA,$6E,$3C,$4E,$08,$E0,$95,$EC,$24,$7A,$9B,$5E,$D2
	.dc.b	$57,$86,$B8,$D6,$C9,$3A,$7C,$91,$A8,$C4,$79,$6B,$AD,$73,$6C,$70
	.dc.b	$6B,$B3,$AC,$6D,$A6,$CA,$6E,$7B,$BF,$02,$79,$B4,$D9,$63,$75,$E7
	.dc.b	$AE,$D1,$8A,$E3,$39,$AF,$04,$E5,$E7,$19,$7C,$55,$C8,$8F,$06,$BA
	.dc.b	$30,$43,$04,$3C,$CF,$06,$9B,$29,$B8,$F1,$38,$23,$82,$57,$B0,$91
	.dc.b	$EA,$6D,$7B,$49,$5F,$81,$C1,$8B,$7F,$F2,$3E,$29,$CC,$61,$B8,$FF
	.dc.b	$48,$6B,$FE,$88,$D7,$FD,$19,$A7,$FA,$23,$4F,$F4,$56,$79,$D7,$FD
	.dc.b	$1D,$7F,$D5,$E7,$2D,$BE,$35,$FD,$43,$0F,$2D,$F8,$73,$FF,$4C,$5E
	.dc.b	$5A,$FF,$61,$CF,$D1,$18,$8F,$E8,$C7,$F3,$D2,$69,$FD,$45,$A7,$F5
	.dc.b	$12,$9F,$D5,$CA,$72,$78,$11,$E0,$D7,$83,$5E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$C1,$27,$FB,$B3,$E4,$6C,$77,$6C,$74,$6B,$CD,$AF,$3D,$7F,$D1
	.dc.b	$33,$C5,$B1,$C5,$AF,$16,$BC,$4F,$F0,$84,$71,$69,$C4,$A7,$12,$9C
	.dc.b	$4F,$54,$74,$9F,$29,$1C,$12,$38,$C6,$5B,$19,$79,$98,$21,$82,$4F
	.dc.b	$9D,$A7,$9C,$A7,$9C,$A7,$9D,$E0,$47,$83,$5E,$0D,$78,$1A,$21,$A2
	.dc.b	$9A,$30,$43,$04,$9F,$80,$F9,$1B,$1D,$DB,$1D,$1A,$F3,$6B,$CE,$7E
	.dc.b	$44,$F1,$6C,$71,$6B,$C5,$AF,$13,$E6,$C7,$16,$9C,$4A,$71,$29,$C4
	.dc.b	$F5,$47,$49,$F2,$91,$C1,$23,$8C,$65,$B1,$97,$99,$82,$18,$24,$F9
	.dc.b	$DA,$79,$CA,$79,$CA,$79,$DE,$04,$78,$35,$E0,$D7,$81,$A2,$1A,$29
	.dc.b	$A3,$04,$30,$49,$F8,$0F,$91,$B1,$DD,$B1,$D1,$AF,$36,$BC,$E7,$E4
	.dc.b	$4F,$16,$C7,$16,$BC,$5A,$F1,$3E,$6C,$71,$7E,$EE,$7F,$E9,$13,$FD
	.dc.b	$25,$4F,$92,$FF,$38,$92,$7F,$CD,$ED,$CF,$FC,$F2,$DB,$FF,$4A,$B6
	.dc.b	$7F,$D2,$6D,$7F,$A0,$35,$FF,$4F,$B6,$BC,$11,$D8,$95,$D9,$B5,$D9
	.dc.b	$A7,$62,$9D,$8A,$76,$2C,$F6,$9E,$84,$9E,$8D,$9E,$8D,$7A,$35,$E8
	.dc.b	$D9,$E9,$3C,$89,$3C,$9A,$F2,$6B,$C9,$B1,$CA,$38,$10,$E0,$53,$81
	.dc.b	$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$2B,$EA,$12,$3E,$4B,$6B,$E0,$36
	.dc.b	$BD,$2D,$9F,$23,$5F,$0B,$5F,$03,$6B,$C1,$1D,$89,$5D,$9B,$5D,$9A
	.dc.b	$76,$29,$D8,$A7,$62,$CF,$69,$E8,$49,$E8,$D9,$E8,$D7,$A3,$5E,$8D
	.dc.b	$9E,$93,$C8,$93,$C9,$AF,$26,$BC,$9B,$1C,$A3,$81,$0E,$05,$38,$15
	.dc.b	$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$BE,$A1,$23,$E4,$B6,$BE,$03,$6B
	.dc.b	$D2,$D9,$F2,$35,$F0,$B5,$F0,$36,$BC,$11,$D8,$95,$D9,$B5,$D9,$A7
	.dc.b	$62,$9D,$8A,$76,$2C,$F6,$9E,$84,$9E,$8D,$9E,$8D,$7A,$35,$E8,$D9
	.dc.b	$E9,$3C,$89,$3C,$9A,$F2,$6B,$F4,$73,$FF,$3A,$BF,$E6,$56,$A9,$FF
	.dc.b	$33,$ED,$FE,$6D,$3F,$CD,$DC,$52,$7F,$CF,$DA,$23,$11,$FE,$7B,$1B
	.dc.b	$FF,$9F,$F7,$FF,$3C,$C7,$F9,$EF,$7F,$F3,$FA,$DB,$FC,$FE,$D9,$DB
	.dc.b	$67,$79,$7F,$9E,$B8,$FF,$D0,$04,$9F,$1B,$FE,$83,$99,$1E,$66,$BF
	.dc.b	$CE,$4F,$E7,$C4,$D7,$F9,$F3,$4D,$38,$23,$82,$1F,$D0,$4E,$05,$76
	.dc.b	$D3,$67,$F4,$53,$65,$76,$D3,$75,$FD,$59,$DB,$63,$6D,$8D,$EF,$FE
	.dc.b	$67,$23,$FE,$64,$6B,$C9,$B3,$C9,$B1,$C8,$E0,$43,$81,$63,$83,$4E
	.dc.b	$06,$88,$E9,$A6,$9C,$11,$C1,$0F,$90,$43,$E2,$16,$7E,$03,$67,$D2
	.dc.b	$D3,$C8,$53,$C2,$59,$F0,$34,$F0,$1D,$89,$3D,$9B,$3D,$9A,$76,$29
	.dc.b	$D8,$B3,$D9,$A7,$69,$E8,$49,$E8,$D8,$E8,$D7,$A3,$67,$A3,$63,$A4
	.dc.b	$F2,$23,$C9,$AF,$26,$CF,$26,$C7,$23,$81,$0E,$05,$8E,$0D,$38,$1A
	.dc.b	$23,$A6,$9A,$70,$47,$04,$3E,$A1,$0F,$92,$59,$F8,$0D,$9F,$4B,$4F
	.dc.b	$21,$4F,$09,$67,$C0,$D3,$C0,$76,$24,$F6,$6C,$F6,$69,$D8,$A7,$62
	.dc.b	$CF,$66,$9D,$A7,$A1,$27,$A3,$63,$A3,$5E,$8D,$9E,$8D,$8E,$93,$C8
	.dc.b	$8F,$26,$BC,$9B,$3C,$9B,$1C,$8F,$CE,$63,$FE,$77,$C5,$82,$61,$25
	.dc.b	$3A,$8C,$9B,$C1,$F2,$7C,$78,$19,$6A,$73,$49,$CB,$DA,$33,$18,$8E
	.dc.b	$95,$CE,$B9,$36,$38,$35,$D9,$D6,$36,$D3,$65,$37,$3D,$DF,$EC,$93
	.dc.b	$CD,$A6,$CB,$1B,$AF,$3D,$76,$8C,$57,$19,$CD,$78,$27,$2F,$38,$CB
	.dc.b	$E2,$AE,$44,$78,$35,$D1,$82,$18,$21,$E6,$78,$34,$D9,$4D,$C7,$89
	.dc.b	$C1,$1C,$12,$BD,$84,$8F,$53,$6B,$DA,$4A,$F0,$D7,$1A,$D9,$27,$4F
	.dc.b	$92,$35,$18,$8F,$2D,$75,$AE,$6D,$8E,$0D,$76,$75,$8D,$B4,$D9,$4D
	.dc.b	$CF,$77,$E0,$4F,$36,$9B,$2C,$6E,$BC,$F5,$DA,$31,$5C,$67,$35,$E0
	.dc.b	$9C,$BC,$E3,$2F,$8A,$B9,$11,$E0,$D7,$46,$08,$60,$87,$99,$E0,$D3
	.dc.b	$65,$37,$1E,$27,$04,$70,$4A,$F6,$12,$3D,$4D,$AF,$69,$2B,$C3,$5C
	.dc.b	$6B,$64,$9D,$3E,$48,$D4,$62,$3C,$B5,$D6,$B9,$B6,$38,$35,$D9,$D6
	.dc.b	$36,$D3,$65,$37,$3D,$DF,$81,$3C,$DA,$6C,$B1,$BA,$F3,$D7,$68,$C5
	.dc.b	$71,$9C,$D7,$82,$72,$F3,$8C,$BE,$2A,$E4,$47,$83,$5D,$18,$21,$82
	.dc.b	$1E,$67,$83,$4D,$94,$DC,$78,$9C,$11,$C1,$2B,$D8,$48,$F5,$36,$BF
	.dc.b	$92,$70,$7F,$A3,$AB,$FD,$1B,$5F,$E8,$AE,$0F,$F4,$73,$FD,$0F,$87
	.dc.b	$FA,$77,$2F,$F2,$B6,$E5,$99,$4E,$5B,$1F,$E9,$48,$FE,$A3,$1F,$E9
	.dc.b	$BC,$1C,$5C,$23,$44,$3F,$D3,$1C,$3F,$E9,$7D,$FF,$D2,$BB,$FF,$A5
	.dc.b	$09,$BF,$FA,$51,$FF,$4B,$EA,$C7,$AF,$37,$82,$31,$92,$31,$3A,$CB
	.dc.b	$FD,$53,$1B,$D7,$FB,$63,$EB,$26,$BE,$F7,$E4,$DF,$F6,$DA,$FE,$C9
	.dc.b	$7E,$CC,$FF,$D3,$B1,$FE,$BB,$9D,$3F,$D8,$0F,$E9,$A4,$BF,$EC,$2D
	.dc.b	$3F,$B0,$94,$E0,$E8,$8E,$9A,$E8,$C1,$0C,$10,$FE,$CE,$75,$2B,$C9
	.dc.b	$A7,$12,$9C,$71,$FF,$02,$BC,$5B,$1C,$5B,$1C,$5A,$F1,$E0,$FF,$0B
	.dc.b	$CE,$DB,$1B,$6B,$B3,$FC,$3D,$1B,$69,$B2,$9B,$3C,$91,$C9,$F4,$92
	.dc.b	$7C,$6D,$8D,$46,$08,$E0,$91,$D8,$E0,$58,$DB,$63,$6F,$88,$C1,$0C
	.dc.b	$12,$7D,$8D,$3D,$85,$3D,$8E,$88,$E9,$AE,$8C,$10,$C1,$0F,$49,$D4
	.dc.b	$AF,$26,$9C,$4A,$71,$9F,$7B,$C5,$B1,$C5,$B1,$C5,$AF,$19,$F8,$B3
	.dc.b	$B6,$C6,$DA,$EC,$F9,$11,$B6,$9B,$29,$B3,$C9,$1C,$9F,$49,$27,$C6
	.dc.b	$D8,$D4,$60,$8E,$09,$1D,$8E,$05,$8D,$B6,$36,$F8,$8C,$10,$C1,$27
	.dc.b	$D8,$D3,$D8,$53,$D8,$E8,$8E,$9A,$E8,$C1,$0C,$10,$F4,$9D,$4A,$F2
	.dc.b	$69,$C4,$A7,$19,$F7,$BC,$5B,$1C,$5B,$1C,$5A,$F1,$9F,$8B,$3B,$6C
	.dc.b	$6D,$AE,$CF,$91,$1B,$69,$B2,$9F,$EA,$3C,$FF,$D4,$D8,$7F,$50,$84
	.dc.b	$7F,$A9,$23,$16,$F9,$1F,$F4,$11,$2B,$FD,$00,$DE,$1F,$F4,$53,$70
	.dc.b	$FF,$53,$36,$7F,$D4,$8D,$7F,$D5,$CD,$55,$DD,$EA,$4A,$EA,$DA,$EA
	.dc.b	$D3,$A9,$4E,$A5,$3A,$D7,$7A,$ED,$5D,$09,$5D,$1B,$5D,$1A,$F4,$6B
	.dc.b	$D0,$ED,$3D,$27,$91,$23,$93,$5E,$4D,$79,$4F,$49,$E5,$1C,$08,$70
	.dc.b	$29,$C2,$79,$47,$08,$D1,$0D,$1C,$1D,$38,$34,$E0,$95,$F1,$C9,$1E
	.dc.b	$E6,$D7,$9D,$B5,$E2,$6C,$F7,$6B,$D9,$AF,$84,$95,$E0,$7B,$12,$BB
	.dc.b	$36,$BB,$34,$EC,$53,$B1,$4E,$B5,$E0,$AE,$D5,$D0,$95,$D1,$B5,$D1
	.dc.b	$AF,$46,$BD,$0E,$D3,$D2,$79,$12,$39,$35,$E4,$D7,$94,$F4,$9E,$51
	.dc.b	$C0,$87,$02,$9C,$27,$94,$70,$8D,$10,$D1,$C1,$D3,$83,$4E,$09,$5F
	.dc.b	$1C,$91,$EE,$6D,$79,$DB,$5E,$26,$CF,$76,$BD,$9A,$F8,$49,$5E,$07
	.dc.b	$B1,$2B,$B3,$6B,$B3,$4E,$C5,$3B,$14,$EB,$5E,$0A,$ED,$5D,$09,$5D
	.dc.b	$1B,$5D,$1A,$F4,$6B,$D0,$ED,$3D,$27,$91,$23,$93,$5E,$4D,$7F,$D0
	.dc.b	$19,$38,$7F,$D1,$B1,$FE,$68,$8C,$3F,$E9,$13,$FD,$19,$1F,$E8,$B2
	.dc.b	$4E,$04,$71,$BF,$F3,$DD,$1A,$48,$D9,$FE,$73,$70,$F1,$32,$79,$9F
	.dc.b	$F4,$BE,$5F,$E8,$EC,$BF,$D2,$3B,$FF,$A3,$C8,$E0,$94,$E0,$F1,$3F
	.dc.b	$D4,$E5,$3E,$89,$CC,$FC,$43,$D2,$43,$C7,$9F,$FA,$61,$C1,$F1,$CA
	.dc.b	$7F,$A3,$4F,$F4,$C1,$DB,$2F,$F5,$A5,$7F,$AC,$C8,$F4,$DB,$FD,$2D
	.dc.b	$37,$3F,$F4,$AB,$F5,$AB,$FD,$32,$FF,$41,$3F,$AA,$92,$BF,$AA,$B4
	.dc.b	$FE,$AA,$53,$FA,$A9,$4E,$4F,$02,$3C,$1A,$F0,$6B,$C0,$D1,$0D,$14
	.dc.b	$D1,$82,$18,$25,$FF,$71,$7C,$8D,$8E,$ED,$8E,$8D,$79,$B5,$E7,$AF
	.dc.b	$FA,$20,$8F,$FA,$0D,$AF,$66,$CF,$66,$C7,$66,$BD,$9A,$F6,$3A,$10
	.dc.b	$E8,$58,$E8,$D3,$A1,$4E,$85,$3A,$1C,$88,$F2,$69,$C8,$A7,$22,$9C
	.dc.b	$9E,$04,$78,$35,$E0,$D7,$81,$A2,$1A,$29,$A3,$04,$30,$49,$FA,$84
	.dc.b	$9F,$92,$D8,$F8,$0D,$7D,$2D,$9F,$23,$63,$C2,$D7,$C0,$D7,$C1,$3D
	.dc.b	$88,$F6,$6B,$D9,$B3,$D9,$B1,$D9,$AF,$66,$BD,$8E,$84,$3A,$16,$3A
	.dc.b	$34,$E8,$53,$A1,$4E,$87,$22,$3C,$9A,$72,$29,$C8,$A7,$27,$81,$1E
	.dc.b	$0D,$78,$35,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$7E,$A1,$27,$E4,$B6
	.dc.b	$3E,$03,$5F,$4B,$67,$C8,$D8,$F0,$B5,$F0,$35,$F0,$E7,$FE,$8E,$7F
	.dc.b	$A3,$E7,$FC,$A9,$FF,$3F,$FC,$A7,$F5,$6D,$58,$7F,$3C,$8C,$36,$7F
	.dc.b	$AC,$36,$77,$1F,$D6,$30,$57,$08,$C3,$5F,$EC,$71,$A7,$99,$A2,$9A
	.dc.b	$8E,$0F,$98,$C9,$67,$C2,$EC,$CC,$F5,$8C,$C6,$EB,$D0,$47,$C1,$1A
	.dc.b	$9E,$13,$99,$E9,$39,$AF,$1D,$6A,$78,$D6,$DE,$C6,$08,$72,$75,$1C
	.dc.b	$63,$06,$8E,$26,$08,$60,$8F,$16,$CF,$90,$8F,$C2,$6D,$7A,$63,$04
	.dc.b	$70,$4A,$F2,$CF,$06,$CE,$DB,$3B,$8F,$45,$78,$C9,$5C,$63,$2D,$7D
	.dc.b	$71,$A7,$A1,$A2,$9A,$8E,$0F,$B8,$C9,$67,$C4,$EC,$CC,$F5,$8C,$C6
	.dc.b	$EB,$D2,$47,$C1,$1A,$9E,$13,$99,$E9,$39,$AF,$1D,$6A,$78,$D6,$DE
	.dc.b	$C6,$08,$72,$75,$1C,$63,$06,$8E,$26,$08,$60,$8F,$16,$CF,$90,$8F
	.dc.b	$C2,$6D,$7A,$63,$04,$70,$4A,$F2,$CF,$06,$CE,$DB,$3B,$8F,$45,$78
	.dc.b	$C9,$5C,$63,$2D,$7D,$71,$A7,$A1,$A2,$9A,$8E,$0F,$B8,$C9,$67,$C4
	.dc.b	$EC,$CC,$F5,$8C,$C6,$EB,$D2,$47,$C1,$1A,$9E,$13,$99,$E9,$39,$AF
	.dc.b	$1D,$6A,$78,$D6,$DE,$C6,$08,$72,$75,$1C,$63,$06,$8E,$26,$08,$60
	.dc.b	$8F,$16,$CF,$90,$8F,$C2,$7F,$2F,$A3,$4F,$F4,$CC,$FF,$A6,$F6,$FF
	.dc.b	$4E,$BF,$CD,$27,$51,$FE,$76,$3F,$D3,$71,$FC,$F7,$87,$FD,$37,$18
	.dc.b	$8F,$E7,$7C,$78,$CB,$FD,$06,$D8,$FE,$88,$D7,$67,$FA,$B6,$3F,$D1
	.dc.b	$35,$FC,$B0,$91,$FC,$E9,$FE,$5F,$7E,$48,$FF,$58,$1E,$92,$9F,$E9
	.dc.b	$9D,$FF,$D3,$BC,$1F,$E9,$8D,$FF,$D4,$93,$FE,$9F,$DB,$FD,$5B,$18
	.dc.b	$9D,$6F,$FE,$A6,$AD,$1F,$0F,$5F,$F4,$7E,$EC,$BF,$D2,$73,$FE,$C3
	.dc.b	$CB,$FD,$23,$2C,$FF,$D7,$3A,$7B,$2B,$FD,$82,$D7,$FA,$74,$69,$F8
	.dc.b	$A7,$FA,$8C,$B5,$FD,$4F,$5F,$F6,$4B,$63,$47,$F6,$A9,$CD,$7F,$AA
	.dc.b	$89,$1F,$5C,$D9,$FE,$A7,$3F,$D5,$53,$E8,$C3,$D7,$1F,$DA,$4F,$E8
	.dc.b	$C4,$9F,$F0,$03,$4F,$EE,$C5,$38,$3A,$23,$A6,$BA,$30,$43,$04,$3F
	.dc.b	$BF,$1D,$4A,$F2,$69,$C4,$A7,$1C,$3F,$C1,$8F,$16,$C7,$16,$C7,$16
	.dc.b	$BC,$77,$FA,$8D,$3E,$C9,$B7,$2F,$80,$91,$C9,$B1,$A8,$C1,$1C,$12
	.dc.b	$3A,$9D,$CF,$43,$96,$9A,$3A,$96,$3A,$36,$38,$4F,$BD,$B1,$E9,$34
	.dc.b	$43,$A9,$23,$83,$96,$B9,$39,$14,$F4,$3A,$23,$A6,$BA,$30,$43,$04
	.dc.b	$3D,$A7,$52,$BC,$9A,$71,$29,$C6,$7E,$1B,$C5,$B1,$C5,$B1,$C5,$AF
	.dc.b	$19,$F8,$ED,$3E,$11,$B7,$2F,$80,$91,$C9,$B1,$A8,$C1,$1C,$12,$3A
	.dc.b	$9D,$CF,$43,$96,$9A,$3A,$96,$3A,$36,$38,$4F,$BD,$B1,$E9,$34,$43
	.dc.b	$A9,$23,$83,$96,$B9,$39,$14,$F4,$3A,$23,$A6,$BA,$30,$43,$04,$3D
	.dc.b	$A7,$52,$BC,$9A,$71,$29,$C6,$7E,$1B,$C5,$B1,$C5,$B1,$C5,$AF,$19
	.dc.b	$F8,$ED,$3E,$11,$B7,$2F,$80,$FE,$05,$AF,$FA,$B5,$1F,$D3,$B4,$FF
	.dc.b	$57,$53,$6F,$F4,$49,$27,$FD,$2E,$D3,$FD,$24,$5C,$39,$C7,$FA,$A2
	.dc.b	$30,$FF,$AB,$DA,$E4,$E6,$47,$9B,$63,$9A,$9F,$23,$93,$A1,$4E,$93
	.dc.b	$C8,$93,$CA,$78,$C6,$27,$8B,$5E,$33,$B8,$F3,$C7,$85,$B3,$CA,$75
	.dc.b	$3E,$13,$45,$8D,$3D,$A3,$73,$DE,$36,$D9,$DC,$78,$8E,$04,$30,$43
	.dc.b	$8C,$F4,$23,$B3,$25,$8E,$AE,$0E,$4E,$09,$3E,$E2,$47,$AD,$A7,$9C
	.dc.b	$B3,$E0,$8E,$71,$87,$C0,$D7,$27,$32,$3C,$DB,$1C,$DF,$31,$27,$CA
	.dc.b	$E8,$EA,$53,$AC,$F3,$24,$F3,$9E,$31,$89,$E2,$D7,$8C,$EE,$3D,$11
	.dc.b	$E1,$6C,$F2,$9D,$4F,$84,$D1,$63,$4F,$68,$DC,$F7,$8D,$B6,$77,$1E
	.dc.b	$23,$81,$0C,$10,$E3,$3D,$08,$EC,$C9,$63,$AB,$83,$93,$82,$4F,$B8
	.dc.b	$91,$EB,$69,$E7,$2C,$F8,$23,$9C,$61,$F0,$35,$C9,$CC,$8F,$36,$C7
	.dc.b	$37,$CC,$49,$F2,$BA,$3A,$94,$EB,$3C,$C9,$3C,$E7,$8C,$62,$78,$B5
	.dc.b	$E3,$3B,$8F,$44,$78,$5B,$3C,$A7,$53,$E1,$34,$58,$D3,$DA,$37,$3D
	.dc.b	$E3,$F3,$98,$FF,$A1,$DF,$F4,$31,$0F,$F3,$E6,$FF,$C8,$97,$7B,$E3
	.dc.b	$BF,$D2,$D1,$86,$C6,$63,$FD,$30,$7F,$34,$71,$2C,$3F,$96,$FE,$0D
	.dc.b	$FE,$AF,$AD,$9F,$D1,$1C,$4E,$E3,$F9,$EC,$62,$78,$F0,$F8,$B3,$FF
	.dc.b	$48,$2E,$8F,$44,$61,$B9,$FF,$AF,$5B,$3B,$C3,$F9,$B4,$FF,$B0,$79
	.dc.b	$F2,$48,$CC,$A3,$8F,$85,$27,$2D,$7F,$AE,$71,$7F,$AB,$B7,$B1,$FE
	.dc.b	$A8,$1C,$3F,$EA,$D8,$CF,$2F,$FA,$BA,$B8,$57,$FA,$B7,$C3,$F3,$91
	.dc.b	$F1,$BE,$AE,$99,$1E,$66,$BF,$D9,$36,$F8,$BD,$5F,$00,$91,$EC,$7E
	.dc.b	$F7,$0F,$D3,$C3,$E8,$4F,$D3,$DB,$FD,$52,$FD,$26,$B9,$3F,$D5,$3C
	.dc.b	$7F,$81,$9E,$31,$F8,$53,$3C,$9C,$1C,$46,$B1,$FB,$2F,$FA,$BE,$FF
	.dc.b	$D9,$7C,$3F,$EB,$0A,$FE,$BA,$7F,$76,$27,$1F,$F6,$26,$9F,$DD,$0A
	.dc.b	$7F,$7F,$29,$C9,$E0,$47,$83,$5E,$0D,$78,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$BF,$F0,$5B,$E4,$6C,$77,$6C,$74,$6B,$CD,$AF,$3D,$FF,$D5,$44
	.dc.b	$7F,$D4,$0D,$7B,$36,$7B,$36,$3B,$35,$EC,$D7,$B1,$D0,$87,$42,$C7
	.dc.b	$46,$9D,$0A,$74,$29,$D0,$E4,$47,$93,$4E,$45,$39,$14,$E4,$F0,$23
	.dc.b	$C1,$AF,$06,$BC,$0D,$10,$D1,$4D,$18,$21,$82,$4F,$D4,$24,$FC,$96
	.dc.b	$C7,$C0,$6B,$E9,$6C,$F9,$1B,$1E,$16,$BE,$06,$BE,$09,$EC,$47,$B3
	.dc.b	$5E,$CD,$9E,$CD,$8E,$CD,$7B,$35,$EC,$74,$21,$D0,$B1,$D1,$A7,$42
	.dc.b	$9D,$0A,$74,$39,$11,$E4,$D3,$91,$4E,$45,$39,$3C,$08,$F0,$6B,$C1
	.dc.b	$AF,$03,$44,$34,$53,$46,$08,$60,$93,$F5,$09,$3F,$25,$B1,$F0,$1A
	.dc.b	$FA,$5B,$3E,$46,$C7,$85,$AF,$F0,$BE,$0F,$F5,$53,$FD,$30,$FF,$47
	.dc.b	$7A,$BF,$EA,$CB,$FE,$58,$DB,$CC,$FF,$AB,$5B,$3A,$8C,$1F,$D1,$E7
	.dc.b	$83,$4C,$9F,$EB,$8B,$46,$A3,$0D,$7F,$D5,$0D,$8D,$C7,$23,$25,$73
	.dc.b	$3E,$32,$57,$98,$93,$E3,$8C,$B6,$32,$F8,$8E,$95,$E1,$23,$82,$4F
	.dc.b	$27,$83,$5F,$2B,$6B,$CB,$5D,$A7,$2D,$8C,$C7,$8C,$87,$48,$D5,$78
	.dc.b	$AB,$A3,$63,$33,$E4,$6C,$EA,$30,$74,$3C,$25,$AF,$0B,$DC,$C5,$75
	.dc.b	$69,$C8,$D1,$0E,$65,$8E,$6E,$08,$E0,$91,$D6,$38,$35,$DB,$63,$71
	.dc.b	$E1,$32,$57,$33,$EF,$25,$7C,$22,$4F,$AA,$32,$D8,$CB,$E2,$3A,$57
	.dc.b	$84,$8E,$09,$3C,$9E,$0D,$7C,$AD,$AF,$2D,$76,$9C,$B6,$33,$1E,$32
	.dc.b	$1D,$23,$55,$E2,$AE,$8D,$8C,$CF,$91,$B3,$A8,$C1,$D0,$F0,$96,$BC
	.dc.b	$2F,$73,$15,$D5,$A7,$23,$44,$39,$96,$39,$B8,$23,$82,$47,$58,$E0
	.dc.b	$D7,$6D,$8D,$C7,$84,$C9,$5C,$CF,$BC,$95,$F0,$89,$3E,$A8,$CB,$63
	.dc.b	$2F,$88,$E9,$5E,$12,$38,$24,$F2,$78,$35,$F2,$B6,$BC,$B5,$DA,$72
	.dc.b	$D8,$CC,$78,$C8,$74,$8F,$F4,$4F,$2F,$FA,$40,$91,$80,$7F,$9D,$FA
	.dc.b	$3F,$D3,$84,$7F,$D4,$A2,$91,$FE,$B2,$DF,$FD,$49,$3F,$E7,$53,$6E
	.dc.b	$1F,$F3,$B5,$7F,$9B,$FA,$FF,$D5,$04,$8E,$B6,$8F,$07,$3F,$F9,$D7
	.dc.b	$0F,$2B,$6B,$FD,$5D,$8F,$A1,$FE,$89,$3A,$8C,$47,$9E,$3D,$BC,$FC
	.dc.b	$49,$3A,$C3,$D7,$5F,$EA,$D8,$FF,$57,$CF,$FA,$C6,$3F,$D3,$5B,$7F
	.dc.b	$AD,$5C,$9D,$B6,$FF,$55,$3F,$29,$AF,$FA,$BB,$A1,$C3,$FE,$AC,$8C
	.dc.b	$37,$1F,$F6,$4F,$2F,$FA,$A2,$7F,$AF,$60,$E2,$4E,$67,$D8,$63,$3F
	.dc.b	$56,$9F,$6A,$7E,$31,$23,$32,$E5,$F9,$44,$8C,$BF,$56,$3C,$F1,$F1
	.dc.b	$77,$FF,$54,$65,$F5,$67,$27,$FA,$F6,$3F,$D5,$9C,$3F,$EB,$9A,$FF
	.dc.b	$72,$CE,$8F,$F5,$B6,$9F,$EB,$49,$D4,$7E,$3C,$D1,$EC,$8F,$BC,$49
	.dc.b	$FF,$70,$BF,$EB,$53,$FD,$D2,$7F,$BB,$B8,$FF,$05,$1F,$E0,$E9,$FF
	.dc.b	$06,$11,$C1,$1F,$F0,$61,$92,$B9,$7F,$BD,$19,$2B,$97,$FC,$12,$64
	.dc.b	$AE,$70,$FF,$07,$C6,$5B,$19,$9F,$3C,$F8,$08,$76,$29,$D0,$B3,$C9
	.dc.b	$B1,$C9,$E0,$43,$81,$63,$83,$63,$83,$A2,$46,$9B,$1A,$8C,$11,$C1
	.dc.b	$23,$F8,$4C,$FB,$5A,$79,$8A,$78,$4A,$76,$2C,$F5,$6C,$F5,$7E,$53
	.dc.b	$67,$B4,$60,$D9,$0C,$11,$D9,$1F,$21,$92,$B9,$7C,$C6,$4A,$E5,$F4
	.dc.b	$19,$2B,$99,$F5,$46,$5B,$19,$9F,$54,$F8,$08,$76,$29,$D0,$B3,$C9
	.dc.b	$B1,$C9,$E0,$43,$81,$63,$83,$63,$83,$A2,$46,$9B,$1A,$8C,$11,$C1
	.dc.b	$23,$E8,$4F,$B5,$A7,$98,$A7,$84,$A7,$62,$CF,$56,$CF,$57,$E5,$36
	.dc.b	$7B,$46,$0D,$90,$C1,$1D,$91,$F2,$19,$2B,$97,$CC,$64,$AE,$5F,$41
	.dc.b	$92,$B9,$9F,$54,$65,$B1,$99,$F5,$4F,$80,$87,$62,$9D,$0B,$3C,$9B
	.dc.b	$1C,$9E,$04,$38,$16,$38,$36,$3F,$03,$BF,$FA,$EA,$3F,$AC,$6F,$FD
	.dc.b	$59,$FE,$D2,$7F,$7F,$F9,$7F,$F5,$E3,$FD,$9D,$AF,$F7,$1B,$46,$23
	.dc.b	$FD,$6C,$D7,$90,$E4,$EB,$3C,$23,$07,$52,$9D,$67,$94,$6E,$78,$C6
	.dc.b	$09,$3C,$0F,$09,$4F,$0C,$70,$71,$3C,$DF,$04,$64,$8E,$08,$78,$CE
	.dc.b	$33,$E0,$8D,$C6,$5A,$F8,$9A,$F8,$8E,$66,$4A,$64,$E4,$62,$7C,$CE
	.dc.b	$08,$71,$29,$C4,$C1,$0C,$11,$E6,$E5,$AE,$67,$D8,$F8,$5A,$F6,$8D
	.dc.b	$92,$33,$1C,$DA,$F3,$3E,$39,$27,$E1,$CF,$18,$C1,$DC,$A7,$79,$E7
	.dc.b	$1B,$9E,$31,$82,$4F,$03,$C4,$53,$C5,$1C,$1C,$4F,$37,$C1,$19,$23
	.dc.b	$82,$1E,$33,$8C,$F8,$23,$71,$96,$BE,$26,$BE,$23,$99,$92,$99,$39
	.dc.b	$18,$9F,$33,$82,$1C,$4A,$71,$30,$43,$04,$79,$B9,$6B,$99,$F6,$3E
	.dc.b	$16,$BD,$A3,$64,$8C,$C7,$36,$BC,$CF,$8E,$49,$F8,$73,$C6,$30,$77
	.dc.b	$29,$DE,$79,$C6,$E7,$8C,$60,$93,$C0,$F1,$14,$F1,$47,$07,$13,$CD
	.dc.b	$F0,$46,$48,$E0,$87,$8C,$E3,$3E,$08,$DC,$65,$AF,$89,$AF,$88,$E6
	.dc.b	$64,$A6,$4E,$46,$27,$CC,$E0,$FA,$5C,$3F,$E8,$5E,$4F,$F5,$DC,$7F
	.dc.b	$A5,$B8,$91,$88,$FF,$52,$9F,$EA,$EE,$2F,$F5,$E4,$7F,$AA,$C6,$BF
	.dc.b	$EA,$3E,$BF,$F6,$1C,$62,$BF,$9F,$46,$27,$51,$FE,$B8,$8F,$F5,$D4
	.dc.b	$FF,$A2,$49,$3F,$E8,$99,$FF,$5E,$E1,$FE,$9B,$E5,$FE,$82,$D8,$FF
	.dc.b	$59,$35,$F5,$47,$F4,$6B,$FF,$5E,$47,$88,$FF,$60,$96,$3C,$CF,$FB
	.dc.b	$6B,$24,$F0,$8F,$F5,$5E,$5F,$D0,$2F,$FD,$59,$9F,$8A,$30,$47,$04
	.dc.b	$9F,$2C,$7F,$4B,$DB,$FD,$7A,$4D,$BF,$B4,$74,$FF,$58,$8C,$70,$FF
	.dc.b	$BB,$23,$FA,$DE,$6C,$BF,$D7,$98,$F8,$48,$76,$29,$F6,$CA,$7A,$8F
	.dc.b	$88,$7F,$AE,$F5,$FA,$7C,$9F,$0A,$3F,$DE,$DA,$FF,$BF,$23,$17,$F1
	.dc.b	$27,$81,$F8,$78,$F8,$44,$7D,$EF,$F7,$6E,$4F,$F6,$09,$23,$FD,$6F
	.dc.b	$3F,$EC,$3C,$BA,$DF,$FB,$18,$93,$A7,$FD,$93,$C3,$23,$8B,$67,$CB
	.dc.b	$97,$FB,$7D,$FF,$5F,$4F,$FB,$00,$FF,$02,$1F,$DE,$48,$60,$87,$F7
	.dc.b	$53,$25,$32,$7F,$81,$0C,$94,$C9,$FE,$16,$32,$53,$37,$FE,$17,$72
	.dc.b	$D7,$3C,$5F,$E2,$87,$A1,$E1,$21,$82,$1D,$4C,$94,$CC,$75,$72,$D7
	.dc.b	$31,$DD,$E0,$48,$D3,$63,$51,$82,$38,$24,$7A,$63,$2D,$8C,$BF,$01
	.dc.b	$AF,$C0,$34,$43,$45,$34,$60,$86,$08,$7B,$9E,$AD,$39,$14,$E2,$53
	.dc.b	$89,$F1,$0D,$94,$D9,$4D,$9F,$14,$D9,$4D,$94,$DC,$7C,$47,$2D,$73
	.dc.b	$1E,$87,$B1,$E5,$21,$82,$1D,$4C,$94,$CC,$75,$72,$D7,$31,$DD,$E0
	.dc.b	$48,$D3,$63,$51,$82,$38,$24,$7C,$28,$CB,$63,$2F,$C0,$6B,$F0,$0D
	.dc.b	$10,$D1,$4D,$18,$21,$82,$1E,$E7,$AB,$4E,$45,$38,$94,$E2,$7C,$43
	.dc.b	$65,$36,$53,$67,$C5,$36,$53,$65,$37,$1F,$11,$CB,$5C,$C7,$A1,$EC
	.dc.b	$79,$48,$60,$87,$53,$25,$33,$1D,$5C,$BF,$BC,$D7,$FD,$71,$05,$B7
	.dc.b	$FA,$E6,$07,$FA,$D7,$C5,$FF,$54,$91,$FF,$50,$B7,$87,$FD,$66,$D8
	.dc.b	$E2,$7F,$6B,$70,$F3,$38,$92,$38,$B4,$E2,$2D,$3C,$A3,$A1,$B3,$A4
	.dc.b	$F2,$23,$CA,$78,$BB,$38,$C6,$9E,$26,$E7,$B3,$E1,$6C,$F2,$9E,$13
	.dc.b	$C8,$E0,$75,$38,$16,$38,$3A,$8E,$2E,$9B,$1A,$8F,$19,$0E,$C5,$38
	.dc.b	$1C,$C8,$65,$E4,$F0,$21,$C0,$D4,$FB,$08,$F9,$DB,$3E,$46,$C7,$63
	.dc.b	$9C,$F6,$7A,$18,$8D,$BD,$8E,$84,$9E,$8D,$3A,$1C,$89,$3D,$49,$3D
	.dc.b	$23,$B1,$C4,$ED,$3D,$08,$F4,$9E,$2E,$CE,$31,$A7,$89,$B9,$EC,$F8
	.dc.b	$9B,$3C,$A7,$84,$F2,$38,$1D,$4E,$05,$8E,$0E,$A3,$8B,$A6,$C6,$A3
	.dc.b	$C6,$43,$B1,$4E,$07,$32,$19,$79,$3C,$08,$70,$35,$3E,$C2,$3E,$76
	.dc.b	$CF,$91,$B1,$D8,$E7,$3D,$9E,$86,$23,$6F,$63,$A1,$27,$A3,$4E,$87
	.dc.b	$22,$4F,$52,$4F,$48,$EC,$71,$3B,$4F,$42,$3D,$27,$8B,$B3,$8C,$69
	.dc.b	$E2,$6E,$7B,$3E,$26,$CF,$29,$E1,$3C,$8E,$07,$53,$81,$63,$F3,$59
	.dc.b	$7F,$AB,$E3,$FD,$07,$C7,$FE,$AD,$7F,$8F,$46,$5F,$E5,$32,$D3,$FD
	.dc.b	$5F,$C1,$FC,$6D,$B3,$FE,$9B,$BF,$F3,$4B,$FE,$AB,$23,$FE,$7B,$E9
	.dc.b	$63,$FE,$B1,$9F,$E6,$B1,$87,$FD,$7F,$7F,$CC,$23,$15,$C6,$B8,$57
	.dc.b	$FA,$97,$B6,$BF,$CD,$5B,$89,$A6,$31,$FF,$55,$4F,$F3,$B6,$CF,$B6
	.dc.b	$3F,$D6,$67,$FB,$3B,$97,$C0,$E0,$63,$E7,$8C,$93,$2F,$89,$3E,$E6
	.dc.b	$E3,$FE,$98,$C3,$FA,$DB,$67,$47,$FB,$7F,$2F,$A5,$8F,$FA,$B7,$97
	.dc.b	$FD,$C3,$3F,$58,$FE,$A7,$BF,$FA,$D1,$72,$FF,$AD,$F8,$3F,$D7,$38
	.dc.b	$FF,$60,$AD,$4E,$78,$BF,$D7,$53,$FE,$ED,$7F,$D2,$B7,$4F,$49,$E5
	.dc.b	$26,$98,$3E,$A4,$FF,$AC,$DB,$1F,$EB,$23,$FB,$BE,$DF,$6F,$97,$E4
	.dc.b	$9E,$F2,$4F,$C6,$6C,$FE,$3B,$83,$FB,$FE,$1F,$EB,$E6,$D7,$F6,$96
	.dc.b	$CE,$C4,$FD,$62,$4F,$91,$B1,$FE,$EF,$7F,$2A,$48,$FF,$80,$2F,$D3
	.dc.b	$23,$F6,$9B,$1F,$14,$87,$FB,$C9,$FF,$78,$1F,$E0,$86,$1F,$DC,$8A
	.dc.b	$7F,$83,$CA,$7F,$82,$0A,$7F,$76,$2E,$BF,$E0,$66,$F1,$7F,$A8,$08
	.dc.b	$FF,$A7,$5A,$E8,$C1,$0C,$10,$F2,$10,$F2,$14,$F2,$14,$F2,$16,$3C
	.dc.b	$8D,$8F,$0B,$4F,$01,$4F,$01,$D8,$87,$62,$9D,$8B,$1D,$9B,$1D,$9A
	.dc.b	$76,$29,$D8,$E8,$43,$A1,$63,$A3,$63,$A3,$4E,$85,$3A,$1C,$89,$1C
	.dc.b	$9B,$1C,$9A,$72,$29,$CA,$38,$12,$38,$36,$38,$35,$E1,$1A,$23,$A6
	.dc.b	$BA,$30,$43,$04,$3E,$A1,$0F,$92,$53,$E0,$14,$F4,$96,$3C,$8D,$8F
	.dc.b	$0B,$4F,$01,$4F,$01,$D8,$87,$62,$9D,$8B,$1D,$9B,$1D,$9A,$76,$29
	.dc.b	$D8,$E8,$43,$A1,$63,$A3,$63,$A3,$4E,$85,$3A,$1C,$89,$1C,$9B,$1C
	.dc.b	$9A,$72,$29,$CA,$38,$12,$38,$36,$38,$35,$E1,$1A,$23,$A6,$BA,$30
	.dc.b	$43,$04,$3E,$A1,$0F,$92,$53,$E0,$14,$F4,$9F,$D4,$35,$FF,$5A,$3F
	.dc.b	$D3,$5A,$FF,$AD,$B5,$FF,$5C,$DF,$FA,$A2,$31,$3A,$7F,$D5,$5F,$31
	.dc.b	$FD,$89,$FF,$52,$35,$FF,$07,$B5,$FF,$60,$E0,$8C,$47,$23,$FC,$10
	.dc.b	$70,$29,$C1,$DC,$71,$8C,$9C,$75,$F3,$46,$5B,$3C,$09,$3C,$27,$4F
	.dc.b	$A0,$C4,$6A,$78,$C6,$08,$E0,$8F,$16,$BE,$96,$BE,$93,$89,$B2,$1B
	.dc.b	$29,$B7,$8B,$ED,$32,$53,$81,$0E,$06,$A7,$DA,$E4,$E0,$68,$C3,$A7
	.dc.b	$04,$36,$47,$4D,$9F,$8C,$F7,$6B,$D1,$AF,$38,$E2,$48,$CC,$73,$3E
	.dc.b	$29,$C4,$A7,$17,$71,$C6,$32,$71,$9F,$8F,$19,$6C,$F0,$24,$F0,$9D
	.dc.b	$3E,$93,$11,$A9,$E3,$18,$23,$82,$3C,$5A,$FA,$5A,$FA,$4E,$26,$C8
	.dc.b	$6C,$A6,$DE,$2F,$B4,$C9,$4E,$04,$38,$1A,$9F,$6B,$93,$81,$A3,$0E
	.dc.b	$9C,$10,$D9,$1D,$36,$7E,$33,$DD,$AF,$46,$BC,$E3,$89,$23,$31,$CC
	.dc.b	$F8,$A7,$12,$9C,$5D,$C7,$18,$C9,$C6,$7E,$3C,$65,$B3,$C0,$93,$C2
	.dc.b	$74,$FA,$4C,$46,$A7,$8C,$60,$8E,$08,$F1,$6B,$E9,$6B,$E9,$38,$9B
	.dc.b	$21,$B2,$9B,$78,$BE,$D3,$27,$FA,$F3,$7F,$F4,$5B,$FE,$8F,$3F,$D1
	.dc.b	$85,$3F,$D7,$F1,$FE,$AA,$23,$FE,$B4,$32,$53,$F9,$AF,$3F,$FA,$C0
	.dc.b	$8E,$45,$A3,$FD,$65,$1F,$EA,$B8,$FF,$58,$71,$FF,$A8,$C9,$AF,$FA
	.dc.b	$31,$FF,$56,$4F,$FB,$1A,$FF,$D7,$3C,$1F,$EB,$1A,$4C,$5E,$7B,$EC
	.dc.b	$33,$E6,$A7,$75,$6B,$4A,$74,$F3,$8C,$C6,$23,$85,$74,$AE,$71,$EE
	.dc.b	$EA,$C3,$FE,$E8,$6B,$FD,$69,$AE,$CF,$04,$7F,$AA,$BC,$FF,$F6,$79
	.dc.b	$1C,$1F,$EC,$CE,$EF,$F6,$41,$1C,$1E,$2D,$7F,$D9,$11,$FD,$73,$A9
	.dc.b	$C0,$8C,$37,$0E,$13,$A9,$FE,$F1,$C9,$FE,$C8,$DF,$FB,$F7,$27,$FB
	.dc.b	$22,$FF,$BD,$B5,$4E,$DF,$EE,$F7,$E0,$EE,$EE,$48,$EF,$8A,$3F,$B8
	.dc.b	$F5,$FD,$A7,$FD,$C8,$4D,$3F,$DD,$6F,$FB,$CA,$64,$7F,$B4,$1B,$5F
	.dc.b	$3C,$7A,$1F,$5D,$F9,$CD,$D5,$BF,$E2,$23,$0D,$BF,$C2,$9F,$8A,$E0
	.dc.b	$C5,$FE,$0D,$AF,$E0,$8E,$F1,$C3,$8B,$F1,$77,$FE,$E0,$8F,$F7,$2F
	.dc.b	$55,$AF,$CD,$1F,$EE,$73,$F4,$BA,$7F,$BA,$DF,$CA,$E4,$8C,$47,$F8
	.dc.b	$89,$AF,$03,$94,$7E,$B5,$FF,$71,$D7,$E0,$F9,$BF,$DC,$F1,$FB,$63
	.dc.b	$27,$67,$2F,$ED,$B8,$6D,$1A,$48,$E2,$7F,$B8,$4F,$B7,$C3,$FE,$E2
	.dc.b	$69,$FE,$25,$3F,$C4,$64,$30,$43,$FC,$70,$64,$A6,$4F,$F1,$11,$92
	.dc.b	$99,$3F,$C8,$46,$4A,$66,$7F,$C8,$0E,$5A,$E7,$9F,$FC,$AC,$F4,$3C
	.dc.b	$24,$30,$43,$A9,$92,$99,$8E,$AE,$5A,$E6,$3B,$BC,$09,$1A,$6C,$6A
	.dc.b	$30,$47,$04,$8F,$4C,$65,$B1,$97,$E0,$35,$F8,$06,$88,$68,$A6,$8C
	.dc.b	$10,$C1,$0F,$73,$D5,$A7,$22,$9C,$4A,$71,$3E,$21,$B2,$9B,$29,$B3
	.dc.b	$E2,$9B,$29,$B2,$9B,$8F,$88,$E5,$AE,$63,$D0,$F6,$3C,$A4,$30,$43
	.dc.b	$A9,$92,$99,$8E,$AE,$5A,$E6,$3B,$BC,$09,$1A,$6C,$6A,$30,$47,$04
	.dc.b	$8F,$85,$19,$6C,$65,$F8,$0D,$7E,$01,$A2,$1A,$29,$A3,$04,$30,$43
	.dc.b	$DC,$F5,$69,$C8,$A7,$12,$9C,$4F,$88,$6C,$A6,$CA,$6C,$F8,$A6,$CA
	.dc.b	$6C,$A6,$E3,$E2,$39,$6B,$98,$F4,$3D,$8F,$29,$0C,$10,$EA,$64,$A7
	.dc.b	$E1,$78,$BF,$DC,$2F,$F6,$AE,$8F,$F7,$04,$62,$3F,$DD,$57,$FB,$8F
	.dc.b	$94,$FF,$6A,$92,$3F,$D9,$CD,$C7,$FD,$AA,$D8,$FE,$FC,$D7,$FA,$3B
	.dc.b	$5F,$F7,$3B,$5F,$F1,$8B,$71,$34,$91,$E0,$24,$F8,$1A,$78,$0A,$78
	.dc.b	$0A,$78,$0A,$78,$0B,$3E,$06,$CF,$82,$7B,$12,$3B,$35,$EC,$D7,$B3
	.dc.b	$5E,$CD,$9E,$CD,$9E,$D3,$D0,$87,$42,$9D,$0A,$74,$2C,$F4,$6C,$74
	.dc.b	$79,$10,$E4,$53,$91,$63,$93,$63,$93,$C0,$87,$02,$C7,$06,$C7,$07
	.dc.b	$44,$8D,$36,$35,$18,$23,$82,$47,$D4,$24,$FC,$96,$9F,$00,$A7,$A4
	.dc.b	$A7,$90,$A7,$84,$B3,$E0,$6C,$F8,$27,$B1,$23,$B3,$5E,$CD,$7B,$35
	.dc.b	$EC,$D9,$EC,$D9,$ED,$3D,$08,$74,$29,$D0,$A7,$42,$CF,$46,$C7,$47
	.dc.b	$91,$0E,$45,$39,$16,$39,$36,$39,$3C,$08,$70,$2C,$70,$6C,$70,$74
	.dc.b	$48,$D3,$63,$51,$82,$38,$24,$7D,$42,$4F,$C9,$69,$F0,$0A,$7A,$4A
	.dc.b	$79,$0A,$78,$4B,$3E,$06,$CF,$82,$7B,$1F,$E9,$4E,$D9,$3F,$ED,$43
	.dc.b	$FD,$17,$C7,$FE,$D8,$7F,$CC,$47,$FB,$36,$FF,$CC,$CD,$DB,$FD,$9B
	.dc.b	$5F,$E7,$FD,$7F,$DA,$6D,$9E,$11,$FE,$C8,$8F,$F4,$27,$2F,$FA,$9B
	.dc.b	$AB,$FD,$4A,$4A,$FF,$4B,$BF,$EB,$E7,$FD,$B7,$DF,$FE,$D0,$8F,$F4
	.dc.b	$63,$78,$7F,$D6,$4D,$9F,$F3,$DB,$5F,$F6,$BF,$A8,$F8,$1C,$42,$30
	.dc.b	$AF,$FD,$35,$18,$6C,$FF,$B8,$BC,$3F,$F6,$5F,$71,$A4,$BF,$F6,$69
	.dc.b	$FE,$8C,$CB,$FD,$87,$DB,$FE,$C2,$24,$7F,$AD,$1F,$76,$3E,$27,$BC
	.dc.b	$1A,$4C,$FF,$D7,$7C,$3F,$EC,$C2,$4E,$08,$78,$78,$7F,$D1,$8E,$0F
	.dc.b	$BB,$1F,$D9,$F8,$FF,$D9,$9C,$9F,$EC,$B9,$FF,$66,$F2,$A7,$2D,$87
	.dc.b	$83,$D1,$DC,$E8,$FF,$6B,$72,$B5,$FF,$71,$CF,$FB,$88,$87,$FB,$50
	.dc.b	$A7,$9C,$A7,$E4,$8F,$8B,$C0,$7F,$DD,$BA,$57,$EE,$C6,$74,$F9,$24
	.dc.b	$9F,$F7,$0B,$FE,$F2,$24,$E0,$E4,$38,$3E,$E3,$FE,$CE,$3B,$38,$E9
	.dc.b	$FF,$6F,$36,$B8,$B6,$7C,$9C,$1F,$8C,$E9,$FB,$64,$AE,$F1,$97,$F3
	.dc.b	$33,$FE,$CC,$E4,$FF,$67,$E8,$69,$8D,$19,$B1,$FD,$35,$7C,$F8,$C5
	.dc.b	$FF,$C0,$09,$CD,$70,$77,$CD,$FE,$D0,$24,$FF,$B3,$85,$FF,$B7,$76
	.dc.b	$FF,$7D,$73,$FF,$B5,$89,$3B,$9F,$E0,$73,$FB,$CC,$7F,$0D,$3F,$91
	.dc.b	$8F,$F6,$B3,$4F,$F0,$E1,$FE,$1F,$21,$82,$1F,$E1,$F3,$25,$32,$7F
	.dc.b	$88,$4C,$94,$C9,$FE,$23,$32,$53,$33,$FE,$25,$72,$D7,$3D,$1F,$EC
	.dc.b	$32,$3F,$E4,$92,$1C,$1C,$B4,$C9,$C8,$A7,$A4,$87,$9C,$A7,$94,$B1
	.dc.b	$E3,$6C,$76,$7A,$1D,$C8,$60,$87,$03,$C2,$77,$21,$DC,$B1,$DD,$B1
	.dc.b	$DD,$E4,$76,$29,$D8,$D9,$D8,$E8,$48,$E8,$D8,$E8,$F0,$3A,$14,$E8
	.dc.b	$6C,$E9,$1C,$89,$1C,$A3,$47,$22,$9C,$A3,$6F,$28,$E1,$18,$8E,$0D
	.dc.b	$78,$46,$88,$F1,$21,$C1,$CB,$4C,$9C,$8A,$7C,$82,$1E,$F2,$9E,$82
	.dc.b	$C7,$8D,$B1,$D9,$E8,$77,$21,$82,$1C,$0F,$09,$DC,$87,$72,$C7,$76
	.dc.b	$C7,$77,$91,$D8,$A7,$63,$67,$63,$A1,$23,$A3,$63,$A3,$C0,$E8,$53
	.dc.b	$A1,$B3,$A4,$72,$24,$72,$8D,$1C,$8A,$72,$8D,$BC,$A3,$84,$62,$38
	.dc.b	$35,$E1,$1A,$23,$C4,$87,$E3,$F8,$FF,$D9,$CF,$F5,$36,$BF,$D5,$DA
	.dc.b	$FF,$66,$3F,$9B,$FC,$07,$F5,$BE,$5B,$1F,$DA,$CF,$EB,$D6,$9D,$47
	.dc.b	$F5,$68,$C3,$B3,$F9,$BC,$69,$FE,$C9,$1A,$9E,$A4,$8E,$53,$B7,$BB
	.dc.b	$5E,$67,$59,$E4,$F0,$31,$1B,$8C,$C7,$82,$30,$47,$64,$9F,$13,$B3
	.dc.b	$07,$57,$2F,$90,$C9,$B7,$2F,$99,$B3,$E2,$8E,$11,$89,$F1,$36,$39
	.dc.b	$9A,$3C,$04,$3B,$14,$D1,$82,$18,$21,$E3,$38,$14,$D9,$4D,$CF,$9D
	.dc.b	$C1,$1C,$12,$7D,$31,$D2,$30,$FA,$8E,$11,$A7,$BC,$6A,$7E,$01,$23
	.dc.b	$94,$ED,$F1,$B5,$E6,$75,$9E,$4F,$03,$11,$B8,$CC,$78,$23,$04,$76
	.dc.b	$49,$F1,$3B,$30,$75,$72,$F9,$0C,$9B,$72,$F9,$9B,$3E,$28,$E1,$18
	.dc.b	$9F,$13,$63,$99,$A3,$C0,$43,$B1,$4D,$18,$21,$82,$1E,$33,$81,$4D
	.dc.b	$94,$DC,$F9,$DC,$11,$C1,$27,$D3,$1D,$23,$0F,$A8,$E1,$1A,$7B,$C6
	.dc.b	$A7,$E0,$12,$39,$4E,$DF,$1B,$5E,$67,$59,$E4,$F0,$31,$1B,$8C,$C7
	.dc.b	$82,$30,$47,$64,$9F,$13,$B3,$07,$57,$2F,$90,$C9,$B7,$2F,$99,$B3
	.dc.b	$E2,$8E,$11,$FC,$13,$5F,$F3,$3F,$4F,$FB,$7A,$7F,$D4,$8D,$7F,$97
	.dc.b	$EF,$FE,$D5,$E0,$FF,$39,$CF,$FB,$3A,$7F,$DC,$13,$97,$FD,$07,$75
	.dc.b	$FF,$4A,$71,$7F,$AB,$AB,$FD,$0B,$39,$9F,$F4,$63,$71,$FF,$3B,$74
	.dc.b	$FF,$B7,$E7,$FD,$6C,$D7,$C6,$7F,$B4,$F1,$FF,$42,$C7,$8C,$8F,$FA
	.dc.b	$2B,$97,$FD,$B7,$C9,$FE,$D7,$8C,$57,$13,$FD,$27,$3F,$EE,$3F,$65
	.dc.b	$F5,$EF,$3F,$0D,$AE,$02,$77,$3A,$3F,$DD,$39,$3C,$0F,$F5,$A7,$86
	.dc.b	$CB,$FA,$B4,$61,$B8,$7F,$B9,$0F,$EB,$B1,$97,$FD,$EE,$7F,$B7,$E3
	.dc.b	$FD,$47,$DB,$F4,$48,$FF,$AF,$F1,$F1,$CE,$CF,$F6,$23,$89,$F1,$46
	.dc.b	$32,$F6,$46,$23,$D9,$18,$AF,$ED,$9C,$5F,$86,$9B,$CE,$E1,$FC,$34
	.dc.b	$7F,$A9,$27,$4F,$D8,$37,$C1,$F6,$5D,$52,$3F,$DD,$D1,$FE,$E8,$E1
	.dc.b	$FF,$03,$76,$FF,$BA,$FB,$7F,$DD,$5C,$FF,$EF,$13,$FD,$E5,$C3,$FE
	.dc.b	$BB,$EC,$FF,$5C,$E5,$FE,$BC,$C7,$FD,$F9,$18,$3F,$D5,$AF,$F8,$23
	.dc.b	$3F,$F6,$2F,$56,$23,$C6,$FF,$B7,$B8,$7E,$4D,$FC,$36,$EF,$F3,$78
	.dc.b	$3F,$05,$39,$6C,$7F,$B5,$A3,$F4,$84,$70,$7D,$EE,$DF,$9D,$87,$FB
	.dc.b	$2B,$8F,$FD,$F1,$18,$D3,$F1,$13,$FE,$ED,$3F,$06,$5C,$7E,$EE,$9E
	.dc.b	$1E,$5F,$F8,$00,$D9,$FF,$70,$63,$E7,$9C,$BF,$A0,$9F,$F7,$74,$FE
	.dc.b	$80,$D9,$5F,$DE,$12,$3F,$DE,$2D,$8F,$9A,$7F,$BA,$48,$60,$FE,$02
	.dc.b	$7F,$B9,$FA,$1C,$48,$C3,$76,$FE,$07,$1F,$EE,$7E,$9C,$61,$FF,$00
	.dc.b	$27,$F8,$71,$1F,$C8,$BF,$67,$3F,$F7,$E3,$7A,$BB,$5F,$F9,$20,$87
	.dc.b	$F8,$D4,$A7,$F8,$B4,$A7,$F8,$AC,$A7,$F8,$F8,$B3,$FE,$4C,$6C,$76
	.dc.b	$7A,$10,$E8,$53,$A1,$4E,$85,$8E,$8D,$8E,$8F,$22,$1C,$8A,$72,$2C
	.dc.b	$72,$6C,$72,$78,$10,$E0,$58,$E0,$D8,$E0,$E8,$91,$A6,$C6,$A3,$04
	.dc.b	$70,$48,$F1,$46,$5B,$19,$7E,$59,$82,$18,$21,$F1,$4C,$94,$C9,$F0
	.dc.b	$0C,$94,$C9,$ED,$32,$53,$27,$B0,$C9,$4C,$C7,$A5,$B1,$E9,$7C,$44
	.dc.b	$3C,$05,$3B,$14,$E8,$58,$E8,$D8,$E8,$F2,$21,$C8,$A7,$22,$C7,$26
	.dc.b	$C7,$27,$81,$0E,$05,$8E,$0D,$8E,$0E,$89,$1A,$6C,$6A,$30,$47,$04
	.dc.b	$8F,$97,$19,$6C,$65,$F9,$66,$08,$60,$87,$C5,$32,$53,$27,$C0,$32
	.dc.b	$53,$27,$B4,$C9,$4C,$9E,$C3,$25,$33,$1E,$96,$C7,$A5,$F1,$10,$F0
	.dc.b	$14,$EC,$53,$A1,$63,$A3,$63,$A3,$C8,$87,$22,$9C,$8F,$EC,$F3,$FB
	.dc.b	$FE,$2F,$F7,$59,$39,$11,$FB,$2F,$47,$FD,$A8,$4A,$FF,$6D,$37,$3F
	.dc.b	$E2,$38,$A3,$13,$FE,$ED,$6B,$FE,$EA,$6B,$FE,$F4,$6A,$8E,$4F,$62
	.dc.b	$4F,$69,$DC,$EA,$31,$3A,$3C,$05,$3C,$05,$3C,$05,$3C,$13,$D0,$8E
	.dc.b	$67,$B3,$D2,$7B,$91,$C1,$0E,$EE,$5A,$64,$F0,$B9,$69,$93,$C6,$78
	.dc.b	$67,$C7,$1B,$8C,$B6,$7C,$B1,$D8,$87,$42,$9C,$0E,$CF,$07,$44,$34
	.dc.b	$70,$74,$E0,$D3,$82,$4F,$9A,$3A,$B5,$E4,$D7,$8B,$5E,$84,$8E,$6F
	.dc.b	$C1,$24,$FB,$E7,$84,$EA,$31,$3A,$3C,$25,$3C,$25,$3C,$25,$3C,$13
	.dc.b	$D0,$8E,$67,$B3,$D2,$7B,$91,$C1,$0E,$EE,$5A,$64,$F0,$B9,$69,$93
	.dc.b	$C6,$78,$67,$C7,$1B,$8C,$B6,$7C,$B1,$D8,$87,$42,$9C,$0E,$CF,$07
	.dc.b	$44,$34,$70,$74,$E0,$D3,$82,$4F,$9A,$3A,$B5,$E4,$D7,$8B,$5E,$84
	.dc.b	$8E,$6F,$C1,$24,$FB,$E7,$84,$EA,$31,$3A,$3C,$25,$3C,$25,$3C,$25
	.dc.b	$3C,$13,$D0,$8E,$67,$B3,$FE,$9A,$E7,$FF,$4C,$C3,$97,$FD,$33,$3F
	.dc.b	$E8,$7E,$4F,$F4,$F3,$FE,$3F,$F2,$1F,$19,$FE,$5B,$F1,$BF,$CF,$71
	.dc.b	$89,$46,$27,$5C,$1C,$5B,$3F,$E5,$6C,$7C,$18,$7F,$99,$A3,$FD,$5B
	.dc.b	$A7,$FA,$70,$8E,$08,$71,$C9,$3C,$63,$0D,$7F,$CE,$19,$79,$7C,$BF
	.dc.b	$24,$1A,$4C,$BA,$C7,$C1,$C3,$C2,$F4,$7F,$A2,$F9,$3F,$E7,$B7,$FD
	.dc.b	$30,$D9,$FF,$32,$64,$8C,$57,$BD,$AF,$FB,$8F,$C6,$7D,$2E,$2D,$18
	.dc.b	$9F,$F4,$4C,$7F,$9F,$0C,$CE,$E7,$58,$7F,$B6,$67,$8F,$D3,$7C,$EB
	.dc.b	$EC,$D8,$CC,$7D,$67,$A5,$27,$84,$FF,$B5,$AB,$E0,$BF,$0C,$F8,$B5
	.dc.b	$E2,$7C,$A7,$FA,$75,$C4,$78,$EB,$CB,$1D,$AF,$E0,$D7,$53,$E7,$94
	.dc.b	$F9,$C6,$31,$FC,$2D,$F8,$A3,$47,$FA,$AF,$93,$FD,$C9,$C9,$FE,$E7
	.dc.b	$9F,$F0,$1C,$FF,$BA,$CD,$46,$5B,$C9,$FE,$B3,$EE,$FF,$76,$CE,$67
	.dc.b	$FD,$EA,$7F,$BB,$08,$7F,$C0,$02,$9F,$8E,$29,$FA,$73,$FD,$49,$97
	.dc.b	$C5,$7E,$16,$5F,$65,$F8,$D9,$FD,$D7,$27,$C9,$3A,$1D,$C8,$60,$E3
	.dc.b	$A7,$FB,$2F,$CA,$F9,$CD,$BF,$B6,$D9,$D8,$9E,$EC,$9F,$A8,$F8,$4F
	.dc.b	$CA,$5F,$9D,$FF,$6C,$6B,$F9,$B7,$47,$F8,$77,$6D,$34,$CC,$6D,$FF
	.dc.b	$7E,$71,$7E,$D9,$F8,$E7,$CE,$3D,$A7,$ED,$B8,$FF,$DC,$DE,$6F,$ED
	.dc.b	$89,$8F,$F0,$99,$47,$4E,$4F,$F5,$6D,$F5,$9E,$67,$53,$E4,$1F,$54
	.dc.b	$CE,$1F,$C4,$1F,$F6,$7F,$27,$F1,$9D,$3F,$82,$93,$1F,$03,$DF,$87
	.dc.b	$FD,$E1,$87,$F1,$07,$FD,$E5,$CF,$FE,$C3,$26,$1F,$E4,$F2,$1F,$E3
	.dc.b	$C2,$1F,$E3,$C2,$9F,$E3,$C2,$9F,$E3,$C2,$9F,$E2,$A2,$E3,$FE,$4C
	.dc.b	$6C,$78,$1A,$78,$0E,$C4,$3B,$14,$EC,$53,$B1,$63,$B3,$63,$B3,$4E
	.dc.b	$C7,$42,$1D,$0A,$74,$2C,$74,$6C,$74,$69,$D0,$E4,$43,$91,$63,$93
	.dc.b	$63,$93,$4E,$47,$02,$47,$06,$C7,$06,$9C,$23,$44,$8D,$36,$35,$18
	.dc.b	$23,$82,$1F,$50,$87,$C9,$29,$F0,$0A,$7A,$4A,$79,$0B,$1E,$16,$C7
	.dc.b	$81,$A7,$80,$EC,$43,$B1,$4E,$C5,$3B,$16,$3B,$36,$3B,$34,$EC,$74
	.dc.b	$21,$D0,$A7,$42,$C7,$46,$C7,$46,$9D,$0E,$44,$39,$16,$39,$36,$39
	.dc.b	$34,$E4,$70,$24,$70,$6C,$70,$69,$C2,$34,$48,$D3,$63,$51,$82,$38
	.dc.b	$21,$F5,$08,$7C,$92,$9F,$00,$A7,$A4,$A7,$90,$B1,$E1,$6C,$78,$1A
	.dc.b	$78,$0E,$C4,$3B,$1F,$BF,$E4,$FF,$73,$57,$FB,$A7,$8B,$FD,$CF,$22
	.dc.b	$1F,$DA,$0F,$F7,$2F,$CC,$7F,$88,$5F,$F0,$FB,$5F,$F1,$E3,$5F,$E9
	.dc.b	$2D,$7F,$DC,$8D,$C0,$D2,$47,$F8,$E4,$D9,$63,$6D,$8D,$F2,$F9,$23
	.dc.b	$04,$70,$47,$CE,$64,$AE,$5F,$49,$92,$B9,$7D,$66,$4A,$E5,$F3,$B6
	.dc.b	$7C,$ED,$9F,$3C,$F8,$08,$76,$29,$D0,$A7,$42,$CF,$46,$C7,$47,$91
	.dc.b	$0E,$45,$39,$16,$39,$36,$39,$3C,$08,$70,$2C,$70,$6C,$70,$74,$48
	.dc.b	$D3,$63,$51,$82,$38,$24,$7C,$D3,$81,$63,$6D,$8D,$CF,$CD,$8C,$11
	.dc.b	$C1,$1F,$8E,$64,$AE,$5F,$84,$64,$AE,$5F,$79,$92,$B9,$7C,$ED,$9F
	.dc.b	$3B,$67,$CF,$3E,$02,$1D,$8A,$74,$29,$D0,$B3,$D1,$B1,$D1,$E4,$43
	.dc.b	$91,$4E,$45,$8E,$4D,$8E,$4F,$02,$1C,$0B,$1C,$1B,$1C,$1D,$12,$34
	.dc.b	$D8,$D4,$60,$8E,$09,$1F,$34,$E0,$58,$DB,$63,$73,$F3,$63,$04,$70
	.dc.b	$47,$E3,$99,$2B,$97,$E1,$19,$2B,$97,$DE,$64,$AE,$5F,$3B,$F8,$1C
	.dc.b	$BF,$DD,$70,$AF,$F6,$FF,$47,$FB,$C2,$73,$F0,$3F,$EB,$1A,$FF,$77
	.dc.b	$9F,$CC,$1C,$4F,$F3,$28,$FE,$69,$18,$70,$D7,$F9,$BF,$8E,$F8,$0F
	.dc.b	$F7,$B4,$71,$8C,$3F,$EF,$63,$11,$A7,$8B,$5F,$F6,$01,$FE,$B4,$21
	.dc.b	$81,$E0,$BF,$02,$7C,$87,$FB,$7A,$30,$D8,$CC,$FF,$B2,$0E,$2E,$34
	.dc.b	$71,$FF,$BA,$BC,$CF,$F6,$9E,$C6,$92,$FF,$DB,$47,$90,$FF,$62,$92
	.dc.b	$3F,$D9,$2F,$31,$AF,$67,$FA,$EC,$75,$73,$5B,$9D,$47,$FA,$8A,$BF
	.dc.b	$B2,$47,$FB,$77,$EA,$7F,$DE,$44,$9F,$65,$7F,$BE,$A3,$FD,$ED,$F4
	.dc.b	$DF,$EF,$1E,$E3,$49,$5E,$A3,$E6,$1E,$9E,$17,$4F,$F6,$16,$E9,$FE
	.dc.b	$CE,$B6,$1E,$18,$F2,$5F,$8A,$72,$FC,$F3,$A5,$F3,$24,$FF,$BB,$74
	.dc.b	$F8,$F3,$F2,$DA,$FF,$80,$7B,$FF,$DC,$FC,$DF,$EE,$9A,$FF,$77,$C6
	.dc.b	$27,$FD,$E0,$39,$FF,$DE,$B5,$B8,$FF,$59,$1F,$EB,$43,$FD,$6C,$60
	.dc.b	$86,$09,$7F,$EB,$77,$FC,$27,$AF,$CD,$7F,$DB,$0D,$7E,$09,$1F,$F5
	.dc.b	$FE,$9F,$EC,$58,$F0,$4F,$37,$FD,$E7,$AF,$FB,$9D,$FC,$8E,$9F,$59
	.dc.b	$FF,$74,$5F,$82,$31,$D3,$F8,$48,$46,$1D,$6D,$FE,$E9,$7C,$86,$2B
	.dc.b	$CF,$3F,$51,$FC,$69,$FC,$17,$83,$FC,$61,$9E,$32,$FC,$A3,$63,$37
	.dc.b	$F9,$B8,$FF,$71,$57,$FB,$B2,$BD,$8F,$CA,$7F,$39,$8F,$18,$F5,$9F
	.dc.b	$A0,$BF,$80,$F8,$4C,$9F,$B4,$E2,$F9,$EF,$F9,$0F,$8F,$F7,$6F,$F9
	.dc.b	$17,$1F,$F2,$2C,$62,$3F,$4E,$D8,$FA,$3C,$3F,$EF,$62,$4F,$F1,$B1
	.dc.b	$CB,$FC,$53,$1C,$36,$3F,$8A,$88,$CF,$2F,$F9,$55,$E5,$CD,$FC,$A7
	.dc.b	$5B,$DF,$F2,$AF,$F9,$2B,$5F,$E5,$8D,$7F,$CB,$DB,$38,$BF,$DF,$91
	.dc.b	$FC,$BE,$31,$9F,$F9,$92,$3F,$83,$CE,$9B,$19,$7F,$C8,$A7,$F9,$4C
	.dc.b	$87,$F9,$40,$A7,$F9,$38,$A7,$03,$44,$34,$53,$46,$08,$60,$87,$F9
	.dc.b	$A0,$EA,$53,$91,$4E,$25,$38,$CF,$F9,$B9,$E2,$D7,$8B,$5E,$2D,$78
	.dc.b	$F4,$7D,$D6,$96,$34,$F7,$23,$82,$47,$18,$CB,$63,$31,$CE,$30,$47
	.dc.b	$04,$3D,$2E,$5A,$64,$F5,$B9,$69,$93,$DA,$E5,$A6,$4F,$69,$D0,$87
	.dc.b	$22,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$7C,$63,$C8,$53,$B9
	.dc.b	$4E,$85,$39,$94,$E7,$1F,$39,$E2,$D7,$8B,$5E,$2D,$78,$C7,$C7,$69
	.dc.b	$E4,$8D,$3D,$C8,$E0,$91,$C6,$32,$D8,$CC,$73,$8C,$11,$C1,$0F,$4B
	.dc.b	$96,$99,$3D,$6E,$5A,$64,$F6,$B9,$69,$93,$DA,$74,$21,$C8,$A7,$02
	.dc.b	$9C,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$18,$F2,$14,$EE,$53,$A1,$4E
	.dc.b	$65,$39,$C7,$ED,$B9,$1D,$3F,$C7,$64,$7F,$7D,$70,$DC,$BF,$80,$4F
	.dc.b	$07,$F8,$17,$D4,$FF,$8F,$67,$FC,$B4,$D7,$98,$70,$F6,$6B,$9E,$2F
	.dc.b	$0F,$47,$F4,$99,$C9,$23,$33,$D9,$AF,$68,$D2,$9E,$53,$A8,$F0,$35
	.dc.b	$F0,$1B,$3A,$14,$E8,$EE,$7A,$CF,$49,$D1,$D0,$A7,$47,$44,$7A,$B5
	.dc.b	$EA,$70,$8E,$AE,$88,$6A,$38,$3D,$DA,$F7,$36,$64,$A6,$4E,$46,$23
	.dc.b	$BB,$82,$1C,$4A,$71,$30,$43,$04,$79,$B9,$6B,$99,$F9,$44,$7E,$03
	.dc.b	$67,$D7,$3C,$09,$19,$9F,$43,$5F,$2C,$69,$E6,$49,$E7,$3B,$8F,$23
	.dc.b	$5F,$21,$B3,$A9,$4E,$AE,$E7,$B4,$F4,$9D,$1D,$0A,$74,$74,$47,$AB
	.dc.b	$5E,$A7,$08,$EA,$E8,$86,$A3,$83,$DD,$AF,$73,$66,$4A,$64,$E4,$62
	.dc.b	$3B,$B8,$21,$C4,$A7,$13,$04,$30,$47,$9B,$96,$B9,$9F,$94,$47,$E0
	.dc.b	$36,$7D,$73,$C0,$91,$99,$F4,$35,$F2,$C6,$9E,$64,$9E,$73,$B8,$F2
	.dc.b	$35,$F2,$1B,$3A,$94,$EA,$EE,$7B,$4F,$49,$D1,$D0,$A7,$47,$44,$7A
	.dc.b	$B5,$EA,$70,$8E,$AE,$88,$7F,$94,$BA,$7F,$CD,$99,$7F,$96,$1B,$AF
	.dc.b	$FB,$29,$BA,$FF,$B8,$AB,$FC,$BD,$3F,$EC,$EF,$0B,$24,$8C,$CA,$3F
	.dc.b	$9B,$C7,$FC,$00,$E5,$76,$7F,$44,$9C,$C0,$CC,$F1,$3F,$D0,$EE,$23
	.dc.b	$6D,$8F,$F5,$77,$99,$FC,$CF,$9F,$FD,$5A,$DE,$93,$4F,$E6,$DA,$7F
	.dc.b	$B5,$AB,$FD,$BB,$B7,$B8,$FF,$73,$7C,$37,$12,$47,$F3,$4C,$3F,$D7
	.dc.b	$31,$D7,$E1,$7F,$DF,$B1,$FE,$BF,$6C,$FF,$A7,$F1,$46,$23,$FD,$8E
	.dc.b	$D7,$C2,$7C,$8B,$F1,$3C,$CF,$F7,$F5,$7F,$B2,$27,$E4,$39,$CF,$FD
	.dc.b	$99,$7D,$AB,$6B,$E1,$37,$1E,$36,$BF,$ED,$86,$D7,$FB,$51,$B2,$F8
	.dc.b	$3F,$ED,$33,$FE,$DB,$6D,$7B,$5A,$FF,$B8,$9A,$FB,$0F,$F4,$FE,$3E
	.dc.b	$C7,$E2,$7C,$1F,$62,$3E,$87,$CA,$F1,$3F,$D5,$FC,$2E,$1F,$EC,$7B
	.dc.b	$7D,$92,$46,$6F,$B5,$70,$AF,$ED,$3C,$DF,$E0,$D8,$78,$3F,$DC,$E3
	.dc.b	$0D,$E8,$74,$7F,$86,$23,$0D,$C3,$FC,$23,$5C,$1B,$5C,$2F,$AC,$62
	.dc.b	$B8,$47,$F8,$8A,$B8,$3F,$EE,$FC,$7D,$8F,$FB,$CD,$AF,$FB,$88,$C6
	.dc.b	$3F,$EE,$1D,$5B,$35,$FF,$6E,$47,$F8,$90,$CB,$8A,$F9,$E7,$FB,$DA
	.dc.b	$32,$FF,$BE,$06,$BF,$ED,$87,$B9,$FC,$00,$FD,$CC,$78,$36,$FF,$7E
	.dc.b	$C7,$DF,$3F,$DF,$87,$E0,$CA,$64,$FC,$AE,$FE,$37,$F2,$07,$8F,$0F
	.dc.b	$CB,$36,$35,$5F,$9F,$8D,$70,$E2,$BF,$3D,$D9,$FA,$08,$E9,$3E,$AD
	.dc.b	$3E,$4B,$FE,$EE,$3D,$67,$EA,$F6,$E0,$FD,$EC,$30,$47,$F7,$1C,$7F
	.dc.b	$ED,$38,$FD,$FC,$7D,$1C,$53,$96,$C7,$FB,$5E,$FC,$2F,$DD,$DB,$07
	.dc.b	$1E,$4E,$04,$8C,$1F,$C9,$8F,$E4,$DC,$DF,$E6,$8D,$3F,$8F,$B6,$3F
	.dc.b	$91,$B5,$FF,$36,$F3,$7F,$29,$8C,$47,$E8,$67,$F9,$C9,$1F,$E6,$AF
	.dc.b	$F3,$CD,$24,$7F,$38,$6B,$FE,$4A,$3F,$CA,$84,$3F,$CB,$65,$34,$60
	.dc.b	$86,$08,$7F,$96,$CE,$05,$36,$53,$67,$F9,$A8,$D9,$4D,$94,$DF,$43
	.dc.b	$A7,$E5,$47,$32,$1B,$2C,$6D,$FF,$47,$C7,$07,$44,$8D,$3E,$26,$BE
	.dc.b	$26,$BE,$28,$E5,$18,$8E,$F1,$82,$3E,$02,$1D,$5C,$B4,$C9,$DD,$CB
	.dc.b	$4C,$9E,$12,$9E,$02,$9E,$B7,$8B,$B3,$B1,$0E,$85,$34,$60,$86,$08
	.dc.b	$77,$38,$14,$D9,$4D,$9E,$33,$CB,$1F,$1D,$C1,$C8,$A7,$22,$9C,$9E
	.dc.b	$04,$8F,$6C,$77,$21,$C0,$B1,$B7,$C3,$1C,$1D,$12,$34,$F8,$DA,$F8
	.dc.b	$DA,$F8,$E3,$94,$62,$3B,$C6,$08,$F8,$08,$75,$72,$D3,$27,$77,$2D
	.dc.b	$32,$78,$4A,$78,$0A,$7A,$DE,$2E,$CE,$C4,$3A,$14,$D1,$82,$18,$21
	.dc.b	$DC,$E0,$53,$65,$36,$78,$CF,$2C,$7C,$77,$07,$22,$9C,$8A,$72,$78
	.dc.b	$12,$3D,$B1,$DC,$FE,$E1,$CD,$FD,$D5,$FF,$29,$F2,$FF,$80,$DB,$19
	.dc.b	$7F,$BC,$1F,$CC,$E9,$F3,$7F,$EF,$72,$47,$FB,$B5,$BC,$FB,$6C,$7F
	.dc.b	$98,$27,$8B,$B2,$4E,$DA,$6E,$78,$A2,$4F,$29,$E0,$48,$E0,$72,$38
	.dc.b	$08,$E0,$EA,$38,$3A,$8C,$B6,$7C,$2F,$81,$AF,$36,$C7,$08,$D1,$E1
	.dc.b	$2C,$F6,$69,$E1,$21,$DC,$E0,$75,$2C,$75,$71,$1C,$DA,$72,$3A,$9C
	.dc.b	$DD,$47,$12,$4F,$A0,$93,$D9,$B3,$C1,$B1,$D4,$EE,$75,$31,$3D,$1E
	.dc.b	$44,$9E,$4D,$37,$3D,$1E,$44,$3D,$24,$3B,$96,$7B,$4F,$42,$47,$43
	.dc.b	$B1,$CC,$E2,$48,$E2,$EE,$38,$BB,$8C,$B6,$7C,$CF,$85,$AF,$36,$C7
	.dc.b	$08,$D1,$E2,$2C,$F6,$69,$E1,$21,$DC,$E0,$75,$2C,$75,$71,$1C,$DA
	.dc.b	$72,$3A,$9C,$DD,$47,$12,$4F,$A0,$93,$D9,$B3,$C1,$B1,$D4,$EE,$75
	.dc.b	$31,$3D,$1E,$44,$9E,$4D,$37,$3D,$1E,$44,$3D,$24,$3B,$96,$7B,$4F
	.dc.b	$42,$47,$43,$B1,$CC,$E2,$48,$E2,$EE,$38,$BB,$8C,$B6,$7C,$CF,$85
	.dc.b	$AF,$37,$FD,$D3,$BF,$FA,$CF,$AB,$FC,$B1,$CE,$8C,$57,$FC,$00,$8C
	.dc.b	$70,$AB,$75,$AA,$FE,$2F,$F1,$EF,$37,$F9,$05,$7F,$98,$E3,$13,$C4
	.dc.b	$FE,$3A,$1C,$C6,$E7,$8C,$EA,$78,$F6,$66,$0D,$23,$FE,$A9,$E5,$7C
	.dc.b	$2F,$F3,$19,$FF,$50,$C6,$27,$89,$B9,$D3,$4C,$9F,$F0,$07,$E1,$7F
	.dc.b	$9D,$6F,$C9,$B1,$DD,$AE,$CF,$F6,$EE,$7F,$ED,$B7,$E5,$35,$FF,$75
	.dc.b	$98,$D1,$5E,$6F,$85,$F8,$33,$FD,$27,$AB,$FD,$8F,$18,$84,$61,$F9
	.dc.b	$86,$F2,$4E,$E7,$FD,$9F,$E9,$7D,$47,$E3,$36,$B8,$35,$EB,$48,$C4
	.dc.b	$7F,$BE,$CF,$F7,$E4,$FF,$B6,$BE,$13,$E7,$3F,$45,$B1,$E3,$6B,$D1
	.dc.b	$B0,$FB,$0F,$5C,$FD,$C8,$FF,$71,$46,$36,$CC,$FC,$08,$C4,$E8,$E6
	.dc.b	$BE,$17,$EB,$C7,$26,$CF,$CB,$9D,$3F,$42,$74,$D3,$EA,$90,$FE,$F5
	.dc.b	$C7,$FA,$2D,$3F,$43,$AF,$D8,$9F,$3F,$87,$E2,$23,$EB,$3C,$0F,$C2
	.dc.b	$7E,$55,$73,$7F,$05,$5D,$08,$E0,$87,$F8,$72,$7C,$0B,$A1,$D0,$EA
	.dc.b	$FF,$08,$56,$A7,$31,$8B,$E4,$4B,$FF,$16,$C6,$5A,$2B,$95,$F5,$8C
	.dc.b	$57,$FB,$D5,$FC,$D1,$FE,$F6,$30,$43,$FD,$D4,$7F,$00,$CB,$F0,$15
	.dc.b	$FE,$E5,$7E,$4C,$3A,$FE,$FC,$6A,$30,$EA,$3F,$07,$0C,$BF,$8A,$BE
	.dc.b	$13,$EB,$F1,$FF,$BE,$AB,$F1,$DB,$7F,$8B,$1F,$BF,$3D,$B6,$FC,$1B
	.dc.b	$67,$4F,$E1,$E7,$CC,$D3,$C6,$5C,$39,$46,$A3,$F3,$C6,$8B,$1E,$CE
	.dc.b	$27,$0E,$EB,$F5,$55,$B3,$D9,$8B,$83,$FD,$C7,$9F,$BA,$7E,$79,$0F
	.dc.b	$DC,$9F,$81,$AF,$DE,$B4,$CE,$FE,$1E,$2F,$F2,$A3,$E3,$E3,$B3,$FC
	.dc.b	$A5,$AF,$F0,$B6,$BF,$C1,$1A,$F9,$B0,$F2,$E3,$E7,$8F,$A5,$83,$37
	.dc.b	$27,$FB,$F2,$30,$D8,$CC,$FF,$44,$3F,$88,$C6,$79,$7F,$CD,$5B,$7F
	.dc.b	$3A,$6C,$FF,$37,$8C,$3F,$89,$3F,$A4,$73,$E9,$FE,$AA,$D7,$FC,$E9
	.dc.b	$CF,$FD,$52,$7F,$A6,$1F,$D3,$75,$93,$C8,$FF,$46,$1F,$D5,$EB,$E0
	.dc.b	$EF,$FD,$92,$7F,$81,$92,$7F,$9E,$3F,$E8,$56,$1F,$E7,$E2,$9F,$E6
	.dc.b	$92,$9F,$E7,$22,$9F,$E6,$B2,$F5,$7F,$A4,$9B,$D5,$F1,$DE,$07,$52
	.dc.b	$1D,$4A,$68,$C1,$0C,$10,$F0,$1C,$0A,$6C,$A6,$E3,$C8,$ED,$AE,$DA
	.dc.b	$EE,$3D,$0F,$22,$46,$9B,$1A,$8C,$11,$C1,$23,$E1,$46,$5B,$19,$7E
	.dc.b	$39,$82,$18,$21,$E9,$78,$34,$D9,$4D,$9E,$D7,$6D,$36,$53,$67,$C1
	.dc.b	$32,$53,$27,$C0,$32,$53,$31,$ED,$6C,$78,$1E,$47,$94,$87,$84,$A6
	.dc.b	$8C,$10,$C1,$0F,$01,$C0,$A6,$CA,$6E,$3C,$8E,$DA,$ED,$AE,$E3,$D0
	.dc.b	$F2,$24,$69,$B1,$A8,$C1,$1C,$12,$3E,$3C,$65,$B1,$97,$E3,$98,$21
	.dc.b	$82,$1E,$97,$83,$4D,$94,$D9,$ED,$76,$D3,$65,$36,$7C,$13,$25,$32
	.dc.b	$7C,$03,$25,$33,$1E,$D6,$C7,$81,$E4,$79,$48,$78,$4A,$68,$C1,$0C
	.dc.b	$1F,$81,$E8,$FF,$03,$BF,$CC,$DA,$FF,$9E,$1A,$8F,$F0,$39,$FE,$7C
	.dc.b	$97,$CC,$7F,$99,$23,$51,$87,$FC,$32,$62,$34,$A3,$5D,$1F,$D0,$DB
	.dc.b	$1D,$67,$71,$88,$F1,$1C,$63,$24,$7B,$99,$2C,$71,$7C,$46,$23,$84
	.dc.b	$F8,$5B,$1E,$13,$44,$8D,$34,$D3,$82,$38,$21,$E8,$9F,$13,$E5,$8D
	.dc.b	$B9,$9F,$0B,$4E,$C5,$8D,$BE,$A6,$BC,$CE,$E6,$4A,$64,$E4,$62,$7C
	.dc.b	$64,$9F,$0C,$71,$36,$43,$04,$3A,$16,$39,$3E,$83,$11,$D1,$D9,$23
	.dc.b	$73,$EA,$6C,$79,$67,$84,$62,$3C,$47,$18,$C9,$1F,$01,$92,$C7,$17
	.dc.b	$C6,$62,$38,$4F,$85,$B1,$E1,$34,$48,$D3,$4D,$38,$23,$82,$1E,$89
	.dc.b	$F1,$3E,$58,$DB,$99,$F0,$B4,$EC,$58,$DB,$EA,$6B,$CC,$EE,$64,$A6
	.dc.b	$4E,$46,$27,$C6,$49,$F0,$C7,$13,$64,$30,$43,$A1,$63,$93,$E8,$31
	.dc.b	$1D,$1D,$92,$37,$3E,$A6,$C7,$96,$78,$46,$23,$C4,$71,$8C,$91,$F0
	.dc.b	$19,$2C,$71,$7C,$66,$23,$84,$F8,$5B,$1E,$13,$44,$8D,$34,$D3,$82
	.dc.b	$38,$21,$F9,$8C,$5C,$3F,$E7,$FE,$BF,$F6,$69,$27,$F8,$E9,$FE,$6F
	.dc.b	$7F,$CC,$DE,$3B,$E1,$7F,$99,$46,$1B,$3C,$27,$4A,$74,$D3,$5F,$1D
	.dc.b	$C1,$FE,$57,$28,$C4,$7F,$3F,$8C,$3C,$4D,$C6,$94,$6B,$BB,$CF,$3F
	.dc.b	$D1,$DB,$26,$9F,$EE,$07,$FD,$E0,$7F,$A3,$72,$C3,$FF,$00,$7E,$1F
	.dc.b	$FD,$F6,$D3,$FD,$F0,$58,$FF,$5E,$46,$A3,$FD,$DF,$97,$FA,$AE,$79
	.dc.b	$C6,$3A,$3F,$A2,$E9,$FE,$F3,$8F,$EB,$9D,$0E,$D6,$3E,$08,$C3,$63
	.dc.b	$33,$E2,$9C,$BE,$43,$B6,$1E,$A8,$E9,$8F,$12,$4F,$1A,$FC,$1B,$7C
	.dc.b	$DE,$C4,$70,$6F,$D8,$F1,$11,$C1,$E1,$CB,$D8,$FF,$B9,$33,$F7,$E1
	.dc.b	$F0,$E3,$F1,$B3,$AC,$7E,$06,$1D,$F2,$FF,$73,$36,$37,$7C,$6F,$EB
	.dc.b	$DF,$67,$DE,$6B,$0E,$E7,$F7,$2E,$3F,$F0,$6F,$80,$CB,$FB,$CC,$61
	.dc.b	$BE,$0B,$9F,$FC,$1D,$5B,$C3,$FC,$21,$7B,$6D,$6D,$4E,$D4,$ED,$B3
	.dc.b	$B9,$D4,$E7,$1F,$F7,$75,$7F,$83,$1A,$FC,$93,$FD,$AC,$E1,$FF,$71
	.dc.b	$18,$CB,$FD,$C4,$FF,$BD,$4E,$07,$F8,$5B,$3F,$9A,$FE,$38,$FC,$7E
	.dc.b	$CD,$75,$18,$6A,$9E,$2D,$3C,$D5,$E0,$6E,$BF,$EF,$42,$3F,$C2,$A7
	.dc.b	$F3,$9B,$74,$6C,$78,$CE,$D8,$27,$4D,$8D,$4F,$88,$F6,$47,$A6,$32
	.dc.b	$D7,$F1,$26,$2F,$99,$1C,$1E,$13,$E5,$9F,$24,$A7,$B0,$F5,$47,$EC
	.dc.b	$1F,$DF,$9E,$0D,$FC,$E7,$C8,$78,$9F,$48,$E4,$75,$CB,$ED,$12,$BE
	.dc.b	$54,$6C,$FE,$1C,$76,$25,$FC,$0A,$ED,$5C,$71,$F9,$F5,$E9,$E2,$FE
	.dc.b	$4F,$D3,$FC,$80,$91,$FC,$C0,$3F,$C9,$38,$24,$65,$AF,$F3,$4E,$5F
	.dc.b	$F3,$84,$7F,$3A,$24,$7F,$3F,$8C,$BF,$B1,$8F,$CA,$5F,$E9,$63,$F0
	.dc.b	$B1,$83,$FC,$B2,$C3,$FC,$BC,$53,$FC,$A0,$53,$FC,$CE,$53,$FC,$D8
	.dc.b	$5E,$7F,$F3,$F3,$79,$FF,$DF,$44,$7F,$DE,$ED,$74,$60,$86,$08,$79
	.dc.b	$08,$79,$0A,$79,$0A,$79,$0B,$1E,$46,$C7,$85,$A7,$80,$A7,$80,$EC
	.dc.b	$43,$B1,$4E,$C5,$8E,$CD,$8E,$CD,$3B,$14,$EC,$74,$21,$D0,$B1,$D1
	.dc.b	$B1,$D1,$A7,$42,$9D,$0E,$44,$8E,$4D,$8E,$4D,$39,$14,$E5,$1C,$09
	.dc.b	$1C,$1B,$1C,$1A,$F0,$8D,$11,$D3,$5D,$18,$21,$82,$1F,$50,$87,$C9
	.dc.b	$29,$F0,$0A,$7A,$4B,$1E,$46,$C7,$85,$A7,$80,$A7,$80,$EC,$43,$B1
	.dc.b	$4E,$C5,$8E,$CD,$8E,$CD,$3B,$14,$EC,$74,$21,$D0,$B1,$D1,$B1,$D1
	.dc.b	$A7,$42,$9D,$0E,$44,$8E,$4D,$8E,$4D,$39,$14,$E5,$1C,$09,$1C,$1B
	.dc.b	$1C,$1A,$F0,$8D,$11,$D3,$5D,$18,$3F,$CB,$E7,$F7,$C9,$AF,$F7,$86
	.dc.b	$F0,$B8,$FF,$8E,$4F,$F8,$22,$77,$6F,$51,$3F,$F0,$01,$BC,$BD,$A7
	.dc.b	$53,$92,$46,$67,$71,$FE,$72,$9D,$CE,$63,$66,$67,$6F,$FB,$F5,$AF
	.dc.b	$8A,$78,$BB,$38,$9B,$29,$B8,$E2,$A3,$8B,$B9,$F1,$11,$E8,$D7,$46
	.dc.b	$08,$60,$87,$73,$C2,$77,$9F,$0C,$F8,$23,$53,$D5,$A7,$42,$9D,$0E
	.dc.b	$47,$48,$E4,$E2,$39,$34,$E4,$53,$94,$70,$24,$F9,$09,$1D,$5B,$3D
	.dc.b	$27,$73,$92,$46,$67,$71,$E0,$9D,$CE,$63,$66,$67,$6F,$91,$AF,$C0
	.dc.b	$21,$E9,$2C,$F4,$79,$1D,$0E,$45,$37,$1D,$1E,$44,$8E,$4F,$09,$F3
	.dc.b	$11,$EA,$D7,$46,$08,$60,$87,$80,$F1,$1D,$E7,$C5,$3E,$08,$D4,$F5
	.dc.b	$69,$D0,$A7,$43,$91,$D2,$39,$38,$8E,$4D,$39,$14,$E5,$1C,$09,$3E
	.dc.b	$42,$47,$56,$CF,$49,$DC,$E4,$91,$99,$DC,$78,$27,$73,$98,$D9,$99
	.dc.b	$DB,$E4,$6B,$F0,$08,$7A,$4B,$3D,$1E,$47,$43,$91,$4D,$C7,$47,$91
	.dc.b	$23,$93,$C2,$7C,$C4,$7A,$B5,$D1,$82,$1F,$E8,$6E,$8F,$F2,$29,$1F
	.dc.b	$F5,$C3,$FC,$CB,$AD,$D7,$FE,$FD,$6C,$FF,$C0,$13,$F9,$6B,$89,$E1
	.dc.b	$3F,$CC,$1F,$F7,$F9,$27,$FC,$C0,$D7,$FD,$5D,$E6,$6D,$FF,$69,$B5
	.dc.b	$C9,$DC,$71,$23,$51,$FE,$DD,$C3,$FD,$FB,$3F,$ED,$33,$FD,$A9,$87
	.dc.b	$A2,$3C,$B3,$FE,$C5,$3D,$38,$7F,$2E,$9D,$47,$FA,$73,$DB,$FF,$7F
	.dc.b	$11,$FF,$7B,$3F,$D2,$B9,$7F,$DB,$D7,$E0,$E6,$FF,$72,$57,$8C,$FF
	.dc.b	$76,$F4,$FC,$3A,$B0,$D3,$FD,$74,$FF,$56,$CF,$FD,$55,$18,$8F,$7D
	.dc.b	$71,$7F,$D8,$F8,$F8,$F1,$E3,$83,$A1,$39,$8F,$2D,$43,$FD,$EF,$97
	.dc.b	$9A,$38,$4E,$E7,$A9,$EE,$3F,$DE,$D1,$FE,$BE,$F5,$BE,$04,$7F,$B6
	.dc.b	$1A,$FF,$B2,$1B,$1E,$36,$D9,$A4,$9E,$A7,$32,$9F,$24,$F3,$F2,$FD
	.dc.b	$77,$F0,$38,$7A,$6B,$A4,$70,$BF,$03,$63,$38,$FC,$66,$CF,$C3,$9E
	.dc.b	$C4,$3A,$DF,$7D,$3F,$15,$87,$DC,$9C,$E1,$FE,$AC,$3F,$C2,$FC,$9F
	.dc.b	$E1,$3E,$3F,$F0,$7F,$5A,$7F,$C2,$94,$8C,$3F,$E1,$D3,$11,$A8,$CC
	.dc.b	$7F,$87,$CC,$9B,$73,$5D,$E3,$26,$E3,$8E,$1F,$ED,$38,$FF,$60,$90
	.dc.b	$FF,$70,$61,$FE,$DE,$7F,$DC,$A7,$FB,$9B,$2F,$92,$FC,$7D,$3F,$E0
	.dc.b	$09,$1F,$F7,$B5,$75,$79,$E7,$FE,$38,$AE,$27,$E1,$E3,$FC,$62,$D7
	.dc.b	$A1,$83,$F0,$06,$08,$EC,$FD,$93,$B0,$FE,$61,$C3,$5F,$BD,$AF,$FB
	.dc.b	$D3,$0F,$75,$F9,$E7,$FE,$00,$9F,$9E,$36,$72,$D7,$F5,$2D,$AF,$F7
	.dc.b	$EE,$DF,$EF,$0C,$3F,$73,$B7,$ED,$70,$4E,$6F,$FC,$9E,$7E,$E0,$4E
	.dc.b	$F1,$F1,$93,$2F,$F2,$ED,$A1,$1F,$2F,$2F,$1D,$F9,$89,$8F,$C4,$7F
	.dc.b	$83,$70,$7F,$34,$7F,$8E,$90,$F1,$87,$4F,$D8,$3F,$08,$7C,$43,$F8
	.dc.b	$97,$17,$F9,$A3,$2F,$F3,$9E,$5F,$CB,$E7,$FC,$DE,$7F,$32,$9D,$3F
	.dc.b	$CE,$0C,$F4,$BB,$73,$1F,$12,$30,$EA,$32,$FF,$9D,$A3,$37,$FC,$D8
	.dc.b	$91,$EA,$C3,$FA,$8C,$62,$3E,$7E,$1F,$27,$3F,$2B,$67,$D3,$A7,$CC
	.dc.b	$6C,$FA,$B9,$7F,$D0,$37,$FE,$7B,$21,$FE,$75,$29,$FE,$75,$29,$FE
	.dc.b	$67,$29,$FE,$5E,$2B,$D1,$E4,$43,$91,$4E,$45,$39,$15,$E4,$F0,$21
	.dc.b	$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$73,$7F,$AC,$A3,$2D
	.dc.b	$8C,$C7,$FA,$C2,$30,$47,$04,$3E,$2B,$96,$99,$3E,$03,$96,$99,$3D
	.dc.b	$AE,$5A,$64,$F6,$39,$69,$93,$D2,$58,$F4,$BE,$22,$1E,$02,$9D,$8A
	.dc.b	$74,$29,$D0,$AF,$47,$91,$0E,$45,$39,$14,$E4,$57,$93,$C0,$87,$02
	.dc.b	$9C,$0A,$F0,$74,$43,$45,$74,$E0,$8E,$09,$1F,$2E,$32,$D8,$CC,$7C
	.dc.b	$B8,$C1,$1C,$10,$F8,$AE,$5A,$64,$F8,$0E,$5A,$64,$F6,$B9,$69,$93
	.dc.b	$D8,$E5,$A6,$4F,$49,$63,$D2,$F8,$88,$78,$0A,$76,$29,$D0,$A7,$42
	.dc.b	$BD,$1E,$44,$39,$14,$E4,$53,$E9,$71,$FF,$7D,$7F,$9C,$B5,$FF,$05
	.dc.b	$98,$E7,$FE,$66,$A5,$F2,$9F,$EF,$82,$47,$F8,$A6,$32,$D8,$CF,$37
	.dc.b	$F4,$68,$E5,$18,$7F,$A4,$CF,$09,$D3,$63,$33,$B7,$FB,$0C,$65,$B1
	.dc.b	$94,$76,$3C,$44,$9E,$B1,$B8,$D4,$61,$D1,$E0,$29,$D8,$A7,$69,$F1
	.dc.b	$36,$7A,$46,$23,$A9,$CE,$7B,$11,$C1,$0E,$CE,$5A,$64,$EC,$78,$8E
	.dc.b	$A6,$88,$69,$E4,$D3,$91,$C0,$C1,$C0,$C1,$23,$D4,$49,$F6,$92,$3C
	.dc.b	$51,$96,$C6,$67,$C3,$1C,$E3,$0F,$8A,$78,$4E,$9B,$19,$9D,$BE,$58
	.dc.b	$CB,$63,$2F,$A8,$87,$73,$C6,$49,$ED,$1C,$23,$51,$87,$47,$84,$A7
	.dc.b	$72,$9D,$A7,$C6,$D9,$E9,$18,$8E,$A7,$39,$EC,$47,$04,$3B,$39,$69
	.dc.b	$93,$B1,$E2,$3A,$9A,$21,$A7,$93,$4E,$47,$03,$07,$03,$04,$8F,$51
	.dc.b	$27,$DA,$48,$F1,$46,$5B,$19,$9F,$0C,$73,$8C,$3E,$29,$E1,$3A,$6C
	.dc.b	$66,$76,$F9,$63,$2D,$8C,$BE,$A2,$1D,$CF,$19,$27,$B4,$70,$8D,$46
	.dc.b	$1D,$1E,$13,$E7,$69,$FE,$41,$E7,$FF,$1E,$92,$3F,$CC,$31,$97,$FD
	.dc.b	$80,$7F,$AE,$8A,$FF,$AF,$B3,$FF,$60,$6B,$FE,$BF,$8C,$5F,$F9,$AB
	.dc.b	$8B,$FD,$DD,$F1,$1F,$C6,$B2,$C7,$22,$31,$97,$FA,$0A,$B6,$FF,$30
	.dc.b	$3F,$D4,$73,$B7,$07,$FA,$71,$FF,$62,$DF,$43,$F9,$39,$63,$FD,$8A
	.dc.b	$6D,$CB,$EA,$C7,$D0,$4C,$BD,$64,$9F,$F6,$C4,$FF,$B8,$6D,$6F,$83
	.dc.b	$FF,$7B,$92,$7F,$DD,$4F,$FB,$F9,$F9,$07,$FB,$D2,$B9,$B4,$F1,$5F
	.dc.b	$9A,$FA,$D7,$3A,$FF,$79,$36,$37,$1C,$AB,$FD,$E7,$1E,$4F,$88,$F8
	.dc.b	$2F,$CB,$C1,$18,$9F,$96,$D7,$E5,$36,$7D,$BD,$F6,$8D,$25,$7F,$BD
	.dc.b	$E7,$B7,$C4,$6C,$91,$93,$EC,$BF,$0F,$6F,$C2,$C7,$A8,$F3,$57,$4C
	.dc.b	$7F,$DA,$F1,$F8,$08,$C5,$EE,$3E,$41,$0F,$EC,$B5,$D1,$FF,$65,$F1
	.dc.b	$7F,$BF,$FB,$3F,$E0,$0F,$1F,$FB,$32,$7F,$E0,$0C,$E7,$1D,$CE,$6B
	.dc.b	$7C,$BF,$DC,$1B,$7C,$27,$55,$C2,$B4,$FF,$B0,$E5,$87,$FB,$A5,$F2
	.dc.b	$F0,$7F,$B6,$5F,$F6,$49,$8E,$0F,$F6,$DC,$7F,$B6,$A3,$F4,$BA,$7E
	.dc.b	$8D,$F8,$47,$FB,$BD,$C3,$FA,$17,$FD,$F1,$3C,$89,$1D,$B4,$FC,$BE
	.dc.b	$BC,$1B,$1F,$EF,$98,$C6,$C6,$DC,$9F,$01,$0F,$AA,$7E,$20,$FF,$0A
	.dc.b	$D7,$E3,$F7,$E6,$48,$C9,$F1,$72,$F8,$8F,$C6,$CB,$E0,$3F,$97,$E0
	.dc.b	$F1,$91,$E2,$7E,$FD,$F7,$BF,$28,$E5,$9F,$C0,$7F,$DE,$C4,$3F,$84
	.dc.b	$C7,$37,$F5,$67,$4C,$3F,$5E,$F6,$39,$9D,$6F,$CE,$4B,$FB,$51,$FA
	.dc.b	$DD,$3F,$C8,$4F,$F0,$D3,$FC,$89,$5F,$C1,$A3,$2D,$8C,$E9,$FC,$6F
	.dc.b	$CA,$FE,$47,$C5,$FE,$3E,$6C,$9A,$78,$F0,$F0,$C7,$CB,$6B,$F8,$B3
	.dc.b	$59,$FF,$BF,$23,$FD,$F5,$1F,$CF,$1A,$FF,$93,$0F,$93,$BF,$F9,$1D
	.dc.b	$87,$F9,$1C,$A7,$F9,$1C,$A7,$F8,$E0,$A7,$F9,$54,$BC,$BF,$CD,$9E
	.dc.b	$07,$32,$18,$21,$CC,$C9,$4C,$9D,$4C,$94,$CB,$DD,$E0,$43,$45,$74
	.dc.b	$E0,$8E,$09,$C9,$FE,$F7,$24,$7D,$E8,$CB,$63,$2F,$C0,$6B,$E6,$8D
	.dc.b	$B9,$38,$10,$C1,$0E,$11,$C9,$F6,$B5,$F5,$35,$F4,$B6,$3B,$C7,$23
	.dc.b	$B1,$4E,$C7,$43,$25,$32,$76,$32,$53,$27,$80,$C9,$4C,$BE,$27,$A1
	.dc.b	$E1,$21,$82,$1D,$4C,$94,$C9,$D4,$C9,$4C,$BD,$DE,$04,$34,$57,$4E
	.dc.b	$08,$E0,$91,$F2,$C9,$1E,$E8,$CB,$63,$2F,$C0,$6B,$E6,$8D,$B9,$38
	.dc.b	$10,$C1,$0E,$11,$C9,$F6,$B5,$F5,$35,$F4,$B6,$3B,$C7,$23,$B1,$4E
	.dc.b	$C7,$43,$25,$32,$76,$32,$53,$27,$80,$C9,$4C,$BE,$27,$A1,$E1,$21
	.dc.b	$83,$F7,$BC,$9F,$EF,$77,$FD,$BE,$D4,$D3,$10,$F9,$5F,$F1,$DC,$FF
	.dc.b	$BA,$9A,$FF,$B6,$DA,$FF,$BF,$DB,$1C,$63,$FD,$FD,$18,$23,$82,$6D
	.dc.b	$FE,$FA,$8C,$B6,$32,$7F,$C0,$08,$CB,$4C,$9E,$52,$9E,$52,$9E,$52
	.dc.b	$CF,$96,$7B,$12,$3A,$35,$E8,$D7,$A3,$5E,$8D,$8E,$91,$C8,$87,$22
	.dc.b	$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA,$70
	.dc.b	$47,$04,$9F,$9F,$1D,$5A,$F2,$6B,$C5,$B1,$C6,$3E,$7C,$60,$8E,$09
	.dc.b	$3F,$26,$32,$D8,$C9,$F1,$23,$2D,$32,$79,$4A,$79,$4A,$79,$4B,$3E
	.dc.b	$59,$EC,$48,$E8,$D7,$A3,$5E,$8D,$7A,$36,$3A,$47,$22,$1C,$8A,$72
	.dc.b	$29,$C8,$AF,$27,$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9,$C1,$1C
	.dc.b	$12,$7E,$7C,$75,$6B,$C9,$AF,$16,$C7,$18,$F9,$F1,$82,$38,$24,$FC
	.dc.b	$98,$CB,$63,$27,$C4,$8C,$B4,$C9,$E5,$29,$E5,$29,$EE,$DF,$FD,$E0
	.dc.b	$FF,$8D,$5A,$FF,$A1,$F9,$BF,$DE,$CF,$F8,$E7,$A1,$E0,$21,$3F,$E3
	.dc.b	$D7,$13,$A7,$26,$5C,$46,$E1,$18,$7F,$D6,$07,$FA,$5F,$0F,$F6,$87
	.dc.b	$C0,$BF,$7E,$7A,$7F,$CC,$23,$0D,$8C,$D7,$FB,$CA,$72,$FF,$B7,$F9
	.dc.b	$1F,$13,$E9,$F9,$3F,$47,$3F,$FB,$BA,$31,$5F,$ED,$B6,$BF,$E9,$0A
	.dc.b	$E2,$AB,$8C,$3E,$27,$FA,$1D,$FF,$BE,$1B,$19,$AF,$14,$7F,$46,$DF
	.dc.b	$FD,$95,$1E,$F3,$FE,$00,$4B,$E1,$38,$92,$3D,$4F,$4F,$53,$FD,$D0
	.dc.b	$48,$FF,$71,$0C,$14,$9F,$F7,$27,$3B,$A1,$C1,$FE,$FD,$C3,$FD,$3A
	.dc.b	$FF,$BF,$FD,$57,$E5,$6F,$8D,$FF,$7D,$46,$1B,$19,$AF,$0C,$E5,$F1
	.dc.b	$E7,$E2,$CB,$CD,$87,$FA,$F2,$31,$3A,$CF,$CC,$FA,$8F,$19,$F8,$42
	.dc.b	$9F,$4E,$B9,$CF,$FA,$34,$E7,$5D,$6B,$9B,$7C,$7F,$CC,$36,$FC,$11
	.dc.b	$8A,$E5,$48,$C4,$FE,$05,$AF,$CD,$8D,$3C,$89,$5F,$ED,$87,$CE,$7D
	.dc.b	$D3,$F5,$DA,$7D,$18,$FD,$16,$C8,$C4,$F8,$AF,$BE,$1F,$69,$B8,$78
	.dc.b	$30,$F5,$1D,$F0,$FC,$5D,$F8,$08,$E0,$CF,$53,$C3,$76,$FE,$EF,$C4
	.dc.b	$FC,$1D,$FF,$89,$27,$2A,$32,$FF,$89,$A3,$31,$8A,$E9,$5C,$F0,$FF
	.dc.b	$60,$46,$2F,$94,$62,$F8,$B6,$B8,$4F,$85,$AF,$FB,$98,$FF,$77,$1F
	.dc.b	$ED,$12,$99,$3E,$6F,$0F,$FB,$83,$3F,$C2,$47,$D9,$3F,$DE,$59,$FD
	.dc.b	$87,$F0,$47,$C9,$D7,$FD,$BA,$F3,$3B,$1C,$4D,$EB,$F5,$5E,$75,$D9
	.dc.b	$FC,$23,$63,$EC,$46,$88,$7E,$57,$6D,$5F,$E8,$23,$07,$9A,$3A,$3E
	.dc.b	$32,$6D,$F8,$7C,$7F,$95,$47,$9E,$FF,$56,$F2,$23,$F9,$CA,$F3,$6F
	.dc.b	$FB,$59,$FE,$0C,$7F,$3B,$71,$3F,$EF,$76,$C6,$78,$3F,$7C,$FC,$7B
	.dc.b	$F0,$E1,$E3,$8F,$CD,$E7,$FC,$35,$B3,$FC,$FF,$0F,$29,$E3,$8F,$71
	.dc.b	$0F,$DC,$71,$7B,$23,$F8,$CB,$73,$FE,$49,$3C,$DA,$F6,$21,$F8,$69
	.dc.b	$FE,$64,$EF,$8F,$91,$1E,$5C,$BF,$D6,$FC,$AF,$E7,$6D,$8F,$E8,$42
	.dc.b	$32,$D3,$FB,$10,$E7,$FE,$9E,$FF,$63,$6B,$A6,$BF,$DA,$5A,$FE,$38
	.dc.b	$87,$DE,$CF,$E5,$3F,$DA,$7B,$3F,$B8,$71,$7F,$69,$EB,$FE,$DC,$D8
	.dc.b	$CC,$EC,$FD,$C9,$FA,$48,$FA,$64,$7F,$BC,$6F,$F8,$B8,$FF,$47,$44
	.dc.b	$3F,$D1,$85,$3F,$D1,$85,$3F,$D1,$85,$3F,$D1,$85,$EE,$FF,$50,$37
	.dc.b	$BB,$F2,$EE,$8E,$A4,$3A,$94,$EA,$53,$81,$A2,$1A,$29,$A3,$04,$30
	.dc.b	$48,$F2,$BD,$5A,$F2,$6B,$C5,$AF,$18,$F5,$C6,$08,$E0,$91,$FB,$68
	.dc.b	$CB,$63,$2F,$C5,$30,$43,$04,$3C,$CF,$06,$9B,$29,$B3,$D4,$ED,$A6
	.dc.b	$CA,$6C,$F7,$3B,$69,$B2,$9B,$3E,$11,$92,$99,$8F,$7B,$63,$A3,$A3
	.dc.b	$CE,$43,$C6,$53,$B9,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$23,$CA,$F5
	.dc.b	$6B,$C9,$AF,$16,$BC,$63,$D7,$18,$23,$82,$47,$C5,$8C,$B6,$32,$FC
	.dc.b	$53,$04,$30,$43,$CC,$F0,$69,$B2,$9B,$3D,$4E,$DA,$6C,$A6,$CF,$73
	.dc.b	$B6,$9B,$29,$B3,$E1,$19,$29,$98,$F7,$B6,$3A,$3A,$3C,$E4,$3E,$CF
	.dc.b	$47,$F8,$1D,$FF,$46,$35,$FF,$03,$1F,$E8,$97,$0F,$FA,$4E,$7F,$CE
	.dc.b	$8F,$F8,$58,$C7,$5F,$F9,$01,$FF,$4B,$46,$BA,$3F,$D2,$4D,$8C,$CE
	.dc.b	$DF,$E7,$B1,$96,$C6,$5E,$A7,$71,$1B,$73,$3D,$A7,$84,$69,$B3,$B8
	.dc.b	$C1,$CD,$C1,$23,$64,$8D,$36,$7B,$46,$08,$E0,$8F,$B8,$C9,$5C,$BE
	.dc.b	$E3,$25,$72,$FB,$0E,$04,$34,$53,$46,$08,$60,$87,$C0,$7A,$B4,$E4
	.dc.b	$53,$89,$4E,$33,$F1,$5F,$39,$89,$F8,$8F,$28,$D4,$F6,$6C,$66,$76
	.dc.b	$F6,$8C,$B6,$32,$F5,$3B,$9E,$32,$47,$07,$53,$DE,$78,$C6,$9B,$3B
	.dc.b	$8C,$1D,$1C,$12,$36,$48,$D3,$67,$B4,$60,$8E,$08,$FB,$CC,$95,$CB
	.dc.b	$EF,$32,$57,$2F,$B0,$E0,$43,$45,$34,$60,$86,$08,$7C,$07,$AB,$4E
	.dc.b	$45,$38,$94,$E3,$3F,$15,$F3,$98,$9F,$88,$F2,$8D,$4F,$66,$C6,$67
	.dc.b	$6F,$68,$CB,$63,$2F,$53,$B9,$E3,$24,$70,$75,$3D,$E7,$8C,$69,$B3
	.dc.b	$B8,$C1,$D1,$C1,$23,$64,$8D,$36,$7F,$53,$8F,$FA,$17,$BB,$FC,$E2
	.dc.b	$4A,$FF,$1F,$EB,$FF,$00,$63,$FC,$EB,$E7,$C9,$CF,$32,$31,$1F,$E5
	.dc.b	$7F,$01,$E2,$A3,$47,$F9,$76,$32,$E0,$E1,$1B,$9F,$F3,$24,$E9,$FF
	.dc.b	$42,$E5,$2B,$C2,$D7,$B7,$33,$9D,$F2,$5F,$CC,$63,$0D,$8C,$CF,$F4
	.dc.b	$19,$CB,$61,$F1,$BE,$9F,$A2,$FF,$3C,$CF,$FB,$EE,$30,$FF,$9F,$2B
	.dc.b	$C1,$18,$8F,$1D,$72,$55,$CA,$9E,$E7,$FA,$87,$2C,$36,$BA,$35,$F6
	.dc.b	$B5,$F9,$9C,$1F,$EF,$7E,$06,$FF,$D3,$63,$FD,$81,$AF,$4F,$9D,$F8
	.dc.b	$58,$1A,$4C,$3E,$14,$3E,$7F,$EC,$7C,$7F,$91,$B3,$E4,$6B,$E4,$6B
	.dc.b	$E4,$CB,$C7,$87,$D5,$8C,$65,$E5,$9C,$C7,$BB,$DE,$F2,$91,$F1,$BE
	.dc.b	$8F,$63,$F1,$5F,$47,$FE,$D2,$96,$FF,$2A,$72,$FC,$86,$C7,$C8,$6B
	.dc.b	$E2,$C7,$C8,$B6,$F8,$EF,$E3,$0F,$A3,$D9,$C2,$FE,$BB,$76,$34,$C7
	.dc.b	$A3,$F9,$99,$FB,$2D,$AF,$B2,$D7,$D3,$48,$C4,$FE,$13,$E7,$AC,$9A
	.dc.b	$4A,$FD,$84,$F9,$DF,$F6,$6E,$1F,$83,$DB,$ED,$FC,$EF,$22,$4F,$03
	.dc.b	$F1,$6F,$8D,$F8,$5B,$FD,$47,$9B,$5F,$90,$77,$29,$A9,$FB,$84,$D3
	.dc.b	$FC,$61,$E5,$3C,$0F,$E0,$5E,$5A,$7F,$C6,$AD,$BF,$E0,$DE,$02,$F8
	.dc.b	$42,$74,$A7,$57,$D2,$F9,$92,$79,$DF,$2B,$E3,$7C,$27,$A5,$7A,$5F
	.dc.b	$DE,$9F,$EE,$BC,$2B,$4C,$E6,$CB,$FD,$DF,$1C,$4F,$E4,$67,$EE,$B3
	.dc.b	$FE,$41,$ED,$7E,$5F,$5F,$DF,$C6,$1F,$DE,$B6,$7E,$98,$9F,$CF,$3E
	.dc.b	$13,$A9,$0F,$E0,$7B,$71,$7F,$16,$DA,$FC,$54,$73,$7F,$85,$C6,$77
	.dc.b	$FC,$84,$E4,$FC,$A9,$4F,$CA,$F0,$6C,$E1,$C3,$FC,$EF,$E9,$3D,$95
	.dc.b	$FA,$2A,$46,$27,$58,$7E,$AA,$72,$FA,$5F,$E0,$67,$CC,$CF,$E5,$B8
	.dc.b	$3F,$74,$70,$3F,$74,$53,$F5,$3C,$7D,$9C,$DF,$93,$97,$EC,$DF,$F2
	.dc.b	$E6,$E2,$69,$F7,$AB,$E7,$1F,$6C,$87,$ED,$B9,$BF,$AD,$78,$BF,$C8
	.dc.b	$E0,$D2,$3F,$CE,$3A,$7B,$91,$FC,$6E,$1F,$D0,$23,$C1,$C3,$F8,$BE
	.dc.b	$2F,$C7,$B6,$BE,$24,$63,$5F,$8B,$7F,$D2,$23,$F1,$B3,$FD,$19,$AF
	.dc.b	$F4,$66,$BF,$C3,$9B,$1F,$D5,$A1,$F4,$BF,$D4,$FB,$7F,$AB,$37,$BB
	.dc.b	$FB,$C3,$67,$67,$F5,$F9,$FE,$D2,$F1,$EF,$F9,$F1,$88,$D4,$7B,$B1
	.dc.b	$FB,$B1,$F3,$8F,$BC,$7F,$80,$0F,$D0,$1F,$E0,$5F,$07,$D3,$AB,$C1
	.dc.b	$FF,$03,$F3,$FF,$81,$27,$DB,$5F,$C6,$35,$E2,$FF,$13,$9F,$92,$47
	.dc.b	$FC,$2F,$DB,$F3,$08,$E0,$FF,$4F,$1F,$EA,$42,$18,$21,$FE,$9F,$32
	.dc.b	$53,$27,$FA,$F8,$C9,$4C,$9F,$EB,$83,$25,$33,$E2,$7F,$B5,$1C,$B5
	.dc.b	$CF,$89,$F9,$78,$C4,$78,$48,$77,$29,$D4,$AF,$07,$44,$34,$57,$4E
	.dc.b	$08,$E0,$91,$E3,$24,$7F,$28,$8C,$B6,$32,$FB,$1A,$FB,$0D,$10,$D1
	.dc.b	$4D,$18,$21,$82,$1E,$E7,$AB,$4E,$45,$38,$94,$E2,$7C,$17,$8B,$4E
	.dc.b	$25,$38,$94,$E2,$7C,$67,$8B,$4E,$25,$38,$94,$E3,$1F,$21,$CB,$5C
	.dc.b	$C7,$48,$C4,$7A,$48,$78,$CA,$75,$2B,$C1,$D1,$0D,$15,$D3,$82,$38
	.dc.b	$24,$78,$C9,$1E,$E8,$CB,$63,$2F,$B1,$AF,$B0,$D1,$0D,$14,$D1,$82
	.dc.b	$18,$21,$EE,$7A,$B4,$E4,$53,$89,$4E,$27,$C1,$78,$B4,$E2,$53,$89
	.dc.b	$4E,$27,$C6,$78,$B4,$E2,$53,$F0,$3D,$BF,$E1,$37,$FD,$1A,$D7,$FD
	.dc.b	$50,$7F,$87,$FB,$FF,$C4,$6E,$9A,$EA,$9F,$3B,$FE,$AF,$EF,$FF,$0D
	.dc.b	$C6,$27,$FC,$52,$7F,$AC,$DC,$46,$A7,$84,$FF,$A7,$A7,$2D,$8C,$BD
	.dc.b	$DA,$F3,$6A,$8E,$2E,$CF,$09,$98,$DC,$6A,$39,$1A,$70,$EC,$C4,$70
	.dc.b	$23,$82,$4F,$AD,$A7,$A4,$A7,$8D,$D1,$1D,$35,$D1,$82,$18,$21,$E8
	.dc.b	$78,$34,$D9,$4D,$9E,$B7,$6D,$36,$53,$73,$F0,$89,$3F,$02,$78,$C6
	.dc.b	$27,$DA,$77,$21,$CC,$AE,$E3,$8C,$F3,$9F,$0C,$E5,$B1,$97,$C6,$D7
	.dc.b	$9B,$5E,$C4,$8E,$4F,$03,$CA,$66,$38,$46,$A3,$99,$A7,$0E,$CC,$47
	.dc.b	$02,$38,$24,$FB,$1A,$7A,$8A,$78,$DD,$11,$D3,$5D,$18,$21,$82,$1E
	.dc.b	$87,$83,$4D,$94,$D9,$EB,$76,$D3,$65,$37,$3F,$08,$93,$F0,$27,$8C
	.dc.b	$62,$7D,$A7,$72,$1C,$CA,$EE,$38,$CF,$39,$F0,$CE,$5B,$19,$7C,$6D
	.dc.b	$79,$B5,$EC,$48,$E4,$F0,$3C,$A6,$63,$84,$6A,$39,$9A,$70,$EC,$C4
	.dc.b	$70,$23,$82,$4F,$F9,$A3,$C5,$91,$FC,$A8,$FF,$23,$38,$7F,$96,$FD
	.dc.b	$33,$CB,$42,$7F,$D8,$73,$98,$C1,$CD,$C4,$6A,$78,$B6,$5E,$35,$23
	.dc.b	$C6,$7B,$C7,$97,$A1,$DB,$FE,$E1,$AE,$27,$F4,$28,$CE,$9C,$23,$13
	.dc.b	$BA,$E4,$D8,$FF,$7B,$1F,$E6,$5C,$BF,$D1,$64,$78,$FC,$FF,$A1,$A7
	.dc.b	$B8,$F8,$78,$79,$23,$51,$8A,$46,$21,$3E,$AF,$A0,$F9,$7F,$4F,$E4
	.dc.b	$6C,$FA,$1A,$F3,$6B,$CA,$FF,$D0,$0F,$F5,$8A,$F8,$BC,$DF,$D4,$1B
	.dc.b	$19,$9D,$E3,$ED,$8F,$ED,$3D,$BF,$EA,$D8,$C4,$7D,$E8,$C3,$F2,$BC
	.dc.b	$BE,$94,$69,$23,$B7,$CF,$FB,$F0,$FA,$ED,$A3,$4F,$89,$3F,$44,$FA
	.dc.b	$77,$E0,$7C,$BD,$6E,$9F,$57,$CF,$73,$24,$FD,$33,$E2,$3D,$FE,$73
	.dc.b	$F2,$FE,$6F,$FB,$C9,$B9,$FE,$11,$B5,$F5,$DA,$FC,$1C,$3C,$1A,$FC
	.dc.b	$48,$F3,$1E,$4C,$7E,$67,$C8,$FE,$93,$EA,$3F,$DF,$ED,$C7,$F2,$2D
	.dc.b	$AF,$C2,$35,$FB,$5F,$0B,$F9,$E9,$FC,$BB,$63,$F1,$4D,$76,$7E,$A2
	.dc.b	$3F,$3B,$DB,$97,$F4,$AD,$E0,$34,$E5,$3E,$A3,$DD,$8F,$F8,$43,$7F
	.dc.b	$A1,$18,$CF,$F4,$FD,$7C,$C9,$1E,$13,$F3,$CF,$49,$E3,$1D,$48,$E0
	.dc.b	$9A,$FC,$C8,$FB,$ED,$7F,$C7,$7D,$AE,$97,$77,$F1,$0F,$3D,$96,$A7
	.dc.b	$36,$9C,$BF,$C5,$AF,$95,$F1,$BE,$0D,$9D,$5F,$2B,$E3,$1F,$E4,$B9
	.dc.b	$D3,$4E,$D5,$C9,$A7,$FB,$D8,$87,$FB,$E4,$98,$B1,$FF,$7E,$B7,$3F
	.dc.b	$E6,$13,$FB,$BC,$FF,$96,$F8,$1F,$C1,$A3,$E6,$87,$0E,$CC,$6F,$FC
	.dc.b	$7A,$78,$BF,$CE,$EF,$B4,$F6,$6B,$9C,$3C,$67,$EB,$5F,$E4,$BB,$72
	.dc.b	$7F,$8D,$35,$FE,$20,$DA,$FC,$7C,$78,$8F,$CB,$94,$FC,$BE,$1E,$88
	.dc.b	$FE,$73,$C1,$C7,$80,$DE,$1F,$E9,$BF,$53,$FA,$AC,$0D,$25,$FA,$75
	.dc.b	$F9,$33,$F3,$B3,$F5,$3F,$BD,$1E,$1F,$F0,$16,$D6,$63,$07,$97,$27
	.dc.b	$3F,$F5,$EE,$1F,$E8,$ED,$BF,$E9,$6D,$7F,$9F,$35,$FC,$0E,$FF,$39
	.dc.b	$FE,$0C,$7D,$1C,$7C,$AF,$F4,$5E,$8E,$19,$7F,$26,$8F,$F4,$BB,$70
	.dc.b	$FE,$AF,$3A,$C7,$F9,$85,$7E,$22,$BF,$B1,$F4,$FC,$26,$C6,$67,$6F
	.dc.b	$87,$4F,$8F,$97,$96,$3C,$F9,$79,$1F,$E9,$9D,$5C,$EF,$FA,$6C,$7F
	.dc.b	$68,$8C,$19,$EB,$FF,$01,$FD,$57,$F5,$99,$34,$8F,$F6,$8D,$7F,$B8
	.dc.b	$F7,$7A,$C9,$1F,$EB,$63,$ED,$67,$E6,$CB,$C9,$A7,$F7,$C9,$FD,$11
	.dc.b	$FD,$F3,$A7,$F4,$ED,$32,$7F,$71,$72,$FF,$86,$BC,$0A,$FF,$86,$BC
	.dc.b	$2F,$F0,$C7,$3F,$F8,$4C,$91,$FE,$11,$3C,$91,$FE,$1D,$CF,$FC,$30
	.dc.b	$48,$FC,$93,$F3,$48,$E0,$FF,$56,$1F,$EA,$F2,$18,$21,$FE,$C1,$32
	.dc.b	$53,$27,$FB,$00,$C9,$4C,$9F,$EC,$13,$25,$33,$E1,$FF,$B7,$5C,$B5
	.dc.b	$CF,$57,$EA,$9E,$07,$84,$87,$72,$9A,$30,$43,$04,$8E,$CF,$06,$BB
	.dc.b	$6B,$B8,$F1,$46,$08,$E0,$91,$FC,$02,$32,$D8,$CB,$EE,$6B,$EE,$34
	.dc.b	$43,$45,$34,$60,$86,$08,$7A,$DE,$AD,$39,$14,$E2,$53,$89,$F0,$5E
	.dc.b	$2D,$38,$94,$E2,$53,$89,$F1,$CD,$94,$D9,$4D,$C7,$C6,$72,$D7,$31
	.dc.b	$E2,$78,$1E,$72,$1E,$12,$9A,$30,$43,$04,$8E,$CF,$06,$BB,$6B,$B8
	.dc.b	$F1,$46,$08,$E0,$91,$F0,$23,$2D,$8C,$BE,$E6,$BE,$E3,$44,$34,$53
	.dc.b	$46,$08,$60,$87,$AD,$EA,$D3,$91,$4E,$25,$38,$9F,$05,$E2,$D3,$89
	.dc.b	$4E,$25,$38,$9F,$1C,$D9,$4F,$F0,$5F,$6F,$F8,$81,$B9,$7F,$2D,$7F
	.dc.b	$C3,$47,$FA,$FE,$34,$FF,$B0,$7F,$77,$7C,$37,$FA,$FB,$BB,$FC,$2B
	.dc.b	$18,$9F,$F1,$29,$FE,$C2,$71,$1A,$9E,$13,$CD,$AF,$FB,$20,$4F,$07
	.dc.b	$47,$43,$31,$C1,$EC,$D7,$26,$E3,$93,$D5,$AF,$03,$51,$C6,$33,$1C
	.dc.b	$A7,$D0,$E8,$8E,$08,$F9,$9A,$F7,$9E,$84,$78,$9B,$21,$B3,$25,$8E
	.dc.b	$6F,$9D,$AF,$94,$E2,$64,$A6,$4E,$46,$27,$DA,$49,$F2,$4F,$48,$C4
	.dc.b	$F8,$8E,$87,$22,$18,$21,$DC,$AF,$28,$E9,$3D,$A7,$C4,$D7,$99,$C4
	.dc.b	$93,$C5,$D9,$E3,$33,$1C,$5E,$ED,$72,$6E,$39,$3D,$9A,$F0,$35,$1C
	.dc.b	$63,$31,$CA,$7D,$2E,$88,$E0,$8F,$99,$AF,$79,$E8,$47,$89,$B2,$1B
	.dc.b	$32,$58,$E6,$F9,$DA,$F9,$4E,$26,$4A,$64,$E4,$62,$7D,$A4,$9F,$24
	.dc.b	$F4,$8C,$4F,$88,$E8,$72,$21,$82,$1D,$CA,$F2,$8E,$93,$DA,$7C,$4D
	.dc.b	$79,$9C,$49,$3C,$5D,$9E,$33,$31,$C5,$EE,$D7,$26,$E3,$93,$D9,$AF
	.dc.b	$03,$51,$C6,$33,$1C,$A7,$D2,$E8,$F1,$73,$7F,$AD,$7C,$5F,$F3,$41
	.dc.b	$23,$FD,$20,$FF,$AE,$35,$78,$E9,$4E,$63,$0D,$8C,$9F,$C8,$C4,$F2
	.dc.b	$72,$7F,$27,$5E,$57,$28,$E1,$18,$76,$7F,$96,$E3,$4F,$16,$B9,$1E
	.dc.b	$3F,$A6,$3D,$0D,$8C,$9D,$A3,$2F,$FB,$DB,$97,$F9,$B6,$3F,$EF,$A6
	.dc.b	$F2,$FF,$BE,$67,$E1,$4E,$73,$FF,$48,$12,$39,$35,$F9,$1C,$AF,$C7
	.dc.b	$BF,$A3,$78,$FF,$16,$31,$3F,$36,$31,$3A,$8F,$E9,$93,$A5,$3A,$F8
	.dc.b	$AF,$4C,$76,$A4,$62,$15,$E1,$8C,$36,$3E,$19,$FD,$67,$0F,$63,$EA
	.dc.b	$21,$9C,$7E,$12,$8F,$F7,$F6,$3E,$8E,$1F,$C3,$7D,$6D,$A3,$48,$F4
	.dc.b	$3E,$5D,$F4,$BF,$0C,$77,$D3,$E8,$93,$0F,$8B,$F0,$3F,$93,$F2,$7F
	.dc.b	$DE,$1A,$A3,$1A,$7C,$88,$C6,$0A,$DD,$6A,$3B,$5F,$D7,$7C,$78,$77
	.dc.b	$7F,$35,$F3,$0F,$2F,$F4,$5F,$29,$F3,$27,$FC,$01,$39,$51,$97,$F5
	.dc.b	$11,$98,$C4,$23,$19,$F9,$FE,$97,$FC,$28,$D9,$63,$EE,$9C,$C7,$DF
	.dc.b	$6B,$F7,$B4,$7D,$2F,$E8,$2F,$E1,$36,$7F,$0C,$D7,$83,$5F,$B9,$BF
	.dc.b	$D3,$26,$9F,$0E,$3C,$07,$92,$BC,$D8,$F9,$63,$06,$9F,$7E,$FF,$72
	.dc.b	$3E,$6C,$7C,$F7,$BD,$75,$72,$7F,$09,$9F,$C6,$BF,$E3,$BE,$B7,$77
	.dc.b	$F1,$3F,$3B,$E2,$CF,$F1,$5E,$DF,$F2,$24,$62,$11,$8C,$7F,$C6,$CD
	.dc.b	$8E,$17,$D1,$57,$19,$D4,$72,$9D,$4E,$5B,$5F,$97,$9F,$E4,$07,$FB
	.dc.b	$F8,$B8,$33,$65,$FF,$00,$49,$3A,$8C,$BF,$C0,$B6,$FE,$02,$FD,$03
	.dc.b	$E7,$13,$4F,$8D,$3C,$22,$D7,$46,$9F,$C3,$35,$FE,$75,$E0,$7E,$A1
	.dc.b	$B7,$F8,$69,$4E,$5B,$1F,$93,$7C,$07,$8D,$C1,$FA,$6E,$2F,$E8,$D7
	.dc.b	$F9,$EE,$1F,$D2,$36,$33,$3B,$AF,$86,$F9,$98,$7C,$B2,$9F,$9D,$E3
	.dc.b	$FE,$A5,$E5,$FE,$EA,$0D,$24,$F0,$8F,$A4,$5D,$7E,$96,$1F,$16,$3B
	.dc.b	$1F,$CA,$79,$FF,$AF,$7C,$1F,$F1,$92,$47,$F0,$B8,$CB,$CF,$6F,$A0
	.dc.b	$F2,$3F,$84,$9F,$6E,$BB,$38,$3D,$78,$7F,$31,$D7,$F9,$D4,$7D,$03
	.dc.b	$F8,$C7,$4F,$D6,$69,$93,$F9,$F1,$FD,$CF,$CC,$FE,$8E,$D3,$F9,$01
	.dc.b	$4D,$87,$C8,$C3,$D5,$7E,$67,$E3,$9E,$B2,$9F,$DB,$6B,$FB,$51,$1F
	.dc.b	$ED,$47,$94,$FE,$E5,$CD,$FD,$BC,$8E,$9F,$F0,$31,$1F,$EF,$FE,$07
	.dc.b	$F8,$36,$31,$1F,$D0,$5F,$94,$47,$F3,$AF,$FA,$71,$87,$FA,$B8,$A7
	.dc.b	$FA,$4C,$A7,$FA,$9C,$A7,$FA,$B8,$BE,$0F,$FB,$05,$BA,$7E,$E9,$E0
	.dc.b	$75,$21,$D4,$A6,$8C,$10,$C1,$0F,$01,$C0,$A6,$CA,$6E,$3C,$8E,$DA
	.dc.b	$ED,$AE,$E3,$D0,$F2,$24,$69,$B1,$A8,$C1,$1C,$12,$3E,$14,$65,$B1
	.dc.b	$97,$E3,$98,$21,$82,$1E,$97,$83,$4D,$94,$D9,$ED,$76,$D3,$65,$36
	.dc.b	$7C,$13,$25,$32,$7C,$03,$25,$33,$1E,$D6,$C7,$81,$E4,$79,$48,$78
	.dc.b	$4A,$68,$C1,$0C,$10,$F0,$1C,$0A,$6C,$A6,$E3,$C8,$ED,$AE,$DA,$EE
	.dc.b	$3D,$0F,$22,$46,$9B,$1A,$8C,$11,$C1,$23,$E3,$C6,$5B,$19,$7E,$39
	.dc.b	$82,$18,$21,$E9,$78,$34,$D9,$4D,$9E,$D7,$6D,$36,$53,$67,$C1,$32
	.dc.b	$53,$27,$C0,$32,$53,$31,$ED,$6C,$78,$1F,$DC,$74,$39,$FF,$C0,$D1
	.dc.b	$FE,$13,$BF,$E0,$D3,$96,$CE,$A7,$FD,$50,$2F,$FC,$A7,$CF,$FE,$06
	.dc.b	$6C,$7F,$A9,$DA,$F0,$6B,$FE,$B1,$6B,$CF,$8F,$FD,$69,$3C,$DB,$1C
	.dc.b	$DA,$F3,$6B,$CD,$AF,$33,$FC,$50,$64,$A6,$63,$B3,$89,$F1,$92,$79
	.dc.b	$B6,$35,$18,$23,$82,$4F,$42,$3E,$93,$25,$72,$FA,$CC,$95,$CB,$E9
	.dc.b	$6B,$E9,$39,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$27
	.dc.b	$E2,$47,$91,$B1,$DD,$AF,$46,$BC,$DA,$F3,$9F,$97,$3C,$DB,$1C,$DA
	.dc.b	$F3,$6B,$CD,$AF,$33,$E5,$99,$29,$98,$EC,$E2,$7C,$64,$9E,$6D,$8D
	.dc.b	$46,$08,$E0,$93,$D0,$8F,$A4,$C9,$5C,$BE,$B3,$25,$72,$FA,$5A,$FA
	.dc.b	$4E,$44,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$49,$F8,$91
	.dc.b	$E4,$6C,$77,$6B,$D1,$AF,$36,$BC,$E7,$E5,$CF,$36,$C7,$36,$BC,$DA
	.dc.b	$F3,$6B,$CC,$F9,$66,$4A,$66,$3B,$38,$9F,$19,$27,$F1,$7C,$5F,$E6
	.dc.b	$EF,$07,$FC,$C4,$4A,$FF,$4A,$BF,$EA,$6F,$05,$E2,$25,$39,$8C,$36
	.dc.b	$32,$7F,$2C,$13,$C9,$C9,$FC,$C1,$78,$3C,$A3,$FC,$AF,$3E,$18,$C4
	.dc.b	$6D,$AF,$8F,$C4,$79,$58,$84,$61,$EE,$6E,$35,$19,$AF,$F6,$F6,$3F
	.dc.b	$EE,$1A,$FF,$6F,$E9,$EB,$23,$EB,$C7,$D5,$18,$8F,$E9,$33,$C4,$E5
	.dc.b	$EF,$BC,$FF,$23,$67,$C6,$1C,$61,$FE,$D4,$6C,$7F,$54,$E8,$78,$2F
	.dc.b	$A5,$FA,$5F,$5D,$F3,$67,$31,$8A,$E1,$5B,$55,$B8,$C3,$7E,$8B,$FA
	.dc.b	$7D,$F6,$6C,$9A,$7F,$52,$CF,$DD,$DB,$FE,$E6,$D7,$FD,$54,$7A,$75
	.dc.b	$FE,$AF,$A7,$BE,$31,$9F,$F5,$68,$C3,$FD,$C3,$B9,$C8,$F0,$3F,$1D
	.dc.b	$E4,$7E,$26,$73,$18,$AE,$15,$B5,$5B,$8C,$3F,$73,$2F,$34,$FF,$BC
	.dc.b	$31,$FB,$FF,$23,$F9,$8D,$FF,$2D,$AA,$31,$5F,$71,$AF,$E8,$63,$4F
	.dc.b	$E8,$A7,$E7,$FC,$EF,$E9,$7E,$0D,$18,$94,$61,$FD,$54,$7D,$C9,$E0
	.dc.b	$EE,$75,$3D,$F1,$F4,$BF,$8E,$6B,$9F,$07,$FC,$03,$87,$FB,$A3,$2F
	.dc.b	$BD,$F3,$DF,$B7,$F4,$79,$E5,$F8,$A8,$C3,$F8,$66,$E1,$E4,$E4,$46
	.dc.b	$23,$D7,$0F,$9F,$FF,$0D,$D7,$E3,$9B,$3C,$5A,$F0,$6C,$7E,$6B,$CC
	.dc.b	$FE,$15,$E7,$F9,$2D,$18,$9F,$CC,$B6,$FC,$8D,$7E,$06,$5F,$77,$4F
	.dc.b	$83,$AF,$C4,$BF,$D3,$E5,$E0,$C7,$D1,$83,$0F,$04,$E4,$F3,$C7,$F0
	.dc.b	$03,$F0,$C4,$3E,$86,$1D,$89,$3E,$13,$F8,$F4,$3C,$3F,$F2,$72,$F3
	.dc.b	$7C,$0F,$F2,$98,$CC,$62,$38,$78,$6F,$13,$F9,$74,$E6,$B7,$7F,$E5
	.dc.b	$66,$DF,$3B,$E0,$F7,$9D,$46,$2B,$93,$63,$D7,$3E,$8E,$16,$AC,$BF
	.dc.b	$DD,$91,$86,$C6,$67,$83,$5F,$D4,$B5,$EA,$62,$3E,$2E,$BF,$52,$3F
	.dc.b	$42,$7E,$1C,$A7,$E1,$CA,$ED,$FD,$18,$AF,$23,$E2,$9F,$29,$1F,$E3
	.dc.b	$DB,$7F,$57,$F1,$FF,$1C,$DB,$FC,$6C,$A7,$2D,$8F,$CE,$3E,$2D,$FF
	.dc.b	$12,$FE,$20,$C6,$FF,$CD,$5F,$D4,$70,$FE,$F2,$BF,$59,$1F,$C0,$9A
	.dc.b	$7E,$7E,$35,$1F,$18,$87,$EE,$CF,$61,$FC,$DF,$93,$FB,$87,$C0,$7E
	.dc.b	$9D,$B5,$FC,$36,$31,$27,$E3,$E9,$FC,$5B,$8A,$F3,$7F,$79,$6D,$7E
	.dc.b	$3E,$FF,$90,$CE,$5F,$B1,$1E,$22,$90,$FF,$01,$75,$7F,$34,$5E,$4F
	.dc.b	$F3,$1E,$BF,$F0,$33,$63,$27,$F3,$88,$CB,$FE,$05,$EC,$FE,$C8,$FF
	.dc.b	$46,$21,$9E,$DF,$D6,$BF,$14,$F1,$14,$F9,$38,$7F,$59,$7F,$1F,$C1
	.dc.b	$FD,$65,$BD,$DC,$A7,$51,$FD,$8F,$C0,$FD,$A3,$FD,$A1,$AF,$F2,$36
	.dc.b	$BF,$C8,$9A,$FF,$79,$BF,$98,$F9,$CF,$F6,$61,$92,$E1,$FE,$01,$D7
	.dc.b	$E6,$47,$E0,$0F,$CE,$73,$7E,$A5,$F6,$1F,$E2,$FF,$0B,$FC,$50,$FF
	.dc.b	$8B,$48,$7F,$84,$FC,$6F,$8A,$F2,$E8,$FD,$13,$F1,$8C,$1F,$C5,$BC
	.dc.b	$4F,$F1,$26,$17,$C5,$F4,$CF,$FB,$8A,$72,$FF,$8E,$5A,$FF,$8C,$4F
	.dc.b	$6E,$1F,$E3,$86,$F8,$DC,$9F,$43,$5F,$F2,$0E,$36,$7F,$C8,$58,$7F
	.dc.b	$91,$C9,$3F,$C1,$DF,$DE,$11,$C1,$FE,$CF,$3F,$DB,$A4,$30,$43,$FD
	.dc.b	$A0,$64,$A6,$4F,$F7,$41,$92,$99,$3F,$DD,$C6,$4A,$67,$C7,$FF,$0B
	.dc.b	$BF,$94,$6B,$B6,$BB,$F1,$7E,$AC,$62,$3C,$64,$3C,$25,$3B,$95,$E0
	.dc.b	$E8,$86,$8A,$E9,$C1,$1C,$12,$3C,$64,$8F,$53,$5F,$53,$5F,$51,$A2
	.dc.b	$1A,$29,$A3,$04,$30,$43,$D8,$F5,$69,$C8,$A7,$12,$9C,$4F,$7B,$C5
	.dc.b	$A7,$12,$9C,$4A,$71,$3E,$23,$C5,$A7,$12,$9C,$4A,$71,$8F,$94,$ED
	.dc.b	$AE,$DA,$EE,$3B,$46,$23,$D6,$43,$CA,$53,$B9,$5E,$0E,$88,$68,$AE
	.dc.b	$9C,$11,$C1,$23,$C6,$48,$F5,$35,$F5,$35,$F5,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$3D,$8F,$56,$9C,$8A,$71,$29,$C4,$F7,$BC,$5A,$71,$29,$C4
	.dc.b	$A7,$13,$E2,$3C,$5F,$F1,$8C,$FF,$4E,$F0,$7F,$C7,$31,$FE,$D9,$6B
	.dc.b	$FE,$2C,$8F,$EA,$7F,$6C,$EB,$FF,$72,$78,$1F,$E4,$28,$C5,$7F,$8F
	.dc.b	$4F,$F7,$43,$88,$D4,$F0,$9F,$F7,$34,$E5,$B1,$97,$FD,$E0,$23,$4E
	.dc.b	$1A,$F6,$8E,$0D,$3C,$47,$17,$2D,$33,$1C,$9C,$4F,$99,$CC,$F0,$9D
	.dc.b	$11,$C1,$23,$64,$9E,$44,$7D,$A6,$4A,$E5,$F6,$47,$38,$C4,$7A,$DA
	.dc.b	$F9,$8E,$24,$38,$9D,$4A,$F0,$9F,$44,$64,$E8,$71,$29,$C0,$C1,$DC
	.dc.b	$C1,$0D,$90,$D1,$5E,$F1,$E1,$9F,$64,$F4,$9C,$B6,$32,$F5,$39,$92
	.dc.b	$36,$E5,$AF,$78,$E2,$D3,$DC,$71,$72,$D3,$31,$C9,$C4,$F9,$DC,$CF
	.dc.b	$09,$D1,$1C,$12,$36,$49,$E4,$47,$DC,$64,$AE,$5F,$64,$73,$8C,$47
	.dc.b	$AD,$AF,$98,$E2,$43,$89,$D4,$AF,$09,$F4,$46,$4E,$87,$12,$9C,$0C
	.dc.b	$1D,$CC,$10,$D9,$0D,$15,$EF,$1E,$19,$F6,$4F,$49,$CB,$63,$2F,$53
	.dc.b	$99,$23,$6E,$5A,$F7,$8E,$2D,$3D,$C7,$17,$2D,$33,$1C,$9C,$4F,$9D
	.dc.b	$CC,$F0,$EB,$FF,$62,$79,$1F,$EB,$12,$4F,$FA,$FD,$FF,$72,$3F,$EE
	.dc.b	$93,$FD,$EE,$7F,$1A,$E1,$7E,$70,$F6,$92,$FB,$3F,$E3,$EF,$FB,$4E
	.dc.b	$74,$E4,$E4,$E2,$37,$5D,$2B,$9C,$72,$E9,$75,$BD,$9F,$E4,$B1,$86
	.dc.b	$C6,$71,$FE,$71,$39,$6C,$3E,$47,$CF,$F6,$9D,$63,$12,$8C,$4E,$A3
	.dc.b	$F9,$CD,$78,$A3,$11,$FC,$F3,$5F,$F7,$76,$FF,$22,$31,$D5,$FE,$F3
	.dc.b	$8E,$DF,$43,$C2,$58,$7B,$23,$0D,$8F,$11,$E0,$C9,$3C,$63,$FD,$0D
	.dc.b	$5D,$DF,$AA,$7D,$2D,$9F,$A7,$3E,$43,$E5,$CF,$D9,$8C,$3F,$4C,$E0
	.dc.b	$E3,$7F,$A6,$D8,$F1,$7D,$1F,$DE,$F4,$3E,$43,$63,$22,$32,$D3,$D5
	.dc.b	$F4,$9E,$27,$F0,$FA,$FD,$C8,$C4,$7C,$4D,$51,$89,$D4,$F9,$CF,$79
	.dc.b	$0F,$86,$7E,$5F,$C7,$7A,$EE,$0E,$B1,$86,$C6,$67,$A4,$E5,$ED,$B3
	.dc.b	$DB,$7C,$0F,$E7,$BE,$DF,$F2,$F1,$8A,$46,$27,$51,$FA,$79,$D4,$FD
	.dc.b	$78,$F3,$9F,$86,$AF,$F0,$71,$DE,$BE,$1B,$63,$C7,$5D,$F8,$3B,$47
	.dc.b	$A3,$E1,$BF,$D7,$5F,$6D,$F8,$C8,$FF,$7E,$C6,$35,$FF,$08,$E6,$8C
	.dc.b	$4E,$BC,$5F,$DC,$DF,$18,$F9,$9C,$FF,$9C,$7F,$81,$1E,$81,$9F,$C4
	.dc.b	$F2,$7E,$26,$5F,$C1,$75,$46,$2B,$F2,$91,$87,$E3,$CE,$E7,$51,$F8
	.dc.b	$1F,$A1,$FD,$AC,$FE,$85,$B1,$F8,$8E,$34,$62,$38,$1F,$B9,$9F,$D4
	.dc.b	$7D,$17,$F0,$A6,$C7,$F1,$56,$EF,$EF,$7F,$86,$CE,$E7,$EA,$74,$FD
	.dc.b	$7D,$3E,$14,$78,$70,$D3,$E0,$C3,$96,$3E,$48,$EC,$6C,$A7,$CC,$E4
	.dc.b	$FE,$08,$FC,$CB,$EA,$E7,$5F,$F1,$FC,$62,$3E,$81,$FE,$5E,$1E,$BB
	.dc.b	$2F,$F2,$E4,$61,$B1,$98,$FE,$6B,$E2,$BD,$77,$8D,$FC,$E2,$B5,$5D
	.dc.b	$5B,$5B,$9D,$5F,$42,$57,$F4,$0C,$3F,$5D,$8F,$F4,$08,$C6,$5F,$86
	.dc.b	$71,$3A,$6C,$65,$FE,$8F,$19,$6C,$65,$FD,$31,$C2,$7B,$6F,$FD,$13
	.dc.b	$6F,$F3,$AB,$F8,$79,$FF,$37,$37,$C6,$FA,$B3,$A9,$E6,$76,$AF,$42
	.dc.b	$C3,$C4,$E8,$FC,$87,$07,$17,$F8,$B3,$70,$FC,$4C,$78,$0D,$94,$FC
	.dc.b	$AF,$DC,$F8,$93,$49,$1F,$A3,$87,$DD,$7E,$CF,$0F,$D7,$46,$27,$53
	.dc.b	$9E,$4F,$D6,$3F,$B9,$9F,$56,$9F,$25,$FA,$77,$EE,$CB,$DB,$18,$3F
	.dc.b	$BB,$F4,$7F,$05,$E7,$FE,$F7,$F7,$7F,$C0,$E3,$12,$8C,$74,$AB,$7A
	.dc.b	$7F,$0E,$8C,$56,$E3,$F9,$16,$9F,$EA,$A8,$A7,$FA,$8C,$BC,$1F,$C7
	.dc.b	$0F,$BC,$C3,$EB,$14,$FE,$BD,$D5,$FC,$C7,$6F,$E4,$64,$8E,$AF,$B3
	.dc.b	$B3,$F9,$F5,$5D,$FC,$55,$FE,$18,$D7,$FA,$04,$E6,$3C,$87,$E0,$4F
	.dc.b	$8C,$78,$88,$7F,$13,$ED,$E6,$FF,$59,$6E,$7F,$D6,$27,$FC,$1D,$18
	.dc.b	$E9,$77,$FF,$8A,$3D,$6F,$EC,$E4,$C3,$FB,$5B,$FD,$CB,$C2,$EB,$C9
	.dc.b	$FD,$F1,$BE,$17,$F7,$AC,$B5,$7E,$83,$CE,$7A,$4F,$C1,$D7,$E0,$75
	.dc.b	$FC,$26,$9F,$E0,$FE,$6F,$F0,$AD,$7E,$50,$FF,$08,$78,$7F,$E3,$C8
	.dc.b	$C9,$F4,$63,$A9,$0E,$A7,$F8,$A0,$E4,$7F,$37,$F1,$39,$3F,$E1,$B6
	.dc.b	$B9,$36,$7F,$80,$7C,$6F,$D9,$DF,$F8,$E6,$3F,$C6,$D1,$8D,$FF,$C7
	.dc.b	$D1,$FE,$48,$F2,$3F,$C9,$BA,$5F,$23,$E3,$C6,$27,$FC,$9F,$1F,$59
	.dc.b	$F3,$9F,$68,$F5,$75,$FE,$91,$F8,$FD,$FF,$C3,$C8,$FF,$07,$7F,$DD
	.dc.b	$8C,$3F,$DC,$65,$3F,$DB,$25,$3F,$DF,$45,$3F,$DF,$05,$B7,$95,$FC
	.dc.b	$4A,$32,$74,$21,$D0,$A7,$42,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$47
	.dc.b	$91,$FE,$4A,$D7,$D1,$1E,$78,$C4,$78,$9A,$F7,$6B,$D9,$AF,$68,$E4
	.dc.b	$47,$BB,$5E,$E6,$08,$60,$87,$85,$E0,$D3,$65,$36,$79,$5D,$B4,$D9
	.dc.b	$4D,$9E,$97,$6D,$36,$53,$67,$B4,$C9,$4C,$C7,$C2,$24,$74,$8D,$1E
	.dc.b	$72,$1E,$32,$9D,$CA,$70,$34,$43,$45,$34,$60,$86,$09,$1E,$47,$D2
	.dc.b	$D7,$D3,$1E,$78,$C4,$78,$9A,$F7,$6B,$D9,$AF,$68,$E4,$47,$BB,$5E
	.dc.b	$E6,$08,$60,$87,$85,$E0,$D3,$65,$36,$79,$5D,$B4,$D9,$4D,$9E,$97
	.dc.b	$6D,$36,$53,$67,$B4,$C9,$4C,$C7,$C2,$24,$7E,$07,$C1,$FF,$18,$4F
	.dc.b	$F8,$AD,$AF,$FB,$B1,$AF,$F9,$00,$FF,$77,$CB,$D7,$FF,$79,$F8,$5F
	.dc.b	$E3,$38,$C4,$FF,$8F,$8F,$F7,$BB,$88,$D4,$F0,$9E,$6D,$7F,$DF,$A2
	.dc.b	$78,$3A,$3A,$19,$8E,$0F,$66,$B9,$37,$1C,$9E,$AD,$78,$1A,$8E,$31
	.dc.b	$98,$E5,$3E,$87,$44,$70,$47,$CC,$D7,$BC,$F4,$23,$C4,$D9,$0D,$99
	.dc.b	$2C,$73,$7C,$ED,$7C,$A7,$13,$25,$32,$72,$31,$3E,$D2,$4F,$92,$7A
	.dc.b	$46,$27,$C4,$74,$39,$10,$C1,$0E,$E5,$79,$47,$49,$ED,$3E,$26,$BC
	.dc.b	$CE,$24,$9E,$2E,$CF,$19,$98,$E2,$F7,$6B,$93,$71,$C9,$EC,$D7,$81
	.dc.b	$A8,$E3,$19,$8E,$53,$E9,$74,$47,$04,$7C,$CD,$7B,$CF,$42,$3C,$4D
	.dc.b	$90,$D9,$92,$C7,$37,$CE,$D7,$CA,$71,$32,$53,$27,$23,$13,$ED,$24
	.dc.b	$F9,$27,$A4,$62,$7C,$47,$43,$91,$0C,$10,$EE,$57,$94,$74,$9E,$D3
	.dc.b	$E2,$6B,$CC,$E2,$49,$E2,$EC,$F1,$99,$8E,$2F,$76,$B9,$37,$1C,$9E
	.dc.b	$CD,$78,$1A,$8E,$31,$F8,$1C,$7F,$D9,$FE,$5F,$FB,$0C,$95,$FE,$BB
	.dc.b	$7F,$DA,$5E,$73,$91,$29,$CC,$61,$B1,$93,$F8,$F8,$9E,$4E,$4F,$E4
	.dc.b	$75,$FE,$F5,$25,$7F,$BF,$23,$2D,$5C,$2E,$07,$D3,$F9,$3C,$8E,$F3
	.dc.b	$98,$C5,$7F,$30,$8C,$29,$DD,$72,$8C,$1B,$5F,$51,$FC,$CA,$BC,$0D
	.dc.b	$8C,$CF,$F4,$39,$CB,$61,$F5,$3F,$23,$CF,$EE,$D9,$EC,$D7,$E6,$D7
	.dc.b	$B6,$31,$1F,$ED,$0E,$0F,$F7,$C1,$1F,$F7,$83,$F0,$3F,$5B,$F8,$B0
	.dc.b	$BC,$5F,$21,$3A,$8C,$CE,$CE,$64,$3E,$8F,$5F,$F5,$4D,$BF,$04,$FB
	.dc.b	$E9,$F4,$5F,$3A,$BD,$31,$87,$E5,$98,$8D,$3F,$6A,$35,$AF,$A7,$CC
	.dc.b	$FC,$6F,$DF,$DA,$34,$8F,$DB,$A7,$93,$D3,$4F,$6B,$63,$33,$E2,$3C
	.dc.b	$11,$E2,$7F,$D7,$7A,$FC,$B7,$DE,$D7,$CD,$C5,$CC,$8F,$3B,$F9,$71
	.dc.b	$EC,$D1,$E4,$BE,$47,$F4,$FE,$97,$D4,$C1,$5A,$9C,$E1,$FE,$04,$8C
	.dc.b	$4A,$B8,$46,$1E,$51,$E8,$3E,$B1,$F8,$52,$CF,$DE,$F7,$BE,$11,$23
	.dc.b	$D1,$A2,$3D,$18,$F9,$E5,$F8,$17,$82,$30,$DC,$BF,$80,$B6,$77,$87
	.dc.b	$89,$F2,$1F,$28,$F2,$E5,$F1,$3E,$3B,$F8,$4E,$1D,$5B,$5F,$06,$11
	.dc.b	$8A,$FE,$1F,$5C,$5F,$9B,$5F,$81,$FA,$BF,$D6,$36,$3F,$40,$DE,$5F
	.dc.b	$A0,$FE,$82,$77,$3F,$96,$FA,$BF,$E1,$13,$F8,$B6,$C6,$4F,$E0,$D1
	.dc.b	$96,$9E,$D2,$1F,$44,$AF,$77,$67,$E7,$E7,$AB,$4C,$C6,$F9,$3E,$EF
	.dc.b	$37,$F2,$2C,$7F,$61,$C3,$FC,$86,$FE,$69,$DC,$86,$09,$DF,$FE,$60
	.dc.b	$E6,$72,$7F,$94,$23,$0D,$F5,$5E,$17,$F3,$CA,$DD,$FF,$3E,$BD,$B6
	.dc.b	$B6,$A7,$6A,$76,$D9,$DC,$EA,$73,$97,$F0,$AA,$FE,$8A,$1C,$3F,$D2
	.dc.b	$1A,$FF,$12,$36,$E1,$FE,$A6,$D7,$96,$3F,$AE,$76,$D7,$F0,$DE,$07
	.dc.b	$E1,$39,$9B,$FF,$9E,$B7,$C9,$2B,$C0,$78,$EF,$93,$82,$1F,$C6,$F8
	.dc.b	$39,$3F,$89,$C3,$B4,$62,$34,$7F,$A0,$6B,$F3,$1E,$3F,$8F,$0F,$ED
	.dc.b	$AD,$93,$4F,$D4,$3F,$C8,$27,$F5,$1C,$7F,$DD,$3A,$72,$48,$F4,$71
	.dc.b	$7C,$EA,$F8,$AD,$D7,$E4,$1F,$2D,$FE,$23,$F8,$1F,$EF,$49,$5F,$B5
	.dc.b	$E8,$FF,$00,$C7,$C9,$E9,$FF,$50,$E1,$FC,$CD,$AF,$F4,$26,$BF,$CF
	.dc.b	$DA,$F8,$8F,$AB,$7E,$3D,$FE,$6C,$7D,$C6,$C7,$F2,$4F,$47,$F8,$81
	.dc.b	$31,$F2,$52,$3E,$E5,$78,$27,$F9,$AC,$FF,$AB,$CF,$E9,$11,$9E,$1F
	.dc.b	$F0,$9F,$0F,$8A,$61,$F2,$0A,$7F,$5C,$D7,$81,$1E,$86,$4A,$7F,$5D
	.dc.b	$E3,$FF,$13,$FE,$09,$FD,$6C,$91,$FD,$66,$32,$FF,$5B,$EE,$F2,$11
	.dc.b	$E7,$97,$F7,$88,$F9,$3C,$1F,$63,$3F,$8F,$5F,$2C,$F4,$1F,$D9,$6B
	.dc.b	$F1,$6D,$3F,$16,$5E,$2F,$F0,$09,$EE,$7E,$86,$DF,$4C,$8F,$F8,$60
	.dc.b	$E2,$7F,$85,$BC,$0F,$F0,$EB,$83,$FC,$6A,$E2,$78,$1F,$E2,$EF,$07
	.dc.b	$F2,$57,$FE,$26,$AC,$1F,$C0,$9C,$3F,$E3,$1F,$0B,$ED,$5F,$F8,$D7
	.dc.b	$7F,$F1,$BC,$60,$F8,$79,$7C,$2C,$FE,$26,$7E,$8C,$3F,$CA,$3E,$1F
	.dc.b	$F9,$43,$AF,$FC,$AA,$48,$CD,$FF,$97,$23,$F2,$64,$8F,$D4,$B5,$FF
	.dc.b	$6C,$9F,$EE,$72,$1F,$EF,$92,$9A,$30,$43,$04,$3F,$DE,$A7,$02,$9B
	.dc.b	$29,$B3,$FE,$00,$1B,$29,$B2,$9B,$C5,$E2,$FF,$01,$8C,$9D,$08,$60
	.dc.b	$91,$C1,$FE,$10,$D7,$C0,$D7,$C0,$D7,$C1,$1D,$E3,$11,$D9,$AF,$68
	.dc.b	$D1,$1E,$ED,$7B,$B5,$EE,$D7,$B9,$C0,$87,$02,$9C,$0A,$70,$34,$43
	.dc.b	$45,$34,$60,$86,$08,$7B,$5F,$23,$4E,$E5,$3A,$14,$E6,$53,$99,$F1
	.dc.b	$4E,$25,$38,$94,$E2,$53,$8C,$7C,$E2,$47,$92,$34,$77,$21,$82,$47
	.dc.b	$07,$C2,$D7,$C2,$D7,$C2,$D7,$C3,$1D,$E3,$11,$D9,$AF,$68,$D1,$1E
	.dc.b	$ED,$7B,$B5,$EE,$D7,$B9,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34
	.dc.b	$60,$86,$08,$7B,$5F,$23,$4E,$E5,$3A,$1F,$B0,$EB,$76,$7F,$8B,$23
	.dc.b	$FD,$CF,$18,$BF,$EF,$0D,$8F,$F1,$3B,$5F,$F7,$6D,$BC,$AF,$F7,$AF
	.dc.b	$77,$F8,$B2,$31,$3F,$E3,$63,$FD,$F2,$E2,$35,$3C,$27,$9B,$5F,$F7
	.dc.b	$F8,$9E,$0E,$8E,$86,$63,$83,$D9,$AE,$4D,$C7,$27,$AB,$5E,$06,$A3
	.dc.b	$8C,$66,$39,$4F,$A1,$D1,$1C,$11,$F3,$35,$EF,$3D,$08,$F1,$36,$43
	.dc.b	$66,$4B,$1C,$DF,$3B,$5F,$29,$C4,$C9,$4C,$9C,$8C,$4F,$B4,$93,$E4
	.dc.b	$9E,$91,$89,$F1,$1D,$0E,$44,$30,$43,$B9,$5E,$51,$D2,$7B,$4F,$89
	.dc.b	$AF,$33,$89,$27,$8B,$B3,$C6,$66,$38,$BD,$DA,$E4,$DC,$72,$7B,$35
	.dc.b	$E0,$6A,$38,$C6,$63,$94,$FA,$5D,$11,$C1,$1F,$33,$5E,$F3,$D0,$8F
	.dc.b	$13,$64,$36,$64,$B1,$CD,$F3,$B5,$F2,$9C,$4C,$94,$C9,$C8,$C4,$FB
	.dc.b	$49,$3E,$49,$E9,$18,$9F,$11,$D0,$E4,$43,$04,$3B,$95,$E5,$1D,$27
	.dc.b	$B4,$F8,$9A,$F3,$38,$92,$78,$BB,$3C,$66,$63,$8B,$DD,$AE,$4D,$C7
	.dc.b	$27,$B3,$5E,$06,$A3,$F8,$4E,$AC,$DA,$BC,$6F,$F4,$49,$2F,$FD,$3E
	.dc.b	$FF,$B3,$3D,$67,$C2,$A7,$81,$FC,$70,$46,$9C,$9F,$C8,$1C,$4F,$39
	.dc.b	$E4,$FF,$A6,$3E,$D6,$57,$FA,$B9,$AF,$F2,$8C,$7F,$DF,$5A,$7F,$A7
	.dc.b	$89,$1F,$CB,$39,$1D,$6F,$A4,$F3,$78,$BE,$59,$CC,$62,$B8,$56,$D5
	.dc.b	$6E,$30,$DF,$AA,$FE,$67,$7D,$9B,$26,$98,$C3,$E3,$78,$DC,$DB,$87
	.dc.b	$81,$AF,$CB,$BF,$4C,$62,$3F,$D1,$2F,$B9,$F8,$5C,$1F,$62,$31,$1F
	.dc.b	$6B,$C3,$78,$9C,$78,$91,$89,$63,$F5,$E7,$35,$BA,$D4,$F4,$AF,$F6
	.dc.b	$ED,$DD,$BF,$11,$87,$F6,$1C,$BE,$7F,$D5,$FC,$BB,$F5,$6A,$8C,$42
	.dc.b	$76,$FC,$E3,$E8,$C6,$A3,$EC,$FD,$67,$E3,$7E,$BF,$E9,$36,$BE,$BB
	.dc.b	$5E,$ED,$7B,$5B,$EB,$BA,$36,$3C,$07,$9A,$32,$79,$9F,$79,$FE,$FE
	.dc.b	$CF,$E0,$6F,$F3,$E3,$F0,$FF,$55,$D0,$8F,$43,$C2,$F8,$3E,$1B,$F4
	.dc.b	$DF,$07,$F0,$67,$E2,$C6,$34,$FC,$8B,$63,$22,$78,$3F,$82,$3C,$84
	.dc.b	$3F,$07,$B7,$F8,$4C,$7C,$7B,$5F,$D0,$E1,$FE,$10,$9C,$C6,$2B,$F7
	.dc.b	$51,$87,$F4,$33,$BA,$E5,$18,$3C,$BF,$82,$F9,$27,$F5,$F8,$FF,$03
	.dc.b	$8C,$6C,$8C,$4F,$BA,$95,$EC,$7E,$79,$EA,$D7,$E7,$E3,$EB,$C7,$F1
	.dc.b	$52,$F4,$7E,$1C,$FB,$B8,$7F,$89,$46,$2B,$6D,$8C,$E3,$FB,$EF,$6F
	.dc.b	$F6,$6D,$C3,$F4,$6D,$E4,$F4,$CE,$A7,$F3,$7F,$5B,$F1,$DF,$C6,$B6
	.dc.b	$BF,$87,$B7,$63,$B9,$F6,$9F,$64,$F7,$1F,$64,$FC,$BF,$D7,$71,$23
	.dc.b	$F3,$CC,$94,$E6,$4E,$6F,$E4,$6F,$33,$1C,$9F,$97,$7B,$9F,$E5,$0F
	.dc.b	$0B,$FC,$C1,$E6,$38,$71,$E2,$BC,$5F,$E7,$93,$9B,$FE,$7D,$7C,$2F
	.dc.b	$75,$A6,$CA,$B8,$2A,$E0,$F2,$BE,$CE,$EB,$8D,$70,$C7,$F8,$6C,$FF
	.dc.b	$4B,$0E,$1F,$E9,$CD,$7F,$42,$6D,$C3,$FD,$55,$AF,$2C,$7F,$AF,$BB
	.dc.b	$6B,$FD,$0B,$2F,$EB,$9E,$5F,$E6,$9B,$B6,$65,$39,$6C,$79,$2B,$9C
	.dc.b	$79,$0C,$72,$7A,$DF,$13,$ED,$9F,$C4,$C7,$23,$F1,$E5,$3F,$1E,$57
	.dc.b	$75,$F9,$68,$FF,$45,$47,$E6,$FC,$6C,$11,$FD,$D7,$1F,$F7,$7F,$47
	.dc.b	$F4,$94,$69,$27,$CF,$1F,$17,$0F,$46,$7F,$28,$F9,$B3,$FC,$02,$5F
	.dc.b	$6D,$FC,$03,$A7,$FC,$08,$D8,$CC,$FF,$01,$8F,$DB,$75,$7C,$C9,$FE
	.dc.b	$23,$D3,$FC,$3B,$5C,$C7,$C1,$CB,$F8,$A4,$7F,$07,$CB,$C5,$B7,$CE
	.dc.b	$AF,$BC,$D7,$E6,$9F,$DE,$73,$FF,$0B,$7C,$37,$8C,$9A,$7F,$13,$7F
	.dc.b	$8B,$F6,$FE,$02,$BF,$9E,$9F,$78,$B8,$FF,$80,$E3,$51,$F7,$48,$78
	.dc.b	$B2,$F8,$F8,$F9,$70,$FF,$09,$F7,$70,$AF,$EA,$AD,$8C,$C7,$73,$FA
	.dc.b	$A1,$4F,$EB,$7A,$FF,$8B,$FF,$06,$FE,$C7,$06,$91,$FE,$D6,$71,$9F
	.dc.b	$ED,$F9,$7F,$86,$F8,$7C,$1D,$5F,$CB,$23,$EF,$54,$3F,$02,$43,$FA
	.dc.b	$2F,$83,$FE,$09,$E4,$FF,$04,$36,$3F,$A6,$35,$DF,$85,$FE,$E4,$79
	.dc.b	$1F,$14,$EC,$7C,$8C,$7F,$C3,$EF,$FB,$8C,$FF,$12,$14,$FF,$05,$F4
	.dc.b	$73,$69,$93,$FB,$FF,$87,$FA,$D9,$FF,$1A,$47,$EB,$72,$FF,$1A,$36
	.dc.b	$75,$1F,$E3,$FC,$BF,$C7,$DE,$2F,$C4,$25,$7D,$13,$F1,$67,$A8,$A7
	.dc.b	$E0,$0F,$DD,$FD,$BF,$E9,$C8,$FF,$00,$7F,$CB,$CF,$F2,$03,$FC,$C5
	.dc.b	$E4,$FF,$9A,$7B,$F2,$48,$FE,$0E,$FE,$C4,$8E,$0F,$F3,$76,$CF,$2F
	.dc.b	$F8,$64,$23,$0F,$F0,$F6,$BF,$EF,$D2,$1C,$A3,$8C,$6A,$30,$71,$8C
	.dc.b	$91,$C1,$0E,$67,$FB,$F0,$A7,$FA,$16,$3A,$39,$3A,$9A,$3D,$11,$D5
	.dc.b	$C4,$70,$79,$46,$5E,$24,$3A,$10,$D9,$E5,$34,$60,$EA,$F3,$78,$1E
	.dc.b	$23,$81,$88,$F1,$39,$38,$9A,$3A,$1D,$8D,$15,$E2,$48,$F0,$47,$03
	.dc.b	$B3,$87,$89,$83,$B1,$83,$81,$0C,$12,$3B,$3C,$1A,$FA,$08,$79,$63
	.dc.b	$C1,$1A,$8C,$1C,$63,$24,$70,$43,$99,$E7,$29,$E5,$8E,$8E,$4E,$A6
	.dc.b	$8F,$44,$75,$71,$1C,$1E,$51,$97,$89,$0E,$84,$36,$79,$4D,$18,$3A
	.dc.b	$BC,$DE,$07,$88,$E0,$62,$3C,$4E,$4E,$26,$8E,$87,$63,$45,$78,$92
	.dc.b	$3C,$11,$C0,$EC,$E1,$E2,$60,$EC,$60,$E0,$43,$04,$8E,$CF,$06,$BE
	.dc.b	$82,$1E,$58,$F0,$46,$A3,$07,$18,$C9,$1F,$C9,$F8,$0F,$03,$F8,$E4
	.dc.b	$7F,$90,$81,$48,$7F,$03,$F8,$0C,$F0,$7F,$8F,$70,$46,$27,$F8,$EB
	.dc.b	$5E,$67,$F1,$D7,$0F,$33,$FC,$9E,$59,$E9,$1A,$8E,$06,$8B,$1A,$71
	.dc.b	$3E,$26,$CF,$18,$C4,$EC,$93,$E6,$36,$47,$2F,$A2,$38,$BB,$8C,$BE
	.dc.b	$73,$04,$38,$4F,$84,$8F,$53,$47,$13,$66,$4B,$1D,$9F,$79,$C0,$A6
	.dc.b	$8C,$4F,$99,$C1,$0E,$C4,$38,$95,$E0,$D9,$F3,$CF,$86,$73,$3C,$08
	.dc.b	$E0,$93,$C2,$3C,$6D,$7C,$47,$07,$0F,$73,$E0,$16,$7C,$11,$A8,$E0
	.dc.b	$68,$B1,$A7,$13,$E2,$6C,$F1,$8C,$4E,$C9,$3E,$63,$64,$72,$FA,$23
	.dc.b	$8B,$B8,$CB,$E7,$30,$43,$84,$F8,$48,$F5,$34,$71,$36,$64,$B1,$D9
	.dc.b	$F7,$9C,$0A,$68,$C4,$F9,$9C,$10,$EC,$43,$89,$5E,$0D,$9F,$3C,$F8
	.dc.b	$67,$33,$C0,$8E,$09,$3C,$23,$C6,$D7,$C4,$70,$70,$F7,$3E,$01,$67
	.dc.b	$C1,$1A,$8E,$06,$8B,$1A,$71,$3E,$26,$CF,$18,$F4,$F6,$3A,$1D,$2F
	.dc.b	$2F,$FD,$4C,$4C,$3F,$D7,$EF,$FB,$67,$C8,$7D,$7F,$F2,$48,$C3,$67
	.dc.b	$F8,$E8,$9D,$BD,$4C,$B8,$9F,$F5,$A7,$99,$27,$FD,$9A,$D7,$F9,$3C
	.dc.b	$F2,$8E,$F0,$69,$1F,$F6,$AE,$9F,$EB,$F2,$38,$3F,$98,$F3,$B9,$9F
	.dc.b	$39,$FC,$FB,$CD,$F4,$CE,$63,$15,$C2,$B6,$AB,$71,$86,$FC,$57,$F3
	.dc.b	$CB,$EC,$D9,$34,$D7,$E6,$1F,$EF,$AF,$33,$C2,$DC,$3C,$ED,$7F,$A9
	.dc.b	$DA,$31,$1F,$EB,$7C,$BD,$EF,$F4,$EE,$1F,$BD,$C1,$FD,$1D,$B3,$DB
	.dc.b	$EA,$BF,$AE,$E3,$6C,$D2,$3F,$24,$E9,$C5,$F8,$8A,$EF,$1E,$13,$CF
	.dc.b	$F5,$7F,$21,$F3,$E4,$8C,$43,$3F,$54,$61,$B1,$F3,$4F,$83,$F5,$9F
	.dc.b	$DC,$BE,$FB,$E2,$B6,$32,$23,$2D,$3E,$4F,$D7,$75,$C0,$D2,$47,$73
	.dc.b	$DB,$B7,$C7,$CF,$DD,$1F,$04,$F7,$1E,$CF,$AF,$E8,$47,$91,$ED,$B4
	.dc.b	$7C,$0F,$8F,$7E,$10,$FA,$5F,$F0,$6F,$AD,$F0,$F0,$FF,$07,$D6,$A7
	.dc.b	$3B,$7F,$82,$23,$12,$AE,$0A,$33,$18,$87,$1F,$E0,$1F,$51,$8C,$FD
	.dc.b	$AF,$C0,$E1,$FD,$CE,$AF,$8A,$47,$E9,$72,$FF,$0C,$CE,$63,$15,$FE
	.dc.b	$26,$8C,$3F,$A5,$9D,$D7,$28,$C1,$EE,$D3,$E1,$2F,$87,$F5,$E9,$EE
	.dc.b	$6C,$E9,$BA,$FF,$8D,$1F,$76,$9E,$DB,$F5,$1E,$E7,$0F,$D3,$FB,$1F
	.dc.b	$F1,$DF,$B9,$F2,$33,$FD,$54,$61,$FD,$7B,$67,$F3,$2D,$7E,$E7,$D9
	.dc.b	$7E,$A2,$7F,$6A,$D8,$FC,$FB,$5D,$9F,$AC,$8F,$DD,$F8,$1F,$22,$3F
	.dc.b	$0C,$D8,$FD,$E3,$78,$FF,$94,$F6,$7B,$B8,$BE,$BD,$FE,$48,$87,$DB
	.dc.b	$E1,$FF,$24,$E1,$FC,$17,$C3,$E4,$49,$FA,$27,$E5,$5F,$13,$C7,$8F
	.dc.b	$F7,$6D,$3E,$55,$F8,$1C,$1F,$38,$87,$F9,$9B,$C4,$FF,$35,$78,$DF
	.dc.b	$E6,$8F,$67,$F0,$12,$9C,$A8,$CC,$61,$B8,$2B,$96,$1F,$E6,$8A,$D3
	.dc.b	$67,$57,$CE,$31,$5C,$23,$C9,$3A,$BF,$1F,$0F,$F2,$49,$FE,$AE,$7C
	.dc.b	$92,$B9,$7F,$AD,$B5,$FE,$0A,$6B,$1F,$ED,$2D,$CF,$9B,$CC,$FE,$73
	.dc.b	$7F,$E8,$AF,$2F,$F2,$1B,$7F,$A3,$E5,$18,$9D,$6B,$C2,$72,$F8,$B0
	.dc.b	$EF,$17,$0E,$F8,$F9,$63,$FD,$2E,$DA,$FC,$5C,$72,$3F,$99,$EF,$D9
	.dc.b	$A6,$4F,$C9,$96,$3F,$A7,$70,$7F,$A8,$BC,$6F,$4B,$63,$F4,$32,$9C
	.dc.b	$B6,$3F,$56,$A7,$E0,$F1,$7F,$AB,$1B,$19,$9D,$C7,$E7,$F1,$F6,$E5
	.dc.b	$F0,$2B,$E8,$9F,$04,$FF,$07,$F3,$FF,$24,$F8,$5F,$E0,$B0,$69,$30
	.dc.b	$FE,$27,$C5,$F0,$67,$FC,$43,$D3,$FC,$57,$83,$F9,$5E,$5F,$C6,$9B
	.dc.b	$19,$BF,$05,$43,$C6,$43,$F9,$97,$4F,$F1,$CF,$67,$C6,$49,$FE,$75
	.dc.b	$19,$7F,$BE,$F6,$7E,$16,$A9,$F8,$42,$E1,$FE,$04,$8D,$47,$73,$F1
	.dc.b	$C5,$E2,$FC,$07,$27,$F5,$E8,$E4,$7F,$72,$ED,$E2,$FF,$55,$6E,$1F
	.dc.b	$E0,$B6,$CB,$BF,$F1,$75,$FD,$BA,$FF,$B6,$91,$FE,$FE,$76,$3F,$B0
	.dc.b	$78,$1C,$88,$E4,$F3,$9F,$96,$E2,$FB,$8D,$DF,$EF,$9F,$B4,$CB,$FC
	.dc.b	$3D,$E1,$7E,$46,$7F,$C4,$6D,$7F,$C1,$AD,$7F,$B8,$35,$EE,$4C,$3F
	.dc.b	$4C,$F8,$A7,$FC,$55,$5C,$5B,$1F,$9B,$3F,$C6,$31,$9C,$BF,$7F,$CD
	.dc.b	$FE,$3E,$9F,$37,$89,$FE,$3C,$6F,$89,$FC,$0B,$0F,$F2,$0C,$6B,$C5
	.dc.b	$F9,$F5,$FE,$4A,$E1,$FF,$24,$C6,$0F,$AB,$3E,$7D,$7E,$2B,$73,$FE
	.dc.b	$1B,$E3,$FF,$9B,$3C,$CF,$F3,$19,$27,$FC,$DB,$D8,$51,$FC,$54,$8F
	.dc.b	$F0,$96,$BF,$E6,$E8,$E1,$DF,$FE,$6E,$24,$74,$3F,$97,$9F,$E2,$67
	.dc.b	$F9,$01,$1F,$E2,$8F,$F2,$CD,$9E,$57,$F2,$49,$FE,$0C,$D8,$CC,$7F
	.dc.b	$28,$6B,$C6,$34,$47,$4D,$74,$60,$86,$08,$7F,$A5,$23,$AB,$63,$93
	.dc.b	$4E,$25,$38,$9F,$2C,$F4,$9A,$31,$1D,$1B,$1D,$1A,$6C,$FE,$54,$E5
	.dc.b	$A6,$4E,$A7,$80,$C9,$4C,$9E,$52,$1D,$8D,$47,$8D,$E2,$68,$A6,$CC
	.dc.b	$1D,$0C,$10,$D9,$0D,$15,$EC,$E0,$8E,$09,$1E,$F8,$CB,$63,$31,$F0
	.dc.b	$5A,$FA,$A3,$44,$74,$D7,$46,$08,$60,$87,$BA,$3A,$B6,$39,$34,$E2
	.dc.b	$53,$89,$F0,$CF,$49,$A3,$11,$D1,$B1,$D1,$A6,$CF,$90,$E5,$A6,$4E
	.dc.b	$A7,$80,$C9,$4C,$9E,$52,$1D,$8D,$47,$8D,$E2,$68,$A6,$CC,$1D,$0C
	.dc.b	$10,$D9,$0D,$15,$EC,$E0,$8E,$09,$1E,$F8,$CB,$63,$31,$F0,$5A,$FA
	.dc.b	$A3,$44,$74,$D7,$46,$08,$60,$87,$EE,$4F,$F2,$97,$7F,$F1,$A7,$FC
	.dc.b	$8E,$7F,$90,$BC,$1F,$F2,$AB,$B6,$BF,$BE,$F9,$6F,$E5,$9D,$BF,$C7
	.dc.b	$5B,$1F,$C3,$5A,$EC,$FF,$30,$9F,$E2,$97,$2D,$32,$7F,$8A,$5C,$B4
	.dc.b	$CE,$08,$CB,$63,$31,$82,$38,$24,$F9,$DA,$79,$CA,$79,$DD,$11,$D3
	.dc.b	$5D,$18,$21,$82,$1E,$D7,$83,$4D,$94,$D9,$EF,$76,$D3,$65,$37,$3F
	.dc.b	$06,$36,$D8,$DB,$5D,$CF,$C4,$9D,$B6,$36,$D7,$67,$C4,$3C,$C4,$3C
	.dc.b	$45,$9E,$CD,$8E,$4D,$39,$1C,$09,$1C,$1B,$1C,$1A,$70,$8D,$12,$34
	.dc.b	$D8,$D4,$60,$8E,$09,$3F,$01,$A7,$C0,$29,$F0,$1D,$11,$D3,$5D,$18
	.dc.b	$21,$82,$1F,$05,$E0,$D3,$65,$36,$7C,$07,$6D,$36,$53,$73,$F0,$63
	.dc.b	$6D,$8D,$B5,$DC,$FC,$49,$DB,$63,$6D,$76,$7C,$43,$CC,$43,$C4,$59
	.dc.b	$EC,$D8,$E4,$D3,$91,$C0,$91,$C1,$B1,$C1,$A7,$08,$D1,$23,$4D,$8E
	.dc.b	$BC,$AD,$DB,$BC,$BF,$F6,$A1,$2B,$FD,$90,$FF,$B8,$3F,$13,$74,$7F
	.dc.b	$91,$E3,$0D,$9F,$E5,$13,$A7,$A1,$88,$D3,$FE,$22,$EF,$AF,$FB,$1D
	.dc.b	$AF,$FB,$15,$B9,$1C,$D7,$5F,$33,$C3,$46,$91,$FF,$70,$7D,$37,$FB
	.dc.b	$BC,$8E,$0F,$F7,$26,$DF,$EF,$32,$38,$21,$FE,$3B,$D7,$E4,$69,$FC
	.dc.b	$C9,$B3,$C2,$7F,$CD,$5D,$2E,$B7,$D6,$FF,$3A,$8C,$36,$33,$3F,$D0
	.dc.b	$27,$2D,$87,$D3,$FF,$9D,$7E,$9F,$FA,$A5,$75,$8C,$57,$F9,$3E,$30
	.dc.b	$A7,$75,$CA,$30,$7B,$0F,$F5,$56,$1E,$96,$99,$18,$FB,$9C,$35,$F8
	.dc.b	$9C,$5F,$8A,$9E,$51,$C2,$30,$F0,$6B,$C3,$3F,$95,$18,$8F,$09,$F2
	.dc.b	$AD,$F5,$FF,$E9,$5F,$C2,$FD,$6D,$9F,$81,$3A,$53,$A6,$9A,$FA,$3F
	.dc.b	$A0,$47,$E7,$3F,$DA,$B2,$F6,$47,$88,$87,$AF,$1F,$39,$32,$F7,$CF
	.dc.b	$09,$7D,$63,$E8,$3F,$D5,$9C,$5F,$E0,$B9,$CC,$61,$FF,$07,$4E,$DF
	.dc.b	$CD,$18,$8D,$3F,$1E,$FB,$EB,$F1,$E3,$C9,$18,$C7,$C9,$1F,$77,$7F
	.dc.b	$C9,$7A,$3F,$EB,$DF,$AF,$47,$C8,$8C,$36,$53,$A7,$E9,$4E,$9F,$11
	.dc.b	$F6,$EB,$BB,$FA,$5F,$94,$F7,$DF,$F8,$A7,$54,$62,$74,$7E,$DE,$75
	.dc.b	$19,$8F,$2F,$03,$EC,$7F,$C5,$35,$F7,$9B,$3F,$85,$6B,$C1,$B1,$FB
	.dc.b	$1F,$AA,$FD,$43,$F8,$66,$CF,$E3,$5A,$F8,$1A,$FC,$C3,$DD,$A7,$E9
	.dc.b	$67,$F7,$95,$DA,$FC,$D4,$AF,$90,$F8,$CC,$57,$A5,$A7,$CA,$21,$F8
	.dc.b	$EC,$7F,$14,$4D,$7E,$F3,$73,$FF,$20,$F7,$3B,$DE,$0F,$FB,$E7,$D3
	.dc.b	$F1,$63,$FE,$FB,$AD,$4E,$73,$FF,$2C,$D6,$AB,$FC,$BB,$39,$78,$5F
	.dc.b	$45,$3B,$BE,$71,$8B,$E6,$44,$7E,$9C,$FD,$06,$3F,$CA,$1B,$97,$F5
	.dc.b	$06,$E5,$FA,$57,$81,$F8,$98,$D3,$63,$53,$CE,$3E,$61,$F4,$33,$FA
	.dc.b	$AF,$F3,$23,$F0,$C5,$3E,$A1,$F5,$8E,$E7,$F1,$33,$C6,$43,$F1,$03
	.dc.b	$EB,$3B,$6A,$DF,$DF,$18,$6D,$70,$8D,$EF,$FC,$D6,$FC,$66,$0F,$24
	.dc.b	$FA,$1F,$26,$1F,$A8,$5E,$2E,$38,$7E,$6B,$63,$35,$FA,$A6,$9E,$B3
	.dc.b	$E4,$1E,$B3,$07,$EF,$07,$A5,$EC,$E3,$FB,$2D,$8C,$CF,$F0,$58,$F2
	.dc.b	$ED,$F5,$27,$99,$F8,$0C,$7E,$A3,$83,$E6,$99,$8E,$0D,$DB,$ED,$C6
	.dc.b	$23,$F9,$29,$F0,$0F,$E4,$45,$3F,$1A,$6B,$1F,$E5,$9C,$DF,$CA,$5F
	.dc.b	$74,$75,$29,$E2,$25,$F9,$9D,$1F,$18,$FE,$8C,$7F,$37,$C7,$FB,$D7
	.dc.b	$47,$E5,$5B,$19,$E0,$4E,$5F,$EA,$5F,$73,$FD,$44,$93,$FD,$5B,$AF
	.dc.b	$F4,$B1,$E4,$C3,$C5,$3E,$4C,$FE,$14,$60,$8E,$09,$A7,$C8,$24,$F3
	.dc.b	$8C,$B6,$3C,$DD,$BF,$96,$8F,$EC,$ED,$7F,$B7,$35,$FE,$62,$DC,$FF
	.dc.b	$C2,$71,$FE,$0A,$EE,$F7,$11,$F0,$9F,$24,$FC,$06,$1E,$8B,$CC,$63
	.dc.b	$0F,$4F,$D6,$FF,$00,$24,$7E,$F5,$FF,$1D,$ED,$23,$FC,$90,$7E,$04
	.dc.b	$FE,$1B,$E0,$FF,$93,$63,$FB,$67,$87,$FE,$45,$9F,$F7,$B1,$FE,$49
	.dc.b	$2B,$9E,$CD,$F8,$7F,$C5,$9F,$F2,$49,$27,$BB,$5F,$F0,$B3,$FC,$58
	.dc.b	$8F,$F1,$37,$FD,$CC,$C3,$FD,$CC,$53,$FD,$F8,$53,$FD,$F4,$53,$FD
	.dc.b	$F0,$59,$FF,$7B,$37,$C6,$FE,$48,$E8,$EA,$43,$A9,$4E,$A5,$38,$1A
	.dc.b	$21,$A2,$9A,$30,$43,$04,$8F,$2B,$D5,$AF,$26,$BC,$5A,$F1,$8F,$5C
	.dc.b	$60,$8E,$09,$1F,$D0,$E3,$2D,$8C,$BF,$14,$C1,$0C,$10,$F3,$3C,$1A
	.dc.b	$6C,$A6,$CF,$53,$B6,$9B,$29,$B3,$DC,$ED,$A6,$CA,$6C,$F8,$46,$4A
	.dc.b	$66,$3D,$ED,$8E,$8E,$8F,$39,$0F,$19,$4E,$E5,$38,$1A,$21,$A2,$9A
	.dc.b	$30,$43,$04,$8F,$2B,$D5,$AF,$26,$BC,$5A,$F1,$8F,$5C,$60,$8E,$09
	.dc.b	$1F,$16,$32,$D8,$CB,$F1,$4C,$10,$C1,$0F,$33,$C1,$A6,$CA,$6C,$F5
	.dc.b	$3B,$69,$B2,$9B,$3D,$CE,$DA,$6C,$FD,$57,$67,$F8,$A1,$FE,$B2,$7F
	.dc.b	$8C,$0F,$F4,$23,$97,$FA,$EB,$5F,$F1,$5C,$BD,$8F,$F1,$6F,$6F,$F2
	.dc.b	$06,$CF,$49,$D3,$FE,$3D,$38,$C6,$9F,$ED,$31,$A9,$E5,$1B,$11,$97
	.dc.b	$A4,$F1,$79,$9B,$8D,$3D,$23,$4F,$26,$C6,$A3,$8C,$F8,$08,$F6,$3A
	.dc.b	$4F,$02,$38,$24,$70,$7C,$06,$0E,$4E,$5E,$E6,$08,$70,$74,$F9,$CE
	.dc.b	$46,$4D,$9C,$8C,$10,$C1,$0E,$4F,$49,$F8,$11,$83,$99,$4E,$45,$39
	.dc.b	$3C,$09,$3F,$0A,$7A,$36,$79,$CE,$9F,$49,$CA,$34,$F6,$8D,$4F,$28
	.dc.b	$D9,$E1,$24,$69,$EB,$3C,$9E,$87,$08,$D3,$D6,$34,$F2,$6C,$6A,$38
	.dc.b	$CF,$84,$8F,$63,$A4,$F0,$23,$82,$47,$07,$C0,$60,$E4,$E5,$EE,$60
	.dc.b	$87,$07,$4F,$9C,$E4,$64,$D9,$C8,$C1,$0C,$10,$E4,$F4,$9F,$81,$18
	.dc.b	$39,$94,$E4,$53,$93,$C0,$93,$F0,$A7,$A3,$67,$9C,$E9,$F4,$9C,$A3
	.dc.b	$4F,$68,$D4,$F2,$8D,$9E,$12,$46,$9E,$B3,$C9,$E8,$70,$8D,$3E,$AD
	.dc.b	$DB,$37,$78,$FF,$EB,$E2,$57,$FA,$ED,$FE,$46,$BC,$A7,$95,$FE,$CA
	.dc.b	$9C,$C6,$27,$99,$FC,$A9,$C4,$6A,$78,$4F,$F2,$3D,$EB,$FC,$81,$AF
	.dc.b	$F2,$C6,$BF,$E4,$79,$F1,$FB,$0F,$AF,$CB,$67,$C2,$7F,$A4,$9C,$4F
	.dc.b	$09,$E7,$A7,$F9,$4F,$C7,$FF,$27,$D9,$A4,$7D,$58,$FF,$BA,$3C,$AF
	.dc.b	$54,$1A,$47,$FD,$C5,$C1,$FE,$E8,$23,$83,$FD,$4B,$D0,$EC,$7C,$FF
	.dc.b	$F3,$98,$C3,$63,$33,$FD,$1E,$72,$D8,$79,$3B,$24,$65,$E5,$F0,$92
	.dc.b	$38,$37,$0F,$2C,$7F,$AE,$39,$3F,$0B,$C7,$FD,$3D,$B3,$A9,$63,$E9
	.dc.b	$F8,$4F,$31,$23,$C8,$FF,$4E,$AE,$31,$F0,$4E,$DC,$BF,$91,$BE,$9D
	.dc.b	$8F,$BB,$7D,$37,$F5,$B8,$C3,$63,$37,$FD,$A6,$72,$FC,$5F,$0B,$FD
	.dc.b	$A3,$E3,$7F,$7A,$D1,$18,$9F,$A2,$D4,$7B,$1C,$37,$DF,$FA,$90,$69
	.dc.b	$23,$E3,$67,$EB,$7D,$C7,$FB,$A7,$3F,$91,$7E,$2E,$7E,$64,$9F,$43
	.dc.b	$FD,$D3,$C6,$70,$FE,$81,$B3,$F9,$C3,$CE,$E3,$A7,$FD,$DB,$3F,$A6
	.dc.b	$E0,$F9,$F5,$F3,$B8,$BC,$A4,$CB,$FC,$3A,$BF,$37,$FC,$4C,$A7,$EE
	.dc.b	$BF,$12,$75,$C9,$FE,$1C,$D3,$FD,$FB,$18,$9E,$11,$8B,$E7,$7C,$A7
	.dc.b	$E6,$FD,$E3,$CF,$FD,$55,$FF,$8B,$A3,$13,$FA,$B8,$C4,$EA,$3E,$A8
	.dc.b	$9D,$F4,$F5,$86,$7F,$01,$FA,$07,$D2,$33,$C1,$F5,$5E,$7B,$BC,$EF
	.dc.b	$C0,$FC,$87,$38,$C4,$A3,$13,$AC,$FF,$C8,$B5,$A9,$F7,$47,$E3,$BE
	.dc.b	$B3,$E3,$5F,$E3,$23,$0F,$C6,$31,$CA,$9D,$C7,$E4,$5B,$3F,$98,$FB
	.dc.b	$4F,$CA,$FA,$BD,$AD,$18,$9F,$CD,$B6,$FA,$B5,$F8,$67,$E0,$F8,$3F
	.dc.b	$74,$FC,$4E,$2F,$D4,$63,$E0,$C3,$C9,$3E,$73,$C8,$5C,$3D,$77,$F4
	.dc.b	$76,$FA,$13,$CE,$3A,$C6,$09,$B7,$D1,$9F,$19,$0F,$F3,$77,$A7,$F9
	.dc.b	$17,$F6,$7E,$3F,$E1,$FC,$9F,$DA,$CE,$67,$78,$7F,$9C,$B9,$15,$A8
	.dc.b	$C3,$67,$9C,$E9,$57,$3B,$ED,$18,$BE,$2D,$3F,$82,$E3,$FC,$91,$FE
	.dc.b	$C6,$D7,$FD,$0E,$D7,$F6,$06,$B1,$FE,$DC,$DD,$7F,$B5,$BF,$44,$C1
	.dc.b	$FC,$5F,$4F,$A2,$FD,$FD,$7F,$A3,$BF,$88,$6B,$F5,$0F,$A5,$5D,$9F
	.dc.b	$1B,$5F,$19,$8A,$FE,$F7,$AF,$E3,$61,$E9,$FE,$37,$7F,$E4,$6D,$8C
	.dc.b	$CF,$E3,$9A,$78,$CF,$CF,$14,$F4,$18,$3C,$CF,$EA,$38,$FF,$86,$36
	.dc.b	$3F,$C2,$4D,$BF,$D7,$C6,$0F,$49,$0F,$A7,$87,$B5,$B1,$F3,$39,$7F
	.dc.b	$85,$F6,$FF,$05,$6D,$FF,$0C,$A4,$E7,$87,$FA,$0B,$FC,$5E,$FF,$8E
	.dc.b	$4D,$E7,$FE,$4A,$DE,$8F,$A3,$3A,$C3,$E6,$4F,$DA,$29,$0F,$E1,$FD
	.dc.b	$3F,$D3,$B9,$7F,$9B,$C7,$E4,$CC,$94,$FE,$7B,$1B,$7F,$AB,$F5,$FE
	.dc.b	$22,$BF,$A5,$47,$DA,$E2,$FF,$10,$3F,$89,$9F,$C6,$BF,$8D,$3C,$47
	.dc.b	$C6,$3C,$C7,$F1,$2E,$CE,$93,$FD,$76,$3A,$46,$0C,$BF,$D9,$BE,$0B
	.dc.b	$F2,$CD,$3F,$B5,$94,$FE,$3A,$53,$E7,$F7,$FF,$78,$7F,$DD,$BB,$7F
	.dc.b	$78,$6C,$65,$FE,$7F,$19,$7D,$27,$AC,$FC,$11,$4F,$F0,$9F,$84,$F0
	.dc.b	$BE,$33,$63,$F3,$ED,$7F,$C0,$8D,$7A,$F4,$7E,$E5,$F8,$C7,$C8,$D3
	.dc.b	$FC,$6D,$AF,$42,$47,$47,$BF,$04,$8E,$ED,$7B,$F8,$7E,$42,$3E,$47
	.dc.b	$EB,$DF,$86,$3D,$84,$71,$E4,$FF,$9B,$73,$FF,$28,$92,$7F,$CA,$A7
	.dc.b	$E7,$0F,$F3,$60,$F3,$3F,$CB,$A4,$72,$FF,$98,$DF,$C8,$0F,$27,$FC
	.dc.b	$D5,$C7,$92,$47,$47,$FC,$CE,$47,$07,$89,$FF,$32,$91,$C1,$FC,$45
	.dc.b	$FE,$44,$47,$07,$F7,$E3,$FD,$FC,$7F,$A4,$3C,$97,$97,$F9,$C6,$C7
	.dc.b	$F2,$A6,$BB,$8D,$11,$D3,$5D,$18,$21,$82,$1E,$38,$EA,$D8,$E4,$D3
	.dc.b	$89,$4E,$27,$F8,$62,$38,$B6,$38,$B4,$E2,$53,$89,$FE,$54,$8E,$2D
	.dc.b	$8E,$2D,$38,$94,$E2,$7C,$A3,$C8,$43,$B1,$4E,$05,$38,$1A,$21,$A2
	.dc.b	$9A,$30,$43,$04,$8F,$4B,$63,$D2,$D3,$D2,$53,$D3,$1C,$09,$1C,$1B
	.dc.b	$1C,$1A,$F0,$8D,$11,$D3,$5D,$18,$21,$82,$1F,$0E,$3A,$B6,$39,$34
	.dc.b	$E2,$53,$89,$F1,$A3,$8B,$63,$8B,$4E,$25,$38,$9F,$2E,$38,$B6,$38
	.dc.b	$B4,$E2,$53,$89,$F2,$CF,$21,$0E,$C5,$38,$14,$E0,$68,$86,$8A,$68
	.dc.b	$C1,$0C,$12,$3D,$2D,$8F,$4B,$4F,$49,$4F,$4C,$70,$24,$70,$6C,$70
	.dc.b	$6B,$C2,$34,$47,$4D,$74,$7F,$8C,$BC,$0F,$F2,$0B,$FC,$45,$AF,$F9
	.dc.b	$08,$C7,$4F,$F2,$65,$4F,$9B,$FF,$64,$F8,$3F,$C5,$DB,$3F,$C2,$5A
	.dc.b	$FF,$03,$6B,$FB,$D6,$BD,$1B,$3D,$63,$FC,$C0,$6C,$A6,$CA,$EE,$7C
	.dc.b	$51,$89,$E6,$47,$04,$8D,$91,$F4,$19,$2B,$97,$D4,$64,$AE,$5F,$33
	.dc.b	$5F,$33,$67,$CD,$1D,$08,$72,$29,$C8,$A7,$22,$BC,$9E,$04,$38,$14
	.dc.b	$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70,$49,$F9,$93,$ED,$6C,$F9,$9A
	.dc.b	$F8,$5A,$F6,$6B,$D5,$B3,$D6,$3E,$71,$B2,$9B,$2B,$B9,$F1,$46,$27
	.dc.b	$99,$1C,$12,$36,$47,$D0,$64,$AE,$5F,$51,$92,$B9,$7C,$CD,$7C,$CD
	.dc.b	$9F,$34,$74,$21,$C8,$A7,$22,$9C,$8A,$F2,$78,$10,$E0,$53,$81,$5E
	.dc.b	$0E,$88,$68,$AE,$9C,$11,$C1,$27,$E6,$4F,$B5,$B3,$E6,$6B,$E1,$6B
	.dc.b	$D9,$AF,$56,$CF,$58,$F9,$C6,$CA,$79,$36,$76,$BA,$DE,$3F,$FB,$3C
	.dc.b	$95,$FE,$C1,$7F,$C5,$CB,$EB,$9E,$6A,$38,$9F,$E3,$38,$CB,$83,$47
	.dc.b	$18,$E1,$3D,$27,$6F,$6E,$36,$BA,$BF,$F1,$F4,$61,$E8,$78,$5C,$4F
	.dc.b	$19,$E1,$3E,$09,$7A,$18,$8E,$0D,$51,$A7,$A9,$88,$D7,$9F,$FE,$4A
	.dc.b	$BF,$86,$D9,$FF,$29,$35,$F7,$B5,$74,$BB,$DF,$69,$F2,$7C,$CF,$F7
	.dc.b	$D3,$6B,$E3,$D6,$E3,$11,$F3,$E3,$0A,$7E,$03,$4F,$8F,$9F,$07,$E3
	.dc.b	$35,$F7,$B7,$83,$F0,$7B,$FF,$46,$6F,$1F,$E0,$1B,$5F,$E6,$F6,$BE
	.dc.b	$3C,$9F,$65,$FD,$5B,$2F,$73,$63,$33,$FD,$52,$72,$D8,$7D,$9F,$E2
	.dc.b	$FC,$CF,$F3,$E5,$7C,$78,$C5,$7C,$4A,$DA,$AD,$C6,$1B,$C7,$E1,$24
	.dc.b	$F7,$7F,$B4,$E3,$EA,$8F,$CD,$35,$34,$D6,$1F,$67,$E7,$BD,$84,$9F
	.dc.b	$53,$D3,$97,$F5,$9A,$7C,$96,$CF,$C9,$6B,$FA,$4B,$F1,$76,$BF,$14
	.dc.b	$FD,$4D,$FF,$B8,$1B,$5F,$ED,$E6,$BF,$B1,$9E,$19,$7C,$FF,$90,$FD
	.dc.b	$A7,$DE,$FC,$46,$D7,$C9,$6B,$F6,$5A,$FE,$5B,$67,$A3,$F6,$49,$1F
	.dc.b	$5D,$F7,$71,$7E,$1F,$5F,$5C,$FC,$A6,$BF,$29,$CE,$FF,$A1,$FA,$DF
	.dc.b	$C2,$12,$7F,$04,$F4,$DB,$E4,$5F,$52,$38,$24,$FC,$B8,$CB,$FC,$39
	.dc.b	$7E,$62,$CF,$FC,$5F,$86,$63,$14,$BF,$DA,$BF,$C6,$EF,$8D,$F0,$8C
	.dc.b	$57,$0A,$DC,$7B,$EB,$C2,$FD,$BC,$7E,$0F,$07,$E2,$5A,$49,$DC,$FF
	.dc.b	$04,$FA,$47,$17,$F2,$BF,$6F,$AE,$89,$CD,$27,$31,$86,$D7,$1A,$FD
	.dc.b	$37,$CA,$7A,$6F,$EE,$36,$3F,$89,$37,$80,$E7,$0F,$C0,$D7,$EB,$6B
	.dc.b	$F8,$9C,$63,$97,$C3,$C3,$FB,$79,$CB,$6B,$E1,$67,$EE,$7F,$04,$FC
	.dc.b	$09,$F0,$C7,$D1,$7C,$ED,$3F,$86,$61,$E5,$3D,$4F,$2D,$BF,$02,$F6
	.dc.b	$6B,$F4,$DA,$EC,$FC,$33,$FD,$6B,$C2,$FE,$9B,$E2,$7F,$59,$F5,$BF
	.dc.b	$05,$5F,$D6,$E7,$2A,$32,$A3,$31,$86,$D7,$F5,$E6,$CF,$3E,$B5,$F7
	.dc.b	$AD,$46,$27,$B4,$62,$75,$1F,$8D,$3F,$46,$43,$F4,$05,$3E,$79,$4E
	.dc.b	$07,$CC,$3F,$17,$B7,$DE,$7F,$A5,$1F,$38,$A7,$E1,$8A,$7D,$83,$A1
	.dc.b	$4F,$09,$F8,$01,$EA,$7E,$11,$B9,$E6,$30,$78,$B3,$FE,$BB,$3E,$13
	.dc.b	$07,$8A,$7F,$0A,$FE,$70,$FE,$D2,$FE,$4B,$83,$FB,$13,$63,$27,$F6
	.dc.b	$5E,$0F,$D3,$C7,$ED,$A3,$D0,$7C,$63,$D2,$52,$1F,$20,$A7,$03,$FC
	.dc.b	$4D,$C9,$FC,$21,$FE,$03,$BF,$F0,$67,$E4,$B5,$CC,$6E,$7B,$C7,$52
	.dc.b	$1D,$49,$C9,$FD,$C9,$F9,$ED,$7F,$B0,$73,$7F,$8C,$9F,$E6,$CD,$9F
	.dc.b	$E5,$91,$FE,$36,$6B,$CB,$97,$FC,$7C,$DC,$35,$AF,$EF,$A3,$FB,$4F
	.dc.b	$1F,$38,$C4,$7F,$6F,$8C,$73,$7F,$8F,$FF,$12,$FE,$99,$26,$91,$FE
	.dc.b	$AD,$D3,$E7,$EA,$FE,$AB,$1F,$E4,$A6,$BF,$D7,$63,$D8,$FC,$2C,$BC
	.dc.b	$EF,$C8,$3E,$96,$7F,$DC,$5F,$4F,$07,$9C,$8F,$C5,$DB,$E4,$3F,$69
	.dc.b	$AF,$D1,$9F,$F1,$95,$FB,$C9,$1F,$11,$F8,$5D,$9F,$E5,$DE,$3F,$F0
	.dc.b	$6B,$67,$FC,$24,$D7,$FC,$44,$D7,$FB,$57,$6F,$CB,$23,$F4,$1F,$7E
	.dc.b	$1E,$48,$F8,$1C,$1E,$9C,$FE,$0C,$7F,$96,$79,$FF,$BF,$3F,$E3,$C1
	.dc.b	$AF,$F9,$00,$8F,$F8,$97,$BF,$6F,$F8,$D5,$AF,$F7,$D6,$EB,$FE,$4A
	.dc.b	$9F,$F2,$DF,$8F,$FE,$7B,$E8,$B7,$FC,$54,$FF,$2D,$C6,$5F,$F2,$FB
	.dc.b	$5F,$F3,$77,$87,$5F,$F3,$69,$C4,$B3,$FE,$7D,$8D,$47,$F9,$48,$D1
	.dc.b	$63,$FC,$B6,$47,$FD,$19,$C1,$C3,$C5,$FF,$46,$CF,$FA,$26,$30,$74
	.dc.b	$DB,$FC,$B7,$1F,$E0,$A2,$57,$F8,$1D,$FF,$4D,$F8,$CF,$1F,$F2,$2D
	.dc.b	$9F,$D7,$35,$DC,$68,$8E,$9A,$E8,$C1,$0C,$10,$FF,$51,$88,$EC,$E1
	.dc.b	$AF,$FA,$A1,$AF,$FA,$BD,$AF,$18,$F0,$BC,$5A,$F1,$6B,$C5,$AF,$13
	.dc.b	$CC,$71,$29,$C4,$A7,$12,$9C,$4F,$61,$B2,$9B,$29,$B3,$B1,$88,$F5
	.dc.b	$92,$3C,$AD,$8E,$ED,$78,$46,$88,$E9,$AE,$8C,$10,$C1,$0F,$19,$0F
	.dc.b	$51,$4F,$51,$4F,$51,$A2,$1A,$29,$A3,$04,$30,$48,$F6,$3D,$5A,$F2
	.dc.b	$6B,$C5,$AF,$18,$F7,$BC,$5A,$F1,$6B,$C5,$AF,$13,$E2,$1C,$4A,$71
	.dc.b	$29,$C4,$A7,$13,$E5,$1B,$29,$B2,$9B,$3B,$18,$8F,$59,$23,$CA,$D8
	.dc.b	$EE,$D7,$84,$68,$8E,$9A,$E8,$C1,$0C,$10,$F1,$90,$F5,$14,$F5,$14
	.dc.b	$F5,$1A,$21,$A2,$9F,$7B,$BB,$FC,$96,$FF,$00,$6B,$FE,$3F,$3F,$CA
	.dc.b	$9B,$FF,$22,$5F,$53,$FE,$94,$EF,$FE,$0F,$18,$E0,$3F,$E5,$42,$CE
	.dc.b	$DF,$E1,$AD,$76,$71,$38,$16,$7A,$C6,$E3,$9C,$F6,$9C,$B6,$3A,$B5
	.dc.b	$EB,$3C,$5E,$86,$8E,$EE,$49,$1C,$63,$B9,$A7,$0F,$23,$AB,$88,$E3
	.dc.b	$1B,$24,$F9,$9A,$78,$5C,$11,$C1,$0E,$A7,$02,$BB,$69,$B3,$CE,$E5
	.dc.b	$A6,$4F,$41,$83,$89,$0C,$10,$D9,$23,$DB,$1D,$63,$13,$F0,$48,$F9
	.dc.b	$CE,$53,$C1,$F1,$35,$F0,$1C,$8E,$05,$9E,$D1,$B8,$E7,$3D,$A7,$2D
	.dc.b	$8E,$AD,$7A,$CF,$17,$A1,$A3,$BB,$92,$47,$18,$EE,$69,$C3,$C8,$EA
	.dc.b	$E2,$38,$C6,$C9,$3E,$66,$9E,$17,$04,$70,$43,$A9,$C0,$AE,$DA,$6C
	.dc.b	$F3,$B9,$69,$93,$D0,$60,$E2,$43,$04,$36,$48,$F6,$C7,$58,$C4,$FC
	.dc.b	$12,$3E,$73,$94,$F0,$7C,$4D,$7C,$07,$23,$81,$67,$B4,$6E,$39,$CF
	.dc.b	$69,$CB,$63,$F3,$D9,$B8,$9C,$6F,$1F,$F8,$BC,$E6,$31,$5F,$E2,$08
	.dc.b	$C4,$EA,$B8,$BF,$C7,$BC,$97,$67,$F1,$E8,$C3,$6F,$FC,$53,$3A,$7A
	.dc.b	$9F,$C7,$9C,$4F,$F2,$8F,$79,$F0,$DC,$6B,$F9,$1C,$61,$E2,$7F,$2C
	.dc.b	$71,$3C,$67,$84,$FF,$8C,$78,$EB,$FE,$39,$6F,$11,$A6,$BC,$B7,$C0
	.dc.b	$D2,$57,$80,$46,$9C,$6B,$F2,$7E,$03,$E3,$6E,$69,$27,$F9,$EF,$03
	.dc.b	$C2,$7C,$3F,$F4,$6F,$37,$FC,$9F,$39,$8C,$57,$0A,$DA,$AD,$C6,$1F
	.dc.b	$0F,$14,$9F,$B5,$C7,$F8,$08,$C4,$7F,$4E,$CF,$F0,$1C,$9E,$82,$47
	.dc.b	$0D,$3F,$AC,$0F,$AF,$7A,$3F,$D7,$3E,$5F,$FC,$CB,$39,$8C,$57,$F5
	.dc.b	$A8,$C2,$9D,$D7,$28,$C1,$ED,$FB,$2F,$E9,$DB,$FC,$56,$EC,$69,$8F
	.dc.b	$83,$FE,$DB,$D3,$F0,$9B,$7F,$1D,$AF,$F7,$2B,$FA,$11,$88,$FE,$EB
	.dc.b	$AF,$C2,$7E,$06,$BF,$09,$F8,$DB,$FE,$A2,$3E,$AF,$99,$F4,$B1,$BC
	.dc.b	$5C,$F2,$4E,$70,$FB,$ED,$79,$EF,$FD,$F3,$2F,$04,$63,$1F,$03,$FB
	.dc.b	$6F,$B1,$7D,$37,$CC,$8F,$F0,$5D,$A7,$31,$F2,$8C,$B8,$E1,$FA,$AD
	.dc.b	$8F,$C5,$7A,$BF,$E1,$4F,$43,$F1,$2D,$8C,$88,$CB,$4F,$C9,$78,$DE
	.dc.b	$35,$AF,$E6,$E3,$0D,$8F,$21,$AE,$34,$F1,$86,$BF,$09,$FA,$7C,$3F
	.dc.b	$3D,$FA,$64,$3F,$C5,$7F,$10,$FB,$3F,$E3,$91,$DA,$31,$3E,$56,$C6
	.dc.b	$67,$7A,$FE,$A3,$C5,$FF,$1A,$FD,$45,$A3,$48,$F8,$4F,$2F,$07,$D9
	.dc.b	$C7,$CD,$1F,$64,$FD,$07,$DA,$3E,$27,$E3,$52,$31,$3F,$12,$31,$3A
	.dc.b	$9C,$E3,$FB,$CF,$AC,$FF,$24,$FC,$37,$CE,$C3,$F3,$5C,$09,$CF,$0F
	.dc.b	$F0,$56,$CE,$9A,$FA,$39,$BF,$1A,$FD,$0D,$3D,$2E,$4F,$E1,$DF,$67
	.dc.b	$FE,$58,$FC,$4F,$C1,$1F,$66,$31,$86,$A1,$3A,$8F,$E3,$93,$A9,$FD
	.dc.b	$87,$C9,$7A,$7B,$34,$DA,$FE,$46,$D9,$DC,$7E,$F0,$CF,$D7,$FF,$0F
	.dc.b	$6C,$7F,$0B,$6E,$DD,$27,$53,$F4,$C8,$7B,$4B,$2A,$F9,$95,$E4,$8E
	.dc.b	$33,$F9,$C9,$CC,$79,$8C,$C7,$C9,$E1,$FC,$11,$34,$F9,$BC,$DF,$9B
	.dc.b	$9D,$46,$35,$F9,$F0,$F3,$9E,$73,$CF,$FF,$43,$FB,$8C,$3F,$D2,$0D
	.dc.b	$BF,$F4,$5D,$F0,$AD,$D6,$89,$3A,$BE,$8A,$F9,$3E,$18,$CC,$62,$FA
	.dc.b	$4E,$EB,$F8,$BB,$FD,$70,$87,$F4,$52,$9A,$30,$43,$ED,$8F,$33,$E7
	.dc.b	$37,$4F,$98,$D3,$71,$F9,$07,$E7,$1D,$8F,$D8,$96,$3E,$D6,$FF,$EA
	.dc.b	$E9,$FC,$14,$F3,$3E,$D4,$66,$7F,$13,$3E,$2F,$95,$91,$E2,$6D,$7E
	.dc.b	$BB,$8B,$FD,$73,$F8,$AF,$E7,$79,$91,$8B,$FD,$14,$E6,$31,$5F,$0C
	.dc.b	$87,$A0,$F8,$87,$F2,$3E,$6F,$F6,$2F,$C1,$75,$6D,$FE,$EE,$53,$96
	.dc.b	$C7,$F0,$EE,$AF,$81,$14,$F9,$C5,$CF,$FB,$AF,$57,$F1,$07,$E5,$47
	.dc.b	$CC,$3C,$65,$3F,$86,$F6,$7F,$B6,$3E,$B3,$F9,$21,$23,$FC,$25,$19
	.dc.b	$7F,$DA,$FD,$FF,$CB,$AB,$FD,$B3,$08,$C3,$AE,$FF,$E6,$F3,$A9,$CC
	.dc.b	$FC,$63,$C4,$7C,$73,$E4,$E5,$FD,$2E,$7F,$C3,$FE,$07,$C0,$9F,$E4
	.dc.b	$6D,$7F,$A7,$36,$7F,$C5,$DE,$17,$FB,$CB,$E6,$BF,$AD,$92,$BF,$AF
	.dc.b	$BF,$D9,$BC,$5E,$5D,$BF,$DB,$5B,$E2,$FF,$6E,$CB,$57,$F3,$8F,$47
	.dc.b	$77,$F9,$A5,$F9,$AD,$7F,$C0,$3E,$39,$74,$7F,$82,$1A,$68,$A7,$F8
	.dc.b	$5F,$C7,$FF,$1F,$C7,$4F,$03,$FC,$F3,$37,$3F,$9E,$FF,$6D,$F2,$3B
	.dc.b	$57,$F8,$8E,$39,$99,$2E,$FF,$E2,$26,$F5,$FF,$8B,$5E,$71,$FE,$31
	.dc.b	$6C,$65,$FF,$27,$F9,$3E,$07,$FD,$17,$E5,$7F,$8E,$F6,$FF,$1E,$F9
	.dc.b	$7F,$2A,$31,$3F,$E4,$36,$BF,$23,$4F,$5B,$F9,$2D,$FE,$43,$FE,$26
	.dc.b	$D7,$FC,$D6,$47,$FC,$A8,$FF,$98,$37,$F9,$9E,$C5,$93,$48,$FF,$96
	.dc.b	$FE,$17,$FC,$EA,$47,$07,$E6,$3D,$1F,$F3,$E7,$8F,$F3,$1B,$19,$F4
	.dc.b	$BE,$D4,$E5,$B1,$FA,$0D,$BF,$D1,$8F,$F3,$03,$F0,$07,$FA,$44,$C1
	.dc.b	$FE,$95,$29,$99,$FF,$4D,$38,$F5,$BF,$D4,$11,$82,$47,$FA,$B3,$D1
	.dc.b	$F0,$BF,$EA,$B6,$BF,$E6,$46,$AF,$37,$FD,$65,$3F,$E9,$62,$47,$F9
	.dc.b	$FD,$AF,$A3,$D7,$FE,$64,$DE,$5F,$C0,$28,$CB,$F9,$D6,$B9,$8D,$C6
	.dc.b	$08,$E0,$91,$F8,$B8,$CB,$63,$2F,$E7,$A3,$2D,$8C,$BF,$C4,$23,$2D
	.dc.b	$8C,$BE,$93,$A1,$0E,$45,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$10
	.dc.b	$F8,$CF,$91,$A7,$72,$9D,$0A,$73,$29,$CC,$F9,$C7,$12,$9C,$4A,$71
	.dc.b	$29,$C6,$3E,$3B,$63,$C8,$E8,$EE,$43,$04,$8E,$2E,$5A,$E6,$39,$C6
	.dc.b	$08,$E0,$91,$E9,$8C,$B6,$32,$FA,$E3,$2D,$8C,$BE,$D8,$CB,$63,$2F
	.dc.b	$B4,$E8,$43,$91,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3E,$33
	.dc.b	$E4,$69,$DC,$A7,$42,$9C,$CA,$73,$3E,$71,$C4,$A7,$12,$9C,$4A,$71
	.dc.b	$8F,$8E,$D8,$F2,$3A,$3B,$90,$C1,$23,$8B,$96,$B9,$8E,$71,$8F,$2B
	.dc.b	$FC,$9B,$3F,$E0,$66,$BF,$C9,$9A,$FF,$96,$6B,$F8,$27,$93,$FD,$A7
	.dc.b	$CA,$FE,$04,$D9,$FE,$00,$D7,$F8,$C3,$5F,$E2,$8D,$7F,$8C,$36,$CD
	.dc.b	$24,$7F,$99,$8E,$25,$38,$96,$38,$B6,$38,$CF,$9A,$30,$47,$04,$7D
	.dc.b	$26,$4A,$E5,$F5,$99,$2B,$97,$CA,$D7,$CA,$D9,$F2,$B6,$7C,$B3,$D8
	.dc.b	$87,$42,$9D,$0A,$74,$2C,$F4,$6C,$74,$79,$10,$E4,$53,$91,$63,$93
	.dc.b	$63,$93,$C0,$87,$02,$C7,$06,$C7,$07,$44,$8D,$36,$35,$18,$23,$82
	.dc.b	$47,$CF,$3A,$94,$E4,$58,$E2,$D8,$E3,$3F,$3E,$30,$47,$04,$7E,$49
	.dc.b	$92,$B9,$7E,$21,$92,$B9,$7C,$AD,$7C,$AD,$9F,$2B,$67,$CB,$3D,$88
	.dc.b	$74,$29,$D0,$A7,$42,$CF,$46,$C7,$47,$91,$0E,$45,$39,$16,$39,$36
	.dc.b	$39,$3C,$08,$70,$2C,$70,$6C,$70,$74,$48,$D3,$63,$51,$82,$38,$3F
	.dc.b	$83,$71,$B9,$1C,$8F,$5F,$FC,$DF,$39,$8C,$61,$FE,$A0,$8C,$4E,$AB
	.dc.b	$8B,$FE,$76,$F6,$3F,$C8,$AD,$F5,$FF,$D0,$8D,$BE,$67,$F9,$22,$76
	.dc.b	$FF,$AB,$38,$5E,$C7,$08,$C3,$5F,$F2,$7C,$69,$E6,$7F,$A4,$5C,$57
	.dc.b	$FA,$23,$85,$F1,$BC,$6B,$FD,$2D,$18,$78,$9F,$E4,$98,$D3,$4C,$F3
	.dc.b	$3D,$AC,$B6,$33,$18,$38,$EB,$F2,$7D,$AF,$8D,$B9,$A4,$AF,$F6,$5E
	.dc.b	$BF,$E6,$C8,$F8,$C4,$8C,$ED,$FE,$AD,$8E,$7C,$D2,$7E,$A7,$07,$DA
	.dc.b	$8C,$47,$CD,$AF,$F5,$A4,$7D,$12,$47,$40,$F9,$B8,$E4,$74,$C7,$FA
	.dc.b	$26,$5F,$83,$9F,$F7,$0F,$33,$D4,$7A,$7F,$EC,$1F,$57,$31,$8C,$51
	.dc.b	$89,$D4,$7F,$A2,$2B,$94,$62,$3F,$D9,$51,$D7,$E3,$3F,$AF,$36,$3F
	.dc.b	$AC,$37,$0E,$F3,$AC,$3F,$39,$5F,$E8,$33,$B3,$8E,$3F,$2E,$DF,$82
	.dc.b	$9C,$B5,$FF,$4D,$E7,$EF,$7E,$16,$1F,$11,$CC,$FE,$A2,$3D,$FF,$65
	.dc.b	$C6,$59,$7C,$58,$C3,$63,$C8,$7B,$B1,$4F,$18,$61,$E1,$7B,$EB,$F4
	.dc.b	$B1,$F0,$47,$90,$F6,$FC,$B7,$C9,$7C,$38,$23,$18,$FD,$76,$C6,$67
	.dc.b	$EA,$9F,$1B,$EC,$FF,$54,$D1,$D1,$F7,$23,$31,$E9,$6B,$E0,$FA,$6E
	.dc.b	$A4,$7A,$9E,$F7,$CC,$FB,$B1,$F5,$BE,$ED,$7E,$72,$35,$9F,$E0,$31
	.dc.b	$7D,$3F,$A8,$93,$D8,$F8,$2F,$BA,$B9,$7A,$DF,$94,$6C,$FE,$11,$AF
	.dc.b	$D3,$6B,$E1,$DB,$E9,$37,$1F,$3B,$71,$FA,$87,$F8,$AC,$64,$F6,$7E
	.dc.b	$A4,$61,$B1,$99,$F4,$4E,$5F,$56,$FF,$AD,$FA,$8F,$85,$F6,$5F,$8C
	.dc.b	$9F,$C4,$46,$27,$F3,$AD,$8C,$89,$E0,$F9,$F3,$F9,$AF,$CC,$7E,$7D
	.dc.b	$7E,$EF,$EA,$FE,$87,$E2,$DC,$23,$14,$8C,$70,$2B,$75,$FA,$E6,$D7
	.dc.b	$E7,$7E,$B7,$F9,$7C,$FE,$C9,$B1,$93,$FD,$F1,$19,$69,$FA,$5F,$AE
	.dc.b	$FE,$19,$1F,$BA,$6C,$64,$FE,$1B,$19,$69,$F8,$B3,$F4,$C5,$21,$FA
	.dc.b	$69,$E6,$D3,$89,$7A,$BD,$F1,$A8,$F9,$07,$43,$E6,$14,$C9,$D8,$A6
	.dc.b	$8F,$F4,$17,$C0,$BC,$7F,$CF,$76,$35,$CC,$61,$B5,$FC,$B9,$B3,$B8
	.dc.b	$FF,$45,$19,$2B,$95,$E4,$2F,$C1,$1B,$32,$E2,$78,$72,$FF,$80,$67
	.dc.b	$FA,$C9,$0F,$E8,$85,$3F,$8A,$94,$FD,$21,$FD,$0B,$1F,$DD,$79,$3F
	.dc.b	$2A,$0D,$26,$9D,$22,$1D,$0A,$7F,$1C,$D7,$9C,$7D,$F9,$FC,$9C,$62
	.dc.b	$7F,$09,$1D,$FD,$29,$5F,$8D,$6C,$FF,$31,$DF,$F8,$77,$E2,$FF,$9C
	.dc.b	$6D,$7E,$6E,$53,$96,$C7,$91,$FE,$2A,$7C,$12,$9F,$B4,$1F,$8C,$7E
	.dc.b	$A3,$97,$F8,$E3,$63,$33,$B8,$FD,$BF,$37,$F0,$3C,$BF,$91,$57,$F2
	.dc.b	$38,$C3,$86,$B9,$E7,$FE,$19,$3E,$3D,$BE,$D4,$F9,$DA,$FF,$1F,$1E
	.dc.b	$0F,$F2,$56,$9F,$C9,$4A,$6D,$FE,$FB,$D9,$F7,$23,$FB,$6B,$5F,$EB
	.dc.b	$8D,$7F,$99,$9F,$80,$3C,$5D,$7F,$E5,$07,$CC,$D7,$FB,$BE,$1C,$73
	.dc.b	$FF,$07,$37,$A3,$F9,$6C,$F3,$3F,$A6,$14,$FE,$C3,$DB,$FD,$27,$F1
	.dc.b	$9F,$EB,$10,$69,$1F,$ED,$27,$1B,$FE,$CC,$DA,$FE,$CD,$A7,$F8,$73
	.dc.b	$C0,$F6,$4F,$DE,$3F,$2C,$43,$F4,$27,$F8,$43,$C1,$FE,$B5,$C9,$FE
	.dc.b	$06,$83,$48,$FC,$4E,$0F,$BE,$F3,$3B,$1F,$DF,$35,$E6,$FF,$86,$0E
	.dc.b	$65,$9F,$F0,$EB,$83,$FC,$1D,$BF,$46,$97,$3F,$F1,$63,$FE,$2A,$31
	.dc.b	$9F,$F8,$E7,$3F,$F1,$C4,$7C,$3E,$7F,$A4,$7C,$2E,$1F,$C5,$C7,$C0
	.dc.b	$3D,$0E,$17,$8C,$F1,$7F,$35,$1F,$E5,$1D,$7F,$D0,$84,$7F,$CA,$0F
	.dc.b	$F9,$A0,$8E,$0F,$F2,$C7,$9F,$FE,$8B,$23,$83,$FC,$B0,$FF,$9C,$08
	.dc.b	$E0,$F9,$BE,$4F,$F9,$E7,$BF,$E2,$36,$33,$E4,$FF,$84,$67,$2D,$8F
	.dc.b	$AB,$E3,$FF,$A2,$67,$FD,$1D,$E5,$66,$7F,$6F,$39,$77,$19,$7F,$D2
	.dc.b	$DE,$65,$7F,$C5,$AD,$F2,$7F,$D3,$97,$F5,$DB,$3F,$E7,$78,$FF,$55
	.dc.b	$11,$C1,$FE,$77,$7F,$D1,$E4,$70,$7E,$33,$CE,$79,$DF,$83,$8F,$F1
	.dc.b	$74,$61,$C3,$5F,$C4,$35,$E5,$1C,$54,$65,$FE,$2E,$D7,$AC,$6E,$31
	.dc.b	$1C,$23,$04,$7C,$2D,$7C,$2D,$7C,$26,$88,$68,$A6,$8C,$10,$C1,$0F
	.dc.b	$3B,$D5,$A7,$22,$9C,$4A,$71,$3D,$A6,$CA,$6C,$A6,$CF,$80,$6C,$A6
	.dc.b	$CA,$6E,$3E,$39,$23,$CD,$1D,$0F,$19,$0C,$10,$E6,$64,$A6,$63,$A3
	.dc.b	$E3,$6B,$E3,$6B,$E3,$8F,$0B,$C8,$91,$A7,$B3,$5E,$D1,$C2,$31,$1C
	.dc.b	$23,$04,$7C,$4D,$7C,$4D,$7C,$46,$88,$68,$A6,$8C,$10,$C1,$0F,$3B
	.dc.b	$D5,$A7,$22,$9C,$4A,$71,$3D,$A6,$CA,$6C,$A6,$CF,$80,$6C,$A6,$CA
	.dc.b	$6E,$3E,$39,$23,$CD,$1D,$0F,$19,$0C,$10,$E6,$64,$A6,$63,$A3,$E3
	.dc.b	$6B,$E3,$6B,$E3,$8F,$0B,$C8,$91,$A7,$B7,$8B,$FE,$4F,$AF,$E1,$CD
	.dc.b	$8F,$EA,$4D,$7F,$AF,$B5,$FE,$13,$D3,$FE,$C1,$F1,$BF,$88,$B6,$3F
	.dc.b	$8D,$35,$D9,$FE,$5F,$3F,$8D,$B9,$69,$93,$F8,$BB,$96,$99,$BF,$F7
	.dc.b	$9B,$63,$51,$89,$F1,$CE,$48,$E0,$8F,$98,$C9,$5C,$BE,$86,$BE,$66
	.dc.b	$BE,$46,$C7,$86,$39,$10,$E4,$53,$91,$4E,$45,$79,$3C,$08,$70,$29
	.dc.b	$C0,$AF,$07,$44,$34,$57,$4E,$08,$E0,$93,$F3,$09,$1F,$18,$EE,$E2
	.dc.b	$7E,$0B,$63,$D4,$D7,$67,$13,$C4,$E5,$A6,$4F,$1B,$96,$99,$9F,$2B
	.dc.b	$63,$51,$89,$F1,$CE,$48,$E0,$8F,$98,$C9,$5C,$BE,$86,$BE,$66,$BE
	.dc.b	$46,$C7,$86,$39,$10,$E4,$53,$91,$4E,$45,$79,$3C,$08,$70,$29,$C0
	.dc.b	$AF,$07,$44,$34,$57,$4E,$08,$E0,$93,$F3,$09,$1F,$18,$EE,$E2,$7E
	.dc.b	$0B,$63,$D4,$D7,$67,$13,$C4,$E5,$A7,$F0,$1E,$37,$53,$89,$E7,$7F
	.dc.b	$99,$67,$31,$8A,$FF,$39,$C6,$27,$55,$C5,$FF,$1B,$6C,$AE,$07,$FA
	.dc.b	$E3,$64,$66,$77,$3A,$3F,$C8,$65,$1C,$0E,$0E,$11,$FE,$4A,$35,$19
	.dc.b	$6C,$69,$FF,$63,$FE,$B2,$EA,$C4,$79,$63,$0F,$76,$DF,$FA,$15,$B1
	.dc.b	$FE,$92,$E9,$7C,$4F,$28,$FF,$7C,$B6,$33,$1E,$96,$D7,$FA,$3D,$7A
	.dc.b	$7B,$24,$ED,$E9,$A7,$C6,$87,$A3,$F1,$88,$FF,$A9,$23,$51,$FE,$70
	.dc.b	$8D,$9F,$15,$C1,$1E,$1A,$FD,$5F,$47,$E8,$92,$32,$1F,$F4,$1F,$D1
	.dc.b	$7D,$BF,$A5,$B2,$69,$1F,$C2,$F0,$FE,$2F,$97,$F0,$BC,$9F,$8F,$8C
	.dc.b	$4E,$A3,$FA,$B7,$0B,$E9,$DF,$4F,$F8,$9E,$5F,$F6,$0C,$E6,$31,$7F
	.dc.b	$8D,$3E,$E5,$70,$7F,$0E,$66,$78,$38,$3F,$28,$BC,$27,$DA,$7F,$65
	.dc.b	$8C,$36,$33,$3E,$66,$E7,$E4,$5F,$01,$F9,$FF,$4F,$0D,$AF,$14,$E9
	.dc.b	$4E,$9A,$7C,$3A,$F8,$E4,$70,$7A,$FE,$7B,$F0,$93,$F1,$1B,$19,$E2
	.dc.b	$F0,$3E,$E8,$CB,$F1,$FA,$FF,$77,$7D,$63,$1C,$1F,$51,$BC,$1F,$52
	.dc.b	$DF,$05,$F4,$27,$A3,$5F,$D6,$46,$9F,$9C,$62,$35,$B7,$CF,$FB,$6F
	.dc.b	$E0,$9F,$8B,$FC,$96,$D7,$C6,$6B,$F5,$B6,$46,$23,$F7,$50,$FB,$7E
	.dc.b	$AD,$8F,$0D,$FD,$D9,$C9,$E7,$86,$BF,$09,$F9,$A7,$C0,$D7,$E2,$3E
	.dc.b	$CF,$AF,$F4,$E8,$69,$23,$BE,$3F,$91,$F6,$FF,$85,$C7,$9B,$97,$F9
	.dc.b	$14,$63,$3F,$C8,$46,$1F,$C2,$B6,$7C,$47,$90,$FE,$55,$97,$9D,$C9
	.dc.b	$FC,$B7,$77,$94,$D3,$F5,$18,$7E,$96,$73,$18,$AE,$15,$B7,$F9,$6D
	.dc.b	$6E,$31,$CF,$E7,$CF,$F5,$9F,$89,$5C,$FF,$9E,$37,$03,$4F,$82,$FC
	.dc.b	$EE,$3F,$C6,$3E,$A6,$B9,$CF,$F8,$21,$FA,$DF,$AE,$FE,$91,$F8,$07
	.dc.b	$85,$BA,$7E,$A9,$B5,$F8,$56,$BF,$CF,$7F,$32,$FE,$B4,$FE,$BA,$7F
	.dc.b	$81,$46,$3A,$3E,$93,$63,$38,$7E,$E5,$DF,$47,$C7,$BE,$8D,$85,$7F
	.dc.b	$35,$DB,$9C,$EE,$BD,$9C,$1F,$8A,$69,$F6,$EF,$8F,$47,$3C,$3F,$51
	.dc.b	$87,$92,$30,$73,$E2,$FC,$05,$FC,$9C,$BE,$3E,$3D,$67,$33,$D8,$FC
	.dc.b	$29,$B8,$FF,$46,$FB,$CF,$2F,$FB,$A7,$A3,$F8,$1A,$79,$9F,$E9,$38
	.dc.b	$C3,$6F,$F9,$73,$67,$FB,$E3,$5E,$37,$D5,$57,$28,$C3,$6F,$C3,$7F
	.dc.b	$C3,$AB,$FA,$29,$0F,$E8,$85,$34,$60,$87,$F3,$AD,$3F,$88,$7A,$7F
	.dc.b	$26,$0D,$26,$DC,$23,$E5,$1D,$4A,$7E,$00,$E2,$D8,$FB,$B1,$83,$C1
	.dc.b	$BF,$F8,$76,$3F,$17,$AF,$F8,$7F,$C7,$B4,$69,$2F,$F3,$50,$FC,$73
	.dc.b	$F3,$CD,$8C,$C6,$0F,$D5,$EF,$FE,$32,$9F,$41,$FA,$93,$D5,$3F,$C3
	.dc.b	$17,$E3,$BF,$B5,$24,$7F,$0E,$8C,$BF,$BF,$D3,$D4,$D8,$F7,$1F,$E4
	.dc.b	$3E,$AF,$E0,$D9,$FF,$1C,$CF,$F8,$5C,$E8,$F1,$1F,$34,$A7,$D8,$3F
	.dc.b	$9D,$74,$FF,$90,$FD,$BF,$E1,$04,$C3,$C8,$FF,$25,$EE,$FE,$5D,$5F
	.dc.b	$68,$FE,$C1,$DD,$F3,$DA,$64,$FE,$BC,$E5,$F0,$1F,$70,$A7,$90,$FE
	.dc.b	$93,$DF,$B6,$9C,$CB,$97,$F7,$D8,$D6,$1F,$D6,$65,$F8,$F7,$F5,$96
	.dc.b	$9F,$D6,$4A,$7F,$59,$F0,$7F,$CC,$51,$FD,$DF,$A7,$9E,$5F,$98,$AF
	.dc.b	$EF,$91,$83,$C8,$F9,$C9,$C3,$F8,$17,$F3,$58,$2B,$FC,$8D,$C7,$FE
	.dc.b	$03,$E4,$FC,$95,$7F,$80,$CF,$C9,$16,$3F,$13,$7C,$E3,$E2,$B5,$FE
	.dc.b	$DB,$E1,$F2,$7F,$B9,$35,$C9,$D0,$C9,$4F,$EE,$BE,$27,$F8,$D5,$FF
	.dc.b	$10,$B5,$FE,$F8,$D7,$66,$71,$FF,$1B,$ED,$FE,$40,$F1,$7F,$C7,$8D
	.dc.b	$E2,$FA,$71,$8A,$FC,$86,$3E,$97,$E3,$F8,$3F,$D0,$9F,$17,$A8,$F2
	.dc.b	$3F,$D2,$9F,$21,$FE,$5A,$24,$EA,$32,$FF,$98,$B9,$FF,$D2,$A4,$70
	.dc.b	$7F,$A4,$7D,$6F,$F3,$81,$1C,$1F,$E6,$A7,$FD,$3A,$47,$07,$FA,$2F
	.dc.b	$CE,$FF,$53,$77,$FF,$A1,$5B,$19,$3F,$96,$1F,$E5,$FE,$2F,$F4,$4B
	.dc.b	$83,$F8,$87,$A1,$C1,$D3,$5C,$9B,$3F,$D2,$E5,$F4,$7F,$D4,$37,$FE
	.dc.b	$9F,$E2,$E1,$29,$CB,$F6,$DA,$FF,$9C,$7C,$5F,$F5,$85,$7F,$A4,$49
	.dc.b	$1F,$EB,$57,$FD,$7A,$47,$FD,$1A,$D7,$FC,$9E,$FF,$A1,$88,$FF,$A1
	.dc.b	$5F,$C4,$7A,$8F,$53,$F1,$12,$8C,$3F,$8B,$6B,$C2,$37,$18,$8F,$CA
	.dc.b	$35,$FF,$45,$35,$FF,$6C,$35,$ED,$1C,$88,$F7,$6B,$DC,$C1,$0C,$10
	.dc.b	$F0,$BC,$1A,$6C,$A6,$CF,$2B,$B6,$9B,$29,$B3,$D2,$ED,$A6,$CA,$6C
	.dc.b	$F6,$99,$29,$98,$F8,$44,$8E,$91,$A3,$CE,$43,$C6,$53,$B9,$4E,$06
	.dc.b	$88,$68,$A6,$8C,$10,$C1,$23,$C8,$FA,$5A,$FA,$63,$CF,$18,$8F,$13
	.dc.b	$5E,$ED,$7B,$35,$ED,$1C,$88,$F7,$6B,$DC,$C1,$0C,$10,$F0,$BC,$1A
	.dc.b	$6C,$A6,$CF,$2B,$B6,$9B,$29,$B3,$D2,$ED,$A6,$CA,$6C,$F6,$99,$29
	.dc.b	$98,$F8,$44,$8E,$91,$A3,$CE,$43,$C6,$53,$B9,$4E,$06,$88,$68,$A6
	.dc.b	$8C,$10,$C1,$23,$C8,$FA,$5A,$FA,$63,$CF,$18,$8F,$13,$FB,$DF,$1D
	.dc.b	$E3,$FF,$25,$9F,$F2,$DD,$FF,$55,$9C,$B6,$75,$4F,$65,$7F,$E6,$B3
	.dc.b	$F8,$4B,$8F,$13,$F8,$CB,$63,$F8,$8B,$5F,$E2,$2D,$78,$9C,$CF,$E3
	.dc.b	$CE,$5A,$66,$7C,$4D,$8D,$46,$27,$C3,$39,$23,$82,$3E,$43,$25,$72
	.dc.b	$F9,$8C,$95,$CB,$E8,$6B,$E6,$6C,$78,$E3,$A1,$0E,$45,$39,$14,$E4
	.dc.b	$57,$93,$C0,$87,$02,$9C,$0A,$F0,$74,$43,$45,$74,$E0,$8E,$09,$3F
	.dc.b	$38,$91,$F1,$0E,$4E,$27,$E0,$B6,$3D,$4D,$7C,$AD,$78,$9C,$CF,$0B
	.dc.b	$96,$99,$9F,$1B,$63,$51,$89,$F0,$CE,$48,$E0,$8F,$90,$C9,$5C,$BE
	.dc.b	$63,$25,$72,$FA,$1A,$F9,$9B,$1E,$38,$E8,$43,$91,$4E,$45,$39,$15
	.dc.b	$E4,$F0,$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$4F,$CE
	.dc.b	$24,$7C,$43,$93,$89,$F8,$2D,$8F,$53,$5F,$2B,$5F,$CA,$73,$39,$9C
	.dc.b	$6F,$53,$FC,$D5,$39,$8C,$57,$F9,$BE,$31,$3A,$AE,$2F,$F9,$C7,$D8
	.dc.b	$78,$DF,$C6,$23,$0D,$BF,$F3,$C4,$E9,$EA,$7F,$1E,$0E,$67,$FC,$F9
	.dc.b	$C0,$F5,$B9,$57,$F9,$15,$AF,$F2,$68,$D3,$D0,$C4,$69,$7C,$7B,$E4
	.dc.b	$30,$D8,$E0,$6A,$32,$71,$E7,$73,$D2,$4F,$08,$CB,$CB,$F3,$57,$C6
	.dc.b	$78,$5B,$19,$8C,$1E,$5E,$E7,$D3,$D2,$4F,$92,$32,$F9,$B4,$F7,$FB
	.dc.b	$5F,$1B,$5F,$E8,$ED,$93,$4F,$F5,$36,$BE,$F6,$EB,$C2,$7F,$D5,$93
	.dc.b	$96,$C6,$5F,$AD,$EB,$FE,$25,$7B,$3F,$EB,$18,$C3,$78,$FF,$AC,$B6
	.dc.b	$76,$7F,$AA,$3A,$72,$D3,$27,$E0,$F4,$FE,$D2,$E0,$FE,$C6,$E2,$7F
	.dc.b	$25,$CD,$FD,$6F,$9B,$F3,$1D,$18,$25,$FE,$37,$B9,$E9,$3E,$D7,$F1
	.dc.b	$13,$98,$C2,$E2,$FF,$67,$57,$0A,$DC,$FF,$74,$8C,$3B,$39,$7D,$B7
	.dc.b	$F6,$F8,$EE,$DB,$34,$C7,$A3,$FD,$FB,$D6,$EF,$28,$C5,$F9,$5A,$F8
	.dc.b	$63,$4A,$7D,$EF,$EA,$F3,$F8,$6F,$EE,$78,$3F,$7F,$A6,$A1,$97,$C4
	.dc.b	$FB,$AE,$B1,$8C,$51,$89,$D4,$F7,$3E,$44,$70,$58,$78,$1F,$16,$1E
	.dc.b	$07,$C5,$B7,$D7,$8C,$47,$AF,$EE,$BE,$6B,$E1,$6E,$1F,$61,$B8,$9B
	.dc.b	$E4,$7F,$89,$BF,$20,$F9,$2D,$BF,$B8,$D7,$E4,$35,$F7,$72,$74,$23
	.dc.b	$F6,$CF,$9C,$F9,$1F,$6E,$3E,$87,$FC,$6B,$9F,$BD,$B1,$97,$79,$FF
	.dc.b	$8E,$73,$FC,$8F,$C1,$7A,$89,$3D,$CF,$A0,$FC,$0F,$31,$E7,$FE,$82
	.dc.b	$BF,$35,$18,$A7,$3F,$DA,$7F,$29,$19,$8C,$47,$DE,$8C,$4F,$D2,$E1
	.dc.b	$FC,$14,$F9,$9F,$C2,$47,$E9,$BB,$7E,$79,$27,$CD,$19,$6D,$78,$FE
	.dc.b	$87,$BC,$A3,$0F,$A3,$1F,$F2,$DC,$EA,$73,$5F,$86,$CF,$F0,$84,$C7
	.dc.b	$C7,$C4,$F9,$5F,$F3,$1F,$E4,$3D,$A3,$16,$8C,$4E,$B3,$FD,$E3,$67
	.dc.b	$F5,$1F,$27,$FA,$F9,$FE,$00,$D8,$C9,$FA,$B8,$CB,$4F,$D8,$7C,$9F
	.dc.b	$F4,$F8,$FE,$10,$D8,$C9,$FC,$26,$32,$D3,$DD,$8F,$D0,$E2,$FC,$44
	.dc.b	$7D,$13,$F8,$97,$DD,$73,$7F,$64,$DA,$FE,$32,$DC,$CE,$E7,$91,$D8
	.dc.b	$FB,$C4,$30,$43,$F2,$FA,$7C,$F7,$FB,$91,$FE,$96,$1F,$00,$C7,$FB
	.dc.b	$B7,$9F,$FE,$98,$9C,$C6,$39,$9E,$7F,$FA,$62,$F8,$57,$FA,$6A,$F8
	.dc.b	$46,$23,$B3,$63,$37,$CE,$B9,$36,$BB,$9E,$83,$FA,$09,$0F,$E1,$A5
	.dc.b	$3F,$18,$53,$81,$F3,$87,$AD,$F8,$BC,$BF,$D5,$ED,$8C,$E9,$B8,$FD
	.dc.b	$5C,$7E,$F0,$E8,$53,$C2,$7E,$01,$FC,$B6,$DD,$27,$04,$9F,$C7,$F9
	.dc.b	$5F,$9B,$E0,$FF,$5F,$B6,$0D,$3F,$3A,$BF,$22,$A4,$AF,$CF,$C6,$5F
	.dc.b	$1B,$F0,$8A,$7C,$23,$F8,$3F,$37,$FB,$3B,$E7,$BF,$68,$D9,$FD,$C4
	.dc.b	$A7,$2D,$8F,$ED,$1D,$1F,$22,$29,$F3,$0B,$9F,$F5,$C8,$D4,$78,$C8
	.dc.b	$78,$C9,$3F,$C2,$6F,$FD,$C3,$F8,$0F,$80,$91,$E4,$7F,$81,$75,$F4
	.dc.b	$AF,$E5,$27,$84,$FE,$CC,$E6,$7F,$98,$C6,$4F,$01,$F6,$4F,$19,$4F
	.dc.b	$E8,$A6,$DF,$F0,$4F,$1F,$3E,$5F,$F0,$AE,$5F,$DE,$23,$1D,$BF,$EF
	.dc.b	$8F,$C5,$3F,$A9,$E0,$69,$1F,$F0,$87,$81,$CC,$8F,$7C,$3F,$AF,$3E
	.dc.b	$73,$C8,$4E,$9F,$F3,$B3,$F9,$DC,$BF,$BA,$F0,$FF,$82,$BC,$1F,$C5
	.dc.b	$CF,$F7,$63,$F1,$65,$8F,$CC,$D7,$28,$FC,$29,$F0,$6B,$FC,$4F,$E1
	.dc.b	$F2,$8F,$F0,$83,$5F,$F0,$33,$63,$71,$D8,$FF,$0B,$78,$9F,$C3,$1A
	.dc.b	$58,$FF,$12,$B7,$1F,$F1,$2B,$7C,$5F,$9C,$47,$E2,$3F,$7D,$F8,$5E
	.dc.b	$1F,$FA,$69,$F5,$B5,$FF,$26,$7B,$3E,$1F,$13,$FC,$96,$DE,$14,$A3
	.dc.b	$0D,$3F,$91,$F9,$3F,$E5,$F8,$FF,$2D,$12,$33,$7C,$A3,$FC,$D7,$0F
	.dc.b	$66,$91,$FF,$35,$99,$3F,$97,$F9,$3F,$E9,$88,$FF,$3C,$12,$33,$E7
	.dc.b	$79,$A3,$D5,$D5,$E8,$F0,$6D,$7F,$30,$3F,$9C,$E9,$FE,$86,$8C,$9F
	.dc.b	$47,$D0,$FF,$48,$B7,$D0,$E5,$5F,$D0,$27,$2D,$8C,$BF,$EB,$5E,$5F
	.dc.b	$F4,$ED,$7D,$B6,$C7,$F4,$86,$BB,$F4,$78,$BF,$3B,$D0,$FF,$57,$57
	.dc.b	$C6,$24,$7C,$47,$FD,$30,$47,$E0,$9E,$D3,$FD,$80,$53,$EC,$FA,$7F
	.dc.b	$EC,$B8,$FF,$67,$C6,$4F,$F6,$5F,$A9,$E2,$69,$9E,$9F,$F6,$F4,$FF
	.dc.b	$A2,$89,$1F,$E5,$67,$F3,$1E,$B3,$D6,$FC,$D4,$FE,$22,$30,$E1,$AF
	.dc.b	$08,$DA,$8C,$BF,$DB,$9A,$FF,$B8,$5A,$F5,$8E,$31,$88,$E5,$18,$23
	.dc.b	$E1,$6B,$E1,$30,$43,$04,$3C,$4F,$06,$9B,$29,$B3,$CC,$ED,$A6,$CA
	.dc.b	$6C,$F5,$19,$29,$93,$D6,$64,$A6,$63,$E0,$12,$3C,$11,$C8,$F2,$90
	.dc.b	$F0,$94,$D1,$82,$18,$21,$E0,$38,$14,$D9,$4D,$C7,$8D,$F4,$35,$F4
	.dc.b	$47,$99,$E0,$48,$D3,$D9,$AF,$66,$BD,$A3,$94,$62,$39,$46,$08,$F8
	.dc.b	$9A,$F8,$8C,$10,$C1,$0F,$13,$C1,$A6,$CA,$6C,$F3,$3B,$69,$B2,$9B
	.dc.b	$3D,$46,$4A,$64,$F5,$99,$29,$98,$F8,$04,$8F,$04,$72,$3C,$A4,$3C
	.dc.b	$25,$34,$60,$86,$08,$78,$0E,$05,$36,$53,$71,$E3,$7D,$0D,$7D,$11
	.dc.b	$E6,$78,$12,$3E,$D7,$95,$FE,$69,$8F,$F3,$7B,$5F,$F2,$F1,$8E,$7F
	.dc.b	$E3,$4F,$F0,$9F,$91,$FE,$AF,$E5,$FF,$14,$6C,$7F,$0B,$6B,$FC,$2D
	.dc.b	$AF,$F1,$76,$BC,$CF,$F3,$91,$92,$99,$8E,$CE,$2D,$3C,$9B,$19,$8C
	.dc.b	$11,$C1,$27,$99,$1F,$41,$92,$B9,$7D,$46,$4A,$E5,$F4,$35,$F4,$1C
	.dc.b	$88,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$93,$F1,$23,$C8
	.dc.b	$D8,$EE,$D7,$A3,$5E,$6D,$79,$CF,$CB,$9E,$6D,$8E,$6D,$79,$B5,$E6
	.dc.b	$D7,$99,$F2,$CC,$94,$CC,$76,$71,$3E,$32,$4F,$36,$C6,$A3,$04,$70
	.dc.b	$49,$E8,$47,$D2,$64,$AE,$5F,$59,$92,$B9,$7D,$2D,$7D,$27,$22,$1C
	.dc.b	$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18,$24,$FC,$48,$F2,$36,$3B
	.dc.b	$B5,$E8,$D7,$9B,$5E,$73,$F2,$E7,$9B,$63,$9B,$5E,$6D,$79,$BF,$81
	.dc.b	$C9,$D0,$F3,$5E,$BF,$F9,$D6,$73,$18,$BF,$F3,$29,$FE,$66,$8C,$CE
	.dc.b	$EB,$93,$FE,$7D,$E9,$7C,$5F,$F9,$B2,$30,$DB,$FF,$2F,$B5,$EB,$3C
	.dc.b	$1F,$01,$E0,$F8,$C7,$BF,$B8,$FF,$3B,$C6,$1D,$98,$8D,$28,$D5,$FF
	.dc.b	$A4,$7D,$C7,$85,$A6,$D7,$20,$E2,$78,$4F,$4E,$87,$45,$25,$F1,$8C
	.dc.b	$BC,$FF,$76,$F9,$37,$FE,$A1,$6C,$7C,$C0,$E1,$F9,$ED,$7F,$D2,$B9
	.dc.b	$BE,$A7,$DA,$D8,$F4,$36,$BE,$94,$FC,$2D,$FE,$DF,$BB,$F5,$F8,$4D
	.dc.b	$26,$1E,$3D,$7E,$1C,$7C,$08,$D9,$9D,$7F,$D4,$F1,$93,$A7,$D4,$C9
	.dc.b	$F6,$61,$F8,$A8,$C4,$7F,$AB,$A9,$E3,$3E,$EB,$FD,$63,$EF,$7C,$39
	.dc.b	$CC,$62,$BF,$23,$18,$53,$BA,$E5,$18,$3C,$3C,$BF,$9D,$FC,$8F,$E0
	.dc.b	$12,$33,$AA,$3E,$27,$44,$9C,$B5,$FD,$0E,$9F,$A4,$8C,$47,$E8,$EF
	.dc.b	$F5,$51,$8C,$BD,$BE,$0F,$F7,$06,$D7,$F6,$E6,$CF,$A6,$75,$93,$ED
	.dc.b	$BF,$DA,$7F,$6B,$ED,$6E,$1F,$0A,$74,$A7,$4D,$3D,$24,$3D,$38,$F9
	.dc.b	$9C,$1E,$9D,$3F,$BF,$E7,$ED,$8C,$87,$0F,$28,$DC,$F3,$8C,$65,$F1
	.dc.b	$1E,$11,$97,$B4,$66,$B9,$C7,$EC,$B0,$7D,$BF,$FB,$CB,$E0,$BE,$E3
	.dc.b	$67,$E0,$35,$FC,$54,$F0,$53,$F8,$3F,$B8,$F9,$64,$8F,$A1,$28,$FF
	.dc.b	$10,$DF,$8A,$7C,$F9,$7B,$09,$8F,$9F,$2F,$54,$EA,$3E,$F7,$D3,$3C
	.dc.b	$CF,$CB,$4F,$E7,$63,$0F,$E4,$CC,$46,$9F,$E2,$31,$AE,$2E,$D8,$78
	.dc.b	$27,$E8,$E3,$E5,$26,$3E,$27,$E4,$F0,$B0,$7D,$EA,$DF,$57,$FC,$71
	.dc.b	$EF,$19,$8C,$57,$18,$C3,$FA,$EA,$E7,$5C,$A3,$C7,$87,$84,$98,$FF
	.dc.b	$91,$BE,$FB,$F6,$3E,$17,$7B,$46,$2F,$F8,$F3,$63,$FA,$5B,$5F,$E6
	.dc.b	$7F,$07,$FC,$A3,$7F,$2B,$6B,$F8,$0B,$67,$F8,$53,$5F,$E7,$7F,$98
	.dc.b	$7F,$56,$E5,$F6,$B6,$BF,$84,$B6,$7F,$89,$35,$FC,$A9,$EE,$CB,$E9
	.dc.b	$3D,$BC,$2F,$F3,$7D,$78,$9B,$19,$18,$FA,$23,$31,$D0,$FD,$29,$D6
	.dc.b	$3E,$F9,$1F,$F3,$F7,$95,$F6,$7E,$13,$9C,$7D,$BF,$29,$E5,$2B,$57
	.dc.b	$FD,$D2,$73,$18,$AE,$15,$B5,$7D,$15,$70,$73,$7C,$EB,$85,$6E,$B4
	.dc.b	$D8,$CC,$FF,$03,$AF,$E9,$84,$3F,$96,$94,$D1,$82,$1F,$24,$F9,$07
	.dc.b	$F4,$53,$18,$FC,$F7,$07,$F1,$73,$A1,$4E,$67,$E0,$8F,$D1,$7D,$AF
	.dc.b	$E0,$9B,$1F,$82,$69,$F8,$22,$9F,$92,$CB,$FA,$7B,$E2,$21,$93,$C6
	.dc.b	$7E,$7B,$BF,$F6,$1A,$7E,$87,$83,$FA,$9B,$63,$37,$FA,$26,$D7,$C4
	.dc.b	$21,$EB,$2C,$7E,$D5,$7D,$6F,$EE,$39,$3F,$6A,$DE,$5F,$E0,$0D,$9D
	.dc.b	$D7,$C4,$7F,$8E,$9F,$20,$A6,$79,$FF,$C7,$EE,$5A,$FF,$03,$E9,$FE
	.dc.b	$19,$EA,$FF,$75,$C7,$F8,$EB,$63,$33,$B1,$1C,$FA,$3F,$C9,$D1,$74
	.dc.b	$FF,$10,$3E,$F2,$1E,$43,$E0,$F6,$7F,$37,$77,$97,$F8,$23,$B3,$8C
	.dc.b	$7F,$49,$6C,$66,$37,$3F,$D1,$7C,$9F,$E9,$70,$69,$1F,$85,$87,$99
	.dc.b	$B8,$76,$D3,$E1,$C7,$D6,$3C,$C7,$A1,$CB,$67,$93,$B8,$CB,$C0,$8F
	.dc.b	$0C,$3F,$C4,$7E,$07,$E6,$27,$FB,$F3,$5F,$EB,$ED,$7F,$B4,$36,$3F
	.dc.b	$C1,$B1,$FE,$0B,$3E,$0F,$83,$F1,$32,$3F,$22,$3C,$38,$79,$9F,$46
	.dc.b	$3E,$F8,$C4,$6A,$30,$4F,$C9,$BF,$C9,$44,$8F,$F1,$5B,$FE,$33,$E2
	.dc.b	$FC,$1C,$7F,$76,$6B,$FD,$FD,$B8,$FF,$8E,$F6,$FF,$2A,$7D,$54,$AF
	.dc.b	$F1,$AB,$5F,$F4,$8F,$8F,$FE,$76,$F5,$FF,$CA,$84,$8F,$F2,$A0,$7F
	.dc.b	$CB,$C4,$7F,$C7,$ED,$7F,$D1,$FE,$4F,$F6,$7E,$EF,$F3,$3C,$7F,$27
	.dc.b	$32,$53,$FC,$E3,$1B,$7F,$CF,$1E,$67,$FA,$92,$7F,$CF,$91,$93,$FC
	.dc.b	$F2,$53,$FB,$A4,$6D,$FF,$40,$71,$FF,$A1,$A7,$F4,$6D,$8C,$F5,$71
	.dc.b	$7F,$30,$DC,$7F,$D2,$F5,$FE,$75,$24,$FC,$27,$F5,$BE,$6F,$FA,$95
	.dc.b	$BE,$77,$38,$C4,$FF,$A4,$5A,$FF,$AC,$BC,$5F,$07,$9D,$FE,$B3,$BF
	.dc.b	$F5,$CC,$FC,$31,$19,$7F,$CD,$71,$FE,$8D,$23,$83,$FD,$3D,$E8,$3D
	.dc.b	$0F,$C9,$36,$3F,$76,$D7,$F8,$2B,$5E,$31,$C0,$8F,$06,$BC,$1A,$F0
	.dc.b	$34,$43,$45,$34,$60,$86,$08,$7F,$B8,$A3,$C8,$D8,$EE,$D3,$A1,$4E
	.dc.b	$65,$39,$9F,$E3,$78,$E6,$D8,$E6,$D3,$99,$4E,$65,$39,$88,$E6,$D8
	.dc.b	$E6,$D3,$99,$4E,$65,$39,$C7,$22,$47,$26,$C7,$26,$BC,$9A,$F2,$8E
	.dc.b	$04,$78,$35,$E0,$D7,$81,$A2,$1A,$29,$A3,$04,$30,$43,$E8,$10,$F9
	.dc.b	$05,$3E,$01,$63,$D2,$D8,$F2,$34,$F0,$94,$F0,$14,$F0,$1D,$88,$76
	.dc.b	$2C,$76,$6C,$76,$69,$D8,$A7,$62,$9D,$8E,$84,$8E,$8D,$8E,$8D,$3A
	.dc.b	$14,$E8,$53,$A4,$72,$24,$72,$6C,$72,$6B,$C9,$AF,$28,$E0,$47,$83
	.dc.b	$5F,$C8,$F8,$BF,$E5,$37,$FC,$C6,$7F,$81,$9C,$3F,$E5,$5A,$FE,$0F
	.dc.b	$E1,$FF,$A7,$CF,$E0,$E5,$8F,$E1,$ED,$7F,$70,$D7,$F6,$0D,$7F,$50
	.dc.b	$DB,$FF,$38,$BD,$23,$FC,$D0,$70,$29,$C0,$B1,$C1,$F0,$98,$8E,$84
	.dc.b	$70,$47,$B1,$1E,$2D,$7D,$26,$CC,$90,$CB,$B7,$CC,$D7,$CC,$D8,$F3
	.dc.b	$47,$03,$91,$C0,$A7,$02,$BC,$1D,$1C,$48,$60,$87,$13,$25,$32,$F3
	.dc.b	$70,$47,$04,$8F,$90,$7A,$4B,$1E,$46,$BD,$A3,$83,$F1,$CC,$46,$9F
	.dc.b	$0B,$5F,$03,$63,$8B,$DE,$39,$1C,$0A,$70,$2C,$70,$7C,$86,$23,$A1
	.dc.b	$1C,$11,$EC,$47,$8B,$5F,$49,$B3,$24,$32,$ED,$F3,$35,$F3,$36,$3C
	.dc.b	$D1,$C0,$E4,$70,$29,$C0,$AF,$07,$47,$12,$18,$21,$C4,$C9,$4C,$BC
	.dc.b	$DC,$11,$C1,$23,$E4,$1E,$92,$C7,$91,$AF,$68,$E0,$FC,$73,$F2,$3D
	.dc.b	$8E,$97,$4B,$D0,$FF,$49,$CE,$63,$14,$20,$8D,$4F,$0A,$E6,$FF,$A7
	.dc.b	$3D,$37,$A7,$C6,$30,$DB,$E6,$27,$6F,$80,$EE,$E2,$7F,$C7,$31,$93
	.dc.b	$C9,$C0,$E1,$E9,$1E,$68,$C3,$B0,$E2,$7F,$D6,$3C,$0F,$07,$8D,$7B
	.dc.b	$E3,$0F,$16,$CF,$09,$DC,$FF,$95,$0E,$DC,$AE,$5C,$B6,$33,$1C,$3D
	.dc.b	$5F,$F4,$9C,$79,$5B,$7F,$EC,$46,$C7,$C6,$6B,$FE,$74,$8F,$F3,$B9
	.dc.b	$1F,$97,$83,$A3,$CB,$BF,$DB,$F4,$FE,$B7,$09,$A4,$BF,$A7,$BF,$D2
	.dc.b	$8F,$AF,$1B,$33,$C1,$F8,$38,$C9,$FE,$78,$D3,$13,$FE,$78,$38,$1F
	.dc.b	$88,$D7,$E1,$4E,$E3,$27,$E3,$6F,$F0,$3E,$9F,$80,$91,$E0,$0F,$FB
	.dc.b	$A3,$E3,$5F,$73,$F9,$5F,$4B,$0D,$AF,$C6,$CE,$94,$E9,$A7,$5F,$57
	.dc.b	$DE,$47,$07,$E5,$B6,$FD,$27,$47,$CA,$24,$67,$9F,$F5,$71,$F9,$CF
	.dc.b	$09,$E8,$3E,$FB,$FB,$84,$61,$B1,$9B,$FE,$DB,$39,$6C,$7E,$4A,$BA
	.dc.b	$36,$3F,$67,$3D,$BA,$39,$9D,$3C,$D9,$3C,$9B,$7E,$88,$FD,$DE,$5E
	.dc.b	$C7,$D5,$CB,$FD,$F2,$31,$1F,$C0,$70,$F5,$FE,$01,$E8,$24,$79,$9F
	.dc.b	$F0,$54,$F0,$8F,$E1,$0D,$7F,$84,$69,$F1,$9E,$F8,$7F,$02,$F1,$5E
	.dc.b	$8B,$E6,$7F,$C2,$D1,$86,$C6,$6B,$FC,$27,$39,$7E,$3F,$57,$F1,$7F
	.dc.b	$77,$E3,$B6,$3E,$3F,$07,$E9,$54,$E9,$A6,$BE,$E3,$EA,$91,$FA,$6F
	.dc.b	$BB,$3F,$54,$7C,$A6,$BF,$01,$AF,$C0,$DF,$ED,$FD,$CF,$E0,$09,$18
	.dc.b	$3F,$C5,$BE,$5B,$89,$E6,$7F,$31,$EA,$FE,$4F,$39,$8C,$57,$F3,$48
	.dc.b	$C3,$FA,$63,$15,$C9,$C1,$E7,$C3,$EF,$63,$E6,$E2,$FA,$C4,$FA,$8F
	.dc.b	$F2,$67,$2F,$E8,$5B,$7F,$88,$6B,$F7,$9B,$0F,$4F,$F9,$DF,$E1,$88
	.dc.b	$FC,$BC,$61,$BE,$17,$F4,$D9,$D5,$FE,$3E,$B4,$F9,$F9,$FF,$30,$FC
	.dc.b	$DC,$FD,$27,$EA,$25,$F7,$7F,$D5,$FE,$0B,$9D,$78,$67,$2A,$33,$97
	.dc.b	$9E,$B8,$3E,$66,$CF,$E8,$79,$7F,$AA,$E3,$C1,$B5,$F8,$CC,$FE,$0C
	.dc.b	$62,$B8,$5F,$F6,$9F,$6F,$83,$6F,$F2,$6D,$9D,$89,$F4,$BF,$89,$C3
	.dc.b	$E0,$B4,$FD,$49,$E1,$0F,$97,$0F,$19,$23,$37,$C1,$F9,$85,$3E,$64
	.dc.b	$F6,$DB,$F0,$7B,$FE,$14,$FC,$DF,$9B,$FB,$6F,$37,$FC,$09,$ED,$F8
	.dc.b	$30,$FF,$03,$4E,$54,$65,$46,$63,$0D,$C3,$FD,$39,$3C,$1B,$3A,$EC
	.dc.b	$5F,$82,$72,$7E,$6E,$BF,$AE,$90,$FE,$36,$53,$46,$08,$7D,$DC,$FF
	.dc.b	$95,$B4,$FE,$40,$53,$E6,$18,$3F,$24,$75,$29,$D0,$FD,$06,$DF,$90
	.dc.b	$F6,$FF,$03,$06,$93,$5F,$B5,$1D,$88,$E0,$FC,$A3,$FA,$B6,$9F,$9A
	.dc.b	$E1,$FC,$EB,$78,$3F,$9C,$3E,$39,$FE,$24,$79,$CA,$7E,$C8,$7E,$19
	.dc.b	$CF,$1F,$F2,$03,$63,$33,$B8,$FC,$AF,$03,$97,$E4,$D7,$F0,$33,$E3
	.dc.b	$1F,$E4,$77,$2D,$33,$1F,$E4,$F7,$C8,$D7,$E5,$73,$FF,$1C,$EE,$FE
	.dc.b	$43,$D1,$FC,$6D,$BD,$3F,$E5,$56,$D1,$FE,$CF,$D7,$FE,$4E,$8F,$DD
	.dc.b	$35,$FA,$AD,$CF,$FB,$BB,$98,$F8,$A7,$E2,$08,$7D,$E3,$F8,$27,$67
	.dc.b	$1E,$13,$49,$1E,$1A,$FF,$01,$F7,$7F,$9A,$FE,$83,$FA,$81,$27,$FA
	.dc.b	$AC,$65,$FF,$37,$E3,$FD,$75,$FE,$CD,$CD,$FD,$85,$BE,$07,$AE,$75
	.dc.b	$1F,$7F,$7F,$94,$FA,$0F,$59,$F9,$4F,$07,$D0,$47,$CD,$A2,$3E,$14
	.dc.b	$F2,$9F,$80,$43,$9E,$7F,$E1,$ED,$FE,$11,$23,$9B,$F1,$7C,$29,$1D
	.dc.b	$9A,$FF,$87,$3C,$3F,$94,$47,$C6,$F8,$BB,$7F,$D2,$F1,$F8,$66,$BF
	.dc.b	$9D,$F2,$3F,$D1,$3E,$BF,$F9,$60,$91,$FE,$49,$7F,$C8,$1C,$B2,$3F
	.dc.b	$C8,$BE,$4F,$FA,$76,$31,$1F,$E9,$0E,$9F,$C6,$3F,$E5,$1E,$0F,$F3
	.dc.b	$19,$1F,$F2,$F7,$AF,$FE,$5B,$23,$83,$FC,$B0,$FF,$9A,$88,$E0,$FF
	.dc.b	$17,$BF,$E7,$42,$38,$3F,$CF,$7C,$9F,$EB,$1E,$FF,$F4,$BB,$63,$27
	.dc.b	$F2,$42,$C7,$FA,$27,$D8,$D4,$7F,$AE,$3D,$1F,$F4,$7C,$E6,$38,$19
	.dc.b	$9D,$BF,$EB,$FF,$3B,$FD,$31,$3F,$E7,$72,$47,$F9,$F5,$FF,$51,$3C
	.dc.b	$8F,$F5,$47,$0F,$25,$E8,$7F,$AA,$A7,$67,$FA,$E8,$91,$FE,$C2,$F4
	.dc.b	$BF,$D6,$F1,$88,$FA,$47,$53,$FD,$8D,$AF,$5E,$1F,$F6,$3C,$FF,$A5
	.dc.b	$8F,$F4,$E9,$27,$FC,$B6,$D7,$FD,$B3,$EA,$7F,$53,$7F,$DA,$2D,$73
	.dc.b	$E2,$FF,$01,$8C,$47,$FB,$85,$A9,$A4,$8E,$04,$8F,$E1,$D1,$96,$C6
	.dc.b	$5F,$E2,$91,$96,$C6,$5F,$2B,$5F,$29,$C8,$87,$02,$9C,$0A,$70,$34
	.dc.b	$43,$45,$34,$60,$86,$08,$7C,$37,$C8,$D3,$B9,$4E,$85,$39,$94,$E6
	.dc.b	$7C,$B7,$9B,$4E,$65,$39,$94,$E6,$53,$9C,$7C,$B7,$2D,$73,$1D,$A3
	.dc.b	$11,$E3,$21,$CC,$AE,$9C,$11,$C1,$23,$A1,$23,$D3,$19,$6C,$65,$F5
	.dc.b	$C6,$5B,$19,$7D,$2D,$7D,$27,$22,$1C,$0A,$70,$29,$C0,$D1,$0D,$14
	.dc.b	$D1,$82,$18,$21,$F1,$1F,$23,$4E,$E5,$3A,$14,$E6,$53,$99,$F2,$DE
	.dc.b	$6D,$39,$94,$E6,$53,$99,$4E,$71,$F2,$DC,$B5,$CC,$76,$8F,$DD,$F8
	.dc.b	$FF,$E5,$79,$1F,$E5,$22,$9F,$E5,$CE,$DF,$E2,$4F,$F9,$69,$B2,$FA
	.dc.b	$CE,$37,$FB,$C6,$D7,$F0,$E6,$BF,$C4,$9A,$FF,$0D,$6B,$FC,$35,$B3
	.dc.b	$D6,$3F,$8B,$36,$7B,$46,$27,$64,$70,$48,$D9,$1F,$21,$92,$B9,$7C
	.dc.b	$C6,$4A,$E5,$F4,$19,$2B,$97,$D4,$64,$AE,$67,$D5,$1E,$02,$1D,$8A
	.dc.b	$74,$29,$C8,$AF,$27,$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9,$C1
	.dc.b	$1C,$12,$7E,$84,$FB,$5B,$3E,$66,$BE,$16,$BD,$9A,$F5,$6C,$F5,$8F
	.dc.b	$94,$D9,$ED,$18,$9D,$91,$C1,$23,$64,$7C,$86,$4A,$E5,$F3,$19,$2B
	.dc.b	$97,$D0,$64,$AE,$5F,$51,$92,$B9,$9F,$54,$78,$08,$76,$29,$D0,$A7
	.dc.b	$22,$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70,$49
	.dc.b	$FA,$13,$ED,$6C,$FF,$0A,$1D,$6E,$87,$96,$F5,$3F,$D2,$D3,$98,$C5
	.dc.b	$71,$8C,$4E,$AB,$8B,$FE,$9D,$D5,$E3,$FF,$1A,$8C,$36,$FC,$53,$A7
	.dc.b	$A9,$FC,$79,$C4,$FF,$A9,$F8,$5D,$9C,$6B,$F9,$1C,$61,$E2,$7F,$28
	.dc.b	$71,$3C,$67,$84,$FF,$2C,$F8,$47,$A7,$88,$46,$1E,$66,$23,$51,$9A
	.dc.b	$F2,$77,$F1,$C0,$D2,$5F,$28,$FF,$4D,$37,$C1,$E0,$47,$BF,$B2,$FA
	.dc.b	$9F,$23,$63,$31,$83,$D1,$BF,$D5,$F5,$BE,$97,$09,$A4,$AF,$9F,$C1
	.dc.b	$FD,$32,$3E,$9C,$6C,$C9,$4F,$A5,$1B,$7F,$11,$97,$CF,$9F,$99,$1B
	.dc.b	$3E,$86,$FB,$7E,$AB,$5F,$EB,$7F,$0C,$E6,$FE,$BB,$EA,$7F,$B8,$E3
	.dc.b	$12,$8C,$4E,$A3,$F3,$13,$A9,$F0,$92,$3C,$B5,$B7,$C2,$73,$38,$D7
	.dc.b	$F6,$CF,$5B,$AC,$1A,$47,$99,$92,$BF,$00,$8F,$E8,$F9,$7F,$53,$CF
	.dc.b	$F4,$89,$19,$D7,$FC,$FD,$E1,$BF,$11,$7C,$47,$F7,$08,$C3,$63,$35
	.dc.b	$C1,$B7,$FB,$9F,$25,$E5,$3C,$9F,$F0,$3F,$AB,$98,$C6,$08,$C4,$EA
	.dc.b	$3F,$C1,$75,$CA,$31,$1F,$E1,$08,$EB,$EE,$FF,$81,$1B,$1F,$E9,$76
	.dc.b	$E1,$DE,$7F,$C0,$FC,$DF,$C1,$E3,$E0,$D7,$F8,$62,$30,$FC,$38,$F3
	.dc.b	$7D,$FF,$5C,$8D,$24,$FF,$A8,$EB,$8B,$CF,$7F,$9E,$F3,$21,$FE,$28
	.dc.b	$D9,$FB,$9B,$3F,$9C,$D9,$FA,$6D,$CB,$F9,$25,$70,$AF,$27,$BB,$FE
	.dc.b	$32,$FC,$AF,$E6,$37,$0F,$93,$A2,$31,$1F,$26,$30,$A7,$EA,$7D,$07
	.dc.b	$BC,$91,$F2,$E5,$1F,$2B,$3F,$6C,$FB,$F8,$FE,$14,$62,$3F,$C8,$BA
	.dc.b	$FC,$47,$DD,$6F,$B8,$F9,$84,$8E,$CF,$5B,$E6,$4D,$7E,$BC,$F4,$C3
	.dc.b	$B3,$FD,$19,$B8,$77,$8C,$57,$FB,$5F,$F0,$07,$C7,$FF,$92,$1B,$39
	.dc.b	$8C,$1E,$1C,$3B,$C7,$26,$DF,$52,$61,$E0,$3E,$06,$DF,$A0,$FB,$AF
	.dc.b	$F3,$2F,$17,$DB,$C3,$E0,$C6,$27,$EA,$B6,$3E,$A3,$5F,$8F,$E2,$FF
	.dc.b	$99,$A7,$ED,$B6,$3E,$D3,$5D,$8C,$7C,$12,$F8,$CF,$E7,$BC,$F8,$6C
	.dc.b	$7E,$B9,$B5,$FA,$D6,$BF,$3F,$6F,$A4,$FE,$68,$C5,$FE,$EF,$0F,$15
	.dc.b	$FF,$00,$6F,$07,$F5,$8A,$F2,$4F,$A9,$A7,$E2,$6B,$E6,$E9,$F2,$F0
	.dc.b	$EF,$39,$3C,$07,$D3,$3F,$C0,$DE,$5F,$F3,$7F,$33,$FD,$27,$F0,$1F
	.dc.b	$06,$7F,$D2,$51,$8C,$3F,$D2,$B5,$B9,$56,$DF,$F4,$B5,$6D,$E2,$6A
	.dc.b	$B8,$F7,$27,$53,$C0,$D1,$C2,$BA,$70,$FF,$85,$63,$FB,$09,$0F,$E6
	.dc.b	$05,$3F,$54,$53,$E8,$1F,$34,$FE,$40,$7E,$00,$A7,$E0,$0F,$BF,$A7
	.dc.b	$CF,$7F,$20,$7E,$88,$E8,$53,$C2,$7E,$07,$C7,$FF,$5D,$FC,$07,$E1
	.dc.b	$5B,$B6,$65,$39,$6C,$7D,$B7,$C3,$BF,$F3,$87,$CC,$60,$F3,$C7,$E9
	.dc.b	$57,$C2,$FA,$38,$FF,$D9,$6D,$8C,$CF,$E9,$9A,$7C,$63,$D2,$52,$1F
	.dc.b	$20,$C9,$4C,$8E,$6F,$E0,$2F,$F9,$3C,$C7,$35,$7F,$DB,$B1,$A7,$FC
	.dc.b	$A8,$D7,$9E,$9F,$75,$EE,$7C,$D3,$3C,$7F,$E5,$87,$E8,$C6,$DF,$E6
	.dc.b	$5C,$FD,$F0,$FE,$CE,$DD,$BF,$93,$4F,$96,$72,$F9,$6B,$93,$ED,$21
	.dc.b	$E1,$3F,$97,$F4,$73,$23,$CE,$BF,$9D,$B7,$AB,$FD,$F7,$F9,$67,$F4
	.dc.b	$A8,$34,$8F,$FB,$FB,$B3,$FA,$93,$F2,$5A,$FF,$56,$ED,$FC,$8B,$FD
	.dc.b	$7D,$AF,$9B,$1F,$4B,$E7,$3D,$19,$7F,$82,$3B,$3E,$11,$23,$D0,$FD
	.dc.b	$57,$84,$F2,$26,$BF,$3E,$3F,$B0,$F7,$7C,$B2,$3F,$05,$F9,$DD,$FF
	.dc.b	$A0,$8F,$EC,$ED,$7F,$B3,$35,$FE,$B2,$DC,$BF,$C4,$71,$F0,$7C,$0F
	.dc.b	$F1,$1B,$FD,$D9,$AF,$C7,$6B,$F8,$4E,$DF,$B5,$87,$9E,$3F,$2B,$E3
	.dc.b	$FF,$A1,$FD,$AF,$F2,$C9,$23,$FC,$7E,$FF,$91,$08,$FF,$92,$03,$87
	.dc.b	$FC,$9B,$87,$F9,$88,$8F,$F9,$21,$FF,$26,$91,$E4,$D7,$FC,$AB,$EC
	.dc.b	$FF,$9A,$88,$FF,$94,$1F,$F3,$A9,$1F,$F3,$AB,$5F,$F0,$23,$FE,$80
	.dc.b	$23,$FE,$4F,$7F,$61,$E6,$7F,$A2,$7B,$AF,$98,$AF,$E5,$C7,$FB,$0E
	.dc.b	$74,$D8,$D4,$FF,$AC,$7B,$AC,$FF,$A0,$9A,$FF,$AC,$9A,$FF,$AC,$5A
	.dc.b	$B2,$FF,$52,$BE,$03,$FD,$73,$E6,$EA,$33,$5F,$EA,$C9,$D1,$FE,$B7
	.dc.b	$F3,$F9,$F7,$FF,$AD,$67,$F8,$63,$67,$FD,$B8,$D7,$FD,$03,$1F,$E9
	.dc.b	$62,$3F,$E7,$C7,$FD,$42,$47,$FD,$42,$D7,$FC,$60,$FF,$AB,$08,$FF
	.dc.b	$9E,$DF,$E3,$9E,$93,$D2,$FD,$43,$67,$F8,$F3,$5D,$C6,$88,$E9,$AE
	.dc.b	$8C,$10,$C1,$0F,$F7,$D4,$75,$6C,$72,$69,$C4,$A7,$13,$F0,$42,$39
	.dc.b	$38,$6B,$F8,$76,$BB,$8E,$CE,$DA,$ED,$AE,$CF,$11,$B2,$9B,$29,$B3
	.dc.b	$CC,$64,$A6,$4F,$59,$0E,$86,$A3,$CE,$48,$F1,$B6,$3B,$B5,$E1,$1A
	.dc.b	$23,$A6,$BA,$30,$43,$04,$3C,$87,$A4,$A7,$A4,$F3,$98,$8F,$13,$63
	.dc.b	$BB,$4E,$C5,$3B,$1C,$88,$77,$29,$DC,$C1,$0C,$12,$3C,$2F,$06,$BB
	.dc.b	$6B,$B8,$F2,$BB,$6B,$B6,$BB,$3D,$26,$CA,$6C,$A6,$CF,$69,$92,$99
	.dc.b	$3E,$11,$0E,$86,$A3,$CE,$48,$F1,$B6,$3B,$B5,$E1,$1A,$23,$A6,$BA
	.dc.b	$30,$7F,$86,$FC,$6F,$F3,$03,$FE,$68,$F1,$7F,$8C,$BF,$E5,$E3,$17
	.dc.b	$FC,$13,$ED,$FF,$99,$78,$FF,$C2,$9B,$3F,$BC,$6B,$FB,$36,$BF,$B3
	.dc.b	$6B,$FB,$96,$E2,$69,$23,$FC,$E4,$71,$29,$C4,$B1,$C5,$B1,$C6,$7C
	.dc.b	$D1,$82,$38,$23,$E9,$32,$57,$2F,$AC,$C9,$5C,$BE,$56,$BE,$56,$CF
	.dc.b	$95,$B3,$E5,$9E,$C4,$3A,$14,$E8,$53,$A1,$67,$A3,$63,$A3,$C8,$87
	.dc.b	$22,$9C,$8B,$1C,$9B,$1C,$9E,$04,$38,$16,$38,$36,$38,$3A,$24,$69
	.dc.b	$B1,$A8,$C1,$1C,$12,$3E,$79,$D4,$A7,$22,$C7,$16,$C7,$19,$F9,$F1
	.dc.b	$82,$38,$23,$F2,$4C,$95,$CB,$F1,$0C,$95,$CB,$E5,$6B,$E5,$6C,$F9
	.dc.b	$5B,$3E,$59,$EC,$43,$A1,$4E,$85,$3A,$16,$7A,$36,$3A,$3C,$88,$72
	.dc.b	$29,$C8,$B1,$C9,$B1,$C9,$E0,$43,$81,$63,$83,$FE,$35,$E9,$72,$B9
	.dc.b	$9E,$97,$FA,$AE,$73,$18,$AF,$EF,$62,$33,$3B,$AE,$4F,$FA,$CF,$D5
	.dc.b	$FF,$54,$36,$FF,$C8,$8D,$BE,$87,$F8,$1A,$76,$F6,$7F,$D9,$AE,$8F
	.dc.b	$F2,$63,$88,$ED,$E3,$74,$9E,$0D,$7F,$C2,$3C,$4F,$5B,$11,$FE,$55
	.dc.b	$8C,$3D,$4F,$F2,$54,$69,$A6,$79,$DD,$F9,$6C,$66,$30,$71,$F5,$DE
	.dc.b	$1D,$24,$F7,$8C,$BE,$1C,$1E,$D7,$85,$B1,$98,$C1,$D7,$A2,$47,$36
	.dc.b	$E9,$F3,$7D,$59,$3F,$2F,$7F,$F3,$A4,$62,$3E,$86,$D8,$7E,$48,$8F
	.dc.b	$9A,$D3,$2F,$FB,$BF,$1F,$92,$DD,$B9,$4F,$D0,$6B,$F3,$DB,$AB,$C6
	.dc.b	$67,$FE,$87,$F5,$3F,$1B,$39,$8C,$5F,$F9,$2A,$30,$A7,$75,$CA,$30
	.dc.b	$7E,$5B,$1F,$BB,$3E,$58,$C9,$E5,$29,$E5,$2D,$76,$F4,$FA,$E1,$FE
	.dc.b	$9B,$6C,$9A,$7E,$91,$7A,$AF,$97,$FF,$51,$7C,$36,$63,$12,$8C,$4E
	.dc.b	$A3,$F5,$B3,$A9,$FD,$66,$FF,$DA,$B9,$FF,$67,$D5,$82,$5F,$C5,$D9
	.dc.b	$F7,$BF,$EB,$1F,$92,$F8,$AD,$9D,$35,$FF,$5B,$6D,$F4,$E3,$11,$F9
	.dc.b	$EA,$7D,$FF,$F8,$12,$BE,$C3,$63,$33,$FE,$00,$9C,$B6,$1F,$79,$FA
	.dc.b	$F8,$F3,$35,$F0,$C6,$9F,$F5,$D1,$A2,$9F,$E8,$AC,$BE,$0B,$FC,$49
	.dc.b	$AE,$7B,$3F,$8D,$67,$F1,$23,$DB,$F7,$5F,$B6,$E4,$FC,$6B,$63,$F1
	.dc.b	$8D,$77,$97,$95,$FE,$4D,$7D,$1F,$03,$5F,$06,$DF,$56,$31,$1F,$C1
	.dc.b	$A9,$E3,$FC,$CB,$46,$25,$BF,$92,$B5,$39,$8F,$3F,$AF,$FE,$EB,$FC
	.dc.b	$9A,$D9,$A4,$7D,$7F,$71,$C8,$8F,$63,$25,$3D,$3A,$FC,$77,$DF,$A7
	.dc.b	$AC,$FB,$98,$FF,$46,$8E,$07,$B7,$4F,$39,$1C,$E9,$EE,$FA,$B4,$7F
	.dc.b	$8F,$CF,$A6,$E3,$C4,$FA,$6D,$8F,$A6,$D7,$CA,$7D,$0C,$3C,$98,$F9
	.dc.b	$EF,$E5,$FC,$D7,$C8,$25,$78,$B2,$47,$CC,$FB,$AF,$1F,$E5,$BF,$CE
	.dc.b	$2B,$EB,$46,$2F,$F0,$6D,$8F,$C4,$B5,$F8,$B8,$7C,$77,$FB,$16,$9F
	.dc.b	$80,$6E,$5F,$A3,$FB,$9F,$BF,$0F,$BE,$F0,$62,$9C,$C6,$1B,$1F,$C7
	.dc.b	$DB,$3F,$9A,$F9,$0F,$EB,$13,$F9,$86,$C6,$47,$2F,$6D,$93,$A9,$FC
	.dc.b	$A7,$CB,$7F,$05,$BF,$D0,$B6,$3F,$4E,$D7,$67,$E5,$A3,$E9,$69,$F0
	.dc.b	$79,$FF,$63,$3F,$10,$F2,$1B,$3C,$9C,$DE,$3C,$3A,$5F,$E8,$9E,$D3
	.dc.b	$F2,$2F,$A6,$DF,$2F,$0E,$B3,$93,$C2,$43,$F0,$A7,$E3,$87,$9C,$F3
	.dc.b	$9E,$77,$E4,$BA,$58,$EE,$B5,$39,$8C,$57,$0B,$5F,$07,$F3,$77,$C1
	.dc.b	$CD,$70,$BE,$51,$AB,$E5,$7C,$63,$CB,$18,$AE,$11,$C8,$FD,$89,$FD
	.dc.b	$00,$87,$E8,$4A,$7F,$10,$29,$C0,$FA,$3E,$57,$CF,$4D,$26,$9C,$E2
	.dc.b	$1C,$CA,$7E,$02,$3F,$15,$B7,$EE,$9B,$1D,$27,$F1,$2D,$3F,$25,$F4
	.dc.b	$7E,$12,$47,$E1,$1F,$CF,$70,$7F,$04,$FB,$DB,$46,$92,$BC,$71,$FA
	.dc.b	$C3,$DE,$53,$F6,$E3,$F2,$EF,$D0,$73,$FF,$10,$6C,$66,$77,$1F,$A5
	.dc.b	$E9,$F6,$4F,$F4,$26,$BF,$D2,$1A,$FF,$3B,$6B,$E1,$3E,$41,$0F,$A4
	.dc.b	$7F,$1E,$1E,$AF,$F1,$E6,$9F,$C7,$8A,$6D,$FE,$AB,$C7,$E0,$BF,$E8
	.dc.b	$ED,$CB,$F9,$96,$9A,$BF,$BA,$7C,$6E,$3F,$F3,$BB,$F1,$9A,$FF,$40
	.dc.b	$8F,$EC,$3D,$BC,$C9,$1C,$8F,$EC,$FC,$5F,$CE,$3F,$2F,$FE,$9E,$48
	.dc.b	$FE,$CF,$19,$7F,$B4,$F8,$1E,$52,$3C,$CF,$EC,$4F,$C8,$C3,$C7,$1F
	.dc.b	$1C,$F9,$47,$C8,$3F,$B4,$67,$FD,$23,$87,$FB,$BB,$63,$FB,$BB,$5F
	.dc.b	$EE,$E2,$3E,$06,$9F,$E9,$B8,$BD,$9F,$E3,$5A,$FF,$28,$78,$5C,$67
	.dc.b	$FC,$3C,$D8,$CC,$73,$3F,$C2,$57,$FE,$4E,$F0,$FF,$C7,$D1,$FE,$25
	.dc.b	$6C,$FF,$88,$23,$47,$7F,$0F,$FC,$66,$FC,$38,$FF,$17,$36,$33,$BF
	.dc.b	$F8,$DA,$3E,$41,$F1,$09,$9F,$C8,$7E,$31,$FE,$4E,$F1,$FF,$B6,$FC
	.dc.b	$B7,$F9,$2A,$0D,$23,$FE,$45,$F2,$BF,$B9,$FB,$5F,$E5,$A2,$46,$43
	.dc.b	$FE,$67,$F6,$A4,$7F,$9A,$FC,$EF,$EF,$71,$88,$FF,$1C,$D6,$1E,$51
	.dc.b	$FE,$66,$23,$C4,$7B,$5F,$E7,$32,$39,$7F,$D1,$E4,$7F,$D2,$6D,$7F
	.dc.b	$D8,$DC,$3F,$E0,$C8,$FF,$65,$12,$3F,$D3,$C7,$63,$FD,$AF,$A7,$FA
	.dc.b	$17,$C0,$FF,$3B,$36,$33,$87,$F8,$5A,$72,$D8,$FF,$45,$7A,$75,$FF
	.dc.b	$65,$B5,$FF,$65,$35,$FE,$E2,$DA,$FF,$5E,$D7,$83,$D4,$C1,$2B,$FA
	.dc.b	$71,$FD,$67,$D5,$FF,$65,$CF,$FA,$2D,$B1,$FE,$F0,$6B,$BE,$2F,$07
	.dc.b	$8F,$FE,$DA,$AF,$F4,$F1,$27,$E8,$3F,$EB,$A2,$3F,$EB,$E6,$BF,$30
	.dc.b	$F1,$9F,$EE,$ED,$3D,$9E,$B7,$EA,$E7,$FD,$F0,$7F,$A5,$09,$3F,$E3
	.dc.b	$56,$BF,$DC,$FD,$7F,$F2,$03,$FD,$D4,$38,$75,$E7,$7E,$B1,$CB,$63
	.dc.b	$51,$FE,$4A,$6B,$A3,$F6,$6E,$09,$1C,$49,$1A,$6B,$FE,$B7,$8D,$B9
	.dc.b	$21,$98,$DB,$E9,$8D,$B9,$36,$66,$36,$FB,$0C,$94,$E0,$43,$81,$A3
	.dc.b	$DC,$F2,$69,$93,$E0,$1C,$4E,$87,$23,$07,$02,$9C,$0D,$10,$E2,$43
	.dc.b	$81,$92,$99,$39,$16,$3E,$1B,$B3,$27,$02,$18,$23,$C1,$E5,$1E,$F7
	.dc.b	$2D,$8D,$47,$26,$BA,$3E,$23,$82,$47,$12,$46,$9A,$FC,$48,$DB,$92
	.dc.b	$19,$8D,$BE,$D8,$DB,$93,$66,$63,$6F,$B0,$C9,$4E,$04,$38,$1A,$3D
	.dc.b	$CF,$26,$99,$3E,$01,$C4,$E8,$72,$30,$70,$29,$C0,$D1,$0E,$24,$38
	.dc.b	$19,$29,$93,$91,$63,$E1,$BB,$32,$70,$3F,$03,$E4,$7F,$99,$DF,$E0
	.dc.b	$4D,$7F,$CC,$26,$3C,$8F,$E2,$2F,$EF,$FE,$A7,$FD,$93,$E4,$FE,$DD
	.dc.b	$B1,$FB,$36,$BF,$C3,$DA,$F1,$3F,$CE,$05,$8E,$4E,$0F,$E2,$71,$C1
	.dc.b	$A6,$6D,$19,$6C,$66,$30,$47,$04,$9E,$E4,$7D,$06,$4A,$E5,$F5,$19
	.dc.b	$2B,$97,$D4,$70,$21,$A2,$9A,$30,$43,$04,$3D,$EF,$56,$9C,$8A,$71
	.dc.b	$29,$C6,$7E,$1C,$71,$6C,$71,$6B,$C5,$AF,$19,$F9,$13,$C5,$B1,$C5
	.dc.b	$AF,$16,$BC,$4F,$8C,$58,$E4,$E0,$F3,$92,$7C,$2D,$9E,$6D,$38,$46
	.dc.b	$89,$1A,$6C,$6A,$30,$47,$04,$9F,$19,$1F,$61,$92,$B9,$7D,$C6,$4A
	.dc.b	$E5,$F7,$1C,$08,$68,$A6,$8C,$10,$C1,$0F,$80,$F5,$69,$C8,$A7,$12
	.dc.b	$9C,$67,$E1,$C7,$16,$C7,$16,$BC,$5A,$F1,$9F,$91,$3C,$5B,$1C,$5F
	.dc.b	$AB,$D2,$EE,$6A,$F6,$3F,$C1,$96,$AD,$4E,$63,$FC,$21,$18,$9D,$B6
	.dc.b	$3F,$C2,$9F,$1C,$F8,$E4,$71,$28,$8D,$39,$36,$E2,$BC,$6D,$8F,$F1
	.dc.b	$07,$03,$DA,$EB,$5E,$88,$C4,$F3,$0E,$23,$53,$C2,$7F,$C5,$5E,$E3
	.dc.b	$A7,$11,$E6,$8C,$3C,$DB,$5F,$19,$B1,$FE,$33,$E4,$7B,$BC,$A3,$E0
	.dc.b	$C6,$1F,$23,$6B,$E7,$B6,$3F,$C7,$9F,$22,$F9,$1E,$50,$8C,$3E,$D6
	.dc.b	$D7,$D8,$6C,$78,$FC,$CE,$37,$FE,$E6,$AE,$F1,$83,$AF,$67,$FA,$B2
	.dc.b	$3F,$CF,$7E,$CE,$09,$7F,$2F,$E7,$1F,$23,$E4,$6C,$66,$30,$7A,$38
	.dc.b	$3E,$9F,$C0,$7D,$09,$34,$8F,$D2,$E0,$FC,$3C,$7C,$D6,$C6,$4E,$25
	.dc.b	$8F,$93,$BE,$A3,$F1,$BA,$FC,$F9,$CC,$70,$33,$3B,$7F,$2B,$EF,$3E
	.dc.b	$A3,$F2,$FF,$03,$FA,$28,$C4,$A3,$13,$A8,$FF,$39,$4E,$A7,$B9,$23
	.dc.b	$F0,$95,$B7,$C2,$73,$3F,$47,$E0,$BE,$3D,$E2,$FE,$A3,$E0,$7F,$CF
	.dc.b	$D3,$98,$C5,$70,$AD,$AA,$DC,$61,$E3,$EA,$BC,$0E,$35,$FB,$4C,$13
	.dc.b	$98,$E4,$79,$1C,$57,$2A,$E3,$3F,$BC,$ED,$FE,$0B,$1F,$C0,$BB,$FF
	.dc.b	$85,$B7,$BF,$FC,$09,$3A,$AE,$9F,$07,$FC,$13,$F1,$3F,$2B,$63,$22
	.dc.b	$32,$D3,$C9,$97,$F0,$CF,$91,$F6,$B6,$32,$23,$2D,$3D,$9F,$80,$FE
	.dc.b	$D1,$A3,$97,$E3,$C6,$63,$E3,$52,$31,$1F,$07,$4F,$7B,$FC,$91,$AF
	.dc.b	$F1,$EC,$BE,$03,$63,$2B,$6F,$A3,$E4,$74,$6C,$78,$CF,$A5,$19,$C0
	.dc.b	$F8,$A3,$F9,$B5,$74,$6E,$BF,$42,$31,$87,$8A,$7C,$17,$FE,$30,$D3
	.dc.b	$F0,$BE,$2F,$D1,$9F,$7B,$5D,$B7,$2F,$C3,$FD,$B7,$F3,$CF,$CC,$7F
	.dc.b	$1A,$DA,$FC,$43,$5F,$92,$D7,$C3,$F7,$7F,$D1,$3E,$0F,$F2,$AD,$8F
	.dc.b	$C8,$35,$FA,$2D,$7E,$86,$BE,$D2,$71,$FE,$3C,$FC,$FF,$CD,$6F,$2F
	.dc.b	$CE,$B7,$53,$4D,$7C,$2F,$E8,$7E,$CB,$F5,$ED,$C7,$F6,$AD,$7F,$50
	.dc.b	$D7,$C7,$87,$85,$FC,$16,$3E,$47,$FA,$FA,$F8,$CF,$E4,$FC,$1F,$81
	.dc.b	$9F,$EC,$B3,$9F,$0B,$F4,$9A,$70,$BD,$B6,$33,$5E,$2E,$8F,$ED,$BF
	.dc.b	$45,$CE,$3F,$19,$9A,$73,$7E,$A6,$CF,$11,$5C,$5F,$C4,$E7,$EB,$7F
	.dc.b	$31,$97,$AD,$FD,$1F,$DC,$7F,$7C,$FC,$CB,$C1,$8F,$F0,$A8,$C3,$FC
	.dc.b	$25,$B3,$C0,$4F,$0E,$07,$DF,$FF,$66,$BF,$E2,$0D,$AF,$D8,$B5,$FC
	.dc.b	$CB,$61,$A3,$E0,$BF,$19,$1F,$61,$B5,$F6,$1A,$FC,$66,$BF,$00,$FB
	.dc.b	$39,$FC,$17,$EE,$9E,$41,$C9,$E9,$C3,$9D,$FF,$2A,$6E,$3E,$CA,$DC
	.dc.b	$F4,$61,$F3,$2F,$B3,$E1,$30,$68,$F0,$9F,$E2,$8F,$55,$EA,$BD,$6F
	.dc.b	$F1,$6F,$C1,$FF,$39,$A5,$6A,$73,$7C,$6F,$85,$2F,$85,$EE,$31,$5B
	.dc.b	$BE,$31,$FE,$3A,$9D,$57,$0B,$E6,$DA,$DD,$7F,$12,$3F,$A1,$10,$FD
	.dc.b	$29,$4F,$C6,$94,$FD,$E9,$FD,$63,$2F,$F2,$4F,$9F,$F3,$E0,$D2,$65
	.dc.b	$D2,$21,$D0,$A7,$DF,$38,$BF,$85,$69,$F8,$52,$9B,$8F,$07,$C1,$49
	.dc.b	$FC,$13,$67,$F3,$9F,$99,$FE,$6F,$8B,$FC,$BE,$D8,$34,$FD,$77,$7F
	.dc.b	$F9,$B2,$3E,$01,$F1,$CF,$3D,$FE,$73,$9B,$FC,$D3,$F4,$3F,$AE,$E4
	.dc.b	$FD,$E5,$23,$13,$AA,$FD,$F4,$E5,$FE,$10,$BD,$AF,$E0,$DC,$FF,$C3
	.dc.b	$E3,$13,$A9,$CF,$57,$F8,$F1,$F3,$F8,$5F,$E8,$6A,$F3,$E7,$E9,$7C
	.dc.b	$F9,$FF,$82,$7E,$1F,$F9,$4F,$37,$FA,$01,$B2,$69,$FC,$D9,$FB,$67
	.dc.b	$F8,$3F,$BF,$EF,$B4,$C9,$FD,$C9,$CB,$F7,$48,$7E,$20,$F9,$3D,$0F
	.dc.b	$03,$71,$FD,$40,$E6,$5E,$4F,$EA,$2E,$36,$FE,$BA,$BF,$18,$FE,$BA
	.dc.b	$D3,$FA,$E9,$4F,$EB,$A5,$3F,$B3,$9F,$3C,$FF,$15,$78,$9F,$EA,$2D
	.dc.b	$FF,$B7,$C6,$4F,$ED,$E5,$3D,$47,$B0,$FC,$21,$EE,$AF,$F0,$2F,$37
	.dc.b	$F8,$F3,$C5,$FF,$56,$B6,$4D,$3E,$BF,$0F,$D9,$8F,$D1,$9D,$0F,$F0
	.dc.b	$1F,$8D,$C9,$FF,$11,$CF,$F8,$89,$E6,$60,$FF,$16,$FC,$F7,$F8,$B0
	.dc.b	$8E,$0F,$F1,$CE,$3F,$E3,$82,$38,$3C,$3C,$7F,$89,$7D,$CD,$7F,$C2
	.dc.b	$BC,$5F,$EC,$8F,$84,$FF,$28,$41,$A4,$7F,$CC,$5E,$6E,$DA,$6C,$B1
	.dc.b	$FE,$64,$8D,$47,$F9,$CB,$E0,$BF,$D0,$A4,$7F,$CE,$E1,$FF,$3F,$B4
	.dc.b	$FF,$3F,$FA,$5F,$ED,$C8,$C9,$23,$8F,$C1,$7F,$A4,$08,$E0,$FF,$4D
	.dc.b	$3F,$E9,$D4,$D2,$3F,$ED,$EF,$7B,$FD,$D3,$1F,$EA,$92,$46,$4E,$27
	.dc.b	$FB,$B0,$7B,$1F,$EB,$4F,$0B,$FD,$F6,$D8,$D3,$5F,$F7,$9B,$5F,$E5
	.dc.b	$FE,$CF,$FB,$E2,$7F,$D8,$71,$FD,$20,$FF,$7C,$F3,$A7,$51,$E1,$8C
	.dc.b	$3B,$3F,$DC,$3E,$DE,$9F,$F5,$AB,$5C,$9B,$3F,$CE,$4E,$5F,$F7,$07
	.dc.b	$BB,$7D,$DF,$F7,$0E,$1F,$ED,$F8,$C1,$BF,$4F,$FD,$DD,$1F,$EC,$02
	.dc.b	$57,$DC,$7E,$81,$1F,$F6,$63,$51,$3D,$C6,$BE,$36,$C7,$9D,$B3,$FE
	.dc.b	$DD,$3E,$94,$3D,$F7,$C0,$7A,$DB,$1E,$F6,$CF,$06,$C7,$F9,$2A,$3F
	.dc.b	$D5,$44,$70,$7D,$FE,$67,$6F,$82,$3F,$87,$35,$FE,$80,$D8,$DC,$6D
	.dc.b	$AF,$A1,$AF,$58,$E0,$47,$83,$5E,$0D,$78,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$3E,$99,$EA,$38,$98,$8F,$3B,$63,$C4,$D3,$67,$C4,$29,$CC,$F0
	.dc.b	$1B,$29,$B2,$9B,$3C,$86,$4A,$64,$E4,$62,$3C,$AE,$CC,$90,$C1,$0E
	.dc.b	$45,$78,$B8,$23,$82,$47,$C0,$8E,$0D,$76,$D8,$DC,$7C,$26,$BE,$E6
	.dc.b	$BE,$98,$E0,$47,$83,$5E,$0D,$78,$1A,$21,$A2,$9A,$30,$43,$04,$3E
	.dc.b	$31,$EA,$38,$98,$8F,$3B,$63,$C4,$D3,$67,$C4,$29,$CC,$F0,$1B,$29
	.dc.b	$B2,$9B,$3C,$86,$4A,$64,$E4,$62,$3C,$AE,$CC,$90,$C1,$0E,$45,$78
	.dc.b	$B8,$23,$83,$3E,$6F,$F9,$E6,$3F,$A1,$35,$FF,$3D,$1F,$C5,$1C,$3F
	.dc.b	$E1,$3F,$2B,$F8,$FF,$47,$F0,$16,$C7,$F0,$D6,$BF,$AF,$6B,$C4,$FF
	.dc.b	$48,$16,$39,$38,$3F,$CA,$11,$C1,$A6,$69,$19,$6C,$66,$30,$47,$04
	.dc.b	$9E,$E4,$7D,$06,$4A,$E5,$F5,$19,$2B,$97,$D4,$70,$21,$A2,$9A,$30
	.dc.b	$43,$04,$3D,$EF,$56,$9C,$8A,$71,$29,$C6,$7E,$1C,$71,$6C,$71,$6B
	.dc.b	$C5,$AF,$19,$F9,$13,$C5,$B1,$C5,$AF,$16,$BC,$4F,$8C,$58,$E4,$E0
	.dc.b	$F3,$92,$7C,$2D,$9E,$6D,$38,$46,$89,$1A,$6C,$6A,$30,$47,$04,$9F
	.dc.b	$19,$1F,$61,$92,$B9,$7D,$C6,$4A,$E5,$F7,$1C,$08,$68,$A6,$8C,$10
	.dc.b	$C1,$0F,$80,$F5,$69,$C8,$A7,$12,$9C,$67,$E1,$C7,$16,$C7,$16,$BC
	.dc.b	$5A,$F1,$9F,$91,$3C,$5B,$1F,$E4,$7E,$97,$6B,$C7,$7B,$FF,$E1,$A9
	.dc.b	$CC,$62,$B8,$56,$EB,$4D,$8F,$F2,$96,$FF,$EC,$79,$FF,$11,$B6,$33
	.dc.b	$5F,$E2,$E9,$CB,$63,$FC,$B9,$1C,$38,$31,$1F,$E0,$2A,$FF,$18,$CE
	.dc.b	$63,$C8,$7F,$81,$BE,$01,$F1,$1C,$23,$FC,$77,$18,$76,$62,$34,$A3
	.dc.b	$53,$FE,$3A,$F8,$27,$89,$A6,$D7,$20,$E2,$78,$4F,$4E,$97,$C1,$D2
	.dc.b	$5F,$18,$CB,$CF,$B3,$FC,$3F,$7F,$E7,$88,$F2,$C6,$1F,$F3,$C9,$E8
	.dc.b	$E8,$78,$14,$91,$E5,$8C,$BE,$7F,$C4,$FE,$FB,$78,$7E,$0C,$63,$0F
	.dc.b	$F2,$CB,$67,$EF,$F0,$BE,$0F,$EF,$B6,$33,$18,$3F,$CC,$3B,$FC,$F6
	.dc.b	$F0,$7F,$9B,$1B,$3C,$4F,$B1,$19,$7F,$D3,$5C,$1F,$3A,$7E,$5C,$6C
	.dc.b	$FF,$3A,$C6,$5D,$F0,$71,$7F,$D2,$BC,$15,$FF,$3D,$F0,$7C,$EB,$FF
	.dc.b	$3A,$B6,$3F,$D3,$13,$BA,$7E,$B5,$FA,$39,$7C,$1F,$FA,$B6,$73,$A7
	.dc.b	$FA,$1A,$31,$5B,$6C,$66,$BE,$4F,$CC,$7C,$82,$38,$3F,$CB,$B7,$E0
	.dc.b	$E5,$FF,$48,$FC,$17,$46,$CE,$B1,$FF,$48,$CE,$60,$FD,$4F,$A9,$91
	.dc.b	$F6,$1A,$FF,$9A,$BA,$3F,$77,$D5,$F6,$89,$1F,$79,$F7,$7D,$E3,$E5
	.dc.b	$3F,$D4,$7D,$18,$6D,$7A,$27,$4A,$74,$D3,$D1,$F8,$07,$F8,$23,$5F
	.dc.b	$9A,$DD,$0D,$31,$F4,$BF,$EB,$2F,$7B,$E6,$B6,$FE,$13,$5F,$F5,$BD
	.dc.b	$FD,$88,$C4,$7F,$AE,$B4,$F7,$BE,$D2,$1F,$0F,$AF,$F9,$06,$BF,$EE
	.dc.b	$9E,$27,$D9,$BE,$C3,$9C,$61,$B1,$9B,$67,$C6,$30,$FC,$C8,$ED,$5F
	.dc.b	$E2,$3C,$7C,$71,$8C,$3E,$7F,$B7,$F3,$23,$D7,$C0,$8C,$47,$F1,$28
	.dc.b	$C3,$F3,$4D,$C6,$B5,$FC,$CF,$D2,$FF,$B6,$BE,$0F,$F2,$AD,$AF,$C5
	.dc.b	$35,$F9,$3C,$28,$C4,$7F,$20,$87,$DA,$75,$6C,$78,$4D,$46,$4F,$3A
	.dc.b	$CF,$DE,$FD,$3E,$0F,$9E,$FD,$23,$1A,$FF,$92,$B2,$68,$F5,$FD,$8D
	.dc.b	$AF,$04,$EA,$72,$79,$7E,$D7,$FD,$F3,$F4,$FF,$CD,$AB,$D5,$18,$AF
	.dc.b	$D5,$36,$3E,$0B,$5F,$1F,$07,$D9,$69,$E7,$C5,$D6,$F7,$3F,$4B,$87
	.dc.b	$FB,$5A,$73,$18,$A1,$FD,$B6,$B8,$3F,$A7,$AE,$11,$8E,$8F,$44,$7F
	.dc.b	$12,$C7,$8F,$E6,$B8,$E1,$46,$27,$D7,$82,$B5,$39,$8F,$C7,$1F,$3C
	.dc.b	$FA,$9B,$FD,$27,$27,$EA,$BE,$01,$F0,$9E,$8F,$C1,$7C,$71,$89,$46
	.dc.b	$27,$58,$FF,$1A,$AD,$4F,$AB,$EC,$DF,$7B,$FB,$BA,$FE,$3F,$18,$C3
	.dc.b	$F5,$31,$89,$D4,$7E,$F2,$75,$D5,$F0,$7E,$E7,$E6,$4F,$E5,$5B,$7F
	.dc.b	$CE,$1B,$B1,$DC,$FC,$0E,$2F,$D9,$6B,$F0,$63,$EE,$9F,$08,$FE,$FF
	.dc.b	$8B,$EF,$BA,$B6,$3E,$C1,$E5,$8C,$9D,$BA,$3F,$3F,$1D,$F9,$7F,$14
	.dc.b	$D3,$E9,$46,$89,$7E,$38,$7A,$6F,$8B,$7A,$DF,$8A,$F8,$B6,$3F,$EB
	.dc.b	$C6,$DF,$E3,$AF,$85,$6E,$B4,$49,$D5,$F4,$57,$C9,$F0,$C6,$63,$17
	.dc.b	$D2,$77,$5F,$C7,$1F,$F0,$19,$0F,$F0,$39,$4D,$18,$21,$F2,$47,$C1
	.dc.b	$FE,$1F,$3F,$D3,$B6,$33,$C1,$F8,$89,$CB,$CA,$3F,$84,$9D,$CA,$7F
	.dc.b	$67,$DF,$6F,$F1,$E6,$BF,$89,$3A,$99,$29,$F8,$2E,$2F,$F0,$4C,$FE
	.dc.b	$56,$1E,$67,$E7,$67,$31,$F9,$76,$BF,$E4,$8E,$5F,$D4,$7E,$0B,$FA
	.dc.b	$78,$34,$97,$EB,$8F,$CB,$1E,$B2,$9F,$BE,$FC,$DB,$F7,$A9,$A4,$8F
	.dc.b	$E1,$7D,$BE,$EE,$5F,$E2,$75,$EC,$6E,$BF,$C9,$5B,$5E,$1E,$4F,$F3
	.dc.b	$FD,$78,$5A,$FF,$0E,$FC,$33,$C2,$49,$C1,$FE,$03,$F0,$3E,$CC,$D3
	.dc.b	$EE,$97,$0F,$F0,$6C,$6A,$3F,$0A,$43,$EC,$96,$7F,$C6,$7E,$16,$E3
	.dc.b	$FA,$79,$D0,$FE,$B2,$E6,$37,$19,$3F,$AD,$78,$7F,$CC,$BE,$3B,$FB
	.dc.b	$1C,$1A,$47,$FC,$39,$E3,$73,$F1,$7F,$B7,$E5,$FD,$DA,$7A,$61,$FD
	.dc.b	$5A,$BC,$91,$F5,$8F,$51,$4F,$59,$FE,$AD,$F1,$FF,$C1,$1C,$9F,$E0
	.dc.b	$7F,$23,$FC,$AB,$3F,$D2,$67,$2D,$8C,$BD,$37,$FA,$6D,$ED,$FF,$2B
	.dc.b	$3F,$E2,$43,$F2,$A5,$BF,$F1,$4B,$83,$FB,$FF,$93,$D1,$FE,$FD,$E5
	.dc.b	$7F,$56,$7F,$CB,$6D,$7F,$C7,$25,$36,$5C,$FF,$BB,$F9,$7F,$1C,$91
	.dc.b	$F1,$B6,$FF,$22,$ED,$F8,$BC,$BC,$F3,$E9,$E4,$E8,$FF,$35,$C3,$FD
	.dc.b	$1A,$47,$FD,$22,$FF,$97,$FE,$1A,$47,$F9,$8F,$CF,$FE,$DB,$18,$8F
	.dc.b	$F1,$7D,$7F,$A5,$88,$FF,$A6,$9F,$F3,$AA,$F8,$5A,$47,$FC,$DD,$1F
	.dc.b	$DE,$23,$FD,$0F,$E9,$7F,$7A,$F8,$5B,$46,$91,$FF,$49,$99,$2B,$FE
	.dc.b	$9A,$23,$F3,$7D,$7F,$F6,$A4,$7F,$A9,$09,$19,$F5,$FC,$0F,$FA,$CA
	.dc.b	$3F,$D6,$24,$70,$7D,$7F,$63,$FD,$7B,$E4,$7F,$AE,$B6,$FF,$08,$B6
	.dc.b	$7F,$A0,$1F,$7E,$76,$FD,$FF,$43,$FD,$8F,$3F,$7D,$B1,$9F,$6B,$9B
	.dc.b	$4E,$67,$F8,$7B,$DB,$FC,$55,$7F,$B6,$63,$22,$32,$FF,$B6,$8C,$F3
	.dc.b	$FF,$B8,$A7,$E6,$92,$3F,$D6,$A7,$2F,$6B,$FD,$E3,$1F,$EC,$42,$47
	.dc.b	$D0,$7F,$D9,$A4,$7E,$51,$E2,$D5,$AF,$AA,$3F,$41,$EE,$38,$F1,$1F
	.dc.b	$E3,$88,$C3,$EA,$6D,$FE,$8D,$B1,$FA,$77,$FD,$91,$19,$E2,$E7,$29
	.dc.b	$CB,$B8,$CF,$53,$81,$1F,$22,$30,$D8,$CD,$FF,$BC,$8F,$DA,$7B,$CF
	.dc.b	$7B,$F5,$78,$23,$0F,$EB,$9A,$F0,$8D,$C6,$23,$F8,$03,$5F,$E0,$8D
	.dc.b	$7F,$CB,$6D,$7B,$47,$22,$3D,$DA,$F7,$30,$43,$04,$3C,$2F,$06,$9B
	.dc.b	$29,$B3,$CA,$ED,$A6,$CA,$6C,$F4,$BB,$69,$B2,$9B,$3D,$A6,$4A,$66
	.dc.b	$3E,$11,$23,$A4,$68,$F3,$90,$F1,$94,$EE,$53,$81,$A2,$1A,$29,$A3
	.dc.b	$04,$30,$48,$F2,$3E,$96,$BE,$98,$F3,$C6,$23,$C4,$D7,$BB,$5E,$CD
	.dc.b	$7B,$47,$22,$3D,$DA,$F7,$30,$43,$04,$3C,$2F,$06,$9B,$29,$B3,$CA
	.dc.b	$ED,$A6,$CA,$6C,$F4,$BB,$69,$B2,$9B,$3D,$A6,$4A,$66,$3E,$11,$23
	.dc.b	$A4,$68,$F3,$90,$F1,$94,$EE,$53,$81,$A2,$1A,$29,$F7,$BC,$DF,$F3
	.dc.b	$A3,$FD,$DD,$AF,$F7,$B6,$BF,$E8,$0A,$FE,$05,$EA,$7F,$B5,$3C,$EF
	.dc.b	$E1,$2D,$9F,$E0,$2D,$7F,$80,$B5,$FE,$02,$D7,$F8,$73,$70,$FF,$63
	.dc.b	$4F,$49,$FF,$46,$1C,$0A,$70,$2C,$70,$7C,$26,$27,$A1,$1C,$12,$7B
	.dc.b	$11,$E2,$D7,$D2,$6C,$C9,$0C,$BB,$7C,$CD,$7C,$CD,$9F,$34,$70,$39
	.dc.b	$1C,$0A,$70,$2B,$C1,$D1,$C4,$86,$08,$71,$32,$53,$2F,$37,$04,$70
	.dc.b	$49,$F9,$13,$E9,$6C,$F9,$1A,$F6,$9E,$0F,$C7,$31,$1A,$7C,$2D,$7C
	.dc.b	$0D,$8E,$33,$DE,$79,$1C,$0A,$70,$2C,$70,$7C,$86,$27,$A1,$1C,$12
	.dc.b	$7B,$11,$E2,$D7,$D2,$6C,$C9,$0C,$BB,$7C,$CD,$7C,$CD,$9F,$34,$70
	.dc.b	$39,$1C,$0A,$70,$2B,$C1,$D1,$C4,$86,$08,$71,$32,$53,$2F,$37,$04
	.dc.b	$70,$49,$FD,$B6,$7F,$E0,$AF,$7B,$FC,$F4,$DA,$FF,$3F,$35,$FF,$41
	.dc.b	$35,$0F,$80,$7C,$0E,$2B,$FD,$09,$18,$94,$62,$75,$1F,$E8,$B9,$D4
	.dc.b	$FF,$8A,$78,$33,$3F,$E3,$26,$BF,$E3,$16,$BF,$E2,$96,$F1,$3E,$6A
	.dc.b	$92,$7C,$27,$F9,$B5,$E3,$C8,$EE,$A4,$9D,$9F,$EA,$18,$D3,$83,$E0
	.dc.b	$FC,$23,$E1,$3C,$6D,$8D,$07,$19,$7F,$A9,$DB,$1F,$E6,$DF,$85,$F5
	.dc.b	$F6,$1A,$48,$E7,$D5,$ED,$6E,$7F,$0A,$31,$87,$FA,$CD,$B3,$FE,$B7
	.dc.b	$5F,$0D,$B2,$47,$47,$E5,$F0,$FF,$93,$3E,$17,$F0,$7D,$3F,$E5,$06
	.dc.b	$F5,$1A,$7D,$BB,$C1,$F6,$F9,$3F,$0B,$19,$24,$7F,$A8,$9F,$F6,$47
	.dc.b	$07,$CC,$6F,$0F,$09,$D1,$FE,$CF,$E0,$F9,$93,$F2,$78,$75,$3F,$ED
	.dc.b	$59,$CB,$B8,$CB,$F9,$6E,$1F,$98,$F9,$8F,$F6,$E3,$8A,$FF,$4E,$70
	.dc.b	$F0,$94,$E5,$FD,$1E,$9F,$E7,$2F,$80,$F8,$64,$8C,$EA,$8F,$88,$48
	.dc.b	$F8,$86,$4A,$FC,$CF,$83,$91,$C9,$AF,$FB,$C7,$9B,$F5,$51,$F3,$C9
	.dc.b	$1D,$1F,$AB,$D3,$23,$C6,$D7,$F6,$FB,$3D,$D7,$CE,$FC,$59,$FF,$7F
	.dc.b	$4E,$63,$11,$C2,$76,$FC,$43,$6E,$34,$FE,$07,$1B,$F7,$7F,$C0,$97
	.dc.b	$DD,$B4,$69,$8F,$88,$FE,$17,$F0,$1C,$AF,$F6,$94,$9C,$C7,$81,$AF
	.dc.b	$95,$AF,$F1,$4E,$6F,$F5,$2E,$7F,$06,$31,$C1,$FE,$AA,$8C,$65,$F0
	.dc.b	$F5,$E2,$DA,$65,$E3,$8C,$37,$25,$78,$29,$8F,$81,$FA,$24,$3F,$94
	.dc.b	$E3,$E1,$7E,$4F,$CC,$FC,$E7,$DE,$DC,$7E,$B3,$70,$37,$DD,$FE,$63
	.dc.b	$F0,$3E,$36,$DF,$D4,$6B,$F6,$5A,$FC,$CF,$96,$E8,$47,$EE,$1F,$7D
	.dc.b	$F2,$3E,$CC,$7D,$0F,$AB,$93,$FC,$6F,$9F,$BE,$31,$B7,$C5,$F8,$1F
	.dc.b	$41,$23,$A9,$F7,$1F,$77,$C0,$7F,$B4,$BE,$53,$F9,$6C,$FF,$33,$8C
	.dc.b	$6F,$F9,$46,$C7,$E6,$9A,$FD,$CC,$7C,$79,$79,$A3,$18,$FE,$06,$5F
	.dc.b	$4A,$F9,$DF,$A1,$18,$6C,$66,$BC,$F3,$97,$D3,$8B,$1F,$F7,$4F,$E1
	.dc.b	$1E,$88,$C7,$97,$FD,$92,$31,$3A,$9F,$54,$E5,$FC,$25,$79,$8A,$7C
	.dc.b	$FB,$FF,$32,$C7,$ED,$FE,$E3,$FB,$7F,$E4,$1D,$F9,$51,$89,$FD,$9B
	.dc.b	$6F,$F7,$0D,$7F,$57,$E0,$7E,$BA,$7F,$72,$D8,$FD,$F3,$5D,$9F,$A2
	.dc.b	$8F,$DF,$FD,$E7,$F2,$B6,$C7,$F2,$66,$F6,$79,$67,$53,$F8,$43,$DB
	.dc.b	$C9,$F8,$77,$E0,$1F,$C3,$7D,$ED,$BF,$C0,$1B,$7F,$C5,$DB,$91,$DC
	.dc.b	$FC,$7E,$2F,$DF,$34,$EF,$C7,$F8,$37,$07,$8C,$87,$FA,$84,$7C,$4A
	.dc.b	$7F,$D4,$DD,$CF,$3F,$FC,$3F,$39,$AE,$0D,$5E,$87,$F1,$D9,$DD,$F2
	.dc.b	$6C,$F0,$8C,$3E,$06,$D7,$3D,$18,$B8,$BF,$92,$5F,$E7,$CF,$EE,$64
	.dc.b	$3F,$A8,$94,$FD,$B1,$4F,$98,$3D,$AF,$C5,$69,$FE,$3A,$6C,$67,$6F
	.dc.b	$E7,$B3,$97,$AC,$43,$A9,$4F,$E6,$FB,$ED,$FE,$90,$D7,$27,$53,$25
	.dc.b	$3F,$1F,$E5,$FE,$77,$8B,$FC,$9A,$D8,$34,$FD,$17,$1F,$F9,$4B,$F2
	.dc.b	$0B,$26,$93,$0F,$34,$7E,$B4,$F8,$85,$3F,$6A,$3F,$0A,$FD,$CF,$3F
	.dc.b	$F9,$79,$B1,$99,$DC,$7F,$18,$D7,$F8,$36,$BF,$D6,$5B,$5F,$D9,$1A
	.dc.b	$FF,$9A,$1A,$FD,$B3,$CB,$BF,$E0,$23,$EF,$C7,$94,$FE,$41,$87,$F9
	.dc.b	$C3,$E6,$FC,$64,$C7,$D2,$FF,$05,$EE,$FB,$95,$FD,$D9,$AF,$F7,$B6
	.dc.b	$BF,$CC,$CF,$C2,$1E,$2E,$1F,$C8,$3F,$1B,$1F,$23,$FD,$87,$BF,$8E
	.dc.b	$1F,$E0,$B6,$F4,$1A,$63,$C0,$FF,$43,$FE,$43,$FD,$5E,$CD,$23,$FD
	.dc.b	$90,$F1,$9F,$E1,$1F,$0F,$A9,$1C,$9F,$24,$FC,$99,$0C,$1F,$E0,$AF
	.dc.b	$13,$F1,$6F,$F8,$61,$AF,$F8,$75,$AE,$CF,$C7,$9F,$0F,$7F,$A7,$9F
	.dc.b	$C2,$8F,$D0,$67,$FE,$2B,$F1,$B8,$B6,$38,$9F,$E1,$52,$9F,$E1,$C8
	.dc.b	$DE,$9F,$E2,$38,$FD,$EC,$FF,$89,$DB,$19,$7F,$C2,$1D,$BF,$84,$9F
	.dc.b	$F1,$B4,$7C,$26,$F1,$7F,$85,$7C,$7F,$A6,$7D,$E9,$FC,$51,$F1,$0A
	.dc.b	$4F,$23,$FD,$6D,$5F,$E7,$B2,$47,$F9,$61,$FF,$40,$91,$FF,$2F,$B5
	.dc.b	$FF,$2B,$79,$DF,$9D,$7F,$CB,$FA,$7F,$A1,$48,$FF,$99,$9F,$F4,$59
	.dc.b	$1C,$1F,$EA,$31,$ED,$FF,$9D,$1A,$65,$C1,$FE,$83,$72,$7F,$A0,$FA
	.dc.b	$FF,$DA,$B1,$92,$47,$F9,$37,$CF,$FF,$5D,$C7,$FA,$42,$38,$9F,$ED
	.dc.b	$C8,$CB,$BC,$39,$BF,$E7,$A9,$FF,$50,$11,$C1,$FE,$82,$7F,$D4,$E4
	.dc.b	$70,$7F,$AC,$FD,$0F,$F7,$4F,$85,$FE,$B7,$6C,$64,$FE,$A0,$58,$FF
	.dc.b	$5E,$7A,$5A,$8F,$F3,$7F,$7F,$FB,$0E,$73,$1C,$0C,$CE,$DF,$F5,$17
	.dc.b	$9F,$5F,$F4,$EB,$7C,$6F,$F5,$2D,$7C,$F3,$E6,$12,$7F,$D5,$87,$53
	.dc.b	$FD,$C3,$A7,$3D,$3F,$DC,$B3,$F2,$C9,$3F,$24,$E0,$3D,$3F,$43,$FE
	.dc.b	$F3,$32,$E1,$FF,$45,$B5,$FF,$80,$07,$2F,$5D,$EA,$FA,$DB,$19,$8F
	.dc.b	$77,$13,$8A,$92,$7C,$71,$97,$CB,$D1,$FC,$C7,$1F,$F6,$F9,$23,$FD
	.dc.b	$7A,$FF,$72,$F5,$DE,$BF,$F3,$69,$46,$1F,$E7,$2D,$78,$46,$E3,$11
	.dc.b	$FE,$9F,$6B,$FE,$AA,$6A,$3A,$C7,$12,$3D,$9A,$F6,$30,$43,$04,$3C
	.dc.b	$0F,$06,$9B,$29,$B3,$C8,$ED,$A6,$CA,$6C,$F4,$47,$A9,$F4,$9B,$33
	.dc.b	$1C,$DA,$64,$D9,$CC,$A7,$90,$87,$80,$A7,$23,$84,$78,$49,$1D,$E3
	.dc.b	$99,$C4,$87,$13,$99,$A3,$06,$8C,$12,$3A,$BE,$96,$BE,$28,$F0,$46
	.dc.b	$23,$AB,$5E,$8D,$7B,$90,$ED,$1C,$88,$F7,$6B,$DC,$C1,$0C,$10,$F0
	.dc.b	$BC,$1A,$6C,$A6,$CF,$2B,$B6,$9B,$29,$B3,$D1,$1E,$B7,$D2,$6C,$CC
	.dc.b	$73,$69,$93,$67,$32,$9E,$42,$1E,$02,$9C,$8E,$11,$E1,$24,$77,$8E
	.dc.b	$67,$12,$1C,$4E,$67,$DE,$F2,$FF,$CF,$8F,$F9,$C9,$AF,$ED,$4F,$F4
	.dc.b	$51,$4F,$F4,$1F,$4F,$F2,$05,$F3,$6F,$82,$FD,$E7,$9B,$FE,$7C,$8C
	.dc.b	$4F,$FA,$2C,$FE,$00,$E2,$35,$3C,$27,$B4,$E5,$B1,$97,$C9,$19,$6C
	.dc.b	$65,$EA,$D7,$CA,$D8,$E1,$1A,$3C,$86,$8A,$E9,$C4,$F9,$9C,$CF,$32
	.dc.b	$38,$24,$6C,$8F,$A8,$C9,$5C,$BE,$A3,$25,$72,$F9,$DA,$F9,$0E,$44
	.dc.b	$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$49,$F8,$C4,$9F,$87
	.dc.b	$3E,$68,$C4,$FB,$0E,$4E,$23,$53,$C2,$7C,$D3,$96,$C6,$5F,$2C,$65
	.dc.b	$B1,$97,$AB,$5F,$33,$63,$84,$68,$F2,$1A,$2B,$A7,$13,$E6,$73,$3C
	.dc.b	$C8,$E0,$91,$B2,$3E,$A3,$25,$72,$FA,$8C,$95,$CB,$E7,$6B,$E4,$39
	.dc.b	$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$27,$E3,$1F,$1F
	.dc.b	$3F,$F5,$F7,$B3,$FE,$FC,$6C,$FF,$09,$6B,$FC,$35,$AF,$F8,$0F,$83
	.dc.b	$FD,$6B,$3F,$C3,$E1,$18,$7F,$89,$35,$FE,$2C,$D7,$FC,$57,$1C,$63
	.dc.b	$FC,$52,$DE,$0F,$F0,$83,$64,$F7,$E1,$7B,$5A,$9F,$E4,$11,$87,$66
	.dc.b	$23,$4F,$F8,$F1,$AF,$F9,$BF,$AE,$AE,$44,$61,$B1,$D0,$FE,$60,$E2
	.dc.b	$FE,$17,$B4,$F8,$4E,$6D,$AE,$07,$F8,$AD,$C4,$F0,$9F,$0F,$B9,$FC
	.dc.b	$EE,$3E,$63,$6B,$F9,$E3,$5F,$E8,$0D,$7F,$C6,$DD,$2F,$77,$D6,$D8
	.dc.b	$F0,$87,$15,$F6,$5B,$1F,$D2,$7F,$73,$FA,$C9,$A4,$F0,$7E,$3F,$BE
	.dc.b	$F3,$29,$2B,$BC,$65,$F0,$F1,$7F,$BF,$3D,$FF,$C0,$5F,$F9,$35,$B2
	.dc.b	$69,$F8,$12,$3F,$8C,$E3,$E0,$D3,$81,$FD,$87,$7F,$A1,$3F,$36,$32
	.dc.b	$7F,$67,$8C,$BB,$DF,$9B,$F3,$0A,$7F,$98,$77,$FE,$DD,$5F,$52,$32
	.dc.b	$23,$2F,$F7,$1D,$7E,$84,$FC,$DD,$F7,$18,$9D,$47,$2C,$D9,$FF,$9B
	.dc.b	$FD,$FF,$86,$4A,$FD,$5C,$65,$F8,$11,$FA,$52,$38,$36,$3C,$AF,$9A
	.dc.b	$47,$2F,$EB,$48,$FC,$B6,$BF,$C0,$77,$F2,$F4,$7F,$07,$6C,$7F,$07
	.dc.b	$6B,$BB,$FF,$05,$72,$3C,$9F,$D1,$7D,$1F,$F8,$56,$31,$C1,$FE,$8D
	.dc.b	$8C,$4E,$AF,$FC,$39,$5A,$53,$C7,$F2,$AF,$F1,$06,$56,$7F,$C1,$94
	.dc.b	$9C,$C7,$26,$BF,$E9,$B8,$D7,$D0,$F8,$63,$4D,$75,$28,$C4,$7F,$04
	.dc.b	$8C,$3E,$43,$E1,$65,$FE,$16,$EB,$FD,$23,$74,$F8,$B3,$FE,$AB,$9C
	.dc.b	$E3,$EB,$F7,$FA,$E5,$E9,$6E,$26,$9E,$2B,$5F,$27,$C3,$7F,$57,$1F
	.dc.b	$01,$F4,$AD,$F4,$BD,$09,$5E,$93,$DE,$F8,$3C,$8F,$F6,$07,$E4,$7F
	.dc.b	$3D,$B3,$F2,$DA,$FC,$66,$BF,$17,$93,$E9,$DF,$F1,$1C,$7E,$C4,$60
	.dc.b	$FF,$20,$65,$E9,$AF,$C4,$4F,$A4,$8F,$D9,$FB,$FF,$86,$47,$D0,$7D
	.dc.b	$37,$E3,$FE,$01,$FE,$D7,$FC,$AF,$F0,$CD,$9F,$C1,$35,$FB,$AD,$7E
	.dc.b	$A7,$07,$D6,$7E,$6B,$70,$F3,$46,$B6,$F9,$9F,$80,$7D,$62,$47,$93
	.dc.b	$87,$F2,$51,$F6,$BC,$CF,$F7,$57,$E1,$DF,$D2,$AB,$F1,$51,$8A,$FC
	.dc.b	$B3,$63,$F2,$2D,$7C,$1A,$FC,$47,$E3,$9F,$80,$3F,$D0,$1F,$83,$3E
	.dc.b	$03,$F8,$55,$E1,$B3,$FE,$87,$9D,$2E,$BF,$22,$8C,$C7,$9E,$9F,$5E
	.dc.b	$CF,$FD,$13,$1F,$B6,$8C,$5B,$1E,$11,$86,$D7,$09,$FD,$6F,$DD,$FF
	.dc.b	$01,$BF,$DB,$36,$32,$7F,$00,$8C,$B4,$F8,$24,$3E,$29,$F0,$CF,$DF
	.dc.b	$5F,$86,$3C,$D5,$E1,$9F,$EF,$B3,$97,$71,$97,$F0,$3A,$FE,$AB,$0F
	.dc.b	$A1,$1D,$8F,$01,$A7,$2D,$3F,$D6,$5E,$6F,$FA,$E3,$E1,$7E,$24,$0F
	.dc.b	$F5,$AF,$A3,$63,$31,$80,$E2,$79,$57,$FA,$FD,$B3,$CD,$AF,$19,$E1
	.dc.b	$E7,$AF,$B5,$FF,$35,$AF,$E0,$C4,$3F,$8C,$94,$D1,$82,$1F,$18,$FE
	.dc.b	$54,$53,$EE,$98,$3E,$6E,$3F,$CE,$9E,$64,$3F,$54,$77,$3F,$96,$69
	.dc.b	$F1,$7E,$97,$F2,$1D,$28,$C6,$DF,$82,$9C,$C6,$2B,$C5,$3F,$79,$F1
	.dc.b	$3F,$98,$BF,$14,$79,$0F,$31,$F9,$FE,$1F,$A9,$B7,$E8,$B8,$7F,$4A
	.dc.b	$D8,$CD,$FE,$AE,$BF,$BB,$BF,$11,$F5,$94,$FC,$49,$FD,$5F,$8B,$F0
	.dc.b	$1F,$9C,$7B,$1B,$3F,$C0,$25,$39,$6C,$7F,$59,$C7,$E2,$45,$19,$FF
	.dc.b	$63,$8C,$C7,$F5,$86,$BE,$43,$E8,$1E,$43,$D1,$3F,$D9,$F9,$FF,$29
	.dc.b	$F9,$CF,$F2,$89,$34,$8F,$F3,$33,$C2,$5B,$FE,$E9,$D5,$E0,$BF,$EE
	.dc.b	$4F,$8A,$7C,$07,$8B,$A3,$FB,$2B,$F8,$19,$FE,$F1,$D7,$FE,$8B,$8F
	.dc.b	$E9,$47,$FA,$36,$32,$DE,$1F,$F2,$E3,$FD,$54,$7E,$4D,$48,$FF,$57
	.dc.b	$32,$7F,$80,$FB,$FC,$D3,$FD,$92,$3D,$67,$F6,$22,$9F,$DF,$E3,$D5
	.dc.b	$87,$A2,$3E,$F3,$5F,$CE,$5F,$F7,$97,$8B,$72,$FE,$FD,$18,$BF,$EF
	.dc.b	$8F,$63,$E2,$9F,$5C,$86,$27,$FC,$2E,$F1,$F0,$7D,$24,$7A,$1F,$26
	.dc.b	$3F,$C4,$4D,$9F,$98,$FF,$86,$4D,$78,$7F,$E9,$5B,$FF,$18,$F8,$9F
	.dc.b	$66,$31,$3F,$E3,$26,$BF,$67,$BF,$FC,$AE,$FE,$E4,$F4,$9F,$EC,$2F
	.dc.b	$2F,$F5,$AD,$3F,$A6,$97,$1F,$F5,$E1,$FE,$80,$23,$FE,$58,$6B,$FE
	.dc.b	$C8,$8F,$DA,$BF,$E7,$FE,$6F,$F6,$21,$1F,$F2,$FB,$FE,$8A,$23,$83
	.dc.b	$FC,$DE,$3D,$BF,$F3,$A3,$4C,$B8,$3F,$CF,$CE,$4F,$F4,$21,$4F,$F4
	.dc.b	$21,$FC,$ED,$D9,$FE,$8F,$23,$FE,$8D,$38,$0F,$33,$FD,$89,$1F,$ED
	.dc.b	$D2,$47,$FA,$6D,$AF,$FB,$A2,$7B,$3F,$E7,$C8,$FF,$53,$91,$FE,$98
	.dc.b	$7F,$4D,$F3,$7F,$D6,$1D,$BF,$EF,$56,$C6,$83,$87,$8B,$5F,$EA,$DE
	.dc.b	$77,$DE,$9F,$F7,$F3,$63,$FD,$8D,$7F,$D6,$E7,$2D,$8F,$9F,$1F,$38
	.dc.b	$8E,$0F,$F6,$D9,$B3,$E6,$BF,$40,$8F,$C9,$7F,$D7,$44,$70,$7A,$FD
	.dc.b	$0F,$F7,$AC,$7B,$63,$FD,$F5,$9F,$57,$FD,$FC,$7F,$63,$38,$39,$5E
	.dc.b	$93,$D2,$C3,$63,$81,$F3,$9C,$57,$76,$C7,$FB,$4E,$3C,$5E,$73,$6E
	.dc.b	$CD,$9E,$67,$F8,$0D,$C4,$FF,$B8,$89,$1F,$EB,$A6,$BF,$EA,$D7,$FC
	.dc.b	$FC,$47,$FD,$22,$FC,$1F,$4D,$EA,$7F,$00,$6C,$FF,$2D,$6B,$B8,$D1
	.dc.b	$1D,$35,$D1,$82,$18,$21,$F8,$D8,$EA,$D8,$E4,$D3,$89,$4E,$27,$E6
	.dc.b	$A3,$8B,$63,$8B,$4E,$25,$38,$9F,$E2,$D8,$E2,$D8,$E2,$D3,$89,$4E
	.dc.b	$22,$38,$B6,$38,$B4,$E2,$53,$8C,$70,$24,$70,$6C,$70,$6B,$C2,$34
	.dc.b	$47,$4D,$74,$60,$86,$08,$7C,$B2,$1F,$14,$A7,$BC,$A7,$A4,$B1,$E4
	.dc.b	$6C,$78,$5A,$78,$0A,$78,$0E,$C4,$3B,$14,$EC,$58,$EC,$D8,$EC,$D3
	.dc.b	$B1,$4E,$C7,$42,$1D,$0B,$1D,$1B,$1D,$1A,$74,$29,$D0,$E4,$48,$E4
	.dc.b	$D8,$E4,$D3,$91,$4E,$51,$C0,$91,$C1,$B1,$C1,$AF,$08,$D1,$F8,$EF
	.dc.b	$1D,$E3,$7F,$06,$9F,$F3,$F3,$5F,$F4,$31,$FE,$8E,$71,$4F,$96,$FF
	.dc.b	$47,$78,$FF,$C0,$DB,$5F,$B5,$6B,$FA,$F6,$BF,$A9,$6B,$FA,$36,$CF
	.dc.b	$58,$FF,$4B,$9B,$29,$B2,$BB,$9F,$14,$62,$79,$91,$C1,$23,$64,$7D
	.dc.b	$06,$4A,$E5,$F5,$19,$2B,$97,$CC,$D7,$CC,$D9,$F3,$47,$42,$1C,$8A
	.dc.b	$72,$29,$C8,$AF,$27,$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9,$C1
	.dc.b	$1C,$12,$7E,$64,$FB,$5B,$3E,$66,$BE,$16,$BD,$9A,$F5,$6C,$F5,$8F
	.dc.b	$9C,$6C,$A6,$CA,$EE,$7C,$51,$89,$E6,$47,$04,$8D,$91,$F4,$19,$2B
	.dc.b	$97,$D4,$64,$AE,$5F,$33,$5F,$33,$67,$CD,$1D,$08,$72,$29,$C8,$A7
	.dc.b	$22,$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$03,$AD,$CC,$E2,$7A
	.dc.b	$BF,$E2,$19,$CC,$62,$BF,$C8,$11,$89,$D5,$71,$7F,$C5,$1E,$C3,$C6
	.dc.b	$EF,$18,$6D,$FF,$92,$67,$4F,$53,$FC,$A2,$E2,$7F,$C4,$3C,$3C,$27
	.dc.b	$FC,$F4,$6C,$FF,$10,$39,$69,$93,$FC,$67,$ED,$65,$A6,$78,$D1,$96
	.dc.b	$C7,$1E,$67,$37,$16,$C6,$63,$07,$5E,$A7,$B5,$49,$3E,$38,$CB,$E5
	.dc.b	$F6,$3F,$CD,$F9,$FF,$A1,$89,$1F,$E5,$C3,$D5,$ED,$F9,$23,$FC,$A8
	.dc.b	$D8,$FF,$39,$35,$FF,$3C,$B5,$FB,$EB,$DC,$F8,$4D,$8E,$B1,$83,$E3
	.dc.b	$7B,$9E,$EA,$FF,$31,$36,$7F,$D0,$0D,$7F,$D0,$ED,$7F,$1A,$BE,$BB
	.dc.b	$99,$23,$EA,$46,$5F,$AD,$D9,$FE,$5B,$F7,$3E,$FC,$1A,$47,$FD,$21
	.dc.b	$BF,$D6,$8F,$F3,$47,$07,$E7,$1B,$3C,$4D,$C6,$5F,$99,$1F,$EB,$42
	.dc.b	$38,$3F,$3B,$C9,$FA,$3F,$A8,$FC,$EF,$39,$A4,$9F,$F5,$17,$D9,$7F
	.dc.b	$6E,$74,$63,$B3,$F5,$F3,$FE,$8B,$9D,$36,$33,$3F,$EA,$BC,$FF,$CF
	.dc.b	$9E,$B7,$C2,$24,$67,$9F,$F7,$31,$EE,$24,$7B,$83,$87,$F4,$FE,$EF
	.dc.b	$CC,$23,$CC,$3E,$35,$EC,$D2,$3E,$B3,$27,$C6,$CF,$FB,$BF,$67,$F0
	.dc.b	$6E,$EF,$8E,$49,$FF,$65,$F5,$3E,$0D,$E8,$7F,$81,$A3,$0D,$8C,$CF
	.dc.b	$F8,$12,$72,$DF,$33,$DA,$DD,$39,$46,$30,$FF,$6A,$52,$31,$5E,$25
	.dc.b	$F7,$FC,$09,$1D,$5F,$F0,$BE,$5E,$B8,$F7,$5F,$AD,$FF,$72,$C6,$5F
	.dc.b	$E5,$77,$ED,$F0,$7D,$44,$8F,$56,$5F,$C7,$AF,$94,$7F,$32,$BE,$4D
	.dc.b	$E4,$FE,$65,$3A,$F9,$77,$D0,$3D,$6F,$F1,$6C,$61,$B1,$9B,$FF,$15
	.dc.b	$CE,$5F,$91,$F3,$1F,$D0,$3E,$FF,$E2,$EE,$8C,$4F,$C5,$6B,$EC,$8D
	.dc.b	$29,$E3,$F8,$17,$D0,$24,$7D,$57,$E4,$E7,$EA,$8F,$92,$43,$E4,$B9
	.dc.b	$7E,$4E,$BF,$51,$F8,$FC,$1F,$83,$23,$F8,$3F,$C1,$DF,$09,$FD,$6B
	.dc.b	$F0,$8F,$BF,$18,$94,$62,$75,$1F,$92,$9D,$4F,$D2,$8F,$EC,$58,$78
	.dc.b	$DF,$A5,$CD,$F9,$C8,$C2,$FC,$69,$F0,$3F,$E6,$4E,$6F,$ED,$77,$FD
	.dc.b	$A6,$72,$B9,$3F,$A3,$B9,$AE,$0F,$F7,$2B,$E9,$18,$8F,$E9,$F9,$3D
	.dc.b	$AF,$EE,$FF,$85,$F2,$C9,$39,$A4,$E6,$30,$DA,$E3,$5F,$46,$BD,$B8
	.dc.b	$FC,$1E,$7F,$C8,$FE,$14,$FD,$DB,$FA,$EF,$D2,$7A,$23,$18,$7F,$82
	.dc.b	$23,$13,$A8,$F8,$33,$AC,$15,$E7,$F8,$77,$91,$ED,$B4,$62,$BE,$34
	.dc.b	$7F,$06,$8C,$65,$C6,$B8,$74,$3F,$09,$FF,$0C,$FC,$67,$C3,$8C,$62
	.dc.b	$8C,$4E,$B1,$FE,$00,$D9,$FC,$8F,$17,$DD,$26,$3F,$B4,$AF,$2E,$3E
	.dc.b	$7C,$BC,$71,$E5,$8C,$1F,$5E,$FF,$13,$7C,$9F,$9C,$62,$FE,$77,$27
	.dc.b	$E3,$63,$06,$BD,$7F,$F5,$AF,$B3,$FE,$34,$F8,$9F,$C4,$4F,$FA,$DA
	.dc.b	$31,$7F,$8E,$AD,$CA,$B6,$A3,$31,$86,$CF,$1F,$25,$7D,$2F,$B1,$CE
	.dc.b	$75,$19,$8F,$69,$FD,$88,$87,$F4,$12,$9F,$30,$A7,$EA,$8F,$E6,$7A
	.dc.b	$7D,$F7,$EE,$1F,$CA,$0F,$E5,$2E,$74,$F9,$D1,$93,$91,$DC,$87,$73
	.dc.b	$EE,$8F,$81,$FC,$0B,$63,$31,$83,$C3,$12,$3B,$B4,$FF,$32,$F1,$7E
	.dc.b	$6F,$E0,$BF,$3F,$3F,$E6,$86,$C9,$A7,$EF,$9F,$8E,$79,$8F,$49,$E7
	.dc.b	$3F,$A3,$F4,$7E,$E7,$F0,$CF,$DB,$41,$A4,$AF,$E1,$BC,$1F,$C1,$A7
	.dc.b	$E2,$35,$FE,$1A,$D8,$DF,$5F,$F1,$29,$F2,$74,$7F,$9E,$9F,$AA,$79
	.dc.b	$8F,$21,$FC,$43,$B7,$E0,$BF,$C9,$DA,$FA,$1A,$EC,$67,$FD,$C3,$BF
	.dc.b	$F9,$D4,$5E,$FF,$E4,$AF,$F8,$0A,$32,$D8,$CB,$E2,$3C,$87,$C8,$26
	.dc.b	$3F,$DE,$FC,$0D,$C7,$F9,$D3,$A9,$18,$9D,$4F,$46,$BF,$D8,$3C,$2F
	.dc.b	$F5,$0F,$DF,$58,$34,$8F,$F3,$CF,$13,$FB,$53,$EA,$6B,$9F,$17,$FB
	.dc.b	$83,$96,$BF,$81,$3D,$64,$DB,$E9,$3F,$0F,$C7,$F4,$6B,$FE,$A2,$8F
	.dc.b	$95,$18,$3D,$EF,$1E,$CF,$F5,$B3,$F6,$AB,$9B,$FD,$DF,$6F,$85,$B7
	.dc.b	$CE,$8C,$1F,$0D,$C3,$F2,$7C,$8D,$91,$D3,$E4,$F2,$64,$7C,$BC,$BF
	.dc.b	$C6,$AF,$DF,$8F,$2E,$1E,$28,$FB,$07,$9F,$93,$EB,$F9,$FF,$E6,$5F
	.dc.b	$87,$FF,$27,$12,$7F,$C9,$8F,$FA,$03,$8A,$47,$F9,$55,$AF,$F6,$EF
	.dc.b	$07,$FD,$6B,$7F,$E7,$E2,$47,$F9,$85,$FF,$44,$91,$FF,$2F,$35,$FE
	.dc.b	$EF,$7F,$E7,$7E,$1F,$F3,$01,$23,$FC,$BC,$FF,$A0,$08,$FF,$97,$DA
	.dc.b	$FF,$80,$3D,$8F,$F0,$0F,$C1,$FE,$69,$B1,$FE,$B9,$8C,$3F,$EF,$07
	.dc.b	$FD,$2A,$47,$FC,$F0,$D7,$FD,$BA,$FF,$A8,$88,$FF,$9D,$DF,$F5,$9F
	.dc.b	$AD,$FE,$FA,$F0,$2C,$FF,$3C,$3F,$D6,$B1,$96,$99,$3F,$D6,$F9,$E5
	.dc.b	$A6,$4F,$F5,$CB,$96,$99,$1E,$CF,$FB,$26,$BF,$DA,$7E,$CE,$E7,$2D
	.dc.b	$9D,$47,$FB,$5B,$D7,$C1,$1E,$91,$FE,$F0,$AF,$8F,$1D,$F2,$FF,$16
	.dc.b	$C6,$09,$1F,$C3,$3D,$AF,$F7,$5C,$7F,$0A,$6C,$67,$DB,$F2,$46,$23
	.dc.b	$FC,$6F,$EE,$7F,$BE,$AB,$F8,$8B,$63,$F8,$BB,$5D,$E7,$ED,$E1,$F3
	.dc.b	$47,$FA,$BE,$BF,$D9,$E4,$7F,$D5,$8F,$FB,$68,$8E,$0F,$E4,$7E,$EB
	.dc.b	$AF,$E4,$36,$3E,$73,$67,$F9,$2B,$63,$FD,$61,$1F,$ED,$42,$38,$3F
	.dc.b	$97,$FB,$CF,$7F,$F8,$4B,$63,$F9,$4B,$5F,$E6,$2D,$78,$C7,$02,$3C
	.dc.b	$1A,$F0,$6B,$C0,$D1,$0D,$14,$D1,$82,$18,$21,$FE,$6C,$8F,$23,$63
	.dc.b	$BB,$4E,$85,$39,$94,$E6,$04,$74,$70,$D7,$31,$D9,$CB,$5C,$9E,$03
	.dc.b	$25,$32,$78,$8C,$94,$C9,$E4,$29,$E7,$21,$C0,$D4,$79,$49,$1E,$36
	.dc.b	$C7,$85,$AF,$76,$BC,$A3,$81,$1E,$0D,$78,$35,$E0,$68,$86,$8A,$68
	.dc.b	$C1,$0C,$10,$F6,$9F,$00,$F7,$18,$8F,$3B,$63,$C4,$D3,$B9,$4E,$C5
	.dc.b	$3B,$1D,$08,$77,$24,$76,$72,$D7,$31,$E0,$72,$D7,$27,$88,$C9,$4C
	.dc.b	$9E,$43,$25,$32,$79,$8A,$7A,$48,$70,$3F,$C5,$3E,$67,$FA,$01,$BE
	.dc.b	$67,$F3,$37,$FD,$00,$D8,$D4,$7E,$A7,$EA,$7F,$41,$E6,$FE,$F9,$B1
	.dc.b	$FC,$19,$AE,$CF,$F4,$59,$FA,$47,$2D,$32,$7E,$A9,$CB,$4C,$E0,$8C
	.dc.b	$B6,$33,$18,$23,$82,$4F,$9D,$A7,$9C,$A7,$9D,$D1,$1D,$35,$D1,$82
	.dc.b	$18,$21,$ED,$78,$34,$D9,$4D,$9E,$F7,$6D,$36,$53,$73,$F0,$63,$6D
	.dc.b	$8D,$B5,$DC,$FC,$49,$DB,$63,$6D,$76,$7C,$43,$CC,$43,$C4,$59,$EC
	.dc.b	$D8,$E4,$D3,$91,$C0,$91,$C1,$B1,$C1,$A7,$08,$D1,$23,$4D,$8D,$46
	.dc.b	$08,$E0,$93,$F0,$1A,$7C,$02,$9F,$01,$D1,$1D,$35,$D1,$82,$18,$21
	.dc.b	$F0,$5E,$0D,$36,$53,$67,$C0,$76,$D3,$65,$37,$3F,$06,$36,$D8,$DB
	.dc.b	$5D,$CF,$CD,$EB,$78,$2E,$B7,$C0,$7F,$AA,$27,$31,$8B,$FF,$57,$C6
	.dc.b	$27,$55,$C5,$FF,$05,$EF,$FD,$8D,$B5,$C4,$FE,$CC,$37,$FF,$0A,$46
	.dc.b	$8F,$F0,$9F,$AD,$C6,$7F,$C3,$6D,$78,$CA,$31,$1F,$DC,$63,$0F,$73
	.dc.b	$FB,$A7,$0B,$A3,$70,$8C,$3B,$31,$1A,$8C,$DF,$FB,$2B,$DD,$7C,$0F
	.dc.b	$26,$DF,$16,$CE,$E7,$BF,$3B,$A2,$92,$FC,$11,$97,$C5,$D0,$EC,$F3
	.dc.b	$B6,$33,$18,$3D,$7D,$7E,$68,$FF,$31,$1E,$B7,$0F,$D0,$6B,$FE,$17
	.dc.b	$EB,$79,$FF,$11,$B1,$F0,$DB,$9F,$FA,$46,$7E,$4F,$C8,$3C,$3A,$4B
	.dc.b	$F9,$51,$97,$E6,$7E,$4D,$F3,$DB,$C1,$F4,$A3,$18,$7F,$BD,$9B,$3F
	.dc.b	$EF,$CA,$7C,$5F,$80,$91,$D9,$FA,$DC,$5F,$E3,$1F,$83,$FF,$7F,$E3
	.dc.b	$F7,$1B,$19,$AF,$C8,$9F,$78,$47,$D7,$69,$AE,$3F,$BB,$19,$24,$7C
	.dc.b	$E7,$9B,$FE,$98,$9E,$CF,$CF,$6B,$FE,$45,$DB,$C8,$FF,$9D,$B9,$BF
	.dc.b	$C9,$1F,$07,$FA,$62,$46,$43,$E8,$F0,$9E,$53,$F0,$CE,$35,$FA,$C9
	.dc.b	$CC,$62,$3F,$CA,$91,$87,$99,$B9,$E4,$E0,$FF,$71,$63,$FE,$5B,$F2
	.dc.b	$FE,$4C,$1A,$47,$E3,$74,$ED,$A6,$CB,$1F,$34,$F5,$57,$E0,$FE,$0A
	.dc.b	$EF,$FE,$69,$6D,$1A,$70,$CB,$51,$E9,$6B,$EE,$AF,$5C,$EE,$39,$1F
	.dc.b	$C6,$3B,$7F,$8F,$78,$5F,$C6,$3C,$0F,$F0,$7F,$81,$FC,$7A,$30,$79
	.dc.b	$B5,$FF,$0A,$ED,$F4,$1B,$97,$F0,$89,$D6,$3F,$E1,$CA,$FA,$BE,$3F
	.dc.b	$FA,$0F,$DB,$F7,$B6,$7E,$83,$5F,$43,$54,$FD,$EF,$C0,$7B,$92,$3B
	.dc.b	$60,$8F,$F1,$2E,$5E,$99,$F7,$67,$EC,$D3,$DF,$18,$CF,$E8,$FC,$3F
	.dc.b	$A8,$91,$D1,$E7,$3C,$23,$F9,$BD,$F5,$6C,$65,$FF,$50,$46,$5F,$4F
	.dc.b	$EE,$5E,$A7,$E2,$36,$FF,$34,$D8,$FC,$C3,$5F,$CB,$BF,$C7,$FF,$3C
	.dc.b	$DF,$83,$FE,$AD,$6C,$9A,$74,$F7,$BE,$BD,$1A,$49,$F8,$99,$CC,$3E
	.dc.b	$FC,$7C,$53,$C7,$E4,$F2,$23,$EA,$7F,$CA,$7F,$24,$FA,$3F,$E9,$6D
	.dc.b	$9F,$E9,$4D,$F0,$BF,$82,$B6,$3B,$1F,$2B,$0F,$13,$83,$E4,$ED,$EA
	.dc.b	$23,$F2,$9F,$F3,$1F,$DC,$3C,$EF,$E1,$3C,$9F,$C7,$23,$15,$FC,$22
	.dc.b	$31,$3A,$8F,$EE,$33,$AE,$DF,$AD,$1F,$6A,$FE,$E2,$E2,$FB,$8D,$8C
	.dc.b	$C7,$DF,$C1,$C0,$FC,$2F,$FD,$C5,$EB,$74,$9C,$C6,$1F,$BE,$DC,$3E
	.dc.b	$FB,$5F,$C4,$FE,$0D,$FE,$8C,$9F,$C2,$B6,$3F,$0A,$D7,$63,$8B,$C1
	.dc.b	$F0,$6F,$0B,$F6,$57,$FC,$96,$31,$4E,$CD,$D6,$A3,$F9,$3C,$EA,$7E
	.dc.b	$2C,$7C,$1E,$1F,$AE,$FD,$9C,$3F,$12,$F9,$0F,$16,$1E,$86,$99,$8F
	.dc.b	$89,$A3,$2F,$54,$F5,$9E,$6F,$D3,$BE,$ED,$8C,$BB,$3F,$10,$7F,$AD
	.dc.b	$BD,$43,$FA,$18,$FF,$17,$FA,$7F,$87,$9C,$F4,$7C,$7A,$DC,$E5,$E0
	.dc.b	$63,$91,$5C,$2F,$9C,$62,$76,$78,$4B,$5C,$EF,$B4,$65,$BC,$3F,$E4
	.dc.b	$C8,$FE,$64,$43,$F8,$C9,$4F,$D3,$14,$F8,$87,$F0,$AC,$BF,$B0,$3F
	.dc.b	$44,$F9,$AE,$1F,$C2,$9F,$6A,$3C,$0F,$62,$1D,$8F,$C3,$6F,$F2,$BE
	.dc.b	$0F,$F0,$B0,$69,$36,$FB,$D1,$E0,$3F,$B7,$9E,$37,$27,$E8,$1F,$F1
	.dc.b	$E3,$7A,$FF,$07,$C5,$F6,$1B,$3A,$3C,$84,$3E,$09,$E8,$3E,$FF,$37
	.dc.b	$F0,$0F,$5F,$F2,$ED,$BF,$E0,$32,$9C,$ED,$F8,$67,$F2,$7C,$DF,$E1
	.dc.b	$A9,$FE,$22,$7F,$87,$2B,$F7,$B1,$96,$C6,$F4,$F9,$6F,$94,$FD,$31
	.dc.b	$F5,$C7,$47,$F2,$5E,$2F,$E4,$7D,$5F,$96,$6C,$FE,$2C,$FF,$42,$CE
	.dc.b	$DF,$E6,$67,$8F,$95,$D7,$E3,$6C,$66,$3D,$4F,$88,$F8,$C5,$3E,$56
	.dc.b	$3F,$C6,$38,$F9,$12,$3B,$3D,$77,$FD,$5F,$E6,$DF,$D4,$09,$3F,$D6
	.dc.b	$23,$2F,$F5,$AA,$F2,$11,$F4,$3E,$4D,$3E,$3C,$7D,$A3,$E4,$38,$8F
	.dc.b	$F1,$EF,$77,$CF,$F0,$0D,$25,$7D,$1D,$FF,$D7,$0F,$CF,$BE,$AF,$BC
	.dc.b	$FF,$05,$D7,$F8,$1D,$FB,$8D,$7F,$C6,$7E,$07,$73,$C1,$E0,$FE,$86
	.dc.b	$3F,$C2,$E7,$E8,$4B,$CD,$FE,$1C,$70,$7F,$8C,$7C,$2F,$41,$1E,$EF
	.dc.b	$C1,$C7,$CF,$1F,$53,$1F,$39,$F0,$71,$FF,$29,$FB,$5F,$E3,$92,$4F
	.dc.b	$F8,$FC,$3F,$E4,$62,$3F,$E5,$BF,$1B,$F9,$0C,$62,$3F,$D9,$FE,$5F
	.dc.b	$FA,$DB,$BF,$FC,$FC,$48,$FF,$2B,$BF,$E5,$E2,$3F,$E8,$6A,$FF,$62
	.dc.b	$C6,$23,$FC,$9D,$E3,$FE,$A5,$F2,$74,$FF,$93,$C8,$FF,$9E,$5F,$F7
	.dc.b	$09,$1C,$1F,$EB,$6F,$67,$FD,$13,$E5,$FF,$B4,$9B,$B7,$F3,$A6,$C9
	.dc.b	$FF,$4B,$11,$FF,$37,$B5,$FF,$3C,$9B,$3D,$51,$CD,$A6,$4D,$C7,$FA
	.dc.b	$6B,$CE,$FF,$57,$C7,$EC,$CF,$E9,$6E,$3A,$F5,$E7,$7E,$E6,$7F,$D6
	.dc.b	$CD,$7F,$DF,$BE,$7F,$FA,$FE,$7F,$D7,$9A,$E6,$7F,$AA,$CE,$5F,$F5
	.dc.b	$9B,$5F,$EB,$22,$FF,$D9,$55,$E5,$F4,$39,$4E,$5B,$3A,$8F,$F5,$EF
	.dc.b	$A5,$FE,$DA,$AF,$F6,$C4,$7A,$C4,$65,$DF,$A5,$D3,$56,$B4,$93,$C6
	.dc.b	$32,$F3,$F4,$FF,$DE,$98,$7F,$B0,$9B,$1F,$EC,$E6,$BE,$C6,$BF,$DD
	.dc.b	$BD,$4F,$F8,$01,$3F,$EC,$A6,$C7,$FB,$59,$AF,$C4,$6B,$FD,$FB,$8D
	.dc.b	$E9,$FC,$46,$C7,$BD,$B5,$FE,$DD,$3F,$D7,$64,$8F,$C1,$B5,$FA,$3E
	.dc.b	$B3,$D7,$F9,$ED,$8F,$94,$D9,$FF,$70,$B6,$3F,$DB,$11,$8F,$2B,$89
	.dc.b	$23,$E9,$75,$3A,$DD,$38,$6D,$FF,$AD,$49,$5F,$EF,$8F,$11,$EC,$7F
	.dc.b	$3E,$6C,$7F,$59,$6B,$FE,$F4,$6B,$C6,$38,$11,$E0,$D7,$83,$5E,$06
	.dc.b	$88,$68,$A6,$8C,$10,$C1,$0F,$CE,$C7,$91,$B1,$DD,$A7,$42,$9C,$CA
	.dc.b	$73,$3F,$C7,$F1,$CD,$B1,$CD,$A7,$32,$9C,$CA,$73,$3F,$B6,$BC,$5A
	.dc.b	$71,$29,$C4,$A7,$18,$F1,$38,$3C,$E4,$3B,$94,$D1,$82,$18,$23,$D0
	.dc.b	$91,$E5,$6C,$79,$5A,$F9,$5A,$F9,$63,$81,$1E,$0D,$78,$35,$E0,$68
	.dc.b	$86,$8A,$68,$C1,$0C,$10,$F8,$11,$E4,$6C,$77,$69,$D0,$A7,$32,$9C
	.dc.b	$CF,$8F,$1C,$DB,$1C,$DA,$73,$29,$CC,$A7,$33,$E7,$BC,$5A,$71,$29
	.dc.b	$C4,$A7,$18,$F1,$3F,$E8,$33,$F9,$D9,$4F,$F4,$4C,$FF,$93,$FC,$DF
	.dc.b	$F4,$84,$7E,$F7,$EE,$3F,$D3,$3E,$7F,$EE,$5B,$19,$3F,$D1,$67,$E8
	.dc.b	$1C,$BF,$AC,$6B,$93,$67,$F9,$21,$CB,$4C,$DA,$32,$D8,$CC,$60,$8E
	.dc.b	$09,$3E,$76,$9E,$77,$04,$70,$43,$D0,$70,$2B,$B6,$9B,$3D,$8E,$5A
	.dc.b	$64,$F6,$39,$69,$99,$F6,$46,$5B,$19,$9F,$6C,$E5,$B1,$93,$CC,$78
	.dc.b	$0F,$29,$0C,$10,$F0,$99,$29,$99,$F0,$B9,$6B,$98,$F0,$B9,$6B,$93
	.dc.b	$C6,$74,$24,$72,$6C,$70,$69,$C2,$34,$48,$D3,$63,$51,$82,$38,$24
	.dc.b	$FC,$76,$9F,$1D,$C1,$1C,$10,$F6,$1C,$0A,$ED,$A6,$CF,$80,$E5,$A6
	.dc.b	$4F,$6B,$96,$99,$9F,$64,$65,$B1,$99,$F6,$CE,$5B,$19,$3E,$7E,$7F
	.dc.b	$DE,$73,$7A,$9F,$BE,$94,$E6,$30,$D8,$FE,$06,$D7,$FC,$EB,$27,$99
	.dc.b	$FC,$23,$83,$FD,$3B,$5D,$5B,$19,$9E,$13,$9F,$8F,$B3,$C5,$AF,$FA
	.dc.b	$09,$AF,$F8,$A5,$AB,$B3,$FC,$7B,$3E,$B6,$C7,$F9,$A2,$31,$CF,$C3
	.dc.b	$99,$E3,$D2,$57,$A8,$FF,$20,$3D,$B9,$DE,$D5,$24,$EC,$FF,$22,$3E
	.dc.b	$4E,$BF,$F3,$1E,$3F,$35,$B3,$FE,$5E,$34,$7A,$BA,$BC,$EF,$F9,$A9
	.dc.b	$AF,$CD,$6A,$6F,$C8,$FC,$C8,$FF,$38,$B5,$FA,$8D,$4D,$F9,$1F,$83
	.dc.b	$1E,$90,$E2,$FF,$08,$D8,$FE,$86,$D7,$FC,$FE,$BC,$BF,$A6,$D8,$E8
	.dc.b	$D9,$E7,$3F,$63,$C1,$FF,$72,$D7,$85,$B3,$F9,$36,$C7,$FA,$3B,$CB
	.dc.b	$FB,$FC,$9F,$99,$F7,$3F,$2B,$D9,$FE,$52,$6F,$29,$A7,$F9,$F7,$87
	.dc.b	$F2,$CD,$C3,$47,$F9,$FE,$32,$D3,$27,$F9,$E7,$83,$2D,$32,$7E,$00
	.dc.b	$A3,$87,$F0,$EE,$A5,$E7,$BF,$09,$FD,$47,$B5,$FD,$BA,$73,$18,$AF
	.dc.b	$F5,$5C,$61,$4E,$EB,$94,$60,$FD,$41,$4F,$D4,$14,$FD,$41,$47,$4B
	.dc.b	$A7,$C1,$1F,$EB,$AF,$7D,$F4,$18,$8F,$DF,$46,$1E,$4D,$C3,$F8,$53
	.dc.b	$63,$E5,$7B,$FF,$3C,$8E,$0F,$F0,$6F,$88,$F9,$2F,$E1,$5F,$23,$FE
	.dc.b	$94,$6C,$E9,$AF,$FA,$79,$AF,$EE,$BE,$BD,$F3,$7F,$C4,$FE,$0F,$E6
	.dc.b	$46,$25,$18,$9D,$47,$CE,$9D,$4F,$86,$1F,$9A,$D2,$3E,$33,$27,$F2
	.dc.b	$4E,$7F,$F5,$1F,$8B,$F9,$22,$46,$7A,$3F,$C4,$3C,$8F,$AD,$7B,$7F
	.dc.b	$CA,$7E,$C7,$F0,$51,$8B,$46,$27,$51,$F8,$49,$D4,$FE,$0F,$ED,$3F
	.dc.b	$C5,$37,$F8,$E6,$F4,$1A,$63,$EC,$7F,$9C,$7A,$7F,$8B,$6E,$3F,$90
	.dc.b	$A4,$62,$3F,$9F,$5F,$E6,$A3,$11,$FD,$07,$BF,$FA,$66,$BF,$02,$3F
	.dc.b	$A2,$1F,$D3,$36,$F8,$AF,$C3,$8F,$7F,$C1,$F2,$C3,$E2,$37,$53,$4E
	.dc.b	$F5,$CF,$7F,$F2,$2E,$3E,$08,$E8,$43,$CF,$4F,$84,$F9,$31,$86,$BF
	.dc.b	$D0,$A3,$4F,$D0,$3E,$74,$6B,$0F,$DC,$E3,$FD,$7F,$F3,$DF,$D9,$36
	.dc.b	$BE,$43,$5F,$8F,$AA,$31,$1F,$D4,$21,$F2,$DD,$5B,$1E,$13,$8C,$64
	.dc.b	$F3,$BF,$06,$7E,$73,$F4,$75,$F9,$AF,$D0,$CF,$E2,$3F,$0B,$6F,$56
	.dc.b	$9F,$72,$3B,$1F,$06,$32,$78,$FE,$E7,$FB,$E7,$D9,$FF,$0F,$6C,$FE
	.dc.b	$15,$AF,$B9,$AF,$8F,$7F,$AB,$8F,$92,$31,$97,$A2,$7F,$29,$E3,$7D
	.dc.b	$26,$CF,$91,$BC,$9F,$E9,$26,$C7,$E5,$7F,$02,$FF,$08,$7E,$7D,$FD
	.dc.b	$6A,$3E,$64,$63,$8B,$50,$9D,$47,$F3,$29,$D4,$FC,$7C,$BC,$CF,$CD
	.dc.b	$3F,$14,$7A,$0F,$D2,$7D,$DF,$F8,$7F,$F2,$5E,$7C,$68,$C6,$1A,$A5
	.dc.b	$6A,$73,$1F,$AA,$D9,$F5,$7F,$B2,$BF,$E7,$F1,$89,$74,$7A,$23,$0D
	.dc.b	$AE,$13,$F8,$EF,$C1,$3F,$85,$5F,$EE,$5B,$19,$3F,$85,$C6,$5A,$7C
	.dc.b	$1D,$BE,$B9,$38,$7F,$1B,$2C,$7C,$9C,$1E,$1B,$E2,$7F,$8A,$A3,$33
	.dc.b	$EA,$C7,$C3,$3F,$24,$ED,$7F,$98,$7E,$71,$83,$E7,$98,$38,$1F,$E4
	.dc.b	$DF,$65,$F1,$AF,$6F,$FC,$A1,$DD,$FA,$DA,$D4,$E5,$FE,$5B,$7F,$ED
	.dc.b	$86,$9A,$AE,$31,$9A,$E5,$2B,$EF,$5A,$8E,$07,$23,$57,$DE,$B8,$F4
	.dc.b	$FF,$9B,$67,$FC,$26,$43,$FA,$11,$4F,$A0,$53,$F5,$07,$F8,$0F,$3F
	.dc.b	$C1,$7A,$BF,$3C,$9A,$7C,$D9,$FF,$21,$3D,$18,$74,$29,$FC,$3F,$7F
	.dc.b	$F2,$54,$7E,$06,$7F,$33,$18,$9F,$C3,$47,$DD,$FA,$E9,$5D,$DB,$3F
	.dc.b	$C7,$38,$BF,$D0,$1F,$9F,$FE,$76,$0D,$25,$78,$A3,$F9,$C1,$EF,$29
	.dc.b	$FB,$6F,$CE,$BF,$57,$D1,$FE,$8B,$6C,$1A,$7E,$FF,$93,$F8,$3E,$5F
	.dc.b	$C1,$7C,$2F,$F2,$5B,$7A,$FF,$CD,$17,$FD,$4A,$B6,$F8,$F7,$FA,$CD
	.dc.b	$EA,$4F,$FA,$8F,$3F,$F1,$BF,$07,$FA,$77,$ED,$FF,$91,$59,$A4,$7F
	.dc.b	$BA,$77,$78,$1F,$E6,$07,$F9,$DC,$C9,$4C,$E7,$F0,$F2,$F1,$C7,$D9
	.dc.b	$21,$FD,$1F,$C0,$E8,$D3,$91,$67,$FC,$89,$1A,$C7,$FA,$BC,$BF,$40
	.dc.b	$FE,$AE,$D3,$FA,$B9,$4F,$EA,$FD,$3F,$EB,$B8,$FF,$09,$F8,$7C,$FC
	.dc.b	$2F,$EC,$B8,$E2,$FC,$38,$7F,$86,$EB,$CB,$1F,$70,$F4,$94,$FC,$E9
	.dc.b	$FE,$19,$F1,$3F,$D9,$1C,$FF,$E0,$28,$34,$8F,$D5,$E0,$FF,$6B,$3C
	.dc.b	$CE,$97,$FE,$13,$78,$9F,$E1,$9D,$BA,$B4,$C9,$FD,$D1,$CB,$FD,$DB
	.dc.b	$8F,$AC,$FF,$8A,$63,$F0,$5E,$0F,$F8,$A5,$B3,$AD,$3F,$C6,$DA,$7F
	.dc.b	$8D,$67,$E1,$C6,$1F,$F1,$B7,$57,$C4,$E1,$FB,$B1,$F0,$8F,$8A,$E1
	.dc.b	$79,$7F,$E6,$35,$FA,$0F,$F9,$4C,$8F,$F9,$45,$FF,$2D,$2F,$8C,$A4
	.dc.b	$7F,$CB,$9E,$5F,$FB,$D6,$3F,$CB,$9F,$3F,$27,$FC,$B4,$D7,$FD,$BE
	.dc.b	$7E,$88,$FF,$51,$70,$7F,$A9,$C8,$FF,$8F,$07,$A1,$FE,$83,$F8,$CF
	.dc.b	$F4,$0F,$A6,$94,$61,$A7,$19,$E0,$FF,$A6,$23,$A3,$FE,$94,$F5,$39
	.dc.b	$46,$3D,$5F,$F4,$B4,$F2,$3F,$D3,$B1,$FE,$A3,$F5,$B7,$1F,$E5,$96
	.dc.b	$BF,$E3,$26,$BF,$E3,$16,$BF,$EB,$7F,$63,$D1,$F1,$9E,$2F,$67,$FD
	.dc.b	$72,$D8,$CC,$FC,$19,$CB,$FE,$C7,$9F,$F6,$61,$1C,$1F,$E8,$17,$FD
	.dc.b	$AA,$47,$07,$FB,$74,$FE,$B0,$7F,$9A,$3D,$BF,$F7,$2F,$97,$FE,$E3
	.dc.b	$F1,$7E,$63,$67,$67,$CF,$F7,$BF,$DE,$11,$FE,$EE,$8C,$9F,$EF,$CF
	.dc.b	$7B,$8B,$FE,$FA,$6B,$FE,$FA,$ED,$AF,$36,$AA,$46,$1B,$1D,$BD,$FF
	.dc.b	$C4,$5F,$DA,$6C,$7E,$1D,$AF,$FA,$11,$AF,$E1,$FB,$1F,$05,$E0,$6C
	.dc.b	$7A,$9B,$5F,$86,$6C,$7E,$4F,$E0,$7F,$D1,$F3,$F8,$A6,$C6,$4F,$5B
	.dc.b	$87,$FD,$F5,$1F,$75,$F8,$FB,$7F,$BE,$23,$07,$E5,$3E,$0B,$DE,$DC
	.dc.b	$FE,$2C,$62,$BF,$48,$D7,$FD,$43,$F0,$7E,$66,$E5,$F3,$23,$11,$FA
	.dc.b	$76,$BF,$A1,$3E,$CF,$27,$E8,$30,$FD,$17,$2F,$D4,$8C,$4F,$EC,$DF
	.dc.b	$96,$FF,$BF,$3A,$34,$47,$F0,$7D,$CF,$33,$FA,$FE,$1F,$B0,$8C,$56
	.dc.b	$A7,$35,$FE,$D9,$AF,$DA,$FC,$2B,$E1,$7F,$99,$CA,$30,$FF,$36,$6B
	.dc.b	$C2,$37,$18,$8F,$E1,$4D,$7F,$D8,$ED,$7F,$85,$35,$ED,$1C,$88,$F7
	.dc.b	$6B,$DC,$C1,$0C,$10,$F0,$BC,$1A,$6C,$A6,$CF,$2B,$B6,$9B,$29,$B3
	.dc.b	$D2,$ED,$A6,$CA,$6C,$F6,$99,$29,$98,$F8,$44,$8E,$91,$A3,$CE,$43
	.dc.b	$C6,$53,$B9,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$23,$C8,$FA,$5A,$FA
	.dc.b	$63,$CF,$18,$8F,$13,$5E,$ED,$7B,$35,$ED,$1C,$88,$F7,$6B,$DC,$C1
	.dc.b	$0C,$10,$F0,$BC,$1A,$6C,$A6,$CF,$2B,$B6,$9B,$29,$B3,$D2,$ED,$A6
	.dc.b	$CA,$6C,$F6,$99,$29,$98,$F8,$44,$8E,$91,$A3,$CE,$7D,$DF,$57,$FD
	.dc.b	$32,$FF,$6C,$6B,$FD,$E5,$AF,$FA,$94,$FD,$9F,$D9,$FF,$79,$F3,$7F
	.dc.b	$7E,$D8,$C9,$FE,$A6,$3F,$6A,$E5,$FD,$CB,$5C,$9B,$3F,$83,$39,$69
	.dc.b	$9B,$46,$5B,$19,$8C,$11,$C1,$27,$CE,$D3,$CE,$E0,$8E,$08,$7A,$0E
	.dc.b	$05,$76,$D3,$67,$B1,$CB,$4C,$9E,$C7,$2D,$33,$3E,$C8,$CB,$63,$33
	.dc.b	$ED,$9C,$B6,$32,$79,$8F,$01,$E5,$21,$82,$1E,$13,$25,$33,$3E,$17
	.dc.b	$2D,$73,$1E,$17,$2D,$72,$78,$CE,$84,$8E,$4D,$8E,$0D,$38,$46,$89
	.dc.b	$1A,$6C,$6A,$30,$47,$04,$9F,$8E,$D3,$E3,$B8,$23,$82,$1E,$C3,$81
	.dc.b	$5D,$B4,$D9,$F0,$1C,$B4,$C9,$ED,$72,$D3,$33,$EC,$8C,$B6,$33,$3E
	.dc.b	$D9,$CB,$63,$FC,$E1,$9F,$F7,$6F,$86,$FE,$F1,$9F,$FB,$61,$BA,$1A
	.dc.b	$7F,$9B,$3C,$57,$C4,$7F,$9D,$63,$0D,$C3,$FC,$F9,$3A,$79,$18,$8D
	.dc.b	$3F,$E7,$D3,$FC,$2D,$3F,$E1,$47,$66,$77,$FF,$3B,$C6,$4F,$F3,$EF
	.dc.b	$0F,$F8,$F1,$BC,$5D,$2B,$FC,$44,$D5,$3B,$7F,$C3,$DF,$15,$9B,$34
	.dc.b	$91,$E2,$3F,$D2,$5C,$CE,$FE,$8D,$8D,$1F,$E8,$12,$B9,$C7,$FC,$81
	.dc.b	$D9,$E3,$E7,$34,$95,$CB,$AB,$CC,$FC,$56,$BF,$E5,$DD,$3D,$91,$88
	.dc.b	$75,$3E,$2F,$E0,$36,$3C,$AD,$C7,$FC,$B9,$3F,$17,$B7,$FC,$E3,$1F
	.dc.b	$EA,$86,$D7,$FA,$D5,$AF,$F9,$F4,$F7,$F7,$FC,$C6,$F0,$7C,$F8,$C5
	.dc.b	$7F,$AC,$9B,$3F,$EB,$FF,$75,$F1,$BF,$55,$B1,$F2,$9B,$5F,$EB,$A6
	.dc.b	$C7,$E6,$39,$9C,$CF,$8A,$FC,$B9,$27,$51,$97,$F2,$05,$3F,$21,$CF
	.dc.b	$F9,$59,$FC,$CC,$64,$FF,$68,$46,$5F,$F5,$B7,$27,$FB,$4E,$7F,$CF
	.dc.b	$C4,$8C,$9C,$47,$1F,$DE,$70,$D8,$FC,$1C,$63,$8F,$A3,$F1,$7E,$27
	.dc.b	$FD,$C1,$F1,$5F,$05,$B1,$91,$19,$69,$FE,$EB,$E0,$FF,$4F,$C7,$E9
	.dc.b	$23,$27,$FA,$20,$A7,$E9,$8B,$E1,$3D,$17,$E1,$DE,$09,$FF,$79,$4E
	.dc.b	$63,$11,$FE,$BA,$9D,$BC,$9B,$5F,$EF,$2F,$2B,$FB,$8F,$C5,$DF,$03
	.dc.b	$FD,$F8,$DB,$34,$FC,$07,$46,$A3,$F1,$0D,$7F,$84,$72,$7E,$76,$77
	.dc.b	$1C,$8F,$E1,$82,$FF,$C0,$3F,$11,$DC,$91,$A8,$CB,$D8,$47,$42,$39
	.dc.b	$7F,$92,$F8,$52,$3F,$1E,$D7,$E3,$E6,$FC,$2D,$0F,$C3,$BC,$BB,$FC
	.dc.b	$F9,$CC,$62,$B8,$56,$CF,$E4,$D5,$B7,$07,$6F,$C3,$7F,$C4,$D1,$F6
	.dc.b	$5B,$26,$98,$FB,$DF,$1E,$BD,$DB,$7D,$A9,$18,$8F,$5B,$5F,$5F,$8D
	.dc.b	$FE,$BE,$CF,$E0,$46,$3A,$FF,$6B,$18,$CB,$F1,$9F,$0F,$C5,$B4,$CB
	.dc.b	$E0,$4E,$72,$3D,$E5,$8F,$77,$E7,$90,$FE,$94,$E5,$FC,$BF,$D2,$7F
	.dc.b	$06,$7E,$38,$C1,$19,$D7,$E9,$B6,$7E,$7B,$5E,$7E,$6F,$DA,$FC,$8F
	.dc.b	$6D,$8D,$B5,$D8,$8E,$BE,$47,$E0,$BE,$2A,$C9,$A4,$7D,$87,$CB,$F0
	.dc.b	$7F,$CC,$19,$FB,$63,$1B,$FA,$3E,$8B,$A4,$7A,$1A,$FC,$26,$BF,$07
	.dc.b	$34,$63,$57,$D0,$BE,$6B,$F2,$FF,$8F,$FE,$5A,$31,$82,$31,$3A,$8F
	.dc.b	$CA,$CE,$A7,$F0,$71,$F6,$39,$3F,$06,$FB,$0C,$69,$FE,$81,$3F,$C1
	.dc.b	$7E,$BF,$E8,$B1,$FF,$02,$46,$27,$F5,$11,$8F,$41,$5B,$9F,$D6,$D6
	.dc.b	$F8,$FE,$F4,$7F,$18,$C9,$F3,$9F,$B2,$FA,$0C,$CF,$E0,$E7,$2A,$33
	.dc.b	$97,$C4,$AE,$0F,$B9,$B3,$F9,$0C,$FD,$AD,$C7,$EB,$C6,$23,$EB,$9F
	.dc.b	$8C,$F9,$8F,$E0,$7F,$A1,$76,$E1,$46,$31,$D5,$2B,$53,$98,$FC,$EF
	.dc.b	$E1,$1F,$BA,$AF,$CB,$B6,$32,$7E,$B6,$32,$D3,$F9,$DF,$C6,$BE,$1F
	.dc.b	$F2,$93,$FC,$3A,$31,$3F,$95,$8C,$4E,$A3,$8B,$5F,$69,$F0,$79,$7F
	.dc.b	$03,$CD,$F9,$7A,$FD,$7E,$1B,$6D,$75,$3C,$FC,$FF,$0E,$7A,$47,$E1
	.dc.b	$AF,$9E,$BF,$2E,$30,$7E,$2F,$0F,$D2,$38,$3C,$47,$FB,$57,$E4,$DE
	.dc.b	$E7,$F3,$2F,$86,$61,$FE,$51,$F7,$3F,$DB,$B1,$8B,$FE,$6F,$18,$9D
	.dc.b	$5F,$29,$56,$D5,$6D,$E3,$3B,$9D,$4E,$67,$73,$9A,$FC,$44,$7F,$9C
	.dc.b	$08,$7F,$86,$8A,$68,$C1,$0F,$BF,$EC,$7C,$94,$D2,$6B,$B8,$FE,$18
	.dc.b	$74,$29,$F7,$8E,$2D,$8F,$D1,$EF,$F8,$46,$EF,$FD,$71,$F0,$7C,$4C
	.dc.b	$AF,$C7,$B6,$FF,$99,$F0,$FF,$62,$FD,$0F,$F3,$70,$69,$2B,$CD,$1F
	.dc.b	$C1,$8F,$31,$4F,$D9,$0F,$A8,$FD,$BF,$4F,$EE,$9B,$D1,$FD,$C1,$B3
	.dc.b	$BA,$FE,$15,$D7,$FD,$D3,$DA,$FE,$07,$D5,$FC,$3D,$BD,$BF,$C3,$AD
	.dc.b	$18,$AE,$17,$E6,$E6,$73,$FF,$4F,$BC,$35,$FE,$25,$DF,$FE,$02,$FC
	.dc.b	$7F,$F9,$4C,$1A,$47,$F9,$D1,$F6,$8B,$C5,$FC,$FD,$C1,$FC,$F4,$A7
	.dc.b	$DA,$3F,$0C,$43,$07,$F4,$1F,$0B,$EC,$3F,$D7,$9A,$FF,$68,$6B,$B3
	.dc.b	$AE,$3F,$DA,$3C,$4F,$F0,$DF,$E8,$9F,$D8,$A4,$D2,$3F,$DA,$B7,$F9
	.dc.b	$CD,$3E,$59,$76,$FE,$E3,$1A,$8F,$41,$F7,$CB,$AF,$CE,$8F,$F1,$EC
	.dc.b	$7F,$94,$7C,$6F,$F1,$6F,$37,$F8,$2A,$0D,$23,$F5,$27,$8B,$D4,$F8
	.dc.b	$B9,$7F,$98,$7C,$9E,$71,$FE,$23,$6B,$FE,$22,$3A,$99,$29,$FE,$33
	.dc.b	$3C,$B1,$FE,$33,$6F,$91,$FE,$33,$6F,$95,$E2,$9F,$F1,$E6,$7F,$E4
	.dc.b	$1F,$2F,$E1,$57,$F9,$B5,$B3,$A8,$F5,$1F,$A4,$D3,$E2,$3F,$E6,$2F
	.dc.b	$8B,$8F,$F9,$6D,$AF,$FB,$BF,$CE,$7A,$3F,$E5,$49,$FF,$39,$12,$7F
	.dc.b	$CC,$91,$97,$FC,$C9,$F1,$3F,$E6,$E2,$38,$36,$1F,$F3,$E1,$1C,$BF
	.dc.b	$E5,$DF,$53,$FC,$C7,$F1,$16,$4D,$23,$FE,$9B,$F8,$1D,$B4,$D9,$63
	.dc.b	$FD,$3C,$7F,$A8,$C8,$FF,$A9,$DA,$FF,$9C,$8F,$F5,$AF,$B3,$FE,$74
	.dc.b	$F8,$8F,$F5,$99,$23,$FD,$13,$EE,$7F,$AE,$63,$F1,$FF,$07,$FE,$FF
	.dc.b	$6F,$BD,$C6,$BF,$D8,$31,$84,$FF,$B3,$08,$FF,$B4,$9A,$FF,$81,$DF
	.dc.b	$CD,$EF,$F9,$D6,$F7,$FF,$B7,$9B,$CF,$D2,$BF,$DD,$5F,$03,$FE,$8C
	.dc.b	$F3,$BF,$DC,$CD,$8C,$9F,$D5,$CF,$F5,$3B,$97,$E1,$7C,$17,$FB,$E6
	.dc.b	$7F,$DE,$DE,$F6,$67,$FD,$2F,$39,$7F,$D5,$6D,$7E,$57,$47,$F8,$BA
	.dc.b	$BF,$66,$D8,$FF,$58,$B5,$DF,$3F,$55,$F0,$6F,$7B,$0D,$8F,$41,$FE
	.dc.b	$A8,$71,$7F,$C0,$1B,$1F,$EB,$EF,$0D,$CB,$D5,$B1,$C8,$FE,$C7,$19
	.dc.b	$3C,$5C,$7F,$EF,$5B,$FF,$79,$92,$7F,$03,$C9,$E7,$E4,$72,$52,$4F
	.dc.b	$58,$CB,$DF,$E1,$1E,$27,$BD,$B1,$98,$C1,$F0,$F8,$7C,$F8,$FF,$BD
	.dc.b	$C9,$3F,$9F,$7F,$8F,$D3,$C4,$E8,$49,$F0,$3F,$37,$E0,$DE,$4D,$24
	.dc.b	$FC,$38,$CB,$F1,$7A,$7F,$96,$61,$F8,$E2,$47,$FB,$AA,$32,$FF,$BB
	.dc.b	$3E,$17,$F8,$4B,$FC,$A9,$AF,$F3,$46,$BB,$EE,$FE,$5E,$A3,$2F,$EC
	.dc.b	$DA,$FF,$B9,$9A,$EE,$38,$C6,$08,$E0,$91,$E3,$6B,$E3,$6B,$E3,$34
	.dc.b	$43,$45,$34,$60,$86,$08,$7A,$5E,$AD,$39,$14,$E2,$53,$89,$EF,$78
	.dc.b	$B4,$E2,$53,$89,$4E,$27,$C5,$36,$53,$65,$37,$1F,$21,$DB,$5D,$B5
	.dc.b	$DC,$79,$1E,$47,$A4,$87,$8C,$A6,$8C,$10,$C1,$23,$B3,$C1,$AE,$DA
	.dc.b	$EE,$3C,$51,$82,$38,$24,$7B,$1A,$FB,$1A,$FB,$0D,$10,$D1,$4D,$18
	.dc.b	$21,$82,$1E,$97,$AB,$4E,$45,$38,$94,$E2,$7B,$DE,$2D,$38,$94,$E2
	.dc.b	$53,$89,$F1,$4F,$F0,$DF,$A8,$F5,$3F,$90,$47,$FA,$3B,$0C,$C6,$27
	.dc.b	$51,$FB,$9F,$3D,$5F,$E9,$B3,$F6,$4E,$3D,$4F,$DE,$36,$3F,$81,$35
	.dc.b	$FD,$D3,$5E,$27,$33,$F8,$03,$96,$99,$C5,$19,$6C,$66,$30,$47,$04
	.dc.b	$9F,$2B,$4F,$29,$4F,$29,$4F,$2B,$C0,$8F,$06,$BC,$1A,$F0,$34,$43
	.dc.b	$45,$34,$60,$86,$08,$7C,$27,$AB,$4E,$45,$38,$94,$E3,$3F,$1A,$38
	.dc.b	$B6,$38,$B5,$E2,$D7,$8C,$FC,$B9,$E2,$D8,$E2,$D7,$8B,$5E,$27,$CB
	.dc.b	$3C,$84,$9E,$CD,$8E,$0D,$38,$46,$89,$1A,$6C,$6A,$30,$47,$04,$9F
	.dc.b	$4B,$4F,$49,$4F,$49,$4F,$4B,$C0,$8F,$06,$BC,$1A,$F0,$34,$43,$45
	.dc.b	$34,$60,$86,$08,$7C,$37,$AB,$4E,$47,$CC,$CB,$FD,$41,$F0,$BF,$E4
	.dc.b	$22,$47,$F7,$5F,$01,$DE,$9F,$F0,$3B,$FB,$C6,$CE,$4F,$F3,$10,$E1
	.dc.b	$E2,$E4,$FF,$33,$38,$8F,$F0,$8F,$0F,$39,$E3,$3F,$E1,$37,$81,$FC
	.dc.b	$29,$C4,$F0,$8C,$8E,$27,$C4,$73,$8F,$F3,$A4,$61,$DB,$67,$FD,$7E
	.dc.b	$D8,$FF,$3A,$F6,$F1,$7F,$8C,$F4,$A3,$10,$8C,$3E,$26,$D7,$F9,$DC
	.dc.b	$91,$F1,$39,$FF,$C7,$71,$88,$FF,$44,$76,$7F,$A2,$63,$FC,$A5,$1D
	.dc.b	$A3,$0F,$CC,$6B,$FE,$8F,$EC,$FF,$29,$CF,$FB,$61,$B1,$FE,$59,$EC
	.dc.b	$F3,$C6,$23,$FD,$2D,$F1,$1F,$45,$FA,$ED,$73,$BF,$BA,$31,$1F,$CE
	.dc.b	$BB,$FD,$AF,$F9,$CB,$4F,$87,$18,$8F,$F3,$99,$F5,$3C,$07,$A7,$F2
	.dc.b	$1B,$19,$8C,$19,$E0,$91,$C5,$BA,$7F,$A0,$BE,$3B,$B9,$27,$F0,$38
	.dc.b	$23,$EA,$F1,$7E,$43,$E2,$3F,$2A,$4A,$FC,$6E,$BF,$D5,$23,$F1,$44
	.dc.b	$8F,$C8,$F2,$65,$A2,$3E,$AB,$4F,$A8,$6C,$FB,$65,$3F,$D1,$C5,$3F
	.dc.b	$D0,$E7,$13,$ED,$70,$67,$83,$FD,$1B,$3F,$6E,$39,$6E,$F7,$9F,$87
	.dc.b	$FF,$6C,$F8,$8F,$D7,$CE,$63,$18,$FF,$B4,$E3,$0A,$77,$5C,$A3,$07
	.dc.b	$E8,$0A,$7F,$B6,$F9,$BF,$3B,$3F,$BC,$8F,$D0,$57,$67,$4D,$7F,$BF
	.dc.b	$E4,$FA,$DC,$42,$30,$F6,$37,$1A,$8C,$D7,$9B,$E2,$FF,$C1,$3F,$15
	.dc.b	$6C,$D2,$3F,$C4,$3E,$25,$F5,$DC,$23,$0D,$7F,$DE,$D1,$A7,$99,$A2
	.dc.b	$9F,$E1,$8C,$BE,$85,$FF,$86,$BE,$2F,$C4,$49,$F9,$DA,$A7,$51,$FC
	.dc.b	$8B,$D3,$FF,$4D,$FD,$AF,$C9,$6C,$67,$C8,$FF,$12,$CE,$5B,$1F,$CC
	.dc.b	$79,$7F,$C3,$DE,$13,$C2,$FE,$67,$E1,$FF,$BE,$09,$5F,$7F,$F0,$B7
	.dc.b	$B3,$FE,$30,$E0,$FC,$2C,$62,$51,$89,$D4,$71,$6B,$F0,$76,$FF,$59
	.dc.b	$DF,$E2,$33,$FF,$19,$70,$7E,$4D,$B3,$AA,$F8,$7E,$57,$F9,$03,$E1
	.dc.b	$FF,$1E,$D8,$C8,$8C,$B4,$F8,$18,$7B,$1F,$F2,$48,$D7,$E1,$6D,$F1
	.dc.b	$36,$F8,$73,$F9,$FF,$BF,$FE,$91,$87,$E9,$1B,$19,$ED,$F1,$3F,$76
	.dc.b	$32,$F6,$C3,$AE,$3E,$18,$FF,$2C,$C6,$1F,$29,$DF,$CF,$F8,$EF,$F4
	.dc.b	$5A,$FF,$79,$F0,$78,$67,$3B,$A7,$35,$F8,$8F,$90,$FF,$32,$FE,$8D
	.dc.b	$FB,$46,$C6,$44,$65,$A7,$E4,$3E,$EB,$AF,$09,$A4,$8F,$BE,$7A,$0F
	.dc.b	$76,$FF,$21,$BC,$1F,$2C,$FF,$3B,$65,$E7,$F8,$BF,$4E,$66,$92,$BA
	.dc.b	$9F,$87,$F3,$BF,$CF,$FF,$A3,$FF,$64,$8F,$ED,$19,$FF,$9C,$23,$18
	.dc.b	$7F,$1B,$AD,$BF,$AA,$6D,$79,$38,$7E,$93,$F8,$1C,$FD,$0F,$D4,$6E
	.dc.b	$6F,$A7,$57,$E8,$70,$FC,$4C,$E6,$31,$7F,$E9,$48,$C3,$FA,$49,$DD
	.dc.b	$72,$8C,$1E,$D1,$FA,$47,$AB,$3F,$F4,$5B,$72,$FF,$4F,$B6,$77,$7F
	.dc.b	$45,$F4,$61,$F3,$F1,$FC,$74,$BE,$A3,$FD,$2B,$F9,$2F,$C6,$6C,$7F
	.dc.b	$3B,$6D,$FF,$15,$6B,$FE,$15,$FB,$F7,$B5,$FB,$1A,$FE,$85,$18,$9F
	.dc.b	$D8,$46,$27,$51,$FC,$7E,$75,$D3,$ED,$FC,$1F,$E3,$CF,$DA,$6D,$FF
	.dc.b	$53,$6E,$C7,$73,$F5,$F4,$F7,$3F,$04,$C6,$9F,$D6,$31,$7D,$9F,$46
	.dc.b	$CF,$D0,$3D,$91,$93,$AF,$57,$EC,$E3,$F0,$C7,$57,$0F,$CE,$AF,$C4
	.dc.b	$C6,$23,$C2,$7F,$95,$BD,$AF,$F6,$BA,$F9,$3F,$C0,$3F,$ED,$D3,$15
	.dc.b	$FC,$4A,$77,$3A,$6C,$67,$B9,$ED,$FF,$B8,$2B,$51,$8B,$E6,$7F,$B9
	.dc.b	$2F,$8B,$63,$53,$E2,$C3,$F8,$FC,$FF,$71,$6B,$FD,$14,$C7,$37,$F9
	.dc.b	$B9,$C1,$0F,$DE,$9A,$D3,$FB,$D6,$ED,$7E,$7F,$C4,$FE,$D1,$B1,$F8
	.dc.b	$8D,$BF,$DF,$33,$96,$C7,$5B,$FC,$D9,$D4,$87,$F6,$4D,$F6,$FF,$4B
	.dc.b	$6B,$F8,$B1,$5F,$8F,$6C,$FE,$55,$FF,$04,$37,$CB,$F3,$71,$FF,$83
	.dc.b	$A3,$07,$E9,$79,$D4,$FC,$97,$F5,$51,$86,$C6,$67,$83,$4F,$8E,$7B
	.dc.b	$4F,$CD,$1E,$C3,$F8,$17,$E9,$3F,$BD,$24,$7F,$19,$84,$7F,$21,$D7
	.dc.b	$F8,$6E,$57,$AB,$F8,$63,$79,$7E,$1C,$EB,$0F,$07,$0F,$DE,$C7,$C6
	.dc.b	$43,$F9,$5E,$7E,$EF,$D2,$BF,$87,$12,$BC,$4F,$F2,$EE,$EF,$E7,$15
	.dc.b	$F3,$8F,$EC,$DD,$DF,$71,$A6,$4F,$EC,$8E,$5F,$01,$F0,$8F,$2F,$27
	.dc.b	$E3,$DF,$F0,$37,$7E,$DF,$E7,$ED,$C3,$FC,$B0,$DE,$AF,$EA,$D3,$DC
	.dc.b	$FE,$8B,$E0,$FD,$8F,$D2,$FF,$AF,$12,$3F,$B1,$3F,$DA,$DE,$17,$FD
	.dc.b	$AF,$C3,$ED,$99,$A7,$8E,$7D,$07,$AB,$6F,$98,$FE,$60,$FF,$02,$1F
	.dc.b	$E5,$52,$F3,$7F,$82,$7C,$5F,$F2,$DC,$60,$F8,$3C,$3F,$E6,$37,$A9
	.dc.b	$CC,$FF,$14,$76,$70,$69,$D4,$B5,$FE,$20,$3F,$C5,$8F,$ED,$35,$78
	.dc.b	$DD,$74,$FF,$17,$4E,$63,$1E,$47,$F8,$EF,$9B,$FC,$75,$3F,$12,$30
	.dc.b	$FF,$8E,$CF,$9C,$79,$89,$A7,$C2,$71,$E5,$78,$17,$DF,$7F,$94,$DA
	.dc.b	$7F,$95,$CA,$7F,$8D,$23,$84,$7F,$9D,$7E,$2E,$4F,$F9,$93,$CD,$FD
	.dc.b	$9C,$62,$3F,$CE,$3E,$7F,$ED,$BE,$2F,$FD,$19,$E6,$FF,$9C,$DB,$19
	.dc.b	$9F,$F3,$A4,$7F,$A0,$48,$F5,$0E,$1F,$F7,$27,$A7,$FC,$0B,$E2,$BF
	.dc.b	$DC,$8D,$8F,$F2,$59,$FE,$93,$70,$F0,$11,$FE,$9F,$F5,$7F,$DF,$AD
	.dc.b	$E1,$DC,$FF,$0E,$9C,$BF,$EA,$99,$FF,$56,$FA,$DA,$AF,$F5,$8C,$62
	.dc.b	$7A,$9F,$E8,$89,$E6,$FF,$B1,$3E,$2B,$FD,$80,$9A,$47,$FD,$9E,$64
	.dc.b	$AF,$FB,$50,$8F,$FB,$61,$FF,$5C,$7B,$4E,$9F,$E4,$B1,$86,$CE,$CF
	.dc.b	$F7,$29,$D4,$FF,$38,$CB,$A7,$71,$86,$BF,$CC,$23,$4F,$13,$F9,$9F
	.dc.b	$B9,$FE,$F5,$7F,$B1,$1F,$E7,$C7,$1E,$77,$FB,$F6,$32,$70,$F7,$3F
	.dc.b	$9D,$C7,$FB,$47,$E0,$1E,$FE,$2B,$ED,$C6,$1E,$21,$C4,$F1,$9E,$13
	.dc.b	$FE,$8C,$97,$C0,$62,$3F,$A4,$C6,$1E,$6D,$AF,$14,$EE,$58,$37,$C3
	.dc.b	$63,$C6,$DA,$FF,$75,$36,$3F,$27,$F0,$3F,$22,$7F,$DD,$CD,$8C,$E9
	.dc.b	$DE,$31,$1F,$EA,$4E,$F7,$BF,$E3,$6C,$7C,$26,$DF,$E3,$0F,$F7,$D7
	.dc.b	$C1,$7F,$AB,$1B,$3E,$B8,$C3,$F8,$E8,$FF,$80,$1F,$07,$FE,$49,$6F
	.dc.b	$A3,$F0,$63,$13,$FA,$C6,$BF,$EB,$8E,$6F,$73,$75,$F9,$11,$88,$FD
	.dc.b	$54,$7C,$BF,$84,$FA,$F1,$FB,$66,$BF,$EC,$26,$BF,$DF,$9A,$B8,$FF
	.dc.b	$3C,$FE,$09,$B8,$FE,$7E,$31,$C9,$F6,$67,$E5,$C7,$E6,$77,$C1,$1F
	.dc.b	$C5,$F8,$1F,$ED,$0B,$FF,$15,$12,$3F,$DE,$CF,$F0,$C5,$F0,$AF,$3F
	.dc.b	$FA,$34,$61,$B1,$98,$FE,$94,$D7,$93,$5F,$E2,$ED,$7F,$DC,$6D,$7F
	.dc.b	$C3,$ED,$7B,$1D,$23,$B9,$1C,$10,$EE,$E5,$A6,$4F,$0B,$96,$99,$3C
	.dc.b	$6E,$5A,$64,$F2,$B9,$69,$93,$CE,$78,$08,$76,$29,$D0,$A7,$22,$9C
	.dc.b	$8E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$48,$FA,$44
	.dc.b	$8F,$81,$19,$6C,$66,$3E,$13,$5F,$5B,$5F,$33,$5F,$1B,$5F,$13,$5E
	.dc.b	$C7,$48,$EE,$47,$04,$3B,$B9,$69,$93,$C2,$E5,$A6,$4F,$1B,$96,$99
	.dc.b	$3C,$AE,$5A,$64,$F3,$9E,$02,$1D,$8A,$74,$29,$C8,$A7,$23,$81,$0F
	.dc.b	$F5,$A4,$FF,$8D,$FD,$8F,$F5,$FC,$7F,$B0,$7C,$8F,$E8,$EF,$FA,$FC
	.dc.b	$FD,$7F,$AB,$FE,$8A,$F4,$BF,$66,$D8,$C9,$FE,$C6,$3F,$62,$E5,$FD
	.dc.b	$AB,$5C,$9B,$3F,$82,$B9,$69,$9C,$7F,$DD,$4D,$8D,$46,$27,$C7,$39
	.dc.b	$23,$82,$3E,$66,$BE,$43,$66,$48,$65,$DB,$E8,$6B,$E6,$6C,$78,$E3
	.dc.b	$89,$B2,$9B,$2B,$B7,$B1,$89,$F6,$91,$F2,$9A,$21,$D4,$A7,$52,$C7
	.dc.b	$57,$44,$34,$57,$4E,$08,$E0,$91,$E3,$9F,$7B,$67,$D0,$79,$8E,$8E
	.dc.b	$5F,$1B,$5C,$9B,$3B,$B9,$69,$99,$F0,$B6,$35,$18,$9F,$1C,$E4,$8E
	.dc.b	$08,$F9,$9A,$F9,$0D,$99,$21,$97,$6F,$A1,$AF,$99,$B1,$E3,$8E,$26
	.dc.b	$CA,$6C,$AE,$DE,$C6,$27,$DA,$47,$F4,$D9,$3D,$0F,$F2,$87,$C2,$FF
	.dc.b	$6D,$8C,$6A,$8C,$4E,$AB,$FC,$FB,$5A,$9F,$EE,$9F,$0C,$F8,$75,$71
	.dc.b	$3F,$D0,$71,$97,$06,$83,$99,$E7,$3C,$9F,$F0,$61,$27,$FD,$1B,$C1
	.dc.b	$FE,$12,$8C,$47,$F8,$67,$A7,$83,$FE,$1D,$6B,$93,$C2,$7F,$87,$5C
	.dc.b	$BF,$E2,$63,$23,$97,$FC,$9D,$3F,$E9,$58,$CC,$61,$FF,$1A,$73,$F2
	.dc.b	$5C,$E8,$C3,$63,$AB,$6B,$FD,$45,$F1,$1C,$6C,$D2,$57,$93,$AB,$FC
	.dc.b	$A3,$5F,$E9,$A2,$47,$FA,$C2,$32,$FF,$AC,$BA,$9F,$13,$E5,$6C,$7B
	.dc.b	$9B,$5F,$E5,$89,$F5,$76,$FB,$6B,$FD,$15,$D9,$82,$3F,$1B,$BD,$DB
	.dc.b	$F0,$5B,$19,$8C,$19,$F0,$24,$71,$6E,$3F,$E7,$8F,$A4,$F0,$92,$7E
	.dc.b	$9E,$28,$F9,$DE,$4B,$D9,$A4,$AF,$B3,$19,$7E,$DF,$27,$F9,$63,$E2
	.dc.b	$BF,$CA,$C4,$8F,$C5,$F8,$5F,$98,$8F,$F2,$76,$9F,$E7,$E6,$F3,$66
	.dc.b	$BF,$18,$23,$2D,$3E,$94,$6D,$FF,$45,$35,$FD,$2F,$17,$57,$67,$13
	.dc.b	$F4,$9C,$5F,$5A,$BE,$A4,$64,$FD,$6C,$65,$FB,$DC,$5D,$78,$9E,$9B
	.dc.b	$F0,$CF,$D9,$FC,$5F,$FB,$D2,$73,$18,$BF,$D9,$46,$14,$EE,$B9,$46
	.dc.b	$0F,$CF,$94,$FD,$D7,$47,$E7,$67,$FD,$33,$19,$3F,$81,$C6,$5E,$F2
	.dc.b	$FC,$37,$C5,$1F,$C2,$2F,$D3,$18,$8F,$F6,$04,$ED,$E9,$3A,$7B,$0D
	.dc.b	$3F,$01,$F1,$9F,$E8,$76,$C7,$E9,$78,$FF,$8A,$CE,$5B,$1F,$C4,$C7
	.dc.b	$1B,$F0,$4C,$C7,$F1,$78,$C2,$AF,$F4,$A4,$61,$B3,$C2,$7E,$FD,$7F
	.dc.b	$24,$F8,$DF,$04,$1A,$47,$E9,$FC,$E3,$C6,$FE,$57,$F1,$DF,$66,$31
	.dc.b	$28,$C4,$EA,$38,$B5,$FE,$5D,$E1,$FF,$3C,$8F,$E7,$7E,$2F,$F4,$2F
	.dc.b	$14,$D3,$FC,$61,$BF,$E2,$BE,$97,$F9,$DE,$FF,$83,$6E,$3F,$69,$AF
	.dc.b	$D9,$E1,$46,$23,$EC,$F8,$9F,$E3,$7B,$FC,$AB,$6B,$F9,$26,$5F,$8F
	.dc.b	$8C,$7E,$21,$FD,$2B,$E2,$BF,$24,$DC,$3F,$12,$D7,$C0,$D7,$FA,$8E
	.dc.b	$BE,$F7,$FA,$BE,$BE,$EC,$3E,$09,$23,$F2,$7F,$11,$C7,$3F,$D4,$B7
	.dc.b	$33,$4E,$D2,$C3,$B3,$F3,$B0,$EC,$F8,$36,$F9,$98,$FE,$AF,$F0,$8F
	.dc.b	$89,$97,$ED,$9B,$91,$A6,$BF,$79,$FE,$CD,$FA,$6F,$EE,$5B,$8F,$EE
	.dc.b	$1A,$FE,$A5,$AA,$D1,$FD,$03,$87,$97,$6F,$52,$57,$E8,$0F,$46,$5F
	.dc.b	$15,$F6,$67,$EA,$7C,$1E,$DF,$94,$8F,$13,$C4,$FA,$7A,$3F,$C0,$7F
	.dc.b	$99,$F8,$E7,$CD,$18,$CF,$D4,$D8,$C8,$9E,$0F,$DC,$DF,$E6,$BE,$4C
	.dc.b	$7C,$CE,$71,$FC,$CE,$8F,$C3,$FE,$51,$27,$CC,$7C,$58,$D3,$8C,$BE
	.dc.b	$3E,$5F,$E1,$AF,$D3,$BC,$31,$89,$46,$2B,$F8,$1B,$6B,$85,$7C,$AA
	.dc.b	$87,$9C,$A7,$87,$F0,$FF,$F1,$3F,$CB,$73,$6C,$FD,$D6,$CF,$DD,$6B
	.dc.b	$E9,$FB,$EF,$D3,$4F,$E0,$DB,$1F,$83,$6B,$B1,$D1,$EC,$F9,$27,$C2
	.dc.b	$7E,$FA,$FF,$A1,$C6,$2B,$F9,$1C,$62,$75,$1F,$CD,$E7,$5C,$9F,$5A
	.dc.b	$3F,$49,$C3,$F5,$9F,$A9,$C3,$F5,$9C,$F0,$FD,$6F,$99,$E7,$1F,$4D
	.dc.b	$B9,$7E,$61,$BC,$27,$73,$CA,$FA,$74,$FE,$CA,$3B,$9F,$38,$D1,$5D
	.dc.b	$2F,$69,$ED,$3D,$BF,$F2,$CF,$CA,$7F,$2E,$AD,$4E,$54,$66,$B8,$D7
	.dc.b	$0C,$3F,$DC,$75,$A6,$D6,$EF,$93,$6B,$85,$7F,$9A,$2F,$9B,$C4,$FC
	.dc.b	$E1,$FC,$E0,$87,$F0,$D2,$9F,$B6,$29,$C0,$FE,$0D,$8F,$F9,$D3,$E2
	.dc.b	$BF,$30,$DD,$33,$29,$CB,$63,$B3,$0E,$C5,$3F,$87,$70,$7E,$1E,$7F
	.dc.b	$19,$C3,$D6,$7F,$0C,$D7,$6D,$AF,$E2,$BC,$7F,$E2,$47,$F3,$53,$FD
	.dc.b	$ED,$BE,$47,$EB,$E7,$53,$FA,$31,$F1,$DF,$A9,$6D,$7E,$A5,$A6,$E3
	.dc.b	$F2,$DC,$1F,$E9,$E8,$F8,$E7,$E5,$0F,$93,$3F,$BF,$FD,$3F,$F7,$90
	.dc.b	$69,$23,$F9,$97,$5F,$CD,$9A,$7C,$B2,$E7,$FC,$E6,$35,$1F,$28,$87
	.dc.b	$D7,$2C,$FF,$0F,$FC,$D7,$C2,$47,$F9,$3C,$23,$F8,$97,$77,$48,$FE
	.dc.b	$38,$D9,$FE,$82,$DE,$53,$B9,$F8,$07,$DB,$29,$E2,$8F,$F4,$3E,$1F
	.dc.b	$D2,$23,$D8,$7F,$4F,$EF,$E8,$75,$7F,$D8,$3F,$A8,$7F,$57,$69,$FD
	.dc.b	$5C,$A7,$F5,$72,$9F,$E3,$EF,$0B,$D1,$E1,$FF,$63,$8E,$E6,$4B,$97
	.dc.b	$F6,$97,$27,$A0,$F9,$44,$30,$FF,$7C,$7F,$C8,$C7,$F8,$00,$A7,$F8
	.dc.b	$67,$C4,$FF,$3A,$B9,$E2,$FC,$2F,$07,$E1,$E3,$E3,$35,$FF,$08,$9C
	.dc.b	$8F,$F2,$BF,$8B,$C9,$CC,$FF,$86,$9D,$5F,$F8,$89,$FF,$28,$1F,$E2
	.dc.b	$62,$B9,$7F,$C4,$86,$30,$FF,$16,$BF,$09,$BA,$7F,$8D,$63,$1E,$3F
	.dc.b	$DD,$C7,$CD,$1F,$0D,$AF,$C4,$F2,$7F,$C9,$7F,$19,$FE,$4A,$83,$48
	.dc.b	$FF,$9A,$47,$C6,$7F,$96,$88,$E5,$FF,$30,$AF,$8C,$A4,$7F,$CD,$DE
	.dc.b	$1F,$F9,$AA,$3F,$CC,$64,$8E,$CD,$7F,$C9,$9E,$81,$7C,$5F,$FA,$56
	.dc.b	$30,$D8,$C9,$FE,$8E,$8C,$BF,$E9,$88,$FF,$4A,$91,$C1,$FE,$94,$F5
	.dc.b	$36,$FF,$96,$DA,$FF,$A4,$9B,$BF,$FA,$9E,$7F,$D6,$BF,$15,$2B,$FD
	.dc.b	$66,$1C,$3F,$EC,$23,$07,$FB,$01,$FF,$4B,$7B,$1E,$D6,$C6,$8F,$80
	.dc.b	$7F,$BB,$A3,$4F,$FB,$4A,$3F,$DA,$24,$7F,$DB,$67,$13,$E4,$7B,$6F
	.dc.b	$6F,$8C,$74,$8C,$3B,$0E,$27,$FD,$3B,$EA,$FD,$1F,$1A,$D7,$F6,$33
	.dc.b	$FD,$49,$19,$7E,$EB,$5F,$4F,$BB,$FE,$F9,$78,$1F,$6D,$C4,$FF,$C0
	.dc.b	$18,$C8,$8C,$BF,$EB,$5F,$80,$7C,$07,$58,$FC,$24,$61,$E2,$DA,$FD
	.dc.b	$6B,$63,$97,$BD,$C5,$42,$30,$D8,$E7,$1B,$BF,$F2,$D5,$7E,$E1,$B3
	.dc.b	$FE,$C9,$6B,$F0,$36,$ED,$EE,$BC,$0A,$49,$F7,$C6,$5F,$83,$D8,$E0
	.dc.b	$F3,$36,$33,$18,$3D,$5F,$04,$F8,$2A,$49,$F9,$71,$97,$E6,$FB,$CF
	.dc.b	$6F,$E1,$36,$33,$18,$3A,$E7,$23,$9B,$75,$FE,$33,$B7,$95,$B9,$7C
	.dc.b	$D8,$C5,$7F,$BB,$5A,$FE,$47,$93,$E9,$72,$7E,$4A,$FF,$91,$B6,$7F
	.dc.b	$37,$1F,$C9,$BE,$11,$F0,$9F,$65,$B3,$F8,$86,$CE,$E7,$EF,$77,$BD
	.dc.b	$6F,$F6,$FC,$FF,$B8,$89,$7F,$EB,$77,$F9,$9F,$C3,$3E,$1B,$F9,$AC
	.dc.b	$FF,$11,$8C,$38,$6B,$C2,$36,$A3,$2F,$EE,$8C,$46,$A3,$31,$FB,$E6
	.dc.b	$BD,$E3,$44,$7C,$0D,$7C,$06,$08,$60,$87,$86,$3A,$BE,$23,$66,$63
	.dc.b	$A3,$4C,$9E,$57,$2D,$32,$79,$CA,$75,$3C,$26,$CA,$6C,$A6,$CF,$01
	.dc.b	$88,$F3,$92,$3C,$31,$C4,$D9,$0E,$C5,$34,$60,$86,$08,$78,$4C,$94
	.dc.b	$CC,$78,$5F,$6B,$5F,$34,$79,$1E,$04,$8D,$3E,$33,$11,$B8,$CC,$78
	.dc.b	$1A,$F8,$23,$44,$7C,$2D,$7C,$26,$08,$60,$87,$86,$3A,$BE,$23,$66
	.dc.b	$63,$A3,$4C,$9E,$57,$2D,$32,$79,$CA,$75,$3C,$26,$CA,$6C,$A6,$CF
	.dc.b	$01,$8F,$53,$FD,$46,$FE,$A5,$AF,$FA,$B8,$C7,$AD,$FD,$21,$FD,$87
	.dc.b	$8B,$FE,$07,$F5,$FF,$76,$D8,$FD,$13,$5F,$D8,$35,$FD,$F3,$5E,$67
	.dc.b	$FB,$0C,$C9,$4C,$C7,$67,$18,$7F,$8D,$A3,$04,$8F,$09,$A9,$ED,$3B
	.dc.b	$8C,$CE,$DF,$29,$B2,$39,$7C,$6D,$7C,$67,$22,$1C,$9D,$47,$78,$C1
	.dc.b	$A3,$99,$4E,$66,$88,$68,$A6,$8C,$10,$C1,$1F,$03,$96,$B9,$9F,$86
	.dc.b	$47,$DE,$78,$E3,$64,$8D,$4F,$B1,$B3,$B6,$BE,$69,$E0,$F4,$31,$1A
	.dc.b	$7C,$8D,$7C,$86,$CC,$94,$CC,$F8,$5C,$4F,$68,$C1,$23,$C8,$6A,$7B
	.dc.b	$4E,$E3,$33,$B7,$CA,$6C,$8E,$5F,$1B,$5F,$19,$C8,$87,$27,$51,$DE
	.dc.b	$30,$68,$E6,$53,$99,$A2,$1A,$3F,$CA,$B9,$FF,$6D,$F6,$1F,$0D,$FE
	.dc.b	$9C,$8C,$37,$54,$EA,$BF,$B2,$56,$9F,$EC,$DC,$1A,$74,$7F,$95,$E7
	.dc.b	$FD,$05,$7F,$E0,$66,$C7,$06,$BF,$E0,$D3,$89,$FE,$62,$E1,$CB,$4C
	.dc.b	$9F,$E6,$07,$2F,$F9,$0F,$BA,$F1,$FF,$93,$6F,$FC,$3F,$18,$37,$D0
	.dc.b	$E6,$A4,$BE,$B1,$97,$BF,$47,$FB,$76,$FF,$D8,$2D,$8F,$F0,$33,$5F
	.dc.b	$F4,$0B,$5F,$F5,$17,$5F,$FA,$12,$7F,$D9,$2D,$8F,$F0,$73,$5F,$F4
	.dc.b	$4B,$5F,$F5,$57,$5F,$F9,$2E,$7F,$D1,$E4,$8F,$F0,$B4,$65,$FF,$0C
	.dc.b	$7C,$47,$D3,$9F,$F6,$BB,$63,$3A,$FC,$88,$C4,$7F,$98,$4F,$9B,$DD
	.dc.b	$F0,$6B,$FC,$D1,$9F,$CE,$8C,$4F,$F9,$F7,$E2,$7E,$8F,$80,$69,$23
	.dc.b	$EB,$C7,$D3,$7F,$D0,$9B,$70,$23,$F7,$3E,$79,$E7,$FD,$C6,$C6,$63
	.dc.b	$07,$E0,$3F,$6E,$FF,$31,$67,$FE,$F4,$6C,$FF,$8E,$5A,$FF,$AD,$1A
	.dc.b	$FF,$B5,$79,$7F,$45,$F1,$3F,$9F,$DC,$D2,$4F,$F9,$8F,$2F,$F6,$E4
	.dc.b	$7E,$32,$36,$64,$A7,$FB,$0B,$93,$FD,$3D,$1F,$E4,$B8,$72,$7E,$36
	.dc.b	$30,$D8,$FF,$28,$35,$FF,$65,$B5,$FF,$76,$70,$FE,$BE,$7E,$EE,$1E
	.dc.b	$49,$FB,$A7,$88,$DB,$97,$F8,$0F,$3F,$EF,$3E,$27,$E0,$12,$3F,$7B
	.dc.b	$19,$7F,$DB,$7C,$6F,$55,$E3,$FF,$BF,$6F,$FC,$CD,$39,$8C,$57,$0A
	.dc.b	$DA,$AD,$C6,$1F,$D1,$C7,$03,$F4,$25,$3F,$52,$51,$CF,$FE,$EE,$8F
	.dc.b	$B1,$F6,$32,$7D,$39,$23,$10,$3F,$80,$5F,$FA,$4B,$E2,$BC,$7E,$26
	.dc.b	$5B,$26,$9F,$85,$FC,$46,$47,$66,$BD,$74,$F8,$DF,$81,$F7,$F8,$BF
	.dc.b	$7B,$67,$E2,$35,$E1,$8A,$31,$1F,$95,$DF,$F9,$8F,$DC,$7E,$54,$91
	.dc.b	$EE,$7F,$31,$E1,$C8,$F8,$2D,$7E,$FE,$DF,$E2,$1E,$7F,$E8,$13,$99
	.dc.b	$ED,$3F,$E2,$7C,$FE,$F6,$FF,$E2,$D6,$CE,$73,$F9,$8D,$9D,$43,$F0
	.dc.b	$CF,$9D,$F8,$57,$DE,$6D,$7E,$0D,$AF,$C0,$6B,$E2,$C7,$C8,$B2,$F3
	.dc.b	$BE,$C2,$1E,$C8,$DC,$6A,$30,$68,$E8,$53,$A6,$9F,$D7,$27,$8B,$63
	.dc.b	$86,$5F,$9F,$F6,$BF,$11,$EB,$7E,$61,$B3,$F9,$D6,$BF,$69,$AF,$E3
	.dc.b	$BE,$53,$E2,$91,$F9,$4F,$F9,$5F,$1F,$44,$79,$8F,$49,$0C,$10,$F4
	.dc.b	$ED,$F6,$F1,$F4,$47,$E2,$BF,$10,$FC,$4E,$5F,$B7,$6E,$C6,$9F,$8E
	.dc.b	$9E,$7A,$7C,$2C,$3A,$11,$C1,$F8,$7B,$FE,$05,$F8,$3F,$EA,$FF,$0C
	.dc.b	$FE,$CD,$7F,$AC,$8C,$4F,$F0,$16,$C7,$E7,$9A,$FA,$F3,$F8,$2F,$88
	.dc.b	$FA,$3F,$88,$3A,$3F,$71,$F3,$FD,$B0,$4E,$69,$39,$8C,$36,$B8,$D7
	.dc.b	$F1,$AC,$5E,$07,$FA,$3E,$30,$DC,$3F,$75,$3A,$59,$F8,$63,$0F,$8F
	.dc.b	$6D,$72,$5B,$FD,$0E,$1F,$C8,$5B,$87,$EF,$9A,$FC,$1E,$0F,$B2,$DD
	.dc.b	$FF,$65,$97,$6B,$F2,$CF,$89,$B9,$78,$F0,$F4,$4F,$18,$FA,$EF,$E7
	.dc.b	$F6,$F9,$A4,$C7,$BC,$3D,$07,$A0,$F4,$BF,$93,$7C,$6F,$C0,$AF,$E5
	.dc.b	$15,$A9,$CE,$3F,$CA,$E7,$95,$F1,$57,$C5,$CD,$70,$79,$CE,$E7,$57
	.dc.b	$D6,$78,$DF,$58,$C4,$EA,$B8,$4F,$F1,$CA,$FE,$F0,$43,$FA,$09,$4D
	.dc.b	$18,$21,$F7,$B2,$FE,$BA,$D3,$F9,$D1,$4F,$9C,$60,$FE,$4E,$75,$29
	.dc.b	$D0,$FE,$9A,$FF,$4F,$FC,$2F,$F0,$90,$69,$34,$FB,$B1,$D8,$8E,$0F
	.dc.b	$EE,$3C,$7F,$9F,$7F,$39,$1F,$E5,$97,$3C,$5F,$D4,$5B,$19,$3E,$09
	.dc.b	$0F,$39,$5F,$DD,$AF,$D4,$FF,$39,$CB,$FD,$9D,$B1,$99,$FD,$E3,$4E
	.dc.b	$87,$F1,$43,$E2,$14,$F8,$85,$37,$C1,$FE,$91,$7F,$C0,$7C,$DF,$DD
	.dc.b	$7E,$06,$C1,$A4,$79,$CF,$16,$F1,$7F,$A4,$5B,$C9,$FE,$2F,$7C,$07
	.dc.b	$C0,$3F,$9A,$1F,$E3,$3C,$3F,$C3,$9D,$1B,$6C,$67,$9D,$39,$5D,$3F
	.dc.b	$E0,$BF,$83,$FE,$89,$46,$91,$F1,$9F,$13,$2F,$43,$F0,$B2,$F4,$37
	.dc.b	$2F,$44,$EA,$3D,$07,$A8,$8E,$5B,$3C,$9F,$76,$5F,$E3,$7E,$CF,$C9
	.dc.b	$47,$F7,$13,$F2,$45,$E2,$FE,$EC,$E0,$FF,$09,$F6,$FB,$C8,$FC,$37
	.dc.b	$E6,$3E,$0C,$7C,$F8,$78,$63,$07,$BD,$C6,$3E,$73,$FD,$A1,$EE,$7F
	.dc.b	$9B,$C9,$3F,$E3,$87,$FC,$4E,$FE,$04,$FF,$13,$94,$FF,$13,$F8,$1F
	.dc.b	$87,$73,$C1,$FE,$79,$FC,$42,$47,$F9,$09,$AF,$F8,$BB,$97,$FC,$85
	.dc.b	$F0,$5F,$E3,$E2,$47,$F9,$5E,$32,$FF,$92,$B3,$FF,$42,$11,$C1,$FE
	.dc.b	$B8,$3F,$C9,$9F,$01,$FE,$5A,$4D,$23,$FE,$63,$F2,$76,$D3,$65,$8F
	.dc.b	$F3,$2C,$6A,$3F,$C7,$31,$FE,$95,$23,$FE,$5E,$7F,$D0,$2B,$E0,$29
	.dc.b	$1F,$F4,$1B,$FE,$8E,$E2,$FF,$31,$36,$38,$1F,$E9,$08,$CB,$4C,$9F
	.dc.b	$E9,$BF,$3B,$2D,$32,$7F,$9B,$7B,$11,$A8,$C3,$5F,$F0,$CD,$7F,$AB
	.dc.b	$C8,$FF,$9E,$1F,$F5,$D7,$BF,$FE,$77,$F0,$AC,$FF,$4D,$3F,$D7,$51
	.dc.b	$97,$FD,$7C,$D7,$FA,$A7,$A7,$FE,$C8,$9F,$F5,$83,$63,$FA,$C0,$70
	.dc.b	$F0,$F4,$F9,$3F,$E8,$7F,$53,$D6,$F8,$4C,$63,$A5,$1A,$BF,$F5,$F3
	.dc.b	$63,$FD,$CC,$71,$37,$EA,$F1,$8F,$F7,$44,$65,$B3,$FE,$F0,$8E,$1E
	.dc.b	$B3,$5F,$03,$63,$31,$E3,$F4,$DC,$34,$93,$E1,$8C,$BE,$3F,$5D,$EF
	.dc.b	$FA,$DB,$19,$8C,$1E,$FF,$53,$FC,$01,$97,$E7,$C9,$3F,$A3,$3D,$DE
	.dc.b	$CB,$E1,$BE,$33,$63,$31,$83,$E5,$7B,$1E,$BA,$FC,$A1,$27,$F2,$4F
	.dc.b	$FB,$BC,$8F,$FB,$BD,$AF,$F8,$4F,$DC,$7B,$9F,$55,$B1,$F4,$5B,$3F
	.dc.b	$EF,$36,$C7,$E1,$1F,$8B,$19,$D7,$91,$27,$ED,$78,$0F,$07,$E7,$D7
	.dc.b	$FB,$F6,$31,$5A,$9C,$D7,$FB,$52,$BF,$22,$BD,$D7,$BD,$FC,$B1,$B1
	.dc.b	$FD,$65,$AF,$06,$BF,$9A,$6B,$FE,$33,$6B,$FE,$DF,$6B,$DE,$3B,$11
	.dc.b	$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B1,$D0,$87,$42,$9D,$0A,$74
	.dc.b	$29,$D0,$A7,$43,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0,$87,$02,$9C
	.dc.b	$0A,$70,$34,$43,$45,$34,$60,$86,$09,$1F,$50,$91,$F2,$5B,$1F,$01
	.dc.b	$AF,$A5,$AF,$91,$AF,$85,$AF,$81,$AF,$82,$3B,$11,$EC,$D7,$B3,$5E
	.dc.b	$CD,$7B,$35,$EC,$D7,$B1,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$A7,$43
	.dc.b	$91,$0E,$45,$39,$14,$FF,$47,$CF,$F9,$0F,$D2,$FF,$47,$47,$F4,$26
	.dc.b	$BF,$E8,$FF,$47,$F9,$B3,$FB,$5E,$B3,$FE,$9B,$3F,$58,$E3,$D1,$FD
	.dc.b	$A3,$63,$F6,$8D,$7F,$68,$D7,$89,$CC,$EC,$5C,$51,$B8,$CC,$F6,$75
	.dc.b	$19,$9D,$C6,$09,$3B,$23,$E4,$32,$57,$2F,$98,$C9,$5C,$BE,$86,$BE
	.dc.b	$43,$91,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$3E,$0C
	.dc.b	$79,$1B,$1D,$DA,$F4,$6B,$CD,$AF,$39,$F9,$84,$7E,$19,$B7,$13,$DD
	.dc.b	$B1,$DD,$AF,$76,$BC,$4E,$67,$72,$CF,$88,$91,$C2,$35,$3D,$DD,$C6
	.dc.b	$67,$71,$82,$4E,$C8,$F9,$4C,$95,$CB,$E7,$32,$57,$2F,$A5,$AF,$90
	.dc.b	$E4,$43,$81,$4E,$05,$38,$1A,$21,$FA,$ED,$3F,$B4,$FB,$DF,$DA,$F4
	.dc.b	$FF,$0D,$B7,$53,$4F,$ED,$7F,$00,$F0,$FF,$6F,$18,$6E,$1F,$E4,$09
	.dc.b	$D3,$C8,$C4,$69,$FF,$05,$70,$D7,$F8,$0F,$0F,$F8,$2A,$FF,$C1,$13
	.dc.b	$99,$4E,$5F,$F2,$6F,$83,$FE,$2A,$AF,$EF,$ED,$8C,$F3,$F1,$8C,$47
	.dc.b	$F0,$E1,$C9,$FE,$3D,$AF,$F3,$B1,$23,$FC,$E8,$FF,$82,$B9,$BF,$80
	.dc.b	$B7,$87,$C9,$18,$9F,$F2,$FB,$5E,$DC,$FD,$EB,$FD,$30,$48,$FF,$1D
	.dc.b	$3F,$E1,$6E,$8F,$2C,$7F,$22,$6A,$69,$89,$79,$1F,$15,$B3,$F3,$CF
	.dc.b	$0C,$64,$F9,$5D,$9E,$F8,$FF,$3A,$B5,$F8,$8D,$7F,$99,$B5,$FF,$38
	.dc.b	$E7,$F4,$FB,$4D,$24,$78,$4F,$F3,$C9,$4F,$F3,$CF,$7F,$12,$39,$A7
	.dc.b	$CC,$F9,$89,$19,$7F,$05,$F0,$1F,$36,$3F,$A3,$B5,$34,$C7,$3F,$F9
	.dc.b	$1B,$E0,$3F,$38,$49,$FF,$4A,$C6,$5F,$CA,$88,$FC,$E3,$4C,$B8,$3F
	.dc.b	$D7,$84,$7F,$D7,$4D,$7F,$1D,$BF,$F4,$88,$FC,$44,$73,$33,$C1,$FD
	.dc.b	$82,$32,$7F,$A4,$C9,$1F,$5B,$C1,$D3,$4E,$47,$F6,$8E,$7F,$D7,$FC
	.dc.b	$07,$A0,$91,$EC,$7F,$D6,$7E,$4B,$27,$9B,$FD,$C2,$73,$18,$A4,$62
	.dc.b	$75,$1F,$A9,$AE,$51,$88,$FD,$65,$7F,$9D,$23,$F8,$51,$23,$E7,$F6
	.dc.b	$7F,$AA,$23,$E6,$92,$3E,$8B,$5F,$DB,$7C,$EB,$E0,$BF,$C0,$BF,$05
	.dc.b	$FE,$7C,$8C,$4A,$31,$3A,$8F,$F6,$8C,$EA,$7E,$DF,$C1,$49,$FB,$ED
	.dc.b	$7F,$C2,$3A,$BE,$6D,$EC,$7F,$85,$3E,$0F,$FD,$B7,$39,$8C,$57,$F0
	.dc.b	$B8,$C2,$9D,$D7,$28,$C1,$F8,$EF,$83,$FC,$69,$1C,$1F,$1B,$C4,$FF
	.dc.b	$0C,$6F,$FE,$9E,$F1,$7F,$9A,$12,$7F,$DD,$BF,$92,$7F,$A6,$A3,$FD
	.dc.b	$3E,$D8,$CF,$97,$FE,$1D,$8C,$47,$BB,$DB,$FF,$68,$D7,$B1,$B1,$EB
	.dc.b	$6B,$BF,$17,$FC,$59,$AA,$5F,$67,$86,$D7,$F8,$E6,$75,$39,$3F,$C5
	.dc.b	$D9,$7A,$E7,$DD,$87,$AD,$FC,$4F,$2F,$FB,$B2,$3C,$9E,$87,$52,$4F
	.dc.b	$5C,$7F,$DF,$73,$C6,$3F,$AE,$57,$17,$A9,$0F,$E8,$19,$FF,$BD,$27
	.dc.b	$E3,$EB,$F1,$49,$3D,$7C,$0F,$F2,$9F,$E9,$9F,$15,$B1,$F1,$5A,$FC
	.dc.b	$B6,$AD,$5F,$88,$D8,$F8,$8D,$C3,$F1,$6D,$8F,$C5,$35,$F8,$78,$7E
	.dc.b	$0B,$2F,$64,$62,$BE,$05,$F6,$F8,$8F,$AB,$81,$A4,$AF,$0F,$C2,$7F
	.dc.b	$9B,$BF,$25,$F0,$B6,$3C,$2D,$7F,$47,$48,$C4,$7C,$EC,$BC,$84,$C7
	.dc.b	$E9,$47,$F0,$DF,$8D,$FB,$E4,$7F,$36,$F9,$BD,$1F,$F3,$FF,$E4,$DF
	.dc.b	$16,$7E,$46,$3F,$EF,$75,$19,$D3,$E5,$D7,$07,$E3,$B6,$7E,$21,$F7
	.dc.b	$CF,$36,$1F,$1B,$E2,$1F,$0F,$FE,$8E,$F8,$3E,$91,$8A,$46,$27,$58
	.dc.b	$7E,$F6,$B5,$3E,$3F,$B4,$FE,$AF,$6C,$31,$E5,$FF,$A5,$F0,$FE,$09
	.dc.b	$5A,$9C,$D6,$E7,$F8,$47,$E2,$4F,$84,$FF,$50,$FC,$5F,$BE,$31,$48
	.dc.b	$C4,$EA,$B8,$B6,$7E,$09,$0F,$82,$7C,$38,$61,$E3,$7C,$B8,$F9,$E7
	.dc.b	$F9,$0C,$E5,$B1,$97,$99,$D4,$FE,$95,$CB,$F9,$47,$24,$33,$C5,$F8
	.dc.b	$57,$EA,$0F,$6D,$ED,$BD,$CF,$F6,$0F,$37,$CB,$AF,$F6,$59,$F9,$1D
	.dc.b	$FF,$D9,$13,$A6,$C6,$4D,$CF,$17,$B5,$75,$74,$62,$B9,$3C,$2B,$F5
	.dc.b	$EF,$F9,$84,$87,$F3,$32,$9A,$30,$43,$E3,$9F,$D2,$8A,$7E,$00,$9A
	.dc.b	$6D,$FD,$51,$D0,$A7,$DC,$1F,$0D,$F8,$06,$C6,$63,$07,$48,$EF,$1F
	.dc.b	$71,$F1,$1F,$C4,$F5,$FC,$BB,$F9,$D1,$EE,$52,$4F,$E7,$F9,$3F,$DE
	.dc.b	$71,$FB,$59,$FB,$67,$9C,$A7,$AE,$7F,$6A,$BF,$4A,$FD,$C7,$47,$EE
	.dc.b	$DB,$CF,$FE,$FE,$6C,$EE,$BF,$C6,$7D,$3F,$E7,$58,$A5,$3F,$D3,$0D
	.dc.b	$69,$7C,$17,$1B,$8F,$0D,$9F,$44,$EA,$7A,$5F,$F3,$6E,$25,$3F,$21
	.dc.b	$FE,$4A,$D3,$F9,$29,$4F,$E4,$B1,$C2,$3F,$9D,$F6,$F9,$6B,$F9,$A5
	.dc.b	$76,$D7,$F9,$A3,$67,$D4,$FC,$A3,$F1,$84,$3F,$06,$7F,$32,$9E,$93
	.dc.b	$F4,$E3,$07,$58,$F0,$78,$2E,$FF,$5F,$EA,$BF,$D6,$68,$D2,$3F,$E0
	.dc.b	$AC,$FC,$7E,$17,$C7,$8F,$EC,$A7,$CF,$70,$FE,$64,$F4,$1E,$62,$61
	.dc.b	$FE,$05,$F0,$FC,$64,$8F,$9F,$8A,$3E,$0E,$DF,$4E,$7E,$AE,$FD,$9E
	.dc.b	$75,$FE,$0C,$7F,$C2,$DE,$27,$B8,$8F,$4F,$17,$E0,$91,$C1,$E9,$F1
	.dc.b	$BD,$A4,$70,$7E,$1B,$48,$F8,$9F,$C7,$7A,$5F,$ED,$AF,$6B,$FD,$06
	.dc.b	$49,$FF,$43,$BF,$E3,$EF,$33,$F1,$3F,$13,$64,$D2,$3F,$E8,$EF,$89
	.dc.b	$FF,$46,$91,$C1,$FE,$58,$F4,$BF,$27,$EA,$7F,$97,$20,$D2,$3F,$E9
	.dc.b	$8C,$BF,$D2,$E4,$70,$7F,$92,$DF,$F7,$E1,$1C,$1F,$E7,$BE,$0F,$F4
	.dc.b	$07,$C4,$7F,$9F,$E0,$D2,$3F,$EA,$67,$FD,$4A,$47,$07,$FA,$57,$E2
	.dc.b	$3F,$71,$EE,$7F,$88,$5B,$CB,$FA,$E6,$C9,$FF,$52,$91,$FF,$75,$FB
	.dc.b	$5B,$69,$B3,$F8,$37,$C1,$75,$6D,$6D,$FF,$7F,$46,$5B,$19,$7F,$D5
	.dc.b	$1F,$09,$57,$12,$30,$D8,$F0,$1F,$C1,$8D,$C7,$1C,$FF,$86,$63,$FC
	.dc.b	$5D,$B3,$FE,$7B,$8C,$73,$FF,$15,$7F,$8E,$7B,$CC,$FF,$88,$F9,$38
	.dc.b	$6D,$FF,$23,$9D,$3F,$D6,$CF,$13,$89,$F9,$FE,$67,$09,$FF,$7D,$B5
	.dc.b	$FF,$7D,$B5,$6D,$FE,$1A,$9F,$E5,$DF,$01,$FC,$26,$7B,$1F,$EF,$A8
	.dc.b	$CF,$7F,$19,$D4,$70,$F8,$0C,$12,$B8,$89,$47,$86,$FF,$CC,$95,$F8
	.dc.b	$22,$57,$FB,$75,$FE,$81,$F0,$2E,$AF,$43,$67,$AB,$67,$73,$ED,$F8
	.dc.b	$27,$6D,$25,$F9,$23,$2F,$23,$A9,$0C,$1E,$AF,$82,$F9,$CF,$E4,$AF
	.dc.b	$FD,$F6,$47,$F2,$27,$C0,$E6,$F6,$B8,$8F,$EB,$51,$88,$FC,$56,$7E
	.dc.b	$29,$FE,$BC,$D8,$FE,$C4,$D7,$78,$7D,$9F,$84,$FE,$F3,$1F,$39,$B5
	.dc.b	$FD,$9D,$B1,$FD,$AD,$AF,$E6,$79,$BF,$39,$3F,$DB,$5B,$1F,$DC,$5A
	.dc.b	$EF,$9B,$F1,$1C,$DF,$DA,$6B,$F3,$7D,$3F,$7A,$31,$5F,$ED,$99,$FE
	.dc.b	$F1,$E5,$FE,$62,$2F,$37,$F6,$08,$C4,$7F,$7F,$6B,$FA,$D6,$BF,$DD
	.dc.b	$1A,$9A,$48,$E8,$48,$FF,$03,$46,$5B,$19,$7C,$66,$08,$60,$87,$95
	.dc.b	$E0,$D3,$65,$36,$7A,$5D,$B4,$D9,$4D,$9E,$D7,$6D,$36,$53,$67,$C1
	.dc.b	$76,$D3,$65,$37,$1F,$05,$B1,$C2,$31,$1E,$92,$1E,$52,$9E,$12,$9D
	.dc.b	$4A,$F2,$78,$10,$E0,$53,$81,$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$23
	.dc.b	$D8,$48,$F8,$71,$96,$C6,$5F,$86,$60,$86,$08,$7B,$9E,$0D,$36,$53
	.dc.b	$67,$B5,$DB,$4D,$94,$D9,$EE,$76,$D3,$67,$FA,$88,$FF,$32,$7A,$5F
	.dc.b	$EA,$8A,$FF,$31,$36,$3F,$CC,$CD,$7F,$59,$F2,$9F,$E5,$CF,$4B,$F6
	.dc.b	$6D,$8C,$9F,$EA,$73,$F5,$CE,$5F,$D9,$35,$C9,$B3,$F6,$4E,$5A,$67
	.dc.b	$04,$65,$B1,$98,$C1,$1C,$12,$7C,$ED,$3C,$EE,$08,$E0,$87,$A0,$E0
	.dc.b	$57,$6D,$36,$7B,$1C,$B4,$C9,$EC,$72,$D3,$33,$EC,$8C,$B6,$33,$3E
	.dc.b	$D9,$CB,$63,$27,$98,$F0,$1E,$52,$18,$21,$E1,$32,$53,$33,$E1,$72
	.dc.b	$D7,$31,$E1,$72,$D7,$27,$8C,$E8,$48,$E4,$D8,$E0,$D3,$84,$68,$91
	.dc.b	$A6,$C6,$A3,$04,$70,$49,$F8,$ED,$3E,$3B,$82,$38,$21,$EC,$38,$15
	.dc.b	$DB,$4D,$9F,$01,$CB,$4C,$9E,$D7,$2B,$1F,$F6,$57,$C2,$FF,$62,$24
	.dc.b	$7F,$B6,$63,$2F,$F6,$BF,$57,$FD,$9B,$3F,$DB,$DB,$19,$11,$96,$9F
	.dc.b	$EE,$3F,$D1,$1F,$7B,$C2,$3F,$CA,$51,$87,$66,$23,$4F,$FB,$72,$35
	.dc.b	$3F,$E0,$5F,$AD,$CC,$B8,$91,$86,$C7,$43,$FD,$CA,$E2,$7F,$C2,$FE
	.dc.b	$4F,$2E,$53,$49,$5C,$47,$37,$FB,$8A,$BF,$C4,$7E,$7F,$8A,$31,$3F
	.dc.b	$E7,$C8,$F2,$F4,$7F,$08,$AF,$F3,$84,$FF,$BD,$A3,$0E,$1A,$FF,$9D
	.dc.b	$5A,$FF,$BF,$3A,$FF,$C9,$13,$FE,$46,$8F,$24,$61,$EA,$D7,$FC,$FD
	.dc.b	$CA,$F4,$7E,$3B,$63,$D8,$DA,$F2,$C7,$F9,$EE,$77,$BC,$8D,$B6,$77
	.dc.b	$F1,$5F,$E3,$A6,$FA,$BF,$52,$31,$5F,$01,$AF,$FA,$47,$C0,$FF,$2C
	.dc.b	$D7,$F9,$E3,$B7,$ED,$C6,$27,$E3,$B5,$FF,$4C,$7E,$0E,$F1,$FF,$02
	.dc.b	$D8,$FA,$CD,$BF,$5B,$63,$E8,$FC,$6F,$FA,$4A,$7D,$ED,$8C,$EB,$F8
	.dc.b	$C8,$C4,$7F,$A9,$FA,$1E,$43,$BD,$F1,$7F,$E4,$A2,$57,$EA,$67,$51
	.dc.b	$FA,$32,$47,$E9,$DA,$FE,$7C,$E6,$7E,$9F,$80,$E0,$D4,$7D,$6E,$AF
	.dc.b	$D4,$3F,$92,$8D,$1C,$39,$9C,$DC,$A7,$44,$9F,$C8,$76,$E5,$A7,$21
	.dc.b	$CF,$FD,$82,$7F,$0A,$D8,$FC,$DF,$CA,$FF,$A6,$FE,$33,$F1,$2D,$8C
	.dc.b	$F8,$DE,$A9,$CB,$63,$DC,$23,$D9,$A6,$63,$E2,$46,$1D,$1F,$37,$67
	.dc.b	$C6,$7F,$9E,$89,$19,$7F,$89,$11,$FE,$2C,$D7,$E8,$3D,$48,$F9,$1F
	.dc.b	$3A,$F8,$AA,$47,$DA,$64,$F5,$BE,$72,$38,$37,$F1,$1D,$08,$E0,$F8
	.dc.b	$6F,$BC,$8E,$0F,$81,$E1,$FF,$A6,$08,$E0,$FF,$7E,$78,$CF,$C2,$1E
	.dc.b	$6F,$F2,$A8,$C3,$63,$33,$FC,$96,$72,$FF,$27,$EB,$7B,$7F,$A0,$87
	.dc.b	$C1,$FF,$1A,$9C,$F4,$7E,$5A,$31,$5B,$6C,$66,$BF,$36,$7E,$E6,$5F
	.dc.b	$59,$FE,$35,$BE,$ED,$8C,$CF,$F8,$F6,$72,$FF,$8E,$7A,$BF,$67,$C1
	.dc.b	$70,$FD,$A6,$BF,$12,$BF,$71,$18,$7C,$37,$EC,$D7,$B1,$23,$B6,$3F
	.dc.b	$12,$F9,$47,$EC,$B1,$EF,$18,$8F,$ED,$58,$75,$7F,$6F,$A3,$DE,$7C
	.dc.b	$87,$D1,$6C,$FC,$F6,$F7,$FF,$44,$AE,$15,$FD,$83,$4F,$A3,$CB,$FC
	.dc.b	$0B,$F5,$6F,$9A,$DC,$7D,$7A,$FF,$69,$9C,$C1,$FA,$7F,$85,$FC,$B2
	.dc.b	$47,$B9,$FF,$30,$6D,$EC,$8F,$C5,$1F,$23,$83,$E6,$1F,$13,$F0,$5F
	.dc.b	$A8,$48,$C1,$AE,$0F,$89,$5C,$E3,$94,$63,$83,$E3,$47,$CB,$6B,$F2
	.dc.b	$AF,$EE,$7E,$1F,$FD,$F2,$CD,$24,$FF,$73,$8E,$CF,$39,$F1,$3E,$A3
	.dc.b	$17,$E1,$8C,$C7,$F2,$CD,$7F,$3B,$3F,$CB,$BD,$BF,$E9,$53,$F9,$58
	.dc.b	$C4,$FC,$E6,$C7,$1D,$51,$88,$FC,$E7,$3F,$FA,$5E,$BE,$AB,$63,$83
	.dc.b	$5D,$9C,$A3,$F2,$BF,$18,$F5,$3F,$59,$18,$6C,$FA,$67,$4F,$ED,$27
	.dc.b	$4D,$3E,$89,$F0,$CF,$9D,$7F,$9C,$8F,$DE,$DF,$8D,$B3,$E3,$3C,$E5
	.dc.b	$3C,$F3,$F1,$9E,$67,$5C,$FE,$9B,$83,$C2,$43,$E9,$9E,$3F,$53,$FD
	.dc.b	$81,$E6,$FC,$A8,$FE,$BF,$18,$AF,$E0,$ED,$8C,$FC,$53,$9B,$FB,$2D
	.dc.b	$F1,$95,$6D,$B5,$B7,$FC,$45,$7D,$23,$15,$C6,$B8,$61,$FC,$32,$FF
	.dc.b	$C4,$6D,$7F,$A8,$18,$E1,$FF,$01,$37,$83,$FC,$52,$FE,$6C,$C1,$0F
	.dc.b	$C1,$9F,$91,$3E,$69,$F9,$0C,$FE,$8B,$77,$6B,$FC,$62,$3F,$68,$73
	.dc.b	$29,$E4,$3F,$03,$F1,$5C,$F7,$FC,$2B,$78,$3F,$C1,$6D,$93,$DE,$BB
	.dc.b	$47,$E5,$0C,$1E,$73,$18,$7E,$8D,$FF,$26,$37,$5F,$46,$FF,$52,$30
	.dc.b	$79,$67,$CE,$43,$F7,$07,$AC,$A6,$CA,$7F,$8C,$78,$BF,$96,$BF,$E9
	.dc.b	$56,$B9,$3E,$F1,$78,$7A,$3F,$4C,$E8,$53,$C4,$53,$B9,$27,$FB,$0F
	.dc.b	$1F,$DC,$6C,$7F,$99,$9A,$FF,$60,$8F,$C0,$BC,$C9,$C1,$F8,$97,$E0
	.dc.b	$4F,$F6,$CE,$4E,$9B,$9A,$49,$ED,$1F,$EA,$5F,$D5,$FF,$9E,$92,$3F
	.dc.b	$A0,$F3,$FF,$91,$A3,$FA,$57,$57,$F4,$EC,$EF,$4F,$C6,$BF,$D4,$4E
	.dc.b	$5F,$F0,$3B,$5F,$87,$9F,$9D,$F8,$67,$C5,$CF,$FB,$43,$E4,$EB,$F4
	.dc.b	$91,$FD,$0C,$F1,$7E,$0F,$37,$E9,$9F,$B9,$5F,$DE,$9F,$F0,$3F,$67
	.dc.b	$C2,$23,$F0,$27,$F2,$EF,$F5,$E6,$BF,$CC,$5A,$FF,$82,$DB,$97,$F8
	.dc.b	$68,$FE,$D9,$DD,$FE,$1E,$6E,$3F,$0D,$D5,$7E,$13,$1F,$41,$2F,$D1
	.dc.b	$1F,$E7,$5F,$0B,$FC,$A4,$47,$FC,$B2,$FF,$91,$88,$FF,$91,$1A,$FF
	.dc.b	$9F,$7E,$03,$FC,$B4,$47,$FC,$C2,$FF,$93,$88,$FF,$93,$1A,$FF,$A1
	.dc.b	$7A,$BF,$CC,$44,$7F,$CD,$2F,$FB,$BC,$8F,$FB,$CD,$AF,$F7,$B7,$F8
	.dc.b	$B1,$1F,$F3,$53,$FE,$6D,$F7,$E4,$7F,$9D,$1A,$FF,$8E,$3C,$55,$FF
	.dc.b	$A5,$27,$FC,$CA,$49,$FF,$31,$BF,$E8,$6F,$23,$FD,$33,$EF,$D9,$34
	.dc.b	$8F,$FB,$9C,$C9,$5F,$F4,$E1,$1F,$F4,$3B,$FE,$83,$23,$83,$FD,$DB
	.dc.b	$ED,$BD,$AF,$0C,$61,$AF,$F9,$4A,$34,$F8,$0C,$46,$9F,$F0,$64,$FF
	.dc.b	$3D,$23,$FE,$91,$7F,$D9,$3E,$87,$EB,$61,$E8,$23,$0D,$8F,$E2,$A7
	.dc.b	$FA,$F1,$C7,$07,$FB,$5B,$93,$97,$A3,$FE,$DB,$8C,$57,$01,$19,$7F
	.dc.b	$DC,$BE,$9F,$87,$D3,$34,$91,$C7,$D4,$7A,$BE,$76,$C6,$63,$07,$AF
	.dc.b	$37,$6D,$24,$F8,$E3,$2D,$F0,$F8,$11,$C1,$E9,$F5,$BF,$12,$DE,$4F
	.dc.b	$91,$18,$D7,$FD,$0B,$3F,$EE,$08,$C3,$FE,$D4,$8F,$C5,$11,$C1,$F8
	.dc.b	$97,$FD,$C4,$47,$07,$4F,$61,$EC,$7D,$76,$C7,$CD,$6C,$F8,$A1,$18
	.dc.b	$7C,$BE,$CB,$DA,$FC,$03,$63,$EB,$36,$B9,$36,$3F,$DB,$F1,$FE,$DB
	.dc.b	$23,$83,$F2,$C7,$DF,$E6,$FC,$AC,$78,$DB,$1A,$8F,$F0,$81,$1F,$F7
	.dc.b	$23,$5F,$F5,$7F,$B6,$E8,$FE,$AA,$D8,$FF,$0B,$B5,$F8,$0D,$7E,$9B
	.dc.b	$5E,$71,$C8,$8F,$26,$BC,$9A,$F2,$6B,$C8,$E0,$43,$81,$4E,$05,$38
	.dc.b	$1A,$21,$A2,$9A,$30,$43,$04,$02,$3D,$CE,$1A,$E6,$3D,$2E,$5A,$E4
	.dc.b	$F3,$19,$29,$93,$CA,$64,$A6,$4F,$31,$92,$99,$3D,$05,$34,$62,$3C
	.dc.b	$E4,$8F,$2B,$63,$C6,$D7,$C2,$D7,$BB,$5E,$91,$C8,$8F,$26,$BC,$9A
	.dc.b	$F2,$6B,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04
	.dc.b	$3E,$39,$0F,$94,$48,$F8,$0E,$5A,$E6,$3F,$C7,$BE,$73,$CF,$FE,$69
	.dc.b	$3F,$E8,$EB,$FF,$40,$CE,$5B,$3A,$FC,$19,$E1,$7E,$43,$CC,$FE,$1D
	.dc.b	$18,$C3,$FD,$34,$7E,$51,$C4,$6A,$78,$4F,$F2,$69,$CB,$63,$2F,$E7
	.dc.b	$04,$69,$C3,$5E,$D1,$C1,$A7,$88,$E2,$E5,$A6,$63,$93,$89,$F3,$39
	.dc.b	$9E,$13,$A2,$38,$24,$6C,$93,$C8,$8F,$B4,$C9,$5C,$BE,$C8,$E7,$18
	.dc.b	$8F,$5B,$5F,$31,$C4,$87,$13,$A9,$5E,$13,$E8,$8C,$9D,$0E,$25,$38
	.dc.b	$18,$3B,$98,$21,$B2,$1A,$2B,$DE,$3C,$33,$EC,$9E,$93,$96,$C6,$5E
	.dc.b	$A7,$32,$46,$DC,$B5,$EF,$1C,$5A,$7B,$8E,$2E,$5A,$66,$39,$38,$9F
	.dc.b	$3B,$99,$E1,$3A,$23,$82,$46,$C9,$3C,$88,$FB,$8C,$95,$CB,$EC,$8E
	.dc.b	$71,$88,$F9,$BE,$1B,$C3,$78,$4F,$4F,$FD,$F0,$4A,$FF,$08,$D6,$E7
	.dc.b	$51,$FA,$BE,$E7,$94,$F8,$9F,$E9,$3E,$AF,$F8,$62,$73,$18,$AE,$15
	.dc.b	$B5,$5B,$8C,$3F,$E0,$FF,$7D,$A7,$1A,$FD,$D4,$61,$E2,$7F,$89,$9C
	.dc.b	$4F,$19,$E1,$2C,$3C,$FE,$86,$1B,$1C,$8F,$EC,$2E,$2B,$C9,$3B,$C1
	.dc.b	$F8,$97,$F6,$6F,$CF,$3D,$0D,$C3,$DA,$D7,$9C,$F0,$53,$EF,$FC,$53
	.dc.b	$51,$E5,$94,$63,$1F,$9A,$D8,$CC,$F2,$3E,$5F,$E0,$9F,$DE,$3B,$3E
	.dc.b	$53,$63,$27,$F8,$E0,$B1,$F3,$FC,$8D,$47,$99,$AF,$8D,$B8,$7D,$E6
	.dc.b	$DB,$F0,$5F,$F3,$C5,$7D,$C6,$BF,$01,$AF,$B5,$AB,$7F,$B5,$5F,$5E
	.dc.b	$3F,$D0,$E7,$DB,$DF,$73,$A8,$FC,$4F,$E0,$BD,$31,$FC,$8B,$71,$34
	.dc.b	$C7,$E0,$DF,$64,$95,$F7,$CF,$A0,$FE,$0F,$F5,$CF,$D1,$C6,$6D,$18
	.dc.b	$E2,$FC,$E3,$67,$F4,$4D,$7F,$12,$FF,$A2,$BF,$2B,$DF,$93,$FE,$2C
	.dc.b	$6C,$9A,$74,$F0,$F5,$1D,$1A,$F4,$6B,$F9,$BA,$FF,$37,$4F,$EA,$79
	.dc.b	$BF,$C7,$3F,$A4,$DA,$34,$8F,$EF,$3A,$FF,$80,$E5,$FE,$7D,$EA,$FD
	.dc.b	$F3,$63,$33,$BE,$AF,$91,$E8,$BF,$7B,$7B,$9F,$AE,$8C,$36,$33,$5C
	.dc.b	$1B,$87,$E5,$A3,$F8,$11,$FA,$6E,$9F,$F2,$8E,$99,$24,$7F,$08,$E7
	.dc.b	$CB,$4E,$27,$E5,$CB,$C5,$F9,$76,$F2,$B9,$3E,$2C,$7F,$12,$6B,$FB
	.dc.b	$66,$BF,$84,$3E,$49,$F7,$B5,$CD,$7E,$11,$BC,$5F,$88,$E1,$FC,$2C
	.dc.b	$63,$F1,$3F,$F3,$5F,$EB,$7F,$88,$6E,$1F,$88,$6B,$F8,$86,$BF,$E7
	.dc.b	$2E,$3F,$E6,$17,$F4,$EF,$C7,$3F,$E1,$B9,$CB,$F5,$67,$FD,$8F,$C5
	.dc.b	$F4,$A3,$61,$C4,$6B,$87,$EA,$B7,$8B,$DB,$B7,$C7,$B7,$C6,$FF,$35
	.dc.b	$DF,$FD,$0A,$D8,$C9,$F2,$CA,$7F,$A2,$B1,$F1,$6B,$D6,$0D,$23,$FA
	.dc.b	$AF,$C4,$3F,$D2,$3F,$AE,$7E,$B9,$B1,$91,$19,$69,$FE,$3D,$AF,$81
	.dc.b	$FA,$EF,$72,$46,$75,$FC,$9C,$7C,$BE,$D9,$3F,$9D,$6B,$FB,$AF,$C2
	.dc.b	$3F,$82,$11,$FE,$09,$C6,$8F,$E1,$1F,$82,$7F,$91,$F9,$7F,$83,$B6
	.dc.b	$4D,$3F,$C9,$1C,$7F,$DA,$F0,$EF,$87,$DF,$EA,$FE,$E7,$18,$84,$ED
	.dc.b	$F3,$10,$FC,$14,$F0,$7F,$89,$FD,$D7,$84,$8F,$F9,$AB,$1F,$E3,$71
	.dc.b	$E2,$9E,$13,$CB,$9B,$FC,$BD,$7D,$E3,$18,$7F,$20,$FC,$1B,$F9,$31
	.dc.b	$23,$E1,$BF,$CA,$7C,$BF,$F6,$4F,$9D,$FC,$99,$B3,$FC,$99,$AF,$F2
	.dc.b	$66,$BF,$C6,$FF,$08,$FF,$68,$ED,$FC,$B9,$B1,$FC,$B9,$AF,$F2,$E6
	.dc.b	$BF,$27,$2F,$4E,$7E,$C8,$C6,$5F,$3B,$5F,$E7,$7F,$65,$FC,$FC,$93
	.dc.b	$F7,$1F,$E8,$5F,$51,$FE,$E2,$FD,$7B,$F9,$FB,$67,$F9,$FB,$5F,$E7
	.dc.b	$ED,$7B,$9F,$54,$A7,$CA,$3F,$A3,$FC,$77,$F4,$F2,$3F,$8C,$7F,$A9
	.dc.b	$7D,$A7,$FB,$CB,$F5,$4F,$F0,$C4,$FF,$4C,$8C,$63,$FD,$41,$B1,$91
	.dc.b	$3C,$1F,$09,$0F,$C5,$E7,$F1,$5B,$A7,$E1,$76,$7D,$B2,$7F,$94,$63
	.dc.b	$FD,$7A,$31,$DD,$D6,$3F,$B2,$C6,$1D,$4F,$03,$97,$CE,$BD,$96,$9F
	.dc.b	$EB,$FA,$FF,$5A,$4E,$63,$15,$C2,$B6,$FF,$90,$AB,$71,$8E,$1F,$16
	.dc.b	$4F,$A0,$FF,$28,$D7,$F7,$46,$DF,$F7,$46,$BF,$DA,$5A,$FC,$2A,$FD
	.dc.b	$87,$3F,$E8,$63,$E4,$E5,$E5,$7C,$47,$97,$2F,$3A,$8D,$57,$D1,$6E
	.dc.b	$1F,$83,$9E,$C5,$BE,$CE,$0F,$C3,$10,$FB,$1E,$CF,$FB,$A5,$7B,$FF
	.dc.b	$E6,$17,$FD,$DC,$63,$0F,$F6,$F4,$EE,$74,$D8,$CF,$3B,$DA,$FB,$D5
	.dc.b	$A8,$C5,$F3,$3F,$05,$7C,$5B,$1A,$9F,$16,$1F,$E8,$59,$FE,$AC,$43
	.dc.b	$FC,$10,$53,$F8,$71,$4F,$DC,$1F,$EF,$53,$E5,$6D,$F8,$07,$E5,$37
	.dc.b	$2F,$C0,$E9,$FE,$91,$9F,$A4,$7F,$9E,$0E,$A7,$80,$86,$0F,$C1,$0F
	.dc.b	$91,$FC,$4E,$DF,$9F,$6C,$66,$BF,$14,$D3,$B9,$E4,$AF,$13,$E8,$6E
	.dc.b	$DF,$87,$8F,$CA,$F0,$7E,$AF,$E2,$2C,$1A,$4B,$F3,$C7,$C4,$29,$FB
	.dc.b	$63,$FC,$A3,$7F,$B5,$FC,$97,$F4,$2D,$9F,$E0,$92,$9C,$B6,$3F,$88
	.dc.b	$BF,$04,$B8,$FF,$97,$39,$FE,$14,$60,$F0,$EB,$F7,$5B,$CD,$FE,$66
	.dc.b	$7F,$91,$8F,$D5,$FF,$93,$B4,$CB,$83,$FC,$D3,$D7,$DA,$3F,$CC,$ED
	.dc.b	$7F,$CB,$8D,$C7,$F9,$BC,$79,$CF,$B4,$7C,$52,$9F,$14,$FE,$8C,$3F
	.dc.b	$56,$FE,$94,$D3,$2E,$0F,$EA,$CB,$E1,$A9,$1F,$EB,$06,$4F,$F3,$FF
	.dc.b	$81,$E5,$E9,$34,$91,$E7,$3C,$C4,$3C,$C7,$AA,$FF,$B9,$BF,$A1,$3F
	.dc.b	$CF,$9E,$0D,$69,$98,$DB,$FA,$93,$99,$D7,$4F,$AC,$FC,$46,$B9,$D3
	.dc.b	$EC,$3F,$E9,$6F,$0B,$B5,$7F,$86,$23,$B6,$3F,$E1,$86,$CE,$A3,$FC
	.dc.b	$41,$1E,$D9,$FF,$52,$F8,$7F,$0C,$91,$F0,$4F,$29,$30,$F1,$BE,$8F
	.dc.b	$1B,$FC,$AD,$FA,$FF,$F9,$14,$91,$FE,$37,$F3,$7F,$A9,$47,$F8,$FF
	.dc.b	$F5,$E9,$3F,$E4,$36,$BF,$E4,$FF,$27,$FD,$1F,$FB,$07,$F9,$3C,$91
	.dc.b	$CC,$3F,$E5,$9F,$22,$47,$F9,$65,$AF,$3C,$3F,$B1,$7E,$99,$FE,$59
	.dc.b	$24,$73,$7F,$CE,$84,$7F,$D4,$CD,$7F,$CF,$87,$80,$FF,$40,$14,$CD
	.dc.b	$7F,$A1,$5C,$1F,$EB,$DF,$1B,$C2,$FF,$A4,$44,$FF,$A3,$9A,$65,$FF
	.dc.b	$3A,$11,$FF,$3A,$35,$FF,$7F,$0F,$C0,$3F,$98,$6F,$FE,$A7,$6F,$A3
	.dc.b	$FE,$F2,$6F,$A9,$FE,$03,$7F,$D7,$04,$AF,$F0,$30,$A5,$3D,$6E,$13
	.dc.b	$A9,$CC,$62,$3E,$71,$FA,$7E,$0F,$F4,$C7,$6F,$FA,$1C,$91,$FE,$D1
	.dc.b	$7F,$D4,$44,$70,$75,$B7,$AD,$F2,$23,$FC,$35,$87,$92,$31,$1E,$79
	.dc.b	$DB,$F2,$8D,$B8,$CF,$FD,$E9,$0F,$CE,$29,$1F,$F7,$B1,$93,$DD,$CC
	.dc.b	$F6,$FF,$C5,$53,$FF,$00,$09,$3A,$8C,$BF,$77,$2F,$B8,$47,$07,$FA
	.dc.b	$B5,$FF,$58,$11,$C1,$F0,$7D,$CF,$F1,$C7,$E1,$1F,$69,$B1,$91,$19
	.dc.b	$69,$FD,$C7,$DE,$79,$FC,$23,$EE,$C6,$1D,$98,$8D,$3F,$E4,$78,$D4
	.dc.b	$FE,$2B,$DF,$CE,$FF,$84,$6E,$C6,$98,$E3,$FF,$28,$76,$FE,$43,$1F
	.dc.b	$E2,$47,$E6,$1C,$1F,$96,$E3,$DC,$7E,$05,$AF,$DA,$6B,$F7,$DB,$87
	.dc.b	$E5,$BE,$07,$FC,$BD,$FA,$DF,$E8,$1B,$5F,$8B,$6B,$F8,$46,$BF,$89
	.dc.b	$F8,$4B,$5F,$A0,$6E,$66,$98,$F8,$5F,$EB,$90,$69,$2B,$F9,$46,$DF
	.dc.b	$3F,$87,$FA,$23,$63,$EB,$70,$7F,$67,$51,$97,$F3,$3C,$DD,$88,$FF
	.dc.b	$37,$C3,$F3,$F7,$FD,$3F,$C0,$78,$4E,$BF,$F2,$0D,$FF,$B0,$FD,$4D
	.dc.b	$4E,$5B,$3A,$9F,$F4,$37,$C4,$3E,$23,$FA,$C4,$23,$0F,$F5,$B6,$BF
	.dc.b	$D7,$DA,$FF,$90,$5A,$FF,$0C,$6B,$DA,$3A,$C6,$23,$B4,$60,$8F,$76
	.dc.b	$BD,$DA,$F7,$6B,$DD,$AF,$73,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0
	.dc.b	$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$7C,$63,$C8,$53
	.dc.b	$B9,$4E,$85,$39,$94,$E7,$1F,$44,$91,$F2,$23,$44,$8C,$BE,$06,$BE
	.dc.b	$06,$BE,$06,$BE,$06,$BE,$08,$ED,$18,$8E,$D1,$82,$3D,$DA,$F7,$6B
	.dc.b	$DD,$AF,$76,$BD,$CE,$44,$39,$14,$E4,$53,$91,$4E,$47,$02,$1C,$0F
	.dc.b	$F2,$07,$AD,$FE,$BD,$7F,$D4,$4D,$7F,$D6,$F7,$FE,$43,$8C,$47,$EC
	.dc.b	$39,$FF,$DE,$DB,$FE,$AD,$B1,$FA,$B6,$BF,$AB,$6B,$C4,$FF,$65,$16
	.dc.b	$39,$38,$3F,$75,$1C,$1A,$67,$0F,$E7,$4D,$8D,$46,$27,$99,$27,$C8
	.dc.b	$6C,$8E,$5F,$31,$83,$6E,$5F,$31,$82,$1C,$1D,$3E,$66,$C7,$8E,$36
	.dc.b	$72,$38,$18,$38,$15,$E0,$E8,$87,$12,$1C,$0C,$94,$CB,$C9,$B3,$F1
	.dc.b	$88,$FB,$CE,$31,$B7,$91,$1C,$12,$38,$CF,$B9,$B3,$E9,$6B,$D1,$AF
	.dc.b	$88,$F0,$16,$39,$38,$3C,$11,$C1,$A6,$67,$C4,$D8,$D4,$62,$79,$92
	.dc.b	$7C,$86,$C8,$E5,$F3,$18,$36,$E5,$F3,$18,$21,$C1,$D3,$E6,$6C,$7C
	.dc.b	$BC,$FF,$AF,$E4,$F8,$8F,$EB,$64,$BF,$F2,$D8,$9D,$C7,$F7,$2B,$4F
	.dc.b	$13,$FB,$99,$4F,$F1,$21,$4F,$F1,$23,$C2,$3F,$CD,$FD,$AF,$8E,$E7
	.dc.b	$F0,$B8,$6D,$FF,$92,$9A,$FF,$96,$A7,$82,$9F,$EE,$77,$23,$D3,$87
	.dc.b	$BA,$31,$1F,$E7,$6F,$85,$7C,$E7,$EE,$DB,$1F,$BF,$6C,$FF,$B4,$DB
	.dc.b	$1F,$E5,$CF,$80,$F8,$3F,$77,$EA,$6C,$64,$46,$5A,$7F,$99,$3E,$6F
	.dc.b	$6F,$AD,$AF,$FA,$27,$34,$63,$AD,$18,$9E,$67,$C3,$AF,$A7,$F1,$7F
	.dc.b	$31,$B1,$FE,$72,$6B,$B1,$1F,$EA,$0F,$5B,$FD,$32,$D8,$D9,$FE,$9B
	.dc.b	$8C,$BF,$E9,$96,$BF,$E8,$C5,$F2,$3F,$84,$F9,$7E,$8D,$9E,$0D,$78
	.dc.b	$35,$FC,$5F,$D4,$BE,$57,$10,$8C,$2A,$FC,$AB,$99,$DC,$EA,$7D,$9E
	.dc.b	$6F,$5D,$18,$FF,$AF,$F0,$E4,$7A,$E7,$51,$9A,$F8,$9F,$86,$FE,$42
	.dc.b	$BE,$13,$6F,$9B,$68,$EE,$5C,$1F,$88,$7F,$21,$C5,$F8,$F2,$3F,$D4
	.dc.b	$2F,$CB,$F8,$37,$EB,$3F,$0F,$FA,$8D,$9F,$A0,$D7,$E8,$35,$FA,$DF
	.dc.b	$86,$DA,$F7,$B7,$23,$4C,$7E,$17,$E9,$25,$7E,$AC,$FB,$CF,$AF,$9F
	.dc.b	$F5,$51,$FA,$CF,$07,$FC,$CD,$D1,$FA,$C8,$FE,$07,$3A,$7D,$E7,$EA
	.dc.b	$F4,$F7,$CF,$EA,$DF,$EE,$39,$7D,$4F,$37,$F8,$5F,$E1,$9F,$93,$6D
	.dc.b	$FE,$3D,$AF,$E3,$DA,$FE,$57,$F0,$2B,$5F,$79,$BA,$9A,$63,$E9,$BE
	.dc.b	$C1,$2B,$E0,$1F,$9B,$7E,$D7,$0F,$D9,$8F,$E3,$ED,$7E,$EE,$5E,$C7
	.dc.b	$F4,$9F,$0B,$F8,$32,$3D,$0F,$D5,$3F,$2F,$6F,$93,$1F,$2F,$C1,$FF
	.dc.b	$07,$6F,$F2,$63,$E5,$CE,$9F,$97,$BF,$C9,$7E,$61,$1F,$DD,$7D,$57
	.dc.b	$E4,$C9,$1E,$1D,$BF,$43,$1E,$CC,$BD,$13,$D0,$FE,$7F,$9F,$BE,$BF
	.dc.b	$4F,$F7,$9F,$9D,$24,$F3,$DD,$1C,$E3,$C4,$72,$BE,$AF,$AB,$4F,$6E
	.dc.b	$7E,$B8,$C6,$7F,$89,$C7,$F6,$BF,$82,$7E,$13,$23,$49,$5F,$A7,$87
	.dc.b	$DA,$EC,$91,$97,$F1,$BF,$57,$FB,$48,$FD,$73,$70,$FE,$BB,$82,$31
	.dc.b	$3F,$D7,$04,$67,$A7,$FC,$81,$C1,$F5,$67,$FB,$26,$7E,$B6,$E9,$EE
	.dc.b	$3C,$DA,$7A,$32,$FE,$5C,$D9,$FC,$04,$7E,$A5,$E2,$FC,$3D,$1F,$82
	.dc.b	$FC,$32,$4F,$C1,$7E,$37,$DE,$7F,$72,$FC,$9F,$F9,$D3,$67,$F9,$B3
	.dc.b	$5F,$E6,$CD,$7F,$9E,$78,$5F,$DE,$AF,$FA,$13,$63,$F9,$F3,$5F,$E7
	.dc.b	$CD,$7E,$C9,$EA,$21,$F2,$74,$FC,$D6,$DF,$CC,$3C,$1F,$B8,$49,$FB
	.dc.b	$4F,$DF,$FA,$3F,$E5,$44,$8F,$C7,$3F,$A7,$DB,$E6,$6B,$F1,$F0,$EB
	.dc.b	$93,$EE,$BF,$D0,$39,$7F,$55,$6E,$3F,$D5,$5A,$FF,$5B,$6C,$7F,$44
	.dc.b	$F9,$EF,$F0,$8B,$FD,$75,$B3,$FD,$75,$AF,$F6,$36,$BF,$60,$F8,$27
	.dc.b	$C4,$3E,$17,$0F,$E7,$E5,$FB,$8F,$E3,$25,$95,$7F,$1F,$19,$8C,$4F
	.dc.b	$03,$CF,$19,$7F,$2D,$F2,$6F,$D7,$1F,$2B,$EE,$8C,$36,$33,$87,$82
	.dc.b	$B8,$C6,$30,$EF,$83,$ED,$FD,$F5,$E8,$6C,$66,$FF,$C6,$13,$FA,$73
	.dc.b	$4E,$0F,$83,$BF,$D8,$7E,$29,$F1,$CF,$F2,$4E,$5E,$67,$C8,$63,$1F
	.dc.b	$25,$79,$B2,$F5,$1F,$6A,$79,$1F,$3C,$86,$09,$BF,$ED,$C9,$A7,$ED
	.dc.b	$FD,$47,$CC,$BD,$6F,$F2,$DF,$CC,$FD,$4C,$3F,$CB,$B5,$A9,$CF,$07
	.dc.b	$FB,$4A,$55,$B5,$5B,$73,$3B,$9D,$4E,$5B,$7E,$1A,$D9,$E0,$71,$1D
	.dc.b	$71,$FF,$3B,$CF,$F8,$81,$AF,$F4,$93,$18,$B6,$FF,$12,$C7,$F1,$43
	.dc.b	$2E,$27,$8C,$7D,$63,$FC,$FF,$B7,$D3,$7F,$28,$DD,$7E,$94,$7F,$A0
	.dc.b	$1A,$FE,$64,$EC,$78,$C8,$75,$3F,$0D,$BF,$E2,$7D,$9F,$C7,$F0,$FF
	.dc.b	$A2,$9B,$26,$9E,$2E,$1F,$CC,$C5,$E1,$FC,$83,$F8,$76,$AC,$7D,$0F
	.dc.b	$9F,$0F,$D5,$B7,$1D,$9F,$AD,$7D,$86,$4A,$7F,$59,$3E,$49,$0F,$69
	.dc.b	$FC,$0C,$F4,$3F,$C2,$5A,$7F,$09,$29,$BC,$3F,$7B,$C9,$FC,$66,$3B
	.dc.b	$CF,$F9,$5B,$93,$CB,$39,$9D,$C7,$B9,$F9,$E7,$CD,$21,$E7,$3F,$CE
	.dc.b	$DC,$BF,$35,$A7,$CD,$2D,$FF,$A0,$63,$51,$FE,$7B,$E6,$FC,$4E,$1F
	.dc.b	$CE,$63,$FC,$DE,$D7,$FD,$11,$1F,$8E,$7C,$67,$84,$F1,$98,$8F,$F3
	.dc.b	$8F,$3F,$EB,$A3,$FA,$69,$FA,$A2,$F1,$7F,$52,$70,$7F,$58,$1F,$94
	.dc.b	$FF,$5E,$69,$FD,$78,$A6,$DF,$F4,$E7,$5F,$AB,$4F,$ED,$73,$EA,$6B
	.dc.b	$A3,$F4,$9C,$9F,$A9,$7E,$F1,$EA,$E3,$FD,$3B,$FE,$A3,$EC,$E8,$47
	.dc.b	$A3,$F0,$E3,$71,$F1,$08,$F2,$CF,$FC,$19,$1F,$33,$BB,$FC,$2C,$FF
	.dc.b	$85,$1A,$FF,$2D,$6B,$FE,$AF,$EF,$F6,$11,$F6,$3E,$37,$FC,$1E,$79
	.dc.b	$08,$7B,$0F,$AC,$43,$CD,$E1,$7F,$B8,$9F,$F2,$71,$0F,$F1,$E1,$F2
	.dc.b	$7C,$2F,$C3,$BF,$EC,$56,$DF,$F9,$22,$3F,$1C,$7F,$B1,$63,$6F,$F9
	.dc.b	$A0,$FE,$2B,$3D,$DE,$A7,$73,$FD,$99,$E1,$F4,$D0,$D2,$47,$63,$FC
	.dc.b	$EB,$E1,$FE,$4E,$BA,$47,$FA,$48,$E8,$7B,$F3,$FA,$A4,$D7,$F0,$11
	.dc.b	$FE,$7A,$3A,$65,$FE,$C8,$F1,$3C,$44,$8F,$01,$EB,$3A,$94,$F5,$9F
	.dc.b	$EE,$7F,$17,$D4,$47,$AE,$68,$F7,$78,$C9,$7E,$B9,$FE,$F5,$8C,$36
	.dc.b	$32,$7F,$BC,$E3,$2F,$FA,$9C,$C9,$FE,$A7,$29,$FE,$FB,$F2,$36,$E6
	.dc.b	$7D,$3A,$7C,$88,$F9,$67,$D6,$D7,$F3,$9A,$7C,$68,$FD,$23,$5F,$C5
	.dc.b	$E9,$FE,$EE,$F2,$BE,$C1,$23,$DE,$FD,$BF,$2F,$DD,$F8,$7F,$F9,$CB
	.dc.b	$6F,$F6,$AD,$23,$13,$A8,$FF,$07,$4E,$A7,$FD,$CB,$1B,$3F,$DC,$A5
	.dc.b	$3F,$DC,$A5,$3F,$CF,$70,$F3,$7A,$47,$E1,$4F,$AE,$52,$1F,$F0,$00
	.dc.b	$FA,$FE,$67,$E4,$C8,$FD,$AF,$43,$EC,$7E,$17,$EB,$6C,$7A,$1A,$FF
	.dc.b	$69,$B4,$62,$3F,$01,$E9,$3B,$F1,$08,$C2,$BF,$F4,$B3,$C2,$77,$3A
	.dc.b	$9F,$99,$E9,$F5,$D9,$8A,$30,$DC,$3A,$1F,$3A,$32,$FE,$4F,$E1,$BE
	.dc.b	$B3,$4F,$A6,$53,$D8,$53,$FD,$53,$EB,$3D,$6C,$60,$8C,$4A,$BF,$D6
	.dc.b	$71,$86,$D7,$08,$7A,$FF,$A5,$E4,$DB,$6F,$FC,$A5,$3A,$7F,$86,$61
	.dc.b	$DE,$75,$1F,$E1,$99,$7B,$18,$8F,$3B,$6B,$CC,$D9,$DC,$FF,$0C,$E6
	.dc.b	$FD,$33,$FB,$F3,$F0,$E5,$3F,$87,$73,$7E,$95,$FE,$82,$FE,$9E,$FE
	.dc.b	$17,$B5,$FC,$A2,$CD,$25,$7F,$0E,$F8,$1F,$F2,$A7,$A9,$F8,$08,$46
	.dc.b	$1F,$C1,$B5,$FF,$6D,$35,$FF,$79,$47,$F9,$08,$8F,$E2,$BB,$D8,$FF
	.dc.b	$92,$67,$B7,$BA,$F7,$BF,$BF,$36,$FF,$C9,$AD,$77,$1A,$23,$A6,$BA
	.dc.b	$30,$43,$04,$3F,$3B,$1D,$5B,$1C,$9A,$71,$29,$C4,$FF,$7D,$47,$16
	.dc.b	$C7,$16,$9C,$4A,$71,$3F,$99,$47,$16,$C7,$16,$9C,$4A,$71,$11,$C5
	.dc.b	$B1,$C5,$A7,$12,$9C,$63,$81,$23,$83,$63,$83,$5E,$11,$A2,$3A,$6B
	.dc.b	$A3,$04,$30,$43,$E5,$90,$F8,$A5,$3D,$E5,$3D,$25,$8F,$23,$63,$C2
	.dc.b	$D3,$C0,$53,$C0,$76,$21,$D8,$A7,$62,$C7,$66,$C7,$66,$9D,$8A,$76
	.dc.b	$3A,$10,$E8,$58,$E8,$D8,$E8,$D3,$FD,$5B,$EA,$FF,$AF,$DF,$E6,$6D
	.dc.b	$7F,$D8,$C6,$3D,$5F,$F5,$F3,$FC,$E3,$F1,$AF,$F6,$8B,$FE,$C9,$3F
	.dc.b	$C0,$0E,$3D,$7F,$DF,$36,$38,$4F,$F8,$16,$72,$D8,$C9,$CC,$4F,$17
	.dc.b	$FC,$18,$62,$38,$38,$8D,$3E,$28,$CB,$4E,$91,$A7,$91,$88,$DB,$89
	.dc.b	$F2,$B9,$23,$82,$3E,$86,$BE,$38,$E7,$3D,$C8,$E0,$91,$D0,$E4,$43
	.dc.b	$91,$92,$CF,$67,$D8,$D7,$D0,$71,$32,$53,$27,$23,$13,$E4,$7B,$1C
	.dc.b	$0D,$10,$C1,$0E,$65,$79,$47,$8D,$B1,$DA,$7D,$F3,$96,$C6,$4F,$09
	.dc.b	$D0,$93,$C9,$F1,$18,$8E,$2E,$23,$4F,$8E,$32,$D3,$AC,$69,$E4,$62
	.dc.b	$36,$E2,$7C,$CE,$48,$E0,$8F,$A5,$AF,$8E,$39,$CF,$72,$38,$24,$7E
	.dc.b	$7F,$3F,$EB,$3E,$BB,$DF,$FE,$CD,$18,$6E,$A9,$D5,$7F,$61,$8C,$4F
	.dc.b	$F6,$DD,$99,$7F,$9C,$E3,$0D,$AF,$ED,$ED,$7F,$B7,$B5,$FF,$3D,$7D
	.dc.b	$5F,$F7,$C7,$B9,$FB,$27,$13,$FD,$ED,$B1,$FE,$3C,$6B,$FE,$83,$F5
	.dc.b	$9E,$AF,$ED,$9B,$A7,$ED,$9B,$3F,$C2,$DB,$87,$A2,$3F,$82,$63,$EC
	.dc.b	$6C,$7E,$E2,$30,$FF,$93,$3E,$C5,$F9,$1F,$F0,$46,$E1,$FC,$11,$B7
	.dc.b	$FC,$65,$B5,$FE,$54,$E6,$7B,$1F,$EF,$DE,$CC,$36,$FF,$89,$CE,$BD
	.dc.b	$5F,$F5,$05,$69,$B3,$AA,$6D,$88,$FF,$53,$5F,$68,$C4,$23,$0F,$F9
	.dc.b	$38,$DC,$F2,$70,$6D,$FF,$10,$AF,$BE,$F3,$F0,$FF,$AC,$DB,$19,$3C
	.dc.b	$91,$97,$CD,$F6,$D5,$EE,$D7,$B3,$6F,$E2,$B6,$DF,$0B,$FE,$C1,$F8
	.dc.b	$4F,$8E,$DB,$F8,$4D,$7D,$6D,$7F,$D9,$1C,$DF,$A2,$FC,$CB,$F3,$FC
	.dc.b	$E6,$92,$7E,$67,$CE,$FF,$B3,$E3,$EA,$B6,$32,$71,$2C,$7F,$A4,$39
	.dc.b	$BF,$27,$F4,$FF,$91,$E8,$34,$95,$F8,$5E,$7F,$C5,$13,$9F,$F1,$4F
	.dc.b	$F5,$0E,$0F,$C3,$FD,$D7,$FB,$87,$C3,$FC,$3B,$6F,$F0,$4D,$7E,$DB
	.dc.b	$5F,$B5,$E2,$DA,$FC,$8B,$78,$8D,$38,$69,$FA,$6A,$FE,$CA,$D7,$F6
	.dc.b	$1A,$7E,$BC,$FE,$C1,$C1,$FC,$2E,$3F,$84,$70,$7E,$F1,$FA,$3C,$9F
	.dc.b	$C2,$9C,$F6,$7F,$AD,$BC,$17,$EC,$3F,$EA,$08,$C3,$67,$FB,$AC,$EB
	.dc.b	$AF,$F8,$53,$63,$B6,$DF,$AC,$C1,$0F,$C2,$7F,$2A,$47,$F2,$A6,$4F
	.dc.b	$CA,$94,$FC,$A9,$4F,$CA,$96,$7F,$2A,$DD,$FF,$84,$65,$FB,$AE,$DF
	.dc.b	$E3,$D5,$FB,$E6,$CE,$AB,$F7,$1F,$35,$82,$3E,$A8,$47,$0F,$8E,$93
	.dc.b	$96,$E1,$F9,$AF,$C0,$64,$FD,$06,$BF,$0B,$4F,$8D,$1F,$33,$5F,$8A
	.dc.b	$D9,$34,$F5,$38,$3F,$87,$65,$C1,$A6,$4F,$E1,$2E,$5F,$06,$DF,$ED
	.dc.b	$AF,$C0,$3B,$12,$3E,$DC,$65,$FE,$35,$F2,$3F,$63,$F0,$EF,$E4,$6D
	.dc.b	$8C,$88,$CB,$4E,$78,$FE,$03,$F0,$1F,$61,$23,$21,$FE,$03,$8C,$8F
	.dc.b	$CB,$B5,$FE,$47,$F2,$9F,$E3,$EE,$1F,$E6,$ED,$E1,$34,$C7,$CA,$CA
	.dc.b	$FF,$2C,$57,$F3,$E8,$C5,$FF,$41,$FC,$19,$F1,$7D,$5B,$7C,$1B,$3B
	.dc.b	$9F,$F2,$4F,$47,$F7,$A8,$F3,$57,$EB,$3A,$3F,$BF,$74,$7F,$79,$9F
	.dc.b	$E9,$7F,$31,$E7,$25,$79,$35,$47,$4A,$E1,$3F,$08,$87,$F9,$93,$0F
	.dc.b	$03,$FD,$3F,$E6,$7E,$11,$1E,$8F,$C5,$F3,$FF,$63,$F8,$5F,$F5,$C6
	.dc.b	$CF,$F5,$76,$BF,$D4,$1A,$FF,$4F,$F9,$BF,$DC,$FB,$3F,$D9,$1B,$1F
	.dc.b	$D7,$DA,$FF,$58,$6B,$F0,$F4,$F8,$AB,$1F,$4B,$EB,$CB,$DE,$FF,$60
	.dc.b	$FA,$FF,$B8,$47,$E3,$3F,$81,$FD,$67,$FC,$49,$0E,$AF,$EF,$51,$86
	.dc.b	$F0,$7F,$10,$6C,$EC,$F9,$67,$94,$A7,$96,$FC,$6F,$F7,$5A,$7C,$F7
	.dc.b	$FA,$6F,$97,$FC,$07,$18,$9F,$F1,$8C,$62,$75,$1F,$E0,$09,$D6,$7F
	.dc.b	$DF,$E3,$1F,$73,$FC,$93,$F0,$DC,$47,$78,$C6,$1D,$9B,$1C,$04,$F0
	.dc.b	$7D,$19,$FE,$40,$97,$F9,$4B,$60,$F6,$FF,$99,$FD,$6F,$82,$73,$18
	.dc.b	$7D,$AD,$C3,$D4,$D5,$E3,$BE,$C9,$F3,$1F,$02,$73,$18,$7E,$29,$FE
	.dc.b	$45,$71,$97,$C0,$AE,$55,$C7,$0F,$3F,$DB,$FF,$48,$FA,$1F,$36,$88
	.dc.b	$C4,$FD,$26,$E5,$F3,$DA,$FD,$D3,$F1,$DA,$FC,$47,$F0,$38,$F8,$1F
	.dc.b	$29,$E8,$C7,$D5,$19,$CB,$CD,$9F,$A3,$1F,$C3,$F2,$7E,$6B,$0E,$C4
	.dc.b	$3E,$9D,$7E,$82,$35,$18,$3F,$DD,$DE,$37,$F6,$4F,$67,$FC,$BD,$0F
	.dc.b	$75,$E9,$2B,$58,$7F,$B9,$27,$31,$8A,$FE,$DB,$18,$57,$D2,$30,$DB
	.dc.b	$E9,$3A,$BE,$91,$89,$D5,$F5,$AF,$C6,$9F,$E3,$62,$1F,$E0,$C2,$9F
	.dc.b	$38,$A7,$F2,$33,$FC,$65,$97,$E0,$DF,$97,$A7,$F9,$B1,$FC,$23,$5F
	.dc.b	$A4,$7F,$9A,$0E,$87,$72,$18,$3F,$00,$3E,$6B,$F0,$FB,$FF,$84,$1B
	.dc.b	$19,$D7,$F1,$0D,$3C,$27,$73,$C2,$79,$8C,$6D,$F8,$77,$F2,$7C,$5F
	.dc.b	$E7,$C9,$FD,$3C,$7F,$9E,$DA,$69,$C1,$E7,$3E,$21,$EA,$29,$FB,$DF
	.dc.b	$C0,$FF,$3F,$CD,$FE,$28,$6C,$1A,$7E,$93,$9F,$E4,$57,$EF,$9B,$C9
	.dc.b	$FC,$56,$B4,$7B,$4F,$B6,$78,$4A,$79,$67,$F8,$3F,$27,$F8,$E3,$E9
	.dc.b	$FF,$94,$92,$7F,$90,$C6,$5F,$F5,$07,$5F,$F8,$F9,$FF,$54,$B5,$C9
	.dc.b	$4F,$F4,$D3,$5F,$82,$43,$C6,$62,$BF,$D6,$BD,$9F,$A7,$8F,$EA,$07
	.dc.b	$E7,$8B,$C7,$FD,$55,$C1,$FD,$6C,$7E,$19,$FD,$85,$A7,$F6,$12,$9B
	.dc.b	$7F,$D8,$3E,$0F,$96,$FF,$D8,$AD,$F0,$CD,$3D,$55,$E8,$27,$57,$FB
	.dc.b	$F9,$FD,$1D,$7F,$B3,$BC,$3E,$44,$8F,$89,$19,$7E,$24,$F1,$7E,$B3
	.dc.b	$5F,$AD,$E2,$FF,$85,$9F,$F0,$A1,$C9,$C3,$FE,$D8,$6B,$FE,$CC,$F1
	.dc.b	$BD,$44,$7C,$38,$A3,$DD,$E3,$C9,$CB,$5F,$DE,$9E,$EF,$0B,$FB,$9B
	.dc.b	$F9,$BE,$CF,$F4,$17,$E0,$7F,$E5,$B2,$47,$F9,$01,$FF,$23,$F4,$C8
	.dc.b	$FF,$25,$B5,$FF,$2B,$F9,$5F,$8D,$68,$E5,$FF,$33,$11,$FF,$30,$FC
	.dc.b	$07,$F9,$98,$8E,$0F,$F2,$7B,$FE,$6D,$23,$83,$FD,$4A,$FF,$B4,$88
	.dc.b	$E0,$FF,$7E,$7A,$3F,$E7,$96,$9F,$CA,$8F,$F7,$BB,$96,$99,$3F,$DF
	.dc.b	$1E,$96,$5A,$67,$85,$BF,$FA,$3A,$30,$DD,$3F,$D0,$24,$8F,$F3,$FB
	.dc.b	$5F,$F5,$2F,$EB,$5F,$AF,$AF,$F5,$1B,$69,$EF,$FF,$AC,$A3,$0D,$AF
	.dc.b	$F5,$8C,$6F,$7C,$47,$FA,$E2,$FF,$D7,$4A,$74,$D3,$FC,$F7,$87,$FA
	.dc.b	$64,$8E,$0F,$F6,$8B,$FE,$FE,$23,$83,$FD,$B7,$EB,$7F,$60,$7F,$83
	.dc.b	$6E,$8C,$47,$FB,$ED,$AF,$0F,$17,$F6,$9E,$BF,$FB,$3B,$DD,$F8,$0D
	.dc.b	$9F,$E2,$99,$7F,$BB,$94,$E9,$A7,$FB,$E0,$A7,$FB,$E0,$A2,$3F,$E0
	.dc.b	$0B,$86,$C7,$F1,$FE,$07,$B7,$E6,$AF,$F7,$0C,$63,$0F,$E4,$0D,$8F
	.dc.b	$E4,$8D,$7D,$DE,$E7,$F2,$5C,$FE,$13,$73,$34,$FE,$29,$8F,$FB,$BB
	.dc.b	$F5,$6F,$C0,$36,$FE,$0F,$0F,$BA,$73,$07,$F9,$A7,$53,$85,$E0,$E6
	.dc.b	$BF,$9D,$4E,$63,$11,$FE,$2C,$AE,$35,$C1,$FC,$87,$BF,$FF,$00,$3E
	.dc.b	$2E,$D1,$A4,$7F,$DF,$FF,$B1,$FE,$51,$34,$8F,$AA,$CE,$8C,$14,$FC
	.dc.b	$C6,$5E,$B7,$07,$D9,$71,$3F,$A5,$F8,$4E,$7C,$96,$8D,$23,$F2,$BC
	.dc.b	$3D,$B4,$D9,$63,$F6,$11,$A8,$FC,$0F,$27,$E0,$D7,$1F,$DF,$7F,$B4
	.dc.b	$35,$DF,$27,$F4,$CF,$85,$FF,$22,$C7,$F6,$B6,$BF,$DB,$DA,$FC,$46
	.dc.b	$F2,$FE,$A6,$7F,$10,$49,$FD,$6F,$2F,$8D,$66,$F8,$8F,$E2,$D1,$86
	.dc.b	$C6,$63,$F8,$E1,$92,$E0,$F4,$78,$B6,$33,$18,$3A,$F2,$FE,$2F,$A7
	.dc.b	$F8,$1C,$FF,$9D,$9A,$FD,$EF,$87,$F3,$13,$3F,$ED,$ED,$8F,$F0,$53
	.dc.b	$5F,$F0,$73,$5F,$F0,$93,$5F,$F0,$B3,$5E,$D1,$D0,$8F,$46,$BD,$1A
	.dc.b	$F4,$6B,$D1,$AF,$43,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0,$87,$02
	.dc.b	$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$23,$E4,$36,$3D,$ED,$3D
	.dc.b	$05,$3C,$65,$3C,$05,$3B,$94,$EF,$1D,$89,$1D,$9B,$1D,$9A,$F6,$6B
	.dc.b	$D9,$AF,$66,$BD,$A3,$A1,$1E,$8D,$7A,$35,$E8,$D7,$A3,$5E,$87,$22
	.dc.b	$1C,$8F,$8B,$E9,$FF,$AD,$DF,$F6,$19,$FE,$65,$70,$FF,$AD,$2B,$F3
	.dc.b	$BD,$5F,$C8,$31,$FC,$EB,$67,$F3,$AD,$7F,$3A,$D7,$F5,$4D,$7E,$D3
	.dc.b	$70,$FE,$97,$1F,$C3,$1B,$3D,$A3,$13,$B2,$38,$24,$6C,$8F,$90,$C9
	.dc.b	$5C,$BE,$63,$25,$72,$FA,$0C,$95,$CB,$EA,$32,$57,$33,$EA,$8F,$01
	.dc.b	$0E,$C5,$3A,$14,$E4,$57,$93,$C0,$87,$02,$9C,$0A,$F0,$74,$43,$45
	.dc.b	$74,$E0,$8E,$09,$3F,$42,$7D,$AD,$9F,$33,$5F,$0B,$5E,$CD,$7A,$B6
	.dc.b	$7A,$C7,$CA,$6C,$F6,$8C,$4E,$C8,$E0,$91,$B2,$3E,$43,$25,$72,$F9
	.dc.b	$8C,$95,$CB,$E8,$32,$7F,$80,$31,$FF,$09,$7C,$0F,$F8,$43,$A8,$D2
	.dc.b	$4F,$FB,$47,$F1,$D7,$C7,$FF,$82,$27,$FD,$A5,$18,$9E,$26,$E7,$4D
	.dc.b	$32,$7F,$B7,$7E,$29,$D9,$A8,$C3,$5F,$F0,$AC,$69,$E2,$62,$34,$B8
	.dc.b	$3F,$B3,$7C,$27,$FB,$95,$B1,$FD,$FF,$4F,$F0,$FC,$E5,$B1,$FE,$5C
	.dc.b	$F4,$1F,$67,$FE,$24,$F6,$BF,$D0,$51,$89,$46,$27,$51,$C5,$AF,$F8
	.dc.b	$2E,$AB,$FE,$0B,$6B,$E7,$6E,$1F,$DF,$6B,$BF,$C7,$70,$F1,$18,$A3
	.dc.b	$0D,$C3,$C6,$7A,$63,$3F,$69,$FE,$3C,$F8,$FF,$4B,$67,$DC,$D7,$FD
	.dc.b	$29,$48,$C4,$7F,$9D,$61,$F6,$B8,$C3,$C6,$D8,$CC,$AF,$E3,$CE,$7A
	.dc.b	$7F,$C9,$BE,$5F,$CA,$6D,$7C,$F6,$BE,$26,$BE,$1F,$B2,$FF,$2A,$FD
	.dc.b	$67,$D2,$6C,$7D,$76,$BE,$A6,$BE,$9F,$81,$FF,$2E,$FE,$75,$F6,$9B
	.dc.b	$1F,$7D,$AF,$C2,$6B,$FA,$5F,$27,$FC,$F3,$F8,$E7,$E9,$49,$1F,$E7
	.dc.b	$A8,$CB,$F8,$9E,$DF,$F3,$5C,$FE,$21,$B1,$91,$19,$69,$F9,$1F,$86
	.dc.b	$FB,$A4,$7E,$E9,$F5,$9F,$C0,$FD,$DC,$8C,$9F,$5D,$C7,$27,$E6,$3E
	.dc.b	$F3,$FC,$FB,$F6,$16,$8D,$23,$FA,$1F,$86,$FF,$41,$FE,$3D,$FA,$06
	.dc.b	$C6,$44,$65,$A7,$E8,$47,$E0,$7F,$A5,$69,$97,$07,$F0,$FC,$24,$7C
	.dc.b	$BF,$0F,$D6,$BC,$4F,$F5,$34,$FF,$A5,$A3,$FD,$2F,$E3,$3F,$0C,$7D
	.dc.b	$4F,$EB,$A3,$0D,$8C,$CF,$C6,$9C,$BF,$23,$B3,$F5,$BE,$5F,$FA,$8B
	.dc.b	$F4,$8F,$DF,$B6,$7F,$82,$35,$FD,$3B,$5F,$CB,$0F,$CD,$3F,$2C,$D3
	.dc.b	$F2,$C5,$3F,$2C,$53,$F5,$DC,$7F,$EB,$2D,$BF,$87,$B6,$3F,$08,$7C
	.dc.b	$A3,$F0,$45,$3F,$00,$5C,$3F,$83,$F1,$F1,$8F,$DE,$B5,$DD,$7F,$1C
	.dc.b	$F9,$9F,$F6,$0F,$E0,$D6,$4D,$23,$F5,$67,$FA,$31,$1C,$1F,$57,$4F
	.dc.b	$9B,$1F,$5E,$11,$87,$5A,$FD,$49,$D4,$E6,$7D,$BF,$6B,$F0,$93,$48
	.dc.b	$F2,$DF,$D2,$D3,$D2,$5C,$3A,$E1,$F8,$27,$F9,$DF,$83,$FE,$E0,$FC
	.dc.b	$1A,$D1,$A4,$7A,$7E,$0D,$CB,$0F,$F7,$3B,$60,$D3,$FB,$57,$91,$23
	.dc.b	$F3,$8D,$7F,$97,$FC,$17,$F3,$A2,$3F,$93,$F1,$91,$FD,$07,$F4,$BF
	.dc.b	$D0,$7A,$3F,$2F,$8D,$18,$BF,$D2,$46,$1F,$ED,$0D,$7F,$B3,$AF,$B2
	.dc.b	$7D,$3F,$F9,$6E,$30,$D8,$CC,$FF,$B9,$A7,$2F,$FB,$9B,$AD,$E1,$BB
	.dc.b	$70,$DD,$7D,$D5,$FC,$A6,$79,$64,$FA,$5F,$03,$67,$31,$83,$FC,$D9
	.dc.b	$87,$6A,$EA,$4B,$EB,$8F,$C2,$FA,$7F,$8C,$4C,$3F,$79,$8A,$3E,$56
	.dc.b	$89,$7E,$47,$F2,$EB,$DE,$D9,$D3,$67,$67,$CE,$97,$D4,$F3,$FD,$97
	.dc.b	$93,$67,$4D,$7C,$18,$FD,$88,$C4,$7C,$5C,$BE,$3B,$FC,$7F,$6F,$94
	.dc.b	$FE,$20,$F0,$7D,$67,$E0,$C8,$FB,$9F,$C3,$7D,$6F,$B7,$F0,$9F,$63
	.dc.b	$67,$D0,$D7,$C7,$9A,$31,$1E,$33,$C0,$7F,$2F,$E0,$FA,$58,$FD,$0A
	.dc.b	$7D,$87,$E5,$49,$5F,$41,$FC,$C7,$D8,$FC,$FF,$C2,$5B,$7C,$98,$C3
	.dc.b	$69,$3A,$7C,$B3,$A7,$C5,$AF,$C4,$6F,$1F,$E0,$DF,$39,$DB,$17,$D9
	.dc.b	$FD,$FF,$9A,$E9,$39,$8C,$65,$F6,$DB,$1F,$11,$AB,$E1,$5F,$AB,$BC
	.dc.b	$EF,$F6,$45,$FF,$53,$9C,$C6,$2B,$F1,$D1,$85,$BF,$8A,$30,$F9,$2B
	.dc.b	$F7,$7F,$6B,$FE,$4F,$7F,$1C,$DA,$FA,$ED,$7D,$CD,$7F,$11,$A7,$C3
	.dc.b	$7E,$2D,$7E,$CA,$FB,$DF,$9F,$1F,$55,$79,$23,$18,$FA,$63,$E0,$4E
	.dc.b	$AF,$E4,$C7,$67,$F2,$FC,$3F,$84,$8F,$A2,$64,$AE,$5F,$19,$FE,$F4
	.dc.b	$F8,$AF,$86,$FE,$CF,$E1,$9E,$D7,$F4,$68,$4E,$94,$EB,$C9,$FD,$D4
	.dc.b	$E6,$31,$3C,$30,$FF,$7F,$34,$ED,$7D,$6B,$83,$67,$71,$DF,$0F,$E2
	.dc.b	$15,$FE,$A3,$21,$FE,$15,$29,$F9,$22,$9F,$C7,$CF,$EC,$1A,$7E,$05
	.dc.b	$FC,$06,$5F,$41,$B9,$FF,$38,$7F,$0F,$19,$D3,$B4,$43,$B1,$4F,$C0
	.dc.b	$8F,$A1,$FC,$57,$13,$83,$F8,$F4,$61,$B5,$F8,$CA,$E0,$FE,$01,$FE
	.dc.b	$2A,$78,$B0,$F2,$3F,$82,$6E,$3E,$B9,$FD,$44,$F7,$6D,$78,$B8,$7F
	.dc.b	$97,$36,$74,$7A,$88,$7C,$73,$D9,$3F,$D7,$39,$3F,$9B,$7E,$15,$EC
	.dc.b	$6C,$7F,$07,$8C,$49,$F8,$3C,$FF,$02,$3F,$9B,$35,$FE,$3E,$D7,$F8
	.dc.b	$33,$5F,$D4,$9F,$68,$A7,$93,$4F,$E4,$CB,$F0,$AF,$E5,$AD,$3F,$96
	.dc.b	$94,$FE,$5B,$D3,$FD,$3E,$3F,$A7,$75,$F8,$71,$FE,$F2,$DB,$34,$F2
	.dc.b	$70,$FE,$4B,$0F,$C0,$1E,$43,$E4,$D7,$F4,$B5,$FB,$37,$F5,$26,$9F
	.dc.b	$D4,$8A,$7F,$40,$ED,$FE,$C5,$1F,$D7,$7F,$67,$93,$FD,$77,$C1,$FE
	.dc.b	$CD,$18,$8F,$E8,$FE,$17,$9C,$8F,$E4,$4F,$EC,$8F,$A3,$BB,$FC,$09
	.dc.b	$1F,$2F,$2F,$53,$E8,$D3,$FC,$00,$FF,$71,$F0,$FF,$C8,$CF,$F5,$46
	.dc.b	$BF,$CE,$DA,$FC,$13,$F1,$07,$E2,$8C,$69,$FE,$16,$7F,$C4,$DE,$27
	.dc.b	$F8,$69,$B9,$F4,$9D,$47,$F7,$B3,$FD,$16,$53,$F8,$91,$FD,$98,$C9
	.dc.b	$4C,$9F,$EE,$5F,$1B,$EE,$78,$FF,$E3,$88,$F8,$87,$F8,$F8,$A6,$CA
	.dc.b	$7A,$C9,$97,$C6,$7D,$A7,$F9,$A3,$CD,$FF,$4C,$FE,$17,$FE,$4F,$24
	.dc.b	$7F,$97,$DF,$F2,$CF,$E1,$72,$3F,$C8,$77,$FE,$F3,$51,$97,$FD,$53
	.dc.b	$5F,$EA,$4F,$C2,$BF,$CB,$44,$8F,$F3,$53,$FE,$71,$F8,$19,$1F,$E9
	.dc.b	$26,$BC,$BC,$EF,$F1,$67,$E1,$5F,$E6,$32,$47,$F9,$CD,$FB,$B9,$FF
	.dc.b	$8E,$DB,$C5,$FE,$49,$6F,$90,$69,$FE,$93,$C6,$5F,$B3,$D5,$FF,$1D
	.dc.b	$46,$23,$FD,$3A,$71,$3E,$0F,$AD,$E4,$69,$93,$75,$FE,$8D,$23,$FE
	.dc.b	$AE,$6B,$FE,$B6,$FC,$DF,$FD,$76,$47,$FD,$76,$FF,$B1,$97,$E9,$B4
	.dc.b	$8F,$FB,$20,$C9,$FE,$95,$F1,$BF,$D1,$04,$70,$7F,$B8,$5F,$86,$47
	.dc.b	$07,$FB,$B6,$BF,$DD,$84,$70,$7F,$BD,$7F,$26,$FF,$7A,$11,$C1,$F8
	.dc.b	$4F,$81,$FF,$31,$7E,$0B,$FE,$8B,$E4,$46,$27,$42,$75,$19,$8F,$9F
	.dc.b	$87,$1C,$FF,$42,$48,$FC,$FF,$3B,$07,$B9,$FA,$62,$4E,$A3,$2D,$F8
	.dc.b	$4F,$F3,$B7,$DF,$F9,$5B,$1F,$9C,$6B,$FE,$6C,$94,$62,$3F,$D2,$13
	.dc.b	$F8,$C2,$3F,$ED,$3E,$4F,$F4,$64,$7C,$1F,$86,$FF,$42,$7E,$81,$68
	.dc.b	$D2,$3F,$EF,$AF,$C0,$BF,$DF,$A4,$70,$7E,$C3,$D3,$FF,$47,$78,$7F
	.dc.b	$35,$B1,$93,$FD,$28,$58,$FA,$5F,$19,$FE,$95,$FB,$DB,$26,$91,$FD
	.dc.b	$EF,$C7,$6D,$A6,$CB,$1F,$76,$35,$1F,$1F,$E3,$FF,$8B,$91,$FA,$4F
	.dc.b	$F6,$3F,$90,$FF,$1A,$F2,$7F,$34,$6C,$FF,$BE,$DA,$FF,$AA,$C3,$87
	.dc.b	$FC,$FF,$C2,$F9,$1F,$E4,$A4,$8F,$C0,$3D,$FD,$9F,$F5,$A7,$CD,$77
	.dc.b	$6C,$F7,$6B,$DE,$FF,$C9,$33,$FD,$23,$E5,$5F,$2B,$FC,$DA,$BF,$D7
	.dc.b	$D1,$87,$F9,$B8,$71,$EF,$71,$9E,$12,$F9,$77,$CB,$E1,$B1,$C9,$B5
	.dc.b	$FD,$61,$B1,$F8,$DF,$1B,$FD,$C9,$3F,$C7,$09,$19,$F1,$BC,$7F,$31
	.dc.b	$F9,$88,$FE,$BE,$D7,$FB,$1B,$5F,$9C,$1C,$3F,$45,$AF,$FB,$75,$AE
	.dc.b	$FB,$3F,$B4,$C6,$08,$E0,$91,$E2,$6B,$E2,$6B,$E2,$8E,$2E,$C8,$6C
	.dc.b	$A6,$E3,$8B,$E7,$6B,$E7,$6B,$E7,$38,$9B,$29,$B2,$9B,$3C,$04,$3C
	.dc.b	$06,$A3,$CE,$E4,$E8,$53,$A1,$4E,$87,$23,$0F,$27,$04,$39,$94,$E6
	.dc.b	$53,$99,$A2,$1A,$29,$A3,$04,$30,$47,$C4,$F0,$6B,$B6,$BB,$8F,$9B
	.dc.b	$18,$23,$82,$47,$C7,$6B,$F0,$1A,$FA,$E3,$8B,$B2,$1B,$29,$B8,$E2
	.dc.b	$F9,$DA,$FB,$7E,$07,$FD,$72,$FE,$79,$AF,$FA,$E8,$C7,$47,$EE,$9F
	.dc.b	$F2,$FF,$9E,$7F,$D8,$C7,$F9,$71,$C7,$0F,$F9,$33,$D4,$46,$27,$F5
	.dc.b	$71,$87,$89,$D0,$AE,$67,$94,$F1,$9F,$E4,$13,$97,$9C,$65,$F1,$1C
	.dc.b	$9C,$3D,$67,$9B,$92,$38,$23,$D0,$C9,$5C,$BE,$28,$EB,$3E,$38,$C1
	.dc.b	$C0,$B3,$C1,$F2,$B5,$F2,$1D,$48,$70,$29,$C0,$A7,$03,$44,$34,$53
	.dc.b	$46,$08,$60,$91,$E9,$72,$D7,$33,$EE,$23,$ED,$36,$E2,$7D,$93,$C4
	.dc.b	$91,$99,$F4,$C6,$1E,$47,$62,$B9,$9E,$73,$C6,$7D,$93,$97,$9C,$65
	.dc.b	$F1,$9C,$9C,$3D,$67,$9B,$92,$38,$23,$D0,$C9,$5C,$BE,$28,$EB,$3E
	.dc.b	$38,$C1,$F9,$CE,$F6,$AD,$1F,$0F,$FE,$70,$25,$7F,$55,$7F,$B9,$3A
	.dc.b	$3F,$B9,$14,$FE,$E5,$8F,$EC,$63,$67,$F6,$B8,$D1,$FD,$CB,$87,$4F
	.dc.b	$F5,$AE,$2F,$EE,$93,$B3,$FD,$03,$F3,$12,$7F,$B3,$35,$FF,$07,$7C
	.dc.b	$CB,$E4,$7F,$80,$4F,$FB,$22,$31,$28,$C4,$EA,$3F,$B4,$CE,$A7,$FB
	.dc.b	$5F,$8A,$F5,$DF,$81,$FF,$0B,$F9,$4C,$C6,$30,$46,$27,$51,$FE,$21
	.dc.b	$AE,$51,$88,$FF,$12,$70,$7D,$0D,$FE,$97,$07,$D1,$E0,$FF,$4F,$92
	.dc.b	$BF,$BE,$FD,$92,$3F,$83,$B6,$3F,$8A,$B6,$3F,$DC,$ED,$8F,$16,$6F
	.dc.b	$4F,$F9,$27,$C8,$7A,$9B,$3A,$6B,$FC,$A7,$4F,$85,$18,$84,$7C,$5F
	.dc.b	$C1,$70,$D8,$FF,$66,$1C,$6D,$3A,$6C,$68,$E6,$BF,$05,$FE,$71,$F3
	.dc.b	$9F,$39,$B3,$EA,$6B,$D5,$AF,$2F,$C1,$AE,$1F,$65,$B6,$69,$8F,$C2
	.dc.b	$BF,$A3,$7C,$EF,$DC,$6D,$FC,$86,$BE,$B6,$BF,$D3,$7C,$4F,$A5,$3F
	.dc.b	$A4,$24,$7E,$91,$FC,$3F,$E1,$1C,$31,$FC,$63,$71,$34,$C7,$89,$F8
	.dc.b	$7B,$F8,$AD,$AF,$CA,$36,$BF,$CA,$3C,$DF,$80,$E2,$FC,$29,$2F,$F0
	.dc.b	$AF,$F5,$1B,$FC,$E7,$DF,$7F,$65,$F9,$EF,$CD,$B6,$BF,$0E,$D7,$ED
	.dc.b	$35,$FA,$FF,$45,$6B,$F4,$CD,$C4,$D3,$1F,$39,$E7,$25,$7E,$A0,$FC
	.dc.b	$AB,$EF,$5D,$3F,$A5,$CB,$F4,$51,$FA,$8E,$8F,$D3,$3F,$E6,$78,$F8
	.dc.b	$0F,$B8,$8E,$0F,$DD,$FE,$03,$F9,$88,$7E,$43,$8C,$FF,$C0,$CD,$8C
	.dc.b	$9F,$B9,$9E,$11,$CB,$F0,$BD,$CE,$5A,$FE,$9E,$51,$88,$FD,$3C,$61
	.dc.b	$EA,$6F,$E5,$FE,$E4,$1A,$48,$F8,$C7,$FA,$AF,$B3,$F7,$8D,$C7,$EF
	.dc.b	$1E,$F7,$D1,$E0,$74,$23,$C8,$F1,$BF,$3F,$4F,$9D,$1F,$2C,$87,$CF
	.dc.b	$EE,$F9,$6F,$F1,$2A,$EE,$D3,$FC,$5B,$A3,$2F,$E9,$71,$DE,$31,$AF
	.dc.b	$B8,$91,$E3,$19,$7B,$1F,$F1,$C3,$5C,$DF,$81,$F9,$DF,$5D,$F2,$08
	.dc.b	$F7,$C1,$1F,$8B,$AE,$44,$AE,$55,$FE,$45,$CB,$E0,$BE,$F8,$FC,$7B
	.dc.b	$9C,$3F,$1E,$FD,$DF,$17,$F5,$58,$FF,$B2,$6B,$F1,$B1,$83,$3F,$5B
	.dc.b	$FD,$72,$BF,$08,$D7,$F0,$6D,$CF,$E4,$37,$ED,$1C,$BF,$B7,$6D,$7C
	.dc.b	$E6,$E1,$FD,$59,$B1,$F6,$0F,$46,$DF,$D3,$09,$3E,$68,$FA,$F9,$7A
	.dc.b	$5F,$CF,$65,$E9,$76,$73,$3F,$67,$9F,$C2,$7A,$D7,$EB,$3D,$6F,$86
	.dc.b	$48,$EA,$FC,$6F,$AA,$FF,$39,$7E,$D1,$FB,$26,$CF,$EB,$5A,$FE,$A5
	.dc.b	$AF,$EA,$2B,$FC,$F5,$DF,$FB,$A6,$C7,$ED,$5A,$FE,$C5,$AF,$98,$F5
	.dc.b	$70,$7D,$E7,$07,$ED,$FE,$C7,$ED,$91,$F8,$AF,$DE,$FA,$BF,$DA,$92
	.dc.b	$3F,$1C,$FE,$A7,$3F,$89,$1F,$1B,$6F,$92,$FE,$2C,$FE,$15,$B7,$FA
	.dc.b	$43,$8A,$C7,$F7,$B6,$E5,$FC,$56,$3F,$83,$CE,$DF,$E1,$9F,$69,$FE
	.dc.b	$11,$AF,$E2,$8D,$8C,$9F,$E1,$28,$CB,$4F,$11,$0F,$82,$47,$F7,$1F
	.dc.b	$6B,$DE,$3F,$8B,$B6,$33,$1E,$73,$F8,$F4,$6F,$E7,$9F,$40,$F8,$EC
	.dc.b	$36,$B8,$09,$DC,$E8,$ED,$8B,$ED,$BD,$95,$FC,$A1,$B8,$FF,$0D,$6C
	.dc.b	$EE,$7D,$C7,$B0,$FA,$59,$FE,$E9,$A7,$FB,$0C,$F1,$63,$E5,$8F,$A1
	.dc.b	$8F,$92,$7C,$ED,$8C,$CE,$F0,$FC,$5E,$BF,$80,$E0,$FC,$66,$1D,$8F
	.dc.b	$9C,$53,$C8,$EB,$D6,$7A,$CF,$67,$FD,$AD,$DF,$FE,$3B,$B5,$6A,$73
	.dc.b	$9F,$A6,$31,$3F,$D1,$E3,$17,$CE,$30,$7F,$B8,$67,$77,$D5,$B5,$C2
	.dc.b	$FA,$D6,$EF,$95,$F8,$F0,$FF,$2B,$4F,$C9,$3F,$86,$94,$FD,$01,$5D
	.dc.b	$BB,$6B,$FD,$2C,$D6,$3F,$E8,$F6,$F1,$73,$D7,$E9,$47,$D0,$E0,$FF
	.dc.b	$35,$3F,$EF,$B8,$CE,$BF,$6D,$B1,$D4,$F0,$5F,$56,$DF,$9D,$F3,$9F
	.dc.b	$DF,$78,$0B,$DE,$FD,$96,$1F,$8F,$8C,$3B,$69,$FD,$FE,$78,$CF,$9E
	.dc.b	$3F,$CD,$07,$9F,$83,$F4,$4E,$49,$5F,$E0,$DE,$2F,$6C,$7E,$C0,$E2
	.dc.b	$53,$C7,$C4,$8D,$C7,$A8,$F8,$04,$3E,$A9,$FC,$0A,$F8,$F1,$F9,$AF
	.dc.b	$04,$9F,$F0,$C5,$F8,$A3,$FB,$D3,$4F,$EF,$91,$E4,$8C,$BE,$63,$C4
	.dc.b	$78,$0F,$11,$87,$FC,$09,$C9,$F1,$A3,$FC,$58,$D7,$FC,$50,$D7,$FC
	.dc.b	$4E,$D8,$FF,$06,$72,$FD,$96,$CF,$E4,$5A,$FF,$8F,$A3,$EF,$3E,$42
	.dc.b	$70,$7E,$49,$F3,$1F,$D3,$47,$BB,$FD,$35,$A7,$F4,$D2,$9F,$D3,$4A
	.dc.b	$7F,$8F,$3A,$3F,$13,$FB,$4F,$F5,$D8,$34,$8F,$F9,$2F,$AF,$D5,$99
	.dc.b	$A4,$8F,$CB,$61,$E7,$7D,$3C,$5F,$9B,$7E,$F6,$7F,$E2,$FE,$CE,$64
	.dc.b	$8F,$89,$82,$3E,$1D,$F3,$25,$F3,$AF,$87,$6E,$DF,$F0,$7C,$7F,$83
	.dc.b	$7B,$BE,$84,$62,$3F,$DB,$8D,$7F,$C3,$A7,$F1,$12,$F5,$7F,$9C,$3B
	.dc.b	$FF,$5A,$D8,$D1,$FE,$64,$CB,$D0,$FF,$85,$48,$7E,$E0,$FE,$0A,$43
	.dc.b	$07,$F9,$44,$FF,$40,$90,$C1,$FE,$79,$F0,$7F,$15,$A7,$F9,$12,$3F
	.dc.b	$7C,$64,$A7,$F9,$30,$A7,$F9,$4B,$B7,$F1,$7E,$0F,$F3,$58,$EA,$7F
	.dc.b	$9C,$4F,$9F,$AF,$62,$3C,$B5,$FF,$30,$BD,$0F,$F2,$D9,$FC,$4E,$7A
	.dc.b	$3C,$CF,$51,$FA,$AD,$BE,$BF,$57,$E1,$2F,$EC,$5F,$F9,$BF,$C2,$F6
	.dc.b	$92,$FC,$19,$23,$F5,$91,$E0,$3A,$9E,$D3,$C5,$97,$E5,$3B,$BE,$39
	.dc.b	$23,$9B,$F2,$BC,$3F,$E4,$FF,$91,$7F,$A8,$E4,$D2,$3F,$EA,$9C,$76
	.dc.b	$D3,$65,$8F,$F5,$81,$E4,$7C,$C7,$CE,$3C,$C7,$E0,$4F,$89,$97,$E6
	.dc.b	$5F,$E1,$DE,$43,$C7,$FE,$10,$D8,$FA,$4D,$ED,$FE,$B5,$82,$31,$3F
	.dc.b	$E0,$58,$FF,$6D,$AF,$8D,$DC,$FF,$B7,$23,$06,$DC,$46,$BC,$9F,$E9
	.dc.b	$6F,$FB,$B2,$76,$7F,$BA,$8A,$2B,$FD,$E6,$E1,$B1,$F0,$F8,$BF,$0B
	.dc.b	$3F,$08,$F8,$93,$E9,$F3,$3D,$E4,$8F,$7B,$F0,$BC,$DF,$EB,$BF,$13
	.dc.b	$EA,$6C,$FA,$5A,$FF,$62,$AF,$89,$18,$8F,$EC,$BE,$83,$8B,$10,$8C
	.dc.b	$2B,$FF,$02,$46,$1B,$3C,$EF,$E8,$79,$77,$53,$49,$87,$33,$E7,$7A
	.dc.b	$3F,$DD,$BE,$1F,$E9,$B6,$32,$23,$2D,$3E,$B7,$A7,$B7,$DD,$68,$C4
	.dc.b	$7D,$38,$C3,$C4,$DC,$6B,$0F,$C4,$7A,$BF,$E0,$8F,$0B,$F0,$AD,$AF
	.dc.b	$C2,$35,$EE,$D7,$B7,$AD,$6B,$F2,$4D,$C0,$D3,$84,$FE,$90,$FE,$0C
	.dc.b	$7E,$94,$C1,$FD,$70,$FE,$8F,$D9,$FC,$1C,$FE,$CD,$2F,$63,$F4,$A4
	.dc.b	$9F,$D2,$BF,$C1,$BB,$1E,$D7,$F8,$93,$E0,$BF,$C7,$D1,$8A,$46,$27
	.dc.b	$51,$FE,$45,$9D,$4F,$F8,$F0,$91,$FE,$32,$6B,$FE,$73,$F7,$7F,$AB
	.dc.b	$3F,$D1,$03,$8F,$13,$FC,$9C,$D8,$FF,$29,$35,$FF,$3E,$72,$FE,$4A
	.dc.b	$7F,$1F,$1C,$E3,$0F,$F9,$65,$AF,$FA,$0F,$C2,$77,$35,$F1,$46,$1B
	.dc.b	$19,$BF,$F2,$B9,$FE,$61,$F7,$9E,$FF,$F8,$41,$B8,$7F,$93,$DA,$EE
	.dc.b	$34,$47,$4D,$74,$60,$86,$08,$7F,$9C,$E3,$AB,$63,$93,$4E,$25,$38
	.dc.b	$9F,$EA,$38,$E2,$D8,$E2,$D3,$89,$4E,$27,$F9,$A2,$38,$B6,$38,$B4
	.dc.b	$E2,$53,$88,$8E,$2D,$8E,$2D,$38,$94,$E3,$1C,$09,$1C,$1B,$1C,$1A
	.dc.b	$F0,$8D,$11,$D3,$5D,$18,$21,$82,$1F,$2C,$87,$C5,$29,$EF,$29,$E9
	.dc.b	$2C,$79,$1B,$1E,$16,$9E,$02,$9E,$03,$B1,$0E,$C5,$3B,$16,$3B,$36
	.dc.b	$3B,$34,$EC,$53,$B1,$FE,$C3,$9F,$F5,$47,$AD,$FE,$D0,$8F,$F6,$13
	.dc.b	$5F,$F6,$51,$FA,$3E,$7F,$F5,$5F,$4D,$8F,$F0,$EE,$09,$CC,$7E,$65
	.dc.b	$AF,$E4,$DA,$FE,$35,$AF,$E1,$DB,$1C,$E3,$AC,$60,$8E,$09,$3E,$32
	.dc.b	$3E,$13,$25,$72,$F8,$CC,$95,$CB,$E5,$32,$57,$2F,$9C,$C9,$5C,$CF
	.dc.b	$A5,$B3,$E8,$6C,$F0,$9D,$3E,$52,$1E,$32,$9E,$12,$9D,$CA,$F2,$78
	.dc.b	$10,$E0,$53,$81,$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$27,$DD,$1E,$46
	.dc.b	$BD,$DA,$F4,$6B,$CD,$B1,$CE,$3E,$34,$60,$8E,$09,$3F,$30,$8F,$84
	.dc.b	$C9,$5C,$BE,$33,$25,$4C,$2E,$7F,$D1,$FD,$FC,$46,$33,$FE,$99,$18
	.dc.b	$84,$ED,$DB,$5F,$EB,$75,$DA,$7F,$AD,$92,$3F,$AD,$46,$5F,$EE,$1C
	.dc.b	$35,$F2,$70,$FF,$77,$AF,$EE,$D0,$69,$FD,$B3,$F0,$3E,$11,$EC,$E1
	.dc.b	$46,$23,$FB,$9B,$63,$31,$CC,$F1,$8F,$80,$FE,$E2,$47,$2F,$C1,$FC
	.dc.b	$D5,$E4,$7C,$A8,$C3,$67,$E4,$CE,$9E,$86,$23,$4F,$F8,$57,$3F,$F0
	.dc.b	$49,$F4,$39,$7F,$CA,$DE,$5F,$F0,$A6,$CF,$FA,$E5,$B3,$FE,$06,$24
	.dc.b	$7A,$9A,$FD,$EF,$61,$FC,$02,$FB,$EF,$E2,$B1,$86,$C6,$67,$FD,$8D
	.dc.b	$39,$7F,$D9,$1D,$AF,$35,$F7,$BF,$8A,$FD,$5F,$FD,$27,$18,$A4,$62
	.dc.b	$75,$1F,$E9,$59,$D4,$F0,$E1,$CC,$FF,$9E,$67,$EA,$CE,$63,$99,$C6
	.dc.b	$76,$FF,$A4,$67,$D1,$F8,$5F,$8B,$3F,$CE,$B6,$4D,$3C,$FF,$57,$A8
	.dc.b	$F2,$35,$F1,$35,$EC,$DC,$9F,$AF,$7F,$8B,$E5,$F9,$DF,$FA,$D2,$30
	.dc.b	$DF,$03,$F5,$ED,$9D,$9F,$AE,$F2,$9E,$66,$A3,$0D,$7F,$6D,$1A,$78
	.dc.b	$98,$8D,$3F,$EB,$3B,$7E,$07,$8C,$7C,$ED,$8C,$CF,$23,$84,$72,$E4
	.dc.b	$FF,$53,$7B,$5F,$91,$E5,$34,$95,$F8,$6E,$5F,$C4,$4F,$F0,$8E,$5F
	.dc.b	$C4,$36,$33,$3F,$C3,$E7,$2F,$FB,$5F,$17,$E0,$BF,$C5,$3F,$56,$F9
	.dc.b	$4D,$9D,$35,$FF,$5F,$4F,$05,$3F,$47,$D1,$D4,$7C,$36,$EB,$F6,$1B
	.dc.b	$47,$7A,$7E,$B2,$BF,$92,$35,$FD,$B7,$4F,$C3,$3F,$BD,$F8,$3F,$C7
	.dc.b	$E3,$F9,$37,$1F,$F0,$B7,$FC,$E9,$E1,$7F,$28,$7F,$92,$B5,$FE,$63
	.dc.b	$F8,$1F,$FB,$47,$87,$F8,$CB,$63,$90,$71,$E0,$FF,$34,$6C,$7E,$AF
	.dc.b	$A7,$F9,$DF,$89,$F9,$F2,$46,$4F,$F0,$71,$F9,$C2,$9F,$9C,$2E,$FF
	.dc.b	$BD,$E9,$E7,$1F,$A4,$33,$C0,$FC,$1F,$F9,$7D,$7E,$89,$B1,$99,$F4
	.dc.b	$CE,$5B,$0F,$7F,$64,$8C,$BC,$BE,$4E,$47,$06,$E9,$FD,$47,$83,$FD
	.dc.b	$39,$E2,$5E,$0F,$F5,$03,$78,$7F,$97,$E1,$F4,$9A,$FC,$BF,$33,$F3
	.dc.b	$6E,$38,$7E,$85,$7C,$D9,$CB,$78,$BF,$9B,$D7,$EB,$77,$BB,$FB,$6B
	.dc.b	$F8,$04,$63,$7F,$85,$1C,$69,$F8,$55,$23,$FD,$C7,$2F,$C0,$47,$F0
	.dc.b	$9F,$95,$FE,$EB,$F8,$6D,$A3,$48,$F1,$D3,$D7,$F8,$6F,$72,$46,$75
	.dc.b	$FE,$FF,$1F,$17,$BA,$4F,$E5,$37,$46,$23,$F5,$BF,$31,$FE,$5A,$E0
	.dc.b	$FE,$3E,$DD,$CD,$31,$F1,$BF,$82,$25,$7E,$01,$FC,$2F,$D2,$71,$24
	.dc.b	$7F,$4C,$A4,$7E,$B3,$8F,$FD,$B5,$8F,$8E,$BF,$5A,$DC,$7F,$98,$FC
	.dc.b	$E7,$8C,$93,$CD,$E5,$9F,$B2,$3D,$D5,$CD,$FF,$0C,$DF,$67,$F8,$97
	.dc.b	$CB,$FB,$48,$F9,$5F,$81,$F3,$BF,$E2,$8F,$DA,$FF,$9F,$36,$7F,$8D
	.dc.b	$35,$FE,$2A,$D7,$F9,$A7,$CF,$FF,$8C,$3E,$03,$FA,$33,$63,$F9,$13
	.dc.b	$5F,$E3,$AD,$7E,$09,$EA,$21,$F9,$DC,$7F,$A8,$7D,$07,$D5,$24,$7C
	.dc.b	$D7,$EC,$7D,$B7,$F9,$03,$A3,$FA,$B3,$67,$F9,$93,$5F,$E5,$AD,$7C
	.dc.b	$27,$73,$C2,$60,$FE,$A1,$EB,$FF,$AA,$38,$3F,$9E,$36,$3F,$9B,$B5
	.dc.b	$D9,$F9,$08,$FE,$85,$F2,$BF,$E5,$4F,$C3,$51,$89,$46,$1F,$E8,$33
	.dc.b	$B9,$D4,$E6,$7C,$87,$DE,$E1,$FC,$13,$E9,$21,$FD,$9F,$E7,$7F,$CC
	.dc.b	$7C,$7C,$6F,$FB,$EC,$61,$FE,$AC,$D9,$FE,$A2,$D5,$E6,$3E,$59,$F5
	.dc.b	$7E,$F8,$C3,$63,$33,$F0,$4F,$F1,$B3,$8E,$0F,$17,$46,$BE,$72,$DF
	.dc.b	$F9,$E1,$B6,$69,$F0,$5F,$81,$5F,$9F,$F1,$1D,$3F,$9F,$9F,$83,$97
	.dc.b	$91,$F1,$1E,$AC,$BC,$5A,$7F,$77,$6C,$79,$B2,$FB,$4F,$63,$F0,$65
	.dc.b	$3E,$69,$4F,$A7,$87,$8B,$6F,$A7,$1F,$5B,$DF,$76,$3E,$03,$FD,$2B
	.dc.b	$EE,$7D,$89,$CC,$63,$6F,$F7,$BC,$E6,$B8,$5F,$F9,$D2,$72,$47,$FD
	.dc.b	$3F,$7C,$9D,$18,$9E,$35,$F8,$6C,$7C,$D3,$FD,$2C,$87,$F2,$72,$9F
	.dc.b	$D3,$4A,$7F,$80,$8F,$F1,$16,$7F,$79,$FC,$A6,$BF,$3D,$B1,$9D,$7E
	.dc.b	$76,$BF,$EB,$37,$B4,$C3,$B1,$4F,$F1,$F4,$7F,$AE,$BE,$33,$F0,$64
	.dc.b	$9F,$F1,$64,$65,$EB,$3F,$EC,$06,$BE,$28,$F0,$3F,$74,$D1,$F9,$E3
	.dc.b	$F6,$71,$FA,$26,$99,$7F,$2F,$C5,$D5,$F3,$47,$EC,$4F,$8A,$53,$F7
	.dc.b	$9F,$86,$7E,$7B,$9B,$FD,$A6,$D8,$34,$FD,$1F,$3F,$9A,$3F,$84,$98
	.dc.b	$3C,$E6,$2B,$C3,$A7,$DA,$8B,$C7,$FE,$77,$7F,$8F,$9F,$E2,$D2,$F8
	.dc.b	$FF,$C9,$1C,$1F,$C8,$8A,$7F,$9B,$7A,$FF,$DC,$4D,$3F,$DC,$97,$FE
	.dc.b	$84,$8C,$B6,$37,$7E,$37,$ED,$90,$FC,$41,$FD,$08,$FF,$13,$17,$CC
	.dc.b	$FE,$8E,$E0,$FE,$8A,$53,$FD,$01,$DF,$FE,$FC,$FD,$B3,$FA,$AC,$1A
	.dc.b	$47,$FA,$F9,$E8,$EA,$78,$5E,$86,$C6,$63,$DE,$FD,$93,$D0,$53,$D2
	.dc.b	$7F,$9D,$3B,$39,$11,$EC,$F2,$C7,$E1,$47,$02,$1C,$F2,$FF,$56,$F8
	.dc.b	$7E,$32,$47,$9D,$F3,$74,$23,$F8,$09,$FE,$84,$71,$D7,$FE,$B3,$6C
	.dc.b	$7F,$A3,$DA,$FF,$87,$8F,$85,$E2,$7C,$12,$3F,$63,$87,$F1,$0F,$C0
	.dc.b	$C7,$D2,$FC,$1E,$1F,$F3,$C7,$E1,$9F,$E3,$F2,$47,$F8,$C4,$3F,$E3
	.dc.b	$BE,$A9,$1F,$EA,$1F,$2B,$FD,$EC,$D3,$FA,$59,$C6,$BF,$C9,$44,$70
	.dc.b	$7F,$AA,$7C,$CF,$85,$E9,$FF,$94,$A0,$D2,$3E,$37,$FC,$A8,$47,$07
	.dc.b	$F9,$0D,$FF,$35,$11,$C1,$FE,$B7,$F3,$FE,$4B,$FE,$CF,$6F,$07,$FB
	.dc.b	$91,$BC,$66,$F3,$4A,$FA,$E7,$FB,$5D,$C7,$0F,$A4,$91,$ED,$37,$7F
	.dc.b	$E8,$B8,$FF,$7C,$C7,$FA,$3C,$FF,$4E,$63,$82,$3F,$EA,$17,$F0,$CF
	.dc.b	$F2,$1F,$75,$E9,$FE,$26,$73,$3F,$E1,$09,$CB,$D4,$C4,$ED,$FB,$95
	.dc.b	$FE,$7D,$23,$83,$FC,$D8,$FF,$A4,$C8,$E0,$FF,$66,$F8,$6F,$53,$F8
	.dc.b	$5E,$BF,$E7,$26,$C6,$67,$CA,$7D,$C7,$0F,$FB,$7C,$7E,$B1,$48,$FF
	.dc.b	$BB,$7D,$6F,$D3,$47,$FB,$B2,$36,$7F,$BB,$0A,$7F,$BB,$3D,$8D,$B9
	.dc.b	$8F,$F7,$D6,$5F,$C8,$38,$DE,$CF,$C6,$AC,$9E,$68,$CC,$FA,$BD,$29
	.dc.b	$3E,$AD,$3F,$D1,$F1,$88,$FA,$3C,$0E,$37,$B5,$FB,$DF,$53,$31,$8A
	.dc.b	$46,$27,$51,$FD,$A6,$75,$3F,$23,$DC,$7B,$9A,$AF,$C1,$C6,$1E,$67
	.dc.b	$E0,$9C,$4F,$19,$E1,$3F,$E5,$4A,$79,$D8,$C7,$D0,$D8,$CC,$F4,$3F
	.dc.b	$B8,$38,$BF,$CC,$FE,$1B,$27,$FD,$DC,$D7,$F2,$DA,$FF,$1D,$8F,$8B
	.dc.b	$CD,$F9,$46,$C6,$67,$FD,$FD,$1A,$F8,$0F,$E4,$9F,$AC,$D9,$34,$8F
	.dc.b	$F8,$3E,$DF,$07,$C2,$30,$D7,$F5,$0D,$7F,$4C,$DE,$2F,$BF,$1F,$81
	.dc.b	$E3,$FC,$23,$FB,$08,$CB,$F8,$63,$F3,$79,$FF,$90,$9F,$DA,$B5,$FC
	.dc.b	$07,$3F,$F5,$08,$C7,$27,$E9,$67,$F1,$53,$FA,$27,$F5,$7D,$1F,$A8
	.dc.b	$3F,$AB,$FC,$23,$D2,$FE,$4B,$3F,$BE,$8C,$3F,$C9,$8C,$46,$9F,$E0
	.dc.b	$F1,$AB,$F3,$DE,$67,$CA,$71,$3D,$2E,$5F,$CA,$B7,$E1,$5B,$FF,$28
	.dc.b	$94,$61,$FE,$46,$62,$35,$19,$EC,$FE,$03,$F0,$CF,$3B,$93,$6F,$8B
	.dc.b	$67,$73,$DF,$5F,$F7,$14,$7A,$AB,$BC,$61,$FE,$42,$D7,$F9,$3F,$8B
	.dc.b	$FE,$77,$9F,$E3,$31,$8F,$1B,$F4,$31,$88,$FE,$E5,$F0,$EF,$83,$FF
	.dc.b	$03,$B6,$3F,$C4,$ED,$7F,$8D,$B5,$FE,$62,$D7,$F9,$BB,$5E,$B1,$D0
	.dc.b	$8F,$46,$BD,$1A,$F4,$6B,$D1,$AF,$43,$91,$0E,$45,$39,$14,$E4,$53
	.dc.b	$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$23,$E4
	.dc.b	$36,$3D,$ED,$3D,$05,$3C,$65,$3C,$05,$3B,$94,$EF,$1D,$89,$1D,$9B
	.dc.b	$1D,$9A,$F6,$6B,$D9,$AF,$66,$BD,$A3,$A1,$1E,$8D,$7A,$35,$E8,$D7
	.dc.b	$A3,$5E,$87,$22,$1C,$8A,$72,$3F,$CF,$FE,$BF,$FB,$25,$FF,$03,$B5
	.dc.b	$FF,$06,$B5,$FF,$66,$57,$FA,$1B,$C5,$FE,$1F,$BF,$F7,$A6,$CF,$F8
	.dc.b	$41,$AF,$F8,$59,$AF,$F9,$75,$A9,$A4,$E2,$FF,$31,$4F,$56,$CF,$56
	.dc.b	$BD,$5A,$F5,$6B,$D5,$B3,$D6,$3F,$9B,$36,$7B,$46,$27,$64,$70,$48
	.dc.b	$D9,$1F,$21,$92,$B9,$7C,$C6,$4A,$E5,$F4,$19,$2B,$97,$D4,$64,$AE
	.dc.b	$67,$D5,$1E,$02,$1D,$8A,$74,$29,$C8,$AF,$27,$81,$0E,$05,$38,$15
	.dc.b	$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$7E,$84,$FB,$5B,$3E,$66,$BE,$16
	.dc.b	$BD,$9A,$F5,$6C,$F5,$8F,$94,$D9,$ED,$18,$9D,$9C,$B3,$67,$FE,$26
	.dc.b	$F8,$2F,$F2,$CC,$63,$B5,$18,$9D,$5F,$F9,$CA,$B5,$3F,$E4,$4F,$D5
	.dc.b	$7F,$CF,$3E,$D5,$AF,$F6,$03,$53,$4F,$F3,$3F,$D3,$7F,$B6,$9B,$E9
	.dc.b	$70,$9F,$F3,$FC,$E5,$B1,$97,$FD,$0F,$E3,$BE,$7B,$F4,$AD,$8F,$F0
	.dc.b	$C3,$76,$FF,$3A,$B6,$3F,$D1,$FE,$23,$27,$DA,$75,$94,$E6,$30,$D8
	.dc.b	$F5,$9C,$A7,$8B,$83,$FD,$37,$D3,$FE,$85,$D3,$FD,$0A,$D8,$FF,$4D
	.dc.b	$B5,$DF,$6F,$85,$FA,$BE,$CF,$F0,$0E,$13,$49,$0D,$3E,$97,$D5,$79
	.dc.b	$B2,$34,$93,$E5,$3F,$C4,$B9,$7F,$83,$C8,$FF,$83,$AB,$E1,$7A,$4F
	.dc.b	$C0,$7E,$94,$FF,$BE,$E3,$1C,$3A,$9C,$CF,$FB,$A9,$BB,$7C,$DF,$8E
	.dc.b	$4F,$A8,$F9,$0E,$34,$FC,$53,$63,$26,$CE,$33,$F5,$FE,$DF,$51,$8D
	.dc.b	$51,$88,$FB,$31,$87,$91,$88,$D4,$FE,$07,$EE,$3C,$BF,$31,$F9,$B6
	.dc.b	$D7,$DD,$6B,$DD,$AF,$6F,$BA,$B5,$F8,$B6,$E0,$69,$8F,$BB,$F8,$7E
	.dc.b	$57,$EA,$5B,$7F,$8B,$6B,$ED,$6B,$FE,$5C,$8F,$F2,$31,$1F,$B0,$7F
	.dc.b	$A1,$8F,$CA,$7D,$EF,$0C,$3F,$3C,$DC,$0D,$31,$F7,$FF,$89,$25,$7E
	.dc.b	$1E,$BF,$4D,$1F,$8D,$FB,$F9,$39,$B4,$63,$93,$F8,$27,$CB,$FE,$03
	.dc.b	$D2,$B7,$FE,$90,$C7,$F6,$13,$98,$FD,$53,$5F,$B9,$F3,$35,$FB,$6D
	.dc.b	$9F,$DC,$37,$13,$7A,$3F,$1F,$F8,$37,$E9,$2C,$D2,$3F,$C6,$FA,$BF
	.dc.b	$D8,$F9,$FF,$1F,$F0,$FF,$4D,$2F,$0F,$F9,$24,$E4,$FC,$DC,$7E,$73
	.dc.b	$DE,$D7,$CE,$7F,$09,$6C,$7E,$BD,$B3,$C2,$11,$87,$F1,$C7,$F1,$93
	.dc.b	$F1,$C5,$3F,$37,$CD,$FA,$AF,$0F,$F2,$D1,$9D,$FF,$20,$DD,$FF,$20
	.dc.b	$DE,$7E,$0F,$F1,$76,$BF,$A9,$E4,$F8,$EF,$F1,$B6,$B9,$3E,$49,$92
	.dc.b	$9F,$B1,$E4,$DB,$FB,$06,$DF,$F2,$D6,$DF,$E1,$A7,$89,$FB,$6F,$9C
	.dc.b	$FE,$0B,$F8,$37,$F4,$B6,$C7,$F2,$D6,$BF,$BF,$6B,$FC,$33,$7F,$EB
	.dc.b	$1C,$5F,$64,$91,$F7,$78,$7B,$EF,$F5,$27,$E9,$36,$7E,$DB,$5F,$B9
	.dc.b	$4E,$1F,$4E,$BF,$C6,$BE,$8B,$FB,$03,$6B,$E2,$46,$13,$DA,$B9,$43
	.dc.b	$3C,$5F,$AB,$0F,$C3,$C6,$27,$47,$F4,$DF,$9F,$FE,$53,$F8,$37,$F6
	.dc.b	$F6,$C6,$44,$65,$A7,$B3,$0F,$0F,$E0,$DD,$49,$19,$C7,$F9,$AC,$7E
	.dc.b	$87,$C5,$93,$F2,$5A,$FF,$2F,$F4,$3F,$A7,$11,$FE,$04,$FF,$51,$F9
	.dc.b	$7F,$E7,$FF,$84,$FF,$81,$5B,$3F,$D9,$DA,$FF,$39,$6B,$FC,$E3,$C2
	.dc.b	$FF,$30,$F3,$7F,$75,$6F,$29,$A7,$F9,$93,$4F,$F0,$C6,$5E,$6B,$FC
	.dc.b	$27,$5B,$97,$FB,$34,$7A,$CF,$EF,$1F,$01,$D0,$8E,$8F,$E9,$8F,$59
	.dc.b	$E5,$5D,$67,$F8,$24,$61,$E8,$7F,$53,$FA,$BF,$EE,$44,$7F,$93,$3F
	.dc.b	$DD,$7E,$9A,$47,$F2,$96,$F7,$7F,$8A,$F9,$FF,$BD,$12,$7F,$97,$3F
	.dc.b	$E0,$5C,$FD,$F1,$E0,$3D,$F9,$FA,$5B,$97,$F0,$FD,$9F,$B7,$7F,$87
	.dc.b	$27,$F9,$E3,$71,$D3,$63,$FC,$4A,$D7,$FB,$8A,$FA,$9F,$EF,$7F,$83
	.dc.b	$F0,$6C,$F0,$6B,$C1,$AF,$90,$F4,$1F,$2B,$1F,$43,$71,$F1,$FA,$BF
	.dc.b	$E3,$0C,$0D,$24,$FA,$BD,$CF,$F0,$6F,$E0,$FE,$D8,$FF,$1C,$D2,$73
	.dc.b	$A7,$B1,$B3,$EC,$6B,$E5,$3F,$9E,$EB,$F3,$9B,$C3,$F7,$DB,$9F,$CF
	.dc.b	$BF,$9B,$3F,$E1,$EF,$A3,$F1,$46,$29,$18,$C7,$FC,$E3,$18,$9F,$9F
	.dc.b	$5C,$5F,$A3,$F5,$9F,$EC,$86,$CF,$F8,$FD,$BA,$F3,$9D,$4F,$D6,$F5
	.dc.b	$3F,$CE,$35,$F6,$9B,$19,$3F,$CE,$71,$96,$9E,$CA,$FC,$F6,$7F,$06
	.dc.b	$3E,$21,$F3,$2F,$BB,$4F,$21,$F8,$67,$35,$F9,$07,$71,$F7,$4F,$C0
	.dc.b	$CF,$57,$07,$84,$CE,$1F,$3D,$F0,$8F,$71,$F2,$4F,$77,$FC,$9B,$C3
	.dc.b	$FB,$99,$CC,$63,$3F,$F8,$03,$5B,$BF,$C5,$C6,$27,$51,$FE,$56,$AE
	.dc.b	$2E,$8C,$57,$1B,$EC,$DB,$E5,$3B,$3F,$38,$7F,$82,$C8,$7F,$53,$29
	.dc.b	$F8,$A2,$9F,$BA,$3F,$67,$97,$FA,$59,$F9,$87,$F9,$88,$A7,$E0,$4C
	.dc.b	$1C,$4E,$C4,$3B,$1F,$7F,$E6,$FF,$25,$AF,$F9,$D1,$B0,$69,$D5,$9A
	.dc.b	$FE,$49,$A7,$F0,$4D,$3F,$CF,$BE,$47,$E7,$5B,$3F,$9C,$9F,$F2,$2B
	.dc.b	$67,$F6,$D1,$F1,$8F,$31,$E9,$3C,$E7,$ED,$47,$E0,$FF,$A9,$E3,$FF
	.dc.b	$45,$B6,$33,$5F,$BD,$69,$FC,$63,$5F,$73,$FC,$71,$AE,$4F,$80,$64
	.dc.b	$A7,$88,$F2,$10,$F4,$1F,$C7,$47,$E0,$FF,$C8,$DA,$7F,$23,$29,$B7
	.dc.b	$F9,$7C,$F8,$A3,$F8,$BB,$4F,$E6,$E5,$E2,$FE,$71,$9F,$E1,$A7,$C2
	.dc.b	$78,$F8,$BF,$DF,$0F,$F4,$93,$D2,$7F,$A8,$BA,$7A,$77,$7F,$85,$23
	.dc.b	$38,$7F,$B8,$67,$3D,$72,$73,$48,$C6,$1F,$EF,$FE,$CE,$64,$AD,$BE
	.dc.b	$B7,$CA,$4C,$7C,$AF,$DE,$C7,$FD,$E1,$DB,$FE,$63,$8F,$ED,$E5,$8C
	.dc.b	$B7,$14,$EE,$3E,$07,$0F,$F7,$D7,$E2,$5F,$67,$FD,$89,$DD,$BC,$3F
	.dc.b	$C1,$51,$CE,$31,$87,$F8,$2E,$3B,$77,$FF,$B5,$BE,$FB,$FC,$31,$06
	.dc.b	$91,$FF,$80,$1E,$17,$CC,$23,$F2,$1F,$85,$8F,$8E,$3C,$FD,$BE,$98
	.dc.b	$C7,$91,$FE,$86,$FC,$17,$FC,$80,$4A,$FF,$1A,$07,$FC,$79,$F8,$2C
	.dc.b	$8F,$F2,$03,$5F,$F2,$BF,$8F,$FE,$EE,$FC,$13,$FC,$94,$48,$E6,$FF
	.dc.b	$95,$7E,$6A,$47,$F9,$ED,$AF,$21,$F8,$17,$F9,$4C,$8F,$91,$FF,$53
	.dc.b	$11,$FF,$37,$35,$FF,$3B,$09,$FF,$3E,$34,$FF,$3E,$14,$DB,$FE,$17
	.dc.b	$DF,$FD,$0D,$37,$B5,$5E,$43,$FD,$63,$C5,$FE,$8D,$70,$7F,$A4,$CA
	.dc.b	$1F,$F6,$CB,$86,$CF,$F9,$DC,$91,$FE,$A5,$1E,$A7,$8F,$D2,$F0,$B6
	.dc.b	$32,$6D,$FF,$59,$91,$FF,$5D,$B5,$F9,$9D,$1F,$E7,$D2,$3F,$79,$FF
	.dc.b	$68,$11,$FF,$4C,$35,$F8,$36,$C7,$C3,$87,$FB,$74,$93,$F1,$23,$2D
	.dc.b	$F5,$7E,$4F,$A3,$F3,$5B,$19,$11,$96,$9F,$EF,$42,$9E,$29,$7A,$DC
	.dc.b	$A3,$0D,$7E,$84,$69,$E2,$7F,$B1,$4F,$E1,$BA,$7F,$14,$23,$FC,$42
	.dc.b	$BF,$8B,$91,$C1,$FC,$53,$D8,$FC,$1F,$B5,$F8,$76,$C7,$C1,$6B,$E7
	.dc.b	$6B,$F8,$AF,$88,$FF,$41,$CF,$F1,$16,$C7,$FB,$D5,$AF,$FB,$F9,$AF
	.dc.b	$B2,$39,$E9,$F4,$DB,$89,$A6,$37,$FC,$CF,$CC,$FE,$24,$97,$F8,$F7
	.dc.b	$FD,$F6,$47,$ED,$7B,$5C,$1A,$23,$F4,$DE,$EF,$E9,$3E,$D2,$C9,$A4
	.dc.b	$7F,$09,$EF,$ED,$A6,$CB,$1F,$B1,$8D,$47,$EC,$3E,$0D,$F0,$4F,$83
	.dc.b	$FE,$B1,$18,$6C,$66,$7F,$A8,$CE,$5F,$EA,$63,$7F,$F5,$9F,$37,$D1
	.dc.b	$F4,$3F,$A4,$46,$27,$E9,$47,$F5,$8E,$89,$1F,$D5,$1B,$5F,$D2,$7C
	.dc.b	$57,$C1,$FF,$45,$A4,$61,$FE,$90,$D7,$84,$6F,$E1,$1E,$8F,$F5,$96
	.dc.b	$CF,$A1,$B7,$FD,$5D,$B1,$FA,$FF,$85,$FE,$DA,$47,$FB,$4B,$FD,$8F
	.dc.b	$E1,$9F,$09,$E5,$6C,$FF,$5B,$6B,$FD,$8D,$AF,$F2,$03,$F9,$0F,$C3
	.dc.b	$FC,$DC,$3F,$B4,$36,$34,$1C,$3C,$5A,$FE,$D0,$FF,$01,$90,$FE,$FE
	.dc.b	$7F,$58,$CF,$FD,$D1,$EC,$FF,$75,$F2,$3E,$5F,$93,$FE,$69,$AF,$F5
	.dc.b	$0C,$E6,$B7,$1E,$2F,$89,$FF,$5B,$BF,$DD,$DA,$FF,$7F,$6B,$BF,$57
	.dc.b	$EC,$46,$23,$E0,$35,$FE,$82,$D4,$D2,$47,$22,$47,$85,$AF,$85,$AF
	.dc.b	$84,$D1,$0D,$14,$D1,$82,$18,$21,$E8,$7A,$B4,$E4,$53,$89,$4E,$27
	.dc.b	$B9,$E2,$D3,$89,$4E,$25,$38,$9F,$11,$E2,$D3,$89,$4E,$25,$38,$C7
	.dc.b	$CA,$76,$D7,$6D,$77,$1D,$A3,$11,$EB,$21,$E5,$29,$DC,$AF,$07,$44
	.dc.b	$34,$57,$4E,$08,$E0,$91,$E3,$24,$7A,$9A,$FA,$9A,$FA,$8D,$10,$D1
	.dc.b	$FE,$4F,$F6,$BF,$D8,$6F,$EB,$9A,$FF,$B0,$0C,$7B,$5F,$A5,$7F,$33
	.dc.b	$F7,$1F,$E1,$8E,$5F,$D0,$36,$3F,$4C,$D7,$62,$36,$D3,$65,$36,$E8
	.dc.b	$8E,$9A,$E8,$C1,$0C,$10,$FD,$6C,$78,$DA,$77,$2C,$74,$69,$CC,$A7
	.dc.b	$33,$F6,$B1,$CD,$A7,$32,$C7,$36,$9C,$CA,$73,$C1,$3C,$DB,$1C,$DB
	.dc.b	$3C,$DB,$1C,$DA,$F3,$9E,$44,$79,$36,$79,$36,$39,$35,$E4,$70,$24
	.dc.b	$70,$69,$C0,$A7,$07,$44,$74,$D7,$46,$08,$60,$87,$D0,$21,$F2,$0B
	.dc.b	$3F,$01,$B3,$E9,$69,$E4,$2C,$F8,$5A,$78,$0F,$F3,$8E,$9F,$D0,$BE
	.dc.b	$2B,$FA,$06,$A6,$92,$7F,$BC,$99,$2B,$FD,$FD,$C1,$FE,$2D,$AF,$F3
	.dc.b	$EC,$7F,$5A,$F0,$DE,$C7,$F8,$57,$E0,$70,$DB,$FF,$40,$B5,$FF,$32
	.dc.b	$B5,$F3,$EE,$C7,$F2,$19,$7F,$5A,$CB,$D2,$4A,$FE,$D3,$4F,$9A,$FC
	.dc.b	$C3,$63,$F3,$0D,$9F,$F0,$84,$FE,$9F,$2F,$3F,$9F,$FE,$96,$CF,$CE
	.dc.b	$DC,$3C,$ED,$7C,$ED,$7C,$F9,$79,$B1,$F4,$63,$E5,$9F,$F0,$16,$5E
	.dc.b	$97,$FB,$FF,$EA,$7F,$E7,$69,$FF,$75,$36,$33,$E5,$7E,$C6,$31,$1F
	.dc.b	$1B,$E6,$BC,$1A,$7C,$E6,$EC,$69,$E5,$96,$3E,$17,$C7,$97,$99,$FF
	.dc.b	$08,$C6,$5F,$5E,$5F,$EF,$DD,$9F,$1B,$FC,$0D,$B3,$FC,$0D,$B3,$B9
	.dc.b	$F9,$98,$FF,$21,$FD,$0F,$F0,$33,$FE,$C2,$6D,$9A,$7E,$1B,$EE,$B5
	.dc.b	$1F,$86,$6B,$F8,$66,$BF,$EC,$5A,$46,$34,$FC,$8F,$67,$FB,$43,$DB
	.dc.b	$FC,$8B,$6F,$F2,$2D,$7F,$22,$D7,$A7,$CB,$DA,$FC,$DB,$70,$34,$C7
	.dc.b	$37,$FB,$7B,$F1,$8F,$CF,$B6,$FF,$3E,$D7,$F3,$ED,$7F,$DC,$BF,$38
	.dc.b	$F7,$F1,$08,$C2,$AF,$F2,$3B,$99,$DC,$EA,$7F,$5F,$F7,$9D,$72,$FD
	.dc.b	$9B,$72,$34,$C7,$DE,$FE,$28,$98,$BA,$BF,$73,$1F,$8B,$E3,$FB,$F3
	.dc.b	$FE,$58,$6B,$FE,$5F,$6B,$FF,$00,$1B,$87,$F0,$4E,$34,$BF,$70,$FE
	.dc.b	$0B,$18,$6C,$64,$FE,$0F,$19,$7F,$84,$FC,$C5,$7F,$84,$B5,$FC,$23
	.dc.b	$6B,$F8,$9B,$73,$7C,$B7,$EA,$09,$5E,$F3,$F8,$AB,$FA,$9E,$7F,$D2
	.dc.b	$47,$EA,$BA,$3F,$50,$FE,$AB,$B3,$F8,$01,$FA,$3C,$BF,$4C,$48,$C1
	.dc.b	$FD,$CF,$2F,$E4,$BF,$19,$F0,$BF,$03,$FE,$4A,$D9,$FE,$4A,$D7,$F9
	.dc.b	$2B,$54,$BE,$53,$F9,$34,$2B,$F9,$94,$61,$B8,$71,$AE,$0D,$97,$D0
	.dc.b	$FF,$0C,$6D,$FC,$33,$EB,$D7,$79,$D1,$F8,$3C,$7D,$DD,$BF,$C0,$6B
	.dc.b	$DE,$F8,$BE,$8B,$A1,$23,$A1,$EA,$C1,$1F,$3F,$4F,$9B,$3F,$C4,$8F
	.dc.b	$E2,$AE,$1F,$A5,$97,$67,$F2,$58,$78,$9A,$7C,$FB,$FE,$85,$C9,$F1
	.dc.b	$38,$FD,$C4,$9F,$19,$FD,$26,$B9,$3F,$9B,$6B,$E7,$FA,$3F,$8C,$47
	.dc.b	$D7,$AF,$D3,$8F,$89,$1E,$53,$91,$4F,$61,$F8,$9D,$3F,$A3,$38,$3E
	.dc.b	$27,$8B,$FD,$57,$2F,$F5,$3C,$7A,$E3,$06,$7D,$9F,$D5,$FE,$05,$F8
	.dc.b	$06,$CF,$FB,$47,$1F,$C3,$46,$23,$ED,$DA,$31,$1F,$4F,$E9,$7C,$04
	.dc.b	$7E,$53,$FE,$43,$E4,$FD,$19,$38,$3E,$A3,$FD,$27,$0F,$CA,$65,$E6
	.dc.b	$BF,$41,$33,$F7,$BF,$69,$B5,$C7,$27,$D3,$7C,$12,$57,$DB,$7E,$27
	.dc.b	$95,$FC,$2B,$BF,$F3,$ED,$9F,$81,$48,$C4,$7C,$06,$BF,$91,$F8,$7F
	.dc.b	$9C,$47,$F1,$0F,$BF,$4F,$69,$39,$3F,$7F,$9F,$EB,$BE,$9F,$EB,$12
	.dc.b	$FF,$20,$FD,$9F,$6F,$F5,$84,$8F,$C3,$E4,$8F,$D0,$5F,$49,$FC,$FB
	.dc.b	$5F,$AF,$F5,$3F,$E9,$3E,$0F,$6B,$63,$27,$FA,$46,$32,$D3,$EF,$FD
	.dc.b	$97,$F8,$76,$3E,$2B,$63,$27,$F3,$F8,$CB,$4F,$CB,$1E,$4C,$BE,$BE
	.dc.b	$7F,$B6,$9F,$E2,$5F,$57,$E3,$AF,$DE,$36,$33,$1E,$33,$25,$7F,$8D
	.dc.b	$7D,$8E,$F4,$34,$93,$B8,$7D,$27,$A6,$3F,$85,$36,$BF,$92,$36,$77
	.dc.b	$3E,$B3,$F5,$C7,$E4,$4F,$61,$0F,$D2,$D7,$83,$1F,$E5,$F1,$83,$C5
	.dc.b	$1E,$F7,$91,$F1,$88,$76,$E3,$FD,$36,$BF,$6E,$3C,$3E,$9F,$E5,$7D
	.dc.b	$5F,$EC,$7E,$AF,$FB,$16,$31,$28,$C7,$22,$B7,$97,$F6,$AA,$5F,$0B
	.dc.b	$DD,$6A,$7F,$B7,$CE,$6B,$83,$63,$37,$D5,$F0,$4F,$F5,$D7,$FD,$0E
	.dc.b	$43,$F2,$A5,$34,$60,$87,$C7,$3E,$49,$FC,$FF,$2F,$E2,$8E,$75,$FD
	.dc.b	$BC,$62,$3F,$CD,$67,$52,$9D,$0F,$BE,$3D,$7E,$B3,$F8,$08,$C3,$FC
	.dc.b	$0B,$6F,$F0,$7C,$EB,$6F,$C1,$B6,$FC,$15,$F9,$2B,$FE,$BE,$79,$48
	.dc.b	$7E,$60,$E5,$BF,$0B,$34,$F5,$47,$F5,$63,$D4,$53,$D5,$3F,$DE,$78
	.dc.b	$7C,$AF,$EC,$30,$F3,$46,$27,$F6,$B1,$E9,$E2,$F0,$12,$3C,$E7,$F6
	.dc.b	$B7,$C3,$C2,$E2,$FF,$02,$CF,$C9,$3C,$47,$D8,$3F,$A1,$F1,$FC,$67
	.dc.b	$F9,$4B,$5C,$9F,$3C,$C9,$4F,$E9,$7C,$9F,$E5,$47,$FB,$E3,$72,$FF
	.dc.b	$0B,$37,$43,$B9,$F1,$9F,$78,$F2,$14,$FE,$6C,$7F,$98,$27,$FC,$41
	.dc.b	$CB,$CC,$91,$9A,$FE,$80,$BA,$7F,$A2,$B4,$FE,$8B,$C5,$FE,$52,$8C
	.dc.b	$9F,$E2,$8E,$8F,$31,$27,$A3,$E7,$D3,$E6,$73,$FB,$DB,$3F,$79,$AF
	.dc.b	$F5,$2E,$9F,$80,$47,$CC,$F3,$DB,$EA,$C7,$C0,$3E,$19,$F1,$34,$FF
	.dc.b	$1F,$63,$F2,$49,$1E,$F7,$E6,$75,$7F,$9B,$3E,$F7,$FC,$05,$26,$91
	.dc.b	$F2,$1E,$22,$1F,$E4,$8C,$3C,$99,$7B,$71,$F2,$CF,$D3,$3C,$BD,$FF
	.dc.b	$EC,$27,$FC,$80,$43,$FC,$62,$7F,$94,$BB,$BF,$14,$FF,$8A,$4F,$C4
	.dc.b	$17,$93,$4E,$0F,$F1,$91,$F8,$B9,$EC,$F4,$21,$FE,$1A,$7A,$E9,$FE
	.dc.b	$63,$6F,$07,$FC,$01,$AF,$F2,$05,$75,$9E,$6F,$F8,$D0,$F4,$14,$FD
	.dc.b	$36,$DF,$69,$FA,$8D,$7F,$CB,$1E,$0F,$62,$3E,$CC,$3F,$CE,$91,$F0
	.dc.b	$23,$D6,$76,$3F,$D4,$9E,$07,$E8,$6F,$C7,$1F,$E6,$3F,$0B,$D2,$47
	.dc.b	$A6,$5F,$E7,$4C,$BF,$CF,$70,$FB,$DA,$47,$FC,$43,$B7,$FA,$BA,$3F
	.dc.b	$D1,$B1,$B3,$FD,$1A,$53,$FD,$1A,$53,$FD,$D9,$E2,$F1,$7E,$79,$E8
	.dc.b	$3E,$A6,$5E,$93,$F0,$EF,$E8,$77,$FC,$3B,$F9,$9F,$1F,$FD,$5E,$FC
	.dc.b	$56,$E3,$FD,$66,$3F,$DF,$5E,$3F,$DA,$9C,$C7,$FA,$F8,$7E,$D9,$BE
	.dc.b	$0F,$F5,$DC,$60,$D4,$66,$5E,$4F,$FB,$3A,$3F,$D9,$64,$9F,$F6,$59
	.dc.b	$C4,$FF,$4E,$CB,$2E,$11,$F0,$0F,$B2,$5E,$0F,$AF,$82,$9E,$5F,$A4
	.dc.b	$95,$EF,$7D,$9E,$67,$FB,$83,$F1,$4F,$F7,$AC,$9A,$47,$E2,$79,$FF
	.dc.b	$F0,$02,$3F,$DF,$ED,$8C,$F9,$5C,$DA,$73,$3F,$D0,$18,$3D,$0D,$47
	.dc.b	$B9,$AF,$FB,$2A,$34,$F1,$34,$53,$5E,$8F,$FB,$E3,$B7,$E7,$36,$3B
	.dc.b	$B5,$F9,$6D,$7B,$7A,$57,$5F,$90,$DB,$34,$C7,$A8,$A7,$E8,$BF,$76
	.dc.b	$30,$D8,$C9,$F0,$23,$2F,$C2,$F3,$2B,$F7,$9A,$FC,$16,$E1,$F6,$DC
	.dc.b	$C7,$E0,$0F,$E0,$84,$3F,$44,$7E,$7F,$AB,$F9,$31,$1F,$C9,$7A,$FF
	.dc.b	$9B,$23,$FA,$17,$F8,$27,$B0,$F6,$7F,$B1,$36,$7F,$B5,$35,$FD,$30
	.dc.b	$70,$FF,$16,$EB,$FD,$7F,$83,$FC,$56,$3F,$61,$ED,$7E,$AC,$8F,$EC
	.dc.b	$5F,$E2,$92,$E8,$FE,$E4,$D9,$FE,$F4,$D7,$C2,$D7,$C1,$97,$BD,$FF
	.dc.b	$0D,$F0,$3E,$1F,$F7,$44,$8F,$E0,$0F,$F2,$3F,$84,$6D,$F7,$7D,$4F
	.dc.b	$6C,$62,$BF,$CA,$47,$F9,$75,$C4,$6A,$78,$4F,$FA,$9A,$3E,$2E,$9F
	.dc.b	$DD,$1B,$93,$2E,$75,$F8,$BA,$D3,$FC,$47,$C7,$FD,$0B,$F3,$9A,$E7
	.dc.b	$9F,$FC,$2D,$18,$8F,$F3,$83,$5F,$CB,$35,$34,$91,$C8,$91,$FE,$27
	.dc.b	$8C,$B6,$32,$F8,$DA,$F8,$CD,$10,$D1,$4D,$18,$21,$82,$1E,$A7,$AB
	.dc.b	$4E,$45,$38,$94,$E2,$7C,$07,$8B,$4E,$25,$38,$94,$E2,$7C,$67,$8B
	.dc.b	$4E,$25,$38,$94,$E3,$1F,$21,$CB,$5C,$C7,$48,$C4,$7A,$48,$78,$CA
	.dc.b	$75,$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$24,$78,$C9,$1E,$E8,$CB,$63
	.dc.b	$2F,$B1,$AF,$B0,$FF,$37,$7A,$9F,$EC,$67,$FB,$53,$5F,$F5,$EF,$99
	.dc.b	$FA,$07,$FD,$7B,$97,$F9,$D7,$F7,$8F,$CA,$C7,$EB,$CF,$F6,$53,$8D
	.dc.b	$3F,$87,$B6,$3F,$BB,$35,$DE,$3F,$CD,$67,$6D,$8D,$B5,$D9,$DC,$FE
	.dc.b	$79,$3C,$DB,$1B,$69,$B1,$1B,$6C,$6D,$A6,$E3,$44,$8D,$36,$35,$18
	.dc.b	$23,$82,$4F,$B1,$A7,$B0,$A7,$B1,$D1,$1D,$35,$D1,$82,$18,$21,$F0
	.dc.b	$1E,$0D,$36,$53,$67,$C0,$76,$D3,$65,$37,$3F,$06,$36,$D8,$DB,$5D
	.dc.b	$CF,$C4,$9D,$B6,$36,$D7,$67,$C4,$3C,$C4,$3C,$45,$9E,$CD,$8E,$4D
	.dc.b	$39,$1C,$09,$1C,$1B,$1C,$1A,$70,$8D,$1B,$CB,$FD,$6F,$EF,$7F,$44
	.dc.b	$D0,$D2,$57,$F9,$4F,$6F,$EA,$51,$FD,$3E,$36,$67,$1F,$F0,$A4,$64
	.dc.b	$FE,$EB,$F3,$EF,$5F,$35,$FE,$5D,$8C,$3C,$8F,$F0,$E0,$73,$3C,$A7
	.dc.b	$8C,$FF,$88,$7A,$1E,$53,$E1,$7F,$C4,$9E,$DF,$F9,$8E,$31,$48,$C4
	.dc.b	$EA,$3B,$D7,$28,$C4,$7F,$6D,$CB,$F5,$59,$FF,$71,$7E,$01,$23,$FC
	.dc.b	$ED,$C8,$F6,$3F,$60,$D8,$FD,$B3,$67,$FC,$0F,$3F,$BB,$C3,$D3,$F6
	.dc.b	$9E,$0B,$46,$1F,$41,$E6,$CD,$3B,$9D,$67,$E3,$AF,$41,$FC,$17,$2F
	.dc.b	$43,$F5,$4F,$E1,$1D,$6F,$9E,$FE,$0C,$D8,$CC,$60,$FE,$21,$F3,$2F
	.dc.b	$47,$FD,$05,$3F,$C5,$E3,$1E,$5E,$A7,$3B,$7F,$A6,$67,$35,$E3,$8F
	.dc.b	$96,$47,$07,$F2,$1F,$AC,$79,$7F,$E7,$09,$FF,$51,$46,$3B,$7F,$91
	.dc.b	$36,$B8,$56,$DF,$8D,$F6,$BF,$E6,$5F,$85,$FC,$13,$63,$22,$32,$D3
	.dc.b	$F0,$7F,$89,$FE,$4A,$3F,$D5,$8D,$8C,$9C,$CF,$C9,$39,$6F,$C4,$FF
	.dc.b	$9D,$7F,$1D,$E2,$D8,$EA,$D7,$62,$3B,$FC,$6E,$63,$F0,$74,$8C,$72
	.dc.b	$23,$13,$C4,$F4,$CE,$A1,$F2,$1F,$E8,$7F,$89,$F5,$B6,$BD,$ED,$7D
	.dc.b	$6D,$7F,$C8,$9D,$5F,$E4,$7F,$C7,$7F,$83,$12,$3F,$DA,$DE,$0A,$75
	.dc.b	$1F,$BA,$E3,$7C,$CB,$E5,$BF,$D3,$18,$7E,$16,$31,$48,$C4,$EA,$39
	.dc.b	$4E,$A7,$FD,$C9,$F6,$4F,$A4,$FF,$2D,$57,$F6,$98,$C7,$8B,$FD,$9D
	.dc.b	$B1,$99,$DF,$47,$E3,$39,$BF,$11,$3F,$EA,$BE,$7F,$C6,$4E,$7A,$3F
	.dc.b	$1F,$5A,$AF,$A7,$F2,$DF,$EB,$1F,$CB,$FF,$0A,$D8,$FC,$4B,$5F,$C2
	.dc.b	$B5,$FB,$7F,$31,$79,$FE,$D3,$79,$8D,$38,$71,$7E,$AB,$3F,$D6,$76
	.dc.b	$7E,$BB,$0F,$9C,$D7,$FD,$89,$3B,$7E,$AF,$27,$F9,$FF,$C7,$FE,$77
	.dc.b	$1F,$CB,$B9,$BF,$A0,$BA,$E3,$FE,$64,$AD,$FB,$97,$F2,$E8,$CC,$63
	.dc.b	$C5,$FE,$67,$1B,$9D,$61,$FC,$05,$B1,$FA,$01,$F8,$27,$E9,$1A,$65
	.dc.b	$C1,$FA,$27,$27,$EB,$BB,$3F,$DC,$11,$92,$47,$EE,$3E,$6F,$C7,$C3
	.dc.b	$FA,$56,$F0,$9A,$63,$AE,$57,$C0,$AF,$D6,$46,$32,$FE,$03,$F3,$BC
	.dc.b	$89,$5B,$C1,$1F,$0A,$1B,$FE,$8F,$77,$E0,$5F,$DB,$49,$3F,$3D,$F8
	.dc.b	$DB,$FF,$BE,$DB,$1C,$0F,$D5,$6F,$F4,$63,$27,$F1,$4D,$78,$4F,$C1
	.dc.b	$38,$1F,$C5,$1C,$B4,$C9,$F5,$E8,$BE,$FF,$C2,$47,$C3,$1A,$8F,$E4
	.dc.b	$7E,$C7,$0F,$C0,$AD,$1A,$47,$88,$FC,$0B,$DA,$47,$2F,$F8,$47,$9E
	.dc.b	$47,$E4,$DA,$FF,$2E,$FA,$2F,$ED,$1C,$06,$92,$3F,$CD,$5C,$7F,$C9
	.dc.b	$7C,$2F,$6F,$CC,$7F,$3A,$6C,$FF,$3E,$6B,$FC,$E9,$AF,$F3,$0F,$A3
	.dc.b	$EE,$47,$AE,$08,$FF,$37,$71,$FD,$FC,$BC,$F5,$FC,$A3,$8F,$FC,$ED
	.dc.b	$39,$8F,$E9,$BF,$6F,$EA,$23,$D1,$E7,$D7,$FE,$F5,$BE,$73,$F8,$C8
	.dc.b	$C3,$F1,$4F,$EA,$FC,$5F,$10,$8F,$99,$F8,$FF,$53,$F7,$BF,$16,$FE
	.dc.b	$B6,$D9,$FE,$BE,$D7,$FA,$DB,$5F,$EA,$3F,$59,$F4,$08,$FC,$27,$CF
	.dc.b	$9F,$BB,$0F,$2C,$FF,$0D,$6B,$F0,$F1,$FE,$D7,$F4,$1F,$68,$91,$F3
	.dc.b	$1F,$BB,$F1,$3F,$8F,$8F,$EE,$6D,$7F,$B7,$B5,$FE,$CA,$D7,$E3,$5F
	.dc.b	$8C,$95,$F5,$38,$BF,$B4,$7D,$6F,$FA,$D5,$FE,$FE,$DA,$FE,$F6,$D7
	.dc.b	$FB,$8B,$63,$FB,$87,$93,$FA,$8F,$C0,$BF,$C6,$F5,$FE,$02,$8C,$4F
	.dc.b	$F8,$29,$B1,$91,$3C,$1F,$88,$7D,$B3,$E4,$61,$E8,$7F,$C2,$7E,$A7
	.dc.b	$ED,$BE,$6F,$86,$1F,$E0,$18,$C3,$A9,$4E,$B9,$7F,$C7,$8F,$0F,$C3
	.dc.b	$7C,$FA,$5B,$F6,$46,$1F,$F3,$1B,$6F,$FC,$DC,$BE,$D3,$FC,$DB,$3A
	.dc.b	$6C,$E9,$AF,$F9,$6A,$78,$3F,$E7,$19,$F7,$3F,$00,$F7,$D7,$D5,$9F
	.dc.b	$04,$71,$9F,$29,$14,$FC,$97,$EF,$BF,$A6,$E6,$FD,$74,$74,$3B,$9A
	.dc.b	$2B,$A7,$FE,$00,$8F,$A2,$FF,$80,$1F,$0A,$F8,$57,$C2,$FF,$26,$9C
	.dc.b	$C6,$2B,$86,$1F,$E9,$8B,$E1,$7B,$8C,$56,$EF,$8C,$E6,$FC,$31,$87
	.dc.b	$99,$FC,$24,$FF,$5A,$10,$FE,$EC,$53,$F0,$E5,$38,$1F,$6B,$1F,$F4
	.dc.b	$EB,$F6,$9A,$FC,$D3,$FD,$54,$64,$A7,$23,$A9,$E0,$21,$F7,$C7,$CB
	.dc.b	$7E,$37,$4F,$E9,$6D,$8C,$CF,$E1,$DA,$77,$3F,$24,$78,$16,$9F,$9E
	.dc.b	$8D,$47,$E7,$4F,$CE,$6F,$FE,$C1,$70,$48,$F3,$90,$F5,$11,$FD,$CA
	.dc.b	$FC,$0B,$F4,$DC,$7F,$D8,$9B,$19,$9F,$DD,$34,$FE,$0A,$79,$F8,$5C
	.dc.b	$9F,$32,$7F,$86,$CE,$63,$E7,$BF,$C5,$0F,$AC,$53,$E5,$9F,$C6,$8F
	.dc.b	$F1,$81,$FC,$80,$A7,$F2,$02,$9B,$F3,$7F,$95,$BF,$EE,$83,$FD,$27
	.dc.b	$CF,$5A,$5A,$FF,$44,$46,$63,$E2,$9E,$33,$E4,$14,$FE,$82,$71,$B7
	.dc.b	$4F,$1A,$FF,$49,$36,$33,$3B,$EA,$FF,$06,$FE,$25,$FD,$42,$0D,$23
	.dc.b	$FD,$68,$ED,$DB,$CC,$8F,$9F,$9F,$FC,$58,$FC,$A6,$BF,$2B,$1F,$F5
	.dc.b	$27,$77,$8C,$91,$E9,$7E,$03,$ED,$24,$F2,$7F,$11,$3F,$EA,$DE,$FF
	.dc.b	$61,$23,$E1,$3E,$FF,$03,$FC,$53,$F3,$BF,$E0,$D9,$34,$8F,$FB,$33
	.dc.b	$C3,$F8,$C4,$7E,$63,$E4,$C3,$C5,$1F,$51,$AF,$D3,$E0,$FF,$38,$2F
	.dc.b	$AB,$FF,$23,$91,$FF,$17,$BF,$E3,$4F,$62,$47,$FB,$73,$C8,$FF,$1F
	.dc.b	$C6,$23,$FD,$0F,$E6,$7F,$9F,$6F,$FC,$A2,$48,$FF,$1C,$BF,$E4,$F5
	.dc.b	$EC,$52,$3F,$E5,$17,$FC,$B2,$47,$FC,$82,$D7,$FA,$EB,$F9,$5C,$7F
	.dc.b	$CC,$CD,$F1,$9E,$27,$6A,$FF,$4A,$4F,$F9,$C9,$FE,$60,$7F,$BC,$3C
	.dc.b	$EF,$F2,$CF,$83,$92,$47,$FB,$EB,$D0,$CB,$4E,$27,$FB,$B1,$CB,$FE
	.dc.b	$FA,$6B,$FE,$F0,$F3,$AF,$27,$FA,$2E,$38,$B6,$BE,$21,$FE,$97,$FC
	.dc.b	$06,$4F,$FB,$C8,$38,$7F,$DF,$EF,$FA,$84,$8F,$FA,$95,$FF,$40,$C4
	.dc.b	$8F,$F3,$4B,$5F,$F0,$2B,$F1,$F2,$F5,$C7,$F8,$82,$30,$7F,$B1,$BE
	.dc.b	$07,$F7,$DB,$3D,$54,$61,$B5,$FC,$1E,$75,$3F,$BB,$F5,$BF,$D0,$BF
	.dc.b	$89,$FF,$0D,$CF,$FD,$AF,$68,$C7,$77,$FB,$62,$30,$7F,$A3,$0F,$F6
	.dc.b	$F9,$83,$FD,$C6,$53,$FD,$D6,$59,$43,$D8,$F0,$CF,$FB,$B0,$E1,$C2
	.dc.b	$F6,$9E,$DF,$F1,$B8,$C3,$63,$33,$FC,$8E,$72,$FF,$25,$E7,$78,$DF
	.dc.b	$90,$8C,$35,$FE,$59,$1A,$E1,$FE,$50,$D8,$F9,$1D,$AE,$5F,$F5,$5F
	.dc.b	$E3,$7F,$CC,$F8,$93,$9A,$4E,$63,$0D,$AE,$33,$FE,$F3,$8F,$BC,$47
	.dc.b	$07,$CD,$DF,$E3,$FB,$DF,$EB,$AF,$99,$FA,$6D,$9D,$70,$A3,$11,$FE
	.dc.b	$C0,$AE,$2A,$BF,$DE,$1F,$8D,$48,$FC,$D3,$5E,$16,$F8,$3F,$F6,$4C
	.dc.b	$FF,$BE,$C9,$3F,$EF,$F7,$F0,$AF,$FB,$3B,$E5,$AA,$69,$1F,$67,$8C
	.dc.b	$F8,$6E,$11,$86,$BE,$98,$D3,$CC,$D1,$4F,$F6,$EF,$1F,$E0,$1F,$EC
	.dc.b	$CD,$7F,$2C,$D7,$7C,$9F,$88,$7F,$B5,$7C,$47,$F9,$41,$FE,$D0,$D7
	.dc.b	$3E,$AF,$F4,$C8,$C4,$7E,$87,$8F,$F5,$6F,$EB,$7B,$3F,$5E,$FE,$D4
	.dc.b	$C7,$67,$F5,$4F,$88,$7A,$34,$93,$FD,$DE,$32,$F8,$B1,$F0,$11,$C1
	.dc.b	$FD,$FF,$47,$C5,$7F,$8D,$23,$E6,$B5,$F9,$4D,$5C,$89,$E3,$1F,$E3
	.dc.b	$E3,$FC,$1D,$E2,$3E,$2F,$FB,$5B,$63,$FC,$18,$D7,$FC,$24,$D7,$FC
	.dc.b	$2D,$B7,$C6,$7F,$83,$35,$07,$0D,$C7,$FC,$21,$6F,$8C,$F4,$63,$E7
	.dc.b	$6C,$66,$BD,$07,$E1,$E3,$DC,$BE,$39,$F1,$DF,$D0,$23,$0D,$8C,$C7
	.dc.b	$F2,$E6,$F8,$BF,$80,$8F,$E6,$39,$7E,$16,$30,$48,$FF,$1A,$79,$2D
	.dc.b	$3F,$AE,$CF,$F3,$69,$CC,$61,$BE,$0F,$F3,$CF,$8F,$7C,$87,$F8,$61
	.dc.b	$B3,$FE,$39,$6B,$FE,$3F,$6B,$FE,$43,$6B,$FC,$CD,$AF,$CE,$6A,$69
	.dc.b	$23,$C0,$48,$F0,$35,$F0,$35,$F0,$35,$F0,$35,$F0,$35,$F0,$36,$3C
	.dc.b	$11,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$A7,$62,$BD,$9E,$84,$3A,$14
	.dc.b	$E8,$53,$A1,$4E,$85,$7A,$3C,$88,$72,$29,$C8,$A7,$22,$BC,$9E,$04
	.dc.b	$38,$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70,$48,$FA,$84,$8F,$92
	.dc.b	$D7,$E0,$35,$F5,$FB,$DF,$EE,$27,$FD,$24,$D7,$FD,$2C,$D7,$FD,$CB
	.dc.b	$87,$E6,$FE,$93,$F8,$0F,$0F,$E6,$DB,$3F,$9B,$6B,$F9,$B6,$BF,$9B
	.dc.b	$6B,$F9,$B6,$E0,$69,$23,$F8,$94,$65,$B1,$99,$F1,$46,$08,$E0,$8F
	.dc.b	$94,$C9,$5C,$BE,$73,$25,$72,$FA,$4C,$95,$CB,$EB,$32,$57,$33,$E9
	.dc.b	$6C,$FA,$67,$C4,$43,$C0,$53,$B1,$4E,$85,$9E,$8D,$8E,$8F,$22,$1C
	.dc.b	$8A,$72,$2C,$72,$6C,$72,$78,$10,$E0,$58,$E0,$D8,$E0,$E8,$91,$A6
	.dc.b	$C6,$A3,$04,$70,$48,$F9,$71,$96,$C6,$67,$E5,$C6,$0F,$A3,$9F,$F3
	.dc.b	$BC,$9F,$13,$FE,$73,$24,$FF,$3A,$7F,$DE,$5F,$AC,$FF,$8C,$63,$67
	.dc.b	$FB,$DD,$C4,$6B,$C3,$FE,$C2,$D9,$D9,$FE,$6B,$D5,$F0,$1A,$8F,$F8
	.dc.b	$01,$18,$76,$7F,$6A,$71,$3F,$E7,$38,$C4,$F1,$7F,$B8,$E5,$FD,$C3
	.dc.b	$E2,$BF,$AE,$12,$3F,$BB,$C6,$5F,$F2,$87,$BF,$DF,$E4,$BF,$D0,$8D
	.dc.b	$8C,$DF,$F7,$A9,$CB,$63,$C5,$F5,$AF,$A8,$C4,$7F,$A1,$E3,$0A,$BD
	.dc.b	$0A,$33,$18,$6C,$7F,$85,$3A,$3F,$80,$F2,$FF,$80,$72,$FF,$08,$C6
	.dc.b	$0F,$E0,$FD,$4C,$FF,$C0,$57,$DC,$F6,$F4,$79,$1F,$F1,$1C,$6B,$2F
	.dc.b	$7C,$E6,$5F,$5E,$FB,$0C,$36,$3A,$1F,$EA,$37,$15,$EC,$6C,$7C,$CC
	.dc.b	$FC,$B8,$7F,$81,$E3,$FC,$97,$8F,$9E,$72,$7F,$20,$E4,$72,$FF,$8C
	.dc.b	$71,$FB,$D1,$8E,$DF,$53,$70,$F1,$9F,$EE,$EA,$E0,$FD,$E3,$27,$F8
	.dc.b	$BC,$BD,$7F,$E6,$1B,$FF,$10,$1F,$8B,$F6,$9F,$62,$FA,$CF,$91,$18
	.dc.b	$6C,$66,$FC,$33,$97,$C7,$A7,$D3,$FB,$AF,$F2,$0F,$EB,$5F,$8C,$6C
	.dc.b	$FE,$41,$AF,$FA,$67,$74,$62,$3E,$BF,$0D,$BF,$BC,$DA,$FA,$97,$F9
	.dc.b	$88,$C7,$CD,$FF,$94,$3E,$87,$F3,$6D,$C3,$F3,$ED,$7C,$0D,$7F,$0B
	.dc.b	$F3,$FF,$E5,$8F,$B1,$FD,$1B,$63,$F4,$ED,$7D,$0D,$7F,$1B,$F4,$3F
	.dc.b	$E6,$0F,$53,$F5,$6D,$8F,$D7,$B5,$F8,$0D,$7F,$CC,$FC,$7F,$C4,$F4
	.dc.b	$FD,$3B,$63,$FC,$97,$18,$E6,$72,$FE,$86,$7F,$D2,$DF,$A6,$FE,$C2
	.dc.b	$BF,$DB,$0D,$8C,$F6,$FF,$9D,$27,$2D,$8F,$E0,$7D,$9F,$A7,$E2,$FD
	.dc.b	$63,$6B,$F1,$13,$A3,$EF,$AF,$A3,$D1,$23,$2F,$1E,$1F,$F4,$1F,$E5
	.dc.b	$9F,$C3,$9B,$3F,$C4,$9A,$FE,$0D,$AF,$EE,$3E,$96,$E3,$FB,$F6,$E0
	.dc.b	$69,$C3,$4F,$D3,$7E,$09,$FA,$AD,$FF,$D3,$8D,$ED,$FF,$4A,$B7,$13
	.dc.b	$F3,$6B,$F5,$35,$FD,$17,$B3,$04,$AF,$F6,$27,$E1,$7F,$EA,$D7,$F6
	.dc.b	$7E,$A2,$31,$3F,$CD,$E3,$1E,$7F,$F1,$36,$D7,$CA,$F8,$1F,$F5,$3F
	.dc.b	$E3,$BF,$CC,$9B,$1F,$CD,$9A,$FE,$9D,$AF,$E6,$07,$CD,$FE,$61,$A7
	.dc.b	$E6,$0A,$7E,$60,$A7,$F2,$4E,$4F,$86,$FD,$DD,$7F,$A1,$C6,$23,$F6
	.dc.b	$CD,$7F,$95,$F2,$5E,$2F,$E6,$4D,$E2,$FC,$24,$FC,$C3,$FA,$7C,$67
	.dc.b	$D8,$FF,$62,$FE,$0D,$FD,$41,$B3,$FD,$51,$AF,$F0,$56,$BF,$A6,$E9
	.dc.b	$FD,$0E,$5F,$5C,$91,$FE,$B0,$7E,$87,$0F,$D5,$6F,$07,$A7,$83,$E8
	.dc.b	$C6,$2B,$CC,$7B,$8F,$F7,$A9,$7D,$7F,$86,$E0,$F9,$05,$9F,$B7,$97
	.dc.b	$32,$47,$E1,$AF,$D3,$AF,$F5,$7F,$AC,$FF,$72,$FE,$0D,$FD,$C1,$B5
	.dc.b	$FD,$D1,$AF,$F2,$56,$AC,$3F,$13,$F8,$37,$A4,$91,$DC,$3F,$9B,$E1
	.dc.b	$91,$FE,$D4,$6B,$FC,$C7,$DB,$FD,$F1,$1F,$DE,$BF,$C0,$7E,$C3,$FD
	.dc.b	$F7,$F8,$AF,$F8,$19,$B3,$FE,$0A,$6B,$FC,$ED,$AF,$F7,$0F,$B2,$FE
	.dc.b	$C6,$47,$FB,$1F,$2A,$3F,$CF,$FD,$5F,$E3,$3C,$7C,$B5,$F7,$FA,$D1
	.dc.b	$8D,$FF,$C0,$1F,$67,$FD,$B8,$95,$D1,$E7,$CD,$FC,$9A,$B9,$CF,$F6
	.dc.b	$E8,$C3,$F0,$C1,$29,$F8,$87,$F2,$E8,$FE,$BA,$DE,$DF,$F1,$94,$EB
	.dc.b	$1F,$E5,$13,$9A,$E6,$27,$9F,$C0,$F6,$FC,$EF,$8D,$23,$15,$C2,$B6
	.dc.b	$AB,$73,$A8,$F3,$53,$ED,$FC,$24,$8F,$3B,$FE,$1D,$D7,$E6,$47,$BC
	.dc.b	$F2,$B8,$6E,$7F,$11,$EB,$F6,$7E,$C2,$3F,$09,$F7,$FD,$C7,$C8,$FC
	.dc.b	$23,$D6,$D9,$F5,$62,$8C,$47,$A1,$AF,$DD,$DF,$EA,$E3,$E3,$8C,$1C
	.dc.b	$BD,$DF,$9A,$47,$E9,$BF,$43,$ED,$BE,$CF,$E1,$1F,$3E,$7E,$2C,$63
	.dc.b	$83,$C6,$D8,$C8,$9E,$0F,$80,$F8,$7C,$3F,$90,$7E,$41,$93,$EE,$66
	.dc.b	$FB,$BF,$C4,$F2,$72,$C7,$F1,$31,$87,$E1,$B6,$BE,$CB,$5E,$7F,$07
	.dc.b	$FE,$E8,$9F,$92,$D8,$FB,$ED,$76,$7E,$92,$3E,$7F,$D4,$FF,$9D,$DB
	.dc.b	$1F,$E7,$16,$F3,$79,$71,$56,$EB,$DF,$BF,$D2,$E3,$FC,$14,$7B,$F0
	.dc.b	$FD,$1E,$3E,$16,$C7,$84,$F2,$94,$F2,$D7,$CC,$7F,$61,$A7,$CB,$7B
	.dc.b	$9F,$30,$CC,$78,$63,$51,$F3,$CC,$9F,$EA,$1F,$87,$F4,$FC,$3F,$EE
	.dc.b	$3E,$0B,$A4,$62,$7F,$D5,$B7,$FB,$28,$C4,$EB,$35,$F1,$AD,$46,$2F
	.dc.b	$89,$FC,$12,$F8,$CE,$AB,$84,$E6,$B8,$5F,$B7,$8F,$FD,$99,$3F,$E0
	.dc.b	$B2,$1F,$EB,$97,$2D,$D3,$F8,$73,$72,$FF,$53,$BF,$20,$E2,$7C,$E3
	.dc.b	$ED,$69,$FC,$1D,$C6,$9F,$88,$7F,$8D,$1F,$9A,$38,$9F,$CA,$CA,$FF
	.dc.b	$1E,$C7,$F9,$0F,$D4,$7E,$15,$B1,$F8,$A8,$C4,$9F,$0B,$D8,$FC,$99
	.dc.b	$82,$47,$E8,$17,$CF,$FE,$47,$87,$F9,$7B,$63,$33,$B8,$FC,$46,$7F
	.dc.b	$1A,$7D,$47,$C5,$3D,$3C,$0E,$4F,$E7,$3F,$84,$7E,$75,$B8,$7F,$00
	.dc.b	$A4,$E5,$B1,$FC,$35,$F6,$17,$2F,$E6,$BC,$DF,$12,$30,$78,$48,$79
	.dc.b	$08,$7F,$04,$E7,$FE,$97,$F8,$47,$F2,$48,$34,$8F,$F2,$6E,$AE,$CF
	.dc.b	$F3,$03,$C6,$5B,$FE,$8F,$1A,$8F,$B4,$78,$8A,$78,$B0,$FE,$B3,$D7
	.dc.b	$B6,$C7,$43,$F8,$CC,$67,$4F,$E9,$51,$99,$FE,$A3,$FA,$FF,$F5,$08
	.dc.b	$34,$8F,$F3,$EE,$3E,$44,$79,$57,$F6,$9C,$7F,$23,$1F,$9B,$3D,$05
	.dc.b	$3F,$D8,$5D,$5F,$A0,$C3,$FB,$EC,$7F,$6A,$32,$53,$FD,$F7,$1F,$A6
	.dc.b	$79,$1F,$08,$86,$21,$E0,$7F,$83,$E3,$94,$63,$C0,$F0,$4E,$67,$FB
	.dc.b	$EB,$5F,$F0,$F0,$FC,$5F,$FC,$44,$D3,$2E,$0E,$3E,$27,$D1,$93,$49
	.dc.b	$3F,$4C,$F4,$93,$93,$F4,$0F,$AF,$C6,$FF,$43,$FE,$13,$FE,$57,$24
	.dc.b	$7F,$8F,$DF,$F2,$52,$FC,$26,$91,$FF,$26,$19,$3F,$C8,$DF,$84,$FF
	.dc.b	$98,$48,$E0,$FF,$5F,$7E,$07,$FE,$60,$23,$83,$FC,$9D,$F8,$4F,$F9
	.dc.b	$A4,$8E,$0F,$E3,$4F,$FA,$AC,$8E,$0F,$F3,$F0,$9F,$F4,$13,$4C,$B8
	.dc.b	$3E,$9F,$A9,$FE,$36,$9B,$26,$91,$FF,$6B,$3F,$E7,$82,$38,$3F,$D4
	.dc.b	$9F,$59,$FE,$A7,$23,$83,$FD,$60,$3F,$1C,$FF,$5A,$34,$CB,$83,$FD
	.dc.b	$15,$C7,$23,$FC,$E4,$D7,$FC,$16,$FF,$B3,$48,$FF,$3F,$7F,$DA,$55
	.dc.b	$23,$FC,$93,$CA,$D9,$ED,$7C,$19,$CC,$61,$F8,$47,$F9,$27,$87,$FD
	.dc.b	$C8,$BF,$2C,$FC,$DF,$B9,$FE,$5E,$6C,$64,$E1,$19,$7F,$DE,$86,$4F
	.dc.b	$F7,$A1,$4F,$F2,$E4,$BD,$DE,$11,$AA,$FA,$B9,$C8,$FF,$60,$9F,$C3
	.dc.b	$6B,$FD,$8B,$1F,$EA,$C2,$47,$A7,$C9,$FD,$41,$FC,$7F,$91,$F0,$0F
	.dc.b	$80,$FF,$3B,$FE,$09,$FE,$ED,$8C,$5A,$31,$3A,$8F,$F3,$9D,$72,$8C
	.dc.b	$47,$F8,$FA,$7F,$E0,$01,$1F,$EE,$4F,$F0,$2F,$82,$FF,$42,$FE,$3B
	.dc.b	$64,$D2,$3F,$EE,$FF,$C1,$FF,$DE,$44,$70,$7D,$5F,$88,$FF,$47,$FE
	.dc.b	$01,$FC,$39,$B1,$91,$19,$69,$F7,$7E,$2B,$FD,$2F,$F3,$9F,$C5,$9B
	.dc.b	$19,$11,$96,$9F,$85,$F8,$DD,$BF,$E9,$FE,$14,$62,$3E,$EB,$7C,$3F
	.dc.b	$E5,$2E,$7E,$39,$F1,$EF,$90,$FE,$A9,$18,$6C,$66,$FF,$A8,$4E,$5F
	.dc.b	$E4,$1F,$23,$FE,$6D,$E6,$FE,$78,$D8,$FE,$80,$D7,$FD,$56,$D5,$D9
	.dc.b	$FA,$5E,$CF,$EC,$8D,$8F,$EB,$CD,$AF,$E9,$4D,$8F,$D3,$7A,$DF,$D0
	.dc.b	$08,$FF,$44,$D5,$1F,$D3,$7E,$49,$C1,$FD,$C6,$91,$87,$F9,$D3,$70
	.dc.b	$FE,$AC,$D8,$FD,$A7,$C9,$FF,$43,$23,$FA,$37,$A5,$BE,$53,$FA,$E3
	.dc.b	$67,$FC,$86,$D7,$FB,$13,$5F,$EC,$DF,$2B,$FA,$2C,$3F,$B3,$36,$33
	.dc.b	$D5,$E9,$8C,$47,$E0,$7D,$BF,$EA,$C4,$7F,$6C,$FB,$F7,$FF,$51,$7C
	.dc.b	$EF,$D6,$EF,$FF,$73,$B6,$8D,$3F,$BB,$7A,$CB,$FF,$76,$FB,$B8,$6D
	.dc.b	$7F,$7D,$9D,$3F,$E7,$13,$FD,$D2,$E2,$7F,$C0,$1F,$34,$F9,$AE,$35
	.dc.b	$FE,$EE,$8C,$3C,$4F,$F7,$DB,$89,$E3,$3C,$27,$FD,$FF,$E3,$FD,$EE
	.dc.b	$5E,$11,$FE,$96,$AF,$F0,$74,$E5,$BC,$1F,$90,$7F,$C2,$5F,$36,$F8
	.dc.b	$0F,$F0,$93,$6F,$FC,$F2,$D7,$71,$A2,$3A,$6B,$A3,$04,$30,$43,$FC
	.dc.b	$49,$1D,$5B,$1C,$9A,$71,$29,$C4,$12,$39,$38,$0E,$1D,$47,$57,$51
	.dc.b	$97,$71,$97,$67,$84,$DB,$97,$66,$4D,$9E,$52,$9E,$82,$1C,$8E,$07
	.dc.b	$23,$81,$4D,$9C,$8E,$06,$A3,$D0,$48,$F2,$36,$3A,$B5,$E1,$1A,$23
	.dc.b	$A6,$BA,$30,$43,$04,$3C,$A7,$9C,$F2,$98,$8F,$03,$63,$AB,$4E,$85
	.dc.b	$3F,$CC,$9F,$03,$FE,$E1,$7F,$DC,$DF,$03,$F9,$C5,$19,$7F,$DB,$47
	.dc.b	$E6,$BE,$7B,$F0,$7C,$36,$3F,$DE,$27,$E0,$A3,$2F,$E4,$1A,$FE,$09
	.dc.b	$AF,$E2,$DA,$F1,$39,$E0,$8C,$B6,$33,$18,$23,$82,$4F,$94,$8F,$8C
	.dc.b	$C9,$5C,$BE,$53,$25,$72,$F9,$CC,$95,$CB,$E9,$6C,$F9,$DB,$3E,$66
	.dc.b	$CF,$23,$81,$E3,$21,$E1,$29,$DC,$A7,$03,$44,$34,$53,$46,$08,$60
	.dc.b	$93,$E7,$7A,$B5,$E4,$D7,$8B,$5E,$33,$ED,$78,$B5,$E2,$D7,$8B,$5E
	.dc.b	$27,$C3,$8E,$84,$8D,$36,$35,$18,$23,$F9,$2C,$FF,$9B,$7C,$E7,$F3
	.dc.b	$2D,$4D,$24,$FF,$82,$30,$FE,$3B,$1F,$D5,$49,$19,$38,$9F,$D4,$78
	.dc.b	$76,$D3,$27,$F5,$27,$2D,$32,$7F,$4A,$DB,$FC,$27,$F3,$9E,$52,$47
	.dc.b	$FB,$67,$1F,$E9,$D1,$FE,$D9,$E9,$7C,$EF,$F3,$28,$C3,$6B,$FD,$C1
	.dc.b	$3A,$53,$A6,$9F,$CD,$FC,$37,$B8,$F5,$BF,$9E,$FC,$37,$FB,$A6,$73
	.dc.b	$18,$AE,$15,$B5,$5B,$8C,$3F,$3F,$69,$3E,$BD,$3F,$C6,$31,$88,$FE
	.dc.b	$91,$E3,$BE,$BF,$F8,$0B,$63,$F8,$5B,$67,$FD,$F4,$D8,$F1,$FD,$CF
	.dc.b	$95,$FE,$AB,$9F,$B2,$31,$08,$C3,$E8,$38,$94,$FE,$B5,$BF,$F7,$DB
	.dc.b	$F3,$47,$D7,$E2,$FC,$66,$5E,$A8,$FE,$C0,$3F,$5B,$7D,$D7,$F1,$A6
	.dc.b	$C7,$F2,$46,$DF,$08,$E1,$3F,$C8,$AD,$F9,$D3,$EB,$FD,$31,$86,$C6
	.dc.b	$6B,$D0,$7F,$97,$1C,$6F,$E8,$8F,$F6,$16,$5E,$77,$DA,$D7,$D9,$8F
	.dc.b	$8F,$E1,$9F,$23,$FC,$C9,$B3,$FC,$F1,$B3,$B9,$FE,$73,$B7,$E2,$BC
	.dc.b	$5F,$F1,$4D,$FE,$19,$AF,$D4,$E3,$FF,$57,$F8,$3F,$87,$9C,$CF,$D5
	.dc.b	$9F,$F4,$04,$FE,$2F,$CE,$CD,$FE,$59,$B3,$AA,$FC,$9F,$E1,$5F,$E1
	.dc.b	$4F,$A3,$FC,$93,$63,$22,$32,$D3,$F3,$7E,$76,$DF,$33,$5F,$23,$70
	.dc.b	$FD,$3B,$6D,$F8,$5F,$F8,$93,$CF,$FD,$13,$6F,$F4,$ED,$7E,$13,$5F
	.dc.b	$F1,$5F,$94,$DF,$10,$8C,$2A,$FF,$34,$11,$FD,$9F,$7F,$3C,$BF,$76
	.dc.b	$DC,$7F,$81,$57,$ED,$63,$1E,$6F,$E9,$2B,$F8,$0B,$6B,$FD,$11,$C9
	.dc.b	$FB,$B8,$FD,$4E,$0E,$3F,$C0,$BF,$A0,$6B,$FE,$41,$6B,$FE,$46,$6D
	.dc.b	$FF,$0E,$FA,$6F,$F2,$4F,$E5,$9F,$C1,$DB,$5F,$C3,$1A,$FE,$1D,$AF
	.dc.b	$E1,$7C,$5B,$5F,$C6,$1B,$97,$F8,$0E,$FF,$89,$47,$0E,$AF,$F7,$1F
	.dc.b	$47,$EA,$BD,$5F,$F5,$AE,$3F,$A8,$9C,$B6,$33,$AF,$EA,$7B,$BF,$81
	.dc.b	$4F,$F1,$9A,$FE,$02,$7E,$8F,$AB,$F4,$C4,$8C,$1F,$60,$FE,$27,$CF
	.dc.b	$FE,$65,$FC,$7B,$F9,$63,$63,$F9,$93,$5D,$9F,$E6,$E8,$DF,$D5,$D8
	.dc.b	$DB,$5E,$AD,$7F,$8F,$4A,$31,$BA,$19,$FE,$05,$B5,$FC,$2A,$75,$39
	.dc.b	$3F,$03,$C3,$F7,$27,$F0,$58,$FB,$F9,$BF,$81,$47,$BC,$F0,$F9,$DD
	.dc.b	$48,$F5,$C7,$FD,$5F,$B7,$CD,$DF,$F8,$94,$FF,$16,$DB,$E6,$E1,$D0
	.dc.b	$91,$F8,$FC,$7C,$0F,$C8,$FB,$2F,$51,$1E,$CF,$A2,$37,$1F,$95,$6B
	.dc.b	$F1,$BC,$0F,$84,$47,$CC,$F1,$7E,$DD,$78,$9E,$47,$98,$FC,$23,$EB
	.dc.b	$BF,$85,$B7,$CF,$FB,$5F,$E4,$BA,$1A,$4B,$DC,$3E,$C3,$FD,$5B,$F1
	.dc.b	$DF,$86,$6C,$FE,$19,$AF,$E3,$1A,$FF,$5D,$FD,$17,$FC,$01,$97,$F7
	.dc.b	$96,$C7,$F7,$F6,$BF,$D9,$9A,$FF,$42,$3D,$64,$3F,$A3,$E5,$E6,$C7
	.dc.b	$D1,$5F,$9A,$6C,$EC,$E4,$70,$8F,$ED,$39,$7E,$47,$ED,$FE,$19,$27
	.dc.b	$E0,$BF,$1B,$CA,$FF,$66,$FD,$57,$E9,$9B,$3F,$A6,$6B,$FA,$C6,$BF
	.dc.b	$A0,$F9,$9F,$9E,$47,$E6,$BF,$E8,$CE,$9F,$DD,$C7,$94,$F4,$1F,$91
	.dc.b	$C7,$F6,$BF,$67,$F6,$09,$1F,$59,$FB,$7F,$75,$FA,$92,$47,$E7,$5F
	.dc.b	$E4,$F5,$DA,$3E,$19,$0E,$C7,$FB,$9F,$ED,$5F,$94,$3E,$6F,$FD,$57
	.dc.b	$B6,$63,$15,$FD,$16,$31,$3A,$8F,$F7,$9D,$72,$8C,$47,$F0,$2F,$BC
	.dc.b	$FF,$1D,$D7,$F0,$E6,$C7,$F1,$46,$BB,$3F,$C7,$71,$F1,$30,$F5,$47
	.dc.b	$A0,$F9,$07,$F1,$CF,$A8,$EF,$3F,$C3,$5B,$CC,$69,$AF,$CA,$5F,$67
	.dc.b	$49,$7E,$D3,$FC,$73,$1A,$70,$7F,$1C,$FB,$EF,$3E,$1F,$C6,$DB,$81
	.dc.b	$A7,$B2,$7D,$07,$D1,$C7,$D6,$7D,$5A,$5F,$7E,$0F,$F2,$35,$78,$B1
	.dc.b	$F2,$C6,$23,$FC,$95,$3B,$AF,$8D,$DB,$FC,$6B,$0E,$AD,$27,$3F,$F0
	.dc.b	$7B,$F9,$20,$FF,$76,$FB,$CF,$80,$C7,$B9,$FD,$27,$DF,$4E,$69,$39
	.dc.b	$8C,$36,$B8,$E9,$FE,$F6,$8C,$57,$08,$C1,$1F,$F2,$E1,$FE,$A5,$21
	.dc.b	$FD,$7C,$A7,$F0,$22,$9C,$0F,$F4,$56,$9F,$5D,$F8,$DA,$7C,$CD,$7F
	.dc.b	$D1,$D1,$F2,$CF,$F3,$A9,$4F,$98,$7F,$41,$3A,$94,$F1,$1F,$D9,$B6
	.dc.b	$F9,$1F,$09,$DB,$83,$F0,$12,$8C,$4E,$AF,$84,$E5,$F3,$DF,$92,$3E
	.dc.b	$DB,$63,$D1,$1E,$63,$25,$7F,$3C,$DB,$FC,$F3,$78,$79,$46,$23,$DB
	.dc.b	$C2,$E1,$70,$FF,$A0,$67,$DC,$7C,$D3,$DE,$7E,$D4,$F4,$17,$3F,$81
	.dc.b	$C5,$ED,$8C,$65,$FC,$11,$F4,$72,$7F,$03,$6F,$2F,$F0,$A7,$F3,$51
	.dc.b	$97,$CB,$8F,$91,$FA,$39,$79,$F2,$F3,$36,$3F,$C3,$3C,$DF,$49,$FF
	.dc.b	$0C,$B5,$FF,$0F,$B5,$FF,$09,$B7,$1F,$F0,$97,$3F,$C1,$6C,$FF,$A8
	.dc.b	$DA,$FF,$32,$3C,$38,$79,$5F,$8A,$7E,$44,$F9,$07,$F8,$BB,$A3,$A6
	.dc.b	$C6,$92,$3B,$47,$FB,$37,$E5,$7F,$A7,$41,$A4,$7F,$C6,$BD,$9D,$08
	.dc.b	$F2,$7E,$D3,$F6,$8F,$21,$0C,$61,$FE,$03,$7C,$3D,$BE,$62,$3E,$ED
	.dc.b	$63,$F4,$B2,$F8,$0F,$F7,$7F,$3F,$E4,$77,$FF,$BF,$63,$FB,$F4,$60
	.dc.b	$F8,$3E,$07,$F0,$DE,$FF,$F0,$54,$FF,$BA,$CF,$F0,$41,$4F,$F2,$3C
	.dc.b	$6D,$FF,$29,$E3,$E4,$7F,$BF,$1E,$83,$CA,$7D,$23,$E0,$13,$C7,$FF
	.dc.b	$63,$BF,$E4,$42,$1F,$E3,$93,$E8,$F8,$7F,$83,$DF,$FD,$B5,$CF,$FE
	.dc.b	$64,$8C,$1F,$89,$9F,$F2,$79,$FA,$7A,$EA,$F3,$3A,$9F,$E6,$EF,$17
	.dc.b	$A1,$1F,$E0,$87,$F9,$1C,$A7,$F8,$E8,$E2,$7F,$A5,$FC,$5F,$B7,$3D
	.dc.b	$A3,$F1,$A7,$DB,$E6,$FD,$13,$F9,$2E,$1F,$F3,$67,$8F,$EB,$24,$76
	.dc.b	$3F,$CC,$65,$3D,$67,$80,$A7,$AC,$FF,$41,$F9,$1F,$14,$8F,$39,$FF
	.dc.b	$3A,$B7,$4F,$F3,$F2,$FC,$83,$FD,$0E,$D3,$FD,$0E,$53,$FD,$0E,$53
	.dc.b	$FD,$0E,$7F,$53,$7F,$D2,$AD,$3F,$D2,$A5,$3F,$D2,$A7,$13,$D2,$7C
	.dc.b	$B3,$D4,$53,$E4,$9F,$86,$C3,$E4,$BF,$33,$93,$FD,$63,$B7,$F6,$38
	.dc.b	$FE,$C9,$18,$70,$D7,$38,$FF,$AE,$A7,$FD,$7F,$0F,$CB,$29,$1E,$99
	.dc.b	$7F,$B1,$DD,$1F,$EC,$F2,$9F,$EC,$F2,$87,$FD,$B0,$E1,$B1,$F6,$74
	.dc.b	$FC,$34,$FE,$24,$FC,$B6,$BF,$C7,$7C,$F7,$A3,$FC,$85,$B3,$F6,$DB
	.dc.b	$8F,$FB,$BE,$31,$CB,$FE,$05,$F9,$1F,$F7,$C4,$1A,$47,$E4,$0F,$51
	.dc.b	$E3,$71,$8C,$C6,$1E,$46,$23,$4F,$D1,$8D,$4F,$4D,$58,$BD,$5D,$DF
	.dc.b	$C5,$8C,$3D,$0F,$F0,$E4,$F0,$9D,$B6,$34,$71,$F5,$BF,$C4,$5F,$1F
	.dc.b	$DE,$D1,$89,$FB,$31,$85,$3B,$9D,$47,$82,$7F,$DB,$CE,$8E,$3C,$BA
	.dc.b	$7D,$0D,$BF,$6B,$65,$EC,$FF,$8D,$7D,$EF,$23,$6F,$C6,$D7,$C0,$D7
	.dc.b	$E2,$7B,$97,$0F,$92,$DB,$34,$E1,$3F,$AA,$3F,$14,$7E,$37,$8B,$FA
	.dc.b	$6F,$3F,$EA,$23,$F8,$97,$81,$FD,$11,$FE,$9F,$D0,$F7,$7F,$95,$C7
	.dc.b	$EA,$1A,$FE,$B1,$AF,$F8,$49,$AB,$47,$BD,$FD,$71,$B1,$FE,$17,$6B
	.dc.b	$F9,$03,$84,$79,$AF,$C9,$1F,$BF,$6E,$1E,$9D,$3F,$17,$EF,$FE,$8C
	.dc.b	$95,$FA,$37,$FA,$07,$C0,$7F,$55,$8F,$E0,$8D,$7F,$CD,$0D,$4D,$AF
	.dc.b	$8D,$1E,$1C,$FF,$AB,$BF,$85,$F8,$1F,$D9,$11,$FD,$93,$FD,$3F,$E0
	.dc.b	$DD,$3F,$E7,$5F,$88,$F7,$46,$29,$18,$9D,$47,$EC,$A7,$53,$F1,$70
	.dc.b	$FF,$15,$37,$4F,$F2,$0B,$71,$E7,$7F,$BA,$AD,$BF,$BC,$F8,$8F,$F3
	.dc.b	$FB,$FE,$81,$3F,$93,$38,$F5,$BE,$93,$63,$87,$AD,$FE,$8D,$9F,$AD
	.dc.b	$D9,$FC,$05,$BD,$B9,$AF,$E0,$47,$53,$FD,$29,$3B,$7C,$99,$FF,$96
	.dc.b	$2B,$D0,$48,$CE,$5C,$57,$C3,$BE,$03,$FC,$4A,$D8,$FF,$30,$35,$FC
	.dc.b	$53,$5F,$E3,$6D,$7F,$A2,$35,$FF,$54,$35,$EF,$1D,$88,$F6,$6B,$D9
	.dc.b	$AF,$66,$BD,$9A,$F6,$6B,$D8,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$53
	.dc.b	$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A
	.dc.b	$21,$A2,$9A,$30,$43,$04,$8F,$A8,$48,$F9,$2D,$8F,$80,$D7,$D2,$D7
	.dc.b	$C8,$D7,$C2,$D7,$C0,$D7,$C1,$1D,$88,$F6,$6B,$FC,$5F,$DA,$FF,$75
	.dc.b	$DF,$FA,$C5,$B1,$9F,$6B,$F3,$4F,$FB,$D2,$7F,$D6,$DF,$BC,$3D,$3F
	.dc.b	$F7,$C6,$BF,$C8,$63,$15,$FD,$B4,$FF,$0D,$38,$8D,$4F,$09,$FE,$EF
	.dc.b	$39,$6C,$65,$E2,$D4,$23,$83,$A3,$C0,$69,$CB,$63,$93,$C5,$B1,$A8
	.dc.b	$C1,$C8,$B3,$E8,$71,$1B,$24,$F2,$24,$69,$B3,$C4,$91,$EE,$34,$47
	.dc.b	$2F,$AD,$AF,$A4,$D1,$C5,$CB,$E6,$30,$43,$91,$4E,$4F,$09,$F4,$C7
	.dc.b	$43,$64,$36,$71,$3B,$19,$29,$A3,$0F,$48,$ED,$3F,$0A,$79,$4E,$5B
	.dc.b	$19,$78,$B5,$E8,$43,$B9,$23,$93,$C0,$F5,$1B,$72,$D8,$E8,$F2,$E0
	.dc.b	$FF,$7E,$7C,$47,$F5,$12,$47,$F3,$A7,$FA,$2A,$69,$1F,$E8,$EF,$F6
	.dc.b	$17,$47,$F6,$1E,$4F,$CB,$C6,$49,$1F,$D6,$FE,$D7,$2D,$38,$9F,$D7
	.dc.b	$1C,$B4,$C9,$FD,$53,$1F,$D1,$FC,$4F,$98,$91,$FD,$69,$FF,$0A,$7E
	.dc.b	$35,$FD,$A1,$BC,$9C,$27,$FC,$CF,$39,$6C,$65,$FF,$2C,$F8,$6F,$C0
	.dc.b	$FF,$63,$F0,$9F,$E6,$38,$C4,$A3,$13,$A8,$FF,$47,$CE,$A7,$E5,$E3
	.dc.b	$27,$FB,$7B,$5F,$F0,$EF,$D9,$FE,$DB,$B4,$D2,$47,$EF,$BB,$D8,$3F
	.dc.b	$04,$F0,$4F,$A1,$AF,$FA,$66,$34,$F9,$CF,$2C,$69,$FF,$44,$61,$FC
	.dc.b	$1B,$AF,$FD,$0C,$F9,$63,$E8,$1E,$1A,$7E,$11,$FC,$57,$0F,$1B,$63
	.dc.b	$33,$B8,$FE,$35,$DC,$F3,$5F,$01,$86,$D7,$F1,$A9,$D4,$E4,$F2,$63
	.dc.b	$27,$F8,$B6,$5E,$67,$E5,$9F,$EC,$9F,$99,$FF,$31,$7D,$47,$F2,$96
	.dc.b	$C7,$F2,$66,$D7,$FA,$A1,$B1,$F6,$3F,$06,$FE,$5F,$F2,$3F,$7D,$B1
	.dc.b	$91,$19,$69,$F8,$0F,$C0,$B6,$FE,$29,$AF,$D3,$6E,$FF,$69,$BB,$FE
	.dc.b	$02,$5F,$86,$FE,$3F,$8B,$FA,$0B,$63,$27,$E3,$A3,$2F,$FB,$7B,$F6
	.dc.b	$5F,$F6,$C3,$A3,$FD,$66,$E2,$7F,$D7,$8D,$8E,$18,$72,$77,$E5,$FF
	.dc.b	$50,$F4,$36,$D8,$EC,$D7,$67,$FB,$18,$FF,$69,$FA,$BF,$82,$9F,$F4
	.dc.b	$33,$63,$3C,$BC,$5F,$F6,$F3,$6D,$87,$F9,$82,$FC,$FD,$59,$8C,$5F
	.dc.b	$BE,$39,$72,$FE,$37,$E2,$BF,$18,$49,$FC,$67,$3A,$3F,$09,$E8,$39
	.dc.b	$A9,$2B,$F0,$D1,$97,$F6,$18,$BC,$DF,$EE,$7E,$BF,$CA,$6C,$FD,$16
	.dc.b	$BF,$41,$AB,$5F,$6F,$9B,$86,$C2,$BE,$AC,$E6,$CF,$2D,$3F,$4B,$5E
	.dc.b	$CE,$BF,$D3,$34,$FE,$F5,$E2,$FF,$3E,$7F,$9E,$10,$FF,$05,$87,$4F
	.dc.b	$F8,$3B,$A9,$FA,$1B,$B3,$F9,$9C,$A7,$31,$86,$C7,$F3,$66,$F8,$BF
	.dc.b	$97,$E1,$93,$F9,$73,$E9,$15,$FC,$9E,$DF,$E1,$E6,$C6,$8F,$CC,$C6
	.dc.b	$5F,$CB,$99,$3E,$66,$AF,$9C,$F3,$CF,$E2,$9B,$19,$9F,$E9,$73,$96
	.dc.b	$C3,$E6,$B6,$48,$CB,$CB,$E7,$64,$70,$6E,$3F,$EB,$BF,$1E,$4F,$FA
	.dc.b	$91,$AF,$CD,$D7,$E6,$28,$FE,$25,$18,$6E,$BF,$D4,$E7,$E8,$4E,$5F
	.dc.b	$DF,$67,$9C,$FD,$57,$FB,$C8,$C6,$7E,$F8,$E7,$4F,$C3,$E9,$1F,$E5
	.dc.b	$18,$79,$23,$F8,$2F,$C7,$FF,$94,$3F,$0F,$B4,$69,$1E,$39,$FF,$95
	.dc.b	$BF,$0F,$EE,$48,$C8,$7F,$DC,$11,$23,$FD,$92,$D7,$F8,$BF,$D1,$7F
	.dc.b	$4D,$CC,$D2,$47,$ED,$27,$FC,$CB,$F2,$DF,$C7,$1B,$1F,$C8,$DA,$FF
	.dc.b	$22,$6B,$F8,$8F,$A9,$EC,$47,$F7,$EF,$F9,$A3,$97,$F6,$3E,$C7,$EB
	.dc.b	$3A,$4D,$25,$FF,$2C,$FA,$6F,$41,$30,$FE,$12,$F3,$AE,$33,$EF,$9F
	.dc.b	$F3,$ED,$F6,$8C,$71,$FF,$22,$FA,$7F,$84,$4A,$FE,$26,$FC,$5F,$4B
	.dc.b	$FD,$0B,$FA,$3F,$F3,$C6,$CF,$F4,$36,$BF,$D0,$9A,$FE,$A3,$F0,$1F
	.dc.b	$9C,$47,$F8,$FB,$E7,$D3,$DB,$1F,$58,$87,$F4,$CF,$BD,$FA,$C4,$7F
	.dc.b	$93,$BF,$67,$EC,$3F,$D3,$3C,$DF,$D4,$1B,$3F,$D5,$DA,$FF,$56,$6B
	.dc.b	$F1,$36,$F8,$EF,$80,$C5,$F9,$1F,$EB,$BF,$07,$FE,$B4,$D7,$FB,$13
	.dc.b	$63,$27,$F1,$B8,$CB,$4F,$EC,$BC,$5F,$EA,$EF,$C3,$FF,$B4,$C7,$F0
	.dc.b	$C8,$C4,$FF,$6D,$6C,$70,$13,$C1,$FE,$59,$87,$A1,$FE,$5F,$C3,$F7
	.dc.b	$5F,$51,$AC,$7F,$A6,$FD,$9F,$FA,$FB,$E1,$3C,$39,$7F,$62,$8C,$3F
	.dc.b	$DF,$5B,$5F,$DF,$1A,$FF,$4D,$FD,$3F,$F2,$AF,$B4,$C5,$FC,$36,$99
	.dc.b	$A4,$F0,$E4,$F0,$AF,$BC,$FF,$30,$CF,$F4,$83,$6E,$2B,$FD,$AB,$18
	.dc.b	$7F,$CC,$8E,$E7,$94,$60,$F8,$79,$FC,$57,$ED,$DF,$E8,$DF,$88,$FF
	.dc.b	$56,$7C,$EF,$93,$0F,$55,$FA,$1B,$7E,$98,$F9,$B9,$FC,$FB,$EF,$5F
	.dc.b	$87,$7B,$9A,$C3,$C4,$F8,$0F,$07,$C1,$3E,$09,$F0,$5F,$EF,$2F,$3F
	.dc.b	$FD,$FD,$39,$8C,$69,$FD,$B2,$31,$6A,$DD,$F2,$8C,$56,$E3,$13,$AB
	.dc.b	$E7,$5F,$F0,$02,$F9,$CE,$63,$89,$FA,$23,$E3,$1F,$8B,$2E,$5F,$C6
	.dc.b	$1B,$8F,$E1,$1F,$99,$19,$6C,$65,$F8,$E7,$0C,$BA,$E7,$F3,$E3,$EF
	.dc.b	$9F,$80,$D3,$A4,$7D,$F9,$FC,$34,$E5,$AF,$9C,$EE,$FA,$1F,$41,$83
	.dc.b	$F0,$11,$F8,$BF,$3F,$F2,$7B,$27,$FD,$15,$3F,$8D,$8C,$4E,$A7,$27
	.dc.b	$84,$F3,$9F,$A0,$3F,$14,$62,$7D,$8F,$EC,$3B,$9C,$1F,$9C,$9F,$D8
	.dc.b	$4F,$F8,$E5,$A6,$E3,$51,$F2,$0F,$51,$0C,$1F,$C1,$FF,$10,$ED,$C9
	.dc.b	$EB,$6C,$1A,$78,$F9,$7E,$4D,$FF,$99,$1B,$81,$A7,$CE,$AF,$96,$78
	.dc.b	$0F,$A0,$59,$FF,$36,$F3,$7F,$8F,$E3,$F9,$91,$FE,$3D,$2F,$95,$FC
	.dc.b	$DD,$C1,$FC,$FC,$FC,$1D,$BA,$7E,$D5,$FF,$9E,$1B,$19,$9F,$BD,$C3
	.dc.b	$F8,$F7,$C8,$79,$8F,$29,$FD,$4C,$EB,$D5,$C8,$8F,$83,$AF,$F0,$5E
	.dc.b	$27,$F5,$F8,$FF,$2B,$9F,$D7,$CA,$7F,$5F,$29,$FE,$46,$EE,$F2,$91
	.dc.b	$F5,$BF,$93,$7C,$84,$BF,$43,$E6,$CB,$FD,$2B,$DF,$EE,$24,$7C,$47
	.dc.b	$94,$F0,$8E,$5C,$DF,$AD,$79,$66,$F0,$3E,$39,$23,$E6,$BF,$2B,$B3
	.dc.b	$F4,$1F,$2F,$FE,$1D,$93,$48,$FF,$B0,$7C,$2F,$A2,$47,$EB,$BE,$4C
	.dc.b	$3C,$51,$E4,$E9,$FA,$B1,$E5,$E2,$FF,$3D,$FD,$67,$F9,$20,$95,$FE
	.dc.b	$3A,$7F,$C8,$7F,$3D,$23,$FD,$A1,$E3,$FE,$D6,$31,$1F,$E5,$FC,$FF
	.dc.b	$1C,$D3,$F1,$67,$F9,$71,$CB,$4C,$F0,$7F,$9D,$7C,$FF,$F2,$E4,$1A
	.dc.b	$47,$FC,$9A,$FF,$97,$88,$E0,$FE,$1E,$FF,$1B,$23,$83,$FD,$D1,$E6
	.dc.b	$7F,$0D,$F0,$BF,$DC,$AD,$8C,$9F,$CE,$4F,$F7,$5B,$97,$FD,$DB,$E6
	.dc.b	$D7,$FD,$26,$DB,$FF,$6D,$B7,$8B,$FD,$1D,$3C,$CF,$F4,$CF,$99,$92
	.dc.b	$39,$AF,$F4,$B6,$1F,$EA,$4F,$A2,$93,$FE,$A2,$6B,$FD,$CB,$DB,$7B
	.dc.b	$7F,$C9,$E7,$F9,$5C,$62,$51,$89,$D4,$7F,$AD,$63,$FD,$0F,$52,$3F
	.dc.b	$CF,$2D,$7F,$B5,$BF,$D1,$48,$FF,$3C,$7F,$BC,$69,$FE,$D6,$8B,$E1
	.dc.b	$BD,$2F,$E8,$13,$99,$FF,$69,$B5,$FF,$05,$E9,$FE,$E0,$5B,$7F,$B9
	.dc.b	$1A,$7F,$84,$E3,$51,$93,$FD,$E9,$1A,$3F,$DE,$45,$3F,$D8,$FE,$A7
	.dc.b	$F4,$B8,$D9,$FE,$0F,$E4,$7A,$BF,$F0,$06,$B0,$D7,$FD,$D3,$1A,$F5
	.dc.b	$BF,$AD,$36,$3F,$B9,$E2,$CF,$FD,$AD,$1F,$D6,$63,$1E,$AF,$C2,$9C
	.dc.b	$CF,$E7,$5B,$C2,$D3,$FC,$15,$2F,$7F,$FB,$2C,$61,$BE,$67,$F6,$56
	.dc.b	$CE,$CF,$F1,$57,$B4,$F6,$B5,$18,$6B,$FD,$B4,$46,$DE,$46,$5C,$4F
	.dc.b	$8B,$DB,$E3,$3F,$EF,$26,$E3,$E9,$6E,$26,$C7,$F7,$BF,$C3,$BF,$DF
	.dc.b	$E4,$AF,$F7,$EB,$F9,$54,$D2,$3F,$E5,$3E,$7F,$CE,$47,$E7,$BD,$EF
	.dc.b	$F0,$37,$C6,$DA,$34,$8F,$1F,$81,$DB,$4D,$96,$3F,$53,$1A,$8F,$F0
	.dc.b	$97,$17,$DF,$7F,$0B,$C7,$F8,$2E,$4F,$C3,$46,$9A,$FD,$5E,$DF,$E9
	.dc.b	$DB,$9A,$48,$FC,$4F,$1F,$E9,$DB,$D1,$F8,$72,$3F,$E3,$4D,$BE,$B3
	.dc.b	$FB,$EE,$3F,$EE,$71,$88,$F8,$D1,$D5,$FF,$23,$E1,$E1,$7F,$C6,$4D
	.dc.b	$73,$87,$2F,$82,$FF,$16,$C7,$CF,$6B,$FC,$3D,$AB,$85,$82,$73,$2F
	.dc.b	$83,$FE,$24,$FF,$96,$1A,$FF,$9F,$DA,$FE,$4D,$AF,$C9,$3C,$F7,$EB
	.dc.b	$7F,$D0,$D8,$7A,$5F,$C2,$7C,$2F,$93,$1F,$1B,$71,$34,$F2,$FC,$37
	.dc.b	$EB,$BA,$BF,$CD,$6D,$BF,$F4,$93,$5F,$D3,$35,$F2,$47,$C8,$7F,$AB
	.dc.b	$1E,$4C,$7F,$80,$34,$C9,$FA,$FF,$8C,$79,$FF,$D2,$61,$18,$7F,$A3
	.dc.b	$18,$8D,$46,$7B,$FF,$D0,$3D,$EF,$88,$E4,$DB,$E2,$D9,$DC,$F7,$DB
	.dc.b	$F0,$51,$E9,$25,$79,$DF,$E7,$AF,$F4,$03,$F9,$F1,$0F,$E1,$1F,$14
	.dc.b	$F8,$AF,$F1,$53,$63,$FC,$FC,$D7,$FD,$2E,$D7,$FD,$7C,$D7,$9C,$72
	.dc.b	$23,$C9,$AF,$26,$BC,$9A,$F2,$38,$10,$E0,$53,$81,$4E,$06,$88,$68
	.dc.b	$A6,$8C,$10,$C1,$0F,$F4,$C4,$7B,$5B,$1E,$66,$9E,$12,$9D,$8A,$75
	.dc.b	$29,$D4,$47,$56,$C7,$56,$9D,$4A,$75,$29,$D4,$A7,$58,$E8,$48,$E8
	.dc.b	$D8,$E8,$D7,$A3,$5E,$8D,$7A,$47,$22,$3C,$9A,$F2,$6B,$C9,$FF,$3B
	.dc.b	$7B,$9F,$EE,$E8,$FF,$7B,$FB,$BF,$A6,$51,$97,$FD,$DD,$7F,$EC,$FF
	.dc.b	$D1,$9E,$A7,$FB,$2F,$7F,$F7,$4C,$62,$BF,$85,$07,$11,$A9,$E1,$3F
	.dc.b	$D9,$CD,$38,$7F,$B5,$4F,$07,$4D,$73,$1B,$7F,$DB,$91,$96,$C6,$5E
	.dc.b	$A7,$92,$34,$E2,$7C,$8E,$67,$84,$E8,$8E,$09,$1B,$24,$F2,$23,$E8
	.dc.b	$8D,$C6,$23,$D2,$70,$34,$53,$2E,$DF,$59,$92,$B9,$7D,$47,$27,$13
	.dc.b	$E8,$8E,$06,$8E,$A6,$C8,$60,$87,$63,$25,$32,$73,$30,$43,$04,$7C
	.dc.b	$51,$E4,$9F,$87,$3D,$0D,$38,$7A,$CF,$07,$4D,$73,$1B,$7C,$11,$97
	.dc.b	$C5,$E9,$BC,$F7,$A0,$F8,$6F,$C8,$52,$B5,$39,$8F,$CA,$46,$27,$6D
	.dc.b	$8F,$F1,$F7,$43,$DF,$78,$09,$7C,$5F,$E7,$6B,$53,$98,$C5,$70,$AD
	.dc.b	$9D,$1C,$47,$9B,$E3,$5F,$59,$C6,$BF,$51,$18,$78,$9F,$E5,$17,$13
	.dc.b	$C6,$78,$4F,$97,$0F,$47,$CA,$FE,$A5,$B1,$FE,$01,$6F,$07,$9A,$75
	.dc.b	$8F,$FA,$82,$BB,$9F,$E1,$77,$18,$7B,$3D,$0E,$F3,$96,$BF,$DD,$3C
	.dc.b	$EB,$D7,$FE,$74,$AF,$A6,$D9,$FE,$08,$D7,$D1,$7F,$DF,$1F,$33,$71
	.dc.b	$F9,$73,$E6,$3F,$82,$7A,$CF,$CD,$3F,$C5,$76,$F3,$3F,$88,$4E,$7C
	.dc.b	$3F,$B1,$18,$AD,$B6,$33,$5F,$C5,$FE,$B9,$F2,$7E,$C6,$C7,$C4,$0E
	.dc.b	$34,$FC,$7B,$63,$FD,$0D,$8F,$85,$7D,$27,$76,$9A,$CB,$D3,$19,$3D
	.dc.b	$3E,$57,$F8,$17,$C7,$BE,$07,$E4,$5B,$7F,$E3,$63,$BC,$ED,$FE,$61
	.dc.b	$F5,$FF,$EE,$AC,$FF,$46,$D8,$CF,$B3,$F6,$23,$11,$F9,$DC,$FF,$DA
	.dc.b	$58,$F7,$8E,$AD,$75,$87,$76,$C6,$5F,$C4,$7D,$EF,$E1,$B9,$FF,$54
	.dc.b	$DE,$43,$4F,$C4,$F8,$BF,$E6,$FD,$3F,$C8,$C4,$BF,$F2,$23,$FA,$CF
	.dc.b	$BE,$E4,$48,$FC,$BE,$1F,$B7,$8E,$DD,$5F,$B7,$FC,$2F,$F6,$6D,$AF
	.dc.b	$D9,$35,$FC,$8B,$5F,$D0,$78,$D6,$FF,$81,$37,$33,$4C,$7C,$C7,$F0
	.dc.b	$5F,$83,$FE,$02,$DB,$FE,$00,$D7,$F3,$ED,$7F,$5D,$F1,$BF,$C3,$7E
	.dc.b	$CB,$F8,$4B,$63,$F8,$43,$5F,$D3,$B5,$FD,$B7,$4F,$F1,$5F,$C2,$FF
	.dc.b	$88,$B6,$3F,$88,$35,$FD,$7B,$5F,$E3,$9E,$27,$F2,$6F,$B3,$FE,$47
	.dc.b	$E4,$7F,$2C,$6F,$90,$69,$FC,$7B,$E7,$3F,$92,$36,$F4,$7F,$1F,$8C
	.dc.b	$B4,$C9,$FC,$03,$C0,$FC,$21,$1F,$B7,$C1,$FB,$CE,$5F,$E5,$7F,$3B
	.dc.b	$F8,$92,$4E,$9F,$C6,$7C,$FF,$F3,$5F,$91,$FE,$62,$D9,$FE,$60,$D7
	.dc.b	$F8,$7B,$5F,$E3,$DE,$95,$E6,$FE,$80,$DC,$7F,$D0,$D7,$FC,$F2,$31
	.dc.b	$D7,$FD,$0F,$F0,$8F,$E7,$ED,$C3,$F9,$F3,$5F,$E3,$AD,$7E,$7F,$07
	.dc.b	$F5,$8D,$3E,$69,$23,$FA,$AC,$65,$FE,$A0,$3D,$D7,$80,$FA,$6F,$E6
	.dc.b	$71,$86,$C6,$67,$F9,$DC,$E5,$F9,$BF,$39,$FD,$6F,$EC,$3F,$AB,$B6
	.dc.b	$3F,$AB,$35,$FE,$6A,$D7,$F2,$E3,$F0,$AF,$CB,$B4,$FC,$B9,$4F,$CB
	.dc.b	$94,$FE,$99,$C7,$F0,$F8,$7F,$B7,$37,$87,$F9,$DE,$FF,$DA,$A3,$1C
	.dc.b	$72,$D3,$F8,$46,$9F,$DB,$63,$97,$DA,$FF,$77,$FC,$27,$FB,$9B,$67
	.dc.b	$FB,$93,$5F,$EA,$2D,$7E,$AF,$47,$47,$E8,$F4,$FF,$82,$B0,$FD,$19
	.dc.b	$23,$EA,$EF,$F3,$63,$FC,$C0,$D8,$CF,$07,$D3,$9C,$B6,$3F,$C0,$99
	.dc.b	$FF,$83,$7E,$33,$E0,$47,$FC,$00,$DB,$E2,$B7,$6F,$EB,$F3,$CB,$8B
	.dc.b	$D1,$97,$DD,$DB,$16,$69,$FE,$14,$FA,$4F,$F0,$D7,$E1,$36,$4D,$23
	.dc.b	$E8,$BF,$F1,$07,$E1,$3D,$09,$19,$0F,$F0,$6F,$4E,$47,$E6,$B8,$7F
	.dc.b	$D0,$FC,$3F,$E3,$29,$FF,$14,$7E,$F3,$F9,$CE,$9B,$7F,$D1,$E3,$0F
	.dc.b	$F8,$C9,$AF,$F8,$75,$7D,$BF,$F8,$CF,$A3,$9B,$67,$83,$5E,$0D,$5C
	.dc.b	$4A,$7D,$5F,$F8,$9D,$B3,$FE,$27,$6E,$3E,$89,$FF,$48,$65,$EB,$AF
	.dc.b	$74,$FC,$07,$F0,$FD,$7F,$E5,$48,$F7,$FD,$6F,$B4,$93,$FD,$63,$04
	.dc.b	$7F,$8A,$AF,$A4,$F2,$D7,$E4,$E1,$CE,$3B,$9F,$17,$EB,$BE,$41,$1E
	.dc.b	$8F,$CB,$E0,$FE,$7D,$A7,$BD,$B9,$7F,$05,$F1,$7E,$9C,$7F,$43,$BF
	.dc.b	$F3,$46,$1D,$DF,$16,$9E,$D7,$F4,$AD,$77,$97,$98,$91,$F7,$BE,$BF
	.dc.b	$EE,$12,$7E,$03,$F7,$FE,$C3,$FC,$D9,$F8,$47,$E1,$1B,$3F,$81,$6B
	.dc.b	$F8,$16,$BE,$23,$E1,$9E,$62,$9F,$7B,$37,$D9,$FE,$34,$91,$F4,$5F
	.dc.b	$C8,$7C,$EF,$F9,$F7,$F0,$8F,$C2,$CF,$E3,$E3,$1B,$FE,$4D,$B1,$91
	.dc.b	$3C,$1F,$21,$E3,$21,$83,$C9,$F6,$DF,$E8,$BF,$A7,$E1,$7F,$93,$8C
	.dc.b	$3F,$A1,$6C,$FE,$85,$AF,$B7,$D7,$FF,$4A,$7E,$69,$CB,$34,$62,$FF
	.dc.b	$54,$DA,$FD,$53,$5F,$8D,$EE,$FF,$93,$67,$F6,$0D,$8F,$D8,$35,$D8
	.dc.b	$F7,$7E,$03,$F8,$2E,$1F,$C2,$73,$FE,$6A,$3E,$05,$FE,$BF,$DB,$FF
	.dc.b	$55,$7D,$EF,$9D,$B1,$FB,$F6,$CF,$EF,$DA,$FC,$93,$B1,$F7,$8A,$7C
	.dc.b	$E2,$9B,$AF,$12,$F8,$CF,$F1,$CF,$C6,$7F,$AE,$FE,$15,$97,$F6,$E9
	.dc.b	$CD,$27,$2F,$FB,$06,$33,$18,$8E,$9E,$DA,$FA,$5F,$38,$C4,$F3,$8C
	.dc.b	$4E,$A3,$A4,$EA,$BA,$5F,$6B,$FE,$05,$5F,$E3,$22,$1F,$CB,$4A,$68
	.dc.b	$C1,$0F,$94,$3F,$22,$FA,$0D,$CF,$E8,$34,$DC,$7D,$CC,$BF,$DD,$31
	.dc.b	$D8,$FE,$CA,$59,$FE,$55,$B7,$FB,$97,$E1,$BF,$24,$D8,$FC,$5C,$A7
	.dc.b	$2D,$8F,$CA,$AF,$CE,$3F,$35,$C3,$FE,$EF,$6C,$66,$77,$1F,$9F,$E4
	.dc.b	$FF,$7A,$16,$4D,$24,$FC,$38,$F4,$94,$FE,$32,$7F,$54,$E7,$FF,$7F
	.dc.b	$7E,$17,$FA,$B6,$CF,$F0,$19,$4E,$5B,$1F,$D5,$FA,$50,$F9,$8F,$EB
	.dc.b	$F1,$86,$C6,$4E,$05,$3E,$D1,$0F,$39,$67,$FA,$C7,$5F,$6F,$C2,$FF
	.dc.b	$94,$49,$A4,$7F,$B5,$77,$78,$5A,$77,$3F,$93,$39,$7F,$B7,$35,$F0
	.dc.b	$90,$F8,$E7,$C9,$E0,$FC,$2B,$7C,$07,$81,$FE,$58,$8F,$47,$67,$F9
	.dc.b	$4F,$F0,$2F,$E9,$F0,$69,$1F,$EB,$87,$97,$3F,$EE,$DE,$27,$8D,$B1
	.dc.b	$9A,$4F,$9E,$3C,$A7,$C9,$29,$EA,$9F,$F0,$07,$8B,$CE,$BF,$B8,$CF
	.dc.b	$38,$C1,$CF,$2F,$6C,$73,$6B,$CE,$7F,$C0,$EF,$F8,$14,$F8,$07,$F5
	.dc.b	$1F,$1A,$B4,$CD,$FF,$84,$7C,$7F,$C3,$FD,$1F,$F8,$6A,$4D,$23,$FE
	.dc.b	$23,$3D,$04,$3E,$11,$E7,$D7,$E8,$BF,$B7,$3E,$09,$F0,$BC,$FF,$F4
	.dc.b	$17,$E1,$3F,$E4,$62,$47,$F8,$ED,$FF,$22,$7E,$13,$23,$FD,$23,$8B
	.dc.b	$D3,$7A,$3F,$9E,$AD,$57,$E8,$30,$FF,$47,$7E,$13,$FE,$52,$24,$7F
	.dc.b	$91,$DF,$F2,$EF,$C2,$C8,$FF,$30,$87,$0F,$3F,$5F,$FD,$35,$F8,$47
	.dc.b	$F9,$70,$91,$FE,$51,$7F,$DA,$A4,$7F,$CA,$0D,$7F,$CF,$A2,$BF,$D0
	.dc.b	$6D,$3F,$D0,$65,$36,$FF,$A3,$62,$47,$FA,$34,$38,$7F,$DB,$46,$0F
	.dc.b	$5B,$FE,$7A,$23,$FE,$A4,$FC,$27,$FD,$4E,$47,$07,$FA,$C0,$7E,$05
	.dc.b	$FE,$B4,$69,$97,$07,$F9,$FB,$B2,$47,$F9,$CD,$AF,$F9,$25,$FF,$7C
	.dc.b	$91,$FF,$3A,$3F,$ED,$87,$D0,$7F,$B5,$8A,$7F,$99,$B5,$7C,$07,$F1
	.dc.b	$CC,$38,$1F,$EE,$39,$78,$14,$8F,$E2,$29,$F0,$3F,$C9,$27,$FD,$EB
	.dc.b	$3C,$0F,$F7,$A9,$4F,$F7,$A9,$4F,$17,$C1,$F1,$7E,$96,$3F,$EC,$12
	.dc.b	$38,$3F,$16,$FF,$AA,$88,$E0,$F2,$7C,$27,$F3,$6F,$A8,$FE,$2D,$C9
	.dc.b	$F9,$D8,$C4,$E8,$4E,$B2,$4E,$A7,$FC,$ED,$3F,$C0,$3A,$DC,$FF,$C8
	.dc.b	$A3,$0D,$9F,$F5,$24,$73,$F8,$7F,$E8,$FD,$F6,$7F,$92,$1F,$26,$32
	.dc.b	$D3,$27,$FB,$B3,$EF,$FF,$DE,$24,$7F,$DE,$2F,$9F,$9B,$FA,$9F,$6D
	.dc.b	$9F,$E2,$E7,$DD,$8C,$BF,$5D,$AF,$F5,$AD,$7F,$80,$7A,$1F,$D7,$3E
	.dc.b	$87,$E1,$36,$7E,$F3,$5E,$1B,$A3,$11,$F8,$BF,$8D,$B7,$FC,$2D,$B4
	.dc.b	$69,$8F,$8F,$FE,$81,$06,$92,$BF,$AD,$0F,$90,$FF,$1B,$F1,$FF,$32
	.dc.b	$6C,$7F,$53,$6B,$FD,$D8,$E1,$E9,$76,$E9,$91,$FD,$47,$0F,$D0,$B6
	.dc.b	$3F,$B2,$76,$BD,$6F,$F3,$7C,$FF,$87,$4E,$91,$9F,$43,$FC,$49,$18
	.dc.b	$9D,$FA,$3F,$D9,$6B,$53,$C4,$F2,$7C,$8B,$E4,$BF,$AE,$36,$3D,$ED
	.dc.b	$D7,$FA,$7B,$63,$B6,$3E,$07,$FC,$27,$7D,$B0,$F0,$63,$E2,$9D,$9E
	.dc.b	$11,$2F,$93,$7C,$47,$F8,$73,$ED,$BD,$11,$8A,$46,$27,$51,$F0,$A7
	.dc.b	$53,$FD,$93,$E5,$5D,$5F,$DD,$DB,$5F,$69,$BA,$EE,$7E,$7D,$F7,$E6
	.dc.b	$ED,$7F,$8D,$3B,$63,$FE,$30,$8C,$9F,$B7,$D3,$FB,$8F,$37,$F5,$96
	.dc.b	$F1,$7B,$EF,$F2,$F1,$FD,$D9,$FF,$2B,$7B,$7F,$E0,$B8,$FF,$41,$47
	.dc.b	$E3,$9B,$9F,$F7,$E6,$C7,$F1,$FF,$97,$FC,$B9,$1F,$C5,$3F,$9A,$F9
	.dc.b	$8F,$09,$23,$FC,$07,$82,$3C,$5F,$33,$FA,$39,$FC,$EB,$70,$FF,$06
	.dc.b	$B6,$3F,$C5,$CD,$7F,$63,$F3,$5F,$A8,$23,$F9,$B7,$FC,$0D,$EA,$FF
	.dc.b	$A5,$23,$F4,$5E,$27,$F1,$88,$C4,$7F,$96,$A3,$1F,$01,$FC,$6A,$73
	.dc.b	$2F,$9C,$7C,$EF,$C9,$6C,$7F,$88,$1A,$FF,$89,$1A,$FF,$91,$5A,$FE
	.dc.b	$CD,$AF,$F9,$A9,$AF,$78,$EC,$47,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B3
	.dc.b	$5E,$C7,$42,$1D,$0A,$74,$29,$D0,$A7,$42,$9D,$0E,$44,$39,$14,$E4
	.dc.b	$53,$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18
	.dc.b	$24,$7D,$42,$47,$C9,$6C,$7C,$06,$BE,$96,$BE,$46,$BE,$16,$BE,$06
	.dc.b	$BE,$08,$FD,$B7,$C1,$7F,$BD,$A3,$FD,$1C,$D7,$FD,$DA,$7F,$9C,$9C
	.dc.b	$3F,$EB,$4F,$AC,$6F,$FE,$B6,$EF,$FF,$3B,$C6,$27,$FD,$F8,$6E,$74
	.dc.b	$FF,$A6,$E3,$53,$C2,$73,$3F,$D2,$E7,$2E,$E3,$2F,$33,$A8,$8D,$BC
	.dc.b	$63,$6E,$5B,$1A,$8E,$53,$E1,$79,$C6,$5E,$06,$27,$89,$1C,$12,$79
	.dc.b	$11,$F4,$19,$2B,$97,$CC,$64,$AE,$5F,$29,$C0,$86,$8A,$68,$C1,$0C
	.dc.b	$10,$F5,$BD,$5A,$72,$29,$C4,$A7,$19,$F8,$84,$9F,$85,$3C,$E3,$13
	.dc.b	$EB,$38,$4E,$9E,$D1,$A9,$E1,$39,$9F,$04,$E5,$DC,$65,$E6,$75,$3C
	.dc.b	$A3,$3F,$E6,$1F,$3D,$FC,$B4,$91,$FE,$E9,$8C,$BF,$E1,$9B,$29,$FF
	.dc.b	$0F,$46,$1B,$19,$38,$14,$FF,$1E,$7B,$2F,$C1,$39,$FC,$7F,$F9,$8A
	.dc.b	$31,$28,$C4,$EA,$3F,$D8,$73,$A9,$FE,$89,$72,$7D,$1E,$6B,$7F,$CF
	.dc.b	$E5,$FD,$36,$BF,$C2,$FC,$FF,$9E,$F0,$BF,$B6,$C6,$30,$FD,$2C,$61
	.dc.b	$EF,$F5,$BE,$17,$D2,$D5,$D4,$D1,$39,$8C,$3B,$9F,$A1,$9F,$97,$0F
	.dc.b	$44,$7F,$62,$8C,$3E,$AC,$FC,$7F,$45,$E7,$24,$60,$FD,$DF,$E4,$0F
	.dc.b	$B0,$FA,$BF,$47,$FE,$4D,$9F,$AF,$39,$51,$95,$19,$8C,$36,$3C,$58
	.dc.b	$F9,$9F,$EE,$98,$F9,$9F,$46,$3E,$2C,$3F,$0B,$F1,$5C,$9B,$3E,$3C
	.dc.b	$3C,$53,$98,$3E,$0C,$3F,$1B,$F1,$B6,$4D,$23,$E4,$21,$E2,$8F,$33
	.dc.b	$EC,$79,$BE,$16,$BF,$E7,$1A,$46,$23,$E3,$B7,$5F,$B9,$5F,$6F,$6F
	.dc.b	$1D,$7E,$54,$FC,$9C,$64,$FB,$DC,$3F,$8F,$8F,$29,$F2,$33,$FC,$9E
	.dc.b	$9F,$26,$3F,$D6,$3E,$C7,$F3,$0F,$53,$FD,$AF,$1F,$CD,$9B,$9F,$D7
	.dc.b	$6C,$7E,$8B,$EA,$3F,$67,$F8,$0F,$EA,$5B,$19,$11,$96,$9F,$AA,$FA
	.dc.b	$AD,$BF,$B4,$6B,$F9,$CD,$51,$8D,$BF,$3C,$D8,$7D,$67,$F0,$4F,$A5
	.dc.b	$FD,$D3,$6F,$F6,$ED,$7F,$80,$B5,$ED,$C9,$70,$FD,$53,$6F,$F2,$F2
	.dc.b	$BF,$E0,$D3,$9F,$9D,$FE,$27,$F4,$9F,$C2,$5B,$5F,$C1,$9A,$FF,$10
	.dc.b	$6B,$FC,$6B,$D5,$FC,$0A,$9F,$E3,$24,$7F,$8C,$BF,$EF,$6F,$A9,$FD
	.dc.b	$FC,$6C,$FF,$5D,$64,$8C,$CE,$E3,$B9,$92,$C3,$EB,$FF,$83,$36,$BF
	.dc.b	$18,$7F,$B8,$5C,$77,$78,$67,$F8,$57,$0F,$E0,$AB,$FD,$94,$D7,$F1
	.dc.b	$67,$E2,$78,$FF,$C0,$38,$79,$7B,$BF,$9C,$FE,$41,$EB,$6D,$7A,$B2
	.dc.b	$FF,$47,$CE,$60,$FC,$1E,$DD,$C7,$95,$AF,$F2,$29,$46,$2F,$E2,$36
	.dc.b	$1C,$FF,$00,$95,$F0,$0E,$4F,$EA,$8F,$D3,$3F,$CA,$3A,$3F,$52,$4E
	.dc.b	$8F,$D2,$67,$FA,$92,$4F,$C1,$3F,$C5,$78,$FD,$5F,$0F,$FA,$CF,$E1
	.dc.b	$1F,$6D,$B3,$F0,$5A,$FD,$16,$BE,$DF,$62,$D7,$DF,$6E,$7F,$EE,$5B
	.dc.b	$FC,$34,$63,$E3,$3E,$E9,$2F,$DD,$7F,$87,$8F,$B7,$8F,$A3,$2F,$5D
	.dc.b	$7A,$5F,$C8,$78,$3F,$86,$24,$72,$CB,$F9,$E7,$1F,$F2,$0E,$EF,$E2
	.dc.b	$13,$FC,$8C,$FE,$22,$38,$FF,$0C,$DC,$7C,$0F,$F4,$78,$FC,$56,$7F
	.dc.b	$9B,$FB,$4F,$CA,$92,$7C,$0F,$A1,$F3,$1C,$6B,$9B,$F9,$9C,$FF,$3A
	.dc.b	$47,$D0,$F1,$3D,$E4,$79,$1F,$74,$F4,$E1,$EC,$7F,$95,$47,$EC,$FE
	.dc.b	$D7,$F3,$B9,$9A,$49,$FC,$0C,$3E,$B3,$FC,$33,$F8,$F7,$EE,$DB,$3F
	.dc.b	$A1,$6B,$FB,$16,$BF,$9D,$F3,$3B,$91,$EA,$FF,$9A,$31,$F3,$47,$E5
	.dc.b	$F1,$FA,$9C,$BF,$CF,$F5,$FE,$0D,$97,$A0,$95,$E6,$79,$4F,$E8,$2F
	.dc.b	$99,$1E,$50,$FB,$8F,$82,$4A,$F7,$BF,$13,$EE,$7F,$C6,$FF,$7D,$FC
	.dc.b	$65,$B3,$FB,$F6,$BF,$C3,$DA,$FF,$87,$3F,$2F,$7C,$C7,$F0,$A6,$C7
	.dc.b	$EB,$1B,$A7,$F8,$3F,$44,$62,$7E,$CE,$7E,$B2,$75,$7D,$4F,$BC,$FB
	.dc.b	$24,$9C,$1F,$63,$EF,$7F,$92,$91,$FD,$03,$FC,$16,$7F,$0D,$1F,$9F
	.dc.b	$21,$FC,$4F,$EC,$BF,$D5,$D8,$FF,$1D,$6C,$7F,$2D,$6B,$B3,$FC,$07
	.dc.b	$1F,$C8,$FE,$97,$FC,$AA,$D8,$FF,$2A,$37,$A7,$FA,$1C,$7F,$21,$9D
	.dc.b	$BE,$23,$F0,$47,$C2,$BF,$E2,$7F,$65,$FE,$55,$8E,$E7,$F9,$B2,$D1
	.dc.b	$9D,$7A,$C6,$2B,$CE,$DA,$E2,$7F,$94,$A9,$F0,$9F,$E7,$6F,$90,$E3
	.dc.b	$18,$94,$62,$75,$5F,$D6,$AB,$53,$9A,$FF,$18,$CE,$6B,$FC,$6A,$D8
	.dc.b	$C9,$E9,$AE,$DE,$87,$C4,$8F,$82,$7C,$3E,$6F,$CD,$63,$F0,$B0,$E7
	.dc.b	$3C,$5A,$FA,$DB,$87,$C4,$73,$1D,$0F,$E0,$67,$CD,$CB,$F5,$F5,$F8
	.dc.b	$48,$C9,$FE,$F4,$F1,$7F,$BF,$AF,$61,$F0,$2F,$6F,$FD,$3B,$5A,$9C
	.dc.b	$C6,$30,$E4,$4B,$FF,$51,$CE,$95,$71,$BE,$91,$8A,$E0,$FF,$5B,$3F
	.dc.b	$CE,$59,$FF,$64,$C7,$FC,$E9,$19,$3F,$D0,$46,$B9,$3F,$10,$E0,$87
	.dc.b	$F9,$C8,$FD,$96,$DF,$7F,$E0,$7E,$D1,$36,$FB,$0F,$D5,$C3,$C1,$5D
	.dc.b	$70,$F0,$C6,$23,$A9,$F4,$C4,$F3,$7E,$C1,$8A,$EA,$7E,$17,$3F,$C5
	.dc.b	$11,$C1,$F8,$FB,$FF,$14,$CF,$E2,$5A,$FE,$6D,$B1,$F9,$F8,$F4,$63
	.dc.b	$F0,$9F,$80,$79,$8F,$90,$7F,$8A,$78,$7F,$DA,$FE,$FF,$EA,$BB,$11
	.dc.b	$8B,$FD,$E4,$E6,$31,$5F,$C1,$4F,$DA,$14,$E5,$CB,$E4,$B3,$E8,$8F
	.dc.b	$09,$FE,$7F,$3E,$61,$0F,$F1,$77,$37,$91,$A2,$FF,$C9,$D1,$98,$FF
	.dc.b	$21,$35,$FF,$17,$73,$FE,$1B,$1F,$F2,$83,$70,$34,$FC,$5C,$F9,$09
	.dc.b	$B7,$DD,$7E,$49,$FC,$47,$A3,$F9,$EB,$FC,$ED,$AF,$F9,$79,$AE,$FA
	.dc.b	$7C,$1D,$48,$7C,$BF,$F4,$F6,$9F,$D3,$CA,$7F,$4F,$29,$FE,$49,$ED
	.dc.b	$F5,$11,$FF,$6B,$67,$FE,$6D,$8F,$5E,$3E,$99,$F9,$C7,$E0,$5C,$4F
	.dc.b	$F6,$E8,$E3,$9F,$F9,$D5,$AF,$C7,$39,$F0,$7E,$2E,$39,$9F,$10,$A7
	.dc.b	$F8,$04,$E7,$DD,$F3,$C8,$F6,$EF,$F8,$5F,$8A,$7F,$83,$60,$D2,$3F
	.dc.b	$E6,$FF,$0B,$EC,$11,$F3,$BE,$4E,$5F,$D9,$E1,$E5,$9F,$B2,$DF,$17
	.dc.b	$FD,$8F,$1F,$E4,$82,$1F,$E2,$A3,$FD,$31,$D5,$F8,$9D,$3F,$D2,$CD
	.dc.b	$D3,$FD,$43,$CB,$AA,$FC,$91,$C4,$87,$FA,$3B,$C4,$FC,$5E,$1E,$08
	.dc.b	$FF,$44,$78,$BC,$C8,$F2,$3F,$D3,$4F,$E6,$5F,$F2,$81,$FC,$7A,$39
	.dc.b	$3D,$4C,$4F,$C1,$7F,$50,$7D,$33,$F5,$B8,$FF,$AB,$FC,$6F,$51,$23
	.dc.b	$C0,$7F,$A4,$DF,$05,$76,$8F,$F5,$29,$F1,$8F,$01,$FE,$BD,$F1,$FE
	.dc.b	$21,$1E,$75,$FE,$77,$6E,$4F,$23,$F4,$1F,$84,$7F,$A0,$24,$D2,$3F
	.dc.b	$E8,$D3,$25,$7F,$D2,$4E,$0F,$F4,$81,$4F,$39,$F4,$8A,$7A,$0F,$D0
	.dc.b	$6D,$F8,$5E,$4F,$D2,$47,$C6,$BF,$F5,$7B,$E8,$9F,$C0,$11,$F5,$F9
	.dc.b	$9F,$B9,$FC,$87,$FD,$7D,$06,$91,$FF,$63,$78,$1B,$69,$B2,$C7,$FB
	.dc.b	$24,$F9,$CF,$A0,$87,$E0,$8F,$F6,$E1,$F3,$3C,$FF,$CA,$11,$F9,$FA
	.dc.b	$7F,$0D,$F9,$4F,$F7,$64,$1A,$47,$F8,$91,$FE,$FD,$E4,$E0,$E0,$7A
	.dc.b	$5F,$F0,$02,$71,$18,$9D,$BC,$A7,$4F,$FB,$FB,$D6,$7A,$9A,$86,$09
	.dc.b	$CC,$79,$4E,$13,$B9,$D3,$65,$EA,$FF,$29,$F9,$4F,$F3,$54,$62,$D1
	.dc.b	$89,$D4,$78,$DA,$FF,$92,$65,$EC,$62,$3C,$B1,$87,$99,$EB,$8D,$61
	.dc.b	$E9,$6C,$BD,$9F,$E7,$9F,$79,$E4,$6D,$FC,$36,$BE,$16,$BE,$0F,$72
	.dc.b	$E1,$F1,$5B,$66,$9C,$27,$F0,$87,$E7,$CA,$7E,$7C,$FE,$05,$CB,$F9
	.dc.b	$E2,$72,$7F,$39,$BF,$A5,$E9,$7F,$1F,$C0,$D2,$5F,$F0,$61,$EE,$B8
	.dc.b	$7F,$AD,$79,$DF,$84,$8C,$4F,$F8,$80,$FC,$53,$88,$D4,$F0,$9F,$D9
	.dc.b	$F8,$3F,$CE,$08,$E8,$E4,$7E,$5A,$B9,$3D,$27,$93,$96,$BF,$DB,$61
	.dc.b	$DD,$FD,$BF,$D3,$F3,$36,$7C,$AD,$7B,$35,$FE,$B6,$7F,$4A,$BF,$2B
	.dc.b	$C0,$FF,$31,$97,$D1,$EA,$E0,$E6,$53,$FC,$01,$F0,$0E,$CF,$F0,$7B
	.dc.b	$63,$FC,$B6,$D7,$F5,$ED,$7E,$39,$FC,$F8,$FE,$DF,$B7,$EA,$FE,$03
	.dc.b	$ED,$12,$3F,$99,$EE,$8F,$E9,$95,$FE,$70,$9E,$FD,$BE,$22,$3E,$E3
	.dc.b	$F1,$3C,$3F,$C9,$CF,$C4,$5B,$E0,$7F,$10,$49,$EE,$FF,$59,$EB,$FC
	.dc.b	$1F,$AD,$FC,$21,$B3,$FC,$35,$AF,$F0,$9E,$2F,$F1,$24,$FF,$9B,$49
	.dc.b	$1F,$C4,$F9,$7F,$D0,$11,$88,$FE,$33,$EE,$3E,$0F,$FC,$72,$D8,$FF
	.dc.b	$42,$B5,$FE,$32,$D7,$F8,$F3,$5E,$71,$C8,$8F,$26,$BC,$9A,$F2,$6B
	.dc.b	$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3F,$C9
	.dc.b	$D1,$ED,$6C,$79,$9A,$78,$4A,$76,$29,$D4,$A7,$51,$1D,$5B,$1D,$5A
	.dc.b	$75,$29,$D4,$A7,$52,$9D,$63,$A1,$23,$A3,$63,$A3,$5E,$8D,$7A,$35
	.dc.b	$E9,$1C,$88,$F2,$7F,$D0,$3E,$C7,$FB,$56,$3F,$B1,$35,$FE,$CE,$D7
	.dc.b	$FD,$B7,$C5,$FE,$D8,$7F,$17,$F2,$DF,$E7,$DC,$BF,$1E,$D8,$C9,$FE
	.dc.b	$EC,$3E,$AB,$97,$F3,$6D,$72,$6C,$F9,$EE,$5A,$67,$04,$65,$B1,$98
	.dc.b	$C1,$1C,$12,$7C,$ED,$3C,$EE,$08,$E0,$87,$A0,$E0,$57,$6D,$36,$7B
	.dc.b	$1C,$B4,$C9,$EC,$72,$D3,$33,$EC,$8C,$B6,$33,$3E,$D9,$CB,$63,$27
	.dc.b	$98,$F0,$1E,$52,$18,$21,$E1,$32,$53,$33,$E1,$72,$D7,$31,$E1,$72
	.dc.b	$D7,$27,$8C,$E8,$48,$E4,$B2,$6C,$E4,$7C,$27,$F8,$D2,$73,$18,$AF
	.dc.b	$E5,$15,$BA,$D3,$63,$F9,$96,$3F,$ED,$39,$FE,$6C,$D8,$C9,$CC,$FE
	.dc.b	$6C,$53,$FD,$B5,$CA,$F2,$B8,$47,$F9,$32,$30,$EC,$FE,$C1,$1A,$78
	.dc.b	$B7,$99,$ED,$3E,$1B,$FD,$CD,$EE,$FF,$94,$E7,$31,$8B,$FF,$2C,$C6
	.dc.b	$14,$EE,$B9,$46,$0F,$5E,$3E,$E7,$F5,$9C,$9F,$E1,$88,$C1,$C7,$E2
	.dc.b	$3F,$65,$87,$91,$B6,$69,$8E,$4F,$26,$1F,$EF,$DA,$F5,$C6,$23,$FC
	.dc.b	$C3,$5F,$D5,$71,$F3,$4E,$67,$FC,$D8,$79,$30,$FD,$F6,$3E,$58,$F4
	.dc.b	$9F,$02,$FF,$C5,$D3,$EF,$EE,$F2,$92,$7F,$87,$7C,$73,$E3,$1F,$00
	.dc.b	$9F,$F3,$41,$67,$51,$9F,$1F,$FD,$05,$18,$9D,$87,$11,$F2,$23,$CA
	.dc.b	$7E,$1B,$0C,$3D,$6F,$C2,$BE,$23,$AC,$61,$B8,$79,$5B,$3B,$3E,$07
	.dc.b	$C8,$FA,$7C,$5F,$BC,$D8,$C9,$FE,$5A,$2C,$78,$9F,$F4,$F1,$F3,$49
	.dc.b	$87,$27,$81,$F6,$74,$57,$58,$C3,$5F,$B6,$DF,$BA,$7C,$E7,$FB,$2A
	.dc.b	$FF,$D5,$51,$8E,$8F,$E7,$8D,$BF,$E8,$4D,$9F,$34,$7E,$09,$BC,$1F
	.dc.b	$81,$8F,$B1,$F7,$EF,$D5,$3F,$87,$37,$BB,$FA,$43,$67,$FA,$73,$6F
	.dc.b	$F4,$9F,$2D,$F8,$1F,$8E,$B4,$69,$1F,$C7,$7C,$BE,$DA,$6C,$B1,$FA
	.dc.b	$F8,$D4,$7C,$9F,$0B,$F1,$DF,$84,$FE,$CD,$B1,$F9,$96,$BB,$11,$FB
	.dc.b	$9F,$9A,$CC,$7E,$A1,$BD,$9F,$C1,$9B,$C6,$6F,$DD,$FE,$77,$F0,$9F
	.dc.b	$E0,$4D,$BF,$D4,$35,$FB,$CD,$7F,$DD,$1E,$95,$7F,$DD,$0D,$F5,$FF
	.dc.b	$D7,$D0,$AF,$F7,$1C,$61,$FE,$21,$F3,$1C,$F2,$FE,$36,$DC,$8D,$3C
	.dc.b	$1C,$BF,$88,$F8,$0F,$C2,$F4,$1A,$4C,$BE,$DF,$37,$E0,$09,$CD,$F8
	.dc.b	$07,$FA,$86,$5F,$CA,$BB,$3F,$7D,$F8,$F7,$F2,$36,$DF,$F0,$96,$BF
	.dc.b	$9F,$6B,$FC,$5B,$E6,$6D,$7F,$34,$6E,$A6,$9C,$34,$FC,$D5,$7F,$0D
	.dc.b	$0E,$1F,$E2,$0D,$73,$A7,$E9,$DF,$EA,$FE,$57,$F4,$7F,$2F,$FA,$DF
	.dc.b	$97,$FD,$62,$31,$DD,$FD,$66,$B9,$F8,$FF,$EC,$FA,$FD,$7B,$63,$C5
	.dc.b	$8F,$F4,$4F,$1B,$FA,$83,$63,$FA,$D6,$BF,$D4,$3A,$BF,$92,$7A,$7F
	.dc.b	$97,$9F,$F5,$41,$F9,$72,$9F,$97,$7F,$2C,$E0,$FC,$B1,$67,$F2,$CD
	.dc.b	$DF,$F2,$CD,$FB,$CF,$EC,$10,$69,$23,$CB,$1F,$E3,$AC,$FF,$AE,$72
	.dc.b	$70,$24,$EB,$CC,$E6,$47,$07,$C8,$DD,$F8,$57,$CB,$24,$7D,$07,$E4
	.dc.b	$ED,$FD,$15,$B1,$C0,$F9,$7B,$FF,$BF,$63,$27,$F6,$FC,$F8,$4F,$F5
	.dc.b	$56,$BF,$CA,$DA,$FF,$62,$6E,$FF,$4F,$E4,$FB,$92,$7B,$E1,$FD,$5E
	.dc.b	$3F,$BE,$7D,$C7,$F5,$9F,$C2,$AD,$1A,$47,$88,$FC,$2B,$D6,$47,$2F
	.dc.b	$E3,$7C,$69,$1F,$92,$6B,$FD,$0F,$EF,$FF,$C0,$24,$7F,$CF,$5C,$7F
	.dc.b	$E1,$88,$FF,$02,$FD,$57,$F6,$DF,$99,$FF,$09,$36,$BF,$BC,$35,$FE
	.dc.b	$9C,$D5,$4F,$95,$7D,$1F,$F8,$49,$B3,$E0,$6E,$3F,$E2,$56,$C7,$A4
	.dc.b	$87,$EC,$FA,$FF,$C8,$B3,$EC,$8E,$FF,$80,$7E,$DE,$3F,$C5,$CD,$8F
	.dc.b	$F1,$6B,$5F,$F0,$E3,$5F,$F0,$54,$F4,$7B,$98,$EC,$FF,$2B,$57,$7C
	.dc.b	$56,$FA,$DF,$F1,$6B,$6F,$E5,$B6,$BC,$73,$AF,$A7,$FF,$0C,$70,$F9
	.dc.b	$9B,$7E,$66,$BE,$66,$BF,$33,$D4,$F9,$AF,$D8,$6D,$FF,$93,$5B,$1F
	.dc.b	$E4,$C6,$BF,$27,$2F,$5B,$EE,$3F,$CE,$8E,$1D,$74,$FA,$BE,$BB,$DA
	.dc.b	$48,$C1,$EE,$FA,$FF,$CE,$3F,$39,$AF,$CE,$6B,$F5,$9A,$F8,$27,$FD
	.dc.b	$89,$B7,$D1,$8F,$0E,$3F,$3F,$57,$DD,$FF,$B4,$F7,$FB,$6D,$AF,$B6
	.dc.b	$D7,$F0,$6D,$8F,$AF,$F3,$9F,$E4,$7F,$C2,$7E,$4D,$7E,$22,$31,$3F
	.dc.b	$86,$6C,$7E,$19,$AF,$E0,$F2,$F3,$EB,$F1,$E3,$19,$79,$FE,$CF,$FC
	.dc.b	$A9,$E8,$76,$C7,$F0,$F1,$87,$F2,$6D,$AF,$C9,$B5,$FD,$4F,$C2,$FF
	.dc.b	$91,$FC,$AE,$4D,$AF,$F4,$3B,$67,$FC,$EC,$D7,$F4,$7F,$69,$FE,$65
	.dc.b	$FB,$9F,$17,$02,$31,$7F,$A4,$6D,$7E,$91,$AF,$B0,$FC,$37,$0F,$D5
	.dc.b	$70,$78,$3D,$DF,$F2,$D5,$79,$1B,$19,$1E,$EF,$AA,$33,$1C,$88,$7E
	.dc.b	$0F,$6F,$DE,$34,$F0,$EF,$F3,$DC,$1F,$DD,$7E,$15,$E6,$BE,$1B,$FD
	.dc.b	$09,$F1,$0C,$35,$39,$B4,$E5,$FF,$45,$5F,$2B,$E3,$7C,$1B,$3A,$BE
	.dc.b	$57,$C6,$17,$C5,$EB,$7C,$5A,$71,$BE,$6F,$E8,$4F,$F0,$21,$0F,$F0
	.dc.b	$49,$4F,$D9,$94,$FC,$91,$FE,$CB,$E0,$FF,$46,$78,$DF,$36,$0D,$26
	.dc.b	$BD,$22,$1D,$0A,$7E,$20,$7D,$CF,$E2,$38,$3F,$D5,$CD,$8C,$CF,$E2
	.dc.b	$5A,$7D,$FC,$FC,$44,$8F,$C2,$47,$D0,$7F,$42,$FF,$46,$69,$F9,$6E
	.dc.b	$4F,$F6,$83,$83,$CB,$1F,$1C,$BC,$2D,$DC,$DF,$E3,$79,$FD,$C0,$FC
	.dc.b	$27,$F4,$3D,$1F,$EC,$76,$C6,$70,$FE,$00,$D3,$F8,$59,$FE,$E1,$29
	.dc.b	$FA,$AE,$AF,$F7,$BB,$88,$F1,$E1,$E2,$8F,$AC,$79,$CF,$E4,$43,$E8
	.dc.b	$7F,$AA,$F6,$7F,$B6,$5B,$19,$3F,$95,$14,$FE,$AF,$DB,$ED,$69,$6F
	.dc.b	$FA,$9C,$66,$3F,$B1,$B5,$F1,$1F,$88,$3C,$F9,$F1,$EE,$FE,$47,$1F
	.dc.b	$D1,$EB,$91,$FE,$EF,$2C,$6A,$3F,$AA,$0F,$9A,$FE,$B4,$D3,$FA,$D1
	.dc.b	$4D,$BF,$D8,$9F,$47,$5F,$F6,$77,$04,$3C,$A7,$CA,$DF,$EF,$61,$C2
	.dc.b	$3C,$B3,$ED,$C3,$FB,$9C,$60,$F8,$15,$C1,$E5,$A7,$D5,$7F,$3F,$3F
	.dc.b	$E1,$5F,$13,$DF,$E2,$7F,$85,$27,$FC,$25,$18,$3E,$1F,$A4,$F1,$92
	.dc.b	$F9,$EF,$F1,$2B,$4F,$F1,$29,$4F,$F1,$29,$4F,$F1,$2C,$BC,$9E,$A4
	.dc.b	$8F,$B9,$19,$7E,$E6,$7E,$B7,$E2,$9F,$B8,$F3,$7F,$D0,$5F,$0F,$FE
	.dc.b	$5E,$24,$7F,$92,$1F,$F2,$6F,$C2,$48,$FF,$4E,$6A,$C1,$E5,$7F,$BD
	.dc.b	$27,$31,$85,$5F,$E6,$AE,$59,$1F,$E5,$46,$BF,$E6,$53,$FD,$4E,$7F
	.dc.b	$9A,$4A,$67,$1F,$F3,$73,$83,$FD,$5B,$E9,$7F,$18,$71,$1F,$E9,$D3
	.dc.b	$FC,$F2,$47,$FD,$5A,$FF,$AA,$C8,$E0,$F4,$FA,$7F,$5B,$C2,$FE,$09
	.dc.b	$9F,$BE,$31,$3A,$3F,$A2,$1F,$72,$34,$E0,$FA,$FE,$A7,$07,$59,$A3
	.dc.b	$11,$A3,$84,$FC,$D8,$CB,$F3,$FD,$3B,$5F,$EA,$7A,$DB,$70,$F8,$A7
	.dc.b	$FA,$D3,$E2,$24,$FD,$4F,$81,$FC,$8C,$62,$3C,$7E,$4E,$1F,$F5,$F4
	.dc.b	$7F,$B0,$48,$FF,$A2,$AB,$FD,$46,$47,$07,$FB,$61,$FF,$6B,$11,$C1
	.dc.b	$FE,$D2,$F6,$7F,$98,$3F,$C4,$F5,$46,$23,$F9,$94,$63,$A7,$F8,$A3
	.dc.b	$63,$F3,$BE,$D7,$EA,$BF,$1B,$FA,$BA,$7F,$BC,$E5,$18,$9D,$47,$E5
	.dc.b	$1A,$FF,$BF,$CD,$9F,$98,$F4,$32,$FE,$B9,$BB,$FE,$7D,$CF,$23,$DD
	.dc.b	$FB,$55,$FC,$B6,$31,$C5,$FC,$AD,$B1,$99,$DB,$EE,$D5,$EF,$7B,$9B
	.dc.b	$1F,$15,$AE,$E3,$F8,$5F,$BF,$FC,$2B,$E1,$FE,$0B,$67,$F6,$2D,$7F
	.dc.b	$1E,$D7,$F8,$8F,$C1,$3E,$0F,$10,$8C,$2A,$FE,$91,$18,$6C,$F3,$D7
	.dc.b	$F8,$17,$AF,$78,$4D,$26,$1C,$CF,$C9,$0F,$C2,$3E,$CF,$47,$F2,$46
	.dc.b	$C6,$4D,$BF,$7B,$A3,$51,$FC,$39,$AF,$E6,$8E,$05,$8F,$C2,$FC,$37
	.dc.b	$F3,$0F,$95,$B2,$69,$1F,$E3,$3F,$13,$B6,$9B,$2C,$7E,$42,$35,$1F
	.dc.b	$CF,$38,$FE,$FB,$FE,$0B,$6B,$FD,$00,$FE,$47,$C7,$F8,$27,$11,$A7
	.dc.b	$5B,$FF,$91,$5F,$F0,$A3,$5C,$FA,$FF,$DD,$63,$10,$E9,$FD,$33,$F8
	.dc.b	$DE,$AF,$C6,$11,$FE,$8B,$F1,$5F,$D4,$BD,$7F,$D4,$B6,$3F,$9C,$B5
	.dc.b	$FE,$2E,$D7,$F2,$18,$78,$48,$FF,$82,$70,$F1,$F1,$FF,$45,$AE,$0D
	.dc.b	$9F,$E6,$C7,$42,$9F,$D7,$23,$B7,$88,$F3,$9D,$9F,$DA,$3E,$4F,$FC
	.dc.b	$6D,$39,$8C,$5F,$CD,$8C,$29,$DD,$72,$8C,$1F,$33,$2F,$4B,$FE,$25
	.dc.b	$CB,$E5,$BF,$A1,$8F,$F1,$25,$F4,$E4,$E7,$8F,$F7,$A6,$D7,$FA,$28
	.dc.b	$F1,$61,$DE,$3F,$8A,$E9,$FE,$02,$24,$FF,$76,$7C,$2F,$8C,$F9,$58
	.dc.b	$61,$FE,$8F,$E0,$FF,$82,$3E,$7F,$F8,$FB,$63,$FB,$7B,$5D,$88,$FE
	.dc.b	$4B,$F1,$DF,$E1,$22,$47,$F8,$19,$EC,$FD,$C3,$C2,$43,$F9,$6F,$C7
	.dc.b	$F5,$7F,$C2,$2D,$E0,$FF,$01,$37,$AC,$EE,$5F,$20,$F1,$FF,$AF,$C6
	.dc.b	$1B,$19,$8F,$ED,$06,$4B,$C3,$FE,$AF,$F1,$BE,$B6,$BC,$A3,$13,$F2
	.dc.b	$67,$FA,$83,$FA,$FF,$2F,$FC,$83,$18,$8F,$F7,$93,$FE,$63,$F9,$1F
	.dc.b	$D9,$47,$ED,$5A,$FF,$8D,$9A,$9A,$4F,$1B,$FC,$CC,$D8,$FF,$35,$35
	.dc.b	$ED,$1A,$23,$A6,$BA,$30,$43,$04,$3F,$D2,$D1,$D5,$B1,$C9,$A7,$12
	.dc.b	$9C,$4F,$73,$B6,$9B,$29,$B3,$E1,$3B,$69,$B2,$9B,$3E,$33,$B6,$9B
	.dc.b	$29,$B8,$F5,$3E,$03,$CE,$43,$04,$3B,$99,$29,$93,$A9,$92,$99,$7B
	.dc.b	$BC,$08,$68,$AE,$9C,$11,$FC,$1F,$BF,$FE,$F7,$8F,$F6,$43,$5F,$F7
	.dc.b	$C1,$8E,$3F,$CD,$BF,$89,$F3,$BF,$D0,$39,$FE,$45,$B1,$F7,$1A,$FE
	.dc.b	$4D,$AF,$E4,$DA,$F3,$3F,$A2,$5F,$F9,$FE,$7A,$46,$27,$73,$98,$EC
	.dc.b	$64,$93,$C0,$8E,$9A,$F8,$CD,$99,$21,$97,$6F,$9C,$D9,$93,$66,$5D
	.dc.b	$BE,$B6,$C7,$9E,$38,$10,$E0,$72,$38,$15,$E0,$E8,$E0,$68,$AE,$9C
	.dc.b	$1B,$21,$82,$3B,$24,$FC,$C2,$47,$B8,$E3,$1A,$70,$72,$8D,$CF,$B9
	.dc.b	$B1,$E9,$6B,$FC,$67,$A9,$EA,$3E,$2F,$FD,$F5,$AF,$FB,$71,$B3,$A6
	.dc.b	$CE,$EB,$FC,$20,$3E,$01,$F4,$1F,$ED,$2A,$CC,$E6,$30,$F0,$9D,$BC
	.dc.b	$CF,$F0,$D3,$89,$E3,$F5,$AF,$A3,$E3,$5F,$E1,$F8,$C3,$C4,$FF,$70
	.dc.b	$B8,$9E,$33,$C2,$7F,$A9,$7C,$77,$F3,$E4,$D2,$3F,$ED,$2F,$69,$F6
	.dc.b	$7F,$EE,$CF,$57,$FD,$6F,$18,$94,$62,$75,$1C,$5A,$FF,$8D,$7E,$BD
	.dc.b	$E5,$E2,$3F,$DE,$F1,$87,$99,$F3,$1C,$57,$CA,$6C,$7F,$B9,$BE,$C3
	.dc.b	$FD,$A9,$C7,$FE,$CA,$6C,$67,$C2,$FD,$EF,$7A,$73,$3F,$EC,$BC,$DF
	.dc.b	$77,$E3,$A4,$62,$7C,$66,$E7,$51,$9C,$BE,$EE,$0F,$86,$F6,$5A,$30
	.dc.b	$F2,$3F,$C8,$91,$A8,$CC,$FC,$CF,$7B,$CF,$66,$92,$FF,$CB,$1E,$43
	.dc.b	$F2,$E7,$D7,$25,$F5,$BF,$E3,$A9,$CC,$60,$FF,$1D,$38,$8D,$4F,$17
	.dc.b	$F2,$FF,$66,$F6,$7D,$6D,$BF,$50,$71,$9F,$09,$FB,$98,$F9,$70,$F2
	.dc.b	$79,$BD,$B0,$FF,$23,$1E,$79,$D3,$80,$E2,$3F,$53,$F4,$0F,$AB,$FE
	.dc.b	$6A,$DA,$FF,$3C,$37,$93,$73,$FC,$EB,$E8,$1F,$55,$F5,$1B,$7E,$A6
	.dc.b	$E9,$FB,$66,$C7,$EF,$30,$E8,$DB,$F0,$91,$FF,$45,$72,$FE,$23,$E4
	.dc.b	$7F,$0A,$48,$FC,$04,$65,$F9,$3E,$53,$3F,$C3,$B6,$33,$18,$3F,$1F
	.dc.b	$F6,$92,$39,$F2,$A3,$1A,$7E,$D3,$E8,$BF,$05,$F9,$D5,$B3,$48,$FE
	.dc.b	$F7,$E7,$B6,$D3,$65,$8F,$DF,$C6,$A3,$F5,$1F,$2D,$F8,$FC,$7F,$81
	.dc.b	$B6,$3F,$84,$35,$D9,$FA,$C3,$F8,$67,$D2,$E5,$6B,$FB,$08,$C3,$63
	.dc.b	$81,$FB,$78,$CD,$7F,$BA,$BE,$C3,$F8,$9F,$C2,$7E,$85,$B4,$69,$FC
	.dc.b	$7C,$8F,$FA,$0D,$AF,$F0,$3F,$A7,$E4,$FF,$21,$6B,$93,$A1,$92,$9F
	.dc.b	$C2,$FE,$9B,$F8,$29,$1F,$AE,$FF,$08,$FA,$99,$1F,$69,$BB,$FF,$4A
	.dc.b	$CD,$F6,$5F,$12,$9F,$8D,$66,$30,$D9,$FE,$00,$D9,$D9,$FE,$E3,$8E
	.dc.b	$7F,$59,$A8,$C5,$23,$11,$FC,$26,$76,$F2,$31,$3B,$70,$7D,$8D,$5C
	.dc.b	$DF,$C4,$7F,$07,$FE,$8A,$D9,$FE,$94,$D7,$F9,$83,$5F,$E3,$1E,$9B
	.dc.b	$D3,$FE,$E7,$A7,$F6,$E2,$4F,$F6,$E9,$D7,$A7,$FC,$AF,$E1,$1B,$BE
	.dc.b	$47,$FA,$64,$61,$B1,$9B,$E0,$DC,$3E,$67,$95,$FC,$A7,$F5,$4B,$46
	.dc.b	$91,$F9,$3E,$EE,$DA,$6C,$B1,$F5,$E3,$51,$F8,$51,$B7,$F6,$AE,$4F
	.dc.b	$85,$C5,$FD,$41,$BC,$26,$E7,$2F,$89,$FD,$D5,$C5,$FE,$17,$83,$FA
	.dc.b	$17,$E1,$16,$8D,$23,$F2,$E3,$FC,$43,$3F,$E1,$C2,$46,$78,$3B,$BF
	.dc.b	$D3,$78,$3E,$63,$80,$E2,$7F,$C4,$6D,$8F,$EA,$86,$B0,$FA,$53,$F0
	.dc.b	$63,$EC,$9F,$E0,$FC,$7F,$AD,$4E,$49,$1C,$AB,$C3,$8F,$69,$FB,$8D
	.dc.b	$E0,$F8,$35,$B9,$CB,$FA,$0F,$6B,$FB,$37,$E1,$1F,$AB,$6C,$64,$46
	.dc.b	$5A,$73,$AF,$ED,$DF,$84,$77,$83,$48,$FC,$9F,$0B,$FC,$8E,$47,$07
	.dc.b	$EC,$BC,$7F,$F0,$87,$47,$F5,$08,$FD,$7C,$60,$CF,$CD,$FF,$7D,$FC
	.dc.b	$27,$F8,$13,$67,$F7,$6D,$7F,$6E,$D7,$F9,$6D,$3E,$C7,$FC,$46,$D8
	.dc.b	$FE,$B2,$DC,$3F,$87,$4F,$FA,$8B,$AB,$F5,$F1,$E9,$AF,$F3,$47,$1F
	.dc.b	$F5,$76,$C7,$F1,$5F,$95,$F5,$92,$3F,$B0,$E0,$8F,$F1,$2D,$F4,$9F
	.dc.b	$86,$72,$29,$E8,$AF,$8A,$47,$9B,$F2,$3D,$FF,$F1,$07,$E1,$BF,$C9
	.dc.b	$5B,$3F,$C7,$9A,$FF,$1A,$6B,$FD,$43,$E7,$9F,$47,$FE,$46,$6C,$7F
	.dc.b	$7C,$6D,$FF,$35,$6C,$79,$86,$1E,$B7,$D4,$43,$E8,$F9,$BF,$78,$8F
	.dc.b	$C4,$7F,$03,$EF,$FF,$8E,$7F,$09,$FE,$7C,$D9,$FE,$6E,$D7,$F9,$9B
	.dc.b	$5F,$E5,$3D,$9F,$E9,$AC,$7E,$B4,$63,$7F,$AC,$D8,$FE,$95,$F6,$DF
	.dc.b	$90,$24,$7D,$67,$F2,$7F,$73,$FE,$4B,$FC,$27,$FA,$94,$FE,$2A,$31
	.dc.b	$E0,$FF,$4A,$6C,$7F,$46,$6B,$DC,$F8,$DA,$FC,$57,$07,$F4,$CF,$BB
	.dc.b	$FF,$2C,$7B,$1D,$AF,$FB,$0C,$61,$FE,$AE,$D9,$E0,$7F,$98,$23,$FA
	.dc.b	$DF,$99,$FE,$61,$FA,$CE,$5C,$FF,$DD,$9B,$19,$9D,$D2,$77,$EE,$7F
	.dc.b	$9A,$76,$B3,$FE,$35,$AF,$EC,$F3,$96,$C6,$4F,$BC,$7E,$0B,$83,$E9
	.dc.b	$3F,$01,$AF,$F6,$AF,$BC,$FF,$55,$4F,$F7,$26,$C6,$4F,$F5,$5C,$65
	.dc.b	$A7,$6D,$7E,$53,$DF,$87,$EE,$B7,$1F,$C0,$B4,$F1,$46,$BE,$31,$F1
	.dc.b	$8F,$8E,$FF,$44,$79,$BF,$BB,$9C,$C6,$30,$FE,$73,$18,$C3,$FD,$3D
	.dc.b	$18,$BE,$51,$83,$FD,$25,$3B,$BE,$93,$A6,$DF,$5A,$DC,$E6,$FF,$86
	.dc.b	$C7,$FB,$B8,$87,$F5,$32,$9A,$30,$43,$E5,$7C,$9F,$E0,$B4,$FF,$52
	.dc.b	$36,$0D,$3F,$3A,$FC,$B3,$A9,$FA,$92,$E9,$FE,$34,$DB,$FC,$59,$1F
	.dc.b	$E2,$46,$BF,$CB,$9B,$1B,$8F,$F1,$61,$FE,$2A,$E0,$FF,$5B,$7D,$67
	.dc.b	$E5,$20,$D2,$4F,$E8,$23,$FD,$AA,$DF,$86,$FC,$EF,$27,$FA,$FD,$B3
	.dc.b	$A3,$E2,$10,$F5,$1F,$14,$FD,$D0,$FC,$1F,$F4,$FC,$FF,$EC,$96,$C6
	.dc.b	$6F,$F7,$CD,$3F,$C8,$7D,$3F,$31,$FF,$25,$B5,$C9,$F3,$8C,$94,$F2
	.dc.b	$1F,$3C,$F2,$14,$FE,$3A,$3F,$07,$FE,$46,$D3,$F9,$19,$4D,$BF,$E6
	.dc.b	$5E,$DF,$14,$7F,$30,$3C,$05,$DB,$4E,$0F,$01,$F1,$8F,$2F,$37,$F9
	.dc.b	$39,$FE,$81,$DF,$FE,$1A,$7F,$CD,$9D,$FF,$D1,$9C,$99,$3F,$A6,$45
	.dc.b	$7F,$B8,$F8,$1F,$EF,$5F,$D9,$BF,$AB,$41,$A4,$7F,$CE,$BC,$BE,$AD
	.dc.b	$7F,$D0,$6D,$ED,$FF,$63,$67,$FE,$83,$8F,$7D,$7A,$89,$9F,$A5,$FC
	.dc.b	$05,$FF,$A3,$3C,$3E,$44,$8F,$8B,$19,$7E,$24,$F1,$7E,$1F,$6B,$B7
	.dc.b	$D9,$39,$9F,$F0,$6F,$89,$EF,$BF,$F4,$CB,$73,$FF,$07,$4F,$F8,$30
	.dc.b	$F8,$D3,$FE,$9E,$F1,$50,$FC,$0F,$FC,$3A,$D3,$FC,$3A,$53,$FC,$3A
	.dc.b	$53,$FC,$3B,$E4,$7C,$52,$3E,$CB,$FF,$7F,$ED,$FB,$C8,$F8,$BA,$7A
	.dc.b	$DF,$DF,$F9,$FF,$E5,$CF,$C2,$3F,$C7,$84,$8F,$F2,$00,$7F,$C8,$9F
	.dc.b	$84,$48,$FF,$24,$35,$FF,$33,$7A,$1F,$E8,$FF,$A7,$FF,$23,$12,$39
	.dc.b	$BF,$E6,$1F,$A5,$91,$FE,$61,$0E,$1E,$7E,$87,$C6,$FC,$1B,$FC,$A0
	.dc.b	$48,$F2,$BF,$91,$E3,$FF,$52,$B7,$4F,$B0,$DE,$43,$4F,$F4,$36,$32
	.dc.b	$FF,$D8,$B5,$F5,$23,$11,$FE,$C5,$F5,$FF,$CE,$F5,$FE,$D7,$6C,$7F
	.dc.b	$A4,$CE,$A5,$8F,$F3,$F1,$1F,$73,$5F,$F6,$6F,$B1,$F8,$2F,$C5,$BF
	.dc.b	$D5,$10,$69,$1F,$F6,$CF,$97,$B6,$9B,$2C,$7F,$AE,$8F,$F3,$B7,$14
	.dc.b	$8F,$B6,$D7,$FC,$B6,$FC,$52,$3F,$EA,$17,$FD,$B6,$FB,$0F,$F6,$C9
	.dc.b	$4F,$F7,$17,$B9,$F5,$5F,$F6,$F9,$9E,$8F,$F7,$5F,$E2,$FF,$EF,$9F
	.dc.b	$7B,$F3,$ED,$93,$4F,$F7,$D1,$1F,$F7,$BC,$F0,$69,$C0,$FA,$17,$F4
	.dc.b	$F8,$FF,$84,$6B,$FE,$AC,$8C,$4F,$EF,$A3,$FD,$98,$DF,$6F,$64,$75
	.dc.b	$F0,$1F,$B4,$FB,$5F,$BC,$D8,$FF,$5D,$E2,$B4,$E6,$7F,$0C,$D7,$F7
	.dc.b	$BF,$12,$F0,$71,$08,$C2,$C3,$F9,$54,$61,$B3,$CF,$0F,$07,$C1,$DC
	.dc.b	$4D,$26,$1C,$CF,$3C,$3E,$13,$F8,$7F,$E1,$5F,$99,$6C,$E9,$AF,$E7
	.dc.b	$DA,$AF,$E0,$7C,$36,$1B,$0A,$FD,$24,$E6,$CD,$F8,$7F,$E3,$FF,$9D
	.dc.b	$FE,$A5,$B7,$E6,$6B,$FA,$F6,$BE,$1A,$7C,$56,$31,$FA,$4D,$8C,$CA
	.dc.b	$FF,$6D,$39,$F8,$BF,$C2,$92,$7F,$08,$FF,$4A,$DF,$FA,$6C,$7F,$5F
	.dc.b	$6B,$FE,$FA,$6B,$F7,$36,$FC,$7F,$34,$8F,$C7,$37,$8F,$FB,$27,$C6
	.dc.b	$3C,$FF,$EC,$94,$8C,$3F,$D9,$DA,$F0,$8D,$FC,$6F,$F3,$A8,$FE,$E4
	.dc.b	$D7,$FA,$48,$70,$FF,$4C,$6B,$DE,$3C,$33,$DA,$3F,$BE,$35,$E6,$D5
	.dc.b	$F1,$EF,$92,$7C,$0F,$F9,$A2,$7C,$33,$98,$C4,$7F,$5C,$8C,$2E,$EF
	.dc.b	$F0,$54,$61,$FF,$04,$7C,$93,$E4,$BF,$BD,$B6,$BF,$C2,$CD,$D3,$73
	.dc.b	$FE,$1E,$BE,$BF,$71,$D3,$2F,$2E,$08,$C6,$5F,$DB,$9B,$19,$3F,$4B
	.dc.b	$F2,$9F,$DE,$79,$4D,$24,$7F,$86,$FE,$57,$D4,$FF,$8A,$5B,$9F,$F8
	.dc.b	$85,$B1,$FE,$27,$6B,$FD,$27,$DD,$FE,$2A,$47,$F8,$A3,$FE,$1F,$F9
	.dc.b	$64,$FF,$82,$DB,$3F,$E0,$86,$BF,$E3,$3E,$94,$62,$3F,$C6,$DF,$2E
	.dc.b	$F9,$8F,$F1,$2B,$63,$FC,$80,$DC,$7F,$C7,$0D,$8F,$E0,$FF,$33,$FC
	.dc.b	$A8,$8F,$F2,$97,$F4,$DD,$1F,$E1,$8F,$4F,$F8,$AC,$9A,$47,$FC,$91
	.dc.b	$F3,$8F,$8D,$FF,$0F,$FC,$57,$F9,$1E,$31,$3F,$EA,$A3,$F3,$6E,$23
	.dc.b	$53,$C2,$7F,$CA,$5F,$3A,$F7,$F1,$1F,$A2,$8C,$3C,$CF,$F1,$6B,$8A
	.dc.b	$FF,$2B,$36,$3F,$CB,$9E,$A3,$D3,$FF,$1F,$57,$F8,$1C,$93,$A8,$CB
	.dc.b	$FE,$63,$F9,$47,$CF,$7E,$F1,$B1,$FE,$61,$6B,$FE,$65,$6B,$C6,$38
	.dc.b	$11,$E0,$D7,$83,$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$E1,$31,$E4
	.dc.b	$6C,$77,$69,$D0,$A7,$32,$9C,$CF,$F2,$DC,$73,$6C,$73,$69,$CC,$A7
	.dc.b	$32,$9C,$C4,$73,$6C,$73,$69,$CC,$A7,$32,$9C,$E3,$91,$23,$93,$63
	.dc.b	$93,$5E,$4D,$79,$47,$02,$3C,$1F,$E7,$BF,$1C,$F8,$EF,$C9,$4F,$FB
	.dc.b	$FB,$0F,$F5,$24,$E5,$B3,$A9,$7E,$5C,$E8,$FF,$42,$F0,$FF,$AB,$63
	.dc.b	$13,$FE,$08,$3F,$A0,$B8,$8D,$4F,$09,$FF,$5C,$4E,$5B,$19,$78,$B5
	.dc.b	$FF,$49,$C6,$DC,$9D,$CC,$C6,$DF,$3B,$63,$84,$68,$E0,$68,$AE,$9C
	.dc.b	$4F,$A1,$CC,$F3,$23,$82,$46,$C8,$FB,$0C,$95,$CB,$EA,$6B,$E5,$6B
	.dc.b	$E2,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$27,$E2
	.dc.b	$12,$7E,$0C,$F9,$A3,$13,$EE,$3A,$B8,$8D,$4F,$09,$F3,$69,$FC,$77
	.dc.b	$D9,$7C,$FF,$FA,$8C,$93,$FC,$76,$32,$FF,$88,$6F,$FD,$AB,$3F,$CF
	.dc.b	$89,$19,$36,$7F,$38,$E1,$DB,$FD,$01,$AE,$4D,$9F,$EE,$0E,$2F,$E8
	.dc.b	$33,$FE,$B5,$F9,$F9,$3E,$36,$BF,$E1,$EF,$65,$F4,$3F,$EE,$89,$FF
	.dc.b	$10,$CF,$FB,$AA,$72,$A3,$2A,$33,$18,$6C,$7F,$9F,$3E,$07,$FC,$13
	.dc.b	$F3,$BF,$DB,$5B,$19,$F0,$FF,$DE,$F3,$96,$C7,$C9,$D3,$FD,$9C,$47
	.dc.b	$07,$F8,$E2,$FF,$51,$E5,$F8,$E3,$F4,$AD,$9F,$3B,$63,$FD,$1F,$E8
	.dc.b	$FF,$BD,$F0,$FF,$00,$36,$33,$BF,$89,$1F,$E7,$78,$D3,$FE,$46,$E0
	.dc.b	$FC,$16,$1E,$58,$C7,$42,$FF,$6F,$5A,$9F,$F2,$7C,$FA,$08,$FA,$1F
	.dc.b	$E0,$5E,$5B,$E1,$FC,$3F,$25,$FE,$60,$6D,$7F,$93,$9A,$FF,$93,$9A
	.dc.b	$FA,$2F,$C0,$F9,$30,$F1,$E1,$E5,$C7,$D1,$3E,$EC,$7C,$0B,$E8,$3C
	.dc.b	$0D,$3C,$05,$3C,$05,$3E,$3D,$3E,$93,$F8,$E3,$63,$F8,$CB,$6F,$F1
	.dc.b	$6D,$8F,$BA,$79,$E3,$CA,$B1,$F5,$BD,$AB,$A9,$1E,$AF,$F3,$1F,$19
	.dc.b	$AB,$E9,$FE,$E5,$7D,$B6,$CE,$9B,$3B,$96,$FE,$2C,$3C,$31,$E3,$CB
	.dc.b	$E5,$3E,$83,$F0,$FF,$84,$3E,$D3,$F9,$E3,$63,$F9,$CB,$74,$FD,$5B
	.dc.b	$63,$F2,$BF,$4D,$F9,$0F,$B6,$FC,$AB,$63,$22,$32,$D3,$F3,$BB,$6D
	.dc.b	$FC,$CE,$DF,$0A,$31,$1F,$9B,$6C,$66,$39,$1C,$35,$7D,$4F,$E8,$BA
	.dc.b	$3F,$4E,$D9,$FD,$5B,$5E,$0D,$7F,$51,$E6,$DB,$FA,$6D,$BF,$F3,$9D
	.dc.b	$FE,$CA,$31,$F5,$7F,$B0,$FC,$9B,$F6,$AD,$C3,$F7,$2D,$7C,$CD,$7F
	.dc.b	$6F,$F6,$7F,$EF,$A5,$3F,$C6,$23,$0D,$F8,$2E,$12,$9C,$BF,$C1,$BE
	.dc.b	$C3,$A6,$7F,$8A,$6E,$66,$98,$F4,$7F,$12,$4A,$FC,$49,$FC,$41,$FC
	.dc.b	$8A,$E1,$FC,$03,$FE,$D0,$6B,$F8,$AE,$3F,$C4,$9F,$9B,$BF,$F1,$14
	.dc.b	$3F,$1A,$FE,$33,$E4,$7F,$16,$6C,$64,$FE,$3F,$19,$7F,$90,$FC,$7D
	.dc.b	$7E,$F3,$5F,$E3,$CD,$CF,$F5,$2D,$B7,$DA,$7E,$98,$95,$EE,$3F,$10
	.dc.b	$FB,$BA,$7F,$4E,$4E,$9F,$D3,$CF,$EF,$4F,$E9,$FA,$7C,$07,$E2,$69
	.dc.b	$F0,$1F,$D3,$BF,$07,$97,$F9,$9F,$E9,$9F,$92,$93,$48,$FF,$3E,$E5
	.dc.b	$DB,$4D,$96,$3F,$20,$E0,$FD,$07,$B9,$FC,$08,$8F,$B4,$FE,$84,$FD
	.dc.b	$AC,$3D,$04,$C3,$D1,$D7,$FD,$33,$BB,$F8,$59,$2F,$97,$07,$F1,$48
	.dc.b	$F8,$DA,$7C,$49,$FE,$70,$D7,$F8,$77,$81,$FC,$E0,$F8,$6C,$BF,$E1
	.dc.b	$38,$78,$23,$FA,$EF,$DB,$7F,$1A,$23,$E0,$E1,$47,$A6,$78,$4F,$E4
	.dc.b	$5A,$FF,$5F,$F9,$2F,$E4,$64,$79,$3C,$5F,$11,$F6,$A3,$8B,$E8,$BF
	.dc.b	$E5,$18,$FA,$F8,$7F,$9C,$E8,$FC,$67,$FD,$7B,$5F,$EF,$16,$DF,$F7
	.dc.b	$01,$3B,$F4,$BF,$9E,$4E,$A7,$F8,$B7,$D8,$FF,$76,$FC,$27,$FB,$DB
	.dc.b	$63,$FB,$FB,$5D,$88,$FF,$00,$F9,$BF,$AD,$24,$7F,$9D,$70,$FE,$93
	.dc.b	$1F,$E7,$5C,$BF,$35,$C1,$F4,$F3,$FE,$D3,$5F,$D3,$B5,$FD,$B1,$30
	.dc.b	$E5,$C1,$F9,$DC,$38,$DF,$C4,$C3,$3A,$FE,$F4,$97,$83,$8F,$87,$FE
	.dc.b	$12,$FC,$13,$FC,$2E,$D8,$FF,$0D,$B5,$FE,$70,$D7,$E4,$FD,$AF,$CA
	.dc.b	$7D,$8D,$C3,$FC,$7F,$A2,$31,$3F,$46,$30,$7C,$1E,$1F,$B4,$4C,$BE
	.dc.b	$14,$EF,$F2,$CF,$B1,$87,$46,$C6,$4F,$EE,$8E,$1F,$EA,$AB,$F0,$4F
	.dc.b	$EE,$6D,$8F,$D3,$B7,$C9,$E9,$3F,$A8,$BF,$0D,$7F,$0B,$21,$F8,$8F
	.dc.b	$C7,$18,$FF,$B0,$32,$F6,$46,$25,$97,$C5,$78,$56,$DF,$ED,$33,$B9
	.dc.b	$D5,$32,$7E,$0B,$FE,$5A,$8F,$84,$DB,$F6,$B5,$FC,$DC,$F0,$7F,$C1
	.dc.b	$B3,$FE,$09,$7F,$50,$5E,$0F,$D3,$CB,$F0,$1F,$24,$F2,$6D,$78,$DB
	.dc.b	$3B,$9F,$6C,$FD,$5F,$C0,$BF,$CB,$7F,$25,$EA,$A4,$62,$75,$4A,$D4
	.dc.b	$E6,$3E,$BB,$D8,$8F,$63,$25,$3F,$8D,$EB,$F5,$9B,$9F,$EA,$30,$F0
	.dc.b	$C6,$F0,$F9,$97,$D2,$BC,$33,$CC,$F8,$A7,$DE,$29,$D0,$F9,$1A,$FE
	.dc.b	$56,$3F,$C9,$3E,$EB,$C5,$7B,$BF,$E7,$CF,$8C,$FF,$80,$13,$98,$C5
	.dc.b	$7F,$11,$8C,$60,$AD,$DF,$28,$C5,$6E,$31,$3A,$BE,$75,$FE,$8A,$BE
	.dc.b	$73,$98,$E2,$7C,$63,$FD,$5B,$BF,$FB,$35,$B8,$FF,$A2,$DB,$AF,$FA
	.dc.b	$C9,$FE,$8A,$60,$87,$FB,$44,$FD,$96,$7F,$BB,$7E,$A9,$F3,$09,$A6
	.dc.b	$D8,$7F,$A1,$CE,$67,$07,$F0,$AF,$03,$07,$E0,$3C,$4F,$AE,$48,$FC
	.dc.b	$5B,$F8,$FE,$0F,$14,$FE,$64,$FF,$2A,$9F,$C0,$9C,$CF,$E6,$E3,$27
	.dc.b	$C5,$3C,$A5,$3C,$C7,$EC,$07,$C2,$FE,$77,$8B,$FD,$72,$D8,$CC,$EE
	.dc.b	$3F,$77,$C7,$FC,$0E,$6E,$7E,$28,$C5,$7F,$08,$8F,$4C,$FD,$03,$C6
	.dc.b	$53,$C6,$7F,$1C,$E4,$F0,$BF,$C9,$5A,$E4,$F3,$99,$29,$FC,$90,$FB
	.dc.b	$D3,$FC,$3D,$BC,$7F,$81,$7F,$85,$97,$6F,$C4,$3E,$46,$BE,$43,$FA
	.dc.b	$1C,$F4,$E6,$FE,$41,$1F,$CF,$DA,$FF,$A2,$1A,$FF,$B8,$79,$FF,$DC
	.dc.b	$7F,$17,$FD,$4A,$0D,$23,$FE,$EE,$EA,$F4,$F5,$9A,$48,$FC,$D6,$3E
	.dc.b	$67,$EE,$10,$F9,$87,$FB,$C7,$B3,$99,$1F,$F0,$31,$FE,$97,$7E,$16
	.dc.b	$BF,$A7,$8F,$84,$D7,$E1,$67,$FD,$F1,$E2,$7F,$BF,$7B,$7C,$E4,$73
	.dc.b	$1F,$EC,$AE,$F4,$3E,$FB,$FC,$1E,$D3,$FC,$1E,$53,$FC,$1E,$FF,$87
	.dc.b	$72,$FB,$7E,$0F,$C3,$9E,$AD,$74,$7C,$4C,$7D,$0F,$C3,$CF,$E3,$3F
	.dc.b	$10,$FF,$33,$9F,$C0,$FC,$5F,$F6,$73,$FE,$44,$31,$A7,$D2,$F0,$FF
	.dc.b	$23,$1F,$E3,$D3,$F1,$E5,$E5,$FF,$01,$EE,$9D,$CF,$63,$A1,$0F,$F2
	.dc.b	$01,$F1,$A3,$FC,$9A,$F4,$DB,$FD,$E3,$E2,$78,$1B,$19,$9D,$DF,$38
	.dc.b	$FF,$29,$1E,$82,$1F,$C4,$8F,$85,$9F,$C4,$7F,$0B,$5F,$E6,$B7,$A7
	.dc.b	$8B,$DE,$CD,$3D,$79,$FC,$69,$F4,$9E,$13,$FD,$60,$7A,$FC,$6E,$44
	.dc.b	$7D,$0F,$5E,$BF,$C1,$FE,$99,$FE,$8A,$24,$FF,$A3,$DF,$F4,$B1,$1F
	.dc.b	$F7,$38,$70,$FC,$03,$C4,$7C,$F3,$CA,$7E,$6B,$BB,$F0,$84,$7E,$BF
	.dc.b	$95,$EF,$23,$F3,$5F,$85,$E5,$FE,$63,$F1,$CF,$BA,$D9,$FF,$61,$88
	.dc.b	$CB,$4F,$F6,$49,$4F,$D0,$EA,$D7,$FD,$A7,$3F,$ED,$02,$4F,$A0,$F6
	.dc.b	$14,$F9,$A7,$E3,$BC,$EF,$AC,$47,$F0,$8F,$D9,$F3,$FF,$5F,$F9,$27
	.dc.b	$E5,$1B,$3F,$EF,$51,$19,$69,$FC,$13,$D1,$7A,$5C,$23,$FD,$19,$18
	.dc.b	$76,$62,$34,$FE,$FA,$35,$3F,$A4,$F5,$33,$9F,$C1,$6E,$66,$98,$E0
	.dc.b	$FE,$0F,$F8,$8F,$FB,$38,$98,$FF,$B4,$DF,$F6,$D1,$1F,$C2,$B5,$F9
	.dc.b	$F4,$F5,$FF,$88,$FB,$7F,$B6,$6C,$FF,$9D,$E7,$4A,$74,$D3,$5E,$CF
	.dc.b	$07,$AB,$5F,$F5,$7C,$A3,$19,$7D,$F7,$53,$F9,$53,$EE,$90,$FB,$A7
	.dc.b	$E5,$CF,$DC,$1F,$B6,$97,$B5,$FB,$52,$47,$F0,$37,$F8,$7F,$4F,$AA
	.dc.b	$3F,$9A,$B5,$FE,$7E,$D7,$FD,$8C,$D7,$F8,$2F,$2F,$E9,$5F,$EE,$B1
	.dc.b	$97,$FD,$07,$97,$F6,$98,$C1,$9F,$6F,$E3,$79,$DF,$D1,$1B,$5F,$D3
	.dc.b	$9A,$FF,$B5,$5A,$F8,$0F,$7F,$81,$FC,$DA,$3E,$87,$C7,$F3,$DB,$FA
	.dc.b	$4B,$68,$D3,$FA,$9F,$BD,$FE,$13,$8F,$EB,$AD,$7F,$DC,$8D,$7F,$A9
	.dc.b	$B5,$FE,$89,$7F,$D1,$89,$8F,$F4,$AD,$FF,$71,$A7,$F2,$52,$57,$F3
	.dc.b	$37,$F9,$FD,$BD,$FF,$F0,$73,$67,$FC,$A0,$D7,$CF,$1F,$DC,$27,$E8
	.dc.b	$47,$F8,$3C,$FE,$6B,$EF,$3E,$03,$FA,$13,$63,$FA,$33,$7C,$2F,$F6
	.dc.b	$9B,$63,$FB,$ED,$BC,$AF,$EE,$3F,$15,$FD,$E2,$31,$28,$C4,$EA,$38
	.dc.b	$B5,$FE,$F2,$35,$FF,$62,$E5,$C6,$3F,$C1,$71,$8A,$FF,$37,$92,$3F
	.dc.b	$C1,$FF,$08,$F8,$5F,$F0,$1F,$B1,$FE,$0A,$8C,$4F,$FA,$14,$EE,$E2
	.dc.b	$35,$3C,$27,$FC,$33,$F2,$9F,$ED,$27,$81,$FE,$0E,$71,$3F,$E1,$66
	.dc.b	$C7,$0F,$6F,$93,$EC,$C7,$D7,$5F,$E8,$62,$47,$B3,$89,$9F,$68,$7C
	.dc.b	$43,$E2,$3F,$C9,$8D,$8F,$F4,$73,$5F,$F1,$3B,$5F,$F1,$8B,$5F,$9A
	.dc.b	$D7,$FC,$4E,$D7,$BC,$76,$23,$D9,$AF,$66,$BD,$9A,$F6,$6B,$D9,$AF
	.dc.b	$63,$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$4E,$87,$22,$1C,$8A,$72,$29
	.dc.b	$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$12
	.dc.b	$3E,$A1,$23,$E4,$B6,$3E,$07,$B9,$FE,$E2,$7F,$CD,$ED,$7F,$DC,$18
	.dc.b	$7F,$B4,$E3,$11,$FE,$66,$FC,$F9,$CD,$FE,$68,$D3,$FC,$DF,$18,$9F
	.dc.b	$D0,$1F,$B4,$71,$1A,$9E,$13,$FE,$7E,$9C,$B6,$32,$F1,$6B,$E1,$6A
	.dc.b	$8E,$2E,$CF,$09,$B2,$9B,$8E,$31,$C2,$35,$3E,$57,$71,$97,$81,$89
	.dc.b	$E2,$47,$04,$9E,$44,$7D,$86,$4A,$E5,$F4,$35,$F1,$9A,$21,$A2,$9A
	.dc.b	$30,$43,$04,$3D,$2F,$56,$9C,$8A,$71,$29,$C6,$7E,$11,$27,$E0,$4F
	.dc.b	$38,$C4,$FB,$4E,$4E,$23,$C1,$A7,$F3,$4F,$89,$FF,$5A,$B6,$BF,$8D
	.dc.b	$68,$9C,$EB,$FC,$71,$FE,$4D,$7F,$EB,$66,$D6,$CF,$E5,$05,$3F,$A1
	.dc.b	$ED,$FC,$C6,$3F,$A4,$C6,$4F,$F5,$FF,$1F,$F4,$E8,$C9,$23,$FD,$A3
	.dc.b	$F1,$4F,$86,$F0,$46,$1A,$FF,$50,$8D,$3D,$43,$89,$F3,$FC,$54,$9F
	.dc.b	$3B,$5F,$EB,$3F,$8D,$BB,$FF,$DA,$73,$FE,$DA,$8C,$4A,$31,$3A,$8F
	.dc.b	$F7,$3C,$EA,$7F,$DD,$3C,$0F,$75,$F8,$4F,$FB,$7B,$E1,$B3,$18,$C1
	.dc.b	$18,$9D,$47,$F6,$CA,$E5,$18,$8F,$9B,$BF,$EC,$B7,$FE,$BC,$FD,$42
	.dc.b	$47,$1F,$C0,$DE,$4F,$EE,$B0,$46,$27,$F6,$F1,$89,$D4,$E6,$BC,$FF
	.dc.b	$5D,$E1,$BF,$E2,$39,$A3,$18,$7B,$DB,$19,$9F,$69,$F8,$6A,$F2,$BE
	.dc.b	$A3,$F8,$2E,$3E,$A7,$E0,$4B,$E7,$FD,$0D,$8F,$3B,$6B,$E5,$CF,$F1
	.dc.b	$6F,$80,$7C,$BB,$ED,$D8,$3E,$47,$FC,$2E,$AB,$73,$A9,$CC,$62,$78
	.dc.b	$1F,$E1,$E7,$11,$EA,$8F,$F0,$66,$1F,$4C,$8F,$F2,$29,$E5,$EE,$74
	.dc.b	$C3,$C9,$3F,$EE,$A3,$A3,$88,$D6,$1E,$87,$FC,$EF,$EF,$7F,$30,$EA
	.dc.b	$34,$97,$F2,$B1,$F0,$9E,$55,$AF,$CD,$7F,$CF,$B3,$C2,$7A,$C7,$F3
	.dc.b	$6A,$EA,$FC,$EF,$2B,$E8,$91,$C1,$DC,$93,$DE,$FB,$47,$F9,$27,$E3
	.dc.b	$BF,$D5,$33,$F4,$3C,$64,$62,$7F,$C9,$31,$8F,$3B,$FA,$7B,$63,$EF
	.dc.b	$FC,$A7,$DA,$F8,$6F,$D4,$B6,$3F,$0E,$D7,$FD,$D7,$C6,$8C,$47,$E6
	.dc.b	$3C,$7B,$CD,$FB,$56,$D7,$F8,$D6,$FF,$6B,$18,$F9,$7F,$C4,$F9,$9F
	.dc.b	$B8,$6E,$1F,$96,$6B,$E0,$6B,$FA,$6F,$9B,$B8,$FF,$03,$6E,$26,$98
	.dc.b	$F9,$CF,$CB,$FC,$AF,$F0,$66,$DF,$E9,$1A,$FB,$9A,$FE,$7B,$C8,$FF
	.dc.b	$38,$FE,$55,$FC,$EB,$C9,$34,$93,$FA,$EF,$9E,$FD,$2C,$7F,$14,$6C
	.dc.b	$64,$E2,$58,$FD,$4F,$E0,$BF,$AD,$FB,$4F,$F0,$2B,$67,$FC,$02,$DE
	.dc.b	$6D,$F6,$2B,$EE,$F3,$FE,$06,$FF,$D2,$2D,$7F,$11,$CF,$A3,$F8,$0F
	.dc.b	$B5,$FB,$AF,$8A,$FE,$52,$D9,$FE,$0E,$D7,$F0,$2D,$7F,$8B,$FC,$B5
	.dc.b	$E4,$FE,$68,$DE,$43,$4C,$76,$7E,$7F,$5F,$E6,$ED,$BF,$EC,$FE,$67
	.dc.b	$CD,$23,$83,$FD,$47,$F8,$4B,$F5,$EF,$DF,$CF,$F1,$08,$C7,$07,$F5
	.dc.b	$AA,$46,$1F,$EC,$CD,$9F,$E8,$FF,$4B,$FC,$6B,$F7,$7B,$46,$91,$FE
	.dc.b	$53,$F0,$DB,$69,$B2,$C7,$F5,$28,$D4,$7E,$A3,$E9,$FC,$9B,$FF,$59
	.dc.b	$6E,$E6,$98,$FA,$89,$5F,$DF,$6B,$FB,$04,$63,$2F,$EB,$FF,$53,$C8
	.dc.b	$95,$BC,$11,$F0,$E3,$FD,$95,$E3,$7F,$AB,$3D,$2F,$CD,$12,$3F,$C2
	.dc.b	$EF,$CB,$E0,$FE,$56,$DE,$0F,$EE,$93,$FD,$03,$83,$E6,$4E,$63,$FB
	.dc.b	$D1,$FD,$3B,$5D,$BF,$D7,$1B,$AF,$C1,$8D,$9F,$14,$A7,$D8,$C3,$F9
	.dc.b	$DF,$E2,$3D,$49,$1F,$82,$8C,$BF,$DE,$FF,$2C,$FF,$3B,$7E,$25,$FE
	.dc.b	$06,$6C,$66,$FF,$AB,$CE,$5B,$1C,$F3,$FE,$B7,$F4,$3E,$B2,$46,$43
	.dc.b	$FC,$07,$9A,$47,$F8,$F9,$AF,$F6,$4A,$60,$FA,$EF,$F0,$33,$67,$FC
	.dc.b	$0A,$DA,$F4,$4E,$BD,$9F,$ED,$BD,$BE,$B6,$DF,$BD,$AF,$7B,$FF,$15
	.dc.b	$CF,$C1,$F7,$3B,$92,$7B,$BF,$E8,$EF,$11,$C7,$FE,$EE,$CB,$D3,$5F
	.dc.b	$C7,$E3,$0F,$B4,$F9,$CF,$A8,$8F,$A9,$E8,$70,$7F,$DF,$CD,$7A,$78
	.dc.b	$9F,$E9,$8A,$F0,$BF,$0B,$DB,$F8,$A4,$9C,$1F,$1B,$ED,$3F,$C1,$4F
	.dc.b	$DF,$6B,$F0,$5A,$FF,$AE,$B3,$46,$36,$FA,$7F,$6D,$F4,$89,$5F,$49
	.dc.b	$F4,$E1,$E5,$CB,$D1,$3F,$A8,$8C,$1F,$63,$ED,$FE,$D9,$23,$07,$DC
	.dc.b	$F8,$AF,$F1,$2F,$37,$E4,$DB,$1F,$97,$6B,$F6,$5A,$FC,$6D,$BE,$9B
	.dc.b	$E1,$7E,$AE,$5F,$8C,$7F,$DA,$CF,$E8,$5B,$1F,$83,$6B,$F0,$DB,$1F
	.dc.b	$94,$FA,$9F,$F1,$CF,$E2,$3F,$ED,$DA,$F9,$11,$89,$FD,$63,$63,$F1
	.dc.b	$CD,$7C,$C7,$E1,$4F,$BF,$97,$C6,$BF,$D9,$7D,$7F,$F9,$1B,$E6,$7B
	.dc.b	$61,$FA,$48,$C3,$FB,$96,$DF,$E6,$DA,$FF,$BD,$3D,$2F,$F2,$3F,$8F
	.dc.b	$C9,$B5,$FE,$7B,$6C,$FE,$C5,$AF,$E0,$FE,$27,$FD,$0D,$3F,$C1,$DB
	.dc.b	$1F,$A9,$6B,$B3,$FD,$13,$1E,$D7,$DE,$7B,$B1,$FA,$75,$FE,$5A,$7C
	.dc.b	$9E,$EF,$CE,$8C,$57,$96,$3E,$9D,$7D,$B3,$A1,$4E,$86,$39,$BF,$4B
	.dc.b	$7F,$82,$6C,$7F,$9B,$BE,$1D,$F5,$0F,$89,$FF,$39,$F6,$30,$FF,$3A
	.dc.b	$DA,$F7,$5A,$9C,$C6,$27,$FD,$6D,$18,$BE,$91,$83,$FC,$FF,$3B,$BE
	.dc.b	$D1,$89,$DC,$62,$3C,$11,$8A,$E9,$7E,$19,$FE,$06,$7F,$8A,$C8,$7F
	.dc.b	$BA,$0A,$7E,$30,$A7,$03,$E6,$1F,$CB,$CF,$C0,$63,$FE,$BD,$69,$93
	.dc.b	$F1,$4E,$5E,$47,$53,$C0,$43,$F0,$83,$E3,$7F,$33,$D4,$DB,$FD,$43
	.dc.b	$18,$6E,$BF,$8D,$AE,$0F,$9B,$E1,$7F,$0B,$1F,$EC,$88,$C1,$1F,$D0
	.dc.b	$07,$BF,$0F,$F4,$B7,$F4,$D1,$96,$9F,$B9,$7F,$8B,$1F,$18,$A7,$AC
	.dc.b	$FD,$E8,$FC,$47,$D7,$CB,$FE,$BA,$6C,$66,$7F,$80,$B4,$FE,$18,$7C
	.dc.b	$B7,$F8,$93,$72,$FC,$04,$7F,$13,$8F,$A0,$78,$8F,$9C,$43,$E8,$9F
	.dc.b	$CD,$F9,$FF,$C1,$8F,$F2,$53,$FC,$72,$5F,$1F,$F9,$53,$83,$FB,$07
	.dc.b	$57,$FB,$41,$A5,$AF,$EE,$31,$98,$FE,$88,$D7,$C8,$78,$8A,$62,$3F
	.dc.b	$A3,$BC,$ED,$D9,$CE,$73,$3F,$E0,$F8,$C4,$FF,$50,$87,$9F,$FE,$13
	.dc.b	$EE,$FF,$73,$B6,$32,$7F,$5B,$29,$FD,$EF,$BF,$99,$1F,$13,$E9,$E8
	.dc.b	$FE,$D3,$1F,$7D,$AF,$DF,$C7,$FC,$23,$DF,$E4,$24,$7A,$DE,$43,$B7
	.dc.b	$E3,$57,$29,$FA,$C7,$C3,$C9,$E0,$7B,$89,$1F,$15,$F8,$3D,$48,$7D
	.dc.b	$17,$F8,$65,$A7,$F8,$64,$A7,$F8,$66,$38,$47,$F8,$8A,$3E,$87,$87
	.dc.b	$F3,$08,$F9,$B2,$F4,$3F,$05,$AF,$E7,$F8,$7F,$D0,$1F,$09,$FE,$55
	.dc.b	$24,$7F,$8E,$DF,$F2,$12,$F5,$A9,$1F,$F2,$41,$93,$FC,$85,$3F,$E5
	.dc.b	$F2,$38,$3F,$CC,$03,$E0,$3F,$98,$BF,$E6,$13,$06,$9C,$EF,$FE,$63
	.dc.b	$7F,$CD,$44,$8F,$F2,$83,$5F,$ED,$AF,$FB,$44,$8F,$6D,$BF,$C5,$71
	.dc.b	$FE,$84,$8F,$E7,$27,$FB,$E3,$CD,$FA,$9D,$D9,$24,$7F,$46,$BF,$F4
	.dc.b	$6B,$C0,$FE,$25,$E7,$3D,$0E,$31,$B8,$C3,$B9,$E3,$5E,$C9,$FF,$4E
	.dc.b	$4E,$5B,$19,$F3,$B9,$CF,$FA,$06,$3F,$D5,$0B,$F7,$7A,$47,$FD,$5E
	.dc.b	$FF,$AD,$88,$F8,$FD,$4F,$C8,$46,$23,$FD,$2D,$E8,$3C,$FE,$F1,$FE
	.dc.b	$8D,$6E,$FF,$EC,$5C,$3F,$D3,$47,$FB,$F4,$93,$F3,$78,$BF,$4D,$18
	.dc.b	$8F,$F3,$F6,$7F,$B8,$7E,$27,$57,$F0,$58,$C4,$6A,$31,$E5,$7E,$97
	.dc.b	$B7,$E3,$36,$7E,$63,$5E,$AD,$7E,$9F,$AB,$78,$18,$23,$0D,$9F,$E2
	.dc.b	$E7,$FB,$E8,$A7,$ED,$78,$DE,$7B,$D7,$FE,$4D,$18,$6C,$66,$7F,$97
	.dc.b	$4E,$5F,$E6,$1C,$FF,$37,$D8,$FE,$5C,$D8,$CF,$0A,$72,$D8,$FE,$6D
	.dc.b	$8B,$DA,$FE,$11,$F9,$5F,$E2,$5B,$3F,$8F,$6B,$F4,$1A,$FF,$A3,$23
	.dc.b	$FA,$17,$57,$FA,$31,$BC,$7F,$C7,$5B,$E0,$9F,$CB,$7B,$9F,$C5,$FE
	.dc.b	$66,$E0,$69,$1F,$D0,$7B,$FF,$C7,$7E,$63,$F3,$CD,$8C,$88,$CB,$4F
	.dc.b	$D4,$7C,$0E,$DF,$C4,$B5,$FD,$7B,$79,$7E,$EB,$76,$7C,$17,$F2,$DF
	.dc.b	$2F,$F5,$CD,$BF,$DA,$35,$FC,$AB,$5F,$DD,$FC,$25,$C3,$F1,$2D,$B3
	.dc.b	$4E,$1C,$BF,$7E,$BF,$8D,$B5,$FF,$26,$B5,$FC,$67,$37,$E1,$DF,$EE
	.dc.b	$9F,$0B,$FE,$43,$7F,$91,$B5,$FF,$2A,$B5,$FE,$90,$DC,$3F,$3B,$D9
	.dc.b	$FA,$B2,$4F,$E2,$1F,$EF,$5C,$6F,$86,$A4,$9F,$F0,$04,$65,$FF,$01
	.dc.b	$5F,$74,$D2,$7A,$DC,$63,$FC,$19,$3C,$5B,$59,$7F,$CD,$BE,$3B,$E2
	.dc.b	$3F,$87,$43,$CB,$FE,$83,$39,$EB,$FF,$3A,$CF,$0A,$DD,$69,$B1,$F1
	.dc.b	$B1,$FF,$0B,$65,$E4,$8F,$4C,$7F,$86,$F0,$E9,$1F,$C4,$5B,$9F,$3C
	.dc.b	$7F,$C1,$F3,$DD,$AF,$F1,$8F,$89,$FE,$2C,$47,$F6,$6F,$82,$E4,$78
	.dc.b	$08,$7F,$3D,$F8,$AF,$F3,$0B,$FD,$0D,$AF,$F2,$46,$BF,$D2,$9B,$B3
	.dc.b	$E2,$DE,$BF,$F8,$AD,$B5,$FE,$39,$6B,$FD,$99,$AF,$C5,$3C,$65,$3F
	.dc.b	$1F,$7F,$CC,$7E,$33,$BF,$29,$A4,$9F,$F3,$E1,$FD,$DB,$E3,$BE,$C3
	.dc.b	$FD,$D8,$FF,$54,$B8,$E5,$FF,$27,$36,$38,$71,$7F,$2B,$7C,$93,$E2
	.dc.b	$79,$1F,$9C,$C3,$4F,$E8,$1A,$FF,$89,$FD,$47,$C7,$FF,$56,$84,$61
	.dc.b	$FE,$E8,$62,$35,$19,$F2,$BF,$CC,$5E,$A3,$E4,$F9,$36,$F8,$B6,$77
	.dc.b	$3D,$F9,$BF,$D7,$71,$E9,$25,$7C,$F7,$FB,$03,$FE,$3D,$F1,$5D,$5F
	.dc.b	$D9,$23,$FB,$37,$8B,$E6,$9C,$BE,$8F,$91,$FF,$4B,$4F,$FB,$51,$AF
	.dc.b	$F9,$1D,$A9,$A4,$E5,$FF,$3C,$36,$3F,$CF,$8D,$7F,$D0,$4D,$7F,$C6
	.dc.b	$2D,$7C,$11,$C8,$8F,$26,$BC,$9A,$F2,$6B,$C8,$E0,$43,$81,$4E,$05
	.dc.b	$38,$1A,$21,$A2,$9A,$30,$43,$04,$3E,$43,$E4,$69,$DC,$A7,$42,$9C
	.dc.b	$CA,$73,$3E,$83,$CD,$A7,$32,$9C,$CA,$73,$29,$CF,$E0,$5F,$03,$F9
	.dc.b	$88,$46,$32,$FF,$5A,$36,$3F,$D7,$4D,$7E,$CF,$BD,$FE,$8B,$E1,$FB
	.dc.b	$AD,$8C,$9F,$D2,$F1,$46,$63,$F1,$CD,$7F,$1C,$D7,$F1,$CD,$8E,$31
	.dc.b	$CE,$30,$47,$04,$9F,$1B,$4F,$1B,$82,$38,$21,$E4,$38,$15,$DB,$4D
	.dc.b	$9E,$83,$65,$76,$D3,$67,$B0,$D9,$5D,$B4,$DC,$FC,$08,$CB,$63,$33
	.dc.b	$F0,$67,$2D,$8C,$9D,$A7,$83,$E9,$21,$E5,$29,$E1,$3F,$D8,$C3,$3F
	.dc.b	$E0,$FF,$13,$FE,$36,$24,$7F,$0C,$7F,$D4,$FE,$DF,$F9,$27,$E4,$6F
	.dc.b	$9D,$FE,$CF,$6D,$1A,$7F,$AC,$9F,$E5,$7F,$49,$B7,$FD,$6B,$CA,$8C
	.dc.b	$47,$92,$30,$F3,$37,$1A,$5E,$13,$CD,$7D,$4F,$E3,$E3,$0D,$8C,$CF
	.dc.b	$F8,$8E,$72,$FF,$89,$78,$D9,$BF,$00,$ED,$18,$6C,$66,$7C,$87,$FB
	.dc.b	$BE,$74,$E0,$FF,$79,$7C,$B3,$E1,$7D,$10,$8C,$3B,$31,$1A,$8C,$EB
	.dc.b	$F4,$3F,$01,$FD,$55,$F9,$9B,$81,$A7,$EB,$7C,$4F,$AD,$8F,$FB,$89
	.dc.b	$AF,$E0,$1A,$8F,$F0,$1F,$07,$9B,$EC,$5F,$43,$49,$3E,$78,$CB,$E9
	.dc.b	$AF,$EF,$3C,$FF,$7D,$B1,$E1,$F0,$BC,$87,$F0,$1E,$E5,$78,$7F,$0C
	.dc.b	$FE,$F0,$4A,$FF,$07,$3C,$EB,$E4,$DF,$F8,$1B,$6C,$57,$86,$7E,$6E
	.dc.b	$1D,$21,$F2,$5F,$E0,$32,$3F,$E1,$97,$F2,$9F,$0C,$F8,$8F,$E2,$8D
	.dc.b	$9F,$F3,$33,$73,$DC,$FF,$19,$FB,$57,$D3,$FD,$C6,$DF,$DB,$6E,$7F
	.dc.b	$A1,$6C,$7E,$C0,$61,$D5,$C3,$5F,$DA,$B5,$FA,$CD,$E3,$FC,$BF,$E0
	.dc.b	$9F,$82,$24,$FE,$6F,$91,$1F,$CB,$FE,$7D,$F1,$9F,$92,$6D,$7E,$45
	.dc.b	$B9,$7E,$DD,$B1,$FC,$1B,$0E,$EF,$5B,$EF,$AF,$F0,$47,$F9,$E7,$E4
	.dc.b	$BF,$EA,$69,$FC,$E3,$70,$FE,$04,$D8,$F8,$4D,$7F,$C6,$87,$E8,$F8
	.dc.b	$3E,$D4,$FF,$8E,$C9,$1F,$27,$CA,$7D,$1F,$EB,$1B,$1F,$AB,$6C,$FF
	.dc.b	$8D,$27,$F7,$3F,$25,$F9,$FF,$C4,$3F,$8D,$B6,$FD,$DB,$A3,$11,$FB
	.dc.b	$D8,$C2,$9F,$E4,$7E,$B7,$EA,$7E,$7E,$C9,$A4,$7F,$97,$F9,$AF,$97
	.dc.b	$FF,$26,$C3,$F4,$2F,$C2,$4E,$7C,$8F,$DA,$46,$2B,$6D,$8C,$D7,$A3
	.dc.b	$E8,$FF,$75,$F2,$1F,$CD,$9B,$19,$11,$96,$9F,$93,$FC,$BD,$F2,$BF
	.dc.b	$C0,$F9,$7F,$9C,$46,$27,$F1,$66,$E7,$4D,$32,$7D,$A9,$FF,$39,$11
	.dc.b	$FF,$44,$3F,$D0,$F1,$7D,$37,$F1,$1F,$A3,$C3,$6B,$F4,$D3,$A5,$3A
	.dc.b	$69,$F9,$5F,$CD,$DB,$7F,$38,$6C,$7F,$33,$EA,$46,$39,$BF,$65,$18
	.dc.b	$AF,$E6,$BF,$A2,$BE,$66,$93,$0F,$E6,$D1,$97,$A6,$1E,$5F,$7F,$9B
	.dc.b	$63,$C9,$A7,$F2,$D9,$CB,$63,$F0,$7F,$BC,$BE,$DF,$F9,$97,$4F,$E0
	.dc.b	$A3,$13,$FD,$50,$DC,$E9,$A6,$4F,$EF,$7D,$DF,$B1,$F2,$7F,$5F,$F1
	.dc.b	$5F,$CB,$67,$F5,$ED,$76,$D7,$FD,$19,$EE,$FC,$AE,$1F,$EF,$1E,$C7
	.dc.b	$FA,$29,$B1,$99,$FF,$47,$C7,$F7,$DE,$3F,$F5,$F4,$7F,$5B,$6C,$67
	.dc.b	$E0,$BF,$B0,$C6,$27,$EC,$78,$3F,$D4,$7F,$08,$F9,$2D,$8C,$88,$CB
	.dc.b	$4F,$CE,$7B,$9F,$9B,$23,$83,$F0,$9C,$DE,$DE,$7F,$BD,$0D,$BE,$BC
	.dc.b	$E7,$97,$F0,$F1,$F8,$DD,$25,$7E,$4B,$93,$FB,$CF,$47,$E4,$E7,$E6
	.dc.b	$CF,$E0,$CF,$96,$53,$FB,$67,$17,$F2,$3E,$8D,$9F,$2B,$2F,$14,$FF
	.dc.b	$AE,$31,$C1,$1F,$E8,$DB,$FC,$C8,$F9,$51,$C0,$FF,$1C,$C6,$5D,$F0
	.dc.b	$7A,$EB,$FB,$E5,$FB,$A7,$C9,$AF,$C2,$8C,$1F,$AB,$C2,$4F,$BC,$EC
	.dc.b	$7F,$12,$D3,$A2,$ED,$E8,$47,$E7,$99,$39,$57,$ED,$C8,$E0,$FF,$6B
	.dc.b	$F8,$DF,$0C,$8E,$0F,$EA,$1F,$91,$7F,$2B,$7F,$74,$DD,$BF,$72,$D8
	.dc.b	$DF,$85,$FC,$07,$E5,$7F,$C3,$9F,$92,$FE,$D1,$B3,$FC,$1D,$AF,$F8
	.dc.b	$E7,$8D,$18,$86,$DF,$C2,$3E,$13,$0D,$85,$FF,$00,$9C,$D9,$E5,$3F
	.dc.b	$E6,$DE,$8F,$E2,$BA,$7B,$A3,$FC,$E3,$D3,$FE,$F6,$8C,$47,$83,$F1
	.dc.b	$FF,$F2,$BE,$DF,$C5,$9B,$1F,$DC,$DA,$FF,$94,$5A,$FE,$EB,$8F,$FC
	.dc.b	$44,$F6,$C3,$C2,$FC,$D3,$11,$FE,$61,$9F,$E1,$F5,$DF,$B3,$FC,$C3
	.dc.b	$18,$AF,$04,$7F,$2D,$F5,$3F,$C9,$1F,$77,$FC,$81,$B3,$FC,$CD,$AF
	.dc.b	$CD,$6B,$F3,$35,$7D,$87,$F9,$71,$B1,$FE,$73,$6E,$9F,$CB,$27,$FD
	.dc.b	$91,$AF,$C7,$AF,$51,$EF,$9F,$93,$C0,$CD,$F1,$7E,$A2,$57,$F4,$38
	.dc.b	$CB,$FD,$17,$B9,$F3,$7F,$E6,$3E,$8F,$C0,$CF,$F9,$F6,$72,$FE,$87
	.dc.b	$0F,$F1,$24,$62,$77,$3F,$88,$8C,$4F,$9A,$3F,$05,$C3,$F5,$1F,$41
	.dc.b	$EA,$8F,$47,$C7,$7D,$02,$47,$9D,$FF,$49,$D7,$8B,$E8,$7B,$D7,$F2
	.dc.b	$5C,$7F,$CF,$51,$89,$D4,$F9,$A7,$2F,$8F,$5F,$95,$87,$E0,$B5,$FC
	.dc.b	$8B,$E6,$6C,$6A,$3C,$E7,$C3,$C7,$FD,$11,$3D,$DB,$5F,$11,$B3,$B3
	.dc.b	$FC,$FF,$1F,$E4,$BF,$BA,$7D,$26,$A7,$F5,$F1,$87,$67,$FA,$B5,$C5
	.dc.b	$FC,$06,$D7,$AF,$9B,$F3,$B5,$EB,$3E,$66,$6C,$7B,$57,$F4,$B6,$C6
	.dc.b	$67,$B9,$F6,$63,$C8,$FC,$33,$B6,$BF,$65,$C1,$DC,$87,$C6,$20,$F8
	.dc.b	$A7,$C6,$3E,$33,$FD,$61,$EF,$FE,$DA,$73,$18,$D7,$FD,$EB,$18,$B5
	.dc.b	$6E,$F9,$46,$2B,$71,$89,$D5,$F3,$AF,$F5,$F5,$F3,$9C,$C7,$13,$F8
	.dc.b	$41,$FD,$54,$87,$FA,$5F,$76,$BF,$E2,$C8,$C3,$5F,$E1,$67,$02,$9F
	.dc.b	$83,$3F,$DC,$C7,$E0,$CF,$C4,$18,$3E,$81,$4F,$D3,$9C,$CF,$29,$E0
	.dc.b	$CB,$FD,$54,$FE,$23,$7F,$F4,$B3,$F8,$76,$B9,$3B,$6C,$DF,$F1,$D1
	.dc.b	$E2,$24,$79,$DA,$7A,$88,$78,$8A,$79,$8F,$21,$E7,$30,$7F,$09,$E0
	.dc.b	$FF,$77,$6B,$FB,$56,$CF,$ED,$67,$F0,$AD,$9F,$DE,$F1,$FE,$57,$E9
	.dc.b	$7F,$A1,$70,$FF,$BD,$DB,$66,$9E,$1D,$7E,$BE,$9F,$1A,$7C,$67,$A4
	.dc.b	$FE,$2D,$97,$FC,$01,$F2,$BC,$90,$69,$33,$FE,$A1,$CD,$E1,$9A,$78
	.dc.b	$4B,$8F,$F5,$78,$D4,$7C,$6C,$7C,$6F,$97,$7F,$C4,$3F,$D0,$F9,$FC
	.dc.b	$D8,$7F,$9D,$9B,$D1,$FD,$12,$7F,$A1,$1F,$80,$9F,$F3,$DF,$4A,$1F
	.dc.b	$A1,$7F,$53,$69,$FD,$4C,$A7,$F5,$32,$9F,$DB,$FB,$3C,$ED,$3D,$87
	.dc.b	$F6,$D7,$35,$FD,$35,$DC,$7C,$C2,$65,$EE,$7F,$06,$7F,$77,$F8,$BE
	.dc.b	$3C,$46,$92,$3B,$9F,$54,$E7,$A7,$C6,$7E,$CE,$1F,$E1,$1F,$C9,$BF
	.dc.b	$C2,$10,$69,$1F,$F0,$F1,$92,$BF,$E2,$37,$07,$F8,$87,$C2,$F9,$51
	.dc.b	$FE,$2C,$F0,$FD,$FE,$27,$F8,$C7,$8B,$17,$E6,$C1,$39,$8F,$79,$F7
	.dc.b	$F1,$F4,$69,$F0,$E3,$EB,$1F,$ED,$FF,$1B,$EC,$7D,$BF,$FA,$EC,$91
	.dc.b	$FE,$51,$7F,$CB,$1C,$B2,$3F,$C9,$AD,$79,$F9,$9F,$EC,$5F,$C2,$3F
	.dc.b	$D8,$84,$8F,$F2,$D3,$FE,$61,$A9,$1F,$E5,$66,$BF,$E6,$AF,$2F,$F1
	.dc.b	$53,$FE,$6E,$24,$7F,$9B,$03,$FE,$79,$23,$FC,$4D,$AF,$FA,$0A,$7F
	.dc.b	$D0,$A4,$7F,$D6,$21,$FF,$6D,$6B,$23,$FC,$E0,$D7,$FC,$A9,$D5,$E4
	.dc.b	$7F,$CD,$B5,$FE,$7B,$23,$FE,$FD,$E2,$FF,$7E,$11,$C1,$FE,$AA,$FD
	.dc.b	$3F,$FD,$52,$47,$07,$B6,$FF,$D1,$A4,$70,$7F,$2F,$7F,$D3,$04,$70
	.dc.b	$7B,$2B,$FD,$46,$47,$07,$FB,$5B,$F3,$4F,$F6,$E7,$37,$ED,$DB,$06
	.dc.b	$9F,$EE,$65,$F9,$BD,$23,$FE,$EC,$32,$7E,$30,$7A,$DF,$EF,$5C,$3F
	.dc.b	$A2,$B6,$34,$D7,$F2,$6D,$7F,$CA,$99,$FE,$FB,$AC,$D2,$47,$ED,$8F
	.dc.b	$CD,$0F,$A2,$FE,$A7,$C0,$E2,$FE,$29,$18,$6C,$FF,$AB,$63,$FA,$CF
	.dc.b	$C1,$3C,$8F,$E3,$5F,$35,$86,$D7,$B6,$74,$A7,$4D,$3E,$17,$B8,$F7
	.dc.b	$38,$C7,$EC,$23,$0E,$CC,$46,$9F,$E5,$51,$A9,$F9,$9D,$6F,$6B,$FD
	.dc.b	$79,$E1,$E1,$B7,$FB,$C9,$D7,$9F,$FC,$CE,$B4,$D9,$FE,$03,$BB,$DF
	.dc.b	$7B,$FF,$CE,$7F,$69,$FB,$53,$98,$C5,$70,$AD,$AA,$DC,$61,$F2,$7C
	.dc.b	$0F,$F4,$3F,$A0,$B4,$69,$1F,$46,$8F,$84,$FE,$95,$F8,$4F,$B5,$B3
	.dc.b	$A6,$BF,$D3,$F6,$F8,$91,$88,$FE,$A5,$F0,$CF,$81,$FE,$8A,$D8,$FC
	.dc.b	$43,$7A,$FF,$C3,$2D,$8F,$BD,$CD,$FD,$33,$9B,$EF,$D7,$DE,$24,$FF
	.dc.b	$0A,$7F,$C4,$3F,$01,$FD,$DE,$3F,$28,$D8,$CF,$8B,$FD,$63,$57,$C2
	.dc.b	$7F,$6E,$6D,$FF,$6B,$6B,$F7,$63,$FC,$10,$BE,$1D,$F1,$1F,$D8,$1B
	.dc.b	$1F,$A3,$6E,$1F,$E3,$46,$C7,$CE,$AF,$01,$1F,$03,$DB,$E2,$5F,$19
	.dc.b	$49,$3F,$E1,$98,$CB,$FE,$1B,$F8,$B7,$C6,$7F,$72,$6C,$7E,$D5,$B8
	.dc.b	$FF,$91,$9B,$1F,$C1,$EB,$9B,$FE,$4A,$6B,$9B,$F0,$BF,$AE,$F9,$B8
	.dc.b	$FD,$60,$E1,$FE,$BB,$EB,$7F,$9D,$9F,$E0,$CD,$C3,$FC,$AE,$D8,$FE
	.dc.b	$30,$D7,$FB,$2F,$07,$EF,$BC,$AF,$DE,$92,$3C,$0F,$2F,$77,$EA,$D7
	.dc.b	$79,$C1,$FE,$3F,$70,$FF,$25,$F8,$F7,$C8,$7F,$86,$1B,$1F,$C6,$DB
	.dc.b	$8F,$F9,$B1,$B1,$E6,$F9,$1F,$F0,$FD,$9A,$48,$FF,$02,$71,$BE,$4B
	.dc.b	$F1,$C4,$8F,$C7,$3F,$8C,$F9,$37,$CA,$7F,$8A,$9B,$3F,$CB,$5B,$97
	.dc.b	$F9,$F9,$B1,$FB,$8F,$95,$FC,$E1,$1F,$0B,$FE,$3A,$F9,$67,$CB,$FF
	.dc.b	$8B,$E7,$FD,$0F,$18,$F5,$BF,$CC,$8D,$8C,$CE,$DF,$F4,$6F,$8E,$F9
	.dc.b	$8F,$F1,$FB,$63,$FA,$13,$72,$DC,$FE,$BF,$69,$5F,$11,$B3,$FE,$98
	.dc.b	$F6,$BF,$CE,$BF,$13,$7D,$CF,$E9,$ED,$F3,$3F,$D7,$B8,$7F,$8A,$2B
	.dc.b	$6F,$F5,$FF,$9A,$7C,$A7,$ED,$9B,$1F,$E7,$F6,$BF,$EA,$66,$BC,$63
	.dc.b	$81,$1E,$0D,$78,$35,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$FF,$21,$C7
	.dc.b	$91,$B1,$DD,$A7,$42,$9C,$CA,$73,$3F,$D0,$51,$CD,$B1,$CD,$A7,$32
	.dc.b	$9C,$CA,$73,$11,$CD,$B1,$CD,$A7,$32,$9C,$CA,$73,$8E,$44,$8E,$4D
	.dc.b	$8E,$4F,$7F,$8B,$FF,$80,$11,$FE,$1D,$6B,$FE,$FE,$C3,$FD,$79,$18
	.dc.b	$8F,$83,$9F,$FA,$8F,$C6,$FC,$03,$63,$F1,$0D,$7E,$8B,$5E,$27,$F8
	.dc.b	$38,$C9,$4C,$C7,$47,$07,$C6,$8E,$4D,$33,$7F,$C1,$DB,$1A,$8C,$4F
	.dc.b	$32,$4F,$98,$D9,$1C,$BE,$46,$BE,$43,$83,$87,$C4,$D8,$EF,$1A,$21
	.dc.b	$CC,$A7,$33,$04,$30,$43,$A9,$C0,$A6,$CA,$6D,$F0,$B9,$6B,$99,$F8
	.dc.b	$84,$7E,$01,$D6,$36,$B1,$FF,$7E,$FC,$D7,$F0,$72,$47,$F9,$8E,$32
	.dc.b	$FF,$57,$8F,$F2,$39,$1C,$1F,$E3,$9F,$65,$F7,$9F,$EE,$BF,$91,$FF
	.dc.b	$20,$C6,$25,$18,$9D,$47,$16,$BF,$C5,$F7,$FF,$34,$E7,$FC,$6C,$91
	.dc.b	$E5,$8C,$BF,$DC,$FB,$BF,$13,$E1,$FF,$69,$AF,$EE,$91,$83,$F2,$DE
	.dc.b	$73,$2F,$EB,$39,$79,$BC,$0F,$0E,$2A,$D4,$61,$B1,$EA,$3C,$5C,$7F
	.dc.b	$9A,$D9,$A7,$92,$7D,$53,$98,$C4,$06,$8F,$33,$D2,$48,$FD,$2C,$65
	.dc.b	$FD,$3F,$82,$F0,$7C,$F0,$8C,$3E,$53,$FD,$CE,$E3,$4F,$EE,$93,$C3
	.dc.b	$2F,$3E,$5E,$52,$6F,$F3,$B0,$FF,$09,$F1,$FE,$D2,$FF,$12,$47,$F6
	.dc.b	$F9,$BE,$73,$D9,$F2,$DF,$E5,$46,$DF,$FA,$B6,$74,$A7,$4D,$3E,$5E
	.dc.b	$1E,$17,$E3,$65,$E6,$8C,$47,$C7,$BF,$95,$F3,$9C,$89,$1D,$E3,$2F
	.dc.b	$F8,$DF,$D7,$FC,$29,$1F,$9A,$FF,$10,$BE,$D1,$E2,$21,$F6,$AF,$FC
	.dc.b	$7D,$3C,$5B,$1C,$8F,$E3,$7F,$0A,$FA,$EF,$E3,$6D,$8C,$C6,$0F,$07
	.dc.b	$17,$E1,$23,$F6,$8D,$C3,$C7,$1E,$13,$FD,$AD,$5C,$EF,$FD,$B1,$08
	.dc.b	$C3,$AB,$EC,$FF,$2E,$F3,$9F,$11,$F8,$86,$C7,$E1,$CF,$CA,$38,$E2
	.dc.b	$EA,$D8,$F8,$67,$D4,$CB,$CC,$4E,$2F,$E7,$7F,$78,$FB,$6F,$E7,$6D
	.dc.b	$AC,$C6,$0F,$D9,$F4,$BE,$4B,$F4,$99,$FF,$08,$6D,$7F,$0D,$6B,$FE
	.dc.b	$A9,$9E,$0F,$A8,$C7,$C5,$69,$FD,$D4,$23,$1D,$5F,$C0,$5B,$19,$9E
	.dc.b	$47,$0F,$B3,$FD,$B7,$CF,$FF,$1A,$6C,$7F,$1F,$6B,$E3,$6B,$E0,$F5
	.dc.b	$6E,$1F,$C3,$DB,$7F,$E8,$EB,$FE,$4B,$1C,$29,$F7,$3F,$EC,$F7,$FD
	.dc.b	$56,$7F,$07,$71,$C1,$F8,$06,$C7,$F0,$B3,$5E,$F7,$DA,$70,$37,$FA
	.dc.b	$5F,$3D,$F8,$86,$C6,$9A,$F3,$6B,$FC,$5F,$D5,$FE,$E5,$3F,$90,$6C
	.dc.b	$76,$6B,$FE,$92,$EE,$FE,$3D,$3B,$F4,$BF,$8B,$D7,$F9,$DB,$9B,$17
	.dc.b	$F9,$11,$19,$7C,$9C,$9F,$87,$7F,$12,$4E,$9F,$F7,$FF,$D2,$BE,$27
	.dc.b	$F9,$67,$AD,$86,$D7,$C5,$E9,$56,$E1,$5B,$69,$FD,$6F,$EE,$BF,$9B
	.dc.b	$72,$FF,$5A,$6C,$64,$FE,$94,$58,$FF,$75,$FD,$47,$F3,$4F,$80,$34
	.dc.b	$93,$EF,$F0,$3F,$7A,$FE,$AF,$2F,$E6,$2F,$EB,$8F,$B1,$AF,$C1,$D7
	.dc.b	$EC,$36,$3F,$B8,$7E,$01,$FD,$2F,$E4,$FF,$B7,$B6,$32,$23,$2D,$3E
	.dc.b	$27,$EA,$2F,$8A,$E1,$1F,$D5,$71,$4E,$63,$81,$97,$15,$F8,$E9,$DD
	.dc.b	$72,$AF,$EB,$DC,$EF,$9A,$FE,$AE,$D8,$FC,$03,$79,$7F,$29,$3A,$CF
	.dc.b	$DD,$5F,$93,$6B,$FD,$A7,$7F,$9A,$FF,$49,$3F,$44,$E3,$C9,$FA,$71
	.dc.b	$93,$87,$07,$F5,$47,$F9,$61,$0F,$A5,$BF,$F4,$EF,$CC,$FF,$DB,$BE
	.dc.b	$97,$F2,$E6,$CF,$EB,$1A,$FE,$C5,$AB,$67,$E9,$AF,$03,$FC,$1B,$18
	.dc.b	$6C,$66,$7F,$99,$4E,$5F,$D2,$76,$7F,$B4,$70,$FE,$7E,$D8,$FD,$EB
	.dc.b	$5F,$E0,$4D,$7C,$55,$D7,$3F,$AC,$D8,$FD,$23,$5F,$C0,$35,$FF,$0B
	.dc.b	$63,$FE,$FE,$FB,$AF,$85,$E9,$7F,$86,$5B,$26,$9F,$DD,$B9,$FF,$3B
	.dc.b	$1F,$D7,$33,$FC,$EB,$FE,$21,$F1,$1F,$82,$7F,$81,$37,$FE,$2F,$18
	.dc.b	$F3,$BF,$C6,$F3,$9A,$FE,$3D,$39,$7F,$D1,$8F,$F8,$EF,$BB,$FC,$59
	.dc.b	$F4,$DF,$E5,$06,$CF,$F9,$55,$AF,$F4,$96,$BF,$D1,$BF,$05,$FF,$1A
	.dc.b	$77,$FF,$96,$1B,$1F,$E5,$D6,$BF,$D4,$5A,$FE,$20,$FD,$C6,$9F,$00
	.dc.b	$FD,$EE,$1E,$EF,$33,$D6,$49,$F4,$E5,$FD,$CB,$4F,$91,$E6,$FF,$6B
	.dc.b	$7F,$C4,$5F,$70,$F8,$6F,$D4,$47,$F3,$C8,$C7,$67,$FA,$C5,$B8,$7F
	.dc.b	$AB,$DB,$3F,$E7,$2E,$0F,$F2,$87,$E7,$F6,$8D,$23,$FE,$7B,$FC,$07
	.dc.b	$FB,$A9,$1F,$9F,$E2,$A3,$FC,$33,$C9,$F7,$27,$E1,$1F,$AB,$DF,$F4
	.dc.b	$1F,$0B,$FE,$70,$ED,$34,$93,$F6,$BC,$5F,$DD,$FD,$5F,$F9,$C6,$3F
	.dc.b	$0D,$18,$BF,$C0,$B6,$33,$E0,$7F,$9A,$EB,$84,$78,$88,$7E,$47,$6F
	.dc.b	$F2,$66,$4F,$AD,$F5,$4A,$78,$CE,$74,$FF,$71,$65,$FD,$26,$F7,$5A
	.dc.b	$6C,$79,$31,$FF,$3E,$7A,$FE,$28,$FE,$53,$18,$75,$8A,$75,$9F,$F1
	.dc.b	$A7,$D3,$1E,$73,$F1,$7C,$9F,$83,$78,$FA,$7C,$30,$FF,$45,$36,$0D
	.dc.b	$3C,$5F,$A8,$3E,$43,$D7,$3F,$B1,$8C,$3F,$07,$2F,$D6,$C6,$34,$E3
	.dc.b	$5C,$2B,$FD,$49,$39,$AD,$CE,$54,$66,$7B,$4F,$17,$D8,$7F,$2F,$29
	.dc.b	$F5,$E7,$8C,$73,$9E,$53,$E0,$8C,$1C,$FA,$BF,$5D,$97,$C5,$9E,$91
	.dc.b	$8D,$FF,$29,$1D,$87,$C1,$7F,$93,$FE,$0F,$FD,$65,$F0,$3F,$81,$9F
	.dc.b	$F5,$A4,$E5,$46,$5E,$3E,$D2,$B8,$EB,$FE,$31,$BE,$17,$B6,$C6,$6B
	.dc.b	$85,$F4,$8C,$5F,$28,$F0,$CE,$AF,$C1,$8F,$FA,$DA,$7F,$CF,$ED,$7F
	.dc.b	$DA,$26,$32,$FE,$A0,$DC,$BF,$D0,$AF,$F3,$13,$04,$3E,$C9,$FC,$CC
	.dc.b	$FA,$E5,$3E,$81,$75,$FA,$AE,$4E,$87,$88,$87,$FB,$80,$FE,$77,$AF
	.dc.b	$E3,$9F,$C5,$9C,$DC,$47,$8A,$32,$47,$04,$7F,$7E,$7F,$BD,$5F,$D5
	.dc.b	$9F,$A1,$21,$E2,$7F,$C5,$87,$BC,$F2,$1F,$B2,$F9,$1F,$D4,$12,$30
	.dc.b	$7E,$FF,$93,$FA,$7C,$FF,$88,$F9,$3F,$83,$BF,$C0,$63,$2F,$F7,$F8
	.dc.b	$CC,$F8,$63,$E2,$9E,$02,$07,$F8,$2B,$FE,$95,$7F,$82,$FB,$7F,$BB
	.dc.b	$23,$FD,$83,$A3,$F9,$4C,$FF,$B0,$23,$F9,$6B,$63,$2F,$F6,$A6,$BE
	.dc.b	$3D,$3F,$10,$F8,$8F,$29,$FE,$A6,$C3,$F0,$F4,$E9,$F7,$CF,$F4,$33
	.dc.b	$C0,$E1,$FF,$59,$75,$97,$E7,$DF,$D3,$48,$FF,$78,$9F,$F8,$03,$1F
	.dc.b	$D6,$7B,$FF,$C8,$53,$7B,$FF,$2F,$DF,$FD,$6A,$30,$7E,$6B,$3F,$4B
	.dc.b	$F7,$0F,$67,$17,$E8,$1F,$F0,$2F,$81,$FE,$BA,$7F,$C1,$CD,$7E,$83
	.dc.b	$74,$FE,$E9,$1D,$67,$93,$F1,$6B,$BB,$F1,$A7,$E2,$78,$5F,$C2,$23
	.dc.b	$FC,$08,$7F,$06,$2F,$2F,$F8,$25,$C1,$FE,$0F,$1F,$47,$FE,$15,$69
	.dc.b	$FE,$15,$29,$B7,$FC,$47,$8B,$7F,$AA,$D9,$FC,$13,$5D,$C7,$F0,$07
	.dc.b	$EE,$93,$1F,$59,$EC,$F2,$5E,$4F,$D9,$D7,$FC,$76,$4A,$FF,$2F,$46
	.dc.b	$5F,$F2,$3F,$1F,$DF,$E0,$FF,$25,$B6,$32,$7F,$85,$CB,$1F,$E5,$8F
	.dc.b	$0F,$F0,$DD,$76,$7F,$91,$9F,$E5,$88,$CB,$FE,$5C,$6B,$FE,$54,$AF
	.dc.b	$C7,$7C,$47,$82,$0D,$23,$E8,$3A,$9F,$E5,$AC,$39,$B4,$C8,$CB,$E1
	.dc.b	$B8,$CF,$E3,$F1,$7E,$12,$3F,$20,$7E,$A7,$CF,$FF,$48,$E9,$FE,$8E
	.dc.b	$24,$7F,$BC,$23,$2F,$FB,$68,$FE,$67,$E9,$3D,$1F,$AC,$D8,$CC,$63
	.dc.b	$9B,$DF,$1F,$EF,$56,$BF,$EF,$56,$BF,$EA,$E6,$F7,$7F,$AC,$A1,$F9
	.dc.b	$67,$FA,$D7,$C7,$FF,$5D,$37,$D4,$FD,$C3,$67,$67,$FB,$21,$FF,$63
	.dc.b	$91,$C1,$EB,$3F,$DA,$67,$A0,$87,$F2,$F3,$E3,$90,$FC,$9E,$5F,$EB
	.dc.b	$8E,$0F,$F5,$A9,$23,$E4,$78,$FE,$CE,$BF,$E2,$5E,$67,$F0,$DF,$C1
	.dc.b	$BF,$DE,$54,$69,$1F,$F7,$F7,$9D,$FE,$FD,$FA,$3F,$F7,$E1,$23,$3E
	.dc.b	$D7,$F1,$78,$F4,$C4,$3D,$84,$3F,$8E,$0F,$6D,$EE,$7D,$08,$C3,$63
	.dc.b	$31,$F9,$43,$89,$A3,$07,$45,$EE,$FF,$29,$FA,$0D,$CF,$78,$C3,$CA
	.dc.b	$76,$A7,$71,$87,$E9,$92,$3F,$20,$D7,$D7,$F0,$0F,$80,$FE,$71,$F3
	.dc.b	$DE,$08,$C4,$A3,$13,$A8,$F5,$CE,$A7,$FD,$6F,$E4,$FF,$B6,$79,$6F
	.dc.b	$81,$FD,$0D,$B6,$69,$C3,$D3,$D4,$7F,$45,$6B,$FE,$C5,$AF,$F2,$DB
	.dc.b	$C4,$D1,$4D,$79,$9F,$D4,$72,$F1,$B6,$3F,$A8,$35,$D9,$EE,$3B,$7A
	.dc.b	$19,$7F,$BA,$52,$31,$CB,$FD,$8D,$B1,$99,$E4,$7F,$61,$EE,$FD,$D3
	.dc.b	$F8,$B3,$F1,$2E,$1F,$C6,$1F,$A3,$F0,$7F,$86,$F6,$FF,$51,$3F,$C6
	.dc.b	$FF,$04,$F6,$3F,$1C,$DA,$FC,$49,$FD,$BD,$C7,$67,$E2,$1B,$1F,$90
	.dc.b	$F8,$57,$C2,$FF,$43,$6C,$7F,$91,$1A,$FE,$41,$AF,$F6,$CC,$3C,$8F
	.dc.b	$EE,$C4,$69,$FD,$2B,$5F,$F0,$13,$5D,$CF,$F5,$3F,$8A,$FF,$50,$47
	.dc.b	$F9,$51,$AF,$E8,$1A,$FE,$A1,$AF,$F7,$3F,$1B,$F9,$93,$FD,$6D,$B1
	.dc.b	$A8,$FF,$09,$35,$34,$C7,$C4,$7F,$69,$F9,$5F,$D7,$B6,$7F,$6E,$D7
	.dc.b	$F8,$0B,$5F,$B1,$5E,$2D,$7E,$CC,$72,$F3,$BF,$B1,$3C,$1A,$FF,$89
	.dc.b	$6B,$FB,$C7,$93,$B8,$FE,$CC,$7F,$0E,$F8,$AF,$DE,$BF,$C0,$9A,$FF
	.dc.b	$0A,$6B,$BE,$EF,$EF,$D8,$78,$A3,$EC,$1F,$C2,$BE,$2D,$5F,$C5,$DB
	.dc.b	$1F,$8A,$0E,$38,$78,$4F,$F1,$DD,$59,$7D,$92,$5F,$D9,$7C,$3A,$7C
	.dc.b	$32,$70,$7E,$22,$7F,$C8,$1E,$6F,$F3,$BE,$5F,$ED,$B9,$7F,$08,$8C
	.dc.b	$1D,$FD,$3F,$C3,$12,$BF,$0C,$F6,$BE,$91,$F8,$37,$FC,$A7,$F1,$9D
	.dc.b	$49,$1F,$A6,$E9,$FF,$2D,$47,$F3,$5F,$6F,$FC,$03,$3F,$CE,$DB,$E1
	.dc.b	$FF,$B0,$7D,$2F,$E5,$71,$AE,$26,$CC,$BF,$C0,$8D,$AF,$F5,$71,$30
	.dc.b	$FE,$7F,$F1,$BF,$59,$FF,$32,$35,$34,$9C,$9F,$E3,$76,$C7,$FA,$E1
	.dc.b	$AF,$F4,$46,$BD,$A3,$81,$1E,$0D,$78,$35,$E0,$68,$86,$8A,$68,$C1
	.dc.b	$0C,$10,$FE,$CB,$1E,$46,$C7,$76,$9D,$0A,$73,$29,$CC,$F8,$EF,$16
	.dc.b	$9C,$4A,$71,$29,$C4,$F9,$AF,$16,$9C,$4A,$71,$3F,$DF,$73,$FE,$E6
	.dc.b	$F7,$7F,$E0,$04,$7F,$78,$6B,$FE,$F6,$3E,$97,$C9,$FF,$A7,$F3,$B0
	.dc.b	$69,$1F,$C1,$C6,$DA,$6C,$B1,$B7,$E5,$C6,$DA,$6C,$B1,$B7,$F1,$B1
	.dc.b	$B6,$9B,$2C,$6D,$FC,$8C,$6D,$A6,$CB,$1B,$C1,$3B,$6C,$ED,$B3,$B9
	.dc.b	$D1,$27,$4D,$8D,$46,$08,$E0,$8F,$C4,$21,$F0,$0A,$7B,$0A,$7A,$0B
	.dc.b	$3E,$46,$CF,$BF,$3F,$DD,$E6,$F8,$8F,$F1,$7C,$61,$BA,$7E,$FD,$B3
	.dc.b	$BB,$FF,$19,$63,$FE,$DF,$9F,$E0,$F1,$93,$FC,$79,$19,$76,$64,$FE
	.dc.b	$85,$D3,$97,$FA,$31,$93,$F9,$7B,$97,$67,$1E,$0F,$F7,$A7,$C8,$7F
	.dc.b	$BB,$FE,$2B,$FC,$62,$D9,$FF,$77,$35,$FF,$15,$35,$FF,$7B,$70,$FE
	.dc.b	$47,$0F,$49,$23,$CE,$FF,$8F,$3C,$0F,$C6,$60,$E7,$FF,$09,$4E,$6B
	.dc.b	$F1,$D3,$9A,$DF,$9D,$FE,$68,$CB,$FC,$C3,$28,$C5,$7F,$9B,$23,$0F
	.dc.b	$95,$AA,$7C,$8F,$CC,$3F,$2D,$8F,$A5,$F3,$9C,$E9,$F8,$57,$88,$91
	.dc.b	$E6,$8C,$BF,$9F,$E1,$FF,$26,$DF,$F9,$C1,$B1,$FE,$78,$6B,$FE,$82
	.dc.b	$6B,$F2,$B2,$FD,$56,$1E,$38,$F3,$47,$A1,$1F,$DB,$BB,$3F,$61,$DB
	.dc.b	$FD,$CA,$7F,$B7,$CF,$EC,$63,$0F,$FA,$23,$ED,$1D,$5C,$BE,$2F,$FD
	.dc.b	$0B,$18,$94,$62,$75,$1C,$5A,$F9,$0F,$8E,$52,$1E,$69,$E9,$F2,$3C
	.dc.b	$E0,$D2,$3F,$E9,$CF,$DA,$3F,$CE,$B8,$FF,$A6,$1B,$19,$F2,$FF,$87
	.dc.b	$46,$23,$B9,$FD,$C4,$FB,$73,$F2,$8D,$CC,$8F,$E2,$93,$C5,$F9,$6D
	.dc.b	$E1,$FC,$67,$BB,$F9,$02,$4F,$DF,$7A,$3D,$89,$7F,$56,$78,$C6,$C9
	.dc.b	$1F,$98,$E1,$57,$F0,$9E,$9F,$F0,$14,$FF,$04,$25,$7F,$AF,$BA,$BF
	.dc.b	$86,$C7,$E0,$63,$DC,$DE,$1F,$C3,$47,$F9,$0F,$EA,$7F,$99,$78,$66
	.dc.b	$92,$3F,$31,$F5,$1F,$8A,$E8,$F2,$36,$3F,$40,$D7,$F1,$FA,$FE,$BE
	.dc.b	$31,$08,$FD,$97,$E1,$9F,$94,$FD,$E7,$DC,$D9,$FD,$53,$5E,$7B,$A3
	.dc.b	$11,$CE,$3F,$D0,$6B,$E9,$BE,$47,$99,$FA,$06,$C6,$4F,$DB,$46,$5F
	.dc.b	$DC,$7D,$75,$7B,$B5,$EC,$DC,$7F,$85,$39,$8F,$F7,$A7,$C6,$FE,$BB
	.dc.b	$E5,$3F,$DC,$6D,$8F,$F7,$43,$5F,$F7,$83,$5F,$DA,$F1,$7F,$44,$F6
	.dc.b	$3F,$9F,$92,$3E,$E7,$27,$F9,$06,$3F,$DE,$BF,$54,$F9,$9F,$DE,$C6
	.dc.b	$1B,$5F,$F0,$02,$74,$A7,$4D,$3F,$DF,$9C,$9F,$C2,$88,$FF,$1E,$7F
	.dc.b	$BB,$53,$E1,$BF,$A2,$4F,$85,$B1,$98,$FE,$94,$DF,$1F,$F3,$5E,$DF
	.dc.b	$F1,$2F,$1A,$D7,$F3,$B3,$F9,$44,$65,$A6,$4F,$C0,$FD,$7E,$5A,$64
	.dc.b	$FC,$13,$96,$99,$1A,$FE,$91,$F9,$7E,$57,$F8,$3E,$31,$1F,$C9,$8D
	.dc.b	$ED,$FA,$E7,$78,$FF,$72,$23,$83,$80,$F5,$FF,$DB,$B1,$88,$C7,$AD
	.dc.b	$F2,$27,$31,$C1,$AF,$E7,$B9,$7F,$9B,$FD,$77,$E6,$09,$1F,$98,$0F
	.dc.b	$F0,$5F,$84,$FE,$7B,$F2,$56,$4D,$23,$ED,$D1,$E7,$BE,$97,$FB,$F4
	.dc.b	$61,$B1,$99,$FF,$04,$4E,$5F,$BB,$C7,$FD,$33,$C7,$F7,$72,$7E,$05
	.dc.b	$B1,$99,$F8,$B3,$97,$E8,$F8,$9F,$C6,$1F,$A1,$C5,$F4,$9F,$D1,$F1
	.dc.b	$7D,$17,$2F,$FA,$87,$0F,$F1,$19,$23,$ED,$BF,$3F,$E2,$DE,$EF,$F6
	.dc.b	$0F,$98,$C3,$6B,$EB,$CE,$94,$E9,$A7,$C7,$D7,$E1,$5F,$DE,$D7,$E1
	.dc.b	$36,$33,$3B,$DB,$E0,$3F,$8B,$D7,$F4,$9F,$77,$F1,$20,$D2,$3C,$4C
	.dc.b	$95,$ED,$3F,$DE,$9B,$1A,$3F,$31,$97,$AF,$F5,$DF,$51,$23,$DF,$A7
	.dc.b	$F8,$0A,$3E,$84,$7C,$EC,$FF,$45,$AF,$CF,$8C,$6F,$FE,$28,$F8,$47
	.dc.b	$83,$FE,$32,$6C,$FF,$8C,$1B,$3B,$9F,$F1,$8F,$CB,$7F,$84,$3E,$2F
	.dc.b	$FA,$D3,$6F,$FC,$4C,$D7,$F9,$EB,$5F,$E7,$9A,$DF,$13,$FC,$70,$DF
	.dc.b	$1B,$E2,$DF,$F6,$58,$E1,$5F,$B8,$D3,$D6,$4C,$BD,$F8,$33,$74,$7A
	.dc.b	$C9,$7E,$B7,$FB,$75,$F1,$8F,$CB,$90,$FF,$15,$FC,$83,$F2,$C7,$81
	.dc.b	$FE,$A1,$8C,$36,$33,$3F,$EA,$39,$CB,$FE,$43,$FB,$FF,$F1,$AF,$DD
	.dc.b	$7F,$7F,$6C,$7F,$92,$DA,$FF,$62,$6B,$FD,$87,$E9,$3F,$CB,$E4,$7F
	.dc.b	$C9,$CF,$AB,$5F,$BF,$1F,$15,$F8,$7A,$2B,$EE,$12,$7E,$E3,$C5,$F8
	.dc.b	$07,$E1,$0A,$73,$2F,$67,$EE,$5C,$1F,$E5,$0D,$1F,$8F,$38,$FF,$DB
	.dc.b	$1C,$1F,$C4,$E3,$15,$FE,$D9,$8C,$4E,$A3,$94,$EB,$2F,$F0,$54,$7F
	.dc.b	$04,$BF,$AC,$D8,$F7,$E1,$E3,$27,$2F,$E7,$67,$CA,$7F,$84,$FE,$E1
	.dc.b	$F2,$DF,$E6,$6D,$3A,$CE,$63,$1A,$7F,$A1,$EB,$73,$F2,$63,$13,$A8
	.dc.b	$FF,$10,$FD,$BB,$EE,$3D,$51,$FE,$21,$8C,$3E,$F3,$71,$AA,$4E,$B3
	.dc.b	$F1,$C3,$0F,$C5,$C6,$1B,$1E,$5E,$4E,$51,$8C,$3E,$21,$F5,$B3,$F8
	.dc.b	$5A,$7F,$2F,$BE,$F9,$78,$72,$F0,$47,$FA,$95,$B7,$E6,$3D,$07,$E0
	.dc.b	$CE,$A4,$30,$43,$E7,$92,$BF,$0D,$F0,$9F,$CD,$BE,$17,$FD,$2B,$EE
	.dc.b	$B5,$FE,$E5,$39,$BF,$F4,$8C,$63,$B9,$7C,$23,$0D,$9F,$F4,$FD,$71
	.dc.b	$BE,$91,$8B,$E2,$F4,$31,$5C,$55,$71,$BF,$C7,$47,$FB,$C8,$87,$F8
	.dc.b	$28,$A6,$8C,$10,$F9,$07,$DA,$C7,$F0,$2D,$CF,$FD,$A4,$EB,$3E,$0F
	.dc.b	$F3,$53,$A9,$4F,$BC,$7F,$9B,$8F,$F3,$7E,$BF,$C3,$9F,$C2,$6D,$D1
	.dc.b	$FC,$F3,$67,$EF,$C7,$13,$F2,$95,$9B,$F2,$CF,$E0,$63,$85,$FE,$83
	.dc.b	$E5,$3F,$21,$C1,$FE,$CC,$6E,$06,$9F,$AA,$7F,$9A,$1E,$A2,$9E,$E2
	.dc.b	$9F,$24,$87,$BC,$FE,$03,$C9,$EB,$E3,$FE,$16,$DE,$4F,$E1,$33,$FB
	.dc.b	$89,$CB,$ED,$3A,$9C,$C8,$60,$9C,$4E,$34,$FF,$8E,$67,$31,$87,$FB
	.dc.b	$27,$26,$75,$FE,$98,$DE,$5F,$E5,$93,$FE,$41,$9C,$BC,$F0,$ED,$85
	.dc.b	$C3,$A4,$FC,$39,$CB,$FE,$C1,$E6,$ED,$DA,$69,$23,$B4,$78,$3F,$35
	.dc.b	$FE,$95,$06,$91,$FF,$2F,$75,$7F,$52,$7F,$CB,$2D,$73,$D7,$F8,$B6
	.dc.b	$F2,$FC,$37,$E1,$E7,$E9,$23,$FE,$07,$EC,$F8,$44,$7E,$16,$DF,$DC
	.dc.b	$1D,$9E,$FD,$7E,$0B,$F8,$0D,$BF,$BF,$AE,$BF,$EF,$CD,$3F,$CE,$D1
	.dc.b	$A8,$C9,$FE,$0E,$8D,$1F,$E0,$C2,$9F,$EC,$7E,$2F,$E1,$EF,$F8,$48
	.dc.b	$CC,$FF,$81,$7B,$FE,$D1,$23,$D2,$FC,$1E,$1F,$AB,$1E,$9C,$FD,$6F
	.dc.b	$F8,$CB,$B0,$BF,$50,$FF,$26,$11,$FF,$19,$99,$3F,$C8,$13,$FE,$3C
	.dc.b	$23,$83,$FC,$D3,$E4,$7F,$6B,$69,$F2,$0F,$F5,$91,$7C,$5F,$F5,$94
	.dc.b	$FF,$91,$BE,$3E,$4F,$F9,$35,$AF,$9C,$71,$FF,$98,$48,$F9,$9F,$F4
	.dc.b	$D6,$5F,$98,$9B,$BF,$E9,$9B,$66,$9F,$BF,$9F,$8F,$7F,$BC,$3F,$61
	.dc.b	$87,$FB,$1E,$7F,$CC,$84,$8F,$6B,$FE,$5E,$F0,$24,$7F,$99,$5A,$FC
	.dc.b	$ED,$9C,$DF,$8D,$FC,$FF,$FD,$44,$49,$FC,$3C,$65,$FF,$55,$7E,$5D
	.dc.b	$FE,$B8,$AE,$3E,$77,$D8,$8C,$52,$31,$5C,$0F,$F5,$C7,$A1,$F7,$B1
	.dc.b	$FF,$5F,$7A,$08,$C4,$E8,$FE,$9E,$7F,$B3,$63,$4E,$0F,$F3,$94,$FC
	.dc.b	$A2,$3F,$ED,$67,$E5,$6F,$F7,$E3,$F6,$27,$9B,$A7,$F1,$FF,$33,$60
	.dc.b	$D2,$3F,$EE,$C7,$FD,$E2,$47,$07,$E3,$2B,$FD,$2C,$47,$07,$FB,$7B
	.dc.b	$E0,$1E,$B7,$F0,$36,$C7,$F2,$76,$BF,$EE,$16,$BF,$EE,$6F,$61,$C5
	.dc.b	$FA,$6F,$C1,$7F,$DC,$D1,$89,$46,$27,$51,$C5,$AF,$F9,$06,$3F,$22
	.dc.b	$47,$FD,$FA,$F8,$7D,$AF,$D9,$FC,$C5,$93,$48,$FE,$1B,$B3,$F7,$3F
	.dc.b	$9F,$FE,$09,$B1,$91,$19,$69,$FE,$F5,$7F,$13,$C5,$C2,$3F,$DF,$86
	.dc.b	$8B,$1F,$8E,$F7,$FF,$83,$FE,$F1,$64,$D2,$3F,$9A,$F8,$2F,$E1,$BF
	.dc.b	$33,$F9,$66,$C6,$44,$65,$A7,$F1,$43,$F8,$AF,$03,$E1,$7F,$9E,$36
	.dc.b	$3F,$B0,$35,$E0,$70,$8F,$C0,$7A,$32,$3F,$00,$D7,$F2,$FD,$FF,$C9
	.dc.b	$1F,$EE,$1D,$9F,$B6,$F8,$6F,$EB,$64,$8F,$EB,$2F,$17,$F1,$D5,$5F
	.dc.b	$47,$2F,$F8,$E9,$FD,$AC,$F1,$69,$C8,$FE,$F1,$F0,$EF,$88,$FE,$B0
	.dc.b	$D8,$CC,$60,$EC,$FF,$38,$AE,$CF,$88,$F6,$9C,$08,$7C,$41,$3D,$5C
	.dc.b	$35,$F9,$3F,$13,$FD,$67,$C2,$E0,$D8,$CF,$93,$FD,$22,$72,$D8,$F1
	.dc.b	$1F,$AB,$CB,$CE,$B0,$F3,$BF,$76,$FF,$B3,$7C,$BF,$28,$34,$8F,$DC
	.dc.b	$F8,$C7,$C6,$FF,$7D,$6C,$7F,$88,$5A,$FB,$DA,$F9,$08,$7D,$6C,$3C
	.dc.b	$BF,$1D,$F3,$FE,$5B,$E3,$36,$7F,$16,$D7,$F2,$8D,$5C,$1F,$CF,$3E
	.dc.b	$47,$FC,$42,$48,$FF,$0F,$BF,$0B,$4F,$83,$1F,$67,$0F,$B7,$5F,$D1
	.dc.b	$23,$FB,$9E,$1F,$CF,$1A,$7F,$44,$29,$FD,$2F,$C7,$FD,$34,$7F,$45
	.dc.b	$F8,$A7,$C7,$FF,$8F,$DB,$5F,$E3,$D6,$BB,$8F,$26,$12,$35,$F2,$5D
	.dc.b	$9F,$6D,$23,$18,$A3,$12,$C3,$C5,$5C,$63,$F9,$8F,$C3,$7F,$4E,$7F
	.dc.b	$79,$D2,$8C,$47,$EF,$9A,$F0,$3F,$A8,$1F,$89,$C7,$3A,$7F,$A9,$BC
	.dc.b	$6F,$C7,$4E,$A3,$0F,$F5,$32,$1F,$EF,$18,$FE,$AE,$FF,$0D,$F9,$37
	.dc.b	$07,$F9,$1D,$B1,$FE,$6B,$6B,$FA,$96,$BF,$C3,$9A,$F3,$8E,$44,$79
	.dc.b	$35,$E4,$D7,$93,$5E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1
	.dc.b	$82,$18,$21,$FE,$44,$8F,$6B,$63,$CC,$D3,$C2,$53,$B1,$4E,$A5,$3A
	.dc.b	$88,$EA,$D8,$EA,$D3,$A9,$4F,$E3,$5F,$03,$FE,$EF,$7F,$9F,$B5,$FF
	.dc.b	$76,$9A,$F5,$3F,$12,$F7,$FB,$6F,$F2,$FE,$5F,$6D,$B1,$F6,$DA,$FE
	.dc.b	$15,$AF,$C3,$6B,$CC,$FE,$37,$82,$32,$D8,$CC,$60,$8E,$09,$3E,$36
	.dc.b	$9E,$32,$9E,$32,$9E,$32,$9E,$37,$91,$1E,$4D,$79,$35,$E4,$D7,$91
	.dc.b	$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$FF,$48,$FA,$5F,$EC
	.dc.b	$EF,$8D,$FF,$3A,$92,$3F,$BD,$C6,$5F,$EF,$BF,$A2,$3B,$FF,$D3,$35
	.dc.b	$FD,$76,$31,$28,$C4,$EA,$38,$B5,$FE,$05,$F6,$15,$FD,$03,$5F,$CB
	.dc.b	$37,$23,$9D,$D9,$7F,$AA,$BE,$33,$F8,$99,$2F,$FB,$A3,$FE,$76,$FC
	.dc.b	$27,$FD,$5F,$F2,$F6,$4D,$23,$FE,$1A,$FB,$2F,$E5,$B1,$FD,$E5,$B1
	.dc.b	$9E,$6E,$6D,$39,$9F,$ED,$0F,$B2,$7D,$AF,$E6,$BA,$D1,$89,$FF,$04
	.dc.b	$C6,$27,$53,$9C,$FF,$D1,$BF,$85,$F6,$BF,$EA,$0D,$7F,$B0,$F7,$23
	.dc.b	$18,$7F,$A7,$9D,$4B,$ED,$9F,$49,$86,$C7,$A0,$FF,$20,$38,$D7,$FC
	.dc.b	$30,$D8,$FF,$52,$FC,$5F,$AF,$13,$49,$1F,$E0,$EE,$77,$B6,$F3,$3B
	.dc.b	$4F,$E4,$23,$18,$FF,$B1,$09,$1F,$E0,$FF,$A0,$F4,$F8,$C6,$92,$39
	.dc.b	$8C,$3F,$2C,$FF,$96,$5A,$F7,$C3,$F2,$8F,$E7,$30,$F2,$75,$7F,$88
	.dc.b	$7E,$C2,$4F,$F8,$B8,$38,$7F,$CC,$9E,$43,$CB,$F9,$AD,$8F,$1B,$74
	.dc.b	$FF,$20,$36,$3A,$7C,$97,$F8,$BC,$8E,$0F,$F3,$4F,$9C,$FB,$5F,$E3
	.dc.b	$ED,$8F,$F1,$CB,$79,$BF,$C9,$33,$FC,$93,$EE,$1F,$13,$F8,$06,$DF
	.dc.b	$C6,$6E,$5F,$E5,$36,$C7,$82,$BB,$91,$EE,$FD,$FF,$7A,$47,$E6,$39
	.dc.b	$7F,$D6,$58,$FF,$B8,$A7,$F9,$AF,$DC,$BD,$EF,$C8,$B6,$FE,$BB,$72
	.dc.b	$FF,$32,$36,$3F,$DE,$35,$DA,$FA,$47,$78,$FE,$87,$EA,$3F,$5A,$78
	.dc.b	$4A,$7D,$1F,$FA,$4A,$B5,$39,$8C,$57,$0A,$D9,$FE,$9F,$AD,$B8,$3F
	.dc.b	$CA,$C7,$E3,$BC,$5F,$E1,$ED,$C7,$F5,$B1,$8C,$FF,$CF,$DD,$3F,$A0
	.dc.b	$E9,$FD,$17,$4F,$E8,$6F,$2D,$C3,$F2,$53,$F2,$FE,$2B,$D5,$F8,$A7
	.dc.b	$FA,$AB,$C2,$71,$A7,$37,$FE,$B5,$9C,$C6,$23,$FC,$A8,$D7,$F8,$2F
	.dc.b	$C4,$7C,$5F,$1B,$A5,$A3,$13,$CE,$30,$FF,$93,$0F,$F2,$F4,$69,$FF
	.dc.b	$30,$7E,$0B,$F3,$FC,$AE,$6D,$8E,$4D,$76,$7F,$38,$3F,$D9,$99,$BF
	.dc.b	$1C,$FE,$23,$F7,$7F,$EC,$88,$C7,$8F,$F6,$E3,$13,$AA,$E2,$D7,$FD
	.dc.b	$AD,$F8,$1B,$E8,$3F,$05,$F9,$DF,$FA,$BE,$31,$28,$C4,$EA,$3F,$DA
	.dc.b	$F3,$A9,$FA,$B8,$BF,$3A,$A4,$AF,$AE,$FF,$AE,$F9,$DF,$81,$7E,$46
	.dc.b	$30,$D9,$FF,$71,$CE,$9E,$86,$23,$4F,$FB,$63,$D8,$FC,$B3,$C1,$AE
	.dc.b	$4F,$CA,$1F,$ED,$17,$2F,$F1,$DE,$7F,$49,$1F,$C6,$BF,$80,$F2,$3F
	.dc.b	$4D,$FA,$0D,$93,$48,$FE,$37,$F2,$FF,$F5,$6A,$FC,$EF,$F2,$91,$86
	.dc.b	$F4,$FE,$C5,$B3,$B3,$F6,$87,$ED,$7F,$02,$7E,$29,$FB,$7F,$CC,$FF
	.dc.b	$DC,$91,$89,$46,$27,$51,$C8,$4E,$E7,$FA,$3C,$7F,$3F,$EE,$FE,$96
	.dc.b	$FE,$B3,$C0,$CB,$F1,$4F,$86,$43,$87,$17,$F7,$2C,$DF,$85,$7F,$0B
	.dc.b	$FC,$47,$DE,$DA,$F8,$8D,$7F,$88,$6A,$AE,$30,$AF,$CC,$F5,$7F,$15
	.dc.b	$FC,$8E,$C9,$A4,$7D,$DF,$33,$FB,$2D,$7E,$04,$23,$0E,$BB,$FE,$2B
	.dc.b	$F3,$A3,$2F,$B3,$93,$F0,$4D,$3E,$01,$F4,$F8,$7F,$94,$7E,$65,$F3
	.dc.b	$E0,$D2,$3F,$C5,$BC,$8F,$D2,$91,$C1,$D0,$FE,$E5,$C5,$FC,$75,$F0
	.dc.b	$10,$FE,$6D,$F5,$AF,$BA,$FE,$71,$F9,$8F,$BA,$31,$28,$C4,$EA,$3F
	.dc.b	$05,$3A,$9F,$6C,$7D,$D3,$B1,$E2,$E4,$FF,$16,$FA,$4F,$97,$FE,$8D
	.dc.b	$F9,$8F,$F2,$D8,$C5,$23,$13,$A8,$FE,$5D,$3A,$9E,$F1,$D7,$6B,$E6
	.dc.b	$7F,$3F,$6E,$9F,$06,$7C,$84,$38,$1F,$1E,$78,$3F,$53,$9F,$F1,$5E
	.dc.b	$C7,$F5,$BF,$CC,$7F,$5E,$D9,$FD,$A3,$5F,$C8,$F4,$FF,$84,$67,$F6
	.dc.b	$BF,$19,$99,$5F,$E5,$63,$0D,$84,$F0,$69,$E5,$C7,$D4,$FA,$0F,$9B
	.dc.b	$8F,$A1,$B1,$A8,$C4,$7B,$DA,$F1,$31,$1B,$26,$DF,$D9,$8F,$D9,$3F
	.dc.b	$DE,$FF,$09,$FE,$1E,$D9,$FE,$28,$D7,$E0,$B5,$F7,$E4,$F8,$4F,$F3
	.dc.b	$83,$63,$F7,$D1,$83,$FD,$97,$87,$98,$95,$E4,$D7,$ED,$E3,$E5,$7F
	.dc.b	$1F,$9F,$EF,$FD,$7F,$F0,$87,$C8,$7F,$26,$CB,$C3,$18,$9F,$E5,$4D
	.dc.b	$8F,$AE,$D7,$E1,$DF,$57,$E8,$35,$F9,$AF,$F1,$58,$FE,$F5,$3D,$25
	.dc.b	$E8,$F8,$63,$C9,$18,$75,$1E,$58,$FA,$44,$CB,$DF,$7F,$E9,$2F,$C0
	.dc.b	$DF,$4F,$FE,$2B,$F9,$6E,$91,$8B,$46,$27,$58,$FF,$84,$EB,$53,$E3
	.dc.b	$FB,$07,$D8,$7F,$C0,$0A,$FF,$0C,$46,$30,$FF,$5C,$C6,$27,$51,$FC
	.dc.b	$B2,$75,$B7,$B7,$87,$FA,$33,$7B,$B7,$5F,$80,$6E,$1F,$CC,$DC,$C7
	.dc.b	$BF,$87,$EE,$3F,$AF,$C3,$AD,$78,$63,$C6,$78,$71,$E7,$3A,$8F,$A7
	.dc.b	$87,$E1,$08,$FD,$87,$F9,$BF,$C8,$BE,$45,$F2,$5F,$E5,$5E,$9F,$F2
	.dc.b	$4D,$AB,$53,$9B,$FE,$CF,$18,$9F,$F4,$74,$62,$F9,$C6,$0F,$F3,$04
	.dc.b	$EE,$FA,$B6,$B8,$5F,$5A,$DD,$F2,$BF,$1F,$17,$F9,$5E,$7F,$C4,$24
	.dc.b	$3F,$CA,$A5,$3F,$B3,$94,$FF,$56,$1F,$E5,$7C,$7E,$E3,$F8,$1C,$BF
	.dc.b	$CA,$CF,$CB,$3F,$B3,$94,$FC,$A4,$F8,$5E,$E4,$3B,$9F,$E1,$3D,$79
	.dc.b	$BF,$7C,$E6,$53,$F0,$86,$34,$5F,$91,$FB,$D1,$E5,$35,$19,$7F,$BD
	.dc.b	$F0,$7E,$7E,$7F,$3D,$1F,$CA,$9F,$CF,$70,$FF,$80,$1B,$19,$3C,$E4
	.dc.b	$3E,$29,$EB,$3F,$82,$8F,$C3,$7F,$39,$CB,$FE,$99,$6C,$66,$7F,$7D
	.dc.b	$1F,$1B,$87,$FD,$D5,$3F,$44,$FE,$BC,$5C,$FE,$53,$E1,$32,$78,$6B
	.dc.b	$F8,$2F,$CD,$F6,$23,$FC,$2D,$ED,$CB,$FE,$10,$9E,$27,$3A,$EA,$FB
	.dc.b	$0C,$5F,$91,$ED,$C9,$FC,$BD,$FF,$23,$F4,$73,$C7,$FC,$4C,$D9,$D4
	.dc.b	$FF,$3C,$1F,$55,$FD,$01,$A6,$5C,$1E,$FB,$F0,$47,$8B,$1F,$46,$3E
	.dc.b	$68,$F4,$9E,$72,$18,$23,$B2,$4F,$48,$DC,$62,$7F,$C6,$78,$FF,$69
	.dc.b	$FC,$C7,$FB,$41,$2B,$FC,$D1,$DB,$FE,$D3,$8F,$EF,$30,$69,$1F,$EF
	.dc.b	$66,$4F,$C3,$1E,$4C,$3C,$4D,$BF,$6C,$63,$0F,$24,$FC,$39,$C9,$1F
	.dc.b	$F0,$D8,$FB,$0A,$47,$FC,$3B,$E0,$FF,$BA,$63,$94,$43,$FC,$5F,$E2
	.dc.b	$7F,$BB,$7F,$31,$FF,$19,$C1,$A4,$7D,$38,$79,$1B,$8B,$17,$CE,$FF
	.dc.b	$AB,$89,$3F,$EE,$27,$FC,$A5,$F9,$7C,$8F,$F4,$07,$A5,$FE,$85,$8C
	.dc.b	$47,$DF,$3C,$C7,$F9,$87,$CC,$2F,$CB,$BF,$CD,$3E,$6F,$FB,$71,$B3
	.dc.b	$A6,$CE,$CF,$F6,$6F,$A1,$CB,$F2,$CF,$F3,$B9,$23,$21,$FF,$42,$AF
	.dc.b	$CF,$29,$1F,$F4,$39,$93,$E1,$E9,$F5,$5B,$F3,$FF,$ED,$02,$3F,$E9
	.dc.b	$FD,$DB,$FC,$99,$FF,$3B,$65,$2B,$E2,$F8,$1E,$B6,$9E,$B3,$FC,$8B
	.dc.b	$CF,$F1,$3F,$3C,$FF,$5A,$92,$3F,$D7,$D1,$97,$FD,$6E,$BE,$19,$ED
	.dc.b	$FC,$CF,$D0,$3F,$D9,$04,$AD,$CE,$94,$7D,$47,$F5,$39,$FE,$6B,$DF
	.dc.b	$7B,$DF,$5B,$9F,$FD,$D4,$DA,$FF,$70,$1F,$86,$2C,$7F,$BB,$1F,$F5
	.dc.b	$11,$FE,$E9,$29,$FE,$CE,$9F,$F7,$77,$2E,$CF,$F7,$D9,$23,$E6,$5F
	.dc.b	$E4,$E3,$11,$FE,$D6,$F8,$27,$C1,$FE,$2B,$F2,$CF,$F5,$8C,$62,$51
	.dc.b	$89,$D4,$7F,$B5,$67,$53,$FE,$BE,$FC,$36,$4F,$E2,$1A,$FF,$B2,$35
	.dc.b	$7C,$47,$E6,$3E,$83,$FD,$BD,$C3,$F4,$63,$15,$F3,$9A,$F1,$6B,$FE
	.dc.b	$7E,$A7,$5F,$F4,$B8,$FA,$AD,$7C,$0D,$78,$36,$BF,$01,$F1,$7F,$AA
	.dc.b	$FC,$97,$C7,$68,$C5,$FE,$01,$AF,$AD,$AF,$E0,$3A,$BF,$3C,$FF,$53
	.dc.b	$E1,$7A,$7F,$B6,$EA,$FF,$6E,$C6,$25,$18,$9D,$47,$FC,$01,$9D,$4F
	.dc.b	$EF,$A3,$F7,$FF,$1C,$F8,$F7,$C8,$7E,$2E,$73,$18,$7F,$18,$62,$34
	.dc.b	$A7,$94,$63,$5F,$D1,$7C,$8F,$F5,$9D,$4D,$25,$FF,$67,$F9,$27,$A3
	.dc.b	$E8,$8F,$34,$61,$F6,$9E,$90,$E7,$3F,$F7,$FB,$63,$F5,$B5,$D9,$FA
	.dc.b	$6D,$79,$7C,$AB,$E5,$3F,$BC,$42,$30,$FF,$7B,$6B,$E1,$35,$1C,$3E
	.dc.b	$55,$F0,$7E,$99,$FE,$57,$18,$7D,$27,$D1,$71,$8F,$36,$C7,$92,$BC
	.dc.b	$77,$E0,$8C,$1E,$38,$F3,$FD,$3B,$DB,$FE,$73,$C9,$E5,$8C,$57,$F7
	.dc.b	$36,$C6,$67,$73,$F2,$FE,$63,$79,$7B,$5B,$19,$9F,$62,$8C,$BF,$8B
	.dc.b	$BE,$FE,$7F,$58,$34,$8F,$F0,$AC,$3F,$CE,$6F,$E4,$FC,$9F,$E6,$F1
	.dc.b	$88,$FE,$A7,$18,$F7,$BE,$83,$63,$F0,$5F,$36,$F9,$BF,$2B,$63,$E7
	.dc.b	$9F,$D7,$1C,$69,$F6,$5B,$1E,$38,$FE,$92,$F8,$FC,$5F,$E9,$0E,$4F
	.dc.b	$E6,$FF,$39,$F9,$B2,$3F,$E1,$CE,$64,$7E,$2B,$E7,$5E,$8F,$C8,$6D
	.dc.b	$79,$03,$8C,$3F,$18,$D8,$FE,$8B,$F3,$DF,$8B,$EA,$FE,$98,$DC,$4D
	.dc.b	$35,$F3,$EF,$A0,$FF,$24,$36,$FF,$C6,$2D,$7F,$93,$B5,$FD,$67,$BD
	.dc.b	$FE,$E4,$7C,$CD,$CB,$F4,$0D,$8D,$FC,$47,$8F,$BE,$47,$D0,$CF,$FB
	.dc.b	$14,$62,$FF,$8A,$4F,$F0,$73,$F8,$A7,$87,$FC,$49,$A6,$4F,$EA,$8E
	.dc.b	$5F,$E3,$1F,$44,$F9,$EF,$F0,$BF,$37,$EC,$E3,$13,$F8,$90,$E2,$35
	.dc.b	$3C,$27,$F8,$57,$D1,$BC,$BC,$47,$F2,$58,$C3,$CD,$B5,$FC,$05,$B1
	.dc.b	$F9,$B8,$FC,$AF,$2F,$F8,$DD,$BD,$3E,$3E,$9F,$CC,$57,$AB,$E9,$3F
	.dc.b	$87,$E6,$69,$23,$FD,$15,$19,$6C,$65,$F8,$0D,$7F,$92,$46,$DC,$9C
	.dc.b	$08,$60,$87,$08,$E4,$F9,$4D,$99,$29,$93,$89,$4D,$18,$3F,$C9,$71
	.dc.b	$E3,$72,$43,$81,$0D,$C7,$89,$A7,$80,$A7,$53,$EE,$BB,$69,$B2,$9B
	.dc.b	$8F,$13,$FE,$FC,$3F,$9D,$7C,$5B,$E2,$FF,$11,$08,$C6,$3F,$EB,$B6
	.dc.b	$C7,$E3,$FE,$73,$FC,$CF,$BD,$8F,$EE,$47,$E5,$A3,$2F,$D4,$6B,$93
	.dc.b	$67,$C6,$72,$D3,$27,$E6,$9C,$B4,$CE,$08,$CB,$63,$31,$82,$38,$24
	.dc.b	$FA,$48,$F9,$CC,$95,$CB,$E9,$6B,$E7,$6B,$E5,$6C,$F8,$DB,$3E,$26
	.dc.b	$CF,$63,$A1,$DC,$86,$08,$7F,$44,$D3,$FD,$49,$F4,$7F,$AE,$24,$7F
	.dc.b	$4D,$7F,$87,$91,$FE,$1A,$D7,$F8,$5F,$07,$FA,$82,$3F,$82,$C1,$A4
	.dc.b	$7F,$AC,$7D,$27,$EF,$08,$E0,$FF,$0A,$7C,$33,$E9,$7F,$D5,$73,$FE
	.dc.b	$07,$8C,$4A,$31,$3A,$8E,$2D,$7F,$D6,$7C,$CF,$89,$7C,$5F,$FA,$DB
	.dc.b	$E7,$7F,$CB,$F1,$8A,$46,$27,$51,$FE,$94,$9D,$4F,$F1,$4D,$64,$FF
	.dc.b	$73,$6B,$FE,$45,$FC,$29,$9F,$E4,$1B,$1F,$94,$6C,$FF,$88,$DB,$1D
	.dc.b	$FE,$37,$C2,$FA,$B0,$46,$21,$97,$B6,$30,$D8,$F5,$9E,$52,$1F,$DD
	.dc.b	$79,$FF,$4B,$CF,$FE,$08,$C7,$C5,$C3,$E7,$25,$79,$DF,$D4,$FA,$9F
	.dc.b	$E6,$7D,$7F,$A9,$92,$7F,$A8,$BF,$09,$FC,$76,$1D,$AB,$FD,$E5,$79
	.dc.b	$95,$FF,$9C,$78,$7F,$19,$1F,$EE,$A8,$C3,$F9,$03,$FC,$B1,$FB,$0F
	.dc.b	$F0,$F7,$FD,$80,$7F,$9C,$DC,$7C,$17,$F9,$35,$B1,$C3,$A7,$FD,$EB
	.dc.b	$3E,$1D,$3E,$04,$FC,$43,$CF,$8F,$1A,$FF,$7D,$FD,$3D,$93,$48,$FF
	.dc.b	$9A,$FE,$34,$7F,$0E,$2B,$FC,$49,$B5,$FE,$61,$6C,$78,$45,$F9,$1F
	.dc.b	$AD,$3F,$1E,$77,$1F,$E9,$16,$A3,$C2,$F8,$4F,$17,$B9,$23,$C4,$D9
	.dc.b	$F8,$5A,$FD,$49,$F1,$5F,$85,$E9,$7F,$EC,$67,$89,$2B,$D5,$7F,$CB
	.dc.b	$3F,$0F,$7C,$87,$E5,$69,$18,$7F,$26,$7F,$AA,$5C,$70,$FF,$A0,$DB
	.dc.b	$1E,$2C,$BC,$8F,$C2,$3E,$36,$FF,$EB,$7C,$7F,$C4,$64,$8F,$F1,$0B
	.dc.b	$FB,$C3,$E2,$78,$96,$3F,$C6,$87,$EA,$E3,$2F,$EF,$5A,$FE,$BF,$E5
	.dc.b	$3E,$67,$E2,$DF,$C1,$9B,$1F,$C0,$DA,$EC,$47,$F0,$CE,$DF,$A9,$F5
	.dc.b	$16,$4D,$23,$FE,$CB,$E6,$7C,$AB,$CD,$E3,$39,$8C,$3C,$8C,$46,$9F
	.dc.b	$F4,$CC,$6A,$BF,$D5,$DF,$84,$BE,$FB,$F0,$DF,$53,$86,$DF,$FB,$0E
	.dc.b	$74,$A7,$4D,$3F,$C9,$BF,$53,$FE,$52,$23,$83,$FD,$CD,$F0,$3F,$93
	.dc.b	$FC,$FB,$FD,$DA,$D8,$C8,$8C,$B4,$CF,$2E,$08,$E0,$FB,$A4,$3F,$CB
	.dc.b	$A7,$F7,$EE,$3F,$C2,$3F,$2F,$27,$C7,$7E,$97,$C8,$FF,$70,$70,$7C
	.dc.b	$38,$C5,$6D,$AF,$FB,$0A,$B8,$BF,$E0,$33,$F9,$97,$AD,$FA,$FF,$CF
	.dc.b	$BF,$A4,$B6,$3F,$A3,$35,$D8,$8F,$CF,$75,$7E,$DF,$E9,$2C,$9A,$47
	.dc.b	$F4,$58,$FF,$29,$27,$47,$E9,$FE,$EB,$FB,$84,$FF,$BE,$FB,$11,$88
	.dc.b	$FD,$73,$5F,$E0,$D1,$A7,$FB,$B1,$FD,$77,$A3,$F8,$57,$D4,$7E,$5C
	.dc.b	$91,$90,$FF,$68,$FC,$3B,$F8,$87,$E5,$56,$4D,$23,$F8,$DF,$57,$F8
	.dc.b	$27,$4F,$FA,$13,$97,$F2,$B1,$83,$F8,$47,$EA,$9F,$DF,$BC,$9F,$E3
	.dc.b	$97,$23,$0E,$03,$87,$5E,$87,$BD,$FE,$23,$C5,$FD,$01,$F8,$CD,$7F
	.dc.b	$89,$35,$F8,$AD,$7F,$99,$F8,$7F,$C4,$9F,$E8,$A7,$D4,$E1,$FA,$4F
	.dc.b	$F8,$4C,$FA,$5C,$3F,$E1,$17,$E7,$B5,$FE,$46,$DD,$3F,$D4,$BC,$7F
	.dc.b	$0A,$7F,$04,$7C,$72,$9F,$0A,$36,$F6,$1F,$85,$78,$88,$E5,$F8,$CB
	.dc.b	$F6,$66,$3F,$D1,$FE,$B7,$31,$89,$46,$27,$51,$F3,$E7,$53,$F0,$3F
	.dc.b	$2D,$93,$E3,$6B,$FC,$DC,$87,$C8,$2F,$9E,$FC,$25,$F6,$FF,$E7,$88
	.dc.b	$C3,$63,$33,$FE,$7E,$9C,$B7,$D3,$FD,$AF,$D6,$7F,$8E,$BC,$0F,$EC
	.dc.b	$4D,$93,$4F,$F1,$FF,$C1,$7F,$67,$F8,$5B,$26,$91,$FB,$C4,$3F,$C9
	.dc.b	$78,$BE,$EF,$CC,$48,$F3,$46,$5F,$DB,$9C,$4F,$84,$3F,$04,$FF,$54
	.dc.b	$7C,$07,$FA,$61,$B1,$FE,$9C,$6F,$37,$E7,$DB,$1F,$E5,$3F,$C5,$3F
	.dc.b	$C0,$1F,$81,$7F,$96,$9B,$19,$11,$96,$9F,$E5,$BB,$FA,$1B,$7F,$B9
	.dc.b	$70,$FF,$31,$C6,$0F,$F5,$3F,$AD,$FE,$03,$AF,$D1,$F8,$BE,$C7,$ED
	.dc.b	$E6,$FC,$43,$EC,$12,$3E,$1B,$F6,$F6,$FF,$0B,$7D,$8F,$F9,$62,$7E
	.dc.b	$5C,$63,$8F,$F7,$E7,$E1,$5C,$46,$84,$F1,$7E,$59,$E4,$CF,$FC,$37
	.dc.b	$8F,$F9,$73,$E0,$3F,$D9,$B8,$7F,$07,$CF,$FC,$FD,$5C,$6F,$FC,$5F
	.dc.b	$5A,$7F,$87,$CE,$AB,$C7,$9F,$E1,$65,$EC,$78,$67,$CD,$18,$77,$9F
	.dc.b	$F8,$CE,$75,$5C,$5F,$05,$FC,$B2,$75,$7F,$0D,$C3,$D4,$72,$1F,$85
	.dc.b	$72,$C3,$FC,$7C,$D8,$CE,$3B,$8F,$27,$07,$FB,$1E,$3F,$D3,$1E,$77
	.dc.b	$B6,$BF,$D0,$F3,$97,$FD,$19,$19,$9F,$F5,$33,$67,$53,$C2,$30,$73
	.dc.b	$7C,$E7,$EF,$8A,$7E,$7A,$79,$AD,$7C,$B3,$CA,$31,$3E,$66,$9C,$F2
	.dc.b	$F8,$BB,$7E,$3E,$3E,$FC,$62,$FF,$D1,$5F,$08,$F7,$5F,$0B,$FE,$63
	.dc.b	$F9,$2F,$06,$1F,$E6,$5A,$D4,$E6,$FF,$A6,$C6,$25,$5C,$14,$66,$30
	.dc.b	$D9,$E3,$7E,$0A,$E3,$18,$BE,$D1,$E1,$9D,$5F,$5C,$3F,$99,$E3,$FC
	.dc.b	$DA,$BF,$5A,$7F,$38,$8C,$B7,$74,$EE,$73,$3F,$9D,$9C,$BB,$9D,$46
	.dc.b	$23,$E7,$1F,$48,$FF,$72,$9F,$82,$29,$F7,$CA,$79,$0E,$D1,$D0,$8F
	.dc.b	$FB,$AB,$4F,$DC,$BC,$CE,$C4,$D3,$C2,$47,$4D,$7F,$DB,$7C,$3F,$AF
	.dc.b	$FC,$2F,$F3,$C4,$9F,$D2,$3F,$AA,$C3,$D9,$5F,$ED,$B3,$E0,$10,$FD
	.dc.b	$56,$7F,$A7,$25,$60,$FE,$09,$C7,$E8,$9F,$F5,$56,$68,$C4,$7F,$06
	.dc.b	$3E,$1C,$65,$B7,$E2,$7F,$DB,$E7,$84,$F2,$1F,$BB,$F1,$BF,$70,$4B
	.dc.b	$C1,$FC,$3B,$9B,$F0,$4F,$F6,$56,$E7,$FC,$B6,$BF,$BD,$CE,$5F,$A2
	.dc.b	$7D,$43,$C0,$78,$8F,$E7,$C7,$0E,$27,$3F,$66,$C6,$63,$1D,$3F,$EC
	.dc.b	$BF,$DB,$FF,$A4,$12,$7F,$A6,$3F,$D4,$F6,$E3,$16,$BF,$35,$D7,$FD
	.dc.b	$6A,$30,$79,$8F,$86,$43,$CC,$7F,$6C,$EC,$F7,$91,$F7,$E7,$FD,$AD
	.dc.b	$F6,$9F,$93,$39,$14,$FE,$3B,$DB,$FE,$E3,$F9,$4F,$EF,$A4,$8F,$E9
	.dc.b	$BA,$27,$51,$8F,$03,$FD,$D5,$F6,$1F,$E0,$92,$46,$43,$FE,$51,$DF
	.dc.b	$FC,$27,$17,$1F,$F0,$BB,$7C,$2F,$B2,$7E,$F5,$F3,$13,$93,$F7,$2F
	.dc.b	$AB,$D3,$FF,$3A,$7E,$3F,$FE,$4C,$24,$7F,$8F,$DF,$F1,$A2,$FA,$BA
	.dc.b	$47,$FC,$6D,$CE,$74,$7F,$91,$F8,$7F,$CA,$E4,$70,$7F,$93,$FE,$A3
	.dc.b	$FC,$A4,$47,$07,$F9,$3F,$E7,$FF,$CB,$E4,$70,$7E,$44,$87,$D4,$CB
	.dc.b	$DA,$FF,$9B,$07,$CC,$7F,$A8,$08,$FF,$9D,$0C,$9F,$E7,$DF,$45,$E8
	.dc.b	$FA,$FC,$2F,$F2,$DC,$62,$7F,$98,$1F,$E8,$02,$9F,$05,$D4,$7F,$9B
	.dc.b	$7E,$46,$4F,$F9,$B5,$AF,$C7,$E2,$73,$3D,$7F,$95,$F9,$5F,$FA,$74
	.dc.b	$95,$F8,$A8,$CB,$FE,$A5,$FA,$EF,$F7,$F1,$1C,$1F,$E7,$7C,$7F,$D0
	.dc.b	$E4,$70,$7C,$83,$F4,$64,$3F,$0F,$C8,$F0,$7E,$E7,$0F,$FB,$20,$93
	.dc.b	$F2,$5F,$F6,$9A,$FD,$02,$91,$FF,$6A,$99,$3F,$DC,$03,$ED,$3F,$DC
	.dc.b	$BF,$01,$FE,$E0,$6F,$67,$E2,$DB,$3B,$3F,$D6,$5F,$03,$FD,$9D,$BE
	.dc.b	$5F,$E7,$E3,$11,$FE,$F6,$3F,$DF,$FC,$51,$E3,$F0,$7F,$97,$8C,$47
	.dc.b	$FB,$02,$53,$F9,$B7,$FD,$36,$62,$0E,$23,$51,$D0,$C9,$E4,$1F,$59
	.dc.b	$CA,$FF,$4C,$D8,$C9,$B7,$FD,$53,$EC,$B3,$FD,$67,$D5,$61,$B5,$FE
	.dc.b	$D3,$9D,$29,$D3,$4F,$F6,$DF,$9C,$F8,$97,$C5,$7E,$DF,$EA,$BF,$DB
	.dc.b	$93,$98,$C5,$70,$AD,$AA,$DC,$61,$FE,$29,$D2,$F8,$5E,$35,$F9,$D8
	.dc.b	$C3,$C4,$FF,$7E,$4F,$09,$DC,$7F,$91,$BE,$37,$F8,$57,$E3,$DD,$1B
	.dc.b	$1F,$CA,$9A,$F5,$6B,$FE,$4E,$DF,$F8,$97,$A5,$E4,$6C,$7F,$32,$6B
	.dc.b	$E5,$6B,$F7,$DF,$C5,$11,$FF,$13,$71,$7E,$31,$FD,$37,$85,$FC,$39
	.dc.b	$FE,$59,$E2,$FF,$73,$7F,$E0,$08,$E5,$FF,$71,$F5,$FE,$AE,$30,$7F
	.dc.b	$11,$8C,$F5,$FE,$4E,$3F,$95,$1F,$CA,$FE,$43,$FD,$02,$FF,$8D,$DA
	.dc.b	$FF,$BB,$9A,$F0,$0E,$1F,$E4,$97,$E5,$C3,$D0,$7A,$67,$53,$DF,$DA
	.dc.b	$FE,$91,$18,$8E,$71,$88,$91,$E3,$6B,$FD,$9A,$7A,$D7,$73,$E2,$4F
	.dc.b	$92,$78,$10,$F5,$3F,$D5,$FE,$4F,$FC,$D9,$1F,$E1,$D6,$F2,$FB,$61
	.dc.b	$18,$7E,$03,$5F,$7C,$7F,$80,$5F,$E3,$91,$D3,$C6,$E4,$D8,$E3,$E9
	.dc.b	$7F,$59,$9C,$B6,$3F,$75,$E0,$FE,$DC,$8E,$0F,$E5,$47,$E0,$AF,$C0
	.dc.b	$FF,$06,$E0,$FB,$8F,$F0,$D6,$BF,$8F,$3F,$06,$64,$A7,$F6,$BF,$95
	.dc.b	$67,$FD,$ED,$B1,$FE,$79,$6B,$F1,$5A,$F9,$34,$FA,$77,$F0,$FC,$6F
	.dc.b	$E9,$FF,$2B,$E2,$6D,$78,$1B,$C3,$FB,$C9,$F3,$CB,$07,$CB,$78,$C9
	.dc.b	$1E,$37,$FC,$21,$8F,$82,$3F,$09,$97,$F9,$5F,$1F,$EB,$FC,$BE,$1E
	.dc.b	$AF,$F2,$E6,$7F,$C2,$A3,$07,$83,$1F,$EE,$1E,$5F,$E0,$5B,$8F,$E7
	.dc.b	$DA,$FE,$A1,$AF,$F9,$B3,$1F,$F4,$CC,$FF,$A5,$49,$1F,$E8,$C7,$F1
	.dc.b	$9F,$31,$FC,$5B,$82,$F5,$FF,$80,$6B,$F5,$93,$96,$C6,$5F,$B1,$3F
	.dc.b	$84,$69,$F8,$53,$FD,$FB,$7F,$EA,$A8,$FF,$53,$12,$3F,$D3,$AF,$8B
	.dc.b	$DB,$FF,$1C,$59,$A4,$9F,$F5,$C4,$65,$B1,$97,$91,$82,$18,$21,$EC
	.dc.b	$78,$34,$D9,$4D,$9F,$C5,$23,$C0,$E0,$FF,$20,$38,$78,$10,$E1,$1D
	.dc.b	$9A,$76,$29,$CC,$47,$37,$27,$33,$BC,$73,$69,$CC,$A7,$38,$E4,$E0
	.dc.b	$FE,$4D,$EE,$FF,$BC,$5B,$C9,$FE,$3A,$7F,$DD,$17,$FE,$E5,$9D,$47
	.dc.b	$F3,$4F,$DF,$9F,$8D,$FF,$97,$F9,$BF,$8A,$46,$27,$FC,$22,$D8,$CC
	.dc.b	$EE,$7F,$9B,$4E,$5B,$19,$7B,$35,$FF,$36,$46,$DC,$9D,$A3,$4A,$34
	.dc.b	$F0,$8C,$C6,$23,$51,$82,$3C,$C9,$3E,$86,$9E,$62,$9E,$27,$44,$74
	.dc.b	$D7,$46,$08,$60,$87,$9C,$EA,$57,$93,$FB,$BE,$67,$C5,$FF,$43,$E5
	.dc.b	$FE,$4B,$18,$9F,$EC,$51,$89,$D6,$CA,$D5,$7E,$D1,$FF,$23,$D7,$2B
	.dc.b	$DC,$62,$B8,$47,$F5,$1F,$C9,$3F,$AD,$37,$AF,$E8,$C6,$30,$FE,$E4
	.dc.b	$D7,$FC,$45,$F0,$9F,$EA,$5F,$9A,$FF,$13,$B6,$3F,$D8,$AD,$76,$23
	.dc.b	$C7,$72,$3D,$B8,$79,$1B,$8F,$F9,$8F,$A5,$F9,$6F,$BE,$5F,$19,$FE
	.dc.b	$C9,$8C,$36,$BF,$D7,$0D,$9D,$9F,$E1,$2A,$E7,$F0,$9F,$93,$6C,$77
	.dc.b	$B4,$E5,$B1,$F2,$6E,$63,$E9,$8F,$A0,$F8,$FE,$C7,$FD,$91,$F2,$36
	.dc.b	$4D,$23,$FE,$1D,$FA,$97,$D5,$7F,$B3,$F8,$FF,$C0,$91,$89,$FB,$A7
	.dc.b	$FB,$4D,$C4,$6A,$78,$4F,$C8,$C3,$C8,$FE,$35,$AA,$BA,$5F,$18,$F2
	.dc.b	$9E,$3A,$F4,$B4,$C9,$FE,$3A,$FC,$EF,$ED,$38,$3F,$DC,$CE,$27,$FC
	.dc.b	$38,$D8,$FF,$1D,$35,$F8,$E7,$9C,$86,$0F,$F2,$3F,$2B,$CA,$FE,$06
	.dc.b	$D8,$FF,$51,$B7,$1F,$F1,$53,$63,$EF,$6A,$FC,$27,$FD,$F5,$E4,$7E
	.dc.b	$1D,$B3,$A6,$BF,$ED,$99,$E0,$FF,$93,$FA,$3F,$54,$F9,$8F,$98,$5E
	.dc.b	$2F,$CA,$FE,$58,$87,$E5,$9F,$95,$8C,$36,$32,$79,$A3,$2F,$A3,$E1
	.dc.b	$FE,$8B,$E3,$6A,$AF,$D1,$C6,$1B,$1F,$4C,$5F,$A1,$F9,$87,$A0,$A6
	.dc.b	$B9,$FF,$60,$E2,$38,$3B,$3C,$18,$FF,$43,$E0,$FD,$86,$BE,$18,$C4
	.dc.b	$FD,$08,$F3,$BD,$79,$7F,$4E,$FC,$93,$1B,$FF,$8F,$71,$FE,$7F,$E4
	.dc.b	$7F,$90,$A3,$13,$FE,$E0,$8C,$3F,$73,$F2,$0F,$CF,$BF,$5D,$AF,$99
	.dc.b	$BA,$7F,$08,$73,$1E,$03,$C5,$A7,$B9,$C1,$FE,$96,$FB,$77,$47,$F5
	.dc.b	$26,$C7,$FB,$C9,$B8,$FF,$9E,$5B,$1F,$90,$FA,$DF,$C1,$7B,$9F,$C5
	.dc.b	$5B,$19,$11,$96,$9F,$98,$F1,$3F,$26,$FD,$06,$E7,$FC,$89,$BC,$1F
	.dc.b	$E8,$CE,$6F,$E4,$5E,$77,$E4,$7F,$1F,$B8,$1A,$47,$F4,$BF,$80,$FE
	.dc.b	$5F,$E9,$3F,$95,$36,$32,$23,$2D,$3F,$5B,$DB,$F3,$B1,$FE,$27,$18
	.dc.b	$3F,$8B,$E7,$FC,$E2,$7E,$8F,$0F,$CF,$27,$27,$E4,$2B,$F1,$D5,$FC
	.dc.b	$FB,$F1,$BE,$84,$8E,$78,$23,$ED,$FC,$03,$ED,$3F,$C0,$ED,$A7,$27
	.dc.b	$FA,$B6,$30,$D9,$FB,$9C,$B2,$BE,$E3,$67,$FA,$67,$D9,$FE,$EF,$F2
	.dc.b	$2B,$46,$91,$FE,$1F,$EF,$7F,$01,$FC,$3F,$FA,$A3,$63,$22,$32,$D3
	.dc.b	$F8,$BF,$E4,$9F,$D6,$1A,$7F,$58,$3F,$8A,$39,$69,$93,$F6,$C7,$E6
	.dc.b	$8B,$C1,$F9,$DE,$AF,$CA,$C7,$E7,$73,$F7,$3F,$EE,$D8,$CB,$FB,$EE
	.dc.b	$6F,$E3,$5F,$90,$7E,$28,$91,$FC,$B2,$32,$FF,$2E,$F8,$DF,$E4,$1F
	.dc.b	$4B,$FD,$CD,$B1,$91,$19,$69,$FE,$FD,$E1,$FA,$AD,$E1,$FB,$64,$7F
	.dc.b	$BA,$7E,$4B,$F8,$F2,$3F,$84,$7F,$25,$F5,$DF,$CC,$7F,$29,$FF,$02
	.dc.b	$36,$7F,$A1,$B5,$FE,$7C,$D7,$E3,$F4,$7E,$4D,$F9,$47,$CB,$E8,$FF
	.dc.b	$08,$FE,$13,$F9,$F2,$47,$E5,$1F,$C9,$E3,$E7,$E7,$FC,$74,$FE,$48
	.dc.b	$F6,$1E,$A1,$F6,$5F,$E1,$9C,$FF,$A4,$36,$32,$7F,$5A,$8C,$BF,$E1
	.dc.b	$9F,$DB,$7F,$A7,$3F,$AE,$6F,$A1,$88,$FE,$8D,$18,$78,$1F,$C9,$17
	.dc.b	$CC,$7F,$57,$FC,$CF,$86,$89,$CD,$27,$31,$86,$D7,$19,$FE,$57,$AF
	.dc.b	$C9,$E7,$FD,$EF,$93,$FE,$39,$C1,$F5,$9F,$E2,$F6,$DE,$63,$06,$7F
	.dc.b	$16,$FE,$D5,$FB,$C7,$81,$B3,$F0,$5A,$FF,$6F,$D3,$D1,$18,$8F,$EE
	.dc.b	$5F,$AD,$FF,$07,$7F,$C5,$21,$C6,$DF,$1E,$31,$E1,$7F,$46,$9C,$C7
	.dc.b	$B3,$3F,$EB,$2F,$F6,$AC,$3E,$37,$D0,$FF,$71,$D0,$D2,$4F,$F7,$3D
	.dc.b	$7E,$1F,$1F,$F8,$C9,$FE,$DB,$7F,$3F,$DD,$F8,$84,$AF,$90,$FC,$7F
	.dc.b	$C6,$FF,$C1,$9F,$A7,$7D,$26,$CF,$E0,$DA,$FC,$3C,$91,$88,$FF,$04
	.dc.b	$7E,$9F,$E6,$5F,$5B,$F8,$08,$C3,$74,$FF,$0D,$36,$76,$7C,$3D,$3D
	.dc.b	$5C,$9F,$CE,$74,$FC,$47,$D4,$FD,$D2,$57,$E0,$5F,$C0,$7D,$97,$F8
	.dc.b	$7A,$3C,$CD,$7F,$2E,$D9,$FC,$93,$98,$FD,$9E,$FF,$29,$FF,$18,$6A
	.dc.b	$EF,$FF,$6C,$5F,$E9,$5B,$3F,$0D,$AF,$E9,$1B,$1F,$9F,$F9,$1F,$F1
	.dc.b	$C7,$E5,$DF,$E2,$DA,$F9,$51,$89,$FD,$83,$63,$22,$78,$3F,$C2,$B8
	.dc.b	$FF,$0E,$FF,$5C,$C1,$F7,$7C,$63,$F6,$AD,$C3,$F6,$0D,$9D,$CF,$59
	.dc.b	$FE,$01,$F5,$DF,$E8,$C6,$C7,$FA,$31,$BE,$DF,$89,$FD,$D4,$EE,$7F
	.dc.b	$09,$F3,$95,$FE,$0A,$D9,$FE,$00,$DB,$3B,$9F,$A3,$C1,$FB,$D2,$65
	.dc.b	$FB,$DF,$BD,$E0,$49,$F2,$1F,$93,$79,$6F,$F6,$63,$FC,$9D,$5C,$B8
	.dc.b	$BF,$17,$7F,$22,$5F,$16,$F6,$5F,$19,$FE,$6D,$FA,$FF,$E5,$95,$A9
	.dc.b	$CA,$8C,$D7,$1A,$E1,$87,$F2,$88,$C5,$6E,$31,$7C,$AF,$8B,$63,$37
	.dc.b	$CA,$BF,$CF,$D7,$CA,$FC,$79,$7F,$BE,$27,$FD,$DE,$43,$FD,$B6,$53
	.dc.b	$FD,$EC,$53,$FD,$10,$F1,$7F,$9F,$67,$FE,$8B,$F9,$4F,$B9,$B2,$31
	.dc.b	$BE,$A9,$5A,$9C,$C7,$63,$C2,$43,$07,$E1,$0E,$33,$FD,$37,$7E,$31
	.dc.b	$F8,$63,$26,$EB,$C7,$C1,$AA,$FC,$7C,$E8,$FE,$A3,$19,$C0,$DE,$2F
	.dc.b	$F5,$37,$E5,$BE,$C6,$DF,$F5,$D6,$DF,$F1,$B6,$BF,$1B,$89,$C5,$FC
	.dc.b	$E6,$3F,$D1,$C7,$CB,$29,$FB,$81,$7F,$2B,$97,$9E,$1F,$C0,$67,$2E
	.dc.b	$E7,$E7,$F3,$7F,$6E,$8F,$E1,$93,$FD,$3C,$FE,$18,$53,$F8,$61,$4F
	.dc.b	$09,$F3,$8F,$45,$F9,$F7,$FB,$D1,$FC,$88,$7E,$65,$FD,$BB,$A3,$FD
	.dc.b	$94,$D8,$C9,$FE,$01,$8C,$BF,$E0,$3E,$9F,$85,$87,$F3,$28,$FE,$B8
	.dc.b	$D7,$FB,$CC,$7C,$77,$CA,$79,$0F,$91,$7E,$47,$FA,$3F,$5F,$FB,$6F
	.dc.b	$EB,$7F,$A1,$12,$3E,$27,$6F,$F4,$B8,$FE,$A9,$0F,$CE,$29,$1F,$EA
	.dc.b	$C6,$4F,$F0,$D7,$83,$E8,$E4,$34,$91,$F1,$F8,$BE,$FB,$E6,$6B,$F9
	.dc.b	$B3,$FC,$3F,$8F,$E9,$DF,$F1,$33,$5F,$EC,$8D,$76,$7E,$A0,$E4,$7C
	.dc.b	$42,$65,$F0,$DF,$F0,$29,$FC,$38,$FF,$18,$F8,$5F,$E0,$DE,$BF,$F0
	.dc.b	$53,$63,$FC,$64,$D7,$FB,$A7,$8A,$5F,$82,$FF,$87,$5A,$7F,$87,$4A
	.dc.b	$70,$29,$FE,$3C,$F2,$3E,$41,$1F,$BF,$A7,$F8,$B1,$F8,$DC,$9F,$24
	.dc.b	$8F,$F8,$E0,$7E,$6D,$48,$FF,$93,$7C,$AF,$34,$7F,$90,$23,$D1,$9D
	.dc.b	$7E,$11,$FE,$4E,$32,$7F,$93,$8C,$F2,$3D,$D7,$9B,$F0,$7E,$63,$FC
	.dc.b	$BC,$4A,$FF,$2F,$07,$FC,$D2,$C3,$FC,$D2,$53,$FC,$81,$2F,$47,$C3
	.dc.b	$18,$6A,$BF,$F3,$CB,$E4,$9E,$0F,$A8,$FB,$FC,$7F,$8C,$7F,$70,$7E
	.dc.b	$9A,$7D,$5E,$97,$D9,$FC,$E3,$FD,$1B,$26,$91,$F5,$E7,$FE,$7B,$23
	.dc.b	$83,$D6,$43,$FC,$F7,$DB,$FE,$D9,$7D,$C7,$A7,$17,$AD,$F8,$9F,$AC
	.dc.b	$FF,$55,$12,$7F,$D5,$CF,$FC,$00,$FC,$9E,$47,$FA,$DC,$38,$7F,$D8
	.dc.b	$47,$F2,$9F,$6D,$EE,$7F,$05,$6C,$66,$3F,$84,$75,$7E,$8C,$9B,$FC
	.dc.b	$79,$FE,$14,$7F,$B6,$3E,$BF,$FD,$17,$9B,$DE,$FD,$04,$61,$B3,$B3
	.dc.b	$FD,$CB,$F9,$3F,$FB,$90,$8E,$0F,$E6,$54,$F8,$3F,$D5,$7E,$CF,$FE
	.dc.b	$F4,$24,$EA,$32,$FF,$B7,$73,$F6,$11,$F6,$1F,$D2,$78,$3E,$83,$F4
	.dc.b	$CF,$DB,$38,$7E,$AF,$BB,$FC,$BB,$0F,$DE,$47,$D5,$F6,$7F,$1F,$CD
	.dc.b	$F9,$F6,$CE,$A7,$F1,$BD,$9F,$B2,$7F,$8C,$57,$E9,$7E,$17,$64,$8F
	.dc.b	$CB,$3C,$BE,$1F,$F8,$47,$D3,$FC,$C6,$CF,$C4,$6B,$FA,$46,$BF,$EF
	.dc.b	$B1,$F9,$37,$D4,$F8,$4F,$E2,$4D,$8C,$9F,$87,$7E,$CF,$B3,$FC,$5C
	.dc.b	$B1,$FE,$33,$3E,$9B,$96,$99,$3F,$6B,$F1,$BF,$C7,$FC,$AF,$C1,$B6
	.dc.b	$3E,$DB,$5D,$9F,$D0,$8F,$B0,$7D,$63,$FD,$6D,$D3,$F9,$46,$F1,$7E
	.dc.b	$53,$A3,$F2,$51,$F3,$8F,$F1,$66,$DF,$C4,$7E,$3B,$E9,$92,$3F,$81
	.dc.b	$3F,$C1,$97,$C7,$AB,$FB,$33,$63,$FC,$1C,$D7,$FD,$90,$D7,$FA,$4F
	.dc.b	$5F,$E9,$38,$BE,$F4,$7F,$16,$F9,$0F,$EA,$18,$FF,$23,$6F,$21,$A7
	.dc.b	$F5,$59,$F5,$1D,$31,$F6,$E1,$D6,$3F,$9E,$9F,$AE,$F9,$1E,$64,$7F
	.dc.b	$AE,$1F,$C9,$9E,$FF,$25,$FE,$08,$8F,$E5,$2D,$7F,$87,$37,$1F,$E7
	.dc.b	$0E,$63,$C6,$7F,$24,$AF,$F0,$34,$FF,$3D,$F9,$4F,$C1,$4F,$F3,$46
	.dc.b	$BF,$C6,$5B,$1B,$8F,$C3,$1F,$D8,$F1,$7C,$AF,$F8,$55,$B1,$FE,$3D
	.dc.b	$6B,$FE,$F7,$6B,$F1,$74,$F9,$EF,$F1,$23,$FB,$70,$F9,$67,$CD,$7C
	.dc.b	$D6,$C7,$C6,$3F,$BA,$38,$F7,$B8,$C6,$27,$EA,$7C,$C3,$B2,$92,$79
	.dc.b	$08,$D3,$8D,$7E,$C3,$E7,$3A,$9B,$9F,$C9,$4F,$0F,$23,$2D,$9F,$1B
	.dc.b	$5F,$EC,$3C,$BF,$8A,$23,$EB,$DB,$FB,$84,$7E,$2F,$0F,$0C,$FD,$9E
	.dc.b	$2F,$E7,$15,$FD,$D7,$E6,$7F,$8E,$12,$7F,$C9,$3C,$88,$FE,$41,$BF
	.dc.b	$F3,$89,$FE,$E2,$D7,$FA,$B3,$6B,$FC,$00,$E6,$3F,$12,$7D,$52,$43
	.dc.b	$E8,$DF,$39,$FE,$25,$EE,$FA,$33,$98,$C5,$7F,$9C,$EB,$6A,$B7,$18
	.dc.b	$7E,$6F,$CE,$FF,$0B,$9B,$DB,$FE,$61,$8C,$47,$DE,$6B,$F1,$71,$FC
	.dc.b	$5B,$FC,$E3,$83,$EB,$ED,$FC,$16,$7F,$C1,$0D,$8C,$ED,$FE,$94,$7E
	.dc.b	$EF,$CF,$7E,$F6,$FF,$D2,$CD,$73,$CD,$FE,$73,$8C,$47,$E4,$1A,$FE
	.dc.b	$69,$AF,$F9,$25,$A9,$A4,$8E,$84,$8F,$0B,$5F,$09,$82,$18,$21,$E3
	.dc.b	$78,$34,$D9,$4D,$9E,$77,$6D,$36,$53,$67,$AD,$DB,$4D,$94,$FD,$3F
	.dc.b	$C4,$7F,$BE,$1F,$DA,$B5,$FF,$7A,$9A,$F0,$3F,$C7,$8F,$C5,$EE,$FF
	.dc.b	$38,$7C,$4F,$D4,$6C,$7C,$F6,$BF,$3D,$AF,$CF,$6B,$CC,$47,$36,$9C
	.dc.b	$CA,$73,$29,$CC,$A7,$37,$91,$1E,$4D,$79,$35,$E4,$D7,$91,$C0,$87
	.dc.b	$02,$9C,$0A,$70,$34,$43,$45,$34,$7F,$B6,$BE,$3F,$F3,$FA,$9A,$48
	.dc.b	$FE,$53,$F3,$EF,$A0,$FF,$50,$4F,$F2,$08,$C4,$F1,$37,$3A,$69,$93
	.dc.b	$F9,$DE,$FF,$B1,$AE,$D1,$8D,$D8,$6E,$3C,$C7,$F4,$2F,$3D,$FA,$8B
	.dc.b	$F1,$1E,$D3,$FE,$54,$8C,$3C,$67,$73,$A8,$46,$1E,$DD,$0F,$71,$F8
	.dc.b	$67,$FA,$F7,$F0,$2F,$E3,$31,$8A,$46,$27,$51,$FD,$5A,$B9,$46,$23
	.dc.b	$D9,$93,$E6,$3E,$79,$23,$FA,$AE,$DF,$C4,$E3,$BF,$DF,$7F,$B3,$7E
	.dc.b	$66,$D1,$A4,$7C,$19,$BD,$2F,$CF,$E7,$EB,$6C,$66,$7F,$33,$39,$7F
	.dc.b	$33,$8F,$F2,$3E,$4F,$C1,$92,$3C,$BC,$3F,$52,$3E,$1F,$07,$FB,$87
	.dc.b	$E5,$2D,$1A,$47,$CB,$F8,$9F,$D1,$C7,$E7,$37,$23,$4F,$AD,$2C,$BC
	.dc.b	$8F,$8B,$7F,$3E,$7E,$38,$C6,$7F,$5B,$E8,$FC,$91,$F1,$1A,$FF,$BD
	.dc.b	$F2,$FB,$D1,$88,$FF,$66,$64,$8C,$7D,$D7,$FA,$7F,$F0,$6F,$B5,$48
	.dc.b	$C3,$F8,$23,$FC,$2A,$E3,$93,$FC,$1E,$D8,$F2,$90,$FC,$1E,$3F,$35
	.dc.b	$F2,$61,$FE,$1A,$E7,$FC,$7F,$95,$F1,$49,$3F,$96,$FA,$6F,$F7,$13
	.dc.b	$FB,$03,$D8,$E3,$AB,$FB,$7B,$63,$F2,$8D,$7C,$27,$8C,$1F,$3F,$6F
	.dc.b	$98,$E8,$EB,$F9,$16,$BE,$97,$AB,$63,$AB,$5D,$88,$F1,$FC,$DF,$E5
	.dc.b	$2C,$D2,$4F,$DC,$E2,$9A,$FC,$58,$F2,$F1,$FE,$09,$FF,$21,$F8,$DF
	.dc.b	$E8,$D9,$FF,$18,$B6,$3F,$C7,$ED,$76,$F8,$7E,$AE,$47,$E8,$B1,$46
	.dc.b	$2B,$C1,$3F,$0F,$0F,$1B,$F1,$F9,$57,$FE,$8B,$AF,$9F,$AF,$B4,$95
	.dc.b	$FD,$1B,$89,$F9,$A3,$F0,$EF,$EA,$71,$86,$C6,$67,$83,$6F,$C7,$AF
	.dc.b	$C1,$9F,$87,$87,$D6,$6E,$7D,$FF,$14,$F9,$B1,$77,$EF,$18,$C3,$F2
	.dc.b	$AD,$9F,$C2,$E5,$F8,$2E,$7F,$C4,$B6,$32,$23,$2D,$3F,$1B,$F7,$3F
	.dc.b	$8A,$FC,$23,$F2,$2D,$8C,$88,$CB,$4F,$CA,$FE,$31,$F9,$2F,$CF,$7F
	.dc.b	$32,$D8,$C8,$8C,$B4,$FC,$EF,$D9,$FE,$6B,$EA,$FF,$42,$D8,$C8,$8C
	.dc.b	$B4,$FD,$2F,$DA,$7E,$8B,$F0,$EF,$D4,$B6,$32,$23,$2D,$3F,$D1,$BF
	.dc.b	$71,$FE,$08,$8F,$F4,$1B,$63,$3B,$F3,$7F,$D3,$71,$97,$F7,$31,$FB
	.dc.b	$2F,$8A,$FD,$CB,$63,$F7,$2D,$76,$23,$F7,$FE,$FF,$F1,$BE,$9F,$CA
	.dc.b	$B7,$A7,$FD,$63,$E2,$EA,$FF,$9B,$1F,$C1,$3E,$DB,$F9,$A4,$9A,$48
	.dc.b	$FE,$21,$C7,$FB,$92,$3F,$A9,$7F,$79,$F8,$0F,$F0,$AF,$C4,$7F,$88
	.dc.b	$B6,$7F,$88,$B5,$FE,$2E,$D5,$97,$EB,$30,$F1,$92,$3F,$90,$3F,$D1
	.dc.b	$3A,$FF,$45,$1F,$B5,$D7,$F4,$BD,$7F,$A3,$BF,$DA,$FE,$0D,$EC,$25
	.dc.b	$7F,$29,$7D,$FC,$DF,$C8,$3C,$BF,$C5,$92,$7F,$1E,$FF,$2A,$FC,$27
	.dc.b	$F9,$2F,$E5,$16,$4D,$23,$FC,$BB,$F0,$9F,$A0,$47,$E7,$9F,$CD,$9F
	.dc.b	$A9,$C3,$FC,$C3,$D1,$FE,$6E,$D9,$FA,$87,$F8,$1C,$B1,$FC,$F3,$C8
	.dc.b	$FE,$6F,$F7,$96,$4D,$23,$FC,$F4,$F8,$C5,$E0,$FE,$15,$B7,$F1,$AF
	.dc.b	$80,$E7,$3F,$D1,$9A,$FF,$4C,$6B,$F5,$B8,$BF,$D6,$F7,$FE,$BC,$FD
	.dc.b	$57,$FC,$6F,$3E,$E3,$ED,$B8,$E7,$D7,$8B,$F9,$58,$C4,$FF,$89,$DF
	.dc.b	$F6,$A3,$FE,$1D,$31,$F0,$5F,$E2,$16,$99,$3F,$BD,$47,$0F,$53,$AC
	.dc.b	$7F,$73,$33,$5F,$DE,$33,$57,$9B,$D1,$D9,$E9,$F8,$4F,$EC,$3F,$AE
	.dc.b	$DC,$0D,$23,$F2,$7A,$7F,$98,$3F,$30,$FE,$CC,$7C,$0F,$8A,$FF,$18
	.dc.b	$E0,$69,$23,$FC,$63,$0F,$8B,$D9,$23,$2F,$2F,$8C,$FE,$EB,$F6,$FF
	.dc.b	$3D,$B3,$F2,$5A,$FC,$46,$BF,$D1,$3B,$BF,$4B,$B7,$F5,$A8,$F9,$3F
	.dc.b	$07,$F4,$C8,$FD,$9D,$FE,$3C,$7F,$89,$63,$DC,$73,$3E,$14,$FA,$BE
	.dc.b	$07,$DE,$48,$C1,$EE,$F9,$0F,$F0,$6F,$E1,$7F,$87,$6C,$7E,$05,$AF
	.dc.b	$DC,$6B,$F8,$AF,$84,$FF,$0A,$FC,$07,$E3,$DB,$1F,$89,$6B,$F8,$46
	.dc.b	$BE,$73,$EA,$9E,$A3,$D7,$97,$E1,$3E,$27,$EB,$92,$3E,$AB,$F6,$BE
	.dc.b	$4F,$FC,$4B,$F7,$3F,$9F,$6C,$FE,$65,$AF,$E5,$1A,$F8,$48,$78,$7B
	.dc.b	$BF,$63,$C6,$F8,$FF,$F6,$DE,$3F,$A3,$6D,$7E,$79,$AF,$EB,$1B,$1F
	.dc.b	$9C,$F9,$6F,$F1,$CF,$DF,$FF,$91,$2B,$F6,$31,$89,$FD,$73,$63,$F5
	.dc.b	$2D,$7C,$B8,$7F,$2C,$27,$47,$F2,$BC,$7F,$A7,$4B,$F4,$1E,$D8,$7E
	.dc.b	$B3,$D0,$FF,$24,$C6,$2B,$6D,$8C,$CF,$EF,$FE,$FB,$FD,$10,$DB,$FF
	.dc.b	$44,$37,$C1,$E5,$3A,$9F,$DF,$FC,$ED,$7F,$86,$B6,$BF,$7A,$DC,$4E
	.dc.b	$E7,$E8,$65,$FB,$8C,$FD,$BC,$FF,$B2,$FA,$0E,$04,$AF,$19,$FC,$39
	.dc.b	$E5,$3C,$3A,$BF,$56,$49,$FD,$1B,$FE,$6A,$67,$C7,$B9,$9F,$23,$FE
	.dc.b	$6E,$F8,$05,$EA,$73,$7F,$E6,$48,$C5,$6D,$B1,$93,$FC,$EF,$7D,$67
	.dc.b	$51,$8A,$E7,$7D,$A3,$13,$A5,$5C,$5E,$87,$F9,$7C,$F5,$1F,$E4,$C2
	.dc.b	$99,$3F,$D6,$46,$30,$67,$FD,$0A,$32,$43,$FD,$C2,$7F,$47,$CB,$FD
	.dc.b	$1F,$F2,$DF,$87,$6E,$FF,$D0,$DB,$3F,$44,$4F,$E7,$5F,$11,$D0,$EC
	.dc.b	$79,$CF,$E9,$9B,$72,$7F,$0E,$72,$AC,$B4,$FC,$69,$4F,$EB,$3C,$1C
	.dc.b	$63,$FA,$4B,$78,$0D,$31,$C5,$FE,$AA,$FC,$07,$F4,$16,$69,$2F,$CF
	.dc.b	$1E,$A3,$D0,$7A,$CF,$DA,$8F,$B3,$FD,$67,$37,$FA,$D9,$B1,$99,$FD
	.dc.b	$EB,$4F,$E1,$07,$FB,$7C,$A7,$ED,$7A,$3F,$2D,$1F,$C3,$E3,$47,$88
	.dc.b	$87,$94,$F3,$3F,$DA,$BA,$7F,$D8,$FF,$47,$FC,$8E,$4D,$23,$FD,$F7
	.dc.b	$B3,$F1,$2F,$F8,$11,$B5,$FC,$C6,$3F,$1A,$D7,$F1,$9C,$3F,$8B,$7E
	.dc.b	$21,$E4,$E2,$FC,$7F,$47,$0E,$0F,$D7,$4B,$C2,$FE,$8C,$D3,$FA,$31
	.dc.b	$4F,$E8,$C5,$3F,$C2,$FD,$FF,$EE,$2F,$CE,$FF,$AB,$41,$A4,$7F,$C3
	.dc.b	$1E,$17,$AA,$CD,$24,$7C,$0C,$7C,$EF,$B3,$1C,$65,$ED,$8F,$A1,$9F
	.dc.b	$F7,$07,$F3,$E7,$F8,$97,$C3,$FD,$3B,$4C,$9F,$DE,$8A,$7C,$59,$E8
	.dc.b	$FD,$9B,$F0,$BF,$E0,$71,$F5,$9F,$E0,$B6,$99,$70,$7F,$8F,$7C,$6F
	.dc.b	$71,$23,$E8,$3F,$07,$C7,$91,$F4,$DA,$FB,$C8,$7E,$82,$BC,$6F,$F8
	.dc.b	$EB,$C9,$2F,$D3,$3F,$C6,$44,$7F,$C8,$D1,$A8,$FF,$1F,$FE,$17,$27
	.dc.b	$FC,$95,$AF,$28,$C4,$7F,$91,$BC,$EE,$BF,$59,$FE,$57,$6C,$7F,$93
	.dc.b	$23,$09,$FF,$25,$F7,$C8,$EA,$7C,$EE,$0F,$F2,$5E,$1F,$52,$3F,$CC
	.dc.b	$C3,$E5,$FF,$94,$CF,$F9,$B6,$30,$7F,$9B,$4A,$70,$8F,$8A,$FE,$78
	.dc.b	$FB,$1C,$1F,$85,$97,$A4,$F5,$3E,$3D,$FF,$A0,$49,$7F,$BC,$9D,$3D
	.dc.b	$8F,$F3,$C7,$AD,$F3,$BE,$AF,$FD,$1F,$06,$91,$FF,$4D,$1E,$F9,$7B
	.dc.b	$1E,$F8,$FF,$41,$37,$1F,$2B,$65,$ED,$7D,$CF,$C7,$7F,$D5,$04,$BF
	.dc.b	$34,$65,$FF,$7E,$F4,$7F,$BF,$08,$E0,$F6,$7B,$DF,$86,$F2,$FE,$6E
	.dc.b	$08,$C4,$E8,$FE,$D6,$7E,$33,$2F,$F6,$1B,$83,$F4,$DC,$1F,$B2,$23
	.dc.b	$D2,$B9,$BC,$5F,$A2,$F1,$9F,$F6,$B3,$63,$37,$FE,$E0,$87,$3D,$23
	.dc.b	$FE,$E2,$32,$7D,$7F,$81,$FE,$47,$46,$92,$3F,$92,$C7,$FB,$D5,$F0
	.dc.b	$6F,$E6,$7E,$E9,$83,$C9,$F0,$5F,$4C,$8F,$DB,$7E,$B7,$A1,$FB,$0F
	.dc.b	$8F,$F9,$6D,$9F,$C5,$35,$F8,$4D,$7F,$6F,$F0,$9F,$E0,$DF,$C2,$FF
	.dc.b	$C0,$AD,$8F,$F0,$4B,$5F,$F1,$23,$5F,$F6,$57,$C3,$FF,$00,$9F,$B6
	.dc.b	$D8,$FC,$D3,$5F,$A4,$D7,$FD,$A5,$F1,$3F,$C2,$3E,$AF,$F0,$6D,$8F
	.dc.b	$D1,$35,$FB,$4D,$7F,$DB,$5F,$0D,$FC,$43,$F4,$1F,$C5,$B6,$3F,$54
	.dc.b	$D7,$F0,$4D,$7F,$DC,$5F,$1D,$FC,$63,$F6,$9F,$C9,$B6,$3F,$64,$D7
	.dc.b	$F1,$4D,$7F,$DD,$5F,$21,$FC,$83,$EC,$FF,$36,$D8,$FD,$D3,$5F,$C9
	.dc.b	$35,$7C,$A3,$E4,$DE,$EF,$F6,$E8,$C3,$63,$33,$FE,$3A,$9C,$BF,$A0
	.dc.b	$3F,$9A,$78,$7F,$D2,$3A,$3F,$9A,$C7,$6F,$94,$63,$D5,$B1,$CD,$AF
	.dc.b	$F9,$22,$D1,$88,$FE,$29,$7F,$9E,$23,$F9,$C7,$FC,$0A,$FF,$20,$EC
	.dc.b	$F2,$3F,$E0,$0E,$CF,$E4,$3C,$FF,$A6,$24,$7E,$91,$E3,$F2,$A2,$92
	.dc.b	$3F,$C2,$2E,$5E,$A7,$F0,$23,$F9,$88,$9F,$E5,$FA,$FE,$D0,$91,$FB
	.dc.b	$27,$F6,$FF,$2F,$FD,$63,$EB,$3F,$89,$B6,$7F,$99,$35,$FE,$14,$D7
	.dc.b	$C0,$7D,$0C,$7F,$7B,$7F,$D8,$BE,$B3,$F8,$EB,$67,$F9,$E3,$5F,$E2
	.dc.b	$8D,$7D,$1F,$38,$F9,$BF,$E0,$92,$8C,$4F,$F1,$96,$F8,$5F,$D5,$9B
	.dc.b	$1E,$33,$E1,$9F,$E5,$5E,$0F,$F5,$5F,$85,$FD,$B6,$3F,$D3,$D1,$85
	.dc.b	$E7,$7E,$07,$E0,$BF,$8D,$46,$0D,$B8,$8F,$07,$CE,$FF,$49,$C0,$D2
	.dc.b	$57,$F8,$70,$FC,$21,$E4,$E6,$FC,$AE,$7E,$DF,$9E,$F7,$D9,$A4,$AF
	.dc.b	$81,$F3,$FD,$C8,$FF,$7B,$94,$7F,$5D,$8E,$E7,$C3,$3F,$CC,$F0,$F9
	.dc.b	$7F,$F2,$CB,$63,$FC,$B2,$D7,$E6,$46,$BE,$3B,$FC,$1F,$1F,$D4,$E5
	.dc.b	$18,$8F,$B2,$DC,$3E,$C3,$98,$FE,$97,$D0,$E3,$FE,$F1,$1F,$EB,$6C
	.dc.b	$FF,$CC,$7E,$7F,$F9,$34,$91,$E8,$FA,$0F,$B0,$F8,$5A,$FE,$21,$B8
	.dc.b	$FE,$1D,$CF,$A2,$FA,$1F,$8B,$5F,$DA,$63,$1B,$7F,$51,$6C,$66,$77
	.dc.b	$CB,$FE,$5F,$24,$79,$F4,$F5,$46,$23,$FD,$D1,$F4,$5F,$E1,$88,$FF
	.dc.b	$67,$B6,$3F,$DA,$6D,$77,$EA,$7F,$61,$7E,$2E,$DF,$DF,$5B,$5D,$5F
	.dc.b	$F2,$6E,$DF,$E8,$88,$FF,$6D,$FD,$1F,$F9,$E2,$7F,$D2,$AD,$73,$EF
	.dc.b	$7F,$14,$8C,$47,$FB,$91,$AF,$FB,$FD,$AF,$F9,$69,$A9,$A4,$8E,$84
	.dc.b	$8F,$0B,$5F,$09,$82,$18,$21,$E3,$78,$34,$D9,$4D,$9E,$77,$6D,$36
	.dc.b	$53,$67,$AD,$FE,$6B,$F1,$AF,$8E,$F9,$D2,$8C,$5F,$FA,$F9,$B1,$FE
	.dc.b	$7C,$23,$87,$03,$F0,$37,$DA,$7F,$A0,$BD,$2C,$CE,$63,$0F,$09,$DA
	.dc.b	$9D,$C6,$1B,$3C,$27,$C5,$39,$79,$46,$5F,$F4,$8C,$65,$DC,$65,$F1
	.dc.b	$1D,$CE,$25,$36,$53,$9B,$C6,$3A,$46,$0D,$1D,$27,$CE,$E2,$36,$E4
	.dc.b	$EF,$1A,$23,$82,$1E,$27,$2D,$32,$7E,$8B,$7F,$E1,$5F,$4B,$FE,$D3
	.dc.b	$24,$7E,$D7,$64,$7F,$A6,$BE,$11,$F6,$7F,$EB,$6F,$C6,$30,$DB,$FB
	.dc.b	$DD,$0A,$DC,$2B,$6D,$37,$0F,$B3,$FF,$60,$FC,$FF,$76,$CE,$9A,$FF
	.dc.b	$9E,$E7,$82,$9F,$F7,$66,$92,$3D,$4D,$7F,$25,$F5,$3F,$8F,$7E,$8B
	.dc.b	$63,$F1,$ED,$8F,$CA,$B5,$F0,$7D,$A7,$E1,$F2,$F7,$37,$03,$4F,$C4
	.dc.b	$73,$37,$FE,$71,$8F,$8F,$0F,$A5,$BF,$E0,$AB,$4F,$E2,$71,$FF,$1E
	.dc.b	$4F,$FB,$93,$E6,$3F,$D9,$7E,$8F,$EA,$DB,$19,$3E,$8C,$65,$BF,$6B
	.dc.b	$97,$FD,$45,$08,$C4,$7F,$87,$E3,$0A,$7B,$CE,$A3,$F9,$6E,$3E,$96
	.dc.b	$9F,$38,$87,$EF,$F1,$F5,$C6,$DB,$1E,$6E,$4F,$0E,$3F,$E5,$96,$E3
	.dc.b	$FB,$CC,$7D,$F1,$E1,$BE,$37,$D7,$3F,$96,$F8,$0F,$97,$8F,$F0,$8F
	.dc.b	$CD,$9E,$07,$EE,$5B,$3F,$B0,$6C,$EE,$7E,$E6,$4F,$51,$6F,$A1,$FA
	.dc.b	$31,$86,$C6,$4F,$A3,$19,$7C,$87,$C6,$3D,$1B,$FD,$FC,$BD,$31,$F5
	.dc.b	$3E,$E3,$C2,$47,$F0,$58,$23,$F8,$B7,$98,$FC,$0B,$CD,$4F,$A6,$FF
	.dc.b	$38,$46,$1B,$97,$A9,$B3,$B3,$CB,$BF,$CC,$7E,$49,$F2,$F8,$BF,$D1
	.dc.b	$9E,$AF,$C2,$EC,$FC,$8B,$67,$F1,$ED,$7F,$1A,$D7,$F1,$9D,$BF,$30
	.dc.b	$8F,$B1,$FD,$2E,$DF,$0A,$3C,$27,$C6,$E2,$FD,$36,$1F,$E9,$DA,$E3
	.dc.b	$1F,$A2,$6C,$FE,$83,$EF,$7E,$BF,$CD,$FE,$99,$B1,$91,$19,$69,$F0
	.dc.b	$B0,$F1,$3F,$5C,$F1,$9F,$A5,$9F,$C1,$FC,$5F,$12,$47,$57,$F6,$BF
	.dc.b	$7D,$F9,$A2,$47,$EA,$CF,$DA,$BF,$9B,$AE,$FA,$FD,$19,$F0,$13,$ED
	.dc.b	$9F,$27,$FD,$45,$B1,$FD,$21,$BA,$7F,$66,$6C,$7F,$02,$F6,$3F,$33
	.dc.b	$FA,$DF,$F0,$86,$C6,$44,$65,$A7,$F0,$AF,$C0,$7F,$43,$F4,$FF,$C4
	.dc.b	$1B,$19,$11,$96,$9F,$C4,$BF,$02,$FD,$4F,$C6,$FF,$18,$6C,$64,$46
	.dc.b	$5A,$7F,$1A,$FC,$0F,$F6,$3F,$A7,$7F,$20,$6C,$64,$46,$5A,$7F,$22
	.dc.b	$EA,$FC,$0B,$FC,$81,$AF,$F1,$F6,$F3,$7F,$BA,$5C,$C3,$E3,$7F,$CB
	.dc.b	$FF,$88,$FF,$84,$9B,$1F,$E1,$86,$BF,$E1,$D6,$BF,$CA,$FF,$06,$FE
	.dc.b	$13,$F8,$F7,$F3,$86,$C7,$F3,$66,$BF,$CD,$1A,$FF,$33,$F1,$3F,$89
	.dc.b	$91,$FE,$06,$FE,$0F,$92,$47,$E0,$DB,$C9,$FD,$1B,$DB,$FE,$33,$F6
	.dc.b	$76,$8D,$23,$FD,$23,$DE,$FE,$3F,$F9,$D7,$F4,$E6,$C6,$44,$65,$A7
	.dc.b	$F5,$0F,$77,$F9,$C9,$1F,$D0,$1F,$D4,$1F,$D0,$79,$1F,$C8,$49,$8F
	.dc.b	$F4,$BE,$AF,$CC,$D7,$F9,$1D,$AF,$F3,$26,$A7,$3C,$4F,$21,$CF,$F0
	.dc.b	$2B,$F1,$DA,$7E,$34,$8F,$F6,$6F,$80,$FE,$7B,$F0,$1F,$DB,$1B,$19
	.dc.b	$3F,$DD,$05,$8F,$EE,$3C,$3F,$4F,$3B,$3F,$50,$FE,$6F,$F0,$3F,$C8
	.dc.b	$73,$F2,$37,$3F,$F6,$EC,$FF,$8B,$0F,$EE,$F1,$9F,$85,$FE,$C2,$48
	.dc.b	$F9,$27,$F7,$57,$E4,$DF,$F6,$BF,$07,$F8,$49,$FC,$A7,$07,$E1,$47
	.dc.b	$C4,$DB,$F0,$1F,$D7,$23,$66,$8F,$36,$5E,$D8,$FF,$05,$B5,$FF,$04
	.dc.b	$E3,$FD,$2D,$F4,$9F,$D2,$CC,$94,$FF,$06,$3B,$7F,$C2,$31,$CD,$FC
	.dc.b	$F1,$F2,$70,$FF,$41,$D7,$AB,$E1,$FF,$B8,$FD,$9E,$31,$46,$2F,$FC
	.dc.b	$9F,$18,$53,$B9,$D4,$7F,$4B,$3F,$9F,$71,$7E,$41,$E3,$F1,$5F,$E3
	.dc.b	$BB,$34,$91,$FE,$33,$87,$C5,$FF,$81,$3E,$CF,$C8,$D9,$F1,$B5,$F9
	.dc.b	$D3,$C1,$4F,$AF,$6F,$01,$23,$B3,$FD,$93,$0F,$34,$7A,$48,$7D,$DA
	.dc.b	$7C,$77,$F8,$A9,$B1,$F8,$8D,$3F,$07,$39,$3F,$13,$1F,$0B,$5F,$ED
	.dc.b	$B4,$F8,$FF,$84,$49,$F8,$0F,$C5,$F9,$1F,$E8,$31,$EC,$6B,$E3,$6E
	.dc.b	$1F,$92,$6F,$73,$E9,$7F,$CD,$0D,$AF,$F2,$F3,$71,$FC,$0B,$63,$CC
	.dc.b	$43,$E5,$FC,$A7,$BC,$8F,$BF,$24,$7C,$3F,$96,$93,$96,$FB,$9F,$A9
	.dc.b	$DB,$E6,$CF,$F7,$ED,$3F,$53,$97,$FB,$45,$FC,$6B,$63,$ED,$B5,$FA
	.dc.b	$4D,$8F,$D8,$FC,$C7,$F9,$DD,$FC,$AB,$67,$F0,$6D,$7E,$D3,$5F,$F0
	.dc.b	$0F,$9F,$FE,$03,$27,$0F,$EF,$74,$FE,$CB,$97,$EE,$EB,$F0,$5E,$5F
	.dc.b	$3C,$7F,$80,$B6,$B7,$1F,$A6,$F9,$2C,$12,$BF,$CF,$D1,$96,$9F,$98
	.dc.b	$EF,$E4,$47,$27,$E4,$5F,$46,$5E,$6D,$7E,$5C,$FF,$8A,$6B,$F8,$9F
	.dc.b	$CC,$F8,$C9,$1C,$A3,$2F,$43,$E0,$E9,$FC,$72,$B9,$DF,$5D,$BE,$C1
	.dc.b	$FE,$74,$A9,$F1,$0F,$8A,$7C,$57,$F9,$E3,$AF,$FD,$E9,$39,$8C,$5F
	.dc.b	$0A,$DD,$FF,$00,$8C,$4E,$A3,$FD,$03,$5C,$5D,$18,$AE,$37,$D9,$B7
	.dc.b	$CA,$76,$30,$FF,$5E,$BE,$93,$FD,$82,$53,$59,$32,$C4,$61,$B3,$CA
	.dc.b	$3F,$DE,$C7,$17,$0F,$FA,$38,$FE,$51,$A7,$FA,$6B,$E7,$BF,$11,$AF
	.dc.b	$D0,$6F,$02,$75,$3F,$CB,$1B,$5C,$CE,$E4,$30,$7E,$1C,$57,$4D,$F8
	.dc.b	$C7,$E1,$CC,$D7,$F3,$88,$CB,$FD,$63,$82,$E0,$69,$2B,$89,$FC,$B3
	.dc.b	$87,$64,$72,$7E,$90,$A7,$C9,$3E,$36,$9F,$35,$BA,$7C,$C8,$C4,$7E
	.dc.b	$E0,$E7,$87,$EF,$1B,$83,$8F,$CE,$7A,$36,$FC,$A5,$7F,$08,$AF,$ED
	.dc.b	$C7,$F0,$82,$9F,$C2,$0A,$79,$4F,$98,$7E,$10,$AF,$F7,$CE,$4F,$B5
	.dc.b	$1F,$D8,$1A,$FF,$4E,$6B,$FD,$15,$B1,$FE,$04,$E5,$B8,$7F,$26,$AF
	.dc.b	$AE,$7F,$6A,$8C,$9C,$9F,$21,$F7,$8F,$21,$65,$CD,$FC,$EE,$3F,$9C
	.dc.b	$F3,$FE,$62,$72,$D9,$D4,$FF,$78,$E8,$FF,$76,$FD,$2F,$F7,$D6,$C6
	.dc.b	$44,$65,$A7,$F8,$03,$AB,$B6,$5F,$D6,$63,$F5,$58,$FF,$88,$A7,$27
	.dc.b	$EA,$A3,$D6,$7A,$88,$62,$3F,$B3,$BF,$10,$FE,$F3,$D7,$F2,$08,$E4
	.dc.b	$E3,$CD,$FB,$06,$F0,$7E,$29,$FF,$00,$BE,$62,$3E,$6C,$11,$EA,$ED
	.dc.b	$4B,$E8,$BF,$C1,$AD,$3F,$C1,$A5,$3F,$C1,$B1,$C2,$3F,$C4,$DD,$5F
	.dc.b	$10,$93,$DD,$F8,$58,$79,$E3,$E1,$91,$F5,$65,$FE,$31,$87,$D0,$FF
	.dc.b	$8A,$DA,$7F,$8C,$3C,$0F,$3C,$64,$FF,$31,$C4,$3F,$C9,$27,$D4,$F1
	.dc.b	$5E,$1F,$E1,$5F,$F3,$1F,$CA,$FF,$90,$1B,$3F,$E4,$C8,$C7,$8B,$F1
	.dc.b	$1E,$91,$FE,$4A,$3A,$57,$33,$F0,$78,$FC,$AF,$0B,$FC,$90,$D9,$FF
	.dc.b	$23,$B5,$FF,$1A,$B5,$FB,$3D,$7F,$C0,$C8,$F9,$7C,$6F,$D5,$E4,$69
	.dc.b	$23,$F0,$95,$FE,$80,$7F,$A6,$9F,$E6,$4B,$F5,$B8,$3F,$CC,$A5,$8F
	.dc.b	$F4,$D6,$1E,$B8,$F6,$90,$FF,$4B,$0F,$C8,$A9,$1F,$F4,$C3,$FE,$EB
	.dc.b	$F8,$59,$1F,$EE,$F6,$BF,$EA,$DF,$47,$F4,$9E,$69,$A4,$8F,$82,$7F
	.dc.b	$14,$3F,$0D,$CB,$F9,$D7,$FD,$1D,$A3,$CE,$F8,$A4,$8F,$87,$19,$6F
	.dc.b	$A1,$F9,$EF,$CE,$3F,$D9,$44,$8F,$F6,$73,$FE,$D7,$CA,$47,$F2,$CA
	.dc.b	$7A,$5F,$A6,$9C,$CF,$AC,$FE,$CF,$9F,$AD,$B9,$FA,$DF,$C4,$9F,$E9
	.dc.b	$02,$ED,$F4,$1F,$CB,$77,$FE,$CB,$F3,$EF,$F7,$C4,$1A,$47,$FB,$D7
	.dc.b	$AB,$ED,$FC,$DF,$FB,$D3,$63,$3B,$FE,$F2,$72,$D8,$F6,$7B,$1F,$E1
	.dc.b	$76,$9F,$E9,$22,$F6,$7E,$96,$35,$1F,$EB,$9D,$DE,$D7,$F0,$8F,$CE
	.dc.b	$BF,$54,$D9,$F9,$0D,$78,$35,$FB,$3E,$BF,$F1,$0F,$8C,$FD,$93,$63
	.dc.b	$E8,$35,$F0,$35,$FC,$0F,$BD,$FC,$E1,$F2,$35,$C9,$FC,$E0,$C9,$4F
	.dc.b	$C4,$FC,$07,$F2,$0F,$DE,$FF,$81,$36,$3F,$00,$D7,$E0,$35,$FE,$4F
	.dc.b	$C9,$F7,$DF,$F1,$53,$5F,$F1,$73,$5F,$C3,$1F,$E0,$02,$1F,$9D,$F0
	.dc.b	$FF,$90,$E8,$F8,$2F,$DF,$92,$7F,$36,$FF,$02,$F8,$4F,$E6,$EF,$F1
	.dc.b	$96,$C7,$E5,$DA,$FD,$76,$BF,$C3,$8F,$09,$FD,$AE,$51,$F7,$A3,$F5
	.dc.b	$AD,$E6,$4E,$A7,$F0,$C7,$52,$9F,$B4,$F8,$6F,$E5,$C4,$7F,$99,$3F
	.dc.b	$D0,$BB,$7F,$75,$1F,$BC,$F0,$BF,$7E,$FE,$36,$78,$3F,$CD,$5A,$FE
	.dc.b	$CF,$E1,$FF,$94,$61,$E4,$6E,$06,$9F,$CB,$67,$B1,$F0,$0F,$EE,$5F
	.dc.b	$16,$E5,$7C,$5F,$F3,$F9,$FF,$3F,$4E,$63,$11,$FD,$9A,$30,$BC,$6F
	.dc.b	$F3,$64,$61,$FE,$99,$DF,$FD,$3C,$93,$FD,$4D,$F8,$5A,$7C,$08,$FA
	.dc.b	$87,$F0,$FB,$FE,$E7,$F3,$9F,$D5,$5B,$19,$11,$96,$9F,$E7,$DF,$8E
	.dc.b	$FE,$BE,$FF,$9E,$9A,$FF,$9F,$DA,$EF,$C6,$F1,$BF,$14,$F8,$E7,$F2
	.dc.b	$1E,$9F,$EA,$24,$7F,$96,$BF,$E3,$7E,$4F,$EF,$84,$E2,$FB,$AD,$CB
	.dc.b	$DD,$5D,$0F,$E1,$5F,$1F,$FD,$70,$8F,$F3,$87,$FA,$FB,$0F,$E7,$65
	.dc.b	$F7,$3C,$D5,$D6,$3E,$E7,$1B,$E4,$5F,$23,$FE,$43,$6C,$FF,$8D,$9A
	.dc.b	$FF,$A8,$E7,$84,$FF,$8D,$FE,$49,$CF,$49,$5F,$EA,$48,$CB,$EC,$C3
	.dc.b	$C4,$FF,$9A,$30,$7C,$A7,$F8,$B3,$BB,$B4,$62,$51,$89,$D4,$7F,$9E
	.dc.b	$A7,$53,$C3,$E5,$BF,$C6,$BF,$47,$68,$D2,$3F,$EB,$DC,$3F,$C7,$BB
	.dc.b	$FF,$7F,$8F,$F5,$D7,$9F,$FE,$90,$23,$FE,$C3,$D1,$C5,$FE,$57,$6C
	.dc.b	$7F,$B0,$9A,$FF,$B1,$DA,$F1,$8E,$04,$78,$35,$E0,$D7,$81,$A2,$1A
	.dc.b	$29,$A3,$04,$30,$43,$FD,$E9,$1E,$46,$C7,$76,$9D,$0A,$73,$3F,$DD
	.dc.b	$07,$FB,$97,$E1,$7F,$DE,$0D,$F8,$5F,$94,$A3,$2F,$83,$A3,$FC,$D3
	.dc.b	$F0,$1F,$15,$B1,$F3,$5A,$EC,$46,$DA,$6C,$A6,$DD,$11,$D3,$5D,$18
	.dc.b	$21,$82,$1F,$E7,$EC,$3F,$8F,$BD,$8C,$4F,$99,$A6,$4F,$D1,$B9,$69
	.dc.b	$98,$E7,$3E,$38,$E2,$FF,$B7,$BE,$4E,$47,$E5,$79,$DA,$27,$33,$FE
	.dc.b	$92,$1F,$26,$FA,$4F,$F4,$AC,$E6,$7F,$D2,$F3,$97,$91,$97,$11,$B8
	.dc.b	$E7,$3C,$9F,$F7,$3F,$B9,$C7,$EF,$FF,$81,$36,$33,$3F,$EA,$49,$CB
	.dc.b	$63,$8F,$D3,$FF,$AA,$3E,$67,$8B,$63,$22,$32,$D3,$E1,$65,$F3,$78
	.dc.b	$FF,$6F,$AF,$9D,$B3,$A8,$F2,$FC,$BF,$A6,$3F,$D6,$FD,$8D,$13,$99
	.dc.b	$4E,$63,$0D,$8F,$8E,$72,$97,$CE,$FE,$06,$7A,$36,$33,$3F,$87,$9C
	.dc.b	$BE,$9E,$0F,$B7,$87,$A5,$B1,$E9,$21,$DF,$3F,$F6,$7F,$C7,$7D,$06
	.dc.b	$C7,$D4,$6B,$B1,$1F,$CD,$FD,$9F,$C4,$C7,$C1,$3F,$11,$87,$F9,$DB
	.dc.b	$5F,$F6,$DE,$5B,$6C,$EA,$B8,$CF,$9E,$7D,$44,$CB,$07,$E2,$7F,$09
	.dc.b	$FD,$AD,$58,$FF,$73,$B6,$FF,$D0,$A7,$E5,$67,$6F,$7F,$AA,$F5,$BF
	.dc.b	$52,$51,$8C,$7F,$D1,$AD,$8C,$CF,$C0,$3E,$49,$FE,$F6,$DF,$F0,$AF
	.dc.b	$E9,$DA,$E7,$5F,$A2,$FE,$1A,$5F,$57,$F7,$5B,$1F,$6D,$B9,$7C,$D9
	.dc.b	$D7,$B8,$A7,$E4,$3F,$46,$30,$D8,$C9,$F3,$63,$2F,$A0,$63,$E6,$6E
	.dc.b	$3E,$67,$D1,$AF,$8F,$EB,$7F,$2F,$99,$A4,$9F,$CC,$9F,$2F,$EB,$FC
	.dc.b	$DF,$3F,$F8,$56,$C7,$DF,$6B,$F0,$DA,$F8,$35,$FB,$EF,$9F,$2F,$97
	.dc.b	$BF,$FB,$03,$EB,$76,$D9,$61,$F9,$39,$CE,$46,$FD,$1F,$DB,$25,$7D
	.dc.b	$B3,$EA,$3F,$84,$7E,$1E,$9F,$91,$7C,$5A,$FC,$5C,$7F,$DA,$35,$F1
	.dc.b	$09,$3F,$0A,$32,$DC,$1F,$21,$F6,$7F,$11,$FD,$43,$67,$F4,$4D,$7F
	.dc.b	$0C,$D7,$E2,$1E,$2C,$7C,$8F,$E2,$FC,$2F,$C2,$75,$7E,$CD,$B1,$FA
	.dc.b	$D6,$BB,$3F,$8A,$9F,$B1,$FB,$1F,$CF,$11,$F8,$87,$E9,$5F,$89,$BF
	.dc.b	$E1,$C9,$AF,$E1,$E7,$E3,$F0,$7F,$9C,$34,$FF,$2F,$B7,$CF,$FE,$B5
	.dc.b	$EA,$7F,$BE,$1B,$3A,$AF,$E0,$5F,$67,$F9,$DF,$B9,$FE,$18,$D8,$C8
	.dc.b	$8C,$B4,$FE,$15,$F6,$BF,$A5,$F9,$EF,$E2,$8D,$8C,$88,$CB,$4F,$E2
	.dc.b	$5E,$37,$EB,$7F,$2D,$FE,$38,$D8,$C8,$8C,$B4,$FE,$35,$F7,$3F,$B5
	.dc.b	$FD,$33,$F9,$23,$63,$22,$32,$D3,$F9,$17,$DD,$FE,$F7,$EA,$BF,$96
	.dc.b	$36,$32,$23,$2D,$3F,$09,$C1,$FA,$87,$F8,$97,$0F,$E5,$F7,$7D,$8F
	.dc.b	$F0,$9F,$AC,$FE,$70,$DA,$FE,$64,$D7,$F8,$B3,$5F,$E4,$DC,$1F,$C5
	.dc.b	$C8,$FF,$17,$7F,$8D,$7A,$9F,$C1,$89,$1F,$A1,$7F,$84,$FA,$FF,$9B
	.dc.b	$ED,$63,$FE,$71,$ED,$FE,$95,$39,$8C,$5F,$F4,$5F,$67,$F9,$07,$E5
	.dc.b	$7F,$D3,$DB,$19,$11,$96,$9F,$D4,$BF,$00,$FF,$0D,$79,$9F,$E8,$D6
	.dc.b	$C6,$4F,$E8,$0E,$1F,$93,$E0,$7E,$B9,$FD,$89,$8E,$1F,$F1,$07,$C0
	.dc.b	$FF,$89,$FC,$4F,$EF,$CD,$9F,$E4,$6D,$F4,$F6,$7F,$13,$8F,$E0,$D8
	.dc.b	$3D,$EF,$E7,$9F,$B3,$79,$5B,$3E,$96,$BF,$E0,$E9,$E0,$A7,$D9,$EC
	.dc.b	$FF,$45,$FD,$12,$C9,$A4,$7F,$B8,$72,$FD,$AD,$3F,$BD,$36,$3F,$BA
	.dc.b	$72,$FF,$2A,$23,$FC,$B8,$C9,$4F,$27,$B3,$F2,$88,$FF,$0A,$E6,$67
	.dc.b	$FE,$5E,$E1,$FC,$B4,$FC,$C2,$1F,$C2,$8F,$B1,$F0,$7F,$D7,$FF,$03
	.dc.b	$FB,$0D,$8C,$88,$CB,$4F,$69,$F9,$52,$67,$EE,$7F,$C5,$16,$F8,$6F
	.dc.b	$E0,$D1,$F3,$5A,$FF,$8E,$23,$4F,$F0,$73,$45,$3F,$4A,$76,$3C,$33
	.dc.b	$DC,$FC,$87,$B9,$FD,$E7,$EA,$AC,$9A,$47,$F7,$64,$3F,$01,$F1,$3F
	.dc.b	$E4,$3B,$34,$91,$EC,$F8,$AF,$F0,$37,$EA,$BF,$9F,$6C,$7E,$8D,$AF
	.dc.b	$E2,$DA,$FB,$FE,$37,$FC,$F6,$FE,$A5,$AF,$E4,$5A,$EC,$FF,$42,$9E
	.dc.b	$73,$ED,$F9,$DF,$DE,$30,$FF,$76,$C7,$E0,$DB,$BF,$EC,$FE,$17,$F6
	.dc.b	$27,$F4,$EF,$13,$FA,$65,$FF,$4E,$ED,$FD,$97,$1F,$E2,$BE,$3F,$E1
	.dc.b	$13,$2F,$B6,$FC,$5F,$90,$FF,$13,$7E,$4D,$FB,$F6,$CF,$F0,$36,$BF
	.dc.b	$AB,$6B,$F7,$FE,$3B,$FC,$96,$FF,$09,$6B,$FB,$16,$BB,$3F,$C9,$E7
	.dc.b	$F0,$53,$E1,$1F,$9B,$3F,$71,$F3,$1F,$C3,$88,$FC,$37,$F8,$8F,$A7
	.dc.b	$23,$E3,$37,$6F,$D3,$C7,$33,$E6,$EB,$FE,$13,$DF,$FC,$BF,$CD,$FE
	.dc.b	$8E,$9F,$E0,$DE,$37,$FB,$8B,$6F,$E4,$4D,$8C,$9F,$EE,$08,$FE,$4D
	.dc.b	$F3,$5F,$E8,$26,$C7,$F9,$F9,$B8,$F5,$8F,$D2,$4E,$DF,$80,$43,$CB
	.dc.b	$C1,$FA,$EF,$96,$D4,$7E,$AD,$B7,$FC,$11,$B9,$1B,$F5,$D2,$BC,$47
	.dc.b	$F2,$17,$18,$F8,$EF,$F8,$6F,$CF,$72,$25,$70,$3F,$74,$F8,$9F,$94
	.dc.b	$4E,$4F,$C7,$6B,$FD,$17,$E7,$BC,$64,$9E,$57,$FC,$6A,$39,$57,$09
	.dc.b	$FB,$ED,$7E,$11,$DB,$E3,$5E,$E3,$E3,$BF,$CF,$5F,$32,$CB,$FD,$DD
	.dc.b	$5A,$9C,$C6,$2B,$85,$AF,$83,$FE,$83,$BE,$0E,$6B,$85,$F2,$8D,$5F
	.dc.b	$2B,$E3,$0B,$E3,$18,$6D,$FF,$62,$9F,$F5,$A1,$0F,$F7,$39,$4D,$18
	.dc.b	$21,$FD,$5F,$1F,$F4,$C7,$C9,$FD,$7E,$04,$63,$4F,$EE,$CD,$AE,$02
	.dc.b	$BF,$20,$FF,$9D,$0F,$01,$FE,$96,$2C,$FE,$09,$E0,$FF,$80,$75,$E9
	.dc.b	$1F,$82,$32,$7F,$6B,$73,$1F,$D5,$B7,$B8,$1A,$4A,$E2,$7E,$6C,$B3
	.dc.b	$F9,$B6,$CF,$E6,$DB,$C1,$E2,$7D,$47,$C6,$3D,$45,$9F,$EC,$FC,$5E
	.dc.b	$22,$47,$47,$CB,$C7,$FE,$EE,$8F,$DD,$9F,$35,$C3,$FD,$65,$AF,$80
	.dc.b	$FA,$E4,$3C,$27,$F8,$37,$97,$EB,$BF,$E2,$96,$B9,$3E,$A1,$92,$9F
	.dc.b	$C7,$8A,$7F,$69,$E6,$E2,$D3,$37,$FE,$04,$8C,$C7,$94,$F1,$90,$C4
	.dc.b	$7F,$82,$B9,$FF,$DC,$5F,$09,$FE,$3B,$6C,$FF,$37,$11,$96,$9F,$D0
	.dc.b	$47,$E4,$DF,$D1,$5A,$65,$C1,$FE,$47,$EC,$FF,$78,$45,$29,$FE,$33
	.dc.b	$6B,$4A,$7A,$C9,$CE,$EA,$75,$6A,$30,$F2,$CB,$FC,$4F,$DB,$F2,$1B
	.dc.b	$7F,$23,$87,$FB,$5F,$27,$F6,$98,$F9,$E7,$6C,$7A,$BF,$A6,$3E,$4C
	.dc.b	$BB,$BF,$88,$C7,$F7,$D3,$F8,$61,$7A,$7F,$CC,$D1,$A8,$FF,$03,$0F
	.dc.b	$C7,$FF,$C1,$8D,$3F,$C1,$85,$36,$FF,$9B,$FC,$2F,$87,$97,$F9,$09
	.dc.b	$B8,$7F,$86,$AB,$E0,$35,$F4,$6B,$F1,$1B,$97,$B1,$FF,$3E,$78,$85
	.dc.b	$F4,$7F,$E2,$C2,$3A,$32,$7F,$91,$0F,$F2,$11,$0C,$1F,$E8,$7F,$1B
	.dc.b	$A7,$D1,$7F,$A1,$DB,$19,$11,$96,$9C,$8F,$26,$7F,$E9,$EF,$33,$E0
	.dc.b	$F4,$FF,$B5,$E7,$C0,$7F,$9E,$FC,$AF,$57,$DD,$7F,$96,$89,$1F,$E5
	.dc.b	$E7,$E0,$F2,$FE,$1F,$5F,$85,$3F,$B9,$C7,$FC,$DB,$E6,$7E,$1F,$80
	.dc.b	$D2,$4F,$EC,$32,$79,$BF,$27,$D3,$FC,$DE,$1F,$EB,$48,$C5,$7F,$9F
	.dc.b	$A7,$35,$C2,$11,$EA,$7F,$CC,$25,$AF,$80,$FD,$CF,$43,$EB,$7E,$89
	.dc.b	$FE,$94,$24,$7F,$AD,$63,$2F,$FA,$7C,$7E,$AB,$FE,$A3,$69,$97,$07
	.dc.b	$FA,$B3,$D6,$FC,$17,$97,$67,$FB,$39,$ED,$F5,$BF,$D6,$11,$93,$E5
	.dc.b	$C4,$3F,$D8,$47,$EB,$4F,$94,$43,$E4,$7B,$1F,$8F,$F9,$AF,$F6,$71
	.dc.b	$23,$FD,$A4,$FF,$B6,$97,$E3,$94,$8F,$F0,$AD,$3F,$DC,$3E,$DF,$E6
	.dc.b	$67,$E6,$F0,$7A,$AB,$FD,$5B,$8E,$1F,$58,$BF,$AC,$E1,$F9,$EF,$D1
	.dc.b	$E0,$FC,$73,$F9,$33,$82,$F7,$7F,$51,$F3,$BF,$71,$B3,$E5,$9D,$29
	.dc.b	$D3,$4F,$EB,$07,$F0,$FE,$6F,$CA,$1F,$C6,$65,$F0,$3F,$B5,$FA,$0F
	.dc.b	$C4,$B6,$BB,$35,$FD,$3C,$F0,$53,$F9,$0F,$84,$FD,$F7,$E9,$B6,$4D
	.dc.b	$23,$E9,$F8,$7F,$F1,$23,$4F,$F0,$F1,$7D,$6F,$CD,$C6,$A3,$C3,$F1
	.dc.b	$3F,$BE,$69,$FB,$C2,$F8,$3F,$A2,$8D,$47,$CC,$F8,$AF,$E2,$5F,$57
	.dc.b	$FA,$66,$C7,$CB,$6B,$EE,$6A,$F8,$C6,$AF,$53,$FA,$4C,$61,$B1,$99
	.dc.b	$FE,$9F,$39,$7E,$89,$F9,$B3,$F8,$24,$3E,$37,$FA,$53,$63,$F8,$14
	.dc.b	$60,$CF,$B9,$FC,$AF,$E0,$3F,$7E,$D9,$FC,$23,$5F,$AE,$D7,$FB,$B7
	.dc.b	$99,$FC,$DD,$FC,$2B,$5F,$BB,$F1,$FF,$9F,$7F,$1C,$D7,$27,$EA,$0C
	.dc.b	$94,$FC,$2F,$C8,$7E,$95,$A7,$E8,$71,$46,$63,$F3,$ED,$7F,$2C,$7C
	.dc.b	$53,$27,$F2,$37,$E2,$91,$FE,$4D,$E0,$FF,$22,$7F,$2A,$D7,$F5,$6D
	.dc.b	$EA,$FB,$2E,$63,$B9,$F0,$BC,$8F,$E6,$EF,$EC,$BD,$7E,$18,$7F,$85
	.dc.b	$3C,$7F,$E5,$B1,$83,$3F,$25,$EB,$C7,$F0,$AD,$A3,$4F,$76,$9F,$3F
	.dc.b	$0F,$A7,$8F,$81,$FF,$14,$D3,$E4,$FF,$B4,$7E,$53,$FC,$F1,$B3,$FC
	.dc.b	$05,$AF,$ED,$1A,$FF,$0A,$F9,$6F,$EB,$64,$7F,$80,$3F,$E2,$07,$E0
	.dc.b	$9F,$E4,$D3,$C3,$F2,$FF,$92,$23,$FD,$4D,$FF,$20,$FC,$C7,$FA,$73
	.dc.b	$5F,$F1,$5B,$67,$FC,$5E,$D7,$FC,$66,$D7,$DF,$E0,$FE,$89,$FF,$35
	.dc.b	$65,$FE,$22,$FC,$77,$FA,$CC,$FF,$81,$DB,$26,$9F,$D8,$BE,$69,$F3
	.dc.b	$7F,$E3,$16,$CF,$F9,$61,$BE,$EE,$E7,$FC,$B3,$C1,$F4,$6B,$E0,$9F
	.dc.b	$E5,$AE,$25,$BE,$73,$FC,$46,$D9,$FF,$20,$35,$F1,$46,$BE,$6B,$F0
	.dc.b	$52,$69,$27,$FA,$29,$E0,$3C,$35,$E0,$F9,$DF,$F1,$37,$C0,$FB,$5B
	.dc.b	$3F,$D4,$ED,$18,$8F,$F0,$CC,$61,$4F,$C5,$F9,$EF,$F1,$87,$E0,$96
	.dc.b	$4D,$23,$FE,$66,$F1,$93,$FB,$36,$C7,$A1,$BE,$77,$F9,$A9,$B1,$FE
	.dc.b	$6E,$D7,$FC,$CB,$3F,$E6,$E6,$C6,$7A,$7F,$CD,$73,$96,$C7,$F9,$DB
	.dc.b	$DF,$FF,$11,$BF,$E7,$46,$BF,$E7,$86,$BB,$F6,$7F,$35,$18,$8F,$F4
	.dc.b	$0B,$5F,$F2,$83,$53,$49,$1C,$89,$1E,$16,$BE,$16,$BE,$13,$44,$34
	.dc.b	$53,$46,$08,$7F,$A5,$3C,$A7,$85,$F8,$48,$FF,$7F,$36,$33,$87,$FA
	.dc.b	$59,$B3,$FE,$9A,$7F,$83,$2F,$8D,$C7,$6F,$F9,$BD,$B1,$93,$C1,$19
	.dc.b	$69,$93,$FD,$53,$3D,$27,$89,$23,$6D,$77,$3C,$DA,$73,$3F,$D0,$13
	.dc.b	$CD,$E2,$43,$73,$CD,$DC,$F3,$69,$CC,$4F,$37,$FD,$7F,$9F,$FA,$D3
	.dc.b	$D7,$FC,$67,$49,$A4,$9F,$F1,$E7,$E8,$7F,$E3,$55,$F9,$C7,$F9,$B2
	.dc.b	$30,$DE,$6F,$F4,$C3,$67,$67,$2F,$AE,$F3,$4A,$91,$86,$CF,$C3,$9D
	.dc.b	$4E,$7E,$1F,$FD,$49,$EF,$F8,$5B,$5E,$66,$BF,$E9,$6B,$FF,$60,$CF
	.dc.b	$9F,$EB,$FF,$D5,$9F,$25,$68,$D2,$3F,$B1,$D3,$E5,$6B,$FE,$DA,$E0
	.dc.b	$F8,$E4,$9F,$F7,$3F,$E0,$4F,$B7,$EB,$18,$6B,$FE,$EE,$8D,$3E,$03
	.dc.b	$45,$3F,$A0,$FD,$AF,$FB,$07,$E5,$3F,$A3,$B6,$3F,$A3,$B5,$D8,$8F
	.dc.b	$23,$F7,$BA,$7F,$DE,$78,$62,$3F,$D9,$98,$FA,$DF,$74,$65,$46,$5E
	.dc.b	$59,$7D,$3D,$BF,$DF,$D1,$FE,$3C,$3F,$0B,$7F,$ED,$7F,$17,$F0,$8D
	.dc.b	$8C,$9B,$39,$E1,$DF,$0E,$D1,$E4,$9F,$3B,$F9,$0F,$B3,$FF,$71,$78
	.dc.b	$1F,$8F,$6C,$64,$FE,$A2,$58,$FC,$A7,$AD,$FE,$E9,$F5,$AC,$9A,$47
	.dc.b	$E0,$90,$F2,$94,$FF,$4A,$66,$FB,$5F,$F7,$A7,$E1,$FF,$A1,$6C,$FE
	.dc.b	$99,$AF,$E4,$5A,$FF,$BE,$FC,$C7,$DD,$7E,$DE,$11,$87,$F7,$0D,$D9
	.dc.b	$10,$E5,$3C,$76,$F9,$31,$F6,$36,$F9,$2F,$A8,$87,$87,$A3,$F2,$6D
	.dc.b	$3F,$20,$5E,$6E,$CD,$2F,$E3,$4D,$FD,$0D,$8F,$E1,$CD,$CB,$8C,$23
	.dc.b	$0F,$CB,$3C,$DC,$7F,$AC,$C7,$CF,$1F,$81,$8E,$77,$D8,$91,$83,$BF
	.dc.b	$2F,$C4,$ED,$EF,$82,$31,$3E,$86,$BF,$26,$34,$FF,$1F,$3C,$9A,$7C
	.dc.b	$A2,$63,$E4,$8F,$55,$7C,$86,$9F,$20,$B9,$F8,$A3,$E6,$F6,$7F,$1D
	.dc.b	$7F,$8E,$12,$7F,$54,$FD,$9A,$ED,$7E,$0C,$FE,$21,$FA,$FF,$B9,$E7
	.dc.b	$C6,$69,$30,$FB,$38,$FF,$2B,$DB,$F1,$ED,$87,$1F,$F3,$79,$CD,$FF
	.dc.b	$5D,$8F,$D9,$65,$E2,$26,$3F,$D2,$7A,$DE,$8F,$C1,$6E,$3F,$05,$BB
	.dc.b	$7E,$01,$B1,$FD,$23,$EE,$BF,$33,$F7,$BF,$D1,$DB,$19,$11,$96,$9F
	.dc.b	$D4,$3B,$BF,$43,$FB,$E7,$F4,$F6,$C6,$44,$65,$A7,$F5,$8F,$BD,$FD
	.dc.b	$4F,$9D,$FD,$5D,$B1,$93,$FC,$84,$7F,$47,$72,$FF,$49,$E9,$FD,$8C
	.dc.b	$7F,$5F,$6C,$7F,$64,$6B,$B1,$1F,$D9,$7E,$FB,$9C,$7F,$0A,$6F,$1F
	.dc.b	$F6,$16,$F1,$1D,$CB,$EC,$FF,$C2,$5F,$8A,$7E,$B1,$B1,$FB,$66,$BF
	.dc.b	$C1,$5A,$FF,$18,$F3,$FF,$84,$FC,$EF,$F7,$76,$C7,$F7,$C6,$BF,$D9
	.dc.b	$DA,$FF,$20,$D3,$F8,$8F,$E9,$3F,$DF,$DB,$1F,$E0,$46,$BF,$DB,$DA
	.dc.b	$FF,$18,$FA,$57,$D2,$7E,$A5,$B1,$FA,$96,$F2,$FF,$08,$6C,$7F,$84
	.dc.b	$3D,$6F,$E4,$3F,$9E,$7F,$83,$DB,$19,$11,$96,$9F,$E1,$8F,$63,$F9
	.dc.b	$4F,$E0,$9F,$E1,$76,$C6,$44,$65,$A7,$F8,$83,$DA,$FE,$63,$F8,$A7
	.dc.b	$F8,$7D,$B1,$91,$19,$69,$FC,$33,$BB,$F5,$7E,$07,$F6,$DB,$7C,$EF
	.dc.b	$F0,$B6,$D7,$F0,$B6,$EB,$FC,$A9,$B1,$FD,$07,$9D,$F8,$53,$C1,$FE
	.dc.b	$8B,$F9,$97,$F3,$89,$CC,$62,$B8,$56,$D5,$6E,$30,$DF,$82,$FE,$9F
	.dc.b	$FB,$85,$A3,$48,$F4,$F1,$3F,$76,$FF,$56,$C7,$CB,$18,$84,$61,$FD
	.dc.b	$E9,$BC,$3F,$85,$F8,$5F,$D2,$3E,$C9,$8F,$F2,$F6,$D7,$F2,$F6,$EB
	.dc.b	$B9,$FE,$5F,$7F,$E7,$CF,$B3,$F6,$7B,$7F,$DA,$1B,$81,$A7,$F2,$0D
	.dc.b	$BF,$97,$F0,$FD,$5E,$1F,$ED,$FF,$69,$F4,$5B,$7F,$D9,$38,$BE,$BC
	.dc.b	$E6,$0F,$CB,$F8,$8F,$EE,$FF,$9E,$D9,$34,$8F,$F6,$ED,$FD,$B5,$EB
	.dc.b	$8E,$66,$76,$F0,$BF,$E0,$19,$DB,$F7,$3D,$3F,$F0,$2F,$EF,$BF,$6D
	.dc.b	$B1,$91,$19,$69,$D4,$FD,$99,$D6,$BF,$C1,$FD,$DE,$32,$47,$E6,$F6
	.dc.b	$F8,$EF,$F8,$97,$91,$F9,$5B,$E8,$3F,$B0,$B6,$7F,$AA,$B7,$97,$F2
	.dc.b	$CD,$8F,$CA,$FC,$77,$F8,$7B,$F1,$CF,$CA,$36,$32,$23,$2D,$3F,$3B
	.dc.b	$F2,$1F,$19,$FC,$7F,$42,$31,$97,$EA,$1B,$19,$9F,$9E,$7F,$89,$8F
	.dc.b	$F0,$AF,$37,$E1,$4F,$D6,$7E,$54,$FA,$4F,$EF,$AD,$9F,$EE,$2D,$D3
	.dc.b	$73,$FE,$07,$F5,$BF,$BE,$5F,$3B,$FD,$D5,$FF,$BE,$2F,$AF,$C4,$F4
	.dc.b	$24,$FF,$7C,$C5,$F2,$7F,$E4,$2F,$C0,$FF,$70,$D9,$FD,$EB,$5F,$D6
	.dc.b	$35,$F8,$7F,$17,$FE,$B0,$AF,$F7,$4C,$7F,$AB,$5B,$97,$F0,$39,$FF
	.dc.b	$0A,$71,$FD,$EA,$F8,$9A,$7F,$9D,$B7,$6A,$F8,$CF,$86,$4A,$F8,$6F
	.dc.b	$ED,$3E,$5B,$FC,$EF,$1F,$BF,$6B,$F5,$9B,$5F,$C7,$1C,$C7,$D2,$3F
	.dc.b	$21,$97,$FA,$66,$7C,$3F,$2D,$F4,$89,$3F,$49,$FE,$07,$F0,$9F,$EB
	.dc.b	$E9,$FE,$42,$D9,$FE,$20,$D7,$F0,$ED,$7C,$07,$C0,$F2,$3F,$9C,$BE
	.dc.b	$22,$1F,$E7,$4F,$9A,$7D,$63,$C2,$FF,$3B,$E5,$FE,$E7,$A5,$6A,$73
	.dc.b	$8F,$F9,$EE,$31,$2A,$E1,$18,$79,$4F,$F4,$0F,$9B,$F4,$36,$BF,$DB
	.dc.b	$2F,$F3,$E8,$CF,$11,$F6,$70,$7E,$4A,$BD,$27,$93,$2F,$43,$E1,$BF
	.dc.b	$24,$63,$0F,$1B,$63,$39,$7A,$AB,$E3,$70,$FE,$1D,$FF,$3D,$5F,$E7
	.dc.b	$DC,$BF,$E8,$4F,$90,$71,$BE,$47,$FD,$3B,$F1,$1F,$06,$FF,$D3,$F5
	.dc.b	$A9,$CD,$FF,$A8,$63,$12,$AE,$0A,$33,$18,$6C,$F1,$BF,$05,$71,$8C
	.dc.b	$5F,$68,$F0,$CE,$AF,$AE,$8E,$0F,$ED,$55,$FE,$A6,$8C,$65,$FE,$A0
	.dc.b	$8F,$F4,$E9,$FE,$A2,$8C,$B6,$78,$BF,$EE,$C3,$13,$C6,$77,$1F,$38
	.dc.b	$FE,$D8,$7E,$3F,$5F,$AE,$FD,$4D,$3F,$C0,$8E,$5A,$FF,$AA,$CE,$A4
	.dc.b	$3D,$17,$E2,$7F,$93,$ED,$FE,$CE,$F9,$5F,$C7,$F0,$7E,$26,$51,$89
	.dc.b	$D7,$07,$E3,$E7,$2F,$9D,$F2,$1E,$83,$F1,$D8,$79,$38,$7F,$3F,$1F
	.dc.b	$A5,$3E,$39,$E4,$D3,$E1,$65,$E5,$8C,$9F,$1C,$D1,$EB,$30,$7E,$F0
	.dc.b	$F1,$97,$1F,$5F,$17,$A2,$31,$8F,$F3,$2E,$3F,$81,$14,$F7,$96,$BF
	.dc.b	$A0,$C6,$A3,$B9,$F3,$4F,$19,$0F,$E3,$C3,$E5,$BF,$C5,$DC,$BF,$EF
	.dc.b	$A6,$C6,$4F,$F1,$AC,$65,$FE,$8B,$CD,$DD,$FF,$14,$F3,$FF,$2F,$7C
	.dc.b	$A7,$F8,$BD,$C6,$1F,$D5,$A7,$84,$FC,$83,$CC,$7A,$0C,$63,$FE,$4A
	.dc.b	$E4,$FC,$C3,$63,$FD,$F2,$D7,$FA,$21,$8E,$1F,$EA,$6B,$F2,$2F,$EA
	.dc.b	$6D,$3F,$A9,$94,$FE,$A6,$53,$D7,$D7,$ED,$AF,$E9,$6D,$C8,$D3,$D7
	.dc.b	$CB,$FA,$4A,$F5,$35,$F6,$9F,$1B,$B7,$91,$1F,$8E,$64,$A7,$C5,$24
	.dc.b	$73,$7E,$2E,$3F,$E0,$5F,$A1,$FF,$33,$77,$A5,$18,$69,$FE,$12,$29
	.dc.b	$FE,$5D,$F0,$78,$7E,$2D,$FE,$18,$24,$7F,$87,$1F,$F1,$1F,$89,$F0
	.dc.b	$FC,$4F,$F1,$47,$16,$2B,$D3,$BF,$F8,$BA,$BE,$1F,$3F,$E9,$75,$F4
	.dc.b	$11,$FD,$47,$93,$FE,$68,$E5,$FF,$26,$13,$0F,$F2,$03,$FE,$43,$FD
	.dc.b	$F6,$47,$F9,$1D,$AF,$F9,$0C,$F5,$9F,$06,$7F,$CE,$7B,$7F,$95,$C9
	.dc.b	$1F,$E8,$EE,$CF,$8D,$F9,$FF,$F9,$84,$91,$90,$FF,$9A,$66,$47,$F9
	.dc.b	$B7,$CF,$F9,$D1,$88,$FB,$F9,$7B,$C9,$DB,$F9,$3F,$47,$EA,$7D,$3F
	.dc.b	$F9,$F8,$95,$FE,$A0,$8C,$BF,$E5,$FE,$1F,$F3,$71,$1C,$1F,$E9,$B1
	.dc.b	$F8,$77,$FA,$97,$D5,$FC,$03,$63,$26,$DF,$F5,$37,$EE,$12,$3F,$D5
	.dc.b	$FE,$7F,$E1,$E3,$11,$FE,$9F,$CB,$F1,$5F,$99,$7F,$AC,$5B,$1F,$EB
	.dc.b	$66,$BB,$11,$FE,$8F,$C6,$47,$C7,$3F,$37,$EA,$FF,$B2,$A3,$F9,$24
	.dc.b	$7E,$90,$FC,$EE,$BF,$CF,$F9,$DF,$0E,$F7,$3F,$39,$F8,$F7,$FB,$64
	.dc.b	$97,$FE,$DD,$7F,$DC,$9F,$87,$48,$FF,$74,$35,$FF,$60,$53,$E0,$7F
	.dc.b	$4D,$FA,$07,$FB,$CC,$93,$FE,$F7,$7F,$DF,$DE,$BF,$F1,$6F,$83,$34
	.dc.b	$93,$FA,$0B,$7C,$27,$EC,$7F,$7B,$FF,$6F,$46,$25,$18,$9D,$47,$FB
	.dc.b	$9E,$75,$3C,$3E,$1F,$F7,$5C,$F6,$8D,$23,$FB,$F3,$F8,$07,$C4,$FF
	.dc.b	$85,$7F,$11,$C4,$63,$D8,$FE,$09,$18,$84,$ED,$DB,$5F,$8F,$F1,$9F
	.dc.b	$C2,$FE,$E3,$D6,$D8,$F4,$35,$D8,$8F,$83,$F1,$FF,$C4,$FC,$FB,$3F
	.dc.b	$CA,$0F,$81,$19,$69,$93,$EB,$FC,$7F,$57,$FD,$EB,$B2,$31,$CF,$F8
	.dc.b	$66,$C6,$67,$C4,$7C,$DF,$91,$FE,$49,$F3,$FF,$51,$B1,$91,$19,$69
	.dc.b	$F5,$7E,$4F,$F5,$AD,$3F,$5A,$7E,$31,$CB,$4C,$9F,$CC,$FE,$53,$EC
	.dc.b	$BF,$CF,$1B,$DB,$F4,$DB,$1F,$61,$AF,$F0,$CF,$03,$FC,$1E,$47,$F1
	.dc.b	$BC,$5F,$CF,$FF,$20,$B0,$69,$1F,$C7,$7C,$BF,$F5,$BD,$0D,$24,$7F
	.dc.b	$50,$F3,$FF,$9B,$61,$DE,$3F,$C1,$07,$D2,$F9,$8F,$EE,$76,$69,$23
	.dc.b	$FC,$9D,$F3,$39,$7F,$AD,$B7,$4F,$C9,$36,$3F,$2E,$D7,$E0,$1F,$C1
	.dc.b	$8F,$EC,$4F,$3F,$84,$E3,$5F,$D7,$DB,$3F,$E6,$83,$F4,$93,$B7,$AF
	.dc.b	$CD,$FF,$98,$54,$23,$0D,$8F,$4C,$6F,$2F,$87,$1E,$37,$FB,$94,$EE
	.dc.b	$3D,$73,$C8,$8F,$F9,$BF,$E3,$3F,$D0,$4F,$F7,$C6,$E3,$FB,$36,$C7
	.dc.b	$EE,$1A,$FC,$23,$E8,$9F,$E0,$35,$CB,$EB,$6C,$77,$8C,$1F,$5B,$E7
	.dc.b	$3F,$0D,$1F,$C3,$DA,$FF,$24,$6B,$FA,$E6,$BF,$CD,$35,$FC,$13,$FE
	.dc.b	$4C,$E6,$FE,$6B,$8F,$F9,$A6,$3F,$6A,$D8,$CC,$7F,$9A,$5A,$7F,$24
	.dc.b	$F9,$EF,$F2,$54,$7F,$34,$6B,$FB,$E6,$C7,$F4,$27,$31,$E3,$3E,$97
	.dc.b	$91,$FD,$1F,$83,$FD,$41,$0C,$FF,$98,$CF,$F8,$8D,$B1,$93,$F9,$84
	.dc.b	$65,$F9,$BB,$7D,$E6,$C7,$F7,$A7,$BE,$5F,$EA,$9F,$A7,$FC,$E7,$C0
	.dc.b	$FF,$2B,$6C,$FF,$30,$6B,$FC,$D5,$AF,$F1,$2E,$9F,$F3,$0F,$49,$A4
	.dc.b	$8F,$D9,$61,$E1,$7C,$78,$7F,$98,$33,$74,$7E,$EC,$93,$FD,$4D,$FD
	.dc.b	$F7,$C4,$FF,$90,$FF,$20,$FE,$A6,$D9,$FE,$A0,$D7,$FB,$2B,$5F,$E5
	.dc.b	$FF,$49,$FB,$E2,$3E,$17,$F8,$0F,$0F,$F9,$3F,$F6,$3F,$EB,$ED,$9F
	.dc.b	$EB,$8D,$7F,$B8,$B5,$FF,$2B,$70,$7F,$B5,$3A,$3F,$D9,$A4,$8F,$F5
	.dc.b	$73,$FE,$F3,$F5,$3F,$71,$F1,$A6,$92,$7F,$DB,$91,$96,$C6,$5E,$46
	.dc.b	$08,$60,$87,$F6,$88,$E4,$E0,$FF,$7E,$47,$37,$4D,$72,$74,$39,$71
	.dc.b	$FF,$13,$7F,$DF,$27,$F7,$19,$46,$63,$FD,$F7,$5E,$AF,$6B,$FC,$F5
	.dc.b	$DF,$B6,$CF,$95,$AF,$F8,$06,$D1,$88,$FF,$41,$46,$14,$F3,$69,$CC
	.dc.b	$A7,$39,$E4,$A7,$94,$F1,$24,$71,$6B,$C4,$E6,$78,$A7,$9C,$F2,$8F
	.dc.b	$F7,$57,$D0,$FF,$84,$F2,$71,$B8,$D3,$98,$C3,$63,$35,$FE,$1E,$FA
	.dc.b	$7F,$5C,$FF,$87,$5B,$19,$CB,$FD,$49,$39,$6C,$7F,$8A,$76,$7C,$2A
	.dc.b	$FF,$D5,$5F,$5F,$FE,$2A,$9C,$C6,$2B,$85,$6D,$56,$E3,$0D,$3E,$A5
	.dc.b	$76,$6C,$70,$9F,$8F,$39,$6C,$3E,$C7,$FD,$79,$F4,$DD,$5B,$3E,$26
	.dc.b	$BF,$EC,$1A,$F5,$C6,$21,$1E,$EF,$B2,$FF,$63,$7C,$B7,$9D,$B3,$EC
	.dc.b	$6B,$CF,$04,$62,$3F,$73,$B7,$D7,$D3,$EA,$12,$3F,$CA,$3B,$FD,$CD
	.dc.b	$FF,$07,$08,$C4,$EE,$3F,$CB,$51,$87,$8D,$75,$7F,$CB,$DF,$70,$FC
	.dc.b	$AD,$F2,$BF,$ED,$FF,$89,$FF,$2F,$CE,$63,$15,$C2,$B6,$AB,$71,$87
	.dc.b	$C5,$1E,$CC,$BC,$F8,$F8,$E3,$CB,$8F,$8D,$FF,$38,$7E,$47,$FB,$A5
	.dc.b	$F7,$FF,$E6,$D8,$C3,$7A,$7F,$DE,$8D,$9D,$9F,$8A,$FB,$DF,$AD,$C5
	.dc.b	$F9,$26,$F0,$9A,$7D,$B9,$F9,$58,$F8,$DF,$47,$07,$A9,$CE,$9F,$8A
	.dc.b	$FB,$84,$3E,$B7,$F2,$D1,$86,$C6,$4F,$CE,$C6,$5F,$CF,$FD,$CF,$2F
	.dc.b	$9A,$FC,$F3,$63,$22,$32,$D3,$ED,$DF,$CC,$C7,$D1,$1E,$6C,$7E,$8E
	.dc.b	$9F,$B2,$E1,$F3,$FD,$77,$EA,$DB,$5F,$B0,$6B,$F9,$56,$BF,$08,$EF
	.dc.b	$5F,$0F,$8B,$ED,$46,$4F,$B7,$8F,$8C,$93,$BC,$7C,$B5,$C2,$BC,$C7
	.dc.b	$F0,$7E,$6F,$F5,$85,$7F,$05,$6C,$64,$FD,$FB,$87,$BC,$75,$BE,$F1
	.dc.b	$B3,$F8,$0C,$ED,$FD,$BB,$72,$FD,$8C,$74,$FC,$2E,$3C,$1B,$7D,$49
	.dc.b	$1E,$7F,$97,$FD,$C5,$7E,$E9,$B6,$69,$8F,$63,$D2,$4A,$F2,$BF,$D1
	.dc.b	$32,$F8,$84,$CF,$E3,$9F,$93,$C3,$F8,$3F,$DF,$76,$25,$BA,$FE,$44
	.dc.b	$78,$BE,$43,$F2,$BF,$87,$FC,$B6,$D7,$CF,$6B,$FC,$77,$44,$62,$3E
	.dc.b	$E1,$0F,$AE,$7D,$03,$F0,$64,$30,$43,$E7,$FC,$E7,$E9,$A7,$EE,$36
	.dc.b	$3F,$00,$D7,$C4,$D7,$EB,$FB,$BF,$AE,$FD,$CF,$F0,$8D,$8F,$C4,$35
	.dc.b	$F5,$35,$FB,$FF,$3F,$FB,$6F,$D9,$7F,$18,$D8,$FC,$83,$5F,$84,$D7
	.dc.b	$F0,$FF,$81,$FE,$FB,$F3,$5F,$CA,$36,$3F,$30,$D7,$E5,$35,$FC,$7F
	.dc.b	$DD,$78,$F8,$3F,$9F,$B7,$43,$4F,$44,$FC,$E3,$E8,$71,$7E,$0F,$9F
	.dc.b	$F3,$7E,$07,$F1,$0F,$D1,$FF,$4C,$DA,$FD,$53,$5F,$B8,$D7,$F4,$3F
	.dc.b	$85,$FF,$18,$FC,$E3,$F5,$CD,$8F,$D9,$35,$FC,$23,$5F,$1F,$3F,$E5
	.dc.b	$30,$FD,$7C,$7E,$27,$ED,$FF,$91,$FC,$67,$EE,$5B,$1F,$BD,$6B,$B1
	.dc.b	$1F,$BF,$FC,$27,$F9,$5F,$E2,$16,$4D,$23,$FC,$1F,$F0,$3F,$E6,$3F
	.dc.b	$87,$FF,$07,$6C,$64,$46,$5A,$7E,$5C,$F5,$E3,$EA,$6F,$81,$FB,$E7
	.dc.b	$F4,$9E,$87,$F3,$EF,$CD,$7F,$8A,$36,$3F,$8C,$35,$D8,$8F,$E3,$5F
	.dc.b	$86,$FF,$46,$FD,$3A,$C9,$A4,$7F,$91,$7E,$1F,$CD,$FE,$2E,$DE,$8F
	.dc.b	$F0,$3B,$7A,$0E,$E7,$F1,$04,$E7,$FC,$46,$7F,$CA,$FF,$10,$FE,$AF
	.dc.b	$F9,$25,$A3,$48,$FC,$53,$97,$2F,$DF,$AF,$89,$1F,$18,$FE,$71,$F8
	.dc.b	$27,$F6,$2F,$DE,$FF,$9C,$36,$32,$23,$2D,$3B,$63,$E6,$7E,$A6,$9F
	.dc.b	$E2,$4F,$9B,$FE,$DB,$F8,$07,$F4,$36,$CF,$F4,$76,$BF,$C6,$5A,$F9
	.dc.b	$4F,$E2,$47,$AB,$8F,$EF,$37,$2F,$5B,$4F,$F2,$C7,$E7,$1F,$E3,$C5
	.dc.b	$F8,$EF,$F9,$5E,$30,$DC,$7F,$C0,$2D,$9D,$9C,$BE,$17,$FA,$2D,$1A
	.dc.b	$48,$FE,$70,$7C,$2D,$BE,$73,$71,$F3,$E9,$F1,$63,$D6,$D7,$DC,$43
	.dc.b	$F8,$46,$2C,$1F,$0D,$FE,$16,$FC,$5B,$D4,$DA,$F7,$B5,$FF,$2F,$4F
	.dc.b	$05,$3F,$07,$E2,$7F,$CD,$F3,$FE,$E6,$9F,$F3,$6B,$75,$D4,$7A,$33
	.dc.b	$F5,$BE,$86,$BE,$8C,$7C,$D1,$FE,$2A,$FC,$B7,$E7,$36,$7E,$9B,$5F
	.dc.b	$2B,$5F,$81,$9F,$C3,$7E,$D1,$DF,$4F,$2F,$C8,$7F,$8D,$BF,$6D,$FB
	.dc.b	$93,$FE,$78,$CD,$39,$CB,$F0,$2D,$9F,$86,$D7,$BE,$3E,$A9,$7E,$19
	.dc.b	$FE,$C0,$F8,$6E,$B1,$8D,$7F,$C8,$B1,$89,$D4,$FF,$A9,$5B,$7F,$43
	.dc.b	$E4,$FF,$DA,$B3,$F9,$26,$C6,$4F,$F6,$7C,$65,$A7,$D8,$F9,$4F,$F4
	.dc.b	$8C,$7E,$69,$B1,$93,$FC,$0D,$19,$69,$F1,$F0,$FD,$EC,$7E,$26,$BC
	.dc.b	$33,$F1,$DB,$19,$8F,$09,$92,$9F,$48,$F8,$FA,$FE,$0C,$FD,$8F,$C5
	.dc.b	$7F,$BD,$BE,$33,$FC,$D3,$C9,$FE,$E6,$8C,$57,$E6,$31,$FF,$7E,$CE
	.dc.b	$7E,$11,$7C,$27,$24,$AF,$F7,$F9,$C4,$B5,$FE,$F2,$E2,$FF,$54,$47
	.dc.b	$FB,$7C,$87,$F9,$A8,$A7,$F9,$68,$A7,$FA,$64,$FE,$05,$9F,$FB,$75
	.dc.b	$FA,$87,$E0,$38,$51,$98,$F8,$87,$E4,$0F,$F4,$49,$D0,$A7,$84,$FA
	.dc.b	$9F,$2D,$CF,$6F,$F4,$6B,$60,$D3,$C1,$3D,$F6,$FA,$93,$83,$FD,$B0
	.dc.b	$79,$9F,$C8,$65,$F4,$1B,$1F,$C1,$9B,$97,$E5,$E7,$53,$F0,$CF,$31
	.dc.b	$48,$7E,$AF,$8B,$FD,$47,$FB,$57,$C3,$6D,$FE,$B2,$53,$96,$C7,$47
	.dc.b	$E7,$14,$F8,$A7,$62,$9A,$3C,$2F,$F0,$A5,$F0,$FF,$C2,$9A,$7F,$0A
	.dc.b	$29,$FC,$28,$A7,$32,$1D,$89,$AF,$E2,$1F,$23,$5F,$0E,$BF,$C8,$DF
	.dc.b	$E4,$24,$3F,$A5,$F3,$FF,$B1,$9F,$E6,$23,$E2,$A9,$1F,$E6,$66,$4E
	.dc.b	$FC,$CE,$4E,$51,$F0,$F0,$F1,$3E,$86,$BE,$89,$E0,$FE,$3E,$32,$48
	.dc.b	$CC,$76,$9E,$AF,$38,$CB,$FD,$13,$AB,$F2,$6F,$F5,$D3,$F2,$65,$E1
	.dc.b	$FE,$CD,$1A,$8F,$E9,$7A,$7F,$B9,$BF,$46,$FE,$CC,$48,$FE,$DF,$19
	.dc.b	$7C,$47,$9E,$FC,$2F,$8F,$0F,$2B,$ED,$6E,$3E,$B6,$CF,$09,$FC,$3D
	.dc.b	$7F,$82,$D7,$D4,$FF,$82,$DA,$7F,$82,$CA,$7F,$82,$CA,$78,$4F,$25
	.dc.b	$7C,$69,$78,$AF,$09,$4F,$CB,$3F,$C3,$E4,$7F,$C4,$6F,$26,$1F,$00
	.dc.b	$87,$FA,$CF,$C5,$E7,$F9,$47,$F8,$DC,$91,$FE,$E7,$8C,$BC,$0F,$31
	.dc.b	$FE,$4A,$1F,$BD,$7F,$BA,$FC,$8F,$2B,$63,$26,$DF,$F2,$AF,$EA,$52
	.dc.b	$3F,$DE,$9E,$5F,$B2,$31,$1F,$E5,$E9,$FF,$31,$BC,$8C,$1F,$E6,$43
	.dc.b	$07,$90,$FC,$C6,$FF,$57,$83,$EC,$47,$C2,$3F,$DA,$5E,$7F,$CA,$FC
	.dc.b	$73,$FC,$F0,$48,$FF,$3E,$3F,$E8,$5F,$C7,$24,$7E,$B7,$87,$E9,$FA
	.dc.b	$7F,$4A,$73,$3F,$E8,$FF,$81,$FF,$4A,$26,$91,$F9,$47,$E3,$4F,$F5
	.dc.b	$28,$FD,$F6,$91,$FF,$53,$99,$3F,$D6,$23,$F4,$BF,$F5,$AB,$4C,$B8
	.dc.b	$3E,$03,$F1,$4F,$F6,$38,$FD,$DB,$F7,$F8,$FF,$B2,$23,$07,$BF,$9B
	.dc.b	$FD,$92,$E0,$FF,$7D,$7B,$DF,$92,$AF,$F6,$BC,$3E,$52,$91,$FF,$6C
	.dc.b	$99,$3F,$DC,$63,$F7,$FF,$F7,$3B,$4C,$B8,$3F,$DE,$6F,$84,$FF,$7B
	.dc.b	$97,$0F,$D2,$FC,$2F,$8E,$3F,$DE,$E6,$71,$FA,$4F,$F8,$C8,$F6,$69
	.dc.b	$FA,$AF,$17,$E0,$CE,$9B,$19,$9F,$AF,$C5,$F8,$A6,$F1,$FE,$34,$FE
	.dc.b	$13,$8F,$ED,$BE,$1F,$F6,$BF,$83,$F4,$6D,$7C,$B6,$BF,$BA,$9E,$0A
	.dc.b	$7C,$7F,$15,$FC,$76,$13,$FC,$0A,$30,$D8,$FE,$4A,$73,$CD,$3C,$63
	.dc.b	$F1,$C7,$EE,$67,$F8,$5F,$C6,$7F,$1E,$6C,$7F,$0F,$A4,$E5,$B1,$9F
	.dc.b	$8D,$F0,$C1,$A4,$8F,$19,$F8,$F3,$C2,$79,$7E,$3B,$A1,$1F,$03,$DD
	.dc.b	$87,$8C,$A7,$E6,$8E,$67,$C4,$9F,$E6,$ED,$36,$7E,$64,$CE,$9F,$65
	.dc.b	$FE,$0B,$E1,$7F,$3A,$23,$D6,$34,$FE,$9F,$17,$CB,$3E,$43,$F9,$8F
	.dc.b	$D4,$7E,$A2,$31,$48,$C4,$EA,$3F,$84,$4E,$A7,$B7,$07,$E2,$71,$FE
	.dc.b	$7D,$2F,$94,$FF,$15,$DF,$E1,$63,$0F,$F8,$F7,$1F,$1C,$EA,$3F,$0B
	.dc.b	$3A,$9F,$27,$CA,$FF,$46,$FC,$1E,$D1,$A4,$7F,$37,$87,$C2,$7C,$FF
	.dc.b	$31,$FD,$3B,$E9,$7D,$CD,$8C,$88,$CB,$4F,$F0,$4F,$A4,$F6,$7F,$AB
	.dc.b	$DF,$F8,$02,$31,$3F,$DC,$1B,$19,$9D,$CF,$90,$FF,$32,$FA,$3F,$C3
	.dc.b	$7E,$6B,$FB,$0F,$D4,$FF,$85,$A3,$12,$8C,$77,$AB,$73,$FE,$25,$AD
	.dc.b	$D7,$D2,$C3,$E0,$4F,$E1,$0F,$E4,$FB,$7E,$03,$97,$F5,$FC,$1F,$DC
	.dc.b	$A7,$F3,$B3,$96,$C6,$7D,$3F,$C2,$7C,$E7,$C7,$9F,$B4,$D7,$EB,$B6
	.dc.b	$BF,$1A,$DA,$7C,$EF,$ED,$5B,$5F,$86,$8C,$19,$F8,$0F,$F0,$1F,$AF
	.dc.b	$F9,$36,$CF,$E4,$1A,$FE,$2D,$AF,$FB,$23,$DA,$FF,$05,$FB,$DF,$E3
	.dc.b	$49,$46,$27,$FC,$87,$18,$7F,$D4,$A6,$23,$4F,$F9,$47,$B1,$D5,$FE
	.dc.b	$4B,$BF,$F1,$FC,$63,$7E,$35,$BA,$D3,$FE,$50,$9D,$57,$F9,$1B,$DF
	.dc.b	$7C,$FF,$EF,$9B,$5F,$E3,$66,$BB,$8D,$11,$D3,$5D,$18,$21,$82,$1F
	.dc.b	$E6,$28,$EA,$D8,$E4,$E7,$9B,$D6,$FF,$9C,$67,$FD,$9A,$E6,$BC,$D3
	.dc.b	$A8,$FF,$1C,$CF,$E9,$F2,$F7,$74,$7F,$8F,$1B,$3F,$3C,$E9,$19,$7C
	.dc.b	$8D,$72,$6C,$EA,$E5,$A6,$44,$F8,$5E,$66,$5C,$3D,$27,$EC,$E5,$FE
	.dc.b	$85,$F9,$FB,$A1,$A4,$AF,$C2,$19,$2B,$F8,$DA,$FF,$2C,$B6,$34,$7F
	.dc.b	$A6,$75,$FF,$6E,$B8,$3F,$CB,$EE,$23,$F2,$DF,$01,$CA,$7F,$2A,$D7
	.dc.b	$FB,$67,$BC,$FA,$9F,$F3,$3F,$E7,$9F,$E1,$E8,$C4,$A3,$13,$A8,$FF
	.dc.b	$1F,$4E,$A7,$87,$4F,$CB,$E7,$4F,$F0,$FA,$FA,$11,$89,$D4,$E7,$3F
	.dc.b	$01,$FE,$78,$FA,$1B,$66,$91,$F0,$FD,$5F,$22,$3D,$A9,$1F,$76,$1C
	.dc.b	$BF,$AD,$BF,$AA,$48,$FD,$67,$07,$FA,$F6,$3F,$D6,$3F,$50,$FA,$AE
	.dc.b	$4D,$AF,$C3,$36,$77,$3D,$F9,$F5,$F6,$FF,$EB,$A6,$DF,$FB,$21,$AF
	.dc.b	$FB,$4D,$AF,$FA,$63,$1F,$33,$E4,$E4,$FC,$A4,$62,$3F,$82,$47,$E5
	.dc.b	$E7,$C9,$F4,$BD,$F0,$FF,$50,$B6,$8D,$3E,$F7,$D5,$7C,$F8,$FB,$6D
	.dc.b	$7F,$0A,$D7,$FC,$5B,$B2,$31,$1F,$33,$5F,$90,$F9,$F4,$F9,$8F,$99
	.dc.b	$FF,$6D,$FD,$5F,$F2,$5F,$97,$B1,$FE,$B6,$6C,$9A,$7E,$4F,$EB,$7F
	.dc.b	$D7,$7F,$51,$64,$D2,$3E,$12,$1E,$12,$9F,$EC,$4D,$7E,$3B,$F4,$F5
	.dc.b	$F8,$ED,$3E,$39,$75,$FF,$79,$FD,$5F,$F7,$9F,$A9,$BD,$DF,$EC,$E6
	.dc.b	$D1,$A7,$C2,$7E,$E6,$DF,$19,$F2,$10,$FF,$05,$61,$F9,$7A,$FC,$64
	.dc.b	$7F,$B6,$8C,$9F,$C5,$5C,$B9,$FA,$D7,$27,$E7,$DB,$1F,$C6,$DB,$BF
	.dc.b	$38,$FF,$72,$C6,$1F,$8A,$7D,$BE,$3F,$C7,$E3,$E7,$8F,$4C,$73,$BE
	.dc.b	$C4,$8C,$12,$FB,$2B,$EC,$F9,$E7,$84,$FC,$58,$F0,$37,$93,$F3,$D1
	.dc.b	$E6,$9D,$BE,$9A,$E6,$E4,$FE,$71,$83,$EA,$FF,$80,$36,$33,$18,$3F
	.dc.b	$82,$5F,$68,$F0,$9F,$09,$C4,$F1,$C3,$D6,$D9,$F9,$4D,$7E,$B3,$5F
	.dc.b	$F4,$B7,$C2,$B8,$90,$F9,$6F,$A5,$EC,$AB,$53,$93,$F0,$11,$9A,$E3
	.dc.b	$5C,$2B,$6F,$C7,$BF,$AD,$7E,$A3,$F1,$83,$EC,$1F,$15,$FC,$8D,$B1
	.dc.b	$FD,$7D,$B9,$7D,$86,$C7,$F3,$EE,$1F,$4F,$E8,$DF,$D1,$5B,$19,$11
	.dc.b	$96,$9F,$D1,$BE,$C3,$E0,$FC,$DF,$F4,$D6,$C6,$44,$65,$A7,$F4,$EF
	.dc.b	$A3,$F9,$3F,$93,$7F,$55,$6C,$64,$46,$5A,$7F,$56,$FB,$5F,$A3,$FB
	.dc.b	$17,$F5,$D6,$C6,$44,$65,$A7,$F5,$EF,$B9,$FB,$3F,$A5,$7F,$65,$6C
	.dc.b	$64,$46,$5A,$7E,$BF,$5F,$9F,$F7,$9F,$37,$83,$F0,$6D,$93,$4F,$EE
	.dc.b	$5F,$8C,$7E,$23,$E9,$6C,$9A,$47,$F7,$9F,$AB,$FF,$78,$FB,$6F,$D8
	.dc.b	$B6,$33,$E7,$FE,$4A,$72,$D8,$F3,$63,$F8,$52,$3F,$85,$7F,$BB,$F7
	.dc.b	$7E,$6B,$F2,$9F,$F0,$33,$67,$FC,$06,$D7,$FC,$1A,$D7,$FD,$EB,$F8
	.dc.b	$F3,$C1,$FD,$1F,$D8,$FF,$13,$F4,$53,$9A,$4E,$63,$0D,$AE,$33,$FC
	.dc.b	$96,$B8,$FE,$07,$FD,$1A,$31,$68,$C4,$7E,$CA,$76,$FF,$4A,$37,$1A
	.dc.b	$7F,$8F,$F6,$7E,$B6,$FF,$91,$9F,$0E,$24,$7D,$92,$3F,$93,$FA,$7F
	.dc.b	$C9,$11,$C1,$E4,$FB,$DF,$E0,$7F,$A9,$FB,$5B,$19,$11,$96,$9E,$CF
	.dc.b	$1B,$F8,$5F,$D8,$FC,$56,$C6,$44,$65,$A7,$DE,$3D,$25,$C3,$D2,$FC
	.dc.b	$5D,$3F,$D1,$3F,$5D,$F4,$7C,$EF,$E3,$2D,$93,$4F,$8D,$52,$3F,$A7
	.dc.b	$7E,$A4,$FC,$53,$FC,$A1,$1F,$C8,$A3,$0F,$F9,$64,$38,$F7,$3F,$A4
	.dc.b	$B6,$3E,$DF,$9D,$FC,$AB,$EC,$7F,$08,$D8,$C8,$8C,$B4,$FE,$C1,$DA
	.dc.b	$F9,$3F,$1C,$7E,$C1,$B1,$98,$F2,$37,$0F,$C9,$7E,$21,$FC,$EB,$EC
	.dc.b	$6D,$1A,$47,$E0,$E7,$FC,$E1,$A7,$F3,$D3,$FB,$E6,$7F,$0B,$CD,$F7
	.dc.b	$CE,$E7,$F9,$CF,$A3,$FD,$54,$9E,$77,$F7,$3E,$A7,$E8,$2F,$90,$FF
	.dc.b	$38,$36,$FF,$CD,$ED,$E2,$FD,$73,$63,$F4,$BF,$89,$7F,$54,$F3,$7F
	.dc.b	$5C,$D8,$C8,$8C,$B4,$FD,$87,$D3,$73,$7F,$58,$D5,$8A,$30,$DC,$7F
	.dc.b	$B8,$E5,$FD,$63,$3F,$6F,$77,$EB,$B4,$F3,$FC,$BF,$B0,$99,$7B,$1E
	.dc.b	$33,$B8,$FE,$F1,$7F,$E1,$7F,$90,$7E,$19,$FE,$4B,$9F,$F6,$04,$63
	.dc.b	$DE,$FF,$60,$36,$33,$3B,$C7,$F8,$87,$0F,$F3,$19,$FE,$16,$D7,$F4
	.dc.b	$CD,$7F,$8E,$B7,$5F,$E3,$7F,$15,$FE,$A3,$EE,$34,$95,$E9,$3E,$FE
	.dc.b	$7F,$0C,$FB,$DF,$25,$F9,$3C,$7F,$93,$36,$7F,$90,$35,$FE,$38,$D7
	.dc.b	$F9,$7F,$E2,$1F,$E1,$0F,$C6,$FF,$95,$36,$3F,$92,$B5,$FE,$00,$D7
	.dc.b	$F9,$09,$F3,$37,$F3,$3F,$CE,$7F,$3C,$7C,$CF,$FB,$CF,$5F,$F2,$B5
	.dc.b	$27,$34,$CB,$F0,$6F,$F8,$8A,$75,$7C,$63,$15,$C2,$B7,$1F,$D2,$3E
	.dc.b	$0F,$FD,$89,$87,$F3,$C6,$C7,$F1,$76,$BB,$3F,$D9,$11,$E9,$9F,$E2
	.dc.b	$91,$FD,$07,$EF,$FF,$A2,$3F,$E8,$F9,$FE,$95,$18,$E3,$F0,$C2,$30
	.dc.b	$EB,$CB,$FF,$4C,$3F,$E3,$6F,$C1,$7E,$ED,$FB,$E3,$1C,$3E,$D8,$7C
	.dc.b	$12,$B7,$97,$BA,$31,$5B,$BF,$EC,$7C,$3F,$E8,$D7,$FA,$DB,$70,$FE
	.dc.b	$6C,$D7,$FB,$6B,$5F,$F2,$11,$F6,$8F,$C6,$EF,$F4,$DF,$29,$E7,$3F
	.dc.b	$CA,$B7,$E2,$70,$7A,$31,$F3,$DF,$0C,$7E,$AC,$F6,$27,$47,$EE,$A3
	.dc.b	$E7,$C7,$84,$83,$E4,$7F,$47,$F2,$5F,$E6,$4F,$B7,$F2,$4E,$75,$4E
	.dc.b	$6F,$FD,$45,$7C,$29,$7C,$23,$0D,$BE,$33,$AA,$E3,$18,$9D,$5F,$5A
	.dc.b	$FF,$3A,$DF,$2B,$FF,$76,$C7,$FA,$6B,$07,$37,$F8,$56,$3F,$DE,$86
	.dc.b	$5C,$3C,$0C,$90,$C1,$F3,$B3,$FC,$3B,$F8,$16,$E7,$F5,$23,$19,$FF
	.dc.b	$87,$5E,$27,$FA,$C8,$F0,$6B,$FE,$AE,$7F,$D7,$07,$F8,$9F,$6F,$F4
	.dc.b	$7F,$CC,$7E,$6B,$A1,$18,$C3,$F1,$74,$AD,$4E,$63,$CD,$87,$85,$FC
	.dc.b	$13,$78,$3F,$0D,$1E,$63,$FC,$58,$7E,$B7,$87,$FD,$3F,$E5,$7E,$AB
	.dc.b	$8F,$F5,$ED,$8C,$E5,$FA,$BA,$4F,$E3,$37,$F9,$FA,$FB,$5B,$3F,$00
	.dc.b	$87,$F0,$31,$F3,$9E,$FE,$6F,$F5,$7B,$63,$38,$6E,$3F,$86,$DF,$F1
	.dc.b	$29,$F4,$9F,$C7,$7A,$3E,$04,$66,$37,$D1,$FC,$71,$F3,$71,$FE,$01
	.dc.b	$FA,$87,$DF,$29,$FC,$B0,$7E,$1F,$FC,$C9,$A7,$F3,$22,$9B,$7F,$9F
	.dc.b	$F9,$7F,$77,$2F,$B2,$DE,$2F,$F0,$B5,$68,$F8,$D7,$E3,$7F,$06,$78
	.dc.b	$CA,$7F,$8F,$78,$BA,$72,$7E,$4A,$3F,$CC,$F1,$83,$3D,$BF,$ED,$8F
	.dc.b	$B3,$FE,$C7,$26,$91,$FF,$6E,$8F,$03,$FB,$63,$83,$1E,$07,$F7,$37
	.dc.b	$11,$A7,$8B,$5F,$CF,$9E,$83,$E5,$93,$4F,$F1,$0F,$83,$F1,$7C,$63
	.dc.b	$49,$3F,$13,$86,$3F,$88,$3F,$57,$5F,$E1,$37,$CC,$7F,$86,$3C,$4F
	.dc.b	$F1,$53,$FE,$95,$6B,$FE,$20,$F1,$3E,$4F,$8D,$FE,$FC,$FA,$5F,$F1
	.dc.b	$29,$27,$FC,$52,$FF,$8C,$27,$CC,$48,$EF,$7F,$E3,$A8,$FD,$D4,$7B
	.dc.b	$C9,$D3,$FA,$07,$ED,$79,$3F,$E6,$8D,$7F,$C8,$F1,$F3,$CF,$F2,$6B
	.dc.b	$83,$FC,$AB,$F3,$F2,$3F,$D4,$7E,$61,$46,$5F,$B3,$E7,$7F,$98,$7C
	.dc.b	$6F,$86,$D8,$FF,$34,$F4,$F1,$9C,$B6,$3F,$CE,$3F,$67,$FE,$70,$23
	.dc.b	$83,$FC,$C3,$3E,$82,$38,$3F,$2D,$9F,$BD,$FC,$0B,$5F,$F6,$4F,$A5
	.dc.b	$E9,$FB,$5F,$F4,$91,$23,$FD,$7F,$19,$7F,$D3,$C7,$FB,$B4,$86,$0F
	.dc.b	$F5,$37,$D8,$FF,$A9,$48,$E0,$FF,$59,$7E,$A3,$FE,$B1,$23,$83,$FD
	.dc.b	$A1,$F2,$DF,$E6,$3F,$2B,$FD,$B4,$D8,$CF,$A3,$F3,$E7,$2D,$8F,$11
	.dc.b	$FA,$8F,$85,$7C,$07,$E0,$9B,$1F,$CF,$5A,$FE,$64,$FD,$19,$F8,$6F
	.dc.b	$0B,$F8,$97,$3F,$FB,$73,$E9,$7F,$DD,$1E,$5F,$DF,$6D,$1A,$7F,$BA
	.dc.b	$D7,$D8,$52,$3F,$EE,$AE,$EF,$C4,$47,$FC,$00,$87,$E5,$94,$8F,$C4
	.dc.b	$F8,$2F,$C7,$47,$D0,$8F,$76,$9E,$D2,$3F,$D6,$7E,$19,$F0,$BF,$E1
	.dc.b	$28,$FE,$11,$18,$7F,$C5,$26,$23,$4F,$F1,$08,$D7,$A5,$F4,$BE,$23
	.dc.b	$F8,$C4,$FD,$66,$BF,$F0,$07,$8D,$18,$8F,$F4,$BC,$61,$FE,$3C,$7F
	.dc.b	$B9,$FF,$0E,$FE,$0B,$C5,$F5,$1B,$97,$EA,$DB,$27,$F0,$7F,$17,$FA
	.dc.b	$FF,$D2,$EC,$9A,$47,$F1,$7F,$1D,$FB,$5F,$DC,$FF,$16,$D8,$C8,$8C
	.dc.b	$B4,$FC,$78,$FC,$B7,$F1,$ED,$32,$E0,$FC,$BF,$C9,$FF,$03,$F8,$0B
	.dc.b	$26,$91,$FC,$FF,$CA,$FF,$09,$FD,$FF,$F3,$ED,$8C,$88,$CB,$4F,$CF
	.dc.b	$9F,$C4,$76,$FC,$DB,$80,$E1,$FE,$03,$D7,$FD,$6F,$C2,$FE,$05,$1F
	.dc.b	$D0,$F8,$DF,$2E,$F9,$8F,$E3,$9E,$5F,$F4,$F9,$CC,$62,$BF,$AA,$C6
	.dc.b	$1F,$E5,$07,$F9,$82,$B9,$38,$3F,$93,$60,$F8,$CF,$E6,$AD,$8F,$F1
	.dc.b	$83,$5F,$E9,$2D,$7F,$8D,$BE,$92,$76,$F0,$9E,$64,$9C,$1F,$E3,$BF
	.dc.b	$99,$76,$FF,$44,$6C,$66,$30,$7F,$48,$D3,$E2,$93,$0E,$F3,$FD,$72
	.dc.b	$B8,$B6,$B8,$BC,$C8,$64,$A7,$F7,$98,$E2,$FB,$30,$FE,$D6,$FF,$83
	.dc.b	$63,$2F,$F8,$09,$AF,$F4,$F6,$A6,$FC,$03,$83,$FB,$0B,$67,$FC,$B8
	.dc.b	$D7,$FB,$AB,$5F,$11,$FC,$8C,$FE,$D1,$F3,$7F,$E4,$A7,$FC,$B8,$7F
	.dc.b	$84,$5C,$7C,$1F,$F7,$E6,$C6,$67,$FC,$CE,$FF,$85,$FC,$9F,$17,$8B
	.dc.b	$FE,$04,$6C,$67,$0F,$F2,$3B,$FD,$AE,$32,$FE,$63,$2F,$8A,$FF,$83
	.dc.b	$BB,$BF,$1B,$E2,$7F,$46,$E3,$FE,$69,$3F,$D2,$9A,$FF,$15,$6B,$FD
	.dc.b	$DF,$E7,$7F,$D7,$19,$FF,$85,$9B,$1F,$E2,$D6,$BF,$E3,$26,$BF,$D1
	.dc.b	$B8,$7F,$09,$8F,$92,$3F,$8F,$FC,$F7,$F9,$99,$FE,$B6,$D7,$F9,$33
	.dc.b	$5D,$9F,$E6,$E3,$E1,$1F,$EC,$21,$9F,$C2,$7E,$26,$9F,$19,$CB,$6F
	.dc.b	$C7,$1F,$65,$FD,$EC,$FE,$07,$ED,$7E,$85,$7F,$85,$DB,$46,$9F,$E3
	.dc.b	$8F,$A0,$FC,$24,$7F,$94,$9A,$FF,$87,$DA,$9B,$7E,$28,$FC,$59,$FD
	.dc.b	$F3,$E8,$7F,$CE,$1B,$9A,$48,$FF,$1C,$C7,$FB,$AF,$E1,$BF,$C0,$3F
	.dc.b	$13,$F1,$F8,$93,$9A,$FF,$03,$CE,$61,$F4,$7D,$89,$3D,$1F,$0F,$95
	.dc.b	$FA,$98,$FF,$07,$B5,$FE,$B4,$D7,$FA,$CB,$57,$D3,$3E,$9B,$FC,$6D
	.dc.b	$1F,$E7,$68,$C3,$FE,$E6,$31,$1A,$7F,$C8,$D1,$AF,$57,$FC,$CA,$B1
	.dc.b	$D4,$F6,$6C,$66,$3A,$1C,$A3,$7A,$FF,$AF,$2B,$FD,$B2,$49,$FF,$43
	.dc.b	$46,$5F,$F4,$47,$C6,$FE,$ED,$FF,$3A,$B5,$CF,$37,$F9,$A2,$31,$1F
	.dc.b	$E9,$26,$BF,$EF,$C8,$D1,$1F,$F7,$44,$69,$34,$91,$D7,$E1,$BD,$73
	.dc.b	$FE,$69,$94,$62,$38,$B6,$33,$7D,$BD,$CF,$F3,$86,$3E,$A6,$DF,$B9
	.dc.b	$AF,$A5,$AF,$76,$BE,$66,$D1,$A6,$B1,$FF,$74,$FC,$9F,$FA,$71,$B5
	.dc.b	$F3,$BC,$07,$4A,$B5,$18,$7E,$FF,$07,$FA,$1D,$B5,$B3,$FD,$79,$B7
	.dc.b	$E6,$9C,$1F,$8D,$9E,$0E,$B9,$2B,$F9,$16,$BF,$8E,$6B,$F9,$86,$E2
	.dc.b	$C3,$CF,$F9,$A7,$9B,$23,$49,$7F,$ED,$2E,$47,$C4,$FE,$8A,$1F,$81
	.dc.b	$FF,$B4,$E7,$3A,$7F,$A8,$E3,$15,$B6,$C6,$6B,$A4,$BE,$AB,$EA,$DF
	.dc.b	$66,$C6,$67,$F4,$47,$FA,$BE,$74,$E0,$3E,$6F,$AC,$FF,$5A,$7C,$B7
	.dc.b	$89,$B3,$E0,$6B,$CE,$91,$88,$E7,$0F,$AD,$FF,$5F,$7D,$47,$B1,$B3
	.dc.b	$EA,$6B,$CD,$AF,$FB,$17,$7F,$A1,$B7,$9E,$11,$8E,$5F,$A4,$48,$FF
	.dc.b	$7F,$76,$BD,$3F,$F6,$77,$E4,$5F,$CD,$5B,$3A,$6B,$FC,$DD,$AF,$F5
	.dc.b	$FE,$5F,$F6,$B7,$AD,$FD,$7D,$B1,$FD,$7D,$AF,$F5,$F6,$BE,$9E,$5F
	.dc.b	$CE,$E3,$E5,$8F,$F7,$01,$EA,$8C,$BF,$A3,$CF,$DD,$9F,$B6,$31,$5C
	.dc.b	$5E,$93,$F6,$78,$7F,$1B,$F1,$1F,$80,$24,$E7,$81,$1F,$B3,$C7,$BC
	.dc.b	$F8,$B1,$EC,$D8,$CB,$DB,$1F,$C9,$FC,$A7,$D4,$C7,$F3,$6D,$C4,$D3
	.dc.b	$1F,$61,$FE,$FD,$FC,$27,$F3,$6D,$C3,$F3,$2D,$7F,$14,$D7,$C0,$7C
	.dc.b	$02,$1E,$5D,$3D,$9F,$1E,$87,$CF,$7E,$A2,$30,$D8,$C9,$FA,$88,$CB
	.dc.b	$FE,$33,$7F,$8D,$F9,$C6,$92,$3E,$9E,$BF,$11,$FB,$17,$82,$6B,$F7
	.dc.b	$23,$C9,$F3,$8C,$FF,$17,$08,$C2,$E4,$E0,$E6,$77,$3A,$9F,$F3,$15
	.dc.b	$4A,$FF,$31,$7A,$5F,$CC,$BD,$2F,$F3,$9C,$FE,$16,$7C,$AD,$E2,$FB
	.dc.b	$B1,$FE,$61,$AE,$04,$7F,$59,$5C,$9F,$CA,$FE,$5D,$FB,$9F,$86,$F9
	.dc.b	$DE,$0F,$B6,$31,$3F,$C8,$0D,$F5,$AB,$71,$86,$BF,$67,$2F,$26,$FF
	.dc.b	$2F,$3F,$2E,$1D,$34,$ED,$8F,$3C,$3E,$31,$CC,$A7,$E0,$67,$83,$FC
	.dc.b	$BF,$E9,$FF,$5B,$8F,$F2,$D6,$C6,$4F,$E4,$EE,$1F,$21,$F8,$E3,$F2
	.dc.b	$26,$0F,$E5,$0B,$EC,$FF,$44,$D8,$77,$77,$9C,$9A,$F8,$EF,$C7,$CF
	.dc.b	$26,$BF,$CC,$29,$18,$EA,$46,$27,$F3,$27,$8C,$F2,$90,$FB,$E7,$E4
	.dc.b	$7F,$70,$FD,$FB,$E4,$31,$D9,$FC,$B1,$46,$B1,$FC,$8B,$63,$E5,$7D
	.dc.b	$AF,$EB,$7E,$B3,$E5,$36,$32,$23,$2D,$3E,$97,$DB,$7E,$D7,$C9,$FA
	.dc.b	$4D,$8C,$9F,$E5,$B2,$C7,$D7,$FB,$7F,$DE,$7E,$51,$64,$D2,$3F,$7F
	.dc.b	$E6,$3F,$81,$7E,$51,$F7,$DB,$19,$11,$96,$9F,$87,$FB,$BE,$ED,$3B
	.dc.b	$96,$BF,$15,$1A,$8F,$E7,$3F,$0A,$F8,$DF,$E2,$5F,$31,$FB,$A8,$C4
	.dc.b	$A3,$13,$A8,$FE,$0D,$3A,$9F,$CB,$7E,$23,$FC,$6F,$CE,$B5,$FD,$B0
	.dc.b	$FC,$E4,$65,$A6,$4F,$CA,$FE,$21,$FC,$8F,$F2,$8F,$D0,$36,$3F,$48
	.dc.b	$D7,$62,$3F,$88,$FE,$25,$FE,$02,$8F,$C4,$1F,$D5,$9C,$72,$EB,$A3
	.dc.b	$F1,$51,$89,$FD,$87,$C0,$7F,$33,$FC,$A3,$F6,$0D,$8C,$88,$CB,$4F
	.dc.b	$DC,$7E,$27,$FC,$EF,$F2,$BF,$DC,$36,$32,$23,$2D,$3F,$80,$79,$5F
	.dc.b	$CC,$1A,$7F,$2F,$2F,$95,$FC,$0E,$35,$1F,$AB,$3F,$52,$5F,$17,$F9
	.dc.b	$3B,$FA,$6E,$EF,$E7,$3D,$DF,$AC,$F1,$7F,$4F,$18,$3F,$C0,$FB,$7E
	.dc.b	$E3,$F1,$4F,$EA,$9F,$B8,$7F,$11,$6E,$1F,$C5,$5A,$FF,$0C,$6B,$FC
	.dc.b	$0F,$F0,$9F,$EB,$9F,$4F,$FC,$65,$B1,$FC,$75,$AF,$F1,$46,$BF,$C2
	.dc.b	$FE,$9F,$E8,$3F,$C6,$1A,$E4,$FA,$06,$4A,$7F,$6B,$E3,$FC,$16,$7F
	.dc.b	$7E,$3F,$97,$99,$CB,$FA,$DC,$EA,$3F,$98,$7D,$8F,$F7,$3F,$C0,$3F
	.dc.b	$98,$36,$32,$23,$2D,$3F,$98,$1F,$48,$FE,$39,$BF,$F5,$8F,$49,$FB
	.dc.b	$63,$F1,$0F,$F0,$07,$E7,$1F,$DB,$E7,$31,$8B,$FF,$16,$C6,$14,$EE
	.dc.b	$B9,$46,$0F,$E9,$1F,$87,$7F,$82,$FF,$2C,$FE,$90,$D8,$C8,$8C,$B4
	.dc.b	$FF,$00,$69,$FA,$9F,$C0,$FE,$41,$23,$3B,$7F,$85,$E3,$FA,$A7,$89
	.dc.b	$FE,$19,$FC,$56,$D1,$A4,$7F,$D1,$5F,$77,$FC,$33,$F3,$7F,$F4,$43
	.dc.b	$63,$35,$FE,$23,$9C,$B6,$3E,$6F,$07,$E7,$09,$EB,$FF,$8F,$B8,$3F
	.dc.b	$CC,$D3,$FE,$70,$6B,$FC,$08,$FF,$5A,$FC,$27,$F9,$63,$C3,$FA,$0D
	.dc.b	$8F,$EE,$1F,$89,$7F,$8C,$FF,$46,$FE,$E0,$D8,$C8,$8C,$B4,$FE,$F1
	.dc.b	$F8,$A7,$C5,$68,$DF,$FB,$D4,$66,$3F,$B9,$B5,$FC,$3E,$5F,$E0,$B2
	.dc.b	$65,$FE,$65,$F5,$B9,$4F,$AF,$87,$FB,$6E,$1F,$DE,$89,$1E,$F3,$F2
	.dc.b	$44,$D7,$E3,$DF,$FA,$9C,$7E,$34,$F8,$0F,$90,$DA,$F9,$13,$A9,$C9
	.dc.b	$FE,$69,$F7,$BF,$97,$74,$FF,$B0,$5B,$5F,$EC,$76,$BF,$EB,$76,$BF
	.dc.b	$EB,$67,$F3,$3E,$E7,$1E,$BF,$F6,$0C,$62,$79,$47,$F9,$FF,$0F,$F2
	.dc.b	$AC,$7E,$0B,$6F,$F3,$EE,$BF,$E3,$7F,$BF,$F2,$12,$7E,$36,$08,$F4
	.dc.b	$7E,$1B,$27,$2D,$E5,$FF,$1D,$E3,$DA,$7F,$96,$7E,$84,$FC,$17,$FC
	.dc.b	$20,$D8,$F9,$CD,$AF,$F3,$0B,$63,$E1,$67,$FE,$C8,$CE,$C7,$F8,$B9
	.dc.b	$B5,$FD,$5E,$3E,$24,$ED,$FE,$74,$FE,$0B,$5F,$29,$23,$FC,$7D,$F4
	.dc.b	$AF,$B3,$7C,$C7,$FA,$2E,$BA,$4E,$63,$18,$FF,$AE,$23,$16,$AD,$DF
	.dc.b	$28,$C5,$7D,$BF,$C4,$7C,$91,$F6,$5B,$1F,$E6,$76,$F0,$1D,$CF,$E4
	.dc.b	$72,$F5,$3E,$8B,$ED,$87,$83,$2F,$2C,$FF,$A8,$F2,$F2,$CE,$63,$C6
	.dc.b	$7C,$4D,$BE,$B3,$FA,$86,$BD,$DF,$F7,$77,$CC,$7C,$4F,$99,$FF,$57
	.dc.b	$71,$7B,$EB,$FD,$61,$39,$7F,$0D,$7F,$ED,$C8,$C4,$EE,$51,$89,$D6
	.dc.b	$2B,$EB,$18,$6D,$76,$9D,$3D,$8C,$46,$9F,$F5,$99,$FE,$AA,$E7,$68
	.dc.b	$CB,$FD,$5B,$3A,$3F,$DD,$C6,$E3,$4D,$8D,$CF,$26,$CE,$A7,$F0,$5A
	.dc.b	$7D,$F7,$E7,$13,$7C,$E3,$FE,$FE,$7B,$C4,$3F,$DB,$47,$73,$F8,$16
	.dc.b	$DF,$ED,$BF,$A5,$FC,$E4,$FE,$16,$51,$8E,$0F,$C3,$46,$27,$F1,$CD
	.dc.b	$3F,$05,$87,$96,$3C,$71,$FE,$E7,$6B,$E7,$3C,$67,$94,$FF,$76,$67
	.dc.b	$F0,$5F,$31,$93,$E5,$1C,$08,$7E,$C8,$F1,$16,$FD,$19,$7B,$6B,$F7
	.dc.b	$71,$FB,$99,$71,$7F,$05,$8F,$E0,$7C,$7F,$0E,$7F,$DD,$93,$96,$C6
	.dc.b	$5F,$09,$E4,$3D,$37,$F5,$DF,$2D,$FF,$B5,$79,$38,$47,$F2,$B6,$C6
	.dc.b	$63,$E7,$9F,$CC,$E3,$39,$FF,$23,$8C,$9F,$CD,$78,$B2,$47,$27,$8B
	.dc.b	$7F,$C4,$91,$F9,$67,$F3,$DE,$5F,$CA,$E1,$FE,$E9,$6D,$FF,$85,$A7
	.dc.b	$F9,$F0,$3F,$A4,$4D,$FC,$D2,$C1,$A4,$7F,$C4,$7D,$3F,$D5,$76,$FF
	.dc.b	$14,$36,$33,$D3,$F1,$32,$FE,$B1,$1E,$A3,$E6,$13,$83,$F0,$8F,$E1
	.dc.b	$F1,$FE,$D6,$FC,$06,$E1,$FE,$F1,$EA,$F9,$BD,$7F,$DC,$A3,$E0,$D7
	.dc.b	$C5,$29,$F6,$6B,$FD,$C3,$D9,$EF,$FA,$BF,$F0,$0C,$9A,$47,$FC,$1B
	.dc.b	$FA,$37,$F8,$30,$8E,$0F,$F2,$1F,$83,$F2,$5F,$AB,$A7,$F8,$7B,$1F
	.dc.b	$F0,$DC,$7C,$19,$D4,$79,$34,$F7,$CF,$E3,$9A,$B2,$78,$9F,$EC,$0F
	.dc.b	$CD,$FF,$C6,$32,$69,$1F,$E0,$9A,$FF,$B1,$BF,$1B,$FF,$20,$92,$33
	.dc.b	$E2,$FD,$F8,$FF,$39,$6D,$27,$FC,$96,$7C,$32,$BF,$E5,$4F,$CA,$E4
	.dc.b	$7F,$96,$1A,$FE,$27,$6F,$F3,$27,$27,$F9,$8F,$9B,$F1,$53,$FE,$68
	.dc.b	$F3,$3F,$CC,$B3,$98,$FC,$61,$FC,$07,$7F,$82,$FE,$0F,$AB,$F5,$CF
	.dc.b	$C3,$D5,$E7,$7E,$5F,$E9,$7F,$CF,$E4,$9F,$F4,$1B,$FE,$8D,$FC,$06
	.dc.b	$47,$B1,$AF,$FA,$57,$F2,$AF,$B0,$FF,$A6,$5B,$C8,$C1,$E9,$FE,$96
	.dc.b	$73,$18,$A4,$7F,$AA,$D7,$E8,$B4,$8F,$FA,$B4,$C9,$FE,$B7,$1F,$BA
	.dc.b	$7F,$AE,$DA,$65,$C1,$FE,$C4,$58,$D2,$3E,$AF,$6D,$ED,$FE,$EE,$7F
	.dc.b	$1D,$B7,$CA,$AF,$99,$C7,$F0,$1F,$99,$EE,$7F,$01,$FC,$23,$FD,$B5
	.dc.b	$06,$91,$FF,$73,$FC,$47,$FB,$98,$8E,$0F,$F7,$9E,$BF,$EF,$22,$38
	.dc.b	$3F,$1B,$DB,$FE,$C1,$23,$83,$F3,$3F,$0B,$FC,$4F,$F2,$CF,$9E,$D8
	.dc.b	$C8,$8C,$B4,$FD,$0F,$C3,$FF,$1B,$D3,$EB,$B6,$32,$7F,$7A,$2C,$7C
	.dc.b	$3C,$5D,$DF,$C8,$FE,$EB,$0D,$AF,$C3,$CE,$94,$E9,$A7,$EC,$BE,$2F
	.dc.b	$F9,$67,$E5,$7F,$88,$6C,$64,$46,$5A,$7E,$EB,$E3,$FF,$9A,$7E,$59
	.dc.b	$F9,$06,$C6,$44,$65,$A7,$F0,$2F,$91,$EE,$FF,$75,$E9,$46,$23,$F5
	.dc.b	$8D,$EB,$FD,$03,$65,$F2,$7F,$D1,$7E,$5F,$F4,$2D,$BF,$E1,$ED,$7B
	.dc.b	$35,$FD,$C7,$CB,$7F,$4D,$FB,$BF,$D4,$B6,$3F,$8B,$B5,$F3,$35,$FE
	.dc.b	$A9,$D9,$F9,$C7,$FA,$43,$79,$BF,$48,$48,$FE,$15,$C7,$F9,$FF,$0B
	.dc.b	$F8,$59,$FC,$B3,$5F,$DE,$7C,$C7,$F5,$FF,$C1,$3F,$76,$DB,$FE,$52
	.dc.b	$D7,$E4,$35,$FE,$25,$AF,$F5,$02,$3F,$D5,$76,$47,$F8,$0A,$3F,$5E
	.dc.b	$7E,$CB,$B7,$F6,$18,$7F,$0D,$8F,$EE,$B5,$FE,$4A,$7F,$8F,$18,$3F
	.dc.b	$C9,$BF,$34,$F9,$B7,$AB,$FB,$D9,$FE,$A1,$39,$8C,$47,$F7,$B8,$C3
	.dc.b	$FC,$20,$DC,$F2,$70,$7F,$51,$D7,$E4,$13,$1F,$E0,$D5,$CB,$F3,$9F
	.dc.b	$E4,$ED,$AE,$78,$7F,$82,$67,$38,$75,$7F,$87,$7C,$F7,$FB,$66,$AF
	.dc.b	$AD,$EF,$8C,$47,$F3,$26,$BE,$3C,$7F,$C3,$8F,$F9,$16,$32,$FF,$85
	.dc.b	$DA,$9A,$63,$E4,$9E,$27,$C4,$9F,$F0,$F4,$61,$F8,$26,$23,$4A,$35
	.dc.b	$E2,$FD,$98,$87,$F4,$8F,$A0,$7D,$07,$D9,$6C,$7F,$A1,$1A,$FF,$2F
	.dc.b	$6B,$FE,$79,$FA,$4D,$F2,$FF,$50,$6C,$66,$BF,$21,$C6,$9C,$CF,$C5
	.dc.b	$3C,$67,$F9,$C5,$7D,$1F,$F8,$EF,$F0,$3E,$8D,$9F,$03,$5E,$0D,$7F
	.dc.b	$C2,$12,$FA,$5F,$F2,$CB,$63,$FC,$64,$DE,$AF,$44,$FE,$7F,$AB,$F5
	.dc.b	$FC,$FF,$89,$BF,$F1,$9E,$5E,$9F,$33,$FC,$A7,$8F,$EE,$E7,$E0,$46
	.dc.b	$0F,$D1,$5F,$2E,$3F,$C0,$CF,$BC,$FD,$EB,$87,$FC,$B1,$F6,$CF,$A7
	.dc.b	$FF,$2D,$79,$5F,$D7,$63,$12,$8C,$4E,$A3,$FC,$03,$3A,$9F,$EC,$9F
	.dc.b	$52,$F7,$7F,$06,$DA,$FC,$38,$71,$CB,$C2,$7F,$13,$97,$A6,$BE,$A9
	.dc.b	$F3,$3A,$FE,$59,$1F,$96,$FC,$EF,$A1,$FF,$39,$7D,$DF,$E0,$DB,$3F
	.dc.b	$87,$6B,$F7,$DA,$FD,$8F,$AC,$F9,$A4,$7F,$87,$F0,$23,$E8,$FD,$6F
	.dc.b	$F9,$FF,$F2,$3F,$C8,$36,$BF,$28,$D7,$F1,$8D,$7F,$D2,$1E,$07,$FB
	.dc.b	$6C,$8F,$F8,$AB,$3F,$F6,$E4,$7F,$83,$65,$F5,$3F,$E8,$DF,$8F,$C3
	.dc.b	$6F,$FC,$63,$3A,$7F,$DE,$67,$F9,$AD,$C4,$F3,$E3,$E1,$39,$6B,$FE
	.dc.b	$99,$9F,$BF,$3B,$8E,$67,$FA,$9D,$FF,$79,$11,$F4,$73,$FF,$BE,$E3
	.dc.b	$FC,$49,$F2,$2F,$B1,$FE,$68,$DA,$FF,$7F,$B5,$DC,$68,$8E,$9A,$F4
	.dc.b	$F9,$1F,$1B,$FE,$AF,$94,$62,$11,$8A,$EF,$39,$87,$E1,$5F,$EB,$7F
	.dc.b	$93,$F3,$B6,$FF,$CB,$73,$A5,$3A,$69,$F0,$3A,$9F,$65,$FE,$43,$E0
	.dc.b	$68,$9C,$D6,$EB,$53,$97,$EF,$7D,$A7,$FB,$25,$B3,$C0,$FB,$51,$97
	.dc.b	$EF,$B5,$FF,$54,$62,$FB,$CC,$47,$C5,$6B,$F1,$1B,$1B,$8E,$C7,$FB
	.dc.b	$8F,$EC,$3C,$B8,$9A,$48,$F0,$99,$2B,$E8,$23,$FE,$BB,$F0,$3F,$B7
	.dc.b	$E1,$FE,$E5,$FD,$27,$FD,$0B,$E6,$A3,$15,$B8,$C2,$9D,$CE,$A3,$FD
	.dc.b	$DD,$5F,$DE,$BF,$25,$FB,$11,$8B,$46,$23,$FD,$85,$18,$7E,$C9,$B8
	.dc.b	$D4,$FF,$81,$3E,$E3,$E4,$B6,$7E,$2B,$75,$F4,$CE,$AF,$F9,$2F,$CE
	.dc.b	$7B,$39,$D8,$23,$0D,$3D,$D1,$A8,$F9,$FC,$52,$7F,$18,$D7,$FD,$C9
	.dc.b	$C5,$F8,$CD,$FD,$31,$DA,$31,$3B,$8C,$71,$7E,$46,$3F,$C1,$FF,$8B
	.dc.b	$7E,$C7,$C9,$BD,$1E,$66,$DF,$F6,$03,$FC,$47,$3B,$7C,$64,$3C,$77
	.dc.b	$D5,$C1,$F0,$3E,$F1,$F7,$9F,$0B,$F1,$2F,$F5,$34,$62,$51,$89,$D4
	.dc.b	$7F,$B2,$A7,$53,$F8,$6C,$BE,$AE,$FF,$B0,$71,$5F,$0B,$8B,$FC,$77
	.dc.b	$1F,$7E,$1A,$FC,$47,$CA,$63,$2F,$43,$EA,$DB,$96,$1F,$CA,$77,$FC
	.dc.b	$AF,$6F,$D5,$AF,$CA,$35,$FC,$73,$5F,$C6,$FD,$F7,$E0,$3D,$3F,$CE
	.dc.b	$B6,$3F,$38,$D7,$F2,$CD,$7E,$BE,$1F,$71,$F3,$37,$2F,$54,$79,$F3
	.dc.b	$FD,$4F,$DF,$FE,$3B,$EE,$AC,$9A,$47,$E0,$D7,$27,$BD,$7E,$1D,$EE
	.dc.b	$D8,$D4,$60,$E3,$AF,$C8,$78,$9F,$80,$DB,$FC,$D7,$C3,$F5,$89,$1F
	.dc.b	$31,$FA,$95,$D0,$95,$CE,$7B,$1F,$B9,$9E,$37,$FC,$19,$B1,$FA,$C3
	.dc.b	$FD,$B9,$E2,$7F,$1C,$F8,$8F,$EE,$91,$FC,$68,$FF,$6E,$46,$4F,$E4
	.dc.b	$38,$78,$34,$F9,$98,$F8,$63,$DF,$7C,$B8,$3F,$DF,$8D,$C3,$C0,$74
	.dc.b	$3F,$74,$78,$DD,$3F,$C4,$7F,$03,$FC,$B9,$1F,$C6,$BF,$0A,$FC,$04
	.dc.b	$C3,$C6,$7E,$4B,$8B,$F1,$9A,$FF,$A5,$9A,$7F,$A3,$F1,$FE,$52,$E3
	.dc.b	$CB,$FE,$1F,$F8,$AF,$F2,$C6,$E5,$FC,$AD,$AF,$F2,$56,$BF,$68,$FC
	.dc.b	$61,$F7,$B0,$F1,$EF,$F8,$3C,$BC,$CF,$F3,$8F,$BD,$FE,$3B,$FA,$2D
	.dc.b	$B3,$48,$FF,$3B,$FC,$23,$F9,$1F,$E5,$7F,$D0,$1B,$19,$11,$96,$9F
	.dc.b	$D0,$FD,$EF,$E5,$7F,$94,$FF,$48,$6C,$64,$46,$5A,$7F,$51,$FC,$03
	.dc.b	$FB,$92,$FC,$A7,$ED,$F1,$27,$31,$86,$CF,$1D,$BF,$9D,$CE,$7D,$CF
	.dc.b	$E7,$BF,$91,$DA,$34,$8F,$CB,$DF,$E5,$5F,$F1,$7C,$1F,$53,$FD,$1B
	.dc.b	$F7,$6F,$03,$6F,$D0,$D7,$FA,$FC,$F0,$53,$E7,$F3,$BE,$1C,$69,$AB
	.dc.b	$4F,$74,$61,$B1,$F1,$8F,$C0,$90,$FC,$21,$1F,$8D,$E1,$FF,$5C,$FC
	.dc.b	$77,$E3,$36,$7E,$73,$5F,$3B,$5F,$16,$0F,$C2,$3F,$B2,$FE,$83,$F3
	.dc.b	$DB,$3F,$5D,$AF,$C0,$6B,$CB,$C9,$FE,$DB,$F4,$FF,$5D,$B1,$F7,$DA
	.dc.b	$FC,$86,$BE,$C3,$F8,$79,$7C,$0F,$E0,$9C,$1F,$58,$FE,$D5,$A7,$E1
	.dc.b	$7D,$3F,$EF,$7F,$15,$F8,$56,$DF,$E3,$5A,$FD,$66,$BF,$03,$F1,$9F
	.dc.b	$F0,$1F,$EA,$3F,$8D,$6C,$7E,$55,$AF,$DE,$6B,$F2,$3D,$7E,$78,$7E
	.dc.b	$5D,$BB,$FF,$98,$25,$8F,$E6,$67,$C3,$1E,$E2,$67,$EE,$7F,$49,$F4
	.dc.b	$FF,$E1,$6F,$C7,$BF,$40,$D8,$C8,$8C,$B4,$FC,$E1,$0F,$C4,$1D,$BC
	.dc.b	$1F,$E0,$EF,$EC,$7C,$FF,$EA,$CD,$3F,$AB,$17,$CB,$FD,$84,$6A,$39
	.dc.b	$E1,$E7,$7F,$29,$A7,$ED,$FF,$1C,$FF,$16,$FC,$6F,$ED,$DB,$3F,$BF
	.dc.b	$6B,$FA,$86,$BF,$7F,$1F,$2B,$D9,$AE,$7B,$3F,$90,$BE,$D3,$F4,$5B
	.dc.b	$BD,$DF,$F1,$EF,$E7,$7F,$C1,$DB,$3F,$C3,$DA,$FE,$E1,$AF,$E8,$BC
	.dc.b	$EF,$F2,$2F,$E0,$1F,$C3,$DB,$1F,$C5,$DA,$FF,$00,$6B,$E5,$E7,$FC
	.dc.b	$B3,$E5,$3E,$86,$3E,$87,$D8,$D7,$81,$EB,$D7,$F9,$07,$C1,$FF,$95
	.dc.b	$BF,$54,$FE,$40,$D9,$FE,$50,$D7,$F8,$8B,$5F,$DD,$FC,$37,$F9,$A3
	.dc.b	$53,$49,$1E,$13,$CB,$8F,$A1,$F2,$BF,$AE,$97,$C5,$FF,$9A,$3E,$F7
	.dc.b	$F9,$93,$6B,$F9,$D3,$5F,$E4,$6D,$7D,$A7,$43,$E2,$71,$FE,$25,$FE
	.dc.b	$23,$F1,$BF,$E7,$4F,$C0,$7F,$92,$47,$F2,$C8,$C6,$BF,$D1,$DB,$19
	.dc.b	$13,$C1,$EE,$49,$7E,$40,$F2,$BF,$D0,$3F,$57,$B9,$CC,$61,$FF,$52
	.dc.b	$4E,$E9,$5B,$8C,$3C,$74,$FF,$3A,$6B,$7D,$FF,$F6,$C6,$1F,$D7,$5B
	.dc.b	$87,$F4,$66,$BF,$CC,$3E,$4B,$FD,$1D,$3F,$D8,$DB,$1F,$D3,$1A,$EC
	.dc.b	$FF,$4C,$47,$CA,$E8,$FD,$2B,$FD,$9E,$BA,$B6,$7F,$DC,$AF,$8D,$A7
	.dc.b	$8C,$A7,$C8,$3F,$06,$75,$E4,$FC,$37,$C4,$7F,$BD,$E1,$F6,$DF,$EB
	.dc.b	$68,$C3,$7C,$17,$C5,$7F,$AB,$A3,$0D,$D3,$FD,$C7,$3A,$BE,$71,$89
	.dc.b	$D4,$7F,$AD,$AF,$AC,$E6,$36,$7F,$9D,$0F,$F7,$19,$0F,$F8,$00,$53
	.dc.b	$FD,$14,$53,$81,$F0,$CF,$B0,$7F,$B6,$0F,$88,$60,$FC,$61,$FE,$80
	.dc.b	$39,$14,$EE,$7D,$91,$F2,$7F,$85,$D7,$FD,$9E,$D8,$CE,$7B,$8E,$CF
	.dc.b	$8C,$91,$E2,$7F,$27,$C3,$FE,$D8,$F2,$BF,$24,$D8,$CC,$FF,$79,$6C
	.dc.b	$FA,$39,$7F,$DA,$93,$E8,$6B,$E8,$3F,$55,$F5,$9F,$9B,$E6,$FF,$72
	.dc.b	$B6,$0D,$3A,$71,$7F,$8E,$23,$91,$F3,$0F,$9E,$60,$F8,$C6,$08,$7F
	.dc.b	$84,$39,$FF,$DE,$7F,$63,$FC,$32,$0D,$23,$CF,$0F,$4B,$C8,$EA,$78
	.dc.b	$48,$60,$CE,$5F,$E4,$BE,$AF,$F4,$3C,$7F,$8E,$1A,$FF,$28,$33,$1B
	.dc.b	$6F,$33,$AD,$0F,$99,$FE,$6A,$D3,$F9,$A9,$4F,$E6,$A5,$3C,$B8,$7A
	.dc.b	$9B,$9F,$CB,$7D,$91,$97,$D4,$D2,$46,$E3,$8F,$3F,$AA,$72,$D9,$D4
	.dc.b	$FC,$2C,$3F,$D6,$7F,$27,$FD,$76,$0D,$23,$FD,$9C,$7E,$69,$FD,$AD
	.dc.b	$A6,$5C,$1F,$8D,$AE,$F8,$78,$A7,$18,$78,$C8,$E0,$87,$F8,$14,$7D
	.dc.b	$5F,$F8,$25,$A6,$5C,$1C,$1F,$8F,$5F,$0F,$4F,$F4,$47,$8F,$F5,$BE
	.dc.b	$67,$FC,$33,$46,$91,$FC,$65,$7A,$09,$87,$9A,$32,$D7,$FC,$FF,$E5
	.dc.b	$FE,$0B,$F1,$EF,$F1,$89,$23,$FC,$6A,$F9,$5F,$AA,$7F,$90,$47,$E4
	.dc.b	$14,$8F,$FA,$8F,$BB,$F1,$B1,$FE,$53,$87,$CD,$D2,$3F,$E5,$53,$27
	.dc.b	$F9,$7C,$EA,$58,$FF,$32,$38,$3F,$CC,$45,$3C,$A7,$B8,$F4,$1F,$90
	.dc.b	$C7,$E6,$70,$BD,$3F,$D0,$FE,$3F,$FE,$7D,$A3,$48,$FF,$A1,$C7,$E7
	.dc.b	$FF,$E8,$B6,$99,$70,$7F,$A6,$17,$E3,$D4,$8F,$FA,$68,$C9,$E7,$21
	.dc.b	$FE,$AD,$1F,$8C,$FF,$AC,$5A,$7F,$B3,$FD,$CF,$DC,$C6,$4F,$F5,$EC
	.dc.b	$3F,$76,$A4,$7F,$D7,$E6,$4E,$E7,$C8,$3F,$D8,$3E,$0F,$F0,$4F,$D6
	.dc.b	$3F,$DA,$04,$8F,$F6,$F4,$65,$FF,$6C,$0F,$CC,$BF,$DB,$9C,$9F,$C3
	.dc.b	$1B,$19,$36,$FF,$BA,$70,$91,$FE,$ED,$6B,$FC,$53,$9D,$F1,$8F,$88
	.dc.b	$FE,$2F,$FA,$37,$FB,$E4,$95,$FE,$FD,$7F,$DD,$F2,$F8,$BF,$E3,$BF
	.dc.b	$A5,$61,$B5,$FC,$80,$FF,$6C,$4E,$D4,$ED,$A7,$C6,$3F,$15,$9F,$EA
	.dc.b	$5F,$90,$7F,$37,$C7,$CF,$F1,$BF,$CB,$3F,$25,$F1,$B6,$7C,$13,$A5
	.dc.b	$3A,$69,$AC,$3E,$0B,$FD,$C8,$C3,$FD,$C9,$A6,$4F,$90,$F0,$F2,$FF
	.dc.b	$43,$E0,$FF,$19,$9E,$9F,$23,$FD,$1B,$0F,$9A,$DB,$34,$FE,$97,$2F
	.dc.b	$92,$7B,$3E,$18,$F9,$51,$87,$C4,$7F,$4C,$71,$8F,$F5,$66,$C7,$CB
	.dc.b	$C7,$C8,$F2,$AF,$EA,$7E,$96,$DB,$1F,$5C,$F1,$38,$78,$1F,$E3,$13
	.dc.b	$E0,$FC,$9F,$AC,$8F,$F6,$2C,$91,$EC,$BF,$0C,$FD,$AC,$FF,$1F,$3C
	.dc.b	$23,$F1,$6D,$7D,$1D,$DF,$CB,$1F,$9A,$D7,$3B,$FE,$16,$BF,$33,$F2
	.dc.b	$9F,$CE,$67,$EE,$35,$FA,$ED,$7F,$26,$D7,$F1,$E7,$F3,$18,$FD,$1F
	.dc.b	$CA,$FF,$78,$E8,$FC,$FB,$67,$F1,$0D,$7F,$06,$D7,$F3,$FF,$2F,$F6
	.dc.b	$5A,$7D,$8A,$FD,$7C,$65,$B1,$F7,$0F,$19,$FC,$21,$FD,$AF,$AB,$F8
	.dc.b	$92,$47,$CF,$7F,$13,$F1,$7F,$42,$3F,$CA,$21,$C4,$FF,$86,$9B,$1F
	.dc.b	$7D,$AF,$F6,$4C,$BF,$60,$FF,$89,$BC,$E7,$CC,$7F,$3B,$87,$CD,$7E
	.dc.b	$1E,$73,$E2,$FF,$87,$A3,$15,$B6,$C6,$6B,$EC,$9F,$72,$FB,$F0,$FF
	.dc.b	$87,$72,$FE,$27,$F3,$5F,$17,$8C,$D2,$5F,$CB,$8F,$1B,$F9,$53,$F8
	.dc.b	$7E,$3F,$99,$69,$9A,$FE,$3F,$19,$7F,$90,$FC,$E7,$ED,$49,$3F,$9B
	.dc.b	$7F,$71,$F3,$BF,$E3,$AF,$D3,$7F,$90,$B6,$7F,$85,$35,$FE,$0A,$D7
	.dc.b	$FC,$83,$F0,$C8,$FF,$21,$FD,$0F,$F9,$32,$31,$28,$C4,$EA,$3F,$69
	.dc.b	$3A,$9F,$F2,$9E,$8E,$76,$DC,$6F,$F7,$53,$98,$C4,$7F,$33,$8C,$57
	.dc.b	$2A,$E2,$FF,$97,$BD,$F7,$93,$F2,$6D,$18,$7E,$71,$88,$D4,$67,$B3
	.dc.b	$D3,$3A,$CB,$DB,$18,$24,$79,$B0,$F4,$C7,$8F,$2E,$B9,$7A,$63,$07
	.dc.b	$97,$E4,$3F,$93,$E2,$69,$30,$FF,$3A,$C7,$FB,$45,$FE,$38,$D7,$FD
	.dc.b	$88,$63,$D1,$F5,$BE,$E7,$FD,$9D,$F0,$3F,$ED,$28,$FF,$64,$99,$BF
	.dc.b	$F6,$7B,$4C,$D7,$FB,$41,$FF,$69,$57,$2E,$9F,$F6,$95,$F3,$23,$82
	.dc.b	$4F,$0B,$EB,$7C,$E3,$B5,$70,$74,$78,$2F,$B3,$88,$D3,$C2,$FA,$B4
	.dc.b	$D3,$8A,$EC,$4A,$E7,$3B,$24,$66,$FC,$B3,$A8,$C4,$F8,$E7,$24,$70
	.dc.b	$4B,$F2,$CF,$06,$D7,$32,$57,$18,$CB,$6F,$D1,$3A,$BE,$91,$89,$D4
	.dc.b	$7B,$88,$7A,$A7,$85,$74,$8E,$26,$AF,$C9,$7C,$E3,$27,$1B,$EA,$D3
	.dc.b	$33,$C1,$E5,$7D,$AB,$95,$F7,$23,$82,$4F,$0B,$EB,$7C,$E3,$B5,$70
	.dc.b	$74,$78,$2F,$B3,$88,$D3,$C2,$FA,$B4,$D3,$8A,$EC,$4A,$E7,$3B,$24
	.dc.b	$66,$FC,$B3,$A8,$C4,$F8,$E7,$24,$70,$4B,$F2,$CF,$06,$D7,$32,$57
	.dc.b	$18,$CB,$6F,$D1,$3A,$BE,$91,$89,$D4,$7B,$88,$7A,$A7,$85,$74,$8E
	.dc.b	$26,$AF,$C9,$7C,$E3,$27,$1B,$EA,$D3,$33,$C1,$E5,$7D,$AB,$95,$F7
	.dc.b	$23,$82,$4F,$0B,$EB,$7C,$E3,$B5,$70,$74,$78,$2F,$B3,$88,$D3,$C2
	.dc.b	$FA,$B4,$D3,$8A,$EC,$4A,$E7,$3B,$24,$66,$FC,$B3,$A8,$C4,$F8,$E7
	.dc.b	$24,$70,$4B,$F2,$CF,$06,$D7,$32,$57,$18,$CB,$6F,$D1,$3A,$BE,$91
	.dc.b	$89,$D4,$7B,$88,$7A,$A7,$85,$74,$8E,$26,$AF,$C9,$7C,$E3,$27,$1B
	.dc.b	$EA,$D3,$33,$C1,$E5,$7D,$AB,$95,$F7,$23,$82,$4F,$0B,$EB,$7C,$E3
	.dc.b	$B5,$70,$74,$78,$2F,$B3,$88,$D3,$C2,$FA,$B4,$D3,$8A,$EC,$4A,$E7
	.dc.b	$3B,$24,$66,$FC,$B3,$A8,$C4,$F8,$E7,$24,$70,$4B,$F2,$CF,$06,$D7
	.dc.b	$32,$57,$18,$CB,$6F,$D1,$3A,$BE,$91,$89,$D4,$7B,$88,$7A,$A7,$85
	.dc.b	$74,$8E,$26,$AF,$C9,$7C,$E3,$27,$1B,$EA,$D3,$33,$C1,$FA,$A7,$F8
	.dc.b	$36,$3F,$C2,$77,$FE,$0E,$9F,$F0,$94,$FF,$7A,$34,$7F,$8C,$63,$44
	.dc.b	$9E,$24,$8D,$35,$E2,$43,$FC,$3A,$43,$B1,$4F,$F0,$C1,$B3,$24,$32
	.dc.b	$6E,$7C,$EE,$0D,$38,$24,$F9,$E3,$CA,$43,$B1,$C4,$91,$C2,$7D,$51
	.dc.b	$DE,$79,$3C,$27,$4F,$33,$53,$C1,$F1,$CF,$07,$26,$CC,$C6,$DF,$0C
	.dc.b	$6D,$C9,$B3,$31,$B9,$F0,$C7,$8E,$78,$3A,$38,$46,$9F,$09,$B9,$E5
	.dc.b	$3C,$27,$53,$E0,$34,$72,$8D,$12,$78,$92,$34,$D7,$89,$0F,$09,$0E
	.dc.b	$C5,$3D,$06,$CC,$90,$C9,$B9,$F3,$B8,$34,$E0,$93,$E7,$8F,$29,$0E
	.dc.b	$C7,$12,$47,$09,$F5,$47,$79,$E4,$F0,$9D,$3C,$CD,$4F,$07,$C7,$3C
	.dc.b	$1C,$9B,$33,$1B,$7C,$31,$B7,$26,$CC,$C6,$E7,$C3,$1E,$39,$E0,$E8
	.dc.b	$E1,$1A,$7C,$26,$E7,$94,$F0,$9D,$4F,$80,$D1,$CA,$34,$49,$E2,$48
	.dc.b	$D3,$5E,$24,$3C,$24,$3B,$14,$F4,$1B,$32,$43,$26,$E7,$CE,$E0,$D3
	.dc.b	$82,$4F,$9E,$3C,$A4,$3B,$1C,$49,$1C,$27,$D5,$1D,$E7,$93,$C2,$74
	.dc.b	$F3,$35,$3C,$1F,$1C,$F0,$72,$6C,$CC,$6D,$F0,$C6,$DC,$9B,$33,$1B
	.dc.b	$9F,$0C,$78,$E7,$83,$A3,$84,$69,$F0,$9B,$9E,$53,$C2,$75,$3E,$03
	.dc.b	$47,$28,$D1,$27,$89,$23,$4D,$78,$90,$F0,$90,$EC,$53,$D0,$6C,$C9
	.dc.b	$0C,$9B,$9F,$3B,$83,$4E,$09,$3E,$78,$F2,$90,$EC,$71,$24,$70,$9F
	.dc.b	$54,$77,$9E,$4F,$09,$D3,$CC,$D4,$F0,$7C,$73,$C1,$C9,$B3,$31,$B7
	.dc.b	$C3,$1B,$72,$6C,$CC,$6E,$7C,$31,$E3,$9E,$0E,$8E,$11,$A5,$F3,$B0
	.dc.b	$00

;--------------------------------------------------------------------------------
;	FMOVE.B FPn,Dr
;	FMOVE.B FPn,<mem>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovebregto_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.B FPN,<EA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.B FPn,<ea>',13,10
;------------------------------------------------
;	d1	actual result
;	d3	fill=$00000000,$FFFFFFFF
;	d4	actual status
;	d6	fpcr=(XRN..DRP)<<4
;	d7	0=failed,1=successful
;	a0	expected result,expected status,...
;	a1	expected result
;	a4	expected status
;	a5	source handle,...
;	fp5	source
;------------------------------------------------
	lea.l	push_decompressed,a3
;decompress data
	move.l	a3,-(sp)
	pea.l	fmovebregto_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a3,a5			;source handle,...
@@:
	add.l	d0,(a3)+		;source handle
	tst.l	(a3)
	bpl	@b
	addq.l	#4,a3			;-1
	movea.l	a3,a0			;expected result,expected status,...
;
55:
	move.l	#0,d6		;fpcr=(XRN..DRP)<<4
66:
;FPn,Dr=fill
	moveq.l	#$00000000,d3		;fill=$00000000,$FFFFFFFF
33:
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	d3,d1			;fill
					;source
	fmove.b	fp5,d1			;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.B FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',Dr='
	puthex8	d3			;fill
	putmsg	' @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result,expected status,...
;
	not.l	d3			;fill
	bne	33b
;FPn,<mem>=fill
	moveq.l	#$00000000,d3		;fill=$00000000,$FFFFFFFF
33:
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	d3,-(sp)		;fill
					;source
	fmove.b	fp5,(sp)		;EXECUTE
	move.l	(sp)+,d1		;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.B FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',<mem>='
	puthex8	d3			;fill
	putmsg	' @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result,expected status,...
;
	not.l	d3			;fill
	bne	33b
;
	add.w	#1<<4,d6		;rprm++
	cmp.w	#176,d6		;fpcr=(XRN..DRP)<<4
	bls	66b
;
	addq.l	#4,a5			;source handle
	tst.l	(a5)			;source handle
	bpl	55b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmovebregto_data_compressed::
	.dc.b	$00,$09,$68,$58,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$E3,$BC,$FF,$C7,$BE,$04,$EA,$6C,$FF,$36,$9C,$CF
	.dc.b	$E1,$23,$0F,$09,$EB,$18,$3A,$9A,$7C,$31,$DC,$EB,$1F,$8E,$9E,$B3
	.dc.b	$CA,$3F,$7B,$3F,$92,$38,$47,$28,$F0,$C7,$8D,$DB,$E9,$7C,$47,$F4
	.dc.b	$E7,$8C,$7C,$37,$C4,$71,$9F,$0C,$7C,$59,$F0,$9E,$A3,$C6,$F8,$5E
	.dc.b	$E7,$C3,$8E,$31,$ED,$8E,$11,$F1,$63,$C1,$1E,$83,$84,$E6,$38,$47
	.dc.b	$9C,$F9,$0F,$07,$B9,$F1,$E3,$83,$EE,$3E,$5C,$F3,$8F,$6C,$76,$7E
	.dc.b	$23,$EC,$3B,$1E,$C9,$ED,$3F,$7A,$38,$47,$D3,$8F,$AF,$1D,$A7,$F0
	.dc.b	$C7,$67,$C8,$F6,$8D,$BF,$42,$78,$3F,$3C,$F4,$CF,$8C,$E9,$1F,$36
	.dc.b	$7C,$51,$D2,$3C,$53,$F8,$38,$F8,$B1,$D1,$F3,$BF,$66,$7C,$31,$A9
	.dc.b	$F3,$3F,$80,$9E,$6F,$D6,$8F,$5B,$F5,$4F,$04,$73,$8F,$04,$7C,$78
	.dc.b	$F0,$3A,$7E,$CB,$EA,$7E,$5C,$7C,$88,$FC,$BC,$78,$1F,$A3,$3F,$5D
	.dc.b	$F0,$1E,$58,$F0,$3E,$57,$E2,$4F,$38,$FC,$FC,$73,$7C,$0F,$91,$E4
	.dc.b	$F9,$23,$E9,$1F,$0A,$3C,$71,$D9,$F1,$C7,$C2,$7F,$1E,$72,$3F,$1C
	.dc.b	$F2,$7C,$E7,$28,$EC,$FC,$F3,$27,$99,$EC,$F9,$A3,$C4,$E6,$38,$BF
	.dc.b	$36,$33,$3F,$AC,$8F,$09,$D4,$F2,$9E,$13,$27,$13,$C2,$F1,$7F,$1D
	.dc.b	$1E,$97,$89,$E1,$38,$C7,$D3,$72,$F1,$3F,$16,$F5,$33,$1F,$9A,$8C
	.dc.b	$9C,$4F,$CC,$C7,$C9,$3C,$27,$A5,$CB,$E9,$32,$75,$72,$F9,$4C,$BD
	.dc.b	$63,$F0,$4F,$95,$FC,$AB,$97,$F5,$C7,$EC,$CC,$9F,$75,$EA,$71,$8F
	.dc.b	$C7,$4F,$E6,$27,$8B,$F8,$D7,$E4,$4F,$07,$E0,$1D,$23,$F2,$AF,$07
	.dc.b	$EF,$BC,$0F,$39,$C2,$7B,$C7,$E5,$DF,$85,$3F,$2A,$7C,$C7,$03,$D5
	.dc.b	$1C,$1F,$C2,$9C,$DF,$CC,$CF,$89,$FE,$06,$FE,$32,$7E,$24,$60,$FB
	.dc.b	$F3,$D8,$F4,$4E,$DC,$46,$CF,$A2,$E1,$FE,$00,$72,$8F,$03,$83,$6E
	.dc.b	$1F,$C1,$4F,$B6,$31,$1F,$02,$7E,$54,$7D,$18,$F0,$4E,$E3,$11,$D6
	.dc.b	$31,$1F,$26,$3F,$0D,$18,$9F,$8A,$7C,$88,$F8,$B3,$C9,$F1,$C7,$82
	.dc.b	$39,$4F,$E0,$E7,$71,$89,$F9,$D1,$FB,$93,$F5,$AE,$0F,$BC,$7C,$A7
	.dc.b	$E2,$47,$77,$DC,$7D,$93,$8C,$79,$5E,$87,$EB,$1C,$1D,$1F,$86,$F9
	.dc.b	$CF,$6C,$7E,$56,$3F,$59,$1F,$87,$8C,$47,$EA,$E3,$11,$F9,$48,$DC
	.dc.b	$79,$DF,$AF,$1E,$77,$E5,$C7,$77,$CF,$1D,$27,$F1,$71,$F4,$E7,$4F
	.dc.b	$A6,$75,$1F,$87,$8F,$9E,$7C,$13,$51,$E3,$7C,$D1,$E3,$9F,$B5,$1E
	.dc.b	$09,$D4,$78,$23,$99,$F5,$A7,$E1,$CE,$9F,$04,$7D,$03,$47,$38,$FC
	.dc.b	$F1,$F9,$27,$DE,$79,$67,$F7,$D3,$F4,$A3,$E3,$9F,$48,$D3,$F3,$27
	.dc.b	$99,$F2,$67,$F1,$53,$F7,$1E,$71,$F1,$0F,$03,$A9,$FB,$D3,$F2,$CE
	.dc.b	$AF,$C3,$3E,$6C,$72,$9F,$71,$CA,$3F,$25,$3E,$63,$91,$F0,$0F,$9F
	.dc.b	$1E,$63,$F3,$91,$ED,$8F,$A2,$7B,$4E,$13,$DE,$78,$4F,$CE,$6C,$FC
	.dc.b	$B3,$F3,$EF,$43,$DC,$7C,$38,$F2,$C7,$63,$CA,$7C,$57,$E5,$3C,$9E
	.dc.b	$CD,$3B,$1F,$96,$8E,$C7,$C2,$3B,$47,$AE,$7F,$0D,$3F,$81,$3B,$47
	.dc.b	$CB,$7E,$69,$EF,$32,$57,$E9,$BF,$8E,$32,$7A,$1C,$BD,$5A,$FE,$28
	.dc.b	$F9,$2F,$B1,$F4,$3E,$C6,$9E,$C8,$FC,$FB,$F5,$8F,$80,$53,$65,$76
	.dc.b	$7D,$57,$6D,$76,$D3,$71,$E3,$6B,$F0,$E3,$CE,$D7,$DE,$7C,$17,$DB
	.dc.b	$1F,$97,$8F,$C7,$19,$29,$F0,$CA,$7C,$33,$99,$93,$F0,$47,$12,$99
	.dc.b	$7E,$CB,$C6,$3C,$27,$E2,$A3,$27,$8E,$38,$9B,$3C,$2E,$4D,$C7,$13
	.dc.b	$F8,$2C,$6C,$E3,$1B,$8C,$B6,$7F,$5D,$1F,$8D,$32,$73,$72,$7E,$80
	.dc.b	$FC,$A9,$FA,$69,$FB,$11,$FB,$A3,$E9,$14,$FA,$07,$B8,$ED,$1D,$0F
	.dc.b	$CF,$1D,$23,$EF,$C7,$27,$F3,$67,$E5,$4F,$82,$F9,$63,$F4,$A7,$95
	.dc.b	$FC,$94,$FE,$0A,$7F,$40,$70,$8F,$0C,$F0,$3C,$2F,$03,$D1,$1C,$0F
	.dc.b	$92,$F7,$8F,$34,$7E,$4A,$3C,$CF,$E3,$8E,$13,$F5,$23,$BC,$F0,$8F
	.dc.b	$24,$FA,$CE,$B3,$A3,$C4,$FC,$03,$F1,$51,$CE,$7D,$D1,$FA,$D9,$F3
	.dc.b	$BC,$E7,$53,$EC,$8D,$1C,$CF,$C9,$4F,$DE,$3C,$0E,$8F,$51,$CD,$D4
	.dc.b	$78,$23,$EF,$9C,$CF,$C7,$4F,$82,$3F,$08,$FE,$D6,$7C,$B3,$EA,$79
	.dc.b	$CF,$96,$39,$9E,$03,$4F,$E1,$4E,$71,$F3,$A3,$F5,$E7,$81,$E7,$1A
	.dc.b	$7F,$1F,$1F,$7E,$3D,$13,$E3,$8F,$43,$FA,$18,$FC,$11,$C6,$3E,$CC
	.dc.b	$F4,$7F,$26,$F4,$7E,$BC,$74,$8F,$21,$F9,$D7,$0E,$E7,$E2,$4F,$23
	.dc.b	$AC,$FB,$5C,$1F,$3D,$C3,$F8,$D9,$E4,$78,$DC,$1C,$A3,$D2,$F2,$3E
	.dc.b	$4B,$87,$67,$57,$13,$F9,$28,$DC,$7E,$56,$7E,$BC,$7E,$1A,$3C,$73
	.dc.b	$B8,$C4,$72,$9D,$C7,$D3,$8D,$C6,$27,$E7,$47,$59,$FA,$33,$FA,$B3
	.dc.b	$F3,$EF,$E0,$4E,$B1,$B7,$07,$57,$C0,$F2,$8F,$CC,$BF,$8A,$3F,$0B
	.dc.b	$3E,$08,$EB,$18,$7C,$0F,$E7,$CC,$3C,$A3,$F0,$D1,$E0,$79,$3B,$8F
	.dc.b	$BD,$3F,$99,$9D,$CF,$28,$FC,$0B,$F5,$5F,$53,$97,$84,$FE,$22,$3F
	.dc.b	$2D,$18,$8F,$BE,$FC,$A8,$F7,$3D,$1F,$B4,$F4,$8F,$1B,$F9,$98,$CC
	.dc.b	$7C,$03,$2F,$E1,$4F,$9A,$E5,$F0,$C7,$17,$F0,$71,$97,$F3,$F1,$E5
	.dc.b	$7D,$91,$99,$FC,$EC,$7E,$38,$E2,$7C,$73,$A9,$E1,$3E,$D4,$FA,$5C
	.dc.b	$9C,$4F,$29,$C4,$F0,$BE,$57,$27,$DF,$9F,$6B,$C4,$EA,$F9,$4F,$D4
	.dc.b	$4F,$E6,$63,$27,$57,$D2,$64,$FB,$EF,$53,$CA,$E6,$3F,$13,$19,$3C
	.dc.b	$A7,$B6,$38,$9F,$84,$8F,$2B,$C5,$FC,$71,$F8,$18,$E2,$7A,$63,$8C
	.dc.b	$65,$F2,$BF,$AB,$3C,$2F,$58,$F2,$C6,$4E,$A7,$E2,$A7,$F4,$F1,$F9
	.dc.b	$B8,$E2,$75,$3E,$29,$C6,$3F,$1D,$1F,$68,$F7,$4F,$89,$FC,$9C,$7E
	.dc.b	$0D,$FC,$79,$F2,$E7,$9B,$A8,$FB,$0F,$D0,$39,$C6,$9F,$23,$E8,$7F
	.dc.b	$1A,$73,$8F,$D1,$3A,$7C,$AF,$E0,$A7,$E1,$3F,$29,$F9,$D1,$A8,$F0
	.dc.b	$46,$8F,$53,$CE,$3D,$4F,$BE,$35,$3E,$58,$FD,$C4,$FA,$A7,$EA,$CF
	.dc.b	$CD,$8F,$CD,$4F,$86,$3E,$54,$74,$8F,$34,$7C,$F8,$F5,$C7,$43,$F1
	.dc.b	$11,$F5,$0E,$87,$E6,$63,$6F,$E2,$8E,$AF,$8D,$E1,$1F,$20,$F9,$B1
	.dc.b	$C1,$FC,$29,$F4,$CE,$F1,$F3,$DF,$B2,$70,$8F,$C3,$BE,$53,$F2,$91
	.dc.b	$D9,$DC,$FC,$69,$ED,$1F,$04,$EC,$70,$9E,$CF,$CD,$7E,$14,$76,$7D
	.dc.b	$91,$C2,$7E,$CC,$FC,$98,$FB,$33,$C2,$3E,$7C,$FD,$A9,$F3,$3C,$23
	.dc.b	$E2,$4F,$82,$7E,$8B,$F3,$1F,$C6,$47,$85,$F8,$A7,$E0,$DE,$27,$E2
	.dc.b	$67,$EF,$C7,$CD,$7E,$04,$71,$8F,$80,$71,$3E,$6B,$C6,$7E,$44,$7C
	.dc.b	$17,$C4,$FC,$F8,$F5,$C7,$E3,$CF,$87,$1C,$CF,$95,$1C,$E3,$D9,$1F
	.dc.b	$11,$E7,$3E,$49,$E6,$7D,$47,$E3,$47,$49,$F8,$EF,$E2,$9F,$3C,$FC
	.dc.b	$F9,$F7,$C7,$DE,$9F,$A0,$7E,$4E,$7E,$04,$F4,$9F,$92,$74,$3E,$D4
	.dc.b	$F5,$7D,$33,$D6,$3E,$1C,$7D,$43,$E6,$3D,$9F,$AB,$1E,$A3,$BC,$7C
	.dc.b	$58,$EE,$FA,$C8,$E0,$87,$D4,$3C,$87,$CB,$9F,$C3,$3F,$02,$7E,$F9
	.dc.b	$E5,$9F,$9B,$3F,$0A,$7E,$C1,$E6,$8F,$9E,$79,$DF,$B4,$FC,$59,$F4
	.dc.b	$11,$C1,$23,$DA,$D7,$89,$82,$18,$24,$FD,$D8,$E0,$D8,$DB,$5D,$C7
	.dc.b	$96,$30,$47,$04,$3D,$64,$8F,$71,$0F,$C1,$39,$69,$93,$E2,$98,$21
	.dc.b	$82,$1F,$7C,$C9,$4C,$9F,$9B,$3A,$19,$21,$93,$89,$4E,$31,$F5,$DD
	.dc.b	$B5,$D1,$88,$FC,$8B,$82,$46,$C9,$1A,$6B,$F8,$43,$04,$30,$49,$FD
	.dc.b	$B1,$0F,$C6,$14,$FC,$79,$27,$F2,$44,$3F,$4E,$E5,$A6,$67,$F3,$2E
	.dc.b	$08,$E0,$93,$FB,$16,$C7,$ED,$08,$7E,$8E,$34,$47,$27,$EA,$08,$FE
	.dc.b	$B9,$FD,$63,$63,$F8,$14,$60,$FD,$8B,$82,$47,$EE,$63,$44,$8C,$CF
	.dc.b	$F1,$79,$D4,$62,$3F,$80,$99,$21,$82,$1F,$AF,$9F,$D1,$CF,$E7,$5F
	.dc.b	$C5,$9F,$79,$FB,$11,$F4,$E3,$E6,$47,$C1,$7E,$21,$F0,$9F,$71,$E4
	.dc.b	$24,$78,$8F,$04,$7E,$C2,$3F,$4C,$FE,$51,$EF,$1F,$82,$21,$F7,$67
	.dc.b	$EA,$CF,$80,$87,$D0,$3E,$5C,$F8,$27,$DE,$FF,$02,$3D,$D1,$FA,$49
	.dc.b	$FC,$BC,$FB,$63,$F0,$B3,$ED,$7A,$1F,$45,$E9,$1E,$A9,$FE,$0D,$1E
	.dc.b	$A9,$FD,$43,$F9,$D9,$F4,$C7,$C8,$3E,$31,$E6,$7E,$24,$79,$4E,$51
	.dc.b	$F4,$5E,$51,$F0,$5F,$21,$F3,$9E,$31,$F6,$A3,$8C,$FB,$27,$C3,$3E
	.dc.b	$C3,$C2,$7A,$DE,$2F,$D3,$3F,$03,$1C,$63,$C1,$1C,$23,$F1,$D1,$F0
	.dc.b	$67,$E5,$CF,$D9,$7C,$EF,$BC,$FA,$AF,$9E,$3E,$9C,$79,$CE,$F1,$F8
	.dc.b	$08,$EF,$1C,$0F,$8C,$F0,$7C,$C7,$D9,$8E,$0F,$73,$EB,$CF,$9A,$3E
	.dc.b	$64,$76,$7C,$AF,$C3,$3B,$1F,$0E,$7B,$4F,$DC,$8E,$11,$F8,$58,$FC
	.dc.b	$8C,$76,$9F,$C1,$9D,$9F,$23,$DA,$36,$FC,$19,$F9,$0F,$D6,$3D,$33
	.dc.b	$E3,$3A,$47,$D4,$9F,$14,$74,$8F,$14,$FE,$0A,$3E,$2C,$74,$7C,$EF
	.dc.b	$E0,$27,$C3,$1A,$9F,$33,$F5,$E7,$9B,$F7,$63,$D6,$FD,$53,$C1,$1C
	.dc.b	$E3,$C1,$1F,$1E,$3C,$0E,$9F,$B2,$FA,$9F,$97,$1F,$86,$8F,$D2,$C7
	.dc.b	$81,$FA,$33,$F5,$DF,$01,$E5,$8F,$03,$E5,$7E,$24,$F3,$8F,$D3,$47
	.dc.b	$37,$C0,$F9,$1E,$4F,$92,$3E,$91,$F0,$A3,$C7,$1D,$9F,$1C,$7C,$27
	.dc.b	$F1,$E7,$23,$F1,$CF,$27,$CE,$72,$8E,$CF,$CF,$32,$79,$9E,$CF,$9A
	.dc.b	$3C,$4E,$63,$8B,$F3,$63,$33,$FB,$38,$F0,$9D,$4F,$29,$E1,$32,$71
	.dc.b	$3C,$2F,$17,$F1,$D1,$E9,$78,$9E,$13,$8C,$7D,$37,$2F,$13,$F1,$6F
	.dc.b	$53,$31,$F9,$A8,$C9,$C4,$FC,$CC,$7C,$93,$C2,$7A,$5C,$BE,$93,$27
	.dc.b	$57,$2F,$94,$CB,$D6,$3F,$04,$F9,$5F,$CA,$B9,$7F,$5C,$7E,$CC,$C9
	.dc.b	$F7,$5E,$A7,$18,$FC,$74,$FE,$62,$78,$BF,$8D,$7E,$44,$F0,$7E,$01
	.dc.b	$D2,$3F,$2A,$F0,$7E,$FB,$C0,$F3,$9C,$27,$BC,$7E,$5D,$F8,$53,$F2
	.dc.b	$A7,$CC,$70,$3D,$51,$C1,$FC,$29,$CD,$FC,$CC,$F8,$9F,$E0,$6F,$E3
	.dc.b	$27,$E2,$46,$0F,$BF,$3D,$8F,$44,$ED,$C4,$6C,$FA,$2E,$1F,$E0,$07
	.dc.b	$28,$F0,$38,$36,$E1,$FC,$14,$FB,$63,$11,$F0,$27,$E5,$47,$D1,$8F
	.dc.b	$04,$EE,$31,$1D,$63,$11,$F2,$63,$F0,$D1,$89,$F8,$A7,$C8,$8F,$8B
	.dc.b	$3C,$9F,$1C,$78,$23,$94,$FE,$0E,$77,$18,$9F,$9D,$1F,$B9,$3F,$5A
	.dc.b	$E0,$FB,$C7,$CA,$7E,$24,$77,$7D,$C7,$D9,$38,$C7,$95,$E8,$7E,$B1
	.dc.b	$C1,$D1,$F8,$6F,$9C,$F6,$C7,$E5,$63,$F5,$90,$0C,$24,$E6,$13,$0C
	.dc.b	$BB,$9E,$47,$2F,$A1,$7C,$06,$EB,$32,$9C,$D7,$18,$E1,$5E,$3A,$ED
	.dc.b	$18,$3C,$93,$E5,$AE,$B3,$E3,$8C,$57,$6A,$B1,$E1,$79,$1E,$E7,$73
	.dc.b	$C6,$3C,$31,$E3,$7C,$D3,$E3,$9F,$83,$1C,$8F,$5D,$78,$A3,$E2,$D7
	.dc.b	$82,$7D,$93,$EE,$9F,$5C,$74,$3D,$67,$C0,$7C,$4F,$82,$7C,$53,$C6
	.dc.b	$7C,$33,$C4,$F5,$1C,$5F,$0B,$EA,$3C,$35,$E9,$8F,$8B,$1E,$03,$DA
	.dc.b	$F8,$27,$E9,$C7,$0A,$F9,$0F,$73,$DF,$5F,$6E,$3D,$D1,$F8,$19,$F3
	.dc.b	$57,$68,$F2,$9D,$9F,$2B,$ED,$3B,$3F,$12,$3F,$0F,$1E,$57,$E6,$47
	.dc.b	$82,$7E,$1D,$7A,$EB,$E9,$47,$DA,$9E,$B5,$EA,$9F,$25,$6D,$F5,$56
	.dc.b	$EB,$A4,$78,$9F,$9B,$5E,$28,$FD,$DC,$FE,$A2,$7F,$9A,$C7,$EA,$23
	.dc.b	$F7,$0F,$F1,$88,$FD,$C3,$FC,$32,$3F,$C4,$33,$EA,$7E,$3C,$7E,$52
	.dc.b	$7E,$05,$7E,$A6,$7C,$86,$A7,$F2,$73,$A7,$E2,$4F,$D1,$7E,$F1,$EE
	.dc.b	$9F,$AF,$5E,$98,$FA,$F5,$F7,$6B,$F4,$91,$D2,$3D,$F1,$E1,$8F,$C5
	.dc.b	$47,$C4,$7A,$3E,$17,$ED,$CF,$AD,$F0,$9E,$B8,$E6,$F9,$9F,$05,$7D
	.dc.b	$83,$CB,$5D,$E7,$F0,$93,$DD,$E4,$FE,$25,$F4,$BF,$72,$3D,$31,$FA
	.dc.b	$58,$F2,$3A,$79,$3E,$D7,$94,$FC,$C8,$F3,$99,$3F,$12,$7C,$07,$E8
	.dc.b	$3D,$63,$D6,$FE,$08,$FE,$62,$FF,$29,$7F,$C2,$11,$FD,$BE,$31,$7F
	.dc.b	$DA,$27,$FA,$84,$FF,$31,$78,$CF,$E2,$9E,$31,$F8,$A9,$FE,$A5,$1A
	.dc.b	$9F,$ED,$87,$F6,$07,$F1,$87,$F4,$33,$FA,$03,$E0,$9F,$ED,$D1,$FD
	.dc.b	$58,$D4,$FF,$84,$63,$FA,$2C,$7E,$64,$FE,$8A,$69,$F5,$4F,$96,$7D
	.dc.b	$53,$CC,$F8,$47,$82,$35,$3E,$08,$FE,$AE,$6A,$7C,$13,$E5,$3D,$4F
	.dc.b	$F6,$78,$FE,$B2,$EA,$39,$9A,$9F,$EC,$D3,$E7,$3E,$59,$E7,$7E,$41
	.dc.b	$F3,$5F,$24,$7C,$83,$F6,$73,$DE,$7F,$B2,$4F,$23,$CC,$FC,$09,$E0
	.dc.b	$7F,$79,$7E,$3C,$7D,$63,$E3,$9C,$1F,$0B,$E3,$32,$7B,$1E,$8F,$F7
	.dc.b	$C8,$E8,$7C,$33,$CB,$1F,$C5,$67,$FA,$C3,$F9,$79,$FE,$4C,$7E,$CA
	.dc.b	$7F,$26,$FE,$7E,$3F,$2F,$7F,$CC,$23,$F3,$67,$EB,$23,$37,$FC,$86
	.dc.b	$7F,$82,$47,$57,$F9,$34,$FE,$C0,$E2,$FE,$F5,$F0,$99,$3F,$94,$9F
	.dc.b	$A7,$3F,$56,$7E,$9E,$BF,$5E,$E4,$E2,$7E,$BC,$E2,$79,$4E,$31,$E9
	.dc.b	$72,$F8,$4F,$2B,$C5,$F4,$BD,$5C,$BE,$13,$D3,$5F,$C9,$5F,$09,$93
	.dc.b	$AB,$98,$E2,$78,$4E,$B1,$F7,$CF,$C3,$1F,$6E,$BE,$FB,$F8,$13,$EF
	.dc.b	$BF,$8C,$7F,$4C,$7D,$53,$DB,$3F,$44,$FE,$41,$3E,$D7,$F8,$DB,$E1
	.dc.b	$72,$79,$4C,$9C,$63,$F9,$65,$7F,$23,$AF,$E5,$8F,$F2,$27,$EE,$4F
	.dc.b	$07,$F9,$7C,$F0,$9F,$E4,$AF,$C3,$AF,$79,$E3,$9E,$73,$F8,$E7,$99
	.dc.b	$87,$EE,$CF,$D9,$3E,$6C,$FA,$5E,$D3,$F9,$03,$C2,$FF,$48,$AF,$64
	.dc.b	$73,$8F,$AF,$5F,$D1,$9F,$EE,$4F,$6A,$F2,$46,$0D,$C7,$C1,$70,$7D
	.dc.b	$C7,$93,$FA,$77,$95,$78,$23,$73,$F0,$23,$11,$B3,$AC,$FC,$69,$F6
	.dc.b	$47,$A6,$3A,$BF,$A3,$8E,$51,$88,$FC,$B4,$FD,$A8,$C4,$7C,$F8,$C4
	.dc.b	$7A,$63,$C0,$7A,$5C,$3D,$63,$71,$E3,$7F,$05,$1E,$38,$DB,$E0,$3E
	.dc.b	$89,$F9,$0A,$FA,$2F,$E4,$CF,$AF,$3C,$9C,$57,$E8,$A3,$13,$F8,$09
	.dc.b	$F8,$F1,$E2,$8F,$B9,$5F,$8F,$9F,$D9,$4F,$E3,$E3,$11,$F8,$5A,$E9
	.dc.b	$5E,$23,$A4,$7C,$08,$F9,$33,$EA,$8F,$8C,$7A,$8F,$3C,$F1,$7B,$BF
	.dc.b	$89,$7E,$44,$7C,$58,$D4,$78,$6B,$C9,$3F,$9B,$8E,$27,$B0,$ED,$5E
	.dc.b	$68,$F4,$BF,$90,$8F,$34,$FC,$17,$E6,$4E,$A3,$F1,$31,$F5,$CE,$6F
	.dc.b	$82,$3F,$22,$7E,$0A,$7F,$17,$3F,$42,$74,$FE,$42,$BE,$6D,$79,$E3
	.dc.b	$D5,$1E,$FA,$F8,$F3,$EF,$3C,$47,$5A,$FB,$35,$EE,$3D,$27,$92,$B9
	.dc.b	$47,$AE,$3E,$6C,$F2,$3E,$95,$79,$AB,$EB,$CF,$B4,$FC,$84,$F8,$4F
	.dc.b	$34,$FC,$78,$D1,$ED,$9E,$EF,$84,$FA,$07,$27,$DF,$3F,$18,$F0,$D7
	.dc.b	$AA,$7E,$09,$C0,$E8,$59,$F2,$BF,$A1,$3A,$1F,$5A,$7C,$B3,$F5,$A3
	.dc.b	$47,$23,$E2,$9C,$0F,$87,$1C,$0F,$99,$3F,$8C,$AF,$74,$F2,$8F,$0B
	.dc.b	$F9,$83,$94,$7E,$2E,$79,$1E,$C9,$E5,$5E,$3A,$E0,$78,$CF,$93,$1D
	.dc.b	$2B,$E9,$47,$DB,$32,$53,$C4,$53,$9C,$71,$9F,$95,$3E,$89,$C9,$B7
	.dc.b	$AB,$F3,$8F,$86,$78,$CE,$A6,$4F,$09,$4C,$BD,$4E,$65,$8F,$09,$EF
	.dc.b	$72,$D3,$35,$F3,$8F,$C9,$1F,$8D,$3F,$08,$59,$FD,$31,$F7,$5C,$D7
	.dc.b	$E2,$A3,$27,$DA,$8C,$BB,$33,$5F,$56,$33,$1B,$33,$5F,$B4,$7F,$86
	.dc.b	$D7,$56,$9C,$CA,$71,$2D,$75,$79,$C7,$E8,$9F,$C2,$CF,$59,$FD,$DC
	.dc.b	$F5,$6C,$F5,$6B,$D5,$A7,$57,$BC,$F3,$3F,$79,$19,$36,$E6,$3F,$64
	.dc.b	$7D,$E9,$FC,$31,$F9,$4A,$FA,$47,$E6,$0F,$A1,$1D,$A3,$D4,$FE,$14
	.dc.b	$FC,$59,$F8,$32,$D7,$E0,$63,$A4,$FE,$84,$FB,$AF,$BE,$BC,$E7,$96
	.dc.b	$BE,$2C,$76,$3C,$B3,$C2,$3E,$24,$F0,$9D,$47,$E6,$E3,$F1,$73,$DC
	.dc.b	$F5,$4F,$79,$F8,$27,$E6,$AF,$FD,$3A,$FF,$A6,$0F,$F1,$F1,$FE,$3C
	.dc.b	$3F,$D2,$07,$F8,$F2,$3F,$C7,$CE,$DF,$2C,$F6,$8F,$29,$D8,$FD,$2D
	.dc.b	$7E,$7E,$7E,$1C,$FC,$CB,$FA,$B1,$EA,$AF,$95,$1F,$A6,$9F,$A6,$7E
	.dc.b	$59,$FB,$95,$F0,$67,$F9,$11,$F0,$6B,$A9,$EF,$7E,$59,$A3,$AC,$7C
	.dc.b	$6A,$F1,$3C,$E3,$F0,$87,$38,$FA,$0F,$E4,$63,$4F,$F0,$2A,$FB,$D3
	.dc.b	$A7,$CB,$1F,$04,$FC,$1C,$FC,$19,$D3,$F7,$E3,$F1,$53,$EF,$9F,$B0
	.dc.b	$7B,$E7,$55,$F3,$A3,$0F,$C7,$9F,$0C,$7D,$1A,$F3,$1E,$18,$E9,$1E
	.dc.b	$E8,$C5,$7C,$1A,$F8,$D1,$E2,$8C,$3F,$68,$F9,$27,$C4,$3E,$4C,$7C
	.dc.b	$E7,$47,$17,$0F,$88,$E2,$7C,$78,$F8,$53,$E7,$AF,$8F,$3A,$3F,$0F
	.dc.b	$1F,$80,$7F,$84,$9F,$66,$3D,$F1,$F0,$A3,$13,$D2,$31,$1E,$79,$DB
	.dc.b	$C8,$C4,$ED,$C1,$E5,$9F,$03,$FB,$F3,$AB,$89,$FC,$BC,$62,$3E,$BB
	.dc.b	$C8,$F4,$B8,$79,$1E,$07,$94,$FE,$2A,$37,$1F,$55,$F8,$47,$E2,$1D
	.dc.b	$BF,$62,$BE,$6D,$7C,$88,$FA,$13,$F8,$C9,$FC,$25,$7B,$67,$C2,$7B
	.dc.b	$67,$C9,$5E,$A8,$F2,$46,$27,$C9,$1E,$18,$C3,$E1,$7B,$3F,$05,$F4
	.dc.b	$3E,$27,$EE,$47,$E6,$23,$F5,$51,$8A,$F2,$BE,$C8,$F7,$D7,$E5,$A7
	.dc.b	$F2,$67,$E7,$E3,$E6,$1E,$39,$EF,$19,$8F,$69,$F2,$9E,$2F,$82,$3F
	.dc.b	$06,$FC,$17,$8B,$93,$AC,$F8,$5F,$CD,$CF,$58,$F8,$D1,$C6,$3A,$BF
	.dc.b	$A2,$9F,$86,$FC,$A7,$AB,$97,$E0,$3F,$1E,$38,$CF,$E5,$23,$D2,$7B
	.dc.b	$63,$35,$F6,$67,$C3,$3E,$D3,$CA,$7A,$4E,$2F,$A5,$FD,$81,$F4,$EB
	.dc.b	$E0,$BC,$4C,$D7,$B6,$BE,$FD,$71,$3F,$19,$19,$9F,$C0,$57,$A9,$F7
	.dc.b	$47,$9A,$3E,$D4,$77,$02,$55,$88,$CD,$7F,$4A,$56,$89,$62,$4F,$49
	.dc.b	$D3,$CE,$79,$47,$17,$84,$F0,$9E,$B1,$E1,$7F,$AE,$96,$78,$4F,$08
	.dc.b	$ED,$3B,$3A,$9E,$48,$E9,$3E,$29,$F5,$9F,$D9,$8F,$EC,$87,$28,$F1
	.dc.b	$CF,$C1,$9F,$34,$F8,$CF,$2C,$7A,$27,$4F,$D4,$AF,$E6,$73,$F6,$0F
	.dc.b	$C0,$5F,$E1,$1F,$C4,$1F,$84,$3F,$21,$5F,$80,$72,$71,$32,$7E,$82
	.dc.b	$7A,$CF,$E0,$23,$AC,$78,$6B,$F8,$7B,$FB,$B8,$E2,$7E,$41,$E2,$66
	.dc.b	$7F,$38,$79,$4F,$C6,$39,$3F,$38,$75,$AF,$D4,$CF,$EB,$4F,$55,$FF
	.dc.b	$85,$1F,$8B,$1F,$55,$FB,$67,$E0,$4F,$A0,$69,$FA,$B7,$F6,$63,$33
	.dc.b	$F7,$A7,$31,$C4,$FC,$2B,$F5,$9F,$C0,$CF,$D5,$7A,$47,$D4,$78,$4F
	.dc.b	$94,$E9,$19,$8F,$C1,$9D,$1F,$CD,$4F,$DC,$3A,$4F,$09,$FC,$79,$D2
	.dc.b	$3F,$04,$6D,$FF,$19,$1F,$42,$FE,$65,$FD,$09,$FE,$57,$7F,$CD,$AB
	.dc.b	$F9,$5C,$E5,$FE,$20,$7E,$16,$BE,$AC,$7D,$19,$FD,$E4,$FD,$13,$F0
	.dc.b	$D3,$A7,$F8,$0C,$7E,$2E,$3E,$DD,$78,$23,$F8,$FC,$6E,$7C,$33,$F8
	.dc.b	$83,$47,$F0,$79,$F3,$47,$F0,$73,$F1,$51,$B7,$CE,$78,$AB,$F0,$EF
	.dc.b	$88,$FC,$C4,$6D,$FE,$26,$7B,$63,$6F,$8A,$7F,$1A,$FF,$21,$AF,$CF
	.dc.b	$3F,$2E,$7E,$14,$FB,$5F,$91,$1D,$0F,$14,$7F,$1E,$AF,$E4,$91,$D6
	.dc.b	$3E,$4B,$D6,$7F,$3D,$1B,$8F,$D5,$9E,$93,$DE,$F9,$1F,$DA,$D7,$90
	.dc.b	$F9,$4F,$C5,$7E,$FD,$7B,$E3,$EB,$46,$EB,$DF,$1B,$7D,$6F,$C6,$AE
	.dc.b	$D3,$F6,$23,$E6,$BC,$63,$CD,$1F,$0C,$F3,$47,$82,$38,$CF,$C4,$9F
	.dc.b	$0B,$EE,$8F,$B7,$5F,$12,$7C,$47,$29,$F1,$BF,$02,$7E,$3B,$F0,$23
	.dc.b	$94,$7E,$29,$F5,$CF,$32,$4F,$CA,$3E,$84,$FD,$78,$F8,$91,$F4,$5F
	.dc.b	$C7,$BF,$83,$8F,$6D,$7C,$C9,$E7,$3E,$F3,$99,$F1,$E7,$CB,$3F,$1E
	.dc.b	$3F,$04,$FC,$78,$F7,$C7,$8C,$F7,$9C,$8F,$73,$E5,$3E,$4C,$F9,$A3
	.dc.b	$E0,$3E,$67,$F1,$B1,$D1,$F2,$1D,$09,$1E,$89,$F5,$47,$86,$31,$5F
	.dc.b	$22,$30,$43,$9C,$6A,$BE,$1C,$EE,$72,$6E,$BA,$B8,$23,$82,$47,$16
	.dc.b	$CF,$43,$F1,$77,$FC,$96,$FF,$B4,$47,$F2,$5B,$FE,$E2,$7F,$B2,$CF
	.dc.b	$E6,$B1,$A7,$F9,$19,$FC,$AA,$79,$47,$F6,$58,$C9,$C8,$FE,$DE,$79
	.dc.b	$AF,$C8,$4A,$F2,$1F,$36,$FC,$ED,$8D,$34,$D9,$F9,$32,$D7,$F1,$97
	.dc.b	$06,$C8,$60,$8E,$C9,$5F,$93,$8D,$C6,$23,$F0,$31,$C1,$D3,$5C,$C6
	.dc.b	$EB,$F2,$53,$96,$C6,$6B,$F5,$6D,$3F,$1D,$1A,$AE,$B1,$97,$F8,$1C
	.dc.b	$F0,$72,$43,$C2,$59,$E4,$E0,$8E,$09,$1C,$E3,$2D,$8C,$CF,$92,$30
	.dc.b	$47,$04,$9F,$2B,$5F,$D9,$C6,$DC,$90,$CC,$6D,$FE,$2E,$D7,$F8,$31
	.dc.b	$B3,$25,$32,$72,$30,$7E,$E0,$B3,$FC,$15,$E2,$6C,$86,$08,$74,$29
	.dc.b	$CC,$D9,$92,$19,$36,$F7,$72,$D7,$46,$27,$F9,$B3,$67,$F8,$84,$EC
	.dc.b	$93,$A9,$E2,$48,$CC,$F0,$9D,$46,$2B,$F8,$D9,$0F,$E4,$F1,$A2,$4F
	.dc.b	$12,$46,$CF,$E4,$87,$07,$2D,$35,$5F,$CC,$DD,$91,$C1,$0F,$E5,$E5
	.dc.b	$AF,$EB,$11,$CC,$FF,$41,$3F,$E9,$28,$FF,$17,$47,$F8,$FD,$FF,$38
	.dc.b	$9F,$E2,$EB,$FE,$1F,$1F,$E4,$F7,$FC,$FD,$1F,$C3,$CF,$F2,$19,$FC
	.dc.b	$4A,$BF,$D0,$A7,$F1,$57,$FC,$AA,$7F,$19,$7F,$C7,$6F,$F9,$30,$EC
	.dc.b	$E1,$F0,$9F,$E9,$37,$F9,$13,$C8,$FF,$47,$47,$F2,$A8,$FF,$22,$46
	.dc.b	$DF,$04,$79,$9F,$E7,$D3,$FC,$3A,$7F,$71,$7F,$A7,$7F,$1E,$7D,$F7
	.dc.b	$EB,$C7,$D0,$8F,$96,$4B,$F9,$71,$F4,$63,$13,$FD,$F2,$73,$1F,$46
	.dc.b	$30,$47,$FC,$04,$43,$FB,$34,$70,$79,$99,$3F,$AF,$9F,$D4,$E3,$F9
	.dc.b	$AC,$FF,$1F,$7D,$51,$FA,$47,$D3,$1F,$7C,$F3,$4F,$53,$F0,$13,$F4
	.dc.b	$E3,$F0,$17,$FE,$B2,$8F,$95,$1F,$7E,$FF,$D1,$11,$FE,$79,$8F,$A3
	.dc.b	$1F,$82,$8F,$E9,$F3,$F8,$69,$FF,$45,$C7,$57,$FC,$FE,$FD,$F3,$AB
	.dc.b	$A7,$F1,$51,$FE,$81,$7E,$F4,$FE,$2A,$7C,$51,$F9,$67,$9B,$F8,$63
	.dc.b	$FB,$41,$E5,$8F,$C1,$B8,$9F,$24,$FF,$6F,$9E,$8F,$85,$AE,$67,$FB
	.dc.b	$8C,$7E,$28,$F8,$33,$CA,$3F,$3C,$6A,$7C,$AF,$BD,$F8,$B3,$FE,$9C
	.dc.b	$3F,$33,$3F,$93,$9E,$13,$E1,$6B,$E1,$7D,$47,$C7,$3D,$53,$FA,$59
	.dc.b	$FA,$D1,$89,$FB,$33,$F4,$E3,$F6,$67,$CA,$3A,$9E,$83,$A9,$F2,$4F
	.dc.b	$B2,$F5,$39,$4F,$D3,$3D,$11,$C1,$F0,$BC,$27,$C3,$3E,$F3,$E4,$47
	.dc.b	$D7,$8F,$86,$FE,$F9,$DB,$5F,$B7,$3D,$23,$D2,$F8,$0F,$A0,$FA,$CF
	.dc.b	$14,$F9,$8F,$9A,$7D,$63,$D4,$78,$E7,$EA,$C7,$E3,$A3,$6F,$95,$FC
	.dc.b	$3B,$EB,$8F,$1C,$75,$3D,$71,$F7,$5D,$BE,$67,$C1,$1E,$C8,$FC,$EB
	.dc.b	$F6,$E7,$D2,$7A,$E7,$CB,$3F,$06,$7E,$2B,$C0,$FA,$D3,$C2,$7A,$CF
	.dc.b	$91,$F6,$3F,$89,$9F,$87,$18,$8F,$03,$DA,$3A,$1F,$26,$7A,$3F,$04
	.dc.b	$FC,$71,$F0,$67,$C6,$FE,$35,$F0,$3E,$79,$F8,$33,$F8,$57,$CB,$3C
	.dc.b	$27,$94,$68,$F9,$EF,$C9,$3B,$9F,$24,$F1,$1F,$93,$30,$6C,$C4,$F7
	.dc.b	$76,$FE,$1C,$C3,$B7,$11,$E0,$8F,$C0,$3F,$69,$B3,$F8,$48,$FE,$03
	.dc.b	$1F,$41,$F9,$0F,$C1,$8F,$63,$E4,$38,$BF,$46,$7E,$5C,$78,$67,$F1
	.dc.b	$13,$D6,$3F,$62,$7D,$18,$EA,$FC,$63,$E1,$CF,$A9,$EB,$3F,$68,$FC
	.dc.b	$63,$E3,$9F,$98,$7A,$CF,$4B,$D6,$36,$F5,$3E,$4C,$7C,$33,$73,$F9
	.dc.b	$39,$DC,$FB,$67,$27,$DC,$3B,$C7,$AE,$3F,$1E,$7E,$0C,$F4,$C7,$57
	.dc.b	$DF,$3D,$67,$F1,$31,$F9,$28,$C4,$7C,$C7,$E0,$4F,$82,$7E,$EB,$F2
	.dc.b	$A7,$E7,$3F,$22,$3C,$C7,$67,$CC,$F8,$9F,$88,$E4,$E6,$FE,$95,$FB
	.dc.b	$D1,$F0,$27,$C3,$1B,$3E,$D4,$ED,$F9,$93,$EA,$7E,$91,$DD,$F8,$93
	.dc.b	$F8,$28,$FB,$27,$68,$E0,$FD,$38,$FC,$AC,$F9,$A3,$E5,$9F,$15,$F5
	.dc.b	$BF,$61,$D4,$75,$7E,$34,$79,$8F,$A4,$6E,$7C,$71,$F1,$E7,$DB,$3D
	.dc.b	$8F,$49,$B9,$F9,$A7,$C3,$9F,$B5,$1B,$3D,$0F,$D3,$8F,$CA,$1E,$C8
	.dc.b	$FA,$66,$CB,$3E,$77,$E6,$1F,$10,$F9,$71,$F8,$19,$EF,$3B,$9E,$4F
	.dc.b	$A0,$C1,$23,$D8,$70,$29,$F1,$8B,$3F,$08,$F6,$39,$69,$93,$D5,$1D
	.dc.b	$E3,$8C,$7C,$43,$2F,$77,$89,$93,$D4,$F5,$76,$D7,$27,$72,$9C,$5F
	.dc.b	$D6,$B6,$7F,$18,$D7,$EB,$3F,$40,$E2,$58,$E2,$D8,$E2,$D7,$89,$F9
	.dc.b	$37,$8B,$63,$8B,$4E,$25,$38,$9F,$61,$DB,$4D,$94,$DC,$FE,$FA,$36
	.dc.b	$D8,$DB,$5D,$C7,$E8,$23,$F4,$B3,$F1,$E3,$89,$93,$F2,$53,$E5,$9D
	.dc.b	$CF,$C7,$6B,$F0,$5B,$1B,$9E,$2F,$9E,$32,$D8,$CC,$79,$4C,$95,$C9
	.dc.b	$E3,$3F,$83,$3F,$E2,$6A,$5F,$F9,$DA,$55,$FE,$90,$7F,$DE,$0F,$FA
	.dc.b	$3D,$FF,$5F,$DF,$F9,$DA,$7F,$D0,$D3,$FE,$84,$6B,$FE,$82,$15,$DA
	.dc.b	$7F,$D7,$0F,$82,$BB,$CC,$37,$5F,$EA,$A9,$FF,$40,$C7,$EC,$EF,$D9
	.dc.b	$5E,$A3,$D9,$1F,$C3,$23,$E0,$C6,$23,$F8,$B1,$FC,$B0,$D3,$F1,$1C
	.dc.b	$1A,$3F,$8B,$14,$F5,$4F,$A6,$FF,$D6,$91,$E6,$9F,$8B,$3F,$EE,$D9
	.dc.b	$FF,$4D,$4F,$FA,$AC,$FF,$52,$4F,$FA,$F2,$3F,$D4,$4F,$0B,$FF,$50
	.dc.b	$C7,$CB,$9D,$9F,$EA,$97,$E3,$C7,$FA,$A2,$3E,$AD,$77,$9E,$11,$F4
	.dc.b	$4E,$27,$CA,$78,$D7,$73,$EC,$C7,$94,$E0,$43,$DF,$5F,$83,$7F,$04
	.dc.b	$7D,$53,$F0,$53,$FE,$CF,$AF,$F5,$B4,$FE,$02,$BD,$C7,$DC,$8F,$BD
	.dc.b	$5C,$5F,$43,$E1,$3B,$D7,$99,$D5,$7C,$47,$55,$D1,$F1,$9F,$EE,$2A
	.dc.b	$E6,$FF,$B4,$8F,$54,$6D,$FA,$C7,$47,$E0,$BB,$3C,$35,$F7,$27,$CF
	.dc.b	$1D,$6B,$C4,$EE,$BA,$4E,$A3,$FD,$B5,$1D,$23,$E1,$C7,$D2,$7F,$11
	.dc.b	$5F,$80,$8F,$13,$F2,$6B,$E4,$47,$DF,$8F,$14,$7D,$F8,$F9,$F1,$F8
	.dc.b	$C7,$C4,$6D,$E8,$F9,$EB,$C5,$3D,$23,$F0,$53,$D2,$7F,$04,$F8,$63
	.dc.b	$CD,$3F,$7D,$F8,$E7,$38,$D3,$F3,$0F,$04,$FD,$D9,$F3,$57,$D2,$3E
	.dc.b	$0C,$EC,$F0,$D7,$99,$F8,$EF,$47,$E8,$B9,$7F,$3B,$3F,$42,$79,$C6
	.dc.b	$AB,$C0,$7A,$AB,$E2,$4E,$A3,$EA,$D7,$90,$E4,$7C,$38,$FA,$71,$CA
	.dc.b	$BE,$9D,$72,$AD,$1F,$0E,$BE,$F4,$F9,$2B,$E6,$BF,$22,$BF,$30,$FC
	.dc.b	$77,$93,$97,$E8,$3D,$9F,$90,$FF,$10,$3F,$C3,$B7,$FC,$61,$C1,$FE
	.dc.b	$1A,$9F,$E3,$CE,$0F,$F0,$BB,$FE,$00,$9F,$E9,$0F,$F1,$B8,$FE,$32
	.dc.b	$74,$70,$FC,$93,$F5,$27,$D2,$8F,$3B,$F5,$EB,$E2,$BA,$3E,$04,$7D
	.dc.b	$79,$F9,$A7,$D1,$7F,$88,$C7,$53,$F1,$27,$57,$31,$C2,$3E,$11,$FB
	.dc.b	$C3,$F1,$67,$73,$F3,$F1,$C1,$C9,$D0,$F1,$BE,$87,$C1,$3F,$59,$E2
	.dc.b	$75,$3F,$23,$3F,$7D,$FA,$D3,$D5,$C9,$E1,$9F,$CD,$5F,$E0,$A7,$EE
	.dc.b	$CF,$AC,$E9,$3F,$19,$F2,$1E,$03,$F6,$A7,$43,$2F,$47,$F0,$91,$F9
	.dc.b	$F7,$84,$7E,$65,$F7,$BE,$B3,$83,$F5,$67,$F2,$87,$E7,$6B,$F2,$71
	.dc.b	$9A,$F3,$BF,$19,$F7,$57,$9C,$E9,$5F,$A4,$9F,$1D,$7C,$57,$DC,$7A
	.dc.b	$A7,$84,$F7,$AF,$C1,$4F,$9A,$31,$1F,$4E,$BE,$F4,$FC,$E3,$E3,$CF
	.dc.b	$A2,$3F,$11,$18,$9F,$3B,$F0,$A7,$E8,$47,$EA,$23,$91,$E3,$8E,$4F
	.dc.b	$E2,$8C,$47,$2A,$F8,$F1,$F8,$49,$F2,$CF,$C2,$78,$C7,$49,$FA,$0F
	.dc.b	$D7,$7E,$2D,$7C,$C8,$E4,$FA,$1E,$F5,$E5,$7E,$3C,$F2,$3E,$3C,$ED
	.dc.b	$C5,$78,$27,$6F,$BA,$BF,$05,$3E,$8A,$F9,$35,$A3,$EA,$9F,$95,$AF
	.dc.b	$B0,$FF,$2D,$BF,$F4,$75,$FF,$9F,$8F,$E0,$75,$FC,$22,$7F,$81,$9F
	.dc.b	$C2,$CF,$F1,$0B,$97,$FC,$58,$FF,$89,$DF,$F1,$9C,$FF,$AC,$AF,$D7
	.dc.b	$1F,$3D,$F5,$D7,$CF,$BF,$1C,$FC,$63,$4F,$B6,$FF,$19,$1F,$93,$3F
	.dc.b	$01,$3F,$25,$C1,$F3,$1F,$87,$3D,$27,$E1,$CE,$AB,$C3,$5C,$63,$F1
	.dc.b	$07,$C6,$3F,$5F,$5C,$5F,$F5,$DD,$7F,$B4,$9F,$79,$FC,$AE,$FF,$D5
	.dc.b	$55,$F2,$63,$FC,$E1,$3F,$14,$E0,$7C,$43,$F9,$83,$FE,$66,$9F,$F3
	.dc.b	$55,$7D,$D8,$F3,$3F,$05,$FB,$13,$F8,$B3,$D6,$FA,$0F,$DC,$57,$AE
	.dc.b	$FF,$05,$1F,$AD,$9F,$C5,$57,$38,$F6,$C7,$37,$47,$D0,$3E,$14,$FE
	.dc.b	$6A,$BF,$6F,$1F,$9D,$AF,$D1,$BF,$4A,$3E,$7D,$73,$8D,$3E,$07,$9C
	.dc.b	$F0,$7E,$B4,$7D,$C7,$D4,$7E,$04,$FD,$64,$79,$EB,$C5,$5F,$C1,$67
	.dc.b	$CE,$68,$EB,$5D,$DF,$87,$1E,$43,$D3,$3F,$92,$8E,$06,$A7,$C5,$1F
	.dc.b	$22,$3D,$27,$C3,$34,$7D,$29,$E7,$5F,$5E,$3D,$35,$F5,$2B,$91,$DE
	.dc.b	$3F,$64,$72,$3E,$8D,$72,$3F,$45,$3F,$26,$78,$4F,$E0,$8F,$69,$E3
	.dc.b	$3E,$31,$D0,$A6,$CE,$CF,$C8,$9F,$E3,$07,$63,$C0,$7E,$1E,$BF,$3D
	.dc.b	$3F,$00,$F8,$A7,$43,$DA,$7D,$03,$C1,$5E,$F8,$E3,$5E,$89,$D9,$CC
	.dc.b	$D9,$C4,$F3,$1D,$5F,$8F,$1F,$2D,$B5,$E7,$3F,$01,$3C,$CE,$35,$E1
	.dc.b	$3E,$45,$71,$3C,$65,$73,$3C,$CF,$86,$7C,$14,$C3,$FC,$5B,$3F,$C8
	.dc.b	$0F,$F1,$3B,$FB,$D7,$FD,$D1,$1F,$BD,$7F,$82,$90,$5F,$F2,$C9,$EB
	.dc.b	$3F,$EF,$C9,$FE,$10,$8F,$04,$7F,$30,$8F,$F5,$34,$FF,$29,$AF,$54
	.dc.b	$FD,$83,$F3,$87,$C4,$BF,$A0,$7C,$87,$D8,$6E,$FF,$3B,$3F,$12,$3E
	.dc.b	$15,$FE,$E6,$72,$7C,$68,$E2,$75,$9F,$0B,$4E,$33,$FB,$A3,$F4,$8E
	.dc.b	$4E,$27,$CE,$3C,$E5,$9E,$EF,$96,$38,$B9,$69,$E5,$29,$DC,$FD,$4B
	.dc.b	$DC,$EA,$EC,$CB,$CD,$A6,$6B,$F4,$8D,$AF,$E2,$91,$FD,$12,$79,$CF
	.dc.b	$78,$DB,$67,$67,$F1,$52,$BF,$B6,$3F,$53,$5F,$95,$AF,$74,$78,$8E
	.dc.b	$51,$FC,$5D,$FD,$14,$FD,$59,$F9,$31,$E5,$9D,$1C,$E3,$E0,$3F,$29
	.dc.b	$B1,$F1,$AB,$D5,$39,$9F,$85,$1E,$A3,$33,$BA,$F5,$57,$68,$FD,$B1
	.dc.b	$FB,$A9,$FC,$6C,$F9,$6B,$E2,$C7,$29,$F6,$34,$ED,$3D,$23,$F8,$C5
	.dc.b	$7A,$6B,$C7,$3F,$22,$7F,$8B,$B5,$F8,$73,$F9,$39,$EF,$3C,$E4,$7F
	.dc.b	$AE,$A3,$F7,$0F,$FB,$00,$FD,$6C,$FE,$B1,$FE,$E6,$7F,$86,$08,$72
	.dc.b	$3F,$C3,$85,$3F,$7F,$1F,$B8,$7F,$BB,$47,$FA,$88,$FE,$F6,$E2,$3F
	.dc.b	$82,$5F,$F7,$A7,$F4,$B7,$F0,$EF,$D9,$7F,$06,$FE,$E1,$F1,$0F,$9F
	.dc.b	$3F,$B9,$AF,$E0,$91,$FE,$C7,$34,$7F,$0C,$9F,$F0,$4B,$FC,$48,$F5
	.dc.b	$9F,$EB,$77,$F8,$A3,$B8,$F2,$C7,$F8,$D6,$3D,$E6,$CE,$26,$08,$7C
	.dc.b	$28,$F1,$BE,$18,$EB,$7F,$C7,$67,$2F,$FB,$46,$7F,$94,$3F,$E2,$48
	.dc.b	$F9,$51,$C4,$EE,$75,$9F,$E4,$8F,$07,$F9,$24,$7D,$C8,$F7,$1F,$56
	.dc.b	$7A,$47,$82,$7E,$A4,$FF,$31,$9F,$24,$74,$8F,$F1,$B9,$D5,$FC,$23
	.dc.b	$D0,$F8,$47,$57,$C3,$3F,$D1,$23,$F1,$26,$63,$8C,$FC,$48,$EA,$48
	.dc.b	$FC,$2B,$F4,$67,$E0,$CF,$C5,$8D,$1E,$79,$FB,$71,$CD,$E2,$FA,$A3
	.dc.b	$F2,$31,$F4,$0D,$9F,$6D,$F9,$46,$8F,$9B,$1F,$1A,$3E,$11,$D0,$E3
	.dc.b	$1F,$A6,$9F,$9D,$3E,$18,$F2,$47,$E2,$1F,$E3,$87,$F7,$A3,$FB,$91
	.dc.b	$FC,$16,$FF,$A2,$C7,$EF,$2B,$F8,$54,$EA,$FF,$C5,$97,$EB,$8F,$C1
	.dc.b	$1F,$2E,$BC,$35,$EB,$9F,$44,$FA,$CD,$4F,$D3,$7E,$C1,$F9,$68,$DB
	.dc.b	$D0,$FB,$67,$F9,$BE,$7E,$B9,$D4,$FB,$B3,$FE,$73,$8F,$69,$FE,$0C
	.dc.b	$3F,$67,$7F,$C7,$E7,$F9,$54,$7E,$E0,$FE,$7A,$69,$FE,$2C,$69,$FE
	.dc.b	$7A,$7E,$E2,$FF,$94,$C6,$A3,$F9,$81,$E0,$3E,$55,$7F,$08,$AF,$E3
	.dc.b	$D1,$F3,$EB,$8C,$FF,$4A,$7F,$96,$C7,$17,$F8,$45,$78,$5F,$E2,$0F
	.dc.b	$D2,$78,$CF,$F4,$69,$FE,$6F,$5D,$9F,$80,$7D,$77,$E0,$3E,$63,$F9
	.dc.b	$AD,$73,$8F,$BF,$1F,$0C,$F3,$38,$7E,$09,$F1,$8E,$91,$F0,$5F,$09
	.dc.b	$E9,$8E,$EE,$0E,$E7,$C9,$7A,$57,$B2,$31,$5F,$1D,$E5,$5D,$67,$E2
	.dc.b	$C7,$23,$C4,$FC,$29,$FC,$B4,$72,$7F,$B7,$BE,$83,$D6,$F5,$79,$18
	.dc.b	$77,$1E,$38,$EB,$5E,$63,$E6,$3F,$72,$77,$18,$9F,$AF,$3F,$34,$FE
	.dc.b	$BF,$3F,$76,$7F,$A3,$C7,$E3,$63,$15,$F4,$5F,$24,$70,$7E,$E5,$7C
	.dc.b	$08,$C4,$7B,$9F,$2B,$C2,$7F,$A9,$42,$55,$FD,$86,$7F,$B2,$47,$C7
	.dc.b	$9F,$ED,$87,$F2,$9A,$FD,$BB,$FC,$25,$FD,$A5,$FF,$09,$8F,$E4,$11
	.dc.b	$FC,$D6,$BF,$8C,$3F,$C2,$E3,$93,$C2,$3F,$68,$7E,$DA,$3A,$47,$F9
	.dc.b	$8D,$FE,$02,$D7,$F7,$A7,$F2,$47,$C6,$F4,$7F,$78,$FD,$83,$F3,$B5
	.dc.b	$EB,$7E,$F1,$8B,$FB,$0F,$E2,$A7,$E7,$BF,$85,$7F,$C2,$2E,$5F,$9F
	.dc.b	$1F,$5A,$3F,$3B,$1C,$1F,$9E,$FE,$01,$E0,$FD,$37,$F1,$4F,$47,$CC
	.dc.b	$7E,$1D,$C3,$F6,$8F,$D7,$1E,$77,$ED,$3F,$85,$7C,$75,$F8,$57,$C1
	.dc.b	$1D,$2B,$F1,$EF,$EB,$4E,$87,$C2,$3C,$13,$FC,$26,$BF,$2C,$7E,$D2
	.dc.b	$7F,$19,$1F,$88,$7F,$41,$3E,$A8,$E0,$FB,$8F,$39,$D0,$FC,$F3,$C6
	.dc.b	$7A,$BF,$B7,$7E,$54,$FD,$53,$AC,$FB,$EB,$D6,$FA,$23,$2F,$53,$D9
	.dc.b	$5D,$5F,$11,$ED,$8F,$82,$7A,$9F,$91,$3F,$4D,$EC,$FD,$09,$E5,$1F
	.dc.b	$26,$BE,$11,$E6,$3F,$0E,$E6,$3D,$47,$D1,$9E,$51,$D8,$F8,$75,$E2
	.dc.b	$AE,$53,$F6,$A7,$F1,$2F,$F5,$D9,$FD,$9C,$7E,$88,$FF,$08,$9F,$A0
	.dc.b	$3F,$BF,$9F,$D7,$CF,$EC,$55,$FD,$71,$D9,$87,$F5,$8F,$E9,$4C,$9F
	.dc.b	$E3,$B9,$FD,$4B,$F3,$6B,$EA,$C7,$E6,$E3,$EF,$1F,$68,$F6,$1F,$36
	.dc.b	$3E,$2C,$FC,$13,$F0,$71,$F1,$67,$93,$F3,$A3,$2F,$73,$DC,$7F,$7C
	.dc.b	$3D,$B7,$FE,$27,$8F,$F0,$C4,$FF,$09,$9F,$F1,$49,$FC,$3A,$3F,$C3
	.dc.b	$AF,$F8,$62,$3F,$C5,$F1,$FE,$44,$8F,$9B,$1F,$7D,$FC,$45,$FE,$0D
	.dc.b	$C9,$F5,$AF,$D6,$FF,$01,$AF,$39,$F8,$47,$D0,$7B,$6B,$F6,$2E,$CF
	.dc.b	$8C,$74,$3E,$34,$7E,$19,$D1,$F9,$83,$E5,$D7,$E7,$0F,$C3,$47,$CB
	.dc.b	$8E,$71,$FA,$5A,$FA,$B5,$E0,$9F,$E1,$EF,$E4,$1F,$04,$F3,$9F,$5C
	.dc.b	$6A,$BD,$6F,$82,$39,$D7,$C5,$AF,$0C,$F9,$9F,$8C,$78,$5F,$D6,$BD
	.dc.b	$6B,$D9,$1D,$4F,$3D,$7E,$99,$F8,$E4,$8F,$94,$47,$C4,$74,$8F,$1B
	.dc.b	$F4,$E3,$AB,$F9,$23,$E4,$1E,$E8,$DB,$F3,$AB,$F1,$EF,$DE,$8F,$D7
	.dc.b	$C7,$F0,$A9,$F0,$4F,$13,$C0,$43,$EE,$47,$C3,$9F,$93,$1E,$1A,$F8
	.dc.b	$71,$ED,$AE,$71,$F1,$67,$89,$E9,$AF,$1C,$7C,$C8,$FF,$67,$3F,$DC
	.dc.b	$27,$FD,$26,$FF,$00,$8F,$DB,$1F,$A3,$BF,$F2,$E9,$FA,$37,$F3,$D5
	.dc.b	$FD,$B5,$FF,$3F,$D7,$F0,$17,$83,$67,$FB,$A1,$FE,$71,$39,$47,$C1
	.dc.b	$3C,$E7,$C9,$9F,$C6,$3E,$FB,$F9,$31,$F8,$DB,$F7,$4F,$DB,$3E,$C9
	.dc.b	$C9,$FC,$24,$7C,$09,$F9,$33,$FC,$04,$FC,$2C,$7F,$21,$9E,$77,$F2
	.dc.b	$AB,$E6,$57,$43,$B1,$E7,$2C,$F0,$75,$3C,$6B,$A5,$72,$9E,$13,$99
	.dc.b	$FC,$14,$70,$21,$A2,$D7,$57,$04,$70,$4A,$EC,$D9,$F0,$90,$EF,$3C
	.dc.b	$67,$64,$8D,$1F,$35,$C3,$C2,$7A,$36,$32,$EC,$E3,$5F,$BC,$AE,$D1
	.dc.b	$87,$E7,$BC,$4C,$3F,$AA,$AE,$B1,$CC,$D1,$0C,$12,$7B,$47,$06,$D7
	.dc.b	$91,$B3,$DA,$7F,$4E,$D9,$FC,$5C,$EA,$3B,$35,$ED,$5E,$85,$1E,$DB
	.dc.b	$FC,$D3,$FC,$F5,$AF,$F9,$3A,$3F,$A0,$B5,$FD,$34,$FF,$92,$8F,$EC
	.dc.b	$33,$FA,$C9,$E4,$7F,$B9,$23,$F6,$31,$FE,$5B,$BF,$74,$78,$E7,$CC
	.dc.b	$47,$04,$3E,$65,$FE,$46,$BC,$B1,$FB,$69,$FE,$E6,$FF,$5A,$8C,$BE
	.dc.b	$53,$F7,$31,$A3,$FB,$0D,$FB,$23,$93,$FD,$E2,$32,$FB,$A3,$FC,$0D
	.dc.b	$3E,$D9,$F8,$47,$F6,$68,$FE,$D0,$7F,$BB,$4D,$BF,$C1,$CF,$F3,$9C
	.dc.b	$F5,$8F,$E1,$F1,$C6,$7C,$F1,$C2,$3C,$CE,$8E,$87,$F8,$46,$79,$3E
	.dc.b	$03,$FC,$30,$7F,$7E,$8F,$A3,$1F,$DF,$5F,$59,$DC,$FA,$91,$F8,$20
	.dc.b	$73,$9F,$E3,$77,$F5,$A3,$E8,$D7,$E3,$EB,$FB,$54,$62,$3F,$1E,$FD
	.dc.b	$79,$FC,$AC,$62,$34,$FE,$49,$B3,$F7,$DA,$71,$BF,$CD,$BE,$4A,$FF
	.dc.b	$06,$5F,$C6,$8F,$85,$7F,$E9,$F7,$FC,$3F,$1F,$CD,$2F,$DA,$7C,$08
	.dc.b	$F9,$27,$C5,$7E,$CB,$88,$FB,$8F,$E0,$A7,$A3,$F9,$89,$FB,$31,$88
	.dc.b	$FC,$74,$76,$7E,$B1,$F4,$A7,$9B,$F9,$17,$F4,$17,$FB,$B8,$F7,$4F
	.dc.b	$CA,$BF,$8F,$7F,$15,$B1,$F1,$5F,$82,$F7,$9F,$AC,$FD,$E9,$FC,$D4
	.dc.b	$77,$8F,$3C,$7F,$AF,$63,$E3,$B5,$F8,$47,$BD,$C5,$FF,$04,$9F,$C8
	.dc.b	$CF,$F2,$09,$FB,$55,$E4,$8F,$49,$FE,$1B,$BF,$AE,$FD,$A2,$46,$DB
	.dc.b	$1A,$8C,$11,$C1,$2B,$E2,$36,$7B,$46,$08,$E0,$87,$F8,$9C,$EE,$E9
	.dc.b	$C4,$F7,$8C,$47,$F7,$36,$BF,$DA,$CE,$27,$43,$45,$35,$1C,$DC,$1F
	.dc.b	$E2,$58,$E6,$D3,$31,$FE,$17,$79,$9E,$49,$F7,$36,$78,$CE,$E3,$BC
	.dc.b	$E9,$E2,$43,$33,$C2,$3F,$C0,$B1,$E1,$3A,$1E,$27,$33,$D6,$34,$73
	.dc.b	$29,$D0,$E1,$3F,$E4,$87,$07,$42,$1C,$63,$A1,$27,$B9,$27,$94,$F6
	.dc.b	$24,$EE,$3F,$C7,$33,$FC,$52,$7F,$8A,$C5,$9F,$F0,$55,$FF,$03,$9F
	.dc.b	$CD,$9F,$A1,$3F,$81,$CF,$E8,$9F,$EA,$AF,$F3,$37,$24,$3F,$7C,$53
	.dc.b	$F9,$21,$FA,$E7,$C4,$F8,$E3,$F4,$D7,$FC,$CA,$BF,$93,$1F,$83,$9F
	.dc.b	$86,$7C,$63,$E2,$5F,$C1,$9F,$F1,$D5,$FC,$2A,$F8,$11,$82,$1F,$E4
	.dc.b	$8A,$E4,$43,$E3,$D7,$38,$D4,$62,$BA,$4E,$67,$FC,$E1,$18,$23,$D8
	.dc.b	$C9,$67,$FC,$DC,$F4,$AF,$2B,$6B,$A4,$EA,$38,$99,$AE,$AD,$3A,$47
	.dc.b	$F9,$8D,$D9,$1E,$44,$34,$5A,$FF,$30,$C7,$32,$1C,$63,$64,$8D,$C6
	.dc.b	$5B,$1C,$EB,$FD,$3B,$18,$8E,$11,$82,$3F,$E9,$93,$FC,$CA,$FF,$39
	.dc.b	$7F,$82,$9F,$78,$FA,$13,$EE,$9E,$84,$9F,$C9,$4F,$F8,$4A,$7F,$BC
	.dc.b	$DF,$E6,$4F,$E8,$F1,$FE,$01,$8C,$DF,$F3,$F9,$D3,$F9,$99,$FF,$42
	.dc.b	$57,$99,$A7,$9E,$3F,$CE,$95,$FD,$2E,$7F,$84,$4F,$E2,$6F,$ED,$CF
	.dc.b	$C5,$9E,$53,$FE,$84,$BE,$F7,$FD,$09,$B5,$E4,$3F,$C1,$93,$FD,$82
	.dc.b	$3F,$D4,$27,$F9,$32,$33,$1F,$A9,$3F,$52,$23,$F5,$31,$FD,$AE,$7C
	.dc.b	$E7,$F5,$FA,$F3,$D7,$CC,$9F,$BE,$FD,$93,$DC,$7F,$A3,$23,$EA,$5F
	.dc.b	$EC,$63,$F5,$F5,$F9,$2A,$F7,$47,$03,$FA,$E3,$F1,$63,$F9,$41,$B8
	.dc.b	$EC,$7C,$C2,$1F,$82,$8F,$F1,$13,$FC,$91,$FC,$05,$7C,$A9,$F5,$CF
	.dc.b	$F6,$63,$F0,$D5,$F4,$1E,$A7,$3A,$FE,$DB,$5F,$02,$75,$3F,$94,$3D
	.dc.b	$0F,$F0,$E8,$D1,$FB,$28,$D1,$E6,$AF,$0B,$F9,$63,$E1,$D7,$99,$E7
	.dc.b	$3F,$20,$E7,$3F,$06,$39,$9E,$59,$F8,$CF,$BD,$FA,$6E,$A3,$BD,$7D
	.dc.b	$18,$FC,$75,$79,$63,$F4,$51,$E0,$AF,$92,$FC,$49,$D1,$FC,$BA,$BF
	.dc.b	$9F,$29,$87,$F3,$4A,$FF,$11,$BF,$8E,$BF,$CA,$BF,$8B,$BF,$CA,$5F
	.dc.b	$F8,$96,$3F,$CF,$51,$FE,$22,$3F,$A1,$C7,$E7,$4F,$F1,$65,$7C,$D7
	.dc.b	$F0,$73,$F1,$E7,$E6,$1F,$8F,$7D,$2F,$CC,$7E,$1C,$FB,$A3,$67,$E3
	.dc.b	$A7,$F3,$87,$49,$DD,$FE,$1E,$7A,$47,$E0,$A7,$4F,$E6,$23,$9D,$79
	.dc.b	$5F,$A5,$19,$7E,$B4,$F6,$72,$7E,$39,$FA,$B5,$F7,$A7,$D9,$3C,$0F
	.dc.b	$CB,$1E,$B7,$A4,$F0,$3C,$71,$CC,$F3,$C7,$E1,$1F,$B5,$1F,$89,$AF
	.dc.b	$85,$1D,$4F,$C9,$1F,$84,$8D,$57,$C1,$8F,$8C,$78,$8D,$57,$9A,$7C
	.dc.b	$51,$EC,$3F,$20,$78,$9F,$80,$F8,$A3,$83,$93,$A1,$FC,$2D,$F6,$B9
	.dc.b	$3C,$A6,$4F,$F4,$84,$FE,$4D,$B1,$F9,$29,$FF,$74,$1F,$CD,$48,$FE
	.dc.b	$4D,$AF,$F9,$D2,$3F,$9E,$3F,$95,$7F,$92,$1F,$E1,$73,$FC,$22,$43
	.dc.b	$FC,$23,$7F,$CA,$E3,$94,$7F,$AD,$A3,$FA,$3B,$67,$FA,$23,$F3,$EF
	.dc.b	$F6,$4F,$AC,$F8,$E7,$CE,$3D,$EA,$FB,$5F,$EC,$27,$FC,$61,$5F,$B2
	.dc.b	$8F,$81,$1F,$12,$3F,$9D,$AA,$E5,$1F,$E3,$88,$FF,$15,$3F,$EB,$17
	.dc.b	$FD,$48,$F9,$8C,$5F,$F4,$87,$E3,$D7,$ED,$A7,$FC,$76,$7C,$87,$FD
	.dc.b	$2B,$1C,$63,$F7,$AF,$FA,$E9,$EA,$FE,$EE,$3E,$91,$0F,$93,$1D,$0F
	.dc.b	$E0,$47,$F7,$37,$EA,$4F,$F0,$A7,$BC,$F3,$78,$C7,$D0,$9E,$44,$70
	.dc.b	$49,$F3,$47,$DE,$3D,$47,$F1,$09,$FE,$D4,$74,$8D,$CF,$86,$75,$3E
	.dc.b	$49,$F8,$A7,$FB,$E5,$F6,$1F,$1E,$7C,$11,$E7,$8C,$3D,$0E,$33,$FE
	.dc.b	$DF,$71,$1C,$9E,$8F,$8D,$E4,$61,$E6,$E2,$39,$35,$FA,$31,$EC,$36
	.dc.b	$58,$E4,$48,$E3,$19,$6B,$EF,$3C,$8E,$1F,$A6,$6D,$EE,$EC,$F4,$19
	.dc.b	$9F,$C6,$3C,$A3,$6F,$23,$F6,$55,$FC,$AE,$3F,$84,$47,$F3,$47,$FA
	.dc.b	$B3,$FC,$CD,$FE,$21,$1F,$A7,$7F,$3A,$7E,$48,$FF,$75,$1F,$62,$3B
	.dc.b	$11,$C1,$0F,$C8,$3F,$E8,$BA,$FE,$45,$3E,$39,$FE,$64,$7F,$2F,$3F
	.dc.b	$70,$FE,$F0,$FF,$02,$9F,$CB,$A3,$F9,$2C,$7A,$A7,$F8,$EB,$C2,$FF
	.dc.b	$A0,$47,$B6,$3F,$A5,$C7,$F8,$06,$7F,$9E,$C7,$F0,$6A,$F1,$3F,$CC
	.dc.b	$AB,$83,$FC,$36,$B9,$C7,$69,$DC,$FF,$0F,$7C,$8F,$F2,$D7,$D1,$3F
	.dc.b	$DF,$A3,$E6,$CF,$F4,$97,$DE,$78,$63,$C8,$FD,$07,$E2,$98,$7D,$C7
	.dc.b	$C5,$9F,$09,$F1,$23,$73,$FD,$DA,$7D,$4F,$9E,$39,$47,$82,$7C,$EF
	.dc.b	$C7,$AF,$87,$18,$3A,$CF,$A6,$7E,$4D,$75,$7E,$04,$78,$23,$F1,$11
	.dc.b	$E6,$3D,$8E,$E7,$CD,$1F,$E5,$3B,$FE,$00,$FE,$10,$FC,$33,$FE,$D2
	.dc.b	$7F,$A1,$DF,$EE,$9F,$CD,$9A,$7F,$D1,$B1,$B7,$F9,$69,$C6,$3F,$A3
	.dc.b	$3F,$E2,$AA,$FC,$84,$7D,$28,$FE,$D2,$FC,$3B,$F5,$4F,$D5,$8F,$92
	.dc.b	$7C,$F8,$F9,$2E,$0F,$6C,$7C,$07,$91,$D4,$FE,$E3,$7F,$01,$F9,$31
	.dc.b	$B7,$94,$7C,$F9,$EB,$1F,$30,$E4,$7F,$58,$7E,$93,$F9,$3A,$FA,$F1
	.dc.b	$E8,$7D,$71,$D2,$30,$F8,$9E,$31,$FA,$88,$F3,$9E,$4A,$F8,$0F,$84
	.dc.b	$C4,$74,$70,$78,$4F,$CB,$1E,$6A,$FC,$FC,$7F,$87,$A3,$15,$FA,$87
	.dc.b	$E5,$57,$E0,$E3,$C7,$1F,$A7,$74,$FC,$C8,$F1,$D7,$C9,$8F,$01,$F8
	.dc.b	$D7,$E4,$47,$3A,$F0,$4F,$F9,$4A,$FF,$1A,$7E,$15,$FE,$30,$7F,$84
	.dc.b	$4F,$C6,$CF,$F2,$78,$FE,$52,$FF,$2C,$3F,$97,$1F,$CF,$0B,$3F,$CE
	.dc.b	$E7,$F2,$67,$FA,$F5,$F9,$EF,$D7,$7C,$6F,$A0,$F9,$E5,$3E,$B1,$F1
	.dc.b	$9F,$B6,$D3,$E6,$9F,$E2,$57,$F3,$A7,$F8,$06,$7C,$F3,$E5,$3F,$98
	.dc.b	$4F,$EA,$C9,$1F,$05,$FE,$E8,$7F,$80,$9C,$57,$F8,$8E,$7F,$C3,$F3
	.dc.b	$FE,$1D,$BF,$D5,$36,$FF,$C3,$C8,$35,$7D,$CB,$E3,$BF,$61,$39,$8C
	.dc.b	$3F,$E2,$A6,$DF,$FA,$7A,$BA,$47,$17,$FC,$66,$7F,$94,$A3,$FD,$49
	.dc.b	$3F,$EA,$63,$FC,$67,$87,$F8,$CA,$39,$4F,$A4,$E2,$79,$2F,$C6,$F6
	.dc.b	$7D,$B7,$E2,$9F,$F5,$84,$E4,$E3,$5F,$16,$3C,$31,$FE,$43,$3F,$DF
	.dc.b	$B1,$E1,$8E,$31,$E1,$3E,$35,$FC,$1B,$F0,$C7,$8E,$3E,$14,$7F,$94
	.dc.b	$A3,$C0,$70,$3E,$04,$FF,$B0,$70,$FA,$77,$DE,$3F,$CB,$13,$B3,$DC
	.dc.b	$FD,$AC,$3E,$7C,$7D,$53,$E0,$61,$FC,$8D,$EE,$75,$9F,$C0,$D7,$F9
	.dc.b	$72,$3D,$B7,$E5,$7D,$87,$E0,$0F,$95,$3D,$F0,$F9,$66,$F0,$FB,$53
	.dc.b	$E5,$BF,$95,$5E,$BC,$3E,$CE,$1D,$67,$FC,$1D,$5E,$BB,$FC,$55,$F5
	.dc.b	$BF,$8E,$7A,$67,$EE,$1F,$5A,$FE,$6D,$FA,$63,$E0,$3F,$1D,$FF,$4E
	.dc.b	$9F,$84,$BE,$B7,$F7,$23,$EA,$1F,$1C,$FA,$CF,$E3,$23,$D7,$1F,$1E
	.dc.b	$38,$4F,$78,$FC,$A4,$7E,$4A,$3D,$71,$F1,$CF,$AB,$1D,$E7,$EA,$C7
	.dc.b	$D5,$7F,$25,$1F,$87,$8F,$63,$F5,$2B,$F0,$27,$C7,$3F,$D6,$D1,$D2
	.dc.b	$3B,$4F,$48,$FC,$85,$FC,$78,$F4,$BF,$E9,$D9,$F4,$D7,$48,$FD,$ED
	.dc.b	$72,$C3,$F2,$F7,$CE,$3F,$13,$7F,$0E,$7C,$38,$7D,$CB,$FD,$A5,$7E
	.dc.b	$21,$E6,$7E,$86,$BE,$D4,$7C,$DB,$F7,$BF,$92,$9F,$24,$FE,$16,$39
	.dc.b	$3F,$2F,$0F,$7D,$7D,$88,$F3,$57,$E1,$1F,$25,$7C,$87,$DB,$1F,$8E
	.dc.b	$3C,$71,$F9,$88,$FA,$58,$7E,$62,$7E,$33,$EC,$BF,$C5,$9F,$16,$7C
	.dc.b	$55,$F8,$99,$F8,$A7,$18,$F0,$C7,$E7,$E7,$8D,$78,$67,$8C,$FE,$76
	.dc.b	$B8,$CF,$A6,$7F,$84,$DF,$D6,$AF,$E0,$33,$F3,$1F,$CD,$DF,$E0,$5F
	.dc.b	$15,$7C,$C7,$F1,$57,$FB,$13,$D5,$1F,$89,$AF,$13,$FC,$5B,$0F,$CA
	.dc.b	$C7,$D5,$8F,$86,$F8,$DF,$C9,$57,$AD,$E4,$FD,$47,$95,$FD,$4B,$F5
	.dc.b	$CF,$DD,$7C,$4F,$D3,$9F,$B8,$FA,$AB,$E9,$BF,$99,$7F,$2E,$F1,$7E
	.dc.b	$9B,$C5,$F8,$4F,$EB,$A3,$C2,$FC,$27,$D2,$F1,$7E,$9B,$FB,$77,$C2
	.dc.b	$F1,$7F,$94,$BF,$9F,$7E,$53,$FA,$87,$8C,$7E,$50,$F0,$C7,$F6,$1B
	.dc.b	$FE,$18,$FF,$24,$8F,$95,$3F,$94,$8F,$13,$F7,$2B,$E9,$BF,$94,$8F
	.dc.b	$94,$F8,$A7,$89,$F0,$A3,$ED,$BF,$09,$FA,$51,$C5,$F9,$4F,$E3,$1F
	.dc.b	$53,$F0,$9E,$2F,$A9,$FC,$59,$FD,$64,$E2,$FC,$A7,$F2,$8F,$C2,$7F
	.dc.b	$7E,$FA,$5F,$A4,$71,$3E,$51,$FC,$B0,$E2,$F8,$5F,$94,$F1,$7D,$2F
	.dc.b	$D2,$7E,$55,$FE,$6D,$E3,$1F,$25,$F4,$BC,$5F,$B5,$3F,$49,$FF,$0A
	.dc.b	$47,$C1,$8F,$93,$1F,$D7,$8F,$A3,$1F,$9A,$78,$9E,$93,$C3,$1F,$92
	.dc.b	$C3,$F8,$63,$F4,$63,$D3,$3C,$63,$F0,$51,$C4,$F0,$C7,$C9,$7E,$8B
	.dc.b	$E9,$78,$9F,$C7,$4F,$E2,$0E,$5F,$9A,$F5,$73,$1E,$17,$27,$13,$FC
	.dc.b	$69,$1F,$B2,$7E,$CB,$F8,$97,$89,$F0,$4F,$92,$F5,$7F,$5C,$64,$F2
	.dc.b	$9F,$DA,$8F,$A3,$1E,$17,$FB,$0C,$7D,$18,$F8,$31,$FE,$43,$3F,$32
	.dc.b	$7E,$04,$F8,$2F,$19,$FE,$7B,$1D,$4C,$9E,$13,$DA,$66,$3E,$CB,$FD
	.dc.b	$C8,$F4,$9F,$07,$0F,$E9,$CF,$D1,$7F,$62,$7F,$1A,$7E,$09,$C4,$F6
	.dc.b	$CF,$F2,$B7,$E0,$99,$3C,$2E,$4F,$C0,$BE,$53,$E0,$D7,$C9,$9F,$0C
	.dc.b	$7C,$13,$E8,$9F,$06,$BD,$35,$FD,$9E,$38,$9F,$65,$EA,$F8,$4C,$9F
	.dc.b	$04,$F4,$BC,$4F,$92,$7A,$67,$C2,$7F,$83,$E3,$F3,$35,$FA,$17,$8E
	.dc.b	$1F,$D4,$E3,$EC,$46,$2F,$F9,$E3,$E3,$9F,$E6,$EF,$CE,$3C,$47,$E3
	.dc.b	$A7,$F8,$73,$FB,$E3,$F8,$F1,$CC,$FD,$69,$CD,$FE,$10,$FB,$2F,$E6
	.dc.b	$4F,$68,$FA,$6F,$F2,$29,$FE,$05,$1F,$AA,$7F,$34,$FD,$69,$F8,$53
	.dc.b	$FD,$2A,$3F,$2A,$7D,$58,$F8,$B1,$FD,$46,$7C,$31,$89,$FE,$2E,$7D
	.dc.b	$73,$D5,$7F,$87,$7B,$57,$C7,$AF,$D3,$57,$D6,$3C,$37,$FC,$C6,$BF
	.dc.b	$32,$7D,$98,$EC,$FF,$1C,$8F,$E2,$97,$F9,$4C,$3C,$97,$FC,$8E,$BF
	.dc.b	$1F,$86,$23,$F8,$14,$7C,$78,$C5,$FE,$A5,$F8,$11,$FC,$60,$F5,$1F
	.dc.b	$80,$9F,$E2,$13,$F9,$09,$F5,$4F,$D3,$9F,$03,$D4,$FE,$CC,$F8,$30
	.dc.b	$F4,$C7,$5C,$3F,$18,$FD,$33,$F3,$87,$DF,$3F,$0D,$1F,$72,$7E,$7D
	.dc.b	$79,$AF,$C4,$7E,$3A,$BF,$17,$7F,$98,$C3,$FC,$55,$7E,$89,$E8,$7E
	.dc.b	$10,$F2,$D7,$C1,$3E,$81,$D1,$FC,$79,$D2,$7F,$C5,$AF,$F1,$23,$F2
	.dc.b	$A7,$DC,$8F,$16,$1E,$7A,$F8,$6F,$C0,$8E,$93,$EC,$C3,$F2,$91,$F6
	.dc.b	$AF,$FA,$A4,$FE,$02,$BF,$07,$7F,$DA,$E3,$B1,$E4,$8E,$D1,$F3,$6F
	.dc.b	$E2,$4F,$F7,$03,$EC,$9F,$34,$F3,$CF,$E3,$E3,$EC,$C7,$EC,$A3,$17
	.dc.b	$EA,$9F,$C2,$CF,$E8,$5E,$AF,$ED,$EB,$F5,$78,$7E,$46,$3E,$59,$BA
	.dc.b	$F9,$6F,$D0,$3F,$26,$68,$FC,$E5,$74,$9F,$B1,$3F,$E4,$8A,$FD,$AD
	.dc.b	$F9,$2F,$E0,$5F,$91,$E3,$5D,$9F,$8A,$7C,$D3,$F6,$77,$F8,$E3,$CC
	.dc.b	$FE,$F0,$E3,$7F,$35,$1F,$EE,$83,$FD,$BC,$7F,$AD,$67,$FD,$84,$D3
	.dc.b	$FD,$BD,$03,$FD,$78,$F0,$3F,$DC,$51,$C1,$FF,$54,$9D,$A3,$81,$FE
	.dc.b	$B7,$78,$1E,$23,$FD,$C4,$77,$30,$F0,$8C,$47,$96,$39,$BF,$EC,$B3
	.dc.b	$C4,$ED,$F0,$9E,$47,$6E,$E3,$FD,$A5,$1B,$8F,$14,$73,$7E,$2B,$EB
	.dc.b	$72,$7C,$28,$F8,$2F,$AC,$EC,$FC,$57,$94,$76,$8D,$BD,$63,$DE,$79
	.dc.b	$A3,$C4,$FC,$43,$C2,$FA,$1F,$82,$7D,$73,$C0,$7A,$1D,$3C,$DE,$EF
	.dc.b	$89,$CB,$EA,$3E,$6B,$A7,$E2,$3A,$7F,$11,$7F,$B4,$8F,$F2,$37,$E4
	.dc.b	$7F,$E7,$23,$EC,$43,$04,$7D,$38,$C3,$F9,$77,$F3,$95,$F9,$73,$F2
	.dc.b	$31,$CD,$FC,$9C,$70,$7F,$D3,$D1,$C6,$7F,$AF,$47,$EB,$0E,$27,$F9
	.dc.b	$02,$3C,$71,$E2,$8E,$35,$FC,$2E,$BC,$31,$FC,$8C,$E2,$7F,$BF,$6B
	.dc.b	$89,$E9,$AF,$5D,$71,$7C,$33,$FC,$FC,$E3,$5F,$EE,$7A,$E3,$1E,$0A
	.dc.b	$FF,$02,$BF,$B6,$3C,$07,$94,$F0,$1F,$02,$B0,$F8,$A3,$C0,$7F,$42
	.dc.b	$BF,$EC,$D5,$F1,$E5,$7F,$1D,$FF,$6A,$C7,$C7,$AE,$07,$FA,$D9,$E1
	.dc.b	$3B,$9E,$D3,$E6,$8F,$8B,$1F,$34,$F3,$3E,$D7,$CD,$3F,$14,$ED,$3F
	.dc.b	$D0,$2F,$FB,$5C,$76,$7D,$73,$C2,$7D,$66,$EF,$D7,$1F,$64,$EA,$7C
	.dc.b	$98,$FF,$03,$CF,$D0,$BF,$ED,$77,$D4,$F1,$4F,$43,$FA,$25,$F4,$BF
	.dc.b	$9A,$7C,$6B,$F1,$B6,$BD,$35,$C0,$F8,$D1,$E3,$34,$78,$CF,$F0,$03
	.dc.b	$B3,$F4,$6E,$EF,$E6,$3E,$F8,$F9,$95,$E9,$3E,$2C,$6D,$F1,$CF,$E9
	.dc.b	$A7,$D5,$7F,$DD,$5F,$1D,$7A,$A7,$AB,$E9,$8D,$BF,$D4,$5D,$C7,$D7
	.dc.b	$8F,$49,$F8,$48,$DB,$FD,$09,$F4,$BD,$6B,$C6,$7E,$82,$3D,$27,$E2
	.dc.b	$8D,$CF,$53,$C6,$7A,$5D,$9F,$74,$F1,$C7,$F4,$07,$F1,$6F,$6B,$FA
	.dc.b	$37,$EF,$9E,$B3,$B9,$FD,$5D,$FA,$67,$A9,$F4,$4F,$F6,$84,$7E,$0E
	.dc.b	$3E,$B1,$FD,$F6,$7E,$85,$7E,$0E,$FD,$11,$E1,$BF,$87,$1F,$C7,$5E
	.dc.b	$75,$F5,$6F,$F1,$B1,$E0,$BF,$33,$A9,$F9,$13,$A8,$FC,$B9,$F1,$CD
	.dc.b	$57,$82,$35,$3F,$8A,$AE,$6F,$F0,$13,$FB,$14,$7A,$DF,$81,$5D,$0F
	.dc.b	$98,$7C,$63,$FB,$5C,$7C,$68,$FD,$DC,$6C,$F8,$CF,$53,$F8,$BD,$FC
	.dc.b	$D7,$F5,$E7,$63,$C8,$7E,$2E,$3B,$1F,$DF,$E7,$B4,$7E,$76,$3B,$57
	.dc.b	$AA,$3F,$DA,$53,$E9,$7C,$67,$43,$F2,$13,$F1,$2B,$C5,$1F,$96,$36
	.dc.b	$7D,$98,$F9,$4F,$A4,$FC,$4B,$F0,$1F,$49,$C2,$FF,$C8,$B7,$E3,$7F
	.dc.b	$AA,$47,$CC,$3E,$E1,$F8,$67,$6F,$BE,$7C,$87,$57,$C8,$7B,$CF,$15
	.dc.b	$FB,$5C,$1F,$00,$EA,$7C,$B3,$D5,$1F,$60,$EA,$7C,$B7,$D2,$72,$9F
	.dc.b	$3B,$F5,$A3,$C5,$1F,$AE,$BF,$DA,$47,$23,$F5,$75,$F8,$CA,$EF,$1F
	.dc.b	$80,$7C,$87,$E5,$8F,$81,$3F,$E2,$F7,$E5,$C7,$8D,$F7,$9F,$C3,$5F
	.dc.b	$C0,$9D,$9E,$27,$F9,$6E,$BF,$50,$FC,$43,$E0,$9F,$4E,$7B,$57,$C7
	.dc.b	$8E,$95,$E9,$8F,$C5,$C6,$0F,$C3,$57,$47,$D2,$FE,$A5,$F0,$9F,$16
	.dc.b	$7D,$11,$F1,$67,$F7,$D5,$F8,$C8,$C4,$FE,$FA,$3F,$D0,$31,$F3,$AB
	.dc.b	$E0,$CF,$B5,$FB,$11,$9A,$F4,$3F,$C7,$67,$D1,$3F,$60,$F4,$1F,$42
	.dc.b	$FF,$21,$7F,$02,$3E,$19,$F7,$5C,$3F,$4C,$EE,$E2,$BD,$D1,$83,$E7
	.dc.b	$47,$A0,$D1,$ED,$3F,$8D,$57,$D6,$BF,$97,$5E,$7A,$FA,$87,$73,$EA
	.dc.b	$1D,$2F,$FD,$40,$FE,$94,$D1,$E7,$31,$1E,$47,$89,$F4,$6B,$E4,$38
	.dc.b	$3F,$D7,$57,$E7,$78,$BE,$7B,$FC,$B5,$7C,$D9,$F8,$A7,$CA,$70,$F8
	.dc.b	$8C,$4F,$FA,$86,$3F,$08,$68,$F2,$47,$9C,$F2,$4F,$C2,$9F,$D3,$C7
	.dc.b	$C2,$9D,$BD,$0F,$84,$7F,$AB,$8F,$3D,$7E,$C2,$7E,$6B,$F8,$08,$FD
	.dc.b	$2C,$FC,$D7,$F2,$B3,$ED,$7F,$2A,$7C,$2B,$F9,$AF,$90,$FC,$44,$F1
	.dc.b	$3D,$53,$C6,$3F,$00,$6C,$EE,$FA,$9F,$C7,$1F,$C5,$CF,$63,$C5,$FA
	.dc.b	$57,$F9,$88,$E1,$3D,$AF,$D9,$3F,$4A,$7E,$61,$F5,$EB,$C1,$5F,$A8
	.dc.b	$AF,$C6,$47,$A2,$79,$9F,$92,$8E,$71,$F0,$2F,$FD,$FE,$FF,$97,$AF
	.dc.b	$C4,$7F,$8C,$DF,$F6,$14,$FF,$0F,$7F,$CE,$27,$F1,$33,$DC,$58,$FF
	.dc.b	$54,$47,$F9,$29,$FF,$54,$1F,$EB,$38,$E2,$7F,$4A,$29,$FD,$AA,$3F
	.dc.b	$D3,$4F,$F9,$49,$FE,$78,$E8,$FF,$48,$BA,$8E,$E7,$86,$3F,$D8,$4F
	.dc.b	$77,$94,$7A,$5F,$F2,$59,$C9,$FF,$60,$3F,$ED,$38,$E6,$FF,$00,$3D
	.dc.b	$91,$DD,$FF,$65,$3E,$D7,$BB,$97,$FD,$6A,$E5,$E8,$E5,$F9,$6E,$5E
	.dc.b	$AF,$89,$F5,$1C,$A2,$BA,$8F,$96,$F8,$A3,$2F,$F3,$88,$F1,$5F,$EA
	.dc.b	$0F,$CC,$5F,$F4,$9F,$C9,$3F,$D5,$F3,$FE,$FC,$7F,$D6,$27,$FA,$A4
	.dc.b	$FE,$66,$FE,$1A,$BF,$3A,$23,$8C,$0C,$47,$FA,$8E,$39,$47,$F9,$28
	.dc.b	$E5,$1F,$C3,$E3,$8C,$70,$8F,$F3,$F4,$70,$8F,$CC,$C7,$EA,$AF,$FC
	.dc.b	$A4,$FF,$82,$E7,$89,$E8,$7F,$25,$7F,$ED,$B8,$F4,$D7,$F4,$EB,$FE
	.dc.b	$07,$5C,$4F,$F5,$C4,$7C,$17,$D0,$72,$8F,$74,$7A,$4E,$57,$FB,$19
	.dc.b	$12,$BC,$6F,$23,$FB,$B5,$F8,$EB,$E3,$1F,$ED,$E8,$E2,$3C,$CF,$D3
	.dc.b	$9F,$AB,$3F,$58,$E9,$FC,$D1,$A7,$FC,$04,$E9,$F1,$BA,$7F,$2A,$F3
	.dc.b	$74,$E5,$FC,$AB,$97,$BB,$97,$F2,$AE,$5F,$24,$7E,$D2,$3C,$51,$E5
	.dc.b	$3A,$C6,$5F,$09,$E6,$32,$FE,$75,$C1,$E2,$29,$D9,$D3,$D4,$D1,$4D
	.dc.b	$3F,$05,$F9,$47,$80,$D1,$F1,$9F,$23,$C8,$F8,$E7,$91,$E5,$1E,$C8
	.dc.b	$C9,$F1,$23,$F8,$9F,$47,$F1,$C3,$F9,$01,$E1,$3F,$87,$19,$8F,$F1
	.dc.b	$B3,$98,$E3,$1F,$C2,$23,$8B,$F6,$5E,$2E,$4E,$26,$63,$8B,$FC,$32
	.dc.b	$38,$BF,$86,$78,$B9,$38,$99,$8E,$2F,$F1,$08,$E2,$FA,$A3,$2F,$FA
	.dc.b	$FE,$32,$FA,$A3,$27,$E6,$A3,$31,$F9,$A7,$C2,$F5,$3C,$27,$58,$FB
	.dc.b	$6F,$EA,$23,$E3,$3C,$4F,$74,$69,$46,$9E,$06,$A3,$F0,$1E,$4F,$F4
	.dc.b	$E3,$FA,$91,$87,$FA,$4B,$98,$FF,$34,$39,$8E,$CE,$63,$F6,$6E,$63
	.dc.b	$AC,$7F,$3E,$8E,$AF,$EB,$DF,$E9,$87,$17,$F8,$1B,$C5,$D1,$C4,$F0
	.dc.b	$9C,$63,$CA,$E4,$F0,$99,$7C,$AE,$4E,$A6,$5F,$2B,$FD,$3A,$3C,$AF
	.dc.b	$F0,$87,$CA,$F1,$0F,$95,$D1,$97,$0F,$23,$FA,$AB,$5F,$93,$1F,$DD
	.dc.b	$3F,$7B,$FE,$72,$7F,$1A,$C1,$9A,$3F,$8C,$C6,$1F,$E7,$34,$8F,$F3
	.dc.b	$D4,$72,$8F,$E5,$71,$B7,$F9,$94,$70,$8F,$F4,$07,$EF,$7F,$D2,$0F
	.dc.b	$05,$1E,$ED,$BF,$E5,$F7,$FE,$D1,$BE,$CF,$86,$7C,$CF,$69,$DB,$93
	.dc.b	$F9,$AC,$F0,$BF,$24,$7F,$3C,$AF,$75,$70,$9F,$5C,$FF,$4C,$8F,$EC
	.dc.b	$95,$E6,$3C,$A7,$AE,$BF,$D7,$57,$EB,$AF,$83,$1E,$B7,$FA,$94,$6D
	.dc.b	$F1,$9D,$6B,$D2,$FA,$EB,$AC,$78,$EB,$FD,$8D,$1E,$97,$73,$E3,$9D
	.dc.b	$3D,$0F,$88,$FF,$56,$9F,$44,$7C,$58,$FF,$12,$DF,$D2,$8F,$15,$FD
	.dc.b	$23,$E8,$9E,$77,$FB,$2D,$78,$4F,$2D,$72,$BF,$2C,$FF,$8A,$A7,$FC
	.dc.b	$94,$7F,$3B,$3F,$C7,$07,$F9,$28,$CB,$FD,$7D,$CB,$C4,$E6,$7F,$68
	.dc.b	$72,$FF,$8C,$9E,$2E,$67,$F3,$4F,$13,$E1,$1E,$27,$2F,$13,$A9,$97
	.dc.b	$F8,$B4,$79,$63,$27,$D0,$8E,$2F,$F8,$FA,$3C,$2F,$18,$F7,$3E,$13
	.dc.b	$E3,$9E,$D7,$8B,$FE,$08,$7E,$CB,$0F,$CA,$BA,$3E,$C3,$E2,$32,$F9
	.dc.b	$9E,$C7,$13,$EA,$BE,$EB,$FD,$CB,$CA,$3F,$18,$7E,$05,$FF,$0C,$3F
	.dc.b	$E2,$0F,$DF,$3F,$C7,$27,$EC,$8F,$D8,$9F,$81,$7F,$2B,$3D,$27,$F4
	.dc.b	$71,$F9,$EA,$FC,$1C,$74,$85,$74,$7F,$46,$F4,$8F,$CB,$C7,$C6,$3F
	.dc.b	$2E,$F4,$7C,$E7,$ED,$67,$C5,$3C,$E3,$F8,$03,$CE,$7F,$25,$1F,$A1
	.dc.b	$AF,$34,$7A,$2F,$E7,$47,$F1,$98,$F0,$CF,$98,$FD,$95,$7C,$17,$FC
	.dc.b	$90,$FC,$1B,$FD,$04,$7C,$93,$F7,$A7,$F9,$02,$3C,$95,$FC,$A1,$F2
	.dc.b	$57,$C3,$AE,$35,$EB,$8F,$0C,$F0,$BF,$14,$FF,$13,$8F,$03,$FC,$16
	.dc.b	$BF,$72,$F1,$AF,$7B,$F6,$4F,$49,$F3,$8F,$4B,$EE,$8F,$0C,$7F,$07
	.dc.b	$8E,$2F,$82,$FE,$3D,$7C,$18,$F1,$4F,$17,$C4,$7D,$F3,$E1,$1F,$CB
	.dc.b	$4F,$11,$F6,$9E,$97,$EC,$8F,$C6,$47,$37,$57,$EC,$7C,$91,$F8,$9A
	.dc.b	$F8,$87,$CA,$8F,$E8,$07,$88,$FC,$29,$C0,$FA,$67,$8A,$FD,$6F,$C3
	.dc.b	$3D,$73,$F0,$2B,$E1,$C7,$E6,$63,$E9,$BF,$28,$FC,$69,$F2,$9E,$0F
	.dc.b	$86,$7F,$A4,$BE,$13,$E1,$1F,$02,$3F,$D0,$B5,$F2,$E3,$E0,$C7,$27
	.dc.b	$E0,$9F,$2C,$FB,$8F,$C1,$7C,$53,$F8,$B7,$C5,$1F,$8B,$7C,$57,$F9
	.dc.b	$C3,$FA,$CC,$7D,$CA,$F9,$33,$C0,$F9,$27,$80,$F9,$27,$C1,$8E,$0F
	.dc.b	$82,$7E,$C3,$FB,$88,$FB,$15,$FA,$CB,$FB,$13,$C6,$7D,$0F,$E0,$6F
	.dc.b	$E0,$3C,$63,$E0,$57,$18,$F8,$07,$C9,$7F,$C1,$EF,$EA,$CF,$14,$FE
	.dc.b	$05,$F9,$B5,$F6,$1F,$EC,$D5,$F6,$1F,$01,$F7,$CF,$9E,$FC,$77,$E0
	.dc.b	$57,$D8,$7D,$07,$E2,$0F,$B1,$1F,$EE,$1A,$F4,$C7,$F0,$E8,$F8,$37
	.dc.b	$FC,$E6,$FE,$0C,$F8,$A7,$E4,$BE,$23,$F1,$27,$E4,$1F,$15,$FC,$18
	.dc.b	$FA,$27,$18,$F9,$31,$F9,$88,$E2,$7A,$8F,$B2,$F1,$3D,$4F,$13,$D5
	.dc.b	$7F,$80,$7C,$07,$0B,$F4,$3F,$40,$F8,$27,$E6,$23,$83,$DC,$F9,$E7
	.dc.b	$E3,$AF,$F9,$EB,$D8,$FA,$A7,$E3,$4D,$9F,$3A,$7F,$29,$1F,$CE,$23
	.dc.b	$67,$90,$DD,$7A,$E7,$77,$F6,$E3,$E1,$1F,$A1,$9E,$0F,$F8,$D1,$ED
	.dc.b	$5E,$47,$D4,$76,$3D,$51,$B3,$B4,$6D,$FC,$7C,$7F,$07,$AF,$EE,$97
	.dc.b	$F8,$09,$FE,$75,$7F,$80,$3F,$B5,$3F,$40,$FE,$94,$FE,$01,$F3,$CF
	.dc.b	$EA,$2B,$F1,$13,$FA,$08,$FA,$31,$F8,$09,$E3,$3F,$80,$9F,$0C,$FE
	.dc.b	$64,$FD,$44,$7E,$22,$38,$CF,$85,$FD,$FC,$71,$3F,$87,$CF,$13,$C2
	.dc.b	$F1,$3D,$31,$F3,$9F,$43,$E0,$8F,$C7,$9E,$03,$F1,$11,$C2,$FE,$44
	.dc.b	$A7,$F6,$C7,$E7,$EB,$EB,$3F,$42,$3F,$C7,$31,$F4,$23,$FA,$3B,$F4
	.dc.b	$1F,$D7,$D7,$F4,$6B,$FD,$F4,$FD,$7A,$F7,$9E,$79,$F7,$1F,$B5,$AF
	.dc.b	$74,$FD,$68,$F9,$C7,$9E,$3B,$CF,$9D,$FB,$D1,$F1,$9F,$79,$F1,$8F
	.dc.b	$7B,$FE,$11,$9F,$9D,$7F,$99,$9F,$75,$7D,$D9,$F0,$57,$CD,$38,$47
	.dc.b	$9E,$7F,$4C,$70,$BF,$6C,$7C,$43,$DB,$7F,$32,$7E,$5D,$FE,$2A,$B7
	.dc.b	$1F,$E1,$BA,$DB,$EC,$36,$FD,$99,$FA,$31,$F6,$4F,$D6,$C7,$07,$FA
	.dc.b	$69,$E4,$8E,$CF,$EE,$5F,$A1,$7F,$26,$36,$7E,$B0,$F5,$1B,$9F,$54
	.dc.b	$6E,$3F,$9D,$47,$CF,$7E,$ED,$7E,$AE,$3F,$C8,$73,$FA,$B3,$E3,$DF
	.dc.b	$C1,$AF,$AB,$1F,$75,$FA,$B7,$F3,$9F,$AB,$1F,$E3,$B9,$E1,$1D,$8F
	.dc.b	$C1,$D7,$69,$FC,$19,$DA,$3E,$41,$D9,$FC,$83,$EB,$3E,$73,$EB,$7F
	.dc.b	$92,$D7,$E5,$EF,$F9,$2C,$7C,$D8,$F8,$F3,$D6,$3E,$3C,$6D,$FE,$24
	.dc.b	$6E,$BE,$3C,$EE,$3E,$39,$F6,$9F,$CA,$4F,$DA,$8F,$CA,$1F,$6A,$3C
	.dc.b	$8F,$F3,$F3,$EA,$47,$DA,$7F,$92,$CF,$69,$F8,$D1,$F0,$0F,$8C,$7C
	.dc.b	$C8,$F8,$D1,$F9,$93,$71,$EA,$8F,$B5,$5F,$02,$3C,$75,$F0,$2B,$71
	.dc.b	$EA,$3E,$64,$FD,$F3,$FC,$B2,$FF,$36,$3C,$4F,$39,$FB,$91,$ED,$9F
	.dc.b	$88,$FD,$0A,$F0,$BF,$28,$F0,$BE,$DB,$FE,$E3,$5F,$CA,$A7,$F3,$93
	.dc.b	$FA,$33,$F6,$C7,$B9,$FC,$4C,$F0,$8F,$E5,$31,$C2,$BE,$8C,$70,$BF
	.dc.b	$4D,$FF,$19,$AF,$88,$7D,$7A,$FC,$B4,$7D,$73,$D6,$7C,$5B,$EC,$7F
	.dc.b	$09,$7E,$B4,$FF,$25,$3F,$1F,$3F,$01,$FB,$91,$F8,$97,$F5,$E7,$DC
	.dc.b	$37,$1F,$41,$F8,$C7,$8D,$FC,$DB,$B3,$FD,$59,$1E,$97,$EB,$1F,$82
	.dc.b	$7D,$17,$F1,$63,$E9,$4F,$57,$C7,$1F,$49,$F1,$9E,$93,$AB,$FC,$35
	.dc.b	$EB,$7F,$76,$3A,$57,$C3,$9F,$F0,$81,$D0,$F4,$4F,$E1,$EB,$F8,$21
	.dc.b	$E1,$9F,$B9,$3F,$9E,$AF,$B9,$5F,$C9,$23,$C3,$5F,$A4,$AD,$9F,$24
	.dc.b	$F9,$F1,$FB,$13,$E4,$DF,$D4,$9F,$F0,$75,$F4,$9F,$E0,$47,$48,$FA
	.dc.b	$95,$D2,$7E,$13,$D2,$7F,$26,$74,$3D,$AE,$DF,$69,$FC,$52,$FF,$17
	.dc.b	$3F,$1E,$3E,$C9,$FA,$83,$55,$F0,$AF,$CC,$FF,$76,$3F,$2B,$5F,$06
	.dc.b	$7F,$DB,$91,$F0,$6B,$F3,$71,$F1,$E3,$6F,$B0,$DB,$EC,$8D,$DF,$C7
	.dc.b	$7E,$14,$FE,$24,$DC,$FD,$38,$D9,$F0,$8D,$C7,$E0,$CD,$DF,$A2,$7F
	.dc.b	$BB,$9F,$94,$8D,$47,$F9,$E6,$BE,$34,$7C,$13,$F8,$A5,$FC,$6A,$F6
	.dc.b	$57,$EA,$6F,$BD,$7D,$B3,$EB,$4F,$96,$74,$FC,$08,$F5,$47,$79,$F4
	.dc.b	$C7,$D4,$3B,$CF,$CF,$7E,$CC,$69,$F2,$5F,$E9,$AF,$C9,$5F,$26,$FB
	.dc.b	$BF,$C0,$AF,$F8,$43,$CE,$FF,$C7,$D1,$CD,$FD,$83,$A8,$FB,$87,$1A
	.dc.b	$FB,$87,$EA,$4E,$EF,$B4,$FC,$B9,$ED,$9F,$A1,$1B,$8C,$9C,$E3,$F0
	.dc.b	$15,$F0,$4F,$3C,$7D,$38,$F9,$47,$92,$7C,$2F,$F2,$FB,$FE,$37,$7F
	.dc.b	$CB,$E3,$E2,$DF,$2A,$FD,$C5,$FA,$8E,$93,$FA,$43,$E4,$D7,$BE,$3F
	.dc.b	$A7,$57,$E1,$67,$F9,$84,$79,$DF,$C8,$57,$AE,$3A,$3D,$9E,$31,$E2
	.dc.b	$7D,$E7,$82,$FC,$F1,$F7,$5F,$B1,$1A,$9E,$53,$FD,$08,$E5,$1D,$9F
	.dc.b	$D8,$B5,$FC,$2B,$67,$9B,$FE,$0C,$3E,$A1,$9B,$F7,$4F,$B2,$3C,$91
	.dc.b	$FC,$88,$FD,$43,$F6,$63,$E8,$DF,$D5,$8F,$34,$78,$9F,$D4,$DF,$A2
	.dc.b	$3F,$79,$1F,$1D,$FD,$59,$FE,$B5,$3F,$B6,$9F,$66,$BC,$0F,$C3,$7C
	.dc.b	$AF,$28,$F2,$CF,$E3,$AB,$EA,$4E,$63,$C1,$7F,$C1,$6B,$D1,$3F,$74
	.dc.b	$FC,$41,$F8,$E3,$C4,$7F,$66,$3F,$03,$3F,$8C,$7F,$A4,$9F,$90,$7E
	.dc.b	$F4,$7D,$C8,$FD,$3D,$7E,$35,$F8,$47,$CB,$9F,$1D,$FC,$AA,$F8,$53
	.dc.b	$F2,$8F,$09,$FD,$48,$F8,$53,$CD,$FB,$71,$F3,$1E,$A7,$F6,$0A,$F8
	.dc.b	$37,$EF,$9E,$B3,$EF,$AF,$3B,$E3,$38,$CF,$F5,$D7,$D3,$3F,$69,$EE
	.dc.b	$6C,$A7,$CD,$38,$DF,$EB,$5E,$31,$F8,$BA,$F9,$15,$FD,$3C,$FC,$EC
	.dc.b	$FC,$C3,$E8,$1F,$E2,$D3,$E8,$1F,$50,$FD,$59,$F0,$A7,$C9,$1F,$9A
	.dc.b	$7E,$83,$FE,$9E,$73,$5F,$86,$8F,$85,$1F,$86,$33,$5F,$D4,$6F,$F3
	.dc.b	$11,$E7,$BF,$E9,$D3,$F1,$9F,$71,$F0,$2B,$E3,$1D,$E7,$84,$79,$8F
	.dc.b	$C2,$9F,$3C,$F4,$3F,$AF,$AF,$AB,$7F,$81,$8F,$21,$CA,$7F,$91,$9C
	.dc.b	$1F,$6C,$7F,$AC,$E3,$C7,$5F,$05,$FA,$8D,$73,$1F,$04,$FA,$91,$E7
	.dc.b	$9F,$C4,$4F,$AE,$7C,$57,$F9,$48,$FD,$73,$F1,$5F,$C9,$9F,$DF,$8F
	.dc.b	$A8,$79,$8F,$7D,$7D,$C7,$27,$E6,$CF,$B8,$7E,$6C,$DC,$FE,$F1,$FC
	.dc.b	$11,$F2,$63,$F0,$D1,$98,$FE,$E6,$7F,$1E,$AF,$4C,$FB,$CF,$BB,$5E
	.dc.b	$99,$E6,$7F,$0C,$3C,$27,$A5,$FB,$91,$C2,$3F,$32,$7C,$03,$F2,$6F
	.dc.b	$F9,$3D,$E0,$7D,$33,$E7,$D7,$CD,$78,$1F,$DD,$8F,$8E,$7F,$91,$4F
	.dc.b	$19,$FC,$1C,$E4,$7D,$69,$F9,$CF,$98,$E0,$7C,$A3,$F1,$47,$C8,$7F
	.dc.b	$81,$4F,$C1,$9F,$BA,$7E,$15,$F6,$4F,$DD,$3C,$A7,$D7,$7E,$6C,$FF
	.dc.b	$27,$8F,$98,$7B,$1F,$DB,$CF,$F8,$5A,$3C,$33,$C2,$BF,$C0,$8F,$E2
	.dc.b	$E3,$81,$FA,$83,$DA,$F8,$8F,$33,$FE,$4D,$3F,$82,$C7,$DA,$9F,$95
	.dc.b	$1F,$18,$EE,$F8,$67,$F1,$97,$FB,$6A,$E1,$7E,$67,$F9,$DC,$FE,$46
	.dc.b	$3E,$ED,$7C,$4A,$FF,$76,$C6,$2F,$E1,$9F,$4C,$F9,$97,$EC,$AF,$94
	.dc.b	$7E,$92,$3F,$0A,$79,$6F,$B3,$EE,$8E,$CF,$F5,$83,$9C,$FE,$ED,$ED
	.dc.b	$3F,$26,$3F,$81,$47,$D0,$3F,$04,$7C,$90,$44,$4A,$CC,$46,$B0,$FF
	.dc.b	$51,$15,$C4,$72,$0E,$63,$81,$E0,$75,$18,$8C,$1D,$9E,$51,$C6,$38
	.dc.b	$BC,$1E,$0F,$82,$3D,$11,$FE,$B1,$3C,$8F,$8D,$D9,$D5,$DB,$ED,$39
	.dc.b	$BB,$7B,$9B,$7D,$6F,$53,$FD,$FE,$ED,$F3,$1F,$2D,$EA,$FB,$0F,$4B
	.dc.b	$B7,$93,$EE,$7A,$1F,$3D,$D3,$E5,$3E,$9B,$B7,$07,$B9,$FB,$2F,$B8
	.dc.b	$DB,$E2,$7D,$27,$89,$DB,$EE,$36,$E9,$F4,$1D,$5F,$AC,$E4,$FB,$31
	.dc.b	$F0,$1F,$87,$1D,$1F,$B1,$3F,$93,$8F,$C5,$AA,$FC,$5F,$5E,$1F,$C1
	.dc.b	$B4,$FC,$3C,$6D,$FE,$27,$1F,$BA,$A4,$71,$AF,$C8,$3C,$63,$F2,$D5
	.dc.b	$C6,$7F,$B9,$C7,$FA,$F6,$78,$47,$E8,$1E,$11,$F9,$E9,$E1,$3F,$AD
	.dc.b	$8F,$F1,$D4,$F0,$8F,$0C,$7E,$AA,$7C,$CF,$82,$BC,$B3,$E0,$8F,$F6
	.dc.b	$35,$71,$3B,$BE,$F8,$D3,$FE,$C5,$8D,$3D,$67,$BB,$E5,$7A,$CF,$96
	.dc.b	$BF,$B2,$57,$0A,$47,$91,$E1,$1F,$2A,$FF,$DC,$8E,$E7,$53,$E3,$AE
	.dc.b	$97,$FD,$96,$B4,$F8,$A7,$53,$FC,$99,$F9,$87,$9D,$FA,$6F,$A6,$7E
	.dc.b	$DC,$FA,$27,$84,$FC,$43,$83,$F4,$DE,$07,$CB,$9F,$43,$D8,$F6,$CF
	.dc.b	$E2,$1D,$9E,$E9,$F9,$6F,$FA,$E2,$7D,$27,$88,$E6,$F8,$4E,$71,$E8
	.dc.b	$7A,$C6,$E3,$9C,$7E,$08,$F3,$BF,$52,$BE,$4C,$7B,$E3,$71,$F5,$23
	.dc.b	$73,$F8,$07,$53,$F3,$AF,$F8,$E4,$7B,$0F,$E6,$33,$E8,$7E,$14,$F4
	.dc.b	$BF,$A9,$1D,$0F,$82,$7D,$43,$FB,$39,$ED,$BF,$D8,$4F,$B5,$F3,$C7
	.dc.b	$BA,$3A,$5F,$8A,$7F,$6D,$3E,$23,$B1,$F1,$0F,$1C,$FA,$4F,$E5,$11
	.dc.b	$C2,$3F,$A6,$DF,$A5,$FA,$93,$B3,$F8,$8C,$EE,$3E,$55,$FD,$B7,$B1
	.dc.b	$C2,$7F,$0E,$FE,$94,$F1,$CF,$A8,$FD,$BC,$6A,$3F,$00,$6A,$7F,$30
	.dc.b	$FD,$08,$D3,$F6,$CF,$C0,$BF,$91,$7C,$27,$CB,$3F,$B4,$9F,$2E,$3C
	.dc.b	$D3,$E1,$7F,$39,$1F,$40,$F3,$1D,$27,$DB,$5F,$8B,$7E,$11,$FC,$86
	.dc.b	$3E,$71,$B8,$FC,$33,$F5,$63,$6F,$F0,$33,$75,$F7,$23,$D0,$F8,$E7
	.dc.b	$F3,$B3,$E2,$8D,$4F,$8A,$3E,$CB,$F0,$A3,$F2,$D1,$E1,$3F,$16,$7C
	.dc.b	$73,$D8,$7C,$C3,$47,$C6,$34,$7C,$63,$C1,$5F,$18,$FA,$2F,$AE,$7C
	.dc.b	$11,$F9,$A8,$E7,$3E,$99,$FC,$99,$F8,$C9,$FE,$8F,$3F,$6E,$BE,$54
	.dc.b	$F2,$BF,$CC,$56,$AB,$9B,$E0,$BF,$C2,$C7,$A8,$FC,$03,$F1,$2F,$E7
	.dc.b	$BA,$8F,$51,$EF,$74,$FC,$43,$D4,$73,$7E,$49,$A7,$E2,$1A,$39,$9E
	.dc.b	$09,$F8,$93,$EA,$8E,$51,$F0,$E7,$91,$F0,$EF,$F0,$33,$FC,$14,$FD
	.dc.b	$69,$F0,$CF,$B6,$72,$BF,$86,$FE,$1A,$BB,$BA,$3C,$75,$E8,$8C,$C7
	.dc.b	$EB,$CE,$55,$F4,$5E,$4F,$F0,$98,$F9,$87,$E8,$23,$9B,$E9,$39,$BF
	.dc.b	$A9,$9F,$F3,$B4,$7F,$79,$BF,$E0,$93,$98,$FF,$31,$3F,$E8,$B3,$DA
	.dc.b	$7D,$A8,$F8,$CF,$6B,$F6,$BA,$AF,$8F,$5E,$4A,$FD,$0C,$FD,$98,$F1
	.dc.b	$BF,$96,$9F,$1D,$7F,$62,$8F,$94,$7C,$13,$B5,$FC,$17,$B5,$72,$7E
	.dc.b	$E5,$72,$7F,$AA,$3E,$2B,$F3,$B9,$8F,$E1,$15,$F2,$67,$35,$F0,$2B
	.dc.b	$FD,$33,$5D,$AB,$4F,$D7,$8E,$4F,$EB,$23,$2F,$9E,$32,$FB,$E3,$2F
	.dc.b	$CD,$3F,$D2,$67,$E4,$5C,$BF,$0E,$7C,$47,$D5,$7E,$19,$C4,$FF,$4A
	.dc.b	$3F,$76,$38,$C7,$C2,$8E,$37,$F2,$E3,$EC,$1E,$98,$CB,$F1,$9F,$95
	.dc.b	$5F,$AF,$3D,$31,$E5,$AF,$E3,$73,$F6,$DF,$E0,$67,$F4,$87,$F7,$D3
	.dc.b	$F3,$E7,$B1,$EE,$7F,$39,$3E,$67,$B4,$F8,$A7,$E9,$CF,$6B,$FA,$71
	.dc.b	$C6,$7D,$D3,$98,$EA,$7C,$2B,$FC,$8B,$FD,$62,$3B,$4F,$E8,$EF,$D6
	.dc.b	$78,$A3,$94,$FD,$79,$F1,$3F,$6E,$32,$FC,$F3,$FD,$E9,$3C,$63,$F9
	.dc.b	$C4,$71,$3C,$D5,$E2,$BF,$D5,$C7,$DB,$33,$3F,$99,$7C,$CF,$67,$47
	.dc.b	$AD,$D4,$72,$8F,$B8,$72,$8F,$A0,$FF,$02,$79,$3A,$39,$3A,$7F,$34
	.dc.b	$EA,$BE,$8C,$73,$8F,$EC,$0E,$9F,$6B,$F0,$6B,$FD,$D4,$E6,$BF,$9A
	.dc.b	$DF,$CD,$3F,$02,$7E,$28,$C4,$79,$6B,$9B,$EE,$3E,$5C,$7D,$C8,$CC
	.dc.b	$7A,$4F,$8A,$E2,$FC,$F1,$E1,$3F,$A0,$38,$79,$DF,$D1,$7F,$4E,$77
	.dc.b	$BF,$95,$3B,$6D,$F9,$A7,$E3,$DF,$C2,$7C,$33,$EE,$3A,$3F,$AA,$3F
	.dc.b	$00,$78,$CF,$E9,$B5,$E3,$BF,$8F,$1C,$8F,$3D,$F7,$70,$70,$71,$5E
	.dc.b	$43,$F9,$3C,$7E,$C6,$FE,$14,$EA,$3D,$A7,$C7,$9E,$13,$FB,$43,$D4
	.dc.b	$E1,$FC,$14,$78,$2F,$FA,$D4,$6A,$39,$C7,$A5,$FF,$35,$DF,$FB,$0C
	.dc.b	$FE,$15,$1F,$BD,$8F,$82,$73,$BF,$E0,$95,$DA,$FC,$0F,$F8,$0C,$FC
	.dc.b	$DC,$F6,$9F,$53,$F2,$CF,$D7,$4F,$07,$FA,$FC,$F3,$78,$47,$D2,$AF
	.dc.b	$14,$70,$33,$3F,$05,$FF,$1A,$1F,$56,$3D,$EE,$0F,$44,$FF,$0E,$75
	.dc.b	$5E,$AA,$F4,$1E,$18,$F3,$1F,$8B,$3B,$B9,$AF,$1C,$7E,$76,$FF,$03
	.dc.b	$7E,$B8,$FF,$03,$1E,$03,$F8,$C9,$4F,$97,$1C,$1E,$B5,$F5,$5F,$09
	.dc.b	$F7,$CC,$9F,$19,$D5,$78,$A3,$47,$07,$F4,$0F,$03,$FB,$5C,$7A,$6F
	.dc.b	$F7,$17,$F3,$5B,$1F,$62,$79,$47,$E6,$6F,$CD,$18,$AE,$15,$FC,$C1
	.dc.b	$AF,$F0,$FA,$E1,$3F,$8C,$3F,$DD,$55,$F8,$C8,$EF,$3E,$33,$91,$E1
	.dc.b	$BF,$61,$F9,$5A,$FF,$3D,$57,$E5,$5F,$EB,$53,$C0,$E4,$78,$6F,$C6
	.dc.b	$FE,$A9,$FD,$6D,$7F,$3B,$7F,$7D,$1D,$EF,$C6,$FC,$C6,$D7,$CF,$3F
	.dc.b	$15,$3F,$76,$BF,$D5,$0D,$9F,$E4,$75,$E3,$3F,$16,$73,$31,$5C,$8F
	.dc.b	$A9,$5C,$9B,$3F,$D5,$1F,$F0,$54,$FA,$CF,$31,$C9,$FC,$6D,$78,$5F
	.dc.b	$E7,$87,$53,$EB,$47,$03,$91,$F1,$8F,$88,$7E,$A5,$E6,$78,$8F,$E7
	.dc.b	$51,$F1,$A3,$83,$5F,$6C,$77,$AF,$AF,$5F,$BE,$6D,$7D,$23,$84,$FC
	.dc.b	$B3,$B9,$E6,$AE,$AF,$F2,$88,$E0,$D8,$F3,$4F,$C7,$8E,$11,$F3,$4F
	.dc.b	$69,$4F,$69,$FE,$78,$8E,$07,$91,$F8,$86,$8E,$13,$E6,$9F,$D3,$C6
	.dc.b	$1F,$13,$E6,$7F,$9D,$47,$99,$B1,$E6,$AF,$C1,$BE,$48,$F8,$D1,$F8
	.dc.b	$C7,$E0,$D7,$8E,$7C,$D3,$C1,$F2,$4F,$07,$F0,$13,$EA,$9E,$4F,$F8
	.dc.b	$79,$FE,$D9,$5E,$98,$D4,$F0,$9E,$93,$97,$F1,$D5,$FD,$3E,$30,$FE
	.dc.b	$00,$F7,$CF,$C1,$7D,$AF,$99,$D1,$87,$55,$FC,$C2,$3D,$B5,$E6,$3B
	.dc.b	$C7,$EC,$63,$A9,$ED,$9F,$8C,$F0,$7D,$0F,$E7,$0C,$3A,$8F,$6B,$67
	.dc.b	$DB,$3D,$0F,$B2,$75,$8F,$96,$F2,$6C,$F2,$6C,$F9,$8F,$97,$3C,$E7
	.dc.b	$55,$F1,$A3,$FA,$4B,$FD,$2D,$FF,$09,$D7,$E3,$23,$51,$FC,$FC,$FE
	.dc.b	$B0,$7E,$60,$FD,$14,$F4,$7B,$57,$A9,$F9,$87,$A6,$79,$DF,$F3,$2A
	.dc.b	$FC,$45,$FD,$0B,$FE,$B5,$7B,$39,$9F,$26,$7B,$3D,$0F,$D8,$81,$30
	.dc.b	$5E,$62,$35,$87,$FB,$44,$AE,$23,$93,$88,$D9,$DD,$ED,$18,$8F,$F6
	.dc.b	$79,$E3,$78,$BD,$E3,$4F,$73,$07,$8E,$36,$F9,$8E,$91,$D9,$A6,$E3
	.dc.b	$DF,$1E,$58,$F2,$BB,$69,$B8,$F8,$6F,$80,$DB,$F1,$DE,$A7,$C7,$7C
	.dc.b	$6F,$B0,$F5,$3B,$7A,$9B,$7C,$6F,$53,$E5,$BB,$7C,$C7,$C5,$76,$FB
	.dc.b	$CF,$1B,$F6,$9F,$19,$B7,$D2,$EC,$EA,$ED,$F1,$9B,$8E,$EF,$A2,$34
	.dc.b	$E1,$F6,$BD,$0E,$6F,$D6,$8F,$EA,$91,$F8,$DF,$DF,$7F,$15,$4B,$FC
	.dc.b	$0C,$61,$3F,$86,$8D,$3F,$CB,$5F,$E1,$87,$19,$FC,$A4,$71,$BF,$CC
	.dc.b	$3C,$CE,$55,$D2,$3F,$69,$3F,$A4,$3C,$87,$E8,$6B,$C9,$1C,$A3,$13
	.dc.b	$E4,$7F,$3F,$3E,$49,$E7,$1F,$AB,$8F,$23,$D0,$3D,$23,$F5,$B5,$E5
	.dc.b	$9E,$75,$E5,$9F,$E9,$95,$EE,$8F,$D7,$57,$B6,$78,$D7,$92,$3D,$A7
	.dc.b	$C5,$9F,$E0,$31,$FE,$50,$8E,$13,$E9,$8F,$1C,$D7,$E3,$47,$AA,$BC
	.dc.b	$11,$F3,$0F,$83,$1F,$0D,$E3,$5E,$08,$E1,$5E,$0A,$F8,$35,$DD,$F8
	.dc.b	$13,$67,$C1,$3E,$98,$E3,$7F,$CA,$6F,$CE,$7C,$59,$FE,$1D,$7F,$3E
	.dc.b	$BE,$2D,$F7,$3E,$2C,$F9,$A3,$E2,$BE,$63,$E6,$D7,$C5,$BE,$D1,$B9
	.dc.b	$F2,$4F,$AD,$EB,$3A,$9F,$8F,$1F,$06,$FF,$27,$1D,$23,$D2,$FF,$9F
	.dc.b	$2B,$D2,$FD,$53,$51,$D1,$D1,$E1,$8F,$73,$F4,$A3,$4F,$FA,$BA,$39
	.dc.b	$BE,$B3,$E1,$DF,$E1,$E7,$FA,$05,$F9,$EF,$E7,$DF,$4B,$F9,$D5,$D1
	.dc.b	$FD,$75,$7F,$1D,$AF,$AF,$5D,$9F,$6C,$F6,$8F,$90,$FA,$0F,$87,$5F
	.dc.b	$3A,$77,$1E,$98,$F1,$BB,$8F,$1B,$F0,$CE,$91,$F2,$1F,$71,$F0,$CD
	.dc.b	$1C,$CF,$EB,$47,$EB,$6F,$EF,$BE,$28,$FB,$2F,$CE,$7C,$2F,$43,$55
	.dc.b	$F0,$DF,$CC,$C7,$85,$F8,$6F,$F3,$3A,$F8,$71,$E3,$3E,$81,$E3,$3F
	.dc.b	$C1,$17,$FD,$F2,$BE,$45,$FE,$72,$7A,$DF,$CF,$7F,$84,$D7,$CF,$9F
	.dc.b	$4D,$FC,$49,$EA,$7D,$6A,$F4,$BE,$47,$FD,$F3,$7D,$8F,$CB,$3A,$8E
	.dc.b	$AF,$A6,$BA,$9F,$CB,$23,$E5,$9D,$4F,$97,$5F,$BE,$7F,$8B,$3E,$33
	.dc.b	$ED,$BB,$3D,$E7,$F2,$73,$F7,$53,$FA,$39,$F8,$13,$FB,$58,$F2,$4F
	.dc.b	$F0,$EB,$F8,$0F,$F0,$B3,$E9,$9E,$43,$E3,$C7,$09,$FE,$89,$3C,$27
	.dc.b	$CB,$5E,$B7,$E3,$9C,$0F,$5B,$FC,$0D,$ED,$3E,$B8,$F2,$9F,$A7,$8D
	.dc.b	$CF,$E9,$E3,$A4,$FC,$09,$E8,$7C,$03,$F2,$07,$48,$D9,$D2,$FF,$4B
	.dc.b	$1F,$8F,$9F,$04,$7D,$68,$EE,$7E,$8E,$78,$C6,$63,$B4,$FD,$37,$E6
	.dc.b	$DF,$EB,$6B,$B3,$F9,$78,$FB,$EF,$E1,$1F,$CB,$4F,$A4,$FA,$4F,$80
	.dc.b	$F2,$4F,$81,$FD,$31,$93,$F1,$D3,$FD,$C5,$F7,$3F,$76,$BC,$D5,$FE
	.dc.b	$48,$3C,$CF,$E4,$E7,$99,$EA,$AF,$A4,$FD,$8A,$E4,$EA,$B9,$4E,$A3
	.dc.b	$DF,$3F,$BA,$7F,$82,$BF,$A9,$7F,$14,$FE,$75,$F2,$BF,$16,$BC,$B3
	.dc.b	$FE,$67,$AF,$2D,$7C,$48,$F9,$35,$CE,$7F,$1A,$F3,$9F,$33,$CD,$F3
	.dc.b	$4F,$33,$CD,$5E,$F8,$F2,$CF,$E5,$E7,$FC,$CA,$F3,$8F,$5C,$FF,$0F
	.dc.b	$AF,$A7,$1F,$16,$FF,$21,$5F,$12,$39,$9F,$4C,$FB,$AF,$D3,$7C,$AF
	.dc.b	$C5,$7F,$89,$4F,$CA,$9E,$73,$E9,$9E,$67,$73,$F2,$0F,$77,$8D,$FB
	.dc.b	$67,$8C,$F5,$9F,$E0,$15,$D5,$E3,$1D,$5E,$27,$FB,$02,$3E,$7B,$FB
	.dc.b	$C3,$F6,$AF,$C0,$39,$3F,$92,$8F,$13,$DA,$BC,$4F,$68,$F1,$3F,$B2
	.dc.b	$9E,$2F,$E5,$8F,$F1,$99,$F8,$28,$DC,$77,$9E,$73,$DE,$7D,$0F,$73
	.dc.b	$D1,$1F,$71,$FC,$35,$FE,$29,$CD,$FC,$48,$FB,$53,$FC,$26,$7E,$61
	.dc.b	$D6,$BC,$31,$9B,$F3,$CF,$DA,$3C,$E7,$DA,$AE,$67,$F0,$87,$D8,$7F
	.dc.b	$8D,$CF,$19,$F9,$23,$C6,$7C,$57,$8D,$7F,$03,$8F,$D2,$BF,$C0,$DF
	.dc.b	$A5,$3F,$C0,$EB,$C8,$7C,$48,$EF,$1F,$D7,$9D,$C7,$E9,$A7,$F3,$97
	.dc.b	$FC,$C6,$7F,$36,$7F,$37,$AF,$9C,$F9,$0F,$85,$7E,$4A,$FD,$75,$FE
	.dc.b	$91,$F2,$57,$F2,$4A,$F6,$1F,$EE,$83,$D8,$75,$9F,$3B,$D4,$F3,$BF
	.dc.b	$D7,$27,$F1,$31,$F8,$58,$F6,$3F,$3C,$F9,$47,$B2,$BE,$E3,$DC,$F3
	.dc.b	$9D,$DF,$A4,$F1,$AF,$88,$79,$DF,$D5,$BF,$71,$F9,$F5,$FB,$57,$E2
	.dc.b	$9F,$A0,$AF,$4B,$DC,$F4,$9F,$97,$AF,$8D,$3F,$38,$FE,$10,$FF,$0D
	.dc.b	$8F,$D6,$9C,$4F,$24,$FC,$19,$FC,$DC,$7C,$13,$E3,$1A,$3B,$1E,$D7
	.dc.b	$F0,$D5,$FC,$81,$FA,$73,$FD,$0E,$7D,$AF,$CA,$7D,$A7,$E6,$0F,$6C
	.dc.b	$FE,$2E,$BE,$3C,$68,$F9,$F1,$F1,$DF,$49,$FE,$CE,$7F,$D0,$37,$FE
	.dc.b	$27,$8F,$CE,$9E,$03,$CE,$7C,$99,$FC,$81,$F0,$DF,$DC,$1F,$0C,$F0
	.dc.b	$4F,$D6,$AF,$54,$78,$27,$83,$F1,$4F,$D8,$1F,$0E,$FF,$2B,$7D,$A3
	.dc.b	$C6,$71,$29,$E0,$3E,$0D,$7E,$95,$FE,$6C,$78,$9F,$A9,$3F,$C0,$AB
	.dc.b	$A4,$FE,$44,$FC,$F9,$C9,$FC,$9B,$F1,$65,$30,$C5,$7F,$11,$9F,$44
	.dc.b	$7E,$3E,$BD,$B7,$FB,$29,$F6,$9F,$26,$7E,$FD,$FE,$C5,$E1,$1E,$C3
	.dc.b	$E2,$1E,$33,$CB,$5D,$27,$F3,$31,$E5,$9F,$25,$FD,$98,$F6,$1F,$64
	.dc.b	$F0,$DF,$E1,$67,$F3,$71,$DE,$3C,$CF,$38,$F7,$CF,$39,$FC,$44,$FC
	.dc.b	$FB,$F1,$1F,$2A,$33,$1E,$89,$FF,$6D,$9C,$6F,$F0,$31,$C4,$F0,$3E
	.dc.b	$AA,$E7,$7F,$12,$BA,$4F,$0A,$F8,$75,$AA,$FD,$29,$FC,$3A,$FD,$37
	.dc.b	$E0,$AF,$F3,$15,$6E,$BE,$6C,$7C,$43,$F1,$45,$3F,$15,$3C,$5F,$F2
	.dc.b	$59,$F2,$2F,$F1,$8F,$EA,$9F,$25,$FC,$8B,$ED,$3F,$90,$8F,$DD,$5F
	.dc.b	$81,$F8,$A6,$E3,$E0,$CF,$BA,$7D,$55,$F7,$2F,$C5,$7F,$DE,$0E,$31
	.dc.b	$F1,$EF,$8D,$7D,$A7,$D4,$7D,$13,$FD,$D1,$5F,$46,$3A,$DF,$77,$F8
	.dc.b	$AD,$F8,$EB,$8B,$F2,$9E,$27,$CA,$3B,$C7,$D0,$38,$DF,$8E,$38,$BE
	.dc.b	$97,$89,$F2,$8F,$CD,$1F,$93,$29,$F4,$A7,$DB,$5F,$C6,$1F,$93,$1F
	.dc.b	$C7,$63,$E6,$BF,$46,$BD,$67,$C1,$9E,$2F,$E2,$4C,$57,$D6,$3C,$D1
	.dc.b	$A8,$F5,$1F,$C0,$CE,$06,$2B,$FA,$0D,$78,$4F,$83,$1F,$6E,$BA,$4F
	.dc.b	$9D,$FD,$D4,$EC,$F8,$71,$F3,$E7,$B1,$D6,$3F,$83,$9F,$87,$9E,$2F
	.dc.b	$B6,$3E,$6C,$F1,$3C,$71,$98,$E6,$7F,$55,$3A,$3F,$C3,$5B,$5F,$BF
	.dc.b	$6D,$7E,$6A,$3F,$B3,$1F,$3A,$7A,$57,$68,$FE,$71,$5F,$38,$E3,$1F
	.dc.b	$C9,$2F,$F2,$B7,$FE,$35,$BF,$7D,$7D,$03,$81,$F1,$CA,$FB,$E7,$D8
	.dc.b	$FC,$23,$C2,$5B,$F0,$CF,$C2,$3E,$A9,$CC,$C1,$E4,$8F,$E4,$A7,$D6
	.dc.b	$3B,$3E,$33,$67,$13,$E4,$BF,$5E,$3F,$09,$1F,$2C,$D5,$FC,$3B,$FE
	.dc.b	$92,$DA,$FE,$C7,$5F,$36,$B8,$47,$3A,$FC,$2C,$78,$4E,$F1,$EE,$3E
	.dc.b	$B3,$C1,$C4,$F3,$79,$47,$C8,$73,$3F,$64,$FF,$01,$4F,$D9,$3E,$3C
	.dc.b	$F6,$6D,$79,$EB,$F8,$D3,$F6,$5F,$DA,$C7,$F0,$DA,$F3,$D7,$CB,$9F
	.dc.b	$DB,$57,$EC,$2B,$F0,$F5,$F9,$67,$AC,$FD,$78,$F8,$AD,$3E,$2B,$E7
	.dc.b	$7F,$C4,$27,$F7,$33,$33,$FD,$B8,$ED,$1C,$A3,$E2,$9F,$6A,$BC,$8F
	.dc.b	$F0,$78,$F3,$C7,$D6,$8F,$01,$FE,$11,$AF,$89,$3C,$DF,$99,$1F,$2A
	.dc.b	$BC,$0F,$E1,$CF,$F4,$39,$FE,$42,$3E,$39,$F8,$C3,$F3,$B5,$E2,$7F
	.dc.b	$A3,$9F,$2A,$BF,$45,$5F,$8A,$8F,$B3,$3F,$90,$6B,$F5,$67,$EC,$9F
	.dc.b	$2C,$F3,$1F,$1E,$02,$AB,$27,$FC,$25,$18,$8D,$B5,$C3,$FE,$17,$8C
	.dc.b	$47,$28,$D4,$72,$8D,$9A,$3F,$D5,$47,$38,$DC,$7F,$88,$9C,$46,$DE
	.dc.b	$CD,$3C,$91,$D5,$F0,$1B,$7F,$C5,$4E,$CF,$63,$D5,$F1,$9E,$88,$EA
	.dc.b	$D3,$AC,$6D,$EA,$6D,$EA,$EC,$EA,$FC,$B7,$A9,$ED,$76,$FB,$0D,$C5
	.dc.b	$75,$1F,$E4,$27,$D0,$7D,$67,$DC,$FD,$23,$DC,$ED,$F1,$1B,$7D,$0E
	.dc.b	$CF,$43,$B7,$C0,$7D,$98,$AF,$D9,$8F,$3B,$EA,$34,$53,$4F,$37,$F1
	.dc.b	$B1,$FE,$5C,$FD,$FB,$F1,$76,$BF,$C3,$46,$13,$F8,$88,$D3,$FC,$E2
	.dc.b	$BF,$6A,$23,$93,$F9,$68,$E5,$7F,$E6,$77,$A1,$CE,$BA,$C7,$F2,$5B
	.dc.b	$FE,$2D,$5E,$03,$F4,$37,$E5,$9E,$51,$8A,$D3,$FA,$19,$F2,$CF,$38
	.dc.b	$FE,$06,$6A,$39,$47,$F9,$EA,$FC,$73,$FA,$79,$E3,$3F,$BD,$8F,$F3
	.dc.b	$8C,$FA,$A3,$84,$78,$63,$D4,$7C,$39,$FE,$1A,$7C,$28,$FD,$CC,$71
	.dc.b	$9F,$15,$7A,$27,$E1,$C7,$9D,$F9,$75,$C5,$DC,$7C,$07,$E7,$4F,$17
	.dc.b	$C5,$1C,$67,$C3,$3F,$C0,$EB,$C0,$7B,$AB,$E3,$47,$D0,$AF,$75,$7F
	.dc.b	$A0,$2B,$75,$ED,$3E,$25,$6E,$BF,$97,$47,$9A,$3B,$3F,$4E,$3C,$D5
	.dc.b	$F8,$68,$E3,$5F,$16,$BF,$8E,$47,$B2,$BE,$5B,$EC,$9F,$97,$1E,$B9
	.dc.b	$DC,$79,$27,$47,$49,$D5,$FE,$6E,$3C,$51,$D2,$FC,$51,$F0,$2B,$C5
	.dc.b	$5B,$8F,$97,$5E,$8B,$F9,$91,$E3,$8E,$AF,$FB,$12,$BC,$57,$D2,$3E
	.dc.b	$CC,$FC,$AA,$FF,$59,$D6,$9E,$75,$F8,$C9,$F9,$2F,$D6,$AF,$D3,$1E
	.dc.b	$88,$FC,$2D,$7C,$27,$A5,$FF,$20,$AF,$C2,$4F,$53,$53,$D2,$FE,$3D
	.dc.b	$F3,$9F,$C0,$3E,$69,$D5,$FC,$1B,$FD,$1C,$78,$27,$E6,$3E,$03,$F9
	.dc.b	$4C,$78,$23,$A1,$F8,$53,$E6,$47,$C6,$7F,$CF,$95,$E5,$79,$C7,$81
	.dc.b	$F9,$47,$23,$ED,$9C,$9F,$01,$FC,$9E,$7C,$11,$F5,$A3,$EE,$5F,$CC
	.dc.b	$3F,$29,$1F,$DB,$2F,$F8,$4D,$F8,$63,$A5,$78,$5F,$29,$E1,$BF,$B4
	.dc.b	$FE,$AA,$3C,$27,$F2,$17,$EA,$9E,$17,$E3,$3D,$0F,$01,$FC,$3A,$7C
	.dc.b	$0F,$CF,$32,$7E,$9C,$E7,$3F,$42,$3E,$AC,$78,$67,$CC,$FF,$31,$AF
	.dc.b	$C0,$5F,$EB,$A3,$C5,$5C,$1F,$CB,$57,$A1,$FE,$48,$76,$9F,$E7,$6F
	.dc.b	$D7,$3C,$87,$F8,$84,$FA,$67,$43,$47,$8C,$F4,$DF,$E5,$DF,$E3,$67
	.dc.b	$C9,$8F,$A5,$5F,$AC,$8E,$73,$ED,$79,$9E,$6A,$F6,$BF,$CE,$6F,$DA
	.dc.b	$F9,$E3,$F6,$55,$F9,$C8,$F0,$DF,$E7,$1F,$31,$F8,$83,$D7,$7F,$E5
	.dc.b	$27,$E7,$1E,$63,$6F,$DE,$7C,$47,$4B,$FD,$14,$FC,$68,$E8,$7C,$13
	.dc.b	$E3,$1F,$88,$39,$BF,$02,$FF,$41,$1F,$5D,$FD,$3C,$7D,$7A,$F8,$53
	.dc.b	$D0,$F0,$9D,$0E,$13,$F9,$F9,$E0,$78,$E7,$D0,$7F,$09,$9F,$41,$F3
	.dc.b	$A3,$47,$8E,$7F,$05,$1F,$D1,$8F,$98,$FE,$A2,$FE,$64,$78,$4F,$A5
	.dc.b	$1E,$87,$6F,$E7,$9F,$9B,$3F,$E0,$87,$C1,$1F,$59,$FD,$E9,$E4,$3E
	.dc.b	$35,$7C,$03,$D3,$3F,$B6,$8F,$01,$FD,$B0,$FC,$55,$79,$27,$ED,$1F
	.dc.b	$8F,$8F,$E8,$D5,$E3,$7E,$A4,$6A,$BF,$08,$F7,$7E,$14,$FE,$3A,$7B
	.dc.b	$1F,$62,$BB,$4F,$AE,$3C,$EF,$CA,$8C,$C7,$FA,$50,$F7,$CF,$C9,$7D
	.dc.b	$E7,$AA,$BF,$B4,$3E,$A3,$2F,$96,$FF,$1B,$3F,$1E,$7F,$84,$1C,$6F
	.dc.b	$FA,$3B,$C4,$CD,$71,$7C,$33,$F0,$5F,$C4,$C6,$5F,$4C,$75,$7E,$B9
	.dc.b	$E9,$32,$FA,$5E,$B3,$F0,$5F,$C8,$C7,$F1,$F7,$F5,$13,$F2,$EF,$E1
	.dc.b	$C7,$FA,$88,$E7,$3E,$39,$FC,$A4,$7A,$EF,$F7,$6F,$2B,$F9,$8F,$C4
	.dc.b	$7C,$0F,$E4,$2B,$F4,$0F,$E4,$2F,$F8,$E5,$F4,$3D,$D1,$E6,$3F,$8E
	.dc.b	$9D,$67,$E0,$3F,$84,$AE,$AF,$E0,$27,$D4,$F8,$4F,$88,$73,$3F,$4C
	.dc.b	$73,$9F,$8C,$FC,$A3,$E0,$1B,$8F,$EA,$67,$47,$E5,$9D,$0F,$64,$7C
	.dc.b	$B7,$F3,$35,$EC,$AF,$D1,$57,$B2,$76,$7B,$0D,$9F,$CA,$1F,$85,$1E
	.dc.b	$C7,$F1,$F1,$F9,$B3,$D8,$79,$8E,$67,$B0,$E6,$7F,$11,$79,$BE,$73
	.dc.b	$E3,$47,$D6,$7D,$67,$33,$35,$F3,$9F,$C8,$C7,$8C,$FC,$CB,$E3,$38
	.dc.b	$9F,$EA,$D3,$FD,$CF,$1F,$0E,$BF,$B2,$C7,$D7,$3C,$55,$EC,$3E,$63
	.dc.b	$F4,$4E,$4F,$FB,$AD,$F6,$47,$27,$C6,$72,$AF,$1C,$F2,$8E,$C7,$AD
	.dc.b	$ED,$1E,$F8,$ED,$3C,$E7,$E2,$C7,$E0,$EB,$F3,$D5,$F6,$0F,$81,$7F
	.dc.b	$9E,$9F,$9B,$5F,$C7,$AF,$07,$3A,$E1,$3F,$9D,$3E,$E3,$87,$C8,$FA
	.dc.b	$0F,$23,$E8,$3A,$BF,$C7,$1F,$C3,$47,$53,$FB,$83,$FB,$47,$F1,$C7
	.dc.b	$89,$FE,$18,$7B,$8A,$75,$AF,$F4,$A1,$FE,$28,$72,$7C,$C3,$E1,$47
	.dc.b	$4B,$FE,$8D,$18,$3D,$71,$E2,$7F,$4D,$5F,$D6,$A7,$F1,$17,$F3,$27
	.dc.b	$D7,$5F,$C3,$1A,$7E,$68,$FB,$F5,$FE,$DA,$7F,$BD,$9A,$3F,$1B,$1F
	.dc.b	$06,$BC,$71,$93,$F1,$17,$EC,$7F,$C6,$87,$B2,$FC,$55,$C6,$BF,$88
	.dc.b	$CF,$AC,$FD,$35,$FD,$57,$C7,$02,$5F,$C7,$8F,$D0,$1F,$34,$A7,$D9
	.dc.b	$AE,$D3,$FD,$4E,$FF,$22,$FC,$4B,$ED,$3A,$AF,$BA,$79,$E3,$C5,$7E
	.dc.b	$B8,$F8,$73,$B7,$F4,$66,$DF,$C0,$3E,$23,$ED,$1E,$AA,$F4,$BF,$10
	.dc.b	$D5,$FC,$47,$94,$7C,$C9,$ED,$3F,$C5,$E7,$F8,$1C,$FF,$26,$7D,$C7
	.dc.b	$07,$E5,$BF,$4C,$FC,$71,$FE,$9D,$AF,$91,$3F,$B5,$8F,$2D,$76,$BF
	.dc.b	$C4,$5F,$59,$F4,$DF,$C1,$BF,$F0,$25,$EE,$7F,$58,$7D,$87,$E0,$C7
	.dc.b	$53,$F2,$13,$C6,$FF,$C0,$E7,$EB,$AB,$C2,$7A,$6F,$F8,$B3,$4F,$F0
	.dc.b	$E1,$F9,$C3,$BD,$7E,$FE,$3E,$34,$6C,$FA,$67,$CC,$9F,$B2,$FC,$13
	.dc.b	$E8,$4F,$96,$3F,$03,$1C,$63,$FA,$6C,$79,$5F,$4C,$71,$3F,$02,$6E
	.dc.b	$3E,$3C,$FF,$16,$BF,$83,$1C,$4F,$0C,$FF,$32,$3F,$1A,$79,$63,$73
	.dc.b	$F9,$36,$BF,$46,$BC,$33,$C4,$F6,$BE,$99,$DC,$F8,$4F,$87,$3D,$5F
	.dc.b	$F3,$2B,$FD,$52,$7D,$B1,$DC,$DB,$CD,$B5,$E9,$AF,$E1,$95,$C6,$3D
	.dc.b	$F1,$FC,$4E,$7F,$1E,$FE,$0E,$7F,$75,$5F,$56,$7F,$CC,$6F,$81,$F6
	.dc.b	$57,$D0,$9F,$34,$72,$7E,$61,$FD,$2E,$FD,$B7,$DE,$7F,$0B,$3C,$1B
	.dc.b	$7E,$97,$BD,$F9,$A7,$F9,$2C,$7B,$0F,$80,$77,$79,$1F,$8E,$3D,$C7
	.dc.b	$E3,$A3,$FA,$03,$F4,$1F,$89,$1C,$CF,$5D,$7C,$69,$FD,$AC,$7D,$33
	.dc.b	$B5,$79,$5F,$3D,$72,$7C,$F3,$DE,$BF,$9A,$1F,$2D,$FD,$D1,$FB,$23
	.dc.b	$F2,$55,$FB,$27,$E3,$9F,$42,$BC,$F1,$F3,$EB,$C4,$7F,$21,$8F,$EB
	.dc.b	$CF,$C2,$3C,$65,$9F,$CA,$CF,$9D,$F8,$87,$B2,$BE,$93,$AA,$F5,$57
	.dc.b	$23,$E1,$9F,$90,$8E,$07,$D2,$7E,$1C,$70,$8E,$51,$E1,$7F,$CB,$4F
	.dc.b	$CD,$3D,$2E,$9E,$A7,$23,$85,$7F,$25,$3F,$9E,$BA,$3C,$87,$28,$FB
	.dc.b	$4F,$09,$FF,$05,$BD,$EB,$DB,$3A,$8E,$4F,$85,$E5,$5D,$EB,$E5,$C6
	.dc.b	$A7,$83,$A8,$FC,$94,$7F,$25,$8F,$89,$3E,$67,$51,$F8,$13,$FB,$95
	.dc.b	$7E,$06,$FF,$94,$1F,$E9,$A9,$ED,$5F,$76,$7C,$ED,$AF,$9D,$3F,$7E
	.dc.b	$BF,$53,$18,$3D,$55,$C2,$FC,$B7,$FB,$7B,$E0,$7C,$3A,$E0,$7B,$2B
	.dc.b	$83,$E7,$3F,$C9,$F7,$E6,$7F,$D4,$F1,$A7,$F2,$8F,$E7,$1F,$CB,$9F
	.dc.b	$D8,$E7,$FC,$1F,$3E,$D7,$E1,$C0,$4D,$64,$7F,$80,$CC,$B9,$29,$4E
	.dc.b	$51,$B8,$C4,$73,$8D,$47,$37,$81,$A8,$E3,$1D,$DC,$47,$08,$FF,$0A
	.dc.b	$3C,$23,$31,$B7,$BC,$6E,$3C,$B1,$E4,$36,$FF,$87,$5E,$A7,$AD,$F2
	.dc.b	$3E,$D8,$F2,$3F,$EF,$B3,$C0,$7C,$48,$F1,$BB,$3A,$BC,$23,$C4,$70
	.dc.b	$8F,$7C,$79,$DC,$1F,$1E,$2B,$F1,$E3,$4F,$53,$C0,$FB,$18,$6D,$F6
	.dc.b	$3F,$08,$F6,$3E,$67,$E9,$1E,$63,$E0,$9D,$23,$C2,$FD,$28,$E6,$F9
	.dc.b	$9F,$8A,$F2,$3E,$7B,$E6,$7F,$2C,$7F,$A6,$FF,$7F,$FC,$55,$B0,$FC
	.dc.b	$24,$61,$3F,$86,$8D,$3F,$AA,$9F,$DD,$57,$F8,$BA,$38,$3F,$93,$79
	.dc.b	$1C,$6B,$9C,$7E,$AE,$78,$BF,$ED,$59,$E2,$FF,$97,$5E,$31,$F9,$9A
	.dc.b	$F0,$CF,$17,$F6,$D1,$FE,$D7,$9E,$33,$E8,$8F,$D4,$4F,$A1,$E5,$5E
	.dc.b	$29,$FE,$50,$7F,$0A,$7F,$51,$5C,$8F,$0C,$FE,$AE,$BC,$35,$C0,$FD
	.dc.b	$AC,$77,$11,$F2,$E3,$E1,$47,$8A,$B4,$FC,$48,$F5,$47,$F5,$63,$C3
	.dc.b	$1B,$8E,$2F,$48,$E2,$F3,$AF,$A5,$5F,$0A,$3F,$DD,$D1,$E9,$74,$7C
	.dc.b	$07,$73,$DE,$3F,$85,$46,$DE,$CF,$D7,$37,$7F,$EA,$57,$55,$F2,$E7
	.dc.b	$E2,$57,$CB,$BF,$E9,$B7,$FC,$7E,$7E,$25,$7C,$39,$FB,$91,$E4,$38
	.dc.b	$DF,$67,$FD,$60,$F6,$BF,$7C,$79,$27,$E3,$DF,$8E,$3E,$34,$74,$7D
	.dc.b	$27,$BD,$DB,$D0,$DB,$E4,$9E,$B7,$F7,$23,$D3,$3F,$CF,$A3,$F4,$8F
	.dc.b	$C6,$38,$DF,$E1,$1F,$B5,$3F,$83,$BF,$81,$3E,$93,$E5,$47,$8A,$7E
	.dc.b	$3C,$FF,$BD,$E7,$51,$E7,$9E,$4F,$86,$FF,$06,$FE,$06,$3F,$07,$7F
	.dc.b	$46,$3C,$31,$E8,$7F,$56,$7E,$8C,$F3,$9F,$3C,$D9,$FD,$80,$EC,$7B
	.dc.b	$4E,$CF,$CB,$8E,$D1,$F8,$F8,$F5,$3F,$C2,$4E,$A7,$D0,$3E,$DB,$A7
	.dc.b	$F4,$17,$EF,$77,$3D,$1F,$B9,$3F,$01,$F4,$9F,$9A,$37,$5F,$58,$FC
	.dc.b	$73,$E9,$37,$3F,$59,$FE,$CD,$7F,$31,$F1,$D7,$59,$FD,$31,$B8,$FC
	.dc.b	$14,$7C,$CA,$ED,$5F,$93,$9F,$AF,$3F,$EF,$C8,$E9,$1F,$28,$FF,$01
	.dc.b	$1F,$28,$F0,$9E,$87,$F0,$31,$E1,$9F,$D8,$4E,$8F,$9F,$1F,$93,$9D
	.dc.b	$BF,$6E,$BC,$F5,$F8,$27,$F5,$51,$BA,$FC,$43,$B9,$EB,$1E,$C8,$FB
	.dc.b	$51,$EC,$7E,$A4,$7F,$BC,$AB,$D6,$FD,$D8,$F3,$5F,$EB,$63,$CB,$3F
	.dc.b	$9F,$75,$3D,$E7,$57,$E4,$AD,$57,$DC,$AD,$3F,$B1,$39,$9F,$93,$39
	.dc.b	$9F,$6E,$7D,$EF,$E4,$DE,$8E,$CE,$86,$EF,$DF,$5E,$A8,$C9,$F5,$0F
	.dc.b	$C2,$C7,$C9,$3F,$AE,$BC,$E7,$F8,$03,$CE,$FD,$73,$A8,$FC,$E1,$F7
	.dc.b	$E7,$FC,$98,$F4,$3C,$D1,$B8,$FC,$E1,$D2,$3C,$4F,$B1,$FC,$03,$EC
	.dc.b	$7A,$9F,$08,$DC,$7B,$1F,$C5,$1F,$78,$F3,$1F,$2E,$39,$1E,$08,$FD
	.dc.b	$2C,$72,$9F,$5D,$FE,$76,$FF,$1B,$3C,$8F,$CE,$C7,$F7,$97,$B9,$FA
	.dc.b	$87,$EC,$BA,$7F,$BB,$3A,$3E,$25,$7F,$9A,$0F,$43,$F7,$8F,$69,$F5
	.dc.b	$2F,$E4,$9F,$0D,$FC,$B4,$FE,$6D,$F0,$1E,$53,$F8,$71,$CE,$3D,$D5
	.dc.b	$FB,$58,$CC,$7C,$2A,$FC,$F5,$7E,$6A,$7E,$34,$7D,$53,$31,$EC,$8E
	.dc.b	$31,$9A,$F8,$B1,$F2,$4E,$33,$F2,$5F,$A4,$64,$F8,$87,$F2,$F3,$CD
	.dc.b	$3D,$9F,$51,$FC,$8A,$3F,$85,$3F,$24,$FC,$05,$7C,$49,$F9,$2F,$AE
	.dc.b	$BC,$6E,$6F,$ED,$DF,$EF,$27,$35,$F4,$67,$33,$F2,$EF,$EA,$9F,$E4
	.dc.b	$03,$C4,$FD,$F7,$CF,$5C,$9D,$1F,$C1,$4F,$31,$F6,$6B,$B3,$F1,$27
	.dc.b	$C4,$FC,$4B,$F3,$39,$8F,$13,$EE,$9E,$2F,$C8,$3F,$9F,$1E,$23,$93
	.dc.b	$A7,$EA,$C7,$D1,$7F,$0A,$7D,$13,$2F,$89,$CC,$F8,$9E,$CF,$F3,$FA
	.dc.b	$FE,$0C,$FE,$D4,$D4,$72,$9F,$3B,$EF,$3E,$1C,$7F,$99,$6F,$FA,$D4
	.dc.b	$79,$CE,$F7,$E7,$7E,$6D,$7E,$99,$D1,$F4,$8F,$F6,$09,$F1,$6F,$F9
	.dc.b	$3D,$F8,$EB,$94,$FD,$13,$89,$F2,$4F,$1B,$FE,$C2,$AF,$C5,$C6,$8F
	.dc.b	$41,$D8,$F8,$31,$FC,$86,$3E,$54,$78,$CF,$F6,$2C,$FF,$4C,$BF,$F0
	.dc.b	$CC,$7E,$02,$3C,$8F,$82,$3E,$0C,$7E,$56,$3E,$B9,$99,$F1,$C6,$63
	.dc.b	$E2,$99,$8F,$C8,$46,$63,$C5,$3F,$E7,$83,$C2,$FA,$E3,$C2,$FA,$CF
	.dc.b	$AD,$7F,$22,$FF,$8F,$CF,$DA,$9F,$A2,$7E,$CD,$EC,$FC,$93,$91,$A3
	.dc.b	$F2,$2E,$9F,$92,$7C,$48,$FF,$60,$3F,$DF,$4F,$E2,$F3,$E1,$3F,$89
	.dc.b	$3C,$E7,$F1,$CF,$C9,$9F,$29,$F7,$23,$F1,$57,$F1,$A3,$F8,$84,$FC
	.dc.b	$A3,$CA,$7E,$06,$7C,$87,$C2,$7C,$27,$F5,$98,$EA,$FF,$52,$9F,$BA
	.dc.b	$FB,$CF,$53,$FE,$C4,$9F,$E6,$15,$FA,$2A,$E3,$5F,$D5,$1F,$1B,$F6
	.dc.b	$0C,$57,$47,$E2,$DF,$D7,$7E,$25,$73,$3F,$0B,$5C,$A3,$F1,$D1,$9A
	.dc.b	$FF,$08,$DF,$C2,$7F,$07,$5B,$AF,$45,$6E,$72,$78,$6B,$E1,$5F,$86
	.dc.b	$BF,$B9,$1E,$69,$FE,$04,$F8,$0F,$A5,$5E,$48,$E5,$7F,$02,$7E,$2C
	.dc.b	$7F,$B0,$5F,$94,$F2,$7C,$67,$27,$EC,$1C,$A7,$F0,$31,$F5,$E7,$D4
	.dc.b	$FE,$2A,$3D,$53,$FC,$0E,$FE,$DD,$78,$6F,$D5,$1F,$52,$FB,$57,$C0
	.dc.b	$3C,$4F,$33,$2F,$F5,$97,$F2,$15,$FB,$D8,$DC,$FE,$1E,$FC,$35,$FA
	.dc.b	$3A,$F2,$CF,$CA,$8F,$F5,$1D,$FE,$C2,$3C,$B5,$F1,$EF,$9D,$FC,$78
	.dc.b	$E7,$7F,$CC,$4F,$09,$F2,$27,$F9,$4D,$78,$63,$CE,$78,$63,$E4,$1F
	.dc.b	$82,$AF,$E3,$4F,$BC,$E2,$F7,$8E,$27,$ED,$9F,$49,$F2,$5F,$49,$F9
	.dc.b	$23,$73,$FB,$27,$F8,$54,$73,$6D,$7D,$89,$FC,$95,$7D,$5A,$EB,$3F
	.dc.b	$06,$3C,$6D,$3F,$80,$3E,$9B,$C5,$FE,$CE,$6B,$FE,$20,$BF,$84,$F8
	.dc.b	$2B,$81,$F8,$A3,$E6,$CF,$09,$FF,$0E,$C6,$4F,$09,$C0,$F8,$D1,$E2
	.dc.b	$6C,$F8,$5E,$11,$97,$F0,$D1,$F9,$57,$E8,$9C,$1F,$35,$FC,$5A,$FC
	.dc.b	$0C,$F7,$3E,$21,$C0,$A7,$A0,$EF,$3E,$17,$FC,$60,$FF,$50,$3F,$09
	.dc.b	$5E,$37,$F0,$51,$C4,$F1,$4F,$C0,$78,$C7,$F3,$E3,$A3,$E0,$3E,$19
	.dc.b	$C2,$BF,$C5,$D1,$F9,$33,$DD,$1D,$4E,$15,$E1,$8F,$C5,$47,$AD,$FF
	.dc.b	$1A,$9C,$1E,$A7,$07,$C2,$7A,$1C,$BC,$2B,$F1,$2E,$67,$B4,$7E,$44
	.dc.b	$F9,$73,$E7,$8F,$F1,$74,$FF,$3E,$9E,$2F,$92,$7D,$0F,$33,$C2,$65
	.dc.b	$F0,$BE,$F8,$E2,$77,$AF,$CC,$CF,$D4,$9F,$BB,$5F,$8D,$9F,$8F,$5F
	.dc.b	$D9,$E7,$E0,$4F,$A2,$7F,$85,$57,$6A,$E8,$FC,$7A,$F0,$4F,$EE,$23
	.dc.b	$E3,$47,$67,$C1,$3F,$6D,$CB,$63,$E4,$4F,$17,$DC,$71,$9F,$8C,$FF
	.dc.b	$63,$3F,$AF,$99,$7F,$37,$3F,$02,$7C,$F5,$E9,$AF,$C7,$9F,$22,$7D
	.dc.b	$35,$F3,$23,$D3,$7F,$0E,$BE,$9C,$76,$AF,$33,$F3,$E3,$A3,$F5,$EF
	.dc.b	$D0,$E4,$EC,$F8,$9A,$78,$8F,$41,$F0,$2B,$C0,$FB,$67,$F9,$5D,$79
	.dc.b	$AF,$E8,$5F,$F8,$1C,$EF,$19,$9F,$1D,$79,$E7,$ED,$9E,$17,$B9,$EE
	.dc.b	$AF,$A7,$1A,$8F,$25,$7B,$8E,$8F,$A0,$F7,$5F,$99,$FF,$01,$1D,$EF
	.dc.b	$E3,$9E,$8A,$F3,$4E,$E7,$FB,$81,$FC,$7E,$FE,$DB,$CA,$3F,$AF,$CF
	.dc.b	$FA,$78,$FC,$BC,$EA,$78,$DF,$EA,$23,$EF,$CF,$C2,$3D,$D1,$E9,$69
	.dc.b	$ED,$9F,$F0,$14,$FC,$A7,$FC,$A0,$FE,$4A,$35,$0B,$30,$B2,$24,$66
	.dc.b	$36,$D7,$0E,$E3,$84,$64,$D9,$C5,$D4,$71,$35,$1D,$23,$93,$88,$ED
	.dc.b	$1C,$23,$B9,$97,$6F,$78,$DC,$7A,$23,$66,$5D,$BE,$13,$C8,$ED,$F8
	.dc.b	$51,$E4,$7E,$34,$70,$79,$9F,$19,$E0,$FB,$8E,$CF,$C2,$7C,$87,$67
	.dc.b	$E8,$3C,$8F,$92,$F5,$7C,$26,$E3,$CC,$F4,$8F,$24,$7B,$9E,$71,$C0
	.dc.b	$D1,$F2,$4E,$47,$73,$DA,$F7,$7D,$07,$AD,$E6,$E0,$F2,$BE,$B7,$C6
	.dc.b	$6E,$2B,$A8,$F2,$BF,$E1,$29,$FC,$47,$9F,$F8,$DB,$47,$E1,$E9,$39
	.dc.b	$8B,$5F,$8D,$AD,$CF,$EB,$63,$FC,$F8,$EE,$BF,$20,$F1,$3F,$21,$3C
	.dc.b	$23,$F8,$0D,$FF,$1E,$9E,$11,$F9,$D7,$85,$77,$BF,$1C,$F0,$8F,$E6
	.dc.b	$53,$C2,$3C,$B0,$BF,$2B,$C6,$7C,$CF,$FA,$A9,$E0,$76,$7F,$55,$1E
	.dc.b	$59,$FE,$01,$1E,$C3,$F6,$71,$F1,$A3,$DB,$18,$AF,$0B,$EF,$8F,$33
	.dc.b	$FC,$5A,$BB,$BB,$7E,$1C,$FB,$27,$BC,$F9,$4E,$8F,$67,$D9,$1D,$9E
	.dc.b	$51,$D9,$FF,$74,$4F,$67,$6F,$63,$E0,$D7,$FA,$DD,$D3,$F1,$A3,$E0
	.dc.b	$D7,$8E,$BF,$8A,$5F,$D5,$9F,$1C,$7D,$57,$D4,$7C,$D7,$C4,$FC,$73
	.dc.b	$CE,$FA,$6B,$B4,$7C,$58,$F7,$C7,$8E,$3A,$C7,$DA,$8D,$C7,$BA,$37
	.dc.b	$5F,$EA,$5B,$F4,$BF,$0E,$7D,$F3,$E2,$7E,$51,$E2,$7D,$0F,$C4,$3D
	.dc.b	$0F,$4B,$FD,$04,$7D,$B3,$F9,$85,$76,$3E,$54,$F5,$BF,$44,$F5,$8F
	.dc.b	$45,$F8,$E7,$F1,$51,$F2,$A7,$F2,$E7,$BA,$BF,$20,$FC,$38,$D3,$F8
	.dc.b	$73,$47,$CE,$3E,$AC,$7D,$AA,$F8,$51,$A3,$D8,$6A,$7C,$D1,$A8,$F3
	.dc.b	$4F,$81,$F9,$93,$E0,$7F,$12,$73,$AF,$A2,$FC,$C8,$F8,$C7,$F3,$C8
	.dc.b	$FE,$0F,$1F,$18,$E1,$5F,$04,$FC,$53,$FE,$2A,$8F,$EC,$8E,$CF,$19
	.dc.b	$ED,$3E,$11,$B8,$FE,$C7,$5F,$46,$3E,$BD,$79,$DF,$C7,$57,$EC,$E3
	.dc.b	$EE,$BE,$83,$E8,$C7,$3B,$FC,$D3,$E0,$3E,$39,$E0,$3F,$C5,$07,$F8
	.dc.b	$5A,$3D,$AF,$F7,$2A,$F1,$3F,$36,$BE,$19,$FD,$2E,$77,$1F,$82,$8F
	.dc.b	$D0,$5F,$C3,$AF,$B1,$7F,$8E,$AF,$DD,$9F,$0E,$FD,$57,$F8,$AA,$FC
	.dc.b	$7B,$D8,$FE,$65,$3B,$BF,$E6,$D3,$F6,$E3,$EB,$4F,$C5,$7F,$43,$5D
	.dc.b	$27,$4F,$C4,$8F,$7B,$F5,$1F,$71,$F6,$A3,$C3,$3C,$E7,$F3,$E7,$39
	.dc.b	$FB,$07,$E2,$8F,$9C,$7E,$78,$F9,$CF,$D3,$AF,$C1,$C7,$81,$FB,$AF
	.dc.b	$73,$EE,$9C,$8F,$D1,$1C,$9F,$95,$3F,$9C,$BF,$95,$5F,$42,$FE,$AD
	.dc.b	$FD,$B8,$FF,$24,$BF,$6E,$FF,$AF,$BD,$CF,$B5,$7E,$D8,$FD,$34,$FB
	.dc.b	$67,$F0,$A7,$B4,$F4,$5F,$E2,$A3,$47,$B6,$FE,$ED,$FE,$96,$FB,$DF
	.dc.b	$A6,$7E,$5B,$ED,$3F,$5E,$64,$F8,$27,$63,$E5,$57,$69,$F8,$35,$F8
	.dc.b	$39,$FE,$2C,$7D,$83,$91,$EC,$3B,$1F,$A6,$8F,$C3,$BF,$60,$FC,$39
	.dc.b	$A3,$F0,$EE,$9F,$82,$68,$F2,$1F,$52,$7C,$95,$EC,$8C,$9E,$C8,$CB
	.dc.b	$EC,$7E,$99,$F8,$F7,$F4,$D5,$FB,$8A,$F6,$BF,$4E,$FE,$EC,$7C,$AA
	.dc.b	$E7,$5E,$D3,$B9,$ED,$AF,$D7,$DF,$8E,$7B,$D7,$E3,$AB,$E2,$9F,$85
	.dc.b	$7F,$CE,$B7,$F6,$DF,$81,$3F,$6E,$FF,$1C,$7D,$BA,$FD,$83,$D9,$F6
	.dc.b	$4F,$27,$EB,$BC,$8F,$5B,$F3,$67,$F5,$31,$9B,$F3,$3C,$67,$E6,$39
	.dc.b	$7A,$9F,$EA,$47,$D5,$3F,$0A,$3D,$46,$6F,$D5,$7E,$58,$F8,$D3,$F0
	.dc.b	$A7,$33,$F3,$EF,$C3,$5C,$4F,$4C,$71,$7A,$9F,$E9,$89,$FF,$5E,$CF
	.dc.b	$EB,$A3,$F4,$95,$D5,$E3,$5D,$5F,$49,$C4,$FF,$5D,$1F,$80,$7F,$8F
	.dc.b	$3A,$3F,$D7,$47,$27,$2F,$AA,$FE,$33,$F8,$D7,$E3,$39,$3D,$47,$CA
	.dc.b	$8F,$53,$E6,$8E,$2F,$E1,$E3,$CC,$F1,$3C,$C7,$F0,$73,$EF,$BA,$7E
	.dc.b	$19,$E6,$3B,$39,$7B,$1E,$23,$CC,$F8,$9F,$D5,$4F,$C7,$78,$C7,$B9
	.dc.b	$EB,$7F,$ED,$97,$C3,$19,$7A,$9E,$69,$FD,$55,$FD,$38,$EB,$7E,$E7
	.dc.b	$F7,$75,$E1,$AF,$E8,$07,$86,$3C,$B1,$E9,$78,$9E,$93,$E3,$3E,$17
	.dc.b	$27,$F0,$F8,$FC,$93,$E4,$3E,$E5,$7F,$4E,$AF,$E7,$F3,$DE,$79,$CE
	.dc.b	$4F,$DE,$BE,$E7,$15,$FA,$F8,$ED,$5F,$79,$F8,$A7,$FA,$64,$E0,$7F
	.dc.b	$1B,$AF,$6C,$7F,$07,$38,$47,$96,$BE,$34,$EA,$3D,$C7,$C6,$AF,$85
	.dc.b	$19,$8F,$0C,$7C,$9A,$F5,$9D,$5C,$57,$F4,$07,$FC,$05,$19,$79,$BC
	.dc.b	$27,$CE,$7F,$2C,$AE,$CF,$C8,$3D,$51,$B7,$EF,$3F,$C0,$0F,$5B,$F4
	.dc.b	$8E,$07,$37,$83,$88,$FE,$35,$3F,$D1,$E3,$FC,$1E,$7B,$1F,$2D,$7D
	.dc.b	$28,$C4,$7D,$F9,$F2,$CF,$C8,$AE,$B7,$F9,$5B,$DC,$FD,$8B,$F8,$31
	.dc.b	$EE,$BF,$0C,$7F,$86,$23,$67,$BA,$7E,$44,$66,$3D,$17,$FC,$A1,$F3
	.dc.b	$98,$9F,$E2,$EF,$23,$E0,$3C,$9F,$EC,$66,$CF,$E1,$8F,$A4,$F7,$39
	.dc.b	$3F,$80,$1F,$BB,$3F,$47,$3F,$06,$7E,$5C,$F1,$8F,$CA,$4F,$13,$F2
	.dc.b	$67,$FA,$06,$FF,$92,$BF,$E0,$D3,$F8,$3C,$F1,$9F,$8B,$1E,$9B,$F8
	.dc.b	$AF,$F4,$5B,$FF,$08,$9F,$7A,$3E,$8C,$73,$3F,$34,$53,$F0,$57,$F4
	.dc.b	$DF,$21,$A3,$FA,$F9,$63,$F6,$EF,$03,$F5,$47,$8E,$7E,$84,$7F,$8A
	.dc.b	$E3,$B3,$E2,$79,$9C,$4F,$80,$7F,$3C,$BF,$75,$FD,$B6,$C7,$D4,$6B
	.dc.b	$FD,$E0,$FD,$84,$7C,$F8,$FA,$A6,$A7,$C5,$3E,$07,$F4,$F7,$EF,$9D
	.dc.b	$5F,$EB,$EF,$91,$F5,$A3,$37,$FB,$EB,$FE,$E3,$3F,$B8,$35,$1E,$EA
	.dc.b	$FB,$ED,$AF,$C6,$3F,$92,$79,$BF,$1D,$FC,$9B,$F4,$0F,$89,$5F,$E7
	.dc.b	$AB,$FC,$1C,$7A,$6B,$AB,$6B,$DE,$F7,$BF,$F0,$79,$F8,$67,$A5,$FD
	.dc.b	$13,$EC,$96,$BF,$0C,$7E,$D2,$3E,$D9,$C1,$EF,$1C,$1A,$70,$AF,$6C
	.dc.b	$7C,$E3,$4F,$92,$7C,$D5,$F1,$9A,$FF,$0B,$9E,$15,$CE,$35,$1C,$9B
	.dc.b	$19,$78,$3F,$5E,$BF,$84,$47,$E1,$EB,$E1,$3F,$06,$3E,$7C,$7E,$D4
	.dc.b	$FE,$92,$7F,$09,$8E,$A7,$DB,$38,$3C,$9F,$F2,$51,$A3,$F8,$F9,$4F
	.dc.b	$C3,$9C,$1F,$6B,$C0,$C5,$7A,$DB,$5E,$B7,$D1,$3C,$0F,$DD,$5F,$C9
	.dc.b	$AE,$13,$A6,$CF,$E8,$CF,$31,$C2,$3C,$97,$F1,$23,$47,$53,$F2,$47
	.dc.b	$03,$F5,$53,$F9,$C9,$F6,$9E,$63,$E2,$1E,$1B,$F4,$4F,$78,$D1,$C0
	.dc.b	$D1,$C1,$F0,$CF,$A1,$EF,$3F,$C5,$A7,$51,$E4,$3C,$D3,$F8,$48,$CB
	.dc.b	$FE,$01,$8F,$DF,$D7,$B5,$E1,$1D,$4E,$13,$E3,$8F,$33,$E4,$30,$F0
	.dc.b	$AF,$5B,$63,$BC,$7F,$88,$67,$E2,$47,$DE,$39,$16,$7D,$B3,$CA,$7C
	.dc.b	$C7,$0A,$F3,$3E,$4A,$E1,$5D,$DF,$1C,$70,$AF,$89,$18,$7F,$3D,$3F
	.dc.b	$97,$8F,$9E,$72,$9E,$17,$F8,$7A,$F1,$1C,$2B,$FA,$ED,$7E,$11,$E4
	.dc.b	$FE,$54,$F5,$DF,$9A,$78,$5F,$F4,$CA,$F3,$57,$79,$F8,$87,$99,$C4
	.dc.b	$7F,$77,$8F,$E8,$27,$22,$BF,$CA,$5E,$15,$DE,$07,$FB,$44,$B2,$24
	.dc.b	$6A,$23,$58,$1E,$11,$C6,$32,$75,$3F,$DA,$85,$39,$BC,$1D,$9C,$CE
	.dc.b	$CF,$48,$EE,$6A,$3C,$11,$C5,$EF,$1E,$48,$E1,$1F,$EF,$A3,$8B,$B7
	.dc.b	$99,$D9,$F7,$BB,$3D,$D1,$EC,$8F,$03,$D5,$DC,$7C,$43,$D3,$1B,$69
	.dc.b	$D6,$3D,$8E,$CF,$3B,$E3,$7C,$0F,$A1,$D1,$E8,$7E,$63,$EE,$3A,$47
	.dc.b	$A9,$E7,$1A,$70,$7D,$C7,$CE,$FD,$13,$F0,$31,$E7,$7A,$1F,$7E,$3C
	.dc.b	$4E,$CF,$9E,$F5,$7C,$EF,$47,$DA,$79,$DF,$8E,$FF,$1F,$A5,$7E,$27
	.dc.b	$F8,$03,$F1,$91,$8A,$FC,$4C,$62,$4F,$E2,$A7,$8B,$FA,$F9,$FE,$0F
	.dc.b	$5F,$95,$8E,$11,$CA,$3F,$3B,$3C,$1F,$CB,$BC,$A3,$F6,$91,$FB,$67
	.dc.b	$FB,$C4,$F2,$7F,$44,$F8,$CF,$15,$79,$23,$8D,$7E,$DC,$F6,$D7,$8C
	.dc.b	$F0,$CF,$E2,$6B,$94,$78,$A3,$95,$7F,$0C,$BF,$DA,$4F,$ED,$2B,$C5
	.dc.b	$3F,$0A,$39,$C7,$88,$F8,$51,$FC,$5E,$3C,$02,$FE,$64,$7A,$A3,$89
	.dc.b	$E1,$9E,$31,$FC,$2A,$3C,$07,$A1,$F9,$4F,$3B,$F1,$3F,$C0,$6B,$E5
	.dc.b	$4F,$A6,$3F,$8E,$9F,$E7,$AA,$F4,$9F,$01,$E1,$5F,$02,$7F,$90,$C7
	.dc.b	$03,$FD,$0F,$3C,$1F,$39,$C1,$F3,$DF,$C6,$8F,$1D,$FB,$AA,$DF,$9A
	.dc.b	$7E,$C4,$7D,$68,$FE,$47,$1E,$E3,$D9,$1E,$E9,$FC,$34,$77,$8E,$15
	.dc.b	$FE,$46,$BF,$24,$FC,$37,$E9,$4E,$A3,$EE,$4F,$FA,$5A,$77,$5D,$27
	.dc.b	$C7,$5F,$1D,$FC,$83,$FE,$A9,$9D,$DF,$A6,$7E,$FC,$F5,$7F,$D3,$F3
	.dc.b	$A3,$C3,$1C,$EB,$F2,$55,$FA,$DA,$FB,$53,$F8,$17,$D0,$7E,$05,$E8
	.dc.b	$FE,$BE,$FD,$0F,$08,$F9,$77,$D9,$FA,$F1,$E8,$3E,$F1,$B8,$FB,$CF
	.dc.b	$BA,$37,$7E,$48,$FC,$AB,$F5,$E7,$71,$EE,$9F,$55,$FD,$68,$FB,$71
	.dc.b	$F5,$63,$E2,$47,$D1,$8F,$4B,$FA,$EB,$F4,$BF,$72,$3E,$05,$FD,$18
	.dc.b	$ED,$1E,$4A,$ED,$7F,$2E,$7E,$24,$F8,$AF,$F0,$11,$ED,$3F,$26,$7B
	.dc.b	$4F,$F0,$8C,$7F,$12,$7F,$19,$7E,$D7,$67,$80,$DC,$7E,$DD,$DD,$FF
	.dc.b	$05,$76,$7B,$63,$C4,$FE,$31,$FF,$6C,$DF,$9E,$3F,$2D,$1A,$7F,$0A
	.dc.b	$79,$CF,$BE,$69,$F8,$17,$EC,$8F,$8D,$1F,$79,$F8,$2E,$9F,$03,$FC
	.dc.b	$91,$E9,$7F,$2D,$F3,$46,$DF,$09,$D2,$BE,$EB,$A3,$E0,$C7,$F0,$F7
	.dc.b	$EE,$BF,$5A,$FE,$ED,$7F,$0C,$8E,$93,$F9,$57,$E0,$BE,$29,$D9,$FC
	.dc.b	$2D,$EB,$7F,$C1,$6F,$E3,$47,$D7,$3F,$C5,$71,$97,$94,$7E,$75,$E2
	.dc.b	$7F,$13,$BF,$E7,$6F,$D6,$BF,$C9,$4E,$E3,$99,$BB,$FD,$83,$F0,$4F
	.dc.b	$B7,$5E,$B7,$A9,$FA,$93,$F8,$F9,$E2,$3E,$89,$A7,$F5,$0F,$86,$BE
	.dc.b	$35,$78,$67,$FC,$B9,$7E,$BA,$F0,$1F,$80,$3E,$04,$69,$F3,$1F,$B2
	.dc.b	$BF,$C8,$3B,$3F,$2C,$7B,$0F,$34,$7C,$27,$EB,$DF,$E1,$DF,$C6,$1E
	.dc.b	$79,$FD,$5B,$EB,$32,$73,$9F,$5B,$CA,$FF,$0B,$1F,$16,$35,$3C,$9F
	.dc.b	$5C,$7E,$3C,$D1,$EB,$7C,$B1,$F4,$23,$F6,$11,$EB,$3F,$20,$74,$BF
	.dc.b	$D0,$46,$8E,$93,$F8,$99,$D1,$FA,$0B,$FE,$68,$FA,$CF,$97,$1F,$E6
	.dc.b	$DA,$E6,$FC,$AA,$E7,$7F,$D5,$AB,$E7,$9F,$66,$FE,$43,$C8,$FF,$3C
	.dc.b	$5F,$28,$FC,$6C,$7E,$08,$F1,$9F,$AC,$AF,$45,$FE,$82,$FB,$4F,$C2
	.dc.b	$3E,$E5,$72,$7E,$8D,$7C,$73,$E8,$CF,$E9,$23,$EF,$C6,$6B,$E8,$3D
	.dc.b	$A7,$F1,$D1,$EF,$9F,$88,$7F,$40,$3E,$AB,$E3,$AF,$E6,$71,$FE,$07
	.dc.b	$BF,$E6,$77,$CE,$3D,$8F,$78,$F8,$13,$F9,$C9,$CD,$77,$7E,$03,$F3
	.dc.b	$4E,$D1,$FC,$7E,$3F,$82,$3F,$C5,$5F,$1D,$FA,$1F,$9A,$7A,$0F,$D7
	.dc.b	$9A,$3B,$9A,$3B,$C7,$C9,$7E,$B1,$FE,$66,$3F,$CB,$53,$F8,$F3,$C9
	.dc.b	$3D,$E3,$47,$FB,$AA,$35,$3C,$E3,$E7,$C7,$B5,$F4,$C7,$E3,$9F,$21
	.dc.b	$F6,$23,$4F,$F5,$D8,$FE,$42,$E8,$F2,$9F,$E8,$CB,$FF,$79,$4F,$82
	.dc.b	$7E,$64,$F3,$3D,$C7,$80,$FB,$53,$F5,$5D,$1E,$A7,$9B,$F8,$C8,$E6
	.dc.b	$F8,$2F,$FC,$AA,$7F,$11,$9E,$51,$93,$E3,$9C,$67,$F4,$2F,$C2,$3C
	.dc.b	$67,$B5,$FB,$51,$CA,$34,$78,$EB,$FD,$DA,$F3,$BF,$E1,$31,$FB,$99
	.dc.b	$FC,$64,$FD,$F7,$F8,$C4,$FE,$7E,$3F,$81,$C6,$9F,$BF,$5E,$C7,$B9
	.dc.b	$F2,$EB,$BB,$FB,$9A,$EE,$7E,$C2,$BF,$57,$1F,$C1,$4E,$37,$F1,$1F
	.dc.b	$E0,$27,$F1,$F3,$FA,$1D,$79,$CF,$BA,$FE,$9E,$FE,$ED,$7C,$18,$F3
	.dc.b	$CE,$A3,$FD,$53,$5F,$AC,$8F,$C8,$CF,$E6,$8F,$8D,$5F,$E1,$37,$C0
	.dc.b	$7F,$23,$AF,$E5,$33,$CC,$FF,$5D,$CF,$E4,$63,$C0,$6A,$7E,$A9,$E5
	.dc.b	$8F,$AB,$3E,$08,$FC,$39,$F1,$E3,$C1,$7F,$3E,$B9,$4F,$92,$7F,$C4
	.dc.b	$0F,$92,$3F,$93,$C7,$C0,$AE,$93,$F9,$CA,$E5,$1F,$1E,$3C,$75,$F0
	.dc.b	$AB,$81,$FA,$D8,$F8,$E7,$E1,$4F,$9F,$5F,$1E,$35,$3E,$49,$F4,$47
	.dc.b	$92,$7B,$9B,$7F,$09,$1D,$0F,$C1,$1D,$1F,$E3,$F3,$E3,$7E,$29,$DA
	.dc.b	$BD,$13,$C0,$FA,$4E,$1F,$84,$FF,$1C,$8F,$F0,$FC,$7E,$8D,$FA,$91
	.dc.b	$F1,$23,$F2,$B5,$E6,$9F,$82,$79,$AB,$D9,$7F,$91,$AF,$E8,$87,$86
	.dc.b	$3E,$7C,$F9,$8E,$17,$E3,$BE,$91,$FC,$B6,$BC,$47,$F0,$69,$F1,$3F
	.dc.b	$0C,$F4,$CF,$F2,$16,$9F,$93,$3E,$6C,$FF,$8F,$E3,$AB,$EC,$3E,$33
	.dc.b	$F3,$DF,$E0,$B1,$D6,$FD,$57,$FA,$E8,$FE,$40,$7C,$23,$F6,$07,$8E
	.dc.b	$BA,$BF,$E1,$39,$FE,$9B,$5D,$2B,$F0,$15,$AA,$FE,$87,$1F,$A0,$BF
	.dc.b	$A9,$7E,$07,$FB,$19,$F0,$23,$8C,$F8,$1A,$F9,$0F,$41,$FC,$80,$FA
	.dc.b	$11,$F7,$5F,$95,$7F,$00,$FB,$27,$96,$3C,$31,$C4,$F9,$67,$EB,$AF
	.dc.b	$D1,$5F,$C7,$23,$C5,$3C,$1F,$EA,$07,$E8,$4F,$C7,$CF,$CD,$3F,$8C
	.dc.b	$D7,$81,$AF,$DF,$8F,$DF,$3F,$54,$FC,$2C,$71,$8F,$D0,$47,$E2,$4F
	.dc.b	$44,$7F,$23,$9F,$8F,$1E,$D3,$51,$D9,$E8,$70,$3F,$53,$1F,$C4,$A7
	.dc.b	$E2,$C7,$80,$F3,$3E,$07,$8C,$F3,$8F,$BC,$7F,$3F,$9F,$EC,$13,$D4
	.dc.b	$F0,$C7,$13,$AB,$93,$F5,$EF,$C0,$33,$3D,$5F,$C2,$CF,$E0,$4F,$D3
	.dc.b	$3F,$42,$3E,$05,$FE,$39,$E1,$7F,$56,$7F,$08,$FE,$18,$FC,$74,$70
	.dc.b	$8F,$41,$F5,$E3,$F6,$B7,$EE,$8F,$41,$E2,$7E,$61,$E8,$8F,$E2,$0E
	.dc.b	$E3,$EF,$BF,$21,$AF,$C2,$6B,$C6,$FE,$84,$75,$8F,$A0,$6E,$7C,$71
	.dc.b	$F1,$0F,$1B,$FD,$A5,$FF,$05,$1C,$63,$E6,$3B,$3B,$96,$39,$BF,$E5
	.dc.b	$C3,$8B,$F1,$CF,$8B,$1E,$97,$66,$67,$E9,$57,$C9,$AF,$EF,$35,$E0
	.dc.b	$7C,$51,$F4,$A7,$F0,$6D,$8F,$E9,$31,$F0,$2B,$A5,$7C,$C9,$F3,$4F
	.dc.b	$ED,$DE,$67,$C0,$7F,$B4,$5F,$E8,$E3,$DA,$68,$B5,$D2,$BE,$59,$F3
	.dc.b	$E7,$84,$7B,$2B,$CB,$5D,$4F,$C1,$3E,$18,$F9,$D1,$FC,$C8,$EC,$F1
	.dc.b	$77,$7F,$CC,$4F,$E0,$B3,$E6,$BF,$94,$F3,$3B,$CF,$99,$B5,$CE,$3F
	.dc.b	$BB,$4F,$F3,$77,$F3,$CF,$2A,$ED,$18,$8F,$E1,$13,$F9,$28,$F9,$33
	.dc.b	$A7,$F4,$35,$FA,$86,$C7,$D9,$7F,$03,$1E,$23,$CC,$FB,$0F,$E9,$51
	.dc.b	$EC,$7E,$CD,$FC,$A9,$DC,$7F,$2E,$9F,$05,$7E,$2A,$FC,$B5,$F7,$CE
	.dc.b	$D3,$A7,$F1,$71,$ED,$AF,$EB,$17,$FD,$04,$FD,$E3,$F5,$4E,$C7,$90
	.dc.b	$F8,$51,$F7,$4E,$4F,$F3,$73,$27,$C4,$3A,$9F,$CF,$6B,$D8,$F4,$38
	.dc.b	$1F,$1A,$3F,$92,$CF,$BE,$7E,$64,$8F,$F7,$C9,$5C,$30,$AA,$61,$98
	.dc.b	$D4,$58,$CB,$93,$51,$CA,$38,$B8,$11,$C6,$3B,$3C,$63,$4F,$82,$38
	.dc.b	$47,$03,$CC,$F8,$1F,$19,$C1,$EE,$F8,$A3,$B4,$71,$8F,$69,$B7,$DC
	.dc.b	$F8,$0E,$AF,$91,$EA,$73,$7A,$BE,$73,$E4,$BE,$37,$A9,$B7,$AB,$E3
	.dc.b	$3A,$BE,$37,$E9,$1E,$37,$DF,$1E,$83,$B4,$7C,$D8,$F0,$BE,$43,$EE
	.dc.b	$3E,$17,$99,$E1,$79,$BE,$08,$E6,$FD,$A8,$E6,$FA,$0E,$6F,$C1,$70
	.dc.b	$6D,$F5,$BF,$06,$FF,$51,$1F,$8A,$57,$F8,$BF,$E0,$1C,$3F,$6D,$A7
	.dc.b	$E1,$A3,$6F,$E5,$EB,$FC,$84,$F0,$AF,$C8,$BC,$23,$F2,$53,$C2,$7F
	.dc.b	$53,$1F,$ED,$C9,$E1,$1F,$99,$78,$47,$E7,$67,$84,$FF,$57,$AF,$E0
	.dc.b	$72,$AE,$33,$63,$F5,$15,$E6,$7C,$35,$B9,$F0,$C7,$FB,$0A,$B8,$9E
	.dc.b	$07,$DF,$1A,$7F,$CC,$51,$A7,$AC,$F7,$7C,$CF,$59,$F3,$5F,$F9,$9E
	.dc.b	$B8,$52,$FC,$8F,$08,$F9,$4F,$FA,$B9,$DC,$6A,$3C,$75,$D1,$F9,$E7
	.dc.b	$89,$F9,$EF,$92,$7E,$B5,$F8,$E7,$FC,$11,$5F,$D0,$6B,$D0,$FE,$02
	.dc.b	$38,$C7,$96,$7D,$8F,$82,$38,$C7,$D0,$AF,$2B,$A9,$F0,$4F,$C7,$AF
	.dc.b	$F5,$85,$68,$F1,$9E,$E9,$FC,$01,$E9,$9F,$C0,$3E,$93,$EB,$C6,$DF
	.dc.b	$AC,$6E,$3D,$35,$E1,$8F,$44,$ED,$FA,$E6,$DF,$B8,$F9,$A7,$E5,$3F
	.dc.b	$3C,$D1,$E1,$3E,$A3,$EB,$9F,$F5,$D4,$FA,$E7,$CB,$39,$BF,$E4,$11
	.dc.b	$F8,$43,$F9,$AC,$ED,$E7,$3E,$07,$E0,$9F,$3D,$FA,$37,$F7,$E3,$CD
	.dc.b	$1F,$42,$BF,$15,$1E,$63,$DE,$78,$A3,$A3,$F1,$8E,$8F,$CC,$76,$78
	.dc.b	$9E,$8F,$E2,$8D,$BF,$32,$BE,$BB,$ED,$3F,$9A,$D7,$F0,$69,$F6,$9E
	.dc.b	$37,$D0,$FD,$28,$E0,$7F,$2A,$8E,$11,$FD,$6E,$7E,$23,$F8,$D3,$FB
	.dc.b	$41,$F4,$CF,$1B,$F6,$1E,$93,$EE,$7E,$E9,$A3,$EC,$1C,$CF,$F1,$4D
	.dc.b	$FE,$5D,$F3,$3F,$87,$AF,$E1,$73,$F0,$A7,$C1,$1F,$A3,$8F,$90,$FB
	.dc.b	$0D,$C7,$98,$FB,$2E,$E3,$E4,$47,$D3,$7A,$3A,$9F,$E5,$4F,$98,$FB
	.dc.b	$A7,$33,$EC,$4F,$3A,$F8,$0F,$82,$3D,$77,$F8,$27,$E0,$1F,$70,$F8
	.dc.b	$11,$FE,$53,$BF,$F3,$33,$F4,$23,$F4,$F1,$F8,$EB,$F8,$47,$F0,$58
	.dc.b	$F1,$1F,$32,$7C,$4F,$9C,$FA,$33,$EC,$7C,$F3,$EC,$7C,$47,$B0,$F3
	.dc.b	$9F,$8D,$9E,$6E,$8E,$67,$81,$FE,$0E,$FD,$6B,$FC,$8D,$7D,$F9,$CC
	.dc.b	$FC,$07,$EA,$57,$A6,$3B,$9F,$69,$F2,$46,$AB,$F8,$23,$C8,$F1,$99
	.dc.b	$AF,$CB,$9F,$D8,$2B,$F3,$A7,$7A,$F8,$F5,$C9,$FE,$07,$1E,$73,$F0
	.dc.b	$07,$AD,$CB,$C6,$3D,$73,$F2,$63,$E7,$4F,$98,$EB,$1E,$28,$FE,$14
	.dc.b	$7E,$9C,$EC,$7E,$E4,$ED,$7F,$C4,$1F,$E7,$07,$A2,$FD,$B5,$D2,$BE
	.dc.b	$53,$E4,$8C,$DF,$92,$B9,$D7,$EF,$AB,$FC,$05,$1E,$03,$F1,$AF,$F9
	.dc.b	$B1,$FB,$8F,$A6,$79,$47,$D2,$3F,$41,$3F,$8F,$8C,$D7,$E5,$6F,$FD
	.dc.b	$6D,$5E,$D3,$E3,$1F,$28,$E4,$7E,$3A,$B9,$39,$9E,$4F,$F9,$DC,$FC
	.dc.b	$EC,$7E,$E2,$FE,$03,$F3,$E7,$F0,$11,$E7,$9E,$4F,$AD,$FC,$44,$F6
	.dc.b	$7E,$71,$E2,$3E,$BB,$97,$E1,$9F,$E9,$93,$8B,$E1,$BF,$7C,$F5,$9F
	.dc.b	$87,$39,$7C,$27,$13,$2F,$F0,$A7,$C3,$5E,$E7,$FD,$74,$7D,$D9,$F5
	.dc.b	$3A,$7B,$3F,$1E,$7B,$39,$3B,$19,$7B,$3E,$A3,$FD,$28,$78,$AB,$F8
	.dc.b	$8C,$FE,$B6,$BC,$47,$F8,$26,$3F,$9C,$CF,$EF,$E3,$33,$FC,$BE,$3E
	.dc.b	$D9,$E1,$3A,$9E,$19,$EA,$7E,$AE,$FF,$35,$5F,$C5,$E3,$AC,$FF,$06
	.dc.b	$3E,$94,$7D,$AB,$FA,$51,$E1,$BF,$82,$FD,$68,$F8,$2F,$B4,$FF,$45
	.dc.b	$4F,$F9,$81,$FD,$69,$F8,$88,$F8,$6F,$33,$E3,$9C,$DC,$BE,$38,$F7
	.dc.b	$DF,$F3,$37,$F5,$71,$F7,$0F,$79,$C6,$FD,$F5,$FB,$D8,$FD,$B9,$FA
	.dc.b	$19,$E2,$FC,$73,$CF,$7F,$ED,$89,$FE,$D4,$FE,$D2,$3E,$69,$ED,$BF
	.dc.b	$F6,$6B,$EB,$BE,$F5,$FC,$02,$7E,$6D,$71,$BE,$F7,$C4,$FF,$6C,$9E
	.dc.b	$43,$91,$FD,$05,$FE,$A4,$7D,$57,$F8,$21,$F1,$CF,$54,$7E,$51,$E8
	.dc.b	$7F,$B7,$6B,$D2,$FE,$65,$FB,$2F,$CB,$8F,$21,$C5,$E8,$78,$8F,$E9
	.dc.b	$13,$E2,$8F,$BE,$FE,$0E,$7E,$FC,$F0,$BF,$C4,$CE,$E3,$C1,$3A,$7C
	.dc.b	$07,$DE,$8E,$51,$F7,$9E,$11,$EC,$8F,$3C,$62,$3C,$8F,$F4,$98,$F9
	.dc.b	$C7,$F2,$98,$EE,$F3,$AF,$D5,$1E,$E7,$9B,$88,$DD,$7D,$89,$FE,$31
	.dc.b	$3F,$CD,$CF,$89,$3E,$83,$C9,$1E,$BA,$F2,$57,$C0,$AE,$95,$F9,$29
	.dc.b	$F6,$CF,$58,$FD,$0B,$B3,$F0,$AF,$88,$FC,$04,$F2,$7F,$27,$5F,$EE
	.dc.b	$7A,$FE,$05,$1F,$82,$3C,$87,$C4,$9F,$E6,$2F,$63,$51,$F2,$6F,$F2
	.dc.b	$F7,$E1,$BF,$91,$5E,$C7,$E4,$47,$92,$30,$FA,$5F,$E0,$77,$F1,$67
	.dc.b	$E5,$CE,$E7,$C4,$FF,$32,$7D,$87,$E2,$A7,$D9,$5E,$28,$F2,$C7,$BE
	.dc.b	$7C,$95,$F9,$89,$F0,$9F,$12,$3F,$8C,$47,$A2,$3B,$9F,$C0,$4A,$7E
	.dc.b	$EE,$39,$C7,$DC,$AF,$34,$62,$7E,$31,$FB,$C3,$F3,$47,$91,$EF,$3E
	.dc.b	$38,$EA,$7F,$1D,$8E,$4F,$EE,$5B,$7F,$C5,$A3,$FB,$BC,$FC,$63,$D1
	.dc.b	$3A,$BF,$CA,$47,$E2,$DF,$8B,$3F,$C1,$2B,$F8,$03,$E5,$3E,$FD,$71
	.dc.b	$7E,$6C,$7E,$8E,$BD,$C7,$68,$F5,$35,$E5,$1F,$98,$73,$1F,$AA,$3D
	.dc.b	$D3,$D6,$3E,$89,$F6,$5F,$92,$F0,$8E,$E7,$07,$07,$CC,$9F,$E3,$EF
	.dc.b	$CC,$3C,$35,$EE,$7F,$8B,$4F,$F9,$05,$EF,$3E,$B7,$94,$FE,$CA,$3D
	.dc.b	$11,$CA,$BE,$AD,$71,$8F,$AA,$77,$9F,$0B,$F2,$0F,$29,$F9,$AB,$F5
	.dc.b	$BC,$E7,$E3,$4F,$03,$B9,$6F,$D0,$FF,$5B,$9F,$C0,$1D,$1F,$AE,$6E
	.dc.b	$FA,$B6,$BF,$A9,$3A,$AF,$61,$FA,$1A,$E4,$FE,$12,$7F,$66,$70,$71
	.dc.b	$5F,$B3,$AE,$53,$F8,$D3,$E9,$57,$AD,$FB,$87,$C5,$9F,$B9,$1F,$18
	.dc.b	$F8,$86,$E3,$ED,$9B,$7C,$67,$E3,$AB,$F8,$6B,$F0,$0E,$07,$F9,$6A
	.dc.b	$3C,$35,$EF,$6C,$FB,$AB,$E1,$B6,$3D,$95,$D6,$7F,$84,$3F,$82,$9E
	.dc.b	$F5,$EE,$AF,$D3,$57,$E0,$2B,$F0,$B1,$F0,$4F,$43,$E4,$3C,$D3,$E3
	.dc.b	$8F,$6C,$70,$79,$9F,$80,$AF,$23,$E8,$3A,$47,$B5,$EB,$3C,$E3,$94
	.dc.b	$7D,$63,$81,$D0,$FA,$2E,$AB,$EC,$57,$DA,$8F,$E3,$93,$E1,$3B,$BF
	.dc.b	$95,$70,$73,$3C,$31,$E6,$69,$C2,$BF,$88,$4F,$F9,$92,$3C,$86,$8E
	.dc.b	$4F,$8C,$F0,$BE,$67,$E3,$CF,$57,$47,$F0,$18,$F1,$9D,$1F,$88,$6A
	.dc.b	$7F,$70,$F2,$8F,$24,$FB,$E3,$94,$77,$9F,$B8,$F3,$8E,$0F,$99,$AE
	.dc.b	$63,$DB,$1F,$E6,$73,$D3,$1E,$D3,$53,$DE,$7C,$67,$09,$FF,$31,$1C
	.dc.b	$0F,$44,$7E,$38,$E0,$57,$FC,$2C,$FF,$5F,$9F,$24,$FA,$2B,$ED,$D7
	.dc.b	$E6,$4F,$EA,$57,$FD,$75,$FC,$94,$ED,$F0,$1E,$C9,$E2,$FF,$81,$EF
	.dc.b	$B5,$FA,$A3,$FD,$C6,$0F,$F8,$00,$7F,$AE,$4C,$12,$7F,$DA,$8E,$0D
	.dc.b	$1C,$03,$97,$4F,$07,$FC,$BC,$77,$81,$23,$81,$C6,$3C,$11,$A3,$0F
	.dc.b	$81,$C4,$70,$8C,$BD,$DF,$69,$E5,$78,$BE,$68,$E8,$F0,$3D,$8F,$47
	.dc.b	$CE,$7A,$9D,$BE,$58,$F7,$46,$E3,$B9,$E8,$7D,$AF,$53,$E5,$BB,$7E
	.dc.b	$49,$B7,$CB,$1D,$1F,$89,$1E,$23,$47,$33,$E5,$3A,$7C,$27,$DB,$8F
	.dc.b	$84,$FC,$78,$D3,$E0,$34,$FB,$1F,$94,$79,$9D,$BF,$70,$F9,$AE,$DF
	.dc.b	$CD,$C7,$F2,$C7,$FD,$37,$1F,$E9,$43,$ED,$52,$BE,$D4,$7F,$1A,$3F
	.dc.b	$25,$3F,$83,$3F,$25,$3B,$8E,$51,$F8,$B8,$C4,$7F,$06,$AF,$CE,$CF
	.dc.b	$82,$3E,$AC,$F1,$8E,$15,$E0,$9F,$DD,$57,$07,$F3,$B5,$D1,$F4,$D7
	.dc.b	$07,$FA,$59,$FC,$39,$FB,$27,$9A,$BA,$36,$3D,$F1,$88,$FD,$91,$FB
	.dc.b	$49,$F8,$4F,$20,$F8,$67,$C7,$3F,$88,$9F,$D7,$4E,$5F,$E1,$75,$F3
	.dc.b	$EB,$67,$C6,$8D,$D7,$86,$BF,$82,$47,$F2,$49,$FF,$0D,$1C,$0F,$8F
	.dc.b	$3D,$E3,$11,$F5,$6B,$E3,$47,$B6,$36,$7B,$8F,$8A,$7F,$18,$9F,$D4
	.dc.b	$C6,$E3,$E5,$BE,$47,$67,$E0,$1F,$B0,$E9,$F0,$BF,$59,$F9,$11,$D1
	.dc.b	$F8,$EF,$B4,$F4,$3D,$E3,$83,$EF,$3E,$B1,$F3,$EB,$E5,$3C,$A3,$D4
	.dc.b	$7E,$2D,$EE,$6E,$7B,$46,$8F,$89,$1A,$3F,$30,$73,$7E,$2B,$F8,$97
	.dc.b	$A3,$E3,$3F,$DE,$4F,$B9,$FC,$41,$B7,$AB,$B9,$FC,$3B,$D9,$F4,$9F
	.dc.b	$9D,$7E,$64,$76,$8F,$EA,$73,$F1,$DD,$BF,$1A,$7C,$D3,$DA,$77,$1E
	.dc.b	$F8,$FC,$53,$E5,$3F,$32,$78,$E7,$A1,$F1,$1F,$44,$7D,$09,$E7,$1E
	.dc.b	$79,$D3,$E6,$9F,$04,$7F,$61,$8D,$47,$81,$FB,$47,$E7,$8F,$F5,$83
	.dc.b	$E3,$9F,$81,$16,$7F,$88,$BA,$38,$9F,$74,$F2,$9C,$CF,$D2,$D7,$FB
	.dc.b	$2E,$B9,$1F,$74,$F5,$4F,$96,$7F,$2B,$1F,$AE,$79,$9F,$12,$BD,$EE
	.dc.b	$8F,$B6,$7B,$E3,$9D,$7E,$DE,$3C,$B3,$EA,$6D,$7A,$9F,$2D,$7F,$B4
	.dc.b	$4F,$2B,$E0,$9F,$DD,$9E,$08,$CC,$77,$3D,$C7,$F0,$AA,$F7,$3D,$A3
	.dc.b	$DC,$E8,$FE,$35,$1A,$7F,$0B,$1E,$F7,$A1,$EF,$AF,$89,$1E,$08,$FA
	.dc.b	$47,$F8,$54,$F4,$BF,$ED,$B3,$4F,$B8,$D1,$E5,$7D,$CF,$94,$F7,$47
	.dc.b	$C3,$75,$1C,$DD,$1E,$4A,$F8,$73,$E4,$8C,$9E,$43,$FC,$FD,$5F,$45
	.dc.b	$F6,$9F,$58,$D1,$FE,$97,$9F,$EA,$D1,$F0,$DF,$D4,$D7,$C8,$9F,$85
	.dc.b	$3C,$E7,$31,$E8,$3D,$A7,$CC,$9F,$6B,$E0,$3F,$28,$7A,$6B,$DC,$EE
	.dc.b	$3D,$CE,$E7,$F0,$CD,$9D,$BF,$C8,$AB,$CD,$1F,$AE,$8F,$5C,$7F,$AC
	.dc.b	$AB,$E8,$C7,$13,$CA,$71,$7C,$B5,$EA,$8F,$21,$A7,$E1,$C6,$AB,$B4
	.dc.b	$FD,$B3,$E6,$9F,$B1,$3C,$D1,$F1,$E7,$EA,$D7,$C7,$AE,$35,$E3,$38
	.dc.b	$9F,$88,$AF,$EA,$AF,$DA,$3B,$1F,$56,$3F,$C7,$11,$F1,$AB,$D1,$3F
	.dc.b	$00,$FF,$65,$9F,$C3,$23,$F9,$9B,$EC,$3B,$9F,$8C,$7B,$BE,$83,$ED
	.dc.b	$1A,$3E,$A3,$A8,$EF,$1E,$4A,$F9,$CF,$9C,$F6,$1E,$77,$EA,$CF,$F1
	.dc.b	$D8,$CC,$70,$7C,$07,$94,$E1,$1F,$3C,$E1,$3F,$CE,$A3,$EA,$57,$C2
	.dc.b	$9F,$23,$F3,$27,$53,$E9,$3E,$F4,$FC,$79,$D1,$FD,$A6,$39,$D7,$9A
	.dc.b	$7E,$3C,$7C,$B3,$FD,$1B,$3F,$D0,$23,$C6,$F1,$8E,$D5,$F8,$69,$EC
	.dc.b	$68,$FE,$FF,$5F,$E0,$A9,$F7,$9D,$67,$FD,$BC,$FF,$64,$3E,$53,$EA
	.dc.b	$7E,$1C,$F4,$3C,$AF,$D1,$3E,$F9,$FC,$C4,$FD,$69,$D6,$BF,$D9,$CF
	.dc.b	$53,$C2,$71,$3D,$A7,$F6,$18,$EA,$FD,$63,$AB,$FA,$A9,$F2,$57,$D9
	.dc.b	$7F,$07,$3B,$39,$9F,$91,$9F,$F0,$E3,$DE,$3D,$E7,$F7,$C7,$D0,$FE
	.dc.b	$F8,$EA,$F8,$0F,$87,$5F,$3D,$F8,$73,$F8,$03,$E7,$1F,$E0,$18,$FE
	.dc.b	$45,$1E,$E3,$C9,$3E,$08,$FB,$53,$EE,$8E,$F3,$E8,$3C,$73,$FE,$07
	.dc.b	$8C,$9C,$63,$E3,$4F,$B1,$EB,$3F,$3A,$7D,$93,$F1,$E3,$D1,$5F,$1D
	.dc.b	$FC,$8C,$7E,$12,$BF,$C5,$47,$81,$FE,$11,$1F,$C7,$AB,$F1,$91,$EF
	.dc.b	$3D,$11,$D2,$7E,$31,$EF,$9F,$3C,$78,$CF,$7C,$F9,$E3,$F1,$8F,$94
	.dc.b	$FD,$01,$98,$E9,$3F,$0E,$34,$7E,$CC,$F0,$9C,$0F,$EC,$CF,$E3,$DF
	.dc.b	$F4,$33,$F8,$17,$2F,$F1,$A7,$99,$F9,$B3,$C0,$7C,$37,$FC,$5C,$7C
	.dc.b	$E7,$8B,$FC,$31,$E2,$7C,$33,$89,$F8,$1A,$FE,$5C,$FE,$10,$EC,$72
	.dc.b	$AF,$83,$3F,$8C,$8F,$04,$7F,$82,$E3,$11,$EF,$3A,$9C,$9F,$ED,$07
	.dc.b	$E6,$EB,$E7,$3E,$73,$27,$9C,$E7,$1C,$5F,$DA,$9C,$5F,$8E,$FE,$04
	.dc.b	$FE,$62,$53,$F9,$23,$F0,$1F,$F0,$C1,$FC,$71,$F8,$47,$B6,$3F,$4C
	.dc.b	$EE,$3F,$80,$4F,$ED,$8F,$C2,$C7,$80,$D3,$F9,$88,$F3,$C7,$47,$D7
	.dc.b	$1C,$23,$FD,$3D,$1E,$F3,$D6,$7B,$CF,$C9,$4F,$89,$E6,$7A,$27,$F3
	.dc.b	$6F,$FA,$50,$FC,$31,$DA,$38,$3C,$A7,$F1,$CE,$5F,$AF,$1F,$BE,$9F
	.dc.b	$63,$C1,$B1,$C2,$7C,$B1,$E9,$8F,$F5,$81,$F0,$27,$F4,$AF,$C9,$38
	.dc.b	$47,$F1,$38,$FD,$1B,$FC,$06,$7C,$E7,$48,$F0,$1F,$BC,$9F,$C4,$1C
	.dc.b	$CF,$CA,$47,$F8,$16,$7C,$B1,$A8,$F6,$35,$E8,$FB,$AB,$F3,$6F,$EC
	.dc.b	$0D,$4F,$95,$F3,$D7,$F1,$73,$CE,$7E,$52,$BC,$F3,$B9,$E3,$3B,$78
	.dc.b	$9F,$C4,$27,$FB,$1B,$A3,$0F,$28,$E1,$19,$8F,$CB,$D7,$F1,$D8,$FD
	.dc.b	$81,$F8,$98,$F5,$BB,$7F,$BF,$46,$E3,$D8,$7E,$96,$BC,$CD,$AE,$EF
	.dc.b	$FA,$A2,$7E,$1B,$FB,$43,$B4,$F0,$6C,$F0,$7F,$D2,$35,$F4,$63,$EB
	.dc.b	$47,$53,$DB,$1F,$AF,$3F,$8D,$3F,$9E,$7E,$03,$A3,$F1,$0F,$E8,$67
	.dc.b	$C0,$FE,$76,$7F,$C2,$8F,$E7,$4E,$87,$F2,$13,$F4,$67,$ED,$1F,$1C
	.dc.b	$F9,$A3,$FC,$84,$71,$7B,$BE,$23,$F8,$2B,$87,$FB,$19,$B3,$AC,$FE
	.dc.b	$ED,$FD,$54,$7D,$17,$DA,$FE,$89,$DC,$FE,$89,$FC,$9C,$7C,$57,$EE
	.dc.b	$9F,$D3,$A7,$F3,$31,$98,$FE,$2A,$72,$3F,$12,$E5,$FF,$46,$BE,$73
	.dc.b	$D1,$5F,$81,$7E,$D4,$7A,$4F,$3B,$C5,$FE,$4C,$FD,$37,$E5,$9F,$C3
	.dc.b	$E7,$C7,$1F,$80,$3E,$DC,$F5,$7F,$64,$73,$9E,$15,$F4,$1F,$B1,$3F
	.dc.b	$A2,$7E,$69,$EB,$3E,$81,$F9,$83,$E1,$47,$17,$ED,$4F,$E8,$1F,$55
	.dc.b	$79,$1E,$B1,$F4,$5F,$C1,$1E,$E7,$D0,$FC,$E3,$FB,$81,$E1,$3D,$F1
	.dc.b	$FD,$61,$E0,$74,$8F,$64,$64,$F0,$47,$92,$3F,$5A,$7E,$35,$FF,$40
	.dc.b	$1E,$69,$F5,$4F,$E1,$CE,$25,$72,$7F,$84,$CF,$CE,$C7,$59,$FB,$2F
	.dc.b	$F9,$34,$06,$15,$C9,$A6,$B0,$DB,$B7,$05,$38,$3D,$1E,$2E,$9C,$BB
	.dc.b	$72,$69,$E0,$F8,$9E,$0F,$81,$A7,$8C,$F1,$BC,$8F,$1B,$86,$9D,$5F
	.dc.b	$01,$C8,$EE,$F3,$34,$D7,$9B,$A7,$B1,$EA,$74,$F8,$DD,$3E,$93,$99
	.dc.b	$C0,$A7,$90,$F7,$3D,$4F,$21,$F1,$5E,$46,$9F,$4B,$F2,$DF,$73,$E2
	.dc.b	$3E,$31,$DC,$E0,$FC,$87,$C2,$7D,$D3,$93,$E8,$7F,$56,$7E,$0A,$7F
	.dc.b	$D3,$CD,$9E,$F1,$FE,$85,$8C,$1F,$B5,$3F,$D8,$C6,$E7,$E6,$C6,$E3
	.dc.b	$F2,$2A,$7F,$07,$1F,$9F,$9F,$DB,$29,$E0,$D9,$F0,$9F,$8B,$8E,$51
	.dc.b	$E1,$3C,$13,$F8,$A8,$E2,$77,$7C,$73,$C4,$F0,$CF,$E9,$E3,$F8,$99
	.dc.b	$EC,$7F,$41,$3C,$63,$F3,$CF,$49,$E2,$FE,$CE,$3D,$2E,$5E,$F3,$FA
	.dc.b	$68,$F3,$C7,$F0,$F8,$FF,$79,$9E,$87,$F5,$F1,$D6,$3D,$87,$C8,$9F
	.dc.b	$F5,$73,$B3,$CC,$70,$7C,$67,$F4,$77,$FB,$EC,$FE,$DE,$7C,$8F,$AA
	.dc.b	$7E,$14,$F6,$3C,$67,$78,$F4,$BF,$0A,$3A,$BF,$DB,$5F,$EF,$4F,$F6
	.dc.b	$B3,$C8,$53,$D0,$75,$9F,$C5,$BD,$67,$C4,$FC,$59,$EA,$FE,$48,$D4
	.dc.b	$7A,$63,$EC,$3E,$48,$F8,$07,$58,$F8,$AF,$03,$E2,$9F,$96,$8F,$A1
	.dc.b	$3E,$28,$FC,$64,$73,$74,$FF,$38,$9E,$71,$A8,$F3,$BF,$0C,$F0,$3F
	.dc.b	$C3,$E7,$E6,$BE,$48,$D3,$E0,$9E,$73,$F8,$B8,$D4,$FA,$9D,$1C,$E7
	.dc.b	$DE,$F8,$23,$4F,$94,$D3,$F2,$5F,$53,$F4,$1E,$67,$E8,$E3,$9B,$E0
	.dc.b	$8F,$7B,$FD,$79,$FD,$C9,$F2,$4F,$B4,$7E,$C4,$F7,$BC,$DF,$03,$E5
	.dc.b	$79,$9F,$BB,$7F,$AE,$BC,$A7,$E0,$3F,$50,$F3,$47,$AC,$E8,$FE,$3D
	.dc.b	$F5,$BE,$17,$D6,$7E,$0D,$F0,$BE,$B7,$9B,$EB,$7F,$58,$F7,$7E,$D3
	.dc.b	$B7,$A3,$E1,$3A,$3E,$17,$D8,$F4,$3D,$8F,$F6,$C8,$FB,$F1,$D5,$FD
	.dc.b	$EC,$73,$31,$3F,$C6,$27,$F3,$C7,$CA,$8E,$EF,$EE,$5F,$7C,$F2,$7D
	.dc.b	$51,$FA,$27,$99,$A9,$FD,$3C,$7F,$84,$9F,$C6,$9D,$DF,$C8,$C7,$8D
	.dc.b	$FA,$51,$97,$F1,$71,$D9,$FE,$02,$7C,$63,$D2,$79,$DF,$09,$F1,$1F
	.dc.b	$92,$F1,$9F,$D6,$1F,$7A,$3B,$9C,$4F,$0C,$7A,$5C,$1F,$8D,$3E,$83
	.dc.b	$F5,$63,$E1,$47,$D0,$8F,$E6,$AF,$27,$C5,$1A,$8F,$E7,$47,$F9,$9A
	.dc.b	$3C,$E7,$39,$F8,$EF,$F9,$D6,$3F,$8B,$9E,$13,$E1,$9D,$CE,$47,$77
	.dc.b	$93,$E8,$9F,$C0,$BE,$E8,$FC,$A9,$E0,$7E,$4B,$87,$91,$87,$EA,$3E
	.dc.b	$13,$C7,$1F,$D0,$63,$84,$64,$FD,$A1,$4F,$A2,$79,$CF,$E7,$EF,$85
	.dc.b	$F2,$3D,$A3,$CF,$1C,$5F,$90,$7F,$59,$8F,$DB,$C6,$67,$F0,$CF,$EF
	.dc.b	$0F,$9A,$76,$3D,$67,$68,$FC,$14,$7E,$49,$FB,$C7,$AA,$7C,$27,$B5
	.dc.b	$F2,$47,$F1,$49,$F8,$91,$C8,$FE,$3A,$FE,$10,$C9,$EF,$9F,$F5,$EB
	.dc.b	$E7,$8E,$E7,$47,$F8,$29,$C0,$F4,$C7,$03,$C7,$3F,$D8,$A7,$CC,$F8
	.dc.b	$E7,$D8,$77,$7F,$AF,$BD,$E3,$F2,$31,$F8,$C8,$FF,$42,$1F,$69,$E9
	.dc.b	$1E,$38,$F5,$1D,$0F,$1C,$74,$9F,$1C,$7F,$06,$9F,$1B,$EA,$38,$47
	.dc.b	$D7,$8F,$E2,$53,$E7,$9F,$B0,$7E,$55,$C4,$FC,$F9,$F9,$87,$F2,$E9
	.dc.b	$FE,$20,$FA,$9E,$07,$73,$9B,$E2,$7D,$26,$1F,$63,$C9,$FD,$29,$B7
	.dc.b	$07,$D1,$8F,$44,$7E,$EA,$3E,$53,$C0,$E6,$FC,$93,$E1,$38,$7F,$43
	.dc.b	$1E,$B7,$EA,$47,$F4,$23,$F4,$0F,$95,$F7,$9F,$CB,$9E,$87,$F2,$47
	.dc.b	$27,$08,$E9,$1E,$43,$D1,$3F,$28,$FB,$73,$C2,$3F,$42,$77,$2C,$F2
	.dc.b	$3E,$DB,$DC,$F3,$9F,$C7,$A3,$C1,$3D,$23,$8B,$5F,$11,$F8,$07,$FC
	.dc.b	$79,$3F,$99,$7C,$91,$F1,$A7,$9B,$4E,$65,$3F,$1E,$FC,$77,$27,$94
	.dc.b	$EE,$7A,$CB,$3E,$B9,$F0,$9F,$97,$3C,$31,$C6,$3D,$EF,$F2,$A3,$8C
	.dc.b	$7A,$E3,$F0,$31,$FC,$B0,$C9,$E5,$78,$BF,$C3,$DA,$FF,$71,$3D,$27
	.dc.b	$C9,$3D,$27,$77,$6F,$F2,$43,$A9,$E5,$8F,$4C,$FC,$18,$C9,$C6,$3D
	.dc.b	$6D,$3F,$45,$3F,$05,$FF,$30,$4F,$18,$EE,$7F,$2A,$9E,$33,$E9,$6B
	.dc.b	$E9,$8F,$82,$FF,$29,$3C,$26,$CE,$F1,$F1,$CE,$2F,$AE,$38,$9F,$81
	.dc.b	$76,$7C,$98,$DC,$78,$E3,$27,$C1,$3D,$27,$C9,$3C,$27,$96,$38,$9F
	.dc.b	$E0,$C7,$F4,$E7,$C3,$3F,$CE,$87,$F8,$1A,$3F,$BF,$3F,$87,$7E,$1C
	.dc.b	$71,$8F,$DB,$BE,$B7,$C2,$7F,$93,$8F,$F0,$71,$F9,$D3,$CE,$77,$3F
	.dc.b	$84,$99,$3C,$31,$FE,$0E,$8F,$EA,$B1,$E9,$38,$9F,$2C,$FE,$2F,$1F
	.dc.b	$65,$E3,$1F,$D5,$67,$F9,$01,$FC,$8E,$3D,$E7,$13,$C2,$7B,$63,$F1
	.dc.b	$2F,$C9,$9F,$7B,$FD,$58,$DB,$DD,$B1,$F0,$5F,$0C,$ED,$EE,$7A,$4F
	.dc.b	$3C,$6D,$E2,$7D,$D3,$E4,$9F,$E5,$C3,$F4,$27,$9C,$E6,$7E,$1D,$FB
	.dc.b	$6F,$F3,$C3,$A9,$CC,$A7,$E1,$4A,$7D,$18,$FF,$13,$3F,$68,$E0,$79
	.dc.b	$0E,$05,$38,$1F,$D8,$8E,$06,$8F,$2C,$FE,$55,$EE,$7C,$F8,$EE,$FE
	.dc.b	$10,$FA,$07,$DC,$3E,$31,$DC,$E0,$77,$38,$47,$D1,$7B,$CF,$E4,$DC
	.dc.b	$BF,$36,$7B,$CF,$E6,$8E,$13,$E8,$3F,$24,$72,$3E,$51,$E1,$3C,$66
	.dc.b	$63,$F1,$4F,$13,$CA,$7F,$A4,$1F,$EC,$11,$F8,$87,$71,$E4,$3E,$69
	.dc.b	$CD,$FE,$C0,$E4,$F9,$47,$53,$C2,$53,$F0,$07,$E1,$8F,$F4,$41,$F8
	.dc.b	$63,$67,$33,$DA,$71,$8F,$A8,$FF,$A2,$E3,$F2,$CE,$67,$E5,$4F,$C7
	.dc.b	$7F,$2D,$1E,$79,$FE,$C8,$FF,$0C,$3C,$67,$DE,$3C,$33,$C5,$FC,$5C
	.dc.b	$7D,$29,$FD,$DC,$70,$3E,$71,$4E,$C7,$F6,$A3,$F8,$54,$F3,$7E,$29
	.dc.b	$ED,$3F,$48,$7A,$4F,$B2,$72,$3C,$CF,$C0,$70,$FF,$2E,$8F,$EB,$D1
	.dc.b	$F9,$13,$C4,$73,$9F,$B3,$3E,$18,$FD,$09,$F7,$CF,$EB,$CF,$DF,$8F
	.dc.b	$F0,$94,$71,$6B,$E7,$3F,$3A,$FC,$33,$FC,$B6,$7F,$7C,$7D,$B1,$DD
	.dc.b	$FC,$6A,$94,$C3,$FC,$8B,$3E,$99,$F7,$3E,$18,$E8,$FA,$23,$94,$71
	.dc.b	$3F,$2F,$3F,$C2,$9F,$7B,$F5,$A3,$9C,$FA,$4F,$B1,$3E,$93,$F9,$69
	.dc.b	$EE,$8F,$44,$7E,$0A,$79,$1F,$E6,$97,$E4,$1F,$D9,$63,$E4,$4F,$9A
	.dc.b	$3E,$D9,$DE,$3F,$2F,$1D,$CF,$F2,$51,$B8,$EC,$FC,$A7,$F1,$07,$F2
	.dc.b	$93,$F0,$33,$F9,$C3,$C6,$FE,$80,$EC,$F9,$67,$E2,$CF,$FA,$99,$F2
	.dc.b	$CF,$EB,$4F,$C2,$9C,$0F,$C7,$9F,$3C,$D3,$F2,$0F,$86,$FC,$83,$F8
	.dc.b	$D3,$83,$53,$E4,$9F,$A2,$7A,$E7,$FC,$46,$7E,$8A,$35,$1F,$3D,$AF
	.dc.b	$C3,$8E,$C7,$CE,$9F,$F0,$53,$FB,$53,$D2,$F8,$67,$E2,$4F,$78,$FD
	.dc.b	$23,$FE,$21,$8F,$C2,$3D,$CD,$38,$7F,$81,$C7,$C6,$9F,$9E,$7C,$D7
	.dc.b	$B3,$F1,$8F,$91,$1E,$E3,$F1,$27,$A6,$3E,$FC,$7C,$69,$F6,$BF,$31
	.dc.b	$F3,$CF,$08,$FB,$E7,$F2,$23,$C2,$F5,$3F,$0E,$F1,$3C,$53,$E4,$8D
	.dc.b	$47,$FA,$08,$FD,$54,$FE,$0C,$FD,$34,$7E,$9A,$32,$7B,$63,$CA,$F8
	.dc.b	$23,$F9,$DC,$F5,$7E,$41,$E7,$8D,$3F,$22,$7E,$54,$79,$63,$C6,$FF
	.dc.b	$88,$CF,$E6,$CE,$1F,$EC,$E7,$DB,$9F,$C4,$C7,$D4,$8F,$E8,$87,$F8
	.dc.b	$E6,$7C,$B3,$F5,$48,$FF,$87,$A3,$F1,$93,$C5,$E9,$1F,$3A,$3E,$4C
	.dc.b	$7E,$48,$FD,$64,$F2,$7E,$1C,$79,$63,$FC,$48,$FE,$DE,$3E,$61,$D0
	.dc.b	$F9,$0F,$77,$DA,$FC,$63,$BC,$75,$8E,$11,$FE,$CF,$7F,$9C,$CF,$C8
	.dc.b	$7E,$59,$F8,$23,$E0,$47,$57,$67,$99,$FA,$B1,$F1,$0E,$53,$DE,$7C
	.dc.b	$73,$F9,$39,$F8,$51,$EA,$3F,$AB,$CF,$E5,$9F,$D7,$47,$F4,$A7,$CF
	.dc.b	$3E,$B7,$81,$E2,$3F,$C8,$AF,$C7,$8F,$ED,$B1,$FC,$24,$F5,$9F,$3C
	.dc.b	$FC,$E4,$F9,$5E,$47,$8A,$3F,$2D,$3F,$CE,$E3,$FA,$9C,$7B,$08,$FC
	.dc.b	$18,$FC,$83,$4F,$C8,$3E,$77,$F2,$0F,$D9,$7F,$24,$78,$0F,$0B,$97
	.dc.b	$F2,$67,$8C,$FB,$AF,$E7,$E7,$F8,$5C,$7F,$9A,$E7,$C6,$FC,$69,$F9
	.dc.b	$CF,$ED,$23,$EA,$3E,$53,$E1,$CF,$27,$F2,$F3,$D6,$79,$C7,$86,$3F
	.dc.b	$0C,$FC,$B9,$F9,$0D,$9F,$D0,$3F,$9A,$78,$BE,$D3,$47,$EA,$A3,$FD
	.dc.b	$25,$3F,$8B,$8C,$9E,$93,$F2,$11,$A8,$F2,$47,$A6,$3F,$19,$1F,$1D
	.dc.b	$FF,$28,$C7,$F2,$28,$FA,$0F,$CA,$75,$3F,$84,$8D,$47,$C6,$3D,$AE
	.dc.b	$0F,$B2,$E4,$FC,$E9,$E7,$74,$7C,$A8,$F8,$2F,$E3,$1F,$82,$FB,$0F
	.dc.b	$95,$1C,$8F,$5B,$EF,$39,$1F,$16,$3F,$D5,$D1,$FB,$87,$9C,$78,$27
	.dc.b	$EC,$3F,$3C,$F8,$91,$A8,$FE,$4D,$3F,$94,$9E,$D1,$F6,$1E,$53,$E8
	.dc.b	$7F,$7B,$3E,$47,$CE,$FC,$13,$93,$83,$E2,$90,$FD,$61,$C9,$F4,$3E
	.dc.b	$42,$1F,$B6,$7E,$AC,$FD,$83,$CF,$1C,$DF,$6B,$F2,$CE,$27,$84,$F8
	.dc.b	$E7,$38,$F3,$3C,$67,$F3,$31,$F8,$08,$FF,$42,$9F,$4C,$FD,$79,$FD
	.dc.b	$8D,$FB,$93,$E2,$8E,$A7,$F1,$83,$A1,$E2,$38,$9F,$C5,$A3,$FB,$E3
	.dc.b	$93,$B1,$F8,$17,$C4,$FF,$94,$A7,$F6,$6F,$D4,$7E,$94,$FE,$7E,$7E
	.dc.b	$4C,$E4,$F8,$2F,$27,$FD,$22,$7D,$A3,$27,$AD,$FB,$4F,$EC,$4D,$1F
	.dc.b	$1C,$F5,$9D,$48,$7B,$CF,$82,$71,$8F,$88,$7D,$47,$EF,$1F,$CC,$8F
	.dc.b	$79,$D8,$E4,$FE,$3C,$E3,$3F,$52,$3C,$D3,$EE,$7E,$E4,$FE,$F1,$C9
	.dc.b	$D4,$E8,$7D,$07,$FC,$97,$1F,$E7,$68,$F4,$9E,$79,$F9,$AF,$89,$FD
	.dc.b	$21,$87,$93,$FE,$12,$7F,$9E,$9F,$64,$EE,$7D,$93,$C6,$7F,$38,$3D
	.dc.b	$F3,$EC,$7F,$71,$3F,$CD,$8F,$63,$E3,$7F,$BA,$1F,$65,$FF,$39,$C7
	.dc.b	$17,$E8,$C6,$5F,$F3,$8B,$F6,$9F,$11,$F8,$F3,$FA,$01,$E7,$8E,$4F
	.dc.b	$CA,$9F,$C5,$9F,$C2,$9F,$95,$1A,$7E,$0C,$FF,$5D,$7F,$D1,$13,$F1
	.dc.b	$A3,$F4,$6F,$CA,$7C,$91,$FE,$5E,$8F,$94,$F9,$27,$E9,$CF,$F8,$EE
	.dc.b	$3E,$99,$F1,$A7,$93,$F2,$8F,$EB,$A7,$C5,$8F,$E2,$07,$69,$FF,$18
	.dc.b	$BF,$06,$3F,$18,$E6,$7F,$C1,$31,$F4,$CC,$9D,$67,$F8,$64,$79,$A3
	.dc.b	$27,$C9,$3F,$AF,$CF,$D6,$7C,$CF,$D6,$79,$47,$F4,$79,$F9,$11,$EF
	.dc.b	$8F,$C4,$3F,$20,$FA,$4F,$78,$FA,$73,$E7,$9F,$C2,$1F,$4A,$7E,$1B
	.dc.b	$E3,$3F,$9D,$1F,$20,$FC,$89,$E8,$9F,$9A,$FE,$58,$D3,$E4,$75,$1E
	.dc.b	$48,$FF,$4A,$C7,$E0,$CF,$2C,$FB,$5F,$01,$FC,$B4,$E7,$3E,$93,$CB
	.dc.b	$3E,$08,$F7,$1A,$9F,$C2,$9C,$E3,$47,$82,$3E,$E4,$7C,$D7,$CA,$FE
	.dc.b	$3E,$35,$1F,$42,$3E,$F4,$7C,$88,$FA,$27,$F3,$83,$E1,$B9,$7D,$27
	.dc.b	$F1,$23,$F0,$4F,$C3,$34,$79,$27,$FC,$95,$1F,$0C,$F5,$4F,$E4,$1E
	.dc.b	$E7,$F3,$57,$DA,$65,$EF,$3F,$CD,$5F,$B5,$1A,$3F,$8D,$1F,$7C,$FF
	.dc.b	$29,$3E,$98,$F1,$CF,$D5,$3C,$71,$F1,$8E,$73,$DE,$7D,$0F,$F4,$C8
	.dc.b	$F3,$BF,$2E,$39,$BF,$2E,$79,$CF,$CB,$8D,$47,$DA,$3C,$0F,$38,$F8
	.dc.b	$44,$3F,$81,$BF,$EF,$58,$FE,$04,$FD,$47,$93,$DC,$E5,$1E,$93,$E8
	.dc.b	$C6,$A3,$94,$7E,$41,$E4,$7E,$1E,$3C,$71,$FC,$68,$F6,$9F,$0A,$3E
	.dc.b	$5B,$F5,$0F,$EB,$F3,$F7,$67,$B4,$FF,$41,$9F,$51,$F0,$4F,$F4,$63
	.dc.b	$F4,$CF,$F0,$7C,$7E,$EC,$E2,$FF,$17,$8E,$27,$E3,$A3,$2F,$EB,$CC
	.dc.b	$CF,$86,$7D,$47,$C3,$8F,$51,$A9,$FD,$54,$68,$F8,$71,$EE,$3E,$3B
	.dc.b	$93,$D4,$FC,$B3,$F9,$29,$FC,$3D,$EC,$FF,$18,$3E,$31,$0F,$F4,$69
	.dc.b	$F1,$CF,$E9,$B1,$E2,$8F,$C6,$3F,$42,$39,$1E,$28,$FA,$A6,$8E,$EF
	.dc.b	$F1,$43,$D8,$FE,$7D,$F9,$A7,$73,$FA,$BB,$F9,$03,$D8,$7C,$D7,$D9
	.dc.b	$3F,$CD,$23,$31,$F3,$4C,$CF,$D5,$8C,$C7,$F1,$03,$C2,$F9,$8F,$C2
	.dc.b	$BF,$B4,$8F,$EB,$CF,$AE,$3C,$47,$AC,$FA,$2F,$27,$27,$29,$FD,$7C
	.dc.b	$7D,$A3,$C6,$7E,$E6,$7F,$06,$F8,$CC,$9F,$C1,$E7,$B1,$F6,$9F,$C9
	.dc.b	$1E,$33,$E3,$9E,$77,$FA,$1B,$92,$1F,$E9,$73,$E6,$1F,$DB,$DF,$A0
	.dc.b	$FD,$E3,$93,$FC,$39,$F1,$3E,$C7,$BC,$FB,$1F,$BE,$7E,$30,$FC,$F4
	.dc.b	$7C,$E8,$FE,$67,$18,$8F,$9C,$F2,$9F,$C4,$3F,$83,$3E,$09,$FB,$47
	.dc.b	$E2,$CF,$23,$F4,$D3,$CA,$7D,$11,$E3,$79,$1F,$18,$F2,$3F,$C7,$23
	.dc.b	$EB,$C7,$63,$F3,$6F,$9E,$7F,$8F,$C7,$9E,$3E,$03,$DA,$3C,$EF,$F5
	.dc.b	$D3,$F0,$07,$E9,$CF,$9C,$7C,$A8,$FE,$0A,$FD,$F7,$E7,$BA,$9F,$43
	.dc.b	$CC,$F2,$1F,$A0,$3E,$34,$7E,$A1,$D1,$ED,$8F,$D0,$47,$E1,$8F,$D2
	.dc.b	$46,$67,$F4,$11,$98,$FD,$21,$93,$C8,$E5,$FC,$91,$93,$F0,$27,$89
	.dc.b	$F9,$2E,$4F,$AC,$FC,$93,$F9,$CB,$93,$E6,$BD,$67,$ED,$4F,$F0,$98
	.dc.b	$F9,$B3,$F0,$E3,$E6,$BF,$DC,$4F,$B2,$ED,$F1,$47,$E8,$DF,$AC,$7D
	.dc.b	$19,$4A,$21,$04,$69,$FE,$BC,$7F,$9B,$A7,$F4,$0F,$B1,$FC,$73,$E5
	.dc.b	$7E,$93,$E5,$8F,$A3,$1E,$A7,$A4,$77,$3B,$CF,$8E,$47,$FB,$DC,$E0
	.dc.b	$FD,$67,$F8,$43,$F0,$A7,$FC,$09,$3F,$BA,$7B,$BF,$0A,$7E,$FB,$F0
	.dc.b	$8D,$4F,$85,$F8,$53,$EF,$9F,$09,$FE,$F0,$8F,$C7,$9F,$C0,$DF,$AD
	.dc.b	$3F,$A2,$9F,$D7,$C7,$9A,$7A,$47,$38,$F5,$47,$D1,$9F,$53,$E5,$3F
	.dc.b	$1E,$79,$4F,$B7,$1F,$87,$7F,$92,$1F,$87,$3F,$40,$7E,$BD,$F9,$F3
	.dc.b	$F0,$A3,$8C,$7D,$48,$F9,$F1,$F7,$5F,$24,$7B,$63,$A1,$ED,$7E,$69
	.dc.b	$D2,$7F,$9E,$C7,$48,$F7,$3F,$66,$79,$BF,$2A,$7F,$D4,$33,$FC,$BA
	.dc.b	$38,$1F,$04,$F8,$B1,$F7,$E3,$E2,$C7,$D4,$3F,$5E,$F8,$8F,$D5,$BF
	.dc.b	$16,$7E,$EB,$F0,$63,$31,$F4,$A3,$B1,$E5,$9E,$D3,$EF,$8E,$51,$F4
	.dc.b	$9E,$51,$FA,$79,$E5,$1E,$33,$91,$FA,$23,$FD,$F4,$7F,$BD,$8F,$92
	.dc.b	$FC,$27,$F3,$67,$E9,$E3,$F8,$EB,$FC,$18,$E8,$7D,$39,$F2,$BF,$24
	.dc.b	$E1,$1F,$B4,$9F,$F7,$D4,$FE,$F6,$7C,$F3,$F9,$C9,$F3,$C7,$CE,$32
	.dc.b	$7E,$31,$C9,$F2,$1F,$DB,$47,$C8,$9E,$91,$C4,$CC,$7C,$38,$CC,$F1
	.dc.b	$3F,$17,$3E,$58,$F4,$BF,$00,$F8,$B1,$E8,$9F,$E7,$6F,$08,$F7,$47
	.dc.b	$EB,$9F,$80,$70,$9F,$F3,$0C,$7E,$3A,$3E,$31,$F4,$4F,$E0,$A7,$F7
	.dc.b	$69,$FF,$6B,$C6,$E3,$C8,$6D,$FF,$29,$18,$7F,$C8,$47,$B9,$F9,$53
	.dc.b	$EE,$3F,$09,$19,$3E,$53,$97,$F2,$A7,$EE,$4F,$E8,$67,$E3,$0C,$9F
	.dc.b	$48,$CC,$FD,$28,$EA,$FC,$38,$FC,$3C,$FC,$37,$E5,$1E,$E3,$FB,$33
	.dc.b	$E9,$7F,$B5,$9F,$8A,$3D,$EF,$18,$FD,$FB,$F0,$23,$E6,$9C,$0F,$61
	.dc.b	$E4,$9F,$63,$F9,$53,$33,$FD,$4A,$7E,$CC,$FA,$0F,$E5,$13,$E8,$8F
	.dc.b	$DC,$BF,$89,$9F,$C3,$4E,$9F,$D2,$1E,$48,$FB,$11,$F8,$88,$FC,$34
	.dc.b	$FF,$2C,$3F,$B2,$CF,$CB,$3C,$91,$F1,$CE,$73,$FC,$3A,$7F,$A9,$CF
	.dc.b	$F6,$A7,$F0,$4F,$E3,$9F,$2C,$F9,$23,$EA,$9E,$E9,$FE,$74,$7C,$27
	.dc.b	$FA,$E4,$FF,$11,$3F,$44,$7C,$27,$8C,$F8,$E7,$D4,$78,$E7,$F7,$F3
	.dc.b	$C8,$FE,$26,$7E,$69,$FA,$D1,$B7,$EE,$47,$F4,$E9,$FC,$DB,$F8,$D7
	.dc.b	$E9,$47,$F3,$B9,$FA,$53,$FC,$A8,$FC,$F1,$F3,$0F,$E6,$13,$EF,$7F
	.dc.b	$8C,$CF,$81,$FE,$1A,$7E,$10,$F0,$4F,$A9,$FE,$6B,$1F,$D4,$5F,$F2
	.dc.b	$CC,$F3,$7D,$87,$E7,$5F,$2C,$7D,$08,$FE,$94,$7A,$67,$E6,$C7,$C9
	.dc.b	$8F,$B7,$1F,$85,$3C,$0F,$F0,$13,$FD,$B1,$1C,$E7,$EF,$C7,$A0,$FE
	.dc.b	$60,$7E,$14,$F9,$91,$F7,$A7,$E2,$C7,$63,$E3,$47,$C1,$7B,$1F,$C0
	.dc.b	$A7,$D8,$FE,$AA,$39,$3F,$7A,$79,$BD,$A3,$FA,$54,$FE,$EE,$34,$76
	.dc.b	$8E,$2F,$84,$87,$13,$F6,$6F,$ED,$5F,$5C,$7F,$9C,$63,$F1,$E7,$DF
	.dc.b	$8E,$A7,$DA,$7A,$BF,$48,$FD,$84,$FC,$B7,$F0,$67,$F0,$83,$EF,$1F
	.dc.b	$D5,$9F,$9E,$FF,$51,$8F,$BC,$FF,$A1,$0F,$98,$7E,$AC,$FB,$4F,$ED
	.dc.b	$27,$E0,$47,$B0,$F9,$11,$F1,$27,$F5,$51,$C4,$FC,$33,$F8,$28,$F6
	.dc.b	$47,$13,$F6,$CF,$13,$F0,$CF,$B6,$7F,$5C,$FB,$1F,$F4,$1B,$FE,$4D
	.dc.b	$9F,$AA,$F7,$3F,$79,$3E,$78,$F1,$C7,$F8,$A0,$C9,$FE,$2C,$3D,$E6
	.dc.b	$67,$BC,$FE,$1E,$3F,$99,$47,$80,$FF,$17,$1F,$88,$78,$4F,$D1,$9E
	.dc.b	$07,$DD,$9F,$E3,$AF,$F0,$B3,$FA,$79,$EE,$9F,$44,$78,$23,$EE,$C6
	.dc.b	$27,$DA,$FC,$63,$F1,$27,$EF,$E3,$F9,$0B,$E2,$3F,$24,$FC,$B3,$E7
	.dc.b	$BF,$2E,$3C,$D1,$DA,$38,$1F,$4C,$FE,$78,$F1,$3F,$65,$1F,$88,$8F
	.dc.b	$F0,$04,$7E,$21,$FF,$57,$9F,$28,$FE,$02,$7E,$01,$FA,$47,$BD,$EE
	.dc.b	$7E,$8C,$FE,$BE,$FC,$78,$FF,$00,$C7,$E4,$0F,$AD,$3F,$90,$7D,$47
	.dc.b	$B8,$F8,$67,$E7,$CF,$D5,$C7,$E0,$1F,$59,$A8,$FB,$51,$FE,$5C,$7E
	.dc.b	$D4,$7A,$0F,$C2,$9F,$C8,$23,$E8,$B9,$8F,$AD,$3F,$16,$3F,$9E,$9F
	.dc.b	$5C,$F2,$1E,$C3,$EB,$3F,$D1,$1D,$1F,$BB,$79,$3F,$DE,$CF,$8A,$7E
	.dc.b	$B0,$F4,$4F,$FA,$7C,$F4,$1E,$49,$FE,$8F,$1F,$3C,$F8,$F1,$F0,$63
	.dc.b	$31,$EC,$3F,$7C,$FC,$B7,$C6,$7D,$37,$FB,$B3,$F2,$DF,$CC,$BE,$E3
	.dc.b	$F2,$47,$B4,$D4,$7D,$87,$D4,$7D,$03,$51,$F4,$27,$FA,$AC,$FA,$4F
	.dc.b	$C6,$4F,$FA,$26,$3F,$18,$FE,$F4,$FE,$42,$7E,$00,$F9,$07,$C4,$7F
	.dc.b	$D2,$C7,$47,$67,$B0,$F9,$47,$8A,$3E,$54,$6E,$3F,$41,$3F,$18,$FB
	.dc.b	$AF,$58,$FF,$33,$1F,$99,$7F,$05,$3F,$AA,$8F,$B3,$1F,$CC,$8F,$09
	.dc.b	$F1,$A7,$E5,$9F,$06,$3F,$6F,$1C,$8D,$9F,$1E,$37,$3E,$69,$DC,$79
	.dc.b	$63,$E4,$4F,$F2,$37,$53,$F8,$D8,$FC,$49,$0E,$91,$FE,$0F,$7F,$A1
	.dc.b	$9F,$B7,$9E,$8F,$E6,$1F,$0C,$FC,$C3,$C3,$3F,$5A,$78,$3F,$D2,$A3
	.dc.b	$F7,$11,$FD,$E0,$FD,$C1,$FA,$93,$DB,$3D,$9F,$B2,$7D,$38,$EC,$7A
	.dc.b	$5F,$88,$7D,$49,$EC,$FE,$AA,$3F,$C4,$EF,$F5,$B3,$E1,$3C,$8C,$BE
	.dc.b	$63,$EF,$1F,$9D,$8F,$84,$FF,$06,$8F,$B9,$3F,$E0,$E8,$FF,$02,$9F
	.dc.b	$E6,$77,$F8,$29,$97,$FA,$1C,$FD,$09,$EF,$3F,$AB,$9F,$A0,$EE,$3E
	.dc.b	$2C,$7F,$83,$E7,$F1,$0F,$E3,$63,$FB,$F9,$FC,$B9,$F3,$9F,$98,$9F
	.dc.b	$A7,$1F,$92,$3F,$9D,$C7,$E7,$4F,$C7,$4F,$B2,$7F,$6F,$1F,$B2,$7C
	.dc.b	$4F,$C9,$3F,$02,$F9,$DF,$C2,$9F,$E2,$D7,$8B,$F9,$B3,$EB,$BE,$29
	.dc.b	$F3,$C7,$F0,$33,$EB,$9F,$34,$E5,$3F,$C0,$DE,$46,$4F,$9A,$7F,$94
	.dc.b	$0F,$79,$EA,$38,$9F,$8B,$3D,$4F,$E0,$9F,$E1,$E6,$4E,$A7,$C7,$7E
	.dc.b	$11,$E1,$9F,$B2,$F1,$7F,$65,$3F,$8F,$7F,$29,$3F,$C3,$4F,$31,$F7
	.dc.b	$A3,$F1,$8F,$C8,$9F,$E5,$87,$E8,$E7,$FC,$71,$3F,$DF,$0F,$AB,$3F
	.dc.b	$82,$9F,$01,$F8,$77,$94,$7E,$42,$39,$07,$F9,$0C,$FF,$83,$A3,$F7
	.dc.b	$F1,$FB,$29,$FC,$FB,$F8,$33,$F1,$06,$4F,$F0,$8B,$F7,$CF,$CF,$3F
	.dc.b	$EA,$D7,$FA,$2B,$E1,$7E,$89,$F8,$23,$EA,$9F,$AD,$7A,$1F,$E0,$77
	.dc.b	$EF,$9F,$91,$38,$1F,$1D,$F2,$3F,$85,$9F,$9A,$FD,$53,$EE,$9F,$A3
	.dc.b	$3F,$82,$CF,$C6,$9F,$9F,$1E,$87,$DA,$78,$67,$9C,$7C,$D3,$E1,$C7
	.dc.b	$E5,$1F,$C3,$9E,$C8,$FD,$0B,$F8,$83,$C0,$7F,$0D,$38,$C7,$D9,$3B
	.dc.b	$9F,$BC,$9F,$A8,$FC,$98,$FE,$38,$71,$7E,$29,$FD,$0A,$7F,$1F,$1E
	.dc.b	$19,$F6,$4F,$C4,$8F,$2B,$C1,$FE,$04,$F8,$5F,$91,$1F,$84,$8F,$C1
	.dc.b	$BF,$E0,$08,$F8,$E7,$53,$D2,$7F,$69,$3E,$EC,$7E,$49,$FA,$E7,$EB
	.dc.b	$4F,$BA,$7F,$25,$7C,$4F,$F3,$37,$CC,$FE,$B0,$F9,$CF,$E3,$1F,$94
	.dc.b	$7F,$25,$8F,$CA,$C7,$D5,$7D,$0F,$86,$7E,$89,$FB,$07,$E6,$1F,$91
	.dc.b	$3C,$2F,$F0,$D7,$E1,$4F,$C6,$7F,$6D,$3F,$29,$FB,$D1,$E5,$3D,$A7
	.dc.b	$E4,$9F,$E5,$EF,$94,$FE,$B9,$1F,$C1,$1F,$E1,$13,$F5,$A7,$D8,$F9
	.dc.b	$63,$F6,$B1,$E9,$9F,$5B,$FE,$79,$8F,$9C,$7E,$08,$FE,$10,$7E,$8E
	.dc.b	$7D,$F3,$FB,$F8,$FC,$D1,$F7,$0F,$AE,$23,$FD,$D2,$E6,$7E,$1C,$F9
	.dc.b	$E3,$F2,$E7,$63,$ED,$1F,$AF,$8F,$87,$3E,$B7,$E3,$C7,$68,$EA,$FC
	.dc.b	$B8,$FA,$27,$E6,$DF,$F2,$1C,$FE,$7A,$11,$FA,$13,$07,$F8,$20,$CB
	.dc.b	$FA,$47,$64,$21,$1D,$38,$23,$87,$06,$1F,$E4,$4E,$0E,$4E,$1C,$18
	.dc.b	$6B,$E1,$7A,$38,$3D,$25,$72,$F0,$6B,$E2,$6B,$E1,$6B,$DC,$E8,$7C
	.dc.b	$07,$2F,$85,$F3,$3A,$72,$ED,$CB,$E3,$6B,$DD,$AF,$B9,$E2,$FD,$07
	.dc.b	$2E,$DF,$1B,$EE,$3C,$0E,$9F,$41,$A7,$9B,$B3,$C4,$E4,$E0,$FD,$37
	.dc.b	$07,$67,$C4,$E1,$C3,$D0,$EA,$E8,$F1,$BC,$5F,$01,$E9,$38,$BE,$07
	.dc.b	$E3,$3B,$3A,$19,$7E,$73,$E1,$61,$F6,$DD,$3C,$8F,$C2,$CF,$F8,$7C
	.dc.b	$FE,$92,$FF,$58,$FC,$13,$FD,$B7,$3F,$DC,$5F,$CD,$CA,$91,$F9,$98
	.dc.b	$C3,$FC,$C6,$7F,$6F,$59,$AF,$A5,$5F,$E8,$A8,$FD,$F4,$7E,$96,$BB
	.dc.b	$D7,$49,$FD,$7D,$74,$7F,$52,$F6,$8E,$B1,$FC,$08,$FE,$D4,$7E,$DA
	.dc.b	$79,$9E,$8A,$E9,$1F,$B3,$AE,$55,$FC,$10,$E7,$19,$9F,$39,$FE,$54
	.dc.b	$7C,$E7,$EF,$2B,$AB,$E6,$8E,$AF,$78,$F8,$53,$D4,$87,$8E,$39,$1F
	.dc.b	$EB,$D7,$E7,$3C,$E3,$E4,$1C,$A3,$F8,$A9,$FD,$CA,$3C,$07,$23,$E5
	.dc.b	$BF,$DD,$23,$C5,$1F,$22,$7E,$BC,$FB,$A3,$FC,$9C,$FF,$12,$AF,$B9
	.dc.b	$1C,$DF,$BD,$1C,$E3,$C8,$73,$AF,$23,$FD,$12,$3C,$8F,$C6,$3E,$91
	.dc.b	$ED,$7E,$8C,$FD,$B8,$F9,$D3,$C9,$FE,$86,$7F,$92,$CF,$E2,$B1,$E4
	.dc.b	$79,$9E,$E9,$F8,$EF,$E2,$A7,$E3,$C7,$B6,$3F,$C6,$CF,$FA,$51,$F6
	.dc.b	$57,$F3,$B9,$F8,$B1,$E2,$9E,$0F,$F8,$05,$E0,$F8,$2B,$E1,$C7,$81
	.dc.b	$F5,$47,$DF,$8E,$31,$FA,$BA,$FD,$F4,$FC,$33,$E0,$BE,$17,$E5,$9E
	.dc.b	$19,$FC,$2B,$F2,$CF,$EC,$0F,$89,$E2,$FD,$37,$E0,$1F,$4D,$F0,$C7
	.dc.b	$F9,$CA,$3E,$4B,$FA,$D7,$E9,$57,$C0,$7D,$13,$E0,$7E,$89,$E7,$3F
	.dc.b	$56,$78,$1F,$81,$3F,$9E,$8F,$CB,$47,$80,$FF,$0C,$1E,$79,$FD,$2C
	.dc.b	$79,$8F,$74,$7E,$69,$F6,$B5,$FC,$03,$F9,$39,$F4,$47,$DF,$8F,$93
	.dc.b	$3F,$79,$F4,$57,$FA,$D0,$FE,$18,$7E,$1E,$7D,$11,$C1,$F8,$CF,$0A
	.dc.b	$FF,$51,$1F,$BD,$9E,$15,$E8,$7E,$ED,$7B,$AB,$83,$FE,$26,$AE,$07
	.dc.b	$9A,$7F,$0B,$3D,$E7,$DE,$7D,$B3,$FC,$1C,$53,$E7,$1F,$E1,$8A,$F0
	.dc.b	$57,$CD,$8F,$CE,$4F,$CD,$3F,$C6,$0F,$DC,$3F,$36,$FE,$0E,$BC,$F1
	.dc.b	$C2,$3C,$EF,$F2,$53,$B1,$F8,$23,$F2,$27,$E0,$9F,$A4,$FF,$92,$63
	.dc.b	$FA,$19,$FE,$FD,$AF,$29,$F4,$67,$71,$FE,$2D,$6D,$7C,$4A,$F2,$C7
	.dc.b	$E2,$2B,$84,$7C,$1A,$E0,$DA,$E0,$FE,$BD,$FD,$33,$F8,$93,$CC,$6C
	.dc.b	$FF,$3F,$D7,$F3,$8A,$F2,$C7,$DE,$8F,$1B,$A7,$E1,$BF,$84,$8F,$D1
	.dc.b	$3F,$76,$36,$FF,$A6,$5D,$C7,$48,$FB,$67,$E9,$AB,$D2,$FE,$F6,$34
	.dc.b	$F9,$8D,$57,$FA,$0A,$BF,$A0,$47,$B5,$FB,$11,$F7,$EB,$FA,$9C,$78
	.dc.b	$A7,$FC,$B8,$7F,$9E,$A3,$E9,$3F,$64,$F8,$87,$A4,$F8,$87,$A6,$39
	.dc.b	$3F,$BA,$AF,$2B,$B3,$CA,$7C,$58,$F2,$BF,$14,$FC,$29,$B9,$FD,$75
	.dc.b	$7A,$E3,$FA,$89,$EB,$9F,$BC,$FD,$13,$C6,$7A,$A7,$F1,$73,$FC,$91
	.dc.b	$FD,$7C,$7A,$DF,$CC,$9F,$09,$E2,$7F,$A2,$63,$FD,$1E,$FF,$9C,$67
	.dc.b	$F8,$33,$F8,$E7,$F6,$0F,$D1,$9F,$3B,$ED,$3C,$E7,$B4,$F9,$27,$E9
	.dc.b	$E3,$E3,$BF,$EB,$C7,$F0,$2F,$96,$7E,$B1,$FD,$AA,$3F,$17,$1F,$DA
	.dc.b	$9F,$E4,$0E,$EB,$F3,$EF,$CE,$3E,$A1,$E2,$7E,$9B,$E1,$3D,$E7,$86
	.dc.b	$3C,$6F,$F8,$CC,$E8,$7C,$D3,$A1,$F3,$67,$E6,$1F,$C4,$CF,$D4,$D7
	.dc.b	$CC,$7D,$13,$D2,$3E,$9C,$FF,$71,$7E,$9B,$FC,$BE,$BF,$4A,$F9,$4E
	.dc.b	$E7,$D3,$3F,$89,$1F,$81,$9F,$A2,$FE,$1E,$BF,$96,$BF,$44,$FD,$31
	.dc.b	$EA,$3F,$57,$5F,$59,$F8,$A6,$8F,$11,$F9,$23,$C4,$ED,$F8,$07,$88
	.dc.b	$FB,$8F,$BE,$BD,$93,$E2,$9F,$F1,$11,$B8,$FA,$33,$F8,$F8,$F6,$9F
	.dc.b	$44,$FA,$C7,$9C,$FC,$64,$F3,$7C,$35,$F9,$27,$E1,$57,$C8,$AE,$87
	.dc.b	$9E,$3E,$44,$ED,$F3,$D7,$E3,$2B,$CF,$1F,$D3,$4F,$F7,$A9,$F9,$D7
	.dc.b	$A4,$7D,$A8,$FC,$39,$FC,$15,$FD,$2B,$FD,$AA,$7F,$26,$FC,$A3,$F3
	.dc.b	$F5,$E4,$9F,$54,$79,$27,$73,$FD,$6A,$BB,$9E,$49,$FE,$94,$7E,$4C
	.dc.b	$E9,$1F,$3E,$7F,$AB,$1F,$E4,$27,$F1,$67,$38,$FB,$4F,$85,$F7,$1F
	.dc.b	$80,$77,$5F,$98,$3E,$24,$F4,$9F,$F0,$39,$F5,$0F,$0C,$6A,$3F,$67
	.dc.b	$1F,$E4,$07,$F8,$DD,$7D,$E9,$F9,$13,$F7,$8F,$90,$7E,$0C,$FB,$C7
	.dc.b	$E6,$4F,$29,$F7,$A7,$E2,$C7,$96,$7F,$7B,$5F,$76,$BF,$10,$7E,$65
	.dc.b	$F8,$0F,$D0,$9F,$0C,$6E,$3C,$35,$F8,$3A,$F3,$1F,$3E,$3F,$D2,$51
	.dc.b	$FE,$2C,$3D,$6F,$98,$FD,$1D,$7D,$A7,$C3,$5F,$2A,$7F,$A6,$D7,$9A
	.dc.b	$B5,$5E,$09,$F5,$9A,$7F,$86,$57,$94,$FE,$0B,$1E,$58,$F0,$47,$D9
	.dc.b	$7F,$06,$EA,$39,$9F,$D0,$A3,$F7,$B3,$E0,$9D,$4F,$AA,$7E,$CD,$79
	.dc.b	$5F,$9E,$7E,$4E,$35,$3F,$28,$FC,$73,$F7,$A3,$FC,$EC,$EA,$BE,$A4
	.dc.b	$7A,$9F,$01,$FA,$47,$FC,$5E,$7E,$66,$7A,$47,$E2,$67,$F9,$23,$B8
	.dc.b	$FE,$FA,$7C,$B8,$F6,$4F,$F0,$23,$CE,$7D,$09,$FE,$61,$1F,$2E,$3E
	.dc.b	$09,$A3,$E8,$3E,$73,$4F,$33,$E5,$C7,$84,$FF,$11,$BF,$C1,$AB,$E3
	.dc.b	$3A,$3F,$0C,$FE,$22,$7F,$4D,$1F,$E4,$E3,$F2,$53,$F1,$1F,$19,$C6
	.dc.b	$3F,$C5,$0F,$DC,$8F,$CA,$9F,$21,$EA,$66,$7F,$28,$FF,$5F,$9F,$51
	.dc.b	$F1,$1F,$D4,$1D,$5F,$ED,$2F,$DE,$3E,$A1,$F0,$EB,$F0,$6F,$EE,$DF
	.dc.b	$A7,$3E,$E8,$FD,$0C,$FF,$98,$0F,$3B,$CC,$F5,$9A,$7F,$88,$52,$21
	.dc.b	$93,$FC,$11,$1F,$96,$3F,$AB,$BF,$AA,$3C,$91,$F4,$5F,$E6,$27,$F3
	.dc.b	$C3,$F8,$CB,$F1,$9D,$9F,$18,$F2,$1F,$6C,$F0,$1F,$4E,$BF,$D2,$8F
	.dc.b	$91,$FF,$43,$BE,$49,$F9,$67,$92,$3F,$5D,$1C,$8F,$F3,$A4,$F9,$5F
	.dc.b	$EB,$73,$FB,$57,$C2,$7D,$23,$84,$FE,$6A,$7E,$1C,$61,$FC,$6D,$7E
	.dc.b	$26,$BF,$59,$5F,$ED,$9A,$FE,$48,$74,$3E,$31,$F7,$63,$F4,$EE,$A3
	.dc.b	$EC,$BF,$78,$D1,$FD,$59,$FF,$2C,$BC,$DF,$6B,$F8,$B3,$55,$FE,$1A
	.dc.b	$7F,$B0,$1F,$42,$BF,$22,$FF,$3D,$3F,$11,$1F,$9D,$7E,$81,$CC,$FD
	.dc.b	$FC,$FC,$E7,$FB,$25,$7C,$EA,$FE,$BA,$FE,$A2,$3E,$69,$EB,$AE,$8F
	.dc.b	$CB,$3A,$57,$DA,$7F,$1C,$F5,$7B,$C7,$A5,$FC,$41,$FD,$06,$BD,$2F
	.dc.b	$F1,$83,$D3,$5F,$C0,$9E,$C7,$B4,$E6,$88,$FE,$1C,$E2,$FE,$C5,$FA
	.dc.b	$A7,$DD,$3F,$C6,$15,$E7,$7F,$DD,$47,$13,$F8,$2C,$78,$5E,$27,$BA
	.dc.b	$BF,$0E,$FE,$5D,$F1,$47,$8A,$3E,$59,$E2,$7F,$31,$1F,$A5,$8F,$C0
	.dc.b	$9E,$A8,$E0,$FC,$63,$E7,$3E,$19,$FC,$1C,$7C,$17,$D0,$7C,$D3,$FD
	.dc.b	$FB,$3F,$CB,$A7,$E0,$47,$09,$F9,$87,$78,$FC,$FC,$70,$8F,$C5,$9F
	.dc.b	$4E,$3F,$01,$1F,$A5,$8C,$BF,$D3,$A3,$F3,$A6,$A7,$F4,$F3,$FC,$8E
	.dc.b	$30,$FC,$E8,$F2,$47,$C2,$AF,$BF,$5F,$AB,$8F,$DB,$3F,$A6,$AF,$E3
	.dc.b	$27,$CF,$3E,$79,$F5,$8F,$9A,$7D,$68,$F7,$D7,$CD,$8F,$BC,$7F,$8A
	.dc.b	$5E,$2F,$13,$E7,$47,$82,$3F,$37,$1F,$78,$F9,$CF,$09,$F9,$CF,$BA
	.dc.b	$3F,$6C,$F7,$7C,$C7,$E1,$23,$6F,$F2,$63,$FA,$B1,$FC,$04,$F2,$9F
	.dc.b	$E2,$99,$F9,$67,$F3,$1A,$FD,$8B,$FC,$42,$BF,$1A,$FF,$3A,$9F,$A7
	.dc.b	$1F,$DD,$5F,$E0,$A7,$CC,$76,$FE,$6E,$BF,$BB,$4F,$D1,$3E,$AC,$7C
	.dc.b	$53,$9D,$7F,$4D,$7D,$67,$98,$FF,$73,$1F,$76,$3C,$87,$F8,$C5,$FC
	.dc.b	$F4,$7A,$23,$D6,$48,$F2,$BF,$3C,$FC,$E5,$7C,$98,$F9,$E7,$DE,$AF
	.dc.b	$92,$FF,$3A,$9F,$88,$FF,$1F,$7D,$2F,$C9,$3C,$26,$5F,$29,$FD,$94
	.dc.b	$E2,$FE,$75,$C9,$F2,$DF,$E2,$71,$F9,$D7,$CF,$1A,$7E,$54,$FE,$6A
	.dc.b	$BD,$D1,$A9,$F7,$3F,$AD,$36,$7C,$C3,$F0,$07,$CC,$3F,$84,$3F,$C2
	.dc.b	$1D,$9E,$29,$FB,$8F,$F9,$CE,$BF,$C0,$6F,$E7,$E7,$ED,$1E,$57,$FB
	.dc.b	$54,$F9,$5F,$85,$1E,$49,$FF,$2C,$4F,$F3,$19,$FC,$99,$FA,$E3,$E8
	.dc.b	$9F,$99,$7F,$12,$FE,$BA,$3F,$13,$5F,$02,$3C,$93,$FD,$BD,$FE,$16
	.dc.b	$F5,$8F,$ED,$15,$E1,$8F,$F1,$43,$F8,$13,$F6,$F1,$F9,$77,$F1,$53
	.dc.b	$ED,$8E,$D5,$E8,$8F,$F0,$53,$FC,$15,$FE,$B6,$7A,$0F,$CB,$1F,$D4
	.dc.b	$2B,$C4,$FE,$5E,$BC,$4F,$F8,$51,$F1,$1F,$28,$CB,$FC,$D2,$7D,$EF
	.dc.b	$CA,$3E,$F1,$0F,$BC,$7C,$59,$FA,$0F,$E3,$67,$9B,$E4,$3F,$CB,$A7
	.dc.b	$E5,$8D,$1C,$E7,$DD,$1E,$4A,$F8,$35,$F2,$23,$FC,$8F,$3F,$06,$BF
	.dc.b	$B1,$9F,$CF,$1F,$E8,$73,$F6,$23,$F8,$EC,$FE,$82,$7F,$39,$1F,$60
	.dc.b	$F9,$73,$FC,$82,$3F,$A3,$1F,$CC,$E7,$BC,$64,$FD,$B4,$F7,$7F,$92
	.dc.b	$1D,$E7,$FA,$54,$7E,$66,$BF,$85,$3F,$54,$FC,$FD,$7F,$5E,$9D,$C7
	.dc.b	$E4,$8F,$B6,$75,$3F,$81,$57,$53,$F9,$84,$7C,$18,$FE,$6C,$6C,$FC
	.dc.b	$C4,$FE,$AD,$FA,$87,$EC,$4F,$CF,$1B,$3E,$A4,$7F,$8C,$0E,$87,$C2
	.dc.b	$9F,$04,$7F,$59,$3C,$73,$D5,$F9,$AF,$F8,$9D,$FC,$31,$F3,$5F,$F3
	.dc.b	$9C,$F4,$9F,$92,$FC,$D8,$FC,$2D,$7E,$9A,$BF,$05,$5F,$7E,$7A,$47
	.dc.b	$E1,$8E,$8F,$D4,$3F,$94,$CF,$EA,$5F,$E3,$C7,$0A,$FB,$E7,$E3,$1E
	.dc.b	$07,$EC,$0F,$E5,$33,$FE,$CB,$3F,$29,$1C,$2B,$C5,$1E,$43,$E5,$1F
	.dc.b	$5A,$3D,$4E,$4F,$C0,$1F,$C7,$8F,$C0,$1F,$B3,$9F,$91,$1F,$80,$AF
	.dc.b	$81,$3E,$C9,$F8,$6F,$ED,$CE,$87,$D6,$AF,$E1,$CF,$F5,$23,$FB,$8C
	.dc.b	$FC,$19,$FE,$75,$1F,$8B,$3F,$4B,$5E,$67,$F6,$E7,$D8,$3D,$F5,$FC
	.dc.b	$BA,$3E,$34,$F8,$A7,$AB,$F1,$9F,$F6,$D1,$F0,$8E,$B3,$F8,$67,$FB
	.dc.b	$51,$F9,$28,$FC,$F4,$7E,$60,$FD,$55,$79,$E7,$F9,$F9,$ED,$3F,$C7
	.dc.b	$F3,$F5,$4F,$E7,$8F,$F9,$C1,$FE,$08,$FE,$3A,$BC,$EF,$C6,$9F,$97
	.dc.b	$18,$8F,$7B,$F3,$DF,$98,$7C,$F3,$81,$87,$81,$FC,$2A,$7E,$BD,$7A
	.dc.b	$5F,$C7,$9F,$0E,$7F,$8D,$9F,$CD,$CF,$24,$7E,$21,$F0,$BC,$63,$D8
	.dc.b	$7F,$24,$8F,$F0,$6D,$7A,$27,$85,$78,$6B,$E4,$D7,$F8,$A6,$B4,$FF
	.dc.b	$39,$8F,$CD,$9F,$CE,$5F,$C3,$9D,$0F,$65,$74,$3F,$8B,$C7,$B2,$3A
	.dc.b	$1A,$3A,$3F,$81,$AF,$DC,$57,$D7,$9F,$ED,$A7,$E0,$4F,$E2,$31,$F8
	.dc.b	$D3,$D1,$3F,$D8,$A3,$A9,$F8,$A7,$ED,$C7,$F9,$AE,$3F,$4E,$7F,$14
	.dc.b	$79,$57,$9A,$7E,$0C,$FF,$88,$5F,$DE,$47,$E5,$CF,$64,$7F,$14,$8D
	.dc.b	$BE,$17,$F0,$EF,$F0,$99,$FB,$B3,$D6,$7C,$D3,$D4,$FD,$1D,$7F,$38
	.dc.b	$7F,$93,$1F,$CE,$23,$6F,$57,$FA,$64,$F8,$E3,$E4,$9F,$8C,$3E,$04
	.dc.b	$7D,$F7,$FB,$B4,$FD,$A7,$FB,$74,$F8,$DF,$D1,$C6,$5F,$D2,$4F,$E3
	.dc.b	$EB,$F9,$B5,$7E,$AA,$30,$FF,$04,$9F,$F0,$01,$FA,$A8,$FC,$23,$F9
	.dc.b	$F3,$F2,$67,$D4,$9F,$BB,$1C,$A3,$FB,$1B,$F9,$E3,$EB,$CF,$F5,$E8
	.dc.b	$CC,$7E,$69,$FC,$79,$EC,$3F,$14,$7B,$0F,$A8,$FC,$87,$F8,$C4,$F8
	.dc.b	$23,$AC,$7F,$4D,$3F,$06,$7F,$2D,$9F,$E1,$8F,$F9,$74,$F4,$9F,$5C
	.dc.b	$FD,$81,$F7,$5F,$C1,$BF,$C4,$63,$C0,$FC,$47,$F0,$73,$FC,$F6,$3E
	.dc.b	$B9,$D0,$FB,$91,$F8,$EA,$FC,$14,$FE,$52,$7D,$67,$D6,$3D,$E7,$D6
	.dc.b	$9F,$5B,$E5,$8F,$F1,$4B,$E5,$3B,$BF,$A6,$79,$1F,$86,$3E,$59,$FD
	.dc.b	$D6,$7F,$04,$FF,$99,$CF,$E9,$53,$F9,$E7,$E5,$D7,$E0,$E3,$C3,$1F
	.dc.b	$1E,$BC,$27,$BD,$FA,$87,$C5,$3C,$2F,$D3,$8F,$CB,$47,$84,$F6,$9E
	.dc.b	$13,$FD,$91,$1F,$CA,$DF,$41,$F8,$D7,$D8,$FF,$84,$0F,$B6,$7F,$46
	.dc.b	$8F,$F2,$EC,$6A,$3F,$A9,$1F,$69,$F3,$3F,$2E,$3C,$71,$FE,$A2,$8F
	.dc.b	$AA,$7E,$D8,$FE,$DC,$7E,$D8,$F8,$B1,$F8,$05,$1F,$A8,$8F,$CE,$1D
	.dc.b	$27,$F1,$B3,$FE,$01,$3C,$4F,$EE,$DF,$8E,$78,$CF,$01,$FC,$5D,$E2
	.dc.b	$FF,$64,$3E,$61,$F0,$4F,$EE,$47,$F2,$67,$6F,$F4,$89,$F8,$47,$CB
	.dc.b	$3E,$93,$F8,$27,$F2,$0F,$CA,$39,$4F,$CA,$9F,$E1,$11,$DE,$BF,$0E
	.dc.b	$FF,$84,$0F,$AF,$1F,$D4,$67,$F8,$1D,$7F,$22,$9F,$CD,$9C,$4F,$BE
	.dc.b	$71,$7E,$83,$F8,$E8,$F9,$2F,$96,$7D,$73,$FE,$B4,$3F,$1E,$FC,$98
	.dc.b	$F9,$07,$E1,$5F,$F4,$2C,$7D,$E8,$FC,$63,$D9,$FD,$B4,$FF,$2C,$AF
	.dc.b	$D8,$D7,$63,$F8,$F4,$FC,$97,$99,$DD,$E6,$ED,$FE,$45,$1F,$E8,$A7
	.dc.b	$F6,$31,$F2,$1F,$C8,$D7,$A2,$7B,$D7,$CD,$9F,$8A,$6A,$BB,$CE,$A7
	.dc.b	$FC,$9F,$3F,$CF,$EB,$EC,$BE,$F3,$8B,$F7,$6B,$F6,$D3,$F9,$13,$B4
	.dc.b	$FD,$D9,$FF,$38,$46,$23,$E4,$C7,$95,$FC,$FB,$F8,$47,$33,$E4,$72
	.dc.b	$FC,$19,$FE,$78,$7F,$20,$7E,$34,$7D,$D7,$51,$F3,$E3,$8B,$E1,$7F
	.dc.b	$B0,$1F,$C1,$A3,$D2,$FE,$06,$3C,$D5,$F9,$08,$FD,$91,$F9,$03,$FA
	.dc.b	$74,$F0,$7E,$21,$FA,$63,$E6,$1F,$E0,$BA,$F8,$35,$F3,$23,$F9,$A1
	.dc.b	$F9,$17,$FA,$83,$C8,$F8,$27,$19,$FF,$46,$9E,$C8,$F9,$C7,$F2,$78
	.dc.b	$FC,$6B,$FC,$F5,$F7,$BD,$E7,$F6,$73,$FA,$7A,$FC,$45,$7F,$02,$3E
	.dc.b	$19,$FA,$73,$F9,$0C,$FE,$AD,$F6,$3F,$7E,$BF,$4F,$1E,$73,$89,$F0
	.dc.b	$0F,$C0,$57,$A2,$3E,$84,$7F,$1A,$3E,$83,$CA,$7B,$47,$C0,$38,$D7
	.dc.b	$F8,$D5,$FF,$33,$4F,$DB,$7F,$8B,$3D,$1F,$A8,$7F,$55,$AF,$C4,$3A
	.dc.b	$8F,$85,$1F,$A5,$AF,$54,$7D,$07,$C3,$3E,$A7,$31,$FC,$32,$7F,$B8
	.dc.b	$CF,$94,$FE,$43,$3F,$72,$7F,$D2,$B1,$C6,$7E,$73,$FE,$68,$3F,$6A
	.dc.b	$7E,$A1,$CB,$FD,$A4,$FA,$47,$C5,$3E,$94,$FE,$C0,$F2,$B9,$3E,$71
	.dc.b	$E2,$8F,$9D,$3E,$E8,$EC,$7E,$59,$F9,$D1,$EA,$8F,$31,$98,$F8,$51
	.dc.b	$98,$ED,$3F,$0A,$BF,$0F,$1F,$04,$FF,$39,$1F,$86,$9E,$0F,$D2,$3F
	.dc.b	$AD,$9F,$D4,$23,$F8,$65,$7C,$68,$EB,$3F,$0A,$BF,$94,$47,$EE,$67
	.dc.b	$FC,$F4,$FD,$B3,$F4,$55,$F8,$39,$FF,$37,$47,$C1,$8C,$57,$F8,$EE
	.dc.b	$BF,$DE,$AF,$ED,$6B,$E4,$D7,$94,$FD,$84,$FD,$D7,$EA,$9F,$9E,$9F
	.dc.b	$21,$FC,$31,$FE,$03,$3F,$CE,$DF,$EE,$30,$7F,$99,$9F,$7C,$FB,$67
	.dc.b	$73,$F6,$A7,$F6,$E3,$FC,$85,$3F,$99,$9E,$87,$F7,$B7,$E9,$C7,$90
	.dc.b	$FC,$85,$7F,$13,$9F,$E4,$D1,$F6,$8F,$8C,$7E,$A2,$7F,$92,$9F,$EA
	.dc.b	$A9,$FB,$31,$FD,$2A,$3E,$6B,$F5,$A7,$F9,$93,$FA,$78,$FA,$CF,$A4
	.dc.b	$F7,$9E,$E3,$F3,$07,$E0,$CF,$F3,$31,$F8,$33,$D0,$79,$4E,$D5,$FB
	.dc.b	$47,$E0,$C7,$E5,$DF,$A0,$7E,$4D,$FA,$8F,$CA,$9F,$8B,$3F,$3E,$7F
	.dc.b	$A0,$57,$E3,$2B,$FD,$DE,$7E,$52,$BF,$51,$5F,$94,$78,$C7,$A5,$FC
	.dc.b	$7B,$F7,$23,$F9,$63,$F5,$4F,$F1,$E5,$79,$5F,$A2,$7F,$82,$8E,$0F
	.dc.b	$CA,$7F,$17,$5F,$39,$FC,$D4,$7F,$35,$7E,$4B,$E8,$7F,$9A,$BF,$3A
	.dc.b	$BC,$AE,$63,$E9,$1E,$C3,$E6,$C7,$B1,$F9,$A7,$EA,$8F,$91,$3E,$17
	.dc.b	$CB,$3E,$19,$FC,$63,$F3,$23,$EC,$1F,$E6,$33,$EC,$1F,$3E,$BE,$1B
	.dc.b	$FB,$93,$CE,$78,$CE,$13,$E3,$78,$4F,$C3,$7F,$3E,$71,$3D,$67,$FB
	.dc.b	$B8,$FB,$E7,$E6,$0F,$DD,$3F,$98,$3F,$DE,$D3,$FB,$93,$ED,$1E,$49
	.dc.b	$FC,$BC,$FE,$2A,$7E,$54,$E4,$FC,$74,$7D,$A8,$C9,$E5,$7E,$9B,$FD
	.dc.b	$2D,$E4,$7B,$67,$FC,$25,$5F,$AF,$AE,$93,$FB,$38,$C5,$7E,$E6,$73
	.dc.b	$3F,$B1,$8F,$F4,$BB,$FE,$71,$AF,$3B,$E3,$3B,$4F,$07,$F1,$A7,$D2
	.dc.b	$3E,$AC,$78,$9F,$C8,$3F,$C0,$1F,$C0,$1F,$26,$BC,$75,$F9,$78,$FD
	.dc.b	$81,$F9,$43,$F8,$A1,$D2,$BC,$0E,$8F,$AF,$5E,$67,$FC,$E7,$3C,$DE
	.dc.b	$E7,$EE,$8F,$E6,$6F,$D7,$7F,$C8,$D3,$F5,$DF,$CA,$9F,$C2,$67,$F2
	.dc.b	$35,$F9,$57,$D5,$19,$9F,$DE,$57,$D3,$8F,$8C,$7E,$1C,$F8,$B3,$F8
	.dc.b	$F7,$D5,$1F,$E3,$67,$E2,$CF,$B1,$FD,$3B,$E3,$7F,$2B,$1F,$A3,$8E
	.dc.b	$EF,$F8,$9E,$3F,$2A,$7F,$83,$9F,$C8,$9F,$DA,$1F,$5C,$FE,$DA,$3C
	.dc.b	$D1,$F1,$9F,$C6,$9D,$5F,$E1,$D3,$FD,$26,$7D,$E7,$19,$F0,$4F,$E4
	.dc.b	$0F,$71,$F4,$CF,$CA,$BA,$7E,$19,$E5,$3E,$7B,$E5,$7D,$AF,$FB,$16
	.dc.b	$3E,$7C,$7E,$00,$FE,$81,$1E,$53,$F8,$73,$E1,$3A,$1F,$E5,$43,$F0
	.dc.b	$73,$FA,$A7,$DC,$7E,$62,$3E,$83,$F9,$73,$AB,$E1,$8F,$9B,$1F,$CA
	.dc.b	$CE,$27,$B8,$FC,$59,$F9,$53,$FA,$59,$C8,$FF,$22,$C7,$F6,$27,$DA
	.dc.b	$61,$FE,$34,$EC,$FF,$13,$C7,$DF,$7F,$0B,$1F,$CE,$6B,$F4,$AF,$D6
	.dc.b	$3C,$D1,$F5,$A3,$FA,$53,$E5,$3F,$28,$64,$FE,$42,$66,$BE,$D3,$ED
	.dc.b	$7F,$C8,$4F,$C5,$9F,$03,$F9,$47,$F8,$31,$FC,$46,$3E,$AC,$7C,$03
	.dc.b	$FA,$3C,$79,$0F,$AA,$FE,$02,$3F,$C3,$EF,$E1,$5E,$27,$F5,$A3,$F1
	.dc.b	$B5,$FC,$2E,$BD,$F3,$E3,$AE,$87,$F8,$0D,$FC,$21,$E0,$3E,$64,$7E
	.dc.b	$D6,$7F,$55,$1F,$CF,$9F,$98,$FC,$F3,$FC,$A3,$1D,$9F,$89,$1E,$17
	.dc.b	$E0,$4F,$F0,$97,$E5,$BE,$C7,$F0,$67,$E4,$2B,$FD,$2A,$FF,$5A,$3C
	.dc.b	$A7,$F5,$A8,$F4,$9F,$A3,$78,$C7,$C4,$7F,$16,$7E,$60,$FE,$2B,$1F
	.dc.b	$CC,$9C,$1F,$D4,$0F,$E3,$47,$F3,$77,$FD,$56,$76,$7F,$5A,$FE,$45
	.dc.b	$FC,$5B,$97,$F4,$C7,$ED,$8F,$C1,$4F,$F9,$CA,$3F,$1D,$1F,$30,$F5
	.dc.b	$CF,$D6,$8F,$C1,$4F,$E9,$A7,$FB,$31,$FB,$39,$F6,$3F,$22,$7F,$87
	.dc.b	$C2,$55,$F9,$5A,$FE,$33,$3F,$E6,$48,$FF,$2A,$47,$F4,$B7,$EB,$3F
	.dc.b	$3E,$7F,$79,$3C,$E7,$94,$F8,$67,$FC,$D5,$5F,$87,$8F,$F1,$FC,$F2
	.dc.b	$AE,$D1,$FE,$61,$7F,$9A,$46,$6B,$8D,$7F,$2A,$3E,$54,$66,$BC,$A7
	.dc.b	$F6,$59,$F1,$9E,$C9,$FC,$35,$7E,$B2,$7F,$A1,$57,$B2,$3B,$C7,$E3
	.dc.b	$A7,$EA,$47,$E2,$A3,$FB,$21,$E5,$AF,$E0,$95,$E1,$8F,$F0,$81,$FA
	.dc.b	$B8,$FA,$53,$FD,$84,$F8,$95,$E2,$8F,$B9,$3F,$AB,$3E,$E1,$F1,$0F
	.dc.b	$11,$93,$F8,$9C,$7F,$35,$7F,$C3,$AF,$E4,$DF,$C4,$1F,$80,$3C,$D5
	.dc.b	$F4,$9F,$E0,$13,$F2,$9F,$C9,$CF,$F8,$E6,$7D,$51,$F9,$A3,$FB,$2C
	.dc.b	$FB,$9F,$19,$F2,$0F,$85,$3E,$39,$E3,$1F,$44,$FD,$3C,$7E,$B6,$BE
	.dc.b	$84,$7E,$1C,$C9,$F9,$E9,$FF,$57,$BF,$B7,$3E,$11,$FD,$0C,$FE,$E3
	.dc.b	$3D,$DF,$19,$CC,$FD,$81,$CD,$FF,$05,$3F,$0C,$FF,$56,$C7,$C3,$7F
	.dc.b	$7F,$5F,$0D,$FD,$8D,$7F,$22,$8F,$D1,$C7,$E3,$0E,$35,$F6,$6B,$CF
	.dc.b	$1F,$84,$3F,$48,$79,$2B,$F3,$6E,$0F,$51,$FC,$86,$7D,$4E,$23,$FC
	.dc.b	$68,$ED,$FF,$0A,$3F,$E9,$1A,$FA,$4E,$DF,$E5,$71,$FD,$22,$3D,$CF
	.dc.b	$E3,$A7,$F7,$B3,$F8,$E9,$FB,$07,$57,$F9,$8B,$F3,$23,$F5,$6F,$A6
	.dc.b	$3E,$24,$FE,$86,$3F,$09,$5F,$A7,$9F,$1C,$F9,$A7,$95,$7E,$2D,$E4
	.dc.b	$7C,$93,$F2,$B1,$CD,$FE,$EC,$FF,$07,$7E,$73,$E9,$8F,$C4,$4F,$A6
	.dc.b	$7E,$F1,$FA,$08,$CD,$7D,$28,$F2,$9F,$94,$8F,$5B,$F1,$DF,$F0,$6D
	.dc.b	$7E,$FA,$31,$1F,$DB,$1F,$88,$FE,$B5,$FA,$0F,$D6,$8F,$74,$7D,$98
	.dc.b	$EA,$FF,$90,$27,$F0,$C7,$BD,$FF,$7D,$28,$86,$63,$F1,$A7,$F2,$C7
	.dc.b	$31,$CA,$39,$3C,$23,$83,$F9,$37,$71,$FC,$82,$33,$1D,$A3,$89,$FC
	.dc.b	$91,$DB,$FC,$EA,$37,$1E,$58,$D9,$E3,$7D,$11,$E3,$8F,$B4,$F8,$DD
	.dc.b	$9F,$04,$F1,$1E,$88,$F1,$3C,$0F,$13,$D1,$D9,$D1,$DB,$DC,$E8,$79
	.dc.b	$CE,$8F,$9D,$E8,$79,$DF,$96,$FD,$03,$6F,$C8,$7C,$2E,$98,$7C,$37
	.dc.b	$C2,$F9,$0F,$09,$CC,$F3,$3E,$17,$CC,$7D,$58,$AF,$38,$D3,$FD,$4A
	.dc.b	$3F,$88,$4F,$E7,$63,$F8,$B4,$FA,$E3,$FB,$94,$7F,$53,$3F,$35,$08
	.dc.b	$FD,$81,$FA,$B3,$FB,$C1,$D2,$BF,$AF,$47,$F0,$0A,$C1,$F5,$EB,$94
	.dc.b	$FF,$66,$AF,$DD,$57,$E5,$1E,$B3,$FB,$29,$F0,$4F,$F1,$F2,$1D,$4F
	.dc.b	$BB,$3E,$89,$FE,$F0,$7A,$1E,$6F,$79,$E7,$3C,$48,$7E,$E8,$E6,$F9
	.dc.b	$63,$91,$FD,$BE,$7E,$04,$72,$9E,$07,$29,$F9,$6F,$F1,$57,$E5,$1F
	.dc.b	$01,$F7,$3C,$CF,$74,$FC,$68,$F6,$9D,$0E,$51,$EC,$39,$3F,$BB,$8F
	.dc.b	$6C,$7D,$48,$E2,$F8,$88,$7D,$03,$C7,$1F,$E2,$F8,$F1,$C7,$D4,$8F
	.dc.b	$69,$E3,$8F,$64,$7F,$3C,$9F,$BB,$1E,$48,$FC,$21,$F1,$8E,$47,$B1
	.dc.b	$F1,$BC,$6B,$A4,$71,$AF,$C0,$BE,$B9,$F8,$87,$D7,$8D,$BE,$A7,$FC
	.dc.b	$35,$5E,$28,$FF,$11,$1E,$C7,$F8,$B4,$7C,$48,$FB,$13,$E2,$8E,$35
	.dc.b	$D2,$7E,$21,$F3,$63,$E2,$4F,$8E,$3D,$67,$8D,$E5,$5F,$AB,$6C,$78
	.dc.b	$5F,$97,$3F,$A4,$8F,$85,$1F,$2A,$3C,$2F,$08,$F0,$4F,$A1,$F8,$27
	.dc.b	$DB,$8E,$11,$DD,$B1,$DD,$F3,$4F,$78,$F3,$46,$9F,$F0,$99,$F4,$5F
	.dc.b	$CE,$BC,$5F,$B2,$FB,$8F,$03,$F9,$17,$E8,$4F,$86,$3E,$C4,$7D,$F9
	.dc.b	$F0,$3C,$0D,$9F,$E2,$C3,$C1,$1F,$C3,$67,$F5,$91,$F9,$A8,$F0,$CF
	.dc.b	$28,$FA,$F1,$E9,$7F,$B7,$1D,$0F,$8E,$FC,$E9,$F4,$C7,$84,$EF,$1E
	.dc.b	$96,$9F,$86,$38,$BF,$22,$38,$C7,$82,$38,$4F,$D6,$7C,$4F,$E5,$9F
	.dc.b	$3C,$F0,$9F,$D4,$47,$DB,$7C,$E7,$80,$FC,$69,$C1,$F4,$4F,$EB,$9E
	.dc.b	$15,$D4,$FF,$0E,$1F,$C3,$8F,$CB,$47,$F1,$D9,$E5,$3E,$98,$E4,$FC
	.dc.b	$79,$E5,$3F,$E7,$03,$F9,$53,$CA,$BE,$8D,$7E,$7A,$BD,$6F,$27,$E0
	.dc.b	$C7,$DF,$AF,$A0,$78,$23,$6F,$E7,$1D,$BF,$7E,$7E,$44,$7E,$1A,$3E
	.dc.b	$F4,$7E,$CD,$F8,$F1,$E7,$7F,$A6,$D7,$C0,$7C,$F1,$F2,$E7,$E8,$1C
	.dc.b	$4F,$E9,$EF,$C8,$8E,$27,$BC,$FE,$7D,$3E,$F7,$F3,$D3,$C6,$BF,$9C
	.dc.b	$47,$B2,$7E,$3B,$FE,$7E,$9F,$89,$1E,$38,$E7,$5F,$A8,$7F,$12,$FC
	.dc.b	$27,$F5,$EF,$C2,$7F,$5F,$3F,$09,$FC,$81,$F2,$5F,$11,$E5,$77,$5F
	.dc.b	$3D,$FC,$3C,$7C,$07,$F8,$19,$F9,$53,$D0,$7C,$03,$D1,$1F,$E0,$17
	.dc.b	$E1,$1F,$8A,$9F,$43,$E1,$39,$9F,$49,$F4,$3F,$98,$9F,$43,$E2,$3F
	.dc.b	$7D,$3F,$9F,$8F,$A3,$3F,$66,$7F,$CA,$EF,$F8,$AD,$FF,$4E,$9F,$69
	.dc.b	$FE,$00,$FD,$23,$C3,$1C,$A3,$C3,$5F,$4A,$3F,$CE,$53,$F4,$27,$F6
	.dc.b	$2F,$8A,$3F,$2C,$FC,$23,$C3,$1F,$E5,$A8,$E3,$1F,$06,$3C,$2F,$D8
	.dc.b	$8F,$A3,$1F,$89,$8E,$33,$F7,$DE,$11,$F9,$99,$FD,$31,$C1,$F8,$0F
	.dc.b	$F9,$E4,$FC,$3C,$7D,$03,$FD,$BF,$1F,$42,$3E,$D4,$7F,$2D,$3E,$D1
	.dc.b	$FC,$15,$F9,$E4,$AF,$DA,$9F,$D0,$E3,$81,$F5,$CF,$81,$1E,$18,$FC
	.dc.b	$04,$7D,$03,$E4,$4F,$A6,$39,$4F,$F2,$D9,$E4,$7D,$D3,$F6,$A7,$D4
	.dc.b	$8F,$D2,$BE,$57,$F2,$91,$FC,$44,$F5,$9C,$CF,$B5,$1F,$9E,$73,$1F
	.dc.b	$87,$39,$BE,$D8,$FE,$0D,$1E,$E9,$FE,$47,$1F,$C2,$CE,$47,$DF,$30
	.dc.b	$78,$CE,$27,$E2,$CF,$8B,$3E,$27,$D8,$FD,$87,$D8,$FE,$BD,$FC,$F4
	.dc.b	$F8,$67,$F9,$54,$76,$9F,$73,$E1,$3F,$7A,$7D,$09,$F9,$F1,$F8,$E8
	.dc.b	$FB,$F3,$F8,$E7,$F5,$C7,$92,$7E,$09,$DE,$3F,$84,$47,$F5,$79,$F8
	.dc.b	$8F,$F0,$23,$D8,$FC,$D9,$E3,$18,$7E,$64,$7A,$EB,$F5,$11,$FE,$8A
	.dc.b	$8F,$F3,$E4,$FC,$28,$FC,$09,$C2,$BF,$29,$1E,$D3,$C5,$1F,$78,$8F
	.dc.b	$79,$FD,$7B,$CE,$7E,$BB,$F1,$0F,$A8,$7A,$4F,$D7,$1F,$41,$EE,$FE
	.dc.b	$38,$F8,$91,$E9,$7F,$8D,$3F,$2D,$FF,$42,$D7,$FA,$1E,$3E,$9C,$71
	.dc.b	$7E,$21,$9A,$F9,$8F,$ED,$6B,$F8,$5C,$F4,$9F,$88,$7F,$91,$DF,$54
	.dc.b	$7E,$01,$F5,$4F,$63,$F1,$0F,$9A,$3F,$83,$47,$EF,$A3,$F5,$C7,$F0
	.dc.b	$63,$F2,$E7,$96,$7E,$74,$FE,$0E,$7F,$46,$7C,$A3,$EF,$CF,$E4,$8F
	.dc.b	$9C,$7E,$46,$7D,$6F,$C9,$9F,$E1,$F1,$F9,$03,$D7,$5F,$06,$78,$9E
	.dc.b	$A8,$E2,$FA,$67,$89,$FE,$F0,$AF,$9C,$7E,$18,$F8,$13,$F8,$67,$CC
	.dc.b	$F2,$7E,$7C,$7C,$28,$F1,$57,$F8,$56,$3E,$EC,$FC,$83,$6F,$C1,$AF
	.dc.b	$C7,$1F,$C8,$23,$E0,$CF,$E6,$0E,$53,$B9,$F9,$53,$E2,$7D,$A7,$E5
	.dc.b	$8F,$E0,$27,$E3,$CF,$D7,$BE,$08,$FE,$40,$7E,$5A,$38,$1B,$9F,$03
	.dc.b	$FD,$56,$36,$7E,$0D,$D9,$D2,$BE,$DC,$7E,$32,$3F,$CB,$13,$FD,$1D
	.dc.b	$FE,$03,$1C,$8F,$DE,$BE,$F9,$F9,$47,$27,$F8,$54,$7D,$38,$FB,$6F
	.dc.b	$CA,$7E,$94,$7E,$0C,$F8,$27,$33,$E0,$9F,$89,$3F,$5E,$7D,$F8,$F1
	.dc.b	$57,$E8,$E3,$BC,$7F,$75,$8F,$C0,$CF,$E8,$DF,$C0,$C7,$C7,$9E,$C7
	.dc.b	$48,$F9,$8F,$28,$FE,$DD,$3E,$87,$EC,$4F,$E0,$63,$DD,$3F,$AE,$9F
	.dc.b	$86,$FF,$50,$9F,$C0,$BF,$B5,$AF,$C0,$CF,$A2,$3F,$40,$D7,$D8,$78
	.dc.b	$63,$E8,$3F,$C1,$A3,$E4,$1E,$97,$B5,$7D,$0A,$FC,$54,$7E,$04,$F0
	.dc.b	$1F,$81,$3F,$72,$FA,$4D,$BF,$81,$6D,$7E,$22,$3F,$B6,$9F,$E6,$48
	.dc.b	$FD,$EB,$F5,$CF,$91,$3F,$0A,$3D,$F1,$F8,$83,$FC,$0C,$7A,$6B,$E4
	.dc.b	$CF,$E0,$23,$F7,$47,$E0,$0F,$59,$F8,$0A,$F1,$3F,$C4,$4F,$9D,$5F
	.dc.b	$7D,$A7,$CA,$3E,$41,$E0,$7F,$80,$4F,$C5,$8F,$41,$EF,$2C,$F6,$7F
	.dc.b	$3A,$7B,$CF,$DD,$9E,$43,$DF,$3F,$19,$F9,$A7,$C6,$3F,$A2,$9D,$4F
	.dc.b	$69,$C0,$EE,$70,$3D,$8F,$F2,$63,$E6,$BC,$0E,$47,$ED,$A3,$CC,$FC
	.dc.b	$BA,$FF,$2D,$CF,$08,$FD,$DB,$5F,$CE,$9F,$A9,$23,$97,$D8,$76,$37
	.dc.b	$3D,$A3,$E4,$C7,$EC,$9E,$07,$F8,$44,$FD,$5C,$FF,$0B,$8E,$27,$B4
	.dc.b	$F9,$67,$C4,$8E,$35,$FD,$FA,$3E,$21,$EF,$AF,$04,$7E,$DC,$FE,$EB
	.dc.b	$1A,$3D,$A7,$D4,$9F,$14,$6A,$3D,$A6,$5E,$CF,$B4,$F7,$CF,$63,$FC
	.dc.b	$4D,$3E,$43,$FA,$74,$77,$3D,$67,$F9,$A0,$FC,$64,$F9,$5F,$CC,$47
	.dc.b	$89,$11,$AC,$A1,$23,$31,$B2,$9C,$1D,$C7,$08,$C9,$B3,$A3,$CD,$E7
	.dc.b	$1A,$7C,$51,$C9,$C4,$76,$8E,$AF,$80,$E1,$1F,$E6,$43,$07,$9A,$36
	.dc.b	$F6,$39,$BD,$5F,$2C,$75,$8F,$24,$7A,$CD,$BE,$B7,$67,$57,$6F,$B4
	.dc.b	$F2,$3D,$5F,$19,$F0,$14,$62,$37,$1D,$5C,$1F,$E8,$D3,$A1,$F1,$9E
	.dc.b	$8F,$AC,$E9,$1C,$1F,$A5,$1A,$7D,$46,$8E,$07,$81,$F9,$31,$F0,$5F
	.dc.b	$2C,$68,$F0,$3E,$57,$0F,$CC,$7A,$9C,$CF,$AE,$7E,$6E,$3F,$2A,$FE
	.dc.b	$6D,$FC,$83,$F9,$73,$F7,$7F,$82,$7E,$26,$BF,$13,$48,$C1,$F5,$A3
	.dc.b	$11,$FD,$30,$FD,$D4,$FF,$04,$AF,$D2,$46,$4F,$EB,$AF,$82,$3A,$1F
	.dc.b	$9F,$9E,$F5,$DA,$3F,$19,$1D,$AB,$FA,$09,$FD,$66,$B9,$1D,$9C,$9E
	.dc.b	$27,$B5,$7A,$E3,$F4,$6F,$5A,$FC,$FD,$75,$9F,$ED,$55,$FD,$EA,$7F
	.dc.b	$A2,$9F,$DB,$0A,$E4,$4F,$EC,$AB,$B5,$7E,$D9,$ED,$5F,$22,$3E,$15
	.dc.b	$7F,$51,$3F,$90,$BF,$12,$BF,$87,$C7,$AE,$3E,$94,$FF,$02,$AE,$B1
	.dc.b	$F2,$23,$F3,$35,$5E,$D1,$F0,$8E,$95,$F8,$03,$A4,$7F,$7A,$9F,$ED
	.dc.b	$4F,$F8,$EA,$BA,$9E,$87,$EA,$3C,$5F,$C5,$57,$A9,$FE,$D0,$7C,$49
	.dc.b	$FE,$F4,$7C,$37,$EE,$47,$D0,$79,$9F,$72,$7A,$3F,$E5,$A3,$FB,$6D
	.dc.b	$79,$9E,$8F,$F8,$3A,$3C,$B1,$C8,$FB,$2F,$28,$FF,$0A,$47,$23,$CF
	.dc.b	$3F,$0E,$3E,$85,$79,$67,$E4,$CF,$95,$F9,$75,$F9,$D3,$A4,$FC,$68
	.dc.b	$FE,$C7,$5E,$53,$EC,$9D,$1F,$AC,$7C,$0A,$FC,$13,$E7,$3E,$AB,$F8
	.dc.b	$E9,$F8,$F1,$FD,$09,$FC,$0B,$E8,$8F,$C0,$BF,$D1,$2B,$E0,$BE,$63
	.dc.b	$E5,$47,$D3,$7F,$2D,$5E,$47,$EA,$B5,$CD,$7C,$18,$FC,$CC,$FC,$17
	.dc.b	$F3,$75,$F0,$63,$A4,$7E,$7E,$7F,$C4,$55,$FE,$0D,$8F,$85,$1C,$CF
	.dc.b	$DC,$C7,$BC,$FC,$4C,$F2,$8F,$7C,$7F,$0F,$9F,$DC,$BE,$F3,$F8,$11
	.dc.b	$F2,$8F,$AB,$3F,$29,$F8,$07,$E0,$1E,$91,$FE,$52,$3C,$71,$F2,$DF
	.dc.b	$E3,$F5,$F2,$1F,$6C,$7C,$89,$FC,$23,$F1,$AB,$DA,$FF,$81,$8F,$CA
	.dc.b	$1F,$E2,$BA,$E5,$1F,$AF,$8F,$6C,$F2,$7E,$69,$E1,$7C,$AF,$F9,$22
	.dc.b	$78,$9F,$36,$38,$57,$CD,$9E,$11,$8A,$F8,$F1,$FB,$F9,$FA,$F3,$F0
	.dc.b	$DF,$5C,$7C,$38,$E1,$1F,$0E,$3E,$61,$FA,$A3,$C4,$7E,$56,$3A,$4F
	.dc.b	$AE,$3F,$43,$1E,$87,$EF,$1C,$A7,$F1,$47,$9D,$FC,$A1,$CE,$BF,$84
	.dc.b	$3C,$A7,$C6,$FE,$CC,$F9,$67,$CD,$9F,$DB,$BF,$34,$E4,$7E,$24,$E5
	.dc.b	$1F,$A7,$8F,$64,$72,$3F,$66,$79,$6B,$F2,$53,$F9,$B7,$D9,$3F,$9C
	.dc.b	$8F,$64,$F8,$67,$F7,$8F,$F5,$73,$91,$F9,$19,$F9,$F3,$F1,$AB,$EC
	.dc.b	$3E,$39,$F6,$4F,$A9,$E9,$1F,$D5,$6B,$FC,$93,$5F,$80,$9F,$C8,$4F
	.dc.b	$17,$E6,$9E,$B3,$C4,$5A,$FE,$B2,$FC,$43,$F4,$27,$F0,$A3,$F9,$D4
	.dc.b	$77,$8F,$E2,$ED,$3D,$F1,$FE,$7A,$9F,$13,$F8,$78,$F1,$3E,$87,$FB
	.dc.b	$45,$7F,$12,$8F,$13,$EA,$7C,$55,$F9,$07,$C4,$7E,$34,$F5,$BF,$0D
	.dc.b	$FA,$A7,$27,$F6,$6F,$28,$FA,$A7,$F7,$17,$D7,$3E,$29,$F5,$B4,$F5
	.dc.b	$C7,$F8,$58,$E5,$5F,$51,$E3,$1F,$D8,$0D,$CF,$F0,$E8,$F1,$47,$F9
	.dc.b	$3E,$3F,$BC,$C7,$DF,$AF,$E3,$2F,$CE,$3F,$7B,$3F,$19,$F5,$CF,$B5
	.dc.b	$FA,$ED,$3E,$BB,$FC,$62,$3F,$10,$FE,$00,$F2,$C7,$F4,$77,$F9,$14
	.dc.b	$FC,$F8,$E4,$FC,$F8,$E5,$5F,$AC,$9F,$CD,$57,$EB,$27,$E7,$D7,$E6
	.dc.b	$A7,$DC,$72,$3E,$7C,$72,$3F,$58,$F2,$9F,$64,$7A,$A7,$91,$E3,$2C
	.dc.b	$F2,$3F,$47,$5E,$38,$FB,$13,$F8,$13,$F1,$55,$F4,$4F,$33,$FE,$0D
	.dc.b	$3E,$B9,$F0,$CE,$07,$CC,$8E,$F1,$F6,$1F,$85,$5E,$09,$F4,$C7,$82
	.dc.b	$3F,$89,$BF,$C9,$9F,$E8,$31,$F3,$1F,$44,$7D,$C7,$F3,$11,$F7,$0F
	.dc.b	$A6,$7F,$41,$3F,$98,$CF,$C4,$8F,$B6,$78,$63,$F0,$87,$F1,$43,$89
	.dc.b	$FD,$00,$F0,$47,$A6,$3C,$2F,$F1,$3A,$FC,$B4,$FC,$03,$E8,$47,$ED
	.dc.b	$CF,$8E,$FE,$DC,$F8,$E7,$ED,$CE,$F3,$FC,$C5,$FF,$3D,$57,$D7,$9E
	.dc.b	$47,$F8,$22,$3F,$9E,$BE,$17,$E8,$4F,$C8,$3F,$95,$9F,$87,$8F,$41
	.dc.b	$FA,$59,$FE,$56,$FF,$4F,$7F,$6F,$5E,$C8,$C3,$EC,$9E,$95,$FC,$92
	.dc.b	$BE,$DD,$7F,$AD,$E7,$F5,$F3,$FD,$20,$FC,$A1,$F3,$1F,$CA,$3F,$E0
	.dc.b	$23,$4F,$D1,$3E,$CB,$F4,$4F,$94,$FE,$82,$7B,$CF,$C1,$AF,$E3,$F5
	.dc.b	$FC,$5A,$BC,$2F,$C4,$9F,$4B,$F1,$2B,$FC,$02,$FE,$4E,$3E,$5B,$E3
	.dc.b	$38,$3F,$BC,$7E,$A1,$FE,$72,$3F,$85,$3F,$3D,$CC,$FC,$47,$F6,$F1
	.dc.b	$D1,$E0,$7B,$A7,$F2,$C7,$CE,$8F,$A5,$1D,$1F,$3B,$D2,$3C,$EF,$E2
	.dc.b	$A3,$F9,$31,$F3,$E3,$EC,$BE,$78,$F8,$51,$8A,$FD,$D1,$FC,$BA,$7F
	.dc.b	$70,$7E,$86,$3F,$62,$79,$23,$EF,$3C,$67,$F6,$71,$C6,$BF,$8E,$47
	.dc.b	$E5,$1F,$C3,$57,$86,$3F,$60,$7C,$03,$E5,$CF,$8A,$3F,$22,$7E,$8A
	.dc.b	$3F,$9F,$1F,$86,$9F,$C0,$9F,$86,$79,$4F,$DE,$7A,$C7,$CF,$7F,$89
	.dc.b	$BE,$48,$D4,$7E,$26,$3F,$07,$18,$7A,$1C,$63,$F0,$53,$F3,$0F,$C6
	.dc.b	$3F,$B4,$3E,$59,$F5,$5F,$73,$D6,$BE,$AC,$FE,$56,$7F,$00,$F8,$27
	.dc.b	$91,$E8,$9F,$98,$FE,$8D,$FE,$3C,$72,$3F,$7D,$3E,$38,$FC,$A5,$7A
	.dc.b	$67,$71,$E3,$8F,$99,$3F,$2E,$BC,$4F,$19,$F9,$27,$88,$F4,$1C,$4F
	.dc.b	$BA,$FD,$43,$F0,$91,$F7,$27,$AC,$7C,$49,$FA,$D5,$F9,$98,$F0,$3E
	.dc.b	$87,$F1,$B1,$E8,$3C,$15,$F4,$27,$E0,$BF,$79,$F5,$C7,$F2,$C9,$E7
	.dc.b	$3E,$18,$F2,$C7,$E7,$DF,$41,$F2,$6B,$F2,$8F,$F3,$D3,$94,$79,$E7
	.dc.b	$F8,$65,$79,$E3,$0F,$E7,$E7,$F8,$21,$FD,$10,$FE,$62,$FF,$9D,$2B
	.dc.b	$E8,$1E,$38,$F0,$4F,$C7,$77,$5F,$ED,$97,$DF,$5E,$C7,$F4,$27,$AA
	.dc.b	$7E,$69,$C5,$FA,$A7,$CC,$AF,$D8,$BE,$EA,$F8,$D1,$EE,$8F,$EA,$47
	.dc.b	$EC,$67,$F8,$4B,$FC,$35,$FE,$28,$FE,$60,$C3,$FC,$31,$C1,$C5,$F9
	.dc.b	$E7,$27,$CE,$7B,$9F,$A0,$F2,$7F,$43,$3F,$6E,$3E,$41,$FC,$DA,$BF
	.dc.b	$85,$CF,$8A,$3F,$40,$7A,$E3,$E9,$3F,$20,$FE,$95,$1F,$A8,$8F,$C2
	.dc.b	$9A,$8F,$8B,$3F,$6A,$BE,$2C,$7D,$E3,$91,$FE,$8D,$3E,$51,$F7,$0F
	.dc.b	$43,$F9,$A7,$67,$F0,$83,$F1,$07,$C0,$8F,$9F,$5F,$0E,$7F,$02,$FD
	.dc.b	$78,$FE,$11,$3F,$C6,$1F,$23,$DC,$E5,$5E,$98,$D9,$FB,$49,$FD,$2C
	.dc.b	$FB,$27,$F2,$4F,$F3,$F3,$E3,$57,$B9,$FA,$D3,$F8,$83,$E9,$BD,$A3
	.dc.b	$47,$9A,$3E,$D9,$0F,$CF,$9F,$4E,$3F,$80,$9F,$10,$FC,$A1,$F1,$0F
	.dc.b	$E4,$6F,$9A,$BF,$8D,$47,$CC,$3C,$C7,$13,$F0,$D1,$E5,$23,$FC,$04
	.dc.b	$E0,$7E,$42,$7C,$D1,$A8,$FC,$44,$FE,$B4,$FA,$31,$EE,$9D,$CF,$94
	.dc.b	$FE,$C5,$1E,$5A,$FC,$3C,$FC,$29,$FD,$8B,$FA,$7A,$FF,$15,$1F,$E4
	.dc.b	$88,$F9,$07,$D7,$8E,$AF,$8A,$BE,$F5,$74,$8D,$1D,$23,$2F,$E8,$2B
	.dc.b	$E2,$9F,$82,$7E,$29,$D8,$AF,$84,$F1,$57,$81,$E3,$5F,$82,$9F,$DA
	.dc.b	$C6,$DF,$F6,$B9,$F3,$DF,$8B,$5F,$CC,$E0,$7F,$A7,$0F,$F1,$81,$82
	.dc.b	$52,$21,$98,$82,$2C,$69,$C9,$A8,$E7,$1C,$9C,$1C,$4E,$8F,$17,$B4
	.dc.b	$69,$FF,$45,$9D,$0D,$C7,$96,$31,$1D,$DF,$3C,$71,$7C,$AE,$E3,$A9
	.dc.b	$E2,$8D,$B4,$EB,$1E,$57,$E1,$9B,$7D,$4F,$73,$71,$E2,$7E,$0C,$69
	.dc.b	$C1,$F3,$CE,$86,$9F,$9C,$F8,$4F,$A4,$FC,$37,$C2,$7B,$5E,$8F,$A8
	.dc.b	$DB,$ED,$76,$75,$7E,$13,$E6,$38,$19,$3C,$07,$E1,$1F,$03,$F8,$43
	.dc.b	$E9,$BE,$07,$F8,$79,$F9,$E3,$F6,$E7,$F3,$83,$F2,$15,$FC,$6C,$FC
	.dc.b	$47,$CA,$7E,$36,$D1,$F5,$A3,$09,$FC,$44,$69,$FD,$F4,$73,$9F,$09
	.dc.b	$FA,$8B,$FD,$24,$76,$8F,$CE,$D7,$6B,$FC,$CD,$F6,$3A,$D7,$78,$FE
	.dc.b	$71,$5F,$B7,$7F,$85,$35,$FE,$46,$7F,$63,$7D,$C7,$9C,$EF,$1E,$B6
	.dc.b	$CF,$63,$D3,$7F,$C3,$E3,$E3,$9F,$CB,$0F,$E0,$07,$F6,$E8,$F8,$75
	.dc.b	$E8,$8E,$67,$EC,$2F,$E2,$47,$A2,$39,$C7,$F6,$D7,$E8,$D7,$78,$FE
	.dc.b	$1D,$1D,$4F,$3C,$F5,$51,$F3,$DE,$D7,$F1,$67,$E1,$C7,$68,$FF,$0B
	.dc.b	$57,$D0,$8F,$A6,$75,$39,$C7,$C3,$6D,$FC,$3A,$F4,$5F,$92,$7E,$5D
	.dc.b	$F3,$7D,$27,$F6,$5A,$F8,$75,$FC,$4D,$FB,$51,$E6,$7E,$6B,$F9,$3A
	.dc.b	$E8,$FC,$DA,$E8,$FD,$73,$CB,$5F,$06,$7E,$04,$7F,$1D,$7E,$6C,$7C
	.dc.b	$07,$99,$F3,$63,$EB,$9E,$83,$F0,$C7,$FB,$0E,$BC,$F3,$DE,$7A,$CF
	.dc.b	$E4,$63,$E0,$DF,$F3,$98,$E4,$7C,$DB,$F5,$BF,$D1,$2F,$99,$F0,$1F
	.dc.b	$34,$FD,$5A,$FC,$A9,$FE,$2E,$7E,$AC,$79,$0F,$F0,$44,$7F,$7B,$BF
	.dc.b	$55,$74,$9F,$C6,$9D,$1F,$0B,$FD,$C4,$E5,$1F,$D4,$27,$95,$7F,$95
	.dc.b	$EB,$95,$7D,$E3,$EA,$4F,$B8,$DD,$F4,$8F,$7D,$FC,$CB,$E5,$1F,$94
	.dc.b	$AF,$BC,$F9,$0F,$90,$FD,$38,$F9,$11,$F7,$4F,$2B,$FD,$C5,$F9,$07
	.dc.b	$38,$FA,$51,$F8,$18,$E7,$1F,$22,$BF,$C7,$D1,$F1,$A3,$F7,$CF,$53
	.dc.b	$EA,$3F,$E5,$FB,$F7,$DF,$DC,$9F,$DF,$CF,$BC,$E3,$1F,$87,$AF,$C8
	.dc.b	$9F,$92,$AE,$37,$E3,$BE,$37,$ED,$3E,$39,$CC,$FB,$95,$F2,$63,$F8
	.dc.b	$45,$72,$3F,$59,$3C,$4F,$0D,$7E,$0A,$7D,$31,$93,$99,$EB,$7E,$34
	.dc.b	$FE,$35,$F8,$A7,$F8,$CE,$FE,$B5,$F8,$5F,$64,$7D,$03,$FC,$11,$3E
	.dc.b	$37,$F0,$35,$DE,$BE,$6D,$7E,$26,$3C,$B1,$CE,$FF,$03,$5F,$B4,$9F
	.dc.b	$C0,$BF,$8B,$3F,$28,$FC,$39,$E7,$3F,$7E,$7F,$2C,$7B,$A7,$FA,$5B
	.dc.b	$E4,$3D,$A7,$E2,$CF,$9E,$FE,$05,$FF,$2D,$9F,$81,$7E,$29,$E9,$73
	.dc.b	$5F,$D6,$1E,$87,$9E,$3F,$1B,$7F,$49,$E9,$1F,$E5,$E3,$E5,$57,$E6
	.dc.b	$2B,$E4,$C7,$9C,$FE,$10,$79,$8F,$80,$7D,$23,$C9,$7F,$01,$FE,$5E
	.dc.b	$7E,$AA,$7A,$1F,$2A,$7F,$B0,$57,$CA,$7F,$2D,$3F,$CC,$A3,$E5,$3E
	.dc.b	$C9,$F9,$4F,$38,$FC,$89,$FE,$88,$BF,$B9,$1F,$4A,$3E,$23,$ED,$3C
	.dc.b	$A7,$33,$CA,$7E,$58,$F6,$57,$88,$FE,$5B,$1F,$19,$FF,$2A,$C7,$C6
	.dc.b	$8F,$09,$9A,$FF,$00,$47,$F9,$62,$BC,$6F,$C5,$AF,$C6,$57,$D8,$38
	.dc.b	$5F,$F6,$1B,$F8,$AF,$F8,$FA,$BF,$AA,$9F,$6A,$BE,$3D,$7D,$A8,$FC
	.dc.b	$CB,$C9,$FE,$99,$5C,$A3,$E4,$C7,$C1,$9F,$C2,$3E,$53,$DF,$3F,$BD
	.dc.b	$7E,$84,$FC,$89,$F7,$C7,$C8,$39,$4F,$D2,$7E,$41,$F8,$DB,$FD,$A3
	.dc.b	$F9,$83,$E3,$57,$EA,$8F,$1D,$74,$3E,$1D,$FE,$1E,$7C,$87,$33,$F4
	.dc.b	$47,$92,$7D,$CF,$F8,$AC,$FD,$E1,$FD,$C8,$FE,$56,$71,$3F,$2A,$78
	.dc.b	$CF,$54,$F8,$6F,$D4,$FE,$78,$FE,$13,$3F,$C3,$E7,$94,$7E,$0E,$7F
	.dc.b	$AD,$47,$B2,$7E,$D3,$6B,$D4,$FF,$70,$3E,$2D,$7D,$8B,$F8,$B1,$EB
	.dc.b	$38,$1F,$93,$3D,$47,$F0,$68,$F1,$D7,$E7,$63,$EB,$CF,$EE,$5E,$27
	.dc.b	$D4,$3F,$37,$3E,$BB,$FC,$15,$71,$9F,$5C,$7D,$F9,$FC,$EC,$72,$7E
	.dc.b	$23,$CA,$3F,$9A,$47,$8A,$7F,$8B,$CF,$D6,$8F,$63,$F5,$9E,$6F,$CE
	.dc.b	$9F,$D1,$BF,$0E,$3C,$93,$FA,$88,$FE,$16,$79,$2B,$D8,$F1,$BF,$D2
	.dc.b	$4E,$E3,$07,$F8,$BC,$F8,$67,$E8,$8E,$75,$E4,$7F,$02,$7C,$F3,$89
	.dc.b	$F8,$93,$F8,$E1,$FA,$E3,$D9,$7F,$E4,$38,$FC,$05,$FE,$CD,$FD,$4C
	.dc.b	$FF,$76,$AF,$F1,$A5,$7C,$B3,$F0,$B1,$D8,$FB,$4F,$D2,$8F,$93,$1C
	.dc.b	$5F,$E4,$11,$C6,$BB,$3F,$E8,$59,$F8,$87,$B4,$E8,$7F,$21,$9F,$D6
	.dc.b	$1F,$76,$BE,$01,$E8,$7F,$34,$E2,$7B,$47,$EC,$9F,$AB,$7F,$DF,$EB
	.dc.b	$E5,$9F,$CA,$67,$CF,$7F,$C6,$A3,$F5,$57,$FB,$87,$F3,$07,$DD,$BF
	.dc.b	$E7,$87,$C9,$3B,$D7,$95,$C5,$7D,$EA,$E5,$5F,$99,$3F,$42,$7F,$A6
	.dc.b	$23,$C0,$FD,$98,$FA,$31,$87,$F1,$E7,$23,$F3,$31,$C8,$FD,$8C,$F8
	.dc.b	$1F,$95,$1C,$9F,$79,$FC,$D6,$3F,$81,$9F,$A0,$3E,$C3,$F8,$C7,$93
	.dc.b	$F1,$4C,$57,$E0,$AB,$D9,$3C,$AF,$FC,$10,$7D,$CA,$FE,$75,$1C,$9F
	.dc.b	$01,$8A,$E4,$F9,$A3,$94,$7D,$79,$FF,$61,$9C,$E7,$F2,$F1,$C1,$F2
	.dc.b	$1F,$9B,$3F,$1C,$FF,$47,$73,$7F,$1E,$BF,$03,$1F,$CA,$8D,$DF,$F1
	.dc.b	$69,$F9,$B3,$F9,$37,$F1,$51,$EE,$9F,$E0,$D7,$EE,$BF,$C7,$DF,$B6
	.dc.b	$7F,$0E,$F1,$30,$FD,$B7,$DB,$1F,$D3,$DF,$E0,$F5,$F7,$5F,$D5,$DF
	.dc.b	$D7,$AF,$D1,$9E,$A8,$E3,$5F,$D1,$23,$FD,$79,$3E,$67,$E5,$1F,$36
	.dc.b	$3F,$5F,$5E,$98,$FC,$34,$FF,$14,$9F,$D5,$9F,$A0,$8F,$C2,$9E,$A3
	.dc.b	$93,$F2,$DF,$B2,$72,$AF,$C4,$47,$F1,$79,$F4,$1E,$39,$F4,$1C,$AB
	.dc.b	$E7,$57,$B8,$F5,$C7,$F3,$37,$E1,$BE,$23,$E4,$1C,$4F,$E6,$0F,$17
	.dc.b	$E9,$D7,$E1,$DF,$E4,$C7,$C8,$3C,$E7,$DE,$38,$CF,$D9,$8F,$A5,$7F
	.dc.b	$8E,$7E,$D9,$F8,$D9,$E3,$3F,$54,$E1,$1F,$46,$3F,$70,$79,$EF,$F1
	.dc.b	$2F,$9C,$FE,$66,$E1,$F9,$15,$C9,$FD,$5C,$FD,$67,$EC,$9F,$9E,$70
	.dc.b	$FC,$A3,$8B,$EF,$7F,$C0,$33,$F3,$DF,$39,$E3,$BF,$8C,$F0,$9F,$7B
	.dc.b	$F1,$A7,$83,$F7,$CF,$04,$7D,$F7,$07,$F1,$FA,$F4,$3F,$AF,$BF,$89
	.dc.b	$3F,$20,$F3,$C7,$D0,$9F,$F5,$6B,$FA,$F9,$DC,$FF,$25,$3E,$74,$76
	.dc.b	$9D,$BF,$D5,$AB,$F1,$77,$FA,$A7,$E5,$47,$DB,$3E,$6C,$7B,$9F,$35
	.dc.b	$7C,$57,$F4,$B1,$C4,$D5,$F8,$A7,$ED,$4F,$FB,$2E,$BF,$C0,$0F,$D4
	.dc.b	$8F,$84,$FE,$48,$F8,$B3,$F9,$A3,$E1,$CF,$F0,$DB,$FC,$55,$7B,$4F
	.dc.b	$87,$5F,$2A,$BC,$0F,$E7,$E7,$FB,$FC,$FE,$CA,$3C,$13,$D6,$3E,$61
	.dc.b	$E2,$AF,$83,$5F,$19,$F3,$D7,$D3,$3F,$01,$5F,$C9,$DF,$8D,$1F,$D9
	.dc.b	$6B,$F0,$31,$FE,$03,$8E,$27,$CE,$8F,$A7,$1E,$F3,$8D,$FF,$5C,$8F
	.dc.b	$8F,$3F,$C2,$5F,$F2,$1C,$F8,$4F,$D8,$1E,$F3,$C6,$FD,$37,$FB,$D3
	.dc.b	$C2,$3C,$0F,$C6,$8F,$05,$7C,$67,$FC,$02,$7D,$5A,$FE,$AA,$F7,$AF
	.dc.b	$CC,$BC,$0E,$F5,$EE,$7F,$C2,$6F,$F1,$59,$FE,$68,$FB,$A7,$F1,$75
	.dc.b	$DD,$FB,$55,$FE,$8C,$9E,$0F,$EE,$E3,$E5,$47,$F2,$78,$F7,$9F,$1E
	.dc.b	$FE,$A5,$F9,$AB,$DB,$5F,$66,$3F,$23,$5E,$68,$F9,$71,$FC,$A9,$FC
	.dc.b	$14,$7B,$63,$F3,$17,$F6,$CF,$33,$FC,$A2,$3E,$2D,$23,$FC,$27,$5E
	.dc.b	$D7,$BC,$F9,$A7,$F1,$2F,$F6,$1B,$FF,$11,$39,$9D,$CF,$CC,$37,$1F
	.dc.b	$B3,$8F,$04,$7B,$2F,$F2,$D7,$FE,$6F,$A0,$60,$8C,$C7,$FB,$D8,$A7
	.dc.b	$FB,$26,$36,$E0,$D1,$C9,$E4,$FF,$91,$CD,$47,$FB,$25,$A7,$48,$F0
	.dc.b	$C6,$23,$A1,$FE,$58,$30,$7A,$23,$6F,$53,$CB,$1D,$5C,$46,$9F,$29
	.dc.b	$EC,$8D,$3E,$88,$D3,$FE,$BE,$3C,$EF,$B9,$F1,$C7,$BE,$37,$1F,$ED
	.dc.b	$53,$E2,$3F,$2D,$F4,$9F,$08,$F9,$6F,$08,$CC,$7B,$4F,$3C,$7D,$38
	.dc.b	$F7,$C7,$B2,$3C,$4E,$1E,$4F,$68,$F5,$BE,$93,$E9,$3C,$9F,$81,$1C
	.dc.b	$9E,$EF,$95,$E0,$69,$F5,$3F,$00,$FD,$11,$FC,$80,$FE,$EA,$FF,$20
	.dc.b	$8F,$C6,$BF,$E8,$AF,$C1,$3F,$13,$5F,$9F,$8F,$C6,$C6,$2C,$FD,$E9
	.dc.b	$FD,$23,$FA,$9B,$EE,$7E,$B6,$BF,$4D,$79,$8F,$D1,$3F,$CA,$67,$A4
	.dc.b	$F6,$9F,$CE,$D7,$47,$F2,$CF,$68,$F4,$47,$A9,$FE,$84,$7F,$01,$AE
	.dc.b	$D3,$FA,$DB,$FE,$B1,$7D,$AB,$F5,$0F,$AC,$F5,$47,$B2,$3A,$DF,$F6
	.dc.b	$0B,$E7,$3F,$C1,$A7,$37,$FC,$5A,$BC,$F3,$FA,$EB,$ED,$5F,$1A,$73
	.dc.b	$1D,$8E,$71,$F1,$0E,$73,$FD,$06,$7E,$B1,$E1,$8F,$CD,$57,$FB,$1A
	.dc.b	$BA,$9E,$99,$F8,$B5,$E9,$AF,$87,$7F,$72,$79,$DF,$9D,$FA,$93,$FD
	.dc.b	$0E,$3F,$BF,$DF,$49,$F3,$9D,$1F,$3C,$EE,$FF,$8C,$9F,$0D,$EF,$3F
	.dc.b	$4D,$FC,$53,$F1,$1F,$B8,$AB,$EE,$3F,$C5,$1F,$B9,$1C,$A3,$E0,$36
	.dc.b	$3D,$07,$F2,$38,$EC,$7D,$03,$F8,$FD,$7A,$5E,$D5,$F4,$5E,$CF,$C5
	.dc.b	$AF,$E9,$95,$F8,$5B,$F6,$C7,$E2,$4F,$A6,$FF,$9E,$EB,$94,$FE,$8E
	.dc.b	$7F,$13,$1F,$18,$EB,$7C,$AB,$E8,$47,$CE,$9E,$87,$AE,$FE,$F4,$7F
	.dc.b	$AA,$EB,$D1,$7F,$12,$3F,$29,$5E,$27,$FC,$AE,$7F,$81,$AF,$C8,$7F
	.dc.b	$56,$7E,$09,$F3,$E3,$CA,$FC,$2A,$FC,$84,$73,$8F,$F2,$F1,$CC,$EB
	.dc.b	$3C,$DF,$E9,$B3,$E4,$BF,$95,$1E,$4A,$F5,$57,$8A,$3F,$7F,$7F,$55
	.dc.b	$FB,$33,$F8,$27,$E9,$BF,$82,$8F,$74,$7E,$92,$3C,$8F,$F9,$B0,$FE
	.dc.b	$0A,$E1,$FF,$32,$DF,$8C,$F5,$C7,$8A,$3D,$27,$8A,$FF,$C2,$31,$F9
	.dc.b	$7B,$FE,$1D,$3F,$16,$3E,$DC,$FB,$23,$C5,$1C,$DF,$8B,$3F,$0D,$F8
	.dc.b	$C7,$C8,$AF,$13,$FA,$59,$F3,$3F,$E2,$53,$D2,$F3,$7F,$D4,$B1,$C1
	.dc.b	$FC,$3C,$70,$BF,$F1,$2C,$7A,$4F,$F1,$15,$FE,$1A,$3F,$80,$47,$A6
	.dc.b	$BF,$80,$3C,$AB,$B3,$FE,$43,$3C,$2F,$E7,$AB,$F6,$CF,$D3,$AF,$CD
	.dc.b	$3F,$0A,$BE,$5B,$B3,$F8,$A1,$D2,$BD,$93,$FA,$78,$FE,$27,$5F,$C7
	.dc.b	$E7,$C8,$7E,$8C,$FC,$34,$73,$3E,$91,$EF,$8F,$E1,$2F,$63,$C8,$FF
	.dc.b	$84,$6B,$DB,$3C,$8F,$E3,$31,$F0,$CF,$BB,$3F,$AC,$3B,$16,$3F,$82
	.dc.b	$1F,$9A,$3C,$A7,$E8,$CF,$B0,$FB,$6B,$E8,$C7,$C8,$3B,$DF,$F0,$1B
	.dc.b	$FB,$11,$E3,$3F,$79,$1F,$BE,$9F,$D0,$1F,$C5,$AF,$CB,$19,$7E,$34
	.dc.b	$FE,$8A,$FF,$6B,$5F,$A2,$8F,$DB,$47,$C5,$AF,$E1,$51,$FC,$78,$EE
	.dc.b	$59,$FC,$1B,$CE,$7E,$D3,$EC,$7E,$1C,$F8,$63,$F1,$6F,$ED,$DF,$71
	.dc.b	$FC,$44,$F8,$B3,$E3,$3F,$99,$CF,$28,$F0,$57,$B1,$A7,$43,$F8,$AD
	.dc.b	$FC,$67,$F5,$73,$C5,$FE,$06,$7C,$C3,$F6,$E7,$8A,$BD,$67,$E5,$27
	.dc.b	$F0,$31,$C8,$E0,$7E,$D0,$FD,$31,$FE,$C6,$3C,$67,$DF,$3E,$23,$C9
	.dc.b	$F9,$B3,$F7,$EB,$F2,$2F,$E5,$0C,$5F,$F0,$69,$E2,$7E,$6C,$F1,$1F
	.dc.b	$48,$F5,$96,$FF,$00,$FC,$33,$F8,$31,$F3,$2B,$F8,$73,$E2,$AF,$89
	.dc.b	$3C,$23,$D4,$74,$7F,$17,$3F,$9F,$8F,$E7,$A7,$90,$FC,$D1,$FC,$06
	.dc.b	$79,$C6,$E3,$9C,$F5,$7F,$8C,$9F,$C4,$8F,$25,$FF,$00,$7E,$69,$FC
	.dc.b	$C6,$3F,$7D,$5F,$10,$FD,$FC,$78,$A3,$EB,$3F,$14,$F6,$1C,$8F,$E2
	.dc.b	$4F,$28,$F2,$DF,$EA,$E7,$C2,$7E,$5C,$E0,$7F,$0E,$BF,$E2,$EE,$2B
	.dc.b	$FC,$2A,$F0,$3E,$01,$98,$FB,$75,$FC,$06,$3E,$A4,$7C,$37,$F9,$FC
	.dc.b	$7F,$51,$8F,$A7,$3F,$9E,$9E,$55,$FB,$07,$E9,$4F,$E7,$DA,$7C,$17
	.dc.b	$F8,$B3,$F0,$67,$93,$E0,$9F,$D7,$47,$C1,$8E,$B5,$F2,$9F,$CB,$47
	.dc.b	$CA,$7E,$D9,$FC,$35,$FC,$0C,$F2,$AF,$87,$3D,$4F,$D7,$9F,$0C,$E0
	.dc.b	$F9,$E7,$F0,$13,$FD,$64,$FE,$B5,$7F,$B5,$AF,$AE,$73,$3F,$58,$7E
	.dc.b	$CE,$B9,$1F,$26,$38,$57,$B8,$FD,$BC,$7C,$29,$F4,$9F,$A6,$8F,$4B
	.dc.b	$FB,$E7,$E1,$3F,$C3,$CF,$CD,$3F,$C3,$E3,$FC,$69,$1F,$CE,$0F,$AF
	.dc.b	$7F,$C9,$A3,$EB,$9A,$8F,$41,$F9,$77,$E2,$CF,$92,$7E,$F5,$7E,$76
	.dc.b	$BC,$13,$F3,$CF,$04,$7D,$F7,$E3,$C7,$F0,$C3,$C3,$3F,$CB,$1F,$C7
	.dc.b	$9C,$AB,$E0,$4F,$29,$FE,$70,$43,$FD,$62,$F8,$DF,$55,$78,$E3,$11
	.dc.b	$D6,$BE,$C9,$FC,$24,$F6,$3F,$D9,$5F,$89,$3F,$4A,$BE,$15,$72,$8F
	.dc.b	$E3,$CF,$D5,$8F,$B3,$7F,$89,$BE,$57,$E9,$7E,$2B,$FB,$13,$0F,$99
	.dc.b	$C5,$7D,$A7,$F0,$AF,$DA,$3C,$0F,$E8,$6B,$C6,$7F,$5E,$8F,$9C,$FD
	.dc.b	$8B,$F9,$CF,$EC,$EF,$F2,$75,$F9,$93,$F1,$11,$C4,$F9,$A7,$1B,$F9
	.dc.b	$87,$F0,$A8,$FD,$09,$F8,$A8,$D4,$7D,$07,$93,$EE,$8F,$B3,$7E,$B7
	.dc.b	$E6,$C7,$F4,$13,$FC,$22,$6D,$FA,$67,$ED,$A7,$C7,$7F,$65,$F9,$B1
	.dc.b	$F9,$D7,$F2,$27,$82,$7E,$54,$78,$23,$11,$F7,$E3,$DE,$7D,$F3,$F9
	.dc.b	$3C,$72,$7F,$87,$47,$8D,$D9,$E3,$70,$FE,$EA,$BD,$8F,$28,$FB,$F1
	.dc.b	$F5,$A3,$F6,$A6,$E7,$C7,$5F,$88,$7E,$CC,$78,$CF,$E6,$55,$CA,$37
	.dc.b	$1F,$7E,$7F,$B3,$57,$DE,$7E,$24,$71,$8F,$F1,$59,$F8,$79,$E3,$1E
	.dc.b	$28,$FC,$B3,$EE,$AF,$8E,$FD,$03,$FA,$B9,$EB,$7E,$B3,$FA,$23,$F1
	.dc.b	$07,$AC,$E4,$FC,$D7,$07,$F6,$29,$FC,$83,$FB,$DA,$FD,$51,$EA,$8F
	.dc.b	$BB,$1F,$A0,$7E,$EB,$FA,$07,$EA,$3F,$46,$7C,$EF,$D4,$3F,$83,$CF
	.dc.b	$CB,$3F,$3F,$7F,$B1,$9F,$4C,$FE,$B9,$F3,$5F,$C2,$8E,$91,$8B,$FA
	.dc.b	$71,$E9,$AE,$53,$FD,$76,$BE,$E1,$FC,$45,$FD,$04,$FC,$2B,$F1,$47
	.dc.b	$EA,$23,$FA,$8B,$A7,$BC,$6A,$31,$1E,$A3,$EF,$3C,$5F,$E5,$E7,$FA
	.dc.b	$C8,$FA,$33,$E9,$78,$1F,$76,$FD,$2F,$DA,$3F,$17,$3F,$0A,$7D,$33
	.dc.b	$FB,$88,$FD,$89,$F3,$27,$EC,$9C,$67,$F1,$73,$F4,$A3,$BC,$7C,$19
	.dc.b	$E2,$7C,$9A,$FA,$33,$FB,$58,$FC,$D3,$F1,$4F,$6B,$F0,$23,$F0,$6F
	.dc.b	$BE,$7F,$57,$27,$FC,$22,$FF,$0F,$9F,$CF,$BF,$28,$E7,$3F,$C5,$5F
	.dc.b	$C4,$3E,$07,$47,$C5,$8F,$C8,$47,$EE,$A7,$E5,$5F,$F7,$7A,$FC,$B1
	.dc.b	$F3,$AF,$E5,$DF,$9A,$FF,$C6,$84,$3E,$0B,$F6,$DB,$3E,$43,$F0,$2F
	.dc.b	$95,$FF,$21,$1E,$B3,$F4,$07,$C5,$3C,$CF,$C3,$9F,$D1,$C7,$0B,$FE
	.dc.b	$00,$F0,$3D,$AF,$D3,$7B,$90,$FD,$05,$79,$5F,$DD,$1E,$13,$F9,$31
	.dc.b	$FC,$26,$7F,$C7,$77,$FE,$99,$8E,$B1,$F8,$77,$8D,$78,$9E,$27,$E9
	.dc.b	$CF,$C0,$5F,$E3,$E7,$F8,$8C,$F9,$67,$C6,$7C,$53,$F3,$A7,$E5,$A7
	.dc.b	$F5,$D3,$C0,$F0,$47,$03,$E2,$C7,$D1,$7D,$D3,$F9,$CB,$FC,$2C,$7E
	.dc.b	$5A,$B8,$CF,$E3,$0F,$CD,$D7,$E1,$67,$55,$C0,$FB,$D1,$F5,$E3,$F9
	.dc.b	$13,$E4,$7E,$B4,$FA,$CF,$03,$F4,$CF,$D4,$57,$CA,$7F,$D4,$A7,$63
	.dc.b	$54,$61,$68,$48,$CC,$7F,$AB,$4A,$7F,$8C,$CA,$E2,$38,$87,$31,$C0
	.dc.b	$F0,$3A,$8C,$47,$53,$FC,$98,$53,$CB,$1B,$76,$F7,$8F,$34,$7F,$BB
	.dc.b	$8F,$1B,$E2,$76,$7A,$9F,$2C,$79,$23,$C8,$F5,$69,$B8,$EA,$FF,$AC
	.dc.b	$8E,$AF,$C6,$7A,$9F,$29,$EA,$F8,$CF,$93,$1D,$5F,$01,$F3,$E3,$AB
	.dc.b	$EF,$3E,$2B,$E9,$7C,$67,$57,$6F,$FB,$30,$DB,$E3,$76,$75,$7E,$93
	.dc.b	$E3,$3E,$5B,$D5,$F1,$9F,$40,$E8,$7D,$D3,$D4,$FB,$1F,$84,$7B,$1E
	.dc.b	$4F,$E9,$E3,$F3,$4F,$F1,$97,$F8,$FC,$7F,$3D,$AF,$E9,$1F,$82,$7E
	.dc.b	$41,$5F,$E0,$E3,$09,$FC,$44,$69,$E7,$3F,$CF,$CF,$F5,$65,$FE,$74
	.dc.b	$FE,$06,$53,$FC,$19,$1D,$23,$F2,$AF,$63,$AD,$77,$8E,$77,$FB,$2A
	.dc.b	$FE,$1F,$3F,$E0,$67,$AC,$65,$FE,$8D,$1F,$A6,$AE,$D5,$FA,$A7,$B4
	.dc.b	$FE,$A2,$FD,$53,$DA,$3E,$33,$5F,$E4,$87,$F8,$8E,$FB,$4F,$C6,$8F
	.dc.b	$DB,$D7,$6A,$F8,$D1,$FB,$B9,$F8,$CF,$7B,$F6,$4F,$F5,$F8,$EE,$D7
	.dc.b	$B9,$FE,$1B,$BE,$F3,$EB,$8E,$8F,$D5,$6D,$FC,$83,$E9,$CF,$F3,$D8
	.dc.b	$FE,$4D,$7F,$C5,$6B,$F0,$91,$FC,$56,$BD,$07,$C4,$7F,$86,$CF,$C4
	.dc.b	$BF,$23,$FC,$F6,$3E,$11,$E8,$8F,$EB,$AF,$F3,$27,$F0,$EF,$C4,$3C
	.dc.b	$EF,$DE,$7C,$E7,$C3,$9F,$3B,$FD,$06,$7F,$94,$BF,$70,$EC,$7F,$B4
	.dc.b	$E3,$CE,$F2,$3C,$F5,$F3,$2B,$C7,$5E,$F9,$FE,$6B,$3F,$D0,$0F,$82
	.dc.b	$FE,$52,$B9,$BF,$88,$7E,$55,$7E,$06,$BE,$04,$FD,$33,$C6,$7C,$CA
	.dc.b	$E8,$FB,$CF,$C8,$CF,$BD,$E8,$7B,$DE,$91,$F9,$98,$F9,$67,$E0,$0F
	.dc.b	$F1,$7D,$FE,$01,$E9,$5F,$B4,$AE,$8F,$E3,$23,$E0,$DF,$CA,$7E,$91
	.dc.b	$FA,$69,$FF,$0E,$5F,$F8,$62,$3D,$35,$F9,$97,$FA,$85,$7C,$28,$E9
	.dc.b	$1F,$B3,$AF,$39,$CA,$FB,$4F,$CD,$9F,$F0,$1C,$7F,$6E,$AE,$57,$FA
	.dc.b	$29,$F2,$DF,$23,$F8,$84,$78,$E3,$E0,$5F,$F9,$66,$FE,$04,$79,$0F
	.dc.b	$C4,$D7,$EE,$5F,$C0,$1E,$57,$F2,$B5,$F2,$E3,$EB,$1D,$1E,$D5,$F6
	.dc.b	$A3,$FC,$B7,$7E,$63,$EF,$D7,$23,$DC,$79,$CF,$E3,$86,$4F,$4D,$FF
	.dc.b	$55,$AF,$95,$5E,$F7,$EB,$4F,$9E,$3F,$02,$7C,$27,$91,$D0,$FC,$BD
	.dc.b	$FE,$2A,$3E,$59,$FB,$53,$E0,$C7,$58,$F9,$91,$F0,$63,$8D,$7F,$3E
	.dc.b	$8F,$80,$FD,$69,$E9,$1C,$67,$F1,$31,$CA,$3F,$84,$C7,$BE,$7F,$84
	.dc.b	$57,$EC,$A7,$F2,$AF,$C9,$3F,$A3,$BE,$4B,$FE,$10,$F8,$CF,$C8,$57
	.dc.b	$C6,$3A,$3F,$BD,$BF,$D8,$BF,$D0,$8E,$47,$DC,$3C,$71,$F1,$AF,$C7
	.dc.b	$3C,$27,$EE,$DF,$8E,$3F,$D3,$B7,$F6,$E7,$F5,$D7,$C5,$F0,$1F,$C0
	.dc.b	$AF,$E3,$5F,$17,$ED,$5F,$F2,$D9,$FD,$CD,$7D,$9A,$FD,$1C,$79,$23
	.dc.b	$E7,$BF,$E3,$83,$F5,$73,$E5,$7F,$1A,$7B,$0F,$C0,$9F,$14,$FE,$11
	.dc.b	$5E,$28,$FC,$C4,$7F,$00,$39,$9F,$C0,$1F,$E7,$CF,$23,$C4,$7C,$E9
	.dc.b	$FC,$53,$67,$E3,$BE,$F3,$F1,$47,$B2,$3F,$15,$7F,$A0,$7D,$71,$FC
	.dc.b	$BC,$FC,$4B,$F6,$1F,$E1,$67,$F6,$C3,$C8,$FE,$2A,$39,$47,$84,$F9
	.dc.b	$E7,$E6,$8E,$97,$F3,$E7,$FC,$0B,$3F,$5E,$7F,$49,$1E,$27,$F1,$F7
	.dc.b	$F1,$2B,$F8,$24,$7C,$43,$F8,$21,$EB,$AF,$C8,$BF,$10,$F9,$A7,$D5
	.dc.b	$3D,$75,$F8,$37,$D7,$3F,$B8,$9F,$14,$7D,$58,$FD,$03,$F8,$EA,$F1
	.dc.b	$1F,$C2,$8B,$7F,$88,$8F,$87,$7F,$87,$9F,$D6,$47,$F4,$B8,$F9,$55
	.dc.b	$FC,$9E,$3F,$C4,$47,$58,$FE,$1E,$7F,$C0,$0B,$F1,$C7,$AE,$7C,$6F
	.dc.b	$E1,$4F,$5C,$FC,$77,$FC,$07,$7E,$37,$E2,$CF,$F1,$29,$FB,$F1,$FA
	.dc.b	$47,$83,$CE,$3E,$A4,$7C,$57,$CD,$5E,$29,$F5,$4E,$6F,$F8,$5D,$FF
	.dc.b	$36,$9F,$59,$FA,$03,$EA,$1F,$BB,$38,$CF,$82,$FE,$DC,$7A,$8F,$C4
	.dc.b	$1F,$52,$BF,$65,$1E,$9B,$FB,$B3,$C8,$DC,$7C,$BB,$F1,$1E,$39,$F4
	.dc.b	$47,$F8,$66,$FC,$17,$F6,$5F,$9E,$62,$BE,$45,$FF,$25,$8F,$B5,$7F
	.dc.b	$9C,$8F,$C5,$9F,$9A,$AF,$C0,$CF,$DC,$7F,$96,$CF,$81,$FF,$09,$4F
	.dc.b	$99,$F8,$EF,$E6,$4F,$8F,$7F,$CA,$9F,$8E,$7E,$F4,$F7,$1F,$AF,$3E
	.dc.b	$DC,$7F,$02,$8F,$E2,$71,$A7,$F0,$A7,$CC,$3E,$DD,$7F,$8C,$A3,$8B
	.dc.b	$FB,$77,$F0,$47,$F8,$AC,$FE,$2D,$1D,$0F,$90,$FA,$6F,$F8,$B4,$FC
	.dc.b	$88,$E0,$7A,$27,$85,$7C,$08,$F4,$1F,$2E,$7F,$22,$FE,$F8,$E8,$7E
	.dc.b	$D6,$7C,$EF,$47,$F0,$0F,$E2,$5F,$1B,$D2,$FE,$71,$FD,$76,$FC,$B5
	.dc.b	$FD,$3C,$FD,$54,$FB,$2F,$F6,$77,$C8,$FE,$01,$19,$BF,$BD,$1F,$2A
	.dc.b	$7F,$0D,$5C,$CF,$95,$5E,$BA,$FC,$9D,$73,$AF,$CB,$9F,$54,$EE,$7C
	.dc.b	$37,$A5,$F9,$EB,$F9,$9C,$7B,$9F,$C0,$3C,$DF,$E1,$77,$F2,$1F,$E2
	.dc.b	$95,$FD,$36,$7E,$F3,$F9,$29,$F7,$CF,$D3,$8F,$0D,$FE,$70,$F8,$F3
	.dc.b	$F9,$7B,$F8,$EF,$F4,$8A,$F9,$87,$C3,$3F,$0A,$60,$E9,$3F,$1A,$BC
	.dc.b	$47,$C6,$9E,$57,$F3,$2B,$67,$27,$07,$BC,$FC,$3D,$7C,$38,$FD,$64
	.dc.b	$FE,$35,$F0,$4E,$EB,$D7,$3F,$D2,$27,$CC,$7E,$2C,$FD,$A9,$FD,$21
	.dc.b	$E5,$3E,$09,$FA,$95,$E9,$AF,$9A,$FD,$A8,$FA,$C7,$B2,$FC,$93,$FE
	.dc.b	$15,$3D,$67,$7A,$E9,$7F,$AB,$8C,$57,$18,$F9,$D1,$F7,$A3,$C2,$FC
	.dc.b	$37,$F0,$F1,$C6,$3E,$E9,$FD,$DC,$F8,$6F,$E4,$5F,$D2,$47,$27,$B9
	.dc.b	$EA,$3F,$4C,$FA,$9E,$47,$C3,$8E,$4F,$CC,$8E,$4F,$80,$FE,$0B,$1F
	.dc.b	$C8,$1F,$31,$C8,$FE,$AC,$F2,$9F,$82,$EE,$3C,$67,$27,$F3,$0F,$F5
	.dc.b	$99,$F8,$91,$FD,$26,$BF,$D0,$D3,$F6,$8F,$11,$FA,$D3,$D7,$7F,$49
	.dc.b	$F0,$CF,$E3,$AB,$E4,$BE,$18,$E3,$3F,$9E,$8E,$D1,$FC,$36,$7F,$4D
	.dc.b	$1F,$BA,$9D,$36,$7C,$47,$9C,$F8,$07,$F6,$59,$E2,$FD,$28,$FB,$0E
	.dc.b	$CF,$81,$3F,$8B,$3D,$0F,$18,$FC,$9C,$FA,$1F,$8B,$1F,$CE,$63,$13
	.dc.b	$C6,$7F,$13,$3F,$82,$7E,$03,$F7,$CF,$E0,$F3,$F7,$DF,$09,$F2,$23
	.dc.b	$8B,$F8,$17,$8B,$A3,$8C,$FF,$05,$7E,$09,$E3,$3E,$0D,$FF,$0E,$9F
	.dc.b	$4C,$78,$63,$F7,$31,$FB,$38,$FD,$29,$F6,$AF,$FB,$9B,$FD,$7E,$BE
	.dc.b	$7B,$F2,$67,$EC,$3F,$24,$FE,$C7,$7F,$C2,$EB,$FC,$B3,$1E,$7B,$EF
	.dc.b	$5F,$C7,$9F,$E0,$97,$FB,$B9,$FF,$27,$1E,$FA,$E2,$FE,$2C,$F7,$D7
	.dc.b	$9E,$B8,$C7,$E7,$8E,$27,$E3,$A7,$8D,$7A,$23,$F5,$15,$F8,$07,$E3
	.dc.b	$1C,$0F,$BC,$FE,$96,$FE,$DC,$7D,$38,$F4,$47,$EC,$23,$E0,$57,$C6
	.dc.b	$7F,$C4,$47,$69,$F8,$AF,$63,$D5,$7F,$0A,$3F,$C3,$07,$DE,$8F,$F2
	.dc.b	$0B,$ED,$AE,$D1,$C6,$FD,$55,$FD,$4A,$B8,$4F,$F8,$DE,$3E,$1B,$FB
	.dc.b	$A3,$BD,$7F,$30,$9F,$D5,$C7,$E4,$EF,$E2,$BF,$88,$AF,$EB,$0F,$D7
	.dc.b	$AE,$15,$E6,$38,$C7,$F2,$B9,$EE,$FE,$E0,$E5,$3F,$9B,$7F,$07,$5F
	.dc.b	$AA,$9F,$F1,$31,$C2,$3F,$11,$7E,$CA,$FE,$1B,$7D,$CF,$DE,$4F,$E5
	.dc.b	$0D,$9D,$A7,$A4,$FE,$3E,$3F,$D2,$E7,$C0,$8F,$C6,$34,$FC,$31,$E4
	.dc.b	$9F,$93,$5E,$4A,$F9,$EF,$C2,$9E,$07,$CB,$BF,$BD,$1F,$E1,$E8,$FB
	.dc.b	$D5,$FC,$0A,$38,$5F,$AE,$7F,$63,$7E,$57,$FC,$C8,$7E,$C5,$F9,$AF
	.dc.b	$EA,$EB,$94,$7C,$9B,$EE,$FF,$18,$9E,$F5,$FC,$36,$BF,$26,$7E,$6E
	.dc.b	$FD,$17,$C2,$3F,$7F,$5F,$6E,$BE,$18,$61,$60,$48,$CC,$7F,$AD,$8A
	.dc.b	$7F,$91,$4A,$E2,$38,$87,$31,$C0,$F0,$3A,$8C,$47,$53,$FD,$CA,$53
	.dc.b	$CB,$1B,$76,$F7,$8F,$34,$7F,$BF,$CF,$1B,$E2,$76,$7A,$9F,$2C,$79
	.dc.b	$23,$C8,$F5,$69,$B8,$EA,$F8,$8F,$0B,$D5,$D9,$F2,$9E,$AF,$8C,$F9
	.dc.b	$31,$D5,$F0,$1F,$3E,$3A,$BE,$F3,$E2,$BD,$5D,$9F,$10,$FF,$69,$9A
	.dc.b	$7E,$BB,$EE,$3C,$4F,$47,$67,$D7,$3C,$8F,$99,$F6,$1F,$09,$F0,$3A
	.dc.b	$79,$47,$D6,$8E,$47,$E6,$5F,$C5,$BF,$B3,$8F,$E3,$07,$E4,$EF,$F8
	.dc.b	$17,$E0,$9F,$8E,$C3,$FD,$B4,$D8,$3F,$85,$84,$62,$79,$9F,$CA,$A3
	.dc.b	$F5,$B5,$DE,$3F,$42,$66,$7B,$C7,$E7,$67,$BB,$63,$BB,$F9,$A8,$EF
	.dc.b	$5F,$DA,$63,$F8,$71,$FC,$BC,$ED,$5F,$C4,$A7,$F8,$C4,$7F,$3B,$9F
	.dc.b	$54,$FE,$81,$F6,$1E,$BA,$F5,$47,$98,$FD,$ED,$F7,$BE,$67,$B2,$FC
	.dc.b	$2F,$F1,$1B,$EB,$5F,$B4,$7A,$CF,$A6,$FA,$CF,$95,$FE,$32,$7F,$31
	.dc.b	$BF,$E2,$53,$FC,$04,$FE,$43,$5D,$2B,$E1,$C2,$FE,$1B,$D6,$7C,$D7
	.dc.b	$D2,$ED,$FD,$D6,$BF,$45,$FF,$78,$CF,$53,$CF,$5F,$C2,$AF,$E7,$D7
	.dc.b	$DC,$3E,$19,$6B,$F8,$E3,$EB,$3B,$1F,$AF,$8F,$C2,$CF,$F1,$8A,$EC
	.dc.b	$FD,$D7,$E8,$1E,$49,$EA,$7D,$1A,$EB,$3F,$CB,$DA,$FF,$B7,$A7,$9B
	.dc.b	$F0,$EF,$E2,$BF,$72,$BC,$CF,$F8,$0D,$F3,$1F,$04,$CB,$FE,$16,$7F
	.dc.b	$6A,$F5,$3F,$24,$FE,$1D,$F4,$1F,$3D,$E3,$7F,$E0,$58,$F3,$1D,$CA
	.dc.b	$75,$8F,$F6,$DD,$7A,$27,$EC,$CF,$CF,$AF,$C2,$BC,$23,$A3,$FD,$28
	.dc.b	$E8,$53,$A1,$FA,$73,$E7,$1E,$8A,$FC,$93,$E8,$3B,$57,$CF,$BF,$97
	.dc.b	$5F,$E8,$CA,$FA,$27,$C6,$2D,$75,$3F,$A9,$57,$A6,$FE,$7C,$7D,$B9
	.dc.b	$F8,$71,$FD,$9A,$3C,$D5,$EA,$3B,$5F,$82,$7F,$AB,$DF,$33,$F0,$77
	.dc.b	$CA,$7F,$8B,$DF,$EB,$AB,$F4,$CF,$EF,$8F,$21,$FA,$4B,$FE,$12,$FE
	.dc.b	$0E,$F9,$DF,$E2,$E3,$E4,$CF,$89,$B5,$FE,$48,$7E,$59,$F8,$13,$F1
	.dc.b	$93,$F8,$17,$A1,$F8,$17,$A3,$C2,$7A,$36,$7A,$47,$C1,$7D,$F1,$F5
	.dc.b	$67,$D1,$1F,$2A,$3D,$11,$F0,$5F,$F1,$0D,$FF,$04,$7F,$C3,$A7,$FA
	.dc.b	$E2,$FE,$43,$E3,$9F,$B1,$3F,$36,$3F,$A0,$1C,$8F,$90,$7B,$6F,$E3
	.dc.b	$9F,$92,$7F,$51,$3E,$C8,$F4,$47,$23,$FC,$AD,$3E,$3A,$F0,$C7,$E3
	.dc.b	$2B,$D0,$E8,$EF,$5F,$10,$E6,$FE,$28,$F8,$53,$FC,$41,$F0,$CF,$A9
	.dc.b	$FA,$A7,$84,$FF,$2C,$1F,$83,$AF,$E3,$2F,$23,$F3,$C7,$23,$F5,$57
	.dc.b	$F9,$3B,$F1,$34,$FA,$A5,$8F,$D4,$9F,$80,$AF,$D4,$BF,$80,$3D,$CF
	.dc.b	$C5,$8E,$11,$FC,$DC,$F8,$07,$96,$3F,$91,$47,$E0,$23,$A4,$7E,$02
	.dc.b	$FF,$2A,$FB,$E3,$F1,$07,$BC,$FF,$34,$9C,$DF,$E2,$B5,$FC,$B2,$3E
	.dc.b	$51,$D2,$7E,$53,$FB,$39,$F3,$5F,$EA,$AB,$FA,$CD,$FE,$AA,$BA,$57
	.dc.b	$F1,$47,$E5,$1E,$F3,$E5,$1F,$BA,$8F,$7B,$F9,$23,$CB,$3D,$2B,$E9
	.dc.b	$C7,$8C,$FA,$67,$8E,$BD,$D3,$F8,$28,$E4,$7F,$79,$3C,$67,$EF,$2F
	.dc.b	$F1,$AF,$8A,$3E,$CD,$7A,$EF,$FA,$F5,$FE,$42,$BF,$48,$7A,$AB,$F3
	.dc.b	$8F,$84,$FC,$1D,$71,$9F,$EA,$E7,$C6,$8F,$ED,$E7,$1A,$FD,$A5,$7F
	.dc.b	$62,$9D,$4F,$DE,$8F,$81,$1E,$18,$FF,$76,$5F,$A1,$F0,$1F,$26,$FC
	.dc.b	$13,$FD,$44,$E7,$1E,$83,$E4,$1F,$44,$E3,$5F,$01,$FD,$D1,$F1,$8E
	.dc.b	$87,$A0,$E2,$FC,$07,$F8,$5C,$76,$7E,$BD,$FC,$88,$C4,$7F,$89,$4F
	.dc.b	$81,$7F,$94,$7E,$39,$E0,$AF,$7C,$F0,$7F,$C4,$11,$C0,$FE,$24,$70
	.dc.b	$3E,$35,$77,$9F,$C0,$BF,$8A,$AF,$8D,$3E,$69,$CD,$FC,$67,$FA,$8C
	.dc.b	$7B,$8F,$F7,$34,$F0,$BF,$A9,$3C,$27,$CD,$7D,$DF,$11,$F3,$67,$F8
	.dc.b	$FC,$7D,$C3,$FB,$7C,$75,$3F,$86,$C7,$BA,$7A,$47,$80,$F7,$1F,$C0
	.dc.b	$9C,$BC,$27,$C0,$F3,$3F,$06,$70,$3F,$34,$FF,$10,$BF,$71,$FC,$DE
	.dc.b	$3F,$85,$1E,$78,$FF,$58,$5F,$EF,$23,$E3,$1E,$C3,$ED,$D7,$E6,$8F
	.dc.b	$1B,$E0,$3F,$07,$3E,$F9,$FF,$00,$47,$F2,$C3,$B9,$CA,$7F,$28,$7F
	.dc.b	$03,$78,$C7,$E0,$DF,$E9,$13,$E4,$7F,$7A,$79,$0F,$80,$7E,$EE,$FE
	.dc.b	$53,$FB,$B3,$E1,$1F,$9C,$8F,$E4,$91,$F5,$27,$B3,$F9,$F9,$FC,$4D
	.dc.b	$7B,$5E,$47,$90,$FC,$79,$E4,$7F,$BB,$47,$E8,$4F,$C2,$C7,$C7,$3C
	.dc.b	$95,$F5,$A3,$F9,$54,$7F,$72,$9E,$33,$F1,$DF,$E7,$31,$FD,$12,$7C
	.dc.b	$CF,$81,$F3,$3F,$81,$8F,$33,$E3,$7F,$A4,$C7,$8E,$7F,$1E,$F8,$2B
	.dc.b	$F3,$51,$83,$F0,$95,$F9,$AA,$E7,$5B,$8F,$92,$7C,$F7,$0F,$C3,$3F
	.dc.b	$28,$7B,$DF,$F4,$B1,$F4,$9F,$24,$7F,$51,$7D,$53,$F4,$8F,$B8,$FF
	.dc.b	$11,$3B,$3E,$47,$E9,$C7,$E1,$9F,$A7,$3F,$11,$F0,$9D,$AB,$F3,$11
	.dc.b	$F4,$A3,$F0,$AF,$92,$BE,$91,$E1,$AF,$A3,$5F,$82,$8C,$47,$CA,$7E
	.dc.b	$A3,$FC,$45,$FE,$B5,$3E,$D3,$E1,$9E,$3A,$FD,$79,$F6,$63,$DB,$1F
	.dc.b	$4E,$BF,$47,$3F,$DC,$0F,$D1,$9F,$C8,$CF,$98,$72,$3F,$B7,$C7,$57
	.dc.b	$F3,$47,$F7,$73,$F1,$B5,$F6,$2B,$D2,$FC,$F8,$F8,$B3,$E6,$9F,$19
	.dc.b	$EC,$9F,$DC,$CF,$27,$E7,$3F,$2C,$FD,$7C,$FC,$B3,$94,$7C,$88,$FB
	.dc.b	$67,$8E,$3E,$74,$78,$E3,$E1,$3B,$7C,$73,$E6,$AF,$AE,$FE,$4E,$39
	.dc.b	$3F,$38,$FA,$EF,$29,$FA,$93,$F5,$E3,$11,$EF,$7E,$41,$FD,$FA,$BC
	.dc.b	$71,$EC,$7D,$2F,$28,$F4,$9F,$3A,$BE,$2D,$75,$8F,$AB,$5C,$AB,$FA
	.dc.b	$2B,$C9,$F0,$47,$D4,$39,$3E,$F7,$C7,$7F,$E6,$B3,$FC,$B4,$FE,$B2
	.dc.b	$7E,$65,$FC,$38,$F9,$EF,$C3,$7C,$2F,$C3,$AF,$99,$1D,$9F,$53,$F3
	.dc.b	$0F,$87,$7F,$30,$FC,$7D,$7F,$88,$6F,$C3,$3F,$8E,$3D,$07,$F1,$3A
	.dc.b	$E6,$FA,$0F,$F3,$5D,$FE,$CD,$FE,$8D,$1E,$D7,$8D,$FF,$0C,$3B,$CF
	.dc.b	$F8,$E2,$BB,$DF,$E6,$6B,$F7,$A7,$98,$F7,$D7,$09,$FE,$AB,$1F,$D1
	.dc.b	$CF,$25,$FC,$E7,$89,$F0,$5C,$57,$CA,$8F,$C4,$CF,$37,$FC,$2D,$1E
	.dc.b	$13,$4F,$1A,$F8,$D5,$FE,$2B,$8F,$D1,$3F,$7A,$FE,$A4,$7F,$15,$AE
	.dc.b	$2F,$92,$B8,$D7,$F0,$5B,$F0,$4F,$F7,$93,$E7,$1E,$83,$81,$F0,$23
	.dc.b	$C1,$5F,$8D,$9F,$EE,$93,$A3,$C1,$7F,$83,$71,$3F,$C2,$EB,$8C,$FE
	.dc.b	$15,$FC,$F3,$E8,$8D,$47,$F9,$EA,$30,$7F,$01,$9F,$E7,$0F,$18,$F8
	.dc.b	$11,$F5,$4F,$0D,$FF,$9B,$AB,$E5,$57,$8E,$FF,$39,$5F,$86,$38,$4F
	.dc.b	$82,$FF,$BE,$3C,$2F,$E2,$CF,$F5,$99,$F3,$BC,$2F,$C8,$FF,$11,$3C
	.dc.b	$31,$EE,$BE,$31,$FE,$0F,$BE,$93,$FD,$A2,$3E,$E1,$F3,$5F,$A8,$71
	.dc.b	$34,$FC,$AA,$EF,$3F,$BB,$BF,$B7,$5F,$C9,$23,$EA,$57,$E4,$E7,$F3
	.dc.b	$D3,$C9,$AF,$62,$1F,$02,$BF,$85,$CF,$F0,$E3,$C1,$3F,$C3,$5F,$E2
	.dc.b	$D1,$F4,$6F,$F0,$33,$D2,$B9,$1F,$AF,$AF,$C5,$4F,$D0,$7F,$CE,$97
	.dc.b	$E9,$7E,$04,$F0,$3F,$76,$FE,$FA,$78,$4F,$F5,$9B,$FC,$1D,$F8,$CF
	.dc.b	$05,$7E,$12,$BE,$7B,$FD,$56,$BE,$7D,$FE,$2E,$3E,$C5,$7A,$EB,$AD
	.dc.b	$FF,$38,$9F,$A6,$7D,$AA,$E0,$FF,$B2,$81,$FE,$02,$30,$60,$CD,$22
	.dc.b	$19,$88,$22,$C6,$9C,$9A,$8E,$8F,$28,$C4,$78,$1E,$0F,$FB,$A4,$E2
	.dc.b	$F8,$1D,$C7,$68,$FF,$7B,$C7,$A0,$EA,$ED,$F0,$46,$E3,$CD,$1D,$4F
	.dc.b	$54,$78,$DE,$A7,$B6,$3C,$6F,$A0,$F0,$BD,$5F,$69,$B7,$C6,$F6,$3C
	.dc.b	$2E,$DF,$03,$D1,$F4,$1F,$39,$D3,$DC,$FA,$4E,$DE,$87,$C4,$7E,$A3
	.dc.b	$E8,$36,$F8,$A3,$E6,$C7,$FA,$80,$F3,$9A,$8E,$6C,$3E,$4B,$D1,$F9
	.dc.b	$27,$B5,$DC,$74,$3E,$44,$68,$FE,$5F,$7F,$CD,$A3,$F3,$71,$FC,$2E
	.dc.b	$3F,$8A,$3F,$90,$7F,$CE,$75,$F8,$F4,$FE,$14,$C1,$FE,$78,$8E,$0E
	.dc.b	$AB,$F8,$14,$7F,$7B,$8F,$E6,$D5,$E0,$AF,$DA,$4F,$E8,$6B,$9B,$F9
	.dc.b	$DA,$EB,$1F,$9C,$7B,$9F,$97,$9F,$1B,$FC,$C4,$FE,$17,$3F,$DF,$CF
	.dc.b	$D6,$C7,$B1,$F7,$47,$F1,$C3,$F4,$C7,$69,$FD,$53,$DA,$3F,$53,$3D
	.dc.b	$A7,$D7,$1F,$2A,$7D,$6F,$B6,$BC,$53,$FC,$4E,$BF,$B1,$D7,$58,$F9
	.dc.b	$35,$F1,$A3,$D4,$7C,$78,$FE,$03,$1F,$C8,$8F,$E4,$95,$F1,$E7,$F8
	.dc.b	$9C,$7F,$27,$7A,$1E,$77,$F8,$05,$79,$EB,$9D,$71,$9F,$E2,$67,$43
	.dc.b	$D1,$1F,$D0,$2B,$9C,$79,$4F,$82,$F9,$63,$F9,$41,$FE,$0C,$7E,$0C
	.dc.b	$7C,$39,$F8,$33,$FC,$B2,$3F,$BE,$4F,$45,$5F,$61,$EB,$1C,$27,$A5
	.dc.b	$53,$FA,$7C,$FF,$A7,$27,$EE,$BF,$E5,$A7,$A5,$73,$8F,$83,$1E,$77
	.dc.b	$85,$7D,$38,$F3,$CF,$F3,$68,$F5,$D7,$6A,$F2,$D7,$E1,$1F,$CD,$CF
	.dc.b	$E4,$5F,$88,$76,$8F,$33,$FD,$30,$EF,$1F,$06,$BF,$5D,$1F,$3A,$3E
	.dc.b	$EC,$FA,$0F,$D4,$47,$81,$E6,$FE,$8C,$E7,$1F,$10,$FC,$8C,$FA,$63
	.dc.b	$F5,$E7,$E9,$CF,$3C,$7E,$69,$F3,$9F,$A8,$78,$CF,$CA,$8F,$F2,$A9
	.dc.b	$E5,$8F,$86,$FA,$CF,$86,$75,$3E,$2B,$F0,$DF,$41,$F6,$9E,$A7,$CF
	.dc.b	$AF,$5D,$7E,$C4,$FD,$F9,$6B,$E7,$57,$4A,$F4,$3D,$0F,$9C,$F4,$3E
	.dc.b	$64,$7D,$A3,$D5,$1F,$E4,$9A,$FC,$F3,$FD,$A2,$3E,$71,$FA,$E3,$C6
	.dc.b	$7F,$01,$7F,$46,$7F,$07,$3F,$60,$57,$8B,$ED,$9F,$D5,$BF,$1C,$FA
	.dc.b	$B3,$F8,$9A,$F6,$D7,$59,$F1,$1F,$52,$BF,$80,$BE,$D3,$EA,$3D,$1F
	.dc.b	$D5,$3E,$F3,$F0,$06,$67,$F0,$11,$F5,$0E,$A7,$C9,$8E,$B1,$FA,$38
	.dc.b	$EB,$1F,$71,$FC,$61,$FC,$8A,$BF,$C3,$8F,$AC,$E8,$72,$AF,$E7,$CF
	.dc.b	$E2,$8F,$E3,$13,$E5,$9F,$81,$3F,$C5,$27,$E5,$47,$D2,$79,$4F,$E1
	.dc.b	$EB,$C9,$1C,$A3,$D9,$5F,$6A,$7F,$4F,$3F,$0A,$72,$7E,$2C,$FE,$2C
	.dc.b	$FC,$B9,$FE,$40,$FB,$8F,$A6,$7E,$66,$7E,$9D,$7B,$A3,$A5,$7F,$4E
	.dc.b	$7E,$4C,$FF,$1D,$AE,$71,$E5,$9F,$91,$3F,$97,$8F,$C3,$9F,$EE,$AA
	.dc.b	$F7,$57,$E1,$A3,$F2,$F5,$E8,$3F,$2F,$3F,$90,$3C,$55,$F6,$8F,$ED
	.dc.b	$B3,$EB,$3E,$71,$FC,$0E,$7E,$73,$EB,$AF,$9C,$F8,$4F,$B1,$3C,$9F
	.dc.b	$C1,$47,$A9,$E5,$3E,$33,$F1,$67,$B2,$3F,$2D,$1F,$A0,$3F,$96,$1E
	.dc.b	$33,$F2,$C7,$C3,$3E,$69,$F8,$A9,$F9,$AF,$F4,$B3,$E6,$9F,$C5,$0F
	.dc.b	$9B,$1F,$8A,$79,$9E,$49,$F1,$57,$C4,$9F,$C1,$1F,$BD,$8F,$C0,$1F
	.dc.b	$CA,$27,$C9,$5F,$8A,$79,$1E,$47,$EB,$3F,$E0,$09,$F6,$D7,$D6,$8E
	.dc.b	$31,$FC,$C2,$7F,$B3,$57,$8A,$7D,$51,$F5,$67,$E7,$C7,$F8,$72,$BF
	.dc.b	$C5,$73,$F3,$CF,$C4,$47,$D8,$79,$BE,$39,$F0,$BF,$59,$E6,$FF,$64
	.dc.b	$3C,$6F,$F1,$5A,$FB,$F1,$F3,$63,$73,$FA,$43,$C9,$1F,$CC,$DF,$0B
	.dc.b	$F3,$1F,$0B,$FB,$D3,$D7,$1F,$32,$BF,$25,$1F,$E0,$93,$C2,$7F,$1D
	.dc.b	$AF,$84,$FE,$E8,$F4,$1F,$A0,$3E,$13,$FB,$27,$F1,$E7,$C1,$9F,$EE
	.dc.b	$0F,$84,$E4,$7E,$10,$C1,$F8,$E3,$89,$E9,$78,$D7,$E3,$23,$8B,$F0
	.dc.b	$1F,$B9,$5F,$EC,$59,$FD,$84,$7F,$01,$3A,$9F,$D7,$0F,$04,$F3,$7C
	.dc.b	$0F,$C8,$3F,$14,$7C,$83,$4F,$17,$C0,$FE,$94,$F8,$35,$F9,$37,$D3
	.dc.b	$1C,$A7,$D2,$7D,$83,$F3,$53,$FC,$08,$FB,$B1,$F9,$69,$FE,$4B,$1F
	.dc.b	$D8,$23,$94,$FD,$69,$F8,$2F,$D9,$9E,$2F,$EE,$0F,$83,$1F,$B5,$3F
	.dc.b	$9C,$B8,$7F,$4A,$7C,$43,$FB,$31,$F7,$0F,$DA,$C7,$F3,$58,$FD,$21
	.dc.b	$EA,$8F,$C7,$3F,$B9,$9F,$2C,$F8,$AB,$F7,$0F,$ED,$8F,$E5,$95,$EC
	.dc.b	$7F,$A5,$57,$E7,$5F,$91,$5F,$46,$3B,$9E,$F9,$FE,$38,$7D,$87,$FC
	.dc.b	$0B,$1B,$79,$4F,$67,$DD,$1F,$65,$FA,$15,$CE,$BE,$BD,$6D,$FC,$F9
	.dc.b	$FC,$D9,$C3,$F6,$A7,$C7,$1E,$D8,$FD,$4B,$ED,$7F,$8E,$1E,$D8,$FE
	.dc.b	$5D,$5E,$18,$F8,$F3,$F2,$E7,$CF,$1F,$C0,$EB,$E3,$CF,$E0,$CF,$87
	.dc.b	$3F,$34,$F0,$B8,$39,$4F,$ED,$4E,$47,$91,$E4,$7A,$AB,$E3,$46,$CF
	.dc.b	$C1,$9F,$00,$FD,$33,$F9,$E7,$F0,$67,$CF,$3D,$31,$F9,$39,$F1,$BD
	.dc.b	$5F,$31,$EC,$8F,$C6,$4F,$F1,$58,$FC,$B4,$FA,$4E,$4E,$E3,$F9,$43
	.dc.b	$CA,$7C,$D1,$E3,$39,$4F,$CF,$8F,$49,$F5,$67,$91,$F8,$43,$93,$F6
	.dc.b	$CF,$8A,$7E,$A2,$39,$4F,$81,$F9,$71,$B3,$AC,$7C,$87,$0F,$8E,$7F
	.dc.b	$73,$3F,$DA,$27,$F0,$31,$F9,$03,$F4,$CF,$EA,$A3,$DE,$FB,$23,$F2
	.dc.b	$71,$F8,$B8,$FF,$26,$1F,$4C,$FD,$C4,$F9,$63,$8C,$78,$A7,$8C,$7F
	.dc.b	$79,$AF,$F1,$7C,$FD,$68,$FD,$6B,$DC,$C4,$FE,$50,$F2,$BF,$C6,$4F
	.dc.b	$BF,$5F,$C7,$23,$D4,$7F,$5C,$3E,$64,$FC,$69,$FC,$3C,$78,$A3,$FA
	.dc.b	$2C,$7F,$2A,$8F,$A2,$E8,$F0,$9F,$D0,$DF,$D9,$3E,$07,$F0,$B1,$F8
	.dc.b	$B3,$B4,$7C,$28,$F0,$C7,$C2,$8E,$87,$F4,$03,$F4,$0F,$D3,$7E,$CC
	.dc.b	$78,$9E,$47,$88,$FE,$98,$F8,$A7,$F1,$B3,$FA,$89,$FB,$67,$D3,$3F
	.dc.b	$93,$C7,$D3,$8F,$9B,$1F,$4C,$FE,$C4,$7A,$DF,$C6,$CF,$A1,$DC,$7A
	.dc.b	$0F,$E0,$C7,$C3,$7F,$04,$FF,$1E,$9F,$A4,$F8,$5F,$E1,$4F,$E2,$E3
	.dc.b	$9B,$EC,$3E,$11,$F9,$D7,$D4,$FC,$27,$E5,$BE,$A9,$F2,$1F,$83,$3E
	.dc.b	$11,$F9,$CA,$FC,$5C,$77,$8D,$4F,$F7,$DA,$FD,$24,$FA,$EB,$FC,$F8
	.dc.b	$7F,$8A,$DF,$DA,$1F,$91,$39,$1E,$F9,$FC,$C3,$FE,$59,$3E,$4C,$A9
	.dc.b	$5F,$91,$9F,$B1,$3F,$D3,$AB,$F8,$51,$F0,$A3,$D8,$72,$AF,$53,$D4
	.dc.b	$F0,$9F,$80,$8F,$54,$7E,$16,$38,$9F,$28,$FA,$4F,$E5,$1F,$C1,$9E
	.dc.b	$C3,$F0,$07,$49,$F2,$4F,$E8,$A3,$FC,$0D,$3C,$5F,$A5,$1F,$99,$7E
	.dc.b	$C1,$F8,$23,$C8,$FC,$47,$FC,$7C,$7A,$EB,$E1,$3D,$E7,$F9,$8B,$E4
	.dc.b	$AF,$A6,$F4,$9F,$A6,$FF,$7B,$3B,$47,$99,$E1,$3F,$06,$3F,$1B,$1F
	.dc.b	$C0,$DF,$D1,$D7,$EA,$DF,$D0,$4F,$E0,$A3,$E8,$B5,$F8,$13,$FA,$17
	.dc.b	$F2,$33,$C2,$3F,$01,$1F,$BD,$3E,$13,$F4,$0F,$04,$FD,$87,$E0,$9E
	.dc.b	$F7,$F9,$EB,$F8,$27,$E0,$BF,$D3,$9E,$13,$FE,$29,$7F,$8C,$CF,$DE
	.dc.b	$3F,$43,$1E,$EA,$F8,$B5,$FD,$74,$EC,$7E,$5C,$F8,$F5,$DA,$7F,$A0
	.dc.b	$4F,$E9,$8F,$A7,$3A,$9F,$2B,$C8,$FC,$29,$EC,$8E,$15,$E3,$3E,$0D
	.dc.b	$7D,$0A,$F3,$C7,$E0,$63,$FC,$85,$3F,$AB,$7E,$ED,$7C,$7A,$DD,$7D
	.dc.b	$09,$FA,$F5,$FB,$8A,$F2,$1E,$BA,$DB,$FD,$88,$FC,$ED,$7F,$16,$7E
	.dc.b	$64,$7A,$9E,$27,$AA,$7E,$3D,$7F,$14,$8F,$D1,$1F,$C6,$E7,$AB,$F4
	.dc.b	$E7,$DF,$5F,$86,$8F,$C0,$9F,$AA,$07,$FB,$04,$B3,$86,$15,$44,$33
	.dc.b	$1A,$8B,$19,$72,$6A,$39,$47,$17,$07,$F0,$F3,$BB,$C5,$E8,$6A,$3C
	.dc.b	$11,$C2,$38,$1E,$77,$83,$E7,$38,$3D,$DF,$14,$79,$A3,$D6,$F9,$A3
	.dc.b	$71,$EE,$7C,$07,$91,$DB,$E4,$39,$BD,$5E,$47,$8D,$DB,$E3,$36,$F5
	.dc.b	$7C,$67,$57,$D6,$FD,$23,$C6,$FB,$E3,$D0,$EA,$21,$F3,$1E,$8F,$88
	.dc.b	$FB,$AF,$C4,$7A,$1E,$27,$A3,$E1,$8E,$0F,$DB,$8F,$63,$EF,$39,$BE
	.dc.b	$C7,$99,$F7,$9D,$BF,$9A,$3F,$40,$7F,$69,$8F,$F0,$F9,$F9,$6A,$F7
	.dc.b	$FE,$09,$F8,$F7,$FA,$85,$7E,$36,$31,$27,$F1,$53,$FD,$55,$FE,$03
	.dc.b	$1C,$CF,$ED,$75,$F9,$E3,$F9,$FD,$7F,$55,$AF,$E5,$B5,$D2,$3B,$4F
	.dc.b	$E8,$A7,$A3,$F9,$C7,$B4,$7F,$07,$39,$57,$AA,$7D,$8F,$89,$B1,$F8
	.dc.b	$F9,$F6,$C7,$59,$FC,$CD,$7B,$0F,$5B,$ED,$AE,$CF,$F3,$B9,$FE,$8B
	.dc.b	$1F,$CA,$23,$C1,$1F,$C0,$0F,$EB,$33,$EB,$9F,$DA,$BD,$63,$D5,$3D
	.dc.b	$67,$F8,$14,$7F,$5E,$7F,$A1,$1E,$C8,$F5,$47,$53,$FA,$24,$F5,$9F
	.dc.b	$A5,$1E,$B8,$F4,$9C,$2B,$F9,$29,$FC,$AE,$3E,$1C,$F8,$23,$FB,$04
	.dc.b	$1F,$8A,$DA,$F8,$AF,$53,$D3,$5D,$63,$FB,$31,$F0,$A7,$BB,$4C,$9F
	.dc.b	$D1,$23,$D0,$7C,$39,$FE,$37,$3F,$72,$BC,$95,$FD,$34,$FC,$4C,$FC
	.dc.b	$53,$E8,$C7,$F4,$68,$FF,$3B,$D7,$49,$F3,$1D,$23,$E7,$4F,$CC,$7E
	.dc.b	$D4,$7E,$49,$B5,$FD,$CE,$3B,$1F,$04,$FE,$61,$5D,$5F,$88,$FC,$EA
	.dc.b	$FC,$2C,$7E,$76,$79,$1F,$30,$B5,$EB,$8E,$67,$E4,$67,$F3,$EF,$43
	.dc.b	$E6,$BD,$1F,$1C,$F2,$9F,$EA,$47,$EB,$63,$99,$F5,$E7,$9C,$FE,$A2
	.dc.b	$3E,$15,$7D,$79,$FC,$C5,$57,$E5,$35,$FC,$81,$FA,$77,$FA,$7C,$7C
	.dc.b	$1A,$E9,$5F,$A7,$AF,$34,$7B,$EB,$EB,$B6,$7F,$88,$1F,$E0,$0A,$E5
	.dc.b	$3F,$98,$8F,$2D,$72,$8F,$8D,$1F,$E0,$73,$FB,$51,$F9,$88,$FE,$22
	.dc.b	$FC,$0A,$FE,$2B,$1F,$AE,$8F,$C0,$9E,$57,$F1,$73,$F2,$23,$91,$F8
	.dc.b	$A8,$F7,$9F,$D0,$67,$F4,$27,$47,$CC,$FC,$08,$FD,$74,$79,$E7,$A4
	.dc.b	$7E,$D6,$7F,$5C,$D8,$FC,$AB,$F7,$E7,$CC,$FD,$F3,$E0,$BE,$67,$F9
	.dc.b	$A9,$C6,$3E,$A1,$F9,$63,$E9,$BC,$E3,$CB,$1F,$2E,$3E,$04,$7E,$7D
	.dc.b	$FC,$A4,$7D,$43,$E6,$BF,$E1,$A3,$FD,$61,$5F,$50,$FC,$E4,$F8,$E3
	.dc.b	$FC,$6F,$3F,$8F,$7E,$A1,$FC,$78,$F1,$4F,$EE,$DF,$09,$F3,$63,$D0
	.dc.b	$FE,$08,$FB,$67,$B6,$7F,$1C,$FD,$B8,$FE,$1E,$FC,$53,$99,$FB,$73
	.dc.b	$D6,$73,$3F,$D0,$EF,$95,$F8,$71,$FA,$28,$F8,$91,$F6,$23,$D8,$F8
	.dc.b	$AB,$F8,$BC,$FC,$E8,$EA,$7F,$8C,$AB,$E2,$3F,$B4,$AE,$2F,$80,$FC
	.dc.b	$59,$E5,$3F,$6F,$3F,$59,$FD,$A4,$7E,$02,$BF,$15,$1F,$C2,$A3,$F1
	.dc.b	$4F,$E0,$23,$C7,$18,$8F,$D4,$1F,$38,$FC,$E9,$EB,$7E,$F3,$E0,$3E
	.dc.b	$21,$FE,$63,$3C,$B1,$F9,$C9,$F9,$AF,$88,$FE,$61,$3C,$9F,$DC,$4F
	.dc.b	$DE,$39,$1F,$5E,$3F,$4D,$3E,$D9,$FD,$35,$7E,$06,$BC,$87,$E0,$4F
	.dc.b	$96,$7E,$3C,$F9,$07,$3A,$FD,$DB,$F3,$CF,$71,$F8,$13,$99,$F3,$67
	.dc.b	$CA,$F3,$3C,$AF,$C8,$8F,$2C,$7E,$2D,$FC,$14,$FE,$71,$F7,$57,$E8
	.dc.b	$8F,$C4,$1F,$89,$8F,$BF,$3F,$B4,$9D,$CF,$EF,$AB,$B4,$62,$7F,$D4
	.dc.b	$95,$C1,$FC,$BC,$FF,$37,$7E,$E5,$7F,$62,$AF,$B9,$1F,$B0,$3D,$F3
	.dc.b	$F4,$E7,$FA,$BC,$7C,$93,$EE,$1F,$4E,$3F,$86,$9E,$27,$FC,$34,$7B
	.dc.b	$E7,$F1,$D3,$FA,$08,$FC,$74,$7E,$61,$F8,$F1,$E3,$8C,$4F,$F8,$CA
	.dc.b	$7C,$73,$C6,$3F,$0B,$1E,$37,$E8,$C7,$C8,$8F,$C6,$BF,$83,$3F,$1A
	.dc.b	$71,$3D,$6F,$D9,$7E,$83,$C0,$FC,$BC,$78,$5F,$D3,$CF,$F1,$CA,$FB
	.dc.b	$11,$E1,$AE,$2F,$E2,$4C,$3F,$EB,$58,$FA,$51,$F8,$8A,$FA,$4F,$F1
	.dc.b	$A3,$C0,$7C,$B9,$FE,$6A,$F8,$27,$F3,$4F,$C8,$AF,$BC,$F1,$3C,$07
	.dc.b	$AE,$7C,$0F,$C8,$9F,$3B,$93,$F0,$33,$FD,$C0,$FC,$0B,$FC,$A4,$F0
	.dc.b	$1F,$BC,$3C,$07,$B0,$E2,$FF,$76,$38,$BF,$E4,$23,$C0,$7F,$15,$7C
	.dc.b	$EF,$F1,$D9,$FB,$D1,$88,$F0,$57,$EE,$1F,$03,$F7,$9F,$41,$EF,$7F
	.dc.b	$1C,$E0,$FC,$C1,$E8,$3F,$6C,$7A,$0F,$F1,$41,$FB,$B3,$4F,$E7,$6B
	.dc.b	$F9,$74,$7E,$70,$FD,$D4,$7E,$04,$F2,$3D,$1F,$21,$FC,$66,$7F,$77
	.dc.b	$1E,$F3,$E3,$1E,$7A,$F7,$BF,$A4,$3C,$53,$D0,$F1,$4F,$F5,$08,$CC
	.dc.b	$FE,$62,$BE,$CD,$7E,$22,$7E,$55,$7D,$93,$F8,$6B,$C1,$F1,$BF,$C1
	.dc.b	$CF,$01,$FD,$49,$FD,$9C,$F4,$AF,$EE,$47,$F0,$B3,$CF,$3F,$20,$E1
	.dc.b	$3F,$A1,$7D,$C6,$67,$F3,$AF,$CF,$AF,$33,$FC,$B1,$FA,$E7,$C7,$9F
	.dc.b	$0C,$72,$8C,$47,$58,$F4,$1F,$74,$E5,$5E,$49,$DD,$7E,$91,$FA,$71
	.dc.b	$FB,$FA,$F8,$35,$FC,$82,$3D,$AF,$E6,$23,$FC,$5A,$7F,$3E,$7D,$B5
	.dc.b	$CE,$7F,$82,$D7,$69,$FC,$9C,$60,$E6,$FD,$DA,$FA,$47,$3A,$F8,$75
	.dc.b	$E4,$9F,$E4,$53,$ED,$7C,$87,$F0,$EA,$FF,$4A,$47,$28,$FA,$F5,$CA
	.dc.b	$7C,$11,$CA,$3C,$C7,$E6,$2B,$F9,$0C,$7F,$0A,$3D,$67,$A2,$3F,$D3
	.dc.b	$CF,$ED,$0F,$9B,$1F,$59,$FE,$63,$5E,$17,$EE,$1B,$3C,$E7,$E3,$5F
	.dc.b	$B2,$F2,$AF,$86,$FE,$52,$3F,$82,$1F,$94,$7C,$67,$EA,$63,$D6,$F3
	.dc.b	$8F,$88,$FF,$25,$8E,$6F,$E0,$23,$C9,$3F,$4E,$39,$57,$B1,$FD,$DD
	.dc.b	$7F,$3D,$8F,$88,$FF,$34,$9F,$C8,$BF,$85,$AF,$4D,$7B,$0F,$2C,$FE
	.dc.b	$CA,$3F,$A8,$CF,$D4,$7D,$67,$DD,$9F,$88,$79,$CF,$A8,$7C,$F7,$F9
	.dc.b	$1C,$FF,$64,$9F,$CC,$1E,$18,$F7,$CF,$E9,$E3,$15,$F7,$27,$8C,$78
	.dc.b	$5F,$CB,$CF,$E1,$0F,$09,$C4,$F0,$9F,$11,$E3,$1F,$08,$FE,$87,$1F
	.dc.b	$1A,$7F,$80,$3C,$2B,$E0,$57,$F3,$38,$FC,$1B,$F0,$1F,$01,$C4,$FD
	.dc.b	$C5,$7A,$1F,$81,$1F,$22,$3C,$11,$FE,$12,$3C,$F3,$FA,$37,$F0,$15
	.dc.b	$EF,$AE,$07,$A0,$FE,$45,$1F,$52,$BE,$E4,$62,$7F,$48,$76,$9F,$D9
	.dc.b	$47,$9D,$FC,$C9,$E7,$3F,$2C,$7B,$E7,$9C,$78,$08,$7D,$D8,$E0,$7E
	.dc.b	$16,$7B,$C7,$94,$DC,$7C,$4A,$FE,$75,$5D,$27,$94,$62,$BE,$7D,$7F
	.dc.b	$7D,$AF,$43,$FD,$62,$74,$FC,$C3,$CD,$3C,$1F,$EA,$73,$F8,$33,$FD
	.dc.b	$32,$FC,$A3,$E8,$D7,$82,$3F,$A4,$47,$33,$C1,$1E,$93,$47,$F2,$67
	.dc.b	$F3,$47,$33,$F1,$4F,$B9,$FE,$64,$7F,$7D,$3E,$29,$C0,$FA,$D3,$E7
	.dc.b	$9F,$F1,$94,$7F,$85,$2B,$F7,$07,$E3,$9F,$91,$1E,$67,$EA,$9E,$6A
	.dc.b	$F0,$4F,$09,$F9,$93,$C2,$7D,$0F,$E8,$4E,$0F,$E6,$67,$FB,$C4,$F5
	.dc.b	$9F,$B4,$7E,$49,$FE,$22,$FC,$C7,$E9,$BF,$E1,$27,$83,$ED,$AF,$C0
	.dc.b	$D7,$92,$7F,$8A,$CF,$C3,$AF,$B2,$FA,$8F,$86,$7D,$13,$B4,$FD,$17
	.dc.b	$D4,$FF,$12,$7B,$1E,$5A,$F1,$BF,$B2,$8F,$E0,$2F,$E7,$27,$BB,$EC
	.dc.b	$3E,$99,$F0,$CE,$F1,$E7,$7F,$16,$78,$E7,$D7,$1F,$1E,$BF,$19,$3F
	.dc.b	$EA,$49,$CB,$E6,$9F,$E2,$F5,$FB,$98,$EB,$1C,$6B,$F9,$94,$F9,$27
	.dc.b	$D6,$79,$1E,$11,$F8,$47,$F9,$9C,$79,$0F,$91,$46,$16,$84,$8C,$C7
	.dc.b	$F8,$EC,$A7,$F9,$F0,$AE,$23,$89,$FE,$F0,$77,$1D,$9C,$C6,$23,$A1
	.dc.b	$FF,$00,$0A,$6D,$E6,$EE,$3B,$47,$96,$3F,$DD,$C7,$89,$DB,$E0,$8D
	.dc.b	$C7,$8E,$39,$3A,$7D,$27,$B9,$F4,$C6,$23,$94,$6D,$EE,$79,$CF,$F5
	.dc.b	$71,$F2,$E3,$DA,$F9,$CE,$91,$E7,$7E,$11,$E2,$7E,$7B,$B3,$E9,$BE
	.dc.b	$27,$B9,$F0,$DD,$BE,$D3,$AB,$ED,$7C,$EE,$9F,$51,$F7,$5F,$13,$F3
	.dc.b	$8F,$43,$ED,$7C,$A7,$89,$DB,$EE,$3F,$00,$7F,$53,$3F,$8C,$BF,$C7
	.dc.b	$8F,$D7,$3F,$A5,$8F,$CE,$BF,$E5,$DF,$C1,$3F,$1F,$4B,$FC,$1C,$61
	.dc.b	$3F,$88,$8D,$3F,$A0,$8F,$E0,$F3,$D4,$FD,$D5,$FF,$57,$6B,$FE,$23
	.dc.b	$9E,$8F,$E5,$1E,$C7,$59,$EF,$1C,$67,$F5,$51,$FC,$62,$FD,$37,$FC
	.dc.b	$95,$FE,$8D,$7D,$2F,$F4,$EF,$4B,$FD,$2D,$7A,$27,$A3,$FB,$78,$F4
	.dc.b	$1D,$AB,$D9,$3F,$CB,$A7,$F8,$05,$75,$8F,$87,$5F,$B7,$9F,$86,$F6
	.dc.b	$BF,$4C,$FF,$90,$6F,$CF,$1F,$56,$BF,$A1,$57,$F1,$B8,$F9,$31,$DA
	.dc.b	$3A,$47,$D2,$6D,$FC,$53,$CB,$7F,$72,$BF,$8A,$5F,$6A,$FE,$18,$7F
	.dc.b	$8F,$6B,$D0,$7C,$37,$F8,$4C,$FC,$3A,$F3,$47,$D6,$7F,$10,$7F,$20
	.dc.b	$BF,$A1,$19,$3F,$D5,$75,$D1,$F2,$9F,$05,$F2,$CF,$1B,$A7,$F8,$7D
	.dc.b	$EA,$D9,$E6,$7C,$08,$FC,$23,$D0,$F3,$D7,$CC,$8E,$31,$FE,$00,$3F
	.dc.b	$9B,$4F,$37,$E6,$47,$D9,$9F,$EB,$F5,$E7,$95,$FD,$87,$AC,$79,$9E
	.dc.b	$87,$2A,$F9,$D5,$F0,$A3,$F0,$D1,$FE,$36,$9E,$95,$CE,$7F,$37,$3F
	.dc.b	$8E,$AF,$BA,$71,$9F,$C8,$1F,$62,$7E,$61,$FB,$33,$E1,$4F,$E6,$9F
	.dc.b	$84,$74,$7E,$BB,$E2,$6C,$74,$9E,$C7,$82,$FF,$90,$D7,$53,$E0,$C7
	.dc.b	$E3,$5B,$7F,$8D,$BF,$41,$F2,$6B,$E2,$4F,$F0,$C3,$E7,$9F,$1A,$FF
	.dc.b	$84,$3F,$D6,$2B,$E1,$DF,$E9,$A3,$E6,$DF,$43,$E5,$DF,$4B,$FD,$19
	.dc.b	$FB,$B9,$FF,$04,$5F,$23,$C9,$5E,$EB,$F6,$3C,$CF,$C7,$16,$FF,$05
	.dc.b	$7E,$4B,$FE,$D9,$7C,$DF,$CB,$56,$DF,$55,$FE,$D4,$FF,$17,$BF,$81
	.dc.b	$8E,$A7,$F1,$97,$EA,$BE,$DA,$FC,$04,$7E,$DA,$3A,$3F,$1C,$E5,$7E
	.dc.b	$4A,$EC,$FF,$18,$8E,$77,$FB,$57,$EF,$57,$E7,$9E,$57,$F1,$5F,$F0
	.dc.b	$39,$F8,$87,$DC,$F8,$EF,$EE,$47,$92,$3E,$EC,$F9,$23,$E7,$CF,$8A
	.dc.b	$7F,$C4,$87,$37,$FC,$1B,$5F,$1E,$3E,$CD,$7D,$13,$F0,$87,$39,$F8
	.dc.b	$F5,$E5,$9F,$E2,$07,$B0,$FC,$C9,$F1,$67,$91,$F8,$23,$95,$FF,$9F
	.dc.b	$CF,$E3,$51,$F6,$0E,$47,$E5,$8F,$B3,$1C,$8F,$AA,$7C,$79,$F6,$C7
	.dc.b	$F2,$23,$FC,$ED,$7E,$B7,$E6,$9F,$9A,$3E,$6D,$FC,$37,$E6,$9F,$AD
	.dc.b	$AF,$EC,$4F,$E7,$4F,$1D,$7C,$D3,$F9,$54,$78,$67,$E1,$9C,$E7,$FC
	.dc.b	$AD,$3F,$B5,$BF,$D9,$BF,$62,$FC,$93,$F1,$6F,$95,$7C,$38,$FA,$87
	.dc.b	$AC,$FF,$2F,$4F,$07,$EE,$CF,$C3,$9F,$C7,$46,$EB,$F8,$73,$EB,$3F
	.dc.b	$97,$9F,$C0,$A7,$F9,$7B,$F3,$4F,$BE,$F8,$9F,$C2,$9C,$4F,$CD,$9F
	.dc.b	$A8,$3E,$21,$CD,$FE,$30,$F3,$3E,$79,$C0,$F9,$E7,$96,$FF,$8B,$C7
	.dc.b	$E3,$0F,$D7,$47,$37,$F3,$97,$FC,$09,$F2,$47,$D6,$AF,$99,$1F,$8A
	.dc.b	$8F,$E1,$27,$D6,$AF,$E2,$4F,$8A,$7F,$2C,$FF,$9D,$9E,$33,$F2,$63
	.dc.b	$D8,$7E,$44,$F1,$57,$ED,$5F,$AC,$7E,$26,$7F,$1E,$FF,$B2,$5A,$FF
	.dc.b	$A8,$8F,$F5,$24,$7D,$78,$FC,$3C,$71,$AF,$C3,$BF,$0E,$FF,$92,$46
	.dc.b	$23,$C5,$3F,$E3,$5B,$FA,$EF,$E4,$67,$FB,$24,$F3,$AF,$C4,$3F,$5E
	.dc.b	$7D,$EF,$E2,$0F,$89,$7F,$2D,$E0,$7D,$E8,$F8,$77,$F3,$A3,$F7,$67
	.dc.b	$C2,$3F,$76,$7E,$41,$D3,$F2,$EF,$F9,$AC,$7C,$B8,$FB,$E7,$AA,$3F
	.dc.b	$8A,$9F,$51,$F9,$67,$E1,$67,$E4,$DF,$F0,$28,$F9,$47,$8A,$3E,$15
	.dc.b	$7E,$39,$F8,$57,$F2,$2B,$E0,$C7,$AA,$BD,$A7,$D2,$3E,$09,$FC,$2A
	.dc.b	$BD,$2F,$F9,$0A,$38,$3E,$3B,$FA,$4F,$F2,$D9,$FA,$31,$E9,$8F,$84
	.dc.b	$76,$3F,$7E,$FC,$28,$EB,$3F,$E3,$D9,$F9,$47,$E9,$4F,$82,$FF,$03
	.dc.b	$7C,$A7,$EC,$8E,$87,$CD,$8F,$D3,$1F,$C5,$1F,$98,$F9,$2F,$EE,$47
	.dc.b	$F7,$F9,$F8,$F3,$FA,$A8,$F8,$07,$43,$E0,$57,$9D,$FE,$0B,$3C,$2B
	.dc.b	$F4,$11,$EE,$3A,$4F,$DA,$8E,$F3,$F8,$E3,$F8,$79,$F2,$8F,$64,$FC
	.dc.b	$07,$EB,$D7,$F8,$A6,$3D,$E7,$ED,$6B,$F9,$43,$EF,$8F,$E8,$C7,$F0
	.dc.b	$49,$F2,$47,$E4,$9F,$D5,$BE,$48,$F8,$4F,$B8,$FB,$57,$F8,$4A,$F7
	.dc.b	$46,$E7,$F1,$51,$F6,$EB,$99,$E8,$3B,$CF,$D7,$3E,$13,$EF,$8F,$E3
	.dc.b	$33,$E6,$3E,$A4,$FE,$1D,$EB,$7F,$1A,$7F,$B6,$CF,$C0,$3F,$02,$FC
	.dc.b	$07,$E9,$CF,$DE,$7E,$9C,$7C,$F7,$D9,$5E,$58,$C4,$7E,$1C,$F7,$9F
	.dc.b	$BB,$7D,$F1,$B9,$F8,$B3,$B8,$C4,$79,$A3,$AC,$F8,$E7,$ED,$BC,$8F
	.dc.b	$4C,$7D,$B7,$E4,$1F,$C4,$9F,$99,$3F,$8F,$7B,$47,$33,$E3,$C7,$F7
	.dc.b	$88,$F6,$BE,$47,$F0,$47,$F4,$C9,$FA,$27,$E2,$23,$F8,$63,$FC,$1A
	.dc.b	$3F,$B2,$9F,$A2,$8F,$53,$E0,$39,$4F,$C6,$7A,$9F,$CD,$DF,$DE,$DF
	.dc.b	$F0,$FB,$FE,$6D,$1E,$BB,$F8,$B7,$EB,$BE,$27,$F1,$43,$F2,$CF,$AD
	.dc.b	$F8,$35,$FA,$89,$FC,$5D,$7C,$87,$D6,$FC,$88,$FE,$31,$3F,$5E,$BF
	.dc.b	$00,$F2,$3F,$3D,$3E,$33,$91,$EB,$7F,$12,$7F,$49,$8F,$EC,$4F,$F1
	.dc.b	$F3,$EB,$4F,$E4,$5F,$87,$1F,$3D,$E2,$7C,$3B,$FD,$34,$78,$AF,$E0
	.dc.b	$D7,$AE,$FC,$57,$CA,$BE,$7B,$F1,$2F,$F9,$C1,$F2,$DF,$CF,$1F,$A8
	.dc.b	$9F,$4C,$F8,$63,$07,$A5,$FB,$B1,$FC,$C2,$BE,$99,$FE,$28,$78,$C7
	.dc.b	$82,$7D,$27,$D3,$3F,$52,$FE,$CD,$FC,$6B,$B7,$B9,$F8,$77,$EE,$3C
	.dc.b	$AB,$D2,$FE,$30,$FD,$D4,$69,$E2,$62,$BE,$21,$F6,$9F,$F0,$A5,$7E
	.dc.b	$9D,$F8,$D3,$C6,$FF,$07,$18,$AE,$33,$F6,$23,$FA,$4B,$F4,$6B,$DE
	.dc.b	$F1,$7E,$8C,$7A,$67,$8C,$FC,$8B,$FB,$2F,$C1,$9F,$C5,$1F,$C6,$6F
	.dc.b	$C2,$F1,$3F,$1B,$5E,$97,$C3,$3C,$67,$F0,$31,$A8,$FE,$70,$FE,$30
	.dc.b	$FE,$0E,$7D,$97,$FA,$FD,$FF,$BC,$23,$F0,$33,$E0,$7E,$41,$CE,$7D
	.dc.b	$0F,$E8,$5F,$04,$FC,$0A,$F6,$C7,$F2,$C7,$F0,$6F,$C7,$9F,$D8,$DF
	.dc.b	$F9,$81,$F6,$1E,$FA,$F8,$EF,$80,$C5,$7B,$DE,$27,$F9,$00,$D5,$FF
	.dc.b	$0F,$BF,$04,$7A,$AB,$9D,$FE,$1A,$3F,$C5,$D1,$E7,$7D,$93,$C1,$F8
	.dc.b	$C7,$F1,$33,$F5,$C7,$BD,$FE,$11,$5F,$3A,$3C,$B1,$F3,$A3,$F8,$3D
	.dc.b	$FE,$E6,$78,$9F,$8E,$9F,$C3,$3F,$8B,$3C,$27,$E8,$1F,$C4,$47,$BE
	.dc.b	$3F,$4B,$1E,$0A,$F4,$47,$82,$3F,$B7,$47,$F0,$D8,$FC,$FD,$FB,$EF
	.dc.b	$FB,$E4,$F0,$7E,$34,$F7,$BF,$AE,$FF,$83,$63,$FA,$4D,$F9,$A3,$EA
	.dc.b	$C7,$07,$F2,$AD,$3F,$B6,$5F,$9D,$FF,$14,$1F,$0A,$3F,$1E,$72,$BF
	.dc.b	$E5,$55,$F8,$38,$F6,$D7,$E0,$DF,$B7,$3A,$9F,$69,$FC,$46,$BF,$29
	.dc.b	$5E,$D9,$FD,$C5,$75,$BF,$A0,$F2,$6C,$EE,$FD,$93,$93,$CA,$53,$F0
	.dc.b	$07,$F1,$5B,$F8,$71,$F9,$E8,$FB,$07,$D1,$7F,$11,$1E,$F8,$E3,$3D
	.dc.b	$E7,$84,$79,$4E,$F5,$FA,$48,$FD,$5C,$7E,$15,$DB,$E3,$3F,$AC,$C6
	.dc.b	$E3,$F6,$F3,$F3,$EF,$F7,$15,$F9,$83,$60,$82,$7F,$D9,$2F,$FA,$9C
	.dc.b	$C9,$BA,$D4,$6E,$32,$7F,$B2,$0E,$8F,$27,$FC,$D8,$71,$78,$BC,$0E
	.dc.b	$C7,$47,$FD,$15,$19,$8E,$11,$DD,$E1,$1E,$08,$EC,$EE,$3F,$C3,$07
	.dc.b	$91,$D4,$7F,$B6,$A3,$B9,$F0,$23,$C0,$F7,$77,$1E,$28,$C3,$E8,$7C
	.dc.b	$27,$C4,$7C,$2E,$9F,$33,$E1,$34,$F8,$5F,$86,$79,$8F,$F1,$89,$F2
	.dc.b	$1D,$47,$AC,$F2,$C7,$80,$FA,$F1,$E0,$7E,$1B,$E5,$7C,$06,$9E,$6F
	.dc.b	$95,$E8,$F9,$1F,$73,$C9,$F4,$BF,$C0,$E7,$F5,$D1,$F9,$A7,$F8,$55
	.dc.b	$FF,$41,$AF,$C8,$57,$E2,$E1,$3F,$88,$FC,$13,$F1,$B1,$FE,$98,$6C
	.dc.b	$1F,$C1,$47,$07,$FC,$09,$3F,$AA,$8F,$E1,$53,$E3,$AF,$DC,$9E,$09
	.dc.b	$4F,$81,$FC,$DD,$76,$BF,$CB,$BE,$03,$BD,$F8,$63,$F9,$A3,$EC,$3F
	.dc.b	$A6,$DF,$5B,$FE,$74,$78,$AF,$A9,$EF,$9F,$CC,$5F,$58,$FD,$1C,$FA
	.dc.b	$EF,$F1,$0F,$96,$3F,$61,$F3,$5F,$8E,$3F,$52,$7E,$64,$FB,$EE,$0F
	.dc.b	$AC,$D7,$F5,$0E,$9F,$DA,$BC,$DA,$7E,$A5,$D3,$FB,$E7,$4D,$35,$3F
	.dc.b	$89,$78,$9F,$89,$51,$E1,$8F,$3C,$6D,$E4,$6E,$34,$F9,$23,$4F,$C0
	.dc.b	$3D,$67,$63,$9B,$F9,$07,$99,$FA,$37,$C2,$EC,$E8,$ED,$E8,$78,$4D
	.dc.b	$1E,$F7,$C2,$F2,$34,$76,$39,$BF,$45,$F2,$99,$39,$3D,$DE,$47,$96
	.dc.b	$3D,$EE,$A3,$F8,$97,$85,$FC,$28,$FE,$4A,$61,$F1,$B9,$8F,$E5,$AE
	.dc.b	$63,$B3,$98,$FE,$6A,$E6,$3A,$C7,$EE,$63,$AB,$FC,$9D,$FB,$EF,$F3
	.dc.b	$D7,$93,$4D,$47,$89,$E4,$78,$8C,$BE,$27,$27,$88,$CB,$E2,$7F,$0D
	.dc.b	$1E,$27,$F9,$E3,$E2,$4F,$89,$A7,$23,$D0,$76,$3C,$C7,$28,$F3,$3D
	.dc.b	$8F,$31,$97,$CC,$E4,$F3,$19,$7F,$A2,$78,$FF,$CF,$0F,$EA,$E6,$1F
	.dc.b	$C2,$B9,$8F,$EB,$EE,$63,$F0,$0E,$63,$FB,$3B,$98,$EB,$1F,$CA,$E3
	.dc.b	$AB,$FD,$65,$F8,$47,$17,$FB,$83,$C5,$D1,$C4,$F0,$9C,$63,$CA,$E4
	.dc.b	$F0,$99,$7C,$AE,$4E,$A6,$5F,$2B,$FB,$B8,$F2,$BF,$DC,$5F,$2B,$C4
	.dc.b	$3E,$57,$46,$5C,$3C,$8F,$DB,$35,$F9,$31,$FB,$0F,$CB,$3F,$99,$3F
	.dc.b	$E0,$D7,$FA,$09,$FC,$C8,$FE,$7F,$1F,$E1,$3F,$CB,$37,$27,$FA,$39
	.dc.b	$FC,$E0,$DC,$64,$FE,$65,$1F,$E1,$48,$DC,$7F,$32,$CF,$94,$FF,$32
	.dc.b	$7F,$C3,$75,$D8,$F1,$CE,$DE,$A6,$5F,$E7,$CF,$9E,$79,$47,$F8,$A2
	.dc.b	$BC,$E7,$81,$FF,$13,$E5,$FD,$8A,$3C,$E6,$A3,$FA,$43,$EB,$9E,$71
	.dc.b	$FE,$33,$AF,$03,$FD,$72,$D1,$E1,$9F,$F1,$15,$FF,$4B,$AC,$9F,$0A
	.dc.b	$3E,$73,$E8,$3D,$8F,$67,$E4,$1D,$5F,$90,$F9,$0F,$91,$7E,$87,$A4
	.dc.b	$70,$9F,$64,$6A,$3A,$57,$A2,$7F,$C0,$2F,$9E,$3F,$BF,$CF,$F9,$3A
	.dc.b	$3F,$C5,$A7,$F9,$20,$FF,$1C,$1F,$C1,$CC,$BF,$58,$FF,$21,$39,$7F
	.dc.b	$C3,$0F,$17,$31,$F8,$F7,$27,$13,$FC,$9D,$1C,$5F,$EB,$8F,$17,$31
	.dc.b	$C5,$C9,$C4,$F3,$39,$7F,$AA,$47,$17,$31,$FD,$16,$3D,$31,$EF,$3C
	.dc.b	$B1,$97,$CA,$64,$F0,$B9,$7C,$26,$4E,$AF,$C6,$78,$9F,$87,$3A,$B9
	.dc.b	$78,$99,$3A,$B9,$9F,$F1,$AB,$FE,$7A,$9F,$D0,$CF,$E6,$2F,$F1,$F3
	.dc.b	$F7,$23,$F0,$33,$A7,$F7,$31,$97,$F1,$4F,$E4,$1F,$CC,$47,$E0,$5F
	.dc.b	$F1,$31,$EF,$F1,$1E,$2F,$F9,$FE,$7F,$D2,$31,$87,$FD,$08,$DA,$FF
	.dc.b	$38,$3C,$2B,$E4,$61,$EA,$9F,$03,$F9,$08,$FC,$FC,$7E,$9B,$0F,$C1
	.dc.b	$4E,$E3,$F4,$F3,$E6,$36,$FE,$59,$F9,$11,$BB,$F8,$11,$FC,$0A,$7E
	.dc.b	$05,$FA,$9F,$95,$1F,$EA,$B8,$F5,$3F,$EA,$47,$D5,$1F,$E9,$97,$D5
	.dc.b	$1D,$AF,$E1,$3F,$B7,$3E,$95,$7F,$12,$7E,$89,$93,$E8,$D7,$B5,$F8
	.dc.b	$73,$C9,$F6,$1C,$A3,$CF,$86,$6F,$CB,$3F,$EB,$C7,$CB,$3F,$4A,$7E
	.dc.b	$3D,$73,$C3,$E6,$5F,$3A,$F0,$3E,$97,$E0,$BE,$03,$CA,$FC,$B7,$E7
	.dc.b	$C7,$CB,$7D,$47,$CB,$8E,$6F,$D4,$BF,$99,$1F,$8B,$7C,$F1,$E5,$C3
	.dc.b	$FD,$47,$1F,$2A,$3A,$DF,$9C,$FB,$AF,$88,$FC,$89,$E7,$AF,$C3,$C6
	.dc.b	$A3,$C5,$1E,$7A,$FE,$35,$39,$9F,$CD,$CF,$CA,$3D,$F5,$E4,$8F,$C0
	.dc.b	$61,$FA,$89,$FB,$CF,$CA,$7E,$2E,$1F,$2A,$7C,$87,$F3,$D3,$EC,$C6
	.dc.b	$AF,$C1,$3F,$CA,$A3,$F5,$CF,$C9,$9F,$9C,$7A,$0F,$69,$F2,$6B,$EF
	.dc.b	$1E,$9B,$FB,$B1,$E6,$3B,$D7,$C2,$C3,$F6,$D1,$E3,$3E,$34,$7C,$C8
	.dc.b	$CD,$FF,$4F,$7E,$29,$8C,$3F,$97,$5F,$D8,$AF,$14,$FE,$EB,$0F,$89
	.dc.b	$1F,$8D,$C3,$C2,$F5,$7F,$62,$7A,$A7,$AB,$83,$C8,$7D,$E8,$F8,$D1
	.dc.b	$E1,$8F,$E1,$93,$F0,$23,$C0,$7E,$19,$FA,$53,$EF,$9E,$98,$7A,$DF
	.dc.b	$D6,$9F,$30,$FB,$F1,$F9,$4A,$F1,$C7,$E8,$63,$F0,$CF,$E0,$1F,$C1
	.dc.b	$D7,$CF,$3D,$F1,$F3,$CF,$97,$3F,$C2,$9E,$B3,$C9,$FE,$52,$FC,$37
	.dc.b	$B1,$F0,$CF,$2D,$7E,$25,$FC,$01,$F1,$27,$F7,$13,$E2,$7E,$0B,$FC
	.dc.b	$A2,$BE,$0B,$D6,$7F,$23,$5F,$05,$F0,$D7,$96,$3F,$21,$3F,$06,$7C
	.dc.b	$38,$7E,$50,$F6,$BE,$13,$E1,$47,$ED,$EF,$E1,$5F,$D1,$9F,$BE,$F5
	.dc.b	$AE,$37,$E1,$3F,$38,$7B,$67,$E2,$9F,$04,$F2,$9F,$C2,$0C,$9E,$11
	.dc.b	$87,$E4,$6B,$F8,$34,$7F,$31,$BF,$3C,$FF,$07,$9F,$B7,$3F,$0A,$FE
	.dc.b	$2C,$79,$4F,$E6,$11,$E0,$3F,$09,$7F,$13,$0F,$B1,$7D,$E7,$F8,$0D
	.dc.b	$7D,$09,$FE,$03,$1F,$00,$F4,$57,$17,$D8,$FC,$0B,$F9,$95,$E0,$8C
	.dc.b	$3F,$9D,$9F,$EF,$57,$DD,$F7,$E1,$F3,$B0,$FB,$D1,$DC,$F3,$E1,$F3
	.dc.b	$8F,$6B,$E2,$70,$EF,$0F,$A0,$79,$5F,$99,$7D,$CF,$E3,$CF,$83,$0F
	.dc.b	$B8,$F9,$30,$FD,$83,$F3,$E3,$2F,$ED,$F0,$FD,$83,$F9,$C3,$EC,$5F
	.dc.b	$B9,$FD,$9E,$1F,$C2,$AF,$9C,$7F,$55,$7E,$C9,$9A,$FC,$F4,$77,$7E
	.dc.b	$0B,$FA,$E3,$DD,$7F,$2E,$BF,$A2,$3F,$2E,$3F,$93,$4F,$27,$B4,$7A
	.dc.b	$0F,$D5,$C7,$E6,$DF,$B8,$FF,$19,$8E,$C7,$CE,$8E,$D1,$FC,$52,$3D
	.dc.b	$67,$F1,$47,$B3,$FD,$65,$F3,$BE,$27,$B3,$EB,$7F,$12,$F6,$BF,$AB
	.dc.b	$5F,$62,$30,$FE,$22,$7F,$AA,$1C,$A3,$D6,$F6,$AF,$5B,$FC,$D2,$3B
	.dc.b	$3E,$B7,$EB,$3F,$C1,$5F,$E1,$B5,$FB,$5A,$F9,$31,$EC,$3F,$01,$3E
	.dc.b	$C7,$F4,$2F,$77,$F8,$1B,$DD,$F1,$BE,$C7,$BD,$7E,$BE,$3F,$AE,$CF
	.dc.b	$69,$FC,$35,$7F,$15,$C3,$F2,$58,$7E,$7E,$F9,$61,$EC,$3F,$7F,$7D
	.dc.b	$A3,$D9,$1F,$BE,$8F,$5E,$1F,$A2,$7E,$C4,$FC,$83,$CC,$FD,$57,$AB
	.dc.b	$EA,$9E,$AE,$63,$AB,$FC,$92,$3F,$18,$77,$7F,$8A,$E1,$E9,$BF,$CD
	.dc.b	$C7,$13,$FA,$21,$F8,$1A,$F0,$BF,$29,$F0,$BF,$C4,$67,$F0,$35,$F8
	.dc.b	$E9,$FD,$5C,$FF,$19,$3F,$1C,$FE,$05,$EC,$7F,$06,$32,$F6,$7F,$74
	.dc.b	$FD,$77,$B5,$FE,$2E,$3F,$A0,$46,$27,$DA,$FC,$33,$F8,$F9,$F8,$A3
	.dc.b	$D6,$FD,$37,$27,$D3,$32,$7C,$73,$F1,$47,$E0,$8F,$DB,$C7,$D3,$8F
	.dc.b	$9F,$1F,$0E,$3E,$7B,$FC,$81,$FC,$83,$EA,$7E,$3D,$7F,$0B,$8F,$C4
	.dc.b	$D7,$DF,$8F,$93,$3D,$67,$CA,$7C,$6C,$3F,$A2,$3E,$D3,$F8,$F1,$ED
	.dc.b	$3E,$51,$FB,$0B,$F8,$6F,$C5,$3D,$6F,$C1,$78,$9F,$04,$E2,$7D,$F7
	.dc.b	$DA,$FA,$67,$DB,$5E,$98,$FB,$87,$73,$A9,$E5,$AF,$E6,$67,$F8,$F1
	.dc.b	$EB,$5E,$18,$CE,$1F,$CA,$27,$F9,$F6,$1F,$77,$0F,$F0,$86,$1D,$E7
	.dc.b	$F4,$A7,$E1,$B0,$F2,$E1,$D4,$F8,$B1,$F4,$6B,$F1,$CF,$D5,$3E,$E9
	.dc.b	$F0,$CF,$A2,$78,$27,$E3,$CF,$EE,$9F,$E5,$71,$C2,$30,$FC,$F7,$BD
	.dc.b	$79,$63,$F5,$77,$F6,$CF,$53,$FE,$05,$9E,$2F,$CF,$7C,$E6,$1F,$BD
	.dc.b	$87,$1B,$FB,$37,$D4,$FE,$C0,$7E,$DF,$0F,$C6,$DF,$C1,$A5,$FF,$BD
	.dc.b	$F3,$78,$BF,$EA,$F6,$E3,$A9,$82,$33,$80,$91,$D2,$39,$BD,$6B,$A1
	.dc.b	$FE,$91,$8E,$33,$C2,$7A,$4F,$6A,$F0,$57,$92,$BF,$D6,$6F,$92,$BC
	.dc.b	$67,$09,$F1,$4F,$39,$F0,$CF,$BA,$79,$D7,$89,$E3,$1C,$6B,$8B,$E9
	.dc.b	$9F,$13,$63,$C3,$1E,$93,$FC,$8F,$5D,$A7,$FD,$E2,$F9,$E7,$E2,$57
	.dc.b	$80,$EF,$1F,$02,$3C,$73,$F1,$5E,$B3,$AA,$EA,$78,$A7,$E2,$CE,$9F
	.dc.b	$83,$5C,$DF,$41,$CE,$3C,$2F,$AE,$BA,$1F,$1A,$3C,$31,$F1,$4F,$73
	.dc.b	$C1,$F9,$87,$6A,$F1,$BD,$63,$DF,$1F,$0E,$7A,$4F,$88,$E8,$7A,$27
	.dc.b	$A5,$7D,$98,$FC,$5B,$E9,$7F,$02,$7E,$DC,$FE,$36,$7E,$CC,$FE,$06
	.dc.b	$7F,$07,$BF,$ED,$13,$FC,$AB,$0F,$D2,$4E,$DF,$D3,$CF,$F5,$68,$C3
	.dc.b	$FE,$2F,$9F,$DF,$C7,$F1,$73,$F7,$11,$B7,$07,$80,$FE,$0D,$1F,$D3
	.dc.b	$E7,$F5,$51,$E7,$3F,$92,$9F,$B4,$30,$FF,$18,$BF,$F1,$94,$7A,$5F
	.dc.b	$E0,$EF,$58,$F4,$BE,$07,$D3,$3F,$B8,$AF,$E4,$13,$FE,$16,$8F,$4C
	.dc.b	$F8,$A3,$11,$E5,$9D,$C7,$F3,$53,$E5,$9D,$5F,$85,$1B,$AF,$2C,$FB
	.dc.b	$9F,$E4,$F3,$C9,$FE,$64,$FD,$08,$F8,$F3,$E5,$9E,$35,$EA,$3C,$93
	.dc.b	$E1,$3E,$29,$F4,$8F,$8B,$5C,$DF,$84,$FD,$18,$E3,$3E,$03,$E3,$BF
	.dc.b	$12,$BE,$0C,$F9,$6B,$E6,$1F,$CB,$DF,$6D,$7A,$8F,$2B,$D4,$E1,$1E
	.dc.b	$F3,$E2,$D7,$81,$E6,$7F,$31,$8E,$6E,$9E,$31,$EF,$69,$D6,$7B,$D7
	.dc.b	$C5,$3F,$C2,$93,$B8,$CC,$7A,$9A,$FD,$33,$73,$CD,$A7,$33,$FC,$57
	.dc.b	$3C,$DC,$9C,$CA,$73,$37,$1C,$DA,$73,$3F,$54,$E5,$A6,$6B,$EF,$BD
	.dc.b	$4E,$25,$36,$57,$6F,$84,$C9,$5C,$D7,$D5,$8F,$3B,$5E,$A6,$E3,$F0
	.dc.b	$AF,$8D,$AF,$76,$BB,$AF,$C3,$36,$BE,$CC,$FE,$48,$FD,$11,$F9,$2A
	.dc.b	$FC,$6C,$6A,$BF,$71,$5E,$4A,$FC,$05,$7D,$63,$CD,$1E,$C6,$C7,$B0
	.dc.b	$DC,$71,$7C,$13,$F8,$D7,$F2,$06,$2B,$F1,$33,$ED,$9F,$4D,$74,$AF
	.dc.b	$6B,$63,$31,$DA,$7E,$3C,$E6,$BD,$47,$EB,$6B,$75,$DE,$7D,$4F,$1A
	.dc.b	$F5,$34,$F5,$1B,$7D,$53,$F1,$1E,$F1,$B3,$CA,$58,$EE,$6E,$3A,$C7
	.dc.b	$37,$8B,$5C,$C6,$E3,$C2,$D7,$A9,$B3,$35,$ED,$3D,$F5,$F6,$A3,$EE
	.dc.b	$1F,$C0,$CA,$7E,$EC,$A7,$F3,$73,$F9,$4B,$F2,$5C,$9F,$C5,$CC,$9E
	.dc.b	$D8,$EE,$F5,$32,$79,$4F,$3C,$FA,$4F,$87,$1E,$78,$EA,$7E,$48,$E3
	.dc.b	$1B,$8E,$B1,$E7,$32,$F1,$6C,$FB,$DA,$78,$CF,$E6,$13,$CD,$EE,$FE
	.dc.b	$44,$FE,$11,$5F,$49,$D4,$72,$3C,$07,$2A,$F0,$36,$BC,$0D,$3C,$0F
	.dc.b	$F3,$76,$9F,$CA,$8F,$9C,$6C,$EA,$ED,$A6,$CF,$A0,$6C,$E2,$ED,$A6
	.dc.b	$CF,$94,$6C,$E2,$ED,$A6,$E7,$E1,$4F,$F1,$F8,$FC,$5C,$FB,$0E,$0F
	.dc.b	$46,$CF,$0A,$FD,$B5,$7D,$6A,$F1,$9F,$BC,$9F,$96,$F8,$CE,$47,$86
	.dc.b	$3F,$A8,$DF,$F9,$1E,$BF,$C5,$77,$FE,$36,$7F,$CA,$75,$FE,$24,$3F
	.dc.b	$C5,$73,$B9,$FF,$26,$47,$F8,$D5,$FF,$18,$BB,$3F,$C6,$B3,$FE,$47
	.dc.b	$7F,$C5,$F8,$7F,$8E,$9E,$6F,$F9,$8D,$E5,$1E,$0A,$FF,$2E,$57,$F9
	.dc.b	$A2,$39,$4F,$F9,$6D,$D5,$7F,$95,$6B,$C9,$3F,$E4,$F3,$9C,$FF,$9C
	.dc.b	$E7,$47,$78,$F1,$D7,$A2,$BB,$C7,$F9,$4A,$BE,$1B,$C9,$FF,$2B,$BE
	.dc.b	$C7,$FC,$97,$5F,$E5,$17,$C7,$1F,$E5,$5A,$F8,$AF,$68,$CB,$F0,$1F
	.dc.b	$5B,$FE,$65,$72,$F5,$7C,$2F,$17,$D4,$FE,$AA,$FF,$9B,$C7,$F5,$1C
	.dc.b	$3F,$7B,$5F,$89,$AF,$BC,$FF,$2E,$7F,$82,$46,$9F,$D2,$9F,$87,$AF
	.dc.b	$D6,$CF,$E3,$9E,$B5,$FD,$66,$7C,$13,$CE,$BF,$1A,$EC,$C5,$72,$9F
	.dc.b	$0D,$7F,$38,$8C,$47,$79,$F3,$34,$F1,$D7,$F0,$4A,$FE,$67,$5C,$23
	.dc.b	$13,$FA,$63,$71,$FA,$97,$F8,$FB,$B3,$11,$FE,$EE,$C3,$F0,$B3,$FE
	.dc.b	$8C,$3E,$65,$7D,$CB,$FC,$73,$F7,$1F,$D1,$3F,$42,$3E,$C3,$97,$F0
	.dc.b	$87,$FA,$3C,$FB,$11,$F8,$07,$F2,$D1,$C5,$C9,$F9,$53,$C2,$71,$8F
	.dc.b	$0C,$7E,$02,$38,$BE,$57,$F2,$91,$D5,$FC,$44,$75,$8E,$26,$63,$FD
	.dc.b	$99,$19,$78,$9E,$13,$A9,$E1,$32,$7C,$53,$27,$EE,$E3,$AB,$C4,$F0
	.dc.b	$C7,$95,$FF,$57,$3E,$97,$27,$53,$EF,$1F,$CB,$8A,$7F,$18,$AF,$BA
	.dc.b	$FD,$A9,$FC,$2C,$60,$FC,$C4,$71,$71,$7F,$C8,$63,$F9,$69,$D0,$B1
	.dc.b	$A7,$F3,$35,$F9,$3A,$EB,$5F,$86,$39,$4F,$F2,$63,$91,$6B,$93,$FD
	.dc.b	$54,$F4,$47,$17,$AD,$7F,$5B,$8F,$C9,$3F,$CD,$A3,$2F,$E7,$2B,$C2
	.dc.b	$E6,$3A,$1E,$89,$E8,$E0,$F1,$38,$7F,$9E,$57,$F4,$39,$FE,$05,$3E
	.dc.b	$A8,$C4,$7E,$E6,$79,$46,$2B,$FC,$1D,$18,$AF,$DF,$C6,$23,$AC,$61
	.dc.b	$FF,$08,$1E,$A3,$F6,$47,$D3,$7C,$91,$FC,$0E,$BE,$74,$FE,$D9,$F7
	.dc.b	$3D,$23,$F8,$39,$F0,$A3,$D3,$1D,$4F,$C1,$1C,$5C,$BE,$78,$E2,$E6
	.dc.b	$3D,$CF,$57,$C2,$F9,$5F,$0C,$79,$5F,$AC,$F8,$5E,$2E,$5F,$AD,$1D
	.dc.b	$63,$E2,$19,$7E,$09,$F8,$28,$EA,$FF,$20,$7A,$BC,$4F,$92,$7A,$5F
	.dc.b	$2B,$C5,$F6,$BE,$97,$8B,$E9,$73,$1E,$18,$F8,$A6,$63,$D2,$F8,$5F
	.dc.b	$4B,$98,$F4,$B9,$3D,$27,$E3,$E3,$D2,$FE,$82,$BF,$0C,$FA,$0F,$24
	.dc.b	$74,$7D,$0F,$81,$E1,$5F,$87,$7F,$A7,$BF,$93,$31,$3F,$D2,$23,$C4
	.dc.b	$7A,$8E,$75,$E6,$79,$BF,$79,$F5,$1E,$37,$E1,$47,$F0,$48,$E2,$64
	.dc.b	$EA,$E5,$F7,$57,$CA,$73,$5E,$5A,$ED,$1C,$8F,$49,$F9,$3A,$E4,$FE
	.dc.b	$CA,$BE,$3C,$7F,$04,$8F,$8A,$7C,$C3,$51,$DD,$F5,$BF,$80,$7C,$6F
	.dc.b	$C9,$8E,$53,$EC,$AF,$24,$EE,$B9,$BB,$8F,$D7,$C7,$A8,$F3,$3C,$23
	.dc.b	$F1,$31,$EB,$AF,$A0,$F8,$A3,$2F,$8A,$BE,$85,$FF,$66,$18,$7F,$87
	.dc.b	$6B,$FA,$55,$7F,$3E,$AF,$E9,$D5,$A7,$FC,$55,$5F,$E2,$EA,$FF,$0E
	.dc.b	$D7,$F5,$27,$FA,$35,$75,$9E,$91,$87,$FC,$69,$3F,$D4,$27,$FC,$66
	.dc.b	$F7,$3F,$A8,$1C,$CF,$E9,$8F,$3A,$FE,$EF,$5C,$CF,$0C,$79,$5F,$F1
	.dc.b	$6C,$7F,$5A,$9F,$EB,$C7,$F5,$63,$FA,$F1,$FD,$8A,$7D,$8F,$F7,$E3
	.dc.b	$27,$8C,$E2,$78,$CC,$BE,$37,$83,$EA,$BF,$BB,$1F,$5C,$FE,$05,$5F
	.dc.b	$49,$FB,$B5,$FA,$19,$DC,$7D,$79,$FD,$83,$F5,$CF,$CF,$19,$3F,$21
	.dc.b	$1F,$B8,$7F,$32,$76,$8E,$2F,$63,$8C,$76,$26,$1F,$8C,$8E,$87,$ED
	.dc.b	$4F,$01,$D0,$FC,$F9,$93,$A1,$9A,$FC,$D3,$F9,$42,$1E,$33,$CF,$1F
	.dc.b	$C1,$0F,$EC,$F5,$FA,$7A,$F1,$BF,$AE,$8F,$24,$62,$35,$18,$23,$B2
	.dc.b	$1A,$29,$FA,$92,$D7,$CF,$8E,$31,$B2,$1B,$2C,$6D,$E3,$5F,$2A,$78
	.dc.b	$C6,$C9,$3B,$78,$CE,$DE,$2F,$C7,$3F,$08,$7C,$CA,$FE,$1E,$7D,$B3
	.dc.b	$F7,$27,$C1,$8E,$B1,$CD,$EA,$F3,$3E,$CC,$72,$AF,$9D,$5F,$D7,$CF
	.dc.b	$25,$72,$AE,$15,$D6,$72,$48,$ED,$3B,$9F,$1D,$7C,$0A,$D5,$72,$8C
	.dc.b	$3C,$23,$33,$C2,$BE,$0B,$5F,$4B,$5E,$D5,$F8,$46,$CF,$A6,$BA,$10
	.dc.b	$F0,$47,$49,$E1,$3E,$06,$CE,$EB,$B4,$E4,$E9,$1C,$AB,$C7,$5C,$23
	.dc.b	$11,$E2,$7F,$BE,$18,$9E,$4E,$2B,$AB,$DE,$B9,$46,$23,$81,$99,$F0
	.dc.b	$C6,$49,$3C,$09,$1A,$6C,$F7,$7F,$D5,$B7,$FC,$46,$3F,$D1,$A7,$F1
	.dc.b	$4C,$3F,$96,$BF,$C7,$E3,$F9,$69,$B7,$F9,$84,$7F,$39,$C3,$FD,$3D
	.dc.b	$1C,$A7,$73,$CA,$3F,$90,$46,$27,$F9,$7C,$FF,$BF,$67,$FD,$3E,$7F
	.dc.b	$26,$9F,$E6,$53,$FD,$29,$F0,$47,$8E,$7C,$13,$FE,$89,$8E,$B1,$8B
	.dc.b	$FE,$9E,$F5,$7F,$9D,$57,$F3,$69,$F3,$4E,$DF,$E8,$F1,$E0,$AF,$F4
	.dc.b	$B6,$1F,$9D,$9F,$A0,$7C,$9B,$F9,$84,$70,$43,$FC,$22,$7D,$82,$47
	.dc.b	$D6,$8F,$F6,$DD,$7B,$27,$D2,$0F,$F6,$4D,$7B,$63,$C9,$1F,$D6,$AF
	.dc.b	$E6,$46,$23,$EC,$4F,$CC,$AF,$ED,$B1,$CC,$F9,$B3,$F0,$5F,$B2,$66
	.dc.b	$3E,$2D,$7F,$6E,$8E,$D5,$CC,$FE,$C6,$FF,$5F,$8D,$3E,$08,$FC,$0B
	.dc.b	$D0,$F4,$CF,$D2,$3C,$6E,$0F,$9F,$3F,$55,$E9,$1F,$DD,$1F,$0C,$F4
	.dc.b	$7F,$00,$79,$CD,$9F,$DA,$27,$ED,$3F,$89,$7A,$3F,$DC,$6B,$E0,$3C
	.dc.b	$8F,$11,$BA,$F2,$BF,$1A,$7D,$B3,$F9,$7A,$F1,$4F,$D6,$38,$C7,$47
	.dc.b	$07,$1A,$F5,$3F,$16,$7E,$D5,$79,$2B,$F0,$07,$D9,$3B,$46,$AB,$C6
	.dc.b	$7D,$43,$9C,$69,$F9,$91,$F5,$2B,$DD,$5E,$83,$E1,$D7,$94,$EC,$E0
	.dc.b	$D4,$7C,$B3,$8D,$76,$8D,$36,$35,$3E,$89,$F5,$1F,$92,$2D,$7E,$1C
	.dc.b	$FD,$C3,$9A,$F5,$D7,$D8,$3F,$62,$74,$8D,$4F,$AE,$BB,$57,$0A,$F2
	.dc.b	$9F,$7C,$E1,$3D,$CF,$F1,$D1,$F5,$A7,$DF,$1A,$8C,$57,$99,$B5,$D6
	.dc.b	$BB,$C7,$86,$7C,$6F,$A8,$ED,$5F,$94,$7E,$B1,$B3,$2F,$DF,$AF,$AB
	.dc.b	$1C,$4E,$E6,$CA,$64,$E6,$57,$8B,$EC,$32,$57,$31,$E3,$6B,$E1,$3F
	.dc.b	$5A,$FB,$67,$D7,$3C,$67,$2F,$8C,$EA,$F3,$32,$F8,$DA,$F1,$7F,$50
	.dc.b	$F1,$9F,$0D,$7E,$56,$76,$73,$37,$3C,$DA,$F3,$8F,$CB,$57,$D5,$8F
	.dc.b	$1C,$FA,$CD,$9C,$E7,$75,$ED,$9F,$A3,$1F,$E6,$9C,$3F,$C1,$75,$FC
	.dc.b	$C1,$FE,$C2,$FF,$34,$C3,$F9,$93,$FC,$DC,$FF,$50,$46,$9B,$7F,$CF
	.dc.b	$5E,$11,$FD,$A3,$0F,$97,$1E,$37,$F5,$6F,$F0,$46,$BF,$C0,$CF,$CC
	.dc.b	$19,$9F,$D6,$34,$C9,$B3,$F4,$A6,$4A,$64,$FE,$5C,$53,$F8,$E1,$4F
	.dc.b	$E2,$6F,$F0,$97,$AB,$FE,$6E,$7F,$CD,$87,$F9,$CE,$BF,$C9,$15,$F0
	.dc.b	$63,$FC,$F3,$1F,$19,$F8,$33,$F0,$DF,$8A,$7C,$93,$73,$C5,$FF,$3E
	.dc.b	$CF,$17,$67,$23,$81,$A8,$E1,$1E,$29,$ED,$3D,$1A,$ED,$B1,$B8,$E6
	.dc.b	$D7,$B1,$0F,$09,$0E,$87,$09,$FF,$2F,$B6,$7F,$DF,$AE,$8E,$45,$39
	.dc.b	$1E,$33,$86,$1F,$EF,$1A,$F1,$46,$9A,$F1,$21,$C4,$EA,$53,$85,$79
	.dc.b	$1C,$C7,$27,$47,$27,$04,$8E,$11,$CA,$BD,$6F,$02,$4E,$E3,$2D,$AF
	.dc.b	$55,$72,$AE,$CD,$78,$B6,$78,$4F,$FB,$55,$AF,$FB,$29,$B3,$FE,$BA
	.dc.b	$AE,$F5,$C4,$8E,$09,$5E,$26,$BD,$63,$6E,$48,$66,$36,$F8,$63,$6E
	.dc.b	$4D,$99,$8D,$CF,$96,$32,$D7,$81,$0E,$06,$9F,$D5,$4F,$68,$C9,$B8
	.dc.b	$E4,$E0,$8E,$08,$72,$9F,$1B,$63,$B5,$7F,$AD,$67,$2D,$8C,$D7,$F8
	.dc.b	$FE,$BF,$D8,$8D,$AF,$E0,$11,$83,$FD,$DA,$53,$F8,$1D,$73,$AF,$44
	.dc.b	$21,$5F,$E0,$48,$FE,$16,$7F,$7B,$3F,$89,$D7,$F0,$F7,$F8,$BB,$5C
	.dc.b	$9B,$3F,$8B,$B9,$69,$93,$F8,$C9,$70,$FE,$F6,$FF,$23,$39,$9F,$E0
	.dc.b	$43,$B9,$BC,$3F,$A5,$4F,$F5,$47,$E1,$47,$FB,$A2,$3E,$19,$1C,$10
	.dc.b	$FF,$71,$D7,$06,$9F,$22,$38,$38,$3E,$4C,$6E,$FF,$D1,$11,$FD,$46
	.dc.b	$7F,$8B,$CF,$E1,$63,$EF,$CF,$CF,$C3,$F0,$71,$F7,$DF,$A3,$19,$8D
	.dc.b	$CF,$53,$FA,$EC,$7F,$2D,$8F,$DA,$D7,$E1,$27,$55,$FE,$DD,$8F,$09
	.dc.b	$99,$F5,$1E,$FA,$F3,$9C,$DF,$DE,$C6,$DF,$69,$C0,$F8,$07,$E2,$48
	.dc.b	$FC,$48,$E7,$5F,$CF,$0F,$81,$3E,$DA,$F7,$CF,$B6,$30,$7B,$8E,$CF
	.dc.b	$2A,$F4,$CF,$E7,$A7,$35,$F0,$8E,$D3,$E7,$8F,$19,$ED,$8D,$47,$29
	.dc.b	$F9,$B1,$F1,$9F,$F7,$CC,$65,$F2,$C6,$5F,$F7,$D4,$FF,$32,$8F,$5D
	.dc.b	$7A,$48,$EE,$BF,$DF,$B3,$B7,$C8,$7C,$58,$F1,$C7,$F4,$8A,$F1,$BE
	.dc.b	$C7,$EC,$39,$7C,$6E,$5F,$3B,$93,$E4,$1F,$58,$87,$C2,$3E,$BC,$F8
	.dc.b	$63,$AC,$79,$5F,$98,$F0,$8D,$47,$68,$D3,$E6,$74,$7C,$D7,$9B,$FD
	.dc.b	$3D,$E6,$F8,$6B,$ED,$3E,$73,$EF,$91,$F9,$2F,$B6,$3E,$7B,$CD,$F0
	.dc.b	$3F,$E1,$DC,$3F,$81,$B7,$0F,$DF,$BF,$E5,$03,$FC,$91,$1F,$E4,$69
	.dc.b	$FF,$2B,$46,$63,$FC,$AC,$FF,$93,$0D,$CF,$F8,$E2,$FF,$8A,$47,$F1
	.dc.b	$23,$F8,$71,$FC,$42,$7F,$D0,$F5,$FE,$5B,$7B,$36,$B9,$4F,$C5,$8F
	.dc.b	$94,$FA,$DF,$B6,$FC,$A3,$EA,$9B,$7E,$0B,$FB,$C7,$E5,$D7,$D4,$73
	.dc.b	$1F,$B6,$72,$F1,$8F,$C8,$BD,$CF,$C2,$9F,$4F,$04,$C1,$29,$98,$7F
	.dc.b	$40,$8F,$E9,$91,$8A,$DD,$71,$8C,$1F,$EA,$F7,$FC,$F3,$87,$CE,$3E
	.dc.b	$2B,$FE,$7F,$C3,$E2,$CF,$CC,$3F,$A0,$94,$FA,$13,$F3,$A7,$85,$FF
	.dc.b	$9A,$27,$F0,$0F,$FA,$06,$3F,$D5,$66,$A3,$FA,$A9,$CE,$7F,$0F,$3F
	.dc.b	$71,$B1,$F1,$CF,$64,$7F,$01,$7E,$3B,$F3,$CF,$96,$FD,$39,$FC,$9C
	.dc.b	$66,$38,$9F,$9D,$C3,$F3,$6F,$DD,$7F,$42,$FE,$AA,$7F,$02,$FD,$37
	.dc.b	$F5,$2F,$57,$33,$D6,$3F,$A5,$1D,$70,$FE,$1C,$F8,$0E,$87,$A0,$F6
	.dc.b	$1C,$0C,$9F,$AF,$78,$B9,$3F,$0C,$64,$FD,$94,$7C,$07,$F1,$D3,$F0
	.dc.b	$1E,$B5,$FD,$1C,$FF,$1E,$BC,$4F,$81,$5F,$C7,$9F,$0C,$7C,$47,$E0
	.dc.b	$4F,$E9,$1F,$80,$7A,$4F,$D1,$92,$3E,$4B,$E6,$79,$57,$C5,$7C,$55
	.dc.b	$FC,$D0,$F6,$B8,$8F,$E6,$33,$E8,$9F,$E8,$F5,$FA,$88,$ED,$1E,$77
	.dc.b	$B3,$C8,$DC,$7C,$18,$DB,$F6,$1F,$8A,$FD,$87,$B4,$FE,$2D,$F8,$87
	.dc.b	$C0,$8F,$DC,$D7,$C8,$3D,$71,$C0,$FB,$C7,$D2,$3E,$F1,$E4,$30,$7D
	.dc.b	$93,$E4,$9D,$A7,$E2,$57,$A2,$BC,$93,$E8,$7E,$83,$E8,$3E,$85,$7C
	.dc.b	$E7,$9B,$E4,$78,$47,$A2,$7E,$74,$78,$A7,$E7,$4F,$CB,$79,$47,$53
	.dc.b	$FC,$95,$3F,$E4,$97,$FC,$6E,$7E,$D7,$0F,$E7,$2F,$F3,$3C,$3F,$84
	.dc.b	$47,$F0,$67,$F8,$14,$7F,$BB,$F0,$FE,$04,$7E,$EE,$78,$CF,$F0,$E9
	.dc.b	$D1,$FC,$09,$E2,$FF,$3E,$9E,$2F,$F9,$1D,$FE,$45,$87,$F2,$18,$4A
	.dc.b	$BF,$05,$5F,$6F,$0F,$88,$7B,$8F,$B5,$86,$AF,$EC,$9F,$A2,$C3,$F0
	.dc.b	$D3,$A8,$F9,$13,$F5,$5F,$A2,$FE,$45,$FA,$2F,$37,$F5,$2D,$7F,$04
	.dc.b	$75,$9F,$B7,$3D,$63,$11,$F5,$EB,$F4,$31,$F6,$9F,$DE,$4E,$E3,$F2
	.dc.b	$91,$8A,$FB,$B1,$CA,$30,$72,$9F,$81,$3A,$9E,$27,$E5,$8F,$49,$EE
	.dc.b	$9F,$91,$1E,$98,$C9,$D4,$E0,$7E,$1E,$7E,$6C,$7C,$38,$C3,$C5,$A7
	.dc.b	$53,$E7,$9E,$23,$F4,$86,$8E,$C6,$23,$E0,$4F,$E9,$8F,$2C,$73,$3E
	.dc.b	$79,$4F,$9E,$75,$8F,$CF,$D7,$F0,$37,$E1,$CF,$C5,$AF,$44,$76,$AF
	.dc.b	$DE,$1D,$0F,$1B,$CA,$BF,$6A,$F2,$7F,$55,$3C,$8F,$9D,$5E,$7A,$F7
	.dc.b	$B5,$E8,$79,$63,$F1,$8F,$F0,$33,$C2,$7A,$A3,$83,$4C,$9F,$50,$E4
	.dc.b	$FC,$47,$CF,$1D,$2B,$F8,$26,$1F,$AC,$3F,$6F,$5F,$AC,$AF,$DE,$3F
	.dc.b	$DD,$6F,$FC,$07,$87,$F0,$19,$FF,$15,$3F,$B0,$9F,$F5,$B4,$7F,$82
	.dc.b	$5A,$FF,$AF,$CF,$F0,$2D,$77,$9E,$51,$87,$93,$FE,$3D,$72,$FF,$8E
	.dc.b	$6F,$E6,$5F,$D7,$C3,$E6,$61,$F5,$E3,$F6,$B1,$F3,$AB,$E1,$9F,$89
	.dc.b	$3F,$C1,$F1,$FE,$4A,$7F,$8B,$5F,$6C,$3F,$C1,$71,$F0,$27,$51,$FC
	.dc.b	$4E,$3F,$88,$1F,$C4,$E7,$CE,$F2,$3E,$01,$E9,$29,$93,$E1,$1F,$E2
	.dc.b	$47,$8C,$75,$3E,$8C,$75,$9F,$F1,$E5,$FC,$A8,$F4,$C7,$F9,$56,$3E
	.dc.b	$B4,$FC,$B9,$EE,$F3,$AF,$61,$FC,$B0,$FE,$64,$78,$0B,$5D,$A7,$D0
	.dc.b	$79,$A7,$E1,$9D,$5C,$4F,$F9,$22,$7F,$CA,$31,$FE,$4F,$9F,$E5,$CF
	.dc.b	$EB,$CF,$C9,$57,$E5,$9A,$7D,$4C,$3F,$58,$FF,$1B,$3F,$84,$B8,$3E
	.dc.b	$FD,$7E,$B3,$0F,$BE,$FE,$B0,$FB,$F5,$F8,$17,$F9,$4C,$7E,$06,$BE
	.dc.b	$D4,$7C,$77,$E6,$3F,$E4,$E7,$FA,$69,$FD,$6A,$FD,$CF,$CE,$8F,$F2
	.dc.b	$93,$9C,$3E,$45,$FD,$06,$C7,$49,$FD,$2B,$F4,$A3,$E6,$C7,$78,$FC
	.dc.b	$49,$CD,$FA,$0F,$E1,$4F,$01,$E4,$3F,$5A,$EA,$BE,$E4,$75,$AF,$14
	.dc.b	$E6,$BB,$BE,$29,$FA,$F3,$FB,$93,$F1,$47,$F6,$F3,$0F,$E0,$0F,$09
	.dc.b	$E6,$3C,$31,$E9,$AF,$CA,$C7,$A6,$31,$5F,$8E,$AE,$73,$EB,$7B,$4F
	.dc.b	$EA,$A7,$CB,$1E,$29,$FE,$F1,$39,$8F,$5B,$D2,$3D,$C7,$CD,$8E,$0F
	.dc.b	$DC,$8F,$AA,$F9,$9F,$80,$FA,$AB,$CE,$E6,$78,$C7,$DE,$7C,$B3,$F8
	.dc.b	$59,$FF,$0D,$E1,$F9,$07,$FB,$DC,$7E,$38,$FC,$81,$F8,$E9,$FC,$83
	.dc.b	$FD,$D4,$FF,$29,$57,$59,$F6,$C7,$57,$C4,$7C,$F9,$F0,$C7,$AA,$3F
	.dc.b	$91,$3F,$6E,$3E,$1B,$EC,$77,$1F,$85,$37,$1F,$1D,$FB,$EF,$FA,$ED
	.dc.b	$FF,$10,$DF,$F6,$FB,$FD,$66,$1F,$AA,$AF,$E5,$47,$B4,$A7,$B4,$F7
	.dc.b	$9E,$D3,$25,$3F,$68,$71,$29,$C4,$DD,$FF,$84,$1E,$6D,$73,$7F,$C0
	.dc.b	$27,$F7,$15,$F4,$63,$FB,$01,$FC,$14,$FF,$00,$46,$8F,$54,$F8,$0F
	.dc.b	$24,$7D,$87,$C4,$66,$BC,$0D,$8F,$04,$7D,$A3,$2F,$18,$E6,$F8,$AB
	.dc.b	$C8,$7F,$87,$23,$E0,$46,$E7,$8B,$BA,$F2,$B9,$3B,$CF,$C2,$6B,$F0
	.dc.b	$9F,$6C,$F8,$DC,$C7,$9D,$CB,$C5,$AF,$1A,$FF,$09,$47,$73,$31,$C6
	.dc.b	$33,$1E,$58,$F4,$B5,$F8,$EF,$18,$D9,$EF,$8C,$C6,$DF,$8B,$19,$6D
	.dc.b	$7E,$3C,$FA,$CF,$68,$D5,$7C,$66,$D7,$B8,$F0,$57,$F8,$54,$E5,$3F
	.dc.b	$0E,$3D,$66,$9F,$F0,$E3,$FE,$1E,$3F,$D0,$00,$FF,$4B,$BF,$E8,$47
	.dc.b	$12,$9F,$E3,$15,$FC,$46,$7F,$87,$1F,$B5,$3F,$72,$FE,$D5,$FE,$F0
	.dc.b	$FE,$D5,$D9,$FC,$7C,$D9,$FB,$53,$6F,$ED,$5F,$E2,$97,$E0,$8F,$ED
	.dc.b	$31,$FB,$63,$FB,$41,$FE,$17,$BF,$ED,$0D,$3F,$B4,$57,$CC,$9F,$A7
	.dc.b	$7F,$2A,$7E,$29,$F5,$6B,$E7,$57,$CA,$AF,$C4,$57,$DC,$8C,$47,$E2
	.dc.b	$5F,$A1,$5E,$29,$FE,$0F,$3F,$66,$76,$FE,$2D,$DC,$FF,$93,$2B,$E9
	.dc.b	$9E,$8A,$E3,$5E,$6A,$E3,$5F,$64,$FF,$30,$9E,$CA,$FB,$D5,$E5,$8F
	.dc.b	$14,$FB,$EB,$F1,$B1,$8A,$DC,$F8,$A3,$F0,$47,$F9,$AA,$3C,$51,$C5
	.dc.b	$F8,$71,$C6,$7E,$1C,$F1,$9D,$9D,$0F,$5C,$F9,$67,$71,$88,$FE,$29
	.dc.b	$5E,$77,$E1,$1E,$C3,$35,$D5,$E5,$5F,$3E,$B9,$57,$F0,$CA,$F8,$B5
	.dc.b	$F6,$4F,$D2,$14,$FB,$B1,$F7,$CF,$03,$F0,$0F,$F4,$11,$E0,$9E,$B3
	.dc.b	$FC,$4A,$39,$4F,$09,$F7,$D7,$B2,$7D,$13,$E3,$7E,$34,$79,$DF,$66
	.dc.b	$1F,$D7,$5F,$EE,$47,$E7,$DF,$F2,$3E,$1F,$E2,$48,$FF,$03,$4F,$F5
	.dc.b	$97,$F4,$A7,$F8,$34,$FE,$B2,$E5,$E6,$64,$FD,$4B,$99,$F0,$57,$96
	.dc.b	$7E,$79,$F7,$CF,$9B,$3F,$95,$BF,$82,$FE,$66,$3D,$6E,$5F,$7E,$1F
	.dc.b	$16,$BE,$63,$F5,$E3,$E5,$1F,$C3,$0F,$E1,$73,$FD,$B6,$7F,$85,$47
	.dc.b	$F8,$58,$F7,$D7,$F8,$3B,$0F,$85,$1F,$E0,$03,$F8,$31,$EE,$C3,$FC
	.dc.b	$2A,$FC,$29,$E1,$7F,$C3,$63,$99,$FC,$4C,$E1,$5F,$C2,$63,$F8,$A5
	.dc.b	$7F,$A6,$2F,$E2,$57,$E0,$CF,$F3,$79,$EA,$3E,$31,$4F,$C3,$BC,$2F
	.dc.b	$F0,$75,$F7,$E7,$73,$97,$E6,$9F,$B9,$33,$87,$C9,$3F,$33,$1F,$2A
	.dc.b	$BE,$8D,$7F,$01,$AF,$A2,$71,$3C,$35,$C5,$C9,$F9,$67,$E9,$57,$86
	.dc.b	$3D,$07,$B1,$C1,$23,$94,$7E,$10,$EB,$3F,$81,$70,$71,$8D,$BF,$1C
	.dc.b	$FE,$5C,$76,$2C,$F5,$9E,$51,$B8,$FE,$3B,$1F,$C5,$23,$A9,$C2,$7A
	.dc.b	$D7,$CC,$9F,$C8,$4F,$C8,$79,$C6,$63,$E2,$1E,$58,$CB,$F4,$0F,$89
	.dc.b	$1F,$32,$3C,$11,$E3,$9F,$E3,$CF,$F0,$8A,$FA,$4F,$E7,$A7,$9B,$ED
	.dc.b	$8F,$4D,$7F,$24,$8F,$4C,$61,$E5,$5F,$70,$8F,$D6,$8E,$AF,$E3,$2F
	.dc.b	$F8,$21,$30,$FD,$0C,$7E,$74,$FC,$D4,$67,$0F,$CF,$47,$E9,$63,$FD
	.dc.b	$3A,$FF,$96,$A3,$FC,$D0,$F1,$3F,$53,$1C,$63,$FB,$66,$1F,$26,$7E
	.dc.b	$29,$EF,$8F,$C8,$3E,$CC,$3E,$2B,$FC,$A9,$F5,$B9,$9F,$E8,$0F,$19
	.dc.b	$F9,$B1,$D7,$0F,$7C,$FE,$32,$7A,$BF,$46,$78,$BF,$76,$33,$1C,$6B
	.dc.b	$BC,$EE,$7F,$98,$3B,$7B,$BF,$7E,$BB,$BF,$83,$8E,$6E,$67,$9D,$75
	.dc.b	$8F,$A9,$18,$74,$60,$87,$BE,$7C,$2F,$AD,$F0,$BF,$85,$8F,$8F,$5F
	.dc.b	$D1,$E3,$99,$C6,$7F,$83,$3C,$6B,$F1,$4F,$B6,$32,$FA,$5E,$B3,$FA
	.dc.b	$B7,$D7,$1D,$E7,$E4,$BC,$CF,$C8,$9E,$83,$A4,$7E,$A9,$D1,$F6,$CD
	.dc.b	$47,$C3,$7C,$D1,$B7,$C5,$1E,$63,$4F,$43,$F6,$27,$EE,$63,$F4,$F7
	.dc.b	$F8,$F7,$FA,$96,$1F,$9C,$9F,$EA,$B5,$FC,$E2,$7F,$27,$1F,$D7,$AB
	.dc.b	$F9,$EB,$A3,$F4,$35,$C6,$3F,$B3,$CF,$E8,$A7,$F3,$D1,$FA,$27,$FC
	.dc.b	$8E,$7B,$0F,$A2,$79,$CF,$CF,$9F,$89,$AF,$63,$FE,$5E,$C1,$50,$4F
	.dc.b	$F9,$96,$7F,$CC,$A7,$F5,$DC,$3F,$CC,$91,$FE,$67,$84,$7F,$5E,$7F
	.dc.b	$69,$3F,$D4,$30,$E6,$FF,$9A,$63,$FA,$B9,$FB,$68,$F7,$1A,$8F,$DD
	.dc.b	$C7,$C3,$34,$FC,$17,$18,$7C,$1B,$FF,$6F,$C7,$F8,$0E,$3E,$3C,$FC
	.dc.b	$E8,$FE,$C6,$7F,$71,$3F,$85,$57,$B9,$FF,$3A,$CF,$37,$EA,$47,$F6
	.dc.b	$97,$FC,$30,$F9,$67,$D9,$5F,$06,$BE,$C9,$E0,$8F,$C0,$9E,$77,$11
	.dc.b	$F8,$18,$C9,$E5,$8F,$C4,$46,$49,$5F,$82,$9E,$55,$E1,$8F,$1C,$F8
	.dc.b	$63,$CD,$1F,$05,$D3,$63,$4F,$C1,$3F,$13,$1F,$05,$FA,$75,$F0,$5F
	.dc.b	$8E,$7D,$A3,$BD,$7D,$97,$E4,$D7,$F8,$7D,$FA,$75,$CD,$C9,$F3,$1C
	.dc.b	$3E,$08,$E1,$5F,$12,$3D,$D5,$F0,$EB,$F0,$75,$E8,$9F,$11,$EB,$9F
	.dc.b	$13,$CC,$ED,$1D,$27,$E0,$D7,$D4,$69,$F1,$9C,$47,$59,$FD,$49,$DA
	.dc.b	$33,$5D,$A3,$DB,$3D,$4E,$07,$E4,$27,$73,$F9,$87,$D9,$3F,$98,$8F
	.dc.b	$61,$F6,$A7,$8C,$7E,$38,$FA,$87,$63,$FC,$95,$3F,$E3,$7C,$3F,$C0
	.dc.b	$B1,$FB,$C9,$FE,$26,$47,$04,$3E,$C9,$EC,$23,$E7,$BF,$84,$48,$F0
	.dc.b	$D7,$A0,$FC,$11,$F9,$47,$EA,$DF,$E6,$0F,$F3,$96,$1F,$CD,$5F,$E6
	.dc.b	$F8,$79,$A7,$F9,$B4,$7F,$48,$7F,$95,$9F,$CD,$8F,$E5,$E7,$F2,$A9
	.dc.b	$FF,$05,$3F,$CD,$4C,$9F,$C6,$0F,$E7,$2F,$F4,$D7,$AC,$7F,$48,$7F
	.dc.b	$99,$3F,$D1,$A3,$C9,$87,$F1,$E9,$DC,$FF,$3D,$76,$FA,$0F,$F0,$FC
	.dc.b	$F0,$8F,$31,$FC,$A5,$FE,$DA,$F0,$9E,$91,$B8,$F4,$1B,$AF,$ED,$8F
	.dc.b	$80,$F7,$3F,$D9,$8F,$F2,$B9,$F1,$E3,$91,$ED,$2B,$AA,$F2,$BB,$23
	.dc.b	$82,$1F,$05,$CB,$4C,$9E,$F3,$04,$30,$4A,$F6,$C7,$28,$C5,$7C,$49
	.dc.b	$FC,$83,$ED,$3E,$B1,$F7,$8F,$86,$EA,$BD,$6F,$F9,$4C,$FA,$97,$F8
	.dc.b	$D9,$FD,$35,$7E,$36,$7F,$90,$C6,$1B,$3F,$94,$C3,$F8,$2D,$7E,$D6
	.dc.b	$3F,$31,$1F,$C3,$AB,$F5,$47,$F1,$C7,$11,$FA,$B9,$F8,$75,$EE,$6D
	.dc.b	$7B,$4F,$D5,$5F,$A1,$FF,$1F,$4E,$E7,$E4,$35,$F8,$B1,$F0,$0E,$38
	.dc.b	$7C,$99,$D1,$23,$B1,$F2,$6B,$F1,$A7,$E8,$8F,$C1,$4F,$18,$DC,$78
	.dc.b	$DF,$EF,$46,$23,$4F,$84,$E8,$FE,$4A,$38,$47,$EA,$A3,$F1,$95,$EA
	.dc.b	$8E,$4F,$C2,$79,$4F,$07,$EE,$3E,$48,$D3,$5E,$51,$C2,$7C,$73,$99
	.dc.b	$F8,$AD,$AF,$85,$5B,$9E,$75,$E4,$76,$D9,$DD,$71,$AF,$3D,$78,$23
	.dc.b	$11,$CE,$31,$5C,$63,$71,$D2,$BF,$25,$18,$7A,$CE,$E3,$94,$ED,$FF
	.dc.b	$10,$47,$F6,$CA,$FF,$0F,$47,$F6,$95,$1F,$E2,$99,$C3,$F9,$38,$FF
	.dc.b	$13,$C7,$F8,$B6,$7F,$C6,$07,$F3,$0C,$3F,$AA,$B6,$FF,$A2,$BF,$CE
	.dc.b	$9F,$E7,$0E,$67,$F4,$13,$C2,$3F,$AC,$3F,$CE,$DA,$6E,$3F,$AA,$46
	.dc.b	$E3,$FA,$19,$F9,$68,$FA,$33,$F1,$23,$E0,$BF,$10,$C5,$7C,$E2,$5F
	.dc.b	$C7,$6C,$6A,$3E,$01,$F1,$CF,$E4,$78,$7E,$FC,$FD,$D5,$FF,$63,$3D
	.dc.b	$73,$F0,$23,$58,$7E,$E0,$F4,$47,$C2,$C3,$DA,$43,$D6,$7C,$73,$FB
	.dc.b	$61,$FD,$2B,$0F,$D6,$E1,$FC,$C6,$7C,$C4,$9F,$DE,$9F,$83,$21,$F0
	.dc.b	$8B,$3E,$E7,$66,$48,$E5,$FD,$A9,$89,$F8,$C4,$9F,$04,$6A,$78,$3F
	.dc.b	$B5,$8C,$B4,$EF,$1A,$7F,$67,$3C,$DC,$B5,$E8,$D7,$91,$89,$EA,$F2
	.dc.b	$6B,$FD,$74,$F7,$C7,$17,$47,$68,$DC,$FF,$17,$9D,$4F,$13,$E0,$4E
	.dc.b	$A3,$E8,$47,$12,$1E,$32,$9E,$62,$1E,$03,$44,$32,$77,$9F,$88,$D8
	.dc.b	$F1,$C7,$96,$7C,$47,$FB,$22,$7A,$1F,$E9,$D8,$FD,$AE,$1F,$EA,$28
	.dc.b	$FD,$A9,$F8,$D9,$FC,$83,$F8,$EC,$3F,$2B,$1F,$E4,$39,$FC,$C9,$0F
	.dc.b	$D7,$BF,$9B,$C3,$FA,$BC,$60,$D1,$FE,$8A,$3F,$82,$52,$93,$FC,$E7
	.dc.b	$0F,$E5,$67,$B0,$F9,$73,$F0,$CF,$59,$0F,$85,$7F,$02,$FF,$C6,$51
	.dc.b	$83,$E7,$9B,$C3,$E3,$38,$34,$75,$29,$F4,$1D,$B9,$6B,$9A,$F2,$B6
	.dc.b	$BE,$83,$C8,$F0,$46,$9B,$1A,$9E,$71,$8A,$F4,$C6,$23,$91,$1C,$12
	.dc.b	$BA,$46,$5B,$19,$AE,$D1,$82,$38,$25,$78,$23,$2D,$8C,$BF,$E5,$19
	.dc.b	$E8,$E4,$86,$67,$8B,$67,$8B,$FE,$53,$9E,$2E,$4E,$26,$67,$8B,$F9
	.dc.b	$77,$F9,$F5,$7E,$4E,$FF,$1B,$3F,$8B,$3F,$1C,$43,$04,$8F,$F5,$DB
	.dc.b	$96,$B9,$7F,$25,$1F,$91,$8F,$E8,$B1,$F8,$5A,$F3,$C7,$96,$FD,$73
	.dc.b	$FA,$83,$C9,$1F,$CF,$4E,$CF,$FA,$E5,$B8,$7F,$AA,$1F,$03,$63,$C6
	.dc.b	$43,$84,$6A,$FE,$0D,$7E,$D0,$93,$FD,$00,$F0,$CF,$FA,$C0,$87,$F9
	.dc.b	$35,$FF,$4D,$C7,$F9,$32,$7F,$A7,$3F,$B7,$9F,$41,$DE,$7F,$B5,$35
	.dc.b	$FE,$08,$77,$8F,$0B,$FD,$44,$95,$FC,$19,$FF,$6A,$90,$FF,$2A,$47
	.dc.b	$F0,$99,$F8,$75,$EC,$9E,$84,$39,$BF,$79,$F8,$95,$F1,$DF,$8A,$62
	.dc.b	$39,$3C,$27,$D6,$E4,$F5,$94,$F5,$9B,$7E,$3D,$7C,$53,$CC,$FC,$18
	.dc.b	$F9,$AF,$CB,$7F,$CE,$C7,$CC,$21,$82,$57,$F8,$02,$3A,$C6,$A3,$11
	.dc.b	$CD,$B3,$E1,$9F,$90,$0F,$F8,$00,$7C,$A3,$26,$0D,$1F,$E8,$22,$C7
	.dc.b	$FA,$1A,$3D,$71,$F0,$4F,$5B,$DD,$E8,$43,$8D,$7F,$84,$9F,$A9,$3F
	.dc.b	$75,$D1,$E1,$21,$E1,$2B,$E1,$7C,$E6,$48,$E5,$E6,$62,$3C,$0F,$D5
	.dc.b	$AF,$8D,$1F,$98,$7F,$A5,$C7,$AE,$BF,$27,$5E,$58,$F4,$C7,$E7,$67
	.dc.b	$E8,$C7,$D4,$AE,$51,$F9,$28,$FC,$29,$F8,$C2,$57,$E3,$C9,$3F,$8E
	.dc.b	$7B,$47,$BA,$3C,$F1,$88,$F5,$57,$86,$BD,$2F,$D2,$8E,$B5,$CC,$FC
	.dc.b	$F3,$83,$C7,$5E,$06,$C6,$EB,$F3,$B1,$B3,$F3,$73,$E1,$3E,$63,$F0
	.dc.b	$A3,$E3,$BB,$3F,$02,$7E,$4E,$7E,$1C,$7E,$7E,$BE,$5C,$79,$2B,$C5
	.dc.b	$5D,$A3,$07,$8E,$BE,$6C,$FE,$6A,$7A,$57,$BE,$38,$C7,$80,$FA,$51
	.dc.b	$9A,$E5,$5F,$1A,$7F,$0F,$5E,$A7,$E7,$57,$CB,$3F,$30,$FB,$67,$ED
	.dc.b	$3E,$57,$E2,$CF,$82,$7D,$71,$8A,$EF,$1C,$AB,$AC,$FC,$6A,$E1,$1F
	.dc.b	$A0,$AF,$7B,$F4,$88,$77,$9F,$83,$1F,$2A,$3F,$28,$71,$8F,$03,$C0
	.dc.b	$F8,$53,$F1,$DF,$86,$7C,$09,$FC,$59,$E0,$7B,$4F,$C5,$7F,$4F,$3F
	.dc.b	$15,$FC,$7C,$7A,$67,$DB,$1F,$48,$E9,$5F,$25,$C1,$CA,$3B,$CF,$9A
	.dc.b	$3B,$CF,$C5,$7A,$C7,$CE,$7D,$33,$EF,$7F,$19,$3F,$3A,$7E,$1B,$F7
	.dc.b	$63,$E0,$39,$83,$FB,$68,$FE,$32,$7F,$1F,$AF,$F2,$1C,$7F,$B1,$2F
	.dc.b	$FD,$E9,$3F,$E1,$D8,$FE,$E2,$A3,$FD,$3F,$18,$6D,$7F,$83,$27,$FB
	.dc.b	$B4,$FF,$82,$E3,$FB,$EC,$7F,$8E,$F0,$FF,$07,$BA,$3F,$D8,$D1,$FE
	.dc.b	$EC,$9F,$F2,$63,$FE,$17,$21,$FE,$03,$3F,$CB,$75,$FE,$E6,$8E,$D0
	.dc.b	$3F,$DC,$AF,$BA,$7F,$9E,$4F,$F9,$CA,$7F,$C3,$D1,$F1,$8F,$C4,$3F
	.dc.b	$29,$FB,$2F,$DD,$74,$7F,$25,$9F,$AD,$7F,$83,$26,$1F,$CA,$AF,$E4
	.dc.b	$9F,$E1,$AA,$F7,$1F,$27,$0F,$F3,$7D,$FB,$B0,$F6,$1E,$E9,$DD,$72
	.dc.b	$AF,$8F,$1C,$4F,$EA,$35,$F0,$2B,$DF,$1F,$E2,$99,$DD,$7D,$49,$E9
	.dc.b	$3C,$4F,$F6,$C4,$FF,$8C,$63,$E8,$47,$1A,$FB,$93,$99,$EB,$3F,$66
	.dc.b	$72,$F8,$8F,$B3,$1A,$7F,$CE,$8F,$E0,$DF,$01,$F9,$23,$71,$FF,$00
	.dc.b	$5F,$04,$6E,$BD,$0F,$C4,$24,$7C,$4A,$F9,$2F,$08,$F6,$CF,$08,$D4
	.dc.b	$7E,$41,$F1,$BA,$9F,$65,$78,$67,$D9,$3E,$19,$CC,$7F,$B5,$A7,$EB
	.dc.b	$3D,$27,$E3,$BC,$1E,$8F,$C3,$3E,$F3,$E1,$7F,$D1,$CE,$9F,$A2,$6E
	.dc.b	$BE,$F4,$FC,$C8,$C9,$0F,$90,$74,$9F,$0B,$EC,$7E,$2B,$F3,$A3,$F0
	.dc.b	$87,$B5,$D1,$F9,$2C,$3F,$AC,$61,$FC,$4A,$BF,$41,$3F,$C5,$6B,$F8
	.dc.b	$C3,$FC,$98,$C4,$7E,$8D,$FE,$2E,$7F,$1C,$3F,$A2,$3F,$E1,$A7,$D9
	.dc.b	$87,$D6,$7E,$34,$79,$B0,$FA,$D1,$E7,$3F,$06,$7C,$F3,$EE,$46,$E7
	.dc.b	$E9,$C7,$FC,$00,$32,$7D,$38,$FC,$83,$F3,$9E,$0F,$89,$FA,$4F,$F1
	.dc.b	$B7,$EF,$39,$7E,$5B,$C4,$E9,$5F,$25,$C1,$E7,$3E,$94,$78,$1E,$8F
	.dc.b	$DD,$7E,$BD,$7D,$C8,$EF,$1F,$6E,$7E,$AB,$EF,$3E,$29,$E0,$30,$7D
	.dc.b	$83,$C1,$3C,$E7,$B4,$F2,$9F,$88,$72,$37,$5F,$45,$E5,$5F,$40,$91
	.dc.b	$F3,$13,$FF,$00,$67,$F2,$53,$EA,$8E,$4F,$C2,$9E,$2F,$C9,$AF,$85
	.dc.b	$1F,$C7,$E3,$C3,$1F,$42,$30,$70,$8F,$84,$7E,$00,$F4,$9E,$69,$F2
	.dc.b	$1E,$39,$F9,$31,$E3,$7A,$47,$F1,$E3,$75,$FB,$69,$F0,$CF,$F4,$38
	.dc.b	$FF,$6E,$1F,$D0,$E3,$F9,$93,$FD,$1E,$3F,$A4,$47,$E9,$6F,$FC,$04
	.dc.b	$FF,$70,$9F,$ED,$8F,$E8,$0F,$D2,$1F,$A0,$3F,$C2,$27,$F4,$53,$F4
	.dc.b	$87,$F8,$7A,$7F,$BD,$47,$D7,$BF,$9E,$F9,$DF,$89,$3E,$D9,$FC,$9B
	.dc.b	$F8,$C7,$F3,$6F,$C8,$3F,$23,$3F,$95,$74,$7E,$24,$F9,$67,$D7,$BF
	.dc.b	$F0,$AB,$E8,$3C,$93,$FE,$51,$7F,$B0,$BF,$E1,$F7,$D8,$7F,$83,$8C
	.dc.b	$CF,$F6,$0B,$FE,$BD,$5F,$84,$9F,$09,$F4,$2B,$F0,$71,$F5,$63,$F0
	.dc.b	$D5,$E7,$BF,$C1,$CF,$D8,$9F,$DD,$9C,$67,$51,$C6,$3E,$04,$FE,$86
	.dc.b	$BE,$1C,$7C,$77,$BD,$7E,$60,$F9,$D5,$F8,$98,$FD,$3C,$FE,$3E,$BE
	.dc.b	$8B,$E0,$7E,$8D,$70,$3E,$A5,$79,$67,$F5,$93,$A7,$94,$75,$AF,$1D
	.dc.b	$71,$9E,$93,$C5,$FC,$E1,$F0,$8F,$DE,$9F,$54,$C1,$F6,$0F,$97,$5D
	.dc.b	$DF,$BE,$70,$33,$1D,$5E,$47,$C6,$7E,$A4,$7C,$E8,$EA,$FC,$07,$D7
	.dc.b	$1C,$5E,$66,$9F,$63,$F8,$B8,$E0,$FA,$8E,$D5,$FB,$2A,$FA,$11,$F1
	.dc.b	$1F,$92,$F3,$6B,$C2,$3F,$3B,$3E,$58,$C4,$7D,$A3,$FC,$7B,$5F,$E5
	.dc.b	$6A,$F9,$93,$FE,$58,$AF,$F1,$D3,$FB,$7A,$FF,$35,$3F,$B7,$3F,$C7
	.dc.b	$2F,$EC,$1F,$CC,$9F,$B4,$3F,$CD,$AF,$E8,$5F,$DC,$0C,$3F,$CC,$6F
	.dc.b	$F9,$25,$FF,$32,$C7,$F0,$78,$FF,$34,$61,$F2,$6B,$D5,$1F,$93,$BF
	.dc.b	$CC,$9F,$11,$FB,$27,$DD,$8D,$C7,$B2,$3E,$61,$C4,$FB,$D1,$A7,$E6
	.dc.b	$1F,$8C,$3E,$61,$4F,$93,$1E,$0A,$FE,$4E,$F5,$8F,$C2,$9D,$4F,$16
	.dc.b	$1F,$AD,$71,$5F,$79,$DC,$F1,$7F,$0A,$76,$3C,$B3,$C6,$35,$3E,$79
	.dc.b	$F7,$BC,$DE,$27,$69,$FC,$6B,$FB,$F3,$9C,$FC,$79,$FD,$F4,$7E,$46
	.dc.b	$77,$1F,$A3,$33,$1C,$4F,$5C,$F1,$9F,$D0,$9E,$58,$F9,$27,$E8,$0F
	.dc.b	$92,$53,$D8,$72,$AF,$92,$F0,$3E,$AC,$FB,$A7,$A4,$7B,$9F,$29,$F8
	.dc.b	$33,$35,$E6,$3D,$0F,$92,$7C,$D5,$FD,$45,$FE,$AB,$5F,$E9,$7A,$FC
	.dc.b	$EB,$ED,$7E,$51,$83,$FD,$40,$FE,$70,$FE,$78,$7E,$70,$FE,$A7,$87
	.dc.b	$F8,$FA,$7F,$C7,$6F,$FB,$84,$98,$7F,$B2,$DF,$F7,$8C,$FE,$86,$3F
	.dc.b	$B0,$CF,$BC,$F3,$DF,$EE,$30,$F6,$9E,$B7,$F6,$F5,$F3,$0F,$ED,$EE
	.dc.b	$1F,$ED,$17,$E0,$AF,$ED,$71,$FB,$73,$FC,$9B,$3F,$B7,$7F,$67,$3B
	.dc.b	$7F,$77,$1F,$B7,$8F,$63,$FE,$53,$8E,$47,$F5,$88,$F5,$9A,$8E,$8D
	.dc.b	$79,$5F,$BA,$7E,$54,$F7,$7C,$73,$A3,$F8,$6D,$7F,$BA,$CF,$EC,$47
	.dc.b	$CC,$AF,$94,$7F,$6E,$AE,$55,$E3,$8C,$1A,$7F,$D9,$71,$E3,$7E,$BC
	.dc.b	$78,$DF,$82,$ED,$F9,$2D,$7D,$E7,$F7,$A3,$8D,$79,$9F,$5C,$6E,$BD
	.dc.b	$11,$FC,$86,$BF,$DA,$F1,$E9,$3E,$99,$A8,$FC,$5C,$6A,$BE,$74,$E6
	.dc.b	$3D,$B3,$E1,$3F,$C0,$C7,$A2,$39,$BE,$26,$B9,$3E,$D9,$E8,$9F,$F4
	.dc.b	$8B,$F1,$CF,$8B,$5F,$0E,$BE,$34,$76,$69,$F7,$A3,$44,$9F,$85,$3D
	.dc.b	$CF,$B6,$77,$7F,$C3,$CF,$A6,$79,$47,$07,$AC,$FD,$8A,$F4,$C6,$2B
	.dc.b	$FD,$7C,$A7,$EC,$61,$FB,$FC,$3F,$81,$5F,$EA,$30,$FE,$B8,$7F,$84
	.dc.b	$B0,$FE,$11,$1F,$E5,$3A,$FD,$E5,$7F,$5F,$91,$A3,$F8,$22,$9F,$F2
	.dc.b	$C5,$6A,$7F,$D2,$98,$79,$5F,$C3,$CF,$A3,$0F,$89,$5F,$3A,$7E,$04
	.dc.b	$7B,$5F,$A6,$7D,$AC,$3D,$4F,$EE,$CF,$E8,$67,$F2,$32,$9E,$53,$FB
	.dc.b	$A6,$1E,$5A,$FE,$01,$1F,$96,$3F,$8A,$1F,$66,$BC,$71,$F4,$A3,$38
	.dc.b	$7D,$28,$F3,$46,$27,$CC,$FD,$A3,$94,$7D,$78,$E7,$87,$ED,$23,$F2
	.dc.b	$F3,$F9,$13,$DB,$5D,$63,$C0,$FE,$75,$FE,$0B,$5C,$A7,$F4,$CF,$FB
	.dc.b	$96,$36,$DA,$ED,$1C,$0F,$11,$0F,$4D,$75,$7F,$85,$18,$9E,$55,$DA
	.dc.b	$BF,$64,$F9,$A3,$DE,$FE,$8C,$FB,$66,$4F,$D2,$CF,$73,$CB,$3C,$A3
	.dc.b	$E3,$57,$BE,$3C,$F3,$E2,$36,$7C,$69,$FD,$31,$DC,$FB,$13,$E9,$7F
	.dc.b	$51,$3C,$9A,$7A,$67,$94,$FE,$46,$38,$D7,$B9,$FE,$16,$D8,$FE,$13
	.dc.b	$1F,$74,$F5,$CF,$C5,$AF,$54,$78,$DF,$E2,$15,$FC,$6E,$3E,$E1,$EE
	.dc.b	$7E,$3B,$F9,$C3,$9C,$FB,$8F,$AC,$75,$3C,$F1,$F7,$E3,$CB,$3E,$07
	.dc.b	$91,$F4,$0E,$C7,$47,$E1,$3F,$92,$9E,$87,$63,$C0,$7C,$F8,$DD,$7D
	.dc.b	$39,$F9,$13,$B9,$CD,$7A,$A7,$E2,$1E,$13,$E1,$1F,$00,$A7,$D8,$3C
	.dc.b	$E7,$D0,$8F,$53,$C8,$C4,$FC,$03,$F1,$11,$E0,$8F,$61,$E4,$8F,$43
	.dc.b	$4F,$BE,$7C,$67,$91,$F0,$0D,$B9,$7D,$8F,$13,$E3,$3F,$11,$E7,$5F
	.dc.b	$82,$3F,$1C,$EA,$7C,$8F,$C1,$3D,$AF,$E0,$8F,$E3,$93,$EA,$8E,$D3
	.dc.b	$F3,$5F,$14,$79,$23,$B5,$7E,$44,$FC,$54,$FC,$93,$E3,$57,$57,$F5
	.dc.b	$C7,$84,$C9,$F4,$4F,$19,$4C,$9C,$E7,$D6,$78,$CF,$0C,$7D,$47,$33
	.dc.b	$F7,$DB,$3F,$48,$F4,$9F,$1A,$3C,$F1,$FA,$D9,$F3,$4F,$E5,$0F,$34
	.dc.b	$72,$6C,$FB,$E7,$E3,$D7,$D7,$9E,$91,$88,$FA,$91,$F4,$4F,$97,$3E
	.dc.b	$53,$93,$F1,$0F,$C5,$C7,$DE,$7C,$E7,$12,$D7,$C1,$9F,$49,$FA,$F7
	.dc.b	$E1,$CF,$E2,$8F,$0B,$BA,$F2,$1F,$82,$AE,$F3,$F6,$EB,$9B,$F8,$18
	.dc.b	$EF,$1E,$79,$F9,$CF,$A2,$3E,$31,$C8,$F8,$27,$AE,$3E,$5B,$FA,$CA
	.dc.b	$F9,$CF,$DD,$7E,$63,$C4,$F7,$3E,$AA,$F8,$4D,$7C,$D3,$EA,$9F,$96
	.dc.b	$7D,$D9,$D3,$E2,$3C,$B1,$E2,$7F,$1A,$7E,$2E,$BD,$4F,$95,$F9,$47
	.dc.b	$C2,$AF,$B9,$1E,$48,$C4,$7C,$C7,$E2,$3F,$3E,$7E,$61,$DE,$7F,$33
	.dc.b	$5F,$01,$F0,$47,$CA,$24,$F1,$8F,$59,$F8,$E8,$CC,$7C,$2A,$F0,$C7
	.dc.b	$69,$F9,$51,$EA,$7D,$B5,$DC,$91,$EC,$9F,$91,$3F,$70,$F9,$11,$D1
	.dc.b	$F0,$9F,$4E,$7F,$10,$FD,$B3,$BC,$F9,$2B,$E9,$C6,$4F,$6B,$EB,$7B
	.dc.b	$1E,$B7,$13,$F5,$A3,$E6,$4F,$92,$30,$78,$23,$EC,$9E,$09,$E3,$5E
	.dc.b	$78,$EA,$79,$DF,$A7,$1F,$1E,$BA,$46,$EB,$AB,$A9,$F4,$D7,$63,$D2
	.dc.b	$F1,$7D,$93,$F2,$A3,$F4,$8F,$28,$F9,$35,$D8,$8F,$0A,$FA,$0D,$AE
	.dc.b	$31,$E3,$24,$F0,$9F,$15,$77,$AF,$C8,$BF,$0E,$3F,$2D,$1C,$A7,$84
	.dc.b	$62,$37,$5E,$D8,$FD,$11,$98,$FC,$C3,$E8,$3C,$8F,$07,$07,$8A,$3D
	.dc.b	$2D,$8F,$03,$C1,$E6,$FD,$D7,$E4,$3B,$7A,$39,$7E,$7B,$E3,$9F,$CD
	.dc.b	$3F,$42,$32,$FE,$6A,$7C,$2E,$CF,$CA,$CF,$EB,$6B,$F8,$4B,$67,$F4
	.dc.b	$53,$FB,$07,$8B,$CE,$BD,$2F,$95,$F4,$BF,$89,$3E,$09,$DC,$FC,$6C
	.dc.b	$F5,$3F,$40,$FC,$C3,$A4,$7B,$9E,$47,$BC,$D1,$E3,$7C,$2F,$5A,$FC
	.dc.b	$53,$B3,$F4,$67,$CD,$9F,$8B,$1F,$83,$9F,$3C,$78,$DF,$D1,$BF,$82
	.dc.b	$9F,$1C,$69,$FD,$43,$F5,$1F,$05,$7E,$41,$EC,$7C,$C3,$71,$F4,$9F
	.dc.b	$43,$ED,$9F,$03,$CE,$3B,$C7,$9A,$7E,$7C,$70,$9F,$13,$F7,$4F,$87
	.dc.b	$5F,$61,$EE,$62,$7E,$CC,$66,$38,$1F,$10,$FB,$15,$DE,$BE,$5C,$78
	.dc.b	$8F,$51,$C2,$7F,$50,$E4,$FB,$67,$80,$E8,$7C,$49,$F9,$4F,$A5,$ED
	.dc.b	$3F,$06,$38,$4F,$47,$47,$81,$FA,$D3,$C2,$7E,$0C,$62,$75,$5F,$1A
	.dc.b	$3D,$C6,$63,$CA,$7A,$63,$F7,$4F,$98,$F5,$47,$E2,$1F,$C7,$92,$BE
	.dc.b	$AD,$7D,$99,$EB,$5D,$CF,$9A,$49,$FC,$DD,$75,$3D,$67,$8A,$7C,$87
	.dc.b	$78,$EB,$1F,$5D,$E3,$1F,$0C,$EF,$5F,$52,$79,$C7,$17,$E4,$D7,$B2
	.dc.b	$38,$3E,$39,$FC,$DC,$FE,$50,$F1,$47,$A6,$BB,$BC,$49,$1F,$B8,$3E
	.dc.b	$2B,$F9,$D7,$F0,$E7,$D0,$9F,$25,$79,$9F,$B9,$1F,$3C,$E4,$7A,$27
	.dc.b	$93,$F8,$53,$C9,$3E,$B7,$81,$FB,$97,$D7,$5E,$68,$E7,$1D,$EB,$F0
	.dc.b	$75,$EB,$7D,$F3,$E6,$3D,$EF,$9A,$3E,$23,$FA,$F8,$F2,$57,$E5,$AB
	.dc.b	$A3,$F8,$F2,$47,$BE,$7D,$B3,$F6,$AB,$D7,$3F,$1A,$B9,$47,$6A,$FC
	.dc.b	$43,$67,$EF,$47,$99,$F8,$47,$9E,$BE,$11,$D6,$B9,$9F,$06,$3E,$25
	.dc.b	$7D,$66,$9F,$11,$EC,$FE,$46,$7D,$6F,$A6,$BE,$39,$F0,$EB,$A4,$7E
	.dc.b	$7C,$FC,$59,$CD,$C9,$F9,$67,$A5,$76,$AF,$AB,$5E,$B9,$F8,$47,$DC
	.dc.b	$9F,$8A,$FC,$E3,$15,$EC,$9E,$11,$93,$E1,$4F,$0A,$F4,$C7,$DA,$9F
	.dc.b	$A3,$3F,$7D,$FB,$53,$A3,$E4,$C6,$63,$E8,$D7,$AA,$BD,$A7,$C8,$21
	.dc.b	$F0,$4F,$C8,$96,$BE,$93,$E8,$21,$E6,$8F,$54,$F8,$AB,$E2,$1F,$56
	.dc.b	$79,$3E,$37,$F0,$31,$B9,$F2,$47,$81,$F9,$D5,$F3,$27,$47,$C6,$7D
	.dc.b	$D5,$F1,$4E,$F5,$F5,$E7,$51,$88,$F6,$D7,$48,$F2,$36,$38,$4F,$62
	.dc.b	$1F,$20,$E7,$1E,$6A,$F8,$73,$F0,$4F,$86,$7E,$68,$87,$E0,$8D,$10
	.dc.b	$C9,$F7,$CA,$FB,$23,$E5,$CA,$64,$AB,$11,$9A,$B3,$9A,$12,$33,$47
	.dc.b	$8D,$61,$DC,$F0,$8E,$F3,$A8,$E2,$68,$B1,$C8,$57,$86,$39,$C7,$23
	.dc.b	$CE,$F2,$32,$F0,$9F,$03,$E4,$9E,$53,$87,$E0,$47,$23,$CB,$1E,$36
	.dc.b	$9E,$38,$E3,$1F,$02,$38,$1D,$DE,$0D,$38,$4F,$B6,$38,$47,$D1,$7D
	.dc.b	$07,$8E,$3C,$F1,$88,$DB,$97,$AC,$FD,$38,$EB,$18,$9F,$99,$1E,$A8
	.dc.b	$F8,$73,$B8,$F8,$53,$E3,$7E,$51,$E3,$38,$4F,$77,$AB,$CA,$3A,$C7
	.dc.b	$BE,$30,$7D,$77,$0F,$AD,$F8,$B3,$F2,$1F,$96,$78,$A3,$D0,$EA,$7C
	.dc.b	$8F,$C9,$35,$1E,$17,$4F,$BC,$F0,$4F,$58,$FC,$3C,$80

;--------------------------------------------------------------------------------
;	FMOVE.B Dr,FPn
;	FMOVE.B <mem>,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovebtoreg_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.B <EA>,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.B <ea>,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d2	source
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source,...
;	a4	expected result handle,expected status,...
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fmovebtoreg_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a0,a2			;source,...
@@:
	addq.l	#4,a0			;source
	tst.l	(a0)
	bne	@b
	addq.l	#4,a0			;0
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;Dr,FPn
	fmove.l	#0,fpcr
	moveq.l	#0,d2
	move.b	(a2),d2			;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fmove.b	d2,fp7			;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)		;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.B Dr='
	puthex2	d2			;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
;<mem>,FPn
	fmove.l	#0,fpcr
	moveq.l	#0,d2
	move.b	(a2),d2			;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.b	d2,-(sp)		;source
	fmove.b	(sp)+,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)		;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.B <mem>='
	puthex2	d2			;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source
	tst.l	(a2)			;source
	bne	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmovebtoreg_data_compressed::
	.dc.b	$00,$00,$64,$08,$10,$AD,$51,$67,$15,$9B,$D6,$1B,$C7,$86,$5C,$73
	.dc.b	$E5,$A7,$3D,$7A,$6D,$D7,$7E,$DC,$1D,$F8,$7C,$1C,$5E,$1E,$3F,$17
	.dc.b	$27,$8F,$97,$C9,$CD,$E5,$E7,$F3,$74,$79,$FA,$7D,$1D,$5E,$9E,$BF
	.dc.b	$57,$67,$AF,$B7,$D9,$DD,$ED,$EF,$F7,$78,$1E,$FF,$07,$E0,$78,$5F
	.dc.b	$07,$C3,$F8,$5E,$27,$C3,$F1,$7E,$27,$8D,$F1,$7C,$7F,$8D,$E4,$7C
	.dc.b	$7F,$27,$E4,$79,$5F,$27,$CB,$F9,$5E,$67,$CB,$F3,$7E,$67,$9D,$F3
	.dc.b	$7C,$FF,$9D,$E8,$7C,$FF,$47,$E8,$7A,$5F,$47,$D3,$FA,$5E,$A7,$D3
	.dc.b	$F5,$7E,$A7,$AD,$F5,$7D,$7F,$AD,$EC,$7D,$7F,$67,$EC,$7B,$5F,$67
	.dc.b	$DB,$FB,$5E,$E7,$DB,$F7,$7E,$E7,$BD,$F7,$7D,$FF,$BD,$F0,$1F,$7F
	.dc.b	$E0,$7F,$01,$F0,$5F,$81,$F8,$3F,$C1,$7C,$27,$E0,$FE,$17,$F0,$9F
	.dc.b	$0D,$F8,$5F,$87,$FC,$37,$C4,$7E,$1F,$E2,$7F,$11,$F1,$5F,$89,$F8
	.dc.b	$BF,$C5,$7C,$67,$E2,$FE,$37,$F1,$9F,$1D,$F8,$DF,$8F,$FC,$77,$C8
	.dc.b	$7E,$3F,$E4,$7F,$21,$F2,$5F,$91,$F9,$3F,$C9,$7C,$A7,$E4,$FE,$57
	.dc.b	$F2,$9F,$2D,$F9,$5F,$97,$FC,$B7,$CC,$7E,$5F,$E6,$7F,$31,$F3,$5F
	.dc.b	$99,$F9,$BF,$CD,$7C,$E7,$E6,$FE,$77,$F3,$9F,$3D,$F9,$DF,$9F,$FC
	.dc.b	$F7,$D0,$7E,$7F,$E8,$7F,$41,$F4,$5F,$A1,$FA,$3F,$D1,$7D,$27,$E8
	.dc.b	$FE,$97,$F4,$9F,$4D,$FA,$5F,$A7,$FD,$37,$D4,$7E,$9F,$EA,$7F,$51
	.dc.b	$F5,$5F,$A9,$FA,$BF,$D5,$7D,$67,$EA,$FE,$B7,$F5,$9F,$5D,$FA,$DF
	.dc.b	$AF,$FD,$77,$D8,$7E,$BF,$EC,$7F,$61,$F6,$5F,$B1,$FB,$3F,$D9,$7D
	.dc.b	$A7,$EC,$FE,$D7,$F6,$9F,$6D,$FB,$5F,$B7,$FD,$B7,$DC,$7E,$DF,$EE
	.dc.b	$7F,$71,$F7,$5F,$B9,$FB,$BF,$DD,$7D,$E7,$EE,$FE,$F7,$F7,$9F,$7D
	.dc.b	$FB,$DF,$BF,$FD,$F7,$E0,$1F,$BF,$FC,$07,$F8,$07,$E0,$5F,$C0,$7F
	.dc.b	$03,$FE,$05,$F8,$27,$F0,$3F,$C1,$7F,$82,$7E,$0D,$FC,$17,$F0,$7F
	.dc.b	$E0,$DF,$84,$7F,$07,$FC,$27,$F8,$47,$E1,$5F,$C2,$7F,$0B,$FE,$15
	.dc.b	$F8,$67,$F0,$BF,$C3,$7F,$86,$7E,$1D,$FC,$37,$F0,$FF,$E1,$DF,$88
	.dc.b	$7F,$0F,$FC,$47,$F8,$87,$E2,$5F,$C4,$7F,$13,$FE,$25,$F8,$A7,$F1
	.dc.b	$3F,$C5,$7F,$8A,$7E,$2D,$FC,$57,$F1,$7F,$E2,$DF,$8C,$7F,$17,$FC
	.dc.b	$67,$F8,$C7,$E3,$5F,$C6,$7F,$1B,$FE,$35,$F8,$E7,$F1,$BF,$C7,$7F
	.dc.b	$8E,$7E,$3D,$FC,$77,$F1,$FF,$E3,$DF,$90,$7F,$1F,$FC,$87,$F9,$07
	.dc.b	$E4,$5F,$C8,$7F,$23,$FE,$45,$F9,$27,$F2,$3F,$C9,$7F,$92,$7E,$4D
	.dc.b	$FC,$97,$F2,$7F,$E4,$DF,$94,$7F,$27,$FC,$A7,$F9,$47,$E5,$5F,$CA
	.dc.b	$7F,$2B,$FE,$55,$F9,$67,$F2,$BF,$CB,$7F,$96,$7E,$5D,$FC,$B7,$F2
	.dc.b	$FF,$E5,$DF,$98,$7F,$2F,$FC,$C7,$F9,$87,$E6,$5F,$CC,$7F,$33,$FE
	.dc.b	$65,$F9,$A7,$F3,$3F,$CD,$7F,$9A,$7E,$6D,$FC,$D7,$F3,$7F,$E6,$DF
	.dc.b	$9C,$7F,$37,$FC,$E7,$F9,$C7,$E7,$5F,$CE,$7F,$3B,$FE,$75,$F9,$E7
	.dc.b	$F3,$BF,$CF,$7F,$9E,$7E,$7D,$FC,$F7,$F3,$FF,$E7,$DF,$A0,$7F,$3F
	.dc.b	$FD,$07,$FA,$07,$E8,$5F,$D0,$7F,$43,$FE,$85,$FA,$27,$F4,$3F,$D1
	.dc.b	$7F,$A2,$7E,$8D,$FD,$17,$F4,$7F,$E8,$DF,$A4,$7F,$47,$FD,$27,$FA
	.dc.b	$47,$E9,$5F,$D2,$7F,$4B,$FE,$95,$FA,$67,$F4,$BF,$D3,$7F,$A6,$7E
	.dc.b	$9D,$FD,$37,$F4,$FF,$E9,$DF,$A8,$7F,$4F,$FD,$47,$FA,$87,$EA,$5F
	.dc.b	$D4,$7F,$53,$FE,$A5,$FA,$A7,$F5,$3F,$D5,$7F,$AA,$7E,$AD,$FD,$57
	.dc.b	$F5,$7F,$EA,$DF,$AC,$7F,$57,$FD,$67,$FA,$C7,$EB,$5F,$D6,$7F,$5B
	.dc.b	$FE,$B5,$FA,$E7,$F5,$BF,$D7,$7F,$AE,$7E,$BD,$FD,$77,$F5,$FF,$EB
	.dc.b	$DF,$B0,$7F,$5F,$FD,$87,$FB,$07,$EC,$5F,$D8,$7F,$63,$FE,$C5,$FB
	.dc.b	$27,$F6,$3F,$D9,$7F,$B2,$7E,$CD,$FD,$97,$F6,$7F,$EC,$DF,$B4,$7F
	.dc.b	$67,$FD,$A7,$FB,$47,$ED,$5F,$DA,$7F,$6B,$FE,$D5,$FB,$67,$F6,$BF
	.dc.b	$DB,$7F,$B6,$7E,$DD,$FD,$B7,$F6,$FF,$ED,$DF,$B8,$7F,$6F,$FD,$C7
	.dc.b	$FB,$87,$EE,$5F,$DC,$7F,$73,$FE,$E5,$FB,$A7,$F7,$3F,$DD,$7F,$BA
	.dc.b	$7E,$ED,$FD,$D7,$F7,$7F,$EE,$DF,$BC,$7F,$77,$FD,$E7,$FB,$C7,$EF
	.dc.b	$5F,$DE,$7F,$7B,$FE,$F5,$FB,$E7,$F7,$BF,$DF,$7F,$BE,$7E,$FD,$FD
	.dc.b	$F7,$F7,$FF,$EF,$DF,$C0,$1F,$DF,$FF,$80,$7F,$C0,$1F,$C0,$10,$7E
	.dc.b	$BA,$FD,$2D,$FB,$F6,$63,$0D,$8C,$CF,$29,$CB,$63,$2F,$F6,$96,$BF
	.dc.b	$E1,$18,$DB,$99,$E8,$47,$93,$5F,$ED,$9F,$82,$F9,$D7,$EB,$28,$C3
	.dc.b	$63,$D8,$D7,$DE,$D7,$99,$A8,$E2,$47,$04,$38,$BE,$1F,$D6,$5E,$C5
	.dc.b	$24,$FC,$B8,$CB,$63,$2F,$D0,$8C,$B6,$32,$FF,$82,$1A,$FF,$8D,$E3
	.dc.b	$6F,$F8,$2B,$F6,$3F,$85,$FA,$C9,$A4,$8F,$B9,$19,$6C,$65,$FC,$0C
	.dc.b	$65,$B1,$97,$B1,$C0,$FC,$37,$EB,$2F,$C6,$BE,$93,$63,$31,$82,$38
	.dc.b	$21,$F9,$08,$E0,$D8,$DB,$4D,$9F,$E2,$6F,$8C,$FC,$07,$EB,$26,$92
	.dc.b	$3F,$35,$19,$6C,$65,$FC,$FC,$65,$B1,$97,$B1,$C0,$FD,$27,$EB,$0F
	.dc.b	$DC,$FF,$20,$D8,$CC,$60,$8E,$08,$7E,$BA,$38,$36,$36,$D3,$67,$F9
	.dc.b	$07,$F2,$DF,$CF,$7E,$B2,$69,$23,$F7,$11,$96,$C6,$5F,$DE,$C6,$5B
	.dc.b	$19,$7B,$1C,$0F,$E0,$5F,$AC,$BE,$8B,$F5,$CD,$8C,$C6,$08,$E0,$87
	.dc.b	$F0,$C8,$E0,$D8,$DB,$4D,$9F,$E5,$C3,$F7,$9F,$AC,$7E,$F5,$34,$8E
	.dc.b	$88,$7F,$19,$72,$D3,$27,$F1,$E7,$A7,$EB,$0F,$D0,$7F,$85,$36,$7F
	.dc.b	$91,$35,$FE,$4A,$D7,$FC,$E1,$1C,$08,$F0,$6B,$C1,$AF,$03,$F9,$7F
	.dc.b	$EF,$EF,$D0,$1F,$AF,$FF,$31,$8C,$36,$33,$3C,$1A,$FF,$AD,$5A,$FF
	.dc.b	$AE,$E7,$8B,$B9,$E4,$47,$04,$39,$3E,$1F,$D1,$9F,$82,$D2,$4F,$A2
	.dc.b	$32,$D8,$CB,$FE,$C7,$8C,$B6,$32,$F9,$63,$2D,$8C,$BF,$ED,$4F,$DF
	.dc.b	$9F,$7A,$F9,$AF,$EA,$91,$86,$C6,$67,$FA,$CC,$E5,$B1,$97,$FD,$AE
	.dc.b	$D7,$FD,$C1,$1B,$73,$3D,$08,$F2,$6B,$CB,$E9,$7C,$24,$7B,$B5,$D1
	.dc.b	$82,$18,$21,$E2,$A4,$71,$8C,$7D,$87,$B8,$8F,$B1,$AF,$B1,$AF,$03
	.dc.b	$44,$34,$53,$FD,$91,$F9,$87,$E8,$08,$FE,$81,$AF,$E8,$1A,$F0,$34
	.dc.b	$43,$45,$3D,$9F,$7B,$F5,$88,$FD,$46,$BF,$51,$AF,$03,$44,$34,$53
	.dc.b	$E4,$65,$F8,$22,$3F,$80,$6B,$F8,$06,$BC,$0D,$10,$D1,$4F,$A9,$F8
	.dc.b	$9F,$E3,$88,$FE,$31,$AF,$E3,$1A,$F0,$34,$43,$45,$3F,$01,$F8,$F7
	.dc.b	$E6,$88,$FE,$61,$AF,$E6,$1A,$F0,$34,$43,$45,$3F,$19,$D9,$FA,$62
	.dc.b	$3F,$A4,$6B,$FA,$46,$BC,$0D,$10,$D1,$4F,$CC,$7E,$69,$FB,$22,$3F
	.dc.b	$B0,$6B,$FB,$06,$BC,$0D,$10,$D1,$4F,$D2,$7E,$7F,$FB,$E2,$3F,$BC
	.dc.b	$6B,$FB,$C6,$BC,$0D,$10,$D1,$4F,$D8,$7E,$91,$FC,$28,$8F,$F0,$86
	.dc.b	$BF,$C2,$1A,$F0,$34,$43,$45,$3F,$79,$FA,$A7,$F1,$62,$3F,$C5,$1A
	.dc.b	$FF,$14,$6B,$C0,$D1,$0D,$14,$FE,$11,$D5,$FC,$88,$8F,$F2,$06,$BF
	.dc.b	$C8,$1A,$F0,$34,$43,$45,$3F,$8A,$72,$7F,$2E,$23,$FC,$B1,$AF,$F2
	.dc.b	$C6,$BC,$0D,$10,$D1,$4F,$E4,$1F,$B5,$7F,$3A,$23,$FC,$E1,$AF,$F3
	.dc.b	$86,$BC,$0D,$10,$D1,$4F,$E5,$9D,$DF,$D1,$88,$FF,$44,$6B,$FD,$11
	.dc.b	$AF,$03,$44,$34,$53,$F9,$C7,$93,$FD,$48,$8F,$F5,$06,$BF,$D4,$1A
	.dc.b	$F0,$34,$43,$45,$3F,$AA,$FE,$0C,$F5,$3F,$B0,$B6,$3F,$AF,$B5,$1C
	.dc.b	$23,$64,$76,$D7,$67,$F6,$F8,$E4,$F1,$36,$43,$FC,$43,$F9,$5F,$F8
	.dc.b	$98,$8F,$F8,$99,$AE,$8C,$10,$C1,$0F,$08,$EE,$47,$28,$F8,$3F,$03
	.dc.b	$F0,$08,$FC,$06,$BC,$BA,$D6,$9C,$CF,$23,$81,$0E,$11,$B7,$07,$D0
	.dc.b	$F8,$DF,$9E,$47,$E7,$B5,$E2,$D7,$81,$A2,$1A,$29,$F7,$3E,$67,$ED
	.dc.b	$91,$FB,$6D,$7C,$AD,$78,$1A,$21,$A2,$9F,$88,$FA,$4F,$C3,$91,$FC
	.dc.b	$3B,$5F,$82,$D7,$81,$A2,$1A,$29,$F9,$4F,$AE,$FC,$99,$1F,$C9,$B5
	.dc.b	$F9,$AD,$78,$1A,$21,$A2,$9F,$A0,$F1,$7F,$3E,$47,$F3,$ED,$7E,$CB
	.dc.b	$5E,$06,$88,$68,$A7,$EB,$3F,$03,$FD,$59,$1F,$D5,$B5,$FC,$2B,$5E
	.dc.b	$06,$88,$68,$A7,$EE,$3F,$0C,$FD,$B9,$1F,$DB,$B5,$FC,$8B,$5E,$06
	.dc.b	$88,$68,$A7,$F0,$4F,$C5,$FF,$81,$91,$FE,$06,$D7,$F3,$AD,$78,$1A
	.dc.b	$21,$A2,$9F,$C4,$3D,$3F,$E1,$E4,$7F,$87,$B5,$FD,$4B,$5E,$06,$88
	.dc.b	$68,$A7,$F1,$CD,$FF,$8D,$91,$FE,$36,$D7,$F6,$AD,$78,$1A,$21,$A2
	.dc.b	$9F,$CA,$3C,$0F,$E4,$E4,$7F,$93,$B5,$FE,$02,$D7,$81,$A2,$1A,$29
	.dc.b	$FC,$D3,$F3,$FF,$E6,$64,$7F,$99,$B5,$FE,$1A,$D7,$81,$A2,$1A,$29
	.dc.b	$FD,$03,$C9,$FE,$7E,$47,$F9,$FB,$5F,$E3,$2D,$78,$1A,$21,$A2,$9F
	.dc.b	$D3,$3F,$52,$FE,$96,$47,$FA,$5B,$5F,$E4,$AD,$78,$1A,$21,$A2,$9F
	.dc.b	$D6,$3D,$8F,$EA,$E4,$7F,$AB,$B5,$FE,$62,$D7,$81,$A2,$1A,$29,$FD
	.dc.b	$93,$D5,$FE,$C6,$47,$FB,$1B,$5F,$E7,$AD,$78,$1A,$21,$A2,$9F,$DC
	.dc.b	$3D,$4F,$ED,$E4,$7F,$B7,$B5,$FE,$92,$D7,$81,$A2,$1A,$29,$FD,$F3
	.dc.b	$F7,$2F,$EF,$64,$7F,$BD,$B5,$FE,$AA,$D7,$81,$A2,$1A,$29,$FE,$08
	.dc.b	$FD,$F7,$FC,$0E,$47,$FC,$0E,$D7,$FB,$0B,$5E,$06,$88,$68,$A7,$FB
	.dc.b	$A3,$F0,$E7,$BF,$FE,$F4,$6C,$7F,$BC,$9A,$FF,$75,$6B,$C6,$38,$11
	.dc.b	$E0,$D7,$83,$5E,$07,$5F,$8A,$F0,$11,$C1,$0F,$03,$96,$99,$3D,$4E
	.dc.b	$5A,$7A,$7E,$57,$D8,$47,$04,$3D,$8E,$5A,$64,$F8,$CE,$5A,$7C,$5F
	.dc.b	$A0,$F9,$04,$70,$43,$E4,$39,$69,$93,$E9,$39,$69,$F4,$7E,$AB,$EA
	.dc.b	$11,$C1,$0F,$A8,$E5,$A6,$4F,$BC,$E5,$A7,$DD,$F6,$BF,$00,$47,$04
	.dc.b	$3F,$00,$E5,$A6,$4F,$C5,$39,$69,$F8,$9F,$BD,$FC,$61,$1C,$10,$FC
	.dc.b	$63,$96,$99,$3F,$2C,$E5,$A7,$E5,$7F,$07,$FC,$C1,$1C,$10,$FC,$C3
	.dc.b	$96,$99,$3F,$44,$E5,$A7,$E8,$73,$FD,$21,$1C,$10,$FD,$23,$96,$99
	.dc.b	$3F,$5C,$E5,$A7,$EB,$7F,$18,$FD,$81,$1C,$10,$FD,$83,$96,$99,$3F
	.dc.b	$74,$E5,$A7,$EE,$7F,$22,$FD,$E1,$1C,$10,$FD,$E3,$96,$99,$3F,$83
	.dc.b	$39,$69,$FE,$80,$3F,$D2,$FE,$23,$A7,$11,$86,$BF,$E8,$56,$C6,$E3
	.dc.b	$6D,$7A,$35,$FF,$46,$B5,$FF,$4F,$B5,$EC,$72,$8E,$A4,$7C,$DC,$0F
	.dc.b	$C0,$DE,$6F,$9A,$30,$D8,$CC,$F0,$6B,$FE,$9E,$9D,$B9,$9E,$04,$70
	.dc.b	$43,$84,$F2,$7F,$D6,$F3,$C9,$E9,$F6,$3F,$1A,$0D,$24,$79,$DA,$FB
	.dc.b	$1A,$F2,$34,$43,$47,$1F,$DE,$A3,$C5,$AF,$89,$AF,$9A,$11,$88,$FF
	.dc.b	$6A,$CF,$9A,$31,$08,$C3,$D4,$F3,$7E,$E3,$E9,$23,$82,$1F,$09,$CB
	.dc.b	$4C,$9E,$37,$2D,$3E,$17,$ED,$1F,$14,$8E,$08,$7C,$E7,$2D,$32,$7B
	.dc.b	$DC,$B4,$F9,$DF,$09,$F4,$48,$E0,$87,$DA,$72,$D3,$27,$CB,$72,$D3
	.dc.b	$ED,$7E,$BD,$F7,$48,$E0,$87,$E1,$9C,$B4,$C9,$F5,$DC,$B4,$FC,$37
	.dc.b	$EA,$BF,$89,$23,$82,$1F,$92,$72,$D3,$27,$E0,$DC,$B4,$FC,$97,$E9
	.dc.b	$DF,$95,$23,$82,$1F,$9E,$72,$D3,$27,$E3,$DC,$B4,$FC,$F7,$C2,$FE
	.dc.b	$84,$8E,$08,$7E,$A9,$CB,$4C,$9F,$9B,$72,$D3,$F5,$5F,$9E,$7E,$B4
	.dc.b	$8E,$08,$7E,$D9,$CB,$4C,$9F,$A7,$72,$D3,$F6,$DF,$99,$FE,$E4,$8E
	.dc.b	$08,$7F,$02,$72,$D3,$27,$EC,$DC,$B4,$FE,$05,$C7,$FC,$14,$8E,$08
	.dc.b	$7F,$0E,$72,$D3,$27,$EF,$DC,$B4,$FE,$1D,$C9,$FC,$44,$8E,$08,$7F
	.dc.b	$1A,$72,$D3,$27,$F0,$B7,$2D,$3F,$8D,$72,$FF,$1D,$23,$82,$1F,$C9
	.dc.b	$9C,$B4,$C9,$FC,$5D,$CB,$4F,$E4,$DC,$DF,$CA,$48,$E0,$87,$F3,$27
	.dc.b	$2D,$32,$7F,$23,$72,$D3,$F9,$97,$3F,$F3,$52,$38,$21,$FC,$F9,$CB
	.dc.b	$4C,$9F,$CB,$DC,$B4,$FE,$7D,$D1,$FD,$04,$8E,$08,$7F,$4A,$72,$D3
	.dc.b	$27,$F3,$B7,$2D,$3F,$A5,$74,$FF,$4D,$23,$82,$1F,$D5,$9C,$B4,$C9
	.dc.b	$FD,$1D,$CB,$4F,$EA,$DD,$5F,$D6,$48,$E0,$87,$F6,$27,$2D,$32,$7F
	.dc.b	$53,$72,$D3,$FB,$17,$5F,$F6,$52,$38,$21,$FD,$B9,$CB,$4C,$9F,$D7
	.dc.b	$DC,$B4,$FE,$DD,$D9,$FD,$C4,$8E,$08,$7F,$7A,$72,$D3,$27,$F6,$B7
	.dc.b	$2D,$3F,$BD,$76,$FF,$7D,$23,$82,$1F,$E0,$67,$2D,$32,$7F,$77,$72
	.dc.b	$D3,$FC,$43,$F7,$3F,$DD,$5B,$D1,$FE,$21,$8C,$47,$F8,$75,$AF,$F8
	.dc.b	$59,$AF,$F8,$01,$AF,$F7,$F6,$C1,$A4,$7B,$12,$3C,$7D,$FE,$12,$38
	.dc.b	$21,$E3,$72,$D3,$27,$95,$CB,$4F,$7F,$81,$ED,$23,$82,$1E,$F7,$2D
	.dc.b	$32,$7C,$17,$2D,$3E,$5F,$83,$F2,$48,$E0,$87,$CB,$72,$D3,$27,$CD
	.dc.b	$72,$D3,$EB,$F8,$5F,$54,$8E,$08,$7D,$77,$2D,$32,$7D,$97,$2D,$3F
	.dc.b	$07,$E1,$FE,$04,$8E,$08,$7E,$0D,$CB,$4C,$9F,$85,$72,$D3,$F1,$FE
	.dc.b	$27,$E3,$48,$E0,$87,$E3,$DC,$B4,$C9,$F9,$17,$2D,$3F,$37,$E2,$FE
	.dc.b	$64,$8E,$08,$7E,$6D,$CB,$4C,$9F,$9D,$72,$D3,$F4,$FE,$37,$E9,$48
	.dc.b	$E0,$87,$E9,$DC,$B4,$C9,$FA,$97,$2D,$3F,$67,$E3,$FE,$C4,$8E,$08
	.dc.b	$7E,$CD,$CB,$4C,$9F,$B5,$72,$D3,$F7,$FE,$37,$EF,$48,$E0,$87,$EF
	.dc.b	$DC,$B4,$C9,$FC,$05,$CB,$4F,$E1,$7E,$47,$F0,$92,$38,$21,$FC,$2D
	.dc.b	$CB,$4C,$9F,$C3,$5C,$B4,$FE,$2F,$E4,$FF,$15,$23,$82,$1F,$C5,$DC
	.dc.b	$B4,$C9,$FC,$65,$CB,$4F,$E4,$7E,$57,$F2,$12,$38,$21,$FC,$8D,$CB
	.dc.b	$4C,$9F,$C9,$5C,$B4,$FE,$5F,$E5,$FF,$2D,$23,$82,$1F,$CB,$DC,$B4
	.dc.b	$C9,$FC,$C5,$CB,$4F,$E7,$7E,$67,$F3,$92,$38,$21,$FC,$ED,$CB,$4C
	.dc.b	$9F,$CF,$5C,$B4,$FE,$8F,$E6,$FF,$45,$23,$82,$1F,$D1,$DC,$B4,$C9
	.dc.b	$FD,$25,$CB,$4F,$EA,$7E,$77,$F5,$12,$38,$21,$FD,$4D,$CB,$4C,$9F
	.dc.b	$D5,$5C,$B4,$10,$BE,$93,$BD,$E1,$66,$75,$39,$8C,$47,$09,$DB,$89
	.dc.b	$D3,$63,$2F,$38,$CB,$63,$2F,$66,$BD,$67,$BA,$9E,$2E,$E3,$31,$88
	.dc.b	$FE,$E5,$E8,$7F,$79,$23,$FD,$F9,$AF,$F7,$B6,$BF,$DE,$DA,$F2,$38
	.dc.b	$10,$FF,$01,$7A,$5F,$E0,$52,$3F,$E0,$A6,$BA,$30,$43,$04,$3F,$C1
	.dc.b	$4F,$07,$FC,$19,$E9,$FF,$84,$48,$FF,$85,$9A,$FF,$84,$DA,$F0,$34
	.dc.b	$43,$45,$3F,$01,$D4,$F0,$1E,$A7,$CE,$8C,$36,$33,$3F,$80,$9C,$B6
	.dc.b	$32,$FE,$1E,$32,$D8,$CB,$E0,$6B,$F9,$28,$DB,$99,$F0,$11,$ED,$EA
	.dc.b	$F8,$C8,$E0,$87,$77,$2D,$32,$78,$5C,$B4,$F5,$7A,$DE,$F2,$38,$21
	.dc.b	$EB,$72,$D3,$27,$B5,$CB,$4F,$8D,$EB,$FC,$B2,$38,$21,$F1,$DC,$B4
	.dc.b	$C9,$F2,$5C,$B4,$FA,$5E,$C7,$D7,$23,$82,$1F,$4D,$CB,$4C,$9F,$55
	.dc.b	$CB,$4F,$BD,$EC,$FE,$0C,$8E,$08,$7D,$F7,$2D,$32,$7E,$05,$CB,$4F
	.dc.b	$C5,$7A,$FF,$8F,$23,$82,$1F,$8B,$72,$D3,$27,$E3,$5C,$B4,$FC,$B7
	.dc.b	$B5,$F9,$B2,$38,$21,$F9,$77,$2D,$32,$7E,$65,$CB,$4F,$D1,$7B,$7F
	.dc.b	$A7,$23,$82,$1F,$A3,$72,$D3,$27,$E9,$5C,$B4,$FD,$77,$B9,$FB,$32
	.dc.b	$38,$21,$FA,$F7,$2D,$32,$7E,$C5,$CB,$4F,$DD,$7B,$BF,$BF,$23,$82
	.dc.b	$1F,$BB,$72,$D3,$27,$EF,$5C,$B4,$FE,$0D,$EF,$7F,$0B,$23,$82,$1F
	.dc.b	$C1,$DC,$B4,$C9,$FC,$25,$CB,$4F,$E2,$5E,$FF,$F1,$72,$38,$21,$FC
	.dc.b	$4D,$CB,$4C,$9F,$C5,$5C,$B4,$FE,$3D,$F0,$1F,$C8,$C8,$E0,$87,$F1
	.dc.b	$F7,$2D,$32,$7F,$21,$72,$D3,$F9,$57,$C0,$FF,$2F,$23,$82,$1F,$CA
	.dc.b	$DC,$B4,$C9,$FC,$B5,$CB,$4F,$E6,$DF,$05,$FC,$EC,$8E,$08,$7F,$37
	.dc.b	$72,$D3,$27,$F3,$97,$2D,$3F,$BF,$7A,$DF,$DC,$BA,$CD,$24,$68,$87
	.dc.b	$F8,$05,$CB,$4C,$9F,$E0,$37,$2F,$F8,$4F,$E0,$5E,$17,$F8,$12,$11
	.dc.b	$87,$FC,$0E,$D7,$FC,$2E,$D7,$94,$71,$8C,$47,$F8,$8D,$AF,$F8,$99
	.dc.b	$AF,$68,$E0,$47,$BB,$E3,$FD,$D9,$EE,$3A,$7F,$C3,$31,$86,$C6,$6B
	.dc.b	$FC,$57,$39,$6C,$65,$FF,$19,$46,$5B,$19,$7F,$C7,$51,$96,$C6,$5F
	.dc.b	$F1,$FC,$65,$B1,$95,$19,$6F,$56,$63,$FC,$0D,$5E,$A8,$C1,$1C,$12
	.dc.b	$7D,$71,$96,$C6,$5F,$5C,$79,$7B,$9F,$8C,$3A,$30,$DB,$FF,$26,$36
	.dc.b	$FF,$C9,$4D,$7F,$CA,$AD,$7F,$CB,$6D,$7F,$C9,$0D,$AE,$F5,$D8,$95
	.dc.b	$D9,$B5,$DB,$A7,$C6,$49,$F1,$B6,$35,$18,$23,$82,$1E,$68,$E0,$FF
	.dc.b	$88,$FE,$1B,$EB,$4F,$F8,$CD,$B8,$9A,$48,$FF,$18,$36,$7F,$C5,$CD
	.dc.b	$7A,$46,$88,$FF,$8C,$FE,$3F,$A9,$23,$FC,$74,$D8,$D4,$60,$8E,$08
	.dc.b	$7F,$8E,$E3,$83,$FE,$3D,$FA,$FF,$49,$27,$FC,$8C,$D8,$FF,$22,$35
	.dc.b	$E1,$1A,$23,$A6,$BF,$E4,$9F,$A6,$F8,$A4,$8F,$F2,$93,$63,$FC,$A0
	.dc.b	$D7,$84,$68,$8E,$9A,$FF,$95,$7E,$77,$E8,$92,$3F,$CB,$CD,$8F,$F2
	.dc.b	$E3,$5E,$11,$A2,$3A,$6B,$FE,$70,$FB,$BF,$F3,$03,$FE,$67,$6B,$93
	.dc.b	$F0,$6E,$1F,$F3,$33,$5F,$E1,$8D,$83,$49,$1C,$89,$1F,$E7,$98,$D7
	.dc.b	$C3,$F6,$24,$65,$B3,$E5,$6B,$E9,$6B,$E4,$6B,$E4,$7D,$3F,$0F,$27
	.dc.b	$C0,$D9,$F8,$0D,$7E,$23,$5F,$7B,$5F,$7B,$63,$DF,$F0,$FF,$18,$8F
	.dc.b	$B5,$B3,$F3,$5A,$FD,$16,$BF,$31,$AF,$CC,$7E,$8F,$C3,$C9,$F9,$0D
	.dc.b	$9F,$B0,$D7,$EE,$35,$FA,$ED,$7E,$BB,$63,$EB,$F2,$7D,$E2,$3F,$55
	.dc.b	$B3,$F8,$56,$BF,$89,$6B,$F8,$46,$BF,$84,$7F,$13,$F0,$F2,$7F,$00
	.dc.b	$D9,$FC,$83,$5F,$CA,$35,$FC,$7B,$5F,$C7,$B6,$3F,$1F,$F0,$FF,$96
	.dc.b	$23,$F8,$D6,$CF,$E7,$5A,$FE,$85,$AF,$E7,$1A,$FE,$71,$FD,$0F,$C3
	.dc.b	$C9,$FC,$C3,$67,$F5,$0D,$7F,$58,$D7,$F4,$ED,$7F,$4E,$D8,$FD,$3F
	.dc.b	$57,$EB,$88,$FE,$95,$B3,$FB,$56,$BF,$B9,$6B,$FB,$46,$BF,$B4,$7F
	.dc.b	$73,$E6,$C9,$FD,$83,$67,$F8,$03,$5F,$E0,$8D,$7F,$7E,$D7,$F7,$ED
	.dc.b	$8F,$DF,$FC,$3F,$F0,$62,$3F,$BD,$6C,$FF,$0D,$6B,$FC,$45,$AF,$F0
	.dc.b	$C6,$BF,$C3,$1F,$EE,$3E,$4C,$9F,$EE,$2D,$9F,$EE,$2D,$7F,$B8,$B5
	.dc.b	$FE,$E2,$D7,$FB,$8B,$63,$F8,$BF,$C3,$FF,$1E,$23,$FC,$55,$B3,$FC
	.dc.b	$95,$AF,$F2,$96,$BF,$C9,$1A,$FF,$24,$7F,$94,$F7,$C9,$FE,$40,$D9
	.dc.b	$FE,$60,$D7,$F9,$A3,$5F,$E5,$ED,$7F,$97,$B6,$3F,$97,$FC,$3F,$F3
	.dc.b	$62,$3F,$CB,$5B,$3F,$CF,$5A,$FF,$41,$6B,$FC,$F1,$AF,$F3,$C7,$FA
	.dc.b	$0F,$C3,$C9,$FE,$70,$D9,$FE,$90,$D7,$FA,$63,$5F,$E8,$ED,$7F,$A3
	.dc.b	$B6,$3F,$A3,$FC,$3F,$F4,$E2,$3F,$D1,$5B,$3F,$D5,$5A,$FF,$59,$6B
	.dc.b	$FD,$51,$AF,$F5,$47,$FA,$CF,$5C,$9F,$EA,$0D,$9F,$EC,$0D,$7F,$B2
	.dc.b	$35,$FE,$BE,$D7,$FA,$FB,$63,$FA,$FF,$C3,$FF,$66,$23,$FD,$75,$B3
	.dc.b	$FD,$B5,$AF,$F7,$16,$BF,$DB,$1A,$FF,$6C,$7F,$B8,$FC,$3C,$9F,$ED
	.dc.b	$0D,$9F,$EF,$0D,$7F,$BE,$35,$FE,$EE,$D7,$FB,$BB,$63,$FB,$BF,$C3
	.dc.b	$FF,$7E,$23,$FD,$D5,$B3,$FE,$05,$6B,$FE,$09,$6B,$FE,$04,$6B,$FE
	.dc.b	$04,$7F,$C1,$3F,$0F,$27,$FC,$00,$D9,$FF,$08,$35,$FF,$0A,$35,$FF
	.dc.b	$07,$B5,$FF,$07,$B6,$1F,$77,$FE,$9D,$53,$0A,$E8,$FF,$2E,$C6,$E3
	.dc.b	$47,$FA,$95,$C3,$C6,$38,$1A,$74,$F2,$21,$93,$6F,$46,$BC,$E3,$AB
	.dc.b	$E4,$6B,$E1,$F8,$7F,$29,$27,$A4,$A3,$13,$D6,$31,$3A,$9C,$CF,$76
	.dc.b	$BD,$DA,$FB,$FE,$1F,$E1,$92,$7B,$B6,$3B,$35,$F5,$35,$F5,$35,$E9
	.dc.b	$1E,$8F,$87,$F9,$E4,$70,$48,$F5,$46,$5B,$19,$7E,$44,$65,$B1,$F1
	.dc.b	$3E,$1F,$ED,$91,$C1,$23,$E3,$46,$5B,$19,$7E,$A4,$65,$B1,$F4,$3E
	.dc.b	$1F,$F0,$E4,$70,$48,$FA,$51,$96,$C6,$5F,$C0,$46,$5B,$1F,$73,$E1
	.dc.b	$FF,$26,$47,$04,$8F,$BD,$19,$6C,$65,$FC,$64,$65,$B1,$F8,$8F,$87
	.dc.b	$FC,$F9,$1C,$12,$3F,$15,$19,$6C,$65,$FC,$C4,$65,$B1,$F9,$4F,$87
	.dc.b	$FD,$59,$1C,$12,$3F,$2D,$19,$6C,$65,$FD,$24,$65,$B1,$FA,$0F,$87
	.dc.b	$FD,$B9,$1C,$12,$3F,$45,$19,$6C,$65,$FD,$84,$65,$B1,$FA,$CF,$87
	.dc.b	$FE,$06,$47,$04,$8F,$D7,$46,$5B,$19,$7F,$79,$19,$6C,$7E,$E3,$E1
	.dc.b	$FF,$87,$91,$C1,$23,$F7,$51,$96,$C6,$5F,$E1,$11,$96,$C7,$F0,$4F
	.dc.b	$47,$F8,$D9,$1C,$12,$3F,$83,$46,$5B,$19,$7F,$8A,$46,$5B,$1F,$C4
	.dc.b	$3E,$1F,$F9,$39,$1C,$12,$3F,$89,$46,$5B,$19,$7F,$90,$46,$5B,$1F
	.dc.b	$C7,$3E,$1F,$F9,$99,$1C,$12,$3F,$8F,$46,$5B,$19,$7F,$96,$46,$5B
	.dc.b	$1F,$CA,$3E,$1F,$F9,$F9,$1C,$12,$3F,$95,$46,$5B,$19,$7F,$9C,$46
	.dc.b	$5B,$1F,$CD,$3E,$1F,$FA,$59,$1C,$12,$3F,$9B,$46,$5B,$19,$7F,$A2
	.dc.b	$46,$5B,$1F,$D0,$3E,$1F,$FA,$B9,$1C,$12,$3F,$A1,$46,$5B,$19,$7F
	.dc.b	$A8,$46,$5B,$1F,$D3,$3E,$1F,$FB,$19,$1C,$12,$3F,$A7,$46,$5B,$19
	.dc.b	$7F,$AE,$46,$5B,$1F,$D6,$3E,$1F,$FB,$79,$1C,$12,$3F,$AD,$46,$5B
	.dc.b	$19,$7F,$B4,$46,$5B,$1F,$D9,$3E,$1F,$FB,$D9,$1C,$12,$3F,$B3,$46
	.dc.b	$5B,$19,$7F,$BA,$46,$5B,$1F,$DE,$FE,$ED,$F1,$08,$C3,$67,$FB,$F4
	.dc.b	$EA,$72,$48,$CC,$ED,$FF,$06,$4E,$DC,$CE,$DC,$CE,$DF,$F0,$CC,$ED
	.dc.b	$CC,$F9,$7E,$23,$51,$E5,$94,$62,$51,$87,$81,$27,$D3,$3B,$9D,$46
	.dc.b	$27,$4F,$A2,$74,$48,$F8,$DE,$37,$12,$38,$24,$FC,$68,$CB,$63,$2F
	.dc.b	$C4,$8C,$B6,$3E,$97,$B1,$E5,$23,$82,$47,$D2,$8C,$B6,$32,$FD,$08
	.dc.b	$CB,$63,$EF,$7B,$3F,$04,$8E,$09,$1F,$7A,$32,$D8,$CB,$F7,$23,$2D
	.dc.b	$8F,$C5,$7C,$47,$CD,$23,$82,$47,$E2,$A3,$2D,$8C,$BF,$88,$8C,$B6
	.dc.b	$3F,$2D,$F1,$1F,$64,$8E,$09,$1F,$96,$8C,$B6,$32,$FE,$52,$32,$D8
	.dc.b	$FD,$17,$C4,$7E,$14,$8E,$09,$1F,$A2,$8C,$B6,$32,$FE,$82,$32,$D8
	.dc.b	$FD,$77,$C4,$7E,$44,$8E,$09,$1F,$AE,$8C,$B6,$32,$FE,$B2,$32,$D8
	.dc.b	$FD,$D7,$C4,$7E,$74,$8E,$09,$1F,$BA,$8C,$B6,$32,$FE,$E2,$32,$D8
	.dc.b	$FF,$40,$7E,$7C,$F0,$3F,$60,$D8,$CC,$60,$8E,$09,$3F,$E8,$79,$E0
	.dc.b	$D9,$DB,$67,$6F,$FA,$63,$E0,$9E,$47,$F0,$28,$46,$27,$FD,$31,$18
	.dc.b	$9D,$4E,$63,$FD,$38,$D7,$A4,$68,$93,$CA,$75,$3F,$EA,$D6,$BF,$EB
	.dc.b	$06,$BD,$E3,$FD,$51,$F0,$2F,$82,$FE,$1B,$3F,$EB,$58,$C7,$17,$F1
	.dc.b	$A6,$D7,$0A,$DD,$7F,$AF,$27,$2F,$FA,$E1,$AF,$38,$DA,$8C,$BC,$63
	.dc.b	$31,$88,$D4,$60,$9C,$9E,$37,$F9,$84,$71,$8C,$12,$BC,$CD,$8F,$24
	.dc.b	$F0,$7A,$C6,$5F,$33,$5F,$67,$C3,$39,$BE,$13,$6F,$E4,$B6,$7E,$03
	.dc.b	$5F,$94,$D7,$9C,$F2,$24,$F2,$6C,$72,$6B,$CB,$9F,$C2,$49,$F0,$B6
	.dc.b	$3C,$2D,$78,$46,$88,$E9,$AF,$9B,$A3,$DA,$48,$F6,$B6,$3D,$AD,$78
	.dc.b	$46,$88,$E9,$AF,$C2,$E9,$F9,$24,$8F,$92,$D8,$F9,$2D,$78,$46,$88
	.dc.b	$E9,$AF,$CE,$EA,$FA,$A4,$8F,$AA,$D8,$FA,$AD,$78,$46,$88,$E9,$AF
	.dc.b	$DA,$EB,$FC,$09,$23,$F0,$2D,$8F,$C0,$B5,$E1,$1A,$23,$A6,$BF,$86
	.dc.b	$EC,$FC,$69,$23,$F1,$AD,$8F,$C6,$B5,$E1,$1A,$23,$A6,$BF,$92,$ED
	.dc.b	$FC,$C9,$23,$F3,$2D,$8F,$CC,$B5,$E1,$1A,$23,$A6,$BF,$9E,$EE,$FD
	.dc.b	$29,$23,$F4,$AD,$8F,$D2,$B5,$E1,$1A,$23,$A6,$BF,$AA,$EF,$FD,$89
	.dc.b	$23,$F6,$2D,$8F,$D8,$B5,$E1,$1A,$23,$A6,$BF,$B6,$F0,$3F,$7A,$48
	.dc.b	$FD,$EB,$63,$F7,$AD,$78,$46,$88,$E9,$AF,$F0,$2F,$07,$F8,$49,$23
	.dc.b	$F8,$4B,$63,$F8,$4B,$5E,$11,$A2,$3A,$6B,$FC,$3B,$C2,$FE,$2A,$48
	.dc.b	$FE,$2A,$D8,$FE,$2A,$D7,$84,$68,$8E,$9A,$FF,$1A,$F0,$FF,$90,$92
	.dc.b	$3F,$90,$B6,$3F,$90,$B5,$E1,$1A,$23,$A6,$BF,$C9,$BC,$4F,$E5,$A4
	.dc.b	$8F,$E5,$AD,$8F,$E5,$AD,$78,$46,$88,$E9,$AF,$F3,$2F,$17,$F9,$C9
	.dc.b	$23,$F9,$CB,$63,$F9,$CB,$5E,$11,$A2,$3A,$6B,$FC,$FB,$AF,$FA,$29
	.dc.b	$23,$FA,$2B,$63,$FA,$2B,$5E,$11,$A2,$3A,$6B,$FD,$2B,$C6,$FE,$A2
	.dc.b	$48,$FE,$A2,$D8,$FE,$A2,$D7,$84,$68,$8E,$9A,$FF,$56,$F1,$FF,$AE
	.dc.b	$92,$3F,$AE,$B6,$3F,$AE,$B5,$E1,$1A,$23,$A6,$BF,$D8,$BC,$8F,$ED
	.dc.b	$24,$8F,$ED,$2D,$8F,$ED,$2D,$78,$46,$88,$E9,$AF,$F6,$E7,$FB,$A9
	.dc.b	$0F,$EE,$A5,$7F,$BA,$B5,$E0,$E8,$8E,$9A,$FF,$88,$FF,$1A,$79,$7F
	.dc.b	$E2,$36,$CF,$F8,$A9,$B3,$FE,$21,$6B,$C6,$78,$12,$78,$36,$38,$35
	.dc.b	$E1,$1D,$7C,$CF,$01,$1C,$12,$3C,$11,$96,$C6,$5F,$54,$65,$B1,$E9
	.dc.b	$F0,$7D,$84,$70,$48,$F6,$46,$5B,$19,$7E,$34,$65,$B1,$F1,$7C,$DF
	.dc.b	$90,$47,$04,$8F,$91,$19,$6C,$65,$FA,$51,$96,$C7,$D1,$F3,$BE,$A1
	.dc.b	$1C,$12,$3E,$A4,$65,$B1,$97,$EF,$46,$5B,$1F,$77,$CF,$FC,$01,$1C
	.dc.b	$12,$3F,$01,$19,$6C,$65,$FC,$54,$65,$B1,$F8,$9F,$43,$F1,$84,$70
	.dc.b	$48,$FC,$64,$65,$B1,$97,$F2,$D1,$96,$C7,$E5,$7D,$1F,$CC,$11,$C1
	.dc.b	$23,$F3,$11,$96,$C6,$5F,$D1,$46,$5B,$1F,$E9,$1C,$BF,$D3,$24,$70
	.dc.b	$48,$FF,$4D,$C6,$5B,$19,$7F,$D5,$11,$96,$C7,$EB,$7D,$2F,$D8,$11
	.dc.b	$C1,$23,$F6,$11,$96,$C6,$5F,$DD,$46,$5B,$1F,$B9,$F4,$FF,$78,$47
	.dc.b	$04,$8F,$DE,$46,$5B,$19,$7F,$83,$46,$5B,$1F,$C1,$7D,$4F,$E1,$04
	.dc.b	$70,$48,$FE,$11,$19,$6C,$65,$FE,$25,$19,$6C,$7F,$11,$F5,$7F,$8A
	.dc.b	$11,$C1,$23,$F8,$A4,$65,$B1,$97,$F8,$F4,$65,$B1,$FC,$77,$D6,$FE
	.dc.b	$40,$47,$04,$8F,$E4,$11,$96,$C6,$5F,$E5,$51,$96,$C7,$F2,$9F,$5F
	.dc.b	$F9,$61,$1C,$12,$3F,$96,$46,$5B,$19,$7F,$9B,$46,$5B,$1F,$CD,$7D
	.dc.b	$8F,$E7,$04,$70,$48,$FE,$71,$19,$6C,$65,$FE,$85,$19,$6C,$7F,$41
	.dc.b	$F6,$7F,$A2,$11,$C1,$23,$FA,$24,$65,$B1,$97,$FA,$74,$65,$B1,$FD
	.dc.b	$37,$DA,$FE,$A0,$47,$04,$8F,$EA,$11,$96,$C6,$5F,$EB,$51,$96,$C7
	.dc.b	$F5,$9F,$6F,$FA,$E1,$1C,$12,$3F,$AE,$46,$5B,$19,$7F,$B3,$46,$5B
	.dc.b	$1F,$D9,$7D,$CF,$ED,$04,$70,$48,$FE,$D1,$19,$6C,$65,$FE,$E5,$19
	.dc.b	$6C,$7F,$71,$F7,$7F,$BA,$11,$C1,$23,$FB,$A4,$65,$B1,$97,$FB,$F4
	.dc.b	$65,$B1,$FD,$F7,$DE,$FF,$00,$11,$C1,$23,$FC,$01,$19,$6C,$65,$FF
	.dc.b	$05,$46,$5B,$1F,$E0,$97,$FC,$18,$47,$04,$7F,$C1,$86,$4A,$E5,$FF
	.dc.b	$0B,$19,$2A,$61,$7D,$17,$C7,$BE,$05,$5A,$AD,$4E,$63,$10,$AE,$35
	.dc.b	$C2,$38,$46,$27,$6D,$8C,$D7,$36,$D7,$17,$CB,$3B,$6C,$ED,$B3,$B7
	.dc.b	$CA,$D7,$C9,$1B,$7D,$FE,$AF,$F8,$C9,$BD,$9B,$51,$97,$E3,$35,$CC
	.dc.b	$6E,$30,$47,$04,$C3,$E2,$CF,$06,$CF,$5C,$3E,$7F,$C1,$79,$DB,$3F
	.dc.b	$1B,$04,$E5,$B1,$98,$C1,$1C,$12,$FE,$6C,$65,$B1,$D7,$E0,$FA,$11
	.dc.b	$C1,$23,$EA,$C6,$5B,$19,$7E,$9C,$65,$B1,$E9,$F8,$4F,$41,$1C,$12
	.dc.b	$3F,$03,$19,$6C,$65,$FB,$F1,$96,$C7,$E4,$3E,$45,$FA,$9B,$83,$F0
	.dc.b	$11,$86,$C6,$67,$83,$6B,$F3,$95,$B9,$CD,$70,$23,$82,$4F,$0A,$E4
	.dc.b	$FE,$AA,$B9,$3D,$39,$1C,$7E,$46,$DF,$A5,$B3,$B9,$D1,$27,$4D,$8D
	.dc.b	$46,$08,$E0,$FD,$9F,$EE,$CF,$13,$F3,$8D,$9F,$C9,$B6,$77,$3A,$24
	.dc.b	$E9,$B1,$A8,$C1,$1C,$1F,$C0,$A1,$EF,$7E,$AD,$B3,$FA,$26,$CE,$E7
	.dc.b	$44,$9D,$36,$35,$18,$23,$83,$F8,$8F,$EE,$CF,$07,$F7,$4D,$9F,$D8
	.dc.b	$B6,$77,$3A,$24,$E9,$B1,$A8,$C1,$1C,$1F,$C8,$3F,$75,$7B,$9F,$C2
	.dc.b	$5B,$3F,$C0,$1B,$3B,$9D,$12,$74,$D8,$D4,$60,$8E,$0F,$E5,$FD,$2D
	.dc.b	$7F,$8C,$36,$7F,$87,$B6,$77,$3A,$24,$E9,$B1,$A8,$C1,$1C,$1F,$CF
	.dc.b	$BF,$75,$7A,$BF,$C9,$DB,$3F,$C7,$9B,$3B,$9D,$12,$74,$D8,$D4,$60
	.dc.b	$8E,$0F,$E9,$BF,$BA,$FF,$39,$7F,$AC,$35,$34,$9D,$7F,$CE,$DB,$3F
	.dc.b	$CC,$9B,$3D,$27,$44,$9F,$E6,$FF,$10,$F4,$3F,$A2,$36,$33,$18,$23
	.dc.b	$82,$57,$F6,$49,$E0,$D9,$DB,$67,$6F,$F6,$DF,$DD,$1F,$23,$FD,$4D
	.dc.b	$B3,$FD,$29,$B3,$FD,$C1,$AF,$19,$E0,$49,$E0,$D8,$E0,$D7,$84,$7F
	.dc.b	$56,$FD,$D1,$E7,$FF,$63,$6C,$66,$30,$47,$04,$9F,$F0,$1C,$F0,$6C
	.dc.b	$ED,$B3,$B7,$FC,$25,$F4,$42,$00

;--------------------------------------------------------------------------------
;	FMOVE.D FPn,<mem>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovedregto_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.D FPN,<EA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.D FPn,<ea>',13,10
;------------------------------------------------
;	d1-d2	actual result
;	d4	actual status
;	d6	fpcr=(XRN..DRP)<<4
;	d7	0=failed,1=successful
;	a0	expected result handle,expected status,...
;	a1-a2	expected result
;	a4	expected status
;	a5	source handle,...
;	fp5	source
;------------------------------------------------
	lea.l	push_decompressed,a3
;decompress data
	move.l	a3,-(sp)
	pea.l	fmovedregto_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a3,a5			;source handle,...
@@:
	add.l	d0,(a3)+		;source handle
	tst.l	(a3)
	bpl	@b
	addq.l	#4,a3			;-1
	movea.l	a3,a0			;expected result handle,expected status,...
@@:
	add.l	d0,(a3)+		;expected result handle
	addq.l	#4,a3			;expected status
	tst.l	(a3)
	bpl	@b
;	addq.l	#4,a3			;-1
;
55:
	move.l	#0,d6		;fpcr=(XRN..DRP)<<4
66:
;FPn,<mem>
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	#-1,-(sp)		;fill=-1
	move.l	#-1,-(sp)
					;source
	fmove.d	fp5,(sp)		;EXECUTE
	movem.l	(sp)+,d1-d2		;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movem.l	([a0]),a1-a2		;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1-d2/d4/a1-a2/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_double
	lea.l	24(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.D FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',<mem> @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d2/d4/a1-a2/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_double
	lea.l	28(sp),sp
@@:
;
	addq.l	#8,a0			;expected
;
	add.w	#1<<4,d6		;rprm++
	cmp.w	#176,d6		;fpcr=(XRN..DRP)<<4
	bls	66b
;
	addq.l	#4,a5			;source handle
	tst.l	(a5)			;source handle
	bpl	55b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmovedregto_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$F1,$F7,$EB,$CF,$80,$FB,$F1,$FC,$D3,$E3,$53,$98
	.dc.b	$CC,$62,$77,$1C,$9B,$19,$8D,$CF,$17,$F1,$71,$96,$C6,$5F,$E5,$F1
	.dc.b	$96,$C6,$5F,$23,$5F,$01,$DA,$7F,$9B,$FD,$77,$FA,$C6,$BD,$3E,$76
	.dc.b	$3D,$3E,$35,$6F,$E0,$CD,$AF,$25,$F1,$C1,$39,$7D,$51,$98,$C5,$E5
	.dc.b	$B7,$FD,$72,$74,$DC,$3C,$2D,$3C,$25,$C3,$C3,$1E,$EA,$E7,$2A,$E6
	.dc.b	$D9,$E6,$D3,$99,$6B,$9B,$89,$F6,$5F,$9E,$3C,$18,$7D,$5F,$87,$EA
	.dc.b	$6A,$3F,$BA,$B8,$C3,$BB,$63,$E8,$9F,$D8,$E3,$D0,$F0,$76,$61,$D1
	.dc.b	$C2,$39,$61,$E6,$3E,$BC,$79,$2F,$CC,$FA,$EF,$E8,$DF,$3C,$3C,$D3
	.dc.b	$9C,$7E,$3C,$7C,$63,$25,$9F,$8E,$D9,$F4,$B7,$0F,$15,$FE,$02,$7C
	.dc.b	$47,$C6,$BE,$B7,$E1,$39,$96,$39,$B6,$B9,$CF,$09,$FA,$53,$C1,$A7
	.dc.b	$02,$CF,$07,$C2,$7B,$4B,$7E,$7B,$F8,$71,$C1,$B3,$F2,$A3,$E4,$61
	.dc.b	$F2,$AB,$E1,$DF,$96,$F6,$7B,$EB,$67,$BC,$EF,$1B,$AE,$4F,$A6,$FB
	.dc.b	$5F,$4A,$F1,$56,$9F,$64,$EA,$B8,$5F,$7B,$E1,$39,$AD,$CE,$6B,$78
	.dc.b	$7C,$9A,$D1,$F1,$6B,$CE,$7C,$52,$DF,$C5,$C3,$C5,$3B,$8E,$F8,$75
	.dc.b	$AE,$F7,$D6,$31,$5B,$C3,$C3,$5F,$08,$F6,$9F,$14,$EE,$FC,$57,$17
	.dc.b	$DA,$FA,$12,$39,$57,$AA,$FE,$8C,$E6,$79,$9D,$8F,$65,$73,$78,$CF
	.dc.b	$98,$E2,$E2,$79,$4E,$E3,$9D,$77,$C3,$E1,$DF,$09,$CB,$CE,$32,$48
	.dc.b	$F5,$E1,$F0,$9D,$E1,$DA,$31,$87,$72,$38,$24,$FC,$E3,$CC,$43,$99
	.dc.b	$63,$8B,$B3,$24,$73,$87,$D5,$8D,$46,$30,$F6,$37,$0F,$54,$FC,$AB
	.dc.b	$F1,$DF,$12,$5F,$08,$CB,$67,$AC,$62,$B9,$DF,$E3,$67,$E2,$9A,$25
	.dc.b	$F4,$9D,$D7,$E1,$67,$C4,$6A,$FB,$56,$E3,$9B,$5D,$B6,$FC,$11,$89
	.dc.b	$E2,$4B,$EF,$1F,$89,$9D,$92,$FC,$4D,$8D,$46,$08,$E0,$97,$F8,$97
	.dc.b	$CB,$7D,$DB,$7E,$09,$D4,$79,$63,$13,$F8,$D8,$D9,$0D,$16,$FC,$95
	.dc.b	$C9,$B3,$C6,$76,$48,$CE,$1F,$93,$8F,$99,$7E,$08,$C5,$FC,$9C,$3D
	.dc.b	$AF,$56,$E1,$D7,$0E,$78,$7E,$4E,$7F,$00,$F9,$5D,$DF,$4A,$F8,$53
	.dc.b	$E3,$9C,$BC,$4F,$19,$C4,$F9,$71,$DE,$37,$1F,$9D,$7B,$BC,$CD,$CF
	.dc.b	$7C,$3D,$91,$D7,$0F,$63,$A3,$17,$E4,$C3,$D7,$1D,$4F,$19,$D6,$7E
	.dc.b	$13,$C4,$D3,$D0,$C9,$27,$D1,$1B,$9F,$D7,$3E,$D9,$D5,$F9,$E3,$15
	.dc.b	$C2,$73,$1E,$6B,$F5,$BE,$09,$F4,$61,$F1,$30,$EB,$3E,$79,$E4,$4C
	.dc.b	$3B,$46,$5B,$3F,$00,$E2,$5A,$F9,$AD,$9F,$86,$DC,$3E,$05,$F5,$9F
	.dc.b	$31,$EA,$BE,$B7,$E5,$39,$96,$3F,$79,$06,$7D,$D2,$FF,$C4,$4D,$8F
	.dc.b	$E3,$CD,$8F,$EF,$0D,$BF,$EE,$8D,$C7,$FC,$46,$E6,$FB,$5A,$30,$F8
	.dc.b	$2F,$F9,$C4,$62,$3F,$CD,$F3,$FE,$76,$BF,$F0,$04,$7F,$9B,$A3,$04
	.dc.b	$7F,$CF,$2D,$78,$98,$21,$83,$FC,$F7,$DD,$D4,$8F,$FA,$09,$AE,$8C
	.dc.b	$10,$C1,$0F,$F4,$0B,$C1,$FF,$41,$FC,$1F,$A4,$8F,$FA,$39,$AF,$FA
	.dc.b	$31,$AF,$03,$44,$34,$53,$FD,$1F,$E8,$FC,$52,$3F,$E9,$A6,$BF,$E9
	.dc.b	$86,$BC,$0D,$10,$D1,$4F,$F4,$DF,$7F,$D1,$23,$FE,$A6,$6B,$FE,$A4
	.dc.b	$6B,$C0,$D1,$0D,$14,$FF,$53,$FC,$17,$E1,$32,$FC,$57,$A3,$74,$FF
	.dc.b	$54,$CD,$AF,$F6,$0C,$D9,$FF,$5F,$CD,$D1,$36,$30,$DA,$FC,$71,$C8
	.dc.b	$B5,$C9,$B5,$C9,$B5,$C9,$A6,$AB,$A3,$F9,$BA,$FF,$5F,$C6,$2F,$D7
	.dc.b	$3C,$0D,$1E,$12,$CF,$85,$B7,$E1,$6F,$81,$ED,$3D,$DF,$03,$63,$CF
	.dc.b	$15,$E3,$16,$FF,$69,$36,$FF,$DA,$73,$6B,$FD,$D7,$36,$7F,$DD,$33
	.dc.b	$6D,$36,$30,$DA,$F0,$9F,$23,$D1,$FA,$5E,$0F,$D3,$6C,$64,$F9,$B1
	.dc.b	$96,$99,$3E,$0B,$96,$99,$C7,$F8,$74,$65,$B1,$9C,$7F,$DD,$F3,$96
	.dc.b	$C6,$4F,$E3,$98,$A7,$51,$87,$F1,$5F,$01,$F8,$BF,$46,$CF,$E1,$1A
	.dc.b	$FD,$66,$E1,$FC,$85,$B8,$7F,$C0,$06,$CF,$81,$B0,$C1,$18,$6C,$FE
	.dc.b	$4A,$FC,$4D,$3C,$45,$3C,$45,$BF,$13,$F9,$AF,$0B,$F4,$87,$E9,$7D
	.dc.b	$FB,$19,$21,$F9,$07,$2F,$E4,$A2,$CF,$F3,$D9,$B8,$FF,$47,$3D,$53
	.dc.b	$8A,$F9,$B3,$67,$31,$DA,$B9,$36,$39,$60,$BE,$0A,$EF,$A3,$FD,$04
	.dc.b	$97,$FD,$28,$E2,$E1,$C5,$FC,$D8,$C4,$7F,$50,$8C,$3C,$4F,$C1,$C6
	.dc.b	$9E,$66,$63,$6F,$4F,$81,$7C,$4F,$F1,$16,$C7,$F0,$8A,$C4,$C8,$FE
	.dc.b	$1B,$3A,$7F,$7B,$39,$7F,$80,$C6,$72,$FE,$D8,$DB,$F1,$CE,$AF,$83
	.dc.b	$EB,$9E,$8D,$8F,$63,$FD,$E6,$32,$FD,$76,$B9,$8D,$CA,$72,$D8,$CC
	.dc.b	$60,$8E,$0C,$FC,$0C,$AF,$91,$28,$C4,$7C,$96,$BE,$96,$BE,$56,$BE
	.dc.b	$96,$D7,$97,$E3,$7C,$64,$9E,$ED,$7E,$B3,$5F,$8C,$D7,$E1,$35,$F8
	.dc.b	$CF,$BB,$C6,$91,$EA,$6B,$F8,$16,$BF,$45,$AF,$CD,$6B,$F4,$5B,$3F
	.dc.b	$37,$E0,$7E,$59,$27,$E3,$B5,$FC,$73,$5F,$BC,$D7,$ED,$35,$FB,$CF
	.dc.b	$D6,$F5,$7F,$29,$E3,$FE,$55,$B3,$F9,$26,$BF,$90,$6E,$9F,$90,$6D
	.dc.b	$FE,$1D,$B3,$F8,$B6,$C7,$E1,$DB,$99,$A4,$9F,$D0,$DF,$8D,$A7,$E9
	.dc.b	$3E,$03,$F5,$47,$EA,$FD,$5B,$19,$21,$FA,$67,$2F,$EA,$22,$E3,$FA
	.dc.b	$89,$B8,$FE,$7E,$7F,$32,$4A,$FC,$AB,$5F,$DE,$35,$FD,$63,$5E,$47
	.dc.b	$02,$1C,$0F,$DE,$F9,$1E,$07,$F6,$0D,$9F,$D6,$37,$23,$49,$3E,$EC
	.dc.b	$3F,$80,$C6,$23,$DC,$76,$35,$5F,$BB,$8D,$4F,$F0,$8B,$FE,$29,$1F
	.dc.b	$C5,$7E,$02,$5F,$F1,$2F,$56,$BF,$C4,$22,$DF,$F1,$09,$B8,$7F,$07
	.dc.b	$9B,$5F,$C2,$E6,$CF,$F0,$79,$B8,$A6,$C6,$1B,$5F,$C8,$A3,$F9,$44
	.dc.b	$62,$3F,$96,$47,$F0,$E2,$3F,$C2,$9A,$E8,$C1,$0C,$10,$FE,$48,$F0
	.dc.b	$BF,$E5,$7E,$4F,$98,$93,$E3,$3F,$9A,$B8,$7E,$21,$83,$47,$F3,$67
	.dc.b	$51,$FC,$DA,$7F,$A1,$46,$30,$F9,$15,$C4,$E6,$71,$39,$D7,$91,$B1
	.dc.b	$FC,$D6,$79,$47,$F3,$9A,$EA,$E8,$FE,$79,$1A,$9F,$E7,$1E,$FF,$F3
	.dc.b	$62,$4F,$F3,$26,$BF,$D6,$5A,$FF,$49,$6B,$FD,$15,$AF,$F4,$97,$FA
	.dc.b	$0F,$AB,$23,$F9,$EB,$5F,$EC,$8D,$7F,$AA,$35,$FE,$A0,$D7,$FA,$A3
	.dc.b	$67,$FA,$87,$8D,$FD,$38,$93,$FD,$29,$AF,$F7,$16,$BF,$D8,$5A,$FF
	.dc.b	$5D,$6B,$FD,$85,$FE,$B3,$E3,$48,$FE,$AA,$D7,$FB,$E3,$5F,$ED,$8D
	.dc.b	$7F,$B4,$35,$FE,$D8,$D9,$FE,$D1,$E3,$7F,$66,$24,$FF,$62,$6B,$FE
	.dc.b	$09,$6B,$FD,$E5,$AF,$F7,$56,$BF,$DE,$5F,$EE,$3E,$34,$8F,$ED,$AD
	.dc.b	$7F,$C2,$8D,$7F,$C0,$8D,$7F,$C0,$0D,$7F,$C0,$8D,$9F,$F0,$07,$8D
	.dc.b	$FD,$F8,$93,$FD,$E9,$AF,$F8,$85,$AF,$F8,$45,$AF,$F8,$35,$AF,$F8
	.dc.b	$45,$FF,$04,$F8,$D2,$3F,$C0,$AD,$7F,$C5,$8D,$7F,$C3,$8D,$7F,$C3
	.dc.b	$0D,$7F,$C3,$8D,$9F,$F0,$C7,$8D,$FE,$16,$24,$FF,$84,$9A,$FF,$8E
	.dc.b	$5A,$FF,$8A,$5A,$FF,$89,$5A,$FF,$8A,$5F,$F1,$0F,$8D,$FE,$3F,$F1
	.dc.b	$7F,$C8,$0D,$9F,$F1,$EB,$5F,$F1,$DB,$78,$7F,$C7,$6D,$BF,$F1,$93
	.dc.b	$67,$FC,$6D,$56,$7F,$C6,$53,$78,$FF,$D6,$35,$FE,$4A,$AF,$0B,$48
	.dc.b	$7F,$92,$9C,$B4,$CD,$FF,$93,$23,$2D,$8C,$DF,$F9,$16,$72,$D8,$C9
	.dc.b	$E6,$3D,$E6,$1F,$F2,$44,$79,$EF,$D0,$D9,$D5,$FA,$9F,$21,$F1,$8C
	.dc.b	$C6,$DC,$DF,$95,$B7,$F2,$1D,$9D,$2F,$BB,$4C,$9C,$0F,$F3,$13,$A7
	.dc.b	$11,$F3,$9D,$CF,$73,$EB,$B8,$8F,$84,$F9,$CE,$37,$EB,$9E,$13,$E5
	.dc.b	$9C,$DF,$AE,$73,$1F,$24,$E3,$7E,$38,$CD,$78,$E7,$89,$EB,$29,$C0
	.dc.b	$F2,$1C,$E3,$CE,$E9,$E7,$6A,$ED,$5E,$A9,$D9,$AB,$F3,$46,$49,$7D
	.dc.b	$5F,$64,$EE,$3C,$8D,$7C,$87,$86,$38,$9E,$B8,$D1,$E4,$AF,$1C,$F5
	.dc.b	$3E,$A9,$EC,$9F,$C5,$BF,$8E,$3A,$1B,$7A,$1D,$AF,$E9,$36,$7E,$F4
	.dc.b	$FF,$A9,$6F,$FD,$32,$47,$FD,$2A,$D7,$FD,$78,$D7,$81,$A2,$1A,$2D
	.dc.b	$FF,$AD,$BC,$6F,$DB,$12,$3F,$3E,$D7,$F0,$ED,$8F,$AA,$DB,$F5,$36
	.dc.b	$7D,$D8,$23,$13,$F1,$27,$FD,$8F,$18,$9F,$83,$1F,$EB,$F8,$C5,$FF
	.dc.b	$B5,$5B,$1F,$ED,$26,$BC,$8F,$39,$0F,$39,$6F,$CE,$DB,$F3,$B4,$F3
	.dc.b	$CF,$49,$F8,$91,$99,$F4,$46,$23,$A9,$E1,$23,$D9,$B1,$FE,$D7,$9F
	.dc.b	$71,$EB,$8E,$97,$F1,$5F,$61,$88,$D3,$DA,$BC,$04,$AF,$4C,$76,$3E
	.dc.b	$41,$C4,$E6,$71,$7C,$4E,$23,$CC,$73,$78,$9B,$2B,$98,$F8,$27,$28
	.dc.b	$F0,$3F,$EE,$F9,$F6,$C7,$58,$F2,$CF,$1A,$F7,$4F,$59,$E0,$EC,$FA
	.dc.b	$27,$A0,$1D,$CA,$79,$8A,$F2,$72,$64,$E0,$43,$04,$78,$3C,$A3,$CA
	.dc.b	$DA,$E6,$FB,$E9,$5E,$41,$4F,$1B,$EB,$CE,$AB,$73,$97,$DD,$5C,$2B
	.dc.b	$6F,$39,$D4,$FC,$C6,$D7,$C9,$AF,$82,$F5,$AF,$0F,$8D,$87,$93,$6F
	.dc.b	$C3,$E2,$D9,$DC,$7C,$CA,$E5,$82,$F4,$D9,$D5,$79,$6F,$35,$27,$27
	.dc.b	$B9,$E3,$19,$85,$F6,$9F,$97,$1B,$3E,$6D,$79,$1F,$B2,$D7,$EA,$47
	.dc.b	$08,$C4,$7B,$8E,$37,$E7,$78,$B5,$E2,$6A,$79,$46,$63,$91,$E6,$9F
	.dc.b	$A5,$E3,$7A,$6F,$E2,$B6,$33,$3F,$01,$F4,$4E,$A3,$15,$E3,$8C,$47
	.dc.b	$53,$E0,$16,$7F,$13,$5D,$63,$F0,$ED,$8C,$C7,$4B,$FA,$B3,$9B,$34
	.dc.b	$93,$B3,$C6,$E1,$F0,$47,$96,$BD,$71,$8A,$EA,$F1,$37,$5C,$6F,$ED
	.dc.b	$47,$43,$4F,$B8,$CB,$C0,$CD,$76,$8C,$9B,$7E,$8B,$63,$E2,$B6,$3E
	.dc.b	$69,$C4,$D9,$63,$6D,$36,$FC,$A3,$C2,$53,$C8,$48,$F1,$C7,$1B,$F9
	.dc.b	$73,$ED,$3E,$21,$67,$C9,$1A,$8C,$9C,$E7,$95,$7C,$76,$C7,$89,$FB
	.dc.b	$6E,$CF,$25,$78,$9B,$1C,$1E,$B1,$E0,$9F,$AB,$18,$9D,$4F,$92,$F2
	.dc.b	$EE,$7C,$17,$E7,$8C,$47,$13,$B9,$2B,$CB,$1E,$36,$9D,$8F,$A3,$1B
	.dc.b	$9E,$91,$F2,$6B,$C2,$EE,$7E,$A9,$BB,$CB,$88,$EA,$E4,$F3,$C6,$A3
	.dc.b	$27,$DC,$2D,$7D,$19,$FE,$71,$3F,$BC,$23,$FC,$11,$AF,$F0,$46,$BC
	.dc.b	$0D,$10,$D1,$6F,$F9,$2F,$8D,$E3,$A3,$49,$3F,$85,$BF,$E7,$51,$88
	.dc.b	$FB,$D1,$FD,$1A,$31,$5F,$8A,$6C,$7E,$45,$B3,$F3,$5B,$1F,$01,$B5
	.dc.b	$F4,$27,$FA,$6C,$60,$FE,$72,$57,$F8,$9B,$5F,$E0,$2D,$AF,$6C,$7A
	.dc.b	$4E,$67,$02,$1E,$A9,$F8,$2E,$23,$8B,$99,$F0,$BF,$CD,$1F,$E5,$E4
	.dc.b	$70,$43,$F9,$BB,$96,$99,$3F,$BA,$39,$69,$FC,$BB,$C5,$EE,$FF,$1C
	.dc.b	$6C,$FF,$5C,$6D,$FF,$6D,$6C,$FF,$74,$6C,$7F,$7A,$6D,$9A,$49,$F8
	.dc.b	$13,$CD,$A7,$32,$BC,$DF,$31,$0F,$A2,$5A,$F3,$36,$3F,$12,$D3,$A1
	.dc.b	$ED,$3A,$D7,$95,$D4,$77,$7A,$9B,$24,$F6,$72,$D2,$47,$AE,$32,$D8
	.dc.b	$CC,$78,$9F,$BE,$62,$3C,$0D,$9F,$C2,$CF,$82,$79,$11,$FC,$69,$93
	.dc.b	$DA,$7F,$8C,$6B,$FB,$A9,$1F,$EF,$AD,$7F,$BE,$B5,$E0,$68,$86,$8B
	.dc.b	$5F,$E1,$EF,$17,$FC,$3B,$3F,$E1,$E6,$CF,$F7,$86,$BF,$D9,$9B,$3C
	.dc.b	$63,$E7,$DA,$7F,$C5,$9F,$AF,$BC,$6A,$63,$C6,$B7,$FE,$07,$9D,$36
	.dc.b	$74,$D2,$1F,$E0,$58,$DB,$67,$9B,$99,$B8,$7F,$90,$E9,$58,$CB,$FC
	.dc.b	$AD,$1C,$A7,$55,$E0,$BE,$47,$82,$32,$F9,$EF,$93,$FE,$60,$A7,$8D
	.dc.b	$FE,$33,$6C,$7F,$8E,$9A,$FF,$8E,$9A,$FF,$9D,$9A,$F9,$23,$91,$1E
	.dc.b	$4D,$79,$35,$E5,$E3,$78,$48,$F8,$5A,$F8,$5A,$F0,$34,$43,$45,$3C
	.dc.b	$DE,$37,$B4,$8F,$B5,$AF,$B5,$AF,$03,$44,$34,$53,$E1,$78,$DF,$24
	.dc.b	$8F,$C9,$6B,$F2,$5A,$F0,$34,$43,$45,$3E,$77,$8D,$F5,$48,$FD,$56
	.dc.b	$BF,$55,$AF,$03,$44,$34,$53,$ED,$78,$BF,$81,$23,$F8,$16,$BF,$81
	.dc.b	$6B,$C0,$D1,$0D,$14,$FC,$37,$8D,$F8,$D2,$3F,$8D,$6B,$F8,$D6,$BC
	.dc.b	$0D,$10,$D1,$4F,$C9,$78,$DF,$99,$23,$F9,$96,$BF,$99,$6B,$C0,$D1
	.dc.b	$0D,$14,$FC,$F7,$8B,$FA,$52,$3F,$A5,$6B,$FA,$56,$BC,$0D,$10,$D1
	.dc.b	$4F,$D5,$78,$DF,$B1,$23,$FB,$16,$BF,$B1,$6B,$C0,$D1,$0D,$14,$FD
	.dc.b	$B7,$8D,$FB,$D2,$3F,$BD,$6B,$FB,$D6,$BC,$0D,$10,$D1,$4F,$E0,$5E
	.dc.b	$37,$F0,$92,$3F,$C2,$5A,$FF,$09,$6B,$C0,$D1,$0D,$14,$FE,$1D,$E3
	.dc.b	$7F,$15,$23,$FC,$55,$AF,$F1,$56,$BC,$0D,$10,$D1,$4F,$E3,$5E,$37
	.dc.b	$F2,$12,$3F,$C8,$5A,$FF,$21,$6B,$C0,$D1,$0D,$14,$FE,$4D,$E3,$7F
	.dc.b	$2D,$23,$FC,$B5,$AF,$F2,$D6,$BC,$0D,$10,$D1,$4F,$E6,$5E,$37,$F3
	.dc.b	$92,$3F,$CE,$5A,$FF,$39,$6B,$C0,$D1,$0D,$14,$FE,$7D,$E3,$7F,$45
	.dc.b	$23,$FD,$15,$AF,$F4,$56,$BC,$0D,$10,$D1,$4F,$E9,$5E,$37,$F5,$12
	.dc.b	$3F,$D4,$5A,$FF,$51,$6B,$C0,$D1,$0D,$14,$FE,$AD,$E3,$7F,$5D,$23
	.dc.b	$FD,$75,$AF,$F5,$D6,$BC,$0D,$10,$D1,$4F,$EC,$5E,$37,$F8,$20,$FF
	.dc.b	$04,$F8,$D6,$3F,$C0,$D1,$5F,$F0,$14,$5E,$AF,$EF,$D3,$7B,$3F,$BE
	.dc.b	$CD,$AF,$EE,$D3,$FD,$D4,$8F,$F7,$56,$BF,$DD,$5A,$F9,$0D,$10,$D1
	.dc.b	$4D,$1F,$E1,$5F,$1B,$C1,$4E,$B4,$61,$B3,$EC,$3F,$C4,$51,$92,$3F
	.dc.b	$E2,$5F,$1B,$FC,$56,$7F,$8B,$3C,$6B,$19,$21,$FE,$25,$72,$FF,$89
	.dc.b	$E2,$E5,$FE,$23,$9B,$87,$F8,$8A,$7F,$C3,$04,$AF,$F0,$C3,$5F,$F0
	.dc.b	$C3,$5F,$F0,$C3,$5E,$47,$02,$1C,$0F,$F1,$F7,$8B,$E0,$7F,$C5,$0D
	.dc.b	$9F,$8C,$DC,$4D,$24,$FF,$91,$FC,$6F,$F2,$69,$FE,$4E,$F1,$AC,$64
	.dc.b	$87,$F9,$1D,$CB,$FE,$49,$8B,$97,$F9,$16,$6E,$1F,$E4,$49,$FF,$1D
	.dc.b	$12,$BF,$C7,$4D,$7F,$C7,$4D,$7F,$C7,$4D,$79,$1C,$08,$70,$3F,$CC
	.dc.b	$1E,$37,$81,$FF,$25,$36,$7F,$10,$DC,$7F,$D1,$38,$7F,$96,$EF,$FC
	.dc.b	$DB,$39,$3F,$CD,$FE,$34,$C3,$FC,$D9,$E2,$D7,$FC,$D1,$16,$FF,$CC
	.dc.b	$93,$71,$FF,$31,$CD,$AF,$F2,$E4,$D9,$FC,$D4,$FF,$97,$49,$3F,$E5
	.dc.b	$D6,$BF,$E5,$D6,$BF,$E5,$D6,$BD,$8E,$04,$38,$01,$49,$C0,$FD,$83
	.dc.b	$FC,$F0,$F3,$9D,$BA,$8C,$C7,$F9,$D6,$BA,$3A,$3A,$46,$A7,$DB,$1E
	.dc.b	$1A,$F0,$3A,$F1,$BE,$15,$6F,$3F,$F6,$74,$F3,$BF,$25,$76,$BE,$31
	.dc.b	$E7,$F1,$B0,$F9,$5B,$5E,$2F,$1A,$CE,$E3,$C4,$ED,$A7,$A2,$77,$8F
	.dc.b	$FA,$32,$FB,$BF,$E9,$E9,$DD,$73,$7C,$D1,$96,$DF,$2B,$91,$F1,$8F
	.dc.b	$01,$6B,$93,$F2,$9A,$E8,$C5,$75,$6C,$F3,$74,$43,$D4,$66,$79,$B8
	.dc.b	$3C,$13,$E7,$BF,$6F,$8D,$E6,$9F,$5B,$63,$31,$F0,$27,$85,$7E,$0E
	.dc.b	$B4,$D9,$D5,$FC,$A9,$F9,$7E,$34,$9E,$5E,$35,$7E,$4C,$5C,$7E,$7C
	.dc.b	$DC,$3F,$21,$36,$BF,$1B,$36,$7F,$31,$3F,$0E,$7F,$22,$D8,$CC,$7C
	.dc.b	$3C,$53,$96,$C6,$63,$04,$70,$43,$E4,$B9,$69,$F1,$7C,$6F,$24,$9A
	.dc.b	$48,$F9,$53,$F8,$E8,$C4,$7E,$2B,$C6,$FC,$89,$F9,$2F,$1A,$C7,$D1
	.dc.b	$8A,$FE,$42,$2E,$5F,$96,$9B,$8F,$ED,$E6,$D7,$ED,$27,$E0,$11,$F8
	.dc.b	$8D,$7F,$0A,$D7,$F2,$2D,$7F,$81,$35,$E8,$72,$3F,$0D,$E3,$78,$2B
	.dc.b	$F8,$63,$72,$34,$93,$ED,$38,$9F,$8F,$F1,$BF,$66,$7E,$D3,$C6,$B1
	.dc.b	$92,$1F,$B2,$72,$FE,$D2,$2E,$5F,$BA,$9B,$87,$F1,$B9,$FC,$11,$2B
	.dc.b	$F0,$CD,$7F,$4E,$D7,$F6,$6D,$79,$1C,$08,$70,$3F,$51,$E3,$78,$1F
	.dc.b	$E4,$AD,$9F,$E6,$0D,$C4,$D2,$4F,$EC,$7C,$6F,$E2,$A7,$F1,$6F,$1A
	.dc.b	$C6,$48,$7F,$14,$72,$FF,$16,$8B,$97,$F1,$C9,$B8,$7F,$3D,$9F,$D2
	.dc.b	$12,$BF,$50,$D7,$F8,$4B,$5F,$E2,$AD,$79,$1C,$08,$70,$3F,$85,$78
	.dc.b	$DE,$07,$FA,$3B,$67,$FA,$93,$71,$34,$93,$FC,$4F,$C6,$FE,$6E,$7F
	.dc.b	$38,$F1,$AC,$64,$87,$F3,$67,$2F,$F3,$88,$B9,$7F,$3E,$9B,$87,$F5
	.dc.b	$F9,$FE,$0C,$4A,$FE,$14,$D7,$F9,$5B,$5F,$E6,$ED,$79,$1C,$08,$70
	.dc.b	$3F,$96,$78,$DE,$07,$FB,$4B,$67,$FB,$A3,$71,$34,$93,$FC,$D7,$C6
	.dc.b	$FE,$B2,$7F,$5A,$F1,$AC,$64,$87,$F5,$87,$2F,$F5,$A8,$B9,$7F,$60
	.dc.b	$9B,$87,$F8,$06,$7F,$94,$12,$BF,$96,$35,$FE,$9A,$D7,$FA,$CB,$5E
	.dc.b	$47,$02,$1C,$0F,$E9,$DE,$2F,$81,$FF,$05,$B6,$7F,$C2,$CD,$C4,$D2
	.dc.b	$4F,$F5,$7F,$1B,$FB,$D9,$FD,$F3,$C6,$B1,$92,$1F,$DE,$9C,$BF,$DF
	.dc.b	$22,$E5,$FE,$02,$9B,$87,$F8,$8E,$7F,$A5,$12,$BF,$A7,$35,$FE,$DE
	.dc.b	$D7,$FB,$DB,$5E,$47,$02,$1C,$0F,$EE,$1E,$2F,$81,$FF,$16,$B6,$7F
	.dc.b	$C7,$0D,$C4,$D2,$4F,$F7,$9F,$1B,$FC,$3A,$7F,$87,$BC,$6B,$19,$21
	.dc.b	$FE,$1C,$72,$FF,$87,$A2,$E5,$FE,$24,$9B,$87,$F9,$16,$7F,$B6,$12
	.dc.b	$BF,$B8,$35,$FF,$08,$B5,$FF,$0E,$B5,$E4,$70,$21,$C0,$FF,$09,$78
	.dc.b	$BE,$07,$FC,$9E,$D9,$FF,$2D,$37,$13,$49,$3F,$E1,$BF,$1B,$FC,$7E
	.dc.b	$7F,$90,$3C,$6B,$19,$21,$FE,$3E,$72,$FF,$90,$22,$E5,$FE,$46,$9B
	.dc.b	$87,$F9,$9E,$7F,$C1,$C4,$AF,$F0,$93,$5F,$F1,$9B,$5F,$F1,$FB,$5E
	.dc.b	$47,$02,$1C,$0F,$F1,$A7,$8B,$E0,$7F,$CE,$2D,$9F,$F3,$E3,$71,$FF
	.dc.b	$36,$E1,$FE,$5C,$BF,$F2,$FC,$E4,$FF,$30,$F8,$D3,$0F,$F2,$FF,$EC
	.dc.b	$3F,$E7,$BC,$13,$99,$FF,$30,$BF,$E6,$E9,$D3,$9C,$3F,$CD,$2E,$7C
	.dc.b	$6C,$65,$FE,$8B,$8F,$F1,$91,$2F,$FC,$6C,$D7,$FC,$AE,$D7,$FC,$C6
	.dc.b	$D7,$91,$C0,$87,$03,$D3,$E2,$F8,$1F,$F5,$06,$1F,$E5,$A7,$FD,$2F
	.dc.b	$3E,$99,$D5,$2B,$4F,$FA,$96,$75,$5F,$E6,$36,$BF,$E9,$79,$FF,$44
	.dc.b	$4F,$FA,$B1,$B1,$E0,$F1,$7E,$37,$8D,$23,$E6,$BF,$26,$31,$1C,$E2
	.dc.b	$E5,$FE,$A0,$A5,$62,$FB,$57,$F9,$7E,$4D,$25,$FF,$A0,$E3,$2D,$8C
	.dc.b	$BF,$E9,$38,$CB,$63,$2F,$FA,$56,$32,$D8,$E3,$7E,$5F,$17,$C9,$3F
	.dc.b	$39,$B7,$F2,$DB,$7F,$E7,$09,$F7,$1F,$8E,$8C,$9F,$8B,$F1,$A4,$7E
	.dc.b	$4B,$C6,$AF,$CF,$8B,$87,$C5,$9B,$87,$D1,$9B,$5F,$80,$9B,$3F,$76
	.dc.b	$7E,$41,$27,$E3,$35,$F8,$8D,$7E,$13,$5E,$C7,$02,$1C,$0F,$D4,$F8
	.dc.b	$BE,$0C,$4D,$24,$7E,$BA,$7F,$55,$18,$8F,$C6,$78,$DF,$AD,$3F,$5D
	.dc.b	$E3,$58,$FD,$AC,$57,$F2,$51,$72,$FC,$44,$DC,$7F,$29,$36,$BF,$47
	.dc.b	$3F,$89,$23,$F8,$76,$BF,$85,$6B,$F8,$36,$BF,$9F,$6B,$D0,$E4,$7F
	.dc.b	$0A,$F1,$7C,$15,$FA,$D6,$E4,$69,$27,$DA,$71,$3F,$89,$F8,$DF,$9F
	.dc.b	$23,$F9,$D6,$BF,$9B,$6B,$F9,$96,$BF,$B5,$6B,$FC,$11,$FD,$57,$8D
	.dc.b	$23,$F5,$0D,$7F,$4C,$D7,$F4,$8D,$7F,$80,$35,$FE,$1E,$DB,$FE,$19
	.dc.b	$E2,$FE,$DC,$93,$FB,$56,$BF,$B3,$6B,$FB,$16,$BF,$C3,$5A,$FF,$1C
	.dc.b	$7F,$81,$78,$D2,$3F,$80,$35,$FD,$F3,$5F,$DE,$35,$FE,$30,$D7,$F9
	.dc.b	$3B,$67,$F9,$27,$8D,$FC,$3C,$93,$FC,$35,$AF,$F0,$B6,$BF,$C2,$5A
	.dc.b	$FF,$25,$6B,$FC,$D1,$FE,$35,$E2,$C8,$FE,$30,$D7,$F8,$B3,$5F,$E2
	.dc.b	$8D,$7F,$98,$35,$FE,$7E,$D9,$FE,$79,$E3,$7F,$57,$3F,$AC,$78,$D6
	.dc.b	$BF,$AF,$C5,$7F,$9E,$45,$E0,$FE,$65,$36,$BF,$9F,$4D,$AF,$E9,$B3
	.dc.b	$FC,$CC,$8F,$F3,$16,$BF,$CB,$DA,$FF,$2D,$6B,$FD,$25,$AF,$43,$91
	.dc.b	$FD,$D3,$C5,$F0,$57,$F5,$76,$EE,$69,$27,$DA,$71,$3F,$BE,$F8,$DF
	.dc.b	$DE,$4F,$EF,$5E,$35,$8C,$90,$FE,$BC,$E5,$FE,$C1,$17,$2F,$EA,$D3
	.dc.b	$70,$FE,$C5,$3F,$D5,$49,$5F,$D4,$DA,$FF,$51,$6B,$FD,$3D,$AF,$23
	.dc.b	$81,$0E,$07,$F8,$5B,$C5,$F0,$3F,$DF,$9B,$3F,$DE,$DB,$89,$A4,$9F
	.dc.b	$F0,$FF,$8D,$FE,$1B,$3F,$C3,$9E,$35,$8C,$90,$FF,$00,$39,$7F,$C0
	.dc.b	$51,$72,$FE,$F1,$37,$0F,$F0,$24,$FF,$77,$25,$7F,$75,$6B,$FD,$CD
	.dc.b	$AF,$F7,$16,$BC,$8E,$04,$38,$1F,$E3,$8F,$17,$C0,$FF,$88,$1B,$3F
	.dc.b	$E1,$D6,$E2,$69,$27,$FC,$83,$E3,$7F,$8F,$4F,$F1,$F7,$8D,$63,$24
	.dc.b	$3F,$C4,$4E,$5F,$F1,$24,$5C,$BF,$C3,$53,$70,$FF,$13,$4F,$F8,$64
	.dc.b	$95,$FE,$17,$6B,$FE,$15,$6B,$FE,$13,$6B,$C8,$E0,$43,$81,$FE,$5A
	.dc.b	$F1,$7C,$0F,$F9,$09,$B3,$FE,$3F,$6E,$26,$92,$7F,$CC,$7E,$37,$F9
	.dc.b	$7C,$FF,$30,$78,$D6,$32,$43,$FC,$88,$E5,$FF,$23,$45,$CB,$FC,$79
	.dc.b	$37,$0F,$F2,$44,$FF,$8E,$C9,$5F,$E3,$96,$BF,$E3,$76,$BF,$E3,$56
	.dc.b	$BC,$8E,$04,$38,$1F,$E7,$CF,$0F,$C0,$FF,$99,$1B,$3F,$E6,$16,$E3
	.dc.b	$FE,$80,$9F,$66,$1F,$E5,$39,$C8,$89,$0A,$C0,$C9,$E2,$EE,$95,$B7
	.dc.b	$84,$E9,$E7,$5C,$A3,$FC,$E6,$66,$38,$C6,$6B,$A5,$73,$8F,$04,$61
	.dc.b	$E5,$E2,$DA,$EF,$36,$77,$1E,$03,$25,$7D,$53,$BB,$F1,$5F,$77,$C9
	.dc.b	$5E,$DA,$E6,$FB,$CE,$65,$72,$EE,$38,$BE,$B3,$81,$C0,$D5,$E1,$C1
	.dc.b	$B3,$A4,$72,$72,$78,$2B,$CD,$E2,$7A,$6B,$D4,$D9,$F7,$57,$FA,$62
	.dc.b	$FD,$37,$C2,$73,$7C,$1A,$70,$BF,$89,$E2,$C8,$E7,$E2,$D7,$EB,$9F
	.dc.b	$7A,$31,$7F,$26,$6E,$1F,$7E,$6D,$7E,$12,$6C,$FE,$2E,$7E,$13,$67
	.dc.b	$E3,$B5,$F8,$71,$AC,$7F,$D5,$31,$A6,$BA,$31,$1B,$23,$82,$1B,$3D
	.dc.b	$73,$F9,$DF,$13,$CB,$46,$92,$7E,$6D,$7E,$26,$31,$1C,$A5,$E2,$FE
	.dc.b	$34,$FC,$77,$8B,$67,$EA,$C5,$78,$45,$CB,$F2,$53,$71,$FD,$44,$DA
	.dc.b	$FD,$6C,$FC,$32,$3F,$0D,$AF,$E2,$9A,$FE,$55,$AF,$EC,$DA,$F4,$39
	.dc.b	$1F,$8A,$F1,$3C,$15,$FC,$05,$B8,$1A,$49,$F6,$9C,$4F,$C9,$F8,$BF
	.dc.b	$AF,$3F,$61,$E2,$D8,$C9,$0F,$94,$E5,$F9,$91,$72,$FD,$B4,$DC,$3F
	.dc.b	$86,$4F,$E1,$49,$5F,$85,$6B,$FA,$C6,$BF,$B7,$6B,$C8,$E0,$43,$81
	.dc.b	$FA,$CF,$17,$C0,$FF,$1E,$6C,$FF,$25,$6E,$26,$92,$7F,$6B,$E2,$FF
	.dc.b	$11,$3F,$89,$78,$B6,$32,$43,$F1,$8E,$5F,$C7,$45,$CB,$F8,$C4,$DC
	.dc.b	$3F,$97,$4F,$E9,$C9,$5F,$A7,$6B,$FC,$39,$AF,$F1,$96,$BC,$8E,$04
	.dc.b	$38,$1F,$C3,$BC,$5F,$03,$FD,$01,$B3,$FD,$1D,$B8,$9A,$49,$FE,$2F
	.dc.b	$E2,$FF,$33,$3F,$9A,$78,$B6,$32,$43,$F5,$CE,$5F,$D8,$45,$CB,$F9
	.dc.b	$D4,$DC,$3F,$A8,$4F,$F0,$92,$57,$F0,$96,$BF,$CC,$1A,$FF,$3B,$6B
	.dc.b	$C8,$E0,$43,$81,$FC,$C3,$C4,$F0,$3F,$D8,$9B,$3F,$DA,$5B,$89,$A4
	.dc.b	$9F,$E7,$3E,$2F,$F5,$53,$FA,$B7,$8B,$63,$24,$3F,$88,$39,$7F,$89
	.dc.b	$45,$CB,$FA,$E4,$DC,$3F,$B9,$4F,$F2,$B2,$57,$F2,$B6,$BF,$D4,$9A
	.dc.b	$FF,$5D,$6B,$C8,$E0,$43,$81,$FD,$4B,$C4,$F0,$3F,$E0,$46,$CF,$F8
	.dc.b	$2D,$B8,$9A,$49,$FE,$B7,$E2,$FF,$77,$3F,$BC,$78,$B6,$32,$43,$F9
	.dc.b	$93,$97,$F9,$A4,$5C,$BF,$BF,$4D,$C3,$FC,$29,$3F,$D3,$49,$5F,$D3
	.dc.b	$5A,$FF,$74,$6B,$FD,$FD,$AF,$23,$81,$0E,$07,$F7,$4F,$13,$C0,$FF
	.dc.b	$89,$9B,$3F,$E2,$D6,$E2,$69,$27,$FB,$EF,$8B,$FE,$19,$3F,$C3,$5E
	.dc.b	$2D,$8C,$90,$FE,$A8,$E5,$FE,$AD,$17,$2F,$F1,$04,$DC,$3F,$C6,$D3
	.dc.b	$FD,$BC,$95,$FD,$BD,$AF,$F8,$59,$AF,$F8,$85,$AF,$23,$81,$0E,$07
	.dc.b	$F8,$5B,$0F,$03,$FE,$48,$6C,$FF,$93,$DA,$7F,$94,$63,$D9,$7F,$E3
	.dc.b	$78,$C9,$FE,$3B,$F1,$64,$FF,$88,$BC,$5A,$FF,$75,$8B,$5F,$E3,$F9
	.dc.b	$B7,$FE,$55,$9B,$5F,$E5,$A9,$B3,$FE,$61,$9F,$F0,$91,$27,$FC,$24
	.dc.b	$D7,$FC,$72,$D7,$FC,$88,$D7,$B1,$C0,$87,$03,$FC,$73,$E2,$78,$30
	.dc.b	$34,$91,$FE,$3E,$9F,$F2,$D4,$62,$3F,$C2,$5E,$2F,$F9,$74,$FF,$2F
	.dc.b	$78,$96,$3F,$C8,$B1,$5F,$F0,$C4,$5C,$BF,$CC,$53,$71,$FF,$3C,$4D
	.dc.b	$AF,$F3,$EC,$FF,$8C,$C8,$FF,$8C,$DA,$FF,$96,$9A,$FF,$99,$5A,$FF
	.dc.b	$A0,$DA,$F4,$39,$1F,$E7,$7F,$D8,$BF,$CE,$B5,$B8,$C7,$89,$E3,$84
	.dc.b	$EA,$79,$1F,$E3,$C7,$11,$FE,$99,$AF,$F4,$34,$63,$34,$62,$F3,$E2
	.dc.b	$CC,$3C,$5E,$2D,$7A,$45,$CF,$FD,$2D,$8F,$F9,$BC,$46,$72,$FF,$54
	.dc.b	$C6,$2B,$99,$ED,$9E,$8F,$FA,$3E,$34,$FF,$A3,$5A,$E6,$3F,$CC,$F1
	.dc.b	$88,$EB,$18,$23,$FE,$6A,$8D,$3F,$EB,$B6,$CF,$C8,$9D,$CF,$77,$E2
	.dc.b	$F8,$9E,$98,$E0,$D9,$F9,$ED,$AF,$F3,$8D,$FC,$2B,$E3,$7E,$D3,$ED
	.dc.b	$57,$D0,$F1,$64,$7D,$DF,$16,$BF,$5E,$2D,$FC,$C9,$B8,$FC,$A9,$B5
	.dc.b	$EB,$9B,$3F,$82,$9F,$8A,$49,$F8,$4D,$7E,$63,$5F,$A0,$D7,$B1,$C0
	.dc.b	$87,$03,$F2,$9E,$27,$83,$2F,$F3,$FC,$7E,$5E,$7F,$1B,$39,$24,$64
	.dc.b	$E2,$7E,$77,$C5,$FB,$64,$7E,$C3,$5F,$C1,$35,$FC,$33,$5F,$C6,$B5
	.dc.b	$FB,$CF,$E1,$BC,$59,$1F,$83,$6B,$F8,$D6,$BF,$91,$6B,$F9,$86,$BF
	.dc.b	$89,$6E,$1F,$B0,$F1,$3F,$26,$49,$FC,$83,$5F,$CD,$35,$FC,$F3,$5F
	.dc.b	$D2,$B5,$FC,$B3,$F9,$EF,$16,$47,$E6,$DA,$FE,$95,$AF,$EA,$5A,$FE
	.dc.b	$C1,$AF,$E8,$5B,$3F,$C2,$3C,$5F,$E0,$67,$F0,$4F,$12,$D7,$F1,$18
	.dc.b	$AF,$F0,$B8,$BB,$7F,$00,$9B,$5F,$BE,$9B,$5F,$AF,$9B,$3F,$C5,$A6
	.dc.b	$EB,$FE,$85,$AE,$C7,$F1,$F8,$CD,$69,$DF,$8B,$FC,$00,$93,$FB,$D6
	.dc.b	$BF,$C1,$DA,$FF,$0B,$6B,$FC,$49,$AF,$F0,$37,$F8,$5F,$8B,$23,$F8
	.dc.b	$43,$5F,$E2,$4D,$7F,$8B,$35,$FE,$3A,$D7,$F8,$73,$70,$FE,$6B,$E2
	.dc.b	$7F,$30,$3F,$98,$F8,$B6,$BF,$9F,$45,$7F,$9C,$45,$CF,$F9,$6C,$DA
	.dc.b	$FE,$57,$36,$BF,$8E,$4F,$F2,$42,$3F,$C8,$5A,$FF,$2B,$6B,$FC,$BD
	.dc.b	$AF,$F3,$66,$BD,$0E,$47,$F5,$3F,$17,$C1,$5F,$D4,$1B,$A7,$F9,$EE
	.dc.b	$7D,$66,$CA,$7F,$53,$8D,$9F,$D4,$BC,$49,$1F,$D7,$FC,$5A,$FF,$59
	.dc.b	$8B,$97,$F4,$E9,$B8,$7F,$4C,$9B,$5F,$CF,$66,$CF,$F6,$49,$FE,$8E
	.dc.b	$49,$FE,$88,$D7,$FA,$73,$5F,$EA,$4D,$7B,$1C,$08,$70,$3F,$BB,$78
	.dc.b	$BE,$0C,$7F,$CF,$51,$FD,$E6,$7F,$B7,$CE,$49,$19,$38,$9F,$DF,$FC
	.dc.b	$4F,$EE,$A7,$F7,$6F,$16,$C6,$48,$7F,$7D,$72,$FF,$7F,$8B,$97,$F7
	.dc.b	$49,$B8,$FF,$72,$9F,$ED,$24,$AF,$EC,$CD,$7F,$B8,$35,$FE,$E8,$D7
	.dc.b	$91,$C0,$87,$03,$FC,$31,$E2,$F8,$1F,$EE,$0D,$9F,$F0,$BB,$72,$FF
	.dc.b	$3F,$4F,$B3,$0F,$F0,$5C,$E4,$FF,$0B,$F8,$92,$7F,$C4,$9E,$2D,$7F
	.dc.b	$C3,$D1,$6B,$FC,$27,$37,$0F,$F0,$8C,$DA,$FF,$01,$4D,$9F,$F1,$4C
	.dc.b	$FF,$83,$09,$3F,$E0,$96,$BF,$E1,$36,$BF,$E1,$76,$BD,$8E,$04,$38
	.dc.b	$1F,$E3,$BF,$07,$C1,$8F,$F9,$EE,$3D,$26,$4F,$F2,$04,$FF,$8D,$A7
	.dc.b	$64,$8F,$F2,$27,$89,$FE,$3A,$3F,$C7,$7E,$2D,$8C,$90,$FF,$21,$39
	.dc.b	$7F,$C8,$91,$72,$FF,$1C,$CD,$C7,$FC,$6F,$3F,$E2,$E2,$57,$F8,$AD
	.dc.b	$AF,$F8,$D5,$AF,$F8,$E5,$AF,$23,$81,$0E,$07,$F9,$77,$C5,$F0,$3F
	.dc.b	$E3,$56,$CF,$F9,$71,$B9,$7F,$A0,$A7,$D9,$87,$F9,$42,$72,$7F,$97
	.dc.b	$3C,$49,$3F,$E6,$9F,$16,$BF,$E6,$38,$B5,$FE,$58,$9B,$87,$F9,$5A
	.dc.b	$6D,$7F,$91,$E6,$CF,$F9,$BA,$7F,$CA,$24,$9F,$F2,$73,$5F,$F2,$C3
	.dc.b	$5F,$F2,$E3,$5E,$C7,$02,$1C,$00,$A4,$E0,$7E,$C7,$FE,$7C,$C7,$FC
	.dc.b	$F3,$5C,$2B,$6F,$09,$D5,$F4,$8E,$97,$CE,$73,$1C,$63,$1E,$37,$C0
	.dc.b	$9C,$CF,$03,$A4,$73,$BE,$F5,$E4,$F1,$70,$F5,$6D,$78,$FC,$6B,$3B
	.dc.b	$8F,$01,$E7,$9C,$E1,$FE,$82,$BE,$AA,$F3,$32,$B8,$DF,$37,$CC,$FC
	.dc.b	$69,$D4,$62,$BF,$D0,$24,$AF,$89,$1F,$16,$B9,$36,$74,$E8,$F3,$16
	.dc.b	$36,$FB,$CC,$47,$07,$AC,$69,$E2,$78,$FC,$6F,$2E,$1E,$E6,$CF,$D5
	.dc.b	$6C,$FD,$B6,$C7,$E0,$9B,$5F,$E9,$28,$FB,$9E,$2F,$CA,$3E,$5F,$8D
	.dc.b	$6F,$E3,$D2,$71,$1F,$16,$6E,$1F,$42,$6D,$FE,$1E,$6D,$7E,$36,$7E
	.dc.b	$23,$4F,$82,$E2,$7F,$2E,$DB,$F0,$4E,$88,$7C,$67,$2D,$8E,$04,$76
	.dc.b	$7D,$57,$F0,$FE,$37,$92,$BF,$4A,$DA,$FF,$4A,$D7,$E0,$6B,$6F,$C2
	.dc.b	$6B,$93,$E2,$CF,$E4,$BC,$69,$1F,$23,$C6,$AF,$E2,$E2,$E5,$F9,$69
	.dc.b	$B8,$FE,$B6,$6D,$7E,$CE,$6C,$FE,$EA,$7E,$21,$27,$F1,$2D,$7E,$23
	.dc.b	$5F,$CA,$35,$EC,$70,$21,$C0,$FB,$FE,$37,$83,$23,$49,$1F,$84,$9F
	.dc.b	$D8,$C6,$23,$F5,$1E,$37,$ED,$0F,$DA,$F8,$D6,$3F,$17,$15,$FD,$74
	.dc.b	$5C,$BF,$73,$37,$1F,$E2,$13,$6B,$F8,$B4,$FE,$14,$8F,$EA,$9A,$FE
	.dc.b	$15,$AF,$ED,$5A,$FF,$1B,$6B,$D0,$E4,$7E,$83,$C6,$F0,$57,$F2,$A6
	.dc.b	$E4,$69,$27,$DA,$71,$3F,$53,$E3,$7F,$16,$3F,$8B,$F8,$D6,$32,$43
	.dc.b	$F8,$83,$97,$F8,$94,$5C,$BF,$8E,$CD,$C3,$F9,$A4,$FE,$9C,$95,$FC
	.dc.b	$31,$AF,$E9,$DA,$FF,$17,$6B,$C8,$E0,$43,$81,$FC,$0B,$C5,$F0,$3F
	.dc.b	$D1,$DB,$3F,$D3,$9B,$89,$A4,$9F,$E1,$7E,$37,$F3,$83,$F9,$CF,$8D
	.dc.b	$63,$24,$3F,$99,$39,$7F,$9A,$45,$CB,$F9,$FC,$DC,$3F,$AB,$4F,$F0
	.dc.b	$92,$57,$F2,$E6,$BF,$C2,$5A,$FF,$39,$6B,$C8,$E0,$43,$81,$FC,$93
	.dc.b	$C5,$F0,$3F,$DA,$5B,$3F,$DC,$1B,$89,$A4,$9F,$E5,$BE,$37,$F5,$A3
	.dc.b	$FA,$DF,$8D,$63,$24,$3F,$AA,$39,$7F,$AB,$45,$CB,$FB,$0C,$DC,$3F
	.dc.b	$BC,$4F,$F2,$B2,$57,$F5,$06,$BF,$CA,$DA,$FF,$5B,$6B,$C8,$E0,$43
	.dc.b	$81,$FD,$1B,$C5,$F0,$3F,$E0,$B6,$CF,$F8,$49,$B8,$9A,$49,$FE,$9F
	.dc.b	$E3,$7F,$7C,$3F,$BE,$F8,$D6,$32,$43,$FB,$B3,$97,$FB,$C4,$5C,$BF
	.dc.b	$C0,$73,$70,$FF,$0D,$4F,$F4,$D2,$57,$F7,$26,$BF,$D3,$5A,$FF,$7D
	.dc.b	$6B,$C8,$E0,$43,$81,$FD,$A3,$C5,$F0,$3F,$E2,$D6,$CF,$F8,$D1,$B8
	.dc.b	$9A,$49,$FE,$E3,$E3,$7F,$87,$8F,$F0,$FF,$8D,$63,$24,$3F,$C3,$0E
	.dc.b	$5F,$F0,$D4,$5C,$BF,$C4,$B3,$70,$FF,$1E,$4F,$F6,$F2,$57,$F8,$51
	.dc.b	$AF,$F6,$F6,$BF,$E1,$F6,$BC,$8E,$04,$38,$1F,$E0,$AF,$17,$C0,$FF
	.dc.b	$93,$DB,$3F,$E5,$66,$E2,$69,$27,$FC,$27,$E3,$7F,$90,$0F,$F2,$0F
	.dc.b	$8D,$63,$24,$3F,$C7,$4E,$5F,$F1,$E4,$5C,$BF,$C8,$F3,$70,$FF,$2F
	.dc.b	$4F,$F8,$44,$95,$FE,$36,$6B,$FE,$11,$6B,$FE,$41,$6B,$C8,$E0,$43
	.dc.b	$81,$FE,$2C,$F1,$3C,$0F,$F9,$C5,$B3,$FE,$78,$6E,$3F,$E7,$89,$F4
	.dc.b	$9C,$4F,$76,$1F,$E6,$09,$FF,$31,$F8,$D2,$7F,$CB,$FF,$B2,$3C,$6C
	.dc.b	$47,$F9,$76,$6D,$FF,$9D,$2B,$53,$FE,$40,$9C,$E3,$FE,$81,$7F,$CE
	.dc.b	$73,$A9,$E4,$6C,$FF,$2D,$C7,$F8,$E4,$8E,$08,$7F,$8E,$5C,$B4,$C9
	.dc.b	$FE,$6B,$72,$D3,$D7,$E2,$F9,$1F,$F4,$FC,$FF,$A0,$6B,$CD,$1A,$3C
	.dc.b	$87,$F9,$FA,$34,$FF,$A7,$EB,$D7,$5C,$E9,$5A,$6F,$8D,$F0,$0F,$83
	.dc.b	$E3,$5C,$3E,$6C,$57,$B4,$5C,$7E,$64,$DA,$FF,$5A,$5F,$F9,$C6,$FD
	.dc.b	$33,$C5,$FF,$46,$47,$FA,$65,$FF,$44,$35,$FF,$42,$B5,$DC,$7F,$9A
	.dc.b	$63,$11,$DA,$30,$47,$9B,$5E,$6D,$AF,$C1,$78,$BE,$79,$F9,$AD,$BF
	.dc.b	$8A,$DB,$FF,$38,$D7,$BA,$BF,$1F,$39,$3F,$0F,$E3,$79,$49,$1E,$C6
	.dc.b	$BE,$C6,$BF,$09,$AF,$C8,$6B,$F8,$07,$E0,$78,$D2,$3E,$3B,$5F,$8E
	.dc.b	$D7,$E6,$B5,$FA,$6D,$7F,$16,$DC,$3F,$07,$E2,$FC,$D2,$4F,$D4,$6B
	.dc.b	$F5,$1A,$FD,$A6,$BF,$80,$6B,$F9,$87,$EC,$78,$D2,$3E,$FB,$5F,$BE
	.dc.b	$D7,$F0,$AD,$7F,$16,$D7,$F4,$6D,$9F,$CD,$F8,$DF,$85,$24,$FE,$31
	.dc.b	$AF,$E3,$1A,$FE,$49,$AF,$E6,$1A,$FE,$C1,$FC,$87,$8B,$23,$F2,$ED
	.dc.b	$7F,$2E,$D7,$F3,$AD,$7F,$46,$D7,$F7,$6D,$9F,$D9,$F8,$DF,$C4,$0F
	.dc.b	$E2,$3E,$35,$AF,$E3,$B1,$5F,$DF,$45,$D7,$F8,$E4,$DA,$FD,$8C,$DA
	.dc.b	$FE,$13,$3F,$A9,$23,$FB,$06,$BF,$B0,$6B,$FB,$66,$BF,$BC,$6B,$D0
	.dc.b	$E4,$7F,$32,$F1,$7C,$15,$FC,$39,$BB,$9A,$49,$F6,$9C,$4F,$E7,$7E
	.dc.b	$37,$F3,$23,$F9,$9F,$8D,$63,$24,$3F,$90,$39,$7F,$91,$45,$CB,$FA
	.dc.b	$04,$DC,$3F,$8A,$CF,$F0,$B2,$57,$F1,$26,$BF,$C4,$9A,$FF,$18,$6B
	.dc.b	$C8,$E0,$43,$81,$FD,$53,$C5,$F0,$3F,$CE,$9B,$3F,$CC,$9B,$89,$A4
	.dc.b	$9F,$EB,$BE,$37,$F5,$43,$FA,$AF,$8D,$63,$24,$3F,$A1,$39,$7F,$A2
	.dc.b	$45,$CB,$FB,$14,$DC,$3F,$9B,$CF,$F2,$D2,$57,$F3,$46,$BF,$CD,$1A
	.dc.b	$FF,$3A,$6B,$C8,$E0,$43,$81,$FD,$DB,$C5,$F0,$3F,$D7,$1B,$3F,$D5
	.dc.b	$1B,$89,$A4,$9F,$EF,$FE,$37,$F7,$63,$FB,$BF,$8D,$63,$24,$3F,$B2
	.dc.b	$39,$7F,$B3,$45,$CB,$FC,$09,$37,$0F,$EB,$33,$FD,$3C,$95,$FD,$59
	.dc.b	$AF,$F5,$66,$BF,$D7,$1A,$F2,$38,$10,$E0,$7F,$86,$3C,$5F,$03,$FD
	.dc.b	$F9,$B3,$FD,$D9,$B8,$9A,$49,$FF,$10,$F8,$DF,$E1,$83,$FC,$33,$E3
	.dc.b	$58,$C9,$0F,$F0,$33,$97,$FC,$11,$17,$2F,$F1,$34,$DC,$3F,$BD,$CF
	.dc.b	$F7,$12,$57,$F7,$86,$BF,$DE,$1A,$FF,$7E,$6B,$C8,$E0,$43,$81,$FE
	.dc.b	$3A,$F1,$7C,$0F,$F8,$81,$B3,$FE,$18,$6E,$26,$92,$7F,$C8,$7E,$37
	.dc.b	$F8,$E8,$FF,$1D,$F8,$D6,$32,$43,$FC,$50,$E5,$FF,$15,$45,$CB,$FC
	.dc.b	$91,$37,$0F,$F0,$EC,$FF,$84,$C9,$5F,$E1,$A6,$BF,$E1,$A6,$BF,$E2
	.dc.b	$06,$BC,$8E,$04,$38,$1F,$E5,$CF,$17,$C0,$FF,$90,$9B,$3F,$E3,$A6
	.dc.b	$E2,$69,$27,$FC,$CB,$E3,$7F,$97,$0F,$F2,$EF,$8B,$63,$24,$3F,$C9
	.dc.b	$4E,$5F,$F2,$64,$5C,$BF,$CD,$53,$70,$FF,$1F,$CF,$F8,$D4,$95,$FE
	.dc.b	$3C,$6B,$FE,$3C,$6B,$FE,$42,$6B,$C8,$E0,$43,$81,$FE,$7E,$F1,$3C
	.dc.b	$0F,$F9,$91,$B3,$FE,$5C,$6E,$26,$92,$43,$05,$60,$7C,$0F,$FA,$06
	.dc.b	$33,$3A,$8C,$4E,$DF,$F3,$51,$88,$D7,$8D,$D9,$ED,$3B,$8C,$57,$48
	.dc.b	$EF,$E3,$61,$D5,$79,$A7,$57,$D6,$7C,$B7,$BA,$F2,$5D,$9C,$C7,$17
	.dc.b	$9B,$E3,$3C,$A5,$29,$97,$8B,$6B,$FD,$44,$FA,$E3,$2D,$8C,$BD,$8C
	.dc.b	$57,$57,$A9,$83,$C1,$3C,$9F,$2D,$FB,$BC,$5F,$55,$FB,$CF,$F4,$24
	.dc.b	$7C,$2B,$CC,$E9,$B5,$EF,$3D,$73,$CD,$C5,$FB,$23,$11,$F5,$0F,$B9
	.dc.b	$E3,$78,$0F,$0F,$8D,$63,$8E,$3F,$E9,$BB,$FC,$2D,$6A,$72,$ED,$E9
	.dc.b	$1A,$78,$D6,$6F,$ED,$D7,$C5,$AD,$57,$CF,$8C,$59,$A4,$8F,$93,$19
	.dc.b	$6C,$65,$E4,$60,$86,$0F,$2C,$3C,$6F,$25,$7E,$6D,$B3,$F9,$F6,$C7
	.dc.b	$E9,$1B,$3C,$67,$E0,$9F,$1A,$3E,$CF,$8D,$23,$E4,$F8,$D5,$F0,$C5
	.dc.b	$C7,$F3,$B3,$70,$FD,$64,$DA,$FD,$84,$D9,$FD,$AC,$FC,$62,$4F,$C4
	.dc.b	$6B,$F9,$96,$BF,$97,$6B,$D8,$E0,$43,$81,$F8,$0F,$1B,$C1,$89,$A4
	.dc.b	$8F,$C2,$CF,$E6,$23,$11,$F3,$BC,$6F,$CD,$9F,$9C,$F1,$AC,$7E,$32
	.dc.b	$2B,$F5,$22,$E5,$FC,$02,$6E,$3F,$C3,$E6,$D7,$F1,$39,$FC,$39,$1F
	.dc.b	$C2,$B5,$FD,$D3,$5F,$DC,$35,$FE,$34,$D7,$A1,$C8,$FD,$0F,$8B,$E0
	.dc.b	$AF,$E4,$AD,$C8,$D2,$4F,$B4,$E2,$7E,$AB,$C6,$FD,$E9,$FB,$EF,$1A
	.dc.b	$C6,$48,$7E,$59,$CB,$F9,$88,$B9,$7F,$24,$9B,$87,$F3,$39,$FD,$49
	.dc.b	$2B,$F4,$ED,$7F,$8E,$35,$FE,$34,$D7,$91,$C0,$87,$03,$F8,$1F,$8B
	.dc.b	$E0,$7F,$A2,$36,$7F,$A4,$B7,$13,$49,$3F,$C3,$3C,$6F,$E3,$E7,$F2
	.dc.b	$0F,$1A,$C6,$48,$7E,$E1,$CB,$FB,$A8,$B9,$7F,$46,$9B,$87,$F5,$59
	.dc.b	$FE,$16,$4A,$FE,$12,$D7,$F9,$F3,$5F,$E7,$8D,$79,$1C,$08,$70,$3F
	.dc.b	$92,$F8,$BE,$07,$FB,$33,$67,$FB,$5B,$71,$34,$93,$FC,$BB,$C6,$FE
	.dc.b	$82,$7F,$42,$F1,$AC,$64,$87,$F1,$A7,$2F,$F1,$C8,$B9,$7F,$68,$9B
	.dc.b	$87,$F7,$79,$FE,$5A,$4A,$FE,$56,$D7,$FB,$03,$5F,$EB,$CD,$79,$1C
	.dc.b	$08,$70,$3F,$A3,$F8,$BE,$07,$FC,$10,$D9,$FF,$06,$B7,$13,$49,$3F
	.dc.b	$D4,$3C,$6F,$EC,$67,$F6,$4F,$1A,$C6,$48,$7F,$3C,$72,$FF,$3E,$8B
	.dc.b	$97,$F8,$2A,$6E,$1F,$E1,$99,$FE,$9E,$4A,$FE,$9A,$D7,$FC,$04,$D7
	.dc.b	$FC,$00,$D7,$91,$C0,$87,$03,$FB,$4F,$8B,$E0,$7F,$C5,$4D,$9F,$F1
	.dc.b	$7B,$71,$34,$93,$FD,$CB,$C6,$FF,$02,$9F,$E0,$6F,$1A,$C6,$48,$7F
	.dc.b	$5E,$72,$FF,$60,$8B,$97,$F8,$B2,$6E,$1F,$E3,$B9,$FE,$E2,$4A,$FE
	.dc.b	$DE,$D7,$FC,$48,$D7,$FC,$44,$D7,$91,$C0,$87,$03,$FC,$17,$E2,$F8
	.dc.b	$1F,$F2,$63,$67,$FC,$A2,$DC,$4D,$24,$FF,$85,$3C,$6F,$F1,$39,$FE
	.dc.b	$28,$F1,$AC,$64,$87,$F8,$01,$CB,$FE,$02,$8B,$97,$F9,$3A,$6E,$1F
	.dc.b	$E5,$D9,$FF,$09,$92,$BF,$C2,$2D,$7F,$C8,$CD,$7F,$C8,$8D,$79,$1C
	.dc.b	$08,$70,$3F,$C5,$BE,$2F,$81,$FF,$37,$36,$7F,$CE,$6D,$C4,$D2,$4F
	.dc.b	$F8,$DB,$C6,$FF,$24,$9F,$E4,$AF,$16,$C6,$48,$7F,$88,$9C,$BF,$E2
	.dc.b	$48,$B9,$7F,$9C,$26,$E1,$FE,$7F,$9F,$F1,$A9,$2B,$FC,$66,$D7,$FC
	.dc.b	$D0,$D7,$FC,$CC,$D7,$91,$C0,$87,$03,$FC,$AF,$FB,$2F,$F9,$86,$3F
	.dc.b	$CF,$8D,$8C,$F8,$9E,$47,$FD,$33,$19,$7F,$CA,$2D,$C3,$FD,$1C,$FF
	.dc.b	$A2,$47,$8D,$FE,$51,$6C,$FF,$9D,$1A,$FF,$A0,$DA,$FF,$A0,$5A,$FF
	.dc.b	$A9,$DA,$F8,$A3,$A1,$1E,$8D,$7A,$3E,$1F,$1A,$47,$85,$AF,$85,$AF
	.dc.b	$85,$AF,$85,$AF,$C2,$6E,$2A,$FF,$33,$4F,$AF,$C5,$F8,$04,$7E,$03
	.dc.b	$5D,$18,$21,$82,$1F,$0D,$F9,$5E,$37,$D6,$3E,$BF,$8D,$67,$ED,$C5
	.dc.b	$7E,$AC,$5C,$BE,$AC,$DC,$3E,$64,$DA,$FB,$B3,$D9,$FA,$CD,$7E,$B3
	.dc.b	$5F,$AC,$D7,$EB,$35,$E6,$72,$21,$F8,$8F,$17,$C1,$5F,$25,$B8,$9A
	.dc.b	$49,$F6,$9F,$98,$7F,$2F,$E3,$48,$FC,$DF,$8D,$5F,$CA,$C5,$CB,$F2
	.dc.b	$B3,$70,$FC,$64,$DA,$FC,$EC,$D9,$FB,$93,$70,$FF,$3A,$57,$EA,$23
	.dc.b	$07,$EA,$2A,$C6,$4F,$19,$4F,$D3,$F8,$BF,$AE,$F1,$BF,$5E,$D8,$FD
	.dc.b	$9B,$5D,$9F,$B6,$8D,$B4,$D9,$4D,$9F,$B4,$76,$D3,$65,$37,$8F,$EE
	.dc.b	$23,$6D,$8D,$B5,$FD,$FF,$8D,$FC,$18,$FE,$0F,$E2,$D8,$FE,$17,$15
	.dc.b	$FE,$0B,$16,$7F,$82,$CD,$CF,$F7,$93,$6B,$F8,$6C,$FF,$07,$23,$FC
	.dc.b	$1D,$AF,$F0,$76,$BF,$C1,$DA,$FF,$07,$6B,$D0,$E4,$7F,$20,$F1,$BC
	.dc.b	$15,$FB,$96,$E8,$69,$27,$DA,$71,$3F,$94,$F8,$DF,$CA,$0F,$E5,$3E
	.dc.b	$2D,$8C,$90,$FE,$4E,$E5,$FE,$53,$17,$2F,$E5,$33,$70,$FE,$41,$36
	.dc.b	$BF,$98,$CD,$9F,$E2,$13,$70,$4D,$8C,$39,$F1,$BF,$9E,$F8,$DF,$CF
	.dc.b	$9B,$87,$F4,$26,$BF,$CF,$1B,$8F,$F3,$C6,$E3,$FC,$CD,$B3,$FD,$11
	.dc.b	$B1,$FC,$8D,$B8,$9A,$49,$FE,$A3,$7E,$36,$9F,$D5,$3C,$5F,$EA,$67
	.dc.b	$F5,$4F,$1A,$C6,$48,$7F,$52,$72,$FF,$54,$8B,$8F,$F5,$49,$B8,$FF
	.dc.b	$4B,$9F,$EA,$84,$AF,$EA,$8D,$7F,$AA,$35,$FE,$A8,$D7,$91,$C0,$87
	.dc.b	$03,$FB,$6F,$8D,$E0,$7F,$B4,$B6,$7F,$A4,$37,$2F,$F3,$AC,$FB,$30
	.dc.b	$FE,$D7,$39,$3F,$BA,$F8,$B2,$7F,$BC,$F8,$D5,$FE,$E7,$16,$BF,$B9
	.dc.b	$CD,$C3,$FB,$54,$DA,$FE,$F7,$36,$7F,$AD,$4D,$C1,$36,$30,$DA,$FF
	.dc.b	$04,$C7,$F8,$22,$38,$78,$DF,$E0,$AF,$17,$FC,$16,$DA,$FF,$07,$B5
	.dc.b	$FF,$04,$B7,$1F,$F0,$4B,$71,$FF,$00,$36,$7F,$C2,$2D,$8F,$EE,$0D
	.dc.b	$C7,$FC,$F7,$3F,$E1,$CB,$F0,$B4,$F0,$94,$F0,$96,$FF,$C3,$3E,$37
	.dc.b	$F8,$88,$FF,$11,$F8,$D6,$7F,$C4,$F1,$5F,$F1,$0C,$59,$E3,$1A,$3F
	.dc.b	$C5,$B1,$9C,$7F,$C3,$55,$FE,$23,$24,$FF,$88,$DA,$FF,$88,$DA,$FF
	.dc.b	$88,$DA,$F2,$38,$10,$E0,$7F,$8F,$3C,$5F,$03,$FE,$38,$6C,$FF,$85
	.dc.b	$DB,$89,$A4,$9F,$F2,$2F,$8D,$FE,$44,$3F,$C8,$BE,$2D,$8C,$90,$FF
	.dc.b	$21,$B9,$7F,$C8,$B1,$72,$FF,$22,$CD,$C3,$FC,$79,$36,$BF,$C9,$B3
	.dc.b	$67,$FC,$59,$37,$0F,$F3,$9D,$77,$BF,$F2,$7C,$E6,$F5,$E2,$7F,$96
	.dc.b	$BC,$5F,$F2,$DB,$6F,$FC,$BE,$D7,$FC,$B2,$DB,$FF,$2C,$B7,$1F,$F2
	.dc.b	$83,$67,$FC,$C2,$D8,$FF,$20,$37,$13,$49,$3F,$E6,$EB,$F1,$B4,$0B
	.dc.b	$5E,$07,$EC,$CF,$1B,$75,$B9,$D4,$E6,$39,$D7,$27,$9C,$7F,$9B,$A3
	.dc.b	$13,$D2,$B9,$C6,$27,$95,$75,$38,$CE,$AB,$BF,$8D,$67,$27,$82,$3C
	.dc.b	$73,$97,$B4,$64,$F1,$CE,$6B,$BD,$F8,$6F,$DB,$5E,$EB,$E2,$E1,$AE
	.dc.b	$6B,$CE,$D9,$E2,$F8,$8B,$3C,$1C,$1C,$E7,$8B,$83,$91,$C1,$CB,$EE
	.dc.b	$3E,$24,$7D,$4F,$1B,$D5,$5E,$B6,$CF,$BE,$74,$DA,$F7,$C7,$D8,$9D
	.dc.b	$CF,$49,$46,$2F,$EA,$3F,$1F,$C6,$95,$CF,$C5,$AF,$18,$B5,$F6,$E6
	.dc.b	$E1,$F8,$69,$B5,$F8,$49,$B3,$F8,$E9,$B8,$26,$C6,$1B,$5E,$B8,$FB
	.dc.b	$51,$C3,$C6,$FB,$9E,$37,$DD,$6D,$7C,$16,$BE,$F6,$E3,$F9,$16,$E3
	.dc.b	$F9,$C6,$CF,$E6,$9B,$1F,$A4,$6E,$26,$92,$7E,$85,$F8,$DA,$7D,$3F
	.dc.b	$1B,$F2,$47,$E4,$FC,$6B,$19,$21,$F5,$5C,$BF,$5E,$2E,$3F,$A9,$9B
	.dc.b	$8F,$ED,$27,$F2,$75,$FB,$26,$C7,$EF,$1B,$89,$A4,$9F,$C9,$35,$FC
	.dc.b	$C3,$5F,$D1,$35,$F0,$1F,$8F,$F1,$BC,$6F,$F0,$26,$D7,$F0,$C6,$D1
	.dc.b	$A4,$9F,$CC,$78,$DF,$B7,$3F,$71,$E2,$D8,$C9,$0F,$CD,$39,$7F,$39
	.dc.b	$17,$2F,$E1,$D3,$70,$FE,$33,$36,$BF,$8B,$CD,$9F,$E4,$33,$70,$4D
	.dc.b	$8C,$39,$F1,$BF,$86,$78,$DF,$C3,$5B,$87,$EC,$5A,$FE,$BD,$B8,$FF
	.dc.b	$21,$6E,$3F,$CB,$1B,$3F,$CA,$9B,$1F,$CD,$1B,$89,$A4,$9F,$E0,$17
	.dc.b	$E3,$69,$FC,$0F,$C6,$FE,$44,$7F,$23,$F1,$AC,$64,$87,$F0,$57,$2F
	.dc.b	$F0,$78,$B8,$FF,$39,$9B,$8F,$F4,$49,$FC,$69,$2B,$F8,$E3,$5F,$E4
	.dc.b	$4D,$7F,$96,$35,$E4,$70,$21,$C0,$FE,$39,$E3,$78,$1F,$E9,$ED,$9F
	.dc.b	$EB,$2D,$CB,$FD,$05,$3E,$CC,$3F,$A8,$4E,$4F,$E8,$9E,$34,$9F,$E5
	.dc.b	$1E,$2D,$7F,$93,$45,$AF,$EB,$13,$70,$FE,$C7,$36,$BF,$B0,$CD,$9F
	.dc.b	$ED,$73,$70,$4D,$8C,$36,$BF,$9C,$47,$F5,$78,$E1,$E3,$7F,$59,$F1
	.dc.b	$BF,$AD,$36,$BF,$9F,$35,$FE,$78,$DC,$7F,$B7,$37,$1F,$EF,$2D,$9F
	.dc.b	$EE,$ED,$8F,$F0,$0B,$71,$34,$93,$FD,$36,$FC,$6D,$3F,$A8,$78,$DF
	.dc.b	$DB,$CF,$EE,$1E,$35,$8C,$90,$FE,$A4,$E5,$FE,$A9,$17,$1F,$F0,$34
	.dc.b	$DC,$7F,$C2,$33,$FC,$68,$95,$FD,$A5,$AF,$F6,$F6,$BF,$DE,$5A,$F2
	.dc.b	$38,$10,$E0,$7F,$69,$F1,$BC,$0F,$F8,$71,$B3,$FE,$26,$6E,$5F,$E8
	.dc.b	$29,$F4,$9C,$4F,$76,$1F,$E1,$F9,$FF,$09,$78,$D2,$7F,$BB,$78,$B5
	.dc.b	$FE,$E9,$16,$BF,$C4,$D3,$70,$FF,$18,$CD,$AF,$F1,$7C,$D9,$FF,$1C
	.dc.b	$CD,$C1,$36,$30,$DA,$FF,$03,$47,$F8,$96,$38,$78,$DF,$E2,$7F,$1B
	.dc.b	$FC,$50,$DA,$FF,$06,$35,$FF,$05,$37,$1F,$F1,$E3,$71,$FF,$23,$B6
	.dc.b	$7F,$C8,$AD,$8F,$F2,$7B,$71,$34,$93,$FE,$1B,$BF,$1B,$4F,$F0,$F7
	.dc.b	$8D,$FE,$3D,$3F,$C7,$DE,$35,$8C,$90,$FF,$10,$39,$7F,$C4,$51,$71
	.dc.b	$FF,$2A,$4D,$C7,$FC,$BF,$3F,$DA,$09,$5F,$E3,$76,$BF,$E3,$D6,$BF
	.dc.b	$E4,$76,$BC,$8E,$04,$38,$1F,$E3,$7F,$17,$C0,$FF,$9A,$9B,$3F,$E7
	.dc.b	$46,$E5,$FE,$80,$9F,$66,$1F,$E6,$B9,$C9,$FE,$5F,$F1,$7F,$CC,$9F
	.dc.b	$B3,$FF,$90,$AB,$FC,$EB,$18,$F1,$75,$1F,$E4,$99,$D6,$3F,$E8,$2A
	.dc.b	$D4,$FF,$92,$DA,$FF,$92,$5B,$87,$F9,$ED,$B3,$FE,$86,$6C,$F8,$0E
	.dc.b	$D7,$E0,$7F,$CB,$B7,$E0,$C5,$5C,$2F,$C5,$5C,$0F,$37,$8D,$EC,$3D
	.dc.b	$BE,$35,$8C,$90,$EA,$E5,$EF,$17,$2F,$F4,$AE,$3F,$E6,$29,$E1,$18
	.dc.b	$9D,$5F,$FA,$96,$B5,$5C,$8D,$C6,$67,$C7,$18,$3F,$D5,$B3,$FE,$A3
	.dc.b	$9D,$47,$CF,$BE,$AE,$BC,$6F,$AE,$7D,$8F,$1A,$C7,$D5,$8A,$FC,$88
	.dc.b	$B8,$FA,$A6,$DF,$9E,$6D,$75,$3F,$D2,$B8,$2B,$FC,$8A,$D8,$CC,$60
	.dc.b	$8E,$08,$7F,$A9,$63,$83,$63,$6D,$37,$7F,$3B,$C6,$F1,$E0,$BF,$F3
	.dc.b	$55,$7B,$4F,$C9,$4E,$A3,$31,$F8,$6B,$FC,$DC,$E4,$FC,$EF,$8D,$F9
	.dc.b	$EF,$16,$D7,$E6,$DA,$FE,$39,$B8,$FE,$01,$B8,$FD,$D6,$CF,$D8,$6C
	.dc.b	$79,$DB,$8F,$F9,$A6,$7F,$51,$7E,$16,$9E,$12,$9E,$12,$DF,$85,$FD
	.dc.b	$87,$8D,$FB,$63,$F6,$FE,$35,$8C,$90,$FD,$3B,$97,$F5,$31,$67,$F3
	.dc.b	$73,$71,$FC,$C4,$DA,$FC,$9C,$D9,$FA,$53,$70,$FF,$3A,$57,$7B,$FD
	.dc.b	$2C,$E6,$F5,$E3,$7F,$0C,$3F,$86,$F8,$D6,$FF,$85,$45,$7F,$7B,$16
	.dc.b	$FF,$5F,$37,$1F,$D6,$4D,$AF,$D3,$CF,$E5,$08,$FE,$61,$AF,$E4,$5A
	.dc.b	$FF,$BE,$71,$46,$23,$F7,$0D,$7A,$9C,$CF,$E4,$1E,$37,$86,$FF,$28
	.dc.b	$DA,$FF,$37,$4F,$B0,$D9,$4F,$E3,$31,$B3,$F9,$7F,$8D,$FC,$C3,$C5
	.dc.b	$B3,$FC,$B5,$AF,$F2,$06,$E5,$FC,$49,$B8,$7F,$0F,$6C,$FF,$0A,$6C
	.dc.b	$7E,$A5,$B8,$1A,$49,$FE,$79,$7E,$36,$9E,$33,$FA,$07,$8D,$FD,$18
	.dc.b	$FE,$8F,$E3,$58,$C9,$0F,$E6,$EE,$5F,$E7,$31,$71,$FE,$57,$37,$1F
	.dc.b	$E5,$13,$6B,$F9,$1C,$D9,$FE,$05,$37,$1F,$F3,$BD,$74,$3C,$17,$FC
	.dc.b	$DA,$7C,$B1,$8A,$FE,$69,$E3,$7F,$5E,$3F,$AF,$F8,$D6,$7F,$AE,$45
	.dc.b	$7F,$A7,$C5,$BF,$E8,$33,$6B,$F9,$F4,$DA,$FE,$73,$3F,$C9,$88,$FF
	.dc.b	$2A,$6B,$FC,$8D,$AF,$E4,$DA,$F9,$8E,$04,$38,$1F,$DC,$7C,$6F,$05
	.dc.b	$7F,$25,$6E,$3F,$E7,$29,$F5,$9B,$29,$FD,$AA,$36,$7F,$80,$3C,$6F
	.dc.b	$F0,$0F,$8B,$67,$FB,$F3,$5F,$EE,$2D,$CB,$FB,$1B,$70,$FE,$C0,$D9
	.dc.b	$FE,$B6,$D8,$FE,$74,$DC,$0D,$24,$FF,$83,$6F,$C6,$D3,$C6,$7F,$84
	.dc.b	$7C,$6F,$F0,$B9,$FE,$18,$F1,$AC,$64,$87,$F8,$21,$CB,$FE,$0A,$8B
	.dc.b	$8F,$F7,$C9,$B8,$FF,$79,$9B,$5F,$DD,$26,$CF,$F5,$39,$B8,$FF,$9E
	.dc.b	$EF,$FC,$0B,$18,$3C,$97,$E2,$8F,$F0,$4F,$8D,$FE,$2E,$3F,$C5,$FE
	.dc.b	$35,$9F,$F1,$64,$57,$FC,$3F,$16,$78,$C6,$8F,$F1,$94,$67,$0F,$F0
	.dc.b	$85,$7F,$76,$24,$FF,$7A,$6B,$FD,$CD,$AF,$F2,$36,$BC,$8E,$04,$38
	.dc.b	$1F,$E4,$1F,$1B,$C0,$FF,$86,$1B,$3F,$DD,$DB,$8F,$F9,$DA,$7D,$07
	.dc.b	$F8,$EA,$32,$7F,$94,$3C,$6F,$F2,$8F,$8B,$67,$FC,$9C,$D7,$FC,$82
	.dc.b	$DC,$BF,$C6,$6D,$C3,$FC,$60,$D9,$FF,$15,$B6,$3F,$C1,$4D,$C0,$D2
	.dc.b	$4F,$F9,$76,$FC,$6D,$3C,$67,$F9,$87,$C6,$FF,$33,$9F,$E6,$8F,$1A
	.dc.b	$C6,$48,$7F,$96,$1C,$BF,$E5,$A8,$B8,$FF,$93,$26,$E3,$FE,$49,$9B
	.dc.b	$5F,$E4,$49,$B3,$FE,$23,$9B,$8F,$F9,$F6,$BA,$1E,$0B,$FF,$2B,$CF
	.dc.b	$96,$31,$5F,$E5,$5F,$13,$FC,$FE,$22,$52,$70,$31,$FF,$3F,$E1,$FE
	.dc.b	$70,$AF,$F4,$15,$F2,$78,$D7,$0A,$DC,$E6,$7F,$CF,$2F,$75,$7D,$DF
	.dc.b	$04,$F0,$7A,$9F,$E5,$29,$FF,$28,$91,$C1,$0F,$F2,$73,$96,$99,$3F
	.dc.b	$C0,$0E,$5A,$7A,$0F,$0F,$89,$E7,$AF,$25,$FC,$18,$C3,$ED,$6C,$F0
	.dc.b	$3C,$E7,$B2,$34,$E2,$79,$DF,$F9,$F6,$3D,$FE,$27,$C3,$9C,$F8,$76
	.dc.b	$FE,$54,$DA,$F8,$B3,$67,$EA,$4D,$BF,$F3,$6D,$F6,$8C,$56,$5B,$7E
	.dc.b	$FB,$E2,$DB,$E3,$1E,$08,$C5,$71,$69,$BA,$E6,$FB,$CF,$59,$A3,$13
	.dc.b	$D9,$B5,$D9,$D9,$8A,$F5,$78,$92,$3C,$9E,$25,$78,$C5,$C7,$E5,$CD
	.dc.b	$C3,$F2,$13,$6B,$F1,$73,$67,$F3,$53,$70,$FF,$43,$DF,$D4,$8C,$1F
	.dc.b	$2E,$AD,$FC,$58,$C3,$4F,$91,$E1,$FD,$13,$E9,$78,$96,$32,$43,$E0
	.dc.b	$B9,$7E,$1C,$5B,$FC,$2C,$DC,$7F,$55,$3F,$A7,$AF,$D4,$47,$EC,$E3
	.dc.b	$11,$A2,$C6,$5B,$8A,$77,$1F,$AE,$6C,$EE,$3B,$35,$EF,$3C,$5E,$73
	.dc.b	$C2,$31,$1C,$1A,$F0,$8F,$CA,$78,$9E,$B8,$FD,$EB,$6B,$F8,$63,$68
	.dc.b	$D2,$4F,$E5,$7C,$4F,$CF,$78,$7F,$9F,$6C,$64,$FC,$D4,$65,$A6,$4F
	.dc.b	$C8,$B9,$69,$9C,$BF,$6B,$19,$6C,$67,$0F,$E3,$93,$96,$C6,$4F,$E3
	.dc.b	$31,$96,$9F,$AD,$F1,$3F,$6C,$7E,$DF,$C4,$B1,$92,$1F,$A6,$72,$FE
	.dc.b	$A2,$2D,$FF,$0A,$9B,$7F,$CB,$E6,$D7,$F2,$A9,$B3,$FC,$E6,$6E,$5F
	.dc.b	$E9,$6A,$E8,$78,$2F,$F8,$AC,$F9,$63,$15,$FC,$EF,$C3,$FE,$1C,$7F
	.dc.b	$0F,$F1,$2C,$FF,$09,$8A,$FE,$FA,$2D,$FF,$20,$9B,$5F,$D1,$66,$D7
	.dc.b	$F4,$09,$FC,$61,$1F,$CC,$B5,$FC,$73,$5F,$CA,$B5,$F3,$1C,$08,$70
	.dc.b	$3F,$8F,$78,$7E,$0A,$FE,$B8,$DC,$4D,$24,$FB,$4E,$27,$F2,$7F,$13
	.dc.b	$F9,$87,$87,$FC,$C5,$B1,$93,$F9,$64,$65,$A6,$4F,$E4,$6E,$5A,$67
	.dc.b	$2F,$E8,$F1,$96,$C6,$70,$FE,$DD,$39,$6C,$64,$FE,$D7,$19,$69,$FC
	.dc.b	$FF,$C4,$FE,$90,$7F,$49,$F1,$2C,$64,$87,$F3,$87,$2F,$F3,$A8,$B7
	.dc.b	$FD,$72,$6D,$FF,$80,$66,$D7,$F7,$C9,$B3,$FE,$0B,$9B,$97,$FA,$5E
	.dc.b	$BA,$1E,$0B,$FE,$CF,$3E,$58,$C5,$7F,$83,$7C,$3F,$EC,$07,$F6,$1F
	.dc.b	$12,$CF,$F5,$B8,$AF,$F5,$08,$B7,$FD,$CA,$6D,$7F,$85,$E6,$D7,$F8
	.dc.b	$4A,$7F,$8F,$11,$FE,$56,$D7,$F9,$03,$5F,$E4,$ED,$7C,$C7,$02,$1C
	.dc.b	$0F,$EE,$1E,$1F,$82,$BF,$C5,$CD,$C4,$D2,$4F,$B4,$E2,$7F,$79,$F1
	.dc.b	$3F,$C0,$5E,$1F,$F8,$0D,$B1,$93,$FB,$F4,$65,$A6,$4F,$EE,$AE,$5A
	.dc.b	$67,$2F,$F0,$CC,$65,$B1,$9C,$3F,$C8,$13,$96,$C6,$4F,$F1,$EC,$65
	.dc.b	$A7,$F8,$47,$C4,$FF,$0D,$1F,$E1,$BF,$12,$C6,$48,$7F,$82,$9C,$BF
	.dc.b	$E0,$C8,$B7,$FE,$2E,$9B,$7F,$E5,$39,$B5,$FE,$4E,$9B,$3F,$E5,$D9
	.dc.b	$B9,$7F,$A5,$EB,$A1,$E0,$BF,$F1,$CC,$F9,$63,$15,$FE,$5F,$AF,$F1
	.dc.b	$91,$FE,$33,$F1,$2C,$FF,$8B,$62,$BF,$E2,$28,$B7,$FE,$44,$9B,$5F
	.dc.b	$E6,$99,$B5,$FE,$64,$9F,$EE,$04,$7F,$BE,$B5,$FE,$E4,$D7,$FB,$CB
	.dc.b	$5F,$31,$C0,$87,$03,$FC,$85,$E1,$F8,$2B,$FD,$00,$D9,$34,$93,$ED
	.dc.b	$38,$9F,$E4,$BF,$13,$FC,$A9,$E1,$FF,$95,$5B,$19,$3F,$CA,$11,$96
	.dc.b	$99,$3F,$C8,$EE,$5A,$67,$1F,$F3,$5C,$65,$B1,$9C,$3F,$D2,$53,$96
	.dc.b	$C6,$4F,$F4,$7C,$65,$A7,$F9,$97,$F6,$87,$8F,$FE,$6F,$6C,$67,$C5
	.dc.b	$C4,$C8,$E0,$EE,$31,$1F,$E6,$29,$E2,$D9,$D3,$C0,$86,$4E,$C6,$30
	.dc.b	$FF,$49,$CF,$26,$CF,$29,$E0,$48,$CB,$C8,$C6,$1F,$EB,$CA,$78,$BE
	.dc.b	$D3,$DD,$E2,$DA,$F8,$51,$5F,$54,$5B,$F1,$CD,$BE,$93,$6B,$FD,$95
	.dc.b	$7D,$9F,$F2,$AB,$5F,$F2,$7B,$5F,$F2,$EB,$5F,$35,$7F,$95,$A3,$11
	.dc.b	$D2,$30,$47,$AB,$5E,$AD,$BF,$93,$E2,$79,$70,$5A,$30,$DB,$F8,$27
	.dc.b	$E0,$E3,$24,$7F,$0D,$E2,$FE,$17,$C4,$FC,$33,$63,$27,$E2,$E3,$2D
	.dc.b	$32,$7E,$15,$CB,$4C,$E1,$F8,$08,$CB,$63,$38,$7D,$B9,$CB,$63,$27
	.dc.b	$D2,$8C,$B4,$FC,$F7,$8B,$F9,$D3,$F3,$DE,$2D,$8C,$90,$FC,$D3,$97
	.dc.b	$F3,$91,$6F,$F2,$B3,$6F,$F1,$D3,$6B,$F0,$73,$67,$E8,$CD,$CB,$FD
	.dc.b	$03,$5D,$EF,$F3,$F3,$9B,$D7,$8B,$FB,$33,$F6,$9E,$2D,$BF,$DC,$45
	.dc.b	$7F,$61,$16,$FF,$53,$37,$1F,$D1,$4D,$AF,$CB,$CF,$E3,$C8,$FE,$61
	.dc.b	$AF,$E6,$1A,$FE,$61,$AF,$EB,$5A,$F4,$39,$1F,$C3,$7C,$4F,$05,$7E
	.dc.b	$49,$B9,$1A,$49,$F6,$9C,$4F,$E2,$DE,$2F,$F1,$5F,$13,$F8,$B3,$63
	.dc.b	$27,$F1,$B8,$CB,$4C,$9F,$C5,$5C,$B4,$CE,$5F,$C4,$23,$2D,$8C,$E1
	.dc.b	$FC,$2E,$72,$D8,$C9,$FC,$0A,$32,$D3,$F9,$77,$8B,$FC,$B4,$FE,$5D
	.dc.b	$E2,$D8,$C9,$0F,$E5,$4E,$5F,$E5,$91,$6F,$F9,$2C,$DB,$FE,$3D,$36
	.dc.b	$BF,$89,$CD,$9F,$E0,$33,$72,$FF,$40,$D7,$7B,$FE,$5F,$39,$BD,$78
	.dc.b	$BF,$D0,$CF,$E8,$9E,$2D,$BF,$E9,$11,$5F,$E8,$11,$6F,$F9,$CC,$DC
	.dc.b	$7F,$99,$4D,$AF,$E4,$F3,$FC,$7C,$8F,$F2,$86,$BF,$CA,$1A,$FF,$28
	.dc.b	$6B,$FC,$F5,$AF,$43,$91,$FD,$77,$C4,$F0,$57,$F2,$26,$E4,$69,$27
	.dc.b	$DA,$71,$3F,$B3,$78,$BF,$D9,$7C,$4F,$EC,$CD,$8C,$9F,$DA,$E3,$2D
	.dc.b	$32,$7F,$65,$72,$D3,$39,$7F,$60,$8C,$B6,$33,$87,$F5,$B9,$CB,$63
	.dc.b	$27,$F5,$28,$CB,$4F,$EF,$DE,$2F,$F7,$D3,$FB,$F7,$8B,$63,$24,$3F
	.dc.b	$BD,$39,$7F,$BE,$45,$BF,$EE,$B3,$6F,$FB,$74,$DA,$FE,$C7,$36,$7F
	.dc.b	$A8,$CD,$CB,$FC,$FF,$5D,$EF,$FB,$FC,$E6,$F5,$E2,$FF,$84,$CF,$F0
	.dc.b	$A7,$8B,$6F,$FC,$31,$15,$FF,$08,$45,$BF,$F0,$4C,$DC,$7F,$C0,$53
	.dc.b	$6B,$FB,$BC,$FF,$6F,$23,$FD,$E1,$AF,$F7,$86,$BF,$DE,$1A,$FF,$83
	.dc.b	$5A,$F4,$39,$1F,$E2,$DF,$13,$C1,$5F,$DC,$9B,$91,$A4,$9F,$69,$C4
	.dc.b	$FF,$1B,$78,$BF,$E3,$5F,$13,$FC,$6C,$D8,$C9,$FE,$3B,$8C,$B4,$C9
	.dc.b	$FE,$35,$72,$D3,$39,$7F,$8C,$23,$2D,$8C,$E1,$FE,$2B,$9C,$B6,$32
	.dc.b	$7F,$88,$A3,$2D,$3F,$C9,$DE,$2F,$F9,$34,$FF,$27,$78,$B6,$32,$43
	.dc.b	$FC,$94,$E5,$FF,$26,$45,$BF,$F2,$2C,$DB,$FF,$1F,$4D,$AF,$F1,$9C
	.dc.b	$D9,$FF,$10,$CD,$CB,$FC,$FB,$5D,$EF,$FC,$9F,$39,$BD,$78,$BF,$E6
	.dc.b	$33,$FC,$C9,$E2,$DB,$FF,$34,$45,$7F,$CC,$11,$6F,$FC,$B3,$37,$1F
	.dc.b	$F2,$94,$DA,$FF,$23,$CF,$F8,$FC,$8F,$F9,$21,$AF,$F9,$21,$AF,$F9
	.dc.b	$21,$AF,$F9,$75,$AF,$43,$90,$14,$AC,$0F,$DA,$7F,$CD,$CF,$39,$E1
	.dc.b	$3B,$78,$4E,$9E,$C7,$49,$E7,$19,$7F,$C9,$F5,$E1,$9D,$1C,$BC,$5F
	.dc.b	$85,$3E,$16,$D7,$11,$E3,$71,$C3,$C5,$5E,$7F,$17,$0F,$82,$6E,$1E
	.dc.b	$69,$B3,$ED,$9B,$7F,$E7,$3B,$E6,$F6,$BF,$7D,$7A,$AF,$77,$C1,$B3
	.dc.b	$A3,$B1,$67,$B3,$6B,$B4,$79,$63,$13,$C9,$F8,$71,$EB,$78,$57,$76
	.dc.b	$9C,$0A,$70,$AE,$AF,$D4,$F1,$BD,$A7,$BB,$C6,$B1,$E3,$8A,$F1,$8B
	.dc.b	$8F,$C2,$9B,$87,$DA,$9B,$5D,$4E,$47,$B5,$CC,$7E,$26,$6D,$FF,$A2
	.dc.b	$EB,$CD,$18,$3C,$8F,$23,$D5,$E3,$7D,$43,$EA,$F8,$D6,$3E,$54,$57
	.dc.b	$E0,$45,$C7,$EC,$CD,$C3,$F2,$33,$6B,$F2,$F3,$F9,$17,$F2,$53,$F9
	.dc.b	$F8,$C4,$7E,$61,$AF,$03,$31,$B6,$E0,$9D,$CE,$67,$B3,$5E,$F3,$B7
	.dc.b	$8C,$F0,$8C,$47,$06,$BC,$23,$F2,$FE,$37,$AE,$BF,$6C,$DA,$34,$95
	.dc.b	$F2,$0F,$C9,$3F,$A0,$F1,$BF,$43,$E2,$D9,$FC,$B3,$5F,$C6,$37,$2F
	.dc.b	$D2,$B7,$0F,$DE,$B6,$7F,$81,$B6,$3F,$85,$37,$03,$49,$3F,$A5,$BF
	.dc.b	$1B,$4F,$19,$FA,$9F,$1B,$F6,$E7,$EE,$3C,$6B,$19,$21,$FA,$37,$2F
	.dc.b	$E9,$62,$E3,$FB,$E9,$B8,$FF,$16,$9B,$5F,$C7,$26,$CF,$F2,$39,$B8
	.dc.b	$FF,$A5,$6B,$BD,$FF,$08,$9C,$DE,$BC,$6F,$E1,$A7,$F0,$EF,$1A,$DF
	.dc.b	$F0,$38,$AF,$EE,$62,$DF,$F1,$29,$B8,$FF,$2A,$9B,$5F,$CC,$27,$F1
	.dc.b	$04,$7F,$26,$D7,$F1,$4D,$7F,$22,$D7,$F9,$AB,$5E,$87,$23,$F8,$C7
	.dc.b	$8D,$E0,$AF,$E8,$8D,$C8,$D2,$4F,$B4,$E2,$7F,$21,$F1,$BF,$97,$9F
	.dc.b	$CC,$3C,$5B,$19,$21,$FC,$6D,$CB,$FC,$76,$2E,$5F,$CD,$A6,$E1,$FD
	.dc.b	$3A,$6D,$7F,$54,$9B,$3F,$D6,$E6,$E0,$9B,$18,$73,$E3,$7F,$44,$F1
	.dc.b	$BF,$A2,$B7,$0F,$E7,$4D,$7F,$98,$37,$1F,$E9,$2D,$C7,$FB,$0B,$67
	.dc.b	$FB,$3B,$63,$FB,$73,$71,$34,$93,$FD,$26,$FC,$6D,$3F,$A6,$78,$DF
	.dc.b	$D6,$8F,$EB,$7E,$35,$8C,$90,$FE,$8C,$E5,$FE,$91,$17,$1F,$EC,$33
	.dc.b	$71,$FE,$F3,$3F,$C5,$49,$5F,$C8,$1A,$FF,$17,$6B,$FC,$79,$AF,$23
	.dc.b	$81,$0E,$07,$F6,$5F,$1B,$C0,$FF,$83,$1B,$3F,$E1,$36,$E4,$69,$27
	.dc.b	$FB,$77,$8D,$FD,$F0,$FE,$FB,$E2,$D8,$C9,$0F,$ED,$0E,$5F,$ED,$51
	.dc.b	$72,$FF,$01,$CD,$C3,$FC,$37,$36,$BF,$C4,$33,$67,$FC,$51,$37,$04
	.dc.b	$D8,$C3,$9F,$1B,$FC,$23,$E3,$7F,$84,$9B,$87,$F8,$1D,$AF,$F7,$D6
	.dc.b	$E3,$FE,$16,$6E,$3F,$E2,$E6,$CF,$F8,$D1,$B1,$FE,$3B,$6E,$26,$92
	.dc.b	$7F,$C2,$D7,$E3,$69,$FE,$19,$F1,$BF,$C4,$E7,$F8,$A3,$C6,$B1,$92
	.dc.b	$1F,$E1,$37,$2F,$F8,$56,$2E,$3F,$E2,$E9,$B8,$FF,$91,$A7,$FB,$11
	.dc.b	$2B,$FB,$6B,$5F,$EC,$8D,$7F,$B5,$B5,$E4,$70,$21,$C0,$FF,$19,$78
	.dc.b	$BE,$07,$FC,$B2,$D9,$FF,$30,$37,$23,$49,$3F,$E3,$BF,$17,$FC,$92
	.dc.b	$7F,$92,$BC,$4B,$19,$21,$FE,$35,$72,$FF,$8D,$E2,$E5,$FE,$50,$9B
	.dc.b	$87,$F9,$A2,$6D,$7F,$9B,$A6,$CF,$F9,$D6,$6E,$09,$B1,$87,$3E,$2F
	.dc.b	$F9,$83,$F6,$AF,$F3,$53,$7C,$6F,$F3,$0B,$71,$FF,$2B,$35,$FF,$26
	.dc.b	$37,$2F,$F3,$2B,$72,$FF,$3E,$B6,$7F,$D0,$6D,$8F,$F4,$53,$72,$34
	.dc.b	$93,$FE,$67,$C3,$FD,$0C,$7F,$9C,$5F,$F3,$1B,$6F,$FC,$FE,$FF,$97
	.dc.b	$AC,$E5,$FF,$4C,$B6,$7F,$D3,$8D,$8D,$F8,$DF,$0F,$C6,$95,$E1,$7E
	.dc.b	$34,$62,$3C,$31,$FE,$3C,$AF,$24,$7F,$92,$1A,$F8,$23,$44,$7F,$C8
	.dc.b	$F1,$A7,$C7,$19,$7F,$C8,$B1,$9A,$F5,$CE,$67,$B3,$F6,$7C,$5F,$26
	.dc.b	$1F,$09,$B5,$F0,$1B,$5F,$01,$B1,$F1,$9B,$67,$F0,$9E,$34,$AF,$A5
	.dc.b	$E2,$D7,$E8,$C5,$C3,$E8,$CD,$C3,$E7,$CD,$AF,$9F,$36,$7E,$AC,$DC
	.dc.b	$3F,$CB,$B7,$F6,$A3,$07,$E5,$6A,$DF,$E0,$E3,$0D,$3F,$0D,$E3,$7E
	.dc.b	$6F,$C6,$FC,$E3,$63,$27,$E2,$E3,$2D,$32,$7E,$31,$CB,$4C,$DF,$E3
	.dc.b	$A3,$2D,$8C,$E3,$F8,$E9,$CB,$63,$27,$E4,$23,$2D,$3F,$3D,$E2,$FE
	.dc.b	$CC,$FD,$A7,$8D,$63,$24,$3F,$42,$E5,$FD,$1C,$5B,$FD,$1C,$DB,$FD
	.dc.b	$0C,$FE,$5C,$95,$F9,$26,$BF,$90,$6B,$F9,$B6,$BC,$8E,$04,$38,$1F
	.dc.b	$B7,$F1,$BC,$0F,$EB,$9B,$3F,$B4,$6E,$86,$92,$7F,$80,$78,$DF,$C5
	.dc.b	$4F,$E2,$DE,$2D,$8C,$90,$FE,$06,$E5,$FE,$0B,$17,$2F,$E0,$B3,$70
	.dc.b	$FE,$07,$36,$BF,$81,$CD,$9F,$E1,$33,$70,$4D,$8C,$39,$F1,$BF,$95
	.dc.b	$78,$DF,$CA,$DB,$87,$F1,$96,$BF,$C6,$1B,$8F,$F1,$86,$E3,$FC,$59
	.dc.b	$B3,$FC,$59,$B1,$FC,$71,$B8,$9A,$49,$FE,$51,$7E,$36,$9F,$CA,$FC
	.dc.b	$6F,$E8,$07,$F4,$1F,$1A,$C6,$48,$7F,$2E,$72,$FF,$30,$8B,$8F,$F3
	.dc.b	$09,$B8,$FF,$2E,$9F,$E4,$84,$AF,$E3,$ED,$7F,$8E,$B5,$FE,$50,$D7
	.dc.b	$91,$C0,$87,$03,$FA,$27,$8D,$E0,$7F,$9D,$B6,$7F,$A0,$B7,$2F,$F3
	.dc.b	$4C,$FB,$30,$FE,$8F,$39,$3F,$B1,$78,$D2,$7F,$A8,$78,$B5,$FE,$9F
	.dc.b	$16,$BF,$A7,$CD,$C3,$FA,$6C,$DA,$FE,$9B,$36,$7F,$A9,$CD,$C1,$36
	.dc.b	$30,$DA,$FE,$C1,$1F,$DD,$63,$87,$8D,$FD,$DF,$C6,$FE,$F0,$DA,$FE
	.dc.b	$CC,$D7,$FB,$2B,$71,$FE,$CA,$DC,$7F,$B1,$B6,$7F,$B1,$B6,$3F,$B4
	.dc.b	$B7,$13,$49,$3F,$DD,$6F,$C6,$D3,$FB,$C7,$8D,$FE,$0D,$3F,$C1,$DE
	.dc.b	$35,$8C,$90,$FE,$F6,$E5,$FE,$FB,$17,$1F,$EF,$B3,$71,$FE,$F7,$3F
	.dc.b	$DC,$49,$5F,$DB,$1A,$FF,$6A,$6B,$FD,$D5,$AF,$23,$81,$0E,$07,$F8
	.dc.b	$47,$C6,$F0,$3F,$E0,$86,$CF,$F8,$39,$B9,$7F,$9A,$67,$D9,$87,$F8
	.dc.b	$52,$72,$7F,$8B,$FC,$69,$3F,$E1,$DF,$16,$BF,$E1,$C8,$B5,$FE,$1C
	.dc.b	$9B,$87,$F8,$6A,$6D,$7F,$86,$A6,$CF,$F8,$82,$6E,$09,$B1,$86,$D7
	.dc.b	$F8,$B6,$3F,$C8,$51,$C3,$C6,$FF,$22,$78,$DF,$E4,$56,$D7,$F8,$CD
	.dc.b	$AF,$F8,$C9,$B8,$FF,$8C,$9B,$8F,$F8,$C1,$B3,$FE,$30,$6C,$7F,$8D
	.dc.b	$9B,$89,$A4,$9F,$F2,$15,$F8,$DA,$7F,$91,$7C,$6F,$F2,$D1,$FE,$5B
	.dc.b	$F1,$AC,$64,$87,$F9,$21,$CB,$FE,$4A,$8B,$8F,$F9,$2A,$6E,$3F,$E4
	.dc.b	$89,$FF,$1F,$12,$BF,$C7,$2D,$7F,$C6,$ED,$7F,$C8,$4D,$79,$1C,$08
	.dc.b	$70,$3F,$CB,$DE,$27,$81,$FF,$2A,$B6,$7F,$CB,$6D,$CB,$FC,$CF,$3E
	.dc.b	$CC,$3F,$CC,$33,$91,$12,$15,$81,$EF,$3C,$7D,$D2,$B6,$F0,$9D,$3C
	.dc.b	$EB,$94,$7F,$9A,$4C,$C7,$18,$CD,$74,$AE,$71,$E0,$8C,$3C,$BC,$5B
	.dc.b	$5D,$E6,$CE,$E3,$C0,$64,$AF,$AA,$77,$7E,$2B,$EE,$D9,$F1,$D7,$08
	.dc.b	$F4,$5F,$02,$46,$6E,$BE,$B3,$C0,$5A,$F2,$36,$B9,$BD,$7C,$6F,$1F
	.dc.b	$8D,$E4,$6C,$64,$F1,$46,$5A,$64,$E8,$E5,$A6,$71,$F6,$C6,$5B,$19
	.dc.b	$BF,$AB,$39,$6C,$64,$FB,$91,$96,$9F,$17,$C5,$F9,$67,$CC,$F1,$AC
	.dc.b	$64,$87,$C1,$72,$FC,$38,$B7,$F4,$A6,$DF,$E2,$27,$F1,$F5,$F9,$08
	.dc.b	$FC,$9C,$62,$34,$58,$CB,$71,$4E,$E3,$F2,$CD,$9D,$C7,$66,$BD,$E7
	.dc.b	$8B,$CE,$78,$46,$23,$83,$5E,$11,$F8,$EF,$1B,$D7,$1F,$A5,$6D,$7E
	.dc.b	$B1,$B4,$69,$27,$F2,$5E,$37,$E6,$3C,$5F,$CC,$B6,$32,$7E,$5E,$32
	.dc.b	$D3,$27,$E4,$5C,$B4,$CE,$5F,$A2,$8C,$B6,$33,$87,$EE,$A7,$2D,$8C
	.dc.b	$9F,$C0,$63,$2D,$3F,$55,$E3,$7E,$C0,$FD,$8F,$8D,$63,$24,$3F,$4C
	.dc.b	$E5,$FD,$44,$5B,$FD,$BC,$DB,$FE,$1B,$36,$BF,$89,$4D,$9F,$E3,$33
	.dc.b	$72,$FF,$45,$57,$7B,$FE,$03,$39,$BD,$78,$DF,$C1,$8F,$E0,$FE,$35
	.dc.b	$BF,$E0,$91,$5F,$DE,$45,$BF,$E1,$73,$71,$FE,$43,$36,$BF,$93,$4F
	.dc.b	$E2,$88,$FE,$5D,$AF,$E3,$1A,$FE,$4D,$AF,$F2,$C6,$BD,$0E,$47,$F1
	.dc.b	$9F,$1B,$C1,$5F,$CE,$9B,$97,$FA,$0E,$7D,$66,$CA,$7F,$17,$8D,$9F
	.dc.b	$CA,$3C,$6F,$E5,$3E,$2D,$9F,$E4,$CD,$7F,$8F,$37,$2F,$E5,$AD,$C3
	.dc.b	$FA,$1B,$67,$FA,$43,$63,$FA,$7B,$70,$34,$93,$FC,$D6,$FC,$6D,$3C
	.dc.b	$67,$F3,$9F,$1B,$F9,$F9,$FD,$03,$C6,$B1,$92,$1F,$CD,$5C,$BF,$CD
	.dc.b	$E2,$E3,$FD,$1A,$6E,$3F,$D7,$26,$D7,$F6,$19,$B3,$FD,$A2,$6E,$3F
	.dc.b	$E8,$CA,$E8,$78,$2F,$FA,$8C,$F9,$63,$15,$FD,$A3,$C6,$FE,$AE,$7F
	.dc.b	$58,$F1,$AC,$FF,$55,$8A,$FF,$4D,$8B,$7F,$D7,$26,$D7,$F7,$29,$B5
	.dc.b	$FD,$DE,$7F,$8D,$91,$FE,$50,$D7,$F8,$EB,$5F,$E4,$8D,$7C,$C7,$02
	.dc.b	$1C,$0F,$ED,$5E,$37,$82,$BF,$C1,$6D,$C7,$FD,$0F,$3E,$B3,$65,$3F
	.dc.b	$B4,$46,$CF,$EF,$3E,$37,$F7,$AF,$16,$CF,$F7,$76,$BF,$DB,$DB,$97
	.dc.b	$F7,$E6,$E1,$FE,$14,$6C,$FF,$86,$5B,$1F,$E2,$06,$E0,$69,$27,$FC
	.dc.b	$0D,$7E,$36,$9E,$33,$FC,$15,$E3,$7F,$84,$0F,$F0,$8F,$8D,$63,$24
	.dc.b	$3F,$C0,$CE,$5F,$F0,$44,$5C,$7F,$C2,$F3,$71,$FF,$16,$CD,$AF,$F1
	.dc.b	$94,$D9,$FF,$1C,$CD,$C7,$FD,$21,$5D,$0F,$05,$FF,$88,$A7,$CB,$18
	.dc.b	$AF,$F1,$CF,$8D,$FE,$28,$3F,$C5,$3E,$35,$9F,$F1,$34,$57,$FC,$3D
	.dc.b	$16,$FF,$C5,$B3,$6B,$FC,$87,$36,$BF,$C9,$13,$FD,$B0,$8F,$F7,$96
	.dc.b	$BF,$DB,$9A,$FF,$75,$6B,$E6,$38,$10,$E0,$7F,$8E,$FC,$6F,$05,$7F
	.dc.b	$97,$1B,$8F,$FA,$2E,$7D,$66,$CA,$7F,$8E,$63,$67,$F9,$2B,$C6,$FF
	.dc.b	$25,$F8,$B6,$7F,$C9,$0D,$7F,$C8,$0D,$CB,$FC,$9E,$DC,$3F,$CC,$AD
	.dc.b	$9F,$F3,$53,$63,$FC,$E2,$DC,$0D,$24,$FF,$95,$EF,$C6,$D3,$C6,$7F
	.dc.b	$96,$FC,$6F,$F3,$09,$FE,$62,$F1,$6C,$64,$87,$F9,$5D,$CB,$FE,$59
	.dc.b	$8B,$8F,$F9,$A2,$6E,$3F,$E7,$E9,$B5,$FE,$83,$9B,$3F,$E8,$A9,$B8
	.dc.b	$FF,$A5,$2B,$A1,$E0,$BF,$F3,$9C,$F9,$63,$15,$FE,$93,$FD,$AF,$FC
	.dc.b	$ED,$1F,$E9,$66,$C1,$78,$BF,$E8,$47,$24,$F1,$B1,$19,$65,$7F,$A1
	.dc.b	$A7,$73,$88,$FF,$3D,$47,$FA,$61,$FF,$29,$35,$F2,$4F,$F9,$2E,$31
	.dc.b	$1E,$58,$C1,$1F,$F2,$B4,$6A,$BD,$13,$99,$FF,$29,$4E,$49,$1E,$BF
	.dc.b	$1B,$C9,$8F,$FA,$9A,$51,$8B,$F8,$CD,$BF,$8C,$D8,$F8,$CD,$9E,$33
	.dc.b	$F0,$7C,$6F,$90,$7C,$9F,$16,$C6,$48,$7C,$47,$2F,$C6,$8B,$87,$96
	.dc.b	$6E,$1F,$66,$6D,$7D,$99,$B3,$F6,$66,$DF,$F9,$B6,$FD,$D1,$83,$C9
	.dc.b	$7E,$28,$F8,$5E,$37,$DF,$F1,$BF,$00,$D9,$FB,$8D,$7E,$DB,$67,$8C
	.dc.b	$7E,$06,$78,$B4,$E2,$53,$89,$63,$8B,$F8,$68,$E2,$D3,$89,$4E,$25
	.dc.b	$8F,$B5,$E3,$7E,$5C,$FC,$C7,$8D,$67,$F2,$91,$5F,$C9,$C5,$9E,$31
	.dc.b	$A3,$F3,$51,$9C,$BF,$51,$5F,$95,$24,$FE,$49,$AF,$E4,$1A,$FE,$79
	.dc.b	$AF,$23,$81,$0E,$07,$EA,$7C,$6F,$03,$FB,$56,$CF,$ED,$5B,$99,$A4
	.dc.b	$9F,$D9,$78,$DF,$B9,$3F,$75,$E2,$D8,$C9,$0F,$DA,$B9,$7F,$6F,$17
	.dc.b	$2F,$D3,$4D,$C3,$F8,$54,$DA,$FE,$15,$36,$7F,$85,$4D,$C1,$36,$30
	.dc.b	$E7,$C6,$FE,$1D,$E3,$7F,$0F,$6E,$1F,$C2,$DA,$FF,$0A,$6E,$3F,$B9
	.dc.b	$6E,$3F,$C7,$5B,$3F,$C7,$5B,$1F,$C7,$5B,$89,$A4,$9F,$E3,$57,$E3
	.dc.b	$69,$FC,$77,$C5,$FE,$48,$7F,$25,$F1,$AC,$64,$87,$F2,$07,$2F,$F2
	.dc.b	$28,$B8,$FF,$11,$9B,$8F,$F3,$59,$FE,$44,$4A,$FE,$3E,$D7,$F8,$EB
	.dc.b	$5F,$E5,$6D,$79,$1C,$08,$70,$3F,$9B,$78,$BE,$07,$FA,$13,$67,$FA
	.dc.b	$13,$72,$34,$93,$FC,$FF,$C6,$FE,$8D,$E2,$FF,$47,$6C,$64,$FE,$89
	.dc.b	$19,$69,$93,$FA,$2B,$96,$99,$CB,$F9,$B4,$65,$B1,$9C,$3F,$AE,$4E
	.dc.b	$5B,$19,$3F,$AF,$46,$5A,$7F,$57,$F1,$BF,$AF,$1F,$D7,$FC,$6B,$19
	.dc.b	$21,$FD,$69,$CB,$FD,$72,$2D,$FF,$4B,$9B,$7F,$DB,$E6,$D7,$F6,$F9
	.dc.b	$B3,$FD,$BE,$6E,$5F,$E7,$AA,$E8,$78,$2F,$FA,$B4,$F9,$63,$15,$FD
	.dc.b	$EF,$C5,$FE,$EC,$7F,$77,$F1,$AC,$FF,$73,$8A,$FF,$72,$8B,$7F,$D8
	.dc.b	$66,$D7,$F8,$16,$6D,$7F,$81,$67,$FB,$A1,$1F,$EE,$2D,$7F,$B7,$B5
	.dc.b	$FE,$FA,$D7,$CC,$70,$21,$C0,$FF,$04,$78,$BE,$0A,$FF,$09,$37,$13
	.dc.b	$49,$3E,$D3,$89,$FE,$11,$F1,$BF,$C3,$07,$F8,$67,$C5,$B1,$92,$1F
	.dc.b	$E1,$47,$2F,$F8,$5A,$2E,$5F,$E0,$59,$B8,$7F,$8A,$66,$D7,$F8,$A6
	.dc.b	$6C,$FF,$8A,$66,$E0,$9B,$18,$73,$E3,$7F,$87,$C9,$7F,$E1,$C6,$BF
	.dc.b	$E1,$A6,$BF,$E2,$86,$BF,$E3,$B6,$BF,$E3,$B7,$FC,$55,$E3,$7F,$8E
	.dc.b	$4F,$F1,$D7,$8B,$67,$FC,$6D,$15,$FF,$1A,$C5,$CB,$FC,$49,$37,$2F
	.dc.b	$F2,$44,$DA,$FF,$24,$4D,$9F,$F2,$44,$DC,$BF,$D0,$55,$FE,$40,$AF
	.dc.b	$0B,$E0,$21,$93,$FC,$91,$3C,$7C,$6F,$F2,$11,$23,$FC,$7E,$D7,$FC
	.dc.b	$7A,$D7,$FC,$96,$D7,$FC,$BC,$D7,$FC,$BC,$FF,$93,$7C,$49,$1F,$E4
	.dc.b	$A6,$BF,$E4,$86,$BF,$E5,$66,$BF,$E6,$96,$BF,$E6,$96,$E1,$FE,$69
	.dc.b	$F1,$BF,$CB,$44,$9F,$F2,$BB,$5F,$F2,$AB,$5F,$F3,$1B,$5F,$F3,$B3
	.dc.b	$5F,$F3,$B2,$61,$21,$FB,$67,$F9,$8C,$9E,$2E,$A7,$51,$87,$02,$78
	.dc.b	$CE,$23,$13,$B7,$94,$E9,$E2,$D7,$31,$B7,$BF,$8D,$CE,$7C,$AD,$9F
	.dc.b	$F4,$6C,$F2,$6B,$CA,$34,$47,$4D,$74,$60,$87,$8B,$C6,$EC,$47,$B3
	.dc.b	$5D,$18,$21,$82,$1F,$21,$E0,$FC,$AF,$17,$D4,$47,$D4,$D7,$E1,$B5
	.dc.b	$E0,$68,$86,$8A,$7D,$6F,$1B,$E3,$11,$F8,$CD,$7E,$7B,$5E,$06,$88
	.dc.b	$68,$A7,$E0,$BC,$6F,$A4,$47,$E9,$35,$FB,$6D,$78,$1A,$21,$A2,$9F
	.dc.b	$8E,$F1,$BE,$F1,$1F,$BC,$D7,$F0,$ED,$78,$1A,$21,$A2,$9F,$9A,$F1
	.dc.b	$7F,$14,$47,$F1,$4D,$7F,$26,$D7,$81,$A2,$1A,$29,$FA,$6F,$1B,$F2
	.dc.b	$C4,$7F,$2C,$D7,$F3,$ED,$78,$1A,$21,$A2,$9F,$B2,$F1,$BF,$44,$47
	.dc.b	$F4,$4D,$7F,$56,$D7,$81,$A2,$1A,$29,$FB,$EF,$1B,$F5,$C4,$7F,$5C
	.dc.b	$D7,$F6,$ED,$78,$1A,$21,$A2,$9F,$C2,$BC,$6F,$DD,$11,$FD,$D3,$5F
	.dc.b	$E0,$6D,$78,$1A,$21,$A2,$9F,$C5,$BC,$6F,$E0,$C4,$7F,$83,$35,$FE
	.dc.b	$1E,$D7,$81,$A2,$1A,$29,$FC,$8B,$C6,$FE,$24,$47,$F8,$93,$5F,$E3
	.dc.b	$6D,$78,$1A,$21,$A2,$9F,$CB,$BC,$5F,$E3,$C4,$7F,$8F,$35,$FE,$4E
	.dc.b	$D7,$81,$A2,$1A,$29,$FC,$EB,$C6,$FE,$54,$47,$F9,$53,$5F,$E6,$6D
	.dc.b	$78,$1A,$21,$A2,$9F,$D1,$BC,$6F,$E6,$C4,$7F,$9B,$35,$FE,$7E,$D7
	.dc.b	$81,$A2,$1A,$29,$FD,$4B,$C6,$FE,$84,$47,$FA,$13,$5F,$E9,$6D,$78
	.dc.b	$1A,$21,$A2,$9F,$D7,$BC,$6F,$E9,$C4,$7F,$A7,$35,$FE,$AE,$D7,$81
	.dc.b	$A2,$1A,$29,$FD,$AB,$C5,$FE,$B4,$47,$FA,$D3,$5F,$EC,$6D,$78,$1A
	.dc.b	$21,$A2,$9F,$DD,$BC,$6F,$EC,$C4,$7F,$B3,$35,$FE,$DE,$D7,$81,$A2
	.dc.b	$1A,$29,$FE,$02,$F1,$7F,$B9,$11,$FE,$E4,$D7,$FB,$DB,$5E,$06,$88
	.dc.b	$68,$A7,$F8,$3B,$C6,$FE,$FC,$47,$FB,$F3,$5F,$F0,$3B,$5E,$06,$88
	.dc.b	$68,$A7,$F8,$6B,$C6,$FF,$05,$11,$FF,$05,$35,$FF,$09,$B5,$E0,$68
	.dc.b	$86,$8A,$7F,$89,$BC,$6F,$F0,$B1,$1F,$F0,$B3,$5F,$F0,$FB,$5E,$06
	.dc.b	$88,$68,$A7,$F8,$CB,$C6,$FF,$11,$11,$FF,$11,$35,$FF,$15,$B5,$E0
	.dc.b	$68,$86,$8A,$7F,$8F,$BC,$6F,$F1,$71,$1F,$F1,$73,$5F,$F1,$BB,$5E
	.dc.b	$06,$88,$68,$A7,$F9,$2B,$C4,$FF,$1D,$11,$FF,$1D,$35,$FF,$21,$B5
	.dc.b	$E0,$68,$86,$8A,$7F,$95,$BC,$4F,$F2,$31,$1F,$F2,$33,$5F,$F2,$7B
	.dc.b	$5E,$06,$88,$68,$A7,$F9,$8B,$C4,$FF,$29,$11,$FF,$29,$35,$FF,$2D
	.dc.b	$B5,$E0,$68,$86,$8A,$7F,$9B,$BC,$4F,$F2,$F1,$1F,$F2,$F3,$5F,$F3
	.dc.b	$3B,$5E,$06,$88,$68,$A7,$F9,$EB,$C4,$FF,$35,$11,$FF,$35,$35,$FF
	.dc.b	$39,$B5,$E0,$68,$86,$8A,$7F,$A2,$FF,$6D,$78,$BF,$E7,$96,$C7,$FA
	.dc.b	$01,$AF,$FA,$11,$AF,$18,$E0,$47,$83,$5E,$0D,$78,$1D,$7C,$3F,$01
	.dc.b	$1C,$10,$F0,$39,$69,$93,$D4,$E5,$A7,$A7,$C3,$F6,$11,$C1,$0F,$63
	.dc.b	$96,$99,$3E,$33,$96,$9F,$17,$C3,$F9,$04,$70,$43,$E4,$39,$69,$93
	.dc.b	$E9,$39,$69,$F4,$7C,$3F,$A8,$47,$04,$3E,$A3,$96,$99,$3E,$F3,$96
	.dc.b	$9F,$77,$C3,$FC,$01,$1C,$10,$FC,$03,$96,$99,$3F,$14,$E5,$A7,$E2
	.dc.b	$7C,$3F,$C6,$11,$C1,$0F,$C6,$39,$69,$93,$F2,$CE,$5A,$7E,$57,$C3
	.dc.b	$FC,$C1,$1C,$10,$FC,$C3,$96,$99,$3F,$44,$E5,$A7,$E8,$7C,$3F,$D2
	.dc.b	$11,$C1,$0F,$D2,$39,$69,$93,$F5,$CE,$5A,$7E,$B7,$C3,$FD,$81,$1C
	.dc.b	$10,$FD,$83,$96,$99,$3F,$74,$E5,$A7,$EE,$7C,$3F,$DE,$11,$C1,$0F
	.dc.b	$DE,$39,$69,$93,$F8,$33,$96,$9F,$C1,$7C,$3F,$E1,$04,$70,$43,$F8
	.dc.b	$43,$96,$99,$3F,$89,$39,$69,$FC,$47,$C3,$FE,$28,$47,$04,$3F,$8A
	.dc.b	$39,$69,$93,$F8,$F3,$96,$9F,$C7,$7C,$3F,$E4,$04,$70,$43,$F9,$03
	.dc.b	$96,$99,$3F,$95,$39,$69,$FC,$A7,$C3,$FE,$58,$47,$04,$3F,$96,$39
	.dc.b	$69,$93,$F9,$B3,$96,$9F,$CD,$7C,$3F,$E7,$04,$70,$43,$F9,$C3,$96
	.dc.b	$99,$3F,$A1,$39,$69,$FD,$07,$C3,$FE,$88,$47,$04,$3F,$A2,$39,$69
	.dc.b	$93,$FA,$73,$96,$9F,$D3,$7C,$3F,$EA,$04,$70,$43,$FA,$83,$96,$99
	.dc.b	$3F,$AD,$39,$69,$FD,$67,$C3,$FE,$B8,$47,$04,$3F,$AE,$39,$69,$93
	.dc.b	$FB,$33,$96,$9F,$D9,$7C,$3F,$ED,$04,$70,$43,$FB,$43,$96,$99,$3F
	.dc.b	$B9,$39,$69,$FD,$C7,$C3,$FE,$E8,$47,$04,$3F,$BA,$39,$69,$93,$FB
	.dc.b	$F3,$96,$9F,$DF,$7C,$3F,$F0,$01,$1C,$10,$FF,$00,$39,$69,$93,$FC
	.dc.b	$14,$E5,$A7,$F8,$27,$C3,$FF,$06,$11,$C1,$0F,$F0,$63,$96,$99,$3F
	.dc.b	$C2,$CE,$5A,$7F,$85,$7C,$3F,$F0,$C1,$1C,$10,$FF,$0C,$39,$69,$93
	.dc.b	$FC,$44,$E5,$A7,$F8,$87,$C3,$FF,$12,$11,$C1,$0F,$F1,$23,$96,$99
	.dc.b	$3F,$C5,$CE,$5A,$7F,$8B,$7C,$3F,$F1,$81,$1C,$10,$FF,$18,$39,$69
	.dc.b	$93,$FC,$74,$E5,$A7,$F8,$E7,$C3,$FF,$1E,$11,$C1,$0F,$F1,$E3,$96
	.dc.b	$99,$3F,$C8,$CE,$5A,$7F,$91,$7C,$3F,$F2,$41,$1C,$10,$FF,$24,$39
	.dc.b	$69,$93,$FC,$A4,$E5,$A7,$F9,$47,$C3,$FF,$2A,$11,$C1,$0F,$F2,$A3
	.dc.b	$96,$99,$3F,$CB,$CE,$5A,$7F,$97,$7C,$3F,$F3,$01,$1C,$10,$FF,$30
	.dc.b	$39,$69,$93,$FC,$D4,$E5,$A7,$F9,$A7,$C3,$FF,$36,$11,$C1,$0F,$F3
	.dc.b	$63,$96,$99,$3F,$CE,$CE,$5A,$71,$F1,$DE,$32,$2F,$88,$F1,$71,$0A
	.dc.b	$DB,$AB,$CC,$62,$D7,$C6,$BA,$4E,$61,$5C,$2B,$C0,$70,$BC,$5F,$53
	.dc.b	$8C,$6A,$BA,$47,$8E,$31,$5C,$1E,$51,$C0,$E4,$6C,$F6,$D6,$EB,$9C
	.dc.b	$7F,$A6,$E3,$FD,$34,$47,$FD,$34,$D7,$FD,$34,$D7,$81,$A2,$1A,$3E
	.dc.b	$49,$EE,$C7,$FD,$42,$DF,$17,$2D,$2D,$A7,$71,$F5,$9B,$19,$9E,$33
	.dc.b	$96,$C6,$63,$9C,$F0,$24,$E9,$C7,$8B,$F2,$53,$63,$0C,$9F,$99,$5E
	.dc.b	$58,$D4,$62,$B8,$B6,$38,$3F,$46,$7C,$8D,$9D,$B6,$7C,$84,$7D,$2D
	.dc.b	$7E,$5C,$F8,$5C,$90,$CC,$F4,$7F,$33,$1B,$73,$3C,$DC,$CB,$C5,$F8
	.dc.b	$84,$8F,$92,$D7,$AB,$5E,$2D,$7E,$F3,$5F,$B0,$FC,$FF,$16,$47,$D4
	.dc.b	$6B,$E8,$6B,$E4,$6B,$F8,$96,$BF,$83,$6D,$FE,$1F,$C5,$FB,$84,$9F
	.dc.b	$C0,$B5,$F8,$AD,$7E,$0B,$5F,$CB,$35,$FC,$83,$F8,$7F,$16,$47,$E3
	.dc.b	$1A,$FD,$06,$BF,$31,$AF,$E8,$5A,$FE,$6D,$B3,$F9,$FF,$17,$F2,$84
	.dc.b	$9F,$CC,$B5,$FB,$AD,$7E,$CB,$5F,$D7,$35,$FD,$43,$F9,$FF,$16,$47
	.dc.b	$E9,$1A,$FE,$21,$AF,$E1,$1A,$FE,$E5,$AF,$EC,$DB,$3F,$B7,$7F,$55
	.dc.b	$E2,$FE,$AD,$B3,$FA,$56,$BF,$A1,$6F,$0F,$F2,$9D,$11,$8B,$F0,$57
	.dc.b	$EB,$6F,$8B,$F9,$87,$F7,$44,$7F,$7E,$D7,$F4,$6D,$78,$1A,$21,$A3
	.dc.b	$F8,$3F,$8B,$FC,$14,$8E,$08,$7E,$B9,$CB,$4C,$9F,$AC,$72,$D3,$F8
	.dc.b	$9F,$8B,$FC,$44,$8E,$08,$7E,$E9,$CB,$4C,$9F,$B8,$72,$D3,$F8,$FF
	.dc.b	$8B,$FC,$74,$8E,$08,$7F,$06,$72,$D3,$27,$F0,$47,$2D,$3F,$95,$F8
	.dc.b	$BF,$CA,$48,$E0,$87,$F1,$27,$2D,$32,$7F,$10,$72,$D3,$F9,$BF,$8B
	.dc.b	$FC,$D4,$8E,$08,$7F,$1E,$72,$D3,$27,$F1,$C7,$2D,$38,$FE,$08,$F8
	.dc.b	$0F,$EA,$9E,$3F,$F7,$E8,$C4,$FF,$4F,$AC,$56,$A3,$FC,$0D,$3A,$D3
	.dc.b	$FA,$B5,$6A,$7F,$AB,$5D,$9C,$C7,$0B,$E4,$AF,$96,$CC,$38,$2C,$2C
	.dc.b	$61,$95,$CC,$EE,$5C,$3B,$BD,$70,$EF,$87,$6A,$E3,$39,$3D,$C7,$1C
	.dc.b	$3C,$0D,$8D,$CF,$9B,$0B,$7E,$AC,$3F,$A4,$D7,$F4,$EB,$C1,$0F,$29
	.dc.b	$FE,$28,$2C,$70,$7F,$C4,$26,$DF,$F0,$04,$73,$8D,$99,$C3,$CA,$F3
	.dc.b	$9F,$F0,$37,$8F,$B6,$C7,$43,$FC,$11,$19,$93,$49,$1B,$9E,$2F,$59
	.dc.b	$DB,$96,$B9,$9E,$2D,$35,$E3,$F7,$23,$82,$1D,$5C,$B4,$CD,$7F,$74
	.dc.b	$8C,$B6,$3C,$5E,$3F,$AC,$8E,$08,$7A,$5C,$B4,$CC,$7F,$80,$23,$2D
	.dc.b	$8F,$77,$8F,$F1,$C8,$E0,$87,$C5,$72,$D3,$31,$FE,$0C,$8C,$B6,$3E
	.dc.b	$57,$8F,$F4,$C8,$E0,$87,$D1,$72,$D3,$31,$FE,$18,$8C,$B6,$3E,$B7
	.dc.b	$8F,$F7,$C8,$E0,$87,$DD,$72,$D3,$31,$FE,$24,$8C,$B6,$38,$ED,$FB
	.dc.b	$8F,$1F,$F6,$F5,$FB,$86,$CF,$EA,$DA,$FE,$BE,$BF,$CB,$91,$8A,$E5
	.dc.b	$3F,$B1,$AE,$4F,$12,$1C,$74,$46,$A3,$15,$6B,$F8,$5C,$DD,$BF,$CE
	.dc.b	$13,$6B,$FC,$D5,$37,$0F,$1C,$76,$BE,$2D,$BE,$2D,$38,$96,$78,$BE
	.dc.b	$23,$D4,$67,$06,$1D,$DF,$4C,$66,$31,$86,$5B,$19,$39,$46,$5A,$64
	.dc.b	$F0,$1F,$15,$D3,$8C,$3D,$6D,$C3,$D6,$D3,$D6,$5C,$3D,$73,$F5,$23
	.dc.b	$A1,$DD,$C1,$D0,$B3,$D1,$B1,$D1,$B3,$D2,$74,$7F,$A8,$EB,$CF,$3A
	.dc.b	$C3,$E2,$DF,$8D,$F9,$2F,$C0,$31,$87,$C6,$69,$99,$E5,$19,$6C,$66
	.dc.b	$FC,$2F,$FA,$AE,$75,$18,$C3,$CD,$39,$3E,$D9,$D8,$F2,$61,$E8,$9E
	.dc.b	$18,$7A,$30,$EF,$5A,$3E,$44,$EA,$B8,$47,$53,$E6,$9D,$67,$E1,$46
	.dc.b	$AF,$C6,$7F,$B0,$DC,$5F,$C4,$AE,$35,$E0,$38,$BE,$03,$B3,$F5,$EF
	.dc.b	$CF,$87,$FB,$04,$93,$FE,$C6,$6B,$FE,$AA,$6B,$FE,$A6,$6B,$F8,$A3
	.dc.b	$07,$13,$FD,$9F,$E3,$FF,$B3,$48,$E0,$87,$FA,$D9,$CB,$4C,$9F,$EB
	.dc.b	$47,$2D,$38,$D7,$F8,$22,$3F,$C0,$E7,$F8,$23,$C7,$B3,$FD,$EE,$2B
	.dc.b	$FC,$26,$2C,$F1,$8D,$1F,$E0,$18,$CD,$FF,$BD,$EA,$CF,$FB,$C6,$6E
	.dc.b	$5F,$EF,$59,$B8,$A6,$C6,$1B,$5D,$8F,$11,$70,$F1,$3F,$C7,$67,$D4
	.dc.b	$E7,$0F,$13,$4F,$11,$70,$F1,$37,$0F,$13,$6B,$C4,$D3,$58,$79,$1F
	.dc.b	$01,$F0,$4D,$18,$3C,$E5,$C3,$CE,$4C,$11,$E4,$34,$FA,$1A,$66,$38
	.dc.b	$2C,$3D,$2E,$96,$1E,$A8,$C2,$C3,$D7,$18,$6D,$74,$AE,$6D,$AE,$4D
	.dc.b	$9D,$DF,$96,$31,$7F,$04,$E8,$7C,$63,$E1,$BC,$F0,$F9,$0D,$8C,$CF
	.dc.b	$33,$E1,$E1,$EA,$AD,$C6,$67,$E1,$36,$7C,$ED,$C3,$CE,$DC,$3C,$F5
	.dc.b	$F1,$E7,$E1,$3E,$03,$99,$98,$EE,$E2,$BA,$B9,$3D,$87,$B8,$F9,$B7
	.dc.b	$C2,$36,$7A,$63,$B3,$8A,$F2,$C7,$33,$96,$1F,$E7,$AA,$FC,$CC,$63
	.dc.b	$0F,$87,$5C,$2F,$A4,$E4,$F3,$46,$4F,$6B,$E4,$3E,$A4,$7E,$39,$FC
	.dc.b	$73,$4E,$65,$8F,$C4,$36,$BF,$CC,$04,$9F,$F2,$D3,$5F,$F2,$F3,$63
	.dc.b	$84,$7F,$97,$FC,$7F,$F3,$41,$1C,$10,$FF,$33,$39,$69,$98,$F2,$46
	.dc.b	$5B,$1F,$E6,$BF,$1F,$FC,$E8,$47,$04,$3F,$CE,$4E,$5A,$66,$3E,$04
	.dc.b	$65,$B1,$FE,$77,$F1,$FF,$D0,$04,$70,$43,$FC,$FC,$E5,$A6,$63,$E6
	.dc.b	$46,$5B,$1F,$E8,$3F,$1F,$FD,$18,$47,$04,$3F,$D1,$4E,$5A,$66,$3E
	.dc.b	$C4,$65,$B1,$FE,$8F,$C7,$FD,$30,$47,$04,$3F,$D2,$CE,$5A,$66,$3F
	.dc.b	$09,$19,$6C,$7F,$A6,$FC,$7F,$F5,$21,$1C,$10,$FF,$51,$39,$69,$98
	.dc.b	$FC,$84,$65,$B1,$C7,$4F,$F6,$EC,$FF,$B7,$0F,$F6,$EF,$8F,$67,$FD
	.dc.b	$A1,$15,$FF,$2F,$45,$9E,$31,$A3,$FD,$AD,$1C,$9C,$1F,$E6,$59,$C3
	.dc.b	$C8,$CC,$F2,$70,$47,$04,$D7,$F6,$F5,$CA,$7F,$58,$FE,$C5,$A5,$3F
	.dc.b	$8A,$35,$A6,$6F,$F9,$DC,$65,$B1,$9C,$13,$96,$C6,$63,$04,$78,$78
	.dc.b	$DB,$23,$82,$1E,$7A,$FE,$97,$1A,$8C,$52,$31,$1F,$C2,$5A,$FF,$18
	.dc.b	$7D,$1F,$B7,$3C,$7A,$49,$F7,$46,$5B,$19,$7C,$51,$96,$C6,$5F,$E3
	.dc.b	$51,$96,$C6,$5F,$E9,$70,$F1,$FC,$64,$8F,$0B,$5E,$ED,$7A,$B5,$E6
	.dc.b	$D7,$FB,$94,$75,$7D,$7E,$3F,$C0,$23,$82,$1E,$C7,$2D,$32,$7B,$1C
	.dc.b	$B4,$F8,$FE,$3F,$CC,$23,$82,$1F,$21,$CB,$4C,$9F,$21,$CB,$4F,$A7
	.dc.b	$E3,$FF,$47,$F1,$7F,$C1,$D1,$EE,$9C,$69,$F3,$1B,$1F,$0A,$AC,$B0
	.dc.b	$46,$15,$53,$11,$2B,$FC,$47,$7E,$0B,$E5,$5B,$9C,$C7,$DC,$AE,$11
	.dc.b	$F8,$67,$F1,$E4,$7D,$64,$34,$53,$F1,$2E,$DC,$90,$FC,$6F,$8F,$F9
	.dc.b	$A2,$3F,$98,$6B,$F9,$66,$BC,$0D,$10,$D1,$4F,$CC,$F8,$FF,$A6,$23
	.dc.b	$FA,$46,$BF,$A2,$6B,$C0,$D1,$0D,$14,$FD,$2F,$8F,$FB,$22,$3F,$B0
	.dc.b	$6B,$FA,$E6,$BC,$0D,$10,$D1,$4F,$D8,$F8,$FF,$BE,$23,$FB,$C6,$BF
	.dc.b	$BA,$6B,$C0,$D1,$0D,$14,$FD,$EF,$8F,$FC,$28,$8F,$F0,$86,$BF,$C1
	.dc.b	$9A,$F0,$34,$43,$45,$3F,$84,$F8,$FF,$C5,$88,$FF,$14,$6B,$FC,$49
	.dc.b	$AF,$03,$44,$34,$53,$43,$4F,$CE,$6A,$9F,$D2,$1F,$A5,$F1,$ED,$7E
	.dc.b	$78,$FE,$1B,$18,$79,$6B,$FC,$72,$71,$7D,$2A,$CE,$6B,$93,$FB,$3A
	.dc.b	$FE,$66,$48,$FE,$62,$D7,$F9,$7B,$5F,$E5,$AD,$7A,$1A,$38,$9F,$CE
	.dc.b	$BC,$7F,$E7,$E4,$70,$43,$F9,$DB,$96,$99,$3F,$9D,$B9,$69,$FD,$1B
	.dc.b	$C7,$FE,$15,$12,$3D,$B8,$A7,$F8,$71,$98,$F8,$E7,$F3,$69,$CB,$C2
	.dc.b	$35,$8F,$D4,$9C,$CF,$0C,$3F,$8D,$57,$F5,$A2,$47,$8C,$87,$F5,$67
	.dc.b	$2D,$32,$7F,$56,$72,$D3,$FB,$0F,$8F,$FD,$98,$8E,$08,$7F,$62,$72
	.dc.b	$D3,$27,$F6,$27,$2D,$3F,$B6,$F8,$FF,$DC,$88,$E0,$87,$F6,$E7,$2D
	.dc.b	$32,$7F,$6E,$72,$D3,$FB,$CF,$8F,$FD,$F8,$8E,$08,$7F,$7A,$72,$D3
	.dc.b	$27,$F7,$A7,$2D,$3F,$C0,$BE,$3F,$F8,$28,$8E,$08,$7F,$81,$9C,$B4
	.dc.b	$C9,$FE,$06,$72,$D3,$FC,$23,$E3,$FF,$85,$88,$E0,$87,$F8,$49,$CB
	.dc.b	$4C,$9F,$E1,$27,$2D,$3F,$C3,$BE,$3F,$F8,$88,$8E,$08,$7F,$87,$9C
	.dc.b	$B4,$C9,$FE,$1E,$72,$D3,$FC,$53,$E3,$FF,$8B,$88,$E0,$87,$F8,$A9
	.dc.b	$CB,$4C,$9F,$E2,$A7,$2D,$3F,$C6,$BE,$3F,$F8,$E8,$8E,$08,$7F,$8D
	.dc.b	$9C,$B4,$C9,$FE,$36,$72,$D3,$FC,$83,$E3,$FF,$91,$88,$E0,$87,$F9
	.dc.b	$09,$CB,$4C,$9F,$E4,$27,$2D,$3F,$C9,$BE,$3F,$F9,$48,$8E,$08,$7F
	.dc.b	$93,$9C,$B4,$C9,$FE,$4E,$72,$D3,$FC,$B3,$E3,$FF,$97,$88,$E0,$87
	.dc.b	$F9,$69,$CB,$4C,$9F,$E5,$A7,$2D,$3F,$CC,$BE,$3F,$F9,$A8,$8E,$08
	.dc.b	$7F,$99,$9C,$B4,$C9,$FE,$66,$72,$D3,$FC,$E3,$E3,$FF,$9D,$88,$E0
	.dc.b	$87,$F9,$C9,$CB,$4C,$9F,$E7,$27,$2D,$3F,$CF,$BE,$3F,$FA,$08,$8E
	.dc.b	$08,$7F,$9F,$9C,$B4,$C9,$FE,$7E,$72,$D3,$FD,$13,$E1,$FF,$A3,$88
	.dc.b	$E0,$87,$FA,$29,$CB,$4C,$9F,$E8,$A7,$2D,$3F,$D2,$BE,$3F,$FA,$68
	.dc.b	$8E,$08,$7F,$A5,$9C,$B4,$C9,$FE,$96,$72,$D3,$FD,$43,$E2,$FF,$A9
	.dc.b	$88,$E0,$87,$FA,$89,$CB,$4C,$9F,$EA,$27,$2D,$3F,$D5,$BE,$2F,$FA
	.dc.b	$C8,$8E,$08,$7F,$AB,$9C,$B4,$C9,$FE,$AE,$72,$D3,$FD,$AD,$FB,$7F
	.dc.b	$FA,$C5,$0E,$74,$66,$30,$D8,$CF,$8D,$FE,$DD,$8C,$11,$C1,$27,$BB
	.dc.b	$5F,$F6,$B4,$70,$F1,$B9,$12,$3C,$AD,$76,$D7,$CE,$79,$1C,$3E,$AA
	.dc.b	$F4,$C6,$23,$47,$C0,$F1,$BD,$04,$70,$43,$CA,$E5,$A6,$4F,$0C,$FC
	.dc.b	$68,$D4,$62,$91,$F1,$7C,$69,$3F,$41,$AF,$C2,$6B,$C5,$AF,$AE,$7F
	.dc.b	$D6,$2F,$12,$1C,$4F,$B7,$E3,$7D,$32,$38,$21,$F4,$1C,$B4,$C9,$E9
	.dc.b	$72,$D3,$F0,$FE,$37,$DF,$23,$82,$1F,$71,$CB,$4C,$9F,$15,$CB,$4F
	.dc.b	$C9,$F8,$DF,$8B,$23,$82,$1F,$88,$72,$D3,$27,$D1,$72,$D3,$F3,$FE
	.dc.b	$37,$E5,$C8,$E0,$87,$E5,$1C,$B4,$C9,$F7,$5C,$B4,$FD,$5F,$8D,$FA
	.dc.b	$32,$38,$21,$FA,$07,$2D,$32,$7E,$25,$CB,$4F,$DB,$F8,$DF,$AF,$23
	.dc.b	$82,$1F,$AC,$72,$D3,$27,$E5,$5C,$B4,$FE,$07,$E3,$7E,$EC,$8E,$08
	.dc.b	$7E,$E1,$CB,$4C,$9F,$A1,$72,$D3,$F8,$7F,$8D,$FC,$1C,$8E,$08,$7F
	.dc.b	$04,$72,$D3,$27,$EB,$5C,$B4,$FE,$37,$E2,$7F,$13,$23,$82,$1F,$C4
	.dc.b	$1C,$B4,$C9,$FB,$97,$2D,$3F,$93,$F8,$9F,$C7,$C8,$E0,$87,$F1,$C7
	.dc.b	$2D,$32,$7F,$05,$72,$D3,$F9,$9F,$89,$FC,$AC,$8E,$08,$7F,$28,$72
	.dc.b	$D3,$27,$F1,$17,$2D,$3F,$9F,$F8,$9F,$CD,$C8,$E0,$87,$F3,$47,$2D
	.dc.b	$32,$7F,$1D,$72,$D3,$FA,$5F,$89,$FD,$0C,$8E,$08,$7F,$40,$72,$D3
	.dc.b	$27,$F2,$97,$2D,$3F,$AB,$F8,$9F,$D3,$C8,$E0,$87,$F4,$C7,$2D,$32
	.dc.b	$7F,$35,$72,$D3,$FB,$1F,$89,$FD,$6C,$8E,$08,$7F,$58,$72,$D3,$27
	.dc.b	$F4,$17,$2D,$3F,$B7,$F8,$9F,$D9,$C8,$E0,$87,$F6,$47,$2D,$32,$7F
	.dc.b	$4D,$72,$D3,$FB,$DF,$89,$FD,$CC,$8E,$08,$7F,$70,$72,$D3,$27,$F5
	.dc.b	$97,$2D,$3F,$C0,$FE,$27,$F7,$F2,$38,$21,$FD,$F1,$CB,$4C,$9F,$D9
	.dc.b	$5C,$B4,$FF,$09,$F8,$9F,$E0,$B2,$38,$21,$FE,$08,$72,$D3,$27,$F7
	.dc.b	$17,$2D,$3F,$C3,$FE,$27,$F8,$5C,$8E,$08,$7F,$85,$1C,$B4,$C9,$FD
	.dc.b	$F5,$CB,$4F,$F1,$5F,$89,$FE,$23,$23,$82,$1F,$E2,$07,$2D,$32,$7F
	.dc.b	$82,$5C,$B4,$FF,$1B,$F8,$9F,$E2,$F2,$38,$21,$FE,$2C,$72,$D3,$27
	.dc.b	$F8,$55,$CB,$4F,$F2,$1F,$89,$FE,$3B,$23,$82,$1F,$E3,$87,$2D,$32
	.dc.b	$7F,$88,$5C,$B4,$FF,$27,$F8,$9F,$E4,$72,$38,$21,$FE,$44,$72,$D3
	.dc.b	$27,$F8,$B5,$CB,$4F,$F2,$DF,$89,$FE,$53,$23,$82,$1F,$E5,$07,$2D
	.dc.b	$32,$7F,$8E,$5C,$B4,$FF,$33,$F8,$9F,$E5,$F2,$38,$21,$FE,$5C,$72
	.dc.b	$D3,$27,$F9,$15,$CB,$4F,$F3,$9F,$89,$FE,$6B,$23,$82,$1F,$E6,$87
	.dc.b	$2D,$32,$7F,$94,$5C,$B4,$FF,$3F,$F8,$9F,$E7,$72,$38,$21,$FE,$74
	.dc.b	$72,$D3,$27,$F9,$75,$CB,$4F,$F3,$99,$FE,$71,$FD,$C3,$FD,$0C,$4F
	.dc.b	$17,$31,$96,$BF,$E8,$F8,$D3,$AF,$11,$18,$8E,$B1,$89,$E2,$D8,$E1
	.dc.b	$3B,$7F,$CE,$AB,$C5,$DB,$63,$71,$82,$38,$21,$E0,$8E,$0D,$8D,$B4
	.dc.b	$D9,$E0,$F1,$7B,$91,$F2,$B5,$E8,$D7,$81,$A2,$1A,$29,$EC,$F1,$7D
	.dc.b	$64,$7E,$0B,$5F,$43,$5E,$06,$88,$68,$A7,$C8,$F1,$3E,$39,$1F,$9A
	.dc.b	$D7,$E2,$35,$E0,$68,$86,$8A,$7D,$4F,$13,$E9,$91,$FB,$2D,$7E,$83
	.dc.b	$5E,$06,$88,$68,$A7,$E0,$3C,$4F,$BE,$47,$F0,$AD,$7E,$E3,$5E,$06
	.dc.b	$88,$68,$A7,$E3,$3C,$4F,$C5,$91,$FC,$8B,$5F,$C4,$35,$E0,$68,$86
	.dc.b	$8A,$7E,$63,$C4,$FC,$B9,$1F,$CE,$B5,$FC,$A3,$5E,$06,$88,$68,$A7
	.dc.b	$E9,$3C,$4F,$D1,$91,$FD,$4B,$5F,$D0,$35,$E0,$68,$86,$8A,$7E,$C3
	.dc.b	$C4,$FD,$79,$1F,$DA,$B5,$FD,$63,$5E,$06,$88,$68,$A7,$EF,$3C,$4F
	.dc.b	$DD,$91,$FE,$02,$D7,$F7,$0D,$78,$1A,$21,$A2,$9F,$C2,$3C,$4F,$E0
	.dc.b	$E4,$7F,$86,$B5,$FE,$08,$D7,$81,$A2,$1A,$29,$FC,$53,$C4,$FE,$26
	.dc.b	$47,$F8,$CB,$5F,$E2,$0D,$78,$1A,$21,$A2,$9F,$C8,$3C,$4F,$E3,$E4
	.dc.b	$7F,$92,$B5,$FE,$38,$D7,$81,$A2,$1A,$29,$FC,$B3,$C4,$33,$C6,$72
	.dc.b	$7F,$41,$9C,$CA,$72,$FF,$3F,$8C,$C6,$27,$F9,$FF,$8B,$87,$95,$75
	.dc.b	$6B,$CD,$B3,$C7,$95,$5C,$6B,$B1,$C4,$A7,$12,$CF,$18,$E1,$5F,$D4
	.dc.b	$2E,$D9,$A6,$26,$DF,$5A,$E0,$DA,$E0,$D9,$DB,$E1,$9D,$B6,$76,$D9
	.dc.b	$F0,$F8,$BF,$D6,$09,$1F,$D7,$9A,$FF,$55,$6B,$FD,$81,$AF,$F7,$36
	.dc.b	$BF,$DB,$DF,$EC,$7E,$2F,$B8,$91,$E9,$F1,$6B,$E5,$8B,$97,$CF,$9B
	.dc.b	$8F,$F7,$F9,$B5,$FD,$EE,$6C,$FF,$7A,$9B,$8A,$6C,$61,$B5,$F2,$63
	.dc.b	$E8,$C7,$D2,$F1,$7E,$9F,$8B,$2B,$E7,$46,$5F,$A1,$19,$7E,$64,$65
	.dc.b	$F9,$D1,$9C,$7F,$11,$39,$9F,$F0,$3B,$FE,$06,$23,$FE,$0C,$6B,$FE
	.dc.b	$03,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$F0,$DE,$2F,$96,$33,$1E,$83
	.dc.b	$26,$CC,$11,$FD,$04,$7E,$5B,$C5,$93,$F9,$17,$F3,$3E,$2E,$23,$F1
	.dc.b	$D1,$6B,$F5,$13,$74,$FF,$1C,$4D,$AF,$F1,$A4,$D9,$FF,$19,$CD,$D1
	.dc.b	$36,$30,$DA,$F1,$1F,$A3,$7F,$57,$E2,$FE,$B3,$C5,$93,$FA,$78,$CB
	.dc.b	$FA,$98,$CB,$FA,$58,$CB,$FA,$78,$CE,$3F,$C0,$67,$2F,$F0,$38,$CE
	.dc.b	$3F,$E5,$39,$CB,$FE,$55,$8C,$D7,$F9,$46,$72,$FF,$94,$E3,$33,$FE
	.dc.b	$52,$9F,$E0,$7E,$2F,$F0,$4F,$16,$5F,$EF,$E3,$2F,$F0,$18,$CB,$FB
	.dc.b	$D8,$CB,$FB,$F8,$CE,$1F,$C6,$67,$33,$FE,$58,$7F,$CA,$E4,$7F,$CB
	.dc.b	$AD,$7F,$CA,$8D,$78,$1A,$21,$A2,$C7,$77,$D0,$7F,$17,$F1,$7C,$B1
	.dc.b	$98,$F4,$19,$36,$60,$8F,$F3,$18,$FE,$4F,$E2,$FF,$28,$F1,$65,$7F
	.dc.b	$23,$8C,$BF,$C9,$63,$2F,$F2,$18,$CB,$FC,$8E,$33,$5F,$CF,$67,$2F
	.dc.b	$F3,$F8,$CE,$5F,$E8,$F9,$CB,$FE,$91,$8C,$D7,$FA,$36,$72,$FF,$A3
	.dc.b	$E3,$33,$FE,$8E,$9F,$E8,$3F,$2C,$F1,$A9,$2F,$F9,$F4,$7F,$A0,$B0
	.dc.b	$FE,$B0,$EE,$33,$3B,$7F,$9F,$46,$EF,$FB,$38,$78,$46,$A7,$85,$74
	.dc.b	$F1,$AD,$F0,$BD,$CF,$82,$BC,$F5,$AB,$FE,$E3,$5A,$BE,$15,$A9,$F2
	.dc.b	$90,$F0,$1E,$CA,$F8,$3E,$37,$C2,$F1,$A4,$FC,$08,$CB,$F0,$A3,$2F
	.dc.b	$BA,$32,$FC,$08,$CE,$3F,$1A,$73,$3F,$E0,$4C,$3E,$14,$7F,$AE,$B1
	.dc.b	$34,$93,$FE,$C3,$8C,$B6,$32,$FF,$B1,$23,$2D,$8C,$BD,$E3,$2D,$8F
	.dc.b	$3D,$79,$63,$EE,$F8,$D8,$BF,$A1,$56,$7E,$6C,$D9,$F4,$C7,$C5,$38
	.dc.b	$9F,$86,$F1,$BF,$17,$E3,$48,$C1,$0F,$C2,$B9,$7F,$0F,$1F,$92,$31
	.dc.b	$87,$E3,$A7,$13,$C6,$30,$FE,$52,$32,$F2,$30,$43,$04,$8E,$2F,$88
	.dc.b	$E9,$5F,$8F,$F1,$BF,$46,$48,$FD,$0F,$8D,$5F,$CE,$C5,$AF,$D1,$CD
	.dc.b	$AE,$55,$FA,$8A,$FD,$6C,$E6,$3F,$2F,$16,$7F,$49,$37,$34,$D8,$C3
	.dc.b	$6B,$F6,$7E,$37,$EE,$3C,$69,$18,$21,$FB,$27,$2F,$ED,$23,$F7,$E6
	.dc.b	$31,$FD,$DC,$E2,$78,$9F,$89,$9F,$CB,$11,$FC,$A3,$5F,$C9,$35,$E0
	.dc.b	$68,$86,$8B,$3F,$C2,$A3,$C1,$1F,$C2,$3C,$6B,$39,$8F,$5B,$FC,$42
	.dc.b	$33,$97,$F0,$5A,$B5,$FC,$0A,$7F,$8D,$F8,$D2,$3F,$8C,$BF,$C7,$BC
	.dc.b	$6C,$47,$F1,$68,$B8,$7F,$1C,$9B,$87,$F1,$19,$B5,$FC,$3A,$6C,$FF
	.dc.b	$1B,$9B,$A2,$6C,$61,$B5,$E2,$3F,$95,$BF,$CC,$3C,$6F,$E6,$3E,$34
	.dc.b	$9F,$E5,$F1,$97,$F9,$8C,$65,$FE,$5B,$19,$7F,$97,$C6,$71,$FE,$6B
	.dc.b	$39,$7F,$9B,$C6,$71,$FE,$59,$39,$7F,$97,$46,$6B,$F9,$5C,$E5,$FE
	.dc.b	$5B,$19,$9F,$E7,$53,$FD,$23,$C6,$FE,$93,$E3,$4B,$FE,$8F,$19,$7F
	.dc.b	$A4,$C6,$5F,$E8,$B1,$97,$FA,$3C,$67,$0F,$E9,$B3,$99,$FE,$34,$FF
	.dc.b	$27,$23,$FC,$95,$AF,$F2,$36,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$AF
	.dc.b	$F8,$DE,$58,$CC,$7A,$0C,$9B,$30,$47,$FB,$2C,$7F,$6C,$F1,$A4,$FF
	.dc.b	$6A,$7F,$B7,$F8,$B8,$8F,$EC,$F1,$6B,$FB,$6C,$DC,$BF,$B1,$4D,$AF
	.dc.b	$EB,$F3,$67,$FB,$64,$DD,$93,$63,$0D,$AF,$11,$FD,$F1,$FF,$00,$F8
	.dc.b	$DF,$E0,$2F,$1A,$4F,$F8,$02,$32,$FF,$80,$A3,$2F,$F7,$E8,$CB,$FE
	.dc.b	$00,$8C,$E3,$FE,$06,$9C,$BF,$E0,$88,$CE,$3F,$DF,$67,$2F,$F7,$F8
	.dc.b	$CD,$7F,$7C,$9C,$BF,$DF,$A3,$33,$FE,$0B,$9F,$F0,$CF,$8B,$FE,$1A
	.dc.b	$F1,$65,$FF,$86,$23,$2F,$F8,$6A,$32,$FF,$85,$A3,$2F,$F8,$62,$33
	.dc.b	$87,$F8,$7A,$73,$3F,$DA,$DF,$EF,$04,$7F,$BB,$35,$FE,$E8,$D7,$81
	.dc.b	$A2,$1A,$2C,$77,$7D,$07,$F8,$C3,$C4,$F2,$C6,$63,$D0,$64,$D9,$82
	.dc.b	$3F,$E3,$68,$FF,$1E,$F8,$B2,$7F,$C7,$6F,$F9,$03,$C4,$C4,$7F,$8E
	.dc.b	$22,$D7,$F8,$FA,$6E,$5F,$E3,$39,$B5,$FE,$30,$9B,$3F,$E3,$D9,$BB
	.dc.b	$26,$C6,$1B,$5E,$23,$FC,$9A,$FF,$94,$BC,$5F,$F2,$9F,$8B,$27,$FC
	.dc.b	$A3,$19,$7F,$CA,$71,$97,$FC,$9F,$19,$7F,$CA,$31,$9C,$7F,$CA,$F3
	.dc.b	$97,$FC,$B3,$19,$C7,$FC,$9D,$39,$7F,$CA,$11,$9A,$FF,$26,$CE,$5F
	.dc.b	$F2,$7C,$66,$7F,$CB,$93,$FE,$6A,$F1,$3F,$CD,$7E,$2C,$BF,$F3,$4C
	.dc.b	$65,$FF,$35,$C6,$5F,$F3,$3C,$65,$FF,$34,$C6,$70,$FF,$37,$CE,$67
	.dc.b	$FC,$78,$FF,$92,$48,$FF,$91,$DA,$FF,$91,$5A,$F0,$34,$43,$45,$8E
	.dc.b	$EF,$A0,$05,$7F,$E7,$CF,$DC,$7F,$CE,$13,$FE,$72,$AF,$55,$72,$8E
	.dc.b	$15,$BF,$1B,$2F,$F9,$F6,$AC,$F4,$BE,$72,$7F,$D1,$51,$9F,$1B,$C6
	.dc.b	$F4,$30,$79,$B0,$F2,$BC,$FC,$5D,$C6,$2F,$A5,$61,$F5,$36,$B9,$9D
	.dc.b	$63,$C5,$59,$78,$C7,$77,$8B,$4E,$35,$DE,$FF,$CE,$AD,$BE,$51,$EC
	.dc.b	$AE,$75,$DE,$37,$1D,$23,$13,$C0,$E3,$76,$3C,$13,$CA,$FB,$46,$2F
	.dc.b	$9B,$F3,$7C,$6F,$0F,$8D,$23,$51,$E4,$30,$F8,$E3,$0E,$8E,$8E,$AB
	.dc.b	$DD,$3A,$9E,$06,$8E,$07,$B5,$E1,$7F,$89,$AE,$15,$CA,$B8,$1C,$8E
	.dc.b	$07,$AF,$C6,$F9,$3E,$34,$8C,$10,$F9,$4E,$5F,$99,$1F,$40,$C4,$FD
	.dc.b	$59,$C4,$F1,$3F,$D6,$D8,$AB,$F1,$F1,$86,$BF,$A1,$6B,$F9,$96,$E1
	.dc.b	$F9,$96,$CF,$39,$E3,$39,$6C,$65,$EF,$19,$6C,$65,$F1,$1C,$27,$D2
	.dc.b	$FB,$8F,$C0,$78,$B8,$C3,$F2,$53,$6F,$F7,$13,$6B,$D0,$7C,$9A,$FC
	.dc.b	$C4,$7E,$5B,$C6,$FC,$BF,$8B,$2B,$F1,$91,$97,$F1,$D1,$97,$F3,$51
	.dc.b	$97,$F3,$91,$9A,$FD,$24,$E5,$FD,$34,$66,$FF,$86,$4E,$5F,$E1,$D1
	.dc.b	$9A,$FE,$0D,$39,$7F,$84,$46,$67,$F8,$44,$FE,$BB,$C6,$95,$FA,$37
	.dc.b	$F6,$3E,$36,$23,$F6,$11,$70,$FD,$C4,$DB,$FE,$3D,$36,$BF,$8B,$4D
	.dc.b	$9F,$E2,$D3,$73,$4D,$8C,$36,$BC,$47,$ED,$5F,$E0,$7E,$37,$F0,$4F
	.dc.b	$1A,$4F,$EE,$63,$2F,$EE,$E3,$2F,$F0,$78,$CB,$FC,$26,$33,$8F,$F1
	.dc.b	$19,$CC,$FE,$71,$FC,$99,$1F,$C8,$B5,$FC,$7B,$5E,$06,$88,$68,$B1
	.dc.b	$DD,$F4,$1F,$C4,$7C,$5F,$2C,$66,$3D,$06,$4D,$98,$23,$FC,$A6,$3F
	.dc.b	$93,$F8,$DF,$CA,$3C,$59,$5F,$C7,$63,$2F,$F1,$F8,$CB,$FC,$AE,$32
	.dc.b	$FF,$2D,$8C,$D7,$F3,$59,$CB,$FC,$DE,$33,$A7,$F5,$D9,$CB,$FD,$7E
	.dc.b	$33,$5F,$D5,$E7,$2F,$F5,$98,$CC,$FF,$59,$9F,$E7,$FE,$34,$AF,$E6
	.dc.b	$8F,$F4,$2F,$1B,$11,$FD,$06,$2E,$1F,$D2,$66,$DF,$F6,$F9,$B5,$FD
	.dc.b	$9E,$6C,$FF,$67,$9B,$B2,$6C,$61,$B5,$E2,$3F,$A3,$3F,$D5,$3C,$6F
	.dc.b	$EA,$BE,$34,$9F,$E9,$51,$97,$FA,$64,$65,$FE,$B1,$19,$7F,$AD,$46
	.dc.b	$71,$FE,$C5,$39,$9F,$E5,$AF,$F2,$42,$3F,$C8,$9A,$FF,$20,$6B,$C0
	.dc.b	$D1,$0D,$16,$3B,$BE,$83,$FB,$17,$8B,$E5,$8C,$C7,$A0,$C9,$B3,$04
	.dc.b	$7F,$BD,$47,$F7,$8F,$1B,$FB,$CF,$8B,$2B,$FB,$74,$65,$FE,$E1,$19
	.dc.b	$7F,$BE,$46,$5F,$EF,$D1,$9A,$FF,$03,$4E,$5F,$F0,$44,$67,$4F,$F1
	.dc.b	$74,$E5,$FF,$18,$46,$6B,$FC,$51,$39,$7F,$C5,$51,$99,$FF,$15,$4F
	.dc.b	$F8,$43,$C6,$95,$FE,$05,$7F,$C2,$7E,$36,$23,$FC,$25,$17,$0F,$F0
	.dc.b	$D4,$DB,$FF,$20,$4D,$AF,$F1,$C4,$D9,$FF,$1C,$4D,$D9,$36,$30,$DA
	.dc.b	$F1,$1F,$E1,$77,$FC,$4B,$E2,$FF,$89,$BC,$59,$3F,$E1,$B8,$CB,$FE
	.dc.b	$1D,$8C,$BF,$E2,$98,$CB,$FE,$2B,$8C,$E3,$FE,$33,$9C,$CF,$F7,$E7
	.dc.b	$FB,$A9,$1F,$EE,$6D,$7F,$B8,$B5,$E0,$68,$86,$8B,$1D,$DF,$41,$FE
	.dc.b	$33,$F1,$3C,$B1,$98,$F4,$19,$36,$60,$8F,$F9,$2E,$3F,$C9,$3E,$2F
	.dc.b	$F9,$2B,$C3,$95,$FE,$3F,$8C,$BF,$E4,$18,$CB,$FE,$4D,$8C,$BF,$E4
	.dc.b	$F8,$CD,$7F,$95,$E7,$2F,$F9,$66,$33,$A7,$F9,$FE,$72,$FF,$A0,$63
	.dc.b	$35,$FE,$79,$9C,$BF,$E7,$B8,$CC,$FF,$9E,$E7,$FC,$C3,$E2,$CA,$FF
	.dc.b	$2B,$3F,$E6,$DF,$DC,$BF,$D0,$3B,$23,$53,$FE,$65,$75,$E3,$63,$2F
	.dc.b	$F3,$74,$5C,$7F,$D2,$E3,$0E,$71,$9B,$FF,$47,$3C,$9F,$09,$0E,$C7
	.dc.b	$F9,$B2,$38,$B9,$AE,$EE,$6F,$D1,$E3,$7A,$7C,$69,$3C,$63,$2F,$38
	.dc.b	$CB,$FE,$8D,$BF,$F4,$11,$23,$37,$FE,$97,$8C,$D7,$F9,$31,$34,$95
	.dc.b	$FE,$4C,$8C,$B6,$32,$FF,$93,$23,$2D,$8C,$BF,$EA,$CC,$7E,$44,$62
	.dc.b	$3E,$5C,$F8,$9F,$61,$F1,$7C,$6F,$6B,$EB,$9B,$7E,$99,$B7,$CA,$6D
	.dc.b	$7F,$B2,$AD,$5F,$81,$F1,$BF,$05,$E2,$CA,$FA,$51,$97,$EA,$46,$5F
	.dc.b	$95,$19,$7E,$64,$67,$1F,$95,$39,$7E,$64,$67,$1F,$6C,$E5,$F7,$C6
	.dc.b	$6B,$CD,$39,$7D,$11,$99,$FC,$74,$FE,$67,$C6,$95,$F8,$97,$F3,$9E
	.dc.b	$36,$23,$F0,$11,$70,$FB,$D3,$6F,$E7,$CD,$AF,$8D,$36,$7F,$39,$37
	.dc.b	$24,$D8,$C3,$6B,$C4,$7E,$6D,$FD,$77,$8B,$FA,$FF,$1A,$4F,$E7,$E3
	.dc.b	$2F,$E8,$63,$2F,$E5,$E3,$2F,$E6,$63,$38,$FE,$5E,$73,$3F,$8B,$C1
	.dc.b	$1F,$99,$24,$7E,$5D,$AF,$E5,$5A,$F1,$6B,$F9,$26,$BF,$89,$7C,$73
	.dc.b	$E1,$8F,$DC,$78,$D8,$CB,$84,$7B,$4C,$94,$F6,$16,$7F,$61,$3F,$C4
	.dc.b	$FC,$6F,$E2,$9E,$2C,$AF,$E0,$91,$97,$F8,$34,$65,$FD,$E4,$65,$FD
	.dc.b	$F4,$66,$BF,$79,$39,$7F,$7D,$19,$BF,$D7,$CE,$5F,$D8,$C6,$6B,$F5
	.dc.b	$13,$97,$F5,$51,$99,$FE,$41,$3F,$CA,$FC,$69,$5F,$C6,$DF,$E5,$DE
	.dc.b	$36,$23,$F8,$94,$5C,$3F,$88,$4D,$BF,$E0,$33,$6B,$F7,$13,$67,$F9
	.dc.b	$74,$DC,$D3,$63,$0D,$AF,$11,$FC,$B5,$FE,$81,$E2,$FF,$41,$F1,$A4
	.dc.b	$FF,$31,$8C,$BF,$CC,$E3,$2F,$F2,$98,$CB,$FC,$AE,$33,$8F,$F2,$99
	.dc.b	$CC,$FE,$4D,$FE,$54,$47,$F9,$43,$5F,$E4,$CD,$78,$1A,$21,$A2,$C7
	.dc.b	$77,$D0,$7F,$49,$F1,$BC,$B1,$98,$F4,$19,$36,$60,$8F,$F4,$18,$FE
	.dc.b	$C9,$E3,$7F,$65,$F1,$65,$7F,$55,$8C,$BF,$D5,$E3,$2F,$F4,$D8,$CB
	.dc.b	$FD,$3E,$33,$5F,$D3,$67,$2F,$F4,$F8,$CE,$9F,$D0,$27,$2F,$F4,$28
	.dc.b	$CD,$7F,$39,$9C,$BF,$CE,$E3,$33,$FD,$C6,$7F,$BE,$78,$D2,$BF,$B6
	.dc.b	$3F,$DF,$FC,$6C,$47,$F6,$38,$B8,$7F,$61,$9B,$7F,$D4,$A6,$D7,$F4
	.dc.b	$99,$B3,$FD,$FE,$6E,$C9,$B1,$86,$D7,$88,$FE,$FC,$FF,$84,$7C,$5F
	.dc.b	$F0,$97,$8D,$27,$FC,$05,$19,$7F,$C0,$91,$97,$FB,$D4,$65,$FE,$F9
	.dc.b	$19,$C7,$FB,$D4,$E6,$7F,$92,$3F,$DE,$C8,$FF,$79,$6B,$FD,$DD,$AF
	.dc.b	$03,$44,$34,$58,$EE,$FA,$0F,$F0,$D7,$8D,$E5,$8C,$C7,$A0,$C9,$B3
	.dc.b	$04,$7F,$C2,$51,$FE,$35,$F1,$BF,$C6,$DE,$2C,$AF,$F1,$34,$65,$FF
	.dc.b	$14,$46,$5F,$F0,$F4,$65,$FF,$10,$46,$6B,$FC,$3D,$39,$7F,$C4,$11
	.dc.b	$9D,$3F,$C2,$33,$97,$FC,$27,$19,$AF,$F0,$54,$E5,$FF,$06,$46,$67
	.dc.b	$FC,$85,$3F,$E4,$DF,$1A,$57,$F8,$F5,$FF,$28,$78,$D8,$8F,$F1,$A4
	.dc.b	$5C,$3F,$C6,$53,$6F,$FC,$47,$36,$BF,$C3,$53,$67,$FC,$A1,$37,$64
	.dc.b	$D8,$C3,$6B,$C4,$7F,$93,$DF,$F3,$17,$8B,$FE,$63,$F1,$A4,$FF,$94
	.dc.b	$E3,$2F,$F9,$56,$32,$FF,$92,$E3,$2F,$F9,$36,$33,$8F,$F9,$2E,$73
	.dc.b	$3F,$DD,$5F,$F2,$61,$1F,$F2,$53,$5F,$F2,$43,$5E,$06,$88,$68,$B1
	.dc.b	$DD,$F4,$1F,$E6,$BF,$17,$CB,$19,$8F,$41,$93,$66,$08,$FF,$98,$E1
	.dc.b	$12,$19,$7F,$A0,$3F,$73,$78,$DA,$A5,$69,$DC,$E9,$E5,$5C,$6B,$85
	.dc.b	$6E,$15,$FE,$68,$8C,$3C,$6B,$AC,$F1,$79,$9D,$AB,$9C,$79,$7C,$5B
	.dc.b	$39,$37,$1D,$A7,$2D,$8C,$D7,$86,$3B,$47,$80,$E4,$57,$83,$63,$83
	.dc.b	$AB,$F4,$D7,$69,$F5,$47,$B2,$F1,$32,$BC,$2E,$E3,$87,$8D,$C7,$C6
	.dc.b	$95,$E6,$8C,$BE,$88,$CB,$E3,$8C,$BE,$58,$CD,$FC,$C9,$CB,$F3,$A3
	.dc.b	$38,$FD,$D9,$CB,$F7,$E3,$35,$F7,$27,$2F,$DE,$8C,$CF,$DA,$9F,$81
	.dc.b	$E2,$FC,$1F,$1A,$5F,$CE,$8C,$BF,$42,$32,$FC,$B8,$CB,$F3,$63,$38
	.dc.b	$7E,$12,$73,$3F,$EA,$BC,$D1,$F9,$38,$C3,$5F,$CC,$35,$FC,$AB,$70
	.dc.b	$FC,$8B,$67,$9C,$F1,$9C,$B6,$32,$F7,$8C,$B6,$32,$F8,$8E,$13,$EC
	.dc.b	$AF,$54,$7E,$43,$C6,$C5,$F9,$A3,$E3,$19,$29,$F1,$4B,$3F,$A6,$9F
	.dc.b	$C7,$78,$D2,$7F,$36,$FE,$47,$C5,$C4,$7E,$5E,$2D,$7E,$BA,$6D,$FE
	.dc.b	$F6,$6D,$7E,$EA,$6C,$FE,$DA,$6E,$3F,$E9,$0A,$EE,$43,$B1,$FA,$CA
	.dc.b	$DC,$F0,$7F,$49,$E3,$7E,$97,$C6,$95,$FB,$28,$CB,$FB,$48,$CB,$FA
	.dc.b	$F8,$CB,$FB,$18,$CE,$3F,$C2,$27,$2F,$F0,$A8,$CE,$3F,$C6,$67,$2F
	.dc.b	$F1,$B8,$CD,$7F,$18,$9C,$BF,$C6,$A3,$33,$FC,$5A,$7F,$79,$E2,$FE
	.dc.b	$F7,$C6,$97,$FC,$2A,$32,$FF,$0C,$8C,$BF,$C1,$E3,$2F,$F0,$98,$CE
	.dc.b	$1F,$C8,$27,$33,$F9,$D7,$F2,$84,$7F,$24,$D7,$F2,$0D,$78,$1A,$21
	.dc.b	$A2,$C7,$77,$D0,$7F,$22,$F1,$BC,$B1,$98,$F4,$19,$36,$60,$8F,$F3
	.dc.b	$88,$FE,$41,$E3,$49,$FE,$5A,$FF,$23,$F1,$71,$1F,$CA,$62,$D7,$F4
	.dc.b	$09,$B9,$7F,$4F,$9B,$5F,$D3,$26,$CF,$F4,$89,$BB,$26,$C6,$1B,$5E
	.dc.b	$23,$F9,$FB,$FC,$D7,$C6,$FE,$6D,$E3,$49,$FE,$87,$19,$7F,$A2,$C6
	.dc.b	$5F,$E8,$11,$97,$FA,$14,$67,$1F,$EB,$33,$97,$FA,$DC,$67,$1F,$ED
	.dc.b	$53,$97,$FB,$64,$66,$BF,$B4,$CE,$5F,$ED,$71,$99,$FE,$CF,$3F,$D3
	.dc.b	$7C,$5F,$E9,$DE,$34,$BF,$EB,$71,$97,$FA,$EC,$65,$FE,$B1,$19,$7F
	.dc.b	$AD,$46,$70,$FE,$E3,$39,$9F,$E5,$CF,$F2,$52,$3F,$C8,$DA,$FF,$21
	.dc.b	$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$FB,$9F,$8D,$E5,$8C,$C7,$A0,$C9
	.dc.b	$B3,$04,$7F,$C1,$31,$FD,$C7,$C6,$93,$FD,$F9,$FE,$E9,$E2,$E2,$3F
	.dc.b	$BD,$45,$AF,$F0,$8C,$DC,$BF,$C4,$13,$6B,$FC,$3B,$36,$7F,$C3,$33
	.dc.b	$76,$4D,$8C,$36,$BC,$47,$F8,$41,$FF,$03,$78,$DF,$E0,$7F,$1A,$4F
	.dc.b	$F8,$52,$32,$FF,$85,$A3,$2F,$F8,$46,$32,$FF,$84,$E3,$38,$FF,$8A
	.dc.b	$A7,$2F,$F8,$B2,$33,$8F,$F8,$EE,$72,$FF,$8F,$63,$35,$FE,$3A,$9C
	.dc.b	$BF,$E3,$C8,$CC,$FF,$8E,$27,$FC,$3D,$E2,$FF,$87,$FC,$69,$7F,$E2
	.dc.b	$C8,$CB,$FE,$2E,$8C,$BF,$E2,$98,$CB,$FE,$2B,$8C,$E1,$FE,$42,$9C
	.dc.b	$CF,$F7,$F7,$FB,$B1,$1F,$EE,$8D,$7F,$B9,$35,$E0,$68,$86,$8B,$1D
	.dc.b	$DF,$41,$FE,$44,$F1,$7C,$B1,$98,$F4,$19,$36,$60,$8F,$F9,$6A,$3F
	.dc.b	$C8,$5E,$2C,$9F,$F2,$7B,$FE,$45,$F1,$31,$1F,$E4,$B8,$B5,$FE,$62
	.dc.b	$9B,$97,$F9,$C6,$6D,$7F,$9B,$A6,$CF,$F9,$AA,$6E,$C9,$B1,$86,$D7
	.dc.b	$88,$FF,$30,$BF,$E5,$7F,$17,$FC,$B1,$E2,$C9,$FF,$32,$C6,$5F,$F3
	.dc.b	$3C,$65,$FF,$31,$46,$5F,$F3,$24,$67,$1F,$F3,$DC,$E5,$FF,$3E,$C6
	.dc.b	$71,$FF,$46,$4E,$5F,$F4,$74,$66,$BF,$D1,$73,$97,$FD,$1B,$19,$9F
	.dc.b	$F4,$4C,$FF,$9F,$2F,$FC,$DF,$1F,$E9,$A3,$E1,$78,$99,$35,$E2,$CC
	.dc.b	$3F,$D0,$B5,$FE,$7A,$AD,$C6,$27,$4F,$FA,$06,$74,$48,$CC,$7F,$A8
	.dc.b	$E7,$E8,$47,$F9,$36,$7F,$D3,$53,$96,$BE,$56,$BF,$EA,$E6,$CF,$28
	.dc.b	$FF,$29,$46,$09,$3D,$49,$1A,$7C,$F8,$FF,$AB,$67,$E1,$CF,$5F,$13
	.dc.b	$37,$E1,$BB,$39,$3E,$01,$0F,$80,$FC,$D9,$F9,$DE,$2C,$9F,$92,$FD
	.dc.b	$1F,$13,$11,$F1,$E2,$E3,$E3,$9B,$87,$C1,$9B,$5F,$06,$6C,$FE,$66
	.dc.b	$6E,$3F,$E7,$0B,$F6,$46,$0F,$05,$F8,$88,$78,$4F,$BB,$7F,$85,$F1
	.dc.b	$7F,$0D,$E2,$C9,$FC,$0C,$65,$FC,$1C,$65,$FC,$04,$65,$FC,$14,$66
	.dc.b	$FE,$74,$E5,$FA,$11,$9C,$7E,$F4,$E5,$FC,$04,$66,$BF,$01,$39,$7F
	.dc.b	$05,$19,$9F,$DD,$4F,$E7,$7C,$5F,$CF,$78,$B2,$FF,$33,$19,$7F,$37
	.dc.b	$19,$7F,$31,$19,$7F,$35,$19,$C3,$F0,$D3,$99,$FC,$B3,$F9,$92,$3F
	.dc.b	$92,$6B,$FE,$A1,$CD,$18,$8E,$26,$C8,$6D,$E3,$3E,$08,$F4,$9F,$B2
	.dc.b	$F1,$3C,$D1,$98,$F4,$99,$36,$60,$8F,$E7,$E3,$F8,$17,$8B,$27,$F7
	.dc.b	$AF,$F0,$5F,$13,$11,$FB,$B8,$B5,$FA,$79,$B9,$FE,$D6,$6D,$7E,$D6
	.dc.b	$6C,$FF,$2D,$9B,$8A,$6C,$61,$B5,$E2,$3F,$87,$BF,$C5,$FC,$5F,$E3
	.dc.b	$1E,$2C,$9F,$E2,$71,$97,$F8,$AC,$65,$FE,$25,$19,$7F,$8A,$46,$71
	.dc.b	$FE,$01,$39,$7F,$81,$46,$71,$FE,$21,$39,$7F,$89,$46,$6B,$F8,$94
	.dc.b	$E5,$FE,$29,$19,$9F,$E9,$93,$FC,$BF,$C5,$FE,$61,$E2,$CB,$FE,$57
	.dc.b	$19,$7F,$96,$C6,$5F,$E5,$51,$97,$F9,$64,$67,$0F,$E3,$13,$99,$FE
	.dc.b	$50,$FF,$2B,$23,$FC,$91,$AF,$E4,$DA,$F0,$34,$43,$45,$8E,$EF,$A0
	.dc.b	$FE,$87,$E2,$79,$63,$31,$E8,$32,$6C,$C1,$1F,$E6,$11,$FD,$4F,$C5
	.dc.b	$93,$FD,$39,$FE,$AD,$E2,$62,$3F,$A6,$45,$AF,$E7,$13,$72,$FE,$8D
	.dc.b	$36,$BF,$A3,$4D,$9F,$EF,$D3,$76,$4D,$8C,$36,$BC,$47,$F6,$07,$FB
	.dc.b	$47,$8B,$FD,$A7,$C5,$93,$FD,$92,$32,$FF,$66,$8C,$BF,$D8,$E3,$2F
	.dc.b	$F6,$58,$CE,$3F,$D4,$67,$2F,$F5,$38,$CE,$3F,$D8,$67,$2F,$F6,$38
	.dc.b	$CD,$7F,$63,$9C,$BF,$D9,$63,$33,$FE,$1D,$9F,$F0,$07,$8B,$FE,$01
	.dc.b	$F1,$65,$FF,$7C,$8C,$BF,$DF,$A3,$2F,$F7,$B8,$CB,$FD,$F6,$33,$87
	.dc.b	$F6,$99,$CC,$FF,$79,$7F,$BE,$11,$FE,$EA,$D7,$F9,$23,$5E,$06,$88
	.dc.b	$68,$B1,$DD,$F4,$1F,$E1,$4F,$13,$CB,$19,$8F,$41,$93,$66,$08,$FF
	.dc.b	$80,$63,$FC,$49,$E2,$C9,$FF,$0F,$BF,$E2,$7F,$13,$11,$FE,$1D,$8B
	.dc.b	$5F,$E0,$99,$B9,$7F,$85,$E6,$D7,$F8,$5E,$6C,$FF,$93,$E6,$EC,$9B
	.dc.b	$18,$6D,$78,$8F,$F1,$8B,$FE,$39,$F1,$7F,$C7,$5E,$2C,$9F,$F1,$AC
	.dc.b	$65,$FF,$1B,$C6,$5F,$F1,$A4,$65,$FF,$1B,$46,$71,$FF,$11,$4E,$5F
	.dc.b	$F1,$24,$67,$1F,$F1,$94,$E5,$FF,$1A,$46,$6B,$FC,$69,$39,$7F,$C6
	.dc.b	$D1,$99,$FF,$37,$4F,$F9,$47,$C4,$FF,$29,$65,$2F,$FC,$9B,$19,$7F
	.dc.b	$C9,$F1,$97,$FC,$99,$19,$7F,$C9,$D1,$9C,$3F,$C7,$53,$99,$FF,$25
	.dc.b	$3F,$E4,$D2,$3F,$E4,$66,$BF,$DD,$5A,$F0,$34,$43,$45,$8E,$EF,$A0
	.dc.b	$FF,$32,$F8,$7E,$58,$CC,$7A,$0C,$9B,$30,$47,$FC,$A5,$1F,$E7,$5F
	.dc.b	$12,$4F,$F9,$C1,$32,$58,$FF,$9E,$7F,$74,$FF,$2C,$BF,$E9,$BA,$E3
	.dc.b	$5F,$E7,$37,$8A,$AE,$2F,$2A,$E3,$8A,$B8,$D7,$F9,$95,$E2,$F0,$B5
	.dc.b	$70,$9D,$C7,$F9,$DD,$DB,$E5,$F1,$33,$87,$F9,$E6,$6C,$F2,$9B,$87
	.dc.b	$86,$BE,$01,$EF,$9E,$4D,$3D,$1E,$2F,$67,$D5,$E2,$D9,$F2,$B7,$1F
	.dc.b	$F4,$C1,$1C,$19,$A9,$5D,$6B,$8B,$69,$5A,$8F,$95,$3F,$46,$B9,$C6
	.dc.b	$E7,$59,$7C,$7B,$F0,$4E,$4E,$95,$AB,$FB,$57,$CA,$7A,$9E,$E3,$C5
	.dc.b	$E2,$FA,$FC,$59,$18,$21,$C2,$BE,$AC,$6A,$31,$82,$BE,$21,$9C,$7C
	.dc.b	$D5,$8A,$E6,$7F,$AD,$EB,$83,$F8,$A3,$F2,$15,$5A,$5B,$FC,$2C,$66
	.dc.b	$3F,$1C,$D8,$CC,$6E,$7A,$BD,$63,$2D,$8C,$BE,$39,$E5,$18,$8F,$09
	.dc.b	$EE,$8F,$53,$F0,$3C,$4B,$87,$D3,$9B,$7F,$A9,$9B,$5E,$63,$E5,$D7
	.dc.b	$D9,$8F,$C5,$78,$B2,$7F,$08,$FE,$37,$C4,$C4,$7C,$E8,$B5,$F8,$19
	.dc.b	$B7,$FB,$79,$B5,$FB,$19,$B3,$FA,$C9,$B8,$FF,$A2,$2B,$B9,$0E,$C7
	.dc.b	$E5,$6B,$73,$C1,$FD,$07,$8B,$FA,$1F,$16,$57,$E6,$E3,$2F,$E7,$63
	.dc.b	$2F,$E1,$A3,$2F,$E2,$23,$38,$FE,$6E,$72,$FE,$76,$33,$8F,$F1,$39
	.dc.b	$CB,$FC,$56,$33,$5F,$C3,$E7,$2F,$F1,$18,$CC,$FF,$0C,$9F,$DC,$78
	.dc.b	$BF,$B9,$F1,$65,$FE,$CE,$32,$FE,$D6,$32,$FE,$7A,$32,$FE,$82,$33
	.dc.b	$87,$EC,$E7,$33,$F9,$D7,$F1,$64,$7F,$28,$D7,$F2,$4D,$78,$1A,$21
	.dc.b	$A2,$C7,$77,$D0,$7F,$0B,$F1,$3C,$B1,$98,$F4,$19,$36,$60,$8F,$F0
	.dc.b	$B8,$FE,$39,$E2,$C9,$FE,$2C,$FF,$1F,$F1,$31,$1F,$C0,$22,$D7,$F1
	.dc.b	$39,$B9,$7F,$49,$9B,$5F,$D0,$E6,$CF,$F3,$E9,$BB,$26,$C6,$1B,$5E
	.dc.b	$23,$F9,$43,$FC,$C7,$C5,$FE,$65,$E2,$C9,$FE,$59,$19,$7F,$97,$46
	.dc.b	$5F,$E2,$F1,$97,$F8,$CC,$67,$1F,$E5,$93,$97,$F9,$74,$67,$1F,$EC
	.dc.b	$93,$97,$FB,$34,$66,$BF,$B0,$4E,$5F,$EC,$51,$99,$FE,$BB,$3F,$D2
	.dc.b	$7C,$5F,$E9,$5E,$2C,$BF,$E8,$91,$97,$FA,$34,$65,$FE,$5F,$19,$7F
	.dc.b	$98,$C6,$70,$FE,$89,$39,$9F,$E5,$CF,$F1,$C2,$3F,$C9,$5A,$FF,$23
	.dc.b	$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$FA,$E7,$89,$E5,$8C,$C7,$A0,$C9
	.dc.b	$B3,$04,$7F,$AE,$47,$F6,$DF,$16,$4F,$F6,$77,$FB,$87,$89,$88,$FE
	.dc.b	$A3,$16,$BF,$B2,$4D,$CB,$FC,$35,$36,$BF,$C2,$93,$67,$FC,$1F,$37
	.dc.b	$64,$D8,$C3,$6B,$C4,$7F,$79,$7F,$C0,$5E,$2F,$F8,$0F,$C5,$93,$FD
	.dc.b	$F6,$32,$FF,$7F,$8C,$BF,$DA,$23,$2F,$F6,$A8,$CE,$3F,$DF,$67,$2F
	.dc.b	$F7,$F8,$CE,$3F,$E3,$59,$CB,$FE,$37,$8C,$D7,$F8,$C6,$72,$FF,$8C
	.dc.b	$E3,$33,$FE,$2E,$9F,$F0,$D7,$8B,$FE,$1B,$F1,$65,$FF,$85,$63,$2F
	.dc.b	$F8,$5E,$32,$FF,$80,$23,$2F,$F8,$0A,$33,$87,$F8,$56,$73,$3F,$DF
	.dc.b	$DF,$ED,$A4,$7F,$BB,$35,$FE,$E8,$D7,$81,$A2,$1A,$2C,$77,$7D,$07
	.dc.b	$F8,$B7,$C4,$F2,$C6,$63,$D0,$64,$D9,$82,$3F,$E2,$D8,$FF,$1F,$78
	.dc.b	$B2,$7F,$C7,$0F,$F9,$07,$C5,$C4,$7F,$88,$A2,$D7,$F8,$D6,$6E,$5F
	.dc.b	$E6,$B9,$B5,$FE,$65,$9B,$3F,$E6,$09,$BB,$26,$C6,$1B,$5E,$23,$FC
	.dc.b	$94,$FF,$94,$FC,$2F,$F2,$A7,$8B,$27,$FC,$9D,$19,$7F,$CA,$11,$97
	.dc.b	$FC,$73,$19,$7F,$C7,$71,$9C,$7F,$C9,$D3,$97,$FC,$A1,$19,$C7,$FD
	.dc.b	$0D,$39,$7F,$D1,$11,$9A,$FF,$41,$4E,$5F,$F4,$24,$66,$7F,$CF,$F3
	.dc.b	$FE,$6B,$F1,$7F,$CD,$DF,$BA,$FF,$A4,$0F,$F4,$8E,$7F,$E7,$9C,$3F
	.dc.b	$C9,$F3,$AF,$1A,$47,$F9,$AE,$39,$3F,$E6,$FA,$E5,$38,$8E,$27,$F9
	.dc.b	$16,$7F,$C7,$E4,$70,$43,$FC,$8C,$E5,$A6,$4F,$F2,$D3,$96,$9E,$32
	.dc.b	$1F,$E6,$99,$FF,$43,$C7,$8F,$2F,$F3,$D3,$F0,$3C,$6C,$65,$FE,$B1
	.dc.b	$8B,$7F,$EB,$4B,$57,$C6,$F1,$A4,$FB,$DF,$93,$E2,$E2,$3C,$11,$71
	.dc.b	$FF,$57,$5F,$F9,$91,$B1,$9C,$7F,$D8,$62,$7F,$D2,$B5,$F2,$EF,$4F
	.dc.b	$8E,$AD,$6E,$6D,$70,$AF,$09,$C4,$F3,$10,$C9,$F1,$63,$F0,$3E,$37
	.dc.b	$E0,$BC,$69,$3F,$6E,$32,$FD,$D8,$CB,$F3,$63,$2F,$CF,$8C,$E3,$F0
	.dc.b	$27,$2F,$C2,$8C,$E1,$F0,$30,$56,$A3,$13,$AB,$FA,$53,$97,$EA,$46
	.dc.b	$67,$E4,$4F,$E6,$BC,$6F,$CD,$F8,$D2,$FF,$29,$19,$7F,$2D,$19,$7F
	.dc.b	$0D,$19,$7F,$11,$19,$C7,$EC,$CE,$67,$FD,$E3,$7F,$5E,$3F,$D4,$58
	.dc.b	$9A,$49,$FF,$55,$46,$5B,$19,$7F,$D4,$D1,$96,$C6,$5E,$F1,$96,$C7
	.dc.b	$9E,$BC,$B1,$FB,$1F,$17,$17,$F3,$6A,$CF,$E6,$E6,$CF,$A6,$3E,$29
	.dc.b	$C4,$FD,$D7,$8D,$FC,$10,$8F,$EF,$7C,$5A,$FE,$B6,$2E,$1F,$9E,$9B
	.dc.b	$5F,$76,$6D,$7E,$AE,$6C,$FE,$7E,$6E,$29,$B1,$86,$D7,$90,$F1,$1F
	.dc.b	$C3,$FC,$6F,$E3,$3E,$34,$8C,$10,$FE,$04,$E5,$FE,$09,$1F,$C8,$0C
	.dc.b	$63,$FB,$69,$C4,$F1,$8C,$3F,$C7,$63,$2F,$23,$04,$30,$48,$E2,$F8
	.dc.b	$8E,$98,$FE,$87,$C6,$FE,$63,$E3,$49,$FE,$55,$1F,$CD,$0C,$3F,$C6
	.dc.b	$E3,$0E,$8F,$E6,$0E,$AB,$F8,$7C,$EA,$7F,$1A,$FE,$58,$8F,$E5,$1A
	.dc.b	$E8,$C1,$0C,$10,$FC,$9B,$C2,$3B,$C7,$8C,$FE,$61,$E2,$F8,$63,$31
	.dc.b	$FC,$CE,$7F,$A7,$CE,$6F,$E1,$4F,$1A,$FE,$4F,$3F,$D5,$3C,$6F,$EA
	.dc.b	$BE,$2C,$AF,$E9,$D1,$97,$FA,$84,$65,$FE,$81,$19,$7F,$A1,$46,$6B
	.dc.b	$F9,$9C,$E5,$FE,$6B,$19,$BF,$E2,$93,$97,$F8,$B4,$66,$BF,$A2,$4E
	.dc.b	$5F,$E8,$D1,$99,$FE,$75,$3F,$DB,$3C,$69,$5F,$D9,$5F,$ED,$FE,$36
	.dc.b	$23,$FA,$B4,$5C,$3F,$A3,$CD,$BF,$E4,$D3,$6B,$FA,$A4,$D9,$FE,$91
	.dc.b	$37,$81,$36,$30,$DA,$F1,$1F,$DD,$DF,$F0,$0F,$8D,$FE,$02,$F1,$A4
	.dc.b	$FF,$7B,$8C,$BF,$DF,$63,$2F,$F6,$D8,$CB,$FD,$BE,$33,$8F,$F6,$09
	.dc.b	$CC,$FF,$1A,$7F,$93,$91,$FE,$4A,$D7,$F9,$1B,$5E,$06,$88,$68,$B1
	.dc.b	$DD,$F4,$1F,$E1,$3F,$17,$CB,$19,$8F,$41,$93,$66,$08,$FF,$78,$8F
	.dc.b	$F1,$2F,$8D,$FE,$26,$F1,$65,$7F,$87,$E3,$2F,$F8,$86,$32,$FF,$84
	.dc.b	$63,$2F,$F8,$4E,$33,$5F,$E0,$49,$CB,$FE,$06,$8C,$E9,$FD,$96,$72
	.dc.b	$FF,$67,$8C,$D7,$F8,$56,$72,$FF,$85,$E3,$33,$FE,$0B,$9F,$F1,$EF
	.dc.b	$8D,$2B,$FC,$6C,$FF,$90,$3C,$6C,$47,$F8,$9E,$2E,$1F,$E1,$89,$B7
	.dc.b	$FD,$DE,$6D,$7F,$89,$66,$CF,$F8,$66,$6E,$C9,$B1,$86,$D7,$88,$FF
	.dc.b	$24,$3F,$E5,$2F,$1B,$FC,$A7,$E3,$49,$FF,$26,$46,$5F,$F2,$74,$65
	.dc.b	$FF,$1F,$46,$5F,$F2,$04,$67,$1F,$F1,$8C,$E6,$7F,$B5,$BF,$DE,$08
	.dc.b	$FF,$76,$6B,$FD,$D1,$AF,$03,$44,$34,$58,$EE,$FA,$0F,$F3,$27,$8B
	.dc.b	$E5,$8C,$C7,$A0,$C9,$B3,$04,$7F,$C9,$31,$FE,$76,$F1,$BF,$CE,$FE
	.dc.b	$24,$AF,$F3,$84,$65,$FF,$39,$46,$5F,$F3,$14,$65,$FF,$32,$46,$6B
	.dc.b	$FC,$AB,$39,$7F,$CA,$F1,$9D,$3F,$C6,$D3,$97,$FC,$71,$19,$AF,$F3
	.dc.b	$34,$E5,$FF,$34,$46,$67,$FC,$B9,$29,$94,$9F,$F3,$0F,$EE,$CF,$1B
	.dc.b	$54,$AD,$3B,$9D,$3C,$AB,$8D,$70,$AD,$C2,$BF,$CF,$35,$B3,$95,$6E
	.dc.b	$78,$C6,$5E,$11,$98,$F0,$78,$D6,$73,$1B,$AF,$1B,$E7,$8E,$6D,$38
	.dc.b	$9B,$AE,$53,$C2,$FC,$ED,$AF,$87,$7A,$A9,$7C,$4F,$04,$61,$ED,$3E
	.dc.b	$46,$D7,$37,$C2,$63,$C6,$F0,$F8,$D2,$3D,$51,$E4,$30,$F2,$8C,$3A
	.dc.b	$3E,$03,$AC,$7D,$33,$A9,$FB,$27,$D7,$6B,$F7,$9A,$E4,$CD,$FD,$78
	.dc.b	$CB,$63,$37,$FE,$9D,$7F,$CE,$2D,$7A,$1A,$9E,$31,$98,$EC,$79,$27
	.dc.b	$A3,$F0,$CF,$61,$F2,$3C,$5F,$55,$FB,$23,$2F,$D6,$8C,$CF,$82,$79
	.dc.b	$1C,$4F,$C2,$78,$DF,$7B,$C5,$91,$82,$1F,$59,$CB,$F6,$23,$F0,$A6
	.dc.b	$2F,$F0,$B3,$89,$E3,$18,$7F,$25,$19,$79,$18,$21,$82,$47,$17,$C4
	.dc.b	$74,$AF,$D9,$F8,$DF,$96,$24,$7E,$7F,$C6,$AF,$E3,$A2,$D7,$E6,$A6
	.dc.b	$D7,$2A,$FD,$35,$7E,$82,$73,$1F,$BB,$8B,$3F,$B7,$9B,$92,$6C,$61
	.dc.b	$B5,$FB,$1F,$1B,$F5,$7E,$34,$8C,$10,$FD,$1B,$97,$F4,$B1,$FB,$23
	.dc.b	$18,$FE,$CA,$71,$3C,$4F,$C7,$CF,$E3,$08,$FE,$51,$AF,$E3,$70,$46
	.dc.b	$23,$89,$B2,$1B,$78,$D7,$F0,$27,$C3,$1F,$BA,$F1,$AC,$E6,$3D,$8F
	.dc.b	$F0,$88,$CE,$5F,$CA,$AA,$D7,$F2,$79,$FE,$1F,$E3,$7F,$10,$F1,$64
	.dc.b	$FF,$19,$8C,$BF,$C6,$E3,$2F,$F0,$A8,$CB,$FC,$32,$33,$87,$F1,$89
	.dc.b	$CB,$FC,$6A,$33,$87,$F3,$C9,$CB,$FC,$FA,$33,$5F,$CE,$E7,$2F,$F3
	.dc.b	$D8,$CC,$FF,$39,$9F,$E4,$FE,$34,$AF,$E6,$0F,$F2,$AF,$1B,$11,$FC
	.dc.b	$82,$2E,$1F,$CB,$26,$DF,$F4,$E9,$B5,$FD,$2E,$6C,$FF,$47,$9B,$9A
	.dc.b	$6C,$61,$B5,$E2,$3F,$A1,$3F,$CF,$3C,$5F,$E7,$BE,$34,$9F,$E8,$D1
	.dc.b	$97,$FA,$44,$65,$FE,$6F,$19,$7F,$9C,$C6,$71,$FE,$8B,$39,$9F,$E4
	.dc.b	$2F,$F1,$E2,$3F,$C9,$9A,$FE,$4D,$AF,$03,$44,$34,$58,$EE,$FA,$0F
	.dc.b	$EB,$DE,$37,$96,$33,$1E,$83,$26,$CC,$11,$FE,$BB,$1F,$D8,$3C,$6F
	.dc.b	$EC,$3E,$2C,$AF,$ED,$51,$97,$FB,$64,$65,$FE,$B7,$19,$7F,$AE,$C6
	.dc.b	$6B,$FB,$4C,$E5,$FE,$D7,$19,$D3,$FC,$1B,$39,$7F,$C1,$F1,$9A,$FF
	.dc.b	$06,$4E,$5F,$F0,$74,$66,$7F,$C1,$53,$FD,$E3,$C6,$95,$FE,$01,$7F
	.dc.b	$BD,$F8,$D8,$8F,$EE,$31,$70,$FE,$FB,$36,$FF,$C3,$F3,$6B,$FC,$39
	.dc.b	$36,$7F,$C3,$13,$76,$4D,$8C,$36,$BC,$47,$F8,$4D,$FF,$06,$F8,$BF
	.dc.b	$E0,$EF,$1A,$4F,$F8,$5E,$32,$FF,$86,$63,$2F,$F8,$22,$32,$FF,$82
	.dc.b	$A3,$38,$FF,$85,$A7,$33,$FD,$C9,$FE,$DE,$47,$FB,$BB,$5F,$E4,$8D
	.dc.b	$78,$1A,$21,$A2,$C7,$77,$D0,$7F,$8B,$FC,$5F,$2C,$66,$3D,$06,$4D
	.dc.b	$98,$23,$FE,$2E,$8F,$F1,$8F,$8B,$FE,$32,$F1,$25,$7F,$8E,$E3,$2F
	.dc.b	$F8,$F6,$32,$FF,$8B,$23,$2F,$F8,$BA,$33,$5F,$E3,$A9,$CB,$FE,$3C
	.dc.b	$8C,$E9,$FE,$5E,$9C,$BF,$E6,$08,$CD,$7F,$97,$67,$2F,$F9,$7E,$33
	.dc.b	$3F,$E5,$B9,$FF,$24,$F8,$B2,$BF,$CA,$4F,$F9,$33,$C5,$C4,$7F,$90
	.dc.b	$A2,$E1,$FE,$4E,$9B,$7F,$E7,$09,$B5,$FE,$6D,$9B,$3F,$E6,$99,$BB
	.dc.b	$26,$C6,$1B,$5E,$23,$FC,$C8,$FF,$97,$BC,$4F,$F2,$FF,$8B,$27,$FC
	.dc.b	$D1,$19,$7F,$CD,$51,$97,$FC,$B3,$19,$7F,$CB,$71,$9C,$7F,$CC,$F3
	.dc.b	$99,$FF,$21,$BF,$E4,$02,$3F,$E4,$86,$BF,$DD,$5A,$F0,$34,$43,$45
	.dc.b	$8E,$EF,$A0,$FF,$43,$7C,$AF,$F9,$B7,$2F,$F4,$75,$79,$EB,$75,$FE
	.dc.b	$74,$AD,$64,$AD,$5F,$FA,$81,$FF,$46,$4F,$33,$FD,$2F,$5C,$7C,$3C
	.dc.b	$9A,$F1,$64,$78,$A3,$2F,$92,$32,$F1,$34,$64,$86,$49,$8F,$FA,$A2
	.dc.b	$BC,$2D,$9E,$B3,$AC,$3F,$D7,$42,$B6,$E1,$B1,$FE,$BD,$57,$CA,$FA
	.dc.b	$4F,$CA,$F1,$7E,$09,$1F,$9B,$E2,$D7,$E5,$45,$9F,$34,$DC,$3C,$13
	.dc.b	$6B,$C1,$36,$7C,$D3,$70,$4D,$8C,$36,$BC,$87,$88,$FB,$FE,$2F,$D6
	.dc.b	$F1,$64,$60,$87,$DF,$72,$FE,$06,$3E,$E9,$8C,$7E,$74,$E2,$78,$9F
	.dc.b	$EE,$0C,$7E,$65,$7F,$A8,$B1,$34,$93,$FE,$A8,$8C,$B6,$32,$FF,$A7
	.dc.b	$E3,$2D,$8C,$BD,$E3,$2D,$8F,$23,$E9,$3F,$3F,$E2,$63,$0F,$C3,$4D
	.dc.b	$BF,$C0,$4D,$AF,$C0,$4D,$9F,$C3,$4F,$E7,$BC,$5F,$61,$2B,$D4,$D7
	.dc.b	$D0,$D7,$CC,$D7,$F1,$CD,$7F,$1C,$FC,$7F,$12,$47,$C5,$6B,$F0,$DA
	.dc.b	$FC,$16,$BF,$99,$6B,$F9,$96,$D7,$E7,$FC,$5F,$A8,$49,$FA,$4D,$7E
	.dc.b	$83,$5F,$9C,$D7,$F4,$CD,$7F,$4C,$FD,$FF,$17,$F8,$1F,$89,$27,$F8
	.dc.b	$7C,$7F,$06,$30,$FF,$0D,$8C,$3A,$3F,$8A,$BA,$CF,$F7,$93,$A9,$FC
	.dc.b	$23,$DD,$FC,$53,$5C,$98,$21,$82,$1F,$89,$78,$34,$F1,$1F,$C8,$5F
	.dc.b	$E3,$BE,$2F,$86,$7C,$73,$A7,$F8,$84,$67,$0F,$E1,$55,$6B,$F8,$54
	.dc.b	$FF,$26,$F1,$7F,$93,$F8,$B2,$7F,$99,$C6,$5F,$E6,$B1,$97,$F9,$94
	.dc.b	$65,$FE,$69,$19,$C3,$F9,$04,$E6,$7F,$48,$FE,$C4,$8F,$EB,$DA,$FE
	.dc.b	$B5,$AF,$03,$44,$34,$58,$EE,$FA,$0F,$E9,$7E,$27,$96,$33,$1E,$83
	.dc.b	$26,$CC,$11,$FE,$71,$1F,$D2,$BC,$5F,$E9,$7E,$2C,$AF,$EA,$F1,$97
	.dc.b	$FA,$CC,$65,$FE,$AD,$19,$7F,$AC,$46,$6B,$FA,$24,$E6,$7F,$84,$3F
	.dc.b	$C5,$48,$FF,$13,$6B,$FC,$45,$AF,$03,$44,$34,$58,$EE,$FA,$0F,$ED
	.dc.b	$FE,$27,$96,$33,$1E,$83,$26,$CC,$11,$FE,$B9,$1F,$DB,$BC,$5F,$ED
	.dc.b	$FE,$2C,$AF,$EF,$71,$97,$FB,$EC,$65,$FE,$F5,$19,$7F,$BE,$46,$6B
	.dc.b	$FB,$44,$E6,$7F,$96,$3F,$CE,$48,$FF,$37,$6B,$FC,$D5,$AF,$03,$44
	.dc.b	$34,$58,$EE,$FA,$0F,$F0,$9F,$8B,$E5,$8C,$C7,$A0,$C9,$B3,$04,$7F
	.dc.b	$C0,$11,$FE,$12,$F1,$3F,$C2,$7E,$2C,$AF,$F0,$FC,$65,$FF,$10,$C6
	.dc.b	$5F,$F0,$F4,$65,$FF,$10,$46,$6B,$FC,$19,$39,$9F,$EA,$0F,$F5,$D2
	.dc.b	$3F,$D6,$DA,$FF,$59,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$FC,$6F,$E2
	.dc.b	$F9,$63,$31,$E8,$32,$6C,$C1,$1F,$F1,$24,$7F,$8D,$BC,$4F,$F1,$BF
	.dc.b	$8B,$2B,$FC,$87,$19,$7F,$C8,$B1,$97,$FC,$85,$19,$7F,$C8,$91,$9A
	.dc.b	$FF,$18,$4E,$67,$FB,$A3,$FE,$01,$23,$FD,$FD,$AF,$F7,$D6,$BC,$0D
	.dc.b	$10,$D1,$63,$BB,$E8,$3F,$CB,$7E,$2F,$96,$33,$1E,$83,$26,$CC,$11
	.dc.b	$FF,$24,$47,$F9,$6B,$C5,$FF,$2D,$F8,$72,$BF,$CC,$F1,$97,$FC,$D3
	.dc.b	$19,$7F,$CC,$D1,$97,$FC,$D1,$19,$AF,$F2,$A4,$E6,$7F,$C3,$0F,$F8
	.dc.b	$94,$8F,$F8,$8D,$AF,$F8,$85,$AF,$03,$44,$34,$58,$EE,$FA,$0F,$F3
	.dc.b	$FF,$8B,$E5,$8C,$C7,$A0,$C9,$B3,$04,$7F,$CD,$90,$89,$0E,$8F,$F3
	.dc.b	$5F,$EE,$FF,$E8,$B3,$FC,$D2,$9A,$67,$C7,$E5,$4A,$E4,$F3,$9D,$F8
	.dc.b	$DA,$9C,$DF,$3B,$C3,$E0,$6D,$F1,$AB,$3E,$16,$DF,$86,$7C,$27,$12
	.dc.b	$9E,$1A,$AD,$33,$5E,$16,$CA,$77,$39,$9E,$AD,$7F,$CC,$93,$B7,$8C
	.dc.b	$F0,$8C,$47,$06,$BC,$23,$D2,$47,$E8,$CF,$D0,$F1,$B1,$1F,$2E,$3D
	.dc.b	$8D,$3D,$45,$7E,$A4,$70,$C3,$E6,$DF,$C5,$AD,$3F,$EA,$07,$B9,$F5
	.dc.b	$7C,$6F,$97,$E3,$48,$CE,$1F,$7E,$7E,$79,$98,$E7,$38,$77,$49,$D5
	.dc.b	$F8,$AB,$55,$E2,$9F,$8C,$47,$E8,$B5,$F9,$0D,$7E,$7B,$5F,$CA,$B5
	.dc.b	$F5,$1C,$8F,$1B,$E0,$9F,$87,$E2,$5C,$38,$D6,$8F,$4C,$65,$E6,$79
	.dc.b	$CC,$9E,$EF,$1B,$F2,$3E,$2C,$8C,$10,$F9,$EE,$5F,$A3,$1F,$98,$31
	.dc.b	$7F,$5A,$71,$3C,$4F,$C6,$CF,$E2,$C8,$FE,$59,$AF,$E3,$5A,$F0,$34
	.dc.b	$43,$45,$9F,$C2,$C7,$82,$3F,$0B,$E2,$D9,$CC,$7A,$DF,$C5,$C6,$70
	.dc.b	$FE,$0B,$56,$BF,$7D,$3F,$B4,$F1,$7F,$6B,$E2,$C9,$FC,$2C,$65,$FC
	.dc.b	$3C,$65,$FC,$AC,$65,$FC,$BC,$67,$0F,$CD,$CE,$67,$F6,$2F,$EB,$C8
	.dc.b	$FE,$E9,$AF,$EC,$5A,$F0,$34,$43,$45,$8E,$EF,$A0,$FC,$EF,$89,$E5
	.dc.b	$8C,$C7,$A0,$C9,$B3,$04,$7F,$67,$1F,$C6,$3C,$5F,$E3,$3E,$2C,$AF
	.dc.b	$D4,$C6,$5F,$D5,$C6,$5F,$DC,$C6,$5F,$DD,$C6,$6B,$F7,$F3,$99,$FE
	.dc.b	$2A,$FF,$13,$23,$FC,$79,$AF,$F1,$56,$BC,$0D,$10,$D1,$63,$BB,$E8
	.dc.b	$3F,$80,$F8,$9E,$58,$CC,$7A,$0C,$9B,$30,$47,$F8,$BC,$7F,$3C,$F1
	.dc.b	$7F,$9E,$F8,$B2,$BF,$86,$C6,$5F,$E1,$F1,$97,$F8,$EC,$65,$FE,$3F
	.dc.b	$19,$AF,$E4,$73,$99,$FE,$72,$FF,$37,$23,$FD,$09,$AF,$F3,$96,$BC
	.dc.b	$0D,$10,$D1,$63,$BB,$E8,$3F,$92,$F8,$9E,$58,$CC,$7A,$0C,$9B,$30
	.dc.b	$47,$F9,$DC,$7F,$60,$F1,$7F,$B0,$F8,$B2,$BF,$98,$C6,$5F,$E6,$71
	.dc.b	$97,$FA,$0C,$65,$FE,$87,$19,$AF,$E8,$F3,$99,$FE,$BA,$FF,$5B,$23
	.dc.b	$FD,$99,$AF,$F5,$D6,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$A4,$F8,$9E
	.dc.b	$58,$CC,$7A,$0C,$9B,$30,$47,$FA,$FC,$7F,$81,$3C,$5F,$F0,$2F,$8B
	.dc.b	$2B,$FA,$AC,$65,$FE,$AF,$19,$7F,$B2,$C6,$5F,$EC,$F1,$9A,$FE,$D7
	.dc.b	$39,$9F,$F0,$0B,$FD,$FC,$8F,$F8,$29,$AF,$F8,$05,$AF,$03,$44,$34
	.dc.b	$58,$EE,$FA,$0F,$ED,$BE,$27,$96,$33,$1E,$83,$26,$CC,$11,$FF,$01
	.dc.b	$C7,$F8,$A3,$C5,$FF,$14,$F8,$B2,$BF,$BC,$C6,$5F,$EF,$71,$97,$FC
	.dc.b	$13,$19,$7F,$C1,$71,$9A,$FF,$07,$CE,$67,$FC,$4A,$FF,$88,$C8,$FF
	.dc.b	$8B,$9A,$FF,$89,$5A,$F0,$34,$43,$45,$8E,$EF,$A0,$FF,$08,$F8,$9E
	.dc.b	$58,$CC,$7A,$0C,$9B,$30,$47,$FC,$4F,$1F,$E4,$CF,$17,$FC,$9B,$E2
	.dc.b	$CA,$FF,$0E,$C6,$5F,$F0,$FC,$65,$FF,$16,$C6,$5F,$F1,$7C,$66,$BF
	.dc.b	$C6,$73,$99,$FF,$24,$BF,$E4,$72,$3F,$E5,$26,$BF,$E4,$96,$BC,$0D
	.dc.b	$10,$D1,$63,$BB,$E8,$3F,$C6,$BE,$27,$96,$33,$1E,$83,$26,$CC,$11
	.dc.b	$FF,$25,$C7,$F9,$C3,$C5,$FF,$38,$F8,$B2,$BF,$C8,$31,$97,$FC,$87
	.dc.b	$19,$7F,$CA,$31,$97,$FC,$A7,$19,$AF,$F2,$BC,$E6,$7F,$CD,$AF,$F9
	.dc.b	$AC,$8F,$F9,$D9,$AF,$F9,$B5,$AF,$03,$44,$34,$58,$EE,$FA,$0F,$F2
	.dc.b	$FF,$EF,$1F,$E9,$F9,$F3,$4E,$A7,$FC,$DD,$39,$78,$1E,$C3,$8F,$8B
	.dc.b	$FE,$96,$87,$8D,$FE,$8A,$6C,$FF,$A4,$1A,$FF,$A3,$1A,$FF,$AA,$BC
	.dc.b	$04,$62,$3F,$D6,$D1,$89,$E8,$48,$E8,$D7,$A3,$5E,$93,$FE,$81,$7C
	.dc.b	$1E,$37,$94,$8F,$95,$AE,$8C,$10,$C1,$0E,$AF,$BF,$C4,$91,$EF,$6B
	.dc.b	$EF,$6B,$E4,$6B,$EE,$9F,$F4,$53,$C4,$87,$13,$E5,$78,$BF,$34,$8E
	.dc.b	$08,$7C,$D7,$2D,$32,$7C,$47,$2D,$3E,$B7,$8B,$F6,$48,$E0,$87,$D9
	.dc.b	$72,$D3,$27,$D0,$72,$D3,$F0,$5E,$27,$E1,$48,$E0,$87,$E1,$5C,$B4
	.dc.b	$C9,$F7,$1C,$B4,$FC,$77,$8B,$F9,$3F,$16,$47,$E7,$A3,$F2,$C6,$1F
	.dc.b	$CE,$C6,$1D,$1F,$A4,$75,$B7,$FB,$E3,$15,$6E,$BF,$34,$FE,$7C,$8F
	.dc.b	$88,$86,$8A,$7E,$51,$DB,$92,$19,$3C,$67,$EC,$9F,$D7,$78,$9E,$1A
	.dc.b	$F2,$45,$DB,$F1,$53,$69,$1F,$AC,$9C,$3F,$AE,$9F,$DC,$78,$BF,$B9
	.dc.b	$F1,$65,$FF,$04,$8C,$BF,$C1,$A3,$2F,$F0,$48,$CB,$FC,$1A,$33,$87
	.dc.b	$C7,$9C,$CF,$EC,$9F,$E0,$84,$7F,$82,$35,$FE,$08,$D7,$81,$A2,$1A
	.dc.b	$2C,$77,$7D,$07,$F1,$CF,$13,$CB,$19,$8F,$41,$93,$66,$08,$FD,$F8
	.dc.b	$FE,$39,$E2,$FF,$1D,$F1,$65,$7F,$28,$8C,$BF,$CA,$A3,$2F,$F2,$88
	.dc.b	$CB,$FC,$AA,$33,$5F,$8B,$9C,$CF,$F1,$67,$F9,$41,$1F,$E5,$0D,$7F
	.dc.b	$94,$35,$E0,$68,$86,$8B,$1D,$DF,$41,$FD,$03,$C4,$F2,$C6,$63,$D0
	.dc.b	$64,$D9,$82,$3F,$A3,$8F,$E8,$1E,$2F,$F4,$1F,$16,$57,$F4,$C8,$CB
	.dc.b	$FD,$3A,$32,$FF,$4C,$8C,$BF,$D3,$A3,$35,$FA,$F9,$CC,$FF,$3A,$7F
	.dc.b	$A6,$11,$FE,$98,$D7,$FA,$63,$5E,$06,$88,$68,$B1,$DD,$F4,$1F,$D9
	.dc.b	$3C,$4F,$2C,$66,$3D,$06,$4D,$98,$23,$FC,$1E,$3F,$B2,$78,$BF,$D9
	.dc.b	$7C,$59,$5F,$DC,$23,$2F,$F7,$28,$CB,$FD,$C2,$32,$FF,$72,$8C,$D7
	.dc.b	$F1,$39,$CC,$FF,$5E,$7F,$B8,$11,$FE,$E0,$D7,$FB,$83,$5E,$06,$88
	.dc.b	$68,$B1,$DD,$F4,$1F,$E0,$8F,$13,$CB,$19,$8F,$41,$93,$66,$08,$FF
	.dc.b	$2B,$8F,$F0,$47,$8B,$FE,$09,$F1,$65,$7F,$85,$23,$2F,$F8,$5A,$32
	.dc.b	$FF,$85,$23,$2F,$F8,$5A,$33,$5F,$CD,$E7,$33,$FE,$02,$7F,$C2,$84
	.dc.b	$7F,$C2,$8D,$7F,$C2,$8D,$78,$1A,$21,$A2,$C7,$77,$D0,$7F,$8B,$3C
	.dc.b	$4F,$2C,$66,$3D,$06,$4D,$98,$23,$FD,$3E,$3F,$C5,$9E,$2F,$F8,$B7
	.dc.b	$C5,$95,$FE,$38,$8C,$BF,$E3,$A8,$CB,$FE,$38,$8C,$BF,$E3,$A8,$CD
	.dc.b	$7F,$5B,$9C,$CF,$F8,$99,$FF,$1C,$11,$FF,$1C,$35,$FF,$1C,$35,$E0
	.dc.b	$68,$86,$8B,$1D,$DF,$41,$FE,$50,$F1,$3C,$B1,$98,$F4,$19,$36,$60
	.dc.b	$8F,$F7,$38,$FF,$28,$78,$BF,$E5,$1F,$16,$57,$F9,$72,$32,$FF,$97
	.dc.b	$A3,$2F,$F9,$72,$32,$FF,$97,$A3,$35,$FD,$FE,$73,$3F,$E4,$A7,$FC
	.dc.b	$B8,$47,$FC,$B8,$D7,$FC,$B8,$D7,$81,$A2,$1A,$2C,$77,$7D,$07,$F9
	.dc.b	$D3,$C4,$F2,$C6,$63,$D0,$64,$D9,$82,$3F,$E1,$78,$FF,$3A,$78,$A9
	.dc.b	$93,$FE,$04,$7F,$C0,$BF,$BC,$BC,$6D,$42,$75,$DA,$AD,$4F,$FA,$22
	.dc.b	$FA,$BD,$27,$FC,$EF,$3A,$8E,$51,$8B,$DC,$7F,$A3,$2B,$BD,$F6,$BE
	.dc.b	$0D,$AD,$BF,$E3,$19,$FF,$46,$12,$BF,$D1,$8D,$7F,$D1,$8D,$7F,$D0
	.dc.b	$2D,$7F,$CD,$CD,$7F,$C4,$BD,$E8,$E5,$3D,$E7,$DC,$7B,$4F,$8F,$E3
	.dc.b	$79,$31,$E7,$1A,$3D,$E6,$8C,$4F,$37,$E0,$1C,$E3,$E6,$4F,$C2,$7E
	.dc.b	$4F,$8D,$67,$EA,$F8,$D6,$7E,$93,$73,$FF,$4E,$61,$F2,$6F,$13,$2F
	.dc.b	$B5,$6A,$FA,$C6,$27,$4F,$A2,$3E,$F5,$75,$3F,$07,$1F,$EB,$5A,$FF
	.dc.b	$55,$36,$32,$F4,$37,$3C,$9C,$B5,$CC,$F2,$78,$1C,$CD,$B9,$7F,$2B
	.dc.b	$3D,$9A,$69,$F3,$3E,$E7,$E2,$78,$D8,$8F,$64,$FC,$17,$89,$88,$D6
	.dc.b	$1F,$26,$39,$1F,$22,$3E,$77,$8D,$F3,$FC,$69,$3F,$32,$32,$FC,$E8
	.dc.b	$CB,$E0,$8C,$BE,$28,$CD,$7E,$7E,$73,$3F,$35,$F7,$63,$FA,$F6,$E0
	.dc.b	$69,$27,$E3,$B5,$F9,$AD,$7A,$1A,$21,$E5,$3D,$47,$E1,$BC,$5F,$3D
	.dc.b	$F9,$63,$27,$CA,$8C,$DF,$EF,$AA,$CF,$F0,$A9,$FC,$A7,$8D,$F9,$5F
	.dc.b	$1A,$4F,$E4,$A3,$2F,$E5,$23,$2F,$D6,$8C,$BF,$62,$33,$87,$F0,$59
	.dc.b	$CC,$FE,$4D,$F8,$A4,$7F,$12,$D7,$F1,$ED,$78,$1A,$21,$A2,$C7,$77
	.dc.b	$D0,$7E,$AB,$C5,$F2,$C6,$63,$D0,$64,$D9,$82,$3F,$C7,$63,$F7,$1E
	.dc.b	$37,$EE,$7C,$69,$5F,$B6,$8C,$BF,$B8,$8C,$BF,$9A,$8C,$BF,$9C,$8C
	.dc.b	$D7,$F2,$99,$CC,$FE,$DD,$FC,$49,$1F,$D6,$B5,$FD,$9B,$5E,$06,$88
	.dc.b	$68,$B1,$DD,$F4,$1F,$C3,$BC,$5F,$2C,$66,$3D,$06,$4D,$98,$23,$FD
	.dc.b	$06,$3F,$8E,$78,$DF,$C7,$7C,$69,$5F,$C6,$A3,$2F,$F1,$C8,$CB,$FB
	.dc.b	$E8,$CB,$FC,$02,$33,$5F,$D3,$67,$33,$FC,$6D,$FD,$69,$1F,$E2,$2D
	.dc.b	$7F,$8B,$B5,$E0,$68,$86,$8B,$1D,$DF,$41,$FC,$CB,$C5,$F2,$C6,$63
	.dc.b	$D0,$64,$D9,$82,$3F,$D9,$63,$FA,$07,$8D,$FD,$07,$C6,$95,$FC,$FA
	.dc.b	$32,$FF,$40,$8C,$BF,$C8,$A3,$2F,$F2,$48,$CD,$7F,$71,$9C,$CF,$F3
	.dc.b	$F7,$F8,$89,$1F,$E6,$AD,$7F,$9D,$B5,$E0,$68,$86,$8B,$1D,$DF,$41
	.dc.b	$FD,$5B,$C5,$F2,$C6,$63,$D0,$64,$D9,$82,$3F,$E0,$98,$FE,$C9,$E3
	.dc.b	$7F,$65,$F1,$A5,$7F,$62,$8C,$BF,$D9,$23,$2F,$F4,$68,$CB,$FD,$22
	.dc.b	$33,$5F,$E1,$59,$CC,$FF,$63,$7F,$9A,$91,$FE,$B2,$D7,$FA,$FB,$5E
	.dc.b	$06,$88,$68,$B1,$DD,$F4,$1F,$DE,$BC,$5F,$2C,$66,$3D,$06,$4D,$98
	.dc.b	$23,$FE,$2D,$8F,$F0,$47,$8D,$FE,$09,$F1,$A5,$7F,$81,$A3,$2F,$F8
	.dc.b	$22,$32,$FF,$6A,$8C,$BF,$DB,$23,$35,$FE,$39,$9C,$CF,$F8,$1D,$FE
	.dc.b	$B2,$47,$FB,$EB,$5F,$F0,$1B,$5E,$06,$88,$68,$B1,$DD,$F4,$1F,$E1
	.dc.b	$EF,$17,$CB,$19,$8F,$41,$93,$66,$08,$FF,$94,$63,$FC,$59,$E3,$7F
	.dc.b	$8B,$7C,$69,$5F,$E2,$A8,$CB,$FE,$2C,$8C,$BF,$E0,$E8,$CB,$FE,$10
	.dc.b	$8C,$D7,$F9,$76,$73,$3F,$E2,$B7,$FB,$E9,$1F,$F1,$0B,$5F,$F1,$3B
	.dc.b	$5E,$06,$88,$68,$B1,$DD,$F4,$1F,$E4,$2F,$17,$CB,$19,$8F,$41,$93
	.dc.b	$66,$08,$FF,$9D,$63,$FC,$A1,$E3,$7F,$85,$09,$3F,$E3,$86,$BF,$E3
	.dc.b	$E6,$BF,$E4,$96,$BF,$E8,$66,$BF,$E9,$27,$FC,$3F,$E3,$48,$FF,$21
	.dc.b	$B5,$FF,$24,$B5,$FF,$2A,$35,$FF,$48,$B5,$FF,$4E,$B7,$B7,$FD,$35
	.dc.b	$E2,$7F,$9A,$49,$3F,$E7,$39,$FF,$48,$CF,$FA,$6A,$73,$3F,$E4,$A7
	.dc.b	$FC,$E5,$3A,$8F,$F5,$4C,$6B,$0F,$F4,$D3,$AF,$1B,$17,$CE,$BF,$CE
	.dc.b	$73,$96,$90,$EC,$78,$E7,$8B,$D8,$D5,$7F,$AD,$2F,$C5,$65,$E2,$FF
	.dc.b	$92,$9B,$5F,$E7,$26,$BF,$E7,$66,$BF,$EC,$09,$E5,$3C,$49,$1C,$5A
	.dc.b	$F1,$6B,$C6,$79,$BF,$EC,$F9,$F2,$78,$9E,$32,$47,$8D,$AF,$8D,$AF
	.dc.b	$36,$BC,$E7,$FD,$09,$3D,$23,$91,$ED,$F1,$7E,$01,$1C,$10,$F8,$0E
	.dc.b	$5A,$64,$F5,$39,$69,$F2,$7C,$5F,$98,$47,$04,$3E,$63,$96,$99,$3E
	.dc.b	$33,$96,$9F,$57,$C3,$FC,$6F,$8B,$23,$F2,$91,$F9,$03,$0F,$FA,$DF
	.dc.b	$04,$E6,$36,$7E,$6D,$DE,$7F,$87,$AD,$4F,$E0,$5F,$C3,$11,$F1,$10
	.dc.b	$D1,$4F,$BC,$ED,$C9,$0C,$9E,$23,$CC,$FE,$A3,$C4,$F0,$CF,$8E,$74
	.dc.b	$FE,$6E,$33,$97,$DF,$AB,$48,$FC,$F4,$FE,$CB,$C3,$FD,$9F,$89,$2B
	.dc.b	$F7,$51,$97,$F7,$91,$97,$E4,$46,$5F,$95,$19,$C3,$F5,$93,$99,$FC
	.dc.b	$E3,$FA,$B2,$3F,$AB,$6B,$FA,$B6,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F
	.dc.b	$89,$78,$9E,$58,$CC,$7A,$0C,$9B,$30,$47,$F8,$24,$7F,$16,$F0,$FF
	.dc.b	$8B,$F8,$92,$BF,$8F,$46,$5F,$E4,$11,$97,$F1,$91,$97,$F1,$D1,$9A
	.dc.b	$FE,$21,$39,$9F,$E0,$0F,$F0,$F2,$3F,$C3,$DA,$FF,$0F,$6B,$C0,$D1
	.dc.b	$0D,$16,$3B,$BE,$83,$F9,$B7,$89,$E5,$8C,$C7,$A0,$C9,$B3,$04,$7F
	.dc.b	$94,$47,$F3,$AF,$0F,$F9,$DF,$89,$2B,$FA,$14,$65,$FE,$89,$19,$7F
	.dc.b	$61,$19,$7F,$65,$19,$AF,$E6,$93,$99,$FE,$48,$FF,$33,$23,$FC,$CD
	.dc.b	$AF,$F3,$36,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$AD,$78,$9E,$58,$CC
	.dc.b	$7A,$0C,$9B,$30,$47,$FA,$64,$7F,$5E,$F0,$FF,$AF,$F8,$92,$BF,$B3
	.dc.b	$46,$5F,$ED,$11,$97,$F8,$A4,$65,$FE,$2D,$19,$AF,$EB,$13,$99,$FE
	.dc.b	$90,$FF,$57,$23,$FD,$5D,$AF,$F5,$76,$BC,$0D,$10,$D1,$63,$BB,$E8
	.dc.b	$3F,$BF,$78,$9E,$58,$CC,$7A,$0C,$9B,$30,$47,$FB,$84,$7F,$80,$BC
	.dc.b	$3F,$F0,$1F,$89,$2B,$FC,$15,$19,$7F,$C1,$91,$97,$F9,$C4,$65,$FE
	.dc.b	$75,$19,$AF,$EF,$93,$99,$FE,$D8,$FF,$7B,$23,$FD,$ED,$AF,$F7,$B6
	.dc.b	$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$C4,$5E,$27,$96,$33,$1E,$83,$26
	.dc.b	$CC,$11,$FF,$0A,$47,$F8,$9B,$C3,$FF,$13,$F8,$92,$BF,$C5,$D1,$97
	.dc.b	$FC,$61,$19,$7F,$AE,$46,$5F,$EB,$D1,$9A,$FF,$10,$4E,$67,$FC,$20
	.dc.b	$FF,$87,$C8,$FF,$87,$DA,$FF,$87,$DA,$F0,$34,$43,$45,$8E,$EF,$A0
	.dc.b	$FF,$23,$78,$9E,$58,$CC,$7A,$0C,$9B,$30,$47,$FC,$71,$1F,$E4,$AF
	.dc.b	$0F,$FC,$97,$E2,$4A,$FF,$29,$46,$5F,$F2,$A4,$65,$FF,$00,$46,$5F
	.dc.b	$F0,$14,$66,$BF,$C8,$93,$99,$FF,$1A,$3F,$E4,$32,$3F,$E4,$36,$BF
	.dc.b	$E4,$36,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$CD,$5E,$27,$96,$33,$1E
	.dc.b	$83,$26,$CC,$11,$FF,$2E,$47,$F9,$BB,$C4,$FF,$37,$F7,$4A,$FF,$3B
	.dc.b	$46,$5F,$F3,$C4,$65,$FF,$12,$46,$5F,$F1,$34,$66,$BF,$CD,$13,$99
	.dc.b	$FF,$2C,$3F,$E6,$72,$3F,$E6,$76,$BF,$E6,$76,$BC,$0D,$10,$D1,$63
	.dc.b	$BB,$E8,$3F,$D1,$DE,$27,$96,$33,$1E,$83,$26,$CC,$11,$FF,$40,$42
	.dc.b	$24,$7F,$89,$5F,$F1,$37,$EF,$5F,$E8,$43,$FC,$FF,$19,$EE,$46,$1F
	.dc.b	$F3,$B3,$BF,$17,$A5,$2B,$84,$EE,$3A,$F8,$BA,$9C,$D5,$BE,$6D,$AD
	.dc.b	$9E,$68,$EF,$59,$7C,$73,$6C,$F8,$AB,$FD,$29,$18,$70,$7B,$A2,$B4
	.dc.b	$CC,$F4,$6C,$F1,$7A,$46,$5B,$19,$7B,$CF,$49,$D4,$60,$E7,$19,$9F
	.dc.b	$91,$1E,$D3,$C4,$7C,$99,$F9,$BE,$2F,$B6,$A4,$EE,$7C,$58,$7C,$EA
	.dc.b	$CC,$F1,$39,$C7,$CE,$78,$5A,$7E,$47,$8B,$F2,$7C,$59,$5A,$8C,$BC
	.dc.b	$23,$2F,$5C,$3F,$07,$3A,$8C,$5A,$B7,$87,$E0,$AB,$55,$F5,$1F,$84
	.dc.b	$47,$E7,$B5,$F9,$AD,$78,$1A,$21,$A2,$C7,$77,$D4,$7C,$0F,$13,$CD
	.dc.b	$19,$8F,$3C,$FE,$26,$73,$87,$E7,$22,$CF,$E9,$27,$F1,$DE,$2F,$E3
	.dc.b	$FC,$59,$3F,$32,$32,$FC,$E8,$CB,$F1,$23,$2F,$C6,$8C,$E1,$FA,$69
	.dc.b	$CC,$FE,$61,$FC,$31,$1F,$C9,$B5,$FC,$8B,$5E,$06,$88,$68,$B1,$DD
	.dc.b	$F4,$1F,$84,$F1,$3C,$B1,$98,$F4,$19,$36,$60,$8F,$EF,$A3,$F6,$5E
	.dc.b	$2F,$EC,$FC,$59,$5F,$90,$8C,$BF,$92,$8C,$BF,$88,$8C,$BF,$8A,$8C
	.dc.b	$D7,$F0,$A9,$CC,$FE,$F1,$FD,$51,$1F,$DB,$B5,$FD,$AB,$5E,$06,$88
	.dc.b	$68,$B1,$DD,$F4,$1F,$A8,$F1,$3C,$B1,$98,$F4,$19,$36,$60,$8F,$F2
	.dc.b	$28,$FE,$2D,$E2,$FF,$17,$F1,$65,$7E,$D2,$32,$FE,$DA,$32,$FE,$B2
	.dc.b	$32,$FE,$BA,$33,$5F,$CB,$A7,$33,$FC,$81,$FE,$1C,$47,$F8,$DB,$5F
	.dc.b	$E3,$2D,$78,$1A,$21,$A2,$C7,$77,$D0,$7F,$0C,$F1,$3C,$B1,$98,$F4
	.dc.b	$19,$36,$60,$8F,$F4,$68,$FE,$75,$E2,$FF,$3B,$F1,$65,$7F,$18,$8C
	.dc.b	$BF,$C6,$A3,$2F,$F1,$08,$CB,$FC,$4A,$33,$5F,$D4,$A7,$33,$FD,$11
	.dc.b	$FE,$64,$47,$F9,$FB,$5F,$E7,$AD,$78,$1A,$21,$A2,$C7,$77,$D0,$7F
	.dc.b	$30,$F1,$3C,$B1,$98,$F4,$19,$36,$60,$8F,$F6,$A8,$FE,$BD,$E2,$FF
	.dc.b	$5F,$F1,$65,$7F,$3C,$8C,$BF,$CF,$A3,$2F,$F3,$48,$CB,$FC,$DA,$33
	.dc.b	$5F,$DD,$A7,$33,$FD,$A1,$FE,$AC,$47,$FB,$1B,$5F,$EC,$2D,$78,$1A
	.dc.b	$21,$A2,$C7,$77,$D0,$7F,$54,$F1,$3C,$B1,$98,$F4,$19,$36,$60,$8F
	.dc.b	$F8,$3A,$3F,$C0,$5E,$2F,$F8,$0F,$C5,$95,$FD,$82,$32,$FF,$62,$8C
	.dc.b	$BF,$D6,$23,$2F,$F5,$A8,$CD,$7F,$86,$A7,$33,$FE,$0C,$7F,$BD,$11
	.dc.b	$FF,$03,$B5,$FF,$02,$B5,$E0,$68,$86,$8B,$1D,$DF,$41,$FD,$E3,$C4
	.dc.b	$F2,$C6,$63,$D0,$64,$D9,$82,$3F,$E3,$28,$FF,$13,$78,$BF,$E2,$7F
	.dc.b	$16,$57,$F8,$12,$32,$FF,$81,$A3,$2F,$F7,$C8,$CB,$FD,$FA,$33,$5F
	.dc.b	$E3,$E9,$CC,$FF,$8C,$1F,$F0,$F1,$1F,$F1,$5B,$5F,$F1,$4B,$5E,$06
	.dc.b	$88,$68,$B1,$DD,$F4,$1F,$E1,$CF,$13,$CB,$19,$8F,$41,$93,$66,$08
	.dc.b	$FF,$95,$A3,$FC,$95,$E2,$FF,$8D,$49,$3F,$E4,$26,$BF,$E4,$06,$BF
	.dc.b	$E4,$E6,$BF,$E6,$46,$BF,$E6,$C7,$FC,$81,$E2,$C8,$FF,$26,$B5,$FF
	.dc.b	$25,$B5,$FF,$2C,$B5,$FF,$37,$B5,$FF,$3B,$B7,$B7,$FC,$EB,$E2,$7F
	.dc.b	$93,$49,$3F,$E5,$A6,$BF,$E5,$86,$BF,$E6,$66,$BF,$E7,$C6,$BF,$E8
	.dc.b	$47,$FC,$B1,$E2,$C8,$FF,$32,$B5,$FF,$31,$B5,$FF,$38,$B5,$FF,$43
	.dc.b	$B5,$FF,$47,$B6,$7F,$D1,$BE,$2F,$F9,$94,$93,$FE,$72,$6B,$FE,$70
	.dc.b	$6B,$FE,$7E,$6B,$FE,$94,$6B,$FE,$9C,$7F,$CE,$9F,$BD,$FF,$9D,$E3
	.dc.b	$FD,$45,$28,$C4,$23,$1E,$2F,$02,$47,$FA,$1A,$77,$3A,$8C,$4E,$9F
	.dc.b	$F4,$3C,$E8,$91,$E2,$F1,$BF,$D3,$51,$FE,$73,$9F,$F3,$E3,$5F,$F3
	.dc.b	$73,$FE,$A9,$AF,$34,$F0,$F1,$A5,$70,$8E,$66,$1E,$31,$87,$47,$A6
	.dc.b	$3C,$4D,$21,$E8,$72,$D3,$27,$FA,$FA,$7C,$6E,$0F,$97,$49,$E3,$B7
	.dc.b	$FB,$26,$3D,$44,$F1,$3C,$95,$EE,$AB,$5B,$9B,$7D,$26,$D7,$D5,$B5
	.dc.b	$7C,$AF,$17,$E5,$F8,$B2,$BE,$5C,$65,$F9,$B1,$97,$E6,$C6,$5F,$9F
	.dc.b	$19,$C7,$D1,$39,$9F,$81,$7F,$E9,$D7,$F0,$51,$F8,$C2,$38,$21,$F8
	.dc.b	$27,$2D,$32,$74,$73,$3E,$08,$F4,$9F,$84,$F1,$3C,$D1,$98,$F4,$99
	.dc.b	$36,$60,$8F,$D1,$8F,$C7,$F8,$BF,$90,$F1,$65,$7E,$42,$32,$FE,$4A
	.dc.b	$32,$FE,$4A,$32,$FE,$52,$33,$5F,$76,$73,$3F,$19,$FD,$79,$1F,$D4
	.dc.b	$35,$FD,$2B,$5E,$06,$88,$68,$B1,$DD,$F4,$1F,$A8,$F1,$3C,$B1,$98
	.dc.b	$F4,$19,$36,$60,$8F,$E5,$63,$F6,$7E,$2F,$ED,$3C,$59,$5F,$B4,$8C
	.dc.b	$BF,$B6,$8C,$BF,$B6,$8C,$BF,$B8,$8C,$D7,$E8,$67,$33,$F8,$A7,$F8
	.dc.b	$99,$1F,$E1,$8D,$7F,$84,$B5,$E0,$68,$86,$8B,$1D,$DF,$41,$FC,$33
	.dc.b	$C4,$F2,$C6,$63,$D0,$64,$D9,$82,$3F,$B9,$8F,$E2,$FE,$2F,$F1,$8F
	.dc.b	$16,$57,$F1,$88,$CB,$FC,$6A,$32,$FF,$1A,$8C,$BF,$C7,$23,$35,$FC
	.dc.b	$16,$73,$3F,$AE,$7F,$9B,$91,$FE,$60,$D7,$F9,$6B,$5E,$06,$88,$68
	.dc.b	$B1,$DD,$F4,$1F,$CC,$3C,$4F,$2C,$66,$3D,$06,$4D,$98,$23,$FC,$76
	.dc.b	$3F,$9D,$F8,$BF,$CF,$3C,$59,$5F,$CF,$23,$2F,$F3,$E8,$CB,$FC,$FA
	.dc.b	$32,$FF,$40,$8C,$D7,$F2,$99,$CC,$FF,$12,$7F,$AD,$91,$FE,$A8,$D7
	.dc.b	$FA,$8B,$5E,$06,$88,$68,$B1,$DD,$F4,$1F,$D5,$3C,$4F,$2C,$66,$3D
	.dc.b	$06,$4D,$98,$23,$FD,$06,$3F,$AF,$F8,$BF,$D8,$3C,$59,$5F,$D8,$23
	.dc.b	$2F,$F6,$28,$CB,$FD,$8A,$32,$FF,$64,$8C,$D7,$F4,$D9,$CC,$FF,$36
	.dc.b	$7F,$BF,$91,$FE,$F0,$D7,$FB,$AB,$5E,$06,$88,$68,$B1,$DD,$F4,$1F
	.dc.b	$DE,$3C,$4F,$2C,$66,$3D,$06,$4D,$98,$23,$FD,$96,$3F,$C0,$7E,$2F
	.dc.b	$F8,$13,$C5,$95,$FE,$04,$8C,$BF,$E0,$68,$CB,$FE,$06,$8C,$BF,$E0
	.dc.b	$88,$CD,$7F,$71,$9C,$CF,$F5,$A7,$FC,$46,$47,$FC,$38,$D7,$FC,$32
	.dc.b	$D7,$81,$A2,$1A,$2C,$77,$7D,$07,$F8,$73,$C4,$F2,$C6,$63,$D0,$64
	.dc.b	$D9,$82,$3F,$E0,$98,$FF,$13,$F8,$BF,$E2,$8F,$16,$57,$F8,$A2,$32
	.dc.b	$FF,$8A,$A3,$2F,$F8,$AA,$32,$FF,$8B,$23,$35,$FE,$15,$9C,$CF,$F7
	.dc.b	$E7,$FC,$8E,$47,$FC,$80,$D7,$FC,$7A,$D7,$81,$A2,$1A,$2C,$77,$7D
	.dc.b	$07,$F9,$03,$C5,$F2,$C6,$63,$D0,$64,$D9,$82,$3F,$E2,$D8,$FF,$25
	.dc.b	$F8,$BF,$E4,$CF,$16,$57,$F9,$32,$32,$FF,$93,$A3,$2F,$F9,$3A,$32
	.dc.b	$FF,$94,$23,$35,$FE,$39,$9C,$CF,$F8,$89,$FF,$35,$91,$FF,$32,$35
	.dc.b	$FF,$30,$B5,$E0,$68,$86,$8B,$1D,$DF,$41,$FE,$64,$F1,$7C,$B1,$98
	.dc.b	$F4,$19,$36,$60,$8F,$F9,$46,$3F,$CD,$FE,$2A,$64,$FF,$9F,$DF,$F4
	.dc.b	$07,$EF,$8F,$1B,$50,$9D,$78,$AA,$D4,$FF,$9E,$2F,$AB,$D2,$7F,$D1
	.dc.b	$B3,$A8,$E5,$18,$BD,$C7,$F9,$FA,$BB,$DF,$6B,$E0,$DA,$DB,$FE,$6B
	.dc.b	$9F,$F4,$D9,$2B,$FD,$28,$D7,$FD,$22,$D7,$FC,$B2,$D7,$FC,$EE,$D7
	.dc.b	$FC,$D8,$F1,$8E,$D1,$ED,$3D,$87,$C6,$F1,$7C,$78,$72,$8D,$1E,$E3
	.dc.b	$46,$27,$9B,$EF,$39,$C7,$CB,$9F,$82,$FC,$8F,$1A,$CF,$D4,$F1,$AC
	.dc.b	$FD,$16,$E5,$FE,$A7,$C3,$E4,$5E,$26,$5F,$6A,$D5,$F5,$8C,$4E,$9F
	.dc.b	$44,$7D,$DA,$EA,$7E,$0A,$3F,$D4,$15,$FE,$B5,$6C,$65,$E8,$6E,$79
	.dc.b	$39,$6B,$99,$E4,$F0,$39,$9B,$72,$FE,$52,$7B,$34,$D3,$E6,$7D,$CF
	.dc.b	$C4,$F1,$31,$1E,$C9,$F8,$2F,$13,$11,$AC,$3E,$4C,$72,$3E,$44,$7C
	.dc.b	$EF,$1B,$E7,$F8,$B2,$7E,$64,$65,$F9,$D1,$97,$C1,$19,$7C,$51,$9A
	.dc.b	$FC,$F4,$E6,$7E,$6B,$EE,$C7,$F5,$CD,$C0,$D2,$4F,$C7,$6B,$F3,$5A
	.dc.b	$F4,$34,$43,$CA,$7A,$8F,$C3,$78,$9E,$7B,$F2,$C6,$4F,$95,$19,$BF
	.dc.b	$DE,$D5,$9F,$E1,$33,$F9,$4F,$17,$F2,$BE,$2C,$9F,$C9,$46,$5F,$CA
	.dc.b	$46,$5F,$AD,$19,$7E,$C4,$67,$0F,$E0,$93,$99,$FC,$9B,$F1,$48,$FE
	.dc.b	$25,$AF,$E3,$DA,$F0,$34,$43,$45,$8E,$EF,$A0,$FD,$57,$89,$E5,$8C
	.dc.b	$C7,$A0,$C9,$B3,$04,$7F,$8E,$47,$EE,$3C,$5F,$DC,$F8,$B2,$BF,$6D
	.dc.b	$19,$7F,$71,$19,$7F,$35,$19,$7F,$39,$19,$AF,$E5,$13,$99,$FD,$BB
	.dc.b	$F8,$92,$3F,$AD,$6B,$FB,$36,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$87
	.dc.b	$78,$9E,$58,$CC,$7A,$0C,$9B,$30,$47,$FA,$04,$7F,$1C,$F1,$7F,$8E
	.dc.b	$F8,$B2,$BF,$8D,$46,$5F,$E3,$91,$97,$F7,$D1,$97,$F8,$04,$66,$BF
	.dc.b	$A6,$4E,$67,$F8,$DB,$FA,$D2,$3F,$C4,$5A,$FF,$17,$6B,$C0,$D1,$0D
	.dc.b	$16,$3B,$BE,$83,$F9,$97,$89,$E5,$8C,$C7,$A0,$C9,$B3,$04,$7F,$B2
	.dc.b	$47,$F4,$0F,$17,$FA,$0F,$8B,$2B,$F9,$F4,$65,$FE,$81,$19,$7F,$91
	.dc.b	$46,$5F,$E4,$91,$9A,$FE,$E1,$39,$9F,$E7,$EF,$F1,$12,$3F,$CD,$5A
	.dc.b	$FF,$3B,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$FA,$B7,$89,$E5,$8C,$C7
	.dc.b	$A0,$C9,$B3,$04,$7F,$C1,$11,$FD,$93,$C5,$FE,$CB,$E2,$CA,$FE,$C5
	.dc.b	$19,$7F,$B2,$46,$5F,$E8,$D1,$97,$FA,$44,$66,$BF,$C2,$93,$99,$FE
	.dc.b	$C6,$FF,$35,$23,$FD,$65,$AF,$F5,$F6,$BC,$0D,$10,$D1,$63,$BB,$E8
	.dc.b	$3F,$BD,$78,$9E,$58,$CC,$7A,$0C,$9B,$30,$47,$FC,$59,$1F,$E0,$8F
	.dc.b	$17,$FA,$61,$27,$FB,$83,$5F,$EE,$CD,$7F,$C0,$2D,$7F,$C7,$AD,$7F
	.dc.b	$C9,$2F,$F5,$7F,$16,$47,$F7,$B6,$BF,$E0,$16,$BF,$E0,$C6,$BF,$E4
	.dc.b	$86,$BF,$E5,$46,$F3,$FF,$94,$7C,$4F,$EC,$84,$9F,$F0,$43,$5F,$F0
	.dc.b	$73,$5F,$F0,$CB,$5F,$F2,$AB,$5F,$F3,$0B,$FD,$BF,$C5,$91,$FE,$13
	.dc.b	$6B,$FE,$19,$6B,$FE,$24,$6B,$FE,$60,$6B,$FE,$6C,$6C,$FF,$9A,$7C
	.dc.b	$5F,$EF,$84,$9F,$F1,$03,$5F,$F1,$33,$5F,$F1,$8B,$5F,$F3,$6B,$5F
	.dc.b	$F3,$CB,$FE,$07,$F1,$24,$7F,$8A,$DA,$FF,$8C,$5A,$FF,$8F,$1A,$FF
	.dc.b	$9E,$1A,$FF,$A1,$1B,$3F,$E8,$1F,$17,$FC,$28,$49,$FF,$1C,$35,$FF
	.dc.b	$1F,$35,$FF,$24,$B5,$FF,$42,$B5,$FF,$48,$BF,$E1,$FF,$16,$47,$F9
	.dc.b	$0D,$AF,$F9,$25,$AF,$F9,$51,$AF,$FA,$41,$AF,$FA,$71,$B3,$FE,$99
	.dc.b	$F1,$3F,$C5,$84,$9F,$F2,$83,$5F,$F2,$B3,$5F,$F3,$0B,$5F,$F4,$EB
	.dc.b	$5F,$F5,$4B,$FE,$39,$F9,$3F,$F5,$04,$7F,$AA,$A5,$18,$8F,$F5,$94
	.dc.b	$63,$C6,$E0,$48,$FF,$34,$4E,$D4,$ED,$CC,$ED,$FF,$38,$4E,$E7,$51
	.dc.b	$89,$F2,$78,$D2,$3A,$B5,$E2,$D7,$D1,$48,$C4,$7A,$23,$0F,$22,$1C
	.dc.b	$8F,$4F,$8B,$F0,$C8,$E0,$87,$A1,$CB,$4C,$9E,$27,$2D,$3E,$2F,$8D
	.dc.b	$F3,$C8,$E0,$87,$C4,$72,$D3,$27,$B9,$CB,$4F,$A3,$E3,$7D,$B2,$38
	.dc.b	$21,$F4,$1C,$B4,$C9,$F2,$9C,$B4,$FB,$BE,$27,$E1,$C8,$E0,$87,$DC
	.dc.b	$72,$D3,$27,$D6,$72,$D3,$F1,$3E,$37,$E4,$C8,$E0,$87,$E2,$1C,$B4
	.dc.b	$C9,$F8,$27,$2D,$3F,$2B,$E2,$FE,$7C,$8E,$08,$7E,$51,$CB,$4C,$9F
	.dc.b	$8E,$72,$D3,$F4,$3E,$27,$EA,$C8,$E0,$87,$E8,$1C,$B4,$C9,$F9,$A7
	.dc.b	$2D,$3F,$5B,$E2,$FE,$DC,$8E,$08,$7E,$B1,$CB,$4C,$9F,$A6,$72,$D3
	.dc.b	$F7,$3E,$2F,$FB,$E3,$43,$4C,$78,$96,$47,$FB,$CA,$76,$E3,$AF,$F8
	.dc.b	$0B,$FC,$0A,$32,$EE,$79,$3C,$6A,$D7,$0E,$1F,$E2,$35,$87,$F7,$53
	.dc.b	$6B,$A4,$D9,$FE,$2F,$36,$D3,$63,$0F,$B7,$C5,$F7,$78,$B3,$0F,$46
	.dc.b	$1F,$C1,$E7,$51,$83,$D5,$3B,$7E,$04,$69,$F8,$51,$9C,$7D,$58,$AB
	.dc.b	$55,$FC,$59,$FE,$60,$47,$F9,$4B,$5F,$E4,$AD,$78,$1A,$21,$A2,$CF
	.dc.b	$82,$3C,$8F,$8B,$C4,$B8,$7B,$6B,$D8,$FF,$3E,$AF,$C0,$4E,$A7,$EC
	.dc.b	$45,$9F,$E8,$73,$6F,$D9,$5C,$9A,$FE,$0E,$35,$5B,$6C,$6E,$3B,$B4
	.dc.b	$87,$CC,$3A,$47,$07,$07,$53,$77,$EF,$6D,$FF,$3F,$9E,$B1,$C2,$74
	.dc.b	$F1,$23,$82,$4F,$18,$F0,$9D,$0F,$86,$60,$E2,$FC,$B7,$B3,$C5,$C9
	.dc.b	$C0,$C9,$FA,$39,$F2,$CF,$82,$34,$FD,$77,$B1,$FD,$72,$3F,$BB,$11
	.dc.b	$FE,$DE,$D7,$FB,$5B,$5E,$06,$88,$68,$B1,$DD,$F3,$CF,$F8,$03,$C5
	.dc.b	$F2,$BE,$B9,$D1,$F9,$C8,$CC,$FA,$CF,$89,$19,$6C,$F1,$23,$FC,$0E
	.dc.b	$78,$C7,$E3,$23,$66,$4A,$E5,$F2,$E0,$9F,$E2,$C7,$25,$78,$9F,$CA
	.dc.b	$4E,$63,$11,$FC,$5E,$30,$E8,$87,$68,$DD,$FA,$E3,$06,$9C,$1F,$12
	.dc.b	$FD,$A7,$C9,$8D,$38,$BF,$2C,$60,$F7,$1D,$0F,$21,$98,$F8,$13,$D6
	.dc.b	$36,$F8,$DF,$F0,$FD,$FF,$8E,$48,$FF,$8C,$9A,$FF,$8B,$9A,$F0,$34
	.dc.b	$43,$45,$8F,$6C,$FF,$8A,$67,$FA,$27,$8B,$F1,$23,$15,$E7,$B4,$E6
	.dc.b	$3C,$F3,$A7,$FA,$34,$66,$3E,$34,$7C,$03,$27,$3A,$D5,$74,$8C,$1D
	.dc.b	$EB,$33,$A6,$C6,$4D,$9D,$6F,$F9,$9C,$6A,$31,$7F,$E4,$58,$C4,$FC
	.dc.b	$09,$FE,$B1,$18,$9F,$54,$F2,$9E,$AE,$E3,$FC,$CF,$3B,$9F,$82,$72
	.dc.b	$37,$1C,$8F,$03,$83,$6F,$27,$0F,$C6,$9F,$04,$7A,$0D,$DF,$E8,$DF
	.dc.b	$E6,$55,$F5,$9E,$31,$F2,$5F,$9A,$71,$AE,$87,$89,$CD,$7A,$E7,$A4
	.dc.b	$62,$75,$39,$BE,$D3,$93,$C6,$E2,$3C,$E6,$E3,$CA,$43,$C5,$5D,$A3
	.dc.b	$75,$D8,$F8,$D3,$EE,$BE,$91,$83,$D3,$1F,$34,$F1,$38,$25,$7C,$1A
	.dc.b	$EB,$3E,$63,$C4,$E2,$77,$7D,$A7,$A4,$FA,$5E,$B5,$CE,$BC,$4F,$A5
	.dc.b	$F4,$47,$07,$35,$CD,$FB,$91,$FE,$A0,$BF,$F5,$81,$1F,$F5,$4B,$5D
	.dc.b	$18,$21,$82,$1F,$EA,$07,$85,$FA,$EF,$CD,$3F,$E3,$1F,$13,$C5,$18
	.dc.b	$BF,$1D,$7F,$85,$A7,$37,$A7,$FC,$85,$E2,$FF,$90,$FC,$59,$8F,$F7
	.dc.b	$F8,$CB,$FE,$01,$8C,$BF,$E4,$38,$CB,$FE,$45,$8C,$DF,$F8,$92,$73
	.dc.b	$3F,$EB,$87,$FD,$A4,$47,$FD,$96,$D7,$FD,$8E,$D7,$81,$A2,$1A,$2C
	.dc.b	$77,$7D,$07,$F8,$5F,$C4,$F2,$C6,$63,$D0,$64,$D9,$82,$3F,$E3,$C8
	.dc.b	$E3,$8F,$FB,$36,$AD,$7F,$97,$E6,$CF,$FB,$AA,$6E,$49,$B1,$86,$D7
	.dc.b	$82,$FB,$36,$FB,$34,$EC,$57,$B3,$6F,$B3,$6F,$B3,$4D,$5F,$81,$F6
	.dc.b	$D7,$FB,$9E,$31,$7F,$14,$F1,$CE,$8F,$F3,$E5,$74,$71,$5E,$26,$DF
	.dc.b	$89,$FF,$43,$5F,$49,$ED,$1F,$3C,$DD,$7F,$A4,$66,$47,$9C,$E2,$6E
	.dc.b	$FC,$95,$F1,$23,$24,$8C,$BE,$63,$84,$78,$1E,$E6,$0F,$96,$60,$E4
	.dc.b	$EE,$39,$BE,$D7,$E1,$91,$F1,$3A,$BF,$80,$E6,$7C,$47,$C4,$AF,$6B
	.dc.b	$DE,$35,$4A,$E2,$F4,$24,$F9,$A7,$72,$BF,$A5,$18,$6D,$7A,$4E,$27
	.dc.b	$A4,$C5,$73,$73,$5C,$E3,$E2,$47,$8E,$FE,$C4,$6C,$91,$CD,$F6,$DF
	.dc.b	$92,$7F,$D8,$B1,$0F,$9A,$62,$7A,$39,$32,$7D,$53,$55,$F8,$96,$CE
	.dc.b	$DD,$92,$79,$BD,$DA,$F0,$3B,$3A,$AF,$34,$7F,$A6,$7C,$5F,$A7,$18
	.dc.b	$8F,$F6,$E4,$62,$7E,$39,$C4,$FF,$70,$9F,$5D,$D4,$7C,$4B,$FB,$2F
	.dc.b	$FB,$86,$33,$3C,$63,$51,$89,$DC,$E4,$E9,$3F,$4A,$32,$49,$E1,$7C
	.dc.b	$5B,$3B,$6B,$F9,$A3,$C4,$73,$8E,$0D,$38,$14,$E1,$5E,$26,$D7,$8A
	.dc.b	$7E,$ED,$FF,$B7,$AF,$BB,$F9,$46,$DF,$C8,$A4,$62,$FE,$09,$DE,$B8
	.dc.b	$36,$B8,$4E,$DF,$89,$5C,$A7,$47,$23,$CB,$7E,$C6,$9E,$C2,$9E,$AA
	.dc.b	$E9,$3F,$2E,$39,$D7,$69,$CD,$7B,$A3,$EF,$D7,$86,$B9,$DF,$A1,$B3
	.dc.b	$E8,$79,$CF,$57,$27,$9E,$7F,$1C,$71,$2C,$F1,$8D,$BE,$37,$9D,$78
	.dc.b	$DF,$B6,$6A,$BB,$57,$D3,$8F,$0D,$EE,$FE,$D4,$F6,$9E,$2E,$DE,$53
	.dc.b	$BA,$78,$BF,$AF,$BF,$E2,$8D,$8C,$D7,$EB,$1A,$FE,$C5,$AF,$EE,$A3
	.dc.b	$81,$1E,$0D,$78,$35,$FE,$03,$E2,$F7,$23,$DD,$AF,$76,$BC,$0D,$10
	.dc.b	$D1,$4F,$E1,$BE,$2F,$AC,$8F,$AD,$AF,$AD,$AF,$03,$44,$34,$53,$F8
	.dc.b	$CF,$8B,$F1,$C8,$FC,$76,$BF,$1D,$AF,$03,$44,$34,$53,$F9,$2F,$8B
	.dc.b	$F4,$C8,$FD,$36,$BF,$4D,$AF,$03,$44,$34,$53,$F9,$8F,$8B,$F7,$C8
	.dc.b	$FD,$F6,$BF,$7D,$AF,$03,$44,$34,$53,$F9,$EF,$8B,$F8,$B2,$3F,$8B
	.dc.b	$6B,$F8,$B6,$BC,$0D,$10,$D1,$4F,$E9,$3E,$2F,$E5,$C8,$FE,$5D,$AF
	.dc.b	$E5,$DA,$F0,$34,$43,$45,$3F,$AA,$F8,$BF,$A3,$23,$FA,$36,$BF,$A3
	.dc.b	$6B,$C0,$D1,$0D,$14,$FE,$C3,$E2,$FE,$E7,$F7,$D7,$8D,$4F,$DB,$44
	.dc.b	$90,$E2,$70,$FF,$56,$BE,$33,$A9,$CB,$A8,$CB,$C6,$7F,$8F,$D7,$58
	.dc.b	$C1,$BB,$F1,$3C,$E7,$6D,$3C,$3C,$2A,$DC,$72,$95,$F8,$30,$F1,$C7
	.dc.b	$F8,$86,$BC,$11,$89,$D4,$FA,$F0,$CB,$E0,$BE,$93,$A7,$CC,$78,$E3
	.dc.b	$A3,$70,$F2,$39,$9E,$13,$E4,$AE,$B3,$EA,$C3,$D3,$87,$8E,$73,$5D
	.dc.b	$2B,$C3,$1A,$8F,$24,$FB,$30,$EF,$5F,$C9,$6F,$F9,$51,$1F,$E5,$4D
	.dc.b	$74,$60,$86,$08,$7E,$18,$CC,$F7,$8F,$2E,$0F,$1B,$CF,$7E,$2A,$D5
	.dc.b	$FC,$19,$CD,$F0,$6C,$7C,$78,$FC,$94,$62,$3E,$D5,$73,$C3,$F1,$D3
	.dc.b	$9B,$EB,$5F,$52,$3F,$35,$E3,$7E,$6F,$C6,$93,$F4,$63,$2F,$D3,$8C
	.dc.b	$BF,$72,$32,$FD,$E8,$CD,$7E,$16,$72,$FE,$1E,$33,$7F,$A1,$9C,$BF
	.dc.b	$A3,$8C,$D7,$E8,$67,$2F,$E8,$E3,$33,$F9,$69,$F0,$CF,$F9,$67,$24
	.dc.b	$7F,$65,$6C,$7F,$93,$9A,$FF,$94,$DA,$F1,$8E,$04,$78,$35,$E0,$D7
	.dc.b	$81,$F1,$C8,$7E,$83,$C6,$F4,$D7,$9E,$33,$87,$EA,$67,$37,$E4,$8F
	.dc.b	$91,$3F,$A1,$9F,$E1,$DE,$37,$F0,$FF,$1A,$4F,$E9,$63,$2F,$E9,$E3
	.dc.b	$2F,$EC,$23,$2F,$EC,$A3,$35,$FB,$99,$CB,$FB,$B8,$CE,$1F,$C5,$67
	.dc.b	$2F,$F1,$78,$CD,$7F,$15,$9C,$BF,$C5,$E3,$33,$FC,$2A,$7F,$93,$78
	.dc.b	$DF,$4C,$95,$F4,$DA,$FD,$36,$BF,$6D,$AF,$F2,$76,$BF,$C9,$5F,$BD
	.dc.b	$E3,$48,$FB,$CD,$7E,$F3,$5F,$C3,$35,$FE,$64,$D7,$F9,$83,$6B,$F9
	.dc.b	$1F,$8D,$F8,$B2,$4F,$E2,$DA,$FE,$2D,$AF,$E4,$DA,$FF,$3F,$6B,$FC
	.dc.b	$F5,$FC,$B7,$8D,$23,$F2,$CD,$7F,$2C,$D7,$F3,$CD,$7F,$A5,$35,$FE
	.dc.b	$90,$D9,$FE,$77,$E3,$7E,$8C,$93,$FA,$36,$BF,$A3,$6B,$FA,$B6,$BF
	.dc.b	$D5,$DA,$FF,$55,$78,$E7,$FD,$0D,$FE,$C8,$7F,$65,$F1,$6C,$FF,$3B
	.dc.b	$8A,$FF,$44,$8B,$3C,$63,$47,$F4,$18,$E4,$FE,$E5,$FD,$E1,$1C,$10
	.dc.b	$FD,$E3,$96,$99,$3F,$83,$39,$7F,$84,$78,$DF,$DE,$C8,$FF,$51,$F1
	.dc.b	$AB,$FD,$6A,$2D,$7F,$61,$9B,$87,$F7,$B9,$B5,$FD,$E6,$6C,$FF,$6C
	.dc.b	$9B,$C0,$9B,$18,$6D,$7F,$61,$8F,$F0,$74,$7F,$83,$FC,$6F,$F0,$87
	.dc.b	$8D,$2B,$FB,$34,$65,$FE,$D1,$19,$7F,$B8,$C6,$5F,$EE,$71,$9C,$7F
	.dc.b	$BD,$4E,$5F,$EF,$91,$9C,$7F,$C2,$D3,$97,$FC,$31,$19,$AF,$F0,$B4
	.dc.b	$E5,$FF,$0C,$46,$67,$FC,$17,$3F,$E2,$7F,$1B,$FC,$51,$E2,$CB,$FE
	.dc.b	$FD,$19,$7F,$C0,$11,$97,$FC,$13,$19,$7F,$C1,$71,$9C,$3F,$C2,$53
	.dc.b	$97,$FC,$29,$19,$BF,$F1,$74,$E5,$FF,$18,$46,$6B,$FC,$5D,$39,$7F
	.dc.b	$C6,$11,$99,$FF,$11,$CF,$F8,$FF,$C6,$95,$FE,$15,$7F,$C8,$5E,$36
	.dc.b	$23,$FC,$3F,$17,$0F,$F1,$34,$DB,$FF,$20,$4D,$AF,$F1,$F4,$D9,$FF
	.dc.b	$18,$CD,$D5,$36,$30,$DA,$F1,$1F,$E2,$77,$FC,$A9,$E3,$7F,$95,$7C
	.dc.b	$69,$3F,$E2,$B8,$CB,$FE,$2D,$8C,$BF,$E3,$68,$CB,$FE,$38,$8C,$E3
	.dc.b	$FE,$3F,$9C,$BF,$E4,$18,$CE,$3F,$E5,$B9,$CB,$FE,$5D,$8C,$D7,$F9
	.dc.b	$6E,$72,$FF,$97,$63,$33,$FE,$50,$9F,$F3,$67,$8D,$FD,$E4,$95,$FD
	.dc.b	$E5,$AF,$F7,$96,$BF,$E0,$16,$BF,$E6,$D6,$BF,$E6,$B7,$FC,$09,$E3
	.dc.b	$48,$FF,$02,$35,$FF,$02,$35,$FF,$06,$35,$FF,$3C,$35,$FF,$3B,$36
	.dc.b	$BF,$CD,$3E,$37,$F8,$44,$93,$FE,$11,$6B,$FE,$11,$6B,$FE,$19,$6B
	.dc.b	$FE,$85,$6B,$FE,$83,$7F,$C3,$9E,$34,$8F,$F0,$E3,$5F,$F0,$E3,$5F
	.dc.b	$F1,$23,$5F,$F4,$83,$5F,$F4,$73,$67,$FD,$03,$E3,$7F,$8A,$49,$3F
	.dc.b	$E2,$96,$BF,$E2,$96,$BF,$E3,$16,$BF,$E9,$D6,$BF,$E9,$B7,$8E,$7F
	.dc.b	$E8,$97,$FD,$54,$7F,$AA,$FC,$6F,$F5,$3C,$BE,$39,$F1,$DC,$E3,$FD
	.dc.b	$09,$5C,$1F,$F4,$35,$70,$AD,$C7,$73,$64,$3F,$D2,$71,$BA,$FF,$59
	.dc.b	$36,$38,$1E,$18,$F0,$34,$FF,$5A,$46,$F0,$FF,$53,$B7,$85,$85,$31
	.dc.b	$12,$BF,$D8,$78,$79,$AF,$E7,$C6,$30,$EF,$39,$C1,$7C,$6B,$C5,$7C
	.dc.b	$1B,$3E,$07,$91,$E6,$2C,$79,$9B,$1B,$78,$E1,$D6,$FC,$17,$E9,$9D
	.dc.b	$CE,$70,$F8,$F1,$82,$1E,$29,$D3,$F2,$AB,$D5,$58,$C3,$C5,$3E,$03
	.dc.b	$8C,$F8,$1F,$2D,$78,$5D,$1D,$4C,$9C,$2F,$E3,$9E,$47,$47,$09,$F0
	.dc.b	$BC,$0C,$B8,$75,$1C,$63,$C5,$18,$3D,$51,$92,$38,$25,$7B,$27,$89
	.dc.b	$D4,$93,$C2,$34,$FA,$27,$8E,$1F,$79,$F3,$57,$07,$37,$F2,$67,$EC
	.dc.b	$4E,$6F,$9D,$7D,$F3,$D0,$FE,$16,$3C,$77,$E1,$AE,$2F,$86,$78,$92
	.dc.b	$33,$5E,$03,$C7,$3B,$8C,$CE,$EF,$EB,$5F,$8E,$7D,$57,$F3,$AF,$86
	.dc.b	$1E,$9B,$DC,$73,$7D,$91,$89,$F5,$5F,$A1,$F8,$B1,$C5,$FC,$01,$E5
	.dc.b	$3B,$1A,$3C,$E7,$2A,$F0,$B8,$C3,$DD,$3D,$DF,$54,$74,$7E,$44,$FC
	.dc.b	$7B,$D6,$1F,$00,$F6,$E1,$E5,$AD,$DF,$8E,$31,$3B,$C3,$D1,$3F,$52
	.dc.b	$39,$5F,$C0,$8F,$04,$74,$C3,$EB,$57,$82,$B5,$5C,$5B,$7F,$3D,$B5
	.dc.b	$F2,$E7,$DD,$7E,$16,$CF,$AA,$FC,$77,$F5,$DB,$7F,$1A,$BE,$5E,$1F
	.dc.b	$19,$F2,$DF,$67,$DD,$3E,$67,$27,$82,$BD,$0D,$32,$74,$C3,$C8,$DA
	.dc.b	$F2,$3E,$63,$E1,$CF,$91,$FB,$86,$CC,$9B,$3E,$63,$BB,$FA,$64,$AF
	.dc.b	$99,$3F,$83,$39,$E1,$F5,$AF,$C3,$1D,$4E,$67,$58,$F8,$CD,$C3,$E3
	.dc.b	$47,$18,$F1,$9C,$5D,$B6,$36,$7E,$2D,$DC,$71,$7B,$9B,$3C,$B1,$B7
	.dc.b	$31,$B3,$A9,$63,$AC,$7B,$6F,$EC,$3F,$87,$3E,$91,$4F,$A5,$87,$D6
	.dc.b	$69,$F5,$AB,$F3,$31,$96,$C6,$6B,$F0,$0F,$38,$CB,$63,$31,$F9,$58
	.dc.b	$F0,$B5,$EE,$6C,$F4,$61,$F8,$48,$F6,$47,$C2,$33,$87,$D6,$8C,$E1
	.dc.b	$F7,$70,$FC,$05,$F1,$C3,$E4,$57,$C6,$69,$F1,$AB,$F7,$F8,$7C,$C8
	.dc.b	$CB,$DE,$33,$3C,$63,$C2,$7D,$89,$DC,$F8,$63,$F0,$87,$1A,$FC,$7C
	.dc.b	$7D,$2A,$F8,$07,$C6,$2D,$7C,$63,$F7,$15,$F1,$23,$2F,$C4,$BF,$A5
	.dc.b	$3B,$8F,$D8,$99,$3E,$A1,$4F,$99,$5E,$18,$EB,$5E,$16,$9C,$E3,$F4
	.dc.b	$D8,$7E,$F5,$B5,$F6,$27,$6F,$5A,$F2,$CF,$58,$F1,$B4,$E2,$FA,$E7
	.dc.b	$8C,$67,$0F,$E4,$33,$D4,$E2,$FC,$83,$ED,$3F,$40,$FA,$C5,$AF,$A9
	.dc.b	$3F,$5A,$BE,$E5,$71,$BF,$C3,$47,$D6,$8E,$73,$BB,$FA,$D5,$EF,$9F
	.dc.b	$1C,$79,$CF,$1C,$72,$7C,$15,$DA,$38,$61,$F9,$03,$EA,$CF,$C5,$9F
	.dc.b	$7D,$7A,$E7,$37,$F0,$5E,$15,$F6,$23,$FC,$47,$F1,$FF,$E1,$32,$5F
	.dc.b	$F8,$6D,$FF,$09,$98,$9F,$F2,$AB,$88,$D1,$FE,$56,$75,$3D,$23,$EF
	.dc.b	$5F,$BB,$1F,$C1,$5F,$B8,$FC,$E1,$EE,$3E,$29,$F5,$6B,$ED,$C7,$06
	.dc.b	$D7,$F8,$BE,$7F,$CE,$F1,$FE,$3C,$6D,$7F,$80,$9B,$8F,$F9,$F1,$A5
	.dc.b	$3F,$C3,$06,$0D,$4F,$37,$B1,$0D,$95,$4B,$C9,$E9,$29,$CC,$61,$B1
	.dc.b	$E1,$6B,$E0,$9E,$4E,$5A,$E6,$79,$46,$23,$DF,$18,$23,$C5,$F7,$7C
	.dc.b	$80,$80

;--------------------------------------------------------------------------------
;	FMOVE.D <mem>,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovedtoreg_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.D <EA>,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.D <ea>,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fmovedtoreg_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a0,a2			;source handle,...
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;<mem>,FPn
	fmove.l	#0,fpcr
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fmove.d	([a2]),fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)		;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.D <mem>='
	puthex16	([a2]),(4,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmovedtoreg_data_compressed::
	.dc.b	$00,$01,$0F,$84,$0C,$9F,$A8,$BE,$FF,$31,$8F,$C1,$EC,$BF,$0F,$BF
	.dc.b	$8B,$E3,$F1,$FC,$FE,$4F,$AF,$CB,$F7,$F9,$FF,$0F,$D3,$F8,$FE,$BD
	.dc.b	$CB,$F6,$EE,$7F,$BF,$76,$FD,$55,$CB,$7D,$CC,$7C,$66,$7E,$93,$5F
	.dc.b	$79,$BF,$C1,$38,$73,$71,$E5,$E5,$C9,$CF,$93,$A7,$27,$5E,$4E,$DC
	.dc.b	$9D,$FF,$3D,$F0,$72,$F8,$7F,$4C,$F1,$7E,$B3,$E3,$FD,$A3,$C9,$CD
	.dc.b	$E5,$FD,$DF,$E1,$FE,$B3,$E6,$ED,$F9,$0B,$CF,$CF,$AF,$AF,$7F,$7D
	.dc.b	$3C,$3E,$3B,$8F,$57,$2F,$9D,$E7,$F4,$BD,$3B,$3A,$FD,$A7,$6F,$BC
	.dc.b	$EF,$DD,$E0,$EE,$F0,$F7,$78,$BB,$BC,$7D,$DE,$4E,$EF,$2F,$77,$9B
	.dc.b	$BB,$CF,$DD,$E8,$EC,$F4,$F8,$5E,$AE,$AF,$BB,$FA,$D3,$D0,$BD,$18
	.dc.b	$E6,$CF,$26,$B8,$B7,$C1,$C3,$6E,$3E,$0F,$2F,$17,$9F,$81,$D3,$C5
	.dc.b	$FC,$4F,$EB,$7F,$17,$C1,$FC,$72,$EE,$CF,$6E,$BB,$37,$D9,$C3,$6E
	.dc.b	$3D,$9C,$BB,$39,$F0,$74,$EC,$EB,$D7,$DB,$C4,$EF,$E2,$78,$3C,$4F
	.dc.b	$CC,$BF,$5F,$C8,$2F,$0F,$3E,$1E,$BC,$3D,$EF,$C3,$5E,$3E,$1F,$2F
	.dc.b	$0F,$9F,$87,$D3,$C3,$EB,$E1,$FE,$9B,$F5,$CF,$95,$D5,$FA,$B5,$E1
	.dc.b	$67,$C2,$D7,$56,$F6,$E1,$E1,$71,$EB,$E5,$D7,$CF,$AF,$A7,$0F,$5E
	.dc.b	$1E,$DC,$3D,$F8,$7C,$1C,$3E,$1E,$1F,$DB,$FE,$BB,$E2,$E7,$3E,$2E
	.dc.b	$3F,$DF,$2E,$1D,$EF,$C3,$6E,$3D,$9C,$BC,$3E,$7E,$1F,$4F,$0F,$AF
	.dc.b	$87,$DB,$C3,$FE,$0B,$FA,$F7,$DD,$D8,$FD,$DD,$BF,$86,$2E,$CD,$ED
	.dc.b	$C3,$A7,$8F,$4F,$2E,$0E,$7D,$3D,$3A,$7A,$F8,$5D,$BC,$2E,$FE,$17
	.dc.b	$83,$C1,$F0,$F8,$5F,$C5,$BF,$5F,$FB,$C7,$DD,$E0,$FE,$38,$F8,$5B
	.dc.b	$A6,$FC,$2E,$1E,$07,$1F,$03,$97,$81,$CF,$C0,$E9,$E0,$75,$F0,$3F
	.dc.b	$91,$FE,$C1,$F3,$4F,$80,$77,$E9,$3F,$3D,$BC,$FC,$34,$E3,$CB,$CB
	.dc.b	$97,$9F,$7F,$4D,$7A,$EB,$DB,$93,$BF,$6F,$83,$B7,$C3,$B7,$F3,$4F
	.dc.b	$D8,$7F,$38,$78,$4F,$A1,$B1,$FB,$DC,$9F,$CF,$57,$57,$1E,$5E,$5B
	.dc.b	$73,$EA,$E9,$E0,$75,$DB,$B7,$37,$7E,$AF,$06,$DE,$1E,$FF,$E8,$BF
	.dc.b	$B1,$7E,$3C,$F8,$47,$CF,$DC,$BB,$B7,$B7,$0F,$0B,$8F,$67,$2D,$B9
	.dc.b	$F7,$F4,$F0,$BA,$ED,$DB,$BB,$BF,$81,$E0,$DB,$FA,$AC,$7E,$44,$FA
	.dc.b	$07,$84,$76,$EB,$B7,$7A,$70,$ED,$E3,$DB,$CB,$4E,$7D,$BD,$3B,$7A
	.dc.b	$E5,$DB,$B7,$FB,$1F,$EC,$7F,$C2,$CF,$D0,$1F,$40,$F9,$FC,$C7,$C3
	.dc.b	$DD,$FD,$B5,$76,$F2,$ED,$E7,$AF,$4E,$DE,$BD,$BD,$B5,$EF,$DB,$E0
	.dc.b	$ED,$F0,$EB,$FD,$D3,$F6,$4F,$DF,$1F,$91,$3F,$02,$7C,$FD,$7F,$BE
	.dc.b	$3F,$80,$6E,$9C,$7A,$F9,$75,$F3,$D3,$A7,$5F,$5E,$BE,$DA,$77,$EB
	.dc.b	$F0,$75,$FF,$81,$3F,$65,$FE,$1C,$7E,$7C,$FC,$01,$E1,$E5,$2E,$BE
	.dc.b	$1D,$7C,$75,$E5,$D7,$CF,$AF,$A6,$BD,$7A,$FB,$75,$F7,$D7,$FC,$27
	.dc.b	$FB,$37,$F3,$B3,$F8,$81,$FA,$F3,$E8,$9F,$43,$53,$E1,$E4,$FF,$0F
	.dc.b	$2E,$BE,$7D,$7D,$34,$EB,$D7,$DB,$AF,$BE,$9E,$0E,$BF,$0F,$5F,$F8
	.dc.b	$93,$F6,$7F,$E9,$87,$F0,$13,$F6,$27,$E0,$8F,$A1,$A9,$FC,$0E,$BF
	.dc.b	$E2,$F7,$C4,$DE,$AE,$9D,$5D,$73,$ED,$D5,$DF,$AB,$C1,$C7,$E1,$E3
	.dc.b	$F1,$69,$E3,$EA,$FF,$1B,$FE,$D1,$FD,$7C,$FE,$84,$7F,$35,$3F,$8B
	.dc.b	$1F,$C5,$4F,$E0,$E7,$ED,$CF,$CC,$9F,$64,$F9,$47,$A8,$F9,$5B,$1F
	.dc.b	$5B,$7A,$7C,$3C,$BE,$2E,$9F,$1F,$37,$93,$83,$CB,$CD,$E6,$E1,$F3
	.dc.b	$F4,$FA,$38,$BD,$3D,$3E,$AE,$8F,$5F,$27,$B3,$A3,$DB,$CB,$EE,$E5
	.dc.b	$F7,$F4,$7C,$0E,$6F,$83,$CD,$F0,$B9,$BF,$CB,$3F,$B4,$FF,$80,$4F
	.dc.b	$EC,$87,$F5,$73,$F9,$C1,$FC,$DC,$FE,$52,$7F,$19,$3F,$76,$7E,$6C
	.dc.b	$FC,$61,$F7,$0F,$86,$7C,$21,$CB,$E0,$E5,$F0,$F2,$F8,$B9,$7C,$7C
	.dc.b	$BE,$4E,$5F,$2F,$2F,$9B,$97,$CF,$CB,$E8,$E5,$F4,$F2,$FA,$B9,$7D
	.dc.b	$7C,$BE,$CE,$1F,$6F,$0F,$F9,$DB,$8F,$FC,$3C,$7F,$80,$4F,$E2,$67
	.dc.b	$EC,$CF,$D2,$1F,$94,$3E,$D9,$F6,$8F,$4F,$47,$FA,$0D,$FC,$AF,$39
	.dc.b	$FF,$42,$46,$78,$3C,$1C,$1E,$1E,$0F,$17,$07,$8F,$83,$C9,$C1,$E5
	.dc.b	$E0,$F3,$70,$79,$F8,$3D,$1C,$1E,$9E,$7F,$57,$3F,$AF,$7F,$67,$3F
	.dc.b	$B7,$8F,$DD,$C7,$FE,$94,$FD,$AB,$FC,$78,$7F,$87,$CF,$F0,$91,$FD
	.dc.b	$CC,$FE,$E4,$7E,$C8,$FC,$C9,$F9,$83,$DF,$C8,$5C,$3D,$78,$7B,$70
	.dc.b	$F7,$E1,$F0,$70,$F8,$78,$7C,$5C,$3E,$3E,$1F,$27,$0F,$97,$87,$CD
	.dc.b	$C3,$E7,$E1,$F4,$70,$FA,$78,$7F,$D5,$BF,$B5,$FF,$95,$4F,$F1,$C1
	.dc.b	$FE,$2B,$3F,$C1,$87,$F8,$2C,$FE,$FA,$7F,$6D,$3F,$A7,$9F,$CB,$CF
	.dc.b	$E4,$07,$F1,$03,$F6,$E7,$ED,$8F,$79,$F7,$0F,$B7,$C4,$5C,$9E,$4E
	.dc.b	$4F,$2F,$1F,$9B,$8F,$CF,$C7,$E8,$E3,$F4,$F1,$FA,$B8,$FD,$7C,$7E
	.dc.b	$CE,$3F,$6F,$1F,$BB,$8F,$DF,$C7,$F0,$38,$FE,$0F,$1F,$C2,$E3,$FF
	.dc.b	$67,$FE,$D9,$FE,$77,$3F,$CB,$47,$F9,$34,$FF,$17,$1F,$E2,$D3,$FC
	.dc.b	$3E,$7F,$83,$CF,$EE,$27,$F5,$13,$FA,$11,$FC,$C8,$FE,$32,$7F,$18
	.dc.b	$3F,$02,$7E,$78,$F8,$67,$FB,$77,$9B,$FD,$C8,$FE,$11,$BB,$F9,$B5
	.dc.b	$F3,$E9,$E8,$CF,$D3,$8F,$FB,$A3,$F6,$DF,$F4,$81,$FE,$65,$3F,$C6
	.dc.b	$A7,$F8,$18,$FE,$AC,$7F,$2A,$3F,$5E,$7D,$2C,$3F,$DE,$8B,$87,$A7
	.dc.b	$0F,$5C,$7B,$6F,$DF,$7F,$06,$3F,$EF,$8F,$DB,$BF,$D0,$47,$F9,$54
	.dc.b	$FF,$22,$9F,$DD,$4F,$EA,$A7,$F3,$A3,$F0,$67,$C1,$E7,$3E,$1C,$D1
	.dc.b	$8E,$3E,$DC,$7D,$F1,$F0,$71,$F8,$78,$D8,$70,$FD,$BF,$FD,$28,$7F
	.dc.b	$99,$4F,$F2,$31,$FE,$07,$3F,$A6,$9F,$B4,$3F,$2A,$7D,$1C,$4B,$8F
	.dc.b	$AC,$63,$2B,$2E,$3B,$C7,$8C,$B7,$FB,$87,$FB,$0C,$FF,$4C,$1F,$E7
	.dc.b	$73,$FC,$72,$7F,$82,$4F,$EB,$E7,$F2,$23,$EA,$60,$7C,$83,$8B,$AF
	.dc.b	$17,$6C,$3B,$F1,$78,$38,$BC,$F8,$FF,$B4,$CF,$F4,$41,$FE,$78,$3F
	.dc.b	$C9,$27,$F8,$20,$FE,$DA,$7F,$32,$3E,$A1,$F4,$F1,$3E,$1D,$BE,$AA
	.dc.b	$E1,$EF,$C3,$E0,$C3,$C3,$C3,$F2,$3F,$71,$FF,$63,$1F,$E9,$93,$FC
	.dc.b	$F0,$7F,$8E,$4F,$F0,$C9,$FC,$98,$FE,$0E,$7E,$14,$FA,$58,$7E,$11
	.dc.b	$70,$F5,$C3,$B7,$07,$7C,$BC,$18,$7D,$8F,$DC,$BF,$D9,$07,$FA,$9C
	.dc.b	$FF,$44,$1F,$E3,$93,$FC,$1E,$7F,$36,$3E,$89,$F8,$33,$C3,$A1,$6F
	.dc.b	$D7,$7E,$D8,$77,$DF,$F0,$FF,$B9,$BD,$EB,$EE,$E3,$DA,$CF,$A5,$AF
	.dc.b	$3B,$7E,$67,$0F,$17,$8F,$85,$CB,$BB,$9F,$57,$4E,$0E,$BD,$5D,$BA
	.dc.b	$BB,$F0,$7E,$5B,$F7,$4F,$F7,$A1,$FE,$C7,$3F,$D6,$E7,$F9,$9C,$FF
	.dc.b	$2C,$1F,$E4,$83,$FC,$5E,$7F,$72,$3F,$68,$7C,$03,$DF,$D7,$FB,$45
	.dc.b	$5D,$FA,$3F,$4B,$FB,$AF,$BE,$BF,$72,$FF,$B5,$CF,$F5,$B1,$FE,$67
	.dc.b	$3F,$C7,$67,$F4,$43,$F6,$07,$E3,$7A,$0F,$E3,$63,$9F,$37,$6C,$7F
	.dc.b	$67,$FB,$B3,$DF,$BC,$D8,$F7,$B3,$EE,$EB,$A3,$7E,$DF,$0E,$9E,$3E
	.dc.b	$B7,$2F,$3B,$9F,$47,$4E,$8E,$BE,$17,$6E,$FE,$FD,$1E,$0F,$07,$F8
	.dc.b	$1F,$EE,$EE,$0B,$B6,$3D,$8C,$FA,$BA,$EF,$DF,$9D,$C3,$CA,$E3,$E0
	.dc.b	$72,$EF,$E7,$DF,$D3,$AF,$AE,$DF,$C4,$1F,$77,$6F,$F2,$47,$C0,$DE
	.dc.b	$AD,$75,$6F,$7E,$1B,$F1,$DF,$96,$FC,$F7,$E9,$BF,$5D,$FB,$6F,$DF
	.dc.b	$7F,$E3,$FF,$BC,$7D,$BF,$00,$B6,$CF,$66,$BC,$4D,$F8,$9C,$3D,$2E
	.dc.b	$3E,$87,$2F,$1B,$9F,$95,$D3,$C5,$EB,$D9,$DB,$6E,$FD,$3E,$0E,$AF
	.dc.b	$E5,$FF,$BC,$F8,$4F,$83,$C2,$5D,$5A,$F1,$37,$E2,$70,$F1,$38,$F8
	.dc.b	$9C,$BC,$4E,$7E,$27,$4F,$13,$AE,$FD,$B5,$FE,$7B,$FB,$D7,$CC,$D8
	.dc.b	$BC,$2C,$F8,$5A,$F0,$B7,$E1,$70,$F0,$B8,$F8,$5C,$BC,$2E,$7D,$3D
	.dc.b	$36,$EB,$E1,$76,$EA,$EF,$C1,$E0,$EA,$F0,$F5,$78,$B8,$BF,$A7,$FE
	.dc.b	$F7,$F4,$4F,$82,$79,$4F,$A5,$E2,$7F,$66,$7E,$1B,$78,$B8,$F8,$5C
	.dc.b	$BA,$F9,$F2,$74,$F0,$BA,$F6,$76,$F0,$BB,$F3,$78,$3C,$2F,$0F,$6F
	.dc.b	$8B,$C2,$F1,$F4,$79,$3C,$2F,$2F,$77,$9B,$AB,$CF,$D5,$E8,$EA,$F4
	.dc.b	$F5,$7A,$BA,$BD,$7D,$5E,$CE,$AF,$6F,$57,$BB,$AB,$FB,$6F,$EF,$9F
	.dc.b	$44,$FA,$1D,$5F,$E0,$27,$FD,$86,$7F,$B0,$8F,$F3,$59,$FE,$4C,$3F
	.dc.b	$C1,$47,$F5,$73,$F8,$51,$FC,$63,$BC,$FF,$0B,$6E,$7E,$0C,$FC,$39
	.dc.b	$F8,$B3,$F1,$E6,$FE,$01,$85,$CB,$FC,$1E,$7F,$84,$3E,$4A,$B0,$35
	.dc.b	$DB,$5E,$05,$70,$F0,$30,$43,$93,$A2,$39,$3C,$7F,$1A,$EA,$FF,$14
	.dc.b	$C7,$8E,$31,$A7,$8D,$B1,$99,$DB,$D8,$7C,$CB,$F0,$8F,$7F,$C8,$3E
	.dc.b	$7F,$AC,$E6,$BE,$0E,$5F,$E3,$28,$C4,$2B,$86,$59,$C3,$CF,$8F,$82
	.dc.b	$AC,$F9,$63,$11,$FE,$3C,$CB,$CE,$FC,$E5,$F8,$B6,$A1,$18,$AF,$74
	.dc.b	$E6,$31,$3F,$33,$1F,$2E,$3E,$48,$F7,$E3,$F6,$72,$F3,$47,$F9,$32
	.dc.b	$75,$5E,$CB,$E7,$3E,$3C,$3B,$13,$1F,$9D,$DD,$E1,$23,$83,$EC,$FC
	.dc.b	$DF,$8B,$5F,$AE,$FC,$8B,$FC,$2B,$99,$FF,$2E,$BE,$1A,$E7,$3F,$75
	.dc.b	$AF,$C5,$FC,$AF,$E0,$12,$BF,$00,$FC,$78,$FC,$8E,$5D,$5F,$82,$7D
	.dc.b	$A2,$9F,$67,$0F,$BF,$AF,$E3,$AB,$A4,$FE,$35,$AF,$D5,$EF,$FA,$04
	.dc.b	$C3,$F2,$0F,$86,$3F,$3D,$87,$84,$F8,$0E,$2F,$F1,$F8,$78,$23,$F0
	.dc.b	$55,$CD,$FC,$E3,$5F,$D9,$FC,$AB,$EE,$3F,$9C,$63,$FB,$38,$C7,$8F
	.dc.b	$F8,$C6,$C6,$67,$75,$F8,$5C,$7C,$33,$F8,$28,$FD,$C9,$F1,$38,$3F
	.dc.b	$07,$3F,$E9,$0B,$E7,$3F,$C0,$9A,$F9,$2F,$AB,$70,$F1,$43,$C5,$A4
	.dc.b	$C7,$C8,$FC,$7E,$2F,$16,$5E,$0C,$3E,$B6,$BF,$27,$3F,$16,$3C,$F1
	.dc.b	$E2,$4A,$F2,$46,$5B,$E5,$78,$89,$3F,$0D,$F0,$47,$BF,$6F,$A4,$FC
	.dc.b	$2C,$3C,$0B,$5F,$A1,$5C,$A3,$AD,$F3,$3E,$77,$8B,$F0,$08,$E0,$FE
	.dc.b	$37,$7D,$27,$B6,$5F,$49,$FC,$AE,$1F,$76,$F8,$C7,$B1,$AF,$CF,$F2
	.dc.b	$BE,$69,$30,$FB,$CF,$B3,$0E,$D3,$FA,$2C,$3B,$47,$F1,$BC,$3F,$15
	.dc.b	$5C,$6B,$E4,$B5,$FE,$7F,$A7,$FB,$11,$FE,$80,$D7,$27,$F1,$57,$0F
	.dc.b	$AF,$97,$F9,$7C,$7F,$11,$7F,$8F,$ED,$F7,$B5,$FE,$73,$1A,$AE,$44
	.dc.b	$CF,$FA,$65,$F4,$8F,$E2,$D5,$DC,$93,$A6,$C7,$99,$AF,$A9,$AF,$CF
	.dc.b	$31,$1C,$4F,$D3,$78,$9F,$98,$23,$83,$FB,$01,$FD,$BB,$C0,$F4,$BF
	.dc.b	$EE,$8F,$DF,$BE,$36,$3F,$D5,$23,$47,$F3,$D3,$57,$C0,$FC,$67,$0F
	.dc.b	$FB,$C6,$3B,$47,$E6,$B0,$F0,$BF,$43,$F3,$3F,$9C,$49,$F8,$AD,$8D
	.dc.b	$46,$08,$E0,$87,$DC,$8E,$0F,$DE,$F1,$7E,$D1,$27,$E8,$B6,$3F,$02
	.dc.b	$D7,$84,$68,$8E,$9A,$FE,$2B,$BB,$F0,$C4,$8F,$BA,$D8,$FC,$6B,$5E
	.dc.b	$11,$A2,$3A,$6B,$F9,$6F,$67,$F2,$44,$8F,$C4,$B6,$3F,$32,$D7,$84
	.dc.b	$68,$8E,$9A,$FE,$8B,$C4,$FC,$F1,$23,$F2,$AD,$8F,$D2,$B5,$E1,$1A
	.dc.b	$23,$A6,$BF,$E2,$CF,$69,$1F,$DF,$DA,$FF,$8B,$5B,$1F,$E3,$16,$BC
	.dc.b	$63,$81,$0E,$05,$78,$35,$E0,$F5,$F5,$BC,$04,$70,$48,$F0,$46,$5B
	.dc.b	$19,$7D,$51,$96,$C7,$A7,$EA,$FD,$84,$70,$48,$F6,$46,$5B,$19,$7E
	.dc.b	$34,$65,$B1,$F1,$7A,$FE,$41,$1C,$12,$3E,$44,$65,$B1,$97,$E9,$46
	.dc.b	$5B,$1F,$47,$EF,$FE,$A1,$1C,$12,$3E,$A4,$65,$B1,$97,$EF,$46,$5B
	.dc.b	$1F,$77,$B3,$F0,$04,$70,$48,$FC,$04,$65,$B1,$97,$F1,$51,$96,$C7
	.dc.b	$F9,$D7,$F5,$37,$EA,$1F,$E4,$B6,$C6,$63,$07,$E4,$7C,$6F,$E2,$D4
	.dc.b	$D1,$1F,$C6,$69,$CA,$F7,$37,$5A,$9F,$E4,$16,$18,$F8,$1F,$F4,$53
	.dc.b	$5F,$F4,$6B,$5F,$F4,$8B,$5E,$53,$D5,$FD,$16,$1E,$7A,$FD,$33,$63
	.dc.b	$31,$82,$38,$21,$FB,$18,$E0,$FF,$87,$7C,$1F,$F3,$87,$69,$A4,$AE
	.dc.b	$7D,$1F,$CF,$F3,$FE,$BD,$1F,$EA,$8C,$7C,$D3,$9E,$AF,$97,$87,$3A
	.dc.b	$FF,$54,$37,$1F,$F1,$D6,$3D,$5F,$F1,$E7,$EA,$DE,$22,$57,$07,$FD
	.dc.b	$09,$F3,$CF,$BD,$FA,$ED,$9F,$A2,$7D,$6F,$6F,$33,$E2,$AE,$77,$DA
	.dc.b	$BA,$1F,$EB,$BB,$E5,$1F,$EB,$F6,$BF,$EC,$26,$BF,$EC,$66,$BF,$7D
	.dc.b	$A9,$A4,$AF,$49,$27,$C0,$D9,$F0,$75,$FB,$89,$3E,$57,$5A,$7D,$1A
	.dc.b	$F6,$CF,$C0,$DB,$D3,$3F,$EC,$0C,$17,$CB,$0E,$B7,$CC,$F8,$1F,$B9
	.dc.b	$7C,$C2,$38,$3C,$A7,$5C,$7B,$CF,$FA,$D2,$3F,$C0,$51,$F4,$A7,$8C
	.dc.b	$FC,$96,$BF,$E6,$7F,$4B,$FC,$CA,$4A,$FF,$3B,$3F,$EC,$47,$F5,$BE
	.dc.b	$BF,$EB,$F3,$37,$C6,$C3,$FE,$0B,$E0,$FF,$42,$5F,$3C,$3F,$CC,$6D
	.dc.b	$7E,$C7,$6F,$E2,$09,$8F,$DF,$6C,$7D,$F6,$BC,$23,$44,$74,$D7,$F0
	.dc.b	$9D,$BF,$94,$24,$7E,$2D,$B1,$F8,$B6,$BC,$23,$44,$74,$D7,$F8,$51
	.dc.b	$FC,$2F,$E8,$7F,$B0,$B8,$6B,$98,$FF,$74,$46,$20,$D2,$46,$88,$76
	.dc.b	$8E,$8E,$1C,$1C,$9E,$BD,$D1,$CC,$76,$8C,$3C,$25,$18,$8E,$D1,$89
	.dc.b	$E2,$D8,$E1,$3B,$7D,$CD,$7E,$17,$77,$B8,$95,$C9,$B1,$E1,$6B,$E0
	.dc.b	$6B,$F2,$5A,$F4,$8F,$1F,$77,$CA,$23,$82,$47,$B6,$32,$D8,$CB,$EE
	.dc.b	$8C,$B6,$3D,$FD,$DF,$58,$8E,$09,$1F,$26,$32,$D8,$CB,$F2,$A3,$2D
	.dc.b	$8F,$97,$DD,$F8,$22,$38,$24,$7D,$58,$CB,$63,$2F,$D6,$8C,$B6,$3E
	.dc.b	$BF,$E2,$FF,$8E,$23,$82,$47,$E0,$63,$2D,$8C,$BF,$82,$8C,$B6,$3F
	.dc.b	$07,$E5,$7E,$68,$8E,$09,$1F,$8D,$8C,$B6,$32,$FE,$3A,$32,$D8,$FC
	.dc.b	$7F,$E5,$7F,$A6,$23,$82,$47,$E6,$63,$2D,$8C,$BF,$9A,$8C,$B6,$3F
	.dc.b	$37,$E5,$FE,$C8,$8E,$09,$1F,$A5,$8C,$B6,$32,$FE,$9A,$32,$D8,$FD
	.dc.b	$3F,$E7,$DF,$BE,$23,$82,$47,$EC,$63,$2D,$8C,$BF,$B2,$8C,$B6,$3F
	.dc.b	$67,$E5,$FF,$0A,$23,$82,$47,$EF,$63,$2D,$8C,$BF,$BE,$8C,$B6,$3F
	.dc.b	$7F,$E2,$7F,$16,$23,$82,$47,$F0,$98,$CB,$63,$2F,$F0,$A8,$CB,$63
	.dc.b	$F8,$5F,$8F,$FC,$88,$8E,$09,$1F,$C5,$63,$2D,$8C,$BF,$C5,$A3,$2D
	.dc.b	$8F,$E2,$FF,$B2,$FF,$2E,$23,$82,$47,$F2,$18,$CB,$63,$2F,$F2,$28
	.dc.b	$CB,$63,$F9,$1F,$2F,$F3,$A2,$38,$24,$7F,$2D,$8C,$B6,$32,$FF,$2E
	.dc.b	$8C,$B6,$3F,$97,$FE,$F3,$FD,$18,$8E,$09,$1F,$CE,$63,$2D,$8C,$BF
	.dc.b	$CE,$A3,$2D,$8F,$E7,$7F,$2D,$FD,$48,$8E,$0F,$E8,$E3,$D0,$FF,$37
	.dc.b	$B4,$FF,$0D,$7B,$D8,$8F,$F3,$9F,$3A,$B9,$DF,$F4,$D6,$BF,$D6,$6F
	.dc.b	$FC,$F0,$76,$F1,$FF,$D9,$B1,$FD,$6A,$32,$D8,$CB,$C6,$FC,$13,$E2
	.dc.b	$3F,$CF,$71,$FE,$D6,$8F,$3C,$F0,$8F,$EE,$B3,$D1,$F1,$46,$5F,$F4
	.dc.b	$14,$65,$DF,$DE,$E5,$FF,$70,$B6,$0D,$31,$5E,$49,$FF,$18,$1F,$EE
	.dc.b	$8E,$F6,$59,$F1,$3D,$78,$72,$3E,$7E,$1C,$DA,$7D,$2F,$5B,$C0,$47
	.dc.b	$07,$D6,$C7,$A3,$EA,$C3,$BB,$F1,$0F,$B1,$8F,$90,$FF,$7D,$D7,$37
	.dc.b	$E8,$35,$F9,$D1,$D4,$8F,$D5,$AF,$7C,$62,$36,$65,$C3,$F4,$23,$6E
	.dc.b	$4F,$BD,$EF,$7C,$B2,$4F,$E4,$1D,$E3,$F8,$48,$7E,$50,$F4,$BD,$F6
	.dc.b	$9C,$C7,$E5,$AF,$9C,$F6,$9E,$87,$E7,$3F,$6D,$FB,$44,$70,$7B,$B0
	.dc.b	$F0,$61,$DE,$3E,$56,$3F,$EC,$D8,$C5,$FC,$B9,$5F,$38,$FC,$D3,$5F
	.dc.b	$D0,$7B,$7F,$89,$24,$FE,$B9,$F7,$BF,$92,$AE,$B8,$7E,$1B,$0F,$06
	.dc.b	$5F,$B2,$BE,$38,$7E,$9D,$AF,$F9,$17,$EA,$BF,$36,$FF,$91,$1A,$E4
	.dc.b	$FC,$E3,$87,$EB,$63,$FE,$B4,$C7,$C2,$E2,$3F,$36,$FC,$E3,$F5,$F3
	.dc.b	$C6,$3F,$DD,$4D,$4D,$27,$9B,$E0,$24,$FF,$95,$21,$1E,$4F,$53,$FD
	.dc.b	$FB,$07,$FD,$1D,$78,$73,$3C,$AF,$DC,$D7,$E1,$79,$3E,$E2,$61,$DD
	.dc.b	$EC,$FA,$AB,$AB,$4F,$6C,$70,$9F,$94,$D7,$E7,$7C,$3F,$CA,$24,$FA
	.dc.b	$DE,$B1,$FA,$FC,$7D,$53,$FE,$C1,$C7,$D4,$FC,$AA,$E3,$5F,$5D,$AF
	.dc.b	$DB,$FD,$3F,$EB,$92,$7E,$43,$DB,$6F,$E4,$B8,$76,$AF,$AF,$C5,$FA
	.dc.b	$8B,$F6,$47,$8E,$F9,$3F,$85,$6B,$F8,$9C,$FD,$2D,$9F,$AD,$18,$8F
	.dc.b	$F4,$FD,$F7,$74,$7E,$3E,$34,$FE,$11,$AF,$E1,$1F,$C8,$FD,$14,$AF
	.dc.b	$C1,$36,$79,$47,$FB,$9F,$2F,$8F,$1E,$C7,$F0,$B3,$B8,$FD,$03,$5F
	.dc.b	$2F,$93,$F9,$E2,$57,$E3,$DF,$90,$FF,$27,$AE,$B9,$FF,$0F,$CD,$1F
	.dc.b	$92,$AE,$37,$FA,$B6,$BF,$EB,$3F,$69,$E4,$FF,$47,$6E,$3F,$EB,$06
	.dc.b	$CF,$FA,$D9,$AF,$C8,$C7,$C1,$87,$8A,$BC,$87,$F5,$CA,$DD,$7F,$B3
	.dc.b	$9A,$83,$86,$DF,$8F,$C7,$95,$E3,$6C,$FC,$E3,$E6,$47,$E1,$B0,$F1
	.dc.b	$BE,$19,$DC,$FB,$DA,$F8,$FC,$6F,$80,$49,$F8,$0D,$8F,$80,$D7,$E4
	.dc.b	$35,$E5,$1C,$08,$FC,$7F,$1B,$E6,$12,$3E,$63,$63,$51,$82,$38,$21
	.dc.b	$F5,$E3,$83,$F6,$7C,$6F,$B0,$49,$FB,$0D,$8F,$B0,$D7,$84,$68,$8E
	.dc.b	$9A,$FE,$17,$C6,$FC,$21,$23,$F0,$8D,$8F,$C2,$35,$E1,$1A,$23,$A6
	.dc.b	$BF,$91,$F1,$BF,$20,$48,$FC,$83,$63,$F2,$0D,$78,$46,$88,$E9,$AF
	.dc.b	$E7,$7C,$6F,$CE,$12,$3F,$38,$D8,$FC,$E3,$5E,$11,$A2,$3A,$6B,$FA
	.dc.b	$9F,$17,$F5,$04,$8F,$D4,$3F,$E3,$1F,$AC,$7E,$25,$FE,$46,$DB,$FC
	.dc.b	$87,$18,$AF,$F2,$0F,$B0,$F8,$9D,$47,$E9,$71,$56,$DF,$F1,$17,$13
	.dc.b	$2F,$04,$7E,$D5,$AF,$EF,$7E,$0F,$F7,$84,$BF,$DE,$3E,$5C,$FB,$CF
	.dc.b	$EC,$F0,$ED,$9F,$F9,$63,$3F,$26,$1F,$E5,$69,$CB,$FB,$BC,$3A,$5F
	.dc.b	$F0,$56,$BF,$C3,$73,$FE,$18,$49,$FE,$18,$F8,$5E,$98,$F8,$A5,$87
	.dc.b	$81,$FF,$20,$61,$E0,$7F,$84,$67,$97,$B5,$F4,$23,$F2,$4F,$67,$C6
	.dc.b	$FD,$32,$46,$63,$2D,$7E,$7E,$1E,$3B,$F0,$47,$93,$1C,$4E,$6F,$9C
	.dc.b	$F3,$A4,$62,$3C,$51,$F6,$5B,$3F,$05,$E0,$D7,$37,$E3,$9E,$11,$F8
	.dc.b	$36,$BF,$5E,$78,$39,$6B,$F6,$FE,$07,$F2,$44,$AE,$8F,$87,$4F,$C9
	.dc.b	$FD,$5F,$B2,$7E,$F3,$F0,$67,$85,$7B,$5A,$FB,$3F,$49,$FC,$F1,$2B
	.dc.b	$D0,$F5,$7D,$15,$D6,$7E,$DB,$F3,$67,$84,$7C,$96,$BF,$23,$E2,$BF
	.dc.b	$54,$4A,$F8,$8F,$EB,$6B,$EB,$1D,$70,$F5,$CF,$E5,$E7,$85,$7D,$56
	.dc.b	$BF,$52,$B9,$4F,$CF,$6C,$7D,$B6,$BB,$8D,$11,$D3,$5D,$1F,$83,$F1
	.dc.b	$3F,$81,$12,$3E,$E3,$63,$51,$82,$38,$21,$F8,$A8,$E0,$FE,$77,$E0
	.dc.b	$1E,$FF,$FA,$FD,$B5,$F9,$B6,$CF,$E7,$DA,$F3,$8F,$F6,$8C,$3E,$2F
	.dc.b	$F4,$17,$FB,$38,$C4,$EA,$FB,$11,$EC,$F8,$7D,$A9,$1E,$16,$E1,$E1
	.dc.b	$C3,$A1,$DA,$7C,$2F,$79,$DC,$FB,$5A,$FF,$23,$BF,$F2,$23,$EB,$F1
	.dc.b	$7E,$01,$2F,$E0,$5F,$67,$F8,$C6,$36,$BF,$D9,$57,$C9,$F9,$57,$C5
	.dc.b	$E9,$5C,$88,$FA,$5A,$FD,$1B,$FA,$04,$8C,$C7,$9E,$30,$F1,$21,$F6
	.dc.b	$7F,$61,$FB,$04,$9F,$B0,$F0,$D7,$FD,$8D,$3F,$E2,$7C,$1F,$15,$EB
	.dc.b	$9C,$CA,$73,$7F,$72,$FA,$4F,$E0,$5A,$FE,$26,$FF,$10,$49,$FC,$43
	.dc.b	$63,$F1,$0D,$78,$46,$88,$E9,$AF,$E5,$7D,$BF,$CA,$12,$3F,$28,$FA
	.dc.b	$23,$FA,$F7,$D8,$3D,$CF,$E8,$4D,$AF,$CA,$DE,$27,$F2,$97,$CE,$FF
	.dc.b	$38,$D7,$FA,$05,$F1,$7F,$9F,$35,$CC,$F8,$5F,$F2,$94,$65,$F1,$C6
	.dc.b	$54,$65,$DC,$66,$31,$1A,$8C,$12,$FF,$A6,$4E,$87,$B9,$C4,$91,$A6
	.dc.b	$D7,$A1,$AF,$B1,$AF,$C8,$6B,$EA,$7C,$9E,$E4,$9F,$0B,$67,$E1,$B5
	.dc.b	$F8,$ED,$7E,$9B,$5F,$8A,$D7,$E1,$F8,$9F,$04,$93,$EE,$6C,$7D,$06
	.dc.b	$BF,$51,$AF,$E0,$1A,$FD,$27,$E6,$7B,$92,$7E,$4B,$67,$ED,$B5,$FB
	.dc.b	$ED,$7F,$16,$D7,$EE,$B7,$DD,$7C,$73,$D8,$FC,$14,$61,$B1,$9B,$FF
	.dc.b	$03,$4E,$5F,$F0,$40,$F3,$1F,$03,$FE,$2A,$8C,$72,$59,$CC,$7F,$A7
	.dc.b	$F9,$3F,$D6,$D3,$FE,$30,$BE,$99,$FF,$85,$DA,$FF,$90,$5A,$FF,$A9
	.dc.b	$DA,$FE,$63,$C8,$FC,$91,$2F,$F1,$ED,$8F,$D0,$B5,$FD,$2B,$5F,$D8
	.dc.b	$B5,$E9,$1F,$98,$F2,$3F,$3C,$47,$04,$8F,$D6,$C6,$5B,$19,$7F,$67
	.dc.b	$19,$6C,$7E,$93,$C7,$FD,$51,$1C,$12,$3F,$73,$19,$6C,$65,$FD,$FC
	.dc.b	$65,$B1,$FB,$0F,$7B,$F6,$C4,$70,$48,$FE,$0B,$19,$6C,$65,$FE,$17
	.dc.b	$19,$6C,$7E,$F3,$DC,$FE,$04,$47,$04,$8F,$E2,$31,$96,$C6,$5F,$E2
	.dc.b	$F1,$96,$C7,$F9,$4F,$EC,$5E,$EF,$F1,$36,$C6,$63,$04,$70,$49,$FF
	.dc.b	$2F,$4F,$06,$CE,$DB,$3B,$7F,$CD,$12,$F7,$7C,$04,$9F,$23,$67,$AB
	.dc.b	$5E,$2D,$7F,$D3,$53,$CE,$78,$92,$3C,$BE,$EF,$B4,$91,$F0,$5B,$1A
	.dc.b	$8C,$11,$C1,$0F,$4C,$70,$7D,$7E,$DF,$C9,$24,$FC,$D6,$C7,$C6,$6B
	.dc.b	$C2,$34,$47,$4D,$7E,$3F,$BB,$F5,$49,$1F,$65,$B1,$F4,$9A,$F0,$8D
	.dc.b	$11,$D3,$5F,$A7,$EE,$FE,$04,$91,$F8,$56,$C7,$DE,$6B,$C2,$34,$47
	.dc.b	$4D,$7E,$FF,$BB,$F8,$D2,$47,$E4,$5B,$1F,$8A,$6B,$C2,$34,$47,$4D
	.dc.b	$7F,$D6,$5F,$64,$7F,$B6,$36,$3F,$D6,$0D,$8F,$F5,$A3,$5E,$27,$02
	.dc.b	$47,$06,$9C,$0A,$70,$7A,$FB,$DE,$02,$38,$24,$78,$23,$2D,$8C,$BE
	.dc.b	$A8,$CB,$63,$D3,$EF,$7B,$08,$E0,$91,$EC,$8C,$B6,$32,$FC,$68,$CB
	.dc.b	$63,$E2,$FB,$3F,$20,$8E,$09,$1F,$22,$32,$D8,$CB,$F4,$A3,$2D,$8F
	.dc.b	$A3,$EF,$7D,$42,$38,$24,$7D,$48,$CB,$63,$2F,$DE,$8C,$B6,$3E,$EF
	.dc.b	$BB,$F8,$02,$38,$24,$7E,$02,$32,$D8,$CB,$F8,$A8,$CB,$63,$F1,$3E
	.dc.b	$EF,$E3,$08,$E0,$91,$F8,$C8,$CB,$63,$2F,$E5,$A3,$2D,$8F,$CA,$FB
	.dc.b	$BF,$98,$23,$82,$47,$E6,$23,$2D,$8C,$BF,$A2,$8C,$B6,$3F,$48,$C1
	.dc.b	$F1,$9F,$E4,$C2,$78,$DF,$E5,$36,$CE,$E7,$F5,$91,$87,$0D,$7A,$35
	.dc.b	$EE,$F5,$F7,$BA,$12,$7A,$4F,$8A,$30,$F4,$6B,$D1,$AF,$46,$BE,$47
	.dc.b	$D1,$E3,$49,$EB,$5F,$E6,$26,$C6,$6B,$D2,$D7,$D2,$D7,$D2,$D7,$E0
	.dc.b	$BF,$17,$C6,$93,$D9,$B3,$F1,$1A,$FC,$46,$BF,$11,$AF,$CC,$6B,$F5
	.dc.b	$3C,$6F,$A4,$49,$F5,$B6,$3E,$8B,$5F,$A2,$D7,$E8,$B5,$FB,$2F,$DD
	.dc.b	$F1,$A4,$FC,$66,$CF,$DC,$6B,$F7,$1A,$FD,$C6,$BF,$84,$6B,$F8,$CF
	.dc.b	$77,$F1,$44,$9F,$A6,$D8,$FC,$4B,$5F,$C4,$B5,$FC,$4B,$5F,$C8,$BF
	.dc.b	$95,$F7,$64,$FD,$E6,$CF,$E5,$1A,$FE,$51,$AF,$E5,$1A,$FE,$71,$AF
	.dc.b	$E9,$3D,$DF,$D1,$12,$7F,$16,$D8,$FD,$0B,$5F,$D0,$B5,$FD,$0B,$5F
	.dc.b	$D4,$BF,$AD,$F7,$64,$FE,$59,$B3,$FA,$C6,$BF,$AC,$6B,$FA,$C6,$BF
	.dc.b	$B4,$6B,$FB,$CF,$6F,$F7,$44,$9F,$D1,$B6,$3F,$72,$D7,$F7,$2D,$7F
	.dc.b	$72,$D7,$F8,$0B,$FE,$D7,$FB,$37,$B9,$49,$5F,$BC,$8C,$B6,$33,$5F
	.dc.b	$C2,$E7,$2D,$8C,$BF,$C5,$63,$2D,$8C,$BF,$C6,$A1,$ED,$F8,$C9,$5E
	.dc.b	$16,$CF,$76,$BD,$5A,$F3,$6B,$FE,$09,$9E,$B3,$EB,$F6,$FE,$01,$1C
	.dc.b	$12,$7D,$91,$96,$C6,$5F,$64,$65,$B1,$F1,$FD,$BF,$98,$47,$04,$8F
	.dc.b	$91,$19,$6C,$65,$F9,$11,$96,$C7,$D3,$F6,$FE,$C1,$1C,$12,$3E,$A4
	.dc.b	$65,$B1,$97,$EA,$46,$5B,$1F,$7F,$DB,$FC,$21,$1C,$12,$3F,$01,$19
	.dc.b	$6C,$65,$FC,$04,$65,$B1,$F8,$BF,$67,$F2,$04,$70,$48,$FC,$64,$65
	.dc.b	$B1,$97,$F1,$91,$96,$C7,$F5,$2F,$8D,$7B,$88,$C3,$67,$FB,$14,$EA
	.dc.b	$72,$48,$CC,$ED,$FE,$DD,$3B,$73,$3B,$73,$3B,$7F,$BA,$CE,$DC,$CF
	.dc.b	$97,$DC,$D4,$7F,$97,$65,$18,$84,$62,$78,$12,$3D,$33,$B9,$D4,$62
	.dc.b	$74,$FA,$27,$44,$8F,$8D,$EE,$71,$23,$82,$4F,$C6,$8C,$B6,$32,$FC
	.dc.b	$48,$CB,$63,$E9,$7A,$5E,$52,$38,$24,$7D,$28,$CB,$63,$2F,$D0,$8C
	.dc.b	$B6,$3E,$F7,$B7,$F0,$48,$E0,$91,$F7,$A3,$2D,$8C,$BF,$72,$32,$D8
	.dc.b	$FC,$57,$B7,$F3,$48,$E0,$91,$F8,$A8,$CB,$63,$2F,$E2,$23,$2D,$8F
	.dc.b	$CB,$7C,$5F,$D9,$23,$82,$47,$E5,$A3,$2D,$8C,$BF,$94,$8C,$B6,$3F
	.dc.b	$45,$EE,$7E,$14,$8E,$09,$1F,$A2,$8C,$B6,$32,$FE,$82,$32,$D8,$FD
	.dc.b	$77,$83,$F9,$12,$38,$24,$7E,$BA,$32,$D8,$CB,$FA,$C8,$CB,$63,$FC
	.dc.b	$97,$F6,$8F,$0B,$F4,$8D,$8C,$C6,$08,$E0,$93,$FE,$5F,$9E,$0D,$9D
	.dc.b	$B6,$76,$FF,$9B,$3D,$77,$E1,$CC,$FF,$6F,$18,$6C,$66,$BF,$CD,$93
	.dc.b	$97,$FC,$DC,$3D,$A7,$91,$FC,$47,$DA,$FE,$25,$39,$F9,$5C,$3F,$ED
	.dc.b	$DA,$DB,$FE,$E1,$9D,$C2,$77,$93,$0F,$03,$FE,$83,$6B,$EB,$C7,$CC
	.dc.b	$E5,$AE,$77,$F5,$93,$1F,$5C,$A3,$F8,$BC,$7A,$F8,$3C,$2D,$3C,$C7
	.dc.b	$A7,$CA,$CE,$49,$F3,$E3,$F3,$2F,$B6,$3E,$36,$BF,$E9,$BD,$DE,$1F
	.dc.b	$E0,$09,$97,$E1,$1B,$3F,$75,$AF,$16,$BF,$EA,$69,$E7,$3C,$49,$1F
	.dc.b	$1F,$C6,$FC,$2B,$F8,$16,$A3,$E7,$38,$6F,$0F,$C4,$CD,$F1,$3F,$03
	.dc.b	$8B,$E0,$B7,$0F,$8B,$7C,$1F,$C3,$E1,$CF,$43,$49,$E4,$7E,$41,$FF
	.dc.b	$50,$57,$9E,$30,$78,$7C,$BF,$9E,$7D,$0D,$AE,$1E,$0C,$3A,$D7,$08
	.dc.b	$F0,$B6,$FF,$4D,$EC,$FE,$B8,$93,$FB,$26,$C6,$A3,$04,$70,$43,$ED
	.dc.b	$47,$07,$E6,$7C,$2F,$AC,$93,$F0,$9F,$95,$D1,$F8,$8E,$8F,$C4,$B6
	.dc.b	$7D,$5A,$2B,$8D,$FD,$96,$BF,$67,$DD,$F9,$04,$AF,$9E,$FE,$6B,$F7
	.dc.b	$FF,$F0,$24,$7D,$40,$E2,$5D,$3F,$B6,$8F,$E1,$2D,$9F,$CD,$E0,$BE
	.dc.b	$D7,$F8,$B6,$BF,$EF,$5E,$3F,$F7,$AD,$7F,$B9,$9B,$1F,$EF,$36,$BB
	.dc.b	$F3,$BF,$C0,$51,$88,$FF,$7F,$B5,$FC,$A3,$5F,$E4,$D1,$C0,$93,$DD
	.dc.b	$B1,$F9,$FF,$8A,$FC,$61,$27,$F2,$6F,$C5,$F4,$3F,$7B,$9F,$E4,$33
	.dc.b	$F8,$51,$8E,$0F,$83,$5C,$70,$E9,$87,$23,$F3,$BC,$1F,$99,$23,$83
	.dc.b	$F9,$2E,$3F,$A0,$7E,$93,$67,$E8,$C7,$EC,$4F,$AB,$E8,$79,$36,$FA
	.dc.b	$37,$D2,$BF,$74,$D7,$E8,$79,$1F,$28,$97,$F4,$5B,$1F,$45,$AF,$08
	.dc.b	$D1,$1D,$35,$FE,$0D,$E6,$FE,$D4,$91,$FB,$C7,$F1,$DE,$87,$CA,$E2
	.dc.b	$F9,$6D,$9F,$CD,$CF,$0A,$FE,$24,$D7,$F8,$97,$C7,$FF,$01,$24,$FF
	.dc.b	$08,$7F,$35,$DB,$F5,$8F,$AF,$C9,$67,$F4,$F3,$C2,$BF,$8F,$35,$FC
	.dc.b	$A7,$AF,$F8,$E2,$4F,$E5,$5B,$1F,$95,$6B,$C2,$34,$47,$4D,$7F,$B5
	.dc.b	$7E,$21,$FE,$C0,$7F,$95,$74,$A3,$11,$FD,$65,$B1,$9F,$43,$F6,$DF
	.dc.b	$03,$F9,$2F,$53,$F2,$6D,$C3,$F8,$E5,$F0,$25,$7F,$76,$71,$F0,$74
	.dc.b	$93,$FD,$DA,$11,$FC,$0F,$E0,$BF,$3E,$7E,$83,$9E,$D7,$69,$E3,$9F
	.dc.b	$F6,$DB,$E3,$3C,$E7,$FC,$F1,$F7,$0F,$87,$A4,$9F,$E9,$57,$04,$EE
	.dc.b	$31,$1F,$D3,$E3,$0E,$1A,$F1,$6B,$D1,$AF,$19,$F0,$3C,$5A,$FD,$4E
	.dc.b	$5F,$8C,$4A,$F8,$8F,$F2,$5F,$63,$F8,$0F,$3F,$F0,$26,$CF,$D0,$9E
	.dc.b	$15,$F0,$DA,$FE,$03,$E4,$7E,$91,$27,$E8,$3F,$CC,$7E,$3B,$F8,$69
	.dc.b	$FC,$3B,$9E,$CF,$DC,$9E,$15,$F3,$DA,$FC,$1F,$43,$E7,$92,$7E,$4C
	.dc.b	$7C,$C8,$C3,$F2,$5A,$F1,$8D,$E8,$9C,$CF,$CB,$6B,$F9,$5F,$5B,$F1
	.dc.b	$84,$AF,$C5,$3F,$D2,$FE,$1B,$F9,$3F,$47,$F2,$86,$CF,$E5,$A7,$85
	.dc.b	$7E,$25,$AF,$E9,$BE,$5F,$F3,$04,$9F,$CB,$3F,$D5,$FD,$CF,$E6,$67
	.dc.b	$F8,$C3,$C8,$7C,$3E,$2B,$F4,$75,$C6,$FF,$2C,$D7,$F1,$1E,$8F,$E5
	.dc.b	$89,$3F,$09,$B1,$F8,$F6,$BC,$23,$44,$74,$D7,$F7,$FF,$05,$FB,$12
	.dc.b	$47,$EB,$DF,$F2,$4E,$1F,$D8,$3C,$8F,$F3,$0B,$67,$F9,$F2,$EB,$F7
	.dc.b	$79,$3F,$E4,$8E,$A6,$1D,$31,$FD,$93,$5F,$E1,$FE,$D7,$F0,$12,$63
	.dc.b	$FB,$F7,$C3,$F0,$FF,$27,$B3,$E5,$36,$7F,$85,$CF,$0A,$FD,$F3,$5F
	.dc.b	$DC,$7D,$CF,$F8,$15,$B3,$FE,$F7,$8C,$47,$EF,$9A,$FF,$92,$5A,$9A
	.dc.b	$48,$E4,$43,$F3,$6E,$5F,$F4,$6F,$E2,$2F,$84,$FF,$49,$B6,$33,$18
	.dc.b	$3E,$7F,$C1,$7E,$1C,$FC,$47,$C4,$59,$FF,$22,$57,$1C,$3F,$D1,$0D
	.dc.b	$7F,$D4,$4D,$7F,$9C,$B5,$FF,$48,$2F,$7B,$C6,$49,$F3,$BF,$80,$F9
	.dc.b	$7F,$CC,$78,$1F,$99,$6C,$F8,$A7,$85,$F8,$9A,$FD,$0F,$37,$E7,$12
	.dc.b	$7E,$2B,$63,$F9,$4E,$9F,$E4,$59,$C9,$23,$33,$B7,$A4,$65,$F7,$FC
	.dc.b	$2F,$CC,$24,$7D,$07,$F3,$3F,$13,$FB,$13,$F6,$5E,$0D,$9F,$97,$3C
	.dc.b	$2B,$E5,$B5,$F9,$7F,$03,$F6,$09,$3F,$71,$FD,$2F,$C2,$FE,$F7,$E2
	.dc.b	$7F,$7C,$D9,$FA,$F3,$C2,$BE,$BB,$5F,$CA,$F8,$FF,$93,$24,$FE,$29
	.dc.b	$B1,$F1,$9A,$F0,$8D,$11,$D3,$5F,$C1,$F6,$FE,$40,$91,$F9,$47,$F7
	.dc.b	$BF,$2B,$FC,$54,$FE,$2D,$E1,$59,$FC,$7C,$F0,$AF,$C7,$B5,$FC,$7F
	.dc.b	$BD,$F9,$C2,$4F,$E8,$1F,$E1,$3E,$8F,$F2,$1F,$0B,$F9,$13,$67,$F3
	.dc.b	$73,$C2,$BF,$36,$D7,$FB,$87,$DD,$FF,$28,$9F,$E8,$ED,$8F,$EE,$2D
	.dc.b	$77,$E2,$27,$6D,$8D,$B5,$DC,$68,$8E,$9A,$E8,$FD,$7F,$BB,$FB,$A2
	.dc.b	$47,$EF,$9F,$E4,$FF,$11,$FC,$FC,$FE,$81,$E0,$D9,$FD,$CC,$F0,$AE
	.dc.b	$71,$C4,$FE,$0F,$E2,$FF,$06,$23,$83,$FC,$1E,$17,$D3,$FF,$4D,$F0
	.dc.b	$7F,$A7,$36,$7F,$83,$CF,$2A,$FE,$0E,$D7,$E6,$F6,$7C,$37,$FC,$3F
	.dc.b	$A2,$31,$1F,$EE,$B8,$C5,$7D,$29,$CC,$A7,$2F,$D4,$8C,$C6,$23,$BC
	.dc.b	$F1,$23,$B6,$BF,$E3,$0F,$55,$F8,$97,$FB,$AA,$30,$D7,$FC,$AA,$D5
	.dc.b	$1C,$3D,$DF,$4D,$3F,$80,$3F,$D2,$95,$FE,$96,$3F,$0C,$E3,$CB,$FC
	.dc.b	$34,$EF,$D4,$FC,$44,$62,$75,$3E,$0B,$F8,$B8,$77,$8F,$F2,$1B,$70
	.dc.b	$FF,$24,$35,$F4,$FB,$BC,$32,$34,$95,$F7,$B0,$7C,$5F,$8A,$3F,$D5
	.dc.b	$91,$FD,$07,$2F,$24,$E7,$0F,$5F,$93,$72,$F9,$38,$F4,$9F,$1E,$5E
	.dc.b	$0C,$BC,$ED,$8F,$D5,$F5,$7D,$82,$4F,$EC,$1B,$1A,$8C,$11,$C1,$0F
	.dc.b	$DC,$C7,$07,$EC,$FB,$7F,$04,$93,$E2,$7E,$BE,$8F,$98,$F6,$4F,$D9
	.dc.b	$6C,$7F,$59,$85,$7D,$B7,$46,$35,$FE,$01,$9F,$E2,$B0,$EF,$5F,$09
	.dc.b	$AF,$E3,$3D,$CF,$A8,$4B,$F8,$AF,$93,$E3,$BC,$3F,$25,$E2,$6C,$F7
	.dc.b	$3F,$27,$3C,$6B,$E7,$B5,$FF,$45,$70,$7E,$49,$A7,$EA,$9C,$3F,$E9
	.dc.b	$06,$BE,$E6,$BF,$84,$38,$12,$7F,$CD,$71,$A2,$46,$4F,$C0,$FB,$7F
	.dc.b	$91,$24,$FE,$09,$FA,$3E,$C7,$CE,$3E,$7F,$8F,$67,$DD,$5F,$AE,$E1
	.dc.b	$5F,$28,$EB,$3C,$CF,$C6,$FB,$5F,$9F,$23,$83,$F3,$DF,$25,$F7,$3C
	.dc.b	$7F,$BA,$D8,$EF,$C1,$FC,$0A,$B8,$DF,$E6,$9A,$FC,$CD,$FE,$51,$2B
	.dc.b	$E1,$B6,$3F,$D6,$FA,$A3,$11,$C6,$76,$47,$71,$A7,$F8,$7F,$B5,$FB
	.dc.b	$92,$4F,$EA,$9F,$CC,$FC,$77,$E5,$8F,$CB,$FC,$75,$9F,$5E,$CA,$E3
	.dc.b	$7F,$B4,$6B,$FC,$87,$D1,$FE,$B8,$FF,$BC,$9A,$FF,$1E,$6E,$9F,$E3
	.dc.b	$BF,$17,$F5,$3F,$1B,$FA,$6F,$2B,$F4,$ED,$CB,$E5,$DF,$02,$57,$F1
	.dc.b	$A7,$AE,$1F,$C4,$2B,$F2,$84,$AF,$C6,$B6,$3E,$5B,$5F,$CE,$B5,$FC
	.dc.b	$D3,$5F,$CD,$3D,$FD,$89,$3F,$CA,$2B,$FD,$EB,$5F,$CA,$1A,$FF,$BC
	.dc.b	$A3,$F9,$D7,$C5,$FF,$02,$3F,$81,$F9,$56,$3D,$B7,$D2,$B9,$35,$F9
	.dc.b	$9E,$BF,$BC,$93,$E1,$7F,$88,$FC,$5F,$F0,$EF,$2B,$F8,$7B,$67,$E4
	.dc.b	$CF,$0A,$F3,$35,$FD,$EF,$91,$FB,$B2,$4F,$ED,$1B,$1F,$92,$6B,$C2
	.dc.b	$34,$47,$4D,$7F,$09,$EB,$FD,$72,$47,$C9,$7F,$94,$FC,$5F,$F2,$63
	.dc.b	$F9,$3D,$59,$FC,$0C,$F0,$AF,$9C,$D7,$F2,$1E,$AF,$E0,$C9,$3F,$55
	.dc.b	$FE,$D7,$FC,$03,$FE,$2D,$8F,$E0,$61,$7C,$67,$F3,$87,$55,$FC,$D5
	.dc.b	$CC,$FE,$4A,$BA,$CF,$DF,$6B,$FC,$F0,$FE,$7B,$F7,$EF,$8A,$FF,$07
	.dc.b	$B6,$B5,$3B,$8C,$38,$0E,$1D,$B5,$E1,$3D,$09,$3C,$63,$8C,$61,$E9
	.dc.b	$3C,$5F,$2C,$7E,$A7,$C9,$95,$F9,$B6,$CF,$9A,$3E,$37,$C6,$FC,$53
	.dc.b	$E3,$79,$B6,$3F,$6B,$5C,$2B,$F8,$13,$5F,$E0,$5E,$B7,$ED,$C9,$3F
	.dc.b	$A9,$7E,$AF,$C6,$7D,$1F,$3B,$E9,$36,$7F,$80,$CF,$0A,$FD,$73,$5F
	.dc.b	$8B,$F1,$3F,$38,$93,$F0,$F4,$46,$27,$E1,$B5,$E3,$3B,$8C,$47,$D5
	.dc.b	$6B,$FC,$67,$D6,$FE,$20,$49,$FE,$04,$FE,$3B,$5F,$C5,$1F,$8B,$F7
	.dc.b	$AC,$FF,$1A,$9E,$15,$FC,$1D,$AF,$F2,$9F,$5F,$F8,$E1,$27,$F8,$73
	.dc.b	$F9,$AF,$8C,$FC,$B7,$A1,$F9,$76,$CF,$F2,$69,$E1,$5F,$C4,$DA,$FE
	.dc.b	$2B,$E1,$FF,$26,$49,$F8,$CD,$8F,$C3,$B5,$E1,$1A,$23,$A6,$BF,$E6
	.dc.b	$1F,$C5,$7F,$DF,$1E,$21,$A4,$F5,$BF,$67,$F1,$BF,$AF,$3F,$61,$E9
	.dc.b	$DC,$3F,$B4,$D6,$DF,$F2,$AB,$70,$FF,$2A,$BF,$E2,$B6,$F8,$9F,$E5
	.dc.b	$C8,$C4,$78,$E3,$F8,$0F,$C6,$FE,$F7,$8B,$F7,$CD,$BE,$F3,$C4,$8F
	.dc.b	$81,$FF,$43,$FE,$02,$F0,$69,$27,$F6,$11,$96,$C6,$70,$FD,$E4,$E5
	.dc.b	$B1,$97,$F8,$1C,$65,$B1,$97,$FC,$8F,$AF,$CB,$F3,$BD,$C4,$BF,$8A
	.dc.b	$FF,$1E,$F8,$1F,$E3,$47,$F1,$BF,$52,$CF,$CC,$9E,$17,$F2,$9A,$FD
	.dc.b	$9F,$57,$E5,$12,$7E,$8B,$FC,$AB,$E3,$BF,$93,$7A,$9F,$C9,$DB,$3F
	.dc.b	$62,$78,$57,$D6,$6B,$EE,$F8,$2F,$96,$49,$F9,$2D,$8F,$8E,$D7,$84
	.dc.b	$68,$8E,$9A,$FE,$47,$D3,$FC,$11,$23,$F1,$2F,$F4,$2F,$8D,$FE,$7C
	.dc.b	$7F,$3F,$F4,$EC,$FE,$42,$78,$57,$E3,$9A,$FE,$77,$D3,$FC,$71,$27
	.dc.b	$F2,$AF,$F4,$EE,$9F,$F4,$0F,$D0,$7F,$BF,$FD,$CF,$F5,$AF,$94,$9C
	.dc.b	$FA,$BF,$E9,$FE,$A6,$1C,$F2,$FC,$EB,$5F,$C2,$F7,$7E,$48,$99,$FE
	.dc.b	$41,$B1,$F8,$E6,$BC,$23,$44,$74,$D7,$F7,$1E,$9F,$EA,$49,$1F,$B0
	.dc.b	$7F,$DA,$7A,$3E,$37,$DB,$F1,$DE,$E6,$D7,$EE,$2B,$8D,$FE,$D9,$AF
	.dc.b	$F6,$5D,$9F,$7D,$FE,$44,$6E,$BF,$EF,$86,$DF,$F6,$36,$D7,$F5,$3C
	.dc.b	$9F,$1D,$F3,$8F,$9F,$EB,$DB,$FF,$3F,$DF,$0B,$F0,$63,$DC,$98,$AB
	.dc.b	$FB,$47,$E0,$4F,$81,$C5,$FF,$73,$8C,$4E,$A7,$33,$FD,$EA,$72,$EE
	.dc.b	$32,$FF,$80,$E3,$2E,$E3,$2F,$FB,$0A,$32,$EE,$32,$A3,$2E,$E3,$31
	.dc.b	$8C,$1F,$BB,$7F,$85,$AB,$FC,$1B,$18,$F5,$7E,$64,$E6,$7E,$93,$5F
	.dc.b	$F6,$9E,$0D,$5F,$21,$F9,$AF,$0B,$F3,$6D,$9F,$9B,$7C,$6D,$8F,$79
	.dc.b	$FB,$13,$98,$C4,$FF,$38,$F6,$3F,$9B,$12,$BF,$9E,$BF,$8E,$F8,$8F
	.dc.b	$D4,$9F,$AA,$F6,$6C,$FF,$44,$9E,$15,$CE,$38,$9F,$6B,$AF,$E2,$91
	.dc.b	$C1,$23,$EE,$46,$5B,$19,$7E,$E4,$65,$B1,$FE,$35,$F1,$5F,$8B,$FC
	.dc.b	$16,$EF,$F7,$1B,$3B,$9F,$8C,$FC,$DF,$8F,$FE,$0B,$ED,$FF,$06,$6C
	.dc.b	$7F,$B2,$27,$A1,$2F,$A3,$F2,$F6,$7B,$5F,$70,$95,$F4,$1F,$DB,$FC
	.dc.b	$77,$F1,$23,$F8,$9F,$C7,$D9,$FB,$33,$C3,$4F,$C3,$DF,$49,$E4,$7E
	.dc.b	$93,$DE,$FC,$79,$1C,$12,$7F,$4D,$19,$6C,$65,$FD,$34,$65,$B1,$FB
	.dc.b	$DE,$9F,$DD,$B7,$AF,$F8,$14,$62,$7F,$36,$D7,$FD,$59,$CE,$F1,$BF
	.dc.b	$97,$FB,$BF,$CC,$1B,$87,$F0,$2B,$E1,$63,$F2,$EE,$1B,$87,$80,$F9
	.dc.b	$FE,$EC,$AF,$B4,$D9,$FE,$25,$F1,$FF,$CF,$8F,$E7,$FE,$ED,$AF,$B5
	.dc.b	$5B,$AF,$F4,$27,$2F,$E2,$23,$11,$FE,$87,$F0,$D1,$FC,$01,$B5,$FE
	.dc.b	$7B,$6C,$FF,$A1,$5A,$FF,$A2,$9A,$F3,$9E,$44,$9E,$4D,$8E,$4D,$7C
	.dc.b	$BC,$6F,$6F,$F1,$E4,$AF,$CD,$3F,$D0,$BE,$3F,$FA,$EF,$B9,$FD,$79
	.dc.b	$B3,$F9,$99,$E1,$7F,$E9,$EC,$BF,$4F,$18,$8F,$E6,$7D,$FF,$78,$95
	.dc.b	$F8,$37,$FA,$7F,$C7,$FF,$6A,$3F,$B5,$F9,$16,$7E,$E4,$F0,$AF,$C2
	.dc.b	$B5,$F8,$9E,$BF,$CB,$24,$FC,$B6,$C7,$CB,$6B,$C2,$34,$47,$4D,$7F
	.dc.b	$94,$E1,$FC,$B0,$91,$FC,$81,$FE,$7B,$F4,$1F,$F0,$06,$7F,$92,$CF
	.dc.b	$E4,$63,$F2,$5F,$1F,$FD,$3B,$C4,$FE,$9F,$1B,$6D,$27,$55,$F9,$99
	.dc.b	$FE,$7F,$7E,$0A,$FD,$CB,$5F,$B7,$EF,$FF,$08,$25,$7F,$13,$7C,$B0
	.dc.b	$F8,$AE,$AB,$1F,$D3,$46,$1F,$D4,$7B,$F6,$78,$5F,$F1,$7B,$EB,$1F
	.dc.b	$7F,$89,$18,$8F,$CB,$7A,$7F,$9D,$25,$7E,$75,$B1,$F9,$D6,$BC,$23
	.dc.b	$44,$74,$D7,$F3,$3E,$F7,$F2,$62,$47,$F3,$17,$DF,$F1,$FE,$9F,$7B
	.dc.b	$D4,$D9,$FE,$61,$3C,$2B,$D4,$D7,$FB,$3E,$1F,$DA,$89,$3F,$D7,$9F
	.dc.b	$97,$E7,$7C,$53,$E3,$7B,$76,$7F,$B3,$CF,$0A,$FE,$5E,$D7,$F8,$A1
	.dc.b	$FC,$57,$F0,$5F,$EE,$33,$FE,$FB,$6C,$68,$D7,$27,$F8,$1A,$33,$19
	.dc.b	$6B,$D1,$AF,$09,$E9,$3C,$48,$E0,$91,$FC,$C3,$CD,$FE,$66,$48,$FE
	.dc.b	$70,$FE,$2B,$E3,$FF,$09,$ED,$FE,$15,$B3,$FC,$C6,$78,$57,$38,$E2
	.dc.b	$7F,$65,$F8,$1F,$EB,$A4,$70,$7B,$E3,$F2,$83,$E3,$BF,$24,$7E,$4F
	.dc.b	$E3,$EC,$FF,$69,$AE,$55,$F9,$36,$BF,$17,$CC,$F9,$04,$9F,$9C,$D8
	.dc.b	$F0,$68,$8C,$47,$19,$D9,$1D,$C6,$9F,$F5,$BF,$23,$E0,$7F,$BC,$92
	.dc.b	$BF,$C0,$8F,$EC,$BE,$3F,$F5,$87,$EB,$7D,$3B,$3F,$E0,$39,$E1,$7F
	.dc.b	$AD,$6B,$FE,$5C,$F4,$FF,$B3,$92,$7F,$B6,$3F,$BE,$F8,$EF,$DC,$7B
	.dc.b	$BF,$B9,$6C,$FF,$65,$9E,$15,$FD,$B9,$AF,$E2,$FC,$3F,$C7,$12,$7F
	.dc.b	$28,$D8,$F8,$ED,$78,$46,$88,$E9,$AF,$F9,$4B,$6F,$F2,$A9,$23,$FC
	.dc.b	$92,$FF,$16,$F8,$EF,$E2,$07,$F1,$1F,$76,$CF,$F9,$4A,$78,$57,$F8
	.dc.b	$69,$AF,$F5,$CF,$84,$FF,$14,$92,$7F,$C6,$8F,$F2,$2F,$8D,$FE,$39
	.dc.b	$F1,$DF,$C7,$5B,$3F,$E3,$39,$E1,$5F,$C7,$5A,$FE,$CF,$F0,$67,$81
	.dc.b	$FE,$78,$6D,$7E,$66,$51,$87,$F3,$8D,$79,$4F,$12,$4F,$19,$D4,$FF
	.dc.b	$06,$6B,$FE,$1B,$6B,$DA,$78,$BF,$CC,$FC,$1F,$F2,$49,$2B,$FC,$A8
	.dc.b	$FF,$42,$F5,$BF,$9E,$1F,$CF,$7D,$CB,$3F,$E5,$39,$E1,$5C,$E3,$89
	.dc.b	$FE,$7E,$E3,$FF,$46,$91,$C1,$FB,$9F,$15,$F8,$37,$FA,$BE,$3F,$AD
	.dc.b	$1F,$D8,$DF,$F4,$C7,$93,$FD,$89,$D5,$FF,$68,$3F,$AA,$3A,$9F,$EC
	.dc.b	$5E,$CF,$EB,$5E,$F3,$FE,$5A,$C3,$C4,$FC,$EE,$9F,$A8,$49,$F9,$8D
	.dc.b	$8D,$46,$08,$E0,$87,$DF,$8E,$0F,$FB,$33,$D7,$FF,$2B,$92,$7F,$CB
	.dc.b	$8F,$F8,$0B,$07,$C8,$7F,$6B,$6C,$FC,$2A,$F6,$FB,$79,$AF,$EE,$79
	.dc.b	$FF,$B9,$B0,$E9,$5F,$E6,$26,$BF,$E4,$CF,$C6,$1F,$07,$FE,$91,$6E
	.dc.b	$3F,$E9,$76,$CF,$F8,$CD,$AF,$F5,$0D,$BE,$AF,$C8,$F9,$7D,$7F,$33
	.dc.b	$6F,$CD,$EA,$B0,$E3,$87,$F9,$AE,$31,$8F,$84,$95,$F9,$9E,$9F,$D2
	.dc.b	$11,$C1,$2B,$F4,$71,$96,$C6,$5F,$D7,$46,$5B,$1F,$D1,$7E,$A1,$F9
	.dc.b	$47,$FB,$4D,$BF,$05,$FE,$EE,$6C,$EE,$7F,$DB,$1A,$FE,$53,$E4,$BE
	.dc.b	$A7,$AD,$F5,$5B,$5F,$56,$F8,$E3,$FD,$C5,$AF,$F7,$86,$BF,$E8,$D6
	.dc.b	$BF,$E8,$EF,$7B,$FD,$36,$49,$FF,$50,$3F,$85,$F9,$2F,$C2,$1F,$85
	.dc.b	$F4,$AC,$FF,$A6,$67,$85,$7F,$A8,$9A,$FF,$80,$FF,$09,$71,$2B,$F8
	.dc.b	$84,$61,$B1,$99,$FE,$D9,$39,$6C,$65,$E8,$D7,$FC,$2B,$1B,$73,$3D
	.dc.b	$08,$F2,$6B,$FE,$1C,$E5,$75,$BE,$3F,$A4,$A7,$31,$86,$D7,$F8,$05
	.dc.b	$B3,$FE,$21,$DF,$F8,$5F,$C9,$FE,$AF,$D8,$FD,$63,$6F,$F5,$97,$C2
	.dc.b	$FC,$58,$F8,$67,$A3,$63,$FD,$C3,$8F,$E1,$FF,$3B,$79,$74,$97,$F8
	.dc.b	$F8,$CB,$FC,$77,$B9,$EF,$7E,$F4,$FD,$F7,$93,$67,$F1,$55,$C3,$1F
	.dc.b	$C5,$35,$FC,$53,$5F,$C5,$35,$FF,$6C,$6E,$DB,$E4,$C2,$31,$5F,$80
	.dc.b	$9C,$C6,$27,$EF,$C6,$09,$3B,$24,$69,$AF,$E3,$63,$F5,$5E,$A7,$FB
	.dc.b	$A6,$7F,$60,$D9,$FD,$23,$5F,$D2,$35,$FE,$7D,$F2,$7F,$C8,$4F,$E4
	.dc.b	$5E,$25,$AF,$D7,$56,$E9,$CF,$FE,$3F,$7C,$24,$9F,$B3,$F2,$3F,$87
	.dc.b	$E1,$34,$91,$F5,$7E,$4F,$F9,$87,$99,$FC,$C5,$B1,$F8,$6A,$E2,$F4
	.dc.b	$25,$FF,$9B,$F9,$3F,$33,$87,$CC,$E4,$95,$FA,$E6,$D7,$EB,$1A,$FE
	.dc.b	$D5,$AF,$EA,$9A,$FD,$C6,$D3,$E2,$1F,$0B,$FA,$36,$DF,$E8,$DB,$7E
	.dc.b	$56,$BF,$CE,$F9,$7F,$39,$7F,$DA,$FC,$AF,$EA,$18,$FF,$51,$7F,$A6
	.dc.b	$DF,$0B,$FF,$66,$35,$FD,$A3,$6B,$C9,$8F,$E4,$3C,$87,$C8,$D2,$67
	.dc.b	$FE,$1F,$8C,$B6,$33,$87,$F9,$EE,$72,$D8,$CB,$FE,$9D,$8C,$B6,$32
	.dc.b	$FF,$2D,$FD,$45,$E1,$7F,$3D,$6D,$FF,$43,$6C,$FF,$37,$6B,$F1,$76
	.dc.b	$FF,$6E,$65,$FD,$90,$FF,$08,$38,$F1,$BF,$DC,$7C,$6B,$E3,$7E,$1C
	.dc.b	$7C,$04,$C7,$E0,$7C,$97,$C5,$23,$82,$4F,$C1,$8C,$B6,$32,$FC,$18
	.dc.b	$CB,$63,$E6,$7C,$07,$D1,$23,$82,$47,$CD,$8C,$B6,$32,$FC,$D8,$CB
	.dc.b	$63,$EC,$7C,$8F,$DD,$23,$82,$47,$D9,$8C,$B6,$32,$FD,$98,$CB,$63
	.dc.b	$F0,$9F,$23,$F8,$92,$38,$24,$7E,$16,$32,$D8,$CB,$F8,$58,$CB,$63
	.dc.b	$F2,$1F,$01,$F9,$52,$38,$24,$7E,$46,$32,$D8,$CB,$F9,$18,$CB,$63
	.dc.b	$F3,$9F,$23,$FA,$12,$38,$24,$7E,$76,$32,$D8,$CB,$F9,$D8,$CB,$63
	.dc.b	$F5,$1E,$17,$EB,$48,$E0,$91,$FA,$98,$CB,$63,$2F,$EA,$63,$2D,$8F
	.dc.b	$DA,$7C,$8F,$EE,$48,$E0,$91,$FB,$58,$CB,$63,$2F,$ED,$63,$2D,$8F
	.dc.b	$E0,$1F,$01,$FC,$14,$8E,$09,$1F,$C0,$63,$2D,$8C,$BF,$C0,$63,$2D
	.dc.b	$8F,$E1,$9E,$97,$F1,$12,$38,$24,$7F,$0D,$8C,$B6,$32,$FF,$0D,$8C
	.dc.b	$B6,$3F,$8C,$7C,$87,$F1,$D2,$38,$24,$7F,$19,$8C,$B6,$32,$FF,$19
	.dc.b	$8C,$B6,$3F,$92,$79,$5F,$CA,$48,$E0,$91,$FC,$96,$32,$D8,$CB,$FC
	.dc.b	$96,$32,$D8,$FE,$61,$F1,$FF,$CD,$48,$E0,$91,$FC,$C6,$32,$D8,$CB
	.dc.b	$FC,$C6,$32,$D8,$FE,$79,$F1,$DF,$D0,$48,$E0,$91,$FC,$F6,$32,$D8
	.dc.b	$CB,$FC,$F6,$32,$D8,$FE,$91,$D5,$FD,$34,$8E,$09,$1F,$D2,$63,$2D
	.dc.b	$8C,$BF,$D2,$63,$2D,$8F,$EA,$9E,$FF,$F5,$92,$38,$24,$7F,$55,$8C
	.dc.b	$B6,$32,$FF,$55,$8C,$B6,$3F,$B0,$7C,$77,$F6,$52,$38,$24,$7F,$61
	.dc.b	$8C,$B6,$32,$FF,$61,$8C,$B6,$0C,$27,$E1,$9F,$E2,$74,$D2,$3F,$E4
	.dc.b	$66,$C7,$18,$E2,$64,$8E,$5E,$F1,$C1,$C3,$D7,$E3,$FA,$12,$78,$4F
	.dc.b	$8A,$30,$F0,$6B,$E1,$94,$62,$3D,$6D,$7C,$2F,$A7,$E3,$E5,$76,$A4
	.dc.b	$62,$BC,$AD,$7B,$35,$F8,$ED,$7D,$AD,$7E,$4F,$C7,$FC,$72,$57,$76
	.dc.b	$C7,$C2,$6B,$EB,$6B,$F5,$1A,$FC,$A7,$E9,$7C,$7C,$9F,$53,$67,$E6
	.dc.b	$B5,$F8,$CD,$7E,$FB,$5F,$AA,$D7,$F0,$3F,$03,$F7,$C9,$3F,$1D,$B1
	.dc.b	$F6,$9A,$FD,$36,$BF,$8C,$6B,$F8,$27,$F1,$5F,$1F,$27,$E9,$36,$7F
	.dc.b	$0A,$D7,$EF,$35,$FC,$BB,$5F,$C6,$B5,$FC,$CF,$C7,$FE,$5C,$93,$F7
	.dc.b	$DB,$1F,$92,$6B,$F8,$B6,$BF,$A4,$6B,$F9,$A7,$F4,$5F,$1D,$27,$F1
	.dc.b	$4D,$9F,$CE,$B5,$FC,$B3,$5F,$D7,$B5,$FD,$2B,$5F,$D8,$FC,$7F,$EB
	.dc.b	$C9,$3F,$97,$6C,$7E,$A9,$AF,$E8,$DA,$FE,$F1,$AF,$EC,$9F,$DD,$79
	.dc.b	$D2,$7F,$44,$D9,$FD,$AB,$5F,$D7,$35,$FE,$0E,$D7,$F7,$AD,$7F,$84
	.dc.b	$FC,$7F,$F0,$72,$4F,$EB,$DB,$1F,$C0,$9A,$FE,$ED,$AF,$F1,$46,$BF
	.dc.b	$C2,$9F,$E2,$5F,$09,$27,$F7,$4D,$9F,$E1,$AD,$7F,$83,$35,$FE,$3E
	.dc.b	$D7,$F8,$AB,$5F,$E4,$3F,$1B,$FC,$7C,$93,$FC,$1D,$B1,$FC,$69,$AF
	.dc.b	$F1,$36,$BF,$CB,$1A,$FF,$22,$7F,$95,$7C,$64,$9F,$E2,$4D,$9F,$E4
	.dc.b	$AD,$7F,$8F,$35,$FE,$6E,$D7,$F9,$6B,$5F,$E7,$3F,$1B,$FC,$DC,$93
	.dc.b	$FC,$7D,$B1,$FC,$C9,$AF,$F2,$B6,$BF,$D1,$1A,$FF,$3A,$7F,$A1,$7C
	.dc.b	$6C,$9F,$E5,$4D,$9F,$E7,$AD,$7F,$9B,$35,$FE,$9E,$D7,$FA,$2B,$5F
	.dc.b	$EA,$3F,$1B,$FD,$3C,$93,$FC,$DD,$B1,$FD,$29,$AF,$F4,$36,$BF,$D7
	.dc.b	$1A,$FF,$52,$7F,$AD,$7C,$6C,$9F,$E8,$4D,$9F,$EA,$AD,$7F,$A7,$35
	.dc.b	$FE,$CE,$D7,$FA,$EB,$5F,$ED,$3F,$0D,$FD,$9C,$93,$FD,$3D,$B1,$FD
	.dc.b	$89,$AF,$F5,$B6,$BF,$DD,$1A,$FF,$6A,$7F,$B9,$7C,$64,$9F,$EB,$4D
	.dc.b	$9F,$ED,$AD,$7F,$B3,$35,$FE,$FE,$D7,$FB,$AB,$5F,$F0,$0F,$C6,$7F
	.dc.b	$7F,$24,$FF,$67,$6C,$7F,$7A,$6B,$FD,$CD,$AF,$F8,$31,$AF,$F8,$09
	.dc.b	$47,$F7,$EF,$C3,$6B,$4C,$CF,$F8,$26,$32,$D8,$CB,$FE,$11,$8C,$B6
	.dc.b	$32,$FF,$87,$63,$2D,$8C,$BF,$E2,$28,$CB,$63,$2A,$32,$D8,$CC,$3E
	.dc.b	$26,$92,$3D,$8D,$9F,$53,$5F,$43,$5F,$33,$5F,$23,$DB,$D4,$93,$F1
	.dc.b	$DB,$3F,$15,$AF,$C3,$6B,$F0,$5A,$FB,$DA,$FB,$7E,$37,$D6,$49,$FA
	.dc.b	$8D,$8F,$A4,$D7,$E8,$35,$F9,$CD,$7E,$63,$F1,$BE,$1A,$4F,$DF,$6C
	.dc.b	$FD,$D6,$BF,$6D,$AF,$D9,$6B,$F5,$DA,$FD,$5F,$8D,$FA,$64,$9F,$C6
	.dc.b	$36,$3F,$14,$D7,$F1,$0D,$7F,$0C,$D7,$F0,$8F,$DE,$F8,$A9,$3F,$97
	.dc.b	$6C,$FE,$55,$AF,$E4,$DA,$FE,$45,$AF,$E3,$DA,$FE,$37,$D2,$FC,$59
	.dc.b	$27,$F4,$8D,$8F,$D1,$35,$FD,$03,$5F,$CF,$35,$FC,$E3,$F9,$6F,$8A
	.dc.b	$93,$FA,$F6,$CF,$EB,$5A,$FE,$AD,$AF,$EA,$5A,$FE,$9D,$AF,$E9,$7E
	.dc.b	$2B,$F4,$64,$9F,$DE,$36,$3F,$74,$D7,$F7,$0D,$7F,$6C,$D7,$F6,$8F
	.dc.b	$EB,$BE,$26,$4F,$F0,$76,$CF,$F0,$56,$BF,$C0,$DA,$FF,$01,$6B,$FB
	.dc.b	$F6,$BF,$BD,$F8,$5F,$DD,$92,$7F,$8A,$36,$3F,$89,$35,$FE,$20,$D7
	.dc.b	$F8,$73,$5F,$E1,$8F,$F0,$6F,$89,$93,$FC,$7D,$B3,$FC,$75,$AF,$F1
	.dc.b	$B6,$BF,$C6,$5A,$FF,$17,$6B,$FC,$57,$E2,$7F,$89,$92,$7F,$96,$36
	.dc.b	$3F,$95,$35,$FE,$50,$D7,$F9,$33,$5F,$E4,$8F,$F1,$EF,$87,$93,$FC
	.dc.b	$DD,$B3,$FC,$D5,$AF,$F3,$36,$BF,$CC,$5A,$FF,$2F,$6B,$FC,$B7,$E1
	.dc.b	$FF,$95,$92,$7F,$A2,$36,$3F,$A1,$35,$FE,$80,$D7,$F9,$F3,$5F,$E7
	.dc.b	$8F,$F3,$6F,$86,$93,$FD,$3D,$B3,$FD,$35,$AF,$F4,$B6,$BF,$D2,$5A
	.dc.b	$FF,$47,$6B,$FD,$17,$E1,$FF,$A1,$92,$7F,$AE,$36,$3F,$AD,$35,$FE
	.dc.b	$B0,$D7,$FA,$B3,$5F,$EA,$8F,$F4,$EF,$85,$93,$FD,$9D,$B3,$FD,$95
	.dc.b	$AF,$F6,$36,$BF,$D8,$5A,$FF,$5F,$6B,$FD,$77,$E1,$7F,$AD,$92,$7F
	.dc.b	$BA,$36,$3F,$B9,$35,$FE,$E0,$D7,$FB,$73,$5F,$ED,$8F,$F6,$6F,$82
	.dc.b	$93,$FD,$FD,$B3,$FD,$F5,$AF,$F7,$B6,$BF,$DE,$5A,$FF,$77,$6A,$FC
	.dc.b	$39,$F0,$7F,$DE,$1B,$49,$84,$8C,$B8,$34,$E1,$C3,$5E,$13,$D4,$93
	.dc.b	$C6,$38,$C6,$1C,$4F,$17,$33,$B7,$D1,$F0,$1E,$72,$46,$65,$18,$9E
	.dc.b	$6D,$7C,$AD,$7D,$8D,$7D,$6F,$C3,$F7,$E5,$75,$6C,$F4,$8F,$8D,$F5
	.dc.b	$6F,$94,$F8,$DE,$FF,$F8,$CA,$31,$5D,$2B,$75,$A6,$C7,$CA,$BE,$AF
	.dc.b	$CA,$6B,$F4,$7E,$07,$EB,$12,$7E,$03,$F6,$70,$F0,$5F,$5A,$61,$E2
	.dc.b	$7E,$4D,$F8,$1F,$C1,$57,$07,$ED,$DF,$57,$E9,$B5,$FB,$B1,$DE,$7E
	.dc.b	$83,$63,$EC,$BF,$E5,$28,$D9,$F8,$B2,$C7,$F9,$3B,$24,$EE,$3D,$06
	.dc.b	$88,$60,$FC,$97,$B7,$F9,$72,$57,$E0,$5B,$1D,$5E,$8F,$E5,$9A,$FE
	.dc.b	$85,$B4,$78,$C7,$E6,$9A,$FE,$87,$DA,$FD,$31,$2B,$F2,$0F,$83,$1F
	.dc.b	$9D,$F8,$7B,$C0,$7B,$3F,$3F,$2E,$D8,$7D,$1A,$D4,$E7,$0E,$98,$FE
	.dc.b	$89,$AF,$EB,$3D,$BF,$DB,$12,$FF,$40,$FF,$9C,$2F,$9C,$FE,$03,$6F
	.dc.b	$DC,$C6,$33,$FC,$16,$5E,$2A,$D1,$FB,$F9,$D5,$7E,$FE,$FB,$4F,$ED
	.dc.b	$1A,$FE,$FB,$DB,$FE,$10,$4C,$7F,$64,$F8,$F4,$F8,$33,$D8,$61,$D9
	.dc.b	$FA,$95,$F0,$E7,$79,$79,$AF,$A5,$7F,$04,$6B,$FC,$47,$D9,$FC,$63
	.dc.b	$FC,$2D,$AE,$4F,$F4,$9B,$87,$E1,$D7,$6C,$AE,$BF,$57,$6F,$D1,$57
	.dc.b	$9A,$B4,$FF,$23,$BE,$4D,$3A,$93,$2F,$E2,$D1,$F8,$06,$DF,$88,$8F
	.dc.b	$F1,$BC,$FE,$15,$77,$9F,$14,$FC,$37,$F6,$D5,$CE,$7D,$D1,$C9,$B3
	.dc.b	$E0,$F6,$BD,$E4,$70,$77,$C3,$D7,$3D,$CF,$A9,$B7,$E0,$F5,$FC,$CC
	.dc.b	$EB,$16,$1C,$EF,$E7,$35,$FA,$5E,$CF,$CE,$26,$3F,$09,$FC,$16,$5E
	.dc.b	$B9,$EF,$A7,$CE,$D3,$E0,$E2,$BF,$DA,$B7,$0F,$B0,$DB,$FF,$6A,$B6
	.dc.b	$3F,$05,$EC,$FD,$E2,$3F,$49,$F0,$CF,$E4,$6F,$E7,$EB,$EF,$DB,$F0
	.dc.b	$33,$F9,$5C,$17,$E2,$26,$1E,$27,$CB,$EC,$FE,$30,$91,$F8,$07,$BE
	.dc.b	$FF,$2B,$5F,$B7,$87,$87,$7F,$99,$5C,$AF,$DA,$D9,$F6,$FB,$3F,$98
	.dc.b	$24,$FE,$31,$EB,$9F,$A3,$1F,$DD,$EF,$F2,$F1,$FB,$15,$F2,$89,$7F
	.dc.b	$29,$F9,$BE,$CF,$E9,$09,$1F,$98,$7A,$E3,$FC,$0E,$BF,$0F,$B7,$C7
	.dc.b	$8C,$69,$F9,$4A,$E7,$7F,$59,$B3,$F5,$BD,$9F,$D8,$92,$7F,$4A,$F6
	.dc.b	$8F,$C9,$67,$EA,$9F,$C3,$3F,$D5,$F7,$57,$E1,$09,$3F,$84,$7F,$0F
	.dc.b	$EC,$FE,$F8,$91,$FB,$27,$B7,$0F,$E0,$36,$F8,$D9,$FE,$B2,$78,$D7
	.dc.b	$E3,$9B,$92,$3F,$C3,$D7,$FC,$B2,$2E,$BF,$C7,$1B,$7F,$AD,$23,$FE
	.dc.b	$1D,$9E,$0F,$E3,$8D,$99,$9E,$4E,$48,$7C,$E3,$97,$99,$FE,$36,$6F
	.dc.b	$67,$F0,$98,$C4,$79,$9A,$FF,$40,$6B,$FC,$F5,$AF,$F3,$26,$D1,$A4
	.dc.b	$9E,$C4,$9F,$67,$7F,$F8,$CD,$FE,$76,$DA,$34,$C5,$7A,$76,$FE,$CA
	.dc.b	$DE,$2F,$36,$1F,$E1,$3F,$C4,$1F,$01,$FE,$50,$BF,$ED,$78,$7F,$94
	.dc.b	$B5,$FA,$6D,$CF,$FA,$EB,$5F,$EB,$4D,$7F,$A7,$F8,$3E,$52,$3F,$CC
	.dc.b	$5F,$2E,$3F,$DC,$F3,$F2,$4F,$C7,$8F,$26,$3E,$3C,$FF,$B5,$DF,$5D
	.dc.b	$0D,$25,$FF,$4F,$87,$83,$F1,$49,$1F,$D0,$5F,$0B,$FC,$57,$1F,$0B
	.dc.b	$F5,$71,$F0,$BF,$E0,$79,$F0,$11,$EC,$FB,$F5,$FB,$B0,$69,$27,$BE
	.dc.b	$BF,$49,$F8,$07,$69,$FE,$ED,$3F,$8C,$D7,$FC,$FC,$D8,$FF,$41,$B5
	.dc.b	$FF,$43,$35,$F0,$7C,$4F,$E1,$60,$D2,$47,$E2,$EF,$C4,$FC,$1B,$F1
	.dc.b	$47,$F9,$5E,$B8,$CF,$F7,$96,$BF,$B5,$3C,$75,$F0,$4F,$85,$F2,$DE
	.dc.b	$12,$47,$FA,$26,$51,$F0,$B5,$FC,$8C,$F8,$4F,$83,$B7,$DF,$7D,$B3
	.dc.b	$D1,$F1,$35,$F1,$1F,$0A,$7E,$01,$1F,$03,$DD,$FC,$61,$DF,$0F,$75
	.dc.b	$77,$8F,$B1,$3F,$6C,$8F,$C1,$7E,$F7,$C3,$7C,$D2,$47,$B5,$EC,$FE
	.dc.b	$24,$ED,$1F,$62,$7B,$60,$AE,$35,$FE,$B3,$8C,$3C,$CF,$91,$E1,$FD
	.dc.b	$B2,$38,$3E,$D5,$FC,$AC,$3E,$D9,$EC,$8F,$DE,$4F,$07,$B3,$5F,$C6
	.dc.b	$F8,$7F,$87,$23,$F5,$DF,$5F,$07,$F1,$5D,$7F,$2F,$3F,$58,$FE,$11
	.dc.b	$3C,$1F,$53,$5F,$CC,$FC,$37,$E4,$C8,$FE,$0D,$EB,$C3,$FB,$0B,$ED
	.dc.b	$3D,$4F,$D1,$CF,$07,$E3,$35,$FD,$2F,$89,$F9,$F2,$3F,$8F,$7A,$E1
	.dc.b	$E8,$CB,$E2,$4F,$EB,$A7,$6F,$D1,$6B,$FB,$0F,$86,$FD,$51,$1F,$CD
	.dc.b	$3F,$AD,$DF,$F1,$18,$FA,$2F,$E8,$EB,$F7,$EF,$F9,$65,$F1,$7E,$F3
	.dc.b	$5F,$DE,$DF,$37,$F4,$CD,$7F,$08,$FF,$0D,$23,$FB,$36,$BA,$38,$90
	.dc.b	$FC,$67,$89,$FC,$0C,$8F,$EC,$DA,$FE,$A8,$FE,$30,$FF,$1B,$23,$FC
	.dc.b	$05,$DB,$FC,$51,$A8,$9F,$88,$FF,$55,$4D,$23,$E0,$C3,$F8,$74,$FE
	.dc.b	$90,$FD,$94,$BB,$6D,$7F,$5F,$8E,$2F,$72,$1D,$8D,$75,$47,$27,$91
	.dc.b	$C3,$88,$EC,$FF,$61,$AF,$13,$FE,$F5,$8C,$BF,$EA,$97,$FC,$0E,$F8
	.dc.b	$5A,$F5,$3B,$9F,$02,$7D,$15,$23,$33,$D1,$F4,$B5,$E9,$1B,$7C,$CD
	.dc.b	$76,$D7,$67,$47,$E3,$78,$12,$3B,$B5,$F8,$0D,$7C,$2D,$7D,$8D,$7E
	.dc.b	$DB,$5F,$B7,$DF,$F4,$C8,$FB,$1A,$FC,$D6,$BE,$E6,$BF,$25,$AF,$E2
	.dc.b	$1F,$BD,$DF,$23,$E3,$B5,$FB,$0D,$7E,$4B,$5F,$A8,$D7,$F2,$6D,$7F
	.dc.b	$27,$DD,$F8,$B2,$3F,$51,$AF,$E1,$5A,$FD,$66,$BF,$81,$6B,$FA,$07
	.dc.b	$F2,$DF,$03,$23,$EF,$B5,$FC,$83,$5F,$C0,$B5,$FC,$63,$5F,$D5,$B5
	.dc.b	$FD,$5F,$C0,$FE,$8C,$8F,$E3,$1A,$FE,$75,$AF,$E3,$9A,$FE,$65,$AF
	.dc.b	$EE,$1F,$D7,$7C,$0C,$8F,$CB,$B5,$FD,$43,$5F,$CC,$B5,$FD,$23,$5F
	.dc.b	$E0,$6D,$7F,$81,$FC,$0F,$EE,$C8,$FE,$91,$AF,$ED,$5A,$FE,$99,$AF
	.dc.b	$EC,$5A,$FF,$10,$7F,$83,$7C,$0C,$8F,$D7,$B5,$FE,$00,$D7,$F6,$2D
	.dc.b	$7F,$78,$D7,$F8,$DB,$5F,$E3,$7F,$03,$FC,$4C,$8F,$EF,$1A,$FF,$0D
	.dc.b	$6B,$FB,$E6,$BF,$C2,$5A,$FF,$28,$7F,$8F,$7C,$0C,$8F,$E0,$ED,$7F
	.dc.b	$8C,$35,$FE,$12,$D7,$F8,$A3,$5F,$E6,$6D,$7F,$99,$FC,$0F,$F2,$B2
	.dc.b	$3F,$C5,$1A,$FF,$25,$6B,$FC,$59,$AF,$F2,$16,$BF,$D0,$1F,$E6,$DF
	.dc.b	$03,$23,$F8,$FB,$5F,$E6,$0D,$7F,$90,$B5,$FE,$58,$D7,$FA,$5B,$5F
	.dc.b	$E9,$7F,$03,$FD,$0C,$8F,$F2,$C6,$BF,$CF,$5A,$FF,$2E,$6B,$FC,$E5
	.dc.b	$AF,$F5,$87,$FD,$59,$D1,$FD,$8E,$3F,$A7,$35,$FE,$BC,$D4,$D2,$78
	.dc.b	$9C,$89,$1F,$D1,$23,$2D,$8C,$BF,$D5,$A1,$F0,$1E,$22,$47,$16,$BB
	.dc.b	$6B,$FD,$8A,$38,$BB,$21,$B2,$9B,$F8,$0F,$79,$1F,$33,$5F,$23,$5E
	.dc.b	$06,$88,$68,$A7,$8F,$CA,$F9,$64,$7E,$13,$5F,$80,$D7,$81,$A2,$1A
	.dc.b	$29,$EF,$F0,$7E,$B9,$1F,$9C,$D7,$E6,$35,$E0,$68,$86,$8A,$7C,$BF
	.dc.b	$80,$FC,$19,$1F,$B4,$D7,$EC,$35,$E0,$68,$86,$8A,$7D,$7F,$80,$FC
	.dc.b	$79,$1F,$C3,$35,$FC,$23,$5E,$06,$88,$68,$A7,$E0,$FE,$03,$F3,$64
	.dc.b	$7F,$24,$D7,$F2,$0D,$78,$1A,$21,$A2,$9F,$8F,$F8,$0F,$D3,$91,$FC
	.dc.b	$F3,$5F,$CE,$35,$E0,$68,$86,$8A,$7E,$6F,$E0,$3F,$66,$47,$F5,$4D
	.dc.b	$7F,$50,$D7,$81,$A2,$1A,$29,$FA,$7F,$80,$FD,$F9,$1F,$DB,$35,$FD
	.dc.b	$A3,$5E,$06,$88,$68,$A7,$EC,$FE,$03,$F8,$59,$1F,$E0,$4D,$7F,$80
	.dc.b	$35,$E0,$68,$86,$8A,$7E,$FF,$E0,$3F,$8B,$91,$FE,$1C,$D7,$F8,$63
	.dc.b	$5E,$06,$88,$68,$A7,$F0,$BF,$80,$FE,$46,$47,$F8,$D3,$5F,$E3,$0D
	.dc.b	$78,$1A,$21,$A2,$9F,$C5,$FE,$03,$F9,$79,$1F,$E4,$CD,$7F,$92,$35
	.dc.b	$E0,$68,$86,$8A,$7F,$23,$F8,$0F,$E7,$64,$7F,$99,$35,$FE,$60,$D7
	.dc.b	$81,$A2,$1A,$29,$FC,$BF,$E0,$3F,$A3,$91,$FE,$7C,$D7,$F9,$E3,$5E
	.dc.b	$06,$88,$68,$A7,$F3,$BF,$80,$FE,$A6,$47,$FA,$53,$5F,$E9,$0D,$78
	.dc.b	$1A,$21,$A2,$9F,$D1,$FD,$7F,$EB,$E4,$7F,$AB,$35,$FE,$A8,$D7,$81
	.dc.b	$A2,$1A,$29,$FD,$4F,$D7,$FE,$D6,$47,$FB,$13,$5F,$EC,$0D,$78,$1A
	.dc.b	$21,$A2,$9F,$D7,$FD,$7F,$EE,$E4,$7F,$B7,$35,$FE,$D8,$D7,$81,$A2
	.dc.b	$1A,$29,$FE,$CC,$FC,$49,$E8,$FF,$AF,$21,$18,$7F,$D7,$ED,$7F,$D9
	.dc.b	$ED,$79,$47,$18,$C4,$7F,$B6,$5A,$FF,$BA,$5A,$F6,$8E,$04,$7B,$BC
	.dc.b	$FD,$89,$1E,$26,$BE,$26,$BE,$B6,$BE,$16,$BE,$16,$BE,$EF,$63,$D0
	.dc.b	$47,$DE,$D7,$DE,$D7,$E4,$35,$F7,$35,$F7,$3F,$0F,$AA,$47,$CA,$6B
	.dc.b	$F2,$9A,$FD,$36,$BF,$25,$AF,$C9,$6B,$F5,$BE,$27,$E8,$11,$FA,$ED
	.dc.b	$7E,$BB,$5F,$C0,$35,$FA,$CD,$7E,$B3,$F6,$FE,$0A,$47,$E0,$9A,$FE
	.dc.b	$09,$AF,$E2,$DA,$FE,$05,$AF,$E0,$5A,$FE,$3B,$DA,$FC,$41,$1F,$C7
	.dc.b	$B5,$FC,$7B,$5F,$CC,$35,$FC,$73,$5F,$C7,$3F,$93,$F6,$64,$7E,$69
	.dc.b	$AF,$E6,$9A,$FE,$8D,$AF,$E6,$5A,$FE,$65,$AF,$E9,$BD,$9F,$D0,$11
	.dc.b	$FD,$3B,$5F,$D3,$B5,$FD,$83,$5F,$D3,$35,$FD,$33,$FA,$BF,$66,$47
	.dc.b	$EC,$9A,$FE,$C9,$AF,$EE,$DA,$FE,$C5,$AF,$EC,$5A,$FE,$FB,$D9,$FD
	.dc.b	$C1,$1F,$DF,$B5,$FD,$FB,$5F,$E1,$0D,$7F,$7C,$D7,$F7,$CF,$F0,$3F
	.dc.b	$62,$47,$F0,$A6,$BF,$C2,$9A,$FF,$13,$6B,$FC,$25,$AF,$F0,$96,$BF
	.dc.b	$C5,$BE,$27,$F8,$81,$1F,$E2,$ED,$7F,$8B,$B5,$FE,$40,$D7,$F8,$B3
	.dc.b	$5F,$E2,$CF,$F1,$BF,$66,$47,$F2,$26,$BF,$C8,$9A,$FF,$2B,$6B,$FC
	.dc.b	$85,$AF,$F2,$16,$BF,$CB,$BE,$27,$F9,$41,$1F,$E5,$ED,$7F,$97,$B5
	.dc.b	$FE,$70,$D7,$F9,$73,$5F,$E5,$CF,$F3,$3F,$89,$91,$FC,$E9,$AF,$F3
	.dc.b	$A6,$BF,$D0,$DA,$FF,$39,$6B,$FC,$E5,$AF,$F4,$6F,$7B,$FA,$01,$1F
	.dc.b	$E8,$ED,$7F,$A3,$B5,$FE,$A0,$D7,$FA,$33,$5F,$E8,$CF,$F4,$BF,$89
	.dc.b	$91,$FD,$49,$AF,$F5,$26,$BF,$D6,$DA,$FF,$51,$6B,$FD,$45,$AF,$F5
	.dc.b	$EF,$86,$FE,$B0,$47,$FA,$FB,$5F,$EB,$ED,$7F,$B4,$35,$FE,$BC,$D7
	.dc.b	$FA,$F3,$FD,$8F,$DD,$91,$FD,$A9,$AF,$F6,$A6,$BF,$DC,$DA,$FF,$69
	.dc.b	$6B,$FD,$A5,$AF,$F7,$6F,$86,$FE,$E0,$47,$FB,$BB,$5F,$EE,$ED,$7F
	.dc.b	$C0,$0D,$7F,$BB,$35,$FE,$EC,$FF,$80,$FF,$13,$FF,$02,$41,$A4,$74
	.dc.b	$40,$42,$3B,$70,$60,$E0,$60,$EC,$60,$DA,$F7,$FC,$84,$8E,$07,$95
	.dc.b	$C3,$CE,$38,$3A,$6B,$98,$DB,$D9,$AF,$A1,$F8,$5F,$0F,$23,$93,$5E
	.dc.b	$CD,$7C,$ED,$7E,$0B,$5F,$8B,$28,$C4,$7C,$DF,$7F,$E8,$11,$F4,$35
	.dc.b	$F6,$35,$F8,$AD,$7E,$7B,$5E,$EF,$DB,$F8,$99,$1F,$0D,$AF,$C7,$6B
	.dc.b	$F4,$1A,$FD,$A6,$BE,$A6,$AF,$CF,$1E,$87,$F9,$2D,$B1,$F5,$9A,$FD
	.dc.b	$F6,$BF,$84,$5E,$97,$E2,$7E,$39,$EA,$7E,$2E,$5E,$A6,$63,$0D,$BF
	.dc.b	$C8,$CF,$27,$F2,$75,$C9,$CE,$1E,$52,$3E,$57,$F4,$1F,$11,$23,$F1
	.dc.b	$0D,$7F,$18,$D7,$F2,$AD,$7F,$3E,$D7,$EF,$B5,$7C,$13,$D7,$FF,$38
	.dc.b	$36,$3F,$36,$D7,$F4,$8D,$7F,$52,$BD,$9F,$81,$1D,$3D,$9B,$26,$CE
	.dc.b	$2B,$B9,$2B,$BB,$FB,$FF,$81,$91,$FA,$76,$BF,$AB,$6B,$FB,$46,$BF
	.dc.b	$BE,$6B,$F9,$E6,$BF,$03,$E0,$3D,$E4,$7D,$EF,$C2,$8F,$4F,$C0,$C9
	.dc.b	$FB,$3B,$FF,$0B,$C9,$5E,$EB,$E5,$5F,$19,$AF,$C6,$9F,$9A,$47,$E6
	.dc.b	$BF,$43,$3F,$8E,$48,$EE,$7F,$1A,$87,$C1,$FC,$C7,$94,$7D,$46,$BF
	.dc.b	$C5,$BD,$EF,$E5,$04,$7F,$88,$35,$FE,$28,$D7,$81,$A2,$1A,$29,$F8
	.dc.b	$0E,$7F,$C3,$91,$FC,$3B,$FE,$B9,$FC,$F1,$D3,$F8,$46,$CF,$B0,$38
	.dc.b	$5C,$5F,$A5,$7F,$2A,$7E,$2A,$BB,$3F,$92,$6B,$FC,$EA,$5D,$7E,$12
	.dc.b	$47,$F3,$48,$CB,$7B,$7C,$9D,$B2,$3C,$92,$9F,$45,$72,$9F,$E8,$93
	.dc.b	$C9,$B1,$FE,$D7,$FC,$5B,$F8,$4C,$7F,$4F,$6B,$FD,$49,$A9,$A4,$E6
	.dc.b	$E4,$48,$FE,$97,$19,$6C,$65,$F1,$F8,$9F,$3C,$8F,$BD,$FA,$5E,$37
	.dc.b	$D1,$24,$7B,$BC,$6F,$A9,$3C,$23,$99,$C4,$F9,$1E,$47,$DB,$23,$83
	.dc.b	$E6,$79,$5F,$77,$CA,$91,$F2,$A3,$F0,$13,$C2,$3E,$4B,$5F,$EE,$39
	.dc.b	$3D,$AF,$94,$48,$F8,$2D,$7F,$BA,$46,$DC,$90,$CC,$6D,$EA,$D7,$FC
	.dc.b	$05,$F9,$E3,$D3,$FE,$EF,$08,$C3,$FD,$E9,$AF,$F8,$0D,$AF,$F8,$21
	.dc.b	$7A,$BF,$C8,$17,$AB,$4C,$46,$6B,$BC,$F2,$AF,$1C,$60,$CF,$AF,$23
	.dc.b	$C4,$D7,$C5,$EC,$F7,$23,$FA,$BA,$7B,$3E,$3A,$E1,$3E,$76,$BF,$95
	.dc.b	$C3,$F3,$04,$7F,$1E,$D7,$E7,$B5,$FA,$ED,$79,$1C,$08,$7C,$7F,$7B
	.dc.b	$DE,$47,$E6,$3F,$0B,$E0,$3E,$47,$C0,$49,$F5,$47,$CB,$9E,$11,$CC
	.dc.b	$E2,$7D,$5E,$5F,$96,$47,$07,$D9,$F2,$BE,$A1,$1F,$8D,$E5,$7D,$79
	.dc.b	$E1,$1F,$75,$AF,$F9,$11,$7A,$DF,$86,$6C,$7F,$8F,$9A,$FF,$91,$DA
	.dc.b	$F1,$8E,$04,$78,$35,$E0,$D7,$81,$FE,$4A,$FC,$F1,$F0,$7F,$96,$6C
	.dc.b	$66,$30,$67,$E1,$7F,$B8,$CB,$E1,$69,$88,$CC,$F0,$AE,$44,$3F,$CC
	.dc.b	$14,$F2,$3C,$04,$8F,$23,$E3,$F2,$7B,$92,$3F,$9E,$4B,$CB,$F1,$D7
	.dc.b	$19,$EA,$D7,$E7,$76,$FC,$D2,$3F,$35,$AF,$CD,$6B,$C0,$D1,$0D,$14
	.dc.b	$FF,$3E,$D3,$E3,$BE,$51,$23,$E7,$3F,$37,$D2,$F9,$3E,$94,$9F,$5C
	.dc.b	$7C,$C9,$E1,$1F,$19,$AF,$6F,$1F,$EB,$11,$FB,$4F,$F0,$EF,$85,$FE
	.dc.b	$BD,$E4,$FF,$5D,$6F,$9F,$FE,$95,$6D,$7F,$A6,$32,$FF,$4A,$D7,$F6
	.dc.b	$37,$FD,$0F,$1D,$CE,$A4,$3F,$25,$E5,$7E,$5C,$8F,$E5,$DA,$E8,$C1
	.dc.b	$0C,$10,$FD,$13,$C1,$FF,$53,$D3,$D8,$F6,$12,$3F,$D4,$31,$96,$FB
	.dc.b	$5F,$E4,$48,$34,$C7,$B5,$9A,$E5,$5F,$EB,$1A,$E4,$D8,$F9,$5B,$C8
	.dc.b	$EE,$D7,$87,$95,$DF,$CA,$91,$FE,$0F,$B4,$F8,$EB,$84,$FB,$1A,$FF
	.dc.b	$B1,$FF,$19,$70,$FF,$46,$6C,$7F,$AF,$9A,$FF,$B1,$DA,$FF,$B3,$1A
	.dc.b	$F3,$8E,$44,$79,$35,$E4,$D7,$E7,$FB,$DF,$89,$23,$F3,$9F,$C6,$7C
	.dc.b	$07,$CD,$24,$7B,$FE,$03,$E8,$4F,$08,$F9,$8D,$7F,$DC,$3F,$9E,$7F
	.dc.b	$53,$7F,$DC,$ED,$4D,$27,$89,$FE,$7E,$7F,$AF,$C7,$F8,$ED,$AF,$F8
	.dc.b	$EB,$93,$FD,$E1,$1F,$D9,$A3,$5E,$4F,$F7,$08,$E8,$D3,$E7,$7C,$54
	.dc.b	$9F,$9C,$D7,$E7,$35,$F9,$CD,$7E,$C3,$5F,$F8,$03,$A2,$31,$1E,$CF
	.dc.b	$87,$F1,$11,$FF,$7C,$DA,$3C,$DF,$11,$EF,$24,$FB,$FC,$EF,$85,$3C
	.dc.b	$60,$7C,$C7,$0F,$CE,$F1,$FE,$01,$1E,$EF,$C3,$F2,$3E,$67,$93,$27
	.dc.b	$E6,$47,$CF,$9E,$11,$D8,$F9,$AE,$1F,$C8,$78,$7F,$97,$23,$F9,$76
	.dc.b	$BF,$97,$6B,$C0,$D1,$0D,$14,$F0,$0F,$CF,$3F,$D8,$4F,$8D,$AF,$C2
	.dc.b	$6B,$BF,$73,$FC,$5F,$1F,$EC,$78,$FF,$41,$35,$FC,$41,$C7,$DD,$FF
	.dc.b	$65,$C7,$FB,$26,$37,$3C,$88,$F9,$0F,$03,$DA,$A4,$64,$F4,$38,$78
	.dc.b	$4B,$C7,$EF,$E4,$49,$F2,$C7,$8A,$B8,$47,$C2,$AF,$EE,$6F,$28,$FF
	.dc.b	$25,$FE,$38,$F3,$29,$27,$C5,$19,$6C,$65,$F5,$46,$5B,$19,$7E,$8C
	.dc.b	$65,$B1,$97,$E9,$7B,$9F,$38,$8F,$C1,$BF,$F5,$5B,$F0,$3D,$EF,$9A
	.dc.b	$47,$E9,$FB,$DF,$42,$B8,$C7,$C6,$6B,$F0,$7C,$9F,$B6,$47,$B3,$F7
	.dc.b	$BB,$BE,$CF,$95,$27,$EF,$C7,$DC,$9E,$11,$F4,$9A,$FB,$B9,$BE,$59
	.dc.b	$1F,$92,$D7,$E3,$B5,$E0,$68,$86,$8A,$7E,$6B,$F3,$CF,$F5,$D3,$7D
	.dc.b	$BF,$F4,$1C,$62,$3F,$5C,$D7,$F7,$AF,$FB,$15,$FF,$6D,$35,$FC,$C8
	.dc.b	$71,$E5,$FF,$B6,$D7,$95,$FB,$98,$C5,$78,$09,$5F,$97,$AF,$03,$CF
	.dc.b	$C8,$F2,$91,$D3,$E8,$F2,$BA,$78,$3E,$8F,$2A,$D7,$A1,$AF,$96,$6D
	.dc.b	$FF,$6E,$8F,$89,$7D,$5F,$D6,$46,$5F,$E2,$1C,$8F,$8A,$FC,$F9,$23
	.dc.b	$F3,$AD,$7F,$36,$D7,$F3,$2D,$79,$B5,$FE,$53,$1D,$5F,$85,$ED,$7D
	.dc.b	$42,$38,$3E,$37,$B7,$EA,$23,$F5,$3D,$CF,$6C,$F0,$8F,$5B,$5F,$C1
	.dc.b	$79,$1F,$80,$23,$F4,$5F,$C1,$F9,$5F,$1B,$CA,$93,$F8,$08,$F9,$33
	.dc.b	$C2,$3E,$3B,$5F,$E4,$FF,$8F,$23,$FD,$9E,$D8,$FE,$28,$D7,$F8,$D3
	.dc.b	$5E,$31,$C0,$8F,$06,$BC,$1A,$F0,$3F,$90,$79,$AF,$63,$FD,$BE,$D8
	.dc.b	$CC,$60,$CF,$B7,$F8,$69,$34,$C7,$B5,$99,$E1,$5C,$88,$7F,$33,$AF
	.dc.b	$ED,$4F,$8B,$C9,$F0,$91,$F2,$F4,$F7,$F2,$A4,$7F,$6A,$E9,$74,$7F
	.dc.b	$5F,$C1,$7E,$5C,$39,$5F,$76,$BF,$43,$D2,$F9,$E4,$7E,$7B,$5F,$9E
	.dc.b	$D7,$81,$A2,$1A,$29,$FD,$8A,$9E,$CF,$CC,$24,$7D,$07,$E7,$FB,$5F
	.dc.b	$2F,$C5,$93,$ED,$CB,$D9,$3F,$3E,$B8,$CF,$C9,$6B,$FD,$C3,$F3,$CF
	.dc.b	$ED,$64,$F2,$BF,$AE,$42,$33,$E5,$FE,$F9,$FE,$D9,$B7,$CA,$9F,$F0
	.dc.b	$AC,$63,$1F,$F0,$53,$FD,$B6,$3F,$8C,$BD,$5F,$ED,$0D,$7F,$2B,$F0
	.dc.b	$1F,$9A,$23,$F9,$A6,$BA,$30,$43,$04,$3F,$4A,$F0,$7D,$DE,$CF,$B2
	.dc.b	$7F,$DA,$51,$E2,$8C,$19,$F6,$FD,$DE,$DC,$9F,$98,$7B,$A3,$E1,$4F
	.dc.b	$08,$E4,$43,$FC,$11,$5F,$D9,$1F,$9D,$E4,$F8,$48,$FC,$2F,$2F,$E6
	.dc.b	$11,$EE,$7D,$1F,$2F,$E8,$4F,$08,$EA,$D7,$FC,$53,$E5,$BE,$17,$EC
	.dc.b	$B6,$3F,$C2,$CD,$7F,$C3,$8D,$78,$C7,$02,$3C,$1A,$F0,$6B,$C0,$FB
	.dc.b	$7E,$E7,$CD,$23,$83,$FC,$4D,$F9,$E7,$FB,$49,$FC,$61,$F8,$B3,$FA
	.dc.b	$4B,$9C,$BF,$DA,$71,$F8,$A9,$E6,$FC,$C7,$F2,$B3,$FD,$68,$9E,$5E
	.dc.b	$0C,$EF,$D0,$95,$F3,$27,$FB,$3B,$67,$83,$E6,$6B,$F1,$B2,$F9,$64
	.dc.b	$7E,$5B,$5D,$18,$21,$82,$1F,$59,$E0,$FF,$98,$25,$ED,$FB,$09,$1F
	.dc.b	$21,$F7,$F4,$7C,$6E,$99,$3E,$B8,$F9,$33,$C2,$3F,$CB,$15,$D1,$E4
	.dc.b	$7D,$2F,$13,$E4,$91,$C1,$F5,$BC,$6F,$A6,$47,$E4,$78,$DF,$5A,$78
	.dc.b	$47,$66,$BF,$92,$F4,$FF,$22,$47,$F2,$2D,$7F,$22,$D7,$81,$A2,$1A
	.dc.b	$29,$FE,$86,$FC,$F3,$FD,$94,$4F,$6F,$F8,$BB,$8F,$73,$FC,$26,$FF
	.dc.b	$B2,$3A,$FF,$8B,$52,$31,$D7,$FE,$9F,$7F,$D9,$51,$FE,$2E,$7A,$3F
	.dc.b	$E7,$D6,$BF,$E8,$37,$FA,$F3,$7C,$4F,$09,$1E,$1E,$37,$88,$8F,$EF
	.dc.b	$25,$E3,$79,$AB,$84,$F3,$3C,$0F,$F3,$2F,$C8,$9E,$1D,$24,$FF,$A4
	.dc.b	$63,$2D,$8C,$BF,$E9,$98,$CB,$63,$2F,$FA,$C2,$32,$D8,$CB,$F1,$7D
	.dc.b	$8F,$9A,$47,$E2,$3F,$43,$D9,$F9,$FE,$D4,$9F,$74,$7D,$29,$E1,$1F
	.dc.b	$25,$AF,$DB,$F1,$BE,$C9,$1F,$A0,$FD,$CB,$FB,$64,$8F,$95,$E3,$FD
	.dc.b	$E9,$E1,$1F,$55,$AF,$B7,$DE,$F9,$44,$7E,$43,$5F,$8C,$D7,$81,$A2
	.dc.b	$1A,$29,$FE,$D8,$F0,$BF,$D6,$ED,$F6,$BF,$B4,$C6,$23,$FD,$D8,$D7
	.dc.b	$FD,$F6,$FF,$B1,$5F,$E7,$EB,$97,$FC,$D6,$D8,$CF,$B7,$8A,$EA,$48
	.dc.b	$FF,$69,$4F,$54,$FF,$5E,$7F,$DB,$4D,$7D,$1E,$37,$98,$F3,$F8,$FE
	.dc.b	$02,$47,$93,$C8,$F1,$CE,$E3,$BB,$5F,$C7,$71,$7E,$5C,$8F,$E5,$5A
	.dc.b	$E8,$C1,$0C,$10,$FC,$A3,$C1,$F9,$DE,$CF,$CC,$23,$F0,$1F,$9F,$ED
	.dc.b	$7C,$8F,$26,$4F,$CC,$8F,$97,$3C,$23,$E1,$B5,$FB,$5E,$3F,$D8,$23
	.dc.b	$F3,$1F,$F1,$1F,$E7,$9F,$D6,$49,$F0,$DF,$5A,$7E,$BC,$F0,$7E,$7B
	.dc.b	$51,$07,$E0,$8F,$5B,$FC,$5E,$D8,$D1,$B7,$0E,$0E,$31,$C0,$8E,$4C
	.dc.b	$96,$38,$BD,$1A,$F5,$8E,$2F,$58,$CB,$E8,$BF,$F5,$B3,$7D,$AF,$86
	.dc.b	$47,$C1,$0F,$6F,$FD,$86,$BD,$CA,$62,$33,$5C,$6B,$D9,$5D,$93,$4F
	.dc.b	$EB,$4E,$BC,$9F,$A6,$47,$C3,$E5,$78,$08,$FF,$8A,$A5,$E5,$F9,$2B
	.dc.b	$84,$F3,$3D,$2F,$D2,$A9,$1F,$25,$AF,$C5,$6B,$F2,$DA,$FD,$76,$BF
	.dc.b	$80,$C9,$18,$8F,$6F,$AF,$F8,$C2,$3F,$29,$FC,$7F,$B3,$F2,$B5,$93
	.dc.b	$EB,$8F,$9B,$3C,$23,$E4,$B5,$FC,$4F,$E7,$8F,$2B,$FC,$F6,$D8,$FC
	.dc.b	$33,$5F,$C6,$35,$E3,$E5,$7F,$BF,$23,$FB,$2F,$95,$FD,$A9,$BE,$57
	.dc.b	$E6,$DF,$ED,$75,$F9,$E7,$93,$9B,$F0,$91,$FC,$A7,$C7,$7E,$80,$8E
	.dc.b	$08,$7E,$4D,$CB,$4C,$9F,$9B,$72,$D3,$DB,$EC,$7C,$02,$38,$3E,$0F
	.dc.b	$B3,$ED,$F6,$A4,$7F,$9C,$30,$FF,$61,$B6,$B8,$BF,$19,$AF,$CE,$F2
	.dc.b	$3E,$69,$1F,$9A,$FD,$0F,$27,$E5,$12,$3B,$79,$5F,$36,$78,$47,$D2
	.dc.b	$6B,$FC,$0B,$F2,$5F,$F3,$1B,$FC,$29,$A9,$A4,$EB,$FF,$5C,$B6,$3F
	.dc.b	$82,$35,$FE,$14,$D7,$B4,$7F,$B0,$3F,$3C,$7B,$94,$93,$FC,$36,$32
	.dc.b	$DF,$73,$F8,$E7,$17,$F8,$BA,$7F,$D8,$0D,$F8,$DC,$F1,$FF,$21,$9F
	.dc.b	$F6,$35,$FF,$24,$79,$BF,$C7,$5A,$FF,$22,$6B,$C7,$0F,$37,$93,$23
	.dc.b	$C8,$D7,$97,$95,$E0,$7B,$35,$57,$F7,$15,$8F,$A7,$CB,$D4,$FF,$6B
	.dc.b	$25,$F5,$AF,$A9,$E5,$FC,$E2,$3F,$39,$AF,$CE,$6B,$C0,$D1,$0D,$14
	.dc.b	$F8,$1E,$C7,$D3,$23,$F4,$5F,$AD,$EE,$F9,$A6,$4F,$FB,$22,$D3,$EA
	.dc.b	$AE,$33,$F3,$5A,$FC,$DF,$1F,$F0,$04,$7E,$F3,$F8,$3F,$23,$E1,$92
	.dc.b	$3B,$79,$3F,$1A,$78,$47,$D9,$6B,$F9,$9D,$7F,$26,$47,$F2,$6D,$7F
	.dc.b	$26,$D7,$81,$A2,$1A,$29,$FD,$9B,$91,$ED,$7F,$61,$84,$61,$FE,$C8
	.dc.b	$D7,$FB,$53,$5E,$5E,$DF,$E1,$63,$FD,$8B,$1F,$84,$E2,$FE,$F1,$18
	.dc.b	$E1,$FF,$45,$AA,$E1,$7E,$18,$E9,$7C,$23,$06,$7C,$89,$1E,$46,$BE
	.dc.b	$4F,$27,$C4,$47,$FC,$13,$87,$83,$CA,$F3,$57,$09,$F8,$4D,$7F,$C1
	.dc.b	$DE,$D7,$FA,$81,$FF,$05,$B5,$CE,$7F,$1E,$31,$1F,$E1,$66,$A6,$92
	.dc.b	$38,$12,$3F,$C2,$11,$97,$E7,$FB,$3F,$0C,$8E,$0F,$A7,$1F,$3F,$DA
	.dc.b	$91,$EE,$AF,$9F,$3F,$4E,$B8,$CF,$D8,$6B,$F8,$1F,$1B,$E8,$11,$FB
	.dc.b	$EF,$D3,$F1,$FE,$E1,$23,$B1,$FE,$34,$DB,$FC,$5B,$A7,$DC,$AE,$53
	.dc.b	$F8,$66,$BF,$2A,$7E,$41,$1F,$82,$D7,$FC,$6D,$87,$E4,$9C,$90,$CC
	.dc.b	$6D,$E8,$64,$FF,$20,$FE,$79,$FE,$CE,$6F,$A1,$F9,$D8,$C4,$7F,$92
	.dc.b	$DA,$FF,$93,$97,$B7,$FC,$7D,$FF,$67,$C7,$C0,$0E,$1F,$CF,$1C,$63
	.dc.b	$84,$F7,$24,$7F,$93,$67,$BB,$CB,$C4,$F2,$11,$D3,$E7,$F1,$FC,$44
	.dc.b	$8F,$14,$BC,$7F,$35,$71,$9F,$F2,$F4,$A3,$11,$F9,$AF,$6B,$F3,$04
	.dc.b	$7F,$22,$D7,$E8,$B5,$E0,$68,$86,$8A,$7D,$0F,$5F,$E7,$11,$F8,$0F
	.dc.b	$D1,$F6,$7E,$61,$23,$D7,$EC,$FC,$F9,$E1,$1E,$A6,$BF,$E7,$CF,$CF
	.dc.b	$3F,$AD,$BF,$E8,$46,$A6,$93,$C7,$FE,$B0,$FF,$6C,$E5,$FF,$45,$36
	.dc.b	$79,$4E,$BC,$6F,$E4,$53,$C9,$A7,$F2,$5F,$CA,$5C,$D4,$97,$FE,$89
	.dc.b	$8C,$B6,$32,$FF,$A4,$63,$2D,$8C,$BF,$E9,$88,$CB,$63,$2F,$A3,$D8
	.dc.b	$F2,$91,$FF,$4C,$5F,$FB,$55,$F6,$FB,$5F,$00,$8F,$E6,$61,$ED,$7C
	.dc.b	$4A,$E5,$3F,$EA,$59,$E4,$FC,$D8,$F8,$7E,$2C,$8E,$ED,$78,$78,$DF
	.dc.b	$3B,$C8,$91,$DE,$3E,$8C,$EE,$3E,$D3,$5F,$F5,$E6,$6F,$95,$F9,$C4
	.dc.b	$8F,$98,$D7,$E5,$35,$F9,$0D,$79,$B5,$FF,$64,$C7,$57,$E3,$F9,$BF
	.dc.b	$78,$8E,$0F,$F6,$17,$E7,$8F,$73,$FC,$40,$D8,$FE,$04,$D7,$F8,$09
	.dc.b	$F9,$13,$F1,$F3,$D1,$F9,$8D,$7F,$DA,$B1,$FD,$79,$FF,$6E,$B5,$34
	.dc.b	$9E,$3F,$8B,$E1,$E4,$F8,$A3,$CB,$E4,$7F,$79,$AE,$0D,$3F,$31,$E3
	.dc.b	$4A,$FC,$B3,$5F,$CA,$35,$FC,$93,$5F,$AA,$D7,$E5,$B5,$F3,$FB,$7E
	.dc.b	$12,$3F,$29,$F2,$7B,$5F,$28,$91,$F2,$BD,$CF,$9B,$3C,$23,$FD,$ED
	.dc.b	$87,$FB,$61,$F8,$F1,$ED,$F2,$24,$7D,$66,$BC,$3C,$6F,$AF,$E4,$C8
	.dc.b	$FA,$F1,$F6,$A7,$71,$F4,$9A,$89,$F9,$5F,$F8,$8D,$34,$8F,$F8,$9D
	.dc.b	$AF,$12,$1C,$0D,$10,$C9,$D8,$C1,$87,$C1,$C5,$FE,$C2,$8F,$23,$5F
	.dc.b	$D3,$7B,$5F,$E3,$78,$C1,$FE,$CC,$8D,$FB,$7F,$E7,$78,$C4,$7E,$B7
	.dc.b	$E4,$31,$5C,$49,$1E,$43,$A9,$67,$FA,$B9,$3C,$9F,$01,$E1,$F2,$BB
	.dc.b	$78,$D2,$BF,$B3,$57,$F6,$A6,$D7,$07,$A3,$5F,$81,$D5,$F2,$C8,$F8
	.dc.b	$F3,$FF,$39,$B6,$32,$F0,$30,$43,$04,$3E,$9C,$72,$7E,$8D,$7F,$83
	.dc.b	$5B,$E5,$FC,$B2,$3F,$07,$DA,$F9,$44,$7D,$BE,$D7,$CD,$9D,$C7,$23
	.dc.b	$E4,$3F,$57,$C7,$91,$F2,$DA,$F8,$BC,$9F,$AD,$E4,$C8,$F9,$31,$F6
	.dc.b	$67,$71,$F8,$CC,$51,$88,$FC,$86,$4F,$9C,$FC,$A1,$23,$E5,$35,$FC
	.dc.b	$73,$5F,$8A,$D7,$9B,$5F,$D1,$BF,$E8,$5F,$CF,$1E,$CD,$24,$FE,$26
	.dc.b	$32,$DF,$57,$FD,$91,$1F,$EC,$7F,$4D,$E7,$7E,$9E,$30,$DF,$6B,$78
	.dc.b	$74,$C3,$F4,$F5,$D1,$B1,$E3,$F2,$64,$78,$5A,$F0,$F3,$3C,$3E,$64
	.dc.b	$8F,$F1,$D5,$A7,$CB,$5C,$27,$F7,$33,$FD,$89,$37,$5F,$F0,$79,$3C
	.dc.b	$EF,$E0,$44,$9F,$CC,$35,$FD,$E3,$5F,$C9,$B5,$FB,$2F,$CD,$F5,$E4
	.dc.b	$7C,$B6,$BF,$0B,$D9,$F9,$64,$7D,$7E,$CF,$CE,$9D,$C7,$A9,$AF,$CF
	.dc.b	$F2,$FE,$D1,$1F,$B0,$FF,$B0,$7F,$3C,$FE,$B9,$C3,$F5,$A7,$F9,$26
	.dc.b	$7F,$C7,$23,$11,$F1,$0F,$E4,$0E,$27,$8F,$0F,$79,$EA,$43,$ED,$DF
	.dc.b	$FB,$3F,$F0,$57,$55,$24,$FF,$13,$8C,$B6,$32,$FF,$2A,$8C,$B6,$32
	.dc.b	$FF,$32,$8C,$B6,$32,$FF,$37,$87,$C2,$78,$C9,$1E,$16,$BD,$DA,$F5
	.dc.b	$6B,$CD,$AF,$F4,$B8,$EA,$FA,$FD,$9F,$80,$47,$04,$3D,$8E,$5A,$64
	.dc.b	$F6,$39,$69,$F1,$FD,$9F,$98,$47,$04,$3E,$43,$96,$99,$3E,$43,$96
	.dc.b	$9F,$4F,$E7,$BE,$C1,$1C,$10,$FA,$8E,$5A,$64,$FA,$8E,$5A,$7D,$FF
	.dc.b	$07,$F0,$84,$70,$43,$F0,$0E,$5A,$64,$FC,$03,$96,$9F,$8B,$F1,$7F
	.dc.b	$20,$47,$04,$3F,$18,$E5,$A6,$4F,$C6,$39,$69,$FD,$A7,$E7,$FF,$CF
	.dc.b	$2F,$F8,$0D,$AF,$F8,$35,$A9,$A4,$F6,$3F,$B9,$46,$5B,$19,$8E,$D1
	.dc.b	$82,$3E,$1F,$95,$D9,$1C,$10,$FF,$08,$CA,$35,$18,$6B,$FE,$1A,$8D
	.dc.b	$3C,$4C,$47,$93,$CA,$91,$F0,$5A,$F6,$6B,$CD,$AF,$C5,$6B,$F1,$5A
	.dc.b	$FC,$5F,$3F,$E0,$91,$F9,$CD,$7D,$6D,$7D,$0D,$7E,$93,$5F,$A4,$FC
	.dc.b	$CF,$7E,$47,$D9,$6B,$F1,$9A,$FC,$36,$BF,$75,$AF,$DD,$6B,$F7,$7E
	.dc.b	$13,$EC,$91,$FC,$33,$5F,$A6,$D7,$E8,$35,$FC,$53,$5F,$C5,$3F,$84
	.dc.b	$F5,$24,$7E,$45,$AF,$DE,$6B,$F6,$DA,$FE,$55,$AF,$E5,$5A,$BF,$01
	.dc.b	$FF,$31,$11,$FC,$AB,$5F,$F2,$43,$5F,$F2,$EB,$5F,$F3,$1B,$5F,$F3
	.dc.b	$33,$D3,$AF,$99,$23,$9B,$5E,$6D,$79,$B5,$E6,$D7,$FC,$F5,$08,$F3
	.dc.b	$FD,$34,$9F,$3B,$5F,$3B,$5F,$3B,$5F,$3B,$5E,$AD,$7E,$37,$B7,$F1
	.dc.b	$08,$FC,$46,$BF,$11,$AF,$C4,$6B,$F1,$1A,$FA,$9F,$9F,$E1,$48,$F9
	.dc.b	$ED,$7E,$7B,$5F,$9E,$D7,$E7,$B5,$F8,$AD,$7E,$F7,$8B,$F7,$08,$FD
	.dc.b	$C6,$BF,$71,$AF,$DC,$6B,$F7,$1A,$FD,$27,$F0,$FF,$37,$23,$F0,$ED
	.dc.b	$7F,$0E,$D7,$F0,$ED,$7F,$0E,$D7,$EE,$B5,$FC,$B7,$83,$F9,$42,$3F
	.dc.b	$94,$6B,$F9,$46,$BF,$94,$6B,$F9,$46,$BF,$8A,$7F,$99,$FE,$02,$F2
	.dc.b	$E9,$27,$FD,$51,$19,$6C,$65,$FF,$5F,$C6,$5B,$19,$7F,$D8,$F1,$96
	.dc.b	$C6,$5F,$F6,$64,$3C,$CF,$19,$23,$C2,$D7,$BB,$5E,$AD,$79,$B5,$FF
	.dc.b	$70,$47,$57,$D7,$E5,$FC,$02,$38,$21,$EC,$72,$D3,$27,$B1,$CB,$4F
	.dc.b	$8F,$CB,$F3,$08,$E0,$87,$C8,$72,$D3,$27,$C8,$72,$D3,$E9,$F9,$DF
	.dc.b	$60,$8E,$08,$7D,$47,$2D,$32,$7D,$47,$2D,$3E,$FF,$7F,$E1,$08,$E0
	.dc.b	$87,$E0,$1C,$B4,$C9,$F8,$07,$2D,$3F,$17,$DF,$F9,$02,$38,$21,$F8
	.dc.b	$C7,$2D,$32,$7E,$31,$CB,$4F,$CB,$FC,$0F,$E7,$08,$E0,$87,$E6,$1C
	.dc.b	$B4,$C9,$F9,$87,$2D,$3F,$47,$E8,$FE,$A0,$8E,$08,$7E,$91,$CB,$4C
	.dc.b	$9F,$A4,$72,$D3,$F5,$FE,$A7,$ED,$08,$E0,$87,$EC,$1C,$B4,$C9,$FB
	.dc.b	$07,$2D,$3F,$77,$F4,$5F,$C0,$08,$E0,$87,$EF,$1C,$B4,$C9,$FB,$C7
	.dc.b	$2D,$04,$27,$DF,$3E,$17,$FC,$82,$D8,$D1,$A8,$C9,$1C,$9D,$23,$83
	.dc.b	$86,$BC,$9A,$F3,$38,$87,$2F,$38,$EC,$F3,$F6,$7C,$A4,$78,$CA,$31
	.dc.b	$1B,$32,$E1,$F1,$35,$F8,$2D,$7D,$5E,$E7,$C2,$23,$D9,$AF,$B5,$AF
	.dc.b	$B5,$AF,$CC,$6B,$D0,$F3,$FD,$17,$CE,$23,$82,$1F,$25,$CB,$4C,$9F
	.dc.b	$2D,$CB,$4F,$87,$EF,$7D,$A2,$38,$21,$F5,$5C,$B4,$C9,$F5,$DC,$B4
	.dc.b	$F9,$FF,$01,$F8,$62,$38,$21,$F8,$17,$2D,$32,$7E,$0D,$CB,$4F,$B7
	.dc.b	$F0,$3F,$92,$23,$82,$1F,$8D,$72,$D3,$27,$E3,$DC,$B4,$FC,$7F,$A2
	.dc.b	$F9,$3F,$F4,$FB,$63,$31,$82,$38,$21,$FA,$08,$E0,$D8,$DB,$4D,$9F
	.dc.b	$AA,$81,$E0,$23,$E4,$6B,$D5,$AF,$16,$BF,$B9,$79,$BC,$48,$79,$7E
	.dc.b	$13,$DA,$47,$E0,$B5,$D1,$82,$18,$21,$E9,$78,$3E,$BF,$5F,$E4,$91
	.dc.b	$F9,$AD,$7E,$33,$5E,$06,$88,$68,$A7,$C7,$F6,$BE,$A9,$1F,$B2,$D7
	.dc.b	$E9,$35,$E0,$68,$86,$8A,$7D,$3F,$67,$F0,$24,$7F,$0A,$D7,$EF,$35
	.dc.b	$E0,$68,$86,$8A,$7D,$FF,$86,$FC,$69,$1F,$C8,$B5,$FC,$53,$5E,$06
	.dc.b	$88,$68,$A7,$E2,$FE,$2B,$F3,$24,$7F,$3A,$D7,$F2,$CD,$78,$1A,$21
	.dc.b	$A2,$9F,$D0,$3F,$39,$7C,$17,$F3,$68,$46,$1F,$E7,$2D,$7F,$A1,$B5
	.dc.b	$E5,$CD,$FD,$3F,$D9,$7B,$1F,$D4,$69,$E9,$66,$30,$DB,$FE,$AF,$2B
	.dc.b	$C4,$78,$B0,$F2,$4F,$78,$E9,$18,$3D,$1F,$79,$FE,$5D,$8F,$EC,$0D
	.dc.b	$7F,$B1,$B5,$34,$9E,$27,$22,$47,$F5,$88,$CB,$63,$2F,$F6,$48,$FF
	.dc.b	$2E,$37,$C4,$F1,$91,$E4,$D7,$46,$08,$60,$87,$F7,$48,$EA,$DF,$0F
	.dc.b	$DE,$47,$CC,$D7,$C8,$D7,$81,$A2,$1A,$29,$DF,$C3,$F9,$64,$7E,$13
	.dc.b	$5F,$80,$D7,$81,$A2,$1A,$29,$EB,$F0,$BE,$B9,$1F,$9C,$D7,$E6,$35
	.dc.b	$E0,$68,$86,$8A,$7C,$7F,$5F,$F0,$64,$7E,$D3,$D7,$7F,$CA,$C7,$E1
	.dc.b	$27,$F1,$F8,$1B,$EB,$62,$F9,$57,$DF,$6B,$F5,$63,$F2,$24,$7F,$10
	.dc.b	$D7,$F0,$CD,$78,$1A,$21,$A2,$9F,$81,$F5,$BF,$3A,$47,$F2,$8F,$E8
	.dc.b	$38,$3C,$F1,$FE,$40,$C7,$DA,$FE,$7A,$BD,$B3,$EE,$AE,$53,$F9,$96
	.dc.b	$BF,$8F,$8E,$8F,$E8,$5A,$FE,$7D,$AE,$CD,$10,$D1,$4D,$1F,$E4,$AF
	.dc.b	$B9,$7A,$FF,$C3,$DB,$1F,$E4,$66,$BB,$8F,$0E,$1E,$CC,$FF,$85,$CF
	.dc.b	$BF,$8F,$F7,$F8,$7F,$08,$C1,$EA,$DC,$7B,$12,$FB,$3E,$1F,$5E,$47
	.dc.b	$85,$AF,$A8,$65,$E1,$C3,$E5,$C7,$84,$FF,$37,$E1,$E0,$70,$15,$74
	.dc.b	$8F,$F3,$7C,$61,$F0,$1F,$07,$D7,$F8,$84,$70,$7C,$5C,$7C,$2F,$93
	.dc.b	$2F,$82,$F8,$2B,$C4,$E4,$F2,$D7,$37,$E9,$35,$F1,$D7,$D3,$23,$DC
	.dc.b	$87,$D9,$72,$D3,$27,$22,$9F,$E9,$19,$F1,$BF,$03,$D7,$FC,$01,$1F
	.dc.b	$C0,$35,$FC,$33,$5E,$06,$88,$68,$A7,$CC,$F5,$FF,$18,$47,$F1,$8F
	.dc.b	$5C,$FF,$4B,$4A,$FC,$41,$FD,$43,$EE,$1E,$B7,$E2,$EF,$96,$1F,$99
	.dc.b	$6B,$F6,$BD,$8F,$CD,$11,$FC,$D3,$FD,$AE,$B9,$C7,$D5,$BE,$EF,$E1
	.dc.b	$A3,$2F,$D7,$3F,$11,$3C,$DF,$D4,$35,$FF,$5B,$C7,$F7,$74,$D2,$3E
	.dc.b	$1B,$EE,$F5,$C3,$E0,$3E,$18,$F8,$0F,$9C,$87,$FA,$E2,$7A,$47,$38
	.dc.b	$C3,$5F,$F6,$2C,$7A,$9F,$2C,$65,$A7,$02,$1C,$1D,$3F,$ED,$78,$7A
	.dc.b	$FF,$24,$91,$E3,$7C,$19,$7F,$1E,$9F,$92,$DA,$FF,$04,$FE,$70,$F1
	.dc.b	$BE,$E5,$F2,$C3,$CE,$D7,$E5,$FB,$1F,$5C,$8F,$C1,$7B,$F0,$79,$70
	.dc.b	$FC,$31,$DA,$FF,$11,$3C,$2B,$E1,$B5,$FA,$FE,$C7,$E0,$C8,$FC,$D7
	.dc.b	$AB,$FE,$F8,$CF,$D7,$87,$78,$FE,$17,$97,$AD,$FC,$45,$72,$7E,$8B
	.dc.b	$5F,$C2,$FB,$1F,$91,$23,$F6,$DE,$EF,$E2,$32,$F5,$BD,$B2,$FD,$14
	.dc.b	$F0,$8F,$BA,$D7,$F2,$3E,$27,$E7,$48,$FE,$1D,$AF,$E1,$1A,$F0,$34
	.dc.b	$43,$45,$3F,$3B,$EC,$7E,$A4,$8F,$E4,$DA,$FE,$41,$AF,$03,$44,$34
	.dc.b	$53,$F5,$3E,$2F,$ED,$48,$FE,$7D,$AF,$E7,$1A,$F0,$34,$43,$45,$3F
	.dc.b	$6B,$E2,$FF,$01,$23,$FA,$B6,$BF,$A8,$6B,$C0,$D1,$0D,$14,$FE,$03
	.dc.b	$EC,$7F,$0D,$23,$FB,$76,$BF,$B4,$6B,$C0,$D1,$0D,$14,$10,$7A,$1F
	.dc.b	$E0,$C7,$06,$36,$FF,$44,$46,$63,$83,$5E,$6D,$78,$1C,$8C,$C7,$22
	.dc.b	$39,$36,$6E,$39,$13,$C3,$F0,$92,$38,$47,$F9,$A1,$F1,$FA,$9F,$C1
	.dc.b	$3E,$22,$47,$F7,$69,$4F,$FA,$41,$E1,$1E,$8A,$E8,$F8,$5F,$7F,$B1
	.dc.b	$23,$BB,$5F,$35,$76,$7F,$9E,$71,$30,$ED,$3D,$EB,$85,$7C,$CC,$11
	.dc.b	$88,$F9,$7E,$C7,$CE,$23,$EE,$7E,$8D,$77,$E7,$FF,$20,$E3,$DE,$BD
	.dc.b	$D3,$ED,$AE,$31,$F2,$DA,$F8,$7D,$8F,$B6,$47,$E5,$BD,$9F,$E2,$79
	.dc.b	$77,$D3,$E6,$C7,$CB,$9E,$0F,$D7,$6B,$ED,$BE,$4F,$C6,$3E,$BB,$87
	.dc.b	$46,$08,$6C,$86,$8A,$7E,$43,$CC,$FC,$99,$1F,$C1,$BE,$0C,$BF,$CC
	.dc.b	$BA,$7E,$02,$7E,$29,$FC,$BB,$C2,$FC,$15,$71,$8F,$C8,$35,$FA,$DE
	.dc.b	$C7,$E8,$08,$FE,$41,$ED,$97,$C4,$8F,$EF,$97,$DD,$FC,$1D,$FE,$3E
	.dc.b	$B8,$C7,$E7,$5A,$FE,$BB,$F3,$07,$B3,$FE,$F5,$6C,$7E,$95,$AF,$EA
	.dc.b	$DA,$FC,$5C,$BD,$F9,$FE,$42,$3D,$C7,$F9,$5A,$76,$FE,$FD,$AF,$F0
	.dc.b	$26,$A6,$9F,$C1,$1F,$16,$3E,$52,$3F,$62,$BC,$2F,$CE,$3E,$09,$E1
	.dc.b	$8D,$BE,$96,$BF,$07,$D8,$F8,$04,$7E,$03,$F0,$FA,$7F,$8B,$57,$FA
	.dc.b	$83,$C4,$6F,$8D,$BF,$8D,$46,$30,$F8,$17,$97,$DD,$87,$57,$E5,$35
	.dc.b	$FA,$5E,$C7,$D1,$23,$F4,$5F,$13,$FE,$80,$AE,$CF,$F8,$ED,$B8,$FF
	.dc.b	$AD,$33,$FE,$F5,$3F,$E3,$F5,$5D,$23,$ED,$35,$FC,$17,$B1,$F8,$12
	.dc.b	$3F,$81,$7C,$2F,$7C,$FD,$8B,$0F,$0E,$7F,$77,$3C,$E2,$C3,$9D,$FE
	.dc.b	$25,$AF,$E4,$7D,$8F,$C8,$11,$FC,$83,$E0,$CF,$E8,$65,$E0,$CB,$BD
	.dc.b	$7E,$27,$1E,$F7,$C6,$3F,$2C,$D7,$F3,$D1,$EA,$69,$F9,$D7,$0F,$A4
	.dc.b	$C1,$0F,$D0,$3A,$7C,$06,$4F,$EA,$FF,$98,$BD,$9F,$E9,$90,$8C,$3F
	.dc.b	$D3,$DA,$FF,$55,$6B,$E4,$3E,$C4,$FC,$A3,$E0,$7A,$7F,$42,$77,$3F
	.dc.b	$DA,$5A,$FF,$6C,$6B,$E1,$AF,$1C,$3D,$9A,$49,$F2,$3F,$3F,$6F,$85
	.dc.b	$1E,$23,$F2,$17,$E1,$9E,$11,$E9,$6B,$F0,$BD,$9E,$E4,$7E,$03,$D5
	.dc.b	$F9,$27,$53,$F3,$53,$ED,$8E,$11,$F1,$5A,$FC,$EF,$67,$D6,$47,$E6
	.dc.b	$3D,$5F,$E1,$59,$7C,$97,$D2,$7F,$BA,$31,$4F,$18,$FA,$4D,$7E,$DF
	.dc.b	$C2,$7C,$82,$3F,$65,$AF,$D9,$6B,$C0,$D1,$0D,$14,$FC,$3F,$B5,$F5
	.dc.b	$08,$FE,$15,$AF,$E1,$5A,$F0,$34,$43,$45,$3F,$27,$E2,$FE,$00,$8F
	.dc.b	$E4,$5A,$FE,$45,$AF,$03,$44,$34,$53,$F3,$FF,$0B,$F8,$C2,$3F,$9D
	.dc.b	$6B,$F9,$D6,$BC,$0D,$10,$D1,$4F,$D5,$F8,$DF,$98,$23,$FA,$96,$BF
	.dc.b	$A9,$6B,$C0,$D1,$0D,$14,$FD,$BF,$8D,$FA,$42,$3F,$B5,$6B,$FB,$56
	.dc.b	$BC,$0D,$10,$D1,$4F,$E0,$7E,$37,$EC,$08,$FF,$01,$6B,$FC,$05,$AF
	.dc.b	$03,$44,$34,$53,$F8,$7F,$8D,$FB,$C2,$3F,$C3,$5A,$FF,$0D,$6B,$C0
	.dc.b	$D1,$0D,$14,$FE,$37,$E3,$7F,$08,$23,$FC,$65,$AF,$F1,$96,$BC,$0D
	.dc.b	$10,$D1,$4F,$F3,$EF,$E6,$5F,$E7,$72,$71,$FF,$28,$6C,$7F,$9D,$5A
	.dc.b	$FF,$9E,$DA,$F2,$8E,$04,$78,$71,$73,$23,$82,$1C,$DC,$B4,$C9,$E6
	.dc.b	$72,$D3,$C9,$E3,$F9,$C8,$E0,$87,$9D,$CB,$4C,$9F,$09,$CB,$4F,$81
	.dc.b	$ED,$7C,$32,$38,$21,$F0,$DC,$B4,$C9,$F3,$9C,$B4,$F9,$9E,$D7,$CF
	.dc.b	$23,$82,$1F,$3D,$CB,$4C,$9F,$69,$CB,$4F,$B1,$E4,$FD,$B2,$38,$21
	.dc.b	$F6,$DC,$B4,$C9,$F8,$67,$2D,$3F,$09,$ED,$7E,$1C,$8E,$08,$7E,$1D
	.dc.b	$CB,$4C,$9F,$92,$72,$D3,$F2,$1F,$11,$F9,$32,$38,$21,$F9,$37,$2D
	.dc.b	$32,$7E,$79,$CB,$4F,$CE,$79,$5F,$9F,$23,$82,$1F,$9F,$72,$D3,$27
	.dc.b	$EA,$9C,$B4,$FD,$47,$B7,$FA,$B2,$38,$21,$FA,$B7,$2D,$32,$7E,$D9
	.dc.b	$CB,$4F,$F7,$47,$E7,$0D,$FF,$C3,$ED,$8C,$C6,$0C,$F0,$FF,$92,$24
	.dc.b	$D3,$1C,$39,$9E,$15,$C8,$87,$F0,$1F,$8B,$FE,$0C,$47,$F8,$33,$FE
	.dc.b	$30,$D1,$F1,$9D,$F9,$65,$07,$86,$E4,$C4,$7F,$B9,$2B,$FC,$BB,$CA
	.dc.b	$C7,$B4,$F8,$4E,$E7,$F1,$5F,$7F,$F8,$C1,$1C,$1D,$0E,$B9,$79,$1F
	.dc.b	$16,$3E,$17,$C9,$B3,$3F,$3C,$79,$B3,$E3,$9F,$93,$1E,$D3,$FC,$8D
	.dc.b	$AF,$7C,$7B,$3E,$13,$D8,$52,$99,$76,$F1,$9F,$94,$47,$84,$69,$F1
	.dc.b	$98,$8D,$3C,$DF,$A5,$F1,$D2,$38,$B5,$D1,$F0,$AF,$9B,$EF,$3E,$B5
	.dc.b	$7A,$B2,$F8,$D0,$C9,$D5,$FE,$92,$A6,$1D,$6B,$E1,$35,$FA,$3F,$57
	.dc.b	$F8,$52,$3E,$C7,$FC,$EF,$F9,$A3,$E9,$7F,$2D,$DF,$2B,$CC,$7E,$26
	.dc.b	$74,$E0,$60,$BE,$B1,$F3,$5A,$FE,$0E,$78,$BF,$55,$AE,$6F,$C6,$47
	.dc.b	$EC,$46,$9F,$CB,$B5,$F2,$46,$DE,$8D,$7A,$1F,$B8,$F5,$BF,$6E,$47
	.dc.b	$07,$68,$FF,$5E,$CF,$9B,$4F,$A2,$DD,$7F,$4F,$AF,$F9,$AA,$FE,$12
	.dc.b	$C3,$A4,$7E,$46,$30,$F6,$3F,$29,$F9,$AB,$B7,$FD,$4E,$D8,$FC,$B3
	.dc.b	$5D,$C6,$88,$E9,$AE,$8C,$10,$C1,$E0,$ED,$EE,$47,$BB,$5D,$18,$21
	.dc.b	$82,$1F,$BD,$94,$71,$8C,$76,$FB,$08,$FB,$1A,$FC,$26,$BC,$0D,$10
	.dc.b	$D1,$4F,$06,$DF,$20,$8F,$C8,$6B,$F3,$9A,$F0,$34,$43,$45,$3D,$9D
	.dc.b	$BF,$50,$8F,$D4,$6B,$F6,$9A,$F0,$34,$43,$45,$3E,$47,$6F,$E0,$08
	.dc.b	$FE,$01,$AF,$E1,$9A,$F0,$34,$43,$45,$3E,$A7,$6F,$E3,$08,$FE,$31
	.dc.b	$AF,$E4,$9A,$F0,$34,$43,$45,$3F,$01,$DB,$F9,$82,$3F,$98,$6B,$F9
	.dc.b	$E6,$BC,$0D,$10,$D1,$4F,$C6,$7C,$4F,$E9,$08,$FE,$91,$AF,$EA,$9A
	.dc.b	$F0,$34,$43,$45,$3F,$31,$E8,$FE,$C0,$8F,$EC,$1A,$FE,$D9,$AF,$03
	.dc.b	$44,$34,$53,$F4,$9D,$3F,$BC,$23,$FB,$C6,$BF,$C0,$9A,$F0,$34,$43
	.dc.b	$45,$3F,$B6,$6E,$F8,$BF,$F0,$89,$23,$FC,$2A,$FF,$5F,$5F,$9C,$3C
	.dc.b	$8F,$F0,$94,$43,$F9,$D4,$65,$FE,$37,$3C,$DF,$E3,$4D,$7F,$BE,$7E
	.dc.b	$6E,$F3,$BB,$36,$3F,$B7,$35,$FE,$E4,$D7,$8C,$70,$23,$C1,$AF,$06
	.dc.b	$BC,$0E,$BE,$9F,$80,$8E,$0E,$47,$F5,$59,$F1,$74,$FF,$6A,$6F,$47
	.dc.b	$F8,$83,$15,$FF,$89,$2D,$F8,$83,$DF,$FF,$12,$E1,$A9,$EC,$7B,$2B
	.dc.b	$C2,$FF,$83,$ED,$F0,$3F,$45,$B1,$FE,$11,$6B,$B8,$D4,$F8,$70,$F0
	.dc.b	$4F,$F8,$DE,$FE,$54,$62,$11,$8C,$3D,$55,$C5,$F6,$5F,$84,$95,$E1
	.dc.b	$7C,$9E,$04,$8F,$23,$5F,$55,$73,$7C,$5A,$7F,$91,$B1,$F1,$63,$E1
	.dc.b	$9F,$F2,$66,$5F,$2B,$4F,$F2,$75,$79,$E3,$FC,$2B,$3D,$88,$7C,$7C
	.dc.b	$7E,$37,$A9,$23,$E3,$35,$F1,$DF,$4D,$3E,$9E,$5E,$36,$CE,$4F,$1E
	.dc.b	$7E,$97,$F4,$37,$CD,$FF,$2A,$62,$8C,$47,$E0,$FE,$0B,$EC,$11,$FB
	.dc.b	$0F,$D6,$C7,$B4,$7D,$2C,$7C,$4F,$C6,$33,$7F,$0D,$FC,$5E,$3D,$5E
	.dc.b	$64,$3F,$11,$7F,$87,$F7,$64,$7E,$1D,$AF,$C2,$39,$9D,$CA,$78,$CF
	.dc.b	$0E,$5F,$06,$D3,$CA,$7E,$0B,$5F,$CE,$F8,$5F,$BA,$23,$FB,$A7,$E0
	.dc.b	$3D,$B6,$FC,$03,$DC,$F0,$D7,$BF,$6F,$C2,$D7,$29,$FE,$02,$D7,$F8
	.dc.b	$4F,$17,$F0,$82,$3F,$C2,$1E,$F3,$F8,$ED,$3E,$6C,$F7,$8F,$C6,$65
	.dc.b	$F7,$71,$FE,$31,$3F,$C3,$AF,$9B,$4F,$F4,$06,$9F,$C8,$DF,$D5,$1F
	.dc.b	$E8,$57,$06,$72,$FB,$1F,$3B,$E1,$9F,$F5,$43,$53,$E2,$DB,$35,$FE
	.dc.b	$AA,$39,$CF,$F3,$16,$BE,$2E,$3C,$11,$C1,$FC,$17,$2F,$66,$9F,$03
	.dc.b	$3F,$9F,$97,$CD,$C3,$B6,$1C,$70,$F5,$35,$FF,$58,$43,$D5,$F1,$12
	.dc.b	$3E,$1B,$F3,$A3,$F5,$31,$F7,$70,$F9,$F7,$DE,$3F,$D9,$13,$CD,$E2
	.dc.b	$43,$E5,$57,$E1,$9F,$D6,$E3,$FC,$AD,$B9,$7B,$48,$FE,$6C,$C9,$4F
	.dc.b	$E3,$6E,$9C,$90,$FE,$71,$F0,$AF,$BC,$FE,$58,$D8,$FE,$64,$D7,$F7
	.dc.b	$ED,$78,$C7,$02,$3C,$1A,$F0,$6B,$C0,$FF,$73,$FE,$0A,$20

;--------------------------------------------------------------------------------
;	FMOVE.L FPn,Dr
;	FMOVE.L FPn,<mem>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovelregto_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.L FPN,<EA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.L FPn,<ea>',13,10
;------------------------------------------------
;	d1	actual result
;	d4	actual status
;	d6	fpcr=(XRN..DRP)<<4
;	d7	0=failed,1=successful
;	a0	expected result,expected status,...
;	a1	expected result
;	a4	expected status
;	a5	source handle,...
;	fp5	source
;------------------------------------------------
	lea.l	push_decompressed,a3
;decompress data
	move.l	a3,-(sp)
	pea.l	fmovelregto_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a3,a5			;source handle,...
@@:
	add.l	d0,(a3)+		;source handle
	tst.l	(a3)
	bpl	@b
	addq.l	#4,a3			;-1
	movea.l	a3,a0			;expected result,expected status,...
;
55:
	move.l	#0,d6		;fpcr=(XRN..DRP)<<4
66:
;FPn,Dr
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	#-1,d1			;fill=-1
					;source
	fmove.l	fp5,d1			;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.L FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',Dr @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result,expected status,...
;FPn,<mem>
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	#-1,-(sp)		;fill=-1
					;source
	fmove.l	fp5,(sp)		;EXECUTE
	move.l	(sp)+,d1		;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.L FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',<mem> @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result,expected status,...
;
	add.w	#1<<4,d6		;rprm++
	cmp.w	#176,d6		;fpcr=(XRN..DRP)<<4
	bls	66b
;
	addq.l	#4,a5			;source handle
	tst.l	(a5)			;source handle
	bpl	55b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmovelregto_data_compressed::
	.dc.b	$00,$04,$C0,$98,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$9F,$EE,$EB,$FC,$CB,$E0,$51,$23,$31,$88,$DC,$7E,$C1
	.dc.b	$D3,$F8,$68,$D3,$C2,$32,$FF,$2A,$8C,$BB,$8C,$BE,$16,$BD,$5E,$F1
	.dc.b	$C2,$39,$46,$A3,$11,$C6,$36,$64,$86,$08,$F9,$5A,$F8,$DA,$F6,$6B
	.dc.b	$D4,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3E,$21
	.dc.b	$D4,$A7,$22,$9C,$4A,$71,$8F,$8E,$E0,$8E,$09,$1F,$22,$3B,$46,$23
	.dc.b	$E1,$C7,$07,$4F,$86,$34,$F0,$8C,$BE,$38,$CB,$B8,$CB,$E5,$6B,$D5
	.dc.b	$EF,$1C,$23,$94,$6A,$31,$1C,$63,$66,$48,$60,$8F,$95,$AF,$8D,$AF
	.dc.b	$66,$BD,$4E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$43
	.dc.b	$E2,$1D,$4A,$72,$29,$C4,$A7,$18,$F8,$EE,$08,$E0,$91,$F2,$23,$B4
	.dc.b	$62,$3E,$1C,$70,$74,$F8,$63,$4F,$08,$CB,$E3,$8C,$BB,$8C,$BE,$56
	.dc.b	$BD,$5E,$F1,$C2,$39,$46,$A3,$11,$C6,$36,$64,$86,$08,$F9,$5A,$F8
	.dc.b	$DA,$F6,$6B,$D4,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$3E,$21,$D4,$A7,$22,$9C,$4A,$71,$8F,$8E,$E0,$8E,$09,$1F,$22
	.dc.b	$3B,$46,$23,$E1,$C7,$07,$4F,$86,$34,$F0,$8C,$BE,$38,$CB,$B8,$CB
	.dc.b	$E5,$6B,$D5,$EF,$1C,$23,$94,$6A,$31,$1C,$63,$66,$48,$60,$8F,$95
	.dc.b	$AF,$8D,$AF,$66,$BD,$4E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3
	.dc.b	$04,$30,$43,$E2,$1D,$4A,$72,$29,$C4,$A7,$18,$F8,$EE,$08,$E0,$91
	.dc.b	$F2,$23,$B4,$62,$3E,$1C,$70,$74,$F8,$63,$4F,$08,$CB,$E3,$8C,$BB
	.dc.b	$8C,$BE,$56,$BD,$5E,$F1,$C2,$39,$46,$A3,$11,$C6,$36,$64,$86,$08
	.dc.b	$F9,$5A,$F8,$DA,$F6,$6B,$D4,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2
	.dc.b	$9A,$30,$43,$04,$3E,$21,$D4,$A7,$22,$9C,$4A,$71,$8F,$8E,$E0,$8E
	.dc.b	$09,$1F,$22,$3B,$46,$23,$E1,$C7,$07,$4F,$86,$34,$F0,$8C,$BE,$38
	.dc.b	$CB,$B8,$CB,$E5,$6B,$D5,$EF,$1C,$23,$94,$6A,$31,$1C,$63,$66,$48
	.dc.b	$60,$8F,$95,$AF,$8D,$AF,$66,$BD,$4E,$04,$38,$14,$E0,$53,$81,$A2
	.dc.b	$1A,$29,$A3,$04,$30,$43,$E2,$1D,$4A,$72,$29,$C4,$A7,$18,$F8,$EE
	.dc.b	$08,$E0,$91,$F2,$23,$B4,$62,$3E,$1C,$70,$74,$F8,$63,$4F,$08,$CB
	.dc.b	$E3,$8C,$BB,$8C,$BE,$56,$BD,$5E,$F1,$C2,$39,$46,$A3,$11,$C6,$36
	.dc.b	$64,$86,$08,$F9,$5A,$F8,$DA,$F6,$6B,$D4,$E0,$43,$81,$4E,$05,$38
	.dc.b	$1A,$21,$A2,$9A,$30,$43,$04,$3E,$21,$D4,$A7,$22,$9C,$4A,$71,$8F
	.dc.b	$8E,$E0,$8E,$09,$1F,$22,$3B,$46,$23,$E1,$C7,$07,$4F,$86,$34,$F0
	.dc.b	$8C,$BE,$38,$CB,$B8,$CB,$E5,$6B,$D5,$EF,$1C,$23,$94,$6A,$31,$1C
	.dc.b	$63,$66,$48,$7F,$B6,$3E,$85,$F0,$1F,$EC,$E6,$C6,$63,$04,$70,$49
	.dc.b	$FF,$73,$CF,$06,$CE,$DB,$3B,$44,$13,$C5,$C1,$89,$E4,$E2,$34,$F1
	.dc.b	$6B,$D1,$AF,$19,$F0,$3B,$9E,$6D,$8C,$BF,$1C,$FA,$CF,$C8,$32,$F3
	.dc.b	$3D,$64,$3C,$91,$D1,$F1,$35,$E0,$6A,$39,$CF,$98,$93,$B6,$C6,$A3
	.dc.b	$04,$7C,$75,$F7,$AB,$F0,$87,$A0,$87,$E2,$E3,$D4,$E0,$FC,$3B,$87
	.dc.b	$46,$4F,$84,$60,$87,$03,$47,$80,$C9,$C0,$D1,$83,$46,$08,$6C,$86
	.dc.b	$8F,$92,$FC,$12,$3F,$3C,$FB,$05,$36,$53,$E7,$9A,$3E,$81,$9A,$F9
	.dc.b	$AD,$AF,$98,$D9,$EE,$FA,$6B,$E7,$1F,$3E,$B8,$C7,$42,$1D,$C8,$79
	.dc.b	$0C,$94,$C9,$E6,$30,$43,$04,$7B,$BC,$1A,$ED,$AE,$EB,$D6,$F8,$0C
	.dc.b	$57,$B1,$E5,$1A,$6C,$65,$E9,$19,$6C,$65,$E6,$7C,$72,$1D,$E7,$C6
	.dc.b	$49,$E5,$1C,$0D,$38,$8D,$4F,$56,$CF,$43,$CA,$53,$C6,$53,$A1,$A3
	.dc.b	$AC,$73,$9E,$C4,$70,$48,$E5,$3E,$07,$89,$89,$E0,$D9,$E0,$E8,$87
	.dc.b	$84,$87,$98,$C9,$4C,$9E,$83,$04,$30,$49,$F2,$3C,$1A,$ED,$AE,$E7
	.dc.b	$CE,$F8,$0C,$4F,$A1,$E5,$1A,$6C,$65,$E9,$19,$6C,$65,$E6,$7C,$32
	.dc.b	$1D,$E7,$C6,$49,$E5,$1C,$0D,$38,$8D,$4F,$56,$CF,$43,$CA,$53,$C6
	.dc.b	$53,$A1,$F3,$2B,$F6,$F3,$FD,$7A,$3F,$C0,$93,$D1,$F0,$C7,$7A,$F1
	.dc.b	$11,$C1,$23,$B4,$F1,$7C,$67,$23,$33,$C5,$B3,$C5,$D1,$0E,$24,$3D
	.dc.b	$66,$4A,$64,$F6,$98,$21,$82,$4F,$95,$CB,$5C,$CF,$95,$EC,$62,$7C
	.dc.b	$CF,$79,$E0,$49,$E1,$19,$6B,$D9,$AF,$33,$E0,$90,$EF,$1C,$E7,$BC
	.dc.b	$71,$6B,$C4,$DC,$F4,$9E,$11,$CC,$F1,$4F,$08,$CC,$70,$8C,$4F,$12
	.dc.b	$38,$24,$73,$3C,$45,$3F,$A9,$12,$BF,$A0,$4F,$28,$D5,$78,$1F,$19
	.dc.b	$88,$D3,$E3,$31,$5E,$07,$07,$F3,$5A,$FC,$BC,$7F,$73,$8F,$F2,$B5
	.dc.b	$75,$7C,$A7,$03,$B9,$4C,$98,$21,$82,$1F,$34,$C9,$4C,$BD,$DC,$11
	.dc.b	$C1,$2B,$DB,$1C,$A3,$13,$ED,$8C,$4E,$9E,$AD,$7A,$36,$78,$C7,$C6
	.dc.b	$21,$ED,$37,$3D,$5E,$6D,$72,$70,$3C,$2E,$63,$AC,$73,$6B,$A3,$81
	.dc.b	$D0,$A7,$43,$84,$F9,$C9,$3D,$5B,$1F,$C4,$EB,$ED,$BF,$E2,$67,$FD
	.dc.b	$1B,$5E,$08,$EC,$74,$AE,$44,$7C,$31,$B7,$C4,$79,$CE,$06,$AB,$C2
	.dc.b	$E0,$86,$C8,$78,$0A,$74,$30,$43,$04,$8F,$11,$23,$BC,$72,$8C,$4F
	.dc.b	$A6,$3C,$27,$38,$CB,$E7,$8D,$3C,$67,$9B,$B6,$BE,$16,$CF,$28,$EA
	.dc.b	$66,$79,$BC,$5A,$FA,$AB,$E9,$BF,$E3,$B8,$FF,$75,$CF,$38,$F1,$47
	.dc.b	$08,$C4,$73,$3D,$04,$3C,$2E,$9F,$11,$93,$F5,$B3,$E0,$7F,$CB,$E7
	.dc.b	$F0,$B9,$FE,$A8,$F8,$DF,$E0,$0F,$F9,$8A,$38,$36,$36,$F2,$8C,$BC
	.dc.b	$E3,$8B,$4F,$C8,$3C,$1E,$4E,$A3,$89,$1D,$47,$13,$C8,$72,$78,$1A
	.dc.b	$8E,$0E,$09,$1B,$21,$C8,$87,$A5,$E4,$43,$8B,$A3,$CE,$72,$38,$3A
	.dc.b	$69,$93,$67,$13,$D4,$68,$C4,$7A,$1F,$F6,$68,$51,$23,$31,$88,$DC
	.dc.b	$6D,$D3,$FE,$D6,$8D,$3C,$23,$F2,$F5,$1F,$C6,$3F,$E2,$29,$F1,$4F
	.dc.b	$92,$79,$BF,$EE,$38,$E6,$F5,$8C,$BF,$EE,$88,$CB,$B8,$EC,$FE,$44
	.dc.b	$FC,$C1,$F9,$69,$CB,$63,$C6,$47,$81,$F9,$A3,$46,$2B,$F3,$C4,$9E
	.dc.b	$71,$FA,$83,$4E,$1F,$09,$93,$99,$0C,$10,$D9,$0F,$11,$92,$99,$8F
	.dc.b	$14,$FA,$63,$A1,$99,$F1,$3C,$09,$3C,$1F,$3C,$F1,$7F,$91,$D7,$D4
	.dc.b	$7F,$10,$7C,$D8,$FC,$0C,$7D,$08,$FA,$51,$F5,$1F,$C2,$3A,$3C,$33
	.dc.b	$DE,$3C,$A7,$8A,$B2,$49,$F1,$47,$A5,$AE,$DB,$3C,$E3,$E2,$35,$E0
	.dc.b	$7B,$08,$78,$1D,$3C,$CC,$9C,$A7,$C4,$E0,$E6,$71,$72,$D9,$F1,$C6
	.dc.b	$23,$94,$60,$8F,$33,$C0,$71,$3F,$B3,$3F,$C0,$5F,$DD,$1F,$28,$FD
	.dc.b	$53,$F2,$C8,$E0,$E9,$3E,$C2,$4F,$8E,$3D,$27,$90,$FF,$63,$57,$DD
	.dc.b	$7F,$83,$9F,$74,$E8,$7E,$BE,$BB,$B4,$FF,$67,$7D,$CB,$E3,$BF,$00
	.dc.b	$D9,$FF,$72,$36,$7F,$86,$B5,$FF,$59,$35,$11,$C3,$FE,$C1,$6A,$6E
	.dc.b	$1F,$EF,$E7,$26,$30,$FC,$94,$1F,$2D,$CB,$F2,$B6,$FD,$6D,$7C,$AD
	.dc.b	$7F,$DA,$B7,$DA,$3F,$DF,$2D,$9E,$67,$F2,$47,$C4,$FF,$26,$BF,$13
	.dc.b	$99,$E9,$5E,$D6,$BF,$ED,$EA,$E8,$F9,$27,$81,$D9,$F9,$73,$E2,$38
	.dc.b	$D7,$C3,$6D,$FC,$23,$AD,$F9,$2B,$F9,$FD,$79,$1E,$77,$D9,$F0,$18
	.dc.b	$29,$F5,$8F,$AD,$7E,$18,$E1,$1F,$21,$AF,$C8,$8F,$1C,$7C,$3B,$F1
	.dc.b	$C7,$19,$EC,$71,$29,$C6,$79,$BE,$06,$BF,$83,$8E,$AF,$B2,$7A,$CE
	.dc.b	$AF,$E5,$CF,$A4,$E1,$1F,$2D,$C4,$69,$F7,$35,$F0,$1F,$08,$CD,$76
	.dc.b	$7C,$E6,$8F,$51,$D1,$D3,$67,$D3,$7F,$5D,$F3,$4E,$A7,$AC,$79,$5A
	.dc.b	$76,$3D,$25,$AF,$B8,$FB,$6B,$EE,$3D,$CD,$57,$57,$CC,$7C,$D3,$35
	.dc.b	$E2,$3C,$B7,$F8,$97,$11,$A2,$1E,$E3,$CA,$77,$21,$82,$1F,$9F,$79
	.dc.b	$BA,$3C,$64,$3F,$45,$3E,$77,$DE,$6C,$A7,$A4,$C5,$FD,$77,$27,$3B
	.dc.b	$F9,$CF,$A0,$F1,$9A,$23,$A7,$9C,$F8,$67,$E6,$C7,$E8,$67,$E6,$BC
	.dc.b	$A3,$C0,$49,$FC,$E4,$72,$AF,$7B,$E0,$AE,$CF,$88,$FE,$43,$1F,$E2
	.dc.b	$C6,$95,$C3,$FD,$B4,$C9,$FE,$32,$72,$47,$81,$0D,$14,$FF,$1F,$BB
	.dc.b	$72,$43,$26,$CF,$D4,$3B,$72,$6C,$C9,$B3,$E6,$57,$F3,$FF,$BA,$FC
	.dc.b	$29,$F1,$EF,$F9,$73,$89,$FC,$04,$FE,$DA,$31,$1C,$A7,$6F,$EB,$63
	.dc.b	$95,$7F,$10,$77,$5F,$46,$3F,$8B,$46,$2B,$F6,$2D,$8F,$C8,$B6,$BF
	.dc.b	$36,$D7,$EC,$1F,$D5,$1C,$3E,$03,$07,$8C,$F4,$CF,$99,$FC,$BC,$6D
	.dc.b	$FD,$23,$C9,$DC,$F9,$9E,$47,$BD,$F6,$38,$AF,$63,$E0,$3A,$9C,$8F
	.dc.b	$E5,$33,$D5,$E4,$6E,$3E,$93,$B3,$C6,$75,$AF,$8A,$DA,$F8,$AD,$3E
	.dc.b	$29,$4F,$8A,$2B,$C1,$5F,$66,$34,$7C,$18,$CD,$7C,$59,$F1,$CF,$98
	.dc.b	$F1,$9E,$C3,$07,$D3,$9F,$94,$D3,$D4,$53,$67,$95,$DB,$4D,$94,$D9
	.dc.b	$DE,$3C,$67,$CE,$7D,$87,$A8,$F9,$11,$DE,$37,$3F,$EA,$08,$F6,$D7
	.dc.b	$D8,$AE,$93,$97,$C1,$5C,$27,$E6,$C6,$9F,$C1,$34,$FB,$13,$C5,$F3
	.dc.b	$CF,$17,$33,$C2,$31,$1E,$83,$91,$F5,$AB,$D1,$5C,$E7,$71,$F1,$A7
	.dc.b	$6E,$AB,$A5,$6C,$FC,$2C,$E8,$F5,$B9,$8F,$5D,$74,$9C,$C7,$23,$33
	.dc.b	$B8,$F4,$90,$F4,$4F,$A9,$FC,$41,$C8,$D5,$7C,$E7,$F3,$E6,$4E,$33
	.dc.b	$D5,$B5,$E0,$3C,$4F,$23,$81,$D8,$F6,$12,$BC,$4F,$E0,$0E,$D5,$F1
	.dc.b	$5F,$D1,$9B,$AF,$04,$79,$63,$F3,$73,$F0,$DF,$F1,$45,$7D,$E7,$F9
	.dc.b	$CC,$FF,$2A,$23,$FC,$A8,$FB,$46,$8C,$3A,$79,$9F,$C0,$CD,$18,$38
	.dc.b	$9F,$89,$32,$43,$04,$8F,$23,$4E,$85,$39,$BA,$23,$F2,$CF,$EE,$B3
	.dc.b	$FE,$02,$3E,$33,$FE,$77,$3F,$C4,$8E,$0D,$3F,$65,$AF,$CB,$8F,$F1
	.dc.b	$8B,$C0,$D1,$4F,$F2,$07,$97,$FD,$90,$8F,$F6,$96,$BF,$E3,$F3,$FD
	.dc.b	$1A,$E1,$C3,$5F,$F2,$23,$5F,$F2,$F3,$5E,$27,$52,$1D,$43,$B6,$9A
	.dc.b	$3B,$94,$EE,$53,$FC,$D9,$CF,$FD,$F8,$8F,$F8,$0D,$AF,$76,$BE,$D6
	.dc.b	$BF,$E5,$D6,$BD,$0E,$44,$39,$14,$E4,$53,$91,$4E,$47,$02,$1C,$0F
	.dc.b	$85,$E4,$7F,$85,$88,$FF,$86,$DA,$E8,$C1,$0C,$10,$FB,$0F,$06,$9B
	.dc.b	$29,$B3,$FC,$F8,$ED,$A6,$CA,$6C,$FF,$4D,$3B,$7F,$D3,$9C,$FF,$E2
	.dc.b	$E2,$3F,$E3,$36,$BF,$7D,$AF,$E3,$5A,$FF,$A3,$5A,$F4,$39,$10,$E4
	.dc.b	$53,$91,$4E,$45,$39,$1C,$08,$70,$3F,$25,$E4,$7F,$91,$88,$FF,$92
	.dc.b	$DA,$E8,$C1,$0C,$10,$FD,$43,$C1,$A6,$CA,$6C,$FF,$56,$3B,$69,$B2
	.dc.b	$9B,$3F,$D9,$4E,$DF,$F6,$67,$F0,$0F,$F9,$78,$8F,$F9,$8D,$AF,$EB
	.dc.b	$DA,$FE,$F5,$AF,$FA,$F5,$AF,$43,$91,$0E,$45,$39,$14,$E4,$53,$91
	.dc.b	$C0,$87,$03,$C9,$C0,$23,$FE,$E6,$8C,$46,$23,$FD,$B1,$18,$8F,$F6
	.dc.b	$5B,$63,$84,$73,$8C,$B6,$32,$F3,$CD,$AA,$B9,$57,$6A,$F0,$BC,$1E
	.dc.b	$67,$19,$E0,$EA,$38,$3C,$DE,$F3,$C2,$24,$73,$21,$C1,$CB,$CA,$3B
	.dc.b	$98,$AF,$3B,$E6,$6C,$F8,$64,$79,$1F,$90,$E2,$7D,$AD,$8E,$0E,$DF
	.dc.b	$C2,$18,$8D,$C6,$4F,$C3,$46,$48,$76,$74,$7E,$51,$C9,$D0,$D4,$74
	.dc.b	$23,$A8,$EA,$E8,$C3,$CD,$C1,$0F,$89,$5E,$18,$57,$DC,$9F,$E9,$CF
	.dc.b	$C6,$7E,$C5,$7C,$9A,$E1,$3B,$AF,$EC,$0E,$EB,$A9,$CA,$76,$7B,$A3
	.dc.b	$53,$F4,$1E,$4F,$D2,$31,$1C,$2B,$C0,$78,$6B,$A1,$D5,$CB,$F8,$BA
	.dc.b	$E7,$5C,$DA,$FE,$21,$F0,$BF,$02,$BC,$B3,$CC,$E4,$F9,$CC,$C7,$B9
	.dc.b	$C5,$78,$E3,$13,$C4,$F8,$E7,$23,$CE,$F1,$8F,$3C,$7E,$8C,$CC,$7C
	.dc.b	$13,$F3,$87,$9E,$7C,$2E,$6B,$D2,$48,$CC,$7D,$CA,$F2,$3B,$8F,$95
	.dc.b	$5E,$C9,$CC,$FC,$59,$CC,$FA,$0F,$24,$7B,$0F,$43,$C4,$F6,$39,$3B
	.dc.b	$4F,$B2,$3A,$1E,$C8,$F2,$3C,$1F,$23,$F1,$CF,$C3,$BC,$CD,$BE,$27
	.dc.b	$67,$89,$F3,$1D,$9E,$07,$C7,$8F,$6B,$F8,$A8,$EC,$FC,$B7,$C8,$ED
	.dc.b	$F2,$3C,$CF,$5B,$EF,$3C,$4E,$4F,$2B,$5E,$C7,$95,$F5,$BE,$83,$9B
	.dc.b	$E8,$76,$64,$E6,$60,$E8,$7E,$55,$F3,$B4,$F7,$9F,$86,$25,$7E,$79
	.dc.b	$D1,$D4,$E8,$43,$27,$80,$A6,$5E,$44,$AF,$C6,$C7,$C7,$9F,$86,$EE
	.dc.b	$32,$F2,$24,$FE,$56,$BE,$9C,$62,$77,$3D,$6B,$EA,$CF,$AA,$72,$48
	.dc.b	$F8,$33,$CD,$FA,$F3,$E1,$73,$3C,$2B,$F2,$53,$93,$E3,$D7,$76,$CF
	.dc.b	$EE,$AB,$9C,$79,$5E,$67,$4A,$E7,$19,$38,$C6,$9F,$C0,$1F,$8D,$32
	.dc.b	$43,$E9,$1E,$D8,$F8,$75,$E3,$8C,$B5,$FA,$8D,$7D,$46,$AB,$F6,$F5
	.dc.b	$F8,$E8,$FC,$3C,$ED,$F3,$1F,$90,$8D,$3C,$8D,$CF,$C7,$8F,$CF,$18
	.dc.b	$3D,$91,$DD,$FA,$B3,$FC,$3D,$CC,$6C,$F9,$75,$F1,$A7,$C4,$4A,$ED
	.dc.b	$5F,$A1,$AF,$29,$F9,$E3,$FB,$8B,$FE,$9D,$69,$5C,$3F,$D5,$CC,$9F
	.dc.b	$EA,$37,$24,$78,$10,$D1,$4F,$F5,$B3,$B7,$24,$32,$6C,$FE,$68,$ED
	.dc.b	$C9,$B3,$26,$CF,$A3,$AF,$F0,$B3,$E3,$3F,$C0,$DC,$3F,$4C,$FE,$06
	.dc.b	$E1,$F8,$A6,$08,$60,$FD,$EB,$BA,$F9,$D3,$FD,$A6,$31,$5F,$A9,$AF
	.dc.b	$F0,$EC,$62,$7E,$EB,$63,$F2,$CD,$7F,$00,$70,$3C,$35,$D9,$C1,$E2
	.dc.b	$3D,$45,$23,$B5,$1D,$5F,$86,$D7,$F3,$AD,$7E,$49,$E0,$AF,$53,$C8
	.dc.b	$C5,$7A,$9A,$7A,$8A,$7A,$8E,$07,$92,$39,$CF,$7A,$FD,$93,$63,$FB
	.dc.b	$7B,$6B,$FB,$AB,$6B,$FB,$03,$5F,$B8,$78,$0F,$69,$4F,$69,$4F,$69
	.dc.b	$E3,$8F,$53,$4F,$51,$4F,$54,$77,$7E,$B3,$F8,$C6,$9F,$4A,$BF,$C5
	.dc.b	$71,$B8,$F8,$6E,$9E,$8F,$17,$24,$32,$73,$AF,$F4,$54,$71,$9E,$B3
	.dc.b	$B7,$EF,$46,$67,$73,$97,$E7,$B6,$BF,$2C,$D3,$DF,$5D,$1F,$09,$1E
	.dc.b	$AE,$CF,$B6,$EC,$93,$DA,$38,$3E,$26,$CF,$CE,$9D,$BE,$28,$C9,$F9
	.dc.b	$C8,$EE,$E0,$EF,$3C,$E3,$8C,$7A,$23,$04,$9E,$93,$A9,$E3,$1E,$19
	.dc.b	$D4,$77,$6B,$A3,$11,$B3,$9C,$6A,$BE,$74,$66,$76,$43,$91,$23,$EF
	.dc.b	$3E,$B7,$8B,$B3,$F2,$E7,$D0,$3F,$96,$BF,$C7,$88,$FF,$1E,$3F,$56
	.dc.b	$68,$C3,$A7,$FC,$F5,$1F,$35,$D1,$83,$F8,$49,$82,$1F,$5C,$AF,$17
	.dc.b	$04,$70,$47,$99,$92,$B9,$7E,$11,$FE,$30,$AF,$F1,$F1,$F0,$9F,$F5
	.dc.b	$09,$93,$F1,$CE,$9A,$64,$FC,$23,$96,$99,$7F,$D5,$5F,$91,$FF,$A0
	.dc.b	$49,$3F,$E7,$F6,$E1,$FE,$66,$7F,$DD,$D5,$FE,$3F,$77,$39,$38,$10
	.dc.b	$C1,$99,$93,$FE,$5F,$9E,$2D,$9E,$0D,$78,$B9,$9E,$2F,$86,$36,$D7
	.dc.b	$FB,$3C,$FF,$B8,$13,$48,$FF,$B6,$CC,$95,$CB,$C8,$C1,$0C,$12,$3F
	.dc.b	$B7,$C7,$06,$C6,$DA,$ED,$FC,$7C,$6D,$B1,$B6,$BB,$AF,$AA,$FF,$AC
	.dc.b	$76,$FF,$5C,$9F,$EB,$AA,$F9,$35,$FE,$71,$8C,$BF,$E9,$88,$CD,$7F
	.dc.b	$B9,$E7,$33,$F3,$0F,$90,$7F,$A4,$5C,$BF,$44,$87,$53,$FD,$46,$F8
	.dc.b	$6B,$E8,$CF,$59,$C9,$E1,$8C,$E2,$55,$DA,$7C,$24,$AE,$4F,$FA,$D2
	.dc.b	$3B,$4F,$9E,$73,$3E,$73,$B4,$66,$BB,$57,$5A,$F9,$33,$93,$C3,$5E
	.dc.b	$63,$17,$FE,$E1,$AE,$17,$E7,$8F,$92,$F4,$9E,$07,$C4,$BE,$AF,$8C
	.dc.b	$F4,$1E,$43,$51,$C5,$C9,$E0,$BF,$B8,$F4,$3E,$14,$78,$63,$31,$C0
	.dc.b	$E6,$FE,$60,$C3,$F5,$4E,$8E,$27,$E3,$C7,$48,$E0,$F9,$C8,$66,$BE
	.dc.b	$13,$E3,$9F,$6B,$C6,$FE,$2D,$72,$8C,$1B,$BF,$44,$7C,$F9,$DD,$FC
	.dc.b	$A8,$FA,$35,$E5,$AD,$BE,$CA,$DC,$FE,$20,$DC,$79,$AF,$C6,$4A,$F9
	.dc.b	$91,$97,$9B,$E4,$3B,$4F,$9E,$78,$3D,$23,$E4,$1C,$5C,$47,$C8,$24
	.dc.b	$F4,$78,$D7,$C2,$25,$7C,$1A,$ED,$5B,$3C,$B1,$D8,$95,$D2,$7B,$CE
	.dc.b	$A3,$CB,$7F,$4A,$BD,$37,$D9,$F5,$C7,$CB,$8C,$47,$A5,$F9,$37,$F4
	.dc.b	$63,$C9,$1A,$7C,$37,$E5,$9F,$39,$E2,$24,$EE,$FC,$D7,$DE,$F9,$90
	.dc.b	$F1,$D7,$D7,$AE,$17,$EB,$24,$75,$AE,$55,$EC,$39,$1F,$45,$CB,$EB
	.dc.b	$32,$7A,$23,$CF,$1F,$16,$34,$73,$3C,$71,$F0,$5C,$1A,$3B,$3E,$77
	.dc.b	$AD,$FD,$3B,$F9,$31,$F9,$33,$37,$EA,$8C,$C7,$E3,$4E,$87,$95,$CC
	.dc.b	$FE,$1A,$38,$D7,$CC,$7F,$0D,$7E,$E7,$81,$AA,$FE,$3E,$F8,$4F,$49
	.dc.b	$FE,$16,$7F,$D9,$C4,$7F,$D9,$8D,$7F,$D8,$ED,$7F,$D8,$6D,$79,$1C
	.dc.b	$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$7B,$B1,$FE,$4A,$7B
	.dc.b	$2F,$FA,$43,$88,$F9,$71,$87,$F9,$F4,$6F,$0F,$E6,$4E,$E3,$E4,$96
	.dc.b	$78,$38,$38,$90,$C1,$FD,$6E,$38,$4F,$DD,$8F,$EB,$31,$8C,$3E,$C4
	.dc.b	$5B,$FF,$2F,$9F,$C0,$27,$11,$FA,$46,$BF,$9D,$6B,$F9,$B6,$BF,$0C
	.dc.b	$F2,$5F,$E4,$5F,$09,$8B,$F2,$3D,$A7,$E1,$5F,$F5,$A7,$9D,$71,$6D
	.dc.b	$7A,$DB,$3F,$91,$6D,$FF,$81,$9B,$7E,$03,$E2,$1F,$4A,$BE,$7B,$E3
	.dc.b	$AF,$86,$DA,$F8,$6F,$C1,$30,$53,$99,$F4,$2F,$EA,$92,$BD,$B1,$C2
	.dc.b	$FD,$B5,$D1,$B1,$FE,$94,$8E,$EF,$CD,$3B,$94,$EF,$5F,$2D,$F9,$07
	.dc.b	$AE,$FE,$43,$C0,$FA,$85,$3E,$A1,$CA,$FE,$0B,$63,$8B,$6B,$F0,$4F
	.dc.b	$A4,$86,$4F,$85,$5E,$F9,$E2,$D9,$F7,$4F,$95,$AF,$1B,$F4,$BF,$26
	.dc.b	$FC,$75,$C6,$3E,$21,$27,$ED,$BC,$23,$51,$67,$E7,$CF,$53,$CF,$3A
	.dc.b	$7C,$47,$33,$31,$E0,$25,$FE,$0A,$39,$12,$79,$3E,$78,$F7,$CF,$0B
	.dc.b	$F6,$D6,$8F,$AD,$5D,$AB,$64,$9D,$D7,$A2,$B8,$3C,$EB,$CC,$7C,$F3
	.dc.b	$F0,$51,$AA,$E5,$1A,$8E,$EF,$9E,$79,$5F,$C5,$AF,$65,$71,$BF,$0C
	.dc.b	$72,$9F,$C9,$9F,$52,$C7,$E1,$8B,$1F,$86,$6C,$FF,$AF,$88,$FF,$B5
	.dc.b	$0E,$04,$8E,$0D,$8D,$CF,$F3,$52,$47,$F3,$56,$BF,$CF,$1A,$F8,$0E
	.dc.b	$D3,$D0,$93,$D2,$39,$10,$FC,$C5,$7F,$96,$8F,$89,$1E,$E7,$11,$F5
	.dc.b	$E3,$0F,$F9,$CE,$36,$FF,$B5,$8F,$F4,$43,$83,$8B,$F4,$67,$FC,$E7
	.dc.b	$18,$17,$FE,$87,$DB,$E6,$D4,$3E,$6B,$5F,$F7,$CB,$5C,$96,$91,$98
	.dc.b	$E8,$65,$C4,$6A,$30,$43,$B3,$96,$C7,$02,$3B,$3C,$D1,$B6,$C6,$9E
	.dc.b	$26,$CF,$49,$4F,$49,$A3,$C3,$B7,$98,$8F,$B5,$AE,$8C,$10,$C1,$0F
	.dc.b	$2B,$C1,$A6,$CA,$6C,$FA,$32,$8E,$11,$86,$BF,$A0,$6B,$FE,$BF,$6C
	.dc.b	$36,$F9,$E4,$9F,$AC,$D7,$E7,$B5,$F9,$6D,$7C,$6D,$7F,$5E,$D7,$FD
	.dc.b	$9C,$D8,$F0,$47,$62,$1D,$8A,$76,$29,$D8,$A7,$EF,$36,$FC,$39,$1F
	.dc.b	$C7,$35,$FC,$3B,$5F,$C1,$B5,$F9,$6D,$7A,$1C,$88,$72,$29,$C8,$A7
	.dc.b	$22,$9C,$8E,$04,$38,$1F,$9A,$FD,$F9,$F6,$2F,$C4,$3F,$5B,$A7,$EC
	.dc.b	$30,$56,$A7,$3A,$7F,$BE,$63,$18,$7F,$03,$8C,$61,$CE,$31,$7C,$A1
	.dc.b	$F8,$87,$28,$C3,$6F,$9D,$31,$F1,$46,$1B,$7D,$67,$58,$78,$31,$E5
	.dc.b	$39,$AD,$C6,$30,$97,$E4,$C3,$A6,$1C,$63,$18,$71,$8C,$47,$6C,$7D
	.dc.b	$58,$76,$39,$C6,$6F,$E5,$E1,$98,$E7,$35,$D6,$1D,$6F,$9C,$68,$F0
	.dc.b	$D7,$4C,$7D,$31,$93,$8D,$F3,$8C,$C6,$0E,$24,$30,$47,$64,$C7,$DA
	.dc.b	$DB,$FA,$73,$C6,$31,$8F,$C6,$97,$17,$F2,$62,$57,$F2,$D6,$BF,$C9
	.dc.b	$9A,$FF,$22,$6B,$FC,$29,$AF,$F6,$66,$A8,$EE,$F5,$21,$D4,$A7,$52
	.dc.b	$9D,$4A,$75,$29,$D7,$8B,$F9,$F9,$1F,$E8,$CD,$7F,$9F,$B5,$FE,$76
	.dc.b	$D7,$F9,$1B,$5F,$EF,$ED,$7B,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A
	.dc.b	$7A,$63,$E3,$3F,$AE,$8F,$91,$1F,$AD,$6C,$66,$36,$FE,$6E,$3C,$AD
	.dc.b	$84,$71,$77,$7F,$C0,$E3,$9C,$EE,$74,$47,$B3,$C6,$24,$60,$87,$12
	.dc.b	$CF,$59,$EF,$87,$B5,$E3,$3E,$5A,$EC,$F1,$21,$F4,$23,$DA,$D7,$EC
	.dc.b	$98,$7A,$1F,$69,$C1,$F0,$0B,$1E,$E8,$E4,$6E,$39,$91,$C1,$0F,$C1
	.dc.b	$BC,$88,$6D,$CB,$5E,$A6,$0E,$46,$0C,$D7,$C5,$8F,$E5,$07,$F8,$7D
	.dc.b	$FE,$6C,$DA,$F6,$CF,$C5,$9C,$BF,$E2,$76,$BF,$29,$E8,$FF,$36,$9E
	.dc.b	$47,$C7,$8C,$C7,$39,$F8,$8F,$F2,$39,$DD,$7E,$12,$B6,$76,$9E,$87
	.dc.b	$86,$BC,$B1,$A7,$9B,$B8,$ED,$5F,$4C,$F7,$BC,$5F,$24,$F2,$23,$82
	.dc.b	$57,$8E,$33,$1E,$C3,$CE,$7E,$38,$ED,$3E,$63,$B9,$F1,$63,$8B,$E1
	.dc.b	$8E,$26,$4F,$01,$87,$F3,$71,$EC,$24,$76,$8F,$6B,$E3,$7D,$93,$E7
	.dc.b	$8D,$57,$C3,$25,$7B,$A7,$53,$C6,$BC,$51,$87,$CC,$74,$8F,$74,$66
	.dc.b	$3D,$A7,$9A,$3B,$CF,$99,$F5,$47,$98,$E6,$76,$7E,$41,$E1,$79,$4F
	.dc.b	$57,$E5,$18,$8D,$BC,$8C,$3E,$27,$F1,$47,$9D,$F9,$87,$57,$F0,$71
	.dc.b	$D5,$D9,$E3,$8E,$55,$F1,$9F,$59,$0E,$4F,$E3,$DC,$47,$8C,$EA,$F9
	.dc.b	$9E,$75,$F0,$DD,$4F,$5A,$FA,$C7,$8D,$F0,$57,$6A,$E1,$5E,$53,$E1
	.dc.b	$9E,$82,$CF,$07,$AC,$7D,$D9,$F1,$46,$A3,$9B,$63,$F3,$A7,$DC,$36
	.dc.b	$F5,$24,$F5,$7D,$71,$F1,$2B,$D1,$5D,$27,$6D,$AF,$8D,$3A,$AF,$CE
	.dc.b	$D7,$99,$AF,$B4,$E2,$E1,$F3,$3F,$0C,$F0,$1E,$69,$F2,$46,$8E,$93
	.dc.b	$F3,$27,$55,$E3,$23,$F0,$0F,$C2,$CF,$8D,$EC,$77,$9E,$4F,$C0,$33
	.dc.b	$1B,$3F,$1C,$7E,$85,$CB,$B3,$C6,$FC,$D3,$0C,$3D,$07,$C1,$AF,$5B
	.dc.b	$F0,$A3,$9B,$F9,$E3,$64,$34,$78,$8F,$D7,$1D,$4C,$9F,$62,$7D,$6F
	.dc.b	$19,$EE,$4A,$FC,$45,$71,$8F,$E5,$52,$3F,$9B,$96,$3F,$B1,$35,$FF
	.dc.b	$6F,$91,$FF,$73,$35,$FF,$6F,$B5,$FF,$6D,$B5,$FF,$61,$B5,$E8,$72
	.dc.b	$21,$C8,$A7,$22,$9C,$8A,$72,$38,$1F,$57,$87,$FA,$19,$EF,$AF,$F0
	.dc.b	$EB,$88,$F8,$D1,$87,$FA,$8C,$6E,$7F,$C4,$EE,$E3,$E7,$9A,$21,$83
	.dc.b	$F8,$AB,$BA,$FA,$13,$CA,$77,$18,$8F,$A1,$3B,$AF,$CC,$C6,$2B,$71
	.dc.b	$85,$3B,$9F,$D5,$47,$80,$DB,$8A,$FE,$F7,$3B,$8F,$CE,$B6,$7F,$3C
	.dc.b	$D7,$E4,$1C,$A3,$6D,$36,$53,$67,$84,$F8,$53,$E0,$74,$7E,$E6,$7B
	.dc.b	$34,$C9,$F9,$E7,$2D,$33,$1F,$C8,$63,$2D,$8C,$D7,$F3,$19,$CB,$63
	.dc.b	$35,$FC,$52,$72,$D8,$CB,$FD,$B6,$32,$D8,$C9,$FD,$9A,$32,$D3,$27
	.dc.b	$CB,$7F,$11,$1F,$66,$3E,$25,$7D,$F7,$EC,$9F,$2E,$78,$3C,$8C,$10
	.dc.b	$E3,$5F,$95,$7E,$6B,$67,$E7,$46,$AB,$F1,$CD,$7C,$51,$E1,$AE,$F3
	.dc.b	$93,$C6,$53,$ED,$C7,$42,$57,$72,$57,$68,$E9,$3E,$77,$71,$99,$DB
	.dc.b	$F8,$A9,$F0,$57,$8A,$B7,$39,$6C,$F9,$1B,$3F,$42,$38,$57,$47,$4D
	.dc.b	$8F,$A3,$18,$AF,$11,$0F,$87,$3B,$8F,$E0,$31,$97,$83,$A3,$13,$E1
	.dc.b	$8F,$74,$F1,$7D,$67,$29,$E0,$47,$99,$27,$93,$C0,$95,$EE,$79,$1B
	.dc.b	$3B,$CF,$89,$CC,$FD,$A7,$F9,$4B,$FC,$9C,$8F,$F2,$76,$BF,$C9,$DA
	.dc.b	$F0,$34,$43,$45,$34,$60,$86,$08,$7E,$54,$B1,$F9,$26,$BF,$A3,$69
	.dc.b	$ED,$3F,$C7,$73,$FE,$78,$3E,$EC,$7F,$B5,$9C,$47,$C2,$8F,$EA,$D1
	.dc.b	$88,$F9,$35,$FE,$D8,$8C,$1F,$2E,$3A,$9F,$E8,$C7,$AC,$F2,$8C,$1F
	.dc.b	$E9,$BE,$2F,$F2,$B9,$23,$FD,$52,$E3,$0A,$62,$24,$F1,$6B,$FE,$9F
	.dc.b	$9E,$4F,$59,$E2,$EC,$86,$E7,$8D,$7F,$AD,$AB,$8C,$F3,$AE,$2E,$CE
	.dc.b	$33,$BA,$FE,$E3,$5F,$ED,$64,$D2,$3A,$24,$FF,$76,$8C,$B6,$32,$FF
	.dc.b	$78,$8C,$B6,$32,$FF,$81,$63,$2D,$8C,$A6,$46,$A3,$11,$E6,$36,$68
	.dc.b	$F0,$CF,$B6,$3F,$DA,$1C,$5F,$EA,$D3,$FD,$5D,$5F,$1A,$7F,$DB,$B1
	.dc.b	$99,$F9,$78,$76,$AF,$F4,$9C,$E6,$7E,$4C,$E6,$FF,$DC,$F3,$99,$F9
	.dc.b	$C4,$BE,$AE,$8E,$B3,$E7,$7C,$A6,$4E,$AE,$5F,$F7,$14,$F5,$9F,$64
	.dc.b	$75,$AF,$87,$39,$7C,$91,$CA,$93,$97,$A5,$72,$3C,$B3,$E2,$38,$5F
	.dc.b	$C0,$7C,$87,$79,$E3,$7E,$08,$CD,$73,$8F,$0C,$7A,$0C,$B8,$8D,$4F
	.dc.b	$DD,$BF,$23,$F6,$EB,$D1,$18,$BF,$AD,$7D,$E7,$CC,$78,$EB,$83,$DC
	.dc.b	$CD,$F6,$8C,$BE,$E8,$D1,$C6,$BD,$CF,$0A,$F3,$12,$BC,$D1,$EA,$7E
	.dc.b	$09,$0F,$9A,$E5,$EF,$1C,$5C,$C7,$57,$84,$F8,$23,$C7,$5E,$78,$C3
	.dc.b	$E7,$31,$3D,$6F,$F3,$2F,$E2,$8E,$B5,$CA,$FC,$17,$C9,$F9,$11,$EF
	.dc.b	$9E,$4F,$19,$D4,$7B,$A3,$F0,$2E,$9F,$43,$DA,$FC,$95,$D1,$F2,$CE
	.dc.b	$AF,$ED,$91,$ED,$5C,$DF,$33,$DD,$AF,$5A,$F4,$BB,$AE,$EE,$DA,$EC
	.dc.b	$F9,$66,$E3,$C8,$47,$83,$5F,$AA,$7D,$9A,$F3,$57,$A4,$FB,$57,$F9
	.dc.b	$09,$E2,$43,$C0,$49,$FC,$B3,$F4,$DE,$73,$EC,$72,$FE,$70,$F3,$47
	.dc.b	$38,$CC,$7E,$6D,$E3,$7E,$E7,$2F,$E6,$DC,$DF,$AD,$CB,$F8,$F7,$35
	.dc.b	$D6,$BD,$6F,$95,$F0,$BE,$97,$27,$94,$F8,$44,$3E,$AD,$FC,$38,$E1
	.dc.b	$19,$BF,$33,$F7,$8C,$5F,$95,$EF,$5C,$6F,$CB,$7C,$A7,$AB,$E5,$78
	.dc.b	$BE,$97,$F2,$33,$C5,$FC,$51,$FE,$CD,$9F,$F6,$59,$1F,$F6,$5B,$5F
	.dc.b	$F6,$5B,$5F,$F6,$5B,$5E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14
	.dc.b	$D1,$82,$5F,$DD,$C7,$FA,$49,$F5,$27,$FC,$96,$49,$FE,$01,$1A,$7E
	.dc.b	$2D,$FF,$3C,$8C,$47,$E3,$CB,$5F,$D7,$23,$A1,$8C,$3E,$A4,$FF,$64
	.dc.b	$8C,$57,$CF,$8F,$E9,$F1,$88,$FA,$D5,$CE,$78,$1E,$2B,$F3,$D7,$68
	.dc.b	$C9,$E3,$2C,$F1,$70,$76,$33,$7E,$47,$04,$8F,$C6,$BA,$BF,$E9,$15
	.dc.b	$DE,$35,$5C,$6B,$B1,$1C,$1F,$D9,$A7,$D9,$7F,$C5,$9B,$5F,$38,$ED
	.dc.b	$7F,$02,$B8,$46,$2B,$83,$4E,$07,$8A,$7E,$55,$F8,$5F,$3C,$7C,$7B
	.dc.b	$EE,$FF,$27,$6B,$FC,$1D,$B3,$F8,$58,$EC,$6C,$A6,$CA,$ED,$F1,$4F
	.dc.b	$A4,$F0,$B8,$AF,$4B,$4F,$4C,$F9,$E7,$81,$F8,$D8,$EB,$7F,$C3,$EF
	.dc.b	$E4,$92,$33,$3F,$C6,$E3,$11,$DD,$F5,$CF,$18,$F9,$51,$AA,$F0,$5F
	.dc.b	$91,$B1,$9A,$FA,$35,$F8,$89,$CB,$C8,$86,$63,$71,$A8,$F1,$1E,$6B
	.dc.b	$FE,$71,$5D,$5F,$8B,$3B,$3F,$1F,$3C,$08,$E6,$BF,$17,$3C,$1E,$E4
	.dc.b	$70,$43,$E5,$4F,$95,$C5,$FE,$A2,$3D,$35,$F0,$9F,$A7,$5C,$EB,$8C
	.dc.b	$F8,$5E,$33,$A7,$9D,$FA,$DD,$D7,$99,$A6,$A7,$BC,$75,$BF,$A0,$DA
	.dc.b	$E5,$1F,$C4,$67,$F2,$07,$C2,$9D,$3F,$2E,$3F,$CD,$93,$F8,$08,$FE
	.dc.b	$68,$FF,$32,$23,$FC,$C9,$AF,$F3,$26,$BC,$0D,$10,$D1,$4D,$18,$21
	.dc.b	$82,$1F,$88,$2D,$7E,$89,$AF,$D4,$6C,$FD,$37,$FC,$DF,$7F,$E7,$C3
	.dc.b	$E1,$95,$F9,$33,$FE,$59,$8C,$47,$C5,$8C,$3F,$E9,$08,$E2,$D3,$8C
	.dc.b	$7F,$9D,$23,$88,$BF,$F4,$EF,$12,$A1,$2B,$FD,$49,$18,$8C,$42,$31
	.dc.b	$1C,$63,$6E,$1A,$F1,$6B,$CA,$38,$BE,$23,$84,$75,$6C,$6A,$38,$C7
	.dc.b	$77,$4F,$56,$BD,$4E,$24,$35,$1C,$E3,$71,$CE,$30,$47,$81,$CC,$E2
	.dc.b	$79,$B8,$B9,$11,$C1,$0F,$23,$96,$99,$3E,$5B,$96,$99,$3D,$CE,$5A
	.dc.b	$64,$FF,$6A,$B9,$69,$99,$FF,$6E,$46,$5B,$1F,$3B,$8B,$E1,$11,$C1
	.dc.b	$0F,$98,$E5,$A6,$4F,$C1,$B9,$69,$93,$EB,$39,$69,$93,$FD,$DA,$E5
	.dc.b	$A6,$63,$FD,$E9,$19,$6C,$7E,$1B,$8B,$ED,$11,$C1,$0F,$C2,$39,$69
	.dc.b	$93,$F3,$6E,$5A,$64,$FC,$73,$96,$99,$11,$23,$51,$88,$46,$1D,$47
	.dc.b	$16,$BF,$99,$DB,$F2,$84,$7F,$46,$D7,$46,$08,$60,$87,$ED,$9E,$0D
	.dc.b	$36,$53,$67,$EA,$DD,$B4,$D9,$4D,$9E,$D7,$6F,$BF,$6F,$D6,$11,$FD
	.dc.b	$DB,$5F,$DB,$35,$FE,$16,$D7,$F7,$8D,$7A,$1C,$88,$72,$29,$C8,$A7
	.dc.b	$22,$9C,$8E,$04,$38,$1E,$49,$FE,$43,$9F,$F2,$57,$EE,$CF,$DE,$9F
	.dc.b	$E4,$D3,$99,$4E,$65,$39,$8C,$35,$EB,$9F,$C7,$6D,$76,$29,$D0,$AF
	.dc.b	$28,$F9,$D1,$82,$57,$92,$3A,$36,$3D,$33,$A8,$F2,$CF,$27,$53,$DD
	.dc.b	$B3,$DE,$3C,$91,$87,$C7,$12,$30,$43,$AC,$70,$7E,$EB,$F8,$22,$3F
	.dc.b	$3C,$86,$8A,$7E,$09,$DB,$92,$19,$36,$7A,$CC,$94,$D1,$87,$C8,$E0
	.dc.b	$86,$C8,$FA,$6B,$ED,$3F,$50,$FE,$C4,$FE,$05,$B5,$E1,$8D,$CF,$CC
	.dc.b	$9C,$9F,$C0,$DF,$A5,$1A,$7E,$3C,$7E,$25,$AF,$03,$EA,$3A,$AF,$A9
	.dc.b	$1B,$9F,$14,$F4,$AF,$11,$FC,$46,$79,$1C,$EB,$89,$C9,$CC,$F1,$7E
	.dc.b	$04,$F1,$8F,$80,$66,$3C,$75,$E0,$8F,$71,$9A,$E8,$F0,$3F,$09,$1E
	.dc.b	$F3,$83,$EE,$38,$4E,$DC,$1C,$2B,$D7,$1F,$26,$3B,$99,$7C,$E6,$5E
	.dc.b	$27,$C7,$7D,$07,$B9,$E0,$FA,$A3,$DD,$1D,$0E,$06,$A3,$83,$82,$4F
	.dc.b	$A8,$93,$E6,$9E,$87,$8E,$32,$F9,$8C,$9F,$66,$3C,$33,$D1,$E5,$1E
	.dc.b	$09,$F9,$11,$89,$F8,$D1,$89,$E8,$6E,$7C,$55,$F4,$AB,$D9,$5B,$AF
	.dc.b	$09,$B3,$C9,$3B,$9F,$93,$3B,$8C,$3E,$A9,$F4,$C6,$2B,$D3,$3B,$9F
	.dc.b	$2C,$60,$F1,$B8,$3E,$23,$8A,$F2,$BB,$8C,$90,$F8,$C5,$AF,$CC,$BC
	.dc.b	$CC,$90,$F6,$18,$77,$3D,$A3,$A1,$A3,$71,$A8,$F8,$33,$E3,$9E,$91
	.dc.b	$DD,$FD,$51,$F8,$C7,$04,$3B,$3F,$18,$C3,$D1,$F0,$BE,$99,$F0,$57
	.dc.b	$CD,$9F,$24,$F0,$8F,$98,$F4,$8C,$B4,$FD,$49,$48,$77,$7E,$11,$E7
	.dc.b	$73,$1D,$DF,$14,$65,$E0,$F6,$79,$4F,$9A,$BE,$24,$E6,$B9,$57,$EB
	.dc.b	$2B,$F0,$93,$EE,$8E,$F5,$F2,$E7,$94,$E6,$7B,$9E,$F9,$FC,$23,$83
	.dc.b	$8C,$6E,$BE,$1C,$64,$8F,$89,$FD,$54,$7E,$91,$E8,$F1,$9F,$13,$EF
	.dc.b	$9E,$04,$8C,$CF,$F0,$13,$EC,$9F,$E8,$F9,$FF,$44,$11,$FF,$4B,$B5
	.dc.b	$FF,$49,$35,$FF,$53,$B5,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1
	.dc.b	$4D,$18,$24,$FE,$37,$5F,$EB,$47,$C4,$9F,$F2,$CB,$89,$F9,$F1,$87
	.dc.b	$FA,$D4,$6E,$0E,$E3,$E8,$D7,$F4,$F8,$C5,$7C,$BA,$FF,$31,$46,$27
	.dc.b	$F1,$35,$67,$2F,$81,$EA,$F8,$0C,$4E,$E3,$F6,$71,$93,$FB,$3C,$79
	.dc.b	$4D,$47,$EF,$5B,$1F,$B9,$6B,$FA,$B6,$BF,$A0,$6D,$7C,$78,$EC,$FA
	.dc.b	$E7,$E2,$47,$80,$F2,$9A,$AF,$6B,$63,$DA,$FC,$66,$CF,$F8,$26,$BD
	.dc.b	$13,$B9,$F6,$BD,$49,$3E,$18,$F8,$B1,$8A,$EE,$D8,$EB,$18,$9E,$D3
	.dc.b	$D6,$7E,$23,$4F,$88,$53,$E2,$14,$F8,$85,$8F,$88,$F7,$8F,$3C,$76
	.dc.b	$8C,$47,$66,$9D,$8A,$76,$29,$D8,$B1,$D8,$E8,$49,$FC,$74,$7A,$EB
	.dc.b	$E4,$BE,$07,$F0,$27,$DA,$8F,$E2,$EE,$23,$81,$EB,$8C,$B8,$39,$19
	.dc.b	$AF,$E8,$F3,$C6,$32,$79,$A7,$8B,$8A,$F3,$57,$D2,$21,$F1,$A7,$D3
	.dc.b	$3B,$6C,$F0,$9F,$C6,$BF,$7E,$BC,$0E,$63,$F0,$2D,$AF,$98,$E6,$77
	.dc.b	$5D,$0F,$C1,$16,$BC,$67,$27,$0F,$81,$AF,$D9,$9E,$95,$F1,$5F,$03
	.dc.b	$F6,$5C,$BE,$B8,$EA,$F9,$27,$9D,$79,$27,$33,$C6,$72,$F5,$7C,$2F
	.dc.b	$18,$F3,$CF,$17,$CC,$7F,$3B,$33,$1F,$C0,$CF,$E6,$D1,$FC,$D0,$8F
	.dc.b	$F3,$46,$BF,$CD,$1A,$F0,$34,$43,$45,$34,$60,$86,$08,$7D,$E2,$C7
	.dc.b	$F0,$26,$BF,$C2,$9A,$FD,$33,$EA,$D7,$FA,$3C,$F8,$0F,$FB,$32,$3F
	.dc.b	$D1,$91,$83,$4F,$C9,$81,$23,$31,$F5,$63,$0F,$23,$83,$87,$FD,$23
	.dc.b	$1F,$E9,$FD,$A4,$7F,$A6,$DA,$FF,$A5,$BF,$24,$FF,$52,$18,$9F,$F5
	.dc.b	$24,$DC,$3F,$D2,$4D,$BF,$28,$78,$DF,$FA,$9E,$F5,$3E,$A9,$AE,$E7
	.dc.b	$C4,$43,$C4,$71,$31,$7E,$27,$47,$F7,$99,$FE,$EE,$47,$FB,$BB,$5F
	.dc.b	$EE,$ED,$78,$1A,$21,$A2,$9A,$30,$43,$04,$3E,$19,$C0,$A6,$CA,$6E
	.dc.b	$7E,$3C,$7C,$2D,$FF,$DA,$47,$C4,$6C,$FF,$B3,$E3,$0D,$21,$FE,$CC
	.dc.b	$8E,$4E,$2B,$E3,$39,$31,$87,$FB,$26,$B9,$C7,$FB,$51,$EE,$EC,$93
	.dc.b	$FE,$E3,$8C,$BD,$09,$3C,$E7,$AD,$FB,$1F,$AF,$19,$7B,$18,$8D,$38
	.dc.b	$24,$F4,$8C,$BE,$18,$EB,$1F,$89,$79,$3F,$A4,$AF,$2D,$7D,$0A,$E5
	.dc.b	$39,$AE,$2F,$E3,$DF,$6C,$F9,$67,$B4,$62,$7C,$33,$92,$38,$21,$D6
	.dc.b	$3D,$31,$E1,$24,$7E,$92,$BC,$AF,$03,$C5,$5E,$02,$4F,$08,$F7,$57
	.dc.b	$47,$93,$F0,$1E,$31,$E4,$7C,$53,$F0,$1C,$9C,$27,$93,$D2,$35,$5E
	.dc.b	$48,$C5,$78,$54,$65,$E6,$68,$E9,$3E,$87,$D9,$7B,$9F,$C6,$CE,$67
	.dc.b	$CB,$5D,$6B,$57,$EB,$6B,$F4,$67,$B1,$C0,$EF,$3C,$63,$ED,$DF,$0B
	.dc.b	$ED,$5E,$89,$E8,$68,$87,$C2,$74,$F7,$7D,$A6,$23,$9C,$F5,$AF,$9D
	.dc.b	$3A,$BC,$3E,$42,$1E,$72,$1B,$BF,$0D,$F4,$6D,$7A,$DF,$25,$F4,$AE
	.dc.b	$2F,$CA,$9F,$0B,$98,$F3,$47,$92,$FF,$80,$D7,$23,$AC,$F9,$0F,$13
	.dc.b	$92,$5F,$8D,$F8,$53,$D6,$E4,$7C,$A3,$77,$E2,$7A,$DF,$D6,$3E,$4C
	.dc.b	$F0,$8D,$C7,$E1,$1D,$BE,$83,$C9,$1F,$8D,$AF,$4D,$F8,$C9,$7E,$DB
	.dc.b	$FD,$05,$75,$9C,$DF,$59,$EE,$FC,$87,$BD,$F4,$BF,$44,$6A,$7C,$47
	.dc.b	$A6,$7A,$4E,$AB,$F1,$71,$C6,$7C,$F5,$FE,$C2,$BF,$F6,$29,$1C,$10
	.dc.b	$FF,$62,$B9,$69,$93,$FD,$A0,$E5,$A6,$4F,$E5,$06,$0E,$24,$30,$43
	.dc.b	$64,$8F,$E0,$0E,$5F,$4A,$A5,$FF,$78,$97,$E4,$AA,$95,$A9,$FE,$FB
	.dc.b	$3F,$91,$35,$3C,$EB,$94,$66,$38,$C6,$1F,$C0,$C7,$0A,$E9,$5F,$E1
	.dc.b	$C8,$E1,$3C,$E3,$BB,$C0,$E6,$E1,$FB,$51,$CE,$BB,$46,$0D,$CF,$57
	.dc.b	$91,$D5,$C3,$B3,$0F,$C2,$77,$5E,$68,$C4,$6E,$31,$5E,$86,$C7,$0A
	.dc.b	$F1,$C7,$E5,$AB,$BB,$A9,$E5,$3A,$3F,$20,$59,$F7,$4F,$17,$C2,$F1
	.dc.b	$31,$5F,$06,$37,$3C,$9A,$43,$F4,$95,$EE,$9E,$11,$E4,$9E,$0D,$38
	.dc.b	$1F,$A6,$24,$70,$7C,$71,$C1,$A7,$0B,$FE,$91,$3B,$3E,$34,$FB,$67
	.dc.b	$C7,$7E,$B8,$C4,$77,$AE,$17,$F6,$A3,$A9,$C2,$77,$7D,$23,$C1,$3F
	.dc.b	$DF,$A3,$8B,$F3,$A3,$2F,$F4,$18,$CD,$FC,$A8,$ED,$1C,$E7,$E1,$1D
	.dc.b	$CF,$86,$F0,$AF,$0C,$F0,$3F,$01,$3A,$70,$7C,$F2,$D7,$77,$04,$7E
	.dc.b	$14,$71,$8F,$43,$E2,$3E,$09,$2B,$C4,$FE,$F2,$78,$BB,$8C,$B6,$7C
	.dc.b	$AD,$7C,$15,$EA,$7B,$C7,$57,$CF,$19,$AF,$41,$DA,$75,$3C,$9F,$9B
	.dc.b	$1F,$DC,$AB,$C3,$1A,$8E,$11,$F3,$63,$15,$D4,$FC,$0D,$FE,$22,$FC
	.dc.b	$33,$F0,$EF,$8C,$7A,$A3,$89,$9A,$E3,$1F,$81,$38,$C7,$BE,$78,$C7
	.dc.b	$BA,$FC,$A4,$7A,$35,$F6,$4F,$D1,$BF,$24,$62,$3F,$81,$BF,$C9,$DF
	.dc.b	$E2,$8D,$7F,$86,$B5,$DB,$FC,$58,$C4,$FF,$2A,$69,$63,$9B,$B2,$38
	.dc.b	$21,$D9,$CB,$4C,$9E,$02,$C7,$78,$DC,$64,$E0,$43,$07,$CE,$3F,$DA
	.dc.b	$3F,$93,$7F,$01,$3E,$35,$FF,$25,$71,$3F,$42,$7F,$2F,$18,$8F,$91
	.dc.b	$56,$B8,$46,$E3,$18,$7F,$53,$AE,$AD,$9F,$E0,$51,$FC,$47,$87,$D5
	.dc.b	$87,$FB,$41,$AF,$FB,$99,$A9,$A4,$9F,$9D,$1B,$6B,$B6,$C6,$DF,$EB
	.dc.b	$F1,$DD,$C9,$0C,$BC,$DA,$F3,$6C,$73,$7F,$08,$D7,$FB,$93,$63,$B4
	.dc.b	$70,$3B,$1F,$8B,$E1,$F6,$91,$F4,$B5,$F0,$B5,$E0,$68,$86,$8A,$68
	.dc.b	$C1,$0C,$10,$FF,$04,$3D,$5A,$72,$29,$C4,$A7,$E8,$F8,$7E,$A9,$1F
	.dc.b	$A2,$D7,$E4,$B5,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$FF,$10,$3D,$5A
	.dc.b	$72,$29,$C4,$A7,$EE,$FF,$7D,$FE,$C4,$47,$FA,$B3,$FD,$97,$F7,$D9
	.dc.b	$18,$21,$FE,$0E,$8E,$0D,$33,$A7,$F8,$3E,$78,$C6,$49,$1F,$E1,$39
	.dc.b	$E2,$D8,$E3,$9F,$F8,$EE,$B8,$CE,$67,$8C,$FF,$8F,$EB,$F4,$A4,$8F
	.dc.b	$D0,$B5,$FC,$CB,$5F,$80,$68,$86,$8A,$68,$C1,$0C,$10,$F1,$1C,$0A
	.dc.b	$6C,$A6,$F1,$12,$3C,$D5,$ED,$3F,$C6,$DF,$BF,$7F,$85,$4F,$8B,$7D
	.dc.b	$23,$37,$FE,$0B,$9C,$DF,$41,$7C,$67,$47,$CC,$8D,$46,$63,$FC,$8F
	.dc.b	$3C,$A7,$C9,$1C,$0C,$9C,$5C,$B8,$25,$7A,$6F,$CC,$7F,$9C,$EB,$B4
	.dc.b	$66,$BC,$35,$D6,$3D,$6F,$DC,$8C,$BD,$2B,$93,$E1,$8F,$6B,$C6,$32
	.dc.b	$F1,$8F,$2C,$F4,$24,$F9,$23,$AC,$61,$F9,$15,$F0,$5E,$B5,$E1,$72
	.dc.b	$F8,$6F,$D7,$7E,$0B,$F6,$57,$82,$30,$47,$BB,$E1,$7F,$03,$7C,$9C
	.dc.b	$90,$CD,$FB,$5C,$1E,$4A,$F4,$3A,$8F,$80,$E6,$BC,$0E,$67,$CA,$E4
	.dc.b	$F5,$CF,$B5,$F4,$1D,$AB,$C8,$E0,$81,$CB,$CC,$CB,$8B,$F9,$4D,$9F
	.dc.b	$B0,$D7,$CE,$DA,$F2,$B6,$FE,$35,$F7,$AF,$84,$F8,$AF,$C3,$5F,$5D
	.dc.b	$B3,$F5,$4D,$CF,$17,$E2,$C6,$9A,$6A,$79,$34,$E4,$7C,$59,$E5,$7F
	.dc.b	$2A,$B9,$36,$78,$1D,$0E,$15,$D1,$A7,$48,$FB,$35,$D2,$FD,$F7,$D1
	.dc.b	$B5,$D2,$BB,$3C,$8E,$CF,$2B,$EC,$D8,$ED,$3F,$82,$BE,$C7,$E2,$AF
	.dc.b	$B3,$67,$B4,$F8,$1E,$D5,$E0,$3A,$3E,$08,$E9,$7E,$06,$C7,$82,$BF
	.dc.b	$1B,$5E,$EB,$FC,$1C,$FD,$46,$BF,$35,$B1,$B8,$FC,$39,$B2,$9B,$2B
	.dc.b	$B7,$F1,$66,$CA,$6C,$AE,$CF,$C7,$BB,$6B,$B6,$9B,$AF,$C9,$DF,$E5
	.dc.b	$6F,$F4,$D5,$C6,$72,$D8,$CB,$CE,$32,$D8,$CD,$75,$6C,$FD,$7B,$F8
	.dc.b	$6D,$AF,$7B,$5F,$5B,$6B,$CE,$D3,$BB,$D6,$BF,$6B,$3E,$56,$BE,$36
	.dc.b	$C6,$DF,$DE,$1F,$88,$3F,$85,$4F,$E1,$63,$F0,$67,$E1,$5F,$C0,$CF
	.dc.b	$E1,$67,$F0,$31,$89,$DB,$E5,$8F,$79,$A2,$1D,$8A,$78,$4F,$1D,$7E
	.dc.b	$61,$D1,$0D,$5F,$E4,$DE,$2E,$DC,$DF,$E5,$5E,$11,$93,$EE,$3F,$0A
	.dc.b	$BD,$04,$7C,$75,$F5,$1E,$B5,$FA,$87,$2F,$D6,$72,$FE,$AA,$7E,$04
	.dc.b	$75,$33,$7F,$11,$E3,$3F,$71,$CC,$F8,$5F,$75,$78,$5F,$4C,$F8,$5C
	.dc.b	$BE,$17,$2F,$85,$EB,$5F,$86,$7A,$90,$E3,$1E,$17,$8C,$79,$5C,$BC
	.dc.b	$5C,$D7,$95,$FA,$D5,$E5,$7E,$7D,$78,$5E,$33,$F9,$43,$85,$FD,$5A
	.dc.b	$E0,$FD,$3A,$FC,$B5,$7E,$4A,$31,$1C,$E3,$F0,$6F,$C1,$7B,$DF,$9D
	.dc.b	$CC,$6E,$78,$CF,$E5,$E7,$E0,$1C,$2F,$D5,$7E,$2B,$E9,$7F,$84,$AF
	.dc.b	$14,$60,$F0,$4E,$DE,$C7,$A1,$EF,$5E,$E8,$FC,$84,$F7,$3C,$35,$D6
	.dc.b	$B7,$18,$9D,$CE,$23,$E7,$9F,$CD,$E3,$B9,$2F,$D9,$1A,$21,$DC,$FE
	.dc.b	$D1,$1E,$93,$85,$FE,$EC,$F3,$38,$8E,$D3,$F2,$EB,$C5,$1F,$32,$31
	.dc.b	$3C,$DF,$0C,$77,$24,$FB,$67,$55,$DA,$B7,$3D,$0D,$4F,$86,$BD,$31
	.dc.b	$FB,$B6,$C7,$FB,$72,$3F,$DC,$64,$70,$43,$FD,$AE,$E5,$A6,$41,$FE
	.dc.b	$48,$2C,$7F,$89,$1A,$FF,$8B,$5A,$FF,$82,$DB,$1D,$5A,$75,$3A,$12
	.dc.b	$3A,$35,$E8,$D7,$E4,$72,$79,$21,$4B,$FF,$3E,$52,$BF,$8E,$1B,$9D
	.dc.b	$47,$FA,$06,$33,$5E,$E9,$CC,$FF,$A4,$27,$33,$DA,$BA,$C6,$2B,$75
	.dc.b	$DE,$BF,$5D,$3D,$EB,$71,$83,$B3,$87,$F6,$D1,$B9,$F2,$46,$CF,$03
	.dc.b	$87,$67,$27,$15,$FB,$B9,$E5,$1E,$99,$E5,$3E,$69,$F4,$57,$86,$38
	.dc.b	$4F,$3A,$E0,$F6,$3C,$2E,$23,$71,$88,$F8,$0E,$E3,$B3,$E0,$9E,$C7
	.dc.b	$F2,$1A,$F2,$D7,$0A,$E6,$70,$AE,$55,$C2,$7F,$88,$D7,$03,$D2,$FC
	.dc.b	$08,$E9,$3E,$98,$E9,$18,$8F,$34,$F0,$9F,$54,$E6,$BC,$07,$89,$F0
	.dc.b	$47,$F0,$49,$F0,$57,$AE,$3C,$07,$91,$F0,$47,$58,$F0,$99,$8F,$09
	.dc.b	$C5,$F0,$99,$AF,$0D,$78,$CF,$0C,$7E,$7A,$7A,$9C,$5E,$A7,$95,$C9
	.dc.b	$F2,$1F,$29,$97,$C2,$65,$F0,$99,$9F,$0B,$E3,$9F,$0D,$FF,$A6,$2F
	.dc.b	$E0,$DF,$1A,$F0,$9E,$57,$AD,$FC,$98,$F0,$99,$7A,$BD,$EF,$AC,$FF
	.dc.b	$A4,$2B,$AD,$71,$AF,$29,$C6,$B7,$5F,$E8,$DA,$DC,$7C,$79,$EF,$7E
	.dc.b	$7A,$F8,$77,$CE,$7C,$F3,$99,$F3,$D7,$19,$EF,$39,$AF,$3C,$78,$4E
	.dc.b	$B5,$E7,$3A,$D7,$18,$EB,$19,$3A,$BF,$2C,$E2,$FA,$4E,$2E,$4F,$0B
	.dc.b	$C6,$7C,$35,$E3,$AF,$83,$1F,$EA,$2B,$F8,$37,$C4,$F0,$DF,$53,$C2
	.dc.b	$FC,$F3,$C3,$3C,$63,$C3,$19,$3C,$2F,$18,$F0,$BE,$3A,$F0,$9F,$EA
	.dc.b	$5A,$F0,$D7,$D6,$6B,$E2,$AE,$D5,$E2,$8F,$C3,$CE,$8F,$D3,$16,$3C
	.dc.b	$71,$C6,$BB,$46,$27,$4F,$3B,$F8,$95,$F8,$E9,$CD,$FA,$2B,$D5,$7F
	.dc.b	$AA,$9C,$BE,$AA,$FF,$33,$C6,$1E,$53,$E2,$BE,$B1,$ED,$8F,$11,$93
	.dc.b	$E2,$3C,$1F,$74,$FC,$29,$E3,$3F,$02,$78,$9E,$93,$8B,$D5,$E2,$75
	.dc.b	$3D,$2F,$5B,$FC,$9D,$75,$9E,$33,$D5,$E2,$75,$38,$BD,$67,$8C,$75
	.dc.b	$7E,$A9,$87,$8B,$F2,$4C,$DF,$C9,$8F,$4C,$66,$3E,$AC,$70,$74,$F8
	.dc.b	$63,$8C,$FA,$67,$31,$C6,$3E,$8C,$F4,$9C,$9F,$56,$3F,$D2,$D5,$F5
	.dc.b	$5F,$54,$65,$FF,$4F,$46,$5F,$85,$5D,$5E,$35,$D5,$F1,$1F,$A2,$AF
	.dc.b	$4C,$FC,$12,$3D,$6F,$A5,$CB,$FE,$60,$D7,$DD,$3C,$8F,$4D,$F6,$3F
	.dc.b	$57,$7F,$09,$F4,$99,$3F,$3C,$7C,$29,$F9,$4E,$6B,$F4,$B1,$E9,$8E
	.dc.b	$27,$ED,$CF,$7D,$78,$1E,$EF,$8D,$E5,$1C,$5F,$E3,$47,$23,$89,$F0
	.dc.b	$CF,$15,$7C,$27,$8C,$7F,$3F,$AF,$81,$3F,$36,$7E,$8C,$F5,$9F,$A5
	.dc.b	$7F,$A0,$77,$7F,$2E,$74,$43,$A1,$6F,$A5,$70,$BF,$0C,$7F,$39,$BC
	.dc.b	$47,$B8,$F3,$DF,$C1,$7D,$66,$4F,$C0,$DF,$CC,$7E,$04,$FC,$C8,$FE
	.dc.b	$C6,$7F,$95,$9F,$F2,$53,$4B,$1F,$E2,$F7,$07,$F9,$3D,$C3,$FE,$39
	.dc.b	$32,$43,$04,$76,$43,$FC,$5A,$ED,$C1,$FE,$2A,$70,$E9,$E4,$D3,$71
	.dc.b	$E8,$8C,$C7,$C2,$1C,$BF,$E2,$E3,$F4,$C5,$8F,$C2,$35,$F8,$D1,$C1
	.dc.b	$FF,$15,$47,$06,$9C,$0A,$70,$79,$5F,$E2,$64,$73,$9F,$F2,$34,$F3
	.dc.b	$69,$CC,$A7,$33,$AC,$F3,$9E,$4F,$F9,$BA,$79,$36,$79,$35,$E4,$74
	.dc.b	$79,$47,$9A,$FF,$10,$DA,$FD,$A3,$5F,$D3,$35,$F9,$8D,$BF,$EB,$ED
	.dc.b	$BF,$EB,$2D,$7F,$B0,$36,$FF,$C1,$6D,$BF,$F1,$EB,$6F,$FC,$4E,$D7
	.dc.b	$FC,$42,$D7,$FC,$5E,$DB,$F2,$3F,$0A,$7F,$D3,$ED,$9F,$A2,$7C,$88
	.dc.b	$F7,$DF,$CC,$6B,$F1,$5B,$7F,$24,$F8,$F1,$B9,$F8,$AD,$7E,$2B,$5F
	.dc.b	$8A,$DB,$F8,$AD,$3E,$29,$4F,$8A,$59,$F8,$AD,$AF,$8A,$DA,$F8,$AD
	.dc.b	$7E,$2B,$5F,$8A,$DA,$F8,$AF,$E1,$27,$E2,$B6,$7E,$2D,$FC,$77,$F2
	.dc.b	$ED,$7E,$41,$F0,$23,$E3,$36,$FE,$35,$7C,$83,$AC,$FE,$42,$33,$3B
	.dc.b	$9C,$C7,$CC,$9E,$73,$F2,$9A,$FC,$A6,$DF,$CA,$69,$F2,$8A,$7C,$A2
	.dc.b	$CF,$CA,$6D,$7C,$A6,$D7,$CA,$6B,$F2,$9A,$FC,$A6,$D7,$CA,$7D,$93
	.dc.b	$EA,$3E,$04,$FC,$A8,$F9,$8F,$CA,$6B,$F2,$8F,$75,$7C,$46,$CF,$37
	.dc.b	$E3,$1C,$E7,$BB,$67,$E4,$1C,$5F,$8C,$DA,$F8,$CD,$9F,$8C,$D7,$E3
	.dc.b	$34,$F8,$C5,$AF,$8C,$DA,$F8,$CF,$7B,$F8,$4E,$63,$E4,$39,$AF,$D1
	.dc.b	$4F,$ED,$5A,$43,$EE,$BF,$8D,$39,$D7,$C1,$AF,$91,$3F,$01,$B5,$D6
	.dc.b	$38,$9F,$32,$B8,$CF,$C5,$AE,$4F,$C4,$6C,$6A,$39,$C7,$D8,$69,$C4
	.dc.b	$86,$08,$73,$32,$7D,$B2,$47,$BE,$7C,$A4,$AF,$41,$F6,$2B,$C7,$1F
	.dc.b	$0E,$3E,$55,$7E,$3D,$DC,$EA,$7F,$22,$6D,$C4,$6A,$7F,$0E,$79,$AB
	.dc.b	$DC,$F9,$6F,$F0,$75,$C0,$91,$EA,$8C,$BE,$C9,$E2,$7D,$5B,$F0,$46
	.dc.b	$5F,$A0,$61,$F0,$38,$3E,$05,$76,$8F,$CE,$9C,$63,$2F,$85,$FC,$AC
	.dc.b	$6E,$3E,$5C,$7F,$19,$3B,$38,$78,$BD,$0F,$0B,$87,$E0,$3D,$1F,$05
	.dc.b	$FC,$77,$ED,$DF,$C0,$78,$C6,$5E,$33,$FA,$6A,$F2,$5F,$CC,$BE,$15
	.dc.b	$E3,$BE,$13,$F1,$EF,$C4,$FE,$0E,$33,$1E,$A8,$D4,$62,$3E,$E9,$C0
	.dc.b	$D1,$F2,$DD,$1F,$86,$AF,$AB,$1A,$3F,$4D,$5E,$3B,$F3,$C7,$16,$BD
	.dc.b	$AB,$9C,$61,$FC,$A4,$FC,$68,$C1,$FA,$57,$A1,$93,$EC,$4F,$C5,$70
	.dc.b	$7C,$A8,$FD,$8D,$7C,$6B,$FF,$16,$50,$FE,$8A,$5A,$FE,$3C,$D7,$F8
	.dc.b	$0B,$5F,$F1,$43,$63,$FC,$7D,$1C,$23,$99,$1C,$10,$E6,$E5,$A6,$4E
	.dc.b	$AE,$5A,$66,$3B,$BC,$08,$68,$AF,$F8,$97,$F7,$FF,$74,$B0,$61,$FE
	.dc.b	$12,$B5,$7E,$D8,$DC,$EA,$3F,$C2,$F1,$9A,$F9,$13,$99,$FF,$25,$CE
	.dc.b	$67,$B5,$75,$8C,$56,$EB,$BD,$7E,$AA,$7B,$D6,$E3,$07,$67,$0F,$EA
	.dc.b	$E3,$73,$E4,$8D,$9E,$07,$0E,$CE,$4E,$2B,$F5,$93,$CA,$3D,$33,$CA
	.dc.b	$7C,$D3,$E8,$AF,$0C,$70,$9E,$75,$C1,$EC,$78,$5C,$46,$E3,$11,$F0
	.dc.b	$1D,$C7,$67,$C1,$3D,$8F,$E2,$15,$E5,$AE,$15,$CC,$E1,$5C,$AB,$84
	.dc.b	$FE,$80,$DB,$87,$F4,$47,$C1,$70,$FE,$A8,$F3,$B8,$BF,$C4,$4F,$BA
	.dc.b	$79,$9D,$6F,$C8,$7C,$57,$A1,$98,$E1,$19,$8E,$A6,$63,$E1,$C6,$63
	.dc.b	$A9,$E6,$36,$43,$CC,$68,$93,$E6,$9E,$67,$CA,$8F,$19,$CC,$E1,$3D
	.dc.b	$C9,$1B,$8E,$D3,$CE,$FE,$69,$F3,$EF,$E2,$57,$F9,$AE,$FE,$0B,$C6
	.dc.b	$FC,$6E,$4F,$82,$66,$3B,$BF,$5E,$3C,$A7,$B2,$39,$C6,$63,$C6,$F3
	.dc.b	$3D,$A6,$5E,$EF,$F9,$52,$38,$D7,$9E,$7A,$DF,$AE,$FD,$37,$B3,$CA
	.dc.b	$73,$BF,$2D,$F1,$BF,$0C,$EC,$F0,$C6,$63,$C2,$FF,$8F,$E7,$C3,$5F
	.dc.b	$98,$AE,$B3,$B9,$EB,$39,$9E,$6D,$9F,$9E,$D3,$CE,$F5,$8E,$71,$9A
	.dc.b	$F6,$C7,$C5,$9E,$31,$F8,$D6,$BF,$81,$AF,$2B,$9A,$EA,$7F,$8D,$23
	.dc.b	$31,$BA,$EE,$EC,$EE,$FC,$3A,$EB,$1F,$66,$36,$FE,$5E,$79,$BF,$1A
	.dc.b	$BB,$4E,$A3,$F2,$6D,$9F,$8B,$1F,$84,$78,$98,$7A,$46,$CF,$C6,$14
	.dc.b	$F9,$6F,$9A,$BE,$44,$62,$38,$B5,$F2,$1F,$E4,$89,$EF,$1C,$E7,$DB
	.dc.b	$7F,$89,$AF,$24,$E6,$FE,$4C,$71,$8C,$3A,$7B,$4F,$18,$F0,$D7,$C8
	.dc.b	$BF,$8B,$39,$9F,$87,$5E,$59,$CB,$F7,$6F,$F8,$BC,$79,$A3,$A9,$99
	.dc.b	$F0,$5F,$58,$F9,$D7,$D6,$B8,$4E,$67,$F4,$07,$A2,$BC,$B7,$F8,$D6
	.dc.b	$9D,$1F,$7B,$FE,$70,$AE,$71,$E4,$9E,$E7,$8E,$78,$90,$F9,$E7,$17
	.dc.b	$C0,$FC,$0A,$F9,$F1,$C6,$31,$3D,$63,$DE,$7E,$45,$C4,$FA,$CE,$31
	.dc.b	$99,$F4,$10,$FD,$DC,$7E,$10,$E1,$3C,$8F,$6C,$7F,$A2,$6B,$DA,$F1
	.dc.b	$3D,$26,$6B,$D3,$1A,$8F,$D4,$47,$D9,$32,$7E,$58,$C1,$2B,$F8,$23
	.dc.b	$A3,$DA,$F9,$6B,$C9,$3E,$1A,$FD,$13,$F1,$0F,$25,$FD,$DB,$E5,$7F
	.dc.b	$08,$E5,$3C,$63,$E7,$CE,$DF,$15,$F6,$9F,$4D,$7C,$47,$84,$FE,$18
	.dc.b	$DB,$E0,$7E,$B5,$FC,$18,$F2,$1C,$AB,$A3,$83,$E4,$DE,$E6,$57,$C5
	.dc.b	$BF,$05,$7E,$1E,$BC,$D1,$88,$E5,$18,$BE,$B7,$FA,$6B,$FC,$05,$7E
	.dc.b	$3E,$3C,$95,$B8,$FF,$11,$CF,$F9,$66,$7F,$C9,$31,$FE,$13,$8C,$1F
	.dc.b	$E4,$63,$FC,$5E,$60,$FF,$24,$BC,$9C,$91,$C1,$0F,$F1,$23,$C4,$86
	.dc.b	$DC,$B5,$F1,$36,$3C,$A4,$7C,$91,$D4,$D9,$F9,$5F,$E0,$0F,$F1,$31
	.dc.b	$F4,$CB,$1F,$AE,$6B,$F6,$A0,$23,$38,$7C,$F6,$E1,$F6,$9B,$3D,$EB
	.dc.b	$FC,$49,$1D,$DA,$77,$29,$DE,$3B,$38,$AE,$CD,$AE,$CD,$9E,$CF,$F9
	.dc.b	$22,$BB,$34,$EC,$53,$B3,$E6,$8F,$E2,$0D,$7F,$24,$DB,$FE,$AA,$D9
	.dc.b	$FE,$A8,$D7,$FA,$33,$6F,$FA,$FB,$6F,$FC,$92,$DB,$FF,$0D,$B5,$FF
	.dc.b	$1C,$35,$FF,$1C,$36,$FC,$8F,$B2,$BF,$D0,$8D,$AF,$F4,$A3,$6B,$EB
	.dc.b	$9F,$2A,$FD,$E7,$C7,$7D,$ED,$7D,$ED,$BF,$7B,$4F,$79,$4F,$79,$6F
	.dc.b	$DE,$DB,$F7,$B6,$BD,$ED,$7D,$ED,$7D,$ED,$AF,$7B,$F6,$2B,$DE,$DA
	.dc.b	$F7,$B6,$BD,$F7,$F0,$5F,$CB,$B5,$F8,$55,$F8,$A9,$FC,$13,$4E,$AF
	.dc.b	$C6,$33,$5F,$8E,$6C,$FE,$3A,$76,$71,$9F,$1C,$F1,$8C,$C7,$16,$CF
	.dc.b	$18,$DC,$71,$8F,$4C,$FA,$DF,$4B,$5F,$4C,$F3,$38,$95,$E3,$1E,$18
	.dc.b	$CC,$77,$33,$3D,$DE,$B3,$B7,$33,$CD,$CC,$ED,$F0,$9C,$CC,$C6,$DC
	.dc.b	$9F,$8C,$9F,$8E,$D9,$F8,$2D,$9F,$2B,$4F,$09,$D4,$E6,$7E,$52,$79
	.dc.b	$B6,$79,$B6,$79,$B4,$E6,$75,$3C,$E7,$5A,$FC,$21,$2B,$E7,$35,$EE
	.dc.b	$FE,$49,$F8,$D1,$C8,$87,$9A,$3F,$6D,$1F,$AB,$78,$57,$99,$E8,$D8
	.dc.b	$D1,$E0,$29,$E0,$30,$52,$3F,$4E,$3A,$D7,$C0,$7A,$B6,$79,$CE,$DB
	.dc.b	$19,$7D,$75,$E9,$72,$FC,$83,$E5,$1E,$EA,$EE,$E6,$FF,$82,$CF,$92
	.dc.b	$BA,$5F,$D4,$9C,$BE,$F9,$D5,$71,$7F,$85,$98,$7B,$3F,$C0,$4F,$1B
	.dc.b	$87,$4F,$90,$F4,$38,$BF,$83,$1E,$97,$CE,$FE,$04,$DC,$F8,$CF,$34
	.dc.b	$F8,$9C,$5F,$A8,$FC,$5D,$FA,$9C,$57,$5A,$F3,$1E,$13,$CE,$FE,$62
	.dc.b	$78,$3B,$8D,$1F,$42,$FC,$B1,$E4,$AF,$D5,$46,$0F,$C5,$CF,$0B,$F2
	.dc.b	$D7,$C9,$8F,$E6,$26,$6B,$CB,$1E,$FA,$E5,$3E,$6A,$E3,$3F,$2D,$AF
	.dc.b	$CB,$9D,$CF,$EB,$AF,$D3,$3B,$BE,$AF,$B6,$7E,$44,$62,$BD,$75,$C2
	.dc.b	$BD,$87,$C1,$BF,$9F,$7E,$87,$C3,$5C,$23,$9C,$EE,$FA,$47,$E8,$AB
	.dc.b	$6F,$AE,$32,$FC,$19,$FD,$E3,$F4,$AB,$E4,$D7,$08,$D4,$FC,$57,$47
	.dc.b	$37,$C0,$7F,$2A,$47,$F3,$C2,$B9,$70,$47,$04,$3F,$96,$BC,$1A,$ED
	.dc.b	$A6,$CF,$EE,$CE,$DA,$ED,$A6,$E7,$FC,$A6,$48,$FF,$1F,$C7,$42,$1C
	.dc.b	$0B,$1F,$E1,$DE,$7F,$6C,$A9,$7F,$E0,$8B,$57,$E5,$8D,$CE,$A3,$FC
	.dc.b	$1B,$19,$AF,$8F,$39,$9F,$F2,$3C,$E6,$7B,$57,$58,$C5,$6E,$BB,$D7
	.dc.b	$EA,$A7,$BD,$6E,$30,$76,$70,$FF,$04,$8D,$CF,$92,$36,$78,$1C,$3B
	.dc.b	$39,$38,$AF,$D1,$CF,$28,$F4,$CF,$29,$F3,$4F,$A2,$BC,$31,$C2,$79
	.dc.b	$D7,$07,$B1,$E1,$71,$1B,$8C,$47,$C0,$77,$1D,$9F,$04,$F6,$3F,$63
	.dc.b	$5E,$5A,$E1,$5C,$CE,$15,$CA,$B8,$4F,$E7,$0F,$0B,$87,$F7,$87,$C1
	.dc.b	$70,$FE,$E0,$D9,$93,$91,$E0,$77,$7F,$8B,$9F,$83,$39,$37,$7E,$73
	.dc.b	$E4,$3C,$CF,$09,$D8,$87,$84,$E4,$62,$7B,$4F,$C8,$9E,$73,$98,$FF
	.dc.b	$2D,$47,$77,$E4,$C7,$73,$A4,$66,$3E,$34,$66,$7A,$DF,$CA,$3E,$65
	.dc.b	$FB,$AB,$A9,$EE,$BF,$0D,$F9,$8E,$B1,$93,$C2,$F9,$5F,$F0,$CC,$78
	.dc.b	$6B,$AC,$F8,$67,$A9,$ED,$38,$BD,$6F,$CB,$7E,$39,$E6,$77,$3D,$73
	.dc.b	$DE,$BF,$C4,$B7,$DE,$BF,$13,$5B,$9F,$0D,$6E,$7F,$C5,$35,$B8,$FB
	.dc.b	$35,$B9,$E2,$F9,$4E,$F1,$E5,$7D,$35,$F2,$67,$6D,$9E,$73,$C6,$3C
	.dc.b	$31,$C6,$3D,$66,$67,$89,$EF,$9F,$5C,$78,$E7,$9C,$75,$7F,$16,$7D
	.dc.b	$3A,$FA,$A4,$9F,$96,$D7,$D2,$D9,$F1,$B6,$BC,$35,$F5,$6B,$CB,$5D
	.dc.b	$E3,$EF,$C7,$9D,$F1,$CF,$9C,$F9,$71,$E7,$8F,$CE,$D7,$CD,$7E,$81
	.dc.b	$FE,$2F,$70,$FA,$63,$83,$F1,$6B,$E0,$C7,$78,$E6,$EA,$7D,$11,$F9
	.dc.b	$F7,$99,$87,$71,$C5,$E9,$5E,$68,$C4,$F0,$7C,$37,$F9,$0A,$E9,$1D
	.dc.b	$CC,$9F,$80,$AF,$E1,$35,$C6,$3B,$57,$1A,$F0,$DF,$C9,$AF,$7C,$E6
	.dc.b	$7E,$05,$78,$8E,$A7,$D0,$BE,$AF,$84,$EA,$78,$63,$CA,$F1,$32,$71
	.dc.b	$73,$7E,$18,$EA,$FC,$C3,$0F,$17,$EF,$57,$96,$FD,$91,$A3,$F0,$F3
	.dc.b	$E3,$71,$7F,$88,$BF,$7B,$E8,$9F,$1C,$F7,$8C,$3C,$27,$F1,$CE,$5F
	.dc.b	$3C,$68,$FC,$63,$F8,$79,$F0,$46,$63,$8B,$F6,$E7,$A4,$78,$8C,$9C
	.dc.b	$67,$DD,$3C,$63,$F2,$67,$FA,$72,$BD,$F1,$98,$FC,$39,$C6,$BA,$BF
	.dc.b	$61,$F8,$F5,$F4,$E3,$C0,$43,$89,$D5,$F4,$BE,$E3,$CA,$65,$F2,$B9
	.dc.b	$3C,$A7,$FB,$12,$BC,$AE,$4E,$A6,$63,$AB,$E1,$8F,$D1,$3E,$7A,$F6
	.dc.b	$5F,$DD,$AD,$CE,$4F,$DD,$9D,$6B,$BB,$FC,$39,$AE,$CF,$2C,$E2,$24
	.dc.b	$FF,$10,$3A,$BF,$A0,$7E,$34,$F1,$24,$79,$8E,$31,$FC,$58,$E2,$7E
	.dc.b	$62,$BE,$F1,$F5,$63,$C7,$5E,$B9,$F9,$35,$F4,$63,$17,$F6,$2B,$C5
	.dc.b	$5F,$A5,$AF,$8C,$FC,$2A,$DD,$FF,$8B,$A3,$FC,$A9,$5F,$E4,$86,$9F
	.dc.b	$E5,$0A,$FF,$0D,$38,$3F,$C9,$AF,$F8,$C1,$C1,$0E,$84,$74,$D7,$FC
	.dc.b	$54,$ED,$C9,$0C,$BB,$7C,$2D,$8F,$33,$4F,$33,$88,$FC,$47,$F0,$0F
	.dc.b	$F8,$AC,$FB,$65,$9F,$C1,$B5,$F9,$70,$8C,$61,$FA,$26,$DF,$E9,$DB
	.dc.b	$3D,$AB,$FC,$4B,$1D,$9A,$76,$29,$D9,$E3,$1F,$A9,$6B,$F8,$76,$DF
	.dc.b	$F4,$86,$CF,$F4,$86,$BF,$CF,$DB,$7F,$D6,$5B,$7F,$E4,$06,$DF,$F8
	.dc.b	$9D,$AF,$F8,$D1,$AF,$F8,$7D,$B7,$E4,$7D,$97,$FE,$5C,$6D,$FF,$9F
	.dc.b	$5B,$7F,$E8,$D6,$D8,$8E,$1B,$5F,$14,$E8,$5A,$E8,$DA,$E8,$D9,$E8
	.dc.b	$7D,$77,$0F,$47,$E7,$1F,$42,$7E,$D3,$67,$ED,$36,$7E,$D3,$E2,$38
	.dc.b	$96,$78,$B6,$78,$B6,$78,$CF,$A2,$38,$B4,$E2,$59,$E2,$D9,$E3,$1E
	.dc.b	$E9,$E2,$D9,$E2,$D3,$89,$67,$89,$F1,$2B,$E7,$3F,$16,$B9,$B9,$7E
	.dc.b	$3B,$9A,$EA,$E6,$36,$FC,$B8,$CB,$6B,$C2,$EE,$32,$F3,$8C,$BB,$3C
	.dc.b	$A5,$3F,$41,$3F,$45,$AF,$CB,$6C,$FC,$56,$CF,$BD,$A7,$85,$EE,$F8
	.dc.b	$4E,$E5,$36,$78,$4F,$D7,$4F,$85,$AF,$85,$B3,$E1,$6C,$F8,$5A,$78
	.dc.b	$4E,$F5,$F4,$09,$5F,$39,$AF,$CC,$8F,$21,$C8,$F2,$5F,$EF,$E3,$C8
	.dc.b	$D7,$C8,$DB,$F2,$36,$FC,$8D,$AF,$21,$F0,$0E,$EE,$5F,$85,$7F,$BD
	.dc.b	$BE,$E7,$F0,$D3,$E2,$39,$7C,$26,$4F,$8C,$E6,$FC,$71,$97,$E4,$46
	.dc.b	$6F,$CB,$39,$7E,$54,$66,$BC,$E7,$C3,$3E,$2B,$97,$E2,$CF,$17,$71
	.dc.b	$97,$DB,$19,$7D,$AF,$A8,$F6,$9F,$C9,$CC,$1E,$A3,$35,$FC,$A5,$CB
	.dc.b	$EA,$73,$5F,$CA,$DC,$BE,$A7,$33,$FC,$B5,$CB,$E6,$72,$FF,$2F,$72
	.dc.b	$F9,$9F,$9A,$79,$4F,$21,$27,$E6,$B9,$7C,$8E,$67,$E6,$B9,$7C,$8E
	.dc.b	$67,$E6,$DF,$E0,$5F,$9B,$1E,$27,$E4,$BC,$5F,$43,$C5,$FA,$2E,$5F
	.dc.b	$45,$78,$8F,$29,$F2,$63,$84,$76,$7D,$B1,$93,$85,$79,$9E,$2F,$CD
	.dc.b	$8E,$37,$F7,$6F,$D0,$FD,$F3,$FA,$AC,$7A,$5C,$3F,$AF,$30,$7D,$0B
	.dc.b	$FE,$B8,$E0,$F2,$CF,$47,$17,$E2,$9F,$B5,$1E,$23,$E3,$4E,$DC,$1E
	.dc.b	$CB,$F4,$57,$B2,$BE,$4B,$CD,$FC,$04,$F0,$8D,$47,$ED,$6F,$D3,$3C
	.dc.b	$E3,$C3,$3A,$7E,$53,$CD,$F5,$C7,$3A,$F9,$47,$F2,$2A,$E7,$5B,$BB
	.dc.b	$3A,$BF,$C1,$DF,$E9,$DE,$B1,$97,$EF,$D7,$D7,$8C,$46,$8E,$33,$F8
	.dc.b	$37,$07,$E1,$23,$F2,$73,$C6,$3F,$17,$23,$FA,$F1,$5F,$EA,$0D,$7F
	.dc.b	$BD,$35,$FE,$B4,$DB,$FF,$1F,$46,$A3,$99,$1E,$6D,$74,$60,$86,$08
	.dc.b	$76,$78,$34,$D9,$4D,$C7,$89,$C1,$F0,$87,$47,$BE,$54,$BF,$F0,$CD
	.dc.b	$6A,$7F,$65,$5B,$AD,$4F,$F8,$7A,$3F,$8F,$19,$AE,$2E,$2B,$E6,$57
	.dc.b	$0A,$FF,$18,$4E,$67,$C3,$5C,$1E,$E7,$07,$0F,$EE,$23,$B4,$F8,$63
	.dc.b	$67,$27,$0E,$CE,$4E,$2B,$F3,$13,$CA,$3C,$73,$CA,$7C,$D1,$F0,$63
	.dc.b	$91,$D5,$DB,$88,$DC,$FA,$DB,$1C,$EB,$D3,$1E,$08,$F1,$D7,$28,$C4
	.dc.b	$75,$8F,$04,$60,$E7,$1D,$E7,$99,$FA,$0A,$F5,$57,$08,$C4,$F6,$8C
	.dc.b	$1F,$C0,$23,$8B,$F0,$DD,$C7,$86,$7D,$51,$A9,$FD,$41,$B2,$18,$3D
	.dc.b	$4E,$DF,$C7,$47,$90,$E6,$F0,$33,$3E,$58,$CC,$72,$3C,$C6,$DC,$BE
	.dc.b	$A3,$C0,$52,$5F,$FA,$5A,$3C,$13,$F4,$67,$C1,$39,$8F,$01,$98,$F0
	.dc.b	$1D,$27,$8C,$7C,$F8,$E2,$FF,$8C,$E3,$37,$F5,$6F,$E0,$5F,$5A,$F9
	.dc.b	$47,$5B,$F2,$BC,$4F,$09,$D5,$F0,$9F,$11,$E2,$65,$E2,$FF,$90,$6F
	.dc.b	$C3,$5F,$62,$FF,$0A,$78,$EB,$CF,$5B,$3B,$C6,$63,$BD,$F9,$E3,$8D
	.dc.b	$73,$3D,$75,$E7,$3C,$71,$C6,$BC,$6F,$F9,$46,$FC,$71,$F6,$0F,$5B
	.dc.b	$F6,$09,$3F,$0D,$AF,$B5,$B3,$D5,$B5,$E3,$AF,$5C,$F7,$AF,$5C,$E6
	.dc.b	$7D,$71,$E7,$9F,$5C,$78,$CF,$C3,$57,$47,$0F,$A1,$FC,$04,$F4,$72
	.dc.b	$D7,$2F,$C2,$AF,$04,$7C,$27,$FC,$E1,$3C,$E3,$BC,$F1,$74,$7F,$92
	.dc.b	$A3,$44,$8D,$5F,$D3,$AF,$D5,$4E,$67,$C1,$5E,$46,$BF,$39,$ED,$1E
	.dc.b	$D8,$C4,$69,$F0,$9F,$E8,$D8,$E9,$7F,$1A,$36,$FD,$19,$E6,$43,$EB
	.dc.b	$15,$E1,$1D,$0F,$F4,$3C,$F4,$AF,$4D,$7B,$0F,$A8,$E1,$F2,$47,$C7
	.dc.b	$BF,$81,$5F,$BF,$9E,$D5,$F9,$27,$F3,$47,$03,$31,$F8,$48,$FA,$31
	.dc.b	$88,$F5,$C7,$1B,$FC,$44,$F1,$23,$EA,$AF,$0D,$70,$3C,$53,$F2,$63
	.dc.b	$31,$F7,$23,$8D,$75,$3C,$2E,$23,$8D,$75,$8F,$83,$18,$9F,$85,$39
	.dc.b	$9F,$6B,$F3,$2F,$E5,$57,$B4,$FC,$44,$FB,$5F,$14,$F1,$3D,$A7,$18
	.dc.b	$D4,$FE,$0E,$77,$5E,$03,$F3,$31,$DA,$3E,$01,$F9,$EA,$E5,$5E,$B8
	.dc.b	$FD,$6B,$E6,$AE,$51,$EB,$3C,$C7,$E3,$63,$C2,$79,$EF,$E4,$DF,$79
	.dc.b	$F1,$9F,$9B,$BE,$E4,$BF,$99,$1F,$02,$FC,$4F,$03,$C7,$3F,$02,$7C
	.dc.b	$71,$EA,$3F,$29,$7D,$EE,$5F,$F1,$13,$CA,$7D,$D3,$EA,$CF,$18,$F3
	.dc.b	$4F,$A2,$77,$7F,$3E,$FE,$FB,$E7,$BF,$84,$79,$23,$27,$91,$F4,$1E
	.dc.b	$49,$E7,$19,$BF,$AC,$79,$A3,$F5,$E7,$C0,$7F,$4C,$F9,$A7,$9C,$79
	.dc.b	$A3,$0F,$1A,$F3,$4F,$C9,$7C,$CF,$B6,$7E,$11,$FD,$38,$FF,$2D,$4F
	.dc.b	$F9,$1C,$8F,$F9,$14,$FF,$18,$9A,$31,$1F,$E2,$E7,$07,$F9,$2D,$C3
	.dc.b	$C0,$8E,$08,$7F,$8A,$9E,$24,$36,$E5,$AF,$F8,$CD,$D3,$C8,$C3,$A8
	.dc.b	$FE,$65,$1F,$E2,$43,$F2,$05,$8F,$D5,$35,$FC,$6B,$67,$E3,$A3,$15
	.dc.b	$F6,$DA,$7E,$00,$D4,$FF,$8B,$DE,$ED,$3B,$94,$EE,$59,$EF,$1D,$0F
	.dc.b	$CD,$14,$FC,$49,$6B,$DB,$3C,$6B,$C8,$D9,$F2,$1B,$7F,$A9,$35,$FE
	.dc.b	$88,$DA,$FE,$76,$D9,$FF,$1D,$34,$FF,$1C,$14,$FF,$0B,$94,$FF,$19
	.dc.b	$16,$BD,$0F,$C2,$AF,$F2,$E3,$6B,$FC,$DA,$DA,$FF,$47,$B4,$FF,$48
	.dc.b	$CF,$C8,$38,$96,$78,$B6,$78,$B6,$78,$BF,$38,$E2,$59,$E2,$D9,$E2
	.dc.b	$D9,$E3,$3F,$52,$3E,$64,$7D,$46,$BF,$51,$B5,$F5,$1E,$F3,$C5,$B5
	.dc.b	$F5,$9B,$5F,$59,$E2,$7D,$62,$9F,$58,$B3,$F5,$9B,$3F,$59,$B3,$F5
	.dc.b	$9A,$FD,$66,$BF,$59,$B5,$F5,$9F,$91,$1F,$9F,$3E,$91,$5C,$BF,$9E
	.dc.b	$7E,$33,$96,$B9,$AF,$80,$7C,$E2,$D7,$CE,$6C,$7C,$E8,$F8,$91,$B6
	.dc.b	$9B,$2C,$6C,$F9,$13,$F2,$9F,$A8,$D7,$EA,$36,$BE,$A5,$75,$9E,$2D
	.dc.b	$78,$B5,$E2,$D8,$E3,$5E,$39,$E2,$D7,$8B,$5E,$2D,$8E,$31,$E9,$36
	.dc.b	$53,$65,$76,$FC,$D9,$F8,$ED,$9F,$86,$D3,$C6,$71,$AF,$C5,$1F,$04
	.dc.b	$B3,$EB,$6C,$F9,$5B,$3E,$33,$F0,$8F,$13,$F2,$47,$CD,$73,$3C,$E3
	.dc.b	$2F,$C9,$8C,$CF,$59,$CB,$F1,$63,$27,$9E,$32,$F8,$5F,$D6,$BF,$C2
	.dc.b	$DA,$43,$C4,$7E,$51,$E6,$7C,$32,$1F,$30,$A6,$8F,$CB,$47,$CA,$69
	.dc.b	$E1,$8E,$A7,$E7,$1E,$07,$8D,$CC,$FB,$DF,$23,$F5,$A3,$99,$E0,$8C
	.dc.b	$CF,$78,$E7,$1F,$B2,$32,$F8,$67,$A4,$78,$4C,$C7,$37,$07,$8A,$7E
	.dc.b	$63,$F0,$CD,$C7,$88,$95,$F8,$C7,$07,$8A,$39,$D7,$80,$95,$E7,$8E
	.dc.b	$73,$F0,$67,$2F,$9E,$7C,$33,$C9,$B3,$A9,$FC,$F9,$E9,$70,$F9,$27
	.dc.b	$F2,$2E,$67,$89,$F8,$99,$FE,$41,$3E,$F9,$F0,$1F,$89,$8C,$9D,$1F
	.dc.b	$BD,$5E,$F7,$DD,$1F,$97,$37,$5D,$A3,$99,$F8,$89,$E1,$5F,$00,$FA
	.dc.b	$C7,$86,$3C,$C7,$03,$F1,$C7,$CC,$78,$1F,$0D,$D1,$F2,$67,$C0,$F2
	.dc.b	$21,$C8,$F3,$BF,$04,$FE,$7D,$1E,$18,$EC,$FE,$D6,$BD,$71,$DA,$3F
	.dc.b	$03,$5E,$A7,$84,$78,$4D,$38,$3F,$10,$53,$A5,$78,$2B,$F2,$70,$3F
	.dc.b	$B3,$96,$7F,$B7,$B5,$FE,$88,$D7,$FA,$43,$5F,$F2,$63,$C8,$8F,$26
	.dc.b	$BC,$9A,$F2,$6B,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$F3,$21,$3E
	.dc.b	$69,$F8,$11,$87,$0F,$E9,$9B,$1B,$7F,$68,$60,$CC,$FE,$9D,$C1,$F0
	.dc.b	$A7,$93,$67,$94,$62,$3F,$CE,$4F,$85,$D4,$F3,$78,$1D,$9F,$D5,$3D
	.dc.b	$8E,$33,$FA,$C7,$07,$F3,$62,$9F,$C5,$8A,$7A,$8E,$A7,$82,$3A,$B8
	.dc.b	$3A,$BE,$37,$B9,$E3,$7E,$64,$7B,$08,$EE,$3F,$18,$27,$DC,$F4,$8E
	.dc.b	$EF,$43,$C4,$F4,$7C,$F1,$F1,$DE,$6F,$27,$DC,$79,$CA,$73,$3B,$47
	.dc.b	$37,$47,$C3,$29,$F0,$CE,$07,$23,$BC,$FB,$23,$CF,$18,$3E,$13,$FE
	.dc.b	$8A,$7C,$F1,$F4,$DE,$B3,$E9,$7F,$1F,$1C,$E7,$C9,$1E,$03,$C8,$53
	.dc.b	$C8,$77,$3E,$2B,$DD,$EB,$1D,$DE,$13,$DD,$FA,$11,$DE,$78,$46,$A3
	.dc.b	$83,$D4,$F8,$F1,$A7,$07,$C3,$3D,$27,$C0,$3F,$D2,$C7,$C0,$7F,$00
	.dc.b	$E8,$F3,$9A,$3C,$EE,$9F,$29,$CC,$FF,$35,$9C,$DF,$9E,$F3,$3C,$A7
	.dc.b	$33,$D4,$F3,$9F,$C2,$47,$CA,$3D,$25,$39,$1E,$B3,$C2,$79,$8F,$09
	.dc.b	$E6,$8F,$0B,$A3,$C2,$68,$F0,$BA,$8F,$0C,$7D,$83,$B9,$E1,$39,$1E
	.dc.b	$17,$67,$D7,$2C,$FD,$48,$E2,$73,$38,$9C,$E3,$73,$F0,$67,$EF,$4F
	.dc.b	$C5,$3E,$09,$C6,$77,$3E,$79,$DC,$FA,$CD,$CF,$8E,$32,$FF,$9E,$63
	.dc.b	$EA,$4F,$E6,$1D,$CF,$9D,$F2,$9D,$DE,$27,$F9,$FD,$E3,$1F,$8D,$7D
	.dc.b	$B3,$DD,$FC,$13,$DD,$FA,$51,$DD,$DC,$F7,$9F,$5C,$E6,$7C,$31,$F2
	.dc.b	$9F,$D9,$9D,$27,$C4,$FE,$A0,$C9,$DA,$7C,$8F,$28,$FE,$38,$E8,$FC
	.dc.b	$0C,$F9,$5F,$D5,$19,$3E,$1C,$FD,$37,$07,$A0,$F1,$3F,$EA,$57,$27
	.dc.b	$73,$E2,$9F,$64,$E0,$79,$CF,$31,$93,$81,$93,$F2,$6F,$C3,$8F,$85
	.dc.b	$19,$3C,$B1,$F8,$17,$C3,$1A,$7D,$B3,$F1,$9F,$69,$C4,$FA,$33,$ED
	.dc.b	$70,$78,$8E,$2F,$B5,$CC,$F5,$7C,$31,$E5,$9F,$9B,$1A,$7F,$0F,$1C
	.dc.b	$63,$C2,$43,$8B,$F5,$A7,$A4,$7C,$63,$E0,$C6,$67,$AC,$71,$3D,$31
	.dc.b	$F0,$5F,$01,$F4,$0F,$D1,$1C,$CF,$CE,$99,$9F,$C0,$38,$38,$9F,$C2
	.dc.b	$5C,$BC,$48,$7E,$2E,$3E,$43,$E3,$3E,$41,$FC,$0E,$7C,$EF,$53,$D6
	.dc.b	$7D,$F9,$F7,$BC,$0F,$5C,$FC,$B7,$ED,$CF,$27,$F8,$33,$5E,$07,$8E
	.dc.b	$3E,$13,$E3,$9E,$A7,$EC,$DF,$69,$E2,$3E,$29,$C9,$DC,$FD,$89,$FF
	.dc.b	$6D,$4F,$88,$F1,$BF,$01,$F6,$10,$F2,$9B,$3F,$3A,$76,$9E,$AF,$28
	.dc.b	$F3,$BC,$4F,$14,$FF,$03,$3D,$D3,$C2,$3E,$A4,$7E,$6D,$E1,$19,$7F
	.dc.b	$2F,$3E,$87,$DE,$6A,$7A,$3E,$53,$E1,$4F,$DE,$7C,$27,$96,$7E,$93
	.dc.b	$C8,$FA,$AF,$E4,$DE,$47,$E6,$8F,$B9,$3E,$82,$3F,$C1,$A3,$C8,$FD
	.dc.b	$58,$CB,$EC,$8F,$D3,$3E,$73,$C4,$F7,$7C,$47,$DF,$7C,$AF,$33,$F3
	.dc.b	$91,$F1,$5F,$CB,$CF,$C5,$7F,$28,$7A,$0F,$59,$E8,$30,$75,$9F,$03
	.dc.b	$D4,$FB,$71,$F1,$CE,$8F,$69,$F9,$AF,$9A,$3F,$9D,$CF,$28,$F6,$1C
	.dc.b	$A7,$CC,$FD,$03,$F8,$44,$7E,$35,$E8,$7B,$1F,$8A,$E0,$DB,$F8,$A7
	.dc.b	$07,$DC,$8E,$B3,$E6,$8F,$C6,$CF,$57,$C0,$6D,$EA,$E0,$FC,$43,$B8
	.dc.b	$E4,$7B,$23,$E4,$1E,$F9,$FD,$44,$8F,$7C,$7C,$93,$F2,$AE,$5F,$39
	.dc.b	$A3,$EA,$9C,$CF,$C4,$3B,$7C,$87,$03,$F5,$E6,$CF,$C5,$3F,$8E,$7A
	.dc.b	$1F,$55,$F5,$47,$E4,$23,$E0,$46,$DE,$87,$C6,$7E,$63,$E0,$3F,$08
	.dc.b	$7C,$03,$D8,$7E,$84,$FD,$F4,$75,$7E,$4C,$72,$9F,$93,$3F,$1A,$3C
	.dc.b	$B3,$D6,$7F,$8B,$3D,$0F,$CA,$9D,$0F,$CA,$BF,$45,$C1,$F0,$4F,$9B
	.dc.b	$1D,$23,$BB,$F0,$23,$8C,$FE,$59,$EC,$7B,$0E,$66,$27,$B1,$F9,$F8
	.dc.b	$F6,$47,$CB,$9F,$64,$62,$39,$4E,$DE,$E7,$DA,$9F,$24,$EE,$7D,$F3
	.dc.b	$83,$47,$D2,$7D,$33,$DA,$3D,$2F,$AD,$FC,$83,$5F,$44,$FB,$5E,$87
	.dc.b	$E1,$23,$F1,$53,$F4,$A3,$F8,$69,$F6,$CD,$9F,$28,$F1,$4F,$9C,$E8
	.dc.b	$E0,$F3,$9F,$65,$D4,$F8,$E3,$DE,$D7,$9B,$5E,$0D,$9F,$7B,$FC,$72
	.dc.b	$3C,$67,$E9,$0F,$69,$E0,$8F,$53,$FC,$0A,$38,$3E,$53,$E0,$3C,$9F
	.dc.b	$B0,$E9,$F4,$1C,$E7,$DC,$FC,$43,$F0,$67,$13,$FC,$C4,$7D,$C3,$A9
	.dc.b	$DC,$FE,$79,$1E,$B8,$F8,$6F,$D9,$8F,$AE,$D7,$EB,$BB,$7C,$13,$F8
	.dc.b	$97,$F3,$87,$C1,$3C,$4F,$E2,$1F,$01,$92,$9F,$C1,$0F,$8E,$79,$8F
	.dc.b	$79,$F1,$E3,$F9,$A3,$C0,$F6,$9E,$03,$91,$E0,$9F,$C0,$BC,$8F,$C3
	.dc.b	$14,$FD,$FC,$7C,$D9,$FA,$B1,$A6,$BA,$8F,$0B,$C4,$EE,$78,$E3,$DC
	.dc.b	$E4,$F2,$9E,$C3,$A4,$FD,$87,$D8,$7E,$6C,$E4,$FA,$23,$C3,$1D,$8E
	.dc.b	$8F,$63,$2F,$63,$E0,$4F,$D5,$7A,$9F,$84,$3A,$B9,$7F,$02,$FD,$D7
	.dc.b	$F4,$2D,$7F,$42,$D7,$F4,$2D,$8F,$D0,$BF,$11,$F1,$9E,$59,$F7,$4F
	.dc.b	$C0,$3F,$24,$7F,$3A,$9F,$DE,$C6,$A3,$E7,$1F,$0C,$F7,$BF,$0C,$D3
	.dc.b	$F0,$E7,$BC,$66,$3F,$87,$CF,$96,$3F,$36,$FD,$D7,$8C,$7C,$73,$E1
	.dc.b	$C6,$DF,$CD,$C7,$F7,$23,$73,$FC,$B1,$C4,$7A,$8F,$CD,$1E,$69,$F9
	.dc.b	$8F,$A8,$FA,$06,$5E,$CF,$D4,$9F,$D4,$3F,$85,$39,$BE,$C8,$E7,$1F
	.dc.b	$C0,$9F,$9D,$3E,$43,$E0,$47,$ED,$A3,$DC,$FC,$26,$B9,$78,$B9,$6B
	.dc.b	$93,$E8,$4F,$86,$3C,$F1,$F8,$67,$9B,$FC,$09,$CB,$EF,$8F,$5B,$F9
	.dc.b	$28,$CB,$E9,$3A,$CF,$A6,$3D,$B1,$FD,$E5,$EA,$FE,$49,$AF,$E4,$9A
	.dc.b	$FE,$49,$B3,$F9,$28,$EA,$7B,$4A,$7B,$4F,$A6,$FB,$63,$8C,$FE,$19
	.dc.b	$F8,$73,$ED,$72,$FE,$11,$FF,$17,$CF,$E4,$23,$F9,$91,$F8,$33,$FC
	.dc.b	$02,$FC,$27,$F1,$B1,$C9,$FC,$04,$FE,$29,$FE,$30,$7E,$4C,$FC,$34
	.dc.b	$6D,$FA,$0F,$91,$F8,$8D,$7E,$23,$5F,$88,$D9,$F8,$91,$F0,$27,$E8
	.dc.b	$4F,$D4,$8F,$A0,$EE,$3E,$14,$7C,$83,$E1,$46,$5F,$84,$FF,$8C,$5F
	.dc.b	$DA,$9F,$08,$F5,$CF,$E9,$E3,$E3,$CF,$E5,$E7,$47,$E7,$CF,$8E,$FC
	.dc.b	$37,$07,$C3,$9F,$AD,$3E,$18,$F9,$2F,$47,$E8,$1B,$9F,$6C,$7C,$98
	.dc.b	$FC,$A4,$75,$7F,$29,$19,$7F,$28,$7C,$53,$F1,$05,$3F,$11,$3F,$6D
	.dc.b	$FA,$B1,$E3,$7E,$A9,$F1,$CF,$AA,$53,$EA,$94,$FA,$A5,$8F,$AA,$75
	.dc.b	$3D,$EE,$67,$E9,$BE,$33,$E9,$99,$3E,$9C,$FF,$2A,$7E,$1C,$FD,$18
	.dc.b	$FA,$CF,$E5,$E3,$FB,$63,$5F,$E6,$51,$FD,$9D,$FF,$4C,$3C,$27,$EC
	.dc.b	$47,$06,$BC,$0F,$EA,$E7,$08,$F1,$3F,$82,$8F,$F3,$43,$A3,$FD,$55
	.dc.b	$1F,$65,$D3,$FC,$0A,$7F,$DF,$6D,$9F,$F6,$A3,$67,$B3,$E7,$79,$BE
	.dc.b	$77,$E0,$C7,$F2,$47,$E3,$9F,$24,$FA,$A5,$3E,$A9,$4F,$AA,$59,$FA
	.dc.b	$B1,$F8,$98,$FD,$DC,$FE,$05,$FC,$6B,$EB,$7F,$33,$3F,$D8,$5F,$C5
	.dc.b	$BF,$76,$7F,$2B,$3C,$67,$F2,$33,$97,$73,$DD,$FC,$79,$F7,$CF,$D0
	.dc.b	$CF,$BD,$F9,$2F,$EE,$E7,$E4,$B6,$7E,$4B,$67,$E4,$9F,$4C,$FB,$47
	.dc.b	$EE,$8F,$96,$FD,$A7,$D7,$1F,$69,$D1,$C2,$7F,$14,$FF,$AA,$DF,$53
	.dc.b	$FA,$88,$F5,$47,$E7,$23,$E6,$47,$E7,$23,$71,$F5,$9A,$FD,$66,$BF
	.dc.b	$59,$AF,$D6,$69,$F5,$A7,$E9,$47,$EF,$9F,$9C,$FF,$1E,$3E,$51,$FB
	.dc.b	$83,$C6,$7E,$A9,$FC,$63,$ED,$9F,$CF,$BE,$D7,$D6,$7D,$A7,$99,$F8
	.dc.b	$28,$FA,$8F,$C8,$7F,$30,$FC,$C7,$AB,$CD,$F3,$BF,$8C,$39,$9F,$84
	.dc.b	$8E,$EF,$E1,$23,$6F,$E1,$0F,$49,$F8,$47,$EC,$3D,$8F,$F1,$34,$79
	.dc.b	$DD,$9F,$06,$7E,$4C,$7D,$33,$BC,$7D,$99,$EF,$3F,$64,$DC,$7C,$73
	.dc.b	$F1,$47,$AE,$3E,$0B,$F4,$4F,$59,$F4,$4F,$09,$F1,$DF,$3B,$C8,$F3
	.dc.b	$9E,$F7,$9B,$F1,$4E,$27,$C5,$7C,$B3,$ED,$9F,$2C,$FC,$39,$F8,$4F
	.dc.b	$FB,$7A,$7D,$AF,$E3,$DA,$FE,$3D,$AF,$E3,$DA,$FE,$3D,$B3,$F8,$F9
	.dc.b	$F2,$CF,$D3,$8F,$0B,$F3,$67,$67,$CD,$9F,$3C,$7B,$5F,$82,$FE,$B0
	.dc.b	$FF,$20,$C7,$AC,$E2,$FC,$F3,$47,$F6,$88,$E6,$FD,$77,$C7,$3F,$5E
	.dc.b	$78,$CF,$C3,$3E,$71,$F0,$E7,$8B,$F0,$DA,$FC,$36,$BF,$0D,$AF,$C3
	.dc.b	$6C,$FC,$37,$D6,$FF,$04,$8F,$B6,$79,$63,$33,$FA,$C8,$C9,$E1,$8F
	.dc.b	$4C,$7C,$17,$AB,$F2,$5C,$C7,$C9,$7E,$EC,$FF,$5F,$7A,$C7,$F4,$27
	.dc.b	$F6,$2F,$EE,$CF,$E8,$A7,$F8,$48,$E3,$3F,$D1,$DF,$81,$1D,$9F,$43
	.dc.b	$A7,$81,$F2,$A7,$F9,$F3,$A3,$13,$FC,$9E,$3E,$14,$7C,$F9,$E8,$FB
	.dc.b	$DF,$3B,$FC,$E9,$FF,$0A,$3F,$A7,$8F,$C5,$BF,$A7,$8F,$13,$FD,$CC
	.dc.b	$F5,$94,$FE,$E0,$79,$8F,$C3,$9F,$E2,$53,$E4,$01,$09,$C4,$43,$51
	.dc.b	$97,$06,$22,$B0,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21
	.dc.b	$08,$6E,$3E,$A9,$ED,$7E,$27,$47,$C2,$8C,$42,$3E,$8B,$F1,$63,$31
	.dc.b	$B9,$F9,$0E,$A7,$07,$CC,$6B,$D8,$FC,$39,$F2,$A7,$9B,$F3,$4E,$06
	.dc.b	$8F,$A2,$E9,$FB,$33,$E0,$74,$79,$5C,$3A,$39,$9F,$84,$8F,$6B,$E4
	.dc.b	$3D,$C7,$84,$F3,$1B,$8F,$BD,$1E,$12,$3F,$BB,$8E,$C7,$99,$F0,$C6
	.dc.b	$A3,$F2,$E7,$94,$F1,$B9,$7F,$32,$71,$33,$1C,$8E,$D1,$E8,$8F,$31
	.dc.b	$E8,$3C,$25,$7E,$EC,$7A,$0E,$51,$D0,$F1,$9F,$06,$63,$F2,$8A,$74
	.dc.b	$7E,$E3,$E0,$73,$3D,$A3,$ED,$4F,$67,$E4,$C6,$5F,$61,$88,$D3,$FE
	.dc.b	$2E,$38,$9E,$53,$47,$98,$F0,$10,$FA,$87,$03,$64,$3A,$1E,$03,$E2
	.dc.b	$3E,$07,$F3,$73,$EA,$8F,$9A,$7C,$78,$F1,$1C,$0F,$11,$F0,$CF,$14
	.dc.b	$FB,$5F,$B8,$78,$67,$E3,$BB,$24,$6D,$F0,$1D,$8F,$19,$D2,$7C,$6F
	.dc.b	$B4,$F1,$1E,$D3,$67,$47,$6F,$BE,$7A,$BF,$8C,$3C,$A7,$22,$1C,$A7
	.dc.b	$F2,$4F,$C8,$37,$1C,$9C,$15,$F6,$C7,$06,$CF,$DC,$6C,$FC,$A6,$BF
	.dc.b	$14,$FC,$71,$EF,$3E,$11,$FB,$A7,$C0,$F9,$63,$EC,$39,$38,$1E,$D3
	.dc.b	$C8,$74,$3E,$21,$D0,$FC,$39,$E1,$29,$EE,$39,$9E,$E2,$C7,$A1,$F9
	.dc.b	$0D,$3C,$AF,$8C,$EE,$78,$CE,$27,$8E,$7E,$33,$67,$E3,$34,$F8,$C5
	.dc.b	$3C,$D1,$F8,$77,$99,$E8,$3D,$E5,$32,$ED,$F1,$9F,$78,$F3,$3F,$79
	.dc.b	$F9,$87,$C5,$7F,$35,$3E,$C8,$F8,$AD,$8C,$C7,$C6,$38,$9F,$93,$3A
	.dc.b	$CF,$D6,$78,$9B,$32,$77,$7F,$2E,$FC,$C3,$E0,$14,$F6,$1E,$A9,$FA
	.dc.b	$0F,$33,$A9,$EE,$3C,$67,$BC,$CC,$FB,$DF,$09,$E7,$33,$3E,$78,$F6
	.dc.b	$C7,$33,$67,$37,$27,$37,$89,$EB,$3F,$9E,$CF,$C7,$9F,$6B,$63,$27
	.dc.b	$A6,$79,$CF,$84,$DC,$65,$AF,$8C,$EE,$7A,$63,$8B,$E1,$3F,$7D,$3D
	.dc.b	$DA,$73,$3C,$27,$38,$E2,$F3,$9E,$F1,$ED,$33,$1F,$06,$32,$79,$DE
	.dc.b	$71,$B8,$CC,$78,$CF,$E1,$53,$E3,$69,$E3,$3C,$A7,$8E,$38,$BE,$33
	.dc.b	$AB,$DD,$DC,$F7,$72,$77,$3F,$DB,$33,$DD,$A7,$73,$C2,$77,$38,$9D
	.dc.b	$CE,$B3,$DD,$DC,$F7,$72,$77,$3F,$B8,$9F,$E0,$92,$18,$3F,$67,$1F
	.dc.b	$9E,$7F,$6F,$1F,$B5,$3F,$81,$BF,$9E,$69,$98,$FE,$0B,$19,$6C,$66
	.dc.b	$3F,$83,$99,$2B,$97,$F8,$4C,$78,$5F,$19,$FD,$6A,$3B,$C6,$63,$F5
	.dc.b	$C7,$EC,$A7,$E5,$3E,$B3,$FC,$1A,$7B,$8A,$74,$3F,$90,$1C,$8F,$21
	.dc.b	$E2,$3C,$11,$EA,$3C,$13,$E2,$8C,$9E,$A8,$C9,$B7,$27,$33,$BC,$71
	.dc.b	$6C,$6D,$A7,$33,$89,$FC,$36,$3E,$33,$63,$BB,$B3,$89,$DC,$FE,$88
	.dc.b	$7B,$63,$8C,$FC,$68,$F8,$64,$8F,$F1,$B3,$67,$C8,$F2,$8F,$43,$63
	.dc.b	$2F,$F1,$A9,$FE,$3E,$FE,$5C,$FC,$89,$F9,$73,$F2,$B1,$F9,$77,$F3
	.dc.b	$E6,$CC,$9F,$97,$3C,$47,$F8,$2C,$FB,$53,$E2,$7C,$07,$E0,$4F,$D4
	.dc.b	$1D,$1F,$D1,$3D,$0F,$11,$98,$F1,$3F,$87,$8F,$82,$48,$FF,$16,$47
	.dc.b	$CB,$3B,$9F,$86,$3E,$91,$F5,$A7,$E6,$4F,$18,$F4,$1D,$67,$E3,$BF
	.dc.b	$CD,$A7,$E3,$CF,$9A,$3C,$87,$C1,$79,$3F,$E4,$33,$E0,$CF,$27,$F9
	.dc.b	$69,$FC,$7C,$FE,$0A,$53,$E8,$94,$F8,$45,$3B,$9F,$10,$F3,$1E,$83
	.dc.b	$E8,$9F,$04,$F4,$1F,$3C,$FB,$07,$23,$A1,$C8,$F4,$14,$F8,$E7,$E5
	.dc.b	$8A,$7E,$2C,$FC,$D1,$4F,$01,$EE,$9F,$8F,$1F,$62,$3C,$A7,$F9,$18
	.dc.b	$F8,$A7,$23,$EB,$CF,$E8,$1F,$C2,$9E,$33,$67,$F9,$44,$C9,$D0,$C9
	.dc.b	$FE,$2F,$3D,$E7,$12,$1F,$AD,$8F,$D9,$BE,$88,$FE,$A7,$3E,$88,$FB
	.dc.b	$93,$F8,$C7,$D0,$7C,$58,$FC,$4C,$66,$3C,$27,$F3,$27,$AB,$F2,$0F
	.dc.b	$6B,$88,$E3,$1F,$66,$7F,$12,$FA,$5F,$C7,$3E,$87,$2F,$53,$F3,$47
	.dc.b	$C2,$72,$EE,$7F,$21,$1E,$C8,$FC,$13,$D1,$F9,$06,$CF,$A4,$7E,$20
	.dc.b	$FD,$44,$F9,$9F,$C2,$9C,$0F,$94,$78,$8F,$9E,$66,$3C,$6E,$8F,$C5
	.dc.b	$47,$37,$E5,$9B,$9F,$39,$FD,$D9,$FC,$A9,$93,$E3,$3A,$7F,$53,$1F
	.dc.b	$5D,$FD,$21,$B3,$CF,$3F,$C4,$5F,$63,$F8,$D7,$CE,$F9,$89,$3D,$63
	.dc.b	$F6,$73,$D5,$F1,$CE,$9F,$19,$F2,$0F,$D0,$1F,$3A,$7F,$41,$1B,$8F
	.dc.b	$1C,$75,$7F,$7B,$3D,$48,$F1,$3F,$22,$76,$38,$9D,$E7,$EE,$3F,$58
	.dc.b	$FC,$14,$FE,$4D,$FB,$87,$E7,$5F,$99,$3D,$A3,$81,$F0,$A3,$83,$F2
	.dc.b	$4F,$CC,$1F,$21,$FC,$C3,$D8,$F2,$1F,$6A,$7E,$73,$F0,$8E,$AF,$E3
	.dc.b	$CF,$84,$EE,$3E,$13,$FC,$4E,$7C,$93,$F5,$4E,$13,$E4,$78,$C7,$E2
	.dc.b	$67,$F5,$E7,$EC,$CF,$21,$F1,$88,$7A,$8F,$24,$75,$9F,$AA,$75,$24
	.dc.b	$7C,$D7,$E3,$BF,$4C,$FC,$6C,$7A,$9F,$E1,$F3,$E3,$7E,$79,$FB,$99
	.dc.b	$EA,$FC,$63,$67,$98,$F8,$47,$CA,$8E,$4F,$D1,$3F,$7C,$7E,$A0,$E0
	.dc.b	$7C,$03,$F9,$5B,$C2,$7C,$91,$FE,$27,$3F,$9A,$1F,$05,$FC,$39,$F0
	.dc.b	$5F,$C6,$9F,$18,$F2,$9F,$83,$3D,$07,$CD,$30,$6C,$C3,$F5,$9D,$9E
	.dc.b	$43,$CE,$7C,$43,$D6,$79,$23,$D1,$3F,$42,$3F,$9E,$1D,$8F,$8A,$7D
	.dc.b	$A8,$DB,$DA,$7F,$1E,$7B,$27,$F2,$B1,$C4,$F8,$D1,$FC,$B2,$78,$C7
	.dc.b	$A4,$FE,$28,$78,$CF,$39,$F4,$4F,$94,$7F,$38,$3F,$8F,$47,$BD,$E4
	.dc.b	$F5,$78,$BF,$49,$EA,$71,$3E,$D1,$E5,$3E,$21,$D8,$FD,$71,$DA,$7F
	.dc.b	$12,$F6,$7F,$A9,$4F,$AA,$3E,$29,$F8,$F3,$ED,$98,$3E,$DC,$F9,$DC
	.dc.b	$1F,$62,$7A,$B8,$3E,$C1,$E5,$3E,$11,$EA,$3C,$A7,$AA,$3A,$3F,$3A
	.dc.b	$3C,$AF,$C2,$7E,$E3,$E5,$3F,$52,$E1,$F2,$BE,$27,$90,$3F,$07,$1D
	.dc.b	$9F,$8E,$78,$CF,$DD,$C7,$E0,$23,$51,$E8,$7F,$0B,$1F,$BA,$8F,$E2
	.dc.b	$6F,$77,$9C,$69,$FE,$96,$7E,$FC,$FE,$7F,$1E,$77,$E4,$1F,$38,$F3
	.dc.b	$9E,$67,$A3,$E1,$7F,$8E,$BD,$9F,$8A,$FC,$D8,$FC,$31,$E1,$7F,$A7
	.dc.b	$9E,$19,$FE,$47,$3E,$D3,$AC,$F0,$7A,$CF,$73,$E4,$1F,$3C,$F4,$9F
	.dc.b	$A7,$7C,$31,$F1,$8E,$51,$FC,$7C,$F8,$27,$C6,$3D,$C7,$A4,$D1,$E6
	.dc.b	$34,$7D,$C3,$B1,$E2,$3F,$7E,$7E,$84,$FD,$71,$F0,$63,$E6,$3F,$6E
	.dc.b	$3F,$74,$FE,$84,$F0,$9E,$D3,$FA,$D1,$E6,$7F,$8D,$3E,$13,$F4,$C7
	.dc.b	$84,$FD,$61,$F8,$63,$A1,$B3,$F4,$0F,$EE,$5F,$21,$EA,$3B,$9E,$F7
	.dc.b	$BB,$F3,$0E,$E7,$E0,$4F,$3B,$F3,$1E,$27,$F7,$E3,$EE,$10,$F9,$27
	.dc.b	$33,$C7,$1F,$58,$FB,$E7,$C6,$8F,$C3,$47,$33,$8C,$FE,$19,$F6,$9C
	.dc.b	$CF,$CA,$9F,$83,$74,$FE,$E6,$7A,$BE,$F3,$FC,$FB,$1F,$C3,$5F,$E6
	.dc.b	$ED,$7F,$A6,$35,$FE,$52,$D9,$FF,$67,$B6,$7F,$DA,$6D,$9F,$F4,$3B
	.dc.b	$5F,$F7,$03,$5F,$09,$FC,$24,$FB,$47,$63,$07,$CF,$38,$9F,$95,$30
	.dc.b	$72,$3F,$10,$7C,$13,$F0,$73,$F5,$5D,$3F,$A3,$9F,$A1,$1E,$63,$A1
	.dc.b	$C4,$FC,$B3,$F4,$CF,$6B,$E4,$35,$1F,$59,$FA,$33,$F2,$9E,$65,$3F
	.dc.b	$1E,$7A,$1F,$D3,$BE,$07,$51,$E7,$8B,$3F,$61,$E0,$7B,$1F,$23,$E9
	.dc.b	$3A,$9E,$47,$F8,$3C,$7F,$8A,$A3,$ED,$9E,$42,$9C,$0E,$E7,$E7,$23
	.dc.b	$E2,$BF,$40,$F9,$A6,$8F,$D6,$1F,$CF,$0F,$14,$7A,$5E,$AF,$CF,$7E
	.dc.b	$EB,$F8,$87,$E5,$1C,$CF,$CD,$1F,$CC,$0F,$03,$CD,$FC,$21,$E7,$3A
	.dc.b	$C7,$C5,$7E,$01,$FC,$60,$FD,$69,$F1,$4F,$E0,$D3,$F8,$E7,$FA,$83
	.dc.b	$53,$48,$ED,$AF,$E5,$9A,$F4,$34,$43,$45,$34,$60,$86,$08,$7F,$2B
	.dc.b	$7A,$B4,$E4,$7E,$48,$F8,$72,$9F,$E6,$4F,$F4,$43,$11,$4F,$8E,$7D
	.dc.b	$68,$DB,$FD,$C6,$36,$D3,$67,$D2,$8D,$A8,$DB,$4D,$9F,$AE,$8D,$C6
	.dc.b	$8F,$BD,$3F,$E4,$47,$CD,$1A,$3D,$8F,$E1,$67,$B4,$65,$F5,$1F,$01
	.dc.b	$D9,$97,$D0,$64,$F7,$1C,$23,$C4,$F9,$A7,$C4,$FC,$19,$E4,$FA,$4E
	.dc.b	$27,$49,$E2,$F5,$3E,$21,$D6,$38,$CF,$BA,$3A,$1F,$DD,$9F,$7B,$B2
	.dc.b	$1C,$88,$7C,$B3,$E0,$9A,$3C,$86,$A7,$9B,$E5,$9F,$86,$FA,$0E,$51
	.dc.b	$E8,$3E,$61,$23,$DE,$F9,$E3,$FC,$CA,$F4,$78,$C7,$47,$C0,$7B,$0F
	.dc.b	$01,$DE,$7E,$5B,$E8,$32,$7B,$1C,$CF,$8E,$7A,$4F,$B0,$EE,$7B,$E7
	.dc.b	$C9,$3F,$8E,$7C,$8C,$8F,$B4,$7F,$A8,$5C,$3F,$54,$E8,$E0,$F8,$0F
	.dc.b	$C5,$71,$3E,$29,$F2,$BF,$6A,$7A,$9E,$B3,$F8,$29,$F0,$4F,$39,$EE
	.dc.b	$3D,$51,$C9,$F5,$1C,$A7,$D4,$F1,$7C,$2F,$C9,$3D,$46,$4E,$13,$FC
	.dc.b	$05,$F1,$4F,$07,$C0,$FC,$68,$D4,$FC,$D9,$F0,$9F,$0E,$7E,$E1,$C5
	.dc.b	$F4,$9E,$83,$C2,$7B,$E3,$A3,$F9,$93,$E0,$47,$18,$CC,$7F,$2A,$32
	.dc.b	$7A,$4E,$47,$A8,$E4,$7A,$8D,$3E,$77,$F0,$C7,$89,$C3,$E5,$3F,$2A
	.dc.b	$70,$9F,$CC,$BD,$4F,$A5,$1E,$57,$07,$B9,$C3,$F0,$CE,$47,$EB,$5F
	.dc.b	$89,$3F,$D4,$DD,$9E,$27,$11,$C0,$EE,$76,$8D,$CF,$E2,$63,$F2,$51
	.dc.b	$87,$E9,$BF,$92,$78,$9F,$18,$8E,$5F,$C7,$98,$8F,$7B,$4F,$C4,$14
	.dc.b	$E7,$1E,$77,$F9,$9B,$E1,$7F,$0E,$60,$F5,$1F,$14,$F5,$18,$7E,$1B
	.dc.b	$B3,$B1,$B3,$E5,$18,$7B,$3F,$5D,$FB,$26,$27,$E4,$C6,$CF,$0B,$88
	.dc.b	$DC,$60,$F0,$38,$7C,$31,$B9,$F8,$73,$EF,$9F,$9E,$7A,$4E,$46,$0F
	.dc.b	$01,$C8,$EB,$1C,$8F,$B1,$1F,$44,$F5,$9F,$9E,$8F,$80,$F8,$CF,$AD
	.dc.b	$3A,$8E,$F1,$C5,$F1,$1D,$CF,$13,$C6,$7B,$9F,$C8,$9E,$51,$F3,$1D
	.dc.b	$3F,$95,$3E,$A9,$EA,$7E,$01,$DD,$C4,$74,$7F,$8F,$1E,$28,$DC,$77
	.dc.b	$8C,$3B,$7F,$9E,$9C,$8F,$CB,$BB,$70,$7E,$28,$F5,$BF,$25,$D9,$E0
	.dc.b	$36,$7F,$0E,$9F,$B0,$FA,$CF,$89,$3E,$93,$DE,$F5,$71,$3D,$63,$0F
	.dc.b	$8E,$3F,$09,$1F,$EC,$17,$FC,$74,$47,$FC,$7C,$D7,$FC,$A2,$60,$86
	.dc.b	$08,$73,$32,$53,$27,$F8,$88,$C1,$0C,$10,$FF,$0B,$99,$29,$93,$F6
	.dc.b	$F1,$ED,$7F,$1A,$FF,$BA,$FF,$08,$FE,$E4,$7E,$12,$7E,$2C,$FF,$38
	.dc.b	$8C,$BF,$68,$F9,$91,$A8,$CC,$7D,$28,$E5,$3D,$63,$FA,$D4,$E1,$E3
	.dc.b	$1F,$E2,$28,$E3,$3E,$03,$52,$11,$A9,$F5,$4E,$A7,$27,$63,$CC,$7C
	.dc.b	$03,$B9,$CC,$D9,$FD,$BE,$79,$47,$F0,$C7,$B4,$70,$39,$10,$E3,$1F
	.dc.b	$E5,$48,$E3,$1E,$03,$33,$F0,$5F,$88,$66,$7A,$BE,$D9,$E2,$FE,$69
	.dc.b	$CB,$F5,$8F,$11,$4E,$07,$BC,$EA,$48,$F2,$39,$7E,$63,$E1,$31,$3E
	.dc.b	$58,$CC,$71,$8C,$11,$F0,$C7,$CF,$9F,$2B,$E8,$3F,$B4,$B8,$8E,$C7
	.dc.b	$AC,$EE,$72,$33,$1E,$67,$E9,$18,$9F,$BA,$F0,$3D,$E7,$94,$E4,$70
	.dc.b	$24,$7A,$89,$1E,$98,$E3,$18,$78,$BE,$12,$47,$82,$3B,$1C,$1E,$4F
	.dc.b	$D1,$24,$7F,$87,$67,$E0,$47,$03,$4E,$23,$C1,$1F,$89,$8C,$4F,$79
	.dc.b	$FD,$54,$FB,$CE,$A7,$C3,$3F,$31,$3C,$1F,$4C,$FE,$EE,$31,$1F,$95
	.dc.b	$9F,$CD,$3C,$DF,$AD,$3F,$AF,$8C,$1D,$4F,$21,$83,$D5,$1E,$D7,$0F
	.dc.b	$A9,$F6,$98,$8F,$51,$E3,$70,$79,$8F,$CC,$CF,$99,$F6,$9F,$85,$3E
	.dc.b	$CC,$72,$78,$1F,$44,$F8,$27,$9C,$E3,$3F,$61,$C9,$C8,$AF,$9C,$FD
	.dc.b	$5C,$C8,$E5,$15,$E0,$77,$3E,$51,$FE,$18,$8D,$47,$78,$C9,$FE,$16
	.dc.b	$8C,$CF,$E1,$A7,$F4,$D3,$F6,$E7,$E0,$47,$D4,$8E,$E7,$F0,$63,$CF
	.dc.b	$3F,$05,$FF,$3F,$4F,$47,$27,$C4,$3F,$9A,$1D,$4F,$9E,$79,$27,$CC
	.dc.b	$F5,$3D,$85,$3E,$4C,$FE,$0D,$F1,$CF,$99,$FE,$34,$F8,$5E,$33,$EC
	.dc.b	$76,$78,$0E,$67,$98,$EF,$1C,$DF,$BD,$1E,$48,$C3,$FA,$08,$E6,$7C
	.dc.b	$47,$B1,$B7,$B3,$F2,$4C,$3B,$3F,$09,$1E,$77,$DD,$1C,$A7,$F7,$33
	.dc.b	$E0,$7F,$2D,$3E,$09,$F8,$E6,$DF,$51,$F1,$E7,$F5,$2E,$1E,$D1,$B8
	.dc.b	$F7,$9E,$63,$E4,$9F,$54,$F3,$1F,$C1,$63,$CC,$74,$78,$9E,$39,$FC
	.dc.b	$13,$F7,$8C,$1E,$49,$F0,$3E,$53,$C0,$7C,$E8,$F7,$B8,$8F,$9C,$7C
	.dc.b	$83,$DE,$FC,$83,$CC,$E1,$F3,$18,$8F,$CE,$3B,$3C,$67,$E1,$0C,$47
	.dc.b	$58,$DC,$72,$36,$7D,$43,$6F,$99,$F8,$B1,$FC,$36,$7E,$2C,$7E,$D8
	.dc.b	$F1,$1F,$44,$C1,$D1,$FE,$F8,$C3,$FB,$E9,$5C,$3F,$C1,$8C,$18,$36
	.dc.b	$7F,$77,$32,$70,$74,$E8,$E4,$6C,$86,$08,$7F,$58,$39,$90,$D9,$92
	.dc.b	$9D,$4C,$1C,$8F,$BC,$7B,$CF,$E8,$DF,$33,$FE,$1B,$95,$7C,$28,$F9
	.dc.b	$33,$B1,$3A,$7F,$C4,$51,$C6,$78,$47,$CE,$3F,$27,$3F,$6D,$D9,$FE
	.dc.b	$33,$79,$3F,$E3,$59,$EF,$1B,$8F,$E1,$F1,$D5,$E2,$78,$E3,$4F,$F1
	.dc.b	$38,$D3,$CC,$F1,$1F,$02,$37,$1A,$83,$E0,$9F,$C5,$4F,$F2,$08,$D1
	.dc.b	$99,$FC,$41,$FC,$C1,$CB,$DA,$B4,$F8,$4F,$8F,$5F,$11,$F4,$3D,$8F
	.dc.b	$F0,$1C,$F2,$8C,$9D,$AB,$C4,$FC,$79,$E5,$1F,$02,$BB,$CF,$A8,$D1
	.dc.b	$F1,$A3,$4F,$94,$C9,$E5,$7E,$8B,$C4,$F0,$9F,$62,$7C,$AF,$03,$51
	.dc.b	$D1,$D9,$88,$F2,$9E,$D7,$D4,$D7,$C0,$E9,$C9,$0F,$01,$E9,$74,$E0
	.dc.b	$87,$DB,$72,$F8,$4E,$46,$4B,$5F,$36,$7D,$04,$9E,$31,$F2,$AB,$FA
	.dc.b	$03,$B8,$F4,$57,$33,$C3,$19,$7B,$CF,$D2,$8C,$47,$9A,$30,$FC,$98
	.dc.b	$EC,$FE,$51,$DB,$D4,$91,$FE,$78,$8C,$BC,$23,$F2,$07,$33,$EE,$1F
	.dc.b	$4A,$36,$E2,$BF,$3E,$EC,$FA,$F3,$EE,$71,$1F,$0D,$F5,$98,$AF,$9B
	.dc.b	$3E,$19,$FC,$9C,$FE,$70,$FA,$C7,$03,$C4,$71,$30,$F5,$7F,$02,$70
	.dc.b	$3E,$31,$E7,$9F,$A6,$D3,$8B,$87,$D2,$7C,$93,$B1,$C8,$F0,$9E,$53
	.dc.b	$B9,$98,$F1,$3F,$87,$32,$52,$4F,$D8,$79,$19,$AF,$9A,$F6,$9F,$D4
	.dc.b	$C7,$DA,$3E,$35,$79,$E3,$33,$E6,$3E,$E9,$FC,$C0,$C4,$F4,$8F,$2B
	.dc.b	$CE,$38,$57,$8E,$79,$C7,$92,$78,$57,$E5,$AB,$84,$E4,$E2,$7C,$B3
	.dc.b	$DF,$3A,$AF,$87,$36,$BD,$07,$3A,$F3,$1C,$C9,$3E,$23,$8B,$EA,$21
	.dc.b	$F5,$8D,$C7,$47,$F0,$07,$C2,$3F,$36,$68,$CD,$7C,$A6,$9D,$27,$BB
	.dc.b	$B3,$D6,$7B,$8F,$A9,$3D,$DF,$BB,$5F,$26,$31,$1F,$78,$F8,$87,$DD
	.dc.b	$3B,$3F,$05,$F2,$CF,$BE,$38,$CF,$39,$F2,$9E,$33,$D9,$3F,$A6,$38
	.dc.b	$3A,$7D,$26,$1F,$49,$C6,$3A,$38,$8D,$3D,$27,$F9,$9B,$DE,$BF,$02
	.dc.b	$FE,$98,$ED,$1E,$67,$DB,$1F,$75,$FA,$35,$D9,$F8,$07,$84,$D9,$F0
	.dc.b	$8D,$98,$3C,$07,$CF,$AF,$CB,$BB,$3E,$79,$B7,$D3,$3E,$68,$FD,$95
	.dc.b	$7D,$28,$C4,$7B,$4F,$49,$0D,$9F,$E6,$A7,$D4,$F6,$3E,$F9,$B3,$EE
	.dc.b	$18,$3E,$54,$7F,$1A,$7C,$8F,$85,$FC,$0D,$74,$7E,$1D,$7D,$69,$D4
	.dc.b	$7C,$53,$A1,$A7,$A3,$8A,$E5,$5F,$35,$FE,$F9,$1F,$E4,$B7,$FC,$82
	.dc.b	$47,$FC,$90,$D7,$FC,$9A,$D7,$81,$A2,$1A,$29,$A3,$04,$30,$43,$FC
	.dc.b	$24,$53,$FC,$56,$53,$FC,$30,$53,$DC,$7C,$D7,$FA,$BF,$E1,$1F,$E2
	.dc.b	$93,$E4,$32,$3F,$C0,$D1,$97,$F0,$87,$23,$E9,$82,$3B,$7F,$1A,$4B
	.dc.b	$FC,$4B,$D0,$86,$6F,$EA,$C7,$E9,$A7,$93,$C2,$74,$7E,$16,$33,$3C
	.dc.b	$09,$1E,$31,$1A,$7E,$01,$EA,$AC,$3E,$83,$04,$3F,$A3,$C6,$9F,$80
	.dc.b	$48,$F9,$2F,$AA,$BC,$C4,$AF,$E2,$B5,$C6,$BE,$65,$7A,$8F,$01,$D8
	.dc.b	$F0,$57,$17,$F9,$6C,$F1,$73,$5D,$1F,$03,$F0,$1F,$A6,$7F,$62,$8E
	.dc.b	$0F,$73,$D0,$70,$3E,$33,$E4,$9F,$04,$71,$34,$74,$38,$9F,$C9,$63
	.dc.b	$AC,$70,$8C,$BF,$E5,$28,$CC,$F9,$67,$E4,$C6,$1F,$A6,$79,$5C,$3D
	.dc.b	$EB,$F7,$51,$88,$EB,$3F,$05,$F9,$07,$B6,$7D,$51,$FE,$C0,$9E,$11
	.dc.b	$89,$F9,$67,$B0,$E6,$78,$AB,$D0,$E6,$38,$3F,$06,$79,$BF,$48,$E6
	.dc.b	$78,$A7,$E1,$4F,$08,$FC,$89,$CC,$E0,$73,$30,$7A,$4F,$D6,$CF,$67
	.dc.b	$07,$CF,$37,$3D,$9D,$98,$AF,$9D,$5F,$94,$8F,$24,$61,$D9,$83,$E0
	.dc.b	$3F,$56,$BC,$15,$D6,$7C,$06,$CF,$A8,$7A,$8F,$69,$93,$E5,$9A,$31
	.dc.b	$1F,$0E,$BE,$DC,$79,$CF,$09,$FB,$B3,$07,$19,$FC,$64,$FE,$C4,$F2
	.dc.b	$D7,$E0,$63,$EB,$CF,$63,$C9,$5B,$7E,$F5,$78,$5A,$43,$E6,$CF,$F0
	.dc.b	$F7,$0F,$43,$E1,$19,$7B,$1E,$F9,$FB,$D3,$DE,$30,$7C,$57,$13,$CE
	.dc.b	$BE,$94,$F2,$8C,$3F,$B2,$3D,$27,$8A,$3F,$34,$FC,$43,$E6,$9F,$42
	.dc.b	$38,$BF,$C2,$9A,$EA,$BC,$33,$CA,$3C,$A7,$C0,$71,$3C,$2B,$E6,$CF
	.dc.b	$D3,$3C,$D3,$A3,$D3,$3B,$8F,$31,$F6,$8F,$E1,$07,$03,$D8,$77,$3F
	.dc.b	$4E,$F1,$7C,$6F,$CE,$8E,$D1,$C2,$34,$65,$E3,$1A,$7E,$03,$5F,$80
	.dc.b	$79,$0C,$9C,$DF,$49,$B9,$F8,$B1,$E9,$3F,$D1,$0E,$1E,$27,$E4,$E3
	.dc.b	$FD,$AB,$19,$8F,$A5,$1D,$1E,$C4,$38,$9A,$3B,$57,$19,$F9,$F1,$FC
	.dc.b	$0C,$F9,$07,$08,$FC,$6B,$EB,$39,$1E,$33,$E1,$9D,$CC,$91,$E2,$F4
	.dc.b	$30,$7E,$54,$E9,$1F,$A5,$74,$F9,$1F,$E6,$E6,$8F,$CA,$9A,$3F,$29
	.dc.b	$1F,$BC,$70,$7D,$F3,$F0,$75,$F6,$AB,$EB,$4F,$92,$31,$3E,$98,$C4
	.dc.b	$FD,$58,$EA,$7F,$18,$9D,$BF,$14,$F5,$1E,$07,$F0,$CF,$82,$39,$3F
	.dc.b	$E0,$33,$F2,$65,$3F,$8C,$96,$B0,$E8,$FD,$41,$93,$27,$03,$24,$38
	.dc.b	$10,$FE,$FA,$71,$3A,$BC,$9E,$46,$CA,$70,$3A,$14,$D3,$C4,$8F,$D5
	.dc.b	$3E,$F7,$E1,$1F,$E2,$1A,$57,$C1,$8F,$C8,$4E,$E7,$2F,$F8,$9E,$38
	.dc.b	$4E,$E3,$4F,$CA,$9F,$E1,$B3,$A9,$E0,$68,$FC,$95,$7F,$89,$E3,$51
	.dc.b	$FE,$45,$7C,$2F,$D7,$AE,$CF,$FB,$D2,$33,$5F,$7A,$3F,$37,$1C,$93
	.dc.b	$A9,$F2,$1F,$C0,$5D,$19,$8F,$19,$FD,$FD,$CB,$E7,$3F,$6B,$5E,$77
	.dc.b	$FC,$B5,$5F,$11,$F4,$1B,$3F,$5B,$1B,$7D,$06,$5F,$E0,$93,$DA,$3D
	.dc.b	$C7,$88,$FE,$CC,$E6,$7D,$D3,$E7,$9E,$E7,$43,$89,$E6,$3D,$47,$9A
	.dc.b	$BE,$4C,$F2,$8F,$35,$75,$9F,$8B,$1E,$19,$F9,$93,$98,$C4,$E9,$EC
	.dc.b	$78,$67,$E1,$3F,$40,$F7,$1F,$C4,$E7,$93,$C2,$3C,$4F,$B8,$F3,$1F
	.dc.b	$86,$3D,$67,$CA,$3D,$67,$79,$F1,$39,$3F,$36,$72,$3E,$2C,$76,$24
	.dc.b	$7C,$22,$47,$B0,$87,$95,$CB,$C9,$EF,$5F,$46,$3D,$C7,$8C,$E2,$78
	.dc.b	$E7,$E4,$BE,$33,$6F,$89,$DC,$7C,$47,$33,$D9,$FA,$A7,$1A,$F0,$4F
	.dc.b	$E2,$23,$07,$CE,$3E,$09,$E3,$3D,$47,$63,$E4,$D7,$D6,$9E,$71,$F5
	.dc.b	$4E,$67,$D0,$9E,$CF,$33,$D9,$5E,$57,$EC,$19,$8E,$6E,$2B,$B3,$E5
	.dc.b	$7E,$1B,$B3,$13,$FB,$37,$C3,$1F,$15,$F4,$9F,$D9,$2B,$E3,$57,$69
	.dc.b	$FE,$1B,$3C,$2B,$C9,$1C,$63,$E5,$9F,$8B,$7C,$6F,$84,$FC,$6D,$7D
	.dc.b	$13,$AB,$F4,$0F,$80,$E1,$D9,$DC,$E9,$3F,$86,$7E,$29,$F8,$28,$E8
	.dc.b	$EC,$F5,$9A,$3F,$18,$72,$21,$83,$75,$FC,$26,$BD,$24,$AF,$04,$7E
	.dc.b	$4E,$BB,$D7,$B6,$BE,$14,$61,$E1,$1F,$83,$3E,$B5,$7A,$67,$DD,$3C
	.dc.b	$27,$D0,$7C,$A3,$C8,$7A,$49,$1D,$67,$C7,$1F,$26,$7B,$BC,$27,$EC
	.dc.b	$4F,$78,$EB,$18,$7E,$09,$DC,$8E,$5E,$47,$03,$F1,$27,$07,$C4,$FB
	.dc.b	$67,$F0,$AF,$DA,$3D,$75,$F1,$DE,$47,$C7,$AF,$C4,$9E,$57,$81,$FA
	.dc.b	$33,$73,$EC,$7C,$D3,$CD,$DC,$66,$30,$EE,$BE,$6D,$76,$AF,$41,$D1
	.dc.b	$C3,$F9,$F3,$07,$C4,$AE,$71,$E5,$AE,$D3,$F3,$E3,$07,$D2,$7F,$2E
	.dc.b	$79,$1C,$47,$DD,$8C,$1D,$4E,$47,$90,$F8,$87,$CB,$8E,$AE,$63,$F5
	.dc.b	$D3,$F4,$23,$EA,$9F,$1C,$EA,$62,$36,$FB,$0C,$4F,$C0,$7F,$BC,$99
	.dc.b	$3E,$69,$E4,$36,$79,$CF,$6C,$7C,$D7,$EE,$9B,$3F,$12,$60,$DC,$7A
	.dc.b	$9F,$CD,$CF,$CC,$7D,$91,$F1,$8F,$25,$7B,$63,$C6,$72,$36,$78,$EB
	.dc.b	$E6,$1E,$37,$FC,$8B,$1F,$E3,$92,$3F,$E4,$46,$BA,$30,$43,$04,$3F
	.dc.b	$CA,$4F,$06,$9B,$29,$B3,$FC,$4E,$60,$86,$08,$7F,$6F,$32,$7B,$0F
	.dc.b	$9E,$41,$F8,$47,$F8,$B4,$F8,$91,$23,$FA,$F4,$65,$FB,$51,$95,$3F
	.dc.b	$E0,$69,$D4,$7E,$18,$FA,$E5,$73,$1F,$2C,$95,$F6,$67,$C6,$78,$0E
	.dc.b	$E4,$3B,$BF,$E7,$B9,$E6,$F1,$AF,$25,$78,$EB,$91,$E3,$8C,$CF,$18
	.dc.b	$CC,$F8,$9F,$8C,$7A,$1E,$4F,$80,$12,$B9,$BC,$2B,$C6,$E8,$F5,$9E
	.dc.b	$13,$D5,$3D,$09,$5E,$98,$CB,$E5,$7C,$07,$73,$27,$73,$D2,$EE,$3D
	.dc.b	$51,$F7,$67,$FC,$25,$5E,$48,$C1,$C8,$E8,$76,$3C,$F3,$C9,$F8,$55
	.dc.b	$F5,$67,$94,$60,$F1,$1F,$E0,$B7,$35,$F1,$23,$F1,$D1,$87,$C0,$71
	.dc.b	$3C,$35,$F0,$9E,$6F,$86,$7B,$CF,$39,$E0,$7F,$4B,$7A,$3F,$34,$FC
	.dc.b	$71,$F1,$0E,$75,$F3,$5F,$74,$70,$AE,$D1,$FC,$56,$31,$3B,$9F,$85
	.dc.b	$18,$9E,$67,$E1,$9C,$3F,$30,$F0,$9D,$8F,$7D,$7F,$0F,$8D,$C7,$AA
	.dc.b	$30,$F9,$CE,$53,$E4,$79,$11,$C1,$2B,$F0,$04,$8F,$29,$C9,$C3,$F3
	.dc.b	$9F,$61,$1F,$CB,$57,$92,$BE,$71,$F4,$23,$D2,$7C,$53,$CC,$7A,$4F
	.dc.b	$B7,$1C,$5D,$9E,$79,$F0,$BB,$3E,$B5,$7D,$F7,$C8,$7A,$0D,$39,$7F
	.dc.b	$27,$1E,$77,$D8,$7A,$0F,$1D,$7A,$1F,$F0,$ED,$78,$DE,$F3,$FE,$3B
	.dc.b	$8F,$64,$79,$4F,$04,$E2,$BC,$13,$CE,$74,$79,$88,$79,$0E,$27,$C4
	.dc.b	$79,$57,$9E,$3F,$2C,$64,$F3,$9F,$20,$C1,$D8,$F8,$2F,$9D,$CC,$FC
	.dc.b	$3A,$FD,$15,$76,$9F,$84,$7B,$8F,$F1,$7B,$97,$CF,$1C,$1E,$75,$F5
	.dc.b	$23,$11,$E1,$9F,$91,$5F,$A0,$8D,$9D,$9F,$C8,$3D,$8C,$9D,$8F,$61
	.dc.b	$CC,$F7,$9E,$53,$E4,$C7,$95,$F7,$9C,$CE,$07,$C3,$9F,$2C,$FD,$48
	.dc.b	$E6,$F8,$9E,$67,$88,$E6,$7F,$97,$27,$B1,$F6,$E3,$94,$76,$79,$3F
	.dc.b	$58,$C5,$7C,$57,$D7,$1C,$0F,$A4,$E2,$3F,$3B,$3F,$21,$FE,$4D,$1E
	.dc.b	$38,$F0,$46,$A3,$CD,$5E,$C8,$C4,$7B,$CE,$87,$CF,$73,$5E,$69,$F5
	.dc.b	$CF,$C2,$9F,$C7,$56,$23,$E7,$1E,$BA,$F4,$4F,$AE,$33,$3B,$9E,$91
	.dc.b	$E7,$3A,$1F,$B3,$8D,$4F,$E6,$EA,$CF,$94,$FD,$09,$F2,$0C,$1F,$83
	.dc.b	$3B,$9D,$4F,$2B,$C9,$D5,$7D,$2A,$C4,$7D,$A3,$D5,$5F,$08,$F5,$4E
	.dc.b	$E3,$F5,$27,$F6,$F3,$FA,$D9,$4F,$EE,$25,$9C,$3F,$A8,$30,$60,$D9
	.dc.b	$FD,$C4,$C9,$C1,$D3,$A3,$91,$B2,$18,$21,$E4,$32,$53,$27,$33,$04
	.dc.b	$30,$7F,$0F,$3F,$3E,$7C,$F3,$F8,$9C,$FF,$83,$E4,$FC,$33,$E5,$CF
	.dc.b	$EB,$DD,$19,$9F,$A9,$3F,$86,$7E,$C9,$C8,$FF,$0F,$B5,$FF,$10,$C7
	.dc.b	$47,$F6,$0F,$47,$FC,$40,$E5,$F3,$1C,$63,$93,$63,$EE,$3E,$09,$F3
	.dc.b	$3E,$08,$F3,$3C,$A3,$F0,$B1,$DE,$3C,$67,$A9,$E6,$FC,$43,$91,$E4
	.dc.b	$34,$77,$3F,$AD,$BC,$DF,$21,$93,$C0,$68,$E6,$78,$0C,$92,$7D,$B1
	.dc.b	$CC,$F8,$4F,$CD,$8E,$0E,$9F,$19,$F7,$1F,$EC,$11,$C6,$32,$7E,$04
	.dc.b	$93,$97,$FC,$71,$1D,$8F,$82,$FA,$8E,$EE,$23,$E4,$C7,$A8,$D1,$D0
	.dc.b	$EE,$F5,$9F,$34,$7B,$67,$93,$C5,$D1,$87,$47,$CF,$3C,$A7,$43,$F0
	.dc.b	$E7,$13,$07,$E1,$9E,$6F,$D1,$8F,$90,$F9,$4E,$64,$73,$3E,$C8,$C4
	.dc.b	$F8,$E7,$31,$DC,$C9,$D6,$34,$7F,$85,$8F,$0B,$F4,$DC,$BF,$A4,$72
	.dc.b	$F8,$E3,$AB,$FD,$E6,$33,$3F,$4E,$38,$1E,$19,$D9,$27,$F3,$AF,$CE
	.dc.b	$31,$1E,$53,$89,$E8,$3C,$86,$4A,$7D,$E3,$E8,$3C,$9C,$9F,$89,$8F
	.dc.b	$2B,$F4,$23,$F1,$EF,$96,$3D,$EF,$09,$F7,$C7,$AE,$39,$BC,$9F,$BE
	.dc.b	$60,$E4,$79,$4F,$8C,$FE,$50,$E1,$3E,$58,$C9,$E8,$8F,$86,$78,$9C
	.dc.b	$9F,$BB,$3E,$11,$F9,$87,$C8,$D3,$C2,$7C,$33,$D4,$7E,$E4,$E2,$F8
	.dc.b	$8E,$C7,$B9,$E2,$78,$08,$E5,$F0,$CF,$E3,$E7,$EC,$4F,$A2,$7D,$47
	.dc.b	$C1,$9F,$F5,$3C,$71,$9D,$CE,$4F,$23,$87,$D0,$7E,$D2,$7B,$4F,$58
	.dc.b	$CB,$F0,$0E,$86,$5F,$44,$78,$1F,$5C,$7B,$DF,$01,$F0,$8C,$9F,$83
	.dc.b	$79,$1D,$1D,$CF,$E1,$9E,$07,$C2,$3F,$8A,$C6,$DF,$4C,$FA,$1F,$86
	.dc.b	$7A,$0E,$C6,$0F,$B0,$73,$3F,$16,$62,$7E,$23,$F0,$CE,$E7,$A0,$E0
	.dc.b	$FE,$2E,$7C,$51,$B3,$9B,$83,$E4,$3C,$63,$AC,$F9,$E3,$CC,$74,$9F
	.dc.b	$C2,$BB,$7E,$74,$6E,$3C,$D1,$87,$CA,$7C,$69,$EA,$E1,$F5,$C7,$D3
	.dc.b	$8F,$CA,$9B,$7E,$8B,$F1,$8F,$1C,$70,$70,$78,$0F,$95,$1E,$C7,$91
	.dc.b	$F3,$67,$CD,$1F,$11,$B1,$EB,$3E,$EC,$7F,$21,$70,$7D,$A3,$D0,$F9
	.dc.b	$1F,$88,$70,$3C,$27,$03,$F2,$27,$E7,$1E,$13,$F0,$A3,$E9,$9B,$8F
	.dc.b	$44,$7B,$63,$A9,$F1,$1C,$12,$3F,$82,$4F,$DB,$9F,$DB,$BE,$C7,$E0
	.dc.b	$9C,$63,$D8,$FE,$60,$FF,$22,$3F,$E3,$A2,$3F,$E4,$16,$BF,$E4,$66
	.dc.b	$BC,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$E1,$C2,$9F,$D9,$CA,$7F,$83
	.dc.b	$4A,$7D,$F3,$F0,$6F,$CC,$3F,$CC,$1F,$84,$7F,$86,$4F,$89,$1F,$08
	.dc.b	$FE,$FA,$E5,$FA,$04,$15,$FE,$34,$8D,$47,$19,$E7,$39,$8F,$F0,$44
	.dc.b	$E2,$7C,$51,$FB,$B8,$E3,$3F,$5E,$3F,$0A,$78,$E3,$33,$C2,$3D,$66
	.dc.b	$A7,$B3,$E4,$8F,$6B,$E7,$8E,$8F,$F2,$69,$E6,$F1,$9E,$8F,$C7,$11
	.dc.b	$A7,$B9,$0D,$39,$78,$47,$B5,$12,$35,$19,$8E,$51,$87,$CD,$1E,$28
	.dc.b	$C4,$7C,$B3,$F4,$CE,$27,$D7,$1E,$03,$E6,$90,$E0,$47,$E2,$92,$7D
	.dc.b	$E7,$8E,$3D,$E7,$18,$EB,$19,$9F,$03,$C2,$7C,$66,$DC,$3D,$DF,$A2
	.dc.b	$FD,$33,$E1,$9D,$8D,$1F,$44,$FC,$11,$CC,$C1,$F3,$DF,$F1,$4C,$78
	.dc.b	$A3,$27,$C1,$3B,$1F,$3C,$E4,$71,$3E,$69,$E2,$79,$39,$3F,$1C,$70
	.dc.b	$3A,$4F,$C2,$71,$1F,$1E,$7F,$A3,$BC,$A3,$71,$F6,$5E,$4E,$DF,$F0
	.dc.b	$2C,$6E,$31,$1E,$83,$E2,$4F,$D5,$7D,$B1,$F4,$5E,$67,$A8,$FC,$49
	.dc.b	$83,$E5,$98,$3A,$4F,$69,$FC,$3B,$E1,$8E,$47,$EF,$8E,$51,$EF,$7E
	.dc.b	$19,$C8,$DC,$FC,$07,$6F,$E0,$23,$AC,$60,$F4,$9F,$86,$39,$1E,$63
	.dc.b	$E3,$9F,$4C,$E0,$7D,$37,$9B,$C1,$F2,$12,$3F,$1A,$FC,$78,$FC,$01
	.dc.b	$B8,$F2,$1F,$E1,$D9,$FD,$B1,$F0,$63,$DB,$3C,$5F,$A3,$3D,$0E,$2F
	.dc.b	$E5,$0E,$87,$9C,$A7,$9D,$EC,$FE,$18,$F5,$9E,$E3,$CB,$1D,$E7,$C4
	.dc.b	$FB,$8C,$47,$09,$F9,$CF,$63,$CA,$78,$63,$E1,$3E,$57,$8B,$B8,$FC
	.dc.b	$43,$C6,$7B,$C7,$E5,$63,$07,$73,$EB,$4F,$23,$E2,$47,$27,$0F,$C5
	.dc.b	$8F,$BA,$7F,$9B,$27,$E6,$BC,$CF,$9B,$1B,$7E,$0B,$5F,$88,$7E,$44
	.dc.b	$F1,$9E,$93,$F2,$B1,$D1,$C1,$F8,$97,$13,$E6,$8D,$9F,$44,$DB,$F1
	.dc.b	$9F,$09,$F2,$67,$C5,$3F,$A5,$6C,$71,$9E,$0F,$95,$F8,$2F,$13,$E0
	.dc.b	$BF,$E4,$E9,$F1,$CF,$B4,$F9,$07,$49,$F1,$BF,$8C,$7C,$64,$3E,$E9
	.dc.b	$5E,$2F,$73,$51,$E7,$70,$70,$23,$97,$67,$AE,$3C,$0F,$AD,$F2,$4F
	.dc.b	$C9,$8F,$98,$6C,$EB,$1E,$27,$99,$EA,$3E,$0B,$C8,$8E,$0F,$CA,$38
	.dc.b	$7D,$87,$D6,$3E,$FC,$78,$4F,$9F,$3F,$2A,$3C,$26,$E3,$E8,$BC,$8F
	.dc.b	$39,$CA,$7E,$83,$B7,$C0,$FC,$57,$D4,$FC,$63,$F2,$D1,$FB,$18,$FA
	.dc.b	$D1,$FD,$58,$FE,$16,$53,$FA,$C1,$67,$0F,$F0,$33,$06,$0D,$9F,$D8
	.dc.b	$8C,$9C,$1D,$3A,$39,$1B,$21,$82,$1E,$43,$25,$32,$73,$30,$43,$07
	.dc.b	$C1,$3E,$41,$F8,$A7,$FB,$AC,$A7,$E1,$7E,$11,$86,$D7,$E1,$4F,$C1
	.dc.b	$CF,$D4,$53,$B7,$FB,$FB,$F6,$DD,$CF,$D4,$8D,$47,$F8,$99,$D3,$E5
	.dc.b	$9E,$91,$A8,$EA,$78,$4D,$4F,$37,$FC,$5F,$3E,$97,$C2,$6E,$BA,$9C
	.dc.b	$EA,$46,$48,$7F,$73,$72,$FC,$63,$D2,$7C,$73,$BD,$7C,$47,$81,$CA
	.dc.b	$BE,$25,$72,$7C,$91,$EF,$35,$3F,$96,$9D,$3C,$E7,$8C,$E8,$F5,$9D
	.dc.b	$8F,$21,$F0,$8F,$21,$97,$BB,$F2,$0D,$CF,$C2,$7D,$27,$48,$DD,$66
	.dc.b	$36,$68,$F4,$1D,$6B,$D0,$F9,$4C,$CF,$27,$07,$94,$FE,$3D,$3E,$E7
	.dc.b	$CF,$1E,$E7,$FA,$F4,$65,$F9,$E4,$39,$1E,$B8,$EC,$F9,$88,$E0,$93
	.dc.b	$E6,$8C,$BF,$34,$8F,$89,$D4,$7F,$80,$E7,$C5,$3E,$0A,$F8,$93,$D2
	.dc.b	$3E,$45,$70,$23,$82,$1E,$99,$E3,$1E,$08,$F8,$EF,$07,$47,$CF,$3D
	.dc.b	$CE,$DF,$11,$F3,$2B,$CB,$3D,$23,$D4,$78,$0E,$26,$0E,$A7,$CE,$3B
	.dc.b	$47,$17,$35,$ED,$7C,$A7,$13,$C2,$71,$3F,$04,$66,$79,$3E,$03,$89
	.dc.b	$D2,$BD,$CE,$49,$1C,$0F,$F5,$55,$7B,$23,$F1,$31,$87,$CE,$60,$FE
	.dc.b	$05,$1B,$7E,$E9,$E7,$71,$5F,$8D,$3F,$85,$C7,$23,$67,$94,$F9,$11
	.dc.b	$E3,$7A,$9B,$8F,$88,$FA,$0F,$A4,$66,$BE,$B4,$F1,$8F,$8A,$7C,$23
	.dc.b	$E0,$39,$9E,$6F,$E3,$8C,$4F,$29,$E8,$7D,$BA,$F9,$B3,$B8,$C9,$E9
	.dc.b	$3F,$5A,$60,$EE,$7C,$73,$C6,$60,$B1,$D1,$E1,$1F,$22,$7E,$BB,$DE
	.dc.b	$36,$7C,$03,$9B,$B7,$F0,$47,$22,$1F,$38,$FF,$0B,$3C,$EB,$E5,$D7
	.dc.b	$86,$79,$9E,$97,$11,$F1,$A3,$07,$85,$C3,$E4,$3B,$4F,$5A,$FD,$45
	.dc.b	$72,$7E,$64,$EE,$BF,$09,$18,$79,$98,$8F,$F1,$6D,$73,$7E,$91,$E1
	.dc.b	$9E,$07,$67,$07,$E2,$8F,$39,$E4,$3E,$1C,$FC,$07,$67,$E1,$67,$E5
	.dc.b	$3F,$6C,$FA,$D1,$E4,$9F,$D4,$D7,$8E,$78,$1F,$76,$74,$F8,$4F,$A1
	.dc.b	$1B,$79,$CF,$27,$C6,$FA,$9F,$B2,$7E,$F6,$79,$BF,$99,$9F,$94,$F5
	.dc.b	$8F,$53,$ED,$3E,$C9,$F3,$67,$F1,$2F,$63,$E4,$9B,$8E,$AF,$EA,$CC
	.dc.b	$47,$8D,$F0,$1F,$18,$F4,$9C,$A3,$EB,$3F,$08,$F0,$B8,$7F,$1E,$78
	.dc.b	$4E,$67,$63,$D5,$1F,$65,$C1,$CE,$3D,$A7,$3A,$F9,$47,$ED,$A3,$89
	.dc.b	$0F,$B1,$5A,$7F,$C8,$51,$FE,$3E,$23,$82,$1F,$E4,$A7,$2D,$32,$7F
	.dc.b	$94,$5C,$B4,$C9,$FE,$11,$30,$71,$21,$82,$1B,$21,$FE,$0C,$32,$7A
	.dc.b	$CF,$9E,$7C,$3A,$FF,$32,$7E,$11,$FE,$33,$3D,$D1,$F4,$23,$FA,$B4
	.dc.b	$66,$3E,$81,$C8,$8E,$55,$FF,$85,$BE,$7F,$FC,$C6,$48,$FF,$35,$35
	.dc.b	$FF,$37,$B5,$FF,$3A,$35,$FF,$3C,$35,$11,$C3,$4E,$A4,$3A,$94,$EA
	.dc.b	$53,$A9,$4E,$A1,$EA,$BC,$7F,$F3,$E9,$1F,$F4,$23,$5F,$F4,$4B,$5F
	.dc.b	$F4,$73,$5F,$F4,$93,$5E,$ED,$7D,$AD,$7C,$07,$62,$1D,$8A,$76,$29
	.dc.b	$D8,$A7,$B3,$CC,$FF,$4A,$91,$FF,$4E,$35,$FF,$50,$B5,$FF,$53,$35
	.dc.b	$FF,$55,$35,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$7F
	.dc.b	$AD,$3C,$2F,$F4,$41,$FE,$89,$F0,$E4,$7F,$8B,$5A,$FE,$4B,$2F,$F6
	.dc.b	$04,$F0,$8F,$F1,$93,$5D,$1F,$80,$9E,$38,$FF,$9B,$9B,$5C,$A7,$89
	.dc.b	$0E,$27,$E5,$2B,$6F,$33,$C8,$43,$61,$F1,$57,$F9,$F1,$FF,$64,$92
	.dc.b	$3F,$DA,$0D,$7E,$11,$82,$18,$21,$FE,$DD,$78,$34,$D9,$4D,$9E,$D2
	.dc.b	$9F,$A2,$78,$B9,$38,$10,$C1,$C6,$3D,$4F,$C5,$6B,$F0,$23,$DF,$3F
	.dc.b	$EF,$3F,$0F,$FD,$84,$62,$7C,$07,$18,$E8,$47,$27,$CA,$8E,$2E,$CC
	.dc.b	$D7,$80,$93,$C6,$3E,$B1,$FE,$73,$9E,$2F,$94,$E4,$68,$C1,$27,$9B
	.dc.b	$FC,$1A,$38,$C6,$23,$83,$5F,$BC,$74,$29,$E9,$3C,$46,$6B,$E1,$3C
	.dc.b	$AB,$D0,$F0,$9F,$F6,$0D,$78,$27,$35,$C2,$B5,$18,$7A,$9D,$CF,$53
	.dc.b	$96,$9F,$21,$EF,$3D,$63,$04,$3C,$85,$3B,$95,$E0,$E8,$86,$8B,$3E
	.dc.b	$78,$DC,$64,$91,$EB,$34,$FA,$AB,$E6,$46,$63,$E4,$19,$AF,$C6,$38
	.dc.b	$3D,$A7,$80,$FA,$55,$E0,$9E,$51,$A3,$C2,$73,$3E,$F1,$CE,$BC,$2D
	.dc.b	$AE,$EF,$CD,$8F,$AA,$F1,$3C,$07,$59,$F7,$D4,$8E,$47,$CB,$8C,$3C
	.dc.b	$EB,$AD,$78,$23,$91,$D4,$57,$08,$D1,$27,$89,$27,$D6,$75,$3C,$07
	.dc.b	$88,$FE,$DB,$3C,$E3,$84,$7A,$A7,$A3,$C9,$F7,$BC,$CE,$27,$33,$E5
	.dc.b	$99,$3C,$E7,$A6,$7E,$8C,$74,$8E,$26,$E3,$EB,$3C,$EB,$C8,$F8,$6B
	.dc.b	$CA,$F6,$AE,$4F,$63,$91,$F1,$E7,$BC,$FE,$7A,$3C,$46,$4E,$C6,$0E
	.dc.b	$27,$D0,$24,$FC,$02,$47,$8A,$BA,$CF,$CC,$38,$CE,$9F,$41,$97,$CF
	.dc.b	$1E,$C8,$F8,$31,$EA,$AF,$D0,$C7,$2A,$F7,$C7,$68,$FD,$6D,$71,$7E
	.dc.b	$05,$71,$7F,$3E,$7A,$CF,$F0,$E9,$F3,$8A,$7C,$72,$C6,$1F,$C3,$98
	.dc.b	$30,$6C,$FD,$09,$93,$83,$A7,$47,$23,$64,$30,$43,$C8,$64,$A6,$4E
	.dc.b	$66,$08,$60,$F8,$35,$F4,$DF,$B5,$3F,$45,$FF,$24,$F8,$7F,$E3,$E8
	.dc.b	$4F,$C4,$8F,$95,$52,$BF,$C9,$F5,$A7,$FC,$8B,$3C,$6B,$84,$76,$7C
	.dc.b	$0F,$D7,$36,$72,$76,$E4,$FB,$EF,$09,$F3,$42,$7B,$3C,$8E,$07,$B8
	.dc.b	$FC,$59,$C8,$A7,$13,$C0,$71,$8E,$B5,$ED,$8F,$DA,$4F,$67,$E2,$9E
	.dc.b	$B3,$9C,$7E,$36,$35,$1D,$4F,$1C,$73,$9F,$8D,$1C,$8F,$F0,$FC,$E6
	.dc.b	$BD,$C7,$EE,$1E,$66,$63,$B9,$FC,$75,$CB,$E3,$3D,$C4,$3E,$11,$5D
	.dc.b	$BE,$43,$81,$A3,$D1,$3E,$66,$CF,$91,$F0,$9F,$3A,$79,$CF,$C8,$8E
	.dc.b	$75,$FD,$62,$73,$5C,$67,$C9,$3F,$1D,$FA,$27,$C3,$AE,$4F,$94,$FA
	.dc.b	$26,$8F,$19,$E2,$76,$49,$FC,$4B,$E9,$AF,$91,$3E,$03,$E4,$1C,$DF
	.dc.b	$2B,$F8,$19,$F3,$3F,$0E,$B2,$F9,$63,$A9,$1F,$13,$C2,$77,$1E,$63
	.dc.b	$81,$A2,$18,$21,$B2,$4F,$A1,$C4,$74,$31,$16,$3D,$44,$3E,$45,$7E
	.dc.b	$9A,$75,$1F,$8E,$3E,$A9,$F0,$8C,$12,$3E,$43,$A3,$E8,$99,$AF,$80
	.dc.b	$FA,$2B,$CD,$5D,$AB,$8C,$F8,$AB,$F3,$4F,$96,$32,$48,$F9,$F5,$F8
	.dc.b	$79,$D4,$76,$7E,$19,$E3,$9E,$0F,$D7,$21,$E3,$35,$1C,$6B,$C8,$F6
	.dc.b	$8F,$81,$3C,$9F,$A3,$3F,$31,$F0,$9D,$0F,$DD,$1F,$D0,$9F,$E7,$04
	.dc.b	$7F,$9F,$B5,$FE,$88,$D7,$81,$A2,$1A,$29,$A3,$04,$30,$43,$F9,$29
	.dc.b	$4F,$E4,$05,$3F,$86,$14,$FC,$49,$EC,$3D,$E5,$1E,$1F,$F1,$83,$E3
	.dc.b	$BF,$06,$2A,$36,$7E,$CE,$32,$F6,$3F,$0C,$7C,$97,$A3,$FB,$9E,$6F
	.dc.b	$E4,$A7,$F2,$6E,$79,$1F,$BD,$6B,$FE,$C2,$8D,$BC,$48,$7F,$B1,$A3
	.dc.b	$77,$FC,$5D,$B8,$7F,$7A,$AE,$35,$CC,$E3,$5B,$7C,$04,$3A,$1A,$9E
	.dc.b	$4F,$80,$FF,$01,$CF,$F7,$92,$3F,$E0,$06,$BF,$E0,$56,$BC,$0D,$10
	.dc.b	$D1,$4D,$18,$21,$82,$1E,$43,$25,$32,$7D,$70,$F8,$9E,$13,$FC,$F6
	.dc.b	$3E,$2B,$63,$E0,$1C,$63,$FB,$07,$3F,$F8,$92,$7F,$AC,$B6,$B9,$92
	.dc.b	$B8,$1C,$8F,$C4,$57,$68,$F2,$4F,$19,$FC,$63,$5D,$9C,$4F,$49,$67
	.dc.b	$93,$C0,$F2,$98,$38,$9F,$7C,$95,$ED,$6D,$78,$9E,$27,$52,$1A,$3D
	.dc.b	$62,$3E,$0B,$C0,$D1,$E0,$7B,$12,$3E,$24,$65,$F0,$3F,$D8,$2B,$E0
	.dc.b	$3E,$E3,$99,$67,$93,$E2,$6A,$8F,$90,$72,$AF,$85,$1C,$6B,$C2,$D3
	.dc.b	$93,$E2,$7D,$8D,$AE,$B1,$B3,$D0,$70,$21,$83,$C4,$7C,$37,$8C,$F8
	.dc.b	$89,$1E,$F3,$53,$E4,$8C,$C7,$C0,$72,$FD,$D3,$E9,$CF,$46,$CF,$9C
	.dc.b	$8F,$90,$F4,$12,$7A,$CF,$D8,$8F,$11,$C8,$E1,$1A,$9E,$71,$EC,$34
	.dc.b	$7B,$CE,$87,$E7,$6B,$E8,$BE,$83,$81,$EB,$38,$BA,$7C,$27,$D0,$3D
	.dc.b	$C6,$C9,$5E,$77,$47,$47,$04,$9F,$95,$1D,$8F,$14,$F7,$79,$C6,$9E
	.dc.b	$71,$99,$E5,$1E,$27,$E7,$3C,$0C,$4F,$C4,$9E,$0F,$39,$D9,$CA,$7C
	.dc.b	$D1,$B7,$E6,$1F,$38,$F8,$31,$B9,$E7,$19,$3C,$07,$23,$C2,$43,$07
	.dc.b	$D3,$AE,$AE,$23,$9B,$A3,$EB,$9D,$1C,$9A,$73,$3E,$AA,$F4,$47,$99
	.dc.b	$FC,$1B,$A7,$E0,$1A,$7E,$59,$F8,$13,$47,$AE,$3D,$CF,$D4,$3C,$E6
	.dc.b	$8F,$41,$93,$F1,$E6,$67,$CE,$F5,$AE,$26,$DF,$09,$E7,$3C,$15,$E6
	.dc.b	$7E,$44,$F2,$7A,$CF,$F2,$78,$C4,$7F,$B7,$9F,$F6,$A1,$1F,$F6,$DB
	.dc.b	$5F,$F7,$03,$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$F7,$8B,$D5,$A7
	.dc.b	$22,$9C,$4A,$7C,$22,$57,$D0,$7E,$21,$FC,$9E,$3F,$AD,$9F,$26,$35
	.dc.b	$1C,$2B,$F3,$D3,$A8,$E2,$FC,$69,$D9,$FD,$89,$D4,$71,$3E,$79,$D6
	.dc.b	$79,$BF,$CF,$E3,$0E,$8E,$A7,$29,$E8,$E0,$87,$53,$F3,$D1,$B9,$FC
	.dc.b	$94,$F9,$1A,$7F,$6E,$8D,$4F,$73,$93,$EB,$8D,$47,$94,$EF,$3C,$DE
	.dc.b	$0F,$91,$FE,$13,$3E,$A7,$F3,$27,$09,$E6,$E8,$C1,$C9,$F2,$BF,$E0
	.dc.b	$08,$F2,$BC,$8F,$29,$D5,$FE,$65,$3E,$13,$C9,$1B,$6C,$6A,$7C,$31
	.dc.b	$89,$E4,$76,$9D,$3E,$32,$1C,$E7,$89,$1C,$C7,$C7,$7F,$56,$F3,$7A
	.dc.b	$9E,$E3,$F1,$C7,$CB,$3D,$66,$5F,$E2,$27,$8C,$E4,$E4,$C1,$C6,$7E
	.dc.b	$04,$FC,$88,$F3,$1F,$24,$D3,$E7,$7C,$27,$80,$91,$C9,$E0,$7D,$52
	.dc.b	$3F,$1D,$FC,$FC,$F8,$9C,$CF,$86,$3D,$31,$E0,$39,$CF,$57,$CA,$48
	.dc.b	$F9,$EF,$92,$35,$18,$3E,$64,$7E,$70,$E8,$6E,$7D,$91,$A3,$FC,$3E
	.dc.b	$79,$67,$9C,$FD,$23,$CE,$79,$27,$8C,$E6,$7F,$A8,$3C,$63,$AC,$E5
	.dc.b	$DB,$E2,$8F,$33,$93,$A1,$0F,$31,$5F,$33,$D4,$C3,$C5,$F7,$1E,$63
	.dc.b	$89,$1E,$2F,$85,$FE,$1A,$FF,$24,$7F,$99,$91,$FE,$66,$D7,$F9,$D3
	.dc.b	$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$DE,$94,$FC,$59,$4F,$C8,$95
	.dc.b	$FB,$C7,$C3,$7D,$CF,$FA,$32,$64,$7B,$C8,$7C,$58,$FC,$34,$6A,$31
	.dc.b	$1C,$CE,$AF,$F9,$AE,$34,$27,$FD,$44,$FF,$A8,$E2,$1F,$E7,$B2,$9F
	.dc.b	$ED,$D6,$B4,$AE,$1E,$07,$07,$04,$76,$43,$45,$3C,$2E,$DC,$90,$C9
	.dc.b	$B3,$FD,$60,$53,$D6,$53,$CC,$7A,$B9,$BC,$C4,$7D,$6D,$74,$60,$86
	.dc.b	$08,$7C,$67,$83,$4D,$94,$D9,$F3,$5D,$B4,$D9,$4D,$9F,$ED,$57,$6F
	.dc.b	$FB,$5E,$7E,$71,$1F,$A6,$D7,$EB,$B5,$FB,$6D,$7E,$F3,$5E,$87,$22
	.dc.b	$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$07,$E2,$B8,$3F,$0C,$47,$F1
	.dc.b	$6D,$74,$60,$86,$08,$7E,$59,$E0,$D3,$65,$36,$7E,$75,$DB,$4D,$94
	.dc.b	$D8,$23,$C1,$C0,$E0,$FD,$01,$1F,$D2,$B5,$FD,$4B,$5F,$D6,$B5,$FD
	.dc.b	$83,$5E,$AD,$7C,$4D,$7C,$07,$62,$1D,$8A,$76,$29,$D8,$A7,$AB,$83
	.dc.b	$F7,$04,$7F,$7A,$D7,$F8,$0B,$5F,$E0,$AD,$7F,$84,$35,$E8,$72,$21
	.dc.b	$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$7F,$14,$DF,$FD,$E4,$4D,$3F
	.dc.b	$8D,$E1,$F5,$A1,$18,$9F,$B9,$18,$7F,$8E,$4E,$E7,$50,$9E,$BA,$61
	.dc.b	$B1,$C0,$E7,$39,$23,$93,$93,$C6,$64,$66,$79,$BF,$ED,$A7,$C8,$D9
	.dc.b	$E6,$D3,$84,$F2,$78,$3F,$CD,$24,$D2,$47,$C7,$6B,$FC,$FA,$36,$E4
	.dc.b	$86,$63,$6F,$F4,$58,$DB,$93,$66,$63,$6F,$D1,$6B,$F4,$1A,$FC,$D6
	.dc.b	$C7,$48,$E4,$78,$B3,$FE,$AC,$F1,$3E,$09,$0F,$82,$57,$6B,$7F,$EB
	.dc.b	$91,$86,$E1,$E5,$9E,$0D,$79,$92,$79,$D7,$C0,$79,$47,$E1,$AB,$9B
	.dc.b	$E5,$38,$57,$39,$E3,$7F,$5D,$B3,$F8,$5A,$E2,$F9,$CD,$57,$12,$3D
	.dc.b	$67,$8B,$D8,$FE,$2D,$3C,$8E,$77,$F8,$C9,$D9,$2B,$9D,$7D,$13,$C4
	.dc.b	$7A,$CF,$09,$DD,$DC,$F3,$78,$D7,$79,$D9,$CC,$F2,$16,$BC,$0F,$33
	.dc.b	$CE,$77,$3E,$89,$93,$84,$F8,$2F,$AD,$77,$9D,$9A,$38,$5F,$95,$E5
	.dc.b	$1B,$7C,$71,$97,$81,$1F,$B9,$5F,$CF,$27,$85,$FA,$2B,$57,$EE,$BE
	.dc.b	$55,$AB,$F3,$57,$92,$7B,$1D,$1F,$79,$F3,$8D,$57,$32,$39,$17,$F8
	.dc.b	$E7,$64,$3D,$71,$F2,$5D,$D7,$A5,$E6,$68,$E1,$3C,$DD,$9E,$42,$9F
	.dc.b	$91,$3C,$4F,$C1,$AF,$5D,$77,$8E,$44,$AF,$05,$76,$7E,$1B,$D2,$FE
	.dc.b	$33,$DE,$78,$3F,$62,$BC,$F1,$88,$F1,$1C,$4F,$05,$74,$8F,$24,$E2
	.dc.b	$7A,$4F,$89,$E0,$78,$CE,$C7,$B1,$D4,$F7,$AF,$D3,$D7,$8A,$7F,$3B
	.dc.b	$5D,$4F,$1D,$79,$6F,$E1,$C7,$E4,$AF,$DA,$EA,$BA,$9E,$72,$3E,$62
	.dc.b	$1B,$2C,$F4,$7F,$53,$3D,$0E,$B1,$9A,$FC,$A4,$66,$7C,$D3,$9A,$F4
	.dc.b	$C7,$BC,$EF,$3E,$E7,$E2,$1F,$A2,$79,$3B,$3F,$3E,$69,$CB,$F3,$0F
	.dc.b	$C6,$12,$BE,$53,$F1,$E3,$B3,$99,$F1,$BC,$67,$6F,$8A,$BF,$56,$F1
	.dc.b	$3D,$77,$F4,$DF,$DD,$99,$3C,$D1,$EF,$72,$7A,$6F,$FD,$24,$47,$FD
	.dc.b	$2E,$D7,$FD,$36,$D7,$FD,$3E,$D7,$FD,$44,$D7,$F1,$86,$0E,$64,$39
	.dc.b	$94,$E6,$53,$81,$A2,$1A,$29,$A3,$E3,$5F,$E2,$9F,$B2,$7C,$DD,$7F
	.dc.b	$A8,$1F,$0D,$D9,$A3,$E1,$3A,$7E,$21,$BC,$3E,$8B,$FB,$58,$D3,$C8
	.dc.b	$FC,$A1,$2B,$F0,$33,$FD,$A6,$76,$F2,$25,$79,$EB,$84,$EA,$31,$5C
	.dc.b	$9B,$39,$3F,$AC,$CF,$96,$33,$1E,$83,$FB,$34,$6D,$F2,$1B,$3E,$01
	.dc.b	$EB,$36,$71,$3D,$93,$E1,$7B,$9A,$32,$73,$3A,$99,$AF,$80,$F3,$7D
	.dc.b	$EF,$81,$CB,$E1,$39,$9B,$AF,$5B,$98,$D9,$C9,$EE,$70,$35,$1F,$D4
	.dc.b	$EB,$C9,$5E,$99,$F9,$A7,$96,$7A,$D7,$39,$FE,$11,$5F,$1D,$E0,$47
	.dc.b	$15,$D0,$EA,$E1,$F4,$1E,$DA,$FB,$31,$CC,$D4,$79,$EB,$F9,$54,$66
	.dc.b	$3C,$87,$12,$1F,$E3,$37,$4F,$B8,$CD,$7A,$1E,$26,$CA,$FA,$9E,$33
	.dc.b	$F2,$AB,$E3,$47,$A0,$F9,$71,$97,$EF,$9D,$CE,$87,$A6,$3E,$49,$C9
	.dc.b	$D4,$AB,$DF,$5F,$36,$7C,$35,$B8,$ED,$1E,$08,$E6,$F8,$23,$35,$CE
	.dc.b	$3D,$67,$D1,$3D,$B5,$D1,$F8,$67,$EF,$CC,$9C,$0F,$74,$7F,$8F,$E3
	.dc.b	$84,$7B,$27,$CC,$FF,$02,$38,$9E,$C3,$C0,$7B,$A3,$8C,$FC,$D8,$DC
	.dc.b	$FA,$27,$6F,$DE,$8E,$67,$F2,$68,$FE,$5A,$47,$F9,$6B,$5F,$E5,$AD
	.dc.b	$78,$1A,$21,$A2,$9A,$30,$43,$04,$3E,$C1,$4F,$BE,$53,$EF,$96,$7E
	.dc.b	$93,$F0,$4F,$7D,$7D,$B3,$FC,$85,$5F,$E7,$D3,$E2,$D4,$9F,$F3,$C4
	.dc.b	$65,$E2,$7D,$F9,$E4,$D2,$1F,$E9,$38,$E0,$FF,$8C,$30,$FF,$23,$3F
	.dc.b	$EB,$87,$FD,$A2,$7F,$B0,$A3,$FD,$36,$D9,$4C,$8C,$37,$1E,$31,$23
	.dc.b	$FD,$A4,$F6,$24,$76,$24,$69,$E0,$F9,$62,$D7,$89,$DC,$7A,$66,$CF
	.dc.b	$99,$DD,$79,$A7,$D8,$78,$AB,$0C,$7F,$D8,$0E,$DA,$FB,$5D,$B9,$6B
	.dc.b	$92,$BC,$1E,$E7,$07,$04,$38,$BA,$3C,$4E,$0D,$3D,$1C,$11,$C1,$0E
	.dc.b	$CE,$5A,$64,$EA,$7C,$12,$1F,$34,$B5,$F3,$5F,$BA,$FF,$B8,$2F,$E0
	.dc.b	$B6,$3F,$D8,$F3,$F6,$CF,$9E,$FC,$33,$A1,$23,$83,$F4,$DA,$F8,$2B
	.dc.b	$E6,$34,$E5,$1E,$37,$94,$7E,$42,$BC,$6D,$8F,$1D,$78,$9F,$A3,$5C
	.dc.b	$49,$3D,$A3,$FD,$23,$5C,$9F,$D9,$9E,$13,$8C,$6E,$78,$C7,$92,$30
	.dc.b	$78,$5F,$63,$5E,$55,$F1,$5A,$FC,$6A,$F0,$1E,$18,$EE,$75,$21,$82
	.dc.b	$47,$FA,$EE,$38,$C7,$C0,$3A,$90,$F8,$E7,$B0,$F7,$9B,$32,$74,$8F
	.dc.b	$90,$75,$7F,$4B,$5F,$59,$DC,$7A,$DA,$FA,$DD,$D7,$C1,$72,$78,$E7
	.dc.b	$E4,$BE,$EA,$F4,$BE,$0A,$8E,$EB,$95,$78,$4E,$33,$E1,$6B,$E1,$9F
	.dc.b	$A2,$7B,$E3,$2F,$33,$E5,$C7,$B5,$DD,$7D,$18,$E9,$5F,$72,$77,$1D
	.dc.b	$5F,$D1,$C7,$90,$8F,$77,$83,$F2,$CF,$CD,$4F,$B9,$E2,$6C,$86,$CF
	.dc.b	$8E,$7D,$63,$CA,$7C,$C3,$83,$E7,$9F,$8F,$1F,$7A,$3B,$3C,$4E,$EE
	.dc.b	$DF,$84,$7D,$13,$A3,$C2,$BE,$75,$7A,$AB,$E1,$D7,$E4,$EB,$53,$F0
	.dc.b	$1E,$91,$E8,$AF,$D2,$C6,$A3,$CE,$7E,$3E,$3E,$9B,$E7,$3C,$24,$3D
	.dc.b	$E7,$B1,$D9,$E4,$79,$9A,$3D,$75,$E6,$9F,$01,$E1,$8F,$99,$3C,$AB
	.dc.b	$D1,$1D,$CF,$8A,$79,$5F,$81,$5F,$7E,$3E,$51,$27,$C4,$F4,$9D,$BC
	.dc.b	$8F,$21,$B7,$F0,$AE,$CF,$C5,$9B,$21,$B3,$F0,$73,$F4,$1E,$15,$EC
	.dc.b	$7D,$07,$CF,$9F,$0B,$C6,$34,$FC,$98,$CD,$7C,$99,$CC,$FC,$8A,$F7
	.dc.b	$4E,$63,$AB,$F1,$DF,$39,$D4,$F0,$9F,$DA,$E7,$FA,$CB,$FC,$34,$C1
	.dc.b	$FD,$71,$C9,$1C,$10,$FE,$D2,$E5,$A6,$4F,$ED,$EE,$5A,$64,$FE,$F2
	.dc.b	$E5,$A6,$4F,$19,$D0,$87,$23,$EB,$C7,$C1,$6D,$7E,$1A,$7F,$62,$62
	.dc.b	$3F,$22,$7D,$89,$E0,$FE,$2A,$36,$E6,$7F,$1E,$F1,$8F,$B4,$69,$F9
	.dc.b	$86,$4F,$A3,$1E,$17,$A1,$CC,$F0,$9C,$0C,$47,$F5,$19,$FE,$E8,$F3
	.dc.b	$88,$68,$FD,$7B,$E1,$8F,$4C,$73,$23,$C9,$F0,$98,$9F,$5B,$CC,$E4
	.dc.b	$FF,$71,$9E,$F1,$B3,$CF,$3C,$DF,$EA,$31,$A8,$F0,$92,$79,$4F,$82
	.dc.b	$30,$F2,$8F,$0B,$C8,$D3,$88,$DC,$FC,$E9,$C9,$C4,$8E,$5E,$47,$9D
	.dc.b	$FD,$2B,$F9,$F7,$DA,$F0,$3E,$A9,$5C,$9C,$23,$C3,$1F,$38,$F0,$C6
	.dc.b	$67,$CC,$7D,$D8,$F0,$CF,$07,$E2,$46,$63,$E4,$11,$F8,$B1,$E7,$3C
	.dc.b	$73,$F8,$18,$CC,$EE,$7F,$D7,$D3,$D1,$F4,$1E,$43,$FD,$23,$1B,$8F
	.dc.b	$34,$7F,$38,$7D,$AE,$A7,$A9,$EA,$79,$9E,$A7,$2F,$80,$FC,$31,$0D
	.dc.b	$19,$2C,$7A,$5F,$9C,$7E,$8D,$F0,$9E,$27,$D4,$FF,$88,$0C,$BE,$B3
	.dc.b	$27,$D1,$3A,$9E,$93,$C1,$3F,$3D,$FD,$24,$72,$76,$72,$31,$1F,$15
	.dc.b	$E5,$1F,$18,$F1,$3F,$2D,$F3,$3E,$F8,$FE,$59,$1C,$A7,$E2,$BC,$67
	.dc.b	$CD,$3D,$63,$8C,$7D,$D3,$81,$E8,$9E,$2F,$E0,$09,$1E,$97,$F9,$8B
	.dc.b	$FC,$A8,$8F,$F2,$86,$BF,$C9,$9A,$F0,$34,$43,$45,$34,$60,$86,$08
	.dc.b	$7E,$58,$A7,$DF,$2B,$FE,$45,$70,$7B,$1F,$79,$F8,$32,$29,$FF,$11
	.dc.b	$9A,$7E,$99,$23,$F2,$F1,$A8,$FA,$84,$3F,$07,$19,$7F,$CC,$71,$FE
	.dc.b	$48,$3F,$AD,$91,$FF,$5A,$B5,$FF,$5F,$B5,$FF,$51,$9A,$21,$A2,$9A
	.dc.b	$30,$43,$04,$3F,$CD,$E7,$02,$9B,$29,$B3,$FD,$54,$7F,$B2,$B5,$FE
	.dc.b	$E6,$47,$04,$3F,$DA,$EE,$5A,$64,$FF,$6E,$B9,$69,$93,$FD,$CE,$E5
	.dc.b	$A6,$4F,$F7,$73,$96,$99,$01,$D3,$87,$FC,$17,$CA,$F4,$5F,$A0,$7F
	.dc.b	$9F,$23,$0D,$8C,$D7,$F9,$B2,$72,$D8,$CB,$F8,$9C,$7C,$33,$A8,$C0
	.dc.b	$7C,$1F,$A0,$F5,$3B,$5C,$9E,$8D,$78,$CF,$3C,$57,$DB,$0F,$0C,$62
	.dc.b	$BB,$43,$0E,$D1,$D0,$87,$48,$E3,$5C,$F0,$F0,$5F,$3C,$3C,$13,$DF
	.dc.b	$0E,$78,$71,$BD,$CF,$18,$EB,$7C,$70,$EB,$1E,$4A,$E3,$39,$BE,$33
	.dc.b	$93,$8D,$F3,$7B,$18,$3D,$B1,$CA,$35,$3D,$EB,$9E,$1E,$8B,$E8,$72
	.dc.b	$3E,$05,$F1,$7E,$41,$F8,$82,$18,$29,$4C,$DF,$F9,$50,$8E,$8F,$C8
	.dc.b	$B8,$70,$D7,$9B,$5E,$67,$13,$B1,$E3,$74,$D3,$4E,$1E,$E6,$08,$78
	.dc.b	$0A,$7A,$B8,$7F,$CD,$64,$7C,$0D,$7D,$4D,$78,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$3E,$6B,$D5,$A7,$22,$9C,$4A,$7D,$2E,$1F,$F4,$19,$1F,$90
	.dc.b	$D7,$E9,$35,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$FC,$2B,$D5,$A7,$22
	.dc.b	$9C,$4A,$7E,$2B,$87,$FD,$36,$47,$F0,$0D,$7F,$14,$D7,$81,$A2,$1A
	.dc.b	$29,$A3,$04,$30,$43,$F3,$AF,$56,$9C,$8A,$71,$29,$FA,$2E,$16,$F2
	.dc.b	$3F,$97,$E9,$FC,$46,$73,$3F,$AB,$8C,$2E,$0C,$65,$FC,$DA,$B9,$C6
	.dc.b	$2D,$18,$7A,$CE,$EB,$94,$60,$DB,$E3,$7F,$97,$CF,$89,$D3,$5C,$C6
	.dc.b	$EF,$CB,$7E,$1A,$F0,$56,$A3,$15,$E0,$8E,$2E,$08,$F6,$7A,$5F,$9E
	.dc.b	$78,$92,$74,$F0,$79,$9C,$67,$F8,$55,$79,$E3,$F8,$A3,$5F,$85,$1A
	.dc.b	$3E,$03,$A6,$BA,$3E,$C3,$87,$E3,$BC,$48,$E9,$AF,$F2,$28,$F0,$3C
	.dc.b	$0D,$10,$D4,$75,$8D,$91,$C1,$0E,$B1,$E7,$7E,$15,$FE,$19,$B7,$E9
	.dc.b	$9F,$8A,$7F,$6A,$E1,$FE,$67,$5F,$D7,$DB,$3B,$8E,$53,$F2,$C8,$72
	.dc.b	$72,$F3,$3F,$9F,$BC,$DB,$1C,$C8,$FC,$13,$93,$8B,$FC,$64,$F1,$9F
	.dc.b	$CD,$57,$91,$E1,$1E,$02,$57,$1B,$FC,$3C,$EA,$3F,$C3,$35,$C4,$E5
	.dc.b	$7D,$A3,$85,$7C,$6A,$E0,$4A,$E0,$FA,$8C,$D7,$16,$CF,$82,$30,$66
	.dc.b	$78,$CF,$58,$CC,$EC,$87,$86,$77,$3F,$3A,$FD,$AF,$69,$E4,$7B,$A7
	.dc.b	$71,$F2,$8D,$1C,$6F,$DA,$FE,$F6,$37,$5D,$2F,$C7,$7D,$0F,$15,$7B
	.dc.b	$EB,$E3,$C7,$9A,$3A,$1D,$8F,$04,$74,$BE,$F1,$E3,$AE,$AF,$09,$D3
	.dc.b	$C6,$FC,$57,$D6,$39,$9E,$13,$E8,$57,$27,$A4,$6E,$7E,$2D,$74,$AD
	.dc.b	$47,$49,$F6,$57,$C1,$3A,$1F,$8C,$29,$EC,$9E,$37,$F9,$6B,$F3,$9D
	.dc.b	$6B,$C1,$1E,$EB,$FC,$B5,$F5,$3A,$10,$E9,$3C,$1F,$C6,$1E,$03,$F5
	.dc.b	$06,$CF,$6B,$E8,$72,$75,$21,$83,$67,$67,$94,$79,$23,$4D,$3E,$23
	.dc.b	$B3,$9B,$99,$FD,$3C,$79,$63,$6F,$63,$9C,$69,$EE,$D7,$A1,$B8,$E8
	.dc.b	$AB,$E5,$BF,$72,$BE,$25,$FD,$08,$D1,$E1,$BE,$8F,$80,$FC,$43,$F4
	.dc.b	$4C,$DF,$6B,$FA,$D7,$F2,$63,$F0,$B5,$CC,$EA,$7A,$0E,$64,$9F,$2B
	.dc.b	$9B,$F2,$1E,$58,$CC,$7C,$8B,$F1,$3C,$4E,$D1,$E2,$7E,$1D,$78,$9F
	.dc.b	$04,$65,$F8,$D3,$F8,$F7,$89,$E7,$AF,$0D,$FE,$26,$BC,$33,$F0,$E7
	.dc.b	$33,$FB,$88,$EA,$7F,$98,$67,$FC,$C4,$FC,$03,$FC,$E2,$E0,$8E,$C8
	.dc.b	$7F,$9C,$5C,$B4,$C9,$C8,$A7,$F9,$F9,$DB,$93,$81,$0C,$10,$E0,$72
	.dc.b	$3F,$D1,$0F,$27,$F0,$0F,$C0,$BF,$BD,$0E,$1F,$DD,$1F,$01,$87,$D5
	.dc.b	$3E,$01,$FD,$55,$DB,$99,$E9,$87,$E2,$89,$1F,$92,$9D,$BE,$23,$EA
	.dc.b	$9D,$8E,$87,$F0,$D7,$15,$E6,$8B,$3F,$AC,$9E,$73,$99,$F1,$1D,$0E
	.dc.b	$44,$3C,$07,$23,$D6,$FC,$28,$E7,$3E,$8A,$F5,$D6,$A3,$0F,$C7,$3A
	.dc.b	$57,$85,$EA,$78,$8E,$67,$C1,$30,$43,$67,$C0,$23,$CC,$91,$E0,$8E
	.dc.b	$72,$7C,$47,$29,$DB,$E7,$24,$6E,$B8,$C7,$C5,$38,$57,$B2,$33,$1F
	.dc.b	$16,$7E,$09,$C5,$F7,$CF,$8D,$F5,$57,$82,$75,$1E,$A9,$D3,$D8,$F1
	.dc.b	$9F,$D4,$EB,$BB,$CA,$37,$5F,$4E,$3B,$9C,$0F,$21,$FB,$C9,$F5,$9C
	.dc.b	$9E,$D5,$FE,$0E,$8C,$C7,$28,$F5,$BE,$13,$07,$CA,$9F,$F1,$74,$F6
	.dc.b	$8F,$E0,$2F,$27,$D8,$78,$CF,$41,$F7,$4C,$CF,$C2,$9E,$71,$A8,$FE
	.dc.b	$6F,$3C,$E7,$33,$E8,$3D,$87,$92,$3D,$91,$F3,$63,$D0,$76,$3D,$15
	.dc.b	$FA,$37,$B1,$C8,$D4,$F2,$8F,$9A,$7D,$F8,$EC,$7C,$23,$E0,$18,$38
	.dc.b	$9D,$0F,$86,$66,$7D,$27,$95,$C9,$C0,$F3,$1D,$0C,$3C,$23,$D1,$1F
	.dc.b	$04,$E3,$19,$7E,$63,$6B,$EC,$C7,$C9,$9E,$2F,$F1,$E3,$F4,$65,$AF
	.dc.b	$E5,$C4,$7F,$97,$35,$FE,$4E,$D7,$81,$A2,$1A,$29,$A3,$04,$30,$43
	.dc.b	$CA,$43,$F0,$C6,$4A,$7C,$2A,$F7,$3F,$70,$F8,$33,$92,$3F,$DF,$23
	.dc.b	$FC,$60,$7C,$37,$64,$AF,$C9,$4F,$07,$FD,$C7,$1F,$E3,$99,$E1,$1A
	.dc.b	$8E,$07,$85,$C1,$C1,$FF,$4A,$47,$FB,$8E,$FF,$CF,$04,$3F,$DA,$4A
	.dc.b	$BF,$8B,$47,$FB,$62,$31,$3F,$EF,$2C,$79,$4F,$0A,$DB,$FE,$90,$3A
	.dc.b	$1A,$37,$12,$7F,$DD,$D1,$D6,$78,$CE,$E7,$33,$CC,$87,$15,$5E,$48
	.dc.b	$E5,$1E,$53,$8C,$F3,$77,$5E,$69,$C4,$7F,$33,$0F,$FB,$50,$87,$FB
	.dc.b	$74,$A6,$DE,$0E,$1C,$1C,$9E,$04,$32,$E5,$A7,$17,$A3,$5E,$AF,$17
	.dc.b	$AB,$97,$D4,$61,$D3,$C9,$B5,$F1,$5F,$93,$1F,$31,$F8,$91,$97,$E1
	.dc.b	$3F,$81,$C3,$CE,$D9,$FF,$68,$4F,$DC,$79,$10,$F9,$A7,$B4,$C4,$71
	.dc.b	$7F,$DC,$33,$D2,$BD,$31,$C2,$BA,$35,$F3,$B5,$F2,$C7,$F0,$68,$FE
	.dc.b	$D5,$5D,$5F,$5C,$74,$8F,$64,$78,$A3,$04,$AF,$11,$E3,$8F,$41,$DC
	.dc.b	$E0,$62,$79,$B9,$3F,$17,$1C,$9F,$8B,$5E,$89,$D9,$2B,$C3,$3C,$DF
	.dc.b	$53,$C8,$E3,$3E,$A7,$44,$7E,$2C,$FB,$E3,$E0,$CF,$28,$ED,$1B,$9F
	.dc.b	$14,$F8,$1F,$F7,$F5,$79,$E7,$FD,$F5,$1D,$27,$4F,$08,$F3,$4E,$DF
	.dc.b	$24,$6D,$EC,$FD,$08,$D4,$7C,$88,$D1,$0F,$D6,$BE,$88,$DB,$F2,$EB
	.dc.b	$D3,$3B,$3A,$47,$EB,$23,$E0,$3B,$AF,$3C,$F9,$0F,$7D,$74,$AE,$71
	.dc.b	$EB,$3A,$57,$84,$E9,$5F,$24,$E6,$FA,$63,$81,$E8,$7E,$B1,$F9,$A3
	.dc.b	$8D,$77,$9E,$31,$E5,$3E,$49,$A2,$1C,$4F,$A3,$3D,$8F,$7B,$93,$F3
	.dc.b	$F3,$D5,$F9,$E7,$96,$32,$F4,$29,$C4,$D1,$0F,$49,$A2,$1F,$0C,$F1
	.dc.b	$BF,$2C,$F1,$BA,$38,$9B,$34,$75,$AF,$AA,$4A,$F9,$95,$FA,$DA,$FA
	.dc.b	$35,$F6,$63,$E4,$1E,$D8,$CD,$7E,$9E,$33,$3E,$D3,$B5,$7C,$E9,$ED
	.dc.b	$3A,$9F,$49,$A3,$E8,$9F,$0C,$F4,$92,$3F,$0B,$5E,$A8,$C5,$7A,$1F
	.dc.b	$01,$F6,$E3,$CB,$19,$8E,$93,$EF,$7E,$A9,$B9,$FC,$44,$79,$A3,$9B
	.dc.b	$B8,$E7,$1E,$63,$E7,$46,$DF,$59,$CE,$3E,$9B,$F4,$48,$7A,$0F,$0C
	.dc.b	$7B,$1C,$BC,$DE,$B3,$F2,$49,$5F,$26,$32,$72,$8C,$BE,$87,$F8,$BB
	.dc.b	$FE,$86,$3F,$AF,$C7,$FB,$B0,$8F,$F5,$E3,$64,$72,$FF,$62,$6B,$C8
	.dc.b	$E0,$E1,$FE,$DC,$D7,$FB,$B3,$5F,$11,$A3,$89,$0E,$C5,$3B,$1F,$62
	.dc.b	$3E,$D3,$F1,$23,$F9,$21,$FA,$A3,$0F,$C1,$7E,$33,$FA,$77,$4F,$CE
	.dc.b	$8E,$6D,$8F,$CA,$1A,$7F,$1A,$F2,$69,$FC,$11,$D4,$FD,$78,$E4,$78
	.dc.b	$4D,$1E,$13,$27,$AE,$3A,$47,$F6,$F7,$9C,$F9,$27,$B9,$B3,$99,$FC
	.dc.b	$72,$3D,$EE,$48,$72,$7E,$8B,$E0,$36,$43,$C8,$73,$34,$43,$C2,$73
	.dc.b	$3F,$CB,$AF,$A2,$3C,$D3,$FA,$18,$D4,$F8,$8E,$51,$97,$BB,$F1,$DE
	.dc.b	$A6,$A7,$FD,$DB,$3C,$A3,$CC,$68,$F8,$6E,$5E,$0F,$8C,$F7,$BE,$B3
	.dc.b	$84,$F7,$24,$F6,$8F,$1B,$EA,$24,$FE,$0A,$32,$F5,$7E,$E1,$FD,$0D
	.dc.b	$F7,$9E,$37,$BC,$7F,$0F,$8C,$C7,$23,$89,$F6,$1C,$BE,$79,$F2,$C7
	.dc.b	$86,$31,$3E,$76,$C7,$8D,$F1,$1D,$4F,$F6,$AB,$F0,$A7,$9C,$7A,$23
	.dc.b	$BC,$7E,$3E,$3E,$49,$93,$94,$7A,$E3,$8C,$76,$3C,$27,$E0,$88,$7D
	.dc.b	$F3,$EA,$9C,$8F,$21,$F6,$9F,$C1,$3C,$8F,$39,$C8,$D4,$FC,$97,$91
	.dc.b	$B3,$33,$FC,$01,$CC,$FA,$DF,$A3,$1F,$26,$30,$49,$F9,$4F,$B2,$7E
	.dc.b	$91,$CA,$32,$7E,$09,$F8,$87,$9E,$7E,$24,$66,$3E,$24,$F5,$9F,$C8
	.dc.b	$C0,$FB,$05,$72,$E0,$8E,$08,$FF,$26,$6B,$FE,$69,$69,$FC,$90,$8F
	.dc.b	$F2,$B3,$47,$12,$18,$23,$C1,$FD,$3B,$5F,$31,$C1,$FC,$3C,$FC,$47
	.dc.b	$F1,$8D,$8F,$F6,$B4,$FF,$A1,$CD,$CE,$89,$1F,$86,$7F,$C3,$F1,$23
	.dc.b	$E8,$46,$9F,$C1,$91,$D3,$FE,$3A,$8F,$F4,$59,$FD,$5D,$FF,$65,$9F
	.dc.b	$D6,$5C,$11,$C1,$0F,$F5,$29,$4F,$F2,$B1,$4F,$F6,$33,$C5,$D9,$0D
	.dc.b	$94,$D9,$C4,$FF,$59,$94,$FF,$7D,$1C,$08,$7F,$B4,$F8,$7D,$44,$7F
	.dc.b	$B9,$35,$FF,$68,$B5,$FF,$6B,$B5,$E4,$70,$21,$C0,$A7,$02,$9C,$0D
	.dc.b	$10,$D1,$4D,$18,$21,$FE,$EB,$AF,$A4,$47,$FC,$14,$D7,$FD,$D2,$D7
	.dc.b	$FD,$DE,$D7,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$87
	.dc.b	$ED,$4F,$DB,$6B,$F8,$B2,$3F,$E2,$36,$BB,$38,$38,$70,$1C,$3B,$6B
	.dc.b	$C0,$E8,$43,$89,$C4,$C1,$D1,$E2,$F9,$8C,$9D,$4A,$75,$39,$BF,$A5
	.dc.b	$D6,$47,$F8,$F5,$AF,$81,$AF,$A9,$AF,$BD,$AF,$C7,$6B,$FC,$75,$A9
	.dc.b	$A4,$7C,$04,$3C,$05,$3C,$07,$F9,$47,$5F,$DE,$11,$C1,$0F,$8E,$E5
	.dc.b	$A6,$4F,$A6,$E5,$A6,$4F,$B4,$E5,$A6,$4F,$C2,$39,$69,$93,$F9,$E3
	.dc.b	$96,$9F,$E6,$9D,$59,$C3,$F1,$D9,$31,$4E,$67,$31,$87,$8E,$98,$9E
	.dc.b	$4D,$8C,$DF,$F4,$DB,$E9,$5A,$8C,$5F,$3A,$FF,$4D,$B5,$E4,$62,$7C
	.dc.b	$53,$C1,$E4,$AB,$AC,$F8,$23,$15,$D6,$F9,$9D,$26,$4F,$37,$E1,$35
	.dc.b	$F0,$1C,$EB,$AA,$9E,$73,$B7,$9B,$E0,$BF,$3B,$67,$AB,$FB,$B8,$CB
	.dc.b	$FD,$C5,$AF,$C1,$3F,$74,$E0,$91,$CA,$38,$35,$E0,$75,$71,$1C,$1F
	.dc.b	$BF,$1D,$1C,$90,$CB,$C5,$AF,$66,$BD,$8C,$3F,$C3,$23,$B3,$4E,$85
	.dc.b	$38,$BF,$12,$BE,$5C,$7C,$86,$CF,$D1,$8C,$5F,$B3,$CA,$79,$74,$F7
	.dc.b	$DF,$0B,$F9,$D7,$F6,$08,$72,$C3,$E9,$DF,$26,$DF,$2C,$24,$79,$5B
	.dc.b	$5F,$63,$0E,$C7,$7B,$E5,$1F,$71,$AF,$83,$0F,$51,$0F,$2B,$E3,$38
	.dc.b	$61,$EB,$C3,$C0,$48,$F0,$61,$E2,$7A,$57,$23,$FC,$71,$7E,$08,$FE
	.dc.b	$8C,$D8,$F0,$61,$E8,$AE,$55,$E1,$3E,$FD,$6E,$8D,$AE,$17,$E3,$AF
	.dc.b	$0C,$61,$F4,$C7,$CA,$34,$F5,$25,$7C,$1A,$EC,$75,$BF,$1C,$79,$9A
	.dc.b	$FB,$4F,$1E,$1F,$16,$BE,$1D,$F6,$9E,$0D,$78,$5F,$6A,$E7,$87,$9E
	.dc.b	$FD,$38,$76,$AD,$61,$C6,$B5,$5E,$8C,$39,$E1,$E1,$8E,$35,$BC,$3A
	.dc.b	$99,$8F,$A3,$38,$9D,$12,$3E,$39,$0E,$67,$B0,$86,$0F,$AB,$1D,$6B
	.dc.b	$EF,$B5,$F8,$D7,$E8,$7E,$74,$67,$0F,$6D,$74,$3D,$95,$F9,$99,$D1
	.dc.b	$CD,$E5,$5D,$30,$CE,$1F,$76,$FD,$07,$28,$F9,$57,$A8,$CD,$7D,$47
	.dc.b	$C0,$F9,$2B,$94,$7B,$1A,$F3,$7E,$D6,$1E,$88,$D1,$D2,$7C,$93,$F2
	.dc.b	$E3,$24,$3E,$8B,$D2,$3C,$51,$F2,$CC,$9C,$E7,$B3,$E1,$3C,$F1,$F9
	.dc.b	$09,$E4,$D8,$F1,$61,$F1,$8F,$8F,$5C,$A7,$EA,$D7,$09,$D6,$1F,$1A
	.dc.b	$39,$D6,$EB,$D7,$7E,$0B,$FA,$95,$AC,$3E,$A6,$1F,$8B,$AF,$2E,$1C
	.dc.b	$4E,$57,$DA,$FE,$4D,$FD,$3A,$F8,$98,$7C,$27,$F3,$97,$C6,$7A,$D7
	.dc.b	$CB,$9F,$EB,$33,$98,$F5,$61,$E1,$9F,$E2,$33,$97,$E1,$C6,$5E,$35
	.dc.b	$F9,$5C,$DB,$FF,$54,$6E,$1F,$D9,$1B,$5F,$D9,$1A,$FF,$52,$6B,$FE
	.dc.b	$3B,$9E,$44,$9E,$4D,$8E,$4D,$79,$35,$E5,$1C,$08,$F0,$6B,$C1,$AF
	.dc.b	$03,$44,$3E,$6E,$3F,$85,$6E,$3F,$15,$B8,$7D,$37,$F2,$26,$2B,$EF
	.dc.b	$11,$FC,$A4,$6E,$BF,$17,$18,$9D,$3C,$4C,$10,$C1,$0E,$E4,$8E,$6E
	.dc.b	$5A,$E5,$F5,$36,$BD,$4D,$AF,$24,$E8,$E2,$43,$04,$9E,$2E,$5A,$E6
	.dc.b	$39,$C6,$08,$E0,$95,$F0,$2B,$C4,$DA,$E9,$3A,$9F,$84,$68,$B1,$A9
	.dc.b	$E0,$F8,$5B,$5E,$06,$D7,$3A,$F8,$D5,$B7,$9B,$5C,$D7,$17,$2D,$73
	.dc.b	$3C,$DF,$21,$8A,$EE,$47,$8C,$6E,$BD,$11,$89,$D3,$C4,$C1,$0C,$10
	.dc.b	$EE,$48,$E6,$E5,$AE,$5F,$63,$6B,$D4,$DA,$F2,$4E,$8E,$24,$30,$49
	.dc.b	$E2,$E5,$AE,$63,$9C,$60,$8E,$09,$5F,$02,$BC,$4D,$AE,$93,$FB,$D7
	.dc.b	$FD,$7D,$75,$FF,$5F,$37,$0F,$E1,$71,$AA,$FD,$EB,$6F,$F7,$87,$EF
	.dc.b	$CF,$E0,$46,$1D,$38,$21,$B2,$3A,$6B,$E2,$8E,$ED,$8E,$8F,$C3,$8F
	.dc.b	$63,$63,$EE,$1F,$00,$AF,$C0,$8D,$E1,$F2,$1B,$87,$C8,$AD,$D7,$37
	.dc.b	$F0,$0F,$D6,$BF,$86,$F8,$E3,$D6,$FA,$48,$65,$F5,$C7,$02,$61,$F3
	.dc.b	$49,$7E,$E8,$CB,$63,$33,$F0,$27,$2F,$9B,$0F,$53,$70,$F5,$3E,$88
	.dc.b	$CB,$63,$2F,$C5,$31,$7F,$08,$8F,$36,$BA,$30,$43,$04,$3A,$13,$0F
	.dc.b	$A0,$EC,$C3,$F3,$5B,$5E,$66,$D7,$9A,$78,$C7,$56,$BA,$AF,$2B,$4E
	.dc.b	$D1,$A2,$1D,$CB,$5D,$DB,$5D,$DD,$12,$34,$D8,$D4,$60,$8E,$09,$5E
	.dc.b	$49,$EA,$D7,$93,$67,$8B,$67,$8D,$7A,$AB,$6D,$AD,$B6,$7D,$17,$F8
	.dc.b	$48,$FC,$9D,$7E,$44,$FC,$9C,$66,$3F,$27,$87,$E3,$1D,$C6,$67,$73
	.dc.b	$99,$DE,$1F,$93,$23,$E1,$7D,$04,$70,$43,$EC,$4F,$A9,$B8,$7A,$9B
	.dc.b	$87,$C0,$6D,$73,$69,$CC,$FB,$C6,$08,$60,$95,$F4,$1E,$0D,$76,$D7
	.dc.b	$6F,$95,$B5,$E5,$6D,$79,$5B,$5E,$56,$9D,$AB,$91,$D4,$86,$09,$5D
	.dc.b	$5C,$B5,$CD,$77,$72,$D7,$35,$E1,$72,$D7,$35,$E3,$AE,$84,$AF,$B9
	.dc.b	$3F,$7A,$BE,$DD,$7E,$0A,$F3,$3F,$89,$26,$1F,$85,$6C,$6A,$30,$47
	.dc.b	$FA,$56,$1D,$EF,$C8,$DB,$F8,$D5,$B9,$F0,$C7,$E3,$8F,$E3,$D5,$D4
	.dc.b	$91,$AB,$F4,$4E,$A3,$11,$E4,$6D,$75,$9D,$10,$F7,$46,$AB,$AC,$64
	.dc.b	$F1,$1A,$31,$5E,$07,$04,$36,$47,$4D,$7C,$8F,$26,$BC,$5D,$90,$CB
	.dc.b	$C6,$BD,$4E,$E3,$26,$E3,$31,$B9,$F5,$BB,$8C,$CE,$E7,$33,$BA,$F8
	.dc.b	$44,$7C,$D5,$C2,$BB,$91,$C1,$2B,$9C,$75,$AE,$55,$E1,$9C,$CF,$16
	.dc.b	$D7,$1A,$DD,$78,$CF,$0C,$E8,$91,$AA,$ED,$3A,$8C,$47,$56,$D7,$59
	.dc.b	$D1,$0F,$24,$6A,$BA,$C6,$4F,$11,$A3,$15,$E0,$70,$43,$64,$74,$D7
	.dc.b	$C8,$F2,$6B,$C5,$D9,$0C,$BC,$6B,$D4,$EE,$32,$6E,$33,$1B,$9F,$5B
	.dc.b	$B8,$CC,$EE,$73,$3B,$AF,$84,$47,$CD,$5C,$2B,$B9,$1C,$12,$B9,$C7
	.dc.b	$5A,$E5,$5E,$19,$CC,$F1,$6D,$71,$AD,$D7,$8C,$F0,$CE,$89,$1A,$AE
	.dc.b	$D3,$A8,$C4,$75,$6D,$75,$9D,$10,$F2,$46,$AB,$AC,$64,$FE,$76,$7F
	.dc.b	$46,$AF,$E5,$6D,$BF,$E3,$4D,$C3,$F9,$A4,$62,$38,$92,$7F,$9A,$BF
	.dc.b	$CC,$E7,$67,$81,$F1,$9E,$18,$F6,$B8,$3F,$01,$3D,$63,$9C,$78,$6F
	.dc.b	$F9,$6D,$74,$69,$23,$C0,$D7,$81,$82,$18,$21,$E9,$9F,$82,$DA,$F1
	.dc.b	$46,$0F,$1C,$66,$BC,$4F,$43,$15,$E1,$69,$99,$E2,$43,$A1,$2B,$B9
	.dc.b	$2B,$AC,$F9,$A7,$84,$E6,$BB,$46,$2B,$75,$E0,$38,$47,$58,$CB,$FD
	.dc.b	$DB,$0F,$C4,$4F,$D1,$3F,$28,$58,$FA,$AF,$EF,$6F,$C0,$FA,$E3,$A5
	.dc.b	$F8,$9B,$19,$9D,$BE,$26,$BE,$52,$47,$A6,$79,$4E,$A3,$13,$CE,$FD
	.dc.b	$31,$DA,$3A,$46,$27,$84,$E6,$7A,$B4,$E6,$F0,$AE,$F5,$C4,$8E,$09
	.dc.b	$1E,$16,$BC,$4C,$10,$C1,$0F,$0D,$7F,$9B,$2F,$E7,$C7,$CE,$25,$FC
	.dc.b	$E2,$1F,$C4,$A7,$D6,$4B,$F2,$CF,$73,$AB,$87,$FD,$73,$5D,$AB,$9B
	.dc.b	$5F,$E1,$11,$FB,$17,$FA,$FD,$75,$BF,$6C,$F4,$6C,$FC,$88,$C1,$E1
	.dc.b	$70,$43,$E7,$12,$7A,$46,$E3,$17,$E7,$8E,$67,$C7,$8E,$53,$C2,$31
	.dc.b	$1C,$2B,$C6,$D8,$E9,$3B,$79,$92,$39,$D7,$76,$D7,$2A,$E1,$18,$AE
	.dc.b	$75,$B8,$ED,$1E,$2B,$F9,$44,$C3,$F8,$E4,$7D,$A9,$F9,$AF,$A2,$FA
	.dc.b	$C7,$28,$F2,$34,$F0,$3B,$BF,$04,$F0,$23,$82,$61,$E9,$6C,$FA,$23
	.dc.b	$04,$70,$43,$B3,$96,$99,$3C,$B3,$E0,$8D,$46,$27,$BB,$C0,$86,$63
	.dc.b	$75,$EC,$83,$29,$4A,$54,$37,$4A,$DC,$FF,$59,$24,$FE,$67,$0F,$A5
	.dc.b	$1F,$D9,$27,$D1,$3E,$BA,$FC,$A5,$6A,$38,$61,$F6,$27,$F4,$AD,$7E
	.dc.b	$CC,$7D,$B9,$E0,$FF,$23,$8C,$9C,$2F,$C7,$19,$9D,$BC,$6B,$A4,$62
	.dc.b	$B8,$4F,$8A,$B8,$BB,$6C,$6A,$3A,$57,$22,$4E,$A7,$C6,$4A,$EC,$D9
	.dc.b	$DD,$77,$21,$D2,$B9,$57,$02,$4F,$09,$D1,$C0,$F3,$1F,$5E,$BE,$EC
	.dc.b	$A7,$F0,$B3,$F8,$33,$E9,$4F,$8E,$BB,$CF,$59,$E6,$7F,$9F,$EF,$EC
	.dc.b	$BF,$8C,$AF,$B3,$3B,$7E,$24,$70,$23,$82,$1D,$DF,$43,$8B,$F6,$DF
	.dc.b	$88,$93,$A9,$E0,$FC,$02,$1D,$9C,$B6,$79,$BC,$5A,$F0,$8F,$05,$7A
	.dc.b	$5E,$26,$2B,$C3,$3B,$8C,$92,$B9,$C7,$7A,$E5,$18,$9E,$53,$E1,$9C
	.dc.b	$92,$BA,$C7,$16,$CF,$08,$F0,$57,$38,$F0,$57,$19,$E9,$5C,$27,$2D
	.dc.b	$AE,$04,$AE,$15,$D1,$AF,$84,$E6,$43,$53,$CE,$37,$5E,$59,$E0,$68
	.dc.b	$E0,$60,$8E,$C8,$79,$9D,$B8,$AF,$35,$77,$24,$EA,$78,$3D,$C8,$76
	.dc.b	$72,$D9,$E6,$F1,$6B,$C2,$3C,$15,$EC,$78,$98,$AF,$0C,$EE,$32,$4A
	.dc.b	$E7,$1D,$EB,$94,$62,$79,$4F,$86,$72,$4A,$EB,$1C,$5B,$3C,$23,$C1
	.dc.b	$5C,$E3,$C1,$5C,$67,$A5,$70,$9C,$B6,$B8,$12,$B8,$57,$46,$BE,$13
	.dc.b	$99,$0D,$4F,$38,$DD,$79,$67,$81,$A3,$81,$82,$3B,$21,$E6,$76,$E2
	.dc.b	$BC,$D5,$DC,$93,$A9,$E0,$F7,$21,$D9,$CB,$67,$9B,$C5,$AF,$08,$F0
	.dc.b	$57,$B1,$E2,$62,$BC,$33,$B8,$C9,$2B,$9C,$77,$AE,$51,$89,$E5,$3E
	.dc.b	$19,$C9,$2B,$AC,$71,$6C,$F0,$8F,$05,$73,$8F,$05,$71,$9E,$95,$C2
	.dc.b	$72,$DA,$E0,$4A,$E1,$5D,$1A,$F8,$4E,$64,$35,$3C,$E3,$75,$E5,$9E
	.dc.b	$06,$8E,$06,$08,$EC,$87,$99,$DB,$8A,$F3,$57,$72,$4E,$A7,$83,$DC
	.dc.b	$87,$67,$2D,$9E,$6F,$16,$BC,$23,$C1,$5E,$C7,$89,$8A,$F0,$CE,$E3
	.dc.b	$24,$AE,$71,$DE,$B9,$46,$27,$94,$F8,$67,$24,$AE,$B1,$C5,$B3,$C2
	.dc.b	$3C,$15,$CE,$3C,$15,$C6,$7A,$57,$09,$CB,$6B,$81,$2B,$85,$74,$6B
	.dc.b	$E1,$39,$90,$D4,$F3,$8D,$D7,$96,$78,$1A,$38,$1F,$A3,$3F,$C9,$07
	.dc.b	$F9,$22,$BF,$C2,$4E,$23,$FC,$2B,$3A,$8D,$36,$33,$1B,$53,$CA,$3F
	.dc.b	$C3,$D1,$89,$E0,$D8,$E1,$3B,$7F,$CB,$71,$D1,$E3,$19,$7C,$27,$09
	.dc.b	$E7,$19,$9E,$13,$C9,$EC,$F2,$9D,$91,$E4,$43,$45,$39,$4F,$59,$F6
	.dc.b	$C8

;--------------------------------------------------------------------------------
;	FMOVE.L Dr,FPn
;	FMOVE.L <mem>,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmoveltoreg_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.L <EA>,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.L <ea>,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d2	source
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source,...
;	a4	expected result handle,expected status,...
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fmoveltoreg_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a0,a2			;source,...
@@:
	addq.l	#4,a0			;source
	tst.l	(a0)
	bne	@b
	addq.l	#4,a0			;0
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;Dr,FPn
	fmove.l	#0,fpcr
	move.l	(a2),d2			;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fmove.l	d2,fp7			;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)		;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.L Dr='
	puthex8	d2			;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
;<mem>,FPn
	fmove.l	#0,fpcr
	move.l	(a2),d2			;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	d2,-(sp)		;source
	fmove.l	(sp)+,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)		;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.L <mem>='
	puthex8	d2			;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source
	tst.l	(a2)			;source
	bne	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmoveltoreg_data_compressed::
	.dc.b	$00,$00,$64,$08,$10,$86,$1B,$18,$9C,$FE,$FD,$AF,$DF,$F7,$FC,$01
	.dc.b	$C3,$F8,$07,$91,$70,$B8,$63,$0C,$E1,$AC,$37,$87,$0C,$3C,$67,$3C
	.dc.b	$58,$E1,$63,$71,$C6,$39,$C7,$58,$EF,$1E,$67,$5C,$6E,$18,$C3,$38
	.dc.b	$6B,$0D,$E1,$C3,$07,$C2,$FC,$43,$E3,$63,$8C,$BE,$4E,$55,$65,$72
	.dc.b	$C6,$59,$CB,$59,$71,$3E,$8E,$5F,$03,$2C,$65,$9C,$75,$8E,$F1,$E1
	.dc.b	$8F,$1C,$7C,$A7,$D9,$C4,$FD,$96,$E5,$F2,$1B,$87,$08,$61,$70,$C6
	.dc.b	$19,$C3,$A9,$E6,$3A,$E5,$8C,$B3,$8E,$B0,$DE,$1C,$30,$E3,$83,$E1
	.dc.b	$BF,$1A,$7C,$A3,$F2,$18,$67,$2F,$C9,$E3,$5F,$2C,$FE,$57,$0C,$61
	.dc.b	$9C,$38,$1F,$9A,$9F,$94,$7E,$2B,$0C,$E1,$AC,$77,$87,$0C,$38,$E1
	.dc.b	$E5,$3F,$45,$87,$C8,$CB,$19,$7E,$42,$3E,$49,$F3,$30,$CE,$1A,$C3
	.dc.b	$78,$73,$3C,$87,$C8,$AF,$AA,$F3,$3E,$BE,$3B,$BE,$17,$C6,$F9,$53
	.dc.b	$F0,$5F,$DA,$9F,$92,$3E,$49,$FB,$9B,$D6,$3F,$BC,$C2,$AC,$2E,$18
	.dc.b	$C3,$67,$F0,$2C,$3F,$24,$7C,$93,$F6,$78,$6B,$0D,$DF,$0B,$E3,$7E
	.dc.b	$53,$F8,$55,$FC,$93,$F2,$39,$E7,$1F,$D2,$C7,$C9,$3F,$2B,$79,$BD
	.dc.b	$5F,$33,$C8,$7E,$42,$7E,$4D,$E7,$1D,$63,$BC,$38,$61,$C7,$07,$E0
	.dc.b	$DF,$C7,$0F,$DC,$9F,$90,$3E,$41,$FC,$86,$B7,$97,$F2,$4B,$AA,$ED
	.dc.b	$E8,$FE,$53,$7F,$B9,$3F,$20,$7C,$83,$F8,$DD,$EE,$F8,$57,$1A,$F2
	.dc.b	$9F,$CC,$6B,$F2,$07,$C8,$3F,$6E,$7E,$3C,$F8,$E7,$EF,$23,$8D,$FF
	.dc.b	$00,$8F,$C6,$BD,$27,$F9,$4C,$7C,$83,$F2,$38,$67,$0D,$61,$BC,$38
	.dc.b	$5B,$F4,$5F,$E8,$67,$F2,$03,$F6,$E7,$E3,$CF,$90,$7E,$D4,$FE,$95
	.dc.b	$1C,$71,$FE,$77,$1F,$CE,$4F,$E3,$C7,$ED,$4F,$C6,$9F,$18,$FE,$83
	.dc.b	$5C,$6B,$E8,$47,$F3,$A3,$F8,$DC,$F9,$63,$FA,$B4,$7E,$38,$FE,$3A
	.dc.b	$7E,$D4,$FC,$69,$F1,$CF,$E4,$6F,$13,$FB,$0B,$FB,$4D,$71,$9E,$72
	.dc.b	$D6,$5B,$C7,$86,$5C,$71,$7E,$8D,$FD,$A0,$FE,$8A,$7F,$1C,$3F,$68
	.dc.b	$7E,$34,$FE,$34,$7C,$63,$FB,$39,$FD,$74,$FE,$84,$7F,$19,$3F,$62
	.dc.b	$7E,$28,$F8,$A7,$F6,$68,$F0,$E7,$FC,$32,$3E,$29,$FC,$64,$FD,$89
	.dc.b	$FD,$00,$FE,$30,$7E,$C0,$FC,$51,$F1,$0F,$C4,$9F,$0C,$FD,$79,$F8
	.dc.b	$83,$F8,$CD,$FC,$0A,$FE,$9F,$75,$FD,$34,$FF,$1E,$7E,$BD,$7D,$8C
	.dc.b	$BE,$CE,$5F,$6B,$2F,$B7,$8F,$DC,$C7,$EE,$E5,$F7,$B1,$FA,$F8,$FD
	.dc.b	$8C,$7E,$CE,$3F,$6B,$1F,$B7,$8F,$DC,$C7,$EE,$E3,$F7,$B1,$FD,$1E
	.dc.b	$3F,$A4,$C7,$F4,$B8,$FE,$9B,$1F,$D3,$E3,$FA,$8C,$7F,$53,$8F,$EA
	.dc.b	$B1,$F5,$63,$EB,$C7,$D9,$8F,$B7,$1F,$76,$3E,$FC,$7E,$06,$3F,$07
	.dc.b	$1F,$F1,$5E,$3F,$E2,$DF,$45,$F0,$8E,$2C,$CA,$73,$18,$6C,$72,$6B
	.dc.b	$FE,$36,$6B,$D6,$7A,$3B,$6B,$B6,$BB,$9E,$D1,$88,$FF,$21,$CF,$46
	.dc.b	$C6,$57,$B1,$FC,$91,$FF,$23,$B5,$FF,$25,$35,$FF,$26,$B5,$E3,$C1
	.dc.b	$F0,$1A,$7C,$28,$D1,$23,$4D,$8D,$47,$1F,$65,$FA,$B5,$24,$FC,$C8
	.dc.b	$CB,$63,$2F,$D4,$8C,$B6,$32,$FF,$98,$A3,$2D,$8C,$BF,$E6,$7F,$71
	.dc.b	$E9,$7A,$DB,$1F,$71,$AF,$E0,$9A,$FF,$9B,$9A,$F3,$8E,$44,$79,$35
	.dc.b	$E4,$D7,$FC,$F1,$D9,$FE,$64,$F0,$FF,$16,$D8,$FC,$8B,$5F,$F3,$EB
	.dc.b	$5E,$3C,$69,$E2,$D8,$E2,$D7,$8B,$5E,$31,$C0,$8F,$07,$9F,$9F,$23
	.dc.b	$9B,$5E,$6D,$7C,$6D,$7F,$D2,$B5,$FE,$80,$8C,$42,$30,$F4,$3F,$5F
	.dc.b	$D8,$FD,$2D,$F9,$17,$EB,$A3,$0D,$8C,$CF,$06,$BF,$EA,$29,$DB,$99
	.dc.b	$E0,$47,$04,$38,$4F,$27,$FD,$59,$3C,$9F,$E0,$7C,$0F,$D8,$BF,$44
	.dc.b	$D8,$FE,$06,$D7,$71,$A2,$3A,$6B,$A3,$04,$30,$7D,$8F,$73,$EE,$91
	.dc.b	$FB,$AD,$74,$60,$86,$08,$7E,$21,$E0,$FE,$2B,$E1,$7F,$12,$47,$F1
	.dc.b	$2D,$7F,$12,$D7,$81,$A2,$1A,$29,$F9,$6E,$FF,$CA,$91,$FC,$AB,$5F
	.dc.b	$CA,$B5,$E0,$68,$86,$8A,$7E,$8B,$E6,$BF,$42,$47,$F4,$2D,$7F,$42
	.dc.b	$D7,$81,$A2,$1A,$29,$FA,$EF,$A6,$FD,$69,$1F,$D6,$B5,$FD,$6B,$5E
	.dc.b	$06,$88,$68,$A7,$EE,$BE,$CB,$F7,$24,$7F,$72,$D7,$F7,$2D,$78,$1A
	.dc.b	$21,$A2,$9F,$9A,$D1,$FA,$87,$EA,$49,$1F,$A9,$6B,$FA,$26,$BC,$5A
	.dc.b	$86,$BF,$B7,$8C,$C6,$27,$BB,$FD,$9B,$E3,$DF,$9C,$7F,$49,$6C,$66
	.dc.b	$30,$47,$04,$3F,$B9,$C7,$06,$C6,$DA,$6C,$FE,$4B,$F8,$07,$F2,$42
	.dc.b	$3F,$C9,$1A,$FF,$24,$6B,$C0,$D1,$0D,$14,$FE,$63,$F8,$67,$F3,$02
	.dc.b	$3F,$CC,$1A,$FF,$30,$6B,$C0,$D1,$0D,$14,$FE,$7B,$F8,$B7,$F3,$C2
	.dc.b	$3F,$CF,$1A,$FF,$3C,$6B,$C0,$D1,$0D,$14,$FE,$93,$F9,$17,$F4,$82
	.dc.b	$3F,$D2,$1A,$FF,$48,$6B,$C0,$D1,$0D,$14,$FE,$AB,$F9,$6F,$F5,$42
	.dc.b	$3F,$D5,$1A,$FF,$54,$6B,$C0,$D1,$0D,$14,$FE,$C3,$87,$F6,$02,$3F
	.dc.b	$D8,$1A,$FF,$60,$6B,$C0,$D1,$0D,$14,$FE,$DB,$FA,$27,$F6,$C2,$3F
	.dc.b	$DB,$1A,$FF,$6C,$6B,$C0,$D1,$0D,$14,$FE,$F3,$9F,$F7,$82,$3F,$DE
	.dc.b	$1A,$FF,$78,$6B,$C0,$D1,$0D,$14,$FD,$B0,$F3,$DF,$9C,$7F,$86,$23
	.dc.b	$0D,$8C,$F6,$FF,$97,$1B,$3F,$E6,$26,$BD,$0E,$31,$B9,$E4,$47,$04
	.dc.b	$8E,$51,$DF,$F3,$96,$BD,$DB,$3D,$DA,$F7,$B4,$62,$3D,$B1,$89,$E4
	.dc.b	$48,$E4,$D7,$94,$F1,$7F,$8D,$4F,$17,$67,$BF,$2F,$F5,$AB,$E1,$6B
	.dc.b	$E0,$6B,$C9,$AF,$CF,$6C,$7F,$20,$F3,$3E,$64,$60,$8E,$09,$3E,$17
	.dc.b	$FC,$77,$E0,$C8,$FF,$1D,$B5,$FF,$1D,$B5,$FF,$21,$B5,$FF,$18,$F8
	.dc.b	$1F,$E6,$26,$C6,$5F,$F1,$94,$65,$BD,$7F,$E4,$A2,$3F,$E4,$A6,$BF
	.dc.b	$E4,$A6,$BF,$E5,$26,$BE,$06,$BD,$0F,$F5,$3F,$23,$9F,$FD,$14,$D8
	.dc.b	$CC,$60,$8E,$08,$7F,$AD,$23,$83,$63,$6D,$36,$7E,$5F,$EA,$1F,$AA
	.dc.b	$7F,$A6,$5B,$1F,$EB,$A6,$BF,$EC,$06,$BC,$63,$81,$1E,$0D,$78,$35
	.dc.b	$E0,$7F,$B1,$BE,$75,$CB,$FE,$A9,$6C,$66,$30,$47,$04,$3F,$DA,$71
	.dc.b	$C1,$B1,$B6,$9B,$3F,$D0,$BF,$B7,$7F,$A1,$08,$FF,$A1,$1A,$FF,$A1
	.dc.b	$1A,$F0,$34,$43,$45,$3F,$D2,$3F,$BD,$7F,$A4,$08,$FF,$A4,$1E,$BA
	.dc.b	$7F,$11,$8F,$C7,$CF,$F1,$48,$C7,$86,$51,$9E,$1F,$F4,$4F,$A1,$FC
	.dc.b	$3E,$76,$30,$EF,$87,$FA,$71,$AF,$FA,$B2,$5E,$CF,$FA,$B1,$B1,$FE
	.dc.b	$AC,$6B,$FE,$AC,$6B,$F0,$4F,$25,$FE,$AB,$0F,$24,$7A,$8F,$F1,$16
	.dc.b	$1E,$AA,$F3,$57,$A2,$FC,$04,$AF,$03,$E3,$F8,$49,$1E,$36,$BF,$38
	.dc.b	$F0,$9D,$CB,$97,$C7,$C7,$C6,$B0,$F0,$CF,$FA,$B5,$57,$48,$F8,$8D
	.dc.b	$7E,$6F,$C7,$7C,$52,$3F,$15,$F9,$1F,$2B,$DA,$9F,$3D,$4C,$46,$67
	.dc.b	$8D,$7C,$B6,$BF,$4F,$E9,$FE,$91,$1F,$A4,$F6,$7E,$5E,$9F,$80,$BE
	.dc.b	$D1,$83,$D9,$7E,$3A,$E5,$1F,$65,$AF,$E0,$7E,$C7,$F0,$04,$7F,$00
	.dc.b	$F7,$7E,$5E,$3E,$C7,$B6,$7F,$2F,$0F,$6C,$62,$55,$CE,$3F,$10,$D7
	.dc.b	$F2,$1F,$79,$F8,$F2,$3F,$8F,$7C,$1B,$FE,$1E,$FD,$B3,$E2,$CB,$E7
	.dc.b	$57,$08,$FC,$A3,$5F,$CE,$7E,$0F,$F9,$B2,$3F,$9B,$6B,$F9,$B6,$BC
	.dc.b	$0D,$10,$D1,$4F,$D4,$7E,$25,$FA,$72,$3F,$A7,$7A,$E1,$F9,$F9,$FC
	.dc.b	$21,$F8,$5F,$C6,$FF,$0D,$3C,$2B,$F5,$8D,$7F,$69,$F9,$37,$EC,$C8
	.dc.b	$FE,$CD,$EB,$BF,$F0,$0B,$EB,$3F,$46,$FF,$25,$5C,$23,$F7,$0D,$7F
	.dc.b	$80,$7E,$61,$FB,$F2,$3F,$BF,$7A,$BF,$A5,$AE,$AF,$E0,$AB,$F3,$D3
	.dc.b	$C2,$3F,$82,$35,$FE,$19,$F9,$DF,$F0,$B2,$3F,$C2,$DF,$E1,$DF,$A2
	.dc.b	$7E,$9F,$C0,$93,$F8,$98,$FD,$54,$F0,$8F,$E2,$0D,$7F,$8C,$75,$FF
	.dc.b	$17,$23,$FC,$5D,$EA,$FE,$4E,$BA,$BF,$9A,$AF,$DB,$4F,$08,$FE,$38
	.dc.b	$D7,$F9,$27,$EC,$3F,$C8,$C8,$FF,$23,$7A,$BF,$A4,$AE,$BB,$7F,$1D
	.dc.b	$BF,$E0,$55,$C2,$BF,$94,$35,$FE,$61,$FB,$67,$F2,$F2,$3F,$CB,$DE
	.dc.b	$B5,$FC,$9E,$FA,$ED,$FC,$FB,$0F,$E1,$D5,$C2,$FF,$9A,$35,$FE,$79
	.dc.b	$FB,$BF,$F3,$B2,$3F,$CE,$DA,$FF,$3B,$6B,$C0,$D1,$0D,$14,$FE,$91
	.dc.b	$F0,$5F,$D1,$C8,$FF,$47,$7A,$E3,$FC,$FE,$7F,$92,$1F,$CC,$BD,$FF
	.dc.b	$EB,$D5,$FC,$32,$B8,$BF,$D3,$5A,$FF,$9D,$3A,$12,$F9,$1F,$F3,$64
	.dc.b	$FF,$BB,$67,$31,$88,$46,$1D,$B5,$F5,$9F,$C7,$2F,$CA,$FF,$1C,$AF
	.dc.b	$2E,$1F,$DC,$1F,$04,$72,$6B,$FE,$7D,$AE,$8F,$9E,$32,$DF,$2F,$C9
	.dc.b	$3E,$A6,$CF,$FA,$36,$31,$3F,$3E,$BE,$43,$FC,$AE,$FE,$46,$9F,$27
	.dc.b	$0E,$24,$9F,$23,$D3,$DD,$91,$E7,$6B,$E7,$C5,$EC,$FD,$3F,$B0,$93
	.dc.b	$F0,$63,$EB,$57,$08,$F6,$35,$FA,$14,$FB,$9F,$59,$23,$E3,$3E,$EA
	.dc.b	$EB,$1F,$DD,$25,$7F,$83,$9F,$C2,$57,$1A,$F3,$35,$F9,$9F,$81,$FC
	.dc.b	$82,$3F,$4D,$EC,$FD,$9A,$EF,$A7,$D9,$BF,$C8,$57,$0A,$F8,$4D,$7E
	.dc.b	$C7,$E1,$BF,$50,$8F,$DF,$7A,$BE,$9B,$EB,$A7,$C2,$BF,$CE,$57,$0A
	.dc.b	$F9,$CD,$7F,$09,$F8,$BF,$E0,$08,$FE,$2D,$AF,$D5,$6B,$C0,$D1,$0D
	.dc.b	$14,$FC,$87,$C1,$FE,$30,$8F,$E5,$DE,$B9,$7E,$6E,$7F,$64,$7E,$CF
	.dc.b	$F2,$9F,$DA,$4F,$0A,$FC,$33,$5F,$CE,$7C,$0F,$E6,$08,$FE,$8D,$EB
	.dc.b	$97,$E1,$2F,$AC,$FE,$AE,$FF,$80,$57,$08,$FC,$93,$5F,$D4,$7E,$79
	.dc.b	$FA,$42,$3F,$AF,$7A,$BF,$BC,$DD,$7D,$A3,$F1,$97,$FC,$36,$B8,$C7
	.dc.b	$E7,$DA,$FE,$D7,$F4,$4F,$D8,$91,$FD,$E3,$FC,$25,$7C,$0F,$F0,$2C
	.dc.b	$BF,$3F,$FA,$77,$F1,$93,$FC,$BB,$3A,$8F,$E3,$95,$CE,$3F,$5C,$D7
	.dc.b	$F8,$27,$A7,$FC,$00,$8F,$F0,$B7,$C1,$7D,$E3,$DD,$7D,$DF,$F1,$E1
	.dc.b	$DC,$FE,$2D,$3C,$9F,$DE,$35,$FE,$25,$FB,$07,$F0,$E2,$3F,$C6,$5E
	.dc.b	$F8,$75,$8F,$DA,$57,$B2,$FC,$5D,$1F,$CC,$2F,$8D,$7F,$09,$6B,$FC
	.dc.b	$7F,$F6,$AF,$E3,$64,$7F,$93,$3D,$9F,$E5,$97,$DF,$87,$FA,$0E,$1D
	.dc.b	$B7,$5F,$1B,$FE,$2C,$D4,$43,$E1,$1F,$D0,$A7,$FA,$5B,$93,$25,$72
	.dc.b	$F1,$6B,$D1,$AF,$03,$59,$7A,$5E,$CD,$7A,$3F,$CE,$BF,$74,$91,$FD
	.dc.b	$15,$AF,$03,$EC,$CF,$DD,$7D,$38,$7F,$4D,$FD,$FB,$FB,$25,$70,$9F
	.dc.b	$E9,$4D,$7F,$AF,$7C,$AF,$F6,$27,$FA,$DB,$5F,$EA,$4D,$7F,$B4,$35
	.dc.b	$F5,$9E,$2C,$3E,$F1,$FD,$CE,$BF,$D4,$B3,$B9,$FE,$D2,$DD,$4D,$27
	.dc.b	$C5,$FF,$5C,$4D,$24,$FF,$73,$CB,$E0,$13,$0F,$04,$7F,$A2,$AB,$8B
	.dc.b	$EC,$AE,$2F,$B1,$ED,$F1,$72,$3E,$0B,$5F,$83,$F1,$7D,$1F,$E9,$D9
	.dc.b	$7B,$23,$DB,$3B,$9F,$8A,$D7,$E6,$FC,$5F,$AC,$8F,$CE,$7D,$D3,$D7
	.dc.b	$5F,$B1,$3E,$98,$F9,$33,$C2,$3E,$9B,$5F,$A7,$9F,$C7,$23,$F6,$9E
	.dc.b	$AF,$F8,$47,$0E,$D9,$FE,$02,$FE,$AD,$70,$AF,$BE,$D7,$EF,$FC,$5F
	.dc.b	$D3,$23,$F8,$67,$AB,$F1,$0E,$C7,$A4,$FA,$4F,$07,$F1,$6D,$7F,$17
	.dc.b	$EE,$FD,$F2,$3F,$92,$6B,$F9,$26,$BC,$0D,$10,$D1,$4F,$CB,$FC,$5F
	.dc.b	$E2,$C8,$FE,$79,$EB,$9B,$1F,$D1,$D7,$E6,$0F,$CC,$FC,$4F,$E6,$AB
	.dc.b	$8D,$FE,$91,$AF,$E9,$3E,$13,$F3,$04,$7F,$56,$F6,$BE,$B1,$F9,$5B
	.dc.b	$EA,$FE,$5E,$78,$3F,$B0,$6B,$FB,$0F,$89,$FD,$21,$1F,$DB,$BD,$5F
	.dc.b	$D6,$CF,$A5,$FD,$16,$1F,$B2,$9E,$11,$FB,$C6,$BF,$BC,$F7,$FF,$60
	.dc.b	$47,$F8,$1B,$FB,$3F,$89,$F3,$C3,$3F,$90,$FE,$F2,$FF,$7F,$3F,$C0
	.dc.b	$2B,$94,$FF,$0A,$6B,$FC,$2B,$E1,$7F,$7C,$47,$F8,$8B,$DD,$F4,$90
	.dc.b	$F0,$67,$FC,$62,$BF,$86,$CF,$19,$FE,$2E,$D7,$F8,$BF,$C4,$FF,$0B
	.dc.b	$23,$FC,$79,$EC,$F4,$3D,$47,$7C,$7D,$BC,$DF,$C5,$AF,$F8,$DD,$72
	.dc.b	$AF,$E4,$AD,$7F,$92,$FC,$4F,$F1,$92,$3F,$CB,$1E,$F7,$D3,$9D,$A7
	.dc.b	$CC,$8C,$61,$FC,$A2,$B8,$DF,$F3,$26,$BF,$CC,$BE,$1F,$F9,$31,$1F
	.dc.b	$E7,$2D,$7F,$9C,$B5,$E0,$68,$86,$8A,$7F,$3E,$F8,$7F,$E6,$44,$7F
	.dc.b	$A2,$BD,$6F,$D5,$3F,$CF,$8F,$E7,$FF,$07,$FD,$02,$78,$57,$F4,$A6
	.dc.b	$BF,$D2,$BE,$17,$F9,$F1,$1F,$EA,$2F,$5D,$FF,$AA,$DF,$59,$FE,$79
	.dc.b	$7F,$D3,$2B,$84,$7F,$56,$6B,$FE,$BD,$F8,$BF,$F1,$86,$E6,$92,$3B
	.dc.b	$1F,$D5,$2F,$B3,$F8,$39,$FE,$AB,$7F,$E0,$5C,$BF,$07,$7C,$6B,$FD
	.dc.b	$82,$D7,$FD,$70,$BE,$1B,$FD,$6A,$D8,$DC,$60,$CF,$C3,$7C,$88,$F6
	.dc.b	$9F,$E1,$4A,$EF,$1E,$1C,$FF,$DA,$97,$CE,$BB,$DF,$37,$A9,$E4,$F8
	.dc.b	$6F,$09,$1C,$1E,$8B,$F0,$3E,$FE,$16,$5D,$E3,$FD,$BF,$82,$B1,$9F
	.dc.b	$CB,$C3,$A5,$7C,$06,$BF,$47,$E0,$BE,$09,$1F,$8E,$FC,$07,$DE,$78
	.dc.b	$70,$EB,$87,$78,$F8,$59,$FC,$0A,$E5,$1F,$39,$AF,$DF,$F8,$3F,$9E
	.dc.b	$47,$EA,$BD,$DE,$C7,$83,$5F,$9A,$F6,$3D,$F1,$C5,$FB,$6D,$7F,$19
	.dc.b	$F0,$DF,$70,$8F,$E0,$9A,$FE,$09,$AF,$03,$44,$34,$53,$F3,$1F,$0B
	.dc.b	$F8,$82,$3F,$8E,$7A,$D7,$E4,$A7,$E3,$9F,$23,$E1,$7E,$4C,$F0,$AF
	.dc.b	$C9,$B5,$FD,$27,$C1,$FE,$50,$8F,$E6,$9E,$B9,$7E,$3A,$FA,$CF,$C5
	.dc.b	$DB,$F5,$58,$7D,$6B,$E3,$3F,$A0,$6B,$FB,$1F,$85,$FD,$09,$1F,$D3
	.dc.b	$BD,$9F,$C6,$DF,$67,$F6,$1A,$AF,$F0,$75,$C6,$7F,$5A,$D7,$F7,$DF
	.dc.b	$09,$FA,$E2,$3F,$B4,$7F,$63,$F0,$BF,$16,$3D,$57,$D5,$B1,$F9,$09
	.dc.b	$E3,$1F,$BA,$6B,$FC,$2F,$E1,$7F,$76,$47,$F8,$0B,$D9,$EB,$87,$77
	.dc.b	$D7,$5F,$9C,$9E,$11,$FC,$19,$AF,$F1,$7F,$85,$FE,$0E,$47,$F8,$6B
	.dc.b	$D5,$F8,$B5,$D7,$7F,$E2,$77,$FA,$8A,$E1,$5F,$C4,$9A,$FF,$23,$F8
	.dc.b	$2F,$E2,$64,$7F,$8C,$BD,$6B,$F8,$ED,$F5,$DF,$F8,$7E,$1F,$B4,$AE
	.dc.b	$17,$FC,$79,$AF,$FA,$97,$E4,$DE,$5F,$FB,$7D,$B1,$FE,$A9,$3F,$D5
	.dc.b	$EE,$1C,$35,$E5,$1C,$48,$F1,$11,$B7,$FD,$70,$D7,$AC,$70,$7A,$BF
	.dc.b	$EF,$1F,$84,$6B,$49,$3D,$63,$2F,$7A,$F3,$67,$FE,$1E,$8F,$0D,$FF
	.dc.b	$BD,$A7,$39,$7F,$34,$BF,$F7,$C9,$CA,$BE,$03,$5F,$F6,$53,$5F,$F6
	.dc.b	$83,$5F,$F6,$9F,$EA,$AF,$5F,$E0,$42,$30,$FC,$36,$BF,$35,$AF,$CA
	.dc.b	$3F,$DF,$B7,$E6,$7E,$29,$FA,$2A,$FE,$7C,$F5,$AE,$F5,$CA,$3F,$DC
	.dc.b	$8D,$7C,$B1,$FD,$73,$E1,$E4,$7F,$62,$6B,$F4,$0F,$0E,$3F,$1C,$FE
	.dc.b	$7B,$7F,$CF,$A7,$73,$FD,$DD,$AF,$F5,$EF,$87,$FE,$D2,$47,$FB,$7B
	.dc.b	$FD,$B3,$E1,$F9,$C7,$AC,$FD,$EB,$FD,$2A,$38,$47,$F7,$46,$BF,$E0
	.dc.b	$5F,$85,$FE,$EA,$47,$FB,$DB,$D5,$E8,$7F,$14,$BF,$EA,$B5,$FD,$5A
	.dc.b	$78,$4F,$F8,$01,$AF,$F8,$47,$E1,$FF,$C0,$24,$7F,$C0,$EF,$57,$A1
	.dc.b	$DB,$6F,$F0,$6D,$7F,$62,$9E,$13,$FE,$0C,$6A,$27,$4B,$E5,$FF,$2D
	.dc.b	$19,$11,$9E,$0F,$D5,$36,$3F,$68,$D7,$FC,$33,$7F,$31,$F2,$9F,$CD
	.dc.b	$1F,$F0,$0D,$76,$8F,$0B,$5E,$ED,$9E,$91,$E3,$7C,$DF,$11,$E2,$93
	.dc.b	$49,$3E,$48,$CB,$63,$2F,$B4,$EE,$E1,$ED,$1C,$1D,$4F,$81,$E4,$75
	.dc.b	$F8,$8F,$11,$1C,$1A,$3F,$C8,$53,$CF,$2F,$F4,$1B,$71,$F8,$3F,$11
	.dc.b	$F1,$2F,$95,$F8,$29,$18,$8F,$2F,$C2,$FC,$12,$3E,$F7,$DB,$87,$6C
	.dc.b	$BE,$79,$F4,$30,$FA,$33,$C2,$BB,$35,$F8,$3F,$0D,$F3,$48,$FC,$B7
	.dc.b	$AE,$5F,$62,$FA,$CF,$F7,$DC,$98,$7D,$EB,$E3,$3E,$B6,$BF,$3B,$E2
	.dc.b	$3E,$D1,$1F,$B0,$FD,$DF,$88,$F4,$47,$5A,$FC,$4C,$7E,$2A,$78,$4F
	.dc.b	$C7,$6B,$F6,$BE,$23,$F0,$C4,$7F,$08,$F5,$7A,$10,$FA,$F3,$F9,$68
	.dc.b	$E1,$1F,$4D,$AF,$E1,$BE,$1B,$F2,$44,$7F,$20,$F5,$7F,$03,$7E,$A7
	.dc.b	$A9,$F1,$E3,$83,$F7,$DA,$FE,$4B,$E2,$3F,$3C,$47,$F3,$8F,$5D,$BF
	.dc.b	$17,$5D,$5F,$C6,$E1,$FA,$E9,$E1,$5F,$8B,$6B,$F9,$EE,$8F,$D5,$11
	.dc.b	$FD,$43,$5F,$C9,$35,$E0,$68,$86,$8A,$7E,$AB,$E2,$3F,$6C,$47,$F6
	.dc.b	$8F,$5B,$FD,$1C,$FF,$04,$3F,$82,$FC,$47,$F0,$69,$E1,$5F,$A3,$6B
	.dc.b	$FB,$6F,$88,$FE,$04,$47,$F8,$03,$D7,$87,$F8,$6D,$F5,$9F,$CB,$5F
	.dc.b	$F1,$2A,$E1,$1F,$AF,$6B,$FC,$0B,$E1,$BF,$87,$11,$FE,$18,$F5,$7F
	.dc.b	$8B,$D7,$57,$F7,$F5,$FC,$7A,$78,$47,$EE,$DA,$FF,$0E,$F8,$8F,$E3
	.dc.b	$44,$7F,$8C,$3F,$C7,$7E,$23,$E1,$C7,$A0,$FD,$73,$FC,$AA,$38,$47
	.dc.b	$F0,$76,$BF,$C6,$BE,$1F,$F9,$31,$1F,$E4,$8F,$57,$A1,$E8,$3A,$9F
	.dc.b	$BB,$78,$3F,$C4,$DA,$FF,$26,$F8,$7F,$E6,$44,$7F,$98,$3D,$5E,$84
	.dc.b	$D7,$F9,$0D,$FF,$42,$9E,$13,$FC,$7D,$AF,$F3,$2F,$85,$FE,$7C,$47
	.dc.b	$F9,$E3,$D6,$BF,$95,$E5,$F1,$A7,$F9,$7D,$FF,$4E,$AE,$11,$FC,$AD
	.dc.b	$AF,$F4,$BD,$5F,$11,$FD,$71,$B1,$FD,$71,$AF,$F4,$26,$BC,$63,$81
	.dc.b	$1E,$0D,$78,$35,$E0,$75,$F8,$8F,$01,$1C,$1C,$8F,$F5,$44,$F8,$B1
	.dc.b	$3F,$DE,$0C,$1F,$DD,$FF,$04,$7C,$27,$FB,$BA,$17,$D6,$FD,$4D,$7E
	.dc.b	$17,$8B,$F0,$48,$FC,$17,$D5,$87,$82,$3F,$C0,$75,$E0,$6E,$3F,$E4
	.dc.b	$2C,$7C,$31,$E8,$BE,$6F,$C9,$6B,$F4,$7E,$23,$E8,$11,$FA,$0F,$82
	.dc.b	$FD,$D7,$EB,$3B,$65,$ED,$9E,$15,$F5,$5A,$FD,$DF,$89,$FB,$84,$7E
	.dc.b	$E3,$FE,$64,$F8,$97,$2F,$49,$FF,$7D,$0C,$FD,$6B,$4F,$F3,$D5,$70
	.dc.b	$8F,$AB,$7D,$5F,$C3,$35,$FF,$07,$71,$F2,$6E,$5F,$E7,$48,$C4,$7F
	.dc.b	$8B,$DA,$FF,$8D,$6B,$B3,$E1,$D3,$EF,$57,$9E,$30,$6F,$2F,$F3,$9D
	.dc.b	$F2,$83,$49,$3F,$E8,$4F,$D4,$9E,$D5,$25,$7F,$A9,$E3,$2F,$FA,$A7
	.dc.b	$1F,$1E,$7F,$E9,$BD,$3E,$2B,$70,$FF,$41,$CE,$B3,$FF,$25,$61,$CB
	.dc.b	$0F,$F5,$D3,$5F,$F1,$93,$5D,$E3,$F5,$7C,$39,$1E,$16,$BF,$EC,$3B
	.dc.b	$EC,$FF,$96,$F2,$F1,$E1,$F1,$A3,$ED,$CE,$93,$FE,$2C,$7A,$10,$E9
	.dc.b	$5F,$EA,$8F,$11,$E7,$7F,$94,$C9,$7F,$E6,$26,$D7,$F9,$89,$AF,$F9
	.dc.b	$21,$AF,$F9,$65,$AF,$F9,$C1,$FF,$30,$7C,$D3,$C3,$A4,$AF,$F6,$C4
	.dc.b	$65,$F0,$1F,$E7,$9D,$BF,$68,$BF,$80,$1E,$C7,$ED,$E7,$FD,$A1,$97
	.dc.b	$AA,$B8,$1F,$E8,$13,$84,$ED,$FF,$40,$5F,$7C,$BF,$DE,$0D,$7F,$CD
	.dc.b	$0D,$77,$8F,$F9,$E7,$F5,$27,$B5,$49,$7F,$EF,$18,$CB,$EA,$AF,$CC
	.dc.b	$BE,$23,$F6,$38,$F9,$E3,$39,$FB,$27,$F6,$B5,$CA,$7F,$CF,$41,$C3
	.dc.b	$FE,$7D,$6B,$FE,$9B,$6B,$FE,$A1,$EF,$7C,$6F,$FA,$45,$B5,$D2,$51
	.dc.b	$87,$BB,$5F,$BD,$1F,$31,$FE,$6B,$7E,$77,$F6,$A7,$FA,$AB,$6F,$9C
	.dc.b	$EB,$5A,$F6,$C3,$C4,$4B,$F1,$61,$DE,$7F,$A0,$70,$7F,$AA,$73,$FE
	.dc.b	$86,$DA,$FC,$E9,$FD,$65,$C5,$F8,$27,$FA,$4B,$E2,$9F,$A1,$5F,$C2
	.dc.b	$A7,$31,$FD,$42,$BA,$11,$E8,$7F,$65,$F8,$CF,$01,$1C,$1F,$DA,$6F
	.dc.b	$C0,$FC,$9E,$0F,$CB,$E9,$F1,$5F,$E9,$98,$78,$31,$F1,$53,$0E,$91
	.dc.b	$FD,$C5,$AF,$F7,$9F,$8D,$F8,$04,$9F,$EF,$0F,$C0,$7F,$B4,$E3,$EF
	.dc.b	$C7,$C1,$1F,$DA,$63,$CB,$3C,$6B,$FB,$F3,$5F,$F0,$37,$C4,$FC,$D2
	.dc.b	$57,$F8,$15,$ED,$1E,$8A,$ED,$AF,$C0,$7E,$0C,$F0,$9F,$F0,$53,$5F
	.dc.b	$F0,$97,$3F,$D9,$25,$FF,$84,$5B,$1F,$E1,$16,$BC,$23,$44,$74,$D7
	.dc.b	$FC,$3D,$F1,$BF,$85,$24,$7F,$87,$5E,$BA,$FD,$28,$7C,$6F,$D8,$BF
	.dc.b	$B2,$FD,$AA,$E3,$7F,$E2,$36,$BF,$E2,$BF,$8B,$FC,$91,$2F,$FC,$54
	.dc.b	$F6,$7F,$C3,$15,$D5,$FF,$0E,$BF,$86,$8E,$13,$FE,$2F,$6B,$FE,$37
	.dc.b	$F8,$DF,$CF,$12,$BF,$C6,$CF,$58,$FF,$0A,$D7,$57,$F1,$6F,$E4,$A3
	.dc.b	$84,$7F,$8E,$DA,$FF,$90,$FE,$37,$F5,$44,$AF,$F2,$13,$FA,$DF,$8D
	.dc.b	$FC,$D9,$F1,$23,$FC,$6A,$FE,$7A,$38,$4F,$F9,$1D,$AF,$F9,$3F,$E2
	.dc.b	$FF,$6C,$4A,$FF,$27,$3D,$5E,$67,$C4,$2C,$7E,$92,$38,$3F,$E5,$36
	.dc.b	$BF,$E5,$BF,$8D,$FE,$04,$49,$FF,$2D,$3D,$63,$F5,$B7,$E9,$6C,$FF
	.dc.b	$93,$A7,$84,$7F,$97,$DA,$FF,$99,$FE,$27,$F8,$71,$27,$FC,$CC,$F5
	.dc.b	$DB,$F1,$57,$D5,$B3,$FE,$5A,$9E,$15,$FE,$6B,$6B,$FE,$73,$F8,$DF
	.dc.b	$E3,$44,$9F,$F3,$93,$63,$FC,$E4,$D7,$84,$68,$8E,$9A,$89,$CC,$3F
	.dc.b	$B1,$39,$32,$E1,$C3,$B7,$F9,$83,$97,$FC,$E5,$3F,$CB,$61,$F1,$BF
	.dc.b	$C8,$23,$0E,$67,$89,$2B,$FA,$CD,$75,$6A,$F8,$CE,$E4,$8C,$BF,$0F
	.dc.b	$1F,$E5,$33,$E0,$9F,$F4,$3E,$1E,$07,$FA,$A4,$7B,$EB,$95,$FB,$CF
	.dc.b	$85,$1D,$E7,$A9,$E2,$F0,$BD,$E4,$70,$7C,$EC,$3C,$38,$FB,$E3,$BD
	.dc.b	$78,$EF,$FA,$63,$FE,$4E,$8E,$31,$E1,$8F,$A5,$7D,$63,$10,$F8,$CF
	.dc.b	$9E,$4A,$F8,$AF,$D2,$F8,$CF,$A0,$7B,$EB,$BB,$F4,$E3,$84,$74,$6B
	.dc.b	$F7,$29,$F1,$9F,$70,$97,$F4,$9E,$CF,$B2,$FB,$65,$F4,$1F,$C0,$4F
	.dc.b	$0B,$F4,$B5,$ED,$E0,$FE,$20,$97,$F7,$9E,$B1,$E8,$8F,$93,$9F,$B1
	.dc.b	$F9,$73,$C2,$7E,$2B,$5F,$57,$C4,$7E,$50,$93,$F8,$A7,$AE,$3E,$78
	.dc.b	$EA,$7A,$CF,$AF,$1C,$27,$E8,$B5,$F8,$DD,$7F,$A0,$24,$FE,$59,$B1
	.dc.b	$F6,$1A,$F0,$8D,$11,$D3,$5F,$A5,$F1,$9F,$AC,$24,$7E,$89,$EB,$7F
	.dc.b	$9E,$F8,$8F,$D7,$5F,$EB,$DF,$D8,$4F,$0B,$FC,$4B,$5F,$BD,$F1,$3F
	.dc.b	$B8,$26,$1F,$AE,$7A,$BF,$C0,$2B,$AB,$F9,$A7,$F7,$91,$C2,$7F,$2A
	.dc.b	$D7,$F1,$5F,$17,$FC,$10,$95,$FB,$A7,$AC,$7E,$BA,$BA,$BF,$A4,$7F
	.dc.b	$84,$47,$08,$FD,$0B,$5F,$CB,$7C,$5F,$F1,$02,$57,$F0,$67,$FD,$E9
	.dc.b	$1F,$EA,$AE,$3F,$F2,$5F,$99,$FB,$B8,$ED,$5F,$ED,$C8,$C6,$9F,$C2
	.dc.b	$2F,$9D,$7E,$C1,$AF,$E9,$7C,$7F,$E3,$E4,$BF,$E2,$AF,$82,$3F,$D7
	.dc.b	$B8,$78,$0F,$A9,$3D,$DF,$E4,$93,$C6,$7F,$7A,$D7,$FD,$82,$73,$3F
	.dc.b	$9B,$61,$FC,$B5,$A6,$43,$96,$CF,$86,$BE,$C1,$E1,$3E,$04,$71,$9E
	.dc.b	$04,$71,$C1,$27,$FD,$C0,$D8,$FF,$6B,$D7,$57,$E5,$9F,$E7,$CA,$F9
	.dc.b	$75,$FE,$DE,$AE,$13,$FE,$DC,$6B,$FD,$33,$E1,$FF,$A4,$92,$7F,$9F
	.dc.b	$B6,$3F,$90,$B5,$FE,$2C,$D7,$94,$70,$23,$FC,$A3,$E2,$7F,$AA,$92
	.dc.b	$3F,$A5,$BD,$73,$FE,$8F,$F1,$3F,$D5,$EF,$FA,$C3,$FD,$66,$78,$5F
	.dc.b	$3B,$E2,$7F,$54,$F8,$7F,$EC,$24,$70,$78,$B0,$EA,$FF,$42,$C3,$AB
	.dc.b	$FD,$12,$78,$57,$F3,$E6,$BF,$CC,$FE,$27,$FB,$69,$27,$FB,$1B,$E9
	.dc.b	$7F,$BA,$E3,$E9,$7F,$A7,$BF,$DC,$63,$84,$FF,$4A,$6B,$FC,$FF,$E2
	.dc.b	$7F,$BC,$92,$BF,$B7,$BF,$EA,$BF,$41,$1F,$EA,$76,$BD,$8F,$ED,$6E
	.dc.b	$27,$FA,$EC,$F3,$7F,$AC,$B5,$FF,$04,$C3,$E2,$3B,$12,$BF,$C0,$31
	.dc.b	$96,$9E,$13,$E4,$CF,$FA,$D6,$3C,$19,$FF,$AA,$1E,$11,$FD,$AD,$AF
	.dc.b	$82,$BB,$39,$F4,$BD,$C4,$BF,$0B,$F0,$5F,$01,$EF,$8F,$03,$F0,$E3
	.dc.b	$84,$73,$8E,$27,$AB,$E1,$3E,$51,$1C,$1E,$C3,$E4,$9E,$CC,$3D,$4F
	.dc.b	$CF,$8E,$13,$EB,$6B,$F2,$BE,$27,$EB,$12,$7E,$4B,$63,$E1,$35,$E1
	.dc.b	$1A,$23,$A6,$BF,$5B,$D4,$FC,$11,$23,$EA,$BD,$6F,$E8,$7C,$47,$E1
	.dc.b	$6F,$F0,$CF,$E1,$E7,$85,$FD,$36,$BF,$82,$F8,$6F,$C7,$13,$0F,$C0
	.dc.b	$BD,$5F,$F2,$AE,$8B,$ED,$1F,$90,$7F,$29,$3C,$6B,$F0,$0D,$7F,$1F
	.dc.b	$F1,$1F,$9B,$25,$7E,$39,$ED,$1F,$E6,$1B,$ED,$3F,$9C,$C7,$D6,$FE
	.dc.b	$7A,$B8,$D7,$E3,$5A,$FE,$73,$E2,$3F,$50,$49,$FC,$DB,$FA,$BF,$88
	.dc.b	$FD,$51,$D2,$7F,$34,$FE,$B6,$38,$47,$E6,$5A,$FE,$A3,$E1,$FF,$68
	.dc.b	$4A,$FD,$3B,$D5,$FD,$9D,$75,$D7,$F4,$8F,$EE,$67,$85,$7E,$95,$AF
	.dc.b	$ED,$3E,$23,$F8,$01,$2F,$F6,$6F,$58,$F3,$C7,$D1,$3B,$9F,$C0,$A3
	.dc.b	$84,$7E,$C5,$AF,$F0,$0F,$86,$FE,$18,$49,$FD,$FB,$D7,$0F,$3C,$75
	.dc.b	$9E,$E7,$EF,$E7,$83,$FB,$D6,$BF,$C3,$3D,$FF,$E3,$04,$9F,$E1,$6D
	.dc.b	$8F,$E0,$8D,$78,$46,$88,$E9,$AF,$F1,$8F,$73,$F9,$21,$23,$F8,$BB
	.dc.b	$D6,$FF,$89,$7B,$FF,$C9,$EF,$F9,$43,$FC,$A6,$78,$5F,$F1,$56,$BF
	.dc.b	$C9,$3E,$1B,$F9,$81,$30,$FE,$46,$F5,$7F,$47,$5D,$5F,$E6,$2F,$F3
	.dc.b	$58,$E1,$3F,$C8,$5A,$FF,$30,$F8,$8F,$E7,$84,$AF,$E5,$EF,$58,$FD
	.dc.b	$15,$75,$7F,$97,$3F,$D0,$63,$84,$7F,$2D,$6B,$FC,$F3,$E2,$3F,$A4
	.dc.b	$12,$BF,$9D,$BF,$D2,$FE,$23,$FA,$51,$F1,$23,$FA,$4B,$FD,$36,$38
	.dc.b	$4F,$F3,$96,$BF,$ED,$0F,$9D,$7C,$3F,$F8,$97,$D0,$46,$30,$FF,$64
	.dc.b	$B6,$BF,$D5,$8D,$7F,$A8,$E1,$E2,$9F,$92,$78,$8F,$F6,$2C,$EE,$7F
	.dc.b	$DC,$AD,$7F,$DC,$CD,$AF,$0D,$F8,$E1,$F1,$34,$97,$E4,$7E,$13,$DE
	.dc.b	$FE,$6B,$77,$FE,$DF,$1D,$A7,$8B,$EA,$6B,$F0,$FE,$27,$C0,$4A,$F8
	.dc.b	$2F,$68,$FE,$CD,$97,$CF,$6C,$FB,$A7,$84,$7C,$66,$BF,$3F,$E2,$7D
	.dc.b	$84,$9F,$9A,$D8,$F9,$AD,$78,$46,$88,$E9,$AF,$DB,$F8,$9F,$90,$48
	.dc.b	$FB,$2F,$5C,$7E,$1E,$9F,$7B,$E1,$FE,$2C,$63,$46,$38,$C3,$EB,$D5
	.dc.b	$C7,$EB,$E3,$D5,$FC,$1B,$5F,$C6,$7C,$27,$D9,$25,$7E,$29,$F1,$3D
	.dc.b	$0F,$C2,$DF,$5C,$BC,$91,$F8,$89,$E3,$5F,$90,$6B,$F9,$9F,$87,$FC
	.dc.b	$31,$27,$F2,$EF,$68,$FC,$6E,$3D,$72,$F6,$D7,$E5,$23,$15,$C6,$BF
	.dc.b	$3A,$D7,$F4,$DF,$0D,$F9,$32,$57,$E9,$1F,$F4,$47,$73,$BF,$E3,$52
	.dc.b	$31,$5D,$C4,$36,$FF,$56,$CE,$A3,$13,$FA,$6B,$EF,$1F,$B0,$6B,$FB
	.dc.b	$9F,$86,$FD,$31,$27,$F6,$EF,$93,$0F,$1D,$F5,$C3,$EF,$65,$89,$F8
	.dc.b	$95,$C7,$0F,$DE,$B5,$FE,$0D,$F0,$DF,$B3,$26,$3F,$C1,$1E,$CF,$B3
	.dc.b	$0F,$5D,$FC,$FC,$3B,$4F,$F0,$2A,$E3,$7F,$C2,$9A,$FF,$13,$F8,$6F
	.dc.b	$E0,$04,$9F,$E2,$2F,$68,$F8,$A7,$CA,$CB,$D9,$3F,$8B,$8C,$9F,$47
	.dc.b	$0F,$C0,$57,$38,$FE,$32,$D7,$E9,$9C,$E3,$B3,$83,$94,$D7,$4D,$72
	.dc.b	$66,$BE,$D3,$C5,$AF,$66,$BC,$8E,$13,$E0,$74,$FF,$21,$F8,$59,$5F
	.dc.b	$CC,$9B,$3C,$23,$F9,$6F,$C3,$7C,$E8,$F8,$BC,$3F,$CE,$A3,$F4,$15
	.dc.b	$C2,$7F,$A0,$35,$FE,$67,$F0,$DF,$CB,$C9,$7F,$D0,$1F,$A1,$AF,$F0
	.dc.b	$5B,$EC,$D9,$FE,$7B,$3C,$2B,$FA,$2B,$5F,$E9,$DF,$0D,$FC,$EC,$93
	.dc.b	$FD,$31,$EA,$FF,$44,$BE,$AD,$3F,$A4,$C7,$09,$FE,$A2,$D7,$FA,$D7
	.dc.b	$C2,$7F,$47,$24,$FF,$58,$7F,$A5,$7C,$37,$F0,$13,$F8,$17,$C3,$59
	.dc.b	$FE,$AB,$3C,$2B,$FA,$EB,$5F,$EC,$DF,$0D,$FD,$4C,$93,$FD,$91,$EA
	.dc.b	$FF,$5E,$BE,$AD,$3F,$B0,$C7,$09,$FE,$D2,$D7,$FB,$97,$C3,$7F,$5F
	.dc.b	$24,$FF,$70,$7A,$C7,$DC,$E7,$5F,$68,$C3,$FC,$76,$78,$C7,$F7,$66
	.dc.b	$BF,$9E,$DD,$EB,$FF,$12,$C3,$F3,$91,$89,$FF,$01,$4E,$5F,$D1,$46
	.dc.b	$67,$FC,$07,$7F,$26,$7C,$99,$DD,$FF,$A6,$E1,$C6,$7C,$0D,$7F,$90
	.dc.b	$E1,$CD,$EB,$19,$51,$E6,$8F,$19,$30,$F3,$BE,$A6,$B9,$9E,$6D,$7C
	.dc.b	$2D,$7C,$2F,$17,$71,$D3,$D9,$F8,$24,$70,$76,$3F,$8F,$1C,$BE,$1B
	.dc.b	$FA,$C5,$73,$C3,$E0,$C6,$2B,$95,$FB,$9A,$FC,$4F,$86,$F9,$E4,$C3
	.dc.b	$DA,$FA,$EF,$9C,$78,$27,$E2,$E1,$DE,$3C,$46,$71,$FB,$B7,$CE,$BE
	.dc.b	$6B,$5F,$A7,$F0,$DF,$78,$95,$F3,$1F,$03,$D2,$3F,$85,$DF,$77,$C3
	.dc.b	$97,$C0,$AE,$31,$F6,$9A,$FE,$03,$E1,$BF,$16,$4B,$FB,$2F,$E3,$BE
	.dc.b	$1B,$E4,$9F,$2B,$E1,$AC,$FA,$F1,$57,$1B,$FC,$3B,$5F,$C6,$FC,$37
	.dc.b	$E6,$09,$5F,$86,$7B,$3C,$F0,$EC,$D3,$B4,$70,$9F,$C9,$B5,$FC,$CF
	.dc.b	$C3,$7E,$90,$93,$F9,$27,$AC,$7C,$6A,$EA,$D9,$F5,$4F,$08,$FC,$FB
	.dc.b	$5F,$D2,$FC,$2F,$EC,$09,$3F,$9E,$7A,$EF,$FA,$CB,$EA,$D9,$F8,$D3
	.dc.b	$C2,$BF,$56,$D7,$F6,$3F,$0D,$FB,$C2,$4F,$EA,$9B,$1F,$B2,$6B,$C2
	.dc.b	$34,$47,$4D,$7F,$7B,$F0,$DF,$C2,$09,$1F,$B6,$7A,$DF,$F0,$4F,$86
	.dc.b	$FD,$3D,$FE,$A1,$FD,$4C,$F0,$BF,$E0,$6D,$7F,$84,$FC,$37,$F1,$42
	.dc.b	$61,$FC,$09,$EA,$FE,$F2,$BA,$BF,$89,$7F,$6B,$1C,$27,$F8,$7B,$5F
	.dc.b	$E2,$BF,$09,$FC,$80,$95,$FC,$39,$EB,$1F,$B0,$AE,$AF,$E9,$1F,$E0
	.dc.b	$31,$C2,$3F,$8D,$B5,$FE,$43,$F0,$BF,$CB,$09,$5F,$C6,$9F,$E5,$FF
	.dc.b	$09,$FC,$28,$FE,$17,$F0,$B6,$7F,$45,$3C,$2B,$F9,$3B,$5F,$E5,$BF
	.dc.b	$0B,$FC,$E0,$93,$FC,$99,$EA,$FE,$AE,$FA,$B4,$FD,$74,$70,$9F,$E6
	.dc.b	$6D,$7F,$9C,$FC,$2F,$F4,$42,$4F,$F3,$27,$AC,$7F,$06,$AE,$AD,$9F
	.dc.b	$DD,$4F,$08,$FE,$7E,$D7,$FA,$A7,$CF,$27,$FD,$A4,$D7,$FA,$AB,$63
	.dc.b	$FA,$9B,$5F,$E9,$B5,$F1,$DF,$E2,$F8,$DD,$FF,$D3,$F5,$B9,$FE,$A2
	.dc.b	$D7,$FA,$7B,$7A,$8D,$1F,$0D,$E4,$26,$3E,$46,$D7,$91,$AF,$B1,$AF
	.dc.b	$81,$AF,$81,$F7,$FC,$34,$9F,$7B,$67,$A4,$7F,$3D,$A7,$C3,$7F,$A9
	.dc.b	$DF,$2C,$66,$31,$5C,$6B,$E6,$B5,$F8,$3F,$09,$F3,$89,$5F,$39,$F5
	.dc.b	$E5,$F2,$2F,$B6,$3D,$DF,$04,$F0,$BF,$A6,$D7,$ED,$FC,$37,$DA,$26
	.dc.b	$1F,$69,$EA,$FD,$2A,$EA,$FA,$4F,$74,$70,$9F,$BE,$D7,$F0,$FF,$0D
	.dc.b	$F8,$62,$4F,$E1,$9F,$C4,$F9,$9F,$14,$F8,$DF,$0D,$67,$E5,$4F,$0A
	.dc.b	$FC,$5B,$5F,$C9,$FC,$37,$E4,$89,$3F,$92,$7A,$BF,$4E,$FA,$B4,$FA
	.dc.b	$D1,$C2,$7F,$2E,$D7,$F3,$FF,$0D,$F9,$E2,$4F,$E7,$9E,$B1,$FE,$7A
	.dc.b	$DF,$FD,$3A,$FE,$5B,$1F,$53,$99,$E2,$FE,$91,$AF,$EB,$3E,$1B,$F5
	.dc.b	$64,$BF,$D5,$BD,$B7,$FD,$2E,$5E,$AA,$F0,$3F,$8D,$9E,$11,$FB,$06
	.dc.b	$BF,$EA,$5F,$29,$EA,$7E,$E6,$3F,$D4,$D1,$8C,$3F,$D4,$CD,$8C,$CE
	.dc.b	$DE,$4D,$7F,$D6,$71,$B7,$33,$CA,$75,$1F,$EE,$A9,$D3,$C2,$32,$A3
	.dc.b	$2E,$E3,$31,$FE,$6A,$9F,$2F,$95,$82,$38,$25,$7B,$23,$2D,$8C,$BF
	.dc.b	$0E,$32,$FF,$B2,$7F,$42,$7B,$FE,$46,$D7,$FB,$25,$B3,$FE,$CD,$6B
	.dc.b	$C6,$78,$12,$78,$36,$38,$35,$E1,$1F,$ED,$5F,$8D,$FF,$67,$C1,$A4
	.dc.b	$CF,$E3,$B6,$78,$46,$08,$E0,$93,$FE,$E3,$9F,$F6,$DF,$3F,$F2,$8F
	.dc.b	$88,$FE,$72,$4A,$FE,$72,$DA,$FE,$72,$D7,$FA,$0B,$5F,$E6,$AD,$7F
	.dc.b	$9A,$BF,$D1,$3E,$22,$4F,$F4,$46,$CF,$F4,$46,$BF,$D3,$1A,$FF,$40
	.dc.b	$6B,$FD,$01,$B0,$FC,$85,$ED,$FE,$3A,$30,$D8,$CC,$2A,$1F,$EA,$B8
	.dc.b	$CB,$83,$FA,$E3,$87,$71,$D0,$D3,$A9,$EC,$48,$C9,$B8,$E8,$D7,$9C
	.dc.b	$F5,$7F,$D7,$92,$ED,$F5,$36,$B9,$B6,$79,$B5,$F2,$B5,$E7,$3C,$89
	.dc.b	$3C,$9B,$1C,$9A,$FF,$7C,$F8,$9F,$EF,$64,$9F,$EF,$6D,$8F,$EF,$6D
	.dc.b	$78,$46,$88,$E9,$AF,$F8,$23,$E2,$7F,$C0,$E4,$8F,$F0,$3B,$63,$FC
	.dc.b	$0E,$D7,$84,$68,$8E,$9A,$FF,$85,$3E,$23,$FC,$26,$48,$FF,$09,$B6
	.dc.b	$3F,$C2,$6D,$78,$46,$88,$E9,$AF,$F8,$83,$E2,$7F,$C3,$E4,$8F,$F0
	.dc.b	$FB,$63,$FC,$3E,$D7,$84,$68,$8E,$9A,$FF,$8B,$3E,$27,$FC,$56,$48
	.dc.b	$FF,$15,$B6,$3F,$C5,$6D,$78,$46,$88,$E9,$AF,$F8,$E3,$E2,$7F,$C6
	.dc.b	$E4,$8F,$F1,$BB,$63,$FC,$6E,$D7,$84,$68,$8E,$9A,$FF,$91,$3E,$23
	.dc.b	$FC,$86,$48,$FF,$21,$B6,$3F,$C8,$6D,$78,$46,$88,$E9,$AF,$EE,$38
	.dc.b	$1F,$43,$FC,$C2,$3F,$79,$18,$E4,$FE,$9E,$DA,$FD,$BB,$67,$F7,$CD
	.dc.b	$7A,$57,$35,$19,$7A,$C6,$63,$11,$A8,$C1,$2B,$FA,$DA,$F8,$7A,$4B
	.dc.b	$FE,$13,$5E,$A8,$C3,$FC,$32,$78,$4E,$9B,$19,$9D,$BF,$C5,$E3,$2D
	.dc.b	$8C,$BF,$C7,$A1,$EF,$79,$09,$5D,$5B,$3E,$16,$BD,$5A,$F3,$6B,$FC
	.dc.b	$AE,$7A,$CF,$F2,$4F,$91,$7B,$EF,$13,$FC,$03,$18,$6C,$66,$78,$36
	.dc.b	$FF,$99,$56,$E7,$35,$C0,$8E,$09,$3C,$2B,$93,$FC,$FA,$B9,$3F,$80
	.dc.b	$B7,$C5,$FE,$19,$B7,$F8,$66,$CE,$E7,$44,$9D,$36,$35,$18,$23,$83
	.dc.b	$C1,$F1,$7D,$C9,$1D,$DB,$1A,$8C,$11,$C1,$0F,$CB,$C7,$07,$F3,$39
	.dc.b	$FA,$C9,$3E,$B6,$C7,$C1,$6B,$C2,$34,$47,$4D,$7F,$4B,$F1,$7F,$1C
	.dc.b	$91,$F1,$DB,$1F,$35,$AF,$08,$D1,$1D,$35,$FD,$8F,$81,$F4,$C9,$1F
	.dc.b	$4D,$B1,$F6,$5A,$F0,$8D,$11,$D3,$5F,$EF,$98,$38,$3F,$C9,$ED,$9F
	.dc.b	$D9,$B6,$7F,$BE,$B5,$E3,$3C,$09,$3C,$1B,$1C,$1A,$F0,$8F,$DF,$FD
	.dc.b	$1B,$E2,$3F,$CB,$ED,$8C,$C6,$08,$E0,$93,$FE,$10,$9E,$0D,$9D,$B6
	.dc.b	$76,$FE,$87,$67,$91,$FB,$02,$4F,$EC,$1B,$3F,$A9,$6B,$C5,$AF,$F8
	.dc.b	$A7,$7F,$DE,$C6,$21,$18,$9F,$D5,$7C,$67,$EB,$48,$E0,$93,$FB,$38
	.dc.b	$CB,$63,$2F,$F3,$48,$CB,$63,$F9,$4F,$D5,$3E,$17,$FD,$12,$D8,$CC
	.dc.b	$60,$8E,$09,$3F,$E3,$A9,$E0,$D9,$DB,$67,$6F,$F9,$17,$87,$F8,$6B
	.dc.b	$FE,$4A,$6B,$FE,$4D,$6A,$69,$38,$FF,$D3,$ED,$9F,$E8,$6D,$9F,$F2
	.dc.b	$6B,$5F,$F2,$A7,$D6,$3E,$2B,$FD,$4E,$DA,$FE,$96,$D9,$FF,$2A,$B5
	.dc.b	$E3,$3C,$09,$3C,$1B,$1C,$1A,$F0,$8F,$EA,$FF,$5A,$F5,$3F,$D6,$ED
	.dc.b	$8C,$C6,$08,$E0,$93,$FE,$68,$9E,$0D,$9D,$B6,$76,$FF,$9D,$FE,$80
	.dc.b	$40

;--------------------------------------------------------------------------------
;	FMOVE.P FPn,<mem>{#k}
;	FMOVE.P FPn,<mem>{Dk}
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovepregto_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+FPSP040+FPSP060,-(sp)
	peamsg	'FMOVE.P FPN,<EA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.P FPn,<ea>',13,10
;------------------------------------------------
;	d1-d3	actual result
;	d4	actual status
;	d5	0=failed,1=successful
;	d6	fpcr=(XRN..XRP)<<4
;	d7	k-factor=-18..18
;	a0	expected result handle,expected status,...
;	a1-a3	expected result
;	a4	expected status
;	a5	source handle,...
;	fp5	source
;------------------------------------------------
	lea.l	push_decompressed,a3
;decompress data
	move.l	a3,-(sp)
	pea.l	fmovepregto_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a3,a5			;source handle,...
@@:
	add.l	d0,(a3)+		;source handle
	tst.l	(a3)
	bpl	@b
	addq.l	#4,a3			;-1
	movea.l	a3,a0			;expected result handle,expected status,...
@@:
	add.l	d0,(a3)+		;expected result handle
	addq.l	#4,a3			;expected status
	tst.l	(a3)
	bpl	@b
;	addq.l	#4,a3			;-1
;
55:
	move.l	#0,d6		;fpcr=(XRN..XRP)<<4
66:
	moveq.l	#-18,d7			;k-factor=-18..18
77:
;FPn,<mem>{#k}
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	#-1,-(sp)		;fill=-1
	move.l	#-1,-(sp)
	move.l	#-1,-(sp)
					;k-factor
					;execute
	jsr	(fmovepregto_execute+18*8,pc,d7.l*8)	;EXECUTE
	movem.l	(sp)+,d1-d3		;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movem.l	([a0]),a1-a3		;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1-d4/a1-a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_packed
	lea.l	32(sp),sp
	move.l	d0,d5			;0=failed,1=successful
;
	move.l	d5,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.P FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',<mem>{#'
	putlong	d7			;k-factor
	putmsg	'} @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d5,-(sp)		;0=failed,1=successful
	movem.l	d1-d4/a1-a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_packed
	lea.l	36(sp),sp
@@:
;
;FPn,<mem>{Dk}
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	#-1,-(sp)		;fill=-1
	move.l	#-1,-(sp)
	move.l	#-1,-(sp)
					;source
	fmove.p	fp5,(sp){d7}		;EXECUTE
	movem.l	(sp)+,d1-d3		;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movem.l	([a0]),a1-a3		;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1-d4/a1-a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_packed
	lea.l	32(sp),sp
	move.l	d0,d5			;0=failed,1=successful
;
	move.l	d5,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.P FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',<mem>{Dr='
	putlong	d7			;k-factor
	putmsg	'} @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d5,-(sp)		;0=failed,1=successful
	movem.l	d1-d4/a1-a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_packed
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a0			;expected result handle,expected status,...
;
	addq.l	#1,d7			;k-factor++
	cmp.l	#18,d7			;k-factor=-18..18
	ble	77b			;signed
;
	add.w	#1<<4,d6		;rprm++
	cmp.w	#48,d6		;fpcr=(XRN..XRP)<<4
	bls	66b
;
	addq.l	#4,a5			;source handle,...
	tst.l	(a5)			;source handle,...
	bpl	55b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

fmovepregto_execute::
	fmove.p	fp5,(4,sp){#-18}
	rts
	fmove.p	fp5,(4,sp){#-17}
	rts
	fmove.p	fp5,(4,sp){#-16}
	rts
	fmove.p	fp5,(4,sp){#-15}
	rts
	fmove.p	fp5,(4,sp){#-14}
	rts
	fmove.p	fp5,(4,sp){#-13}
	rts
	fmove.p	fp5,(4,sp){#-12}
	rts
	fmove.p	fp5,(4,sp){#-11}
	rts
	fmove.p	fp5,(4,sp){#-10}
	rts
	fmove.p	fp5,(4,sp){#-9}
	rts
	fmove.p	fp5,(4,sp){#-8}
	rts
	fmove.p	fp5,(4,sp){#-7}
	rts
	fmove.p	fp5,(4,sp){#-6}
	rts
	fmove.p	fp5,(4,sp){#-5}
	rts
	fmove.p	fp5,(4,sp){#-4}
	rts
	fmove.p	fp5,(4,sp){#-3}
	rts
	fmove.p	fp5,(4,sp){#-2}
	rts
	fmove.p	fp5,(4,sp){#-1}
	rts
	fmove.p	fp5,(4,sp){#0}
	rts
	fmove.p	fp5,(4,sp){#1}
	rts
	fmove.p	fp5,(4,sp){#2}
	rts
	fmove.p	fp5,(4,sp){#3}
	rts
	fmove.p	fp5,(4,sp){#4}
	rts
	fmove.p	fp5,(4,sp){#5}
	rts
	fmove.p	fp5,(4,sp){#6}
	rts
	fmove.p	fp5,(4,sp){#7}
	rts
	fmove.p	fp5,(4,sp){#8}
	rts
	fmove.p	fp5,(4,sp){#9}
	rts
	fmove.p	fp5,(4,sp){#10}
	rts
	fmove.p	fp5,(4,sp){#11}
	rts
	fmove.p	fp5,(4,sp){#12}
	rts
	fmove.p	fp5,(4,sp){#13}
	rts
	fmove.p	fp5,(4,sp){#14}
	rts
	fmove.p	fp5,(4,sp){#15}
	rts
	fmove.p	fp5,(4,sp){#16}
	rts
	fmove.p	fp5,(4,sp){#17}
	rts
	fmove.p	fp5,(4,sp){#18}
	rts
	.cpu	68000

	.align	4
fmovepregto_data_compressed::
	.dc.b	$00,$16,$34,$CC,$10,$86,$36,$BD,$5A,$D5,$C2,$F9,$AC,$2F,$A6,$E3
	.dc.b	$E4,$72,$F5,$39,$FC,$07,$4F,$8C,$EB,$F3,$1D,$BE,$8F,$BF,$D7,$78
	.dc.b	$3E,$EB,$C3,$F8,$37,$8B,$F1,$2F,$1F,$E3,$DE,$4F,$CA,$BC,$BF,$9B
	.dc.b	$79,$BF,$42,$F3,$FE,$9D,$E8,$FD,$6B,$D3,$FB,$37,$AB,$F7,$2F,$5F
	.dc.b	$EF,$DE,$DE,$C7,$87,$7C,$CC,$7A,$F9,$F8,$2D,$7C,$6E,$FE,$63,$87
	.dc.b	$D1,$F1,$FA,$DE,$5F,$6F,$CF,$F0,$2E,$9F,$86,$F5,$FC,$63,$B7,$E4
	.dc.b	$7D,$FF,$2E,$F8,$BF,$81,$FC,$6F,$C3,$7E,$3F,$E2,$FF,$23,$F2,$1F
	.dc.b	$93,$F9,$5F,$CA,$FC,$D7,$E5,$FE,$81,$F3,$3F,$4B,$F4,$7E,$77,$E9
	.dc.b	$FD,$17,$D5,$FA,$7F,$AF,$F5,$AF,$67,$EC,$BE,$DF,$DC,$3D,$DF,$BD
	.dc.b	$FD,$BF,$69,$EE,$5F,$84,$C7,$C7,$67,$E6,$35,$F4,$5B,$FA,$CE,$1F
	.dc.b	$6D,$C7,$DD,$E5,$EE,$73,$EA,$E9,$D5,$D7,$AB,$B7,$57,$7E,$AF,$07
	.dc.b	$57,$87,$AB,$C5,$D5,$F8,$6F,$D6,$BF,$0F,$FB,$27,$E2,$3F,$6E,$FC
	.dc.b	$4F,$EF,$1F,$8D,$F8,$DF,$C9,$7C,$19,$7C,$2E,$7E,$17,$5F,$09,$BF
	.dc.b	$84,$E1,$F0,$9E,$FD,$3E,$06,$9F,$07,$4F,$85,$A7,$5F,$84,$FC,$F7
	.dc.b	$C3,$FE,$78,$EB,$F0,$FF,$A2,$5A,$6B,$4D,$E9,$C3,$4E,$38,$72,$F8
	.dc.b	$6E,$7F,$0D,$D3,$E1,$7A,$FC,$0F,$6F,$7B,$BF,$BD,$E0,$F7,$BF,$5D
	.dc.b	$E3,$FE,$4B,$83,$F6,$2B,$DC,$CF,$B9,$AF,$73,$7E,$E7,$0F,$73,$8F
	.dc.b	$9F,$FB,$7F,$89,$72,$5F,$80,$C6,$F9,$DF,$5B,$FE,$F7,$E2,$BC,$1C
	.dc.b	$07,$C0,$FF,$02,$5B,$6F,$6E,$1B,$71,$DB,$97,$BF,$CF,$DF,$E9,$EF
	.dc.b	$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BF,$F7,$7A,$FE,$F7,$5F,$DF,$F8
	.dc.b	$8F,$27,$C0,$79,$7E,$03,$CD,$F0,$1F,$83,$EF,$FC,$27,$C4,$7E,$17
	.dc.b	$E2,$3F,$0D,$F1,$1F,$87,$F8,$8F,$C4,$78,$7E,$DF,$84,$FE,$35,$F0
	.dc.b	$AF,$67,$F1,$73,$FC,$77,$E2,$DC,$FF,$8F,$F8,$9F,$C8,$7C,$4F,$E4
	.dc.b	$7C,$4D,$75,$6F,$AB,$87,$57,$1E,$AE,$5D,$5F,$C9,$FE,$31,$F0,$3F
	.dc.b	$CA,$7D,$57,$37,$F2,$BF,$8D,$78,$57,$C2,$C7,$85,$F9,$FF,$23,$F4
	.dc.b	$1E,$47,$E8,$7C,$8E,$1E,$2F,$1E,$6E,$5C,$7C,$F8,$FA,$71,$FF,$33
	.dc.b	$F8,$6F,$C5,$7B,$27,$ED,$B7,$C8,$FE,$79,$F1,$8F,$86,$FE,$7D,$ED
	.dc.b	$BC,$9F,$E8,$18,$70,$3C,$FE,$6B,$1E,$9C,$7A,$CB,$7C,$9C,$23,$8F
	.dc.b	$27,$8C,$F2,$70,$F4,$3A,$FC,$4F,$F4,$DF,$8F,$7A,$DE,$9E,$4F,$57
	.dc.b	$27,$AF,$93,$FA,$9F,$8A,$FD,$71,$97,$BF,$9B,$97,$4F,$3E,$9E,$9D
	.dc.b	$3F,$D6,$7E,$47,$F3,$A7,$E7,$39,$4F,$E7,$1F,$67,$33,$C1,$F9,$5E
	.dc.b	$0F,$1E,$6E,$5C,$DC,$F9,$BA,$6D,$D7,$8F,$B7,$1F,$7E,$3F,$07,$0F
	.dc.b	$87,$3F,$16,$7E,$3F,$3F,$C9,$E7,$F9,$7C,$FF,$37,$97,$E7,$F8,$EF
	.dc.b	$47,$C7,$7A,$7D,$7F,$57,$6F,$F7,$3F,$63,$F9,$CF,$37,$F7,$67,$EF
	.dc.b	$37,$E3,$75,$E2,$6F,$E3,$78,$79,$5C,$7E,$37,$97,$C6,$F3,$F8,$DE
	.dc.b	$9E,$A7,$5F,$8D,$ED,$F1,$BD,$FE,$37,$C1,$F1,$BE,$1F,$8D,$F1,$7C
	.dc.b	$6F,$8F,$E3,$7C,$9F,$1B,$E5,$F8,$DF,$37,$C6,$F9,$FE,$37,$FC,$15
	.dc.b	$F2,$5F,$E0,$A3,$F9,$6F,$31,$FE,$5A,$FF,$84,$17,$C6,$70,$F8,$9E
	.dc.b	$3E,$9F,$2F,$8C,$E7,$F1,$3D,$3E,$27,$AF,$B5,$DB,$E2,$7B,$FB,$FE
	.dc.b	$0F,$89,$F0,$FC,$4F,$8B,$E2,$7C,$7F,$13,$E4,$F8,$8F,$2F,$C4,$79
	.dc.b	$BE,$23,$CF,$EF,$7A,$3D,$EF,$4F,$BD,$FE,$23,$F9,$3F,$E4,$E7,$E2
	.dc.b	$CF,$C5,$73,$9F,$D3,$B7,$DC,$E1,$EE,$71,$F7,$39,$7B,$9C,$FD,$CE
	.dc.b	$9E,$E7,$5F,$73,$B7,$B9,$DF,$DC,$F0,$7B,$9E,$1F,$73,$C5,$EE,$78
	.dc.b	$FD,$CF,$27,$B9,$E5,$F7,$3C,$DE,$E7,$9F,$DC,$FF,$1D,$7C,$A7,$E2
	.dc.b	$39,$0F,$E2,$1B,$EE,$6B,$DC,$DF,$B9,$C3,$DC,$E3,$EE,$72,$F7,$39
	.dc.b	$FB,$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B
	.dc.b	$DC,$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$FE,$51,$F8
	.dc.b	$FF,$F2,$7F,$C5,$57,$1E,$EE,$7D,$DD,$7B,$BB,$F7,$78,$7B,$BC,$7D
	.dc.b	$DE,$5E,$EF,$3F,$77,$A7,$BB,$D7,$DD,$ED,$EE,$F7,$F7,$7C,$1E,$EF
	.dc.b	$87,$DD,$F1,$7B,$BE,$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$FE,$67
	.dc.b	$F9,$5F,$ED,$67,$F0,$F3,$F0,$A7,$E1,$07,$BB,$BF,$77,$87,$BB,$C7
	.dc.b	$DD,$E5,$EE,$F3,$F7,$7A,$7B,$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE
	.dc.b	$F8,$7D,$DF,$17,$BB,$E3,$F7,$7C,$9E,$EF,$97,$DD,$F3,$7B,$BF,$E7
	.dc.b	$DF,$96,$FC,$21,$F8,$31,$EE,$67,$DC,$D7,$B9,$BF,$73,$87,$B9,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$E9
	.dc.b	$3F,$97,$FF,$49,$9F,$C2,$8F,$E1,$3E,$F1,$FE,$12,$FF,$A6,$9F,$F1
	.dc.b	$1B,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6,$BC
	.dc.b	$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37,$B5
	.dc.b	$E7,$F6,$BD,$1E,$D7,$A7,$DA,$FF,$56,$78,$FF,$E3,$03,$FA,$A1,$FC
	.dc.b	$28,$FE,$12,$7E,$1B,$93,$FD,$66,$BD,$8E,$3E,$C7,$2F,$63,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$F3,$FB,$1F,$EC,$3F,$98,$FE
	.dc.b	$14,$7F,$09,$3F,$0F,$C9,$FE,$C8,$7F,$C1,$8D,$F5,$F8,$7A,$FC,$7D
	.dc.b	$7E,$5E,$BF,$3F,$5F,$A7,$AF,$D7,$D7,$ED,$EB,$F7,$F5,$FC,$1E,$BF
	.dc.b	$87,$D7,$F1,$7A,$FE,$3F,$5F,$C9,$EB,$F9,$7D,$7F,$37,$AF,$E7,$F5
	.dc.b	$FF,$DB,$5F,$33,$FE,$DA,$3F,$A8,$9F,$88,$3F,$0F,$C8,$7F,$0E,$DF
	.dc.b	$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE,$BF
	.dc.b	$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6,$F5
	.dc.b	$FC,$FE,$BF,$F8,$EF,$CD,$FF,$1E,$79,$BE,$AF,$5F,$FD,$E9,$F3,$5F
	.dc.b	$E8,$7F,$95,$78,$73,$A3,$FD,$7A,$BC,$3D,$FC,$37,$0F,$8D,$E3,$F1
	.dc.b	$1C,$BE,$03,$9F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73
	.dc.b	$C3,$EE,$78,$BD,$CF,$1F,$2F,$93,$97,$CB,$EE,$79,$BD,$CF,$3F,$B9
	.dc.b	$E8,$F7,$3D,$3E,$E7,$AB,$DC,$74,$4E,$5F,$EA,$C7,$F5,$5F,$C5,$7F
	.dc.b	$17,$D4,$FD,$73,$FC,$F5,$5F,$E3,$8D,$3F,$CC,$1A,$72,$8F,$F4,$BF
	.dc.b	$BF,$71,$FC,$9C,$7F,$A6,$BD,$EC,$7B,$D9,$F7,$B5,$EF,$6F,$DE,$E1
	.dc.b	$EF,$71,$F0,$B9,$78,$5C,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE
	.dc.b	$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$37,$CE,$F9,$F9,$0F,$F8
	.dc.b	$90,$FE,$36,$7E,$50,$FC,$99,$FE,$66,$7F,$CA,$FD,$FF,$E7,$FE,$FE
	.dc.b	$7E,$EF,$4F,$77,$AF,$BB,$FE,$83,$F0,$BF,$D0,$9E,$17,$83,$DD,$F9
	.dc.b	$FF,$22,$F9,$DF,$F2,$5B,$7E,$3B,$CB,$F0,$1E,$6F,$80,$F3,$FC,$07
	.dc.b	$0B,$FD,$FD,$F1,$F8,$5F,$5F,$C0,$FB,$3E,$07,$DB,$F0,$3E,$EF,$81
	.dc.b	$F7,$FC,$0F,$C0,$F8,$1F,$83,$F0,$3F,$0B,$E0,$7E,$1F,$C0,$FD,$9F
	.dc.b	$9E,$FF,$69,$9F,$E7,$73,$FC,$5A,$7F,$8B,$0F,$E4,$9D,$67,$F9,$23
	.dc.b	$F8,$A7,$FD,$0A,$DF,$73,$A7,$B9,$D7,$DC,$ED,$EE,$77,$F7,$3C,$1E
	.dc.b	$E7,$87,$DC,$F1,$7B,$9E,$3F,$67,$C9,$EC,$F9,$7C,$5F,$37,$8B,$E7
	.dc.b	$F0,$BD,$1E,$17,$E4,$7E,$7F,$FC,$5C,$7F,$8B,$4F,$E4,$BC,$27,$F2
	.dc.b	$63,$D8,$E1,$EC,$71,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7
	.dc.b	$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5
	.dc.b	$F6,$3E,$D7,$07,$DB,$E0,$FB,$9C,$1F,$77,$83,$EF,$70,$7D,$FE,$0F
	.dc.b	$C0,$70,$7E,$07,$83,$F0,$5C,$1F,$83,$E0,$F8,$1E,$C7,$C1,$F6,$3F
	.dc.b	$65,$F1,$7F,$9D,$3F,$CF,$E7,$F7,$63,$F9,$79,$FC,$B8,$FC,$E7,$2F
	.dc.b	$FA,$03,$D8,$E3,$EC,$72,$F6,$39,$FB,$1D,$3D,$7E,$BE,$BF,$6F,$5F
	.dc.b	$F2,$9D,$7F,$95,$EB,$FC,$B7,$5F,$E5,$FA,$FF,$31,$D7,$F9,$9E,$1F
	.dc.b	$CD,$70,$FE,$8B,$E5,$3C,$DE,$AF,$A3,$D5,$F4,$FA,$BE,$AF,$57,$D7
	.dc.b	$EA,$FB,$3D,$5F,$6F,$AB,$C7,$E3,$B9,$7C,$77,$C0,$F5,$FE,$0F,$AF
	.dc.b	$FC,$4F,$E8,$1F,$2B,$7E,$23,$1F,$01,$9F,$88,$D7,$C4,$6F,$E2,$38
	.dc.b	$7C,$47,$1F,$7B,$97,$BD,$CF,$DE,$F4,$79,$1E,$9F,$90,$ED,$EF,$F7
	.dc.b	$F7,$FC,$1E,$FF,$87,$DF,$F1,$7B,$FE,$3F,$7F,$C9,$EF,$FC,$1F,$8B
	.dc.b	$F8,$5E,$87,$C3,$F4,$3E,$27,$A1,$F1,$7D,$0F,$8D,$E8,$7C,$7F,$43
	.dc.b	$E4,$7A,$1E,$DF,$80,$F7,$7C,$07,$BF,$E0,$3E,$07,$C0,$7F,$33,$FA
	.dc.b	$1F,$F4,$81,$FE,$8F,$3F,$46,$7F,$34,$3F,$45,$EB,$7F,$42,$5E,$FF
	.dc.b	$1F,$7F,$97,$BF,$CF,$DF,$E9,$EF,$F5,$F7,$FE,$D7,$A9,$F6,$FD,$4F
	.dc.b	$B9,$EA,$7D,$DF,$53,$EF,$7A,$9F,$7F,$D4,$FC,$07,$A9,$F8,$1F,$4B
	.dc.b	$F0,$5E,$97,$F2,$4F,$95,$FA,$FC,$7E,$3F,$80,$F5,$7C,$07,$AF,$E0
	.dc.b	$3D,$9F,$01,$ED,$F8,$0F,$77,$C0,$7B,$FE,$03,$E0,$7C,$07,$C1,$F8
	.dc.b	$0F,$85,$F0,$1F,$D5,$7E,$8B,$FA,$01,$FE,$E2,$3F,$CB,$27,$F8,$48
	.dc.b	$FF,$08,$9F,$A4,$1E,$4F,$1F,$27,$97,$93,$CF,$C9,$E9,$E4,$F5,$F2
	.dc.b	$7B,$79,$3D,$FB,$BC,$1D,$DE,$1E,$EF,$17,$77,$8F,$BB,$C9,$DD,$E5
	.dc.b	$EE,$F3,$77,$79,$FB,$BF,$B6,$7C,$2F,$C4,$EE,$2F,$77,$3E,$EE,$BD
	.dc.b	$DD,$FB,$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$ED,$F6
	.dc.b	$F6,$FB,$FB,$7E,$0F,$6F,$C3,$ED,$F8,$BD,$BF,$1F,$B7,$F5,$F8,$7E
	.dc.b	$C7,$0F,$9B,$DA,$F3,$FB,$5E,$8F,$6B,$EE,$72,$7D,$DE,$4F,$5F,$B5
	.dc.b	$F7,$F9,$7F,$01,$CB,$FE,$09,$FA,$3F,$F7,$31,$FE,$E5,$3F,$AF,$FA
	.dc.b	$8E,$7F,$C3,$73,$EF,$A3,$87,$C2,$71,$F6,$F9,$7B,$7C,$FD,$BE,$9E
	.dc.b	$DF,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD
	.dc.b	$BF,$27,$B7,$E5,$F6,$FC,$DE,$DF,$9F,$DB,$F4,$7B,$7F,$E2,$1F,$A4
	.dc.b	$FF,$06,$0F,$84,$C7,$C2,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$01,$FA
	.dc.b	$1F,$43,$F4,$5E,$87,$F8,$D3,$E3,$9F,$AD,$B8,$7F,$C5,$DD,$0F,$23
	.dc.b	$C1,$F0,$5E,$1F,$82,$F1,$7C,$17,$8F,$E0,$BC,$9F,$05,$E5,$F8,$2F
	.dc.b	$37,$C1,$79,$FE,$0B,$D1,$F0,$5E,$9F,$82,$F5,$7C,$17,$AF,$E0,$BD
	.dc.b	$9F,$05,$ED,$F8,$2F,$77,$C1,$7F,$8F,$FE,$9B,$F2,$87,$E4,$FE,$0F
	.dc.b	$FC,$96,$BD,$CD,$7B,$5B,$F6,$B8,$7B,$3C,$7D,$9F,$77,$E3,$EF,$36
	.dc.b	$F9,$DD,$7D,$8E,$DE,$C7,$F0,$5F,$84,$FE,$0D,$F0,$9E,$1F,$17,$C5
	.dc.b	$E2,$F8,$E7,$C9,$3E,$5F,$84,$F3,$7C,$27,$9F,$E1,$3D,$1F,$09,$E9
	.dc.b	$F8,$3F,$57,$C1,$FA,$FC,$7F,$67,$C1,$FF,$96,$F9,$9E,$65,$F8,$EC
	.dc.b	$7C,$86,$7E,$0F,$5F,$07,$BF,$83,$E1,$F0,$7C,$7E,$0F,$97,$C1,$F3
	.dc.b	$F4,$BA,$7C,$1F,$5F,$83,$ED,$DB,$DF,$BB,$C1,$F0,$7E,$1F,$83,$F1
	.dc.b	$7A,$FE,$3F,$83,$F2,$7C,$1F,$97,$E0,$FC,$DF,$07,$FE,$71,$FA,$8F
	.dc.b	$F2,$E1,$F9,$1F,$37,$FC,$FA,$FF,$BD,$8F,$F5,$79,$FE,$AE,$3F,$C0
	.dc.b	$27,$F1,$E7,$7E,$E7,$3F,$73,$A7,$B9,$D7,$DC,$ED,$EE,$77,$F7,$3C
	.dc.b	$1E,$E7,$87,$DC,$F1,$7B,$9E,$3F,$73,$C9,$EE,$79,$7D,$CF,$37,$B9
	.dc.b	$E7,$F7,$3D,$1E,$E7,$FA,$27,$EA,$7F,$8F,$1F,$8B,$E2,$3F,$8B,$7F
	.dc.b	$D2,$EB,$C9,$E1,$ED,$F1,$F6,$F9,$7B,$7C,$FD,$AE,$9E,$D7,$5F,$6B
	.dc.b	$B7,$B5,$DF,$DA,$F0,$7B,$5E,$1F,$6B,$C5,$ED,$78,$FD,$AF,$27,$B5
	.dc.b	$E5,$F6,$BC,$DE,$D7,$9F,$DA,$F4,$7B,$5F,$EA,$1F,$AA,$FC,$3F,$37
	.dc.b	$FA,$B1,$7B,$59,$F6,$B5,$ED,$6F,$DA,$E1,$ED,$71,$F6,$B9,$7B,$5C
	.dc.b	$FD,$AE,$9E,$D7,$5F,$6B,$B7,$B5,$DF,$DA,$F0,$7B,$5E,$1F,$6B,$C5
	.dc.b	$ED,$78,$FD,$AF,$27,$B5,$E5,$F6,$BC,$DE,$D7,$9F,$DA,$FF,$5C,$7C
	.dc.b	$17,$FA,$8C,$FC,$47,$41,$7C,$26,$BE,$13,$71,$C3,$E1,$38,$FC,$27
	.dc.b	$2F,$7B,$9F,$BD,$D3,$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$7B,$C3
	.dc.b	$EF,$78,$BD,$EF,$1F,$BD,$E4,$F7,$BC,$BE,$F7,$9B,$DE,$F3,$FB,$DE
	.dc.b	$8F,$7B,$FD,$A1,$F5,$7F,$D9,$4F,$C4,$1F,$87,$E7,$3F,$87,$7F,$DB
	.dc.b	$EF,$FA,$68,$FF,$4C,$BA,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F,$6F
	.dc.b	$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB,$ED
	.dc.b	$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$DD,$1F,$59,$F8,$8E,$23,$F8,$87
	.dc.b	$FD,$EA,$FF,$A6,$8F,$F7,$59,$FE,$38,$77,$EB,$F2,$F5,$F9,$FA,$FD
	.dc.b	$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5,$FC,$3E,$BF,$8B,$D7,$F1
	.dc.b	$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F,$AF,$E8,$F5,$FF,$E0,$07
	.dc.b	$D6,$FF,$BA,$CF,$C4,$9F,$88,$E3,$05,$EB,$F0,$F5,$F8,$FA,$FC,$BD
	.dc.b	$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB,$F8,$3D,$7F,$0F,$AF
	.dc.b	$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F,$5F,$CF,$EB,$FA,$3D
	.dc.b	$7F,$4F,$AF,$E6,$E6,$FF,$62,$7C,$7D,$F6,$BF,$89,$1F,$8B,$3F,$15
	.dc.b	$1F,$EC,$9F,$80,$D7,$C0,$6F,$DA,$E1,$ED,$71,$F6,$B9,$7B,$5C,$FD
	.dc.b	$AE,$9E,$D7,$5F,$6B,$B7,$B5,$DF,$DA,$F0,$7B,$5E,$1F,$6B,$C5,$ED
	.dc.b	$78,$FD,$AF,$27,$B5,$E5,$F6,$BD,$1F,$5D,$F8,$93,$F1,$1C,$87,$FD
	.dc.b	$C2,$7F,$92,$1F,$A6,$FE,$24,$FF,$3F,$B8,$F6,$79,$FB,$3D,$3D,$9E
	.dc.b	$BE,$CF,$6F,$67,$BF,$B3,$E0,$F6,$7C,$3E,$CF,$8B,$D9,$F1,$FB,$3E
	.dc.b	$4F,$67,$CB,$EC,$F9,$BD,$9F,$3F,$B3,$E8,$F6,$7E,$87,$D7,$FD,$1E
	.dc.b	$23,$FC,$48,$FC,$53,$8F,$67,$7E,$CF,$0F,$67,$8F,$B3,$CB,$D9,$E7
	.dc.b	$EC,$F4,$F6,$7A,$FB,$3D,$BD,$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F
	.dc.b	$67,$C7,$EC,$F9,$3D,$9F,$2F,$B3,$E6,$F6,$7C,$FE,$CF,$A3,$D9,$FC
	.dc.b	$47,$D8,$7F,$9D,$4F,$EB,$07,$F1,$13,$F0,$E3,$D8,$FF,$4F,$7A,$1F
	.dc.b	$E9,$FF,$43,$8F,$B1,$CB,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8F
	.dc.b	$F5,$37,$AD,$FA,$CF,$69,$FA,$E7,$F6,$5F,$98,$7C,$83,$DA,$FF,$56
	.dc.b	$7B,$5F,$EA,$DF,$6B,$F3,$DC,$5F,$83,$F1,$DF,$64,$F0,$BF,$45,$F6
	.dc.b	$7F,$E3,$BC,$F9,$FB,$9F,$EB,$5F,$73,$3F,$1D,$AF,$8A,$DE,$7C,$3A
	.dc.b	$3F,$7F,$87,$FA,$57,$E3,$6F,$CA,$7E,$53,$37,$57,$6C,$FF,$61,$E6
	.dc.b	$8D,$FC,$57,$0F,$8A,$E3,$F0,$BA,$F1,$37,$E2,$70,$F1,$3F,$71,$F6
	.dc.b	$BF,$EA,$53,$D1,$AF,$E7,$37,$FF,$53,$7C,$1F,$8B,$E0,$F7,$CB,$C3
	.dc.b	$97,$8F,$2F,$2E,$AF,$E0,$5F,$6F,$FC,$5B,$AF,$F8,$CA,$F2,$B3,$E5
	.dc.b	$6B,$87,$7C,$3C,$38,$7D,$FF,$09,$F0,$3E,$13,$E0,$FC,$27,$F0,$DF
	.dc.b	$89,$FF,$5D,$0F,$33,$F8,$87,$DC,$3D,$DF,$E2,$5F,$72,$F5,$7E,$4F
	.dc.b	$C3,$FC,$AF,$87,$F9,$7F,$0F,$F3,$3E,$1F,$E6,$FC,$3F,$CE,$F8,$7E
	.dc.b	$3E,$DF,$2F,$6F,$9F,$B7,$F4,$BE,$1F,$E9,$FC,$3F,$D4,$F8,$7E,$FD
	.dc.b	$7E,$0E,$BF,$0F,$5F,$8B,$AF,$C7,$D7,$FC,$97,$EE,$BF,$DA,$47,$F8
	.dc.b	$D4,$FE,$DC,$7F,$24,$3F,$3D,$DD,$FD,$01,$FE,$DA,$7E,$01,$C7,$BB
	.dc.b	$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7C,$5F,$F3,$3F,$57,$FA,$F7,$DD,$BE
	.dc.b	$13,$F9,$B7,$DE,$7F,$7D,$3F,$99,$9E,$0C,$CF,$F7,$C3,$C0,$E3,$3E
	.dc.b	$39,$F2,$CF,$9E,$7D,$3C,$2E,$BE,$CF,$F4,$3F,$BD,$FE,$E8,$7F,$2C
	.dc.b	$1E,$AE,$7D,$5D,$7A,$BB,$F2,$38,$7A,$5C,$7D,$2E,$5E,$6F,$E8,$BC
	.dc.b	$AE,$9E,$6F,$5F,$37,$B7,$95,$DF,$CA,$FD,$4F,$D7,$FF,$53,$F5,$0F
	.dc.b	$F9,$ED,$BF,$27,$AF,$8D,$DF,$C6,$F0,$F6,$B8,$FC,$6F,$2F,$8D,$E7
	.dc.b	$F1,$BD,$3E,$0B,$D9,$EA,$7B,$7D,$4F,$77,$A9,$EF,$F5,$3E,$07,$A3
	.dc.b	$F0,$7D,$1F,$85,$E8,$FC,$3E,$7F,$EC,$DA,$FF,$37,$3F,$C8,$E7,$F3
	.dc.b	$63,$F5,$1E,$31,$FF,$25,$3F,$E0,$27,$F7,$CD,$79,$ED,$D3,$6E,$BB
	.dc.b	$76,$F2,$3B,$FC,$6F,$83,$E3,$7C,$3E,$BF,$8B,$D7,$F1,$FA,$FF,$22
	.dc.b	$FE,$4D,$FC,$AB,$F9,$77,$F3,$2F,$E6,$DF,$CE,$F5,$FE,$7F,$AF,$EC
	.dc.b	$7D,$AF,$BB,$D7,$F7,$FA,$FF,$03,$D7,$F8,$3E,$BF,$F8,$1F,$EF,$9F
	.dc.b	$17,$7E,$07,$1F,$03,$9F,$81,$D7,$C0,$EF,$DB,$E1,$F0,$9C,$7D,$AE
	.dc.b	$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DD,$ED,$F2,$3D,$FE,$3B,$C1,$EE,$F8
	.dc.b	$7E,$3B,$C5,$F1,$9E,$3F,$8C,$F2,$7C,$3F,$97,$E0,$7C,$DF,$13,$FE
	.dc.b	$1D,$FB,$FF,$F2,$71,$FE,$06,$3F,$37,$E4,$FF,$8D,$9F,$E8,$87,$F9
	.dc.b	$34,$FE,$55,$EB,$BB,$BF,$2B,$DD,$CF,$DE,$E9,$EF,$75,$F7,$BB,$7B
	.dc.b	$CF,$E0,$18,$5C,$FF,$BF,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42
	.dc.b	$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10
	.dc.b	$84,$21,$F1,$3D,$CF,$E1,$3E,$27,$F9,$77,$E9,$9E,$FF,$F0,$E8,$C3
	.dc.b	$6B,$FC,$EF,$3A,$55,$C9,$CC,$ED,$4E,$E7,$51,$89,$D4,$E7,$65,$F8
	.dc.b	$2B,$AC,$61,$EA,$62,$35,$7D,$2B,$94,$62,$B8,$57,$73,$81,$CA,$B8
	.dc.b	$5F,$88,$91,$A3,$84,$69,$F1,$F8,$DE,$0F,$B2,$FD,$EC,$73,$5E,$2E
	.dc.b	$4D,$78,$BB,$F1,$78,$78,$BC,$7C,$59,$87,$1F,$17,$A7,$8B,$D7,$C5
	.dc.b	$ED,$E2,$F7,$F1,$7C,$1E,$2F,$87,$C5,$F7,$72,$4C,$BC,$5E,$2F,$94
	.dc.b	$8E,$08,$7B,$38,$51,$A8,$C3,$79,$BE,$C4,$EA,$78,$C6,$27,$5C,$BF
	.dc.b	$E3,$CA,$D5,$72,$37,$19,$9D,$C7,$E2,$27,$75,$F8,$1E,$4F,$99,$E2
	.dc.b	$FC,$DF,$17,$E7,$78,$BF,$2B,$0F,$9D,$CB,$F2,$9F,$A1,$E2,$FD,$3F
	.dc.b	$17,$EA,$72,$FD,$5F,$17,$EB,$78,$BF,$2A,$FE,$B7,$27,$CA,$7F,$05
	.dc.b	$F6,$5F,$EE,$DD,$FD,$91,$E8,$3F,$04,$FE,$31,$FC,$47,$77,$E3,$23
	.dc.b	$11,$F8,$68,$C7,$1F,$E2,$BB,$BF,$10,$48,$CC,$F4,$9D,$46,$27,$5C
	.dc.b	$BF,$6E,$BA,$36,$FB,$92,$FF,$CD,$78,$7F,$AF,$23,$EB,$F8,$FF,$91
	.dc.b	$BF,$CA,$77,$FE,$9A,$BF,$9D,$C7,$E7,$1F,$F3,$F5,$7E,$D5,$7C,$27
	.dc.b	$E5,$27,$F0,$DD,$DC,$7E,$13,$F2,$B1,$F9,$FE,$EE,$9E,$37,$5F,$1B
	.dc.b	$B7,$8D,$F9,$5A,$F8,$FB,$FA,$77,$FC,$14,$7C,$DE,$07,$6F,$E0,$A7
	.dc.b	$33,$F8,$92,$6F,$F5,$1B,$1A,$8C,$11,$C1,$23,$F1,$DD,$DF,$90,$BE
	.dc.b	$ED,$BF,$04,$E8,$FB,$BD,$DF,$92,$7F,$79,$DB,$F9,$27,$F7,$DE,$27
	.dc.b	$EE,$DF,$E0,$1E,$27,$F0,$3E,$DF,$C9,$47,$EE,$CF,$E0,$9D,$7F,$92
	.dc.b	$7F,$83,$78,$7F,$BB,$7F,$84,$78,$7F,$57,$E4,$3C,$9A,$FC,$79,$FC
	.dc.b	$31,$36,$FC,$2B,$F3,$08,$7E,$18,$BC,$1F,$85,$78,$76,$7E,$34,$9C
	.dc.b	$5D,$6B,$9C,$F1,$25,$7E,$CD,$FE,$39,$D9,$F8,$B7,$F8,$47,$67,$EC
	.dc.b	$DF,$E1,$5E,$0F,$EC,$DF,$C3,$9F,$C8,$FA,$FF,$66,$FF,$25,$F1,$7F
	.dc.b	$8E,$3F,$08,$FD,$99,$F0,$8F,$E5,$5E,$2F,$97,$4F,$D2,$73,$FD,$6D
	.dc.b	$3E,$CD,$7E,$26,$30,$4D,$BE,$4C,$65,$AF,$E2,$0F,$BC,$59,$E6,$D9
	.dc.b	$E4,$FE,$08,$FE,$79,$E2,$FF,$13,$7F,$9F,$78,$BF,$07,$7F,$E4,$BA
	.dc.b	$FC,$17,$F9,$3F,$2F,$E0,$AB,$F9,$77,$89,$FC,$4D,$FE,$8D,$E2,$7F
	.dc.b	$13,$7F,$99,$78,$9F,$C4,$DF,$DA,$72,$FA,$30,$FB,$5A,$FD,$66,$EF
	.dc.b	$F6,$AB,$EF,$F6,$FF,$0A,$E6,$F9,$04,$8C,$9F,$82,$70,$FD,$E8,$FE
	.dc.b	$13,$F3,$5F,$D6,$F9,$FE,$F6,$1F,$D0,$39,$FF,$84,$7C,$47,$F0,$9F
	.dc.b	$88,$FE,$15,$E3,$7F,$05,$AF,$E8,$BD,$5F,$C1,$73,$FB,$C7,$F0,$53
	.dc.b	$F9,$6F,$89,$FC,$15,$FE,$5F,$E1,$FF,$05,$7C,$47,$E7,$7A,$7E,$8B
	.dc.b	$F4,$FA,$BE,$16,$FF,$BB,$E2,$FA,$F9,$7E,$FE,$BE,$DB,$63,$33,$BB
	.dc.b	$E7,$39,$6C,$73,$E2,$78,$7F,$BF,$C3,$FB,$27,$87,$FB,$F7,$FB,$37
	.dc.b	$87,$FB,$F7,$FB,$47,$87,$FB,$F7,$FC,$01,$E4,$7E,$79,$FE,$D9,$E2
	.dc.b	$7E,$FD,$FE,$DD,$E2,$7E,$FD,$FE,$E1,$E2,$7E,$FD,$FC,$B1,$F8,$F3
	.dc.b	$FA,$56,$5F,$E1,$92,$39,$DB,$F6,$8E,$8C,$6B,$FE,$C4,$9D,$C6,$49
	.dc.b	$3C,$09,$3B,$8E,$27,$F8,$70,$BB,$FF,$B3,$2B,$9D,$70,$9D,$1D,$2B
	.dc.b	$93,$4C,$C6,$EB,$94,$74,$DB,$E7,$B6,$FA,$DF,$2B,$E1,$5A,$AE,$97
	.dc.b	$C9,$B1,$99,$DD,$F2,$9F,$0D,$F3,$23,$82,$5F,$3A,$E3,$1E,$32,$1D
	.dc.b	$AB,$A4,$EE,$B9,$B6,$79,$DF,$8E,$72,$D8,$CD,$72,$8C,$57,$12,$38
	.dc.b	$24,$7F,$77,$F2,$A5,$7F,$85,$9A,$FF,$79,$6E,$1F,$DE,$5B,$87,$F8
	.dc.b	$11,$B3,$FE,$76,$F2,$51,$8C,$51,$89,$EC,$43,$B1,$6F,$B3,$6F,$F8
	.dc.b	$4B,$FC,$98,$FF,$15,$FC,$47,$F8,$3E,$3F,$C2,$1F,$11,$FE,$11,$F8
	.dc.b	$4F,$F0,$97,$C2,$7F,$84,$FE,$13,$FC,$29,$F0,$9F,$E1,$5F,$84,$FF
	.dc.b	$0B,$7C,$27,$F8,$5F,$E1,$3F,$C3,$1F,$09,$FE,$7E,$F8,$86,$D7,$E8
	.dc.b	$71,$1E,$BE,$1F,$54,$FD,$93,$ED,$F0,$FD,$E7,$04,$70,$4E,$0F,$B3
	.dc.b	$C9,$F5,$C9,$3C,$A3,$2D,$3E,$E9,$78,$FE,$DC,$F0,$8E,$D3,$C1,$A7
	.dc.b	$02,$CF,$07,$97,$2F,$E0,$9B,$5C,$EF,$CD,$1C,$5B,$5C,$63,$9D,$71
	.dc.b	$6D,$71,$BF,$15,$FF,$AC,$F1,$7C,$4F,$CD,$25,$FF,$A9,$63,$2D,$8C
	.dc.b	$BF,$EA,$48,$CB,$63,$33,$D5,$B5,$F6,$31,$5F,$0A,$D5,$F5,$21,$CC
	.dc.b	$AE,$9C,$1E,$1F,$89,$FC,$27,$C4,$FE,$17,$E2,$7F,$0D,$F1,$3F,$87
	.dc.b	$F8,$9F,$C4,$7C,$4F,$E2,$7E,$27,$F1,$5F,$13,$F8,$BF,$89,$FC,$67
	.dc.b	$C4,$FE,$37,$E2,$7F,$1D,$F1,$3F,$8F,$F8,$9F,$C8,$7C,$4F,$E4,$7E
	.dc.b	$1F,$F2,$5F,$0F,$EC,$E2,$FF,$70,$F2,$7A,$F9,$BE,$69,$23,$EB,$37
	.dc.b	$9B,$EB,$73,$7D,$8E,$9F,$AC,$DE,$9F,$AD,$7C,$09,$5C,$2F,$95,$F0
	.dc.b	$6D,$70,$9D,$57,$F1,$EC,$3C,$8F,$9B,$A7,$D1,$3F,$49,$B3,$A8,$C4
	.dc.b	$FD,$0B,$FB,$13,$92,$57,$02,$4E,$CE,$C5,$AE,$6F,$D7,$3F,$19,$D1
	.dc.b	$F5,$DF,$C7,$73,$FD,$77,$F2,$1C,$DF,$5D,$FC,$97,$2F,$D7,$7F,$29
	.dc.b	$C9,$F5,$DF,$CB,$71,$FD,$77,$F3,$1C,$5F,$5D,$FC,$D7,$0F,$99,$FC
	.dc.b	$C7,$B3,$FC,$2C,$8F,$E7,$DA,$E8,$C1,$0C,$13,$8B,$F5,$51,$C1,$B1
	.dc.b	$B6,$BB,$E0,$F9,$73,$B6,$C7,$EB,$F8,$7F,$8E,$AA,$4F,$EA,$7E,$C9
	.dc.b	$EC,$EA,$73,$4E,$15,$3D,$9D,$61,$C7,$D0,$B7,$BA,$7C,$06,$30,$E3
	.dc.b	$E8,$D9,$7A,$F7,$DF,$C7,$C2,$67,$E1,$B5,$E9,$EF,$E0,$38,$7A,$9C
	.dc.b	$7E,$17,$97,$B5,$CF,$E1,$7E,$07,$07,$EA,$39,$EF,$07,$F0,$46,$F8
	.dc.b	$3F,$EE,$32,$38,$24,$FC,$DE,$4F,$92,$FC,$D3,$A1,$92,$99,$AE,$CE
	.dc.b	$5F,$8F,$ED,$7B,$FA,$3E,$07,$C2,$FC,$1F,$85,$F8,$5E,$E7,$C3,$F7
	.dc.b	$3E,$27,$37,$C5,$F8,$5F,$91,$F6,$4F,$86,$BF,$0D,$8F,$86,$CF,$C3
	.dc.b	$7C,$4E,$0F,$EB,$3E,$17,$D8,$E1,$63,$FE,$F0,$9C,$F1,$7F,$27,$87
	.dc.b	$C4,$76,$F8,$6F,$83,$5F,$85,$AD,$7C,$47,$88,$91,$C8,$FC,$25,$A3
	.dc.b	$31,$C4,$CB,$8C,$7C,$93,$B9,$E5,$8F,$9A,$B5,$7E,$09,$CD,$6D,$E9
	.dc.b	$5B,$73,$5C,$1E,$E6,$08,$70,$9E,$57,$FD,$57,$C6,$FF,$00,$1F,$93
	.dc.b	$F1,$BF,$29,$E3,$7E,$57,$C6,$FC,$B7,$8D,$F4,$BE,$1F,$F2,$D9,$7A
	.dc.b	$3C,$77,$8D,$F9,$78,$C7,$8B,$F9,$77,$D3,$F0,$DF,$97,$78,$78,$7F
	.dc.b	$97,$7C,$BB,$7C,$4E,$0F,$A2,$FE,$F0,$9C,$5F,$92,$6C,$FE,$41,$BC
	.dc.b	$5F,$90,$8F,$C9,$78,$5F,$9C,$7F,$24,$D7,$3E,$17,$6B,$FE,$01,$E1
	.dc.b	$7E,$56,$7D,$9E,$0F,$E5,$5F,$77,$81,$F9,$57,$E0,$77,$FE,$55,$FA
	.dc.b	$BD,$DF,$94,$7E,$BF,$6F,$E5,$1F,$8D,$E2,$FE,$6B,$E2,$3F,$37,$DD
	.dc.b	$F8,$F9,$F2,$EB,$E3,$7D,$9A,$FB,$C8,$E0,$9B,$FB,$F8,$3E,$0F,$07
	.dc.b	$B5,$B7,$CE,$75,$7C,$9B,$7C,$AF,$FC,$CF,$CD,$FC,$82,$3E,$E7,$5F
	.dc.b	$E0,$9F,$BD,$D5,$F8,$27,$E0,$9F,$81,$F2,$7F,$8B,$BF,$83,$F2,$7F
	.dc.b	$8B,$BF,$85,$F2,$7F,$68,$FE,$63,$07,$93,$FC,$62,$3F,$15,$E4,$F9
	.dc.b	$F3,$F5,$6B,$E7,$6C,$F9,$F6,$D1,$1F,$92,$43,$EE,$9F,$95,$E8,$FB
	.dc.b	$8F,$87,$A3,$EE,$3E,$3E,$8F,$B8,$F9,$7A,$3E,$E3,$FA,$7E,$8F,$B8
	.dc.b	$FE,$A7,$A3,$EE,$3E,$BE,$8F,$B8,$FE,$DB,$9F,$EE,$3E,$23,$F6,$1D
	.dc.b	$3F,$B5,$E7,$FD,$C4,$60,$FD,$0E,$FF,$BF,$70,$4E,$8F,$D2,$3A,$3F
	.dc.b	$7A,$66,$38,$46,$A3,$06,$9C,$12,$36,$7F,$84,$3C,$1F,$E0,$33,$F8
	.dc.b	$03,$F7,$9E,$0F,$F0,$17,$F7,$DE,$0F,$F0,$17,$F0,$07,$F0,$1F,$03
	.dc.b	$F8,$03,$F7,$7C,$0F,$E0,$0F,$F0,$FF,$03,$F8,$03,$F8,$1F,$03,$F8
	.dc.b	$03,$E6,$D3,$E1,$E9,$F9,$88,$FB,$4D,$7E,$53,$49,$A7,$E1,$79,$BF
	.dc.b	$05,$3A,$8F,$DD,$BF,$C5,$7C,$9F,$E9,$EF,$F1,$7F,$27,$FA,$7B,$FC
	.dc.b	$67,$C9,$FE,$9E,$FF,$1B,$F2,$7F,$A7,$F3,$7E,$03,$2F,$E6,$FC,$DF
	.dc.b	$80,$7F,$A6,$77,$39,$FF,$03,$87,$F2,$DE,$7F,$C0,$D7,$9A,$3F,$30
	.dc.b	$7B,$35,$F4,$34,$C9,$E9,$2C,$71,$DF,$F6,$E4,$EC,$FE,$F1,$CF,$F9
	.dc.b	$C9,$F1,$F8,$7F,$B7,$BF,$D7,$78,$7F,$B7,$7F,$38,$7D,$83,$F6,$5C
	.dc.b	$FF,$62,$BD,$7C,$FF,$61,$F6,$EB,$F6,$1F,$7F,$37,$D8,$7C,$99,$FE
	.dc.b	$D3,$4F,$27,$37,$D2,$9C,$B7,$6F,$7E,$7E,$DA,$DB,$FB,$53,$EB,$1F
	.dc.b	$CC,$0F,$AC,$7F,$30,$3E,$B1,$FC,$40,$FE,$15,$C7,$F5,$9F,$EC,$BE
	.dc.b	$4F,$F8,$0D,$FE,$CF,$E4,$FF,$10,$C7,$FB,$7F,$8D,$FC,$C2,$7F,$B5
	.dc.b	$F9,$3F,$C9,$7C,$9F,$EC,$79,$FE,$92,$7F,$50,$DC,$FF,$2A,$D9,$DC
	.dc.b	$E8,$93,$A6,$C6,$A3,$04,$70,$4E,$FF,$E0,$8E,$F6,$72,$FE,$D2,$FF
	.dc.b	$94,$72,$FD,$DD,$3F,$23,$CB,$F7,$5F,$F0,$5F,$93,$FD,$D2,$BE,$E9
	.dc.b	$FC,$CC,$FE,$E6,$7F,$33,$3F,$B9,$9F,$DE,$7C,$6F,$E6,$4F,$F3,$5F
	.dc.b	$1B,$C7,$A7,$D9,$8F,$C9,$ED,$F6,$B0,$FB,$B1,$F9,$9D,$BE,$DB,$F7
	.dc.b	$75,$FC,$20,$E4,$FF,$2F,$78,$DF,$97,$C3,$FC,$DD,$E3,$7F,$32,$C7
	.dc.b	$F5,$7E,$37,$F3,$27,$F5,$BE,$37,$F3,$27,$FC,$4F,$E0,$7E,$CE,$FF
	.dc.b	$C7,$1E,$07,$EC,$DF,$F3,$C7,$67,$EC,$AB,$FC,$5B,$DF,$FB,$29,$F4
	.dc.b	$1F,$E3,$FF,$87,$FF,$25,$93,$7F,$F1,$6B,$5F,$F1,$33,$6B,$FC,$6E
	.dc.b	$D9,$E5,$3C,$08,$70,$2C,$F0,$6C,$70,$74,$48,$FF,$21,$C7,$F9,$7B
	.dc.b	$E1,$FF,$CB,$FF,$0F,$FE,$60,$F8,$7F,$F3,$0F,$C3,$FF,$98,$BE,$1F
	.dc.b	$FC,$C7,$F0,$FF,$E6,$4F,$87,$FF,$32,$FC,$3F,$F9,$9B,$E1,$FF,$D1
	.dc.b	$DF,$65,$FE,$9E,$DB,$FC,$8E,$BE,$23,$3D,$5A,$F7,$B7,$F0,$FC,$3D
	.dc.b	$FF,$76,$1F,$E5,$CE,$4E,$7F,$0F,$23,$FC,$AF,$92,$31,$1D,$9B,$CB
	.dc.b	$FE,$71,$9D,$4F,$09,$CF,$2A,$3F,$D5,$93,$A7,$91,$B2,$18,$25,$75
	.dc.b	$AE,$71,$C9,$BD,$FF,$E8,$0B,$F8,$7F,$01,$F3,$2B,$FC,$D7,$CB,$F3
	.dc.b	$1F,$5F,$C3,$FC,$C7,$DB,$F0,$3F,$31,$FF,$38,$F1,$7C,$C7,$E6,$F1
	.dc.b	$7C,$EF,$87,$F9,$FF,$05,$F2,$A7,$E2,$FC,$3F,$8B,$83,$D9,$C1,$F8
	.dc.b	$39,$FC,$06,$DF,$83,$8C,$11,$C1,$23,$E9,$1F,$8A,$29,$C8,$B1,$C5
	.dc.b	$B1,$C7,$7F,$F5,$0F,$0F,$DC,$6C,$F3,$56,$F8,$3F,$C9,$4F,$DE,$F8
	.dc.b	$7F,$C5,$FC,$27,$E3,$3E,$1F,$F2,$53,$F8,$2F,$85,$FC,$93,$F8,$4F
	.dc.b	$87,$FC,$8F,$C3,$7E,$4B,$CC,$FC,$9F,$A7,$F9,$0A,$FC,$4F,$6F,$FB
	.dc.b	$8F,$C8,$FF,$28,$47,$E5,$BB,$7C,$DE,$A3,$8B,$E5,$77,$4A,$FC,$1F
	.dc.b	$2F,$E0,$5B,$CB,$F8,$BE,$5F,$5D,$6E,$7F,$0F,$39,$E5,$FA,$6D,$9D
	.dc.b	$5F,$3A,$D9,$F8,$B3,$A9,$FE,$BE,$9F,$C5,$BF,$EB,$4F,$1B,$F6,$CF
	.dc.b	$FA,$DB,$C6,$FD,$8F,$77,$E2,$E3,$F6,$3D,$DF,$8B,$7F,$D6,$5D,$DF
	.dc.b	$8B,$7F,$57,$DD,$F8,$B7,$F5,$BD,$BF,$8B,$7F,$73,$DB,$E6,$CF,$EE
	.dc.b	$70,$FC,$59,$FB,$4D,$7F,$09,$C1,$F7,$9B,$19,$DB,$F0,$D3,$96,$C6
	.dc.b	$6B,$AF,$0F,$D2,$25,$F6,$8F,$C2,$BF,$EC,$BD,$FE,$0B,$FE,$CD,$DF
	.dc.b	$F0,$AF,$FB,$3F,$C5,$FC,$2B,$FE,$D1,$F1,$3F,$8B,$3F,$C2,$8F,$D5
	.dc.b	$9F,$C3,$3C,$4F,$E2,$8F,$F0,$EF,$13,$F8,$A3,$EE,$CB,$E1,$3F,$54
	.dc.b	$FA,$FA,$FE,$01,$FA,$ED,$73,$CB,$F8,$07,$86,$DF,$7C,$9A,$FE,$0E
	.dc.b	$38,$CE,$EB,$95,$6A,$78,$D6,$DF,$F7,$63,$F8,$47,$E2,$1F,$C3,$8F
	.dc.b	$C2,$1F,$C3,$8F,$F7,$DE,$7F,$CD,$B8,$FF,$09,$4A,$FC,$D6,$EE,$3F
	.dc.b	$C3,$61,$FC,$EB,$8F,$F0,$CF,$F3,$CE,$3F,$C3,$61,$FC,$BF,$8F,$E2
	.dc.b	$E9,$F7,$72,$FB,$77,$F2,$48,$7E,$13,$6F,$BA,$D3,$37,$F8,$48,$CB
	.dc.b	$FC,$EF,$C1,$FC,$0B,$FC,$F7,$C1,$FE,$0B,$5F,$D1,$3C,$0F,$C0,$CF
	.dc.b	$F4,$1F,$03,$F0,$2F,$F4,$3F,$03,$F8,$2D,$7F,$4A,$EF,$FE,$0A,$FE
	.dc.b	$04,$FE,$0A,$7F,$45,$9F,$88,$B4,$76,$FF,$08,$9F,$E9,$DD,$BE,$6D
	.dc.b	$BE,$A4,$7D,$7A,$FB,$31,$82,$5F,$DA,$24,$69,$BC,$DF,$81,$27,$6F
	.dc.b	$CC,$6C,$7E,$1A,$3F,$A7,$F2,$F9,$A3,$FB,$7F,$89,$FD,$3D,$63,$EA
	.dc.b	$C1,$4F,$13,$FA,$95,$7F,$76,$EF,$FE,$11,$7F,$DA,$7A,$7F,$0B,$A7
	.dc.b	$E1,$3A,$7F,$07,$3F,$DB,$7A,$7F,$06,$FF,$6F,$E9,$FC,$1D,$7F,$1F
	.dc.b	$E9,$FC,$1F,$13,$D4,$FE,$60,$4D,$3E,$75,$7F,$29,$9D,$46,$0F,$A0
	.dc.b	$E1,$F5,$C6,$E7,$94,$60,$DB,$8C,$3E,$05,$71,$6C,$66,$B9,$57,$7D
	.dc.b	$FE,$EC,$E6,$39,$99,$BE,$55,$C2,$3C,$15,$C1,$D7,$C1,$FF,$4D,$F5
	.dc.b	$7F,$A7,$7C,$1F,$F4,$FF,$3F,$FA,$87,$C3,$7F,$51,$F7,$FF,$A9,$7C
	.dc.b	$37,$F5,$3F,$80,$FE,$A9,$F0,$DF,$D5,$7D,$3F,$EA,$DE,$D7,$F5,$7F
	.dc.b	$81,$FE,$B1,$ED,$FF,$59,$F6,$BF,$AD,$7B,$53,$87,$CD,$CB,$E9,$E3
	.dc.b	$FC,$BC,$E6,$7C,$FC,$A8,$C7,$37,$C6,$6F,$37,$D3,$6D,$7D,$26,$DF
	.dc.b	$48,$F8,$91,$8B,$E9,$18,$21,$E0,$8D,$1F,$DC,$79,$BF,$B9,$7C,$27
	.dc.b	$F7,$3F,$6F,$FB,$A7,$C2,$7C,$6B,$FA,$1F,$07,$FE,$37,$7E,$7F,$D9
	.dc.b	$3D,$EC,$6D,$F9,$D8,$7C,$26,$37,$FE,$E9,$1F,$B8,$BF,$B5,$CB,$F3
	.dc.b	$5F,$B9,$C9,$B7,$F7,$D2,$F8,$4C,$71,$21,$F0,$9E,$AA,$F6,$E7,$F2
	.dc.b	$5F,$A2,$43,$E7,$F2,$7D,$3E,$4F,$C4,$C7,$E3,$0C,$9B,$9F,$CA,$B7
	.dc.b	$4F,$C7,$46,$23,$64,$70,$48,$D9,$23,$B4,$7C,$5F,$81,$FC,$A4,$7E
	.dc.b	$6F,$BF,$F2,$8F,$E7,$7B,$FF,$28,$FE,$7F,$B7,$F2,$8F,$E8,$7B,$3E
	.dc.b	$1E,$AF,$84,$FC,$AC,$FE,$97,$B7,$F2,$AF,$E9,$FB,$3F,$2C,$FE,$B6
	.dc.b	$BF,$DA,$9B,$7C,$2E,$1F,$D4,$4F,$E1,$4F,$97,$BF,$E4,$5A,$7E,$44
	.dc.b	$B1,$F9,$17,$81,$37,$FA,$4D,$9E,$4D,$39,$16,$79,$4F,$02,$4F,$07
	.dc.b	$F8,$4F,$5F,$E5,$AB,$F8,$17,$57,$E5,$9F,$E0,$9E,$07,$F9,$2F,$AB
	.dc.b	$E2,$F5,$7C,$6E,$AF,$8F,$A7,$E3,$EF,$F6,$3D,$5F,$BB,$7E,$27,$4F
	.dc.b	$CC,$E9,$FC,$64,$79,$6B,$F5,$6F,$DF,$DB,$EE,$6F,$F2,$37,$FC,$04
	.dc.b	$FD,$E9,$CD,$7C,$D2,$74,$FE,$44,$93,$F8,$D3,$81,$2F,$A5,$72,$9F
	.dc.b	$DB,$4F,$F2,$1F,$07,$F8,$DB,$F8,$73,$E0,$9F,$C9,$7C,$1F,$E3,$6F
	.dc.b	$F2,$7F,$07,$F8,$DB,$FB,$03,$F0,$C7,$EC,$0F,$E5,$9E,$0F,$E1,$9F
	.dc.b	$81,$E0,$F9,$9A,$7E,$03,$3F,$87,$A7,$D6,$8F,$B6,$DD,$3E,$DE,$9F
	.dc.b	$64,$93,$A6,$C6,$A3,$04,$70,$4A,$EF,$3C,$17,$2F,$E0,$EB,$F9,$5F
	.dc.b	$2F,$E0,$DF,$E5,$BC,$BF,$83,$7F,$41,$C9,$F8,$37,$F4,$57,$F8,$3C
	.dc.b	$BF,$AA,$F1,$FE,$0E,$7F,$8E,$F1,$FE,$0D,$F8,$1A,$FE,$12,$BE,$FE
	.dc.b	$BF,$5A,$7E,$C6,$1F,$79,$FB,$91,$92,$46,$63,$F1,$47,$0E,$2F,$9D
	.dc.b	$5D,$27,$4F,$6E,$67,$77,$F0,$8B,$FE,$93,$DD,$FC,$21,$FE,$97,$DD
	.dc.b	$FC,$21,$FE,$9B,$DD,$FC,$21,$FE,$9F,$DD,$FC,$23,$BB,$F8,$3C,$7F
	.dc.b	$52,$EE,$FE,$0E,$FF,$54,$EE,$FC,$16,$9F,$CE,$FE,$0F,$CB,$D1,$F4
	.dc.b	$78,$BE,$13,$6F,$E2,$90,$FC,$39,$2B,$F0,$9C,$1F,$6A,$3E,$F3,$78
	.dc.b	$BE,$F7,$47,$DB,$AF,$F0,$3F,$47,$DB,$7F,$C1,$3D,$1F,$6F,$0F,$EA
	.dc.b	$3D,$1F,$6D,$FE,$A7,$8F,$DB,$CB,$FC,$35,$D1,$F6,$F0,$FE,$A9,$D1
	.dc.b	$F6,$EF,$C9,$5E,$97,$FA,$DF,$C2,$7F,$2C,$23,$FC,$9D,$AE,$8C,$10
	.dc.b	$C1,$2B,$FB,$14,$70,$6C,$6D,$AE,$E7,$FB,$2C,$ED,$B1,$FC,$DF,$E1
	.dc.b	$FF,$9F,$FA,$FF,$D0,$27,$FA,$0F,$C2,$FF,$42,$F7,$BF,$A1,$F9,$D3
	.dc.b	$0F,$E7,$DE,$F7,$F4,$6F,$85,$FE,$8F,$EE,$49,$FE,$85,$F0,$BF,$D2
	.dc.b	$BE,$03,$FA,$5F,$A3,$FD,$BB,$EC,$BF,$BF,$57,$F8,$B1,$7A,$5E,$58
	.dc.b	$FE,$EB,$C3,$7A,$BF,$BA,$37,$C5,$34,$93,$F0,$78,$7D,$AF,$C1,$39
	.dc.b	$99,$29,$9A,$EA,$E5,$AE,$69,$C7,$E9,$F8,$1F,$8D,$3F,$E4,$BC,$9F
	.dc.b	$05,$ED,$F8,$2F,$8D,$3F,$E1,$AE,$3F,$81,$EE,$7C,$78,$FE,$CD,$C5
	.dc.b	$F1,$DF,$85,$ED,$FC,$77,$FB,$57,$0F,$C7,$7E,$47,$B7,$EE,$7E,$47
	.dc.b	$B5,$F2,$C8,$F8,$DA,$E8,$C1,$0C,$13,$AF,$FD,$11,$AA,$78,$C6,$1B
	.dc.b	$C4,$AF,$F5,$35,$6D,$E9,$5B,$6C,$EF,$89,$5B,$9F,$C0,$7B,$5F,$90
	.dc.b	$7E,$57,$C3,$7E,$41,$FF,$05,$75,$FE,$1F,$DB,$FC,$47,$B7,$FE,$14
	.dc.b	$ED,$FC,$57,$C1,$FE,$2F,$E0,$FF,$19,$F0,$DF,$E1,$8E,$FF,$C7,$7C
	.dc.b	$27,$E3,$FE,$13,$F2,$1F,$09,$F9,$1F,$07,$D7,$B7,$AB,$97,$E6,$72
	.dc.b	$FE,$A0,$9C,$BF,$8A,$6C,$FE,$21,$BC,$BF,$86,$6F,$81,$F9,$49,$FC
	.dc.b	$53,$5F,$C5,$78,$3D,$AF,$A1,$23,$FC,$A7,$E0,$FE,$35,$FF,$15,$47
	.dc.b	$E9,$FC,$0F,$C6,$47,$F8,$B7,$BF,$F1,$B8,$FD,$8E,$FF,$C6,$4F,$F9
	.dc.b	$A3,$BF,$F1,$93,$F7,$3B,$FF,$19,$3F,$E6,$CE,$FF,$C6,$3F,$99,$EF
	.dc.b	$F2,$3E,$DC,$FC,$9B,$FB,$5B,$BF,$AF,$1F,$7E,$FC,$2B,$57,$C4,$8E
	.dc.b	$09,$5C,$6F,$9D,$71,$8F,$D6,$D7,$F9,$27,$C5,$FD,$6B,$FE,$82,$F1
	.dc.b	$7F,$89,$3F,$04,$87,$FA,$1B,$AF,$E0,$E5,$FE,$85,$DB,$E0,$E5,$FE
	.dc.b	$84,$E8,$F8,$33,$FE,$89,$DB,$E0,$E1,$FC,$47,$97,$C5,$AC,$8F,$B2
	.dc.b	$D7,$EC,$98,$21,$82,$57,$FA,$57,$93,$ED,$47,$FA,$6B,$1F,$B4,$FF
	.dc.b	$07,$F0,$BE,$CC,$7F,$A6,$39,$3F,$82,$CF,$F9,$A7,$C5,$F4,$BF,$E9
	.dc.b	$6D,$7F,$94,$72,$7A,$AF,$FD,$3F,$AF,$AB,$3F,$DC,$F2,$7D,$3F,$0F
	.dc.b	$F5,$D1,$FB,$3C,$7F,$7D,$7F,$B3,$9C,$B7,$A3,$F4,$7C,$DF,$B6,$6C
	.dc.b	$EA,$BF,$4C,$7E,$F0,$A7,$42,$DF,$72,$5F,$58,$E1,$7D,$5F,$E5,$1E
	.dc.b	$1F,$F0,$17,$FA,$17,$87,$FC,$05,$FF,$60,$F8,$7F,$C0,$5F,$E5,$DE
	.dc.b	$17,$F0,$17,$F9,$87,$85,$FC,$05,$FE,$65,$E0,$FF,$01,$7F,$9A,$78
	.dc.b	$3F,$C0,$5F,$E6,$DE,$07,$F0,$17,$C7,$B7,$B7,$4F,$CA,$F4,$7D,$FD
	.dc.b	$FF,$3B,$5F,$79,$AF,$DE,$DF,$F3,$2E,$5A,$FF,$B6,$3A,$3F,$03,$3F
	.dc.b	$ED,$79,$FC,$0C,$FF,$5F,$E8,$FC,$0B,$FE,$D2,$E8,$FC,$0B,$FE,$E0
	.dc.b	$E8,$FC,$0E,$5F,$ED,$3E,$8F,$C0,$CF,$FA,$B7,$C5,$FE,$D6,$FF,$51
	.dc.b	$3F,$A2,$9E,$33,$F9,$AF,$C3,$7F,$6F,$23,$FD,$5D,$AF,$F7,$16,$E7
	.dc.b	$FD,$09,$B9,$FE,$1D,$B3,$D0,$E4,$43,$91,$67,$93,$67,$93,$FD,$9F
	.dc.b	$E1,$7F,$C0,$5F,$0D,$FE,$F0,$FB,$27,$C4,$DF,$73,$1E,$96,$7E,$03
	.dc.b	$5E,$E6,$FD,$CE,$1E,$A4,$D3,$FA,$A6,$CD,$BF,$A8,$68,$F8,$0E,$DE
	.dc.b	$EF,$7F,$80,$F0,$7B,$D3,$1E,$38,$70,$F5,$BD,$FC,$FF,$D6,$39,$7C
	.dc.b	$DE,$EC,$BD,$89,$68,$9D,$4B,$5F,$23,$B9,$D3,$CA,$B8,$D7,$0A,$DC
	.dc.b	$62,$75,$5C,$1E,$98,$FB,$AF,$83,$63,$71,$87,$C0,$DA,$E9,$1C,$8C
	.dc.b	$9C,$0E,$FF,$09,$F5,$7E,$07,$EB,$7B,$FF,$56,$7C,$7E,$F7,$D9,$F8
	.dc.b	$6F,$AB,$1E,$6E,$4F,$AC,$FA,$3D,$FF,$BD,$F0,$DF,$4A,$3D,$7C,$7F
	.dc.b	$81,$F8,$6F,$F0,$64,$78,$F5,$FF,$58,$FC,$0E,$38,$BE,$1F,$5B,$E1
	.dc.b	$BD,$5C,$9F,$0E,$7F,$1F,$87,$E4,$88,$E0,$9C,$9F,$92,$8C,$B6,$33
	.dc.b	$1F,$9D,$32,$57,$38,$FE,$0E,$3F,$38,$66,$79,$11,$C1,$FE,$26,$F0
	.dc.b	$3F,$3E,$FD,$AE,$FF,$64,$FF,$89,$72,$7C,$17,$E8,$2B,$EF,$F7,$E9
	.dc.b	$FC,$74,$BE,$0B,$19,$7F,$B5,$BE,$1B,$F4,$55,$F2,$75,$78,$1F,$AD
	.dc.b	$F0,$24,$FE,$3F,$1F,$F5,$FB,$F8,$CE,$FF,$D1,$3C,$7B,$FF,$43,$E5
	.dc.b	$BC,$0F,$9F,$E0,$4A,$FC,$31,$DD,$C3,$4C,$F2,$7E,$6B,$4F,$AD,$C5
	.dc.b	$F9,$59,$D1,$27,$55,$C9,$B1,$CA,$7F,$77,$E0,$7E,$4E,$3F,$47,$DF
	.dc.b	$F9,$37,$E0,$77,$FE,$4D,$FD,$3F,$85,$F9,$27,$E2,$77,$FF,$0F,$EF
	.dc.b	$FC,$9C,$7E,$B3,$C2,$FE,$08,$FE,$BB,$BB,$E6,$F1,$BB,$AF,$77,$97
	.dc.b	$6F,$BD,$D3,$F8,$3C,$3F,$0E,$DE,$0F,$C3,$F2,$7C,$99,$FC,$3C,$E6
	.dc.b	$B8,$CE,$67,$8B,$6B,$8F,$27,$D2,$24,$F2,$97,$77,$E3,$23,$F0,$1D
	.dc.b	$DF,$0E,$FE,$86,$9F,$CA,$67,$E1,$9F,$B5,$3F,$85,$F5,$7E,$22,$7E
	.dc.b	$B7,$57,$ED,$67,$F1,$5D,$FF,$84,$9F,$B5,$D7,$FC,$DF,$AF,$C7,$5F
	.dc.b	$67,$6F,$B9,$5F,$62,$30,$7F,$0C,$2E,$FF,$C4,$23,$44,$9E,$24,$9E
	.dc.b	$11,$96,$C6,$67,$93,$7A,$FF,$1B,$3A,$3F,$97,$75,$7E,$1D,$FD,$27
	.dc.b	$57,$C5,$AF,$E8,$BD,$5F,$87,$9F,$D4,$75,$7E,$1F,$2F,$CC,$75,$7E
	.dc.b	$1E,$7F,$98,$F8,$1F,$16,$7F,$39,$BF,$F1,$99,$FD,$67,$47,$E1,$9F
	.dc.b	$CC,$1E,$3D,$BF,$00,$FF,$04,$69,$39,$BF,$01,$A7,$E1,$39,$FE,$6E
	.dc.b	$3F,$82,$BF,$C0,$46,$27,$51,$83,$FB,$61,$FE,$D3,$F1,$3F,$B6,$73
	.dc.b	$FE,$6B,$C1,$FE,$19,$3F,$C1,$7C,$1F,$CE,$F8,$3F,$9E,$F0,$7F,$05
	.dc.b	$8F,$F6,$8E,$AF,$E0,$F7,$FC,$2F,$AB,$F4,$7F,$0D,$F8,$3A,$FE,$DB
	.dc.b	$D7,$F8,$3A,$F1,$1F,$7B,$AB,$E9,$F1,$7D,$68,$F8,$51,$82,$78,$5F
	.dc.b	$AB,$8C,$B6,$78,$12,$77,$3F,$32,$31,$D7,$F7,$78,$FF,$87,$75,$FD
	.dc.b	$DA,$FE,$DD,$D7,$F7,$67,$F8,$97,$5F,$DD,$D3,$FC,$0B,$D7,$F7,$6F
	.dc.b	$F9,$47,$5F,$DD,$9F,$F0,$4F,$5F,$DD,$9F,$EA,$9D,$BF,$C0,$DF,$E5
	.dc.b	$5D,$FE,$78,$7C,$37,$F0,$82,$4F,$F2,$F6,$BC,$9B,$9F,$F5,$C6,$E1
	.dc.b	$FD,$0D,$B3,$FD,$01,$BB,$AF,$BC,$F5,$21,$D4,$B7,$D5,$FE,$81,$F0
	.dc.b	$DF,$C7,$3C,$EF,$E3,$BE,$77,$F1,$EF,$7F,$F8,$FE,$7F,$C8,$3D,$FF
	.dc.b	$E4,$3E,$B7,$F2,$2F,$43,$F9,$1F,$BF,$33,$FE,$95,$E8,$FF,$26,$F4
	.dc.b	$7F,$93,$FB,$FF,$CA,$3D,$2F,$E5,$3E,$EF,$F5,$6F,$B0,$FE,$FB,$87
	.dc.b	$94,$9C,$5E,$56,$ED,$F0,$5F,$4B,$5C,$9F,$0C,$C9,$4C,$F2,$7C,$5E
	.dc.b	$5F,$8F,$CB,$F1,$A7,$2D,$BF,$01,$2F,$B3,$67,$FC,$79,$83,$E0,$3D
	.dc.b	$5F,$0F,$F2,$27,$FB,$57,$27,$B7,$E1,$3D,$DE,$C7,$BF,$E1,$3E,$07
	.dc.b	$C1,$FC,$1F,$83,$F8,$5F,$07,$FD,$33,$9B,$E4,$65,$FE,$41,$D5,$F0
	.dc.b	$7F,$26,$31,$F0,$7F,$27,$E1,$FD,$7C,$9F,$2F,$E1,$25,$71,$6B,$FE
	.dc.b	$29,$C1,$18,$E3,$FF,$25,$C6,$27,$81,$0E,$13,$BE,$44,$7C,$28,$C3
	.dc.b	$4E,$15,$D2,$33,$CC,$BE,$D5,$D0,$FC,$89,$FE,$11,$F0,$3F,$09,$F0
	.dc.b	$9F,$85,$F8,$7F,$C8,$CF,$F8,$57,$C1,$FC,$8B,$FE,$17,$F0,$3F,$20
	.dc.b	$FE,$CB,$C0,$FC,$67,$C2,$FE,$37,$E1,$7F,$23,$3F,$2F,$E1,$FF,$22
	.dc.b	$FC,$D3,$CB,$A7,$93,$7F,$BD,$B7,$E1,$E7,$F1,$6D,$7F,$0F,$C3,$F4
	.dc.b	$1C,$F7,$FE,$27,$8B,$F0,$E4,$8F,$C9,$5F,$49,$D4,$62,$75,$7D,$6B
	.dc.b	$C3,$87,$E3,$6B,$F0,$5D,$DF,$8D,$7F,$87,$F7,$7E,$35,$FC,$37,$6F
	.dc.b	$E3,$5F,$C4,$76,$7E,$35,$FC,$57,$5F,$E3,$5F,$F1,$F7,$87,$FB,$1E
	.dc.b	$BF,$C6,$47,$E1,$7A,$FF,$18,$FB,$38,$3D,$CF,$C0,$30,$43,$04,$D7
	.dc.b	$DD,$1C,$1B,$1F,$1B,$C6,$FE,$1C,$FF,$93,$FC,$6F,$E1,$CF,$F9,$47
	.dc.b	$87,$F8,$73,$FE,$53,$F1,$BF,$87,$3F,$E5,$5E,$7F,$BF,$87,$FA,$03
	.dc.b	$BF,$F4,$F3,$FC,$CB,$C6,$F5,$CF,$FA,$13,$C5,$FE,$1D,$3E,$3C,$7E
	.dc.b	$7F,$3F,$D2,$E7,$FA,$11,$8D,$B5,$5F,$54,$93,$A6,$D7,$12,$4F,$09
	.dc.b	$D1,$23,$B7,$3F,$DB,$AF,$D9,$73,$FD,$BC,$BF,$D3,$3C,$FF,$6E,$7F
	.dc.b	$6D,$CF,$F6,$DF,$DC,$73,$FD,$B7,$F7,$5C,$FF,$6D,$FE,$A3,$CF,$F6
	.dc.b	$DF,$F3,$D7,$8D,$FC,$B7,$9F,$ED,$C7,$D5,$E0,$FD,$CE,$FF,$B3,$F0
	.dc.b	$7F,$42,$49,$FD,$CC,$EB,$7F,$D2,$36,$75,$3F,$B9,$26,$DF,$BF,$78
	.dc.b	$9F,$C0,$CC,$CF,$17,$37,$E1,$7F,$80,$9F,$DA,$BC,$1F,$E0,$2F,$F2
	.dc.b	$1F,$03,$F8,$0B,$FD,$BB,$C0,$FE,$02,$FF,$70,$F0,$3F,$07,$87,$FA
	.dc.b	$9F,$A7,$F8,$0C,$FF,$29,$EF,$FE,$02,$FF,$21,$EF,$FE,$02,$FF,$23
	.dc.b	$3D,$7A,$7E,$59,$F8,$C7,$DE,$E8,$FB,$9D,$1F,$7A,$3F,$41,$3B,$26
	.dc.b	$3F,$89,$8F,$E0,$FC,$1F,$7E,$7F,$96,$F4,$7D,$FA,$FF,$5E,$74,$7D
	.dc.b	$F9,$FE,$63,$D1,$F7,$DF,$E6,$7D,$1F,$7D,$FE,$6B,$D1,$F7,$DF,$F6
	.dc.b	$FF,$47,$DF,$CF,$FC,$23,$D1,$F7,$DF,$E9,$1E,$17,$E4,$F5,$7C,$37
	.dc.b	$F6,$F2,$4F,$F4,$D6,$BF,$D1,$5B,$87,$F4,$46,$EB,$F8,$36,$CF,$F6
	.dc.b	$26,$F4,$7F,$3D,$6D,$F8,$27,$B1,$0E,$C5,$BE,$CD,$BE,$CD,$3B,$16
	.dc.b	$FB,$36,$BB,$3D,$09,$5D,$1B,$5D,$1A,$74,$2D,$74,$6D,$74,$AE,$44
	.dc.b	$AE,$4D,$9E,$4D,$AE,$4D,$AE,$55,$C0,$95,$C1,$B5,$C1,$B3,$C2,$74
	.dc.b	$49,$D3,$63,$51,$82,$38,$25,$FD,$42,$4F,$C9,$6B,$FE,$05,$F7,$3F
	.dc.b	$C6,$84,$7F,$C2,$4D,$74,$60,$86,$09,$87,$F8,$32,$38,$36,$36,$D7
	.dc.b	$78,$7F,$0F,$9D,$B6,$3F,$C4,$5F,$01,$FE,$47,$F7,$7F,$C2,$30,$B5
	.dc.b	$7F,$1C,$5E,$EE,$9C,$CB,$3E,$1F,$0D,$9B,$E1,$CB,$6B,$56,$F7,$71
	.dc.b	$87,$7F,$4A,$CB,$D2,$BE,$DE,$3D,$BC,$FC,$1E,$BD,$4D,$FB,$7C,$3E
	.dc.b	$0B,$8F,$B9,$CB,$DB,$F7,$EF,$FE,$4C,$E6,$FA,$7A,$FE,$27,$9E,$F1
	.dc.b	$FE,$11,$B5,$F8,$16,$F0,$FE,$01,$BD,$1F,$80,$9F,$BC,$43,$F3,$B1
	.dc.b	$A2,$39,$BF,$03,$6F,$C0,$D3,$C0,$5B,$F0,$36,$FC,$0F,$63,$17,$D9
	.dc.b	$C1,$2F,$BB,$6F,$BB,$4E,$E7,$DF,$7E,$F7,$C3,$7D,$22,$47,$F9,$52
	.dc.b	$7E,$FE,$1F,$85,$9C,$CF,$13,$EE,$B8,$9D,$87,$15,$DA,$7A,$35,$E5
	.dc.b	$5D,$5C,$9D,$0A,$E5,$DD,$75,$72,$D7,$35,$E0,$79,$1C,$3D,$CF,$C7
	.dc.b	$FB,$9F,$90,$F8,$6F,$C8,$FC,$37,$E4,$BD,$EF,$C9,$FC,$37,$E5,$BE
	.dc.b	$BF,$FD,$A5,$8F,$99,$7C,$06,$7E,$17,$5F,$0B,$BF,$85,$E1,$F0,$BC
	.dc.b	$7C,$EE,$5E,$EF,$3F,$5B,$A7,$BB,$E5,$E5,$F6,$72,$FD,$82,$73,$7D
	.dc.b	$86,$EF,$F6,$1B,$3F,$86,$8F,$B8,$D8,$CF,$47,$E2,$1B,$D1,$F8,$76
	.dc.b	$D7,$E2,$88,$7C,$83,$E5,$F4,$79,$63,$E6,$1F,$71,$C1,$23,$64,$39
	.dc.b	$12,$38,$39,$6B,$9C,$3F,$05,$18,$3F,$13,$D1,$F5,$A7,$EE,$F3,$FD
	.dc.b	$67,$EF,$F3,$7D,$67,$F0,$BC,$BF,$5A,$FF,$2E,$FD,$68,$CF,$27,$D6
	.dc.b	$77,$C7,$F5,$9E,$3C,$5F,$59,$E7,$C3,$F5,$9F,$46,$9E,$63,$FC,$7F
	.dc.b	$83,$D0,$FC,$31,$27,$C6,$D7,$F1,$0D,$CB,$CA,$DA,$E6,$D9,$FE,$59
	.dc.b	$5D,$5E,$64,$39,$96,$B9,$B6,$B9,$BF,$11,$FC,$D7,$AD,$F9,$BF,$77
	.dc.b	$F3,$9E,$BF,$E7,$7D,$7F,$CF,$7B,$BF,$9F,$F4,$BF,$41,$EE,$FE,$87
	.dc.b	$D3,$FD,$17,$BB,$FC,$E3,$E0,$FF,$A3,$EB,$FB,$87,$FC,$69,$AF,$F9
	.dc.b	$89,$7B,$3B,$78,$7B,$3C,$7E,$0F,$DD,$DC,$E3,$F7,$63,$EC,$27,$1F
	.dc.b	$B1,$B9,$7B,$1A,$7C,$6E,$4E,$75,$CA,$B8,$12,$B8,$36,$B8,$1F,$1C
	.dc.b	$A7,$4B,$F2,$5F,$4A,$F8,$1E,$D7,$C1,$F8,$3F,$93,$7F,$90,$E4,$F8
	.dc.b	$9E,$DF,$C9,$8F,$C9,$F1,$7C,$7F,$81,$F9,$1F,$03,$F2,$67,$E5,$7C
	.dc.b	$27,$CB,$E3,$91,$FC,$7B,$8D,$C7,$F2,$23,$CD,$C1,$E4,$75,$EF,$7D
	.dc.b	$32,$38,$21,$F5,$5C,$B4,$CF,$17,$F3,$CB,$4E,$A3,$0D,$E1,$FB,$8D
	.dc.b	$AF,$F0,$05,$70,$79,$CE,$9A,$6B,$E0,$BF,$22,$FE,$DB,$AF,$F2,$4F
	.dc.b	$EE,$3A,$FF,$22,$FC,$1F,$84,$FC,$77,$C1,$FE,$46,$3F,$7B,$D5,$F9
	.dc.b	$1F,$84,$FC,$97,$C2,$7E,$4F,$E1,$3F,$1D,$5F,$27,$CA,$FC,$73,$E2
	.dc.b	$E1,$6F,$E7,$8F,$BD,$DB,$F8,$7E,$0F,$C2,$92,$3F,$07,$C5,$F8,$7B
	.dc.b	$FA,$11,$C3,$B7,$F1,$E4,$7E,$AD,$F3,$78,$DF,$3A,$E3,$7C,$EB,$FC
	.dc.b	$53,$5F,$A9,$ED,$FC,$63,$F8,$B2,$1F,$AD,$EB,$FC,$63,$FD,$47,$AF
	.dc.b	$F1,$8F,$E1,$FA,$BF,$18,$FE,$27,$AB,$F1,$6F,$ED,$7A,$BF,$14,$FE
	.dc.b	$DF,$AB,$D0,$FA,$CF,$4E,$D8,$D3,$CC,$DD,$BD,$5A,$FB,$76,$F5,$C6
	.dc.b	$27,$57,$F0,$09,$3F,$BD,$9F,$C0,$C7,$F6,$9E,$8F,$D4,$BF,$DA,$FA
	.dc.b	$3F,$02,$FF,$6D,$F0,$3F,$88,$BF,$CF,$3C,$0F,$E2,$2F,$E9,$4F,$D4
	.dc.b	$9F,$A7,$D7,$F0,$2F,$B4,$FD,$57,$81,$E4,$CB,$CF,$1E,$4D,$3D,$44
	.dc.b	$D3,$4D,$8F,$86,$43,$E3,$13,$5F,$8A,$FF,$2D,$3F,$71,$C7,$F6,$9F
	.dc.b	$F0,$37,$7F,$BA,$7F,$5D,$C9,$F6,$9F,$EA,$BC,$9F,$6A,$BF,$95,$72
	.dc.b	$7D,$A7,$F9,$67,$27,$DA,$AF,$DB,$6D,$F6,$9F,$DC,$72,$7B,$79,$7F
	.dc.b	$39,$1F,$B8,$BF,$D0,$46,$39,$7F,$69,$1F,$C0,$49,$3B,$6C,$6A,$30
	.dc.b	$47,$04,$8F,$E0,$3D,$DF,$C0,$F8,$7F,$5A,$DB,$F0,$4F,$F0,$13,$FB
	.dc.b	$AF,$85,$FC,$01,$FC,$11,$FC,$04,$FE,$F3,$E1,$7F,$4D,$EF,$FE,$03
	.dc.b	$1F,$81,$3F,$80,$9F,$C9,$7B,$7F,$80,$BF,$C9,$FB,$3F,$03,$A7,$F1
	.dc.b	$BE,$CF,$E0,$33,$EC,$3C,$E7,$DB,$8F,$B2,$FD,$BD,$7F,$34,$E5,$AE
	.dc.b	$4F,$C4,$73,$FE,$18,$FF,$22,$47,$F4,$48,$FE,$71,$E1,$7D,$F8,$FF
	.dc.b	$02,$F3,$7F,$44,$9F,$F0,$AF,$5F,$E7,$1F,$ED,$1E,$17,$F4,$47,$F3
	.dc.b	$87,$EE,$4F,$CE,$10,$FC,$E1,$31,$F4,$9F,$CE,$7D,$4F,$EE,$84,$8F
	.dc.b	$E9,$4D,$7F,$BB,$36,$38,$46,$88,$68,$AE,$9C,$11,$C1,$30,$FF,$00
	.dc.b	$4F,$57,$FC,$05,$F0,$9F,$E0,$7F,$84,$FF,$5F,$FD,$77,$FB,$2B,$0F
	.dc.b	$F3,$DA,$F8,$4C,$FC,$26,$BE,$13,$7F,$09,$2F,$7F,$09,$CB,$E1,$24
	.dc.b	$72,$F8,$4E,$BF,$09,$DB,$E1,$3B,$FC,$24,$AF,$5F,$07,$99,$A7,$F5
	.dc.b	$AE,$1F,$43,$4C,$F8,$1F,$D2,$23,$2D,$8C,$F7,$A7,$2D,$8C,$C6,$08
	.dc.b	$F4,$E1,$7C,$27,$C4,$F8,$4F,$8B,$F0,$BF,$1B,$E1,$7E,$1D,$FC,$6E
	.dc.b	$4F,$84,$FF,$A9,$B0,$71,$FC,$CE,$3F,$87,$39,$E3,$F9,$FC,$7F,$12
	.dc.b	$3E,$7F,$1F,$D2,$F7,$3E,$9F,$B9,$E4,$9E,$9F,$0B,$F5,$88,$FD,$66
	.dc.b	$BA,$30,$43,$04,$EB,$FF,$63,$E8,$9E,$31,$86,$F2,$A9,$CA,$BE,$11
	.dc.b	$D6,$F8,$36,$B8,$56,$EF,$FC,$A9,$D7,$F8,$FA,$FC,$6F,$BB,$F9,$0E
	.dc.b	$BF,$C8,$FC,$2F,$E3,$67,$F2,$3E,$F7,$E5,$3E,$17,$C5,$1F,$ED,$2F
	.dc.b	$AE,$FF,$45,$69,$D9,$FF,$45,$46,$7E,$03,$7B,$7F,$B9,$25,$F0,$B8
	.dc.b	$BE,$1F,$03,$F9,$3C,$3F,$3F,$E0,$7E,$4D,$F8,$D3,$EA,$E2,$F7,$C7
	.dc.b	$EB,$63,$F2,$CD,$3F,$28,$5E,$2F,$C9,$37,$C0,$FC,$83,$7C,$0F,$B9
	.dc.b	$A7,$ED,$63,$8B,$6B,$8B,$6B,$8B,$9C,$3F,$2B,$5F,$B6,$EF,$FC,$AB
	.dc.b	$FB,$8E,$EF,$CA,$BF,$93,$EE,$FC,$AD,$FE,$03,$BB,$F2,$93,$FB,$EE
	.dc.b	$DF,$CA,$CF,$E1,$3B,$7F,$29,$3F,$9B,$ED,$FC,$A3,$FC,$13,$B3,$D3
	.dc.b	$C1,$F0,$AF,$CD,$C1,$8A,$F5,$CE,$48,$E0,$9C,$1E,$E8,$CB,$63,$3C
	.dc.b	$1F,$02,$72,$D8,$CD,$7E,$31,$FF,$3A,$78,$7F,$09,$FD,$97,$57,$C2
	.dc.b	$C7,$F3,$FE,$1F,$C2,$C7,$EE,$6F,$F0,$AF,$F6,$87,$C3,$3F,$8C,$1F
	.dc.b	$B4,$3F,$91,$F8,$7F,$85,$BF,$D4,$F2,$FB,$F5,$F4,$13,$5F,$33,$75
	.dc.b	$D4,$60,$8E,$09,$87,$F1,$43,$F8,$97,$87,$FC,$45,$FE,$67,$E1,$FF
	.dc.b	$11,$7F,$88,$10,$FE,$21,$9F,$B5,$FD,$BF,$3B,$C3,$FB,$BB,$7E,$6F
	.dc.b	$97,$F8,$8D,$7B,$4F,$BC,$7F,$40,$F0,$FF,$88,$BE,$33,$F4,$67,$4F
	.dc.b	$0F,$F8,$2B,$4F,$E0,$05,$37,$B7,$F0,$17,$2D,$73,$8F,$EB,$38,$3F
	.dc.b	$84,$C6,$2B,$AD,$6E,$72,$4C,$7F,$AC,$D7,$E0,$0F,$E1,$67,$F5,$8E
	.dc.b	$3F,$E1,$6F,$F5,$AE,$3F,$EB,$BE,$07,$F0,$B8,$FE,$39,$7F,$C1,$7E
	.dc.b	$27,$F8,$24,$7F,$1E,$EF,$FE,$17,$3F,$C2,$7C,$0C,$E6,$9F,$E1,$75
	.dc.b	$CB,$3F,$E1,$6F,$E1,$8F,$E1,$64,$3D,$87,$CB,$BF,$D3,$B6,$3F,$3B
	.dc.b	$C9,$F8,$28,$FE,$0C,$64,$FC,$F9,$6F,$F9,$97,$27,$E1,$63,$FA,$87
	.dc.b	$27,$E1,$5F,$E0,$07,$E1,$70,$FE,$00,$FF,$7C,$E4,$FD,$04,$7F,$42
	.dc.b	$E3,$FC,$2C,$FE,$23,$AF,$F8,$0C,$FF,$46,$EB,$FE,$03,$3F,$D7,$7A
	.dc.b	$FF,$41,$B3,$E1,$BE,$2B,$67,$E2,$1F,$2E,$32,$47,$3A,$FC,$B8,$C4
	.dc.b	$7F,$3D,$6E,$1F,$CE,$5B,$BF,$F3,$F6,$CF,$53,$9D,$F8,$5E,$46,$23
	.dc.b	$77,$DE,$37,$F0,$7E,$8F,$86,$F4,$FC,$24,$9E,$7F,$0D,$EC,$F3,$BD
	.dc.b	$B9,$FB,$BD,$4F,$7F,$C0,$FC,$0F,$3F,$E0,$FC,$0F,$C2,$F5,$BE,$1F
	.dc.b	$A3,$F1,$3D,$CF,$8B,$F0,$3F,$1B,$E0,$FC,$7C,$9E,$7E,$56,$BF,$DA
	.dc.b	$3E,$BB,$FA,$FF,$0A,$72,$FF,$72,$8C,$C6,$39,$51,$8A,$E5,$5F,$98
	.dc.b	$8C,$3C,$4C,$10,$FA,$6E,$A7,$9C,$64,$F1,$4F,$47,$66,$6B,$C5,$5C
	.dc.b	$49,$19,$9E,$0F,$F7,$9E,$5F,$A0,$F8,$3D,$EF,$37,$BD,$F8,$AE,$5F
	.dc.b	$A1,$3E,$4F,$2B,$E8,$3F,$7B,$93,$D9,$EF,$7D,$08,$FC,$0F,$1F,$D0
	.dc.b	$7D,$7F,$07,$F0,$7E,$0B,$E8,$47,$A0,$87,$E1,$89,$BF,$E3,$DF,$C4
	.dc.b	$10,$CE,$DF,$91,$71,$1B,$72,$7E,$4C,$BC,$1F,$8F,$6F,$0F,$DF,$AE
	.dc.b	$51,$8E,$CF,$CA,$F1,$3D,$BF,$B1,$F0,$5F,$93,$AF,$F1,$07,$6F,$DB
	.dc.b	$F7,$3E,$E7,$C1,$7E,$52,$7E,$CF,$C1,$7D,$FF,$86,$FC,$A4,$7D,$CF
	.dc.b	$83,$FC,$17,$C3,$7E,$0F,$E1,$3F,$29,$3F,$A8,$EE,$FC,$37,$C3,$7E
	.dc.b	$53,$83,$EC,$4F,$E1,$CF,$8F,$C9,$F7,$9C,$F2,$7E,$19,$CF,$1F,$DE
	.dc.b	$9D,$46,$0E,$47,$E2,$8A,$E9,$C4,$F3,$6F,$1F,$E2,$9E,$51,$99,$E0
	.dc.b	$7F,$8F,$7B,$BF,$17,$3C,$3B,$BF,$16,$FE,$E8,$FE,$0B,$E2,$FE,$8F
	.dc.b	$B7,$F1,$71,$FE,$4B,$ED,$FC,$5B,$FC,$2B,$C3,$FD,$D3,$FC,$33,$C3
	.dc.b	$FD,$57,$6F,$E2,$E3,$F8,$7F,$85,$F0,$B6,$FD,$47,$6F,$DD,$9F,$C1
	.dc.b	$6D,$F8,$52,$70,$7E,$01,$B1,$A8,$C1,$1C,$13,$83,$F1,$53,$C1,$B3
	.dc.b	$B6,$CE,$EB,$C5,$18,$3F,$AC,$70,$7C,$6B,$FF,$32,$75,$FE,$35,$FF
	.dc.b	$33,$78,$3F,$B0,$C3,$FC,$DD,$8F,$C6,$9F,$F3,$56,$FF,$8C,$9F,$F3
	.dc.b	$96,$FF,$19,$FF,$3A,$6F,$F8,$C7,$FC,$ED,$CB,$FB,$07,$F3,$C7,$DA
	.dc.b	$DB,$EE,$3F,$78,$F9,$44,$34,$53,$46,$08,$60,$98,$7E,$0D,$FE,$87
	.dc.b	$E3,$7F,$21,$E4,$FC,$1C,$7F,$A2,$74,$F7,$57,$F1,$8E,$3F,$76,$BF
	.dc.b	$E8,$BD,$7F,$07,$7E,$E3,$F8,$69,$F8,$33,$F8,$69,$FC,$A3,$C0,$F7
	.dc.b	$5F,$E5,$2B,$F4,$47,$D5,$E2,$FA,$64,$E6,$FD,$75,$FF,$0A,$9D,$46
	.dc.b	$23,$E3,$92,$BF,$01,$5C,$63,$81,$FD,$0F,$C0,$FE,$0E,$FF,$62,$E2
	.dc.b	$FB,$98,$7F,$4B,$EF,$FE,$0F,$3F,$D9,$A3,$FA,$27,$77,$F0,$78,$FE
	.dc.b	$D3,$E2,$FF,$47,$EE,$FE,$0F,$1F,$D1,$3B,$7E,$E6,$9F,$EA,$BE,$5F
	.dc.b	$E0,$F3,$E2,$E9,$FC,$77,$37,$DC,$AF,$85,$3F,$BE,$CF,$E1,$B4,$95
	.dc.b	$FB,$B8,$CB,$63,$35,$CE,$72,$D8,$C9,$FC,$4B,$77,$4F,$E7,$6B,$FB
	.dc.b	$57,$37,$E0,$31,$FC,$ED,$FF,$02,$3F,$A6,$1F,$E0,$8F,$1B,$FB,$DB
	.dc.b	$FD,$8B,$C4,$FE,$98,$FF,$B3,$FC,$4F,$E9,$8F,$F1,$4E,$EF,$E0,$4F
	.dc.b	$FB,$4F,$BB,$F2,$D8,$7F,$62,$F8,$5F,$EE,$04,$9F,$ED,$AD,$7F,$8E
	.dc.b	$B6,$38,$46,$88,$68,$AE,$9C,$11,$C1,$33,$FF,$05,$4F,$57,$FC,$19
	.dc.b	$F0,$BF,$E0,$3F,$43,$FC,$09,$F0,$BF,$E0,$5F,$77,$FC,$BB,$F0,$4F
	.dc.b	$4A,$FB,$18,$F7,$F3,$EC,$EB,$D5,$9B,$FF,$8E,$B0,$FF,$47,$37,$D5
	.dc.b	$E7,$EA,$F4,$F5,$64,$ED,$F2,$3F,$E1,$BE,$1B,$D9,$FE,$1A,$6F,$61
	.dc.b	$A4,$9F,$87,$C7,$EE,$69,$CC,$B5,$CD,$B5,$CD,$B5,$DB,$8F,$E2,$47
	.dc.b	$AF,$DF,$F8,$8F,$B7,$87,$E1,$BE,$FE,$0F,$8F,$EF,$FC,$8F,$63,$E4
	.dc.b	$FC,$0F,$CA,$F8,$0F,$97,$C3,$F3,$3E,$07,$E0,$E3,$FD,$7B,$8D,$C5
	.dc.b	$8E,$AF,$B1,$3E,$93,$5F,$03,$F5,$08,$FC,$F6,$BA,$30,$43,$04,$8F
	.dc.b	$F2,$35,$27,$8C,$61,$BC,$8A,$72,$2F,$84,$75,$BE,$0D,$AE,$15,$BB
	.dc.b	$FC,$5F,$B9,$F8,$CF,$80,$FC,$7D,$FF,$93,$BB,$3F,$1F,$F0,$BF,$90
	.dc.b	$F8,$5F,$C8,$FC,$17,$E3,$AB,$F1,$9D,$9F,$8E,$7E,$47,$C2,$FF,$AC
	.dc.b	$BE,$B9,$F0,$9F,$8E,$9F,$C9,$77,$E7,$E2,$75,$F0,$BE,$C9,$F5,$F2
	.dc.b	$7A,$9F,$D4,$71,$FE,$25,$A7,$D1,$E4,$FA,$7C,$9F,$89,$8C,$B6,$33
	.dc.b	$E0,$FE,$2E,$72,$D8,$CF,$83,$F9,$79,$CB,$F8,$F7,$FC,$DF,$E2,$FE
	.dc.b	$C3,$C1,$FC,$7C,$7E,$17,$C1,$FC,$7B,$F8,$7F,$07,$F1,$EF,$EB,$FC
	.dc.b	$0F,$C7,$BF,$B1,$EF,$FC,$7B,$FB,$3E,$EF,$C6,$6D,$F5,$37,$FC,$7C
	.dc.b	$FE,$C4,$FD,$09,$0F,$29,$75,$F4,$B6,$7E,$34,$62,$36,$47,$07,$FA
	.dc.b	$0B,$C2,$FD,$5C,$7F,$8E,$BA,$FD,$B3,$F9,$2D,$3D,$AF,$E5,$34,$FC
	.dc.b	$05,$7E,$97,$A3,$F0,$13,$F9,$8D,$3D,$B3,$FA,$9E,$5F,$6B,$F7,$3B
	.dc.b	$EF,$27,$D5,$D7,$CE,$49,$FB,$8D,$8D,$46,$08,$E0,$9C,$9F,$6E,$78
	.dc.b	$36,$76,$D9,$DF,$27,$DF,$CB,$E1,$DF,$F3,$0B,$FB,$E7,$F0,$D3,$FD
	.dc.b	$5F,$E2,$FF,$30,$DB,$E1,$CF,$D7,$DB,$EF,$D7,$F0,$7E,$3F,$BE,$FF
	.dc.b	$09,$DB,$E1,$E3,$FB,$CE,$3F,$83,$C7,$F7,$AB,$F8,$76,$FE,$3C,$BF
	.dc.b	$05,$E0,$7E,$8F,$7F,$DB,$57,$EE,$4F,$D2,$C6,$6F,$F4,$D8,$FE,$FA
	.dc.b	$7F,$85,$5F,$27,$85,$F2,$72,$72,$33,$5C,$A7,$84,$6A,$78,$57,$F1
	.dc.b	$4F,$03,$F8,$34,$FF,$05,$21,$FC,$9F,$BF,$F8,$33,$FD,$3F,$BB,$F0
	.dc.b	$F5,$FC,$73,$97,$F0,$F8,$FE,$7F,$BB,$F8,$37,$6F,$F0,$5C,$3F,$10
	.dc.b	$7F,$1D,$BF,$F7,$47,$73,$BB,$F8,$3C,$FA,$37,$F9,$0F,$DE,$E7,$BB
	.dc.b	$FD,$E8,$FD,$21,$92,$B9,$7F,$49,$9F,$E2,$3A,$FF,$47,$97,$F3,$2E
	.dc.b	$CF,$D1,$CF,$ED,$7A,$3F,$03,$7F,$C5,$BA,$3F,$03,$7F,$CE,$3A,$3F
	.dc.b	$02,$FE,$8E,$7F,$B7,$1F,$BF,$3F,$78,$7E,$00,$AF,$E4,$9F,$E7,$7B
	.dc.b	$7F,$3B,$7E,$16,$1F,$CE,$23,$11,$FE,$B2,$6C,$FC,$6E,$5F,$E6,$04
	.dc.b	$9E,$27,$47,$0F,$16,$C7,$18,$DE,$8A,$EB,$7E,$1B,$DB,$6B,$73,$A7
	.dc.b	$B1,$89,$DB,$EC,$F8,$7F,$E8,$9E,$BF,$F4,$5F,$3B,$FA,$37,$B9,$FD
	.dc.b	$1F,$D7,$FE,$91,$EE,$7F,$49,$F7,$FF,$A5,$69,$FD,$2F,$D1,$FE,$99
	.dc.b	$EE,$7F,$4D,$F6,$3F,$A7,$7B,$FF,$D3,$FD,$3F,$EA,$1E,$FF,$F5,$1F
	.dc.b	$86,$9C,$3E,$32,$1F,$27,$D9,$FE,$AC,$47,$D6,$D7,$46,$08,$60,$9C
	.dc.b	$DF,$96,$8E,$0D,$8D,$B5,$DE,$9F,$6E,$76,$D8,$F9,$7C,$7F,$DA,$7D
	.dc.b	$5F,$8D,$1F,$3B,$8B,$FC,$00,$FF,$BC,$BE,$BA,$BF,$83,$E2,$F9,$2F
	.dc.b	$D4,$C3,$E4,$3F,$5B,$83,$E4,$3F,$63,$7E,$7A,$2B,$FC,$0B,$83,$D9
	.dc.b	$F9,$77,$F8,$48,$7C,$3F,$0A,$DF,$C1,$7A,$AB,$F3,$3B,$FE,$03,$AF
	.dc.b	$F1,$5D,$57,$AF,$F0,$D1,$8A,$FA,$7A,$7E,$45,$FC,$A1,$8C,$7F,$1C
	.dc.b	$E0,$91,$B2,$46,$9B,$1F,$92,$37,$19,$3D,$BE,$DF,$E4,$A3,$D9,$ED
	.dc.b	$FA,$DF,$F1,$5E,$AF,$82,$FC,$A4,$7C,$1F,$82,$FC,$A3,$F0,$FE,$1F
	.dc.b	$86,$6B,$7C,$1E,$17,$C3,$DF,$84,$C7,$C3,$FE,$5F,$1F,$99,$F0,$BF
	.dc.b	$96,$7F,$79,$DB,$F1,$F8,$FE,$66,$7F,$B3,$E4,$BC,$86,$92,$FF,$1F
	.dc.b	$C7,$F8,$B7,$F1,$E7,$D4,$32,$53,$35,$D5,$CB,$5C,$D3,$B7,$F2,$91
	.dc.b	$FC,$03,$C4,$FD,$CB,$F8,$53,$F2,$27,$E4,$CF,$C9,$1F,$C1,$BC,$2F
	.dc.b	$DC,$BF,$C2,$3C,$2F,$DC,$BF,$8D,$EB,$FC,$7B,$F9,$DE,$BF,$95,$D7
	.dc.b	$ED,$8F,$B6,$7D,$DE,$BF,$BF,$A7,$E2,$E3,$2D,$8C,$E9,$F9,$02,$6F
	.dc.b	$F3,$6B,$F1,$C7,$3D,$BF,$0C,$D2,$5F,$38,$CB,$FC,$97,$6F,$C5,$4F
	.dc.b	$F3,$3D,$BE,$1B,$FC,$D7,$C4,$F8,$79,$7F,$5A,$F1,$3F,$8D,$CF,$C3
	.dc.b	$3F,$68,$7C,$33,$F6,$87,$F2,$CF,$13,$E1,$CF,$F6,$2E,$3F,$C3,$3E
	.dc.b	$CD,$3E,$23,$FC,$34,$9A,$7E,$05,$B7,$F7,$DB,$A7,$DD,$6E,$BF,$8B
	.dc.b	$9F,$C0,$B5,$FC,$0F,$1F,$6B,$E8,$48,$FE,$78,$7F,$42,$F0,$FF,$8B
	.dc.b	$BF,$D5,$7C,$3F,$D1,$3F,$0A,$7F,$AD,$78,$7F,$0A,$3F,$AD,$73,$7E
	.dc.b	$11,$FE,$B9,$CD,$F8,$47,$F8,$F7,$37,$C2,$BF,$E2,$BC,$DF,$84,$9F
	.dc.b	$3E,$7F,$6D,$FC,$0E,$9F,$79,$FB,$BB,$7D,$97,$ED,$C6,$5D,$EF,$C7
	.dc.b	$7F,$99,$1C,$27,$51,$FC,$96,$3F,$82,$F8,$3F,$80,$8C,$5F,$F0,$1F
	.dc.b	$07,$F8,$07,$83,$FC,$22,$FF,$9B,$78,$3F,$C2,$1F,$E7,$1E,$0F,$F0
	.dc.b	$5F,$07,$F8,$37,$81,$FC,$16,$7F,$97,$77,$FF,$05,$7F,$9F,$77,$FE
	.dc.b	$03,$8B,$FC,$05,$D3,$E7,$7E,$96,$BF,$6D,$BC,$1F,$67,$7F,$BB,$C5
	.dc.b	$F6,$71,$FC,$15,$6A,$31,$7C,$67,$24,$70,$4B,$E9,$5F,$E1,$3E,$EF
	.dc.b	$E0,$75,$FA,$13,$FA,$B1,$FC,$0C,$FF,$02,$F4,$7E,$05,$FF,$0E,$CF
	.dc.b	$F5,$47,$FC,$35,$E1,$FF,$54,$7F,$C3,$9E,$1F,$F2,$9E,$8F,$C0,$47
	.dc.b	$F3,$4E,$8F,$C0,$3F,$94,$3F,$AA,$7C,$3F,$F2,$F2,$38,$21,$FD,$9D
	.dc.b	$CB,$4C,$ED,$F6,$BC,$44,$EA,$30,$DC,$BF,$B2,$B6,$BF,$A3,$B7,$6F
	.dc.b	$E8,$4D,$B7,$C3,$7F,$43,$F8,$7F,$E8,$9E,$77,$F4,$5F,$3B,$FA,$37
	.dc.b	$BD,$32,$FE,$DD,$EF,$7F,$49,$F7,$3F,$A6,$FD,$6B,$D1,$BF,$0F,$8F
	.dc.b	$4B,$3F,$0F,$AF,$67,$7E,$FF,$0F,$85,$E3,$F0,$5E,$3E,$49,$B7,$C0
	.dc.b	$E6,$F5,$B4,$F5,$97,$9B,$D6,$DE,$6F,$8A,$DA,$E6,$FC,$B2,$73,$FC
	.dc.b	$87,$47,$F9,$33,$83,$FD,$41,$1F,$A8,$5E,$C7,$C4,$9D,$7C,$3F,$C0
	.dc.b	$F6,$7E,$0F,$C3,$FC,$2F,$83,$F8,$95,$CF,$DC,$F8,$BF,$05,$F1,$BE
	.dc.b	$1F,$E1,$CF,$7E,$4F,$86,$F8,$78,$FE,$5F,$C1,$FA,$37,$F1,$7C,$3F
	.dc.b	$CE,$25,$79,$DA,$A3,$6E,$48,$67,$90,$B9,$1C,$4B,$E7,$7C,$AB,$81
	.dc.b	$B9,$D3,$D2,$34,$4B,$EB,$7D,$2F,$9C,$62,$F8,$DF,$4A,$D3,$FE,$39
	.dc.b	$F0,$7F,$24,$FD,$3F,$84,$FC,$93,$FE,$3D,$EF,$FC,$93,$FE,$4D,$EF
	.dc.b	$FC,$8F,$C2,$7E,$4B,$BF,$F2,$33,$F6,$FB,$BF,$C4,$3D,$CF,$84,$BF
	.dc.b	$09,$8F,$87,$CF,$B3,$F9,$2C,$3D,$9C,$3E,$7E,$2F,$A4,$FE,$4A,$3F
	.dc.b	$12,$DE,$3F,$C5,$46,$0F,$CA,$97,$8B,$F2,$AD,$F0,$3F,$2A,$DF,$03
	.dc.b	$AB,$C2,$B9,$92,$FE,$47,$C3,$7E,$BB,$C1,$FC,$95,$7C,$9E,$77,$83
	.dc.b	$F9,$38,$FC,$D7,$81,$F9,$37,$FC,$C9,$E2,$FE,$DB,$C0,$FC,$9C,$7E
	.dc.b	$7F,$BF,$F2,$6F,$3E,$FF,$C9,$BD,$7B,$FF,$26,$FE,$98,$F4,$17,$6F
	.dc.b	$33,$E9,$DB,$E2,$11,$C1,$38,$7E,$1C,$F0,$6C,$ED,$B3,$BE,$1F,$8F
	.dc.b	$5B,$6D,$7F,$A0,$BB,$BF,$0F,$5F,$E8,$FF,$1F,$F0,$EF,$FA,$47,$BB
	.dc.b	$F0,$F8,$7F,$29,$EE,$FD,$AC,$FC,$9E,$DF,$C3,$BF,$2F,$B7,$E1,$D7
	.dc.b	$F0,$1D,$BE,$1B,$FC,$0F,$6F,$C3,$57,$C2,$24,$79,$79,$FE,$DE,$FE
	.dc.b	$A2,$6F,$E8,$AF,$64,$EA,$31,$3A,$8C,$1F,$CB,$39,$FF,$01,$5F,$E9
	.dc.b	$DD,$BF,$97,$F2,$FD,$F8,$FC,$17,$2F,$DF,$7F,$D6,$1C,$BF,$7D,$FE
	.dc.b	$39,$E0,$FF,$11,$7F,$8F,$77,$7F,$33,$7F,$90,$78,$BF,$CC,$DF,$E4
	.dc.b	$5E,$17,$AB,$AB,$F6,$FC,$1F,$BD,$9F,$E0,$5D,$9F,$BC,$69,$FA,$82
	.dc.b	$9F,$95,$75,$C1,$FB,$E8,$C1,$B7,$13,$D5,$BD,$3F,$B1,$9D,$5F,$8A
	.dc.b	$31,$1F,$EC,$DE,$9F,$C0,$DF,$F3,$7F,$0B,$F0,$2F,$F6,$7E,$AF,$C0
	.dc.b	$DF,$FB,$23,$AB,$F8,$3E,$1F,$CF,$FA,$BF,$03,$7F,$CF,$FA,$BF,$83
	.dc.b	$F8,$3F,$C1,$AB,$F9,$3C,$FF,$8A,$B8,$5E,$07,$F0,$89,$FE,$91,$E2
	.dc.b	$FB,$B7,$F3,$47,$DE,$3F,$4D,$A7,$E7,$7A,$FE,$F9,$3A,$FE,$FC,$E8
	.dc.b	$E0,$7E,$32,$78,$47,$E0,$E3,$FA,$B7,$81,$FC,$0B,$3F,$EA,$DD,$5F
	.dc.b	$83,$7F,$AC,$75,$7E,$0D,$FF,$04,$75,$7E,$0E,$FF,$84,$75,$7E,$0D
	.dc.b	$FE,$15,$BF,$E0,$DF,$E1,$9D,$5F,$83,$7F,$87,$75,$7E,$72,$7F,$D5
	.dc.b	$D1,$F1,$35,$FE,$7F,$18,$8F,$E7,$8D,$C7,$E4,$79,$DF,$CC,$89,$3C
	.dc.b	$4F,$EA,$53,$BA,$E7,$3A,$9E,$6D,$AE,$71,$C7,$A9,$7C,$63,$9D,$6E
	.dc.b	$F9,$C6,$E7,$55,$C2,$B5,$3E,$2F,$86,$FE,$8F,$F0,$7F,$D2,$3D,$FF
	.dc.b	$E9,$3F,$07,$FD,$2B,$D2,$FE,$97,$F0,$7F,$D3,$34,$FE,$9B,$F0,$3F
	.dc.b	$D3,$BD,$4F,$E9,$FF,$03,$FD,$43,$D5,$FE,$A3,$F0,$7F,$D4,$BE,$1B
	.dc.b	$FA,$9F,$B5,$FD,$53,$DA,$FE,$AB,$F0,$9E,$2E,$4F,$27,$1F,$9F,$93
	.dc.b	$1C,$BE,$82,$38,$24,$6C,$87,$C8,$2F,$37,$C8,$70,$47,$04,$E0,$F8
	.dc.b	$F1,$F6,$B9,$BE,$1B,$F3,$FE,$B3,$FC,$12,$DF,$6F,$E2,$3F,$37,$97
	.dc.b	$E2,$3B,$F6,$FE,$23,$C7,$E1,$B8,$E5,$FE,$94,$C1,$F0,$DF,$1A,$BE
	.dc.b	$B7,$16,$B4,$FF,$05,$D3,$E1,$BE,$45,$7D,$BE,$2F,$82,$FF,$82,$E5
	.dc.b	$F0,$DC,$63,$FC,$3D,$18,$7D,$51,$9E,$24,$E6,$79,$4E,$63,$19,$F9
	.dc.b	$E3,$04,$9E,$64,$38,$96,$38,$37,$1F,$5C,$F1,$8D,$92,$7E,$27,$C3
	.dc.b	$78,$7E,$1B,$F2,$53,$F9,$9E,$DF,$27,$C3,$7E,$4A,$3E,$A7,$C3,$7E
	.dc.b	$49,$FA,$DF,$0F,$F9,$27,$F4,$1D,$9F,$93,$7E,$D7,$C3,$FE,$49,$FB
	.dc.b	$9F,$0F,$F9,$27,$CC,$7C,$8D,$7E,$6E,$BF,$85,$8F,$D6,$9F,$8C,$DB
	.dc.b	$F0,$F1,$9E,$0F,$CC,$CF,$E3,$67,$24,$78,$99,$2C,$F1,$EB,$FC,$EC
	.dc.b	$75,$33,$3F,$C3,$BA,$BF,$27,$1F,$7C,$FA,$DD,$5F,$92,$7F,$88,$F5
	.dc.b	$7E,$49,$FC,$27,$57,$E4,$9F,$C3,$75,$7D,$DE,$AF,$C8,$47,$F1,$8E
	.dc.b	$AF,$C0,$75,$7E,$3A,$3F,$8D,$F5,$7C,$2D,$3D,$5B,$FB,$B7,$F4,$75
	.dc.b	$E6,$FC,$E4,$9D,$57,$0E,$0F,$6C,$FC,$19,$CB,$B8,$CD,$F4,$8F,$E3
	.dc.b	$1C,$1F,$02,$3F,$A9,$78,$7F,$C5,$27,$F6,$07,$F1,$DF,$0F,$E0,$4F
	.dc.b	$F5,$7F,$0F,$E0,$3F,$D6,$7C,$2F,$C1,$57,$F3,$6F,$0B,$F8,$A3,$FC
	.dc.b	$97,$C2,$F8,$15,$FD,$7F,$C2,$F6,$6B,$E6,$26,$BE,$6D,$BD,$9B,$EA
	.dc.b	$31,$8F,$AE,$30,$47,$EE,$57,$F3,$8D,$FD,$97,$FB,$E9,$FB,$67,$B0
	.dc.b	$FB,$67,$F3,$97,$F6,$9C,$BE,$C7,$F8,$2F,$2F,$DB,$9F,$E7,$DA,$FB
	.dc.b	$1F,$E0,$9C,$7E,$CA,$FE,$AF,$C7,$EC,$F0,$BF,$7B,$1F,$B4,$AF,$D0
	.dc.b	$E7,$FC,$13,$A7,$F6,$DC,$1F,$BA,$9F,$DB,$56,$DA,$7E,$9C,$E3,$7D
	.dc.b	$1C,$1A,$BE,$8E,$AF,$AD,$73,$BF,$E0,$33,$FC,$BF,$C1,$FE,$02,$FF
	.dc.b	$6C,$F0,$7F,$80,$BF,$7C,$FE,$02,$7F,$5D,$E5,$FE,$31,$DF,$FC,$06
	.dc.b	$3F,$B9,$F7,$FE,$03,$0F,$E6,$1D,$FF,$C0,$67,$F9,$47,$7F,$F0,$17
	.dc.b	$CD,$B7,$A7,$6F,$BF,$CB,$F6,$67,$F3,$1C,$DF,$7A,$73,$1F,$72,$30
	.dc.b	$49,$E1,$8F,$F8,$A7,$9B,$EF,$D7,$F3,$43,$EF,$9F,$78,$91,$FE,$2F
	.dc.b	$E0,$FB,$F1,$FD,$E7,$C3,$FE,$69,$CD,$F7,$E3,$F3,$47,$F4,$83,$FA
	.dc.b	$B7,$2F,$DF,$7F,$C1,$BC,$BF,$7F,$CC,$FE,$78,$48,$FE,$B8,$D7,$FA
	.dc.b	$53,$71,$FE,$AC,$DC,$7F,$A9,$B6,$7F,$A6,$B7,$5F,$E9,$6D,$DB,$FC
	.dc.b	$FA,$47,$1E,$BF,$F4,$DF,$86,$FE,$9D,$ED,$4C,$3F,$B7,$7C,$37,$F5
	.dc.b	$13,$FA,$97,$B3,$23,$FB,$97,$B3,$FD,$57,$CF,$FE,$B5,$F5,$8F,$6A
	.dc.b	$FC,$3E,$3D,$BC,$F4,$EB,$E0,$37,$E9,$F9,$78,$6E,$9F,$E2,$BC,$93
	.dc.b	$88,$FF,$7B,$46,$1D,$46,$79,$5C,$BF,$F0,$02,$B5,$5C,$2B,$53,$96
	.dc.b	$C3,$95,$5C,$2F,$9C,$E6,$31,$7C,$63,$C2,$43,$B1,$67,$8D,$74,$BF
	.dc.b	$8B,$F0,$1F,$1B,$DB,$F9,$D7,$FE,$07,$E4,$F9,$3F,$01,$F2,$BD,$AF
	.dc.b	$97,$ED,$7C,$CF,$87,$FF,$01,$73,$7C,$EF,$6F,$E7,$F3,$7D,$0F,$87
	.dc.b	$FA,$3F,$09,$F3,$B4,$F9,$5C,$BE,$EE,$1F,$14,$79,$78,$FF,$12,$FE
	.dc.b	$40,$87,$E4,$0B,$C7,$F9,$06,$C7,$07,$44,$8D,$36,$35,$18,$23,$82
	.dc.b	$47,$E6,$0E,$A7,$F8,$9B,$B7,$F2,$8F,$E0,$FD,$DF,$C6,$FC,$27,$E4
	.dc.b	$E3,$F2,$A7,$F9,$0F,$B7,$F2,$8F,$E1,$BD,$EF,$CA,$3F,$8B,$F8,$4F
	.dc.b	$C9,$BF,$8D,$ED,$FC,$BF,$C2,$7E,$63,$E1,$FF,$C6,$5F,$58,$F8,$6F
	.dc.b	$CA,$5F,$C7,$3F,$05,$DD,$F2,$FB,$FF,$07,$18,$35,$C5,$F3,$9F,$C1
	.dc.b	$93,$8B,$F0,$73,$C2,$75,$18,$AE,$84,$AE,$75,$B2,$77,$FE,$26,$3F
	.dc.b	$CC,$5D,$DF,$8A,$BE,$BD,$DF,$89,$9F,$D5,$76,$FE,$25,$FD,$B9,$FA
	.dc.b	$DE,$BF,$C4,$BF,$AF,$EA,$FC,$4B,$FB,$1F,$07,$F6,$CF,$8B,$4F,$C5
	.dc.b	$78,$DF,$C2,$63,$EB,$ED,$FC,$39,$B1,$97,$F0,$47,$DD,$2C,$6D,$B1
	.dc.b	$BC,$FF,$0B,$1B,$6C,$ED,$B2,$E9,$FC,$2D,$7E,$A4,$F8,$67,$E1,$0F
	.dc.b	$F3,$97,$8B,$F8,$4A,$FF,$3F,$74,$7E,$AA,$7F,$D1,$5E,$2F,$C2,$7F
	.dc.b	$86,$F8,$BF,$C5,$DF,$84,$7E,$A8,$FB,$B8,$F9,$FB,$7F,$81,$47,$DD
	.dc.b	$6E,$1F,$6E,$3E,$41,$0D,$1F,$87,$2E,$7F,$82,$9E,$24,$8D,$4F,$37
	.dc.b	$7C,$9F,$83,$D7,$F8,$5C,$7F,$27,$F1,$3F,$05,$7F,$E9,$6E,$4F,$E1
	.dc.b	$73,$FC,$AF,$C4,$FE,$16,$FF,$2D,$F1,$3E,$13,$FC,$B3,$93,$E1,$67
	.dc.b	$F9,$6E,$4F,$C1,$CF,$F3,$0E,$4F,$C1,$BE,$2D,$3F,$07,$E2,$FE,$9A
	.dc.b	$3E,$DB,$6F,$EB,$B7,$4F,$DF,$EF,$F6,$DB,$C1,$F2,$EF,$F0,$11,$8A
	.dc.b	$FF,$50,$72,$FD,$EC,$3F,$53,$E0,$FF,$02,$BF,$BD,$1F,$EB,$9E,$6F
	.dc.b	$BD,$87,$FA,$DB,$9B,$FA,$9F,$81,$FC,$0A,$FF,$A4,$F8,$1F,$C0,$9F
	.dc.b	$E9,$7E,$07,$F0,$27,$FA,$87,$7F,$F0,$2E,$EF,$44,$7D,$2C,$7E,$A7
	.dc.b	$07,$D7,$8C,$11,$C1,$3A,$7E,$B4,$7E,$07,$5F,$B4,$4A,$E7,$3A,$9F
	.dc.b	$98,$7F,$00,$EE,$FD,$F3,$F9,$D2,$57,$EF,$9F,$B8,$7E,$F8,$FB,$87
	.dc.b	$EF,$8F,$B8,$7F,$6B,$F1,$3F,$A3,$3F,$DB,$7C,$4F,$E8,$CF,$F6,$CF
	.dc.b	$13,$F8,$83,$F9,$4F,$09,$F1,$1F,$C9,$09,$01,$87,$F5,$E9,$DC,$66
	.dc.b	$31,$3C,$23,$79,$F1,$D1,$7C,$27,$27,$F7,$37,$17,$CA,$3A,$57,$19
	.dc.b	$CB,$6F,$B5,$6E,$B9,$DF,$69,$CD,$6E,$79,$9D,$E3,$2D,$33,$3C,$5B
	.dc.b	$7E,$07,$47,$67,$11,$B8,$C4,$F4,$8C,$5F,$94,$97,$E3,$AD,$9D,$AB
	.dc.b	$85,$F4,$23,$82,$46,$C8,$72,$25,$76,$7C,$ED,$BF,$1D,$F8,$63,$17
	.dc.b	$DA,$B9,$C6,$2F,$BC,$74,$AE,$04,$7A,$B6,$FA,$C6,$08,$E0,$97,$DE
	.dc.b	$3A,$1A,$71,$7E,$4F,$87,$FE,$EA,$49,$FA,$6D,$7E,$93,$63,$F2,$8D
	.dc.b	$BF,$C4,$36,$7F,$1C,$DB,$FC,$9B,$6C,$D3,$1F,$11,$FE,$06,$F3,$FF
	.dc.b	$C0,$FE,$7F,$F8,$23,$E1,$BF,$C1,$3E,$FF,$F8,$2B,$D7,$FF,$05,$FC
	.dc.b	$17,$F8,$33,$E0,$3F,$C1,$BB,$7F,$83,$BD,$9F,$F0,$7F,$AB,$FE,$10
	.dc.b	$F8,$0F,$F0,$8F,$B3,$FE,$12,$F7,$7F,$C2,$7F,$03,$FE,$14,$F8,$1F
	.dc.b	$17,$2F,$D4,$CB,$E7,$72,$7C,$66,$F3,$FD,$6E,$7F,$B1,$5B,$3E,$64
	.dc.b	$66,$77,$19,$E7,$FA,$73,$98,$D9,$9B,$EC,$E0,$FA,$F3,$C1,$D4,$F1
	.dc.b	$8C,$46,$A7,$8C,$62,$FC,$B3,$9A,$E1,$5A,$23,$82,$47,$62,$1C,$4B
	.dc.b	$3D,$09,$7E,$E8,$FD,$B7,$B1,$FE,$3E,$23,$FB,$C6,$BA,$30,$43,$04
	.dc.b	$9F,$E4,$11,$C1,$B1,$B6,$BB,$9F,$E3,$13,$B6,$C7,$F0,$CF,$81,$FE
	.dc.b	$58,$FF,$18,$F8,$2A,$DF,$83,$CE,$1F,$E8,$C9,$7C,$1E,$30,$FE,$59
	.dc.b	$0F,$7B,$18,$7F,$A7,$60,$63,$5B,$EF,$58,$FF,$77,$0F,$86,$C7,$C1
	.dc.b	$E7,$E1,$B5,$F0,$DB,$F8,$5E,$1F,$0B,$C7,$E1,$79,$7A,$9E,$DD,$7F
	.dc.b	$9D,$72,$7D,$87,$EB,$13,$8B,$EA,$BF,$8B,$6E,$FF,$85,$E1,$FB,$11
	.dc.b	$8E,$6F,$C1,$D7,$DE,$9C,$91,$E4,$DA,$E5,$18,$3F,$86,$5F,$96,$3F
	.dc.b	$A8,$B6,$8D,$24,$FD,$BA,$FA,$EF,$DB,$3D,$A6,$4A,$66,$BA,$B9,$6B
	.dc.b	$FD,$1F,$97,$E8,$3F,$47,$97,$E9,$72,$FD,$3E,$5F,$A9,$C9,$F5,$79
	.dc.b	$7E,$B7,$C3,$7D,$7E,$6F,$AB,$8F,$F5,$6A,$72,$FD,$68,$C7,$27,$D6
	.dc.b	$75,$C7,$F5,$9F,$27,$0F,$C0,$79,$7A,$BF,$83,$23,$82,$1F,$95,$72
	.dc.b	$D3,$3C,$7F,$95,$8C,$B6,$33,$C5,$F9,$69,$CB,$63,$27,$E5,$E3,$2D
	.dc.b	$33,$C5,$F9,$A8,$CB,$63,$39,$FF,$A1,$5B,$3A,$8C,$47,$9C,$97,$E5
	.dc.b	$6D,$F8,$DA,$78,$4B,$7D,$DB,$5D,$2B,$91,$2B,$93,$67,$93,$6B,$93
	.dc.b	$6B,$95,$70,$25,$70,$6D,$70,$6C,$F0,$9D,$12,$74,$D8,$D4,$60,$8E
	.dc.b	$09,$7F,$1C,$93,$F2,$8F,$9B,$F0,$DF,$C1,$C8,$FE,$45,$AF,$F1,$96
	.dc.b	$E1,$FC,$5D,$B8,$7F,$16,$6C,$FF,$15,$6E,$1F,$C5,$9B,$81,$A6,$3D
	.dc.b	$EF,$E3,$1E,$B7,$F1,$9F,$5B,$F8,$D7,$C3,$7F,$1B,$F8,$2F,$E3,$DF
	.dc.b	$58,$F4,$AF,$C1,$E3,$D8,$CF,$BD,$AF,$53,$7E,$F7,$0F,$5F,$8F,$C1
	.dc.b	$F2,$F8,$5E,$7F,$09,$D3,$E1,$7C,$7C,$DF,$14,$FC,$1F,$4F,$C8,$69
	.dc.b	$9E,$9F,$BD,$19,$6C,$67,$A7,$ED,$CE,$5B,$19,$3E,$D4,$65,$A6,$7A
	.dc.b	$7E,$DC,$65,$B1,$9E,$9F,$BA,$D9,$D4,$62,$3C,$E4,$BF,$2B,$6F,$C6
	.dc.b	$D3,$C2,$5B,$EE,$DA,$E9,$5F,$9F,$E3,$7C,$2F,$F5,$EF,$BC,$FA,$96
	.dc.b	$C1,$5A,$8F,$F4,$74,$EA,$73,$5C,$79,$7F,$D1,$57,$BB,$5E,$E3,$0D
	.dc.b	$AE,$13,$DE,$73,$5C,$A7,$37,$DA,$BC,$33,$96,$9C,$67,$75,$CA,$30
	.dc.b	$6A,$B9,$3E,$4A,$E4,$D3,$65,$8D,$DF,$9E,$39,$57,$59,$F8,$BF,$0D
	.dc.b	$E0,$F8,$6F,$0F,$BF,$E2,$F8,$5F,$1F,$C4,$79,$3E,$1B,$CB,$F0,$DE
	.dc.b	$6F,$88,$F3,$FC,$3F,$BB,$EB,$5F,$11,$7D,$7C,$7C,$3E,$7E,$1F,$5F
	.dc.b	$0F,$BF,$87,$E1,$F0,$FF,$03,$A2,$73,$7E,$2C,$87,$E0,$7B,$3F,$1A
	.dc.b	$47,$3D,$5F,$8C,$27,$67,$E3,$1B,$1F,$8C,$6F,$47,$E3,$1B,$D9,$F8
	.dc.b	$CA,$FE,$47,$D5,$F1,$49,$8F,$93,$0F,$99,$18,$7E,$5C,$67,$AF,$E7
	.dc.b	$46,$A3,$1D,$9F,$5A,$72,$47,$04,$9F,$B3,$19,$6C,$7E,$53,$B3,$EB
	.dc.b	$47,$DA,$EB,$FA,$CF,$DC,$EA,$FA,$CF,$DE,$E9,$FA,$CF,$E0,$3A,$3E
	.dc.b	$B7,$3F,$D5,$8F,$C6,$74,$7F,$A5,$1B,$CD,$F5,$9C,$F2,$FD,$67,$DE
	.dc.b	$43,$FD,$C7,$2F,$87,$FC,$11,$23,$AB,$5F,$CA,$46,$DC,$90,$CC,$6F
	.dc.b	$6F,$DB,$4E,$E7,$26,$E3,$33,$BD,$93,$FC,$06,$B8,$56,$A3,$84,$EB
	.dc.b	$91,$5A,$8F,$C4,$71,$4A,$FC,$43,$6B,$F1,$0D,$AE,$13,$DF,$0F,$13
	.dc.b	$4E,$E5,$3E,$3B,$AA,$F0,$B6,$79,$C6,$27,$8B,$5F,$82,$4C,$3D,$8D
	.dc.b	$C3,$D9,$5C,$89,$7D,$49,$3F,$C0,$B8,$3F,$16,$FE,$73,$83,$F1,$6F
	.dc.b	$E7,$B8,$3F,$16,$FE,$83,$83,$F1,$6F,$E8,$B8,$3F,$16,$FE,$2B,$A7
	.dc.b	$F7,$EF,$A3,$2F,$0B,$72,$FC,$17,$2B,$7F,$3F,$C3,$FF,$09,$25,$7E
	.dc.b	$51,$AF,$E2,$1B,$BF,$E1,$1B,$BF,$29,$E0,$43,$81,$67,$83,$63,$83
	.dc.b	$A2,$47,$F4,$5E,$6F,$E0,$91,$F8,$22,$70,$FE,$09,$B1,$A8,$FC,$5F
	.dc.b	$C3,$FF,$23,$F8,$7F,$E4,$99,$7F,$25,$F8,$7F,$3E,$1F,$95,$CF,$FD
	.dc.b	$51,$B7,$C6,$8F,$B4,$7E,$1F,$3F,$BA,$4D,$3E,$EE,$DF,$6A,$77,$A7
	.dc.b	$E1,$6B,$51,$C6,$B6,$FD,$12,$1C,$3A,$5C,$BF,$82,$AF,$EB,$DC,$BF
	.dc.b	$82,$7F,$B0,$72,$FE,$09,$FE,$C5,$CB,$F8,$27,$FB,$27,$2F,$E0,$B1
	.dc.b	$FD,$EF,$2F,$E0,$9F,$2E,$DE,$DA,$F2,$EB,$E7,$D7,$CB,$5F,$A4,$3C
	.dc.b	$B3,$AA,$E1,$3A,$8C,$6D,$F0,$5B,$5C,$27,$44,$7E,$11,$2F,$EC,$13
	.dc.b	$93,$EC,$13,$AB,$FC,$3D,$8F,$F2,$3A,$F4,$E3,$DF,$0E,$98,$7E,$CF
	.dc.b	$9B,$F3,$55,$FC,$3F,$E1,$FF,$B1,$11,$FE,$5A,$D7,$F9,$0B,$73,$FE
	.dc.b	$3A,$DB,$FD,$CB,$67,$A1,$C8,$87,$22,$CF,$26,$CF,$27,$F7,$C7,$F8
	.dc.b	$03,$C2,$FE,$FA,$FF,$80,$BC,$2F,$EF,$AF,$F8,$13,$C2,$FE,$FA,$FF
	.dc.b	$81,$BC,$2F,$EF,$AF,$F4,$6E,$EF,$E9,$0F,$F3,$AF,$0B,$FC,$01,$F0
	.dc.b	$FF,$E0,$1F,$87,$FF,$01,$7C,$3F,$F8,$0F,$E1,$FD,$5C,$3F,$CB,$B8
	.dc.b	$3F,$3F,$C5,$F2,$F2,$FA,$9C,$7F,$79,$B1,$9A,$FC,$34,$60,$FB,$C5
	.dc.b	$3E,$A9,$FE,$25,$E5,$FB,$4F,$F8,$13,$BF,$F9,$C4,$62,$3F,$D5,$F7
	.dc.b	$DE,$BA,$91,$FD,$97,$2F,$DC,$8F,$44,$FC,$C3,$E9,$9F,$44,$9C,$BF
	.dc.b	$45,$BC,$BF,$45,$A7,$08,$D1,$23,$4D,$8D,$46,$08,$E0,$93,$FE,$38
	.dc.b	$7E,$E1,$FE,$03,$3E,$E1,$FE,$03,$3E,$E1,$FE,$03,$3E,$E1,$FE,$03
	.dc.b	$3E,$E1,$FC,$E4,$FF,$01,$F0,$7D,$C7,$FC,$0B,$BF,$DC,$7F,$C0,$FB
	.dc.b	$7D,$CC,$9B,$7F,$1A,$F8,$7F,$F2,$11,$2B,$FC,$1A,$D7,$FC,$02,$DC
	.dc.b	$FF,$BE,$B7,$3F,$EB,$2D,$9F,$EF,$4D,$CF,$FB,$BB,$70,$34,$93,$E0
	.dc.b	$21,$E0,$2D,$F8,$1B,$7E,$06,$9E,$02,$DF,$81,$B7,$E0,$6D,$F8,$1F
	.dc.b	$F1,$1B,$6F,$B4,$78,$48,$E0,$97,$E1,$8C,$B6,$33,$1E,$33,$25,$73
	.dc.b	$7E,$58,$CB,$63,$37,$E7,$9C,$B6,$33,$7E,$7B,$F0,$12,$BB,$36,$FA
	.dc.b	$36,$F9,$36,$BE,$F6,$9F,$A9,$97,$C1,$FF,$A1,$49,$87,$F9,$41,$AF
	.dc.b	$F9,$11,$B8,$FF,$90,$1B,$87,$F8,$71,$B3,$FE,$3D,$6E,$3F,$E3,$A6
	.dc.b	$DF,$82,$7B,$10,$EC,$7F,$98,$3E,$0F,$FD,$6B,$F5,$BF,$E4,$88,$FF
	.dc.b	$57,$2F,$84,$CF,$C2,$6B,$E1,$37,$C7,$C3,$E1,$38,$FC,$27,$2F,$84
	.dc.b	$E7,$F1,$1D,$3E,$17,$AF,$C4,$76,$F8,$6E,$FF,$11,$E0,$F8,$7F,$0F
	.dc.b	$C4,$7A,$39,$3F,$55,$C5,$F8,$C2,$1F,$63,$A7,$EB,$3F,$50,$9C,$DF
	.dc.b	$51,$BC,$DF,$51,$B3,$C2,$74,$49,$D3,$63,$51,$82,$38,$24,$F9,$3A
	.dc.b	$3F,$3B,$D3,$F7,$E7,$33,$E7,$35,$D3,$F8,$56,$9E,$42,$E1,$E1,$6E
	.dc.b	$1E,$06,$D7,$4C,$3F,$2F,$18,$21,$FA,$DC,$3F,$9C,$74,$3C,$7F,$BA
	.dc.b	$49,$FB,$AD,$7F,$DC,$D1,$B7,$24,$33,$1B,$A6,$5F,$E8,$9A,$E1,$5A
	.dc.b	$38,$4E,$96,$49,$DC,$E6,$31,$7D,$6F,$A3,$6F,$8D,$F0,$9F,$07,$C3
	.dc.b	$FE,$63,$E1,$FF,$33,$F0,$FF,$9A,$F8,$7F,$CD,$FC,$3F,$E7,$3E,$1F
	.dc.b	$F3,$9A,$F7,$F7,$7C,$9A,$23,$1F,$0F,$AD,$3F,$D4,$94,$F8,$7C,$5F
	.dc.b	$AA,$1F,$0F,$8E,$14,$3E,$1F,$0B,$E1,$EF,$C3,$E3,$E1,$F3,$F0,$FA
	.dc.b	$F8,$7F,$76,$3F,$0B,$3F,$F5,$CF,$47,$E2,$DF,$BB,$C1,$F8,$07,$F1
	.dc.b	$84,$70,$4D,$FF,$1D,$19,$6C,$66,$3F,$20,$64,$AE,$6B,$F2,$51,$96
	.dc.b	$C7,$E7,$BC,$FF,$AB,$3F,$7A,$AB,$F8,$12,$4F,$DD,$E8,$FB,$11,$F6
	.dc.b	$BA,$7D,$F1,$8C,$3E,$04,$F0,$BE,$77,$C6,$F6,$49,$F0,$1C,$CF,$E2
	.dc.b	$FD,$3F,$83,$7F,$8C,$F4,$7E,$0D,$FE,$37,$CF,$F8,$37,$FD,$CB,$7F
	.dc.b	$83,$C3,$EB,$F3,$7D,$8E,$6F,$B3,$CD,$F6,$B9,$BE,$DF,$2F,$DA,$C7
	.dc.b	$8E,$1F,$86,$E4,$FB,$71,$83,$F0,$DC,$7E,$75,$C7,$FE,$2B,$E6,$7C
	.dc.b	$3F,$E1,$89,$3C,$78,$FF,$17,$18,$8F,$CB,$63,$F9,$48,$C6,$FF,$9B
	.dc.b	$6C,$66,$79,$5F,$66,$CE,$9F,$CD,$36,$FB,$C6,$88,$78,$0D,$98,$BF
	.dc.b	$03,$6F,$C0,$EE,$79,$B6,$FC,$4D,$BF,$04,$62,$F9,$CE,$4E,$87,$80
	.dc.b	$C1,$1D,$92,$FC,$38,$7C,$DB,$EA,$4A,$E7,$87,$92,$F8,$47,$56,$C6
	.dc.b	$6F,$D7,$18,$C3,$CD,$18,$25,$71,$6D,$71,$8C,$11,$C1,$27,$9C,$65
	.dc.b	$B1,$9B,$F4,$4F,$BE,$BA,$D6,$89,$3A,$C3,$F1,$9F,$0D,$FC,$40,$97
	.dc.b	$F8,$26,$BF,$4D,$B8,$FF,$23,$6E,$1F,$C8,$1B,$3F,$C7,$DB,$8F,$F1
	.dc.b	$E6,$E2,$69,$8F,$7F,$F9,$0F,$B5,$FC,$8B,$DF,$FE,$47,$ED,$FF,$24
	.dc.b	$F7,$FF,$92,$FB,$9F,$C9,$BD,$FF,$E4,$FE,$EF,$F2,$8F,$7F,$F9,$67
	.dc.b	$D6,$BD,$FB,$F0,$18,$F7,$73,$F0,$FA,$E1,$DF,$C3,$F0,$F8,$8F,$1F
	.dc.b	$2F,$F2,$AF,$1B,$E4,$F5,$7C,$CE,$9F,$B8,$DE,$9C,$D7,$E1,$B1,$F9
	.dc.b	$AD,$9D,$3F,$5D,$BD,$3F,$5E,$34,$43,$EC,$1B,$31,$7E,$06,$DF,$E6
	.dc.b	$30,$F0,$C7,$0A,$E8,$DC,$3C,$6D,$C3,$C3,$18,$C3,$A4,$E4,$EA,$78
	.dc.b	$4C,$11,$D9,$30,$F0,$B6,$38,$3F,$B4,$EA,$FF,$61,$70,$3E,$07,$EA
	.dc.b	$92,$BF,$8C,$46,$5B,$19,$7F,$82,$46,$5B,$19,$AF,$F0,$14,$E5,$B1
	.dc.b	$9B,$57,$F8,$2A,$B5,$18,$C9,$18,$9D,$57,$16,$9E,$42,$1E,$4F,$85
	.dc.b	$FC,$7F,$C4,$7E,$43,$E2,$3F,$23,$B7,$E4,$BE,$23,$F2,$7F,$07,$F9
	.dc.b	$4F,$87,$FC,$AF,$C3,$7E,$5B,$E1,$FF,$2F,$F0,$FF,$98,$F5,$BF,$33
	.dc.b	$EC,$7E,$6B,$E1,$FF,$C0,$DF,$5B,$FE,$6D,$E0,$FE,$4C,$FF,$9B,$5F
	.dc.b	$E8,$0B,$E1,$FD,$1C,$9E,$CE,$3F,$F3,$A7,$3F,$F1,$DB,$FB,$B3,$F8
	.dc.b	$6E,$7F,$B0,$DE,$7F,$B6,$49,$FB,$6D,$8D,$46,$08,$E0,$9D,$1F,$8D
	.dc.b	$9E,$0D,$9D,$BF,$94,$BF,$C2,$E1,$E8,$E8,$FA,$58,$7A,$6F,$E7,$E3
	.dc.b	$F7,$1B,$8F,$D9,$AD,$92,$F9,$12,$76,$7D,$A3,$F1,$26,$2F,$B3,$F6
	.dc.b	$CF,$C9,$F3,$FD,$B7,$F2,$BC,$DF,$6D,$FC,$BF,$2F,$DB,$7F,$33,$C9
	.dc.b	$F6,$DF,$CD,$F1,$FD,$B7,$F3,$BC,$5F,$6D,$FA,$E7,$DC,$F8,$8F,$BB
	.dc.b	$C3,$E7,$8F,$EF,$F8,$FE,$57,$89,$F1,$1F,$82,$24,$FE,$01,$AF,$E1
	.dc.b	$1B,$C3,$FE,$65,$9E,$33,$B2,$1B,$2C,$6E,$78,$F0,$F8,$5B,$3F,$E7
	.dc.b	$2A,$E6,$F1,$25,$77,$F8,$8F,$CB,$FC,$47,$E6,$3E,$23,$F3,$3F,$11
	.dc.b	$F9,$AF,$88,$FC,$DF,$C4,$7E,$73,$E2,$26,$5F,$9A,$F8,$8F,$CF,$FC
	.dc.b	$47,$E8,$3E,$22,$4F,$E7,$BE,$23,$F4,$7F,$03,$FA,$4F,$81,$FD,$2F
	.dc.b	$C3,$F9,$BB,$5C,$7E,$6F,$01,$CF,$F2,$FE,$0B,$F5,$64,$CF,$F5,$2D
	.dc.b	$7F,$5A,$DE,$6F,$94,$DE,$8E,$53,$C0,$87,$02,$CF,$06,$C7,$07,$44
	.dc.b	$8F,$AD,$F0,$FF,$22,$3F,$77,$F0,$7F,$C0,$3E,$1F,$FD,$0D,$F5,$BF
	.dc.b	$E9,$0B,$FF,$50,$2F,$85,$F9,$37,$FC,$13,$9B,$F8,$3F,$37,$CA,$8F
	.dc.b	$E0,$BC,$BF,$29,$FF,$17,$5F,$FA,$55,$FE,$13,$C9,$F2,$DF,$E1,$7C
	.dc.b	$7E,$F7,$F0,$1B,$FF,$0A,$E0,$FC,$24,$7E,$23,$82,$BF,$82,$6E,$BF
	.dc.b	$8B,$8D,$1F,$89,$ED,$FA,$DC,$7F,$8A,$9D,$4F,$13,$47,$0E,$4F,$B7
	.dc.b	$5C,$09,$1F,$C8,$3B,$7E,$2F,$C2,$7E,$3E,$3F,$91,$F6,$7E,$3D,$FE
	.dc.b	$4B,$D7,$F2,$BE,$13,$E5,$FC,$27,$CC,$F8,$8F,$9B,$F0,$BF,$3B,$E2
	.dc.b	$3E,$7F,$C3,$7F,$30,$F0,$3E,$8F,$AD,$F4,$BE,$1B,$E9,$FC,$37,$8B
	.dc.b	$5F,$DB,$9F,$18,$FB,$E5,$3E,$F7,$2F,$E0,$79,$3F,$09,$18,$34,$E0
	.dc.b	$91,$B2,$46,$9B,$C9,$F9,$EE,$7F,$AC,$D9,$F0,$B5,$FE,$8F,$E0,$FE
	.dc.b	$35,$FE,$93,$E0,$FE,$35,$FF,$44,$76,$7E,$37,$0F,$EC,$8F,$E3,$63
	.dc.b	$FA,$6F,$7F,$E3,$B2,$FE,$19,$DF,$F8,$D9,$FE,$A7,$DF,$F8,$D9,$FC
	.dc.b	$AF,$C3,$FE,$DA,$7F,$AA,$F8,$9F,$B6,$7C,$BD,$CD,$FF,$19,$1F,$80
	.dc.b	$3F,$05,$BF,$E0,$1F,$C1,$4F,$E1,$F5,$FC,$69,$23,$86,$3F,$8A,$24
	.dc.b	$66,$B9,$4F,$4A,$E4,$F0,$9D,$46,$27,$75,$CA,$BF,$90,$76,$FC,$98
	.dc.b	$FE,$59,$E3,$FF,$1C,$9F,$EE,$3E,$3F,$F1,$C7,$E3,$90,$FE,$EF,$B7
	.dc.b	$C7,$AF,$E6,$9B,$7C,$7C,$FF,$9F,$73,$7C,$7B,$FE,$FB,$B7,$C7,$7D
	.dc.b	$1E,$07,$C6,$D7,$EE,$63,$F8,$69,$FC,$21,$23,$38,$7E,$1B,$6F,$9D
	.dc.b	$3F,$43,$7F,$C3,$C6,$23,$F2,$55,$C1,$E5,$19,$BE,$D3,$92,$46,$6F
	.dc.b	$FA,$B7,$27,$C5,$7F,$AC,$72,$7E,$37,$0F,$E1,$F1,$FC,$6B,$C4,$FC
	.dc.b	$53,$FD,$77,$93,$F8,$BC,$FF,$86,$BC,$7F,$88,$FF,$4F,$E0,$FE,$43
	.dc.b	$CB,$F1,$75,$FE,$B9,$C1,$F1,$5F,$2E,$BE,$BF,$1F,$F9,$FE,$1F,$74
	.dc.b	$9B,$FD,$CD,$3E,$FF,$0F,$EE,$70,$FB,$D3,$9A,$DE,$1F,$83,$25,$6E
	.dc.b	$F8,$CF,$E5,$23,$FC,$6B,$CD,$FC,$4A,$3F,$C6,$7C,$DF,$85,$7F,$C6
	.dc.b	$FC,$DF,$85,$7F,$98,$F8,$1F,$C4,$9F,$D8,$F7,$FF,$12,$7F,$67,$DD
	.dc.b	$FC,$49,$FE,$C0,$7E,$DB,$B7,$F8,$93,$FB,$8E,$CF,$E2,$5C,$1F,$03
	.dc.b	$7F,$AB,$D3,$F7,$89,$5F,$81,$21,$F8,$0E,$0F,$B4,$FE,$02,$36,$E5
	.dc.b	$FC,$3F,$4F,$DD,$8F,$F0,$17,$4F,$DD,$E3,$FE,$A7,$C7,$F7,$67,$F4
	.dc.b	$11,$FE,$11,$E8,$FB,$B1,$FE,$58,$E8,$FB,$B9,$FF,$76,$E8,$FB,$B3
	.dc.b	$FD,$48,$FE,$0E,$7F,$52,$3C,$FA,$FB,$7B,$1F,$0F,$FD,$90,$93,$FC
	.dc.b	$69,$AF,$F5,$E6,$E5,$FD,$A1,$B9,$FF,$1E,$6C,$B1,$5F,$68,$E8,$43
	.dc.b	$A1,$6F,$A3,$6F,$A3,$FC,$B3,$E1,$FF,$BC,$FC,$3F,$F7,$AF,$87,$FE
	.dc.b	$F7,$F0,$FF,$DF,$3E,$1F,$FB,$EF,$C3,$FF,$7E,$F8,$7F,$F4,$BE,$2C
	.dc.b	$D9,$FF,$A5,$FA,$FF,$DB,$93,$FE,$9C,$9C,$BE,$98,$7C,$37,$29,$CE
	.dc.b	$E9,$7C,$36,$70,$E9,$F0,$D7,$1E,$5E,$ED,$C3,$83,$70,$7C,$35,$F8
	.dc.b	$6C,$7B,$DF,$2A,$FE,$4F,$37,$F1,$7E,$6F,$95,$97,$C8,$26,$DF,$21
	.dc.b	$BA,$FC,$86,$9F,$4E,$78,$CE,$C9,$3B,$6C,$EE,$BC,$55,$C2,$B4,$49
	.dc.b	$D1,$F5,$CE,$DC,$DF,$59,$F9,$91,$F3,$FE,$1B,$E0,$7B,$FF,$07,$E1
	.dc.b	$BE,$BD,$68,$FA,$1E,$B7,$C5,$F8,$6F,$8D,$CB,$F1,$F9,$7E,$47,$2F
	.dc.b	$C9,$E5,$98,$71,$E1,$FA,$9F,$03,$F3,$7E,$1B,$E7,$7C,$17,$AB,$8F
	.dc.b	$FA,$07,$43,$E1,$BE,$91,$31,$E5,$19,$6C,$65,$FC,$57,$1F,$E2,$63
	.dc.b	$1C,$DF,$84,$AE,$15,$A3,$F1,$53,$AA,$E1,$1D,$08,$6F,$3F,$C6,$4E
	.dc.b	$E3,$89,$AB,$ED,$1F,$89,$F8,$3F,$C5,$7C,$37,$E8,$EB,$FC,$23,$DF
	.dc.b	$FA,$37,$D7,$F0,$9F,$A3,$7D,$BF,$0F,$F9,$13,$F2,$5F,$0F,$F9,$5F
	.dc.b	$AB,$FF,$72,$67,$F9,$E7,$FD,$C9,$5E,$57,$FD,$C8,$FC,$6F,$03,$F3
	.dc.b	$0F,$C8,$EF,$F4,$6D,$EA,$81,$FE,$3C,$7F,$36,$FE,$1D,$AF,$E5,$9B
	.dc.b	$7F,$92,$6E,$FF,$8F,$6C,$FE,$40,$E1,$DF,$F9,$57,$15,$C6,$31,$3C
	.dc.b	$08,$E0,$9D,$FF,$97,$AF,$A5,$C4,$EF,$FC,$BC,$7E,$13,$BB,$F2,$EE
	.dc.b	$BB,$BF,$2C,$FE,$23,$B7,$F2,$CF,$E2,$BB,$3F,$2A,$F4,$EC,$FE,$01
	.dc.b	$D9,$F9,$58,$FC,$7F,$5F,$E5,$5F,$8F,$E3,$7F,$35,$EC,$C4,$7B,$9B
	.dc.b	$C3,$ED,$6F,$17,$C5,$E1,$F6,$46,$2B,$E1,$CE,$63,$FD,$43,$C3,$EF
	.dc.b	$BF,$F5,$0F,$8B,$FC,$AA,$FF,$CB,$3E,$2F,$BA,$7F,$D4,$BE,$27,$E1
	.dc.b	$E7,$DC,$7E,$9B,$C1,$FD,$9B,$FA,$8D,$FD,$F5,$FA,$3E,$8F,$C2,$CF
	.dc.b	$CC,$E8,$F5,$EF,$ED,$8F,$C1,$6F,$E8,$6E,$DE,$9E,$8F,$AE,$D9,$D4
	.dc.b	$FC,$33,$67,$2D,$FD,$9C,$1F,$C3,$E3,$F7,$7E,$1F,$F0,$F7,$FC,$EB
	.dc.b	$E2,$FF,$31,$E7,$FB,$31,$F8,$0E,$7F,$B2,$FE,$0B,$9F,$D7,$87,$F4
	.dc.b	$7E,$1F,$B3,$3F,$C1,$BC,$3F,$E1,$AF,$88,$FC,$F7,$5F,$F0,$2E,$3F
	.dc.b	$DB,$C7,$EE,$6F,$F8,$0B,$F8,$EF,$0F,$F4,$F5,$B9,$CD,$70,$AD,$1C
	.dc.b	$4F,$E0,$66,$08,$EC,$91,$C8,$91,$C2,$3F,$8B,$78,$7F,$C2,$23,$FD
	.dc.b	$2F,$D3,$F8,$17,$FD,$B7,$D3,$FD,$57,$C1,$FE,$11,$1F,$A4,$F0,$7F
	.dc.b	$04,$FE,$9B,$C1,$FE,$10,$FF,$09,$F8,$7F,$E1,$5F,$0F,$FD,$77,$B3
	.dc.b	$F0,$3A,$7F,$25,$EC,$F7,$F0,$7F,$5B,$EA,$FB,$84,$C7,$E5,$57,$E7
	.dc.b	$5B,$C1,$F9,$96,$9F,$9C,$27,$5F,$F8,$3B,$AB,$F0,$31,$FD,$E7,$C2
	.dc.b	$FD,$CB,$FB,$CF,$0B,$F3,$97,$FE,$01,$EA,$FC,$0E,$3F,$CE,$7A,$BF
	.dc.b	$02,$FF,$3B,$EA,$FC,$0E,$1F,$DE,$8F,$E8,$BE,$17,$EE,$5F,$E0,$FE
	.dc.b	$17,$E7,$30,$FE,$3F,$AB,$C3,$FF,$51,$13,$D7,$FE,$5C,$4B,$FE,$B1
	.dc.b	$19,$6C,$65,$FE,$60,$DC,$BF,$9A,$37,$4E,$F5,$C6,$B6,$72,$21,$83
	.dc.b	$F9,$D7,$AF,$FC,$FF,$E1,$FF,$A0,$7B,$1F,$D0,$7E,$0F,$FA,$17,$B3
	.dc.b	$FD,$0F,$E0,$E6,$1E,$6F,$4B,$FA,$37,$BF,$FD,$1F,$D3,$FE,$91,$E9
	.dc.b	$FF,$49,$F7,$FF,$A5,$7C,$34,$C3,$DA,$FF,$4D,$F7,$BD,$11,$FE,$45
	.dc.b	$E0,$71,$7F,$AC,$DF,$57,$1F,$BA,$7D,$2D,$7D,$4D,$27,$1F,$C2,$AE
	.dc.b	$BC,$9F,$45,$B7,$D2,$75,$7D,$1A,$F4,$6D,$FF,$61,$E4,$FE,$C9,$F0
	.dc.b	$DF,$D9,$7D,$AF,$EC,$DE,$D7,$C6,$C3,$EC,$F1,$FC,$67,$FB,$33,$FD
	.dc.b	$AF,$DA,$F8,$D1,$F7,$BE,$1F,$FC,$85,$F5,$7F,$E4,$E8,$FE,$D9,$C3
	.dc.b	$F1,$DA,$7E,$13,$87,$D3,$C1,$FE,$D7,$A7,$C1,$F2,$25,$7F,$88,$63
	.dc.b	$2D,$8C,$BF,$E0,$D8,$CB,$63,$3B,$FF,$97,$27,$2D,$8C,$F0,$7E,$72
	.dc.b	$72,$D8,$CD,$7F,$9F,$E7,$2D,$8C,$F1,$27,$2D,$F8,$3F,$C9,$57,$E8
	.dc.b	$BA,$FF,$24,$FA,$FE,$23,$E3,$FA,$5F,$23,$E2,$3F,$C2,$DD,$9F,$92
	.dc.b	$AF,$81,$F0,$BF,$92,$7F,$57,$D5,$F3,$BE,$23,$E7,$FC,$37,$D0,$F8
	.dc.b	$8F,$A3,$F0,$FE,$4E,$0F,$87,$EF,$7F,$AF,$FB,$FF,$0D,$97,$E5,$79
	.dc.b	$BE,$8B,$F8,$73,$05,$E3,$FC,$73,$67,$8C,$62,$3F,$18,$70,$2B,$9E
	.dc.b	$6F,$CC,$C6,$5F,$C8,$3F,$C1,$3B,$BF,$20,$FF,$06,$ED,$FC,$83,$F8
	.dc.b	$E3,$F2,$24,$3F,$C6,$DD,$9F,$90,$7F,$C7,$1D,$9F,$90,$EA,$FC,$76
	.dc.b	$1F,$E3,$DE,$CB,$1F,$C2,$BB,$3D,$F7,$FE,$37,$D3,$D6,$4D,$73,$87
	.dc.b	$A3,$5F,$56,$1F,$A8,$25,$FA,$DB,$1A,$8C,$11,$C1,$2F,$F9,$3F,$7F
	.dc.b	$E1,$27,$FC,$A1,$D7,$F8,$47,$FC,$A5,$E1,$FE,$BF,$1F,$F2,$B7,$7F
	.dc.b	$E1,$27,$F9,$5F,$87,$F0,$67,$FC,$B5,$E1,$7F,$15,$9F,$E2,$64,$3F
	.dc.b	$5E,$7F,$97,$76,$F8,$3C,$9F,$83,$DF,$DF,$BF,$AB,$6F,$85,$7F,$7E
	.dc.b	$72,$FA,$E3,$2F,$DE,$6E,$3F,$72,$3F,$CD,$5C,$9F,$71,$FF,$36,$72
	.dc.b	$7D,$C9,$F5,$C7,$DB,$6B,$F7,$0F,$E8,$FE,$1F,$F9,$DF,$93,$EE,$47
	.dc.b	$AE,$7E,$E3,$72,$FB,$8F,$9C,$FF,$26,$F8,$DF,$B9,$7F,$83,$10,$FE
	.dc.b	$08,$59,$FE,$04,$D9,$E0,$E8,$91,$A6,$C6,$A3,$04,$70,$48,$FE,$0C
	.dc.b	$75,$3F,$85,$78,$5F,$81,$9F,$F4,$97,$1F,$E0,$5F,$E2,$DE,$17,$F0
	.dc.b	$4A,$FE,$41,$E0,$FE,$05,$FF,$4B,$71,$FF,$2E,$F0,$3F,$82,$47,$FA
	.dc.b	$67,$C0,$FE,$08,$FF,$27,$EF,$FE,$08,$FF,$29,$EE,$FE,$09,$E1,$FF
	.dc.b	$53,$E5,$FB,$BD,$9F,$95,$6F,$37,$DC,$E1,$FC,$24,$60,$98,$7F,$03
	.dc.b	$EC,$FE,$92,$FF,$7E,$F1,$3F,$A4,$BF,$E0,$0E,$DF,$CD,$5F,$F7,$AE
	.dc.b	$DB,$3F,$D9,$7B,$FF,$35,$3F,$DF,$BA,$3E,$FC,$FE,$63,$97,$EF,$B6
	.dc.b	$7E,$FB,$F9,$1C,$FF,$A4,$7C,$47,$F5,$02,$47,$F4,$76,$BF,$D1,$5B
	.dc.b	$5F,$D0,$DB,$5F,$D0,$5B,$3F,$CF,$DB,$B7,$F3,$D6,$F8,$26,$98,$F8
	.dc.b	$8F,$EC,$3E,$87,$F6,$2F,$43,$FB,$1F,$C3,$FF,$64,$F6,$FF,$B2,$FC
	.dc.b	$27,$F6,$6F,$73,$FB,$3F,$C0,$7F,$68,$F8,$4F,$F6,$5E,$DF,$EE,$7F
	.dc.b	$BF,$7B,$D8,$E2,$71,$7F,$8C,$A5,$EB,$66,$31,$F0,$B7,$07,$22,$DE
	.dc.b	$BE,$63,$0D,$87,$C1,$D3,$DB,$5F,$AC,$3E,$21,$F3,$48,$7C,$D2,$E5
	.dc.b	$F3,$5B,$8F,$CD,$69,$C8,$E0,$48,$E0,$D8,$E0,$D3,$84,$68,$91,$F0
	.dc.b	$FD,$DF,$57,$C4,$7D,$1A,$FF,$04,$72,$7B,$7D,$EF,$A3,$1F,$EF,$8C
	.dc.b	$3E,$5B,$F1,$B2,$FF,$72,$3F,$E0,$FE,$1F,$A8,$EF,$2F,$A1,$F0,$BF
	.dc.b	$1F,$83,$E4,$7B,$BF,$27,$79,$7B,$3A,$79,$DF,$37,$E2,$3E,$5C,$3D
	.dc.b	$34,$FD,$BF,$7B,$E9,$12,$79,$72,$7D,$98,$C4,$7E,$15,$BB,$FA,$A7
	.dc.b	$53,$C3,$75,$6B,$6E,$D3,$98,$C3,$F0,$5B,$5C,$A3,$B1,$B2,$1F,$E2
	.dc.b	$FE,$DF,$CF,$BD,$FE,$1B,$F3,$EF,$F8,$DF,$B3,$F1,$9E,$FF,$E7,$E3
	.dc.b	$C9,$F0,$3F,$8F,$F8,$6F,$C8,$76,$7E,$47,$E2,$3F,$25,$DB,$2F,$D5
	.dc.b	$F0,$FF,$93,$7D,$9D,$DE,$6E,$1F,$6C,$7E,$57,$7F,$DB,$3F,$8E,$6B
	.dc.b	$F8,$96,$F1,$7E,$22,$3F,$17,$C5,$F8,$E8,$C4,$EA,$30,$49,$D9,$23
	.dc.b	$4D,$F1,$7F,$82,$3E,$16,$BF,$BB,$EF,$FC,$AB,$F9,$42,$1F,$BF,$ED
	.dc.b	$FC,$AB,$FC,$07,$E1,$3F,$82,$BF,$E6,$1F,$07,$F6,$DD,$DF,$91,$8F
	.dc.b	$BD,$DB,$F9,$07,$F8,$47,$6F,$EE,$FA,$FF,$21,$1E,$3F,$1B,$F9,$3B
	.dc.b	$48,$7C,$5D,$3C,$FB,$FC,$26,$EF,$F0,$77,$F7,$92,$33,$5F,$06,$FF
	.dc.b	$3B,$7C,$5F,$F4,$1F,$5F,$E1,$A7,$FC,$ED,$E0,$FE,$19,$FF,$3C,$75
	.dc.b	$FC,$0B,$FF,$3B,$78,$7F,$02,$7F,$CF,$9C,$1F,$86,$7F,$39,$E0,$FE
	.dc.b	$15,$F7,$9F,$9F,$DF,$F0,$CF,$F2,$9E,$6F,$D7,$F8,$1F,$AE,$8F,$31
	.dc.b	$F0,$31,$F3,$37,$83,$DD,$BF,$B3,$83,$DD,$C1,$EC,$9C,$B6,$BF,$05
	.dc.b	$CE,$E7,$FB,$B1,$EE,$3E,$EE,$3F,$B5,$97,$3F,$BF,$0F,$E7,$5C,$FE
	.dc.b	$FA,$FF,$4B,$70,$7D,$EC,$3D,$E7,$DE,$3D,$C4,$3F,$A3,$72,$FD,$E7
	.dc.b	$ED,$67,$FD,$2B,$6F,$81,$AF,$EE,$F0,$F1,$78,$1F,$A9,$25,$FE,$F2
	.dc.b	$32,$DC,$FF,$83,$BF,$C0,$89,$1A,$6C,$6A,$30,$47,$04,$8F,$F5,$57
	.dc.b	$37,$E0,$E7,$E2,$78,$1F,$83,$7F,$D6,$DC,$DF,$C2,$2B,$FD,$81,$CD
	.dc.b	$FD,$57,$BF,$F8,$44,$7F,$1D,$EE,$FE,$10,$FF,$1F,$ED,$FC,$26,$1F
	.dc.b	$3F,$B7,$F0,$6F,$D1,$E8,$F7,$D7,$F3,$1E,$2F,$B6,$FE,$04,$FC,$29
	.dc.b	$F2,$8C,$70,$7E,$09,$BD,$1F,$85,$27,$1F,$E0,$E7,$FB,$77,$47,$E0
	.dc.b	$B1,$FF,$68,$75,$FF,$53,$C3,$F7,$C7,$FB,$6F,$C4,$FE,$E1,$D1,$F8
	.dc.b	$1E,$5F,$F6,$97,$47,$E0,$6B,$FC,$05,$D9,$FD,$39,$FC,$06,$3F,$8B
	.dc.b	$EC,$FD,$F4,$7F,$3F,$F0,$BF,$3F,$C1,$FE,$05,$F8,$7F,$EF,$84,$9F
	.dc.b	$E8,$8D,$7F,$C1,$CD,$BF,$CB,$37,$2E,$53,$C0,$87,$02,$CF,$06,$C7
	.dc.b	$07,$44,$8F,$EA,$DF,$0B,$FE,$1E,$84,$61,$FF,$0A,$43,$E1,$EF,$C3
	.dc.b	$E3,$CB,$CF,$A3,$AA,$DF,$B7,$C3,$E1,$F8,$FA,$5C,$BD,$2E,$7F,$0D
	.dc.b	$D3,$DB,$EB,$EF,$F6,$F8,$3E,$FE,$E7,$BF,$83,$FC,$05,$C7,$FA,$9E
	.dc.b	$44,$7F,$8A,$B6,$7C,$27,$98,$91,$FD,$DA,$32,$D8,$CB,$EE,$E5,$E7
	.dc.b	$DC,$AD,$CE,$4E,$D5,$C3,$9B,$E3,$B6,$F8,$D7,$09,$CD,$77,$21,$CE
	.dc.b	$FB,$5F,$18,$EA,$6A,$7E,$83,$E9,$3E,$8F,$C2,$7D,$07,$D9,$F0,$9F
	.dc.b	$41,$F7,$61,$F4,$1F,$81,$ED,$FD,$8F,$84,$FB,$3E,$E7,$D0,$9F,$89
	.dc.b	$C7,$F7,$3D,$CF,$A1,$1F,$E1,$CF,$AC,$FF,$18,$BE,$7C,$7D,$9B,$FB
	.dc.b	$B6,$FA,$53,$F5,$0F,$C6,$94,$FC,$64,$7E,$3F,$6F,$C9,$46,$0D,$38
	.dc.b	$24,$6C,$91,$A6,$EF,$FE,$67,$E1,$FC,$23,$67,$C2,$D7,$E7,$62,$F7
	.dc.b	$FF,$CD,$8B,$E1,$7F,$37,$DF,$2B,$F0,$DD,$DA,$F8,$5D,$FC,$47,$0F
	.dc.b	$86,$FC,$D5,$7E,$2F,$BB,$F3,$4F,$C7,$F8,$6F,$CD,$3F,$8F,$ED,$EF
	.dc.b	$F1,$1F,$9A,$8F,$2F,$B6,$E1,$F9,$3C,$3F,$82,$9F,$C0,$71,$7C,$D6
	.dc.b	$9F,$84,$3E,$97,$27,$E1,$E7,$84,$60,$DB,$8E,$4F,$C8,$C6,$48,$E0
	.dc.b	$95,$DE,$32,$D8,$FC,$8C,$7E,$04,$FC,$89,$FA,$BF,$17,$F8,$1F,$67
	.dc.b	$E4,$63,$F0,$DC,$FF,$91,$7F,$11,$D9,$F9,$17,$F6,$5D,$7F,$91,$7F
	.dc.b	$19,$D7,$F9,$17,$F6,$DE,$07,$D6,$CF,$DD,$87,$EC,$BD,$AF,$EC,$FC
	.dc.b	$1F,$26,$FE,$5F,$07,$E0,$A3,$11,$F8,$DD,$FF,$05,$BF,$E3,$23,$F1
	.dc.b	$44,$8C,$E3,$F8,$CA,$E3,$5B,$9C,$D6,$E7,$35,$BB,$FE,$47,$30,$FE
	.dc.b	$1D,$D3,$FB,$57,$F5,$1D,$3F,$17,$1F,$E5,$B9,$FC,$57,$F9,$7F,$89
	.dc.b	$FC,$92,$BF,$8B,$78,$9F,$16,$7F,$99,$F8,$9F,$C9,$27,$E2,$9F,$C6
	.dc.b	$FC,$4F,$81,$9F,$E2,$B1,$FD,$AC,$FE,$1E,$FE,$EB,$74,$FB,$84,$70
	.dc.b	$4D,$BF,$11,$19,$6C,$66,$BA,$C6,$08,$E0,$9C,$DF,$62,$7E,$21,$FC
	.dc.b	$27,$C3,$F8,$98,$7F,$4E,$E6,$F8,$98,$FF,$28,$C3,$E2,$3F,$CA,$B9
	.dc.b	$BF,$11,$87,$F1,$1E,$6F,$89,$3F,$CB,$B9,$BF,$11,$3F,$C5,$79,$BF
	.dc.b	$10,$FE,$8B,$0F,$C4,$78,$9F,$5E,$3F,$0E,$43,$F0,$DC,$DF,$71,$FB
	.dc.b	$DC,$1F,$31,$D6,$BF,$83,$9D,$70,$7E,$10,$8E,$3C,$1F,$E1,$D5,$FC
	.dc.b	$A7,$C1,$FE,$18,$FF,$29,$E8,$FE,$CC,$FF,$2D,$F0,$7F,$86,$BF,$CB
	.dc.b	$FC,$0F,$E1,$AF,$F2,$FF,$03,$F8,$6B,$FC,$6B,$C0,$FE,$1A,$FF,$33
	.dc.b	$F0,$3F,$86,$BF,$83,$3D,$9D,$DF,$E4,$CE,$1F,$A7,$1F,$0F,$A7,$F5
	.dc.b	$64,$70,$43,$F8,$09,$F7,$DD,$38,$DF,$EF,$BF,$73,$83,$FC,$0F,$D3
	.dc.b	$F7,$39,$BF,$B0,$F4,$FD,$CA,$FE,$E9,$8F,$DC,$7F,$BB,$74,$FD,$CC
	.dc.b	$3F,$C1,$FD,$3F,$72,$7F,$BD,$74,$FD,$C9,$FF,$09,$F4,$FD,$C9,$F1
	.dc.b	$61,$F7,$16,$8E,$67,$C3,$FF,$3E,$24,$FF,$5B,$6B,$FD,$11,$B8,$7F
	.dc.b	$BD,$67,$8C,$EC,$86,$CB,$1B,$9E,$3A,$FF,$82,$2B,$8D,$6C,$95,$B7
	.dc.b	$8C,$EF,$E2,$3F,$A9,$FC,$47,$F5,$4F,$4A,$57,$F7,$1F,$73,$FA,$BF
	.dc.b	$C3,$C8,$FE,$E9,$E9,$FF,$5A,$F8,$7F,$EB,$7E,$EF,$F5,$CF,$56,$57
	.dc.b	$F7,$9F,$77,$FA,$FF,$C3,$7F,$60,$F5,$BF,$C4,$F3,$FE,$11,$FC,$01
	.dc.b	$EF,$7A,$A3,$E5,$F1,$FA,$BB,$BF,$D2,$B4,$F6,$38,$F2,$B9,$7F,$C4
	.dc.b	$F3,$9C,$3F,$C6,$73,$97,$EA,$C6,$5F,$05,$72,$AE,$2D,$8D,$B7,$9D
	.dc.b	$5C,$6F,$A4,$62,$75,$39,$BE,$77,$E2,$6B,$E0,$30,$76,$2D,$73,$6C
	.dc.b	$F3,$74,$71,$25,$F8,$A3,$53,$E6,$F7,$BC,$7E,$CF,$93,$DD,$F2,$FB
	.dc.b	$9F,$E1,$0E,$7F,$B5,$97,$F8,$53,$9F,$FC,$27,$CF,$F6,$E3,$C3,$F1
	.dc.b	$1F,$69,$FF,$0C,$F3,$7F,$86,$39,$BD,$FF,$07,$F6,$E7,$CF,$CB,$F0
	.dc.b	$B8,$E4,$7F,$99,$39,$3F,$D9,$BC,$BF,$94,$E4,$FC,$94,$FE,$53,$97
	.dc.b	$F3,$1D,$BF,$81,$E5,$FD,$53,$6F,$84,$EA,$BF,$32,$7E,$1A,$74,$47
	.dc.b	$35,$C9,$B5,$C9,$FC,$6F,$6F,$E8,$1F,$F2,$77,$67,$E8,$1F,$CA,$7B
	.dc.b	$DF,$73,$E1,$7E,$EF,$C2,$FD,$EF,$4B,$EF,$FC,$07,$E8,$2F,$EB,$FC
	.dc.b	$2F,$E0,$BE,$17,$F0,$7F,$11,$FA,$09,$FB,$7F,$0D,$F8,$6F,$88,$F3
	.dc.b	$47,$FB,$34,$FB,$BC,$7F,$6F,$97,$E4,$C7,$DE,$E4,$FC,$4C,$E6,$3F
	.dc.b	$0B,$3A,$8C,$46,$CE,$73,$C0,$8E,$09,$C7,$F8,$F7,$83,$67,$99,$F8
	.dc.b	$DF,$07,$F1,$73,$F9,$4F,$03,$F1,$6F,$F0,$53,$FC,$ED,$DD,$F8,$CA
	.dc.b	$FE,$35,$DD,$F8,$B9,$FD,$F7,$77,$E2,$E7,$F8,$F7,$77,$E2,$E7,$F3
	.dc.b	$FD,$BF,$8B,$7F,$82,$76,$FB,$75,$FD,$4F,$6F,$E2,$BB,$7E,$4E,$3F
	.dc.b	$8A,$24,$FE,$16,$FF,$18,$7E,$14,$FC,$0C,$F0,$8D,$4F,$09,$D1,$C3
	.dc.b	$7F,$C4,$37,$B7,$F1,$11,$FC,$F7,$B7,$F1,$13,$FE,$7C,$ED,$FC,$45
	.dc.b	$FF,$39,$EF,$F8,$58,$FF,$33,$E1,$FD,$6D,$FF,$A6,$FC,$6F,$C4,$3F
	.dc.b	$AC,$3F,$D3,$DC,$1F,$0D,$FF,$43,$F8,$BF,$0F,$1F,$DF,$70,$7E,$23
	.dc.b	$83,$F9,$1F,$37,$C5,$BF,$B7,$E2,$7F,$02,$24,$67,$83,$EE,$6B,$F8
	.dc.b	$26,$D7,$09,$D5,$70,$8F,$C6,$90,$FC,$2B,$D7,$9F,$F0,$71,$FC,$E7
	.dc.b	$9F,$F0,$6F,$FA,$7B,$83,$E0,$BF,$CE,$B8,$3E,$0B,$FE,$A2,$E0,$FE
	.dc.b	$4F,$CD,$F8,$3D,$3F,$AC,$70,$FC,$1E,$0F,$E3,$9C,$DF,$06,$7F,$9E
	.dc.b	$F0,$F9,$B7,$FF,$4B,$F2,$FD,$42,$73,$7D,$DD,$3E,$DE,$3F,$77,$8B
	.dc.b	$ED,$B6,$B7,$3F,$84,$21,$F7,$EB,$95,$7D,$F8,$FF,$69,$78,$FF,$7B
	.dc.b	$2F,$ED,$DC,$DF,$D0,$3C,$2F,$E0,$D5,$FC,$E3,$C1,$FE,$0C,$FF,$3A
	.dc.b	$F0,$3E,$FE,$9F,$CD,$FC,$0F,$E0,$D3,$FC,$FB,$BF,$F8,$33,$FD,$8F
	.dc.b	$BF,$F8,$33,$E4,$EB,$FB,$EF,$D3,$BF,$DA,$E9,$F6,$FA,$FE,$C4,$FD
	.dc.b	$6E,$2F,$8C,$EE,$75,$18,$8D,$19,$3F,$13,$D7,$F7,$2F,$FA,$67,$5F
	.dc.b	$DC,$7F,$39,$3F,$BE,$21,$FE,$DC,$EE,$FD,$F4,$7F,$53,$E9,$FB,$93
	.dc.b	$FD,$EB,$A7,$EE,$65,$FD,$DB,$A7,$EE,$4F,$F7,$4E,$9F,$26,$DF,$DF
	.dc.b	$BE,$23,$FB,$59,$27,$FB,$1B,$5F,$F0,$43,$72,$FE,$C4,$DC,$3F,$C0
	.dc.b	$6D,$9F,$F0,$0B,$7A,$3F,$BD,$B7,$80,$D3,$1F,$11,$FD,$F3,$DA,$97
	.dc.b	$FD,$DB,$EB,$7F,$C3,$2B,$2F,$F0,$1C,$3D,$9F,$F0,$42,$F8,$1B,$F0
	.dc.b	$38,$F8,$8C,$FA,$1A,$F8,$8D,$FC,$1F,$0F,$73,$8F,$C1,$F2,$F5,$B9
	.dc.b	$FC,$1F,$4F,$77,$CD,$C1,$FD,$B7,$D1,$F7,$BF,$18,$87,$C6,$2F,$0F
	.dc.b	$C6,$6F,$3F,$07,$44,$8D,$36,$35,$18,$23,$82,$47,$CE,$3A,$9F,$E6
	.dc.b	$4E,$17,$B1,$F1,$3E,$0F,$E2,$FB,$DF,$1B,$DE,$F8,$FE,$F7,$C8,$E6
	.dc.b	$F8,$F8,$FC,$FE,$5F,$8E,$FD,$1F,$7B,$E6,$FC,$47,$C7,$8F,$A9,$C7
	.dc.b	$F1,$DF,$AD,$EF,$FD,$2F,$88,$F8,$F1,$FD,$BF,$C0,$7B,$BF,$E4,$58
	.dc.b	$FB,$5C,$47,$FC,$F7,$19,$E3,$FF,$1F,$B6,$75,$5C,$9F,$F2,$34,$66
	.dc.b	$B8,$C6,$08,$FF,$8F,$A3,$53,$C6,$31,$3A,$7C,$67,$23,$24,$33,$1C
	.dc.b	$5B,$3D,$09,$3F,$E5,$4E,$FF,$CC,$CF,$F9,$BF,$BF,$F2,$AF,$E5,$BB
	.dc.b	$E3,$E5,$F8,$8F,$5F,$C2,$FB,$3E,$17,$F2,$D3,$FA,$1E,$FF,$7F,$C4
	.dc.b	$7C,$0F,$86,$F8,$3F,$0D,$F9,$6A,$FD,$37,$7F,$AF,$87,$E4,$EF,$FE
	.dc.b	$80,$F2,$3F,$7D,$5F,$86,$E2,$FD,$9B,$7C,$8F,$D9,$B7,$C8,$FB,$38
	.dc.b	$7E,$42,$3F,$1C,$7E,$54,$AE,$6B,$9C,$65,$B1,$9A,$EB,$3F,$2B,$C1
	.dc.b	$FC,$AB,$FB,$DF,$03,$F2,$AF,$EF,$FB,$FF,$2A,$FF,$01,$EE,$FC,$AB
	.dc.b	$C7,$BB,$F2,$AF,$3E,$EF,$CA,$BF,$C1,$FC,$5F,$DD,$BF,$C2,$7C,$5F
	.dc.b	$E0,$BD,$DF,$95,$8F,$71,$FB,$B3,$D7,$C3,$FA,$CD,$FD,$2D,$DB,$D3
	.dc.b	$BF,$AF,$83,$E2,$CE,$5B,$7C,$89,$7C,$9B,$7C,$AB,$A7,$67,$E1,$B0
	.dc.b	$FF,$46,$76,$7E,$1A,$FF,$90,$76,$7E,$D5,$FE,$45,$E4,$7F,$1B,$7F
	.dc.b	$92,$79,$1F,$86,$7E,$67,$5F,$ED,$5F,$9D,$E0,$FE,$D5,$FA,$1E,$3F
	.dc.b	$F1,$B7,$E0,$9F,$B5,$D7,$C8,$F9,$B7,$F5,$EF,$E9,$24,$E9,$B1,$A8
	.dc.b	$C1,$1C,$1F,$D0,$7A,$3E,$FD,$7E,$07,$A3,$EF,$BF,$83,$E8,$FB,$EF
	.dc.b	$E1,$7A,$3E,$FF,$0F,$F3,$EE,$8F,$BE,$FF,$40,$E8,$FB,$F5,$F8,$BE
	.dc.b	$8F,$BE,$FE,$37,$A3,$EF,$BF,$D2,$7C,$2F,$E2,$7D,$4F,$1F,$F8,$5F
	.dc.b	$87,$FB,$8A,$FD,$F6,$5F,$A8,$6F,$5E,$8F,$D6,$15,$C9,$FC,$1C,$FE
	.dc.b	$06,$E9,$C4,$73,$72,$70,$E3,$FE,$15,$19,$24,$FF,$BB,$3A,$FF,$08
	.dc.b	$FF,$BB,$BA,$FF,$09,$7F,$A9,$EB,$FE,$12,$FF,$12,$F0,$FF,$84,$BF
	.dc.b	$C7,$FC,$2F,$C2,$63,$FD,$9B,$AF,$F8,$4C,$FF,$23,$F0,$7F,$84,$BF
	.dc.b	$DB,$7A,$FF,$08,$F9,$89,$C1,$FC,$21,$FD,$E7,$5F,$DD,$7F,$13,$87
	.dc.b	$E8,$DB,$AF,$E8,$DA,$7E,$2C,$BB,$FE,$2F,$AB,$F0,$75,$FD,$EB,$AB
	.dc.b	$F0,$6F,$F7,$CE,$AF,$C1,$BF,$DF,$BA,$BF,$07,$8F,$F4,$2B,$75,$7E
	.dc.b	$12,$31,$D5,$F8,$4B,$FF,$02,$F5,$7E,$11,$FF,$03,$F5,$7E,$13,$5F
	.dc.b	$E2,$5D,$5F,$84,$EA,$FC,$1F,$C4,$7F,$4C,$24,$FF,$49,$6B,$FD,$01
	.dc.b	$B7,$FD,$59,$B9,$72,$9E,$04,$38,$16,$78,$36,$38,$3A,$24,$23,$F0
	.dc.b	$91,$FD,$9E,$3F,$83,$B6,$3F,$0C,$FF,$89,$CF,$C2,$C7,$C9,$7F,$84
	.dc.b	$9F,$DF,$4F,$E1,$27,$F8,$B7,$C9,$FF,$11,$BF,$CA,$4F,$EF,$A7,$F2
	.dc.b	$93,$F0,$C7,$F4,$E3,$D0,$7F,$78,$AF,$E9,$D7,$EA,$AF,$6C,$FA,$AF
	.dc.b	$04,$78,$1C,$B8,$7F,$C0,$B7,$F2,$A3,$27,$12,$D7,$42,$57,$3A,$E3
	.dc.b	$3C,$FB,$17,$CE,$5B,$FE,$42,$FF,$15,$C5,$F9,$07,$F1,$9C,$5F,$90
	.dc.b	$7F,$1D,$C5,$F9,$0C,$BF,$C5,$9C,$5F,$8F,$7F,$C5,$DC,$5F,$8F,$AF
	.dc.b	$CA,$71,$7E,$3D,$FC,$B7,$17,$E3,$F7,$FF,$2B,$F1,$78,$F8,$3C,$B8
	.dc.b	$FF,$51,$DF,$FC,$4B,$ED,$7F,$87,$09,$1F,$E1,$96,$BF,$E1,$06,$DF
	.dc.b	$F8,$99,$B8,$FF,$1C,$6C,$FF,$8B,$1B,$5D,$A3,$A1,$0E,$85,$AE,$8F
	.dc.b	$E3,$44,$39,$1D,$BF,$C5,$A7,$30,$ED,$FE,$2F,$7F,$E7,$6E,$DF,$C9
	.dc.b	$E5,$FC,$C3,$B7,$F8,$BE,$1F,$E2,$AE,$DF,$E2,$F3,$FB,$DE,$DF,$E2
	.dc.b	$F3,$FE,$2E,$ED,$FE,$2E,$FF,$5A,$3F,$24,$7A,$09,$D9,$FC,$F5,$FC
	.dc.b	$3D,$7E,$17,$83,$E7,$C7,$E2,$F7,$FC,$3B,$63,$3A,$7E,$32,$38,$92
	.dc.b	$77,$19,$6D,$75,$EA,$FB,$84,$AF,$F4,$C7,$95,$FE,$71,$7F,$D3,$5E
	.dc.b	$57,$F9,$C5,$FF,$4E,$79,$5F,$DF,$5F,$F1,$27,$89,$FD,$F5,$FF,$13
	.dc.b	$78,$9F,$E9,$CE,$AF,$C7,$47,$F9,$F7,$AB,$F1,$CF,$F9,$FF,$AB,$F1
	.dc.b	$DA,$7F,$25,$EA,$F1,$4F,$F9,$C7,$81,$F0,$1F,$E6,$82,$47,$F9,$95
	.dc.b	$AF,$F9,$71,$BC,$3F,$E7,$26,$DF,$F6,$86,$CF,$F9,$E1,$BC,$3F,$E7
	.dc.b	$E6,$DF,$82,$7B,$10,$EC,$7F,$9D,$7D,$BF,$F3,$D9,$FE,$7C,$F6,$FF
	.dc.b	$CF,$BE,$FF,$F9,$FB,$DF,$FF,$3F,$FA,$9F,$E8,$0F,$7F,$FD,$03,$F0
	.dc.b	$FF,$E8,$2F,$57,$FD,$07,$F0,$BF,$E8,$4F,$82,$FF,$42,$FC,$3F,$FA
	.dc.b	$37,$5F,$F3,$CF,$D6,$BD,$8C,$72,$7F,$2D,$9F,$CE,$CB,$DD,$CF,$2F
	.dc.b	$FA,$42,$1E,$EE,$2B,$6F,$BB,$97,$F9,$0C,$7C,$33,$DD,$87,$C9,$6D
	.dc.b	$FC,$96,$9B,$3E,$19,$23,$83,$63,$83,$4E,$11,$A2,$46,$9B,$1A,$8F
	.dc.b	$57,$C4,$7A,$79,$D8,$BD,$EC,$3F,$EB,$6B,$FF,$54,$72,$E3,$DE,$CF
	.dc.b	$BD,$AF,$84,$DF,$C4,$7D,$1B,$FF,$57,$72,$7D,$17,$FD,$61,$C7,$D3
	.dc.b	$E1,$7A,$FA,$FD,$BE,$17,$BF,$C2,$FC,$9B,$FE,$45,$B3,$E2,$3C,$64
	.dc.b	$8F,$E2,$5C,$3E,$88,$C4,$7D,$C6,$F2,$7D,$B9,$D4,$F0,$AE,$8D,$AD
	.dc.b	$F2,$7D,$D6,$C7,$E0,$9B,$5C,$A3,$B1,$C4,$87,$FB,$0B,$C0,$FC,$CB
	.dc.b	$FE,$CF,$F0,$3E,$5F,$C3,$7E,$66,$3F,$DA,$5E,$07,$E6,$5F,$F6,$A7
	.dc.b	$7F,$D0,$F8,$6F,$A3,$F1,$1F,$4B,$E0,$3E,$9F,$B1,$F5,$3C,$1F,$AB
	.dc.b	$E0,$FD,$6F,$06,$65,$E2,$E3,$F8,$1C,$7E,$E8,$FE,$01,$E0,$FC,$D7
	.dc.b	$F1,$1C,$5F,$B5,$7F,$22,$D7,$3E,$37,$ED,$DB,$3C,$6F,$F1,$75,$B6
	.dc.b	$D6,$DB,$3F,$8A,$F0,$BF,$17,$E1,$7E,$32,$BF,$77,$87,$D3,$F0,$7F
	.dc.b	$16,$FD,$5F,$03,$F1,$6F,$FB,$DF,$C4,$FD,$DB,$FE,$F9,$F1,$3F,$76
	.dc.b	$FD,$BE,$FF,$C5,$BF,$77,$BB,$F1,$6F,$B3,$1F,$57,$A3,$FD,$0F,$7F
	.dc.b	$CF,$FA,$FF,$9E,$25,$7E,$15,$AF,$F0,$66,$F0,$7C,$C6,$F7,$FF,$03
	.dc.b	$6C,$FF,$08,$6D,$76,$8E,$84,$3A,$16,$BA,$3F,$C8,$48,$1F,$CD,$60
	.dc.b	$ED,$F9,$33,$9C,$3F,$D3,$3E,$47,$F2,$48,$FD,$2D,$3C,$8F,$E4,$B1
	.dc.b	$F2,$CF,$DA,$9F,$2C,$FD,$A9,$FA,$5E,$8F,$C6,$BF,$A7,$E7,$F9,$6F
	.dc.b	$A0,$9A,$FC,$BE,$7F,$EF,$D1,$EB,$CB,$E2,$34,$F8,$85,$D7,$E2,$3C
	.dc.b	$F5,$F8,$2D,$9E,$2D,$38,$96,$78,$CF,$02,$4F,$06,$CF,$F3,$AE,$6F
	.dc.b	$C8,$BF,$C0,$74,$F9,$33,$FA,$6E,$5F,$C8,$CF,$F4,$4E,$5F,$C8,$BF
	.dc.b	$34,$FF,$0F,$D3,$97,$F2,$71,$9C,$92,$E5,$FC,$AC,$FF,$4E,$E5,$FC
	.dc.b	$AB,$FD,$43,$97,$F9,$87,$D6,$FF,$53,$8F,$66,$FF,$29,$70,$7E,$5E
	.dc.b	$5E,$C7,$02,$4F,$F2,$28,$CB,$63,$2F,$96,$32,$D8,$CE,$FF,$C9,$67
	.dc.b	$2D,$8C,$F0,$AB,$FC,$7F,$5A,$8C,$6A,$8C,$4E,$AB,$8B,$4F,$21,$0F
	.dc.b	$E4,$C7,$E7,$AF,$F3,$BC,$DF,$C9,$A3,$F3,$3C,$DF,$9B,$7E,$76,$FF
	.dc.b	$9B,$7F,$93,$9F,$CE,$FB,$BF,$93,$BF,$CF,$7B,$7F,$93,$BF,$9B,$3F
	.dc.b	$A0,$76,$7F,$26,$7F,$C0,$3E,$27,$9B,$5F,$56,$9F,$DE,$1F,$C6,$ED
	.dc.b	$F8,$B2,$57,$E4,$B5,$FC,$6C,$EA,$31,$3A,$8C,$13,$83,$EE,$57,$06
	.dc.b	$CE,$DB,$5C,$89,$5C,$7A,$FF,$95,$47,$F5,$9E,$CF,$E5,$4F,$F7,$43
	.dc.b	$F9,$51,$FD,$77,$C9,$FE,$E8,$FF,$5F,$F2,$7F,$BA,$3F,$2B,$C9,$FE
	.dc.b	$E8,$FC,$CF,$27,$FB,$A3,$FD,$97,$C9,$FE,$E8,$FE,$34,$FE,$37,$9F
	.dc.b	$E5,$79,$FF,$29,$EE,$7E,$68,$93,$F9,$36,$BF,$91,$6E,$FF,$8F,$6E
	.dc.b	$FF,$8A,$E8,$46,$2B,$F1,$ED,$E0,$FC,$73,$67,$73,$DC,$87,$73,$F1
	.dc.b	$95,$F9,$6E,$67,$93,$FE,$4B,$9F,$90,$7E,$33,$A3,$F1,$D9,$7F,$2E
	.dc.b	$F0,$FF,$93,$CF,$E4,$3B,$BF,$7F,$3F,$CC,$BB,$BF,$7A,$FF,$6E,$F0
	.dc.b	$FF,$93,$D7,$F9,$33,$C3,$FD,$EB,$E7,$DB,$F1,$9B,$7A,$23,$CC,$4B
	.dc.b	$F8,$8F,$A9,$AE,$76,$F8,$D5,$CE,$35,$18,$69,$F2,$0B,$BF,$C8,$6D
	.dc.b	$F8,$6F,$A9,$FE,$06,$F0,$FF,$92,$CF,$F9,$93,$83,$F9,$2E,$1F,$E5
	.dc.b	$FF,$0F,$F9,$2D,$FF,$83,$3C,$3F,$E4,$AF,$F5,$4D,$FE,$2D,$7C,$FD
	.dc.b	$FE,$2B,$FE,$67,$F0,$FF,$90,$D7,$F8,$53,$C3,$FE,$43,$E8,$FA,$F4
	.dc.b	$FE,$3F,$D1,$F8,$3F,$73,$F9,$29,$23,$F8,$F3,$5F,$E3,$8D,$BF,$E3
	.dc.b	$4D,$C7,$94,$F0,$21,$C0,$B3,$C1,$B1,$C1,$D1,$23,$FB,$8F,$3F,$F8
	.dc.b	$F2,$3F,$89,$F8,$BF,$C5,$DF,$C5,$1F,$E3,$C3,$FB,$8F,$3F,$E2,$AB
	.dc.b	$FC,$67,$E2,$FF,$16,$9F,$F4,$8F,$8B,$FC,$59,$FF,$49,$F4,$FE,$2A
	.dc.b	$BF,$C7,$3D,$3F,$8A,$9F,$29,$FC,$CB,$4F,$C5,$F4,$FC,$9A,$FB,$7A
	.dc.b	$FE,$0E,$BF,$10,$7E,$0C,$BD,$3F,$83,$6F,$4F,$E0,$E7,$85,$72,$9E
	.dc.b	$0D,$AE,$0D,$AE,$13,$AA,$FF,$45,$BF,$C5,$4F,$ED,$C7,$EB,$A7,$FB
	.dc.b	$73,$FC,$54,$FF,$1F,$1F,$C5,$4F,$F0,$A1,$FC,$54,$FF,$0A,$1F,$C5
	.dc.b	$4F,$F1,$F1,$FC,$54,$FF,$1F,$1F,$D8,$78,$7F,$15,$D7,$FD,$9F,$D2
	.dc.b	$FE,$DA,$49,$FE,$CC,$D7,$FB,$23,$6B,$FB,$13,$6F,$F8,$A3,$67,$FB
	.dc.b	$0B,$71,$FE,$C0,$DC,$4D,$24,$F8,$08,$78,$0B,$7E,$06,$DF,$81,$A7
	.dc.b	$80,$B7,$E0,$6D,$F8,$1B,$7E,$07,$F5,$67,$13,$C0,$4B,$F0,$36,$FC
	.dc.b	$0D,$38,$46,$89,$1A,$6C,$6A,$30,$47,$04,$BF,$43,$D5,$B3,$C9,$B3
	.dc.b	$C5,$B3,$C6,$FD,$CF,$16,$D7,$16,$D7,$16,$CF,$1B,$F8,$95,$F0,$30
	.dc.b	$FF,$16,$7C,$07,$F8,$C8,$95,$FE,$2B,$6B,$FE,$29,$6C,$FF,$89,$DB
	.dc.b	$87,$F5,$96,$CF,$F8,$99,$B8,$FF,$89,$5B,$89,$A6,$3D,$9F,$F2,$1F
	.dc.b	$A7,$FE,$44,$F8,$2F,$F2,$2F,$A9,$FE,$46,$F8,$2F,$F2,$3F,$C2,$7F
	.dc.b	$92,$3D,$AF,$F2,$4F,$C1,$7F,$92,$BD,$BF,$F2,$5F,$C1,$7F,$93,$3D
	.dc.b	$CF,$F2,$6F,$B9,$FE,$4E,$F8,$8F,$F2,$7F,$C2,$FF,$94,$3D,$EF,$F8
	.dc.b	$01,$F5,$AE,$6F,$1F,$37,$F8,$FF,$A3,$F5,$4F,$E3,$08,$7D,$1E,$9F
	.dc.b	$97,$3F,$3D,$B1,$F3,$CD,$F4,$FD,$58,$CF,$4F,$DC,$27,$4F,$D6,$6C
	.dc.b	$6A,$30,$47,$04,$AF,$37,$89,$E1,$7B,$E1,$5F,$31,$83,$C8,$60,$98
	.dc.b	$79,$5D,$61,$E2,$6E,$1E,$16,$E1,$E1,$78,$DF,$7B,$FC,$66,$1F,$E6
	.dc.b	$4C,$9F,$07,$F3,$89,$7F,$E6,$BA,$FA,$0B,$25,$70,$8D,$61,$CE,$F8
	.dc.b	$C7,$FA,$3E,$74,$D8,$CD,$F4,$BE,$71,$8B,$E5,$5D,$0F,$05,$F2,$BE
	.dc.b	$0D,$9D,$CF,$53,$85,$F4,$6C,$ED,$E7,$7F,$8C,$F8,$9F,$C6,$FC,$27
	.dc.b	$E3,$BE,$27,$F1,$FF,$0B,$F9,$0F,$89,$FC,$8F,$C3,$7E,$4B,$E2,$3F
	.dc.b	$27,$F0,$1F,$94,$F3,$7F,$2B,$E6,$FE,$5B,$E1,$FF,$2F,$EB,$FE,$63
	.dc.b	$E1,$7F,$33,$E8,$7E,$6B,$E1,$7F,$37,$EA,$FB,$F8,$FC,$67,$E1,$8F
	.dc.b	$C0,$F3,$FD,$F7,$EC,$98,$E7,$FB,$44,$70,$4E,$7F,$C2,$3C,$1B,$3C
	.dc.b	$C9,$3C,$63,$2D,$EA,$FC,$55,$71,$7F,$8D,$F5,$BA,$BE,$7E,$1F,$4B
	.dc.b	$A7,$D5,$8F,$CF,$C3,$E9,$12,$74,$FD,$29,$DF,$57,$D3,$AE,$15,$AA
	.dc.b	$E1,$97,$62,$3F,$18,$C9,$63,$81,$F8,$4E,$AF,$B7,$1F,$80,$E9,$FB
	.dc.b	$6F,$E0,$BA,$3E,$DB,$F8,$4E,$7F,$B6,$FE,$1B,$9B,$ED,$BF,$88,$E5
	.dc.b	$FB,$6F,$E2,$B9,$3E,$DB,$F8,$CE,$3F,$B6,$FB,$48,$7F,$5C,$D7,$F4
	.dc.b	$FE,$E7,$F1,$A2,$47,$E5,$1A,$FE,$9D,$BA,$FE,$79,$BB,$7E,$DD,$B3
	.dc.b	$FA,$46,$D7,$68,$E8,$43,$A1,$6B,$A3,$FA,$4F,$5B,$F9,$69,$FC,$BB
	.dc.b	$DC,$FF,$75,$42,$7F,$6D,$D8,$F4,$F3,$18,$F8,$3B,$9F,$FB,$BA,$9E
	.dc.b	$A6,$30,$FE,$11,$0F,$73,$1A,$21,$F0,$18,$5E,$ED,$31,$EE,$E7,$E0
	.dc.b	$35,$EF,$6F,$E0,$38,$7B,$FC,$7B,$FE,$1F,$73,$9B,$FD,$CD,$9B,$DF
	.dc.b	$EC,$4E,$0F,$E0,$F1,$96,$C6,$5F,$85,$C5,$ED,$8C,$72,$7C,$46,$C6
	.dc.b	$67,$89,$0E,$27,$C2,$8D,$B4,$D4,$F3,$6C,$FF,$35,$E7,$FA,$B1,$E6
	.dc.b	$F8,$3F,$8D,$F0,$FF,$1F,$9F,$E4,$7C,$3F,$C9,$F8,$2F,$95,$F0,$FF
	.dc.b	$27,$0F,$6F,$35,$7C,$7D,$1F,$47,$9B,$E5,$C7,$C2,$F8,$3F,$A5,$F0
	.dc.b	$FF,$4F,$E1,$BC,$F7,$F5,$F8,$BF,$0B,$E4,$FF,$03,$C3,$F2,$47,$E3
	.dc.b	$78,$BF,$2C,$D3,$27,$E4,$89,$C7,$F9,$26,$C7,$06,$9C,$23,$44,$8D
	.dc.b	$36,$35,$18,$23,$FD,$5F,$EB,$7F,$36,$FC,$05,$F1,$1F,$9F,$8F,$C0
	.dc.b	$F8,$3F,$94,$9F,$F4,$3E,$5F,$95,$F0,$7F,$29,$E0,$FE,$57,$C1,$D7
	.dc.b	$C2,$EF,$E1,$FF,$2B,$97,$F5,$0C,$7F,$C0,$C7,$F8,$A5,$FC,$6F,$83
	.dc.b	$F9,$67,$F1,$FE,$07,$E5,$BB,$FF,$2B,$C5,$F3,$BB,$FF,$25,$5F,$59
	.dc.b	$BC,$5F,$85,$8C,$4E,$C8,$FE,$20,$FC,$61,$A3,$1B,$FD,$96,$D7,$4E
	.dc.b	$2F,$CC,$46,$CF,$EF,$1D,$FF,$90,$7C,$5D,$FF,$90,$7F,$4F,$DD,$F9
	.dc.b	$07,$F5,$3D,$BF,$90,$7F,$1E,$43,$D5,$D9,$FB,$97,$F5,$FE,$17,$EF
	.dc.b	$FB,$3F,$1F,$1F,$B2,$EB,$F5,$BF,$CA,$7C,$1F,$DD,$3F,$8B,$DB,$F8
	.dc.b	$F3,$F8,$66,$F0,$7E,$0D,$BB,$7E,$09,$B3,$F8,$62,$38,$25,$76,$8E
	.dc.b	$8D,$AE,$8D,$AE,$8B,$6F,$8D,$3F,$E1,$4F,$13,$F8,$D7,$5F,$ED,$70
	.dc.b	$FE,$1B,$D7,$F8,$A7,$F8,$7F,$57,$EC,$EF,$F9,$6E,$FF,$8A,$9F,$C0
	.dc.b	$77,$FE,$D1,$FC,$17,$47,$E2,$5F,$C2,$74,$7E,$25,$FF,$04,$6F,$F7
	.dc.b	$5F,$B5,$87,$E1,$C9,$BF,$DE,$6C,$7E,$19,$B3,$C5,$BC,$DF,$85,$9D
	.dc.b	$57,$12,$38,$25,$71,$8C,$B6,$3F,$C7,$3B,$FC,$69,$FE,$8D,$CD,$F8
	.dc.b	$B9,$FC,$F7,$37,$E2,$DF,$D0,$73,$7E,$2D,$FC,$49,$F8,$BC,$BE,$3B
	.dc.b	$FC,$E3,$5F,$8D,$1F,$D5,$B9,$7F,$17,$3F,$A9,$E5,$FC,$5F,$77,$F1
	.dc.b	$7D,$7E,$FF,$4F,$C9,$AF,$BE,$DC,$3E,$FB,$75,$FB,$ED,$9F,$C4,$F3
	.dc.b	$FE,$0D,$FC,$5F,$57,$F0,$EB,$FF,$2E,$3F,$C3,$A3,$FC,$A3,$D1,$F8
	.dc.b	$47,$F8,$37,$C3,$FF,$5F,$E8,$FC,$24,$7E,$FF,$A3,$F8,$6B,$FC,$2F
	.dc.b	$B3,$F0,$98,$F4,$F5,$FF,$8B,$7C,$3F,$F6,$1A,$FE,$B3,$E3,$7F,$61
	.dc.b	$7F,$4D,$AF,$E6,$F8,$7E,$07,$4F,$DB,$E2,$FA,$84,$70,$4E,$FF,$D8
	.dc.b	$E7,$F8,$26,$F1,$FD,$C9,$D6,$3F,$C2,$A7,$2F,$DE,$3F,$C1,$87,$F5
	.dc.b	$73,$FB,$6F,$8D,$FA,$37,$F8,$E7,$8D,$FD,$5D,$FE,$E3,$E3,$7E,$8F
	.dc.b	$0F,$F0,$57,$81,$67,$FC,$13,$D5,$F7,$A7,$FC,$1D,$D5,$F7,$B4,$FF
	.dc.b	$0A,$F5,$7D,$EE,$F6,$1F,$D0,$7E,$1F,$F9,$E1,$1F,$E6,$CD,$7F,$A2
	.dc.b	$36,$BE,$CB,$70,$E7,$C0,$EB,$4E,$6F,$B1,$D0,$87,$42,$DF,$46,$DF
	.dc.b	$4A,$E7,$1F,$D2,$7E,$1F,$FA,$97,$C3,$FF,$53,$F8,$7F,$EA,$9F,$0F
	.dc.b	$FD,$57,$E1,$FF,$AB,$7C,$3F,$F5,$7F,$87,$FE,$B1,$F0,$FF,$D6,$7E
	.dc.b	$1F,$FA,$D7,$C3,$FF,$5B,$F8,$79,$AF,$F5,$5F,$87,$FE,$BD,$F0,$FF
	.dc.b	$D7,$FE,$1F,$FC,$67,$F5,$BF,$EB,$EA,$F4,$13,$C1,$FD,$44,$7F,$9D
	.dc.b	$C9,$F0,$1C,$09,$3F,$EB,$08,$CB,$63,$2F,$F8,$42,$32,$D8,$CF,$37
	.dc.b	$FB,$0E,$72,$D8,$CF,$37,$D2,$9C,$BF,$27,$97,$D3,$F0,$1F,$1F,$0F
	.dc.b	$EE,$5C,$9F,$1D,$FE,$E9,$C7,$EE,$F8,$8F,$7F,$C0,$FC,$79,$FE,$F1
	.dc.b	$C5,$F1,$9F,$F0,$37,$17,$C4,$F1,$7E,$2F,$C0,$FC,$79,$F7,$7C,$47
	.dc.b	$91,$FF,$72,$6F,$FE,$DE,$E2,$FE,$1D,$D9,$F7,$27,$E5,$F1,$FE,$05
	.dc.b	$BC,$5F,$81,$ED,$FC,$57,$77,$E3,$5B,$7C,$6F,$81,$2B,$85,$FE,$28
	.dc.b	$FC,$89,$D0,$B1,$B6,$C7,$07,$E9,$7C,$17,$E3,$9F,$B1,$F1,$1F,$82
	.dc.b	$F8,$3F,$C1,$FC,$47,$E1,$3E,$13,$F1,$D5,$FE,$1D,$EF,$FC,$3F,$C4
	.dc.b	$7E,$23,$E1,$7F,$13,$F1,$1F,$8A,$F8,$6F,$F1,$57,$85,$F8,$EC,$3F
	.dc.b	$09,$F1,$1F,$8E,$DF,$DB,$C5,$FE,$4A,$F2,$7E,$E7,$85,$F2,$F9,$5C
	.dc.b	$FF,$7B,$1F,$9B,$39,$8C,$4E,$A3,$04,$9F,$C8,$10,$FC,$71,$79,$7F
	.dc.b	$18,$DC,$7F,$14,$FE,$67,$C1,$FC,$53,$FE,$47,$F0,$7F,$14,$FF,$92
	.dc.b	$7C,$1F,$C5,$3F,$9F,$F0,$3F,$14,$FF,$93,$7C,$0F,$C5,$63,$FD,$0B
	.dc.b	$C0,$FC,$54,$FE,$97,$BF,$F1,$4F,$E8,$FB,$FD,$0F,$FB,$1F,$BB,$F0
	.dc.b	$1A,$FD,$7C,$3E,$D6,$FF,$85,$AF,$C0,$35,$FC,$03,$63,$75,$CA,$7F
	.dc.b	$02,$49,$D5,$75,$24,$66,$78,$4E,$A3,$1E,$17,$E1,$DF,$F3,$57,$91
	.dc.b	$FC,$51,$FF,$30,$79,$1F,$C5,$1F,$F3,$17,$91,$FC,$51,$FF,$32,$79
	.dc.b	$1F,$C5,$1F,$F3,$36,$DF,$1B,$3F,$E4,$5D,$3F,$87,$9F,$F3,$B7,$8F
	.dc.b	$F8,$77,$DF,$AF,$EA,$FC,$5F,$E2,$7C,$FF,$23,$83,$ED,$D7,$DE,$23
	.dc.b	$82,$63,$F8,$E8,$FC,$01,$F8,$C2,$F0,$7E,$2D,$BD,$1F,$8E,$25,$FF
	.dc.b	$A1,$78,$7F,$87,$CF,$F9,$FF,$87,$E1,$3F,$E8,$DE,$1F,$84,$FF,$1F
	.dc.b	$F1,$FF,$87,$BF,$08,$FE,$1E,$7C,$22,$1F,$C9,$7C,$7F,$E1,$EF,$FA
	.dc.b	$2F,$7F,$E6,$BC,$9F,$86,$DF,$F8,$6F,$67,$EA,$AB,$E2,$B5,$FC,$1F
	.dc.b	$0F,$DA,$24,$67,$87,$F0,$AD,$E1,$FC,$0F,$0F,$DF,$3F,$63,$E3,$FD
	.dc.b	$FA,$FE,$69,$E3,$FF,$05,$C7,$F9,$F7,$8D,$FC,$15,$FB,$E7,$F0,$53
	.dc.b	$FA,$17,$89,$FC,$15,$FE,$89,$E1,$FF,$05,$7F,$A9,$1F,$EA,$7F,$23
	.dc.b	$FA,$8F,$87,$F7,$E3,$D3,$AF,$F0,$57,$EB,$1F,$73,$5F,$B4,$D3,$39
	.dc.b	$FD,$E6,$C6,$A3,$1A,$FD,$D9,$E0,$D8,$CF,$47,$DE,$9F,$C0,$F4,$7F
	.dc.b	$13,$E0,$74,$7E,$02,$7F,$42,$7F,$4E,$3F,$42,$7F,$4E,$3F,$B1,$73
	.dc.b	$7E,$01,$FF,$68,$75,$FE,$85,$FE,$9A,$7E,$03,$1F,$EC,$1C,$FF,$80
	.dc.b	$8F,$D0,$9F,$D3,$8F,$19,$FE,$7E,$CB,$FC,$A1,$9F,$F2,$6F,$6F,$FB
	.dc.b	$C9,$27,$FB,$93,$5F,$EE,$0D,$BF,$ED,$CD,$D7,$FB,$63,$67,$FA,$DB
	.dc.b	$6B,$B4,$74,$21,$D0,$B5,$D1,$B5,$D1,$A7,$42,$CF,$47,$91,$27,$93
	.dc.b	$67,$93,$4E,$45,$9E,$53,$C0,$E3,$CD,$FE,$18,$BF,$56,$1F,$20,$8E
	.dc.b	$09,$7E,$48,$CB,$63,$2F,$78,$CB,$63,$35,$E1,$C3,$CE,$D9,$F3,$B7
	.dc.b	$0F,$3C,$E8,$93,$A6,$C6,$A3,$04,$7E,$46,$1F,$E3,$4F,$7F,$FC,$72
	.dc.b	$49,$FF,$19,$35,$FF,$18,$36,$FF,$C5,$CD,$BF,$F1,$63,$67,$FC,$36
	.dc.b	$DB,$FE,$E6,$DC,$7F,$DC,$13,$D8,$87,$63,$FC,$85,$EE,$7F,$9C,$96
	.dc.b	$0C,$3F,$CD,$7F,$5A,$F8,$2D,$4E,$69,$92,$9F,$01,$AC,$38,$FB,$D6
	.dc.b	$F7,$4F,$85,$C6,$1C,$7D,$FB,$2F,$5A,$FC,$46,$3D,$7C,$FC,$2E,$BE
	.dc.b	$17,$7E,$FF,$0F,$4B,$8F,$BD,$CB,$E1,$79,$FB,$FF,$0F,$A3,$F8,$B6
	.dc.b	$FF,$5C,$9D,$1F,$86,$E6,$FC,$03,$79,$7E,$EB,$79,$3E,$E3,$6B,$8F
	.dc.b	$57,$E3,$2B,$6D,$AD,$B6,$77,$1F,$97,$21,$F9,$52,$DF,$8A,$7C,$F7
	.dc.b	$DC,$8F,$36,$CF,$0B,$F0,$B4,$F0,$96,$FC,$2F,$72,$5F,$77,$55,$D1
	.dc.b	$CD,$74,$5D,$5F,$EF,$5E,$47,$BF,$F6,$09,$3F,$EB,$CC,$3F,$0D,$8F
	.dc.b	$E1,$67,$33,$F8,$EC,$3F,$1B,$18,$AD,$DF,$4A,$E5,$5A,$69,$F8,$AA
	.dc.b	$E1,$19,$8E,$86,$4D,$DF,$85,$B5,$D5,$DC,$65,$B1,$9B,$F1,$DF,$69
	.dc.b	$FC,$CF,$C2,$FE,$5B,$E0,$3F,$2F,$F0,$1F,$98,$F8,$8F,$CC,$FC,$0F
	.dc.b	$E7,$3E,$0F,$F5,$98,$76,$5F,$11,$9F,$87,$D7,$C4,$6F,$E2,$38,$7C
	.dc.b	$47,$1B,$E5,$E6,$F3,$F2,$FA,$7C,$3F,$5F,$67,$C9,$CB,$E9,$E7,$FE
	.dc.b	$55,$B7,$DA,$E9,$FC,$31,$1C,$10,$FB,$A7,$E1,$8B,$3F,$86,$E6,$FC
	.dc.b	$24,$63,$A7,$F0,$91,$82,$74,$FE,$17,$0F,$C2,$74,$79,$3A,$7E,$54
	.dc.b	$FD,$5A,$F9,$78,$7D,$2A,$D1,$1C,$10,$FA,$6E,$5A,$66,$39,$B8,$3E
	.dc.b	$CF,$4F,$D5,$7E,$FF,$47,$D5,$7F,$03,$CF,$F5,$5F,$AD,$CF,$F4,$DF
	.dc.b	$C3,$F1,$B9,$BE,$A4,$63,$97,$EA,$3A,$E5,$FA,$8F,$0E,$4F,$A8,$F9
	.dc.b	$76,$FC,$A7,$03,$B9,$3F,$CA,$32,$7C,$0F,$E1,$C9,$1E,$26,$BF,$89
	.dc.b	$6F,$07,$EB,$27,$8C,$EC,$86,$CB,$1B,$9E,$38,$78,$B9,$17,$CA,$F8
	.dc.b	$12,$F8,$3C,$AB,$84,$EE,$73,$3B,$9C,$DF,$68,$F7,$36,$FB,$B6,$F9
	.dc.b	$57,$0A,$EC,$70,$2D,$70,$6C,$F0,$74,$7F,$14,$C3,$93,$E3,$33,$87
	.dc.b	$98,$97,$A6,$C6,$A3,$04,$70,$4B,$F1,$3D,$5B,$3C,$9B,$3C,$5B,$3C
	.dc.b	$6F,$D1,$18,$23,$82,$61,$F1,$49,$3F,$01,$AF,$86,$36,$FE,$33,$1F
	.dc.b	$C7,$3F,$8D,$F6,$BF,$89,$11,$FC,$EB,$5F,$E2,$8D,$C7,$F1,$8D,$C7
	.dc.b	$EF,$B6,$7A,$1C,$88,$72,$2C,$F2,$6C,$F2,$7F,$5B,$F0,$3F,$C9,$7D
	.dc.b	$EF,$E4,$DF,$0B,$FC,$9F,$D1,$FE,$51,$F0,$3F,$CA,$7E,$03,$F9,$57
	.dc.b	$C0,$7F,$2B,$F6,$7F,$96,$7C,$07,$F8,$23,$EB,$7F,$C2,$5A,$7F,$3A
	.dc.b	$5E,$D6,$7E,$13,$5E,$D6,$FE,$13,$87,$B7,$C7,$E1,$3D,$9D,$3F,$A2
	.dc.b	$E6,$FF,$02,$F4,$7C,$CD,$3E,$77,$4F,$D0,$E8,$FB,$D3,$96,$C6,$7A
	.dc.b	$3F,$01,$39,$6C,$66,$BF,$04,$DC,$3F,$02,$DA,$E1,$3A,$3F,$00,$43
	.dc.b	$F6,$BD,$3F,$7A,$B8,$C6,$70,$F1,$B7,$0F,$0B,$4E,$87,$B0,$C4,$F2
	.dc.b	$70,$49,$E6,$D9,$E6,$D3,$9C,$68,$FE,$07,$5E,$5E,$07,$C2,$7D,$72
	.dc.b	$57,$F7,$48,$CB,$63,$2F,$DB,$6D,$7F,$82,$A7,$73,$93,$81,$0C,$12
	.dc.b	$38,$4F,$2B,$FE,$41,$5C,$A7,$35,$C5,$ED,$F0,$9F,$8D,$F8,$4F,$C7
	.dc.b	$7C,$47,$E3,$FE,$17,$F2,$1F,$11,$F9,$1F,$86,$FC,$97,$C4,$7E,$4F
	.dc.b	$E1,$FF,$29,$F1,$1F,$95,$F8,$8F,$CB,$7A,$3F,$97,$F6,$3F,$31,$F0
	.dc.b	$FF,$E6,$0F,$AD,$FF,$36,$70,$FF,$85,$9F,$F3,$4B,$7E,$1B,$D0,$FF
	.dc.b	$9B,$B8,$FE,$0F,$37,$E0,$7A,$3F,$0F,$D1,$F6,$79,$BE,$97,$3F,$D6
	.dc.b	$6E,$1F,$5A,$30,$47,$04,$9E,$51,$96,$C6,$6B,$B9,$3A,$7F,$30,$7F
	.dc.b	$56,$C3,$F0,$1D,$3E,$8B,$65,$F3,$F1,$FA,$33,$98,$C6,$3E,$D8,$C1
	.dc.b	$1F,$A9,$3C,$1A,$6C,$B5,$DA,$78,$47,$F0,$3E,$9F,$B4,$FE,$3F,$A3
	.dc.b	$ED,$3F,$91,$E7,$FB,$4F,$E4,$F9,$BE,$D3,$F9,$5E,$5F,$B4,$FE,$5F
	.dc.b	$93,$ED,$3F,$99,$E3,$FB,$5F,$11,$F6,$FE,$23,$EE,$72,$78,$B8,$7F
	.dc.b	$6D,$CB,$FE,$9F,$C3,$EF,$7C,$47,$E0,$48,$E0,$87,$F9,$52,$38,$36
	.dc.b	$36,$D3,$7C,$DF,$E6,$E9,$DB,$67,$6D,$9D,$F1,$FF,$80,$6B,$6D,$AD
	.dc.b	$BF,$81,$E6,$FB,$D5,$FC,$19,$AF,$84,$97,$D5,$B3,$CD,$EE,$7F,$B1
	.dc.b	$AF,$BB,$6F,$FD,$67,$18,$E4,$34,$91,$E4,$9E,$07,$49,$DC,$65,$B1
	.dc.b	$99,$E2,$D9,$D4,$62,$3C,$E7,$EB,$78,$FF,$0E,$FF,$4B,$E3,$FC,$3B
	.dc.b	$FD,$37,$8F,$F0,$EF,$F4,$FE,$3F,$C3,$BF,$D4,$78,$FF,$0E,$F4,$CB
	.dc.b	$C6,$DC,$7C,$1A,$7C,$72,$69,$F1,$E7,$FC,$34,$7C,$7F,$88,$FE,$0C
	.dc.b	$47,$F8,$33,$5F,$C3,$36,$7F,$04,$D9,$FB,$CD,$9E,$87,$22,$1C,$8B
	.dc.b	$3C,$9B,$3C,$9F,$C1,$73,$FF,$00,$69,$FE,$36,$BF,$F7,$C3,$FC,$6F
	.dc.b	$E1,$FF,$8F,$7C,$3F,$F1,$FF,$87,$FE,$41,$F0,$FF,$C8,$7E,$1F,$C9
	.dc.b	$AF,$E0,$35,$FC,$1F,$1F,$D6,$26,$5F,$64,$87,$CB,$72,$D3,$3B,$7C
	.dc.b	$C7,$F0,$E7,$F0,$33,$FA,$87,$85,$FC,$F1,$FE,$A5,$E1,$7F,$3C,$7F
	.dc.b	$AA,$78,$5F,$CF,$1F,$EA,$DE,$17,$F3,$C7,$FA,$C7,$85,$E0,$C3,$B6
	.dc.b	$7F,$79,$BC,$3F,$7B,$8F,$EE,$75,$7F,$02,$D7,$84,$7D,$FA,$DC,$63
	.dc.b	$1F,$3C,$61,$F5,$CE,$C8,$E9,$B8,$78,$5B,$D5,$FB,$6A,$FE,$83,$3F
	.dc.b	$CE,$89,$E0,$B3,$F0,$47,$1C,$3A,$11,$C1,$33,$FE,$27,$9F,$F1,$CF
	.dc.b	$4F,$FA,$D1,$2B,$FA,$D3,$5F,$E3,$4D,$C7,$F8,$B3,$73,$FE,$24,$D9
	.dc.b	$FE,$0C,$DC,$3F,$81,$B7,$0F,$F2,$EC,$F4,$3B,$90,$FB,$C7,$F9,$0F
	.dc.b	$83,$EF,$3F,$E4,$5E,$0F,$BC,$FF,$91,$F8,$3E,$F3,$FE,$49,$E0,$FB
	.dc.b	$CF,$F9,$2F,$5F,$EE,$DF,$0F,$FD,$DF,$E1,$FF,$BC,$7C,$3F,$F7,$9F
	.dc.b	$7F,$FB,$D7,$BF,$FD,$EF,$E1,$FC,$DB,$FE,$6B,$2F,$B2,$7C,$BD,$FF
	.dc.b	$08,$DE,$6F,$A3,$18,$E1,$F9,$D1,$83,$EA,$F3,$7D,$BE,$3F,$E6,$9D
	.dc.b	$BF,$B1,$C3,$FD,$25,$87,$12,$5F,$18,$FF,$42,$63,$F3,$9F,$8D,$CB
	.dc.b	$F4,$9B,$86,$5F,$98,$48,$D3,$63,$51,$82,$38,$24,$7D,$A3,$A9,$4E
	.dc.b	$45,$8F,$F1,$7B,$FB,$43,$FB,$B1,$FB,$43,$FB,$B1,$FB,$43,$FB,$B1
	.dc.b	$FB,$43,$FB,$B1,$FB,$43,$FC,$55,$C1,$F5,$DF,$F1,$66,$FF,$5D,$FF
	.dc.b	$17,$6D,$F5,$DF,$2E,$BF,$8E,$7E,$39,$FD,$5F,$E0,$3F,$C6,$C4,$7F
	.dc.b	$C6,$CD,$7F,$BB,$37,$0F,$EE,$4D,$C7,$FB,$73,$67,$A1,$C8,$87,$22
	.dc.b	$CF,$26,$CF,$26,$9C,$8E,$04,$8E,$0D,$8E,$0D,$38,$46,$89,$1A,$6C
	.dc.b	$6A,$30,$47,$07,$87,$9F,$FC,$3B,$5E,$DB,$FE,$F2,$DD,$4D,$25,$7A
	.dc.b	$A7,$6D,$9D,$B6,$76,$FA,$67,$6D,$9D,$B6,$77,$87,$AA,$B6,$DA,$DB
	.dc.b	$67,$78,$7B,$6B,$6D,$AE,$B8,$7F,$8E,$BE,$1F,$FC,$F0,$4C,$7F,$CF
	.dc.b	$0D,$7F,$C7,$0D,$C7,$FC,$68,$DB,$FF,$18,$36,$7F,$C4,$8D,$C3,$FC
	.dc.b	$42,$DC,$0D,$31,$F0,$2A,$98,$30,$C6,$0F,$C0,$75,$5A,$9C,$C3,$E0
	.dc.b	$B3,$7C,$69,$7C,$7E,$23,$38,$70,$F8,$3B,$5A,$C5,$F1,$18,$C3,$C1
	.dc.b	$F0,$96,$5C,$F7,$E1,$31,$F1,$19,$F8,$5D,$7C,$46,$FA,$F8,$7A,$FC
	.dc.b	$7E,$1B,$97,$C3,$7C,$1D,$FF,$6B,$B7,$D6,$7F,$26,$D2,$73,$FD,$EE
	.dc.b	$2F,$B5,$D3,$F7,$AB,$73,$96,$C6,$6B,$8B,$67,$51,$89,$FC,$A1,$0F
	.dc.b	$C8,$17,$97,$F1,$8D,$BF,$0B,$FB,$EE,$9F,$D3,$B9,$3C,$25,$AE,$CF
	.dc.b	$42,$57,$46,$D7,$46,$9C,$8B,$3C,$A7,$81,$27,$83,$67,$85,$3B,$7F
	.dc.b	$0B,$F0,$DF,$64,$93,$F8,$93,$F0,$0E,$15,$7E,$27,$0F,$C8,$4E,$63
	.dc.b	$1C,$88,$C6,$3D,$AF,$A5,$F0,$6D,$6D,$B5,$C1,$F0,$98,$21,$E2,$AE
	.dc.b	$8D,$36,$5A,$FD,$27,$C3,$7E,$57,$E1,$BF,$2D,$F0,$DF,$97,$F8,$6F
	.dc.b	$CC,$7C,$37,$E6,$7E,$1B,$F3,$9E,$A7,$E9,$72,$E4,$BE,$1B,$3F,$0D
	.dc.b	$AF,$86,$DF,$C3,$70,$F8,$6E,$3F,$0D,$CB,$E1,$79,$FC,$07,$4F,$5B
	.dc.b	$D9,$C7,$F0,$75,$FF,$40,$BF,$67,$93,$EB,$92,$3E,$F7,$3F,$D9,$9D
	.dc.b	$46,$27,$51,$82,$75,$7E,$68,$9D,$7F,$9D,$25,$7E,$46,$32,$D8,$FE
	.dc.b	$BD,$D7,$F8,$58,$F9,$78,$7D,$CE,$9F,$4F,$57,$D7,$9F,$9C,$D9,$D4
	.dc.b	$7D,$66,$BA,$9E,$24,$75,$97,$DA,$9D,$91,$C1,$23,$F8,$97,$57,$DB
	.dc.b	$7F,$13,$D3,$F6,$DF,$C5,$F4,$7D,$B7,$F1,$FC,$FF,$6F,$0F,$CC,$3F
	.dc.b	$6E,$33,$CD,$F6,$DD,$F2,$FD,$B7,$8F,$27,$DB,$79,$F1,$FD,$B7,$D2
	.dc.b	$7E,$43,$7F,$91,$C4,$C5,$F0,$1F,$8C,$25,$79,$1A,$FE,$39,$B9,$F9
	.dc.b	$9B,$BF,$F3,$EA,$E7,$5C,$48,$71,$2D,$71,$6C,$F1,$AE,$71,$FD,$02
	.dc.b	$B9,$BC,$49,$5C,$5B,$5C,$5E,$67,$1A,$E7,$7E,$F6,$DF,$4B,$E4,$4B
	.dc.b	$E5,$7D,$0E,$47,$4B,$E4,$FF,$40,$AE,$53,$C2,$B9,$4F,$03,$95,$70
	.dc.b	$6D,$70,$75,$3C,$27,$4D,$9D,$46,$2B,$B1,$2B,$8B,$63,$51,$82,$38
	.dc.b	$24,$F4,$25,$FC,$42,$4F,$BE,$32,$D8,$CB,$FC,$12,$1E,$BF,$F1,$42
	.dc.b	$47,$E7,$1A,$FF,$16,$6E,$1F,$9E,$6E,$1F,$89,$6C,$FE,$05,$B8,$7E
	.dc.b	$5D,$B7,$E0,$9E,$C4,$3B,$1F,$AC,$F8,$0F,$E4,$FE,$C7,$F2,$8F,$80
	.dc.b	$FE,$53,$EC,$FF,$2A,$F8,$0F,$E5,$7E,$D7,$F2,$CF,$85,$FE,$5B,$F0
	.dc.b	$DF,$CB,$BE,$1B,$FC,$15,$F8,$13,$E1,$EF,$C3,$E3,$E1,$F3,$EF,$6B
	.dc.b	$E0,$F7,$F1,$5C,$38,$FC,$FC,$9F,$E2,$0E,$9F,$D8,$C7,$CC,$E5,$FC
	.dc.b	$1F,$4F,$CC,$E9,$FC,$03,$7A,$7E,$E5,$6E,$73,$5F,$4C,$87,$22,$BA
	.dc.b	$70,$47,$04,$9E,$B5,$F7,$2F,$CA,$F8,$6B,$F1,$37,$E1,$AD,$C7,$66
	.dc.b	$BD,$5B,$7D,$63,$8D,$73,$E9,$FC,$8B,$6F,$A4,$E8,$E8,$53,$A1,$FA
	.dc.b	$BC,$3F,$8C,$F0,$3E,$23,$EA,$92,$7F,$BC,$C6,$5B,$19,$7F,$B8,$46
	.dc.b	$5B,$19,$C7,$FC,$11,$39,$6C,$67,$2F,$F2,$DC,$E5,$B1,$9A,$FF,$2B
	.dc.b	$CE,$5B,$19,$C9,$39,$6F,$C4,$7E,$3B,$E2,$3F,$1F,$F1,$1F,$90,$F4
	.dc.b	$7F,$23,$F0,$9F,$92,$F8,$7F,$C9,$FC,$2F,$E5,$3E,$1F,$F2,$BF,$0D
	.dc.b	$F9,$6F,$87,$FC,$BF,$9D,$F9,$8F,$87,$FC,$CF,$C3,$FF,$98,$FF,$03
	.dc.b	$FF,$38,$F1,$FF,$84,$5F,$F3,$83,$FE,$20,$5F,$0F,$E7,$E3,$F8,$DC
	.dc.b	$FF,$84,$AF,$AB,$C9,$F6,$9B,$D1,$F6,$A3,$11,$F6,$1B,$D1,$F8,$36
	.dc.b	$F4,$7E,$18,$93,$F7,$E7,$55,$CC,$9D,$3F,$87,$3F,$2D,$CB,$F4,$2B
	.dc.b	$EE,$90,$FB,$65,$AF,$B2,$DB,$E0,$E8,$91,$A6,$C6,$A3,$04,$70,$48
	.dc.b	$FB,$A7,$53,$F0,$3D,$1F,$6D,$FC,$9F,$3F,$DB,$7F,$2B,$CF,$F6,$DF
	.dc.b	$CB,$F3,$7D,$B7,$F3,$3C,$BF,$6D,$FC,$DF,$27,$DB,$7F,$3B,$C7,$F6
	.dc.b	$DF,$B9,$F1,$1F,$77,$8F,$EC,$75,$7F,$0F,$DF,$FD,$63,$CA,$F8,$8F
	.dc.b	$C1,$13,$0F,$C0,$35,$FC,$23,$78,$7F,$CD,$53,$C6,$76,$43,$65,$8D
	.dc.b	$CF,$1E,$1F,$0B,$67,$FD,$77,$5C,$DE,$24,$AE,$FF,$11,$F9,$7F,$88
	.dc.b	$FC,$C7,$C4,$7E,$67,$E2,$3F,$35,$F1,$1F,$9B,$F8,$8F,$CE,$7C,$47
	.dc.b	$E7,$7E,$23,$F3,$DF,$11,$34,$FC,$E7,$C4,$7E,$87,$E2,$3F,$45,$EF
	.dc.b	$49,$FD,$07,$BD,$E5,$F0,$5C,$5E,$5E,$2F,$5D,$7F,$AF,$A3,$1C,$9E
	.dc.b	$C8,$F8,$0D,$7D,$4D,$9E,$31,$83,$E0,$F2,$77,$AE,$71,$83,$99,$4E
	.dc.b	$17,$E4,$9E,$07,$1F,$88,$F8,$D3,$FB,$9F,$7F,$E3,$3F,$BB,$F8,$8F
	.dc.b	$E0,$1F,$01,$F1,$A3,$F7,$DC,$3F,$15,$FE,$05,$C3,$FC,$17,$87,$E3
	.dc.b	$47,$F0,$3F,$88,$FF,$75,$7E,$09,$FE,$55,$E0,$FC,$5F,$0F,$8A,$BF
	.dc.b	$05,$E4,$7F,$93,$2B,$FC,$B1,$1F,$E9,$D6,$F1,$7F,$AC,$DB,$69,$FF
	.dc.b	$7E,$C6,$1B,$F1,$3E,$02,$3F,$EB,$48,$D7,$11,$A4,$9D,$CF,$85,$B7
	.dc.b	$E1,$6D,$F8,$6F,$94,$62,$BC,$27,$F2,$3E,$CF,$83,$F0,$7F,$92,$8F
	.dc.b	$E4,$DD,$9F,$92,$7F,$94,$75,$7C,$6F,$89,$F8,$FF,$09,$F2,$3E,$27
	.dc.b	$E4,$FC,$2F,$CA,$F8,$9F,$97,$F0,$DF,$91,$C7,$F9,$97,$77,$CE,$F8
	.dc.b	$9F,$9F,$E8,$78,$F8,$7C,$B5,$F9,$1E,$FF,$A0,$FE,$13,$C0,$FC,$4C
	.dc.b	$FE,$16,$3F,$12,$D7,$53,$C3,$97,$F2,$64,$E5,$FC,$B1,$27,$F2,$11
	.dc.b	$96,$C7,$F0,$AF,$03,$F1,$F1,$FD,$27,$C0,$FC,$7B,$FD,$2F,$C0,$FC
	.dc.b	$7B,$FD,$2F,$C2,$FD,$A3,$FD,$37,$BF,$F2,$18,$FF,$0C,$EF,$FC,$7C
	.dc.b	$FF,$53,$EF,$FC,$7E,$7F,$D5,$FB,$FD,$0F,$EB,$7B,$FF,$D3,$3C,$1F
	.dc.b	$00,$9B,$EB,$0F,$39,$38,$3E,$16,$7E,$B6,$E1,$EF,$6C,$EE,$F9,$BF
	.dc.b	$29,$AF,$CF,$ED,$F8,$37,$FC,$93,$B3,$E0,$BF,$C9,$BC,$3F,$E2,$95
	.dc.b	$F0,$4F,$D7,$9F,$04,$87,$F7,$1D,$BE,$0C,$FF,$2E,$E6,$FC,$23,$FC
	.dc.b	$C3,$6F,$66,$7F,$6F,$6F,$83,$B6,$35,$F2,$DF,$B5,$BB,$7B,$5B,$9F
	.dc.b	$B6,$B8,$11,$FF,$00,$72,$7B,$2B,$FA,$3F,$27,$B1,$FE,$93,$C9,$F7
	.dc.b	$2B,$FC,$11,$C9,$F7,$1F,$E1,$27,$B3,$2F,$B8,$FF,$82,$FC,$3F,$F0
	.dc.b	$87,$27,$DC,$8F,$E5,$E7,$CD,$CF,$FA,$86,$BF,$C2,$7C,$0F,$D0,$78
	.dc.b	$3F,$BB,$AF,$E0,$11,$82,$38,$21,$FB,$F2,$E3,$FB,$F6,$E9,$FB,$FA
	.dc.b	$E9,$1B,$24,$FF,$50,$7F,$87,$78,$3F,$BF,$7F,$C4,$DC,$DF,$BF,$7F
	.dc.b	$C5,$1E,$0F,$EF,$DF,$E2,$9E,$07,$EF,$DF,$E2,$DD,$FF,$BF,$7F,$8C
	.dc.b	$77,$7D,$CC,$FF,$B9,$F4,$7E,$FE,$7F,$8E,$76,$FB,$16,$FF,$79,$F8
	.dc.b	$C4,$3F,$76,$5D,$7F,$01,$C1,$F6,$E3,$24,$8F,$C2,$73,$7D,$D8,$FE
	.dc.b	$E3,$CD,$F7,$5F,$EE,$7C,$DF,$76,$BF,$BF,$F3,$7D,$D7,$FC,$03,$DB
	.dc.b	$FB,$8D,$7F,$C1,$7D,$7F,$98,$7F,$C1,$BC,$FF,$77,$1F,$F2,$67,$5F
	.dc.b	$8F,$6F,$E8,$9E,$C7,$F4,$A2,$4F,$F3,$B6,$BF,$D1,$DB,$97,$F4,$76
	.dc.b	$E1,$F6,$5B,$3F,$D2,$1B,$D9,$FD,$0D,$BD,$3F,$EC,$59,$E8,$77,$21
	.dc.b	$FD,$57,$E1,$FF,$AE,$9F,$D7,$BE,$1F,$FA,$FF,$C3,$FF,$60,$F8,$7F
	.dc.b	$EC,$3F,$0F,$FD,$8B,$E1,$FF,$B1,$FC,$3C,$C7,$FA,$A7,$C3,$FF,$66
	.dc.b	$F8,$7F,$EB,$90,$E0,$70,$7F,$4F,$EE,$FF,$65,$4F,$F6,$19,$CA,$E1
	.dc.b	$4B,$E1,$B9,$D6,$A7,$3F,$0D,$8C,$3A,$7C,$35,$C7,$96,$2F,$86,$F8
	.dc.b	$7C,$BF,$E4,$39,$CC,$6F,$1E,$ED,$AE,$13,$E5,$C3,$D1,$39,$24,$64
	.dc.b	$E6,$5A,$E6,$DA,$E6,$FC,$3A,$E6,$E6,$B9,$B9,$39,$96,$B9,$B6,$B9
	.dc.b	$E5,$F0,$DB,$7D,$27,$57,$D2,$B5,$F0,$DE,$DF,$80,$FB,$D5,$FE,$0B
	.dc.b	$E4,$FB,$CF,$9F,$D7,$FB,$CF,$F8,$3F,$8F,$E2,$7C,$07,$C5,$F8,$6F
	.dc.b	$BD,$3E,$DF,$87,$F9,$1C,$72,$3E,$07,$17,$87,$7F,$26,$DE,$7D,$FE
	.dc.b	$97,$0F,$E5,$EF,$F2,$73,$F9,$F6,$D7,$E7,$DA,$7E,$7C,$B3,$C5,$FC
	.dc.b	$E9,$27,$93,$67,$93,$4E,$45,$9E,$53,$C0,$FF,$1B,$75,$7E,$92,$7E
	.dc.b	$D7,$C1,$7E,$17,$DB,$FC,$37,$C1,$7E,$92,$7E,$F7,$C3,$FE,$8D,$FF
	.dc.b	$20,$75,$FF,$92,$3A,$FF,$19,$F0,$7F,$8D,$F8,$7F,$C7,$7C,$27,$E3
	.dc.b	$FB,$66,$1F,$87,$F8,$7F,$CB,$BF,$E4,$5F,$11,$AF,$CA,$E2,$93,$F8
	.dc.b	$6E,$1F,$98,$FE,$2C,$87,$E0,$8A,$E9,$C1,$1C,$12,$3A,$F1,$7E,$42
	.dc.b	$31,$3D,$E7,$3D,$DF,$90,$8F,$D7,$F7,$7E,$35,$FF,$2B,$78,$7F,$C2
	.dc.b	$9F,$F2,$C7,$87,$FA,$DE,$EF,$C6,$C7,$F9,$97,$BB,$F1,$AF,$E9,$FB
	.dc.b	$7F,$1A,$FE,$7F,$B7,$F1,$8F,$E8,$7B,$7F,$6D,$DB,$F8,$D8,$F2,$1F
	.dc.b	$A8,$E1,$FE,$5E,$FE,$1B,$C4,$FE,$45,$C1,$F7,$78,$3F,$06,$D9,$FC
	.dc.b	$1B,$78,$7F,$07,$3A,$24,$E9,$B1,$A8,$C1,$1C,$12,$BC,$33,$FD,$C7
	.dc.b	$B7,$F1,$95,$FE,$85,$E0,$F8,$CF,$FA,$1F,$83,$F1,$8F,$C5,$3F,$D0
	.dc.b	$1D,$5F,$8D,$CB,$FD,$15,$D5,$F1,$9F,$E0,$5B,$FC,$67,$F8,$27,$3F
	.dc.b	$C6,$BF,$C0,$E7,$E8,$D3,$FA,$26,$DF,$8C,$9F,$93,$B7,$DF,$AF,$BB
	.dc.b	$DD,$F3,$63,$13,$AE,$6F,$C0,$46,$08,$E0,$93,$C5,$B3,$E2,$39,$47
	.dc.b	$F2,$FE,$6F,$C4,$47,$C3,$3F,$8A,$1F,$88,$3F,$8A,$1F,$C4,$C8,$7F
	.dc.b	$23,$EF,$FE,$28,$FF,$A9,$7C,$3F,$E7,$5B,$7C,$3C,$BF,$A8,$72,$FC
	.dc.b	$3A,$FD,$04,$7B,$9F,$D7,$74,$FC,$5F,$0F,$F6,$31,$FC,$17,$7F,$B2
	.dc.b	$FD,$D6,$ED,$F7,$67,$7C,$3F,$89,$9C,$92,$33,$87,$D1,$E6,$FC,$0D
	.dc.b	$7F,$39,$F0,$7F,$86,$63,$FD,$1F,$C1,$FC,$0C,$FF,$3D,$F0,$7F,$01
	.dc.b	$87,$F8,$E7,$9F,$F8,$65,$FF,$4D,$F0,$7F,$86,$3F,$D3,$FC,$0F,$E1
	.dc.b	$6F,$F4,$7E,$8F,$EB,$9E,$07,$B3,$A7,$F0,$1C,$3F,$63,$A3,$EB,$12
	.dc.b	$FE,$B4,$65,$BD,$3F,$69,$F9,$27,$DA,$33,$3C,$5F,$EA,$DD,$3F,$73
	.dc.b	$7F,$F7,$07,$67,$E7,$9F,$F7,$17,$67,$F0,$3A,$FB,$71,$FE,$E5,$EF
	.dc.b	$FE,$07,$1F,$D7,$FB,$FF,$3D,$C5,$FE,$09,$EB,$FB,$93,$FD,$FB,$AF
	.dc.b	$EE,$61,$FC,$5B,$6F,$44,$FF,$92,$BA,$FF,$C2,$1F,$0B,$FE,$14,$8F
	.dc.b	$F0,$AC,$FF,$83,$A3,$11,$A2,$C6,$5B,$C0,$9D,$C7,$F8,$39,$B3,$B8
	.dc.b	$EC,$D7,$BC,$F1,$79,$CF,$08,$C4,$70,$6B,$C2,$34,$47,$89,$0E,$0E
	.dc.b	$5E,$5F,$0F,$FE,$1B,$F8,$6F,$F0,$F7,$E0,$BF,$07,$2F,$F1,$52,$F8
	.dc.b	$7C,$FC,$3E,$BD,$3D,$FB,$FC,$3E,$1E,$67,$FE,$1D,$C1,$ED,$F4,$F8
	.dc.b	$7E,$BF,$03,$DB,$DD,$EF,$EB,$78,$3E,$1B,$C3,$EB,$FF,$8B,$78,$FC
	.dc.b	$DE,$FB,$DC,$FA,$1A,$FA,$8F,$61,$F0,$4A,$68,$C1,$B2,$18,$21,$B2
	.dc.b	$1F,$5C,$E6,$53,$89,$AE,$7F,$B0,$FF,$95,$F9,$BE,$53,$FE,$3B,$E6
	.dc.b	$F9,$7E,$CF,$CC,$F8,$1F,$9B,$EF,$7C,$8A,$F8,$DC,$9F,$21,$FF,$1D
	.dc.b	$72,$3D,$AF,$A7,$EF,$7C,$89,$F9,$9E,$F7,$D6,$F7,$BE,$44,$7C,$FF
	.dc.b	$87,$F9,$0F,$DF,$E1,$F9,$0F,$A4,$F7,$94,$F5,$77,$FE,$29,$BC,$7F
	.dc.b	$85,$77,$C5,$EE,$78,$11,$C1,$27,$94,$65,$B1,$99,$FD,$0E,$8F,$7F
	.dc.b	$FC,$D9,$F8,$37,$F9,$F3,$0F,$82,$7E,$4B,$C0,$FC,$7B,$F8,$BE,$FF
	.dc.b	$C8,$3F,$E6,$99,$7C,$1F,$2D,$3F,$33,$DF,$F8,$FA,$FC,$97,$7F,$E3
	.dc.b	$DF,$CA,$76,$FE,$41,$FC,$B7,$67,$C8,$7F,$0D,$7F,$23,$7F,$44,$7E
	.dc.b	$09,$AF,$E0,$8F,$76,$FF,$1D,$F9,$26,$F8,$3E,$09,$23,$51,$C0,$D7
	.dc.b	$5F,$E2,$1F,$F3,$C7,$5F,$E2,$1F,$D5,$75,$7E,$21,$FD,$67,$81,$F8
	.dc.b	$7F,$0B,$F6,$BD,$7F,$B8,$EB,$FC,$3D,$7E,$C7,$C1,$FD,$DF,$5F,$E1
	.dc.b	$E3,$F6,$9E,$07,$EF,$BA,$FE,$04,$7E,$9B,$87,$E0,$73,$BE,$23,$EA
	.dc.b	$92,$7E,$83,$5F,$F4,$84,$6D,$C9,$0C,$C6,$F4,$FF,$5A,$4E,$E7,$26
	.dc.b	$E3,$33,$BF,$1D,$3F,$EB,$FA,$E1,$5A,$8E,$13,$AE,$45,$6A,$30,$BE
	.dc.b	$13,$F1,$FF,$11,$2F,$F1,$5C,$DF,$92,$F8,$8F,$C9,$FC,$37,$E5,$3E
	.dc.b	$23,$F2,$BE,$3F,$E5,$BD,$DF,$CB,$FB,$BF,$98,$F4,$BF,$33,$F0,$BF
	.dc.b	$9A,$F4,$FF,$37,$F0,$FF,$9C,$F5,$3D,$5B,$7A,$79,$7E,$47,$AF,$FD
	.dc.b	$B3,$8D,$EE,$7E,$88,$95,$F9,$D6,$BF,$39,$BC,$FF,$29,$BC,$DE,$E6
	.dc.b	$CF,$C7,$6F,$3F,$C4,$6D,$F8,$27,$B1,$0E,$C7,$EB,$79,$FE,$33,$F7
	.dc.b	$B9,$FF,$6F,$F0,$7F,$1E,$3F,$65,$CD,$F1,$DF,$DA,$7C,$17,$FB,$9F
	.dc.b	$F0,$7F,$F7,$CC,$7E,$1B,$97,$1E,$BE,$7D,$DF,$91,$3F,$EE,$3A,$7C
	.dc.b	$2F,$1F,$77,$97,$BB,$E8,$E0,$FF,$61,$F0,$7D,$DC,$3E,$61,$F8,$22
	.dc.b	$9F,$8C,$27,$1F,$E1,$DB,$C7,$F8,$4E,$5F,$AB,$C9,$F8,$86,$DF,$58
	.dc.b	$C5,$75,$6B,$C8,$EF,$7D,$45,$FE,$4F,$47,$C4,$7C,$3F,$87,$F8,$9F
	.dc.b	$0F,$FC,$83,$C2,$F8,$DF,$01,$F1,$FE,$1B,$E4,$7C,$07,$E4,$34,$FE
	.dc.b	$3D,$E1,$7C,$BF,$86,$FC,$84,$7F,$21,$F0,$BE,$77,$C3,$7E,$42,$3F
	.dc.b	$92,$78,$3E,$AE,$4F,$C7,$E3,$F8,$59,$FB,$DC,$5F,$89,$6F,$16,$5F
	.dc.b	$C0,$92,$34,$D8,$D4,$60,$8E,$09,$1F,$91,$3A,$94,$E4,$58,$FE,$07
	.dc.b	$E0,$7E,$39,$FE,$0B,$E0,$7E,$3A,$7F,$9F,$78,$FF,$B4,$7F,$A0,$78
	.dc.b	$FF,$D2,$3C,$0F,$C7,$7C,$1F,$ED,$27,$F1,$5E,$07,$E3,$5F,$E8,$DE
	.dc.b	$37,$E4,$FC,$0F,$C6,$47,$E3,$FB,$FF,$18,$FF,$9F,$07,$6B,$B5,$F0
	.dc.b	$DF,$9B,$24,$FE,$5D,$AB,$7F,$C0,$BC,$A7,$8B,$B2,$1B,$E1,$FC,$35
	.dc.b	$72,$AE,$37,$E1,$8C,$4E,$AB,$8D,$6A,$FA,$C7,$33,$C1,$5C,$1A,$70
	.dc.b	$AF,$E4,$7C,$3F,$3D,$FE,$4B,$C3,$F9,$39,$F9,$E7,$EF,$0F,$9E,$7F
	.dc.b	$6F,$F2,$7F,$91,$3F,$DC,$7C,$4F,$C9,$BF,$D0,$BA,$3F,$76,$FF,$44
	.dc.b	$F2,$3F,$91,$3F,$A9,$CF,$CF,$A7,$A7,$4F,$E7,$B3,$FE,$71,$3E,$47
	.dc.b	$C4,$7F,$09,$23,$FC,$15,$AF,$BF,$6F,$AD,$18,$AF,$88,$DD,$7F,$64
	.dc.b	$DA,$EA,$73,$21,$CF,$04,$6E,$FA,$CE,$AF,$AB,$FC,$94,$FF,$09,$6B
	.dc.b	$F2,$DF,$EA,$DE,$47,$F2,$57,$FB,$2F,$91,$FC,$95,$FE,$CF,$CD,$F9
	.dc.b	$87,$E5,$1F,$98,$3F,$91,$79,$1F,$2B,$0F,$BD,$C9,$F9,$89,$F3,$72
	.dc.b	$79,$75,$F3,$C7,$E6,$74,$FC,$43,$F7,$CF,$C0,$9E,$DD,$FF,$09,$1A
	.dc.b	$8C,$1B,$71,$87,$D5,$8C,$57,$D8,$DF,$F2,$51,$C0,$93,$B6,$C6,$A3
	.dc.b	$07,$E5,$FC,$5F,$E4,$53,$FC,$FF,$C4,$FC,$83,$F4,$7C,$3F,$E4,$4F
	.dc.b	$F4,$3E,$DF,$E4,$4F,$F8,$3B,$C2,$FE,$44,$FF,$90,$7C,$AF,$EE,$2F
	.dc.b	$F4,$9F,$0B,$F9,$13,$FE,$14,$F0,$BF,$9E,$EB,$E9,$8F,$87,$E5,$7F
	.dc.b	$B7,$23,$E4,$C7,$E0,$7A,$3E,$61,$1F,$C0,$6F,$F8,$A6,$9F,$8A,$38
	.dc.b	$18,$E0,$FC,$3B,$82,$70,$7E,$21,$B3,$CD,$A7,$F8,$77,$B7,$F1,$55
	.dc.b	$FE,$1A,$E9,$FE,$45,$3F,$E2,$EF,$0B,$F9,$13,$FE,$30,$F0,$BF,$91
	.dc.b	$3F,$DA,$7C,$2F,$E4,$4F,$E6,$FC,$2F,$E4,$4F,$F8,$DB,$C2,$FE,$44
	.dc.b	$FF,$6F,$F2,$BF,$B6,$BE,$9D,$FE,$1F,$07,$C5,$E9,$FC,$3C,$7C,$9D
	.dc.b	$7F,$0B,$BF,$DD,$25,$FE,$3B,$83,$EF,$C6,$27,$73,$98,$C5,$73,$25
	.dc.b	$F6,$24,$FE,$B4,$FF,$08,$1F,$AE,$3F,$C6,$E7,$F1,$63,$F5,$A7,$F1
	.dc.b	$63,$FC,$A7,$E1,$FF,$16,$7F,$C2,$07,$E1,$CF,$ED,$07,$F4,$43,$FD
	.dc.b	$15,$E0,$FF,$16,$7F,$BE,$61,$E8,$3F,$4D,$97,$E1,$F8,$FF,$09,$3F
	.dc.b	$62,$7E,$FE,$7F,$74,$99,$FD,$D6,$CF,$E1,$DB,$3C,$27,$44,$9D,$36
	.dc.b	$35,$18,$23,$FE,$5E,$E3,$FC,$3F,$92,$E4,$FC,$3E,$1F,$E6,$EE,$4F
	.dc.b	$C3,$BF,$E7,$0E,$4F,$C3,$E3,$FC,$7F,$93,$F0,$F3,$FD,$2F,$93,$F0
	.dc.b	$EF,$F9,$DB,$93,$F0,$F5,$FC,$97,$93,$F0,$EF,$C1,$E4,$FC,$3F,$4F
	.dc.b	$DC,$E5,$FB,$33,$FC,$08,$FC,$01,$38,$3E,$65,$7E,$02,$3F,$86,$CF
	.dc.b	$0C,$7F,$13,$1B,$9D,$3C,$EB,$FB,$A7,$67,$F0,$A8,$FE,$71,$D9,$FC
	.dc.b	$2A,$FF,$B2,$76,$7F,$0A,$9F,$E7,$99,$7F,$0A,$9F,$F4,$8F,$67,$F0
	.dc.b	$A9,$FF,$4C,$F6,$7F,$0A,$7F,$A1,$76,$7F,$0A,$AF,$F4,$BF,$67,$F0
	.dc.b	$AD,$BF,$BF,$F5,$7E,$0E,$FF,$51,$D5,$F8,$4B,$FB,$BC,$1F,$C3,$1B
	.dc.b	$19,$EA,$FB,$93,$96,$C6,$6B,$EE,$CE,$5F,$E0,$87,$E8,$4F,$D1,$9F
	.dc.b	$EB,$0F,$0B,$FA,$75,$FF,$AB,$3C,$2F,$E9,$CF,$FA,$BB,$C2,$FE,$9D
	.dc.b	$5F,$A2,$3F,$82,$1F,$D3,$8F,$D1,$1F,$D8,$7C,$1F,$E9,$D3,$FE,$0D
	.dc.b	$F0,$7F,$3F,$A7,$F2,$7F,$87,$FF,$3E,$12,$FF,$CF,$0D,$7F,$B5,$36
	.dc.b	$BF,$C2,$CD,$D7,$FC,$C8,$D9,$FF,$13,$37,$9F,$FC,$60,$DD,$4D,$35
	.dc.b	$DF,$FE,$BB,$AF,$89,$87,$F9,$21,$FE,$0A,$7F,$27,$3F,$C3,$47,$F2
	.dc.b	$73,$FD,$1D,$F0,$FF,$E9,$5F,$63,$FD,$2D,$EC,$7F,$A5,$FD,$8F,$F0
	.dc.b	$9D,$FF,$8C,$BD,$BF,$F0,$AE,$1F,$E0,$F5,$EF,$EB,$6F,$F6,$84,$3E
	.dc.b	$1F,$15,$FE,$4A,$87,$C0,$63,$81,$0F,$6F,$D9,$5F,$DE,$61,$ED,$ED
	.dc.b	$B1,$FE,$E2,$CF,$C9,$39,$24,$67,$2F,$1C,$62,$3D,$0D,$9E,$31,$B3
	.dc.b	$CD,$19,$9E,$31,$9C,$3C,$F3,$98,$D9,$9B,$F2,$BD,$8C,$47,$BF,$83
	.dc.b	$EB,$47,$E6,$30,$FA,$CF,$E6,$4F,$E7,$4F,$E6,$5F,$E7,$4F,$D1,$F7
	.dc.b	$FE,$B4,$77,$BF,$89,$ED,$FC,$58,$F8,$DB,$7C,$7F,$87,$F9,$1B,$7C
	.dc.b	$9F,$80,$F4,$6B,$EA,$BF,$F1,$37,$37,$E3,$1F,$CC,$90,$FC,$C9,$74
	.dc.b	$FC,$CB,$77,$FC,$CB,$4F,$CC,$96,$BF,$32,$DA,$EC,$F4,$25,$74,$6D
	.dc.b	$74,$7F,$CA,$5E,$37,$F1,$D9,$FF,$3F,$78,$DF,$C7,$6B,$FD,$99,$E3
	.dc.b	$7F,$1D,$9F,$F4,$17,$8D,$FE,$EF,$E9,$FC,$47,$B9,$F9,$CC,$BF,$DD
	.dc.b	$9D,$1F,$8B,$F8,$0F,$C6,$7C,$3F,$E3,$7A,$A5,$7D,$BE,$AF,$C9,$BF
	.dc.b	$E7,$3E,$1F,$90,$FD,$D2,$69,$F8,$A6,$ED,$F8,$86,$9F,$86,$2C,$F2
	.dc.b	$9E,$04,$9E,$0D,$9E,$0D,$9E,$13,$A2,$4F,$F2,$43,$F1,$8F,$FA,$69
	.dc.b	$FC,$C7,$4B,$C3,$FE,$4B,$29,$FF,$5E,$CB,$C3,$FE,$4F,$3F,$BA,$EB
	.dc.b	$FC,$7D,$7E,$FF,$AF,$F1,$EF,$E7,$CF,$C7,$9F,$A1,$E9,$FC,$7B,$FA
	.dc.b	$3E,$8F,$C7,$CF,$E9,$3C,$07,$2F,$F0,$6F,$81,$FD,$E1,$2F,$F5,$4D
	.dc.b	$7F,$87,$37,$7F,$E2,$4D,$CF,$94,$F0,$21,$C0,$B3,$C1,$B1,$C1,$D1
	.dc.b	$23,$F8,$F7,$C3,$FF,$0F,$F8,$29,$3F,$BD,$F8,$7F,$E2,$5F,$07,$FC
	.dc.b	$4F,$E1,$FF,$8A,$7C,$27,$F1,$5F,$86,$FE,$39,$F8,$47,$FB,$83,$0F
	.dc.b	$F5,$A2,$F8,$7C,$FC,$3C,$9F,$E1,$92,$F8,$2E,$3E,$EF,$2F,$77,$9F
	.dc.b	$C2,$FB,$EF,$F9,$57,$1F,$FB,$1B,$97,$E5,$C6,$0F,$69,$4F,$67,$27
	.dc.b	$BF,$93,$E1,$46,$0D,$38,$24,$6C,$91,$A6,$F6,$FF,$38,$BF,$23,$67
	.dc.b	$C2,$D7,$E0,$FB,$DF,$29,$FE,$91,$CB,$FC,$EF,$97,$E5,$C7,$F2,$9E
	.dc.b	$4F,$93,$D9,$F2,$E3,$E2,$E5,$F3,$3D,$EF,$91,$1F,$33,$8F,$E7,$FC
	.dc.b	$07,$D0,$F6,$FE,$44,$FC,$CE,$1F,$2F,$63,$83,$E4,$76,$7E,$87,$85
	.dc.b	$CE,$F8,$3F,$B2,$4C,$7E,$AB,$5F,$B6,$DE,$46,$FC,$EB,$95,$70,$21
	.dc.b	$C0,$B3,$C1,$5F,$8A,$FA,$5F,$36,$CE,$AF,$AB,$CC,$97,$F9,$87,$F1
	.dc.b	$7E,$E7,$E4,$BE,$0F,$F2,$7E,$FF,$E5,$3E,$0B,$F2,$BE,$F7,$E6,$2F
	.dc.b	$F2,$3F,$0F,$F9,$8F,$23,$F3,$3E,$07,$E6,$BD,$DF,$EE,$DF,$84,$BE
	.dc.b	$13,$F3,$BE,$1F,$E5,$B1,$FC,$CF,$87,$AF,$84,$F3,$F2,$FA,$E7,$FB
	.dc.b	$D7,$3F,$D2,$23,$F8,$6E,$8F,$A8,$D8,$CB,$F9,$59,$FC,$57,$27,$E5
	.dc.b	$23,$F2,$44,$8C,$F3,$FE,$52,$F9,$CF,$E4,$0F,$EE,$9E,$0F,$E4,$6F
	.dc.b	$FC,$35,$E0,$FE,$42,$7F,$C3,$5E,$0F,$E4,$1F,$EF,$5E,$2F,$EE,$B1
	.dc.b	$FC,$84,$7E,$DB,$C0,$FC,$8E,$1F,$67,$3F,$C4,$F8,$1F,$91,$AF,$DD
	.dc.b	$F8,$BF,$B8,$7D,$87,$F2,$FF,$17,$F6,$F4,$E4,$7C,$2F,$CE,$25,$7F
	.dc.b	$83,$E3,$2D,$8C,$AC,$7C,$53,$A8,$C6,$7F,$9D,$8C,$4E,$AB,$8B,$6B
	.dc.b	$A1,$0E,$27,$E5,$27,$8D,$73,$AE,$2D,$9E,$15,$CD,$FF,$22,$79,$3F
	.dc.b	$3E,$7F,$C8,$7C,$1F,$BF,$9F,$F1,$37,$93,$FC,$A5,$FF,$14,$79,$3F
	.dc.b	$8E,$F8,$1F,$C7,$FC,$2F,$E4,$3E,$0B,$F2,$3E,$D7,$E4,$BE,$23,$F2
	.dc.b	$7E,$DF,$E5,$3D,$BF,$CA,$FC,$47,$E5,$BC,$CF,$2E,$9E,$9E,$2F,$DF
	.dc.b	$BE,$BF,$88,$FC,$D9,$1F,$9C,$D7,$E0,$F2,$FD,$58,$C7,$27,$CE,$6F
	.dc.b	$1F,$CD,$6D,$75,$39,$90,$E7,$CA,$8D,$DF,$59,$D5,$F5,$7F,$A1,$FA
	.dc.b	$5F,$25,$FF,$34,$72,$FC,$97,$FC,$E1,$CB,$F2,$5F,$E9,$BE,$97,$C9
	.dc.b	$7F,$11,$C9,$F2,$5F,$C5,$7C,$47,$C9,$7F,$19,$F1,$1F,$25,$FC,$77
	.dc.b	$0F,$C9,$7C,$7B,$FB,$5F,$CC,$70,$FD,$F7,$D7,$C3,$F3,$B7,$FC,$D3
	.dc.b	$78,$7F,$0F,$3F,$44,$FC,$2D,$F1,$E0,$FC,$31,$1C,$12,$38,$1D,$8E
	.dc.b	$26,$23,$83,$63,$FD,$1F,$D3,$FD,$C3,$CF,$FE,$41,$3F,$DB,$3C,$EF
	.dc.b	$E4,$0F,$F6,$EF,$37,$F9,$03,$FE,$99,$F0,$7F,$8E,$FC,$47,$F1,$EF
	.dc.b	$88,$FC,$7C,$FE,$C7,$B3,$F1,$EF,$EC,$FE,$07,$F9,$1F,$C4,$7F,$24
	.dc.b	$F8,$8F,$81,$C3,$F8,$EE,$2F,$B9,$7F,$80,$EC,$FB,$FC,$7F,$81,$8F
	.dc.b	$9A,$7E,$1C,$D1,$89,$E3,$C7,$F8,$38,$C9,$1C,$13,$8F,$EB,$46,$5B
	.dc.b	$19,$8E,$CF,$F1,$7F,$55,$DB,$F8,$DB,$FF,$58,$F6,$FE,$36,$BF,$D7
	.dc.b	$DD,$BF,$8D,$78,$76,$FE,$37,$1F,$CD,$F6,$FE,$35,$FC,$EF,$6F,$E3
	.dc.b	$5F,$CF,$F6,$FE,$35,$FD,$0F,$6F,$E3,$5F,$31,$F0,$7C,$4F,$E4,$7C
	.dc.b	$0F,$82,$FE,$9E,$49,$FE,$44,$D7,$F5,$6D,$CF,$F9,$0B,$78,$7F,$90
	.dc.b	$36,$7F,$8D,$37,$7F,$E6,$0D,$BF,$04,$F6,$21,$D8,$FF,$5E,$77,$7C
	.dc.b	$7F,$2D,$C3,$F1,$F1,$FF,$6F,$F0,$FC,$77,$FD,$C3,$E3,$7F,$B7,$7F
	.dc.b	$0A,$FF,$73,$EB,$FD,$1D,$FF,$73,$E5,$FE,$CB,$7F,$DC,$CD,$F8,$6E
	.dc.b	$1F,$03,$C7,$E2,$79,$7C,$0C,$AE,$1A,$FA,$78,$3F,$21,$8B,$E2,$7C
	.dc.b	$04,$AE,$CD,$7F,$8E,$C6,$DC,$90,$CC,$6F,$3F,$43,$75,$FF,$14,$57
	.dc.b	$1A,$D9,$C6,$B6,$D9,$DD,$71,$8F,$F0,$D7,$4F,$CB,$E0,$F9,$51,$FE
	.dc.b	$22,$C9,$C1,$F2,$E3,$FC,$41,$D3,$F2,$DF,$99,$F1,$3F,$37,$E0,$FE
	.dc.b	$77,$C1,$FC,$89,$F9,$9C,$5F,$47,$E2,$3E,$97,$C4,$7D,$3C,$FC,$5D
	.dc.b	$2E,$2F,$57,$57,$E4,$B9,$3E,$66,$5F,$82,$E4,$FC,$6B,$79,$3F,$1A
	.dc.b	$D3,$F1,$A5,$C3,$F1,$AD,$EA,$FC,$6B,$F8,$B2,$57,$46,$D7,$46,$9F
	.dc.b	$0F,$AB,$F8,$E3,$FE,$86,$F8,$1F,$F1,$DB,$FE,$88,$F8,$1F,$F1,$DB
	.dc.b	$FD,$A8,$FF,$04,$7A,$7F,$91,$F8,$8F,$C9,$7C,$47,$E4,$FB,$BF,$29
	.dc.b	$ED,$FE,$36,$FF,$25,$DD,$F8,$D7,$F0,$B9,$FC,$27,$FD,$C1,$C5,$F7
	.dc.b	$F6,$FB,$B1,$F8,$02,$38,$26,$DF,$80,$8F,$C1,$1F,$E5,$E9,$E5,$19
	.dc.b	$8E,$2D,$9E,$31,$BE,$EF,$AF,$5C,$23,$07,$4E,$1F,$B5,$3F,$EB,$0E
	.dc.b	$EF,$C6,$4F,$FA,$DC,$FF,$07,$9F,$E8,$D3,$FD,$5F,$D9,$F8,$B7,$F6
	.dc.b	$5D,$9F,$8B,$7F,$69,$D7,$F8,$B7,$F4,$FD,$7F,$8B,$7F,$71,$D5,$F8
	.dc.b	$B7,$C5,$9F,$E2,$BD,$2F,$D2,$F7,$FF,$01,$24,$FE,$FD,$AF,$EC,$9B
	.dc.b	$B7,$F0,$56,$EF,$CA,$78,$10,$E0,$59,$E0,$D8,$E0,$E8,$91,$A6,$C6
	.dc.b	$A3,$04,$70,$48,$FB,$13,$E4,$69,$DC,$A7,$42,$CF,$36,$CF,$3D,$FF
	.dc.b	$61,$5F,$C6,$0E,$8E,$27,$A3,$5E,$8D,$AE,$8D,$AE,$98,$7F,$34,$AE
	.dc.b	$2D,$78,$B6,$B8,$B6,$78,$DF,$8A,$31,$5E,$82,$5F,$76,$C6,$A3,$04
	.dc.b	$70,$49,$E8,$49,$F2,$BF,$CE,$3D,$BF,$E8,$C4,$8F,$E8,$8D,$7F,$9B
	.dc.b	$B7,$0F,$E9,$4D,$C3,$F9,$9B,$67,$F5,$CD,$C3,$F8,$F3,$70,$34,$C7
	.dc.b	$81,$FD,$5F,$DB,$FE,$B1,$E1,$FF,$59,$F4,$58,$CC,$FF,$D4,$1E,$BF
	.dc.b	$FB,$66,$3F,$CC,$EB,$E0,$77,$18,$F7,$2D,$70,$F7,$EE,$FF,$EF,$1C
	.dc.b	$1E,$FE,$30,$FF,$71,$C3,$DD,$C7,$0A,$1F,$0B,$87,$FC,$37,$5E,$5C
	.dc.b	$3A,$74,$E7,$E1,$F5,$F0,$1B,$F8,$7F,$81,$7F,$E5,$C3,$ED,$F3,$FD
	.dc.b	$57,$F0,$ED,$DF,$F0,$ED,$C7,$F0,$F3,$F6,$88,$70,$2C,$F0,$6C,$70
	.dc.b	$74,$48,$D3,$63,$F7,$3E,$57,$FB,$1B,$A3,$F1,$53,$C8,$F3,$97,$6F
	.dc.b	$C8,$61,$E5,$8C,$9C,$CB,$87,$62,$61,$E3,$6C,$6A,$30,$47,$04,$AF
	.dc.b	$DA,$61,$FE,$DA,$D5,$F0,$FF,$4C,$95,$F4,$CF,$B6,$E1,$FC,$3B,$70
	.dc.b	$FC,$5B,$70,$FC,$23,$67,$89,$CF,$0F,$C1,$46,$5B,$19,$BF,$1B,$67
	.dc.b	$51,$88,$FC,$BF,$BF,$F8,$9F,$2B,$F1,$5E,$FF,$E2,$FC,$BF,$C6,$7B
	.dc.b	$FF,$8D,$F8,$7F,$C7,$7C,$3F,$E3,$FE,$1F,$F2,$7F,$85,$BE,$22,$FC
	.dc.b	$46,$3D,$FC,$FA,$7A,$F7,$77,$F1,$1C,$3E,$13,$8F,$1F,$93,$93,$F9
	.dc.b	$17,$27,$DB,$EA,$FB,$EF,$D9,$E6,$FB,$AE,$7A,$BE,$F4,$62,$35,$3C
	.dc.b	$3A,$7E,$DC,$E5,$B1,$9E,$9F,$B9,$39,$6C,$66,$BF,$9D,$F5,$7D,$DC
	.dc.b	$7C,$D1,$F2,$C8,$7D,$A2,$E1,$F6,$9B,$7F,$69,$A7,$DA,$38,$12,$38
	.dc.b	$36,$38,$34,$E1,$1F,$C4,$3A,$BE,$C4,$7D,$FE,$9F,$B0,$FE,$07,$A3
	.dc.b	$EC,$3F,$83,$E7,$FB,$0F,$E2,$79,$BE,$C3,$79,$7E,$C3,$9E,$4F,$B0
	.dc.b	$EF,$8F,$EC,$3C,$78,$BE,$1A,$DD,$E5,$7F,$26,$E2,$64,$F8,$8F,$C3
	.dc.b	$92,$FC,$4D,$5C,$0E,$1E,$55,$C6,$76,$43,$71,$AA,$E5,$5C,$65,$7E
	.dc.b	$29,$CD,$6E,$79,$56,$EB,$51,$89,$D3,$D8,$E7,$7E,$37,$13,$C1,$B0
	.dc.b	$F8,$8F,$D2,$FB,$FF,$A6,$CB,$F4,$FF,$07,$FA,$8F,$82,$FD,$4F,$C1
	.dc.b	$7E,$AB,$E2,$3F,$57,$F1,$1F,$AC,$F8,$8F,$D6,$FC,$47,$EB,$BE,$23
	.dc.b	$F5,$FF,$07,$FD,$0B,$F0,$CF,$F5,$5F,$17,$CF,$6F,$0F,$C0,$7F,$D5
	.dc.b	$AF,$99,$FF,$56,$BE,$6E,$4F,$A1,$CB,$F6,$23,$87,$C0,$F5,$23,$FC
	.dc.b	$3E,$91,$88,$E8,$DE,$6F,$95,$3A,$9E,$13,$9C,$3C,$CD,$8E,$E6,$88
	.dc.b	$74,$2D,$74,$8E,$1C,$CB,$7B,$5F,$3A,$FF,$39,$C9,$F1,$7E,$23,$E3
	.dc.b	$7C,$0F,$CE,$9F,$D0,$71,$FC,$E7,$F4,$5C,$7F,$39,$F7,$7C,$37,$CE
	.dc.b	$7F,$4D,$C3,$F3,$FE,$07,$E8,$7C,$47,$D1,$F8,$5F,$1E,$DE,$FB,$FD
	.dc.b	$16,$BF,$88,$E1,$FC,$67,$6F,$E1,$DB,$CB,$F9,$2A,$FC,$5B,$78,$BF
	.dc.b	$15,$3A,$9F,$C9,$1A,$2C,$6B,$93,$B1,$23,$47,$DD,$37,$7E,$18,$D4
	.dc.b	$FE,$D3,$B7,$F1,$5F,$03,$F8,$BF,$88,$FC,$AE,$1F,$C1,$7B,$7F,$1D
	.dc.b	$F0,$5F,$8F,$F8,$7F,$C8,$7C,$3F,$E5,$6B,$F0,$9F,$0F,$F9,$57,$F0
	.dc.b	$DE,$77,$E5,$58,$7E,$27,$E1,$BF,$31,$F0,$DE,$FE,$0F,$D5,$78,$1F
	.dc.b	$91,$E0,$FB,$9E,$07,$CC,$26,$1F,$82,$E2,$FC,$4B,$4F,$C4,$9F,$87
	.dc.b	$27,$1F,$E1,$DD,$57,$46,$F3,$7D,$79,$D4,$70,$AE,$AF,$E9,$FC,$0F
	.dc.b	$C6,$3F,$BC,$E7,$FC,$63,$FA,$3E,$FF,$C6,$3F,$C0,$3B,$FF,$18,$FF
	.dc.b	$02,$EF,$FC,$63,$FB,$1E,$FF,$C6,$3F,$C1,$BB,$FF,$18,$FE,$B7,$BB
	.dc.b	$F1,$8F,$9B,$7F,$F2,$4F,$07,$C5,$DD,$F0,$DF,$C0,$49,$5F,$B5,$6B
	.dc.b	$FB,$F6,$F0,$7E,$D1,$BC,$1F,$C1,$DB,$3F,$C1,$DB,$C3,$FA,$D6,$DF
	.dc.b	$82,$7B,$10,$EC,$7F,$05,$F8,$6F,$E2,$9F,$0D,$FC,$57,$E1,$BF,$CA
	.dc.b	$FF,$86,$FF,$99,$73,$FE,$5C,$FF,$99,$6F,$FC,$34,$BE,$1B,$7F,$0D
	.dc.b	$27,$3F,$0D,$CB,$E1,$B9,$FC,$37,$4F,$88,$EB,$F0,$FD,$BE,$1F,$BF
	.dc.b	$C3,$F8,$3E,$1F,$D3,$C7,$F2,$BB,$BE,$06,$9E,$9E,$0F,$88,$DE,$6C
	.dc.b	$BE,$C2,$46,$9B,$1A,$8C,$11,$C1,$23,$E6,$1D,$4A,$72,$2C,$7F,$28
	.dc.b	$E4,$F9,$31,$FC,$BF,$8F,$E4,$D7,$CA,$F7,$BE,$5F,$1F,$C5,$AF,$EA
	.dc.b	$3C,$8E,$4F,$9F,$EF,$7D,$0F,$87,$FA,$3E,$FF,$C7,$BF,$9F,$F0,$FF
	.dc.b	$53,$E0,$3E,$AF,$C3,$FD,$6F,$81,$F8,$F5,$FE,$83,$EE,$68,$F8,$7F
	.dc.b	$B6,$49,$FB,$2D,$7C,$6D,$E3,$FB,$ED,$E1,$FF,$02,$57,$3A,$E2,$43
	.dc.b	$89,$6B,$8B,$67,$8D,$73,$8F,$EF,$B5,$CD,$FD,$C1,$FE,$65,$EA,$FF
	.dc.b	$0E,$BF,$DD,$97,$C1,$7F,$AA,$97,$C3,$7E,$5B,$E2,$31,$F0,$99,$F8
	.dc.b	$7D,$7C,$2F,$E6,$BC,$3F,$C9,$65,$F9,$8F,$0B,$97,$C3,$FE,$4A,$3F
	.dc.b	$37,$E1,$7E,$49,$FC,$EF,$83,$E7,$E3,$F9,$7C,$3F,$93,$9F,$C0,$F1
	.dc.b	$7E,$1E,$7F,$61,$1F,$87,$6B,$F8,$76,$F1,$FE,$3E,$38,$92,$33,$3C
	.dc.b	$FC,$6F,$DC,$4F,$16,$D7,$1A,$DC,$F1,$9C,$D7,$F8,$2F,$C0,$FC,$9C
	.dc.b	$FF,$82,$7C,$0F,$C9,$BF,$B8,$F0,$3F,$26,$FE,$EB,$BF,$F2,$99,$7E
	.dc.b	$FB,$BF,$F2,$53,$FB,$EE,$EF,$C9,$3F,$98,$EE,$FC,$93,$FC,$0B,$B7
	.dc.b	$F2,$4F,$C1,$D7,$D3,$E3,$FF,$45,$EB,$FF,$70,$4F,$DD,$F8,$6F,$B4
	.dc.b	$47,$04,$3F,$00,$E5,$A6,$73,$FF,$20,$4F,$48,$C9,$23,$33,$C5,$B1
	.dc.b	$C7,$83,$FC,$BD,$5C,$67,$33,$F8,$FF,$77,$F1,$DE,$EF,$E3,$FE,$1B
	.dc.b	$F2,$1E,$37,$E4,$7E,$1B,$F2,$5E,$EF,$E4,$FE,$1A,$63,$F9,$9F,$53
	.dc.b	$F2,$DE,$A7,$E5,$FE,$1B,$F3,$1F,$0D,$F9,$9F,$86,$FC,$D7,$AC,$C3
	.dc.b	$FD,$89,$F8,$77,$F9,$0B,$BC,$FF,$90,$B8,$BC,$A7,$9B,$9B,$E6,$73
	.dc.b	$7B,$79,$BD,$6D,$C3,$FC,$D3,$CD,$ED,$9C,$C7,$C1,$33,$3B,$C5,$5B
	.dc.b	$9C,$9B,$8C,$CE,$E7,$51,$89,$D4,$60,$CE,$3F,$E5,$8C,$7E,$BF,$2F
	.dc.b	$CB,$AF,$7F,$AD,$F0,$5F,$F3,$9D,$7F,$91,$1F,$87,$F0,$FF,$31,$FE
	.dc.b	$05,$C7,$BC,$3F,$D0,$14,$F6,$BE,$7D,$7F,$09,$E2,$D6,$DF,$E9,$BA
	.dc.b	$7B,$5F,$46,$BE,$7F,$C3,$E9,$FF,$47,$CB,$DB,$C7,$07,$D2,$F8,$6F
	.dc.b	$81,$C5,$FC,$06,$3E,$AF,$C0,$6A,$FE,$B6,$7D,$E3,$04,$C7,$EE,$B5
	.dc.b	$F7,$B7,$6F,$5B,$67,$94,$F0,$21,$C0,$B3,$C1,$B1,$C1,$FC,$D1,$F8
	.dc.b	$CF,$86,$F3,$7C,$37,$E6,$A3,$F9,$77,$6F,$E6,$9F,$C8,$FC,$37,$AF
	.dc.b	$E1,$BF,$97,$77,$7B,$7E,$1B,$DD,$F0,$DF,$9A,$BF,$CB,$FC,$37,$E6
	.dc.b	$9F,$CC,$9E,$93,$E5,$ED,$EF,$ED,$FC,$BC,$7E,$1B,$8B,$E5,$B4,$87
	.dc.b	$EE,$77,$FC,$BB,$4F,$C9,$97,$8B,$F2,$0D,$ED,$FC,$85,$7E,$6F,$8F
	.dc.b	$F2,$17,$FB,$EE,$DF,$C8,$5F,$E3,$BB,$7E,$C7,$C2,$FD,$9E,$EF,$C7
	.dc.b	$73,$FF,$0C,$EE,$FB,$9D,$DF,$8C,$C3,$FA,$97,$77,$DF,$EE,$FC,$54
	.dc.b	$7E,$63,$BB,$F0,$5D,$DF,$83,$EE,$F6,$CF,$FA,$30,$FC,$6F,$C0,$FE
	.dc.b	$1C,$8F,$F0,$06,$BF,$A2,$6E,$7F,$9E,$6F,$07,$E6,$DB,$3D,$0E,$44
	.dc.b	$39,$16,$79,$36,$79,$3F,$C4,$3D,$0F,$CD,$7C,$2F,$E6,$FD,$1F,$D0
	.dc.b	$FE,$1F,$FE,$41,$BF,$E2,$DD,$08,$FF,$21,$61,$FE,$EA,$7F,$C8,$4F
	.dc.b	$F1,$BC,$7F,$DE,$91,$FE,$1D,$8C,$57,$28,$C4,$71,$F6,$3B,$FC,$37
	.dc.b	$83,$D9,$F0,$F5,$F8,$BE,$1B,$C7,$EC,$F9,$3E,$1B,$E0,$EB,$FC,$9B
	.dc.b	$87,$E4,$F7,$FF,$A4,$F5,$7B,$5E,$A2,$6D,$E2,$17,$EB,$8C,$C7,$AD
	.dc.b	$BC,$1E,$A6,$F0,$7A,$5B,$3C,$AF,$BB,$D6,$FA,$36,$32,$F5,$31,$3C
	.dc.b	$A7,$71,$93,$F9,$1F,$1F,$D4,$7F,$A2,$F1,$FD,$4F,$67,$EA,$FB,$FF
	.dc.b	$5B,$E1,$BE,$BF,$B7,$2F,$C0,$FC,$C7,$C5,$C9,$F3,$1F,$27,$B7,$F7
	.dc.b	$BE,$1B,$E6,$47,$9F,$DC,$F2,$6F,$E8,$E2,$FF,$5F,$78,$DF,$E0,$EA
	.dc.b	$FC,$81,$F8,$E2,$71,$FE,$39,$BC,$7F,$8E,$69,$C2,$34,$48,$D3,$63
	.dc.b	$51,$82,$38,$27,$17,$E0,$5F,$9D,$EE,$7E,$63,$E1,$BF,$35,$1F,$46
	.dc.b	$BF,$37,$EE,$FE,$73,$E1,$BF,$3B,$EF,$FF,$3E,$FC,$43,$FC,$9F,$9F
	.dc.b	$E6,$3C,$0F,$D0,$99,$F8,$6D,$7C,$37,$E4,$F3,$FB,$9D,$FF,$94,$7F
	.dc.b	$58,$79,$38,$3F,$C4,$3E,$FF,$E1,$E3,$EF,$90,$FB,$D8,$FE,$35,$C3
	.dc.b	$5C,$ED,$C5,$A7,$E3,$FB,$BE,$A9,$1E,$31,$B9,$E4,$47,$FB,$17,$77
	.dc.b	$E2,$E7,$FB,$7F,$77,$E2,$DF,$CB,$76,$FE,$2D,$FC,$49,$0F,$DF,$76
	.dc.b	$7E,$2D,$FE,$01,$D9,$F8,$B7,$F8,$17,$67,$E2,$DF,$D0,$78,$1F,$C2
	.dc.b	$BB,$3C,$D1,$FC,$8F,$A9,$CA,$F8,$6F,$A4,$49,$FD,$54,$65,$B1,$95
	.dc.b	$BB,$7F,$E2,$B5,$B9,$CC,$62,$B8,$56,$E7,$F8,$0D,$6E,$31,$7E,$4A
	.dc.b	$DD,$F1,$BF,$09,$0E,$A7,$28,$DB,$4D,$57,$4B,$7C,$37,$E3,$FE,$1B
	.dc.b	$F2,$1F,$0D,$F9,$1F,$53,$F2,$5F,$0B,$F9,$3F,$57,$F2,$9F,$0B,$F9
	.dc.b	$5E,$89,$A7,$C4,$F7,$BF,$31,$F0,$BF,$99,$F5,$7F,$35,$EE,$FE,$6F
	.dc.b	$E1,$7F,$39,$EB,$79,$39,$7D,$98,$FA,$7C,$96,$7F,$23,$E1,$7F,$44
	.dc.b	$49,$FA,$2D,$7D,$F5,$F6,$23,$1C,$BE,$D6,$F3,$F3,$AE,$24,$38,$F3
	.dc.b	$27,$71,$F4,$9B,$7D,$63,$85,$F5,$7F,$07,$CD,$FB,$7F,$85,$F9,$31
	.dc.b	$F8,$6F,$85,$F9,$2F,$E2,$39,$3F,$C7,$3E,$C7,$F9,$AF,$3F,$7C,$FE
	.dc.b	$33,$5C,$FB,$DF,$2A,$3F,$1F,$C9,$C3,$E2,$7E,$54,$7F,$3B,$C1,$EF
	.dc.b	$F4,$F7,$FC,$F9,$7A,$EB,$FD,$E0,$FE,$2F,$8B,$F1,$11,$F0,$3B,$BE
	.dc.b	$94,$E6,$3F,$0C,$64,$FC,$3C,$F0,$E1,$FC,$45,$70,$24,$E8,$E3,$1A
	.dc.b	$71,$C5,$F7,$9F,$E9,$27,$EA,$BB,$BE,$1F,$BF,$F8,$F8,$FD,$6F,$77
	.dc.b	$C6,$F7,$FF,$1F,$1E,$FF,$7F,$E4,$EF,$FC,$97,$BB,$E5,$FB,$FF,$8F
	.dc.b	$AF,$DA,$F7,$7C,$EF,$7F,$F1,$F1,$E2,$E1,$F7,$6D,$E3,$ED,$FA,$2F
	.dc.b	$DD,$9F,$C1,$F6,$FC,$A2,$3F,$80,$E3,$FC,$53,$A3,$07,$E1,$CB,$3C
	.dc.b	$F8,$BF,$0D,$1B,$23,$82,$4F,$47,$F6,$9D,$BF,$8B,$9F,$E1,$FE,$1F
	.dc.b	$F4,$3E,$DF,$C5,$C7,$E6,$3B,$7F,$16,$FE,$C8,$FC,$DF,$6F,$E2,$9F
	.dc.b	$C4,$90,$FD,$3F,$17,$F9,$49,$FE,$1F,$E0,$7E,$53,$AF,$F1,$51,$EA
	.dc.b	$E0,$FC,$53,$FC,$96,$9F,$0D,$F9,$A2,$47,$E4,$DA,$F6,$6E,$BF,$C6
	.dc.b	$A7,$8C,$EC,$86,$CB,$1B,$9E,$3B,$7F,$8F,$AB,$8D,$6C,$95,$B7,$8C
	.dc.b	$EF,$E1,$BF,$5F,$F0,$DF,$B0,$F8,$6F,$D8,$FC,$37,$EC,$BE,$1B,$F6
	.dc.b	$7F,$0D,$FB,$4F,$82,$FD,$AF,$A7,$FB,$6F,$4F,$F6,$FF,$0D,$35,$FF
	.dc.b	$15,$7E,$23,$FE,$51,$7F,$C6,$EF,$F9,$43,$7F,$E6,$EF,$F9,$42,$3F
	.dc.b	$8A,$BF,$E4,$C8,$CF,$B7,$EC,$9F,$F1,$FF,$0F,$B3,$B7,$F2,$76,$F8
	.dc.b	$7F,$F5,$FB,$6B,$C3,$18,$23,$82,$1E,$07,$2D,$33,$87,$9A,$32,$D8
	.dc.b	$CF,$1F,$F9,$B2,$BC,$11,$C4,$93,$B7,$DB,$EA,$F9,$BE,$1F,$E7,$47
	.dc.b	$C6,$F8,$7F,$9C,$FF,$17,$E2,$F5,$FB,$9F,$3A,$3F,$8D,$70,$7C,$D7
	.dc.b	$F2,$BC,$1F,$03,$4F,$83,$ED,$7C,$2F,$86,$F8,$7E,$DF,$C4,$F6,$BD
	.dc.b	$35,$E7,$E1,$FF,$71,$71,$FE,$1B,$87,$F0,$99,$FE,$1A,$BF,$15,$D9
	.dc.b	$F4,$7C,$55,$EE,$73,$5F,$8A,$3E,$AD,$6E,$75,$18,$35,$5C,$A7,$44
	.dc.b	$8C,$CB,$B7,$F2,$71,$F8,$AF,$73,$F2,$6F,$E3,$3E,$1F,$F2,$6F,$F3
	.dc.b	$DE,$BF,$C0,$FB,$BF,$82,$F8,$7F,$C1,$FB,$DF,$84,$F8,$6F,$C2,$FC
	.dc.b	$37,$E1,$BE,$1B,$F0,$FF,$0D,$F8,$8F,$86,$FC,$4F,$85,$E3,$D3,$FC
	.dc.b	$3D,$E0,$FD,$CE,$5F,$B4,$4E,$5F,$95,$3F,$82,$3F,$3B,$CD,$F3,$9A
	.dc.b	$67,$99,$7C,$EB,$8C,$FE,$19,$B7,$D6,$38,$99,$2B,$9A,$E7,$7F,$A0
	.dc.b	$F0,$7F,$19,$5F,$D7,$FC,$6F,$D1,$F8,$3F,$8C,$8F,$E3,$7E,$0F,$E3
	.dc.b	$1F,$EC,$9E,$27,$EE,$5F,$EC,$DE,$07,$E3,$73,$FD,$AE,$9F,$8C,$9F
	.dc.b	$DD,$F7,$FE,$31,$FD,$EF,$77,$BF,$6F,$1C,$0F,$F2,$37,$E2,$5F,$E7
	.dc.b	$4B,$FF,$68,$B6,$3F,$D7,$2D,$93,$49,$F1,$BC,$E7,$FC,$33,$18,$7F
	.dc.b	$D6,$CD,$73,$1B,$C7,$FC,$87,$39,$6C,$66,$FC,$6D,$AF,$1B,$6F,$C6
	.dc.b	$DB,$E5,$5E,$AF,$88,$C7,$C4,$67,$DE,$D7,$A9,$BF,$7A,$67,$E6,$F5
	.dc.b	$79,$7B,$BC,$FE,$1F,$A7,$AD,$D7,$E1,$FB,$7A,$FD,$FD,$DF,$07,$C1
	.dc.b	$F8,$7E,$1F,$C5,$EF,$78,$F9,$26,$BF,$3F,$D5,$FE,$EB,$CB,$E6,$8F
	.dc.b	$59,$F1,$A7,$6F,$C6,$30,$43,$04,$E7,$F8,$EE,$5A,$E7,$9B,$E4,$B9
	.dc.b	$6B,$93,$C2,$7D,$8F,$6F,$E5,$7C,$3F,$C7,$8F,$B3,$CB,$F1,$9F,$BB
	.dc.b	$CB,$F3,$FD,$EF,$A1,$F0,$FF,$1E,$7E,$FF,$BF,$F1,$DF,$F7,$B7,$E2
	.dc.b	$7F,$85,$7F,$0B,$C7,$F2,$1C,$FC,$47,$C8,$7C,$C7,$93,$87,$FC,$37
	.dc.b	$5F,$DF,$7B,$8D,$27,$C4,$7E,$0D,$B5,$E0,$8C,$11,$C1,$0F,$F3,$7C
	.dc.b	$70,$6C,$6D,$A6,$F8,$7F,$D3,$33,$B7,$F3,$3D,$FF,$8E,$9F,$7F,$C1
	.dc.b	$78,$FE,$23,$C9,$F1,$1E,$5F,$88,$FC,$B7,$83,$F8,$EB,$F8,$BF,$11
	.dc.b	$F8,$E7,$E3,$FB,$9F,$8E,$7E,$4F,$C4,$7B,$7E,$23,$F1,$D1,$F3,$3D
	.dc.b	$FF,$36,$DE,$4E,$0F,$D9,$71,$7E,$3B,$B7,$F1,$73,$F8,$42,$76,$FE
	.dc.b	$09,$B8,$7E,$09,$BC,$9F,$AC,$8F,$A6,$D7,$F1,$0D,$8D,$F7,$7E,$6E
	.dc.b	$31,$5D,$49,$5F,$C2,$BB,$BF,$1A,$FF,$04,$F1,$FF,$66,$FF,$06,$ED
	.dc.b	$FC,$76,$3F,$C1,$3B,$7F,$1B,$3F,$C2,$BC,$6F,$DF,$76,$FE,$36,$3E
	.dc.b	$C9,$FB,$33,$F8,$7F,$87,$F8,$3E,$DF,$C6,$47,$A3,$66,$FF,$1B,$C4
	.dc.b	$FE,$45,$C1,$F8,$CF,$7F,$F1,$44,$BF,$CC,$B5,$FC,$3B,$75,$FC,$1B
	.dc.b	$78,$79,$F7,$FF,$65,$AE,$24,$38,$96,$B8,$B6,$78,$D7,$38,$5F,$95
	.dc.b	$FD,$5F,$C3,$FE,$7F,$D1,$FD,$07,$C3,$FE,$87,$D1,$FD,$17,$C3,$FE
	.dc.b	$8F,$E0,$3F,$49,$E9,$FE,$97,$E1,$7F,$4D,$EA,$7F,$29,$E7,$FF,$20
	.dc.b	$6A,$7F,$C8,$1B,$FC,$47,$FC,$81,$C1,$FE,$F8,$7F,$C8,$11,$9F,$67
	.dc.b	$97,$B1,$CF,$D8,$F4,$61,$E7,$E2,$77,$FE,$EB,$93,$3C,$BF,$16,$7E
	.dc.b	$19,$34,$F8,$6D,$CF,$E1,$BF,$14,$F8,$73,$C6,$76,$49,$DC,$F1,$37
	.dc.b	$5E,$02,$3F,$28,$FE,$2B,$C9,$F2,$9F,$E2,$FC,$7F,$17,$E1,$FE,$54
	.dc.b	$7F,$67,$C3,$E1,$7B,$3F,$27,$E1,$FE,$5C,$FF,$1E,$E2,$F9,$9E,$CF
	.dc.b	$CD,$F8,$5F,$9D,$F0,$BF,$3F,$E1,$7E,$87,$C2,$FD,$1F,$85,$F1,$E3
	.dc.b	$F1,$BB,$FE,$87,$C2,$FD,$52,$47,$87,$8F,$2D,$3F,$0B,$D4,$9D,$C3
	.dc.b	$83,$EC,$46,$1B,$1C,$79,$FF,$01,$39,$BE,$B3,$C5,$B5,$C2,$30,$78
	.dc.b	$27,$83,$4E,$97,$E3,$6C,$7F,$75,$F2,$7F,$21,$F0,$BF,$91,$F8,$5F
	.dc.b	$C9,$7C,$2F,$E6,$2F,$8F,$C2,$FE,$61,$FE,$93,$E0,$7F,$7F,$FC,$53
	.dc.b	$FC,$25,$9F,$E6,$0F,$CB,$CF,$C2,$F8,$5F,$CE,$7C,$3F,$E7,$7C,$3F
	.dc.b	$CA,$CF,$93,$C2,$F3,$F1,$7C,$08,$FE,$E9,$C7,$F8,$8E,$2F,$C5,$47
	.dc.b	$E2,$0F,$C8,$4F,$E6,$E3,$2D,$E3,$E0,$49,$E0,$7E,$46,$34,$E2,$78
	.dc.b	$F8,$3F,$99,$AE,$15,$A2,$46,$6B,$F3,$FE,$47,$EF,$DF,$ED,$7E,$47
	.dc.b	$EF,$DF,$9F,$E0,$7E,$59,$FA,$B7,$FB,$BF,$03,$F2,$D1,$F9,$52,$1A
	.dc.b	$E0,$77,$7E,$5E,$3E,$CF,$6F,$E5,$DF,$B7,$D9,$F9,$77,$E1,$E9,$FB
	.dc.b	$9F,$11,$C1,$FC,$AB,$95,$EE,$7F,$7F,$BC,$35,$FE,$37,$1A,$75,$AA
	.dc.b	$31,$1F,$E0,$78,$C4,$F1,$6C,$70,$9D,$BF,$E5,$18,$CB,$C6,$32,$F8
	.dc.b	$8E,$13,$CE,$33,$1E,$48,$C1,$1D,$90,$E8,$E5,$ED,$F0,$DF,$95,$F4
	.dc.b	$BF,$2D,$F0,$1F,$97,$F4,$FF,$31,$ED,$7E,$67,$E1,$BF,$35,$C5,$F9
	.dc.b	$BF,$86,$99,$7E,$CF,$47,$C3,$7E,$7F,$7F,$D0,$7C,$17,$E8,$7D,$6F
	.dc.b	$D1,$7C,$37,$E8,$FE,$1B,$F4,$9E,$47,$E9,$78,$FC,$9E,$1F,$DC,$D3
	.dc.b	$D8,$7A,$4F,$8D,$A7,$C0,$E5,$F8,$F1,$8A,$F6,$34,$E0,$7C,$93,$E0
	.dc.b	$98,$3A,$98,$21,$C4,$A7,$13,$04,$3F,$11,$CB,$F2,$5F,$85,$CB,$FB
	.dc.b	$FF,$86,$FE,$01,$ED,$FF,$01,$F8,$6F,$93,$5F,$1F,$8F,$E4,$BF,$A0
	.dc.b	$CD,$EE,$7F,$A1,$7F,$15,$FF,$28,$78,$AF,$17,$F2,$DC,$B9,$F7,$75
	.dc.b	$EF,$7C,$BC,$BF,$92,$60,$F8,$7F,$99,$1F,$81,$E4,$F9,$9C,$9F,$22
	.dc.b	$3D,$84,$3E,$19,$ED,$29,$97,$83,$82,$38,$26,$BF,$54,$9C,$9F,$52
	.dc.b	$79,$46,$1F,$D4,$E0,$F8,$7F,$81,$F0,$FF,$92,$9F,$C6,$78,$1F,$92
	.dc.b	$7F,$1D,$DF,$F9,$37,$DF,$F0,$FF,$1F,$E1,$FF,$97,$78,$1F,$91,$9F
	.dc.b	$CA,$78,$1F,$91,$7F,$2D,$DF,$F9,$27,$F3,$1D,$DF,$19,$FC,$17,$0F
	.dc.b	$C6,$DF,$CD,$1F,$82,$3D,$EE,$1F,$8D,$C3,$F0,$38,$7D,$D1,$A6,$C6
	.dc.b	$63,$F2,$0D,$7A,$3F,$C2,$7B,$BF,$15,$1F,$83,$EE,$FC,$53,$FA,$DE
	.dc.b	$DF,$C5,$3F,$AF,$F0,$7F,$13,$D7,$F8,$58,$FE,$2B,$DB,$FD,$3F,$B7
	.dc.b	$F1,$51,$FB,$23,$F0,$BD,$7F,$88,$7F,$71,$E0,$7E,$3F,$AF,$E1,$47
	.dc.b	$E8,$78,$3F,$0F,$D6,$F8,$7F,$E3,$4D,$9F,$B3,$18,$23,$82,$1F,$C8
	.dc.b	$E3,$83,$63,$6D,$36,$7F,$23,$8D,$B6,$36,$D3,$67,$F9,$8A,$36,$D8
	.dc.b	$F3,$FC,$3F,$EA,$FE,$1E,$4F,$E9,$BE,$1F,$F5,$DF,$03,$FA,$FF,$46
	.dc.b	$4F,$EA,$BE,$07,$F6,$47,$EC,$FD,$2F,$DA,$7C,$0C,$9F,$D7,$FC,$2F
	.dc.b	$ED,$FE,$16,$47,$EC,$BE,$17,$F7,$5B,$7B,$B8,$3E,$5B,$EA,$9F,$31
	.dc.b	$EA,$3F,$CC,$FC,$1E,$A6,$C6,$DA,$6F,$AD,$3B,$6C,$ED,$B3,$B9,$D1
	.dc.b	$27,$4E,$3D,$AF,$8B,$3F,$DC,$A1,$C3,$FC,$4F,$E0,$D3,$FE,$31,$FC
	.dc.b	$5B,$FC,$E2,$7F,$86,$B1,$3F,$E7,$0F,$37,$EA,$70,$FC,$AA,$E1,$EE
	.dc.b	$72,$7F,$BF,$D3,$DC,$C7,$6F,$E6,$F1,$E7,$F0,$DF,$C8,$F8,$F3,$9F
	.dc.b	$FA,$83,$07,$BB,$8C,$BF,$D1,$F0,$F7,$71,$8F,$93,$DD,$FA,$75,$F9
	.dc.b	$4E,$4F,$A6,$FC,$5E,$4F,$86,$47,$04,$3F,$2C,$43,$F2,$45,$3E,$D9
	.dc.b	$4F,$AE,$5A,$FA,$FD,$9E,$4C,$5F,$0D,$7D,$EF,$E8,$1D,$DF,$94,$C3
	.dc.b	$F9,$DF,$6F,$E5,$1F,$8F,$F0,$DC,$7D,$FF,$CA,$47,$F4,$0E,$EF,$CA
	.dc.b	$3F,$33,$E1,$BF,$26,$FD,$BE,$FF,$E9,$9D,$FF,$0A,$3D,$7C,$3F,$09
	.dc.b	$F3,$77,$7E,$15,$F8,$07,$B8,$87,$B8,$A7,$B8,$BC,$3F,$90,$79,$F1
	.dc.b	$7C,$B6,$9F,$D0,$3B,$FF,$1F,$D5,$F7,$EF,$F1,$F3,$FD,$6B,$BB,$E3
	.dc.b	$74,$FC,$7F,$8A,$F9,$19,$FE,$2E,$BF,$2D,$DF,$F2,$FB,$FE,$67,$7F
	.dc.b	$CD,$EF,$FC,$45,$7E,$5F,$BB,$E8,$77,$7D,$1E,$EF,$85,$3F,$6B,$2F
	.dc.b	$EC,$BD,$AF,$81,$FF,$1C,$36,$7F,$C7,$ED,$7F,$C7,$CD,$7F,$C7,$4D
	.dc.b	$7F,$2D,$1C,$88,$F2,$6B,$C9,$AF,$26,$BC,$8E,$04,$38,$14,$F0,$FC
	.dc.b	$3F,$E3,$7E,$1E,$47,$ED,$FE,$1F,$F2,$1F,$0F,$F9,$1F,$87,$93,$FB
	.dc.b	$BF,$87,$FD,$3E,$8F,$88,$BE,$A6,$3D,$4C,$FA,$9A,$F8,$6D,$FB,$D3
	.dc.b	$2F,$F5,$A6,$2F,$86,$FC,$F7,$0F,$BB,$8F,$E6,$BE,$CA,$FD,$BB,$7C
	.dc.b	$1F,$E1,$2D,$E4,$F9,$53,$A3,$DC,$7C,$33,$E0,$90,$F9,$06,$8E,$06
	.dc.b	$4E,$06,$8C,$09,$60,$F7,$FE,$4C,$FF,$16,$E4,$FD,$C7,$27,$CA,$8F
	.dc.b	$7F,$AF,$F2,$3D,$7F,$93,$EE,$FC,$AE,$49,$5F,$13,$8F,$E6,$FB,$BF
	.dc.b	$26,$3E,$3F,$C1,$FD,$0F,$86,$FA,$3C,$7F,$22,$7E,$F7,$1F,$C8,$7D
	.dc.b	$9C,$3F,$19,$BC,$5F,$1A,$31,$1E,$C6,$BE,$F2,$1F,$08,$87,$C6,$32
	.dc.b	$53,$2F,$26,$C7,$71,$D7,$F8,$8F,$7F,$F1,$3E,$D7,$E2,$BD,$FF,$C5
	.dc.b	$FC,$37,$E3,$B0,$FE,$7F,$DD,$FC,$87,$BB,$F1,$FE,$FF,$E3,$A7,$F1
	.dc.b	$3D,$FF,$8E,$7F,$17,$F0,$DF,$8D,$7F,$D8,$BD,$8E,$EF,$CB,$FC,$07
	.dc.b	$C7,$9F,$C1,$F1,$7C,$77,$E1,$70,$F9,$FB,$FF,$18,$48,$F9,$2D,$76
	.dc.b	$D7,$E1,$F1,$FC,$17,$E2,$1F,$87,$C7,$F0,$CF,$F6,$3E,$EF,$C3,$F9
	.dc.b	$BF,$D6,$70,$77,$FE,$1E,$BD,$3D,$FF,$87,$7F,$4F,$DD,$F8,$67,$DB
	.dc.b	$DD,$FA,$F7,$F5,$7D,$BF,$87,$7E,$0F,$6F,$E1,$DF,$D7,$F6,$7C,$17
	.dc.b	$F3,$9C,$1F,$E7,$EF,$C5,$FF,$DB,$11,$F9,$D6,$F0,$1A,$48,$FF,$6B
	.dc.b	$35,$FF,$6E,$B5,$FF,$6E,$35,$FF,$6C,$35,$FF,$02,$C7,$57,$99,$0E
	.dc.b	$65,$39,$9E,$0F,$81,$FE,$EF,$26,$FC,$3E,$3E,$0B,$3F,$0F,$2F,$FC
	.dc.b	$27,$C5,$FE,$07,$C1,$F0,$FC,$BE,$1F,$9F,$C3,$F4,$F5,$FA,$FC,$17
	.dc.b	$6F,$84,$EF,$F0,$33,$3E,$7E,$BF,$AB,$37,$C3,$79,$27,$D0,$7C,$03
	.dc.b	$FC,$35,$C7,$F3,$DC,$9C,$4A,$7D,$12,$F1,$FC,$57,$E1,$90,$F8,$65
	.dc.b	$3B,$14,$E8,$53,$A1,$5E,$8F,$23,$EB,$F1,$FC,$C2,$3E,$E3,$F0,$9C
	.dc.b	$5F,$43,$E0,$BE,$8F,$B1,$F4,$B8,$BE,$9F,$C3,$7C,$AB,$E1,$8F,$E2
	.dc.b	$7E,$07,$E5,$C6,$35,$FE,$F2,$B9,$7E,$A7,$0F,$DC,$F8,$6F,$BB,$C1
	.dc.b	$F2,$EB,$D1,$7F,$E4,$35,$C1,$F2,$DC,$9F,$22,$3F,$1C,$FC,$03,$DC
	.dc.b	$7C,$62,$9F,$18,$A6,$CD,$10,$D1,$4D,$18,$3F,$37,$E1,$7E,$F9,$F6
	.dc.b	$71,$7E,$17,$E0,$7F,$2B,$1F,$07,$E0,$7F,$28,$E8,$FD,$87,$56,$35
	.dc.b	$FC,$B7,$57,$E7,$BE,$07,$F2,$95,$F2,$7E,$03,$FC,$AF,$F8,$C7,$FA
	.dc.b	$EE,$39,$71,$FE,$DF,$AB,$F2,$D1,$8D,$D1,$F2,$E3,$DF,$3F,$82,$E0
	.dc.b	$F9,$51,$8A,$F9,$33,$F8,$68,$F8,$0D,$7E,$01,$F0,$88,$E5,$B1,$C8
	.dc.b	$8F,$D5,$3F,$74,$7E,$0F,$AB,$F1,$EF,$9F,$AB,$F1,$CF,$E1,$FA,$7F
	.dc.b	$1C,$FE,$27,$A3,$F1,$CF,$B7,$A3,$F1,$AF,$BF,$A3,$F7,$0F,$E3,$FA
	.dc.b	$3F,$16,$FE,$44,$F1,$1F,$E9,$6D,$1F,$0D,$FC,$53,$77,$C1,$7E,$3E
	.dc.b	$B0,$D7,$FC,$05,$1A,$75,$BA,$31,$1F,$97,$8C,$4F,$16,$C7,$09,$DB
	.dc.b	$FE,$9F,$8C,$BC,$63,$2F,$88,$E1,$3C,$E3,$31,$E4,$8C,$11,$D9,$0E
	.dc.b	$8E,$5E,$D7,$F8,$CF,$82,$91,$FB,$AA,$FF,$58,$37,$E0,$7F,$22,$47
	.dc.b	$5F,$03,$F9,$4F,$81,$FC,$AF,$6F,$E5,$BE,$1F,$F2,$FA,$CB,$E9,$F0
	.dc.b	$FF,$9A,$3F,$37,$AF,$8E,$3E,$16,$7F,$2B,$5F,$AB,$1E,$D3,$D4,$6A
	.dc.b	$78,$3F,$14,$F6,$9A,$30,$6C,$C1,$CC,$D9,$0D,$14,$E2,$43,$81,$A3
	.dc.b	$CF,$F0,$FF,$21,$FD,$9D,$36,$FD,$BF,$C3,$FE,$E3,$D2,$FD,$CF,$BD
	.dc.b	$FB,$AF,$4F,$F7,$7C,$3F,$BC,$F4,$BE,$4E,$5F,$47,$E1,$7E,$43,$D3
	.dc.b	$83,$F8,$0F,$B7,$FC,$0B,$DB,$F9,$33,$EC,$E7,$FC,$4F,$17,$C8,$7F
	.dc.b	$17,$1E,$73,$D4,$E1,$BC,$1F,$8A,$8C,$47,$B8,$CF,$17,$CE,$70,$48
	.dc.b	$D9,$27,$91,$27,$84,$7E,$96,$BF,$DD,$FF,$8C,$FF,$A2,$6F,$FA,$3A
	.dc.b	$F6,$FF,$8E,$FC,$3F,$E3,$67,$E5,$F5,$5F,$1F,$FE,$00,$5B,$E0,$FF
	.dc.b	$21,$5C,$BE,$1F,$F2,$0F,$4F,$87,$EF,$EF,$7E,$3E,$3F,$22,$7E,$77
	.dc.b	$6F,$2E,$FF,$8F,$E2,$F7,$CF,$CB,$8F,$53,$67,$64,$70,$43,$F1,$07
	.dc.b	$C5,$8E,$2F,$EE,$FA,$FF,$0D,$A3,$CF,$FE,$21,$D9,$F8,$6A,$FD,$87
	.dc.b	$5F,$E1,$9F,$D9,$75,$7E,$19,$FD,$A7,$4F,$E1,$9F,$C7,$F4,$FE,$1B
	.dc.b	$1E,$91,$FC,$5B,$5F,$C3,$BF,$BB,$E8,$FC,$3B,$E6,$E8,$FC,$3C,$FF
	.dc.b	$8D,$39,$3F,$D6,$93,$F5,$3D,$7F,$AE,$47,$04,$3F,$A8,$C7,$06,$C6
	.dc.b	$DA,$6F,$0F,$B3,$19,$6C,$67,$6F,$14,$E5,$B1,$F8,$6F,$87,$FE,$C3
	.dc.b	$C1,$F8,$AF,$77,$F1,$7F,$0F,$F8,$CF,$7B,$F1,$BE,$EF,$E3,$BD,$DF
	.dc.b	$C7,$FC,$3C,$CB,$F1,$FF,$0F,$F9,$2F,$87,$FC,$9F,$C3,$FE,$53,$E1
	.dc.b	$FF,$2B,$F0,$FF,$96,$F8,$7F,$CB,$FC,$3F,$E6,$38,$7C,$99,$7F,$B0
	.dc.b	$F9,$7E,$07,$C5,$37,$FF,$2E,$FC,$1F,$F8,$5F,$9F,$E9,$F4,$A3,$1C
	.dc.b	$FA,$C4,$7F,$8A,$23,$2F,$FB,$3D,$B3,$C6,$36,$7F,$80,$23,$24,$73
	.dc.b	$5E,$08,$C4,$75,$33,$5E,$17,$04,$3A,$1A,$3C,$45,$AF,$13,$E7,$F6
	.dc.b	$31,$4F,$86,$BE,$CE,$3D,$FC,$FC,$1E,$BE,$03,$7F,$0D,$F4,$F2,$FC
	.dc.b	$4F,$4F,$2F,$86,$FA,$91,$F0,$F3,$EB,$ED,$FD,$48,$FE,$0F,$CF,$E0
	.dc.b	$F8,$0F,$0E,$FE,$29,$87,$CC,$E4,$FA,$8F,$D7,$F0,$BE,$CF,$13,$9B
	.dc.b	$F3,$1C,$BF,$92,$AF,$CB,$91,$C1,$39,$7F,$27,$19,$6C,$66,$3B,$1F
	.dc.b	$9B,$29,$B2,$BB,$AF,$17,$3F,$E7,$27,$F9,$1F,$83,$F9,$C7,$F9,$2F
	.dc.b	$81,$F9,$C7,$F9,$3F,$81,$F4,$3E,$17,$E8,$FC,$37,$F2,$CF,$0B,$E9
	.dc.b	$FC,$0F,$E7,$2F,$F9,$6F,$83,$F5,$BE,$07,$F3,$91,$FC,$C3,$C1,$FC
	.dc.b	$E3,$E7,$DB,$E2,$71,$7E,$3B,$8F,$E4,$C7,$E0,$27,$F0,$A7,$E1,$F8
	.dc.b	$7F,$1C,$FD,$03,$24,$72,$EF,$8B,$F2,$0E,$5B,$1A,$8C,$1F,$90,$38
	.dc.b	$14,$CC,$FE,$AF,$93,$F2,$3E,$07,$E4,$27,$F5,$5D,$FF,$90,$7F,$59
	.dc.b	$DD,$F9,$07,$FA,$7F,$6F,$E4,$71,$FD,$37,$6F,$E4,$27,$F5,$9D,$BF
	.dc.b	$90,$7F,$1E,$43,$F6,$1D,$7F,$90,$7C,$BA,$7C,$5D,$BE,$3F,$5F,$E3
	.dc.b	$37,$FC,$29,$2B,$F0,$2D,$AF,$9F,$1F,$46,$78,$3A,$24,$69,$FC,$53
	.dc.b	$72,$FC,$54,$F0,$8C,$57,$58,$C1,$FE,$95,$9F,$E2,$7E,$73,$C7,$FE
	.dc.b	$55,$87,$F7,$2F,$1F,$F9,$53,$FD,$D3,$C4,$FE,$54,$FC,$82,$1F,$D0
	.dc.b	$3A,$7E,$46,$1F,$DD,$3C,$6F,$E5,$53,$F1,$E3,$FB,$BF,$8D,$FC,$AA
	.dc.b	$3F,$9D,$1E,$FE,$7F,$C7,$61,$F8,$2D,$64,$FD,$FB,$FC,$21,$23,$3B
	.dc.b	$7E,$4E,$31,$3B,$9C,$F0,$FF,$20,$8F,$C3,$B6,$77,$5C,$6B,$7C,$FF
	.dc.b	$8A,$BF,$E6,$5E,$37,$F1,$97,$F9,$A7,$8D,$FC,$65,$FE,$6D,$E3,$7E
	.dc.b	$26,$FF,$B4,$73,$FF,$19,$9F,$F0,$DF,$8F,$FC,$93,$4F,$C5,$47,$F6
	.dc.b	$6F,$1B,$F8,$D3,$F8,$A3,$F9,$51,$F8,$A8,$FF,$6F,$73,$FE,$2B,$83
	.dc.b	$F8,$BF,$3F,$DA,$69,$2B,$E5,$10,$FC,$2C,$6D,$C7,$07,$E1,$BC,$6F
	.dc.b	$E1,$D1,$FD,$E3,$C6,$FE,$1C,$FF,$7A,$F1,$BF,$87,$3F,$DF,$3C,$6F
	.dc.b	$E1,$CF,$F5,$5F,$17,$F8,$73,$FD,$5F,$C5,$FE,$1C,$FF,$80,$BC,$5F
	.dc.b	$E1,$CF,$F8,$F3,$A7,$F0,$0F,$F8,$1B,$2F,$42,$DB,$F8,$6E,$FF,$57
	.dc.b	$8F,$EC,$CF,$F0,$BD,$BE,$BC,$E6,$3E,$F1,$3A,$3E,$D3,$63,$51,$82
	.dc.b	$38,$1D,$1F,$7E,$7F,$B5,$74,$7D,$F7,$FB,$67,$47,$DF,$7F,$B7,$74
	.dc.b	$7D,$F7,$FC,$01,$D1,$F7,$DF,$F1,$0F,$47,$DF,$7F,$BD,$78,$FF,$7B
	.dc.b	$4F,$F2,$BF,$4F,$F5,$79,$FB,$E7,$E5,$0F,$D0,$9F,$DF,$FE,$07,$FD
	.dc.b	$8D,$A9,$A4,$8F,$F6,$7C,$65,$B1,$97,$FD,$A1,$19,$6C,$65,$FF,$67
	.dc.b	$46,$5B,$19,$7F,$C3,$2D,$7C,$87,$62,$1D,$0F,$1F,$C0,$FF,$80,$BE
	.dc.b	$1B,$FC,$07,$F0,$DF,$E0,$4F,$86,$FF,$02,$FC,$34,$BF,$F0,$F3,$FE
	.dc.b	$0B,$FC,$6B,$FC,$AF,$0D,$3F,$C4,$F1,$FE,$EA,$6F,$BD,$BF,$5B,$87
	.dc.b	$BB,$C7,$DD,$E5,$F0,$1C,$FE,$1B,$FC,$27,$C1,$E9,$EA,$FF,$18,$3F
	.dc.b	$E2,$8C,$7F,$D4,$9C,$5F,$07,$5F,$79,$EC,$3E,$19,$4C,$9E,$F2,$1A
	.dc.b	$29,$A3,$04,$30,$43,$E5,$9C,$0A,$6C,$FF,$20,$71,$7C,$5F,$86,$F9
	.dc.b	$71,$FE,$46,$DB,$E2,$FC,$37,$C9,$F5,$FE,$57,$C3,$7C,$BF,$3F,$E6
	.dc.b	$78,$1F,$37,$DE,$F9,$DF,$0D,$F3,$FD,$FF,$A1,$F0,$DF,$27,$4E,$97
	.dc.b	$FE,$52,$E2,$74,$39,$7E,$B7,$C0,$7C,$CB,$F8,$87,$AB,$93,$E5,$F2
	.dc.b	$FC,$FE,$5F,$99,$3E,$C3,$E0,$63,$F4,$9E,$2D,$7E,$13,$5F,$84,$D7
	.dc.b	$8C,$74,$8C,$47,$52,$38,$24,$7F,$94,$E3,$FC,$D9,$DB,$F8,$FF,$80
	.dc.b	$FC,$A4,$7E,$48,$87,$F9,$AF,$B3,$F2,$9F,$01,$F9,$5F,$86,$FC,$B6
	.dc.b	$7F,$91,$AF,$8E,$32,$F8,$FD,$BF,$9B,$F8,$1F,$F3,$37,$E1,$FF,$ED
	.dc.b	$DA,$FC,$E7,$17,$CD,$E2,$77,$7C,$B9,$F8,$3C,$1F,$B7,$CF,$E5,$C7
	.dc.b	$C6,$8F,$80,$4E,$1F,$9C,$48,$D3,$5E,$6D,$7E,$5C,$72,$EE,$FC,$64
	.dc.b	$7F,$A3,$FC,$2F,$D9,$76,$FE,$2E,$3D,$BB,$FE,$0B,$B3,$F1,$91,$8D
	.dc.b	$3F,$6F,$DD,$FB,$6E,$CF,$C6,$CE,$7A,$FF,$1A,$EF,$C0,$FD,$AB,$C7
	.dc.b	$AB,$E0,$3E,$4D,$FF,$69,$DF,$F9,$1E,$47,$C3,$FF,$A7,$AB,$FD,$3F
	.dc.b	$B2,$73,$19,$3E,$93,$87,$84,$6E,$75,$1A,$6C,$66,$39,$35,$E3,$3C
	.dc.b	$DF,$0B,$5F,$34,$F2,$8C,$47,$53,$9B,$89,$E9,$1C,$A7,$81,$1D,$B5
	.dc.b	$F0,$FC,$3F,$E2,$BE,$1F,$F1,$7F,$0F,$F8,$CF,$87,$FC,$6F,$B5,$2F
	.dc.b	$EB,$E5,$FE,$41,$3F,$C7,$DD,$5F,$2F,$D9,$FC,$9F,$BD,$F9,$4F,$6B
	.dc.b	$F2,$BE,$E7,$E5,$BD,$BF,$CB,$FA,$3F,$98,$F7,$BF,$33,$C3,$E3,$D3
	.dc.b	$D0,$7F,$B5,$BC,$07,$27,$C2,$8F,$49,$0F,$93,$3C,$5D,$18,$34,$74
	.dc.b	$30,$43,$04,$38,$94,$F9,$C6,$CC,$90,$FC,$1F,$1F,$C9,$7E,$2F,$A5
	.dc.b	$FB,$4F,$7B,$F6,$BE,$DF,$ED,$BD,$3F,$DB,$FB,$7F,$B8,$F6,$E6,$1F
	.dc.b	$33,$E1,$7E,$43,$F3,$B8,$7F,$DC,$5F,$8D,$FF,$BD,$39,$DC,$5F,$26
	.dc.b	$BE,$9E,$FF,$05,$F3,$F0,$FC,$87,$D2,$7A,$FB,$3F,$0A,$FA,$5A,$E8
	.dc.b	$C5,$FE,$2C,$95,$D6,$76,$48,$CC,$FC,$E3,$8C,$EF,$C8,$79,$1E,$AF
	.dc.b	$73,$F1,$B5,$F8,$AE,$DF,$C6,$3F,$C6,$0F,$D6,$F6,$E3,$C5,$77,$7E
	.dc.b	$DE,$3F,$8F,$F7,$7E,$3A,$BF,$84,$F0,$7F,$31,$79,$6C,$87,$B9,$F9
	.dc.b	$19,$DF,$C2,$FC,$CF,$77,$F2,$31,$E9,$3E,$4F,$6F,$A9,$C6,$1E,$FE
	.dc.b	$1F,$85,$C5,$F0,$27,$F0,$87,$C9,$24,$ED,$E3,$1B,$7E,$93,$5F,$E2
	.dc.b	$5D,$BF,$89,$7D,$30,$ED,$FC,$54,$7E,$C3,$B3,$F1,$4F,$EC,$BA,$FF
	.dc.b	$12,$F0,$EB,$FE,$87,$D7,$F8,$A8,$E7,$D7,$F8,$A7,$F6,$27,$EE,$BB
	.dc.b	$BE,$0B,$EA,$D3,$F8,$A7,$5F,$E2,$5F,$95,$F0,$BF,$C6,$B9,$0D,$25
	.dc.b	$7F,$1C,$8C,$B6,$32,$FF,$26,$8C,$B6,$32,$FF,$30,$8C,$B6,$32,$FF
	.dc.b	$81,$E3,$2D,$8C,$BF,$D9,$63,$2D,$8C,$CF,$E0,$FD,$EF,$D5,$FC,$2F
	.dc.b	$EB,$3B,$BF,$5B,$EF,$FE,$BB,$E1,$66,$3F,$8C,$7F,$63,$F1,$1F,$B2
	.dc.b	$F8,$4F,$D9,$FC,$27,$ED,$3E,$23,$F6,$BE,$6F,$ED,$BC,$5F,$DB,$FB
	.dc.b	$BF,$B8,$E0,$F0,$F0,$F9,$38,$7E,$5E,$BF,$C7,$7B,$7D,$4F,$C7,$21
	.dc.b	$F1,$4A,$7C,$32,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$1C,$EA,$7C
	.dc.b	$BF,$5B,$E3,$BF,$AE,$F8,$4F,$8E,$FE,$C3,$DD,$3F,$DE,$BC,$FF,$EE
	.dc.b	$FE,$3F,$F8,$F5,$FF,$36,$3F,$67,$8B,$E5,$4F,$EE,$38,$7E,$53,$FB
	.dc.b	$AE,$0F,$94,$FE,$F3,$7F,$94,$F9,$35,$F9,$2F,$C0,$E8,$FC,$B6,$FF
	.dc.b	$34,$9C,$1F,$40,$93,$C6,$32,$D8,$CB,$F1,$CF,$84,$5A,$EE,$FC,$83
	.dc.b	$F1,$DD,$1F,$01,$FE,$1F,$AB,$0F,$C6,$47,$F1,$4E,$7F,$C5,$BF,$C5
	.dc.b	$31,$73,$63,$C8,$FE,$3B,$3F,$E3,$DC,$1F,$09,$D6,$73,$9A,$5E,$E6
	.dc.b	$78,$BE,$77,$AB,$71,$E5,$F0,$17,$0E,$1E,$E5,$C5,$F0,$97,$DD,$F9
	.dc.b	$B8,$7B,$F7,$F8,$AF,$AF,$AF,$F0,$B8,$FC,$28,$B3,$A7,$E2,$46,$5F
	.dc.b	$92,$62,$34,$F0,$E4,$F9,$EE,$5A,$FE,$F7,$AF,$F1,$EF,$C7,$EB,$FC
	.dc.b	$7B,$F2,$7A,$FD,$5C,$1E,$BF,$6F,$F6,$B3,$FD,$0B,$BF,$F6,$AF,$F4
	.dc.b	$4E,$FF,$DA,$BF,$D1,$BB,$FF,$6A,$FF,$48,$CF,$C2,$FE,$03,$AB,$F9
	.dc.b	$07,$17,$F6,$98,$FF,$1E,$B5,$FF,$22,$B5,$FC,$F7,$07,$F8,$FA,$31
	.dc.b	$1F,$6E,$30,$47,$FC,$83,$1A,$E0,$FB,$93,$99,$EE,$D7,$BB,$5E,$E7
	.dc.b	$27,$0F,$73,$AF,$BB,$F7,$FE,$13,$F0,$1F,$09,$F8,$1F,$7B,$F0,$5F
	.dc.b	$09,$F8,$3F,$7E,$61,$FB,$9F,$84,$FC,$37,$C2,$7E,$1F,$E1,$3F,$11
	.dc.b	$F0,$9F,$8B,$FC,$73,$FD,$1D,$E1,$7E,$51,$FF,$31,$37,$D5,$FC,$77
	.dc.b	$07,$93,$87,$DB,$9F,$9C,$FE,$40,$79,$CF,$79,$92,$90,$F8,$26,$4A
	.dc.b	$64,$F8,$65,$3A,$1A,$21,$A2,$9A,$30,$43,$FD,$5D,$6F,$84,$F9,$51
	.dc.b	$FC,$77,$83,$E1,$7C,$27,$CA,$8F,$F4,$CD,$3D,$EF,$8D,$EF,$7C,$7F
	.dc.b	$57,$E4,$7B,$DF,$2E,$FF,$93,$6F,$F3,$1E,$3D,$1F,$37,$E1,$3E,$4C
	.dc.b	$7B,$CF,$26,$DF,$29,$FD,$A1,$E9,$23,$97,$DA,$D7,$8C,$FE,$3C,$9B
	.dc.b	$FD,$56,$C7,$73,$9B,$87,$E6,$4F,$37,$3D,$3F,$83,$F5,$BF,$09,$F0
	.dc.b	$9F,$8D,$9F,$87,$F0,$5F,$8C,$7F,$92,$F4,$FF,$11,$E9,$FC,$6C,$7C
	.dc.b	$8F,$63,$F1,$9E,$E7,$E3,$7E,$2F,$F1,$DE,$FF,$E3,$FA,$BE,$1D,$F8
	.dc.b	$B7,$F2,$78,$3F,$B6,$E1,$F5,$12,$7D,$5C,$3E,$72,$57,$AE,$3E,$2B
	.dc.b	$5F,$8B,$3C,$5E,$CF,$FA,$E3,$A3,$F0,$51,$FD,$77,$C1,$FD,$5B,$F8
	.dc.b	$0E,$AF,$C1,$3F,$E5,$09,$FC,$13,$F8,$4E,$9F,$C1,$3F,$86,$E8,$FC
	.dc.b	$13,$FD,$23,$A3,$F0,$4F,$1C,$BF,$04,$FF,$6B,$CB,$C8,$FE,$83,$4F
	.dc.b	$F1,$A6,$2F,$85,$F3,$12,$7F,$D9,$B1,$96,$C6,$5F,$F6,$BC,$65,$B1
	.dc.b	$97,$FD,$C3,$19,$6C,$65,$FE,$FF,$AA,$75,$18,$6B,$FA,$A9,$E0,$D3
	.dc.b	$C6,$43,$C7,$EF,$FF,$0D,$F8,$5F,$E1,$DF,$01,$FC,$3F,$E0,$3F,$AD
	.dc.b	$7E,$3B,$FE,$F5,$F0,$FF,$B1,$AF,$86,$CF,$C5,$EB,$E1,$F7,$F0,$FC
	.dc.b	$3E,$1F,$8F,$9B,$CB,$E0,$39,$FC,$37,$4F,$81,$FE,$39,$C5,$E6,$E4
	.dc.b	$F6,$3F,$1F,$97,$E7,$B9,$E5,$F6,$CE,$B9,$7D,$47,$C6,$29,$F0,$E7
	.dc.b	$81,$1C,$9F,$0C,$FA,$26,$23,$A3,$82,$1F,$53,$97,$E3,$3F,$C4,$39
	.dc.b	$7E,$33,$FE,$32,$E3,$F8,$BF,$0D,$F2,$38,$FE,$4F,$C1,$7C,$AF,$86
	.dc.b	$F8,$B7,$BF,$63,$ED,$F3,$BD,$7F,$9F,$F0,$DF,$1E,$B4,$36,$53,$8F
	.dc.b	$EA,$FB,$1F,$2A,$BD,$5C,$5F,$18,$FC,$77,$1F,$CA,$8C,$F1,$FC,$37
	.dc.b	$DA,$7C,$73,$24,$32,$7C,$73,$E7,$18,$C7,$E3,$B8,$21,$CC,$FC,$FF
	.dc.b	$A7,$F8,$AF,$86,$FC,$6C,$7B,$27,$D7,$EC,$FE,$3A,$3F,$1F,$F0,$DF
	.dc.b	$8C,$75,$AF,$FB,$29,$BE,$F7,$E3,$5C,$F5,$FE,$5B,$DE,$FC,$BF,$BD
	.dc.b	$F8,$D9,$FE,$8B,$AF,$BD,$F4,$ED,$F0,$DF,$56,$FF,$1E,$75,$1E,$93
	.dc.b	$D6,$7C,$13,$04,$30,$4D,$7E,$1B,$96,$B9,$9E,$CF,$F0,$6E,$BF,$C5
	.dc.b	$BF,$C6,$7A,$FF,$16,$FF,$BA,$B5,$75,$7E,$2E,$3E,$1F,$4F,$E2,$DF
	.dc.b	$8B,$DF,$FB,$6E,$8F,$C4,$C7,$C8,$EE,$FD,$7B,$F2,$B9,$FF,$0C,$FC
	.dc.b	$0D,$BF,$5D,$E2,$7F,$0C,$FC,$7B,$FC,$C7,$3F,$E6,$D3,$F2,$0E,$1F
	.dc.b	$F3,$87,$02,$31,$1C,$7B,$38,$34,$B3,$B8,$CC,$74,$6C,$66,$37,$3D
	.dc.b	$1D,$18,$9D,$C6,$67,$84,$F9,$DA,$F9,$A3,$44,$9E,$24,$8E,$FE,$FF
	.dc.b	$F9,$DE,$7F,$D4,$F1,$F0,$38,$F1,$F0,$F9,$F8,$7D,$7C,$3E,$FE,$1F
	.dc.b	$87,$C3,$F1,$F8,$7E,$5F,$0F,$CF,$7E,$9E,$F7,$5F,$53,$B7,$07,$87
	.dc.b	$6F,$46,$FE,$57,$DC,$7F,$8D,$B8,$7D,$6F,$CB,$37,$3C,$5F,$73,$5D
	.dc.b	$18,$39,$9C,$0A,$64,$C1,$0C,$10,$E6,$43,$E9,$9E,$8E,$1F,$90,$7C
	.dc.b	$9F,$82,$F9,$5F,$0D,$F2,$FE,$1B,$E3,$D7,$EF,$F1,$70,$7C,$88,$C7
	.dc.b	$BD,$F4,$7D,$EF,$A5,$F0,$DF,$4F,$E1,$BE,$45,$71,$D7,$DC,$F8,$8F
	.dc.b	$31,$E4,$E8,$71,$79,$F8,$3E,$4C,$EC,$C6,$9F,$3D,$F8,$06,$2F,$AC
	.dc.b	$F0,$24,$EE,$7B,$C6,$27,$73,$F8,$EF,$63,$FD,$81,$C9,$FE,$76,$9F
	.dc.b	$D6,$63,$F8,$AC,$3E,$17,$46,$BD,$DD,$FB,$BC,$3C,$EF,$C5,$57,$C7
	.dc.b	$C7,$F1,$4F,$C9,$E8,$EB,$EE,$FE,$2A,$3E,$D9,$E9,$3E,$FE,$5F,$09
	.dc.b	$F8,$04,$DB,$D0,$F0,$31,$B7,$B5,$E4,$7C,$43,$3A,$FB,$88,$F1,$9F
	.dc.b	$8B,$D1,$F8,$28,$FB,$DD,$1F,$82,$7F,$01,$CF,$F8,$25,$EA,$7F,$BB
	.dc.b	$31,$FC,$1C,$7E,$17,$B3,$F6,$5D,$1F,$83,$8F,$C4,$73,$FE,$0D,$FC
	.dc.b	$57,$5F,$C1,$7C,$DA,$7B,$1F,$CD,$ED,$F0,$61,$C3,$FE,$44,$9F,$E0
	.dc.b	$CD,$7F,$8D,$B5,$FF,$16,$B5,$FE,$E0,$D8,$E7,$F0,$DF,$C6,$23,$04
	.dc.b	$70,$49,$FE,$43,$19,$6C,$65,$F1,$B5,$F1,$B5,$F1,$FB,$DF,$8C,$F8
	.dc.b	$6F,$C6,$FB,$FF,$8E,$F7,$FF,$1F,$F0,$DF,$90,$F8,$6F,$C8,$FC,$37
	.dc.b	$E4,$BE,$1B,$F2,$7F,$0D,$F9,$4F,$86,$FC,$AF,$95,$F9,$6F,$4F,$F2
	.dc.b	$FE,$9F,$E6,$38,$3C,$1C,$1F,$01,$BC,$5F,$02,$3F,$8A,$71,$7A,$DF
	.dc.b	$49,$EC,$21,$82,$1E,$C3,$25,$32,$7D,$09,$EF,$1D,$63,$44,$34,$72
	.dc.b	$3E,$BF,$C2,$FC,$57,$FC,$19,$0F,$85,$F8,$D1,$8E,$03,$FE,$11,$FC
	.dc.b	$7F,$FC,$B9,$E0,$7E,$03,$87,$E4,$4F,$2E,$1D,$BF,$98,$9F,$F3,$8D
	.dc.b	$7F,$98,$63,$97,$C3,$7C,$B7,$C3,$B7,$CB,$7C,$C7,$CB,$2E,$DF,$2D
	.dc.b	$FF,$65,$B5,$F7,$1B,$3E,$29,$4C,$BB,$7E,$19,$A7,$0E,$CC,$9C,$34
	.dc.b	$F9,$8E,$08,$6C,$91,$FA,$8E,$7F,$83,$DC,$CB,$E0,$FC,$37,$E4,$AF
	.dc.b	$E5,$7A,$BA,$9F,$8B,$F0,$DF,$93,$9F,$9D,$EE,$E5,$FF,$2D,$6C,$F5
	.dc.b	$BF,$2B,$1F,$63,$A3,$F2,$AF,$D5,$F8,$2E,$1B,$2D,$F0,$DF,$36,$BE
	.dc.b	$07,$07,$B1,$F9,$1E,$07,$F0,$18,$C5,$FC,$42,$38,$27,$4F,$E4,$09
	.dc.b	$5F,$5A,$32,$D8,$FE,$01,$D3,$F8,$F8,$FC,$21,$FC,$1B,$A7,$F1,$EE
	.dc.b	$2F,$F8,$8F,$4F,$AB,$3F,$C7,$CF,$F0,$CE,$9F,$C7,$BB,$C9,$D3,$BF
	.dc.b	$86,$E1,$D5,$F8,$DA,$D7,$57,$C1,$7E,$E6,$DE,$87,$F1,$79,$7F,$92
	.dc.b	$1F,$F1,$8B,$5F,$F1,$33,$5F,$F1,$DB,$5F,$CE,$FB,$9F,$E3,$48,$C4
	.dc.b	$7D,$88,$C1,$1E,$AD,$7A,$B5,$EA,$D7,$A9,$C0,$87,$03,$CF,$EE,$7C
	.dc.b	$7F,$86,$F9,$1E,$EF,$C9,$F8,$6F,$95,$EF,$7C,$BF,$86,$F9,$9F,$0D
	.dc.b	$F3,$7E,$1A,$65,$FB,$DF,$86,$FA,$1F,$0D,$F4,$7E,$1B,$F1,$7F,$90
	.dc.b	$7F,$97,$BC,$6F,$CE,$5F,$B4,$F1,$9E,$82,$9E,$83,$E6,$F0,$7A,$DE
	.dc.b	$07,$A4,$87,$12,$9C,$4C,$10,$C1,$0E,$66,$4A,$67,$AB,$F4,$51,$E1
	.dc.b	$6C,$7F,$9B,$AB,$FC,$B3,$05,$DB,$F0,$3E,$17,$E5,$57,$F2,$1D,$FE
	.dc.b	$1F,$BB,$F1,$3E,$17,$E4,$CF,$E6,$79,$9C,$1F,$99,$E0,$F9,$73,$FC
	.dc.b	$A7,$7F,$97,$F0,$7F,$9D,$DF,$C9,$87,$E2,$37,$F2,$DF,$C5,$DB,$FB
	.dc.b	$A4,$E4,$95,$C0,$93,$F4,$0F,$82,$4A,$E6,$FC,$56,$BF,$16,$79,$BC
	.dc.b	$63,$2F,$18,$73,$FE,$36,$3F,$A1,$73,$7E,$39,$FF,$6A,$70,$BD,$6F
	.dc.b	$C7,$47,$F4,$7E,$5F,$C4,$FC,$2F,$E2,$BB,$FF,$17,$F0,$5F,$8C,$F7
	.dc.b	$BF,$1B,$EB,$FE,$3B,$CD,$F8,$98,$F8,$F6,$F2,$BF,$0E,$FD,$1C,$1F
	.dc.b	$02,$3F,$0E,$79,$DC,$3E,$F9,$E2,$D8,$CB,$C8,$F8,$1A,$FE,$08,$FC
	.dc.b	$0B,$EC,$E9,$FC,$0B,$F7,$7A,$7F,$02,$FC,$0E,$8F,$C0,$F5,$7E,$93
	.dc.b	$93,$F9,$87,$4F,$E0,$6B,$FB,$47,$5F,$EA,$DF,$8F,$D1,$EE,$7D,$35
	.dc.b	$F4,$09,$AF,$F1,$1E,$8F,$F1,$A6,$AF,$73,$FD,$82,$DB,$FC,$D4,$60
	.dc.b	$8E,$08,$7F,$AF,$E3,$83,$63,$6D,$36,$7F,$B6,$23,$6D,$8D,$B4,$D9
	.dc.b	$FE,$06,$BF,$D5,$46,$A3,$12,$F8,$5F,$E1,$7E,$EC,$9F,$C4,$FB,$9F
	.dc.b	$C3,$BE,$17,$F8,$7F,$BB,$FC,$43,$E1,$65,$7E,$3B,$E1,$7F,$AE,$FE
	.dc.b	$43,$FE,$87,$86,$A8,$7C,$36,$BE,$1B,$F8,$CE,$FE,$3B,$DE,$FE,$C7
	.dc.b	$CB,$19,$C3,$D3,$1F,$EC,$B9,$D3,$C2,$75,$18,$9D,$CF,$C9,$21,$F1
	.dc.b	$4A,$76,$3E,$39,$CC,$AF,$27,$81,$0E,$07,$23,$81,$FE,$F9,$C1,$F0
	.dc.b	$DF,$13,$E1,$BE,$2F,$99,$FC,$E3,$82,$5F,$ED,$B7,$F9,$6F,$F6,$FC
	.dc.b	$9B,$FF,$3E,$DF,$E6,$FC,$4F,$CE,$E0,$97,$FB,$ED,$FE,$09,$1F,$C6
	.dc.b	$F2,$7E,$5A,$31,$1E,$7E,$0F,$63,$EA,$3D,$87,$A8,$AE,$9C,$1B,$21
	.dc.b	$82,$3B,$26,$5F,$4C,$9C,$9F,$8D,$9F,$77,$27,$E2,$3D,$EF,$C7,$47
	.dc.b	$F1,$8E,$5F,$C5,$FC,$2F,$E3,$3E,$17,$F1,$BF,$0B,$F8,$EF,$5F,$F1
	.dc.b	$FF,$0B,$F8,$BC,$3E,$11,$F6,$34,$F0,$6D,$F0,$37,$F8,$53,$98,$F8
	.dc.b	$6F,$BA,$76,$E4,$87,$C7,$72,$D8,$E2,$FE,$B7,$A7,$EF,$BF,$E7,$4E
	.dc.b	$FF,$E2,$10,$E9,$FC,$0C,$FE,$B0,$87,$DF,$E7,$FC,$0D,$7F,$17,$C3
	.dc.b	$F0,$33,$FA,$D3,$A7,$8B,$DF,$9B,$F0,$31,$FE,$35,$C7,$F5,$AF,$C2
	.dc.b	$3C,$C7,$A8,$FD,$06,$7F,$EC,$B8,$7B,$1F,$E8,$49,$FF,$42,$F3,$7E
	.dc.b	$DE,$31,$1E,$D6,$BF,$BC,$9E,$2E,$A7,$83,$63,$2F,$E8,$E7,$8C,$EA
	.dc.b	$30,$6E,$33,$3C,$23,$C6,$D7,$C4,$6C,$87,$22,$1C,$23,$99,$D3,$D8
	.dc.b	$FE,$2B,$F0,$B2,$7E,$59,$EB,$F7,$BF,$8C,$9F,$C6,$BE,$17,$F8,$DF
	.dc.b	$C2,$FF,$1C,$F6,$BF,$8E,$FB,$DF,$C7,$BD,$EF,$F3,$DF,$9B,$FE,$62
	.dc.b	$DB,$FC,$A5,$C7,$F8,$0D,$BC,$3A,$79,$8F,$21,$4F,$59,$FE,$A8,$DF
	.dc.b	$DC,$FC,$D8,$E2,$FB,$0D,$9C,$8C,$1B,$3A,$14,$CD,$7C,$F7,$81,$0E
	.dc.b	$64,$38,$19,$29,$F9,$DD,$FD,$FE,$F7,$C0,$F7,$BE,$3C,$FC,$AA,$FE
	.dc.b	$7D,$C1,$F1,$3E,$17,$E2,$FC,$2F,$C9,$AD,$66,$DF,$E5,$47,$1D,$BE
	.dc.b	$51,$F2,$E7,$E1,$E3,$E9,$9F,$33,$5F,$29,$EF,$E4,$FC,$E6,$DF,$22
	.dc.b	$3D,$53,$B2,$57,$42,$4E,$E7,$AC,$61,$F9,$73,$C2,$75,$1E,$08,$C3
	.dc.b	$FA,$1B,$7C,$2F,$E3,$63,$F6,$9C,$BF,$86,$F8,$5F,$C3,$FC,$2F,$E3
	.dc.b	$67,$E1,$F3,$7E,$2B,$E1,$7F,$17,$F0,$BF,$8C,$CB,$F1,$BE,$FF,$E3
	.dc.b	$BE,$17,$FB,$6E,$DE,$2D,$BC,$39,$7D,$ED,$FF,$5F,$C1,$F0,$A3,$D3
	.dc.b	$1E,$B7,$55,$C6,$31,$1F,$19,$AF,$C2,$8D,$11,$E7,$CF,$F8,$27,$EC
	.dc.b	$73,$7E,$09,$FE,$25,$CD,$F8,$17,$EE,$72,$FE,$05,$FE,$2D,$CB,$F8
	.dc.b	$07,$FD,$99,$B7,$E0,$1F,$69,$E8,$C7,$C0,$FE,$F7,$37,$B3,$FE,$1B
	.dc.b	$A7,$C0,$7F,$0C,$6D,$7D,$38,$C1,$1C,$10,$FE,$41,$B2,$78,$C6,$1A
	.dc.b	$FF,$24,$6B,$E1,$9E,$8E,$DA,$ED,$AE,$CF,$01,$93,$F1,$1F,$0B,$FC
	.dc.b	$2B,$DA,$91,$FA,$3F,$85,$FE,$1A,$7F,$0E,$F8,$1F,$E1,$FE,$DF,$F1
	.dc.b	$0F,$84,$FE,$23,$EE,$7F,$12,$F8,$46,$1F,$B3,$FC,$4B,$FD,$89,$E4
	.dc.b	$7F,$90,$5F,$F5,$5E,$DE,$A9,$FF,$1A,$B7,$7F,$85,$BF,$A2,$72,$FB
	.dc.b	$B8,$7E,$8C,$60,$F5,$9E,$D3,$E1,$18,$34,$60,$87,$C1,$29,$C4,$C1
	.dc.b	$0C,$10,$E6,$64,$A6,$67,$ED,$70,$7C,$28,$78,$9F,$E8,$1C,$7F,$D1
	.dc.b	$31,$8C,$BE,$27,$C0,$7C,$E8,$FA,$07,$F1,$3E,$0F,$9E,$FC,$BF,$86
	.dc.b	$E9,$87,$3F,$81,$B8,$7C,$DF,$81,$B7,$FE,$60,$CB,$FD,$60,$FD,$3F
	.dc.b	$88,$FE,$8F,$BF,$CD,$DB,$E7,$CF,$C7,$6B,$F1,$F7,$F9,$F1,$88,$F8
	.dc.b	$0D,$7E,$07,$07,$D9,$25,$75,$6C,$75,$6B,$F4,$0F,$92,$49,$E8,$D8
	.dc.b	$FC,$26,$0F,$86,$FC,$AC,$FE,$E3,$9E,$ED,$F3,$7E,$1B,$E0,$E2,$C3
	.dc.b	$F0,$F7,$F3,$CF,$89,$F0,$DF,$99,$7F,$A0,$73,$7E,$65,$FC,$C7,$07
	.dc.b	$2A,$7C,$37,$C8,$9F,$56,$DE,$D3,$F2,$9D,$7F,$09,$C4,$68,$87,$BF
	.dc.b	$2F,$69,$37,$F8,$71,$F3,$1A,$F3,$3E,$47,$3F,$8F,$87,$C9,$CF,$F8
	.dc.b	$79,$FE,$33,$D7,$FB,$57,$FA,$D7,$5F,$A7,$A3,$D5,$D1,$EB,$E8,$F6
	.dc.b	$7C,$3F,$B7,$A7,$D9,$87,$97,$5F,$0B,$F4,$77,$FC,$D6,$BF,$82,$E2
	.dc.b	$FF,$4E,$13,$D6,$FF,$15,$B6,$BF,$C6,$4D,$7F,$8E,$B5,$FF,$1E,$B5
	.dc.b	$FE,$09,$1C,$88,$F2,$6B,$C9,$AF,$27,$C3,$EF,$FD,$7F,$7E,$4F,$F1
	.dc.b	$17,$ED,$7C,$1F,$DB,$F7,$FE,$E7,$B3,$F7,$7E,$0F,$EF,$7C,$1F,$E2
	.dc.b	$3F,$22,$FF,$64,$74,$9F,$F2,$DE,$5F,$89,$8C,$7C,$3F,$E1,$75,$F1
	.dc.b	$CF,$A8,$F4,$10,$F4,$1F,$C9,$37,$F9,$51,$C2,$3F,$6C,$DA,$FC,$C3
	.dc.b	$77,$F8,$D3,$A3,$A1,$F0,$4A,$78,$4E,$E4,$34,$53,$46,$0F,$A7,$BF
	.dc.b	$E7,$CF,$85,$F0,$FF,$0F,$DF,$F9,$33,$FC,$EF,$7F,$8D,$F0,$FF,$26
	.dc.b	$3F,$C6,$1A,$BE,$1B,$E5,$7C,$37,$CB,$F8,$6F,$99,$F0,$DF,$1E,$FC
	.dc.b	$E7,$AC,$F2,$14,$F3,$F0,$7C,$67,$7C,$1F,$11,$BB,$FC,$48,$C4,$72
	.dc.b	$6B,$ED,$3D,$E4,$3B,$15,$DB,$63,$A1,$FE,$F9,$E8,$78,$3F,$8C,$9F
	.dc.b	$C1,$4B,$A3,$1E,$43,$C5,$FE,$C3,$8F,$E2,$7E,$1B,$F1,$9F,$0D,$F9
	.dc.b	$0C,$7E,$3F,$A1,$F9,$07,$E4,$FC,$2F,$E4,$7E,$1F,$F2,$5D,$7F,$1E
	.dc.b	$7D,$9A,$F9,$3A,$FF,$47,$C3,$F1,$67,$31,$F8,$72,$3F,$19,$AE,$8C
	.dc.b	$10,$C1,$FB,$6E,$BF,$C1,$3F,$73,$AF,$F0,$4F,$F7,$BE,$FF,$D5,$3F
	.dc.b	$80,$EA,$FC,$13,$FA,$5E,$AF,$C1,$3F,$84,$E9,$FC,$13,$F8,$6E,$8F
	.dc.b	$63,$F5,$4F,$A6,$4E,$3F,$A6,$FE,$17,$4F,$EA,$36,$F5,$FF,$D7,$2D
	.dc.b	$9F,$F6,$03,$5F,$F1,$5B,$5F,$F6,$5B,$5F,$AA,$D5,$5F,$BF,$9E,$A4
	.dc.b	$8E,$AD,$7A,$B5,$EA,$D7,$AB,$5E,$B3,$DE,$7B,$47,$A3,$E0,$3F,$88
	.dc.b	$7A,$F2,$3F,$57,$F0,$DF,$C4,$FD,$FF,$E2,$9F,$0D,$FC,$57,$D9,$FE
	.dc.b	$2D,$F0,$DF,$D7,$BF,$23,$FF,$60,$0F,$87,$C7,$BD,$FC,$6F,$7F,$16
	.dc.b	$DE,$0D,$3D,$27,$B0,$86,$07,$07,$D0,$9E,$3C,$1E,$92,$38,$24,$F2
	.dc.b	$3E,$19,$4F,$86,$53,$E1,$BC,$1F,$01,$C0,$FB,$3C,$1E,$DF,$87,$F7
	.dc.b	$7B,$FE,$FF,$87,$F8,$1F,$0F,$F0,$7E,$1F,$E2,$61,$FC,$4B,$83,$E2
	.dc.b	$7C,$3F,$C5,$F8,$7F,$8D,$E5,$7C,$3A,$F1,$70,$7C,$37,$C0,$43,$CE
	.dc.b	$7C,$3E,$0F,$84,$F0,$E0,$F8,$2D,$E0,$F5,$3C,$4F,$69,$92,$1E,$E3
	.dc.b	$A9,$BA,$F1,$38,$3F,$84,$F1,$BE,$1F,$F0,$91,$FA,$DE,$9F,$EA,$DD
	.dc.b	$3F,$85,$8F,$E6,$3D,$1F,$83,$F8,$7F,$C2,$7A,$9F,$85,$9F,$C3,$74
	.dc.b	$FE,$11,$FE,$6F,$A7,$A1,$F0,$1E,$83,$CC,$5D,$BE,$0B,$8D,$7C,$EE
	.dc.b	$CF,$56,$DF,$11,$D3,$5C,$D7,$77,$81,$E0,$9E,$64,$9F,$C7,$74,$7E
	.dc.b	$06,$7F,$DF,$9B,$B9,$FF,$05,$18,$E6,$FC,$13,$FD,$4B,$AB,$F4,$AF
	.dc.b	$0E,$5F,$C1,$3C,$B9,$3F,$04,$F9,$B4,$F2,$69,$E1,$7F,$C3,$FE,$EF
	.dc.b	$D2,$F3,$BF,$CF,$55,$FE,$7B,$A6,$9F,$D1,$23,$27,$F0,$A7,$0F,$FA
	.dc.b	$F1,$AF,$16,$BD,$4E,$31,$D2,$79,$C7,$36,$C6,$63,$73,$C9,$D1,$89
	.dc.b	$DC,$66,$78,$4F,$8E,$72,$47,$81,$0F,$47,$C1,$7F,$0D,$F7,$7F,$87
	.dc.b	$7C,$27,$F0,$FF,$7B,$F8,$87,$C3,$7F,$11,$F8,$2F,$E2,$5F,$0D,$FE
	.dc.b	$70,$F4,$BF,$D7,$F9,$7F,$2C,$5E,$EE,$7A,$35,$F0,$5F,$C6,$76,$F1
	.dc.b	$5F,$94,$F1,$1F,$1D,$CB,$4F,$F1,$57,$0F,$C9,$78,$10,$D1,$4D,$18
	.dc.b	$21,$82,$1F,$28,$E0,$53,$65,$37,$C1,$F3,$1C,$B5,$FF,$00,$E2,$F8
	.dc.b	$2F,$95,$1F,$EE,$4F,$5F,$FC,$0C,$FC,$5F,$82,$F9,$1F,$05,$F2,$7E
	.dc.b	$0B,$E5,$7C,$37,$CC,$AF,$D0,$F0,$7C,$DF,$86,$F9,$51,$CF,$DC,$EF
	.dc.b	$C1,$F2,$E3,$E4,$9E,$6E,$9F,$C1,$13,$87,$E7,$F1,$7F,$AF,$23,$13
	.dc.b	$C0,$F7,$13,$87,$D8,$D8,$DB,$F1,$A3,$24,$8C,$C7,$C4,$6B,$E3,$AF
	.dc.b	$CA,$F4,$FE,$1F,$8F,$F1,$B3,$FB,$EE,$7F,$C6,$BF,$C0,$39,$BF,$1C
	.dc.b	$FB,$CF,$C6,$69,$F8,$FE,$69,$19,$CD,$CC,$F8,$6F,$89,$3E,$7D,$7C
	.dc.b	$6F,$AF,$5F,$87,$A7,$96,$3E,$16,$1E,$B9,$FC,$8C,$64,$E4,$E1,$E4
	.dc.b	$70,$AE,$B3,$C6,$39,$E1,$F8,$37,$F8,$DF,$57,$E9,$F9,$BF,$03,$1E
	.dc.b	$99,$FF,$7B,$BF,$E8,$F7,$FD,$C0,$FE,$08,$86,$78,$BF,$06,$EF,$8B
	.dc.b	$DE,$FD,$1C,$BC,$D8,$F8,$E3,$F4,$FC,$5F,$EF,$CC,$3F,$DB,$B1,$FC
	.dc.b	$49,$AF,$F2,$09,$F9,$7F,$0D,$FC,$4A,$73,$3F,$C8,$DA,$FE,$82,$34
	.dc.b	$47,$A3,$5E,$91,$C1,$FE,$0D,$19,$7A,$35,$E8,$72,$21,$C9,$F2,$7B
	.dc.b	$1F,$C4,$7D,$8F,$E2,$5E,$FF,$F1,$3F,$63,$F8,$A7,$C1,$7F,$15,$F8
	.dc.b	$69,$8F,$CF,$F8,$6F,$E3,$1F,$0C,$8F,$C5,$7E,$49,$FE,$C2,$3F,$D8
	.dc.b	$3E,$0F,$EE,$37,$F8,$73,$E2,$3C,$E5,$3C,$E7,$94,$F3,$F0,$7C,$FD
	.dc.b	$FD,$91,$F0,$0E,$05,$38,$10,$EA,$53,$A9,$82,$18,$21,$CC,$C9,$4C
	.dc.b	$CF,$A9,$FA,$7B,$FC,$17,$FD,$BD,$83,$E0,$3E,$5C,$7E,$57,$6D,$65
	.dc.b	$F0,$E9,$EF,$E3,$B3,$FD,$43,$0F,$87,$C6,$1F,$E6,$E8,$7C,$07,$F4
	.dc.b	$7E,$0F,$A3,$3F,$43,$E1,$F7,$B2,$FE,$8E,$3E,$8E,$6F,$C4,$4F,$A6
	.dc.b	$FD,$AD,$75,$C1,$FD,$E9,$F8,$04,$8D,$36,$79,$92,$7E,$AC,$E8,$91
	.dc.b	$99,$FA,$F1,$89,$DC,$E6,$7E,$B1,$FD,$87,$C7,$FC,$A3,$F8,$CF,$81
	.dc.b	$F6,$E2,$F8,$7F,$CA,$CF,$F0,$1E,$6F,$CA,$BF,$4B,$E0,$B7,$8F,$0F
	.dc.b	$85,$E3,$F0,$FF,$DB,$F4,$F0,$92,$FC,$B5,$FB,$AE,$8F,$C8,$46,$27
	.dc.b	$4F,$9B,$6F,$53,$92,$19,$AF,$63,$F3,$9A,$FD,$3E,$8F,$5F,$47,$E1
	.dc.b	$A3,$F8,$C7,$47,$BB,$9F,$DF,$CF,$F0,$39,$FF,$01,$5F,$C7,$FA,$7E
	.dc.b	$1F,$3F,$C4,$F8,$6F,$64,$F8,$8F,$06,$DE,$8D,$DE,$27,$E3,$BE,$1F
	.dc.b	$E5,$92,$BF,$72,$D7,$F3,$6D,$EC,$FC,$C3,$7A,$F9,$4F,$02,$1C,$0B
	.dc.b	$3C,$1B,$1C,$1D,$12,$34,$D8,$D4,$60,$8E,$09,$1F,$A5,$9F,$23,$4E
	.dc.b	$E5,$3A,$16,$79,$B6,$79,$F1,$FE,$A6,$BF,$79,$F2,$2E,$5C,$F5,$27
	.dc.b	$BD,$6F,$0E,$ED,$7B,$B7,$1E,$ED,$C7,$9D,$72,$BE,$97,$C9,$AF,$26
	.dc.b	$DF,$26,$D7,$29,$E1,$5C,$C8,$E0,$87,$37,$2F,$D2,$C7,$DC,$FF,$0F
	.dc.b	$F4,$BF,$5E,$47,$F9,$EB,$5F,$E3,$ED,$C3,$F8,$E3,$6F,$F8,$DB,$67
	.dc.b	$A1,$C8,$FE,$8F,$EC,$7E,$F3,$DF,$FD,$EF,$BF,$FB,$EF,$87,$FD,$FF
	.dc.b	$C0,$7F,$00,$F6,$BF,$CC,$DD,$9F,$C0,$BE,$0F,$F8,$FF,$E4,$BF,$E8
	.dc.b	$ED,$BF,$A7,$3F,$EF,$BD,$3F,$C3,$8B,$E2,$37,$F1,$1C,$3E,$03,$8F
	.dc.b	$C4,$72,$F8,$1E,$7F,$11,$DF,$1F,$EB,$7B,$78,$DF,$EB,$BC,$1E,$77
	.dc.b	$D2,$7A,$CF,$86,$7F,$63,$E5,$F9,$7D,$1F,$7E,$3F,$1C,$47,$04,$E8
	.dc.b	$FB,$71,$96,$C6,$7A,$3E,$BF,$07,$E0,$5B,$D1,$F8,$1B,$FB,$87,$DE
	.dc.b	$9D,$36,$35,$18,$8D,$91,$C1,$FB,$2C,$3D,$38,$7E,$4A,$7F,$D7,$4D
	.dc.b	$7F,$74,$62,$FF,$D5,$F1,$98,$FD,$81,$93,$FD,$6E,$E7,$0F,$D3,$46
	.dc.b	$4F,$F1,$44,$64,$93,$D0,$91,$C9,$AF,$06,$BC,$C8,$79,$88,$79,$0B
	.dc.b	$3E,$26,$CF,$81,$EC,$43,$B1,$4E,$A7,$32,$47,$F0,$1E,$BF,$E5,$F1
	.dc.b	$89,$E8,$F2,$21,$CA,$7C,$4E,$EB,$D6,$F9,$1A,$E4,$F1,$19,$29,$93
	.dc.b	$A9,$92,$99,$3C,$13,$E4,$6C,$76,$9F,$19,$27,$D3,$5E,$C8,$C4,$EE
	.dc.b	$79,$57,$73,$99,$E4,$9D,$B6,$36,$F8,$CE,$04,$30,$48,$E2,$FA,$9A
	.dc.b	$F6,$31,$1D,$23,$75,$E7,$8C,$12,$BD,$04,$AF,$23,$5E,$6D,$7B,$10
	.dc.b	$F3,$57,$89,$B3,$D5,$E6,$43,$99,$F8,$BA,$F2,$57,$EE,$A3,$F8,$03
	.dc.b	$6B,$F8,$04,$7E,$E4,$C9,$5C,$BF,$B9,$C7,$F7,$A4,$76,$64,$B8,$FE
	.dc.b	$FA,$35,$3C,$A3,$F8,$26,$3F,$C6,$38,$9F,$05,$FC,$B8,$98,$7F,$2C
	.dc.b	$6A,$AF,$E6,$31,$86,$C6,$C8,$6F,$0F,$F4,$46,$0B,$EF,$3A,$AE,$2E
	.dc.b	$C4,$EF,$E2,$3F,$A0,$FC,$1F,$F4,$2F,$88,$FE,$87,$F1,$1F,$D1,$3E
	.dc.b	$23,$FA,$2F,$AD,$FD,$1B,$E0,$3F,$A3,$FB,$7F,$D2,$3E,$1B,$FA,$4F
	.dc.b	$C0,$FF,$4A,$F8,$6F,$E9,$7F,$05,$FD,$33,$E0,$FF,$CD,$7F,$93,$39
	.dc.b	$2F,$C2,$63,$A7,$AF,$27,$F4,$ED,$FC,$2F,$F5,$0E,$1F,$2B,$FD,$4B
	.dc.b	$8F,$E2,$61,$F0,$39,$BD,$F2,$E8,$FE,$5F,$D3,$F8,$1E,$AF,$E8,$24
	.dc.b	$70,$49,$FC,$14,$65,$B1,$97,$F0,$1D,$1F,$77,$0F,$BF,$5B,$9C,$B7
	.dc.b	$A7,$F0,$91,$FE,$27,$C3,$F0,$5D,$5E,$88,$FA,$39,$7A,$2B,$EA,$93
	.dc.b	$0F,$65,$6F,$1F,$B8,$DA,$E1,$3A,$BF,$BC,$D7,$EF,$57,$DD,$7F,$BC
	.dc.b	$F5,$7D,$D7,$FB,$DF,$57,$DD,$7F,$BE,$F5,$7D,$D7,$FB,$FF,$4F,$DD
	.dc.b	$7F,$C0,$3D,$1F,$75,$FF,$01,$F3,$FD,$D7,$FD,$B1,$AF,$FB,$57,$9B
	.dc.b	$EF,$46,$34,$FC,$AC,$39,$BF,$01,$3E,$03,$F0,$07,$90,$FC,$01,$E8
	.dc.b	$3F,$00,$7B,$8F,$93,$BF,$F9,$33,$CA,$EB,$F0,$1F,$EC,$99,$FF,$65
	.dc.b	$70,$30,$4E,$67,$65,$9D,$46,$63,$F4,$B3,$C1,$D3,$5E,$D3,$C1,$CC
	.dc.b	$F1,$8C,$46,$A7,$83,$E4,$6B,$E2,$35,$3C,$63,$31,$E2,$8C,$11,$EE
	.dc.b	$43,$B7,$B9,$FA,$EF,$82,$FD,$7F,$C3,$FE,$C3,$E1,$FF,$63,$F0,$FF
	.dc.b	$B2,$F8,$5F,$D9,$FA,$5F,$B4,$F8,$7F,$DA,$FA,$9F,$B6,$F3,$FF,$6F
	.dc.b	$AF,$7D,$7C,$A7,$73,$D4,$E5,$A7,$8F,$C2,$FE,$B9,$BF,$AB,$83,$E1
	.dc.b	$CE,$8E,$64,$30,$43,$E2,$14,$F8,$47,$84,$C9,$0C,$9C,$CC,$3E,$06
	.dc.b	$C7,$8C,$7C,$37,$F1,$4F,$86,$F8,$F3,$F9,$73,$F0,$3E,$EF,$C8,$71
	.dc.b	$7F,$8B,$F8,$6F,$E3,$7F,$07,$FC,$73,$E1,$BE,$4D,$6B,$27,$BB,$F2
	.dc.b	$A3,$D0,$79,$8F,$66,$1E,$57,$07,$9B,$53,$35,$F9,$71,$F3,$E3,$47
	.dc.b	$49,$E4,$4B,$F7,$4E,$E7,$A9,$27,$51,$96,$C7,$39,$F8,$DC,$FF,$8B
	.dc.b	$8F,$83,$C4,$F8,$3F,$C6,$47,$BF,$9B,$F1,$8F,$C1,$9F,$C6,$3C,$3D
	.dc.b	$DF,$C6,$3F,$17,$2F,$06,$3E,$68,$F0,$46,$0F,$82,$B3,$F8,$2E,$88
	.dc.b	$7D,$F3,$D9,$3C,$DC,$67,$EC,$6C,$65,$FA,$2D,$7F,$0B,$C5,$F8,$37
	.dc.b	$ED,$74,$7F,$58,$E2,$FC,$14,$7F,$2C,$A7,$17,$E0,$A3,$F0,$1C,$3F
	.dc.b	$82,$75,$C3,$F8,$17,$C3,$A7,$B5,$F4,$9F,$A3,$D3,$FC,$85,$B3,$E0
	.dc.b	$BF,$85,$36,$7F,$8B,$35,$FE,$22,$D7,$F8,$FB,$5F,$F0,$1B,$5F,$EF
	.dc.b	$ED,$D8,$D2,$4F,$80,$93,$E0,$6B,$E0,$6B,$E0,$6B,$E0,$7D,$3F,$05
	.dc.b	$F6,$BC,$0F,$E8,$77,$E0,$F8,$2C,$78,$3E,$6F,$82,$B8,$F9,$BE,$0A
	.dc.b	$DF,$99,$B9,$BE,$1A,$FC,$36,$3E,$1B,$3F,$0D,$AF,$17,$7C,$1E,$4C
	.dc.b	$FE,$0B,$73,$F8,$2D,$CB,$E0,$CF,$EB,$DF,$53,$F0,$48,$7C,$12,$9F
	.dc.b	$04,$A7,$03,$44,$34,$53,$46,$08,$60,$87,$A0,$EA,$7D,$AE,$0F,$87
	.dc.b	$BF,$C4,$F4,$7E,$2F,$C3,$7C,$6F,$86,$F8,$FB,$FC,$8F,$86,$F8,$D8
	.dc.b	$FE,$EB,$DC,$F9,$7F,$0D,$F1,$A3,$AE,$DF,$11,$BA,$FC,$47,$D0,$78
	.dc.b	$CF,$DA,$3E,$68,$E1,$1A,$6B,$EC,$38,$15,$CB,$C6,$3B,$39,$6B,$9D
	.dc.b	$BE,$B1,$F5,$FD,$EF,$C1,$FC,$37,$E1,$39,$BF,$0B,$CD,$2F,$F8,$BF
	.dc.b	$C3,$7E,$19,$FE,$73,$F9,$3F,$F8,$17,$AB,$F2,$79,$B9,$FF,$11,$5F
	.dc.b	$C7,$B4,$EE,$FC,$26,$BF,$73,$AF,$F6,$F1,$88,$F4,$35,$F4,$1C,$48
	.dc.b	$7C,$5A,$F6,$BF,$10,$D9,$FB,$CE,$7F,$BA,$FF,$30,$E6,$FB,$AF,$F3
	.dc.b	$2E,$5F,$BA,$FF,$34,$E4,$FB,$AF,$F9,$02,$DC,$9F,$7A,$3D,$87,$88
	.dc.b	$F3,$94,$FC,$C6,$5F,$ED,$3D,$5F,$15,$FE,$1B,$6C,$FF,$89,$DA,$FF
	.dc.b	$88,$9A,$FF,$8C,$1A,$FF,$A5,$1A,$FF,$A4,$1B,$C8,$69,$27,$C0,$49
	.dc.b	$F0,$35,$F0,$35,$F0,$35,$F0,$3E,$9F,$88,$FE,$17,$2F,$88,$BF,$11
	.dc.b	$8F,$88,$98,$7E,$1B,$E2,$37,$F1,$1C,$3D,$7E,$3F,$05,$CB,$DF,$E7
	.dc.b	$AF,$7C,$3D,$8D,$DB,$D8,$DD,$7D,$93,$FC,$34,$D9,$EC,$21,$EC,$29
	.dc.b	$EC,$29,$EC,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$F5,$FE,$0F
	.dc.b	$E3,$ED,$F2,$3E,$03,$E4,$FC,$27,$CA,$F8,$1F,$97,$F1,$1F,$33,$E0
	.dc.b	$BE,$36,$3F,$AE,$F8,$8F,$8C,$FE,$C3,$4F,$84,$47,$D3,$AF,$C6,$7E
	.dc.b	$0F,$37,$E4,$67,$51,$E7,$9D,$37,$6F,$96,$D9,$DB,$EA,$3D,$A7,$C0
	.dc.b	$31,$3C,$9B,$3F,$90,$E6,$FC,$2F,$C1,$FE,$12,$7F,$83,$F3,$7E,$11
	.dc.b	$FF,$55,$D3,$9B,$F0,$B1,$FC,$33,$9B,$F0,$AF,$F0,$EF,$88,$FC,$2B
	.dc.b	$EE,$AF,$3B,$E0,$CF,$DB,$A7,$BE,$35,$1E,$A7,$D0,$43,$D6,$4D,$3D
	.dc.b	$4E,$5A,$E6,$BB,$B9,$6B,$9E,$4F,$BA,$FF,$24,$E3,$FB,$AF,$F2,$6E
	.dc.b	$2F,$BA,$FF,$28,$E1,$FB,$AF,$C2,$E1,$FB,$AF,$C4,$CB,$D0,$F8,$5B
	.dc.b	$97,$D8,$CF,$F8,$66,$5F,$B2,$25,$7F,$34,$6B,$F3,$9B,$E0,$FF,$82
	.dc.b	$34,$46,$3C,$EF,$CB,$57,$12,$1C,$4B,$5C,$67,$84,$7F,$11,$6E,$1F
	.dc.b	$C4,$5B,$7D,$A7,$8D,$F6,$7A,$12,$FA,$3A,$3A,$16,$FA,$36,$BA,$3C
	.dc.b	$6B,$A5,$72,$8C,$1C,$87,$C3,$3C,$DD,$D4,$9D,$57,$12,$46,$70,$F0
	.dc.b	$CF,$46,$CE,$A3,$8C,$EF,$0F,$2C,$E6,$B7,$39,$C7,$E2,$92,$FD,$8D
	.dc.b	$7B,$5F,$29,$E1,$1A,$21,$A9,$E1,$8F,$B6,$3F,$A2,$7C,$2F,$F3,$62
	.dc.b	$3F,$DE,$1A,$FE,$C9,$B8,$FE,$15,$B7,$FD,$3D,$B3,$D0,$E4,$7F,$80
	.dc.b	$BE,$23,$FC,$9B,$F9,$42,$AF,$C3,$E3,$6C,$FA,$9A,$F7,$FF,$A4,$F5
	.dc.b	$F0,$F8,$7E,$3E,$E7,$2F,$7F,$9F,$C3,$F4,$F8,$7E,$BE,$4F,$6F,$87
	.dc.b	$EF,$F0,$FE,$0F,$87,$E9,$C7,$FE,$CC,$E2,$6F,$E1,$8C,$70,$F9,$B8
	.dc.b	$BF,$AC,$F1,$FB,$3A,$7F,$B4,$72,$FF,$A9,$B9,$FE,$54,$7D,$8C,$3E
	.dc.b	$E9,$1C,$13,$A7,$EB,$46,$5B,$19,$E9,$FB,$9D,$3F,$83,$9C,$92,$33
	.dc.b	$97,$DE,$BF,$B8,$7D,$EA,$E1,$3A,$8E,$CD,$BE,$D1,$BA,$FD,$76,$1E
	.dc.b	$AC,$3F,$4F,$3F,$E0,$BB,$FF,$09,$11,$FF,$09,$35,$FF,$09,$35,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$10,$FE,$04,$53,$F6,$C5,$BF,$D7,$B6,$7D
	.dc.b	$2F,$02,$1C,$0B,$1C,$1B,$1C,$1D,$12,$34,$D8,$FE,$17,$D3,$FC,$B2
	.dc.b	$3D,$67,$2A,$F8,$4D,$32,$71,$72,$D3,$35,$D8,$95,$E3,$6C,$F0,$6C
	.dc.b	$F0,$9D,$12,$74,$D8,$D4,$60,$8E,$0F,$F1,$24,$F4,$7E,$91,$E9,$29
	.dc.b	$E7,$2B,$B7,$E8,$47,$47,$33,$E1,$6C,$65,$D9,$E1,$32,$53,$26,$08
	.dc.b	$7E,$1E,$BE,$83,$83,$CC,$70,$21,$82,$1E,$93,$25,$33,$5E,$B6,$D7
	.dc.b	$A5,$DC,$65,$E0,$47,$E4,$4F,$93,$1F,$D8,$CF,$EB,$4F,$D8,$96,$7F
	.dc.b	$62,$FE,$E0,$91,$C1,$B8,$FE,$F6,$30,$47,$04,$8E,$A7,$02,$C7,$3C
	.dc.b	$7F,$8A,$EC,$F8,$3F,$E6,$24,$AF,$E4,$4D,$55,$FC,$D5,$60,$AE,$51
	.dc.b	$C0,$87,$0C,$0F,$08,$DB,$93,$17,$DE,$FB,$5F,$58,$FE,$8B,$F0,$FF
	.dc.b	$D1,$BE,$13,$FA,$3F,$C1,$FF,$48,$F8,$7F,$E9,$3F,$09,$FD,$2B,$E1
	.dc.b	$3B,$E7,$D7,$27,$E5,$2F,$88,$CE,$BD,$A1,$F1,$18,$E0,$F3,$DF,$0F
	.dc.b	$6F,$18,$FC,$0B,$7B,$78,$AE,$BF,$0B,$65,$F0,$17,$E0,$31,$F0,$99
	.dc.b	$F8,$7D,$7C,$3E,$F8,$3D,$1A,$7F,$51,$E2,$F6,$3F,$D4,$F9,$3E,$03
	.dc.b	$FD,$57,$9B,$E3,$70,$F9,$BA,$3E,$5F,$4F,$F3,$7E,$CF,$E7,$9D,$9F
	.dc.b	$D1,$70,$FC,$73,$5F,$C2,$71,$FE,$43,$9F,$F1,$51,$F8,$6C,$7F,$15
	.dc.b	$39,$6D,$71,$AD,$CE,$6B,$81,$1C,$1F,$9C,$C3,$F1,$FD,$7E,$89,$FB
	.dc.b	$E7,$E1,$AB,$EF,$E5,$F8,$46,$CE,$EF,$F0,$55,$BC,$7E,$04,$E6,$F8
	.dc.b	$47,$C3,$CF,$E3,$D7,$E0,$65,$A7,$CC,$8C,$6B,$F4,$1F,$C3,$72,$7E
	.dc.b	$15,$FC,$BE,$7F,$85,$E8,$FB,$9D,$1F,$77,$A3,$EF,$74,$7D,$FE,$8F
	.dc.b	$BD,$87,$0D,$BF,$13,$93,$9F,$F0,$13,$93,$F1,$5C,$DF,$81,$7F,$C2
	.dc.b	$1D,$3F,$9A,$7F,$C2,$5D,$5F,$9E,$7F,$C2,$9D,$7F,$A2,$7E,$11,$F2
	.dc.b	$8F,$F2,$AF,$99,$E0,$DF,$F2,$44,$8F,$07,$37,$E8,$A3,$11,$F0,$1B
	.dc.b	$C3,$EF,$8D,$E9,$FA,$46,$CE,$B5,$ED,$18,$21,$CC,$B5,$CD,$C1,$1C
	.dc.b	$12,$39,$C6,$5B,$19,$8F,$D8,$4F,$85,$A7,$69,$D9,$1C,$9C,$CB,$3C
	.dc.b	$EF,$E1,$47,$A3,$9F,$FC,$6F,$87,$BE,$38,$4E,$9B,$19,$AE,$4D,$BF
	.dc.b	$05,$6E,$73,$87,$9E,$31,$3A,$AE,$17,$E4,$8C,$56,$E7,$35,$C2,$B4
	.dc.b	$47,$04,$8D,$92,$79,$12,$7E,$31,$C6,$34,$FC,$1C,$7F,$2F,$F0,$5F
	.dc.b	$C6,$49,$1F,$82,$6B,$FA,$66,$E3,$FC,$A5,$B8,$F2,$9E,$04,$3F,$5B
	.dc.b	$5F,$C9,$7E,$0B,$F9,$37,$C3,$FF,$27,$F8,$3F,$E5,$1F,$07,$FC,$A7
	.dc.b	$E0,$FF,$95,$7C,$3F,$FB,$13,$57,$C3,$FF,$2D,$F8,$7F,$E5,$DF,$0F
	.dc.b	$FC,$C7,$C3,$FE,$7F,$B7,$F3,$F5,$EB,$E7,$D6,$D7,$C1,$EF,$E1,$FA
	.dc.b	$DF,$E7,$B5,$F0,$BF,$A0,$DF,$CA,$FE,$8B,$87,$F9,$E7,$17,$BA,$3F
	.dc.b	$A1,$72,$7F,$9F,$E3,$EC,$C7,$CE,$E7,$FA,$9C,$FF,$46,$3F,$12,$DE
	.dc.b	$7F,$C2,$C6,$F8,$FE,$9B,$A3,$13,$BE,$7E,$B1,$92,$57,$26,$C6,$A3
	.dc.b	$07,$EC,$7A,$3F,$3B,$5F,$1A,$BF,$5E,$7F,$66,$C3,$F3,$CD,$C3,$F4
	.dc.b	$4D,$9E,$0F,$EB,$0C,$1C,$E7,$8B,$D0,$86,$CA,$E9,$C1,$1C,$12,$7C
	.dc.b	$31,$C1,$AE,$DB,$1B,$9F,$2B,$67,$C9,$3A,$24,$E9,$B1,$A8,$C1,$1C
	.dc.b	$10,$FC,$6C,$FF,$03,$F2,$DE,$9F,$96,$B8,$56,$AF,$AB,$63,$2F,$C8
	.dc.b	$30,$75,$30,$43,$A1,$92,$9C,$08,$6C,$F8,$8F,$67,$66,$6F,$E2,$3C
	.dc.b	$0C,$90,$E2,$77,$3F,$94,$DF,$B9,$C9,$C4,$A7,$5A,$F7,$B6,$BC,$4E
	.dc.b	$8E,$D5,$E2,$74,$77,$70,$EA,$79,$B6,$79,$CE,$A3,$A1,$1F,$A9,$7B
	.dc.b	$8E,$95,$C9,$D3,$67,$91,$1C,$12,$BB,$B6,$79,$C6,$08,$E0,$87,$C5
	.dc.b	$9E,$91,$FE,$52,$BF,$95,$1F,$C1,$27,$F7,$A4,$3F,$72,$7F,$35,$9F
	.dc.b	$E0,$0D,$35,$8F,$EF,$32,$FD,$F1,$23,$47,$2A,$EC,$DC,$3F,$86,$D7
	.dc.b	$18,$DB,$FC,$17,$3F,$F2,$FF,$23,$E0,$7F,$96,$12,$BF,$C1,$31,$96
	.dc.b	$C6,$5F,$E6,$0D,$AF,$F3,$CC,$EE,$72,$70,$21,$FC,$DF,$E0,$7F,$9D
	.dc.b	$7C,$0F,$F3,$BF,$87,$FE,$79,$F0,$FF,$CF,$7E,$1F,$F9,$F6,$1F,$CF
	.dc.b	$FE,$0F,$FA,$07,$C3,$FF,$41,$F8,$4F,$E8,$5F,$0F,$FD,$0F,$E1,$7F
	.dc.b	$A2,$7C,$2F,$F4,$5F,$87,$FE,$8D,$F0,$FF,$D1,$FD,$BF,$E9,$1E,$E7
	.dc.b	$3E,$2F,$6E,$BD,$DF,$7E,$FE,$37,$E0,$F0,$FF,$4F,$E2,$F7,$74,$7C
	.dc.b	$8E,$5F,$E8,$55,$F3,$23,$EA,$74,$7D,$2E,$6F,$A0,$D8,$CF,$47,$D4
	.dc.b	$6F,$4F,$E0,$27,$57,$CC,$8E,$09,$5C,$63,$2D,$8F,$C0,$F6,$3A,$BC
	.dc.b	$D5,$F4,$31,$FA,$B9,$7A,$A3,$FD,$77,$B7,$D7,$9C,$C6,$F3,$E2,$FD
	.dc.b	$A2,$46,$9B,$1F,$8A,$D3,$ED,$C7,$E2,$3A,$BE,$DB,$F8,$AE,$AF,$B6
	.dc.b	$FE,$33,$AB,$ED,$BF,$8E,$E9,$FB,$6F,$E4,$3A,$3E,$DB,$F9,$2E,$7F
	.dc.b	$B6,$FE,$53,$9B,$ED,$BC,$CF,$B6,$77,$3E,$D9,$E3,$3E,$D9,$F9,$CC
	.dc.b	$BE,$17,$0B,$6F,$CC,$70,$FC,$9A,$FF,$6F,$FE,$55,$FE,$E1,$7F,$A1
	.dc.b	$B5,$FE,$D4,$D7,$FD,$92,$D7,$8E,$FF,$D5,$A3,$16,$69,$23,$44,$9F
	.dc.b	$ED,$D1,$96,$C6,$5F,$1B,$5F,$1B,$5F,$F2,$AC,$3D,$EF,$F6,$C2,$ED
	.dc.b	$A6,$3D,$FC,$FC,$46,$BE,$23,$7F,$11,$C3,$E2,$38,$E7,$D7,$3F,$69
	.dc.b	$E4,$21,$E4,$29,$E4,$33,$8F,$F2,$4C,$BD,$91,$B2,$1B,$29,$F3,$27
	.dc.b	$C1,$A7,$C9,$6C,$F2,$9E,$04,$38,$1E,$22,$9D,$0F,$49,$FD,$FB,$67
	.dc.b	$85,$F1,$BE,$23,$E3,$FC,$47,$C8,$F1,$7E,$35,$FF,$AB,$30,$6F,$F3
	.dc.b	$3E,$23,$E6,$FC,$47,$C7,$AD,$69,$E6,$23,$E3,$3C,$27,$9C,$F1,$ED
	.dc.b	$F1,$63,$71,$E7,$DB,$E0,$3C,$5A,$E7,$1F,$84,$FB,$8E,$91,$A6,$C6
	.dc.b	$9F,$09,$B2,$1F,$E8,$F7,$F1,$0F,$BE,$BF,$0E,$D3,$E1,$7C,$47,$E1
	.dc.b	$DF,$89,$C3,$F8,$87,$E0,$9E,$7A,$F0,$5F,$8E,$3B,$B5,$F1,$9F,$0B
	.dc.b	$4C,$CF,$02,$63,$F1,$27,$D6,$D7,$6D,$79,$D7,$52,$70,$7D,$E8,$FB
	.dc.b	$5B,$FD,$E7,$EE,$61,$F7,$97,$73,$0F,$DE,$EF,$F7,$E7,$F0,$5B,$7D
	.dc.b	$F7,$73,$E2,$7E,$03,$E3,$CB,$C5,$9F,$E8,$36,$FD,$17,$C2,$7F,$05
	.dc.b	$25,$FF,$00,$6B,$FC,$41,$AF,$F1,$06,$BF,$C4,$1A,$FF,$18,$6D,$7F
	.dc.b	$0F,$6D,$1A,$63,$E1,$3F,$82,$FB,$3F,$C1,$BE,$0F,$F8,$3F,$B7,$31
	.dc.b	$FD,$07,$C1,$FF,$0A,$F7,$51,$38,$3E,$32,$D1,$F9,$5B,$E1,$F7,$3F
	.dc.b	$18,$D4,$65,$FD,$66,$BE,$47,$E1,$10,$F8,$45,$3E,$11,$9D,$FE,$5B
	.dc.b	$93,$E2,$14,$F8,$85,$BF,$88,$DD,$FE,$23,$6B,$C4,$F8,$08,$78,$0A
	.dc.b	$76,$29,$FD,$23,$7F,$7E,$5F,$12,$FE,$1F,$97,$F3,$6B,$E4,$70,$FC
	.dc.b	$A9,$F8,$FF,$0B,$C2,$31,$A3,$E1,$71,$7F,$23,$E1,$7E,$93,$C8,$F9
	.dc.b	$7F,$0B,$AC,$7E,$8F,$C4,$FF,$40,$DB,$D7,$5E,$92,$1E,$92,$9F,$18
	.dc.b	$F8,$1C,$1F,$31,$F6,$D7,$C8,$6F,$07,$C8,$6D,$76,$7A,$10,$E8,$76
	.dc.b	$9E,$4D,$9E,$4D,$9F,$CC,$E6,$F8,$6F,$CA,$CF,$F3,$1E,$7F,$CA,$3F
	.dc.b	$8B,$C3,$FB,$67,$37,$F5,$EE,$6F,$85,$83,$E2,$6F,$34,$C3,$F1,$E7
	.dc.b	$DF,$21,$EA,$2E,$BF,$05,$F2,$1A,$29,$B2,$18,$21,$F0,$71,$F7,$B7
	.dc.b	$5F,$7B,$FE,$08,$E6,$F3,$F2,$FE,$16,$3D,$99,$39,$7D,$7C,$9E,$CE
	.dc.b	$4F,$C0,$D6,$B9,$3D,$FC,$9F,$7E,$3C,$C7,$7E,$4F,$A8,$FD,$23,$F3
	.dc.b	$1B,$7F,$A5,$B2,$7C,$3F,$FB,$B1,$B3,$FE,$FF,$6B,$FE,$E4,$6B,$FB
	.dc.b	$CD,$FF,$7F,$3C,$49,$1C,$5A,$F1,$6B,$C6,$79,$BF,$A0,$8C,$B4,$E4
	.dc.b	$43,$91,$4E,$5E,$D7,$E0,$3D,$3F,$C0,$FB,$7F,$82,$F8,$7F,$C1,$FB
	.dc.b	$BF,$84,$F8,$7F,$C2,$FB,$FF,$86,$F8,$7F,$C3,$FC,$3F,$E2,$34,$F0
	.dc.b	$E9,$FC,$59,$BB,$7B,$A3,$C0,$79,$4F,$19,$37,$F9,$93,$C3,$7F,$6C
	.dc.b	$F3,$38,$94,$E2,$7C,$38,$E0,$47,$75,$E1,$76,$78,$4E,$06,$A7,$BC
	.dc.b	$7F,$32,$DF,$E2,$4F,$F0,$78,$FE,$6B,$F0,$FF,$16,$31,$4F,$87,$F8
	.dc.b	$D1,$BF,$CB,$3F,$9C,$BF,$94,$D3,$F8,$74,$3E,$23,$E4,$4F,$87,$3F
	.dc.b	$43,$E6,$3D,$C7,$98,$F8,$DE,$4F,$A5,$86,$8B,$9F,$BD,$E0,$43,$39
	.dc.b	$7B,$89,$5E,$08,$D1,$CC,$93,$CD,$E3,$1E,$C7,$F9,$96,$0F,$88,$F6
	.dc.b	$FC,$47,$E2,$A7,$E1,$7C,$47,$E2,$9F,$89,$F0,$3F,$8A,$7E,$37,$B5
	.dc.b	$F8,$A7,$E5,$E3,$E3,$7B,$9F,$14,$87,$F9,$CF,$4C,$BF,$1F,$3F,$33
	.dc.b	$E8,$D3,$DA,$F0,$3D,$B1,$A2,$46,$A3,$93,$5D,$BF,$1F,$87,$F0,$71
	.dc.b	$FC,$3B,$87,$F0,$4F,$F1,$0E,$1F,$C0,$BF,$CB,$77,$FC,$0F,$95,$FD
	.dc.b	$6F,$07,$17,$E0,$B8,$BF,$03,$7E,$0D,$7E,$03,$E9,$7F,$07,$C6,$DB
	.dc.b	$E7,$FC,$1F,$D1,$AF,$F0,$46,$68,$C6,$E8,$C4,$FD,$78,$C4,$EA,$73
	.dc.b	$3F,$96,$9C,$BB,$8C,$BF,$C5,$DA,$F8,$8D,$4F,$32,$38,$21,$D6,$38
	.dc.b	$BE,$88,$DB,$96,$BD,$FE,$0B,$F4,$FF,$09,$FA,$8F,$83,$FD,$4F,$A9
	.dc.b	$FA,$AF,$83,$FD,$5F,$B5,$FA,$CF,$83,$FD,$6F,$C4,$7E,$BB,$E1,$3F
	.dc.b	$5F,$A7,$4D,$3D,$07,$73,$CA,$7B,$5D,$3E,$53,$D3,$C1,$F4,$5C,$1C
	.dc.b	$48,$71,$39,$9D,$8C,$1A,$30,$43,$99,$A2,$19,$DF,$E9,$BC,$4D,$99
	.dc.b	$DF,$E5,$C7,$DF,$DF,$F8,$69,$F1,$E7,$F8,$97,$C4,$7C,$77,$F8,$A6
	.dc.b	$9F,$C4,$BE,$17,$E3,$C7,$F1,$7C,$FE,$3B,$EA,$3B,$93,$0F,$1B,$E6
	.dc.b	$31,$8F,$88,$FC,$5E,$5F,$1E,$77,$3E,$7C,$FE,$6C,$62,$38,$C6,$2B
	.dc.b	$93,$49,$1E,$39,$E0,$49,$DC,$7F,$6F,$EA,$FC,$3B,$FC,$C7,$E2,$3F
	.dc.b	$0E,$FF,$33,$F8,$6F,$E6,$3F,$11,$F8,$78,$FE,$79,$E4,$FF,$9C,$9F
	.dc.b	$F1,$23,$FE,$BA,$7F,$75,$97,$89,$FA,$59,$78,$63,$11,$F7,$B3,$F8
	.dc.b	$19,$F9,$23,$6D,$7E,$45,$71,$7D,$91,$97,$E3,$18,$AE,$AE,$0F,$D5
	.dc.b	$70,$7D,$E7,$FA,$77,$2F,$F4,$CE,$1F,$BD,$1F,$C4,$78,$7E,$F3,$FD
	.dc.b	$4F,$8B,$F4,$8F,$F5,$5C,$3A,$BF,$48,$F0,$10,$FE,$5C,$7F,$0E,$C3
	.dc.b	$C0,$FF,$1D,$6F,$67,$F9,$DF,$24,$63,$D9,$FE,$8A,$DA,$E4,$7F,$65
	.dc.b	$71,$A7,$E6,$63,$04,$3B,$16,$FB,$4F,$0B,$EF,$39,$BE,$ED,$3B,$9C
	.dc.b	$89,$5C,$9D,$5F,$85,$CC,$F2,$69,$CA,$78,$38,$9E,$11,$83,$3F,$0E
	.dc.b	$DF,$94,$66,$E9,$2F,$A3,$6F,$9B,$70,$F6,$CF,$76,$D7,$38,$E2,$4A
	.dc.b	$E2,$D9,$DD,$73,$C7,$E4,$12,$7E,$0C,$E9,$E1,$19,$7C,$0D,$BF,$05
	.dc.b	$F6,$25,$79,$B1,$F8,$17,$FE,$DB,$C5,$BF,$FA,$B5,$B1,$FB,$08,$C1
	.dc.b	$1C,$10,$FD,$9B,$96,$99,$C3,$FD,$8D,$3D,$23,$F7,$3F,$0D,$DF,$DF
	.dc.b	$F0,$7C,$37,$87,$E0,$BC,$5F,$0D,$E3,$F8,$3F,$27,$C3,$7F,$A8,$3A
	.dc.b	$BF,$84,$E5,$FC,$0B,$C1,$F4,$7C,$37,$A7,$E1,$7D,$5F,$0D,$EB,$F5
	.dc.b	$BD,$9E,$EF,$B7,$E0,$F9,$E3,$FC,$33,$4E,$EF,$F0,$ED,$BC,$6F,$F1
	.dc.b	$0E,$0F,$89,$C3,$ED,$E3,$F5,$F2,$7F,$C0,$0E,$5F,$97,$CF,$F3,$BA
	.dc.b	$3E,$97,$4F,$DB,$CB,$EC,$C6,$1F,$B2,$60,$87,$DA,$2E,$1F,$6B,$A7
	.dc.b	$F0,$17,$D6,$73,$7D,$5E,$04,$AE,$17,$DE,$BF,$51,$87,$9B,$0F,$CB
	.dc.b	$DF,$F7,$93,$FC,$15,$5F,$E0,$F2,$38,$21,$FC,$70,$A7,$13,$04,$30
	.dc.b	$43,$F6,$C5,$3F,$7E,$5A,$FD,$E3,$67,$CA,$F0,$21,$C0,$B1,$C1,$B1
	.dc.b	$C1,$D1,$23,$4D,$8D,$46,$08,$E0,$FD,$DE,$3F,$D1,$AB,$E1,$D7,$B4
	.dc.b	$EE,$E1,$F4,$35,$C9,$F1,$8A,$71,$3C,$4E,$DA,$6C,$A6,$EB,$CB,$1B
	.dc.b	$6C,$6D,$AE,$EB,$D5,$3B,$7D,$7C,$9E,$29,$F3,$9C,$63,$6D,$9E,$04
	.dc.b	$70,$43,$B3,$96,$99,$3A,$98,$21,$82,$1F,$12,$FE,$73,$C4,$F3,$10
	.dc.b	$F8,$67,$02,$99,$BF,$82,$E5,$AE,$6B,$DA,$F8,$8C,$57,$A5,$E1,$1A
	.dc.b	$7F,$39,$5E,$47,$F5,$99,$7E,$BF,$2F,$D6,$C7,$EC,$8F,$DD,$9A,$2B
	.dc.b	$A7,$18,$7E,$E5,$B7,$FB,$C2,$61,$FB,$76,$C7,$59,$E1,$5C,$CF,$EB
	.dc.b	$3D,$29,$FE,$75,$F0,$DF,$CE,$48,$FF,$8A,$1A,$FF,$8A,$1B,$3F,$CD
	.dc.b	$1B,$8F,$3B,$FE,$C9,$18,$AF,$E8,$DF,$09,$FD,$17,$E1,$3F,$A3,$7C
	.dc.b	$37,$86,$58,$B1,$E5,$F9,$6B,$E1,$75,$39,$A6,$6A,$7C,$26,$B0,$E3
	.dc.b	$F0,$96,$F7,$4F,$84,$C6,$1C,$7E,$12,$CB,$E1,$2F,$C2,$63,$E1,$F3
	.dc.b	$F0,$FA,$F5,$B7,$E2,$F0,$E6,$E3,$F0,$DC,$B3,$E7,$AF,$4D,$FA,$F0
	.dc.b	$F6,$E3,$F8,$3C,$3F,$E5,$4E,$5F,$ED,$1C,$DF,$1F,$8F,$E4,$CF,$D5
	.dc.b	$EC,$FE,$61,$D5,$F7,$49,$1F,$77,$AB,$EF,$BF,$84,$8C,$B4,$D3,$8E
	.dc.b	$8F,$C1,$74,$7E,$4C,$93,$F9,$06,$E1,$FD,$03,$87,$EE,$0E,$AF,$33
	.dc.b	$6B,$F0,$C4,$3F,$0C,$5B,$FC,$33,$6F,$9D,$72,$CF,$F0,$93,$FE,$76
	.dc.b	$EA,$F8,$CF,$E5,$6F,$F0,$18,$FD,$FD,$FE,$64,$E7,$1F,$C3,$F0,$FD
	.dc.b	$18,$C1,$F8,$0E,$9F,$B6,$E2,$FF,$11,$D1,$F7,$23,$1D,$4E,$8F,$BB
	.dc.b	$18,$DF,$EE,$BA,$E6,$FB,$AF,$0E,$6F,$BA,$F2,$E6,$FB,$AF,$FA,$3B
	.dc.b	$9B,$EE,$BF,$04,$FD,$79,$F7,$6F,$F3,$FF,$0D,$FE,$FC,$C8,$D2,$4F
	.dc.b	$FB,$F6,$32,$D8,$CB,$FE,$FD,$8C,$B6,$32,$FE,$BF,$4F,$D4,$46,$23
	.dc.b	$C2,$73,$21,$C4,$A7,$19,$EA,$D8,$F4,$7C,$37,$EA,$BE,$1B,$F5,$7F
	.dc.b	$0D,$FA,$CF,$77,$F5,$BF,$0D,$FA,$EF,$81,$FD,$7F,$C0,$FE,$CB,$F2
	.dc.b	$EF,$E0,$59,$FF,$02,$E7,$F1,$BE,$12,$1D,$CF,$19,$A3,$06,$8C,$1C
	.dc.b	$BC,$4F,$F6,$B6,$DE,$88,$C9,$FC,$13,$6F,$84,$D3,$A9,$4E,$85,$3B
	.dc.b	$90,$E7,$BF,$CC,$78,$B5,$E2,$D7,$81,$CC,$FE,$3F,$BF,$C3,$F8,$1F
	.dc.b	$89,$F0,$FF,$17,$E1,$FE,$4D,$7E,$5B,$7F,$92,$FE,$6F,$5F,$93,$DB
	.dc.b	$FA,$6D,$3D,$D3,$D4,$99,$F8,$5F,$3C,$6A,$31,$1A,$8C,$1F,$8A,$D7
	.dc.b	$E4,$EB,$F1,$AB,$64,$9D,$C6,$5B,$5F,$09,$AF,$C2,$6B,$F0,$A7,$94
	.dc.b	$F5,$C1,$F0,$FF,$8B,$AF,$DA,$F1,$FE,$2D,$F8,$1F,$0F,$F8,$B7,$F7
	.dc.b	$3C,$7F,$8B,$F8,$7F,$E7,$78,$F8,$C9,$3E,$36,$BF,$8D,$D3,$DA,$E4
	.dc.b	$87,$9E,$36,$F9,$F4,$F4,$BF,$2A,$32,$D3,$E0,$71,$7D,$B7,$EB,$70
	.dc.b	$FD,$B7,$ED,$10,$FD,$01,$FD,$5B,$93,$F4,$8F,$A8,$F1,$9E,$1A,$EE
	.dc.b	$FD,$77,$F8,$0E,$0F,$2F,$FB,$96,$EF,$67,$F5,$B7,$86,$BF,$DA,$A3
	.dc.b	$4E,$B8,$51,$88,$FD,$AC,$62,$78,$B6,$38,$4E,$DF,$E1,$CD,$7C,$06
	.dc.b	$A7,$8C,$66,$3C,$51,$82,$3B,$21,$E2,$8E,$13,$E1,$8C,$4F,$4F,$1B
	.dc.b	$F8,$67,$C3,$7F,$0D,$F7,$BF,$87,$7C,$37,$F0,$FF,$86,$FE,$21,$EF
	.dc.b	$FF,$11,$F8,$6F,$E2,$5F,$01,$FC,$4F,$2F,$0E,$7D,$0F,$09,$EC,$72
	.dc.b	$D3,$27,$B8,$E0,$7F,$B9,$36,$C3,$B3,$25,$21,$A2,$1F,$10,$EA,$53
	.dc.b	$89,$AD,$7E,$13,$C8,$D1,$5C,$BB,$DB,$E4,$8F,$85,$FE,$65,$F0,$5F
	.dc.b	$1E,$7F,$86,$6B,$FC,$D7,$E1,$BE,$3C,$7F,$0F,$F8,$6F,$E7,$1F,$0D
	.dc.b	$F1,$E3,$CB,$97,$67,$C2,$77,$D3,$E3,$B4,$FB,$7C,$9F,$97,$76,$49
	.dc.b	$E4,$4D,$3C,$F5,$CA,$75,$18,$7C,$2D,$7E,$29,$A8,$E1,$3C,$DF,$C3
	.dc.b	$47,$F2,$9E,$3F,$EA,$5F,$0D,$FE,$44,$FC,$BF,$FC,$9D,$3F,$B4,$BF
	.dc.b	$D1,$F2,$67,$E2,$BF,$11,$3F,$CC,$78,$FE,$13,$E0,$AE,$CF,$93,$2F
	.dc.b	$0E,$7E,$58,$FC,$16,$9F,$06,$B7,$1C,$67,$4F,$C7,$31,$5C,$88,$F5
	.dc.b	$21,$A3,$F6,$7C,$5F,$71,$FE,$8D,$CD,$FA,$27,$FA,$47,$0F,$DC,$7D
	.dc.b	$D6,$E1,$FB,$B1,$FD,$37,$0E,$EF,$52,$1F,$8D,$9F,$A1,$97,$AF,$63
	.dc.b	$FC,$19,$B3,$FC,$35,$AF,$F0,$C6,$BF,$E1,$56,$BF,$E8,$B6,$BF,$E8
	.dc.b	$06,$E6,$69,$27,$C0,$49,$F0,$35,$F0,$35,$F0,$35,$F0,$3E,$9F,$73
	.dc.b	$F0,$9E,$EF,$E1,$7E,$23,$F0,$DE,$FF,$E1,$FE,$23,$F1,$1F,$03,$F8
	.dc.b	$9F,$88,$FC,$57,$5F,$E2,$F2,$E9,$9F,$A5,$BA,$FA,$5B,$AF,$A6,$7C
	.dc.b	$E4,$3F,$85,$1C,$4C,$1E,$82,$9E,$82,$9B,$3D,$E6,$27,$C0,$D8,$F0
	.dc.b	$3A,$21,$D4,$A7,$52,$9F,$4F,$5F,$8A,$FF,$9B,$61,$A7,$C6,$8C,$7C
	.dc.b	$47,$EB,$7E,$23,$E3,$46,$FE,$23,$E3,$3E,$3C,$BE,$19,$1C,$1F,$0D
	.dc.b	$FC,$71,$97,$C7,$97,$B1,$B8,$FB,$1F,$51,$0F,$51,$4F,$8C,$58,$E4
	.dc.b	$D8,$E4,$F0,$3F,$B6,$F3,$7F,$07,$F8,$8F,$E1,$1F,$11,$FC,$27,$E2
	.dc.b	$3F,$09,$5F,$0F,$87,$F8,$67,$C1,$7E,$12,$3E,$36,$7E,$97,$B3,$5E
	.dc.b	$C7,$90,$FA,$0F,$C0,$7E,$19,$EE,$78,$3B,$26,$7E,$A7,$9B,$5C,$E7
	.dc.b	$EC,$6C,$F2,$9F,$D6,$70,$7D,$C9,$FE,$45,$5F,$71,$FB,$DC,$9F,$C9
	.dc.b	$F8,$3E,$E4,$7E,$07,$8F,$F9,$57,$07,$DC,$8F,$31,$E8,$3D,$E5,$3D
	.dc.b	$BC,$7F,$BF,$F8,$2F,$E1,$04,$8F,$E1,$0D,$7F,$84,$35,$FE,$10,$D7
	.dc.b	$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$F8,$A7,$B3,$FD,$17,$E0,$91
	.dc.b	$FE,$CB,$C5,$ED,$E2,$3F,$A4,$43,$E0,$B1,$EA,$FF,$B6,$A1,$EE,$E3
	.dc.b	$DA,$E7,$C4,$97,$C1,$67,$14,$E7,$DF,$FE,$AD,$B7,$BE,$7F,$92,$36
	.dc.b	$7F,$92,$37,$8C,$D2,$4F,$AB,$97,$F9,$C4,$62,$3E,$01,$F0,$CF,$8C
	.dc.b	$60,$EE,$7C,$82,$99,$AF,$0B,$96,$B9,$AF,$1B,$96,$BC,$71,$7C,$17
	.dc.b	$CF,$8F,$9D,$BF,$C3,$BF,$8C,$DA,$7C,$45,$F8,$3C,$69,$F1,$B0,$ED
	.dc.b	$9F,$B5,$F5,$90,$F7,$10,$F5,$9F,$03,$4F,$8A,$F2,$21,$EB,$8E,$8F
	.dc.b	$1C,$BD,$EF,$53,$99,$9D,$3D,$EE,$63,$6E,$67,$FB,$47,$1F,$B7,$E2
	.dc.b	$3D,$DF,$09,$EF,$E4,$F8,$1C,$7F,$07,$8B,$E1,$7C,$47,$BF,$2F,$21
	.dc.b	$E6,$3B,$69,$E0,$7B,$1E,$AD,$7D,$AF,$02,$1C,$34,$F6,$BE,$A2,$6B
	.dc.b	$E8,$6D,$75,$6C,$F0,$9D,$1F,$9D,$E3,$FB,$93,$FA,$6E,$3F,$B4,$7D
	.dc.b	$BE,$2F,$B5,$1F,$AD,$E1,$F6,$3E,$33,$D8,$7A,$8F,$39,$72,$FF,$16
	.dc.b	$71,$BE,$17,$FD,$2E,$D9,$FF,$4F,$35,$FF,$4E,$B5,$FD,$D6,$5F,$B3
	.dc.b	$9E,$24,$8E,$2D,$78,$B5,$E3,$3C,$DF,$DE,$B5,$FD,$DB,$67,$B4,$74
	.dc.b	$21,$D0,$A7,$4F,$88,$FD,$07,$B1,$FA,$1F,$6F,$F4,$5F,$0D,$FA,$3F
	.dc.b	$81,$FD,$27,$C3,$7E,$97,$E1,$3F,$4D,$A7,$97,$2F,$1B,$5F,$1B,$73
	.dc.b	$F1,$C7,$84,$87,$B2,$3E,$06,$BF,$39,$E1,$AF,$AC,$8E,$09,$3E,$A3
	.dc.b	$C2,$53,$C2,$5A,$F0,$BD,$CC,$4F,$47,$EC,$EB,$FC,$07,$E0,$BF,$81
	.dc.b	$7B,$FF,$0A,$7F,$9A,$69,$F0,$9F,$E6,$D9,$FC,$27,$9E,$5E,$47,$DA
	.dc.b	$79,$0A,$79,$0B,$97,$C4,$CF,$E0,$47,$22,$1A,$CF,$CC,$DA,$EC,$FA
	.dc.b	$CC,$1C,$0C,$10,$FE,$87,$F9,$87,$F4,$B7,$FA,$5F,$07,$DF,$7F,$A8
	.dc.b	$F8,$0F,$03,$7F,$0F,$FC,$96,$DF,$05,$7E,$1F,$1F,$0F,$F0,$71,$F0
	.dc.b	$90,$F0,$1E,$F2,$ED,$F4,$9F,$88,$7A,$76,$F8,$2F,$13,$24,$39,$16
	.dc.b	$39,$38,$23,$82,$6D,$F2,$23,$C2,$D7,$93,$F8,$DE,$3F,$C0,$47,$F7
	.dc.b	$1E,$2F,$C0,$3F,$DC,$38,$BF,$00,$FF,$72,$E2,$F8,$0F,$4E,$2F,$A6
	.dc.b	$F4,$6B,$E4,$C3,$F8,$26,$5F,$EB,$5D,$9F,$0F,$F4,$EB,$EA,$B5,$FE
	.dc.b	$EB,$1A,$75,$A2,$31,$1F,$6E,$31,$3C,$5B,$1C,$27,$6F,$F1,$26,$BE
	.dc.b	$03,$53,$C6,$33,$1E,$28,$C1,$1D,$90,$F1,$47,$09,$F0,$C6,$27,$A7
	.dc.b	$C3,$FE,$2F,$D8,$FC,$67,$C3,$7E,$37,$E0,$7F,$1D,$EE,$FE,$3F,$E0
	.dc.b	$FF,$21,$F0,$3F,$91,$C7,$C1,$97,$33,$C0,$7A,$DC,$B4,$C9,$ED,$38
	.dc.b	$13,$5F,$A5,$1C,$23,$47,$13,$65,$21,$A2,$1F,$14,$EE,$53,$89,$AD
	.dc.b	$3E,$1B,$C8,$D1,$5C,$BF,$B6,$F8,$6F,$D8,$7C,$37,$EC,$7E,$0B,$F6
	.dc.b	$5E,$DF,$C6,$AF,$DB,$F6,$7E,$F3,$4F,$DB,$7C,$37,$C7,$9F,$1E,$5D
	.dc.b	$1E,$E7,$8F,$8F,$EB,$13,$4F,$85,$1A,$3D,$5A,$FC,$67,$9C,$ED,$B3
	.dc.b	$C2,$BB,$12,$75,$3F,$08,$E9,$19,$9E,$13,$A9,$F8,$4B,$8F,$F0,$D1
	.dc.b	$EE,$3F,$47,$C5,$F8,$77,$17,$FA,$3F,$83,$FE,$2D,$F0,$7F,$88,$9C
	.dc.b	$E1,$D5,$F8,$18,$F8,$08,$FA,$8F,$CD,$65,$EF,$7D,$18,$79,$DD,$18
	.dc.b	$AE,$8F,$B1,$B1,$B8,$D5,$76,$78,$70,$FD,$C7,$CB,$C1,$F7,$1F,$3F
	.dc.b	$1F,$E8,$5F,$4E,$DF,$6D,$E6,$78,$8F,$8F,$7D,$5A,$7F,$9C,$7D,$5F
	.dc.b	$E7,$FF,$0D,$FB,$B2,$47,$F0,$36,$BF,$B8,$6B,$FC,$15,$AF,$23,$81
	.dc.b	$0E,$05,$38,$14,$E0,$68,$87,$F1,$1F,$86,$FF,$38,$7E,$63,$FD,$3E
	.dc.b	$3E,$DA,$F8,$7C,$EF,$AF,$77,$7F,$0D,$FD,$4F,$1F,$2E,$1F,$C9,$9B
	.dc.b	$C1,$FC,$95,$BC,$26,$92,$7D,$58,$7F,$33,$8C,$47,$98,$F4,$9E,$C3
	.dc.b	$07,$73,$DC,$53,$35,$E1,$72,$D7,$35,$E3,$72,$D7,$39,$FC,$4F,$85
	.dc.b	$F8,$BE,$FF,$C6,$F8,$6F,$85,$5F,$1B,$4F,$84,$F2,$CB,$D2,$F9,$C8
	.dc.b	$7A,$88,$79,$CC,$9E,$9C,$FE,$13,$CC,$87,$A2,$3A,$BC,$B3,$F6,$3C
	.dc.b	$9A,$E7,$3F,$63,$96,$B9,$9F,$EA,$DB,$3E,$0B,$F0,$7D,$7F,$84,$E0
	.dc.b	$FC,$0D,$FE,$13,$E0,$7F,$02,$FE,$F3,$0F,$1B,$EB,$3A,$1E,$83,$C0
	.dc.b	$FD,$CC,$FD,$B1,$EB,$34,$70,$30,$43,$89,$A3,$B1,$33,$F7,$B7,$37
	.dc.b	$07,$D9,$8F,$CD,$EF,$F6,$5F,$B0,$41,$E6,$7E,$86,$F9,$92,$FE,$3C
	.dc.b	$F5,$C7,$DF,$39,$6C,$FE,$27,$17,$1F,$FC,$00,$DE,$E1,$FE,$08,$6B
	.dc.b	$FE,$07,$6B,$FE,$94,$6B,$C6,$3E,$A9,$1E,$4D,$79,$35,$E4,$D7,$91
	.dc.b	$C0,$87,$02,$9C,$0F,$37,$87,$FC,$13,$E1,$BF,$82,$FC,$17,$F0,$6F
	.dc.b	$86,$FE,$0F,$F0,$D2,$FF,$83,$E1,$DD,$6A,$CB,$D7,$18,$6C,$66,$78
	.dc.b	$36,$7C,$67,$A0,$F3,$10,$F3,$0C,$FE,$6C,$71,$8D,$1C,$4D,$90,$C1
	.dc.b	$0F,$24,$F8,$AB,$CA,$47,$04,$8F,$29,$C0,$B3,$EF,$CF,$F9,$5F,$C3
	.dc.b	$7C,$89,$FE,$57,$97,$C7,$7F,$1F,$C9,$F8,$5C,$FF,$98,$E3,$D6,$7D
	.dc.b	$27,$52,$E3,$ED,$7D,$46,$4A,$E5,$F9,$19,$FC,$4C,$FD,$51,$E6,$6B
	.dc.b	$D1,$AF,$26,$BF,$16,$7A,$3D,$CC,$46,$9C,$12,$3F,$01,$F0,$DF,$89
	.dc.b	$8F,$E9,$7B,$7F,$4E,$F7,$3F,$A7,$FC,$27,$F5,$0F,$86,$F8,$35,$E5
	.dc.b	$CF,$E1,$3E,$36,$BE,$07,$F4,$19,$FB,$5C,$C7,$87,$3F,$26,$7E,$82
	.dc.b	$38,$24,$F0,$8C,$B6,$1C,$1F,$6A,$BF,$B1,$F0,$FE,$85,$FE,$CB,$B7
	.dc.b	$DB,$E4,$FE,$11,$7D,$27,$D4,$43,$E7,$61,$D4,$3F,$E0,$53,$F4,$FF
	.dc.b	$99,$3D,$EF,$F4,$43,$63,$FD,$1E,$D7,$FD,$1C,$D7,$99,$CA,$38,$91
	.dc.b	$E2,$D7,$8B,$5E,$39,$A7,$A4,$79,$A3,$13,$B8,$C4,$E9,$F0,$1D,$88
	.dc.b	$76,$3D,$3F,$09,$E6,$C7,$F9,$74,$BE,$13,$1C,$9D,$1B,$7F,$13,$E1
	.dc.b	$2E,$5E,$5F,$80,$B8,$3E,$0A,$E9,$E5,$C3,$C6,$78,$48,$7B,$A7,$D1
	.dc.b	$7E,$C8,$F3,$99,$9E,$51,$98,$E5,$C5,$FC,$3A,$B8,$92,$B6,$D9,$E6
	.dc.b	$7C,$68,$C9,$1C,$9E,$63,$C0,$62,$BC,$0E,$0F,$49,$F4,$35,$F8,$27
	.dc.b	$C2,$F8,$2F,$87,$9F,$C4,$F8,$7F,$8B,$9F,$C4,$BE,$99,$79,$5E,$67
	.dc.b	$84,$F3,$1A,$CB,$CC,$FC,$0E,$0C,$BC,$33,$F2,$B8,$3C,$C4,$30,$43
	.dc.b	$B1,$1E,$2D,$AF,$23,$C4,$C5,$78,$9F,$B1,$F0,$FF,$80,$DE,$47,$FA
	.dc.b	$42,$9F,$0F,$F8,$4E,$0F,$C2,$FC,$3F,$C1,$AF,$45,$76,$7C,$99,$78
	.dc.b	$89,$19,$8F,$09,$F9,$AE,$1F,$C0,$67,$E3,$8E,$33,$BA,$E3,$39,$9E
	.dc.b	$27,$81,$C4,$6A,$3E,$9F,$0F,$DA,$7D,$5C,$3F,$69,$F6,$70,$7D,$A7
	.dc.b	$DD,$7C,$B0,$F0,$11,$C1,$FB,$DC,$7F,$2D,$EC,$FF,$0A,$25,$7F,$0A
	.dc.b	$6B,$FC,$29,$AF,$E9,$5A,$FE,$F9,$AF,$F1,$66,$E5,$FC,$1D,$B6,$69
	.dc.b	$8F,$77,$F6,$E7,$EE,$3E,$0B,$F7,$3F,$0D,$FB,$AF,$86,$FD,$DF,$BF
	.dc.b	$FB,$CB,$E3,$97,$9C,$87,$9C,$A7,$9C,$A7,$9C,$A7,$9C,$FE,$5D,$97
	.dc.b	$9D,$CB,$5E,$84,$3A,$14,$E8,$53,$A1,$5E,$4F,$02,$1C,$0A,$70,$2B
	.dc.b	$C1,$FE,$57,$8F,$C6,$7F,$39,$F0,$DF,$C7,$BE,$03,$F8,$FF,$C3,$7F
	.dc.b	$20,$F8,$2F,$8D,$5E,$AD,$BE,$83,$72,$F9,$2F,$96,$32,$D8,$FC,$0E
	.dc.b	$FF,$86,$8E,$04,$34,$53,$CF,$3C,$9B,$3D,$9A,$6C,$FE,$2C,$7F,$5E
	.dc.b	$ED,$FF,$36,$7C,$07,$F5,$68,$FF,$10,$9F,$D8,$38,$7F,$06,$FE,$EF
	.dc.b	$83,$F0,$6F,$EF,$6F,$B3,$F2,$C8,$7C,$B2,$EF,$F2,$DC,$11,$C1,$F7
	.dc.b	$F2,$F8,$11,$EE,$21,$EA,$29,$D3,$2F,$80,$DA,$F1,$BD,$63,$2D,$8F
	.dc.b	$EF,$3B,$FD,$B8,$F7,$DF,$F6,$DD,$FE,$E4,$7F,$19,$DB,$EE,$3E,$F3
	.dc.b	$E5,$1C,$CB,$3E,$AC,$3B,$B7,$0F,$B7,$D7,$FD,$DF,$57,$C3,$7F,$68
	.dc.b	$6E,$1F,$4F,$BF,$53,$92,$46,$4F,$F3,$CC,$F0,$7D,$2D,$E0,$FB,$73
	.dc.b	$C2,$72,$D8,$CC,$73,$C5,$5C,$6F,$C3,$39,$8C,$3E,$0B,$E5,$3A,$8E
	.dc.b	$A7,$F3,$FD,$7E,$66,$1F,$E0,$A7,$71,$C4,$E9,$58,$C3,$C2,$DC,$3B
	.dc.b	$C6,$27,$64,$3B,$92,$FD,$6D,$C3,$C9,$F1,$7F,$B0,$AE,$53,$93,$C4
	.dc.b	$57,$35,$C8,$95,$ED,$6C,$F0,$6C,$F0,$9D,$12,$74,$D8,$D4,$60,$8F
	.dc.b	$D0,$F8,$7F,$C1,$7C,$3F,$E0,$FE,$1F,$E1,$67,$2F,$E0,$63,$E0,$CF
	.dc.b	$E2,$92,$75,$3E,$B8,$C3,$F1,$F3,$F8,$50,$F6,$BF,$20,$49,$FB,$CD
	.dc.b	$7F,$16,$DB,$FE,$0A,$FE,$5F,$E0,$BF,$31,$F0,$FF,$99,$F8,$3F,$CD
	.dc.b	$7C,$3F,$E6,$FE,$03,$F3,$9F,$0B,$F9,$DC,$FF,$3D,$F0,$7F,$9F,$F8
	.dc.b	$7F,$D0,$7C,$27,$E8,$7E,$1F,$F0,$FC,$DF,$CB,$B8,$3F,$14,$7E,$47
	.dc.b	$E1,$7F,$4D,$F0,$FB,$8F,$CA,$5F,$37,$F2,$D8,$F7,$7F,$31,$9F,$8D
	.dc.b	$FC,$D6,$BE,$77,$F3,$9B,$FA,$DF,$82,$41,$C7,$F2,$39,$3E,$6F,$27
	.dc.b	$D1,$6E,$BF,$77,$83,$E7,$DF,$D1,$9C,$BF,$7A,$33,$5C,$63,$F5,$9C
	.dc.b	$FF,$80,$7F,$5D,$CD,$C3,$A3,$F0,$E4,$BE,$38,$F7,$C7,$51,$8B,$ED
	.dc.b	$1F,$BF,$CB,$D1,$1F,$CE,$CF,$CF,$9F,$E8,$A9,$FE,$10,$D9,$D3,$A3
	.dc.b	$F7,$86,$4E,$57,$FA,$A7,$07,$F4,$77,$0F,$02,$1C,$C8,$E9,$B3,$E2
	.dc.b	$77,$19,$3F,$D3,$DD,$3F,$84,$8F,$F5,$8B,$FC,$6D,$F0,$90,$E3,$3C
	.dc.b	$DE,$B5,$EC,$AE,$93,$96,$C6,$63,$04,$70,$7F,$AE,$3B,$DD,$7D,$A3
	.dc.b	$37,$F1,$2B,$83,$63,$2F,$C5,$8C,$B6,$32,$FA,$4F,$39,$C4,$86,$CA
	.dc.b	$6E,$FF,$84,$3F,$CD,$7A,$FF,$84,$3F,$C9,$4D,$57,$81,$B5,$E0,$7D
	.dc.b	$47,$42,$9D,$C8,$7D,$EC,$3E,$0B,$E6,$32,$70,$3A,$9C,$CD,$18,$36
	.dc.b	$60,$E0,$43,$EE,$9F,$C0,$0F,$EE,$7D,$AC,$BF,$01,$1B,$C7,$EC,$BD
	.dc.b	$DA,$F6,$6B,$E1,$3C,$55,$EA,$8C,$C7,$3C,$3E,$33,$70,$F8,$CF,$9C
	.dc.b	$93,$C1,$B3,$F8,$8D,$BF,$75,$1F,$DC,$B6,$F3,$BA,$BE,$77,$C4,$97
	.dc.b	$E0,$9D,$1F,$80,$C3,$CD,$1F,$BD,$9F,$E0,$55,$FC,$00,$95,$FD,$13
	.dc.b	$6F,$E0,$8D,$3F,$81,$96,$78,$BF,$37,$7F,$EE,$8F,$F1,$03,$57,$DC
	.dc.b	$97,$DA,$3F,$8E,$1A,$7F,$A7,$EF,$FD,$33,$E1,$FF,$C3,$44,$9F,$F0
	.dc.b	$43,$5F,$F0,$AB,$6F,$FC,$9E,$FF,$88,$7D,$EF,$F1,$17,$C0,$2B,$FC
	.dc.b	$29,$F9,$9B,$4C,$47,$F8,$DA,$1F,$07,$8C,$75,$2D,$54,$DB,$58,$75
	.dc.b	$9C,$FC,$2E,$2F,$85,$BE,$23,$15,$CB,$E2,$2C,$BE,$16,$FC,$46,$3E
	.dc.b	$23,$FC,$75,$CD,$F8,$2D,$BF,$CD,$7C,$DC,$3E,$23,$8F,$5E,$F6,$F6
	.dc.b	$E3,$D3,$2E,$BA,$76,$DB,$BE,$FE,$0E,$1F,$0F,$17,$8B,$8B,$E3,$E5
	.dc.b	$FE,$35,$E4,$F9,$91,$FE,$4D,$E5,$F8,$FC,$DF,$D0,$AB,$EC,$93,$87
	.dc.b	$FA,$93,$6B,$F1,$2D,$3F,$12,$5E,$5F,$C4,$BF,$B4,$E9,$FE,$69,$1F
	.dc.b	$CD,$A3,$F0,$84,$38,$74,$FE,$29,$C5,$F8,$0F,$0E,$1F,$89,$8F,$C3
	.dc.b	$90,$FC,$39,$63,$F0,$ED,$C3,$DC,$FC,$13,$FC,$AD,$D1,$F1,$5F,$C4
	.dc.b	$E9,$F2,$5F,$C5,$ED,$F3,$5F,$C6,$F0,$7D,$17,$E9,$9F,$90,$E7,$FB
	.dc.b	$6F,$F2,$D3,$F4,$C7,$E4,$79,$7E,$DB,$B3,$89,$F9,$5E,$3F,$B6,$FE
	.dc.b	$6B,$8F,$ED,$BF,$9C,$E3,$FB,$6F,$E7,$B4,$FB,$6F,$F9,$EB,$8F,$ED
	.dc.b	$BF,$E7,$EE,$3F,$B7,$E4,$FE,$67,$A5,$F0,$1F,$D7,$5B,$5F,$DD,$9A
	.dc.b	$FF,$75,$6B,$FA,$AE,$5F,$D2,$CF,$12,$47,$16,$BC,$5A,$F1,$9E,$6F
	.dc.b	$8E,$32,$D3,$91,$0E,$45,$39,$7C,$27,$EB,$3E,$07,$F5,$BF,$0D,$FA
	.dc.b	$EF,$82,$FD,$7F,$C3,$FE,$C3,$0E,$B8,$7E,$D9,$B9,$7A,$A3,$A1,$E1
	.dc.b	$3B,$90,$F4,$47,$0C,$BD,$31,$CC,$FE,$21,$BF,$F4,$58,$CD,$75,$9E
	.dc.b	$47,$53,$99,$E0,$39,$1A,$29,$A3,$0F,$37,$04,$3A,$10,$E2,$57,$83
	.dc.b	$F8,$1C,$FE,$43,$FC,$9B,$2F,$E2,$BF,$0F,$FC,$5B,$E1,$FE,$44,$F9
	.dc.b	$6F,$8B,$D0,$8E,$09,$87,$C0,$26,$5F,$1C,$FC,$1E,$5F,$15,$C5,$75
	.dc.b	$9D,$B6,$32,$FC,$08,$CB,$63,$2F,$C1,$3C,$70,$F8,$7F,$E5,$FF,$0F
	.dc.b	$FC,$CB,$F3,$47,$BB,$F8,$5B,$FE,$8B,$BF,$B9,$E8,$78,$48,$60,$87
	.dc.b	$7C,$FE,$E6,$7E,$B8,$D9,$9C,$BC,$ED,$CF,$CE,$FD,$38,$CB,$63,$2F
	.dc.b	$A9,$FD,$A6,$FF,$5A,$3F,$AC,$6F,$F5,$9F,$EB,$BB,$7A,$5E,$C7,$88
	.dc.b	$E8,$53,$A1,$EC,$C3,$FC,$35,$A3,$E0,$3E,$31,$27,$FD,$F9,$19,$6C
	.dc.b	$65,$77,$FF,$68,$D1,$5B,$9C,$F3,$FF,$67,$8C,$56,$D7,$C2,$7D,$5F
	.dc.b	$87,$FA,$DF,$09,$F5,$FE,$1E,$63,$E2,$F8,$4F,$B5,$F0,$9F,$6F,$E1
	.dc.b	$FE,$E7,$C3,$CA,$F3,$FC,$3F,$F6,$EE,$4F,$C0,$77,$7E,$07,$CB,$FC
	.dc.b	$17,$C1,$FE,$0F,$E1,$FF,$09,$7C,$70,$F3,$9F,$E0,$1E,$16,$5D,$E3
	.dc.b	$E1,$E9,$F8,$BD,$7C,$B1,$C3,$93,$E0,$EF,$EA,$8F,$93,$C3,$F9,$1E
	.dc.b	$2F,$91,$C7,$F0,$79,$7F,$99,$D7,$7F,$84,$FC,$B1,$1F,$A2,$D7,$B6
	.dc.b	$A8,$C5,$7C,$B7,$B4,$F8,$0F,$B3,$CB,$FA,$1F,$87,$FD,$17,$C2,$FE
	.dc.b	$8F,$E1,$FD,$95,$D3,$A3,$F4,$DF,$0D,$FA,$7F,$87,$F8,$B3,$FE,$2B
	.dc.b	$CF,$EF,$7E,$6B,$FE,$2A,$8F,$1F,$57,$F8,$7A,$1E,$CD,$F6,$71,$ED
	.dc.b	$F2,$BF,$4E,$3D,$9F,$C6,$E7,$E2,$CB,$F0,$9A,$79,$A7,$F2,$3B,$7C
	.dc.b	$B7,$E0,$F2,$7B,$9F,$87,$C3,$F1,$7D,$66,$FF,$8F,$E4,$F9,$FE,$05
	.dc.b	$BF,$CC,$90,$F9,$71,$A7,$F0,$0F,$AF,$DA,$F6,$7C,$47,$DE,$D7,$F3
	.dc.b	$1D,$37,$0F,$CE,$75,$FE,$53,$1F,$8D,$1F,$E4,$5E,$57,$C4,$7E,$37
	.dc.b	$0F,$D0,$77,$7E,$46,$FF,$CA,$BE,$1F,$C9,$F7,$BF,$2D,$1F,$B0,$F0
	.dc.b	$F4,$FE,$03,$C3,$E5,$9F,$EA,$F0,$FA,$19,$78,$6B,$F0,$BA,$79,$5F
	.dc.b	$CD,$6B,$7C,$16,$CD,$FD,$D5,$F0,$F7,$FC,$A3,$FE,$69,$E3,$F9,$6F
	.dc.b	$E5,$4F,$C9,$90,$FA,$25,$8F,$C9,$37,$4F,$E0,$1C,$9F,$5E,$7F,$80
	.dc.b	$72,$FD,$B7,$F8,$17,$37,$DF,$BF,$CE,$F3,$FE,$0F,$0F,$E1,$1D,$1F
	.dc.b	$87,$9F,$C4,$9F,$BE,$3E,$79,$FE,$77,$D3,$F2,$98,$7F,$0D,$E8,$FC
	.dc.b	$A7,$67,$E4,$EB,$7B,$FC,$2E,$BF,$CA,$5F,$F1,$3E,$BE,$CF,$F1,$5C
	.dc.b	$BC,$57,$FC,$0F,$4F,$34,$FA,$23,$F5,$A7,$EB,$CF,$DA,$F8,$3F,$B4
	.dc.b	$D7,$E0,$C7,$CB,$D3,$F2,$4F,$E4,$0F,$A1,$18,$AF,$A3,$19,$23,$83
	.dc.b	$F9,$45,$7F,$27,$8F,$09,$FC,$13,$8F,$EF,$78,$2E,$5F,$C2,$5F,$AE
	.dc.b	$BF,$9D,$F3,$FE,$29,$FE,$3D,$E1,$FF,$27,$7F,$9F,$E7,$F9,$27,$F8
	.dc.b	$7F,$57,$E4,$AF,$F9,$C5,$71,$E1,$FE,$7F,$87,$C6,$7F,$A0,$F8,$7F
	.dc.b	$C9,$70,$FE,$93,$A7,$93,$C3,$F4,$47,$C6,$3E,$09,$FB,$03,$E3,$1F
	.dc.b	$8F,$3F,$1A,$4C,$7F,$1A,$DD,$FE,$6B,$4E,$11,$F1,$9F,$A8,$FF,$4F
	.dc.b	$9F,$E4,$51,$FD,$73,$C3,$FE,$44,$FF,$12,$21,$FC,$BB,$C3,$FE,$44
	.dc.b	$FE,$3C,$FE,$7D,$CF,$F8,$EC,$3F,$AE,$6F,$F8,$E9,$FC,$6C,$BC,$3F
	.dc.b	$E4,$31,$C7,$2F,$C7,$3F,$DB,$30,$F9,$1C,$4E,$5F,$1C,$FF,$3E,$CF
	.dc.b	$CE,$FF,$40,$D3,$E4,$E1,$ED,$8F,$81,$BF,$F7,$0E,$BF,$E9,$6F,$F7
	.dc.b	$2F,$0F,$F8,$E3,$67,$E7,$B5,$FC,$6E,$FF,$91,$23,$FA,$BF,$0F,$F9
	.dc.b	$0B,$F6,$88,$7E,$3C,$FE,$E8,$7F,$64,$3E,$0F,$81,$F8,$5D,$7F,$B4
	.dc.b	$C7,$F8,$2B,$C1,$FC,$73,$F2,$BC,$2F,$C7,$69,$FE,$03,$E8,$FC,$76
	.dc.b	$5F,$E0,$7B,$E5,$CF,$FC,$7B,$0E,$C7,$F1,$E3,$C7,$7F,$CD,$1B,$5F
	.dc.b	$CC,$7C,$2F,$E3,$B1,$EE,$3F,$1C,$3C,$2F,$E3,$BB,$7C,$98,$FE,$30
	.dc.b	$7E,$2A,$BF,$1B,$7F,$90,$23,$82,$47,$F8,$1F,$B3,$EA,$C7,$F8,$7F
	.dc.b	$8B,$EC,$D7,$F8,$7F,$8F,$EE,$BF,$B3,$BF,$EE,$07,$E0,$CF,$F1,$7F
	.dc.b	$77,$F1,$4B,$FE,$95,$1F,$E1,$A3,$F1,$BE,$07,$F1,$B7,$F2,$7E,$06
	.dc.b	$DF,$F0,$C5,$7F,$1A,$AF,$E5,$31,$F9,$CB,$FE,$DA,$FE,$84,$FE,$DA
	.dc.b	$7F,$0A,$7F,$9A,$BF,$E4,$0C,$BF,$8D,$3F,$B3,$D3,$E3,$F5,$3E,$23
	.dc.b	$FC,$05,$3F,$D7,$DA,$FF,$33,$6E,$DF,$D4,$1B,$3C,$67,$8C,$E5,$B1
	.dc.b	$97,$FC,$1D,$19,$6C,$65,$FF,$69,$35,$F1,$1C,$A7,$B9,$1C,$10,$E6
	.dc.b	$E5,$A7,$8B,$E0,$3F,$C3,$9F,$11,$FE,$1D,$F8,$2F,$F0,$F7,$BF,$FE
	.dc.b	$1F,$F8,$3F,$F1,$06,$1C,$70,$F1,$B5,$F1,$9D,$08,$68,$F1,$14,$C9
	.dc.b	$C0,$C7,$96,$F8,$2C,$44,$3E,$71,$C8,$C6,$9F,$1D,$E8,$43,$3A,$7C
	.dc.b	$77,$89,$E0,$29,$B2,$9C,$2F,$D8,$E0,$FE,$E1,$F1,$1F,$19,$FE,$43
	.dc.b	$9F,$C6,$7F,$91,$E5,$F1,$9F,$31,$C7,$0F,$33,$D4,$CB,$87,$58,$79
	.dc.b	$E3,$04,$7F,$0D,$97,$C5,$9D,$E3,$E6,$AD,$36,$B9,$11,$C1,$27,$C3
	.dc.b	$19,$6C,$71,$D7,$FC,$BD,$F0,$BF,$E5,$FF,$88,$FF,$30,$7C,$47,$F9
	.dc.b	$87,$E2,$3D,$D8,$7C,$BC,$BC,$2F,$22,$19,$CB,$B3,$72,$33,$3F,$65
	.dc.b	$72,$8D,$1C,$8F,$AC,$E5,$B5,$D1,$B5,$ED,$78,$9E,$1D,$BE,$C3,$FD
	.dc.b	$73,$6F,$B0,$FF,$5E,$DB,$D6,$F2,$3C,$07,$D7,$39,$99,$3A,$E3,$E8
	.dc.b	$C1,$EE,$FF,$00,$24,$FE,$F9,$AF,$EF,$1A,$FE,$F1,$AF,$F9,$AD,$AF
	.dc.b	$F8,$FD,$B9,$7F,$97,$1B,$3E,$09,$EC,$43,$B1,$FC,$3B,$DE,$FF,$4C
	.dc.b	$7C,$37,$D0,$8F,$F5,$D7,$A8,$E9,$C7,$4F,$FB,$76,$1F,$0D,$89,$FA
	.dc.b	$F9,$A5,$A7,$FA,$8F,$4F,$19,$31,$F1,$1E,$D2,$9E,$D2,$9E,$D2,$D7
	.dc.b	$B5,$FC,$27,$36,$DF,$59,$0E,$85,$34,$60,$86,$08,$77,$38,$14,$D9
	.dc.b	$4D,$C7,$8D,$DB,$5F,$E6,$7A,$7C,$C7,$FC,$DF,$87,$CB,$6E,$3F,$25
	.dc.b	$57,$C9,$C7,$E4,$47,$C0,$D7,$ED,$3D,$9A,$F9,$27,$C0,$F9,$23,$2D
	.dc.b	$CB,$E8,$3F,$53,$2C,$3D,$63,$2F,$39,$E2,$FB,$C9,$5D,$9B,$1A,$8C
	.dc.b	$11,$C1,$27,$F2,$7C,$1E,$F9,$FC,$37,$C3,$5C,$DF,$05,$76,$C7,$C3
	.dc.b	$7C,$2C,$FE,$06,$5E,$67,$D2,$60,$F5,$10,$F7,$1E,$67,$83,$4D,$1E
	.dc.b	$F7,$38,$FA,$5B,$9F,$A6,$7E,$96,$FF,$6D,$FD,$2E,$FE,$93,$D5,$BF
	.dc.b	$9E,$3A,$1D,$8F,$39,$83,$D0,$47,$F8,$EE,$8F,$84,$F8,$84,$9F,$C4
	.dc.b	$B5,$FD,$AB,$78,$7F,$68,$DB,$FD,$9D,$73,$AE,$24,$38,$96,$B8,$B6
	.dc.b	$78,$D7,$38,$FF,$66,$61,$F6,$71,$34,$91,$FE,$80,$6B,$D8,$97,$D9
	.dc.b	$B7,$D9,$F0,$1F,$A2,$C3,$D0,$E8,$C6,$1E,$3C,$FF,$87,$36,$77,$DB
	.dc.b	$FE,$5A,$78,$E1,$DC,$98,$77,$C3,$A6,$1C,$9F,$0B,$5C,$D7,$29,$E1
	.dc.b	$5C,$A7,$C6,$70,$9D,$38,$38,$E1,$F9,$DF,$87,$FC,$F7,$C3,$FE,$7F
	.dc.b	$E1,$FF,$41,$CB,$FA,$1F,$7F,$58,$73,$C3,$D8,$47,$04,$3A,$9F,$6F
	.dc.b	$5F,$81,$EE,$FE,$AC,$91,$FC,$41,$AF,$F3,$C6,$E9,$FC,$EF,$E1,$FF
	.dc.b	$65,$F0,$1F,$B3,$F8,$7F,$DA,$7C,$1F,$ED,$7E,$1F,$F6,$DF,$07,$FB
	.dc.b	$7F,$87,$FD,$C7,$C3,$FE,$EF,$F3,$67,$C4,$5F,$6F,$1E,$77,$E0,$B6
	.dc.b	$FE,$41,$D4,$EC,$FC,$4F,$17,$E1,$C8,$7F,$B3,$3B,$38,$76,$7F,$4C
	.dc.b	$CD,$8F,$5A,$FF,$02,$E3,$D2,$7C,$31,$FC,$BB,$4F,$33,$FC,$C3,$6F
	.dc.b	$53,$FC,$CB,$83,$DC,$FF,$34,$E2,$FE,$25,$C7,$F1,$33,$FD,$4F,$27
	.dc.b	$FB,$2B,$B3,$EE,$F0,$7E,$17,$5F,$C1,$36,$BF,$0A,$4C,$7F,$05,$3A
	.dc.b	$AE,$24,$E7,$FB,$93,$F6,$89,$3B,$EC,$FC,$3C,$7F,$48,$EB,$FC,$3B
	.dc.b	$FD,$2B,$AB,$F0,$EF,$0C,$7C,$19,$78,$E3,$C3,$18,$CB,$C1,$1F,$89
	.dc.b	$C7,$C9,$97,$F0,$BE,$AF,$D4,$3F,$E9,$6A,$FD,$8D,$7F,$80,$F0,$FD
	.dc.b	$23,$63,$2F,$EA,$CC,$4F,$12,$3F,$E9,$C8,$DB,$FA,$53,$66,$63,$6F
	.dc.b	$FA,$7A,$37,$3E,$08,$EC,$43,$91,$FE,$10,$E5,$FC,$5B,$F8,$93,$F1
	.dc.b	$78,$7E,$42,$32,$DE,$5F,$C6,$CF,$22,$57,$76,$DF,$88,$98,$79,$A3
	.dc.b	$D2,$79,$DC,$3E,$59,$FE,$15,$DB,$FC,$F2,$73,$3D,$1B,$3D,$C9,$5D
	.dc.b	$A7,$E2,$10,$EC,$53,$C2,$43,$C0,$F1,$7E,$56,$DF,$8A,$7F,$C7,$7B
	.dc.b	$7E,$29,$FF,$1E,$ED,$F8,$AC,$FF,$9A,$47,$33,$CC,$77,$3A,$E3,$E8
	.dc.b	$C3,$B4,$6F,$1F,$4C,$FF,$92,$72,$F8,$13,$C4,$E6,$7A,$88,$71,$AE
	.dc.b	$6F,$63,$85,$74,$6C,$69,$FA,$FA,$7D,$CD,$BF,$05,$AF,$DC,$BF,$E1
	.dc.b	$CD,$9D,$57,$88,$F0,$63,$EB,$23,$D9,$AF,$9F,$1F,$5B,$E8,$3C,$C5
	.dc.b	$3A,$18,$3A,$9C,$C8,$6E,$FD,$8F,$D6,$3F,$C8,$BD,$BF,$D1,$5F,$F2
	.dc.b	$3F,$6F,$F4,$57,$FC,$93,$BF,$F8,$05,$EA,$73,$D3,$E2,$5F,$48,$C9
	.dc.b	$0F,$EC,$18,$7A,$B5,$FE,$03,$1F,$C0,$F5,$FE,$0D,$9F,$F1,$49,$CC
	.dc.b	$FF,$0D,$3F,$85,$92,$FF,$87,$13,$0F,$E3,$B8,$FC,$FD,$65,$7A,$0E
	.dc.b	$C7,$19,$FE,$26,$F5,$AE,$58,$78,$71,$F1,$CF,$FB,$0B,$81,$F0,$1F
	.dc.b	$D0,$FE,$1F,$FA,$81,$27,$FA,$1D,$65,$A7,$F9,$B3,$A5,$3B,$86,$BF
	.dc.b	$D2,$63,$FA,$D7,$AB,$FD,$6F,$E0,$BF,$AE,$7C,$2F,$F5,$DF,$84,$FE
	.dc.b	$BD,$F0,$5F,$D7,$FD,$BF,$EC,$1F,$0B,$FD,$87,$6F,$EC,$5C,$5F,$D8
	.dc.b	$FD,$EF,$EC,$9E,$F7,$EB,$78,$7F,$C6,$7D,$3F,$77,$87,$FD,$5B,$D3
	.dc.b	$F7,$5F,$F5,$7F,$47,$07,$D3,$CE,$E5,$65,$DA,$73,$1F,$11,$BA,$F9
	.dc.b	$5C,$EF,$E9,$77,$C3,$ED,$78,$F1,$FC,$3E,$5F,$85,$1F,$3F,$87,$E8
	.dc.b	$BF,$57,$87,$EB,$CF,$E1,$09,$C1,$A6,$CF,$DD,$E7,$FB,$4F,$F9,$20
	.dc.b	$FE,$67,$3F,$B6,$69,$93,$F9,$C1,$DE,$38,$ED,$F8,$48,$FC,$4B,$5F
	.dc.b	$ED,$1C,$FF,$06,$3E,$37,$3F,$AE,$FF,$05,$5F,$79,$B8,$7B,$A3,$07
	.dc.b	$E1,$67,$83,$F8,$7C,$FE,$33,$AB,$FA,$5A,$7C,$B8,$C6,$ED,$BE,$84
	.dc.b	$63,$7F,$C2,$3F,$5B,$6F,$C2,$3F,$63,$93,$F0,$8F,$F9,$7F,$93,$F0
	.dc.b	$9A,$FF,$BB,$79,$3F,$08,$FF,$BB,$F9,$3F,$09,$5E,$0E,$4F,$C1,$3E
	.dc.b	$2E,$4F,$C9,$BF,$84,$E5,$FC,$BB,$F8,$6E,$6F,$CD,$BF,$88,$E7,$FC
	.dc.b	$FB,$F8,$AE,$4F,$D2,$73,$7C,$BE,$27,$C1,$AA,$1F,$E0,$38,$CB,$83
	.dc.b	$FC,$6D,$15,$A5,$E5,$FC,$D4,$66,$3F,$68,$D8,$CC,$6E,$7A,$BD,$63
	.dc.b	$2D,$8C,$BE,$39,$E5,$18,$8F,$09,$B8,$C9,$1E,$CD,$74,$78,$7E,$1F
	.dc.b	$FC,$B9,$EF,$FF,$97,$7E,$1F,$FC,$BD,$F0,$DF,$E5,$FB,$E1,$7E,$13
	.dc.b	$99,$0E,$64,$34,$53,$CA,$60,$86,$0F,$4F,$73,$93,$11,$F2,$0E,$C7
	.dc.b	$03,$07,$52,$99,$CF,$E4,$B7,$3F,$8C,$E8,$EB,$1A,$6C,$6A,$31,$1B
	.dc.b	$23,$FC,$57,$DF,$FF,$7E,$7E,$6F,$F8,$3E,$6F,$F3,$A6,$98,$F8,$2F
	.dc.b	$91,$87,$A0,$EB,$A7,$CA,$69,$93,$AE,$5E,$12,$46,$E3,$DF,$A7,$C5
	.dc.b	$77,$19,$CF,$CE,$4A,$E3,$3D,$63,$13,$B9,$CC,$FC,$99,$CB,$B8,$CB
	.dc.b	$FA,$1D,$FF,$0A,$FF,$36,$CD,$F1,$1F,$86,$8F,$E6,$B8,$75,$23,$CD
	.dc.b	$AF,$7B,$EA,$FC,$21,$8F,$B7,$1F,$1C,$6E,$B8,$BF,$6D,$AF,$13,$07
	.dc.b	$72,$1D,$F5,$FA,$EF,$E7,$4F,$56,$7F,$59,$EC,$70,$39,$98,$21,$83
	.dc.b	$F3,$B7,$E7,$C9,$F0,$7F,$BD,$6C,$FF,$03,$6B,$FB,$D6,$BF,$E8,$F6
	.dc.b	$BF,$C4,$71,$46,$21,$18,$9E,$C4,$8E,$CD,$7B,$35,$EC,$D7,$B4,$F5
	.dc.b	$9E,$91,$C8,$87,$9F,$E2,$3F,$3F,$F0,$BF,$A0,$F8,$8F,$D0,$FC,$3F
	.dc.b	$E8,$B1,$E9,$7E,$6C,$7A,$12,$33,$8F,$69,$F2,$9D,$8F,$29,$4F,$2C
	.dc.b	$70,$9F,$93,$97,$CF,$8C,$12,$78,$1E,$17,$0F,$62,$1D,$8A,$76,$38
	.dc.b	$D7,$95,$D1,$0D,$1F,$0C,$FD,$17,$C0,$7F,$02,$F8,$5F,$E0,$7F,$0F
	.dc.b	$F0,$E7,$9E,$3F,$05,$B8,$79,$1E,$87,$9C,$C9,$0F,$D1,$F3,$78,$1D
	.dc.b	$E3,$E5,$70,$43,$91,$EC,$29,$C2,$BC,$4E,$63,$93,$A3,$93,$82,$47
	.dc.b	$08,$FC,$B7,$C1,$7E,$09,$FE,$05,$AF,$E0,$9F,$E0,$98,$7B,$5E,$75
	.dc.b	$C5,$B1,$97,$C0,$4B,$EF,$1F,$EA,$BC,$37,$3F,$0C,$F5,$47,$19,$DC
	.dc.b	$64,$D9,$ED,$AF,$03,$D5,$FC,$9E,$FF,$CE,$34,$FB,$13,$FC,$6F,$3F
	.dc.b	$B0,$F1,$C3,$B5,$F2,$C3,$BC,$E4,$F0,$1F,$43,$5F,$E0,$5F,$0F,$FC
	.dc.b	$04,$91,$FC,$05,$AF,$F0,$16,$BF,$C0,$5A,$FE,$D1,$AF,$EA,$DB,$3F
	.dc.b	$C1,$1B,$46,$98,$F8,$3F,$EA,$BF,$05,$FD,$5B,$E1,$FF,$AB,$FC,$3F
	.dc.b	$F5,$8A,$DE,$3E,$52,$1E,$52,$9E,$52,$9E,$52,$9E,$52,$E3,$E5,$7F
	.dc.b	$C5,$59,$7C,$47,$39,$79,$DB,$5E,$27,$C0,$43,$C0,$53,$C0,$53,$C0
	.dc.b	$5A,$F0,$36,$7B,$37,$E1,$FE,$1B,$FD,$67,$E1,$FE,$1B,$FD,$6E,$FC
	.dc.b	$4F,$80,$87,$80,$A7,$80,$B7,$E0,$6D,$F8,$1B,$3F,$9B,$78,$1C,$C8
	.dc.b	$60,$93,$CD,$CB,$5C,$CF,$57,$2D,$73,$3E,$A7,$2F,$E0,$DF,$EF,$1F
	.dc.b	$05,$F8,$37,$FB,$D7,$A7,$F8,$37,$E9,$90,$F9,$53,$CD,$EA,$64,$FA
	.dc.b	$B9,$7D,$D7,$66,$4F,$09,$4E,$26,$88,$7C,$63,$58,$7F,$60,$CB,$EB
	.dc.b	$C7,$FA,$4F,$16,$5F,$62,$31,$87,$96,$FE,$31,$2B,$E2,$36,$35,$18
	.dc.b	$23,$FB,$4C,$3F,$BD,$7C,$3F,$F8,$F0,$97,$FE,$1F,$6B,$E3,$6F,$82
	.dc.b	$F4,$FF,$B5,$C6,$1B,$3C,$48,$71,$2C,$F1,$7F,$B4,$4F,$F9,$33,$39
	.dc.b	$1F,$E2,$F6,$BF,$0D,$B3,$C7,$2F,$91,$3B,$21,$B2,$C6,$F2,$5F,$9E
	.dc.b	$7C,$27,$93,$0F,$4B,$E7,$25,$71,$9E,$17,$FE,$16,$7E,$2F,$8B,$FE
	.dc.b	$D5,$C3,$AD,$F3,$3A,$9C,$B0,$EB,$87,$38,$D6,$1C,$DB,$5C,$9E,$07
	.dc.b	$29,$E0,$E2,$78,$3F,$ED,$9F,$CE,$3E,$63,$7E,$1B,$FD,$77,$5F,$1E
	.dc.b	$7F,$C9,$F3,$FE,$BF,$8C,$7C,$46,$B1,$FF,$19,$53,$E2,$31,$7F,$E3
	.dc.b	$28,$7C,$46,$32,$43,$D9,$C2,$C6,$F0,$63,$4C,$E7,$F2,$B5,$F7,$6B
	.dc.b	$EC,$24,$7C,$62,$69,$F0,$5B,$9F,$C0,$9D,$CF,$CE,$D9,$E8,$7A,$BD
	.dc.b	$2F,$19,$27,$E0,$61,$96,$9E,$D8,$D3,$F2,$FD,$3F,$5F,$B9,$EC,$F0
	.dc.b	$7D,$BE,$E7,$BB,$DD,$F7,$FB,$DF,$03,$DF,$F8,$3F,$11,$F0,$BE,$07
	.dc.b	$F2,$FA,$FF,$94,$3A,$EE,$BF,$E4,$76,$F1,$7E,$6A,$3F,$30,$4E,$2F
	.dc.b	$CC,$4E,$B8,$CF,$59,$F4,$D7,$D2,$C7,$C5,$18,$E7,$69,$E7,$8C,$6D
	.dc.b	$EB,$75,$C1,$F6,$38,$7E,$0E,$9F,$87,$E2,$54,$E4,$FC,$F7,$5F,$D9
	.dc.b	$24,$7E,$61,$AF,$E5,$77,$FB,$BC,$1F,$61,$B1,$F6,$C9,$BF,$E2,$A7
	.dc.b	$F0,$D1,$E9,$E7,$FC,$5F,$47,$E3,$3A,$FF,$1B,$D7,$F8,$EE,$BC,$DE
	.dc.b	$CF,$81,$D7,$F8,$E7,$C8,$43,$C9,$A7,$9B,$7F,$C4,$E9,$F8,$BD,$3F
	.dc.b	$80,$ED,$EE,$BE,$15,$AD,$BD,$37,$B9,$F4,$C6,$09,$87,$5C,$3E,$03
	.dc.b	$FC,$23,$93,$F6,$B8,$7D,$CB,$FE,$2B,$5F,$85,$69,$F8,$52,$E1,$F8
	.dc.b	$59,$E1,$1F,$85,$9F,$DB,$9D,$DF,$1D,$DA,$F3,$13,$0F,$2C,$FF,$34
	.dc.b	$C7,$F4,$EF,$EF,$7E,$23,$F7,$24,$7F,$14,$D7,$F8,$73,$FB,$EF,$83
	.dc.b	$FD,$FF,$C4,$7F,$00,$F8,$5F,$E0,$9D,$AF,$88,$BF,$11,$8F,$88,$CF
	.dc.b	$BB,$AF,$87,$DF,$BF,$F5,$F5,$FE,$30,$7D,$B3,$F8,$E7,$4F,$E0,$5F
	.dc.b	$CE,$74,$7E,$05,$FC,$F4,$7F,$0E,$7F,$02,$7E,$87,$93,$AB,$FA,$6D
	.dc.b	$18,$78,$A3,$19,$79,$9D,$69,$EB,$7F,$55,$B7,$BD,$F8,$64,$1C,$3F
	.dc.b	$87,$7F,$1A,$7E,$23,$1F,$C0,$B4,$FA,$A5,$E3,$FA,$9C,$9F,$BD,$F0
	.dc.b	$64,$FE,$03,$97,$F7,$ED,$8C,$BF,$81,$39,$9E,$02,$73,$7E,$5E,$BF
	.dc.b	$05,$5F,$D2,$F1,$FB,$F8,$7F,$38,$E6,$FB,$97,$FD,$15,$B9,$7F,$42
	.dc.b	$6E,$3F,$D0,$79,$BE,$FB,$FC,$37,$9B,$EF,$ED,$F2,$61,$87,$F7,$5E
	.dc.b	$17,$07,$CF,$9C,$E5,$FE,$FF,$9F,$A7,$1F,$57,$93,$F0,$57,$FD,$5F
	.dc.b	$A3,$EE,$4F,$E0,$8F,$E3,$9D,$5F,$C0,$9F,$E3,$DD,$5F,$C0,$9F,$EC
	.dc.b	$5D,$7F,$C2,$9F,$EC,$99,$7E,$09,$F0,$8E,$06,$7E,$28,$CC,$63,$A7
	.dc.b	$F3,$11,$F8,$9E,$AF,$7B,$F0,$4F,$C2,$9F,$CB,$F8,$7E,$3D,$7C,$D8
	.dc.b	$FE,$51,$E2,$FB,$FE,$1F,$F9,$5C,$43,$C5,$8A,$31,$1B,$2D,$79,$A3
	.dc.b	$31,$C1,$B1,$98,$DC,$F3,$6C,$EA,$31,$1E,$19,$E0,$D8,$CB,$E6,$8C
	.dc.b	$B6,$32,$F9,$4E,$13,$D4,$8F,$80,$87,$5F,$81,$FE,$0B,$F0,$FF,$C1
	.dc.b	$BE,$0F,$F8,$3D,$70,$AE,$67,$03,$C0,$E5,$A6,$4F,$11,$83,$89,$0C
	.dc.b	$10,$F4,$6C,$F2,$78,$C4,$3E,$61,$D4,$A7,$33,$59,$7C,$46,$E5,$F1
	.dc.b	$1D,$9D,$CA,$6A,$FD,$4D,$3B,$4F,$E4,$B2,$FE,$53,$F0,$FF,$16,$7F
	.dc.b	$8C,$57,$07,$99,$2B,$C4,$DB,$F1,$46,$08,$E0,$93,$CE,$32,$FF,$AD
	.dc.b	$DB,$3E,$49,$DB,$67,$84,$62,$3E,$21,$E1,$8C,$91,$E0,$7E,$6B,$C2
	.dc.b	$FF,$26,$3F,$CC,$F4,$FC,$1B,$EE,$36,$74,$BF,$1B,$6F,$C0,$FC,$D8
	.dc.b	$CB,$63,$C9,$9F,$DD,$8F,$21,$D0,$D6,$1E,$57,$A3,$5C,$9B,$3E,$B1
	.dc.b	$E1,$CB,$EA,$BF,$D2,$F4,$FC,$CB,$B3,$B9,$D0,$C6,$1E,$67,$99,$D8
	.dc.b	$5F,$9E,$3F,$89,$EB,$FB,$72,$4F,$EF,$1A,$FE,$DD,$AF,$ED,$5A,$FF
	.dc.b	$2D,$6B,$FC,$39,$B1,$DA,$3A,$10,$E8,$53,$A1,$FC,$33,$E1,$FF,$BE
	.dc.b	$FC,$34,$8F,$ED,$95,$D3,$3F,$F0,$5C,$72,$21,$C8,$A7,$94,$EC,$53
	.dc.b	$27,$9C,$C9,$4E,$98,$7C,$78,$EF,$1D,$48,$73,$29,$C4,$B8,$7C,$17
	.dc.b	$BB,$5C,$C7,$47,$91,$0E,$45,$8E,$0D,$8E,$0F,$F3,$CC,$3F,$C4,$67
	.dc.b	$F8,$93,$E1,$FF,$C4,$B8,$7C,$72,$4F,$BA,$F9,$BD,$88,$E0,$97,$D9
	.dc.b	$B3,$F8,$4C,$3E,$23,$D2,$BD,$B5,$B2,$46,$63,$DE,$D7,$D4,$D7,$C3
	.dc.b	$1C,$7E,$07,$F0,$11,$FD,$9A,$19,$7B,$23,$D5,$1D,$9B,$5C,$F0,$F3
	.dc.b	$C6,$08,$7D,$CE,$1F,$0B,$83,$CA,$53,$27,$B2,$7A,$57,$78,$C1,$E2
	.dc.b	$CF,$E8,$BF,$E4,$1C,$FC,$CE,$88,$76,$29,$D8,$C1,$9C,$FF,$21,$F0
	.dc.b	$9F,$DA,$C9,$1F,$DC,$1A,$FF,$6B,$6B,$FD,$A5,$AF,$F8,$85,$AF,$F8
	.dc.b	$09,$B9,$7F,$85,$1B,$81,$A6,$3E,$1F,$FC,$DE,$7F,$9C,$3E,$1B,$FC
	.dc.b	$E3,$1A,$BF,$21,$0F,$21,$4F,$21,$4F,$21,$4F,$21,$6F,$C8,$F3,$C3
	.dc.b	$E1,$B9,$C3,$CC,$DA,$F1,$3E,$02,$1E,$02,$9E,$02,$9E,$02,$D7,$81
	.dc.b	$B3,$D9,$B8,$7C,$27,$FC,$FD,$F0,$FF,$09,$F1,$10,$E3,$7E,$47,$C4
	.dc.b	$43,$C4,$53,$C4,$79,$F0,$F7,$BC,$8E,$86,$F0,$F3,$37,$0F,$33,$6B
	.dc.b	$C0,$F6,$21,$D8,$B3,$C9,$B2,$F8,$5F,$F2,$4F,$73,$AF,$FD,$EB,$6F
	.dc.b	$CE,$7F,$C9,$93,$FE,$71,$D7,$E1,$63,$E7,$21,$E3,$3D,$27,$6C,$BC
	.dc.b	$AF,$8C,$FA,$59,$FC,$07,$81,$C8,$E0,$72,$38,$61,$E9,$6E,$7F,$09
	.dc.b	$D9,$C4,$CD,$7F,$95,$75,$FB,$51,$FC,$37,$5F,$4B,$E0,$21,$C8,$A7
	.dc.b	$23,$C5,$7E,$13,$F1,$18,$7F,$6D,$CD,$F0,$DF,$E6,$86,$D7,$F9,$B9
	.dc.b	$AF,$F9,$95,$AF,$F9,$8D,$AF,$D0,$6B,$FE,$8B,$6D,$7D,$39,$EC,$49
	.dc.b	$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$FA,$7E,$23,$F0,$7B,$78,$DB,$3E
	.dc.b	$3F,$7E,$DF,$8E,$3C,$24,$3C,$25,$3C,$25,$3C,$25,$3C,$25,$7A,$3C
	.dc.b	$88,$72,$3F,$C3,$BC,$9C,$9C,$94,$D1,$0C,$10,$EE,$70,$29,$B2,$9B
	.dc.b	$8F,$1B,$96,$B9,$C7,$EA,$92,$FE,$2A,$6D,$E3,$1F,$A1,$7C,$AF,$B9
	.dc.b	$1C,$12,$FA,$CF,$06,$CF,$C7,$6B,$F1,$0E,$4E,$1F,$43,$5F,$44,$73
	.dc.b	$23,$CD,$AF,$C5,$9F,$14,$7E,$AE,$07,$FB,$72,$5F,$05,$F0,$3E,$1F
	.dc.b	$DD,$3C,$C8,$73,$9F,$13,$E6,$AE,$EF,$99,$FC,$D6,$9F,$AD,$73,$1E
	.dc.b	$83,$99,$B2,$99,$3B,$3C,$1E,$C7,$EC,$F2,$FA,$A7,$D4,$7D,$7C,$5F
	.dc.b	$C8,$5D,$9C,$8E,$A6,$0F,$01,$92,$1F,$95,$AF,$E0,$9F,$09,$F8,$22
	.dc.b	$47,$6D,$BF,$7D,$18,$8F,$94,$DA,$F1,$69,$FC,$C6,$31,$5C,$2B,$91
	.dc.b	$FC,$16,$32,$FF,$07,$C3,$DA,$EA,$BF,$9D,$46,$23,$EC,$36,$FB,$E5
	.dc.b	$F9,$79,$E0,$D7,$66,$27,$86,$1F,$C1,$63,$04,$36,$4A,$EB,$87,$9B
	.dc.b	$0F,$64,$F6,$8C,$60,$69,$F1,$7C,$B7,$47,$C1,$AF,$8E,$71,$21,$82
	.dc.b	$61,$E0,$72,$D7,$38,$F3,$8C,$1F,$0B,$E1,$FF,$73,$F0,$BF,$BA,$F8
	.dc.b	$7F,$DD,$FC,$37,$EF,$3E,$1F,$F7,$BF,$01,$FB,$EF,$87,$FD,$FF,$C3
	.dc.b	$FF,$00,$F8,$7C,$6B,$F2,$AB,$83,$F3,$30,$E8,$FC,$EC,$BF,$83,$E7
	.dc.b	$F0,$23,$CF,$AF,$94,$8F,$DF,$32,$53,$E0,$6F,$F1,$9F,$E9,$1F,$0F
	.dc.b	$FC,$48,$8F,$DC,$6B,$FB,$9F,$63,$F8,$BF,$C2,$FF,$18,$F8,$7F,$E3
	.dc.b	$3F,$0F,$FC,$6B,$DB,$FE,$7F,$F9,$D7,$E9,$70,$72,$FE,$5A,$1E,$B7
	.dc.b	$E1,$27,$F2,$47,$E0,$E7,$F1,$BD,$79,$EB,$FC,$9C,$E7,$7F,$C5,$3F
	.dc.b	$87,$6D,$70,$8D,$1D,$0D,$15,$D3,$E4,$8F,$F0,$CE,$1E,$48,$FD,$13
	.dc.b	$E9,$9F,$CF,$E3,$EE,$9F,$CF,$E7,$F0,$9F,$D0,$EB,$F1,$AB,$F4,$BB
	.dc.b	$7F,$20,$96,$FF,$3F,$3F,$CD,$C7,$E3,$0F,$A9,$5F,$8C,$CF,$EC,$CE
	.dc.b	$A3,$13,$A9,$CF,$5F,$E8,$63,$EE,$D6,$E7,$FC,$09,$C3,$F9,$49,$FD
	.dc.b	$AF,$4F,$E5,$1F,$DB,$F4,$FE,$51,$FD,$CF,$4F,$E5,$1F,$DD,$F4,$7E
	.dc.b	$51,$FE,$71,$D1,$F9,$47,$F9,$D7,$47,$E5,$1F,$E7,$9D,$1F,$94,$7C
	.dc.b	$7A,$F9,$5F,$76,$BE,$57,$D4,$47,$04,$8E,$1B,$FC,$77,$96,$FF,$0A
	.dc.b	$75,$D7,$FA,$1B,$EB,$5A,$E0,$F8,$B3,$9B,$FD,$B5,$7E,$12,$FF,$16
	.dc.b	$FE,$11,$B1,$F8,$42,$57,$E3,$09,$3F,$89,$6C,$F0,$6D,$FF,$21,$AE
	.dc.b	$35,$FE,$33,$E0,$F1,$57,$92,$3E,$2E,$18,$7F,$69,$8F,$E1,$79,$5D
	.dc.b	$9F,$AA,$24,$FE,$51,$AF,$E6,$46,$69,$7B,$FF,$B2,$F5,$FF,$67,$F0
	.dc.b	$1F,$B4,$F6,$3F,$6B,$F0,$3F,$B6,$F8,$2F,$DB,$FC,$1F,$D9,$CF,$ED
	.dc.b	$9F,$88,$3F,$89,$1F,$88,$3F,$8B,$1F,$88,$3F,$A0,$F8,$5F,$C6,$DC
	.dc.b	$1F,$88,$38,$1F,$88,$3A,$1D,$8F,$F0,$27,$0F,$89,$FD,$8D,$79,$9F
	.dc.b	$D9,$E1,$EA,$7F,$6B,$97,$F0,$EC,$FE,$1F,$17,$F0,$DD,$FF,$61,$C1
	.dc.b	$F8,$9E,$1F,$9D,$C3,$F8,$8C,$7E,$89,$0F,$C4,$C7,$F4,$CE,$FF,$E2
	.dc.b	$1C,$9F,$6F,$93,$EE,$C6,$09,$5F,$8B,$9D,$3F,$85,$25,$FE,$15,$B1
	.dc.b	$A8,$F1,$47,$80,$F1,$65,$F9,$D9,$FC,$D4,$7E,$13,$C3,$FE,$2B,$9F
	.dc.b	$BB,$1B,$3F,$83,$21,$F8,$3C,$3F,$05,$A7,$E0,$63,$F8,$DF,$27,$E0
	.dc.b	$5F,$E3,$BC,$9F,$81,$7F,$8F,$F2,$7E,$07,$4F,$EF,$BC,$DF,$BD,$7F
	.dc.b	$BF,$F3,$FF,$01,$7F,$C0,$3D,$1F,$C1,$5F,$F0,$1F,$3F,$E0,$70,$FF
	.dc.b	$1F,$75,$7F,$0D,$7F,$C8,$1D,$7F,$C4,$5F,$F2,$17,$67,$F1,$57,$F0
	.dc.b	$27,$88,$FC,$09,$E6,$3F,$02,$7A,$8F,$C0,$9F,$CD,$B6,$F8,$8F,$E8
	.dc.b	$BA,$FF,$52,$FF,$94,$39,$FF,$59,$E4,$7F,$33,$8F,$E6,$9C,$5F,$A7
	.dc.b	$8C,$47,$F4,$26,$BC,$DB,$1B,$8D,$B5,$E8,$D7,$AD,$27,$8B,$63,$8B
	.dc.b	$5E,$2D,$78,$C7,$02,$3C,$1A,$F0,$6B,$C0,$EB,$F0,$9F,$DC,$BC,$DF
	.dc.b	$EE,$73,$A9,$F7,$1C,$C8,$73,$29,$CC,$C1,$0C,$10,$E6,$64,$A7,$1E
	.dc.b	$6F,$81,$87,$C5,$8C,$90,$F9,$C7,$80,$C3,$B7,$27,$52,$9E,$12,$9C
	.dc.b	$30,$F9,$4D,$AF,$80,$E8,$93,$FC,$9F,$0F,$8D,$1F,$E0,$09,$DC,$F8
	.dc.b	$49,$5D,$27,$8B,$63,$33,$C2,$30,$47,$07,$D8,$BF,$87,$19,$AF,$80
	.dc.b	$74,$9D,$B8,$AE,$AD,$9E,$0F,$C3,$8C,$B6,$32,$FE,$4B,$E1,$7F,$C6
	.dc.b	$7F,$11,$F8,$28,$E5,$7D,$9E,$0D,$7E,$65,$F4,$6D,$F4,$7E,$FE,$1E
	.dc.b	$AA,$E1,$1F,$3C,$EA,$43,$A9,$4E,$6F,$07,$E0,$1F,$98,$3F,$C6,$39
	.dc.b	$79,$DE,$24,$38,$9C,$CF,$01,$92,$9F,$91,$8F,$F1,$EF,$C3,$FE,$C8
	.dc.b	$91,$FB,$26,$BF,$B2,$6B,$FB,$A6,$BC,$8E,$04,$38,$14,$E0,$53,$81
	.dc.b	$A2,$1F,$C1,$BE,$23,$FC,$CF,$F0,$BF,$E6,$89,$EF,$3F,$C6,$DB,$97
	.dc.b	$F7,$D6,$E4,$69,$27,$C0,$71,$2C,$F1,$7D,$19,$7E,$4A,$78,$47,$90
	.dc.b	$F3,$9E,$A3,$07,$32,$CF,$36,$D7,$8D,$EE,$D7,$33,$D9,$E8,$7F,$26
	.dc.b	$F7,$FF,$D0,$FF,$0F,$F0,$63,$8C,$ED,$F1,$90,$F1,$1C,$8C,$5F,$8D
	.dc.b	$C1,$A3,$37,$E3,$7D,$98,$7C,$17,$B9,$0F,$39,$4E,$73,$D5,$E7,$5E
	.dc.b	$57,$B1,$23,$4D,$A6,$5F,$7A,$3F,$D1,$B5,$C6,$B6,$4A,$F3,$4F,$63
	.dc.b	$C4,$53,$CA,$4B,$F5,$38,$3C,$F1,$C9,$F1,$57,$78,$C5,$77,$8C,$12
	.dc.b	$7E,$CE,$3F,$51,$F3,$9E,$23,$99,$67,$93,$6B,$A9,$2B,$F1,$B7,$FC
	.dc.b	$83,$E1,$BF,$B1,$92,$BF,$B1,$B5,$FE,$C6,$D7,$FB,$5B,$5F,$F5,$BB
	.dc.b	$5F,$F0,$03,$6F,$FC,$C4,$DB,$34,$D7,$E7,$6F,$87,$FF,$5B,$74,$BC
	.dc.b	$06,$3F,$E6,$CA,$7C,$27,$FB,$73,$0E,$53,$E8,$21,$E8,$29,$E8,$29
	.dc.b	$E8,$29,$E8,$2E,$5E,$86,$E5,$E8,$7C,$79,$FC,$97,$35,$E2,$7C,$04
	.dc.b	$3C,$05,$3C,$05,$3C,$05,$AF,$03,$6B,$C0,$D9,$F8,$F9,$FC,$6C,$BE
	.dc.b	$2D,$B2,$7C,$3F,$C5,$C3,$CA,$43,$91,$4F,$29,$0F,$29,$4F,$29,$67
	.dc.b	$CB,$97,$C1,$8E,$64,$35,$97,$A1,$B9,$7C,$56,$9E,$13,$B9,$0E,$E5
	.dc.b	$AF,$F5,$56,$DE,$F8,$FC,$16,$9E,$D7,$C0,$6E,$FC,$0D,$BF,$03,$4F
	.dc.b	$01,$D8,$97,$EC,$E2,$D4,$74,$33,$5D,$1B,$5D,$1A,$74,$AF,$B2,$FB
	.dc.b	$8E,$4F,$D5,$D7,$D4,$FD,$7C,$FD,$53,$B2,$1E,$22,$9D,$8E,$77,$E3
	.dc.b	$9E,$F1,$F8,$CC,$FF,$C7,$BA,$3E,$1B,$BB,$67,$F5,$7B,$FE,$B5,$FD
	.dc.b	$64,$E8,$8E,$B7,$FC,$6D,$71,$8F,$F4,$E3,$6B,$96,$1F,$91,$BE,$35
	.dc.b	$AC,$FC,$24,$7A,$1F,$BB,$70,$EC,$CC,$FC,$76,$DF,$F0,$56,$CF,$0B
	.dc.b	$F0,$CE,$67,$91,$23,$33,$BC,$3D,$6D,$9F,$11,$D0,$87,$18,$F8,$38
	.dc.b	$7B,$2B,$BB,$63,$33,$E6,$8F,$83,$D0,$F1,$33,$3F,$22,$78,$3F,$40
	.dc.b	$E0,$43,$45,$BE,$EF,$C2,$F8,$9F,$B7,$F1,$1F,$73,$E1,$7E,$EF,$C2
	.dc.b	$FD,$EF,$86,$FB,$FF,$0D,$F8,$0F,$88,$FC,$0E,$9F,$82,$F8,$5F,$C1
	.dc.b	$FC,$37,$E1,$3E,$1A,$EB,$F3,$A7,$6F,$D0,$BE,$6F,$D2,$C7,$BB,$F5
	.dc.b	$33,$F1,$BE,$53,$D1,$C3,$F0,$37,$F4,$C7,$B0,$8E,$0C,$F0,$FC,$78
	.dc.b	$FE,$7D,$F0,$FF,$98,$23,$F8,$27,$F3,$7F,$0B,$F9,$CF,$85,$FC,$EF
	.dc.b	$C3,$7E,$7B,$E1,$BF,$3F,$F0,$FF,$77,$AB,$F1,$58,$FD,$B3,$EF,$95
	.dc.b	$CE,$7F,$86,$8F,$C0,$B5,$D7,$57,$E3,$63,$04,$D7,$F1,$73,$C1,$AF
	.dc.b	$0E,$AF,$15,$71,$9C,$D7,$E6,$30,$F4,$5F,$E7,$B2,$F6,$3F,$A0,$D3
	.dc.b	$E0,$BF,$B0,$D7,$FB,$1A,$DB,$E4,$69,$FB,$18,$FA,$27,$E3,$4F,$9F
	.dc.b	$97,$D3,$23,$9C,$BE,$94,$62,$7F,$A5,$74,$FD,$CA,$FD,$FF,$07,$E0
	.dc.b	$1F,$E0,$3C,$3F,$84,$7F,$81,$F1,$7E,$21,$FE,$0B,$C7,$F8,$C7,$F8
	.dc.b	$3F,$25,$CF,$F6,$F3,$B7,$F7,$35,$CD,$FD,$DD,$F7,$7F,$7B,$87,$8D
	.dc.b	$FD,$FE,$3F,$AA,$CB,$E0,$65,$EE,$3D,$99,$FA,$F8,$7F,$8B,$3F,$B2
	.dc.b	$D7,$E3,$C6,$27,$64,$AE,$78,$7C,$48,$C4,$EE,$73,$7D,$70,$FB,$BC
	.dc.b	$52,$FF,$04,$7E,$3B,$0F,$C3,$36,$BF,$02,$47,$04,$BF,$C9,$C6,$5B
	.dc.b	$19,$C3,$F1,$4D,$BF,$E7,$B8,$7C,$39,$F3,$13,$AB,$07,$F1,$4E,$7F
	.dc.b	$F0,$3F,$9B,$F8,$AC,$19,$FF,$3A,$8C,$36,$3F,$AB,$C6,$33,$E3,$F0
	.dc.b	$98,$F5,$F3,$F0,$FA,$F8,$7D,$FC,$3F,$93,$1F,$E7,$3E,$07,$F1,$77
	.dc.b	$F9,$DF,$81,$FC,$5D,$FE,$7B,$E0,$7F,$17,$7F,$9F,$ED,$FC,$5D,$FE
	.dc.b	$83,$E0,$7F,$17,$69,$FC,$5C,$D9,$FC,$5C,$E6,$7F,$17,$3B,$9F,$C5
	.dc.b	$CF,$19,$FC,$5C,$F8,$3A,$FC,$4C,$7D,$51,$FD,$4B,$66,$9E,$F8,$C6
	.dc.b	$DF,$11,$87,$F9,$5B,$97,$F1,$BC,$3F,$42,$3F,$67,$CB,$F8,$87,$F1
	.dc.b	$71,$97,$F8,$31,$F8,$88,$FB,$E4,$E3,$FB,$ED,$E5,$FC,$74,$F0,$27
	.dc.b	$27,$E3,$A7,$51,$CE,$B8,$92,$BC,$15,$F9,$DA,$F2,$12,$7F,$11,$5E
	.dc.b	$BD,$FF,$0E,$FB,$8F,$8E,$25,$A7,$E1,$63,$0D,$8C,$E7,$F2,$67,$8C
	.dc.b	$70,$D3,$E7,$C6,$29,$AF,$E2,$63,$FC,$1F,$C5,$F8,$97,$FC,$23,$C5
	.dc.b	$F8,$97,$FC,$27,$C5,$F8,$97,$FC,$2B,$9F,$E2,$5F,$F0,$BF,$17,$E2
	.dc.b	$5A,$7E,$24,$D9,$F8,$93,$99,$F8,$93,$B9,$F8,$93,$C6,$7E,$27,$6F
	.dc.b	$D4,$63,$EA,$7F,$55,$9F,$BD,$FD,$66,$BF,$25,$FD,$65,$7C,$D3,$FD
	.dc.b	$39,$D5,$E5,$8F,$F0,$DB,$78,$FF,$C6,$2D,$9D,$9F,$E2,$47,$1B,$FF
	.dc.b	$A0,$5B,$C0,$69,$27,$81,$0F,$E6,$A5,$BF,$0B,$6F,$C2,$D3,$C2,$70
	.dc.b	$24,$70,$6C,$70,$69,$C2,$34,$48,$D3,$63,$AF,$1F,$F3,$6A,$F4,$47
	.dc.b	$2C,$3D,$CF,$8D,$AE,$70,$F8,$0E,$5A,$E6,$79,$DF,$7A,$E2,$DA,$DB
	.dc.b	$67,$78,$79,$89,$7E,$B6,$DF,$26,$DF,$26,$D7,$2A,$E0,$4A,$F4,$E1
	.dc.b	$FD,$EB,$85,$ED,$FF,$30,$25,$7F,$13,$7F,$99,$FB,$FF,$CD,$3D,$8F
	.dc.b	$E6,$BE,$A7,$F3,$6F,$63,$F9,$BF,$AD,$FC,$E3,$D8,$FE,$73,$97,$F3
	.dc.b	$AF,$63,$F9,$DF,$C0,$7F,$3C,$F8,$1F,$E7,$BF,$05,$FC,$FB,$E1,$FF
	.dc.b	$9F,$FC,$27,$F4,$0F,$87,$FE,$83,$F0,$DF,$E2,$BC,$8E,$78,$BF,$93
	.dc.b	$57,$17,$F9,$46,$1D,$5F,$E5,$59,$78,$5F,$E5,$97,$E5,$7F,$97,$6B
	.dc.b	$E9,$7F,$98,$6F,$ED,$7F,$99,$70,$FC,$27,$F8,$37,$17,$F8,$8B,$8F
	.dc.b	$F0,$D1,$F7,$F9,$7E,$77,$2F,$D3,$8F,$9D,$CF,$F4,$DB,$3A,$8C,$4E
	.dc.b	$C8,$E0,$91,$B2,$4F,$72,$1F,$8D,$2F,$47,$E2,$5B,$D1,$FA,$3C,$3C
	.dc.b	$F1,$FD,$C7,$0F,$CC,$BF,$E3,$C6,$BF,$A9,$31,$87,$E6,$9B,$87,$F7
	.dc.b	$97,$44,$3F,$6A,$68,$D4,$F4,$72,$6D,$C1,$27,$B4,$65,$AF,$02,$46
	.dc.b	$E7,$C5,$18,$9D,$CE,$63,$C3,$18,$24,$F0,$9D,$4F,$C3,$9C,$BC,$27
	.dc.b	$51,$88,$D4,$7F,$98,$BD,$0F,$F6,$FB,$A2,$47,$12,$1C,$0A,$7C,$53
	.dc.b	$04,$30,$43,$E1,$3D,$1D,$18,$39,$94,$C9,$F0,$DE,$0F,$33,$07,$02
	.dc.b	$9C,$2B,$E1,$BB,$7E,$2C,$FB,$A3,$84,$69,$F0,$1B,$AF,$8C,$DA,$F0
	.dc.b	$BA,$21,$E0,$34,$74,$8E,$2E,$A7,$B3,$B3,$11,$A7,$07,$AA,$BC,$CF
	.dc.b	$AE,$35,$3C,$88,$E0,$95,$E6,$9E,$73,$A8,$C3,$63,$CD,$18,$9E,$31
	.dc.b	$82,$1E,$18,$D1,$F8,$7F,$53,$F9,$9C,$7E,$D0,$FE,$06,$5A,$FE,$03
	.dc.b	$3F,$BE,$8F,$DE,$13,$2F,$DE,$37,$2E,$6D,$39,$96,$B9,$BD,$70,$FE
	.dc.b	$16,$FF,$08,$CB,$FD,$73,$F9,$E2,$FF,$D7,$B8,$5A,$FF,$17,$4D,$7F
	.dc.b	$C5,$2D,$8D,$60,$8C,$52,$B8,$FC,$47,$48,$C5,$6F,$4C,$57,$6F,$81
	.dc.b	$B3,$CF,$E1,$AE,$2F,$85,$BF,$0F,$8F,$86,$CF,$C4,$6B,$E1,$F7,$C7
	.dc.b	$C3,$E1,$F8,$FC,$2F,$2F,$87,$E7,$F0,$FD,$3E,$13,$AF,$C2,$DE,$2F
	.dc.b	$47,$37,$B3,$81,$7C,$67,$3C,$BE,$A9,$65,$DE,$72,$7C,$8D,$3C,$8E
	.dc.b	$38,$FE,$36,$FE,$98,$C5,$B8,$7D,$FC,$5F,$03,$8F,$E0,$F2,$7C,$2E
	.dc.b	$5F,$87,$CD,$F1,$36,$99,$7F,$8E,$F4,$FB,$EF,$E0,$7A,$3E,$FB,$7A
	.dc.b	$7F,$19,$C5,$F8,$A2,$5F,$E2,$9B,$3F,$8A,$6D,$F4,$6D,$F4,$6D,$F4
	.dc.b	$BF,$E9,$DE,$77,$C5,$7C,$B8,$7E,$21,$B5,$F3,$71,$F6,$4F,$C2,$CF
	.dc.b	$E0,$3F,$13,$5F,$BE,$FC,$6E,$9F,$BE,$FC,$8E,$9F,$BE,$FC,$AE,$9F
	.dc.b	$BE,$FC,$CE,$9F,$BE,$FC,$EE,$9F,$BE,$D3,$E8,$F4,$7D,$D7,$F3,$BD
	.dc.b	$1F,$6D,$FC,$FF,$47,$DA,$7F,$43,$D1,$F6,$5F,$D1,$F3,$FD,$87,$ED
	.dc.b	$F3,$FD,$87,$EE,$F3,$FD,$87,$EF,$F1,$FC,$4E,$AF,$D8,$FC,$3F,$F5
	.dc.b	$3E,$A3,$49,$5F,$DA,$A3,$2D,$8C,$BF,$DE,$E3,$2D,$8C,$BF,$DF,$63
	.dc.b	$2D,$8C,$BF,$9B,$6B,$FC,$01,$BD,$5F,$BD,$6C,$BE,$1B,$F6,$B1,$82
	.dc.b	$57,$94,$87,$8C,$A7,$84,$A7,$84,$A7,$84,$B5,$E1,$79,$E2,$D7,$84
	.dc.b	$75,$21,$D4,$A6,$8C,$10,$C1,$0F,$01,$C0,$A6,$CA,$6E,$7C,$8E,$DA
	.dc.b	$FC,$FB,$F8,$AF,$AB,$0F,$A2,$F1,$6B,$D0,$93,$C9,$D6,$1F,$31,$ED
	.dc.b	$3C,$27,$51,$C2,$3F,$03,$87,$C5,$8C,$57,$17,$D5,$5C,$09,$5C,$C9
	.dc.b	$3B,$9F,$09,$E9,$70,$F7,$F8,$7F,$6B,$E4,$8C,$D7,$46,$D7,$49,$E4
	.dc.b	$43,$81,$FD,$1E,$FD,$15,$C6,$34,$7A,$27,$93,$C0,$86,$6B,$B9,$1F
	.dc.b	$D4,$61,$E6,$74,$74,$3D,$A7,$22,$9A,$30,$7C,$A3,$CD,$8F,$F4,$1F
	.dc.b	$86,$FD,$C1,$23,$F6,$CD,$7F,$68,$D7,$F6,$4D,$79,$1C,$08,$70,$29
	.dc.b	$C0,$A7,$03,$44,$3F,$CF,$5F,$9E,$BD,$7F,$ED,$33,$DE,$7F,$BD,$B7
	.dc.b	$2F,$EF,$0D,$C8,$D2,$4F,$80,$E2,$59,$E2,$D9,$E2,$FC,$63,$67,$94
	.dc.b	$F4,$1E,$B3,$06,$8C,$10,$D9,$0D,$16,$3C,$2F,$26,$BC,$4D,$91,$CB
	.dc.b	$FE,$12,$BF,$8C,$FF,$72,$AF,$7B,$CE,$BD,$6D,$8C,$BD,$6B,$9B,$D6
	.dc.b	$76,$E5,$AF,$9D,$E7,$9F,$E7,$9C,$C7,$1A,$EA,$ED,$AE,$4F,$9C,$E2
	.dc.b	$BC,$24,$9E,$6E,$2D,$F1,$1F,$82,$BE,$24,$3D,$85,$3D,$17,$D6,$F9
	.dc.b	$91,$FC,$E6,$1E,$97,$44,$9E,$47,$A4,$A7,$A4,$EE,$43,$F2,$F8,$F9
	.dc.b	$5C,$1C,$4A,$71,$3C,$A7,$4C,$7E,$04,$FD,$8A,$FE,$13,$8B,$E2,$3F
	.dc.b	$D7,$CD,$BF,$F6,$9B,$5F,$F7,$13,$5F,$D4,$C7,$02,$3C,$1A,$F0,$6B
	.dc.b	$C0,$D1,$0D,$14,$D1,$82,$1F,$B8,$F6,$BF,$CC,$33,$E3,$8F,$CB,$4F
	.dc.b	$66,$9C,$CA,$72,$29,$D8,$86,$08,$30,$4F,$78,$C3,$63,$33,$E1,$9C
	.dc.b	$BF,$E9,$B6,$9E,$32,$9E,$32,$9E,$22,$9E,$28,$EA,$F1,$6B,$C5,$AF
	.dc.b	$16,$BC,$67,$CC,$49,$FB,$FE,$EF,$CA,$7D,$15,$C2,$BC,$24,$70,$4A
	.dc.b	$E5,$1E,$36,$BE,$36,$BE,$37,$F1,$58,$7E,$42,$BB,$C6,$A7,$CD,$18
	.dc.b	$23,$D5,$AF,$56,$BD,$70,$F5,$3B,$21,$B2,$9E,$38,$E2,$F8,$0D,$D7
	.dc.b	$95,$D9,$93,$91,$0E,$46,$8F,$8D,$5E,$B6,$CF,$BB,$0F,$2C,$71,$3A
	.dc.b	$1C,$8C,$1C,$0E,$98,$7C,$F9,$F8,$37,$FB,$8A,$FD,$09,$2B,$FD,$51
	.dc.b	$68,$C4,$3C,$0F,$F6,$8C,$E6,$37,$3A,$9C,$B7,$8F,$FA,$03,$67,$F3
	.dc.b	$ED,$C7,$F4,$6D,$BE,$93,$DE,$B8,$BD,$88,$64,$E6,$62,$FC,$2D,$8E
	.dc.b	$2E,$EF,$BB,$93,$89,$9A,$F0,$D7,$28,$C5,$F9,$48,$F1,$AE,$6B,$8D
	.dc.b	$E0,$F8,$2F,$91,$31,$C1,$2B,$C3,$19,$6C,$66,$FA,$CE,$5B,$19,$C7
	.dc.b	$CE,$DC,$7C,$B7,$E8,$3C,$D3,$A7,$A9,$2B,$9B,$63,$51,$82,$38,$24
	.dc.b	$F6,$BF,$24,$62,$BE,$AE,$3F,$09,$FE,$3B,$ED,$FF,$2E,$23,$F8,$E7
	.dc.b	$F9,$8F,$BD,$FC,$CB,$DC,$FE,$67,$EF,$FF,$34,$F7,$7F,$9A,$FC,$07
	.dc.b	$F3,$6D,$FF,$9B,$FC,$0F,$F3,$8F,$7F,$F9,$CF,$95,$FC,$EB,$E0,$BF
	.dc.b	$9D,$FC,$1F,$F3,$CF,$87,$FE,$7B,$F0,$BF,$CF,$BE,$1F,$F9,$FF,$C3
	.dc.b	$FF,$40,$EB,$BC,$7F,$A0,$9D,$BF,$A2,$BE,$6F,$E9,$31,$EE,$FE,$9B
	.dc.b	$3F,$1B,$FA,$8D,$7C,$EF,$EA,$B7,$F5,$BF,$AC,$E1,$F7,$BF,$04,$87
	.dc.b	$E2,$67,$ED,$3F,$63,$9F,$E7,$74,$E2,$3E,$E1,$3A,$3E,$8C,$EA,$3E
	.dc.b	$D3,$7A,$7E,$CB,$7A,$7E,$B4,$F1,$BE,$97,$CA,$3C,$04,$34,$74,$3E
	.dc.b	$EE,$1F,$AC,$7F,$CC,$37,$FD,$09,$FE,$FE,$D7,$FB,$FB,$5D,$9A,$21
	.dc.b	$A2,$9A,$30,$43,$04,$3C,$44,$3F,$76,$64,$A6,$70,$FD,$BB,$96,$B9
	.dc.b	$C3,$F8,$8B,$96,$B9,$9F,$53,$4F,$39,$FC,$3B,$AB,$F9,$63,$83,$C2
	.dc.b	$4A,$EE,$DA,$EA,$DA,$EA,$D3,$94,$F0,$24,$F0,$6C,$F0,$6C,$F0,$9D
	.dc.b	$12,$74,$D8,$D4,$60,$8E,$0F,$AB,$5E,$C7,$E9,$1F,$04,$A7,$9C,$B5
	.dc.b	$E2,$6D,$77,$6D,$76,$69,$DC,$A4,$3E,$93,$B6,$9B,$29,$C2,$BC,$CE
	.dc.b	$08,$79,$CA,$79,$CB,$5E,$76,$D7,$9D,$B5,$E1,$79,$10,$E4,$59,$E4
	.dc.b	$D9,$FE,$B7,$5F,$27,$0F,$D7,$4F,$EB,$2B,$F6,$AF,$EB,$E3,$24,$C7
	.dc.b	$F6,$A4,$8F,$D9,$1C,$6F,$F7,$73,$B8,$CB,$6B,$AC,$6F,$0F,$E0,$44
	.dc.b	$47,$F1,$13,$FC,$AD,$F9,$F6,$1C,$D3,$0D,$DA,$B5,$5F,$DB,$5B,$5B
	.dc.b	$8E,$7F,$09,$CE,$31,$5C,$7E,$23,$18,$74,$C1,$F0,$D8,$C7,$C7,$F1
	.dc.b	$17,$0F,$07,$C4,$5B,$7B,$37,$DA,$C7,$C4,$67,$DC,$D7,$C4,$6F,$DE
	.dc.b	$E1,$F1,$1C,$7E,$03,$97,$C4,$73,$F8,$2E,$9F,$11,$38,$7E,$0F,$1F
	.dc.b	$96,$77,$18,$3E,$3D,$F4,$71,$C3,$F1,$72,$F0,$C6,$38,$1A,$F9,$A3
	.dc.b	$1B,$FA,$F8,$3D,$9C,$3E,$DE,$2F,$77,$1F,$BF,$93,$E0,$73,$7C,$1C
	.dc.b	$61,$F6,$BC,$BF,$EE,$1B,$7E,$03,$9F,$F0,$5D,$3F,$86,$E3,$FC,$54
	.dc.b	$E5,$B1,$9B,$FC,$4C,$63,$A7,$F0,$D1,$F8,$93,$F2,$D1,$96,$9C,$09
	.dc.b	$5F,$90,$C3,$C7,$C3,$F8,$4A,$FB,$D5,$E8,$3F,$11,$97,$B5,$F8,$DA
	.dc.b	$7C,$17,$E4,$6D,$F1,$5F,$8E,$7C,$BE,$9F,$BC,$FC,$DE,$9F,$BC,$FC
	.dc.b	$FE,$9F,$BC,$FD,$1E,$9F,$BC,$FD,$3E,$9F,$BC,$FE,$67,$A7,$EE,$BF
	.dc.b	$9B,$E9,$FB,$8F,$E7,$7A,$7E,$DB,$F9,$FE,$9F,$B4,$FE,$87,$A7,$CC
	.dc.b	$FE,$8F,$5F,$53,$FA,$5D,$BD,$CF,$C0,$3F,$08,$7C,$3E,$2F,$8F,$F0
	.dc.b	$9F,$9D,$24,$7E,$25,$AF,$EB,$DF,$AB,$CF,$F7,$B9,$AF,$27,$D5,$AF
	.dc.b	$B0,$49,$D3,$F6,$39,$BF,$04,$FD,$8E,$7F,$F2,$0F,$3B,$1E,$F7,$FB
	.dc.b	$88,$C5,$78,$C8,$7F,$09,$BF,$E0,$95,$C3,$0F,$DE,$36,$7F,$86,$B6
	.dc.b	$7F,$85,$74,$FE,$82,$7F,$4D,$D5,$FB,$B7,$EF,$1F,$A9,$E0,$FA,$39
	.dc.b	$78,$F8,$7E,$AB,$E5,$E2,$FB,$2F,$9F,$8F,$EE,$BE,$9E,$4F,$C0,$E1
	.dc.b	$FC,$5F,$97,$F0,$AF,$F1,$9E,$6F,$C4,$BF,$C6,$F9,$FF,$1A,$FF,$1D
	.dc.b	$E8,$F2,$3F,$C7,$F5,$F4,$3F,$C8,$76,$F6,$71,$7C,$68,$FA,$C7,$C6
	.dc.b	$CF,$E5,$70,$FF,$2D,$E5,$FD,$2F,$C4,$7F,$1B,$25,$7F,$07,$F8,$6F
	.dc.b	$E3,$DF,$11,$FC,$7F,$E2,$3E,$EC,$FD,$7E,$9B,$CF,$F6,$89,$D3,$FB
	.dc.b	$B6,$C6,$A3,$04,$7A,$4F,$F1,$5B,$FE,$59,$87,$8A,$3F,$9E,$63,$E7
	.dc.b	$F0,$3E,$DC,$7A,$FA,$AB,$EE,$26,$DE,$E7,$F9,$07,$07,$C6,$8F,$C5
	.dc.b	$70,$FC,$A7,$F1,$9C,$5F,$39,$FC,$77,$1F,$D2,$7F,$21,$C9,$F5,$9F
	.dc.b	$C9,$72,$FD,$A7,$EE,$63,$FB,$B3,$F9,$77,$3F,$E0,$B3,$FE,$99,$D1
	.dc.b	$F8,$67,$FA,$77,$4F,$E2,$9F,$EA,$1D,$5F,$BA,$C3,$F3,$DD,$BF,$BA
	.dc.b	$7F,$41,$DB,$FB,$A7,$D9,$97,$BA,$FE,$0C,$60,$CF,$4F,$EE,$A7,$E3
	.dc.b	$1F,$A0,$2F,$47,$E9,$9B,$D6,$E3,$FA,$DD,$12,$7F,$6D,$D3,$F5,$DB
	.dc.b	$19,$7F,$79,$97,$F3,$CB,$E5,$1D,$3A,$FF,$26,$7F,$08,$CB,$FA,$2F
	.dc.b	$87,$FD,$D6,$7F,$DC,$8D,$D3,$FB,$AB,$70,$FE,$FE,$BE,$1B,$FB,$92
	.dc.b	$F0,$AF,$C3,$7C,$5A,$FE,$F5,$C1,$F2,$5F,$EF,$9C,$3F,$35,$F9,$F9
	.dc.b	$7F,$37,$3F,$B2,$71,$FD,$57,$FB,$59,$FC,$DC,$FB,$67,$DD,$3F,$C1
	.dc.b	$DE,$07,$E0,$67,$FC,$11,$CF,$F8,$57,$FC,$15,$D1,$F8,$97,$FC,$19
	.dc.b	$D3,$F8,$D7,$FC,$1D,$D5,$F9,$17,$FC,$21,$D7,$F9,$57,$FC,$25,$CF
	.dc.b	$EC,$CB,$FB,$DF,$07,$BD,$BC,$3F,$0A,$A4,$7F,$71,$EC,$FC,$1C,$48
	.dc.b	$FC,$01,$F8,$72,$F1,$FE,$1E,$3F,$9B,$3F,$CD,$23,$ED,$9F,$BD,$3E
	.dc.b	$F9,$0C,$F4,$FE,$F1,$C4,$7F,$8B,$3A,$7F,$9C,$F3,$FE,$02,$7F,$04
	.dc.b	$79,$45,$32,$FB,$D1,$86,$C6,$70,$FB,$F9,$7E,$03,$9B,$E2,$5F,$F1
	.dc.b	$F3,$FB,$A9,$FC,$E0,$FE,$EA,$7F,$27,$3F,$9C,$1F,$7C,$FE,$D4,$7F
	.dc.b	$75,$3F,$C4,$9D,$7F,$B7,$7F,$CA,$3D,$9F,$CE,$2F,$FC,$89,$C9,$F7
	.dc.b	$E7,$FC,$AB,$C9,$F7,$DF,$F2,$BF,$27,$DF,$7F,$CB,$3C,$9F,$7D,$FF
	.dc.b	$2D,$F2,$7D,$F7,$FC,$BB,$C9,$F7,$DF,$F2,$69,$F7,$BA,$3D,$CF,$DE
	.dc.b	$3F,$03,$97,$DE,$7F,$AF,$70,$3D,$CF,$D1,$92,$3F,$0F,$F0,$DF,$A6
	.dc.b	$F7,$BF,$4F,$F0,$DF,$A8,$F8,$0F,$D4,$FC,$37,$EA,$BD,$EF,$D5,$FC
	.dc.b	$37,$EB,$3E,$0F,$F5,$BF,$0D,$FA,$EF,$85,$FD,$7F,$C3,$7E,$C3,$E2
	.dc.b	$3F,$63,$F0,$DF,$B2,$F8,$6F,$89,$DF,$FD,$07,$C9,$76,$E7,$8F,$F3
	.dc.b	$37,$C5,$FC,$DE,$3D,$5F,$CE,$E7,$E1,$7F,$3F,$AF,$95,$FD,$0E,$FE
	.dc.b	$97,$F4,$7C,$3E,$D7,$DE,$7E,$9B,$93,$E1,$3F,$1B,$E1,$FF,$83,$91
	.dc.b	$FD,$5F,$C2,$7F,$9E,$FF,$3F,$7C,$35,$F8,$8C,$7C,$3E,$7D,$ED,$7A
	.dc.b	$3B,$F8,$7E,$1F,$0F,$C7,$DC,$E5,$F0,$FC,$FD,$EE,$9F,$0F,$D7,$D2
	.dc.b	$ED,$F0,$FF,$1B,$AD,$E6,$FE,$37,$8F,$F1,$E7,$F9,$E3,$89,$5C,$63
	.dc.b	$18,$75,$75,$97,$85,$E1,$A7,$95,$E5,$B7,$A5,$E9,$C1,$ED,$7B,$71
	.dc.b	$7F,$2A,$E3,$F8,$91,$FC,$CB,$93,$FB,$73,$FB,$3E,$5F,$97,$E9,$FF
	.dc.b	$34,$7F,$98,$F8,$DF,$46,$7F,$9C,$F4,$7D,$57,$F9,$DF,$4F,$D9,$7F
	.dc.b	$9E,$F5,$7D,$D7,$F9,$FF,$5F,$E0,$5F,$E8,$3D,$9F,$85,$7F,$A1,$F6
	.dc.b	$FE,$25,$FE,$1F,$E9,$FF,$82,$FC,$87,$85,$F9,$29,$FE,$4B,$E1,$FE
	.dc.b	$59,$FE,$4F,$E2,$7E,$69,$FE,$53,$E2,$FE,$79,$FE,$57,$E3,$7E,$89
	.dc.b	$FE,$5B,$E3,$FE,$99,$FE,$5F,$E4,$7B,$9F,$C6,$10,$FC,$61,$FC,$D3
	.dc.b	$97,$E4,$F2,$FF,$58,$E6,$FC,$67,$3F,$CF,$A4,$FF,$B4,$3B,$9D,$3F
	.dc.b	$56,$73,$E4,$FF,$B7,$25,$D7,$F6,$E7,$27,$F2,$88,$76,$7E,$06,$33
	.dc.b	$E4,$A5,$DD,$F8,$79,$CC,$3B,$FF,$19,$18,$F0,$3F,$29,$E0,$FE,$4F
	.dc.b	$0F,$F7,$67,$93,$F9,$8A,$FC,$A7,$89,$F9,$C7,$F2,$DE,$2F,$E8,$1F
	.dc.b	$CC,$78,$DF,$A4,$7F,$35,$E3,$FE,$A1,$FC,$E7,$91,$FA,$C7,$F3,$DE
	.dc.b	$4F,$C6,$7F,$66,$7C,$B2,$1F,$C7,$BC,$8F,$C8,$F1,$FF,$80,$9F,$A9
	.dc.b	$1F,$E0,$8E,$9F,$B0,$FF,$82,$BA,$BE,$E3,$FE,$0C,$EB,$FC,$03,$FE
	.dc.b	$0E,$EC,$FC,$23,$FE,$10,$ED,$FC,$43,$FE,$12,$EE,$FC,$63,$FE,$14
	.dc.b	$EF,$FE,$3B,$E0,$FF,$96,$FC,$9F,$E3,$B3,$FD,$D2,$77,$DD,$FE,$18
	.dc.b	$BE,$6F,$F8,$6B,$0E,$EF,$F8,$73,$1F,$1B,$FE,$1E,$CB,$CE,$FF,$88
	.dc.b	$33,$F5,$BF,$E2,$2D,$3D,$EF,$EC,$72,$FC,$97,$B1,$FE,$2A,$F0,$BF
	.dc.b	$23,$C1,$FE,$26,$7F,$79,$3F,$BE,$3E,$79,$FC,$08,$FA,$67,$F0,$63
	.dc.b	$EB,$9F,$C2,$8F,$B6,$7F,$0E,$3E,$F9,$FC,$48,$FC,$19,$FC,$5F,$AB
	.dc.b	$F1,$2F,$F8,$96,$FF,$D4,$0E,$4D,$9F,$92,$EB,$FC,$A4,$7E,$8F,$B7
	.dc.b	$F3,$4F,$E9,$7B,$BF,$40,$FE,$9F,$BF,$F4,$CF,$EA,$7C,$0F,$D6,$3F
	.dc.b	$AB,$F0,$7D,$F8,$FF,$19,$3F,$13,$AF,$F9,$3B,$9F,$F1,$31,$FC,$6C
	.dc.b	$FE,$8E,$7E,$24,$FE,$96,$7F,$02,$3F,$A7,$9F,$C1,$8F,$EA,$67,$F0
	.dc.b	$A3,$FA,$B9,$FC,$38,$FE,$B6,$7F,$12,$3F,$AF,$9F,$85,$3F,$C4,$BB
	.dc.b	$1F,$F1,$0F,$3F,$EF,$9D,$1F,$C8,$0E,$47,$F2,$43,$B1,$FC,$A0,$F1
	.dc.b	$1F,$CB,$0F,$31,$FC,$C0,$F5,$1F,$CD,$0F,$E3,$67,$C0,$C3,$F1,$4F
	.dc.b	$F9,$C2,$BF,$9F,$F7,$FF,$8C,$F8,$3F,$17,$3F,$E3,$37,$E6,$D7,$CF
	.dc.b	$8F,$C5,$9F,$C3,$33,$FA,$B8,$FF,$19,$D3,$EC,$BF,$C6,$F5,$FB,$AF
	.dc.b	$F1,$DD,$BF,$02,$FF,$1F,$DF,$F8,$9E,$3F,$C9,$78,$3F,$17,$97,$F9
	.dc.b	$B8,$FF,$38,$9F,$C5,$CF,$C5,$C7,$F2,$B9,$E7,$1F,$C9,$6F,$BB,$FC
	.dc.b	$9F,$0F,$1B,$FC,$A7,$1F,$3B,$FC,$AF,$2F,$C5,$E5,$FC,$C7,$83,$DE
	.dc.b	$FE,$AE,$7F,$17,$DD,$FC,$D7,$C3,$FF,$04,$12,$3F,$BC,$7C,$3F,$F8
	.dc.b	$2F,$E0,$BF,$C1,$9F,$0F,$FE,$0D,$F8,$4F,$F0,$77,$C3,$FF,$83,$FE
	.dc.b	$1B,$FC,$21,$F0,$FF,$83,$C7,$F9,$CF,$E8,$15,$BF,$0F,$9E,$56,$FF
	.dc.b	$89,$E8,$5B,$E2,$35,$39,$87,$C4,$63,$1F,$F3,$7C,$7F,$AD,$1E,$FF
	.dc.b	$07,$9C,$39,$FB,$16,$9F,$07,$9E,$FF,$F7,$D4,$63,$0F,$CD,$69,$E4
	.dc.b	$7A,$DF,$1C,$79,$7C,$17,$3C,$FA,$6B,$D7,$6E,$D8,$7C,$23,$FC,$41
	.dc.b	$C1,$F1,$9F,$F1,$17,$17,$CA,$7F,$C4,$9C,$9F,$E3,$7E,$5F,$A1,$1F
	.dc.b	$D7,$39,$BF,$33,$E8,$FF,$A2,$E3,$FC,$55,$1A,$DF,$FB,$85,$7D,$59
	.dc.b	$FF,$17,$73,$FD,$57,$FC,$61,$CF,$F5,$5F,$F1,$97,$3F,$99,$F7,$F3
	.dc.b	$FC,$DD,$BD,$71,$83,$E6,$F0,$7C,$07,$1C,$FF,$5B,$8F,$E6,$72,$7C
	.dc.b	$DE,$7F,$ED,$DE,$57,$F9,$0A,$F7,$CF,$F3,$5F,$C1,$D7,$CB,$73,$CF
	.dc.b	$F2,$DD,$F3,$FC,$B7,$89,$F2,$CF,$EF,$FC,$FF,$2D,$FF,$00,$F3,$FC
	.dc.b	$B7,$FC,$07,$CF,$F2,$DF,$8E,$41,$E1,$BA,$BF,$19,$ED,$FE,$14,$93
	.dc.b	$E8,$F8,$4F,$C4,$7B,$BF,$89,$F8,$4F,$C5,$7B,$FF,$8B,$F8,$1F,$C6
	.dc.b	$7C,$0F,$E3,$7D,$7F,$C7,$7B,$1F,$8F,$F8,$4F,$C8,$7C,$2F,$E4,$7E
	.dc.b	$1B,$F2,$5F,$0F,$F9,$3F,$88,$FC,$B7,$E8,$3F,$EC,$4E,$BF,$EE,$3C
	.dc.b	$3F,$ED,$26,$F8,$DF,$93,$EF,$57,$FB,$2B,$B5,$87,$4C,$3E,$7E,$5E
	.dc.b	$07,$E8,$E9,$E4,$7E,$9E,$DE,$87,$EA,$F0,$7B,$1F,$AF,$C5,$F0,$1F
	.dc.b	$84,$7D,$AE,$5F,$8A,$FC,$AE,$6F,$E1,$0F,$CD,$DB,$FC,$85,$E3,$FD
	.dc.b	$18,$FA,$67,$E5,$0F,$D0,$9F,$94,$3F,$4A,$7E,$50,$FB,$E7,$E2,$3C
	.dc.b	$2F,$C9,$BF,$C2,$BC,$2F,$C9,$3F,$BE,$3F,$1B,$E0,$7E,$45,$FE,$ED
	.dc.b	$E1,$FF,$B7,$1F,$C8,$F8,$5F,$96,$7F,$27,$E1,$F2,$CB,$E6,$71,$FF
	.dc.b	$0C,$9F,$C8,$9E,$13,$C6,$7F,$93,$BC,$4F,$D5,$3F,$E5,$0F,$17,$F5
	.dc.b	$CF,$F9,$4B,$C6,$FD,$93,$FE,$54,$F1,$FF,$6D,$9F,$C6,$8F,$90,$7F
	.dc.b	$97,$F7,$CB,$4F,$E5,$E7,$F0,$43,$E4,$1F,$9E,$3E,$41,$FA,$23,$E4
	.dc.b	$1F,$CB,$8F,$D4,$1F,$CC,$3C,$6F,$D6,$3F,$CC,$BC,$7F,$D8,$3F,$B2
	.dc.b	$3E,$29,$FB,$CF,$27,$F9,$1E,$FF,$1E,$3E,$71,$C4,$FD,$DF,$0F,$D4
	.dc.b	$DF,$E4,$F1,$7F,$0B,$9F,$8C,$7D,$C3,$E3,$1F,$80,$3E,$31,$ED,$3E
	.dc.b	$31,$F8,$9C,$FE,$37,$8B,$FB,$58,$F8,$87,$F2,$33,$47,$FA,$1F,$5F
	.dc.b	$9F,$7F,$E9,$1D,$BE,$9B,$FE,$93,$DF,$EB,$BF,$E9,$5E,$0F,$33,$E8
	.dc.b	$BF,$BE,$7E,$34,$FE,$26,$7E,$18,$FC,$69,$F1,$8F,$EF,$04,$7F,$13
	.dc.b	$A7,$C7,$7F,$17,$C5,$F8,$DC,$FF,$AC,$6B,$F1,$DF,$E5,$D7,$F8,$CA
	.dc.b	$F2,$3F,$1D,$F4,$3F,$1D,$F6,$3F,$1D,$FE,$7E,$FC,$2E,$DF,$F5,$27
	.dc.b	$17,$F8,$7B,$8D,$CD,$FB,$6B,$FC,$EB,$FD,$87,$C5,$FD,$E6,$1F,$DD
	.dc.b	$3A,$3F,$80,$3F,$DD,$BA,$7F,$82,$3F,$DE,$3A,$BF,$84,$3F,$DE,$BA
	.dc.b	$FF,$86,$3F,$DF,$3B,$3F,$88,$3F,$DF,$BB,$7F,$8D,$EF,$FE,$8C,$F3
	.dc.b	$3F,$B7,$1F,$DC,$FC,$0F,$E3,$71,$FD,$FF,$C0,$FE,$35,$5C,$B7,$FD
	.dc.b	$9E,$3D,$9F,$DA,$E5,$E2,$7F,$6F,$9F,$99,$FD,$CE,$9E,$A7,$F7,$7A
	.dc.b	$FB,$9F,$DE,$ED,$F0,$F3,$FE,$7F,$B7,$F9,$5F,$AB,$F1,$DC,$7F,$2A
	.dc.b	$3E,$67,$95,$FD,$BD,$FC,$A9,$FC,$68,$FC,$C9,$FB,$33,$F3,$A7,$EC
	.dc.b	$CF,$D0,$9F,$B3,$3F,$85,$1F,$B3,$3F,$52,$7E,$38,$FD,$6E,$1F,$C4
	.dc.b	$BC,$0F,$E3,$51,$F8,$DC,$3F,$8D,$3F,$E2,$AE,$8F,$C7,$47,$F8,$8F
	.dc.b	$A3,$F1,$D3,$FD,$D8,$FC,$DD,$FE,$75,$F1,$1E,$47,$FD,$ED,$E5,$7F
	.dc.b	$69,$7F,$DF,$1E,$57,$F6,$97,$FD,$F5,$E5,$7F,$39,$7F,$9D,$D7,$F6
	.dc.b	$9D,$7E,$3F,$2B,$E2,$7F,$AC,$92,$7F,$C0,$B1,$96,$C6,$5F,$F7,$DC
	.dc.b	$65,$B1,$9A,$FE,$BD,$39,$6C,$67,$3F,$C4,$36,$9B,$FF,$64,$8C,$36
	.dc.b	$3F,$71,$5F,$DA,$7C,$2F,$E3,$B1,$E4,$21,$E2,$2C,$78,$1B,$7E,$06
	.dc.b	$9E,$0C,$3C,$D1,$98,$E4,$49,$E4,$D8,$D4,$63,$1F,$97,$8F,$A2,$F9
	.dc.b	$DF,$58,$C1,$1F,$0E,$3E,$D7,$A1,$B3,$33,$CA,$35,$7E,$22,$61,$FB
	.dc.b	$1F,$89,$FF,$1B,$F9,$3F,$E1,$27,$B6,$1D,$2F,$8D,$7A,$09,$1D,$CD
	.dc.b	$61,$DC,$98,$F2,$C7,$BC,$ED,$F8,$1E,$3F,$F0,$1F,$89,$FF,$10,$92
	.dc.b	$7F,$13,$EA,$FF,$89,$3E,$27,$FC,$4B,$EB,$FF,$89,$BE,$27,$FC,$4F
	.dc.b	$EC,$FF,$8A,$3E,$27,$FC,$53,$ED,$FF,$8A,$BE,$27,$FC,$57,$EE,$FF
	.dc.b	$8B,$3E,$27,$FC,$5B,$EF,$FF,$8B,$BE,$27,$FC,$5F,$F0,$3F,$E3,$0F
	.dc.b	$89,$FD,$7F,$1F,$E1,$7C,$BF,$E2,$FC,$5F,$88,$E7,$FE,$2D,$5C,$23
	.dc.b	$F3,$B8,$74,$7F,$3F,$97,$81,$FD,$0E,$9E,$47,$F4,$7B,$7A,$1F,$D2
	.dc.b	$F0,$7B,$1F,$D3,$F1,$7F,$90,$B8,$FE,$14,$7E,$9B,$93,$FD,$B3,$CF
	.dc.b	$F3,$23,$EE,$F3,$FC,$DD,$BE,$9F,$47,$D6,$9F,$9D,$18,$AE,$11,$F5
	.dc.b	$AB,$83,$A3,$F3,$5D,$3F,$82,$7F,$72,$EA,$FA,$12,$B9,$F1,$7D,$C8
	.dc.b	$C5,$EE,$73,$87,$86,$3F,$80,$75,$FE,$7B,$1F,$F6,$11,$F8,$93,$FD
	.dc.b	$8D,$7F,$9B,$6D,$FF,$05,$6D,$70,$78,$1F,$AF,$32,$74,$9E,$4E,$48
	.dc.b	$70,$23,$A6,$CF,$77,$71,$FE,$D2,$F0,$7F,$07,$0C,$7F,$0C,$FE,$14
	.dc.b	$E3,$7E,$5A,$E8,$4B,$E9,$3A,$3C,$CE,$23,$53,$C6,$30,$47,$04,$9E
	.dc.b	$2F,$FB,$B7,$27,$47,$A2,$B8,$56,$88,$7F,$BC,$6F,$BB,$83,$E4,$99
	.dc.b	$8E,$2D,$BF,$03,$4E,$07,$77,$D1,$7F,$89,$8F,$E7,$AF,$F3,$F7,$A9
	.dc.b	$0E,$07,$23,$E2,$1D,$48,$6C,$C9,$6B,$D2,$A3,$E0,$BA,$3A,$9C,$8A
	.dc.b	$71,$AF,$6B,$6B,$DA,$D3,$C9,$5F,$6B,$CD,$FF,$6C,$63,$F8,$09,$FC
	.dc.b	$16,$87,$FA,$8F,$73,$9F,$F8,$5C,$F8,$26,$1E,$52,$61,$EA,$7C,$07
	.dc.b	$9C,$CE,$5F,$D9,$73,$F6,$CE,$E7,$9B,$5D,$E7,$EA,$73,$5E,$E6,$C7
	.dc.b	$D6,$7F,$D4,$A7,$F6,$8F,$33,$FD,$EC,$FD,$83,$FB,$19,$E4,$21,$C8
	.dc.b	$F8,$B3,$D5,$F0,$98,$9F,$CA,$6B,$FC,$06,$3F,$8B,$90,$FE,$24,$5D
	.dc.b	$3F,$85,$57,$F0,$88,$FE,$29,$87,$F0,$C7,$FA,$DE,$3F,$B2,$AF,$A4
	.dc.b	$7F,$81,$9F,$E2,$18,$49,$E9,$97,$85,$EE,$4A,$EF,$3F,$CB,$3E,$0F
	.dc.b	$FD,$E2,$49,$FF,$0C,$FC,$1F,$FB,$D3,$E2,$7E,$F4,$7E,$00,$FF,$07
	.dc.b	$F4,$BD,$FC,$E3,$FE,$F5,$87,$C2,$E3,$3F,$C3,$6A,$97,$C0,$E7,$1E
	.dc.b	$7F,$0D,$6B,$5C,$76,$DF,$0F,$7E,$13,$1F,$11,$9F,$86,$D7,$C4,$EF
	.dc.b	$E2,$38,$79,$7F,$DE,$F5,$FB,$FA,$3A,$BE,$FF,$17,$C9,$9F,$CD,$3E
	.dc.b	$9A,$E5,$18,$E1,$F4,$EC,$CB,$C1,$39,$3E,$1E,$9E,$57,$19,$7B,$F6
	.dc.b	$F5,$6F,$EB,$E0,$F6,$70,$FB,$78,$BD,$DC,$7F,$3F,$1F,$A2,$7F,$92
	.dc.b	$B9,$BE,$A3,$FE,$E8,$CB,$F0,$8F,$DB,$DB,$EF,$70,$7D,$C8,$FC,$09
	.dc.b	$F8,$AE,$9F,$C6,$57,$09,$D4,$FF,$9A,$CF,$F7,$06,$9F,$8D,$BF,$AF
	.dc.b	$D9,$FC,$01,$EA,$49,$E9,$D3,$F8,$F7,$C4,$60,$F0,$E1,$FE,$35,$EE
	.dc.b	$FF,$33,$E7,$EA,$CB,$FC,$BA,$FD,$FC,$DD,$5F,$07,$0F,$C7,$61,$F8
	.dc.b	$7D,$7E,$34,$7C,$CE,$4F,$C4,$B7,$87,$E6,$B9,$E2,$FA,$4F,$29,$71
	.dc.b	$FD,$88,$C7,$27,$DE,$75,$CB,$F8,$77,$FD,$20,$7F,$05,$F3,$31,$BF
	.dc.b	$E1,$63,$C5,$D1,$F9,$07,$C9,$D3,$F9,$47,$CD,$D5,$F9,$9D,$3C,$B1
	.dc.b	$E3,$D7,$F0,$2F,$AF,$A3,$F0,$2F,$FA,$23,$AF,$E2,$71,$7C,$7E,$7F
	.dc.b	$F6,$37,$2F,$EC,$B9,$BF,$D6,$FC,$6F,$89,$FD,$31,$31,$F8,$DE,$E7
	.dc.b	$EA,$7E,$07,$F5,$5E,$87,$EA,$FE,$07,$F5,$9E,$97,$EB,$7E,$07,$F5
	.dc.b	$DE,$A7,$EB,$FE,$07,$F6,$5F,$A1,$7F,$00,$CF,$E9,$AF,$83,$CF,$C2
	.dc.b	$EB,$E1,$37,$F0,$BC,$3D,$BF,$8F,$CB,$FC,$03,$CD,$FE,$89,$97,$E3
	.dc.b	$8F,$F1,$69,$FA,$6C,$95,$CA,$31,$87,$67,$59,$79,$1F,$31,$F8,$3D
	.dc.b	$3D,$4F,$E1,$76,$F7,$3F,$87,$E0,$F8,$4F,$C4,$3F,$15,$C7,$F1,$DF
	.dc.b	$D6,$64,$F7,$3E,$09,$23,$CD,$E7,$FC,$4F,$85,$F8,$BE,$FF,$C6,$F8
	.dc.b	$AF,$8F,$F0,$3F,$23,$E2,$BE,$4F,$C1,$FC,$AF,$8A,$F9,$7F,$0B,$F3
	.dc.b	$3C,$AF,$9B,$F0,$DF,$3B,$E2,$BE,$7F,$C4,$7D,$0F,$8A,$FA,$3F,$15
	.dc.b	$FE,$0B,$F3,$DD,$BF,$CA,$B9,$53,$BE,$4F,$95,$87,$37,$E6,$65,$DD
	.dc.b	$F9,$DA,$78,$DF,$A1,$B7,$9D,$FA,$5C,$1E,$B7,$EA,$71,$7B,$DF,$82
	.dc.b	$7D,$7E,$5F,$88,$FF,$13,$F1,$7F,$24,$FC,$C2,$1F,$B1,$F1,$7E,$8B
	.dc.b	$FB,$3E,$9F,$AA,$FE,$D7,$AB,$EC,$BF,$B7,$EB,$FB,$AF,$E4,$0F,$C9
	.dc.b	$F6,$7E,$0D,$FC,$9F,$6F,$E1,$DF,$CA,$F7,$7E,$41,$FF,$09,$FB,$1F
	.dc.b	$C6,$DF,$F3,$E9,$F8,$F8,$E1,$CD,$F3,$23,$FD,$D3,$CA,$BE,$D3,$9F
	.dc.b	$55,$3F,$A1,$AF,$CC,$1E,$69,$FE,$C1,$97,$A9,$FE,$C5,$9F,$B9,$FE
	.dc.b	$C9,$A7,$C2,$7F,$B3,$6B,$F9,$17,$FB,$47,$81,$FB,$DF,$4B,$EF,$70
	.dc.b	$FC,$E9,$F1,$67,$F9,$5C,$FE,$99,$FB,$4F,$03,$F2,$6F,$EA,$FC,$0F
	.dc.b	$C9,$BF,$AD,$F0,$3F,$26,$FE,$BF,$C0,$FC,$9B,$FB,$1F,$03,$F2,$6F
	.dc.b	$EC,$FC,$0F,$C9,$BF,$B5,$F0,$3F,$26,$FF,$6A,$F6,$BF,$B4,$F8,$3F
	.dc.b	$93,$F0,$7F,$25,$3F,$C1,$BC,$1F,$C9,$3F,$C2,$3C,$1F,$C9,$3F,$C2
	.dc.b	$BC,$1F,$C9,$3F,$C3,$3C,$1F,$C9,$3F,$C3,$BC,$1F,$C9,$3F,$C4,$3C
	.dc.b	$1F,$C9,$3F,$C4,$BC,$2F,$89,$9F,$C6,$8F,$F2,$97,$95,$FC,$93,$A7
	.dc.b	$E6,$4F,$F7,$87,$FC,$15,$D5,$F4,$63,$FB,$DF,$5F,$D5,$7F,$BE,$F6
	.dc.b	$7D,$97,$FB,$FF,$6F,$DD,$7F,$61,$8F,$E4,$8F,$E3,$1E,$37,$F2,$17
	.dc.b	$F8,$D7,$8D,$FC,$85,$FE,$3A,$7F,$5A,$F1,$BF,$90,$B7,$C4,$FC,$B7
	.dc.b	$85,$F8,$CF,$13,$F2,$78,$FE,$F7,$C4,$FC,$8B,$FB,$FF,$17,$F9,$06
	.dc.b	$1F,$D8,$7C,$7F,$E6,$EF,$F6,$3F,$23,$F9,$DB,$FD,$97,$C9,$FE,$7E
	.dc.b	$FF,$67,$F2,$BE,$0B,$FB,$4F,$1F,$F1,$EF,$C8,$F2,$7E,$53,$FC,$FB
	.dc.b	$C9,$F9,$AF,$E9,$3A,$7E,$8D,$7E,$4B,$AB,$EA,$BF,$94,$EB,$FB,$2F
	.dc.b	$E5,$BB,$3E,$EB,$F9,$8E,$DF,$C0,$BF,$9A,$EE,$FC,$2B,$F9,$CE,$FF
	.dc.b	$C4,$BF,$9E,$F0,$3F,$90,$71,$FE,$37,$4F,$C9,$1F,$C7,$8F,$CB,$1F
	.dc.b	$C7,$8F,$CD,$1F,$C7,$8F,$CF,$1F,$C7,$8F,$D1,$1F,$C7,$8F,$D3,$1F
	.dc.b	$C7,$8F,$EA,$5E,$47,$F6,$0D,$7E,$0F,$3F,$F0,$0D,$FF,$69,$5F,$E6
	.dc.b	$0F,$03,$F1,$AD,$3F,$A6,$1F,$E0,$FE,$2F,$A2,$FF,$84,$78,$FE,$AB
	.dc.b	$FE,$13,$E4,$FB,$2F,$F8,$57,$97,$EE,$BF,$DB,$B9,$BF,$02,$FF,$14
	.dc.b	$E7,$FC,$2B,$FC,$5B,$A3,$F1,$3D,$5F,$8C,$E2,$FF,$41,$F0,$FF,$9B
	.dc.b	$BD,$7F,$F6,$25,$FF,$1F,$3F,$2A,$7F,$6C,$3F,$32,$7F,$6C,$3F,$8D
	.dc.b	$9F,$8D,$3F,$B6,$1F,$8D,$3F,$B6,$1F,$8D,$3F,$B6,$1F,$8D,$3E,$0C
	.dc.b	$FE,$31,$FF,$18,$6D,$F2,$38,$3E,$54,$7F,$49,$21,$FB,$B3,$E8,$9F
	.dc.b	$4C,$FF,$07,$F9,$3F,$C6,$9F,$F0,$8F,$93,$FC,$69,$FF,$09,$F9,$3F
	.dc.b	$C6,$9F,$F0,$AF,$93,$FC,$69,$FF,$0B,$F9,$3F,$C6,$9F,$F0,$CF,$93
	.dc.b	$FC,$69,$FF,$1E,$E3,$FE,$83,$AC,$67,$FC,$BA,$B8,$3F,$CC,$2F,$A3
	.dc.b	$FC,$CA,$7C,$0F,$F3,$4C,$3C,$8F,$F3,$6C,$7D,$0F,$F3,$8C,$BF,$9A
	.dc.b	$E5,$FC,$DC,$FE,$73,$1F,$CF,$39,$BF,$B5,$F9,$0F,$07,$FD,$4F,$FA
	.dc.b	$1F,$F5,$5B,$FF,$31,$B6,$FF,$AA,$37,$87,$7E,$CF,$FA,$A6,$3F,$03
	.dc.b	$5C,$3D,$EC,$7B,$F9,$F8,$0D,$7C,$56,$FE,$0B,$87,$C5,$71,$F8,$4E
	.dc.b	$5F,$09,$CF,$E1,$7A,$7C,$57,$5F,$5F,$B7,$C5,$7C,$8E,$1F,$F1,$F7
	.dc.b	$A1,$FD,$BF,$B7,$FB,$3F,$95,$AE,$2F,$F5,$1F,$1F,$FB,$05,$B9,$75
	.dc.b	$73,$A7,$85,$DE,$DE,$57,$8F,$07,$A5,$E7,$C5,$EF,$E3,$F8,$1F,$0F
	.dc.b	$F0,$79,$7E,$16,$F2,$FF,$8F,$0D,$9A,$FC,$98,$C3,$63,$33,$CE,$7D
	.dc.b	$37,$DD,$F0,$9E,$CE,$5F,$96,$FB,$B9,$7E,$5B,$F0,$39,$7E,$5B,$F0
	.dc.b	$B9,$7E,$5B,$F1,$39,$7E,$33,$F7,$BE,$23,$F9,$0B,$E4,$73,$CF,$F8
	.dc.b	$09,$E3,$3F,$84,$AE,$AF,$E1,$AF,$C2,$FE,$23,$0F,$2B,$F8,$AC,$7F
	.dc.b	$06,$FD,$4D,$FE,$4B,$F5,$B7,$F9,$0F,$E4,$35,$FD,$D6,$EF,$8A,$FC
	.dc.b	$71,$27,$EE,$FC,$57,$E4,$7D,$DF,$C9,$7B,$DF,$93,$F7,$FF,$DF,$9F
	.dc.b	$A2,$3E,$26,$FC,$16,$3E,$2F,$3F,$09,$AF,$03,$7F,$0B,$C3,$E1,$B8
	.dc.b	$FC,$57,$2F,$88,$E7,$F1,$5F,$91,$E1,$FF,$1F,$9F,$E8,$D3,$FD,$2F
	.dc.b	$E2,$E9,$FF,$79,$EC,$AE,$71,$8C,$3B,$BA,$CB,$C6,$F0,$D3,$CE,$F2
	.dc.b	$DB,$D8,$FE,$73,$83,$E0,$3F,$08,$FC,$FF,$1F,$C5,$7F,$DF,$1C,$9F
	.dc.b	$8D,$27,$2F,$E3,$63,$F7,$BD,$FF,$8D,$7F,$7F,$E0,$FE,$7D,$FE,$03
	.dc.b	$E1,$7E,$8D,$FE,$07,$E1,$FE,$9D,$FE,$0B,$E2,$7E,$AD,$FE,$0F,$E2
	.dc.b	$FE,$DD,$FD,$8F,$8D,$FC,$17,$C5,$FC,$6F,$8D,$FA,$27,$F1,$BA,$7F
	.dc.b	$10,$F1,$7F,$1B,$E3,$7F,$0E,$F1,$BF,$1A,$FF,$10,$F1,$BF,$1A,$FF
	.dc.b	$12,$F1,$BF,$1A,$FF,$14,$F1,$BF,$1B,$A7,$F1,$13,$F5,$27,$E3,$4F
	.dc.b	$E0,$78,$FE,$3B,$C8,$FE,$41,$1F,$B2,$8F,$C8,$D7,$F1,$E7,$E3,$3F
	.dc.b	$C8,$38,$FE,$33,$FC,$8B,$8F,$E3,$3F,$C9,$38,$FE,$33,$FC,$9B,$8F
	.dc.b	$E3,$3F,$CA,$38,$FE,$33,$FC,$AB,$8F,$E3,$65,$FC,$BB,$8F,$F1,$9C
	.dc.b	$6F,$4B,$FB,$76,$BF,$C7,$AB,$FD,$B1,$DD,$F2,$1F,$E1,$3C,$9F,$22
	.dc.b	$7F,$9C,$F2,$7C,$87,$F9,$DF,$27,$C8,$7F,$9E,$F2,$7C,$87,$F9,$FF
	.dc.b	$27,$C8,$DB,$F9,$EF,$91,$FB,$4C,$BF,$A0,$F9,$1F,$12,$FF,$A2,$9F
	.dc.b	$B4,$31,$1F,$24,$F9,$C7,$E3,$0F,$DF,$1F,$C6,$0F,$E0,$47,$F2,$C3
	.dc.b	$F8,$C1,$FC,$C0,$FE,$30,$7F,$34,$3F,$8C,$1F,$CE,$0F,$E3,$07,$E3
	.dc.b	$0F,$E7,$BC,$FF,$8C,$E8,$3F,$17,$83,$F9,$7E,$DF,$16,$BF,$97,$ED
	.dc.b	$F1,$5F,$E6,$3B,$7C,$57,$F9,$9E,$DF,$15,$FE,$6B,$B7,$C5,$CF,$FB
	.dc.b	$2E,$BF,$17,$6F,$EC,$7B,$7C,$2D,$3F,$9C,$E1,$FD,$01,$FE,$7F,$D3
	.dc.b	$FB,$3A,$FE,$D2,$FF,$67,$F2,$3F,$73,$5F,$DB,$BA,$FF,$7A,$AB,$FD
	.dc.b	$85,$E0,$FF,$19,$8F,$E1,$7E,$37,$F1,$97,$F8,$6F,$8D,$FC,$65,$FE
	.dc.b	$1F,$E3,$7F,$19,$7F,$88,$F8,$DF,$C6,$5F,$E2,$7E,$37,$E3,$BA,$7F
	.dc.b	$1B,$1F,$C6,$8F,$E3,$26,$8F,$E3,$3E,$37,$F1,$87,$C9,$C4,$7F,$AE
	.dc.b	$66,$F1,$BF,$8D,$C6,$8C,$F8,$DF,$C7,$1C,$71,$29,$E3,$7F,$1E,$9C
	.dc.b	$C7,$BD,$FF,$08,$F8,$FF,$CF,$1F,$EF,$7E,$47,$C9,$7F,$8F,$90,$FE
	.dc.b	$3F,$97,$D3,$EE,$FE,$41,$CF,$F5,$E3,$EC,$CF,$E4,$F9,$3F,$84,$3F
	.dc.b	$93,$3F,$AC,$1F,$93,$3F,$AE,$1F,$93,$3F,$B0,$1F,$93,$3F,$C4,$BD
	.dc.b	$BF,$8F,$BF,$C8,$C7,$FB,$5F,$CE,$FE,$40,$FE,$5C,$D9,$C6,$BF,$8F
	.dc.b	$3D,$5F,$E3,$CF,$85,$FE,$3C,$F9,$5F,$3B,$FC,$D8,$FD,$69,$FA,$F3
	.dc.b	$F6,$E7,$F1,$E2,$1F,$D0,$BB,$FC,$6F,$F7,$F3,$F5,$24,$30,$48,$FF
	.dc.b	$1F,$C7,$06,$C6,$DA,$EF,$0F,$F2,$4C,$ED,$B1,$B6,$BF,$ED,$33,$FB
	.dc.b	$44,$F2,$24,$69,$B1,$FE,$13,$39,$14,$F0,$15,$DE,$DF,$D3,$63,$6D
	.dc.b	$76,$D8,$DE,$1F,$62,$7E,$31,$C1,$C4,$6A,$FD,$93,$DE,$FC,$C4,$70
	.dc.b	$4B,$ED,$19,$6C,$66,$FA,$CE,$5B,$1F,$E4,$7F,$88,$63,$FA,$F8,$FF
	.dc.b	$44,$CE,$AF,$99,$1C,$13,$0F,$3C,$65,$B1,$99,$F7,$46,$0F,$C2,$F9
	.dc.b	$3F,$D9,$2B,$F6,$DF,$15,$FE,$2A,$25,$7F,$87,$3C,$CF,$F1,$6F,$A7
	.dc.b	$FE,$2E,$F8,$9F,$F1,$7F,$B9,$FE,$30,$F8,$9F,$F1,$8F,$BB,$FE,$32
	.dc.b	$F8,$9F,$F1,$9F,$BD,$FE,$34,$F8,$9F,$F1,$AF,$BF,$ED,$E5,$72,$FF
	.dc.b	$0D,$E0,$7B,$3F,$07,$83,$F8,$4A,$F7,$F8,$56,$E1,$F1,$38,$AF,$F3
	.dc.b	$84,$3E,$03,$F9,$07,$81,$FE,$87,$F8,$A7,$A5,$BC,$FB,$D5,$BE,$78
	.dc.b	$7F,$1F,$D3,$CA,$F4,$D3,$4F,$A8,$FF,$1F,$ED,$EE,$7F,$C8,$3C,$1F
	.dc.b	$09,$FF,$21,$F1,$7C,$67,$FC,$8B,$C9,$F2,$9F,$98,$7F,$92,$39,$FE
	.dc.b	$7B,$F9,$43,$F3,$3D,$5F,$65,$BD,$5F,$65,$FB,$64,$32,$5C,$FF,$34
	.dc.b	$FF,$15,$31,$C7,$F9,$19,$E5,$1C,$09,$3C,$1B,$3B,$7F,$84,$75,$7E
	.dc.b	$1E,$FF,$55,$1F,$97,$E0,$FD,$83,$63,$2F,$F0,$23,$18,$7D,$D7,$04
	.dc.b	$3F,$7C,$68,$D4,$F4,$74,$D7,$66,$27,$BB,$82,$46,$CF,$F1,$77,$1F
	.dc.b	$FB,$26,$77,$19,$69,$2B,$CD,$5C,$A3,$0F,$91,$AF,$E5,$E7,$94,$F0
	.dc.b	$6C,$6E,$31,$3C,$3D,$97,$AD,$E9,$9C,$92,$B8,$10,$ED,$5E,$07,$E3
	.dc.b	$18,$8E,$0D,$38,$1A,$21,$DC,$C9,$FB,$1E,$EF,$E2,$2F,$E8,$73,$C7
	.dc.b	$77,$F4,$A8,$E2,$7C,$3A,$EE,$E3,$2F,$7B,$E1,$21,$9C,$3D,$0F,$84
	.dc.b	$C9,$FD,$1B,$2F,$86,$F5,$3C,$A5,$36,$53,$A5,$78,$9E,$86,$0E,$35
	.dc.b	$E4,$7F,$2D,$E3,$FF,$88,$A3,$F0,$71,$FD,$8C,$E0,$76,$78,$3D,$08
	.dc.b	$78,$4A,$E9,$FE,$AD,$5E,$A8,$C1,$2F,$E0,$CF,$12,$46,$6F,$D5,$18
	.dc.b	$9D,$CB,$2F,$AB,$3F,$D1,$2B,$73,$C6,$3D,$F5,$C0,$93,$A6,$C6,$A3
	.dc.b	$FA,$75,$F9,$E7,$F8,$76,$DF,$C0,$49,$9F,$F0,$1B,$FE,$06,$7F,$5B
	.dc.b	$BF,$E1,$33,$A8,$C1,$FC,$3C,$DE,$1F,$D4,$DF,$E2,$26,$8C,$DF,$82
	.dc.b	$FB,$57,$3A,$FE,$F5,$B7,$F5,$5E,$57,$C5,$7F,$37,$26,$3F,$97,$3F
	.dc.b	$A0,$FC,$17,$F3,$BF,$77,$F9,$E7,$C1,$FF,$3D,$F8,$4F,$E7,$DF,$15
	.dc.b	$FC,$FF,$E0,$7F,$A0,$7C,$57,$F4,$1F,$83,$FE,$85,$F1,$5F,$D0,$F5
	.dc.b	$FE,$89,$F1,$5F,$D1,$7A,$3F,$A3,$7C,$57,$F4,$7F,$87,$FE,$91,$F1
	.dc.b	$5F,$ED,$EB,$39,$E1,$F2,$72,$7F,$5C,$BE,$51,$8A,$63,$DE,$31,$97
	.dc.b	$8D,$D6,$9E,$77,$86,$DE,$B7,$97,$07,$BD,$E9,$C5,$F1,$1F,$ED,$3C
	.dc.b	$7F,$E0,$BE,$4F,$E7,$1D,$1F,$63,$9F,$E7,$92,$BE,$A7,$3F,$D1,$E8
	.dc.b	$FA,$91,$89,$D7,$0F,$F8,$4B,$AF,$F5,$47,$EA,$EB,$71,$F7,$EB,$EE
	.dc.b	$E1,$DD,$E4,$48,$D6,$1F,$D3,$F1,$FB,$D9,$7A,$63,$FB,$8C,$FF,$81
	.dc.b	$23,$FB,$B6,$7E,$FB,$FF,$02,$E9,$F0,$DF,$F0,$3E,$BF,$1D,$FF,$04
	.dc.b	$ED,$F2,$DF,$F0,$5E,$FF,$3D,$FF,$06,$F0,$7D,$47,$FC,$1F,$C5,$F6
	.dc.b	$5F,$F0,$8F,$1F,$DD,$31,$AF,$93,$8D,$5C,$63,$17,$D5,$D6,$1E,$17
	.dc.b	$86,$3E,$57,$96,$5E,$97,$A6,$7E,$D7,$B6,$9F,$0F,$8F,$FC,$41,$B7
	.dc.b	$DB,$F3,$7F,$C4,$5F,$13,$FE,$19,$27,$07,$E3,$7C,$7F,$F0,$F7,$E8
	.dc.b	$BF,$E1,$A8,$FF,$59,$AF,$47,$3D,$FA,$F8,$0D,$FC,$4F,$0F,$82,$E3
	.dc.b	$F1,$3C,$BE,$13,$9F,$C4,$F4,$F8,$6E,$BE,$47,$6F,$87,$EF,$F1,$3F
	.dc.b	$A4,$DF,$F8,$17,$37,$C7,$EE,$CC,$7E,$93,$85,$7C,$A3,$18,$F6,$75
	.dc.b	$9F,$89,$E1,$AF,$99,$E5,$BF,$A9,$E9,$C3,$F0,$39,$BF,$C7,$9C,$7F
	.dc.b	$12,$7F,$C9,$7C,$9F,$25,$77,$B5,$7B,$5F,$20,$93,$E7,$F8,$9F,$97
	.dc.b	$E1,$FC,$CF,$89,$F9,$BF,$13,$F3,$BD,$EF,$9F,$F1,$3F,$43,$E0,$3F
	.dc.b	$C9,$DF,$A3,$7F,$B7,$7A,$3F,$36,$BE,$2B,$3F,$09,$AF,$8A,$DF,$AD
	.dc.b	$C3,$E2,$B8,$FC,$3F,$CB,$E1,$FF,$75,$FB,$1F,$D0,$B3,$FD,$2B,$7A
	.dc.b	$3F,$CC,$9D,$FC,$2B,$F8,$8E,$AC,$7A,$C6,$33,$F0,$BA,$D7,$CF,$F1
	.dc.b	$7E,$98,$FA,$DC,$1E,$D7,$EC,$71,$7C,$17,$ED,$72,$7E,$3B,$97,$E3
	.dc.b	$C7,$E3,$B9,$BE,$77,$47,$F9,$F3,$A7,$E9,$71,$7E,$A5,$FA,$B5,$F8
	.dc.b	$3E,$1F,$CB,$F5,$FD,$B8,$FC,$BF,$67,$DF,$7F,$33,$DB,$F8,$37,$F3
	.dc.b	$7D,$DF,$87,$7F,$3B,$DF,$F8,$B7,$F3,$FE,$07,$E4,$1F,$F6,$1F,$A7
	.dc.b	$FC,$63,$AF,$FD,$4F,$E0,$3C,$2F,$CD,$57,$E5,$7C,$3F,$CF,$3F,$97
	.dc.b	$F1,$3F,$44,$FE,$67,$C5,$FD,$33,$F9,$BF,$1B,$F5,$4F,$E7,$7C,$7F
	.dc.b	$5B,$ED,$3F,$68,$7F,$41,$E3,$FC,$A7,$A5,$FE,$B1,$CF,$F2,$33,$FA
	.dc.b	$5E,$7F,$97,$CD,$FC,$8A,$5D,$1F,$92,$AF,$D5,$78,$FF,$92,$7F,$59
	.dc.b	$E3,$FE,$49,$FD,$77,$8F,$F7,$1F,$BC,$7E,$13,$C7,$FC,$83,$F8,$6E
	.dc.b	$3F,$C8,$3F,$88,$F1,$FF,$21,$9F,$E3,$BD,$5F,$F5,$47,$91,$F8,$FA
	.dc.b	$FF,$57,$F9,$1F,$8F,$55,$F9,$EF,$1F,$F1,$F1,$8F,$03,$FD,$7D,$E4
	.dc.b	$3C,$8F,$C8,$CE,$4F,$D9,$C7,$E9,$5F,$CC,$F8,$FF,$93,$7F,$8C,$F8
	.dc.b	$FF,$93,$7F,$8D,$F8,$FF,$93,$7F,$8E,$F8,$FF,$93,$7E,$51,$07,$47
	.dc.b	$F8,$2F,$A7,$F2,$79,$FF,$2A,$9F,$AA,$7F,$26,$7E,$CB,$FC,$A3,$AB
	.dc.b	$EE,$BF,$CA,$BA,$FF,$02,$FF,$2C,$EC,$FC,$2B,$FC,$BB,$B7,$F1,$2F
	.dc.b	$F3,$0E,$EF,$C6,$BF,$CC,$BB,$FE,$6E,$3F,$E6,$E8,$FF,$03,$F2,$FC
	.dc.b	$DF,$1F,$FD,$C7,$CD,$F9,$9E,$6F,$C1,$78,$5F,$9D,$7F,$09,$E1,$FE
	.dc.b	$85,$FC,$37,$89,$FA,$57,$F1,$1E,$2F,$EA,$5F,$C5,$78,$DF,$AD,$7F
	.dc.b	$19,$E3,$FF,$27,$D3,$E0,$C7,$ED,$5F,$CB,$03,$E4,$F9,$5F,$EF,$7F
	.dc.b	$1B,$F2,$F1,$FB,$F3,$E7,$4F,$E5,$FC,$6F,$CB,$47,$F5,$0F,$1B,$F2
	.dc.b	$CF,$F5,$2F,$1B,$F2,$CF,$F5,$4F,$1B,$F2,$CF,$E0,$F8,$5E,$07,$E0
	.dc.b	$E3,$0B,$C6,$FC,$CC,$63,$C6,$FC,$CB,$AF,$1B,$E6,$BF,$CF,$B9,$BE
	.dc.b	$61,$F3,$7D,$3F,$F1,$BC,$7E,$7C,$FE,$4F,$E3,$7E,$63,$7F,$E6,$FE
	.dc.b	$37,$E5,$9F,$E7,$3E,$57,$99,$FD,$A6,$DE,$A7,$F6,$DB,$FB,$9F,$DC
	.dc.b	$70,$7C,$27,$E2,$63,$F1,$BA,$FE,$57,$93,$FC,$9B,$47,$87,$F9,$4A
	.dc.b	$FE,$AB,$18,$7E,$AD,$7D,$13,$F2,$B7,$FC,$17,$C3,$FC,$AB,$FC,$1F
	.dc.b	$C3,$FC,$AB,$FC,$27,$C3,$FC,$AB,$FC,$2F,$C3,$FC,$AB,$FC,$37,$C3
	.dc.b	$FC,$AB,$FC,$3F,$C3,$FC,$06,$5F,$CA,$39,$BF,$DD,$1E,$2F,$E4,$F4
	.dc.b	$FC,$17,$87,$F9,$37,$F0,$9E,$1F,$E4,$DF,$C3,$78,$7F,$93,$7F,$11
	.dc.b	$E1,$FE,$4D,$FC,$57,$87,$F9,$37,$F1,$9E,$1F,$E4,$F3,$FC,$17,$1B
	.dc.b	$CB,$F8,$9E,$0F,$F4,$B9,$FC,$A7,$8D,$FC,$A7,$A3,$F9,$3E,$3F,$32
	.dc.b	$7F,$BC,$1F,$D5,$8F,$E4,$E7,$F5,$A3,$F9,$39,$FD,$78,$FE,$4E,$7F
	.dc.b	$10,$3F,$93,$9F,$90,$F0,$7F,$1F,$C7,$FD,$EB,$8B,$F1,$EF,$F7,$CF
	.dc.b	$07,$F1,$F9,$7F,$AD,$FC,$AF,$E4,$E6,$3C,$34,$F0,$CB,$F9,$1E,$1D
	.dc.b	$1F,$E4,$B8,$FF,$82,$3C,$0F,$C7,$47,$EB,$FC,$0F,$C7,$3F,$B1,$F0
	.dc.b	$3F,$1C,$FE,$CF,$C0,$FC,$73,$FB,$5F,$13,$E1,$EF,$FE,$1E,$F1,$7F
	.dc.b	$6F,$C4,$F1,$BF,$99,$FC,$5F,$F8,$58,$95,$FE,$08,$F8,$AF,$F0,$CF
	.dc.b	$C5,$FF,$86,$BD,$DF,$F0,$DD,$FF,$87,$3E,$2B,$FC,$3B,$EF,$FF,$87
	.dc.b	$BE,$2B,$FC,$3F,$F0,$3F,$E2,$0F,$8A,$FF,$10,$FC,$1F,$F8,$8B,$E2
	.dc.b	$BF,$C4,$7F,$0B,$FE,$24,$F8,$AF,$C6,$F7,$FF,$9F,$BF,$47,$FC,$7F
	.dc.b	$0F,$F0,$15,$F1,$1F,$2A,$3F,$92,$F8,$1C,$3A,$3F,$8A,$78,$2B,$DD
	.dc.b	$7F,$AC,$1E,$78,$7F,$8C,$31,$F0,$CF,$F8,$93,$3F,$2B,$FE,$26,$D7
	.dc.b	$D2,$FF,$8A,$37,$F6,$BF,$E2,$AE,$1F,$82,$FF,$8B,$38,$FE,$2B,$F1
	.dc.b	$CF,$F1,$7F,$37,$CA,$7F,$93,$9F,$D7,$1C,$65,$FE,$42,$7F,$C8,$BC
	.dc.b	$DC,$A7,$F8,$ED,$76,$7F,$8F,$DF,$89,$FE,$43,$87,$99,$FE,$47,$8F
	.dc.b	$A9,$FE,$4B,$97,$B9,$FE,$4F,$9F,$C7,$7F,$CB,$FA,$7C,$4D,$BE,$4B
	.dc.b	$F2,$23,$FC,$9B,$BF,$CF,$7E,$97,$07,$1E,$0F,$9D,$1F,$64,$FF,$26
	.dc.b	$F0,$FD,$D7,$FC,$9F,$C5,$F8,$17,$FC,$A3,$C7,$F8,$57,$FC,$A7,$C9
	.dc.b	$F8,$97,$FC,$AB,$CB,$F8,$D7,$FC,$AF,$CD,$F1,$1F,$96,$43,$E5,$E3
	.dc.b	$F7,$FE,$2F,$C9,$A9,$A6,$3E,$2B,$3E,$EE,$BD,$ED,$FC,$57,$0F,$80
	.dc.b	$E3,$F1,$5C,$BE,$0B,$9F,$C5,$74,$F8,$4E,$BF,$15,$DB,$E1,$BB,$FA
	.dc.b	$1E,$0F,$87,$F0,$F5,$7F,$00,$F8,$7F,$F2,$67,$27,$FB,$F7,$E1,$32
	.dc.b	$FE,$4E,$B8,$BF,$95,$C3,$AB,$F9,$7C,$BC,$2F,$E6,$74,$F2,$BF,$9B
	.dc.b	$DB,$D2,$FE,$77,$83,$DA,$FE,$7F,$8B,$E0,$BF,$A1,$E3,$FF,$38,$FA
	.dc.b	$2F,$1B,$F2,$3D,$FF,$C0,$2B,$F3,$07,$F6,$CE,$9F,$A4,$FF,$6E,$EA
	.dc.b	$FA,$CF,$F7,$0E,$BF,$B4,$FF,$72,$EC,$FB,$CF,$F7,$4E,$DF,$C1,$3F
	.dc.b	$DD,$BB,$BF,$0C,$FE,$EC,$FA,$FD,$FF,$8D,$F1,$FF,$C9,$91,$FB,$78
	.dc.b	$FB,$6F,$08,$FF,$53,$CF,$47,$FD,$53,$5E,$07,$FD,$57,$7E,$47,$FD
	.dc.b	$5B,$87,$A1,$FF,$57,$E3,$EC,$7F,$D6,$39,$7E,$37,$1F,$F0,$5E,$9F
	.dc.b	$B7,$F2,$3F,$92,$47,$ED,$CF,$C8,$57,$ED,$DF,$8E,$F2,$7F,$81,$1F
	.dc.b	$9A,$3F,$83,$1F,$9E,$3F,$85,$1F,$A2,$3F,$87,$1F,$A6,$3F,$89,$1F
	.dc.b	$AA,$3F,$D6,$FE,$17,$C2,$7E,$21,$FE,$EA,$DB,$F6,$FB,$66,$3F,$7A
	.dc.b	$71,$3F,$80,$9D,$4F,$E0,$A7,$84,$FE,$12,$79,$4F,$E1,$A7,$A4,$FE
	.dc.b	$22,$7E,$C8,$F7,$C7,$E1,$DF,$8E,$7F,$70,$CB,$17,$F1,$63,$27,$F8
	.dc.b	$DB,$C8,$FE,$39,$1F,$EE,$0F,$23,$F8,$E3,$FE,$E2,$F2,$3F,$8E,$3F
	.dc.b	$EE,$4F,$23,$F8,$E3,$FE,$E6,$F2,$3F,$8E,$3F,$EE,$8F,$23,$F8,$E3
	.dc.b	$FC,$EF,$1F,$E7,$C7,$E3,$0F,$8C,$7F,$1E,$D3,$E3,$46,$E3,$F2,$87
	.dc.b	$D0,$3F,$30,$7D,$43,$F3,$87,$D8,$3F,$40,$7D,$C3,$F4,$87,$E0,$0F
	.dc.b	$D4,$1F,$18,$FC,$41,$F0,$B6,$FC,$5F,$91,$FC,$5E,$3F,$68,$43,$FA
	.dc.b	$F7,$47,$EE,$9F,$EC,$1D,$3F,$BE,$7F,$B1,$75,$7F,$02,$7F,$B2,$75
	.dc.b	$FF,$06,$7F,$B3,$76,$7F,$0A,$7F,$B4,$76,$FF,$0E,$7F,$B5,$77,$7F
	.dc.b	$12,$7F,$C5,$7D,$FF,$CE,$FC,$0F,$E2,$F1,$83,$FB,$47,$83,$FC,$75
	.dc.b	$FE,$D5,$E1,$7F,$21,$7F,$B6,$78,$7F,$C9,$5F,$ED,$DE,$27,$F2,$97
	.dc.b	$FB,$87,$8B,$FC,$B5,$FE,$E5,$E3,$7F,$31,$7F,$BA,$78,$FF,$CD,$5F
	.dc.b	$85,$CB,$FE,$01,$F1,$FF,$8B,$F8,$3F,$C5,$A7,$E5,$CB,$C0,$FF,$64
	.dc.b	$F7,$FE,$31,$FF,$65,$F7,$FE,$31,$FF,$66,$F7,$FE,$31,$FF,$67,$F7
	.dc.b	$FE,$31,$FF,$68,$F7,$FE,$31,$FF,$69,$F7,$FE,$31,$FF,$6A,$F8,$3F
	.dc.b	$88,$7F,$5F,$9F,$E3,$3B,$BF,$C1,$DE,$06,$63,$FD,$AF,$5C,$5F,$F6
	.dc.b	$CD,$F5,$7F,$DB,$78,$78,$5F,$F6,$EE,$3E,$57,$FD,$BF,$97,$A5,$FF
	.dc.b	$70,$E7,$ED,$7F,$DC,$7A,$7C,$3C,$FF,$D4,$7B,$7F,$9D,$3C,$0F,$C6
	.dc.b	$F9,$0F,$88,$FE,$C6,$4C,$3F,$C4,$4D,$7F,$C5,$8D,$E1,$FF,$18,$37
	.dc.b	$87,$FC,$06,$D9,$FF,$03,$B7,$1F,$F1,$73,$6F,$C1,$3D,$8F,$E2,$E7
	.dc.b	$F8,$28,$FF,$1B,$98,$BE,$EE,$CF,$19,$5C,$B8,$23,$82,$5F,$9A,$38
	.dc.b	$35,$DB,$63,$71,$EA,$3F,$8C,$14,$D9,$5D,$DF,$BA,$3A,$99,$9E,$84
	.dc.b	$70,$48,$E0,$F4,$6C,$F4,$AF,$8F,$39,$5F,$15,$FD,$8E,$FF,$CB,$7F
	.dc.b	$0D,$FF,$00,$67,$35,$E2,$C3,$9D,$78,$5B,$1E,$19,$EB,$87,$7C,$7E
	.dc.b	$B1,$27,$94,$EA,$F9,$3E,$4C,$BF,$17,$DB,$FE,$2C,$25,$7F,$9A,$FE
	.dc.b	$13,$FC,$5F,$EB,$FF,$90,$F8,$BF,$69,$23,$FD,$F0,$BD,$8D,$7C,$26
	.dc.b	$FD,$AE,$1F,$09,$C7,$DC,$E5,$F0,$9C,$FD,$EE,$9F,$09,$D7,$E0,$3B
	.dc.b	$7C,$27,$7F,$82,$FE,$7D,$B3,$9B,$C3,$CE,$F2,$BF,$8D,$C7,$FA,$2F
	.dc.b	$C2,$E1,$5E,$58,$DF,$13,$1E,$D3,$9E,$8F,$F4,$5C,$B4,$F2,$4E,$46
	.dc.b	$49,$6D,$EB,$9D,$4E,$78,$3E,$04,$62,$9C,$5F,$27,$8F,$E5,$65,$F3
	.dc.b	$2B,$E7,$1F,$E9,$9E,$6F,$A2,$FF,$98,$3A,$3E,$C3,$F7,$FA,$7E,$E3
	.dc.b	$F8,$B6,$EF,$F8,$18,$FB,$F1,$8B,$FB,$EF,$F9,$A3,$AF,$F2,$13,$EC
	.dc.b	$EB,$FC,$84,$6E,$3B,$1C,$88,$77,$2D,$F7,$70,$79,$EB,$FD,$76,$FE
	.dc.b	$D4,$FC,$31,$FB,$12,$ED,$FA,$86,$E3,$FA,$66,$CF,$27,$81,$CC,$86
	.dc.b	$09,$3C,$DC,$B5,$CC,$F5,$72,$FE,$B9,$FE,$30,$4A,$FF,$6A,$5F,$F2
	.dc.b	$98,$F1,$9E,$CA,$E8,$4A,$E9,$1D,$AF,$D1,$39,$6C,$EA,$31,$3C,$E3
	.dc.b	$04,$9D,$9F,$23,$4F,$CC,$4F,$53,$84,$69,$A4,$3F,$DD,$75,$D9,$B5
	.dc.b	$D5,$D1,$EC,$33,$3C,$DA,$68,$F8,$F7,$F8,$8A,$63,$F8,$89,$FE,$78
	.dc.b	$73,$21,$F3,$AB,$A3,$83,$D4,$76,$21,$99,$F4,$36,$FF,$94,$E1,$F0
	.dc.b	$A7,$33,$D4,$F3,$1D,$0E,$15,$E0,$74,$76,$29,$A3,$E9,$71,$7F,$89
	.dc.b	$7E,$02,$47,$F5,$9D,$36,$C3,$B1,$6F,$C6,$E0,$8F,$D1,$C3,$D1,$8F
	.dc.b	$D1,$9F,$11,$A7,$0F,$8C,$CC,$F4,$24,$6A,$3E,$99,$FE,$15,$7F,$35
	.dc.b	$A7,$F6,$A8,$E4,$43,$CD,$1C,$5F,$06,$5F,$15,$A7,$F3,$BB,$F3,$C7
	.dc.b	$F0,$3B,$FE,$9E,$4A,$FE,$26,$D8,$D4,$60,$8E,$0F,$F0,$46,$9F,$89
	.dc.b	$D3,$F1,$18,$7F,$08,$C7,$F8,$51,$32,$FE,$2F,$1F,$C3,$9A,$ED,$BD
	.dc.b	$9F,$E6,$EE,$07,$C2,$7F,$55,$25,$7F,$31,$F8,$3F,$EB,$1E,$EF,$F5
	.dc.b	$9F,$84,$FE,$B5,$EF,$FF,$5B,$F8,$7F,$EB,$9F,$03,$FD,$77,$E2,$3F
	.dc.b	$AF,$7C,$1F,$F1,$8D,$FF,$A7,$FE,$90,$F8,$AC,$70,$FF,$61,$E6,$4B
	.dc.b	$E1,$B3,$C1,$FD,$22,$1F,$0F,$8A,$DF,$C4,$D9,$75,$DF,$8A,$FE,$6D
	.dc.b	$87,$DD,$E9,$C3,$FD,$3A,$B6,$DB,$E5,$8F,$3C,$BA,$69,$EA,$AF,$EA
	.dc.b	$5B,$FB,$9F,$EA,$9C,$3F,$09,$FE,$AD,$C7,$F1,$9F,$90,$7F,$59,$E6
	.dc.b	$F9,$6F,$F3,$D3,$EA,$1F,$D1,$8F,$A8,$5E,$1F,$B4,$DE,$1F,$B3,$18
	.dc.b	$E1,$FF,$35,$CF,$F6,$8F,$13,$FB,$CD,$7F,$9C,$23,$9C,$FE,$35,$EE
	.dc.b	$D7,$3D,$9F,$8E,$72,$D5,$D5,$F8,$5C,$7D,$38,$FF,$80,$9F,$F0,$D7
	.dc.b	$4F,$BE,$BF,$C3,$39,$FC,$37,$FC,$37,$A7,$C7,$7F,$C3,$BA,$FC,$B7
	.dc.b	$FC,$25,$A7,$E0,$78,$BF,$09,$1F,$57,$83,$E9,$BF,$54,$FB,$1C,$BF
	.dc.b	$69,$E2,$7F,$83,$36,$FD,$CD,$FF,$90,$BD,$AF,$E5,$D1,$89,$FC,$57
	.dc.b	$37,$E1,$EF,$F1,$1C,$BD,$A7,$9F,$17,$E4,$B2,$FC,$0C,$7F,$84,$79
	.dc.b	$FF,$38,$FF,$84,$FA,$3F,$40,$FF,$85,$7A,$7F,$48,$FF,$85,$FA,$BE
	.dc.b	$2B,$F8,$12,$1F,$33,$E0,$BE,$07,$C5,$7E,$64,$91,$F9,$5F,$3F,$F3
	.dc.b	$9E,$87,$E7,$7E,$13,$F3,$DE,$97,$E7,$F0,$FD,$07,$A7,$FA,$1F,$85
	.dc.b	$FD,$17,$87,$FA,$3F,$86,$FD,$27,$AD,$FA,$5F,$87,$FD,$37,$B1,$FA
	.dc.b	$7F,$88,$FD,$47,$B5,$FA,$6E,$0F,$EF,$1D,$BF,$25,$FF,$19,$F6,$7C
	.dc.b	$9E,$9F,$DA,$D7,$19,$FD,$56,$1D,$5F,$D6,$65,$E1,$7F,$5D,$A7,$95
	.dc.b	$FD,$86,$DE,$97,$F6,$5C,$1E,$D7,$F6,$9C,$5F,$BB,$E3,$F8,$71,$F5
	.dc.b	$3B,$9C,$BF,$EB,$9E,$4F,$E1,$1F,$A4,$FF,$89,$2B,$EC,$B7,$C5,$FF
	.dc.b	$2E,$B7,$8B,$7E,$77,$F0,$98,$7B,$97,$E2,$71,$EF,$67,$E2,$75,$F0
	.dc.b	$1B,$F8,$9E,$1F,$05,$C7,$E2,$79,$7C,$27,$3F,$89,$E9,$F0,$DD,$7E
	.dc.b	$27,$B7,$C4,$77,$F8,$9F,$E7,$1E,$7D,$E5,$FF,$2F,$F1,$2B,$84,$63
	.dc.b	$0E,$8E,$B2,$F0,$3C,$34,$F2,$3C,$B6,$F4,$3D,$38,$3D,$8F,$6E,$2F
	.dc.b	$81,$C7,$F0,$62,$1F,$EB,$CF,$07,$F9,$8F,$47,$D0,$E7,$FE,$6E,$FD
	.dc.b	$39,$FA,$A7,$F2,$9F,$1F,$F2,$6F,$F2,$BE,$9F,$C9,$BF,$CB,$7C,$7F
	.dc.b	$C9,$BF,$CB,$FC,$7F,$C9,$BF,$CC,$7C,$7F,$C9,$BF,$CC,$FC,$7F,$C9
	.dc.b	$BF,$90,$F6,$7F,$C1,$2F,$ED,$3D,$7F,$E6,$EF,$ED,$BC,$9F,$CA,$78
	.dc.b	$2F,$27,$F2,$93,$F8,$7F,$27,$F2,$8F,$E2,$7C,$9F,$CA,$3F,$8B,$F2
	.dc.b	$7F,$28,$FE,$37,$C9,$FC,$A3,$F8,$FF,$27,$F2,$8F,$E4,$7C,$9F,$84
	.dc.b	$FE,$4F,$93,$F2,$BE,$BB,$93,$F2,$73,$F3,$0F,$C9,$9F,$40,$FA,$47
	.dc.b	$E0,$3C,$8F,$C7,$BF,$82,$F2,$3F,$1E,$FE,$13,$C8,$FC,$7B,$F8,$6F
	.dc.b	$23,$F1,$EF,$E2,$3C,$8F,$C7,$BF,$8A,$F2,$3F,$1E,$FE,$33,$C8,$FC
	.dc.b	$7B,$4F,$C7,$F8,$DF,$8E,$7F,$4F,$E3,$7E,$39,$FD,$4F,$8D,$F8,$E7
	.dc.b	$F5,$7E,$37,$E3,$9F,$D6,$F8,$DF,$8E,$7F,$5F,$E3,$7E,$39,$FD,$8F
	.dc.b	$8D,$EF,$7E,$09,$F9,$E3,$F7,$07,$F7,$BE,$3F,$C7,$79,$3F,$C8,$A3
	.dc.b	$F9,$01,$F8,$E3,$F8,$07,$89,$F8,$E7,$F8,$17,$27,$E3,$9F,$E0,$9E
	.dc.b	$27,$E3,$9F,$E0,$DE,$27,$E3,$9F,$E1,$1E,$27,$E3,$9F,$E1,$5E,$27
	.dc.b	$E3,$9F,$C5,$78,$9F,$19,$A7,$F8,$17,$C1,$FC,$5F,$8B,$FE,$18,$F0
	.dc.b	$BF,$16,$FF,$86,$BC,$2F,$C5,$BF,$E1,$CF,$0B,$F1,$6F,$F8,$7B,$C2
	.dc.b	$FC,$5B,$FE,$20,$F0,$BF,$16,$FF,$88,$BC,$2F,$C5,$BF,$AE,$CB,$FC
	.dc.b	$29,$C7,$F1,$BC,$67,$91,$FC,$66,$7F,$DD,$2D,$8C,$C7,$CF,$AF,$E3
	.dc.b	$FC,$DF,$4D,$FE,$43,$CF,$F5,$DF,$E4,$7D,$1F,$6D,$FE,$4B,$D3,$F7
	.dc.b	$DF,$E4,$FD,$5F,$83,$7F,$94,$F5,$FE,$1D,$FE,$57,$D9,$FC,$67,$2F
	.dc.b	$E8,$E7,$FB,$87,$C9,$C4,$EA,$3F,$28,$7F,$1B,$3F,$30,$7F,$1B,$3F
	.dc.b	$38,$7F,$1B,$3F,$40,$7F,$1B,$3F,$48,$7F,$1B,$3F,$50,$7E,$A8,$FE
	.dc.b	$7F,$E4,$FE,$BE,$3F,$9F,$C7,$C9,$F1,$FF,$A3,$BF,$30,$87,$F4,$BF
	.dc.b	$2B,$FB,$6B,$FC,$6C,$FE,$A6,$7F,$1B,$3F,$AB,$9F,$C6,$CF,$EB,$67
	.dc.b	$F1,$B3,$FA,$F9,$FC,$6C,$FE,$C6,$7F,$1B,$3F,$14,$7F,$1B,$3F,$C6
	.dc.b	$DE,$8F,$F1,$CA,$FF,$72,$C6,$8F,$CA,$9C,$8F,$CC,$9D,$8F,$CE,$9E
	.dc.b	$23,$F4,$27,$98,$FD,$29,$EA,$3F,$52,$7E,$AD,$FF,$38,$CF,$EC,$23
	.dc.b	$F6,$47,$F4,$1E,$1F,$C6,$11,$FC,$67,$83,$FD,$BB,$87,$F1,$8F,$F7
	.dc.b	$0E,$1F,$C6,$3F,$DC,$B8,$7F,$18,$FF,$74,$E1,$FC,$63,$FD,$DB,$87
	.dc.b	$F1,$8F,$F7,$8E,$1F,$C6,$3F,$DE,$B8,$7F,$12,$FF,$84,$72,$FC,$6F
	.dc.b	$17,$E8,$BC,$2F,$E3,$B1,$FD,$FB,$C3,$FE,$42,$FF,$80,$3C,$4F,$E4
	.dc.b	$AF,$F8,$0B,$C5,$FE,$52,$FF,$81,$3C,$6F,$E5,$AF,$F8,$1B,$C7,$FE
	.dc.b	$62,$FF,$82,$3C,$8F,$F0,$27,$17,$C0,$8F,$D7,$9F,$CE,$FC,$9F,$E7
	.dc.b	$CF,$EC,$BA,$BF,$C1,$44,$7F,$BD,$FB,$7F,$E0,$DF,$13,$FC,$1D,$EE
	.dc.b	$FF,$83,$FE,$27,$FC,$21,$F1,$3F,$E9,$9F,$D2,$9C,$97,$E0,$71,$F0
	.dc.b	$39,$F8,$AD,$7C,$1E,$FE,$2B,$87,$C2,$F1,$F8,$AE,$5F,$0F,$F2,$38
	.dc.b	$7F,$64,$0F,$93,$E1,$BB,$37,$19,$F1,$7C,$2B,$B9,$C0,$C7,$BC,$EF
	.dc.b	$83,$FD,$6D,$2D,$3C,$B3,$93,$94,$36,$F5,$C6,$70,$EB,$C1,$F0,$70
	.dc.b	$F8,$73,$FE,$22,$E2,$F8,$EF,$F8,$93,$93,$FC,$6D,$CB,$F3,$63,$FC
	.dc.b	$7B,$CD,$FC,$97,$80,$E7,$7F,$F1,$FB,$C2,$79,$1F,$E2,$EE,$6F,$A4
	.dc.b	$FF,$8C,$39,$BC,$4F,$97,$76,$BE,$88,$FF,$41,$C7,$AD,$D7,$2F,$D3
	.dc.b	$78,$72,$FD,$37,$97,$2F,$F4,$BC,$FE,$6F,$0B,$E1,$FC,$F5,$F0,$DA
	.dc.b	$F6,$AE,$4F,$82,$FB,$3E,$2C,$3F,$08,$F9,$F9,$FF,$10,$FA,$78,$FE
	.dc.b	$5B,$FE,$62,$E3,$F9,$6F,$F9,$93,$8F,$E2,$3F,$28,$FF,$28,$F0,$FF
	.dc.b	$4F,$D9,$F1,$5F,$95,$24,$FC,$5F,$89,$FC,$C7,$C5,$7E,$67,$E2,$3F
	.dc.b	$35,$F1,$1F,$9B,$F7,$FF,$39,$F1,$1F,$9D,$F8,$1F,$CF,$7C,$47,$E7
	.dc.b	$FE,$0F,$F4,$1F,$11,$FA,$1F,$85,$FF,$7D,$61,$FE,$C0,$D7,$FD,$62
	.dc.b	$FF,$C0,$17,$F1,$0F,$FC,$00,$7F,$C3,$47,$F9,$0F,$D3,$FE,$55,$C9
	.dc.b	$FE,$82,$F1,$39,$FC,$57,$4F,$87,$EB,$F1,$5D,$F0,$FB,$D9,$F8,$DF
	.dc.b	$C0,$6B,$E7,$7F,$05,$BF,$AD,$FC,$27,$07,$BD,$FC,$37,$17,$EE,$78
	.dc.b	$FE,$2C,$7E,$1B,$93,$F2,$9C,$DF,$93,$DF,$FD,$2A,$FD,$0E,$FF,$F4
	.dc.b	$9F,$47,$D4,$7F,$D2,$BD,$3F,$61,$FC,$3F,$4F,$ED,$31,$FC,$8F,$81
	.dc.b	$F9,$39,$CF,$8B,$FB,$09,$FC,$AF,$8D,$FB,$67,$F2,$FE,$3F,$EF,$9F
	.dc.b	$E5,$3E,$17,$E3,$AF,$F2,$6D,$D7,$F9,$74,$78,$3C,$2F,$C8,$BE,$2F
	.dc.b	$0B,$F2,$2F,$93,$C2,$FC,$8E,$1F,$A4,$F0,$BF,$22,$FE,$9B,$0F,$51
	.dc.b	$FE,$B4,$DF,$DC,$FF,$AD,$B8,$3E,$13,$F1,$27,$F2,$27,$F5,$1E,$1F
	.dc.b	$90,$FF,$2D,$3F,$95,$9F,$BA,$3F,$D7,$BE,$2F,$EE,$9F,$F5,$FF,$8B
	.dc.b	$FB,$A7,$E3,$9F,$A5,$3F,$79,$C3,$F1,$F4,$F9,$BC,$3F,$1D,$F9,$FC
	.dc.b	$3F,$1D,$FF,$63,$F7,$7E,$E7,$3F,$C6,$61,$FB,$C3,$F8,$35,$71,$8F
	.dc.b	$E4,$77,$D5,$FE,$4B,$87,$D7,$7F,$93,$F1,$F9,$71,$FF,$69,$E5,$E9
	.dc.b	$7F,$DA,$B9,$FF,$47,$CB,$DE,$7E,$D4,$9E,$47,$F5,$0F,$1B,$FC,$87
	.dc.b	$B7,$C6,$6C,$F0,$BF,$C6,$BF,$C7,$CF,$E5,$87,$F1,$F3,$F8,$31,$FC
	.dc.b	$7F,$1F,$33,$FC,$E3,$3F,$53,$FC,$EB,$4F,$73,$FC,$F3,$5F,$8D,$AF
	.dc.b	$E2,$09,$1F,$D4,$0F,$EE,$FC,$CE,$5F,$C5,$CF,$F4,$8E,$5F,$C5,$BF
	.dc.b	$D2,$B9,$7F,$17,$A7,$E8,$39,$7F,$16,$FE,$8A,$7F,$17,$8F,$F9,$83
	.dc.b	$97,$F1,$6F,$F9,$8B,$97,$F1,$6F,$E1,$CF,$8F,$F0,$BF,$D5,$BA,$3F
	.dc.b	$6B,$87,$EB,$1F,$DC,$4F,$EE,$8F,$E8,$1E,$3F,$F1,$A7,$FA,$17,$89
	.dc.b	$FC,$69,$FE,$89,$E3,$FF,$1A,$7F,$A3,$78,$FF,$C6,$9F,$E9,$1E,$3F
	.dc.b	$F0,$E7,$F8,$BF,$8B,$FC,$49,$FE,$2F,$97,$F4,$2F,$88,$FF,$6C,$B7
	.dc.b	$C8,$FE,$F1,$D4,$CF,$F8,$F5,$FF,$22,$C3,$A3,$FD,$4B,$1F,$03,$FD
	.dc.b	$53,$2F,$23,$FD,$5B,$3F,$43,$FD,$62,$FD,$8F,$F5,$AD,$3F,$9C,$BF
	.dc.b	$CE,$CF,$EC,$1A,$FF,$1A,$EC,$FC,$77,$93,$FC,$16,$7E,$66,$DF,$DB
	.dc.b	$1F,$C7,$71,$FE,$ED,$FC,$71,$FD,$3C,$FC,$71,$FC,$0C,$FC,$71,$FB
	.dc.b	$33,$FB,$5F,$5F,$E3,$B7,$FF,$07,$F5,$FE,$39,$FF,$08,$F5,$FE,$3A
	.dc.b	$BF,$19,$87,$F1,$8E,$DF,$DA,$47,$F8,$3B,$AF,$F1,$AF,$F8,$43,$AF
	.dc.b	$F1,$AF,$F8,$4B,$AF,$F1,$BA,$7F,$86,$3A,$FF,$1A,$FF,$86,$8F,$C6
	.dc.b	$E7,$FE,$AE,$EB,$FC,$6B,$FE,$B0,$EB,$FC,$6B,$FA,$EC,$3F,$1A,$79
	.dc.b	$2B,$F4,$31,$FD,$F7,$3F,$2B,$67,$FB,$F4,$63,$C1,$F9,$93,$99,$FE
	.dc.b	$6B,$39,$23,$C0,$86,$8B,$AF,$C3,$6D,$F8,$AF,$A9,$2B,$9C,$F1,$27
	.dc.b	$17,$F0,$B9,$D4,$72,$3D,$84,$C3,$CD,$87,$78,$E4,$D2,$61,$E1,$C3
	.dc.b	$BC,$62,$79,$5F,$67,$8D,$6D,$F0,$5F,$68,$C4,$68,$E3,$7D,$23,$38
	.dc.b	$7F,$23,$E5,$7C,$4F,$F8,$28,$97,$FE,$08,$EB,$FF,$06,$FC,$57,$F8
	.dc.b	$3B,$DD,$FF,$07,$FB,$7F,$E1,$0F,$7B,$FC,$23,$F1,$3F,$E1,$2F,$80
	.dc.b	$FF,$09,$F8,$7F,$E1,$4F,$81,$FF,$0A,$FC,$4F,$F8,$5B,$E0,$FF,$C2
	.dc.b	$FF,$13,$FE,$18,$F8,$5F,$F0,$CF,$C4,$FE,$EB,$C9,$FD,$EF,$2F,$F9
	.dc.b	$C7,$9D,$18,$E3,$FF,$7C,$F9,$DC,$1F,$F0,$EE,$1D,$1F,$F0,$FE,$5E
	.dc.b	$07,$FC,$43,$A7,$91,$FF,$11,$ED,$E8,$7F,$C4,$BC,$1E,$C7,$FC,$4F
	.dc.b	$C5,$F0,$1F,$84,$7F,$8A,$B9,$7E,$2A,$E7,$FC,$17,$47,$CD,$8F,$9E
	.dc.b	$7D,$4E,$9F,$B1,$D5,$F6,$F8,$3E,$B4,$E8,$FA,$CE,$09,$3C,$38,$7E
	.dc.b	$DD,$71,$8C,$57,$18,$C5,$F8,$6F,$8B,$6B,$FC,$CB,$87,$9A,$FF,$D3
	.dc.b	$84,$AF,$D5,$36,$7F,$CD,$6D,$AD,$B4,$A7,$F5,$73,$06,$A7,$9B,$FE
	.dc.b	$D1,$32,$43,$04,$39,$15,$E2,$E0,$8E,$09,$3E,$68,$E0,$D7,$6D,$8D
	.dc.b	$CF,$A2,$72,$D8,$CC,$7B,$1A,$FA,$E7,$91,$27,$83,$FE,$90,$ED,$79
	.dc.b	$FE,$29,$C9,$2B,$A5,$71,$6C,$65,$EC,$D7,$E3,$1E,$63,$64,$36,$53
	.dc.b	$77,$E8,$6C,$F7,$76,$43,$33,$DD,$F4,$1C,$CA,$6C,$A7,$DE,$BF,$7B
	.dc.b	$E5,$34,$53,$A9,$CC,$C1,$C8,$86,$09,$5E,$97,$2D,$73,$5E,$B7,$A9
	.dc.b	$E0,$8D,$12,$3E,$04,$F9,$A3,$15,$EF,$AE,$6D,$8C,$C7,$9D,$AF,$5A
	.dc.b	$F0,$3E,$46,$BC,$8E,$0E,$23,$8C,$79,$88,$7C,$CA,$F2,$4F,$EE,$A7
	.dc.b	$F6,$F3,$FC,$1E,$BF,$70,$47,$04,$3F,$83,$CF,$12,$46,$E3,$2D,$C7
	.dc.b	$FA,$86,$3F,$CF,$FE,$1F,$FD,$8E,$4A,$FF,$62,$FC,$4F,$FB,$2B,$E2
	.dc.b	$7F,$82,$DF,$FB,$C3,$F4,$BF,$E0,$2D,$F8,$AC,$EF,$FC,$D6,$5E,$EE
	.dc.b	$31,$6C,$97,$BD,$98,$C7,$C4,$D9,$7C,$05,$F8,$9C,$7C,$16,$7E,$27
	.dc.b	$5F,$09,$BF,$89,$E1,$F0,$DC,$7E,$27,$97,$C4,$4D,$C3,$8F,$D5,$3C
	.dc.b	$23,$31,$ED,$BE,$B1,$8E,$2F,$66,$5E,$28,$C1,$EC,$D9,$AF,$A7,$6F
	.dc.b	$56,$FE,$BE,$1F,$67,$17,$B7,$8F,$DD,$8F,$CD,$3E,$79,$FE,$F4,$E6
	.dc.b	$FA,$4F,$F4,$BF,$23,$F9,$FE,$B2,$3F,$9F,$37,$97,$F0,$F1,$F7,$8F
	.dc.b	$C2,$70,$7F,$42,$6C,$EF,$AB,$F0,$B1,$83,$99,$2B,$9D,$71,$8E,$87
	.dc.b	$8C,$FB,$7E,$6B,$1F,$3C,$B0,$FE,$ED,$39,$AF,$69,$F4,$63,$3D,$1F
	.dc.b	$0A,$3E,$5E,$1F,$85,$CF,$E3,$C6,$38,$7F,$0E,$DD,$FE,$73,$9E,$1F
	.dc.b	$A8,$F5,$3E,$A7,$17,$DA,$71,$87,$E4,$78,$FF,$01,$18,$C1,$C9,$F8
	.dc.b	$78,$C7,$2F,$E2,$DF,$17,$37,$E3,$DF,$27,$3F,$E4,$DF,$37,$47,$E5
	.dc.b	$DF,$47,$47,$E0,$9F,$57,$47,$B1,$F0,$70,$7C,$17,$CB,$C6,$E2,$FD
	.dc.b	$22,$E9,$F9,$BC,$9F,$33,$9B,$FB,$BF,$57,$FB,$67,$85,$F1,$3F,$A9
	.dc.b	$26,$5F,$27,$C6,$FD,$67,$6F,$EB,$7E,$27,$F5,$DE,$87,$EB,$FE,$07
	.dc.b	$F6,$1E,$97,$EC,$7E,$0B,$F6,$9F,$A6,$7D,$BC,$BE,$AA,$F8,$3C,$F8
	.dc.b	$BA,$F6,$F7,$EB,$F0,$F6,$F8,$FB,$3F,$27,$1F,$F0,$DF,$17,$CE,$79
	.dc.b	$C6,$DE,$B5,$CD,$EF,$87,$8B,$1F,$1F,$2F,$96,$7F,$09,$AF,$A5,$FC
	.dc.b	$36,$FE,$D7,$F1,$1C,$3F,$C2,$F8,$BE,$1C,$7D,$FE,$3F,$93,$E4,$7F
	.dc.b	$92,$F8,$5E,$DF,$C0,$25,$7C,$0F,$6F,$E1,$FB,$BF,$13,$CD,$F8,$BF
	.dc.b	$09,$F1,$BD,$FF,$8F,$F0,$BF,$23,$BF,$E4,$FC,$37,$CA,$F8,$2F,$97
	.dc.b	$F0,$FF,$33,$E1,$3E,$6F,$C4,$7C,$EF,$86,$FE,$45,$FA,$6B,$E2,$BE
	.dc.b	$47,$AD,$FC,$BF,$D0,$FC,$A3,$FC,$63,$8F,$F4,$F5,$B7,$EA,$DF,$47
	.dc.b	$EB,$E3,$E0,$7E,$CE,$7E,$47,$ED,$EB,$E8,$7E,$EE,$FE,$C7,$EF,$F0
	.dc.b	$FC,$07,$E1,$1F,$82,$E4,$F8,$AF,$F0,$3F,$4F,$F9,$E7,$47,$CE,$D7
	.dc.b	$FA,$33,$F4,$67,$E9,$9F,$BB,$F1,$FF,$22,$FE,$F7,$C7,$FC,$8B,$FB
	.dc.b	$FF,$1F,$EF,$BF,$90,$F0,$7F,$1E,$FE,$4B,$C7,$FC,$7B,$F9,$4F,$1F
	.dc.b	$F1,$EF,$F1,$BC,$FF,$8C,$F8,$7F,$77,$C6,$FC,$74,$FF,$02,$EF,$78
	.dc.b	$DF,$8F,$8C,$78,$DF,$8F,$7F,$3F,$7F,$8F,$7F,$83,$78,$DF,$8F,$7F
	.dc.b	$84,$78,$DF,$8F,$7F,$85,$78,$DE,$F7,$F1,$C7,$F1,$9E,$2F,$C7,$BF
	.dc.b	$D8,$B2,$FC,$7B,$F2,$C8,$79,$7C,$4F,$A0,$F9,$F9,$BE,$A3,$E9,$E7
	.dc.b	$FB,$0F,$AF,$A3,$EE,$3E,$DE,$9F,$C0,$3E,$FE,$AF,$C2,$3F,$07,$AF
	.dc.b	$F1,$FD,$DF,$8E,$E0,$FE,$F1,$DD,$F8,$E7,$FB,$D7,$A5,$AB,$FC,$69
	.dc.b	$FC,$0B,$C0,$FC,$D3,$FC,$13,$C1,$FC,$F3,$FC,$1B,$C2,$FD,$13,$FC
	.dc.b	$23,$C3,$FD,$33,$FC,$2B,$C4,$FD,$53,$FC,$33,$C5,$FE,$1F,$C7,$F8
	.dc.b	$CF,$0F,$F8,$87,$6B,$9B,$E3,$DF,$E3,$FC,$8F,$E3,$EF,$CD,$21,$FC
	.dc.b	$EF,$C8,$FE,$3E,$FE,$3C,$FE,$0A,$7E,$3C,$FE,$12,$7E,$3C,$FB,$E7
	.dc.b	$E3,$CF,$E2,$27,$E3,$CF,$E2,$A7,$E3,$CF,$E3,$27,$F5,$5D,$FF,$C3
	.dc.b	$5E,$BF,$F8,$E6,$BF,$8F,$4F,$E3,$FC,$0F,$C8,$E5,$EC,$C7,$F2,$33
	.dc.b	$D4,$FC,$89,$E1,$3F,$2D,$C5,$F9,$19,$FF,$18,$77,$7E,$45,$FF,$19
	.dc.b	$77,$7E,$45,$FF,$1A,$77,$7E,$45,$FD,$8D,$FC,$4F,$29,$E5,$7F,$1F
	.dc.b	$F2,$7E,$76,$56,$BF,$7A,$7E,$97,$8F,$EA,$BF,$A7,$E4,$FB,$2F,$EA
	.dc.b	$79,$7E,$EB,$FA,$BE,$6F,$C0,$BF,$AD,$E7,$FC,$2B,$FA,$FE,$8F,$E3
	.dc.b	$CF,$F1,$8E,$AF,$C8,$3F,$E2,$9E,$AF,$C8,$3F,$E2,$BE,$7F,$C8,$3F
	.dc.b	$97,$3F,$20,$7F,$1C,$3F,$20,$7F,$1C,$3F,$20,$7F,$1C,$3F,$20,$7F
	.dc.b	$1C,$3F,$20,$7F,$1C,$3F,$20,$7C,$0E,$0F,$A9,$B7,$F1,$CE,$0F,$ED
	.dc.b	$9E,$0F,$F1,$BE,$0F,$F2,$E3,$F3,$5F,$F2,$F7,$17,$D1,$7F,$CC,$1C
	.dc.b	$7F,$55,$FF,$31,$72,$7D,$97,$FA,$71,$FC,$37,$2F,$BE,$7F,$13,$F1
	.dc.b	$7F,$8D,$BF,$C5,$7C,$5F,$E3,$6F,$F8,$C3,$4F,$D1,$F5,$32,$CC,$7E
	.dc.b	$2E,$B8,$BF,$8D,$BF,$E4,$F7,$FC,$7F,$C6,$FE,$56,$FF,$87,$FC,$7F
	.dc.b	$E5,$EF,$F8,$87,$C8,$FE,$66,$FF,$88,$FC,$9F,$D4,$BF,$AB,$3F,$5B
	.dc.b	$87,$E3,$1F,$E8,$1D,$9F,$B4,$8F,$93,$1F,$C6,$9F,$CF,$78,$3F,$C6
	.dc.b	$9F,$D0,$78,$FF,$C6,$9F,$D1,$78,$FF,$C6,$9F,$D2,$78,$FF,$C6,$9F
	.dc.b	$D3,$78,$FF,$C6,$9F,$D4,$78,$FF,$C6,$9F,$D5,$78,$FF,$86,$7F,$B2
	.dc.b	$65,$FE,$38,$3F,$C9,$1D,$7F,$8E,$F1,$DE,$57,$F6,$79,$FC,$61,$F9
	.dc.b	$53,$F1,$87,$E6,$4F,$C6,$1F,$9D,$3F,$18,$7E,$84,$FC,$61,$FA,$53
	.dc.b	$F1,$87,$EA,$4F,$C6,$73,$FE,$B5,$FE,$A1,$D9,$FD,$0A,$BC,$7F,$13
	.dc.b	$FE,$00,$24,$FF,$7C,$F8,$BF,$F0,$1F,$B7,$FE,$04,$F8,$9F,$F0,$2F
	.dc.b	$C4,$FF,$81,$BD,$EF,$F0,$3F,$C4,$FF,$82,$3E,$03,$FC,$13,$F1,$3F
	.dc.b	$E0,$AF,$82,$FF,$05,$F9,$FF,$E0,$CF,$83,$FF,$06,$FC,$4F,$F8,$3B
	.dc.b	$E1,$7F,$C1,$FF,$13,$FE,$10,$F8,$7F,$E3,$3E,$3C,$E2,$FF,$06,$C6
	.dc.b	$9F,$F0,$7D,$72,$7F,$C2,$38,$76,$7F,$C2,$79,$78,$9F,$F0,$AE,$9E
	.dc.b	$67,$FC,$2F,$B7,$A9,$FF,$0C,$F0,$7B,$9F,$80,$7F,$87,$38,$FE,$1B
	.dc.b	$FE,$0D,$E7,$46,$3A,$7F,$A2,$C6,$27,$89,$FE,$7E,$5C,$9F,$26,$3F
	.dc.b	$C4,$FC,$9F,$25,$FF,$14,$F2,$7C,$97,$FC,$57,$C9,$F2,$5F,$F1,$6F
	.dc.b	$27,$C9,$7F,$C5,$FC,$9F,$25,$FF,$18,$F2,$7C,$97,$F1,$E4,$3C,$5C
	.dc.b	$7F,$25,$F2,$71,$FC,$97,$CD,$C7,$F2,$5F,$47,$1F,$C9,$7D,$5C,$7F
	.dc.b	$25,$F6,$71,$FC,$97,$DD,$C7,$F2,$5F,$81,$CB,$F1,$7A,$3F,$DB,$FC
	.dc.b	$FF,$8E,$9F,$F7,$0F,$99,$FD,$09,$FF,$79,$F2,$BE,$27,$FC,$A2,$48
	.dc.b	$FF,$26,$FB,$7F,$E5,$4F,$73,$FC,$AB,$EE,$FF,$BD,$FD,$BF,$F8,$03
	.dc.b	$9F,$65,$EF,$67,$E2,$35,$F0,$1B,$F8,$8E,$1F,$05,$C7,$E2,$39,$7C
	.dc.b	$27,$3F,$88,$E9,$F0,$DD,$7E,$23,$F7,$1E,$C7,$F6,$BF,$73,$1C,$7F
	.dc.b	$EC,$6D,$D7,$C6,$31,$8F,$57,$59,$F8,$5E,$1A,$F9,$5E,$5B,$FA,$DF
	.dc.b	$69,$FE,$6C,$E2,$FF,$39,$F1,$FC,$38,$FC,$67,$27,$EE,$79,$7F,$21
	.dc.b	$1F,$8F,$F2,$3E,$83,$F3,$3A,$7E,$A7,$91,$F9,$6E,$BF,$B0,$FE,$63
	.dc.b	$B3,$EE,$3F,$9A,$ED,$FC,$03,$F9,$CE,$EF,$C2,$3F,$9E,$EF,$FC,$43
	.dc.b	$FA,$0F,$03,$F1,$8F,$E8,$BC,$1F,$ED,$3E,$0B,$E1,$5E,$67,$E5,$76
	.dc.b	$FD,$27,$89,$F9,$97,$F4,$DE,$2F,$E7,$5F,$D4,$78,$DF,$A1,$7F,$55
	.dc.b	$E3,$FE,$95,$FD,$67,$91,$FA,$97,$F5,$DE,$4F,$EB,$5F,$D8,$79,$5F
	.dc.b	$E9,$AE,$5F,$87,$AF,$E4,$67,$F7,$5C,$D8,$E7,$F9,$3C,$3F,$BA,$BF
	.dc.b	$8F,$E4,$7E,$62,$5D,$1F,$22,$31,$D1,$F2,$1D,$74,$7C,$87,$87,$47
	.dc.b	$C8,$79,$74,$7C,$87,$A7,$47,$C8,$7B,$74,$7F,$1F,$D9,$AF,$F6,$EF
	.dc.b	$2F,$E4,$3F,$C3,$B9,$FE,$43,$FC,$43,$9F,$E4,$3F,$C4,$B9,$FE,$43
	.dc.b	$FC,$53,$9F,$E4,$3F,$C5,$B9,$FE,$43,$FC,$63,$9F,$E4,$3F,$C6,$B9
	.dc.b	$FE,$46,$7F,$C8,$48,$7C,$82,$C7,$F2,$0E,$6D,$C7,$D0,$3F,$95,$9F
	.dc.b	$C7,$8F,$E5,$E7,$F1,$E3,$F9,$99,$FC,$78,$FE,$6E,$7F,$1E,$3F,$9D
	.dc.b	$9F,$C7,$8F,$E7,$E7,$F1,$E3,$FA,$19,$FC,$76,$3F,$DA,$8E,$0F,$C6
	.dc.b	$9C,$0F,$C6,$9D,$0F,$C6,$9E,$03,$F1,$A7,$90,$FC,$69,$E8,$3F,$1A
	.dc.b	$7B,$0F,$71,$FC,$78,$F8,$7B,$7E,$1E,$7F,$69,$D5,$FB,$7A,$FE,$8A
	.dc.b	$FE,$EA,$BF,$78,$7F,$BA,$3C,$AF,$E0,$0F,$F0,$23,$F8,$67,$95,$FC
	.dc.b	$61,$FE,$1D,$E5,$7F,$18,$7F,$88,$79,$5F,$C6,$1F,$E2,$5E,$57,$F1
	.dc.b	$87,$F8,$A7,$95,$FC,$63,$93,$F1,$99,$FF,$85,$0F,$F4,$DF,$90,$9E
	.dc.b	$15,$F9,$0D,$BF,$95,$61,$D6,$30,$3B,$D6,$CB,$C9,$3A,$9C,$E7,$E9
	.dc.b	$8C,$53,$4F,$E7,$1E,$27,$C0,$9F,$EC,$DB,$7F,$3D,$7F,$9F,$83,$F8
	.dc.b	$F9,$75,$F9,$6F,$F4,$AE,$5F,$A0,$FD,$23,$FA,$5F,$0F,$E5,$27,$F3
	.dc.b	$91,$F6,$5F,$D0,$3F,$75,$FD,$23,$F8,$17,$F5,$0F,$E1,$5F,$D6,$3F
	.dc.b	$89,$7F,$60,$FE,$35,$FD,$A3,$F9,$17,$F2,$6F,$F9,$04,$FE,$E2,$7E
	.dc.b	$64,$FE,$E2,$7E,$74,$FE,$E2,$7E,$84,$FE,$E2,$7E,$94,$FE,$E2,$7E
	.dc.b	$A4,$FE,$E2,$7E,$B4,$FE,$E2,$7C,$5D,$BF,$8C,$71,$7C,$17,$E2,$C7
	.dc.b	$F4,$53,$E4,$1F,$6F,$37,$C4,$7F,$6B,$24,$7F,$5D,$6B,$F6,$9B,$7F
	.dc.b	$8A,$6D,$7F,$71,$6C,$F5,$6E,$DF,$88,$6D,$F8,$27,$B1,$F2,$6B,$FC
	.dc.b	$0F,$E5,$FF,$A3,$63,$D8,$7E,$45,$C3,$F3,$0F,$21,$C8,$A4,$C3,$CC
	.dc.b	$E5,$AE,$67,$D0,$E5,$AE,$4F,$51,$92,$9D,$BC,$CF,$F0,$D6,$3E,$D8
	.dc.b	$D3,$5E,$4D,$AF,$09,$2B,$BB,$67,$AB,$67,$84,$E8,$93,$A6,$C6,$A3
	.dc.b	$EE,$79,$9F,$DC,$BC,$6F,$EE,$31,$FE,$17,$8F,$A3,$9C,$8F,$3E,$7F
	.dc.b	$4D,$F3,$B7,$1F,$13,$71,$EE,$D9,$F1,$13,$1F,$11,$F0,$72,$F8,$B3
	.dc.b	$E2,$9F,$E0,$7E,$37,$F8,$E4,$FF,$1D,$7B,$DF,$E3,$B9,$FF,$1E,$74
	.dc.b	$7F,$8F,$7D,$7F,$F1,$F7,$BD,$FE,$3F,$F6,$7F,$C8,$1E,$F7,$F9,$07
	.dc.b	$DB,$FE,$0D,$9F,$F1,$1F,$D3,$9B,$E3,$82,$CF,$F1,$99,$7B,$B9,$AF
	.dc.b	$F2,$AC,$3E,$23,$15,$BF,$7E,$F1,$A9,$F1,$38,$5F,$03,$FD,$DB,$1F
	.dc.b	$CC,$F0,$FF,$30,$7F,$BE,$E3,$F9,$AC,$39,$57,$1C,$39,$63,$CF,$2E
	.dc.b	$99,$F5,$C7,$DD,$8F,$F9,$3F,$6F,$84,$FF,$94,$78,$3E,$33,$FE,$53
	.dc.b	$D3,$E5,$3F,$E5,$5E,$3F,$9C,$FF,$95,$F9,$3F,$D8,$1C,$9F,$67,$7F
	.dc.b	$AE,$D2,$73,$7E,$16,$BF,$06,$DE,$2F,$C1,$B7,$4F,$C1,$BF,$E7,$3E
	.dc.b	$7F,$C8,$63,$6F,$F1,$B3,$B7,$F0,$D8,$76,$72,$43,$B9,$70,$FD,$BF
	.dc.b	$3F,$E2,$AB,$EF,$CF,$E0,$32,$FD,$73,$6F,$F3,$4D,$9D,$BF,$BC,$30
	.dc.b	$62,$78,$B9,$21,$FB,$E3,$67,$81,$DB,$93,$6E,$09,$3E,$28,$FC,$09
	.dc.b	$FD,$97,$A7,$F8,$CE,$1F,$E3,$D8,$FE,$40,$73,$25,$72,$BF,$4C,$62
	.dc.b	$FA,$36,$33,$3B,$8F,$64,$EA,$31,$1E,$98,$C1,$27,$F5,$3C,$2F,$7F
	.dc.b	$D7,$39,$AE,$75,$DA,$31,$1C,$E3,$04,$76,$43,$45,$3E,$59,$0E,$EF
	.dc.b	$5D,$3F,$13,$1F,$C5,$CF,$E7,$86,$CF,$80,$73,$38,$94,$C9,$E0,$21
	.dc.b	$C7,$0F,$86,$FD,$CC,$3E,$03,$C0,$D5,$FB,$9B,$7E,$D7,$99,$E0,$29
	.dc.b	$C4,$C1,$DF,$D2,$79,$9F,$C0,$E3,$FA,$D7,$0F,$F0,$77,$81,$31,$F2
	.dc.b	$BC,$88,$67,$0F,$23,$E1,$C7,$D3,$3C,$23,$58,$F9,$DF,$49,$F1,$63
	.dc.b	$44,$AE,$71,$D7,$2F,$A2,$FF,$8E,$1F,$F0,$0B,$C8,$F7,$CF,$1A,$ED
	.dc.b	$1E,$68,$C3,$DE,$BF,$0F,$7E,$67,$F8,$04,$7E,$F8,$9A,$7F,$0C,$7F
	.dc.b	$80,$70,$7F,$0C,$69,$FC,$23,$0F,$F0,$4E,$3F,$2F,$16,$DF,$AC,$BF
	.dc.b	$F2,$44,$7F,$95,$A7,$F8,$83,$FC,$7A,$AE,$3E,$2C,$7C,$18,$F8,$B0
	.dc.b	$F0,$3E,$5D,$FF,$BE,$F4,$3E,$27,$FA,$01,$27,$F9,$E4,$3D,$BF,$E8
	.dc.b	$9F,$03,$FD,$17,$E2,$7F,$A3,$72,$7F,$47,$F8,$9F,$E9,$1F,$07,$FD
	.dc.b	$27,$E2,$7F,$A5,$7C,$27,$F4,$BF,$89,$FE,$99,$F0,$BF,$D3,$7E,$27
	.dc.b	$FA,$77,$C3,$7F,$4F,$F8,$9F,$EA,$1F,$0F,$FC,$B7,$87,$FC,$9D,$CB
	.dc.b	$FE,$50,$3F,$CF,$BD,$3A,$8F,$2D,$72,$7C,$F8,$76,$7D,$39,$78,$9F
	.dc.b	$5E,$9E,$67,$DB,$B7,$A9,$F7,$F0,$7B,$9F,$83,$C5,$F0,$DF,$EE,$5C
	.dc.b	$9F,$D3,$BA,$3E,$79,$23,$ED,$F4,$FD,$36,$F4,$FD,$3E,$9F,$A3,$C1
	.dc.b	$FD,$E2,$9D,$DF,$AA,$BF,$ED,$5D,$7F,$7D,$FD,$59,$FC,$AA,$7A,$3F
	.dc.b	$6C,$FC,$41,$DB,$1F,$13,$98,$DF,$5F,$DE,$CF,$D5,$8F,$87,$AF,$F0
	.dc.b	$17,$EE,$3E,$FF,$5F,$DD,$7F,$03,$D7,$F7,$5F,$C1,$F5,$FD,$D7,$F0
	.dc.b	$BD,$7F,$75,$FC,$3F,$5F,$DD,$7F,$13,$D7,$F7,$5F,$C5,$F5,$FD,$DC
	.dc.b	$FE,$E7,$17,$E7,$3C,$AF,$F1,$54,$7D,$EA,$D9,$F7,$8E,$67,$DE,$3B
	.dc.b	$9F,$78,$F1,$9F,$78,$F3,$9F,$78,$F5,$9F,$78,$F7,$9F,$04,$FB,$A7
	.dc.b	$E8,$36,$FF,$18,$EF,$FE,$62,$FD,$3D,$F1,$5F,$E2,$86,$D7,$F8,$85
	.dc.b	$AF,$FB,$1D,$79,$DB,$F8,$3E,$1E,$7F,$1F,$83,$E5,$ED,$73,$F8,$3E
	.dc.b	$9E,$E7,$5F,$83,$ED,$EF,$77,$F8,$3F,$07,$AF,$E1,$F8,$3F,$17,$C0
	.dc.b	$F8,$FE,$0F,$E6,$73,$FF,$8D,$B8,$BF,$17,$E7,$61,$FF,$6F,$71,$AB
	.dc.b	$8C,$7A,$70,$EA,$FA,$F2,$F0,$BE,$DD,$3C,$AF,$BF,$6F,$4B,$F0,$78
	.dc.b	$3D,$AF,$C3,$E2,$F8,$DC,$7F,$1F,$92,$4F,$FA,$DB,$C1,$6A,$F8,$3F
	.dc.b	$98,$49,$F9,$9E,$87,$CF,$F8,$3F,$A1,$F0,$9F,$47,$E1,$7E,$97,$BF
	.dc.b	$FE,$5A,$FD,$43,$FC,$53,$DF,$FA,$D5,$F0,$39,$F8,$2D,$7C,$3E,$FE
	.dc.b	$13,$87,$C4,$71,$F8,$6E,$5F,$13,$CF,$E2,$3A,$7C,$57,$F4,$8D,$3A
	.dc.b	$39,$E2,$FF,$30,$6A,$BE,$51,$8C,$7B,$3A,$CF,$C4,$F0,$D7,$D0,$FA
	.dc.b	$8F,$C5,$6F,$EE,$7F,$19,$C3,$F8,$FE,$2F,$89,$1F,$8D,$E3,$FC,$AF
	.dc.b	$85,$FD,$13,$97,$F2,$3B,$FF,$0D,$6F,$83,$9E,$8F,$A7,$87,$D5,$3F
	.dc.b	$31,$E3,$7E,$4D,$FC,$D7,$8D,$F9,$37,$F3,$9E,$37,$E4,$DF,$CF,$78
	.dc.b	$DF,$93,$7F,$41,$9F,$E4,$DF,$D1,$78,$DF,$93,$7F,$C3,$C7,$F9,$17
	.dc.b	$C6,$FE,$B9,$EA,$BD,$DD,$4F,$E9,$EB,$93,$FA,$9B,$EC,$FE,$AF,$0F
	.dc.b	$13,$FA,$DC,$7C,$CF,$EB,$F2,$F5,$3F,$B1,$CF,$F6,$0F,$ED,$7C,$1F
	.dc.b	$C8,$BF,$B5,$D7,$E4,$BF,$A4,$E1,$F9,$BC,$3F,$25,$FA,$11,$F4,$8F
	.dc.b	$DF,$78,$1F,$91,$7F,$80,$78,$1F,$69,$FB,$87,$E1,$3C,$0F,$C7,$BF
	.dc.b	$86,$F0,$3F,$1E,$FE,$23,$C0,$FC,$7B,$F8,$AF,$03,$F1,$EF,$FA,$AF
	.dc.b	$CB,$FE,$47,$CF,$FC,$97,$DF,$CC,$B5,$FC,$F5,$F2,$8C,$57,$F8,$1B
	.dc.b	$C1,$63,$E0,$9C,$8F,$0D,$2C,$FC,$F3,$A9,$CE,$9F,$AD,$EF,$F7,$4F
	.dc.b	$F1,$DD,$BE,$13,$FC,$7F,$7F,$DB,$3F,$B8,$3F,$92,$70,$7E,$63,$CB
	.dc.b	$FE,$55,$C5,$F4,$B8,$3F,$2C,$FE,$52,$BF,$94,$78,$BF,$C2,$9F,$E5
	.dc.b	$5E,$37,$F0,$E7,$F9,$67,$8F,$FC,$49,$FE,$5D,$E4,$7F,$16,$7F,$98
	.dc.b	$79,$3F,$C6,$9F,$E6,$5E,$57,$F1,$E7,$F9,$A7,$97,$FC,$8F,$8B,$FC
	.dc.b	$5B,$E1,$FE,$53,$C2,$FC,$9E,$DF,$A9,$F0,$BF,$26,$FE,$AF,$C2,$FC
	.dc.b	$9B,$FA,$DF,$0B,$F2,$6F,$EB,$FC,$2F,$C9,$BF,$B1,$F0,$BF,$26,$FE
	.dc.b	$CF,$C2,$FC,$9E,$BF,$D1,$7C,$2F,$85,$DF,$FB,$58,$FE,$49,$E2,$7F
	.dc.b	$23,$7E,$49,$0F,$E9,$9E,$2F,$CE,$7E,$81,$FC,$20,$FC,$99,$FC,$50
	.dc.b	$FB,$07,$E4,$CF,$E2,$47,$E4,$CF,$E2,$C7,$E4,$CF,$E3,$7E,$57,$F1
	.dc.b	$D7,$F0,$5E,$0B,$CB,$FE,$45,$87,$E3,$5D,$67,$FC,$BE,$FF,$90,$DF
	.dc.b	$0F,$2F,$F9,$0B,$CB,$CB,$FE,$42,$F4,$F0,$BF,$17,$7F,$C5,$FC,$2F
	.dc.b	$C5,$BF,$C6,$7C,$2F,$73,$FC,$6F,$5F,$DA,$E9,$FC,$7F,$C2,$FD,$CF
	.dc.b	$4F,$F4,$AF,$07,$E5,$E7,$FD,$79,$FC,$66,$1F,$C8,$FC,$1F,$C6,$3F
	.dc.b	$C9,$7C,$1F,$C6,$3F,$C9,$FC,$1F,$C6,$3F,$CA,$7C,$1F,$C6,$3F,$CA
	.dc.b	$FC,$1F,$C6,$3F,$CB,$7C,$1F,$C6,$3F,$CB,$FC,$1F,$C6,$7A,$DF,$E3
	.dc.b	$9D,$7F,$1B,$AF,$F2,$2A,$D3,$FC,$98,$FE,$E0,$7F,$2A,$3F,$B8,$1F
	.dc.b	$CB,$8F,$EE,$07,$F3,$23,$FB,$81,$FC,$D8,$FE,$E0,$7F,$3A,$3F,$9E
	.dc.b	$1F,$8D,$3F,$C2,$9A,$FF,$20,$F0,$FF,$C3,$5B,$FC,$A9,$FE,$40,$7E
	.dc.b	$DC,$FF,$0C,$79,$7F,$D4,$5F,$F0,$D7,$99,$FD,$55,$FC,$19,$F8,$C3
	.dc.b	$F0,$BE,$37,$F0,$8C,$7F,$B8,$78,$FF,$C3,$DF,$EE,$5E,$47,$F1,$57
	.dc.b	$F8,$BD,$7F,$68,$ED,$FC,$5B,$FC,$4F,$BB,$F1,$CF,$F2,$33,$F2,$67
	.dc.b	$E5,$4F,$C6,$1F,$99,$3F,$18,$7E,$74,$FC,$61,$FA,$13,$F4,$67,$A8
	.dc.b	$FE,$72,$7F,$3B,$3F,$56,$7E,$B6,$7F,$A0,$3F,$E1,$58,$7C,$57,$F8
	.dc.b	$34,$91,$FE,$00,$F5,$FF,$C2,$1E,$4F,$F8,$47,$CC,$FF,$09,$7C,$0F
	.dc.b	$F8,$4F,$CF,$FF,$0A,$79,$BF,$E1,$5F,$8A,$FF,$0B,$7C,$1F,$F8,$5F
	.dc.b	$E2,$BF,$C3,$1F,$0B,$FE,$19,$F8,$A6,$B3,$83,$FD,$0D,$CE,$F0,$F3
	.dc.b	$3F,$CD,$61,$F0,$F8,$AD,$FC,$3F,$F4,$67,$FD,$1D,$DD,$F9,$33,$F2
	.dc.b	$9C,$BF,$3B,$0E,$55,$C7,$1F,$F6,$6E,$3D,$27,$C5,$18,$3C,$CF,$F8
	.dc.b	$83,$3F,$53,$FE,$22,$D7,$DC,$FF,$89,$37,$F8,$4F,$F8,$9B,$87,$E3
	.dc.b	$3F,$E2,$8E,$3F,$F1,$C7,$27,$CC,$8F,$E1,$FC,$BF,$BC,$D7,$E8,$C7
	.dc.b	$F8,$BA,$35,$BF,$F8,$F6,$BE,$8C,$FF,$8C,$B9,$7E,$8B,$FE,$34,$E5
	.dc.b	$FA,$2F,$F8,$DB,$97,$E8,$BF,$E3,$8E,$5F,$A2,$FF,$8E,$B9,$7E,$13
	.dc.b	$FD,$97,$8B,$E8,$3F,$08,$C7,$2E,$BB,$BF,$CB,$1C,$B6,$9F,$11,$F9
	.dc.b	$C9,$FF,$2E,$39,$9D,$56,$F9,$BE,$5E,$1F,$CB,$B9,$BE,$5B,$FC,$C3
	.dc.b	$9B,$E5,$BF,$CC,$B9,$BE,$5B,$FC,$D3,$9B,$E5,$BF,$CD,$B9,$BE,$5B
	.dc.b	$F1,$C8,$3B,$9E,$A7,$DE,$F8,$9F,$41,$27,$DD,$BF,$F9,$CF,$E2,$BD
	.dc.b	$9E,$D7,$B7,$E1,$3D,$DF,$05,$EF,$F8,$5F,$81,$F0,$7F,$07,$E1,$BE
	.dc.b	$17,$C2,$7C,$3F,$87,$F8,$9F,$0B,$F1,$7E,$23,$E3,$7C,$37,$C7,$F8
	.dc.b	$9F,$91,$F0,$FF,$97,$EA,$FF,$4C,$71,$7F,$91,$CF,$E0,$C7,$93,$B9
	.dc.b	$3C,$63,$17,$D5,$D6,$3E,$17,$86,$7E,$57,$96,$BE,$97,$A6,$FE,$D7
	.dc.b	$B7,$0F,$DB,$E2,$F8,$91,$FA,$FE,$4F,$CA,$78,$DF,$92,$DB,$FC,$5C
	.dc.b	$FD,$09,$FA,$47,$E4,$CF,$D0,$1F,$93,$3F,$48,$7E,$4C,$FD,$41,$F9
	.dc.b	$33,$F5,$87,$E4,$CF,$81,$1F,$88,$7E,$17,$77,$E3,$FB,$FF,$1D,$C1
	.dc.b	$F9,$09,$C9,$F1,$FB,$FF,$1C,$F3,$3F,$1C,$77,$3F,$1C,$78,$CF,$C7
	.dc.b	$1E,$73,$F1,$C7,$AC,$F6,$9F,$E0,$4F,$17,$E1,$3F,$CE,$71,$F8,$AF
	.dc.b	$E4,$3C,$6F,$F5,$94,$7C,$57,$FD,$69,$DB,$F9,$87,$FD,$6D,$DD,$F9
	.dc.b	$C7,$FD,$71,$DF,$FA,$07,$FD,$75,$E0,$7E,$91,$FF,$5E,$78,$3F,$A8
	.dc.b	$7F,$92,$EF,$F1,$72,$FE,$36,$7E,$20,$FC,$51,$23,$FD,$72,$FC,$63
	.dc.b	$F9,$56,$9F,$19,$FF,$0F,$F8,$5F,$C1,$1F,$F1,$0F,$87,$FC,$21,$FF
	.dc.b	$11,$F8,$9F,$C3,$1F,$F1,$2F,$8B,$FC,$41,$FF,$13,$F8,$DF,$C5,$1F
	.dc.b	$8C,$7C,$3C,$BF,$0E,$FF,$4D,$E0,$F9,$3E,$27,$F9,$C9,$F8,$D3,$FE
	.dc.b	$79,$E1,$F8,$CF,$F9,$EF,$87,$E3,$3F,$E7,$DE,$1F,$8C,$FF,$9F,$F8
	.dc.b	$7E,$33,$FE,$81,$E1,$F8,$CF,$F3,$9C,$BF,$08,$7C,$1E,$1F,$E9,$BC
	.dc.b	$9F,$8A,$3F,$8B,$46,$8F,$E9,$D3,$C9,$FE,$ED,$5D,$9F,$EF,$17,$E2
	.dc.b	$7F,$BD,$61,$E6,$7F,$BE,$63,$EA,$7F,$BF,$65,$EE,$7F,$10,$7F,$16
	.dc.b	$F1,$BF,$8A,$BF,$CC,$79,$FF,$63,$A7,$F6,$27,$F6,$91,$FB,$63,$F8
	.dc.b	$27,$8B,$FC,$51,$FE,$0D,$E2,$FF,$14,$7F,$84,$78,$BF,$C5,$1F,$E1
	.dc.b	$5E,$2F,$F1,$47,$F8,$67,$8B,$FC,$51,$FE,$1D,$E2,$FF,$14,$7F,$15
	.dc.b	$97,$F6,$AD,$BF,$A0,$79,$1F,$DA,$BC,$9D,$57,$F7,$3B,$E4,$FF,$75
	.dc.b	$C3,$B3,$FD,$DF,$1F,$13,$FD,$E7,$2F,$33,$FD,$EF,$3F,$53,$FD,$F7
	.dc.b	$4F,$E6,$2F,$F2,$1F,$03,$E1,$F2,$7F,$39,$8F,$C5,$1F,$8D,$3E,$59
	.dc.b	$F9,$03,$F6,$FB,$7F,$5A,$E7,$FF,$5F,$F4,$7E,$5E,$31,$E4,$BA,$BF
	.dc.b	$39,$18,$EB,$FD,$03,$AE,$CF,$D2,$3C,$3B,$7F,$50,$FF,$85,$BB,$BF
	.dc.b	$13,$9F,$F8,$57,$B7,$F1,$8F,$E3,$F5,$FF,$06,$CF,$07,$FD,$AD,$7D
	.dc.b	$1F,$F6,$C6,$1E,$07,$FD,$B5,$8F,$91,$FF,$6E,$65,$E9,$7F,$DB,$DA
	.dc.b	$7E,$BF,$3F,$F0,$26,$5F,$0F,$5F,$D6,$3F,$E2,$CE,$9F,$C6,$C7,$F7
	.dc.b	$48,$FE,$99,$AF,$F2,$B6,$90,$F5,$16,$BF,$4E,$DC,$3F,$C4,$F3,$C6
	.dc.b	$36,$49,$DB,$67,$6F,$13,$FC,$1D,$3F,$8C,$E8,$91,$9B,$F3,$3C,$4D
	.dc.b	$99,$F0,$3F,$A0,$39,$8D,$E1,$E0,$26,$1E,$09,$D5,$70,$9C,$C7,$43
	.dc.b	$38,$7F,$AB,$27,$55,$C6,$B7,$7D,$A3,$04,$BE,$53,$AA,$E3,$18,$9D
	.dc.b	$57,$18,$C1,$1C,$12,$78,$B8,$F8,$66,$7F,$E3,$16,$D7,$C3,$C7,$E9
	.dc.b	$63,$E1,$6E,$3E,$0C,$79,$92,$33,$5E,$18,$C3,$CC,$FC,$16,$7F,$88
	.dc.b	$87,$C5,$7F,$8F,$C9,$3F,$E1,$5F,$88,$FF,$22,$7E,$A2,$F4,$EF,$C1
	.dc.b	$E3,$DB,$CF,$C2,$6B,$DC,$DF,$C2,$F0,$F7,$78,$FC,$37,$2F,$7B,$9F
	.dc.b	$C3,$F4,$F7,$FA,$FB,$1D,$BD,$FF,$F1,$47,$0B,$AB,$FD,$9E,$FF,$5F
	.dc.b	$E7,$FF,$7B,$3F,$E3,$CE,$25,$72,$8C,$61,$D9,$D6,$5E,$27,$86,$9E
	.dc.b	$67,$96,$DE,$A7,$A7,$07,$B9,$ED,$C5,$FE,$63,$E3,$F8,$B1,$FE,$97
	.dc.b	$E4,$FC,$67,$3F,$D0,$DF,$EF,$37,$9B,$E8,$D7,$E4,$8F,$A4,$52,$74
	.dc.b	$7D,$D7,$F4,$DD,$3F,$82,$51,$FC,$7E,$30,$EE,$38,$F1,$7E,$0E,$FC
	.dc.b	$04,$8E,$F3,$B2,$4F,$EA,$7A,$7F,$05,$5F,$EB,$2A,$FF,$59,$E3,$FB
	.dc.b	$66,$E3,$FB,$46,$CE,$DF,$D9,$18,$31,$3C,$5C,$90,$FD,$A1,$B3,$C0
	.dc.b	$ED,$C9,$B7,$04,$9F,$14,$67,$1F,$87,$87,$FB,$3A,$91,$89,$E6,$48
	.dc.b	$E5,$7E,$98,$C5,$F4,$6C,$66,$77,$1E,$C9,$D4,$62,$3D,$31,$82,$4F
	.dc.b	$17,$F8,$DF,$A4,$F5,$7D,$91,$AA,$E9,$5B,$23,$82,$1D,$9C,$B4,$C9
	.dc.b	$F2,$C8,$76,$76,$E0,$FD,$66,$1F,$17,$0F,$8F,$D9,$FC,$FE,$75,$1E
	.dc.b	$D6,$DF,$9D,$F0,$10,$CD,$F6,$7C,$07,$90,$A3,$1F,$84,$FC,$E3,$45
	.dc.b	$3A,$1C,$0E,$A6,$AB,$C8,$E4,$87,$DD,$F3,$1D,$DF,$D4,$23,$FA,$C9
	.dc.b	$C8,$EA,$77,$30,$43,$9C,$F6,$74,$7F,$4F,$AF,$4C,$64,$98,$7C,$59
	.dc.b	$E4,$70,$24,$6B,$0F,$8D,$18,$CB,$E9,$1F,$4E,$B9,$47,$89,$AF,$26
	.dc.b	$D7,$38,$F8,$91,$83,$F1,$F9,$FF,$03,$8F,$E0,$3A,$57,$F8,$39,$FD
	.dc.b	$17,$0F,$E0,$CE,$2F,$F8,$CB,$C0,$FC,$3E,$9F,$D5,$4D,$ED,$FC,$2A
	.dc.b	$39,$1B,$9E,$2F,$3C,$7F,$90,$CE,$E7,$FC,$03,$B7,$F8,$63,$A1,$F1
	.dc.b	$1F,$D3,$89,$87,$F3,$DF,$80,$FE,$A3,$8F,$F5,$2F,$81,$FE,$A7,$F0
	.dc.b	$3F,$D5,$3E,$0B,$FA,$AF,$C1,$7F,$56,$F8,$36,$3F,$F0,$03,$F5,$2F
	.dc.b	$E3,$9B,$FF,$91,$D7,$C2,$EA,$7F,$96,$F1,$FF,$09,$97,$C2,$E6,$F8
	.dc.b	$7C,$35,$AD,$7C,$35,$C3,$FA,$B6,$0F,$87,$C7,$2A,$1F,$0F,$85,$F1
	.dc.b	$1F,$2E,$FC,$F9,$FF,$3D,$E8,$D4,$EE,$F8,$63,$C7,$2E,$5A,$73,$D7
	.dc.b	$A6,$DD,$73,$F8,$47,$F6,$5E,$0F,$8C,$FF,$67,$E2,$F9,$4F,$F6,$9E
	.dc.b	$4F,$EF,$BC,$BF,$42,$3E,$19,$FD,$C7,$9F,$EA,$BF,$E4,$FF,$23,$F1
	.dc.b	$5C,$3F,$6E,$FE,$EF,$26,$09,$5F,$84,$3F,$05,$D5,$F8,$88,$FF,$13
	.dc.b	$F7,$FE,$DA,$7E,$0C,$7F,$39,$8C,$47,$03,$F1,$95,$D9,$E9,$7E,$1B
	.dc.b	$E9,$18,$BF,$EB,$98,$7F,$43,$C7,$D3,$3E,$CE,$64,$63,$0F,$F4,$53
	.dc.b	$6F,$86,$7F,$90,$BF,$F0,$77,$4F,$E4,$1F,$F0,$87,$4F,$E4,$1F,$F0
	.dc.b	$97,$4F,$D1,$7F,$21,$C3,$F5,$5F,$C9,$71,$FD,$97,$F2,$9C,$BF,$75
	.dc.b	$FC,$B7,$3F,$E0,$DF,$37,$6F,$F3,$5E,$AF,$C3,$C7,$E6,$FA,$7F,$06
	.dc.b	$FE,$77,$A7,$F0,$6F,$E7,$FA,$7F,$06,$FE,$84,$FC,$19,$F0,$78,$5D
	.dc.b	$3F,$84,$8C,$74,$FE,$11,$D7,$47,$E1,$1F,$F1,$77,$5F,$E1,$7B,$3F
	.dc.b	$D1,$FA,$3E,$2B,$F3,$84,$AF,$D8,$7B,$9F,$9F,$F8,$2F,$D0,$7C,$1F
	.dc.b	$E8,$7E,$03,$F4,$5E,$0F,$E8,$FE,$07,$F4,$9E,$A7,$E9,$7C,$6F,$D3
	.dc.b	$7A,$BF,$A7,$F8,$2F,$D4,$7A,$FF,$A9,$F8,$3F,$D5,$7B,$3F,$AE,$FD
	.dc.b	$4D,$E2,$FE,$A3,$87,$F1,$67,$F1,$82,$F8,$FF,$D7,$9F,$9E,$62,$78
	.dc.b	$BF,$C1,$EB,$B3,$FC,$27,$0F,$13,$FC,$2F,$2F,$33,$FC,$37,$4F,$53
	.dc.b	$FC,$3F,$6F,$73,$FC,$47,$83,$E1,$3F,$10,$FE,$29,$C7,$F1,$D7,$2F
	.dc.b	$FA,$EB,$37,$B7,$ED,$24,$F9,$FC,$DF,$81,$E5,$FC,$1F,$84,$F8,$5E
	.dc.b	$F7,$C3,$F8,$5F,$89,$F0,$1F,$17,$E1,$BE,$37,$C1,$7C,$7F,$4F,$E4
	.dc.b	$7C,$1F,$C9,$F8,$7F,$95,$F0,$BF,$2F,$E2,$3E,$67,$C3,$FC,$BE,$0F
	.dc.b	$C9,$F3,$7E,$57,$D6,$FE,$55,$E7,$EA,$7E,$55,$F2,$7E,$66,$3D,$9F
	.dc.b	$9D,$9F,$89,$FA,$1A,$F9,$9F,$A5,$BF,$A9,$FA,$9C,$3E,$E7,$E0,$1F
	.dc.b	$5F,$93,$E1,$BF,$23,$D7,$FE,$71,$D1,$F2,$F5,$FE,$7C,$FF,$2F,$F1
	.dc.b	$BE,$8D,$7F,$01,$F1,$DD,$7F,$5A,$31,$C3,$F6,$9D,$76,$7D,$E7,$87
	.dc.b	$6F,$E0,$9E,$5D,$DF,$86,$7A,$77,$FE,$4F,$C1,$FF,$41,$7B,$1F,$C6
	.dc.b	$B6,$FC,$6F,$85,$FE,$9A,$7F,$D9,$FE,$0F,$E4,$30,$E2,$7F,$57,$F0
	.dc.b	$FF,$3E,$FF,$59,$F1,$3F,$46,$FF,$5B,$F1,$7F,$4E,$FF,$5D,$F1,$BF
	.dc.b	$56,$FF,$5F,$F1,$FF,$5E,$FF,$61,$F2,$3E,$13,$F9,$02,$1F,$24,$FF
	.dc.b	$57,$63,$F9,$07,$E7,$10,$FE,$50,$FF,$8E,$5F,$D4,$78,$DF,$91,$7F
	.dc.b	$55,$E3,$7E,$45,$FD,$67,$8D,$F9,$17,$F5,$DE,$37,$E4,$5F,$D8,$78
	.dc.b	$DF,$91,$7F,$65,$E3,$7E,$45,$FD,$A7,$8D,$F9,$1E,$1F,$EC,$91,$FB
	.dc.b	$AF,$47,$F9,$27,$8F,$F9,$18,$F2,$F8,$FF,$91,$7C,$FE,$3F,$E4,$5F
	.dc.b	$4F,$8F,$F9,$17,$D7,$E3,$FE,$45,$F6,$F8,$FF,$91,$7D,$FE,$3F,$E4
	.dc.b	$5F,$83,$E4,$7E,$43,$4F,$B7,$C9,$F1,$E7,$E5,$75,$FF,$7F,$F1,$BF
	.dc.b	$25,$DD,$F9,$49,$FE,$00,$7E,$47,$A3,$EA,$BF,$93,$E9,$FB,$2F,$E5
	.dc.b	$7A,$BE,$EB,$F9,$7E,$BF,$C0,$BF,$99,$EC,$FC,$2B,$F9,$BE,$DF,$E4
	.dc.b	$6F,$F1,$BC,$7E,$41,$FE,$A3,$E6,$FF,$26,$F8,$78,$9F,$BF,$3C,$1E
	.dc.b	$47,$F1,$F3,$F2,$07,$F1,$F3,$F2,$07,$F1,$F3,$F2,$07,$F1,$F3,$F2
	.dc.b	$07,$F1,$F3,$F2,$07,$C0,$DF,$F6,$6F,$E4,$3C,$9F,$E3,$DE,$4F,$CB
	.dc.b	$AF,$F5,$C7,$0F,$CB,$BF,$9B,$C5,$F4,$5F,$9F,$C7,$F5,$5F,$A3,$C9
	.dc.b	$F6,$5F,$A7,$CB,$F7,$5F,$AB,$CD,$F8,$17,$EB,$F3,$FE,$15,$FB,$3D
	.dc.b	$1F,$89,$7F,$8D,$ED,$F8,$DD,$FF,$8F,$6B,$FE,$43,$E9,$D3,$E6,$BE
	.dc.b	$4F,$A3,$0E,$CF,$AB,$1F,$13,$EC,$CB,$CC,$FB,$B3,$F5,$3F,$03,$4F
	.dc.b	$73,$F0,$B5,$F8,$9A,$7E,$D3,$7F,$F2,$B7,$07,$F7,$0E,$8F,$C6,$CF
	.dc.b	$F9,$41,$FD,$EB,$F8,$D3,$E9,$1F,$8D,$3E,$B1,$F8,$D3,$ED,$1F,$8D
	.dc.b	$3E,$F1,$F8,$D3,$F0,$47,$E1,$0F,$E2,$A7,$F0,$2B,$FE,$3B,$CB,$FE
	.dc.b	$AD,$E1,$FE,$37,$1F,$8B,$7F,$CD,$DC,$9F,$96,$7F,$CE,$1C,$BF,$9A
	.dc.b	$7F,$CE,$55,$F9,$E7,$FC,$E9,$CD,$FA,$27,$FC,$ED,$CF,$FA,$67,$FC
	.dc.b	$F1,$D1,$EE,$7F,$17,$97,$F3,$BC,$FE,$2F,$93,$F8,$AD,$BE,$4F,$53
	.dc.b	$A7,$F1,$9B,$FC,$DC,$3F,$8C,$9F,$D3,$8F,$E3,$27,$F5,$23,$F8,$C9
	.dc.b	$FD,$58,$FE,$32,$7F,$5A,$3F,$8C,$9F,$D7,$8F,$E3,$27,$F6,$23,$F8
	.dc.b	$C9,$FE,$0E,$3F,$B4,$9F,$8D,$31,$8F,$F0,$AE,$BF,$E4,$11,$FC,$BB
	.dc.b	$B3,$F9,$23,$FC,$C3,$B7,$F9,$43,$FC,$CB,$BB,$F9,$63,$FC,$D3,$BF
	.dc.b	$F9,$83,$FC,$DB,$C0,$FE,$68,$FF,$38,$F0,$7F,$C4,$DD,$1F,$13,$5F
	.dc.b	$E3,$1D,$1F,$E0,$FF,$89,$FF,$15,$12,$FF,$C4,$9F,$13,$FE,$2D,$F8
	.dc.b	$AF,$F1,$77,$B9,$FE,$2F,$F7,$7F,$D3,$BC,$2F,$7A,$FC,$4E,$3E,$03
	.dc.b	$3F,$13,$AF,$82,$DF,$C4,$F0,$F8,$4E,$3F,$13,$CB,$E1,$B9,$FC,$4F
	.dc.b	$4F,$88,$FD,$8F,$17,$F8,$E4,$FF,$3C,$6C,$8D,$C7,$F8,$F2,$15,$D2
	.dc.b	$31,$87,$81,$D6,$5E,$47,$86,$9E,$87,$96,$DE,$D7,$FC,$89,$C1,$F0
	.dc.b	$5F,$86,$7F,$91,$F8,$FE,$33,$FD,$03,$95,$18,$ED,$FE,$8A,$FF,$95
	.dc.b	$79,$79,$57,$F6,$AB,$EC,$FB,$30,$F1,$3E,$EC,$7C,$CF,$C0,$CB,$D4
	.dc.b	$FC,$2C,$FD,$CF,$C4,$D3,$E1,$BF,$1B,$6F,$93,$DF,$FE,$5F,$8C,$1F
	.dc.b	$1F,$83,$E8,$3F,$27,$87,$EA,$3F,$2F,$8B,$EC,$3F,$37,$8F,$EE,$3F
	.dc.b	$3F,$93,$F0,$3B,$7A,$B8,$BF,$B1,$F0,$7B,$9F,$EC,$BC,$3F,$0F,$BF
	.dc.b	$F9,$EF,$1F,$ED,$B1,$6C,$E5,$78,$3F,$8C,$24,$FD,$DF,$89,$FC,$7F
	.dc.b	$B7,$F9,$0F,$89,$FC,$8F,$BB,$F9,$2F,$89,$FC,$9F,$BF,$F9,$4F,$89
	.dc.b	$FC,$AF,$C0,$FE,$5B,$E2,$7F,$DE,$7F,$AA,$7F,$BF,$F4,$FF,$61,$2F
	.dc.b	$8A,$CF,$C3,$6B,$E2,$B7,$F1,$1F,$D9,$F8,$FF,$C7,$1C,$BF,$ED,$5F
	.dc.b	$53,$18,$AB,$86,$5F,$E8,$E8,$63,$E0,$CB,$C3,$3E,$33,$F3,$1A,$79
	.dc.b	$DF,$CD,$6D,$EB,$7F,$39,$C1,$EF,$7F,$3D,$C5,$FB,$0E,$3F,$8B,$1F
	.dc.b	$A4,$E4,$F9,$9B,$FF,$1F,$E6,$FA,$13,$FE,$3C,$7F,$24,$FF,$A7,$FC
	.dc.b	$7F,$DD,$BF,$EA,$1F,$1F,$ED,$BF,$74,$FD,$29,$F8,$13,$F5,$27,$E1
	.dc.b	$4F,$D6,$9F,$89,$3F,$65,$E1,$FE,$4B,$AF,$FC,$BF,$D5,$FB,$79,$FE
	.dc.b	$40,$7F,$22,$3F,$26,$7E,$F3,$C2,$FC,$9B,$FB,$EF,$0B,$F2,$6F,$F0
	.dc.b	$0F,$0B,$F2,$6F,$F0,$2F,$0B,$F2,$6F,$F0,$4F,$0B,$D2,$FF,$95,$B3
	.dc.b	$F6,$BF,$E5,$8D,$3E,$0B,$FE,$5A,$D7,$F7,$3E,$47,$F2,$18,$FD,$E1
	.dc.b	$0F,$F2,$FF,$83,$F9,$BE,$CF,$F0,$AF,$85,$F9,$F5,$3F,$9A,$DF,$F9
	.dc.b	$13,$F2,$4F,$DD,$1F,$24,$FD,$D1,$F2,$4F,$DD,$1F,$24,$FD,$D1,$FB
	.dc.b	$53,$F8,$E7,$89,$FB,$AF,$7D,$EF,$FF,$25,$9D,$1C,$0F,$90,$7D,$43
	.dc.b	$E4,$1F,$60,$F9,$07,$DC,$3E,$41,$FC,$80,$FC,$11,$FC,$80,$FC,$31
	.dc.b	$FC,$80,$F9,$07,$C5,$CB,$E4,$38,$27,$1F,$F0,$07,$89,$FC,$7C,$EA
	.dc.b	$7F,$1F,$3E,$33,$F8,$6F,$27,$F8,$FB,$F8,$8F,$27,$F8,$FB,$F8,$AF
	.dc.b	$27,$F8,$FB,$F8,$CE,$7F,$C7,$6B,$F4,$BA,$3F,$85,$6B,$F1,$F4,$F8
	.dc.b	$D9,$FE,$34,$FC,$A9,$F3,$F0,$E8,$FD,$37,$C0,$FD,$77,$C8,$F9,$AF
	.dc.b	$B6,$DF,$7A,$3F,$1C,$7C,$73,$F1,$C7,$C7,$9F,$E8,$2F,$C7,$E9,$E2
	.dc.b	$7E,$E2,$3F,$4A,$FE,$EE,$3F,$7A,$7F,$5B,$F2,$3F,$8E,$BF,$D7,$7C
	.dc.b	$8F,$E3,$AF,$F5,$FF,$23,$F8,$EB,$FD,$87,$C8,$FE,$1A,$FF,$11,$F0
	.dc.b	$FF,$8D,$BF,$C4,$FC,$8F,$E3,$6F,$F8,$43,$4F,$E3,$4F,$E6,$B3,$FF
	.dc.b	$6A,$FC,$07,$FB,$5E,$7F,$CC,$13,$FC,$6F,$C8,$FE,$3A,$FF,$28,$F1
	.dc.b	$FF,$8E,$3F,$C7,$CF,$E3,$87,$F7,$6F,$1F,$F8,$E3,$FD,$E3,$C7,$FE
	.dc.b	$38,$FF,$7A,$F1,$FF,$8E,$3F,$DF,$3C,$7F,$84,$FF,$1B,$3F,$BE,$71
	.dc.b	$7F,$1C,$F2,$FF,$B6,$C7,$ED,$75,$F9,$AF,$ED,$8F,$C7,$BF,$D4,$9F
	.dc.b	$C7,$BF,$C4,$BC,$BF,$ED,$AF,$DC,$8F,$D1,$9F,$B4,$3F,$4E,$7E,$D0
	.dc.b	$FD,$59,$FA,$D7,$F8,$DF,$91,$F8,$C7,$FD,$A7,$1F,$C6,$DF,$E3,$FD
	.dc.b	$1F,$94,$8F,$E0,$9D,$5F,$98,$E0,$FC,$4F,$5F,$E7,$1F,$C5,$F6,$7E
	.dc.b	$81,$FC,$6F,$6F,$99,$FC,$7E,$7F,$C3,$F0,$FF,$7A,$F8,$1F,$C4,$DF
	.dc.b	$F7,$BF,$83,$FC,$FB,$BF,$FA,$04,$7F,$14,$F0,$3F,$B5,$BF,$DB,$BC
	.dc.b	$A7,$C5,$7F,$62,$24,$7F,$57,$6B,$FD,$91,$BC,$0C,$FF,$5D,$5C,$AB
	.dc.b	$81,$0E,$05,$9E,$0F,$86,$B9,$D7,$26,$C6,$4F,$E3,$E7,$F4,$69,$F9
	.dc.b	$A7,$CC,$7A,$4F,$16,$9C,$4F,$09,$23,$C3,$82,$FD,$13,$9A,$E4,$D3
	.dc.b	$95,$70,$9D,$B6,$FC,$F7,$CD,$F8,$F8,$7C,$E8,$C5,$F9,$1A,$F5,$6D
	.dc.b	$73,$8D,$5F,$6C,$3C,$73,$99,$E1,$39,$7C,$B3,$A2,$38,$25,$7B,$FE
	.dc.b	$2B,$F9,$14,$7F,$84,$DF,$4D,$73,$9F,$1D,$72,$6D,$72,$9D,$13,$1E
	.dc.b	$24,$8F,$4B,$F0,$B9,$3F,$0B,$F0,$5F,$E2,$E2,$4F,$F8,$8F,$D7,$FF
	.dc.b	$18,$FC,$4F,$F8,$CB,$D9,$FF,$19,$FC,$47,$F8,$D3,$DB,$FF,$1A,$F9
	.dc.b	$3F,$E3,$6F,$73,$FC,$6F,$F1,$1F,$E3,$8F,$57,$FC,$73,$F1,$1F,$E3
	.dc.b	$AF,$7F,$FC,$77,$F1,$1F,$E3,$CF,$81,$FF,$1E,$FC,$47,$ED,$79,$BF
	.dc.b	$8E,$F0,$BA,$3F,$8E,$E5,$98,$FD,$25,$71,$7F,$C7,$18,$75,$7F,$C7
	.dc.b	$59,$78,$5F,$F1,$E6,$7E,$57,$FC,$7D,$AF,$A5,$FF,$20,$6F,$ED,$7F
	.dc.b	$C8,$5C,$3F,$E5,$0E,$2F,$87,$1F,$E5,$0E,$3F,$F1,$BF,$37,$CD,$8F
	.dc.b	$96,$43,$F0,$3C,$DF,$3D,$A6,$77,$FC,$31,$38,$3E,$D5,$3A,$3F,$03
	.dc.b	$5F,$E5,$E7,$95,$6E,$3E,$F5,$E2,$F8,$B6,$32,$77,$8F,$D3,$F4,$FE
	.dc.b	$7F,$0F,$D5,$CF,$F8,$88,$FD,$91,$7A,$7F,$42,$D8,$DB,$FB,$03,$17
	.dc.b	$FE,$73,$70,$43,$99,$67,$9B,$67,$9B,$A2,$46,$9B,$1F,$0D,$FE,$19
	.dc.b	$7F,$E0,$7F,$4B,$FC,$57,$1F,$A5,$3F,$8D,$1E,$43,$C9,$5C,$C9,$5E
	.dc.b	$06,$DF,$A6,$7B,$B6,$33,$1C,$A7,$81,$27,$4D,$8F,$F2,$5F,$B6,$F8
	.dc.b	$6F,$75,$68,$93,$CD,$AF,$FA,$6E,$FB,$B6,$FA,$36,$76,$FB,$4C,$1C
	.dc.b	$C8,$60,$E5,$A7,$E2,$B4,$FE,$48,$FF,$46,$8E,$44,$3C,$E7,$23,$07
	.dc.b	$C4,$30,$68,$EC,$6A,$FD,$38,$FC,$18,$E0,$68,$E0,$6F,$0F,$6B,$70
	.dc.b	$F6,$BA,$21,$C8,$FC,$A7,$A9,$FE,$93,$7F,$82,$69,$FC,$1B,$2F,$E0
	.dc.b	$B1,$FD,$48,$FE,$96,$75,$3D,$27,$33,$AE,$3F,$41,$D1,$E7,$34,$7F
	.dc.b	$55,$C7,$D9,$1B,$8E,$38,$FA,$9B,$8F,$A9,$E3,$3B,$6C,$FE,$2B,$2F
	.dc.b	$A9,$8F,$F7,$47,$EA,$4F,$23,$CF,$5C,$49,$5D,$E3,$9D,$F7,$9F,$EA
	.dc.b	$F8,$79,$AB,$F8,$A5,$7F,$10,$26,$3F,$C2,$6B,$FA,$39,$27,$71,$96
	.dc.b	$E3,$FC,$32,$BF,$AF,$E3,$E8,$7F,$C2,$8F,$F0,$FB,$E5,$8D,$BE,$35
	.dc.b	$FC,$48,$F0,$DF,$37,$C9,$AF,$F7,$1F,$D5,$7F,$C1,$F1,$F7,$06,$08
	.dc.b	$CE,$5F,$E0,$46,$DF,$FA,$25,$7C,$27,$F1,$B9,$D6,$96,$B5,$EF,$DC
	.dc.b	$7A,$E0,$F8,$5C,$65,$DF,$E0,$6C,$BE,$1A,$FC,$1E,$3E,$1F,$3F,$0B
	.dc.b	$AE,$5D,$FC,$37,$0F,$88,$E3,$F1,$1C,$BE,$27,$AF,$37,$F1,$BE,$0F
	.dc.b	$F2,$87,$17,$F1,$BE,$FC,$F4,$EE,$7F,$CF,$93,$ED,$C3,$A4,$60,$F4
	.dc.b	$E5,$E1,$71,$C5,$F1,$F5,$F3,$46,$38,$5C,$1E,$CE,$1F,$6F,$17,$BB
	.dc.b	$8F,$DF,$C9,$F0,$39,$7E,$0F,$3F,$C2,$DE,$67,$FE,$55,$F2,$BE,$D7
	.dc.b	$47,$E1,$FA,$BF,$00,$49,$FE,$A0,$7E,$06,$34,$FF,$8B,$7A,$FF,$75
	.dc.b	$1F,$B8,$8F,$C0,$70,$BD,$BF,$F4,$D4,$7F,$BD,$7D,$3E,$8F,$E2,$8E
	.dc.b	$99,$78,$88,$E0,$95,$FE,$3A,$ED,$FC,$74,$FA,$A7,$BF,$1F,$E2,$9F
	.dc.b	$C1,$EB,$F0,$5F,$91,$BF,$C5,$7E,$57,$0F,$C9,$7E,$67,$1F,$CE,$79
	.dc.b	$F2,$7D,$37,$AF,$2F,$D8,$7B,$F3,$7D,$D7,$FB,$E1,$FD,$70,$FC,$FF
	.dc.b	$77,$F5,$C7,$27,$9B,$AF,$F1,$6F,$DB,$EC,$FC,$7B,$F7,$7B,$7F,$26
	.dc.b	$FD,$FE,$EF,$CB,$BF,$81,$EB,$FB,$EF,$E0,$FA,$FE,$FB,$F8,$5E,$BF
	.dc.b	$BE,$FE,$1F,$8F,$E3,$72,$7E,$D7,$A7,$F4,$64,$9F,$C6,$7C,$17,$E9
	.dc.b	$BC,$EF,$D3,$FC,$1F,$EA,$3D,$0F,$D4,$FC,$27,$EB,$3F,$56,$FF,$78
	.dc.b	$77,$FF,$9F,$17,$C3,$67,$D5,$D7,$C3,$EF,$D7,$E1,$F1,$1C,$7D,$9E
	.dc.b	$5F,$13,$CF,$DB,$FD,$3F,$0F,$C9,$F1,$5D,$BF,$27,$6F,$EB,$F3,$FA
	.dc.b	$4D,$97,$CA,$31,$8F,$67,$59,$F8,$9E,$1A,$FA,$76,$F5,$7B,$7E,$C9
	.dc.b	$FD,$67,$0F,$C0,$7F,$5D,$C7,$FC,$23,$93,$E3,$47,$E3,$B9,$7F,$C5
	.dc.b	$DC,$DF,$E8,$FD,$9E,$E7,$C5,$25,$79,$FC,$8F,$91,$E5,$FC,$9F,$8A
	.dc.b	$F9,$5E,$FF,$CB,$F8,$AF,$99,$F0,$3F,$37,$E2,$BE,$77,$C1,$FC,$FF
	.dc.b	$8A,$FA,$1F,$0B,$F4,$7E,$2B,$F8,$EF,$C0,$7F,$8F,$F6,$FD,$7A,$F8
	.dc.b	$AC,$FC,$46,$BE,$2B,$F9,$F7,$82,$A4,$6A,$FF,$87,$C7,$16,$9D,$CF
	.dc.b	$A1,$8F,$8D,$FA,$59,$F9,$DF,$A9,$AF,$AD,$FA,$DB,$FB,$DF,$B1,$C3
	.dc.b	$F8,$9E,$2F,$8B,$1F,$C1,$B9,$3E,$4B,$FC,$4C,$FE,$77,$CD,$F4,$3C
	.dc.b	$0F,$E8,$CF,$D3,$8F,$AA,$7F,$23,$F1,$BF,$2C,$FF,$25,$F1,$BE,$EB
	.dc.b	$F8,$7E,$EF,$CA,$3F,$89,$F1,$BF,$28,$FE,$2F,$C6,$FC,$A3,$F8,$DF
	.dc.b	$1B,$F2,$8F,$F9,$4C,$FE,$37,$EB,$FF,$1A,$E5,$FE,$39,$C9,$F9,$89
	.dc.b	$F5,$78,$DF,$93,$7D,$9E,$37,$E4,$DF,$77,$8D,$F9,$37,$E0,$78,$DF
	.dc.b	$93,$7F,$39,$E3,$7E,$4D,$FE,$37,$E3,$7E,$4D,$FE,$3B,$E3,$7C,$17
	.dc.b	$F2,$47,$F4,$3E,$3F,$C9,$F3,$A3,$FB,$7F,$37,$CC,$7E,$77,$89,$F4
	.dc.b	$39,$FF,$71,$D1,$F9,$49,$FD,$67,$89,$F9,$47,$F5,$DE,$27,$E5,$1F
	.dc.b	$D8,$78,$9F,$94,$7F,$65,$E2,$7E,$51,$FD,$A7,$89,$F9,$47,$F2,$BE
	.dc.b	$27,$E3,$5F,$CA,$F8,$7F,$92,$E7,$FC,$8C,$7E,$43,$C4,$DB,$F9,$03
	.dc.b	$99,$F9,$03,$B9,$F9,$03,$C6,$7E,$40,$F3,$9F,$90,$3D,$67,$E4,$0F
	.dc.b	$D9,$1F,$C2,$76,$FC,$8F,$8B,$FC,$8E,$3E,$51,$0F,$E9,$7E,$37,$F2
	.dc.b	$37,$F1,$C7,$F0,$63,$F1,$C7,$D8,$3F,$1C,$7F,$0E,$3F,$1C,$7F,$12
	.dc.b	$3F,$1C,$7F,$16,$3F,$1C,$7F,$1B,$EA,$FC,$57,$0B,$C8,$FF,$05,$4F
	.dc.b	$F2,$3F,$6F,$FC,$1D,$3F,$CB,$EB,$F1,$CF,$F3,$1E,$CF,$C7,$3F,$CC
	.dc.b	$FB,$3F,$1C,$FF,$35,$EC,$FC,$76,$1F,$AF,$EC,$FC,$74,$FF,$8C,$7B
	.dc.b	$3F,$1C,$FF,$8C,$FB,$3F,$1C,$FE,$C2,$BE,$26,$DF,$C7,$F7,$FC,$56
	.dc.b	$DF,$8F,$ED,$75,$FE,$3F,$0F,$C6,$E1,$F4,$4F,$DF,$75,$7E,$35,$FE
	.dc.b	$01,$D5,$F8,$D7,$F8,$17,$57,$E3,$5F,$E0,$9D,$5F,$8D,$7F,$83,$75
	.dc.b	$7E,$35,$FE,$2F,$D5,$F8,$DC,$7E,$44,$7F,$93,$BA,$7F,$1A,$FF,$8D
	.dc.b	$0F,$EE,$7E,$DF,$F2,$37,$F2,$07,$23,$F2,$07,$63,$F2,$07,$88,$FC
	.dc.b	$81,$E6,$3F,$20,$7A,$8F,$C8,$1E,$E3,$E0,$4F,$C2,$7F,$62,$7E,$44
	.dc.b	$FE,$3D,$AF,$F9,$8F,$BB,$F8,$F6,$FF,$E6,$17,$E8,$BF,$E6,$3E,$2F
	.dc.b	$AA,$FF,$99,$78,$FE,$CB,$F4,$78,$7F,$B9,$B7,$C3,$FE,$3F,$8F,$F2
	.dc.b	$DF,$0F,$F8,$FB,$FC,$BF,$C3,$FE,$3E,$FF,$31,$F0,$FF,$8F,$BF,$E1
	.dc.b	$6D,$7F,$C7,$44,$C3,$FD,$39,$1F,$97,$3F,$32,$7F,$1E,$3F,$3A,$7F
	.dc.b	$1E,$3F,$42,$7F,$1E,$3F,$4A,$7E,$44,$FD,$49,$FC,$78,$FD,$69,$FC
	.dc.b	$78,$FF,$09,$9F,$B3,$8F,$C9,$2E,$7F,$C9,$3F,$2C,$F9,$A7,$F4,$BF
	.dc.b	$07,$F8,$E3,$F4,$CF,$E9,$DE,$37,$F5,$57,$FA,$87,$8F,$FD,$65,$FE
	.dc.b	$A5,$E4,$7F,$5D,$7F,$AA,$79,$3F,$D8,$5F,$ED,$A7,$F1,$63,$FB,$7F
	.dc.b	$89,$F8,$B7,$F1,$DC,$9F,$E1,$8E,$5C,$69,$F9,$48,$FE,$E1,$5C,$63
	.dc.b	$1C,$DF,$DE,$5F,$F1,$57,$47,$E7,$DF,$F1,$67,$4F,$E8,$DF,$F1,$77
	.dc.b	$57,$E9,$DF,$F1,$87,$5F,$EA,$DF,$F1,$97,$67,$C0,$7F,$20,$7F,$8D
	.dc.b	$7B,$7F,$8F,$F8,$DF,$B7,$C9,$F1,$5F,$E0,$A2,$6F,$FE,$02,$F6,$FF
	.dc.b	$C1,$BE,$E7,$F8,$3B,$DD,$FF,$07,$FB,$DF,$E1,$0F,$7F,$FC,$23,$F1
	.dc.b	$3F,$E1,$2F,$81,$FF,$09,$FC,$4F,$F8,$53,$9F,$FC,$2B,$F1,$3F,$E1
	.dc.b	$6F,$84,$FF,$0B,$EF,$FE,$18,$F8,$5F,$F0,$CF,$C4,$FC,$6F,$0B,$F8
	.dc.b	$FF,$0F,$CA,$AF,$DC,$F2,$FC,$6A,$E1,$3F,$E1,$69,$E8,$FF,$86,$31
	.dc.b	$F0,$3F,$E1,$AC,$FC,$8F,$F8,$73,$5F,$43,$FE,$1E,$DF,$D8,$FF,$88
	.dc.b	$38,$7F,$C5,$7C,$5F,$0A,$3F,$C5,$7C,$7F,$DB,$1F,$E1,$8E,$74,$FB
	.dc.b	$2F,$08,$E4,$7F,$8A,$B8,$FE,$3B,$FE,$2C,$E3,$F8,$EF,$F8,$BB,$8F
	.dc.b	$E3,$BF,$E3,$0E,$3F,$8E,$FF,$8C,$B8,$FE,$3B,$FE,$34,$E3,$F8,$EF
	.dc.b	$F9,$97,$D7,$FF,$32,$72,$7C,$8D,$3F,$1D,$18,$9F,$E4,$5C,$9F,$21
	.dc.b	$FF,$22,$F2,$7C,$87,$FC,$8F,$C9,$F2,$1F,$F2,$4F,$27,$C8,$7F,$C9
	.dc.b	$7C,$9F,$21,$FF,$26,$F2,$7C,$87,$FC,$9F,$C9,$F0,$5F,$8E,$7F,$92
	.dc.b	$B8,$FE,$47,$6F,$F9,$3F,$E1,$FF,$CA,$E4,$9F,$F2,$7F,$C4,$FF,$96
	.dc.b	$BE,$27,$FD,$EF,$FA,$BF,$FA,$E6,$CB,$E2,$73,$EE,$EB,$E2,$77,$EF
	.dc.b	$F0,$F8,$9E,$3F,$03,$CB,$E2,$79,$FC,$1F,$4F,$89,$EB,$F0,$BD,$BE
	.dc.b	$27,$F2,$1C,$2D,$FF,$C7,$6F,$FB,$C3,$E1,$3F,$C7,$DD,$5F,$92,$EE
	.dc.b	$FE,$BD,$C4,$C3,$9D,$FF,$BD,$A1,$97,$82,$31,$A7,$91,$D6,$DE,$87
	.dc.b	$87,$07,$B1,$E5,$C5,$F0,$B2,$F8,$91,$F1,$8F,$F3,$AF,$2F,$C9,$7F
	.dc.b	$91,$78,$DF,$BB,$E7,$FC,$A5,$59,$FD,$D1,$F9,$43,$FD,$0B,$E2,$FE
	.dc.b	$51,$FF,$43,$F8,$BF,$94,$7F,$D1,$3E,$2F,$E5,$1F,$F4,$5F,$8B,$F9
	.dc.b	$47,$FD,$1B,$E2,$FE,$51,$FF,$47,$F8,$BF,$94,$7F,$90,$F8,$9F,$91
	.dc.b	$E7,$FE,$45,$1F,$EB,$BE,$67,$89,$F9,$3F,$17,$F9,$0C,$7E,$4D,$FE
	.dc.b	$47,$E2,$FE,$4D,$FE,$4B,$E2,$FE,$4D,$FE,$4F,$E2,$FE,$4D,$FE,$53
	.dc.b	$E2,$FE,$4D,$FE,$57,$E2,$FE,$4D,$FD,$AE,$5F,$93,$3F,$B7,$E3,$FC
	.dc.b	$81,$C1,$F9,$83,$FD,$53,$E3,$7F,$20,$7F,$74,$7F,$07,$3F,$74,$7F
	.dc.b	$0B,$3F,$74,$7F,$0F,$3F,$74,$7F,$13,$3F,$74,$7F,$17,$3F,$74,$7F
	.dc.b	$AB,$B3,$FC,$8E,$9F,$1A,$3F,$90,$7C,$27,$CB,$C3,$FA,$2E,$FF,$3D
	.dc.b	$FE,$8F,$C1,$F4,$DF,$E9,$3C,$3F,$5D,$FE,$97,$C5,$F6,$DF,$E9,$BC
	.dc.b	$7F,$7D,$FE,$9F,$C9,$ED,$CB,$DE,$7C,$28,$F8,$EF,$E3,$0F,$E3,$E6
	.dc.b	$79,$7F,$18,$F0,$3F,$80,$9F,$18,$FE,$0A,$7C,$63,$F8,$49,$F1,$8F
	.dc.b	$E1,$A7,$C6,$3F,$88,$9F,$18,$FE,$2A,$7C,$63,$FD,$B7,$AF,$FB,$63
	.dc.b	$97,$F1,$87,$E3,$78,$3F,$17,$3B,$BF,$C5,$BC,$DF,$C5,$BD,$DF,$C5
	.dc.b	$BE,$37,$F1,$6F,$9D,$FC,$5B,$EB,$7F,$16,$FE,$A8,$FC,$33,$F1,$BD
	.dc.b	$47,$47,$F6,$BE,$9F,$DA,$4F,$F7,$C7,$FD,$EB,$E4,$7E,$EA,$D3,$FE
	.dc.b	$03,$F0,$3F,$8F,$1F,$C7,$63,$F8,$57,$8F,$FC,$75,$FE,$19,$E3,$FF
	.dc.b	$1D,$7F,$87,$78,$FF,$C7,$5F,$E2,$1E,$3F,$F1,$D7,$F8,$97,$8F,$FC
	.dc.b	$75,$FF,$7D,$F4,$7F,$B0,$DF,$F6,$CE,$F7,$3F,$F1,$E4,$7F,$99,$E7
	.dc.b	$83,$E6,$E2,$FE,$3F,$7D,$63,$07,$F2,$7D,$18,$F8,$E3,$3D,$3F,$CE
	.dc.b	$33,$F4,$46,$29,$A7,$B6,$31,$AF,$F4,$0F,$07,$F9,$04,$7F,$B6,$75
	.dc.b	$FF,$12,$F9,$1F,$2A,$3F,$90,$90,$FC,$B6,$4D,$FF,$79,$1F,$5B,$6F
	.dc.b	$CB,$DF,$F8,$BB,$B3,$F4,$0F,$E8,$A7,$F7,$87,$F2,$23,$F7,$87,$F2
	.dc.b	$23,$F7,$87,$F2,$23,$F7,$87,$F2,$23,$F7,$87,$F2,$23,$F7,$87,$F7
	.dc.b	$5E,$AF,$F6,$F7,$3F,$F1,$FB,$FF,$18,$61,$FC,$7D,$FF,$19,$78,$3F
	.dc.b	$C7,$DF,$F1,$A7,$83,$FC,$7D,$FF,$1B,$78,$3F,$C7,$DF,$F1,$C7,$83
	.dc.b	$FC,$7D,$FF,$1D,$78,$3F,$C7,$DF,$79,$FB,$6E,$BF,$CA,$E9,$FE,$A0
	.dc.b	$3C,$84,$3F,$9B,$E9,$FA,$D2,$39,$D7,$FC,$17,$18,$8D,$99,$AF,$F0
	.dc.b	$81,$1E,$C6,$89,$3C,$9E,$06,$9C,$3B,$AF,$3E,$FF,$C8,$E7,$24,$AE
	.dc.b	$D1,$96,$F6,$7F,$04,$C3,$CD,$5B,$C3,$B5,$6A,$72,$D8,$F1,$DF,$37
	.dc.b	$64,$70,$4B,$ED,$1C,$1A,$FB,$A7,$A3,$F9,$4E,$57,$C3,$FF,$87,$09
	.dc.b	$1F,$E0,$7F,$89,$FF,$0F,$FC,$4F,$F8,$83,$DF,$FF,$10,$FC,$3F,$F8
	.dc.b	$8B,$E0,$3F,$C4,$7F,$13,$FE,$24,$F8,$1F,$F1,$2F,$C4,$FE,$E7,$5F
	.dc.b	$F7,$87,$EB,$1F,$EC,$0D,$3F,$C3,$4B,$E0,$B5,$18,$F8,$6B,$C4,$A7
	.dc.b	$C1,$E1,$7C,$3D,$F8,$4F,$EA,$BE,$BF,$EB,$71,$FF,$12,$77,$FF,$58
	.dc.b	$B5,$71,$BE,$58,$73,$CB,$A6,$7D,$7E,$1F,$D3,$97,$F8,$9F,$7F,$6B
	.dc.b	$FE,$29,$E1,$F8,$2F,$F8,$AF,$8F,$E2,$BF,$1C,$FF,$17,$73,$7C,$A7
	.dc.b	$F8,$0F,$47,$D2,$EA,$FC,$07,$17,$D7,$EA,$FB,$F5,$F6,$EB,$5C,$9F
	.dc.b	$5A,$F7,$18,$9D,$57,$0B,$E5,$1F,$7E,$31,$3B,$9C,$CF,$E0,$4E,$27
	.dc.b	$FB,$F2,$13,$FC,$C6,$3F,$99,$4F,$F7,$F6,$CE,$9D,$1F,$AA,$33,$38
	.dc.b	$70,$7F,$A9,$8C,$9C,$27,$A3,$92,$1C,$08,$E9,$B3,$E1,$7C,$D1,$92
	.dc.b	$46,$63,$B9,$A7,$0F,$68,$DC,$F9,$DB,$1E,$59,$DB,$E0,$9D,$46,$27
	.dc.b	$73,$D6,$72,$47,$93,$57,$8A,$F2,$7C,$57,$CC,$95,$BA,$F4,$90,$DB
	.dc.b	$96,$9D,$88,$78,$CA,$71,$3D,$46,$0E,$26,$C8,$6E,$FD,$AD,$9F,$2B
	.dc.b	$C4,$E6,$70,$39,$9F,$42,$FD,$CE,$8E,$55,$E9,$70,$70,$21,$82,$57
	.dc.b	$89,$CB,$5C,$D7,$91,$F1,$18,$9E,$D3,$B2,$4E,$E7,$CD,$7D,$23,$11
	.dc.b	$6B,$8C,$62,$3A,$1F,$06,$B9,$CE,$89,$19,$9E,$13,$A8,$C4,$E9,$F3
	.dc.b	$C6,$5B,$3F,$9D,$BF,$93,$3F,$B8,$C3,$F8,$0D,$FE,$F0,$97,$FB,$F2
	.dc.b	$1F,$BD,$9E,$18,$FF,$09,$24,$7F,$00,$35,$87,$F0,$52,$57,$F3,$07
	.dc.b	$FA,$3F,$6F,$F5,$EE,$57,$C4,$7F,$31,$25,$FF,$88,$70,$FE,$7B,$F0
	.dc.b	$DF,$CD,$7E,$27,$F9,$B7,$B9,$FC,$DF,$E1,$FF,$9C,$7A,$5F,$CE,$7D
	.dc.b	$EF,$E7,$5F,$0F,$FC,$EF,$E0,$3F,$9E,$7C,$3F,$F3,$DF,$82,$FE,$7D
	.dc.b	$F0,$FF,$CF,$FE,$13,$FA,$07,$C3,$FF,$41,$F8,$6F,$E8,$BF,$AC,$FF
	.dc.b	$17,$E9,$71,$38,$FF,$DC,$D2,$F8,$AD,$9C,$23,$91,$E9,$3F,$A9,$E1
	.dc.b	$DD,$C7,$57,$F5,$1C,$FC,$91,$8E,$76,$DE,$98,$C7,$07,$B5,$D7,$17
	.dc.b	$C1,$78,$72,$7C,$57,$97,$37,$CA,$7F,$CA,$78,$7C,$E7,$F9,$A7,$57
	.dc.b	$DE,$7E,$BB,$78,$3F,$07,$1F,$80,$6B,$A3,$1C,$1F,$80,$D7,$F0,$71
	.dc.b	$F8,$19,$D9,$27,$FC,$07,$D3,$F7,$FA,$7C,$93,$F3,$2B,$FB,$BC,$7F
	.dc.b	$71,$CB,$D5,$1E,$8C,$FD,$CF,$AB,$4F,$84,$FB,$35,$F8,$CF,$BB,$6F
	.dc.b	$96,$FC,$0D,$FE,$83,$F0,$B8,$3E,$AB,$FE,$76,$E2,$FB,$DB,$7D,$DD
	.dc.b	$D1,$F7,$67,$EB,$B5,$E0,$72,$DB,$D5,$3F,$E0,$BB,$EF,$18,$E1,$63
	.dc.b	$E4,$8C,$65,$E8,$75,$9F,$B1,$E1,$A7,$C0,$79,$6B,$F1,$1E,$9B,$7C
	.dc.b	$AD,$BF,$C4,$3B,$FF,$A7,$78,$3F,$C9,$3C,$0F,$89,$FC,$D9,$38,$7D
	.dc.b	$7E,$E7,$E7,$BE,$13,$F3,$FE,$C7,$E8,$3D,$3F,$D0,$FC,$07,$E8,$BD
	.dc.b	$5F,$D1,$FC,$0F,$E9,$3D,$7F,$D2,$FC,$17,$E9,$BD,$9F,$D3,$F6,$7E
	.dc.b	$A3,$DA,$FD,$4F,$C1,$FE,$AB,$D4,$FD,$27,$07,$F9,$3B,$8B,$E7,$BF
	.dc.b	$AA,$8D,$BF,$3A,$B9,$BF,$43,$0E,$EF,$D2,$CB,$C6,$FD,$4D,$3C,$EF
	.dc.b	$D6,$DB,$D6,$FD,$8E,$0F,$DD,$F0,$FC,$18,$FC,$07,$17,$C7,$E9,$FC
	.dc.b	$06,$AF,$84,$FE,$02,$4A,$FC,$0F,$BB,$FC,$13,$E1,$3F,$82,$FC,$27
	.dc.b	$F0,$6F,$80,$FF,$25,$F9,$1F,$9B,$BF,$F6,$0B,$FE,$DB,$6F,$9D,$AF
	.dc.b	$85,$DF,$C1,$F0,$F8,$6E,$3F,$0B,$CB,$E1,$F9,$FC,$3F,$4F,$88,$F9
	.dc.b	$39,$7E,$5F,$CE,$FE,$1E,$FE,$50,$D1,$FC,$33,$4F,$F6,$9B,$6F,$AB
	.dc.b	$9C,$7C,$2E,$F3,$F4,$E9,$EA,$F8,$8F,$5F,$8D,$ED,$AF,$D0,$F0,$7C
	.dc.b	$17,$E1,$9F,$A4,$E3,$F8,$CF,$FB,$4B,$93,$F2,$0F,$EA,$39,$BE,$87
	.dc.b	$91,$FA,$69,$74,$7E,$46,$BF,$5B,$E2,$FE,$45,$FD,$7F,$8B,$F9,$17
	.dc.b	$F6,$3E,$2F,$E4,$5F,$D9,$F8,$BF,$91,$7F,$6B,$E2,$FE,$45,$FD,$BF
	.dc.b	$8B,$F9,$17,$F8,$B9,$FE,$9B,$CB,$FC,$F4,$FF,$01,$8E,$0F,$F0,$39
	.dc.b	$E8,$FF,$05,$AF,$03,$FC,$1E,$FC,$8F,$F0,$9C,$3D,$0F,$F0,$BC,$7D
	.dc.b	$8F,$F0,$DC,$BE,$0E,$FF,$8E,$D7,$F2,$13,$FD,$2B,$6F,$93,$B7,$C5
	.dc.b	$7F,$A8,$F6,$FC,$F9,$FE,$2D,$C1,$F4,$DF,$C6,$E1,$F5,$DF,$C0,$71
	.dc.b	$7E,$35,$FC,$17,$5F,$E3,$5F,$C2,$75,$FE,$35,$FC,$37,$5F,$E2,$1F
	.dc.b	$E3,$67,$F1,$EE,$AF,$C6,$B8,$3F,$37,$D7,$F9,$47,$F1,$87,$43,$F1
	.dc.b	$87,$80,$FC,$61,$E4,$3F,$29,$CF,$F8,$C7,$F9,$5F,$67,$EA,$9F,$E5
	.dc.b	$BD,$BF,$01,$FC,$61,$0F,$C6,$70,$FC,$A7,$FB,$8B,$FB,$C7,$F7,$C7
	.dc.b	$E2,$8F,$E0,$47,$E2,$8F,$E0,$C7,$E2,$8F,$E1,$47,$E2,$8F,$E1,$C7
	.dc.b	$E2,$8F,$E2,$47,$E2,$8F,$E8,$3E,$2F,$EC,$B8,$BF,$AE,$47,$F8,$08
	.dc.b	$FC,$71,$F8,$B3,$89,$F8,$B3,$A9,$F8,$B3,$C2,$7E,$2C,$F2,$9F,$8B
	.dc.b	$3D,$27,$E2,$CF,$69,$F8,$B3,$F5,$84,$3F,$61,$C5,$F8,$AE,$0F,$E3
	.dc.b	$6F,$CB,$8F,$EB,$D8,$7C,$F7,$FB,$06,$3F,$4D,$FE,$2F,$97,$F0,$27
	.dc.b	$EC,$9F,$C6,$B7,$FC,$53,$FC,$73,$7F,$C5,$3F,$C7,$B7,$FC,$56,$7F
	.dc.b	$BD,$8F,$C4,$BF,$BF,$DF,$E4,$69,$FD,$B7,$B9,$C1,$F8,$BB,$FE,$4F
	.dc.b	$DD,$FC,$59,$FE,$53,$DD,$FC,$59,$FE,$57,$DD,$E5,$7C,$E7,$F1,$FC
	.dc.b	$3F,$49,$97,$F0,$DE,$5F,$D4,$3F,$C3,$F9,$BE,$25,$7E,$BC,$FF,$1D
	.dc.b	$78,$DF,$D9,$B9,$BF,$19,$E7,$FF,$80,$1F,$C6,$4F,$F8,$17,$C2,$FE
	.dc.b	$96,$FF,$81,$FC,$3F,$E9,$EF,$F8,$27,$C4,$FE,$A6,$FF,$82,$FC,$5F
	.dc.b	$EA,$EF,$F8,$37,$C6,$FE,$B6,$FF,$83,$FC,$7F,$EB,$EF,$F8,$47,$B7
	.dc.b	$FC,$61,$C1,$FD,$49,$CE,$FF,$C9,$EB,$8B,$FC,$A6,$FA,$BF,$CA,$F0
	.dc.b	$F0,$BF,$CB,$71,$F2,$BF,$CB,$F2,$F4,$BF,$CC,$73,$F6,$BF,$CC,$F4
	.dc.b	$FE,$C9,$B7,$F8,$23,$B7,$E3,$EF,$FE,$22,$7F,$15,$5F,$CE,$7B,$7F
	.dc.b	$14,$FF,$3B,$ED,$FC,$53,$FE,$22,$E8,$FE,$C0,$DE,$EF,$E2,$9C,$7F
	.dc.b	$8B,$3F,$83,$9F,$8B,$3F,$85,$9F,$8B,$3F,$AD,$74,$FE,$21,$FF,$37
	.dc.b	$F5,$7E,$33,$97,$FB,$BC,$7E,$41,$FC,$67,$5F,$E5,$30,$F1,$76,$7E
	.dc.b	$61,$F2,$76,$F8,$9F,$25,$FF,$14,$3F,$14,$7F,$14,$3F,$AF,$9F,$C5
	.dc.b	$0F,$EB,$E7,$F9,$1F,$6F,$EB,$FC,$0F,$1F,$F3,$9E,$BF,$F8,$08,$95
	.dc.b	$FD,$CF,$E2,$7F,$C0,$BE,$CF,$F8,$1B,$E2,$3F,$C0,$FF,$11,$FE,$08
	.dc.b	$F7,$3F,$C1,$3F,$11,$FE,$0A,$F7,$BF,$C1,$7F,$11,$FE,$0C,$F8,$0F
	.dc.b	$F0,$6F,$C4,$7F,$A4,$74,$70,$FF,$94,$3F,$5A,$7C,$1E,$63,$1F,$09
	.dc.b	$7A,$3F,$CB,$94,$F8,$5C,$60,$F0,$12,$F8,$6C,$C6,$3E,$1F,$FD,$71
	.dc.b	$EB,$4B,$FF,$4C,$DA,$76,$F1,$8C,$E1,$8C,$B3,$A7,$AA,$7F,$C2,$DB
	.dc.b	$7B,$9F,$F0,$C7,$07,$C2,$7F,$C3,$5C,$5F,$19,$FF,$0E,$72,$7C,$A7
	.dc.b	$E6,$1F,$E1,$FE,$7F,$9E,$BA,$7E,$A3,$ED,$F0,$53,$AE,$3F,$CB,$3F
	.dc.b	$E3,$8E,$9E,$95,$FE,$38,$C3,$C0,$FF,$8E,$B1,$F2,$3F,$E3,$CC,$BD
	.dc.b	$0F,$F8,$FB,$3F,$63,$F5,$39,$3E,$7E,$BF,$06,$3E,$19,$F2,$39,$3E
	.dc.b	$3F,$17,$E6,$E3,$F3,$A7,$CC,$7E,$5C,$7C,$CE,$1E,$53,$D4,$FB,$1C
	.dc.b	$5F,$80,$7F,$C8,$1C,$BF,$87,$7F,$C8,$5C,$DF,$8B,$7F,$C8,$9C,$FF
	.dc.b	$8F,$7F,$C8,$DD,$1F,$93,$7F,$C9,$1D,$3F,$0D,$F9,$84,$3E,$67,$57
	.dc.b	$E2,$BE,$23,$EF,$92,$3F,$07,$ED,$7E,$0B,$DB,$FC,$1F,$C3,$FE,$13
	.dc.b	$DD,$FC,$2F,$C3,$FE,$1B,$1F,$C3,$FC,$3F,$E2,$3E,$03,$F1,$3F,$0F
	.dc.b	$F8,$AF,$82,$FC,$5F,$81,$F8,$CF,$83,$FC,$6F,$C3,$FE,$3B,$E1,$BF
	.dc.b	$82,$78,$FF,$ED,$03,$FD,$6F,$C3,$93,$F2,$B3,$C5,$FC,$BD,$F5,$7F
	.dc.b	$33,$8F,$85,$FC,$DE,$7E,$57,$F3,$BA,$FA,$5F,$CF,$EF,$ED,$7D,$E7
	.dc.b	$E8,$B8,$BE,$13,$F1,$8F,$C8,$78,$7F,$8F,$EE,$FE,$0B,$1F,$98,$3F
	.dc.b	$D0,$BC,$BF,$45,$FF,$43,$F3,$7D,$57,$FD,$13,$CF,$F6,$5F,$F4,$5F
	.dc.b	$47,$DD,$7F,$D1,$BD,$3F,$AC,$7F,$67,$E3,$7E,$1D,$FC,$71,$F8,$B3
	.dc.b	$FC,$78,$43,$F1,$5C,$5F,$BD,$76,$63,$AF,$F8,$33,$D0,$FD,$FF,$7F
	.dc.b	$E7,$BC,$5F,$E7,$DE,$0F,$E8,$9F,$E8,$1E,$17,$E9,$9F,$E8,$5E,$1F
	.dc.b	$EA,$9F,$E8,$9E,$27,$EB,$9F,$E8,$DE,$2F,$C5,$C7,$E3,$9F,$DD,$3B
	.dc.b	$FE,$3B,$FE,$AA,$E4,$D6,$AE,$A5,$F3,$E5,$FF,$06,$65,$DD,$FF,$07
	.dc.b	$67,$E3,$7F,$C2,$1A,$79,$DF,$F0,$96,$BE,$B7,$F6,$F9,$FB,$CF,$85
	.dc.b	$E1,$FE,$C9,$F9,$47,$C8,$F8,$87,$91,$FC,$8E,$3E,$47,$8F,$B7,$89
	.dc.b	$CF,$87,$F9,$B7,$17,$78,$FE,$29,$97,$8D,$FE,$2D,$9F,$9D,$FE,$31
	.dc.b	$A7,$AD,$FE,$35,$AF,$BD,$FE,$39,$B7,$C4,$DF,$F8,$EC,$7C,$93,$E5
	.dc.b	$9F,$24,$FD,$D6,$5F,$41,$FD,$F9,$FC,$B0,$FE,$06,$7F,$30,$3F,$83
	.dc.b	$9F,$CD,$0F,$E1,$67,$F3,$83,$F8,$79,$F8,$43,$DD,$BF,$F4,$ED,$FE
	.dc.b	$14,$FF,$55,$D7,$F8,$F7,$03,$E1,$9E,$77,$F2,$2C,$7F,$23,$7F,$D6
	.dc.b	$F0,$FE,$3E,$FC,$88,$EE,$7F,$2F,$3C,$67,$F3,$33,$CE,$7F,$37,$3D
	.dc.b	$67,$F3,$B3,$DE,$7F,$BD,$33,$F9,$1C,$7F,$EB,$2C,$7F,$B0,$F6,$7E
	.dc.b	$EA,$FF,$B6,$3F,$D7,$3C,$8F,$E0,$18,$7E,$97,$BB,$F8,$23,$FA,$7E
	.dc.b	$FF,$E1,$0F,$EA,$7C,$0F,$E1,$8F,$EA,$FC,$1F,$E2,$0F,$EB,$78,$BF
	.dc.b	$8A,$3F,$AF,$F0,$BF,$8C,$3F,$EA,$FC,$FF,$D1,$BA,$7F,$1C,$8F,$ED
	.dc.b	$FE,$1F,$F2,$27,$FB,$8F,$89,$FC,$99,$FE,$E7,$E2,$FF,$2A,$7F,$BA
	.dc.b	$F8,$DF,$CB,$9F,$EE,$FD,$FF,$CC,$9F,$EF,$3E,$3F,$F3,$67,$FB,$DF
	.dc.b	$91,$EF,$7F,$8E,$10,$FE,$39,$97,$EC,$79,$3E,$4C,$7F,$1B,$D3,$FC
	.dc.b	$81,$E4,$7F,$1B,$7F,$C8,$5E,$47,$F1,$B7,$FC,$89,$E4,$7F,$1B,$7F
	.dc.b	$C8,$DE,$47,$F1,$B7,$FC,$91,$E4,$7F,$1B,$C3,$F8,$97,$91,$FC,$6D
	.dc.b	$FE,$29,$E4,$FE,$1D,$FE,$33,$DF,$F8,$A7,$F8,$E5,$FE,$3D,$FC,$5F
	.dc.b	$83,$F9,$37,$F2,$A7,$F6,$8F,$1B,$F6,$5B,$FF,$0B,$F1,$BF,$64,$FF
	.dc.b	$0D,$F1,$BF,$64,$FF,$0F,$F1,$BF,$64,$FF,$11,$F1,$BF,$64,$FF,$13
	.dc.b	$F0,$FE,$1E,$3F,$1B,$C9,$7C,$47,$F8,$90,$91,$FE,$28,$6B,$FC,$BD
	.dc.b	$BA,$FF,$8E,$DB,$AF,$F8,$E1,$B3,$FE,$00,$6E,$1F,$E3,$86,$DF,$82
	.dc.b	$7B,$1F,$C6,$2F,$F1,$71,$FE,$37,$8C,$9E,$12,$C7,$85,$B7,$E1,$69
	.dc.b	$E1,$2D,$F8,$5B,$5E,$17,$A1,$2B,$A3,$8F,$2F,$FC,$6B,$5C,$63,$57
	.dc.b	$DD,$A7,$72,$C6,$EB,$99,$27,$83,$67,$83,$67,$84,$E8,$93,$A4,$FF
	.dc.b	$98,$FF,$5B,$FF,$37,$3F,$8E,$F2,$7F,$8C,$C6,$2F,$A6,$3E,$F6,$E5
	.dc.b	$F4,$89,$7F,$11,$AF,$AA,$78,$61,$E2,$C7,$E1,$65,$F1,$31,$FC,$E5
	.dc.b	$FF,$76,$6E,$2B,$F3,$DA,$73,$18,$5E,$AF,$F9,$E9,$7C,$1D,$F5,$F1
	.dc.b	$F0,$99,$F6,$75,$F0,$BB,$F6,$F8,$7C,$37,$1F,$77,$97,$C3,$F3,$F7
	.dc.b	$FA,$7A,$DD,$7E,$03,$B7,$C4,$7E,$D7,$8F,$F1,$FE,$C6,$39,$7E,$6C
	.dc.b	$7A,$5F,$9D,$5C,$9C,$F2,$7A,$71,$EF,$18,$3E,$A6,$7E,$47,$1C,$BF
	.dc.b	$07,$6F,$4C,$63,$91,$C3,$EE,$8C,$71,$FC,$3E,$4F,$89,$CB,$F1,$7A
	.dc.b	$65,$7F,$69,$CB,$F0,$EF,$E1,$48,$E0,$9D,$1F,$6A,$32,$D8,$CF,$4F
	.dc.b	$E0,$38,$7F,$0B,$5F,$83,$3F,$53,$A7,$E2,$F8,$3E,$DF,$81,$F6,$A7
	.dc.b	$FD,$06,$7E,$04,$E5,$1F,$87,$C7,$B1,$30,$CC,$F8,$9B,$8F,$8A,$3C
	.dc.b	$07,$F1,$58,$FC,$EB,$FA,$F3,$FD,$99,$1F,$B3,$6C,$FE,$ED,$B3,$C1
	.dc.b	$FD,$A1,$83,$9C,$F1,$7A,$10,$D9,$5D,$38,$23,$82,$4F,$F1,$78,$FF
	.dc.b	$18,$72,$FF,$8B,$F2,$FE,$7D,$1F,$C3,$8F,$11,$FB,$8A,$E3,$7E,$76
	.dc.b	$C7,$9A,$FD,$8F,$8D,$AF,$89,$B1,$C2,$3A,$46,$09,$3F,$EF,$3F,$7F
	.dc.b	$FC,$63,$5B,$AF,$19,$0E,$D3,$E3,$AE,$31,$E3,$8C,$11,$C1,$0F,$0B
	.dc.b	$96,$99,$3E,$CE,$7F,$8B,$7F,$13,$87,$EB,$EF,$FC,$61,$A7,$F0,$39
	.dc.b	$FE,$A2,$79,$4F,$81,$7D,$5C,$4F,$AD,$F4,$9D,$CA,$78,$88,$7D,$2C
	.dc.b	$7E,$1E,$3E,$E8,$C9,$C0,$EA,$E5,$B5,$E1,$74,$43,$CA,$7D,$9E,$5F
	.dc.b	$C1,$69,$FC,$8B,$DA,$FF,$39,$CF,$F5,$F8,$E1,$8F,$A6,$38,$9E,$22
	.dc.b	$9E,$03,$D8,$60,$F6,$E3,$EA,$C3,$C9,$19,$24,$F5,$26,$3E,$B8,$F0
	.dc.b	$B5,$FC,$EE,$9F,$56,$7E,$9C,$7F,$86,$30,$FE,$70,$FD,$43,$8D,$F8
	.dc.b	$27,$AB,$6F,$A4,$62,$7D,$2F,$F6,$BC,$3D,$11,$FC,$03,$4F,$E2,$2F
	.dc.b	$F5,$1B,$FE,$0B,$7F,$C2,$63,$58,$7F,$0D,$9F,$E1,$38,$F9,$F2,$F9
	.dc.b	$B5,$FB,$23,$F9,$71,$CA,$63,$C8,$FE,$24,$5C,$BF,$89,$3E,$13,$F9
	.dc.b	$75,$3E,$0B,$FA,$B1,$23,$F9,$D7,$B9,$FD,$67,$E0,$FF,$AD,$7C,$1F
	.dc.b	$F5,$BF,$84,$FE,$B9,$F0,$9F,$DE,$2F,$FD,$B7,$FA,$E7,$F4,$D6,$FC
	.dc.b	$4E,$78,$7F,$45,$2E,$2C,$70,$FF,$6A,$87,$C3,$E3,$07,$02,$5F,$09
	.dc.b	$98,$C7,$C4,$59,$7C,$35,$F8,$9C,$7C,$47,$EB,$F1,$FB,$C7,$ED,$33
	.dc.b	$FB,$EF,$F5,$AB,$4F,$18,$C5,$F5,$C3,$B6,$5D,$F3,$F0,$69,$E1,$DB
	.dc.b	$C5,$9F,$C4,$C7,$FB,$17,$07,$C8,$7F,$B2,$71,$7F,$80,$F8,$FE,$74
	.dc.b	$7F,$83,$79,$3F,$D4,$5D,$1F,$83,$8F,$AC,$4E,$3F,$C4,$93,$1F,$C3
	.dc.b	$47,$DD,$9E,$13,$FE,$24,$E9,$FD,$AC,$FC,$AE,$9F,$C4,$CF,$F3,$83
	.dc.b	$94,$7E,$18,$ED,$7D,$1A,$66,$37,$CB,$F8,$AF,$2B,$FC,$4D,$97,$AA
	.dc.b	$7F,$8F,$74,$FE,$2E,$FF,$12,$7C,$03,$FC,$09,$D3,$F8,$97,$FC,$0D
	.dc.b	$D3,$F2,$1F,$A5,$C3,$F3,$1F,$A9,$C7,$F4,$1F,$AD,$CB,$F7,$F9,$BE
	.dc.b	$B4,$7E,$43,$9F,$EF,$3F,$E4,$73,$EF,$ED,$FE,$35,$7F,$06,$7D,$FE
	.dc.b	$9E,$4F,$43,$F0,$5D,$1F,$75,$FC,$27,$47,$DD,$7F,$0D,$C5,$F7,$5F
	.dc.b	$F2,$27,$47,$DD,$7F,$C8,$DD,$1F,$75,$FF,$11,$71,$FD,$EE,$4F,$F3
	.dc.b	$7F,$4B,$E2,$7F,$30,$4A,$FC,$97,$B3,$F9,$BF,$37,$F3,$9E,$CF,$E7
	.dc.b	$7C,$FF,$CF,$7B,$3F,$9F,$F4,$7F,$41,$EC,$FE,$87,$D3,$FD,$17,$B3
	.dc.b	$FA,$3F,$57,$F4,$9E,$CF,$E9,$E9,$EB,$DF,$67,$1F,$13,$F2,$3C,$4F
	.dc.b	$F6,$FF,$89,$FC,$96,$D3,$BE,$2F,$F1,$FC,$72,$7F,$51,$8F,$77,$F5
	.dc.b	$59,$F8,$DF,$D6,$4F,$9D,$FD,$76,$DE,$B7,$F6,$1C,$1E,$F7,$E0,$9F
	.dc.b	$B3,$E3,$F8,$8F,$FB,$82,$9E,$D7,$AC,$91,$E4,$F1,$FD,$DF,$13,$EF
	.dc.b	$F8,$9F,$81,$EE,$FC,$1F,$89,$F8,$5E,$FF,$C3,$F8,$9F,$89,$F0,$3F
	.dc.b	$17,$E2,$7E,$37,$C1,$FC,$7F,$89,$F9,$1F,$0B,$F2,$7E,$27,$E5,$7C
	.dc.b	$3F,$CB,$F8,$9F,$E2,$5E,$DF,$F9,$1F,$CE,$FE,$97,$C6,$ED,$FF,$00
	.dc.b	$73,$2F,$8C,$FC,$AC,$7A,$BF,$33,$3F,$0B,$F3,$B5,$F2,$BF,$43,$7F
	.dc.b	$4B,$F4,$B8,$7D,$AF,$D4,$E3,$F8,$2F,$C3,$3E,$BF,$37,$C6,$5E,$3F
	.dc.b	$E4,$BA,$3E,$6E,$DF,$CD,$1F,$E5,$5E,$37,$D3,$AF,$E7,$7D,$5F,$5D
	.dc.b	$FE,$7B,$D7,$F6,$DF,$E7,$FC,$3F,$7D,$FE,$83,$D9,$F8,$37,$FA,$1F
	.dc.b	$6F,$E4,$7C,$1F,$C5,$47,$C5,$F2,$FF,$D5,$37,$F8,$E8,$FE,$B6,$7F
	.dc.b	$A3,$3D,$6F,$E2,$F3,$FE,$0D,$F6,$FF,$2C,$F0,$F0,$B9,$EF,$F3,$B1
	.dc.b	$FC,$FE,$1F,$CE,$FC,$5F,$D1,$BF,$CF,$7C,$6F,$D3,$BF,$CF,$FC,$7F
	.dc.b	$D5,$BF,$D0,$7C,$8F,$D7,$BF,$D0,$FC,$9F,$86,$FE,$4C,$87,$E4,$F9
	.dc.b	$BE,$66,$5F,$79,$FD,$77,$8F,$F4,$A7,$DC,$7F,$5A,$E0,$74,$7D,$88
	.dc.b	$CF,$7A,$5D,$5F,$7A,$73,$0E,$BF,$C1,$C6,$3B,$3F,$0E,$EB,$B7,$F1
	.dc.b	$6F,$0E,$EF,$C7,$BC,$BB,$FF,$2F,$ED,$C8,$FD,$DC,$6B,$7F,$CB,$D7
	.dc.b	$27,$F3,$37,$D9,$FC,$DE,$1E,$27,$F3,$B8,$F9,$DF,$CF,$E7,$ED,$7F
	.dc.b	$43,$AF,$C3,$7F,$47,$B7,$C8,$E7,$FC,$0F,$07,$F8,$27,$C0,$FC,$A6
	.dc.b	$DF,$3A,$FF,$28,$FF,$04,$3F,$C1,$BC,$5F,$59,$FF,$07,$F1,$FD,$A7
	.dc.b	$FC,$23,$C9,$F7,$9F,$F0,$9F,$2F,$E0,$9F,$F0,$AF,$37,$E1,$9F,$9B
	.dc.b	$CF,$F8,$AD,$3F,$19,$1F,$92,$3F,$C5,$C4,$CB,$FA,$EB,$FC,$A0,$FE
	.dc.b	$47,$E5,$7E,$75,$FE,$46,$7E,$44,$FE,$46,$7E,$44,$FE,$46,$7E,$44
	.dc.b	$FE,$46,$7E,$44,$FE,$46,$7E,$44,$F8,$11,$F0,$9F,$F1,$6F,$83,$F1
	.dc.b	$FC,$6F,$93,$1F,$BE,$21,$F8,$8E,$0F,$A1,$87,$EB,$B8,$7E,$A3,$FB
	.dc.b	$0E,$2F,$B0,$FE,$CB,$8F,$EE,$3F,$B4,$E4,$FC,$03,$FB,$6E,$5F,$C2
	.dc.b	$3F,$B8,$E6,$FC,$43,$FB,$AE,$7F,$C6,$67,$FD,$C4,$FC,$81,$FD,$23
	.dc.b	$C4,$FE,$54,$FF,$4A,$F1,$7F,$97,$3F,$D3,$3C,$6F,$E6,$4F,$F4,$ED
	.dc.b	$FF,$9B,$3F,$D4,$3C,$7F,$E7,$4F,$F5,$2F,$23,$F9,$F3,$FD,$53,$C9
	.dc.b	$FD,$8B,$F0,$B9,$3F,$68,$8F,$E4,$37,$FE,$4F,$E4,$FE,$43,$AD,$9F
	.dc.b	$9E,$FF,$94,$FA,$FF,$1D,$7F,$E5,$8E,$BF,$C7,$3F,$E5,$AE,$BF,$C7
	.dc.b	$3F,$E5,$CE,$BF,$C7,$3F,$E5,$EE,$BF,$C7,$3F,$C5,$71,$FC,$39,$FD
	.dc.b	$AB,$C1,$FE,$36,$FF,$8D,$7C,$2F,$F1,$BF,$85,$FE,$17,$8D,$1C,$0F
	.dc.b	$CB,$9F,$8E,$3F,$36,$7E,$38,$FC,$F9,$F8,$E3,$F4,$67,$E3,$8F,$D3
	.dc.b	$9F,$8E,$3F,$56,$7C,$03,$FC,$DF,$E5,$FF,$85,$5F,$E3,$FE,$03,$83
	.dc.b	$F1,$F1,$FC,$53,$C1,$FD,$D3,$FC,$77,$BF,$F1,$EF,$F1,$FE,$FF,$C7
	.dc.b	$BF,$C8,$7B,$FF,$1E,$FF,$23,$EF,$FC,$7B,$FC,$97,$BF,$F1,$EF,$F2
	.dc.b	$7E,$FF,$C7,$BF,$CA,$7B,$FF,$1E,$FF,$2B,$EF,$FC,$7F,$17,$FB,$B7
	.dc.b	$CA,$FE,$E5,$E2,$7F,$6F,$AF,$C8,$1F,$DB,$CF,$C8,$1F,$DB,$CF,$C8
	.dc.b	$1F,$DB,$CF,$C8,$1F,$DB,$CF,$C8,$1F,$DB,$CF,$C8,$1F,$DB,$CF,$C8
	.dc.b	$1F,$DB,$CF,$88,$7E,$CB,$83,$E4,$3F,$D0,$7E,$27,$FC,$64,$47,$FC
	.dc.b	$57,$EE,$7F,$A6,$FF,$5D,$FF,$6A,$1F,$E0,$B1,$EF,$67,$DF,$D7,$C0
	.dc.b	$6F,$E2,$B8,$7C,$17,$1F,$8A,$E5,$F0,$9C,$FE,$2B,$A7,$C3,$75,$F8
	.dc.b	$AE,$DF,$11,$FE,$19,$DF,$FB,$77,$83,$F9,$0E,$3F,$91,$1F,$D5,$38
	.dc.b	$15,$C6,$3F,$55,$CA,$F2,$3F,$59,$E1,$2D,$9F,$86,$F7,$5A,$96,$BE
	.dc.b	$79,$C9,$FE,$BE,$86,$FE,$D8,$CE,$1D,$78,$7E,$14,$62,$BE,$2B,$7D
	.dc.b	$2F,$94,$FF,$97,$F9,$7E,$73,$FD,$AB,$9B,$F9,$43,$F4,$CF,$F2,$DB
	.dc.b	$AD,$7F,$C6,$F3,$CA,$7A,$1D,$F9,$BC,$2F,$C5,$C7,$CC,$FC,$7C,$FD
	.dc.b	$8F,$C9,$D3,$E1,$3F,$2F,$6F,$94,$FC,$DD,$FF,$01,$C3,$F9,$AE,$0F
	.dc.b	$97,$D1,$F8,$18,$FA,$F5,$EF,$E3,$FB,$6F,$C1,$E4,$FB,$EF,$C3,$E5
	.dc.b	$FC,$1B,$F1,$79,$BF,$0E,$FC,$7E,$7F,$C7,$70,$7E,$3F,$93,$DD,$3F
	.dc.b	$E0,$EE,$2F,$85,$AF,$F8,$D3,$93,$E3,$F3,$7D,$DF,$89,$FC,$B1,$2F
	.dc.b	$E9,$7C,$57,$E6,$7D,$CF,$CD,$7C,$57,$E6,$FD,$EF,$CE,$7C,$4F,$E7
	.dc.b	$7E,$03,$F3,$DF,$13,$FE,$F9,$FD,$3F,$F1,$59,$7D,$E5,$F1,$59,$F8
	.dc.b	$4D,$7C,$56,$FE,$1B,$87,$C5,$71,$F8,$AF,$CD,$63,$FD,$D7,$93,$FC
	.dc.b	$8C,$FF,$87,$0E,$91,$C1,$ED,$5D,$1F,$06,$1E,$3C,$7C,$97,$E6,$9F
	.dc.b	$C7,$69,$EA,$7F,$21,$B7,$B9,$FC,$97,$07,$EE,$78,$7E,$24,$7E,$A7
	.dc.b	$8B,$E4,$F8,$CE,$5F,$9B,$3F,$91,$3E,$81,$F9,$B3,$EA,$1F,$8F,$3E
	.dc.b	$C1,$F9,$FF,$13,$F1,$FE,$2F,$C8,$E7,$FD,$87,$8B,$F9,$08,$C7,$8D
	.dc.b	$FA,$EB,$FC,$31,$FA,$7F,$17,$F2,$33,$AF,$17,$F2,$33,$FE,$E6,$F1
	.dc.b	$7F,$20,$FE,$B6,$37,$C5,$FB,$9B,$FC,$DD,$F6,$F1,$BC,$13,$FB,$DC
	.dc.b	$BC,$AF,$EF,$F4,$FD,$3D,$79,$3C,$0F,$6E,$5F,$C3,$4F,$DB,$1F,$B2
	.dc.b	$21,$FE,$BF,$DB,$E3,$ED,$FE,$65,$77,$1C,$4F,$DA,$9F,$C5,$E7,$BE
	.dc.b	$1E,$18,$FD,$11,$FC,$1F,$1F,$3B,$FC,$3F,$2F,$5C,$FB,$63,$F5,$C7
	.dc.b	$C1,$3F,$14,$7E,$34,$FE,$EC,$63,$1F,$94,$F0,$3F,$7C,$7D,$03,$F8
	.dc.b	$11,$FB,$53,$C4,$79,$0F,$E1,$47,$DE,$3D,$47,$B1,$FB,$6F,$C0,$7F
	.dc.b	$D9,$78,$7F,$1C,$F1,$FF,$A6,$E5,$F1,$6B,$E4,$D6,$9F,$E9,$A7,$E3
	.dc.b	$4E,$97,$FC,$CF,$2F,$8F,$1F,$CC,$73,$FB,$2F,$F3,$3C,$FE,$3E,$1F
	.dc.b	$CF,$73,$F8,$F3,$FC,$DF,$4F,$C2,$BF,$CE,$75,$FC,$5B,$FE,$41,$CF
	.dc.b	$FB,$A7,$83,$87,$E3,$C7,$F4,$DF,$1B,$F8,$DB,$FD,$3F,$C6,$FE,$36
	.dc.b	$FF,$51,$F1,$BF,$8D,$BF,$D4,$FC,$6F,$E3,$6F,$F5,$5F,$1B,$F8,$DB
	.dc.b	$EC,$3F,$A5,$1F,$1F,$87,$F8,$D3,$F8,$CE,$2F,$8F,$CA,$E7,$FD,$BE
	.dc.b	$9F,$A8,$7F,$75,$5F,$BC,$3F,$A7,$78,$BF,$C6,$DF,$EA,$1E,$2F,$F1
	.dc.b	$B7,$FA,$97,$8B,$FC,$21,$FE,$35,$DF,$FC,$65,$FE,$39,$E2,$7F,$19
	.dc.b	$7F,$8F,$78,$9F,$C6,$5F,$C5,$6B,$FC,$61,$FC,$77,$3F,$FB,$0F,$C8
	.dc.b	$D4,$7F,$B1,$BC,$65,$F3,$8C,$61,$DD,$D6,$3E,$37,$FB,$D6,$5E,$77
	.dc.b	$FB,$E6,$7E,$B7,$FB,$F6,$9F,$CE,$5F,$E7,$67,$F6,$DD,$7F,$8D,$F0
	.dc.b	$7E,$3B,$BF,$F8,$D4,$FE,$3B,$7F,$9A,$FF,$1A,$3F,$68,$7F,$82,$79
	.dc.b	$FE,$AB,$F5,$CF,$E3,$47,$E8,$78,$7F,$8B,$F5,$FE,$3B,$8F,$F8,$C4
	.dc.b	$FE,$0C,$FE,$30,$7E,$B6,$7F,$12,$7F,$1E,$F2,$3F,$B4,$43,$91,$E0
	.dc.b	$7F,$1A,$8F,$ED,$C7,$1C,$BF,$7C,$F5,$3F,$6A,$7E,$3C,$FD,$A9,$F8
	.dc.b	$F3,$CE,$7A,$4F,$DA,$9E,$D3,$F5,$87,$C1,$BF,$D8,$3F,$E0,$F3,$FD
	.dc.b	$21,$A7,$E4,$30,$FD,$15,$7F,$88,$DB,$3F,$E2,$16,$E1,$FE,$2B,$6E
	.dc.b	$26,$92,$7F,$C4,$4D,$7F,$C7,$0F,$2F,$82,$E1,$1F,$E3,$E3,$64,$36
	.dc.b	$5C,$7E,$63,$70,$F3,$34,$F3,$17,$0F,$33,$70,$EE,$F4,$25,$74,$71
	.dc.b	$8F,$C3,$78,$63,$E9,$9D,$13,$0F,$0C,$F2,$6C,$66,$3C,$66,$4A,$E7
	.dc.b	$0F,$2C,$65,$B1,$FC,$52,$3F,$C9,$B9,$7E,$1E,$38,$63,$EA,$6E,$1E
	.dc.b	$5C,$3C,$24,$BE,$ED,$C3,$AB,$6F,$93,$6B,$F9,$2E,$3F,$CE,$39,$5F
	.dc.b	$13,$FE,$23,$26,$3F,$E2,$3F,$8A,$FF,$13,$7B,$9F,$E2,$7F,$89,$FF
	.dc.b	$14,$7B,$BF,$E2,$9F,$89,$FF,$15,$7B,$DF,$E2,$BF,$89,$FF,$16,$7B
	.dc.b	$FF,$E2,$DE,$7F,$F1,$77,$BF,$FE,$2F,$F8,$9F,$F1,$87,$C0,$7F,$8C
	.dc.b	$7E,$27,$FD,$C9,$BB,$9F,$FC,$19,$FA,$7B,$E2,$BF,$C5,$B1,$FE,$0F
	.dc.b	$EB,$F5,$D7,$F6,$23,$FD,$FA,$71,$E8,$FF,$41,$DF,$18,$FF,$1E,$61
	.dc.b	$E0,$7F,$C7,$D9,$79,$1F,$F2,$06,$9E,$87,$FC,$85,$B7,$B1,$FF,$22
	.dc.b	$70,$7C,$07,$FC,$8D,$C5,$F1,$1F,$8C,$7F,$92,$79,$7E,$4B,$FE,$FE
	.dc.b	$3E,$A9,$3A,$BE,$AB,$7A,$BD,$71,$82,$38,$27,$57,$E0,$67,$83,$67
	.dc.b	$6F,$F1,$27,$F8,$8F,$6E,$6B,$8D,$7E,$78,$87,$E4,$DC,$B6,$B8,$1F
	.dc.b	$86,$F8,$1F,$D0,$C7,$D6,$BF,$CD,$B6,$BF,$5E,$DA,$FB,$6D,$29,$FA
	.dc.b	$63,$06,$A7,$9B,$D8,$86,$CA,$E9,$C1,$1C,$12,$7C,$71,$FD,$6F,$8B
	.dc.b	$F1,$15,$FE,$04,$8F,$09,$0E,$B3,$C1,$E8,$DA,$E9,$3E,$A9,$CB,$63
	.dc.b	$31,$EE,$6B,$ED,$9E,$47,$F1,$4F,$59,$9F,$82,$F9,$4E,$9A,$4A,$E9
	.dc.b	$19,$6C,$65,$EC,$D7,$D8,$79,$8D,$90,$D9,$E7,$EE,$67,$F8,$9A,$ED
	.dc.b	$3A,$21,$D3,$1F,$6B,$63,$6D,$BF,$13,$ED,$3C,$07,$E0,$31,$F8,$2F
	.dc.b	$94,$EA,$6C,$A6,$4E,$87,$23,$46,$0D,$D7,$91,$FC,$07,$1C,$79,$1C
	.dc.b	$08,$60,$91,$C5,$B7,$EC,$7F,$A0,$D7,$A2,$74,$49,$D4,$F0,$BF,$82
	.dc.b	$D3,$D0,$53,$CF,$8F,$D0,$7F,$A0,$E3,$F9,$57,$CB,$7D,$6B,$9B,$63
	.dc.b	$9B,$5F,$83,$7F,$63,$0F,$33,$FB,$CC,$FF,$A4,$CF,$F0,$06,$BF,$BF
	.dc.b	$31,$97,$F4,$56,$9F,$C3,$8F,$F5,$A6,$3F,$C1,$1C,$E1,$FC,$41,$B7
	.dc.b	$DE,$B9,$47,$52,$77,$7F,$A1,$B9,$1F,$15,$FD,$03,$2F,$E7,$ED,$BF
	.dc.b	$E7,$2D,$BF,$F1,$65,$3E,$13,$19,$21,$F0,$78,$5F,$07,$7E,$17,$1F
	.dc.b	$09,$9F,$86,$D7,$C2,$EF,$E1,$F8,$7C,$37,$1E,$8E,$5F,$0D,$CF,$E2
	.dc.b	$3A,$7C,$3F,$5F,$89,$ED,$F1,$1F,$CB,$FB,$3E,$D7,$27,$F5,$13,$CF
	.dc.b	$1B,$46,$78,$FD,$17,$D6,$33,$CF,$FD,$83,$2F,$14,$63,$9F,$FA,$DE
	.dc.b	$BE,$78,$C1,$F1,$37,$F6,$38,$C7,$E2,$70,$FC,$18,$C7,$27,$C4,$6F
	.dc.b	$2F,$C7,$E6,$F9,$1C,$F2,$7F,$C8,$FC,$7F,$6B,$6F,$C3,$93,$5C,$12
	.dc.b	$7F,$0C,$7D,$F8,$D3,$87,$F9,$8B,$FE,$41,$35,$1C,$73,$FC,$54,$60
	.dc.b	$E1,$19,$E9,$FC,$64,$60,$FF,$17,$F4,$FE,$2F,$1F,$44,$7C,$DD,$3F
	.dc.b	$1A,$FC,$0C,$FF,$09,$A7,$E1,$7A,$7E,$24,$EF,$6F,$92,$F1,$DF,$E7
	.dc.b	$3C,$F8,$3E,$9B,$D7,$87,$EC,$3D,$F8,$BE,$F3,$E1,$E3,$FC,$27,$57
	.dc.b	$E1,$A3,$E8,$F3,$7E,$29,$FA,$7C,$FF,$8E,$7E,$AF,$37,$E4,$9F,$AF
	.dc.b	$D1,$F9,$67,$EC,$F4,$FE,$69,$FB,$7D,$5E,$B7,$ED,$9E,$F3,$EF,$F2
	.dc.b	$7C,$4E,$5F,$D8,$7C,$57,$E9,$49,$3F,$84,$F5,$FF,$51,$F0,$5F,$A9
	.dc.b	$F8,$9F,$D6,$FA,$9F,$DE,$BB,$9D,$BF,$9A,$EE,$4F,$E8,$27,$38,$FF
	.dc.b	$A8,$61,$F0,$7C,$BD,$BE,$7F,$07,$D3,$DD,$EB,$EB,$F6,$F7,$BB,$FC
	.dc.b	$1F,$83,$E0,$3C,$3F,$07,$F2,$74,$FF,$7D,$6F,$E1,$F2,$3C,$0F,$F8
	.dc.b	$FB,$BB,$73,$FA,$1D,$D8,$74,$8C,$65,$E0,$75,$A7,$91,$E1,$B7,$C0
	.dc.b	$DF,$E0,$F7,$7C,$29,$F8,$87,$EA,$B9,$3F,$87,$72,$FC,$A8,$FC,$87
	.dc.b	$37,$F9,$6B,$9D,$D9,$F1,$F6,$7C,$17,$CC,$25,$F9,$FE,$0F,$E7,$F9
	.dc.b	$9F,$43,$DE,$FA,$3F,$09,$F4,$BE,$17,$E9,$E3,$F5,$3E,$1B,$EA,$FC
	.dc.b	$17,$D6,$F8,$7F,$AF,$F0,$9F,$C9,$3F,$52,$FA,$7A,$FF,$0F,$5F,$0D
	.dc.b	$9F,$57,$5F,$0F,$F3,$7C,$4F,$F4,$95,$FC,$E3,$F8,$C9,$FC,$2E,$D3
	.dc.b	$C6,$31,$7D,$FB,$FC,$31,$F4,$72,$F2,$BF,$4F,$4F,$4B,$F5,$76,$F6
	.dc.b	$BF,$5F,$83,$E0,$BF,$0C,$FB,$5C,$7F,$19,$FF,$35,$78,$BF,$CF,$F9
	.dc.b	$FE,$86,$DF,$E8,$67,$E9,$CF,$D5,$3F,$8F,$F8,$DF,$65,$FC,$27,$5F
	.dc.b	$DD,$7E,$F9,$F8,$7F,$1B,$F2,$8F,$E2,$7C,$6F,$CA,$3F,$8B,$F1,$BF
	.dc.b	$28,$FE,$37,$C6,$FF,$6E,$75,$7F,$52,$F7,$BF,$8A,$EB,$FA,$49,$DD
	.dc.b	$7F,$3F,$EA,$61,$D2,$31,$8F,$81,$D6,$5F,$A7,$7F,$35,$E0,$FE,$49
	.dc.b	$FC,$E1,$F9,$23,$F8,$F7,$83,$F0,$1F,$C8,$9F,$CF,$B7,$FC,$97,$7F
	.dc.b	$E4,$78,$BE,$57,$17,$ED,$B6,$FD,$CC,$7E,$FB,$BF,$E8,$E1,$E7,$E4
	.dc.b	$FA,$AF,$A7,$97,$EC,$BE,$BE,$6F,$BA,$FB,$79,$FF,$02,$FE,$3F,$A3
	.dc.b	$F2,$5D,$3F,$87,$8F,$C9,$75,$7E,$3E,$7F,$75,$D3,$FE,$97,$F5,$9E
	.dc.b	$3F,$E5,$2F,$F2,$1D,$DC,$9F,$C9,$5F,$6E,$1F,$57,$4B,$4F,$1C,$63
	.dc.b	$5F,$3B,$AD,$BD,$6F,$0D,$FD,$EF,$2E,$0F,$8B,$8F,$C7,$CB,$FB,$09
	.dc.b	$FE,$10,$E4,$FC,$7C,$7F,$B3,$1F,$A1,$5F,$ED,$6E,$5F,$A8,$F9,$2B
	.dc.b	$EC,$1F,$81,$E6,$FB,$8F,$DE,$3F,$88,$9F,$8F,$3F,$8A,$9F,$8F,$3F
	.dc.b	$8D,$78,$FF,$C7,$1F,$E9,$87,$F8,$C8,$FE,$F7,$C1,$9A,$F8,$73,$FC
	.dc.b	$7E,$7F,$21,$E3,$FF,$1F,$7F,$25,$E3,$FF,$1F,$7F,$29,$E3,$FE,$8D
	.dc.b	$FC,$B7,$2F,$E9,$DF,$ED,$5E,$0F,$F4,$47,$FA,$31,$F8,$D2,$1F,$B4
	.dc.b	$E4,$FE,$3A,$FE,$E0,$87,$E8,$78,$3E,$7F,$07,$F2,$3E,$1F,$A6,$FF
	.dc.b	$25,$E2,$FA,$EF,$F2,$7E,$3F,$B6,$FF,$29,$E4,$FC,$03,$FD,$47,$C2
	.dc.b	$FE,$3B,$C3,$F9,$B7,$F8,$EE,$5F,$88,$3F,$8C,$74,$7E,$3D,$75,$FF
	.dc.b	$B9,$1F,$F3,$79,$F2,$8D,$9C,$4F,$E5,$1D,$5F,$C7,$5F,$E9,$9E,$1F
	.dc.b	$F2,$F7,$FA,$77,$89,$FC,$CD,$FE,$A1,$C1,$FC,$DD,$FE,$A5,$E2,$FF
	.dc.b	$3B,$7F,$AA,$78,$BF,$AF,$7E,$2E,$9F,$9A,$EF,$FD,$BC,$7E,$3C,$F9
	.dc.b	$B1,$FC,$75,$FA,$27,$F0,$03,$F8,$E9,$FC,$10,$FE,$3A,$7F,$08,$3F
	.dc.b	$8E,$9F,$8F,$3F,$02,$7E,$3C,$FC,$29,$F8,$F3,$F8,$B7,$4F,$E2,$F7
	.dc.b	$FE,$FD,$EB,$B0,$FE,$3C,$4A,$FF,$27,$46,$5F,$E4,$87,$E6,$0E,$97
	.dc.b	$FB,$6E,$AF,$C9,$3F,$B8,$EA,$FC,$93,$FB,$AE,$AF,$C9,$3F,$BC,$EA
	.dc.b	$FC,$93,$FC,$5F,$C0,$75,$FE,$4F,$1F,$F1,$F7,$91,$FB,$47,$E4,$1F
	.dc.b	$93,$DF,$F7,$6F,$F8,$4D,$FA,$0F,$F3,$3D,$7E,$A3,$FC,$D7,$83,$EC
	.dc.b	$3F,$CD,$F8,$7E,$E3,$FC,$E7,$8B,$F0,$0F,$F3,$BE,$3F,$C2,$3F,$86
	.dc.b	$3F,$11,$A7,$E4,$F9,$BF,$21,$CD,$F8,$F9,$FE,$D8,$FF,$A5,$FB,$7F
	.dc.b	$8F,$47,$F7,$5F,$07,$F8,$F3,$FC,$A4,$FC,$79,$FD,$AC,$FC,$79,$FD
	.dc.b	$AC,$FC,$79,$FD,$AC,$FC,$79,$FD,$AC,$FC,$79,$FD,$AC,$FE,$A5,$AF
	.dc.b	$F6,$BF,$23,$FC,$2B,$1F,$A4,$EA,$7C,$4F,$F8,$24,$91,$FE,$01,$F8
	.dc.b	$9F,$F0,$67,$C5,$7F,$83,$7D,$CF,$F0,$77,$C4,$FF,$83,$FE,$27,$FC
	.dc.b	$21,$EF,$7F,$84,$7E,$27,$FC,$25,$F0,$1F,$E1,$3F,$89,$FF,$0A,$7C
	.dc.b	$17,$F8,$57,$E2,$7F,$C2,$DF,$09,$FE,$17,$F8,$9F,$F0,$C7,$C3,$7F
	.dc.b	$86,$7E,$27,$FC,$D3,$C3,$FE,$11,$72,$FF,$84,$E7,$8B,$FE,$15,$BE
	.dc.b	$AF,$F8,$5F,$1F,$0B,$FE,$19,$CF,$CA,$FF,$86,$F5,$F4,$BF,$E1,$DD
	.dc.b	$FD,$AF,$BC,$FF,$10,$71,$7C,$27,$FC,$23,$C7,$F1,$9C,$90,$FF,$17
	.dc.b	$71,$F2,$7F,$C6,$15,$D9,$FF,$19,$5F,$89,$FF,$1A,$61,$E6,$7F,$C6
	.dc.b	$D8,$FA,$9F,$F1,$C6,$5E,$E7,$FC,$75,$9F,$C2,$7F,$C7,$9A,$7C,$66
	.dc.b	$9F,$E3,$7D,$7E,$63,$FE,$39,$DB,$E8,$3F,$E3,$BD,$FE,$A3,$FE,$3D
	.dc.b	$E0,$FB,$0F,$F8,$FF,$87,$EE,$3F,$E4,$1E,$2F,$C0,$3F,$E4,$3E,$3F
	.dc.b	$F2,$77,$0F,$C5,$EA,$FE,$65,$C9,$FC,$6B,$A5,$CA,$F8,$8F,$F2,$A1
	.dc.b	$31,$FF,$6B,$7E,$A6,$F8,$AF,$F7,$DB,$FE,$B8,$8F,$F7,$33,$FE,$FF
	.dc.b	$77,$F1,$3A,$F8,$9D,$FC,$07,$0F,$89,$E3,$F0,$5C,$BE,$27,$9F,$C2
	.dc.b	$74,$F8,$9E,$BE,$F7,$6F,$89,$EF,$F0,$FF,$B6,$F8,$5F,$E3,$5B,$7F
	.dc.b	$BD,$5B,$CB,$FC,$47,$C8,$DC,$FA,$2F,$9B,$EA,$C7,$BB,$EC,$CF,$C6
	.dc.b	$FB,$B5,$F3,$BF,$03,$7F,$7F,$07,$C0,$F8,$AF,$83,$F1,$1F,$0B,$93
	.dc.b	$E3,$5F,$F9,$E7,$97,$E6,$73,$7E,$46,$3F,$CF,$9D,$1F,$47,$D9,$FE
	.dc.b	$C9,$D5,$F9,$1E,$2F,$EC,$9E,$47,$E4,$5F,$EC,$DE,$47,$E4,$5F,$ED
	.dc.b	$1E,$47,$E4,$5F,$ED,$5E,$47,$E4,$5F,$ED,$9E,$47,$E4,$5F,$ED,$DE
	.dc.b	$47,$E4,$5F,$EE,$1E,$47,$E4,$5B,$AF,$E4,$4F,$DB,$1F,$91,$3F,$6C
	.dc.b	$7E,$44,$FD,$B1,$F9,$13,$F6,$C7,$E4,$4F,$DB,$1F,$91,$3F,$6C,$7E
	.dc.b	$44,$FD,$B1,$F0,$8F,$D9,$79,$3F,$C7,$3C,$2F,$C8,$EF,$F1,$67,$F8
	.dc.b	$F3,$F9,$87,$91,$FB,$63,$B1,$FB,$63,$C4,$7E,$D8,$F3,$1F,$B6,$3D
	.dc.b	$47,$ED,$8F,$71,$FB,$63,$E1,$1F,$B4,$3F,$A0,$FC,$47,$FB,$61,$FE
	.dc.b	$61,$87,$CB,$E0,$FF,$0E,$65,$F3,$DF,$F0,$F6,$7F,$4D,$FF,$10,$69
	.dc.b	$F5,$DF,$F1,$16,$BF,$6D,$FF,$12,$6D,$F7,$DF,$F1,$36,$FF,$83,$7F
	.dc.b	$0B,$97,$F1,$7D,$FF,$6E,$FF,$97,$B6,$F8,$EE,$67,$FC,$5D,$B7,$C7
	.dc.b	$8F,$F6,$57,$89,$FC,$09,$FF,$66,$78,$BF,$C1,$9F,$F6,$77,$8D,$FC
	.dc.b	$29,$FF,$68,$78,$FF,$C3,$9F,$F6,$97,$91,$FC,$49,$FF,$6A,$79,$3F
	.dc.b	$C5,$9F,$F6,$B7,$95,$FD,$05,$FF,$2F,$1F,$27,$5F,$F1,$F4,$6D,$FF
	.dc.b	$20,$4F,$37,$FC,$85,$5D,$DF,$F2,$25,$F8,$DF,$F2,$36,$1E,$77,$FC
	.dc.b	$91,$8F,$AD,$FF,$25,$65,$EF,$7E,$2E,$5F,$8D,$3F,$A0,$F4,$FE,$D2
	.dc.b	$3F,$CA,$0F,$ED,$E3,$F8,$DF,$5F,$EE,$DF,$E2,$E7,$EF,$CF,$E3,$FE
	.dc.b	$4F,$F1,$77,$F9,$0F,$93,$FC,$5D,$FE,$47,$E4,$FF,$17,$7F,$92,$F9
	.dc.b	$3F,$C5,$DF,$E4,$FE,$4F,$F1,$7F,$4B,$FC,$65,$1F,$E8,$43,$F9,$3C
	.dc.b	$6D,$FE,$53,$3C,$DF,$E5,$75,$DD,$FE,$5B,$7E,$37,$F9,$7E,$1F,$CB
	.dc.b	$5F,$49,$FC,$CB,$C2,$FE,$2A,$FF,$38,$F0,$BE,$17,$37,$F9,$CF,$5F
	.dc.b	$E2,$BC,$DF,$AF,$E7,$FC,$5D,$FF,$64,$3F,$5F,$C1,$FC,$51,$FD,$7B
	.dc.b	$FC,$51,$FD,$7B,$FC,$51,$FD,$7B,$FC,$51,$FD,$7B,$FC,$51,$FD,$7B
	.dc.b	$FC,$51,$FD,$7B,$FC,$49,$FD,$7F,$AD,$F8,$B7,$F1,$A7,$F1,$43,$F2
	.dc.b	$27,$F1,$43,$F2,$A7,$F1,$43,$F3,$27,$F1,$43,$F3,$A7,$F1,$43,$F4
	.dc.b	$27,$F1,$43,$F4,$A7,$F1,$48,$FC,$49,$F8,$AF,$07,$F8,$A7,$98,$F8
	.dc.b	$AF,$EB,$E4,$AF,$EC,$6D,$7F,$B0,$B6,$BF,$DF,$13,$C6,$76,$43,$65
	.dc.b	$8D,$CF,$1C,$D3,$FE,$A2,$AE,$4F,$02,$57,$C9,$13,$FE,$F6,$8F,$F5
	.dc.b	$39,$FB,$37,$9E,$2A,$E7,$5E,$0A,$F1,$56,$A7,$2D,$BF,$24,$F8,$DB
	.dc.b	$7C,$DB,$5C,$27,$47,$42,$9E,$8A,$E8,$FF,$1B,$8E,$15,$CE,$38,$35
	.dc.b	$E0,$72,$BF,$43,$88,$D4,$60,$95,$DA,$75,$5D,$89,$7E,$8F,$8A,$FE
	.dc.b	$C9,$5F,$05,$F5,$C4,$AF,$1D,$70,$6C,$65,$EA,$DB,$F2,$E3,$F8,$AF
	.dc.b	$37,$FC,$4E,$4A,$FF,$15,$FC,$37,$F8,$AB,$D7,$FF,$15,$FC,$37,$F8
	.dc.b	$B3,$D9,$FC,$0D,$31,$FB,$AF,$FB,$93,$F5,$EC,$B9,$DB,$EE,$C7,$F5
	.dc.b	$7D,$CB,$E1,$F9,$DF,$29,$CF,$BB,$8B,$E3,$F1,$17,$1E,$DE,$FD,$BE
	.dc.b	$7F,$13,$73,$7C,$0D,$F8,$AC,$7C,$27,$EE,$75,$ED,$ED,$FF,$8D,$35
	.dc.b	$FC,$8B,$74,$FE,$D6,$7E,$7A,$7C,$98,$AF,$A4,$E6,$FD,$B8,$F8,$72
	.dc.b	$F1,$69,$E3,$D7,$C9,$B7,$97,$7F,$37,$4F,$9F,$2F,$98,$7F,$96,$B8
	.dc.b	$BE,$83,$F4,$8F,$F2,$EF,$2F,$D5,$5C,$DF,$C0,$B9,$FE,$E6,$DF,$8B
	.dc.b	$E1,$FC,$6D,$7E,$4E,$30,$47,$04,$E4,$FC,$AC,$65,$B1,$9E,$8F,$C6
	.dc.b	$3F,$96,$E9,$FC,$9B,$FC,$05,$F0,$BB,$C2,$D7,$22,$57,$2C,$3C,$B1
	.dc.b	$89,$DC,$FF,$0A,$C7,$D9,$7F,$8A,$3F,$6A,$7E,$42,$BF,$60,$D9,$FE
	.dc.b	$22,$DA,$E0,$F0,$3F,$5C,$64,$E9,$3C,$9C,$90,$E0,$47,$4D,$9E,$EE
	.dc.b	$E3,$F8,$C6,$1F,$11,$B1,$FC,$AC,$F0,$1F,$A4,$79,$92,$79,$C7,$5A
	.dc.b	$F4,$CE,$5B,$3A,$8C,$4F,$AA,$30,$49,$D9,$FB,$6F,$75,$E4,$79,$E7
	.dc.b	$55,$C6,$B6,$D2,$1E,$78,$DB,$63,$6D,$36,$7B,$88,$77,$7D,$38,$7C
	.dc.b	$59,$3F,$D5,$67,$67,$C1,$39,$9C,$4A,$64,$F0,$10,$E3,$87,$C4,$73
	.dc.b	$87,$C0,$78,$1A,$AF,$5B,$6F,$D6,$D3,$BD,$73,$70,$77,$3F,$01,$EB
	.dc.b	$3D,$4F,$F1,$4E,$9F,$C3,$3C,$7F,$E1,$DB,$7A,$6F,$81,$E2,$30,$74
	.dc.b	$21,$82,$1F,$D5,$33,$F5,$63,$E3,$8C,$B5,$E8,$47,$04,$CF,$DF,$1F
	.dc.b	$81,$D3,$E9,$1F,$45,$E7,$1A,$79,$36,$79,$69,$F1,$A7,$6F,$E3,$75
	.dc.b	$FE,$11,$5F,$BD,$C6,$CF,$EF,$09,$5F,$D5,$08,$7F,$54,$3F,$DF,$38
	.dc.b	$79,$1C,$E7,$FC,$1A,$FA,$47,$F0,$D3,$37,$D9,$A6,$4F,$F0,$96,$DF
	.dc.b	$EB,$9E,$97,$C5,$FF,$45,$24,$FF,$39,$3F,$AC,$7B,$DF,$D2,$7E,$2F
	.dc.b	$FA,$57,$07,$F4,$BF,$8B,$FE,$99,$F0,$BF,$D3,$7E,$2F,$FA,$77,$C3
	.dc.b	$7F,$4F,$F8,$BF,$EA,$1F,$0F,$FD,$47,$E2,$FF,$A9,$7C,$47,$F5,$5F
	.dc.b	$27,$FB,$2E,$DF,$D8,$C7,$42,$1F,$17,$BE,$AF,$EB,$9C,$DF,$A6,$BF
	.dc.b	$F4,$CF,$3F,$F8,$E3,$A3,$51,$E9,$C7,$96,$1D,$FA,$3C,$33,$9E,$76
	.dc.b	$BE,$68,$C6,$FE,$A7,$5C,$3E,$E7,$87,$1F,$C2,$79,$72,$FF,$6D,$E6
	.dc.b	$F9,$31,$EB,$D3,$FB,$DF,$3F,$F8,$13,$A3,$F9,$CC,$FD,$D2,$70,$7D
	.dc.b	$BE,$9F,$C1,$C6,$37,$FB,$4D,$3F,$AA,$76,$7E,$B5,$F7,$78,$FF,$DD
	.dc.b	$72,$36,$75,$3C,$67,$FA,$63,$F8,$46,$B9,$E4,$FC,$5C,$63,$1F,$24
	.dc.b	$FF,$B6,$78,$BF,$0E,$FC,$9D,$3D,$F3,$FE,$02,$D7,$F0,$D3,$F3,$7A
	.dc.b	$FF,$0C,$FC,$FE,$BF,$C3,$3F,$47,$AF,$F0,$CF,$D3,$EB,$FC,$33,$F5
	.dc.b	$7A,$FE,$C3,$F6,$FA,$FF,$06,$DE,$BF,$C1,$F5,$FE,$96,$3E,$84,$A7
	.dc.b	$F1,$DD,$7C,$A7,$EA,$6E,$C3,$BC,$63,$1F,$1B,$AC,$BC,$EF,$0C,$FD
	.dc.b	$6F,$2D,$3D,$EF,$4D,$7E,$1F,$27,$F8,$F3,$83,$F0,$1E,$97,$F9,$EB
	.dc.b	$77,$C4,$FE,$84,$98,$7B,$7C,$DF,$D2,$7C,$1F,$E9,$7D,$7F,$D3,$7B
	.dc.b	$BF,$A7,$F6,$7F,$51,$EF,$7E,$A7,$C4,$FD,$57,$BF,$FA,$BF,$73,$F5
	.dc.b	$9F,$01,$FA,$DF,$7B,$F5,$DF,$03,$FA,$FF,$80,$FD,$87,$C1,$7C,$8E
	.dc.b	$0F,$D4,$F2,$FF,$9A,$5F,$9D,$1B,$7E,$85,$73,$7E,$96,$1D,$DF,$A9
	.dc.b	$97,$8D,$FA,$DA,$79,$DF,$B1,$B7,$AD,$FB,$5C,$1F,$C0,$78,$7E,$0C
	.dc.b	$7D,$8E,$2F,$8E,$FF,$02,$E3,$7C,$17,$F0,$72,$47,$E0,$FC,$DF,$E1
	.dc.b	$5E,$DF,$F9,$7F,$F5,$F7,$BB,$7E,$0F,$1E,$FE,$7E,$17,$5F,$0D,$BF
	.dc.b	$82,$E1,$F0,$FC,$7E,$13,$97,$C4,$73,$F8,$6E,$9F,$13,$D7,$C3,$ED
	.dc.b	$F1,$5F,$E1,$BD,$CE,$78,$BF,$88,$42,$B9,$46,$30,$EC,$EB,$2F,$13
	.dc.b	$C3,$4F,$33,$CB,$6F,$53,$D3,$83,$DE,$FC,$13,$F4,$9C,$7F,$11,$FF
	.dc.b	$30,$E9,$F8,$FE,$5F,$9B,$E8,$7F,$A2,$79,$BF,$23,$5F,$8F,$3E,$99
	.dc.b	$FA,$DF,$17,$F1,$EF,$EB,$FC,$5F,$C7,$BF,$B1,$F1,$7F,$1E,$FE,$CF
	.dc.b	$C5,$FC,$7B,$FB,$5F,$17,$F1,$EF,$ED,$FC,$5F,$C7,$BF,$D6,$8F,$E2
	.dc.b	$FE,$47,$F8,$13,$C6,$FC,$84,$7E,$E3,$C6,$FC,$83,$FB,$AF,$1B,$F2
	.dc.b	$0F,$EF,$3C,$6F,$C8,$3F,$BE,$F1,$BF,$20,$FF,$00,$F1,$BF,$20,$FF
	.dc.b	$02,$F1,$BF,$20,$FF,$04,$F1,$BF,$20,$FF,$0B,$E4,$FC,$8F,$AC,$E6
	.dc.b	$FD,$B7,$3F,$CB,$E7,$F9,$0F,$F5,$7F,$1B,$E8,$E1,$FC,$F7,$BF,$F1
	.dc.b	$EF,$E0,$3C,$5F,$C7,$BF,$82,$F1,$7F,$1E,$FE,$13,$C5,$FC,$7B,$F8
	.dc.b	$6F,$17,$F1,$EF,$E2,$3C,$5F,$C7,$BF,$8A,$D3,$F1,$EF,$ED,$8F,$EC
	.dc.b	$B8,$FE,$41,$D1,$F9,$0F,$13,$F1,$EF,$97,$C4,$FC,$7B,$E7,$F1,$3F
	.dc.b	$1E,$FA,$7C,$4F,$C7,$BE,$BF,$13,$F1,$EF,$B7,$C4,$F7,$F1,$7E,$27
	.dc.b	$87,$E2,$76,$7C,$6A,$1F,$90,$3F,$71,$CB,$F9,$08,$FC,$94,$7E,$3C
	.dc.b	$FA,$47,$E3,$CF,$E0,$A7,$E3,$CF,$E1,$27,$E3,$CF,$E1,$A7,$E3,$CF
	.dc.b	$E2,$27,$E3,$CF,$E2,$A7,$E3,$CF,$E3,$27,$E3,$CF,$EE,$5F,$15,$FE
	.dc.b	$ED,$3F,$DD,$DE,$A7,$F8,$26,$3F,$20,$71,$3F,$20,$75,$3F,$20,$78
	.dc.b	$4F,$C8,$1E,$53,$F2,$07,$A4,$FC,$81,$ED,$3D,$E7,$F5,$4F,$27,$E2
	.dc.b	$3F,$18,$F9,$3D,$1F,$91,$7F,$90,$74,$7C,$E6,$9F,$BF,$3F,$AE,$F9
	.dc.b	$3F,$C7,$F5,$FA,$FA,$B3,$FA,$7E,$A3,$91,$C1,$F7,$AB,$58,$FE,$2B
	.dc.b	$87,$F0,$71,$8B,$E1,$C5,$F8,$88,$C6,$0E,$3F,$C6,$C6,$39,$3F,$90
	.dc.b	$63,$FA,$3F,$2B,$F9,$21,$89,$D4,$7E,$74,$FE,$44,$7E,$84,$FE,$44
	.dc.b	$7E,$94,$FE,$44,$7E,$A4,$FE,$44,$7E,$B4,$FE,$44,$7F,$80,$F6,$FD
	.dc.b	$A3,$FC,$88,$87,$E4,$F2,$FF,$25,$72,$7D,$06,$C7,$E4,$EB,$F9,$27
	.dc.b	$67,$F0,$67,$F9,$37,$6F,$F0,$A7,$F9,$47,$77,$F0,$E7,$F9,$57,$7F
	.dc.b	$F1,$27,$F9,$67,$81,$FC,$59,$FE,$5D,$E0,$FF,$1A,$7F,$98,$78,$5F
	.dc.b	$DB,$F2,$FF,$22,$7A,$DF,$E3,$73,$FC,$71,$D7,$F9,$3A,$DC,$FE,$4D
	.dc.b	$E6,$FE,$4D,$EE,$FE,$4D,$F1,$BF,$93,$7C,$EF,$E4,$DF,$5B,$F9,$37
	.dc.b	$DE,$FC,$1E,$7F,$DA,$BF,$DC,$4F,$C9,$F7,$7E,$F5,$FF,$15,$F5,$7E
	.dc.b	$4F,$83,$FC,$E9,$DD,$FC,$12,$FF,$1F,$CF,$FC,$8A,$FF,$92,$78,$7F
	.dc.b	$C8,$9F,$E4,$DE,$1F,$F2,$27,$F9,$47,$87,$FC,$89,$FE,$55,$E1,$FF
	.dc.b	$22,$7F,$96,$78,$7F,$C8,$9F,$E5,$DA,$FF,$22,$7F,$77,$EF,$FF,$10
	.dc.b	$69,$FD,$C3,$AF,$F2,$1A,$79,$3B,$3F,$2E,$F9,$BB,$7F,$38,$FA,$3B
	.dc.b	$BF,$42,$FA,$BB,$FF,$4C,$FB,$3C,$0F,$D6,$BE,$EF,$07,$E0,$65,$FE
	.dc.b	$15,$E3,$F8,$9E,$67,$F8,$5A,$3E,$49,$FC,$DB,$E2,$BF,$C2,$84,$7F
	.dc.b	$C0,$BE,$CF,$F8,$5F,$D8,$FF,$0C,$7B,$FF,$E1,$9B,$FF,$0D,$7C,$57
	.dc.b	$F8,$6F,$E0,$7F,$C3,$9F,$15,$EF,$EE,$FD,$B7,$EC,$1F,$07,$3F,$F4
	.dc.b	$72,$E6,$D7,$07,$F0,$28,$7C,$26,$2B,$FD,$5B,$0F,$85,$C7,$80,$87
	.dc.b	$C3,$61,$7C,$3D,$F8,$8F,$A5,$A7,$F0,$7F,$4B,$F9,$66,$B8,$D7,$C1
	.dc.b	$5C,$6F,$96,$3C,$F3,$E9,$AF,$5C,$3B,$6B,$EE,$3F,$C5,$3B,$FC,$27
	.dc.b	$FC,$57,$C3,$F1,$9F,$F1,$6F,$1F,$F8,$FF,$93,$E6,$47,$F7,$3E,$5F
	.dc.b	$F2,$4F,$17,$F2,$17,$3C,$3F,$C6,$9E,$13,$C8,$FF,$1B,$F2,$FD,$17
	.dc.b	$FC,$73,$CB,$F4,$5F,$F1,$DF,$2F,$99,$FF,$66,$72,$FC,$F6,$EF,$ED
	.dc.b	$73,$C3,$F0,$5D,$F1,$FC,$EE,$5F,$8F,$1F,$DE,$F5,$FA,$13,$B8,$FB
	.dc.b	$75,$CD,$FB,$B7,$F1,$DF,$09,$F1,$CF,$F0,$97,$0F,$C7,$7F,$C2,$9C
	.dc.b	$3F,$1D,$FF,$0B,$70,$FC,$07,$E3,$1F,$E5,$6D,$FE,$3E,$8F,$89,$FC
	.dc.b	$01,$27,$DF,$F1,$7F,$83,$F8,$BF,$C2,$7B,$BF,$85,$F7,$BF,$0D,$F0
	.dc.b	$FF,$87,$F6,$7F,$11,$F1,$1F,$89,$F8,$1F,$C5,$7C,$4F,$E2,$FE,$0F
	.dc.b	$F1,$9F,$15,$F8,$DF,$85,$FC,$77,$C5,$FE,$3F,$E1,$FF,$15,$C0,$EA
	.dc.b	$FF,$6D,$E9,$FE,$47,$8F,$F1,$0F,$99,$B7,$E9,$DF,$37,$EA,$E3,$DD
	.dc.b	$FA,$F9,$F8,$DF,$B3,$AF,$9D,$FB,$7B,$7A,$DF,$BB,$C1,$EF,$7E,$09
	.dc.b	$F8,$0E,$3F,$88,$FF,$4D,$F4,$DC,$DF,$8E,$F1,$BF,$91,$BF,$3F,$C2
	.dc.b	$FE,$D3,$D3,$F4,$DF,$ED,$7D,$5F,$5D,$FE,$DB,$D7,$F6,$DF,$DE,$E3
	.dc.b	$F7,$CF,$C0,$9F,$90,$F1,$7F,$56,$FE,$4B,$C6,$FD,$A3,$F8,$BF,$07
	.dc.b	$F1,$AF,$F0,$4F,$07,$F1,$97,$FC,$23,$C1,$FC,$63,$FC,$2B,$C1,$EC
	.dc.b	$F8,$2B,$F3,$E7,$EC,$8F,$D1,$9F,$B2,$3F,$4E,$7E,$A4,$FF,$31,$78
	.dc.b	$BF,$05,$FE,$40,$7F,$22,$EA,$FD,$93,$F8,$F2,$1C,$0E,$53,$FE,$0E
	.dc.b	$F1,$7F,$65,$1F,$E6,$DF,$17,$F6,$4F,$F9,$BF,$C5,$FD,$93,$FC,$83
	.dc.b	$C6,$FE,$16,$FC,$53,$F6,$27,$C5,$3F,$62,$7F,$45,$3F,$61,$CB,$FE
	.dc.b	$69,$7E,$31,$B3,$F6,$27,$33,$E8,$9D,$CF,$D8,$9F,$18,$FD,$89,$F1
	.dc.b	$8F,$D8,$9F,$18,$FD,$89,$FD,$00,$F8,$38,$7F,$13,$7E,$34,$7E,$29
	.dc.b	$CF,$2F,$E2,$9E,$07,$23,$FD,$21,$8F,$C6,$AF,$F1,$AE,$3F,$19,$FF
	.dc.b	$1B,$E3,$F1,$9F,$E8,$F8,$FC,$6B,$FE,$8F,$C3,$F8,$A7,$FA,$4F,$0F
	.dc.b	$E2,$9F,$ED,$64,$AF,$E7,$D5,$FC,$64,$D3,$FD,$4A,$79,$3F,$D5,$2B
	.dc.b	$B3,$E0,$AF,$8E,$7D,$93,$F1,$87,$DD,$3F,$18,$7B,$0F,$C1,$E5,$FE
	.dc.b	$0C,$D7,$F1,$9A,$FE,$2A,$7F,$C7,$9E,$37,$F1,$77,$F6,$84,$3F,$B1
	.dc.b	$F8,$DF,$B9,$7F,$B2,$F4,$FE,$F5,$FE,$CF,$D5,$FC,$05,$FE,$D3,$D7
	.dc.b	$FC,$15,$FE,$D7,$D9,$FC,$25,$FE,$DB,$DB,$FC,$35,$FE,$DF,$DD,$FC
	.dc.b	$45,$FF,$30,$1F,$ED,$C2,$18,$3F,$BA,$F8,$3F,$C7,$9F,$E9,$3E,$17
	.dc.b	$F2,$27,$FA,$5F,$87,$FC,$99,$FE,$9B,$E2,$7F,$2A,$7F,$A7,$F8,$9F
	.dc.b	$CB,$9F,$EA,$3E,$2F,$F3,$27,$FA,$9F,$8D,$FD,$DF,$83,$F8,$AF,$2F
	.dc.b	$F8,$1B,$C5,$FE,$C7,$DB,$F8,$DB,$FE,$82,$49,$FF,$3E,$F3,$7F,$84
	.dc.b	$E3,$F3,$11,$FC,$59,$FC,$E3,$FC,$59,$FD,$03,$FC,$1D,$FD,$21,$FA
	.dc.b	$67,$F0,$4F,$EC,$8F,$C6,$9F,$B2,$3F,$19,$CF,$FE,$96,$F0,$7F,$8B
	.dc.b	$CF,$E3,$A7,$4F,$E3,$5E,$4F,$47,$B4,$FF,$87,$7A,$FF,$3C,$FF,$94
	.dc.b	$BB,$3F,$44,$FF,$95,$3B,$7F,$4C,$FF,$95,$BB,$BD,$D7,$FC,$44,$F8
	.dc.b	$51,$FA,$F7,$FD,$11,$8F,$92,$3C,$D8,$7E,$86,$7F,$97,$35,$F5,$9F
	.dc.b	$D5,$E7,$83,$63,$58,$FB,$63,$7D,$8A,$DB,$6B,$77,$E0,$AE,$E4,$38
	.dc.b	$D7,$02,$4F,$08,$F0,$1B,$F8,$0F,$F6,$65,$F5,$8E,$58,$78,$C9,$87
	.dc.b	$78,$EB,$7C,$EB,$84,$60,$E3,$3B,$AE,$31,$E0,$34,$47,$33,$C9,$B8
	.dc.b	$7A,$27,$99,$C6,$33,$7E,$19,$FF,$17,$4F,$FA,$EF,$F6,$1F,$F6,$CA
	.dc.b	$AF,$F6,$C4,$61,$B8,$7F,$97,$21,$F1,$57,$E2,$F1,$F0,$59,$F8,$4D
	.dc.b	$7C,$1E,$FE,$2B,$87,$C2,$71,$F8,$AE,$5F,$0B,$CF,$E2,$BA,$7C,$37
	.dc.b	$5F,$8A,$ED,$F0,$FD,$FE,$2B,$FC,$33,$CD,$FE,$18,$E4,$FE,$07,$C9
	.dc.b	$FE,$EA,$F5,$B5,$03,$D5,$CB,$FE,$F8,$E5,$63,$D6,$75,$7E,$EC,$FC
	.dc.b	$51,$8B,$E1,$AF,$9E,$30,$7F,$8A,$70,$6F,$ED,$8C,$E6,$97,$0F,$C3
	.dc.b	$9C,$C3,$8F,$E4,$46,$39,$7E,$66,$52,$3F,$BF,$73,$7D,$B7,$F0,$11
	.dc.b	$F8,$52,$3F,$75,$B8,$FE,$0D,$B3,$C7,$5F,$C0,$D7,$E1,$2B,$4F,$12
	.dc.b	$38,$24,$F1,$AE,$7C,$FF,$89,$25,$73,$9F,$C5,$F4,$7E,$91,$FC,$6C
	.dc.b	$7F,$B2,$EF,$F3,$ED,$8C,$BF,$A5,$31,$5F,$EB,$17,$04,$3F,$44,$68
	.dc.b	$D4,$F4,$74,$D7,$66,$27,$BB,$82,$46,$C9,$1A,$6C,$F8,$E7,$73,$92
	.dc.b	$46,$67,$73,$E4,$9C,$B6,$75,$18,$8F,$F6,$74,$F0,$6C,$67,$A1,$CB
	.dc.b	$E9,$BE,$24,$AE,$84,$38,$B9,$69,$93,$E4,$18,$21,$82,$1D,$CE,$05
	.dc.b	$36,$53,$67,$C4,$7C,$0E,$2F,$D8,$DB,$F8,$0D,$9D,$BF,$C9,$6F,$DE
	.dc.b	$F3,$3C,$25,$3C,$25,$39,$1A,$AE,$EF,$94,$86,$CA,$E9,$C1,$1C,$12
	.dc.b	$BC,$F1,$C1,$FA,$97,$6F,$D3,$3C,$67,$6D,$8C,$D7,$36,$CE,$A3,$13
	.dc.b	$E9,$8C,$12,$76,$48,$D3,$5F,$87,$3F,$D2,$AF,$E4,$BF,$B8,$9F,$DA
	.dc.b	$CF,$F0,$83,$F7,$B5,$FB,$B6,$C6,$B2,$FE,$0B,$1B,$AE,$4F,$EF,$67
	.dc.b	$59,$7F,$02,$8C,$10,$FF,$2E,$E5,$FD,$93,$57,$C4,$7F,$47,$24,$FF
	.dc.b	$22,$F8,$AF,$E9,$5F,$13,$FD,$2F,$E1,$FF,$A6,$7B,$7F,$D3,$7E,$2B
	.dc.b	$F9,$D5,$FF,$3A,$3F,$DB,$14,$F7,$73,$BB,$5F,$E8,$1B,$A9,$F1,$1A
	.dc.b	$9C,$C3,$DF,$C6,$3D,$3E,$22,$E1,$C7,$E0,$6D,$BE,$22,$FC,$1E,$3E
	.dc.b	$23,$3F,$0B,$AF,$88,$DF,$C3,$CD,$3F,$A1,$E2,$9D,$C6,$2F,$9B,$AC
	.dc.b	$7B,$E5,$E0,$D3,$C3,$AF,$8B,$7F,$1F,$0F,$C5,$C3,$FA,$5F,$1F,$C9
	.dc.b	$7F,$A6,$E9,$FD,$D7,$97,$E7,$C7,$C9,$E1,$FE,$A9,$CD,$F3,$3A,$3F
	.dc.b	$9B,$74,$FD,$CA,$FC,$19,$F6,$F9,$BF,$06,$4E,$6F,$C0,$47,$E1,$7A
	.dc.b	$BF,$15,$8F,$E1,$A3,$17,$CE,$31,$7C,$67,$A9,$A3,$FB,$3D,$7F,$48
	.dc.b	$F8,$8F,$EE,$D9,$7A,$2B,$FC,$22,$FA,$EB,$DA,$7F,$6A,$EA,$FC,$3B
	.dc.b	$FD,$B3,$AB,$E2,$BF,$43,$87,$E4,$BF,$4B,$8F,$E6,$BF,$3C,$FA,$BD
	.dc.b	$5F,$6D,$FA,$FD,$5F,$6D,$FB,$3D,$5F,$6D,$F9,$5C,$7F,$6D,$D9,$F6
	.dc.b	$CE,$67,$53,$7C,$FF,$6D,$FC,$1F,$3F,$DB,$7F,$0B,$CF,$F6,$DF,$AD
	.dc.b	$D6,$E8,$FB,$91,$8C,$3E,$13,$AE,$0F,$BB,$CF,$FE,$52,$97,$1F,$E6
	.dc.b	$09,$5F,$92,$F1,$7F,$37,$F1,$1F,$9C,$F3,$FF,$3B,$F1,$1F,$9E,$F3
	.dc.b	$BF,$3F,$F1,$1F,$A0,$F4,$3F,$43,$F1,$1F,$A2,$F4,$BF,$4B,$FB,$17
	.dc.b	$ED,$34,$FC,$63,$FE,$4A,$ED,$FD,$94,$3B,$D2,$F8,$7E,$3F,$0D,$CB
	.dc.b	$E2,$3E,$75,$FF,$0F,$E3,$F9,$8F,$48,$DB,$FE,$2C,$BE,$4D,$F0,$FC
	.dc.b	$6F,$94,$FD,$7E,$9E,$97,$F6,$3B,$7B,$5F,$D9,$F0,$7C,$17,$F6,$BC
	.dc.b	$5F,$BE,$E3,$F8,$F1,$F7,$B9,$3F,$4C,$7F,$0C,$C9,$F1,$1F,$08,$93
	.dc.b	$E5,$F8,$9F,$8B,$ED,$7C,$6F,$6F,$E3,$FB,$9F,$23,$E2,$3E,$4F,$BD
	.dc.b	$F2,$BE,$23,$E5,$FC,$07,$CC,$F8,$8F,$9B,$F0,$5F,$3B,$E2,$3E,$7F
	.dc.b	$C2,$7D,$0F,$88,$FA,$3F,$0D,$F2,$3B,$FE,$7F,$7F,$F1,$FF,$2F,$F2
	.dc.b	$D1,$8E,$EF,$96,$EB,$99,$7C,$F9,$7E,$66,$5D,$DF,$9D,$A7,$8D,$FA
	.dc.b	$1B,$79,$DF,$A5,$C1,$EB,$7E,$A7,$17,$BD,$FA,$DC,$9F,$86,$E5,$F8
	.dc.b	$B1,$F8,$6E,$6F,$96,$FF,$3A,$E7,$F9,$FB,$79,$DF,$A5,$1F,$50,$FE
	.dc.b	$53,$E2,$7E,$59,$FE,$57,$E2,$7E,$59,$FE,$5B,$E2,$7E,$59,$FE,$5F
	.dc.b	$E2,$7E,$11,$FC,$31,$F8,$9F,$13,$F1,$6F,$E3,$4F,$E3,$DE,$9F,$DD
	.dc.b	$F1,$33,$9F,$C7,$AF,$CF,$F8,$9C,$EF,$F8,$0E,$1D,$DF,$E0,$78,$F8
	.dc.b	$DF,$E0,$B9,$79,$DF,$E0,$F9,$FE,$C1,$FD,$91,$FD,$13,$4F,$C7,$F8
	.dc.b	$DF,$EC,$AE,$7F,$C6,$CF,$E7,$F7,$F9,$3A,$FD,$9E,$0F,$C6,$CF,$E8
	.dc.b	$FC,$0F,$C6,$BF,$A5,$F0,$3F,$1A,$FE,$9F,$C0,$FC,$6B,$FA,$9F,$03
	.dc.b	$F1,$AF,$EA,$FC,$0F,$C6,$BF,$AD,$F0,$3F,$1A,$FF,$58,$F5,$FF,$17
	.dc.b	$CF,$FD,$A3,$C0,$FC,$83,$FA,$3E,$FF,$C5,$57,$F0,$2E,$FF,$C5,$3F
	.dc.b	$C1,$3B,$FF,$14,$FF,$06,$EF,$FC,$53,$FC,$23,$BF,$F1,$4F,$F0,$AE
	.dc.b	$FF,$C5,$3F,$C3,$3B,$FF,$14,$FC,$32,$0E,$7F,$E4,$1D,$1F,$8A,$D3
	.dc.b	$FB,$EB,$F8,$A9,$FE,$6F,$E1,$FE,$FD,$FE,$73,$E2,$7F,$03,$7F,$9D
	.dc.b	$F8,$BF,$C1,$DF,$E7,$BE,$37,$F0,$B7,$F8,$E9,$F8,$23,$F1,$47,$F1
	.dc.b	$3F,$1F,$F8,$D3,$F1,$CF,$EB,$7E,$17,$FB,$F7,$C5,$FD,$8E,$7F,$8D
	.dc.b	$9E,$07,$23,$F9,$4C,$FE,$37,$B7,$F7,$1E,$1F,$E3,$5F,$DD,$78,$7F
	.dc.b	$8D,$7F,$79,$E1,$FE,$35,$FD,$F7,$87,$F8,$DC,$3F,$AA,$F8,$FF,$0D
	.dc.b	$FD,$71,$FE,$E3,$E2,$FC,$77,$87,$F8,$CD,$3F,$8E,$3F,$3A,$7E,$81
	.dc.b	$F8,$D3,$F7,$A7,$E3,$4F,$E0,$27,$E3,$4F,$E0,$A7,$E3,$4F,$E1,$27
	.dc.b	$E3,$4F,$E1,$A7,$E3,$4F,$E7,$5D,$5F,$8A,$7F,$C7,$7D,$7F,$8E,$7F
	.dc.b	$C7,$BD,$9F,$92,$7F,$A6,$F6,$FE,$59,$FE,$9F,$DD,$F9,$A7,$FA,$8F
	.dc.b	$7F,$E7,$9F,$EA,$7E,$07,$E8,$9F,$EA,$BE,$0F,$E9,$9F,$EA,$FE,$17
	.dc.b	$EA,$9F,$EB,$3E,$1F,$C3,$7F,$AC,$78,$9F,$B2,$7F,$1D,$DD,$F8,$C7
	.dc.b	$F7,$04,$3E,$79,$FC,$5C,$FD,$F9,$FD,$38,$FE,$2E,$7F,$52,$3F,$8B
	.dc.b	$9F,$D5,$8F,$E2,$E7,$F5,$D3,$F1,$C7,$F6,$CF,$17,$F8,$9F,$0F,$E2
	.dc.b	$63,$FC,$50,$7F,$8E,$FD,$4F,$EE,$11,$9E,$6F,$E7,$DD,$9F,$DD,$AF
	.dc.b	$9C,$7F,$41,$BF,$CE,$3F,$D1,$7B,$7F,$40,$FF,$47,$EE,$FD,$23,$FD
	.dc.b	$27,$BF,$F5,$0F,$F4,$BF,$03,$F5,$8F,$F3,$7B,$F8,$5D,$3F,$8C,$7F
	.dc.b	$6B,$1F,$D3,$FA,$7F,$18,$FF,$51,$F0,$7F,$7A,$FF,$53,$F0,$BF,$80
	.dc.b	$BF,$D5,$7C,$3F,$E0,$AF,$F5,$7F,$13,$F8,$4B,$FD,$67,$C5,$FE,$1A
	.dc.b	$FF,$5B,$F1,$BF,$88,$BF,$D7,$7B,$FF,$19,$E4,$7F,$63,$D3,$FC,$72
	.dc.b	$7E,$D4,$FF,$0B,$C7,$0E,$4F,$EF,$97,$D1,$FE,$FD,$87,$81,$FF,$00
	.dc.b	$63,$E4,$7F,$C0,$59,$7A,$1F,$F0,$26,$7E,$C7,$FC,$0D,$A7,$C0,$7F
	.dc.b	$5B,$97,$F1,$8D,$BF,$9F,$7C,$47,$F8,$A8,$91,$FE,$1D,$F8,$8F,$F4
	.dc.b	$8F,$EC,$7F,$E2,$A8,$FF,$3C,$2F,$77,$3E,$E6,$BD,$ED,$FC,$47,$0F
	.dc.b	$80,$E3,$F1,$1C,$BE,$0B,$9F,$C4,$74,$F8,$4E,$BF,$09,$DB,$E1,$7B
	.dc.b	$FC,$47,$83,$E1,$FF,$CF,$1B,$9C,$F0,$0F,$1D,$72,$71,$DB,$FE,$12
	.dc.b	$A6,$3E,$09,$CF,$2F,$FA,$D2,$5A,$79,$A7,$27,$F9,$1E,$1B,$7B,$23
	.dc.b	$3D,$A9,$70,$FC,$29,$CC,$7C,$56,$E1,$FE,$56,$E4,$F9,$B1,$FE,$56
	.dc.b	$E5,$FF,$08,$E1,$FE,$54,$73,$A7,$F8,$95,$E1,$3C,$8E,$BC,$BD,$DF
	.dc.b	$A7,$97,$95,$FA,$B9,$FA,$DF,$AF,$A7,$C1,$7E,$CE,$BF,$1D,$F9,$27
	.dc.b	$A3,$83,$E8,$3F,$96,$3F,$D1,$BE,$86,$1F,$C0,$C7,$D7,$7D,$FC,$5F
	.dc.b	$6D,$F8,$3C,$7F,$7D,$F8,$7C,$9F,$83,$7E,$2F,$2F,$E1,$DF,$8F,$CD
	.dc.b	$F8,$ED,$BF,$1F,$C7,$F0,$27,$E3,$1F,$E6,$FE,$0F,$8F,$C8,$F8,$8F
	.dc.b	$CA,$92,$7E,$8F,$B5,$F9,$8F,$86,$FC,$CF,$B5,$F9,$AF,$86,$FC,$DF
	.dc.b	$BB,$F9,$CF,$86,$FF,$76,$FE,$C9,$FE,$F9,$C7,$EE,$BF,$EF,$66,$FC
	.dc.b	$16,$BE,$1B,$7F,$09,$C3,$E2,$38,$FC,$3F,$2F,$89,$FD,$FF,$A9,$FC
	.dc.b	$8B,$1F,$F4,$83,$7C,$26,$FF,$DC,$A7,$FD,$29,$A2,$F9,$C7,$8B,$1E
	.dc.b	$EF,$93,$3F,$3E,$9E,$8F,$89,$F4,$FC,$47,$AE,$BF,$23,$C1,$EF,$7F
	.dc.b	$27,$9F,$C3,$7E,$29,$F9,$6E,$4F,$90,$FF,$4C,$3F,$A7,$F3,$7E,$4B
	.dc.b	$9F,$FD,$1C,$FD,$3F,$13,$F5,$91,$F5,$E7,$EC,$9F,$A0,$F1,$3F,$25
	.dc.b	$3F,$B1,$F1,$3F,$24,$FE,$CF,$C4,$FC,$93,$FB,$5F,$13,$F2,$4F,$ED
	.dc.b	$FC,$6F,$DE,$E1,$FC,$3B,$C7,$FE,$E5,$CD,$FC,$A6,$30,$7E,$5C,$FD
	.dc.b	$7C,$F2,$CF,$F7,$98,$76,$7F,$7D,$8F,$89,$FE,$01,$97,$99,$FE,$05
	.dc.b	$9F,$EA,$F3,$FE,$11,$DF,$F9,$2C,$BF,$91,$1F,$91,$21,$F1,$F3,$FD
	.dc.b	$C3,$89,$FD,$C3,$C0,$FE,$4F,$7F,$1D,$FE,$53,$E0,$7E,$85,$FE,$57
	.dc.b	$E0,$FE,$95,$FE,$5B,$E1,$7E,$A5,$FE,$5F,$E1,$FE,$B5,$FE,$63,$E2
	.dc.b	$7E,$C5,$FD,$99,$FC,$43,$87,$F9,$39,$F2,$B2,$F9,$91,$FB,$63,$F8
	.dc.b	$AF,$07,$C5,$7F,$8F,$70,$7C,$57,$F9,$07,$07,$C5,$7F,$91,$70,$7C
	.dc.b	$5D,$BF,$94,$70,$7C,$59,$FF,$77,$F8,$FF,$87,$7F,$12,$7F,$8F,$7D
	.dc.b	$7F,$C4,$BF,$16,$AB,$C0,$FE,$38,$7F,$4C,$AE,$CF,$80,$FE,$0E,$7F
	.dc.b	$2C,$C3,$CD,$C5,$FC,$C7,$2F,$53,$FC,$CF,$3F,$C1,$57,$F3,$9E,$3F
	.dc.b	$85,$8F,$F3,$FD,$BF,$17,$CC,$F2,$3F,$8E,$70,$7F,$47,$F2,$3E,$63
	.dc.b	$FD,$27,$8B,$E8,$3F,$D2,$F8,$FE,$A3,$FD,$37,$93,$EC,$3F,$D3,$F9
	.dc.b	$7E,$E3,$FD,$47,$9B,$F0,$0F,$EA,$36,$FC,$23,$F8,$67,$F8,$E9,$F8
	.dc.b	$BF,$23,$F8,$E3,$FB,$52,$1F,$DC,$3C,$8F,$DD,$3F,$DC,$B5,$FD,$F3
	.dc.b	$FD,$D3,$AF,$F8,$12,$BF,$F7,$9F,$85,$FC,$76,$31,$C3,$FD,$A7,$91
	.dc.b	$3F,$C7,$E7,$27,$F1,$4E,$FF,$E4,$0F,$F2,$4F,$17,$F9,$03,$FC,$9B
	.dc.b	$C5,$FE,$40,$FF,$86,$F1,$FF,$80,$3E,$2F,$F1,$F3,$F9,$06,$9F,$C7
	.dc.b	$F2,$EF,$E2,$FF,$1E,$BE,$F1,$E2,$F1,$7F,$8F,$C6,$34,$78,$BF,$C8
	.dc.b	$23,$F9,$AF,$8B,$FC,$81,$FE,$DF,$E3,$7F,$3C,$7F,$B8,$F8,$FF,$D0
	.dc.b	$1F,$EE,$7E,$47,$F8,$6B,$93,$F8,$FE,$FF,$E0,$5F,$1F,$FB,$8F,$B0
	.dc.b	$F2,$3F,$35,$BF,$EF,$31,$FC,$B1,$FB,$A3,$FB,$89,$FC,$80,$FE,$E2
	.dc.b	$7E,$58,$FE,$E2,$7F,$60,$3F,$C0,$7E,$1F,$E5,$BA,$7F,$C2,$71,$FB
	.dc.b	$33,$FB,$89,$FB,$A3,$F8,$E9,$FD,$C0,$FE,$E5,$8F,$F8,$6F,$C1,$FC
	.dc.b	$DC,$7E,$74,$E5,$1F,$CA,$DF,$EE,$27,$80,$F1,$1E,$43,$F8,$79,$FC
	.dc.b	$DC,$F5,$3F,$C8,$1F,$D9,$AE,$9F,$C3,$64,$F1,$FE,$2B,$F1,$EB,$F7
	.dc.b	$C7,$F5,$AF,$89,$FE,$D6,$47,$FB,$5B,$5F,$B8,$DB,$FC,$AB,$6F,$F2
	.dc.b	$9C,$88,$C5,$75,$39,$90,$E6,$5A,$E6,$FF,$23,$3F,$DC,$24,$3E,$6B
	.dc.b	$A7,$89,$1C,$12,$BC,$73,$E2,$D3,$F4,$ED,$BF,$35,$F8,$6B,$6D,$3D
	.dc.b	$07,$57,$2F,$F4,$98,$E2,$E4,$DD,$F8,$5B,$3C,$23,$17,$E2,$9E,$CD
	.dc.b	$79,$36,$37,$5D,$9B,$7E,$9F,$81,$FE,$47,$E2,$7F,$88,$9F,$64,$4C
	.dc.b	$3A,$DF,$1C,$7C,$44,$8F,$29,$31,$F0,$46,$5B,$19,$9E,$4F,$C4,$F3
	.dc.b	$1E,$57,$E0,$32,$FF,$19,$92,$7F,$C6,$71,$FE,$36,$F8,$2F,$F1,$BF
	.dc.b	$AD,$FE,$38,$F8,$3F,$F1,$CF,$B1,$FE,$3A,$F8,$4F,$F1,$DF,$B5,$FE
	.dc.b	$3C,$F8,$5F,$F1,$EF,$B9,$FE,$3E,$F8,$6F,$F4,$46,$9F,$C4,$3E,$1A
	.dc.b	$B7,$E1,$F3,$CD,$FA,$18,$71,$A9,$F0,$1A,$E5,$FE,$55,$0F,$81,$FE
	.dc.b	$97,$1F,$CB,$7C,$9F,$E4,$59,$63,$1E,$77,$B7,$FD,$16,$F4,$6C,$75
	.dc.b	$76,$7A,$4F,$F2,$AE,$9E,$D7,$FC,$AF,$B7,$C1,$7F,$CB,$3C,$1F,$15
	.dc.b	$FF,$2D,$F1,$7C,$97,$E5,$9F,$E5,$EE,$5F,$9C,$FE,$4F,$A3,$F0,$0F
	.dc.b	$E2,$BA,$3E,$BF,$07,$DA,$23,$82,$4F,$0E,$1F,$C6,$D7,$E2,$23,$F1
	.dc.b	$CF,$F1,$AD,$FF,$A8,$79,$3F,$DC,$63,$FC,$DF,$D5,$C6,$7F,$11,$87
	.dc.b	$49,$E4,$43,$C9,$7B,$7F,$18,$FF,$0D,$EA,$FC,$5C,$7E,$C0,$FB,$F9
	.dc.b	$7E,$95,$B5,$FB,$06,$D7,$E0,$9F,$DA,$98,$31,$3C,$5C,$90,$EC,$57
	.dc.b	$4E,$08,$E0,$93,$E1,$8F,$F0,$A1,$FC,$57,$2F,$ED,$3C,$1F,$87,$AD
	.dc.b	$DF,$96,$B9,$DF,$16,$C6,$67,$71,$E8,$9D,$46,$23,$F3,$ED,$7D,$C6
	.dc.b	$23,$67,$F1,$4F,$39,$E9,$7B,$63,$37,$DD,$B5,$DE,$30,$47,$04,$3B
	.dc.b	$39,$69,$93,$C6,$43,$C1,$A7,$F1,$67,$F0,$E7,$F4,$13,$47,$C2,$39
	.dc.b	$1E,$78,$EC,$EB,$0F,$80,$E4,$D9,$C4,$61,$F0,$1C,$1C,$6F,$DE,$FA
	.dc.b	$08,$66,$BB,$BD,$4C,$94,$D1,$F8,$CF,$01,$E0,$7F,$9E,$23,$D8,$6C
	.dc.b	$87,$88,$B8,$7A,$9C,$11,$FE,$85,$87,$A5,$D1,$30,$F4,$CF,$88,$F2
	.dc.b	$9C,$5C,$BD,$9A,$FF,$B5,$B1,$FA,$4F,$F8,$D5,$FE,$F8,$F3,$21,$C7
	.dc.b	$1F,$86,$F5,$6D,$78,$DF,$E6,$77,$E7,$BF,$E0,$15,$FB,$E3,$F8,$06
	.dc.b	$9F,$C3,$1C,$11,$C1,$36,$FE,$13,$8F,$F2,$9C,$5C,$DF,$CC,$AF,$F1
	.dc.b	$27,$1A,$FE,$1F,$97,$F0,$BA,$FE,$3A,$DC,$7C,$78,$F5,$BF,$E1,$E7
	.dc.b	$F8,$82,$BF,$A2,$7C,$47,$7E,$73,$4C,$7C,$16,$7E,$0F,$5E,$EE,$FE
	.dc.b	$13,$87,$BF,$C7,$E1,$79,$7C,$0F,$3F,$86,$E9,$F0,$7D,$7E,$1F,$B7
	.dc.b	$C2,$F7,$F8,$8F,$07,$C3,$F8,$78,$3F,$BD,$71,$7F,$2F,$3F,$AC,$F7
	.dc.b	$FF,$2F,$7E,$C1,$FD,$42,$38,$BF,$D4,$AB,$AB,$FD,$53,$0F,$0B,$FD
	.dc.b	$5B,$2F,$2B,$FD,$63,$4F,$4B,$FD,$6B,$6F,$6B,$FD,$73,$83,$E0,$BF
	.dc.b	$D7,$B8,$BE,$23,$F2,$79,$7E,$6F,$3F,$D2,$8F,$BE,$D7,$EB,$1F,$40
	.dc.b	$BC,$BF,$59,$FF,$10,$F6,$7E,$A9,$FE,$55,$3F,$DC,$7B,$BF,$95,$5F
	.dc.b	$69,$E5,$D1,$F6,$F0,$E5,$1D,$27,$51,$89,$FC,$87,$7F,$F4,$C8,$FB
	.dc.b	$F5,$EB,$21,$FE,$0B,$E9,$FB,$CF,$F7,$CE,$9F,$BC,$FF,$7E,$E9,$FB
	.dc.b	$CF,$F8,$03,$A7,$EF,$3F,$E0,$2E,$9F,$BC,$FF,$81,$3A,$7E,$F3,$FE
	.dc.b	$06,$E9,$FB,$CF,$E6,$B4,$FF,$0B,$66,$E9,$FB,$D1,$86,$F0,$7F,$9B
	.dc.b	$3B,$38,$DF,$F8,$87,$0E,$AF,$F8,$8F,$1F,$0B,$FE,$25,$CB,$CA,$FF
	.dc.b	$89,$F3,$F4,$BF,$E2,$9D,$3D,$AF,$F8,$AF,$5F,$82,$FC,$33,$E2,$E9
	.dc.b	$E6,$EA,$FD,$2C,$7E,$A5,$7C,$47,$EA,$49,$1F,$E2,$4F,$47,$F5,$9E
	.dc.b	$87,$EE,$FF,$65,$FF,$1A,$4F,$FB,$21,$7C,$3E,$7C,$2D,$7C,$46,$FE
	.dc.b	$0B,$87,$C4,$71,$F8,$4E,$5E,$BF,$3F,$85,$E9,$F1,$1D,$7C,$BE,$DF
	.dc.b	$11,$FA,$AD,$BF,$DA,$BE,$5F,$F0,$97,$F5,$27,$86,$38,$23,$C9,$5D
	.dc.b	$5C,$72,$7D,$6E,$06,$5E,$39,$CF,$67,$FB,$CE,$5A,$FA,$67,$27,$F9
	.dc.b	$4A,$1B,$FB,$E3,$3D,$C9,$F8,$71,$87,$E3,$E3,$F4,$78,$FF,$8B,$72
	.dc.b	$7C,$EA,$FF,$31,$72,$FF,$99,$3C,$0F,$CF,$60,$F8,$8F,$A0,$4A,$E9
	.dc.b	$E6,$FD,$3F,$73,$EA,$7B,$BF,$57,$DE,$FA,$DE,$27,$D7,$F7,$FE,$C7
	.dc.b	$C4,$7D,$9F,$81,$FE,$69,$FB,$33,$E2,$AF,$C2,$63,$E2,$B3,$F0,$DA
	.dc.b	$F8,$AD,$FC,$47,$0F,$8A,$FE,$A3,$D3,$FA,$0C,$91,$AE,$3F,$87,$0B
	.dc.b	$E7,$18,$C7,$BB,$AC,$FC,$8F,$C7,$D7,$D0,$FC,$9D,$FD,$8F,$CB,$E1
	.dc.b	$F8,$0F,$C2,$3E,$77,$27,$C5,$5C,$DF,$C1,$FC,$8F,$CE,$C7,$F9,$FB
	.dc.b	$A3,$E8,$6F,$FA,$3E,$9F,$CE,$CF,$D7,$EE,$FC,$F7,$91,$F7,$63,$F3
	.dc.b	$DD,$DF,$83,$7F,$41,$DF,$F8,$97,$F4,$5E,$07,$E3,$DF,$C8,$9F,$81
	.dc.b	$F1,$3F,$36,$FF,$07,$C3,$F9,$23,$FC,$98,$FE,$F8,$7E,$1F,$C3,$FC
	.dc.b	$DB,$C8,$FD,$77,$8B,$FA,$87,$F1,$9E,$37,$EB,$1F,$C7,$78,$FF,$B0
	.dc.b	$7F,$21,$E4,$7A,$9F,$61,$FE,$9E,$F1,$7F,$22,$FC,$32,$1F,$23,$E2
	.dc.b	$3F,$45,$CB,$F2,$F8,$BF,$50,$FF,$01,$F1,$7E,$8D,$7F,$01,$E7,$FA
	.dc.b	$AF,$F0,$3E,$8F,$B2,$FF,$05,$E9,$FB,$AB,$0F,$C4,$75,$7E,$0A,$31
	.dc.b	$C3,$FB,$9F,$1D,$DC,$ED,$FC,$4D,$68,$FF,$61,$C7,$E3,$5F,$C7,$F8
	.dc.b	$1F,$95,$7F,$B8,$57,$F3,$3F,$77,$11,$C6,$3A,$CF,$E7,$63,$F2,$FE
	.dc.b	$17,$E8,$5F,$CC,$F8,$7F,$A6,$CF,$D3,$A7,$F0,$9D,$7D,$AF,$F0,$BD
	.dc.b	$BE,$0B,$FC,$37,$7F,$8A,$FF,$0F,$E0,$F9,$5C,$1F,$33,$2F,$E6,$1D
	.dc.b	$FF,$92,$8F,$F6,$CB,$F5,$23,$FD,$C9,$C9,$FC,$BE,$7E,$D4,$7F,$2F
	.dc.b	$E5,$FB,$CF,$F3,$1E,$6F,$C1,$3F,$CC,$F9,$FF,$0C,$FE,$20,$FE,$25
	.dc.b	$E4,$FE,$33,$3F,$E4,$5D,$5F,$CB,$5F,$F4,$FC,$FE,$4F,$91,$E9,$BD
	.dc.b	$2F,$CC,$CF,$F3,$FE,$DF,$E6,$EF,$F1,$DE,$DE,$F8,$7F,$2D,$CF,$C6
	.dc.b	$FF,$2F,$D3,$CE,$FF,$31,$D7,$D6,$FF,$33,$DB,$F6,$38,$7D,$5F,$0D
	.dc.b	$E1,$FC,$58,$FF,$7B,$F1,$7F,$25,$EF,$F9,$51,$FD,$47,$BF,$E6,$BF
	.dc.b	$CD,$79,$7E,$8E,$3F,$CE,$F9,$BE,$AB,$FC,$F7,$9F,$EC,$BF,$CF,$FA
	.dc.b	$3E,$EB,$F9,$1E,$EF,$E4,$98,$7E,$08,$FE,$30,$7E,$18,$FE,$3B,$E2
	.dc.b	$FF,$1F,$7F,$19,$1F,$C8,$9F,$C8,$DF,$F8,$01,$FC,$BF,$8D,$BE,$0F
	.dc.b	$E5,$17,$F9,$C7,$F9,$57,$83,$FA,$37,$F4,$B5,$FD,$27,$C7,$FE,$72
	.dc.b	$FF,$4B,$F2,$3F,$9E,$BF,$D3,$7C,$9F,$E8,$2F,$F4,$FF,$2B,$F6,$CF
	.dc.b	$F2,$42,$1F,$BD,$3E,$59,$FC,$02,$FF,$C0,$9C,$7F,$91,$7F,$AB,$47
	.dc.b	$F0,$83,$FB,$F7,$93,$FC,$8E,$3F,$B5,$79,$3F,$C8,$DF,$ED,$9E,$4F
	.dc.b	$F2,$37,$FB,$77,$93,$FC,$8D,$FE,$E1,$E1,$FE,$47,$8F,$FB,$77,$87
	.dc.b	$F9,$17,$F9,$1E,$5F,$DF,$BD,$4C,$3F,$DB,$E7,$83,$FD,$C6,$BA,$57
	.dc.b	$F7,$7C,$3C,$0F,$F7,$9C,$7C,$8F,$F7,$BC,$BD,$0F,$F7,$DC,$FF,$59
	.dc.b	$87,$F2,$33,$FC,$3C,$7C,$38,$FD,$9B,$FE,$8D,$F3,$3F,$B9,$3F,$2B
	.dc.b	$4F,$E4,$6F,$F5,$13,$FC,$45,$B7,$D2,$7F,$C4,$9B,$FD,$67,$FC,$4D
	.dc.b	$C1,$F6,$9F,$F1,$47,$0F,$F0,$D7,$F1,$C7,$E2,$B8,$BF,$09,$97,$E4
	.dc.b	$F8,$FF,$10,$FE,$57,$93,$FB,$8B,$6B,$F9,$09,$F9,$43,$F2,$C7,$33
	.dc.b	$AE,$7F,$8C,$7C,$27,$F6,$F3,$F1,$87,$F6,$F3,$F1,$87,$F6,$F3,$F1
	.dc.b	$87,$BD,$F8,$2F,$EC,$4F,$F5,$97,$33,$E2,$BF,$C1,$A4,$8F,$F0,$1F
	.dc.b	$BD,$FE,$10,$F1,$7F,$C2,$3E,$6F,$F8,$4B,$BF,$FC,$27,$F1,$3F,$E1
	.dc.b	$4F,$80,$FF,$0A,$FC,$4F,$F8,$5B,$E0,$BF,$C2,$FF,$13,$FE,$18,$F8
	.dc.b	$4F,$F0,$CF,$C4,$FF,$86,$BE,$1B,$FC,$37,$F1,$3F,$EC,$FF,$D9,$FF
	.dc.b	$C8,$5C,$1F,$E1,$2E,$5F,$F6,$97,$B3,$FE,$49,$E5,$FF,$0C,$46,$E7
	.dc.b	$89,$FE,$18,$C3,$AB,$FE,$1A,$CB,$C2,$FF,$87,$34,$F2,$BF,$E1,$ED
	.dc.b	$BD,$2F,$F8,$83,$83,$DA,$FF,$88,$B8,$BF,$C5,$FC,$7F,$0E,$3F,$83
	.dc.b	$F2,$7F,$1B,$7F,$AD,$B9,$D3,$FC,$58,$F0,$8E,$47,$F8,$B3,$93,$E4
	.dc.b	$3F,$E2,$EE,$4F,$90,$FF,$8C,$39,$3E,$43,$FE,$32,$E4,$F9,$0F,$F8
	.dc.b	$D3,$93,$E4,$3F,$E3,$6E,$4F,$84,$FF,$96,$B9,$1E,$27,$E3,$A1,$E9
	.dc.b	$A7,$31,$B3,$E4,$9F,$C8,$B9,$3E,$4B,$FC,$93,$93,$E4,$BF,$C9,$B9
	.dc.b	$3E,$4B,$FC,$A3,$93,$E4,$BF,$CA,$B9,$3E,$4B,$FC,$B3,$93,$E1,$3F
	.dc.b	$20,$FF,$26,$F1,$7C,$9C,$FC,$9F,$15,$F8,$12,$4F,$AF,$E2,$BF,$09
	.dc.b	$F0,$9F,$85,$F8,$3F,$C3,$7C,$2F,$E1,$FD,$FF,$C4,$7C,$37,$E2,$7E
	.dc.b	$07,$F1,$5F,$0F,$F8,$BF,$83,$FC,$67,$C4,$7E,$37,$E1,$7F,$1D,$F1
	.dc.b	$3F,$8F,$F8,$7F,$C8,$7C,$57,$F8,$3B,$C2,$FE,$D7,$C3,$FC,$6C,$FC
	.dc.b	$91,$F3,$63,$83,$F3,$EB,$A3,$F4,$70,$F0,$3F,$4F,$2F,$23,$F5,$74
	.dc.b	$F4,$3F,$5F,$6F,$63,$F6,$78,$3F,$45,$C3,$F0,$A3,$FC,$41,$D6,$E3
	.dc.b	$F9,$1E,$37,$ED,$E7,$E5,$90,$FE,$B7,$E2,$7E,$DD,$FC,$81,$F9,$D3
	.dc.b	$F2,$07,$E8,$4F,$C8,$1F,$A5,$3F,$20,$7E,$A4,$FC,$81,$FA,$D3,$F2
	.dc.b	$07,$E2,$4F,$C8,$77,$7E,$3F,$9F,$F1,$CF,$E5,$23,$FB,$6F,$7F,$E6
	.dc.b	$1F,$ED,$FE,$07,$E7,$1F,$EE,$3E,$0F,$E8,$1F,$EE,$7E,$17,$E9,$1F
	.dc.b	$EE,$BE,$1F,$EA,$1F,$EE,$FE,$27,$B5,$F7,$9F,$07,$9F,$F6,$2F,$F1
	.dc.b	$CE,$EF,$DA,$78,$FF,$E0,$57,$F2,$73,$FE,$A1,$F0,$3F,$2E,$FF,$A8
	.dc.b	$FC,$1F,$CD,$BF,$EA,$5F,$0B,$F3,$EF,$FA,$9F,$C3,$FD,$1B,$FE,$A9
	.dc.b	$F1,$3F,$4E,$FF,$AA,$FC,$5F,$D5,$AC,$BF,$CC,$9C,$FF,$C7,$34,$FF
	.dc.b	$07,$F8,$DF,$C7,$39,$24,$7E,$38,$FB,$E7,$2A,$FF,$0E,$61,$D9,$FF
	.dc.b	$0F,$63,$E2,$7F,$C4,$19,$79,$9F,$F1,$16,$7E,$A7,$FC,$49,$A7,$B9
	.dc.b	$FF,$13,$6B,$F0,$9F,$89,$8F,$E2,$7E,$21,$58,$8F,$DC,$13,$6F,$F6
	.dc.b	$47,$3F,$D1,$7F,$D9,$5C,$3F,$55,$FF,$66,$71,$7D,$97,$FD,$9D,$C7
	.dc.b	$F7,$5F,$F6,$87,$27,$E0,$5F,$F6,$97,$2F,$E1,$5F,$F6,$A7,$37,$E2
	.dc.b	$5F,$95,$97,$F1,$AE,$5F,$E3,$2F,$FA,$0F,$D7,$FF,$1B,$74,$7F,$A1
	.dc.b	$E3,$F2,$67,$E5,$4F,$C8,$9F,$99,$3F,$22,$7E,$74,$FC,$89,$FA,$13
	.dc.b	$F2,$27,$E9,$4F,$C8,$9F,$A9,$3F,$22,$7F,$3C,$3F,$22,$7C,$5E,$0F
	.dc.b	$91,$C1,$F2,$A3,$D1,$D5,$FB,$BC,$7D,$77,$FC,$82,$7F,$8F,$1F,$C0
	.dc.b	$4F,$E7,$5E,$47,$F1,$E7,$F9,$E7,$91,$FC,$79,$FE,$7D,$E4,$7F,$1E
	.dc.b	$7F,$A0,$79,$1F,$C7,$9F,$E8,$5E,$47,$F1,$E7,$FA,$27,$91,$FC,$79
	.dc.b	$FC,$76,$5F,$C7,$89,$AF,$F8,$4E,$3F,$87,$F9,$1F,$C7,$9F,$E2,$3E
	.dc.b	$47,$F1,$E7,$F8,$9F,$91,$FC,$79,$FE,$2B,$E4,$7F,$1E,$7F,$8B,$F9
	.dc.b	$1F,$C7,$9F,$E3,$3E,$47,$F1,$E7,$F8,$DF,$91,$FC,$79,$FE,$B1,$C7
	.dc.b	$FC,$7F,$93,$E4,$72,$7F,$B2,$BC,$6F,$DC,$D7,$F6,$E3,$FC,$26,$7D
	.dc.b	$1D,$FF,$22,$FD,$57,$F2,$2F,$D9,$7F,$22,$FD,$D7,$F2,$2F,$E0,$5F
	.dc.b	$C8,$BF,$85,$7F,$22,$FE,$25,$FA,$FC,$FF,$90,$7E,$CF,$3F,$E4,$0F
	.dc.b	$C8,$C6,$78,$7F,$7F,$5F,$B9,$3F,$DB,$FE,$27,$EE,$5F,$F7,$0F,$89
	.dc.b	$FB,$97,$FD,$C7,$E2,$7E,$E5,$FF,$72,$F8,$9F,$B9,$7F,$DC,$FE,$1F
	.dc.b	$EE,$5F,$F7,$4F,$89,$FB,$97,$E2,$1F,$B6,$DF,$E4,$E7,$E3,$8F,$D4
	.dc.b	$78,$BF,$E2,$D2,$3F,$A6,$3F,$A6,$B8,$69,$9D,$BE,$0C,$65,$B1,$9D
	.dc.b	$7F,$8E,$CE,$5B,$19,$3C,$27,$FB,$57,$C9,$F2,$70,$FF,$46,$8D,$35
	.dc.b	$F0,$92,$FB,$B6,$BA,$B4,$EA,$59,$E5,$3C,$09,$3C,$1B,$3C,$31,$F8
	.dc.b	$F8,$7A,$63,$9C,$EA,$F9,$DF,$12,$4E,$9B,$1A,$8C,$11,$C1,$31,$F8
	.dc.b	$24,$BF,$04,$7F,$19,$9F,$E3,$9E,$2E,$2B,$51,$F1,$9A,$FB,$DB,$87
	.dc.b	$9D,$B8,$79,$1B,$3E,$37,$F9,$4E,$3F,$86,$F8,$9F,$F3,$81,$27,$FC
	.dc.b	$95,$EF,$7F,$9C,$FD,$9F,$F3,$A7,$BD,$FE,$75,$F8,$5F,$F3,$AE,$7F
	.dc.b	$C6,$BC,$7F,$E6,$B8,$23,$1E,$B6,$B5,$4F,$F9,$4A,$9F,$01,$98,$C7
	.dc.b	$B1,$78,$54,$F8,$1C,$2F,$6A,$FC,$16,$3D,$CC,$FC,$1E,$BD,$EF,$F1
	.dc.b	$D6,$BF,$ED,$4D,$FF,$8E,$EB,$FE,$FA,$3C,$6F,$A6,$78,$C6,$27,$CF
	.dc.b	$A7,$8B,$0E,$F3,$93,$D5,$97,$93,$4F,$2E,$BE,$6D,$BC,$FB,$FA,$32
	.dc.b	$F9,$18,$7F,$9D,$78,$7E,$63,$F3,$8F,$F3,$C7,$27,$D1,$7F,$DF,$DA
	.dc.b	$7E,$1F,$A3,$F0,$44,$8F,$C1,$35,$FC,$13,$77,$FC,$13,$70,$F5,$F8
	.dc.b	$1F,$C5,$FA,$71,$7F,$83,$AB,$87,$78,$EA,$43,$A9,$70,$FD,$9F,$47
	.dc.b	$EA,$23,$FC,$38,$7E,$C8,$FF,$5C,$CF,$EE,$DB,$D3,$FD,$E1,$B7,$C1
	.dc.b	$E0,$7E,$F0,$C9,$D2,$79,$39,$21,$C0,$8E,$9B,$3D,$DD,$C6,$4F,$D7
	.dc.b	$5F,$F3,$AC,$3F,$90,$4F,$80,$FF,$1B,$4E,$EB,$94,$74,$BF,$44,$E5
	.dc.b	$B3,$A8,$C4,$FA,$63,$04,$9D,$92,$34,$FF,$B8,$36,$7C,$27,$A2,$B8
	.dc.b	$56,$88,$79,$9C,$B6,$BA,$34,$D9,$ED,$21,$D5,$DB,$83,$E5,$E7,$F8
	.dc.b	$77,$FA,$31,$9C,$3D,$AF,$52,$19,$AF,$1B,$6B,$AB,$DC,$F1,$94,$FE
	.dc.b	$07,$75,$E0,$68,$FA,$05,$AF,$0B,$A3,$C2,$71,$21,$F8,$4E,$BF,$E9
	.dc.b	$E7,$F2,$87,$81,$0D,$C7,$46,$DF,$A9,$B7,$F3,$31,$FA,$B1,$E4,$3C
	.dc.b	$25,$3A,$14,$E2,$7A,$67,$A3,$93,$E7,$E5,$FC,$87,$2F,$F0,$5D,$FC
	.dc.b	$89,$D6,$1E,$56,$DF,$28,$C1,$1C,$1F,$E6,$6C,$3C,$F1,$FB,$D9,$FD
	.dc.b	$D9,$33,$FE,$0E,$FE,$F7,$5F,$E9,$18,$7F,$04,$C3,$FA,$4D,$F1,$BF
	.dc.b	$D4,$5F,$F2,$38,$FD,$5D,$F0,$8E,$47,$F0,$D2,$C7,$18,$D4,$78,$5B
	.dc.b	$3F,$C3,$3B,$7F,$98,$F4,$FC,$EA,$FE,$99,$58,$DB,$FA,$19,$FD,$13
	.dc.b	$E0,$3F,$A2,$FC,$27,$F4,$6F,$84,$FE,$8F,$F0,$3F,$D2,$3E,$13,$FA
	.dc.b	$4F,$C1,$7F,$4A,$E2,$FE,$97,$F0,$5F,$D3,$3E,$13,$FA,$6F,$C1,$FF
	.dc.b	$51,$FD,$A1,$F0,$F7,$E2,$31,$F0,$D9,$F8,$9D,$7C,$3F,$FB,$7E,$CE
	.dc.b	$79,$3E,$24,$2B,$94,$63,$0E,$EF,$C9,$E5,$67,$E4,$8C,$6B,$E8,$75
	.dc.b	$BF,$B1,$E1,$C3,$F0,$1E,$5C,$7F,$11,$E9,$C9,$FD,$C1,$FE,$97,$D3
	.dc.b	$F6,$89,$D5,$F7,$49,$1F,$5A,$32,$FD,$8E,$DF,$D4,$CE,$A3,$EC,$1F
	.dc.b	$74,$BD,$7F,$75,$C7,$5F,$E3,$1F,$C7,$F7,$7F,$9F,$37,$FB,$93,$FE
	.dc.b	$09,$CB,$D1,$5F,$78,$FF,$03,$67,$ED,$71,$87,$F8,$13,$4F,$85,$18
	.dc.b	$C1,$AF,$C7,$8C,$6D,$F2,$DD,$6F,$F4,$1E,$7C,$5F,$79,$FA,$FC,$BF
	.dc.b	$79,$FB,$3C,$BF,$79,$FC,$F5,$FF,$23,$78,$F2,$FD,$F7,$9F,$2F,$DE
	.dc.b	$7A,$F2,$F7,$E1,$FF,$13,$67,$E3,$7F,$C5,$18,$79,$DF,$F1,$56,$9E
	.dc.b	$C7,$FC,$59,$B7,$C1,$7F,$C5,$DC,$1F,$19,$F9,$18,$FF,$8B,$38,$BF
	.dc.b	$D4,$FB,$BE,$2B,$F3,$44,$9F,$C9,$FC,$37,$E7,$7D,$3F,$CF,$7C,$37
	.dc.b	$E7,$FD,$0F,$D0,$78,$7F,$A1,$F0,$BF,$45,$F0,$DF,$A3,$F4,$BF,$49
	.dc.b	$F0,$DF,$A5,$F5,$3F,$4D,$F0,$DF,$A7,$F5,$BF,$51,$DB,$FA,$9F,$5F
	.dc.b	$E4,$70,$7F,$96,$8F,$93,$A6,$39,$3F,$DC,$3C,$6A,$E3,$5F,$A8,$C3
	.dc.b	$AB,$FA,$AC,$BC,$2F,$EB,$34,$F2,$BF,$AE,$DB,$D2,$FE,$C3,$7F,$6B
	.dc.b	$FB,$2E,$1F,$DE,$71,$7C,$38,$FC,$0F,$1F,$E9,$79,$3E,$4F,$C3,$7F
	.dc.b	$01,$24,$FF,$1D,$FD,$A7,$F5,$4F,$E4,$9F,$F4,$C8,$F6,$BF,$CB,$0B
	.dc.b	$E2,$6F,$B9,$8F,$88,$CF,$BD,$AF,$88,$DF,$C0,$70,$F8,$8E,$3F,$05
	.dc.b	$CB,$E2,$39,$FC,$27,$4F,$88,$F9,$1E,$BF,$F8,$EB,$BB,$FA,$31,$8E
	.dc.b	$17,$9C,$E1,$5F,$29,$F5,$63,$D9,$F6,$67,$E2,$7D,$DA,$F9,$9F,$81
	.dc.b	$BF,$A9,$F8,$5C,$3E,$EE,$2F,$7F,$A1,$F0,$27,$E3,$9F,$AC,$E4,$F9
	.dc.b	$9D,$5F,$E6,$AE,$FF,$CA,$4F,$F2,$4E,$8F,$A5,$E0,$7E,$0B,$AB,$F2
	.dc.b	$93,$FC,$A3,$C5,$FC,$A3,$FC,$AB,$93,$F2,$8F,$F2,$CF,$17,$F2,$8F
	.dc.b	$F2,$EF,$17,$F2,$8F,$F3,$0F,$17,$F2,$8F,$F3,$2F,$17,$F2,$8F,$FB
	.dc.b	$3B,$D1,$FE,$36,$FF,$B4,$35,$FC,$93,$B3,$F2,$47,$33,$F2,$47,$73
	.dc.b	$F2,$47,$8C,$FC,$91,$E7,$3F,$24,$7A,$CF,$C9,$1F,$B2,$3E,$11,$0F
	.dc.b	$F3,$FA,$DF,$E5,$EB,$FA,$67,$E7,$4F,$E4,$F8,$3E,$93,$F5,$0F,$CB
	.dc.b	$77,$FE,$3D,$FC,$C7,$7F,$E3,$DF,$CD,$77,$FE,$3D,$FC,$E7,$7F,$E3
	.dc.b	$DF,$CF,$77,$FE,$3D,$FE,$B3,$E7,$7F,$5D,$30,$FE,$86,$38,$3F,$A3
	.dc.b	$9E,$8F,$E9,$6B,$C0,$FE,$9E,$FC,$8F,$EA,$70,$FD,$3F,$47,$AA,$3F
	.dc.b	$59,$97,$B9,$FC,$51,$FD,$AB,$3F,$C5,$BF,$9E,$D7,$F1,$6F,$EE,$8F
	.dc.b	$95,$AF,$E2,$DF,$9E,$7F,$3E,$E4,$FC,$5B,$FD,$03,$93,$F1,$6F,$F4
	.dc.b	$2E,$4F,$C5,$BF,$D1,$39,$3F,$16,$FF,$46,$E4,$FC,$5B,$FD,$23,$93
	.dc.b	$F1,$6F,$F4,$23,$F8,$E9,$0C,$1F,$D3,$BC,$7F,$E3,$AF,$F5,$0F,$1F
	.dc.b	$F8,$EB,$FD,$4B,$C7,$FE,$3A,$FF,$54,$F1,$FF,$8E,$BF,$D5,$BA,$FF
	.dc.b	$8E,$BF,$D6,$3C,$7F,$E3,$AF,$F5,$AF,$1F,$F8,$EB,$F0,$BC,$6F,$D8
	.dc.b	$77,$7C,$8F,$19,$C9,$F8,$DA,$F9,$84,$3F,$8E,$1F,$BC,$3F,$90,$78
	.dc.b	$1F,$59,$FC,$69,$FC,$10,$FC,$69,$FC,$20,$FC,$69,$FC,$30,$FC,$69
	.dc.b	$FC,$40,$FC,$69,$FC,$AF,$C5,$FE,$31,$5F,$E4,$88,$FE,$33,$E5,$B9
	.dc.b	$73,$E3,$FF,$1A,$9E,$27,$F3,$1F,$1F,$F8,$D3,$FC,$CF,$C7,$FE,$34
	.dc.b	$FF,$35,$F1,$FF,$8D,$3F,$CD,$FC,$7F,$E3,$4F,$F3,$9E,$FF,$C8,$76
	.dc.b	$B8,$7F,$DF,$F9,$7F,$3C,$AF,$D8,$C7,$F1,$9E,$EF,$95,$C3,$F9,$0B
	.dc.b	$FE,$B3,$C5,$FC,$6A,$BF,$8F,$78,$FF,$C6,$B0,$FC,$81,$FC,$08,$FC
	.dc.b	$81,$FC,$18,$FC,$81,$FC,$28,$FC,$81,$FC,$38,$FC,$81,$FC,$48,$FC
	.dc.b	$81,$FE,$4C,$EA,$FF,$1E,$5F,$A6,$3F,$A0,$FA,$5F,$C7,$1F,$F1,$3F
	.dc.b	$91,$FE,$78,$CB,$F8,$EC,$72,$3F,$24,$7F,$1D,$3F,$24,$7F,$1D,$3F
	.dc.b	$24,$7F,$1D,$3F,$24,$7F,$1D,$3F,$24,$7F,$1D,$3F,$24,$7F,$1D,$3F
	.dc.b	$68,$4F,$27,$FB,$6B,$F8,$FB,$FE,$38,$FF,$6D,$3F,$20,$7F,$6D,$E7
	.dc.b	$FC,$83,$FC,$1C,$FC,$81,$FC,$2C,$FC,$81,$FC,$3C,$FC,$81,$FC,$4C
	.dc.b	$FC,$81,$FC,$5C,$FC,$81,$FC,$70,$FD,$B4,$7F,$A6,$9F,$DB,$BF,$2F
	.dc.b	$9B,$FB,$57,$3F,$F6,$99,$FC,$81,$F9,$C3,$F2,$07,$E8,$0F,$C8,$1F
	.dc.b	$A4,$3F,$20,$7E,$A0,$FC,$81,$FA,$C3,$F9,$D1,$F0,$4F,$89,$AF,$E3
	.dc.b	$DF,$F1,$E7,$57,$E5,$3E,$1B,$FC,$1A,$48,$FF,$0A,$7C,$47,$F8,$43
	.dc.b	$E2,$3F,$C2,$3E,$E7,$F8,$4B,$E2,$3F,$C2,$7F,$11,$FE,$14,$F7,$FF
	.dc.b	$C2,$BF,$11,$F8,$8D,$19,$7F,$82,$3F,$6A,$FC,$53,$FE,$6E,$5F,$05
	.dc.b	$B8,$D7,$C1,$5E,$15,$3E,$0F,$0B,$E1,$2F,$C2,$E3,$E2,$73,$F0,$FF
	.dc.b	$E1,$8C,$BE,$8E,$9E,$2C,$B9,$CE,$E7,$FD,$99,$8B,$0E,$B8,$F6,$CB
	.dc.b	$BE,$7E,$0D,$3C,$39,$FB,$B1,$FF,$1C,$6D,$F0,$9F,$F1,$D7,$07,$C6
	.dc.b	$7F,$C7,$9C,$5F,$E3,$8E,$3F,$99,$1F,$E3,$DE,$4F,$E5,$1E,$09,$CF
	.dc.b	$47,$E1,$63,$13,$F4,$8E,$67,$F9,$1F,$9B,$E9,$3F,$E4,$9E,$6F,$A4
	.dc.b	$FF,$92,$F9,$BC,$EF,$BB,$6F,$5B,$F0,$38,$3D,$EF,$C2,$E2,$F8,$8F
	.dc.b	$D5,$E4,$F9,$4F,$F9,$B4,$C1,$F2,$39,$BE,$8B,$F2,$B9,$3E,$43,$F3
	.dc.b	$39,$3E,$43,$F3,$B9,$3E,$43,$FE,$5D,$E4,$F9,$0F,$F9,$7F,$93,$E4
	.dc.b	$3F,$E6,$1E,$4F,$90,$FF,$99,$78,$FE,$4F,$53,$A3,$EE,$73,$BE,$27
	.dc.b	$F0,$A4,$BF,$BB,$F1,$3F,$88,$F8,$4F,$C4,$FB,$BF,$8A,$F8,$9F,$C5
	.dc.b	$FB,$FF,$8C,$F8,$9F,$C6,$FC,$0F,$E3,$BE,$27,$F1,$FE,$8F,$E4,$3E
	.dc.b	$27,$F2,$3F,$09,$F9,$2F,$89,$FC,$B4,$7F,$AF,$76,$FF,$64,$AF,$89
	.dc.b	$FE,$A3,$E5,$7F,$27,$F1,$7F,$D2,$7F,$0D,$C2,$78,$FC,$3F,$3C,$3F
	.dc.b	$27,$97,$77,$F2,$B9,$F8,$DF,$CB,$EB,$E7,$7F,$33,$BF,$AD,$FC,$DF
	.dc.b	$0F,$BD,$F8,$27,$E7,$B9,$3E,$22,$E7,$F9,$4F,$F9,$F3,$C7,$FC,$7B
	.dc.b	$F9,$53,$E8,$1F,$91,$3E,$A1,$F9,$13,$F3,$E7,$E4,$4F,$D1,$9F,$91
	.dc.b	$3F,$00,$7E,$44,$FC,$21,$F9,$13,$F1,$07,$E4,$4F,$C6,$1F,$B4,$F0
	.dc.b	$7F,$23,$BF,$F2,$93,$F9,$47,$42,$3F,$2F,$18,$F0,$BF,$37,$5F,$C3
	.dc.b	$7C,$3F,$CF,$BF,$C3,$FC,$4F,$D1,$BF,$EA,$DF,$17,$F4,$EF,$FA,$BF
	.dc.b	$C6,$FD,$5B,$FE,$B1,$F1,$FF,$90,$3F,$07,$1F,$DA,$78,$1F,$92,$F0
	.dc.b	$3F,$94,$4F,$C6,$3E,$41,$FE,$B9,$F2,$38,$BC,$CF,$8E,$77,$3E,$39
	.dc.b	$E3,$3E,$39,$E7,$3E,$39,$EB,$3E,$39,$FC,$A7,$4F,$83,$E2,$FE,$8B
	.dc.b	$6F,$8E,$FF,$AE,$67,$E5,$1F,$30,$E0,$72,$C7,$E5,$ED,$F1,$9F,$9B
	.dc.b	$B7,$C6,$7F,$89,$6D,$F1,$9F,$E2,$9B,$7C,$67,$F8,$B6,$DE,$E7,$F8
	.dc.b	$C6,$9F,$C6,$B2,$F8,$B9,$7E,$3F,$5B,$5F,$2A,$3E,$43,$C1,$FE,$3A
	.dc.b	$7C,$73,$F8,$21,$F8,$D3,$F8,$41,$F8,$D3,$FA,$54,$FF,$0E,$7F,$A6
	.dc.b	$78,$DF,$C4,$9F,$E9,$DE,$3F,$F1,$67,$FA,$87,$91,$FD,$19,$FE,$8A
	.dc.b	$7F,$91,$3D,$7C,$3A,$3F,$28,$72,$3F,$30,$76,$3F,$38,$78,$8F,$26
	.dc.b	$1F,$CE,$7C,$9F,$E3,$AF,$F3,$BF,$27,$F8,$EB,$FC,$F7,$C9,$FE,$3B
	.dc.b	$5F,$88,$8F,$88,$FC,$73,$FB,$47,$4F,$EE,$23,$F4,$2F,$F6,$2F,$27
	.dc.b	$F7,$D3,$FD,$AF,$AF,$F8,$13,$FD,$B7,$B3,$F8,$33,$FD,$BF,$B7,$F8
	.dc.b	$53,$FD,$C7,$BB,$F8,$73,$FD,$CF,$BF,$F8,$93,$FC,$6B,$C0,$FE,$2C
	.dc.b	$FF,$86,$CF,$F1,$FF,$C1,$7F,$75,$F1,$0F,$F9,$BB,$C5,$FE,$41,$3F
	.dc.b	$E0,$0D,$BF,$C3,$D1,$FC,$95,$FD,$CF,$8D,$FC,$AA,$7F,$B8,$CF,$F2
	.dc.b	$E7,$FB,$9F,$8F,$FC,$C9,$FE,$EB,$E4,$7F,$36,$7F,$BB,$F9,$3F,$CE
	.dc.b	$9F,$EF,$3E,$57,$C3,$7F,$8F,$10,$F9,$5E,$5B,$C8,$FE,$3B,$9F,$CE
	.dc.b	$27,$6F,$F9,$D3,$BF,$F3,$4F,$F9,$DB,$C0,$FC,$F3,$FE,$78,$F0,$7F
	.dc.b	$44,$FF,$9E,$BC,$2F,$D3,$3F,$D4,$BC,$3F,$D5,$3F,$D5,$3C,$4F,$E2
	.dc.b	$78,$FF,$15,$D3,$F6,$4F,$F1,$AF,$17,$F6,$F1,$F9,$13,$F6,$E7,$E4
	.dc.b	$4F,$DB,$9F,$91,$3F,$81,$1F,$91,$3F,$83,$1F,$91,$3F,$85,$1F,$A6
	.dc.b	$3F,$87,$1F,$AA,$3F,$59,$1F,$91,$7F,$8E,$57,$ED,$0F,$DB,$F9,$3F
	.dc.b	$D1,$36,$71,$BE,$23,$FA,$F9,$2F,$EF,$5F,$F8,$B6,$75,$18,$15,$CA
	.dc.b	$B8,$47,$F6,$66,$E1,$F8,$66,$DF,$36,$CF,$2B,$EA,$F1,$3B,$5F,$56
	.dc.b	$C6,$AF,$A4,$7C,$C7,$FA,$55,$FF,$25,$8F,$F0,$D9,$CC,$91,$C1,$B1
	.dc.b	$E3,$7A,$DF,$92,$FC,$11,$A3,$CD,$19,$BF,$15,$71,$9D,$57,$0B,$ED
	.dc.b	$87,$CF,$BD,$D7,$23,$B9,$0E,$2E,$AB,$95,$F5,$BF,$41,$1C,$12,$BA
	.dc.b	$1C,$B0,$F3,$1F,$E1,$AF,$89,$FC,$9D,$FF,$B9,$CF,$C0,$BA,$9F,$15
	.dc.b	$F5,$79,$B6,$7C,$71,$8B,$F0,$46,$09,$87,$A4,$98,$75,$AF,$DB,$C3
	.dc.b	$CD,$FF,$16,$92,$3E,$EF,$A9,$FE,$30,$F5,$3F,$C6,$3E,$DF,$F8,$CB
	.dc.b	$E1,$3F,$C6,$7E,$EF,$F8,$D3,$CA,$FF,$1A,$F9,$DF,$E3,$6F,$84,$FF
	.dc.b	$1B,$FB,$FF,$E3,$8F,$84,$FF,$1C,$FC,$0F,$F8,$EB,$E1,$3F,$C7,$7F
	.dc.b	$07,$FE,$3C,$F8,$4F,$E8,$7C,$DF,$EB,$7F,$1F,$FC,$75,$CF,$FD,$BF
	.dc.b	$8F,$F1,$F5,$C1,$FC,$8E,$1D,$1F,$C9,$E5,$E0,$7F,$2B,$A7,$91,$FC
	.dc.b	$BE,$DE,$87,$F3,$3C,$1E,$C7,$F3,$7C,$5F,$E4,$EE,$3F,$85,$1F,$E4
	.dc.b	$EE,$4F,$F5,$6B,$F2,$39,$BE,$56,$FF,$4F,$9B,$E6,$CF,$D8,$DF,$E9
	.dc.b	$C6,$23,$F0,$46,$4F,$B0,$4E,$7F,$AF,$7F,$85,$8F,$EC,$F1,$F8,$59
	.dc.b	$74,$78,$2F,$AD,$DB,$E5,$5D,$F0,$F2,$4E,$70,$EB,$5F,$9D,$E9,$FD
	.dc.b	$14,$FB,$23,$DB,$3F,$AE,$6D,$FF,$02,$6C,$ED,$FD,$71,$83,$13,$C5
	.dc.b	$C9,$0F,$D1,$9B,$3C,$0E,$DC,$9B,$70,$49,$F1,$47,$F2,$1C,$7F,$53
	.dc.b	$E5,$37,$F9,$13,$FC,$34,$EC,$43,$A5,$72,$7B,$92,$FB,$61,$ED,$8C
	.dc.b	$4E,$E7,$31,$EE,$8C,$12,$7F,$0B,$EA,$BC,$4E,$F1,$CE,$F8,$DF,$C0
	.dc.b	$BE,$0F,$39,$D4,$62,$35,$18,$23,$B2,$1A,$29,$EF,$3F,$9D,$65,$F1
	.dc.b	$B1,$F8,$BA,$7F,$2F,$9D,$61,$E4,$6E,$1E,$16,$E1,$E0,$7B,$10,$C9
	.dc.b	$E0,$3F,$98,$65,$F0,$5F,$93,$1C,$5D,$9C,$4F,$8A,$71,$36,$5A,$F2
	.dc.b	$3A,$38,$1F,$D7,$BD,$8F,$F4,$A3,$FD,$53,$9F,$F8,$3B,$9E,$EF,$EC
	.dc.b	$2F,$80,$87,$32,$9E,$03,$AD,$78,$1C,$61,$EB,$8C,$13,$1F,$8D,$3C
	.dc.b	$C8,$73,$3A,$C6,$F1,$FB,$BA,$7D,$39,$FE,$D2,$F9,$9E,$33,$B7,$E1
	.dc.b	$57,$67,$A1,$A7,$11,$FD,$92,$FC,$D1,$FC,$07,$83,$F8,$2D,$7F,$0F
	.dc.b	$AF,$E0,$2F,$F1,$22,$69,$FC,$19,$D5,$71,$7F,$85,$E1,$E7,$7F,$DD
	.dc.b	$EF,$F0,$B8,$D5,$74,$AE,$55,$D2,$FC,$2D,$3F,$8A,$E3,$FC,$AB,$83
	.dc.b	$FB,$97,$BD,$FE,$FC,$24,$FF,$83,$3D,$FF,$E5,$F8,$30,$4F,$F8,$7B
	.dc.b	$F6,$BF,$E3,$13,$98,$C7,$9B,$C2,$B7,$9A,$DE,$F6,$6F,$97,$C0,$DA
	.dc.b	$D7,$C3,$5B,$7C,$1D,$F8,$7C,$7C,$2E,$7E,$23,$5F,$0F,$BF,$89,$E1
	.dc.b	$F1,$3C,$7E,$2B,$F5,$FC,$1F,$E2,$1E,$2F,$F1,$1F,$07,$ED,$0F,$F1
	.dc.b	$5E,$AA,$E3,$18,$C3,$AB,$AC,$BC,$2F,$0D,$3C,$DA,$F9,$F6,$F4,$70
	.dc.b	$7A,$78,$7D,$5C,$5E,$BC,$3D,$DC,$EE,$4F,$9F,$0D,$FF,$D1,$3C,$BF
	.dc.b	$57,$CB,$FB,$31,$F7,$09,$C5,$82,$57,$DC,$3F,$13,$D1,$F8,$53,$FB
	.dc.b	$DC,$7D,$2E,$CF,$DA,$F4,$3B,$7F,$73,$7F,$85,$8F,$C5,$57,$46,$C7
	.dc.b	$53,$5A,$7E,$40,$FE,$D9,$D5,$F8,$A8,$FF,$29,$E9,$EC,$CF,$FD,$47
	.dc.b	$AF,$C1,$9F,$9F,$B7,$C5,$7E,$8F,$07,$C9,$7E,$9F,$17,$DF,$7E,$AF
	.dc.b	$57,$DF,$7E,$BF,$57,$DF,$7E,$CF,$57,$DF,$7F,$76,$7F,$78,$F0,$3F
	.dc.b	$AA,$38,$3E,$77,$07,$E5,$3A,$FF,$01,$1C,$8F,$C0,$1D,$8F,$C0,$1E
	.dc.b	$23,$F0,$07,$98,$FC,$4F,$3F,$E0,$1F,$C5,$F3,$FE,$01,$FC,$69,$F8
	.dc.b	$0D,$BE,$25,$FF,$A2,$78,$3E,$F4,$FE,$AF,$E2,$3F,$40,$47,$F2,$BF
	.dc.b	$0F,$FA,$3F,$57,$F4,$9F,$05,$FA,$5F,$5B,$F4,$DF,$07,$FA,$7F,$5F
	.dc.b	$F5,$1F,$09,$FA,$9F,$63,$F5,$9F,$B6,$7E,$D7,$3F,$82,$7F,$A5,$97
	.dc.b	$B3,$AF,$23,$7E,$CF,$0F,$87,$E3,$F1,$1F,$AA,$C3,$FD,$70,$72,$F0
	.dc.b	$BF,$C4,$9E,$5E,$A3,$FD,$2B,$92,$F9,$C6,$31,$F1,$E5,$E4,$E5,$F3
	.dc.b	$4F,$EC,$B5,$F5,$3F,$B4,$DF,$DC,$FE,$DB,$87,$E1,$3F,$10,$FD,$CF
	.dc.b	$27,$C7,$7F,$DC,$39,$BE,$23,$E0,$92,$3C,$BE,$6F,$C4,$F8,$8F,$8B
	.dc.b	$F1,$1F,$1B,$D1,$F8,$FF,$11,$F2,$3E,$07,$E4,$FC,$47,$CA,$F8,$3F
	.dc.b	$97,$F1,$1F,$33,$E1,$7E,$6F,$8B,$F3,$BC,$EF,$9F,$F1,$1F,$43,$E1
	.dc.b	$FF,$8E,$7E,$DA,$F8,$9F,$E7,$5E,$8F,$F2,$7E,$F4,$E7,$93,$E4,$DF
	.dc.b	$27,$E5,$E3,$D9,$F9,$B9,$F8,$9F,$9F,$AF,$99,$FA,$3B,$FA,$9F,$A7
	.dc.b	$C3,$EE,$7E,$AF,$1F,$E1,$39,$3E,$24,$7D,$BE,$6F,$90,$FF,$3E,$F4
	.dc.b	$BF,$DC,$9E,$63,$A7,$E7,$70,$FA,$63,$15,$F4,$CF,$C2,$2F,$1F,$F2
	.dc.b	$F1,$FC,$BB,$C7,$FC,$BB,$FC,$C3,$C7,$FC,$BB,$FC,$CB,$C7,$FC,$23
	.dc.b	$F8,$AF,$0B,$F2,$8F,$E3,$3C,$7F,$CA,$3F,$8E,$F1,$FF,$29,$D0,$F5
	.dc.b	$3F,$8E,$F3,$FF,$9D,$AB,$47,$E5,$3C,$7F,$C9,$BE,$FF,$1F,$F2,$6F
	.dc.b	$E6,$3C,$7F,$C9,$BE,$AF,$1F,$F2,$6F,$B3,$C7,$FC,$9B,$EE,$F1,$FF
	.dc.b	$26,$FC,$0F,$27,$E2,$DF,$F0,$DE,$6F,$DE,$EA,$E8,$F9,$B7,$F3,$FC
	.dc.b	$8F,$A3,$D1,$F8,$39,$FE,$66,$DE,$BF,$B1,$5F,$57,$B3,$EE,$3F,$5F
	.dc.b	$B7,$F0,$0F,$D9,$EE,$FC,$23,$F6,$FB,$FF,$10,$FD,$DE,$4F,$C6,$3F
	.dc.b	$7F,$C0,$FC,$AF,$85,$F9,$4E,$0F,$EE,$FE,$17,$E4,$EB,$47,$EF,$3C
	.dc.b	$3F,$CF,$3F,$BE,$F1,$3F,$44,$FF,$00,$F1,$7F,$4C,$FF,$02,$F1,$BF
	.dc.b	$54,$FF,$04,$F1,$FF,$5C,$FF,$06,$F2,$3F,$64,$FE,$D0,$FE,$13,$CB
	.dc.b	$F9,$4F,$1F,$F2,$3D,$1F,$32,$7F,$93,$1F,$8F,$E6,$FC,$97,$8D,$F8
	.dc.b	$E8,$FC,$8F,$8D,$F8,$E7,$F2,$7E,$37,$E3,$9F,$CA,$F8,$DF,$8E,$7F
	.dc.b	$41,$E3,$7E,$39,$FD,$17,$8D,$F8,$E7,$F4,$9E,$37,$E3,$B8,$BF,$A5
	.dc.b	$C7,$F2,$16,$E7,$FB,$BE,$8F,$C8,$4F,$03,$F2,$07,$E6,$4F,$E4,$47
	.dc.b	$E7,$4F,$C8,$1F,$A1,$3F,$91,$1F,$A5,$3F,$91,$1F,$A9,$3F,$91,$1F
	.dc.b	$E1,$8D,$3F,$C4,$DE,$07,$C5,$F0,$3F,$C5,$DB,$FE,$43,$C0,$FC,$7E
	.dc.b	$3F,$8D,$E2,$FC,$7F,$4F,$F8,$D2,$7F,$1E,$DF,$03,$F1,$F5,$F9,$3F
	.dc.b	$03,$F1,$EF,$E5,$7C,$0F,$C7,$BF,$97,$F0,$3F,$1E,$FE,$67,$C0,$FC
	.dc.b	$7B,$F9,$BE,$0F,$C7,$BF,$9D,$F0,$3F,$1F,$C3,$FD,$D7,$DB,$AF,$E3
	.dc.b	$8F,$E2,$5D,$FF,$8E,$7F,$8A,$72,$7E,$39,$FE,$2D,$DF,$F8,$E7,$F8
	.dc.b	$C7,$7F,$E3,$9F,$E3,$5D,$FF,$8E,$7F,$8E,$77,$FE,$39,$FE,$3D,$E4
	.dc.b	$FC,$27,$FB,$17,$2F,$ED,$BC,$9F,$E3,$71,$FC,$6B,$9F,$E6,$43,$A7
	.dc.b	$F7,$FC,$1F,$C8,$7C,$4F,$E0,$6F,$F2,$3F,$17,$F8,$3B,$FC,$97,$C6
	.dc.b	$FE,$16,$FE,$F7,$AB,$F8,$7E,$5F,$C4,$4F,$C2,$1F,$CC,$7C,$8F,$ED
	.dc.b	$4F,$F9,$F8,$FF,$56,$7A,$9F,$E6,$5F,$28,$FF,$20,$EE,$78,$3F,$97
	.dc.b	$C7,$F8,$57,$85,$F9,$B7,$F8,$67,$87,$F9,$F7,$F9,$C7,$89,$FA,$37
	.dc.b	$F9,$D7,$8B,$FA,$77,$F9,$E7,$8D,$FA,$B7,$F5,$BA,$7E,$BF,$6F,$E3
	.dc.b	$8F,$F7,$03,$FA,$37,$AF,$F8,$BE,$6F,$99,$E3,$7F,$1F,$7F,$A7,$4F
	.dc.b	$F5,$03,$F8,$01,$FD,$50,$FE,$08,$7F,$58,$3F,$84,$1F,$D7,$0F,$E1
	.dc.b	$87,$F6,$03,$F8,$81,$FD,$90,$FE,$28,$7F,$68,$3F,$D2,$B7,$FD,$B7
	.dc.b	$C9,$FE,$34,$FF,$A0,$29,$DD,$FB,$8C,$7F,$8F,$F9,$5F,$DD,$F0,$E6
	.dc.b	$7F,$27,$3B,$9F,$CA,$CF,$19,$FC,$BC,$F3,$9F,$CC,$CF,$59,$FC,$DC
	.dc.b	$F7,$9F,$DB,$7C,$8F,$86,$FF,$41,$3F,$8F,$9F,$28,$5F,$F8,$13,$E2
	.dc.b	$BF,$C1,$24,$8F,$6D,$FF,$36,$F6,$3F,$C1,$BE,$77,$F8,$3B,$DD,$FF
	.dc.b	$07,$FB,$DF,$E1,$0F,$7F,$FC,$23,$F1,$3F,$E1,$2F,$81,$FF,$09,$FC
	.dc.b	$4F,$F8,$53,$CB,$FF,$0A,$F2,$FF,$85,$BE,$0F,$FC,$2F,$F1,$3F,$E1
	.dc.b	$8F,$85,$FF,$0C,$FC,$4F,$ED,$B9,$FE,$57,$3F,$C9,$E3,$F3,$F4,$7C
	.dc.b	$EE,$4F,$E9,$37,$C6,$7C,$E6,$B1,$EC,$E3,$9B,$FC,$43,$A7,$8E,$31
	.dc.b	$83,$6F,$44,$63,$83,$D8,$EB,$8B,$E0,$3C,$39,$3F,$C5,$FC,$BF,$1E
	.dc.b	$3F,$CA,$D8,$39,$BF,$B9,$B6,$3B,$C6,$F8,$3F,$11,$5F,$32,$7C,$7C
	.dc.b	$DF,$31,$F2,$F3,$7C,$C7,$CF,$CD,$F3,$1F,$4F,$37,$CC,$7D,$7C,$DF
	.dc.b	$31,$F6,$F3,$7C,$C7,$F9,$11,$0F,$3F,$27,$C9,$7D,$3C,$9F,$21,$F5
	.dc.b	$F1,$FC,$77,$DB,$C7,$F1,$9F,$7F,$1F,$C6,$7E,$0F,$1F,$C6,$7E,$1F
	.dc.b	$1F,$C6,$7E,$09,$07,$AE,$C3,$FC,$07,$F0,$FF,$E5,$A2,$4F,$D6,$F8
	.dc.b	$9F,$F2,$EF,$C4,$FF,$97,$BD,$CF,$F2,$FF,$BB,$FE,$FB,$F0,$57,$7C
	.dc.b	$7C,$7B,$D9,$F8,$4D,$7C,$06,$FE,$13,$87,$C3,$71,$F8,$5E,$5F,$0F
	.dc.b	$CF,$E1,$BF,$25,$E2,$FF,$B6,$F3,$FE,$48,$63,$4F,$D4,$EA,$AE,$31
	.dc.b	$F8,$F8,$61,$DA,$31,$97,$89,$D6,$9E,$67,$86,$1E,$B7,$DA,$7E,$5B
	.dc.b	$7F,$82,$FC,$33,$F3,$3C,$5F,$19,$F9,$7C,$7F,$E2,$CF,$0F,$F1,$DA
	.dc.b	$D9,$FC,$FF,$85,$F8,$F7,$F2,$FE,$17,$E3,$DF,$CC,$F8,$5F,$8F,$7F
	.dc.b	$37,$E1,$7E,$3D,$FC,$EF,$85,$F8,$F7,$F3,$FE,$17,$E3,$DF,$D0,$F8
	.dc.b	$5F,$8F,$7F,$C3,$13,$FE,$3F,$7F,$D7,$7E,$1F,$F2,$7F,$0F,$F2,$0F
	.dc.b	$F2,$9F,$0F,$F2,$0F,$F2,$BD,$FF,$20,$FF,$2D,$F0,$FF,$20,$FF,$2F
	.dc.b	$F0,$FF,$20,$FF,$31,$F0,$FF,$20,$FF,$33,$F0,$FF,$21,$97,$ED,$38
	.dc.b	$FF,$6D,$E1,$FE,$3E,$3F,$85,$C7,$C6,$69,$F1,$8F,$DF,$1F,$B4,$3F
	.dc.b	$81,$1F,$B4,$3F,$83,$1F,$B4,$3F,$85,$E3,$E8,$E4,$FF,$2E,$E7,$EC
	.dc.b	$7F,$CB,$FA,$7C,$07,$F8,$CD,$7F,$12,$F2,$3E,$37,$8D,$FE,$B6,$7E
	.dc.b	$56,$9F,$D3,$78,$7E,$73,$FD,$3F,$8B,$E9,$3F,$D4,$78,$BE,$B3,$FD
	.dc.b	$4F,$8F,$ED,$3F,$D5,$79,$3E,$F3,$FD,$5F,$97,$F0,$4F,$F1,$2C,$FF
	.dc.b	$68,$FF,$5B,$E0,$FE,$32,$7C,$5D,$3F,$A3,$4F,$C5,$3F,$7A,$7C,$53
	.dc.b	$F8,$09,$F1,$4F,$E0,$A7,$C5,$3F,$84,$9F,$C6,$0F,$E6,$3B,$FC,$57
	.dc.b	$F9,$C1,$FC,$60,$FE,$6B,$CB,$F1,$3C,$7F,$E8,$55,$FD,$B3,$4F,$8B
	.dc.b	$C1,$FE,$8A,$DF,$E2,$BF,$E8,$CD,$FE,$2B,$FE,$8E,$9F,$8A,$FF,$A4
	.dc.b	$37,$F8,$AF,$FA,$4B,$7F,$8A,$FF,$A5,$37,$F8,$AF,$F6,$1D,$3E,$2B
	.dc.b	$FA,$B7,$F5,$E7,$F5,$4E,$8F,$D9,$70,$7F,$80,$DF,$F7,$9F,$8D,$FB
	.dc.b	$89,$FE,$8B,$D5,$FB,$C5,$5F,$EF,$5E,$FF,$E3,$31,$FC,$08,$FE,$15
	.dc.b	$E2,$7F,$19,$7F,$86,$78,$9F,$C6,$5F,$E1,$DE,$27,$F1,$97,$F8,$87
	.dc.b	$89,$FC,$65,$FC,$61,$FE,$3C,$21,$FD,$5B,$A3,$4F,$8B,$5F,$E5,$3C
	.dc.b	$8C,$97,$D6,$B4,$7F,$20,$C3,$C4,$E2,$B8,$63,$E7,$8C,$5B,$2F,$E6
	.dc.b	$91,$FC,$94,$F8,$59,$FF,$8A,$74,$F8,$CF,$F8,$AF,$5F,$97,$A7,$F4
	.dc.b	$4A,$FF,$0C,$65,$F9,$27,$F7,$87,$D4,$3F,$C0,$BD,$FF,$61,$FF,$03
	.dc.b	$F0,$7D,$C7,$FC,$13,$C3,$F8,$07,$FC,$17,$C5,$F8,$47,$FC,$1B,$C7
	.dc.b	$F8,$87,$FC,$1F,$C9,$F8,$C7,$FC,$23,$CB,$F9,$07,$F8,$DE,$7F,$23
	.dc.b	$9B,$F2,$A4,$BD,$68,$F1,$D8,$73,$8C,$63,$DD,$D5,$78,$DE,$19,$79
	.dc.b	$DE,$59,$FA,$DE,$9A,$7B,$DE,$DA,$FE,$DF,$2F,$F0,$86,$BF,$96,$9F
	.dc.b	$D4,$3F,$E0,$1D,$FF,$C5,$B9,$7E,$B4,$91,$9D,$BD,$91,$88,$FE,$76
	.dc.b	$64,$F8,$24,$AE,$AE,$89,$5D,$9E,$11,$A8,$C4,$6A,$BB,$BA,$DB,$FD
	.dc.b	$C3,$DB,$FC,$DE,$F8,$56,$AB,$81,$E1,$AF,$A0,$F1,$C3,$D1,$87,$62
	.dc.b	$57,$16,$CF,$19,$DC,$E4,$91,$93,$A1,$0F,$0D,$F9,$1F,$F3,$5D,$FA
	.dc.b	$3A,$BF,$57,$C8,$F8,$8F,$F0,$B1,$2B,$FC,$0D,$F0,$FF,$E1,$9F,$89
	.dc.b	$FF,$0D,$7B,$5F,$E1,$BF,$67,$FC,$39,$F0,$FF,$E1,$DF,$6B,$FC,$3D
	.dc.b	$E8,$FF,$87,$FD,$AF,$F1,$06,$9F,$E2,$1F,$6B,$FC,$45,$EA,$7F,$0A
	.dc.b	$E4,$71,$FE,$F3,$F6,$EB,$C1,$FE,$63,$5E,$B6,$E7,$5F,$13,$FE,$1D
	.dc.b	$E6,$7B,$7F,$EC,$2C,$3F,$84,$E3,$FE,$9F,$F1,$F8,$47,$FA,$2B,$26
	.dc.b	$18,$C1,$1D,$A7,$C1,$8F,$99,$FF,$13,$69,$EA,$7F,$C5,$1B,$7B,$9F
	.dc.b	$F1,$57,$07,$C2,$7F,$C5,$9C,$5F,$19,$FF,$17,$72,$7F,$90,$79,$7E
	.dc.b	$64,$7F,$10,$E6,$FE,$57,$D5,$F8,$9E,$9F,$C2,$CF,$D8,$6F,$0F,$E0
	.dc.b	$B7,$FB,$44,$8C,$ED,$F8,$28,$C7,$57,$E1,$AF,$F0,$87,$37,$11,$A8
	.dc.b	$FC,$35,$F5,$BE,$71,$C2,$3F,$43,$87,$96,$FF,$A0,$CF,$F4,$2B,$FD
	.dc.b	$5B,$6D,$3A,$8F,$D1,$19,$70,$FE,$49,$B3,$CA,$36,$43,$47,$37,$4E
	.dc.b	$0F,$1B,$82,$4F,$8A,$38,$35,$DB,$5F,$21,$A9,$F1,$C6,$63,$C5,$3D
	.dc.b	$27,$24,$8E,$04,$3A,$B9,$69,$93,$B9,$82,$18,$21,$FC,$27,$85,$F1
	.dc.b	$3E,$2B,$F2,$4F,$C1,$36,$53,$55,$E1,$73,$7E,$97,$C4,$68,$A7,$12
	.dc.b	$3B,$7C,$A6,$CA,$6A,$7C,$71,$B8,$EF,$18,$24,$F8,$E3,$91,$A7,$E3
	.dc.b	$5F,$BA,$3A,$46,$DA,$79,$AB,$B3,$C8,$86,$8F,$2B,$A7,$07,$73,$95
	.dc.b	$79,$1B,$3D,$E7,$81,$27,$53,$D1,$AF,$DA,$AF,$44,$EE,$B9,$56,$AB
	.dc.b	$CB,$39,$7B,$90,$E9,$3C,$08,$E4,$F4,$1A,$30,$74,$9E,$CF,$B8,$FA
	.dc.b	$D5,$E4,$8F,$DD,$D7,$EF,$E7,$F8,$11,$FC,$03,$2F,$E1,$13,$A2,$5F
	.dc.b	$F0,$66,$CE,$CF,$E0,$93,$C2,$74,$48,$FF,$6E,$E5,$FC,$EA,$FF,$B7
	.dc.b	$E5,$FD,$97,$89,$ED,$7F,$31,$6D,$FF,$3F,$8C,$7B,$B9,$F7,$35,$EE
	.dc.b	$6F,$DF,$E1,$EE,$F1,$F8,$0E,$5C,$5C,$FE,$07,$A7,$C4,$F5,$F8,$3E
	.dc.b	$DF,$13,$DF,$E1,$7C,$1F,$13,$E1,$E7,$F1,$7C,$4F,$8F,$E2,$27,$1F
	.dc.b	$F5,$18,$D3,$FD,$4E,$B9,$3F,$D5,$70,$EC,$FF,$57,$CB,$C4,$FF,$59
	.dc.b	$D3,$CC,$FF,$5B,$DB,$D4,$FF,$5D,$E0,$F7,$3F,$D7,$F8,$BE,$27,$1C
	.dc.b	$8F,$F2,$5E,$BF,$49,$FB,$24,$E7,$FE,$71,$D1,$F4,$63,$1D,$3F,$55
	.dc.b	$FB,$1D,$3F,$81,$AD,$C6,$2F,$A4,$62,$3A,$F4,$FD,$7C,$7C,$51,$FE
	.dc.b	$69,$E9,$FA,$ED,$CF,$D2,$E7,$4F,$6B,$BD,$7E,$0B,$C7,$6F,$8A,$F3
	.dc.b	$DF,$E4,$BD,$78,$3E,$6B,$DF,$87,$E8,$BE,$1E,$2F,$AA,$F8,$F8,$FE
	.dc.b	$CB,$FE,$07,$E0,$FB,$AF,$F8,$27,$93,$F0,$2F,$F8,$2F,$97,$F0,$AF
	.dc.b	$F8,$37,$9B,$F1,$2F,$F8,$3F,$9F,$F1,$AF,$F8,$47,$A3,$F2,$2F,$F8
	.dc.b	$4F,$A7,$F2,$AF,$E5,$C8,$7F,$96,$78,$DF,$13,$F9,$D2,$47,$E9,$3A
	.dc.b	$7F,$D3,$3F,$B7,$BC,$4F,$F0,$93,$FE,$62,$EA,$FD,$42,$F3,$B3,$EF
	.dc.b	$EB,$D0,$DF,$C0,$70,$F4,$B8,$FC,$0F,$2F,$53,$9F,$C0,$F4,$F1,$3A
	.dc.b	$FA,$9D,$BD,$6E,$FF,$03,$F2,$38,$BF,$0E,$7F,$89,$7B,$72,$F9,$C7
	.dc.b	$8E,$E1,$5F,$49,$D4,$E7,$1F,$0C,$62,$99,$F9,$A3,$1A,$FA,$9D,$6F
	.dc.b	$EE,$78,$70,$FC,$4F,$81,$F8,$BC,$9F,$2A,$7F,$25,$CD,$FE,$28,$E0
	.dc.b	$F9,$7E,$CF,$CA,$24,$F8,$7E,$07,$E6,$FC,$0F,$CE,$F7,$3E,$7F,$C4
	.dc.b	$7D,$0F,$7B,$E8,$FC,$47,$F9,$0F,$F7,$0F,$F0,$EF,$6F,$F1,$97,$FC
	.dc.b	$99,$97,$FB,$25,$FF,$63,$3F,$01,$7C,$37,$1F,$83,$E5,$F0,$FC,$FE
	.dc.b	$17,$A7,$C4,$7C,$FC,$7F,$92,$78,$8F,$33,$F3,$71,$8E,$FD,$E9,$D7
	.dc.b	$0E,$6F,$7C,$BC,$D9,$F9,$FE,$1B,$D1,$F1,$1E,$9A,$F5,$9F,$47,$83
	.dc.b	$DE,$FD,$3E,$2F,$86,$FC,$53,$EB,$72,$FC,$87,$F3,$5C,$FF,$3F,$5F
	.dc.b	$3B,$F4,$A3,$EE,$74,$FE,$63,$AB,$EC,$47,$FB,$B7,$B9,$D9,$F7,$63
	.dc.b	$1D,$BF,$81,$75,$DD,$F8,$67,$87,$7F,$E2,$DE,$5E,$0F,$E4,$9F,$CA
	.dc.b	$F8,$3F,$C6,$7C,$7F,$D4,$F8,$7F,$9B,$9F,$C7,$78,$5F,$9F,$7F,$21
	.dc.b	$E2,$7E,$8D,$FC,$97,$8B,$FA,$77,$F2,$9E,$37,$EA,$DF,$CB,$78,$FE
	.dc.b	$B7,$DA,$7E,$67,$C7,$F8,$59,$7F,$A5,$79,$3F,$21,$3F,$A4,$E2,$FF
	.dc.b	$73,$BF,$28,$87,$C3,$F1,$FE,$7B,$F1,$7A,$3E,$9B,$F1,$FA,$7E,$BB
	.dc.b	$F8,$D3,$ED,$9F,$8D,$EB,$FB,$EF,$E0,$4F,$C8,$78,$7F,$8D,$7F,$25
	.dc.b	$E1,$FE,$37,$3F,$F5,$57,$85,$F8,$C7,$FD,$59,$E1,$7E,$32,$BF,$35
	.dc.b	$E1,$7E,$31,$FC,$E7,$85,$D9,$FC,$F6,$5E,$27,$C8,$7E,$EF,$C5,$FD
	.dc.b	$43,$FB,$DF,$1B,$F5,$8F,$EF,$FC,$7F,$E3,$7C,$1F,$8B,$F0,$BF,$8D
	.dc.b	$F8,$7F,$23,$B3,$FD,$5F,$CB,$F3,$30,$FF,$5A,$3F,$8D,$7F,$15,$E1
	.dc.b	$FD,$37,$F1,$9D,$1F,$5D,$FC,$77,$4F,$DB,$7F,$21,$D5,$F7,$DF,$C9
	.dc.b	$75,$FE,$0D,$FC,$A7,$67,$E1,$DF,$CB,$76,$FC,$77,$FC,$36,$7F,$21
	.dc.b	$3F,$8C,$3B,$D7,$F9,$9C,$F3,$7F,$9A,$D7,$77,$F9,$BD,$F8,$DF,$E7
	.dc.b	$38,$79,$DF,$E7,$78,$FA,$DF,$E7,$B9,$7B,$DF,$D6,$E5,$FE,$DF,$CF
	.dc.b	$E3,$76,$7E,$32,$3F,$59,$B7,$CB,$E7,$FE,$93,$C1,$F8,$CC,$3F,$A3
	.dc.b	$F6,$7E,$31,$FE,$93,$D9,$F8,$C7,$F8,$D6,$1F,$6D,$FE,$39,$CD,$F7
	.dc.b	$DF,$E3,$DC,$FF,$83,$7F,$90,$74,$7E,$1D,$FE,$29,$5F,$EF,$3E,$AF
	.dc.b	$C6,$3F,$1C,$FE,$49,$D7,$F9,$37,$F9,$37,$67,$F2,$37,$F9,$37,$8B
	.dc.b	$FC,$5D,$FE,$51,$E2,$FF,$17,$7F,$18,$7F,$17,$3F,$18,$7F,$17,$3F
	.dc.b	$18,$7F,$17,$3E,$19,$FA,$EF,$5F,$FC,$99,$C7,$F8,$BE,$BF,$C6,$CF
	.dc.b	$EE,$4F,$DD,$9F,$C5,$7C,$6F,$E2,$FA,$7F,$6E,$F1,$BF,$8B,$3F,$DC
	.dc.b	$3C,$6F,$E2,$CF,$F7,$2F,$1B,$F8,$B3,$FD,$D3,$C6,$FE,$2C,$FF,$76
	.dc.b	$F1,$BF,$8B,$3F,$E2,$93,$FC,$C1,$E7,$FF,$B3,$DF,$EF,$7E,$17,$E2
	.dc.b	$FA,$BF,$C0,$3E,$17,$E2,$DF,$F0,$1F,$85,$F8,$B7,$FC,$0B,$E1,$7E
	.dc.b	$2D,$FF,$03,$F8,$5F,$8B,$7F,$C1,$3E,$17,$E2,$DF,$F0,$5F,$85,$F8
	.dc.b	$B7,$F8,$B7,$17,$E2,$C0,$E5,$65,$FD,$A2,$3F,$8B,$E9,$FB,$67,$E7
	.dc.b	$3F,$E1,$2F,$0F,$F7,$B1,$FC,$F7,$C4,$FE,$02,$FF,$3F,$F1,$7E,$D3
	.dc.b	$FD,$97,$A7,$EF,$3F,$D9,$FA,$BF,$04,$FE,$10,$FE,$B8,$7F,$18,$3F
	.dc.b	$CD,$C7,$E3,$48,$7F,$8F,$CD,$1F,$C6,$0F,$E4,$27,$F2,$33,$B1,$FC
	.dc.b	$9C,$FC,$69,$FD,$D4,$FD,$16,$7F,$E3,$0F,$07,$F4,$CF,$F8,$CB,$C2
	.dc.b	$FD,$53,$FA,$C9,$F8,$6F,$C5,$7E,$46,$9F,$2B,$AD,$E4,$BE,$23,$FC
	.dc.b	$10,$4A,$FF,$02,$7C,$4F,$F8,$2F,$DB,$FF,$06,$7B,$9F,$E0,$DF,$77
	.dc.b	$FC,$1D,$F1,$1F,$E0,$FF,$7F,$FC,$21,$F1,$1F,$E1,$1F,$81,$FF,$09
	.dc.b	$7C,$47,$F8,$4F,$E0,$FF,$C2,$9F,$11,$FE,$15,$F8,$5F,$F0,$B7,$C4
	.dc.b	$7D,$CD,$DD,$9F,$EA,$FE,$77,$36,$63,$1F,$11,$63,$FD,$09,$5C,$1F
	.dc.b	$65,$27,$53,$FE,$40,$8E,$6F,$F8,$7E,$FC,$4F,$F8,$87,$1F,$33,$FE
	.dc.b	$23,$CF,$D4,$FF,$89,$75,$F7,$3F,$E2,$7D,$FE,$13,$FE,$29,$E1,$F8
	.dc.b	$CF,$C8,$3F,$C5,$9C,$9F,$2D,$FA,$27,$F9,$4B,$97,$E8,$3A,$21,$FE
	.dc.b	$35,$E5,$E6,$FF,$8D,$EB,$BB,$FE,$39,$BF,$1B,$FE,$3B,$C3,$CE,$FF
	.dc.b	$8F,$71,$F5,$BF,$E3,$FC,$BD,$EF,$F9,$07,$3F,$8B,$CC,$D7,$E2,$CF
	.dc.b	$CB,$3E,$46,$DF,$C9,$A5,$EC,$7F,$8F,$E7,$FC,$CF,$5F,$3D,$BC,$1F
	.dc.b	$71,$FF,$26,$71,$7E,$01,$FF,$27,$71,$FE,$11,$FF,$28,$72,$7E,$21
	.dc.b	$FF,$29,$72,$FE,$31,$FF,$2A,$73,$7E,$41,$FF,$2B,$73,$FF,$97,$79
	.dc.b	$3E,$67,$3F,$F9,$77,$9F,$FB,$A6,$AF,$87,$F6,$B7,$0F,$3C,$63,$E1
	.dc.b	$F3,$F0,$FA,$F7,$37,$F0,$FC,$3D,$EE,$3F,$0F,$CB,$E0,$39,$FC,$3F
	.dc.b	$4F,$82,$EB,$F0,$FD,$BE,$13,$BF,$C3,$F8,$3E,$1B,$C3,$F0,$FF,$DD
	.dc.b	$7B,$BF,$C8,$1C,$5F,$D3,$F9,$7F,$3D,$3B,$8F,$97,$7C,$DF,$9B,$8F
	.dc.b	$77,$E7,$E7,$E3,$7E,$8E,$BE,$77,$E9,$EF,$EB,$7E,$AF,$0F,$BD,$F8
	.dc.b	$24,$3F,$27,$9F,$F0,$2E,$5F,$97,$DB,$FC,$17,$C6,$FC,$AD,$7E,$4C
	.dc.b	$FA,$27,$E7,$CF,$AA,$7E,$8C,$FB,$27,$E9,$CF,$BA,$7E,$AC,$FC,$09
	.dc.b	$FA,$F3,$F0,$A7,$EC,$CF,$C4,$9F,$8C,$F0,$3F,$21,$E0,$7E,$3E,$3F
	.dc.b	$29,$87,$E5,$E7,$EE,$DF,$E3,$E7,$FD,$39,$E0,$7E,$3D,$FF,$4F,$78
	.dc.b	$1F,$8F,$7F,$D4,$1E,$07,$E3,$DF,$F5,$17,$81,$F8,$F7,$FD,$49,$E0
	.dc.b	$7E,$3D,$FF,$53,$78,$1F,$8F,$7F,$6B,$CD,$F8,$F7,$FA,$9F,$9D,$FD
	.dc.b	$49,$FD,$E7,$07,$CA,$69,$FC,$0C,$E6,$7C,$93,$B9,$F2,$4F,$19,$F2
	.dc.b	$4F,$39,$F2,$4F,$59,$F2,$4F,$79,$F2,$4F,$8B,$E0,$7E,$EA,$BF,$93
	.dc.b	$C7,$FB,$73,$D5,$FE,$F5,$B7,$F4,$16,$C7,$EF,$8F,$A2,$7F,$05,$3E
	.dc.b	$A9,$FC,$24,$FB,$27,$F0,$D3,$EE,$9F,$C4,$4F,$C0,$9F,$C5,$4F,$C2
	.dc.b	$9F,$C6,$4F,$C4,$9F,$8B,$3F,$DA,$57,$FC,$97,$0F,$95,$C1,$F9,$39
	.dc.b	$F9,$F7,$F9,$3A,$FA,$51,$F2,$8E,$87,$CA,$3C,$07,$CA,$3C,$87,$CA
	.dc.b	$3D,$07,$CA,$3D,$87,$CA,$3E,$01,$F2,$A3,$F8,$4F,$07,$E4,$BD,$FB
	.dc.b	$3F,$1B,$2F,$F7,$67,$89,$FC,$B1,$FF,$77,$78,$BF,$CC,$1F,$F7,$87
	.dc.b	$8D,$FC,$D1,$FF,$79,$78,$FF,$CE,$1F,$F7,$A7,$91,$FC,$F1,$FF,$7B
	.dc.b	$79,$3F,$D0,$1F,$F7,$C7,$93,$F8,$BD,$FE,$26,$7F,$B8,$E2,$FE,$38
	.dc.b	$FE,$03,$C9,$FE,$3E,$7E,$FD,$FC,$1F,$67,$F0,$37,$F0,$BD,$BF,$C1
	.dc.b	$DF,$C3,$F7,$7F,$0B,$7F,$13,$DF,$FC,$3D,$FC,$5F,$81,$FC,$4D,$FC
	.dc.b	$6F,$83,$FC,$5D,$FC,$7F,$85,$FC,$77,$AB,$F1,$7A,$9B,$E0,$BE,$1B
	.dc.b	$F9,$1D,$FF,$48,$F1,$3F,$93,$BF,$D2,$BC,$5F,$E5,$6F,$F4,$CF,$1B
	.dc.b	$F9,$7B,$FD,$3B,$C7,$FE,$66,$FF,$50,$F2,$3F,$9B,$BF,$D4,$BC,$9F
	.dc.b	$E7,$6F,$F5,$4F,$2B,$E1,$BF,$C7,$FC,$3F,$C6,$F9,$5F,$C7,$3A,$BF
	.dc.b	$75,$3F,$92,$3F,$6C,$FF,$81,$B8,$3F,$C4,$BE,$2F,$F0,$38,$C5,$3C
	.dc.b	$6F,$E1,$11,$8F,$1F,$F8,$63,$AF,$23,$F8,$83,$C3,$C9,$FE,$28,$F2
	.dc.b	$F2,$BF,$8C,$3F,$E1,$FF,$27,$F1,$BA,$7F,$1E,$E6,$FE,$41,$3E,$36
	.dc.b	$CF,$F8,$7B,$C9,$FE,$43,$5F,$C0,$6B,$A3,$FC,$15,$F0,$3F,$C2,$5F
	.dc.b	$23,$FC,$35,$F4,$3F,$C4,$5F,$6E,$7E,$FA,$FE,$85,$AF,$EC,$27,$E3
	.dc.b	$9F,$23,$7F,$DB,$F8,$3F,$7F,$0F,$F0,$1F,$4F,$F7,$02,$38,$21,$FC
	.dc.b	$59,$CB,$4C,$D7,$F9,$06,$32,$D8,$CE,$1F,$E3,$E9,$CB,$63,$FA,$41
	.dc.b	$FC,$7F,$D9,$FF,$0E,$CF,$E5,$2F,$FC,$2D,$3D,$C9,$7D,$5B,$7C,$DA
	.dc.b	$7F,$92,$6B,$BB,$4F,$19,$4D,$F1,$7F,$8A,$E3,$6D,$8F,$83,$8F,$C8
	.dc.b	$9C,$E1,$F2,$67,$8B,$70,$F2,$13,$0F,$23,$67,$C0,$DC,$3B,$47,$2B
	.dc.b	$EA,$76,$EA,$FE,$35,$1F,$E3,$87,$2F,$AA,$78,$D7,$3C,$3C,$58,$78
	.dc.b	$09,$5E,$06,$D7,$06,$CF,$09,$D1,$27,$CB,$8F,$E2,$3C,$AF,$F1,$A9
	.dc.b	$27,$FC,$97,$EF,$7F,$95,$3F,$71,$7C,$45,$F5,$B1,$F0,$19,$F6,$35
	.dc.b	$F0,$1B,$F6,$B8,$7C,$07,$1F,$73,$97,$C0,$73,$F7,$BA,$73,$F5,$F6
	.dc.b	$3B,$7C,$07,$EC,$39,$FF,$DB,$F3,$F8,$83,$1C,$FF,$EA,$7E,$15,$71
	.dc.b	$E3,$FF,$69,$DF,$F0,$79,$FE,$42,$31,$F0,$CE,$AF,$96,$7E,$68,$C0
	.dc.b	$E6,$5B,$6F,$6C,$EA,$73,$C1,$F0,$A3,$14,$E2,$F9,$1C,$7F,$E6,$0E
	.dc.b	$4F,$99,$5F,$E0,$4E,$5F,$F3,$2F,$4F,$E0,$23,$F0,$C7,$E0,$7A,$3E
	.dc.b	$C6,$7F,$6E,$3E,$C4,$60,$93,$C7,$83,$F0,$CF,$E3,$67,$FD,$BF,$1F
	.dc.b	$C5,$EF,$FC,$E3,$D3,$BE,$0F,$C2,$DD,$BE,$AD,$C7,$C0,$4C,$7B,$CF
	.dc.b	$96,$7D,$2F,$FB,$1C,$FD,$B1,$FE,$C9,$9F,$D0,$B6,$7F,$58,$D9,$E0
	.dc.b	$F0,$21,$FB,$93,$46,$A7,$9B,$A6,$BB,$31,$3D,$9B,$3D,$9F,$F1,$61
	.dc.b	$FE,$58,$7F,$B2,$9F,$C4,$FB,$3F,$90,$C7,$80,$F6,$57,$62,$57,$58
	.dc.b	$F0,$12,$33,$5D,$DB,$1E,$99,$E1,$18,$8E,$B1,$83,$F8,$87,$A2,$F3
	.dc.b	$BD,$33,$C6,$FA,$56,$DA,$48,$D1,$0E,$AE,$5A,$64,$F7,$10,$EE,$79
	.dc.b	$32,$FE,$2A,$FF,$96,$9F,$8A,$E8,$E0,$7C,$19,$F2,$3D,$4D,$10,$D4
	.dc.b	$F8,$DF,$9D,$7E,$F7,$27,$12,$9E,$82,$1E,$0C,$3E,$53,$6B,$C2,$E8
	.dc.b	$F0,$47,$D8,$C7,$F8,$54,$7F,$53,$3F,$AA,$9F,$D4,$4E,$57,$E5,$78
	.dc.b	$D7,$47,$27,$96,$BC,$2E,$A4,$CB,$F6,$4E,$A3,$D6,$72,$3C,$C6,$0E
	.dc.b	$47,$5A,$F0,$3E,$2C,$3E,$9E,$9F,$E2,$39,$FF,$1D,$4F,$F7,$A3,$A1
	.dc.b	$0E,$67,$6A,$F1,$12,$3A,$C7,$F6,$7B,$F3,$C7,$EF,$A3,$F8,$99,$2B
	.dc.b	$F8,$6B,$FC,$04,$86,$73,$FE,$16,$FF,$19,$C3,$E7,$BF,$AF,$E6,$FE
	.dc.b	$22,$F0,$AE,$6F,$13,$F8,$89,$9A,$E6,$E7,$1F,$E2,$59,$7C,$E8,$FF
	.dc.b	$73,$F0,$BE,$03,$FA,$B1,$23,$F9,$9F,$C3,$FF,$59,$F8,$7F,$EB,$5F
	.dc.b	$05,$FD,$6F,$DE,$FE,$B9,$F0,$5F,$D7,$78,$3F,$AF,$7C,$17,$5D,$FF
	.dc.b	$CA,$1F,$B9,$7F,$1D,$6F,$C1,$E7,$87,$EF,$F0,$A9,$F1,$19,$DF,$FA
	.dc.b	$54,$3E,$13,$15,$BF,$88,$B2,$F8,$6B,$F1,$1F,$1B,$0F,$F2,$87,$67
	.dc.b	$0F,$37,$FC,$9F,$1F,$D5,$78,$3F,$02,$FF,$BC,$B0,$5F,$4C,$3A,$E3
	.dc.b	$DB,$3E,$FA,$78,$36,$F0,$E3,$F0,$0F,$EB,$39,$7C,$47,$FA,$DF,$17
	.dc.b	$C8,$7F,$AE,$F2,$7F,$80,$F9,$7E,$74,$38,$BF,$B8,$73,$7F,$44,$8F
	.dc.b	$B1,$87,$E2,$E7,$EF,$93,$8F,$EE,$B6,$7F,$02,$70,$EC,$FC,$4C,$7F
	.dc.b	$80,$A3,$F6,$EE,$F9,$3F,$07,$1F,$87,$38,$C6,$6F,$C1,$7D,$8F,$C5
	.dc.b	$F9,$7F,$ED,$CE,$9F,$C4,$5F,$F7,$8C,$BD,$78,$7F,$82,$73,$FC,$44
	.dc.b	$FF,$7C,$E9,$FC,$43,$FD,$FB,$A7,$F1,$0F,$F8,$03,$A7,$E5,$BF,$E0
	.dc.b	$5E,$77,$17,$D0,$8C,$72,$7D,$47,$5C,$DF,$80,$7F,$C8,$47,$F7,$8D
	.dc.b	$FF,$1B,$D1,$F8,$0E,$27,$17,$E0,$37,$FC,$AF,$47,$DF,$C3,$C1,$D1
	.dc.b	$F7,$DF,$17,$47,$DF,$7C,$87,$DF,$3E,$C6,$5F,$E2,$BC,$3F,$01,$18
	.dc.b	$3F,$C4,$5C,$8E,$8F,$C1,$46,$73,$FF,$12,$F4,$7E,$0E,$3E,$51,$07
	.dc.b	$3B,$C7,$FD,$3F,$1B,$E2,$3F,$42,$4A,$F9,$1F,$11,$FA,$4F,$82,$FD
	.dc.b	$2F,$9F,$FA,$6F,$83,$FD,$3F,$A3,$FA,$8F,$84,$FD,$4F,$AD,$FA,$AF
	.dc.b	$85,$FD,$5F,$AF,$FA,$CF,$86,$FD,$6F,$B1,$FA,$EF,$87,$FD,$7F,$B3
	.dc.b	$FB,$0F,$88,$FD,$77,$77,$E3,$78,$7F,$8B,$1F,$48,$FD,$64,$70,$7E
	.dc.b	$E5,$74,$7E,$F6,$1E,$07,$F0,$19,$79,$1F,$C1,$69,$E8,$7F,$09,$B7
	.dc.b	$B1,$FC,$37,$07,$F0,$2E,$1F,$85,$1F,$BB,$E2,$FE,$3D,$CB,$F2,$63
	.dc.b	$FC,$A9,$FB,$9B,$B3,$F8,$5B,$63,$FC,$A8,$D7,$EB,$3F,$EA,$ED,$FF
	.dc.b	$DD,$4B,$CE,$E3,$EE,$F2,$F8,$9E,$7E,$FF,$4F,$89,$EB,$F0,$3D,$BE
	.dc.b	$27,$BF,$C1,$F8,$3D,$6F,$0F,$C2,$78,$BE,$27,$C7,$F0,$DE,$4F,$89
	.dc.b	$F2,$FC,$47,$F5,$8E,$3F,$E4,$FC,$28,$D7,$1F,$A2,$F9,$3E,$AC,$7B
	.dc.b	$3E,$CC,$FC,$4F,$BB,$5F,$33,$F0,$37,$F5,$3F,$0B,$87,$E2,$F1,$7C
	.dc.b	$6F,$5F,$E3,$EB,$2B,$FB,$BF,$2F,$CD,$F2,$3F,$9D,$73,$FD,$2E,$7F
	.dc.b	$E8,$2F,$D5,$AF,$AE,$7E,$F7,$C5,$FC,$A3,$FB,$FF,$17,$F2,$8F,$F0
	.dc.b	$1F,$17,$F2,$8F,$F0,$3F,$17,$F2,$8F,$F0,$5F,$17,$F2,$8F,$F0,$7F
	.dc.b	$17,$F2,$8F,$E4,$72,$FC,$A1,$35,$FE,$9F,$E5,$FF,$01,$7F,$57,$E3
	.dc.b	$7E,$56,$BF,$A1,$78,$DF,$95,$7F,$A2,$78,$DF,$95,$7F,$A3,$78,$DF
	.dc.b	$95,$7F,$A4,$78,$DF,$95,$7F,$A5,$78,$DF,$95,$7F,$A6,$78,$DF,$05
	.dc.b	$FC,$A1,$FC,$77,$8F,$F2,$BE,$47,$DC,$D3,$E6,$4F,$E4,$1F,$A0,$FE
	.dc.b	$53,$A3,$F1,$EF,$E5,$BC,$6F,$C7,$BF,$98,$F1,$BF,$1E,$FE,$6B,$C6
	.dc.b	$FC,$7B,$F9,$CF,$1B,$F1,$EF,$E7,$BC,$6F,$C7,$BF,$A0,$F1,$BF,$1E
	.dc.b	$FF,$29,$3F,$D7,$5E,$7F,$F2,$9F,$1F,$F2,$11,$B3,$F2,$07,$33,$F2
	.dc.b	$07,$73,$F2,$07,$8C,$FC,$81,$E7,$3F,$20,$7A,$CF,$C8,$1E,$F3,$F6
	.dc.b	$7E,$07,$C4,$E8,$FD,$8E,$DF,$22,$7F,$C0,$BC,$1F,$8F,$8F,$E4,$AF
	.dc.b	$CF,$9F,$F0,$8F,$0F,$D3,$7F,$C2,$7C,$5F,$5D,$FF,$0A,$F1,$FD,$B7
	.dc.b	$FC,$2F,$C9,$F7,$DF,$F0,$CF,$2F,$E0,$DF,$F0,$DF,$37,$E1,$DF,$F0
	.dc.b	$EF,$3F,$F2,$2C,$FF,$C3,$7E,$CF,$FA,$93,$DA,$C6,$DF,$E1,$AA,$E1
	.dc.b	$5F,$CE,$B0,$E8,$FF,$3C,$C7,$C0,$FF,$3E,$CB,$C8,$FF,$40,$CF,$D0
	.dc.b	$FF,$42,$D3,$D8,$FF,$44,$D7,$E0,$3F,$0B,$2F,$C7,$1F,$B6,$E9,$FC
	.dc.b	$73,$FC,$7A,$3E,$61,$FC,$33,$83,$F7,$F1,$F4,$8F,$E4,$3D,$1F,$8E
	.dc.b	$7F,$91,$F4,$7E,$39,$FE,$4B,$D1,$F8,$E7,$F9,$3F,$47,$E3,$9F,$E5
	.dc.b	$3D,$1F,$8E,$7F,$95,$F4,$7E,$39,$FF,$1A,$9F,$8F,$F0,$5D,$1F,$8E
	.dc.b	$8D,$9F,$8E,$39,$9F,$8E,$3B,$9F,$8E,$3C,$67,$E3,$8F,$39,$F8,$E3
	.dc.b	$D6,$7B,$79,$FF,$9D,$F5,$7E,$BA,$3F,$60,$7C,$53,$E4,$1F,$8E,$3E
	.dc.b	$59,$FB,$A3,$FC,$91,$E5,$7F,$6F,$7F,$8E,$1F,$D5,$8F,$E3,$87,$F5
	.dc.b	$A3,$F8,$E1,$FD,$78,$FE,$38,$7F,$62,$3F,$8E,$1F,$D9,$8F,$E3,$87
	.dc.b	$F8,$8F,$9F,$F1,$78,$3A,$7F,$1D,$3F,$C7,$F8,$7F,$8D,$BC,$0F,$E3
	.dc.b	$67,$43,$F8,$D9,$E0,$3F,$8D,$9E,$43,$F8,$D9,$E8,$3F,$8D,$9E,$C3
	.dc.b	$F8,$D9,$FE,$0E,$3E,$1E,$3F,$CF,$9F,$8F,$5F,$6A,$7F,$1B,$1F,$DC
	.dc.b	$1F,$9D,$3F,$42,$3E,$94,$7F,$5C,$E0,$FC,$74,$FF,$6E,$E0,$FC,$73
	.dc.b	$FD,$C3,$83,$F1,$CF,$F7,$2E,$0F,$C7,$3F,$DD,$38,$3F,$1C,$FF,$76
	.dc.b	$E0,$FC,$73,$FD,$F7,$37,$5F,$F9,$0C,$FE,$39,$E5,$7F,$6A,$D7,$FB
	.dc.b	$FF,$95,$FD,$A9,$FF,$00,$F9,$5F,$DA,$9F,$F0,$1F,$95,$FD,$A9,$FF
	.dc.b	$02,$F9,$5F,$DA,$9F,$F0,$3F,$95,$FD,$A9,$FF,$04,$F9,$5F,$AC,$7F
	.dc.b	$C1,$7D,$3F,$13,$4F,$8D,$1F,$23,$A5,$F1,$3F,$E1,$42,$47,$F8,$53
	.dc.b	$9F,$FC,$2F,$EE,$7F,$86,$3D,$CF,$F0,$CF,$BB,$FE,$A4,$5E,$F5,$F7
	.dc.b	$F1,$F1,$19,$F8,$1D,$7C,$46,$FE,$0F,$87,$C4,$71,$F8,$5E,$5F,$11
	.dc.b	$CF,$E1,$FE,$4F,$89,$FC,$73,$CF,$FF,$45,$BF,$E1,$2E,$DD,$71,$FF
	.dc.b	$BE,$3B,$FE,$77,$0B,$1E,$93,$9B,$FF,$5F,$CB,$3F,$14,$E4,$73,$A5
	.dc.b	$B7,$A6,$75,$39,$E0,$F7,$46,$2B,$E0,$BF,$E3,$AE,$3F,$8C,$FF,$8F
	.dc.b	$39,$3F,$C7,$FC,$BF,$32,$3F,$C7,$BC,$DF,$D2,$F1,$FE,$22,$E7,$6F
	.dc.b	$F1,$FB,$C2,$79,$1F,$E4,$6C,$7E,$8B,$E1,$E6,$F1,$BF,$E7,$D8,$67
	.dc.b	$E9,$8C,$69,$F0,$1D,$6B,$F1,$9E,$1B,$FC,$C7,$97,$0F,$F9,$05,$F8
	.dc.b	$B8,$23,$59,$FD,$0A,$E4,$FD,$2B,$EC,$FD,$4C,$3C,$4F,$D6,$C7,$CE
	.dc.b	$FD,$8C,$BD,$79,$7F,$98,$32,$FC,$7B,$F9,$13,$FC,$CB,$9F,$C9,$F6
	.dc.b	$3F,$C8,$13,$FE,$6D,$9F,$81,$E5,$7E,$4C,$FC,$A7,$C4,$7E,$57,$E2
	.dc.b	$3F,$2D,$F1,$1F,$97,$F7,$7F,$31,$F1,$1F,$99,$F7,$FF,$35,$F1,$1F
	.dc.b	$9B,$F6,$7F,$39,$F1,$1F,$F0,$03,$F7,$46,$B7,$E2,$71,$F0,$B9,$F8
	.dc.b	$9D,$7C,$3F,$E6,$2B,$F1,$7D,$BF,$C9,$9F,$F3,$DF,$83,$A7,$F1,$78
	.dc.b	$AF,$9C,$63,$1F,$03,$E2,$3F,$1D,$9F,$99,$FC,$86,$BE,$A7,$F2,$5B
	.dc.b	$FB,$9F,$CA,$70,$FC,$27,$F2,$DC,$7F,$15,$FE,$01,$CD,$F8,$FF,$07
	.dc.b	$FC,$FA,$FE,$3E,$3F,$D2,$3E,$07,$E6,$9F,$F4,$9F,$83,$FB,$FE,$AF
	.dc.b	$B3,$7F,$AC,$EC,$FB,$AF,$EB,$BB,$7F,$02,$FE,$C3,$BB,$F0,$AF,$EC
	.dc.b	$BB,$FF,$12,$FE,$D3,$C0,$FC,$77,$83,$FB,$17,$39,$FE,$42,$B8,$BF
	.dc.b	$92,$BE,$AF,$E5,$30,$F0,$BF,$96,$C7,$CA,$FE,$63,$2F,$C6,$F8,$3F
	.dc.b	$EA,$AF,$03,$F1,$B3,$FE,$AE,$E9,$F8,$4F,$FA,$DB,$83,$F6,$9E,$3F
	.dc.b	$F2,$48,$FC,$89,$0E,$07,$ED,$0F,$CC,$1F,$C0,$4F,$DA,$1F,$C1,$7C
	.dc.b	$7F,$E4,$6F,$F0,$7F,$1F,$F9,$1B,$FC,$27,$C7,$FE,$46,$FF,$0B,$F1
	.dc.b	$FF,$91,$BF,$C3,$7C,$7F,$C6,$71,$FF,$71,$E3,$FE,$47,$3F,$EC,$CE
	.dc.b	$FF,$8F,$3F,$C6,$79,$3E,$3B,$FC,$6F,$93,$E3,$BF,$C7,$79,$3E,$3B
	.dc.b	$FC,$6E,$FC,$C7,$A0,$FB,$E7,$F0,$F3,$DD,$C5,$FE,$D2,$E1,$F8,$F3
	.dc.b	$FC,$62,$7F,$16,$7C,$59,$FE,$32,$FE,$E4,$FC,$59,$CC,$EA,$7F,$17
	.dc.b	$E3,$FA,$F3,$E3,$8F,$E3,$5B,$FC,$67,$F8,$E6,$FF,$19,$FE,$3D,$BF
	.dc.b	$C6,$7F,$90,$6F,$FD,$4F,$7F,$C5,$F1,$FF,$17,$9F,$94,$7F,$37,$DF
	.dc.b	$E2,$C7,$F4,$38,$FA,$47,$F3,$BD,$BE,$2C,$7F,$2A,$8F,$8A,$79,$CF
	.dc.b	$BC,$7E,$00,$F6,$9F,$84,$3F,$9E,$9F,$87,$F1,$FF,$8B,$F4,$FE,$C6
	.dc.b	$3F,$8B,$9F,$D2,$7C,$7F,$E2,$CF,$F4,$BF,$1F,$F8,$B3,$FD,$37,$C7
	.dc.b	$FE,$2C,$FF,$4F,$F1,$FF,$8B,$3F,$D4,$7C,$7F,$E0,$EF,$F0,$93,$F8
	.dc.b	$6F,$8F,$FC,$55,$FE,$1F,$E3,$FE,$2B,$9F,$F8,$A7,$3F,$F8,$57,$AF
	.dc.b	$FC,$C0,$FF,$85,$FC,$75,$71,$B4,$E7,$0F,$E4,$51,$FC,$68,$FE,$2C
	.dc.b	$7F,$51,$F1,$7F,$8B,$3F,$D4,$FC,$5F,$E2,$CF,$F5,$5F,$17,$F8,$B3
	.dc.b	$FD,$5F,$C5,$FE,$2C,$FC,$42,$71,$FF,$BF,$74,$FD,$94,$7C,$C2,$65
	.dc.b	$F9,$37,$F2,$BE,$1F,$F3,$6E,$DF,$F3,$DF,$37,$F0,$08,$FF,$00,$F7
	.dc.b	$7E,$3A,$3E,$E5,$7E,$38,$FC,$01,$F8,$23,$FB,$29,$FB,$33,$F1,$07
	.dc.b	$E2,$8F,$EC,$A7,$FA,$27,$CF,$FF,$04,$3F,$E0,$3E,$BF,$D9,$CF,$F8
	.dc.b	$4E,$39,$1D,$0F,$CD,$9E,$03,$C4,$7F,$79,$F1,$3C,$D7,$E9,$8F,$F0
	.dc.b	$AE,$7F,$EE,$3F,$1B,$F9,$B6,$9F,$AE,$9F,$89,$AF,$F6,$4C,$BF,$45
	.dc.b	$1E,$32,$1E,$82,$4F,$F8,$A2,$37,$18,$D3,$F9,$A3,$C2,$74,$D8,$CB
	.dc.b	$B3,$FC,$0B,$3C,$A7,$71,$F8,$89,$FF,$1D,$3F,$E2,$AE,$5F,$DA,$3C
	.dc.b	$76,$FE,$80,$F5,$39,$95,$CB,$BD,$78,$DF,$62,$5F,$91,$B3,$C8,$E1
	.dc.b	$87,$96,$7D,$38,$FC,$79,$ED,$7D,$27,$91,$23,$95,$F4,$C3,$C6,$DA
	.dc.b	$E1,$3A,$AE,$15,$A6,$CE,$A3,$13,$B2,$3F,$E8,$68,$FF,$07,$C7,$FB
	.dc.b	$58,$F4,$9E,$6C,$3D,$4F,$08,$E8,$48,$CC,$F6,$24,$66,$79,$1F,$84
	.dc.b	$C1,$D7,$FA,$9E,$67,$C4,$FF,$88,$49,$5F,$E1,$8F,$89,$FF,$12,$7C
	.dc.b	$0F,$F8,$97,$DA,$FF,$13,$7B,$9F,$E2,$7F,$87,$FF,$14,$7B,$DF,$E2
	.dc.b	$9F,$87,$FF,$15,$7C,$07,$F8,$AF,$E1,$FF,$C5,$9F,$05,$FE,$2D,$F8
	.dc.b	$7F,$F1,$77,$C2,$7F,$8B,$FE,$1F,$FC,$61,$F0,$DF,$C5,$F9,$FF,$AF
	.dc.b	$F3,$FF,$BD,$BD,$2C,$F1,$FF,$8A,$AF,$8B,$FE,$2C,$C7,$AB,$FE,$2E
	.dc.b	$CF,$C2,$FF,$8C,$35,$F2,$BF,$E3,$2D,$FD,$2F,$F8,$D3,$87,$DA,$FB
	.dc.b	$CF,$F1,$BF,$27,$C2,$5C,$FF,$89,$7E,$A9,$3A,$BF,$09,$B7,$D2,$E9
	.dc.b	$FA,$13,$9E,$AF,$C1,$CE,$49,$1F,$C2,$E6,$BB,$3E,$F4,$62,$BA,$C7
	.dc.b	$DA,$35,$C1,$F7,$A7,$F7,$78,$79,$67,$F1,$73,$FA,$A6,$DF,$EC,$9B
	.dc.b	$5F,$90,$69,$4F,$D5,$18,$35,$3C,$DF,$E0,$66,$48,$60,$87,$22,$BC
	.dc.b	$5C,$11,$FF,$0F,$3F,$C1,$E3,$F8,$AB,$D0,$D4,$FA,$5E,$6D,$73,$1E
	.dc.b	$96,$9E,$B9,$F0,$12,$78,$36,$78,$36,$7F,$D0,$3F,$0C,$F8,$2E,$F5
	.dc.b	$AB,$E2,$47,$04,$AF,$04,$65,$B1,$97,$E5,$C6,$5B,$19,$7E,$05,$F8
	.dc.b	$DF,$A3,$E1,$7F,$50,$7F,$83,$65,$FC,$C1,$DD,$F3,$6D,$F3,$7D,$27
	.dc.b	$88,$87,$80,$A7,$02,$9F,$D0,$31,$F8,$4F,$CC,$38,$10,$E0,$53,$B9
	.dc.b	$4E,$06,$88,$68,$FC,$BF,$77,$F8,$5B,$4C,$44,$38,$97,$2F,$88,$F2
	.dc.b	$32,$47,$FA,$4D,$7A,$1D,$C7,$1C,$3E,$33,$70,$F8,$CF,$18,$CB,$63
	.dc.b	$33,$FC,$DE,$3F,$D8,$B8,$B0,$FF,$09,$65,$F5,$63,$EA,$4F,$53,$E5
	.dc.b	$9C,$CB,$5E,$47,$07,$C7,$7F,$01,$8F,$A7,$1F,$E1,$75,$FC,$0A,$BF
	.dc.b	$80,$C6,$30,$FE,$1B,$1F,$C1,$67,$44,$8C,$C7,$E2,$1F,$3C,$FF,$0D
	.dc.b	$73,$3B,$9F,$E2,$27,$38,$C9,$1C,$9F,$E0,$DD,$3F,$A5,$AC,$1D,$AF
	.dc.b	$87,$FF,$76,$12,$7F,$DC,$7F,$0F,$FD,$83,$1F,$95,$1C,$7F,$75,$AF
	.dc.b	$F8,$2C,$C7,$BB,$BC,$FF,$C8,$12,$D7,$15,$C7,$DE,$BA,$29,$F1,$18
	.dc.b	$5F,$01,$7E,$23,$1F,$05,$9F,$88,$D7,$C2,$6F,$A3,$87,$C2,$F1,$F8
	.dc.b	$8E,$5F,$0F,$FE,$DA,$CC,$E7,$7F,$C2,$F0,$F9,$AF,$94,$60,$F7,$63
	.dc.b	$DD,$C6,$FE,$7D,$3C,$91,$8C,$7C,$ED,$DF,$D5,$C1,$EB,$E1,$F6,$71
	.dc.b	$7B,$74,$F8,$3C,$CE,$5F,$A1,$8F,$CB,$D3,$EA,$F3,$7F,$84,$39,$FF
	.dc.b	$AE,$73,$FE,$06,$7F,$A6,$57,$E0,$4F,$BF,$19,$E8,$FC,$84,$63,$2F
	.dc.b	$C2,$57,$F8,$47,$AF,$F7,$53,$92,$47,$47,$F3,$3D,$7F,$91,$73,$7D
	.dc.b	$27,$58,$FE,$31,$F3,$CF,$C6,$D3,$F1,$8D,$CF,$DF,$A7,$C1,$8F,$91
	.dc.b	$AF,$C5,$7E,$56,$FF,$81,$E0,$FC,$17,$57,$CE,$9F,$89,$C4,$E3,$FA
	.dc.b	$91,$8E,$4F,$B2,$EB,$97,$EF,$3C,$39,$BF,$04,$63,$97,$F3,$7C,$1F
	.dc.b	$91,$AE,$31,$8C,$58,$76,$8C,$63,$E3,$75,$97,$9F,$2F,$49,$F8,$3E
	.dc.b	$6F,$BE,$FE,$17,$9B,$E0,$BF,$0C,$87,$F8,$63,$CC,$FD,$3F,$C4,$7E
	.dc.b	$84,$91,$F8,$DF,$17,$F4,$9F,$0B,$FA,$5E,$AF,$D3,$79,$FF,$A7,$EA
	.dc.b	$FD,$47,$C3,$7E,$B3,$F7,$67,$A7,$7E,$1F,$1E,$AE,$7E,$1B,$5F,$05
	.dc.b	$BF,$86,$E1,$F0,$FC,$7E,$1F,$F5,$3C,$3F,$E2,$13,$F3,$DD,$B9,$7F
	.dc.b	$47,$A7,$FB,$DD,$FC,$C2,$BE,$B1,$AC,$7C,$2F,$0C,$FC,$DA,$7A,$23
	.dc.b	$F5,$5B,$7B,$1F,$D6,$70,$7C,$07,$E1,$1F,$AF,$E3,$F8,$AF,$C9,$3F
	.dc.b	$8E,$65,$FE,$BD,$7C,$7F,$0D,$F0,$CF,$89,$F0,$FF,$17,$C7,$F8,$DC
	.dc.b	$FF,$1F,$E1,$FE,$47,$BD,$F2,$7E,$1F,$E5,$7C,$07,$CB,$F8,$7F,$99
	.dc.b	$F0,$5F,$37,$E1,$FE,$77,$C2,$7C,$FF,$87,$FE,$37,$E7,$7F,$8D,$FC
	.dc.b	$5F,$C6,$AF,$88,$F9,$7C,$5F,$CF,$4F,$E6,$5E,$B7,$F5,$3F,$17,$F9
	.dc.b	$3F,$6F,$2A,$E7,$5D,$4F,$9F,$97,$85,$FA,$3A,$79,$5F,$A7,$B7,$A5
	.dc.b	$FA,$BC,$1E,$D7,$EB,$F1,$7C,$17,$EC,$F2,$7C,$57,$ED,$F2,$FF,$0F
	.dc.b	$9F,$C9,$C7,$CF,$21,$FB,$2F,$17,$EA,$3F,$B4,$E9,$FB,$0F,$ED,$BA
	.dc.b	$BE,$E3,$F8,$4E,$BF,$C9,$3F,$88,$F0,$BF,$24,$FE,$2B,$C2,$FC,$93
	.dc.b	$F8,$CF,$0B,$F1,$CF,$FB,$4F,$0F,$E2,$FE,$67,$FA,$A7,$B9,$E1,$FE
	.dc.b	$62,$BF,$27,$E1,$FE,$41,$FC,$9F,$87,$F9,$07,$F2,$BE,$1F,$E4,$1F
	.dc.b	$E7,$1E,$27,$EA,$5F,$E7,$5E,$2F,$EB,$5F,$E7,$9E,$37,$F5,$3E,$2F
	.dc.b	$C7,$F7,$FF,$05,$F1,$BE,$35,$7E,$43,$9B,$E5,$3F,$90,$3E,$69,$FA
	.dc.b	$6F,$1B,$F1,$EF,$EA,$3C,$6F,$C7,$BF,$AA,$F1,$BF,$1E,$FE,$B3,$C6
	.dc.b	$FC,$7B,$FA,$EF,$1B,$F1,$EF,$EC,$3C,$6F,$C2,$BF,$D5,$3C,$6F,$C7
	.dc.b	$3F,$B6,$3F,$70,$7F,$21,$35,$E1,$F0,$7F,$6F,$3D,$1F,$DC,$D7,$81
	.dc.b	$FD,$DD,$F9,$1F,$DE,$E1,$E8,$7F,$7F,$8F,$B1,$FE,$03,$97,$EB,$9F
	.dc.b	$C5,$90,$FD,$B7,$3F,$F7,$FD,$7E,$5F,$8B,$F8,$E9,$FC,$61,$F4,$0F
	.dc.b	$E0,$27,$D4,$3F,$82,$9F,$60,$FE,$23,$DB,$F8,$C7,$F8,$8F,$6F,$E3
	.dc.b	$1F,$E2,$7D,$BF,$8C,$7F,$8A,$F6,$FE,$33,$8F,$F9,$F7,$8B,$FD,$12
	.dc.b	$7F,$8D,$F6,$7E,$2D,$FE,$37,$D9,$F8,$BA,$FE,$8F,$DA,$ED,$FC,$64
	.dc.b	$63,$B7,$F1,$97,$FD,$5B,$B7,$F1,$8F,$F5,$8E,$DF,$C6,$3F,$D6,$BB
	.dc.b	$7F,$18,$FE,$B2,$FF,$18,$8F,$8F,$8F,$F1,$B8,$F0,$F6,$7E,$39,$F1
	.dc.b	$F8,$3F,$BB,$AF,$E6,$3E,$0F,$EF,$DF,$E6,$7E,$17,$F0,$37,$F9,$AF
	.dc.b	$87,$FC,$1D,$FE,$6F,$E2,$7F,$0B,$7F,$9C,$F8,$BF,$C3,$DF,$F0,$97
	.dc.b	$85,$FC,$6F,$A7,$C9,$D7,$FE,$47,$F0,$BF,$8D,$EB,$FD,$AB,$C6,$FE
	.dc.b	$35,$87,$F1,$CF,$1B,$38,$70,$C7,$F2,$B3,$D0,$ED,$7F,$33,$2F,$13
	.dc.b	$F3,$B3,$F3,$3F,$43,$4F,$53,$F4,$B5,$F7,$3F,$53,$6F,$84,$FC,$4C
	.dc.b	$3F,$20,$FC,$87,$E6,$7A,$5F,$E1,$7F,$31,$E0,$7E,$4A,$7F,$CB,$8F
	.dc.b	$D4,$7E,$B1,$F9,$23,$ED,$1F,$92,$3E,$F1,$F9,$23,$FA,$E1,$FC,$74
	.dc.b	$FE,$C0,$7F,$1D,$3F,$8A,$9F,$C7,$4F,$C7,$9F,$E4,$CD,$BF,$8E,$F9
	.dc.b	$1F,$E2,$DE,$5F,$E3,$B3,$B3,$F2,$67,$33,$F8,$E9,$DC,$FE,$3A,$78
	.dc.b	$CF,$29,$FC,$CC,$FD,$51,$FC,$DC,$FD,$71,$FB,$03,$FC,$5B,$5F,$DB
	.dc.b	$30,$FF,$1E,$4F,$E4,$78,$3F,$02,$DA,$FE,$40,$7F,$1E,$3F,$80,$9F
	.dc.b	$D3,$8F,$E3,$C7,$F5,$23,$F8,$F1,$FD,$58,$FE,$3C,$7F,$5A,$3F,$8F
	.dc.b	$1F,$D7,$8F,$E3,$C7,$F1,$D3,$FB,$21,$FC,$74,$FE,$D5,$C7,$FE,$BF
	.dc.b	$7F,$8F,$1F,$98,$E3,$FE,$3D,$A7,$EE,$3C,$7F,$E3,$CF,$EE,$BC,$7F
	.dc.b	$E3,$CF,$EF,$3C,$7F,$E3,$CF,$EF,$BC,$7F,$E3,$CF,$F0,$0F,$1F,$F8
	.dc.b	$F3,$FC,$0B,$C7,$FE,$3D,$97,$F3,$F2,$6D,$FE,$63,$E8,$7C,$3F,$F8
	.dc.b	$10,$93,$FC,$DF,$E2,$7F,$C0,$FF,$13,$FE,$08,$F6,$BF,$C1,$3F,$11
	.dc.b	$FE,$0A,$F8,$8F,$F0,$5F,$BB,$FE,$0C,$F8,$8F,$F0,$6F,$BF,$FE,$0E
	.dc.b	$F8,$8F,$F0,$7F,$C0,$FF,$84,$3E,$23,$FC,$23,$F0,$7F,$E1,$2F,$88
	.dc.b	$FF,$09,$FC,$2F,$F8,$53,$E2,$3F,$D5,$FE,$4F,$ED,$BB,$BE,$47,$47
	.dc.b	$F8,$5A,$77,$C9,$FD,$2F,$0E,$6F,$F4,$DC,$BB,$BF,$D3,$F4,$F1,$BF
	.dc.b	$D4,$76,$F3,$BF,$D4,$F8,$3D,$6F,$F5,$5E,$2F,$7B,$F0,$4F,$EB,$1C
	.dc.b	$BF,$11,$FE,$8B,$CD,$F2,$1C,$90,$FE,$27,$CD,$C9,$FE,$2B,$5D,$9F
	.dc.b	$E2,$F7,$E2,$7F,$8C,$E1,$E6,$7F,$8D,$E3,$EA,$7F,$8E,$E5,$EE,$7F
	.dc.b	$8F,$E7,$F0,$9F,$E4,$3A,$7C,$8F,$0E,$47,$F6,$F8,$D6,$DF,$E3,$EA
	.dc.b	$E4,$FF,$90,$2F,$B3,$FE,$42,$C3,$C4,$FF,$91,$31,$F3,$3F,$E4,$6C
	.dc.b	$BD,$4F,$F9,$23,$3F,$73,$FE,$4A,$D3,$E3,$3F,$E4,$CD,$BF,$C2,$3B
	.dc.b	$FF,$BE,$7E,$1F,$FC,$A2,$4D,$BF,$C0,$9F,$11,$FE,$54,$E1,$FF,$2A
	.dc.b	$FC,$1F,$FB,$C3,$F7,$7F,$F7,$CE,$5F,$EA,$25,$F0,$99,$F7,$B5,$F0
	.dc.b	$9B,$F8,$4E,$1F,$03,$C7,$E2,$79,$7C,$1F,$3F,$89,$E9,$F0,$BD,$7E
	.dc.b	$27,$FA,$0F,$0B,$8D,$E8,$FF,$8E,$BD,$2F,$F7,$9C,$FF,$BB,$38,$17
	.dc.b	$CA,$31,$8F,$67,$59,$F8,$9E,$1A,$F9,$9E,$5B,$FB,$78,$3D,$DE,$8F
	.dc.b	$C0,$9F,$F1,$37,$27,$F9,$DB,$97,$E3,$47,$F5,$EE,$6F,$97,$3F,$91
	.dc.b	$8F,$EC,$3C,$FF,$42,$FF,$AD,$74,$7D,$47,$FA,$E7,$4F,$D8,$7F,$AF
	.dc.b	$75,$7D,$C7,$FB,$07,$5F,$E0,$1F,$EC,$5D,$9F,$84,$7F,$B2,$76,$FE
	.dc.b	$21,$FE,$CD,$DD,$F8,$C7,$FB,$47,$7F,$E4,$34,$FE,$3F,$AA,$35,$1F
	.dc.b	$36,$15,$CE,$31,$7D,$DD,$61,$E3,$78,$63,$E7,$79,$65,$EB,$7A,$67
	.dc.b	$EF,$7B,$69,$F0,$DF,$06,$BF,$20,$C7,$4B,$4F,$95,$DD,$F9,$9A,$FF
	.dc.b	$2B,$79,$1F,$C7,$DF,$F2,$C7,$91,$FC,$7D,$FF,$2D,$79,$1F,$C7,$DF
	.dc.b	$F2,$E7,$91,$FC,$7D,$FF,$2F,$79,$1F,$C7,$DF,$F3,$07,$91,$FC,$7D
	.dc.b	$FF,$31,$79,$1F,$C7,$DF,$F3,$0E,$5F,$C7,$5C,$9F,$C7,$4E,$27,$F1
	.dc.b	$D3,$A9,$FC,$74,$F0,$9F,$C7,$4F,$29,$FC,$74,$F4,$9F,$C7,$4F,$69
	.dc.b	$FC,$74,$FC,$49,$FC,$67,$C9,$FE,$89,$B7,$C6,$D7,$E2,$F3,$FE,$32
	.dc.b	$B8,$9F,$C7,$0E,$A7,$F1,$C3,$C2,$7F,$1C,$3C,$A7,$F1,$C3,$D2,$7F
	.dc.b	$1C,$3D,$A7,$F1,$C3,$E0,$9F,$C7,$0F,$8A,$FE,$39,$CD,$FF,$1B,$78
	.dc.b	$BF,$C6,$DE,$AF,$F1,$B7,$C2,$FF,$1B,$7C,$AF,$F1,$B7,$D2,$FF,$1B
	.dc.b	$7D,$AF,$E0,$DF,$8A,$4B,$FE,$D1,$D1,$FB,$49,$FF,$74,$BF,$B7,$9F
	.dc.b	$DC,$9F,$E9,$FF,$1F,$F7,$AF,$F0,$9E,$BF,$E3,$0F,$F0,$BF,$1F,$F8
	.dc.b	$C3,$FC,$37,$C7,$FE,$30,$FF,$0F,$F1,$FF,$8C,$3F,$C4,$7C,$7F,$E3
	.dc.b	$0F,$F8,$37,$26,$9F,$E6,$0F,$1F,$F8,$C7,$C0,$EA,$95,$FC,$A6,$F9
	.dc.b	$C6,$2F,$FC,$CF,$E3,$B1,$F0,$CE,$4F,$E4,$90,$CB,$CF,$19,$E9,$4B
	.dc.b	$4F,$6C,$E6,$3F,$9B,$BF,$CA,$70,$FE,$3C,$FF,$9E,$38,$7E,$43,$FC
	.dc.b	$74,$FF,$63,$70,$7E,$DD,$FE,$D9,$A7,$F4,$B7,$F3,$0F,$C0,$7F,$86
	.dc.b	$74,$FE,$DD,$FF,$41,$F4,$FE,$DD,$FF,$42,$F4,$FE,$DD,$FF,$43,$F4
	.dc.b	$FE,$DD,$FF,$44,$F4,$FE,$DD,$FF,$45,$F4,$FE,$DD,$FF,$46,$F4,$FE
	.dc.b	$DD,$FF,$1D,$F9,$5F,$E1,$5D,$7F,$8F,$47,$E5,$36,$FE,$4A,$FE,$51
	.dc.b	$FE,$52,$FE,$51,$FE,$5A,$FE,$51,$FE,$62,$FE,$51,$FE,$6A,$FE,$51
	.dc.b	$FE,$72,$FE,$51,$FD,$93,$F9,$4D,$7E,$23,$FE,$D8,$DD,$F0,$FF,$DA
	.dc.b	$49,$1F,$D6,$5A,$FD,$56,$CF,$F6,$D6,$D7,$E1,$1B,$3F,$82,$6F,$17
	.dc.b	$E0,$1B,$7E,$09,$EC,$7F,$84,$4F,$DC,$9F,$D0,$7C,$3D,$CC,$8D,$9E
	.dc.b	$32,$E1,$E3,$6E,$1E,$36,$9E,$32,$E1,$E3,$6E,$1E,$37,$E2,$5F,$77
	.dc.b	$B1,$2F,$B3,$6F,$B3,$4E,$C5,$BE,$CD,$BE,$D5,$C8,$95,$C9,$FF,$0C
	.dc.b	$7C,$4F,$F6,$57,$FD,$A5,$E6,$7F,$87,$63,$C9,$1C,$2F,$BB,$70,$EE
	.dc.b	$DC,$3B,$E1,$D4,$95,$C1,$B5,$C1,$B3,$C2,$7D,$7D,$AF,$5B,$F8,$0F
	.dc.b	$5F,$F8,$BC,$93,$FE,$1F,$F5,$BF,$C6,$5F,$11,$FE,$33,$F7,$BF,$C6
	.dc.b	$9D,$FF,$E3,$5F,$7F,$FC,$6D,$C1,$FE,$37,$F8,$06,$33,$83,$F6,$DC
	.dc.b	$0F,$DE,$1F,$13,$AA,$FE,$15,$0F,$83,$C5,$F0,$87,$C4,$E3,$93,$FD
	.dc.b	$FB,$0F,$85,$C6,$1F,$E3,$98,$7C,$4E,$33,$F0,$7C,$3D,$97,$C4,$FE
	.dc.b	$EF,$3F,$2F,$0B,$DB,$E0,$DC,$BF,$34,$79,$AB,$96,$1C,$F2,$E9,$9F
	.dc.b	$5D,$7B,$6D,$DF,$7F,$89,$8F,$F9,$27,$87,$E4,$3F,$E4,$BC,$FE,$63
	.dc.b	$FE,$4D,$E4,$FF,$30,$72,$FD,$28,$FE,$55,$CD,$FE,$6A,$E9,$FB,$5C
	.dc.b	$7F,$76,$7F,$00,$47,$04,$8D,$93,$AF,$F1,$FC,$BF,$84,$6D,$71,$9D
	.dc.b	$F5,$FE,$5E,$BE,$DD,$FF,$AC,$9F,$E8,$F1,$C2,$3B,$5D,$AE,$D7,$FB
	.dc.b	$28,$C1,$1F,$C2,$F5,$FE,$45,$FD,$69,$F8,$3A,$FD,$6B,$6B,$F7,$4D
	.dc.b	$AF,$C4,$34,$A7,$23,$F7,$26,$4D,$CF,$47,$24,$38,$11,$D3,$67,$C2
	.dc.b	$EE,$3F,$AF,$DF,$C5,$8F,$F4,$47,$77,$FB,$5E,$93,$FC,$BA,$7C,$44
	.dc.b	$39,$47,$47,$B5,$FA,$9A,$7B,$2F,$C6,$49,$E0,$D9,$E0,$D9,$FD,$8F
	.dc.b	$AD,$FE,$96,$79,$9F,$0A,$B8,$92,$B8,$B6,$3C,$0D,$78,$46,$88,$E9
	.dc.b	$AE,$8C,$10,$F0,$5F,$C7,$BF,$C5,$47,$F8,$32,$3E,$33,$F1,$27,$9B
	.dc.b	$F1,$CF,$49,$67,$DC,$F7,$21,$9A,$F4,$3F,$81,$BF,$85,$3C,$23,$55
	.dc.b	$D9,$F0,$1E,$73,$25,$32,$76,$29,$E3,$EF,$FF,$00,$BF,$E2,$3D,$FF
	.dc.b	$84,$37,$A3,$F8,$76,$BC,$22,$1E,$C2,$E1,$E5,$6E,$5E,$A7,$C3,$97
	.dc.b	$AA,$31,$1C,$49,$3C,$5B,$1A,$8C,$11,$FC,$0E,$7F,$4F,$4F,$CC,$3F
	.dc.b	$4A,$B8,$E9,$F1,$5A,$7C,$32,$9F,$0C,$B8,$79,$63,$F7,$D1,$FD,$95
	.dc.b	$BD,$3F,$D6,$5B,$97,$F1,$19,$FE,$EF,$7E,$3E,$2F,$EC,$91,$FC,$24
	.dc.b	$87,$32,$DF,$36,$D7,$37,$43,$7F,$E1,$DA,$7F,$51,$F8,$9F,$E8,$44
	.dc.b	$9F,$EA,$D5,$FD,$33,$E2,$7F,$A3,$7C,$4F,$F4,$7E,$0F,$E9,$1E,$0F
	.dc.b	$F4,$9F,$7F,$FA,$57,$C3,$7F,$4B,$F8,$1F,$E9,$9F,$0F,$FD,$37,$E0
	.dc.b	$FF,$A7,$7C,$47,$F4,$FF,$85,$FE,$A1,$F1,$3F,$D4,$7E,$1F,$FA,$A7
	.dc.b	$EF,$3F,$C6,$B9,$7F,$C0,$BC,$5F,$E2,$6E,$4F,$E1,$AF,$90,$29,$E1
	.dc.b	$C7,$FD,$8A,$FA,$D6,$EB,$FB,$76,$3E,$28,$C7,$90,$D3,$CF,$18,$DB
	.dc.b	$D6,$EB,$83,$DE,$F0,$E2,$F8,$6F,$2E,$4F,$90,$FF,$78,$E5,$FE,$A9
	.dc.b	$D3,$F7,$23,$FA,$53,$74,$FE,$8B,$1F,$58,$C9,$5F,$F7,$97,$99,$FD
	.dc.b	$C3,$09,$3F,$DC,$38,$7E,$F3,$F7,$1A,$4E,$0F,$C0,$36,$3A,$9F,$82
	.dc.b	$E8,$FC,$0E,$EC,$FD,$35,$FE,$0D,$7F,$C0,$9A,$7B,$AF,$FC,$27,$8F
	.dc.b	$C2,$7F,$C2,$BA,$FC,$67,$FC,$2F,$B7,$CA,$7F,$C3,$3B,$FC,$F7,$FC
	.dc.b	$37,$C1,$F5,$1F,$F0,$EF,$0F,$E0,$1F,$CE,$9F,$E8,$AE,$9B,$A7,$F8
	.dc.b	$9B,$8F,$EE,$CF,$F8,$6E,$B9,$4F,$F8,$8E,$FB,$3F,$E2,$5C,$3C,$4F
	.dc.b	$F8,$9F,$1F,$33,$FE,$29,$CB,$D4,$FF,$8A,$F3,$F7,$3F,$E2,$DD,$3E
	.dc.b	$1F,$07,$F8,$B3,$6F,$F1,$A7,$07,$EA,$36,$7C,$4F,$E6,$09,$BF,$E3
	.dc.b	$E1,$EE,$7E,$73,$DD,$FC,$EF,$AD,$F9,$EF,$7B,$F3,$FE,$BF,$E8,$3D
	.dc.b	$FF,$D0,$FB,$1F,$A2,$F8,$0F,$D1,$FB,$3F,$A4,$F8,$1F,$D2,$FB,$5F
	.dc.b	$A6,$F8,$2F,$D3,$FB,$7F,$A8,$F8,$3F,$8F,$CB,$FE,$14,$F2,$DD,$FF
	.dc.b	$D3,$7C,$CF,$EE,$3C,$BE,$5B,$E2,$F9,$F1,$EA,$FA,$73,$F0,$BE,$BD
	.dc.b	$7C,$AF,$B7,$7F,$4B,$EF,$E1,$F6,$BF,$07,$8F,$F7,$7C,$9F,$12,$3F
	.dc.b	$DC,$9C,$BF,$21,$FF,$0F,$F9,$3F,$E2,$A8,$7B,$9F,$C0,$89,$3F,$2F
	.dc.b	$C4,$FE,$0B,$EE,$FF,$06,$F8,$4F,$E0,$FE,$8F,$F0,$8F,$84,$FF,$2E
	.dc.b	$78,$FF,$EE,$1C,$3F,$DE,$4F,$FB,$B9,$BE,$8E,$BE,$03,$7F,$0B,$C3
	.dc.b	$E0,$B8,$FC,$37,$2F,$84,$E7,$F0,$FD,$3E,$1B,$F9,$96,$5F,$C2,$F3
	.dc.b	$46,$A3,$15,$C9,$D6,$1D,$9E,$19,$79,$B3,$F3,$FC,$3F,$A3,$C0,$F5
	.dc.b	$57,$E2,$77,$F7,$3F,$8B,$E1,$FE,$4B,$C5,$F1,$23,$E5,$F1,$FE,$53
	.dc.b	$97,$E5,$78,$FF,$3F,$CB,$FD,$43,$F4,$AB,$EA,$1F,$95,$F2,$3F,$24
	.dc.b	$FE,$5F,$C8,$FC,$93,$F9,$9F,$23,$F2,$4F,$E6,$FC,$8F,$C9,$3F,$9D
	.dc.b	$F2,$3F,$24,$FE,$7F,$C8,$FC,$93,$FC,$BF,$C6,$FC,$7B,$FD,$FC,$FE
	.dc.b	$65,$ED,$7F,$80,$B4,$FE,$7B,$1F,$84,$F1,$FF,$26,$FE,$1B,$C7,$FC
	.dc.b	$9B,$F8,$8F,$1F,$F2,$6F,$E2,$BC,$7F,$C9,$BF,$8C,$F1,$FF,$26,$FE
	.dc.b	$3B,$C7,$FC,$9B,$F9,$0F,$1F,$E1,$3F,$92,$E3,$FC,$A7,$95,$FB,$BE
	.dc.b	$6F,$99,$AF,$F0,$F7,$E7,$D7,$D1,$3F,$79,$E3,$7E,$41,$FD,$F7,$8D
	.dc.b	$F6,$5F,$C2,$76,$FD,$D7,$EF,$9F,$87,$F1,$BF,$1C,$FE,$27,$C6,$FC
	.dc.b	$73,$F8,$BF,$1B,$F1,$CF,$F5,$FF,$1B,$F1,$AE,$4F,$C6,$9C,$4F,$C6
	.dc.b	$9D,$4E,$E7,$E7,$8F,$19,$F9,$7F,$07,$F1,$AF,$EC,$BC,$1F,$C6,$BF
	.dc.b	$B4,$F0,$7D,$EF,$E3,$0F,$E3,$5A,$FE,$35,$FE,$DF,$E3,$FE,$E5,$F9
	.dc.b	$47,$F1,$FE,$1F,$9B,$C1,$FC,$DF,$8B,$E8,$BF,$CE,$78,$FE,$AB,$FC
	.dc.b	$EF,$93,$EC,$BF,$CF,$79,$7E,$EB,$FC,$FF,$9B,$F0,$2F,$F4,$1E,$7F
	.dc.b	$C2,$BF,$D0,$FA,$3F,$8F,$E7,$FE,$8E,$F2,$7F,$8F,$7B,$5F,$92,$8F
	.dc.b	$E1,$FD,$BF,$96,$7F,$88,$F7,$7E,$69,$FE,$27,$DF,$F9,$E7,$F8,$AF
	.dc.b	$81,$FA,$27,$F8,$BF,$83,$FA,$67,$F8,$CF,$85,$FC,$ED,$FE,$7D,$E5
	.dc.b	$7E,$BF,$A7,$F6,$31,$FC,$70,$F8,$FC,$FF,$8C,$8F,$E3,$CF,$CE,$DF
	.dc.b	$FA,$E7,$47,$D2,$7F,$AF,$74,$FD,$65,$5F,$EC,$EE,$AF,$E3,$91,$8C
	.dc.b	$3F,$C4,$31,$FC,$76,$31,$EA,$21,$E3,$FF,$1F,$AF,$C3,$1F,$C9,$3C
	.dc.b	$9F,$E2,$CF,$F2,$6F,$2B,$F8,$D3,$FC,$A3,$CB,$FE,$3E,$D3,$F2,$46
	.dc.b	$CF,$C9,$1C,$CE,$B8,$F8,$B0,$79,$7F,$A3,$9F,$E6,$7E,$27,$E9,$DF
	.dc.b	$E8,$7E,$2F,$EA,$DF,$E8,$BE,$37,$F3,$E7,$FA,$01,$F9,$22,$1F,$E5
	.dc.b	$9F,$2B,$E7,$BF,$92,$26,$3F,$E0,$7E,$3F,$AD,$1F,$60,$FE,$AC,$7F
	.dc.b	$1F,$3F,$AD,$1F,$C7,$CF,$EB,$C7,$F1,$F3,$FB,$11,$FC,$7C,$FE,$CC
	.dc.b	$7F,$1F,$3F,$B5,$1F,$C7,$FA,$BF,$C6,$AF,$F1,$F1,$3F,$C7,$8F,$CC
	.dc.b	$3C,$4F,$E3,$C7,$53,$F8,$F1,$E1,$3F,$8F,$1E,$53,$F8,$F1,$E9,$3F
	.dc.b	$8F,$1F,$CE,$CF,$E7,$A7,$F6,$E9,$FC,$A3,$FB,$8F,$0F,$F7,$5C,$3F
	.dc.b	$E6,$7E,$5F,$CA,$4F,$E5,$8F,$9E,$7F,$03,$3F,$8F,$9F,$D4,$CF,$E3
	.dc.b	$E7,$F0,$B3,$F8,$69,$FC,$3C,$FE,$DF,$AF,$F5,$F7,$F0,$87,$F7,$2F
	.dc.b	$1B,$F8,$FB,$FD,$D3,$C6,$FE,$3E,$FF,$8F,$CF,$ED,$9C,$3F,$92,$CF
	.dc.b	$F1,$5C,$5F,$96,$7F,$19,$C7,$D5,$FC,$76,$3F,$8D,$7F,$21,$C5,$F8
	.dc.b	$D7,$FC,$91,$C5,$F8,$D7,$FC,$95,$C5,$F8,$DC,$BF,$C2,$7E,$5F,$C2
	.dc.b	$DF,$E2,$47,$EE,$5F,$D5,$7C,$47,$F8,$30,$91,$FE,$01,$F5,$BF,$C1
	.dc.b	$FC,$9F,$E1,$0F,$73,$FC,$23,$EE,$FF,$84,$BD,$EF,$F0,$9F,$C4,$7F
	.dc.b	$85,$3D,$0F,$F0,$AF,$C4,$7F,$85,$BC,$7F,$F0,$BF,$C4,$34,$FF,$4B
	.dc.b	$7E,$F4,$F8,$3C,$47,$E2,$BC,$2D,$70,$A9,$F0,$9A,$E5,$FE,$03,$0F
	.dc.b	$85,$C5,$F0,$87,$C3,$7D,$0A,$FE,$0F,$EC,$FF,$91,$B2,$FF,$53,$35
	.dc.b	$FE,$00,$78,$4F,$F5,$7D,$72,$7C,$F8,$76,$73,$8D,$CB,$CD,$9F,$BA
	.dc.b	$7F,$C5,$1A,$FC,$27,$FC,$55,$BF,$C6,$7F,$C5,$9C,$3F,$29,$FF,$17
	.dc.b	$71,$FC,$E7,$E8,$1F,$E3,$1E,$6F,$A6,$FF,$03,$3F,$91,$B8,$DB,$F8
	.dc.b	$93,$FE,$44,$E6,$E5,$3F,$E4,$4A,$EC,$FF,$91,$AF,$C4,$FF,$92,$30
	.dc.b	$F3,$3F,$E4,$AC,$7D,$4F,$C3,$CF,$DA,$FD,$8D,$3E,$06,$DF,$46,$3E
	.dc.b	$DF,$07,$F2,$AD,$7E,$7E,$1F,$17,$83,$E7,$47,$D8,$9E,$3C,$1C,$B1
	.dc.b	$E7,$C1,$E0,$AF,$E5,$98,$79,$1F,$E5,$DA,$7A,$1F,$E6,$1A,$FB,$1F
	.dc.b	$E6,$5B,$7C,$07,$F9,$A6,$FF,$17,$87,$FA,$67,$0F,$F0,$4E,$07,$C3
	.dc.b	$FA,$89,$C3,$F9,$4F,$88,$F6,$FC,$4F,$BB,$DD,$F7,$FB,$DF,$03,$E2
	.dc.b	$7E,$0F,$C0,$7C,$2F,$2F,$E1,$FC,$0F,$C4,$F8,$9F,$8B,$F0,$7F,$1B
	.dc.b	$E2,$7E,$3F,$C2,$FC,$8F,$89,$F9,$3F,$0F,$F9,$5E,$4F,$CF,$F4,$FF
	.dc.b	$91,$DB,$A6,$78,$BF,$27,$5C,$5F,$D4,$E1,$D5,$FD,$5E,$5E,$17,$F5
	.dc.b	$BA,$79,$5F,$D7,$DF,$A5,$FD,$8E,$FE,$D7,$F6,$7C,$3F,$73,$8B,$E1
	.dc.b	$C7,$E0,$79,$3E,$3B,$FC,$1F,$D3,$73,$7C,$D8,$FC,$99,$F9,$83,$FD
	.dc.b	$27,$E3,$7F,$07,$7F,$D1,$5E,$37,$F0,$77,$FD,$19,$E3,$7F,$07,$7F
	.dc.b	$D1,$DE,$37,$F0,$77,$FD,$21,$E3,$7E,$11,$FC,$81,$FB,$23,$F1,$E7
	.dc.b	$EE,$BC,$0F,$C8,$72,$7F,$B0,$DF,$CA,$F8,$9F,$8E,$9E,$27,$AF,$C2
	.dc.b	$FC,$75,$7F,$83,$EB,$F1,$CF,$F8,$47,$9F,$F1,$CF,$F8,$4F,$C2,$FC
	.dc.b	$73,$FE,$15,$F0,$BF,$1C,$FF,$85,$FC,$1F,$85,$E3,$7F,$2A,$CB,$E3
	.dc.b	$70,$E3,$8B,$E3,$C7,$C5,$F0,$FF,$C1,$1C,$7F,$16,$3F,$D7,$3C,$7F
	.dc.b	$15,$FF,$5D,$F1,$FC,$57,$FD,$7B,$C7,$F1,$5F,$F5,$FF,$1F,$C5,$7E
	.dc.b	$B7,$89,$EF,$7E,$C7,$27,$E2,$75,$78,$DF,$B2,$E0,$F8,$D5,$FD,$48
	.dc.b	$F9,$67,$DF,$F1,$BF,$68,$FF,$36,$E6,$F8,$EF,$F8,$CB,$9B,$E3,$BF
	.dc.b	$E3,$4E,$6F,$8E,$FF,$8D,$B9,$BE,$3B,$FE,$38,$E6,$F8,$EF,$F8,$EB
	.dc.b	$9B,$E2,$73,$7F,$76,$E6,$F8,$B3,$FE,$DA,$8F,$91,$3C,$36,$FF,$1F
	.dc.b	$5F,$47,$FC,$81,$87,$81,$FF,$21,$63,$E4,$7F,$C8,$99,$7A,$1F,$E3
	.dc.b	$38,$FC,$63,$F1,$47,$C6,$3F,$14,$43,$FB,$B1,$93,$E2,$1F,$8A,$E6
	.dc.b	$FE,$A3,$A7,$5D,$3E,$64,$FF,$95,$B8,$FE,$C5,$7F,$91,$79,$3E,$E3
	.dc.b	$FE,$47,$E5,$FC,$03,$F8,$2A,$FC,$27,$0F,$F1,$67,$FA,$01,$FD,$D7
	.dc.b	$9F,$F6,$2F,$EC,$FC,$6F,$DA,$D7,$F6,$2E,$9F,$E2,$F3,$FD,$B3,$C6
	.dc.b	$FE,$2E,$FF,$6E,$F1,$BF,$8B,$BF,$DC,$3C,$6F,$E2,$EF,$F7,$2F,$1B
	.dc.b	$F8,$BB,$FD,$D3,$C6,$FE,$2E,$FF,$76,$F1,$BF,$8B,$BF,$87,$3F,$C8
	.dc.b	$19,$7F,$9C,$B9,$7F,$8C,$47,$F9,$DB,$CE,$DB,$FC,$26,$B9,$BF,$C2
	.dc.b	$EF,$BB,$FC,$37,$0F,$1B,$FC,$3F,$1F,$3B,$FC,$47,$2F,$5B,$FC,$4F
	.dc.b	$3F,$7B,$FC,$57,$4F,$87,$BF,$F8,$07,$6F,$C6,$75,$7E,$3A,$BE,$69
	.dc.b	$FE,$8A,$F0,$3E,$8D,$7F,$A3,$F8,$BE,$AB,$FE,$91,$E3,$FB,$2F,$FA
	.dc.b	$4F,$93,$F6,$1C,$7F,$E7,$FF,$03,$F6,$0F,$F8,$4B,$C0,$FD,$83,$FD
	.dc.b	$77,$2F,$C5,$3F,$D8,$9B,$7F,$B0,$7F,$19,$1F,$E1,$AE,$EF,$D7,$C7
	.dc.b	$F1,$53,$F9,$09,$FE,$0A,$3F,$92,$9F,$E0,$A3,$F9,$49,$FE,$0A,$3F
	.dc.b	$96,$9F,$E0,$A3,$FC,$5D,$1F,$8D,$7F,$BD,$74,$3B,$BE,$24,$7F,$82
	.dc.b	$FA,$FF,$82,$F6,$7F,$2A,$8F,$F6,$71,$E8,$29,$FA,$4B,$F6,$37,$1F
	.dc.b	$64,$60,$8E,$09,$1C,$A3,$2D,$8C,$D7,$8A,$3C,$0D,$7B,$1B,$23,$99
	.dc.b	$E6,$D9,$E7,$DF,$FB,$48,$C5,$70,$8C,$12,$FF,$D0,$ED,$8F,$0B,$5F
	.dc.b	$0C,$F2,$24,$66,$FC,$2D,$BF,$0D,$FC,$BA,$DB,$67,$57,$E3,$8C,$11
	.dc.b	$F0,$B5,$FA,$DD,$8F,$89,$FF,$18,$92,$3F,$CD,$79,$FF,$94,$FF,$7B
	.dc.b	$FF,$35,$C7,$F8,$E5,$7A,$99,$F5,$75,$F0,$BB,$F6,$B8,$7C,$2F,$1F
	.dc.b	$73,$97,$C2,$F3,$F7,$BA,$7C,$2F,$5F,$80,$ED,$F0,$BD,$FE,$0B,$FA
	.dc.b	$77,$95,$FC,$0B,$98,$DE,$27,$95,$BE,$3F,$CA,$F4,$7F,$9F,$B9,$19
	.dc.b	$75,$9C,$8C,$52,$D3,$C9,$3A,$9C,$ED,$E9,$8C,$53,$83,$DD,$18,$E2
	.dc.b	$F8,$4E,$B9,$3E,$43,$FE,$5C,$D7,$E6,$3F,$E2,$1E,$5F,$C2,$F3,$FD
	.dc.b	$3E,$9F,$BE,$DD,$FE,$FD,$7D,$82,$1F,$88,$E9,$FB,$FB,$7E,$0A,$33
	.dc.b	$7C,$9B,$5C,$9E,$17,$D5,$FC,$31,$FE,$B7,$BF,$F4,$C3,$FE,$A8,$BF
	.dc.b	$D3,$36,$32,$FF,$0F,$31,$3F,$84,$70,$43,$F7,$A6,$8D,$4F,$47,$4D
	.dc.b	$76,$62,$7B,$B8,$24,$6C,$91,$A6,$CF,$8E,$77,$39,$24,$66,$77,$3E
	.dc.b	$49,$CB,$67,$51,$88,$FD,$0C,$F0,$6C,$67,$E0,$3F,$C6,$73,$99,$F8
	.dc.b	$06,$AB,$99,$1E,$C4,$34,$53,$89,$0F,$88,$F1,$76,$43,$47,$C0,$74
	.dc.b	$E2,$3C,$8F,$23,$B1,$4E,$45,$38,$57,$C1,$6C,$FF,$BB,$2B,$DF,$39
	.dc.b	$9E,$04,$3C,$47,$62,$9D,$8A,$74,$35,$5E,$67,$27,$37,$04,$8F,$04
	.dc.b	$68,$91,$9A,$F8,$D5,$E6,$7E,$1C,$F3,$AE,$35,$B8,$F0,$C6,$09,$3C
	.dc.b	$23,$AC,$61,$DC,$66,$79,$47,$B5,$AF,$B4,$C4,$F1,$55,$E5,$7F,$6E
	.dc.b	$7E,$D0,$FD,$B9,$89,$FD,$F3,$82,$61,$FB,$F6,$E1,$FB,$F8,$C1,$1C
	.dc.b	$12,$3B,$47,$1F,$88,$79,$3F,$CA,$7E,$1B,$FA,$01,$2B,$F9,$0F,$C2
	.dc.b	$FF,$43,$F8,$6F,$E8,$9F,$09,$FD,$17,$DC,$FE,$8D,$F0,$DF,$D1,$FD
	.dc.b	$EF,$E9,$1F,$0D,$FD,$27,$E0,$19,$7F,$2D,$FD,$F1,$C1,$88,$70,$61
	.dc.b	$FE,$6D,$2F,$82,$D4,$E7,$89,$3F,$2A,$BA,$7C,$1E,$B0,$EF,$5A,$F8
	.dc.b	$9C,$57,$4F,$85,$B3,$CB,$E2,$2E,$2F,$87,$A6,$23,$35,$AC,$37,$97
	.dc.b	$0E,$8E,$3A,$F2,$D3,$E0,$69,$FD,$23,$7F,$88,$FF,$4A,$E1,$F9,$0F
	.dc.b	$F4,$CE,$3F,$98,$FC,$E3,$FA,$7F,$37,$D1,$7F,$C5,$9D,$3F,$D0,$09
	.dc.b	$D5,$FC,$EF,$8B,$F0,$BC,$7F,$7E,$BE,$EE,$7F,$7E,$72,$DB,$E3,$3F
	.dc.b	$8D,$6B,$F8,$A3,$89,$27,$F2,$3D,$5F,$86,$8F,$ED,$5E,$03,$2F,$36
	.dc.b	$3F,$56,$31,$3E,$B3,$FC,$11,$D3,$F8,$87,$FB,$5F,$4F,$E2,$1F,$ED
	.dc.b	$BD,$3F,$88,$7F,$B7,$F4,$FC,$B7,$F1,$1C,$5F,$3D,$FB,$E7,$E2,$F9
	.dc.b	$7E,$AB,$F8,$DE,$7F,$B2,$FE,$3F,$A3,$ED,$3F,$91,$E8,$FB,$4F,$23
	.dc.b	$F2,$3C,$5F,$88,$E9,$FC,$64,$60,$FC,$9E,$3E,$37,$FC,$45,$97,$9D
	.dc.b	$FF,$12,$67,$EB,$7F,$C4,$DA,$7B,$DF,$82,$7C,$3D,$FE,$37,$07,$DD
	.dc.b	$E1,$F9,$9C,$5F,$9C,$F8,$8F,$C9,$92,$FF,$51,$E2,$FE,$5B,$D6,$FC
	.dc.b	$BF,$9F,$F9,$8F,$5B,$F3,$3F,$01,$F9,$AF,$5F,$F3,$7F,$01,$F9,$CF
	.dc.b	$63,$F3,$BE,$6F,$E7,$BD,$8F,$CF,$F7,$FE,$83,$D8,$FD,$18,$F8,$0B
	.dc.b	$EC,$FC,$8E,$0F,$F2,$BF,$53,$BB,$31,$FE,$27,$BE,$0F,$EF,$5E,$AF
	.dc.b	$EF,$F1,$F0,$BF,$C0,$73,$F2,$BF,$C0,$F5,$F4,$BF,$C1,$77,$F6,$BF
	.dc.b	$C1,$F8,$7F,$6B,$C5,$F0,$E3,$EA,$72,$7C,$77,$F8,$6F,$2F,$CA,$CD
	.dc.b	$EC,$FB,$09,$3E,$7F,$81,$F7,$FB,$7F,$03,$DC,$F8,$3F,$05,$F0,$BD
	.dc.b	$EF,$87,$F0,$7F,$13,$E0,$3E,$2F,$C2,$7C,$6F,$4B,$E3,$FC,$2F,$C8
	.dc.b	$F8,$3F,$93,$F0,$DF,$2B,$E1,$7E,$5F,$C3,$FC,$9F,$1B,$F8,$C7,$6F
	.dc.b	$F0,$DE,$2F,$CA,$1F,$97,$3E,$54,$F1,$7E,$65,$F5,$7E,$76,$3E,$17
	.dc.b	$E8,$67,$E5,$7E,$96,$BE,$97,$EA,$6F,$ED,$7E,$B7,$0F,$DE,$E2,$F8
	.dc.b	$71,$F2,$B9,$3E,$3B,$F2,$B8,$3F,$45,$E2,$FC,$FD,$FD,$2F,$D2,$9F
	.dc.b	$A8,$7F,$3B,$F1,$7F,$24,$FF,$3D,$F1,$7F,$24,$FF,$3F,$F1,$7F,$24
	.dc.b	$FF,$41,$F1,$7F,$24,$FF,$43,$F1,$7F,$24,$FE,$2F,$C5,$FC,$6B,$FC
	.dc.b	$4F,$B7,$FD,$4D,$AB,$C5,$FC,$BD,$70,$AF,$C1,$F3,$FF,$AD,$47,$89
	.dc.b	$F9,$08,$CF,$89,$F9,$07,$7E,$27,$E4,$1E,$3E,$27,$E4,$1E,$7E,$27
	.dc.b	$BD,$EB,$C5,$F9,$1F,$0F,$F1,$DE,$1F,$CA,$E5,$FE,$A9,$BF,$CA,$BF
	.dc.b	$99,$CF,$F8,$E7,$F4,$DE,$1F,$E3,$9F,$D4,$78,$7F,$8E,$7F,$55,$E1
	.dc.b	$FE,$39,$FD,$67,$57,$E3,$9F,$D7,$78,$7F,$8E,$7F,$61,$E1,$FE,$39
	.dc.b	$FC,$51,$F0,$BC,$DF,$E4,$2F,$C4,$F1,$B4,$FD,$F9,$E4,$FE,$06,$BB
	.dc.b	$3F,$83,$BF,$13,$F8,$5C,$3C,$CF,$E1,$F1,$F5,$3F,$89,$CB,$DC,$FC
	.dc.b	$03,$F8,$26,$DF,$8E,$DF,$F8,$EF,$33,$87,$F1,$9E,$67,$E3,$AF,$F8
	.dc.b	$E9,$F8,$C3,$ED,$F7,$FE,$31,$FB,$BD,$FF,$8C,$7E,$FF,$7F,$E3,$1F
	.dc.b	$C0,$F7,$FE,$31,$FC,$1F,$7F,$E1,$79,$FE,$AF,$6F,$E3,$1F,$D9,$F2
	.dc.b	$FF,$BA,$F5,$FE,$39,$7F,$C6,$FB,$BF,$1A,$FF,$1B,$3F,$2C,$7F,$1B
	.dc.b	$3F,$34,$7F,$1B,$3F,$3C,$7F,$1B,$3F,$44,$7F,$1B,$3F,$4C,$7B,$6B
	.dc.b	$F8,$1F,$8F,$F0,$A3,$FC,$3F,$BF,$F8,$8F,$83,$F1,$B1,$F2,$88,$7C
	.dc.b	$D3,$F7,$1E,$0F,$EF,$5F,$DD,$78,$5F,$C0,$5F,$DE,$78,$7F,$C1,$5F
	.dc.b	$DF,$78,$9F,$C2,$5F,$E0,$1E,$2F,$F0,$D7,$F8,$17,$8D,$FC,$45,$FE
	.dc.b	$09,$E3,$FF,$19,$E7,$FE,$C9,$DF,$F8,$CD,$FE,$4D,$7F,$18,$3F,$26
	.dc.b	$7F,$18,$3F,$2E,$7F,$18,$3F,$36,$7F,$18,$3F,$3E,$7F,$18,$3F,$46
	.dc.b	$7F,$18,$3F,$4E,$7F,$18,$3F,$9C,$1F,$09,$FD,$7F,$4F,$C7,$ED,$FC
	.dc.b	$54,$FE,$D8,$FF,$27,$70,$7E,$E6,$BF,$98,$F8,$5F,$BD,$7F,$99,$F8
	.dc.b	$7F,$C0,$5F,$E6,$BE,$27,$F0,$57,$F9,$BF,$8B,$FC,$25,$FE,$73,$E3
	.dc.b	$7F,$5C,$C7,$F8,$7F,$8D,$F8,$67,$F9,$57,$47,$F8,$CF,$D1,$76,$FE
	.dc.b	$3F,$87,$FC,$A7,$7F,$D9,$AF,$89,$F8,$C3,$F8,$CE,$1D,$E7,$FC,$25
	.dc.b	$97,$8D,$FF,$0A,$67,$E7,$7F,$C2,$DA,$7A,$DF,$F0,$C6,$BE,$F7,$E1
	.dc.b	$EB,$FC,$E3,$1F,$E7,$83,$C8,$67,$FC,$5E,$7E,$5C,$FC,$DA,$F3,$F9
	.dc.b	$1F,$BC,$9F,$E3,$BE,$27,$F0,$07,$F8,$FF,$8B,$FC,$11,$FE,$43,$E3
	.dc.b	$7F,$08,$7F,$91,$F2,$7F,$0C,$7F,$92,$F8,$FF,$C4,$1F,$E4,$FE,$47
	.dc.b	$F1,$47,$F1,$79,$FF,$19,$3F,$B2,$9F,$C6,$4F,$C7,$1F,$C6,$4F,$C7
	.dc.b	$1F,$C6,$4F,$C7,$1F,$C6,$4F,$C7,$1F,$C6,$4F,$C7,$1F,$C6,$4F,$C7
	.dc.b	$1F,$C6,$4F,$C7,$1F,$AD,$7E,$1B,$FD,$97,$9F,$F1,$CF,$C9,$D5,$F0
	.dc.b	$FF,$E2,$82,$47,$F8,$8F,$E2,$3F,$C5,$7E,$D7,$F8,$B3,$DB,$FF,$47
	.dc.b	$FE,$FB,$FE,$9F,$AF,$F0,$CA,$F7,$73,$F1,$1A,$F7,$F7,$F1,$1C,$3E
	.dc.b	$07,$8F,$C4,$72,$F8,$3E,$7F,$11,$D3,$E1,$7A,$FC,$47,$6F,$87,$FF
	.dc.b	$4E,$F8,$BF,$92,$DF,$FD,$63,$CD,$FE,$A9,$8C,$47,$8A,$78,$BA,$EE
	.dc.b	$FF,$54,$F0,$B1,$EF,$39,$3F,$15,$0C,$FC,$B1,$9E,$F4,$B6,$F5,$CE
	.dc.b	$61,$C1,$F0,$23,$1C,$5F,$19,$F9,$07,$F9,$37,$93,$FC,$AB,$CB,$F3
	.dc.b	$A3,$FC,$31,$CD,$F5,$32,$FC,$DB,$9E,$0F,$EE,$6F,$09,$E4,$7F,$95
	.dc.b	$F9,$BB,$3E,$03,$C9,$CD,$F4,$5F,$37,$37,$D1,$7D,$1C,$DF,$45,$F5
	.dc.b	$73,$7D,$17,$D9,$CD,$F4,$5F,$CC,$E5,$FE,$40,$F2,$31,$1F,$80,$9E
	.dc.b	$0F,$E0,$AB,$A3,$F8,$4B,$F0,$3F,$86,$C3,$C8,$FE,$23,$1F,$C5,$EF
	.dc.b	$EA,$8F,$F1,$D6,$7F,$90,$7F,$24,$7F,$9D,$74,$F9,$1E,$87,$FA,$3B
	.dc.b	$89,$F1,$1F,$94,$25,$7D,$5F,$6B,$F2,$FF,$11,$F9,$8F,$82,$FC,$CF
	.dc.b	$C2,$7E,$6B,$DE,$FC,$DF,$C2,$FE,$73,$E0,$3F,$3B,$F0,$DF,$9E,$F8
	.dc.b	$2F,$F7,$6F,$EF,$DF,$EF,$BD,$3F,$CE,$EF,$FB,$ED,$FC,$33,$FE,$AE
	.dc.b	$E9,$FF,$7E,$2F,$89,$E3,$F0,$FF,$D4,$6B,$FD,$A5,$EF,$FF,$76,$79
	.dc.b	$C7,$69,$EF,$F1,$3E,$0F,$87,$F0,$F8,$DE,$3B,$FC,$86,$9E,$77,$F2
	.dc.b	$5B,$7A,$DF,$CA,$70,$7B,$DF,$CB,$71,$7E,$E3,$8F,$E2,$C7,$EE,$39
	.dc.b	$3F,$21,$E4,$FF,$02,$8F,$9C,$43,$FC,$D9,$E1,$7D,$37,$EA,$9F,$A1
	.dc.b	$3E,$C9,$F9,$03,$F4,$9E,$27,$E9,$9F,$E0,$BE,$2F,$EB,$1F,$E0,$FE
	.dc.b	$37,$EC,$9F,$E1,$3E,$3F,$EE,$1F,$DD,$1F,$90,$DF,$FB,$BB,$FA,$FF
	.dc.b	$23,$5C,$E9,$FD,$ED,$73,$8C,$57,$EF,$2F,$C1,$18,$F1,$FF,$77,$8F
	.dc.b	$96,$31,$DB,$FB,$AB,$F5,$1F,$C7,$3C,$1F,$CA,$4F,$F9,$93,$C1,$F8
	.dc.b	$4F,$F9,$9B,$6F,$DB,$69,$FD,$93,$C1,$F9,$66,$30,$FE,$5C,$FE,$E5
	.dc.b	$FF,$4B,$79,$1F,$C1,$1F,$E5,$27,$CA,$3F,$48,$7C,$A3,$F7,$27,$F0
	.dc.b	$FC,$FD,$75,$FC,$73,$5F,$80,$FC,$29,$FD,$C9,$F1,$CF,$E3,$5E,$4F
	.dc.b	$F2,$6C,$BF,$94,$7A,$FF,$BA,$7E,$0F,$93,$FB,$A7,$E1,$F9,$3F,$BA
	.dc.b	$7E,$2F,$93,$FB,$A7,$E3,$F9,$3F,$BA,$7E,$4F,$93,$E9,$7E,$5E,$5E
	.dc.b	$D7,$FD,$17,$B7,$C1,$7F,$8D,$C7,$E3,$7C,$9F,$F6,$F7,$07,$C7,$E5
	.dc.b	$FE,$43,$3F,$C0,$0F,$8E,$7F,$1F,$3F,$20,$7F,$1F,$3E,$39,$FD,$4B
	.dc.b	$A3,$EE,$70,$7F,$40,$E2,$F8,$F3,$F8,$33,$F8,$F9,$FC,$E7,$9F,$F1
	.dc.b	$73,$FD,$1B,$A3,$FC,$88,$FE,$42,$7F,$26,$7F,$1F,$38,$1C,$8F,$CD
	.dc.b	$9F,$CB,$0F,$AC,$7D,$83,$C8,$7E,$33,$3F,$C7,$F3,$FE,$32,$7F,$C5
	.dc.b	$7C,$FF,$8C,$7F,$C5,$BC,$FF,$0F,$A7,$F6,$98,$37,$FE,$41,$D5,$FB
	.dc.b	$AC,$BF,$32,$FE,$F6,$7F,$7E,$7F,$55,$F2,$7F,$8F,$BF,$D5,$FC,$9F
	.dc.b	$E3,$EF,$F5,$9F,$27,$F8,$FB,$FD,$6F,$C9,$FE,$1E,$FF,$1C,$F1,$3F
	.dc.b	$89,$BF,$C7,$BC,$6F,$E3,$6F,$FA,$DB,$2F,$C5,$BF,$8E,$ED,$FE,$3D
	.dc.b	$CC,$F2,$5E,$47,$F1,$DA,$FE,$4F,$E4,$7F,$1D,$7F,$94,$F1,$FF,$1D
	.dc.b	$7F,$C1,$7E,$47,$F1,$D7,$FC,$1B,$E4,$7F,$1D,$7F,$C1,$FE,$47,$F1
	.dc.b	$D7,$FC,$23,$E4,$7C,$17,$F8,$E1,$FD,$D7,$93,$F6,$6F,$F1,$BD,$BF
	.dc.b	$22,$FF,$1C,$3F,$B5,$9F,$C7,$0F,$D9,$9F,$8D,$7F,$3A,$FE,$CC,$FE
	.dc.b	$B4,$7D,$E3,$FA,$F1,$FE,$03,$F1,$BF,$66,$FF,$83,$3B,$3F,$1A,$FF
	.dc.b	$83,$BB,$3F,$1A,$FF,$96,$7C,$97,$95,$FB,$3D,$BF,$C2,$FE,$3F,$EC
	.dc.b	$DF,$F0,$CF,$8F,$FB,$37,$FC,$37,$E3,$FE,$CD,$FF,$0E,$F8,$FF,$B3
	.dc.b	$7F,$C3,$FE,$3F,$EC,$DF,$F1,$0E,$5F,$B3,$7F,$CF,$3E,$3F,$EC,$DF
	.dc.b	$E3,$11,$FB,$3F,$38,$FF,$9B,$7A,$1F,$11,$FD,$88,$91,$FD,$81,$AF
	.dc.b	$F6,$46,$F1,$7F,$62,$6D,$F4,$EA,$FC,$24,$62,$BB,$36,$33,$1D,$48
	.dc.b	$75,$3F,$8E,$0A,$FC,$8B,$6B,$B7,$95,$FC,$7D,$FD,$E4,$72,$9F,$23
	.dc.b	$63,$C9,$87,$BA,$30,$79,$0B,$7E,$47,$C4,$4A,$E6,$DA,$E7,$87,$AE
	.dc.b	$FA,$4F,$0F,$0D,$96,$A3,$AC,$EB,$0E,$64,$C3,$53,$D4,$F2,$17,$1F
	.dc.b	$24,$EE,$32,$D8,$CC,$F2,$8C,$57,$78,$F8,$3F,$07,$F9,$4A,$71,$FE
	.dc.b	$53,$0F,$F0,$EC,$FA,$F3,$91,$E8,$6B,$E3,$C3,$A3,$E2,$C3,$C0,$47
	.dc.b	$04,$C3,$E6,$67,$F1,$23,$F2,$1C,$7F,$E3,$02,$3F,$E2,$EF,$6F,$FC
	.dc.b	$67,$F0,$7F,$E3,$4F,$5F,$FC,$6B,$F1,$1F,$E3,$6F,$67,$FC,$6F,$F1
	.dc.b	$1F,$E3,$8F,$6F,$FC,$73,$F1,$1F,$E3,$AF,$77,$FC,$77,$F1,$1F,$E3
	.dc.b	$CF,$7F,$FC,$7B,$E4,$FF,$8F,$BD,$0F,$D0,$EF,$F8,$B3,$FD,$9D,$FB
	.dc.b	$FF,$FB,$7F,$A7,$FC,$0F,$E1,$7F,$A5,$2E,$57,$E5,$27,$FC,$1D,$5B
	.dc.b	$E6,$FC,$6E,$1D,$AB,$FC,$8F,$8F,$89,$FF,$24,$E7,$E6,$7F,$C9,$7A
	.dc.b	$FA,$9F,$F2,$6E,$FE,$E7,$FC,$9F,$C3,$F0,$9F,$88,$7F,$94,$B9,$3E
	.dc.b	$3B,$FD,$1F,$AF,$E8,$73,$FE,$0A,$3E,$87,$47,$D6,$8C,$11,$C1,$3A
	.dc.b	$7E,$EF,$07,$E2,$1B,$BF,$49,$DF,$57,$E2,$7A,$FF,$11,$8F,$F9,$9A
	.dc.b	$36,$F8,$2E,$E3,$DD,$B7,$DE,$7F,$18,$43,$D4,$7E,$80,$FA,$B3,$FB
	.dc.b	$36,$C7,$EB,$9B,$3F,$ED,$57,$F5,$26,$0C,$4F,$17,$24,$3B,$15,$D3
	.dc.b	$82,$38,$24,$F8,$63,$F8,$79,$F8,$98,$76,$7F,$13,$84,$63,$93,$FA
	.dc.b	$04,$74,$AE,$4D,$BF,$53,$4F,$4C,$6A,$BC,$33,$A6,$CE,$A3,$13,$B2
	.dc.b	$38,$3F,$3F,$E5,$37,$F8,$13,$AA,$E4,$43,$FE,$00,$5F,$66,$C6,$DB
	.dc.b	$7D,$5E,$07,$D1,$32,$60,$EE,$F9,$B2,$FC,$56,$FF,$EB,$07,$FA,$6C
	.dc.b	$F1,$21,$C0,$E4,$7C,$53,$A9,$0D,$99,$2D,$FA,$71,$F8,$11,$93,$A1
	.dc.b	$C4,$A7,$0C,$3D,$CD,$C3,$DC,$EC,$E4,$7E,$1B,$D8,$73,$FF,$06,$8F
	.dc.b	$C2,$6B,$F8,$3C,$55,$A9,$FE,$BE,$75,$3D,$B1,$F3,$DC,$1D,$DC,$3E
	.dc.b	$62,$1F,$D1,$B1,$F6,$C6,$A3,$CA,$D2,$63,$EC,$8E,$31,$8C,$7D,$B1
	.dc.b	$F0,$72,$FA,$AF,$F4,$43,$CE,$70,$AE,$4F,$66,$D7,$38,$F5,$C6,$0F
	.dc.b	$EC,$B7,$E6,$7F,$81,$47,$F4,$BB,$FE,$0A,$47,$F8,$27,$07,$F1,$36
	.dc.b	$F0,$7F,$13,$C7,$E7,$4F,$F8,$10,$FF,$79,$BC,$67,$A1,$C0,$FE,$28
	.dc.b	$5C,$7F,$8A,$37,$0F,$EF,$3D,$DF,$D5,$65,$D0,$F8,$1F,$E8,$F8,$FF
	.dc.b	$40,$6E,$5F,$E1,$F9,$D6,$3F,$E2,$CA,$FE,$8B,$ED,$EA,$3F,$A6,$FC
	.dc.b	$0D,$C7,$AF,$C4,$5C,$1F,$05,$7E,$23,$1F,$07,$9F,$88,$D7,$C2,$6F
	.dc.b	$E2,$38,$76,$F1,$F8,$8E,$5F,$0D,$CF,$E2,$3A,$7C,$3F,$EB,$78,$7F
	.dc.b	$DD,$C7,$F9,$4F,$BB,$0F,$F1,$9D,$D3,$C6,$3D,$5C,$BE,$DC,$3B,$46
	.dc.b	$0F,$EC,$F9,$78,$DC,$71,$7C,$8D,$7D,$11,$8C,$5B,$FB,$63,$1C,$3F
	.dc.b	$07,$8B,$E1,$71,$FC,$3E,$4F,$89,$CD,$2F,$FC,$AF,$CF,$F5,$7A,$3F
	.dc.b	$13,$D1,$F8,$59,$FB,$E4,$70,$4E,$1F,$EB,$5D,$5F,$85,$8F,$EF,$B3
	.dc.b	$FE,$9B,$21,$FE,$C2,$78,$3F,$88,$3A,$4F,$27,$F1,$13,$CE,$31,$1F
	.dc.b	$D6,$B8,$3F,$13,$8F,$A6,$3E,$73,$ED,$7E,$26,$7F,$05,$F8,$67,$C7
	.dc.b	$E9,$FC,$0B,$F2,$7A,$7F,$02,$FC,$BE,$9F,$C0,$BF,$37,$A7,$F0,$2F
	.dc.b	$CF,$E9,$FC,$0B,$FB,$93,$FA,$B7,$83,$FD,$E3,$66,$BF,$EE,$17,$F5
	.dc.b	$33,$F5,$FA,$7F,$04,$FE,$17,$A7,$F0,$2F,$E1,$FA,$7F,$00,$FE,$27
	.dc.b	$A7,$EF,$BF,$8B,$E8,$FB,$CF,$E3,$7A,$3D,$CF,$E3,$F8,$7E,$13,$F1
	.dc.b	$0F,$C3,$71,$7F,$8D,$39,$BF,$65,$D2,$F8,$8F,$D3,$12,$BF,$19,$F0
	.dc.b	$FF,$A9,$F8,$3F,$D5,$7A,$1F,$AB,$F8,$4F,$D7,$7C,$0F,$F8,$FA,$FF
	.dc.b	$BB,$2F,$85,$CF,$A9,$AF,$87,$DF,$AD,$C3,$E1,$F8,$FB,$1C,$BD,$CE
	.dc.b	$7E,$D7,$4F,$73,$F5,$3D,$9F,$C4,$7B,$7F,$D2,$7E,$86,$1F,$F5,$16
	.dc.b	$CA,$E3,$BF,$E9,$A1,$8F,$68,$C6,$7E,$27,$5A,$F9,$9E,$1B,$FB,$1F
	.dc.b	$71,$FA,$DE,$1F,$84,$FC,$43,$F6,$1C,$9F,$1D,$FC,$A6,$2F,$73,$E2
	.dc.b	$92,$3C,$9E,$87,$C8,$F7,$7E,$4F,$C4,$7C,$AF,$7F,$E5,$FC,$47,$CC
	.dc.b	$F8,$1F,$9B,$F1,$1F,$3B,$E0,$FE,$7F,$C4,$7D,$0F,$85,$FA,$3B,$7F
	.dc.b	$1F,$FE,$00,$FC,$7E,$DF,$C1,$97,$C4,$E7,$D9,$D7,$C4,$FF,$50,$AF
	.dc.b	$F1,$65,$23,$51,$C6,$BA,$75,$76,$8F,$99,$8F,$89,$F9,$D9,$F9,$9F
	.dc.b	$A1,$AF,$A9,$FA,$5B,$FB,$9F,$A9,$C3,$F8,$8E,$2F,$89,$1F,$88,$E3
	.dc.b	$FC,$A7,$97,$FC,$FF,$97,$E6,$76,$FF,$A0,$9F,$E7,$5E,$27,$D2,$BF
	.dc.b	$E5,$9D,$1F,$59,$FC,$89,$F9,$4E,$9F,$B8,$FE,$53,$AB,$F0,$0F,$E5
	.dc.b	$BA,$FF,$08,$FE,$63,$B3,$F1,$2F,$E6,$BB,$7F,$1C,$FF,$93,$BD,$CF
	.dc.b	$F0,$71,$FE,$0F,$F5,$DE,$DF,$E5,$E7,$F2,$1E,$17,$E6,$DF,$C9,$78
	.dc.b	$7F,$9F,$7F,$29,$E2,$7E,$8D,$FC,$B7,$8B,$E7,$7F,$30,$7A,$F8,$BF
	.dc.b	$D4,$3C,$1E,$F7,$FD,$47,$C3,$F1,$3A,$FE,$35,$7F,$38,$E4,$FD,$1F
	.dc.b	$2F,$CC,$BF,$C1,$FA,$EF,$07,$F2,$97,$F9,$23,$E9,$9F,$A7,$F1,$7F
	.dc.b	$24,$FE,$A7,$C5,$FC,$93,$FA,$BF,$17,$F2,$4F,$EB,$7C,$5F,$C9,$3F
	.dc.b	$AF,$BF,$C3,$BF,$89,$3F,$25,$E2,$FE,$41,$FE,$CB,$8F,$FB,$93,$DB
	.dc.b	$C4,$7E,$4A,$7F,$35,$E2,$72,$8F,$E1,$77,$D9,$FE,$1B,$87,$89,$FE
	.dc.b	$1F,$8F,$99,$FE,$23,$97,$A9,$FE,$27,$9F,$EB,$DF,$80,$7E,$3C,$FE
	.dc.b	$0D,$AF,$E4,$1F,$F6,$AF,$77,$CD,$69,$FB,$F3,$FB,$D7,$93,$FC,$89
	.dc.b	$FE,$E1,$E4,$FF,$07,$7F,$84,$9F,$70,$FE,$1A,$7E,$00,$FE,$22,$7E
	.dc.b	$10,$FE,$2D,$DB,$F9,$0E,$1F,$AF,$AF,$F1,$B9,$FE,$99,$1F,$C8,$5F
	.dc.b	$C9,$79,$3F,$94,$E2,$FB,$9D,$DC,$79,$7F,$91,$E3,$F9,$CC,$39,$F7
	.dc.b	$FE,$81,$EB,$E0,$7E,$91,$EF,$E0,$FA,$73,$FF,$0B,$6B,$ED,$7D,$E7
	.dc.b	$C1,$9F,$D9,$3F,$C8,$0F,$C8,$1F,$EB,$2D,$7F,$1A,$FC,$D2,$47,$B3
	.dc.b	$C2,$FA,$51,$F5,$0F,$E0,$A7,$E3,$CF,$E1,$27,$E3,$CF,$E1,$A7,$E3
	.dc.b	$CF,$E2,$27,$E3,$CF,$E3,$D9,$FE,$21,$FC,$79,$FC,$87,$C8,$FE,$3D
	.dc.b	$5F,$A2,$F2,$7F,$8F,$4E,$8F,$E4,$3D,$1F,$8F,$9F,$A3,$8F,$E3,$DF
	.dc.b	$A7,$D1,$F8,$F7,$EA,$F4,$7E,$3D,$FA,$FD,$1F,$8F,$7E,$CF,$47,$B9
	.dc.b	$F8,$18,$FE,$38,$87,$F8,$67,$CA,$FE,$E0,$FC,$B3,$FC,$8D,$D1,$FC
	.dc.b	$79,$FF,$1E,$74,$7D,$3A,$FD,$DD,$38,$7E,$C4,$7D,$AA,$FC,$81,$FC
	.dc.b	$7C,$FC,$81,$FC,$7C,$FC,$81,$FC,$7C,$FE,$2E,$7F,$67,$E6,$FC,$67
	.dc.b	$8B,$FC,$7B,$A7,$F1,$DB,$FE,$4E,$BF,$BE,$F8,$7F,$C7,$5F,$EF,$FE
	.dc.b	$1F,$F1,$D7,$FC,$03,$E1,$FF,$1D,$7F,$C0,$7E,$1F,$F1,$D7,$F9,$A9
	.dc.b	$FA,$83,$FB,$61,$EE,$3F,$1D,$CF,$FC,$71,$F8,$67,$E3,$BB,$7F,$8D
	.dc.b	$77,$7F,$1D,$9F,$E3,$F3,$FB,$CD,$BF,$53,$E3,$7F,$4F,$8F,$F0,$9F
	.dc.b	$8F,$FD,$4D,$FF,$0A,$F9,$1F,$D5,$DF,$F0,$BF,$93,$FD,$6D,$FF,$0C
	.dc.b	$F9,$5F,$D7,$DF,$EC,$B8,$FE,$35,$FE,$2C,$7E,$34,$FE,$37,$E9,$FF
	.dc.b	$8F,$E1,$0D,$1D,$DF,$90,$8E,$27,$E4,$0F,$CC,$9F,$9B,$3C,$27,$E7
	.dc.b	$CF,$29,$FA,$33,$D2,$7E,$9C,$F6,$9F,$CE,$4F,$E7,$67,$F8,$CB,$C0
	.dc.b	$F8,$CF,$ED,$76,$FE,$3D,$5E,$4F,$89,$FF,$04,$12,$7F,$C1,$1E,$D7
	.dc.b	$F8,$2F,$CC,$FF,$06,$79,$3F,$E0,$DF,$89,$FF,$07,$7C,$4F,$F8,$3F
	.dc.b	$E0,$3F,$C2,$1F,$13,$FE,$11,$F8,$2F,$F0,$97,$C4,$FF,$84,$FE,$13
	.dc.b	$FC,$29,$F1,$3F,$E1,$5F,$86,$FF,$0B,$7C,$4F,$F8,$5F,$E2,$3F,$A0
	.dc.b	$F6,$FF,$1D,$E3,$78,$9F,$EF,$BE,$7F,$91,$3F,$B9,$AE,$3C,$9F,$E1
	.dc.b	$DC,$7A,$BF,$E1,$FC,$FC,$2F,$F8,$87,$5F,$2B,$FE,$23,$DF,$D2,$FF
	.dc.b	$89,$78,$7D,$AF,$F8,$9F,$8F,$FC,$55,$C9,$F0,$E3,$FC,$4B,$CB,$F2
	.dc.b	$35,$39,$D7,$F8,$63,$C2,$79,$1F,$E3,$1E,$5F,$90,$FF,$8C,$F9,$7E
	.dc.b	$43,$FE,$35,$E5,$F9,$0F,$F8,$DF,$97,$E4,$3F,$E3,$9E,$5F,$90,$FF
	.dc.b	$8E,$F9,$7E,$43,$FE,$3D,$F1,$3F,$8E,$F4,$FE,$43,$9F,$F1,$F1,$A9
	.dc.b	$FF,$1E,$D7,$27,$FC,$7F,$7D,$9F,$F2,$0E,$1E,$27,$FC,$87,$8F,$99
	.dc.b	$FF,$22,$E5,$EA,$7F,$C8,$F9,$FE,$15,$FC,$39,$FE,$4F,$D3,$E4,$7B
	.dc.b	$3F,$EF,$DD,$DF,$13,$FE,$55,$25,$7F,$B6,$BF,$80,$5F,$17,$FE,$F7
	.dc.b	$7F,$DF,$B0,$E2,$FF,$0C,$C3,$E2,$75,$EF,$EF,$E2,$B8,$7C,$0F,$1F
	.dc.b	$8A,$E5,$F0,$7C,$FE,$2B,$A7,$C2,$F5,$F8,$AE,$DF,$05,$DF,$E2,$BC
	.dc.b	$1F,$11,$F9,$2E,$EF,$F4,$6F,$0F,$FB,$5B,$87,$FA,$47,$BF,$A8,$F4
	.dc.b	$5F,$27,$D5,$8F,$67,$D9,$9F,$89,$F7,$6B,$E6,$7E,$06,$FE,$A7,$E1
	.dc.b	$70,$FC,$1F,$88,$F8,$5C,$9F,$1A,$7F,$CE,$BC,$BF,$91,$E8,$FC,$86
	.dc.b	$FF,$7F,$9F,$F2,$75,$F4,$8F,$F3,$DF,$85,$F9,$C7,$FC,$FB,$E1,$FE
	.dc.b	$81,$FF,$3F,$F8,$9F,$A4,$7F,$D0,$3E,$2F,$EA,$1F,$F4,$1F,$8D,$FA
	.dc.b	$C7,$FD,$0B,$E3,$FE,$C1,$FF,$43,$F9,$1F,$E8,$FF,$1B,$F2,$31,$FC
	.dc.b	$7F,$3F,$DC,$4E,$FA,$3F,$B8,$5F,$37,$FB,$96,$1D,$DF,$EE,$98,$F8
	.dc.b	$DF,$EE,$D9,$79,$DF,$EF,$19,$FA,$DF,$EF,$5A,$7E,$BF,$3F,$DA,$9F
	.dc.b	$91,$3F,$A3,$67,$F1,$1C,$92,$7F,$C0,$5E,$47,$E6,$63,$F3,$67,$F0
	.dc.b	$33,$F6,$C7,$F0,$73,$F6,$C7,$F0,$B3,$F6,$C7,$F0,$F3,$F6,$C7,$F1
	.dc.b	$33,$F6,$C7,$FA,$C7,$4F,$C6,$3F,$E3,$83,$F1,$9B,$7D,$9E,$9F,$DB
	.dc.b	$5F,$F8,$57,$C7,$FD,$B3,$FE,$17,$F1,$FF,$6C,$FF,$86,$7C,$7F,$DB
	.dc.b	$3F,$E1,$BF,$1F,$F6,$CF,$F8,$77,$C7,$FD,$B3,$FE,$1F,$F1,$FD,$CF
	.dc.b	$C0,$CF,$E3,$1F,$13,$C7,$FD,$B7,$3F,$E3,$23,$ED,$3F,$36,$7E,$7C
	.dc.b	$7F,$1B,$3E,$31,$FC,$10,$F8,$C7,$F0,$83,$E3,$1F,$C3,$0F,$8C,$7F
	.dc.b	$10,$3E,$31,$FC,$50,$F8,$C7,$F1,$B3,$FB,$5D,$7E,$35,$C9,$B3,$F2
	.dc.b	$67,$D0,$3F,$2E,$7D,$43,$F3,$67,$D8,$3F,$3E,$7D,$C3,$F4,$67,$AC
	.dc.b	$FD,$39,$FA,$98,$FE,$76,$FC,$6A,$FE,$35,$D1,$FB,$39,$FF,$25,$3F
	.dc.b	$B6,$9F,$E3,$5D,$5F,$C6,$7A,$FF,$79,$1F,$C6,$78,$3F,$9A,$78,$EE
	.dc.b	$DF,$E0,$53,$93,$FC,$0D,$0E,$EF,$E1,$11,$9F,$21,$2F,$03,$F8,$74
	.dc.b	$E6,$1E,$0F,$F1,$38,$C7,$85,$FC,$5D,$FF,$7E,$F3,$FF,$85,$1E,$9E
	.dc.b	$4F,$F8,$F5,$FF,$0D,$1C,$FC,$4F,$E4,$8F,$F3,$5F,$17,$F9,$43,$FC
	.dc.b	$DF,$C6,$FE,$58,$FF,$39,$F1,$FF,$98,$3F,$CE,$FC,$8F,$E6,$8F,$F3
	.dc.b	$DF,$27,$E2,$3F,$18,$FE,$81,$E4,$FC,$BE,$3F,$E3,$5C,$DF,$C7,$67
	.dc.b	$F2,$BC,$FF,$49,$F8,$5E,$13,$C2,$FC,$DD,$FF,$21,$EB,$FC,$FB,$FC
	.dc.b	$8F,$B7,$F4,$6F,$F2,$5E,$EF,$D3,$BF,$C9,$FB,$FF,$56,$FF,$29,$F0
	.dc.b	$3F,$5E,$FF,$2B,$F0,$7F,$66,$FF,$2D,$F2,$7F,$8E,$F6,$7F,$6C,$D7
	.dc.b	$F2,$7B,$7E,$DD,$FC,$DB,$FB,$77,$F3,$EF,$ED,$DF,$D1,$BF,$B7,$7F
	.dc.b	$4E,$FE,$DD,$FD,$5B,$FB,$77,$FC,$12,$7C,$0D,$7F,$69,$3F,$E3,$78
	.dc.b	$FE,$3B,$5F,$A5,$7E,$2B,$6B,$FD,$3C,$DB,$F5,$4F,$EB,$4D,$96,$76
	.dc.b	$D8,$DB,$A2,$46,$9B,$1A,$8F,$F1,$C7,$3F,$F8,$DE,$BC,$15,$FD,$0D
	.dc.b	$B7,$F2,$63,$89,$93,$C0,$53,$2E,$EB,$C0,$4B,$F0,$12,$78,$47,$97
	.dc.b	$C1,$FF,$08,$5F,$83,$83,$C9,$5D,$B0,$E9,$87,$16,$C7,$28,$C4,$F0
	.dc.b	$8F,$1B,$70,$F1,$46,$C9,$4F,$33,$FD,$05,$1F,$E8,$16,$C7,$62,$57
	.dc.b	$3B,$EF,$87,$9C,$97,$E7,$9D,$3D,$9B,$5F,$2F,$17,$47,$F0,$EF,$8A
	.dc.b	$FF,$15,$12,$FF,$C5,$DF,$15,$FE,$2D,$F7,$7F,$C5,$DF,$11,$FE,$2F
	.dc.b	$F7,$BF,$C6,$1F,$13,$FE,$31,$F7,$BF,$83,$70,$B8,$3F,$C9,$9F,$BA
	.dc.b	$7F,$09,$CF,$F8,$A2,$F8,$6D,$CE,$BD,$BB,$5F,$EA,$1A,$7C,$3E,$38
	.dc.b	$90,$F7,$70,$BE,$22,$FB,$F8,$F8,$9F,$ED,$37,$FE,$84,$E8,$E5,$D5
	.dc.b	$FE,$85,$F2,$75,$A7,$FA,$C3,$83,$D0,$F8,$71,$EB,$97,$6D,$3B,$EB
	.dc.b	$E0,$DB,$C3,$C1,$E2,$C3,$E1,$1F,$E5,$5E,$2F,$8C,$FF,$95,$F9,$3E
	.dc.b	$53,$F3,$0F,$F2,$D7,$3F,$CF,$5D,$3F,$92,$7E,$D7,$07,$DC,$7F,$16
	.dc.b	$DD,$BF,$17,$C3,$F8,$48,$C7,$27,$E1,$2B,$89,$FD,$90,$A7,$22,$D7
	.dc.b	$2E,$AF,$C2,$92,$B9,$CF,$E2,$5F,$E1,$18,$79,$63,$EE,$61,$FB,$57
	.dc.b	$FA,$33,$5F,$E0,$66,$29,$5F,$83,$8C,$90,$FD,$D1,$B7,$4E,$0D,$4F
	.dc.b	$57,$04,$70,$49,$EF,$1D,$E3,$3D,$7F,$AC,$3F,$5B,$3A,$9F,$5B,$5F
	.dc.b	$11,$C5,$C3,$E2,$6B,$E1,$6C,$F8,$6B,$D2,$D9,$E5,$18,$9D,$91,$C1
	.dc.b	$FD,$BF,$E2,$9E,$BF,$78,$CD,$72,$21,$E8,$72,$D3,$27,$A8,$C1,$0C
	.dc.b	$10,$EC,$70,$3E,$7E,$1F,$1F,$C4,$FE,$91,$1B,$3E,$04,$F8,$1C,$63
	.dc.b	$EF,$7B,$1D,$0A,$70,$3B,$14,$61,$F0,$1C,$1D,$31,$F8,$CF,$09,$EC
	.dc.b	$E6,$BC,$4F,$73,$06,$E7,$8B,$F9,$8E,$7F,$F0,$D3,$FD,$5C,$D9,$E6
	.dc.b	$AE,$AF,$32,$1E,$42,$BA,$7F,$A9,$5F,$A6,$30,$4A,$F0,$CF,$62,$1C
	.dc.b	$9C,$B6,$FD,$6F,$F4,$3C,$3F,$91,$63,$F9,$89,$CD,$F5,$6D,$F2,$BF
	.dc.b	$31,$27,$4F,$C3,$9F,$C7,$65,$FC,$1E,$3F,$82,$13,$4F,$E0,$93,$FC
	.dc.b	$0A,$BF,$87,$4F,$F0,$8D,$78,$56,$A3,$F8,$71,$C6,$32,$47,$F8,$51
	.dc.b	$C4,$A6,$CB,$1F,$DC,$FB,$BF,$A4,$FC,$17,$F3,$E2,$4F,$F3,$7A,$FE
	.dc.b	$9B,$F0,$9F,$D0,$BD,$DF,$E8,$7E,$F7,$F4,$4F,$83,$FE,$8B,$F0,$DF
	.dc.b	$D1,$BE,$13,$FA,$3F,$C3,$FF,$48,$F8,$5F,$E9,$3F,$11,$FD,$2B,$E1
	.dc.b	$BF,$A5,$FC,$4F,$F4,$CF,$87,$FE,$9B,$F1,$1F,$D5,$3F,$7C,$70,$FF
	.dc.b	$46,$3F,$CE,$4E,$3E,$2F,$FD,$B6,$D8,$F3,$07,$8F,$17,$48,$CF,$37
	.dc.b	$F5,$CC,$3C,$31,$8F,$29,$9F,$9A,$31,$AF,$A9,$D6,$FE,$E7,$87,$0F
	.dc.b	$C2,$79,$71,$FF,$66,$E4,$F9,$31,$FD,$CB,$97,$FC,$F9,$DF,$FC,$EF
	.dc.b	$A3,$ED,$4F,$D3,$DB,$EF,$B7,$AF,$FA,$23,$F6,$63,$33,$C0,$FF,$1D
	.dc.b	$F5,$7E,$1E,$B7,$5C,$63,$04,$79,$47,$F3,$1B,$FF,$09,$C7,$F8,$07
	.dc.b	$0F,$3D,$FF,$74,$C3,$FC,$03,$1E,$B9,$F6,$9F,$1F,$A7,$F0,$0F,$C9
	.dc.b	$E9,$FC,$03,$F2,$FA,$7F,$00,$FC,$DE,$9F,$C0,$3F,$3F,$A7,$F0,$0F
	.dc.b	$D1,$E9,$FC,$03,$F7,$F3,$FF,$02,$F4,$7D,$D8,$FF,$57,$76,$7F,$25
	.dc.b	$7E,$FC,$66,$77,$1C,$7A,$7F,$C4,$77,$DA,$3F,$C4,$78,$78,$9F,$F1
	.dc.b	$2E,$3E,$67,$FC,$4F,$97,$A9,$FF,$14,$E7,$EE,$7F,$C5,$7A,$7E,$85
	.dc.b	$FC,$47,$17,$C8,$CB,$F5,$31,$F9,$8E,$0F,$9B,$AF,$E6,$BB,$BD,$A4
	.dc.b	$9F,$48,$F2,$7E,$0F,$C2,$7C,$2F,$43,$E1,$FC,$37,$C4,$F0,$BE,$2F
	.dc.b	$C3,$FC,$6F,$5F,$E3,$F8,$9F,$23,$D7,$F9,$3F,$11,$F2,$BD,$8F,$97
	.dc.b	$F1,$3F,$33,$D9,$F9,$BF,$15,$F3,$BC,$9F,$D4,$F0,$9C,$BE,$49,$E2
	.dc.b	$F9,$AF,$AB,$E8,$C7,$C2,$FA,$B3,$F2,$BE,$CD,$7D,$2F,$BB,$7F,$6B
	.dc.b	$F0,$38,$7E,$13,$FB,$C3,$FA,$99,$F2,$7C,$AF,$E0,$3C,$9F,$C0,$BD
	.dc.b	$AF,$C2,$92,$7E,$4F,$C5,$7E,$23,$C5,$FC,$4F,$B9,$FC,$27,$F8,$01
	.dc.b	$EF,$5F,$8B,$C7,$C5,$E7,$E0,$75,$F1,$7B,$F8,$3E,$1F,$17,$C7,$E1
	.dc.b	$79,$7C,$5F,$3F,$87,$E9,$F1,$5F,$27,$87,$F9,$8F,$9D,$F9,$0E,$DC
	.dc.b	$C7,$E5,$38,$17,$CA,$31,$8F,$67,$59,$F8,$9E,$1A,$F9,$9E,$5B,$FA
	.dc.b	$F8,$3D,$B1,$F8,$8E,$2F,$D3,$F1,$FC,$38,$F9,$5C,$9F,$C3,$B9,$BE
	.dc.b	$4C,$7C,$C3,$E7,$78,$BF,$43,$83,$F4,$FD,$1F,$52,$7E,$97,$4F,$D8
	.dc.b	$7E,$A7,$57,$DC,$7E,$B7,$5F,$E0,$1F,$B1,$D9,$F8,$47,$ED,$76,$FE
	.dc.b	$21,$FB,$9D,$DF,$8C,$7F,$D8,$FE,$D7,$F8,$0B,$3F,$F5,$AF,$7F,$F3
	.dc.b	$EF,$03,$F2,$D5,$FA,$4F,$07,$F3,$4F,$E9,$BC,$2F,$CF,$3F,$A8,$F0
	.dc.b	$FF,$44,$FE,$AB,$C4,$FD,$33,$FA,$CF,$17,$F5,$4F,$EB,$BC,$6F,$E1
	.dc.b	$BC,$7F,$07,$C8,$F8,$BC,$BF,$1E,$BF,$A4,$0E,$7F,$99,$BF,$F0,$C7
	.dc.b	$E7,$C7,$D1,$3F,$73,$E3,$FD,$57,$EB,$9F,$81,$F1,$FF,$1C,$FE,$0F
	.dc.b	$C7,$FC,$73,$F8,$5F,$1F,$F1,$CF,$E1,$FC,$7F,$C7,$3F,$89,$F1,$FF
	.dc.b	$1C,$FF,$20,$EB,$FF,$09,$98,$CF,$F9,$2F,$8C,$AE,$35,$FD,$9A,$17
	.dc.b	$DA,$31,$87,$89,$D6,$3E,$67,$86,$5F,$A7,$7F,$35,$E0,$FE,$41,$FD
	.dc.b	$B7,$83,$F0,$9F,$C7,$9F,$C6,$B6,$F9,$5D,$3F,$C7,$E3,$F8,$EF,$07
	.dc.b	$CF,$7E,$6F,$17,$E3,$F7,$FE,$6F,$DF,$F8,$F7,$F9,$CF,$0F,$E3,$DF
	.dc.b	$E7,$7D,$FF,$8F,$7F,$9E,$F7,$FE,$3D,$FE,$7F,$DF,$F8,$F7,$FA,$0F
	.dc.b	$7F,$E3,$DF,$DA,$65,$FB,$6E,$7F,$D1,$F8,$F9,$7F,$4B,$5C,$5F,$D3
	.dc.b	$DF,$57,$F5,$38,$78,$5F,$D5,$E3,$E5,$7F,$5B,$97,$A5,$FD,$7E,$7E
	.dc.b	$D7,$F6,$3A,$7C,$1D,$BE,$26,$9F,$ED,$0F,$07,$F6,$BC,$1F,$8E,$BB
	.dc.b	$3F,$C7,$0F,$C7,$1F,$C9,$7A,$7F,$1C,$FD,$DF,$0F,$F8,$33,$F7,$FC
	.dc.b	$4F,$E1,$4F,$E0,$7C,$5F,$E1,$CF,$E0,$FC,$6F,$E2,$4F,$E1,$7C,$7F
	.dc.b	$E2,$EF,$F3,$1E,$4F,$E8,$66,$39,$7F,$1B,$3C,$1F,$C7,$D7,$47,$F2
	.dc.b	$37,$E0,$7F,$27,$87,$91,$FC,$AE,$3E,$87,$FB,$AE,$5F,$CD,$F2,$F7
	.dc.b	$1F,$8A,$27,$91,$FD,$A1,$FE,$31,$3F,$DF,$39,$FE,$54,$7F,$67,$3F
	.dc.b	$78,$7F,$4D,$3F,$80,$1F,$D4,$4F,$E0,$87,$F5,$53,$F8,$41,$FD,$64
	.dc.b	$FE,$18,$7F,$5D,$3F,$88,$1F,$D8,$4F,$EE,$BE,$37,$F9,$49,$FE,$33
	.dc.b	$D3,$6B,$FC,$4D,$E3,$FF,$B6,$DF,$E3,$72,$AF,$E3,$6F,$F5,$DF,$1F
	.dc.b	$F8,$DB,$FD,$7F,$C7,$FE,$36,$FF,$61,$F1,$FF,$8D,$BF,$D8,$FC,$7F
	.dc.b	$E3,$6F,$F6,$5F,$1F,$F8,$DB,$FD,$9F,$C7,$FE,$36,$FE,$B8,$87,$E3
	.dc.b	$4F,$E3,$7C,$FF,$8D,$69,$F4,$33,$FF,$13,$EB,$FB,$D7,$F8,$5F,$83
	.dc.b	$FC,$06,$7F,$81,$9F,$D5,$0F,$E3,$47,$F5,$83,$F8,$D1,$FD,$70,$FE
	.dc.b	$34,$7F,$61,$E6,$FC,$57,$47,$EC,$23,$F6,$47,$F2,$EE,$8C,$BF,$E7
	.dc.b	$7F,$1B,$FB,$3B,$C8,$FC,$C1,$D8,$FC,$E1,$E2,$3F,$40,$7F,$2F,$3F
	.dc.b	$98,$9F,$C8,$6F,$D9,$5F,$D8,$B1,$FE,$31,$8F,$F3,$BA,$FE,$30,$7F
	.dc.b	$83,$CF,$17,$57,$F8,$24,$8F,$F7,$BF,$8A,$FF,$06,$7C,$57,$F8,$37
	.dc.b	$DD,$FF,$07,$7C,$57,$F8,$3F,$E2,$BF,$C2,$1F,$01,$FE,$11,$F8,$AF
	.dc.b	$F0,$97,$C1,$7F,$84,$FE,$2B,$FD,$39,$AF,$E2,$F4,$FF,$4D,$AF,$84
	.dc.b	$CF,$C2,$EB,$E1,$B7,$F0,$FC,$3E,$23,$FD,$3F,$67,$38,$7F,$98,$A1
	.dc.b	$5C,$A3,$18,$78,$1F,$11,$FE,$15,$CB,$CC,$FF,$85,$F4,$F5,$3F,$E1
	.dc.b	$9D,$BD,$CF,$F8,$6F,$83,$FC,$59,$C3,$F1,$23,$FA,$77,$17,$F4,$0E
	.dc.b	$6F,$F4,$5C,$60,$FF,$11,$47,$07,$91,$FE,$25,$E3,$F9,$2F,$F8,$9F
	.dc.b	$8F,$E4,$BF,$E2,$9E,$3F,$92,$FF,$8A,$F8,$FD,$4F,$B0,$F8,$1C,$7F
	.dc.b	$1D,$F8,$5C,$7F,$1D,$FC,$7F,$92,$E3,$F9,$11,$A3,$E4,$71,$FC,$77
	.dc.b	$E9,$F1,$F6,$7F,$B5,$63,$E2,$7F,$B6,$65,$E6,$7F,$B7,$67,$EA,$7F
	.dc.b	$B8,$69,$F8,$67,$F1,$07,$C2,$21,$FB,$5C,$BC,$9F,$13,$F8,$42,$47
	.dc.b	$DB,$F8,$AF,$C3,$FC,$0F,$E2,$3D,$AF,$C4,$FC,$27,$E2,$BD,$CF,$C5
	.dc.b	$FC,$2F,$E3,$3D,$EF,$C6,$FC,$37,$E3,$BE,$03,$F1,$FF,$0F,$F9,$0F
	.dc.b	$82,$FC,$8F,$C4,$7E,$4B,$E1,$3F,$27,$F1,$5F,$E0,$BF,$17,$FD,$DF
	.dc.b	$F1,$0F,$4B,$FC,$6D,$1F,$B9,$E3,$FC,$BD,$71,$7F,$21,$87,$57,$F2
	.dc.b	$59,$78,$5F,$CA,$69,$E5,$7F,$2D,$B7,$A5,$FC,$C7,$07,$B5,$FC,$D7
	.dc.b	$17,$E9,$B8,$FE,$1C,$7E,$9B,$93,$F1,$DD,$FF,$91,$ED,$FF,$10,$BF
	.dc.b	$8D,$9F,$F1,$07,$85,$F9,$87,$FC,$45,$E1,$FE,$71,$FF,$12,$78,$9F
	.dc.b	$A0,$7F,$C4,$DE,$2F,$E9,$1F,$D9,$9F,$B6,$F0,$3F,$09,$8F,$FB,$48
	.dc.b	$FD,$6C,$7E,$24,$FD,$CF,$8D,$FB,$43,$F6,$6F,$EF,$3C,$2E,$3C,$BF
	.dc.b	$E9,$AA,$FD,$F4,$FE,$DC,$F0,$3F,$E3,$AC,$3C,$8F,$F8,$F3,$1F,$43
	.dc.b	$FE,$3E,$CB,$D8,$FF,$90,$33,$F8,$0F,$C2,$BF,$C6,$FA,$9F,$ED,$0F
	.dc.b	$1B,$F6,$73,$99,$FD,$A1,$FE,$45,$F1,$FF,$2C,$FE,$60,$FE,$00,$7E
	.dc.b	$CC,$FE,$08,$7E,$CC,$FE,$10,$7E,$CC,$F5,$1F,$B3,$3F,$8D,$F8,$FF
	.dc.b	$B3,$7F,$8E,$F8,$FF,$B3,$7F,$CB,$40,$A7,$CA,$3E,$65,$7C,$FD,$BE
	.dc.b	$4C,$3C,$3F,$DD,$70,$FE,$D6,$BF,$CC,$3E,$3F,$ED,$5F,$F3,$1F,$8F
	.dc.b	$FB,$57,$FC,$CB,$E3,$FE,$D5,$FF,$33,$F8,$FF,$01,$FF,$34,$EB,$F8
	.dc.b	$BE,$4F,$F6,$7E,$FF,$C6,$B8,$3E,$57,$36,$76,$FE,$8A,$7E,$F0,$FC
	.dc.b	$79,$FC,$6C,$F9,$27,$F0,$43,$E4,$9F,$C2,$0F,$92,$7F,$0C,$3E,$49
	.dc.b	$FC,$40,$FE,$36,$7F,$14,$3E,$1D,$7E,$27,$9B,$F1,$D1,$83,$E2,$63
	.dc.b	$F9,$47,$E8,$9F,$4D,$FC,$C3,$F8,$E9,$F6,$F3,$7E,$39,$F7,$F3,$7E
	.dc.b	$39,$F8,$3C,$DF,$8E,$7E,$1F,$37,$E3,$9F,$8B,$97,$F4,$1C,$CF,$E7
	.dc.b	$79,$FF,$6F,$9F,$E9,$DF,$DD,$67,$FB,$C3,$FA,$B7,$8F,$FC,$69,$FE
	.dc.b	$B1,$E3,$FF,$1A,$7F,$AD,$78,$FF,$C6,$9F,$EB,$9E,$3F,$F1,$A7,$FA
	.dc.b	$F7,$8F,$FC,$69,$FE,$C1,$E3,$FF,$1A,$7F,$19,$97,$FA,$01,$FE,$81
	.dc.b	$AF,$F1,$CF,$1B,$F8,$DC,$7F,$84,$4F,$ED,$1E,$37,$F1,$B7,$FB,$9F
	.dc.b	$8D,$FC,$6D,$FE,$EB,$E3,$7F,$1B,$7F,$BB,$F8,$DF,$C6,$DF,$EF,$3E
	.dc.b	$37,$F1,$B7,$FB,$DF,$8D,$FC,$6D,$FE,$FB,$E3,$7C,$27,$F8,$D1,$FD
	.dc.b	$DF,$8F,$F8,$DB,$FD,$13,$5F,$DB,$1F,$94,$7F,$B3,$9F,$B5,$3F,$8D
	.dc.b	$9F,$B5,$3F,$8D,$9F,$B5,$3F,$8D,$9F,$B5,$3F,$8D,$9F,$83,$D7,$F8
	.dc.b	$97,$8B,$FB,$59,$FE,$39,$E2,$FE,$D5,$4F,$F4,$2F,$88,$7B,$BF,$ED
	.dc.b	$0C,$3F,$C1,$DD,$DF,$C7,$AF,$F6,$D1,$F9,$83,$F3,$47,$80,$FC,$F1
	.dc.b	$E4,$3F,$44,$7A,$0F,$D3,$1E,$C3,$DC,$FF,$1D,$7F,$5C,$FC,$4B,$FC
	.dc.b	$8C,$7F,$8D,$FC,$6F,$EF,$9F,$11,$FE,$2C,$24,$7F,$88,$DA,$FF,$8B
	.dc.b	$9B,$5F,$CE,$DB,$5F,$CD,$5B,$3F,$AC,$6D,$76,$8E,$84,$3A,$1F,$10
	.dc.b	$FF,$07,$D7,$F8,$39,$C9,$1F,$03,$C5,$AE,$6B,$C4,$E5,$AE,$4F,$21
	.dc.b	$92,$99,$9F,$33,$E0,$EC,$FD,$AB,$AA,$EE,$4A,$EA,$D8,$D4,$60,$8E
	.dc.b	$09,$5D,$A7,$83,$67,$6D,$9D,$CA,$FF,$D5,$15,$FD,$06,$9F,$C0,$1F
	.dc.b	$90,$BF,$F1,$BB,$5F,$65,$F1,$7E,$04,$F5,$24,$66,$37,$7F,$62,$30
	.dc.b	$79,$8B,$7E,$66,$CF,$9A,$7F,$D7,$D8,$7C,$8B,$F2,$F9,$DE,$8A,$FE
	.dc.b	$A5,$8A,$31,$3A,$9F,$E2,$38,$FB,$6B,$FD,$B2,$DF,$7B,$70,$F4,$AF
	.dc.b	$BF,$8F,$53,$3B,$EB,$D5,$DF,$C0,$70,$F5,$F8,$FC,$0F,$2F,$67,$9F
	.dc.b	$C1,$74,$F6,$FA,$FC,$1F,$EE,$F8,$3F,$90,$7B,$3E,$9F,$67,$FC,$78
	.dc.b	$E7,$8F,$D7,$5C,$5F,$49,$EE,$C3,$B3,$8E,$4F,$AF,$9F,$8E,$31,$C6
	.dc.b	$DB,$D1,$18,$E0,$F6,$3A,$E2,$F8,$3C,$7F,$0B,$93,$E1,$FB,$DF,$13
	.dc.b	$9A,$5F,$EF,$7A,$7E,$AE,$BF,$7B,$A7,$EE,$12,$7F,$07,$3A,$24,$66
	.dc.b	$3F,$06,$69,$FA,$3C,$7F,$C6,$E3,$FC,$23,$3F,$91,$AF,$C9,$CF,$18
	.dc.b	$FC,$45,$58,$FC,$43,$6B,$F1,$07,$52,$1F,$E7,$D9,$FC,$A3,$FE,$1F
	.dc.b	$3F,$6E,$5A,$FD,$BB,$63,$6F,$F0,$03,$13,$FE,$C7,$70,$43,$99,$67
	.dc.b	$9B,$67,$9B,$A2,$46,$9B,$B7,$EB,$E7,$F6,$5C,$6E,$BF,$E8,$35,$F8
	.dc.b	$48,$C6,$FF,$2A,$3F,$91,$1E,$63,$CD,$87,$52,$5F,$89,$B8,$FA,$E7
	.dc.b	$C7,$39,$24,$66,$77,$3D,$27,$2D,$9F,$99,$A3,$9F,$E0,$47,$1A,$DD
	.dc.b	$7F,$BB,$EF,$9B,$63,$2F,$A4,$C4,$F1,$23,$DC,$86,$8A,$71,$3E,$E6
	.dc.b	$7F,$2F,$4F,$ED,$59,$7E,$3E,$FF,$1D,$39,$AF,$83,$7D,$9E,$84,$3A
	.dc.b	$1E,$C3,$B9,$4D,$14,$D9,$FD,$13,$2F,$88,$F6,$38,$94,$E2,$53,$8E
	.dc.b	$3F,$19,$B7,$F1,$9E,$C7,$E3,$7A,$FF,$86,$78,$DF,$E2,$CA,$FF,$77
	.dc.b	$47,$F7,$09,$E3,$87,$B5,$A7,$8A,$7B,$3D,$0F,$11,$5F,$EB,$78,$7A
	.dc.b	$9D,$C6,$88,$E0,$98,$79,$E3,$2D,$8F,$CA,$63,$F4,$75,$FC,$DD,$7E
	.dc.b	$77,$6F,$CF,$BF,$E4,$83,$F9,$FE,$BE,$B7,$C9,$87,$32,$65,$ED,$8C
	.dc.b	$B5,$F0,$BF,$96,$DB,$F8,$4C,$7F,$06,$9F,$E1,$AD,$AC,$BF,$C6,$09
	.dc.b	$1A,$7F,$BF,$63,$E9,$C3,$CF,$3F,$B8,$6D,$7F,$51,$3F,$8B,$65,$FC
	.dc.b	$A6,$FB,$1F,$C6,$0E,$B1,$E0,$AE,$C7,$7E,$CF,$EC,$58,$3D,$EF,$EB
	.dc.b	$24,$BF,$E8,$3E,$F7,$F5,$CF,$89,$FE,$BB,$D7,$FD,$7B,$E0,$3F,$B2
	.dc.b	$FB,$E2,$FC,$0E,$3E,$17,$3F,$07,$AF,$86,$DF,$C2,$F0,$F8,$7E,$3F
	.dc.b	$0F,$CB,$E2,$39,$FC,$47,$F1,$1E,$0F,$E7,$3C,$BF,$AE,$7F,$A9,$BA
	.dc.b	$7F,$AC,$6A,$AE,$71,$8C,$3B,$BA,$CB,$C6,$F0,$D3,$CE,$F2,$BF,$63
	.dc.b	$FD,$83,$7F,$80,$FC,$23,$FB,$1F,$17,$C5,$7F,$A1,$74,$FD,$2E,$6F
	.dc.b	$EA,$B1,$F4,$B9,$FE,$A3,$F4,$8F,$A9,$1F,$4B,$5F,$A8,$7F,$B0,$63
	.dc.b	$F9,$97,$6F,$EB,$E7,$FC,$9C,$6E,$38,$CF,$06,$92,$BB,$CE,$DF,$0F
	.dc.b	$47,$DE,$C7,$D1,$3F,$C5,$9F,$F0,$BF,$3F,$BA,$FF,$49,$7F,$E3,$8C
	.dc.b	$FE,$1C,$63,$4F,$BC,$FD,$0E,$5F,$BC,$FD,$2E,$5F,$BC,$FD,$4E,$5F
	.dc.b	$BC,$FD,$6E,$5F,$BC,$FD,$8E,$5F,$BD,$AF,$E7,$7B,$BF,$95,$70,$7E
	.dc.b	$76,$B4,$7D,$EE,$6F,$BA,$F8,$F9,$BE,$EB,$E5,$E6,$FB,$AF,$9F,$9B
	.dc.b	$EE,$BE,$9D,$FD,$79,$7F,$36,$E0,$F8,$0F,$F3,$8E,$2F,$8B,$9F,$F8
	.dc.b	$F7,$8F,$FC,$8F,$ED,$FF,$96,$F0,$7C,$4F,$E8,$09,$97,$E4,$A7,$F4
	.dc.b	$7F,$05,$FA,$4F,$7B,$F4,$BF,$07,$FA,$6F,$80,$FD,$3F,$C2,$7E,$A3
	.dc.b	$E0,$BF,$53,$F0,$BF,$AA,$F8,$4F,$D5,$FC,$37,$F9,$43,$F8,$07,$FC
	.dc.b	$19,$AF,$D4,$5F,$11,$9F,$67,$5F,$13,$F2,$B9,$3F,$4D,$7F,$A8,$3F
	.dc.b	$8E,$9F,$E3,$DB,$4F,$18,$C5,$F7,$7C,$27,$EB,$31,$F2,$BF,$AE,$CF
	.dc.b	$D2,$FE,$C3,$5F,$6B,$FB,$2D,$FE,$0B,$F0,$CF,$DA,$F1,$7C,$65,$E1
	.dc.b	$FF,$9C,$73,$7B,$5E,$F2,$4F,$9F,$CE,$F8,$5E,$DF,$C3,$F8,$AF,$89
	.dc.b	$EE,$FC,$5F,$89,$F8,$DE,$FF,$C7,$F8,$9F,$91,$F0,$3F,$27,$E2,$7E
	.dc.b	$57,$C1,$FC,$BF,$89,$F9,$9F,$0B,$F3,$7E,$27,$E7,$7C,$3F,$CF,$F8
	.dc.b	$9F,$C8,$F0,$BD,$BF,$E9,$2F,$D1,$8D,$BF,$2E,$B9,$BF,$37,$0E,$EF
	.dc.b	$CF,$CB,$C6,$FD,$1D,$3C,$EF,$D3,$DB,$D6,$FD,$5E,$0F,$7B,$F0,$4F
	.dc.b	$B1,$C7,$F1,$1F,$93,$E2,$7E,$4F,$9F,$E6,$79,$1F,$E5,$DF,$1B,$F2
	.dc.b	$B5,$F4,$4F,$DC,$78,$DF,$93,$7F,$93,$F8,$DF,$93,$7F,$94,$F8,$DF
	.dc.b	$93,$7F,$95,$F8,$DF,$93,$7F,$96,$F8,$DF,$85,$7F,$0E,$7E,$2F,$C6
	.dc.b	$FC,$8B,$FC,$D0,$FE,$33,$E4,$FF,$82,$0F,$CA,$2F,$51,$E0,$7E,$4A
	.dc.b	$39,$1D,$0F,$CF,$1F,$92,$3F,$88,$78,$BF,$92,$7F,$89,$78,$BF,$92
	.dc.b	$7F,$8A,$78,$BF,$92,$7F,$8B,$78,$BF,$92,$7F,$8C,$71,$FC,$6F,$23
	.dc.b	$E4,$40,$F9,$67,$97,$1F,$C9,$BF,$40,$87,$EC,$7C,$5F,$AA,$FE,$CF
	.dc.b	$9F,$EC,$BF,$B5,$E8,$FB,$AF,$ED,$FA,$7F,$02,$FE,$E7,$AB,$F0,$AF
	.dc.b	$EE,$FA,$FF,$12,$FE,$F7,$B3,$F1,$AF,$E3,$CF,$F0,$C7,$9F,$F9,$4E
	.dc.b	$2F,$93,$DD,$AE,$CF,$F4,$E6,$CC,$39,$C6,$31,$EE,$EB,$2F,$1B,$C3
	.dc.b	$3F,$3B,$CB,$4F,$5B,$D3,$5F,$7B,$DB,$6F,$C9,$3E,$0E,$EF,$E4,$BB
	.dc.b	$FD,$DE,$DF,$C8,$BF,$D1,$FB,$BE,$83,$9E,$3F,$A9,$7F,$97,$E4,$FB
	.dc.b	$0F,$E6,$79,$7E,$E3,$F9,$BE,$6F,$C0,$3F,$9D,$E7,$FC,$23,$F0,$30
	.dc.b	$FE,$4D,$E4,$7F,$1B,$7F,$93,$79,$3F,$90,$F0,$7F,$73,$5F,$D3,$67
	.dc.b	$F2,$C6,$8F,$C8,$9F,$CB,$6B,$F3,$CF,$F5,$FE,$DF,$D1,$3F,$D8,$7B
	.dc.b	$BF,$4C,$FF,$63,$EF,$FD,$53,$FD,$97,$C0,$FD,$73,$FB,$03,$F6,$58
	.dc.b	$7E,$D7,$83,$F9,$17,$93,$FC,$87,$8F,$E5,$F7,$FE,$4F,$BB,$F4,$B1
	.dc.b	$FD,$07,$9B,$E8,$CE,$7B,$5D,$1F,$5A,$31,$D3,$F6,$9D,$75,$7D,$E7
	.dc.b	$87,$5F,$E0,$9E,$5D,$9F,$86,$7A,$76,$FE,$2D,$ED,$DD,$FC,$8B,$9F
	.dc.b	$E4,$C6,$38,$BF,$A5,$78,$7F,$CA,$5F,$E9,$9E,$27,$F2,$D7,$FA,$77
	.dc.b	$8B,$FC,$C5,$FE,$A1,$E3,$7F,$35,$7F,$A9,$78,$FF,$CE,$5F,$EA,$9E
	.dc.b	$47,$F3,$D7,$FA,$B7,$93,$F0,$5F,$D8,$F4,$FE,$3D,$FD,$C7,$43,$C4
	.dc.b	$FD,$DC,$7E,$3C,$FE,$9E,$7D,$0D,$7F,$BC,$79,$3F,$52,$FF,$93,$76
	.dc.b	$7D,$87,$F9,$47,$6F,$DC,$7F,$95,$77,$7E,$01,$FE,$59,$DF,$F8,$47
	.dc.b	$F5,$FE,$07,$E3,$DF,$D8,$F8,$3F,$8F,$F5,$7F,$D4,$DE,$CF,$FB,$07
	.dc.b	$C2,$FE,$3F,$3B,$D7,$ED,$61,$CD,$FF,$06,$63,$DD,$FF,$07,$65,$E3
	.dc.b	$7F,$C2,$19,$F9,$DF,$F0,$96,$9E,$B7,$DB,$B7,$BC,$FD,$97,$87,$F8
	.dc.b	$EE,$0F,$C4,$78,$7F,$8E,$9F,$E3,$CF,$CD,$7F,$C1,$9C,$7F,$45,$FF
	.dc.b	$07,$72,$7D,$57,$FC,$21,$CB,$F6,$5F,$F0,$97,$37,$DD,$7F,$C2,$9C
	.dc.b	$FF,$81,$7F,$C2,$DD,$1F,$85,$7F,$C3,$1D,$3F,$DB,$5F,$E3,$64,$3F
	.dc.b	$8F,$1A,$3F,$B6,$1F,$8C,$3F,$B6,$1F,$8C,$3F,$B6,$1F,$8C,$3F,$B6
	.dc.b	$1F,$8C,$3F,$B6,$1F,$8C,$3F,$B6,$1F,$8C,$3F,$B6,$1F,$E2,$BC,$BF
	.dc.b	$B6,$74,$FF,$84,$FE,$27,$FC,$60,$49,$FF,$14,$7B,$1F,$D1,$DF,$F3
	.dc.b	$2F,$F0,$0F,$DD,$1E,$FE,$7B,$1C,$7F,$BB,$97,$BF,$98,$C7,$C0,$5C
	.dc.b	$3F,$B2,$53,$E0,$71,$DA,$87,$C1,$61,$7C,$55,$F4,$B1,$F1,$59,$F8
	.dc.b	$5D,$7C,$2E,$FE,$1B,$87,$C5,$7D,$1D,$7F,$C8,$9E,$4F,$EF,$1F,$EC
	.dc.b	$1E,$56,$9F,$EC,$9C,$7E,$A7,$CB,$7D,$1C,$F1,$7B,$72,$F0,$C6,$0F
	.dc.b	$5E,$9E,$7D,$7D,$1B,$FA,$78,$7D,$5C,$5E,$BE,$3F,$67,$27,$B7,$3F
	.dc.b	$F3,$07,$2F,$D1,$CF,$FC,$C1,$CD,$FD,$3B,$9B,$FC,$12,$E7,$83,$FB
	.dc.b	$0B,$C2,$7C,$53,$F2,$6F,$AC,$63,$9B,$E9,$B7,$3F,$23,$9D,$7D,$0E
	.dc.b	$F7,$F6,$BF,$80,$E0,$F8,$4F,$E0,$B8,$7E,$43,$F3,$3E,$1F,$FD,$EA
	.dc.b	$FF,$92,$5B,$C9,$F3,$F4,$FB,$9C,$BF,$56,$BC,$9C,$DF,$65,$F3,$73
	.dc.b	$FD,$D7,$D1,$CD,$F2,$9F,$57,$37,$CA,$7D,$9C,$DE,$E7,$F2,$BC,$5F
	.dc.b	$0D,$F9,$27,$FA,$07,$8F,$E5,$71,$BE,$23,$F1,$84,$9F,$8D,$F1,$5F
	.dc.b	$8F,$F8,$AF,$C8,$7B,$DF,$91,$F7,$FF,$25,$F1,$5F,$93,$F7,$FF,$29
	.dc.b	$F1,$5F,$98,$FE,$01,$FF,$62,$E5,$FE,$CA,$5F,$17,$9F,$84,$D7,$C5
	.dc.b	$EF,$E1,$B8,$7C,$5F,$1F,$88,$FC,$B7,$87,$F9,$1E,$D7,$B2,$DB,$FC
	.dc.b	$99,$AF,$FA,$6F,$C7,$E0,$FE,$03,$66,$5D,$63,$1A,$78,$5D,$6D,$E7
	.dc.b	$F8,$7F,$4C,$7E,$17,$1F,$6B,$F8,$7E,$2F,$82,$FE,$27,$93,$F6,$5C
	.dc.b	$BF,$1E,$3F,$81,$F3,$7E,$57,$E0,$9D,$3F,$95,$EA,$FE,$F0,$FD,$3E
	.dc.b	$3F,$D6,$78,$3F,$95,$C3,$F2,$9E,$4F,$E8,$DF,$CA,$1F,$78,$FC,$A1
	.dc.b	$F8,$23,$F2,$87,$E1,$8F,$CE,$78,$BF,$94,$AE,$BD,$AF,$17,$F2,$B1
	.dc.b	$FE,$21,$AF,$CB,$78,$DF,$E4,$BA,$78,$BF,$93,$9F,$E3,$BE,$2F,$E4
	.dc.b	$DF,$E3,$FE,$2F,$E4,$DF,$E4,$3E,$2F,$E4,$DF,$E4,$7E,$2F,$E4,$DF
	.dc.b	$E4,$BE,$2F,$E4,$F3,$F1,$F0,$FC,$13,$E1,$CF,$E4,$CF,$EF,$7E,$9E
	.dc.b	$1F,$CB,$92,$7F,$CB,$BE,$57,$F0,$58,$F8,$E7,$E8,$0F,$8E,$78,$8F
	.dc.b	$8E,$79,$8F,$8E,$7A,$8F,$8E,$7E,$C0,$FE,$09,$BF,$C6,$BF,$F0,$F4
	.dc.b	$7C,$63,$F1,$C7,$F0,$9D,$7E,$35,$FC,$ED,$7E,$33,$F4,$35,$F8,$CF
	.dc.b	$D2,$D7,$E3,$3F,$53,$5F,$8C,$FD,$6D,$7E,$36,$5F,$D1,$CF,$8D,$97
	.dc.b	$FA,$4F,$8F,$F9,$17,$97,$FD,$56,$3E,$41,$F1,$67,$FC,$89,$B7,$C5
	.dc.b	$8F,$EA,$BC,$7F,$8B,$7F,$AB,$F9,$7F,$C7,$9F,$EB,$3E,$5F,$F1,$E7
	.dc.b	$FA,$DF,$27,$F1,$ED,$3F,$13,$CB,$F8,$BC,$BF,$AD,$71,$FC,$5B,$FE
	.dc.b	$39,$7F,$E4,$5F,$2F,$F8,$E9,$FC,$83,$9F,$F8,$EC,$FC,$C3,$E7,$1F
	.dc.b	$CA,$4F,$CC,$1E,$03,$F3,$87,$90,$FD,$01,$E8,$3E,$F1,$EC,$3F,$9D
	.dc.b	$5F,$C6,$8F,$E8,$71,$F1,$36,$FF,$22,$3F,$DB,$3A,$7F,$6D,$9F,$E9
	.dc.b	$1F,$DC,$C7,$EE,$CF,$E8,$9E,$5F,$F1,$D7,$FA,$37,$97,$FC,$75,$FE
	.dc.b	$91,$E5,$FF,$07,$7F,$84,$9F,$C3,$FC,$BF,$E3,$6F,$F1,$1F,$2F,$F8
	.dc.b	$DB,$FC,$4F,$CB,$FE,$37,$AF,$FB,$3F,$CB,$FE,$37,$E2,$7F,$1A,$AF
	.dc.b	$F1,$0E,$BF,$C1,$3C,$96,$1C,$A7,$F9,$56,$3D,$9F,$E5,$99,$7F,$2F
	.dc.b	$F1,$FC,$91,$FE,$14,$D3,$D0,$FF,$85,$B5,$F6,$3F,$E1,$8D,$BF,$9E
	.dc.b	$3F,$CF,$8F,$EC,$9B,$FF,$1B,$E0,$FC,$74,$7E,$34,$FC,$91,$FE,$22
	.dc.b	$3E,$89,$FB,$23,$F7,$E7,$EC,$B5,$FC,$F3,$FB,$27,$F4,$4F,$EC,$9F
	.dc.b	$D3,$3F,$B2,$3F,$87,$9F,$87,$3F,$C5,$BE,$07,$E3,$79,$24,$7E,$C4
	.dc.b	$D4,$FE,$C5,$E4,$F4,$7F,$18,$7F,$01,$3F,$18,$7F,$05,$3F,$18,$7A
	.dc.b	$0F,$C6,$1F,$DA,$0F,$71,$F0,$78,$FF,$1A,$AF,$F1,$BC,$5F,$E2,$92
	.dc.b	$72,$7E,$8B,$0F,$3D,$7F,$09,$26,$1F,$CC,$49,$C9,$F8,$3A,$E1,$3F
	.dc.b	$E6,$39,$CB,$FA,$E3,$13,$BA,$EB,$3A,$BF,$0C,$62,$75,$5C,$0E,$47
	.dc.b	$89,$CB,$4C,$D7,$3B,$57,$FB,$F2,$FA,$D7,$36,$BC,$5B,$5C,$67,$C5
	.dc.b	$3D,$2F,$BC,$E6,$39,$5F,$C0,$79,$36,$35,$18,$AE,$77,$DA,$F9,$92
	.dc.b	$7B,$1E,$EC,$3F,$80,$E2,$E3,$7C,$5F,$F8,$54,$95,$FE,$07,$F8,$AF
	.dc.b	$F0,$C7,$C5,$FF,$86,$7D,$7F,$F0,$D7,$8B,$FE,$1B,$F7,$FF,$C3,$9F
	.dc.b	$15,$FE,$1D,$F6,$BF,$C3,$DF,$15,$FE,$1F,$F7,$3F,$C4,$1F,$15,$FE
	.dc.b	$21,$F7,$BF,$C4,$5F,$15,$FE,$23,$F8,$0F,$F3,$DF,$27,$FA,$27,$83
	.dc.b	$F7,$EF,$F0,$3F,$63,$FC,$39,$1F,$E6,$9E,$6F,$E0,$AF,$F5,$01,$D1
	.dc.b	$FE,$08,$AE,$4F,$F0,$7B,$EF,$3F,$E2,$1C,$7C,$6F,$F8,$8F,$3F,$3B
	.dc.b	$FE,$25,$D7,$D6,$FF,$89,$F7,$F7,$BF,$E2,$9E,$1F,$86,$FC,$53,$FC
	.dc.b	$59,$C9,$F2,$1F,$D1,$71,$7D,$1D,$7E,$A4,$7D,$26,$93,$A7,$EE,$6D
	.dc.b	$B6,$CE,$DB,$3B,$E9,$FB,$F1,$82,$38,$24,$FE,$29,$AF,$90,$FC,$F9
	.dc.b	$FE,$B7,$94,$FE,$72,$3E,$91,$A3,$F4,$66,$DD,$3F,$5E,$38,$CF,$08
	.dc.b	$E0,$48,$CB,$D1,$AF,$19,$E6,$F5,$9E,$53,$A8,$C1,$C6,$76,$F9,$08
	.dc.b	$77,$72,$D9,$E7,$18,$9E,$93,$C1,$AE,$CE,$D3,$C0,$8E,$08,$7A,$5C
	.dc.b	$B4,$C9,$E8,$3F,$DB,$BE,$7F,$7A,$F1,$12,$78,$CE,$C8,$6D,$CB,$4E
	.dc.b	$A5,$3C,$04,$38,$9E,$B3,$06,$CC,$D7,$C0,$24,$F8,$E3,$A9,$C4,$DB
	.dc.b	$97,$99,$D7,$CC,$5F,$58,$87,$12,$D7,$6B,$F5,$C6,$49,$19,$7C,$2D
	.dc.b	$7B,$B5,$F1,$C7,$28,$C4,$76,$BF,$33,$A3,$9B,$E9,$BF,$3C,$78,$23
	.dc.b	$6F,$1A,$E9,$18,$8D,$46,$08,$F4,$6B,$E6,$8E,$15,$E4,$8E,$26,$C8
	.dc.b	$6E,$78,$D7,$92,$7F,$7D,$3F,$B9,$21,$FC,$FC,$BC,$3F,$D0,$89,$8F
	.dc.b	$F0,$9B,$FD,$FC,$62,$7F,$85,$46,$27,$51,$C9,$AF,$53,$F8,$37,$52
	.dc.b	$1F,$03,$FC,$D8,$92,$3F,$98,$C5,$9E,$1F,$17,$FC,$F3,$E0,$FF,$9E
	.dc.b	$FB,$9F,$CF,$BE,$2F,$F9,$FF,$C2,$7F,$40,$F8,$BF,$E8,$3F,$0B,$FD
	.dc.b	$0B,$E2,$FF,$A1,$FC,$37,$F4,$4F,$8B,$FE,$8B,$F0,$FF,$D1,$BE,$2F
	.dc.b	$FA,$3F,$C4,$7F,$48,$F8,$BF,$E9,$3F,$13,$FD,$2B,$E2,$E7,$27,$97
	.dc.b	$9F,$FA,$24,$EE,$31,$CB,$E7,$18,$75,$8C,$8E,$5F,$85,$9F,$8E,$33
	.dc.b	$1F,$17,$5F,$44,$62,$DB,$FB,$63,$1C,$3F,$05,$D7,$1F,$C5,$78,$72
	.dc.b	$FC,$C7,$E5,$9E,$1B,$FA,$2F,$FB,$08,$FB,$67,$D9,$29,$F6,$75,$FC
	.dc.b	$07,$47,$E2,$A3,$13,$F7,$A7,$3D,$5F,$D2,$27,$95,$6E,$73,$D3,$F8
	.dc.b	$79,$EB,$E6,$7F,$82,$30,$F1,$C7,$9B,$6F,$C3,$75,$7A,$70,$FE,$E5
	.dc.b	$9F,$B6,$B5,$6D,$3E,$14,$63,$5F,$8C,$EB,$6F,$96,$F0,$DF,$E8,$3C
	.dc.b	$B8,$3E,$B3,$D3,$87,$EE,$3D,$B8,$BF,$02,$F8,$38,$FF,$0A,$FD,$5E
	.dc.b	$4F,$C4,$BF,$5F,$97,$F1,$AF,$D9,$E6,$FC,$8B,$F6,$F9,$FF,$2A,$FD
	.dc.b	$DE,$8F,$CC,$BF,$7F,$A7,$F3,$AF,$E0,$7A,$BF,$C3,$7C,$7F,$0F,$8F
	.dc.b	$F1,$DC,$BF,$A7,$AF,$D6,$C3,$E2,$BF,$C4,$24,$8F,$C5,$7C,$57,$F8
	.dc.b	$93,$DF,$FF,$12,$FB,$FC,$7A,$7F,$69,$5F,$95,$ED,$AF,$D3,$5E,$FE
	.dc.b	$F2,$FD,$A4,$3C,$0C,$57,$1F,$7E,$F7,$29,$F0,$18,$5F,$01,$7E,$07
	.dc.b	$1E,$1E,$7E,$0B,$5F,$05,$BF,$1F,$E7,$E9,$FE,$69,$CF,$F8,$57,$7E
	.dc.b	$63,$F5,$39,$AF,$94,$63,$1E,$CE,$B3,$F1,$69,$E3,$D7,$C9,$B7,$97
	.dc.b	$83,$CD,$C1,$F1,$B0,$FD,$97,$1F,$F9,$33,$93,$E6,$47,$E4,$B9,$7E
	.dc.b	$97,$8D,$FE,$62,$DD,$F0,$5E,$F2,$57,$9B,$D9,$F8,$5E,$D7,$C3,$F8
	.dc.b	$5F,$89,$F0,$DF,$17,$DD,$F8,$DF,$0F,$F1,$FD,$FF,$91,$F0,$BF,$27
	.dc.b	$E0,$7E,$6F,$5B,$E2,$2F,$C1,$E3,$E1,$33,$F1,$3A,$F8,$6F,$93,$E2
	.dc.b	$7F,$96,$FC,$FF,$CD,$BF,$C6,$B8,$FF,$88,$E4,$AE,$31,$8C,$3A,$BA
	.dc.b	$CB,$C4,$FD,$0D,$3C,$CF,$D2,$DB,$D4,$FD,$4E,$0F,$73,$F5,$B8,$BF
	.dc.b	$05,$C7,$F1,$23,$F1,$DC,$9F,$2B,$D7,$FE,$55,$D1,$F3,$B7,$FF,$39
	.dc.b	$3F,$46,$BE,$99,$FC,$97,$C7,$FA,$EF,$E0,$BB,$7E,$DB,$F8,$4E,$FF
	.dc.b	$BE,$FE,$04,$FC,$3F,$8F,$F9,$07,$F1,$3E,$3F,$E4,$1F,$C5,$F8,$FF
	.dc.b	$C5,$FB,$7F,$C9,$B9,$FE,$9B,$9B,$F3,$93,$BA,$FD,$35,$F3,$7F,$51
	.dc.b	$87,$E7,$FC,$4F,$0C,$7E,$AF,$2F,$2B,$FA,$DC,$FD,$2F,$F1,$CD,$3F
	.dc.b	$6C,$FE,$E0,$FE,$85,$AF,$E4,$3B,$FF,$1D,$C1,$F2,$76,$F2,$6F,$F9
	.dc.b	$0B,$FC,$71,$F3,$CF,$D2,$F7,$FE,$39,$FD,$3F,$7F,$E3,$9F,$D4,$F7
	.dc.b	$FE,$39,$FD,$5F,$7F,$E3,$BA,$BF,$03,$1F,$CD,$79,$1D,$7F,$86,$8C
	.dc.b	$76,$7E,$3B,$C2,$FE,$C5,$DD,$F8,$E7,$FB,$27,$8F,$F9,$3A,$FC,$7F
	.dc.b	$81,$F9,$77,$F2,$3E,$0F,$77,$C2,$7E,$34,$F2,$9F,$8D,$3D,$27,$E3
	.dc.b	$4F,$69,$EF,$3F,$8C,$76,$FC,$5B,$FE,$2F,$C7,$F8,$D9,$FE,$C5,$CB
	.dc.b	$F8,$D9,$FC,$7C,$7C,$F3,$FA,$6F,$87,$FC,$05,$FE,$45,$1F,$C1,$5F
	.dc.b	$E1,$FE,$27,$F0,$97,$F8,$8F,$8B,$F7,$DF,$C6,$9F,$C4,$4F,$C6,$9F
	.dc.b	$C5,$BC,$7F,$E3,$0F,$F4,$D3,$FB,$87,$0F,$F1,$C7,$F2,$26,$CE,$27
	.dc.b	$F3,$3F,$1F,$F8,$DB,$FC,$D7,$BB,$F1,$BD,$5F,$82,$EE,$FC,$6B,$F8
	.dc.b	$4E,$EF,$C6,$BF,$ED,$BE,$EF,$C6,$DF,$F6,$AF,$1F,$E1,$3F,$10,$FF
	.dc.b	$02,$65,$FC,$E7,$87,$E5,$47,$F4,$57,$F7,$2F,$CF,$BF,$C8,$F6,$FE
	.dc.b	$36,$3F,$A4,$76,$FE,$35,$FE,$95,$DB,$F8,$D7,$FA,$67,$6F,$DF,$7F
	.dc.b	$19,$CF,$F8,$37,$F1,$DD,$1F,$87,$7E,$31,$FD,$C4,$FF,$12,$75,$7E
	.dc.b	$3D,$FF,$16,$69,$FC,$65,$FC,$69,$FC,$64,$FC,$69,$FC,$64,$FC,$69
	.dc.b	$FD,$77,$C7,$FE,$32,$FF,$5F,$F1,$FF,$8C,$BF,$D8,$7C,$7F,$E3,$2F
	.dc.b	$F6,$3F,$1F,$E0,$BF,$C6,$0F,$EA,$3C,$1F,$8D,$C3,$FC,$BD,$D9,$F3
	.dc.b	$1B,$3F,$8D,$E1,$FE,$15,$D6,$EC,$FC,$74,$7E,$EA,$3F,$1D,$3F,$DC
	.dc.b	$7B,$3F,$1C,$FF,$73,$EC,$FC,$73,$FC,$38,$FC,$71,$FC,$48,$FC,$41
	.dc.b	$FD,$E7,$C6,$FE,$35,$EB,$FF,$BB,$B0,$FE,$0F,$1F,$C6,$5F,$E1,$3E
	.dc.b	$3F,$F1,$97,$F8,$5F,$8F,$D9,$FE,$1B,$87,$89,$FE,$1F,$8F,$99,$F4
	.dc.b	$63,$FC,$CC,$FE,$6A,$7E,$34,$FE,$32,$43,$FD,$07,$E3,$7F,$17,$7E
	.dc.b	$51,$23,$F8,$DF,$85,$FD,$9A,$3F,$7A,$7D,$33,$F8,$09,$F5,$CF,$E0
	.dc.b	$A7,$DB,$3E,$EC,$7F,$37,$DB,$FB,$34,$7E,$08,$FE,$35,$E2,$FF,$19
	.dc.b	$9D,$78,$BF,$8B,$AF,$F4,$B7,$3F,$F1,$9B,$FF,$34,$78,$BF,$93,$BE
	.dc.b	$9D,$3F,$97,$7B,$75,$7E,$6D,$FE,$67,$D7,$F9,$F7,$F9,$AF,$67,$E8
	.dc.b	$DF,$E6,$FD,$BF,$A7,$7F,$53,$8F,$EA,$F9,$BF,$C5,$BB,$FE,$C3,$5F
	.dc.b	$D9,$4F,$ED,$17,$C5,$7F,$80,$89,$1F,$DE,$BE,$23,$FC,$0B,$ED,$7F
	.dc.b	$81,$BE,$1F,$FC,$0F,$F0,$FF,$E0,$8F,$77,$FC,$13,$E7,$7F,$82,$BD
	.dc.b	$EF,$F0,$5F,$C3,$FF,$83,$3E,$03,$FC,$1B,$F0,$FF,$E0,$EF,$82,$FF
	.dc.b	$07,$FC,$3F,$F8,$43,$E1,$3F,$C2,$3F,$0F,$FE,$12,$F8,$6F,$F2,$0F
	.dc.b	$81,$FD,$9B,$9B,$FD,$74,$DE,$3F,$DA,$CF,$07,$FC,$1F,$7D,$1F,$F0
	.dc.b	$8E,$3E,$07,$FC,$27,$9F,$91,$FF,$0A,$EB,$E8,$7F,$C2,$FB,$FB,$1F
	.dc.b	$F0,$CF,$0F,$C0,$7E,$11,$FE,$1C,$E4,$F8,$AF,$FB,$4B,$97,$E4,$B9
	.dc.b	$21,$FD,$77,$97,$93,$FD,$7E,$BB,$3F,$D8,$6F,$C4,$FF,$63,$C3,$CC
	.dc.b	$FF,$65,$C7,$D4,$FF,$67,$CB,$DC,$FF,$69,$CF,$E1,$3F,$DA,$F4,$FC
	.dc.b	$76,$7F,$90,$8F,$F1,$9C,$6B,$9F,$F3,$55,$C9,$FC,$E5,$F6,$7F,$3D
	.dc.b	$87,$89,$FD,$06,$3E,$67,$F4,$59,$7A,$9F,$D2,$67,$EE,$7F,$4D,$A7
	.dc.b	$C6,$7F,$51,$AF,$F1,$BF,$E0,$1F,$F4,$76,$BF,$EB,$56,$CF,$F8,$8D
	.dc.b	$BC,$1F,$EF,$77,$FD,$77,$19,$F6,$BF,$D0,$70,$F6,$EF,$C3,$E3,$DD
	.dc.b	$CF,$C3,$EB,$DF,$DF,$C3,$F0,$F8,$1E,$3F,$0F,$CB,$E0,$F9,$FC,$3F
	.dc.b	$4F,$85,$EB,$F0,$FF,$D0,$FB,$7F,$20,$7E,$47,$B7,$FB,$67,$27,$A6
	.dc.b	$78,$3F,$EF,$0E,$16,$1D,$63,$19,$78,$5D,$69,$E5,$78,$6D,$E9,$79
	.dc.b	$70,$7B,$B8,$7D,$FC,$5F,$03,$E1,$BE,$0D,$43,$F7,$5C,$DF,$3F,$99
	.dc.b	$1F,$D6,$A3,$FC,$FB,$CF,$F5,$67,$3C,$01,$D1,$F6,$E3,$30,$E9,$FC
	.dc.b	$04,$63,$AB,$F0,$8E,$BA,$FF,$10,$F0,$EC,$FC,$63,$CB,$B7,$F2,$0F
	.dc.b	$4E,$EF,$CA,$AF,$37,$F8,$06,$DF,$E8,$8F,$03,$F8,$23,$FE,$8A,$F0
	.dc.b	$7F,$84,$3F,$E8,$CF,$0B,$F8,$63,$FE,$8E,$F0,$FF,$88,$3F,$E9,$0F
	.dc.b	$13,$F8,$A3,$FE,$92,$F1,$7F,$8C,$3F,$E9,$4F,$1B,$E1,$BF,$B6,$F1
	.dc.b	$BF,$2A,$FC,$A3,$1A,$7C,$2E,$6D,$C7,$F9,$5E,$B9,$BF,$E5,$9B,$EE
	.dc.b	$FF,$96,$F0,$F1,$BF,$E5,$DC,$7C,$EF,$F9,$7F,$2F,$5B,$FE,$61,$CF
	.dc.b	$DE,$FF,$98,$F4,$F8,$6F,$FB,$57,$91,$BF,$CA,$8F,$F0,$AF,$07,$CE
	.dc.b	$7F,$C2,$FC,$3F,$49,$FF,$0C,$F1,$7D,$67,$FC,$37,$C7,$F6,$9F,$F0
	.dc.b	$EF,$27,$DF,$7F,$C3,$FC,$BF,$85,$7F,$C4,$3C,$DF,$8B,$7F,$14,$7C
	.dc.b	$53,$FC,$F1,$C7,$9D,$3F,$15,$3C,$5F,$C6,$57,$57,$F1,$D7,$E1,$7F
	.dc.b	$21,$87,$95,$FC,$96,$3E,$97,$F2,$99,$7B,$5F,$CB,$67,$F0,$5F,$CC
	.dc.b	$69,$F8,$AE,$6F,$8F,$CD,$FD,$D7,$7F,$96,$FF,$77,$E0,$F9,$EF,$F7
	.dc.b	$9E,$1F,$A6,$FF,$7B,$E2,$FA,$EF,$F7,$DE,$3F,$B6,$FF,$7F,$E4,$FB
	.dc.b	$EF,$F8,$07,$97,$F0,$6F,$E2,$39,$7E,$19,$FE,$5E,$EF,$FE,$32,$FE
	.dc.b	$B5,$BB,$1A,$7F,$1C,$E7,$FE,$3D,$D3,$FB,$FA,$FE,$02,$7F,$0B,$F1
	.dc.b	$BF,$8D,$BF,$C3,$7C,$6F,$E3,$6F,$F0,$FF,$1B,$F8,$DB,$FC,$47,$C6
	.dc.b	$FE,$36,$FF,$13,$F1,$BF,$8D,$BF,$C1,$34,$FF,$7B,$25,$97,$F6,$EF
	.dc.b	$57,$FC,$78,$DF,$83,$FF,$0D,$6A,$F1,$FF,$8F,$E1,$FC,$E7,$C7,$FE
	.dc.b	$3C,$FF,$3B,$F1,$FF,$8F,$3F,$CF,$7C,$7F,$E3,$CF,$F3,$FF,$1F,$F8
	.dc.b	$F3,$FD,$07,$C7,$F8,$0F,$F4,$3E,$3F,$E3,$CF,$F2,$B3,$F8,$F1,$FD
	.dc.b	$BF,$C5,$FC,$87,$2F,$F7,$08,$FE,$DC,$7E,$F4,$FE,$F2,$7F,$01,$3F
	.dc.b	$BE,$9F,$C1,$4F,$F0,$09,$FC,$24,$FF,$02,$9F,$C3,$4F,$F0,$49,$FC
	.dc.b	$44,$FF,$06,$9F,$C5,$4F,$ED,$67,$FA,$20,$87,$F7,$13,$47,$F7,$53
	.dc.b	$91,$FD,$E4,$EC,$7F,$7D,$3C,$47,$F8,$04,$F3,$1F,$E0,$53,$D4,$7F
	.dc.b	$82,$4F,$71,$F0,$0F,$C8,$93,$E1,$9D,$3F,$DA,$7E,$1F,$FB,$39,$27
	.dc.b	$FA,$B3,$5F,$EE,$4D,$E3,$FE,$E4,$DE,$2E,$53,$C0,$87,$02,$CF,$06
	.dc.b	$C7,$07,$E2,$1F,$C6,$FC,$0E,$8F,$E9,$6A,$E3,$FA,$56,$F1,$1A,$49
	.dc.b	$EA,$71,$2C,$F1,$6C,$F1,$7F,$69,$3C,$5F,$41,$30,$F4,$37,$0F,$44
	.dc.b	$60,$8E,$09,$1E,$13,$81,$63,$6D,$8F,$F0,$D7,$A3,$FC,$5E,$3F,$C0
	.dc.b	$CF,$8E,$77,$87,$A6,$79,$B6,$33,$87,$AE,$72,$D8,$CE,$1C,$F0,$F2
	.dc.b	$DF,$3C,$FF,$07,$87,$F1,$3F,$07,$FC,$52,$47,$FC,$31,$ED,$7F,$8B
	.dc.b	$3E,$07,$FC,$5B,$ED,$FF,$8B,$BE,$0B,$FC,$17,$5F,$C1,$F2,$FF,$04
	.dc.b	$F9,$BF,$C2,$23,$F8,$64,$63,$E0,$F7,$B6,$75,$5B,$D9,$CD,$F2,$F8
	.dc.b	$4B,$8F,$F8,$4E,$9E,$DE,$30,$E3,$F0,$B6,$5E,$ED,$F0,$31,$EF,$67
	.dc.b	$E1,$BF,$A1,$E9,$FC,$67,$3F,$28,$F5,$FF,$C7,$59,$FF,$6C,$8C,$BE
	.dc.b	$6A,$E7,$18,$CB,$F9,$1B,$E5,$CB,$C5,$9F,$8F,$4F,$26,$BE,$5D,$BC
	.dc.b	$DC,$1E,$7D,$7E,$56,$3F,$E4,$5E,$2F,$9C,$FD,$03,$FC,$91,$CB,$F4
	.dc.b	$D7,$37,$D8,$DF,$F0,$57,$F7,$3A,$3F,$00,$DE,$8F,$C0,$5F,$E3,$C9
	.dc.b	$5C,$1B,$5C,$1B,$3C,$27,$47,$F5,$5E,$9F,$CB,$4F,$F9,$D5,$D5,$71
	.dc.b	$9D,$B4,$95,$A2,$57,$EC,$9F,$E1,$D8,$7A,$E7,$F3,$53,$F8,$CA,$FD
	.dc.b	$03,$63,$2F,$F0,$93,$15,$FE,$88,$70,$43,$F6,$46,$8D,$4F,$47,$4D
	.dc.b	$76,$62,$7B,$B8,$24,$7F,$83,$9F,$C4,$F1,$FF,$A0,$BC,$4F,$C1,$CE
	.dc.b	$67,$F8,$59,$E1,$3F,$D3,$AF,$72,$57,$66,$DF,$AE,$7B,$CE,$49,$19
	.dc.b	$9D,$CF,$B2,$72,$D9,$FD,$AE,$CF,$80,$F6,$47,$1A,$DD,$79,$1A,$F9
	.dc.b	$C8,$70,$3A,$10,$E1,$1C,$C8,$E4,$E2,$7F,$87,$F3,$FC,$4F,$47,$FA
	.dc.b	$ED,$FE,$9D,$1A,$C7,$D6,$F4,$39,$1E,$13,$04,$38,$5F,$A1,$D1,$F3
	.dc.b	$F1,$F8,$0F,$CA,$C3,$CA,$F1,$36,$43,$47,$80,$A6,$CA,$79,$EB,$F8
	.dc.b	$33,$FD,$64,$FE,$C2,$79,$23,$BB,$4E,$91,$EE,$70,$75,$70,$FB,$E7
	.dc.b	$D3,$18,$8F,$09,$E7,$70,$4A,$F8,$51,$A2,$47,$E3,$AF,$E8,$EB,$FD
	.dc.b	$42,$7C,$E7,$3A,$E3,$5C,$F0,$F4,$CE,$5A,$F4,$9F,$EA,$18,$79,$9F
	.dc.b	$DF,$C7,$EF,$48,$7F,$54,$9F,$E0,$2D,$33,$8F,$F5,$67,$FB,$09,$FA
	.dc.b	$CC,$FF,$96,$E9,$FE,$ED,$B6,$7C,$AF,$85,$F3,$23,$82,$67,$FC,$8E
	.dc.b	$3F,$C2,$BA,$FF,$29,$E5,$FC,$2C,$7F,$59,$3F,$A7,$6D,$FD,$2C,$FE
	.dc.b	$99,$F0,$3F,$D3,$7D,$AF,$E9,$DF,$05,$FD,$3F,$E1,$BF,$A8,$7C,$27
	.dc.b	$F5,$1F,$7B,$FA,$97,$C2,$FF,$53,$F8,$0F,$EA,$9F,$0D,$FD,$57,$E0
	.dc.b	$BF,$AC,$7F,$00,$BE,$22,$FC,$2E,$3E,$23,$3F,$0F,$FD,$77,$A3,$DF
	.dc.b	$C7,$FE,$4A,$E7,$FE,$D1,$CF,$FE,$E5,$8F,$8E,$7A,$7B,$59,$B0,$E9
	.dc.b	$3F,$D6,$A1,$97,$8B,$3F,$24,$FF,$5B,$D7,$D0,$FF,$5D,$DF,$D8,$FF
	.dc.b	$5F,$E1,$F8,$0F,$F6,$1E,$3F,$88,$FF,$63,$E5,$F9,$0F,$F6,$5E,$6F
	.dc.b	$F0,$2F,$47,$F3,$DE,$CF,$B1,$C1,$F4,$EB,$EB,$B7,$87,$EB,$C7,$F5
	.dc.b	$23,$F0,$B1,$FE,$18,$E1,$FD,$7C,$FF,$83,$7A,$FF,$98,$78,$FF,$DF
	.dc.b	$2F,$81,$D0,$B3,$D2,$BF,$AA,$E1,$E2,$9C,$E1,$FE,$53,$C7,$F0,$99
	.dc.b	$7A,$A7,$E6,$75,$FE,$1B,$1F,$F0,$C2,$D3,$E0,$CE,$B5,$F8,$AF,$0D
	.dc.b	$BE,$4B,$CB,$7F,$9A,$F4,$E0,$FA,$2F,$D9,$E3,$FC,$23,$F5,$F8,$FE
	.dc.b	$CB,$9E,$4F,$C2,$76,$7E,$EA,$3F,$96,$1F,$76,$37,$AF,$F8,$B6,$FE
	.dc.b	$FC,$FE,$07,$97,$BC,$FA,$B1,$F1,$BE,$CC,$BC,$EF,$BB,$3F,$5B,$F0
	.dc.b	$34,$F7,$BF,$0B,$5F,$8C,$FC,$4D,$FE,$FF,$98,$CD,$C9,$F9,$4F,$88
	.dc.b	$FD,$09,$23,$F2,$9E,$77,$E9,$3E,$1F,$F4,$BE,$E7,$E9,$BE,$1F,$F4
	.dc.b	$FE,$F7,$EA,$3E,$1F,$F5,$39,$FE,$AB,$E1,$FF,$57,$F0,$3F,$AC,$F8
	.dc.b	$7F,$D6,$FC,$1F,$EB,$BE,$1F,$F5,$FF,$0B,$FB,$0F,$87,$F9,$3C,$1F
	.dc.b	$C3,$7C,$3F,$F2,$6F,$89,$FC,$37,$BF,$53,$FA,$7B,$E4,$FE,$A7,$1E
	.dc.b	$CF,$EA,$F3,$F1,$3F,$AD,$D7,$CC,$FE,$BF,$7F,$53,$FB,$1E,$1F,$73
	.dc.b	$F0,$0F,$DA,$72,$7C,$37,$FC,$7B,$5F,$27,$97,$F7,$7F,$0F,$FC,$20
	.dc.b	$93,$FB,$BF,$6F,$FC,$C5,$FC,$03,$FE,$72,$8F,$E0,$4F,$F9,$B1,$FF
	.dc.b	$2C,$AF,$77,$7F,$13,$C3,$DF,$E3,$F1,$3C,$BE,$07,$9F,$C4,$F4,$F8
	.dc.b	$3E,$BF,$13,$DB,$E1,$7B,$FC,$4F,$83,$E1,$FC,$3F,$13,$F9,$3E,$26
	.dc.b	$E8,$D7,$07,$8A,$F9,$3E,$4C,$7B,$3E,$6C,$FC,$4F,$A3,$5F,$33,$EA
	.dc.b	$DF,$DF,$C1,$F0,$3D,$0F,$83,$3F,$CB,$38,$BE,$2B,$FA,$BE,$3F,$C9
	.dc.b	$BF,$92,$F3,$9C,$DF,$3F,$C8,$FD,$33,$FC,$D3,$C7,$FA,$B5,$FB,$3E
	.dc.b	$9F,$B2,$FE,$D7,$AB,$EE,$BF,$B7,$EA,$FC,$0B,$FB,$9E,$BF,$C2,$BF
	.dc.b	$BB,$EC,$FC,$4B,$FB,$DE,$DF,$C6,$BF,$E6,$93,$F2,$44,$30,$7F,$03
	.dc.b	$F0,$7F,$32,$FF,$39,$F0,$BF,$3A,$FF,$3B,$F0,$FF,$42,$FF,$3D,$F1
	.dc.b	$3F,$4A,$FF,$3F,$F1,$7F,$52,$FF,$41,$F1,$BF,$5A,$FF,$43,$F1,$FE
	.dc.b	$03,$F9,$22,$1F,$92,$3E,$47,$27,$CC,$E3,$FE,$18,$FC,$F8,$FC,$69
	.dc.b	$F4,$CF,$C7,$F3,$FE,$39,$FC,$8F,$89,$F8,$E7,$F2,$7E,$27,$E3,$9F
	.dc.b	$CA,$F8,$9F,$8E,$7F,$2F,$E2,$7E,$39,$FC,$CF,$89,$F8,$E7,$FB,$17
	.dc.b	$A5,$FA,$53,$FB,$13,$FA,$0F,$0F,$F1,$EF,$E8,$BC,$3F,$C7,$BF,$A4
	.dc.b	$F0,$FF,$1E,$FE,$9B,$C3,$FC,$7B,$FA,$8F,$0F,$F1,$EF,$EA,$BC,$3F
	.dc.b	$C7,$BF,$AA,$F0,$FF,$1E,$FE,$C4,$FE,$D9,$C3,$F9,$0F,$4B,$FD,$69
	.dc.b	$E1,$FE,$3A,$7E,$5F,$27,$E3,$9F,$E0,$07,$F3,$FE,$5F,$A6,$FF,$41
	.dc.b	$E6,$FA,$EF,$F4,$3E,$7F,$B6,$FF,$45,$E8,$FB,$EF,$F4,$7E,$9F,$C1
	.dc.b	$BF,$D2,$7A,$BF,$0E,$FE,$24,$FF,$02,$E3,$FD,$9F,$DB,$FF,$15,$9F
	.dc.b	$E9,$6F,$57,$FC,$1B,$3B,$3F,$2C,$7E,$3C,$FE,$40,$7E,$3C,$FE,$40
	.dc.b	$7E,$3C,$FE,$40,$7E,$3C,$FE,$40,$7E,$3C,$FE,$40,$7E,$3C,$F8,$33
	.dc.b	$F0,$DF,$EC,$1E,$2F,$ED,$BC,$8F,$E3,$F1,$F3,$08,$7F,$65,$F1,$FE
	.dc.b	$8B,$FC,$7B,$8B,$EA,$BF,$5C,$FC,$79,$FC,$24,$FC,$79,$FC,$34,$FC
	.dc.b	$79,$FC,$44,$FC,$79,$FC,$54,$FC,$79,$FC,$A3,$A3,$F8,$E7,$91,$FC
	.dc.b	$6E,$32,$7E,$34,$E2,$7E,$34,$EA,$7E,$34,$F0,$9F,$8D,$3C,$A7,$9C
	.dc.b	$FE,$6F,$E0,$7E,$A5,$FE,$72,$7E,$B4,$FE,$34,$7C,$2E,$4F,$C6,$E5
	.dc.b	$FC,$67,$AF,$FA,$37,$2F,$E3,$EB,$FC,$15,$E4,$7F,$1A,$DB,$FC,$05
	.dc.b	$E4,$7F,$1A,$7F,$C0,$9E,$47,$F1,$A7,$FC,$0D,$E4,$7F,$1A,$7F,$C1
	.dc.b	$1E,$47,$F1,$A7,$FC,$15,$E4,$7F,$1A,$7F,$C1,$9E,$47,$F1,$A7,$FD
	.dc.b	$75,$97,$F1,$97,$F1,$DE,$5F,$F6,$E8,$FC,$64,$70,$7F,$18,$7F,$25
	.dc.b	$3F,$18,$7F,$29,$3F,$18,$7F,$2D,$3F,$18,$7F,$31,$3F,$18,$7F,$35
	.dc.b	$3F,$18,$7C,$03,$F5,$FA,$FC,$4D,$BF,$19,$B3,$BB,$F8,$D4,$FC,$87
	.dc.b	$E7,$5A,$3F,$B8,$71,$7F,$6B,$8C,$61,$FE,$1A,$8F,$ED,$8F,$D8,$3F
	.dc.b	$AA,$1F,$72,$37,$CD,$FD,$8B,$C0,$FE,$1B,$18,$A7,$83,$FC,$4A,$31
	.dc.b	$E1,$7F,$16,$75,$E1,$FF,$1A,$78,$74,$7E,$48,$C6,$7E,$69,$E0,$FA
	.dc.b	$2B,$A3,$EA,$BF,$0B,$EC,$C7,$F8,$EE,$1E,$3F,$0F,$F5,$2F,$97,$AB
	.dc.b	$F5,$AF,$F9,$27,$AF,$E3,$ED,$FD,$08,$FF,$1E,$75,$39,$FE,$3F,$C4
	.dc.b	$FF,$84,$89,$3F,$E1,$0F,$5B,$FC,$2B,$E6,$7F,$85,$BD,$EF,$F0,$BF
	.dc.b	$C4,$FF,$86,$3E,$27,$FC,$33,$F0,$3F,$E1,$AF,$89,$FF,$24,$63,$FE
	.dc.b	$9B,$FE,$01,$FA,$2D,$F8,$4C,$F0,$7C,$E9,$7C,$26,$39,$7F,$C5,$10
	.dc.b	$F8,$5C,$61,$FE,$52,$87,$C3,$63,$B5,$0F,$87,$C2,$F8,$8F,$EA,$17
	.dc.b	$FE,$C5,$F6,$7F,$D1,$EF,$F0,$53,$FC,$BD,$A7,$8E,$78,$DF,$2C,$79
	.dc.b	$E5,$D3,$4E,$BA,$F6,$D7,$E0,$65,$FE,$36,$DF,$E2,$3F,$E3,$8E,$1F
	.dc.b	$90,$FF,$8E,$B8,$FF,$C7,$FC,$9F,$3A,$3F,$C7,$FC,$BF,$CA,$7C,$03
	.dc.b	$9E,$0F,$C8,$3C,$27,$91,$FE,$44,$E6,$FA,$6F,$F9,$1B,$9B,$E9,$BF
	.dc.b	$E4,$8E,$6F,$33,$EE,$D7,$E6,$ED,$EB,$8C,$1F,$37,$83,$E0,$38,$E6
	.dc.b	$F9,$BC,$7F,$33,$93,$E6,$F8,$9F,$E4,$7E,$73,$9B,$E1,$1F,$4D,$BC
	.dc.b	$DF,$29,$CF,$37,$CA,$77,$A7,$CA,$7F,$CC,$3C,$DF,$29,$FF,$31,$F3
	.dc.b	$7C,$A7,$FC,$CB,$CD,$F0,$DF,$92,$7F,$99,$B8,$FF,$A9,$64,$F8,$AF
	.dc.b	$C1,$92,$7C,$BF,$15,$F8,$6F,$73,$F0,$FE,$EF,$E2,$3E,$2B,$F1,$3E
	.dc.b	$FF,$E2,$BE,$2B,$F1,$7F,$03,$F8,$CF,$8A,$FC,$6F,$C1,$FE,$3B,$E2
	.dc.b	$BF,$1F,$F0,$BF,$90,$F8,$AF,$CA,$7C,$4F,$FB,$0B,$7F,$F7,$E3,$FE
	.dc.b	$BA,$7F,$C3,$C7,$F2,$FE,$6F,$F5,$77,$8A,$F6,$FE,$5C,$FE,$54,$E3
	.dc.b	$5D,$3E,$23,$B4,$7C,$EC,$7C,$4F,$D0,$CF,$CC,$FD,$2D,$7D,$4F,$D4
	.dc.b	$DF,$DC,$FD,$6E,$1F,$84,$FC,$43,$EC,$F2,$7C,$77,$FC,$95,$DF,$FE
	.dc.b	$84,$E7,$FC,$A6,$FF,$C2,$27,$F8,$31,$F9,$43,$FD,$33,$E2,$7E,$51
	.dc.b	$FF,$4D,$F8,$9F,$94,$7F,$D3,$BE,$27,$E5,$1F,$E0,$87,$E1,$0F,$C4
	.dc.b	$F8,$7F,$92,$7F,$85,$78,$7F,$8D,$7F,$86,$78,$5F,$91,$31,$E2,$FF
	.dc.b	$0B,$9E,$0F,$F0,$DA,$FC,$F5,$76,$3F,$8A,$65,$F8,$F9,$FE,$79,$E0
	.dc.b	$FE,$3D,$FE,$7D,$E0,$FE,$3D,$FE,$81,$E0,$FE,$3D,$FD,$9D,$FE,$3C
	.dc.b	$FE,$57,$E4,$7C,$37,$F2,$64,$8F,$E9,$1E,$47,$F2,$A8,$FD,$C1,$FC
	.dc.b	$14,$FD,$C1,$FC,$24,$FD,$C1,$FA,$43,$F5,$7E,$3F,$AA,$7D,$91,$FA
	.dc.b	$C3,$F5,$FA,$7C,$2C,$FF,$B0,$FA,$3F,$EE,$4E,$CF,$F2,$3E,$FF,$C8
	.dc.b	$B0,$FE,$3D,$87,$03,$91,$FC,$DF,$7E,$D8,$7F,$06,$8F,$11,$F6,$0F
	.dc.b	$31,$F7,$0F,$51,$F8,$03,$F0,$53,$F0,$1F,$8C,$7C,$4F,$2B,$FA,$83
	.dc.b	$FD,$CF,$93,$F1,$6D,$3F,$16,$7E,$F8,$F8,$E7,$F0,$23,$E3,$9F,$C1
	.dc.b	$8F,$8E,$7D,$C3,$E3,$9F,$8B,$3D,$87,$BA,$7F,$0C,$FF,$16,$F2,$7F
	.dc.b	$8D,$BF,$E9,$41,$CD,$FE,$0B,$E6,$CC,$7F,$5B,$9F,$E8,$AF,$3C,$7F
	.dc.b	$AB,$E5,$F2,$23,$FA,$B6,$FF,$9D,$C7,$FC,$11,$CB,$FA,$17,$FC,$15
	.dc.b	$CD,$FA,$57,$FC,$19,$97,$E1,$6F,$FE,$00,$F1,$7C,$4D,$3F,$61,$1F
	.dc.b	$8E,$3F,$B1,$73,$FE,$D6,$FF,$B5,$3F,$D8,$3C,$6F,$DD,$CF,$F7,$1E
	.dc.b	$9F,$DF,$BF,$DC,$FA,$BF,$81,$BF,$DD,$7A,$FF,$83,$BF,$DD,$FB,$3F
	.dc.b	$85,$BF,$DE,$7B,$7F,$87,$BF,$C6,$BB,$BF,$8E,$F7,$FE,$2F,$93,$F8
	.dc.b	$CF,$4F,$E3,$AB,$F8,$C6,$DF,$E2,$AA,$FE,$40,$DB,$E6,$FF,$80,$70
	.dc.b	$EE,$FF,$80,$F1,$F1,$BF,$E0,$5C,$BC,$EF,$F8,$1F,$3F,$5B,$FE,$09
	.dc.b	$D3,$DE,$FF,$82,$F5,$F8,$9A,$FF,$82,$B6,$FE,$D9,$CF,$F2,$FC,$AF
	.dc.b	$ED,$BB,$FE,$3B,$0F,$DE,$BF,$8E,$7F,$80,$BF,$8E,$7F,$82,$BF,$8E
	.dc.b	$7F,$C3,$17,$FD,$66,$3F,$C3,$DE,$4F,$F5,$D7,$FC,$41,$E5,$7E,$19
	.dc.b	$FC,$41,$FD,$93,$B3,$F8,$C3,$F9,$13,$F6,$47,$03,$F9,$09,$F9,$9B
	.dc.b	$FF,$13,$74,$FE,$2F,$0F,$F2,$EF,$4F,$E2,$DF,$F2,$FF,$4F,$E2,$DF
	.dc.b	$F3,$0F,$4F,$E2,$EF,$F6,$58,$7F,$13,$9F,$E2,$DD,$9F,$B2,$E9,$FF
	.dc.b	$14,$E7,$F9,$EB,$FF,$1A,$36,$BF,$C6,$6D,$BF,$F1,$93,$6B,$8C,$7B
	.dc.b	$48,$7F,$36,$2D,$74,$6D,$74,$69,$D0,$FE,$C8,$3C,$3F,$EC,$97,$DF
	.dc.b	$CA,$FE,$CB,$3D,$49,$7D,$5B,$7D,$5A,$75,$2E,$1E,$86,$D7,$4A,$E4
	.dc.b	$4A,$E4,$F5,$C7,$E4,$4E,$E7,$24,$AE,$D5,$C5,$B1,$9B,$F0,$4E,$5B
	.dc.b	$19,$AE,$95,$C0,$93,$A6,$C6,$A3,$07,$F1,$78,$FE,$34,$E2,$34,$4C
	.dc.b	$3E,$51,$27,$E8,$CE,$E3,$11,$F1,$1B,$87,$E1,$FA,$9F,$15,$FE,$1E
	.dc.b	$25,$7F,$91,$E1,$F1,$5F,$E2,$2F,$5B,$FC,$47,$EE,$7F,$89,$3E,$13
	.dc.b	$FC,$4B,$EE,$FF,$89,$BE,$13,$FC,$4F,$EF,$7F,$8A,$3E,$13,$FC,$53
	.dc.b	$E5,$7F,$8A,$BE,$13,$FC,$57,$F0,$1F,$E2,$CF,$84,$FF,$16,$FC,$0F
	.dc.b	$F8,$BB,$E1,$3F,$89,$70,$FE,$1F,$CA,$FE,$24,$FF,$8B,$8F,$F1,$8C
	.dc.b	$70,$7C,$D5,$D1,$F4,$61,$E0,$7D,$59,$79,$1F,$66,$9E,$87,$DD,$B7
	.dc.b	$B1,$F8,$1C,$1F,$E4,$0E,$1F,$87,$1F,$E4,$0E,$3F,$F2,$9F,$2F,$CC
	.dc.b	$E9,$FC,$07,$47,$D3,$AF,$A2,$D7,$E8,$18,$9E,$1D,$3F,$85,$7F,$60
	.dc.b	$BA,$FF,$D4,$4D,$9E,$04,$39,$92,$FC,$35,$D2,$3F,$16,$DE,$AF,$D2
	.dc.b	$75,$7E,$7A,$3F,$D7,$73,$FE,$BC,$9F,$D4,$37,$1F,$D9,$B6,$76,$FE
	.dc.b	$C4,$C1,$89,$E2,$E4,$87,$ED,$4D,$9E,$07,$6E,$4D,$B8,$24,$F8,$A3
	.dc.b	$9F,$5F,$FA,$F9,$FE,$0C,$74,$21,$CA,$38,$3D,$49,$3D,$2B,$D7,$18
	.dc.b	$9D,$CE,$63,$D9,$18,$24,$F0,$9F,$E1,$DE,$03,$D8,$F0,$4E,$6B,$77
	.dc.b	$CA,$30,$47,$04,$3E,$24,$70,$6C,$6D,$A6,$CF,$79,$E9,$7C,$99,$FE
	.dc.b	$1E,$19,$A1,$1F,$E2,$27,$51,$FC,$B6,$39,$E3,$F0,$5E,$E4,$33,$7E
	.dc.b	$56,$DF,$95,$A7,$94,$F8,$67,$83,$1F,$8C,$F9,$CD,$14,$EC,$70,$29
	.dc.b	$93,$A9,$8A,$F3,$38,$3D,$A7,$F0,$79,$FF,$33,$1F,$E3,$E7,$47,$52
	.dc.b	$9D,$4C,$10,$C1,$E5,$9F,$4C,$F8,$23,$4D,$76,$D2,$57,$C7,$8D,$B5
	.dc.b	$FA,$5A,$7E,$F8,$F3,$3A,$BF,$98,$D3,$E3,$D7,$48,$CC,$78,$2B,$F1
	.dc.b	$17,$E6,$7F,$88,$61,$FC,$12,$7F,$84,$35,$FE,$11,$8F,$F1,$17,$2D
	.dc.b	$72,$79,$8F,$EA,$33,$D2,$35,$3C,$C8,$E0,$93,$FC,$48,$EA,$7F,$77
	.dc.b	$D3,$FA,$87,$C1,$7F,$BB,$09,$1F,$D3,$FE,$2B,$FD,$DF,$F0,$5F,$EF
	.dc.b	$0F,$73,$EE,$E3,$F9,$BA,$FF,$3D,$FF,$00,$FF,$1A,$8C,$6C,$87,$BD
	.dc.b	$C3,$4F,$D0,$CE,$7E,$17,$18,$72,$F8,$0B,$3C,$7E,$2E,$E2,$F8,$2B
	.dc.b	$F1,$78,$F8,$4C,$FC,$5E,$BE,$1B,$7F,$17,$C3,$E2,$3E,$E6,$BF,$DE
	.dc.b	$32,$46,$A3,$F6,$7C,$3E,$89,$5F,$49,$CF,$07,$A2,$59,$78,$A7,$27
	.dc.b	$C2,$D3,$CF,$AF,$A3,$7F,$4F,$07,$AB,$87,$D7,$7F,$2C,$FE,$F9,$C7
	.dc.b	$F0,$3C,$D7,$2F,$D2,$9F,$E1,$3A,$FC,$89,$73,$7F,$60,$9F,$B7,$CB
	.dc.b	$FE,$0D,$6F,$47,$E1,$5F,$C0,$1F,$8A,$8D,$36,$3F,$BD,$F1,$7D,$BE
	.dc.b	$DF,$F2,$1C,$F2,$EB,$FE,$B1,$5C,$EF,$8D,$FE,$3A,$BA,$B6,$3A,$F5
	.dc.b	$7E,$32,$7F,$C3,$18,$FA,$EB,$FD,$31,$A7,$E3,$27,$FC,$39,$D5,$F1
	.dc.b	$1F,$97,$87,$E2,$77,$F9,$31,$8E,$57,$17,$CE,$8C,$72,$7D,$37,$E3
	.dc.b	$9F,$94,$E5,$FB,$4E,$30,$FC,$A7,$37,$E0,$63,$18,$7E,$11,$FF,$1E
	.dc.b	$F1,$FF,$34,$3F,$01,$CF,$BD,$BF,$0D,$7D,$1F,$C4,$63,$E1,$7F,$15
	.dc.b	$97,$97,$0E,$B9,$FA,$9E,$FA,$7B,$DF,$27,$23,$6F,$89,$18,$DF,$E4
	.dc.b	$6B,$F8,$1E,$0F,$E5,$BC,$3F,$E8,$DE,$07,$C5,$FE,$98,$98,$7C,$7F
	.dc.b	$80,$FD,$4F,$AB,$FA,$AF,$7B,$F5,$7F,$05,$FA,$CF,$07,$F5,$BF,$07
	.dc.b	$FA,$EF,$53,$F5,$FF,$09,$FB,$0F,$5B,$F6,$7F,$C0,$3F,$BF,$D3,$E5
	.dc.b	$2F,$63,$3E,$3E,$BE,$0F,$7E,$D7,$EC,$6F,$E5,$E9,$5E,$31,$B7,$EF
	.dc.b	$E0,$BE,$F8,$78,$39,$7C,$55,$F8,$1C,$FC,$CF,$E0,$F5,$F5,$3F,$85
	.dc.b	$DB,$DC,$FE,$1F,$83,$F8,$4F,$0F,$C4,$8F,$BF,$C5,$F2,$BD,$7F,$E2
	.dc.b	$7A,$BE,$17,$DE,$4A,$F2,$FC,$2F,$C2,$F1,$7E,$1F,$BD,$F1,$3D,$FF
	.dc.b	$8B,$E7,$FC,$6F,$80,$F8,$FF,$15,$F2,$3E,$0B,$E4,$FC,$57,$CA,$F8
	.dc.b	$4F,$97,$F1,$5F,$33,$E1,$BE,$6F,$C5,$7C,$EF,$88,$F8,$DE,$67,$F2
	.dc.b	$AE,$BF,$C9,$71,$7F,$41,$3F,$2F,$E7,$ED,$F9,$37,$CD,$F9,$78,$F7
	.dc.b	$7E,$6E,$7E,$37,$E7,$EB,$E7,$7E,$8E,$FE,$B7,$E9,$F0,$FB,$DF,$82
	.dc.b	$7D,$6E,$4F,$88,$BC,$6F,$C9,$74,$7C,$DD,$BF,$95,$3F,$42,$7E,$91
	.dc.b	$FB,$9F,$17,$F2,$2F,$EE,$FC,$2F,$C8,$BF,$BD,$F1,$7F,$22,$FE,$FF
	.dc.b	$C5,$FC,$13,$F8,$6F,$0B,$F1,$EF,$E2,$3C,$5F,$C7,$BF,$ED,$EF,$4B
	.dc.b	$FC,$A9,$9F,$FB,$07,$B9,$E3,$7E,$3E,$BF,$33,$E3,$7E,$39,$FC,$8F
	.dc.b	$47,$E3,$9F,$E0,$7E,$37,$E3,$9F,$E0,$BE,$37,$E3,$9F,$E0,$FE,$37
	.dc.b	$E3,$9F,$E1,$3E,$37,$BD,$FC,$69,$FC,$5F,$8F,$F1,$CB,$D1,$69,$F8
	.dc.b	$F8,$F9,$84,$3F,$5B,$E2,$FD,$17,$F5,$FD,$1F,$55,$FD,$8F,$4F,$D9
	.dc.b	$7F,$67,$D5,$F7,$5F,$DA,$F5,$FE,$05,$FD,$BF,$67,$E1,$5F,$DC,$F6
	.dc.b	$FE,$25,$FC,$59,$FB,$63,$F9,$77,$76,$79,$7F,$BD,$E8,$AE,$51,$8B
	.dc.b	$EC,$EB,$0F,$13,$C3,$1F,$33,$CB,$2F,$53,$D3,$3F,$73,$DB,$4F,$C6
	.dc.b	$BE,$0E,$EF,$93,$B7,$F8,$2F,$7F,$C7,$72,$7F,$73,$7E,$85,$7F,$79
	.dc.b	$E1,$FA,$8F,$F7,$BE,$2F,$B0,$FF,$7D,$E3,$FB,$8F,$F7,$FE,$4F,$C0
	.dc.b	$3E,$EC,$3F,$09,$C9,$F8,$63,$F1,$C7,$F1,$6E,$6F,$DA,$C7,$F8,$68
	.dc.b	$FD,$AE,$7F,$E7,$DE,$CC,$D7,$DF,$E5,$FC,$8B,$C8,$E8,$7F,$2A,$EA
	.dc.b	$FC,$EB,$FD,$4B,$AF,$F4,$2F,$F5,$4E,$CF,$D2,$BF,$D5,$AF,$F5,$2F
	.dc.b	$F5,$8E,$DF,$E7,$AF,$F3,$F3,$F2,$04,$3F,$6D,$E3,$7F,$1E,$E3,$F9
	.dc.b	$BA,$59,$FD,$E9,$FC,$F3,$C6,$FE,$3A,$FF,$34,$F1,$BF,$8E,$BF,$CD
	.dc.b	$BC,$6F,$E3,$AF,$F3,$8F,$1B,$F8,$EB,$FC,$EB,$C6,$FE,$3A,$FF,$3C
	.dc.b	$F1,$BF,$8E,$BF,$CF,$BC,$6F,$8B,$C7,$F1,$E3,$FC,$69,$E1,$7E,$3F
	.dc.b	$4F,$E6,$7E,$17,$E3,$9F,$E6,$BE,$17,$E3,$9F,$E6,$FE,$17,$E3,$9F
	.dc.b	$E7,$3E,$17,$E3,$9F,$E7,$7E,$17,$E3,$9F,$E7,$BE,$17,$E3,$9F,$E7
	.dc.b	$FE,$17,$C2,$CF,$EB,$72,$7E,$3B,$C8,$FE,$D7,$1F,$27,$BB,$F8,$DB
	.dc.b	$F3,$4F,$DF,$9F,$C6,$CF,$EA,$07,$F1,$B3,$FA,$A1,$FC,$6C,$FE,$B0
	.dc.b	$7F,$1B,$3F,$87,$9F,$C6,$CF,$EC,$3E,$27,$F1,$58,$FD,$7F,$7F,$F9
	.dc.b	$8B,$D2,$FF,$1A,$E5,$FE,$C8,$F1,$BF,$C6,$B8,$7F,$86,$E5,$D5,$FC
	.dc.b	$8A,$3F,$BA,$F8,$FF,$C9,$9F,$C7,$1E,$03,$F1,$C7,$90,$FC,$71,$E8
	.dc.b	$3F,$1C,$7B,$0F,$C7,$1F,$CE,$8F,$F1,$3F,$93,$FB,$27,$FD,$03,$E4
	.dc.b	$7E,$DB,$C4,$FE,$39,$1F,$E0,$6D,$FE,$77,$47,$F2,$0E,$1F,$A5,$3F
	.dc.b	$A5,$E2,$FA,$CF,$E9,$F8,$FE,$D3,$FA,$9E,$4F,$BC,$FE,$AF,$97,$F0
	.dc.b	$4F,$EB,$79,$BF,$0C,$FE,$BF,$9F,$F1,$4F,$FA,$53,$A3,$FD,$02,$FF
	.dc.b	$90,$79,$3F,$6F,$D1,$FE,$3F,$8F,$F0,$DF,$4F,$E5,$9F,$F0,$EF,$57
	.dc.b	$E6,$9F,$F0,$FF,$5F,$E7,$9F,$F1,$0F,$67,$E8,$9F,$F1,$1F,$6F,$E9
	.dc.b	$9F,$F1,$2F,$77,$EA,$9F,$D6,$1F,$AE,$ED,$FE,$38,$FF,$3F,$F2,$5E
	.dc.b	$07,$F8,$4B,$E2,$BF,$C5,$E4,$9F,$F3,$1F,$F0,$0B,$E2,$FF,$D4,$AB
	.dc.b	$D0,$BE,$E6,$3D,$DC,$FB,$DA,$F8,$AD,$FC,$07,$0F,$8A,$E3,$F0,$5C
	.dc.b	$BE,$2B,$9F,$C2,$74,$E0,$EB,$F0,$BD,$BE,$2B,$BF,$C3,$FC,$AE,$5F
	.dc.b	$F0,$A7,$A5,$8E,$4F,$2D,$70,$7C,$F8,$74,$7D,$39,$78,$1F,$5E,$9E
	.dc.b	$47,$DB,$B7,$A1,$F7,$F0,$7B,$1F,$83,$C5,$F0,$39,$3E,$24,$7F,$94
	.dc.b	$B9,$7F,$20,$63,$5F,$EA,$EE,$E3,$89,$F2,$B9,$3E,$33,$F3,$39,$3E
	.dc.b	$33,$F3,$B9,$3E,$33,$F4,$39,$3E,$33,$F4,$B9,$3E,$33,$F5,$39,$3E
	.dc.b	$33,$F5,$B9,$3E,$37,$77,$F4,$5E,$5F,$8D,$3F,$E9,$1E,$67,$37,$C7
	.dc.b	$8C,$73,$7C,$77,$5C,$DF,$1D,$E1,$CD,$F1,$DE,$5C,$DF,$1D,$E9,$CD
	.dc.b	$F1,$DE,$DC,$DF,$05,$F8,$C7,$F9,$BB,$93,$FB,$5E,$6F,$8A,$FC,$59
	.dc.b	$27,$F1,$FF,$13,$F8,$EF,$73,$F1,$FE,$EF,$E4,$3E,$13,$F2,$3E,$FF
	.dc.b	$FB,$DB,$DB,$FF,$80,$1C,$9F,$EF,$C7,$FD,$FF,$8F,$FA,$E9,$7C,$2E
	.dc.b	$FE,$0B,$87,$C3,$71,$F8,$4E,$5F,$13,$CF,$E1,$BA,$7C,$4F,$F6,$AF
	.dc.b	$77,$FC,$77,$A7,$FA,$CB,$4E,$93,$B7,$B5,$F3,$7C,$18,$F7,$7C,$59
	.dc.b	$F9,$FE,$1F,$D1,$1E,$A3,$F5,$BB,$FB,$9F,$D7,$F0,$FE,$BB,$8B,$E2
	.dc.b	$47,$D9,$E3,$FC,$7B,$F2,$C8,$7D,$DF,$1B,$F6,$BC,$FF,$4F,$C6,$6E
	.dc.b	$AF,$F2,$17,$3B,$A7,$ED,$56,$A7,$3D,$5F,$7E,$31,$EB,$AD,$D9,$F8
	.dc.b	$59,$CC,$3B,$7F,$15,$18,$EE,$FC,$73,$AE,$FF,$C9,$BF,$BA,$CB,$FD
	.dc.b	$DF,$A7,$F8,$6B,$07,$83,$F9,$C9,$FC,$C7,$85,$FA,$07,$F3,$5E,$1F
	.dc.b	$E9,$1F,$CE,$78,$9F,$A8,$7F,$3D,$E2,$FE,$B1,$FD,$07,$8D,$FB,$07
	.dc.b	$F4,$5E,$3F,$C6,$7F,$8E,$F2,$FE,$E7,$4F,$E4,$51,$F1,$FC,$8F,$E4
	.dc.b	$AF,$F8,$AF,$C7,$D3,$4E,$27,$EF,$8E,$A7,$F0,$A3,$F7,$C7,$F0,$E3
	.dc.b	$F7,$C7,$F1,$23,$F7,$C7,$F1,$63,$F7,$C7,$F1,$A3,$F7,$C7,$F1,$E3
	.dc.b	$F7,$C7,$F9,$9C,$FD,$E6,$BF,$1F,$7F,$A7,$B7,$C7,$7E,$AE,$DF,$1D
	.dc.b	$FA,$FB,$7C,$77,$EC,$ED,$F1,$DF,$B7,$B7,$C7,$7E,$EE,$DF,$1F,$2F
	.dc.b	$C7,$1F,$90,$8F,$E3,$EF,$C7,$E1,$F9,$2F,$F0,$02,$47,$E1,$B7,$E4
	.dc.b	$F4,$8F,$8C,$7D,$63,$E3,$1F,$68,$F8,$C7,$DE,$3E,$31,$F8,$23,$E3
	.dc.b	$1F,$86,$3E,$31,$F1,$0F,$8C,$FE,$5B,$4F,$8D,$BF,$E8,$F9,$3F,$39
	.dc.b	$B7,$C7,$8C,$70,$B7,$F9,$11,$8D,$FE,$43,$AD,$FE,$43,$C3,$7F,$90
	.dc.b	$F2,$DF,$DE,$F4,$DB,$F1,$39,$FF,$AC,$B8,$3F,$91,$F4,$7E,$FB,$7F
	.dc.b	$D5,$BF,$C0,$78,$3F,$81,$9F,$DC,$BC,$7F,$E4,$4F,$F7,$4F,$1F,$F8
	.dc.b	$5B,$FC,$7F,$BB,$F9,$27,$7F,$F1,$18,$61,$FB,$AF,$1D,$E0,$FF,$17
	.dc.b	$9C,$ED,$FC,$A7,$C2,$FE,$39,$18,$3F,$D9,$14,$F0,$FF,$92,$7C,$57
	.dc.b	$F9,$12,$7F,$C3,$C7,$F2,$43,$F9,$87,$8D,$FC,$B1,$FE,$65,$E3,$FF
	.dc.b	$30,$7F,$9A,$79,$1E,$67,$D0,$7F,$37,$3F,$92,$1F,$D5,$FC,$7F,$E4
	.dc.b	$8F,$F5,$9F,$1F,$E2,$3F,$C8,$CF,$F0,$AF,$1F,$F2,$47,$F1,$FA,$FE
	.dc.b	$4B,$7F,$C8,$47,$ED,$5F,$C8,$3F,$D6,$0F,$DA,$9F,$D7,$0F,$DA,$9F
	.dc.b	$D8,$0F,$DA,$9F,$D9,$0F,$DA,$9F,$DA,$0F,$DA,$9F,$DB,$7C,$1F,$E3
	.dc.b	$B3,$FD,$C6,$3F,$8E,$F9,$38,$7F,$98,$CF,$F2,$49,$FC,$7F,$83,$FC
	.dc.b	$A1,$FC,$8F,$87,$FC,$B5,$FC,$9F,$89,$FC,$CD,$FC,$AF,$8B,$FC,$E1
	.dc.b	$FC,$BF,$8D,$FC,$FD,$FC,$CF,$8F,$FD,$C7,$4F,$DA,$E7,$FB,$3F,$1F
	.dc.b	$F8,$DF,$73,$9D,$F1,$3F,$DA,$C9,$5F,$DA,$5A,$FE,$0E,$78,$46,$2F
	.dc.b	$F1,$8D,$C7,$FB,$93,$6B,$F1,$2D,$AE,$31,$D4,$87,$5B,$F1,$BF,$C7
	.dc.b	$8F,$F3,$47,$07,$F8,$71,$FE,$B3,$E4,$73,$9D,$C7,$1B,$C1,$1E,$B8
	.dc.b	$F9,$9B,$8F,$99,$A7,$98,$B8,$F9,$9E,$87,$F1,$D3,$46,$09,$3D,$5B
	.dc.b	$3D,$5A,$70,$8D,$12,$34,$D8,$D4,$60,$8E,$09,$3F,$67,$E2,$BE,$23
	.dc.b	$FE,$DB,$8F,$95,$1F,$05,$F0,$C7,$AE,$BB,$36,$BA,$B6,$7A,$E1,$F3
	.dc.b	$2B,$6D,$AD,$B6,$77,$87,$F8,$35,$FF,$7F,$9F,$C2,$3D,$3F,$F1,$B1
	.dc.b	$1F,$F1,$A7,$BB,$FE,$39,$F7,$BF,$C7,$5F,$15,$FE,$3B,$F8,$5F,$F1
	.dc.b	$E7,$C5,$7F,$8F,$78,$3F,$C7,$DF,$15,$FE,$3F,$F8,$8F,$F2,$07,$C5
	.dc.b	$7F,$90,$7E,$2B,$C1,$C3,$FE,$16,$FD,$FF,$FC,$35,$B7,$E5,$DF,$CF
	.dc.b	$3F,$E2,$18,$D7,$C5,$F0,$E3,$FE,$85,$4F,$88,$FD,$B7,$B1,$FE,$40
	.dc.b	$C3,$F1,$E7,$1A,$DC,$F1,$87,$32,$D8,$66,$31,$CD,$67,$CB,$97,$F3
	.dc.b	$3C,$FD,$78,$7F,$95,$F5,$F7,$BF,$E5,$9D,$FE,$1B,$FE,$5B,$E1,$F8
	.dc.b	$EF,$F9,$77,$8F,$FC,$C5,$C9,$F3,$63,$F5,$5C,$BF,$4F,$A7,$F7,$FA
	.dc.b	$FD,$7D,$BE,$F1,$1C,$13,$9F,$F1,$1B,$7E,$01,$BC,$5F,$80,$6E,$3F
	.dc.b	$86,$E8,$FC,$8C,$B6,$FF,$3A,$46,$B1,$ED,$77,$1E,$D8,$7E,$B9,$AF
	.dc.b	$E5,$9B,$8F,$86,$3F,$61,$5F,$B3,$7F,$1C,$D7,$F5,$C6,$27,$F8,$1B
	.dc.b	$6B,$F2,$8D,$29,$FB,$A3,$06,$A7,$A3,$93,$A9,$0C,$11,$D9,$27,$C3
	.dc.b	$1F,$E9,$3B,$F8,$B8,$7C,$48,$FE,$3E,$78,$08,$75,$AE,$57,$E6,$8E
	.dc.b	$D7,$D1,$B1,$99,$DC,$7B,$1A,$FA,$E7,$84,$7F,$1B,$E2,$7C,$27,$79
	.dc.b	$C9,$2F,$A3,$6F,$A3,$5E,$91,$A2,$3A,$6B,$A3,$04,$30,$79,$71,$F8
	.dc.b	$AF,$FA,$DD,$FE,$9E,$FC,$43,$DE,$71,$31,$87,$B9,$B5,$E5,$6D,$78
	.dc.b	$9A,$7F,$A8,$70,$F8,$2F,$43,$C6,$68,$A7,$C5,$8E,$4D,$3C,$45,$36
	.dc.b	$73,$3F,$2B,$EB,$FF,$A6,$9F,$E1,$03,$9B,$F8,$66,$1F,$EA,$BE,$BF
	.dc.b	$7C,$F3,$3C,$44,$30,$43,$B1,$63,$EE,$67,$72,$F8,$53,$71,$F4,$BC
	.dc.b	$09,$5F,$25,$D1,$F7,$B3,$FA,$6F,$F4,$A3,$F2,$E6,$8F,$7C,$71,$79
	.dc.b	$D7,$81,$B5,$E0,$9F,$2C,$7F,$46,$3F,$80,$47,$F0,$32,$38,$21,$FC
	.dc.b	$0D,$FE,$FD,$5E,$4D,$7F,$CB,$4F,$09,$D1,$C2,$3F,$84,$38,$23,$83
	.dc.b	$FB,$B6,$FF,$E2,$2E,$67,$C5,$FF,$35,$25,$7D,$8C,$7F,$9E,$7C,$47
	.dc.b	$F3,$9F,$88,$FE,$75,$F1,$5F,$CE,$FE,$27,$F9,$E7,$BF,$FC,$F7,$83
	.dc.b	$F9,$F7,$C0,$7F,$3F,$F8,$9F,$E8,$1F,$05,$FD,$07,$E2,$7F,$A1,$7C
	.dc.b	$27,$F4,$3F,$89,$FE,$89,$F0,$DF,$D1,$7E,$27,$F8,$37,$0F,$F8,$80
	.dc.b	$FF,$24,$9F,$A5,$3C,$9C,$68,$E3,$18,$AE,$AE,$B0,$F0,$BC,$32,$F2
	.dc.b	$BC,$B4,$F4,$BD,$36,$F6,$BD,$B8,$3F,$AC,$70,$FC,$48,$FE,$D9,$C7
	.dc.b	$FD,$37,$CF,$FE,$89,$1F,$47,$BE,$BF,$41,$BC,$FF,$45,$A4,$8F,$F0
	.dc.b	$DF,$6F,$E9,$DF,$F4,$87,$6F,$E9,$A3,$15,$C6,$7F,$14,$75,$2B,$9E
	.dc.b	$8F,$E9,$F1,$E3,$37,$D3,$F7,$F2,$F5,$57,$F8,$1D,$F6,$CF,$BC,$F9
	.dc.b	$BD,$3F,$75,$F9,$FD,$3F,$75,$FA,$3D,$3F,$75,$FA,$7D,$3F,$75,$FA
	.dc.b	$BD,$3F,$75,$FA,$FD,$3F,$75,$FB,$47,$DD,$26,$5F,$9C,$7E,$EF,$47
	.dc.b	$DC,$7F,$C3,$5D,$1F,$71,$FF,$0E,$74,$7D,$C7,$FC,$3D,$D1,$F7,$1F
	.dc.b	$F1,$07,$47,$DC,$7F,$C4,$5D,$1F,$71,$FF,$12,$74,$7C,$07,$ED,$9F
	.dc.b	$8C,$E1,$FB,$8F,$F8,$73,$A5,$F1,$1F,$E1,$92,$47,$E4,$BE,$03,$FC
	.dc.b	$3D,$FC,$02,$F8,$2B,$E7,$E3,$E0,$F3,$E8,$EB,$E1,$37,$E9,$F0,$F8
	.dc.b	$5E,$3E,$AF,$2F,$86,$E7,$EB,$F4,$F8,$7E,$BE,$37,$6F,$88,$FD,$2F
	.dc.b	$63,$C7,$FF,$57,$71,$FC,$A7,$FB,$83,$FB,$9E,$25,$72,$8C,$61,$D9
	.dc.b	$D6,$5E,$27,$86,$9E,$67,$96,$DE,$A7,$A7,$07,$B9,$ED,$C5,$FE,$3D
	.dc.b	$E3,$F8,$B1,$FE,$51,$E4,$F9,$7E,$2B,$57,$B5,$F2,$09,$5E,$7F,$37
	.dc.b	$E5,$FB,$7F,$33,$E2,$7E,$6F,$BB,$F3,$BE,$27,$E7,$FB,$FF,$43,$E2
	.dc.b	$7F,$C9,$DF,$C0,$3F,$E1,$EE,$8F,$89,$9F,$F9,$52,$1F,$15,$AF,$8A
	.dc.b	$DF,$C2,$70,$F8,$AE,$3F,$0D,$CB,$E2,$B9,$FC,$47,$F2,$FE,$F6,$7C
	.dc.b	$E3,$4F,$F2,$5C,$D7,$CE,$31,$8F,$77,$59,$F9,$9F,$41,$F5,$76,$F6
	.dc.b	$3F,$5F,$83,$E0,$3F,$67,$8B,$F1,$BC,$7F,$1A,$3F,$DA,$5C,$BF,$29
	.dc.b	$FF,$22,$F8,$DF,$96,$7F,$92,$F4,$7D,$2E,$3F,$F4,$27,$4F,$E5,$BA
	.dc.b	$BE,$C5,$72,$EB,$FB,$8F,$4E,$CF,$C0,$3D,$BB,$7F,$08,$F8,$3B,$BF
	.dc.b	$10,$F8,$BB,$FF,$18,$F9,$3C,$0F,$CA,$F4,$7C,$8F,$5F,$FB,$FF,$B6
	.dc.b	$F6,$BF,$35,$7F,$95,$F1,$3F,$3C,$FE,$5F,$C5,$FD,$13,$F9,$9F,$1B
	.dc.b	$F4,$CF,$E6,$FC,$7F,$D5,$3F,$9D,$F2,$3F,$5C,$FE,$7F,$C9,$F7,$3F
	.dc.b	$B5,$3E,$19,$0F,$F5,$AF,$93,$F9,$3F,$27,$E6,$73,$FE,$22,$31,$3F
	.dc.b	$93,$3E,$91,$FA,$BF,$27,$F2,$6F,$EB,$7C,$9F,$C9,$BF,$AF,$F2,$7E
	.dc.b	$F3,$F8,$03,$F0,$DE,$4F,$E4,$1F,$C4,$79,$3F,$90,$7F,$15,$E4,$FE
	.dc.b	$41,$FE,$BE,$7E,$EC,$FC,$81,$A3,$F2,$1E,$3F,$E3,$D6,$7E,$9F,$21
	.dc.b	$E4,$7E,$46,$73,$D1,$F9,$AD,$95,$FA,$49,$FC,$D1,$F9,$43,$F1,$BE
	.dc.b	$3F,$E5,$1F,$C7,$F8,$FF,$94,$7F,$23,$E3,$FE,$51,$FE,$1D,$C5,$F9
	.dc.b	$5F,$63,$FC,$19,$E2,$FE,$47,$3F,$F6,$6B,$FC,$16,$BF,$83,$9F,$5C
	.dc.b	$FE,$16,$7D,$B3,$F8,$79,$F7,$CF,$E2,$67,$E0,$CF,$E2,$E7,$E1,$CF
	.dc.b	$E3,$67,$E2,$CF,$E4,$1D,$DF,$92,$7F,$DA,$B3,$FE,$B5,$7E,$6E,$DF
	.dc.b	$D3,$DF,$CC,$1C,$4F,$CE,$1D,$4F,$D0,$1E,$13,$F4,$87,$94,$FD,$41
	.dc.b	$E9,$3F,$58,$7B,$4F,$DB,$F8,$9F,$05,$FE,$48,$7E,$4F,$07,$67,$D5
	.dc.b	$F1,$FF,$79,$3F,$BE,$CF,$F2,$B8,$E7,$C8,$FE,$4B,$3F,$93,$3F,$84
	.dc.b	$1F,$93,$3F,$86,$1F,$93,$3E,$E9,$F9,$33,$F0,$67,$F2,$5E,$6F,$CA
	.dc.b	$38,$E2,$FC,$2F,$37,$E5,$63,$1C,$6E,$6F,$CB,$47,$E5,$6B,$FA,$97
	.dc.b	$BB,$8E,$0F,$C9,$57,$17,$F2,$98,$76,$7F,$2D,$8F,$F3,$8F,$2B,$F4
	.dc.b	$F8,$FF,$1C,$E9,$FD,$5B,$FC,$7B,$AB,$F5,$EF,$F2,$0E,$BF,$82,$FF
	.dc.b	$22,$DB,$E2,$AD,$7F,$72,$FE,$FB,$CA,$FE,$48,$FD,$02,$67,$FC,$91
	.dc.b	$FC,$99,$FC,$B3,$B7,$F8,$5B,$FC,$BB,$BB,$F8,$7B,$FC,$C3,$BF,$F8
	.dc.b	$9B,$FC,$CB,$C0,$FE,$2E,$FF,$34,$F0,$7F,$8D,$BF,$CD,$BA,$3F,$8F
	.dc.b	$BF,$E8,$B3,$FC,$7B,$F1,$2F,$6B,$FA,$6C,$61,$FC,$C6,$BF,$D9,$FB
	.dc.b	$7F,$96,$EB,$FC,$57,$AD,$E3,$FF,$32,$8C,$79,$1F,$CD,$9D,$79,$3F
	.dc.b	$CE,$9E,$1E,$57,$F3,$E7,$97,$97,$FD,$09,$E9,$E1,$7E,$55,$FE,$45
	.dc.b	$CB,$F2,$1B,$E1,$FF,$7A,$CF,$F2,$B1,$E3,$F0,$BE,$8E,$DF,$DF,$7A
	.dc.b	$3E,$AC,$F9,$FA,$7E,$CB,$E9,$EA,$FB,$AF,$AF,$AF,$F0,$2F,$F1,$9E
	.dc.b	$CF,$E3,$0F,$F1,$BF,$1F,$F8,$F3,$FC,$77,$C8,$FE,$E3,$CF,$F9,$4F
	.dc.b	$37,$FB,$B3,$FC,$98,$FC,$BF,$85,$F9,$1F,$17,$F9,$07,$8A,$F1,$3F
	.dc.b	$25,$18,$F1,$3F,$25,$7F,$A6,$3F,$24,$7E,$A8,$FC,$91,$FA,$E3,$F2
	.dc.b	$47,$F4,$43,$F2,$47,$F2,$22,$1F,$D3,$A9,$F1,$5F,$E1,$62,$47,$F8
	.dc.b	$27,$DE,$FF,$0C,$F9,$9F,$E1,$AF,$80,$FF,$0D,$FC,$0F,$F8,$73,$E0
	.dc.b	$BF,$C3,$BF,$15,$FE,$1E,$F8,$4F,$F0,$FF,$C5,$7F,$88,$3C,$EF,$F1
	.dc.b	$0F,$C5,$7F,$88,$BE,$1F,$FC,$47,$F1,$5F,$E2,$4F,$89,$FC,$6F,$47
	.dc.b	$F5,$DE,$3F,$F1,$2F,$0F,$C9,$3F,$91,$F2,$7F,$92,$A7,$FA,$57,$3F
	.dc.b	$08,$FE,$91,$5C,$27,$FA,$75,$F6,$BF,$F1,$16,$3E,$27,$FC,$49,$9F
	.dc.b	$99,$FF,$13,$6B,$EA,$7F,$C5,$1B,$FB,$9F,$F1,$57,$0F,$C2,$7E,$21
	.dc.b	$FE,$2D,$E4,$F8,$EF,$CB,$6E,$5F,$D7,$DF,$F2,$0F,$27,$19,$FE,$2F
	.dc.b	$5D,$5F,$E3,$37,$E1,$7F,$8D,$E1,$E5,$7F,$8E,$E3,$E9,$7F,$8F,$E5
	.dc.b	$ED,$7F,$90,$E7,$F0,$DF,$E1,$DA,$7C,$3D,$7E,$41,$23,$FA,$7D,$EF
	.dc.b	$BF,$F9,$2C,$7F,$BA,$EB,$83,$FE,$55,$BE,$DB,$7F,$79,$C7,$C4,$FF
	.dc.b	$7B,$C3,$CC,$FF,$7D,$CB,$D4,$FF,$7F,$CF,$DC,$FF,$80,$74,$F8,$4F
	.dc.b	$F8,$0F,$5F,$8F,$AF,$E8,$B6,$FF,$39,$EA,$F5,$FC,$CD,$DB,$CD,$18
	.dc.b	$F8,$9C,$FC,$56,$BD,$CD,$FC,$57,$0F,$7B,$8F,$A3,$CB,$DF,$E7,$F1
	.dc.b	$3D,$3E,$07,$AF,$C4,$F6,$F8,$3E,$FF,$13,$E0,$F8,$5F,$0F,$C4,$FF
	.dc.b	$49,$E4,$FF,$24,$7B,$5F,$C9,$78,$FF,$CE,$3C,$FF,$E7,$CF,$85,$4E
	.dc.b	$67,$F3,$47,$F5,$9E,$66,$1D,$63,$E8,$65,$E1,$7E,$96,$9E,$57,$EA
	.dc.b	$6D,$E9,$7E,$B7,$07,$B5,$FB,$1C,$5F,$05,$FB,$5C,$9F,$15,$F8,$E4
	.dc.b	$3F,$2C,$7E,$64,$87,$F6,$AE,$8F,$AB,$BF,$F1,$4E,$9F,$B2,$FF,$16
	.dc.b	$EA,$FB,$AF,$F1,$8E,$BF,$C0,$BF,$C6,$BB,$3F,$0A,$FF,$1C,$ED,$FC
	.dc.b	$4B,$FC,$7B,$BB,$ED,$BF,$C9,$3B,$FF,$23,$9F,$F7,$7E,$3F,$C8,$C7
	.dc.b	$F9,$0A,$B8,$CF,$FA,$9E,$FA,$BF,$EA,$9C,$3C,$2F,$FA,$AF,$1F,$2B
	.dc.b	$FE,$AD,$CB,$D2,$FF,$AB,$F3,$F6,$BF,$EB,$1D,$3F,$22,$FF,$AC,$FB
	.dc.b	$FF,$D8,$7E,$3F,$F2,$48,$F8,$BD,$1F,$BB,$7F,$80,$9F,$E1,$0F,$1F
	.dc.b	$F9,$23,$FE,$AF,$F1,$FF,$92,$3F,$EB,$1F,$1F,$F9,$23,$FE,$B3,$F1
	.dc.b	$FF,$92,$3F,$EB,$5F,$1F,$F9,$23,$FE,$B7,$F1,$FF,$92,$3F,$B3,$E5
	.dc.b	$FC,$83,$FB,$83,$F9,$19,$FD,$2F,$2F,$E5,$11,$C2,$3F,$C4,$73,$D1
	.dc.b	$FF,$12,$D7,$81,$FF,$13,$DF,$91,$FF,$14,$E1,$E8,$7F,$C5,$78,$FB
	.dc.b	$1F,$71,$F0,$31,$F8,$4F,$F8,$C3,$4F,$8E,$7C,$AE,$4F,$F3,$CB,$FC
	.dc.b	$82,$3F,$DA,$5A,$73,$AF,$F4,$15,$F7,$7F,$D0,$98,$78,$DF,$F4,$36
	.dc.b	$3E,$77,$FD,$11,$97,$AD,$FF,$45,$67,$EF,$7E,$0E,$3F,$13,$3F,$C5
	.dc.b	$C7,$FA,$9F,$4F,$F3,$64,$FE,$35,$E0,$7F,$70,$AE,$8F,$F7,$2B,$F0
	.dc.b	$3F,$DD,$30,$F2,$3F,$DD,$B1,$F4,$3F,$DE,$32,$F6,$3F,$DE,$B3,$F8
	.dc.b	$18,$FC,$53,$F5,$BF,$05,$FE,$6A,$7F,$9A,$F8,$FF,$C7,$08,$FF,$31
	.dc.b	$F1,$FF,$8D,$BF,$CC,$FC,$7F,$E3,$6F,$F3,$5F,$1F,$F8,$DB,$FC,$DF
	.dc.b	$C7,$FE,$36,$FF,$39,$F1,$FF,$8D,$BF,$CE,$FC,$7F,$E3,$6F,$F3,$DF
	.dc.b	$1F,$F8,$DB,$FC,$63,$5F,$F0,$8F,$91,$FD,$27,$C8,$FE,$33,$E1,$DF
	.dc.b	$23,$F8,$D7,$3F,$F7,$0C,$3F,$A3,$79,$1F,$C6,$5F,$E3,$9E,$47,$F1
	.dc.b	$97,$F8,$F7,$91,$FC,$65,$FE,$41,$E4,$7F,$19,$7F,$91,$63,$FC,$65
	.dc.b	$FE,$49,$E4,$7F,$19,$7F,$93,$79,$1F,$C6,$5F,$E5,$1C,$DF,$19,$FE
	.dc.b	$32,$7C,$D2,$67,$FD,$2D,$FC,$79,$FD,$E0,$FD,$A1,$FD,$F0,$FD,$A1
	.dc.b	$FE,$00,$3F,$68,$7F,$81,$0F,$DA,$1F,$E0,$83,$F6,$87,$C7,$3F,$B2
	.dc.b	$1F,$B5,$3F,$21,$BF,$E3,$A3,$F2,$1E,$2F,$F6,$D8,$C1,$FE,$16,$3F
	.dc.b	$23,$3F,$E8,$3F,$05,$DF,$F9,$28,$C7,$7F,$E4,$9D,$77,$FE,$49,$E1
	.dc.b	$DF,$F9,$27,$97,$7F,$E4,$9E,$9D,$FF,$92,$D3,$F6,$E7,$EE,$3B,$FF
	.dc.b	$25,$E0,$7E,$9B,$5F,$F0,$23,$6F,$F8,$71,$FE,$A9,$8F,$4B,$8A,$F6
	.dc.b	$BF,$03,$6F,$E7,$53,$AA,$E4,$47,$04,$8D,$92,$79,$10,$F0,$39,$69
	.dc.b	$9B,$F4,$CF,$38,$E1,$7E,$47,$64,$70,$48,$F7,$9C,$63,$4E,$2B,$C1
	.dc.b	$3B,$9F,$81,$39,$7B,$B5,$E2,$68,$97,$F0,$6B,$F0,$44,$80,$BF,$A7
	.dc.b	$E4,$3E,$1F,$FC,$3E,$49,$FF,$1A,$7C,$4F,$F8,$8B,$E2,$7F,$C4,$7E
	.dc.b	$BF,$F8,$93,$E2,$7F,$C7,$5F,$C0,$2F,$6E,$FC,$3E,$3D,$DC,$FC,$3E
	.dc.b	$BD,$FD,$FC,$3F,$0F,$81,$E3,$F0,$FC,$BE,$0F,$9F,$C3,$FF,$50,$F2
	.dc.b	$7F,$81,$F1,$7E,$85,$AF,$F8,$73,$A5,$D3,$EC,$E9,$70,$B0,$E9,$7F
	.dc.b	$A9,$9C,$E5,$E1,$8C,$53,$4F,$34,$63,$6F,$53,$AE,$0F,$7F,$0F,$C1
	.dc.b	$8F,$F2,$C7,$1F,$C5,$7E,$39,$FE,$5B,$E6,$F9,$4F,$CF,$F8,$7F,$B1
	.dc.b	$BF,$DD,$D7,$EB,$4F,$DC,$3E,$CE,$FF,$85,$9D,$34,$E2,$48,$DF,$57
	.dc.b	$E2,$23,$15,$D2,$7F,$0A,$DA,$E7,$5B,$9E,$27,$E1,$3A,$BF,$3B,$1F
	.dc.b	$EF,$08,$FF,$78,$D7,$F8,$99,$B7,$FD,$71,$B3,$B7,$E9,$46,$5B,$19
	.dc.b	$7F,$7A,$70,$38,$4F,$57,$66,$27,$93,$C0,$C9,$E5,$8E,$0E,$48,$E0
	.dc.b	$91,$C9,$B3,$E3,$6C,$F7,$8C,$4E,$C8,$E0,$91,$B2,$1E,$38,$DB,$83
	.dc.b	$DC,$66,$38,$AF,$49,$E7,$FB,$23,$64,$30,$43,$DB,$5D,$9E,$87,$73
	.dc.b	$57,$EC,$72,$43,$A1,$CA,$FC,$6E,$23,$83,$67,$B3,$67,$6F,$68,$CB
	.dc.b	$63,$2F,$53,$E2,$DF,$B9,$C1,$E3,$38,$1E,$93,$27,$73,$15,$E4,$76
	.dc.b	$71,$36,$60,$E0,$65,$E0,$E8,$8E,$09,$5E,$A8,$F2,$CF,$9E,$73,$5E
	.dc.b	$C6,$CF,$92,$B8,$46,$23,$C1,$18,$24,$F1,$6C,$71,$8C,$11,$C1,$0E
	.dc.b	$6E,$5F,$C4,$5F,$EF,$A7,$F8,$24,$7E,$E8,$FD,$FB,$8A,$FE,$00,$EE
	.dc.b	$BF,$80,$46,$27,$86,$3F,$C1,$1F,$E2,$51,$92,$57,$32,$57,$F4,$B9
	.dc.b	$57,$F6,$1C,$7F,$BE,$F0,$BE,$1B,$FA,$29,$2B,$F9,$7E,$9F,$D2,$3E
	.dc.b	$1F,$FA,$4F,$C2,$FF,$4A,$F8,$5F,$E9,$7F,$11,$FD,$33,$E1,$BF,$A6
	.dc.b	$FC,$4F,$F4,$EF,$87,$FE,$9F,$F1,$5F,$D4,$3E,$23,$FA,$DE,$9E,$2E
	.dc.b	$2F,$E8,$5F,$C0,$1F,$C6,$78,$BF,$A2,$CB,$91,$2F,$89,$E3,$5A,$9C
	.dc.b	$F9,$18,$C3,$9F,$C4,$DC,$79,$62,$F8,$7C,$61,$D3,$E2,$6C,$9B,$38
	.dc.b	$BC,$E3,$AC,$F7,$8F,$AC,$FE,$A7,$AF,$BD,$FE,$AB,$BF,$C3,$7F,$AB
	.dc.b	$F0,$FC,$77,$FA,$CF,$1F,$CB,$7E,$69,$FD,$73,$9B,$E8,$3F,$D4,$4F
	.dc.b	$BF,$BF,$E0,$DF,$BE,$FD,$C6,$F4,$FE,$1A,$31,$1F,$8C,$8C,$11,$D9
	.dc.b	$0F,$A4,$7F,$8E,$BA,$7F,$07,$5E,$2E,$2F,$9D,$D3,$F8,$5A,$F3,$9C
	.dc.b	$38,$FF,$BC,$F4,$7E,$12,$3F,$BA,$F4,$7E,$11,$FE,$EF,$D1,$F8,$47
	.dc.b	$FB,$CF,$47,$E1,$1F,$EF,$7D,$1F,$31,$FF,$00,$74,$7D,$F6,$F2,$FD
	.dc.b	$37,$3C,$FF,$63,$A2,$46,$0E,$3D,$1F,$61,$E7,$D1,$D5,$F4,$6E,$CB
	.dc.b	$C5,$18,$CF,$CC,$EB,$4F,$53,$C3,$5F,$73,$CB,$6F,$F1,$5F,$4F,$C7
	.dc.b	$E5,$FF,$0E,$65,$F1,$EB,$ED,$70,$FE,$6B,$E2,$7F,$1E,$48,$F7,$FB
	.dc.b	$BF,$92,$F8,$3F,$C9,$FC,$27,$E5,$3D,$0F,$CA,$FC,$2F,$E5,$BD,$2F
	.dc.b	$CB,$FC,$37,$E6,$3D,$4F,$CC,$FC,$3F,$E6,$BD,$6F,$CD,$F8,$1F,$9C
	.dc.b	$F5,$FF,$3B,$F1,$1F,$9E,$F6,$7F,$3B,$C1,$FC,$23,$C0,$FD,$23,$F4
	.dc.b	$23,$6F,$D2,$AE,$6F,$D4,$C3,$BB,$F5,$B2,$F1,$BF,$63,$4F,$3B,$F6
	.dc.b	$B6,$F5,$BF,$73,$83,$F5,$FC,$3F,$06,$3F,$15,$9B,$8B,$F8,$2F,$87
	.dc.b	$FE,$90,$87,$C4,$FE,$D8,$95,$FE,$45,$FE,$01,$FE,$0E,$FE,$4D,$7B
	.dc.b	$BF,$BE,$7F,$BC,$46,$FD,$EC,$7C,$56,$7E,$03,$5F,$15,$BF,$82,$E1
	.dc.b	$F1,$5C,$7E,$13,$97,$A5,$CF,$E1,$7A,$7C,$57,$5F,$87,$F9,$7A,$FA
	.dc.b	$BC,$DF,$C9,$BE,$CF,$3B,$4F,$9E,$B9,$3E,$9C,$3B,$3E,$BC,$BC,$4F
	.dc.b	$B7,$4F,$33,$EF,$DB,$D9,$BF,$B7,$D6,$F7,$71,$7B,$FC,$7F,$8B,$7F
	.dc.b	$33,$93,$E5,$F9,$1F,$CB,$BA,$3E,$87,$A3,$FC,$A7,$A3,$E8,$5F,$E4
	.dc.b	$8F,$AC,$7E,$CB,$C7,$FC,$93,$FB,$4F,$1F,$F2,$4F,$ED,$BC,$7F,$C9
	.dc.b	$3F,$B8,$F1,$FF,$24,$FE,$EB,$C7,$FC,$93,$FB,$CF,$1F,$F2,$4F,$F1
	.dc.b	$1C,$BF,$CD,$99,$7F,$B7,$23,$47,$E4,$4E,$47,$E4,$4E,$C7,$E4,$4F
	.dc.b	$11,$F9,$13,$CC,$7E,$44,$F5,$1E,$C3,$F1,$FE,$07,$C2,$DF,$E2,$47
	.dc.b	$F9,$AF,$6F,$EA,$7B,$FC,$AC,$3E,$33,$F3,$67,$F2,$1C,$1F,$45,$FC
	.dc.b	$61,$F5,$4F,$CA,$77,$7E,$31,$FC,$B7,$77,$E3,$1F,$CC,$77,$7E,$31
	.dc.b	$FC,$D7,$77,$E3,$1F,$CE,$77,$7E,$33,$2F,$AD,$DD,$F8,$C7,$EC,$77
	.dc.b	$7E,$32,$BF,$41,$DD,$F8,$C7,$F4,$5D,$DF,$8C,$7F,$49,$DD,$F8,$C7
	.dc.b	$F4,$DD,$DF,$8C,$7F,$51,$DD,$F8,$C7,$F5,$9D,$BF,$8C,$7F,$5D,$C9
	.dc.b	$F0,$CF,$8A,$FC,$73,$FB,$6F,$6F,$CB,$69,$F8,$B3,$F7,$BE,$0F,$F0
	.dc.b	$27,$F7,$FE,$17,$F0,$67,$F8,$0F,$87,$FC,$29,$FE,$07,$E2,$7F,$0E
	.dc.b	$7F,$82,$F8,$BF,$C4,$9F,$E0,$FE,$37,$F1,$67,$F8,$4F,$8F,$F1,$F3
	.dc.b	$FF,$55,$F7,$7F,$1C,$9C,$9F,$8C,$38,$9F,$8C,$3A,$9F,$8C,$3C,$27
	.dc.b	$E3,$0F,$29,$F8,$C3,$D2,$7E,$30,$F6,$9E,$FE,$4F,$D6,$6B,$FC,$77
	.dc.b	$C7,$FE,$35,$DB,$F2,$6B,$FC,$26,$FE,$2D,$FE,$31,$DB,$F8,$B7,$F8
	.dc.b	$D7,$8D,$FC,$59,$FC,$59,$FC,$0C,$FC,$59,$FC,$1C,$FC,$59,$FC,$2C
	.dc.b	$FC,$59,$FC,$3C,$FC,$59,$FD,$94,$FE,$2C,$7F,$6E,$E9,$F9,$0F,$F8
	.dc.b	$BF,$C7,$75,$FE,$4E,$3F,$96,$76,$7E,$5D,$FE,$5D,$DB,$F9,$B7,$F9
	.dc.b	$87,$77,$E7,$DF,$E6,$5D,$FE,$97,$D6,$7E,$30,$FD,$49,$F8,$C3,$FB
	.dc.b	$FE,$9F,$C6,$35,$FE,$81,$B7,$E2,$FC,$8F,$ED,$15,$F3,$78,$BF,$C8
	.dc.b	$FD,$3F,$8B,$E2,$FE,$35,$D3,$F8,$B7,$F8,$E7,$4F,$E2,$DF,$E3,$DD
	.dc.b	$3F,$8B,$7F,$90,$74,$FE,$2D,$FE,$45,$D3,$F8,$B7,$F9,$27,$4F,$F8
	.dc.b	$9F,$3F,$C4,$9F,$D5,$BC,$3F,$E3,$6F,$F5,$8F,$13,$F8,$FB,$FD,$6B
	.dc.b	$C5,$FE,$46,$FF,$5C,$F1,$BF,$93,$BF,$D7,$BC,$7F,$E5,$6F,$F6,$0F
	.dc.b	$23,$FC,$00,$FB,$4F,$D4,$9F,$D9,$38,$FF,$9C,$BF,$88,$3E,$39,$0F
	.dc.b	$EC,$1C,$5F,$B7,$7F,$B1,$78,$1F,$BB,$7F,$7A,$7E,$28,$FE,$02,$7E
	.dc.b	$28,$FE,$0A,$7E,$28,$FE,$12,$7E,$28,$FE,$1A,$7E,$28,$FE,$B8,$7E
	.dc.b	$24,$86,$0F,$EE,$5D,$7F,$89,$DB,$FC,$ED,$D8,$EC,$FC,$54,$63,$B3
	.dc.b	$F1,$4E,$BB,$3F,$14,$F0,$EC,$FC,$53,$CB,$BB,$F4,$CF,$4E,$FF,$D5
	.dc.b	$3F,$CD,$B3,$F8,$9D,$AF,$1D,$1F,$9D,$F8,$AF,$F0,$21,$1F,$F0,$57
	.dc.b	$C4,$FF,$81,$FD,$9F,$F0,$47,$C4,$7F,$82,$7E,$23,$FC,$15,$EE,$7F
	.dc.b	$82,$FE,$23,$FD,$13,$F1,$0F,$7E,$FC,$06,$23,$3F,$03,$AF,$88,$DF
	.dc.b	$C1,$F0,$F8,$8E,$3F,$0B,$CB,$E2,$3F,$D6,$9C,$BF,$E9,$BE,$F7,$8B
	.dc.b	$9E,$4F,$4C,$FF,$A7,$F4,$61,$CE,$73,$7E,$EF,$39,$4C,$FC,$55,$A3
	.dc.b	$FD,$59,$0D,$7D,$11,$9A,$E3,$B7,$BB,$7F,$81,$3F,$E2,$FE,$1F,$88
	.dc.b	$FF,$8C,$78,$FE,$43,$F2,$8F,$F1,$A7,$37,$CD,$7F,$CC,$3D,$E8,$C6
	.dc.b	$FF,$84,$7F,$C6,$DC,$FC,$AB,$FC,$75,$7D,$9F,$F1,$E6,$1E,$27,$FD
	.dc.b	$21,$AF,$D1,$7D,$0B,$3F,$5C,$6B,$5F,$82,$F0,$DB,$E4,$3C,$B8,$3E
	.dc.b	$73,$FC,$90,$9A,$7C,$FE,$2F,$AF,$1D,$B8,$FE,$DB,$E0,$E4,$FB,$EF
	.dc.b	$8B,$3F,$C1,$BE,$4E,$1F,$46,$1F,$A2,$D7,$D8,$FE,$93,$6F,$80,$FE
	.dc.b	$9B,$7F,$8B,$C7,$F6,$F8,$BE,$46,$3F,$DB,$F8,$FE,$EF,$C3,$FE,$48
	.dc.b	$91,$F0,$BE,$0F,$F2,$BD,$5F,$96,$F7,$3F,$2F,$F0,$9F,$98,$F7,$BF
	.dc.b	$33,$F0,$BF,$9A,$F8,$0F,$CD,$FC,$37,$E7,$3E,$0B,$F3,$BF,$0F,$F9
	.dc.b	$EF,$84,$FF,$73,$FF,$00,$FF,$BD,$78,$BF,$1C,$FF,$AD,$77,$FF,$61
	.dc.b	$2F,$88,$FE,$97,$C4,$AF,$CD,$D7,$E3,$5A,$FF,$07,$E5,$FF,$7D,$5F
	.dc.b	$18,$FC,$93,$D6,$33,$1E,$33,$EF,$E3,$E7,$7F,$03,$9F,$AD,$FC,$1E
	.dc.b	$BE,$F7,$F0,$BB,$FC,$37,$F0,$FC,$3F,$BA,$E2,$F9,$31,$F8,$CE,$3F
	.dc.b	$E5,$FA,$B9,$3E,$8C,$FF,$A3,$7C,$1F,$AA,$FF,$67,$E6,$FB,$3C,$1F
	.dc.b	$5F,$9F,$EE,$BF,$67,$A3,$F0,$2F,$ED,$7A,$BF,$31,$7F,$9A,$F0,$7F
	.dc.b	$2F,$C7,$FB,$DF,$07,$F2,$EF,$EF,$FC,$1F,$CB,$BF,$C0,$7C,$1F,$CB
	.dc.b	$BF,$E4,$4B,$FE,$62,$FE,$5D,$FD,$1C,$70,$C7,$C1,$5D,$1F,$15,$F8
	.dc.b	$1F,$26,$1F,$BF,$BF,$F5,$67,$87,$FC,$81,$FF,$57,$78,$9F,$C9,$1F
	.dc.b	$F5,$87,$8B,$FB,$87,$F7,$E4,$3F,$7F,$C1,$F1,$B8,$BE,$53,$B8,$FF
	.dc.b	$5B,$4F,$37,$FD,$71,$5D,$DF,$F5,$D5,$F8,$DF,$8C,$7E,$B4,$FD,$FE
	.dc.b	$5E,$B7,$F8,$DE,$7E,$F7,$F8,$EE,$9F,$1B,$4F,$E4,$87,$E4,$35,$FE
	.dc.b	$5D,$E2,$E6,$B6,$7E,$F4,$FE,$17,$7D,$75,$FE,$57,$8F,$C7,$9F,$E5
	.dc.b	$3B,$79,$70,$FE,$BB,$9F,$A5,$FE,$BF,$A7,$B5,$FE,$C3,$AF,$C1,$7F
	.dc.b	$90,$5F,$C7,$F1,$BF,$AB,$3F,$D5,$4F,$95,$1A,$7F,$DB,$9C,$7F,$8E
	.dc.b	$7F,$C6,$DC,$7F,$8E,$7F,$C7,$1C,$7F,$8E,$7F,$9E,$F8,$7F,$C4,$9F
	.dc.b	$E4,$FC,$9F,$C5,$9F,$E5,$3E,$27,$F1,$A7,$F9,$5F,$8B,$FC,$79,$FE
	.dc.b	$5B,$E3,$7F,$A3,$F9,$7F,$A6,$7B,$CF,$7B,$E4,$72,$FF,$4E,$E8,$FA
	.dc.b	$0F,$F5,$0E,$7F,$91,$A7,$F4,$DC,$BE,$C1,$F6,$8F,$E7,$3A,$F9,$EB
	.dc.b	$F5,$DB,$7A,$DF,$D8,$6F,$EF,$7F,$65,$C1,$F8,$BE,$AF,$D9,$F1,$FC
	.dc.b	$89,$FE,$47,$8F,$F4,$2E,$BF,$DD,$F2,$7F,$67,$EC,$FE,$47,$3F,$D1
	.dc.b	$BC,$8F,$E4,$6F,$F4,$8F,$23,$F9,$1B,$FD,$2B,$C8,$FE,$46,$FF,$4C
	.dc.b	$F2,$3F,$91,$BF,$D3,$BC,$8F,$E4,$6F,$F2,$4F,$87,$FE,$4B,$E4,$FF
	.dc.b	$20,$E9,$FC,$71,$F9,$0F,$0B,$F9,$1F,$93,$FC,$76,$FF,$93,$DF,$27
	.dc.b	$A1,$FC,$70,$F0,$1F,$C7,$0F,$21,$FC,$70,$F4,$1F,$C7,$0F,$61,$FC
	.dc.b	$70,$FE,$84,$7C,$32,$1F,$B4,$3F,$1B,$5F,$D3,$7C,$4F,$CA,$3F,$D3
	.dc.b	$FC,$0F,$CC,$3F,$D4,$7C,$1F,$CE,$3F,$D4,$FC,$2F,$D0,$3F,$C2,$2F
	.dc.b	$EE,$F1,$7F,$19,$7F,$B2,$1F,$AA,$3F,$58,$7F,$6F,$3F,$BB,$F8,$1F
	.dc.b	$8D,$5C,$AF,$03,$F1,$D1,$A3,$F8,$D9,$CA,$BF,$92,$3D,$8F,$EF,$C7
	.dc.b	$8A,$FC,$8F,$E3,$8F,$E3,$67,$F8,$64,$FE,$36,$7F,$86,$4F,$E3,$67
	.dc.b	$F8,$64,$FE,$7C,$7E,$D7,$C5,$FE,$BF,$F1,$1F,$D8,$89,$1F,$D5,$9A
	.dc.b	$FF,$62,$6D,$7F,$5C,$6D,$72,$9E,$04,$38,$16,$78,$36,$38,$3F,$B4
	.dc.b	$3F,$B5,$70,$7F,$92,$23,$FA,$66,$5D,$A7,$E5,$F0,$30,$C6,$68,$C4
	.dc.b	$F5,$37,$3A,$8C,$C7,$90,$98,$F9,$1B,$87,$92,$30,$47,$CF,$D3,$FB
	.dc.b	$5A,$E7,$39,$AE,$E7,$87,$0F,$26,$1E,$12,$46,$6B,$8E,$1D,$C9,$1C
	.dc.b	$8F,$19,$6F,$B5,$FB,$FE,$1B,$F1,$CF,$E3,$7C,$3F,$F0,$DC,$76,$AE
	.dc.b	$33,$E2,$7C,$18,$E6,$F9,$E3,$E2,$9E,$D8,$F8,$AF,$84,$72,$9D,$46
	.dc.b	$32,$F8,$59,$7F,$8A,$89,$5F,$E1,$EF,$53,$FC,$5B,$F0,$FF,$E2,$EF
	.dc.b	$81,$FF,$17,$FA,$FF,$E3,$0F,$81,$FF,$18,$FB,$3F,$E3,$2F,$81,$FF
	.dc.b	$19,$FB,$7F,$E3,$4F,$81,$FF,$1A,$FB,$BF,$E3,$6F,$81,$FF,$1B,$F3
	.dc.b	$FF,$8E,$3E,$07,$FC,$73,$F0,$1F,$CE,$39,$7F,$8A,$F0,$FE,$C1,$BC
	.dc.b	$9F,$E8,$8E,$9D,$CF,$F8,$CA,$F9,$BF,$E3,$4C,$7B,$BF,$E3,$6C,$FC
	.dc.b	$6F,$F8,$E3,$5F,$3B,$FE,$3A,$DF,$D6,$FF,$8F,$38,$7D,$EF,$C1,$3F
	.dc.b	$C7,$FC,$9F,$11,$72,$FE,$27,$A3,$E9,$6D,$F3,$F9,$FE,$69,$23,$35
	.dc.b	$F5,$09,$BF,$D4,$E8,$FB,$33,$A8,$C5,$FF,$96,$BA,$7F,$0B,$D7,$FC
	.dc.b	$43,$0F,$F2,$DC,$6F,$1E,$D1,$F8,$42,$1F,$7F,$0D,$11,$FF,$3B,$47
	.dc.b	$EA,$9F,$F6,$89,$F9,$72,$E1,$F8,$86,$C6,$DF,$DF,$98,$9E,$2D,$29
	.dc.b	$CC,$D9,$0C,$12,$7B,$39,$6B,$99,$F0,$3F,$EC,$4E,$2F,$E1,$8F,$E1
	.dc.b	$79,$3F,$BF,$CE,$AB,$89,$1C,$12,$FC,$B5,$D5,$B3,$B6,$CE,$E7,$D7
	.dc.b	$39,$6C,$65,$7C,$03,$E1,$BD,$53,$99,$F8,$87,$C6,$AE,$4D,$8D,$3C
	.dc.b	$2F,$C2,$47,$BB,$5E,$0D,$7A,$9F,$5B,$2F,$C3,$EF,$FC,$2A,$3F,$9D
	.dc.b	$1A,$3B,$1D,$08,$74,$29,$B3,$E3,$18,$38,$9B,$26,$3F,$05,$D9,$93
	.dc.b	$96,$1F,$01,$B8,$7C,$07,$66,$30,$F6,$3D,$8E,$3E,$83,$4F,$C0,$DF
	.dc.b	$FA,$E7,$4F,$C0,$DF,$0B,$E6,$71,$21,$82,$4F,$8D,$F5,$E3,$E9,$8C
	.dc.b	$46,$B1,$F3,$46,$08,$E0,$95,$E1,$8F,$C8,$67,$F4,$AB,$FD,$8E,$FF
	.dc.b	$72,$7D,$C7,$CB,$3B,$4F,$1A,$E7,$19,$AE,$F3,$F8,$1B,$F3,$57,$F0
	.dc.b	$0A,$FD,$F1,$FC,$03,$7F,$E1,$6D,$DF,$F8,$A5,$FF,$01,$8D,$9F,$C2
	.dc.b	$8F,$EE,$E4,$30,$6A,$79,$B6,$79,$BD,$4F,$E3,$75,$E3,$9D,$9F,$DC
	.dc.b	$F8,$3F,$B7,$7C,$17,$FB,$D8,$95,$FE,$EA,$F8,$3F,$E2,$35,$FE,$F3
	.dc.b	$8F,$F1,$7F,$F0,$0D,$51,$9D,$12,$F7,$B8,$E6,$BE,$3A,$E1,$E7,$EF
	.dc.b	$DC,$38,$FC,$2D,$C5,$F0,$37,$E1,$B1,$F0,$79,$F8,$7D,$7C,$2E,$FA
	.dc.b	$B8,$7C,$37,$1F,$88,$FD,$76,$DE,$6E,$0F,$F4,$9F,$67,$F7,$CE,$77
	.dc.b	$4E,$EB,$CD,$CB,$EB,$E0,$65,$D6,$73,$8F,$AF,$8B,$DF,$AF,$8E,$72
	.dc.b	$7C,$6D,$FD,$0E,$27,$E2,$F0,$FB,$78,$BD,$DC,$7E,$FE,$4F,$81,$CB
	.dc.b	$F0,$73,$F8,$9C,$0E,$6F,$AB,$9F,$AC,$FF,$1B,$F3,$FF,$46,$E8,$FE
	.dc.b	$C1,$1F,$81,$D7,$FA,$BB,$F8,$2E,$9F,$C7,$71,$FE,$0B,$2F,$C3,$CF
	.dc.b	$E2,$4D,$1F,$E6,$93,$B5,$7E,$ED,$FC,$94,$FD,$FE,$CF,$E0,$11,$DF
	.dc.b	$3F,$C7,$5F,$4A,$ED,$3D,$0E,$D7,$E4,$25,$7E,$9F,$1F,$C9,$CF,$C7
	.dc.b	$86,$5F,$02,$FF,$1D,$1F,$0E,$7E,$67,$27,$E4,$E7,$EC,$6B,$F2,$E7
	.dc.b	$3B,$FC,$F7,$7C,$3F,$51,$F6,$DB,$8B,$ED,$46,$38,$FF,$00,$EB,$93
	.dc.b	$F0,$CF,$EC,$B2,$FD,$07,$6F,$F8,$05,$BC,$FF,$90,$7F,$15,$D1,$F9
	.dc.b	$4A,$F3,$F4,$FE,$61,$F4,$F5,$7E,$71,$F5,$F5,$FE,$87,$6F,$44,$79
	.dc.b	$B7,$F6,$3E,$8E,$0F,$C3,$3F,$B7,$3E,$47,$17,$E1,$B8,$FF,$DB,$DC
	.dc.b	$4F,$88,$FD,$61,$2B,$E4,$FA,$1F,$AF,$F6,$FF,$61,$E8,$FE,$C7,$DC
	.dc.b	$FD,$97,$A5,$FB,$3F,$77,$F6,$9E,$9F,$ED,$FF,$80,$5E,$FD,$F8,$4C
	.dc.b	$78,$39,$F8,$5D,$7C,$2E,$FE,$0F,$87,$C2,$FC,$5F,$17,$FD,$BD,$C5
	.dc.b	$F8,$C7,$FC,$58,$FE,$9F,$34,$F1,$8C,$5F,$57,$58,$F8,$5E,$19,$F9
	.dc.b	$9F,$C2,$EB,$EA,$7F,$0F,$BF,$B9,$FC,$4F,$0F,$F0,$FE,$2F,$89,$1F
	.dc.b	$84,$E3,$FD,$67,$B5,$FB,$FD,$5F,$0B,$F0,$49,$5E,$7F,$3B,$E2,$7C
	.dc.b	$27,$C5,$F8,$4F,$8D,$87,$C7,$F8,$4F,$91,$F0,$DF,$27,$E1,$7E,$57
	.dc.b	$C3,$FC,$BF,$86,$F9,$9F,$11,$F3,$7E,$1F,$E7,$7C,$4F,$CF,$F8,$8F
	.dc.b	$E4,$DE,$B7,$F8,$F3,$CD,$FE,$60,$BE,$1F,$FA,$E7,$17,$EE,$23,$0F
	.dc.b	$F1,$17,$87,$77,$AE,$FA,$CF,$CF,$C7,$C2,$FD,$1C,$FC,$AF,$D3,$D7
	.dc.b	$D2,$FD,$5D,$FD,$AF,$D7,$E1,$F8,$2F,$C3,$3E,$D7,$27,$C6,$7F,$CE
	.dc.b	$BC,$5F,$CA,$7C,$5F,$CA,$47,$EE,$B9,$FE,$8F,$07,$F3,$1E,$8F,$CA
	.dc.b	$4F,$EF,$BC,$5F,$CA,$3F,$C0,$3C,$5F,$CA,$3F,$C0,$BC,$5F,$CA,$77
	.dc.b	$FE,$0E,$3F,$8C,$F6,$3C,$0F,$C4,$46,$3C,$1F,$C6,$3A,$F0,$BF,$28
	.dc.b	$FF,$96,$4F,$CA,$1F,$EE,$4F,$67,$F3,$2F,$E5,$7C,$4F,$CE,$BF,$97
	.dc.b	$F1,$7F,$42,$FE,$67,$C6,$F1,$BE,$53,$F2,$67,$A4,$FC,$99,$ED,$3F
	.dc.b	$26,$7E,$D0,$F8,$84,$3E,$48,$F5,$99,$FE,$52,$3E,$79,$0F,$83,$E2
	.dc.b	$7D,$47,$E1,$F3,$FD,$87,$E2,$F4,$7D,$C7,$E3,$F4,$FE,$01,$F9,$3D
	.dc.b	$5F,$84,$7E,$5F,$5F,$E2,$1F,$9B,$D9,$F9,$47,$F2,$BD,$FF,$92,$E9
	.dc.b	$FC,$AB,$98,$FD,$E4,$F1,$9F,$F0,$46,$8B,$ED,$18,$C3,$C4,$EB,$1F
	.dc.b	$33,$C3,$2F,$53,$CB,$3F,$73,$D3,$4F,$C8,$3D,$BB,$7F,$77,$A7,$F8
	.dc.b	$6B,$AF,$F1,$FD,$DF,$93,$7F,$81,$47,$D2,$3F,$A2,$F5,$FE,$41,$FE
	.dc.b	$8F,$D7,$F9,$07,$FA,$4F,$5F,$DE,$7F,$1F,$C9,$F8,$27,$F2,$3C,$BF
	.dc.b	$86,$7F,$27,$CD,$F8,$A7,$FA,$56,$5F,$C8,$9F,$ED,$EF,$FB,$B3,$D6
	.dc.b	$FF,$0A,$C7,$F2,$3E,$9F,$CB,$ED,$F2,$3A,$BF,$36,$FC,$AE,$BF,$D0
	.dc.b	$3F,$33,$B3,$F9,$15,$FF,$5E,$F2,$3F,$91,$3F,$D8,$3C,$8F,$E4,$4F
	.dc.b	$F6,$2F,$23,$E0,$BF,$C8,$4F,$C8,$79,$2F,$0F,$F9,$17,$37,$CC,$DE
	.dc.b	$D7,$EF,$CF,$D7,$F9,$3F,$C8,$5F,$E7,$55,$FC,$85,$FE,$79,$E4,$FF
	.dc.b	$21,$7F,$9F,$79,$3F,$C8,$5F,$E8,$1E,$4F,$F2,$17,$FA,$17,$93,$FC
	.dc.b	$85,$FE,$34,$7E,$38,$FF,$16,$7A,$7F,$DC,$72,$FF,$1E,$C7,$F1,$FF
	.dc.b	$17,$F2,$0F,$E5,$8E,$67,$F1,$F3,$B9,$FC,$7C,$F1,$9F,$C7,$CF,$39
	.dc.b	$FC,$7C,$F5,$9F,$C7,$CF,$EA,$D9,$FF,$63,$EE,$F8,$91,$FB,$5F,$13
	.dc.b	$F6,$8F,$EE,$3C,$6F,$F3,$2F,$81,$F9,$08,$FC,$91,$FB,$F3,$E9,$1F
	.dc.b	$C0,$CF,$AC,$7F,$07,$3E,$D1,$FC,$2C,$FE,$3B,$CB,$F8,$0C,$BF,$AB
	.dc.b	$73,$FE,$11,$FE,$B1,$D1,$F8,$87,$FA,$D7,$4F,$F1,$EC,$7F,$C7,$98
	.dc.b	$7E,$43,$DA,$C6,$7F,$DC,$23,$F8,$F1,$FD,$C0,$FE,$3C,$7F,$70,$3F
	.dc.b	$8F,$1F,$DC,$0F,$27,$57,$F9,$A7,$AB,$F1,$EF,$F9,$AF,$AB,$F1,$EF
	.dc.b	$F9,$B7,$AB,$F1,$EF,$F9,$BF,$7F,$EE,$3D,$5F,$8F,$C3,$E5,$75,$7C
	.dc.b	$CD,$3F,$8F,$BF,$3E,$3E,$89,$F4,$CF,$E7,$FD,$5F,$8F,$7F,$A0,$F5
	.dc.b	$7E,$3D,$FE,$87,$D5,$F8,$F7,$FA,$2F,$57,$E3,$DF,$E8,$FD,$5F,$8F
	.dc.b	$7F,$B5,$1F,$8F,$3F,$CC,$7D,$5F,$8F,$7F,$99,$37,$27,$97,$FD,$C2
	.dc.b	$78,$9F,$C8,$0F,$C8,$9F,$CA,$8F,$C8,$9F,$CB,$8F,$C8,$9F,$CC,$8F
	.dc.b	$C8,$9F,$CD,$8F,$C8,$9F,$CE,$8F,$82,$7F,$3E,$3F,$A0,$6D,$FC,$7D
	.dc.b	$FF,$58,$6B,$F2,$EB,$CB,$F1,$1F,$E0,$A2,$4F,$F8,$17,$D9,$FF,$06
	.dc.b	$F9,$7F,$E0,$EF,$5B,$FC,$1F,$F1,$1F,$E1,$0F,$88,$FF,$08,$FB,$DF
	.dc.b	$E1,$2F,$88,$FF,$09,$FC,$07,$F8,$53,$E2,$3F,$C2,$BF,$05,$FE,$16
	.dc.b	$F8,$8F,$F0,$BF,$C2,$7F,$86,$3E,$23,$FC,$33,$F0,$DF,$2F,$93,$FD
	.dc.b	$91,$E2,$3C,$3F,$F6,$13,$78,$FF,$74,$7F,$86,$6B,$93,$FE,$1B,$C3
	.dc.b	$B3,$FE,$1D,$CB,$C4,$FF,$87,$F4,$F3,$3F,$E2,$1D,$BD,$4F,$F8,$8F
	.dc.b	$83,$DC,$FF,$89,$78,$BF,$C5,$BC,$7F,$12,$3F,$C5,$5C,$9F,$D2,$39
	.dc.b	$7E,$4C,$7F,$8A,$E3,$5B,$FF,$8C,$AB,$E4,$CF,$F8,$BF,$97,$E4,$BF
	.dc.b	$E3,$1E,$5F,$92,$FF,$8C,$F9,$7E,$4B,$FE,$35,$E5,$F9,$2F,$F8,$DF
	.dc.b	$97,$E4,$BF,$E3,$9E,$5F,$92,$FF,$8E,$F9,$7F,$B7,$72,$FF,$B4,$F9
	.dc.b	$FE,$44,$FF,$90,$79,$BE,$3B,$FE,$43,$E6,$F8,$EF,$F9,$17,$9B,$E3
	.dc.b	$BF,$E4,$7E,$6F,$8E,$FF,$92,$79,$BE,$3B,$FE,$4B,$E6,$F8,$EF,$F9
	.dc.b	$37,$9B,$E3,$BF,$0C,$87,$F1,$FC,$9F,$0F,$FE,$58,$24,$7F,$94,$FE
	.dc.b	$23,$FC,$B7,$EE,$7F,$97,$3E,$1D,$3F,$EE,$7F,$E0,$1F,$F5,$E7,$07
	.dc.b	$FB,$BD,$7B,$FA,$9F,$F6,$AC,$7E,$C1,$E1,$F0,$19,$EB,$5F,$1F,$89
	.dc.b	$C3,$CF,$E0,$AD,$FF,$8A,$F0,$7C,$1E,$3C,$04,$7F,$B1,$67,$C9,$F0
	.dc.b	$99,$87,$C4,$5F,$86,$FC,$E6,$5F,$DE,$F4,$FC,$EF,$5E,$31,$ED,$5C
	.dc.b	$6F,$96,$3C,$FE,$27,$A6,$9D,$75,$ED,$B7,$7D,$FC,$1C,$BE,$1E,$1F
	.dc.b	$16,$FF,$27,$5F,$F3,$F7,$1F,$FA,$1B,$93,$E7,$C7,$FA,$0F,$97,$F3
	.dc.b	$58,$3C,$4F,$CC,$E1,$F8,$97,$ED,$D7,$DD,$3F,$34,$7E,$9C,$FC,$D1
	.dc.b	$FA,$B3,$F3,$47,$EB,$CF,$CD,$1F,$B3,$3F,$34,$7E,$DC,$FC,$D1,$FB
	.dc.b	$B3,$F3,$47,$F7,$AF,$0F,$E8,$78,$DF,$CA,$EF,$F3,$7C,$BF,$E9,$EE
	.dc.b	$FF,$CD,$BF,$EA,$0E,$FF,$CD,$BF,$EA,$2E,$FF,$CD,$BF,$EA,$4E,$FF
	.dc.b	$CD,$BF,$EA,$6E,$FF,$CD,$BF,$EA,$8E,$FF,$CD,$BF,$EA,$AE,$FF,$82
	.dc.b	$FF,$AB,$36,$FE,$05,$E3,$7F,$29,$D3,$E4,$6F,$F9,$EA,$D3,$FC,$A4
	.dc.b	$FE,$14,$7C,$83,$F5,$07,$C8,$3F,$58,$7C,$83,$F6,$07,$C8,$3F,$68
	.dc.b	$7C,$83,$F7,$07,$C8,$3F,$78,$7F,$24,$3F,$BC,$1F,$E9,$6E,$FF,$EA
	.dc.b	$3E,$C7,$F0,$68,$FF,$61,$78,$5F,$C2,$9F,$F6,$27,$87,$FC,$39,$FF
	.dc.b	$63,$78,$9F,$C4,$9F,$F6,$47,$8B,$FC,$59,$FF,$65,$78,$DF,$C6,$9F
	.dc.b	$F6,$67,$37,$F1,$E7,$FD,$9D,$E3,$FE,$29,$FF,$4A,$6B,$F1,$9F,$E0
	.dc.b	$5B,$E7,$5F,$E5,$31,$C0,$F8,$E7,$43,$F1,$C7,$80,$FC,$71,$E4,$3F
	.dc.b	$1C,$7A,$0F,$C7,$1E,$C3,$F1,$C7,$C0,$3F,$1C,$7F,$B8,$FC,$5F,$E4
	.dc.b	$EF,$99,$F8,$DB,$FF,$BA,$B6,$F8,$CF,$FB,$B3,$6F,$8C,$FF,$BB,$B6
	.dc.b	$F8,$CF,$FB,$C3,$6F,$8C,$FF,$BC,$B6,$F8,$CF,$FB,$D3,$6F,$8C,$FF
	.dc.b	$45,$CB,$F1,$1D,$3F,$AF,$7F,$BA,$F9,$9F,$CE,$7A,$FF,$6F,$5F,$EF
	.dc.b	$C7,$F7,$53,$FB,$C3,$FD,$F5,$E4,$7F,$00,$7F,$93,$77,$FF,$04,$7F
	.dc.b	$90,$9F,$CA,$7C,$1F,$E1,$4F,$F2,$BF,$0F,$F8,$73,$FC,$68,$FE,$5D
	.dc.b	$E3,$7F,$19,$7F,$19,$8F,$F1,$87,$F1,$9C,$BF,$E1,$B8,$FE,$57,$1F
	.dc.b	$C6,$8E,$27,$33,$F9,$9F,$8B,$FC,$69,$FE,$75,$E2,$FF,$1A,$7C,$A7
	.dc.b	$F3,$8A,$FE,$34,$FF,$3F,$DB,$F8,$D2,$BE,$FC,$DF,$E2,$CF,$13,$E1
	.dc.b	$CF,$F1,$B3,$FA,$3C,$FC,$77,$E5,$9F,$92,$E8,$FC,$86,$1F,$93,$9F
	.dc.b	$A5,$3F,$C7,$4F,$C8,$1F,$B6,$3F,$20,$7E,$D8,$FC,$81,$FB,$63,$F2
	.dc.b	$07,$ED,$8F,$C8,$1F,$B6,$3F,$20,$7E,$D8,$FC,$81,$FC,$74,$86,$0D
	.dc.b	$3F,$91,$7F,$32,$FE,$45,$FC,$EB,$F9,$17,$F4,$2F,$E4,$5F,$D2,$BF
	.dc.b	$91,$7F,$52,$FE,$45,$FD,$6B,$F0,$0F,$E2,$EF,$F8,$70,$F8,$F7,$FE
	.dc.b	$58,$E8,$FE,$C1,$5F,$CC,$31,$FF,$06,$46,$09,$5F,$CE,$0F,$55,$FD
	.dc.b	$6A,$E1,$1F,$AD,$35,$19,$9E,$31,$C2,$BA,$BA,$6B,$93,$91,$FC,$78
	.dc.b	$93,$CD,$E0,$69,$C3,$C6,$FD,$8E,$9A,$E6,$FC,$F1,$89,$E2,$4A,$F1
	.dc.b	$B6,$BB,$47,$9B,$AF,$FA,$26,$1E,$CA,$D5,$F3,$AE,$24,$BE,$24,$8D
	.dc.b	$37,$0F,$19,$2F,$A5,$71,$AD,$92,$7D,$3E,$1F,$FA,$44,$FE,$D1,$5C
	.dc.b	$2F,$A3,$6B,$83,$6B,$B4,$FC,$56,$BF,$16,$34,$7F,$98,$7C,$87,$C4
	.dc.b	$FF,$88,$49,$3F,$E3,$7F,$89,$FF,$12,$7A,$1F,$E2,$5F,$88,$FF,$13
	.dc.b	$7A,$5F,$E2,$7F,$82,$FF,$14,$7A,$9F,$E2,$9F,$82,$FF,$15,$7A,$DF
	.dc.b	$E2,$BF,$82,$FF,$18,$72,$3D,$CB,$F0,$58,$F7,$73,$F0,$5A,$F7,$BF
	.dc.b	$C5,$9C,$DF,$ED,$AE,$9F,$E3,$1C,$7F,$C2,$7A,$7F,$CC,$9E,$66,$E3
	.dc.b	$F8,$3C,$7F,$93,$A9,$A3,$1E,$D5,$A1,$3A,$CF,$CB,$A7,$9E,$7F,$CA
	.dc.b	$DB,$7A,$DF,$F2,$C7,$07,$BD,$FF,$2D,$71,$7C,$37,$FC,$B9,$C9,$FE
	.dc.b	$49,$E5,$F9,$31,$FE,$5A,$E6,$FF,$1A,$F4,$7D,$08,$FA,$8D,$7E,$A1
	.dc.b	$82,$18,$27,$0F,$E4,$BA,$BF,$09,$E3,$71,$8F,$CA,$F5,$FD,$E6,$DE
	.dc.b	$78,$FF,$1D,$39,$6C,$67,$0F,$D9,$F1,$FE,$81,$FF,$39,$1F,$B8,$3F
	.dc.b	$CE,$9C,$5F,$AE,$6D,$FF,$A5,$1B,$5C,$1E,$04,$3F,$7E,$68,$D4,$F3
	.dc.b	$74,$D7,$66,$27,$B3,$67,$B3,$A3,$FD,$EF,$8F,$F9,$38,$FE,$26,$F7
	.dc.b	$3D,$75,$D4,$95,$D2,$39,$46,$2B,$AB,$63,$CF,$3B,$8C,$47,$48,$C1
	.dc.b	$27,$8B,$FE,$B8,$1E,$BF,$A6,$79,$CE,$88,$E0,$87,$57,$2D,$32,$7B
	.dc.b	$48,$75,$34,$43,$27,$C6,$3F,$AB,$E5,$FC,$B1,$CD,$FB,$1B,$7E,$B7
	.dc.b	$B9,$D4,$CD,$79,$1A,$68,$FE,$7B,$1F,$01,$F9,$AF,$37,$C8,$73,$21
	.dc.b	$A8,$E8,$E2,$7C,$8D,$9F,$A5,$97,$F3,$E3,$F8,$EC,$78,$8E,$A7,$33
	.dc.b	$C9,$3C,$9C,$1A,$70,$49,$FB,$18,$77,$BF,$1D,$70,$3B,$10,$D1,$DA
	.dc.b	$7D,$CE,$4F,$A0,$7F,$19,$2E,$1F,$45,$E2,$78,$E3,$8B,$83,$95,$F9
	.dc.b	$6B,$C6,$FF,$60,$9F,$9C,$FF,$42,$9F,$E1,$35,$FC,$09,$B8,$7F,$00
	.dc.b	$69,$FD,$28,$B3,$C5,$F3,$2C,$FF,$87,$C6,$AF,$F8,$53,$C4,$FE,$2B
	.dc.b	$5D,$5A,$66,$FC,$4F,$F8,$97,$37,$8F,$FC,$DD,$7C,$17,$F4,$62,$4F
	.dc.b	$F3,$7C,$BF,$9D,$FB,$FF,$D2,$BD,$EF,$E9,$7F,$01,$FD,$33,$E0,$7F
	.dc.b	$A6,$FC,$1F,$F4,$EC,$FF,$A7,$FC,$27,$F5,$0F,$82,$FE,$A3,$F0,$BF
	.dc.b	$D4,$BE,$0F,$FA,$9F,$C3,$7F,$54,$F8,$4F,$EA,$BF,$0F,$FD,$5B,$E1
	.dc.b	$7F,$AB,$FC,$47,$FA,$33,$87,$FA,$5B,$9E,$2F,$27,$22,$F9,$46,$31
	.dc.b	$EC,$EB,$3F,$13,$C3,$5F,$33,$CB,$7F,$53,$D3,$87,$DC,$F6,$E3,$F8
	.dc.b	$6F,$D4,$CB,$E3,$BF,$E3,$EE,$7F,$A7,$AF,$D9,$E6,$F9,$E4,$E7,$FA
	.dc.b	$8D,$AF,$A9,$18,$23,$FD,$91,$FD,$67,$4F,$F4,$5A,$E9,$1A,$6C,$F2
	.dc.b	$23,$82,$5A,$93,$FE,$54,$C3,$D2,$FA,$FA,$3D,$B8,$FF,$A1,$25,$97
	.dc.b	$C2,$AF,$81,$9F,$C6,$7E,$16,$9F,$29,$F8,$9A,$FC,$E7,$E3,$6D,$F4
	.dc.b	$9F,$91,$BF,$D6,$7E,$57,$07,$DB,$7E,$67,$0F,$DF,$CC,$E7,$4F,$25
	.dc.b	$71,$7C,$D7,$D5,$F4,$61,$E1,$7D,$58,$F9,$5F,$66,$5E,$97,$DD,$9F
	.dc.b	$B5,$F8,$1A,$7C,$2E,$1F,$D2,$6D,$F2,$36,$FF,$0D,$F3,$7F,$87,$7F
	.dc.b	$80,$7F,$BA,$E1,$FE,$AA,$6D,$E6,$B8,$4E,$7C,$6B,$3C,$7C,$EB,$5F
	.dc.b	$E2,$2B,$7B,$B8,$E6,$43,$DE,$C2,$F0,$2F,$BF,$8F,$6F,$3F,$01,$AF
	.dc.b	$77,$7F,$03,$C3,$DF,$E3,$F0,$5C,$BE,$07,$9F,$C1,$FE,$BB,$83,$FA
	.dc.b	$9F,$17,$F8,$E7,$C8,$ED,$1F,$AE,$EF,$DB,$FC,$0F,$81,$87,$48,$C6
	.dc.b	$5E,$07,$5A,$79,$1E,$1B,$7A,$1E,$5C,$1E,$CE,$1F,$6F,$1F,$BB,$93
	.dc.b	$DF,$CB,$F0,$2E,$1F,$EA,$CD,$1F,$07,$F1,$09,$1E,$5F,$84,$F8,$FE
	.dc.b	$1F,$C8,$F8,$4F,$93,$F0,$BF,$2B,$DE,$F9,$7F,$0B,$F3,$F5,$FF,$78
	.dc.b	$E3,$FC,$C1,$7C,$2E,$7C,$9D,$7C,$36,$FE,$1B,$87,$C1,$F1,$F8,$6E
	.dc.b	$5F,$0B,$F2,$F2,$FE,$73,$D9,$FC,$E1,$FC,$E9,$FC,$AF,$34,$F1,$8C
	.dc.b	$5F,$57,$58,$F8,$5E,$19,$F9,$FB,$7D,$31,$F4,$F6,$F6,$BF,$57,$83
	.dc.b	$E0,$BF,$0C,$FB,$1C,$7F,$19,$FC,$D9,$F9,$13,$E6,$9F,$D4,$39,$3F
	.dc.b	$36,$FE,$57,$97,$EA,$3F,$81,$D7,$EC,$3F,$68,$FC,$27,$85,$F9,$37
	.dc.b	$F0,$DE,$17,$E4,$DF,$C4,$78,$5F,$93,$7F,$15,$E1,$7E,$4D,$FC,$67
	.dc.b	$85,$FE,$AB,$DB,$F9,$0C,$7F,$B3,$73,$4E,$E3,$CD,$5C,$DF,$45,$F7
	.dc.b	$7D,$58,$78,$DF,$66,$3F,$AA,$ED,$F4,$C7,$BF,$3F,$6B,$F0,$74,$F8
	.dc.b	$7E,$27,$EF,$63,$F8,$17,$9D,$FD,$A1,$FD,$C6,$FF,$2E,$7F,$81,$3F
	.dc.b	$B6,$ED,$FA,$33,$F2,$F8,$7E,$AB,$F3,$78,$BE,$CB,$F8,$FE,$2F,$C9
	.dc.b	$71,$FD,$F8,$FC,$97,$27,$E0,$DF,$CA,$72,$FE,$1D,$FC,$B7,$37,$E3
	.dc.b	$F4,$61,$FB,$9E,$0F,$F3,$95,$7E,$E4,$FC,$85,$71,$AD,$C7,$FA,$CB
	.dc.b	$81,$87,$79,$CF,$8B,$FC,$53,$2F,$24,$62,$BF,$4B,$7F,$9B,$CB,$D7
	.dc.b	$3F,$C8,$F4,$F7,$BF,$C9,$75,$F8,$6F,$F2,$7D,$BE,$45,$7F,$2B,$DF
	.dc.b	$E6,$F8,$DF,$CA,$71,$FE,$1C,$7D,$29,$FE,$61,$C5,$F5,$B0,$FE,$61
	.dc.b	$9F,$DA,$7F,$99,$70,$7F,$0D,$BF,$CD,$78,$5F,$C4,$5F,$CE,$78,$7F
	.dc.b	$C5,$5F,$CF,$78,$9F,$C6,$5F,$D0,$78,$BF,$C7,$9F,$DB,$9F,$BA,$BF
	.dc.b	$96,$E3,$0F,$E8,$1D,$7F,$93,$7F,$A1,$75,$FE,$4D,$FE,$89,$D7,$F9
	.dc.b	$37,$FA,$37,$5F,$E4,$DF,$E9,$1D,$7F,$93,$C3,$F9,$67,$8B,$FC,$89
	.dc.b	$FE,$5D,$E2,$FF,$22,$7E,$19,$0F,$DC,$F4,$FF,$46,$E5,$F9,$71,$F9
	.dc.b	$43,$F8,$01,$FB,$BF,$17,$F9,$13,$FC,$E7,$C5,$FE,$44,$FF,$3B,$F1
	.dc.b	$7F,$86,$3F,$C3,$8F,$E2,$DE,$2F,$F2,$07,$F8,$C7,$8B,$FC,$81,$FE
	.dc.b	$35,$E2,$FF,$20,$7F,$1C,$7C,$73,$FA,$59,$FE,$FE,$F0,$FF,$BD,$BF
	.dc.b	$92,$38,$9F,$C8,$0E,$A7,$F2,$03,$C2,$78,$CF,$D3,$1E,$73,$F9,$01
	.dc.b	$EB,$3F,$90,$1F,$D5,$71,$FD,$93,$FC,$80,$87,$C8,$F1,$3F,$8E,$E9
	.dc.b	$FD,$BA,$3F,$1C,$7F,$1F,$3E,$89,$FC,$7C,$FA,$A7,$F1,$F3,$F8,$59
	.dc.b	$FD,$70,$FE,$3E,$7F,$60,$3F,$8F,$9F,$D9,$0F,$E3,$E7,$F6,$9E,$6F
	.dc.b	$C7,$70,$FE,$BB,$3F,$E3,$F3,$FE,$80,$EC,$FF,$25,$E1,$FC,$82,$38
	.dc.b	$1C,$8F,$CC,$9D,$8F,$CE,$9E,$23,$F4,$27,$F3,$43,$D1,$5F,$E1,$7E
	.dc.b	$1F,$C7,$47,$F3,$FE,$1F,$C7,$3F,$D0,$4F,$E8,$7A,$7F,$1F,$CF,$F1
	.dc.b	$D1,$F2,$AF,$F8,$FB,$F8,$E7,$FC,$45,$A7,$E3,$A3,$FA,$57,$07,$E3
	.dc.b	$9F,$E9,$9C,$1F,$8E,$7F,$B7,$78,$9F,$C3,$35,$FF,$2A,$E3,$FC,$3E
	.dc.b	$7F,$06,$7F,$20,$3F,$0E,$7F,$20,$3F,$1F,$8F,$F1,$6F,$17,$FB,$73
	.dc.b	$FE,$59,$F1,$BF,$29,$3F,$DE,$F8,$BF,$30,$FF,$7D,$E3,$FC,$E3,$FD
	.dc.b	$FF,$8B,$F9,$8B,$FC,$6C,$FC,$71,$FC,$DC,$FC,$71,$FC,$EC,$FC,$71
	.dc.b	$F0,$72,$FE,$34,$FE,$3B,$7F,$91,$DC,$F8,$7F,$F0,$71,$27,$FC,$05
	.dc.b	$F0,$FF,$E1,$1F,$5F,$FC,$25,$EC,$7F,$84,$FD,$9F,$F0,$A7,$B5,$FE
	.dc.b	$15,$F8,$6F,$F0,$B7,$B9,$FE,$17,$F8,$6F,$F0,$C7,$BD,$FE,$19,$F8
	.dc.b	$6F,$F0,$D7,$C0,$7F,$86,$FE,$1B,$FC,$95,$FC,$02,$F8,$3B,$F0,$98
	.dc.b	$F8,$7F,$F5,$64,$9C,$F1,$7F,$AF,$E3,$83,$FE,$17,$BE,$AF,$F8,$67
	.dc.b	$1F,$0B,$FE,$1B,$CF,$CA,$FF,$87,$75,$F4,$BF,$E1,$FD,$FD,$AF,$F8
	.dc.b	$87,$87,$FC,$61,$C5,$F0,$E3,$FC,$61,$C7,$FD,$09,$FE,$A8,$E7,$AF
	.dc.b	$FA,$83,$C2,$39,$1F,$E2,$BE,$4F,$90,$FF,$8B,$79,$3E,$43,$FE,$2F
	.dc.b	$E4,$F9,$0F,$F8,$C7,$93,$E4,$3F,$E3,$3E,$4F,$90,$FC,$1E,$4F,$86
	.dc.b	$FF,$BC,$B8,$BE,$33,$FC,$73,$B7,$FD,$E9,$E6,$7C,$B8,$F9,$1C,$DF
	.dc.b	$42,$F9,$C7,$F9,$2F,$0E,$EF,$F9,$37,$1F,$1B,$FE,$4F,$CB,$CE,$FF
	.dc.b	$94,$73,$F5,$BF,$E5,$3D,$3F,$0C,$FE,$20,$FF,$25,$6B,$F2,$7A,$3F
	.dc.b	$6D,$B3,$E1,$BF,$02,$4A,$FB,$9F,$0F,$F8,$4F,$6B,$F0,$BE,$DF,$E1
	.dc.b	$BE,$1F,$F0,$FE,$EF,$E2,$3E,$1F,$F1,$3E,$FF,$E2,$BD,$9F,$C5,$FC
	.dc.b	$07,$E3,$3E,$1F,$F1,$BF,$05,$F8,$EF,$87,$FC,$7F,$C2,$7E,$43,$E1
	.dc.b	$FF,$C2,$9E,$47,$EF,$0F,$E3,$BD,$FF,$B5,$E2,$FC,$74,$F0,$7F,$21
	.dc.b	$7D,$1F,$C9,$63,$E0,$7F,$29,$9F,$91,$FC,$B6,$BE,$87,$F3,$1B,$FB
	.dc.b	$1F,$CD,$70,$FE,$8B,$8B,$E1,$47,$DB,$E3,$FC,$77,$7F,$FA,$C3,$9F
	.dc.b	$F1,$DB,$FF,$12,$F2,$3F,$71,$7F,$3C,$FF,$42,$78,$5F,$9A,$7F,$D0
	.dc.b	$DE,$1F,$E7,$9F,$F4,$47,$89,$FA,$27,$FD,$15,$E2,$FE,$99,$FF,$46
	.dc.b	$78,$DF,$AA,$7F,$D1,$DE,$3F,$E1,$DF,$C4,$9F,$B1,$F2,$3F,$6A,$7E
	.dc.b	$4B,$9F,$FC,$67,$DD,$FB,$49,$E2,$7E,$64,$FD,$A1,$F9,$D3,$F6,$87
	.dc.b	$E8,$4F,$DA,$1F,$A5,$3F,$68,$7E,$A4,$FD,$59,$FE,$A1,$F2,$3E,$1B
	.dc.b	$FB,$20,$7F,$84,$35,$F8,$AF,$EE,$0F,$C9,$9F,$1E,$BF,$C8,$DA,$FC
	.dc.b	$77,$FC,$91,$AF,$C7,$7F,$C9,$5A,$FC,$77,$FC,$99,$AF,$C7,$7F,$C9
	.dc.b	$DA,$FC,$77,$FC,$A1,$AF,$C7,$C7,$F6,$9E,$3F,$EC,$5F,$F0,$81,$27
	.dc.b	$E3,$ED,$F2,$77,$FE,$9B,$BF,$C9,$9E,$75,$F2,$4F,$AA,$7C,$93,$EC
	.dc.b	$9F,$24,$FB,$A7,$C9,$3F,$02,$7C,$93,$F0,$A7,$C1,$9F,$C5,$70,$7E
	.dc.b	$D7,$9E,$CF,$C6,$3E,$41,$F3,$CF,$A3,$5F,$E6,$DF,$0B,$F8,$0B,$FE
	.dc.b	$6F,$F0,$FF,$82,$BF,$E7,$1F,$13,$F8,$4B,$FE,$73,$F1,$7F,$86,$BF
	.dc.b	$E7,$5F,$1B,$F8,$8B,$F8,$8F,$1F,$E1,$62,$E8,$6D,$FC,$72,$3E,$46
	.dc.b	$DF,$E8,$97,$F2,$47,$E5,$0F,$E4,$55,$D2,$3F,$4D,$87,$81,$FD,$46
	.dc.b	$3E,$47,$F5,$59,$7A,$1F,$D6,$67,$EC,$7F,$5D,$A7,$F1,$CC,$7E,$11
	.dc.b	$F1,$34,$FE,$7E,$FE,$43,$C7,$FE,$3B,$D5,$FB,$A9,$FD,$1B,$FB,$D9
	.dc.b	$FD,$F9,$FD,$67,$C7,$FE,$38,$FF,$5B,$F1,$FF,$8E,$3F,$D7,$7C,$7F
	.dc.b	$E3,$8F,$F5,$FF,$1F,$F8,$E3,$FD,$87,$C7,$FE,$38,$FF,$63,$F1,$FF
	.dc.b	$8E,$3F,$E1,$0D,$7F,$8E,$7A,$3F,$A4,$8F,$F1,$C3,$FC,$81,$F9,$DC
	.dc.b	$8F,$23,$F8,$F4,$7F,$4B,$F2,$3F,$8F,$3F,$D3,$7C,$8F,$E3,$CF,$F4
	.dc.b	$FE,$9F,$E3,$CF,$F5,$1F,$23,$F8,$F3,$FD,$4F,$C8,$FE,$3C,$FF,$55
	.dc.b	$F2,$3E,$13,$FC,$74,$FE,$F5,$C7,$FC,$7B,$C6,$FD,$BF,$2F,$CC,$9F
	.dc.b	$9D,$BF,$F1,$FF,$1F,$E9,$4F,$E4,$67,$F8,$29,$FD,$4C,$FE,$12,$7F
	.dc.b	$57,$3F,$86,$9F,$D6,$CF,$E2,$27,$F5,$F3,$FC,$10,$7F,$6B,$3F,$14
	.dc.b	$FE,$D1,$FE,$D5,$E7,$3C,$AF,$DD,$4F,$F1,$EF,$0F,$F7,$CF,$ED,$AB
	.dc.b	$A1,$FB,$63,$C0,$7E,$D8,$F2,$1F,$B6,$3D,$07,$ED,$8F,$ED,$47,$EB
	.dc.b	$B0,$F8,$0F,$EC,$E3,$E2,$75,$7F,$6A,$F8,$6F,$F1,$49,$27,$FC,$48
	.dc.b	$D7,$FC,$08,$D9,$FE,$6A,$DA,$FE,$1E,$D9,$FC,$1B,$7C,$2F,$D8,$B7
	.dc.b	$94,$DF,$13,$FA,$0D,$78,$08,$78,$0B,$87,$81,$B8,$78,$1A,$78,$0B
	.dc.b	$87,$81,$B8,$78,$1B,$87,$B7,$1F,$85,$18,$21,$E1,$26,$1E,$16,$E1
	.dc.b	$E1,$69,$E1,$2E,$1E,$16,$E1,$E1,$6F,$C3,$FF,$95,$0F,$89,$8C,$8F
	.dc.b	$0E,$1D,$C9,$87,$76,$C7,$76,$E1,$DD,$B8,$77,$7D,$18,$FF,$8A,$3D
	.dc.b	$0F,$F3,$39,$2F,$FA,$FC,$B0,$FF,$4A,$FF,$00,$FF,$B2,$07,$C1,$7F
	.dc.b	$01,$9D,$57,$F2,$D9,$7A,$D9,$BE,$3C,$76,$B5,$EB,$DD,$58,$7E,$8E
	.dc.b	$7C,$37,$E0,$F6,$F5,$2F,$82,$BE,$EE,$3E,$0B,$3E,$FE,$BE,$0B,$7F
	.dc.b	$03,$C3,$E0,$3F,$5B,$9F,$F1,$8D,$3A,$9F,$D0,$D7,$8B,$A8,$F5,$4F
	.dc.b	$27,$7B,$79,$B0,$EF,$18,$D5,$9F,$92,$31,$AF,$A3,$6F,$4F,$5F,$AB
	.dc.b	$83,$D7,$C3,$EC,$E2,$F6,$F2,$7B,$B1,$FA,$47,$F9,$C7,$97,$FC,$6F
	.dc.b	$CF,$F6,$78,$BF,$00,$49,$FC,$03,$78,$3F,$00,$D9,$E1,$3A,$24,$E9
	.dc.b	$B1,$A8,$C1,$1C,$13,$A7,$F2,$D3,$F9,$4A,$ED,$D3,$F9,$78,$FC,$EB
	.dc.b	$7A,$3F,$22,$D9,$DC,$FF,$5D,$F0,$BF,$50,$FF,$4B,$C7,$F4,$CD,$C7
	.dc.b	$F4,$CD,$BF,$EA,$0D,$29,$FA,$F3,$06,$A7,$9B,$D8,$86,$CA,$E9,$C1
	.dc.b	$1C,$12,$58,$FC,$29,$FE,$08,$78,$08,$74,$8F,$4B,$6B,$CE,$F5,$8C
	.dc.b	$B6,$33,$5D,$DB,$3A,$8C,$4F,$F9,$53,$DA,$7B,$3D,$63,$E1,$CE,$9F
	.dc.b	$85,$5C,$49,$7D,$09,$1C,$23,$2D,$8C,$BC,$9A,$FB,$8F,$56,$1F,$0F
	.dc.b	$3F,$E4,$91,$9B,$F1,$36,$FA,$BC,$CE,$C4,$30,$43,$85,$F8,$1F,$E5
	.dc.b	$D8,$FC,$07,$E4,$1C,$4D,$94,$F8,$87,$32,$9A,$30,$72,$2D,$7F,$91
	.dc.b	$BB,$7F,$8B,$69,$FE,$BA,$3F,$CD,$6F,$F6,$39,$E4,$43,$A9,$0E,$86
	.dc.b	$4A,$7A,$B3,$D5,$7C,$68,$B8,$FC,$67,$84,$65,$E2,$47,$F2,$99,$FD
	.dc.b	$1C,$38,$E5,$F1,$EB,$57,$CD,$C9,$F1,$4C,$C7,$DE,$CB,$E6,$61,$FC
	.dc.b	$06,$3F,$71,$3F,$C1,$9B,$19,$8F,$E0,$46,$4A,$FF,$09,$C3,$C9,$19
	.dc.b	$3F,$82,$9A,$BF,$E1,$B5,$D1,$B1,$98,$FE,$E9,$A7,$F5,$2F,$82,$FE
	.dc.b	$86,$49,$FE,$65,$F0,$5F,$D1,$BE,$0F,$FA,$3F,$67,$F4,$8F,$84,$FE
	.dc.b	$93,$F0,$7F,$D2,$BE,$17,$FA,$5F,$C2,$7F,$51,$FE,$01,$FE,$B3,$93
	.dc.b	$97,$FA,$DF,$0A,$5F,$0B,$BE,$3E,$1F,$0F,$C7,$E1,$79,$7C,$47,$3F
	.dc.b	$87,$FE,$B9,$CD,$FC,$1B,$4F,$F1,$FF,$81,$8E,$8D,$47,$F4,$CD,$18
	.dc.b	$73,$8C,$65,$DD,$D6,$9E,$7E,$9F,$4C,$7F,$51,$E1,$F6,$BF,$D4,$F8
	.dc.b	$FE,$0B,$FD,$57,$97,$FB,$4F,$37,$C7,$8F,$75,$FC,$B7,$F9,$91,$F5
	.dc.b	$78,$FF,$9B,$BF,$44,$87,$D8,$23,$A7,$FC,$37,$D7,$FA,$6E,$79,$1F
	.dc.b	$EA,$A3,$75,$FD,$01,$A7,$77,$15,$CC,$8E,$FC,$A7,$57,$9E,$BD,$27
	.dc.b	$F8,$3F,$AB,$D9,$7F,$63,$C5,$67,$F8,$3A,$FE,$ED,$97,$E0,$DF,$EF
	.dc.b	$1D,$7F,$83,$7F,$BD,$75,$FC,$D7,$E7,$9F,$57,$AB,$EF,$BF,$5F,$AB
	.dc.b	$EF,$BF,$67,$AB,$EE,$3F,$78,$FF,$03,$F5,$7E,$05,$F3,$F5,$FE,$2E
	.dc.b	$78,$C6,$3A,$57,$DA,$31,$87,$E5,$79,$FC,$91,$CB,$66,$7E,$98,$C6
	.dc.b	$9E,$D7,$5A,$FC,$17,$86,$DF,$1B,$8F,$E5,$71,$FC,$C9,$FF,$72,$6A
	.dc.b	$F8,$9F,$D2,$12,$3F,$2D,$F0,$5F,$A7,$F3,$FF,$51,$E0,$7E,$A7,$DD
	.dc.b	$FD,$57,$C3,$FE,$AF,$DF,$FD,$67,$C3,$FE,$B7,$E0,$7F,$5D,$F0,$FF
	.dc.b	$AF,$F8,$3F,$D8,$7C,$3F,$EC,$7E,$17,$F6,$5F,$0F,$FE,$43,$F6,$9F
	.dc.b	$0F,$F3,$38,$7F,$DE,$A7,$FB,$E7,$DB,$FF,$23,$3F,$33,$C1,$F9,$DE
	.dc.b	$17,$EE,$2F,$8C,$FE,$B7,$0E,$CF,$EB,$F2,$F1,$3F,$B1,$D3,$CC,$FE
	.dc.b	$CF,$6F,$53,$FB,$5E,$0F,$73,$FB,$7E,$2F,$84,$FC,$43,$F7,$5C,$BF
	.dc.b	$1D,$73,$FC,$DF,$09,$A3,$DB,$F7,$12,$BF,$C9,$B0,$F1,$BE,$13,$F0
	.dc.b	$FE,$23,$E2,$7B,$BF,$17,$E2,$3E,$37,$BF,$F1,$FE,$23,$E4,$7C,$0F
	.dc.b	$C9,$F8,$8F,$95,$E6,$FC,$BF,$88,$F9,$9F,$09,$F3,$7E,$23,$E7,$7C
	.dc.b	$37,$CF,$F8,$8F,$F3,$97,$77,$CC,$ED,$FE,$42,$DF,$21,$C8,$E6,$F5
	.dc.b	$DF,$28,$F4,$E3,$D9,$F5,$E7,$E2,$7D,$BA,$F9,$9F,$7E,$FE,$A7,$E0
	.dc.b	$F0,$FB,$9F,$87,$C7,$F0,$9F,$8B,$CB,$F1,$DF,$D5,$F2,$FC,$B7,$F8
	.dc.b	$CE,$9F,$CD,$FA,$3E,$97,$07,$F3,$AF,$23,$F3,$95,$F5,$CF,$D8,$78
	.dc.b	$FF,$9A,$7F,$99,$78,$FF,$9A,$7F,$9A,$78,$FF,$9A,$7F,$9B,$78,$FF
	.dc.b	$9A,$7F,$9C,$78,$FF,$9A,$7F,$9D,$78,$FF,$9A,$7E,$DF,$8F,$F9,$9C
	.dc.b	$FF,$A5,$F3,$FF,$82,$67,$E5,$78,$FF,$98,$8F,$9B,$BF,$E1,$AF,$95
	.dc.b	$7F,$05,$C3,$B3,$FC,$1F,$1F,$13,$FC,$27,$2F,$33,$FC,$2F,$3F,$53
	.dc.b	$FC,$37,$4F,$73,$F0,$0F,$85,$A7,$C5,$DF,$F3,$1E,$0F,$E5,$F8,$7E
	.dc.b	$65,$7C,$EF,$07,$E8,$5F,$EB,$78,$FE,$A4,$FD,$2E,$4F,$B0,$FD,$4E
	.dc.b	$5F,$B8,$FD,$6E,$6F,$C0,$3F,$63,$9F,$F0,$8F,$DA,$E8,$FC,$43,$F7
	.dc.b	$3A,$7F,$18,$FD,$EE,$AF,$CA,$FB,$32,$3F,$4F,$1A,$D3,$F8,$05,$72
	.dc.b	$7F,$81,$5F,$67,$F8,$26,$1E,$27,$F8,$36,$3E,$67,$F8,$46,$5E,$A7
	.dc.b	$F8,$56,$7E,$E7,$E0,$65,$FB,$6D,$7F,$25,$BF,$EF,$7A,$FF,$1B,$3F
	.dc.b	$CE,$F8,$7F,$23,$D3,$F7,$25,$E4,$FF,$08,$AF,$E8,$1E,$27,$F0,$C7
	.dc.b	$FA,$17,$8B,$FC,$41,$FE,$89,$E3,$7F,$14,$7F,$A3,$78,$FF,$C6,$1F
	.dc.b	$E9,$1E,$47,$F1,$C7,$FA,$57,$93,$FC,$A7,$BF,$F2,$1A,$7E,$33,$C3
	.dc.b	$FD,$D5,$7F,$29,$E9,$CB,$FB,$BE,$BF,$E4,$EF,$23,$F3,$67,$F2,$73
	.dc.b	$F3,$E7,$F2,$73,$F4,$67,$F2,$73,$F4,$E7,$F2,$73,$F5,$67,$F2,$73
	.dc.b	$FC,$3D,$E1,$FF,$8B,$F7,$F8,$B1,$FD,$43,$87,$F8,$6B,$FC,$A3,$C6
	.dc.b	$FD,$1D,$FE,$4B,$97,$F9,$E7,$2F,$E4,$B4,$CF,$6F,$D5,$77,$CF,$F6
	.dc.b	$5E,$3D,$1F,$75,$E7,$D3,$F8,$17,$AF,$57,$E1,$5E,$FD,$7F,$89,$7C
	.dc.b	$3D,$9F,$8D,$7C,$7D,$BF,$91,$DF,$FA,$9F,$87,$FC,$89,$FE,$73,$E2
	.dc.b	$FF,$2D,$7F,$9D,$F8,$DF,$CC,$5F,$E7,$BC,$5F,$CD,$5F,$E7,$FE,$3F
	.dc.b	$F3,$97,$FA,$0F,$91,$FC,$F5,$FE,$87,$E4,$FF,$41,$7F,$A2,$F2,$FF
	.dc.b	$65,$E3,$F8,$51,$F6,$79,$3E,$4F,$87,$F9,$3E,$7F,$9B,$B7,$F2,$3A
	.dc.b	$FD,$FF,$07,$E4,$FA,$7F,$81,$F4,$7F,$25,$F1,$3F,$83,$BF,$C9,$FC
	.dc.b	$5F,$E1,$6F,$F2,$9F,$1B,$F8,$7B,$FC,$AF,$C7,$FE,$26,$FF,$2D,$F2
	.dc.b	$3F,$8B,$BF,$DC,$7C,$1F,$C6,$E8,$CF,$F9,$02,$F2,$7F,$90,$6B,$F9
	.dc.b	$1A,$E3,$5F,$A7,$C3,$AB,$FA,$9C,$7C,$2F,$EA,$F2,$F2,$BF,$AD,$CF
	.dc.b	$D2,$FE,$BF,$4F,$6B,$EF,$C7,$E0,$9F,$8F,$3F,$BB,$9F,$D2,$38,$BF
	.dc.b	$1D,$1F,$90,$E1,$F9,$BA,$FF,$78,$E2,$FC,$75,$FE,$9F,$BF,$F1,$CF
	.dc.b	$EA,$7B,$FF,$1C,$FE,$AF,$BF,$F1,$CF,$EB,$7B,$FF,$1C,$FE,$BF,$BF
	.dc.b	$F1,$CF,$EC,$7B,$FF,$1C,$FE,$CF,$BF,$F1,$DC,$DF,$DC,$BC,$1F,$EE
	.dc.b	$D3,$F8,$E3,$FB,$79,$F8,$E3,$FB,$79,$F8,$E3,$FB,$79,$F8,$E3,$FB
	.dc.b	$79,$F8,$E3,$FB,$79,$F8,$E3,$FB,$79,$F8,$E3,$FB,$79,$FC,$7C,$87
	.dc.b	$ED,$4F,$E1,$1F,$11,$FE,$37,$23,$FE,$1A,$F6,$FF,$C7,$5E,$DF,$F8
	.dc.b	$EF,$DC,$FD,$EE,$1F,$EA,$CF,$E0,$1F,$DF,$DF,$F9,$91,$7B,$DA,$E5
	.dc.b	$FE,$C7,$0F,$7F,$15,$FE,$11,$87,$C0,$63,$99,$0F,$81,$C3,$FC,$82
	.dc.b	$33,$F0,$58,$F8,$3C,$FC,$4E,$BE,$17,$7F,$13,$F3,$B3,$FD,$EF,$A1
	.dc.b	$C7,$BD,$CF,$A9,$FF,$11,$F1,$7A,$F0,$E7,$18,$E1,$F4,$E7,$E2,$D3
	.dc.b	$C7,$CF,$E4,$E0,$F2,$F0,$F9,$B8,$BC,$FC,$7E,$8C,$BE,$41,$FE,$44
	.dc.b	$E5,$FF,$32,$73,$7C,$E8,$FF,$10,$F3,$FF,$00,$F3,$7F,$93,$B9,$E0
	.dc.b	$FF,$31,$3C,$27,$91,$FE,$4E,$E8,$EC,$F9,$CF,$99,$8F,$8D,$C7,$3F
	.dc.b	$CC,$D3,$D1,$18,$E8,$F9,$83,$83,$E6,$46,$7A,$3E,$63,$BE,$8F,$CD
	.dc.b	$F2,$7F,$53,$E3,$B3,$FD,$4C,$D1,$D7,$97,$E5,$BF,$85,$E5,$F9,$4F
	.dc.b	$86,$7C,$6F,$E2,$73,$F4,$3F,$8B,$D7,$DA,$FF,$91,$B6,$FC,$9B,$F9
	.dc.b	$53,$FC,$DF,$BF,$CB,$F2,$BF,$CE,$DB,$BE,$1F,$F1,$24,$AF,$81,$F1
	.dc.b	$3F,$8C,$F8,$9F,$C6,$FC,$27,$E3,$BD,$CF,$C7,$FC,$2F,$E4,$3D,$EF
	.dc.b	$C8,$FC,$37,$E4,$BE,$03,$F2,$7F,$0F,$F9,$4F,$82,$FC,$C7,$F0,$07
	.dc.b	$F2,$5D,$7E,$7A,$F8,$5C,$FC,$36,$BE,$27,$F2,$9E,$07,$F9,$2B,$4F
	.dc.b	$F7,$6E,$1F,$E2,$53,$8C,$F0,$79,$DF,$6E,$CF,$04,$7D,$9C,$FC,$8F
	.dc.b	$DB,$D7,$D0,$FD,$DD,$BD,$8F,$DF,$E0,$F8,$0F,$C2,$3F,$05,$C7,$F1
	.dc.b	$5F,$E0,$9E,$07,$FA,$7B,$9F,$F2,$59,$7E,$52,$7E,$81,$FE,$78,$F0
	.dc.b	$FF,$3A,$FF,$01,$3F,$42,$7E,$37,$C4,$FD,$2B,$F8,$FF,$17,$F5,$2F
	.dc.b	$E4,$7C,$3F,$C7,$BF,$C2,$7C,$3F,$C7,$BF,$C2,$FC,$3F,$C6,$E6,$F2
	.dc.b	$3F,$81,$13,$C2,$FD,$D5,$7F,$11,$F1,$3F,$1F,$CF,$FC,$57,$C4,$FC
	.dc.b	$7C,$FF,$15,$AF,$C7,$CF,$F1,$9F,$13,$F1,$EF,$F1,$BF,$13,$F1,$EF
	.dc.b	$FB,$47,$C4,$FC,$7D,$FF,$A5,$BC,$6F,$86,$FF,$28,$3F,$A8,$72,$63
	.dc.b	$5F,$91,$1F,$1A,$FF,$DB,$5C,$7F,$19,$FE,$51,$1F,$18,$FE,$51,$C7
	.dc.b	$F1,$9F,$E5,$5C,$7F,$19,$FE,$59,$C7,$F1,$AB,$F9,$87,$1F,$C6,$7F
	.dc.b	$99,$71,$FE,$33,$93,$F6,$97,$FC,$03,$93,$F9,$07,$8F,$F1,$AF,$F8
	.dc.b	$F4,$FE,$F4,$FE,$05,$C9,$F1,$A3,$F9,$E7,$27,$C6,$C3,$F9,$CF,$27
	.dc.b	$C6,$7F,$9D,$F2,$7C,$67,$FC,$7B,$C9,$F1,$AF,$FD,$F1,$C9,$F0,$FC
	.dc.b	$87,$2F,$F1,$DD,$3E,$41,$FC,$80,$FD,$C9,$F1,$F0,$FF,$23,$F2,$7C
	.dc.b	$77,$FA,$4C,$7C,$7C,$3F,$A8,$72,$7C,$77,$FA,$97,$27,$C7,$7F,$AA
	.dc.b	$72,$7C,$7B,$FE,$A3,$C9,$F1,$DF,$EA,$7C,$9F,$8C,$E4,$FC,$56,$5F
	.dc.b	$1F,$8B,$E3,$57,$F5,$8E,$6F,$C5,$BF,$D0,$A7,$F3,$07,$F1,$93,$F3
	.dc.b	$87,$F2,$BE,$3F,$8C,$FF,$32,$3F,$97,$1E,$B3,$FC,$ED,$B7,$F3,$B7
	.dc.b	$E1,$13,$C7,$FE,$33,$D1,$FB,$3D,$BF,$B7,$3F,$D8,$BC,$6F,$DC,$D7
	.dc.b	$F6,$FE,$AF,$DE,$BF,$DC,$7A,$FF,$80,$BF,$DC,$FB,$3F,$82,$BF,$DD
	.dc.b	$7B,$7F,$84,$AC,$7F,$8C,$76,$BC,$2F,$E3,$53,$9D,$FF,$DB,$71,$FC
	.dc.b	$6E,$3F,$8E,$F8,$5F,$C6,$DF,$F5,$C9,$FD,$3A,$7F,$C8,$BD,$5F,$DB
	.dc.b	$A3,$F8,$D9,$C0,$FE,$36,$7F,$26,$8F,$1E,$1F,$E0,$BC,$7C,$EF,$F8
	.dc.b	$37,$2F,$5B,$FE,$0F,$CF,$DE,$FF,$84,$74,$F8,$6F,$F8,$4F,$5F,$91
	.dc.b	$87,$F8,$43,$6F,$C8,$6D,$FB,$4C,$FE,$7D,$FF,$6C,$7F,$20,$7D,$43
	.dc.b	$FB,$61,$F6,$0F,$EC,$DD,$BF,$72,$7F,$B5,$47,$E0,$0F,$E1,$87,$E1
	.dc.b	$0F,$EB,$C7,$EB,$4F,$E2,$87,$F6,$43,$FC,$27,$C9,$F9,$13,$F1,$EF
	.dc.b	$F8,$8F,$93,$F1,$EF,$F8,$97,$BB,$F6,$99,$7F,$8A,$AB,$F6,$93,$FE
	.dc.b	$29,$E5,$F2,$78,$FF,$E3,$08,$F4,$1F,$D9,$CF,$F0,$39,$FC,$64,$FD
	.dc.b	$79,$F8,$FF,$23,$F8,$3B,$E4,$3F,$44,$43,$04,$D3,$FC,$B4,$E5,$AE
	.dc.b	$70,$E7,$18,$23,$82,$47,$F8,$1C,$E0,$58,$DB,$63,$79,$FF,$41,$C7
	.dc.b	$E4,$4F,$86,$73,$D1,$FB,$A8,$E7,$5B,$7D,$5C,$EC,$77,$87,$29,$DD
	.dc.b	$F1,$C7,$C3,$87,$19,$CD,$6E,$73,$5B,$C7,$9D,$78,$CF,$41,$2F,$95
	.dc.b	$70,$BE,$55,$C2,$BA,$9C,$6B,$71,$9C,$7D,$12,$9F,$DB,$7C,$47,$F8
	.dc.b	$40,$91,$FE,$10,$F8,$AF,$F0,$9F,$C5,$7F,$85,$3D,$CF,$F0,$AF,$C1
	.dc.b	$7F,$85,$BD,$DF,$F0,$BF,$C4,$7F,$86,$3D,$8F,$F0,$CF,$C4,$7F,$86
	.dc.b	$BD,$AF,$F0,$DF,$C4,$7F,$87,$3D,$CF,$F0,$EF,$C4,$7F,$87,$BD,$EF
	.dc.b	$F0,$FF,$C4,$7F,$4F,$E0,$F2,$72,$7F,$00,$F4,$33,$1F,$E2,$2A,$E2
	.dc.b	$FF,$89,$30,$EA,$FF,$89,$B2,$F0,$BF,$E2,$8D,$3C,$AF,$F8,$AB,$6F
	.dc.b	$4B,$FE,$2C,$E0,$F6,$BE,$F3,$FC,$5F,$C7,$F0,$9F,$F0,$CF,$47,$D5
	.dc.b	$7E,$93,$79,$7E,$6C,$63,$9F,$E9,$B4,$FA,$53,$B2,$46,$A7,$89,$23
	.dc.b	$33,$C2,$75,$18,$8F,$BA,$7E,$AB,$9F,$ED,$D7,$E3,$60,$7E,$C8,$A7
	.dc.b	$E9,$4A,$ED,$D1,$0D,$15,$D3,$82,$38,$24,$AB,$FD,$73,$3C,$63,$0E
	.dc.b	$27,$73,$A8,$E7,$3A,$7F,$D8,$33,$A7,$84,$65,$F4,$B5,$ED,$3C,$E7
	.dc.b	$84,$72,$9D,$3D,$A7,$51,$89,$DC,$F4,$9C,$91,$F9,$7E,$33,$97,$BC
	.dc.b	$79,$27,$4F,$76,$BC,$CD,$10,$D1,$4D,$18,$21,$82,$1E,$03,$25,$32
	.dc.b	$7B,$4E,$86,$48,$64,$E2,$53,$8D,$7C,$8A,$F7,$46,$4E,$31,$F1,$9B
	.dc.b	$7F,$24,$95,$F0,$23,$2D,$79,$12,$79,$36,$78,$BE,$83,$13,$C6,$76
	.dc.b	$49,$DC,$AB,$CF,$39,$AE,$15,$AA,$F6,$B5,$F2,$9D,$E7,$83,$63,$4F
	.dc.b	$88,$C4,$F4,$23,$82,$1F,$32,$79,$BF,$6E,$BC,$B5,$FB,$9E,$9F,$E7
	.dc.b	$13,$FB,$D2,$1F,$BC,$72,$DE,$2F,$E7,$91,$82,$38,$24,$7F,$0F,$38
	.dc.b	$16,$36,$FF,$65,$CB,$F8,$3F,$8E,$F8,$0F,$F6,$79,$27,$FD,$9F,$F1
	.dc.b	$1F,$C2,$2F,$FC,$77,$FC,$02,$F8,$3C,$65,$FC,$3B,$54,$BE,$0F,$38
	.dc.b	$FF,$4D,$87,$C1,$E2,$B7,$E6,$D9,$7C,$1D,$F8,$5C,$7C,$26,$7E,$1B
	.dc.b	$5F,$0B,$BF,$87,$E1,$F0,$DC,$7C,$9E,$5F,$0D,$CF,$E2,$3A,$7C,$3C
	.dc.b	$E1,$FE,$47,$CB,$EA,$9D,$C6,$39,$3D,$5C,$1F,$D9,$23,$FC,$AC,$7A
	.dc.b	$72,$F0,$3B,$3D,$9A,$79,$5C,$71,$7C,$2D,$FD,$51,$8B,$F6,$B7,$8B
	.dc.b	$E0,$71,$FC,$1E,$4F,$85,$CB,$FF,$00,$39,$BE,$95,$F8,$B2,$FF,$35
	.dc.b	$73,$FF,$94,$3A,$BF,$15,$BF,$E0,$27,$FA,$47,$67,$F4,$62,$5F,$E3
	.dc.b	$38,$BF,$03,$3A,$8C,$57,$1B,$E7,$1F,$87,$8C,$5F,$28,$FE,$9F,$D7
	.dc.b	$F8,$8C,$7C,$B8,$FF,$9F,$BA,$FF,$15,$5F,$88,$3E,$06,$9E,$D7,$E1
	.dc.b	$6D,$F7,$F7,$FC,$07,$5F,$C6,$9F,$77,$0B,$8B,$E6,$46,$38,$FE,$8B
	.dc.b	$AE,$4F,$C0,$3F,$37,$9F,$F0,$0F,$CF,$E7,$FC,$03,$D3,$9F,$EF,$BD
	.dc.b	$BA,$3F,$12,$FD,$5E,$9F,$C6,$BF,$5F,$AB,$F2,$2F,$D9,$EB,$F3,$BF
	.dc.b	$64,$F5,$9F,$77,$AF,$EC,$BF,$04,$FB,$1C,$7F,$11,$BC,$BF,$E8,$8E
	.dc.b	$07,$C4,$FE,$84,$93,$F8,$7F,$1B,$F4,$9E,$FF,$E9,$7E,$03,$F4,$DE
	.dc.b	$DF,$E9,$FE,$07,$F5,$5F,$07,$FD,$F3,$0F,$C3,$3F,$E7,$76,$FC,$16
	.dc.b	$BD,$5D,$FC,$17,$0F,$07,$8F,$C1,$72,$F6,$39,$FC,$17,$CA,$C7,$FC
	.dc.b	$BF,$C2,$E9,$CC,$7E,$7B,$45,$F2,$8C,$63,$D9,$D6,$7E,$6D,$3C,$FE
	.dc.b	$DF,$A2,$7D,$47,$E9,$77,$F7,$3F,$A7,$E1,$FE,$09,$C5,$F1,$23,$F0
	.dc.b	$BC,$7F,$EB,$0D,$1E,$DF,$C2,$24,$F9,$7E,$0B,$E2,$FB,$BF,$1B,$DE
	.dc.b	$F8,$FF,$11,$F2,$3E,$03,$E4,$FC,$47,$CA,$F8,$2F,$97,$F0,$DF,$33
	.dc.b	$E1,$3E,$6F,$C4,$7C,$EF,$86,$F9,$FF,$11,$FC,$67,$F8,$07,$FD,$B7
	.dc.b	$E9,$7F,$B7,$9F,$F6,$D7,$07,$E3,$CF,$C9,$10,$C1,$FB,$B7,$71,$C7
	.dc.b	$D4,$EB,$3F,$36,$FC,$2F,$CF,$C7,$CA,$FD,$1C,$FD,$2F,$D3,$D7,$DA
	.dc.b	$FD,$5D,$FE,$0B,$F0,$CF,$B1,$C5,$F1,$97,$A9,$FE,$48,$EF,$FC,$9C
	.dc.b	$7F,$17,$E6,$FA,$1A,$FF,$40,$E7,$FC,$9C,$FF,$1B,$F1,$3F,$26,$FF
	.dc.b	$1D,$F1,$3E,$E3,$F8,$5E,$CF,$C0,$3F,$82,$3F,$11,$E2,$7E,$45,$FC
	.dc.b	$57,$89,$F9,$17,$F1,$9E,$27,$F0,$FC,$BF,$D4,$FD,$A8,$D4,$FE,$7A
	.dc.b	$B9,$3F,$A0,$BF,$CF,$3F,$95,$F0,$3F,$1E,$FE,$5C,$FC,$79,$FC,$3B
	.dc.b	$C0,$FC,$7B,$FC,$43,$C0,$FC,$7B,$F0,$48,$7C,$6F,$2F,$FA,$F6,$5F
	.dc.b	$90,$8F,$96,$43,$F6,$3E,$0F,$D0,$7F,$67,$CB,$F5,$1F,$DA,$F3,$7D
	.dc.b	$87,$F6,$FC,$FF,$71,$FD,$CF,$47,$E0,$1F,$DD,$F4,$FE,$11,$FC,$47
	.dc.b	$57,$E4,$7E,$27,$F6,$DC,$7F,$DD,$3C,$DF,$E3,$EF,$F5,$DA,$FE,$A7
	.dc.b	$DB,$F8,$EB,$FE,$EC,$FE,$61,$F5,$F7,$7E,$76,$7F,$79,$DF,$FA,$17
	.dc.b	$F7,$DE,$07,$E9,$5F,$E0,$1E,$0F,$EA,$5F,$E0,$5E,$17,$EB,$5F,$E0
	.dc.b	$9E,$1F,$F1,$4E,$3F,$C7,$68,$CF,$E4,$F8,$7F,$C8,$7C,$1F,$C6,$C7
	.dc.b	$F7,$07,$E8,$E1,$F4,$CF,$C7,$1F,$C1,$4F,$C7,$1F,$6C,$FC,$71,$F7
	.dc.b	$CF,$C7,$1F,$83,$3F,$89,$9F,$8E,$3F,$8C,$78,$DF,$C6,$9F,$D9,$DF
	.dc.b	$F6,$4F,$0F,$F8,$F4,$7E,$48,$D9,$C4,$FE,$5B,$E3,$7F,$1D,$7F,$97
	.dc.b	$F8,$DF,$9E,$7F,$98,$F8,$5F,$A2,$7F,$AD,$F8,$7F,$A6,$7F,$AE,$F8
	.dc.b	$9F,$AA,$7F,$AF,$F8,$BF,$09,$FF,$13,$69,$FC,$77,$6F,$93,$B7,$C8
	.dc.b	$9F,$DD,$1F,$CF,$B8,$3E,$83,$FD,$03,$87,$EA,$3F,$D0,$B8,$BE,$C3
	.dc.b	$FD,$13,$8F,$EE,$3F,$D1,$B9,$3F,$00,$FF,$48,$E5,$FE,$39,$8F,$E3
	.dc.b	$63,$FC,$31,$E3,$FE,$32,$3F,$C4,$BD,$1F,$90,$7F,$17,$D3,$C2,$BF
	.dc.b	$A2,$5F,$4E,$6F,$C0,$43,$D7,$78,$0C,$FC,$75,$A9,$CE,$9E,$88,$C5
	.dc.b	$35,$F6,$C6,$36,$F8,$2E,$B7,$FC,$86,$5F,$B3,$8E,$3C,$1F,$8F,$8F
	.dc.b	$DD,$C7,$F9,$6B,$87,$F1,$EF,$F9,$73,$C2,$FE,$06,$FF,$97,$BC,$3F
	.dc.b	$E0,$EF,$8A,$BF,$85,$9F,$DB,$0F,$C7,$BE,$6F,$17,$F8,$9B,$E8,$F1
	.dc.b	$BF,$8B,$BE,$AF,$1F,$F8,$E7,$27,$F9,$9F,$D7,$FF,$6F,$4E,$70,$FC
	.dc.b	$79,$FC,$74,$FC,$C9,$F9,$B3,$F1,$E7,$E7,$CF,$C7,$9F,$A3,$3F,$1E
	.dc.b	$7E,$9C,$FC,$79,$FD,$AC,$FF,$06,$1F,$0A,$7F,$60,$8F,$E3,$C7,$D7
	.dc.b	$E0,$FE,$3C,$FF,$83,$35,$FC,$9B,$9F,$03,$F8,$F5,$FD,$6F,$1B,$F8
	.dc.b	$F3,$F6,$3C,$6F,$E3,$CF,$DA,$F1,$BF,$8F,$3F,$73,$C6,$FE,$3C,$FD
	.dc.b	$EF,$1B,$F8,$F3,$FD,$8F,$1F,$E3,$3D,$7F,$8F,$7F,$D1,$5D,$9F,$93
	.dc.b	$7F,$D1,$9D,$BF,$97,$7F,$92,$9F,$92,$3F,$B5,$1F,$C7,$0F,$ED,$47
	.dc.b	$F1,$C3,$FB,$51,$FC,$70,$FE,$D4,$7F,$1C,$3F,$B5,$1F,$C7,$0F,$D8
	.dc.b	$D7,$E3,$1F,$E8,$1E,$4B,$C8,$FE,$5B,$EC,$FF,$81,$49,$3F,$CC,$B8
	.dc.b	$7F,$C1,$1E,$DF,$F8,$27,$E2,$3F,$C1,$5F,$11,$FE,$0B,$F7,$BF,$C1
	.dc.b	$9F,$11,$FE,$0D,$F8,$0F,$F0,$77,$C0,$7F,$83,$FE,$07,$FC,$21,$F1
	.dc.b	$1F,$E1,$1F,$83,$FF,$09,$7C,$47,$F8,$4F,$E1,$7F,$C2,$9F,$11,$FE
	.dc.b	$15,$F8,$7F,$F5,$77,$83,$F8,$EF,$07,$F2,$1E,$A6,$79,$3F,$A2,$DF
	.dc.b	$17,$FA,$3E,$3D,$5F,$E9,$39,$F8,$5F,$E9,$7A,$F9,$5F,$E9,$BB,$FA
	.dc.b	$5F,$E9,$FC,$3E,$D7,$FA,$8F,$1F,$F8,$9F,$93,$E1,$C7,$F8,$87,$97
	.dc.b	$F2,$0F,$E4,$3C,$D7,$AF,$FE,$FB,$8F,$EB,$33,$C3,$AB,$FC,$5D,$7F
	.dc.b	$2A,$7F,$AE,$F4,$7C,$A7,$FA,$FF,$47,$CA,$7F,$B0,$F4,$7C,$A7,$FB
	.dc.b	$1F,$47,$CA,$7F,$B2,$F4,$7C,$A7,$FB,$3F,$47,$CA,$7F,$B4,$F4,$7E
	.dc.b	$3F,$6F,$F6,$EF,$67,$C9,$9F,$F1,$D7,$37,$C9,$7F,$C7,$9C,$DF,$25
	.dc.b	$FF,$1F,$73,$7C,$97,$FC,$81,$CD,$F2,$5F,$F2,$17,$37,$C9,$7F,$C8
	.dc.b	$9C,$DF,$25,$FF,$23,$73,$7C,$97,$E1,$90,$F9,$1C,$CF,$88,$FF,$2A
	.dc.b	$12,$3F,$C0,$3E,$D7,$F9,$5F,$E2,$3F,$DE,$1F,$C0,$3F,$EF,$99,$FF
	.dc.b	$6B,$BF,$EF,$76,$FC,$46,$BE,$23,$7F,$11,$C3,$E0,$B8,$FC,$47,$2F
	.dc.b	$84,$E7,$F1,$1D,$3E,$1B,$AF,$C4,$76,$F1,$FF,$71,$D5,$FE,$CF,$D7
	.dc.b	$FB,$7F,$8D,$FC,$3E,$1B,$FF,$BB,$FC,$4F,$E8,$4B,$0E,$97,$E6,$CB
	.dc.b	$C0,$FA,$34,$F2,$3E,$AD,$BD,$0F,$A8,$F6,$FC,$3F,$C0,$F8,$9F,$85
	.dc.b	$1F,$10,$FF,$12,$F2,$FC,$77,$E5,$F8,$FF,$8F,$DB,$F7,$71,$FB,$93
	.dc.b	$F3,$67,$E4,$0F,$CF,$9F,$90,$3F,$46,$7E,$40,$FD,$39,$F9,$03,$F5
	.dc.b	$67,$E4,$0F,$D7,$9F,$90,$3F,$66,$7E,$40,$FF,$44,$6B,$FE,$3D,$70
	.dc.b	$7E,$40,$E0,$7E,$40,$E8,$7E,$40,$F0,$1F,$90,$3C,$87,$E4,$0F,$41
	.dc.b	$F9,$03,$D8,$7E,$40,$F8,$07,$ED,$CF,$C8,$79,$5F,$C7,$DF,$EF,$9D
	.dc.b	$1F,$B7,$74,$49,$FD,$BB,$F2,$0F,$F5,$07,$8F,$FB,$77,$FD,$45,$E3
	.dc.b	$FE,$DD,$FF,$52,$78,$FF,$B7,$7F,$D4,$DE,$3F,$ED,$DF,$F5,$47,$8F
	.dc.b	$FB,$77,$FD,$55,$E3,$FE,$DD,$FF,$56,$78,$FF,$C8,$3C,$99,$1F,$23
	.dc.b	$3F,$F5,$B7,$8F,$FB,$77,$FD,$71,$E3,$FE,$DD,$FF,$5D,$78,$FF,$B7
	.dc.b	$7F,$D7,$9E,$3F,$ED,$DF,$F5,$F7,$8F,$FB,$77,$FD,$81,$E3,$FE,$DD
	.dc.b	$FF,$61,$78,$FF,$B7,$CB,$E1,$72,$FE,$29,$FC,$69,$FB,$99,$FF,$64
	.dc.b	$73,$6D,$E3,$1F,$49,$EA,$FD,$67,$C2,$FD,$A7,$CA,$FD,$E7,$D2,$FE
	.dc.b	$09,$F6,$BF,$86,$7E,$0B,$F0,$E7,$FC,$76,$FE,$39,$FF,$42,$ED,$F2
	.dc.b	$E7,$FD,$B9,$BF,$CF,$7F,$DB,$DC,$1F,$4D,$FF,$70,$70,$FD,$77,$FD
	.dc.b	$C5,$C5,$F6,$DF,$F7,$27,$1F,$DF,$7F,$DC,$DC,$9F,$CE,$73,$FF,$71
	.dc.b	$79,$3F,$AF,$7F,$17,$AF,$F4,$4E,$7F,$C7,$78,$1F,$D1,$3A,$7F,$6D
	.dc.b	$8F,$FB,$BD,$FF,$79,$79,$1F,$BC,$C1,$3F,$C2,$BA,$FF,$8F,$F6,$7F
	.dc.b	$20,$F1,$FF,$82,$D6,$B9,$3F,$89,$78,$4E,$FF,$E1,$73,$93,$FC,$B1
	.dc.b	$0F,$03,$F8,$8C,$67,$5F,$E3,$3E,$0F,$F1,$68,$C5,$7F,$19,$7F,$96
	.dc.b	$F8,$7F,$C8,$BC,$AF,$EE,$51,$FE,$63,$6D,$BE,$17,$FD,$23,$3C,$AB
	.dc.b	$F9,$5B,$FC,$DB,$C6,$FE,$5E,$FF,$38,$F1,$FF,$99,$BF,$CE,$BC,$8F
	.dc.b	$E6,$EF,$F3,$CF,$27,$E0,$3F,$C8,$8F,$E7,$FC,$5F,$1F,$8F,$FC,$E1
	.dc.b	$E4,$39,$7E,$6F,$1F,$FB,$1B,$9B,$E9,$6B,$F9,$9B,$FC,$BE,$3F,$99
	.dc.b	$AF,$F6,$5F,$4F,$DB,$9F,$BB,$3F,$C9,$0F,$DE,$9F,$C9,$0F,$DE,$9F
	.dc.b	$C9,$0F,$DE,$9F,$C9,$0F,$DE,$9F,$C9,$0F,$DE,$9F,$C9,$0F,$DE,$9F
	.dc.b	$97,$3F,$C8,$67,$E6,$CF,$E2,$7E,$3F,$EF,$6B,$FD,$BF,$E3,$FE,$F5
	.dc.b	$FF,$70,$F8,$FF,$BD,$7F,$DC,$7E,$3F,$EF,$5F,$F7,$2F,$8F,$FB,$D7
	.dc.b	$FD,$CF,$E3,$FE,$F5,$FF,$74,$F8,$FF,$B7,$7E,$21,$FB,$DE,$FF,$CB
	.dc.b	$F5,$3E,$27,$FB,$79,$35,$FE,$DE,$D7,$F0,$ED,$DB,$FB,$0B,$6F,$F2
	.dc.b	$AD,$9F,$C9,$B7,$AB,$FB,$4B,$6F,$C1,$3D,$8F,$89,$3F,$DC,$6F,$EF
	.dc.b	$F9,$1B,$B9,$5B,$3C,$65,$9F,$1B,$70,$F1,$B4,$F1,$96,$FC,$6D,$C3
	.dc.b	$C6,$FF,$8F,$5B,$5D,$23,$B9,$1C,$12,$BB,$C6,$5B,$19,$8F,$09,$92
	.dc.b	$B9,$AF,$1C,$65,$B1,$FE,$0E,$F3,$7F,$C7,$9D,$8F,$37,$FC,$4B,$1F
	.dc.b	$12,$F8,$63,$E8,$C7,$C2,$48,$EE,$DC,$7A,$B6,$78,$4E,$89,$3A,$6C
	.dc.b	$7C,$6D,$3F,$13,$87,$F8,$D8,$93,$FE,$36,$F6,$FF,$C7,$3F,$05,$FE
	.dc.b	$3A,$F7,$3F,$C7,$7F,$07,$FE,$3C,$F7,$7F,$C7,$BF,$09,$FE,$6C,$E0
	.dc.b	$FF,$76,$7F,$00,$FF,$9B,$B2,$FE,$30,$BE,$1B,$5C,$1F,$E3,$38,$7C
	.dc.b	$4E,$2B,$FD,$F1,$0F,$88,$C7,$12,$1F,$05,$87,$F8,$D8,$F8,$2C,$7C
	.dc.b	$4F,$F9,$9F,$4F,$F7,$F7,$97,$FC,$80,$FE,$43,$E3,$3A,$F7,$A7,$FB
	.dc.b	$4E,$BD,$38,$75,$F6,$3B,$67,$DF,$5F,$06,$DE,$1E,$0F,$16,$3F,$08
	.dc.b	$FF,$2E,$F1,$7C,$67,$FC,$BF,$C9,$F2,$9F,$98,$7F,$98,$B9,$FE,$7B
	.dc.b	$F9,$7E,$CF,$C1,$6B,$F8,$BE,$AF,$B7,$C9,$F6,$48,$E0,$93,$C3,$97
	.dc.b	$F0,$4D,$BF,$BF,$D7,$F9,$0C,$7F,$1F,$1F,$A7,$9D,$61,$D1,$B8,$74
	.dc.b	$9D,$10,$FC,$AC,$7F,$36,$7F,$46,$7F,$AB,$27,$F4,$2D,$9F,$D9,$B6
	.dc.b	$7F,$08,$FE,$B8,$C1,$89,$E2,$E4,$87,$62,$BA,$70,$47,$04,$9F,$0C
	.dc.b	$7F,$83,$8F,$BD,$87,$FB,$51,$A7,$F6,$5E,$0F,$C4,$47,$69,$E3,$3D
	.dc.b	$5B,$5D,$6F,$D7,$39,$6C,$66,$3D,$93,$C0,$93,$A6,$C6,$A3,$FD,$AB
	.dc.b	$E9,$3C,$1F,$14,$6C,$93,$B2,$1F,$AF,$BF,$03,$6F,$AB,$67,$8B,$4E
	.dc.b	$27,$D2,$30,$43,$ED,$67,$F8,$97,$FD,$72,$FF,$4B,$78,$10,$F3,$9D
	.dc.b	$0A,$73,$29,$B3,$E4,$1B,$3F,$96,$E1,$F0,$5C,$9C,$4A,$71,$BF,$7B
	.dc.b	$6F,$DE,$DB,$F7,$B4,$ED,$3F,$F0,$03,$93,$FC,$EF,$1F,$E2,$DE,$6F
	.dc.b	$E0,$B1,$ED,$8F,$39,$0F,$09,$67,$C2,$DB,$F0,$BF,$D2,$B0,$F5,$4E
	.dc.b	$A3,$89,$1C,$12,$78,$C6,$5B,$1F,$9B,$C7,$E9,$4F,$F9,$01,$F3,$3B
	.dc.b	$C7,$DE,$F8,$DA,$E4,$F8,$35,$D0,$FE,$CF,$7E,$67,$F8,$04,$7E,$F8
	.dc.b	$FE,$01,$B7,$F1,$3A,$FE,$0A,$48,$CE,$9F,$D4,$70,$F9,$75,$FC,$C3
	.dc.b	$5F,$E1,$F5,$CC,$91,$CD,$E2,$7F,$54,$BE,$8E,$AF,$FB,$9F,$6F,$F9
	.dc.b	$D3,$A5,$F0,$7F,$D0,$49,$5F,$BA,$D7,$FA,$37,$C1,$FF,$45,$F8,$3F
	.dc.b	$E8,$DF,$09,$FD,$1F,$E1,$3F,$A4,$7C,$2F,$F4,$9F,$85,$FE,$95,$F0
	.dc.b	$DF,$D2,$FE,$1B,$FA,$67,$C3,$FF,$4D,$F8,$7F,$E9,$DF,$11,$FD,$3F
	.dc.b	$E2,$3F,$A8,$7C,$4F,$F5,$4F,$11,$CB,$EE,$E7,$FF,$74,$71,$FF,$9D
	.dc.b	$FC,$CF,$EF,$CF,$EA,$0F,$35,$70,$7D,$1C,$CC,$3B,$46,$32,$F1,$3A
	.dc.b	$D3,$CC,$F0,$DB,$D4,$F2,$E0,$F7,$3D,$38,$BE,$13,$DB,$93,$E3,$BF
	.dc.b	$DD,$39,$BF,$9D,$73,$7F,$41,$8F,$A6,$D7,$E9,$1F,$54,$CF,$5F,$E1
	.dc.b	$B8,$3E,$D4,$7F,$87,$FC,$7F,$ED,$8F,$D3,$EF,$FE,$E1,$D9,$FC,$B2
	.dc.b	$39,$1D,$30,$ED,$7C,$89,$C9,$FD,$6A,$7B,$F6,$7F,$96,$7B,$3F,$05
	.dc.b	$A7,$AA,$FF,$C1,$EF,$F8,$0F,$5F,$7D,$FF,$84,$36,$F8,$6F,$F8,$4B
	.dc.b	$7F,$8E,$FF,$85,$38,$3E,$5B,$FE,$16,$E1,$F9,$EF,$F8,$63,$8B,$E9
	.dc.b	$BF,$E1,$AE,$3F,$C0,$6D,$F7,$CF,$F6,$DC,$7D,$FF,$07,$0B,$83,$FC
	.dc.b	$2F,$3F,$88,$8F,$BF,$D1,$F8,$C7,$F0,$3D,$3F,$90,$7F,$07,$D5,$F9
	.dc.b	$47,$F0,$BD,$7F,$98,$7F,$0F,$D9,$F9,$C7,$F1,$3D,$BF,$05,$FC,$5F
	.dc.b	$1F,$DE,$E5,$FF,$6D,$F6,$3E,$27,$F3,$04,$AF,$C9,$F8,$1F,$9B,$F5
	.dc.b	$3F,$39,$E7,$FE,$77,$E0,$BF,$3D,$E8,$FE,$7F,$E0,$FF,$41,$E9,$FE
	.dc.b	$87,$E1,$3F,$45,$E4,$FE,$8F,$E1,$7F,$49,$EB,$7E,$97,$E1,$BF,$4D
	.dc.b	$EC,$7E,$9F,$E1,$FE,$2F,$67,$FA,$17,$B3,$FD,$C7,$E6,$E7,$8F,$F4
	.dc.b	$B7,$C5,$FD,$3E,$3D,$5F,$D4,$E7,$E1,$7F,$57,$AF,$95,$FD,$6E,$FE
	.dc.b	$97,$F5,$FC,$3E,$D7,$DE,$7E,$CB,$93,$E1,$2E,$6F,$8F,$ED,$7E,$F4
	.dc.b	$91,$FB,$8F,$3F,$F8,$07,$B9,$FC,$07,$DD,$FE,$05,$F1,$3F,$E5,$6F
	.dc.b	$E0,$17,$C0,$5F,$89,$C7,$C4,$E7,$E0,$F5,$F1,$3B,$F8,$5E,$1F,$13
	.dc.b	$C7,$E1,$F9,$7C,$4F,$3F,$89,$E9,$F1,$5F,$D0,$3D,$7F,$E1,$1D,$ED
	.dc.b	$FF,$E0,$07,$22,$73,$0B,$E7,$3F,$C9,$61,$8F,$82,$31,$9F,$91,$D6
	.dc.b	$BE,$87,$86,$FE,$DE,$0F,$7C,$7F,$1A,$E2,$F8,$6F,$C5,$3F,$8E,$F2
	.dc.b	$FC,$87,$E6,$7C,$51,$FD,$27,$93,$F9,$37,$E9,$10,$F5,$78,$3B,$E0
	.dc.b	$FF,$3F,$F9,$3F,$66,$72,$72,$EC,$FB,$CE,$27,$FD,$0B,$DB,$F8,$38
	.dc.b	$C6,$0E,$EF,$C4,$46,$3B,$FF,$18,$EB,$C0,$FC,$83,$C3,$C1,$FC,$A3
	.dc.b	$FD,$BC,$FE,$39,$ED,$7F,$5E,$D7,$F3,$31,$FC,$F3,$C4,$FC,$FB,$FC
	.dc.b	$FB,$C5,$FD,$1B,$FD,$03,$C6,$FD,$3B,$FD,$0B,$C7,$FD,$5B,$FD,$13
	.dc.b	$C8,$FD,$7B,$FD,$1B,$C9,$FD,$9B,$FD,$23,$CA,$FE,$A1,$C9,$F9,$8E
	.dc.b	$FF,$C1,$F1,$7C,$BA,$F9,$E7,$CE,$E7,$FA,$53,$A7,$FA,$C7,$8F,$F6
	.dc.b	$27,$E2,$F4,$FD,$C7,$E5,$77,$FE,$69,$FC,$09,$F8,$9F,$17,$F3,$4F
	.dc.b	$E2,$FC,$5F,$CD,$3F,$8D,$F1,$7F,$34,$FE,$3F,$C5,$FC,$D3,$F9,$1F
	.dc.b	$17,$F3,$5C,$DF,$77,$1F,$CC,$EB,$F9,$8F,$17,$F2,$F7,$E2,$F1,$7F
	.dc.b	$2C,$F9,$3C,$5F,$CB,$3E,$6F,$17,$F2,$CF,$A3,$C5,$FC,$B3,$FD,$BB
	.dc.b	$C6,$FD,$B3,$FD,$C3,$C7,$F8,$2F,$E5,$88,$7C,$8F,$13,$FD,$9B,$E4
	.dc.b	$FC,$DD,$BF,$25,$CF,$F3,$76,$F9,$FD,$1F,$93,$7E,$8F,$8D,$F9,$27
	.dc.b	$E9,$F8,$DF,$91,$7E,$AF,$8D,$F9,$07,$EB,$F8,$DF,$90,$7E,$CF,$8D
	.dc.b	$F9,$07,$ED,$F8,$DF,$90,$7F,$7F,$EE,$BC,$BF,$C8,$70,$7F,$A5,$CF
	.dc.b	$F0,$61,$F9,$EF,$1F,$F2,$2F,$E8,$3C,$7F,$C8,$BF,$A2,$F1,$FF,$22
	.dc.b	$FE,$93,$C7,$FC,$8B,$FA,$6F,$1F,$F2,$2F,$EA,$3C,$7F,$C8,$BF,$AA
	.dc.b	$F1,$FF,$22,$FE,$DF,$4F,$C8,$F3,$7F,$82,$B9,$5E,$3F,$E4,$39,$FE
	.dc.b	$4B,$F4,$2F,$F6,$7D,$3F,$51,$FD,$AF,$57,$D8,$7E,$D1,$F8,$F3,$F8
	.dc.b	$99,$F8,$F3,$F8,$B9,$F8,$F3,$F8,$E7,$77,$E2,$32,$FC,$54,$7E,$3C
	.dc.b	$FF,$22,$7B,$FF,$E3,$27,$FB,$B1,$93,$F3,$FE,$07,$1C,$BF,$29,$7F
	.dc.b	$90,$BF,$E5,$7E,$07,$E4,$1F,$E5,$BE,$07,$E4,$1F,$E5,$FE,$07,$A5
	.dc.b	$FD,$9E,$7E,$D7,$F6,$BA,$7C,$17,$F6,$59,$FE,$43,$CC,$FE,$EB,$C1
	.dc.b	$F8,$F8,$FE,$4A,$FC,$E7,$E8,$1F,$D6,$8F,$E0,$87,$F5,$E3,$F8,$41
	.dc.b	$FD,$88,$FE,$18,$7F,$66,$3F,$88,$1F,$DA,$8F,$E2,$87,$F6,$E3,$F8
	.dc.b	$C1,$F8,$A9,$FF,$3C,$3F,$C7,$8C,$69,$FA,$2E,$7F,$C6,$CF,$F1,$0E
	.dc.b	$7F,$C6,$BF,$C4,$B9,$FF,$1A,$FF,$14,$E7,$FC,$6B,$FC,$5B,$9F,$F1
	.dc.b	$AF,$F1,$8E,$7F,$C6,$BF,$C6,$B9,$FE,$03,$F8,$CC,$FF,$A2,$6B,$FD
	.dc.b	$D1,$FD,$BC,$7C,$A7,$F9,$EF,$93,$FC,$6E,$3F,$91,$F9,$3F,$C6,$DF
	.dc.b	$EC,$7E,$67,$F7,$3A,$FC,$61,$FC,$24,$FC,$61,$FC,$34,$FC,$61,$FC
	.dc.b	$44,$FE,$34,$7F,$69,$E7,$FE,$E6,$FE,$33,$D4,$FE,$B8,$FE,$30,$FE
	.dc.b	$53,$E3,$FF,$20,$7F,$95,$F9,$1F,$C9,$1F,$E5,$BE,$4F,$F2,$87,$F9
	.dc.b	$7F,$95,$FC,$B1,$FE,$63,$E5,$F9,$DF,$E6,$79,$7A,$F9,$BF,$89,$EB
	.dc.b	$EF,$7F,$8A,$ED,$FA,$FA,$FE,$80,$7F,$6F,$DD,$E6,$FF,$2F,$3F,$C2
	.dc.b	$C4,$8F,$F0,$27,$B9,$FE,$19,$F0,$FF,$C3,$5C,$1F,$E1,$BF,$8A,$FF
	.dc.b	$0E,$7C,$57,$F8,$77,$E0,$BF,$C3,$DF,$15,$FE,$1F,$F8,$4F,$F1,$07
	.dc.b	$C5,$7C,$2E,$8F,$F0,$AF,$F0,$0F,$C4,$D7,$FB,$A2,$F8,$7D,$71,$7F
	.dc.b	$7D,$87,$C4,$62,$BF,$CD,$10,$F8,$9C,$7A,$C8,$79,$DF,$4A,$BF,$DF
	.dc.b	$1E,$C7,$F2,$4A,$FF,$12,$77,$70,$BD,$D7,$FB,$15,$E8,$DC,$75,$9E
	.dc.b	$E3,$D6,$7F,$89,$F6,$F7,$BF,$E2,$9E,$0F,$86,$FF,$8A,$F8,$BE,$3B
	.dc.b	$FE,$2D,$E4,$FF,$21,$72,$FC,$D8,$FE,$D7,$CD,$FE,$AE,$F6,$3E,$9C
	.dc.b	$7F,$8C,$E3,$5C,$7F,$E3,$8A,$FA,$53,$FE,$37,$E7,$FA,$4F,$F8,$E7
	.dc.b	$9F,$E9,$3F,$E3,$BE,$7F,$A4,$FF,$8F,$79,$FD,$4F,$BF,$7F,$85,$C1
	.dc.b	$F0,$F9,$BE,$27,$37,$C5,$E6,$FE,$E7,$7F,$41,$CB,$F1,$E7,$8B,$CC
	.dc.b	$F9,$47,$C6,$3F,$C0,$FC,$7F,$19,$FF,$04,$F1,$FC,$67,$FC,$17,$C7
	.dc.b	$F1,$9F,$F0,$6F,$1F,$C6,$7E,$11,$0F,$EE,$5C,$8F,$8A,$FB,$44,$8F
	.dc.b	$B5,$F1,$7F,$77,$DB,$FB,$DF,$11,$F7,$FD,$DF,$C0,$7C,$47,$E0,$7D
	.dc.b	$FF,$C1,$7C,$47,$E0,$FE,$07,$F0,$9F,$11,$F8,$5F,$83,$FC,$37,$C4
	.dc.b	$7E,$1F,$E1,$7F,$11,$F1,$1F,$89,$F8,$7F,$C7,$78,$3F,$C9,$38,$7F
	.dc.b	$1E,$7F,$49,$7F,$3F,$1C,$23,$F1,$D5,$D1,$FC,$86,$1E,$07,$F2,$59
	.dc.b	$79,$1F,$CA,$69,$E8,$7F,$2D,$B7,$B1,$FC,$C7,$07,$E6,$F8,$7E,$14
	.dc.b	$7E,$B3,$8B,$F1,$9E,$17,$E2,$BC,$6F,$DF,$4F,$EF,$0F,$CA,$1F,$8B
	.dc.b	$3F,$30,$7E,$2C,$FC,$E1,$F8,$B3,$F4,$07,$E2,$CF,$BE,$7E,$04,$FD
	.dc.b	$DF,$6F,$E2,$DF,$D6,$76,$FE,$2D,$FE,$48,$43,$F8,$0F,$6F,$E3,$1F
	.dc.b	$E0,$7D,$BF,$8C,$7F,$82,$F7,$7E,$61,$FF,$28,$F7,$FE,$71,$FF,$29
	.dc.b	$F8,$1F,$A0,$7F,$CA,$BE,$0F,$E9,$1F,$F2,$BF,$85,$FA,$87,$FC,$B3
	.dc.b	$E1,$FC,$37,$FD,$E5,$E2,$7E,$C1,$FF,$1F,$37,$A7,$FD,$4F,$C1,$BF
	.dc.b	$1B,$FD,$59,$7C,$E7,$FD,$57,$87,$77,$FD,$5B,$8F,$8D,$FF,$57,$E5
	.dc.b	$E7,$7F,$D6,$39,$FA,$DF,$E4,$7A,$7B,$DF,$E4,$BA,$FC,$37,$FD,$6F
	.dc.b	$E3,$7F,$19,$F8,$97,$07,$E4,$38,$FF,$93,$71,$7C,$FE,$2F,$A2,$7D
	.dc.b	$3C,$3F,$97,$BF,$B3,$7F,$D8,$3E,$3F,$EC,$DF,$F6,$1F,$8F,$FB,$37
	.dc.b	$FD,$8B,$E3,$FE,$CD,$FF,$63,$F8,$FF,$B3,$BF,$E2,$E7,$EC,$FC,$8F
	.dc.b	$E2,$FC,$BF,$EC,$FE,$6F,$8D,$CF,$F2,$75,$FF,$3D,$73,$FC,$77,$FC
	.dc.b	$F9,$CF,$F1,$DF,$F3,$F7,$3F,$C7,$7F,$D0,$1C,$FF,$1D,$FF,$41,$73
	.dc.b	$FC,$77,$FA,$27,$3F,$C7,$7F,$A3,$73,$FC,$77,$FA,$47,$2F,$F1,$8D
	.dc.b	$7F,$A2,$39,$DF,$FA,$4D,$71,$7F,$A5,$DF,$57,$F0,$78,$78,$5F,$C2
	.dc.b	$E3,$E5,$7F,$0F,$97,$A5,$FC,$4E,$7E,$D7,$F1,$7A,$7C,$2D,$BF,$C0
	.dc.b	$3B,$FC,$6F,$23,$F8,$C5,$7F,$4F,$EA,$FD,$BF,$1F,$F6,$4E,$BF,$E3
	.dc.b	$13,$FD,$4F,$C8,$FE,$30,$FF,$55,$F2,$3F,$8C,$3F,$D5,$FC,$8F,$E3
	.dc.b	$0F,$F5,$9F,$23,$F8,$C3,$FD,$6F,$C8,$FE,$30,$FF,$5D,$F2,$3F,$8C
	.dc.b	$3F,$C5,$7A,$3F,$DD,$5D,$FF,$8C,$8F,$F1,$E1,$FE,$16,$3F,$B2,$79
	.dc.b	$1F,$C6,$5F,$EE,$FE,$47,$F1,$97,$FB,$CF,$91,$FC,$65,$FE,$F7,$E4
	.dc.b	$7F,$19,$7F,$BE,$F9,$1F,$C6,$5F,$EF,$FE,$47,$F1,$97,$FC,$03,$E4
	.dc.b	$7F,$19,$7E,$21,$0F,$D9,$F3,$FC,$B7,$F8,$C1,$FD,$18,$FE,$30,$7F
	.dc.b	$4A,$3F,$8C,$1F,$D3,$8F,$E3,$07,$F5,$23,$F8,$C1,$FE,$10,$3F,$BE
	.dc.b	$9F,$DF,$CF,$EB,$87,$F1,$83,$FC,$09,$DF,$F8,$BE,$7F,$8D,$1F,$E1
	.dc.b	$8F,$1B,$F6,$1C,$FF,$E1,$CF,$23,$9C,$F5,$3B,$C7,$EC,$1F,$1B,$FB
	.dc.b	$07,$CE,$FE,$C1,$F5,$BF,$B0,$7D,$EF,$F1,$27,$F6,$18,$7E,$79,$F2
	.dc.b	$90,$FE,$56,$58,$FF,$17,$E1,$FC,$C6,$32,$48,$CB,$BE,$1E,$8E,$5B
	.dc.b	$1A,$8C,$1F,$E0,$33,$81,$FE,$CF,$3F,$B0,$9A,$F0,$BF,$62,$E9,$AE
	.dc.b	$72,$F9,$73,$D4,$95,$E0,$24,$F7,$38,$5F,$AE,$77,$19,$7C,$DD,$7F
	.dc.b	$AF,$9D,$57,$2C,$3C,$35,$C1,$B1,$9C,$3C,$6D,$AE,$38,$75,$8C,$3E
	.dc.b	$B9,$DB,$92,$53,$DF,$FF,$66,$60,$B5,$2B,$5D,$DD,$9F,$1C,$72,$C3
	.dc.b	$AC,$60,$8E,$09,$8F,$16,$CF,$96,$30,$7E,$07,$2F,$E2,$5F,$13,$FE
	.dc.b	$61,$24,$F9,$32,$F3,$3E,$5F,$E0,$1A,$A3,$38,$A5,$ED,$FF,$9C,$97
	.dc.b	$C5,$5F,$5B,$1F,$15,$9F,$63,$5F,$15,$BF,$6B,$87,$C5,$71,$F7,$39
	.dc.b	$7C,$57,$3F,$7B,$A7,$C5,$75,$F8,$0F,$EC,$1C,$5F,$8F,$E2,$FF,$18
	.dc.b	$1F,$D9,$9F,$57,$47,$CE,$E3,$F3,$57,$29,$C9,$ED,$C3,$BB,$89,$F6
	.dc.b	$65,$E4,$8C,$71,$35,$F4,$C6,$37,$F6,$BA,$E1,$F8,$5C,$5F,$0F,$8F
	.dc.b	$E2,$6B,$F3,$AB,$FD,$3F,$C9,$FD,$13,$AB,$F1,$71,$F9,$C3,$EC,$10
	.dc.b	$C1,$38,$3F,$31,$D1,$F7,$A3,$17,$F9,$DE,$AF,$C7,$BF,$05,$B1,$F8
	.dc.b	$BE,$9F,$C1,$BD,$6F,$9C,$62,$39,$9B,$35,$5F,$C1,$3A,$7F,$59,$1F
	.dc.b	$EB,$93,$F6,$87,$FB,$7B,$8B,$F7,$AD,$C7,$F9,$3B,$6F,$83,$C0,$FE
	.dc.b	$1E,$64,$E9,$3C,$9C,$90,$E0,$47,$4D,$9E,$EE,$E3,$23,$C4,$FD,$59
	.dc.b	$F2,$63,$C0,$7F,$B4,$E3,$73,$CA,$3A,$57,$A2,$72,$D9,$D4,$62,$7D
	.dc.b	$31,$82,$4E,$C9,$1A,$7F,$81,$BF,$A8,$E7,$F4,$4F,$08,$D1,$0F,$33
	.dc.b	$96,$CF,$46,$9B,$3D,$A4,$3A,$BB,$70,$7C,$BC,$FF,$0F,$7F,$C5,$23
	.dc.b	$FA,$61,$A7,$DC,$F6,$21,$9A,$F2,$36,$BB,$3E,$03,$C8,$53,$E0,$1F
	.dc.b	$38,$C9,$4E,$A6,$CE,$86,$A7,$C6,$E0,$D9,$C4,$FC,$A6,$1F,$80,$9F
	.dc.b	$EC,$51,$B2,$1D,$48,$71,$2C,$F6,$7D,$1B,$B1,$CE,$1F,$02,$6D,$FC
	.dc.b	$07,$91,$2F,$93,$67,$83,$FD,$0F,$1F,$A2,$7D,$28,$CC,$F0,$23,$82
	.dc.b	$57,$78,$CB,$63,$FC,$2F,$87,$9E,$FF,$78,$7E,$FA,$BF,$A5,$D7,$F0
	.dc.b	$1D,$3F,$81,$CF,$F0,$03,$6E,$1E,$18,$79,$BA,$F8,$6B,$FC,$6A,$B4
	.dc.b	$47,$04,$AE,$6D,$C3,$FC,$07,$AF,$FA,$2B,$91,$F1,$5F,$D2,$C9,$7F
	.dc.b	$CC,$BE,$0B,$FA,$77,$C4,$FF,$4F,$F7,$3F,$A8,$7C,$27,$F5,$1F,$7B
	.dc.b	$FA,$97,$C2,$FE,$B3,$6F,$EA,$FF,$04,$F8,$1C,$70,$7F,$45,$87,$C3
	.dc.b	$63,$5F,$EE,$90,$F8,$3C,$71,$21,$F0,$F8,$71,$5F,$D2,$57,$C4,$67
	.dc.b	$E1,$B5,$F1,$3B,$F8,$8F,$F4,$A5,$1C,$DF,$1A,$E9,$7D,$6B,$B6,$1D
	.dc.b	$F2,$F0,$67,$E1,$D7,$DF,$8F,$F5,$6E,$0F,$86,$FF,$58,$E2,$F8,$EF
	.dc.b	$C9,$3F,$AD,$F2,$FC,$C5,$CF,$FC,$F1,$FC,$27,$67,$F4,$DE,$3F,$B5
	.dc.b	$1F,$60,$9C,$5F,$88,$7F,$0C,$D7,$FC,$81,$DD,$FA,$F7,$35,$CF,$93
	.dc.b	$F0,$0D,$8E,$6D,$AE,$57,$E0,$7F,$0B,$DF,$FD,$EF,$AB,$F0,$B7,$EA
	.dc.b	$21,$FE,$15,$C3,$5D,$5F,$02,$30,$7F,$86,$F5,$F8,$AE,$30,$FF,$0D
	.dc.b	$6D,$F2,$A3,$1C,$7F,$87,$7E,$E7,$0F,$D1,$7E,$F7,$1F,$D5,$7F,$01
	.dc.b	$C9,$F7,$DF,$C1,$73,$FD,$F7,$FC,$19,$D9,$FA,$77,$F0,$DD,$1F,$80
	.dc.b	$7B,$F4,$7E,$01,$FC,$56,$5F,$80,$7F,$19,$D1,$F8,$07,$F1,$DD,$1E
	.dc.b	$7E,$4F,$F1,$9F,$1B,$7F,$6C,$63,$87,$E1,$3A,$E2,$FC,$03,$C3,$A3
	.dc.b	$FC,$25,$8F,$E8,$3E,$27,$F3,$84,$7F,$39,$F1,$3F,$9F,$E7,$FD,$07
	.dc.b	$BF,$FA,$1F,$86,$FD,$17,$81,$FA,$3F,$87,$FD,$27,$C1,$7E,$97,$E2
	.dc.b	$3F,$4D,$F0,$9F,$A7,$F8,$9F,$D4,$7C,$37,$EB,$3F,$80,$7F,$B0,$ED
	.dc.b	$F4,$D7,$C4,$67,$D8,$F9,$53,$F5,$FB,$7E,$53,$F5,$BB,$74,$FF,$8E
	.dc.b	$6F,$8B,$4E,$E7,$ED,$71,$F1,$BF,$B7,$CF,$CE,$FE,$E7,$5F,$5B,$FB
	.dc.b	$BD,$FD,$EF,$EF,$78,$7F,$80,$71,$7C,$58,$FC,$3F,$1F,$CA,$F0,$99
	.dc.b	$BE,$27,$DC,$4A,$F3,$7B,$5F,$07,$CC,$F8,$5F,$13,$F0,$FE,$27,$E2
	.dc.b	$7B,$BF,$17,$E2,$7E,$37,$BF,$F1,$FE,$27,$E4,$7C,$0F,$C9,$F8,$9F
	.dc.b	$95,$F0,$7F,$2F,$E2,$7E,$67,$C2,$FC,$DF,$89,$F9,$3E,$2F,$F9,$E3
	.dc.b	$B7,$E4,$BF,$C9,$3B,$75,$C7,$F2,$6F,$93,$F2,$F1,$EC,$FC,$DC,$FC
	.dc.b	$4F,$CF,$D7,$CC,$FD,$1D,$FD,$4F,$D3,$E1,$F7,$3F,$00,$FA,$DC,$9F
	.dc.b	$0D,$FE,$23,$D3,$FC,$CF,$A3,$E6,$6B,$FC,$E1,$F9,$F3,$F4,$4F,$E5
	.dc.b	$DE,$27,$E4,$5F,$E6,$1E,$27,$E4,$5F,$E6,$5E,$27,$E4,$5F,$E6,$9E
	.dc.b	$27,$E4,$5F,$E6,$DE,$27,$E4,$5F,$C9,$78,$9F,$8A,$7F,$8C,$F5,$FF
	.dc.b	$40,$F5,$3F,$8A,$77,$3C,$5D,$D7,$E3,$4E,$67,$E3,$4E,$E7,$E3,$4F
	.dc.b	$19,$F8,$D3,$CE,$7E,$34,$F5,$9E,$D3,$F8,$8F,$83,$F0,$BD,$3F,$89
	.dc.b	$1F,$EC,$8F,$55,$C5,$FE,$9D,$C3,$F1,$D3,$F3,$08,$7E,$AF,$C2,$FA
	.dc.b	$2F,$EB,$79,$7E,$AB,$FA,$FE,$6F,$B2,$FE,$C7,$9F,$EE,$BF,$B3,$E8
	.dc.b	$FC,$0B,$FB,$5E,$9F,$C2,$BF,$B7,$EA,$FC,$4B,$FB,$33,$F8,$FF,$37
	.dc.b	$F6,$8E,$CF,$F2,$5F,$9F,$FD,$7E,$7F,$6D,$DD,$F9,$67,$F7,$1D,$FF
	.dc.b	$9A,$7F,$75,$E0,$7E,$79,$FD,$E7,$83,$FA,$27,$F7,$DE,$17,$E9,$9F
	.dc.b	$E0,$1E,$1F,$EA,$9F,$E0,$5D,$5F,$C3,$78,$FF,$1D,$E0,$FF,$B6,$39
	.dc.b	$7E,$3D,$7E,$40,$FE,$40,$7F,$79,$F0,$7E,$83,$FE,$02,$EA,$FE,$40
	.dc.b	$FD,$63,$FA,$2F,$8B,$FC,$19,$FE,$8F,$E3,$7F,$0A,$7F,$A4,$F8,$FF
	.dc.b	$C3,$9F,$E9,$7E,$47,$F1,$27,$F8,$F1,$F8,$A3,$F9,$07,$7F,$ED,$BB
	.dc.b	$FF,$D7,$1C,$FF,$24,$C4,$BC,$8F,$E1,$B5,$C7,$6F,$64,$F5,$7F,$C3
	.dc.b	$F8,$78,$5F,$F1,$0E,$3E,$57,$FC,$47,$97,$A5,$FF,$12,$E7,$FC,$7A
	.dc.b	$FF,$B0,$F7,$FE,$C1,$FE,$3A,$7C,$53,$F2,$4F,$E4,$7C,$7F,$9B,$AD
	.dc.b	$9F,$DF,$1F,$CE,$F8,$BE,$A3,$FC,$F7,$8F,$EC,$3F,$CF,$F9,$3E,$E3
	.dc.b	$FD,$07,$97,$F0,$0F,$F4,$3E,$6F,$C2,$3F,$D1,$79,$FF,$10,$FF,$47
	.dc.b	$E8,$F8,$BC,$7F,$E9,$78,$FE,$8D,$C9,$8E,$75,$70,$E1,$FC,$16,$1D
	.dc.b	$1F,$C2,$63,$E0,$7F,$0D,$97,$91,$FC,$46,$7E,$87,$F1,$5A,$7B,$1F
	.dc.b	$C6,$6B,$F8,$FD,$3F,$63,$1F,$8F,$DF,$FA,$27,$4B,$A3,$F1,$BC,$1F
	.dc.b	$E4,$B7,$E7,$57,$F9,$5B,$8B,$E9,$3F,$53,$0F,$E0,$E7,$E3,$8F,$E1
	.dc.b	$67,$E3,$8F,$E1,$E7,$E3,$8F,$E2,$67,$E3,$8F,$EC,$9E,$1F,$F1,$D8
	.dc.b	$71,$A7,$F1,$9E,$AF,$F8,$DD,$FF,$12,$78,$9F,$C8,$3A,$BF,$1D,$5F
	.dc.b	$C3,$E7,$F1,$CF,$F9,$A7,$AB,$F1,$CF,$F9,$AF,$AB,$F1,$CF,$F9,$B7
	.dc.b	$AB,$F1,$CF,$F9,$BF,$AB,$DA,$FF,$9C,$75,$F8,$2F,$F1,$EC,$3F,$1E
	.dc.b	$7E,$D7,$7F,$ED,$7C,$3F,$8E,$8F,$D1,$74,$FE,$EB,$7F,$DE,$76,$7F
	.dc.b	$1E,$7F,$7D,$E1,$FF,$1E,$7F,$80,$78,$7F,$C7,$9F,$E0,$5E,$1F,$F1
	.dc.b	$E7,$F8,$27,$87,$FC,$79,$FE,$0D,$E1,$FF,$1E,$7F,$84,$78,$7F,$C7
	.dc.b	$9F,$ED,$19,$FF,$1E,$21,$FC,$84,$FC,$A1,$FD,$AC,$FE,$3A,$7F,$6B
	.dc.b	$3F,$8E,$9F,$DA,$CF,$E3,$A7,$F6,$B3,$F8,$E9,$FD,$AC,$FE,$3A,$7F
	.dc.b	$6B,$3F,$8E,$9F,$AE,$7F,$1C,$FF,$40,$F1,$3F,$A2,$BF,$1B,$E1,$FF
	.dc.b	$C5,$C4,$7F,$C4,$DF,$13,$FE,$31,$F8,$9F,$F1,$97,$BD,$FE,$9E,$FE
	.dc.b	$01,$FC,$2D,$7E,$17,$1B,$C6,$97,$C0,$6F,$DF,$E1,$F0,$3C,$7E,$2B
	.dc.b	$97,$C1,$F3,$F8,$AE,$9F,$0B,$D7,$E2,$BB,$7C,$3F,$7F,$87,$F0,$7C
	.dc.b	$47,$F1,$FE,$0F,$97,$C1,$E0,$DD,$3B,$9F,$27,$7F,$E6,$78,$7C,$98
	.dc.b	$F5,$9C,$9F,$C9,$ED,$9F,$8E,$33,$5E,$BD,$7D,$11,$8A,$6F,$ED,$8C
	.dc.b	$70,$FC,$5C,$BE,$3C,$7F,$93,$F8,$FF,$CB,$5C,$9F,$36,$3F,$CB,$5C
	.dc.b	$BF,$E4,$3E,$E3,$9E,$0F,$EA,$4F,$09,$E4,$7F,$96,$79,$BB,$3E,$03
	.dc.b	$C9,$CD,$F3,$9F,$37,$37,$CE,$7D,$1C,$DF,$39,$F5,$73,$7C,$E7,$D9
	.dc.b	$CD,$F3,$9F,$9B,$B7,$F4,$E7,$E0,$73,$7C,$D7,$FD,$49,$C4,$E6,$F9
	.dc.b	$91,$8E,$6F,$94,$EB,$97,$E4,$BC,$39,$7E,$43,$CB,$97,$D6,$FF,$9F
	.dc.b	$78,$3F,$20,$FE,$48,$FF,$3C,$70,$FC,$9E,$7F,$CC,$F3,$FE,$3F,$E2
	.dc.b	$BF,$28,$4A,$FA,$DF,$15,$F9,$7F,$73,$F3,$1E,$EF,$E6,$7E,$2B,$F3
	.dc.b	$5E,$FF,$E6,$FE,$2B,$F3,$9F,$03,$F9,$DF,$8A,$FF,$7D,$FF,$00,$FF
	.dc.b	$AF,$73,$FF,$61,$2F,$8B,$CF,$C3,$EB,$E2,$F7,$EB,$F0,$F8,$BF,$E9
	.dc.b	$FE,$AF,$F2,$7F,$13,$F8,$1E,$5C,$A7,$6F,$E3,$B2,$61,$D2,$31,$97
	.dc.b	$8B,$E2,$7C,$91,$F9,$1D,$7D,$0F,$E4,$F7,$F6,$3F,$95,$E1,$F8,$0F
	.dc.b	$C2,$3F,$31,$C9,$F1,$5F,$E5,$1E,$37,$E4,$1F,$9A,$43,$FD,$5D,$E3
	.dc.b	$FD,$27,$EA,$1F,$A1,$F0,$BF,$81,$76,$7D,$AE,$0F,$D9,$76,$FD,$E7
	.dc.b	$F6,$9D,$DF,$82,$7F,$6D,$DF,$F8,$67,$F7,$1E,$07,$E2,$DF,$DD,$78
	.dc.b	$5F,$BD,$CB,$FB,$A9,$F9,$03,$F2,$A7,$EC,$67,$96,$5F,$BD,$F0,$BF
	.dc.b	$73,$3F,$9E,$3F,$6D,$8F,$8E,$75,$97,$9D,$E1,$9F,$E4,$32,$FF,$20
	.dc.b	$78,$5E,$F7,$FC,$85,$AF,$EC,$EB,$FC,$D5,$E1,$7E,$DC,$C6,$5F,$CA
	.dc.b	$4F,$CB,$1F,$ED,$5F,$23,$F8,$0B,$FA,$69,$F8,$EF,$ED,$8F,$E1,$59
	.dc.b	$79,$5F,$E2,$59,$FC,$79,$F6,$6F,$EE,$9F,$E2,$DB,$7C,$79,$F8,$1B
	.dc.b	$FC,$7D,$3F,$D7,$1F,$03,$FB,$9B,$FD,$AF,$8B,$FB,$D7,$F6,$FE,$37
	.dc.b	$F0,$17,$F7,$3E,$3F,$F0,$57,$F7,$7E,$47,$F0,$97,$F7,$BE,$4F,$A1
	.dc.b	$F5,$1F,$C4,$4F,$E2,$67,$E1,$8F,$C4,$63,$FC,$61,$FF,$20,$9F,$14
	.dc.b	$FE,$3E,$6B,$3F,$F4,$A7,$17,$CF,$7F,$D2,$DC,$7F,$4D,$FA,$B3,$F1
	.dc.b	$CF,$E1,$07,$F1,$C3,$EE,$9F,$7C,$FE,$20,$7F,$1C,$3F,$0C,$7E,$20
	.dc.b	$FE,$2F,$AF,$C7,$8F,$E3,$91,$83,$57,$F9,$47,$91,$D0,$F8,$E7,$80
	.dc.b	$F8,$E7,$DA,$3C,$CF,$F5,$3E,$4F,$54,$7F,$84,$AF,$DC,$FF,$85,$30
	.dc.b	$FC,$4F,$90,$E7,$FD,$94,$FF,$86,$F9,$DD,$1F,$B7,$CB,$FA,$F3,$FD
	.dc.b	$C7,$C8,$FD,$ED,$7F,$32,$EA,$FE,$02,$FF,$34,$EB,$FE,$0A,$FF,$36
	.dc.b	$EC,$FE,$12,$FF,$1A,$ED,$FE,$3B,$DD,$FC,$3E,$3F,$8E,$F7,$FF,$14
	.dc.b	$7F,$8B,$1F,$D8,$BC,$3F,$F5,$A3,$FE,$40,$F6,$FF,$DC,$07,$F9,$B3
	.dc.b	$C2,$78,$7C,$77,$FF,$01,$C7,$F1,$CA,$FE,$3D,$E2,$F8,$1F,$E4,$19
	.dc.b	$F9,$2B,$FB,$D6,$9E,$87,$FB,$E6,$BE,$C7,$FB,$F6,$DF,$01,$FF,$00
	.dc.b	$6F,$F1,$79,$FF,$81,$70,$FE,$D7,$2F,$C7,$1F,$91,$AF,$DD,$BF,$98
	.dc.b	$7F,$7E,$FF,$1B,$3F,$B5,$1F,$A0,$3E,$E1,$F7,$8F,$EB,$47,$E3,$DF
	.dc.b	$E3,$6F,$E1,$9F,$C4,$3F,$B0,$3F,$8C,$1F,$C6,$88,$FF,$1E,$7F,$1F
	.dc.b	$3F,$C6,$7B,$FF,$7D,$CB,$FE,$29,$F0,$3F,$81,$57,$F0,$43,$F4,$07
	.dc.b	$E8,$8F,$31,$FD,$A4,$F5,$1F,$CD,$8F,$E2,$47,$F1,$4B,$FD,$A3,$FB
	.dc.b	$23,$FC,$D1,$BB,$E2,$BF,$AF,$92,$3F,$AA,$35,$FE,$C2,$D9,$FB,$AD
	.dc.b	$C7,$FC,$5D,$5C,$EB,$89,$0E,$25,$AE,$2D,$9E,$35,$FC,$66,$7F,$8D
	.dc.b	$CF,$F8,$48,$FF,$1B,$75,$F7,$7F,$DB,$D7,$DD,$EA,$49,$EA,$DB,$EA
	.dc.b	$F7,$38,$DF,$3C,$57,$CE,$F8,$92,$F8,$DF,$33,$89,$FC,$7A,$B8,$DF
	.dc.b	$3B,$E5,$5C,$2F,$95,$F0,$39,$5F,$07,$C8,$D7,$35,$C9,$E1,$3C,$A7
	.dc.b	$4D,$9D,$47,$AB,$E2,$FF,$8F,$52,$3F,$DC,$15,$FB,$49,$F3,$61,$2B
	.dc.b	$C3,$7C,$89,$19,$8F,$1B,$6B,$C0,$FF,$B1,$63,$F1,$7F,$15,$FE,$28
	.dc.b	$23,$FE,$1D,$F7,$BF,$C5,$7F,$0B,$FE,$2C,$F6,$3F,$C5,$BF,$15,$FE
	.dc.b	$2E,$F6,$BF,$C5,$FE,$47,$F8,$C3,$DB,$FF,$18,$FC,$57,$F8,$CB,$DD
	.dc.b	$FF,$19,$FC,$57,$F8,$D3,$DF,$FF,$1A,$FC,$57,$87,$87,$FD,$23,$E5
	.dc.b	$F8,$F8,$11,$8F,$83,$FC,$67,$3F,$FA,$0F,$D4,$FF,$1A,$3F,$DA,$AB
	.dc.b	$FC,$FF,$6A,$E9,$3F,$E8,$98,$E9,$7F,$C7,$30,$F1,$5F,$F8,$E7,$2F
	.dc.b	$33,$FE,$3B,$D3,$D4,$FF,$8F,$76,$F7,$3F,$E3,$FE,$0F,$84,$FF,$90
	.dc.b	$78,$BF,$C9,$FC,$7F,$22,$3F,$C9,$FC,$9F,$37,$D9,$FF,$2F,$F3,$FD
	.dc.b	$49,$FA,$FC,$FF,$4F,$83,$F1,$C4,$8F,$C3,$B7,$A7,$F0,$ED,$AE,$53
	.dc.b	$F7,$0F,$C0,$F4,$FE,$22,$BF,$0F,$5F,$BC,$8E,$D3,$D0,$93,$D1,$B5
	.dc.b	$D3,$87,$F0,$B1,$E6,$21,$DC,$FE,$23,$87,$D7,$7F,$5A,$7E,$2E,$BF
	.dc.b	$44,$DA,$FD,$F3,$67,$83,$FA,$B3,$07,$39,$E2,$F4,$21,$B2,$BA,$70
	.dc.b	$47,$04,$95,$FC,$5F,$8A,$6F,$FE,$86,$8F,$E4,$D7,$E2,$3F,$25,$7C
	.dc.b	$70,$F3,$B6,$3C,$D8,$7B,$1F,$1B,$5F,$13,$63,$84,$74,$8C,$12,$7F
	.dc.b	$65,$E7,$BD,$2E,$71,$E0,$9D,$57,$82,$30,$47,$04,$3B,$39,$69,$93
	.dc.b	$DC,$43,$B9,$EC,$C3,$E3,$63,$F1,$63,$FA,$11,$B3,$DB,$3D,$5C,$63
	.dc.b	$EE,$6D,$FA,$1F,$09,$0C,$CF,$97,$1F,$81,$1D,$0F,$31,$E4,$38,$14
	.dc.b	$E2,$68,$E0,$68,$A7,$93,$CC,$77,$7F,$88,$20,$7F,$88,$CF,$EF,$FA
	.dc.b	$FF,$50,$7A,$9C,$F1,$F6,$3E,$12,$1D,$8A,$FD,$EB,$F5,$BC,$23,$44
	.dc.b	$70,$4C,$7E,$34,$78,$08,$7E,$33,$3F,$A6,$FF,$B3,$DF,$CB,$BE,$08
	.dc.b	$E2,$DC,$BE,$A3,$4F,$8D,$5D,$E3,$F1,$FA,$7F,$0B,$7F,$82,$C7,$F0
	.dc.b	$32,$57,$F0,$3D,$B5,$7F,$C0,$B2,$FE,$0A,$FF,$03,$7F,$AD,$63,$E6
	.dc.b	$8F,$F7,$7B,$D2,$B9,$47,$F0,$EC,$30,$47,$A5,$F6,$5B,$BB,$BF,$9D
	.dc.b	$7C,$1F,$F3,$C9,$FE,$C9,$D2,$8C,$68,$8C,$4F,$F4,$4F,$83,$CC,$3E
	.dc.b	$0E,$F9,$98,$F8,$3C,$FC,$2E,$BE,$13,$7F,$0D,$C3,$E1,$78,$FC,$3F
	.dc.b	$2F,$86,$E7,$F1,$1D,$3E,$1F,$AF,$C4,$F6,$F8,$8F,$D4,$70,$FF,$92
	.dc.b	$39,$BF,$D2,$FD,$5F,$DE,$63,$FA,$BC,$F0,$7F,$AC,$DF,$47,$FA,$DE
	.dc.b	$3E,$07,$FA,$EE,$7E,$47,$FA,$FE,$BE,$87,$FB,$0E,$FE,$C7,$FB,$1F
	.dc.b	$0F,$C0,$E3,$F8,$3C,$B2,$7F,$C4,$1D,$3F,$53,$9B,$EF,$F4,$FD,$39
	.dc.b	$FB,$9C,$1F,$4F,$83,$EF,$91,$C1,$FD,$C6,$3F,$54,$48,$FF,$33,$3C
	.dc.b	$8F,$C0,$1F,$78,$A6,$6B,$C2,$E5,$AF,$E0,$BA,$3F,$02,$FF,$7E,$C7
	.dc.b	$D4,$FF,$81,$AF,$FB,$F6,$5F,$7B,$3F,$85,$39,$3F,$0F,$A7,$C7,$71
	.dc.b	$8F,$E1,$F5,$F9,$91,$8E,$86,$FF,$46,$31,$C1,$F5,$5D,$70,$FD,$97
	.dc.b	$87,$17,$E0,$38,$3F,$CC,$31,$F8,$13,$FC,$BF,$D5,$F8,$57,$FC,$2F
	.dc.b	$C9,$F8,$97,$FC,$33,$CB,$F8,$D7,$FC,$37,$CD,$F9,$17,$FC,$3B,$CF
	.dc.b	$F9,$57,$FC,$3F,$D1,$F9,$97,$FC,$43,$D3,$F0,$5F,$D0,$1F,$1C,$87
	.dc.b	$CB,$3F,$53,$F1,$5F,$A4,$23,$FE,$2A,$F8,$5F,$D3,$FB,$DF,$A8,$F8
	.dc.b	$5F,$D4,$FC,$07,$EB,$BF,$80,$5D,$77,$E0,$B1,$E1,$E7,$E1,$B5,$D3
	.dc.b	$BF,$86,$E1,$F0,$BC,$7D,$8E,$5F,$0D,$CF,$DA,$F9,$19,$FE,$98,$9C
	.dc.b	$3F,$20,$FE,$6F,$D5,$E0,$7F,$CC,$9B,$2B,$9C,$6A,$BF,$9E,$F8,$0A
	.dc.b	$63,$E2,$AD,$1C,$B3,$F3,$B8,$AE,$1A,$FB,$23,$16,$DF,$E1,$63,$F1
	.dc.b	$27,$FC,$A3,$C3,$F2,$1F,$94,$7F,$95,$39,$3E,$6B,$F4,$71,$F2,$79
	.dc.b	$3F,$2C,$91,$DB,$CA,$F9,$DE,$DF,$CF,$F7,$3E,$87,$C5,$7D,$1F,$7B
	.dc.b	$E9,$7C,$57,$D3,$F8,$0F,$A9,$F1,$5F,$57,$E0,$BE,$B7,$C5,$7F,$24
	.dc.b	$FE,$01,$7C,$2D,$F8,$AC,$7C,$3E,$7E,$2B,$5F,$15,$FC,$E7,$BB,$E7
	.dc.b	$72,$B1,$4E,$F9,$3E,$34,$30,$E9,$18,$CB,$C2,$FC,$8D,$3C,$AF,$CA
	.dc.b	$DB,$D2,$FC,$CE,$0F,$6B,$F3,$B8,$BE,$0B,$F0,$CF,$A3,$CB,$F1,$9F
	.dc.b	$95,$E0,$B9,$FE,$6E,$FF,$E8,$A7,$F9,$FF,$8B,$F4,$EB,$F7,$9D,$3F
	.dc.b	$5D,$FE,$03,$D9,$F8,$B6,$F6,$7D,$F7,$3D,$BF,$84,$77,$DD,$F8,$97
	.dc.b	$8F,$7F,$E4,$1E,$7E,$07,$E6,$9F,$F5,$C7,$C2,$FF,$A5,$9F,$CD,$1A
	.dc.b	$3F,$17,$E1,$7E,$8D,$FC,$6F,$87,$FA,$77,$F1,$FC,$1F,$AB,$7F,$23
	.dc.b	$E2,$7E,$BD,$FC,$9F,$89,$EB,$E2,$FF,$54,$F0,$7B,$DF,$F5,$5F,$0F
	.dc.b	$C4,$9F,$F5,$17,$17,$FA,$4F,$D7,$FD,$2F,$1F,$F8,$C3,$1F,$87,$C9
	.dc.b	$F3,$71,$FC,$0C,$FF,$8A,$27,$F4,$3E,$27,$E4,$DF,$D1,$F8,$9F,$93
	.dc.b	$7F,$4B,$E2,$7E,$4D,$FD,$3F,$89,$F9,$37,$F5,$3E,$27,$E4,$FB,$BF
	.dc.b	$0F,$1F,$CC,$BC,$0F,$C5,$3F,$8C,$F0,$FF,$26,$FE,$CF,$DD,$FF,$3B
	.dc.b	$B7,$DD,$CB,$C6,$7F,$36,$FE,$53,$C6,$EC,$F8,$0F,$C9,$1E,$43,$F2
	.dc.b	$47,$A0,$FC,$91,$EC,$3F,$24,$7C,$03,$F2,$47,$C6,$3F,$26,$7F,$92
	.dc.b	$BD,$CF,$F0,$B3,$FC,$B0,$FE,$00,$43,$FA,$67,$2F,$D5,$E5,$FF,$6A
	.dc.b	$73,$7D,$97,$E9,$9F,$76,$7F,$2B,$CF,$F8,$17,$F2,$FD,$1F,$85,$7F
	.dc.b	$33,$D3,$F8,$97,$F3,$7D,$5F,$CA,$DF,$F2,$31,$F1,$C8,$7F,$8D,$0F
	.dc.b	$E3,$7D,$9F,$99,$E8,$FB,$FD,$BF,$9D,$7F,$03,$DD,$FA,$17,$F0,$7D
	.dc.b	$FF,$A5,$7F,$0B,$E0,$7F,$22,$C3,$FD,$E3,$E4,$FF,$47,$7F,$DE,$7E
	.dc.b	$57,$EC,$9F,$85,$5F,$B5,$7E,$47,$81,$F9,$2F,$03,$F2,$1C,$36,$7E
	.dc.b	$7D,$7E,$4B,$BF,$F2,$11,$FB,$8E,$FF,$C8,$3F,$BA,$EF,$FC,$83,$FB
	.dc.b	$CE,$FF,$C8,$3F,$BE,$EF,$FC,$1B,$F8,$FE,$AF,$C3,$BF,$89,$3F,$8D
	.dc.b	$F9,$1F,$C8,$73,$FE,$93,$DF,$FC,$82,$7F,$28,$7F,$20,$F2,$3F,$33
	.dc.b	$AF,$F2,$69,$FE,$5B,$3F,$8F,$3F,$86,$F9,$1F,$C8,$1F,$E1,$FE,$47
	.dc.b	$F2,$07,$F8,$8F,$91,$FC,$81,$FE,$27,$E4,$7F,$20,$7F,$8A,$F9,$5F
	.dc.b	$11,$FE,$BF,$C7,$FB,$97,$F1,$C7,$F8,$5F,$C8,$FD,$F3,$FE,$64,$F0
	.dc.b	$BF,$1C,$FF,$99,$BC,$2F,$C7,$3F,$DA,$37,$FC,$76,$DF,$BE,$F0,$BF
	.dc.b	$1C,$FF,$00,$F0,$BF,$1C,$FF,$02,$F0,$BF,$1C,$FF,$04,$F0,$BF,$1D
	.dc.b	$87,$E2,$CF,$F2,$06,$1F,$DC,$3C,$97,$97,$FC,$8F,$CB,$FE,$E5,$E2
	.dc.b	$FE,$3E,$BF,$88,$F8,$BF,$8F,$7F,$89,$F8,$BF,$8F,$7F,$8A,$F8,$BF
	.dc.b	$8F,$7F,$8B,$F8,$BF,$8F,$CF,$FC,$6D,$E2,$FE,$3D,$FF,$1C,$78,$BF
	.dc.b	$8F,$7E,$21,$FB,$1F,$1B,$F6,$BA,$7F,$70,$8F,$DC,$9F,$BB,$3E,$79
	.dc.b	$FB,$F3,$E9,$9F,$C0,$CF,$AE,$7F,$07,$3E,$D9,$FC,$2C,$FB,$E7,$E0
	.dc.b	$78,$3F,$09,$E5,$7E,$1B,$2F,$C5,$F7,$FF,$69,$AF,$E4,$1E,$07,$E3
	.dc.b	$D6,$1F,$DB,$E3,$58,$7E,$36,$B9,$56,$BC,$0F,$C7,$BF,$9D,$3F,$8E
	.dc.b	$9F,$A1,$3F,$8E,$9F,$A5,$3F,$8E,$9F,$A9,$3F,$8E,$9F,$D8,$BC,$6F
	.dc.b	$F1,$E7,$6F,$E3,$F1,$F8,$D3,$FE,$17,$A7,$C5,$7F,$82,$C9,$1F,$E0
	.dc.b	$2F,$73,$FC,$1D,$E9,$FF,$83,$FD,$FF,$F0,$87,$C0,$7F,$84,$7E,$07
	.dc.b	$FC,$25,$E6,$FF,$84,$FE,$0B,$FC,$29,$F1,$5F,$E1,$5F,$47,$FC,$2D
	.dc.b	$F1,$5F,$E1,$7F,$85,$FF,$0C,$7C,$57,$F8,$67,$E1,$FF,$C3,$5F,$15
	.dc.b	$FD,$03,$9F,$FC,$DA,$4E,$9F,$DB,$F4,$EB,$8F,$FC,$2B,$7C,$9F,$F0
	.dc.b	$BE,$3D,$9F,$F0,$CE,$7E,$27,$FC,$37,$AF,$99,$FF,$0E,$EF,$EA,$7F
	.dc.b	$C3,$FC,$3E,$E7,$E0,$1F,$E2,$2E,$4F,$86,$FC,$76,$E9,$FC,$15,$FF
	.dc.b	$19,$72,$71,$9F,$EC,$15,$D5,$FE,$C5,$7E,$17,$FB,$26,$1E,$57,$FB
	.dc.b	$36,$3E,$97,$FB,$46,$5E,$D7,$FB,$56,$7F,$05,$FF,$48,$73,$3D,$7F
	.dc.b	$ED,$31,$86,$EB,$FC,$67,$83,$E7,$4F,$F8,$F3,$87,$E9,$3F,$E3,$EE
	.dc.b	$2F,$AC,$FF,$90,$38,$FE,$D3,$FE,$42,$E4,$FB,$CF,$F9,$13,$97,$F0
	.dc.b	$4F,$F9,$1B,$9B,$E0,$BF,$1C,$87,$E3,$7E,$27,$FC,$AC,$47,$FC,$E5
	.dc.b	$FC,$02,$7F,$DF,$AB,$E2,$EF,$C4,$E3,$E2,$73,$F1,$3A,$F2,$B7,$F1
	.dc.b	$3C,$3E,$03,$8F,$C4,$F2,$F8,$2E,$7F,$13,$D3,$E1,$3A,$FC,$4F,$6F
	.dc.b	$86,$FC,$77,$8B,$FC,$6F,$3F,$17,$0F,$F1,$E3,$F6,$5E,$2E,$DF,$3D
	.dc.b	$F3,$7D,$38,$F7,$7D,$79,$F8,$DF,$6E,$BE,$77,$DF,$B7,$AD,$F8,$3C
	.dc.b	$1F,$03,$87,$E0,$C7,$C5,$3F,$C5,$1C,$7F,$D0,$FB,$7F,$CE,$DE,$2F
	.dc.b	$E4,$37,$FE,$B2,$FD,$0A,$FF,$17,$73,$7D,$47,$FC,$61,$CF,$F6,$1F
	.dc.b	$F1,$97,$47,$DC,$7F,$C6,$9D,$3F,$80,$7F,$C6,$DD,$5F,$84,$7F,$C7
	.dc.b	$1D,$7F,$88,$7F,$C7,$5D,$9F,$C6,$F4,$FC,$87,$83,$F2,$3B,$BF,$6D
	.dc.b	$5F,$C6,$6B,$84,$7F,$69,$BE,$8F,$F6,$BC,$3C,$0F,$F6,$DC,$7C,$8F
	.dc.b	$F6,$FC,$BD,$0F,$F7,$1C,$FD,$8F,$F7,$3D,$3E,$03,$FB,$1C,$BF,$1F
	.dc.b	$E5,$FF,$42,$F1,$FF,$8D,$C7,$E4,$A7,$4F,$F9,$4F,$C7,$FD,$B4,$7F
	.dc.b	$7D,$F1,$FF,$6C,$FF,$7F,$F1,$FF,$6C,$FF,$80,$7C,$7F,$DB,$3F,$E0
	.dc.b	$3E,$1F,$DB,$3F,$E0,$5F,$1F,$F6,$CF,$F8,$1F,$C7,$FD,$B3,$FE,$09
	.dc.b	$EF,$F9,$1A,$7F,$83,$F4,$F9,$67,$F1,$A3,$F6,$87,$F1,$A3,$F6,$87
	.dc.b	$F1,$A3,$F6,$87,$F1,$A3,$F6,$87,$F1,$A3,$F6,$87,$F1,$A3,$F6,$87
	.dc.b	$F1,$A3,$FC,$E5,$CD,$FC,$6B,$6F,$8C,$FE,$2E,$3F,$A8,$3F,$33,$9F
	.dc.b	$FC,$4B,$C1,$F4,$1F,$F1,$3F,$0F,$D4,$7F,$C5,$3C,$5F,$61,$FF,$15
	.dc.b	$F1,$FD,$C7,$FC,$5B,$C9,$F8,$07,$FC,$5F,$CB,$F8,$47,$FC,$63,$CD
	.dc.b	$F8,$BF,$1F,$F8,$BC,$7F,$BF,$1B,$A7,$E2,$CF,$8C,$7E,$50,$F8,$C7
	.dc.b	$E6,$0F,$8C,$7E,$70,$F8,$C7,$E8,$0F,$8C,$7E,$90,$F8,$C7,$EA,$0F
	.dc.b	$8D,$C1,$F8,$63,$F0,$FB,$7E,$2F,$A7,$F6,$91,$FE,$43,$7F,$6F,$3F
	.dc.b	$C6,$BA,$BF,$8B,$BF,$C7,$3C,$6F,$E2,$EF,$F1,$EF,$0B,$F8,$BB,$FC
	.dc.b	$83,$C6,$FE,$2E,$FF,$22,$F1,$BF,$8B,$BF,$C9,$3C,$6F,$E2,$EF,$F2
	.dc.b	$6F,$1B,$F8,$BB,$FE,$D0,$BF,$D8,$6D,$FC,$66,$3F,$CB,$A6,$CF,$E2
	.dc.b	$E7,$33,$F8,$B9,$DC,$FE,$2E,$78,$CF,$E2,$E7,$9C,$FE,$2E,$7A,$CF
	.dc.b	$E2,$E7,$BC,$FE,$72,$7C,$47,$FC,$D7,$9F,$F1,$7E,$7F,$D8,$CF,$E4
	.dc.b	$0F,$D9,$4F,$F9,$C3,$BF,$F6,$2F,$F9,$CB,$BF,$F6,$2F,$F9,$D3,$BF
	.dc.b	$F6,$2F,$F9,$DB,$BF,$F6,$2F,$F9,$E3,$BF,$F6,$2F,$F9,$EB,$BF,$F6
	.dc.b	$2F,$F9,$F3,$BF,$F6,$2F,$E2,$E7,$F1,$C6,$32,$FF,$17,$4F,$07,$FC
	.dc.b	$61,$5D,$1F,$F1,$94,$F8,$1F,$F1,$A5,$F9,$1F,$F1,$B6,$1E,$87,$FC
	.dc.b	$71,$8F,$B1,$FF,$1D,$65,$FA,$DF,$03,$F9,$CC,$7F,$A9,$0F,$F2,$1F
	.dc.b	$3F,$E7,$A3,$F9,$5C,$7F,$AE,$B9,$BF,$95,$C6,$23,$FB,$E6,$5E,$A7
	.dc.b	$5A,$FB,$5A,$74,$8D,$3C,$0C,$CF,$3A,$EB,$7E,$12,$57,$08,$CB,$6F
	.dc.b	$B5,$F3,$12,$8C,$F3,$FF,$9C,$AF,$9D,$78,$1A,$F1,$AE,$AE,$6F,$B3
	.dc.b	$E9,$30,$43,$85,$F8,$A7,$8C,$60,$E0,$5C,$3D,$D3,$B2,$46,$A7,$89
	.dc.b	$CB,$0F,$F4,$4F,$8E,$F8,$9F,$F0,$99,$2B,$FC,$2F,$F1,$1F,$E1,$6F
	.dc.b	$89,$FF,$0B,$FB,$BF,$E1,$8F,$77,$FC,$33,$E8,$FF,$86,$BD,$DF,$F0
	.dc.b	$DF,$A7,$FB,$6C,$7F,$D0,$3D,$5F,$1D,$FF,$10,$6D,$78,$3F,$D5,$7E
	.dc.b	$A6,$DF,$F5,$6F,$BB,$6D,$EB,$5F,$77,$1E,$C6,$7D,$DF,$E0,$B8,$3D
	.dc.b	$CF,$F5,$AE,$9D,$B9,$FF,$83,$3F,$E9,$43,$07,$92,$B9,$B8,$AF,$F5
	.dc.b	$EE,$1E,$1C,$7C,$59,$F8,$F4,$F2,$63,$EE,$C3,$FC,$77,$BF,$C2,$7F
	.dc.b	$C7,$BC,$3F,$19,$FF,$1F,$F1,$FF,$8F,$B9,$3E,$64,$7E,$2A,$FF,$C8
	.dc.b	$5A,$39,$BE,$94,$FE,$9E,$3E,$DB,$F6,$67,$ED,$C6,$09,$D3,$F8,$EE
	.dc.b	$3F,$BB,$3A,$8C,$37,$A7,$F1,$F5,$C1,$B3,$AB,$EC,$49,$D3,$F9,$3B
	.dc.b	$EA,$E6,$78,$4E,$A1,$D3,$FA,$3B,$FF,$26,$C7,$F9,$3A,$7F,$78,$DA
	.dc.b	$FF,$3C,$B6,$76,$FE,$C0,$C1,$89,$E2,$E4,$87,$E8,$8D,$9E,$07,$6E
	.dc.b	$4D,$B8,$24,$F8,$A3,$2D,$78,$12,$37,$3E,$68,$C4,$EE,$73,$1E,$78
	.dc.b	$C1,$27,$84,$EA,$7C,$D3,$97,$84,$EA,$31,$0D,$DE,$F7,$9A,$B4,$49
	.dc.b	$E2,$43,$AB,$96,$99,$3E,$39,$0E,$86,$4A,$70,$21,$B3,$E0,$BD,$5C
	.dc.b	$5F,$B5,$B7,$EA,$6C,$ED,$EA,$60,$F4,$5F,$B9,$EA,$78,$CA,$78,$8E
	.dc.b	$24,$32,$73,$34,$60,$D5,$79,$5B,$5E,$56,$D7,$95,$E0,$49,$E0,$F8
	.dc.b	$AB,$CD,$18,$25,$73,$9E,$0D,$8C,$D7,$56,$CF,$B5,$AF,$A1,$AF,$AC
	.dc.b	$87,$98,$E1,$3D,$4F,$C3,$D7,$91,$FD,$A9,$2B,$F7,$E4,$9F,$DE,$65
	.dc.b	$FC,$15,$A7,$EF,$63,$4F,$F4,$D9,$E4,$EA,$FF,$80,$65,$FC,$11,$FE
	.dc.b	$F1,$93,$CD,$FE,$3B,$CE,$F7,$3F,$9E,$1F,$CF,$7D,$EF,$E7,$DE,$F7
	.dc.b	$F3,$FD,$FF,$A0,$7B,$7F,$D0,$7D,$CF,$E8,$5E,$FF,$F4,$3F,$87,$FE
	.dc.b	$89,$F0,$3F,$D1,$7E,$1F,$FA,$37,$C1,$FF,$47,$F8,$7F,$E9,$1F,$0B
	.dc.b	$FD,$27,$E1,$FF,$A6,$7F,$00,$BE,$1E,$FC,$46,$3E,$22,$74,$FF,$41
	.dc.b	$A4,$6A,$31,$5C,$DF,$E8,$DC,$AC,$7C,$11,$8C,$FC,$8E,$B5,$F4,$3C
	.dc.b	$37,$F6,$3C,$B8,$7E,$03,$D3,$8F,$E2,$3D,$B9,$7E,$3B,$FD,$0F,$AB
	.dc.b	$E8,$BF,$4F,$AF,$EA,$F5,$FD,$19,$CB,$63,$3D,$9F,$76,$B8,$1F,$66
	.dc.b	$B7,$19,$AE,$13,$FD,$47,$6F,$B0,$F8,$CF,$83,$D3,$F6,$1F,$87,$D3
	.dc.b	$F6,$1F,$8B,$D3,$F6,$1F,$8F,$D3,$F6,$1F,$93,$D3,$F6,$1F,$97,$D3
	.dc.b	$F6,$1F,$9B,$D3,$F6,$39,$BE,$94,$7F,$67,$1D,$8E,$9F,$AF,$1F,$67
	.dc.b	$A7,$ED,$BB,$9D,$75,$73,$8E,$A7,$DA,$EA,$FB,$2E,$F8,$1D,$5F,$6A
	.dc.b	$31,$D5,$F6,$9D,$75,$7D,$A7,$87,$57,$DA,$79,$75,$7D,$A7,$A7,$57
	.dc.b	$DA,$7F,$B2,$F5,$FF,$80,$37,$7C,$4F,$E5,$89,$3F,$57,$B3,$EC,$BF
	.dc.b	$B3,$30,$43,$ED,$93,$AB,$EC,$B7,$AF,$EE,$11,$F9,$37,$E0,$7F,$41
	.dc.b	$7E,$07,$CA,$69,$C5,$7F,$8D,$E7,$84,$FB,$8F,$C0,$9F,$9B,$3F,$02
	.dc.b	$7E,$7C,$FC,$09,$FA,$33,$F0,$27,$E9,$CF,$C0,$9F,$4C,$FA,$A7,$D9
	.dc.b	$7F,$70,$FE,$D7,$87,$EE,$CF,$8F,$8E,$E1,$F2,$E7,$EF,$57,$DB,$E5
	.dc.b	$FC,$63,$F7,$79,$BF,$20,$FD,$FE,$7F,$CA,$3F,$81,$E8,$FC,$C3,$F8
	.dc.b	$3E,$9F,$CE,$3F,$85,$EA,$FD,$03,$E5,$DB,$E2,$E7,$F6,$D4,$7F,$8E
	.dc.b	$76,$7C,$1F,$F1,$23,$F8,$9F,$C1,$FD,$CE,$BF,$E5,$38,$7D,$56,$DF
	.dc.b	$DF,$3E,$D1,$78,$7E,$E9,$27,$87,$77,$EE,$A7,$51,$FE,$46,$C7,$FD
	.dc.b	$A7,$5F,$67,$1F,$15,$79,$72,$F4,$BF,$7C,$FE,$24,$7D,$C3,$F3,$A7
	.dc.b	$DC,$3F,$42,$7D,$C3,$F4,$A7,$DC,$3F,$55,$D5,$FA,$CA,$FF,$3F,$75
	.dc.b	$FF,$27,$9F,$BB,$1F,$8B,$ED,$FD,$EB,$F6,$E3,$F7,$44,$3F,$05,$C9
	.dc.b	$F7,$23,$9F,$27,$DC,$7A,$F2,$7D,$C7,$BF,$27,$DC,$7C,$3C,$9F,$71
	.dc.b	$F1,$F2,$7D,$C7,$CB,$C9,$F7,$1F,$CF,$6D,$F1,$38,$FF,$87,$3F,$73
	.dc.b	$4F,$D2,$C6,$1F,$D4,$90,$FD,$21,$F6,$1F,$B6,$4E,$3F,$AC,$DE,$6F
	.dc.b	$C1,$46,$08,$E0,$F2,$72,$7E,$36,$BB,$61,$E6,$7F,$1D,$E8,$7F,$5E
	.dc.b	$3F,$AF,$FC,$1F,$08,$FF,$58,$E7,$EF,$BF,$EE,$7A,$7C,$37,$FB,$AE
	.dc.b	$BF,$1D,$FE,$EF,$B7,$CB,$7F,$BC,$EF,$F3,$DF,$EF,$7C,$1F,$4D,$FD
	.dc.b	$86,$3F,$5F,$4F,$B4,$FE,$E7,$A3,$EE,$CF,$F1,$FE,$3F,$ED,$53,$FC
	.dc.b	$BF,$BF,$F0,$D3,$FE,$06,$E6,$FE,$61,$87,$E2,$3B,$FF,$98,$3F,$8A
	.dc.b	$EF,$FE,$60,$FE,$33,$BF,$F9,$83,$F8,$EE,$FF,$83,$97,$F7,$BE,$4F
	.dc.b	$88,$FE,$7C,$F9,$39,$7C,$B7,$EB,$73,$7D,$07,$F9,$65,$FE,$B5,$FD
	.dc.b	$59,$39,$FE,$C3,$AA,$FD,$8F,$3F,$D7,$8E,$A7,$EA,$CF,$E0,$C7,$D6
	.dc.b	$3C,$87,$F8,$3F,$9F,$EB,$53,$AB,$FC,$C1,$3F,$B0,$3F,$89,$1F,$CA
	.dc.b	$CF,$E2,$C7,$F2,$B3,$F8,$D1,$FC,$AC,$FE,$3C,$7F,$2B,$3F,$91,$1F
	.dc.b	$E1,$A3,$F9,$51,$F5,$EF,$FA,$97,$17,$F2,$A8,$FE,$8B,$D5,$F7,$5F
	.dc.b	$F7,$CF,$27,$E0,$5F,$F7,$DF,$2F,$E1,$5F,$F7,$EF,$37,$E2,$5F,$F7
	.dc.b	$FF,$3F,$E3,$5F,$F8,$03,$D1,$F9,$17,$F2,$67,$F0,$E3,$F9,$F6,$BF
	.dc.b	$59,$FE,$81,$C3,$FA,$FF,$07,$FC,$95,$1F,$D6,$3D,$2F,$F1,$D9,$FE
	.dc.b	$3C,$F8,$3F,$F1,$EF,$A9,$FE,$3E,$F0,$BF,$C7,$FE,$AF,$F9,$03,$E0
	.dc.b	$FF,$C8,$3C,$DF,$E4,$2F,$83,$FF,$21,$FB,$1F,$E4,$4F,$83,$FF,$22
	.dc.b	$FB,$5F,$E4,$6F,$83,$FF,$23,$FB,$BF,$BF,$ED,$75,$FF,$BD,$BF,$80
	.dc.b	$3F,$DB,$71,$8E,$2F,$F0,$9C,$3E,$13,$85,$6F,$E1,$7E,$67,$81,$FE
	.dc.b	$17,$EF,$B6,$9D,$E1,$CB,$0B,$6A,$EF,$87,$F8,$1F,$2F,$35,$FF,$95
	.dc.b	$F4,$F5,$3F,$E5,$9D,$BD,$CF,$F9,$6F,$83,$E1,$3F,$E5,$DE,$2F,$8C
	.dc.b	$FF,$97,$F9,$3E,$53,$FE,$61,$E6,$F9,$AF,$F5,$AE,$7F,$A3,$AF,$8F
	.dc.b	$E1,$7C,$BB,$1B,$F0,$D8,$F0,$B3,$F0,$FA,$F8,$7D,$FC,$47,$0F,$88
	.dc.b	$E3,$F1,$3C,$BE,$27,$9F,$C5,$74,$F8,$AE,$BF,$13,$DB,$DD,$EF,$EF
	.dc.b	$78,$3E,$27,$C3,$F0,$1F,$33,$DB,$FF,$20,$F3,$7F,$0F,$3F,$B8,$F8
	.dc.b	$5F,$57,$9B,$E5,$57,$19,$CF,$7F,$DB,$C7,$B4,$60,$FB,$79,$F8,$DC
	.dc.b	$5F,$D3,$D7,$D1,$18,$E7,$70,$7B,$63,$1C,$5F,$05,$D7,$27,$C5,$78
	.dc.b	$73,$7C,$AE,$79,$1F,$E0,$7E,$9F,$A2,$FF,$88,$7C,$8F,$CE,$BF,$E9
	.dc.b	$5F,$23,$F3,$AF,$FA,$5F,$C8,$FC,$EB,$FE,$99,$F2,$3F,$3A,$FF,$A6
	.dc.b	$FC,$8F,$CE,$BF,$E9,$DF,$23,$F3,$AF,$A4,$FB,$7E,$37,$DC,$F1,$FF
	.dc.b	$23,$E8,$FF,$56,$E2,$FC,$B5,$7B,$EB,$F0,$2F,$23,$FC,$6D,$7D,$9F
	.dc.b	$F1,$C6,$1E,$27,$FC,$75,$8F,$99,$FF,$1E,$65,$EA,$7F,$C7,$D9,$FB
	.dc.b	$9F,$80,$7E,$4C,$87,$F7,$0C,$7F,$20,$FD,$1D,$BF,$22,$FD,$3F,$07
	.dc.b	$F2,$2F,$D5,$F0,$7F,$22,$FD,$7F,$07,$F2,$2F,$D9,$F0,$7F,$22,$FD
	.dc.b	$BF,$07,$F2,$2F,$DD,$F0,$7F,$22,$FD,$FF,$07,$F2,$2F,$F2,$93,$FC
	.dc.b	$D7,$E4,$FE,$F3,$C0,$FC,$74,$7F,$A4,$75,$77,$FE,$3A,$31,$DF,$F8
	.dc.b	$D7,$5D,$DF,$8C,$78,$77,$7E,$2D,$E5,$DD,$F8,$B7,$A7,$77,$E2,$DE
	.dc.b	$DD,$DF,$8A,$7C,$1C,$9F,$CA,$79,$7F,$D1,$DD,$DF,$8C,$F0,$7F,$CF
	.dc.b	$5D,$FF,$3D,$FF,$3E,$4F,$D3,$7F,$CF,$DD,$3F,$5D,$FF,$40,$75,$7D
	.dc.b	$B7,$FD,$05,$D7,$F7,$DF,$F4,$27,$67,$E0,$DF,$D8,$71,$FF,$18,$D3
	.dc.b	$F8,$E1,$FA,$7E,$FF,$DC,$73,$7F,$22,$7F,$23,$1F,$93,$3F,$DE,$5E
	.dc.b	$4F,$F1,$E7,$FD,$05,$E4,$FF,$1E,$7F,$D0,$9E,$4F,$F1,$E7,$FD,$0D
	.dc.b	$E4,$FF,$1E,$7F,$D1,$1E,$4F,$F1,$E7,$FD,$15,$E4,$FF,$1E,$7F,$D1
	.dc.b	$9C,$7F,$AE,$F2,$7F,$8E,$EB,$F8,$C9,$F9,$27,$ED,$8F,$C6,$1F,$BA
	.dc.b	$3F,$18,$7E,$F8,$FC,$61,$FC,$0F,$A7,$ED,$6F,$F0,$BA,$BE,$F3,$F1
	.dc.b	$3A,$FF,$04,$FC,$6E,$CF,$C3,$BF,$23,$B7,$E2,$E1,$F8,$DF,$1F,$F8
	.dc.b	$CB,$F9,$13,$F8,$85,$72,$F2,$3F,$AA,$61,$D9,$FE,$AD,$8F,$89,$FE
	.dc.b	$B1,$97,$E2,$F2,$FA,$BD,$FF,$8B,$7E,$BF,$7F,$E2,$DF,$B3,$DF,$F8
	.dc.b	$B7,$ED,$F7,$FE,$2F,$4F,$EE,$47,$ED,$23,$F1,$6F,$F4,$B3,$F8,$B1
	.dc.b	$FD,$3C,$FE,$2C,$7F,$53,$3F,$8B,$1F,$D5,$CF,$E2,$C7,$F5,$B3,$F8
	.dc.b	$B1,$F8,$23,$F0,$87,$EA,$FC,$7F,$E2,$CF,$F8,$C4,$FE,$E3,$D3,$F8
	.dc.b	$A7,$F6,$07,$E4,$F3,$FF,$77,$75,$FE,$5D,$FF,$78,$76,$7E,$6D,$FF
	.dc.b	$79,$76,$FE,$7D,$FF,$7A,$77,$7E,$8D,$FF,$7B,$77,$FE,$9D,$FF,$7C
	.dc.b	$78,$1F,$AB,$7F,$5B,$8F,$F6,$79,$FF,$03,$6B,$F2,$67,$FC,$11,$B7
	.dc.b	$CD,$7F,$C1,$5B,$FD,$17,$FC,$19,$C1,$F5,$5F,$AF,$87,$F1,$53,$FA
	.dc.b	$77,$17,$DD,$7F,$AA,$9F,$C5,$4F,$C1,$9F,$85,$3F,$88,$73,$7E,$29
	.dc.b	$FE,$C8,$60,$D7,$17,$E4,$5E,$47,$F1,$43,$F9,$11,$E0,$3C,$47,$F1
	.dc.b	$43,$F3,$E7,$F1,$43,$F4,$67,$F1,$43,$F4,$E7,$F1,$43,$F9,$B1,$0F
	.dc.b	$F5,$4F,$93,$FD,$E2,$3F,$D0,$E4,$BF,$F2,$9C,$FE,$86,$75,$18,$8F
	.dc.b	$5C,$ED,$FE,$62,$6F,$C1,$F8,$11,$9E,$8F,$F1,$84,$E4,$87,$2A,$E9
	.dc.b	$3B,$BE,$F3,$D0,$D1,$63,$57,$DA,$B8,$92,$74,$DB,$F0,$C7,$D2,$EA
	.dc.b	$FF,$15,$E1,$DA,$73,$87,$4A,$F2,$B5,$F0,$C6,$AF,$A5,$71,$26,$1E
	.dc.b	$3A,$F0,$92,$B8,$36,$76,$F6,$AE,$75,$C6,$B6,$4A,$DD,$78,$9B,$7E
	.dc.b	$5C,$3F,$94,$FC,$4F,$F9,$10,$FF,$22,$FC,$17,$F9,$1B,$E2,$7F,$CB
	.dc.b	$7F,$C0,$3F,$E6,$7C,$58,$7F,$8D,$61,$F0,$BA,$F8,$4D,$FC,$37,$0F
	.dc.b	$87,$E3,$F1,$5C,$BE,$27,$9F,$C3,$F4,$F8,$8E,$BE,$F7,$6F,$85,$EF
	.dc.b	$F0,$1F,$DD,$B7,$FF,$79,$79,$2E,$9C,$C7,$F8,$F3,$83,$FD,$5F,$C3
	.dc.b	$E3,$C7,$A4,$E7,$3F,$03,$BD,$7C,$8F,$1D,$FD,$0F,$3E,$1F,$7F,$4F
	.dc.b	$C1,$8F,$F2,$37,$27,$C5,$7F,$C9,$1C,$BF,$CA,$FA,$3E,$77,$4F,$D1
	.dc.b	$E9,$F9,$B8,$7D,$4B,$FA,$C7,$D9,$9D,$11,$CD,$72,$6D,$72,$78,$18
	.dc.b	$9E,$0F,$DB,$7F,$D1,$F1,$FE,$6A,$E9,$FC,$AC,$6B,$7F,$C9,$13,$0F
	.dc.b	$25,$F1,$3A,$10,$C1,$23,$83,$E2,$31,$1C,$A3,$38,$7A,$67,$27,$EC
	.dc.b	$6F,$FD,$23,$C0,$F6,$FF,$16,$7E,$33,$E0,$7F,$1B,$EE,$FE,$3B,$E0
	.dc.b	$BF,$1F,$EF,$FE,$43,$E0,$FF,$23,$F0,$3F,$92,$F8,$4F,$C9,$FC,$1F
	.dc.b	$FA,$BB,$E2,$1F,$0D,$7E,$1B,$1C,$99,$F8,$7D,$7C,$46,$EF,$87,$C3
	.dc.b	$F1,$F8,$89,$CB,$FE,$86,$94,$6A,$31,$1C,$9D,$5F,$67,$86,$3E,$3C
	.dc.b	$BC,$B1,$F9,$5D,$3D,$2F,$E5,$F6,$F6,$BF,$99,$E0,$F8,$2F,$E6,$F8
	.dc.b	$BE,$2B,$F9,$DE,$3F,$C7,$74,$FE,$0F,$9F,$E7,$74,$7D,$08,$C1,$38
	.dc.b	$BF,$01,$D3,$F5,$49,$87,$D9,$AE,$13,$A8,$C4,$EA,$FA,$B6,$FA,$4F
	.dc.b	$F6,$51,$2E,$8F,$B9,$1E,$73,$D4,$7E,$D7,$A3,$EE,$3F,$B7,$E8,$FB
	.dc.b	$8F,$EE,$7A,$3E,$E3,$FB,$BE,$8F,$94,$FC,$C3,$EA,$67,$F6,$DF,$AD
	.dc.b	$D1,$F6,$DF,$B1,$D1,$F6,$9F,$E2,$BA,$9F,$3F,$3F,$DC,$7E,$F7,$47
	.dc.b	$DD,$76,$6A,$3E,$F3,$8D,$90,$E8,$FC,$04,$E6,$1D,$1E,$88,$E1,$C0
	.dc.b	$D7,$DB,$18,$DB,$E0,$BA,$DF,$E2,$BC,$38,$3F,$48,$FC,$B3,$9F,$1F
	.dc.b	$F0,$FC,$9E,$D7,$E8,$4F,$D1,$7A,$3F,$A3,$F6,$FF,$49,$E9,$7E,$97
	.dc.b	$DC,$FD,$37,$A9,$FA,$7F,$77,$F5,$1E,$B7,$EA,$7D,$EF,$D5,$7B,$1F
	.dc.b	$AB,$F7,$FF,$59,$ED,$7E,$B7,$E0,$3E,$57,$0F,$C3,$FE,$01,$FF,$42
	.dc.b	$75,$17,$BD,$AF,$0B,$FD,$F1,$1F,$ED,$0A,$F8,$DD,$BF,$56,$7F,$DF
	.dc.b	$9C,$BF,$4E,$7A,$7B,$DC,$7B,$FE,$C6,$5D,$72,$C5,$7E,$43,$2F,$0E
	.dc.b	$0C,$FC,$F3,$FB,$9D,$7D,$6F,$EE,$F7,$F7,$BF,$BD,$E1,$F8,$6F,$EF
	.dc.b	$F8,$FE,$3B,$FC,$07,$97,$E5,$BF,$C0,$F9,$FE,$73,$FA,$1D,$7F,$07
	.dc.b	$18,$8F,$C1,$73,$FD,$17,$F8,$57,$3F,$D1,$7F,$86,$73,$FD,$17,$F8
	.dc.b	$77,$3F,$D1,$7F,$88,$73,$FD,$17,$F8,$97,$3F,$D1,$7F,$35,$CD,$F3
	.dc.b	$1F,$87,$9F,$C5,$73,$CB,$F2,$39,$3E,$85,$7C,$EE,$7F,$A1,$3F,$33
	.dc.b	$97,$E8,$4F,$D6,$E7,$E7,$7F,$6B,$0E,$67,$CF,$3C,$47,$E7,$F2,$F3
	.dc.b	$3F,$A1,$CF,$D4,$FE,$8F,$4F,$73,$FA,$5D,$7E,$13,$FA,$7D,$BE,$33
	.dc.b	$F2,$0F,$9C,$41,$E4,$A3,$D9,$EF,$79,$8F,$3F,$BD,$E8,$F8,$8F,$4F
	.dc.b	$BF,$EA,$F8,$8F,$5F,$C0,$7B,$3E,$23,$DB,$F0,$5E,$EF,$88,$F7,$FC
	.dc.b	$1F,$C0,$F8,$8F,$83,$F0,$BF,$0B,$E2,$3E,$1F,$C3,$FC,$4F,$88,$F8
	.dc.b	$BF,$01,$FC,$9F,$BF,$F4,$0E,$79,$7E,$3C,$7E,$8E,$B9,$46,$3A,$98
	.dc.b	$F7,$8C,$67,$E3,$75,$AF,$9D,$E1,$BF,$AD,$E5,$C3,$EF,$7A,$71,$FC
	.dc.b	$37,$B7,$27,$F0,$5F,$07,$F2,$9D,$BF,$AE,$F0,$7E,$7B,$FB,$0E,$8F
	.dc.b	$A6,$FE,$CB,$AB,$EB,$BF,$B4,$EB,$FB,$6F,$ED,$BB,$3E,$FB,$FB,$8E
	.dc.b	$DF,$C1,$BF,$93,$CF,$DD,$E4,$7E,$2B,$BF,$F1,$F1,$FC,$C3,$C2,$BC
	.dc.b	$FF,$C1,$E7,$F1,$F5,$F9,$0A,$E7,$E1,$7E,$57,$1E,$EF,$E5,$F2,$F1
	.dc.b	$BF,$99,$CF,$CE,$FE,$6F,$4F,$5B,$F9,$DD,$7D,$EF,$E7,$F6,$F8,$6F
	.dc.b	$E8,$77,$F8,$D9,$7E,$3F,$7F,$8C,$FF,$01,$3F,$72,$7F,$05,$3F,$72
	.dc.b	$7F,$09,$3F,$72,$7F,$0D,$3F,$72,$7F,$11,$3F,$72,$7F,$82,$FC,$8F
	.dc.b	$73,$F1,$8F,$C2,$6D,$F1,$3C,$3F,$D9,$E1,$FB,$5E,$1F,$95,$5B,$3F
	.dc.b	$98,$1F,$18,$FE,$B7,$BF,$C6,$7F,$AE,$EF,$F1,$9F,$EB,$FB,$FC,$67
	.dc.b	$FB,$0E,$FF,$19,$FE,$C7,$BF,$C6,$7F,$B2,$EF,$F1,$9F,$EC,$FB,$FC
	.dc.b	$67,$FA,$F3,$FC,$91,$FE,$C1,$E2,$7F,$28,$7F,$B1,$78,$BF,$CB,$1F
	.dc.b	$EC,$9E,$37,$F3,$07,$FB,$37,$8F,$FC,$D1,$FE,$D1,$E4,$7F,$38,$7F
	.dc.b	$B5,$71,$FC,$67,$FA,$77,$2F,$C2,$7F,$A4,$9F,$DE,$08,$7F,$48,$3E
	.dc.b	$37,$2F,$E4,$DE,$67,$D2,$3B,$9F,$58,$F1,$9F,$68,$F3,$9F,$78,$F5
	.dc.b	$9F,$82,$3D,$E7,$F9,$17,$4F,$8D,$CD,$FD,$0A,$3F,$A2,$77,$3A,$3F
	.dc.b	$69,$1F,$D9,$FA,$7F,$70,$FF,$69,$EA,$FD,$E3,$FD,$AF,$AF,$F8,$03
	.dc.b	$FD,$B7,$B3,$F8,$23,$FD,$BF,$B7,$F8,$43,$FD,$C7,$BB,$F8,$6B,$FD
	.dc.b	$CF,$BF,$F8,$93,$FD,$D7,$C0,$FE,$2D,$A7,$FA,$9E,$3F,$89,$78,$3F
	.dc.b	$C7,$1F,$E2,$9E,$17,$F2,$07,$F8,$B7,$87,$FC,$91,$FE,$31,$E2,$7F
	.dc.b	$28,$7F,$8D,$78,$BF,$CB,$1F,$E3,$9E,$37,$F3,$07,$F8,$F7,$8F,$FC
	.dc.b	$D1,$FE,$26,$43,$F9,$DF,$7F,$EC,$5F,$D8,$78,$BF,$B5,$7F,$60,$7F
	.dc.b	$45,$3F,$60,$7F,$49,$3F,$60,$7F,$4D,$3F,$60,$7F,$51,$3F,$60,$7F
	.dc.b	$55,$3F,$60,$7F,$31,$3F,$87,$9F,$B0,$3F,$13,$D1,$FD,$95,$B1,$FD
	.dc.b	$A7,$C7,$DC,$7E,$C0,$E6,$75,$3F,$16,$78,$4F,$C5,$9E,$53,$F1,$67
	.dc.b	$A4,$FC,$59,$ED,$3F,$16,$7F,$61,$3F,$5A,$7F,$61,$F2,$3F,$B0,$3F
	.dc.b	$E1,$FF,$6B,$FB,$A9,$1F,$E9,$8D,$7F,$C6,$8D,$BF,$F1,$8B,$6F,$FC
	.dc.b	$08,$D9,$E8,$72,$21,$C8,$B3,$C9,$F8,$A7,$F3,$B2,$F5,$7F,$90,$67
	.dc.b	$A3,$63,$8D,$78,$1A,$78,$0E,$C4,$AE,$CD,$8E,$0D,$38,$46,$89,$1A
	.dc.b	$BF,$8F,$5E,$1A,$E0,$DA,$DC,$72,$AE,$B1,$89,$D4,$60,$93,$B2,$46
	.dc.b	$9B,$C2,$C3,$E0,$B6,$3F,$C0,$1E,$DF,$F8,$B0,$8F,$C2,$AE,$44,$9E
	.dc.b	$11,$E8,$6E,$1E,$56,$E1,$E4,$AE,$18,$78,$09,$7D,$6B,$9E,$3F,$11
	.dc.b	$7C,$47,$F8,$F8,$FF,$24,$FF,$00,$B1,$BE,$F6,$3E,$27,$3F,$01,$AF
	.dc.b	$89,$DF,$C1,$70,$F8,$9E,$3F,$09,$CB,$E2,$79,$FC,$37,$4F,$89,$EB
	.dc.b	$F1,$3D,$BD,$8E,$FF,$09,$E0,$F6,$BF,$B0,$7B,$3F,$E0,$4E,$24,$6B
	.dc.b	$9B,$FC,$57,$1F,$EC,$C7,$FD,$64,$B0,$ED,$1A,$CB,$C4,$F0,$D3,$CC
	.dc.b	$F2,$DB,$D4,$F4,$E0,$F7,$3D,$B8,$BE,$1F,$1F,$C5,$8F,$F1,$C7,$27
	.dc.b	$FA,$83,$9B,$E5,$73,$FC,$E9,$FA,$CD,$E9,$FA,$CD,$BF,$AD,$D1,$F6
	.dc.b	$2B,$F0,$F3,$97,$FD,$35,$D5,$F8,$47,$FD,$58,$E7,$AB,$EE,$C7,$DB
	.dc.b	$9C,$57,$39,$CC,$F8,$8D,$10,$FE,$5E,$7E,$BC,$FF,$5C,$BF,$A5,$6C
	.dc.b	$7E,$41,$B3,$FE,$BE,$69,$4E,$47,$ED,$0C,$9B,$9E,$8E,$48,$70,$23
	.dc.b	$A6,$CF,$85,$DC,$7F,$BB,$B8,$FF,$0B,$1F,$C5,$0E,$E4,$37,$3C,$5E
	.dc.b	$75,$E7,$69,$E9,$AF,$01,$27,$83,$67,$83,$67,$84,$FE,$D3,$A9,$CF
	.dc.b	$E0,$BD,$CF,$22,$38,$25,$72,$6B,$F1,$E3,$6E,$48,$66,$36,$FB,$CF
	.dc.b	$8F,$8F,$E1,$F3,$FF,$76,$BF,$D0,$63,$58,$FA,$5E,$47,$43,$37,$E0
	.dc.b	$72,$43,$37,$E0,$7D,$C7,$93,$1F,$82,$FC,$C3,$81,$A2,$99,$39,$9E
	.dc.b	$13,$46,$0D,$D7,$D4,$F6,$BF,$CC,$31,$FD,$34,$FF,$1F,$9F,$D1,$8E
	.dc.b	$04,$39,$90,$F0,$16,$3E,$13,$DE,$7D,$31,$88,$D5,$FC,$37,$A1,$C8
	.dc.b	$91,$B6,$C7,$DF,$C3,$E9,$69,$FE,$EF,$7E,$94,$F4,$C7,$E5,$3C,$1B
	.dc.b	$7C,$C9,$7E,$19,$F9,$58,$79,$A3,$F7,$C4,$AF,$E0,$24,$9F,$DF,$65
	.dc.b	$FC,$41,$B8,$FF,$10,$9F,$2C,$7F,$0E,$3F,$A9,$9A,$3F,$89,$97,$0F
	.dc.b	$E1,$75,$CE,$3F,$86,$11,$C1,$FD,$DB,$5F,$ED,$FC,$AF,$85,$FE,$96
	.dc.b	$7F,$4C,$F7,$3F,$A6,$FC,$37,$F4,$EF,$7B,$FA,$7F,$C3,$FF,$50,$F8
	.dc.b	$06,$7F,$EE,$8F,$E0,$17,$C4,$E2,$1A,$FF,$55,$5B,$29,$F0,$9B,$AD
	.dc.b	$4E,$7E,$23,$1C,$5D,$A3,$87,$C3,$63,$1E,$78,$BE,$23,$15,$D7,$A2
	.dc.b	$CB,$E2,$2F,$5E,$3E,$0F,$3E,$E7,$CE,$C1,$D4,$EC,$C6,$2A,$E1,$A7
	.dc.b	$BF,$1E,$B9,$76,$CF,$BE,$9E,$0D,$7C,$3B,$F8,$B8,$7C,$7C,$5F,$27
	.dc.b	$2F,$EC,$BC,$9F,$35,$FE,$CF,$CD,$F4,$5F,$ED,$3D,$1F,$51,$FF,$4A
	.dc.b	$F5,$FE,$25,$FB,$9C,$BF,$87,$7F,$05,$C7,$F8,$F6,$CF,$F5,$3D,$BF
	.dc.b	$6B,$C9,$FC,$D6,$7F,$C8,$35,$B7,$99,$F8,$C2,$C7,$E3,$2F,$B4,$62
	.dc.b	$78,$73,$7E,$2F,$CC,$FA,$DD,$7F,$8B,$AF,$49,$FE,$02,$EB,$FC,$5B
	.dc.b	$FE,$04,$EB,$FC,$5B,$FE,$06,$EB,$F8,$AF,$E2,$F8,$BE,$4B,$F8,$DE
	.dc.b	$4F,$9A,$FE,$3F,$6F,$A2,$FD,$33,$F2,$5D,$7F,$71,$FD,$39,$FD,$D0
	.dc.b	$BA,$7E,$3B,$AB,$EE,$47,$03,$91,$F9,$8D,$FF,$21,$1F,$89,$8F,$01
	.dc.b	$F9,$FE,$7F,$BC,$FE,$87,$9F,$EF,$3F,$A3,$E7,$FB,$CF,$F8,$83,$9F
	.dc.b	$EF,$3F,$E2,$2E,$7F,$BC,$FF,$89,$39,$FE,$F3,$FE,$00,$E6,$FD,$37
	.dc.b	$C1,$FE,$68,$FC,$DF,$37,$E7,$3E,$0F,$F3,$BF,$09,$F9,$EF,$84,$FC
	.dc.b	$FF,$C2,$FE,$83,$E1,$7F,$43,$F0,$DF,$A2,$F8,$6F,$D1,$FC,$3F,$E9
	.dc.b	$3E,$1F,$F4,$FE,$67,$F8,$1F,$A3,$F6,$2B,$E1,$F3,$EB,$EB,$E1,$F7
	.dc.b	$F0,$FF,$A7,$F1,$7F,$4D,$CB,$F3,$3B,$7F,$6D,$8A,$B8,$4F,$F0,$6C
	.dc.b	$BA,$E3,$E0,$CB,$C3,$3E,$33,$F8,$56,$9E,$77,$F8,$66,$DE,$B7,$F8
	.dc.b	$77,$07,$BD,$FE,$21,$C5,$F0,$DF,$E2,$5C,$9F,$1D,$FE,$29,$CB,$FE
	.dc.b	$41,$8F,$2F,$B5,$F0,$0F,$83,$F0,$FF,$0B,$DC,$F8,$7F,$0F,$F1,$3D
	.dc.b	$EF,$8B,$F0,$FF,$1B,$E0,$3E,$3F,$C3,$FC,$8F,$82,$F9,$3F,$0F,$F2
	.dc.b	$BE,$13,$E5,$FC,$3F,$CC,$F8,$6F,$9B,$F0,$FF,$3B,$D4,$F9,$FE,$C7
	.dc.b	$D0,$F7,$27,$2F,$C5,$EF,$F9,$31,$A8,$C1,$FC,$A3,$BF,$FC,$90,$30
	.dc.b	$ED,$1A,$F1,$D2,$CF,$C9,$39,$86,$BE,$98,$C6,$FE,$D7,$5C,$3F,$05
	.dc.b	$E1,$C7,$F1,$5E,$5C,$BF,$25,$E9,$CD,$F3,$5F,$E6,$7D,$1F,$45,$FD
	.dc.b	$51,$FC,$DB,$A3,$EB,$38,$D7,$FA,$17,$4F,$DB,$8C,$79,$CE,$BF,$C0
	.dc.b	$46,$3B,$3F,$08,$EB,$B7,$F1,$0F,$3E,$1F,$CD,$72,$FE,$4B,$BB,$F3
	.dc.b	$33,$F0,$7C,$3F,$CC,$BF,$97,$3F,$B6,$1F,$C1,$4E,$5E,$17,$E6,$1E
	.dc.b	$9E,$0F,$E5,$DF,$94,$78,$3C,$2F,$E1,$59,$79,$1F,$E1,$99,$FA,$1F
	.dc.b	$E1,$DA,$7B,$5F,$E2,$1B,$7C,$27,$F8,$97,$07,$C7,$7F,$8A,$70,$FC
	.dc.b	$BE,$1F,$B7,$C5,$F3,$F3,$F7,$71,$FD,$37,$E0,$72,$7D,$77,$E1,$72
	.dc.b	$FD,$B7,$E2,$73,$7D,$F7,$E3,$73,$FE,$11,$F9,$1D,$1F,$8A,$7E,$57
	.dc.b	$4F,$E5,$DF,$C4,$75,$FE,$59,$F9,$DD,$77,$7F,$E0,$F3,$B9,$E2,$7E
	.dc.b	$3F,$B7,$F4,$6F,$E4,$7B,$BF,$4E,$FE,$4F,$BF,$F5,$6F,$E5,$7C,$0F
	.dc.b	$D7,$BF,$97,$F0,$7F,$66,$FE,$67,$C2,$FD,$BB,$F9,$02,$1F,$C9,$FC
	.dc.b	$17,$37,$CB,$F0,$3F,$9C,$F4,$7C,$F7,$F9,$DF,$4F,$D3,$7F,$9E,$F5
	.dc.b	$7D,$77,$F9,$FF,$5F,$DB,$7F,$A0,$F6,$7D,$F7,$F8,$EE,$5F,$83,$3F
	.dc.b	$0A,$7E,$34,$FD,$77,$77,$F2,$5E,$2F,$F0,$67,$83,$FC,$8E,$74,$70
	.dc.b	$3F,$31,$E0,$FF,$22,$7F,$C5,$5E,$27,$F3,$67,$FC,$59,$E2,$FF,$3A
	.dc.b	$7F,$C5,$DE,$37,$F3,$E7,$FC,$61,$E3,$FF,$42,$7F,$C6,$5E,$47,$F4
	.dc.b	$67,$F9,$09,$0F,$E9,$98,$BC,$3F,$C6,$75,$FF,$40,$F0,$FF,$18,$FF
	.dc.b	$42,$F0,$FF,$18,$FF,$44,$F0,$FF,$18,$FF,$46,$F0,$FF,$18,$FF,$48
	.dc.b	$F0,$FF,$18,$FF,$4A,$F0,$FF,$18,$FF,$4C,$F0,$FF,$0C,$FE,$23,$4F
	.dc.b	$E2,$9E,$1F,$E2,$F9,$BF,$90,$C7,$E4,$0F,$E3,$1E,$3F,$F2,$A7,$F1
	.dc.b	$87,$F1,$BC,$3B,$CF,$F5,$0C,$7C,$6F,$F5,$2C,$BC,$EF,$F5,$4C,$FD
	.dc.b	$6F,$F5,$6D,$3D,$EF,$F5,$8D,$7E,$1B,$FD,$1B,$1F,$F2,$EF,$B4,$ED
	.dc.b	$FF,$01,$F7,$7E,$36,$7F,$8E,$1F,$40,$FE,$A6,$7D,$43,$FA,$B9,$F6
	.dc.b	$0F,$EB,$67,$DC,$3F,$AF,$9F,$80,$3F,$86,$9F,$CD,$FA,$3F,$8C,$BF
	.dc.b	$C7,$7C,$8F,$E3,$2F,$F1,$7E,$1F,$F6,$4B,$FC,$6F,$C7,$FE,$83,$D9
	.dc.b	$FC,$66,$7F,$B7,$CF,$33,$A9,$FC,$BB,$AF,$F1,$CF,$F3,$0E,$BF,$C7
	.dc.b	$3F,$CC,$BA,$FF,$1C,$FF,$34,$EB,$FC,$73,$FC,$DB,$AF,$F1,$CF,$F3
	.dc.b	$8D,$7F,$A0,$6F,$F1,$E3,$FA,$C7,$57,$CC,$7F,$AD,$71,$7D,$07,$FA
	.dc.b	$E7,$1F,$D4,$7F,$AF,$72,$7D,$87,$FB,$07,$2F,$DC,$7F,$B1,$73,$7E
	.dc.b	$01,$FE,$C9,$CF,$F8,$EE,$7F,$F0,$8F,$57,$E3,$9F,$F0,$9F,$67,$E3
	.dc.b	$6B,$FB,$57,$5D,$E7,$FE,$CF,$3B,$9F,$ED,$D7,$CD,$EA,$7F,$19,$3F
	.dc.b	$B5,$63,$E3,$AF,$ED,$F9,$79,$DF,$EE,$39,$FA,$DF,$EE,$7A,$7B,$DF
	.dc.b	$EE,$BA,$FC,$37,$F9,$EE,$1F,$8E,$EA,$FF,$4F,$7F,$00,$FF,$A9,$E3
	.dc.b	$FB,$33,$7A,$3F,$D4,$AB,$DE,$FF,$4F,$AF,$76,$FB,$F8,$F8,$8C,$F8
	.dc.b	$3A,$F8,$8D,$FC,$17,$0F,$88,$E3,$F0,$9C,$BE,$13,$9F,$C2,$F4,$F8
	.dc.b	$8E,$BF,$11,$DB,$E2,$3B,$FB,$1F,$27,$47,$6F,$CB,$E1,$FF,$40,$72
	.dc.b	$7A,$2F,$8B,$FE,$15,$86,$3D,$A3,$19,$F8,$9D,$6B,$E6,$78,$6F,$EA
	.dc.b	$79,$70,$7B,$9E,$9C,$5F,$0B,$8F,$E1,$F7,$49,$F9,$87,$FB,$43,$3F
	.dc.b	$95,$1F,$38,$E1,$1C,$8F,$37,$27,$C9,$7D,$1B,$FC,$97,$D5,$C9,$F2
	.dc.b	$5F,$67,$27,$C9,$7D,$DC,$9F,$25,$F8,$1C,$9F,$25,$F8,$5C,$9F,$63
	.dc.b	$E1,$24,$7C,$57,$E6,$BF,$1F,$93,$E4,$BF,$82,$E4,$F9,$0F,$E1,$39
	.dc.b	$3E,$43,$F8,$6E,$4F,$90,$FE,$23,$93,$E4,$3F,$8A,$E4,$F9,$0F,$E3
	.dc.b	$39,$3E,$0B,$F8,$EE,$2F,$90,$FF,$A4,$78,$DE,$D7,$E3,$8F,$C7,$FC
	.dc.b	$27,$E4,$3D,$CF,$C8,$FC,$27,$FB,$C3,$F8,$07,$FE,$00,$E1,$FE,$53
	.dc.b	$7F,$DF,$AD,$F8,$0D,$7C,$26,$FE,$0B,$87,$C2,$71,$F8,$7E,$5F,$11
	.dc.b	$CF,$E1,$BA,$7C,$4F,$5F,$43,$B7,$8B,$F9,$1F,$13,$F7,$5B,$7F,$0F
	.dc.b	$DB,$FC,$98,$FF,$80,$9F,$15,$F3,$7C,$98,$F7,$7C,$D9,$F8,$DF,$46
	.dc.b	$BE,$BD,$BD,$9F,$11,$ED,$F8,$8F,$7D,$7E,$6F,$8B,$E1,$BF,$9D,$E4
	.dc.b	$F8,$CF,$E4,$3C,$87,$37,$E4,$23,$F2,$5C,$FF,$43,$C8,$FD,$DF,$4F
	.dc.b	$D4,$7F,$7B,$D5,$F6,$1F,$DF,$F5,$FD,$C7,$F8,$0F,$67,$E0,$1F,$E0
	.dc.b	$7D,$BF,$84,$7F,$82,$F7,$7E,$21,$FE,$0F,$DF,$F8,$CC,$FF,$85,$F8
	.dc.b	$1F,$90,$69,$F9,$43,$F2,$07,$E6,$0F,$DC,$57,$5D,$3F,$8A,$E1,$E1
	.dc.b	$7F,$8B,$E3,$E5,$7F,$8C,$E5,$E9,$7F,$8D,$E7,$ED,$7F,$8E,$E9,$F0
	.dc.b	$5F,$D9,$63,$FC,$15,$FF,$30,$6B,$F1,$F6,$F9,$31,$B9,$FD,$D5,$73
	.dc.b	$7F,$79,$7D,$DF,$DF,$61,$E3,$7F,$80,$63,$E7,$7F,$81,$65,$EB,$7F
	.dc.b	$82,$67,$F1,$F1,$FE,$63,$AF,$C7,$DB,$FD,$C7,$D3,$F8,$FB,$F8,$C7
	.dc.b	$F1,$D3,$E3,$1F,$B5,$3F,$80,$9F,$C7,$0F,$E0,$A7,$F1,$C3,$F8,$49
	.dc.b	$FC,$70,$FE,$1A,$7F,$1C,$3F,$88,$9F,$C4,$CF,$E8,$3E,$47,$ED,$27
	.dc.b	$FA,$0D,$FC,$63,$07,$F4,$88,$FE,$87,$AF,$C5,$DB,$F2,$3B,$7C,$57
	.dc.b	$F2,$7B,$7C,$57,$F2,$BB,$7C,$57,$F2,$FB,$7C,$57,$F3,$3B,$7C,$5D
	.dc.b	$3F,$95,$6D,$F1,$74,$FE,$AF,$C1,$F8,$BD,$BF,$AA,$B9,$AD,$9F,$1A
	.dc.b	$3F,$2A,$FC,$63,$F8,$D1,$F1,$8F,$E3,$47,$C6,$3F,$8D,$1F,$18,$FE
	.dc.b	$34,$7C,$63,$F8,$D1,$FD,$37,$6F,$86,$FE,$B8,$FD,$97,$07,$F1,$A7
	.dc.b	$F6,$C7,$F5,$3F,$07,$F8,$D3,$FD,$57,$C1,$FD,$F3,$FC,$67,$9B,$F8
	.dc.b	$13,$FC,$10,$FE,$39,$E0,$FF,$18,$7F,$8F,$78,$3F,$C6,$1F,$E4,$1E
	.dc.b	$0F,$F1,$87,$F9,$17,$83,$FC,$61,$FF,$2F,$90,$FF,$55,$9F,$C9,$FC
	.dc.b	$1F,$E3,$CF,$F1,$83,$99,$FC,$60,$EE,$7F,$18,$3C,$67,$94,$FE,$43
	.dc.b	$D9,$FC,$61,$FE,$65,$D9,$FC,$61,$FF,$05,$F6,$7F,$18,$7F,$C1,$BD
	.dc.b	$9F,$C6,$1F,$E8,$1C,$7F,$2B,$9B,$F8,$CF,$17,$E3,$27,$E8,$71,$FF
	.dc.b	$16,$7E,$A3,$FC,$59,$FB,$0F,$F1,$67,$EE,$3F,$C5,$9F,$C0,$3F,$82
	.dc.b	$7F,$C3,$BE,$1F,$F6,$57,$FC,$3F,$E1,$FE,$2F,$0F,$C6,$C7,$F6,$53
	.dc.b	$F6,$35,$FC,$71,$FC,$64,$7F,$8A,$3A,$3F,$31,$CC,$E9,$FC,$E2,$EA
	.dc.b	$FD,$02,$EB,$FD,$22,$EC,$FD,$43,$FE,$2B,$C7,$F1,$87,$F1,$52,$1F
	.dc.b	$D9,$7A,$FC,$CF,$F8,$35,$BB,$FF,$40,$6C,$9A,$49,$F6,$1F,$CC,$89
	.dc.b	$A7,$EA,$9C,$B5,$CE,$9F,$11,$A7,$84,$DB,$98,$E0,$47,$04,$8E,$07
	.dc.b	$2B,$F4,$3C,$8C,$CF,$16,$CF,$19,$DC,$7A,$7D,$2F,$F4,$04,$62,$B8
	.dc.b	$36,$B8,$61,$F0,$EF,$B4,$E5,$EA,$47,$04,$8E,$11,$E7,$6D,$76,$6D
	.dc.b	$76,$C3,$E7,$56,$DB,$5B,$6C,$EE,$7C,$75,$C9,$F1,$36,$FC,$55,$FD
	.dc.b	$A7,$0F,$7F,$88,$F4,$FF,$C4,$47,$F8,$8F,$DF,$FF,$12,$7A,$BF,$E2
	.dc.b	$5F,$7F,$FC,$4D,$EB,$FF,$89,$FD,$FF,$F1,$47,$B3,$FE,$29,$F7,$FF
	.dc.b	$C5,$5E,$DB,$82,$71,$38,$B1,$C2,$FE,$01,$D5,$6A,$73,$0F,$7F,$37
	.dc.b	$C6,$97,$C7,$E2,$33,$87,$0F,$80,$B5,$AC,$5F,$01,$8C,$3C,$1F,$01
	.dc.b	$65,$EC,$FF,$65,$BF,$E3,$63,$3F,$05,$F9,$E7,$85,$F1,$C3,$96,$3C
	.dc.b	$F3,$E9,$A7,$5C,$BE,$19,$FE,$4B,$DB,$E3,$BF,$E4,$DE,$0F,$96,$FF
	.dc.b	$93,$F8,$BE,$7B,$FE,$51,$E4,$FA,$6F,$F9,$4F,$97,$FC,$1B,$D1,$F8
	.dc.b	$08,$FB,$AD,$DF,$EE,$C7,$E0,$A3,$1D,$3F,$81,$23,$82,$46,$C8,$7E
	.dc.b	$3C,$BC,$3F,$8F,$6F,$1F,$E3,$DF,$D4,$78,$DF,$BB,$7E,$DE,$1F,$AC
	.dc.b	$6E,$1F,$A0,$6E,$1F,$80,$69,$4F,$DD,$98,$35,$3C,$DE,$C4,$36,$57
	.dc.b	$4E,$08,$E0,$93,$E3,$8E,$0D,$76,$D8,$DC,$F9,$67,$2D,$8C,$C7,$A5
	.dc.b	$AF,$A2,$79,$12,$78,$36,$38,$35,$FF,$71,$77,$79,$27,$64,$8D,$B5
	.dc.b	$EC,$D7,$E4,$10,$E4,$70,$21,$C0,$A6,$CF,$69,$83,$89,$B2,$1B,$AF
	.dc.b	$7B,$6B,$D4,$D9,$F2,$3C,$8C,$1F,$C1,$7C,$FF,$EB,$CF,$80,$EA,$6C
	.dc.b	$86,$4E,$26,$8C,$1A,$AF,$1B,$6B,$C6,$DA,$F1,$BC,$09,$3C,$1B,$3C
	.dc.b	$1F,$E9,$B5,$E6,$9D,$12,$74,$DA,$EA,$DA,$F9,$47,$48,$CC,$F0,$23
	.dc.b	$82,$47,$09,$E4,$FB,$DA,$FC,$2A,$F2,$57,$EF,$23,$F6,$A7,$F0,$42
	.dc.b	$CF,$F0,$47,$F7,$84,$CB,$F7,$8D,$CB,$F7,$91,$82,$38,$24,$75,$38
	.dc.b	$16,$36,$FF,$74,$EC,$FE,$D7,$CE,$F8,$3F,$E6,$07,$F3,$1F,$6F,$F9
	.dc.b	$97,$C2,$7F,$33,$F7,$7F,$9A,$7C,$2F,$F3,$5F,$7F,$F9,$B7,$C3,$7F
	.dc.b	$37,$F8,$1F,$E7,$1F,$0F,$FC,$E7,$E1,$7F,$9D,$7C,$47,$F3,$BF,$85
	.dc.b	$FE,$79,$F0,$FF,$CF,$7E,$1F,$F9,$F7,$BD,$FD,$07,$C6,$FF,$5A,$0F
	.dc.b	$4B,$04,$E5,$FE,$A4,$7F,$4E,$8F,$F5,$D7,$22,$BA,$4E,$BA,$3F,$AD
	.dc.b	$63,$E1,$8C,$53,$3F,$34,$63,$5F,$53,$AD,$FD,$CF,$0E,$1F,$84,$F2
	.dc.b	$E3,$F8,$CF,$4E,$5F,$92,$FF,$A1,$3A,$FF,$9C,$BF,$4B,$4F,$BD,$D1
	.dc.b	$F4,$9B,$D3,$F6,$1B,$3F,$5E,$31,$D5,$F8,$98,$E4,$DA,$E4,$D9,$DD
	.dc.b	$F8,$27,$FC,$79,$D5,$F7,$B1,$F2,$E1,$FD,$A6,$FF,$BD,$F5,$7D,$E8
	.dc.b	$F8,$BD,$5F,$79,$F8,$FD,$5F,$79,$F9,$3D,$5F,$79,$F9,$7D,$5F,$79
	.dc.b	$F9,$BD,$5F,$79,$F9,$FD,$5F,$79,$FA,$9D,$3F,$7B,$AB,$EE,$47,$8C
	.dc.b	$FB,$8F,$E0,$F4,$E1,$19,$E2,$FF,$0A,$5F,$58,$C7,$43,$1F,$14,$63
	.dc.b	$2F,$33,$AC,$FD,$4F,$0D,$3D,$CF,$2D,$7E,$13,$D3,$6F,$8F,$C3,$FE
	.dc.b	$0E,$CB,$F4,$D3,$FE,$29,$E1,$FF,$29,$F0,$7F,$9E,$E3,$CC,$DA,$FF
	.dc.b	$0A,$AF,$87,$FC,$F7,$9D,$F9,$FF,$87,$FD,$07,$A1,$FA,$1F,$87,$FD
	.dc.b	$17,$A5,$FA,$3F,$87,$FD,$27,$87,$FA,$5F,$87,$FD,$37,$AB,$FA,$7F
	.dc.b	$5F,$F5,$1F,$07,$FA,$9F,$63,$F5,$5D,$FF,$AB,$F6,$7E,$6F,$17,$8B
	.dc.b	$BE,$F7,$FF,$8F,$B9,$3F,$DB,$1C,$BE,$8C,$39,$3F,$5B,$2E,$CF,$D8
	.dc.b	$D3,$C4,$FD,$AD,$BC,$CF,$DC,$E0,$F5,$3F,$7B,$8B,$DC,$FE,$03,$93
	.dc.b	$F7,$DC,$BF,$BF,$F2,$64,$FC,$BF,$2B,$C9,$F0,$7F,$C0,$CF,$E0,$9E
	.dc.b	$DF,$F0,$5F,$83,$FF,$23,$7F,$00,$BD,$EB,$F0,$98,$F8,$0C,$FC,$2E
	.dc.b	$BE,$17,$7F,$0D,$C3,$E1,$B8,$FC,$3F,$2F,$87,$E7,$F1,$1D,$3E,$23
	.dc.b	$AF,$0F,$6F,$6B,$E3,$F1,$FE,$50,$FC,$CF,$7E,$5F,$E1,$DC,$2A,$E5
	.dc.b	$18,$C3,$B3,$AC,$BC,$4F,$0D,$3C,$CF,$2D,$BD,$4F,$4E,$0F,$7F,$0F
	.dc.b	$C1,$8F,$D4,$F1,$FC,$57,$F5,$7C,$9F,$E7,$6F,$23,$F2,$7C,$DF,$32
	.dc.b	$7F,$27,$CF,$F9,$07,$F6,$3E,$47,$E4,$1F,$D9,$F9,$1F,$90,$7F,$6B
	.dc.b	$E4,$7E,$41,$FD,$BF,$91,$F9,$07,$F7,$3E,$47,$E4,$1F,$DD,$F9,$1F
	.dc.b	$90,$7F,$7B,$E4,$7F,$11,$F1,$FF,$1C,$E6,$3F,$4B,$3C,$5F,$D3,$D7
	.dc.b	$57,$F5,$37,$E1,$7F,$57,$87,$95,$FD,$6E,$3E,$97,$F5,$F9,$7B,$5F
	.dc.b	$79,$FC,$3B,$4F,$85,$AF,$ED,$B6,$F9,$13,$FB,$6D,$FE,$63,$FC,$37
	.dc.b	$9F,$F1,$8F,$D1,$3E,$EF,$77,$E3,$1F,$BF,$DD,$F8,$C7,$F0,$3D,$DF
	.dc.b	$8C,$7F,$07,$DD,$F8,$C7,$F0,$BD,$DF,$8C,$7F,$0F,$DD,$F8,$CF,$47
	.dc.b	$FD,$3B,$DF,$73,$FF,$5F,$57,$E3,$3B,$FF,$16,$F9,$7B,$FF,$16,$F9
	.dc.b	$FB,$FF,$16,$FA,$7B,$FF,$16,$FA,$FB,$FF,$16,$FB,$7B,$FF,$16,$FF
	.dc.b	$32,$F0,$3F,$5A,$FE,$28,$83,$97,$F1,$6F,$CB,$3F,$9C,$F7,$FC,$FF
	.dc.b	$27,$F8,$EF,$4F,$D3,$7F,$8F,$F5,$7D,$77,$F9,$0F,$5F,$DB,$7F,$91
	.dc.b	$F6,$7D,$F7,$F9,$2F,$6F,$E0,$DF,$E4,$FD,$DF,$87,$7F,$94,$F7,$FC
	.dc.b	$7F,$27,$FC,$09,$A7,$EC,$8F,$C7,$9F,$D3,$3C,$9F,$E3,$2F,$F3,$2F
	.dc.b	$27,$F8,$CB,$FC,$D3,$C9,$FE,$32,$FF,$36,$F2,$7F,$8C,$BF,$CE,$3C
	.dc.b	$9F,$E3,$2F,$F3,$AF,$27,$F8,$CB,$FC,$F3,$C9,$FE,$32,$FC,$33,$F9
	.dc.b	$FF,$8F,$FB,$1E,$CF,$E3,$31,$F2,$B6,$FE,$83,$E3,$7E,$39,$FF,$6B
	.dc.b	$F7,$7E,$3A,$7E,$A1,$F8,$E3,$F8,$11,$F8,$E3,$F8,$31,$F8,$E3,$F8
	.dc.b	$51,$F8,$E3,$F8,$71,$F8,$E3,$F1,$07,$E3,$8F,$8C,$7F,$54,$EF,$FC
	.dc.b	$77,$83,$F9,$38,$FD,$07,$85,$F9,$77,$F4,$5E,$1F,$E6,$DF,$D2,$78
	.dc.b	$9F,$9F,$7F,$4D,$E2,$FE,$8D,$FD,$47,$8D,$FC,$C9,$FE,$47,$E4,$FB
	.dc.b	$FC,$9F,$E9,$DB,$7E,$C3,$6F,$EC,$7E,$4F,$C7,$BF,$F1,$D7,$0F,$F1
	.dc.b	$79,$FE,$CF,$E0,$FE,$37,$0F,$EA,$DE,$0F,$E3,$5F,$EB,$1E,$0F,$E3
	.dc.b	$5F,$EB,$5E,$0F,$E3,$5F,$EB,$9E,$0F,$E3,$5F,$EB,$DE,$0F,$E3,$5F
	.dc.b	$EC,$1E,$0F,$E3,$5F,$E2,$BA,$7F,$9C,$BD,$37,$81,$FD,$A6,$3F,$18
	.dc.b	$7F,$1E,$3F,$B2,$1F,$C8,$8F,$EC,$87,$F2,$63,$FB,$21,$FC,$A8,$FE
	.dc.b	$C8,$7F,$2E,$3F,$B2,$1F,$CC,$8F,$EC,$87,$EB,$0F,$C6,$6B,$FE,$08
	.dc.b	$DB,$E3,$79,$4F,$09,$C3,$F2,$F2,$F9,$B8,$7E,$EC,$FA,$27,$F1,$93
	.dc.b	$EA,$9F,$8C,$3E,$C9,$F8,$C3,$EE,$9F,$8C,$3F,$02,$7E,$30,$FC,$29
	.dc.b	$F8,$C3,$F9,$37,$4F,$EC,$FA,$BF,$17,$19,$CB,$F3,$BD,$5F,$8B,$7F
	.dc.b	$3F,$D5,$F8,$B7,$F4,$3D,$5F,$8B,$7F,$47,$D5,$F8,$B7,$F4,$BD,$5F
	.dc.b	$8B,$C3,$F9,$A7,$57,$E2,$DF,$F4,$07,$57,$C1,$7F,$D0,$5B,$7C,$6E
	.dc.b	$B7,$B7,$FE,$06,$3F,$C0,$FF,$0F,$FE,$08,$F7,$7F,$C1,$3F,$0F,$FE
	.dc.b	$0A,$F7,$FF,$C1,$7F,$0F,$FE,$0C,$F8,$1F,$F0,$6F,$C3,$B4,$FF,$45
	.dc.b	$7F,$00,$D6,$FC,$25,$E2,$74,$EF,$AD,$6F,$85,$D4,$E6,$1F,$0D,$8C
	.dc.b	$57,$C3,$E1,$F0,$F2,$F8,$8B,$7F,$81,$AF,$1F,$C0,$E3,$DB,$91,$E6
	.dc.b	$77,$1C,$3C,$1E,$37,$CB,$1E,$79,$74,$D3,$AE,$DD,$B2,$F8,$27,$F8
	.dc.b	$73,$83,$E2,$BF,$E1,$EE,$2F,$92,$FF,$88,$39,$3E,$6B,$FE,$22,$E6
	.dc.b	$FA,$0F,$F8,$EC,$F6,$F3,$FD,$47,$47,$F8,$A3,$9F,$EA,$3F,$E2,$AE
	.dc.b	$7F,$A8,$FF,$8B,$39,$FE,$A3,$FE,$2E,$E7,$F3,$3F,$48,$FA,$BA,$FB
	.dc.b	$1F,$AF,$BF,$C0,$7E,$CF,$0F,$C4,$7E,$31,$0F,$F4,$67,$23,$93,$51
	.dc.b	$89,$E4,$EA,$BB,$3C,$2F,$E4,$BF,$86,$3E,$49,$FE,$0B,$E0,$F9,$2F
	.dc.b	$F8,$37,$83,$E4,$BF,$E0,$FE,$0F,$92,$FF,$84,$78,$3E,$4B,$FD,$C7
	.dc.b	$CC,$FC,$47,$C1,$7E,$08,$FC,$1F,$BB,$F8,$4F,$83,$FC,$2F,$BF,$F8
	.dc.b	$6F,$84,$FC,$3F,$C0,$FE,$23,$E1,$7F,$13,$F0,$7F,$8A,$F8,$6F,$C5
	.dc.b	$FC,$2F,$E3,$3E,$1F,$F1,$BF,$0F,$F8,$EF,$88,$FC,$8F,$F0,$0F,$E7
	.dc.b	$38,$7F,$DD,$4F,$F7,$07,$FD,$64,$FF,$A7,$7C,$5F,$F4,$57,$9F,$FC
	.dc.b	$0E,$77,$E8,$F2,$8E,$73,$D6,$7F,$57,$87,$85,$FD,$6E,$5E,$57,$F5
	.dc.b	$FA,$7A,$5F,$D8,$ED,$ED,$7F,$67,$C1,$F0,$5F,$DA,$F1,$7C,$57,$F6
	.dc.b	$FC,$7F,$C0,$F9,$7E,$5F,$91,$FD,$C3,$C2,$FC,$7D,$7F,$8F,$FC,$2F
	.dc.b	$C7,$BF,$E4,$1F,$0B,$F1,$EF,$F9,$0F,$C2,$FC,$7B,$F8,$6F,$13,$F5
	.dc.b	$0F,$F0,$4F,$17,$F5,$8F,$F0,$6F,$1B,$F6,$0F,$ED,$8F,$C7,$65,$F2
	.dc.b	$FC,$8F,$E0,$BE,$2F,$EE,$27,$F2,$3E,$47,$07,$F2,$77,$D1,$FC,$AE
	.dc.b	$1E,$07,$C4,$7E,$3C,$FD,$21,$FB,$73,$F5,$07,$ED,$CF,$D6,$1F,$B7
	.dc.b	$3F,$60,$7E,$C8,$FC,$81,$FB,$7B,$79,$1F,$B7,$8C,$9F,$E6,$5D,$BE
	.dc.b	$4E,$5F,$E1,$AD,$BE,$4B,$FE,$1C,$DB,$E4,$BF,$E1,$ED,$BE,$4B,$F0
	.dc.b	$7C,$86,$FE,$B8,$C6,$9E,$F7,$5A,$FC,$37,$86,$DF,$1D,$E5,$BD,$E8
	.dc.b	$FE,$59,$3B,$7F,$97,$57,$37,$F9,$85,$F7,$7F,$99,$61,$E4,$7F,$9C
	.dc.b	$6E,$CF,$D3,$18,$D3,$DA,$EB,$5F,$82,$F0,$DB,$F1,$AF,$2E,$1F,$C8
	.dc.b	$70,$7C,$DD,$FF,$20,$FD,$13,$F2,$07,$D5,$3F,$20,$7D,$93,$F2,$07
	.dc.b	$DD,$3E,$F9,$FC,$3C,$FC,$81,$FC,$78,$FC,$81,$FC,$78,$FC,$81,$FC
	.dc.b	$73,$7F,$8C,$FF,$4E,$E0,$FE,$3C,$EA,$7E,$A7,$91,$FC,$79,$FA,$DE
	.dc.b	$47,$F1,$E7,$EC,$79,$1F,$C7,$AB,$F9,$39,$FC,$7A,$3F,$CB,$7E,$47
	.dc.b	$F1,$E7,$FC,$BB,$E4,$7F,$1E,$7F,$CB,$FE,$47,$F1,$E7,$FC,$C3,$A7
	.dc.b	$F1,$EF,$82,$FF,$74,$61,$E5,$E9,$FD,$D3,$FC,$D3,$AB,$F7,$CF,$F3
	.dc.b	$6E,$BF,$E0,$4F,$F3,$8E,$CF,$E0,$CF,$F3,$AE,$DF,$E1,$4F,$F3,$CE
	.dc.b	$EF,$E1,$CF,$F3,$EE,$FF,$E2,$4F,$F1,$AE,$5F,$E2,$CF,$F1,$83,$FC
	.dc.b	$CF,$AF,$ED,$F8,$FF,$90,$37,$C0,$FE,$4B,$E0,$F0,$AE,$47,$F8,$1B
	.dc.b	$C3,$FE,$56,$FF,$82,$3C,$4F,$E5,$EF,$F8,$2B,$C5,$FE,$66,$FF,$83
	.dc.b	$3C,$6F,$E6,$EF,$F8,$3B,$C7,$FE,$76,$FF,$84,$3C,$8F,$E7,$EF,$F1
	.dc.b	$C2,$1F,$D1,$4F,$ED,$BC,$DF,$92,$7F,$B7,$F7,$7E,$59,$FE,$E3,$DF
	.dc.b	$F9,$A7,$FB,$9F,$81,$F9,$E7,$FB,$AF,$83,$FA,$37,$F0,$DE,$1F,$E9
	.dc.b	$DF,$C4,$78,$9F,$AB,$7F,$15,$E2,$FE,$BD,$FC,$67,$8D,$F8,$EF,$1F
	.dc.b	$F8,$EC,$7F,$20,$E7,$FD,$AB,$F9,$53,$F8,$E9,$F9,$93,$F8,$E9,$FD
	.dc.b	$E8,$FE,$F8,$7E,$38,$FE,$3A,$7F,$6B,$3F,$8E,$9F,$DA,$CF,$E3,$A7
	.dc.b	$F6,$B3,$F8,$E9,$FC,$63,$9B,$F1,$EF,$F1,$DF,$21,$ED,$FF,$5E,$24
	.dc.b	$7F,$5E,$6B,$FD,$79,$B3,$FC,$E6,$78,$CE,$C8,$6C,$B1,$B9,$E3,$5F
	.dc.b	$E4,$5A,$E3,$5B,$25,$7F,$1C,$7E,$41,$FB,$7F,$23,$8B,$FE,$46,$9E
	.dc.b	$33,$D6,$3B,$D7,$5B,$DE,$0B,$DC,$E6,$B6,$73,$AD,$CE,$A3,$06,$AB
	.dc.b	$94,$E8,$FF,$49,$CE,$E3,$33,$B8,$C1,$27,$84,$65,$AF,$C0,$AF,$23
	.dc.b	$4F,$09,$6B,$D5,$EE,$7F,$85,$09,$1C,$C8,$6E,$FC,$0E,$30,$F8,$CE
	.dc.b	$70,$F0,$C6,$09,$87,$8B,$C3,$FE,$27,$8F,$C9,$F7,$3F,$C4,$67,$F8
	.dc.b	$93,$DD,$FF,$12,$FB,$BF,$E2,$6F,$7B,$FC,$4F,$EF,$7F,$8A,$3D,$FF
	.dc.b	$F1,$4F,$BF,$FE,$2A,$F8,$0F,$F1,$5F,$C0,$7F,$8B,$3E,$07,$FC,$5B
	.dc.b	$F0,$3F,$E2,$EF,$80,$FF,$17,$FC,$17,$F8,$C3,$D9,$FF,$18,$FB,$5F
	.dc.b	$E3,$2F,$83,$FD,$67,$B3,$FC,$E3,$A3,$F9,$E1,$FE,$34,$8D,$BF,$E3
	.dc.b	$6A,$E6,$FF,$8E,$30,$EE,$FF,$8E,$B2,$F1,$BF,$E3,$CD,$3C,$EF,$F8
	.dc.b	$FB,$6F,$5B,$FE,$40,$E0,$F7,$BF,$E4,$2E,$2F,$86,$FF,$91,$38,$FF
	.dc.b	$07,$D1,$F3,$39,$FE,$DF,$47,$CC,$AF,$B6,$DE,$7F,$C3,$C7,$D3,$3E
	.dc.b	$B4,$F0,$23,$C1,$B1,$C2,$7F,$65,$D5,$F8,$2A,$FF,$29,$BE,$17,$81
	.dc.b	$27,$85,$72,$EA,$FC,$2C,$F8,$49,$7E,$58,$CB,$70,$F6,$CF,$FB,$04
	.dc.b	$FC,$B9,$FE,$C3,$BF,$5B,$6B,$F7,$AD,$9E,$0F,$02,$1F,$9E,$34,$6A
	.dc.b	$79,$BA,$6B,$B3,$13,$D9,$B3,$D9,$FE,$0D,$87,$C5,$BF,$8F,$1F,$C5
	.dc.b	$8E,$E7,$AE,$FA,$92,$FA,$47,$72,$46,$6F,$B3,$63,$D1,$3C,$23,$11
	.dc.b	$D6,$30,$49,$FF,$75,$79,$6D,$7D,$35,$CA,$7A,$D6,$E3,$04,$3A,$46
	.dc.b	$88,$E4,$F6,$90,$EA,$70,$34,$7F,$33,$C7,$E3,$65,$63,$E3,$1C,$0E
	.dc.b	$47,$C2,$AF,$23,$CC,$86,$B1,$F8,$2D,$8D,$B6,$FE,$0E,$1D,$62,$1D
	.dc.b	$63,$D6,$F4,$33,$5D,$1B,$5D,$1F,$C3,$71,$FF,$05,$CF,$FA,$54,$7F
	.dc.b	$06,$3F,$A7,$1E,$13,$E0,$9D,$CB,$5D,$5C,$1E,$12,$9C,$4F,$81,$9F
	.dc.b	$D9,$72,$79,$30,$F4,$BB,$21,$93,$BE,$1E,$97,$E6,$67,$F4,$9F,$CC
	.dc.b	$9F,$48,$EA,$4B,$F3,$12,$79,$36,$BC,$33,$F8,$0B,$F3,$4F,$F0,$28
	.dc.b	$FE,$09,$5F,$BE,$CF,$F8,$25,$FF,$09,$AD,$12,$75,$5C,$27,$51,$FD
	.dc.b	$7B,$0F,$45,$7F,$56,$AF,$E3,$66,$CA,$F3,$AE,$24,$8C,$D7,$F8,$13
	.dc.b	$AF,$FA,$8F,$BE,$F7,$55,$34,$FF,$65,$4E,$63,$19,$EB,$24,$BE,$13
	.dc.b	$8D,$2F,$8F,$BF,$87,$9F,$C2,$DC,$79,$60,$F8,$1C,$60,$CB,$AF,$C3
	.dc.b	$61,$EF,$54,$7C,$3D,$F8,$4C,$7C,$46,$7E,$1B,$5F,$0D,$BF,$87,$E1
	.dc.b	$E9,$71,$F4,$B9,$7C,$17,$CC,$E7,$74,$FC,$ED,$53,$BE,$0F,$82,$79
	.dc.b	$63,$FD,$B1,$C9,$F1,$B2,$EF,$3A,$E0,$F8,$BA,$F9,$23,$16,$DF,$D5
	.dc.b	$C1,$EB,$8F,$67,$0F,$B7,$8B,$DD,$C9,$EF,$E5,$F8,$1C,$DF,$07,$9E
	.dc.b	$67,$FE,$4F,$E9,$FB,$CF,$F4,$EE,$8F,$C7,$BF,$80,$8C,$F0,$FF,$4F
	.dc.b	$8F,$F1,$87,$5F,$EE,$EB,$EE,$36,$3F,$C8,$87,$12,$1D,$4F,$C8,$17
	.dc.b	$87,$F2,$0E,$BA,$BF,$1B,$8F,$9A,$3F,$09,$7F,$80,$CB,$D9,$1F,$13
	.dc.b	$4F,$80,$FC,$6D,$BE,$23,$F2,$38,$3E,$F3,$F2,$BA,$BE,$F3,$F3,$3A
	.dc.b	$BE,$F3,$F3,$BA,$BE,$C3,$E0,$3E,$C7,$3F,$DF,$E8,$FB,$CE,$B8,$7F
	.dc.b	$3F,$5C,$AB,$87,$07,$E7,$B0,$EF,$18,$E5,$65,$E4,$8C,$67,$E8,$75
	.dc.b	$A7,$B1,$E1,$AF,$C0,$79,$6D,$F1,$74,$FD,$7F,$07,$EA,$B8,$7F,$6B
	.dc.b	$E9,$FE,$7C,$FD,$07,$C1,$FE,$87,$D4,$FD,$17,$C2,$7E,$8F,$D5,$FD
	.dc.b	$37,$F0,$0F,$F8,$27,$27,$3F,$ED,$E1,$EB,$6B,$E1,$B7,$EB,$F0,$F8
	.dc.b	$7E,$3E,$C7,$2F,$80,$E7,$EC,$F4,$F6,$7A,$FC,$3F,$6F,$87,$F9,$5C
	.dc.b	$FF,$9F,$F2,$7F,$52,$FE,$87,$64,$F0,$8C,$5F,$47,$58,$F8,$1E,$19
	.dc.b	$F9,$1E,$5A,$FB,$36,$F6,$C7,$BC,$FD,$5F,$17,$C3,$7F,$5B,$C9,$F1
	.dc.b	$DF,$D7,$F2,$FC,$BE,$EF,$85,$EE,$7C,$63,$E3,$FC,$3F,$C8,$F7,$BE
	.dc.b	$4F,$C3,$FC,$AF,$80,$F9,$7F,$0F,$F3,$3E,$0B,$E6,$FC,$3F,$CE,$F8
	.dc.b	$4F,$9F,$F0,$FF,$43,$E1,$BF,$8B,$7F,$00,$FD,$0E,$0F,$E0,$4F,$FB
	.dc.b	$F3,$BF,$F9,$42,$F5,$F7,$E8,$F0,$F4,$38,$FB,$33,$0E,$2E,$5E,$73
	.dc.b	$DA,$BB,$FC,$47,$83,$1F,$15,$7D,$1C,$FC,$CF,$D3,$D7,$D4,$FD,$5D
	.dc.b	$FD,$CF,$D7,$E1,$F8,$4F,$D9,$E3,$F8,$CF,$DB,$E4,$F9,$4F,$F2,$1E
	.dc.b	$6F,$9C,$FE,$8F,$9F,$E9,$3F,$A5,$E8,$FA,$CF,$EA,$BB,$FF,$29,$AF
	.dc.b	$E4,$FA,$BF,$25,$3F,$87,$F1,$3F,$24,$FE,$27,$C4,$FC,$93,$F8,$BF
	.dc.b	$13,$F2,$4F,$E3,$7C,$4F,$C9,$3F,$91,$E7,$FC,$9B,$DF,$C3,$FC,$83
	.dc.b	$E1,$F0,$FF,$20,$F8,$FC,$3F,$C8,$3E,$5F,$0F,$F2,$0F,$9F,$C3,$FC
	.dc.b	$83,$F9,$B3,$D7,$E2,$FF,$64,$E1,$F7,$BF,$D9,$B8,$BE,$1B,$FD,$A3
	.dc.b	$8F,$E3,$FA,$0E,$47,$2F,$CC,$C7,$E2,$73,$7D,$07,$E3,$73,$FD,$47
	.dc.b	$E4,$74,$7D,$87,$E5,$74,$FD,$C7,$E6,$75,$7E,$01,$FE,$2F,$DF,$F9
	.dc.b	$16,$F6,$7E,$1D,$FC,$49,$F9,$4E,$6F,$C7,$3C,$3B,$EF,$07,$E6,$77
	.dc.b	$61,$FB,$CA,$EB,$DF,$CD,$EA,$7E,$7B,$BB,$C5,$7C,$78,$5A,$79,$E3
	.dc.b	$1A,$FA,$DD,$6D,$EF,$78,$6F,$F0,$DE,$5C,$1F,$1D,$E9,$C3,$F3,$3D
	.dc.b	$07,$1F,$D0,$D3,$FC,$17,$C9,$F5,$1F,$F0,$6F,$2F,$D8,$7F,$91,$6F
	.dc.b	$F7,$0F,$BC,$7E,$44,$FE,$22,$7E,$44,$FE,$2A,$7E,$44,$FE,$32,$7E
	.dc.b	$44,$FE,$3D,$D5,$F9,$0C,$BF,$D4,$31,$F9,$13,$47,$E4,$4E,$47,$E4
	.dc.b	$4E,$C7,$E4,$4F,$11,$E4,$3F,$51,$DF,$FC,$81,$FF,$04,$F9,$1F,$D0
	.dc.b	$1F,$F0,$5F,$93,$FD,$11,$FE,$3E,$41,$E5,$3C,$9F,$E4,$31,$FB,$B3
	.dc.b	$F7,$C7,$F3,$7F,$27,$F9,$0B,$FC,$E7,$C9,$FE,$42,$FF,$3B,$F2,$7F
	.dc.b	$90,$BF,$CF,$7C,$9F,$E4,$2F,$F3,$FF,$27,$F8,$93,$FC,$83,$C0,$FC
	.dc.b	$36,$DF,$8B,$ED,$76,$FF,$1B,$AF,$C8,$1F,$C7,$88,$7E,$6B,$C0,$FE
	.dc.b	$4A,$FF,$2A,$F2,$7F,$8F,$BF,$8F,$3C,$07,$F1,$F3,$C8,$7F,$1F,$3D
	.dc.b	$07,$F1,$F3,$D8,$7F,$1F,$3E,$01,$F0,$8F,$E2,$DD,$3F,$B5,$E0,$FE
	.dc.b	$E3,$1F,$2F,$1F,$DF,$3F,$40,$FE,$A2,$7D,$43,$F8,$31,$FD,$87,$8F
	.dc.b	$F8,$EB,$FC,$87,$C9,$FE,$3A,$FF,$23,$F2,$7F,$8E,$BF,$C9,$7C,$9F
	.dc.b	$E3,$AF,$F2,$7F,$27,$F8,$EB,$FE,$12,$F2,$BF,$65,$8F,$E3,$89,$D5
	.dc.b	$FD,$F7,$AB,$F8,$FC,$7E,$C7,$CA,$FE,$3E,$FE,$CF,$CA,$FE,$3E,$FE
	.dc.b	$D7,$CA,$FE,$3E,$FE,$DF,$CA,$FD,$23,$FB,$9C,$BF,$50,$7F,$97,$FB
	.dc.b	$FF,$58,$FF,$98,$7C,$0F,$D8,$3F,$D0,$CF,$F4,$4F,$17,$E3,$DF,$F3
	.dc.b	$D7,$91,$FC,$81,$FC,$81,$FC,$80,$FA,$27,$F2,$03,$EA,$9F,$C8,$0F
	.dc.b	$B2,$7F,$20,$3E,$E9,$FC,$80,$FC,$09,$FC,$83,$5F,$C2,$BF,$C8,$B9
	.dc.b	$7F,$12,$FE,$2C,$FC,$6E,$3F,$93,$F2,$BF,$8F,$D7,$F7,$4E,$7D,$CF
	.dc.b	$E5,$EF,$9B,$F9,$99,$FE,$3A,$7F,$8A,$FC,$9F,$E3,$8F,$F8,$B7,$C9
	.dc.b	$FE,$38,$FF,$8B,$FC,$9F,$E3,$8F,$F8,$C7,$C9,$FE,$38,$FF,$8C,$FC
	.dc.b	$9F,$E3,$8F,$ED,$34,$FD,$BF,$8B,$FE,$06,$F6,$FF,$D5,$71,$FE,$AC
	.dc.b	$F2,$FF,$99,$C6,$23,$45,$8C,$B7,$C1,$4E,$E3,$FD,$59,$18,$75,$19
	.dc.b	$9E,$31,$E0,$6B,$E1,$9E,$0E,$5A,$E6,$79,$46,$23,$CA,$6E,$32,$6B
	.dc.b	$E2,$3E,$8C,$79,$38,$6D,$FF,$50,$9F,$F1,$5D,$70,$AF,$39,$0E,$85
	.dc.b	$3C,$45,$3C,$26,$0D,$90,$C1,$0D,$9E,$8E,$0F,$F7,$DE,$1E,$B8,$F9
	.dc.b	$33,$E3,$79,$1C,$0C,$9E,$12,$9A,$29,$DC,$A7,$1B,$F9,$ED,$9F,$04
	.dc.b	$7C,$38,$FC,$CE,$5F,$E4,$AC,$7E,$3D,$31,$F8,$F5,$E1,$BE,$6F,$62
	.dc.b	$38,$24,$F6,$26,$1E,$D9,$F3,$4E,$A3,$11,$F1,$7D,$CF,$34,$62,$3A
	.dc.b	$D7,$27,$C4,$D7,$E1,$57,$37,$33,$C6,$3C,$7E,$EF,$F6,$BC,$94,$97
	.dc.b	$6F,$F2,$29,$FF,$35,$D7,$C0,$AE,$B8,$77,$7D,$2D,$7C,$E7,$80,$87
	.dc.b	$0C,$3E,$E6,$9F,$80,$7C,$93,$C6,$7D,$B1,$82,$3C,$8E,$EE,$1F,$D3
	.dc.b	$67,$F5,$B4,$FA,$EF,$9E,$36,$71,$3C,$A7,$8C,$C1,$E0,$2E,$7F,$B6
	.dc.b	$C3,$F2,$B5,$D6,$7F,$0A,$7E,$AC,$B3,$FE,$3A,$8D,$47,$F7,$02,$71
	.dc.b	$FF,$23,$77,$8A,$77,$39,$9D,$E7,$F9,$17,$73,$A8,$C4,$68,$F1,$16
	.dc.b	$FC,$4F,$02,$19,$BF,$03,$8B,$F0,$73,$7F,$4B,$AE,$33,$BC,$3C,$0F
	.dc.b	$6A,$F3,$13,$0F,$1B,$63,$51,$82,$38,$25,$76,$9E,$0D,$9D,$B6,$77
	.dc.b	$87,$A1,$F2,$1E,$99,$DC,$66,$B9,$B6,$33,$7E,$AB,$E8,$49,$E9,$87
	.dc.b	$C1,$5F,$11,$FE,$CE,$3F,$D9,$FB,$FF,$B4,$3E,$23,$FD,$A3,$F0,$1F
	.dc.b	$ED,$2F,$88,$FF,$69,$FC,$17,$FB,$53,$E2,$3F,$DA,$BF,$09,$FE,$D6
	.dc.b	$F8,$8F,$F6,$BF,$C3,$7F,$B6,$3E,$23,$FD,$B3,$F1,$1F,$B0,$E4,$72
	.dc.b	$7F,$A0,$FF,$80,$5E,$DE,$63,$1E,$E5,$E7,$FD,$BF,$32,$DF,$09,$9C
	.dc.b	$7F,$75,$0F,$7B,$F7,$D1,$FB,$DE,$EF,$DE,$3A,$8E,$17,$CE,$ED,$B1
	.dc.b	$B9,$FA,$30,$FF,$77,$EB,$EC,$7F,$DE,$3B,$FC,$07,$FD,$E7,$C3,$F1
	.dc.b	$1F,$F7,$AF,$1F,$C8,$7F,$DE,$FC,$BF,$31,$FF,$7C,$F3,$FC,$F7,$FC
	.dc.b	$3B,$DF,$F6,$1F,$BC,$7E,$0B,$AF,$ED,$75,$7D,$D8,$C1,$AE,$0F,$B9
	.dc.b	$3A,$24,$66,$78,$4E,$A3,$13,$A7,$F1,$A4,$E4,$FD,$2D,$F8,$E3,$1C
	.dc.b	$96,$7F,$1F,$5C,$5B,$3D,$08,$70,$3C,$E5,$BF,$13,$70,$F2,$92,$BA
	.dc.b	$5F,$87,$AF,$F6,$DC,$4F,$85,$FA,$FF,$01,$F6,$3E,$03,$EC,$FC,$0F
	.dc.b	$DA,$F8,$6F,$B7,$F0,$7F,$73,$E1,$FE,$EF,$0F,$DE,$F8,$8F,$BF,$F0
	.dc.b	$DF,$80,$F8,$9F,$C0,$FC,$47,$E0,$BE,$23,$F0,$7E,$9F,$E1,$3E,$1F
	.dc.b	$F0,$BE,$AF,$E1,$BE,$22,$74,$7F,$4A,$EA,$FC,$D4,$62,$3F,$3B,$CF
	.dc.b	$F9,$8E,$65,$F3,$9E,$13,$9C,$7C,$11,$83,$F3,$79,$F9,$5C,$75,$7E
	.dc.b	$87,$6F,$54,$63,$1F,$F3,$23,$E0,$3F,$47,$C3,$F0,$9D,$71,$FC,$67
	.dc.b	$87,$2F,$CA,$79,$73,$FE,$88,$FD,$B7,$47,$F0,$5E,$9F,$A7,$3F,$6B
	.dc.b	$87,$EF,$BF,$66,$33,$A7,$DC,$9C,$93,$7F,$C4,$12,$B9,$36,$78,$37
	.dc.b	$AB,$F0,$53,$F8,$AC,$3F,$03,$8F,$9E,$3D,$D7,$FC,$0F,$2F,$64,$63
	.dc.b	$56,$9F,$06,$31,$AF,$E0,$9F,$91,$CD,$F8,$27,$E5,$73,$7E,$09,$F9
	.dc.b	$9C,$DF,$82,$7E,$77,$37,$E0,$9F,$A1,$CD,$F7,$F8,$7F,$03,$18,$3E
	.dc.b	$A7,$37,$DF,$7C,$9C,$DF,$79,$F3,$73,$7D,$D7,$D1,$CD,$F7,$1F,$57
	.dc.b	$3F,$E6,$5F,$BD,$D1,$F9,$D7,$F0,$1D,$3F,$A1,$7F,$05,$D5,$F6,$5F
	.dc.b	$B6,$7F,$0C,$3E,$79,$39,$7E,$7B,$7A,$3E,$7B,$F6,$CF,$DC,$EB,$F6
	.dc.b	$5D,$75,$7D,$E9,$DC,$E6,$B9,$92,$33,$D5,$F6,$EB,$F1,$79,$FE,$8F
	.dc.b	$A7,$ED,$61,$FD,$07,$A7,$EC,$BF,$D0,$FA,$3E,$C3,$FD,$17,$A3,$EC
	.dc.b	$69,$F7,$7A,$3E,$C3,$F7,$FA,$3E,$C3,$F8,$1E,$8F,$B0,$FE,$0F,$A3
	.dc.b	$EC,$3F,$85,$E8,$FB,$1B,$FE,$D5,$B1,$FD,$2B,$A3,$EC,$67,$FD,$57
	.dc.b	$A3,$EC,$3F,$D5,$FA,$3E,$C3,$FD,$67,$A3,$EC,$3F,$D6,$F8,$BE,$C6
	.dc.b	$1F,$93,$E8,$FB,$0F,$E5,$7A,$3E,$C3,$F9,$7E,$8F,$B0,$FE,$67,$97
	.dc.b	$E3,$BF,$40,$FA,$7B,$7C,$B6,$99,$3E,$C1,$31,$FA,$9C,$FF,$4A,$35
	.dc.b	$18,$E7,$FB,$47,$F8,$E3,$AB,$F5,$B3,$F8,$A3,$F8,$29,$F8,$E3,$F8
	.dc.b	$49,$F9,$23,$F8,$87,$37,$D4,$AF,$EF,$5C,$DF,$51,$FE,$F9,$CD,$F5
	.dc.b	$1F,$EF,$DC,$DF,$51,$FF,$00,$73,$7D,$47,$F5,$38,$7D,$53,$F5,$45
	.dc.b	$3F,$79,$CB,$F5,$35,$FE,$51,$1F,$B6,$3F,$91,$72,$7D,$48,$FE,$61
	.dc.b	$C9,$F5,$1F,$E6,$5C,$9F,$53,$0F,$F0,$87,$27,$D4,$7F,$C2,$5C,$9F
	.dc.b	$51,$FF,$0A,$72,$7D,$47,$FC,$2D,$C9,$F5,$3A,$7F,$C2,$4F,$CA,$D3
	.dc.b	$E9,$C7,$E8,$79,$31,$1F,$4F,$3F,$B0,$D3,$E9,$B8,$BF,$A9,$1F,$DE
	.dc.b	$FA,$BF,$AF,$3F,$BA,$3F,$DF,$27,$F8,$7B,$0F,$D4,$D7,$F4,$DE,$AF
	.dc.b	$D4,$BF,$D3,$FA,$BF,$52,$FF,$51,$E6,$FA,$CF,$F5,$3E,$6F,$AC,$FF
	.dc.b	$55,$E6,$FA,$CF,$F5,$7E,$6F,$AC,$FF,$59,$E6,$FA,$CF,$E9,$F3,$FA
	.dc.b	$BA,$FD,$82,$47,$F5,$FE,$6F,$AD,$3F,$E3,$4E,$9F,$D4,$CF,$F6,$3E
	.dc.b	$7F,$AD,$3F,$E3,$8E,$7F,$AC,$FF,$8E,$B9,$FE,$B5,$7F,$69,$E7,$FA
	.dc.b	$CF,$F6,$BE,$7F,$AC,$FF,$6D,$E7,$FA,$CF,$F0,$DB,$FE,$41,$CF,$F4
	.dc.b	$DF,$AB,$D3,$FA,$38,$C4,$7D,$36,$ED,$F4,$A3,$13,$B2,$74,$7D,$C9
	.dc.b	$D1,$FD,$8B,$A7,$F5,$35,$FB,$B3,$F5,$27,$EF,$CF,$D4,$9F,$C0,$CF
	.dc.b	$D4,$9F,$DE,$74,$F8,$38,$7F,$7D,$8F,$EA,$67,$E7,$0F,$CF,$1F,$D2
	.dc.b	$8F,$EA,$67,$F4,$E3,$E9,$E3,$F5,$5F,$D4,$5F,$D4,$7F,$8E,$1C,$0E
	.dc.b	$47,$F2,$73,$B1,$FC,$AC,$FF,$06,$6B,$E4,$E9,$FF,$06,$CF,$F3,$39
	.dc.b	$FE,$04,$7F,$6F,$E7,$FC,$8B,$FE,$72,$E8,$FC,$AB,$F9,$73,$F3,$59
	.dc.b	$FF,$AB,$FC,$EF,$F4,$14,$7F,$7A,$5F,$01,$FE,$74,$5E,$C5,$F8,$0C
	.dc.b	$7B,$59,$F8,$0D,$7B,$9B,$F8,$0E,$1E,$F7,$1F,$80,$E5,$F0,$1C,$FE
	.dc.b	$07,$A7,$B7,$D7,$E2,$3B,$7B,$BD,$FE,$23,$C1,$EF,$CE,$3F,$F1,$7F
	.dc.b	$12,$31,$1E,$79,$E2,$EA,$FA,$BC,$31,$F0,$BC,$B3,$F2,$BD,$35,$F4
	.dc.b	$BD,$B7,$F6,$BE,$0E,$1F,$85,$C5,$F0,$F8,$FE,$24,$C3,$F3,$BA,$FC
	.dc.b	$AE,$77,$57,$FA,$9A,$3E,$5C,$FA,$AF,$E5,$3E,$CE,$5F,$94,$FB,$B9
	.dc.b	$7E,$53,$F0,$39,$7E,$53,$F0,$B9,$7E,$53,$F1,$39,$7E,$53,$F1,$B9
	.dc.b	$7E,$53,$F2,$39,$7E,$3E,$BE,$48,$75,$B5,$F9,$91,$B3,$E7,$F3,$7C
	.dc.b	$C7,$F0,$9C,$DF,$31,$FC,$37,$37,$CC,$7F,$11,$CD,$F3,$1F,$C5,$73
	.dc.b	$7C,$C7,$F1,$9C,$DF,$31,$FC,$77,$37,$CC,$7E,$EF,$37,$CB,$7F,$24
	.dc.b	$7F,$3D,$CD,$F1,$1F,$93,$F8,$1F,$CA,$7C,$47,$E5,$7E,$0F,$F2,$DF
	.dc.b	$11,$FE,$CA,$FE,$01,$F7,$63,$FE,$54,$5F,$13,$9F,$84,$D7,$C4,$EF
	.dc.b	$E2,$38,$78,$FC,$76,$E5,$F0,$FC,$FE,$23,$A7,$B1,$D7,$E2,$3F,$C2
	.dc.b	$BD,$7F,$EE,$1F,$66,$F9,$47,$5C,$1F,$E8,$6E,$0F,$F7,$DB,$71,$EA
	.dc.b	$E7,$3F,$0B,$BD,$7C,$AF,$1D,$FD,$9C,$1E,$DF,$1F,$DF,$3F,$A4,$E3
	.dc.b	$F8,$6F,$E9,$B9,$7F,$71,$CD,$23,$F2,$A7,$E6,$4F,$C9,$1F,$9D,$3F
	.dc.b	$24,$7E,$84,$FC,$91,$FA,$53,$F2,$47,$EA,$4F,$C9,$1F,$AD,$3F,$24
	.dc.b	$7E,$C4,$FC,$91,$F8,$A3,$F1,$87,$F0,$DF,$03,$F2,$77,$FA,$57,$F2
	.dc.b	$F1,$FC,$47,$C1,$FC,$DB,$FC,$4F,$C2,$FC,$FB,$FC,$57,$C3,$FD,$1B
	.dc.b	$FC,$5F,$C4,$FD,$3B,$FC,$67,$C5,$FD,$5B,$FC,$6F,$C6,$FD,$7B,$FC
	.dc.b	$77,$C7,$F8,$4F,$E3,$8F,$E1,$A7,$E3,$CC,$F8,$FF,$A7,$E8,$FD,$34
	.dc.b	$F2,$8C,$72,$2F,$BC,$63,$0F,$1B,$AC,$7C,$EF,$0C,$BD,$6F,$2C,$FD
	.dc.b	$EF,$4D,$3E,$1B,$DB,$5F,$8F,$AF,$EE,$33,$FD,$BB,$F2,$CF,$9C,$7C
	.dc.b	$93,$E9,$1F,$24,$FA,$C7,$C9,$3E,$D1,$F2,$4F,$BC,$7C,$93,$F0,$47
	.dc.b	$C9,$3F,$0C,$7C,$93,$F6,$E7,$C8,$21,$FE,$24,$D7,$F8,$63,$F2,$B6
	.dc.b	$FE,$45,$B7,$C9,$7F,$92,$6D,$F2,$1F,$E4,$DB,$7C,$87,$F9,$46,$DF
	.dc.b	$21,$FE,$55,$B7,$C8,$7F,$96,$6D,$F2,$1F,$E5,$DB,$7C,$87,$F1,$99
	.dc.b	$7E,$43,$6F,$94,$D8,$FE,$69,$E3,$FF,$1D,$9F,$B7,$E3,$FF,$1D,$7E
	.dc.b	$EF,$8F,$FC,$75,$FB,$FE,$3F,$F1,$D7,$F0,$3E,$3F,$F1,$D7,$F0,$7E
	.dc.b	$3F,$F1,$D7,$F0,$BE,$3F,$F1,$DC,$FF,$B6,$78,$FF,$C7,$67,$F1,$31
	.dc.b	$FC,$77,$9F,$F8,$E3,$FD,$7B,$C7,$FE,$38,$FF,$60,$F1,$FF,$80,$3F
	.dc.b	$C2,$FB,$3F,$8D,$3F,$C3,$7C,$7F,$E3,$4F,$F0,$FF,$1F,$F8,$D3,$FC
	.dc.b	$47,$C7,$FE,$34,$FF,$13,$F1,$FF,$8A,$BF,$E1,$4E,$4F,$F0,$B6,$9F
	.dc.b	$E1,$5E,$CF,$E3,$93,$FC,$67,$C7,$FE,$42,$FF,$19,$3A,$1F,$C6,$4F
	.dc.b	$01,$FC,$64,$F2,$1F,$C6,$4F,$41,$FC,$8E,$BF,$8C,$BF,$E0,$FF,$07
	.dc.b	$F8,$CB,$FE,$11,$F0,$7F,$8C,$3F,$E1,$3E,$0F,$DA,$3F,$90,$3F,$24
	.dc.b	$7F,$4A,$3F,$8C,$9F,$D3,$8F,$E3,$27,$F5,$23,$F8,$C9,$FD,$58,$FE
	.dc.b	$32,$7F,$5A,$3F,$8C,$9F,$D7,$8F,$E3,$27,$F8,$34,$FC,$41,$FB,$1F
	.dc.b	$07,$F6,$78,$FF,$69,$7F,$21,$1F,$B5,$EA,$FC,$A3,$FB,$7E,$BF,$CC
	.dc.b	$3F,$B9,$EC,$FC,$E3,$FB,$BE,$DF,$D0,$3F,$BD,$EE,$FD,$23,$FB,$FE
	.dc.b	$FF,$D4,$3F,$C0,$7C,$0F,$D6,$3F,$C0,$CF,$ED,$1E,$43,$CC,$47,$F7
	.dc.b	$EF,$1D,$ED,$7F,$8F,$A3,$3B,$A7,$31,$98,$C4,$EE,$39,$36,$33,$1B
	.dc.b	$9E,$2F,$F9,$2A,$32,$D8,$CB,$E4,$39,$9C,$67,$A4,$6A,$3C,$AD,$76
	.dc.b	$73,$8D,$9C,$3E,$23,$FC,$01,$EE,$49,$FE,$E0,$D7,$FC,$05,$1F,$E0
	.dc.b	$59,$E3,$1E,$12,$1D,$4A,$79,$EF,$D6,$DB,$F2,$46,$23,$CF,$3D,$1E
	.dc.b	$44,$38,$39,$69,$F5,$BC,$87,$3F,$28,$F0,$94,$E4,$53,$89,$83,$E2
	.dc.b	$98,$21,$B2,$1A,$2D,$7A,$5F,$C3,$61,$F1,$FA,$3F,$B0,$62,$CB,$1D
	.dc.b	$1F,$51,$FF,$17,$57,$CB,$BF,$39,$D4,$98,$75,$25,$FC,$82,$67,$F0
	.dc.b	$E7,$C7,$18,$79,$CF,$09,$FC,$46,$7F,$1A,$7A,$D6,$DF,$85,$5B,$78
	.dc.b	$C6,$5F,$86,$70,$AF,$03,$4F,$CD,$7C,$47,$F9,$27,$DE,$91,$FE,$39
	.dc.b	$6B,$FE,$4B,$8F,$80,$F8,$2B,$B3,$EA,$AE,$B7,$E3,$8D,$46,$27,$71
	.dc.b	$E1,$18,$7B,$6B,$84,$7A,$1A,$F5,$3D,$B5,$CC,$8E,$0F,$DA,$63,$F6
	.dc.b	$1F,$E0,$59,$5A,$FE,$00,$FD,$A8,$FF,$2B,$57,$27,$CE,$73,$29,$0F
	.dc.b	$26,$1E,$AA,$F7,$61,$E9,$87,$C4,$7E,$F0,$95,$FC,$1D,$AF,$F0,$56
	.dc.b	$BF,$AD,$6B,$FC,$09,$AF,$F1,$B6,$D7,$F1,$97,$F8,$4F,$C0,$7F,$A0
	.dc.b	$7E,$23,$FD,$05,$F0,$52,$FF,$CE,$4D,$7F,$D0,$91,$FE,$87,$9F,$F4
	.dc.b	$45,$74,$C3,$D3,$5E,$72,$1E,$72,$9E,$72,$9E,$72,$9E,$72,$E1,$E7
	.dc.b	$79,$E3,$F1,$EB,$C0,$48,$DD,$78,$9F,$01,$0F,$01,$4F,$01,$4F,$01
	.dc.b	$FC,$C6,$FE,$23,$F0,$88,$7F,$A9,$71,$F8,$9C,$1F,$C0,$32,$F8,$93
	.dc.b	$FE,$99,$C3,$C8,$F5,$CB,$D4,$DC,$7C,$CF,$90,$87,$90,$A7,$90,$B1
	.dc.b	$C9,$F5,$E5,$F0,$DF,$83,$5E,$07,$75,$D9,$E8,$43,$A1,$6B,$A3,$63
	.dc.b	$83,$FD,$8F,$E2,$3F,$A9,$63,$FE,$7D,$FE,$00,$7C,$2E,$38,$FF,$B2
	.dc.b	$C3,$E1,$B1,$87,$FB,$A2,$30,$FF,$6E,$BE,$53,$AC,$95,$AA,$57,$7C
	.dc.b	$3F,$DA,$D8,$FE,$42,$7D,$E4,$3E,$66,$7F,$05,$F5,$35,$F7,$9F,$7B
	.dc.b	$4F,$90,$F0,$39,$1F,$22,$38,$63,$F1,$23,$8B,$49,$A7,$C6,$7F,$C0
	.dc.b	$9B,$FE,$22,$7F,$0C,$7E,$23,$2F,$E8,$BB,$FE,$23,$6F,$EE,$FB,$FE
	.dc.b	$22,$BF,$DE,$19,$78,$32,$E8,$47,$04,$8F,$39,$F9,$AC,$FF,$91,$6C
	.dc.b	$F8,$7F,$F0,$83,$67,$FC,$2C,$D7,$FC,$20,$D7,$FB,$C3,$5F,$B7,$1C
	.dc.b	$88,$F2,$6B,$C9,$AF,$27,$C3,$F1,$1F,$76,$5F,$11,$7D,$EC,$7B,$73
	.dc.b	$0F,$C2,$6B,$F8,$A8,$C4,$7E,$0A,$3F,$1B,$3C,$EB,$A6,$1D,$71,$F0
	.dc.b	$DF,$A0,$F3,$10,$F3,$14,$F3,$14,$E0,$68,$86,$8C,$65,$F3,$A3,$BC
	.dc.b	$7E,$75,$BC,$3F,$9A,$6E,$7F,$2E,$74,$77,$29,$D8,$F0,$9D,$CF,$A7
	.dc.b	$7E,$F7,$E2,$9F,$AB,$F7,$FE,$67,$BB,$23,$38,$7E,$AA,$31,$1F,$24
	.dc.b	$E0,$6A,$FE,$9B,$6F,$E5,$3E,$5A,$F5,$36,$7D,$51,$88,$F4,$1E,$A3
	.dc.b	$D0,$7A,$88,$64,$AF,$1C,$7E,$54,$75,$C7,$E3,$91,$C1,$0E,$F3,$E1
	.dc.b	$7F,$2F,$F0,$1F,$8D,$F8,$8F,$C6,$47,$B7,$7F,$C4,$F2,$7F,$17,$6F
	.dc.b	$07,$E2,$A7,$E1,$57,$E5,$6F,$E0,$C7,$6C,$FD,$E4,$CB,$C9,$19,$6B
	.dc.b	$F0,$0F,$71,$F7,$33,$F6,$BB,$3D,$46,$F3,$F5,$3F,$12,$BB,$3A,$31
	.dc.b	$A7,$E0,$5F,$B5,$3F,$C9,$B6,$FB,$BA,$FF,$33,$DA,$EB,$FC,$C9,$FD
	.dc.b	$84,$F1,$9F,$E7,$18,$7B,$B0,$F1,$63,$E0,$7C,$C4,$33,$9F,$B9,$F3
	.dc.b	$3F,$E4,$CE,$EF,$E7,$FD,$FF,$C3,$89,$3F,$C3,$9A,$FF,$0E,$6B,$FC
	.dc.b	$39,$AF,$23,$81,$0F,$E2,$BF,$11,$FC,$27,$E0,$7F,$85,$7C,$47,$F0
	.dc.b	$BF,$84,$95,$F9,$B6,$BF,$1B,$24,$70,$8F,$D0,$57,$F1,$3B,$FE,$29
	.dc.b	$87,$8A,$7F,$AB,$37,$83,$FA,$9B,$70,$FF,$63,$4F,$94,$F5,$14,$F5
	.dc.b	$1E,$F3,$07,$43,$FD,$31,$96,$9E,$26,$C8,$60,$86,$C8,$73,$36,$62
	.dc.b	$BC,$EF,$F3,$EB,$F9,$0F,$EF,$F2,$FE,$65,$F1,$12,$3F,$81,$35,$F9
	.dc.b	$0F,$C7,$7E,$56,$5F,$1D,$FE,$11,$97,$C3,$23,$EE,$C3,$C2,$F9,$0F
	.dc.b	$31,$EA,$29,$E7,$CB,$E5,$BF,$07,$8B,$FA,$CD,$74,$8E,$55,$D1,$E4
	.dc.b	$D7,$94,$6D,$CD,$3E,$1B,$FA,$87,$C4,$7F,$51,$F8,$8F,$C2,$F1,$7F
	.dc.b	$86,$78,$2F,$2F,$F8,$61,$B8,$FF,$85,$EB,$8B,$FC,$9F,$0F,$C3,$BE
	.dc.b	$FC,$7D,$2F,$83,$2F,$13,$E4,$34,$E1,$FB,$1A,$7B,$9C,$90,$F7,$CF
	.dc.b	$47,$91,$89,$FC,$97,$07,$DC,$8F,$CF,$70,$7D,$9D,$3F,$C6,$9C,$17
	.dc.b	$5F,$F1,$9B,$6B,$FC,$65,$5C,$6B,$FC,$75,$7E,$C9,$EC,$78,$CF,$61
	.dc.b	$4F,$19,$83,$EC,$70,$FF,$90,$3E,$1B,$FC,$1E,$48,$FF,$07,$B5,$FF
	.dc.b	$07,$B5,$FF,$07,$B5,$E4,$70,$04,$2F,$0B,$F8,$07,$11,$FC,$91,$7B
	.dc.b	$BF,$E4,$B5,$E6,$5F,$7B,$1E,$74,$BF,$EC,$F4,$8C,$47,$37,$83,$AA
	.dc.b	$E9,$7D,$63,$FC,$39,$8F,$A5,$FF,$2F,$37,$1F,$F2,$DB,$78,$4D,$24
	.dc.b	$FB,$CF,$8A,$7F,$06,$39,$99,$3E,$39,$4F,$8E,$70,$21,$A2,$9A,$30
	.dc.b	$7C,$B9,$F9,$4F,$B0,$F8,$79,$7C,$07,$E6,$7B,$FF,$2C,$8F,$C8,$6B
	.dc.b	$F4,$63,$86,$CE,$1F,$AF,$AF,$D4,$AF,$9E,$7C,$83,$D4,$7B,$8F,$39
	.dc.b	$4F,$39,$E9,$21,$92,$E1,$F3,$9E,$6D,$5C,$7F,$E6,$4C,$BE,$7C,$76
	.dc.b	$9D,$3D,$8E,$84,$72,$F6,$CF,$EC,$BF,$90,$11,$F9,$3F,$43,$F2,$1F
	.dc.b	$01,$F9,$1D,$FF,$17,$A7,$C3,$BD,$E7,$F8,$F7,$F2,$3C,$1A,$D7,$E3
	.dc.b	$B6,$36,$F3,$7F,$2F,$87,$E7,$71,$F9,$51,$EB,$CF,$CA,$DD,$3D,$EF
	.dc.b	$B7,$4F,$5B,$ED,$9D,$B9,$78,$EB,$F1,$E3,$E5,$C6,$9F,$85,$5D,$1A
	.dc.b	$69,$F8,$8F,$B8,$FC,$61,$F8,$53,$F1,$A7,$E2,$48,$7E,$38,$BA,$7C
	.dc.b	$7B,$FD,$F6,$FF,$8C,$D3,$F1,$7B,$F4,$AD,$6D,$FC,$09,$FF,$6A,$63
	.dc.b	$E5,$C7,$DC,$DC,$7E,$09,$32,$F6,$4E,$8F,$D9,$66,$E1,$FC,$CF,$A5
	.dc.b	$EC,$BF,$E2,$0D,$C8,$D2,$5F,$E1,$67,$6D,$9D,$B6,$76,$FF,$1A,$9D
	.dc.b	$B6,$5F,$03,$FC,$6F,$D4,$FE,$39,$F0,$5F,$C7,$7D,$6F,$E3,$DF,$07
	.dc.b	$30,$FE,$36,$D7,$F9,$23,$77,$F8,$EF,$18,$FE,$43,$5D,$5F,$E4,$78
	.dc.b	$7F,$29,$CF,$CB,$1F,$C1,$F2,$F4,$BE,$58,$CB,$F1,$8F,$61,$0D,$14
	.dc.b	$E4,$FC,$0D,$7E,$74,$7E,$92,$39,$B4,$E0,$75,$39,$90,$E6,$68,$E0
	.dc.b	$7C,$9A,$F6,$3F,$D1,$6F,$E0,$3F,$D1,$FD,$8F,$E8,$DF,$09,$23,$FA
	.dc.b	$1B,$5F,$E9,$8F,$F4,$B8,$E3,$9F,$E4,$2B,$AB,$F9,$2B,$F9,$17,$E3
	.dc.b	$DB,$ED,$BE,$63,$D0,$7A,$CC,$10,$C1,$0F,$59,$C7,$2F,$8F,$1D,$23
	.dc.b	$4D,$76,$63,$3F,$AE,$F5,$32,$48,$E0,$7F,$6B,$F6,$BF,$B4,$7B,$7F
	.dc.b	$DA,$7D,$CF,$ED,$5F,$09,$F8,$9E,$7F,$E7,$7C,$37,$1F,$E3,$6D,$C7
	.dc.b	$FB,$A3,$72,$5F,$3A,$FE,$DF,$8F,$F7,$5C,$BE,$3C,$7C,$43,$C9,$5E
	.dc.b	$87,$D5,$C3,$F8,$4C,$7C,$F3,$99,$DC,$FE,$2F,$93,$71,$F4,$67,$55
	.dc.b	$C2,$B9,$46,$08,$FE,$EF,$6F,$B7,$A7,$F5,$8E,$0F,$BE,$FF,$5A,$E1
	.dc.b	$FC,$27,$0F,$F1,$0E,$2B,$3F,$D7,$9B,$9F,$F3,$EA,$E2,$FF,$40,$BE
	.dc.b	$AF,$F4,$2C,$3E,$0B,$F5,$73,$F3,$3E,$9C,$BC,$E4,$70,$7F,$81,$63
	.dc.b	$FC,$45,$EF,$7F,$13,$24,$7F,$0F,$6B,$FC,$2D,$AF,$EE,$DA,$FF,$76
	.dc.b	$F8,$4F,$F1,$76,$9F,$CC,$7B,$3F,$AC,$4F,$FB,$41,$7C,$26,$B6,$FF
	.dc.b	$19,$C7,$F6,$E9,$FF,$1B,$D7,$1F,$84,$F8,$79,$7F,$8D,$F7,$BC,$06
	.dc.b	$9C,$EF,$FB,$5E,$0B,$C5,$A3,$B4,$F8,$F2,$CC,$79,$B1,$FE,$87,$9F
	.dc.b	$F9,$07,$4F,$6D,$FF,$70,$6E,$1F,$22,$3E,$31,$0F,$8C,$53,$E3,$14
	.dc.b	$FF,$0D,$C7,$FB,$85,$E8,$64,$A7,$22,$1C,$8A,$72,$38,$47,$F9,$6E
	.dc.b	$7D,$7A,$7B,$3E,$1B,$DB,$F0,$BE,$EF,$86,$F7,$FC,$34,$C7,$E7,$DF
	.dc.b	$D2,$8C,$47,$D7,$6E,$7F,$6E,$B9,$CF,$03,$E7,$E1,$E0,$7F,$CC,$78
	.dc.b	$FC,$97,$CB,$C0,$D3,$E4,$B6,$33,$18,$23,$F2,$FA,$FF,$8B,$CE,$E3
	.dc.b	$E0,$9F,$18,$F8,$84,$3E,$21,$AD,$7E,$B3,$76,$7C,$3F,$E0,$3E,$1F
	.dc.b	$F0,$3C,$3F,$82,$C7,$F0,$7C,$5F,$8C,$C3,$F9,$D9,$0F,$E7,$65,$DF
	.dc.b	$F2,$CF,$C3,$CB,$F1,$D5,$D2,$31,$93,$1F,$C7,$E5,$F1,$67,$CD,$9F
	.dc.b	$B5,$F5,$1E,$D3,$D6,$63,$0F,$5B,$63,$83,$A3,$E0,$E1,$F2,$34,$F9
	.dc.b	$71,$D9,$AF,$17,$FC,$25,$AF,$D9,$8F,$DB,$6D,$F7,$9F,$DC,$6F,$F8
	.dc.b	$2E,$1F,$C3,$47,$B3,$87,$0F,$EC,$1B,$9F,$EB,$EB,$96,$FF,$C0,$F0
	.dc.b	$EC,$FE,$BA,$FF,$D5,$39,$79,$1F,$51,$F0,$49,$7F,$01,$CB,$5C,$E9
	.dc.b	$F0,$9F,$D4,$E9,$FD,$7B,$7F,$F0,$2C,$7F,$89,$5A,$FF,$A6,$5A,$FF
	.dc.b	$12,$E8,$FF,$1B,$46,$23,$F8,$5C,$60,$8F,$F8,$D6,$1F,$05,$FC,$2B
	.dc.b	$E1,$3F,$85,$FC,$27,$F0,$CF,$85,$FE,$1B,$F0,$DF,$C3,$BD,$FF,$E1
	.dc.b	$FF,$07,$32,$FC,$2B,$5F,$E4,$CD,$E0,$FE,$55,$3C,$A7,$F1,$77,$D9
	.dc.b	$FC,$6E,$3F,$C6,$B2,$FE,$37,$A7,$9A,$7E,$03,$5F,$80,$D7,$E0,$1E
	.dc.b	$B7,$0F,$C0,$30,$7F,$47,$D3,$E2,$6F,$FC,$AE,$76,$D9,$D9,$E0,$29
	.dc.b	$E0,$3E,$45,$7C,$77,$F4,$BA,$FF,$32,$F8,$2F,$74,$7F,$47,$F8,$3F
	.dc.b	$E6,$BF,$09,$23,$F5,$6D,$7F,$A6,$37,$6F,$EA,$13,$C9,$FE,$A5,$5F
	.dc.b	$1E,$BF,$A9,$69,$F1,$DF,$DA,$E7,$EA,$D3,$DA,$47,$04,$8F,$59,$F2
	.dc.b	$0F,$F6,$2B,$5F,$71,$F0,$48,$60,$86,$C8,$7F,$B0,$BF,$80,$5F,$0D
	.dc.b	$FD,$61,$7C,$3D,$F7,$BF,$19,$5F,$C3,$E5,$EF,$FE,$36,$3F,$16,$43
	.dc.b	$F1,$65,$D3,$FB,$A7,$27,$E3,$7B,$3F,$B8,$F2,$F7,$BE,$B6,$BF,$F5
	.dc.b	$56,$0C,$7C,$99,$79,$73,$FC,$7D,$FC,$43,$F1,$07,$C0,$3D,$BC,$9F
	.dc.b	$85,$70,$68,$F8,$65,$3E,$67,$3F,$EE,$76,$FA,$11,$C2,$79,$3C,$1A
	.dc.b	$F6,$DF,$F6,$6F,$E9,$0F,$C6,$9F,$CC,$B8,$7F,$09,$BF,$F8,$43,$8B
	.dc.b	$F1,$0F,$F8,$4B,$8E,$D7,$F8,$29,$B8,$7F,$82,$5F,$F0,$3D,$72,$9F
	.dc.b	$F0,$7D,$FE,$37,$7F,$F1,$07,$1F,$E3,$67,$FC,$41,$9F,$A9,$F4,$67
	.dc.b	$F8,$D7,$E4,$65,$EA,$7E,$11,$1F,$91,$A7,$F9,$E3,$81,$F0,$BF,$EC
	.dc.b	$B6,$D7,$FB,$1D,$AF,$FB,$0D,$A8,$94,$8C,$C6,$1A,$FE,$4E,$7B,$7C
	.dc.b	$37,$E4,$FE,$1B,$F2,$9F,$07,$F9,$5F,$84,$FC,$B7,$C2,$FE,$5F,$E1
	.dc.b	$BF,$31,$F0,$3F,$99,$F7,$E6,$7F,$97,$6B,$F8,$DE,$6F,$04,$63,$8F
	.dc.b	$F1,$F5,$CE,$BF,$3D,$87,$77,$F4,$19,$78,$DF,$D1,$69,$FA,$7D,$7E
	.dc.b	$21,$23,$E2,$35,$F8,$8D,$7E,$21,$EF,$36,$43,$F7,$A6,$4E,$26,$4E
	.dc.b	$3B,$FC,$B8,$E9,$1F,$2D,$AF,$EF,$67,$D0,$FC,$13,$F8,$06,$BE,$D7
	.dc.b	$15,$F5,$70,$F8,$51,$8D,$D9,$7C,$7D,$7E,$4C,$FC,$D2,$1B,$D7,$4F
	.dc.b	$F0,$D9,$E7,$AF,$EF,$30,$F0,$3F,$BE,$CB,$CA,$FF,$00,$CF,$E6,$5F
	.dc.b	$BC,$F6,$14,$F8,$87,$C2,$CB,$F1,$39,$FC,$38,$D3,$5F,$92,$7C,$C7
	.dc.b	$0F,$1D,$3E,$3C,$62,$3B,$9C,$0D,$1F,$9F,$F8,$6F,$E4,$DF,$03,$F9
	.dc.b	$48,$FE,$49,$F0,$BF,$CA,$7E,$0F,$F9,$57,$C2,$7F,$2B,$F8,$6F,$E5
	.dc.b	$9F,$0D,$2F,$F9,$53,$5F,$9C,$DD,$BF,$90,$CF,$29,$FE,$61,$7F,$CD
	.dc.b	$70,$F0,$47,$D5,$C7,$F9,$C6,$5F,$22,$3D,$07,$A8,$F4,$98,$3E,$49
	.dc.b	$92,$18,$3F,$21,$CF,$FC,$07,$5F,$8B,$B7,$CE,$23,$83,$F6,$F3,$FD
	.dc.b	$17,$9F,$EE,$F2,$FE,$F7,$8B,$F6,$F3,$FD,$27,$9F,$F6,$EF,$E2,$37
	.dc.b	$FE,$7F,$CD,$F8,$8E,$3F,$E0,$CD,$CC,$F0,$AF,$CA,$5F,$47,$F2,$D8
	.dc.b	$78,$1F,$EB,$99,$7F,$5B,$E4,$F8,$91,$F0,$88,$7C,$0C,$7F,$00,$FC
	.dc.b	$16,$93,$1F,$7C,$7F,$B6,$57,$C0,$7F,$23,$24,$7F,$23,$6B,$FC,$8D
	.dc.b	$FD,$FB,$FB,$8F,$E0,$1B,$A3,$5F,$E4,$B0,$F8,$7F,$F6,$4A,$F8,$7B
	.dc.b	$F0,$93,$0F,$F4,$04,$BE,$1F,$7F,$0D,$C3,$E1,$F8,$F8,$12,$F6,$D7
	.dc.b	$F7,$F7,$FE,$D5,$72,$74,$9E,$6F,$6B,$F0,$3C,$30,$F2,$E3,$E6,$CB
	.dc.b	$CF,$3F,$E0,$9D,$7E,$0B,$FB,$56,$BF,$CF,$DA,$FF,$84,$5B,$BF,$F4
	.dc.b	$57,$FD,$AC,$D8,$F9,$84,$3E,$61,$4E,$C7,$CE,$0B,$2F,$9F,$AF,$CE
	.dc.b	$BF,$4D,$7C,$AC,$7D,$BB,$7D,$38,$E3,$6D,$BE,$1C,$7C,$8D,$7E,$55
	.dc.b	$6F,$2E,$DB,$63,$3F,$B6,$DB,$E6,$FD,$57,$EC,$E1,$D2,$F9,$62,$CB
	.dc.b	$C7,$87,$E1,$35,$F9,$D8,$7E,$00,$F9,$C7,$C1,$3E,$61,$4F,$98,$5C
	.dc.b	$BE,$53,$F2,$09,$CD,$FD,$F7,$6F,$A3,$C1,$F9,$C9,$E0,$D7,$81,$F9
	.dc.b	$9F,$73,$F2,$BE,$EF,$E5,$BE,$1B,$F2,$FE,$FF,$E6,$3E,$1B,$F3,$3C
	.dc.b	$BF,$9A,$F8,$5F,$CD,$FC,$17,$EE,$FB,$26,$7F,$C0,$78,$7F,$A5,$BF
	.dc.b	$94,$9E,$0F,$E4,$4F,$CF,$E1,$D9,$FD,$0E,$5E,$27,$F4,$7A,$7C,$47
	.dc.b	$CE,$7C,$42,$9F,$10,$BC,$1F,$11,$FC,$49,$F1,$3C,$8E,$0E,$08,$60
	.dc.b	$87,$33,$F7,$87,$D6,$3F,$5B,$C5,$F6,$9F,$C5,$F1,$FD,$FC,$FF,$7F
	.dc.b	$CD,$F8,$39,$FD,$9F,$3F,$E1,$F1,$FC,$5C,$7E,$D0,$C9,$79,$FF,$0F
	.dc.b	$D1,$CE,$3F,$2F,$7D,$EB,$F4,$18,$FE,$7A,$7F,$7F,$D3,$FB,$67,$FC
	.dc.b	$F5,$CF,$F8,$02,$3F,$80,$6B,$FD,$9F,$A5,$F0,$5F,$AD,$8F,$D2,$B7
	.dc.b	$0F,$D7,$36,$77,$3F,$B3,$9C,$B6,$1F,$07,$FC,$7B,$E1,$3F,$8F,$FC
	.dc.b	$37,$F2,$0F,$86,$FE,$43,$EE,$FF,$22,$F8,$1F,$F1,$3F,$0F,$EB,$FF
	.dc.b	$80,$7F,$6D,$9F,$F8,$61,$B1,$9F,$86,$DF,$2F,$F9,$42,$51,$8A,$FE
	.dc.b	$0B,$18,$7F,$CA,$D5,$C5,$B5,$C6,$B8,$6C,$BE,$15,$E1,$9C,$C7,$8A
	.dc.b	$BC,$07,$A6,$7F,$9B,$E3,$ED,$7F,$9C,$E7,$F0,$5F,$E7,$7A,$FF,$40
	.dc.b	$DF,$E5,$12,$3E,$43,$5F,$E9,$79,$7C,$C6,$EF,$F4,$1B,$3B,$F0,$9D
	.dc.b	$1F,$D3,$A3,$94,$FF,$53,$6B,$93,$EB,$61,$E5,$7E,$75,$7B,$BE,$1B
	.dc.b	$EA,$C7,$C6,$3E,$5F,$C1,$7C,$2D,$64,$67,$3F,$91,$BF,$C6,$24,$67
	.dc.b	$7F,$97,$C1,$F4,$63,$E6,$19,$9E,$6F,$DF,$BF,$9F,$87,$8A,$3F,$05
	.dc.b	$8F,$D2,$CB,$D1,$87,$F2,$CC,$FE,$65,$7C,$A2,$1F,$18,$A7,$C3,$29
	.dc.b	$F0,$CD,$1F,$A3,$D3,$E3,$BF,$A2,$36,$53,$EA,$E6,$F8,$7F,$C2,$FC
	.dc.b	$27,$E7,$A7,$FB,$CF,$27,$E2,$3E,$17,$F0,$D1,$F9,$BF,$86,$FC,$5F
	.dc.b	$C3,$FE,$33,$8E,$4F,$C7,$6B,$FA,$16,$EB,$FA,$57,$8C,$7C,$DA,$EB
	.dc.b	$3F,$A9,$BF,$C9,$4F,$F8,$0B,$97,$CA,$FD,$3C,$FE,$3B,$F0,$0F,$88
	.dc.b	$53,$E2,$18,$21,$83,$E4,$6F,$F9,$07,$E2,$1A,$29,$B2,$9F,$DC,$37
	.dc.b	$FB,$3C,$5F,$B9,$9F,$C8,$4F,$EA,$8F,$C4,$71,$7E,$13,$2F,$E0,$95
	.dc.b	$F8,$C9,$FC,$49,$23,$F1,$84,$8F,$DE,$47,$E3,$4F,$C9,$3C,$CF,$E1
	.dc.b	$13,$F8,$D9,$FE,$21,$C5,$F8,$DA,$FF,$18,$F4,$7F,$8D,$B1,$F8,$F1
	.dc.b	$ED,$E5,$FD,$A3,$79,$BF,$68,$DE,$5F,$C4,$CF,$EE,$74,$F8,$CF,$F0
	.dc.b	$DF,$84,$FE,$59,$CC,$69,$2F,$F9,$64,$3D,$AF,$E2,$3F,$09,$FC,$4B
	.dc.b	$E1,$7F,$89,$FC,$37,$F1,$4F,$85,$FE,$2B,$F0,$DF,$C5,$BE,$03,$F8
	.dc.b	$BF,$C0,$FF,$18,$F8,$2F,$E3,$BF,$04,$F8,$3A,$4D,$BF,$75,$19,$6C
	.dc.b	$65,$FF,$64,$C6,$5F,$E6,$93,$D9,$FD,$F5,$F8,$9F,$E0,$18,$F9,$9F
	.dc.b	$E0,$59,$FA,$9F,$E0,$9A,$FF,$2C,$DF,$E0,$47,$EB,$1A,$FF,$B8,$B7
	.dc.b	$F8,$CF,$C4,$8C,$BF,$D0,$5F,$F2,$5C,$6A,$39,$1B,$21,$FB,$89,$F1
	.dc.b	$3F,$D3,$6B,$E2,$FC,$27,$C6,$F8,$5F,$8F,$F0,$DF,$23,$DD,$F7,$5F
	.dc.b	$C2,$DF,$E2,$75,$FF,$91,$78,$3E,$23,$63,$3D,$AA,$FF,$2B,$D6,$A3
	.dc.b	$1C,$C8,$C4,$EA,$B8,$BF,$D6,$B8,$FC,$4F,$D6,$CB,$CC,$FD,$8D,$3F
	.dc.b	$08,$FF,$2F,$E6,$FE,$AF,$BF,$CD,$8F,$82,$7C,$BE,$1F,$92,$FF,$66
	.dc.b	$6E,$7F,$95,$8F,$8F,$CF,$F9,$8E,$2F,$99,$1C,$4E,$67,$D1,$3F,$C0
	.dc.b	$5A,$BD,$AF,$C7,$FB,$7F,$90,$8F,$C1,$4F,$F7,$BE,$5F,$C3,$3F,$DF
	.dc.b	$79,$BF,$2B,$EE,$7E,$5E,$3F,$27,$F0,$BF,$98,$23,$82,$1F,$94,$72
	.dc.b	$FE,$50,$FC,$DC,$7E,$8D,$FE,$A5,$D5,$FA,$77,$FA,$A7,$5F,$95,$FD
	.dc.b	$0E,$5F,$A6,$CF,$E4,$47,$E2,$88,$7C,$12,$9F,$04,$A7,$CB,$3F,$01
	.dc.b	$AF,$C9,$DF,$E6,$ED,$F2,$A7,$81,$F9,$DE,$0F,$C9,$F3,$FF,$88,$7A
	.dc.b	$BF,$27,$3F,$B4,$EA,$FC,$9B,$FB,$6E,$AF,$C9,$D7,$E3,$E3,$F6,$E4
	.dc.b	$3F,$2C,$57,$4E,$08,$E0,$FF,$18,$74,$7E,$3A,$3F,$7F,$D1,$F8,$E7
	.dc.b	$F8,$0F,$47,$E3,$9F,$F1,$A7,$47,$E3,$9F,$EF,$DA,$FB,$88,$FE,$18
	.dc.b	$F8,$87,$CB,$DB,$F9,$1F,$3B,$DE,$FF,$27,$36,$7F,$17,$18,$F8,$6F
	.dc.b	$E3,$3E,$D7,$F1,$AF,$80,$FE,$37,$F0,$3F,$C7,$3E,$0B,$F8,$EF,$C1
	.dc.b	$FF,$1E,$F8,$6F,$E3,$FF,$0B,$FC,$83,$E1,$BF,$90,$FA,$FF,$E3,$EF
	.dc.b	$E0,$1F,$F6,$3E,$DF,$EC,$A6,$F8,$86,$92,$38,$10,$FE,$4F,$3D,$5F
	.dc.b	$E5,$37,$E1,$7F,$95,$E3,$E5,$7F,$96,$E7,$E9,$7F,$97,$EB,$FC,$D3
	.dc.b	$6F,$E6,$BB,$FC,$22,$4F,$C2,$21,$FE,$44,$E0,$F8,$CD,$8F,$F7,$5C
	.dc.b	$62,$3F,$88,$1F,$C4,$23,$9B,$C6,$7E,$5B,$E4,$20,$E7,$FF,$70,$68
	.dc.b	$F6,$7E,$47,$BD,$F2,$7D,$BF,$95,$F0,$9F,$2F,$DD,$F9,$B8,$FF,$90
	.dc.b	$F0,$7B,$DF,$3C,$91,$96,$BD,$FC,$CF,$E4,$73,$C2,$34,$7C,$E3,$A5
	.dc.b	$78,$1E,$D8,$79,$1F,$06,$5F,$84,$7D,$79,$FF,$35,$DF,$E7,$47,$F5
	.dc.b	$ED,$BE,$1B,$F1,$B9,$3F,$99,$46,$23,$F9,$AF,$07,$CB,$6C,$7C,$AE
	.dc.b	$9F,$CC,$BB,$8F,$9C,$7F,$A8,$F4,$7B,$FF,$8F,$F8,$0F,$C0,$4F,$F6
	.dc.b	$AE,$AF,$C2,$3F,$DB,$39,$BF,$10,$FF,$6E,$E7,$FC,$63,$F1,$4F,$CB
	.dc.b	$7C,$37,$E4,$5C,$6B,$F9,$77,$1B,$FD,$86,$F1,$FC,$C6,$9C,$8F,$EF
	.dc.b	$FD,$5F,$A6,$7F,$C0,$3D,$7F,$AA,$7F,$C0,$79,$F9,$AF,$FA,$F7,$67
	.dc.b	$AE,$7E,$F6,$DF,$98,$7F,$2A,$7C,$BE,$2F,$90,$FC,$43,$F2,$C6,$8F
	.dc.b	$96,$7F,$26,$F6,$EB,$A2,$1C,$FA,$FF,$2E,$FF,$84,$BB,$3E,$F3,$F0
	.dc.b	$B8,$FF,$2C,$FE,$47,$B3,$F2,$CF,$E4,$FB,$3F,$2C,$FE,$2C,$E1,$D9
	.dc.b	$F9,$2D,$7F,$3E,$DE,$1F,$C9,$B6,$7F,$83,$3F,$9E,$7C,$1D,$9F,$93
	.dc.b	$7C,$5D,$9F,$93,$7C,$9D,$9E,$6C,$3F,$C0,$DA,$7A,$9F,$F0,$46,$BF
	.dc.b	$26,$BD,$E7,$C2,$3E,$57,$57,$E3,$79,$BF,$2B,$1F,$C0,$BA,$FF,$C9
	.dc.b	$7C,$0F,$82,$FF,$7B,$36,$7F,$25,$F0,$DF,$C7,$FE,$13,$F9,$07,$C3
	.dc.b	$7F,$21,$F8,$6F,$E4,$5E,$BF,$F2,$3F,$67,$F9,$27,$B5,$FC,$97,$DE
	.dc.b	$FE,$4D,$F0,$5F,$C9,$FE,$17,$F9,$47,$C3,$4D,$7F,$7C,$D7,$E9,$37
	.dc.b	$93,$F8,$4B,$7A,$BF,$7B,$5D,$2B,$F8,$36,$1E,$07,$F8,$46,$5E,$47
	.dc.b	$F8,$56,$9E,$87,$F8,$66,$DF,$CE,$77,$FE,$75,$C3,$F0,$09,$3F,$B1
	.dc.b	$6B,$FE,$EA,$E1,$F8,$8D,$8F,$E4,$BC,$5F,$29,$DC,$7F,$03,$95,$7C
	.dc.b	$98,$FE,$39,$C5,$F2,$5F,$F7,$07,$F0,$03,$6F,$95,$18,$F8,$7F,$94
	.dc.b	$EB,$2F,$94,$F0,$F8,$1F,$E6,$BC,$5C,$61,$CC,$C5,$5A,$C7,$FC,$8F
	.dc.b	$18,$AE,$44,$3F,$81,$B9,$6C,$F1,$7A,$C6,$6F,$F9,$EF,$16,$70,$F0
	.dc.b	$63,$E8,$BE,$79,$7B,$1E,$B9,$7C,$17,$F6,$19,$FC,$57,$F6,$5A,$7D
	.dc.b	$47,$E9,$10,$FA,$05,$3E,$76,$DF,$5B,$0F,$A5,$19,$3D,$FF,$05,$E5
	.dc.b	$D7,$FD,$DD,$1F,$43,$E1,$F1,$F0,$BF,$87,$8F,$E0,$5D,$1F,$8B,$7F
	.dc.b	$82,$74,$7E,$3D,$F8,$BF,$0D,$F9,$37,$FC,$01,$D3,$78,$7F,$C0,$0D
	.dc.b	$E5,$34,$F1,$57,$3E,$4F,$EC,$B8,$77,$7F,$B3,$E3,$E3,$7E,$E6,$1F
	.dc.b	$6F,$2F,$56,$7E,$BF,$86,$F6,$6D,$F1,$F5,$F8,$A4,$3E,$29,$F0,$CF
	.dc.b	$F0,$B7,$07,$C7,$F1,$7F,$B3,$BF,$98,$8F,$ED,$3C,$5F,$4F,$3F,$EF
	.dc.b	$3C,$7F,$5D,$F1,$F1,$7E,$3E,$BF,$94,$75,$7C,$DE,$6F,$C0,$56,$B9
	.dc.b	$7F,$2F,$5F,$8A,$3F,$1D,$1C,$BA,$3E,$A1,$23,$3C,$FF,$96,$8C,$47
	.dc.b	$F3,$5E,$8F,$B5,$7D,$AB,$F3,$78,$7E,$37,$0F,$2C,$79,$CF,$F0,$1E
	.dc.b	$7E,$B7,$DB,$87,$F8,$37,$5F,$80,$FC,$43,$E3,$71,$FF,$8E,$63,$2D
	.dc.b	$8F,$E9,$C7,$E8,$BE,$1F,$FC,$4A,$FF,$80,$76,$70,$7E,$3F,$D3,$FC
	.dc.b	$87,$B7,$F9,$1F,$57,$F2,$5E,$DF,$E4,$FD,$FF,$CA,$7C,$24,$CB,$F8
	.dc.b	$1F,$C0,$FE,$5F,$E1,$3F,$31,$EE,$FE,$67,$E1,$A5,$7F,$08,$6B,$FB
	.dc.b	$86,$F2,$7D,$66,$F1,$FF,$00,$AE,$95,$D8,$FE,$23,$8F,$85,$FE,$03
	.dc.b	$97,$99,$FE,$2B,$A7,$A9,$FE,$2F,$B7,$EA,$F7,$FD,$67,$0F,$C2,$9F
	.dc.b	$E1,$AD,$7E,$31,$9E,$2F,$E2,$91,$96,$C6,$78,$93,$FC,$B2,$B9,$CE
	.dc.b	$5F,$98,$FF,$B8,$F6,$FD,$2F,$F0,$0F,$FB,$83,$02,$F7,$BF,$8A,$2F
	.dc.b	$81,$FA,$17,$C7,$83,$FD,$D3,$2F,$80,$DF,$C0,$F0,$F8,$3F,$8F,$7F
	.dc.b	$C9,$38,$F9,$FC,$24,$8E,$47,$F3,$0C,$11,$98,$FF,$16,$C6,$1D,$E9
	.dc.b	$E6,$9D,$6B,$EA,$7E,$FC,$FF,$44,$E6,$F1,$3C,$2D,$97,$B3,$3F,$5C
	.dc.b	$F6,$D7,$DF,$B7,$C0,$F8,$7F,$8F,$3F,$CF,$38,$3E,$BB,$F3,$F8,$7E
	.dc.b	$97,$17,$D0,$8C,$10,$FC,$EE,$FF,$59,$77,$7E,$79,$FB,$A6,$B5,$7B
	.dc.b	$3F,$43,$E1,$BF,$11,$3F,$D3,$39,$FF,$18,$FC,$4F,$87,$FC,$87,$27
	.dc.b	$F0,$CE,$AF,$E6,$DD,$9F,$63,$D3,$FB,$3E,$A7,$E6,$B0,$FE,$A3,$DD
	.dc.b	$77,$FE,$BC,$DE,$03,$4E,$33,$F2,$6F,$AB,$FA,$02,$1E,$33,$FB,$9F
	.dc.b	$7F,$ED,$DF,$EE,$BE,$07,$AD,$FA,$5C,$1F,$29,$F8,$B8,$7C,$96,$F1
	.dc.b	$FC,$C2,$3F,$2F,$8F,$E4,$67,$F2,$E3,$F8,$47,$27,$D6,$E0,$FC,$DF
	.dc.b	$2F,$DA,$9F,$C0,$73,$FD,$E9,$FC,$FF,$47,$E5,$A7,$F0,$9D,$BF,$96
	.dc.b	$9F,$E6,$BD,$BF,$8C,$EE,$FC,$EF,$75,$BF,$CB,$37,$4F,$C9,$47,$E3
	.dc.b	$63,$1B,$AF,$A5,$FE,$9B,$1F,$05,$FE,$73,$2F,$24,$FE,$AB,$3F,$44
	.dc.b	$FE,$83,$4F,$64,$FF,$84,$35,$F9,$13,$FC,$9F,$7F,$8A,$FC,$83,$F2
	.dc.b	$06,$4F,$E5,$FA,$FE,$F3,$E1,$7F,$77,$EF,$7E,$F3,$D7,$FD,$EF,$BF
	.dc.b	$FB,$EF,$84,$FD,$FF,$C0,$FF,$00,$F8,$4F,$E0,$3F,$07,$FC,$0B,$E1
	.dc.b	$BF,$C2,$DE,$A7,$FB,$D8,$FF,$3F,$EB,$FE,$A2,$5F,$07,$AF,$80,$93
	.dc.b	$FB,$89,$46,$23,$83,$7A,$BF,$BB,$CE,$A7,$84,$F8,$2B,$C3,$F0,$1E
	.dc.b	$29,$F2,$1F,$BD,$C7,$D0,$FE,$FF,$3F,$63,$FC,$07,$5F,$80,$FF,$03
	.dc.b	$DF,$F8,$EF,$07,$B3,$E0,$FF,$96,$68,$69,$2B,$F9,$94,$65,$F9,$BE
	.dc.b	$4A,$BE,$6C,$7F,$0F,$E2,$F8,$FF,$05,$E2,$8F,$E3,$79,$FC,$AF,$83
	.dc.b	$F9,$7F,$09,$EA,$9F,$94,$7B,$CF,$91,$C1,$F4,$3E,$1A,$47,$CB,$6B
	.dc.b	$F4,$5B,$C7,$F2,$9B,$D4,$9F,$F2,$EE,$DF,$59,$EA,$7D,$53,$C5,$5F
	.dc.b	$4B,$0F,$34,$7C,$B3,$F9,$57,$3F,$B1,$F7,$1F,$2C,$FE,$5D,$B7,$CB
	.dc.b	$7E,$3F,$07,$C9,$7F,$91,$E1,$F2,$5B,$1F,$D5,$DF,$2F,$C3,$7E,$3B
	.dc.b	$D4,$FC,$7F,$B5,$F9,$69,$FC,$57,$C1,$FE,$4B,$D7,$FB,$F1,$FD,$1B
	.dc.b	$A3,$F2,$BF,$09,$F9,$6F,$86,$FC,$3C,$FE,$28,$87,$E4,$0B,$C1,$F9
	.dc.b	$0E,$CF,$C9,$71,$7C,$7E,$F7,$07,$E4,$EF,$81,$2F,$83,$F9,$DE,$DF
	.dc.b	$03,$F9,$FC,$BF,$52,$FF,$51,$EF,$F4,$3E,$A3,$F2,$87,$E9,$77,$F8
	.dc.b	$0F,$E9,$F8,$7E,$53,$F1,$8F,$96,$43,$07,$F4,$CF,$07,$FB,$9F,$27
	.dc.b	$E4,$A3,$FB,$4F,$67,$D7,$7F,$B5,$F3,$7E,$F9,$FD,$A7,$57,$E4,$9F
	.dc.b	$EC,$DE,$07,$EB,$5F,$C2,$9F,$BD,$EB,$FC,$5B,$F8,$AE,$BB,$C5,$F1
	.dc.b	$9B,$C4,$69,$F9,$B9,$E4,$7E,$3C,$EE,$7F,$05,$EB,$FC,$7B,$FA,$63
	.dc.b	$F1,$E7,$F0,$9E,$DF,$DC,$BF,$E0,$4E,$FF,$DC,$BF,$C3,$7B,$7F,$1C
	.dc.b	$FE,$7C,$FC,$FB,$87,$EE,$61,$F9,$1F,$27,$FA,$5A,$F7,$7F,$8E,$FC
	.dc.b	$0F,$F1,$EF,$86,$FE,$3F,$F0,$7F,$C8,$3D,$FF,$E4,$3F,$0B,$FE,$B7
	.dc.b	$FE,$00,$FF,$3D,$77,$7F,$A3,$97,$C0,$E7,$4D,$7C,$16,$FE,$0F,$E1
	.dc.b	$EF,$FB,$92,$1F,$B9,$2E,$1F,$B9,$6F,$85,$E8,$E5,$E5,$1F,$E2,$2C
	.dc.b	$BF,$D9,$53,$F9,$77,$FC,$1B,$19,$CB,$CD,$3E,$88,$FE,$65,$A7,$B1
	.dc.b	$FE,$69,$B7,$C0,$7F,$9B,$70,$7B,$7E,$13,$DD,$F0,$BF,$E4,$FC,$8D
	.dc.b	$25,$7F,$95,$63,$2F,$F5,$0E,$3F,$98,$F6,$21,$FD,$1F,$8B,$E5,$47
	.dc.b	$96,$1B,$7C,$BD,$3E,$76,$BF,$3F,$E1,$BE,$87,$9F,$F0,$AF,$75,$F0
	.dc.b	$7C,$1F,$A8,$48,$F8,$CD,$76,$7F,$BF,$B3,$FF,$19,$38,$EB,$46,$32
	.dc.b	$FB,$95,$C5,$FB,$DC,$7E,$2B,$FA,$59,$79,$9F,$A9,$8F,$AE,$BF,$09
	.dc.b	$A7,$CF,$9F,$EC,$DC,$3F,$8D,$7E,$79,$F6,$F8,$3E,$7B,$F2,$F8,$3E
	.dc.b	$6B,$FE,$4A,$D3,$E6,$B6,$3F,$94,$6B,$E8,$F8,$5F,$B5,$3E,$BA,$7B
	.dc.b	$3F,$97,$F8,$1F,$CD,$4F,$E2,$7D,$BF,$CD,$7C,$37,$E6,$FE,$07,$F3
	.dc.b	$13,$F8,$DE,$AF,$CF,$7C,$0F,$E6,$23,$F2,$C4,$E1,$FC,$B3,$78,$BF
	.dc.b	$2C,$FE,$60,$FE,$0C,$7E,$5E,$7F,$46,$FE,$67,$B3,$C0,$FE,$6F,$1F
	.dc.b	$D6,$65,$E6,$8F,$56,$7F,$AF,$D3,$DB,$97,$EB,$F6,$FD,$97,$C3,$7C
	.dc.b	$8B,$F8,$87,$C9,$21,$83,$F9,$31,$FD,$8B,$AF,$E9,$3F,$50,$FF,$0B
	.dc.b	$75,$FE,$51,$FF,$0C,$77,$7E,$A9,$FD,$61,$F8,$03,$F6,$5D,$7F,$93
	.dc.b	$C7,$F0,$F1,$F9,$02,$1F,$4F,$AF,$F1,$5C,$3F,$97,$8C,$CF,$07,$47
	.dc.b	$EF,$7A,$FF,$1C,$FD,$BE,$BF,$C7,$3F,$76,$BF,$1C,$7F,$01,$EB,$FC
	.dc.b	$75,$FF,$8A,$BA,$FD,$AF,$F8,$B3,$5F,$DE,$5F,$F0,$9E,$AF,$C6,$BF
	.dc.b	$C9,$FA,$BF,$43,$18,$8F,$E1,$8F,$E3,$BA,$BF,$CB,$91,$FC,$47,$E0
	.dc.b	$7F,$91,$1F,$CB,$F0,$67,$FD,$A7,$F8,$07,$FD,$98,$7F,$B2,$EB,$FD
	.dc.b	$72,$FF,$B2,$DB,$EE,$71,$9E,$1F,$0F,$65,$9A,$DE,$F6,$63,$1F,$03
	.dc.b	$67,$FD,$53,$5D,$BE,$1B,$04,$70,$43,$F8,$04,$23,$51,$86,$E7,$EC
	.dc.b	$8F,$1C,$F6,$78,$D7,$87,$0F,$16,$3E,$3C,$FC,$9F,$01,$E5,$F8,$7F
	.dc.b	$37,$C0,$FC,$8C,$7E,$51,$FC,$EF,$E0,$BD,$5F,$0F,$EB,$F8,$6F,$67
	.dc.b	$C3,$FF,$BF,$F6,$34,$97,$FE,$7E,$E4,$FA,$AF,$F0,$1E,$4E,$CF,$BB
	.dc.b	$2F,$8D,$EB,$FC,$7F,$43,$CD,$3F,$53,$D7,$F5,$3F,$5B,$E1,$7E,$67
	.dc.b	$B1,$F3,$7E,$12,$4F,$DA,$6B,$F2,$9B,$CF,$F1,$FA,$7E,$C4,$63,$AB
	.dc.b	$F0,$8F,$4A,$EC,$7E,$13,$A3,$ED,$65,$E4,$8F,$AB,$9F,$A1,$FC,$4E
	.dc.b	$BE,$CC,$BF,$B8,$6D,$FD,$CF,$93,$E4,$D7,$E3,$B8,$FE,$4B,$FE,$03
	.dc.b	$3F,$29,$C5,$F2,$1F,$F7,$AF,$1F,$CC,$7F,$D4,$EB,$DC,$FC,$67,$C0
	.dc.b	$FE,$37,$DA,$FC,$87,$F0,$0F,$3D,$3D,$D7,$F9,$DC,$3F,$17,$D5,$AF
	.dc.b	$80,$FC,$1C,$7F,$9C,$6B,$E2,$5F,$F9,$C6,$3F,$12,$7E,$93,$E2,$3F
	.dc.b	$1F,$CB,$FA,$9F,$02,$CF,$E4,$1B,$CF,$FA,$36,$E7,$F9,$99,$FC,$E6
	.dc.b	$3E,$4C,$3F,$49,$E0,$7A,$3D,$DF,$1D,$7F,$5F,$CF,$FD,$73,$7E,$AD
	.dc.b	$7F,$2D,$3E,$7D,$3F,$CC,$55,$EF,$7F,$79,$B7,$C4,$9F,$DD,$F0,$7C
	.dc.b	$4F,$82,$F8,$BE,$E7,$D0,$9F,$9B,$C7,$F3,$DF,$E1,$DB,$FC,$F7,$F9
	.dc.b	$9F,$0F,$F7,$0E,$5F,$B9,$9F,$F8,$6F,$9B,$F4,$F5,$FC,$23,$BB,$F3
	.dc.b	$0F,$F0,$83,$F3,$07,$E2,$4F,$C5,$9F,$BC,$ED,$FC,$83,$FB,$8E,$DB
	.dc.b	$8F,$E5,$23,$F3,$44,$3F,$34,$5E,$2F,$BB,$3F,$97,$8F,$E0,$FD,$DF
	.dc.b	$A6,$7F,$84,$F6,$FE,$5E,$BF,$07,$DB,$F9,$79,$FD,$81,$FB,$E3,$F6
	.dc.b	$87,$EF,$CF,$E4,$7E,$07,$EF,$DF,$E4,$BC,$1F,$1D,$FC,$B9,$F9,$93
	.dc.b	$F8,$8B,$FE,$00,$E5,$7B,$BF,$94,$F7,$E5,$7D,$93,$F9,$4C,$BD,$FF
	.dc.b	$CC,$FB,$DF,$9A,$F8,$0F,$CD,$FC,$37,$E7,$3E,$0B,$F3,$BF,$0D,$F9
	.dc.b	$EF,$86,$FC,$FF,$C2,$FE,$83,$E1,$A6,$9D,$79,$9A,$A7,$33,$F8,$D6
	.dc.b	$F1,$FE,$81,$BD,$5F,$8C,$7A,$CF,$3A,$63,$E2,$8C,$67,$E6,$75,$AF
	.dc.b	$A9,$E1,$BF,$C0,$7F,$39,$C1,$FB,$6E,$1F,$DB,$FA,$1F,$B8,$F8,$6F
	.dc.b	$DC,$FA,$5F,$BA,$F6,$7E,$66,$1F,$19,$BC,$3E,$C7,$E7,$7C,$37,$5E
	.dc.b	$2F,$DC,$E7,$FC,$0B,$4F,$E0,$7E,$D7,$F5,$3E,$7F,$F4,$1F,$27,$FB
	.dc.b	$25,$FF,$41,$E7,$F1,$DF,$F3,$AB,$FB,$1E,$6F,$9C,$E7,$E0,$FE,$2B
	.dc.b	$F6,$39,$AE,$DF,$39,$BC,$DF,$3E,$7E,$44,$60,$F1,$57,$48,$EC,$78
	.dc.b	$7E,$0B,$CD,$F0,$DE,$7C,$FD,$33,$F2,$8F,$A5,$B7,$BE,$3F,$80,$F0
	.dc.b	$7B,$7D,$BF,$95,$5F,$5F,$8F,$E5,$3F,$95,$D7,$E4,$37,$57,$1F,$E5
	.dc.b	$F9,$BE,$37,$C3,$7C,$7F,$6B,$ED,$5F,$F1,$3E,$3F,$CA,$3F,$83,$ED
	.dc.b	$FC,$A3,$F2,$7D,$BF,$9D,$EE,$7E,$22,$3F,$CF,$14,$F7,$BE,$91,$23
	.dc.b	$E8,$35,$DB,$7A,$BF,$AB,$4F,$19,$D9,$F9,$3F,$03,$BB,$E1,$3F,$26
	.dc.b	$79,$4F,$C9,$9F,$99,$8F,$5C,$F4,$CB,$DF,$3F,$9D,$D3,$F2,$6F,$E0
	.dc.b	$FD,$EF,$C4,$7C,$07,$CA,$8F,$EA,$DC,$3F,$29,$BC,$3F,$29,$1F,$CA
	.dc.b	$17,$27,$D7,$8F,$DF,$F3,$7E,$55,$FD,$67,$67,$DF,$7F,$25,$D1,$F9
	.dc.b	$7E,$9F,$D7,$47,$F4,$3F,$0B,$F1,$2F,$E3,$08,$7E,$EC,$BB,$FE,$ED
	.dc.b	$BD,$5F,$BB,$69,$F9,$93,$F3,$87,$EF,$67,$BC,$FD,$0A,$FC,$9C,$FE
	.dc.b	$96,$BF,$26,$FF,$55,$EE,$F6,$6B,$FC,$3B,$3F,$EE,$3A,$7E,$4E,$BF
	.dc.b	$84,$77,$FE,$F1,$FE,$15,$D5,$FB,$AD,$FE,$44,$7C,$A8,$FF,$53,$F2
	.dc.b	$7F,$A9,$23,$F6,$9F,$05,$FB,$9F,$83,$FD,$D7,$C2,$49,$FD,$C7,$C3
	.dc.b	$7E,$F7,$E1,$BF,$7D,$F0,$BF,$BF,$F4,$BF,$80,$7A,$5F,$E5,$3F,$E0
	.dc.b	$17,$C0,$DF,$84,$99,$7F,$27,$9F,$F0,$37,$47,$F8,$95,$BC,$FF,$73
	.dc.b	$99,$18,$AF,$F3,$0B,$63,$37,$D2,$7F,$DB,$13,$93,$C2,$FF,$0F,$C7
	.dc.b	$CC,$FF,$11,$CF,$D4,$FB,$0F,$E2,$9B,$7B,$DF,$D5,$EF,$FC,$6F,$83
	.dc.b	$F8,$E7,$C3,$FB,$7D,$CF,$77,$C3,$7B,$FE,$1F,$E0,$7C,$37,$C1,$F8
	.dc.b	$1F,$E5,$3B,$1F,$9F,$8F,$3C,$FE,$17,$C1,$7C,$7F,$83,$F9,$1E,$BF
	.dc.b	$C9,$F8,$6F,$2D,$7E,$F3,$A3,$D2,$FE,$FB,$8F,$E7,$7B,$1F,$07,$A3
	.dc.b	$E5,$73,$5E,$8F,$95,$BF,$C3,$8C,$61,$F2,$9B,$D1,$F2,$A3,$EA,$61
	.dc.b	$F6,$9F,$E6,$D9,$78,$31,$FE,$0B,$D5,$E5,$9F,$B5,$AF,$A5,$F5,$9F
	.dc.b	$8A,$21,$F5,$3D,$EF,$8E,$FE,$07,$93,$E3,$BF,$83,$E3,$F8,$E7,$E5
	.dc.b	$39,$FF,$15,$ED,$7E,$4E,$7E,$A9,$0F,$E3,$FD,$BF,$92,$7E,$E1,$F8
	.dc.b	$DF,$6F,$F2,$5D,$D2,$3F,$67,$C1,$FD,$25,$BE,$0F,$F4,$97,$F7,$84
	.dc.b	$DF,$F7,$8D,$E4,$FD,$E3,$F8,$EB,$E8,$FE,$43,$0F,$13,$FC,$CB,$BF
	.dc.b	$FB,$16,$5F,$8B,$C3,$F9,$6F,$7F,$AE,$BF,$2F,$A7,$E3,$1F,$D2,$77
	.dc.b	$7E,$A7,$DC,$FD,$57,$BD,$FA,$BF,$7B,$F1,$95,$FD,$37,$AB,$F1,$8F
	.dc.b	$F2,$2F,$25,$E7,$7F,$6F,$F0,$7F,$1D,$3F,$A4,$F0,$7E,$B3,$FA,$6E
	.dc.b	$AF,$B4,$FF,$48,$EB,$FB,$DC,$BF,$DD,$FB,$3F,$05,$3F,$AC,$ED,$FC
	.dc.b	$3E,$9F,$8B,$8F,$DD,$77,$60,$8E,$0F,$C5,$9F,$B1,$3F,$16,$43,$C0
	.dc.b	$7F,$03,$F0,$3F,$42,$FF,$05,$F0,$7F,$85,$69,$EA,$E6,$FE,$AB,$B7
	.dc.b	$F0,$DF,$03,$F1,$77,$FC,$03,$C0,$FC,$5B,$FC,$0B,$C0,$FC,$5B,$F1
	.dc.b	$78,$BF,$47,$F0,$5F,$C5,$7E,$1B,$F8,$B7,$C2,$7F,$BA,$7F,$80,$7F
	.dc.b	$CF,$97,$FE,$EE,$7F,$D0,$7C,$9F,$C4,$DF,$F4,$B6,$BF,$EB,$B5,$F0
	.dc.b	$BC,$7E,$16,$5F,$F9,$0E,$7F,$CA,$6D,$F5,$7B,$11,$C1,$05,$F5,$8D
	.dc.b	$46,$30,$F1,$46,$27,$53,$C5,$F2,$5F,$8A,$38,$1E,$BC,$3D,$DF,$0D
	.dc.b	$EF,$F8,$7F,$81,$3F,$08,$FE,$69,$A7,$F3,$9D,$7E,$27,$A5,$F1,$7D
	.dc.b	$3F,$8D,$E8,$7C,$7F,$85,$F9,$1E,$F7,$C9,$F8,$6F,$F2,$3F,$11,$FE
	.dc.b	$EB,$9F,$FA,$73,$B7,$FD,$94,$FC,$5F,$85,$FA,$1E,$C7,$D1,$F7,$7C
	.dc.b	$93,$F2,$78,$7E,$A7,$C2,$FA,$F9,$7D,$B3,$F3,$7E,$1B,$EC,$7B,$52
	.dc.b	$3E,$83,$5F,$83,$EA,$36,$4E,$6F,$ED,$37,$AB,$EF,$36,$FF,$05,$87
	.dc.b	$86,$7F,$05,$D3,$E4,$9F,$31,$F7,$FA,$3D,$4F,$E0,$76,$FC,$66,$FF
	.dc.b	$2E,$3F,$09,$F0,$9F,$2D,$FC,$37,$37,$CB,$7F,$11,$C9,$F2,$DF,$F2
	.dc.b	$56,$DF,$2D,$FC,$D1,$F9,$5F,$6F,$F3,$4F,$F6,$BE,$DF,$B6,$FE,$47
	.dc.b	$DC,$FC,$E7,$BD,$F9,$DE,$EF,$A7,$C6,$E1,$FE,$DD,$FC,$03,$FE,$76
	.dc.b	$C1,$CA,$8F,$F3,$BC,$7C,$AD,$3F,$36,$DF,$03,$F2,$57,$F9,$48,$FC
	.dc.b	$93,$5D,$18,$3F,$59,$1C,$1A,$66,$FC,$1E,$FF,$9A,$BC,$A7,$A6,$7F
	.dc.b	$59,$9F,$ED,$F4,$F7,$C7,$F8,$03,$5F,$86,$F9,$F6,$FC,$CB,$FB,$3E
	.dc.b	$FF,$CC,$BF,$EB,$7D,$7F,$D2,$3B,$7C,$38,$F8,$30,$F8,$7F,$A5,$1F
	.dc.b	$E5,$2D,$FE,$93,$F9,$BF,$07,$EE,$BF,$7C,$FE,$01,$DB,$F9,$A7,$E4
	.dc.b	$76,$FE,$1D,$F9,$5D,$5F,$8B,$7F,$1A,$7E,$60,$87,$CF,$EC,$FC,$97
	.dc.b	$1F,$E6,$5B,$8F,$E6,$A7,$7A,$7E,$69,$BC,$5F,$C5,$FB,$3B,$61,$FA
	.dc.b	$1C,$7C,$57,$FC,$67,$2F,$33,$FC,$6F,$3F,$D9,$E3,$FB,$53,$FC,$6D
	.dc.b	$A7,$F2,$5E,$9F,$CC,$47,$E0,$FA,$7F,$30,$FE,$17,$A7,$E3,$BF,$C0
	.dc.b	$33,$FD,$87,$7F,$F0,$2F,$43,$FA,$AC,$FC,$EF,$87,$FB,$CF,$D0,$F8
	.dc.b	$7B,$7F,$91,$F8,$4B,$CC,$F8,$7B,$F0,$F8,$F8,$7C,$F9,$5A,$F8,$79
	.dc.b	$9F,$DD,$F8,$3E,$24,$70,$43,$F8,$2E,$3F,$E6,$16,$C6,$DA,$6F,$97
	.dc.b	$F0,$DA,$7E,$96,$75,$18,$A5,$F8,$AF,$AE,$1E,$77,$F2,$19,$7B,$33
	.dc.b	$F6,$E9,$EE,$D7,$DF,$B7,$C0,$F4,$26,$1F,$97,$F8,$7F,$87,$EF,$FC
	.dc.b	$4F,$80,$F8,$BE,$D4,$AF,$F5,$1F,$C9,$BD,$AF,$DE,$FC,$17,$CA,$F7
	.dc.b	$3E,$5F,$C3,$78,$6F,$C6,$7E,$8F,$A3,$E5,$BF,$A5,$E4,$F5,$73,$7C
	.dc.b	$0F,$6F,$FC,$D7,$3F,$CB,$1F,$F3,$91,$3D,$DF,$B2,$D8,$FF,$5A,$CF
	.dc.b	$19,$C9,$23,$3E,$2F,$F9,$FE,$B8,$47,$5F,$85,$F0,$CF,$F1,$6C,$3C
	.dc.b	$B3,$F6,$31,$FC,$1B,$FB,$BE,$7F,$6B,$FB,$DD,$3F,$13,$AF,$93,$E0
	.dc.b	$FC,$BF,$11,$F8,$CE,$5F,$97,$7F,$C0,$F9,$3E,$53,$F8,$4E,$4F,$EB
	.dc.b	$3C,$9F,$2E,$3F,$30,$7F,$09,$E8,$FB,$2F,$A8,$FE,$53,$C2,$F7,$7E
	.dc.b	$1F,$C4,$7C,$4F,$7B,$E2,$FC,$07,$C6,$AF,$8F,$F0,$5F,$8A,$E5,$FE
	.dc.b	$67,$E1,$5C,$BF,$89,$B7,$C3,$FC,$84,$7E,$4E,$76,$D3,$65,$8D,$BF
	.dc.b	$5E,$BC,$0F,$9E,$16,$C3,$CD,$39,$86,$5F,$73,$3F,$7C,$FE,$A3,$4F
	.dc.b	$CB,$BF,$CA,$FB,$BF,$2E,$FE,$17,$E0,$BF,$07,$DD,$F8,$4F,$82,$FC
	.dc.b	$2F,$C3,$4A,$FC,$BC,$FE,$59,$FB,$07,$F3,$5F,$03,$F2,$CF,$ED,$BC
	.dc.b	$4F,$D5,$3F,$AC,$3F,$2C,$43,$F1,$07,$A7,$C0,$FC,$73,$F9,$76,$F8
	.dc.b	$3F,$93,$8D,$10,$D1,$5D,$3D,$0F,$E0,$FE,$07,$E5,$1F,$E8,$FE,$2F
	.dc.b	$92,$7F,$81,$5F,$A1,$FE,$09,$87,$EC,$6B,$FA,$87,$77,$E4,$DF,$9D
	.dc.b	$DB,$F9,$37,$FA,$A7,$87,$FB,$D7,$FA,$B7,$67,$E5,$3B,$3F,$25,$D1
	.dc.b	$F6,$3E,$1F,$F5,$FF,$0F,$FB,$0F,$87,$FD,$8F,$C3,$FE,$CB,$D0,$FD
	.dc.b	$9F,$C3,$FE,$D3,$DB,$FE,$5B,$FC,$03,$FE,$5E,$C8,$D3,$FC,$BF,$AF
	.dc.b	$F8,$09,$FF,$2F,$F8,$5F,$EF,$36,$C6,$7D,$5E,$64,$7F,$19,$1A,$AE
	.dc.b	$B3,$99,$FF,$19,$CE,$49,$F0,$9E,$48,$87,$E4,$70,$F4,$BF,$93,$CB
	.dc.b	$DA,$FE,$57,$4F,$82,$FE,$5F,$6F,$6F,$C1,$7B,$BD,$8F,$7F,$C1,$7C
	.dc.b	$0F,$7B,$E0,$FB,$1F,$0B,$E2,$26,$3F,$92,$3E,$07,$C4,$74,$7C,$FE
	.dc.b	$D7,$81,$FE,$41,$AF,$CA,$F6,$BE,$64,$7C,$03,$D2,$7B,$BE,$17,$E0
	.dc.b	$77,$FF,$48,$E3,$F8,$1C,$98,$E6,$F9,$8D,$BF,$98,$7C,$A7,$1C,$DF
	.dc.b	$2A,$3E,$6E,$1D,$B0,$E9,$1E,$23,$E6,$E3,$F8,$17,$F7,$5D,$1F,$80
	.dc.b	$D3,$D9,$5E,$3D,$7E,$3D,$7D,$8F,$86,$FC,$2F,$BD,$F8,$6F,$87,$FC
	.dc.b	$3F,$C0,$7C,$6A,$FB,$3C,$5F,$C9,$3C,$8F,$E7,$1D,$3F,$91,$9F,$A8
	.dc.b	$7E,$0B,$E0,$BF,$1F,$F0,$3F,$6A,$3F,$03,$F0,$9F,$92,$F8,$5F,$C9
	.dc.b	$F8,$12,$7F,$61,$CF,$FD,$0F,$CC,$FD,$BB,$63,$27,$F2,$E8,$CB,$4C
	.dc.b	$F3,$FF,$1C,$8E,$91,$F9,$0A,$F0,$3F,$96,$C3,$C9,$C7,$F5,$72,$FD
	.dc.b	$2E,$7E,$BA,$FE,$BD,$E0,$7F,$5C,$D7,$F5,$3F,$0D,$F8,$D9,$FC,$FF
	.dc.b	$7F,$E3,$5F,$D0,$F7,$7F,$86,$BF,$80,$7F,$CD,$9B,$9F,$F2,$74,$BC
	.dc.b	$6F,$DD,$F7,$BA,$BE,$AF,$1F,$EB,$FA,$FF,$21,$C3,$FD,$C7,$C0,$FC
	.dc.b	$86,$3F,$A5,$3F,$20,$43,$F2,$07,$E2,$48,$7E,$C3,$C0,$FC,$5E,$3F
	.dc.b	$92,$8C,$D7,$E5,$67,$27,$F0,$6F,$03,$F1,$D8,$FF,$01,$AF,$C7,$57
	.dc.b	$E6,$FC,$0F,$C7,$47,$F8,$0F,$C0,$FC,$73,$FD,$27,$C6,$FE,$1F,$AF
	.dc.b	$E3,$A3,$F8,$0F,$83,$F8,$E7,$F8,$3F,$81,$F8,$E7,$F8,$4F,$7F,$E3
	.dc.b	$F8,$7F,$8F,$F1,$7F,$A4,$B4,$FF,$18,$B7,$55,$FF,$A6,$23,$0D,$F8
	.dc.b	$8F,$A4,$47,$FB,$04,$6B,$93,$EA,$4E,$67,$FD,$8F,$39,$24,$66,$FC
	.dc.b	$2D,$BF,$0D,$F2,$BE,$B3,$97,$C4,$60,$87,$67,$51,$C6,$30,$6A,$BB
	.dc.b	$B6,$BB,$DF,$92,$30,$76,$38,$92,$78,$B6,$78,$DF,$EE,$63,$84,$E8
	.dc.b	$93,$A8,$EB,$7E,$59,$CC,$70,$9D,$3D,$4C,$12,$78,$46,$A7,$8C,$62
	.dc.b	$75,$7E,$B7,$94,$64,$91,$99,$E2,$DA,$E8,$4A,$E7,$7F,$32,$75,$18
	.dc.b	$3E,$8E,$1F,$8C,$F8,$6F,$DB,$7C,$47,$ED,$FE,$17,$F7,$1E,$BF,$EE
	.dc.b	$7E,$1B,$F7,$5F,$09,$FB,$BF,$4F,$F7,$9F,$0B,$FB,$DF,$57,$F4,$FA
	.dc.b	$7E,$AB,$7F,$E7,$6D,$3F,$9D,$9F,$B1,$31,$C3,$FC,$ED,$C1,$38,$7F
	.dc.b	$3B,$5D,$A3,$F2,$F8,$78,$9F,$CC,$E5,$E6,$7F,$37,$A7,$A9,$FC,$EE
	.dc.b	$DF,$C4,$7D,$6F,$E2,$5F,$01,$FC,$4F,$DB,$FE,$29,$F0,$7F,$C5,$7D
	.dc.b	$6F,$E2,$DF,$07,$FC,$5F,$D7,$99,$7F,$63,$E0,$FA,$3D,$1F,$2C,$91
	.dc.b	$F8,$4E,$8F,$9B,$3A,$8C,$4E,$9F,$B7,$3A,$24,$67,$8B,$ED,$B7,$8F
	.dc.b	$ED,$5F,$3B,$E3,$39,$7E,$D1,$F9,$EE,$8F,$E0,$B1,$99,$E1,$7D,$2B
	.dc.b	$C6,$43,$99,$5D,$38,$23,$82,$5F,$93,$0F,$4C,$F1,$8C,$5F,$86,$7F
	.dc.b	$35,$87,$F1,$8F,$67,$FA,$17,$C2,$FF,$43,$F6,$FF,$DB,$BF,$C0,$3F
	.dc.b	$EA,$AC,$FF,$D9,$6B,$D2,$CF,$C3,$6B,$DD,$DF,$C0,$F0,$F8,$2E,$3F
	.dc.b	$0D,$38,$7F,$CF,$58,$A3,$10,$CB,$AC,$61,$B1,$C2,$31,$C3,$FC,$6A
	.dc.b	$B9,$4E,$4E,$B7,$E2,$7B,$E3,$E6,$7C,$39,$FB,$3E,$1F,$DD,$1F,$CB
	.dc.b	$B6,$FE,$C5,$BF,$C0,$F8,$6F,$83,$F0,$DF,$0B,$E1,$FE,$1F,$C3,$FC
	.dc.b	$4F,$7B,$E2,$FC,$3F,$C6,$F8,$79,$9F,$E4,$FA,$3E,$8B,$F5,$B8,$BE
	.dc.b	$8B,$F4,$CF,$C2,$F2,$7D,$86,$C6,$9E,$06,$09,$D5,$F8,$88,$D4,$F1
	.dc.b	$8C,$4E,$AB,$C3,$3F,$CB,$3A,$7E,$F4,$FB,$CF,$E9,$BD,$3F,$0D,$F8
	.dc.b	$A7,$D6,$E9,$FB,$8F,$D8,$E9,$FB,$8F,$CF,$3E,$C6,$5F,$4F,$A7,$EA
	.dc.b	$C3,$37,$4F,$D6,$8D,$99,$71,$9F,$DE,$8C,$CF,$0A,$E6,$D8,$F0,$4E
	.dc.b	$CF,$C9,$F4,$FD,$F8,$E3,$6E,$9F,$C0,$46,$3A,$7D,$CF,$F6,$8D,$7F
	.dc.b	$3E,$FF,$6E,$E7,$FC,$03,$F9,$6E,$6F,$C0,$3F,$98,$E5,$FC,$03,$F9
	.dc.b	$AE,$4F,$BE,$F8,$39,$3F,$AB,$6D,$F0,$3D,$AF,$D4,$FA,$12,$BF,$43
	.dc.b	$ED,$FE,$B3,$D2,$FD,$6F,$AF,$FA,$EF,$87,$FD,$7F,$C3,$FF,$81,$AF
	.dc.b	$E1,$F9,$9F,$EA,$13,$FD,$3F,$99,$A4,$F8,$1E,$1C,$EE,$9F,$EF,$F3
	.dc.b	$9B,$FF,$75,$35,$EB,$5C,$27,$6D,$8F,$B5,$3D,$9F,$29,$FB,$AC,$7D
	.dc.b	$2F,$EF,$33,$FE,$0F,$A7,$BE,$3E,$9E,$BF,$C2,$B6,$E9,$A7,$A7,$D9
	.dc.b	$B9,$3E,$1E,$FC,$3E,$3D,$FC,$FC,$3E,$BE,$03,$7E,$CC,$D7,$FA,$06
	.dc.b	$3F,$E5,$37,$DB,$F0,$1D,$3D,$EF,$9D,$1E,$53,$E0,$FB,$FE,$A7,$F0
	.dc.b	$FC,$FE,$2F,$7B,$C7,$CD,$27,$E3,$B5,$F8,$E6,$08,$60,$9D,$1F,$45
	.dc.b	$F9,$31,$C6,$31,$C5,$F6,$6B,$C5,$1F,$CB,$2F,$E1,$E1,$E8,$C3,$F3
	.dc.b	$7C,$DE,$B7,$DA,$7D,$AF,$7F,$E4,$F2,$7C,$AE,$4F,$97,$F0,$FF,$26
	.dc.b	$BE,$FF,$C0,$7C,$97,$F0,$3C,$5F,$25,$FF,$3E,$79,$9F,$CA,$FA,$3E
	.dc.b	$C6,$DF,$CF,$BA,$7E,$DD,$7E,$23,$E1,$FE,$FB,$FA,$EE,$CF,$B3,$F0
	.dc.b	$9F,$6A,$3F,$6F,$E4,$7E,$F3,$BF,$F8,$24,$7E,$E1,$AF,$ED,$BC,$AD
	.dc.b	$91,$D1,$F9,$03,$A9,$F8,$F3,$FA,$4F,$83,$F8,$8C,$7C,$B1,$F8,$E3
	.dc.b	$FA,$6E,$5F,$8F,$7F,$A7,$E9,$F8,$EF,$85,$FC,$6C,$7E,$8B,$E1,$BF
	.dc.b	$1A,$FE,$93,$E1,$FF,$27,$F0,$F2,$3F,$1B,$CF,$F9,$6E,$FF,$A5,$1F
	.dc.b	$50,$FC,$61,$FB,$6E,$AF,$B7,$BF,$F5,$5E,$BF,$C6,$4F,$E1,$B9,$3F
	.dc.b	$03,$E0,$7E,$C5,$BD,$FF,$85,$BF,$D8,$D6,$DA,$FE,$31,$B3,$C2,$3F
	.dc.b	$77,$DD,$E0,$7F,$96,$61,$F8,$A7,$CB,$3F,$DE,$3B,$BD,$2F,$8B,$3F
	.dc.b	$D8,$E9,$F8,$AA,$FE,$0B,$DB,$F8,$A7,$F8,$3F,$87,$FB,$6E,$DF,$C5
	.dc.b	$47,$F1,$4E,$DF,$C5,$7A,$2E,$EF,$F4,$6F,$F0,$0F,$FB,$28,$FF,$4C
	.dc.b	$13,$E0,$3F,$80,$A6,$FC,$3E,$3E,$0B,$3F,$0F,$AF,$84,$DF,$C3,$CD
	.dc.b	$3F,$D5,$3C,$5F,$AA,$C9,$39,$9F,$E9,$31,$87,$7C,$DE,$49,$D5,$71
	.dc.b	$6C,$66,$76,$7F,$59,$AE,$4F,$4B,$F2,$BA,$F2,$12,$CB,$D7,$39,$8F
	.dc.b	$73,$74,$F8,$3E,$F4,$8D,$63,$F1,$3A,$BE,$2F,$C3,$FC,$6F,$87,$F8
	.dc.b	$FF,$0F,$F2,$3E,$1F,$E4,$FC,$3F,$CA,$F8,$79,$97,$F9,$87,$5F,$A1
	.dc.b	$F0,$FF,$3B,$E1,$FE,$7E,$F2,$BC,$9F,$0F,$F3,$DF,$37,$B9,$F3,$DF
	.dc.b	$47,$C3,$7D,$62,$38,$21,$F3,$9C,$B4,$CF,$3F,$D5,$8C,$BF,$5F,$9B
	.dc.b	$F0,$35,$E1,$AF,$AF,$7E,$57,$EC,$E1,$E9,$7E,$56,$5F,$88,$E3,$FC
	.dc.b	$4F,$BD,$F1,$A7,$E9,$7F,$00,$FF,$95,$67,$F2,$3C,$5F,$1E,$7E,$97
	.dc.b	$0F,$C8,$69,$FE,$87,$AF,$F2,$D3,$F5,$F9,$BE,$A3,$FE,$12,$87,$C2
	.dc.b	$7D,$C8,$FC,$6F,$67,$F5,$A9,$FF,$07,$47,$DF,$EC,$C7,$C2,$7E,$32
	.dc.b	$3F,$15,$DB,$7B,$7F,$0E,$DE,$7F,$C6,$F2,$FE,$4E,$BF,$11,$39,$38
	.dc.b	$62,$C3,$F2,$33,$E5,$F8,$3F,$25,$FE,$8F,$3F,$26,$9E,$A8,$FD,$36
	.dc.b	$BE,$7F,$83,$FC,$76,$1F,$23,$E1,$3F,$1C,$FE,$5F,$B7,$D9,$EB,$7B
	.dc.b	$7E,$1F,$F1,$B3,$FA,$CE,$EF,$F0,$EF,$0F,$E3,$A3,$F3,$DD,$DF,$67
	.dc.b	$4F,$DA,$75,$7E,$FB,$AF,$EF,$47,$ED,$FB,$3F,$1D,$7F,$87,$24,$7E
	.dc.b	$1D,$FC,$4F,$57,$F0,$AE,$EF,$C4,$B6,$32,$7E,$D6,$32,$D3,$F6,$1D
	.dc.b	$DF,$8E,$7F,$65,$E1,$78,$DF,$D9,$10,$FE,$0F,$9F,$F1,$4E,$FF,$C7
	.dc.b	$47,$EE,$7B,$FF,$1C,$FE,$0F,$E2,$3F,$66,$FE,$F7,$C1,$FC,$63,$FC
	.dc.b	$43,$C0,$BE,$2F,$F2,$38,$FF,$31,$77,$3E,$1B,$F1,$1F,$11,$F8,$9F
	.dc.b	$87,$FC,$57,$91,$F8,$BF,$87,$FC,$67,$C3,$FE,$37,$DA,$FC,$77,$B7
	.dc.b	$F8,$FF,$87,$FC,$87,$A9,$35,$FD,$0B,$5F,$CC,$37,$A7,$FC,$1F,$B2
	.dc.b	$31,$D2,$AF,$2C,$61,$A7,$EB,$2F,$C2,$FE,$BB,$1F,$2B,$FB,$0C,$FD
	.dc.b	$2F,$EC,$B5,$F6,$BF,$B4,$DF,$F4,$DE,$F7,$E9,$FE,$07,$F5,$1E,$BC
	.dc.b	$AF,$E5,$5F,$C0,$3F,$E5,$F7,$FD,$02,$FF,$97,$89,$ED,$7F,$9B,$37
	.dc.b	$F9,$53,$C3,$E1,$3E,$53,$FC,$0B,$93,$FC,$9B,$7F,$E4,$B8,$FE,$0B
	.dc.b	$C9,$8F,$88,$F5,$C7,$6F,$63,$7F,$11,$23,$C6,$D7,$C6,$DE,$6F,$F1
	.dc.b	$CE,$08,$C7,$37,$CB,$F4,$D3,$9B,$E6,$43,$C0,$7F,$18,$E8,$F2,$3E
	.dc.b	$63,$E1,$FB,$5F,$31,$FE,$25,$CB,$EC,$DF,$F8,$87,$2F,$CC,$9F,$E4
	.dc.b	$3C,$9F,$35,$F9,$7F,$11,$F0,$BD,$AF,$87,$CB,$27,$E7,$BF,$2F,$B3
	.dc.b	$EB,$C7,$D0,$F7,$3E,$4F,$BB,$F9,$38,$FA,$BF,$11,$F9,$37,$EB,$FC
	.dc.b	$07,$E4,$DF,$C8,$90,$FC,$89,$79,$7F,$23,$E0,$7E,$37,$91,$D7,$F9
	.dc.b	$3A,$D7,$37,$C9,$8C,$47,$22,$5F,$E6,$BC,$0F,$BB,$8F,$8E,$7F,$0F
	.dc.b	$97,$9D,$FC,$4E,$7F,$82,$D3,$DB,$87,$E8,$B5,$FC,$2F,$C0,$7E,$1B
	.dc.b	$E2,$3F,$25,$87,$F3,$FE,$FF,$C4,$BF,$8A,$F8,$1F,$C5,$FC,$34,$8F
	.dc.b	$E9,$1C,$FF,$4D,$FD,$01,$F9,$13,$EC,$9F,$6C,$FC,$DF,$7F,$E4,$33
	.dc.b	$FC,$4F,$7F,$E2,$27,$FD,$CD,$DD,$F8,$77,$F1,$E7,$E1,$E7,$F1,$FD
	.dc.b	$FB,$E6,$FC,$6B,$6B,$84,$EA,$B8,$73,$7E,$4E,$58,$78,$63,$18,$FE
	.dc.b	$EA,$BF,$5F,$DD,$F8,$FC,$3F,$83,$F7,$7B,$67,$FB,$16,$9F,$8F,$7F
	.dc.b	$B2,$76,$FE,$3D,$FD,$BF,$6F,$E3,$DF,$C6,$90,$FD,$DF,$67,$FB,$93
	.dc.b	$5F,$3E,$9E,$5D,$BD,$44,$AF,$49,$F1,$1C,$3F,$10,$99,$FB,$DB,$3C
	.dc.b	$DE,$3B,$FC,$66,$DF,$81,$E8,$62,$FA,$DF,$32,$5F,$39,$EA,$70,$2D
	.dc.b	$F5,$6D,$F5,$8C,$57,$29,$E0,$F9,$FC,$5F,$E8,$93,$AB,$ED,$3B,$8C
	.dc.b	$57,$09,$C9,$1C,$12,$36,$4B,$F1,$13,$0F,$6D,$71,$9D,$46,$2B,$95
	.dc.b	$6D,$F8,$84,$3D,$B1,$B9,$E3,$19,$BF,$67,$85,$F8,$FF,$88,$FE,$63
	.dc.b	$F0,$FF,$CC,$BE,$1F,$F9,$9F,$C3,$FF,$34,$F8,$7F,$F8,$03,$DB,$FE
	.dc.b	$E5,$C9,$97,$FB,$92,$1E,$F6,$BD,$7D,$FC,$07,$F0,$9F,$17,$F8,$2D
	.dc.b	$FF,$0C,$C3,$F8,$49,$FD,$60,$FE,$0F,$C5,$FC,$2F,$C6,$E0,$48,$FF
	.dc.b	$3C,$64,$C3,$B4,$63,$2F,$2D,$F9,$E3,$FA,$AE,$9E,$B7,$FA,$BE,$DF
	.dc.b	$D3,$B7,$F5,$FA,$9E,$CF,$84,$F6,$FA,$DE,$EF,$84,$F7,$FC,$3F,$C0
	.dc.b	$F8,$7F,$83,$F0,$33,$3F,$EE,$5D,$7F,$84,$EB,$FA,$51,$F4,$0F,$9D
	.dc.b	$BF,$D9,$8C,$F2,$7E,$1D,$B1,$9E,$BF,$AD,$18,$AF,$BB,$18,$23,$D0
	.dc.b	$C9,$7A,$FE,$EB,$E1,$3F,$61,$87,$96,$78,$C6,$70,$F1,$46,$1B,$1C
	.dc.b	$EF,$C5,$7D,$27,$33,$C7,$0F,$43,$BB,$E5,$18,$8D,$1E,$61,$87,$EA
	.dc.b	$B9,$5F,$0F,$F8,$CF,$5F,$F1,$BE,$BF,$E3,$BE,$03,$F1,$FE,$DF,$E4
	.dc.b	$3D,$CF,$C8,$FB,$FF,$92,$F7,$FF,$27,$EF,$CD,$7F,$0A,$D7,$F2,$BB
	.dc.b	$7F,$7D,$D5,$39,$E0,$FC,$C3,$67,$4D,$A7,$0A,$F8,$C7,$63,$F2,$78
	.dc.b	$79,$1F,$CA,$E5,$E8,$7F,$2F,$A7,$B1,$FC,$CE,$DF,$AA,$DF,$F5,$7E
	.dc.b	$FF,$F8,$EE,$7C,$E7,$83,$F8,$07,$FD,$48,$DF,$85,$D6,$ED,$FF,$C3
	.dc.b	$15,$FE,$9E,$AF,$F6,$E3,$FE,$9C,$79,$EC,$8F,$F4,$94,$6B,$E1,$79
	.dc.b	$D7,$2F,$85,$B3,$E1,$7F,$A0,$73,$7D,$C8,$FB,$58,$FE,$08,$95,$F7
	.dc.b	$DB,$1A,$8C,$11,$C1,$31,$FC,$64,$65,$B1,$9E,$7F,$C8,$92,$FE,$37
	.dc.b	$3F,$E2,$76,$FC,$45,$7F,$04,$E7,$FC,$43,$FC,$1B,$9F,$F1,$0F,$F0
	.dc.b	$8E,$7F,$C4,$3F,$64,$F8,$BC,$DF,$71,$BC,$FF,$81,$D3,$F0,$0D,$9D
	.dc.b	$6D,$F8,$8A,$E4,$D7,$89,$8D,$BF,$21,$CF,$F8,$58,$EF,$CF,$F8,$57
	.dc.b	$C3,$CF,$F8,$57,$C7,$CF,$F8,$57,$F8,$FF,$3F,$D4,$D3,$ED,$7C,$1D
	.dc.b	$E7,$FF,$70,$D7,$DE,$7F,$04,$DF,$87,$D7,$C4,$6F,$A7,$87,$4C,$AF
	.dc.b	$E3,$3E,$67,$E2,$BE,$1F,$A7,$A3,$D7,$DB,$ED,$E9,$F7,$AF,$07,$C2
	.dc.b	$F8,$7E,$13,$C5,$F0,$DF,$17,$AA,$67,$F0,$4F,$D4,$F5,$7C,$8E,$9F
	.dc.b	$D6,$C6,$79,$3E,$6B,$67,$4F,$0A,$EA,$FE,$73,$0F,$0B,$F9,$EC,$BC
	.dc.b	$AF,$E8,$34,$F4,$BF,$A2,$DB,$E3,$7C,$2F,$C7,$F8,$3F,$91,$E8,$FC
	.dc.b	$9F,$86,$F9,$5E,$CF,$CB,$F7,$FE,$77,$F0,$0F,$FA,$6F,$C1,$FD,$76
	.dc.b	$7F,$C0,$A1,$ED,$68,$FC,$9E,$DF,$E8,$98,$7C,$3F,$8A,$BF,$63,$1F
	.dc.b	$E9,$B6,$FB,$DF,$2D,$FD,$BF,$3F,$CB,$7F,$73,$CF,$F2,$9F,$C5,$93
	.dc.b	$2F,$C7,$13,$9F,$E5,$C6,$5B,$19,$8F,$94,$7C,$FE,$E5,$77,$8F,$C4
	.dc.b	$5F,$8D,$F2,$9F,$EC,$DA,$F8,$3D,$0C,$BD,$51,$8C,$3F,$D7,$75,$EE
	.dc.b	$CF,$E7,$BC,$72,$F8,$9E,$FF,$D0,$8C,$1F,$2F,$CB,$F9,$FE,$F7,$D0
	.dc.b	$F8,$7F,$A3,$3F,$4B,$B2,$3F,$96,$3F,$92,$74,$FD,$6F,$85,$FC,$B4
	.dc.b	$7F,$20,$EE,$FB,$3F,$0D,$F6,$BE,$1F,$ED,$FC,$3F,$DC,$F8,$7F,$E3
	.dc.b	$5C,$9F,$C3,$89,$E9,$FF,$0D,$C7,$FB,$57,$2A,$B8,$C7,$A2,$BE,$16
	.dc.b	$1D,$A3,$C0,$7E,$44,$D5,$33,$FD,$B6,$9F,$8C,$F8,$7F,$C9,$56,$BE
	.dc.b	$1F,$F2,$4F,$F3,$8E,$EF,$C9,$3C,$BD,$EF,$C7,$3D,$3B,$BF,$1E,$FE
	.dc.b	$4F,$A3,$F1,$DD,$5F,$5B,$4F,$AB,$D3,$F9,$3E,$DF,$B9,$3F,$5F,$B3
	.dc.b	$F0,$0F,$D9,$ED,$FC,$73,$F8,$7E,$2F,$D9,$E5,$F8,$B3,$F1,$25,$73
	.dc.b	$1B,$E8,$FD,$BC,$63,$1F,$C8,$9F,$D5,$2F,$BB,$FD,$5B,$0F,$1B,$E5
	.dc.b	$C3,$E3,$FC,$3F,$AA,$3F,$55,$A7,$E2,$A7,$F5,$3D,$BF,$89,$7F,$57
	.dc.b	$DB,$F8,$87,$E7,$F6,$7E,$21,$FA,$3A,$7E,$21,$FF,$07,$F7,$F9,$1F
	.dc.b	$5E,$DE,$9F,$1F,$E0,$E7,$9D,$FD,$55,$A8,$C6,$1E,$E2,$57,$09,$D1
	.dc.b	$2B,$89,$27,$71,$FC,$76,$78,$4E,$5B,$3A,$9F,$9C,$DB,$F3,$47,$F5
	.dc.b	$7F,$5F,$FC,$51,$5C,$09,$1D,$48,$72,$72,$D3,$33,$D1,$C1,$1C,$13
	.dc.b	$0F,$83,$3C,$A3,$85,$7F,$A8,$B8,$9F,$01,$FC,$A3,$E1,$FF,$94,$FC
	.dc.b	$17,$F2,$AF,$87,$FE,$B1,$FC,$03,$FE,$87,$C7,$F8,$5B,$FE,$86,$D3
	.dc.b	$F5,$8B,$E2,$37,$F0,$FF,$B6,$F0,$FF,$69,$5F,$E4,$AE,$1F,$DB,$F2
	.dc.b	$7F,$BA,$F0,$FF,$41,$4F,$2C,$3B,$FC,$3F,$83,$8B,$C5,$3F,$B4,$D7
	.dc.b	$CC,$FE,$DB,$7F,$2F,$C3,$F9,$BE,$1F,$CF,$F0,$FE,$8F,$87,$F4,$FC
	.dc.b	$3F,$AB,$E1,$FD,$7F,$0F,$EC,$F8,$69,$9F,$F1,$5E,$2F,$B8,$FC,$ED
	.dc.b	$BE,$4B,$4F,$C3,$75,$FD,$D8,$CC,$7D,$16,$F3,$7D,$18,$DF,$5F,$D1
	.dc.b	$9C,$92,$32,$75,$BF,$0B,$CC,$FD,$CF,$5F,$E2,$1D,$61,$E1,$78,$9B
	.dc.b	$9E,$38,$78,$5B,$7D,$DB,$3C,$27,$55,$CE,$72,$47,$04,$C3,$F2,$78
	.dc.b	$7F,$BA,$B9,$5F,$01,$F8,$6F,$86,$FC,$3F,$C1,$7E,$23,$E1,$BF,$13
	.dc.b	$F0,$9F,$8A,$F8,$6F,$C5,$FC,$37,$E3,$33,$FC,$6F,$C3,$4E,$0F,$C5
	.dc.b	$B5,$FF,$30,$E8,$8C,$70,$7E,$49,$BC,$2A,$FF,$3F,$D6,$DE,$64,$39
	.dc.b	$CF,$85,$FC,$CE,$1E,$57,$F3,$79,$7A,$5F,$CE,$E9,$ED,$7F,$3F,$B7
	.dc.b	$F8,$EF,$DF,$FF,$3E,$F1,$1F,$F1,$DE,$9F,$DC,$1F,$F1,$DB,$7E,$17
	.dc.b	$87,$C2,$F1,$F8,$5E,$5F,$0B,$CF,$D6,$99,$FE,$4F,$9F,$E8,$3F,$80
	.dc.b	$E3,$FC,$53,$F6,$78,$FE,$E4,$EA,$3F,$04,$DC,$FE,$ED,$70,$D7,$EE
	.dc.b	$36,$75,$7F,$E1,$2E,$7F,$B9,$BF,$BF,$1F,$A3,$C1,$F0,$DF,$A7,$C3
	.dc.b	$F1,$DF,$AB,$C5,$F2,$DF,$9A,$7C,$FD,$3E,$B3,$71,$FA,$DA,$39,$3E
	.dc.b	$B4,$F0,$25,$70,$6D,$70,$9D,$CE,$49,$19,$D7,$F1,$B5,$FC,$53,$97
	.dc.b	$F2,$2F,$F1,$6E,$6F,$CA,$BF,$C6,$39,$FF,$8E,$ED,$F6,$FE,$1B,$EE
	.dc.b	$7C,$3F,$DD,$F8,$AF,$BD,$D1,$F6,$30,$FF,$05,$74,$BA,$3E,$CC,$63
	.dc.b	$9F,$EC,$BF,$D6,$0F,$F6,$8C,$FF,$8B,$B8,$3F,$08,$FE,$27,$DF,$FC
	.dc.b	$57,$9D,$F8,$BF,$4F,$F1,$9F,$0F,$F8,$DF,$57,$F1,$DF,$0F,$F8,$FF
	.dc.b	$87,$99,$F8,$9A,$FF,$B1,$B2,$46,$33,$FC,$93,$79,$FE,$8C,$EA,$B9
	.dc.b	$10,$E5,$3E,$07,$DF,$87,$91,$F8,$39,$7A,$1F,$87,$A7,$E8,$EF,$F4
	.dc.b	$9F,$11,$FA,$5F,$88,$FD,$37,$C4,$7E,$9F,$DE,$FD,$47,$C4,$7F,$A8
	.dc.b	$FF,$80,$7F,$CE,$DB,$7E,$11,$FF,$3A,$E7,$FE,$37,$7F,$CE,$6D,$F7
	.dc.b	$78,$1F,$E2,$AF,$03,$F6,$19,$F3,$F7,$BA,$7B,$BE,$5B,$F8,$9C,$1D
	.dc.b	$FD,$7F,$07,$C4,$78,$7D,$EF,$17,$95,$2F,$C0,$D7,$FD,$E3,$AA,$31
	.dc.b	$CF,$FE,$6C,$8C,$4F,$02,$1C,$27,$7B,$FB,$5F,$04,$7B,$70,$F2,$3F
	.dc.b	$30,$FC,$6F,$57,$A9,$F8,$5A,$FC,$C7,$F2,$3C,$9F,$35,$F8,$DF,$09
	.dc.b	$F3,$1F,$91,$F1,$1F,$37,$E0,$BE,$77,$C3,$7C,$FF,$88,$95,$FD,$7F
	.dc.b	$6F,$F3,$33,$F9,$CE,$AF,$CA,$3F,$47,$E1,$FE,$B7,$C3,$FE,$06,$3F
	.dc.b	$1D,$DD,$F8,$57,$F0,$E7,$E2,$78,$FF,$1E,$DE,$3F,$C7,$F7,$FE,$46
	.dc.b	$3F,$46,$DE,$AF,$E5,$0D,$CB,$F2,$93,$D0,$FC,$E7,$81,$FA,$37,$F6
	.dc.b	$3E,$0F,$91,$F3,$1F,$92,$3D,$47,$E2,$7E,$23,$F1,$DF,$11,$F9,$28
	.dc.b	$FD,$CF,$83,$F9,$1F,$7F,$F2,$51,$0F,$C8,$7C,$47,$E3,$9F,$C8,$1F
	.dc.b	$E0,$5C,$BF,$75,$AF,$E5,$FA,$7E,$B4,$FE,$C7,$AB,$F7,$4F,$E6,$7B
	.dc.b	$BF,$1E,$FE,$6F,$BB,$F0,$6B,$85,$CD,$F8,$A8,$FC,$41,$F8,$AE,$3F
	.dc.b	$C7,$77,$70,$AD,$B6,$B8,$4E,$AB,$86,$7F,$8F,$7C,$07,$E3,$CF,$21
	.dc.b	$F8,$F3,$D0,$7E,$BF,$C0,$FD,$53,$FC,$7B,$C1,$FC,$73,$FC,$37,$BF
	.dc.b	$F7,$1D,$FF,$8F,$8F,$D2,$F7,$FE,$3D,$FD,$C7,$77,$E3,$9F,$D4,$F7
	.dc.b	$7E,$DF,$47,$C3,$FF,$01,$F8,$2F,$E0,$5F,$0F,$FD,$E6,$DB,$FF,$B6
	.dc.b	$3F,$80,$7F,$CA,$1C,$1F,$D2,$1F,$F2,$84,$EA,$31,$F0,$DC,$6B,$87
	.dc.b	$C4,$59,$E1,$F0,$D6,$FE,$86,$0F,$86,$C7,$32,$31,$3F,$EC,$76,$BD
	.dc.b	$CE,$33,$C2,$33,$8F,$9E,$30,$49,$E4,$D7,$95,$77,$8C,$4F,$BE,$FA
	.dc.b	$63,$D7,$2E,$DF,$0F,$DF,$3F,$8F,$97,$F2,$5D,$7F,$95,$6D,$E3,$F8
	.dc.b	$7F,$27,$C3,$F9,$7E,$1F,$CD,$F0,$FE,$7F,$87,$98,$FB,$7E,$1F,$D5
	.dc.b	$ED,$C8,$FF,$31,$72,$3D,$CF,$AD,$1F,$CC,$F9,$BC,$6F,$C5,$E0,$F3
	.dc.b	$3F,$1F,$E0,$FE,$27,$27,$B3,$8B,$EC,$73,$5A,$F9,$27,$C4,$21,$F1
	.dc.b	$32,$F9,$73,$C2,$33,$CD,$F3,$5B,$19,$78,$5F,$D3,$C3,$C7,$1C,$F1
	.dc.b	$F3,$BF,$CE,$B2,$F5,$BD,$F3,$FB,$5E,$F7,$C8,$8F,$C0,$7C,$1F,$DD
	.dc.b	$F8,$0F,$BD,$F0,$7F,$7F,$E0,$BF,$01,$C5,$F8,$1F,$87,$98,$7F,$2B
	.dc.b	$E9,$FC,$5F,$F0,$0F,$F9,$10,$FF,$21,$F0,$FE,$64,$CF,$C1,$6B,$E0
	.dc.b	$BE,$FD,$FF,$96,$79,$15,$FE,$26,$B7,$C4,$74,$F8,$39,$7B,$AD,$46
	.dc.b	$23,$F1,$B5,$FE,$15,$74,$7F,$0F,$70,$EA,$38,$5F,$82,$35,$18,$AE
	.dc.b	$B1,$CB,$C2,$5F,$2B,$E1,$18,$C3,$D7,$5E,$2C,$7D,$EF,$EA,$F0,$F8
	.dc.b	$BF,$09,$F1,$BE,$17,$F3,$33,$FA,$FF,$07,$E4,$FC,$47,$E6,$63,$F6
	.dc.b	$5E,$07,$E6,$1F,$C3,$78,$1F,$3B,$E1,$BF,$31,$1F,$89,$F0,$7F,$C0
	.dc.b	$9E,$07,$E6,$63,$F4,$C7,$E6,$4F,$CC,$10,$FC,$87,$8B,$F8,$97,$F8
	.dc.b	$07,$83,$F9,$7E,$1F,$DB,$47,$E4,$0F,$CC,$9F,$97,$F0,$B5,$7F,$98
	.dc.b	$8C,$F1,$FE,$5D,$D1,$C6,$B9,$BF,$98,$3F,$50,$7F,$0F,$C3,$F2,$F3
	.dc.b	$FC,$87,$C1,$FD,$FC,$FF,$84,$BC,$5F,$E2,$1E,$0F,$E6,$23,$F8,$B7
	.dc.b	$8B,$FC,$4F,$C1,$FC,$C4,$7F,$19,$F0,$3F,$30,$FE,$FC,$FE,$39,$DD
	.dc.b	$F9,$87,$FC,$AD,$A7,$DC,$F3,$7F,$33,$EA,$7E,$6B,$E1,$BF,$37,$EA
	.dc.b	$7E,$73,$E0,$FF,$3B,$EE,$7E,$7B,$DF,$FC,$FF,$BD,$FA,$0F,$7E,$69
	.dc.b	$F8,$7F,$13,$FB,$DF,$2F,$E1,$9B,$E1,$7E,$76,$B7,$5F,$98,$6D,$6D
	.dc.b	$B3,$B9,$E9,$1E,$07,$B6,$1E,$47,$C1,$97,$A1,$F1,$69,$EC,$7C,$9B
	.dc.b	$7E,$DB,$D5,$FD,$BF,$C3,$7E,$E3,$DE,$FD,$CF,$C3,$7F,$44,$F0,$BF
	.dc.b	$CA,$B8,$7F,$77,$7F,$CA,$AF,$F3,$A7,$FC,$AA,$DF,$86,$E0,$7F,$95
	.dc.b	$61,$EF,$78,$67,$F6,$B9,$F9,$5E,$5E,$DF,$A5,$FD,$CE,$DE,$1F,$7B
	.dc.b	$E0,$C7,$EE,$78,$2F,$47,$EC,$9B,$C9,$F3,$27,$47,$C6,$21,$F3,$4B
	.dc.b	$C9,$F3,$5F,$DE,$68,$C3,$EE,$47,$D9,$E6,$F8,$59,$F9,$EF,$ED,$69
	.dc.b	$EB,$7F,$80,$E1,$EE,$BF,$C0,$71,$7C,$9F,$86,$F9,$5E,$FF,$CC,$9E
	.dc.b	$E7,$CD,$F7,$FE,$77,$C0,$7C,$88,$F1,$F1,$FF,$9F,$77,$FF,$17,$C7
	.dc.b	$D1,$F8,$0F,$A9,$F0,$1F,$57,$E0,$3F,$29,$3E,$AF,$80,$FB,$1F,$07
	.dc.b	$F8,$58,$FD,$89,$38,$3F,$68,$4E,$EF,$E9,$4D,$8F,$DC,$C6,$08,$E0
	.dc.b	$9C,$FF,$BA,$9F,$CE,$77,$F8,$27,$C4,$7E,$48,$F9,$B8,$FA,$1F,$9F
	.dc.b	$9F,$F4,$6E,$EF,$C7,$7C,$2F,$E4,$A7,$F5,$1D,$DF,$93,$7E,$AF,$C2
	.dc.b	$FE,$4F,$E1,$BF,$29,$F0,$DF,$95,$F8,$69,$43,$F7,$7D,$5F,$4D,$FE
	.dc.b	$87,$E3,$7E,$ED,$FD,$A7,$8D,$F7,$1F,$BC,$7E,$A8,$FC,$1F,$81,$F8
	.dc.b	$CF,$06,$47,$E2,$39,$3F,$64,$FE,$2C,$9C,$7F,$8B,$6C,$6A,$30,$7E
	.dc.b	$1A,$BF,$1B,$3F,$D5,$BC,$1F,$13,$FD,$63,$0F,$33,$E8,$3F,$2B,$E1
	.dc.b	$FE,$DD,$FC,$BF,$7F,$E3,$1F,$E1,$DE,$27,$ED,$DF,$C6,$1F,$9C,$EE
	.dc.b	$FC,$63,$FB,$53,$E3,$7C,$43,$0F,$36,$92,$7F,$52,$7A,$8B,$A7,$AA
	.dc.b	$35,$D5,$FB,$19,$F8,$86,$A3,$3D,$DF,$B3,$9D,$4F,$36,$D7,$38,$E0
	.dc.b	$48,$CD,$72,$DF,$E8,$CF,$1B,$E8,$48,$CD,$F3,$BE,$35,$CE,$B4,$F9
	.dc.b	$A3,$27,$93,$1F,$85,$1B,$BE,$78,$78,$63,$85,$F2,$7C,$53,$A7,$81
	.dc.b	$A7,$17,$DA,$78,$61,$E5,$6D,$F5,$BE,$67,$12,$18,$26,$1E,$6B,$70
	.dc.b	$FF,$A2,$A7,$F8,$CF,$9B,$FC,$DF,$E0,$FF,$CF,$7F,$C0,$3F,$E8,$89
	.dc.b	$FF,$57,$3F,$E8,$8C,$FF,$C2,$EB,$D1,$DF,$C3,$FF,$2F,$AF,$EE,$DE
	.dc.b	$9F,$F7,$BC,$FF,$7D,$AF,$F0,$26,$CF,$F0,$29,$D1,$3D,$4F,$F0,$19
	.dc.b	$FE,$09,$D7,$F9,$64,$FF,$8A,$72,$63,$E6,$CB,$CF,$EA,$7A,$33,$F5
	.dc.b	$5F,$B0,$FE,$6B,$F0,$3E,$CF,$5B,$DB,$9F,$BB,$DB,$F7,$FA,$DF,$03
	.dc.b	$E1,$FE,$0E,$DF,$0B,$DB,$F8,$7F,$07,$33,$FD,$0F,$4F,$D1,$E4,$FA
	.dc.b	$9D,$3F,$63,$A7,$EA,$D7,$DC,$3E,$94,$E9,$FA,$AD,$73,$C1,$CC,$8F
	.dc.b	$18,$D1,$23,$3C,$DF,$85,$BF,$01,$FC,$FC,$B5,$DA,$77,$7E,$4A,$E1
	.dc.b	$39,$3B,$4E,$AB,$8C,$79,$0C,$92,$78,$B9,$6D,$F9,$63,$F8,$0F,$4F
	.dc.b	$E3,$F9,$9E,$DF,$E3,$3D,$BF,$C6,$FC,$3F,$E3,$BE,$1F,$F1,$FE,$EF
	.dc.b	$E4,$3E,$1F,$F2,$3E,$F7,$E4,$BE,$1E,$69,$F8,$16,$BF,$EC,$2C,$D1
	.dc.b	$8D,$7F,$2A,$DE,$0F,$E7,$9B,$AB,$77,$CC,$87,$3A,$E0,$FE,$83,$0F
	.dc.b	$23,$F9,$1C,$BD,$0F,$E4,$F4,$F6,$3F,$95,$DB,$F5,$3E,$FF,$FB,$93
	.dc.b	$F8,$05,$F0,$77,$E1,$31,$F1,$19,$F8,$6D,$7C,$3E,$FE,$23,$87,$C4
	.dc.b	$71,$F8,$89,$AF,$E4,$0F,$A5,$D7,$F6,$C9,$D7,$F6,$39,$FE,$97,$37
	.dc.b	$D4,$E7,$FB,$97,$F7,$E3,$0F,$2A,$E1,$3A,$8F,$B9,$7D,$1F,$E1,$7D
	.dc.b	$7F,$6F,$8F,$DF,$1F,$47,$93,$E1,$BF,$4F,$97,$E3,$BF,$57,$9B,$E5
	.dc.b	$E7,$F4,$79,$EF,$47,$D2,$AF,$A0,$7D,$2D,$BE,$CB,$75,$FB,$33,$B9
	.dc.b	$CC,$F0,$23,$83,$F8,$AE,$1E,$27,$F8,$BE,$3E,$67,$F8,$CE,$5E,$A7
	.dc.b	$F8,$DE,$7F,$62,$FF,$D6,$DC,$FF,$61,$BC,$DF,$61,$CF,$2F,$D8,$77
	.dc.b	$C9,$F6,$1E,$3C,$7F,$61,$FF,$42,$DF,$F8,$CB,$80,$BD,$2F,$C3,$7C
	.dc.b	$27,$E1,$FB,$BF,$11,$F1,$1F,$89,$F4,$FF,$15,$F0,$FF,$8B,$F7,$E6
	.dc.b	$7E,$66,$AE,$1F,$F6,$5E,$69,$CC,$61,$BC,$DF,$A4,$9D,$57,$32,$1C
	.dc.b	$E3,$53,$C2,$1A,$FA,$B0,$F3,$47,$C8,$CB,$D4,$FC,$AD,$3F,$45,$AF
	.dc.b	$E8,$FD,$DF,$D2,$75,$FE,$97,$E1,$BF,$4D,$F1,$1F,$A7,$F7,$BF,$51
	.dc.b	$EB,$FF,$8C,$F2,$F1,$1E,$0F,$E0,$1F,$F4,$5E,$C7,$FD,$17,$C5,$FE
	.dc.b	$EA,$7F,$D1,$6D,$F8,$6E,$4D,$AF,$D5,$57,$D2,$D7,$CB,$1F,$EF,$5B
	.dc.b	$FF,$20,$57,$3F,$89,$F6,$CF,$8A,$FE,$CF,$2F,$D4,$8E,$1E,$EF,$6E
	.dc.b	$5F,$E1,$F8,$FF,$58,$9E,$0D,$C7,$CD,$5B,$9F,$63,$70,$F6,$36,$B8
	.dc.b	$5F,$48,$C1,$2B,$C2,$43,$BB,$E5,$7D,$1C,$CC,$7D,$78,$FB,$67,$E0
	.dc.b	$7B,$BF,$09,$DE,$9F,$59,$FD,$1F,$0F,$D6,$7E,$37,$BB,$E5,$F8,$9F
	.dc.b	$34,$79,$FE,$06,$3F,$2F,$87,$D4,$7D,$BF,$1B,$FC,$25,$AF,$E8,$F8
	.dc.b	$1F,$13,$FA,$2B,$FD,$8F,$6F,$E8,$9F,$A9,$F0,$BF,$0F,$E2,$7F,$25
	.dc.b	$CB,$FA,$AE,$FB,$97,$E6,$AF,$DF,$18,$23,$F9,$CE,$3F,$D1,$37,$9B
	.dc.b	$F3,$AF,$E0,$70,$FA,$18,$F8,$63,$E4,$E5,$E5,$7E,$DE,$7F,$9B,$7F
	.dc.b	$83,$F6,$7D,$7F,$87,$FB,$1F,$13,$F9,$B9,$FE,$17,$D7,$F6,$FE,$27
	.dc.b	$EE,$7C,$4F,$DD,$F8,$6F,$BD,$F1,$32,$FF,$2B,$C5,$FC,$15,$FE,$27
	.dc.b	$D7,$F8,$3E,$CF,$C2,$77,$FD,$AE,$3F,$DD,$77,$7D,$FE,$5F,$C6,$E1
	.dc.b	$F8,$36,$BF,$84,$31,$3F,$8B,$69,$F9,$22,$F3,$FE,$49,$BD,$1F,$92
	.dc.b	$74,$7E,$33,$C5,$F1,$3F,$C1,$34,$FC,$73,$FA,$BC,$BD,$0F,$A8,$87
	.dc.b	$E5,$3C,$3F,$E0,$4F,$F3,$0F,$0F,$F4,$1E,$07,$E3,$63,$F6,$87,$E3
	.dc.b	$0F,$D5,$F7,$FE,$31,$FC,$71,$FB,$CC,$9F,$13,$FA,$93,$F5,$5F,$13
	.dc.b	$FA,$BF,$83,$FD,$67,$C4,$CB,$FE,$01,$F1,$3F,$B1,$F0,$3F,$CC,$1C
	.dc.b	$26,$93,$8B,$44,$AF,$F6,$47,$1A,$3F,$C1,$F3,$B9,$CC,$60,$E0,$FF
	.dc.b	$0B,$8D,$36,$3F,$80,$5F,$96,$3F,$84,$E1,$E9,$7F,$8B,$65,$ED,$7F
	.dc.b	$8C,$67,$E9,$F8,$6F,$57,$C3,$FA,$FE,$23,$D9,$F1,$1E,$DF,$88,$F7
	.dc.b	$7C,$47,$BF,$E2,$26,$5E,$63,$F9,$EE,$EF,$88,$F8,$87,$EE,$7C,$8F
	.dc.b	$E3,$BA,$F8,$6B,$F8,$EE,$DF,$27,$E2,$3D,$1C,$7F,$CC,$F8,$7D,$75
	.dc.b	$D7,$93,$E6,$7C,$47,$CF,$24,$67,$44,$62,$3E,$33,$6F,$E7,$47,$9A
	.dc.b	$31,$CB,$F4,$1B,$19,$3A,$3E,$27,$E5,$9F,$CF,$F0,$F4,$3F,$30,$F9
	.dc.b	$64,$3E,$B7,$C4,$7E,$17,$DA,$FC,$37,$C4,$7C,$C9,$F4,$71,$FE,$27
	.dc.b	$DC,$F9,$71,$F7,$78,$FF,$19,$F1,$12,$3F,$A8,$EC,$F7,$BF,$21,$F1
	.dc.b	$1F,$91,$F8,$0F,$BD,$5F,$85,$F8,$2F,$CA,$7C,$47,$E2,$38,$FE,$0F
	.dc.b	$27,$E5,$79,$30,$7E,$36,$FE,$53,$87,$ED,$B6,$38,$46,$88,$7E,$44
	.dc.b	$B3,$DB,$3F,$CA,$B6,$FF,$4D,$87,$9A,$35,$8F,$AA,$FE,$D6,$7F,$93
	.dc.b	$7E,$E7,$6F,$E4,$DF,$C1,$7F,$00,$FF,$8B,$5F,$ED,$8B,$E0,$FF,$2B
	.dc.b	$1A,$F8,$8F,$EC,$92,$F8,$4F,$C9,$CF,$E1,$FB,$BF,$28,$FF,$81,$7B
	.dc.b	$7F,$29,$A7,$F8,$0F,$B7,$EE,$CF,$E0,$09,$1F,$E0,$7F,$0F,$F0,$FC
	.dc.b	$5F,$94,$9F,$C3,$90,$FC,$89,$67,$E2,$76,$F0,$E3,$FC,$BC,$E7,$93
	.dc.b	$F2,$4D,$3F,$24,$7F,$80,$B0,$FD,$2C,$7F,$81,$BB,$7C,$B8,$79,$EB
	.dc.b	$F2,$0F,$F8,$6F,$3F,$E2,$9D,$BF,$91,$9F,$D3,$76,$7E,$45,$FE,$15
	.dc.b	$D9,$F9,$17,$F5,$5D,$7F,$92,$C3,$ED,$75,$FE,$46,$5E,$DB,$C0,$FF
	.dc.b	$52,$F2,$BE,$23,$E7,$FC,$2F,$D0,$F8,$8F,$A3,$F0,$DF,$4B,$DC,$FA
	.dc.b	$7E,$AF,$D4,$F8,$5F,$AB,$94,$D3,$EE,$B5,$FB,$AD,$DB,$F3,$3A,$78
	.dc.b	$67,$64,$36,$58,$DC,$F1,$E5,$F3,$DF,$8A,$BF,$17,$8F,$99,$FC,$6E
	.dc.b	$7E,$A7,$F1,$FA,$FE,$2B,$E1,$BF,$17,$EB,$7E,$33,$D6,$FC,$6F,$C0
	.dc.b	$FE,$3B,$D5,$FC,$7F,$C2,$7E,$43,$D6,$FC,$8F,$C3,$7E,$4B,$DF,$9A
	.dc.b	$7F,$92,$B8,$FE,$43,$FC,$77,$F8,$07,$FC,$AB,$0D,$FF,$C7,$91,$FE
	.dc.b	$56,$E2,$F9,$FC,$38,$F7,$B3,$F0,$BE,$DC,$7F,$CD,$B8,$FC,$57,$E1
	.dc.b	$93,$9B,$E1,$BF,$17,$0F,$B3,$CD,$C7,$C5,$5F,$1B,$DD,$F1,$BD,$BC
	.dc.b	$4F,$F3,$DE,$6C,$7C,$91,$F3,$0F,$67,$B5,$F8,$37,$DC,$7E,$D3,$9F
	.dc.b	$E6,$3F,$07,$DE,$F9,$8F,$C3,$F7,$7E,$0F,$C3,$FC,$2F,$7B,$E1,$FC
	.dc.b	$3F,$CC,$AF,$91,$F0,$DF,$18,$FD,$DF,$03,$5F,$EE,$53,$FC,$23,$C1
	.dc.b	$FB,$CF,$F0,$8F,$07,$E6,$FC,$3F,$E0,$E3,$F8,$5F,$83,$FB,$5F,$36
	.dc.b	$47,$EC,$F1,$FD,$F3,$FB,$C2,$72,$7E,$F1,$D7,$9B,$FB,$C7,$3C,$BF
	.dc.b	$BC,$7A,$9F,$C6,$BC,$4F,$0B,$FC,$73,$3F,$2D,$FE,$7B,$4F,$C0,$E1
	.dc.b	$F8,$2F,$87,$FC,$8D,$FE,$1F,$E1,$FF,$22,$FF,$24,$F0,$BF,$22,$FF
	.dc.b	$26,$F0,$7F,$13,$F0,$FF,$90,$8F,$E4,$FE,$0F,$C6,$ED,$FC,$8C,$7F
	.dc.b	$A5,$79,$3F,$23,$E0,$FE,$43,$B7,$F7,$13,$F9,$5F,$03,$F1,$EF,$E0
	.dc.b	$79,$3F,$0F,$E0,$7E,$27,$1F,$C4,$72,$FE,$1A,$BF,$10,$7E,$52,$74
	.dc.b	$47,$94,$F0,$7F,$17,$E0,$7E,$29,$FE,$7B,$E0,$F8,$DF,$DA,$9F,$8A
	.dc.b	$3F,$6A,$7F,$44,$F1,$FF,$4F,$E0,$FE,$2A,$3D,$9E,$0F,$E2,$9F,$D7
	.dc.b	$78,$1F,$8A,$7F,$A5,$F8,$7F,$89,$7F,$16,$7F,$AF,$B8,$5F,$0F,$FB
	.dc.b	$2F,$87,$FD,$9F,$A1,$2B,$F7,$1E,$9F,$ED,$BE,$17,$F6,$FE,$9C,$9F
	.dc.b	$DE,$35,$F9,$CD,$F0,$3F,$B9,$E4,$8C,$72,$2B,$F5,$B1,$86,$9D,$08
	.dc.b	$74,$3F,$85,$5F,$F0,$8C,$3C,$D1,$F7,$71,$F5,$3F,$C3,$F3,$FE,$1B
	.dc.b	$EA,$7F,$B2,$7F,$80,$7F,$CB,$23,$E1,$B1,$ED,$E7,$E1,$B5,$F0,$DB
	.dc.b	$F5,$F8,$7C,$37,$1F,$63,$97,$B7,$35,$FF,$73,$E6,$F8,$6E,$DE,$77
	.dc.b	$7F,$73,$C1,$F0,$DE,$1F,$77,$C5,$F0,$DE,$FF,$17,$F7,$5D,$17,$AB
	.dc.b	$E3,$69,$F0,$3A,$FE,$4C,$E4,$91,$9E,$BF,$95,$7C,$A3,$07,$CA,$3F
	.dc.b	$DE,$9B,$31,$F1,$47,$F9,$3E,$19,$F9,$E3,$1A,$FC,$7D,$BE,$4C,$6F
	.dc.b	$1F,$95,$CD,$F2,$FD,$EF,$91,$3C,$F9,$7E,$43,$D7,$E0,$BE,$87,$C3
	.dc.b	$7C,$78,$F1,$F2,$7F,$1A,$E2,$FC,$8C,$7F,$45,$EF,$FB,$7C,$7F,$88
	.dc.b	$ED,$FB,$D3,$F8,$03,$D1,$F0,$9F,$6C,$8E,$08,$7C,$07,$2D,$33,$3F
	.dc.b	$E4,$6E,$24,$EA,$30,$DD,$FE,$6C,$EA,$BC,$47,$F0,$BE,$FF,$3C,$FD
	.dc.b	$2C,$BF,$D5,$B3,$EB,$3F,$24,$7C,$FF,$85,$FC,$97,$C3,$7E,$86,$19
	.dc.b	$7F,$96,$3D,$9F,$F2,$03,$E5,$F0,$34,$E7,$E1,$AE,$5F,$80,$A7,$C3
	.dc.b	$63,$07,$5A,$7F,$C8,$91,$F7,$FC,$0F,$CD,$3F,$81,$F0,$2B,$FE,$20
	.dc.b	$F3,$BD,$FD,$7F,$9A,$8F,$C3,$78,$3F,$9A,$7F,$11,$E0,$FE,$69,$FC
	.dc.b	$7E,$5F,$91,$7F,$6C,$4D,$BF,$32,$EB,$8B,$F3,$AE,$76,$F9,$1E,$0F
	.dc.b	$E7,$23,$07,$E6,$6B,$8B,$F9,$AF,$07,$F3,$2F,$E7,$3C,$1F,$CC,$3F
	.dc.b	$9E,$F0,$7F,$30,$FF,$83,$FC,$4F,$E0,$18,$FE,$62,$3F,$41,$E0,$FE
	.dc.b	$61,$FD,$2F,$83,$F9,$87,$F4,$FE,$0F,$E6,$1F,$D3,$78,$1F,$99,$F8
	.dc.b	$5F,$F1,$A7,$07,$CD,$F8,$1F,$A0,$FC,$FF,$6E,$F4,$FF,$87,$70,$FF
	.dc.b	$5E,$35,$FB,$FA,$7D,$9F,$73,$7E,$87,$0F,$76,$57,$06,$BC,$0F,$BB
	.dc.b	$68,$CF,$27,$E2,$5B,$87,$E1,$5B,$5C,$1B,$97,$E1,$5B,$6C,$3D,$38
	.dc.b	$FA,$BD,$69,$87,$AB,$3F,$6E,$9E,$EC,$3D,$FD,$7F,$03,$D2,$F8,$3F
	.dc.b	$0B,$F0,$BE,$17,$E1,$FC,$2F,$C4,$F7,$7E,$2F,$C2,$FC,$6F,$7A,$6B
	.dc.b	$F1,$9F,$F0,$EE,$6F,$57,$E5,$FB,$BF,$3B,$F8,$07,$F5,$17,$F0,$F9
	.dc.b	$51,$FE,$49,$8F,$D0,$AF,$77,$E6,$C7,$FB,$B6,$3F,$C7,$4D,$F7,$3A
	.dc.b	$11,$C1,$0E,$D1,$C1,$B1,$B6,$9B,$E9,$FF,$68,$CE,$DB,$3B,$7F,$7D
	.dc.b	$CF,$EB,$BF,$67,$BD,$ED,$C3,$DE,$7E,$23,$3F,$9C,$FE,$2B,$4F,$9C
	.dc.b	$FE,$33,$8F,$E7,$3E,$DF,$77,$E7,$3F,$90,$E0,$F9,$1E,$EF,$C9,$F8
	.dc.b	$69,$3F,$DC,$7B,$BF,$37,$8F,$F8,$9B,$6F,$E2,$5D,$5F,$80,$AF,$9D
	.dc.b	$3F,$99,$9F,$E2,$DD,$9F,$4F,$DE,$FC,$4F,$74,$BF,$CC,$10,$E2,$E5
	.dc.b	$A6,$6B,$F3,$0D,$EF,$FC,$C3,$F5,$EB,$F2,$6F,$EA,$BB,$FC,$95,$E6
	.dc.b	$8F,$D6,$F7,$FE,$17,$E1,$BF,$27,$1F,$7F,$E1,$BF,$26,$FE,$CB,$C0
	.dc.b	$FC,$9B,$F8,$3F,$85,$FC,$67,$C3,$7E,$4A,$3D,$FD,$F7,$8B,$E2,$F4
	.dc.b	$7D,$49,$FC,$19,$F6,$0F,$C6,$75,$7E,$39,$FC,$19,$35,$FC,$0B,$F8
	.dc.b	$5E,$0F,$EE,$FD,$BF,$85,$6F,$6F,$E1,$5B,$3B,$3F,$37,$19,$69,$9E
	.dc.b	$EF,$D0,$C7,$86,$FF,$23,$87,$96,$BC,$E7,$F1,$2E,$FF,$E2,$B9,$7E
	.dc.b	$32,$3E,$EF,$3F,$E3,$1F,$BF,$DB,$F8,$C7,$F4,$3D,$BF,$8C,$7F,$07
	.dc.b	$DB,$F8,$C7,$F5,$9D,$9F,$8C,$7F,$A5,$1F,$B2,$F8,$6F,$DC,$7C,$37
	.dc.b	$EE,$7E,$1A,$4F,$EC,$FE,$13,$F7,$9F,$01,$FB,$DF,$80,$93,$FB,$76
	.dc.b	$BF,$B2,$6F,$0F,$EB,$9B,$C9,$F3,$1B,$3F,$C0,$9B,$C9,$FC,$19,$BC
	.dc.b	$AC,$3F,$86,$63,$E4,$BF,$DF,$E5,$E8,$7F,$84,$69,$FC,$47,$DE,$91
	.dc.b	$FC,$2F,$E1,$7F,$8A,$7A,$9F,$C5,$7D,$1F,$E8,$7F,$C0,$2D,$EF,$C2
	.dc.b	$63,$DF,$CF,$C3,$6B,$D4,$99,$FE,$CF,$D2,$FC,$2E,$DC,$BE,$1B,$9F
	.dc.b	$BF,$D3,$DF,$EB,$EB,$7C,$8C,$3F,$CD,$5D,$3D,$DF,$F0,$0E,$FF,$E5
	.dc.b	$37,$3B,$FF,$91,$A3,$11,$FC,$A5,$B3,$C7,$A5,$18,$AE,$1E,$FF,$B0
	.dc.b	$97,$D0,$EE,$E1,$E8,$DB,$E9,$5C,$E7,$67,$1B,$EF,$5E,$B8,$F2,$61
	.dc.b	$EF,$75,$4C,$BE,$24,$7D,$43,$F9,$E7,$2F,$D4,$7D,$9E,$D7,$D4,$78
	.dc.b	$FB,$FF,$53,$DC,$FA,$BE,$FF,$D6,$F7,$BE,$9D,$76,$F7,$BE,$C9,$FE
	.dc.b	$EE,$EF,$78,$3F,$83,$9F,$C2,$9F,$D2,$FB,$BF,$12,$F9,$BD,$EF,$C6
	.dc.b	$F8,$52,$3F,$34,$43,$F3,$45,$E4,$FC,$D3,$79,$7F,$34,$D3,$F3,$45
	.dc.b	$E5,$FA,$B7,$E0,$8F,$19,$2B,$FA,$8E,$3E,$77,$E8,$FC,$37,$E5,$3E
	.dc.b	$1B,$FB,$57,$85,$F9,$6F,$85,$FC,$C6,$1F,$3B,$3F,$CB,$BF,$43,$C0
	.dc.b	$FC,$DF,$C2,$FE,$5A,$3F,$11,$E0,$7C,$5E,$FF,$A7,$1F,$8B,$EA,$FB
	.dc.b	$0F,$DA,$3F,$B8,$78,$DF,$83,$C7,$F1,$71,$F7,$3B,$B0,$FE,$10,$FC
	.dc.b	$49,$92,$18,$26,$FF,$8C,$8E,$3C,$5F,$8A,$9D,$CF,$F7,$2B,$FD,$3C
	.dc.b	$7E,$2B,$BF,$F5,$6F,$E3,$3C,$0F,$F0,$66,$7F,$89,$BF,$EF,$3D,$DF
	.dc.b	$BA,$EE,$FC,$55,$FE,$9F,$BB,$F1,$4F,$E5,$3B,$7F,$14,$FE,$5B,$C2
	.dc.b	$FC,$4B,$F8,$B3,$F7,$87,$F9,$D3,$F8,$07,$FC,$FB,$C1,$FE,$B8,$5F
	.dc.b	$0D,$FE,$4A,$7F,$CF,$73,$FA,$57,$FC,$F2,$FF,$87,$DF,$EE,$D3,$C7
	.dc.b	$DE,$E0,$49,$E0,$D7,$83,$74,$FD,$F5,$23,$15,$CE,$B8,$90,$E2,$5A
	.dc.b	$E3,$3E,$7F,$6B,$D1,$F1,$3E,$9F,$87,$F5,$65,$EB,$F8,$EF,$67,$BD
	.dc.b	$32,$F6,$7C,$0F,$BF,$E1,$3E,$07,$C1,$7C,$1F,$5B,$E1,$79,$DF,$0F
	.dc.b	$D6,$F8,$9E,$FF,$C5,$F8,$7F,$8D,$F0,$F3,$3F,$9B,$DE,$E6,$7C,$3F
	.dc.b	$8A,$7E,$57,$B3,$E6,$7D,$5D,$1E,$F7,$DA,$4E,$8F,$6B,$A3,$EC,$74
	.dc.b	$7C,$66,$99,$C7,$C1,$19,$6C,$67,$1F,$9C,$D9,$F1,$3F,$6F,$0F,$33
	.dc.b	$F7,$71,$F8,$CF,$DF,$E3,$FC,$27,$C0,$7E,$17,$DC,$F8,$F3,$F8,$5F
	.dc.b	$77,$E3,$BF,$CA,$FB,$FF,$CC,$BA,$FD,$EE,$1F,$91,$39,$F8,$7F,$F4
	.dc.b	$BD,$1F,$F6,$97,$C0,$FF,$A7,$37,$FC,$AD,$7E,$5F,$B3,$EE,$3F,$93
	.dc.b	$EE,$FC,$03,$FA,$EE,$3F,$DD,$BF,$B4,$6E,$7F,$B5,$8D,$1F,$BD,$70
	.dc.b	$EC,$C4,$70,$23,$82,$71,$FE,$F5,$FD,$15,$78,$9F,$D2,$5F,$99,$FD
	.dc.b	$37,$67,$BB,$E0,$3D,$FF,$03,$FA,$DE,$EF,$C8,$57,$C7,$F8,$1F,$C8
	.dc.b	$3F,$AF,$ED,$FC,$83,$F2,$FE,$1F,$E3,$FC,$17,$E3,$E3,$F6,$5D,$9F
	.dc.b	$23,$BF,$F7,$5D,$7F,$6B,$8B,$F5,$BD,$9F,$79,$FC,$5D,$FE,$37,$8F
	.dc.b	$F2,$11,$82,$38,$24,$7E,$38,$E0,$58,$DB,$63,$7C,$7F,$90,$8D,$B6
	.dc.b	$76,$FE,$97,$C3,$FE,$1D,$7E,$7B,$F4,$DF,$F0,$FE,$DF,$C8,$E1,$FC
	.dc.b	$1B,$B7,$F2,$13,$FC,$4F,$C1,$FE,$2D,$DB,$F9,$08,$FF,$59,$76,$7E
	.dc.b	$43,$1F,$E3,$9D,$BF,$90,$9C,$F6,$FE,$40,$F9,$0B,$65,$BE,$1F,$F2
	.dc.b	$BF,$0F,$F9,$6F,$87,$FC,$BF,$C3,$FE,$63,$E1,$FF,$33,$E6,$CC,$BF
	.dc.b	$40,$D7,$C6,$DE,$1F,$F0,$E6,$48,$C6,$CA,$F1,$4E,$5E,$64,$39,$96
	.dc.b	$B9,$D7,$28,$E3,$7E,$67,$F6,$58,$FA,$9F,$DA,$67,$FB,$0F,$6B,$F6
	.dc.b	$3E,$EF,$EC,$BE,$1F,$F6,$7E,$9F,$ED,$3D,$EF,$DA,$FA,$5F,$B6,$F6
	.dc.b	$FF,$6F,$F0,$FF,$B8,$F8,$7F,$46,$9F,$EE,$1F,$E0,$1F,$F2,$03,$7E
	.dc.b	$23,$3C,$5F,$E3,$B9,$47,$D0,$E7,$F9,$FC,$1A,$E5,$F8,$50,$A7,$BD
	.dc.b	$F4,$71,$FF,$1B,$43,$D6,$FA,$51,$8F,$77,$93,$F6,$E5,$18,$BE,$8D
	.dc.b	$8C,$CF,$12,$1E,$B8,$D1,$1C,$E7,$EB,$8C,$46,$CC,$E3,$EA,$AE,$B1
	.dc.b	$F9,$4E,$6F,$06,$3E,$B8,$F9,$19,$7F,$22,$CF,$C9,$ED,$7D,$49,$FE
	.dc.b	$55,$CB,$E7,$F7,$7D,$1E,$E7,$D4,$9F,$87,$F1,$1F,$51,$FE,$61,$C5
	.dc.b	$ED,$DB,$DD,$EF,$49,$FE,$AB,$BF,$E7,$9F,$E6,$DD,$FF,$0F,$DE,$FC
	.dc.b	$44,$7D,$BF,$87,$FC,$67,$2F,$E9,$3B,$EF,$81,$F9,$5A,$FC,$99,$39
	.dc.b	$FF,$26,$FE,$57,$4F,$B7,$3B,$9C,$CE,$E7,$35,$D0,$EF,$9B,$1F,$1D
	.dc.b	$FE,$27,$2F,$CD,$61,$F8,$FC,$7F,$34,$FE,$13,$DF,$FC,$D3,$FD,$47
	.dc.b	$BB,$F3,$4F,$E2,$3E,$1F,$F3,$4F,$E2,$BE,$1F,$F0,$5F,$0D,$F8,$3F
	.dc.b	$86,$FC,$27,$74,$AF,$EF,$79,$7F,$03,$7F,$23,$D7,$F6,$75,$FC,$E7
	.dc.b	$67,$E4,$A7,$F0,$44,$8F,$C3,$3F,$84,$3F,$29,$DD,$F8,$88,$CC,$6F
	.dc.b	$8F,$F1,$13,$96,$C7,$1E,$4F,$CC,$B7,$8F,$F1,$B7,$FC,$33,$BB,$F1
	.dc.b	$B3,$FC,$00,$F3,$9F,$C0,$0F,$D4,$F6,$FE,$35,$FE,$4F,$DB,$F8,$D7
	.dc.b	$F5,$BD,$9F,$8D,$7F,$79,$D9,$F8,$D7,$F9,$6F,$67,$E3,$5F,$D9,$F6
	.dc.b	$7E,$35,$FF,$77,$E9,$FB,$9F,$86,$FD,$77,$C3,$7E,$BF,$E1,$BF,$65
	.dc.b	$FC,$03,$FD,$F2,$8F,$F9,$42,$FF,$D6,$4F,$F9,$41,$6F,$FE,$C3,$8C
	.dc.b	$37,$DE,$E6,$49,$E2,$D7,$8F,$17,$F0,$CA,$E2,$D9,$D5,$73,$AE,$24
	.dc.b	$38,$9F,$D1,$6B,$8B,$67,$D7,$F0,$FE,$CF,$7F,$DD,$3F,$C5,$B3,$FE
	.dc.b	$1D,$A7,$C0,$F5,$7E,$0F,$C0,$7C,$2F,$84,$F8,$7E,$B7,$C4,$F7,$FE
	.dc.b	$2F,$C3,$FC,$6F,$5F,$E3,$FB,$FF,$23,$DC,$F9,$3F,$0F,$F2,$BD,$99
	.dc.b	$BF,$F9,$43,$8D,$F0,$FF,$42,$3E,$5F,$BF,$E5,$7E,$6F,$BF,$F0,$3A
	.dc.b	$3E,$14,$7B,$FA,$3E,$13,$F0,$3A,$7E,$AF,$4E,$E3,$E0,$1B,$AE,$2E
	.dc.b	$F2,$F8,$B5,$C2,$34,$70,$35,$5C,$E3,$F0,$38,$79,$67,$96,$3E,$97
	.dc.b	$A6,$5F,$88,$F7,$FF,$13,$EF,$FC,$99,$FC,$3F,$27,$C9,$7F,$13,$C7
	.dc.b	$F2,$5F,$C5,$F1,$7C,$97,$F1,$BC,$5F,$92,$F8,$79,$1F,$C9,$3A,$7F
	.dc.b	$2B,$F0,$1F,$6A,$3E,$17,$F0,$0F,$F8,$2D,$FC,$17,$77,$F4,$B8,$FF
	.dc.b	$2C,$47,$F7,$9E,$57,$2F,$F5,$A9,$CC,$F2,$F8,$1E,$04,$78,$4E,$E3
	.dc.b	$11,$B3,$27,$F8,$56,$79,$D7,$68,$C4,$F0,$8C,$72,$2B,$77,$D6,$F9
	.dc.b	$CE,$6F,$84,$FE,$97,$0F,$65,$7C,$1C,$3F,$61,$8F,$E6,$A3,$F6,$BD
	.dc.b	$BF,$9A,$79,$E4,$F8,$6F,$89,$F0,$DF,$17,$E1,$BE,$37,$C3,$7C,$7F
	.dc.b	$86,$F9,$1E,$F7,$C9,$F7,$7F,$37,$97,$F0,$0E,$FF,$CD,$BF,$C2,$BB
	.dc.b	$FF,$36,$FE,$AC,$FC,$D9,$FC,$1B,$C6,$FE,$17,$DF,$F8,$B9,$FC,$89
	.dc.b	$3C,$0C,$12,$BF,$27,$CD,$F9,$57,$F2,$66,$27,$83,$6B,$F2,$9C,$7F
	.dc.b	$C4,$7C,$0F,$CA,$3F,$C4,$FC,$0F,$CA,$57,$F1,$8E,$FF,$CA,$3F,$A3
	.dc.b	$EF,$FC,$A3,$FA,$5E,$FF,$CA,$3F,$A7,$EF,$FC,$A3,$FA,$9E,$FF,$C9
	.dc.b	$BF,$AB,$EF,$FF,$13,$74,$BE,$17,$F2,$1F,$0D,$F9,$1F,$43,$F2,$5E
	.dc.b	$FF,$E4,$FE,$1B,$F2,$9E,$8C,$CB,$EA,$B5,$FC,$DB,$78,$7F,$0F,$92
	.dc.b	$31,$E1,$FE,$65,$B5,$F9,$16,$F3,$7E,$35,$B5,$C6,$7F,$07,$87,$91
	.dc.b	$FC,$2E,$5E,$87,$F0,$FA,$7E,$A7,$E1,$BF,$55,$EF,$7E,$AF,$E1,$BF
	.dc.b	$59,$F0,$DF,$AD,$F8,$6F,$D7,$7C,$37,$EB,$FD,$5F,$F1,$D6,$7F,$E4
	.dc.b	$EE,$1F,$CF,$BF,$E4,$CD,$3F,$3A,$D7,$F8,$D3,$FE,$4A,$73,$EE,$7F
	.dc.b	$94,$63,$FC,$95,$1C,$3D,$BE,$BE,$D7,$6F,$86,$F9,$93,$F2,$88,$7C
	.dc.b	$A2,$E1,$F2,$B8,$FE,$19,$23,$3C,$BF,$2D,$A7,$CB,$2F,$2F,$CB,$7E
	.dc.b	$49,$EA,$BF,$1B,$E5,$3D,$BE,$EF,$C9,$7D,$FE,$E7,$C4,$F7,$3E,$4C
	.dc.b	$7E,$D3,$87,$E3,$FB,$9F,$26,$3E,$2F,$BD,$F2,$BD,$F9,$1F,$23,$E1
	.dc.b	$7E,$69,$F9,$0E,$AF,$B1,$1F,$C8,$B8,$FE,$E3,$F3,$3E,$1B,$F5,$8F
	.dc.b	$EA,$89,$C5,$FD,$28,$9C,$3F,$D1,$9B,$1F,$D0,$DB,$C3,$FB,$36,$F9
	.dc.b	$3F,$B2,$9E,$75,$FE,$84,$C3,$FC,$89,$1F,$C4,$3B,$FF,$44,$FF,$12
	.dc.b	$F0,$3F,$9A,$67,$F8,$6F,$86,$FC,$6C,$FF,$16,$ED,$FC,$4F,$C3,$7E
	.dc.b	$2B,$E1,$BF,$1B,$3F,$88,$F8,$6F,$C6,$E9,$F8,$EF,$86,$FC,$7F,$BB
	.dc.b	$F9,$0E,$F9,$7F,$EB,$7F,$88,$79,$3E,$5E,$DF,$CA,$F8,$3F,$62,$BF
	.dc.b	$23,$DF,$F8,$A7,$F1,$04,$E6,$FC,$43,$F8,$2F,$23,$F5,$4E,$75,$FC
	.dc.b	$43,$4F,$C4,$17,$A3,$F1,$0D,$9E,$35,$D2,$31,$E4,$7E,$6D,$FD,$2F
	.dc.b	$87,$E6,$7F,$16,$7F,$3A,$F2,$3F,$5B,$E1,$FE,$2E,$3F,$9E,$F8,$FF
	.dc.b	$B3,$7F,$5B,$E1,$7E,$2D,$FD,$6F,$85,$F8,$A7,$F5,$FE,$17,$E2,$9F
	.dc.b	$C6,$1F,$1F,$8B,$ED,$7C,$2F,$EE,$BE,$1B,$FD,$69,$FC,$03,$FE,$43
	.dc.b	$3F,$C8,$55,$FE,$11,$7F,$C1,$64,$F5,$36,$48,$FE,$B7,$19,$6C,$65
	.dc.b	$E4,$DE,$0F,$EC,$B3,$B9,$C9,$C0,$87,$87,$E0,$3C,$BE,$B7,$9E,$3F
	.dc.b	$86,$E3,$E8,$F7,$FD,$3F,$0F,$EA,$F8,$7F,$5F,$AB,$EC,$F8,$7F,$6F
	.dc.b	$AD,$EE,$F8,$7F,$7F,$AF,$F0,$38,$A6,$9E,$AF,$87,$F8,$7E,$EC,$8F
	.dc.b	$90,$3B,$9E,$7B,$83,$C9,$3F,$1F,$87,$E5,$FB,$7F,$33,$DC,$93,$F1
	.dc.b	$9A,$FC,$D6,$F3,$7C,$46,$F8,$5F,$EF,$2E,$74,$62,$BC,$51,$87,$A1
	.dc.b	$0E,$85,$AE,$8F,$D6,$BF,$96,$FF,$36,$E9,$FE,$77,$97,$CB,$8F,$B7
	.dc.b	$EE,$FE,$13,$E1,$FF,$0B,$EF,$7C,$B9,$FB,$FC,$DF,$2D,$FC,$0F,$2F
	.dc.b	$CA,$7F,$11,$C9,$F8,$CE,$4F,$C6,$FB,$D2,$7E,$5F,$23,$81,$F0,$C9
	.dc.b	$FF,$12,$FF,$00,$FF,$8F,$E8,$FF,$8A,$A3,$E0,$F7,$7E,$2F,$97,$E3
	.dc.b	$77,$D9,$F8,$AD,$BE,$CD,$C7,$F2,$1C,$BF,$92,$8C,$12,$BF,$23,$19
	.dc.b	$7F,$3B,$87,$95,$FF,$3A,$63,$E3,$C3,$CB,$1F,$AA,$CB,$57,$FE,$42
	.dc.b	$AF,$F1,$A4,$FE,$9B,$BB,$F2,$CF,$EA,$3B,$77,$A7,$C1,$C7,$DF,$F0
	.dc.b	$DF,$98,$AF,$87,$F0,$DF,$98,$7E,$2F,$C3,$70,$7F,$29,$93,$E1,$B1
	.dc.b	$87,$C8,$F8,$6F,$CD,$57,$EE,$FC,$0F,$EE,$1D,$7F,$98,$DF,$F6,$7D
	.dc.b	$FF,$98,$AF,$C4,$6D,$F9,$1E,$FF,$C8,$65,$F5,$DB,$CB,$F9,$1A,$FC
	.dc.b	$79,$30,$FC,$7B,$6F,$93,$D0,$FC,$CD,$72,$39,$F7,$FE,$52,$3F,$C0
	.dc.b	$FD,$FF,$94,$7F,$86,$F8,$9F,$C5,$BC,$0F,$CA,$47,$F1,$0E,$FF,$CA
	.dc.b	$3F,$C5,$FB,$FF,$28,$FF,$19,$EF,$FC,$A6,$9F,$C4,$3B,$FF,$29,$3F
	.dc.b	$C6,$3B,$BF,$28,$FF,$BA,$39,$9F,$0D,$F2,$9F,$F2,$04,$7E,$53,$DC
	.dc.b	$C7,$87,$9F,$76,$4F,$E6,$1A,$FE,$19,$B9,$7E,$0D,$BC,$7F,$7F,$04
	.dc.b	$62,$BE,$DB,$79,$7F,$42,$D9,$DC,$F7,$3F,$4F,$87,$9F,$1F,$47,$BB
	.dc.b	$E9,$CF,$D5,$E8,$7A,$FE,$0F,$D9,$E8,$7B,$7E,$1B,$DD,$EF,$4C,$FF
	.dc.b	$63,$E8,$FC,$13,$E1,$7C,$1C,$8F,$DA,$FC,$37,$C5,$F8,$69,$1F,$BC
	.dc.b	$FE,$01,$FF,$22,$72,$7B,$77,$F2,$4F,$98,$FF,$50,$CF,$F9,$17,$97
	.dc.b	$FD,$4E,$DA,$FD,$13,$67,$F9,$83,$79,$4D,$27,$A7,$FA,$06,$CF,$68
	.dc.b	$C1,$1C,$10,$FF,$77,$47,$07,$F7,$BC,$9E,$87,$F7,$F8,$F5,$C3,$E5
	.dc.b	$C7,$F1,$6E,$2F,$96,$FF,$18,$E0,$F9,$4F,$F0,$6E,$0F,$94,$FC,$C3
	.dc.b	$F8,$EE,$FF,$2D,$FE,$16,$FC,$A7,$F8,$6E,$DF,$C3,$FD,$CF,$F0,$EC
	.dc.b	$7E,$FB,$93,$DD,$EE,$FB,$FD,$DF,$81,$EE,$FE,$17,$AA,$5F,$E1,$4F
	.dc.b	$C9,$17,$93,$F2,$4D,$DF,$F2,$53,$C2,$36,$49,$DB,$67,$EA,$D7,$86
	.dc.b	$3E,$DE,$1F,$8F,$7E,$EF,$B7,$F8,$F7,$F9,$DF,$4F,$D5,$F7,$7E,$B7
	.dc.b	$B9,$F5,$FE,$1F,$EC,$7B,$BF,$67,$DD,$FB,$5F,$0F,$F6,$FE,$1F,$F6
	.dc.b	$5D,$DF,$77,$E1,$A6,$9F,$8F,$E0,$FC,$FB,$FB,$53,$F1,$C7,$DC,$EE
	.dc.b	$FC,$03,$4F,$C0,$71,$7E,$42,$32,$D8,$CC,$FE,$4A,$72,$D8,$C9,$F9
	.dc.b	$48,$CB,$4C,$F7,$FF,$58,$EE,$FC,$55,$FF,$43,$EE,$FC,$55,$7E,$24
	.dc.b	$87,$F4,$CF,$07,$F6,$8F,$F6,$2E,$DF,$C5,$D7,$F4,$BF,$07,$F1,$73
	.dc.b	$F8,$83,$F1,$64,$3F,$12,$7F,$6A,$ED,$FC,$57,$8B,$FA,$CF,$86,$FD
	.dc.b	$37,$C3,$7E,$9F,$E1,$BF,$51,$F0,$DF,$A9,$F8,$69,$8F,$ED,$9A,$FE
	.dc.b	$99,$B9,$FE,$E1,$BC,$5F,$B8,$6C,$FE,$E1,$BC,$3F,$B8,$6F,$8E,$6E
	.dc.b	$1E,$3A,$FE,$0B,$97,$EE,$F3,$FD,$E7,$AB,$FB,$DF,$47,$FA,$EE,$CF
	.dc.b	$56,$FB,$F8,$F8,$0C,$FA,$9A,$F7,$E6,$9F,$E0,$DE,$27,$17,$F6,$FC
	.dc.b	$9F,$0B,$D0,$95,$B9,$D7,$C0,$78,$0F,$97,$1F,$C5,$38,$3E,$5B,$FC
	.dc.b	$73,$97,$E5,$BF,$24,$87,$C9,$2E,$3F,$25,$B7,$F2,$5A,$7C,$92,$F2
	.dc.b	$FC,$96,$F2,$FC,$96,$F2,$FC,$A7,$C4,$7C,$A2,$0B,$F8,$2F,$F9,$6F
	.dc.b	$47,$AB,$F3,$63,$E0,$19,$F8,$5F,$9A,$E3,$1F,$F5,$7F,$AD,$F3,$A3
	.dc.b	$16,$E0,$F9,$F1,$DF,$4F,$9D,$C1,$F3,$23,$07,$85,$FB,$7F,$0B,$23
	.dc.b	$FA,$57,$47,$DE,$C3,$EF,$FC,$27,$E0,$3D,$EF,$D7,$E9,$FD,$0B,$C2
	.dc.b	$BC,$5F,$CE,$78,$7F,$75,$97,$EF,$C8,$E0,$9A,$FE,$EE,$32,$D8,$CC
	.dc.b	$7E,$EC,$C9,$F1,$F0,$F1,$BF,$03,$2F,$CB,$3F,$D7,$3A,$FF,$2B,$EE
	.dc.b	$7E,$5B,$AF,$F2,$FE,$EF,$E5,$6B,$E3,$FB,$BF,$9A,$F8,$5F,$CD,$FB
	.dc.b	$DF,$9C,$F7,$BF,$3B,$F0,$BF,$E3,$AF,$E0,$1F,$F3,$16,$AE,$FF,$C9
	.dc.b	$65,$FD,$BF,$BF,$F2,$2F,$F5,$AE,$FF,$C8,$BF,$8F,$27,$17,$E3,$DB
	.dc.b	$C7,$F8,$F6,$9F,$8F,$2F,$37,$E3,$DB,$CB,$F8,$F6,$F7,$FE,$E8,$FC
	.dc.b	$83,$E1,$3F,$1F,$7F,$DF,$31,$FD,$BF,$7F,$E3,$67,$F0,$38,$7F,$BE
	.dc.b	$3B,$FF,$1B,$18,$D5,$DF,$F8,$D8,$C7,$7F,$E3,$32,$FC,$CE,$1F,$F0
	.dc.b	$03,$BF,$F1,$91,$8C,$3F,$77,$7F,$94,$EF,$FC,$62,$F6,$FF,$B5,$E0
	.dc.b	$E3,$FF,$3D,$1F,$E7,$69,$F2,$AF,$86,$FA,$5F,$0D,$27,$F4,$52,$8C
	.dc.b	$47,$06,$E5,$E6,$9D,$4F,$09,$CF,$1F,$F9,$FE,$73,$3C,$4D,$10,$C1
	.dc.b	$2B,$A5,$72,$8F,$F6,$A4,$F9,$DF,$2F,$C4,$FE,$27,$E2,$BF,$15,$F1
	.dc.b	$1F,$8B,$F8,$8F,$C6,$7C,$47,$E3,$7E,$0B,$F1,$DE,$97,$E3,$FE,$1B
	.dc.b	$F2,$1E,$A4,$CF,$DF,$F0,$5F,$93,$23,$F0,$BE,$1B,$F2,$DF,$01,$23
	.dc.b	$FD,$7B,$BF,$CC,$E3,$FC,$D1,$E7,$E4,$FA,$1C,$56,$7E,$0F,$1F,$B2
	.dc.b	$FE,$03,$5F,$81,$C9,$F1,$EB,$81,$23,$71,$96,$9F,$12,$3C,$2F,$D8
	.dc.b	$C3,$E2,$3F,$6B,$8B,$F5,$DE,$BF,$C4,$8F,$BB,$F0,$DF,$B1,$F6,$3F
	.dc.b	$65,$EC,$7C,$39,$FC,$2F,$0F,$C2,$7F,$0F,$C3,$75,$FC,$07,$C3,$7E
	.dc.b	$E7,$DB,$95,$F1,$3D,$AF,$DE,$75,$FF,$AD,$BF,$80,$7F,$CC,$F2,$E4
	.dc.b	$47,$F9,$7C,$9E,$A6,$C9,$3F,$CC,$E3,$2D,$8C,$BF,$AE,$8C,$B6,$33
	.dc.b	$97,$EA,$A7,$2D,$8C,$F2,$7F,$B4,$E7,$CB,$7F,$A7,$C3,$F8,$A6,$3E
	.dc.b	$9F,$85,$FC,$66,$5E,$3F,$87,$FC,$63,$FA,$6E,$BF,$C6,$3E,$7F,$87
	.dc.b	$F8,$1F,$0F,$F0,$7E,$1F,$E1,$7C,$3F,$C3,$F8,$7F,$E4,$1D,$FF,$8B
	.dc.b	$C3,$F6,$7D,$FF,$DC,$37,$FD,$C7,$89,$F7,$27,$F6,$FD,$DF,$83,$7F
	.dc.b	$0E,$7E,$04,$FC,$1C,$FE,$41,$BE,$07,$E4,$39,$3F,$11,$1C,$27,$4D
	.dc.b	$9D,$3C,$CC,$78,$1F,$99,$BF,$17,$3F,$EF,$F1,$FD,$7D,$7F,$06,$EF
	.dc.b	$FC,$63,$FC,$33,$BF,$F1,$8F,$F0,$AE,$EF,$C6,$5F,$F0,$7E,$EF,$C5
	.dc.b	$6B,$FC,$17,$BA,$DF,$EF,$BB,$BF,$16,$FF,$16,$EE,$F8,$9D,$AF,$87
	.dc.b	$FC,$67,$C3,$FE,$37,$E1,$FF,$1D,$F0,$F3,$0F,$C4,$35,$FC,$33,$75
	.dc.b	$FC,$23,$74,$FC,$13,$67,$F2,$D8,$A3,$1C,$BF,$98,$6C,$66,$7B,$90
	.dc.b	$FC,$A6,$1E,$67,$F3,$59,$7E,$8F,$E1,$FF,$49,$F0,$5F,$A5,$F4,$FF
	.dc.b	$4D,$EB,$7E,$9F,$E1,$FF,$51,$EB,$7E,$A7,$E0,$66,$5F,$A3,$F8,$73
	.dc.b	$FE,$17,$FE,$01,$FF,$23,$37,$E0,$B2,$FF,$52,$87,$C4,$63,$93,$FA
	.dc.b	$6C,$3E,$23,$1C,$BD,$39,$53,$9C,$3B,$FA,$BF,$D3,$21,$F1,$1E,$D9
	.dc.b	$ED,$ED,$72,$7F,$7D,$3F,$D5,$9F,$17,$B5,$9C,$3A,$B6,$75,$9F,$92
	.dc.b	$30,$4A,$F7,$35,$E6,$60,$86,$09,$1C,$DC,$B5,$F3,$9F,$C3,$7A,$BC
	.dc.b	$58,$FD,$58,$FE,$21,$CB,$E5,$F7,$FC,$DE,$F7,$9F,$DE,$F4,$7B,$DF
	.dc.b	$52,$FF,$89,$72,$DC,$FE,$6F,$27,$B4,$93,$F5,$F8,$FE,$01,$F9,$CE
	.dc.b	$BF,$85,$EE,$FE,$6E,$7F,$15,$7F,$96,$6F,$2F,$E5,$9A,$7E,$5B,$BF
	.dc.b	$F2,$64,$73,$C9,$F9,$76,$F7,$FE,$5D,$FD,$19,$89,$F0,$3F,$87,$C3
	.dc.b	$F9,$C6,$3F,$57,$DE,$FA,$DF,$11,$F9,$7A,$FE,$69,$DD,$F9,$77,$F9
	.dc.b	$B7,$6F,$DB,$F8,$8F,$B9,$F1,$1F,$D0,$7B,$FE,$F7,$C4,$7D,$FF,$88
	.dc.b	$FC,$07,$83,$31,$FC,$D3,$FC,$87,$BF,$F1,$B1,$F5,$4F,$CF,$78,$1F
	.dc.b	$81,$E5,$FB,$44,$E5,$FC,$C5,$7E,$29,$BE,$06,$DB,$7F,$82,$6B,$F8
	.dc.b	$66,$D7,$38,$FC,$CC,$F1,$69,$C4,$A7,$E2,$8F,$EA,$5E,$07,$E2,$9F
	.dc.b	$EB,$1E,$27,$EF,$9F,$DB,$77,$FE,$29,$FD,$C7,$77,$E2,$9F,$EB,$DD
	.dc.b	$BF,$89,$CB,$FA,$A3,$F8,$88,$FD,$D7,$6D,$CF,$F7,$3D,$BF,$8A,$9F
	.dc.b	$DA,$F4,$7C,$2F,$88,$FD,$67,$C4,$7E,$B7,$E2,$3F,$5D,$F1,$12,$FF
	.dc.b	$84,$35,$FE,$0C,$DC,$7F,$82,$B7,$4F,$9A,$D9,$FD,$EB,$7C,$1F,$E0
	.dc.b	$ED,$E3,$34,$F1,$4F,$F1,$7C,$7F,$80,$65,$FC,$0F,$C9,$FE,$2F,$3F
	.dc.b	$E6,$85,$E8,$E7,$E0,$35,$E9,$EF,$D4,$E1,$F0,$7C,$7E,$1E,$63,$FE
	.dc.b	$B9,$C5,$1D,$7E,$1F,$B7,$C3,$CA,$D7,$C0,$F8,$4F,$C8,$6B,$E3,$F8
	.dc.b	$7F,$27,$AF,$E5,$F5,$E5,$F5,$6B,$E7,$6F,$37,$F9,$E3,$44,$63,$9B
	.dc.b	$F9,$64,$62,$B9,$10,$E4,$5A,$E5,$5C,$63,$FC,$51,$3C,$5D,$9F,$1E
	.dc.b	$FD,$0F,$F4,$3C,$7E,$67,$BB,$F3,$7D,$DF,$9D,$EE,$7C,$AA,$F8,$DE
	.dc.b	$E7,$CA,$7E,$47,$0F,$CA,$7E,$93,$F5,$7E,$1F,$E4,$47,$D5,$E1,$FB
	.dc.b	$1E,$EC,$8F,$C8,$9E,$6F,$77,$F2,$2F,$F5,$5E,$EF,$C1,$3F,$90,$21
	.dc.b	$F9,$02,$F0,$FE,$41,$BC,$5F,$90,$69,$F9,$02,$F7,$7E,$41,$BD,$DF
	.dc.b	$90,$6F,$77,$E4,$1F,$EC,$5D,$DF,$8F,$CB,$F2,$11,$F9,$1E,$EF,$C7
	.dc.b	$3F,$89,$F7,$7F,$1C,$FE,$2F,$B7,$F2,$DF,$0D,$FD,$67,$F8,$07,$FC
	.dc.b	$8E,$3E,$23,$F1,$D5,$FD,$D3,$7F,$F5,$CC,$BE,$23,$87,$C4,$7E,$3E
	.dc.b	$BF,$22,$7C,$8E,$0F,$ED,$FD,$7F,$6A,$3E,$E1,$F8,$3E,$2F,$C4,$F7
	.dc.b	$7E,$0E,$9D,$DF,$83,$CF,$F0,$F7,$B3,$F0,$F5,$B9,$D5,$F2,$9D,$5F
	.dc.b	$2A,$D1,$23,$34,$EE,$F2,$61,$FE,$08,$C7,$F1,$B5,$FA,$FE,$DF,$C6
	.dc.b	$BF,$B1,$F0,$FF,$79,$DB,$F8,$D8,$FC,$59,$F8,$E2,$1F,$BF,$EC,$FC
	.dc.b	$6E,$1F,$A9,$EC,$FC,$64,$FF,$03,$EC,$FF,$32,$71,$3E,$23,$E9,$7C
	.dc.b	$47,$D3,$F8,$89,$7F,$47,$8F,$FC,$35,$69,$CC,$FF,$BD,$23,$0E,$CF
	.dc.b	$F4,$A3,$87,$51,$C3,$2F,$B6,$DA,$EA,$D9,$E2,$DE,$35,$F3,$BE,$35
	.dc.b	$DE,$73,$5B,$A6,$1F,$8D,$C7,$F1,$DF,$11,$F8,$FF,$88,$FC,$87,$BB
	.dc.b	$F9,$1F,$4B,$F2,$5C,$D3,$4E,$99,$7F,$7D,$D7,$F2,$3E,$FF,$E5,$CF
	.dc.b	$CC,$7A,$7F,$99,$F8,$3F,$CD,$7C,$3F,$E6,$FD,$FF,$CE,$7C,$3C,$CB
	.dc.b	$F2,$99,$7C,$F7,$AF,$C0,$7E,$87,$D6,$91,$E0,$6B,$F3,$5B,$A7,$E8
	.dc.b	$9B,$CD,$F5,$1B,$3F,$41,$BC,$DF,$21,$B8,$1A,$63,$0F,$8D,$5F,$03
	.dc.b	$DE,$FF,$26,$7F,$00,$FF,$98,$A3,$EF,$71,$FC,$77,$E8,$47,$F9,$89
	.dc.b	$BE,$BF,$C5,$7E,$4F,$15,$7F,$0D,$C5,$D3,$E1,$FE,$3C,$7C,$EE,$1F
	.dc.b	$8C,$FE,$2F,$87,$FD,$71,$8B,$DD,$FC,$AC,$FD,$4E,$BF,$BC,$47,$04
	.dc.b	$3F,$1E,$5E,$3F,$C3,$B8,$23,$82,$71,$FE,$4A,$38,$35,$DB,$FA,$2E
	.dc.b	$BF,$8B,$7F,$8C,$9F,$C3,$F4,$FC,$8F,$77,$E4,$FB,$BF,$8B,$9F,$D3
	.dc.b	$75,$7C,$CF,$87,$F9,$BE,$F7,$CE,$F8,$7F,$9F,$F0,$DF,$8B,$BF,$9B
	.dc.b	$EF,$FD,$2F,$86,$FA,$7F,$0D,$27,$F1,$7E,$27,$F2,$9E,$DF,$B7,$1F
	.dc.b	$89,$21,$F8,$92,$F1,$FE,$25,$BC,$7F,$89,$69,$F8,$92,$F8,$1F,$89
	.dc.b	$6F,$81,$F8,$96,$F8,$1F,$89,$7F,$4F,$8F,$F3,$BF,$03,$F1,$3F,$05
	.dc.b	$FA,$E9,$FD,$5F,$89,$F8,$CF,$83,$FC,$6F,$85,$F8,$5D,$7F,$5B,$5F
	.dc.b	$88,$3F,$63,$E0,$FE,$1D,$FD,$9F,$89,$FA,$D7,$F6,$BE,$07,$E2,$3E
	.dc.b	$21,$CB,$FE,$AF,$F0,$5F,$0F,$F9,$AF,$87,$FC,$DF,$C3,$CC,$7F,$32
	.dc.b	$D5,$A3,$9B,$94,$E6,$31,$2E,$66,$2A,$D5,$F0,$AE,$A4,$3A,$B9,$69
	.dc.b	$CE,$79,$3B,$BE,$F7,$DA,$FA,$CE,$DB,$1E,$2C,$3F,$6F,$8F,$EE,$3D
	.dc.b	$1F,$DC,$FB,$9F,$BA,$F7,$FF,$77,$EF,$7F,$B1,$FF,$80,$7F,$C8,$BA
	.dc.b	$76,$6E,$DF,$ED,$77,$FC,$8F,$AF,$F2,$D5,$F1,$1C,$4E,$5E,$AF,$3F
	.dc.b	$77,$A7,$C4,$75,$F8,$3E,$DF,$11,$31,$FE,$E9,$C9,$F5,$1F,$E1,$3C
	.dc.b	$5E,$A8,$F8,$31,$F1,$38,$FE,$03,$F0,$F5,$F9,$1B,$7C,$4A,$DB,$79
	.dc.b	$39,$12,$B8,$1F,$1C,$B7,$D1,$ED,$5C,$AF,$A9,$23,$EA,$61,$F3,$1F
	.dc.b	$E3,$FC,$9F,$27,$D7,$F9,$91,$FC,$8B,$8B,$E6,$BF,$1F,$E2,$3E,$77
	.dc.b	$C4,$7C,$78,$F9,$5C,$75,$F7,$72,$38,$FE,$A7,$BB,$F2,$67,$E8,$7B
	.dc.b	$DF,$5C,$3F,$E1,$DA,$7C,$47,$DB,$F7,$FE,$E7,$C4,$4A,$FB,$0D,$7B
	.dc.b	$B7,$97,$F0,$0D,$E2,$FC,$03,$67,$F3,$D8,$23,$1C,$88,$C4,$F6,$21
	.dc.b	$D8,$B7,$D9,$FC,$56,$1F,$91,$C7,$F2,$D1,$FD,$2B,$AF,$F2,$EF,$E3
	.dc.b	$FE,$03,$F2,$CC,$3F,$25,$F0,$DF,$99,$F8,$1F,$CD,$7C,$0F,$E4,$A7
	.dc.b	$F2,$DF,$0F,$F9,$27,$F3,$1F,$0F,$F9,$17,$E7,$F6,$FF,$4C,$EA,$FE
	.dc.b	$09,$1F,$84,$EC,$FC,$13,$F8,$73,$F0,$BD,$BF,$8B,$69,$B7,$1C,$1F
	.dc.b	$91,$8C,$10,$D9,$1F,$CC,$76,$FE,$2E,$3F,$23,$7F,$8B,$7F,$B5,$F8
	.dc.b	$3F,$8A,$7F,$2B,$DB,$FB,$7E,$DF,$C5,$C7,$E2,$88,$7E,$2B,$B3,$F6
	.dc.b	$4F,$ED,$FA,$FF,$14,$FE,$3B,$AF,$F1,$4F,$E2,$7E,$1F,$F1,$6F,$EB
	.dc.b	$E3,$F8,$EB,$7E,$1F,$3A,$FF,$12,$9F,$F3,$A4,$FF,$8B,$DA,$EC,$C4
	.dc.b	$F0,$ED,$FC,$BF,$67,$FA,$E2,$31,$3F,$CA,$6B,$8D,$F5,$9C,$D6,$EB
	.dc.b	$99,$0E,$C4,$9E,$0E,$5A,$E6,$FC,$37,$D9,$A7,$F2,$1C,$24,$FA,$FE
	.dc.b	$1F,$C3,$4F,$87,$BE,$8C,$AF,$81,$F0,$1A,$F8,$1D,$FC,$3F,$0F,$81
	.dc.b	$95,$F1,$7E,$1B,$9F,$A9,$D3,$E0,$3A,$FC,$34,$AF,$A4,$7C,$BE,$0F
	.dc.b	$0F,$1C,$8F,$64,$7A,$A3,$11,$ED,$3E,$53,$8E,$2F,$82,$49,$E1,$3C
	.dc.b	$8D,$97,$8F,$E7,$3C,$DB,$3C,$63,$EB,$4F,$C6,$7F,$8B,$F1,$FC,$6E
	.dc.b	$3F,$8F,$EA,$7C,$59,$FC,$57,$17,$CA,$F8,$7F,$8D,$1F,$87,$F5,$7E
	.dc.b	$2B,$F8,$FE,$1F,$9F,$C3,$F1,$A3,$F1,$9C,$1F,$1D,$A7,$FA,$EF,$99
	.dc.b	$F0,$DF,$5B,$D6,$FA,$FE,$EC,$AF,$CA,$35,$FC,$E3,$78,$7F,$34,$DE
	.dc.b	$1F,$C2,$B6,$7F,$28,$DE,$26,$28,$C3,$67,$B9,$0E,$E7,$E9,$70,$FC
	.dc.b	$5E,$3F,$8C,$F7,$3F,$1B,$EE,$7E,$3B,$AF,$F3,$FF,$C0,$3F,$E6,$DD
	.dc.b	$7F,$BF,$BF,$E6,$BC,$7F,$69,$D9,$F9,$43,$F1,$EF,$FA,$7E,$1E,$DF
	.dc.b	$2F,$88,$E7,$F1,$1D,$3E,$23,$AF,$C4,$7E,$3F,$0F,$DF,$77,$7F,$00
	.dc.b	$F7,$FF,$C4,$F1,$FC,$FB,$BB,$F4,$3C,$DF,$8B,$3F,$13,$1F,$85,$E4
	.dc.b	$FD,$53,$F8,$A3,$1D,$FB,$71,$C1,$F8,$9E,$7F,$C5,$DF,$28,$C5,$F2
	.dc.b	$9C,$D7,$0A,$D1,$1F,$C8,$9F,$E1,$1C,$3F,$25,$7F,$C4,$FB,$FF,$23
	.dc.b	$3F,$C4,$FB,$BF,$25,$3F,$C5,$FB,$BF,$23,$3F,$C7,$7B,$BF,$23,$3F
	.dc.b	$C6,$FB,$BF,$23,$3F,$C6,$FC,$0F,$C8,$3F,$C8,$FB,$AF,$07,$C6,$D9
	.dc.b	$CE,$F8,$8F,$B3,$F1,$1F,$6B,$E2,$26,$3F,$55,$AF,$F8,$6E,$91,$8D
	.dc.b	$3E,$FB,$78,$BF,$93,$CE,$AB,$91,$0E,$53,$B7,$FC,$A5,$1B,$72,$6E
	.dc.b	$BA,$B6,$BA,$DF,$E4,$30,$93,$F8,$BF,$88,$FC,$9F,$C1,$FE,$53,$E2
	.dc.b	$24,$FE,$3F,$E2,$3F,$2F,$E7,$FE,$63,$D0,$FC,$CF,$A3,$F9,$AF,$85
	.dc.b	$FC,$DF,$B7,$F9,$CF,$4F,$F3,$BE,$9F,$F3,$CF,$E0,$1F,$F2,$6E,$7F
	.dc.b	$E6,$97,$FA,$8A,$E6,$F9,$DC,$9F,$36,$3F,$3D,$C5,$EF,$E6,$FA,$3C
	.dc.b	$D6,$BE,$07,$37,$BA,$FE,$09,$1C,$13,$9B,$E6,$D7,$12,$4F,$C5,$3E
	.dc.b	$5B,$87,$EC,$61,$F1,$9F,$B5,$CD,$FB,$8E,$6F,$8F,$1F,$B3,$E5,$F8
	.dc.b	$EF,$DF,$E5,$F8,$1E,$DF,$C7,$8F,$F1,$CE,$7F,$E7,$CD,$FF,$CD,$71
	.dc.b	$FB,$BE,$3F,$8D,$F0,$DF,$1F,$E0,$BE,$47,$C3,$7C,$9A,$87,$F9,$0B
	.dc.b	$8F,$F2,$98,$3E,$1B,$E7,$7B,$7F,$B3,$E5,$FD,$69,$3C,$5F,$D8,$13
	.dc.b	$C6,$D3,$71,$FE,$88,$D7,$F7,$A6,$23,$67,$67,$D7,$CD,$FD,$FB,$1F
	.dc.b	$14,$60,$FB,$DF,$0D,$F8,$1F,$6F,$F0,$5F,$0D,$F8,$3F,$6F,$F0,$99
	.dc.b	$7E,$17,$E0,$FF,$1B,$87,$1D,$19,$3E,$13,$F1,$F3,$98,$FF,$28,$BF
	.dc.b	$AD,$F0,$3F,$20,$FF,$25,$EF,$FC,$87,$C2,$48,$F9,$1E,$2F,$EB,$BB
	.dc.b	$BF,$73,$1F,$81,$CB,$F1,$AD,$F0,$7F,$1B,$C5,$F8,$88,$C7,$83,$F8
	.dc.b	$5E,$CF,$C6,$47,$E2,$0C,$90,$FC,$94,$F1,$69,$FB,$8F,$07,$F1,$93
	.dc.b	$FD,$03,$1F,$C6,$4F,$F3,$DC,$BF,$17,$3F,$D1,$3C,$0F,$E8,$FE,$07
	.dc.b	$E3,$2B,$FA,$17,$89,$FA,$DF,$7F,$F6,$51,$F1,$7C,$0F,$C5,$3F,$1F
	.dc.b	$BF,$F1,$2F,$FA,$B7,$93,$F6,$3D,$FF,$8A,$F8,$4F,$DB,$7C,$24,$AF
	.dc.b	$9C,$D7,$83,$75,$FC,$D3,$78,$7F,$04,$D9,$FD,$F3,$79,$3E,$DB,$71
	.dc.b	$34,$93,$F7,$70,$FE,$11,$F0,$9F,$C2,$7E,$13,$F8,$57,$A5,$FE,$46
	.dc.b	$FE,$01,$FF,$2B,$E9,$FE,$E6,$7F,$A6,$0F,$4F,$5E,$DE,$FD,$4E,$1E
	.dc.b	$DF,$1F,$84,$E5,$F0,$9C,$FD,$5E,$9F,$07,$D7,$D6,$ED,$EB,$4E,$2F
	.dc.b	$C4,$9F,$F0,$07,$57,$B3,$E3,$F8,$49,$3F,$99,$94,$62,$3F,$DB,$91
	.dc.b	$87,$51,$9E,$0F,$53,$7A,$3F,$CD,$35,$BA,$E3,$39,$7C,$4D,$E8,$4E
	.dc.b	$EF,$9D,$76,$21,$C8,$F3,$61,$F3,$9F,$D6,$72,$7C,$E7,$D5,$ED,$7C
	.dc.b	$FF,$43,$E8,$72,$FC,$C9,$F7,$72,$7C,$C7,$E0,$7B,$5F,$57,$E0,$BE
	.dc.b	$B7,$C1,$FC,$C9,$F8,$9F,$09,$F2,$DF,$87,$C5,$FE,$13,$F1,$BF,$85
	.dc.b	$76,$FE,$0A,$7F,$1E,$7E,$48,$86,$09,$C1,$F9,$27,$2D,$73,$C9,$F9
	.dc.b	$47,$2D,$72,$7E,$58,$C9,$4C,$F1,$FF,$12,$AF,$C8,$7C,$27,$E4,$7E
	.dc.b	$13,$FD,$2F,$E4,$7F,$8E,$70,$FA,$E6,$3E,$23,$3F,$0B,$AF,$85,$DF
	.dc.b	$C2,$F0,$F8,$5F,$C9,$6B,$F8,$7F,$03,$F2,$4F,$F2,$1C,$9F,$0B,$F9
	.dc.b	$38,$C7,$9F,$E0,$3E,$43,$F9,$3E,$DF,$C1,$77,$7E,$07,$6F,$C4,$F7
	.dc.b	$E0,$95,$F8,$2E,$4F,$C2,$BF,$83,$23,$82,$4F,$28,$E0,$D7,$6D,$8D
	.dc.b	$C7,$8C,$FE,$6D,$E3,$FE,$E5,$FE,$71,$E3,$FE,$E7,$C0,$FC,$64,$7E
	.dc.b	$9F,$BF,$F1,$8F,$E9,$BB,$FF,$18,$FE,$A3,$BF,$F1,$8F,$EB,$7B,$BF
	.dc.b	$18,$FE,$B3,$BB,$F1,$99,$FF,$03,$EE,$F8,$BD,$AE,$47,$A1,$F6,$3D
	.dc.b	$19,$8F,$CF,$6A,$C1,$C3,$FE,$DA,$BF,$E8,$72,$C3,$A5,$F0,$BD,$CE
	.dc.b	$6B,$75,$A8,$EC,$43,$B5,$F5,$BE,$4D,$8D,$5F,$4A,$E5,$1C,$0E,$D5
	.dc.b	$C1,$A7,$E2,$70,$FC,$AF,$C1,$7E,$5B,$C1,$FC,$BE,$5F,$98,$F4,$BF
	.dc.b	$33,$F0,$BF,$9A,$F4,$FF,$37,$F0,$BF,$9C,$F8,$3F,$CE,$FB,$3F,$9E
	.dc.b	$F8,$5F,$CF,$FB,$5F,$EF,$6F,$E0,$1F,$F1,$DF,$09,$7C,$36,$7D,$AD
	.dc.b	$7B,$53,$8F,$F9,$4D,$EF,$E1,$B9,$7B,$72,$7F,$C0,$18,$A3,$11,$EF
	.dc.b	$AE,$D1,$8E,$7F,$89,$5B,$AE,$33,$9E,$84,$7D,$79,$D3,$CD,$B5,$CA
	.dc.b	$B8,$4E,$DB,$7E,$08,$EA,$43,$94,$7C,$3C,$3E,$27,$C3,$7C,$5F,$86
	.dc.b	$F8,$DF,$0D,$F1,$FE,$1B,$E4,$7C,$37,$C9,$F8,$6F,$9F,$97,$C0,$F8
	.dc.b	$6F,$99,$F0,$DF,$3E,$3E,$1F,$C3,$7C,$FF,$89,$FA,$1C,$BF,$36,$7E
	.dc.b	$3F,$C3,$FD,$33,$F3,$FD,$FF,$9E,$8F,$C4,$F2,$FE,$2D,$F9,$51,$F9
	.dc.b	$42,$78,$7F,$89,$E4,$FC,$6F,$2F,$E3,$E7,$35,$C8,$FC,$69,$CA,$B8
	.dc.b	$46,$A3,$06,$EB,$9D,$F8,$6F,$A7,$87,$F9,$4A,$FE,$3D,$E1,$7E,$51
	.dc.b	$FE,$41,$E1,$7E,$51,$FE,$45,$E0,$FE,$3F,$DA,$FC,$A4,$7F,$25,$F0
	.dc.b	$3F,$25,$F0,$9F,$94,$8F,$C4,$9F,$95,$3F,$2D,$E7,$7E,$45,$FC,$07
	.dc.b	$7F,$F0,$4E,$F7,$77,$DE,$7F,$07,$BF,$E2,$38,$3F,$01,$7F,$78,$FC
	.dc.b	$07,$07,$E1,$DB,$3B,$AE,$31,$F8,$93,$25,$39,$12,$79,$36,$79,$77
	.dc.b	$F8,$CF,$26,$1F,$8C,$9F,$D2,$F7,$7E,$31,$FD,$3F,$6F,$E3,$1F,$D4
	.dc.b	$F6,$7E,$31,$FC,$F7,$67,$E3,$1F,$E9,$9E,$17,$EE,$7B,$3F,$19,$1F
	.dc.b	$B0,$EB,$FC,$63,$FA,$5E,$BF,$C6,$3F,$B8,$DD,$EF,$7F,$01,$F3,$E4
	.dc.b	$FE,$E1,$AF,$EB,$5B,$C3,$FB,$66,$F0,$FC,$76,$CF,$ED,$5B,$C5,$FC
	.dc.b	$2D,$B8,$1A,$49,$FE,$0F,$87,$EF,$9F,$E0,$DF,$C0,$3F,$E4,$4C,$3E
	.dc.b	$D2,$F7,$F5,$97,$F2,$B8,$FF,$23,$57,$FA,$3D,$7C,$2E,$AF,$8F,$A1
	.dc.b	$6B,$FA,$05,$3E,$17,$1A,$A1,$F0,$B8,$5E,$D5,$F8,$5C,$7C,$2E,$7E
	.dc.b	$17,$5E,$9E,$FD,$BE,$1E,$D7,$1F,$6A,$6D,$FC,$A3,$37,$B3,$F0,$63
	.dc.b	$E7,$1F,$44,$86,$09,$C7,$F4,$5C,$B5,$CF,$17,$D3,$72,$D7,$27,$D5
	.dc.b	$32,$53,$3C,$1F,$06,$BE,$07,$C2,$FC,$1F,$6B,$E1,$7B,$5F,$0F,$E1
	.dc.b	$7E,$27,$C2,$FC,$8F,$E0,$1F,$F1,$46,$5F,$33,$9B,$E4,$19,$F8,$6F
	.dc.b	$8F,$1F,$57,$93,$E3,$BF,$EC,$5A,$7C,$37,$3F,$86,$E9,$F0,$D2,$BF
	.dc.b	$31,$DF,$FA,$B7,$F6,$7C,$3F,$BC,$E3,$FD,$89,$38,$FF,$63,$3A,$F0
	.dc.b	$7F,$77,$5B,$9C,$CF,$ED,$CF,$DE,$57,$16,$C6,$A3,$13,$C6,$72,$7F
	.dc.b	$53,$BF,$C3,$57,$E1,$FC,$0F,$C4,$7C,$17,$E2,$7C,$1F,$C5,$78,$3F
	.dc.b	$8B,$F0,$7E,$47,$C3,$7E,$37,$C2,$FC,$66,$7F,$0F,$E1,$BE,$67,$C3
	.dc.b	$7C,$DF,$63,$F1,$93,$F9,$DF,$07,$F1,$6F,$E9,$BC,$1F,$F1,$2F,$83
	.dc.b	$F8,$C8,$FB,$BC,$5F,$B1,$6F,$27,$EC,$5B,$E3,$7E,$47,$97,$F1,$44
	.dc.b	$E4,$FC,$43,$67,$F0,$CD,$BE,$91,$F8,$8B,$E8,$DB,$FD,$47,$85,$F8
	.dc.b	$A8,$FD,$A7,$83,$F8,$A7,$F6,$DE,$07,$E2,$9F,$C1,$7C,$67,$E2,$1F
	.dc.b	$DD,$78,$1F,$89,$7F,$79,$E3,$7E,$BD,$FD,$F7,$8D,$F9,$1E,$FF,$C4
	.dc.b	$47,$E2,$FB,$BF,$10,$BE,$25,$ED,$7F,$43,$F0,$BF,$C7,$64,$F7,$3F
	.dc.b	$34,$4A,$FC,$A3,$5F,$F0,$C4,$6D,$C9,$0C,$C6,$F8,$FF,$C7,$D3,$B9
	.dc.b	$C9,$B8,$CC,$EF,$93,$C8,$FE,$AF,$E1,$FF,$59,$87,$EB,$7D,$DF,$D7
	.dc.b	$7B,$BF,$AF,$F8,$7F,$D8,$7B,$DF,$B1,$F8,$79,$A7,$EA,$FE,$1F,$F6
	.dc.b	$9E,$FF,$ED,$7D,$6F,$DB,$7C,$3F,$F1,$EF,$4F,$FC,$7E,$7F,$8F,$AF
	.dc.b	$FC,$5B,$C3,$FE,$55,$87,$C3,$EF,$DF,$95,$FE,$40,$9F,$ED,$EF,$0E
	.dc.b	$3F,$ED,$D1,$88,$FF,$1F,$36,$FA,$7B,$DE,$12,$4F,$13,$FC,$E5,$3B
	.dc.b	$AE,$73,$A9,$E6,$DA,$E7,$1C,$7C,$2F,$F7,$CC,$E6,$B8,$CE,$6F,$A3
	.dc.b	$DA,$74,$D3,$4F,$D1,$78,$7B,$FF,$1F,$E1,$FE,$47,$C0,$7D,$19,$E9
	.dc.b	$F0,$FF,$41,$FC,$2F,$17,$CD,$F8,$1F,$A3,$1F,$C7,$78,$BE,$8B,$E3
	.dc.b	$3E,$97,$C3,$FD,$3E,$1F,$A9,$F0,$52,$7F,$CD,$BD,$BF,$89,$E2,$FC
	.dc.b	$3C,$7E,$27,$87,$F0,$ED,$3F,$D3,$53,$B8,$CC,$7E,$40,$93,$C9,$B1
	.dc.b	$A8,$C1,$1C,$12,$3A,$F6,$FE,$5D,$A7,$F3,$9E,$CF,$CB,$3F,$CE,$FA
	.dc.b	$FF,$2E,$FF,$3D,$EA,$FC,$BB,$F4,$FE,$0B,$FA,$47,$5C,$8F,$E8,$5D
	.dc.b	$7F,$95,$F8,$7F,$CB,$7C,$1F,$E5,$FE,$0F,$F3,$1F,$0F,$F9,$2B,$FC
	.dc.b	$6F,$77,$FB,$13,$B7,$EE,$F7,$7E,$0A,$7E,$D9,$0F,$C4,$97,$8B,$F1
	.dc.b	$2E,$08,$E0,$9C,$5F,$8A,$8C,$B6,$33,$1F,$89,$32,$57,$33,$E3,$8F
	.dc.b	$E1,$3E,$17,$EC,$7B,$BF,$19,$1F,$95,$ED,$FC,$63,$FA,$8E,$DF,$C6
	.dc.b	$61,$FC,$2B,$B7,$F1,$93,$F8,$B2,$1F,$C5,$3C,$0F,$DE,$3F,$8C,$3F
	.dc.b	$41,$D7,$F8,$B7,$F6,$7D,$7F,$1F,$E1,$3F,$81,$12,$3F,$4C,$D7,$F7
	.dc.b	$2D,$C3,$F8,$23,$76,$FD,$FB,$67,$F7,$2D,$D9,$BA,$30,$D9,$EE,$43
	.dc.b	$B9,$6F,$F8,$97,$C3,$FF,$17,$F8,$7F,$E3,$1F,$0F,$FC,$67,$E0,$3F
	.dc.b	$8D,$7A,$5F,$E3,$8F,$E0,$1F,$F2,$36,$3F,$EF,$B6,$E8,$7F,$C9,$11
	.dc.b	$FC,$1D,$FF,$23,$BF,$ED,$36,$E5,$FE,$F5,$7F,$C9,$31,$BF,$7B,$B7
	.dc.b	$BD,$DF,$D1,$F0,$7B,$BE,$1F,$88,$97,$FC,$97,$87,$DE,$47,$D9,$C5
	.dc.b	$F2,$1A,$7C,$82,$E9,$F2,$27,$84,$7C,$42,$69,$F4,$9D,$F1,$7C,$B7
	.dc.b	$37,$D5,$F6,$7A,$7F,$17,$E1,$3E,$4C,$7F,$23,$E1,$F9,$1E,$A7,$C9
	.dc.b	$C3,$E5,$7B,$9F,$2F,$E0,$FE,$67,$BB,$F3,$7D,$BF,$9D,$F0,$5F,$3F
	.dc.b	$E2,$3E,$1E,$7F,$2F,$93,$E1,$BF,$37,$DD,$FA,$87,$F9,$EF,$D9,$FE
	.dc.b	$E3,$1F,$57,$E2,$3E,$C1,$1F,$A8,$D7,$EA,$37,$8F,$EA,$37,$8F,$FD
	.dc.b	$79,$D4,$8C,$57,$53,$99,$0E,$65,$AE,$6D,$AE,$73,$CB,$9B,$F3,$3D
	.dc.b	$DF,$91,$AF,$E9,$5D,$DF,$8F,$F7,$BF,$23,$1F,$8B,$C3,$F2,$47,$E4
	.dc.b	$FE,$23,$F2,$9F,$0D,$27,$FA,$B7,$F0,$0F,$F9,$9B,$2F,$CA,$78,$1F
	.dc.b	$E9,$08,$7C,$4D,$F8,$1C,$7C,$4F,$E3,$30,$FC,$B7,$83,$FB,$97,$F0
	.dc.b	$3E,$0F,$E1,$B8,$BF,$09,$1F,$82,$3F,$18,$5C,$FF,$18,$F0,$21,$9E
	.dc.b	$DF,$C5,$4F,$22,$6F,$F8,$B6,$CF,$EB,$7C,$1F,$C6,$3F,$DB,$7C,$5F
	.dc.b	$DA,$3F,$AD,$F0,$3F,$18,$FE,$BF,$C0,$FC,$63,$FB,$1E,$FF,$C6,$3F
	.dc.b	$B3,$EE,$FC,$63,$FB,$5F,$13,$F5,$CF,$ED,$FC,$4F,$C5,$BF,$8D,$07
	.dc.b	$F8,$F7,$D8,$FF,$7F,$72,$BE,$0B,$E6,$12,$7E,$8B,$56,$CD,$F8,$CE
	.dc.b	$63,$12,$E1,$5F,$58,$E6,$43,$9D,$F2,$71,$5C,$AB,$8D,$F5,$AD,$57
	.dc.b	$03,$B5,$F5,$71,$5C,$1E,$73,$C2,$FA,$56,$DB,$F1,$3F,$8B,$F8,$9F
	.dc.b	$C6,$7C,$4F,$E3,$7E,$27,$F1,$DE,$F7,$E3,$FD,$4F,$C8,$6D,$F9,$1F
	.dc.b	$86,$FC,$97,$BF,$36,$FC,$BF,$C3,$7E,$57,$E0,$3F,$2D,$EA,$FE,$5F
	.dc.b	$DF,$FC,$C7,$C1,$FE,$67,$E0,$26,$1F,$53,$CD,$FD,$FF,$37,$C8,$F6
	.dc.b	$3F,$3C,$49,$FD,$1B,$5F,$7F,$32,$31,$CD,$EE,$6F,$3F,$DE,$6D,$7C
	.dc.b	$86,$DF,$78,$EA,$43,$AC,$6A,$FB,$46,$6F,$E4,$BF,$B6,$E6,$FC,$6B
	.dc.b	$F4,$BF,$80,$6B,$E6,$E6,$F9,$6F,$EF,$39,$77,$7F,$7E,$BF,$CD,$93
	.dc.b	$C7,$D9,$CE,$9F,$98,$D1,$4F,$77,$37,$F8,$F8,$FF,$0D,$9F,$ED,$81
	.dc.b	$F0,$DF,$4A,$3F,$85,$F1,$FD,$27,$F8,$6F,$15,$7F,$48,$7E,$33,$B3
	.dc.b	$F2,$0D,$3F,$16,$53,$E8,$BA,$C7,$F2,$11,$83,$6E,$34,$FC,$8C,$EF
	.dc.b	$83,$F2,$5A,$7E,$54,$97,$D6,$B6,$F2,$9D,$BE,$FF,$77,$E0,$E1,$F1
	.dc.b	$3D,$DF,$CC,$CF,$F9,$F7,$57,$C3,$69,$7C,$35,$F7,$FF,$37,$5F,$CB
	.dc.b	$BB,$7F,$34,$FE,$33,$B7,$F9,$77,$6F,$E6,$E3,$A7,$C2,$FE,$6D,$ED
	.dc.b	$F0,$BF,$9A,$7F,$27,$DB,$FD,$67,$C0,$FC,$84,$7E,$1F,$8E,$BF,$8C
	.dc.b	$30,$7E,$4C,$BC,$1F,$93,$74,$4E,$3F,$CE,$3A,$E3,$FC,$9B,$9A,$EB
	.dc.b	$3A,$8C,$1D,$4B,$5D,$63,$F3,$31,$FA,$AE,$CF,$8C,$FF,$01,$3F,$A8
	.dc.b	$F8,$3F,$C0,$5F,$99,$F0,$9F,$2F,$B3,$E6,$76,$7C,$DF,$87,$F9,$DF
	.dc.b	$09,$F3,$FB,$BF,$21,$87,$ED,$BB,$BF,$1D,$A7,$F1,$2E,$EF,$C8,$78
	.dc.b	$5F,$C0,$3A,$DF,$0D,$F5,$C9,$87,$DB,$6B,$FA,$D6,$DF,$EA,$5B,$C5
	.dc.b	$FA,$7A,$E7,$5C,$48,$71,$2D,$71,$6C,$F1,$AE,$71,$FA,$D6,$F1,$7E
	.dc.b	$CB,$E1,$BF,$19,$E9,$7E,$37,$DC,$FC,$77,$B9,$F8,$FF,$4B,$F2,$1F
	.dc.b	$09,$F9,$1F,$86,$FC,$BF,$6F,$F9,$38,$9F,$0B,$8F,$77,$3E,$9C,$DF
	.dc.b	$FD,$AD,$4F,$77,$8F,$A9,$CB,$E0,$E5,$7F,$29,$F0,$BF,$83,$5B,$D5
	.dc.b	$EE,$49,$E6,$D7,$FA,$3C,$6D,$C9,$0C,$C6,$F9,$BF,$AA,$4E,$E7,$26
	.dc.b	$E3,$33,$BE,$6F,$1B,$67,$FB,$25,$71,$AD,$C7,$1A,$F4,$FB,$7F,$32
	.dc.b	$3E,$0F,$B5,$F3,$1F,$87,$F0,$DF,$31,$F8,$BE,$DF,$CC,$F7,$3E,$6F
	.dc.b	$B9,$F3,$BD,$CF,$99,$5F,$2B,$8B,$E6,$3F,$03,$E1,$BE,$5B,$F3,$BE
	.dc.b	$1B,$EA,$9F,$B9,$7F,$00,$7E,$10,$A7,$E2,$49,$B7,$E2,$5B,$B7,$E2
	.dc.b	$5E,$1D,$DF,$CE,$3B,$FF,$13,$E0,$7E,$52,$76,$74,$BE,$44,$BE,$51
	.dc.b	$AB,$E4,$F4,$A7,$81,$F8,$CF,$86,$FC,$94,$FE,$23,$E1,$7F,$24,$FE
	.dc.b	$2B,$E1,$7F,$24,$FE,$33,$1F,$C9,$9F,$94,$ED,$FC,$AF,$C2,$FF,$2C
	.dc.b	$FE,$01,$FF,$28,$1F,$E4,$EA,$FC,$9F,$76,$7E,$17,$F2,$31,$F8,$AE
	.dc.b	$EF,$91,$BB,$B3,$EF,$67,$F8,$2A,$FB,$C7,$E0,$BB,$7F,$09,$5B,$6D
	.dc.b	$70,$9D,$1C,$27,$4D,$8D,$46,$23,$F2,$64,$E2,$FC,$6B,$6F,$C4,$D3
	.dc.b	$F2,$7D,$BF,$8F,$7F,$47,$DB,$F8,$F7,$F4,$BD,$BF,$8F,$7F,$4F,$DB
	.dc.b	$F8,$F7,$F6,$BD,$9F,$8F,$7F,$6F,$E0,$7E,$3B,$2F,$D3,$F6,$7F,$00
	.dc.b	$EC,$FC,$7D,$7E,$C3,$B3,$F1,$EF,$EE,$4F,$F2,$47,$23,$E1,$7E,$91
	.dc.b	$23,$FC,$B3,$87,$F9,$16,$75,$18,$DF,$FC,$0D,$18,$8F,$F2,$1B,$67
	.dc.b	$85,$73,$9C,$B6,$33,$1C,$CF,$1C,$66,$BA,$B6,$32,$6F,$81,$7D,$EB
	.dc.b	$73,$98,$C3,$C0,$F2,$FB,$DF,$8F,$F8,$5F,$C8,$7A,$9F,$91,$F8,$5F
	.dc.b	$C9,$7C,$17,$E4,$FD,$6F,$CA,$7C,$17,$E5,$7C,$79,$97,$5F,$77,$F3
	.dc.b	$1E,$EF,$E6,$7D,$D9,$3E,$1F,$85,$FC,$E7,$AB,$F9,$DF,$85,$93,$F4
	.dc.b	$78,$FD,$3C,$9E,$D6,$C7,$A6,$31,$C7,$EA,$9C,$CF,$CA,$34,$43,$04
	.dc.b	$E5,$F9,$AD,$E5,$F8,$AD,$3E,$24,$68,$91,$A7,$F0,$1C,$BF,$B5,$F5
	.dc.b	$BF,$6D,$EF,$7C,$7B,$F9,$DC,$BF,$B9,$F8,$5F,$F1,$2F,$EE,$BF,$E5
	.dc.b	$2C,$7F,$D1,$8F,$F9,$39,$BE,$F7,$EF,$F9,$3E,$4E,$1F,$5F,$8F,$E4
	.dc.b	$BF,$E8,$CB,$FF,$28,$3F,$7D,$7B,$BF,$2A,$3F,$1F,$C5,$FC,$0A,$5E
	.dc.b	$EF,$88,$93,$C5,$AF,$56,$E5,$E6,$6F,$07,$F9,$86,$B9,$D7,$12,$1C
	.dc.b	$4B,$5C,$5B,$3C,$6B,$9C,$7F,$AD,$6B,$9B,$C4,$F7,$7B,$BF,$93,$7F
	.dc.b	$31,$D9,$F2,$BE,$1B,$F2,$71,$F9,$BE,$CF,$C9,$BF,$9D,$EB,$FC,$9B
	.dc.b	$FB,$07,$E8,$FC,$37,$D2,$F8,$6F,$A7,$F0,$DF,$CB,$7B,$7F,$27,$7F
	.dc.b	$2F,$E1,$BE,$B9,$FC,$AB,$B3,$F1,$C4,$ED,$FC,$89,$27,$F1,$DC,$5F
	.dc.b	$86,$7F,$13,$C5,$F8,$72,$38,$27,$6F,$E6,$27,$83,$67,$6D,$9D,$DF
	.dc.b	$F9,$C2,$3F,$7B,$D9,$F9,$09,$FC,$79,$0F,$E1,$FD,$5F,$90,$7F,$23
	.dc.b	$F0,$9F,$92,$EB,$FC,$74,$7F,$03,$F0,$3F,$1B,$9F,$F1,$BE,$BF,$CB
	.dc.b	$FC,$27,$E6,$3B,$3F,$19,$3F,$C7,$FB,$3F,$19,$F0,$DF,$9C,$26,$3F
	.dc.b	$93,$6B,$F9,$76,$CF,$E8,$1B,$C1,$F8,$B6,$CF,$E0,$DB,$BB,$B1,$18
	.dc.b	$6C,$F7,$21,$DC,$B7,$FA,$EF,$3F,$F6,$3E,$DF,$EC,$BD,$AF,$D9,$FC
	.dc.b	$24,$BF,$D5,$7C,$27,$ED,$BE,$13,$F6,$FF,$09,$FB,$8F,$84,$95,$FA
	.dc.b	$FF,$84,$FF,$4F,$FE,$D2,$E1,$A6,$3D,$EC,$FA,$9A,$F5,$26,$1F,$C7
	.dc.b	$78,$BD,$2F,$9C,$9C,$5F,$15,$F4,$E1,$F1,$5A,$7C,$52,$F2,$7C,$5E
	.dc.b	$4F,$88,$48,$CD,$F6,$BE,$91,$DA,$FA,$34,$5B,$3F,$F2,$F4,$7F,$4B
	.dc.b	$5F,$0B,$F2,$63,$F8,$87,$17,$C9,$7F,$8A,$71,$7C,$97,$F8,$A7,$0F
	.dc.b	$C6,$F7,$3E,$3F,$B9,$FC,$8B,$8B,$E3,$D7,$5D,$2A,$D7,$E6,$3E,$4D
	.dc.b	$1E,$E7,$D0,$F7,$24,$FF,$47,$97,$B9,$F5,$09,$1E,$96,$BD,$70,$FF
	.dc.b	$7D,$4F,$08,$D1,$0D,$1E,$19,$E3,$87,$93,$83,$E3,$46,$2F,$9B,$6B
	.dc.b	$81,$CE,$B8,$5F,$78,$EB,$7C,$9D,$5F,$27,$F9,$E7,$57,$E6,$1F,$A5
	.dc.b	$EE,$7E,$43,$E1,$7F,$23,$EE,$FE,$4B,$E1,$7F,$27,$EF,$7E,$53,$DE
	.dc.b	$FC,$BE,$1F,$D1,$FB,$7F,$2F,$F0,$9F,$98,$F8,$4F,$CC,$FC,$27,$E5
	.dc.b	$EB,$EF,$77,$FE,$59,$F2,$F0,$BB,$FF,$7F,$1F,$8C,$EF,$FC,$73,$F8
	.dc.b	$9E,$3F,$C9,$39,$21,$9E,$FF,$C9,$72,$7E,$43,$97,$F2,$64,$70,$49
	.dc.b	$FC,$F1,$D7,$1F,$C9,$BE,$4E,$FF,$C9,$3F,$D8,$FB,$FF,$25,$7F,$DB
	.dc.b	$3B,$FF,$25,$3F,$D9,$FB,$FF,$24,$FE,$6F,$BB,$F2,$4F,$E4,$48,$7C
	.dc.b	$0C,$BF,$7B,$DB,$F8,$E8,$F8,$7D,$BF,$E6,$CD,$1F,$09,$FC,$08,$93
	.dc.b	$FA,$76,$BF,$29,$B9,$FE,$09,$B8,$7E,$49,$B3,$F9,$F6,$EF,$FB,$36
	.dc.b	$DF,$82,$7B,$10,$EC,$5B,$FC,$B7,$92,$F8,$B4,$FF,$91,$63,$FD,$14
	.dc.b	$FE,$49,$7C,$26,$A3,$1E,$8D,$BE,$6D,$CF,$9E,$6C,$3F,$AA,$0F,$84
	.dc.b	$D6,$1F,$C3,$21,$ED,$E2,$B8,$7B,$F6,$5F,$09,$7D,$CC,$7C,$26,$7D
	.dc.b	$D9,$97,$A7,$E0,$78,$61,$23,$F8,$17,$1F,$C6,$23,$F0,$F6,$F9,$CD
	.dc.b	$3E,$71,$74,$F9,$CD,$E3,$F9,$CF,$CC,$27,$0F,$D9,$75,$7D,$9C,$DF
	.dc.b	$C4,$C3,$E6,$D7,$D2,$E2,$F7,$71,$4B,$FA,$5F,$07,$F0,$7D,$5F,$85
	.dc.b	$C7,$F0,$F8,$FE,$27,$B9,$F1,$7E,$13,$E3,$7A,$BF,$0B,$1F,$B1,$C7
	.dc.b	$F2,$78,$FE,$14,$7D,$BD,$FF,$6E,$FE,$C7,$C1,$FD,$BF,$07,$EB,$E3
	.dc.b	$F5,$AD,$E4,$FD,$CC,$6C,$FD,$CC,$67,$C1,$FD,$CC,$70,$E3,$FD,$B5
	.dc.b	$70,$AD,$5F,$DB,$EA,$FB,$9E,$DF,$DD,$F8,$4F,$BD,$EE,$7E,$07,$F8
	.dc.b	$07,$FC,$C5,$97,$E2,$9F,$F3,$13,$F2,$7B,$B5,$EE,$6F,$DC,$FC,$37
	.dc.b	$81,$F8,$5D,$FF,$35,$DF,$CF,$DC,$E9,$F0,$BD,$7E,$17,$B7,$C2,$CB
	.dc.b	$FC,$29,$F6,$08,$7D,$83,$F0,$65,$E4,$FC,$1B,$7C,$3F,$C1,$F2,$FE
	.dc.b	$06,$31,$E1,$FE,$06,$30,$4E,$0F,$C7,$CE,$8F,$D8,$7C,$57,$E7,$FC
	.dc.b	$4F,$C1,$47,$EB,$8F,$C1,$9F,$B5,$F0,$BF,$06,$FE,$DF,$C1,$FC,$1B
	.dc.b	$FB,$9F,$03,$F0,$6F,$EE,$FB,$FF,$09,$97,$F0,$0E,$FF,$C1,$BF,$C0
	.dc.b	$BB,$FF,$79,$F1,$2A,$FA,$1F,$0F,$F4,$C9,$87,$FB,$27,$AD,$18,$8F
	.dc.b	$EA,$F1,$87,$51,$9C,$3F,$B2,$4E,$67,$73,$9D,$FB,$C7,$02,$1A,$2D
	.dc.b	$72,$9E,$11,$FE,$14,$AE,$93,$98,$F3,$7C,$3F,$E2,$FE,$1F,$F1,$9F
	.dc.b	$0F,$27,$D5,$E1,$FE,$3F,$DE,$FC,$87,$C2,$FE,$47,$E1,$7F,$25,$E9
	.dc.b	$FE,$4F,$D3,$FC,$A7,$C2,$FE,$57,$D4,$FC,$B7,$B7,$F9,$7F,$83,$FE
	.dc.b	$31,$FC,$03,$FE,$14,$D7,$FB,$CA,$F8,$6C,$9F,$3B,$9B,$DF,$3E,$D2
	.dc.b	$71,$FA,$5B,$CD,$A8,$C1,$1C,$12,$3E,$17,$37,$C4,$F0,$5C,$FF,$12
	.dc.b	$B5,$18,$9D,$4E,$6B,$89,$F3,$AB,$70,$E7,$F9,$53,$F8,$1E,$6F,$94
	.dc.b	$FE,$03,$9B,$F7,$07,$C2,$F6,$FE,$54,$7E,$1F,$93,$E5,$3F,$D3,$B4
	.dc.b	$FF,$13,$6F,$FE,$09,$8C,$67,$FD,$22,$3F,$BD,$CF,$E2,$38,$7E,$75
	.dc.b	$68,$FF,$29,$CF,$CF,$9E,$7E,$D7,$D2,$3E,$74,$EB,$E1,$BE,$A9,$1E
	.dc.b	$5C,$0B,$4E,$67,$EC,$B6,$FC,$ED,$C7,$CC,$D9,$EC,$74,$AE,$6D,$8D
	.dc.b	$5F,$76,$CE,$A7,$B9,$B3,$97,$B7,$F8,$CF,$6F,$F1,$BE,$DF,$E3,$BE
	.dc.b	$1B,$F1,$FF,$0D,$F9,$9B,$FD,$A7,$6F,$E4,$BE,$1B,$F2,$7F,$0D,$F9
	.dc.b	$4F,$86,$FC,$CD,$7E,$E7,$B7,$F2,$FF,$0D,$F9,$8F,$86,$FC,$CF,$C3
	.dc.b	$7E,$5A,$BD,$5D,$FF,$2B,$8F,$F1,$11,$F8,$53,$F0,$E7,$E5,$8B,$CB
	.dc.b	$F9,$56,$F2,$FE,$2B,$BB,$F1,$E4,$8E,$55,$D1,$E4,$DA,$E4,$DA,$FC
	.dc.b	$9F,$6F,$E4,$AB,$F8,$7F,$6F,$E4,$AF,$F8,$6F,$6F,$E4,$A7,$E9,$76
	.dc.b	$FE,$49,$FE,$2B,$DB,$F9,$27,$F6,$BF,$09,$FB,$8F,$89,$FD,$DC,$7F
	.dc.b	$1D,$F0,$BF,$77,$A7,$E2,$E3,$EE,$77,$FC,$3E,$E7,$C3,$FF,$00,$24
	.dc.b	$FE,$05,$AF,$F0,$26,$EB,$FA,$06,$DF,$E7,$DB,$3F,$9C,$6F,$07,$E5
	.dc.b	$5B,$7E,$09,$EC,$43,$B1,$6F,$F5,$BF,$0F,$FC,$53,$D2,$FF,$16,$7F
	.dc.b	$00,$FF,$8F,$EB,$F4,$2D,$7F,$D1,$EF,$F8,$53,$2F,$F1,$CB,$FE,$40
	.dc.b	$8D,$FA,$5C,$BE,$1F,$9F,$C3,$F4,$F7,$E5,$74,$F8,$7E,$FE,$AF,$83
	.dc.b	$DE,$F0,$FC,$47,$8B,$DF,$97,$FE,$50,$F5,$7F,$A8,$3F,$04,$F5,$6D
	.dc.b	$F1,$5A,$64,$F7,$93,$87,$E6,$3A,$E4,$F9,$0D,$E3,$F9,$0E,$EB,$A3
	.dc.b	$4E,$95,$C9,$C5,$72,$6F,$0F,$99,$FE,$7B,$C5,$F3,$5F,$97,$EB,$FC
	.dc.b	$EF,$5F,$E7,$FB,$3F,$43,$E2,$3E,$45,$7D,$0F,$6B,$E3,$BF,$43,$93
	.dc.b	$EA,$F2,$7D,$6F,$88,$FA,$FF,$0D,$F6,$3E,$1B,$EC,$FC,$34,$C3,$E4
	.dc.b	$7B,$3F,$E6,$EF,$3F,$FC,$4F,$1F,$EF,$66,$F1,$1A,$48,$FF,$37,$35
	.dc.b	$FE,$46,$D7,$FD,$A2,$DE,$3F,$F6,$E3,$79,$3C,$13,$C8,$87,$22,$9C
	.dc.b	$8B,$3F,$94,$7B,$FC,$4F,$E5,$3E,$22,$47,$E5,$3E,$07,$F2,$FE,$04
	.dc.b	$8F,$2F,$C4,$7E,$41,$FC,$CF,$81,$EA,$F8,$8F,$5F,$C4,$7E,$42,$7F
	.dc.b	$3B,$E0,$FE,$41,$FC,$FF,$81,$F9,$05,$E0,$7E,$43,$87,$EF,$6F,$F8
	.dc.b	$12,$70,$7E,$19,$B5,$F8,$68,$C1,$1C,$12,$79,$C6,$5B,$19,$F0,$3F
	.dc.b	$27,$5D,$27,$24,$8C,$D7,$1F,$03,$F2,$15,$FD,$A3,$C0,$FC,$86,$1F
	.dc.b	$B8,$F0,$3F,$20,$FE,$F3,$BF,$F2,$0F,$EF,$BC,$3F,$E0,$3D,$FF,$90
	.dc.b	$C1,$97,$EF,$BB,$FF,$23,$7F,$C1,$3B,$BF,$22,$FF,$06,$ED,$FC,$83
	.dc.b	$FC,$1B,$B7,$F9,$17,$8B,$FC,$50,$91,$FA,$F6,$BF,$AA,$E0,$F5,$BF
	.dc.b	$06,$33,$97,$C1,$9C,$F0,$FC,$19,$FE,$3F,$E2,$FB,$AB,$CF,$6A,$71
	.dc.b	$7F,$19,$9C,$C6,$3B,$3F,$8F,$F6,$7E,$26,$FF,$37,$F1,$1F,$A0,$3F
	.dc.b	$43,$F1,$1F,$A2,$F8,$8F,$D1,$FA,$5F,$A4,$F4,$BF,$4B,$F0,$FF,$A6
	.dc.b	$F4,$FF,$4F,$E9,$CC,$BF,$4E,$FC,$27,$E0,$10,$F8,$1C,$3E,$EE,$7F
	.dc.b	$C1,$71,$FB,$89,$19,$9F,$88,$71,$E1,$FC,$27,$17,$E0,$2F,$FC,$49
	.dc.b	$D9,$F0,$DF,$E6,$BD,$9F,$80,$BF,$E7,$3D,$9E,$D7,$F7,$7C,$5F,$BE
	.dc.b	$F8,$7F,$74,$7E,$FB,$87,$F8,$0F,$B9,$FC,$0B,$E1,$FF,$81,$FA,$BE
	.dc.b	$DA,$FE,$03,$C3,$7C,$1F,$D3,$79,$5F,$C0,$67,$F8,$71,$FB,$CD,$BF
	.dc.b	$81,$BF,$C0,$48,$7F,$01,$2F,$0F,$F0,$A7,$81,$C8,$CC,$72,$78,$12
	.dc.b	$38,$3A,$38,$19,$8F,$F2,$47,$67,$F1,$CF,$2B,$F8,$5C,$FF,$5C,$EC
	.dc.b	$FC,$1B,$FD,$68,$FE,$3F,$F0,$FF,$C8,$BE,$1F,$F0,$71,$FC,$8B,$DD
	.dc.b	$FE,$4B,$F0,$DF,$83,$8F,$E4,$BD,$9F,$83,$7F,$93,$FA,$E7,$FC,$97
	.dc.b	$FC,$01,$FE,$53,$D7,$1E,$1A,$3F,$D0,$9C,$3F,$C3,$FB,$BF,$03,$7F
	.dc.b	$7E,$FF,$0B,$DD,$F8,$3C,$3F,$45,$39,$3F,$0A,$7E,$30,$C3,$A7,$FB
	.dc.b	$CF,$A3,$F8,$57,$FC,$CB,$DD,$FE,$00,$EE,$FC,$37,$2F,$E9,$BB,$DD
	.dc.b	$FF,$87,$BF,$D2,$9F,$D0,$BB,$BF,$0E,$FF,$42,$EE,$FC,$3B,$FD,$1B
	.dc.b	$B7,$F0,$EF,$E8,$88,$7E,$20,$FE,$AA,$7A,$30,$FD,$F6,$BE,$DD,$FE
	.dc.b	$0C,$FA,$C9,$3A,$7D,$B1,$A8,$C4,$7F,$AB,$3A,$BD,$DB,$FE,$FE,$3F
	.dc.b	$C3,$7D,$3F,$79,$FF,$0D,$FA,$3F,$55,$5F,$F0,$6E,$4F,$F5,$A4,$7D
	.dc.b	$8F,$6F,$53,$9F,$6E,$F2,$FD,$AB,$FF,$49,$B7,$DA,$CE,$1F,$C4,$E7
	.dc.b	$FC,$F5,$7D,$7E,$1B,$3A,$7E,$07,$E1,$AD,$3E,$1A,$FC,$34,$C7,$E3
	.dc.b	$DF,$DB,$C7,$DE,$4C,$FF,$0B,$3F,$13,$8B,$E5,$11,$C1,$FE,$3C,$EA
	.dc.b	$F8,$BC,$6E,$AF,$C1,$4F,$F9,$03,$DC,$FF,$0A,$BF,$98,$F8,$6F,$2F
	.dc.b	$1F,$9B,$E1,$BC,$FF,$03,$E8,$E4,$F4,$F8,$FE,$AE,$5F,$5F,$C3,$7B
	.dc.b	$3E,$1B,$DB,$F0,$D3,$4F,$5D,$FF,$07,$E4,$FF,$09,$47,$F0,$06,$BF
	.dc.b	$BB,$6B,$FB,$B3,$F7,$BC,$9F,$C0,$1C,$46,$A3,$04,$E4,$FE,$10,$D9
	.dc.b	$E3,$18,$23,$83,$FD,$93,$D9,$FC,$1D,$FC,$49,$0F,$F3,$1F,$A7,$FD
	.dc.b	$26,$BF,$11,$F0,$BF,$6B,$E1,$BE,$DF,$6F,$DC,$E9,$FB,$B8,$FD,$EF
	.dc.b	$87,$FB,$FD,$1F,$80,$EE,$FC,$0F,$6F,$DF,$D7,$F2,$7B,$7E,$37,$B7
	.dc.b	$FA,$54,$7E,$8A,$3E,$FB,$4F,$D5,$38,$DB,$F0,$91,$82,$70,$7E,$26
	.dc.b	$7F,$C1,$1D,$BF,$82,$CB,$F0,$35,$FE,$7E,$EC,$FC,$0B,$E2,$3F,$35
	.dc.b	$D9,$F8,$27,$F8,$DF,$5F,$E0,$9F,$E3,$BD,$5F,$82,$7F,$8F,$F5,$7E
	.dc.b	$09,$FE,$43,$D3,$F8,$17,$A6,$9F,$C8,$7A,$7F,$D9,$5F,$C0,$1F,$DD
	.dc.b	$F6,$FE,$4A,$DA,$FE,$F8,$DA,$34,$9E,$0F,$EB,$1B,$3F,$EF,$A6,$BE
	.dc.b	$06,$F0,$7F,$81,$1B,$7E,$0A,$E4,$4A,$E4,$D7,$93,$DF,$AB,$C1,$E0
	.dc.b	$F8,$7D,$DF,$17,$C1,$78,$FD,$DF,$27,$C2,$F9,$7E,$23,$CD,$F0,$5E
	.dc.b	$7F,$88,$9A,$79,$FE,$0F,$D5,$D5,$EB,$F3,$FD,$9D,$5E,$DF,$3B,$DD
	.dc.b	$F1,$13,$0F,$E0,$BC,$9E,$A7,$E3,$F1,$F9,$F9,$7D,$CD,$E5,$F7,$4F
	.dc.b	$B0,$87,$B0,$B3,$C1,$B1,$C1,$D1,$23,$4D,$8D,$46,$0F,$9D,$F0,$9F
	.dc.b	$57,$E1,$3E,$35,$7D,$5F,$88,$F8,$CF,$D7,$E3,$FB,$5F,$0B,$F1,$A3
	.dc.b	$ED,$71,$FD,$DF,$85,$FB,$DC,$7F,$1A,$7E,$F7,$17,$C6,$7F,$B8,$69
	.dc.b	$FE,$40,$FE,$01,$FF,$4F,$F0,$38,$7E,$45,$7F,$9F,$B9,$3F,$51,$C3
	.dc.b	$23,$F4,$FE,$8F,$F4,$D8,$C7,$6F,$EA,$67,$3E,$57,$F2,$D7,$F6,$51
	.dc.b	$F9,$4E,$EF,$C1,$C7,$8A,$BF,$BF,$E6,$F7,$BF,$0B,$3F,$95,$ED,$FE
	.dc.b	$87,$2F,$89,$FC,$34,$FE,$67,$B7,$F0,$AF,$E6,$7B,$77,$F1,$3C,$3E
	.dc.b	$27,$8F,$C2,$F2,$F8,$9F,$C3,$5F,$E8,$7C,$0A,$FB,$CF,$C2,$9F,$70
	.dc.b	$A7,$D9,$2F,$37,$D9,$DF,$ED,$92,$32,$DE,$1F,$D9,$F8,$1F,$8D,$9C
	.dc.b	$92,$32,$F7,$6B,$DD,$B5,$DE,$3F,$7F,$E0,$7E,$1D,$FD,$D7,$87,$FB
	.dc.b	$DF,$03,$F0,$F1,$FB,$BF,$03,$F0,$EF,$E1,$88,$7E,$FF,$BB,$F0,$EF
	.dc.b	$F0,$3E,$AF,$D6,$3F,$C1,$7C,$2F,$A2,$FE,$13,$2F,$E0,$9D,$9F,$C5
	.dc.b	$35,$F6,$6B,$E6,$DB,$E0,$76,$FE,$BE,$31,$87,$BE,$30,$4A,$E2,$D8
	.dc.b	$F8,$4E,$BA,$FF,$0F,$1F,$C5,$77,$FC,$2B,$FC,$5F,$C3,$FE,$12,$FF
	.dc.b	$1B,$F0,$BF,$0F,$F1,$3F,$88,$F6,$FF,$13,$F1,$3F,$8A,$F8,$1F,$C5
	.dc.b	$FC,$4F,$E3,$3E,$1F,$F1,$BF,$01,$F8,$EF,$89,$FC,$7F,$C0,$CD,$3F
	.dc.b	$84,$F0,$F9,$88,$F9,$9A,$FD,$7E,$1F,$53,$63,$3C,$BF,$67,$8F,$E1
	.dc.b	$92,$9D,$7F,$6E,$FD,$C7,$F0,$13,$F9,$B7,$17,$B9,$FE,$8B,$E2,$3E
	.dc.b	$27,$DF,$1F,$AB,$3F,$D7,$D0,$E3,$F8,$31,$9C,$52,$E1,$F8,$73,$98
	.dc.b	$7C,$4F,$EC,$3E,$27,$F6,$3F,$13,$F0,$EB,$C3,$C1,$FD,$C3,$7F,$ED
	.dc.b	$79,$7F,$01,$27,$99,$FC,$16,$BF,$81,$9F,$E2,$E3,$F8,$1B,$98,$FF
	.dc.b	$7C,$F0,$3A,$BE,$FD,$FF,$15,$F2,$FE,$FB,$FD,$4F,$CB,$FE,$07,$E5
	.dc.b	$FE,$73,$2F,$87,$F1,$3F,$7D,$F5,$7C,$4F,$F9,$2B,$A3,$FD,$19,$1F
	.dc.b	$EE,$17,$FC,$B2,$FC,$AE,$AF,$C0,$BF,$06,$38,$7C,$34,$8F,$C0,$9F
	.dc.b	$7C,$87,$DD,$2E,$5F,$77,$5F,$C0,$B7,$AB,$F0,$3E,$0F,$E9,$1F,$CF
	.dc.b	$F5,$FD,$F9,$FE,$DD,$E0,$FE,$E7,$0F,$BE,$7F,$3F,$3E,$A7,$5F,$DE
	.dc.b	$7F,$03,$D7,$F7,$1F,$C1,$F5,$FD,$A7,$F0,$BD,$5F,$65,$FB,$9D,$3F
	.dc.b	$67,$1F,$E8,$1D,$3F,$B9,$9F,$4E,$BE,$56,$9F,$B7,$D7,$D2,$D3,$3D
	.dc.b	$1F,$99,$8C,$BF,$9B,$E8,$FB,$35,$FE,$16,$DB,$EC,$BF,$E1,$8E,$CF
	.dc.b	$CD,$65,$FE,$0D,$F0,$7D,$8F,$F5,$3F,$07,$D8,$FF,$55,$DF,$D8,$FF
	.dc.b	$57,$E6,$FA,$FA,$FE,$8B,$9B,$D9,$97,$F4,$AD,$FD,$93,$F5,$49,$7F
	.dc.b	$4D,$F4,$F2,$FD,$6A,$F8,$11,$A6,$CE,$E3,$FC,$49,$CB,$F5,$E7,$D9
	.dc.b	$1F,$B5,$3F,$89,$1F,$5C,$FD,$AF,$2F,$B2,$3F,$AE,$F2,$FB,$1F,$EB
	.dc.b	$FA,$7B,$1F,$EC,$3C,$BF,$5E,$FF,$7B,$CB,$EC,$7F,$77,$A7,$EB,$EB
	.dc.b	$FB,$7F,$67,$EA,$49,$E7,$FF,$6F,$D7,$F6,$71,$8B,$FD,$84,$2B,$FD
	.dc.b	$DB,$2A,$FF,$1B,$57,$F9,$0F,$D4,$FE,$DF,$DD,$F6,$23,$FC,$9B,$DD
	.dc.b	$FB,$07,$F9,$71,$FC,$1F,$B7,$F6,$0F,$F1,$4E,$DF,$D8,$65,$FB,$1F
	.dc.b	$83,$FE,$5D,$3F,$C3,$7C,$2F,$E5,$CF,$F1,$3E,$67,$7F,$EB,$E3,$FB
	.dc.b	$97,$77,$E3,$5F,$AF,$DF,$F8,$EE,$9F,$B1,$1F,$93,$3E,$C1,$78,$BE
	.dc.b	$C1,$33,$FF,$33,$F4,$FD,$99,$FF,$33,$F7,$7E,$BD,$FB,$27,$F6,$73
	.dc.b	$F5,$E7,$F2,$3E,$9F,$B4,$FF,$2C,$F1,$3F,$A5,$3F,$C9,$FC,$4F,$F0
	.dc.b	$AF,$57,$D9,$CB,$FC,$59,$D5,$F6,$76,$FE,$5F,$D5,$F9,$49,$F2,$61
	.dc.b	$E7,$D7,$F1,$E4,$70,$4A,$F8,$4F,$B9,$B3,$FC,$3F,$C1,$FD,$7C,$FB
	.dc.b	$0F,$F1,$41,$FC,$28,$F6,$1F,$E7,$DF,$03,$D7,$5F,$D0,$BC,$0F,$CA
	.dc.b	$4F,$F4,$8E,$EF,$CA,$3F,$D2,$BC,$0F,$D7,$E3,$EC,$8F,$B1,$F0,$5F
	.dc.b	$CB,$4F,$66,$7E,$9C,$7C,$AD,$E4,$FA,$F1,$8A,$F6,$46,$09,$C9,$F5
	.dc.b	$67,$FD,$33,$C9,$F5,$71,$FF,$49,$78,$5F,$C2,$31,$FF,$23,$F8,$5E
	.dc.b	$AF,$17,$FA,$F6,$7F,$D6,$79,$BE,$AB,$FD,$9B,$9B,$EA,$BF,$D7,$7C
	.dc.b	$5F,$EB,$6F,$E4,$7E,$1B,$FA,$31,$FA,$D7,$FA,$B9,$FD,$5B,$C4,$FD
	.dc.b	$66,$5F,$B0,$6C,$6B,$4F,$D6,$2D,$BF,$AD,$F3,$7D,$47,$FB,$7F,$87
	.dc.b	$FA,$B7,$FD,$6B,$E1,$FE,$AE,$BE,$A1,$FA,$B3,$FB,$97,$3F,$F9,$83
	.dc.b	$C2,$FD,$5C,$7E,$9F,$C2,$FD,$5B,$FE,$02,$F0,$BF,$56,$FF,$B2,$3C
	.dc.b	$27,$4F,$D5,$8F,$EB,$27,$E2,$7A,$7E,$87,$83,$FA,$B2,$47,$EB,$88
	.dc.b	$7F,$86,$FA,$BE,$A4,$BA,$BE,$A7,$1F,$F3,$BA,$FE,$41,$C5,$FE,$8B
	.dc.b	$E9,$FA,$9E,$0F,$E3,$38,$BF,$C1,$FD,$BF,$8C,$7F,$C2,$3E,$0F,$E9
	.dc.b	$F1,$FF,$40,$F8,$3F,$4B,$2F,$A8,$7A,$32,$FD,$3D,$FE,$24,$9A,$7A
	.dc.b	$A7,$CF,$3A,$8C,$47,$FB,$6F,$9F,$E9,$47,$A6,$3F,$4A,$7F,$00,$9F
	.dc.b	$F6,$F7,$97,$FE,$07,$F6,$FF,$C1,$1F,$05,$FE,$09,$F7,$3F,$C1,$5F
	.dc.b	$05,$FE,$0B,$F8,$2F,$F0,$67,$C1,$7F,$83,$7E,$07,$FC,$1D,$F0,$3F
	.dc.b	$E0,$FF,$85,$FF,$08,$7C,$34,$DB,$D5,$C7,$FA,$3E,$AF,$9F,$1F,$4D
	.dc.b	$BC,$1E,$96,$CF,$F9,$AB,$AB,$E8,$F5,$7D,$0B,$FF,$7F,$F9,$3F,$A3
	.dc.b	$7F,$C3,$9C,$5E,$87,$F9,$67,$17,$F8,$87,$E0,$FD,$11,$FC,$BF,$83
	.dc.b	$D0,$FF,$31,$DF,$D0,$FF,$33,$DB,$D0,$FF,$32,$F4,$7F,$A5,$3F,$A3
	.dc.b	$3F,$A5,$EB,$FA,$36,$9F,$A3,$7F,$D8,$7C,$DF,$E3,$B8,$FE,$98,$7F
	.dc.b	$93,$7C,$5F,$D1,$43,$C5,$FD,$17,$17,$E2,$6B,$FA,$27,$C1,$FD,$17
	.dc.b	$FA,$37,$3F,$D2,$7F,$A4,$7C,$37,$F8,$F7,$C8,$FF,$1F,$7C,$37,$F8
	.dc.b	$FF,$E1,$95,$FE,$5F,$FE,$01,$74,$E2,$19,$FE,$E3,$C0,$56,$A7,$EB
	.dc.b	$9F,$C5,$AB,$E9,$B4,$FC,$7F,$07,$DB,$69,$FE,$9D,$EA,$FA,$EF,$F8
	.dc.b	$17,$0F,$E3,$2F,$E9,$A7,$CB,$83,$C7,$74,$FD,$9A,$FD,$D7,$4F,$D9
	.dc.b	$7F,$B3,$77,$7E,$41,$FE,$D1,$DD,$F9,$07,$FB,$57,$77,$D8,$CF,$D9
	.dc.b	$8B,$A7,$F9,$AD,$7A,$4F,$C8,$1E,$8C,$FF,$57,$B7,$B9,$B1,$FE,$B2
	.dc.b	$F0,$7F,$53,$DF,$F6,$3B,$FF,$21,$5E,$9D,$FD,$19,$F8,$3C,$0F,$AD
	.dc.b	$8F,$F1,$3D,$FD,$0F,$F1,$5D,$FD,$0F,$F1,$7D,$FE,$A5,$FD,$3E,$3F
	.dc.b	$C4,$4F,$D6,$E2,$F8,$F1,$8F,$8A,$7B,$7F,$46,$BD,$64,$BF,$D5,$3E
	.dc.b	$DE,$DF,$E0,$71,$FE,$75,$E6,$FA,$B3,$D4,$FE,$71,$2F,$77,$FD,$D3
	.dc.b	$5F,$C1,$27,$FC,$2D,$BF,$BB,$3F,$B3,$CC,$E7,$FA,$D5,$EF,$3F,$5C
	.dc.b	$7D,$63,$F5,$C7,$D6,$21,$F0,$4F,$CB,$E9,$EC,$7F,$54,$4E,$2F,$F5
	.dc.b	$5F,$57,$F5,$06,$C7,$EC,$1F,$F5,$86,$9F,$23,$3F,$CE,$2F,$67,$FC
	.dc.b	$77,$5F,$C8,$27,$FD,$54,$FF,$AF,$FD,$AF,$A3,$A7,$F6,$C8,$FD,$89
	.dc.b	$FE,$35,$E5,$FB,$75,$F8,$CE,$EF,$F5,$67,$77,$EC,$6B,$FC,$75,$DD
	.dc.b	$FB,$17,$F6,$1D,$7D,$1F,$EC,$9C,$DF,$67,$9B,$EC,$63,$F5,$A7,$EC
	.dc.b	$67,$82,$3F,$97,$8F,$E7,$9C,$DF,$61,$F6,$F3,$7D,$6A,$FD,$5F,$03
	.dc.b	$C6,$FE,$75,$87,$F2,$22,$1F,$D9,$7C,$6F,$E4,$6F,$F6,$7F,$1B,$F9
	.dc.b	$1B,$FD,$A7,$C6,$FE,$46,$FD,$52,$73,$FD,$5E,$6F,$D2,$C7,$A0,$F3
	.dc.b	$65,$E8,$8C,$E9,$ED,$3F,$AA,$70,$7E,$AA,$BE,$95,$7F,$69,$7F,$C6
	.dc.b	$B3,$FB,$8E,$AF,$C7,$47,$FB,$5F,$97,$E8,$D7,$E4,$3B,$7F,$57,$7F
	.dc.b	$BE,$ED,$FD,$5B,$FC,$03,$B7,$E8,$67,$FC,$73,$3B,$CB,$F4,$73,$F3
	.dc.b	$92,$7C,$F9,$7A,$78,$BF,$4E,$DF,$03,$9C,$39,$7E,$94,$FD,$0E,$5F
	.dc.b	$A5,$87,$EA,$B9,$7E,$8B,$FC,$68,$FA,$3A,$FF,$14,$E5,$FA,$2F,$F1
	.dc.b	$6E,$5F,$A2,$FF,$18,$E5,$FA,$2F,$AE,$BF,$4F,$CB,$F4,$7C,$6F,$EA
	.dc.b	$73,$FE,$9E,$ED,$FD,$24,$63,$4F,$EA,$97,$FD,$5C,$FD,$F7,$07,$E9
	.dc.b	$B8,$BF,$C6,$9E,$AB,$C7,$FE,$A7,$E3,$FF,$52,$C3,$F4,$58,$7F,$8A
	.dc.b	$FC,$7F,$EA,$4F,$F8,$B7,$BF,$E9,$BF,$E3,$FF,$0D,$E0,$7E,$8E,$3F
	.dc.b	$C6,$5E,$07,$E8,$DF,$0F,$57,$ED,$71,$FE,$37,$AF,$F3,$4D,$BE,$8D
	.dc.b	$7D,$3E,$9F,$A2,$DE,$AF,$E6,$91,$FD,$56,$3F,$C6,$87,$F5,$EE,$3F
	.dc.b	$DF,$FC,$47,$D6,$F1,$BF,$C0,$D8,$2D,$E3,$7F,$38,$DF,$FA,$7E,$3F
	.dc.b	$53,$4F,$E3,$7E,$37,$F1,$AC,$FF,$A4,$F8,$3F,$8E,$9F,$E3,$FE,$0F
	.dc.b	$E3,$9F,$E4,$3E,$0F,$D6,$6E,$1F,$DA,$34,$F5,$93,$4F,$5B,$77,$F5
	.dc.b	$9F,$8C,$3F,$DA,$33,$F5,$A7,$F8,$C3,$ED,$AF,$F7,$7B,$9E,$6F,$C7
	.dc.b	$DF,$B7,$BF,$F3,$BE,$37,$E7,$B8,$7F,$3F,$5F,$A0,$E1,$FD,$0D,$7E
	.dc.b	$8B,$E1,$7F,$47,$F1,$1F,$A4,$F8,$6F,$D2,$FC,$47,$E9,$BC,$99,$AF
	.dc.b	$AF,$7F,$49,$38,$3D,$31,$FD,$35,$FF,$6A,$78,$FF,$A4,$8F,$E9,$7E
	.dc.b	$3F,$9A,$BF,$8A,$47,$F8,$04,$FE,$47,$BF,$9D,$FF,$0B,$7C,$47,$99
	.dc.b	$FF,$0C,$7C,$17,$EE,$3E,$13,$CD,$1F,$E1,$BF,$83,$F2,$BF,$8C,$D3
	.dc.b	$FD,$31,$CB,$FD,$12,$3F,$41,$D5,$FA,$57,$FA,$37,$8D,$FA,$57,$FA
	.dc.b	$63,$FF,$00,$5F,$9C,$7F,$62,$EF,$FD,$2F,$8B,$FA,$48,$FE,$8C,$7F
	.dc.b	$8A,$FA,$7F,$85,$F0,$FC,$E8,$FF,$17,$74,$7F,$0E,$F0,$3E,$74,$7F
	.dc.b	$8C,$79,$FE,$7B,$FE,$33,$E6,$FE,$27,$CB,$23,$E7,$73,$7C,$C7,$F9
	.dc.b	$66,$FF,$87,$7E,$71,$FD,$EF,$2F,$9D,$1F,$9D,$AF,$9D,$D3,$FE,$14
	.dc.b	$F0,$7F,$CF,$D7,$FB,$0E,$7F,$E4,$3C,$FF,$36,$3F,$67,$CF,$F3,$5F
	.dc.b	$F2,$37,$4F,$F2,$6E,$7F,$9B,$CD,$FE,$13,$E7,$FF,$24,$79,$1F,$D0
	.dc.b	$73,$7C,$07,$F2,$C2,$63,$FE,$4F,$6B,$FC,$01,$B7,$FE,$1D,$6F,$77
	.dc.b	$F8,$71,$B3,$FD,$99,$BC,$1F,$E5,$46,$DF,$82,$7B,$10,$EC,$5B,$FF
	.dc.b	$15,$7C,$27,$F4,$2F,$E0,$1F,$F3,$04,$FF,$22,$7F,$90,$3F,$80,$C1
	.dc.b	$0F,$47,$87,$C4,$F1,$F8,$9E,$5F,$13,$CF,$E2,$7A,$7B,$7D,$7E,$07
	.dc.b	$B7,$C2,$CD,$3A,$66,$F7,$7C,$47,$8F,$E1,$7C,$9F,$0D,$2B,$FC,$43
	.dc.b	$CD,$FE,$64,$B7,$C1,$7A,$49,$3D,$1A,$FA,$DB,$C7,$FE,$7B,$9E,$33
	.dc.b	$B2,$1B,$2C,$6E,$78,$F3,$78,$5B,$3F,$EF,$BA,$E6,$F1,$25,$71,$7D
	.dc.b	$FE,$8F,$CF,$F8,$7F,$A1,$E4,$7D,$1F,$82,$F9,$75,$F1,$39,$3E,$A7
	.dc.b	$C1,$7D,$5F,$87,$F9,$73,$F2,$38,$FE,$63,$F2,$BE,$0F,$E4,$BF,$33
	.dc.b	$8F,$EE,$7C,$4F,$DD,$F8,$79,$3F,$28,$FC,$51,$F8,$3E,$2F,$C7,$34
	.dc.b	$C9,$F8,$82,$70,$7E,$55,$D7,$67,$E4,$5B,$DD,$F9,$17,$75,$D1,$A7
	.dc.b	$4A,$E4,$E2,$B9,$36,$9D,$DF,$95,$8F,$F0,$BF,$F0,$0B,$0F,$CC,$77
	.dc.b	$E3,$E2,$33,$F1,$5F,$92,$BF,$C6,$F7,$F0,$F8,$9E,$3F,$15,$F9,$19
	.dc.b	$FC,$F7,$81,$77,$FF,$5F,$69,$FD,$82,$7F,$45,$E0,$7E,$45,$FD,$27
	.dc.b	$81,$FB,$EF,$03,$F0,$51,$F8,$6D,$EB,$F8,$06,$E1,$F8,$18,$C1,$F8
	.dc.b	$7F,$03,$F1,$3C,$7F,$AC,$6F,$81,$F8,$AA,$E0,$4B,$E6,$4B,$E4,$78
	.dc.b	$4A,$76,$BF,$DB,$F8,$1F,$8F,$8F,$D7,$F7,$FE,$3D,$FD,$8F,$87,$FB
	.dc.b	$97,$F6,$7D,$DF,$8F,$7F,$6B,$E1,$7E,$39,$FE,$09,$DD,$FC,$1F,$BB
	.dc.b	$F1,$F1,$FC,$1F,$BB,$F1,$EF,$EF,$3B,$7F,$1E,$BC,$5F,$E4,$9E,$2F
	.dc.b	$F1,$62,$47,$E2,$1A,$FE,$05,$BC,$1F,$86,$27,$07,$E2,$DB,$3F,$8B
	.dc.b	$8C,$11,$FF,$45,$F7,$7E,$25,$F8,$27,$F1,$0E,$EF,$D7,$BF,$E4,$0E
	.dc.b	$8F,$F5,$B3,$FC,$7D,$7C,$3F,$E4,$3E,$07,$F2,$3E,$F7,$E4,$BE,$1B
	.dc.b	$F2,$7E,$FF,$E5,$3E,$1B,$F2,$BF,$01,$F9,$6F,$89,$FC,$BF,$C4,$7E
	.dc.b	$63,$E1,$A6,$BF,$B1,$ED,$FC,$03,$F6,$5A,$FD,$9E,$EF,$C1,$71,$7D
	.dc.b	$F9,$CF,$1F,$E0,$EB,$EF,$B6,$33,$3B,$AE,$35,$A8,$C7,$7F,$E3,$89
	.dc.b	$1F,$C5,$0F,$E6,$7E,$77,$F1,$47,$F9,$AF,$9D,$F1,$B1,$FD,$2F,$C3
	.dc.b	$7E,$CB,$E2,$3E,$25,$7C,$7E,$3F,$88,$FC,$9E,$2F,$88,$FC,$BE,$1F
	.dc.b	$DC,$FC,$3F,$EE,$BE,$2B,$E1,$CF,$D1,$E1,$FD,$63,$F6,$C9,$C3,$FC
	.dc.b	$13,$4F,$BE,$DE,$0F,$BC,$D9,$DF,$9F,$FC,$46,$B6,$DA,$DB,$67,$7D
	.dc.b	$9F,$8E,$9C,$BF,$DA,$BB,$3F,$11,$3F,$85,$9F,$E1,$EF,$E1,$7C,$FF
	.dc.b	$E2,$7F,$11,$FC,$53,$B3,$FD,$57,$F0,$BF,$E7,$1C,$FF,$8B,$5B,$44
	.dc.b	$FF,$9F,$63,$0F,$F9,$69,$FF,$48,$BF,$E1,$49,$E3,$3C,$3E,$23,$AF
	.dc.b	$C1,$FE,$36,$FF,$2D,$D7,$E0,$F7,$64,$7E,$37,$B3,$F0,$38,$7E,$12
	.dc.b	$3F,$04,$47,$04,$8D,$93,$B3,$F0,$B3,$F9,$36,$E5,$F8,$F9,$D4,$FE
	.dc.b	$65,$E3,$E6,$7F,$0F,$8F,$F0,$57,$99,$FC,$3D,$FF,$06,$79,$9F,$CE
	.dc.b	$BB,$7F,$15,$1F,$AF,$F0,$FF,$5A,$FE,$C7,$C3,$FD,$6B,$FB,$3F,$0F
	.dc.b	$F5,$AF,$ED,$7B,$3F,$14,$FE,$DF,$AF,$F1,$7E,$37,$F0,$E8,$FE,$18
	.dc.b	$43,$F8,$61,$76,$FC,$2F,$5F,$E0,$63,$F0,$66,$7A,$BF,$06,$FF,$5C
	.dc.b	$EB,$FC,$1D,$7F,$82,$77,$FC,$1C,$FB,$0F,$E5,$67,$F8,$9F,$A1,$F1
	.dc.b	$1F,$84,$F8,$3F,$C2,$FB,$9F,$86,$F8,$3F,$C3,$FA,$3F,$88,$F7,$E6
	.dc.b	$3C,$7D,$FF,$C5,$E5,$F8,$CF,$7E,$4F,$DD,$EB,$FA,$7C,$1F,$BC,$8F
	.dc.b	$A5,$18,$7E,$F7,$5F,$D6,$7E,$D1,$A2,$18,$3F,$C8,$1D,$7F,$67,$87
	.dc.b	$DD,$1F,$C0,$8F,$69,$EF,$3F,$2F,$F0,$5E,$D7,$E0,$7C,$27,$B5,$F8
	.dc.b	$5B,$7E,$93,$E0,$FF,$4B,$CD,$FA,$6D,$FD,$B5,$F1,$F6,$F6,$BF,$D1
	.dc.b	$EF,$E8,$BF,$60,$99,$7D,$87,$EB,$1F,$60,$CE,$DC,$23,$51,$83,$FA
	.dc.b	$19,$FE,$50,$3F,$C1,$71,$F6,$A3,$FC,$03,$E3,$7E,$56,$7E,$D7,$2F
	.dc.b	$EF,$39,$FE,$CC,$7F,$98,$FF,$80,$7F,$CC,$F7,$FB,$67,$FC,$CF,$1F
	.dc.b	$BA,$E4,$FB,$6A,$75,$E9,$23,$97,$83,$5D,$67,$9F,$09,$8B,$E0,$B5
	.dc.b	$8F,$77,$F9,$3C,$FE,$E7,$9F,$F0,$0D,$C3,$EE,$D7,$ED,$E7,$3C,$FF
	.dc.b	$83,$8C,$4E,$B1,$FC,$53,$F8,$23,$FA,$19,$FD,$A0,$FF,$42,$7A,$1F
	.dc.b	$D0,$E3,$FC,$E5,$9F,$F2,$0E,$7F,$C1,$57,$F1,$AE,$7F,$C1,$3F,$C7
	.dc.b	$38,$FF,$04,$FE,$97,$9B,$F0,$4F,$E9,$FA,$7F,$03,$9F,$F2,$05,$CD
	.dc.b	$EE,$E1,$7C,$1F,$C2,$D7,$E3,$46,$30,$FA,$93,$97,$F9,$84,$65,$F9
	.dc.b	$A6,$23,$5C,$3F,$39,$B3,$BC,$7E,$E4,$60,$95,$D8,$87,$4A,$E4,$E2
	.dc.b	$78,$B6,$FC,$F5,$CA,$38,$5F,$79,$CC,$7E,$37,$E0,$3C,$67,$93,$DA
	.dc.b	$F2,$FB,$52,$7F,$25,$EF,$7A,$3D,$AF,$4F,$BD,$EA,$F7,$BD,$7D,$1E
	.dc.b	$CF,$06,$61,$F9,$9F,$67,$DF,$D3,$F0,$3D,$EF,$83,$F0,$72,$BF,$C2
	.dc.b	$DC,$2F,$7F,$E2,$92,$3F,$44,$D7,$D6,$DE,$4F,$D2,$37,$8F,$EB,$B6
	.dc.b	$7E,$AB,$79,$7F,$3E,$DB,$F0,$4F,$62,$1D,$8B,$7F,$B1,$E4,$F8,$F1
	.dc.b	$FB,$5F,$83,$FB,$7F,$01,$F7,$BF,$80,$3F,$B9,$CF,$D7,$5E,$37,$C9
	.dc.b	$8F,$F6,$F4,$3E,$13,$87,$C2,$7C,$A9,$CF,$C2,$73,$F8,$CF,$C3,$F2
	.dc.b	$7C,$A9,$FE,$09,$C7,$F2,$5F,$BB,$C7,$F9,$9E,$DF,$C0,$70,$7E,$12
	.dc.b	$7F,$03,$18,$E4,$FC,$0C,$60,$FC,$41,$F9,$72,$F2,$7E,$39,$D1,$2B
	.dc.b	$A1,$27,$9B,$63,$91,$B8,$C9,$FC,$3F,$6F,$F3,$0C,$7C,$5F,$73,$F2
	.dc.b	$0F,$F1,$FE,$DF,$C8,$3F,$98,$ED,$FD,$39,$F9,$07,$F9,$27,$5F,$CF
	.dc.b	$F5,$FF,$1F,$1F,$9E,$EB,$FA,$5F,$05,$F4,$FD,$8F,$A9,$F0,$5F,$57
	.dc.b	$D9,$97,$F9,$0E,$EF,$C1,$11,$FB,$FB,$FE,$29,$A7,$E2,$8B,$C3,$F8
	.dc.b	$A6,$F7,$7E,$29,$FC,$41,$38,$BF,$2A,$EA,$FB,$3F,$B4,$EE,$FC,$5B
	.dc.b	$FB,$6E,$EF,$C5,$BF,$86,$C7,$F1,$9D,$BF,$89,$8F,$C6,$76,$FE,$21
	.dc.b	$FE,$8D,$E0,$7E,$47,$B7,$F1,$11,$FB,$13,$FA,$57,$81,$FB,$17,$FA
	.dc.b	$67,$5F,$E2,$7D,$77,$91,$F5,$7E,$0B,$F3,$44,$AF,$CB,$B5,$FC,$E3
	.dc.b	$78,$3F,$C9,$3D,$A8,$C6,$DC,$EB,$89,$0E,$25,$AE,$33,$C2,$3F,$CB
	.dc.b	$55,$CD,$D5,$73,$78,$9F,$AD,$F8,$4F,$D8,$1F,$B1,$F7,$BF,$65,$ED
	.dc.b	$C9,$FD,$7F,$BB,$FB,$5F,$84,$FD,$B7,$BD,$FB,$7F,$84,$FE,$6B,$FC
	.dc.b	$03,$FE,$6B,$CF,$FB,$23,$FE,$69,$EB,$3F,$E6,$9D,$FF,$CC,$EF,$F9
	.dc.b	$A6,$77,$F0,$7C,$BE,$0F,$9F,$C1,$CD,$7F,$D6,$5C,$3E,$B7,$D3,$8F
	.dc.b	$C0,$7D,$A7,$C0,$E2,$F8,$59,$AB,$85,$6E,$B8,$E5,$F0,$E3,$04,$70
	.dc.b	$48,$E8,$76,$AE,$97,$DA,$FA,$D6,$E7,$C1,$C5,$F3,$2F,$F8,$BF,$17
	.dc.b	$C6,$F7,$7E,$3F,$C1,$7F,$1F,$E3,$F9,$3F,$05,$F3,$2F,$F8,$EF,$17
	.dc.b	$CC,$F8,$8F,$9B,$F1,$5F,$3B,$E1,$3E,$7F,$C4,$FD,$0E,$AF,$91,$87
	.dc.b	$F8,$5F,$8F,$E6,$F2,$E2,$3F,$C6,$C7,$EC,$4F,$D6,$71,$7E,$D5,$A6
	.dc.b	$67,$F5,$EE,$38,$BF,$6C,$DF,$17,$F6,$D3,$A2,$57,$36,$CF,$38,$C1
	.dc.b	$1C,$12,$BA,$CF,$E1,$3E,$0B,$F1,$7F,$09,$F8,$CF,$73,$F2,$57,$FD
	.dc.b	$0F,$BB,$F1,$FE,$E7,$E4,$E3,$E0,$65,$F9,$0F,$7F,$F2,$7E,$FF,$E5
	.dc.b	$3E,$17,$F2,$BF,$01,$F9,$6F,$8A,$FC,$BF,$C3,$7E,$63,$C2,$FF,$06
	.dc.b	$F7,$FF,$91,$B2,$FC,$CF,$85,$F9,$1D,$3E,$4F,$1F,$E0,$1F,$C1,$78
	.dc.b	$3F,$80,$7F,$0A,$7E,$30,$BC,$9F,$8B,$6D,$FE,$16,$78,$57,$39,$E0
	.dc.b	$DA,$E0,$DA,$E1,$E1,$7E,$3C,$FC,$86,$3F,$8D,$BF,$EC,$DE,$17,$E3
	.dc.b	$5F,$86,$7F,$B6,$FC,$37,$85,$F8,$F8,$CE,$3F,$B4,$F0,$7F,$21,$1F
	.dc.b	$DB,$FC,$1F,$C8,$3A,$BF,$DD,$0E,$FF,$C8,$C6,$7B,$BF,$1D,$B7,$E4
	.dc.b	$23,$87,$77,$F9,$83,$83,$DE,$4D,$FD,$53,$F2,$9A,$FC,$28,$D3,$92
	.dc.b	$19,$8D,$D7,$F2,$8D,$FE,$13,$FC,$AB,$7F,$C3,$BF,$CB,$3A,$BE,$05
	.dc.b	$FF,$82,$FC,$5F,$7C,$FF,$30,$DB,$DF,$3F,$E4,$5D,$BD,$FC,$DE,$E8
	.dc.b	$C7,$2F,$BB,$93,$F0,$B1,$F1,$79,$7D,$DB,$79,$27,$EB,$CF,$A3,$2F
	.dc.b	$B7,$BF,$B2,$3D,$46,$67,$6E,$09,$3E,$D7,$FC,$3F,$BF,$F3,$57,$EA
	.dc.b	$78,$9F,$C2,$9F,$E6,$A7,$D7,$E4,$F6,$DF,$F8,$1F,$6F,$6E,$1F,$CC
	.dc.b	$3C,$2F,$E1,$57,$F7,$7C,$2F,$E1,$4F,$D8,$21,$F0,$F9,$3F,$CF,$9E
	.dc.b	$7F,$F7,$0B,$FD,$EF,$47,$EE,$32,$FD,$E3,$6F,$F7,$9E,$1F,$F0,$69
	.dc.b	$CB,$63,$37,$CE,$72,$D8,$FE,$09,$C8,$F0,$FF,$81,$C7,$E4,$FC,$2F
	.dc.b	$BB,$7F,$E2,$3E,$8F,$E0,$73,$FE,$44,$E8,$FE,$A7,$E0,$FF,$02,$8F
	.dc.b	$CD,$78,$1F,$C0,$9F,$CE,$75,$7F,$02,$7F,$3D,$D5,$FC,$09,$FE,$01
	.dc.b	$C1,$FD,$5B,$AB,$EE,$C7,$D7,$3F,$76,$4E,$1F,$DB,$37,$A7,$EF,$57
	.dc.b	$E0,$3A,$BF,$09,$39,$6D,$70,$25,$7E,$9B,$AB,$F3,$7B,$FF,$8F,$FA
	.dc.b	$FF,$79,$87,$E5,$7B,$BF,$37,$97,$F6,$AE,$BF,$BF,$3F,$E0,$0E,$BF
	.dc.b	$BF,$87,$F7,$2E,$BF,$BF,$3F,$9D,$EB,$FB,$EF,$DD,$3E,$FE,$1F,$B9
	.dc.b	$7F,$8F,$E9,$EA,$BF,$CA,$CF,$AE,$7E,$16,$BE,$8D,$BD,$77,$B9,$F7
	.dc.b	$1A,$71,$3F,$D9,$23,$F3,$67,$F6,$43,$FC,$F7,$CB,$F7,$1F,$F0,$CF
	.dc.b	$8B,$FC,$F5,$FF,$41,$78,$DF,$E0,$F7,$F8,$A7,$8D,$FD,$91,$FB,$67
	.dc.b	$E6,$CF,$B6,$7B,$33,$FF,$7B,$71,$7F,$C0,$0D,$11,$8F,$87,$FB,$B5
	.dc.b	$FB,$08,$F7,$91,$F5,$99,$2C,$FC,$37,$19,$FC,$37,$FA,$2D,$FD,$A7
	.dc.b	$FC,$67,$C7,$F6,$AB,$FC,$43,$C7,$F6,$B0,$FF,$13,$F1,$FD,$AC,$7F
	.dc.b	$C3,$1C,$7F,$6B,$1F,$E7,$5C,$7F,$6B,$1F,$F1,$4F,$1F,$DA,$7D,$72
	.dc.b	$B6,$8F,$87,$FE,$BB,$B4,$AF,$F3,$C1,$FE,$6B,$21,$FD,$F7,$BB,$F6
	.dc.b	$57,$FB,$B2,$77,$FE,$E5,$B1,$A8,$C1,$1F,$F6,$37,$7F,$EE,$6B,$FA
	.dc.b	$57,$91,$FD,$ED,$FB,$E4,$3F,$98,$79,$1F,$7F,$4F,$F2,$2F,$77,$EE
	.dc.b	$67,$F9,$A7,$77,$EE,$5F,$E6,$DD,$DF,$B9,$7F,$9A,$9F,$7C,$FE,$C1
	.dc.b	$E1,$7F,$88,$7C,$9F,$F3,$AC,$7D,$8E,$6F,$B9,$BF,$D7,$8F,$B3,$DB
	.dc.b	$F9,$62,$77,$7E,$70,$91,$F7,$0F,$CC,$9F,$DC,$0F,$E6,$67,$F6,$7E
	.dc.b	$9F,$B8,$FF,$71,$E9,$FB,$94,$E9,$FB,$98,$7F,$99,$3A,$7E,$E6,$5F
	.dc.b	$E6,$8E,$9F,$B8,$FE,$63,$A3,$EC,$64,$E2,$FB,$F8,$7D,$EE,$84,$CD
	.dc.b	$FD,$F3,$F8,$2F,$85,$FF,$0E,$92,$7F,$C0,$3D,$48,$C4,$75,$CB,$F4
	.dc.b	$F1,$8C,$7B,$D6,$EB,$8C,$E6,$B7,$18,$21,$CC,$FC,$A4,$F1,$71,$7E
	.dc.b	$57,$77,$DA,$FA,$47,$06,$CE,$E3,$8D,$73,$AE,$11,$CA,$FC,$6E,$AF
	.dc.b	$8D,$6E,$73,$5C,$2B,$47,$13,$C0,$5B,$F0,$46,$CE,$18,$7C,$EB,$E9
	.dc.b	$5C,$0E,$64,$9E,$17,$E0,$24,$6E,$32,$D8,$CD,$F8,$1B,$5D,$A3,$13
	.dc.b	$B2,$38,$24,$6C,$98,$7C,$22,$4F,$B6,$32,$D8,$CB,$E9,$AF,$0C,$7E
	.dc.b	$1B,$0F,$81,$F0,$FF,$E6,$82,$57,$E3,$1A,$FD,$66,$E3,$F8,$A6,$DF
	.dc.b	$F3,$46,$CF,$E1,$9B,$8F,$F1,$A6,$E2,$69,$27,$F3,$7E,$F7,$F9,$EB
	.dc.b	$D4,$FF,$3D,$FC,$37,$F9,$F3,$DE,$FF,$3E,$FB,$DF,$E7,$EF,$86,$FF
	.dc.b	$3F,$FC,$37,$FA,$03,$E0,$3F,$D0,$3E,$BF,$FA,$0B,$DF,$FF,$41,$FC
	.dc.b	$1F,$FA,$13,$D8,$FF,$42,$FC,$1F,$FA,$1B,$D9,$FF,$43,$F9,$1F,$E8
	.dc.b	$8F,$76,$71,$7E,$C7,$D6,$F8,$EF,$CB,$21,$F2,$CF,$AF,$CB,$F4,$1C
	.dc.b	$73,$7D,$8E,$AF,$A3,$D3,$F8,$48,$C4,$F2,$8C,$57,$16,$D7,$0E,$9F
	.dc.b	$0D,$74,$AD,$46,$2B,$73,$FB,$6E,$AF,$CD,$5F,$18,$F2,$B7,$0F,$0C
	.dc.b	$73,$8C,$61,$D2,$36,$79,$70,$EC,$47,$27,$13,$46,$30,$F1,$B9,$26
	.dc.b	$1F,$AD,$F8,$6F,$F5,$A9,$23,$F8,$CB,$5F,$E0,$6D,$9F,$E2,$ED,$AF
	.dc.b	$F0,$0B,$67,$F8,$9B,$71,$FE,$CE,$DC,$0D,$24,$FF,$AA,$3F,$6C,$FE
	.dc.b	$DE,$FF,$1A,$5E,$FF,$F3,$28,$C7,$BF,$72,$FE,$6F,$4F,$87,$C6,$7F
	.dc.b	$CE,$33,$4B,$E0,$33,$8F,$F7,$B8,$7C,$3E,$2B,$7F,$13,$65,$F1,$17
	.dc.b	$E2,$31,$F1,$19,$F8,$8D,$7C,$46,$FE,$23,$87,$A3,$C7,$97,$97,$B5
	.dc.b	$37,$FC,$D1,$F5,$78,$7E,$D6,$FF,$82,$E3,$FA,$E4,$8C,$DF,$E0,$E7
	.dc.b	$51,$8E,$8F,$C3,$92,$FF,$03,$3A,$7B,$11,$C1,$F9,$0E,$8F,$C1,$47
	.dc.b	$E0,$08,$7E,$00,$B8,$7E,$01,$B7,$F8,$06,$9F,$80,$2D,$7E,$01,$B8
	.dc.b	$7E,$01,$BD,$1F,$80,$7B,$12,$FE,$DD,$FD,$1E,$8F,$A5,$D1,$F4,$FA
	.dc.b	$3E,$A7,$47,$D5,$E8,$FA,$DD,$1F,$5F,$9F,$EB,$67,$F8,$BE,$6F,$AC
	.dc.b	$FE,$37,$97,$EB,$3F,$8F,$E4,$FA,$CF,$E4,$79,$3E,$B2,$E3,$FE,$FF
	.dc.b	$D0,$F7,$BF,$08,$49,$FC,$C3,$5F,$CC,$37,$8B,$F2,$AD,$E2,$FC,$A5
	.dc.b	$73,$AE,$24,$38,$96,$B8,$B6,$78,$D7,$38,$FC,$D3,$79,$7F,$3E,$DB
	.dc.b	$ED,$3D,$09,$7D,$1B,$7D,$1E,$C7,$42,$DF,$46,$D7,$4A,$E4,$4A,$E5
	.dc.b	$3F,$27,$AA,$CE,$63,$A1,$E0,$2B,$9B,$ED,$19,$6C,$66,$FC,$17,$D0
	.dc.b	$95,$C9,$B5,$C1,$B3,$C2,$74,$49,$D3,$63,$51,$82,$38,$26,$1F,$34
	.dc.b	$93,$F1,$1A,$FC,$FC,$7F,$13,$EF,$7F,$0B,$24,$7F,$1F,$6B,$FC,$7D
	.dc.b	$B8,$FF,$1C,$6E,$3F,$88,$6C,$FF,$19,$6E,$3F,$C7,$9B,$89,$A4,$9F
	.dc.b	$E5,$DF,$03,$FC,$77,$DF,$FE,$3D,$E9,$FF,$21,$FE,$01,$7C,$05,$F5
	.dc.b	$71,$ED,$67,$D6,$D7,$C1,$6F,$DB,$E1,$F0,$5C,$7E,$17,$97,$BB,$CF
	.dc.b	$DD,$E9,$F0,$DD,$7A,$7B,$7C,$34,$E3,$FE,$47,$CD,$F3,$BA,$BE,$4F
	.dc.b	$4F,$DF,$6F,$4F,$DA,$6D,$7D,$76,$D7,$18,$FA,$44,$3F,$15,$D3,$F6
	.dc.b	$1C,$5F,$56,$DF,$56,$9D,$4F,$2C,$F1,$79,$DF,$82,$31,$5C,$E7,$27
	.dc.b	$2B,$F2,$38,$24,$6C,$97,$E1,$8C,$B6,$33,$1E,$33,$27,$EA,$BA,$BF
	.dc.b	$D2,$7C,$CF,$83,$FB,$44,$9F,$F4,$CD,$7F,$A7,$27,$51,$8B,$FE,$CD
	.dc.b	$18,$8F,$F4,$4B,$67,$85,$73,$9C,$B6,$33,$1C,$CF,$F7,$34,$F1,$AE
	.dc.b	$B5,$B6,$CE,$F0,$56,$E3,$8D,$6D,$A6,$E3,$53,$C3,$E1,$BF,$31,$F0
	.dc.b	$9F,$99,$F8,$4F,$CD,$7C,$4F,$E6,$FE,$17,$F3,$9F,$13,$F9,$DF,$89
	.dc.b	$FE,$FD,$FC,$03,$FE,$0B,$D3,$FD,$4E,$BE,$1B,$3F,$0F,$AF,$4F,$7F
	.dc.b	$0F,$C3,$E1,$F8,$FC,$3F,$2F,$87,$E7,$F0,$F3,$3F,$C3,$F5,$7E,$14
	.dc.b	$8F,$E0,$FA,$BF,$0F,$D1,$F7,$49,$1F,$77,$A7,$EF,$F3,$FE,$21,$B5
	.dc.b	$C2,$75,$5C,$2B,$4D,$9D,$47,$E4,$3A,$BF,$0B,$D5,$F2,$E3,$E7,$B7
	.dc.b	$AB,$E5,$91,$C1,$30,$FA,$0D,$7F,$02,$60,$86,$09,$8F,$E1,$1F,$C5
	.dc.b	$F5,$7D,$67,$ED,$74,$FD,$67,$EE,$74,$7D,$67,$EF,$73,$FD,$67,$EB
	.dc.b	$F3,$7D,$47,$F1,$1C,$FF,$EB,$73,$FD,$30,$E3,$93,$EA,$BA,$E3,$FA
	.dc.b	$8F,$DE,$E3,$FF,$25,$70,$7F,$BE,$AD,$F0,$FF,$84,$25,$76,$6B,$F8
	.dc.b	$66,$EC,$E3,$FF,$27,$57,$2A,$E0,$43,$81,$67,$82,$E2,$4F,$18,$D9
	.dc.b	$27,$73,$AB,$F0,$DF,$81,$B5,$CE,$F9,$5F,$06,$D6,$E7,$C4,$D7,$B3
	.dc.b	$67,$85,$74,$BF,$05,$7D,$0F,$87,$FD,$49,$1C,$10,$FD,$5B,$96,$9F
	.dc.b	$CF,$B0,$7C,$3F,$EC,$89,$1F,$71,$AF,$ED,$1B,$87,$AD,$B8,$FD,$06
	.dc.b	$CF,$C1,$6E,$1F,$31,$B7,$E0,$9E,$C4,$3B,$16,$FE,$47,$C3,$FF,$06
	.dc.b	$F8,$7F,$E0,$FF,$0F,$FC,$23,$E1,$FF,$84,$FC,$3F,$F0,$AF,$87,$FE
	.dc.b	$17,$F0,$FF,$C3,$3E,$1F,$F8,$6F,$C3,$FC,$1C,$FE,$97,$0D,$E1,$FC
	.dc.b	$3B,$76,$FB,$0D,$E1,$FC,$1B,$F9,$6F,$23,$EA,$54,$AF,$BD,$C7,$F5
	.dc.b	$78,$BE,$EC,$60,$FB,$A5,$3E,$E9,$71,$FB,$AE,$B1,$FB,$AE,$9B,$1A
	.dc.b	$8C,$4F,$69,$CD,$F8,$DB,$3D,$63,$13,$B2,$5F,$9C,$F4,$7C,$1F,$FA
	.dc.b	$BF,$83,$FD,$7F,$8F,$E7,$9F,$F5,$FE,$1F,$D1,$57,$C1,$6F,$E0,$B8
	.dc.b	$7C,$17,$1F,$82,$97,$F8,$DE,$8F,$BE,$47,$EE,$6B,$F8,$76,$9F,$87
	.dc.b	$2E,$DF,$87,$6E,$DF,$87,$7F,$0A,$4D,$3F,$24,$EA,$FB,$3F,$D2,$BA
	.dc.b	$3F,$10,$FF,$16,$E7,$FC,$43,$FC,$63,$9F,$F1,$0F,$F1,$AE,$6F,$C4
	.dc.b	$3F,$C7,$39,$7F,$05,$93,$96,$92,$7F,$5D,$9F,$E3,$1F,$D6,$72,$7C
	.dc.b	$D7,$EE,$57,$D0,$6C,$65,$FB,$6D,$9F,$BC,$4D,$BE,$CC,$65,$B3,$C8
	.dc.b	$87,$23,$D0,$79,$1F,$CE,$B4,$FC,$E9,$76,$FC,$EB,$6F,$DC,$FD,$93
	.dc.b	$E6,$10,$FA,$13,$F6,$5A,$7D,$92,$CF,$D9,$6C,$FD,$9B,$FA,$F1,$8D
	.dc.b	$BE,$BC,$60,$97,$FC,$EB,$6F,$B3,$5F,$E0,$4D,$BE,$CB,$FE,$06,$DB
	.dc.b	$EC,$BF,$E0,$8D,$BE,$CB,$FE,$0A,$DB,$F3,$9F,$05,$F9,$DF,$82,$FC
	.dc.b	$F7,$C3,$FE,$7F,$E0,$FF,$41,$F0,$FF,$A1,$F8,$4F,$D1,$7C,$3C,$CB
	.dc.b	$ED,$9F,$3B,$83,$EA,$3F,$AD,$27,$1F,$EC,$09,$C7,$F7,$E3,$EA,$12
	.dc.b	$78,$CE,$B9,$3E,$E4,$F2,$AE,$15,$A6,$F1,$FE,$02,$79,$3F,$71,$FE
	.dc.b	$76,$49,$FC,$93,$63,$51,$83,$3E,$07,$F6,$FE,$5F,$A3,$3F,$6E,$7E
	.dc.b	$8D,$7D,$B6,$E3,$F6,$DB,$87,$DB,$AE,$04,$AE,$0D,$AE,$0D,$9E,$13
	.dc.b	$A2,$4E,$9F,$ED,$1C,$BF,$80,$BF,$ED,$1C,$BF,$80,$7F,$B5,$72,$FE
	.dc.b	$01,$FE,$D9,$CB,$F8,$07,$FB,$77,$2F,$E0,$30,$FE,$31,$C9,$F8,$07
	.dc.b	$F8,$D7,$1F,$E0,$1F,$E3,$9C,$5F,$7D,$5F,$F2,$1E,$2F,$E8,$11,$FE
	.dc.b	$05,$EB,$7C,$2F,$F2,$72,$47,$F2,$56,$BC,$B8,$4F,$18,$CD,$FF,$2D
	.dc.b	$6C,$EB,$81,$7D,$AF,$A1,$0E,$87,$33,$17,$CD,$B5,$CE,$79,$47,$0A
	.dc.b	$EB,$7E,$08,$E4,$66,$77,$7C,$E3,$13,$BA,$E3,$1D,$8E,$33,$B2,$5F
	.dc.b	$86,$B8,$5F,$48,$C5,$71,$BE,$6F,$63,$DD,$C6,$EE,$DC,$64,$98,$F6
	.dc.b	$7C,$37,$A8,$C5,$F4,$6E,$1E,$1A,$D9,$27,$55,$C4,$91,$99,$E1,$3A
	.dc.b	$8C,$63,$EC,$23,$EB,$9D,$92,$33,$1E,$8A,$EB,$5C,$C9,$5B,$9F,$83
	.dc.b	$87,$D9,$F8,$7F,$EE,$84,$8F,$EE,$4D,$7E,$C3,$71,$FE,$F0,$DC,$39
	.dc.b	$4F,$02,$1C,$0B,$3C,$1B,$1C,$1D,$12,$34,$FE,$1F,$E1,$BF,$C1,$1E
	.dc.b	$7F,$F8,$27,$E1,$BF,$C1,$5F,$0D,$FE,$0B,$F8,$6F,$F0,$67,$C3,$7E
	.dc.b	$8B,$0F,$C9,$7F,$00,$BE,$1F,$19,$FF,$9B,$21,$F0,$F8,$D7,$F4,$90
	.dc.b	$F8,$7C,$61,$FE,$77,$87,$C3,$E3,$8D,$0F,$87,$C2,$F8,$7B,$F0,$F8
	.dc.b	$F8,$7C,$FC,$3E,$BE,$1E,$67,$FB,$8E,$BF,$C9,$3F,$80,$3E,$F7,$5F
	.dc.b	$DC,$E1,$FC,$6B,$78,$3E,$EF,$5F,$DB,$AF,$C0,$1C,$4B,$BF,$E0,$2B
	.dc.b	$84,$66,$F9,$BD,$88,$70,$AE,$91,$A8,$C4,$FF,$10,$C3,$CF,$7D,$67
	.dc.b	$8D,$6C,$98,$79,$AF,$C3,$5B,$AE,$67,$09,$D3,$4F,$33,$87,$64,$AE
	.dc.b	$AE,$89,$3F,$BE,$BF,$2E,$CF,$87,$FA,$A4,$AF,$E4,$F7,$F8,$38,$C4
	.dc.b	$7E,$21,$B1,$98,$DE,$1F,$86,$9C,$B6,$32,$7E,$1E,$32,$D3,$38,$7E
	.dc.b	$22,$32,$D8,$CD,$F9,$DB,$3A,$8F,$C2,$C7,$E7,$3E,$1F,$F1,$FE,$F7
	.dc.b	$E4,$3D,$EF,$C8,$FC,$3F,$E4,$BD,$FF,$C9,$FC,$3F,$E5,$3E,$03,$F2
	.dc.b	$BF,$0F,$F9,$6F,$81,$FC,$D7,$F0,$0F,$E9,$F4,$FE,$70,$BE,$0B,$3F
	.dc.b	$11,$AF,$88,$DF,$C4,$70,$F8,$59,$87,$E6,$7C,$BF,$AD,$CF,$F5,$39
	.dc.b	$BE,$F3,$7A,$BF,$17,$D1,$A2,$5F,$D8,$8C,$B7,$AB,$F0,$F1,$8B,$E7
	.dc.b	$5C,$1B,$5B,$3F,$10,$FE,$07,$0F,$A9,$1F,$6C,$F9,$64,$30,$4C,$3E
	.dc.b	$F3,$6B,$8C,$60,$8E,$09,$1F,$74,$E0,$58,$DB,$63,$75,$F7,$E3,$F8
	.dc.b	$AD,$7F,$08,$E7,$FB,$71,$F8,$2E,$6F,$B6,$FE,$13,$97,$ED,$BF,$86
	.dc.b	$E4,$FB,$6F,$E2,$38,$FE,$DB,$F7,$38,$BE,$CB,$F9,$0E,$37,$17,$DA
	.dc.b	$8C,$70,$FD,$A5,$D5,$FB,$CA,$7C,$47,$E1,$89,$3C,$DA,$FE,$21,$BC
	.dc.b	$2E,$1F,$F0,$75,$72,$AE,$04,$38,$16,$78,$2E,$14,$F1,$8D,$92,$77
	.dc.b	$3A,$BF,$0D,$F8,$1B,$5C,$EF,$95,$F0,$6D,$6E,$7C,$4D,$7B,$36,$78
	.dc.b	$57,$4B,$F0,$57,$03,$B4,$66,$FA,$C6,$0D,$4F,$07,$D7,$87,$CF,$AE
	.dc.b	$04,$8D,$D7,$2A,$E1,$7E,$29,$CD,$6D,$F3,$9D,$CB,$3C,$30,$F5,$12
	.dc.b	$35,$5C,$DD,$DF,$72,$5F,$59,$D1,$2B,$95,$F5,$BE,$71,$8B,$E3,$1E
	.dc.b	$63,$F8,$31,$F4,$FE,$1B,$F8,$89,$1F,$CB,$35,$FE,$26,$DB,$FB,$0D
	.dc.b	$AF,$C6,$B6,$7A,$1C,$88,$72,$2C,$F2,$6C,$F2,$69,$F8,$7F,$86,$FE
	.dc.b	$4B,$ED,$7F,$26,$F7,$BF,$93,$FB,$7F,$CA,$3D,$AF,$E5,$3F,$0D,$FC
	.dc.b	$AB,$DB,$FE,$57,$D5,$FC,$B3,$DB,$FE,$5B,$F0,$DF,$CB,$BD,$CF,$E5
	.dc.b	$FF,$0D,$FC,$C3,$DD,$FE,$63,$F0,$DF,$E0,$BF,$E0,$1F,$D0,$F0,$FF
	.dc.b	$9A,$57,$C5,$64,$FF,$42,$F3,$FC,$F2,$73,$FC,$EA,$FB,$3D,$1F,$5F
	.dc.b	$9B,$E9,$36,$3E,$A9,$2B,$8C,$65,$B1,$F5,$CE,$65,$E7,$FC,$03,$C6
	.dc.b	$FC,$11,$89,$E3,$3C,$09,$D3,$F9,$D7,$96,$1E,$46,$9C,$F0,$F1,$61
	.dc.b	$D2,$75,$3C,$F0,$ED,$39,$8E,$07,$2C,$3C,$33,$AB,$E6,$49,$D5,$7F
	.dc.b	$0B,$C3,$FC,$33,$93,$E2,$3E,$91,$2B,$FB,$BC,$65,$B1,$97,$EB,$37
	.dc.b	$0F,$E2,$73,$B9,$C9,$C0,$86,$09,$1C,$27,$96,$5F,$EB,$5A,$E5,$39
	.dc.b	$AE,$2D,$7F,$0D,$F1,$1F,$89,$F8,$8F,$C5,$7C,$47,$E2,$FE,$23,$F1
	.dc.b	$9F,$0B,$F8,$DF,$88,$FC,$77,$C3,$7E,$3F,$E2,$3F,$21,$F0,$FF,$91
	.dc.b	$F8,$7F,$C9,$79,$DF,$93,$F8,$7F,$CA,$7C,$3F,$E5,$7E,$1F,$F2,$DF
	.dc.b	$0F,$F9,$7F,$87,$9C,$5F,$BC,$EA,$FA,$3C,$DF,$63,$97,$E9,$4F,$D0
	.dc.b	$EA,$FA,$4F,$DB,$9D,$11,$FB,$66,$4B,$3C,$9C,$57,$67,$33,$C6,$30
	.dc.b	$7F,$2D,$E9,$FB,$BD,$3F,$32,$BF,$A6,$74,$FC,$CB,$F9,$D3,$9B,$4E
	.dc.b	$6B,$73,$98,$C4,$EA,$30,$49,$FA,$B5,$D1,$A7,$22,$CF,$17,$F1,$BD
	.dc.b	$3F,$65,$FC,$7F,$47,$D9,$7F,$23,$CF,$F6,$5F,$C9,$F3,$7D,$97,$F2
	.dc.b	$BC,$BF,$65,$FC,$BF,$27,$D9,$7F,$33,$C7,$F6,$5F,$CD,$F1,$7D,$87
	.dc.b	$F8,$77,$17,$F0,$CD,$DF,$0F,$FC,$4C,$93,$FA,$26,$BF,$C5,$5B,$C5
	.dc.b	$FA,$06,$F0,$7E,$65,$B3,$F4,$1B,$C1,$F3,$9B,$7E,$09,$EC,$43,$B1
	.dc.b	$6F,$F5,$BF,$0F,$FE,$B1,$F6,$BF,$D7,$93,$FE,$6C,$6A,$9F,$F0,$2C
	.dc.b	$7F,$AF,$A3,$5F,$0B,$C7,$E1,$79,$7C,$2F,$3F,$85,$E9,$F0,$B2,$FA
	.dc.b	$7C,$2F,$7F,$85,$F0,$7C,$2C,$9F,$07,$C2,$F8,$FD,$F9,$1F,$A3,$E0
	.dc.b	$F6,$3E,$A2,$6F,$F1,$DF,$67,$1F,$C7,$69,$F1,$CB,$C7,$F1,$DB,$5D
	.dc.b	$2B,$94,$74,$AE,$4D,$39,$16,$B9,$37,$85,$EF,$FC,$BF,$87,$93,$F2
	.dc.b	$FE,$1F,$E7,$7C,$37,$CF,$F8,$2F,$8F,$3F,$3F,$0F,$8C,$FC,$FC,$8F
	.dc.b	$F8,$E6,$1C,$5F,$5F,$8B,$E5,$4F,$D7,$E1,$F9,$4F,$13,$87,$C1,$7D
	.dc.b	$DF,$86,$91,$FC,$37,$07,$C1,$FE,$04,$91,$F8,$16,$BC,$F8,$13,$C2
	.dc.b	$34,$43,$4E,$5B,$3C,$B8,$7F,$12,$DD,$3D,$B7,$D2,$31,$5E,$F6,$C6
	.dc.b	$63,$B5,$72,$7F,$A5,$F5,$7E,$67,$E1,$BF,$35,$F0,$9F,$22,$7D,$1F
	.dc.b	$C0,$3F,$EB,$3B,$FC,$CF,$6F,$E5,$E7,$FD,$93,$3F,$EB,$57,$F4,$1D
	.dc.b	$7C,$6B,$FC,$C7,$2F,$87,$C6,$BE,$88,$FF,$40,$DF,$2F,$87,$CB,$FE
	.dc.b	$6F,$87,$C2,$E3,$64,$7F,$94,$47,$C2,$FE,$C7,$BB,$F3,$F7,$F2,$BB
	.dc.b	$BF,$85,$78,$5F,$BC,$24,$7E,$61,$AF,$E4,$FB,$FF,$35,$A7,$E6,$36
	.dc.b	$FC,$D9,$1C,$12,$7F,$37,$19,$6C,$67,$3F,$BF,$DF,$F9,$A9,$FB,$BD
	.dc.b	$FF,$C1,$FB,$FF,$37,$1F,$C1,$FB,$BF,$36,$FE,$0B,$A7,$F3,$79,$7F
	.dc.b	$08,$EE,$FC,$C4,$FF,$0D,$F0,$3F,$2F,$3F,$C3,$3B,$BF,$89,$77,$7E
	.dc.b	$5A,$BF,$17,$DD,$F0,$9F,$80,$43,$E0,$70,$FC,$2B,$F5,$6F,$FB,$46
	.dc.b	$F0,$FC,$4A,$F6,$92,$75,$C3,$EF,$9D,$46,$27,$50,$ED,$F8,$57,$FC
	.dc.b	$8B,$87,$E1,$57,$F2,$23,$F9,$47,$85,$F0,$A7,$F9,$3F,$5F,$C2,$CB
	.dc.b	$F3,$FC,$3F,$BA,$BF,$E4,$67,$C3,$3F,$74,$43,$F5,$7B,$7E,$27,$A1
	.dc.b	$F0,$B4,$F8,$93,$F0,$8F,$89,$8F,$C0,$DF,$E2,$63,$ED,$6C,$FB,$1B
	.dc.b	$BF,$BC,$9B,$FC,$2C,$FF,$8D,$D7,$F4,$3D,$BE,$0E,$1E,$AE,$5F,$C0
	.dc.b	$CF,$F0,$0E,$5F,$81,$3E,$EB,$F8,$13,$FC,$67,$97,$F0,$2F,$F0,$6E
	.dc.b	$5F,$C0,$D7,$C4,$E5,$F7,$7C,$2F,$BF,$E1,$65,$FF,$17,$F0,$BF,$81
	.dc.b	$E5,$FC,$37,$87,$F8,$14,$FF,$0E,$3F,$82,$E3,$F3,$AB,$64,$8C,$9F
	.dc.b	$C3,$4B,$C1,$FC,$35,$E4,$4A,$EA,$4A,$E9,$1F,$C4,$0F,$EC,$DE,$17
	.dc.b	$F1,$07,$F8,$59,$FD,$AF,$C1,$FE,$1C,$FF,$12,$F8,$7F,$EC,$8F,$F1
	.dc.b	$0F,$0B,$EE,$7C,$3F,$DD,$F0,$FE,$F7,$87,$F7,$FD,$3F,$C0,$78,$9F
	.dc.b	$A4,$C3,$F8,$BF,$5F,$E0,$5F,$EC,$9D,$7F,$6B,$AF,$F3,$24,$70,$48
	.dc.b	$FC,$2E,$FF,$83,$69,$F8,$33,$FC,$01,$E1,$7D,$E7,$FA,$87,$67,$E7
	.dc.b	$9F,$E7,$BE,$3F,$F5,$97,$FC,$0D,$E4,$7F,$82,$79,$FE,$FF,$37,$F4
	.dc.b	$6E,$CF,$BF,$CF,$FE,$08,$EC,$FB,$F8,$7F,$5B,$EC,$FB,$EF,$F3,$2E
	.dc.b	$BF,$BE,$BD,$3F,$F4,$5F,$07,$A6,$3F,$3C,$4C,$3F,$3C,$FA,$B2,$FC
	.dc.b	$CC,$EE,$73,$3F,$C9,$7C,$4F,$65,$7F,$5A,$DF,$D9,$5F,$D1,$2B,$F7
	.dc.b	$27,$B0,$FE,$86,$7F,$88,$F4,$F6,$EA,$F2,$7F,$B8,$CF,$F6,$0D,$7D
	.dc.b	$B8,$FF,$3E,$EF,$FD,$C7,$C3,$CA,$FA,$E7,$9F,$97,$EB,$37,$6F,$47
	.dc.b	$2F,$D5,$8F,$73,$75,$F5,$9F,$C7,$B4,$FD,$AC,$FF,$34,$3F,$6C,$7F
	.dc.b	$83,$CF,$F1,$D7,$27,$D7,$7F,$BA,$F2,$7D,$77,$D8,$43,$FB,$CE,$DE
	.dc.b	$CB,$FE,$69,$1F,$C4,$DF,$EC,$7A,$7F,$13,$F8,$79,$3F,$B5,$3F,$57
	.dc.b	$8F,$ED,$72,$FD,$3C,$7E,$AD,$B7,$FA,$B8,$FD,$91,$92,$B9,$DB,$F6
	.dc.b	$8D,$AE,$93,$AA,$E6,$47,$04,$AE,$31,$FE,$C1,$E2,$FF,$03,$4F,$F9
	.dc.b	$77,$CB,$FE,$58,$FE,$DC,$FC,$09,$FB,$73,$F0,$27,$F8,$2B,$BB,$F6
	.dc.b	$EF,$F7,$F3,$FC,$13,$D9,$FB,$77,$FC,$17,$D7,$FB,$77,$FC,$23,$C7
	.dc.b	$F8,$17,$F3,$9E,$57,$DB,$CA,$47,$EC,$4F,$BE,$5D,$7E,$FC,$6B,$4F
	.dc.b	$CE,$CF,$E1,$CF,$F2,$DE,$BF,$82,$AF,$EA,$91,$F8,$23,$FC,$67,$E2
	.dc.b	$FF,$36,$8F,$F3,$CF,$8B,$FC,$DA,$7F,$C9,$95,$F9,$CE,$8F,$DB,$E3
	.dc.b	$FD,$5B,$B3,$F3,$72,$E9,$FD,$BD,$FF,$8B,$38,$FF,$02,$FB,$0F,$F1
	.dc.b	$07,$C3,$FF,$57,$23,$FE,$2E,$6B,$FE,$34,$6E,$7F,$E2,$E6,$E3,$FE
	.dc.b	$2D,$6C,$F4,$39,$10,$E4,$59,$E4,$D9,$E4,$D3,$FC,$69,$F0,$FF,$DA
	.dc.b	$FE,$1F,$FB,$67,$C3,$FF,$6D,$F8,$7F,$ED,$DF,$0F,$FD,$BF,$E1,$FF
	.dc.b	$B8,$7C,$3F,$F7,$1F,$87,$FF,$1D,$7F,$00,$B4,$BF,$11,$8F,$80,$CF
	.dc.b	$C0,$6B,$E1,$E7,$07,$F9,$EB,$87,$82,$EA,$64,$F8,$1E,$BF,$0F,$2B
	.dc.b	$FD,$03,$E5,$BB,$BD,$71,$F0,$09,$8F,$C0,$6E,$DA,$8C,$11,$C1,$23
	.dc.b	$E6,$1C,$0B,$1B,$6C,$6F,$2F,$F2,$CF,$3F,$D4,$6C,$FC,$D7,$FC,$BB
	.dc.b	$CF,$F1,$BE,$0B,$E6,$CF,$F9,$BB,$9F,$E6,$BF,$0F,$0F,$F4,$0E,$CF
	.dc.b	$87,$F9,$BF,$07,$F3,$BE,$1F,$E7,$FC,$27,$CE,$C3,$81,$F2,$6D,$CD
	.dc.b	$F4,$23,$3A,$FC,$CE,$4F,$A3,$1F,$E9,$9A,$7C,$3F,$DA,$F8,$59,$3F
	.dc.b	$3B,$8F,$C7,$F0,$FF,$78,$91,$D2,$D1,$88,$E8,$DE,$4F,$BB,$97,$DC
	.dc.b	$8C,$70,$74,$AE,$51,$87,$D4,$D9,$E1,$1A,$21,$D8,$B5,$DA,$FC,$71
	.dc.b	$E5,$F8,$6F,$CE,$3E,$7F,$87,$FC,$E3,$E9,$F8,$7F,$CE,$3F,$EA,$0E
	.dc.b	$EF,$CE,$3E,$DC,$FF,$3B,$F0,$FF,$9B,$8F,$81,$EA,$FE,$83,$D1,$FF
	.dc.b	$4D,$7F,$00,$BD,$8F,$D1,$F8,$1F,$97,$BF,$9D,$E0,$7F,$AC,$7C,$0F
	.dc.b	$CC,$47,$E2,$89,$C5,$F8,$87,$F2,$0D,$E2,$FC,$83,$75,$FC,$87,$C1
	.dc.b	$BC,$1F,$C8,$36,$FF,$20,$D7,$9B,$6F,$9B,$6B,$9D,$72,$F0,$7F,$2D
	.dc.b	$7F,$95,$21,$FE,$C8,$F1,$BF,$7F,$E0,$7E,$52,$3F,$05,$DF,$F9,$47
	.dc.b	$F0,$9D,$DF,$95,$7F,$1D,$E1,$FE,$43,$4F,$AD,$DF,$FC,$23,$BF,$F1
	.dc.b	$F5,$F9,$3E,$FF,$8D,$DF,$E4,$8F,$5E,$DE,$E2,$38,$21,$EE,$72,$D3
	.dc.b	$38,$7F,$B6,$36,$FC,$1C,$7F,$B8,$BC,$2F,$C1,$BE,$D3,$FD,$C3,$E3
	.dc.b	$7F,$1E,$7F,$DC,$7E,$37,$F1,$EE,$2F,$66,$1F,$8F,$F1,$BF,$5D,$5F
	.dc.b	$AB,$DB,$DD,$5F,$BA,$DB,$F0,$BE,$7C,$AF,$D6,$1F,$80,$E5,$FA,$4D
	.dc.b	$DB,$D3,$97,$D6,$6E,$1F,$0E,$7D,$84,$9D,$3E,$E8,$FF,$79,$6D,$F6
	.dc.b	$A3,$DC,$7D,$A9,$FF,$7C,$F2,$FD,$A8,$FD,$D7,$8D,$FC,$35,$FD,$E7
	.dc.b	$8D,$EE,$C3,$F8,$9F,$2F,$BB,$0F,$B3,$5F,$C0,$76,$FB,$4F,$F0,$EE
	.dc.b	$D7,$C0,$FE,$90,$FB,$71,$FB,$C3,$F6,$F8,$7F,$01,$7F,$74,$4B,$FD
	.dc.b	$D7,$8F,$FC,$07,$C7,$FD,$DC,$E6,$B8,$56,$A3,$99,$D4,$D1,$5D,$38
	.dc.b	$EC,$FE,$16,$DA,$EB,$3F,$C2,$7C,$7F,$C2,$BF,$C2,$FC,$7F,$C2,$BF
	.dc.b	$C3,$7C,$7F,$E0,$F6,$8F,$E3,$BC,$8F,$1F,$F8,$4C,$FF,$2D,$CF,$F8
	.dc.b	$4B,$FC,$9F,$A3,$FB,$3F,$8D,$FC,$26,$3E,$47,$8B,$FC,$21,$F9,$5D
	.dc.b	$3F,$87,$7F,$C1,$5B,$FF,$01,$7F,$0C,$7D,$F2,$9F,$83,$C3,$F0,$D9
	.dc.b	$FE,$12,$30,$69,$C1,$FC,$0B,$A3,$F0,$CF,$AB,$B7,$FA,$AC,$FE,$8C
	.dc.b	$FE,$8D,$E1,$7E,$FD,$FE,$91,$E1,$7E,$FD,$FE,$95,$D1,$F8,$77,$FA
	.dc.b	$97,$85,$FA,$27,$FA,$A7,$4F,$E1,$9F,$91,$99,$FF,$00,$74,$FE,$21
	.dc.b	$FC,$57,$4F,$E2,$36,$F8,$04,$E4,$FD,$3B,$7B,$7F,$4B,$5E,$DC,$3E
	.dc.b	$09,$1C,$12,$78,$3F,$C9,$B8,$3E,$FC,$F8,$E5,$6D,$FF,$03,$39,$87
	.dc.b	$6F,$E0,$B8,$7F,$03,$CD,$F8,$07,$FA,$95,$7C,$3C,$3F,$C3,$7C,$DF
	.dc.b	$A9,$8F,$EF,$BB,$FC,$59,$FF,$0F,$F7,$7D,$FB,$FE,$CF,$C0,$E5,$F8
	.dc.b	$5A,$7C,$02,$70,$7C,$0C,$3D,$7C,$1F,$79,$B8,$7D,$EC,$FE,$14,$61
	.dc.b	$B7,$CF,$83,$F0,$31,$ED,$E0,$FC,$0D,$FF,$06,$C3,$FB,$F7,$2F,$E0
	.dc.b	$63,$FB,$77,$2F,$E0,$5F,$EE,$1C,$BE,$FE,$B7,$37,$B5,$71,$7E,$06
	.dc.b	$BD,$BC,$DF,$C1,$72,$FE,$47,$BF,$ED,$27,$F6,$E4,$D7,$F7,$E4,$9F
	.dc.b	$DE,$36,$78,$37,$0F,$DD,$B7,$B3,$F8,$51,$2B,$F8,$3D,$70,$79,$91
	.dc.b	$FF,$28,$F6,$7E,$12,$3F,$C5,$9D,$9F,$82,$7F,$C5,$DD,$9F,$81,$C3
	.dc.b	$F0,$9D,$3F,$81,$7F,$C7,$27,$F0,$09,$FF,$17,$74,$FF,$00,$7F,$C5
	.dc.b	$1E,$37,$F8,$23,$B3,$F8,$04,$7E,$90,$FD,$1F,$4F,$F0,$1F,$13,$FC
	.dc.b	$0F,$AF,$D9,$9F,$CD,$74,$FE,$09,$FE,$05,$19,$24,$66,$3F,$08,$DE
	.dc.b	$1F,$C2,$47,$FA,$00,$FF,$1D,$79,$1F,$80,$D3,$F0,$51,$FE,$48,$AF
	.dc.b	$CF,$BF,$82,$3F,$A3,$9F,$E0,$03,$FA,$39,$FE,$00,$21,$FE,$00,$3F
	.dc.b	$A3,$10,$FF,$28,$76,$FE,$05,$F8,$18,$3B,$7F,$81,$47,$F2,$0A,$F4
	.dc.b	$5F,$E6,$3B,$7F,$41,$18,$9F,$87,$18,$23,$F1,$21,$DB,$F6,$F0,$FD
	.dc.b	$EF,$6F,$DB,$7F,$7F,$DD,$F6,$EF,$FD,$25,$7F,$CE,$DF,$F0,$09,$FE
	.dc.b	$43,$3D,$BA,$FF,$80,$5F,$F3,$17,$5F,$E7,$FA,$FF,$3D,$A7,$F0,$BD
	.dc.b	$5F,$0D,$FE,$46,$3F,$77,$B7,$9B,$0F,$AA,$DC,$7E,$AE,$9F,$6A,$31
	.dc.b	$87,$EE,$36,$FD,$BD,$F9,$B6,$FB,$0F,$A3,$6F,$EF,$8F,$F9,$F7,$C2
	.dc.b	$FE,$F8,$FF,$9F,$FC,$2F,$DB,$77,$FF,$13,$8F,$E6,$A7,$D5,$D3,$FB
	.dc.b	$EB,$F5,$36,$7B,$3F,$E5,$E3,$F9,$A7,$7F,$EB,$A7,$F5,$0D,$DF,$F5
	.dc.b	$84,$70,$4E,$FF,$D9,$4F,$F7,$16,$E1,$FB,$F8,$E7,$18,$DF,$F7,$9D
	.dc.b	$FF,$66,$DD,$FF,$65,$77,$FD,$9C,$BF,$9C,$F7,$FE,$CF,$2F,$F5,$26
	.dc.b	$3F,$B3,$CF,$EC,$47,$FA,$87,$C3,$FE,$E4,$FB,$FB,$BE,$CB,$FD,$53
	.dc.b	$9F,$EC,$3F,$C5,$BA,$BF,$21,$97,$D5,$8F,$C7,$74,$7D,$67,$F2,$66
	.dc.b	$08,$7E,$7A,$7F,$26,$FE,$C0,$FF,$1A,$65,$FB,$07,$FB,$86,$5F,$E9
	.dc.b	$BF,$0F,$F9,$84,$FF,$B1,$3C,$3F,$E6,$0F,$F8,$D2,$BF,$26,$FF,$B2
	.dc.b	$3B,$3F,$25,$1F,$6E,$3F,$25,$C9,$F6,$63,$D5,$97,$9F,$3F,$54,$7E
	.dc.b	$48,$F3,$E5,$F9,$2C,$FD,$E4,$AF,$EF,$13,$EB,$EB,$FD,$64,$FF,$67
	.dc.b	$8F,$E1,$1E,$5F,$FB,$71,$F8,$1E,$37,$F9,$AD,$F8,$5E,$37,$F9,$AD
	.dc.b	$F4,$9F,$ED,$9F,$17,$F2,$99,$FF,$B5,$E3,$D4,$7F,$B7,$7C,$1F,$A6
	.dc.b	$FE,$57,$8B,$F5,$6F,$A3,$7F,$E0,$CD,$3E,$99,$ED,$26,$DF,$E2,$17
	.dc.b	$E9,$1F,$E6,$62,$3F,$D8,$5F,$F1,$09,$FC,$80,$FF,$10,$9F,$C8,$0F
	.dc.b	$F1,$09,$EA,$B7,$3F,$F6,$28,$F4,$C7,$AD,$FF,$43,$1F,$E6,$7E,$0F
	.dc.b	$A8,$FF,$8F,$A7,$F4,$CF,$F5,$66,$BF,$A7,$31,$D5,$FA,$57,$F4,$E6
	.dc.b	$08,$6C,$FF,$2F,$75,$7E,$A7,$AB,$F5,$1D,$5F,$A7,$E2,$FE,$3D,$5F
	.dc.b	$50,$FE,$3C,$7F,$34,$C1,$E5,$FF,$B8,$63,$F5,$07,$F5,$A3,$FC,$B2
	.dc.b	$7F,$8E,$F4,$FF,$79,$6F,$F5,$9F,$F0,$DE,$FF,$5A,$7E,$8E,$FF,$59
	.dc.b	$A7,$D6,$9F,$C8,$72,$FE,$57,$2F,$F0,$B4,$FD,$69,$FF,$2B,$3F,$A8
	.dc.b	$3F,$90,$1F,$D5,$4F,$E4,$07,$F5,$53,$F9,$01,$F5,$72,$76,$FE,$39
	.dc.b	$FF,$0A,$C7,$D6,$6B,$FC,$87,$83,$EA,$F8,$9F,$DC,$B0,$FD,$14,$FE
	.dc.b	$37,$8B,$F5,$38,$FA,$23,$13,$A7,$FB,$8F,$3F,$E9,$BC,$2F,$EA,$13
	.dc.b	$FD,$B2,$7F,$4A,$7F,$6F,$EE,$FE,$31,$87,$E2,$B6,$FF,$16,$73,$79
	.dc.b	$FC,$3F,$DE,$DF,$F3,$5F,$86,$FD,$9E,$9F,$E4,$87,$E7,$3E,$38,$F9
	.dc.b	$B8,$79,$B0,$FA,$13,$92,$3F,$E5,$DE,$1E,$6F,$FB,$2A,$BE,$83,$FC
	.dc.b	$56,$7E,$8C,$7F,$16,$8F,$F3,$47,$03,$BB,$F8,$BC,$B0,$FD,$1F,$77
	.dc.b	$F1,$88,$C5,$7D,$27,$EA,$77,$7B,$31,$FE,$A2,$7F,$98,$BD,$2F,$1C
	.dc.b	$FF,$A1,$9F,$EA,$3D,$9F,$A1,$C7,$FA,$73,$70,$FD,$3D,$FF,$4B,$9C
	.dc.b	$DF,$E9,$23,$13,$AC,$FF,$C8,$D3,$FA,$FE,$CF,$D3,$3F,$C6,$8F,$CE
	.dc.b	$47,$BB,$AF,$F5,$1C,$5F,$5A,$D3,$FD,$34,$F8,$7C,$5F,$E3,$FB,$FD
	.dc.b	$57,$91,$FE,$3F,$79,$9F,$C6,$CF,$F7,$47,$17,$BB,$4F,$F1,$13,$F6
	.dc.b	$AF,$EB,$92,$7F,$4C,$D9,$FB,$91,$83,$FC,$33,$C5,$CF,$93,$FA,$7C
	.dc.b	$7E,$EF,$C7,$FF,$03,$D7,$F8,$8A,$3F,$C7,$DE,$3F,$F8,$23,$83,$D7
	.dc.b	$1F,$5B,$C7,$FF,$04,$C6,$2B,$F8,$EF,$8F,$FE,$0A,$AF,$F1,$09,$FA
	.dc.b	$9E,$8F,$F4,$E7,$8E,$E6,$F6,$36,$7F,$CF,$EB,$DA,$C9,$23,$E2,$E1
	.dc.b	$FE,$43,$8C,$47,$F8,$4D,$B8,$FF,$A9,$9B,$BF,$F8,$BD,$B3,$FE,$CC
	.dc.b	$6D,$73,$8E,$B7,$E1,$6C,$EA,$31,$1D,$CC,$9C,$3D,$3F,$5F,$A5,$EC
	.dc.b	$F8,$8F,$6F,$9D,$EE,$F7,$FD,$FE,$EF,$C0,$F5,$FE,$0F,$AF,$F0,$BE
	.dc.b	$1F,$E1,$FC,$47,$C4,$F8,$8F,$8B,$F1,$1F,$1B,$E0,$26,$FF,$0B,$C3
	.dc.b	$FE,$DE,$FF,$99,$97,$C4,$7E,$1D,$B1,$F3,$38,$FD,$44,$7D,$27,$AC
	.dc.b	$9C,$DF,$25,$CB,$5C,$F3,$7C,$B7,$2D,$72,$7C,$D3,$25,$33,$CD,$FE
	.dc.b	$7B,$E5,$F8,$F1,$EB,$F8,$1F,$8E,$FB,$7E,$23,$D5,$F1,$1F,$1E,$3E
	.dc.b	$07,$C1,$FC,$77,$E1,$7C,$47,$C7,$7E,$27,$C2,$7C,$67,$FC,$AD,$C1
	.dc.b	$F0,$DF,$89,$F1,$11,$FF,$0F,$9F,$CD,$3C,$5F,$E9,$84,$70,$7F,$3B
	.dc.b	$E1,$7C,$37,$CD,$F8,$8F,$BD,$3F,$4F,$E1,$FE,$F3,$F8,$AE,$CF,$BE
	.dc.b	$FE,$33,$AF,$EF,$BF,$8E,$EA,$FA,$DF,$0F,$F7,$A3,$EE,$67,$F6,$7D
	.dc.b	$5F,$B5,$F0,$FF,$6F,$AE,$57,$DB,$3E,$6F,$47,$E6,$5F,$9B,$E2,$7E
	.dc.b	$5B,$AF,$EC,$13,$AF,$EB,$36,$7E,$A3,$6B,$93,$6B,$95,$70,$25,$70
	.dc.b	$6D,$70,$6C,$BA,$FE,$ED,$7E,$57,$AF,$EE,$BF,$97,$EB,$FB,$AF,$EC
	.dc.b	$3C,$0F,$CC,$75,$FD,$D8,$FC,$E7,$5F,$DD,$7F,$3D,$D7,$F7,$75,$FE
	.dc.b	$21,$9B,$AF,$EF,$57,$E9,$7A,$FE,$F2,$F5,$FF,$DC,$F5,$EC,$8F,$51
	.dc.b	$38,$3D,$4D,$8D,$46,$08,$E0,$9D,$DF,$AB,$9F,$A2,$DC,$7E,$53,$FC
	.dc.b	$27,$83,$E1,$C7,$A3,$C4,$F8,$6F,$AB,$C3,$FC,$3D,$FF,$AB,$78,$3F
	.dc.b	$0E,$FF,$AB,$FC,$3F,$E0,$AF,$F1,$6F,$0F,$F1,$0F,$E7,$8F,$E3,$3B
	.dc.b	$FC,$27,$F8,$27,$AF,$FC,$0C,$FF,$3A,$EF,$F0,$B5,$F3,$73,$7D,$86
	.dc.b	$D7,$C1,$6F,$07,$C0,$6F,$07,$C0,$9F,$C8,$B5,$FF,$62,$70,$7B,$AD
	.dc.b	$1F,$DF,$FC,$27,$3F,$C0,$9C,$E2,$97,$3F,$E1,$2F,$E1,$1F,$C0,$0F
	.dc.b	$E6,$BB,$FF,$1D,$E6,$FC,$27,$1F,$F5,$1B,$73,$7E,$17,$2F,$F7,$77
	.dc.b	$0D,$C7,$C5,$F0,$1F,$C9,$4F,$ED,$4F,$F6,$5D,$3F,$B3,$93,$CE,$FF
	.dc.b	$1B,$65,$FE,$22,$9D,$43,$B7,$F9,$77,$89,$FC,$0E,$7F,$A4,$F4,$FF
	.dc.b	$31,$F0,$FF,$81,$C7,$FB,$9F,$C3,$FB,$B8,$FE,$63,$C3,$FB,$AF,$E6
	.dc.b	$BC,$3F,$E0,$75,$FD,$4B,$C6,$FE,$75,$9F,$F0,$38,$FE,$80,$7D,$B3
	.dc.b	$FA,$17,$5F,$DC,$F1,$FE,$D4,$7D,$4E,$BF,$AE,$47,$04,$3F,$2E,$5C
	.dc.b	$7F,$2E,$FF,$5B,$3F,$B1,$F9,$1F,$D2,$FA,$BE,$C4,$7D,$BE,$AF,$AD
	.dc.b	$5F,$B2,$EF,$FD,$CC,$FD,$63,$F9,$A9,$F9,$73,$F9,$A9,$F9,$62,$1F
	.dc.b	$99,$3F,$AD,$72,$FD,$57,$F8,$F1,$E6,$24,$F9,$9B,$CF,$F0,$1A,$7E
	.dc.b	$64,$F6,$39,$8F,$E3,$9D,$9F,$95,$9F,$F2,$97,$6B,$97,$DB,$D0,$87
	.dc.b	$81,$F6,$31,$FC,$BC,$33,$FB,$31,$8E,$EF,$DA,$E2,$CF,$F8,$AF,$81
	.dc.b	$FB,$67,$F8,$BF,$67,$E4,$F3,$EC,$FF,$77,$F3,$65,$FC,$0A,$F7,$13
	.dc.b	$1F,$51,$23,$DC,$7C,$0B,$FD,$B3,$F0,$0C,$CF,$C0,$7E,$D9,$FD,$BC
	.dc.b	$FF,$0F,$F2,$7F,$81,$F8,$7E,$DD,$FB,$B4,$FB,$75,$FC,$44,$FB,$67
	.dc.b	$B6,$3E,$1E,$7F,$E0,$FE,$EF,$66,$DF,$CC,$EB,$EE,$E5,$FB,$5F,$80
	.dc.b	$FF,$0A,$1F,$B0,$D3,$F5,$DD,$1F,$DB,$EB,$FD,$27,$DF,$FB,$0C,$FF
	.dc.b	$CB,$4F,$EB,$E3,$FC,$82,$D7,$FC,$41,$DF,$FB,$17,$C8,$7D,$CE,$4F
	.dc.b	$D9,$3F,$E3,$CE,$4F,$B4,$FE,$03,$BB,$EC,$BF,$82,$EE,$FB,$3A,$FF
	.dc.b	$0B,$F2,$3E,$C6,$7F,$C5,$BC,$1F,$04,$FF,$0F,$F2,$3F,$97,$3F,$C4
	.dc.b	$79,$FE,$CE,$3F,$D3,$F9,$FE,$B7,$0F,$D4,$8F,$AD,$C7,$FC,$9E,$FE
	.dc.b	$E4,$7D,$6C,$B7,$39,$6F,$77,$F4,$99,$FB,$27,$F6,$B3,$EC,$9F,$E2
	.dc.b	$73,$FC,$89,$E5,$7F,$89,$DF,$F2,$27,$8F,$FD,$AF,$0F,$E4,$DE,$2F
	.dc.b	$F2,$8C,$3D,$9D,$1F,$63,$B3,$F2,$9C,$0E,$9F,$AF,$97,$F3,$AF,$19
	.dc.b	$C1,$F0,$FD,$4F,$F2,$71,$27,$FC,$A8,$D7,$FC,$A8,$DD,$3F,$9B,$37
	.dc.b	$3E,$53,$C0,$87,$02,$CF,$06,$C7,$07,$44,$8D,$3F,$E6,$3F,$80,$FF
	.dc.b	$54,$7F,$00,$7F,$A6,$23,$F0,$BC,$68,$FF,$67,$46,$17,$AB,$C3,$D9
	.dc.b	$E3,$E8,$F2,$F8,$8E,$7F,$11,$D3,$D6,$EB,$EE,$F6,$F5,$FB,$FB,$BE
	.dc.b	$0F,$63,$C3,$F0,$3E,$2F,$88,$F1,$FB,$F3,$83,$E6,$73,$7F,$62,$E1
	.dc.b	$FE,$C5,$1F,$E9,$E6,$BE,$5E,$9F,$59,$1D,$35,$D1,$82,$18,$27,$2F
	.dc.b	$FA,$DE,$7A,$B6,$79,$36,$78,$B5,$E3,$CD,$E7,$9F,$9B,$F0,$9F,$3A
	.dc.b	$3E,$7F,$C1,$7D,$0E,$6F,$95,$7E,$BF,$81,$F9,$4F,$B7,$E1,$3E,$AF
	.dc.b	$C3,$7C,$A8,$F8,$1C,$7F,$63,$E0,$BE,$54,$7C,$3E,$2F,$96,$FF,$B0
	.dc.b	$BF,$80,$5F,$0F,$FE,$E5,$6C,$7C,$C3,$FC,$21,$DF,$F8,$F6,$9F,$8A
	.dc.b	$29,$F8,$9D,$BF,$19,$C3,$F8,$A2,$38,$24,$70,$8C,$B6,$33,$DD,$FE
	.dc.b	$1B,$AF,$03,$67,$C0,$D7,$F2,$C7,$AF,$E2,$3D,$5F,$13,$EB,$F8,$9F
	.dc.b	$67,$C4,$FE,$4A,$BF,$09,$E0,$FE,$49,$FC,$37,$7F,$C1,$F8,$8F,$85
	.dc.b	$F1,$1F,$92,$9F,$C5,$78,$1F,$17,$CD,$FC,$94,$7E,$37,$BF,$F2,$4F
	.dc.b	$F1,$71,$B7,$E9,$7B,$7F,$07,$C3,$F7,$E7,$F0,$6D,$7F,$06,$70,$EE
	.dc.b	$FC,$5C,$67,$8F,$F1,$04,$70,$49,$E1,$5D,$9A,$76,$29,$DA,$FC,$8E
	.dc.b	$5F,$DD,$77,$7E,$46,$3F,$51,$DD,$F9,$17,$F5,$5E,$2F,$F0,$2E,$EF
	.dc.b	$C8,$C7,$EF,$FB,$BF,$22,$FE,$BF,$C3,$FD,$D3,$FB,$1E,$DF,$C8,$BF
	.dc.b	$C1,$7B,$7F,$22,$FF,$0A,$EC,$FC,$7B,$7B,$7F,$C6,$4F,$CA,$8F,$F5
	.dc.b	$13,$F9,$7F,$77,$F2,$C4,$70,$43,$A3,$96,$99,$E1,$70,$FF,$9A,$2B
	.dc.b	$AC,$E8,$93,$A7,$C5,$5C,$E3,$1C,$0A,$E3,$18,$6D,$F5,$AE,$0D,$BF
	.dc.b	$D8,$7C,$07,$EB,$FD,$7F,$D8,$78,$9F,$B1,$F5,$FF,$65,$EB,$FE,$CF
	.dc.b	$DC,$FD,$A7,$B1,$FB,$5F,$7B,$F6,$DF,$07,$FB,$7F,$7B,$F7,$1F,$07
	.dc.b	$FB,$9F,$53,$FD,$0F,$B7,$F3,$6F,$E0,$1A,$DF,$87,$CF,$27,$AA,$7F
	.dc.b	$D1,$73,$FE,$8E,$8C,$7C,$16,$B8,$FF,$D1,$B5,$FE,$8C,$AF,$E0,$9C
	.dc.b	$C8,$7B,$7B,$9F,$D0,$C6,$1E,$95,$CA,$75,$5C,$5B,$5B,$AF,$44,$60
	.dc.b	$8F,$23,$AB,$87,$66,$67,$A3,$67,$A6,$1E,$FA,$E0,$D9,$D4,$62,$FC
	.dc.b	$D1,$83,$EB,$9F,$C9,$F9,$3E,$C3,$D7,$DB,$F4,$FB,$7F,$5E,$3C,$1E
	.dc.b	$DF,$B3,$E1,$FE,$BC,$7F,$2F,$E1,$FA,$EF,$F3,$1E,$0F,$83,$F0,$3F
	.dc.b	$0B,$E2,$7E,$1F,$C3,$7C,$4E,$29,$7F,$43,$BB,$F1,$AF,$E5,$0F,$CE
	.dc.b	$6F,$F8,$D7,$F3,$2D,$E1,$FC,$A7,$77,$E4,$A7,$81,$A2,$C6,$9E,$66
	.dc.b	$27,$91,$1C,$13,$93,$F4,$55,$D5,$B1,$F4,$7E,$23,$F3,$AF,$F4,$FE
	.dc.b	$DF,$CE,$BF,$57,$E1,$3F,$3A,$FF,$53,$EB,$FC,$1F,$C3,$7E,$13,$E1
	.dc.b	$7F,$0B,$F1,$3F,$86,$F8,$9F,$C3,$DF,$E2,$3B,$7F,$13,$F1,$3F,$8A
	.dc.b	$F8,$89,$3F,$96,$EF,$FC,$0F,$0F,$DC,$8F,$C3,$93,$8B,$F0,$4D,$8D
	.dc.b	$46,$08,$E0,$91,$F9,$23,$F0,$C7,$13,$05,$E1,$FC,$9B,$6B,$C6,$EA
	.dc.b	$BF,$84,$78,$DF,$8C,$8F,$E4,$9D,$FF,$A4,$EF,$FC,$6C,$7F,$77,$F1
	.dc.b	$3F,$4F,$DF,$F8,$D8,$FE,$F5,$E1,$FF,$03,$7F,$BE,$78,$7F,$C0,$DF
	.dc.b	$D1,$77,$7E,$35,$FD,$27,$6F,$E3,$1F,$E0,$DD,$BF,$CC,$37,$FB,$31
	.dc.b	$F8,$56,$ED,$F8,$58,$C7,$6F,$E4,$A7,$3C,$3F,$85,$6C,$FE,$16,$74
	.dc.b	$4E,$DF,$C3,$36,$3F,$1F,$3C,$09,$1C,$38,$7E,$26,$1F,$CB,$7C,$6F
	.dc.b	$E4,$D7,$FE,$19,$F1,$BF,$60,$FF,$4B,$F1,$BF,$93,$3F,$10,$FD,$81
	.dc.b	$F1,$0F,$D8,$1F,$C1,$4F,$C5,$1F,$10,$FE,$11,$A7,$C5,$F8,$6F,$F3
	.dc.b	$BC,$7D,$7D,$7F,$03,$A7,$DC,$CB,$F0,$B3,$92,$46,$67,$7C,$DF,$80
	.dc.b	$8C,$4E,$A3,$04,$F1,$FF,$89,$46,$5B,$3C,$09,$B3,$6F,$8B,$87,$F5
	.dc.b	$BE,$7F,$C4,$61,$FC,$87,$C7,$F8,$B9,$FF,$3F,$E6,$FC,$44,$FF,$71
	.dc.b	$E6,$FC,$43,$FD,$CF,$9B,$F1,$0F,$F4,$9E,$6F,$C4,$3F,$D2,$F9,$BF
	.dc.b	$11,$87,$F4,$68,$FE,$29,$CD,$F7,$08,$FD,$CD,$BE,$F3,$F8,$88,$CB
	.dc.b	$4D,$38,$9F,$C7,$13,$5F,$C6,$70,$BC,$3F,$E1,$77,$FE,$09,$F0,$FF
	.dc.b	$85,$BF,$D7,$FC,$3F,$E1,$6F,$F8,$37,$C3,$FE,$16,$FF,$43,$F0,$BF
	.dc.b	$85,$BF,$D1,$7C,$1F,$E1,$6F,$F4,$7F,$07,$F8,$5B,$FD,$27,$C0,$FE
	.dc.b	$16,$FF,$8F,$7A,$3E,$96,$9F,$5E,$3F,$80,$ED,$F8,$29,$CF,$4F,$D8
	.dc.b	$AD,$57,$F0,$18,$C1,$1E,$3B,$FE,$1A,$3E,$FF,$0F,$DD,$9F,$EA,$04
	.dc.b	$3F,$BE,$78,$1F,$C1,$1F,$F1,$57,$81,$FC,$11,$FE,$C1,$E3,$FF,$50
	.dc.b	$7F,$C5,$DE,$3F,$F5,$07,$FC,$61,$E3,$FF,$50,$7F,$B3,$78,$FF,$D4
	.dc.b	$1F,$D0,$90,$F8,$1A,$7F,$B6,$DF,$ED,$9F,$0F,$FD,$E0,$8E,$08,$7F
	.dc.b	$84,$5C,$B4,$CE,$FF,$E1,$18,$CB,$63,$3B,$7F,$84,$27,$2D,$8C,$9F
	.dc.b	$E0,$E8,$CB,$4C,$EF,$FE,$15,$8C,$B6,$33,$7E,$B7,$C8,$4B,$F1,$36
	.dc.b	$FC,$0D,$3B,$16,$BA,$3C,$BC,$1F,$E0,$F1,$8A,$E6,$4A,$E6,$DA,$E6
	.dc.b	$D9,$E1,$3A,$24,$E9,$B1,$A8,$C1,$1C,$12,$FD,$D5,$D5,$B5,$C9,$B5
	.dc.b	$C5,$B5,$C7,$0F,$93,$18,$23,$82,$4F,$CC,$7E,$FE,$1F,$0A,$5E,$FF
	.dc.b	$F8,$F8,$95,$FE,$35,$6B,$FE,$57,$6E,$3F,$E5,$56,$E3,$FE,$52,$6C
	.dc.b	$FF,$93,$9B,$8F,$F9,$4D,$B7,$E0,$9E,$C4,$3B,$16,$FF,$CC,$5F,$0F
	.dc.b	$FE,$54,$E5,$FF,$40,$4A,$7F,$CF,$F0,$F4,$B5,$F0,$7B,$F4,$F8,$7B
	.dc.b	$BC,$7E,$1F,$97,$C3,$F3,$F7,$FA,$7A,$BD,$7E,$13,$B7,$BD,$DF,$E1
	.dc.b	$FC,$1E,$BF,$87,$E1,$FC,$5E,$FC,$E2,$FF,$0D,$74,$FC,$E7,$F2,$1C
	.dc.b	$BF,$49,$FA,$04,$33,$1B,$E4,$FC,$03,$7A,$3F,$00,$D9,$FC,$03,$7A
	.dc.b	$3F,$01,$1C,$EB,$8B,$6B,$8B,$FB,$EE,$9F,$BD,$7C,$70,$F2,$4F,$88
	.dc.b	$E0,$76,$2E,$1D,$9D,$13,$0E,$ED,$C3,$BB,$4E,$45,$9E,$53,$C0,$E3
	.dc.b	$87,$EC,$B9,$FF,$DD,$75,$FE,$F4,$6D,$1A,$4F,$7B,$F0,$44,$9F,$F4
	.dc.b	$64,$65,$B1,$97,$F0,$D8,$1E,$D1,$9C,$3F,$12,$D9,$D5,$F8,$63,$15
	.dc.b	$C0,$EA,$43,$A9,$E9,$F7,$BF,$2D,$EF,$7E,$5F,$DE,$FC,$C7,$C3,$FE
	.dc.b	$67,$DF,$FF,$56,$FF,$00,$FF,$BB,$34,$FF,$6B,$AF,$88,$CF,$C4,$6B
	.dc.b	$E2,$37,$F1,$1C,$3E,$23,$8F,$C4,$72,$F8,$8E,$7E,$BF,$4F,$7F,$AF
	.dc.b	$BF,$38,$FF,$2A,$7D,$CE,$9F,$C0,$EF,$F4,$F9,$FE,$CB,$67,$EC,$B7
	.dc.b	$9F,$EC,$CE,$89,$3A,$6C,$6A,$30,$47,$04,$AF,$C8,$F4,$FE,$07,$A3
	.dc.b	$E6,$CF,$CF,$AF,$99,$87,$D4,$6E,$1F,$53,$A7,$EF,$47,$D0,$AE,$0D
	.dc.b	$8C,$D7,$18,$C5,$F6,$9C,$DF,$38,$C1,$9E,$9F,$B3,$5F,$8C,$E8,$FB
	.dc.b	$2F,$E3,$B9,$FE,$CF,$C3,$FD,$AE,$7F,$AF,$3F,$ED,$DE,$87,$3F,$D8
	.dc.b	$8C,$73,$7D,$87,$5C,$BF,$61,$E1,$C9,$F5,$DF,$65,$B9,$3F,$D0,$0F
	.dc.b	$F9,$7A,$33,$F0,$9F,$8A,$23,$E1,$6B,$A3,$04,$30,$4E,$2F,$C8,$C7
	.dc.b	$06,$C6,$DA,$EF,$87,$F4,$B5,$E3,$9C,$92,$33,$5C,$DB,$5C,$DA,$F3
	.dc.b	$1C,$6B,$E9,$3A,$BE,$95,$A9,$CB,$6F,$B3,$4E,$C5,$3B,$5F,$58,$F9
	.dc.b	$67,$07,$11,$C1,$AF,$09,$EF,$18,$AF,$1B,$6F,$CC,$4A,$F0,$4E,$AB
	.dc.b	$89,$2F,$C0,$D9,$F0,$35,$F0,$46,$88,$E9,$AE,$8C,$1F,$C7,$79,$3F
	.dc.b	$C3,$B2,$F6,$BF,$87,$92,$7F,$2C,$D7,$F1,$ED,$BF,$E2,$6D,$C3,$EF
	.dc.b	$B6,$7E,$93,$71,$FB,$0D,$BF,$04,$F6,$21,$D8,$B7,$FC,$83,$E1,$3F
	.dc.b	$91,$FA,$7F,$C9,$3E,$13,$F9,$2F,$B7,$FC,$9B,$E0,$7F,$93,$FA,$BF
	.dc.b	$CA,$3E,$07,$F9,$4F,$AD,$FC,$EF,$3F,$E0,$5F,$C0,$35,$BE,$EE,$75
	.dc.b	$70,$FF,$3C,$A7,$C0,$E6,$31,$F0,$B7,$0F,$C9,$D7,$FB,$0C,$7B,$99
	.dc.b	$E4,$43,$E1,$70,$BE,$16,$9F,$8E,$BF,$AA,$7D,$72,$67,$F8,$5E,$8F
	.dc.b	$B6,$D8,$DB,$5D,$E7,$F8,$DC,$7F,$21,$1F,$8E,$8C,$73,$7E,$26,$B8
	.dc.b	$12,$33,$7D,$9B,$5E,$4E,$8F,$C3,$CE,$63,$A1,$30,$EE,$4C,$3B,$5F
	.dc.b	$29,$F2,$9B,$8C,$B6,$BA,$57,$16,$CE,$A3,$13,$F0,$2B,$A6,$1E,$98
	.dc.b	$E4,$7F,$19,$C3,$FC,$F3,$93,$DE,$FA,$24,$9F,$EF,$D5,$F8,$07,$ED
	.dc.b	$C6,$5F,$C3,$B6,$35,$87,$DC,$8C,$61,$F8,$76,$D7,$23,$AC,$65,$A6
	.dc.b	$4E,$A6,$27,$A3,$6F,$D0,$DA,$EB,$3A,$8F,$CD,$FC,$2F,$E3,$3D,$FF
	.dc.b	$C6,$FC,$2F,$E3,$BE,$17,$F1,$FF,$11,$F9,$0F,$86,$FC,$8F,$C4,$7E
	.dc.b	$4B,$CC,$FC,$9F,$C4,$7E,$53,$E2,$3F,$2B,$EB,$7E,$5B,$E1,$BF,$2F
	.dc.b	$F0,$DF,$9B,$FD,$A7,$FC,$D7,$CB,$F0,$9F,$F3,$57,$0F,$A5,$7C,$36
	.dc.b	$CF,$CE,$47,$E0,$38,$FE,$CB,$7A,$3E,$E1,$3A,$7E,$F9,$23,$EF,$1F
	.dc.b	$70,$E0,$E5,$A6,$A7,$AE,$1F,$A3,$E9,$FC,$0F,$3F,$C8,$E8,$F9,$77
	.dc.b	$F4,$09,$5F,$72,$32,$D9,$E0,$49,$FA,$44,$38,$F4,$FD,$D6,$9F,$76
	.dc.b	$79,$3B,$6B,$F6,$0F,$BD,$D1,$F6,$1F,$C0,$73,$FD,$87,$F0,$5C,$DF
	.dc.b	$61,$FC,$27,$2F,$D8,$7F,$0D,$C9,$F6,$1F,$C4,$71,$FD,$87,$EA,$65
	.dc.b	$F6,$7E,$1F,$EA,$47,$D7,$E2,$FE,$49,$BF,$FB,$7B,$8D,$F0,$FF,$7C
	.dc.b	$95,$F7,$5A,$FF,$8A,$E3,$6E,$48,$66,$37,$BF,$E1,$DB,$8F,$F9,$A2
	.dc.b	$B8,$D6,$CE,$35,$B6,$CE,$EB,$8C,$7F,$80,$AB,$8D,$6E,$39,$11,$C1
	.dc.b	$2B,$94,$74,$39,$57,$4E,$25,$F4,$AD,$D7,$BF,$E1,$FF,$44,$47,$F4
	.dc.b	$0D,$74,$7F,$6D,$E4,$F7,$7C,$3F,$EA,$49,$1F,$A7,$6B,$F2,$DB,$97
	.dc.b	$EB,$5B,$7C,$A7,$81,$0E,$05,$9E,$0D,$8E,$0E,$89,$1A,$7E,$D7,$C3
	.dc.b	$FE,$F7,$E1,$FF,$7D,$F0,$FF,$BF,$F8,$7F,$E0,$1F,$0F,$FC,$07,$E1
	.dc.b	$FF,$81,$7C,$3F,$F0,$3F,$87,$FE,$09,$F0,$FF,$C1,$7E,$03,$F8,$37
	.dc.b	$C0,$7C,$3D,$BF,$83,$71,$5E,$0F,$C0,$37,$83,$ED,$3F,$92,$E2,$F9
	.dc.b	$F5,$F5,$E7,$EA,$9F,$5F,$93,$EA,$F2,$FD,$7E,$5F,$AD,$3A,$3F,$01
	.dc.b	$8F,$22,$39,$BE,$97,$CA,$3A,$57,$28,$C4,$72,$38,$57,$28,$CD,$72
	.dc.b	$AE,$04,$BF,$21,$2F,$AB,$FC,$A7,$E1,$FF,$95,$7C,$0F,$F2,$BF,$87
	.dc.b	$FE,$59,$F0,$5F,$E0,$7F,$5F,$FD,$7D,$A7,$FB,$31,$7C,$17,$C5,$CB
	.dc.b	$F4,$1D,$7F,$84,$69,$F7,$CA,$7D,$EA,$FC,$0E,$BF,$7C,$8E,$09,$1C
	.dc.b	$23,$2D,$8C,$ED,$FA,$4A,$F0,$36,$7C,$0D,$7F,$16,$7F,$43,$EA,$FC
	.dc.b	$5B,$FD,$17,$A7,$F1,$6F,$F4,$7E,$8F,$C5,$BF,$D2,$79,$FF,$16,$FF
	.dc.b	$4B,$E6,$FC,$5B,$FA,$E2,$1F,$AE,$26,$7F,$8E,$73,$D9,$FC,$6F,$8F
	.dc.b	$E9,$47,$DB,$CF,$EC,$EB,$F6,$F9,$3E,$C4,$FD,$62,$1B,$72,$DA,$E7
	.dc.b	$87,$E0,$5B,$5C,$DB,$3F,$88,$25,$F8,$1F,$E3,$1C,$FF,$A1,$8F,$E1
	.dc.b	$FC,$F7,$13,$4F,$C0,$47,$CD,$C7,$EA,$B7,$1F,$AB,$4E,$0F,$AA,$DC
	.dc.b	$7E,$AB,$5F,$C2,$1F,$74,$95,$CD,$B1,$BB,$EE,$DB,$EF,$7D,$BB,$3F
	.dc.b	$84,$5F,$F8,$07,$B3,$F8,$43,$FE,$03,$EC,$FE,$10,$FF,$81,$7B,$3F
	.dc.b	$84,$3F,$E0,$7E,$CF,$E1,$0F,$F8,$27,$6F,$D1,$7C,$17,$E8,$FE,$1F
	.dc.b	$F4,$9F,$07,$FA,$5F,$87,$FD,$37,$C2,$4C,$3F,$97,$1F,$3B,$83,$EB
	.dc.b	$F0,$FD,$58,$FB,$6D,$7E,$D1,$8E,$1F,$B0,$D3,$97,$27,$DA,$E3,$FB
	.dc.b	$04,$BE,$B5,$C2,$B9,$92,$FF,$24,$FE,$52,$3F,$8F,$BF,$98,$8C,$9A
	.dc.b	$E5,$FE,$E1,$C9,$F6,$09,$3F,$61,$AF,$D8,$C3,$EB,$13,$0F,$A9,$1F
	.dc.b	$62,$77,$7F,$61,$FA,$D1,$FB,$33,$EB,$1F,$B3,$3E,$B1,$FB,$33,$EB
	.dc.b	$1F,$B3,$3E,$B1,$FB,$33,$EB,$1F,$E5,$7E,$C6,$FF,$5E,$31,$B7,$D7
	.dc.b	$75,$AF,$D7,$59,$BC,$1F,$E3,$DF,$0F,$F9,$EC,$4D,$25,$7D,$ED,$53
	.dc.b	$A8,$C3,$5E,$B5,$D1,$CD,$72,$72,$2B,$9C,$66,$FC,$93,$9B,$E9,$18
	.dc.b	$AD,$CF,$42,$1D,$1C,$B6,$78,$DF,$8E,$31,$7D,$E3,$17,$D6,$B7,$5E
	.dc.b	$1A,$D3,$4E,$F3,$B8,$CD,$F8,$23,$13,$AB,$EA,$47,$91,$E2,$34,$62
	.dc.b	$38,$B9,$23,$82,$47,$39,$F0,$36,$FD,$31,$8A,$E9,$39,$BF,$3D,$68
	.dc.b	$8E,$09,$3E,$4A,$E4,$49,$DC,$65,$AF,$9A,$BB,$39,$AE,$6E,$4F,$B9
	.dc.b	$F0,$BF,$8F,$23,$F8,$A6,$BF,$65,$B8,$7D,$56,$D7,$F7,$56,$CF,$43
	.dc.b	$91,$0E,$45,$9E,$4D,$9E,$4D,$3F,$21,$F0,$FF,$A3,$F8,$6F,$D2,$7C
	.dc.b	$37,$E9,$6B,$F4,$DF,$0D,$FA,$7F,$86,$FD,$47,$C3,$7E,$A7,$E1,$BF
	.dc.b	$55,$B7,$EA,$FE,$1B,$F5,$9F,$0D,$FA,$DF,$86,$FD,$77,$C3,$7F,$4C
	.dc.b	$FE,$01,$70,$5F,$87,$C7,$C3,$C8,$F9,$44,$3E,$57,$37,$D3,$E6,$F9
	.dc.b	$BC,$BF,$84,$9F,$BE,$49,$D3,$C6,$35,$18,$8E,$7A,$FD,$A9,$CB,$78
	.dc.b	$BE,$DC,$EA,$FB,$C6,$2B,$71,$89,$F0,$57,$16,$C6,$6F,$C1,$5C,$27
	.dc.b	$C9,$7C,$C8,$F0,$32,$53,$C4,$62,$79,$C7,$06,$C6,$E7,$B3,$FC,$17
	.dc.b	$27,$C3,$FD,$12,$47,$F1,$D8,$CB,$63,$2F,$16,$CF,$F1,$39,$DC,$E4
	.dc.b	$E0,$43,$04,$8E,$13,$CA,$FF,$03,$6B,$E7,$5A,$BE,$4D,$7D,$5F,$15
	.dc.b	$F8,$9F,$88,$FC,$57,$C4,$7E,$2F,$E2,$3F,$19,$EF,$FE,$37,$E2,$3F
	.dc.b	$1D,$F0,$1F,$8F,$F8,$8F,$C8,$7C,$0F,$E4,$7E,$23,$F2,$5F,$05,$F9
	.dc.b	$3F,$88,$FC,$A7,$C1,$FE,$57,$E2,$3F,$2D,$F0,$9F,$97,$F8,$89,$C5
	.dc.b	$F4,$7A,$7E,$F7,$37,$D0,$23,$82,$74,$FE,$05,$B3,$C6,$30,$47,$04
	.dc.b	$3E,$F9,$39,$BE,$11,$3A,$3E,$4F,$47,$C8,$6F,$4F,$CD,$AF,$E9,$D3
	.dc.b	$96,$C6,$71,$FA,$0D,$9D,$46,$27,$E8,$D7,$46,$9C,$8B,$1B,$7E,$F4
	.dc.b	$6D,$FA,$E7,$E3,$7A,$3E,$BB,$F8,$FE,$7F,$AE,$FE,$47,$9B,$EB,$BF
	.dc.b	$93,$E5,$FA,$EF,$E5,$79,$3E,$BB,$F9,$7E,$3F,$AE,$FE,$67,$8B,$EB
	.dc.b	$BF,$9B,$E1,$FA,$EF,$E8,$7E,$17,$F8,$81,$1F,$D0,$35,$D1,$82,$18
	.dc.b	$27,$07,$E8,$63,$83,$63,$6D,$77,$AF,$D6,$9D,$B6,$36,$FD,$AF,$88
	.dc.b	$FF,$37,$FF,$00,$BD,$1B,$F0,$F8,$F8,$79,$87,$F8,$4E,$9E,$A7,$0F
	.dc.b	$77,$8F,$A7,$CB,$DD,$E7,$F0,$5D,$3D,$EE,$BF,$05,$DB,$DF,$EF,$F0
	.dc.b	$5E,$0F,$80,$F0,$FC,$14,$DB,$CA,$7E,$2B,$93,$D2,$D3,$3D,$DF,$B9
	.dc.b	$8C,$B6,$33,$DE,$9C,$B6,$33,$18,$23,$83,$DD,$F0,$3F,$13,$97,$E1
	.dc.b	$C7,$C1,$E4,$F8,$6F,$F8,$87,$93,$E1,$BF,$17,$8F,$E1,$BF,$1F,$8B
	.dc.b	$E1,$BF,$27,$87,$E1,$BF,$2F,$87,$E1,$BF,$37,$E0,$BE,$13,$D7,$7F
	.dc.b	$AB,$F0,$52,$3E,$93,$5F,$13,$7A,$BF,$C7,$19,$A3,$1D,$5F,$C8,$23
	.dc.b	$15,$C8,$87,$22,$D7,$2A,$E3,$1F,$E4,$99,$E2,$EC,$93,$B7,$8D,$7E
	.dc.b	$35,$FC,$47,$C3,$FE,$3B,$E1,$FF,$1F,$F0,$FF,$90,$F8,$7F,$F5,$FF
	.dc.b	$F0,$0F,$FC,$00,$BF,$F3,$03,$FE,$6E,$6F,$C3,$EB,$E0,$F7,$F0,$FC
	.dc.b	$3E,$13,$8F,$9B,$F8,$DB,$FC,$B7,$7F,$E3,$5E,$7F,$09,$D8,$FB,$EF
	.dc.b	$E9,$8F,$C6,$17,$BB,$F1,$4D,$EE,$FC,$43,$7B,$B8,$F1,$FE,$BC,$9C
	.dc.b	$7F,$AD,$9D,$1F,$8B,$3A,$72,$3B,$BF,$19,$1F,$AB,$F1,$3F,$17,$7F
	.dc.b	$63,$BB,$F6,$7E,$27,$EB,$6B,$F6,$1E,$27,$EB,$5F,$D9,$78,$9F,$AD
	.dc.b	$7F,$14,$43,$F2,$9E,$27,$EF,$3B,$BF,$17,$8F,$E1,$3B,$BD,$3B,$79
	.dc.b	$E3,$D3,$7F,$09,$BC,$3F,$09,$BC,$3F,$09,$BC,$3C,$AB,$85,$F3,$23
	.dc.b	$82,$5F,$38,$CB,$FD,$C7,$C5,$F8,$73,$FA,$BE,$0F,$87,$5F,$97,$E0
	.dc.b	$F8,$73,$FA,$D3,$E1,$9F,$AD,$3E,$19,$FA,$D3,$E1,$63,$FA,$2E,$4F
	.dc.b	$D6,$C7,$ED,$FC,$5F,$87,$7F,$A3,$D7,$D5,$9F,$96,$BC,$FA,$FB,$EB
	.dc.b	$CE,$D8,$CE,$BC,$23,$15,$C4,$8E,$09,$1B,$3F,$97,$F1,$FC,$0C,$FF
	.dc.b	$6F,$AF,$C0,$C3,$F5,$DA,$FC,$07,$F8,$77,$89,$FC,$31,$FE,$21,$E2
	.dc.b	$7F,$0C,$7F,$89,$78,$9F,$C3,$1F,$E2,$9E,$27,$C0,$C7,$F8,$1F,$1F
	.dc.b	$C0,$9F,$DD,$70,$7E,$F3,$77,$47,$EF,$2F,$F7,$55,$FB,$CE,$1F,$E1
	.dc.b	$1E,$26,$EB,$51,$89,$FD,$F9,$FC,$16,$74,$47,$37,$DA,$31,$1D,$4E
	.dc.b	$07,$F9,$0F,$97,$FA,$9F,$89,$FC,$1E,$3F,$A8,$78,$7F,$C1,$DF,$E0
	.dc.b	$BF,$11,$FC,$23,$E1,$BF,$84,$F8,$7F,$C1,$67,$F8,$4F,$85,$FC,$15
	.dc.b	$FE,$17,$E0,$FF,$05,$7F,$9C,$76,$7F,$05,$7F,$DB,$5E,$23,$C2,$FE
	.dc.b	$0D,$1F,$B9,$E2,$FC,$FB,$78,$7F,$07,$D9,$F6,$E3,$07,$E1,$0B,$C5
	.dc.b	$F8,$28,$FE,$AB,$C7,$FC,$06,$7F,$B5,$78,$FF,$D3,$1F,$E9,$5D,$7F
	.dc.b	$81,$BF,$A7,$E1,$FF,$01,$9F,$E9,$DE,$1F,$F0,$17,$FA,$87,$87,$FC
	.dc.b	$05,$FB,$E7,$E8,$33,$FE,$7D,$E1,$FF,$01,$8F,$EF,$1D,$5F,$81,$C7
	.dc.b	$C9,$1E,$6F,$0B,$F2,$AD,$DB,$D8,$FF,$1A,$26,$FF,$09,$B8,$FF,$3D
	.dc.b	$7F,$B1,$6D,$EB,$9F,$E3,$51,$FD,$9C,$FF,$04,$6B,$EB,$8F,$EA,$BA
	.dc.b	$FA,$DF,$EA,$FA,$FA,$DF,$EB,$3D,$BF,$5D,$FE,$B7,$B7,$D7,$7F,$8D
	.dc.b	$61,$EC,$3D,$04,$3C,$A7,$D3,$2E,$BF,$01,$BA,$FC,$07,$3C,$1F,$55
	.dc.b	$F5,$9F,$CD,$4F,$51,$F5,$8F,$59,$FC,$4C,$FA,$A7,$F1,$33,$EA,$9F
	.dc.b	$C4,$CF,$AA,$4A,$F5,$BF,$DF,$CF,$51,$FD,$FC,$FD,$3C,$7E,$B9,$FD
	.dc.b	$29,$FA,$7A,$FD,$53,$AE,$CF,$D5,$D7,$EE,$09,$B7,$ED,$67,$85,$F3
	.dc.b	$BE,$31,$8B,$E3,$5B,$BF,$F5,$77,$67,$EC,$AB,$FC,$81,$E3,$FF,$78
	.dc.b	$7F,$C8,$5E,$3F,$F7,$8E,$CF,$D9,$47,$F7,$63,$FB,$CF,$8F,$FD,$D9
	.dc.b	$FE,$F7,$E3,$FF,$76,$7F,$BA,$90,$FE,$68,$7F,$76,$21,$F5,$B9,$7F
	.dc.b	$27,$97,$E4,$49,$9F,$EC,$A7,$F0,$11,$87,$F0,$0F,$F5,$AD,$BE,$E3
	.dc.b	$FE,$24,$E5,$FC,$B5,$7F,$4D,$AF,$BB,$3F,$E2,$8E,$5F,$CB,$4F,$F8
	.dc.b	$9F,$97,$F2,$CF,$F8,$A7,$97,$F2,$CF,$F8,$AF,$97,$F9,$85,$FD,$CC
	.dc.b	$3F,$C8,$9E,$0F,$EB,$F3,$73,$FF,$87,$BE,$1A,$57,$F8,$79,$AF,$F1
	.dc.b	$26,$DB,$1F,$F6,$BF,$5A,$73,$18,$6D,$39,$D5,$B8,$EC,$43,$B1,$6F
	.dc.b	$B3,$C6,$38,$3A,$24,$6A,$7A,$D7,$81,$EF,$F0,$DF,$E3,$AF,$86,$FF
	.dc.b	$1D,$E3,$FE,$3C,$F8,$6F,$F1,$EF,$C3,$7F,$8F,$BE,$1B,$FC,$7F,$F0
	.dc.b	$DF,$E4,$0F,$63,$FC,$83,$F0,$DF,$E4,$2F,$86,$FF,$21,$FC,$37,$F9
	.dc.b	$13,$E1,$A7,$08,$70,$F8,$FC,$2C,$46,$63,$1F,$0D,$87,$5F,$0D,$78
	.dc.b	$3F,$E0,$0D,$A3,$17,$F2,$63,$FD,$A6,$4F,$7B,$3C,$9F,$E6,$B6,$C6
	.dc.b	$67,$C1,$7D,$27,$24,$8C,$9D,$5C,$3E,$A8,$E0,$EA,$BB,$36,$BB,$3B
	.dc.b	$39,$19,$AE,$78,$7C,$36,$DF,$BB,$E1,$BE,$A5,$7F,$9B,$23,$EA,$BF
	.dc.b	$E6,$EE,$2F,$5F,$BF,$EC,$F8,$6F,$A9,$3E,$1F,$80,$F7,$FC,$37,$C0
	.dc.b	$F8,$1F,$A7,$3F,$DA,$38,$7E,$1F,$C3,$7C,$4F,$82,$F8,$BF,$0D,$2B
	.dc.b	$F1,$DA,$FE,$5D,$A7,$E3,$FB,$BF,$15,$DF,$F9,$0E,$2F,$C7,$36,$B8
	.dc.b	$57,$E4,$9B,$3A,$AE,$24,$9E,$BC,$9F,$96,$6D,$F3,$7C,$07,$FA,$3B
	.dc.b	$BF,$ED,$FC,$1F,$DC,$F8,$6F,$CC,$4F,$D1,$F8,$4F,$CC,$3F,$E9,$4E
	.dc.b	$FF,$C0,$FC,$37,$E6,$23,$FD,$2F,$DD,$F9,$87,$EC,$7C,$2F,$E1,$BE
	.dc.b	$1B,$F0,$F1,$F8,$8E,$FF,$C4,$F7,$C8,$FB,$44,$3E,$F6,$FF,$82,$6F
	.dc.b	$7F,$E0,$B8,$BE,$F4,$6D,$AE,$BB,$FF,$02,$FE,$18,$C1,$BE,$4F,$C7
	.dc.b	$B8,$24,$F2,$24,$69,$B1,$FA,$5E,$FF,$C5,$BF,$E2,$89,$FC,$5B,$FE
	.dc.b	$2A,$EF,$FC,$5B,$F9,$9E,$EF,$C5,$BE,$AE,$EF,$C5,$BF,$9D,$ED,$FC
	.dc.b	$5E,$3F,$80,$ED,$FC,$5C,$FE,$87,$B3,$F1,$4F,$C2,$EC,$FB,$CF,$E0
	.dc.b	$CF,$B6,$5D,$FF,$00,$DE,$BF,$C0,$46,$30,$FC,$44,$FE,$13,$B3,$9C
	.dc.b	$62,$38,$46,$36,$FC,$79,$FC,$CB,$B3,$E0,$D7,$E5,$A3,$F0,$F5,$F0
	.dc.b	$4F,$F6,$CF,$87,$FC,$79,$FF,$6D,$F8,$7F,$87,$7F,$C9,$FD,$3F,$87
	.dc.b	$7F,$54,$7F,$B8,$36,$FC,$3B,$FC,$07,$9F,$F0,$CF,$E0,$B9,$FE,$EE
	.dc.b	$BF,$85,$CF,$EA,$B7,$3F,$B0,$4A,$FC,$14,$EB,$9F,$F1,$53,$9A,$E1
	.dc.b	$5A,$8C,$4E,$A3,$04,$9D,$92,$1C,$DF,$12,$BF,$AB,$9F,$84,$BF,$DD
	.dc.b	$F3,$7E,$13,$0F,$CD,$F3,$7E,$12,$7F,$9D,$73,$7E,$12,$7F,$3F,$CD
	.dc.b	$F8,$4C,$3F,$B0,$ED,$F8,$49,$FD,$1F,$37,$E1,$30,$FE,$85,$CD,$F5
	.dc.b	$E7,$F0,$71,$F6,$C9,$7F,$6A,$32,$DD,$BF,$08,$DE,$6F,$C3,$93,$9B
	.dc.b	$A4,$EA,$B8,$91,$C1,$25,$CD,$FC,$32,$7F,$9D,$76,$FE,$1B,$1F,$EB
	.dc.b	$1D,$BF,$C3,$27,$FD,$23,$DB,$FC,$31,$FE,$85,$83,$C3,$FE,$C9,$18
	.dc.b	$F0,$FF,$A8,$F7,$7F,$0D,$8F,$E8,$C7,$F0,$D3,$FA,$47,$47,$E2,$32
	.dc.b	$FC,$07,$6F,$E2,$3A,$7E,$E6,$FF,$66,$30,$4B,$FE,$0C,$7D,$F9,$D9
	.dc.b	$23,$51,$FC,$53,$A7,$F0,$0F,$F8,$27,$8B,$F0,$07,$F8,$63,$A7,$F0
	.dc.b	$18,$FF,$77,$E9,$FC,$04,$FF,$72,$E9,$FC,$04,$FF,$7B,$E9,$FC,$06
	.dc.b	$1F,$E1,$3E,$9F,$C0,$4F,$F7,$FE,$9F,$C0,$61,$FD,$EB,$A7,$EA,$F2
	.dc.b	$7D,$7C,$3F,$78,$43,$F7,$65,$BF,$C1,$BF,$66,$38,$3A,$24,$7F,$52
	.dc.b	$E8,$F6,$F0,$FF,$8B,$30,$FB,$47,$B6,$BF,$7E,$7F,$45,$3F,$7E,$7D
	.dc.b	$AE,$5F,$B2,$FF,$B3,$B9,$7D,$B5,$FE,$36,$D3,$F7,$F3,$FC,$FB,$AB
	.dc.b	$EC,$61,$FE,$16,$E5,$F9,$F9,$FD,$AA,$FA,$B8,$7D,$96,$F2,$FD,$56
	.dc.b	$F2,$F0,$C3,$F7,$6D,$8F,$B9,$CB,$EC,$DF,$FC,$83,$1F,$CE,$CF,$F0
	.dc.b	$21,$FE,$11,$F0,$FF,$C0,$8F,$B0,$FE,$74,$7F,$85,$7C,$3F,$F0,$23
	.dc.b	$FC,$64,$87,$F5,$8F,$0F,$FC,$08,$FF,$86,$F8,$FE,$CB,$F4,$7B,$7E
	.dc.b	$C9,$1F,$AD,$9F,$EC,$E3,$11,$F5,$23,$0F,$F7,$CE,$4F,$D7,$3F,$D9
	.dc.b	$BB,$7E,$AB,$FE,$3B,$ED,$FD,$77,$0F,$F8,$33,$B7,$F5,$D3,$FF,$00
	.dc.b	$39,$FE,$AC,$FF,$84,$39,$FF,$5B,$3F,$E3,$1F,$07,$EA,$CF,$F9,$32
	.dc.b	$3F,$5A,$43,$FB,$4E,$7F,$39,$FA,$24,$32,$7E,$AF,$8F,$E9,$72,$FD
	.dc.b	$98,$CB,$6B,$EF,$46,$39,$7F,$96,$BF,$C1,$3C,$5F,$D6,$BF,$AF,$3E
	.dc.b	$B9,$FE,$06,$E4,$FA,$EF,$F8,$2B,$95,$E0,$FF,$96,$3B,$7F,$5F,$A7
	.dc.b	$F9,$8B,$B7,$F2,$74,$EC,$FD,$7C,$FF,$99,$BB,$7E,$BE,$CD,$5D,$1F
	.dc.b	$53,$7F,$D5,$CF,$F2,$1E,$8C,$61,$F5,$B6,$FB,$07,$DF,$3E,$E9,$4F
	.dc.b	$EE,$DD,$1F,$93,$8F,$B3,$5F,$E9,$AE,$FF,$B3,$1F,$EA,$2E,$7F,$E1
	.dc.b	$37,$FE,$97,$E7,$FB,$33,$FE,$A6,$E7,$F5,$EF,$F6,$23,$FC,$FF,$DD
	.dc.b	$FA,$CB,$FF,$3F,$77,$7D,$2E,$0F,$D4,$47,$CF,$E8,$C1,$1F,$B5,$C1
	.dc.b	$F5,$09,$87,$F6,$6E,$EF,$44,$FF,$9B,$4F,$A7,$3F,$E9,$5E,$8F,$A6
	.dc.b	$FF,$97,$3A,$3E,$94,$FE,$A4,$FF,$11,$1F,$A8,$8F,$F1,$13,$FE,$81
	.dc.b	$ED,$FD,$4B,$FE,$21,$3F,$54,$7C,$AE,$2F,$A0,$DE,$7F,$E3,$75,$F5
	.dc.b	$1B,$97,$D3,$8C,$72,$FD,$6A,$FA,$51,$FC,$7D,$FF,$48,$78,$9F,$E6
	.dc.b	$37,$FA,$9F,$8B,$FE,$A0,$EE,$FA,$71,$FE,$C7,$E6,$FD,$5D,$FF,$9E
	.dc.b	$39,$BE,$9C,$FF,$B2,$F9,$BF,$56,$FF,$96,$BB,$BE,$A3,$FE,$AF,$E6
	.dc.b	$FA,$9B,$FE,$37,$9B,$E8,$4F,$D3,$3E,$8E,$FF,$AD,$70,$4C,$BE,$9B
	.dc.b	$FA,$D3,$67,$F1,$E3,$F5,$07,$D5,$3F,$8F,$1F,$E3,$73,$F5,$07,$F9
	.dc.b	$5C,$FE,$3C,$FF,$85,$1F,$AC,$7F,$85,$0F,$AD,$1F,$E1,$47,$EB,$1F
	.dc.b	$E5,$73,$EA,$9F,$5C,$F3,$3E,$98,$F3,$D7,$EF,$E3,$04,$9F,$51,$C0
	.dc.b	$B3,$C3,$9B,$F2,$ED,$E6,$F8,$ED,$9F,$83,$7D,$5E,$37,$D6,$30,$4B
	.dc.b	$E7,$19,$6C,$F0,$24,$EE,$FB,$D6,$DB,$3A,$AE,$35,$FE,$46,$E2,$FC
	.dc.b	$8C,$FE,$3A,$FF,$24,$47,$04,$9F,$CD,$BF,$C2,$1B,$87,$E6,$23,$FD
	.dc.b	$C9,$F0,$DF,$EF,$2F,$3F,$FD,$E7,$EA,$FF,$BD,$BF,$80,$1E,$8D,$F8
	.dc.b	$5C,$7B,$59,$F4,$F5,$F0,$BB,$F5,$38,$7C,$2F,$1F,$57,$97,$C2,$F3
	.dc.b	$F5,$BA,$7B,$FD,$7D,$EE,$DF,$0F,$38,$FE,$7F,$47,$C0,$27,$4F,$C4
	.dc.b	$E7,$F9,$13,$A8,$C4,$EA,$30,$43,$F0,$13,$C1,$B1,$B6,$CF,$22,$3C
	.dc.b	$79,$FF,$04,$E8,$C7,$47,$D7,$69,$F5,$E7,$E4,$7A,$F2,$07,$FB,$4B
	.dc.b	$99,$19,$8C,$36,$D7,$C3,$15,$F0,$BE,$75,$A5,$3A,$BE,$57,$C6,$F8
	.dc.b	$57,$63,$8C,$66,$77,$1C,$A7,$77,$D6,$FA,$4F,$03,$55,$C2,$33,$1D
	.dc.b	$49,$3C,$69,$F0,$FF,$8B,$F6,$BF,$19,$9F,$E3,$7E,$1F,$F1,$DF,$01
	.dc.b	$F8,$FF,$87,$FC,$87,$C0,$FC,$8C,$FF,$DF,$3F,$C0,$2F,$82,$C6,$BE
	.dc.b	$58,$7C,$46,$38,$7E,$44,$3E,$23,$18,$7C,$68,$7C,$26,$38,$90,$F8
	.dc.b	$8C,$2F,$85,$BF,$11,$8F,$88,$CF,$A7,$32,$FC,$0E,$3F,$8D,$DB,$F0
	.dc.b	$E4,$D7,$F0,$D3,$F8,$92,$46,$5F,$C4,$1C,$B9,$3F,$29,$19,$AF,$9A
	.dc.b	$FC,$BE,$8F,$A1,$D1,$F5,$B0,$FB,$58,$7D,$5B,$DC,$62,$3E,$9B,$6F
	.dc.b	$94,$6E,$B9,$3F,$70,$C5,$F6,$23,$82,$1D,$0F,$C4,$74,$7D,$D7,$A6
	.dc.b	$3F,$75,$ED,$CD,$F7,$5F,$07,$2F,$CE,$E5,$F9,$FC,$BF,$43,$97,$E8
	.dc.b	$F2,$FD,$2E,$5F,$A7,$C9,$F4,$B1,$F7,$72,$B9,$3E,$94,$7E,$1B,$93
	.dc.b	$EC,$FB,$52,$36,$D7,$F1,$6D,$E3,$FC,$53,$79,$3F,$12,$D9,$FC,$43
	.dc.b	$79,$7F,$0E,$DC,$7F,$DA,$B3,$D8,$87,$62,$DF,$67,$D3,$EF,$FE,$3B
	.dc.b	$0F,$C7,$FA,$9F,$90,$F6,$FF,$23,$F0,$BF,$92,$F5,$7F,$27,$EC,$FE
	.dc.b	$53,$D6,$FC,$AF,$C2,$FE,$5B,$D7,$FC,$BF,$BB,$35,$FC,$17,$F0,$0F
	.dc.b	$EA,$56,$FF,$BB,$87,$B1,$FA,$15,$EC,$5F,$1E,$5F,$A3,$87,$E1,$B4
	.dc.b	$F8,$65,$E5,$F8,$6D,$E4,$F8,$6F,$C1,$26,$5F,$05,$B5,$D1,$A7,$42
	.dc.b	$D7,$46,$D7,$4A,$E4,$7F,$01,$E4,$FD,$B9,$F2,$A7,$E0,$FB,$DE,$DF
	.dc.b	$7B,$DD,$F0,$BE,$FF,$83,$F8,$1F,$0B,$F0,$7E,$13,$E1,$7C,$17,$C9
	.dc.b	$C7,$F8,$7F,$1F,$C5,$7E,$37,$1F,$C7,$F8,$4F,$91,$F0,$92,$7F,$47
	.dc.b	$BF,$EB,$1F,$D2,$70,$FE,$C9,$FD,$61,$8F,$33,$F5,$F1,$92,$38,$27
	.dc.b	$99,$FB,$09,$E0,$DA,$EC,$4E,$4F,$E8,$DC,$EF,$84,$FC,$65,$FE,$0B
	.dc.b	$E1,$7F,$18,$FE,$A3,$BF,$F1,$8F,$EA,$BB,$BF,$18,$FE,$B3,$B7,$F1
	.dc.b	$8F,$E2,$BE,$17,$F0,$DF,$0F,$F8,$7F,$87,$FC,$47,$C3,$FE,$27,$C0
	.dc.b	$FC,$3D,$FE,$3B,$6F,$1F,$81,$FA,$5F,$1A,$4F,$E1,$5A,$FD,$C6,$F0
	.dc.b	$FD,$CA,$FC,$23,$67,$F0,$8D,$E4,$FC,$23,$7C,$1F,$C2,$57,$F4,$0F
	.dc.b	$07,$F0,$77,$FC,$1B,$C6,$FC,$F7,$85,$F8,$38,$FD,$2F,$83,$F8,$37
	.dc.b	$F4,$FE,$07,$E0,$DF,$D4,$F7,$FE,$0F,$3F,$E9,$DD,$FF,$83,$9F,$D6
	.dc.b	$F7,$FE,$0D,$FD,$7F,$77,$E0,$DF,$61,$EF,$21,$82,$57,$A7,$83,$E0
	.dc.b	$ED,$F0,$09,$3F,$01,$BC,$1F,$01,$BC,$0E,$0F,$BF,$5F,$D9,$B8,$3E
	.dc.b	$FB,$FD,$A3,$A7,$EF,$BF,$DA,$BA,$7E,$FB,$FD,$B3,$A7,$EF,$BF,$A0
	.dc.b	$3F,$96,$EF,$EF,$7E,$E1,$FE,$0E,$EC,$FE,$12,$FF,$0E,$E6,$FC,$0E
	.dc.b	$FF,$C3,$79,$BD,$73,$E9,$26,$FE,$96,$ED,$F6,$1B,$C1,$C2,$74,$49
	.dc.b	$D3,$63,$51,$82,$6F,$F8,$0A,$F8,$47,$F0,$23,$F9,$1F,$85,$FC,$09
	.dc.b	$FE,$4B,$E1,$7F,$02,$7F,$93,$F8,$5F,$C0,$9F,$82,$43,$E0,$E2,$D7
	.dc.b	$F9,$66,$9F,$CA,$F9,$7F,$05,$5A,$9F,$F7,$DB,$F8,$43,$F8,$2F,$27
	.dc.b	$F7,$12,$3F,$C1,$2B,$F7,$4D,$8C,$DF,$F0,$4F,$03,$F8,$04,$63,$C1
	.dc.b	$FE,$0B,$5D,$0E,$04,$AE,$0E,$5F,$47,$83,$F8,$6C,$3F,$A1,$78,$3F
	.dc.b	$86,$7F,$C5,$5C,$FF,$86,$7F,$C5,$9C,$FF,$86,$7F,$C5,$DC,$FF,$86
	.dc.b	$7F,$A5,$78,$3F,$C1,$32,$FE,$9F,$E0,$7F,$04,$7F,$08,$7F,$52,$E7
	.dc.b	$FC,$2E,$DF,$98,$E7,$FB,$24,$9F,$D2,$E1,$FA,$36,$CE,$DA,$7E,$18
	.dc.b	$FE,$D2,$43,$EF,$9F,$47,$BF,$F7,$CF,$D3,$EF,$FD,$F3,$F5,$7B,$FF
	.dc.b	$7C,$FD,$7E,$FF,$DF,$5F,$F9,$23,$B7,$F4,$18,$FD,$DA,$FE,$D7,$E1
	.dc.b	$7E,$F6,$3F,$B5,$78,$1E,$99,$FD,$DD,$F9,$DA,$79,$CB,$CB,$F9,$97
	.dc.b	$D6,$D7,$FA,$EF,$6F,$D5,$9F,$F2,$CF,$81,$F5,$5F,$F2,$DF,$2F,$D5
	.dc.b	$7F,$CB,$BC,$BF,$55,$FF,$2F,$F2,$FD,$57,$D5,$87,$E6,$73,$FF,$00
	.dc.b	$3F,$D6,$8F,$AA,$7E,$5F,$8F,$EA,$BE,$78,$FA,$4F,$A3,$BF,$F5,$DC
	.dc.b	$9E,$F9,$C9,$2F,$E0,$36,$BF,$7B,$A7,$EC,$E3,$FC,$ED,$A7,$F3,$29
	.dc.b	$FD,$99,$FC,$C8,$FD,$99,$FC,$C8,$FD,$99,$FC,$C8,$F5,$57,$B1,$FD
	.dc.b	$71,$EB,$3F,$61,$BF,$D5,$8F,$EC,$7B,$FE,$A2,$7F,$49,$D3,$FA,$87
	.dc.b	$FB,$59,$FA,$57,$18,$7E,$CB,$83,$F5,$33,$A2,$46,$67,$77,$FD,$C3
	.dc.b	$A7,$EC,$57,$F2,$5E,$9F,$B0,$FF,$A3,$F8,$BE,$C3,$FE,$91,$E2,$FB
	.dc.b	$0F,$FA,$4F,$8B,$EC,$3F,$CB,$7A,$7F,$58,$FF,$0B,$E9,$FD,$63,$FC
	.dc.b	$37,$C0,$FA,$FB,$7F,$13,$E5,$FA,$9A,$FD,$22,$57,$DB,$6C,$FE,$3F
	.dc.b	$7F,$AC,$D8,$CF,$2F,$F2,$AA,$FD,$65,$7D,$87,$F9,$51,$FC,$77,$C0
	.dc.b	$FE,$54,$FF,$1F,$F0,$3F,$95,$3F,$C8,$7C,$0F,$E5,$4F,$F4,$B2,$57
	.dc.b	$E4,$9F,$ED,$27,$EB,$0F,$E4,$C7,$F2,$9E,$4F,$F2,$A7,$C3,$48,$FF
	.dc.b	$2D,$B5,$FE,$5C,$D9,$FF,$2E,$37,$2F,$F1,$1B,$67,$FA,$C3,$73,$FE
	.dc.b	$74,$DF,$00,$D2,$4F,$80,$FF,$35,$FC,$37,$F9,$9F,$E1,$BF,$D9,$1C
	.dc.b	$2F,$86,$BF,$0D,$2F,$FD,$65,$28,$DF,$C3,$70,$F8,$6E,$3F,$0D,$CB
	.dc.b	$E1,$B9,$FC,$37,$4F,$86,$EB,$F0,$DD,$BE,$1B,$BF,$C3,$78,$3E,$1A
	.dc.b	$6D,$DB,$E1,$BC,$64,$70,$43,$FA,$CE,$E8,$D4,$61,$BC,$FF,$EF,$F9
	.dc.b	$D4,$F1,$8C,$4E,$B9,$D2,$E7,$5F,$08,$E8,$71,$9D,$CE,$AB,$94,$F1
	.dc.b	$AE,$17,$DA,$3E,$79,$F2,$BC,$FF,$9F,$C9,$F4,$3E,$0B,$E8,$FC,$37
	.dc.b	$D2,$F8,$3F,$9B,$7F,$43,$97,$EA,$FC,$37,$D6,$F8,$4F,$9B,$3F,$53
	.dc.b	$93,$EC,$FC,$37,$CD,$8F,$AF,$F0,$BF,$31,$F8,$FC,$7F,$82,$7F,$1E
	.dc.b	$7E,$0C,$A4,$DF,$F2,$2E,$5A,$E7,$8F,$F2,$6E,$5A,$E4,$FC,$A9,$92
	.dc.b	$99,$E3,$FC,$BB,$95,$FC,$02,$F8,$6F,$CA,$4F,$FC,$01,$87,$C3,$7E
	.dc.b	$4E,$3F,$03,$DD,$71,$63,$F5,$2D,$F0,$DF,$72,$FF,$C3,$F0,$F8,$6F
	.dc.b	$CB,$C7,$E7,$3B,$B5,$87,$28,$FC,$15,$3E,$1B,$38,$F5,$F8,$6F,$CD
	.dc.b	$D7,$97,$E1,$FF,$36,$F9,$FE,$1F,$6F,$27,$F0,$BC,$1F,$87,$8F,$C5
	.dc.b	$1F,$87,$EE,$CF,$77,$E2,$31,$FC,$6B,$67,$47,$0A,$E4,$FE,$50,$9A
	.dc.b	$7E,$3D,$B5,$CA,$FA,$92,$39,$79,$0F,$1F,$F8,$0F,$81,$F9,$49,$FD
	.dc.b	$FF,$7F,$E5,$1F,$E0,$3D,$DF,$94,$7F,$63,$DD,$F9,$47,$F8,$2F,$6F
	.dc.b	$E5,$1F,$DA,$F6,$FE,$51,$FE,$13,$E1,$7F,$0E,$ED,$FC,$A6,$FF,$B5
	.dc.b	$ED,$F8,$FC,$3F,$82,$F8,$7B,$D0,$69,$31,$FC,$23,$5F,$BA,$DB,$FB
	.dc.b	$55,$CE,$76,$43,$65,$8D,$D7,$78,$C7,$1F,$E2,$EF,$A3,$63,$27,$E1
	.dc.b	$A3,$CF,$ED,$FA,$3D,$BF,$4D,$7A,$BD,$B9,$7F,$9B,$7D,$BF,$0F,$EE
	.dc.b	$F6,$3D,$FF,$0F,$F0,$3D,$9F,$83,$F0,$FF,$0B,$DA,$F8,$7F,$0F,$F1
	.dc.b	$3E,$03,$E3,$FF,$00,$7F,$9E,$35,$FE,$26,$FF,$A6,$B5,$FF,$3E,$B7
	.dc.b	$8B,$E1,$AF,$73,$89,$27,$F9,$0C,$EA,$31,$1C,$9B,$19,$8D,$D7,$82
	.dc.b	$72,$D8,$C9,$FC,$AE,$BC,$0E,$39,$11,$88,$EC,$65,$C5,$F7,$8C,$91
	.dc.b	$C1,$2B,$F8,$07,$17,$D1,$9E,$5E,$E7,$C6,$F8,$5F,$8F,$EE,$7C,$8F
	.dc.b	$85,$F9,$3E,$EF,$CA,$F8,$5F,$97,$EF,$7D,$1C,$7F,$88,$71,$FD,$17
	.dc.b	$F8,$97,$17,$D0,$F8,$5F,$A3,$1F,$4B,$E1,$A3,$F8,$4E,$2F,$C8,$3F
	.dc.b	$8E,$27,$0F,$E3,$BB,$3F,$17,$1F,$89,$38,$76,$FE,$45,$DF,$1F,$E3
	.dc.b	$DA,$6C,$B7,$DA,$FA,$47,$2B,$EB,$39,$AF,$C2,$FC,$37,$E2,$3E,$07
	.dc.b	$F2,$B7,$F9,$32,$1F,$C9,$BB,$BF,$2A,$FF,$31,$ED,$FC,$AB,$FC,$CF
	.dc.b	$B7,$F2,$3F,$0B,$F9,$2F,$85,$FC,$9F,$C2,$FE,$56,$BE,$D7,$C4,$7E
	.dc.b	$5B,$D0,$91,$F7,$7B,$FF,$0E,$FE,$15,$BB,$FE,$11,$BD,$FF,$82,$E0
	.dc.b	$FC,$65,$7E,$05,$BC,$5F,$81,$8E,$47,$52,$1B,$32,$5B,$EC,$EA,$39
	.dc.b	$47,$F5,$2E,$FF,$C5,$4F,$EA,$BB,$FF,$14,$FE,$57,$BB,$F1,$4F,$E5
	.dc.b	$FB,$7F,$14,$FE,$67,$B3,$F1,$4F,$F5,$0E,$CF,$C5,$3F,$D4,$BB,$3F
	.dc.b	$14,$FE,$7C,$FC,$49,$FB,$8E,$BF,$E9,$2B,$F4,$37,$97,$FD,$6E,$24
	.dc.b	$61,$C1,$FE,$1A,$70,$E1,$AF,$08,$DB,$B6,$BD,$DB,$1C,$23,$FB,$53
	.dc.b	$5E,$07,$27,$0F,$13,$11,$C8,$8E,$08,$79,$5C,$B4,$C9,$E7,$29,$E2
	.dc.b	$76,$E4,$E0,$43,$04,$38,$1C,$8F,$49,$B3,$25,$32,$71,$29,$A3,$07
	.dc.b	$BC,$E0,$53,$31,$F1,$09,$1E,$88,$E4,$43,$A9,$0E,$46,$4A,$66,$39
	.dc.b	$BE,$B3,$24,$72,$F9,$1A,$F9,$23,$C4,$F3,$6B,$C5,$B1,$C2,$3C,$4D
	.dc.b	$78,$1E,$17,$0F,$13,$11,$C8,$8E,$08,$79,$5C,$B4,$C9,$E7,$29,$E2
	.dc.b	$76,$E4,$E0,$43,$04,$38,$1C,$8F,$49,$B3,$25,$32,$71,$29,$A3,$07
	.dc.b	$BC,$FD,$77,$C1,$FE,$A0,$8F,$E9,$5A,$E8,$C1,$0C,$10,$FD,$EB,$C1
	.dc.b	$A6,$CA,$6C,$FD,$AB,$B6,$9B,$29,$B3,$F8,$33,$B6,$9B,$29,$B3,$F8
	.dc.b	$93,$B6,$9B,$29,$B9,$46,$DB,$1B,$6B,$B8,$D1,$1D,$35,$D1,$82,$18
	.dc.b	$21,$F1,$08,$7C,$02,$9E,$C2,$9E,$82,$C7,$91,$B1,$E1,$69,$E0,$29
	.dc.b	$E0,$3B,$10,$EC,$53,$B1,$63,$B3,$63,$B3,$4E,$C5,$3B,$1D,$08,$74
	.dc.b	$2C,$74,$6C,$74,$69,$D0,$A7,$43,$91,$23,$93,$63,$93,$4E,$45,$39
	.dc.b	$47,$02,$47,$06,$C7,$06,$BC,$23,$47,$F5,$4D,$BF,$C8,$BF,$13,$FE
	.dc.b	$5E,$24,$FF,$91,$5A,$FF,$97,$9B,$C1,$FE,$5E,$6F,$0F,$F9,$91,$B3
	.dc.b	$FE,$53,$6D,$76,$8E,$84,$3A,$16,$BA,$36,$BF,$CD,$FF,$0F,$FE,$72
	.dc.b	$F4,$BF,$CE,$7F,$0F,$FE,$9A,$FE,$01,$7C,$25,$26,$1F,$E6,$18,$7B
	.dc.b	$FB,$F8,$8E,$1C,$1C,$77,$E5,$EF,$73,$F8,$8E,$9E,$AF,$5F,$88,$ED
	.dc.b	$EB,$77,$F8,$89,$AF,$5F,$5B,$C4,$47,$04,$1C,$9F,$EA,$6D,$D3,$B9
	.dc.b	$CF,$21,$72,$AB,$71,$CE,$F9,$5F,$16,$C6,$6F,$95,$F1,$AD,$46,$2F
	.dc.b	$9D,$78,$0E,$EE,$2F,$9C,$66,$F9,$3D,$4F,$F4,$77,$2F,$D3,$7E,$77
	.dc.b	$BB,$F4,$FD,$CF,$A9,$EE,$7D,$5F,$88,$FA,$DE,$EF,$D7,$F8,$8F,$B1
	.dc.b	$EF,$7D,$1C,$7E,$B7,$2F,$D1,$7E,$C7,$27,$DD,$F8,$8F,$A3,$1F,$6F
	.dc.b	$8F,$E8,$BE,$5E,$1F,$47,$17,$B7,$8F,$DF,$3E,$C6,$F1,$7A,$A3,$84
	.dc.b	$62,$B8,$C6,$CF,$74,$EF,$8F,$E1,$92,$FB,$4E,$AB,$8D,$6E,$73,$3D
	.dc.b	$8F,$F5,$77,$1F,$C7,$7D,$1C,$7F,$1D,$FC,$B7,$17,$C7,$7F,$31,$C3
	.dc.b	$F1,$DF,$CD,$70,$7C,$77,$E0,$70,$7C,$77,$E1,$70,$7C,$77,$F4,$1C
	.dc.b	$9F,$1B,$6F,$C8,$70,$7E,$CB,$C8,$77,$D2,$5F,$EB,$CF,$DC,$17,$83
	.dc.b	$F7,$0F,$ED,$89,$7F,$B6,$6F,$07,$F0,$17,$85,$F6,$6D,$FF,$02,$E1
	.dc.b	$7B,$FF,$9F,$9F,$C7,$78,$8F,$88,$FD,$EC,$7E,$E7,$C0,$FD,$DB,$F7
	.dc.b	$DF,$E0,$1E,$07,$EE,$E3,$F7,$DD,$FF,$BC,$7F,$DC,$16,$F8,$8E,$DB
	.dc.b	$36,$FD,$2C,$BE,$23,$31,$8F,$88,$FE,$01,$87,$2F,$88,$CB,$FB,$4B
	.dc.b	$7C,$46,$38,$D1,$89,$F0,$FC,$46,$60,$D2,$46,$88,$7C,$27,$2D,$33
	.dc.b	$9F,$A5,$F1,$46,$A3,$15,$E2,$8C,$65,$EE,$6D,$70,$39,$C6,$5A,$7B
	.dc.b	$7E,$23,$D1,$EE,$FA,$7D,$2F,$57,$C3,$CA,$F9,$AF,$B6,$FD,$DE,$E7
	.dc.b	$BF,$E1,$FE,$07,$BB,$F0,$7D,$DF,$85,$F0,$FF,$0F,$83,$E2,$7C,$3F
	.dc.b	$C5,$F5,$7E,$37,$C3,$CD,$7F,$91,$71,$FB,$3F,$80,$3F,$D6,$4D,$8F
	.dc.b	$A1,$18,$DB,$F8,$FB,$63,$F9,$4B,$79,$0D,$24,$72,$21,$E0,$9E,$6E
	.dc.b	$0F,$E4,$51,$C1,$D1,$D8,$D5,$79,$5C,$D7,$81,$CD,$74,$7F,$2B,$C7
	.dc.b	$EA,$F5,$FE,$74,$70,$C3,$FD,$E4,$FA,$BD,$BF,$7F,$C4,$7C,$F8,$FC
	.dc.b	$EF,$1F,$C2,$F7,$3E,$1F,$C4,$7C,$F9,$FD,$0F,$17,$C6,$F7,$7E,$7C
	.dc.b	$75,$F8,$8F,$9C,$FE,$C3,$87,$CA,$FA,$CF,$11,$8E,$0F,$53,$82,$1F
	.dc.b	$10,$D6,$BF,$15,$C1,$AE,$1F,$88,$E0,$8E,$09,$C3,$F1,$63,$2F,$ED
	.dc.b	$B8,$7E,$34,$7E,$F3,$A7,$F0,$3E,$FF,$E0,$B8,$BE,$2C,$FF,$00,$BF
	.dc.b	$C9,$BF,$2F,$83,$E2,$BF,$C1,$39,$FF,$26,$FC,$FD,$FE,$2B,$F7,$B7
	.dc.b	$F8,$BA,$7E,$B3,$9D,$C1,$FA,$AE,$0F,$D7,$5F,$EA,$1A,$FE,$DC,$95
	.dc.b	$C7,$B3,$85,$FF,$02,$9E,$0D,$77,$A7,$EC,$DA,$7E,$DC,$B7,$DA,$FA
	.dc.b	$47,$2B,$FD,$1F,$BF,$FB,$88,$FC,$57,$C4,$7E,$DD,$FE,$5B,$DB,$6F
	.dc.b	$F5,$9F,$0F,$FA,$AF,$87,$FD,$C5,$7F,$28,$EE,$FD,$6F,$C3,$FE,$BB
	.dc.b	$E1,$FF,$71,$3F,$95,$F8,$7F,$DC,$3F,$97,$F8,$7F,$D9,$9F,$B4,$F8
	.dc.b	$79,$1F,$9B,$6B,$FA,$A6,$E7,$FB,$D6,$F0,$FE,$E9,$B3,$FB,$A6,$F0
	.dc.b	$FE,$E9,$BD,$86,$92,$7C,$07,$EA,$FD,$1F,$E0,$FE,$DF,$F0,$8F,$86
	.dc.b	$FE,$13,$F0,$DF,$CE,$7F,$80,$7F,$C9,$37,$FB,$87,$FC,$8E,$DF,$87
	.dc.b	$D7,$A3,$BF,$87,$E1,$EE,$71,$F8,$79,$7C,$7D,$DE,$9F,$0F,$D7,$DE
	.dc.b	$ED,$EF,$4A,$F2,$9F,$0C,$BC,$BF,$0D,$BC,$BF,$0D,$A7,$C3,$2F,$2F
	.dc.b	$C3,$6F,$2F,$C3,$6F,$2F,$C3,$7B,$12,$FB,$36,$FF,$94,$F2,$FF,$2C
	.dc.b	$3E,$34,$7C,$3F,$83,$F8,$DF,$07,$F1,$FE,$1F,$E4,$7C,$27,$C9,$F8
	.dc.b	$7F,$95,$F0,$BF,$2F,$E1,$FE,$1E,$3F,$E4,$9E,$97,$2F,$CF,$E5,$F8
	.dc.b	$B3,$F3,$FE,$1B,$E9,$7C,$3C,$8F,$49,$0F,$31,$F0,$4B,$C9,$F0,$5B
	.dc.b	$C9,$F0,$5B,$C9,$F0,$79,$3D,$E4,$E4,$F9,$93,$A3,$E4,$F2,$7C,$17
	.dc.b	$F9,$97,$6F,$E1,$79,$7E,$0C,$7E,$17,$93,$E0,$BF,$87,$E3,$F8,$2F
	.dc.b	$E2,$78,$BE,$0E,$5F,$DD,$38,$BE,$0C,$FE,$14,$FC,$77,$47,$E4,$78
	.dc.b	$7E,$04,$7C,$7E,$1F,$CA,$7C,$2C,$8F,$CA,$35,$F9,$8D,$C3,$B3,$77
	.dc.b	$FF,$40,$E3,$FE,$2E,$AE,$24,$38,$96,$B8,$B6,$78,$D7,$38,$FF,$49
	.dc.b	$D7,$37,$89,$2B,$FC,$23,$FC,$03,$FE,$39,$8F,$EB,$5E,$07,$F8,$96
	.dc.b	$3F,$C6,$51,$FD,$77,$C0,$FD,$87,$C2,$FE,$C7,$E1,$7F,$65,$E1,$7E
	.dc.b	$BA,$BF,$65,$E0,$FE,$B9,$FE,$D7,$87,$F8,$DD,$BF,$07,$D7,$E0,$FB
	.dc.b	$7C,$1F,$EB,$E7,$F7,$5E,$17,$EB,$DF,$BC,$4D,$BF,$04,$FE,$2D,$BD
	.dc.b	$BF,$8B,$6F,$6F,$E2,$DB,$DB,$F8,$BE,$4F,$C4,$C6,$2F,$F2,$CF,$0E
	.dc.b	$DF,$E1,$F1,$F8,$A3,$E0,$1F,$C3,$78,$7E,$03,$FC,$34,$FE,$21,$E2
	.dc.b	$FF,$0D,$7F,$89,$78,$BF,$8A,$C3,$ED,$F8,$3F,$A9,$9F,$C4,$9F,$C5
	.dc.b	$FC,$5F,$76,$5F,$8F,$DB,$CD,$3E,$DB,$F3,$91,$C1,$36,$F8,$11,$96
	.dc.b	$C6,$76,$F8,$53,$96,$C6,$76,$EF,$5C,$09,$3F,$E8,$19,$FF,$14,$ED
	.dc.b	$FC,$31,$FE,$09,$E0,$7E,$B1,$FE,$0D,$E0,$7E,$B1,$FE,$11,$E0,$7E
	.dc.b	$B1,$FE,$15,$E0,$7C,$37,$F2,$7B,$FC,$37,$F8,$77,$81,$FA,$B7,$F9
	.dc.b	$D7,$7F,$EB,$38,$3E,$B4,$7E,$07,$7F,$56,$FE,$8D,$BD,$7B,$FA,$23
	.dc.b	$04,$AF,$BB,$5F,$07,$7F,$E0,$D1,$FE,$44,$F0,$7F,$83,$3F,$D2,$7C
	.dc.b	$5F,$E4,$CF,$F4,$BF,$17,$F9,$33,$FD,$37,$C5,$FE,$4C,$FF,$4F,$DF
	.dc.b	$D9,$9B,$C2,$78,$BF,$C9,$AB,$F9,$57,$8B,$EB,$D7,$F7,$BC,$3F,$B5
	.dc.b	$C3,$87,$8B,$71,$FD,$B7,$17,$EF,$31,$FD,$B6,$1F,$BC,$7F,$7C,$7E
	.dc.b	$F2,$B8,$11,$CD,$F4,$6D,$F4,$AE,$51,$D0,$E4,$7E,$DB,$E1,$BF,$42
	.dc.b	$B0,$FE,$0B,$F0,$DF,$DC,$23,$F4,$47,$F6,$FF,$17,$F8,$0B,$FA,$23
	.dc.b	$F8,$09,$FA,$23,$F8,$0F,$C3,$7F,$0B,$F0,$FF,$80,$47,$F0,$4F,$0B
	.dc.b	$F8,$03,$FD,$3F,$BB,$F4,$79,$FD,$E8,$FB,$27,$E5,$CB,$AF,$DF,$6F
	.dc.b	$4F,$E0,$89,$D3,$F8,$18,$FC,$41,$FE,$8C,$E9,$F6,$ED,$FE,$95,$E9
	.dc.b	$F6,$4F,$F5,$FF,$0B,$F7,$AF,$F6,$1F,$1B,$F8,$E3,$FD,$8F,$C6,$FE
	.dc.b	$38,$FF,$65,$E9,$FC,$0F,$2F,$E0,$23,$F7,$A7,$F6,$9F,$0B,$D6,$FF
	.dc.b	$32,$E0,$F2,$67,$F9,$C2,$70,$7B,$E7,$CF,$C1,$ED,$6F,$07,$B1,$F7
	.dc.b	$F0,$7E,$6A,$FF,$D5,$DC,$1F,$B5,$7F,$AC,$78,$5F,$B5,$7F,$AD,$78
	.dc.b	$5F,$B5,$7F,$AE,$78,$5F,$B5,$7F,$AF,$78,$5E,$AE,$0F,$DA,$47,$F6
	.dc.b	$1F,$0B,$F6,$99,$FF,$5D,$3F,$6A,$78,$88,$79,$B3,$F6,$ED,$EB,$AF
	.dc.b	$6C,$FD,$93,$D7,$A7,$A6,$7F,$D8,$3C,$BF,$B2,$9F,$F0,$87,$85,$FB
	.dc.b	$27,$FC,$25,$E1,$7E,$C9,$FF,$0A,$78,$5F,$B2,$7E,$91,$FB,$33,$FB
	.dc.b	$B1,$FC,$30,$FF,$0B,$1E,$83,$D4,$7E,$87,$8B,$F4,$CD,$E5,$FD,$33
	.dc.b	$FA,$43,$1D,$FF,$AD,$69,$FA,$C7,$1A,$7E,$DC,$99,$FE,$B9,$B3,$CD
	.dc.b	$FE,$DF,$DF,$F9,$29,$4B,$07,$8D,$FD,$C2,$3F,$DB,$3E,$3F,$F7,$07
	.dc.b	$F6,$07,$F5,$13,$FC,$6D,$E3,$7F,$51,$7F,$C7,$1E,$37,$F7,$9F,$07
	.dc.b	$F6,$11,$FD,$CB,$C7,$FE,$DE,$FF,$51,$3F,$B7,$90,$FE,$01,$97,$93
	.dc.b	$2F,$1F,$A7,$FE,$03,$24,$FF,$7C,$6B,$C0,$F5,$67,$FE,$A5,$D7,$FC
	.dc.b	$1B,$3A,$6C,$F1,$21,$C6,$BB,$B8,$8D,$B9,$AE,$97,$E5,$6C,$78,$2F
	.dc.b	$B4,$E5,$B1,$93,$A4,$65,$A6,$4E,$D1,$C2,$FC,$71,$83,$6E,$2F,$CA
	.dc.b	$4A,$E9,$3B,$24,$65,$FC,$14,$7E,$14,$8E,$09,$3F,$94,$9F,$C4,$B6
	.dc.b	$76,$D9,$DE,$FF,$98,$8C,$11,$C1,$37,$FC,$BD,$72,$C5,$EE,$FF,$8F
	.dc.b	$FE,$1B,$FC,$81,$EF,$7F,$90,$7D,$4F,$F2,$15,$FF,$90,$FE,$07,$FC
	.dc.b	$89,$F0,$DF,$E4,$5F,$53,$FC,$8D,$EE,$FF,$91,$FE,$07,$FC,$91,$ED
	.dc.b	$FF,$92,$7E,$1B,$FC,$95,$F0,$DF,$E4,$BF,$7B,$F3,$FC,$9F,$7F,$C5
	.dc.b	$AD,$F8,$6C,$F3,$B7,$FC,$95,$27,$D3,$5F,$03,$96,$F3,$FB,$DB,$CF
	.dc.b	$EF,$9F,$DB,$10,$FD,$C9,$31,$FF,$30,$F3,$7C,$18,$FF,$2B,$F2,$FC
	.dc.b	$17,$FC,$B3,$9F,$C1,$7F,$CB,$7C,$7F,$05,$FF,$2E,$F1,$7C,$17,$FC
	.dc.b	$BF,$C3,$F0,$5F,$F3,$0F,$07,$C1,$7F,$7F,$97,$B5,$FE,$01,$BF,$C2
	.dc.b	$8C,$6C,$8F,$81,$3C,$32,$F8,$7B,$FE,$57,$9F,$F4,$37,$F9,$39,$FC
	.dc.b	$8E,$1F,$8F,$8C,$5F,$E7,$63,$15,$C0,$9D,$1C,$C9,$3F,$9E,$8C,$B6
	.dc.b	$32,$FE,$72,$FB,$91,$EC,$DB,$F0,$92,$B9,$CF,$13,$64,$8D,$57,$6B
	.dc.b	$F3,$B6,$FA,$5F,$29,$CD,$F9,$48,$75,$BE,$6E,$23,$6E,$67,$95,$74
	.dc.b	$BE,$D3,$E3,$9C,$B6,$33,$7D,$23,$94,$62,$77,$39,$BE,$E4,$BE,$D7
	.dc.b	$E3,$6B,$C8,$C1,$DC,$B3,$CE,$37,$3C,$E7,$47,$12,$39,$9E,$11,$A8
	.dc.b	$C5,$79,$89,$7E,$92,$4F,$82,$32,$D8,$CB,$DA,$FC,$64,$7F,$37,$F0
	.dc.b	$9F,$BA,$8F,$DC,$B6,$32,$7E,$6C,$A7,$F0,$37,$58,$7E,$85,$B9,$1A
	.dc.b	$49,$E5,$3B,$6C,$ED,$B3,$B7,$B1,$82,$18,$26,$3F,$B2,$9E,$51,$8B
	.dc.b	$FE,$27,$F0,$9E,$4F,$86,$F2,$FC,$2F,$9B,$E1,$BC,$FF,$0D,$E8,$F8
	.dc.b	$6F,$4F,$C3,$7A,$BE,$1B,$D7,$F0,$DE,$CF,$86,$F6,$FC,$37,$BB,$E1
	.dc.b	$BD,$FF,$0D,$F0,$3E,$1B,$E0,$FA,$7F,$0B,$D3,$9C,$7E,$8E,$5F,$37
	.dc.b	$37,$B5,$B1,$9E,$5F,$A2,$D3,$E7,$C6,$9F,$61,$88,$DB,$8E,$7F,$98
	.dc.b	$DE,$7F,$87,$5C,$49,$5C,$1B,$3F,$C7,$3E,$1B,$EC,$FA,$9F,$6B,$D0
	.dc.b	$FB,$BF,$C0,$2F,$80,$BF,$07,$8F,$81,$CF,$A9,$AF,$81,$DF,$AB,$C3
	.dc.b	$DB,$E3,$F0,$FC,$BD,$C9,$BF,$E0,$B4,$7C,$17,$6F,$82,$EF,$F0,$32
	.dc.b	$BF,$34,$43,$F3,$47,$E9,$79,$3F,$35,$CD,$FA,$82,$79,$3F,$A2,$9F
	.dc.b	$E1,$98,$BD,$FF,$6C,$7E,$43,$93,$DC,$F8,$F8,$FD,$FF,$03,$F0,$3E
	.dc.b	$03,$D9,$3E,$7F,$81,$F8,$7F,$03,$F1,$3B,$3D,$73,$FE,$23,$CD,$C5
	.dc.b	$F2,$3E,$1F,$E4,$FC,$3F,$CA,$F6,$65,$FA,$FB,$FF,$7B,$E8,$FE,$D9
	.dc.b	$B1,$9E,$4F,$6F,$1F,$AE,$7F,$A7,$E7,$EB,$AF,$D4,$F7,$FA,$9F,$EE
	.dc.b	$DC,$7F,$59,$F9,$DC,$5E,$B7,$E8,$76,$7D,$CE,$2F,$5C,$7D,$3E,$1F
	.dc.b	$5E,$5D,$B8,$7D,$53,$F5,$F8,$3D,$2F,$BF,$83,$F3,$5B,$FE,$2C,$91
	.dc.b	$F9,$52,$73,$7E,$53,$0F,$C7,$B6,$76,$D9,$DE,$3F,$98,$9C,$B6,$33
	.dc.b	$E0,$F8,$1A,$7E,$9E,$BF,$B9,$78,$3F,$C5,$3F,$80,$5F,$09,$7E,$23
	.dc.b	$F3,$7E,$26,$7E,$17,$F3,$58,$FE,$4F,$C4,$E1,$F1,$1F,$9A,$8F,$CB
	.dc.b	$78,$9F,$9A,$7F,$8C,$E4,$F8,$6E,$DF,$11,$FA,$6F,$17,$F3,$55,$F9
	.dc.b	$DF,$13,$F3,$4F,$EC,$F8,$7D,$1C,$3E,$F2,$71,$7B,$5B,$1A,$8C,$11
	.dc.b	$C1,$27,$E0,$46,$5B,$C5,$F9,$E7,$F0,$67,$EB,$F8,$3F,$08,$FE,$5F
	.dc.b	$C8,$FD,$CB,$F8,$43,$F6,$9E,$47,$EE,$5F,$DB,$79,$1F,$84,$E6,$FF
	.dc.b	$47,$78,$1F,$9E,$9F,$DD,$6D,$F0,$1F,$D1,$6D,$F0,$3A,$3F,$37,$1F
	.dc.b	$B8,$DF,$F6,$BE,$17,$E7,$F1,$FC,$EB,$67,$F7,$87,$F8,$37,$1F,$CF
	.dc.b	$3F,$E5,$1F,$07,$F3,$D8,$7F,$3F,$F0,$7D,$6F,$E9,$3C,$1F,$CF,$57
	.dc.b	$EC,$FC,$1F,$CF,$3F,$C3,$FC,$4F,$DC,$3F,$C4,$7C,$1F,$63,$FE,$BB
	.dc.b	$C9,$E2,$7E,$E2,$3F,$8B,$78,$3F,$A0,$CB,$D6,$4D,$BD,$33,$FB,$73
	.dc.b	$D3,$BF,$B9,$C2,$C3,$D7,$5F,$D5,$F6,$F5,$DF,$F9,$A3,$6F,$5B,$F3
	.dc.b	$B6,$F5,$D7,$F0,$0D,$BD,$79,$7F,$0F,$DB,$D6,$FF,$11,$DB,$D7,$A7
	.dc.b	$7D,$BD,$53,$FB,$73,$D6,$7E,$23,$A3,$F1,$4F,$E5,$4C,$10,$C1,$31
	.dc.b	$FD,$CF,$47,$E3,$E3,$1A,$7E,$DB,$A3,$F6,$1E,$0F,$EB,$E7,$F2,$07
	.dc.b	$EA,$3C,$0F,$D7,$BF,$C9,$FC,$0F,$D7,$BF,$90,$3F,$5E,$7E,$40,$FD
	.dc.b	$7F,$6F,$EB,$9F,$C7,$9F,$D2,$7C,$8F,$C7,$63,$F9,$4E,$8F,$5E,$9E
	.dc.b	$6D,$6C,$FF,$0A,$6D,$FF,$08,$3D,$25,$9E,$0F,$C1,$7F,$C1,$9A,$7E
	.dc.b	$AD,$FC,$84,$7B,$0F,$E8,$1E,$3F,$B2,$3F,$9A,$E7,$FD,$22,$7F,$A2
	.dc.b	$76,$7E,$AD,$FE,$8D,$D9,$F5,$BA,$7F,$55,$1F,$D2,$7B,$3F,$55,$87
	.dc.b	$F3,$ED,$7D,$B3,$F8,$CF,$1B,$FA,$16,$DF,$90,$8F,$D6,$46,$1D,$76
	.dc.b	$FE,$AF,$9B,$FC,$F7,$97,$E4,$2F,$FD,$6B,$D9,$F9,$0B,$FF,$4D,$F6
	.dc.b	$7E,$41,$FD,$17,$67,$E4,$2B,$F7,$FD,$9F,$90,$D7,$FA,$8F,$67,$E4
	.dc.b	$1F,$E8,$D7,$FA,$B3,$FB,$E7,$5F,$E4,$23,$FA,$BF,$6F,$EA,$70,$F4
	.dc.b	$47,$EA,$6B,$CE,$7E,$9B,$7F,$56,$DF,$A9,$8F,$E1,$1B,$7A,$5F,$F5
	.dc.b	$8E,$DE,$99,$FF,$58,$6D,$E9,$7F,$5D,$B7,$A6,$BF,$79,$B7,$A7,$3F
	.dc.b	$E3,$3B,$7A,$5F,$E3,$7D,$BF,$4F,$7F,$44,$7F,$1E,$DF,$D1,$87,$E1
	.dc.b	$63,$F1,$4F,$E1,$5A,$FE,$2B,$2F,$E7,$2E,$4E,$04,$72,$FF,$BA,$F9
	.dc.b	$FF,$4E,$FF,$4C,$F1,$3F,$AA,$BF,$A7,$3F,$C4,$DD,$FF,$CD,$97,$17
	.dc.b	$FB,$43,$C2,$78,$9F,$D5,$E7,$FC,$25,$E2,$7F,$57,$77,$99,$66,$BF
	.dc.b	$EB,$51,$FD,$73,$C4,$FE,$B4,$FE,$B4,$FE,$B4,$43,$DD,$E1,$FF,$01
	.dc.b	$CB,$F4,$F1,$8E,$FF,$D6,$E1,$EE,$9C,$93,$1F,$8E,$FF,$0A,$DB,$F5
	.dc.b	$91,$FC,$86,$FF,$59,$3F,$E1,$EE,$FF,$D6,$4F,$F5,$0E,$FF,$D6,$4F
	.dc.b	$F8,$8B,$BF,$F5,$8F,$F8,$93,$BF,$F5,$9C,$BE,$AC,$7F,$C6,$9E,$07
	.dc.b	$EA,$B4,$FE,$65,$DF,$FA,$5E,$EF,$C7,$93,$BF,$04,$BF,$F5,$DF,$7F
	.dc.b	$E4,$6F,$FD,$FD,$DF,$F9,$1E,$37,$7F,$E4,$69,$DF,$F9,$1B,$FE,$81
	.dc.b	$DF,$F9,$1C,$7F,$A7,$77,$FE,$45,$FE,$A1,$BF,$E4,$5F,$F1,$71,$FC
	.dc.b	$D8,$FE,$70,$7E,$FB,$4F,$23,$4F,$D4,$38,$E1,$FF,$65,$65,$E7,$D1
	.dc.b	$AF,$9F,$5F,$36,$BE,$5A,$69,$E5,$D7,$FA,$B6,$BE,$57,$FA,$C6,$BE
	.dc.b	$5D,$7F,$CB,$DE,$2F,$E9,$27,$FA,$E7,$7F,$E0,$35,$FC,$27,$37,$DE
	.dc.b	$25,$FE,$86,$32,$DE,$9E,$07,$1E,$9F,$D1,$C7,$E6,$0F,$CB,$F5,$BB
	.dc.b	$FF,$4B,$19,$CB,$F3,$5E,$1F,$E1,$A3,$FB,$D7,$5F,$F4,$C7,$F9,$89
	.dc.b	$FD,$30,$FE,$62,$7F,$4C,$3F,$40,$7F,$4C,$3F,$0E,$7E,$7F,$97,$F9
	.dc.b	$93,$FE,$40,$E6,$FF,$26,$11,$C1,$0F,$26,$28,$D4,$61,$B3,$FB,$96
	.dc.b	$EB,$FB,$5C,$7F,$69,$18,$AF,$82,$D8,$CC,$77,$AE,$4D,$AE,$57,$E1
	.dc.b	$8F,$F2,$BF,$BF,$FE,$63,$F8,$7F,$F3,$27,$C0,$7F,$99,$7E,$1E,$57
	.dc.b	$F9,$73,$0F,$F3,$47,$C3,$FF,$9A,$7E,$0B,$FC,$D5,$F0,$FF,$E6,$BF
	.dc.b	$83,$FF,$36,$7C,$3F,$F9,$B7,$E1,$3F,$CD,$DF,$0F,$FE,$6F,$F8,$5F
	.dc.b	$F3,$87,$C3,$FF,$0F,$D3,$EB,$7F,$00,$D6,$9E,$BE,$2F,$33,$F1,$4F
	.dc.b	$51,$F0,$0B,$C9,$F0,$1D,$13,$93,$E0,$B7,$83,$E0,$F2,$F5,$24,$72
	.dc.b	$BE,$F5,$C0,$95,$F4,$70,$7C,$3F,$C7,$BF,$F3,$FF,$27,$C6,$7E,$EF
	.dc.b	$27,$FA,$36,$FE,$3C,$7F,$A1,$38,$FE,$3B,$FE,$86,$E2,$F8,$EF,$FA
	.dc.b	$23,$87,$E4,$3F,$E8,$AE,$0F,$8D,$A3,$0F,$17,$5E,$1F,$25,$53,$F9
	.dc.b	$CE,$BF,$2E,$BD,$31,$F1,$CF,$84,$57,$3C,$1F,$25,$BB,$FC,$2B,$F8
	.dc.b	$F5,$F1,$78,$3E,$3B,$CA,$F8,$4E,$A3,$15,$C5,$ED,$19,$BF,$54,$79
	.dc.b	$B8,$3E,$53,$FA,$4E,$0F,$95,$3F,$D6,$78,$3E,$54,$FF,$AB,$39,$7F
	.dc.b	$20,$FF,$AB,$B9,$7F,$20,$FF,$AC,$39,$7E,$4B,$FA,$EE,$1F,$C8,$3F
	.dc.b	$D7,$70,$F0,$EA,$DF,$FD,$58,$7C,$7D,$7E,$5D,$6D,$FC,$07,$C3,$FC
	.dc.b	$DA,$E0,$DA,$34,$95,$C4,$FA,$6E,$1F,$CF,$B6,$FC,$73,$CA,$78,$57
	.dc.b	$66,$C6,$5E,$26,$2B,$95,$F9,$A3,$04,$36,$4B,$FF,$02,$74,$7A,$7E
	.dc.b	$1F,$D5,$F0,$FE,$BF,$87,$F6,$7C,$3F,$B7,$E1,$FD,$DF,$0F,$EF,$F8
	.dc.b	$7F,$81,$F0,$FF,$07,$E1,$FE,$17,$C3,$FC,$3F,$87,$FD,$FE,$FF,$EE
	.dc.b	$2F,$0B,$F7,$EF,$AB,$E1,$FE,$41,$F8,$AE,$1F,$C7,$F7,$5E,$3F,$C3
	.dc.b	$12,$BF,$12,$4E,$6F,$C7,$B7,$BB,$F1,$D1,$82,$38,$25,$75,$9F,$E8
	.dc.b	$FC,$BF,$90,$C3,$FC,$4D,$DB,$F9,$0C,$7F,$92,$F6,$FE,$42,$FF,$9B
	.dc.b	$F6,$FE,$41,$FE,$73,$DB,$F9,$07,$F9,$DF,$6F,$E4,$31,$FF,$18,$6F
	.dc.b	$F9,$09,$FE,$7F,$E1,$FE,$1D,$87,$EA,$C8,$7E,$A3,$5F,$36,$FE,$A6
	.dc.b	$EF,$EA,$9D,$E1,$F2,$1B,$AF,$F1,$5E,$FF,$63,$FC,$FB,$7F,$64,$FF
	.dc.b	$66,$DF,$D9,$7F,$E4,$5D,$FD,$8F,$F9,$1F,$7F,$63,$FE,$49,$DF,$D9
	.dc.b	$8F,$F4,$8E,$FF,$64,$FE,$63,$C0,$F5,$E5,$FC,$E0,$F1,$9F,$57,$6F
	.dc.b	$49,$1C,$EF,$ED,$6E,$FE,$D7,$E1,$1F,$A0,$DF,$D6,$FF,$7E,$DF,$D7
	.dc.b	$5F,$E5,$5D,$FD,$7A,$7F,$5C,$DF,$D6,$FF,$5E,$DF,$D6,$FF,$60,$DF
	.dc.b	$D7,$8F,$F8,$23,$7F,$5C,$FF,$2C,$8F,$D7,$9E,$CD,$FF,$6B,$DF,$FA
	.dc.b	$E8,$FD,$D9,$FA,$D8,$CB,$67,$F6,$4F,$EB,$DB,$1A,$8C,$1C,$8B,$9F
	.dc.b	$EF,$DB,$E0,$7E,$FE,$78,$12,$B9,$92,$7F,$36,$7F,$06,$F0,$3F,$76
	.dc.b	$FF,$08,$EF,$FD,$DB,$FC,$2B,$BB,$F7,$6F,$F0,$CE,$DF,$DD,$BF,$C3
	.dc.b	$BB,$3F,$76,$FF,$10,$EB,$FD,$DB,$FE,$18,$E9,$FD,$DB,$FD,$E3,$AF
	.dc.b	$F7,$6F,$DA,$E8,$FB,$CF,$DA,$26,$BF,$97,$6E,$BF,$BD,$DF,$F0,$51
	.dc.b	$83,$F7,$9D,$7F,$80,$8F,$F4,$97,$3F,$E0,$3A,$5D,$1F,$80,$BF,$F2
	.dc.b	$0F,$47,$E0,$1F,$F2,$1F,$47,$E0,$1F,$F2,$2F,$47,$E0,$31,$FF,$4E
	.dc.b	$6F,$F8,$09,$FF,$24,$F5,$FD,$FD,$7F,$A0,$C7,$8E,$BC,$BB,$7A,$49
	.dc.b	$3E,$9D,$3F,$31,$18,$AF,$D0,$1F,$0C,$FC,$FF,$3F,$F8,$21,$FE,$7B
	.dc.b	$2F,$13,$FC,$5F,$18,$DB,$DD,$87,$F8,$22,$3F,$C8,$FE,$27,$F8,$21
	.dc.b	$FF,$24,$F8,$9F,$E0,$87,$FA,$F1,$FE,$09,$3F,$C1,$07,$F4,$8E,$EF
	.dc.b	$63,$FE,$04,$3E,$AE,$7E,$7D,$3D,$B8,$7A,$F9,$32,$DA,$F5,$1F,$0F
	.dc.b	$4F,$E5,$FA,$FA,$E3,$FC,$B7,$AF,$A9,$FF,$10,$9E,$B3,$FC,$90,$7F
	.dc.b	$79,$3F,$C4,$47,$F7,$93,$FC,$44,$7F,$79,$3D,$47,$B0,$FF,$33,$F8
	.dc.b	$9F,$DB,$F3,$F5,$E5,$FD,$E6,$7F,$68,$FE,$A7,$3F,$D3,$47,$EB,$5B
	.dc.b	$19,$CB,$F5,$0D,$9D,$46,$31,$FD,$83,$7B,$7A,$C6,$27,$64,$AE,$44
	.dc.b	$8D,$3F,$EE,$9E,$6F,$ED,$D3,$FD,$EB,$C4,$FE,$DC,$FF,$7C,$F1,$3F
	.dc.b	$B7,$3F,$DF,$BC,$4F,$ED,$CF,$F8,$03,$C4,$FE,$DC,$FF,$80,$BC,$4F
	.dc.b	$ED,$CF,$F6,$D2,$1F,$D3,$8F,$ED,$C4,$3F,$CD,$BF,$0F,$FE,$17,$23
	.dc.b	$FE,$0C,$6B,$A3,$04,$30,$4D,$BF,$C2,$F1,$C1,$B1,$B6,$BB,$DB,$FC
	.dc.b	$D1,$3B,$6C,$6D,$AE,$CF,$F2,$CC,$6D,$A6,$CA,$6F,$4F,$F1,$D4,$6D
	.dc.b	$B1,$B6,$BB,$D5,$3B,$6C,$7F,$8B,$BE,$1F,$FC,$78,$47,$04,$3F,$D2
	.dc.b	$EE,$5A,$67,$0F,$F1,$DC,$65,$B1,$9C,$3F,$D1,$53,$96,$C6,$4F,$F3
	.dc.b	$CC,$65,$A7,$F9,$0B,$E1,$FF,$CA,$1E,$5F,$F9,$47,$D5,$FF,$29,$7A
	.dc.b	$BF,$E5,$59,$7C,$25,$F7,$31,$F0,$19,$F3,$B5,$EB,$6F,$DD,$E1,$EF
	.dc.b	$F1,$F7,$B9,$7B,$1C,$FE,$0B,$A7,$BF,$D7,$DF,$9C,$5E,$7E,$8F,$F3
	.dc.b	$57,$27,$9D,$B1,$9E,$74,$E5,$B1,$98,$C1,$1C,$10,$F9,$A4,$E7,$FF
	.dc.b	$33,$F1,$FC,$28,$FF,$32,$F2,$FC,$27,$FC,$D3,$1F,$09,$BC,$7F,$09
	.dc.b	$CF,$17,$C2,$77,$C3,$F0,$9E,$3C,$1F,$09,$E7,$BF,$C2,$7A,$ED,$F0
	.dc.b	$9F,$C6,$74,$7E,$2C,$8F,$E4,$76,$FC,$6B,$A3,$11,$A7,$04,$D7,$F5
	.dc.b	$24,$E8,$FD,$2B,$63,$F4,$2D,$E0,$FC,$EB,$6F,$BB,$6F,$B3,$6B,$C2
	.dc.b	$43,$C1,$7D,$88,$F6,$6D,$F6,$6D,$F6,$6D,$F6,$6D,$75,$BF,$03,$D8
	.dc.b	$E8,$4B,$E8,$DB,$E8,$DB,$E8,$D3,$A5,$F6,$7A,$57,$22,$57,$26,$D7
	.dc.b	$26,$CF,$23,$A1,$CA,$78,$12,$78,$36,$38,$4F,$27,$84,$E8,$8E,$A7
	.dc.b	$83,$A3,$11,$A7,$04,$BF,$8E,$4B,$F7,$36,$3C,$ED,$BF,$13,$6F,$BB
	.dc.b	$F8,$5D,$9F,$0F,$FC,$14,$97,$FA,$66,$A9,$FF,$72,$CF,$08,$D1,$0D
	.dc.b	$60,$58,$27,$84,$68,$CB,$8A,$EB,$7E,$06,$CF,$56,$BC,$6B,$A4,$EE
	.dc.b	$FC,$11,$89,$E1,$18,$BE,$CD,$F7,$FF,$92,$7C,$3F,$F2,$5F,$87,$FE
	.dc.b	$4D,$F0,$1F,$C9,$FD,$EF,$E5,$1F,$0B,$FC,$A7,$E0,$7F,$95,$7C,$0F
	.dc.b	$FB,$FF,$F8,$05,$F1,$17,$E0,$F1,$F0,$79,$F8,$7D,$7C,$46,$FE,$23
	.dc.b	$87,$A9,$C7,$CF,$9C,$BF,$EF,$CC,$FF,$C8,$0F,$C9,$F9,$5F,$E2,$DC
	.dc.b	$E9,$CC,$A7,$31,$86,$FA,$9F,$46,$75,$E5,$7C,$4A,$D3,$67,$55,$C2
	.dc.b	$BC,$13,$96,$9F,$6A,$B8,$B6,$B8,$CE,$E3,$B9,$93,$8D,$FA,$67,$71
	.dc.b	$93,$71,$97,$75,$DA,$31,$1A,$8C,$12,$79,$CE,$AF,$D5,$5B,$9C,$D6
	.dc.b	$E3,$AC,$60,$8E,$08,$77,$F7,$B8,$FB,$32,$3C,$AD,$7C,$8D,$ED,$75
	.dc.b	$FC,$F3,$C2,$49,$F6,$B5,$F6,$37,$1F,$B6,$DC,$F9,$4F,$02,$1C,$0B
	.dc.b	$3C,$1B,$1C,$1D,$12,$34,$FC,$BF,$7B,$E2,$FB,$5F,$1B,$DE,$F8,$FE
	.dc.b	$DF,$C8,$F7,$BE,$4F,$C3,$7C,$AF,$7F,$E5,$FB,$BF,$33,$DF,$FA,$3F
	.dc.b	$C0,$3F,$8F,$D3,$E4,$1F,$80,$7F,$D0,$5A,$FE,$69,$7B,$FC,$3D,$FE
	.dc.b	$3F,$0F,$CB,$E0,$26,$1F,$A3,$E3,$F8,$4F,$E4,$5B,$D3,$FA,$36,$F4
	.dc.b	$FD,$09,$F8,$26,$CB,$3B,$6C,$6D,$F8,$9C,$BF,$1F,$4F,$8A,$47,$04
	.dc.b	$8E,$87,$C9,$3F,$2F,$C9,$F2,$5F,$CC,$F1,$FC,$97,$F3,$7C,$5F,$25
	.dc.b	$FC,$EF,$0F,$C9,$7E,$31,$F9,$6E,$8F,$97,$F1,$1F,$33,$87,$E3,$4F
	.dc.b	$CB,$E0,$F8,$AF,$EA,$38,$9C,$3F,$C4,$FA,$BF,$37,$3F,$A1,$3F,$37
	.dc.b	$BF,$E9,$09,$BF,$ED,$DB,$1F,$B3,$6F,$0F,$EA,$E7,$44,$9D,$36,$35
	.dc.b	$18,$23,$82,$70,$FE,$C6,$3C,$84,$AE,$F1,$96,$DF,$A9,$CF,$57,$EB
	.dc.b	$DF,$E3,$ED,$C3,$F8,$E3,$FB,$1E,$3F,$25,$F3,$9B,$7C,$DB,$87,$C4
	.dc.b	$6B,$E3,$6E,$1E,$28,$D1,$EC,$AE,$8D,$AE,$8D,$37,$8F,$C0,$8C,$B6
	.dc.b	$3F,$84,$90,$FE,$12,$5A,$F6,$B9,$EA,$FC,$E4,$FE,$66,$3F,$2E,$7E
	.dc.b	$4A,$7F,$2C,$DC,$3F,$2D,$3B,$9C,$B6,$33,$87,$E8,$67,$2D,$8C,$E3
	.dc.b	$FA,$C2,$5F,$EA,$3A,$BF,$83,$BF,$CD,$BA,$BF,$83,$BF,$CE,$3A,$BF
	.dc.b	$83,$BF,$CE,$BA,$BF,$83,$BF,$CF,$3A,$BF,$83,$BF,$9F,$3E,$06,$3F
	.dc.b	$69,$A7,$F4,$FD,$FE,$F3,$4F,$E0,$87,$EE,$B2,$F2,$BF,$13,$5F,$73
	.dc.b	$F0,$C8,$E0,$9A,$7B,$E3,$2D,$8C,$CF,$4D,$7E,$2B,$FD,$2B,$5F,$83
	.dc.b	$7F,$D4,$F5,$F8,$2F,$F5,$5D,$7E,$0B,$FD,$5F,$5F,$82,$FF,$59,$D7
	.dc.b	$E0,$BF,$D7,$B9,$7D,$B9,$FF,$7D,$D6,$EA,$69,$FB,$DA,$FC,$6C,$7E
	.dc.b	$4A,$7F,$19,$1F,$99,$3F,$24,$57,$3A,$FE,$6E,$31,$3C,$49,$C1,$F9
	.dc.b	$66,$C7,$68,$C1,$1F,$CE,$F1,$3E,$23,$FA,$0F,$C1,$7F,$42,$F8,$8F
	.dc.b	$E8,$7F,$07,$FD,$13,$E2,$26,$1F,$E4,$3E,$6F,$CD,$E2,$F8,$4F,$E9
	.dc.b	$24,$9F,$E9,$0D,$7F,$8F,$37,$5F,$E9,$4D,$E7,$FF,$1F,$D7,$3A,$E2
	.dc.b	$43,$89,$6B,$8B,$67,$8D,$73,$8F,$F2,$BD,$73,$78,$92,$B8,$B6,$B8
	.dc.b	$BC,$CE,$35,$CF,$75,$F3,$BE,$24,$BE,$37,$CC,$E2,$73,$AE,$37,$CE
	.dc.b	$FC,$98,$7C,$AF,$88,$FE,$E2,$4B,$FE,$E0,$D7,$FA,$33,$71,$FE,$E4
	.dc.b	$DC,$7E,$4B,$67,$E0,$B7,$1F,$4B,$70,$34,$93,$FD,$FF,$E1,$7F,$C0
	.dc.b	$7F,$11,$FE,$04,$F8,$8F,$F0,$2F,$B1,$FE,$06,$F8,$0F,$F0,$3F,$AF
	.dc.b	$FE,$08,$F8,$7F,$F0,$4F,$C0,$FF,$82,$BE,$1F,$FC,$77,$FC,$03,$FE
	.dc.b	$F5,$D3,$FC,$8A,$BE,$0F,$3F,$11,$AF,$84,$DF,$C4,$70,$F8,$5E,$3F
	.dc.b	$11,$31,$FC,$87,$91,$EE,$7E,$29,$0F,$8A,$5E,$8F,$8A,$DE,$8E,$0E
	.dc.b	$89,$1A,$6C,$6A,$30,$47,$04,$8F,$9A,$75,$29,$FD,$EF,$97,$E4,$BF
	.dc.b	$E2,$9E,$4F,$92,$FF,$8A,$F8,$FE,$4B,$FE,$2D,$E2,$F9,$2F,$F8,$BF
	.dc.b	$87,$E4,$E1,$F2,$B8,$3E,$3C,$FF,$9C,$B8,$5C,$1F,$22,$31,$BF,$C7
	.dc.b	$7E,$7E,$FF,$E8,$9E,$9F,$DB,$F5,$7E,$44,$93,$F9,$5E,$9F,$CE,$67
	.dc.b	$F9,$AE,$1F,$CE,$4F,$E6,$8F,$CE,$75,$7E,$69,$E1,$7C,$9C,$DF,$27
	.dc.b	$85,$72,$AE,$11,$AA,$E1,$39,$AE,$15,$A8,$C4,$EA,$FC,$84,$BE,$AD
	.dc.b	$AE,$B8,$7B,$A7,$CA,$43,$C0,$5B,$EA,$DA,$E5,$87,$88,$98,$78,$67
	.dc.b	$B9,$30,$F3,$57,$09,$D4,$62,$75,$5C,$9E,$C6,$27,$71,$FC,$0B,$8F
	.dc.b	$94,$EA,$B9,$12,$79,$47,$A8,$98,$7A,$9B,$87,$26,$9D,$CE,$93,$CA
	.dc.b	$78,$12,$78,$36,$78,$1F,$CA,$31,$F9,$8F,$4B,$F2,$63,$EA,$24,$76
	.dc.b	$6B,$BC,$3C,$18,$76,$24,$76,$C3,$C1,$7D,$23,$91,$06,$1F,$E2,$AD
	.dc.b	$5F,$0D,$FC,$38,$93,$FA,$56,$BF,$C4,$1B,$7F,$EF,$29,$E3,$3B,$21
	.dc.b	$B2,$C6,$E7,$8E,$09,$60,$BE,$71,$C4,$97,$C5,$E7,$5C,$6B,$85,$FF
	.dc.b	$26,$F8,$8F,$E4,$FF,$0F,$FC,$A3,$E2,$3F,$94,$FC,$47,$F2,$AF,$87
	.dc.b	$FE,$57,$F0,$FF,$CB,$3E,$1F,$F9,$6F,$C3,$FF,$2E,$F8,$7F,$E5,$FF
	.dc.b	$0F,$FC,$C3,$E1,$FF,$98,$FC,$3F,$F3,$2F,$87,$FF,$76,$FA,$6F,$87
	.dc.b	$BF,$0F,$8F,$87,$9C,$DF,$CB,$8F,$6D,$3E,$1F,$89,$23,$E1,$74,$3A
	.dc.b	$3D,$95,$B9,$CC,$62,$B8,$42,$B8,$3D,$A3,$31,$8E,$85,$F3,$AE,$17
	.dc.b	$CD,$EA,$71,$2C,$F0,$BE,$73,$B9,$E2,$6C,$B3,$B7,$8D,$77,$BF,$13
	.dc.b	$6B,$94,$62,$FA,$D7,$38,$C3,$CA,$33,$7D,$63,$BD,$F3,$78,$5F,$42
	.dc.b	$38,$24,$F1,$BE,$B1,$CA,$FA,$DF,$C5,$EA,$FF,$46,$AF,$87,$FB,$E4
	.dc.b	$AF,$6B,$5F,$1B,$6F,$DE,$DC,$7E,$DB,$67,$F0,$6D,$C7,$EA,$B6,$FC
	.dc.b	$13,$D8,$87,$62,$DF,$CC,$F8,$7F,$C9,$FC,$3F,$E5,$3E,$1F,$F2,$BE
	.dc.b	$BF,$E5,$BD,$AF,$CB,$FB,$5F,$98,$F1,$3F,$33,$EF,$7E,$6B,$E1,$BF
	.dc.b	$37,$EF,$FE,$73,$E1,$BF,$3B,$F0,$1F,$9E,$F8,$6F,$CF,$F0,$FE,$83
	.dc.b	$E1,$BF,$43,$F0,$53,$8F,$F3,$DC,$BE,$A7,$CB,$CD,$F0,$1A,$7C,$02
	.dc.b	$C6,$DF,$61,$39,$7E,$4B,$AC,$3E,$1B,$9A,$E8,$D3,$A1,$6B,$A5,$72
	.dc.b	$8C,$53,$97,$E3,$C7,$E5,$B9,$3E,$3B,$F9,$8E,$3F,$8E,$FE,$6B,$8B
	.dc.b	$E3,$BF,$9C,$E1,$F8,$EF,$E7,$B8,$3E,$3B,$FA,$0D,$FE,$3B,$FA,$2D
	.dc.b	$FE,$33,$F9,$8D,$FF,$C1,$58,$7E,$5B,$6F,$CA,$6F,$76,$FC,$E4,$62
	.dc.b	$FF,$33,$18,$27,$47,$E6,$09,$1F,$A7,$21,$F9,$D2,$F0,$7E,$80,$97
	.dc.b	$DE,$78,$CE,$C9,$3A,$7C,$31,$A8,$C5,$F9,$49,$7E,$4A,$EF,$7F,$C1
	.dc.b	$FA,$3F,$57,$87,$18,$F0,$4E,$8E,$A5,$BE,$73,$B8,$C9,$E4,$9E,$2D
	.dc.b	$33,$87,$93,$0F,$14,$66,$F9,$BD,$63,$33,$BB,$ED,$5A,$AF,$85,$87
	.dc.b	$A1,$E1,$3A,$BE,$38,$78,$B0,$E5,$5A,$21,$E1,$9D,$BE,$69,$E0,$47
	.dc.b	$04,$C3,$C5,$5D,$23,$95,$F8,$4F,$C4,$BB,$BE,$B5,$C2,$FA,$36,$75
	.dc.b	$5C,$5F,$2D,$F5,$25,$70,$BE,$64,$AE,$15,$E2,$38,$CF,$DD,$C7,$FB
	.dc.b	$97,$F0,$0F,$F7,$B9,$FE,$92,$D7,$FA,$2B,$70,$FE,$A8,$DA,$E3,$3C
	.dc.b	$67,$2D,$8C,$BD,$0C,$1F,$E0,$89,$E4,$D3,$38,$A3,$2D,$8C,$C6,$08
	.dc.b	$E0,$87,$72,$1E,$A3,$8F,$C3,$F2,$F8,$7E,$7F,$11,$D3,$E2,$3A,$FC
	.dc.b	$47,$6F,$87,$EF,$F0,$FE,$0F,$87,$F0,$FC,$3F,$8B,$E1,$FC,$7E,$BF
	.dc.b	$93,$E1,$FC,$BF,$03,$E6,$F8,$7F,$3F,$C1,$7A,$3E,$1E,$71,$7F,$30
	.dc.b	$E3,$7C,$1F,$B0,$91,$F4,$9A,$FD,$26,$F3,$FD,$26,$F3,$FD,$A6,$CF
	.dc.b	$CD,$6F,$3F,$CB,$6D,$F8,$27,$B1,$0E,$C5,$BF,$BF,$F0,$FF,$43,$E1
	.dc.b	$3E,$8F,$C3,$FD,$2F,$85,$FA,$FF,$C0,$3F,$ED,$DB,$FF,$0C,$3F,$ED
	.dc.b	$DC,$3F,$C2,$6B,$E2,$37,$F0,$DC,$3E,$23,$8F,$C3,$F2,$F8,$7E,$7F
	.dc.b	$0B,$D3,$E1,$7A,$FC,$2F,$6F,$85,$EF,$F0,$B3,$5F,$EB,$3A,$BE,$17
	.dc.b	$C6,$48,$F0,$B5,$FF,$0F,$C6,$DC,$90,$CC,$6F,$7F,$5B,$79,$FF,$CB
	.dc.b	$15,$C6,$B6,$71,$AD,$B6,$77,$5C,$63,$FC,$CF,$5C,$6B,$E6,$C7,$EC
	.dc.b	$38,$FE,$73,$F2,$3E,$17,$E6,$FC,$2F,$CE,$F8,$8F,$9F,$F0,$DF,$43
	.dc.b	$E1,$BE,$55,$FC,$EE,$3F,$94,$FD,$0E,$2F,$AB,$F0,$DF,$5B,$E1,$BE
	.dc.b	$54,$FD,$4E,$1F,$94,$FF,$A3,$3C,$FF,$3B,$E9,$3C,$FC,$3E,$9E,$F7
	.dc.b	$17,$A7,$8F,$75,$F0,$89,$C7,$F0,$9B,$1A,$8C,$11,$C1,$23,$B9,$E1
	.dc.b	$AE,$F7,$E1,$BF,$04,$FF,$09,$ED,$F8,$EF,$F1,$6E,$4F,$C9,$F2,$7C
	.dc.b	$88,$FC,$8F,$1F,$C8,$7F,$27,$C5,$F2,$1F,$C1,$F1,$7C,$87,$F0,$BC
	.dc.b	$5F,$21,$FC,$CF,$0F,$C8,$7F,$13,$C3,$F2,$3D,$27,$17,$EC,$67,$F5
	.dc.b	$1C,$35,$FD,$59,$38,$BF,$70,$E8,$FD,$B1,$78,$BF,$6C,$DE,$FF,$DB
	.dc.b	$3C,$6B,$A5,$72,$8C,$1C,$8B,$5C,$9B,$5C,$A7,$F6,$BF,$0D,$FB,$A9
	.dc.b	$FE,$7F,$DF,$FB,$AC,$3F,$77,$F0,$DF,$BC,$F0,$3F,$6F,$5F,$E8,$CF
	.dc.b	$E0,$1F,$F5,$43,$7E,$1B,$F9,$BC,$3E,$1A,$FA,$DF,$B9,$AF,$E0,$3E
	.dc.b	$06,$BE,$1B,$7F,$01,$27,$F9,$17,$91,$FC,$50,$8F,$C1,$6B,$EA,$6F
	.dc.b	$07,$C1,$26,$DF,$10,$93,$F0,$E3,$2D,$8F,$EB,$FD,$BF,$88,$9F,$EB
	.dc.b	$7E,$47,$F1,$BA,$FE,$2F,$5F,$AF,$7F,$8C,$EF,$F0,$9F,$E1,$3E,$3F
	.dc.b	$F1,$87,$F8,$EF,$8F,$FC,$61,$FE,$3F,$E3,$FF,$18,$7F,$90,$F8,$FF
	.dc.b	$AE,$CB,$F9,$2F,$5F,$F8,$0B,$A7,$F3,$B5,$F9,$FF,$0F,$F3,$B1,$88
	.dc.b	$FD,$49,$FA,$42,$B9,$AF,$D5,$46,$5B,$19,$D3,$F5,$8D,$9D,$46,$2B
	.dc.b	$B9,$3C,$4F,$DF,$75,$B8,$3E,$35,$7F,$81,$BC,$4F,$D6,$E1,$FC,$F3
	.dc.b	$C7,$F8,$CF,$F2,$CF,$1F,$F8,$A3,$FD,$03,$87,$E3,$3F,$CC,$38,$7E
	.dc.b	$33,$FC,$CB,$C7,$FE,$28,$FC,$63,$F8,$A1,$0F,$E2,$87,$A0,$9B,$7A
	.dc.b	$C9,$B7,$A6,$3E,$56,$DF,$C1,$C8,$E0,$87,$CA,$8E,$0F,$CC,$DB,$E0
	.dc.b	$3F,$CC,$C8,$7F,$48,$F1,$FF,$86,$3F,$D2,$BC,$7F,$E1,$8F,$F4,$CF
	.dc.b	$1F,$F8,$67,$17,$F6,$2F,$1F,$F8,$63,$FD,$93,$C7,$FE,$19,$5F,$D4
	.dc.b	$BC,$7F,$E1,$93,$F8,$48,$F8,$1E,$3F,$F0,$C7,$F3,$04,$DB,$F7,$93
	.dc.b	$F9,$C6,$F2,$FE,$81,$B3,$BD,$7F,$80,$4E,$5B,$19,$EC,$FD,$5C,$E5
	.dc.b	$B1,$93,$FB,$CF,$8D,$FC,$1D,$FF,$21,$F8,$DF,$C1,$DF,$EF,$BE,$37
	.dc.b	$F0,$77,$FB,$FF,$91,$FC,$13,$E1,$FF,$82,$F9,$1F,$C1,$BC,$7F,$E0
	.dc.b	$B5,$FA,$63,$F8,$29,$FC,$37,$C5,$FE,$08,$B7,$6A,$86,$EF,$23,$FD
	.dc.b	$89,$1F,$ED,$A5,$F0,$FF,$C5,$09,$1F,$C6,$5A,$F5,$AE,$95,$CE,$73
	.dc.b	$7E,$0A,$EF,$7D,$1E,$44,$39,$1A,$9E,$37,$D2,$78,$4E,$AF,$B3,$C6
	.dc.b	$FB,$46,$23,$53,$C2,$72,$DA,$E9,$7D,$AB,$CB,$F0,$5F,$CC,$BE,$1F
	.dc.b	$F9,$9F,$C1,$FF,$34,$F8,$7F,$E6,$BE,$77,$F3,$6F,$87,$FE,$6F,$F0
	.dc.b	$BF,$CE,$3E,$1F,$F9,$CF,$C3,$7F,$3A,$F8,$7F,$E7,$7F,$0F,$FC,$F3
	.dc.b	$E1,$7F,$9E,$FC,$3F,$F3,$EF,$87,$FF,$3A,$F8,$3F,$E7,$41,$F0,$F8
	.dc.b	$3F,$8E,$2F,$87,$D9,$38,$FE,$D7,$3F,$C0,$9D,$46,$39,$BE,$9C,$62
	.dc.b	$3E,$43,$67,$85,$73,$9C,$B6,$33,$1C,$CF,$C3,$CF,$1A,$EB,$5B,$6C
	.dc.b	$EF,$A1,$5B,$8E,$35,$B6,$9B,$8D,$4F,$0F,$87,$FA,$53,$FD,$9F,$97
	.dc.b	$E9,$3F,$DA,$79,$3E,$4F,$C3,$FD,$28,$E5,$F0,$FF,$33,$E1,$FE,$94
	.dc.b	$7F,$6F,$E2,$F9,$FF,$0F,$F4,$A3,$FB,$97,$0F,$D3,$7C,$5F,$0F,$F5
	.dc.b	$0F,$A3,$BF,$9F,$6F,$5B,$77,$F3,$91,$C1,$2B,$D8,$D7,$E6,$98,$21
	.dc.b	$82,$70,$FD,$07,$EB,$F0,$7C,$27,$F3,$7C,$1F,$09,$FC,$EF,$07,$C2
	.dc.b	$7F,$C1,$9D,$1F,$8C,$E0,$F8,$51,$FE,$0F,$E6,$FC,$9B,$F3,$F7,$F8
	.dc.b	$4F,$E9,$B7,$F8,$4F,$E1,$B7,$F8,$4F,$F0,$3E,$DF,$D3,$70,$7E,$B7
	.dc.b	$6F,$D5,$C6,$3B,$BF,$59,$5F,$B9,$3F,$69,$CB,$FA,$E6,$99,$3F,$6B
	.dc.b	$3C,$48,$EA,$FB,$4F,$18,$D9,$FE,$29,$EE,$FD,$4F,$C3,$FE,$AB,$E1
	.dc.b	$FF,$57,$F0,$FF,$AC,$F8,$7F,$DB,$5F,$F8,$BF,$C0,$FD,$B3,$FE,$31
	.dc.b	$EF,$FD,$8F,$C3,$FE,$CB,$E1,$FF,$6D,$3F,$E3,$5F,$07,$F6,$BF,$0F
	.dc.b	$FB,$68,$FD,$A1,$F9,$7F,$07,$F6,$BF,$0F,$FB,$A2,$47,$E6,$DA,$FE
	.dc.b	$C1,$BC,$9F,$B7,$6F,$0F,$ED,$9B,$3F,$AE,$6F,$17,$EA,$9B,$DE,$69
	.dc.b	$27,$F5,$5E,$7F,$F0,$CF,$7F,$F8,$6F,$BF,$FC,$3B,$E1,$3F,$87,$FA
	.dc.b	$3F,$D0,$6F,$F9,$57,$F0,$0F,$FB,$3F,$4F,$E7,$2F,$FB,$3E,$33,$F0
	.dc.b	$BB,$DB,$F9,$6C,$BD,$2C,$5F,$0F,$80,$B8,$7F,$4C,$A7,$B3,$8E,$14
	.dc.b	$3E,$03,$0B,$DA,$A4,$BF,$E1,$50,$F8,$5D,$9F,$42,$3E,$71,$0F,$9C
	.dc.b	$5D,$3E,$73,$73,$F9,$CD,$3E,$71,$71,$F9,$CD,$E4,$F9,$CD,$E4,$F9
	.dc.b	$CF,$62,$5F,$F3,$1E,$4F,$A1,$1E,$8F,$87,$FA,$0F,$AB,$E1,$BD,$FF
	.dc.b	$0D,$F0,$3E,$03,$E0,$FC,$1F,$C2,$E4,$F8,$7F,$05,$F1,$39,$3E,$2F
	.dc.b	$C3,$7C,$3C,$BE,$BF,$25,$DB,$E9,$FC,$37,$CA,$3E,$24,$F9,$38,$FD
	.dc.b	$ED,$3D,$E5,$E3,$F7,$B7,$8F,$DE,$DE,$3F,$7E,$1F,$2A,$30,$7C,$32
	.dc.b	$1F,$D3,$38,$FE,$0B,$F8,$CE,$3F,$82,$FE,$3B,$8F,$E0,$BF,$73,$8B
	.dc.b	$E0,$BF,$7B,$87,$E0,$BF,$80,$E0,$F8,$2F,$E0,$B7,$F8,$1B,$7E,$59
	.dc.b	$F8,$11,$F8,$6D,$FF,$D0,$5C,$AF,$84,$FC,$69,$27,$F1,$6D,$7F,$42
	.dc.b	$DD,$BF,$3A,$DC,$3F,$46,$D9,$FC,$E3,$78,$7F,$36,$DB,$F0,$4F,$62
	.dc.b	$1D,$8B,$7F,$AA,$F8,$6F,$D0,$F1,$7E,$DA,$7F,$3F,$F0,$7F,$B6,$7F
	.dc.b	$43,$97,$E9,$8F,$D3,$F5,$C8,$FD,$2F,$5F,$EA,$BE,$0F,$F5,$7D,$9F
	.dc.b	$AF,$FE,$01,$FF,$66,$D7,$F8,$9D,$7C,$1E,$7E,$0F,$F5,$73,$FA,$EE
	.dc.b	$FF,$C4,$EF,$F7,$C9,$A7,$E2,$9B,$3F,$8A,$6F,$17,$E2,$9B,$BF,$E2
	.dc.b	$9B,$D7,$F8,$AB,$E4,$4B,$E4,$DB,$E4,$DB,$FF,$36,$C7,$F8,$2B,$C3
	.dc.b	$F8,$D5,$FE,$1A,$F0,$BE,$33,$FC,$6A,$FF,$58,$FF,$08,$E1,$FD,$63
	.dc.b	$F1,$8F,$C6,$1F,$C3,$37,$FC,$5B,$FC,$5F,$9F,$E2,$3F,$E1,$9E,$07
	.dc.b	$C4,$FC,$B3,$E2,$E1,$E5,$D7,$D9,$5E,$A2,$69,$EA,$6E,$BF,$0E,$B8
	.dc.b	$47,$EA,$36,$F8,$AD,$8C,$BF,$AB,$3F,$C4,$9B,$7F,$26,$ED,$FD,$5D
	.dc.b	$7F,$32,$ED,$FD,$5D,$7F,$3A,$ED,$FD,$5B,$FC,$F3,$B7,$F5,$6F,$F3
	.dc.b	$EE,$DF,$D5,$BF,$C2,$8F,$D6,$10,$FC,$37,$C2,$7E,$24,$F7,$BE,$AD
	.dc.b	$BD,$FB,$7B,$49,$B7,$B6,$74,$7A,$CA,$7F,$08,$26,$7F,$01,$F7,$1F
	.dc.b	$E5,$4D,$7D,$CF,$F9,$0F,$93,$FA,$2E,$BE,$EA,$FE,$A3,$AF,$B9,$FE
	.dc.b	$A7,$AF,$B9,$FE,$AB,$AF,$B9,$FE,$AF,$C5,$F7,$B4,$FE,$A2,$FB,$72
	.dc.b	$FB,$DB,$7F,$08,$ED,$FD,$D1,$23,$F6,$FC,$3F,$C0,$9A,$7F,$02,$2D
	.dc.b	$FF,$02,$6F,$6F,$F0,$27,$F8,$01,$31,$FE,$1A,$EA,$FB,$3F,$DC,$BB
	.dc.b	$7F,$81,$BF,$E6,$3F,$17,$FB,$2F,$77,$E7,$AF,$FB,$8E,$5F,$C0,$EB
	.dc.b	$F3,$C4,$3F,$BD,$F7,$7F,$03,$7F,$3C,$7E,$6F,$E1,$7F,$82,$77,$7E
	.dc.b	$7A,$3F,$82,$74,$FE,$75,$FF,$62,$7C,$3F,$FB,$9F,$7F,$BB,$D3,$F8
	.dc.b	$0E,$9F,$BD,$18,$AF,$BC,$D7,$EE,$E7,$F9,$AD,$3E,$F1,$1F,$DE,$78
	.dc.b	$5F,$D1,$A3,$FD,$31,$BF,$E7,$27,$FD,$2F,$E3,$FB,$6F,$F3,$87,$B6
	.dc.b	$BE,$FB,$FB,$B3,$F3,$87,$B8,$FC,$E1,$1F,$F6,$1F,$6B,$C1,$FE,$49
	.dc.b	$F1,$32,$BD,$FE,$EF,$9F,$19,$1F,$98,$C3,$D0,$D8,$CE,$9E,$F9,$CB
	.dc.b	$63,$3A,$FF,$18,$1D,$FF,$9C,$BF,$E3,$07,$F8,$0C,$FF,$0E,$F9,$3F
	.dc.b	$C6,$1F,$F1,$9F,$93,$FC,$61,$FF,$1A,$F9,$3F,$E0,$37,$FC,$47,$E4
	.dc.b	$FE,$6F,$B3,$FC,$51,$1F,$C5,$F7,$FE,$81,$EB,$48,$F6,$9E,$92,$6B
	.dc.b	$E7,$6E,$9F,$70,$9C,$BF,$7C,$91,$F7,$A3,$5C,$BF,$67,$7F,$F2,$3F
	.dc.b	$8B,$F6,$6B,$FC,$81,$E2,$7E,$DE,$7F,$C9,$7E,$27,$ED,$DF,$F2,$6F
	.dc.b	$89,$FB,$77,$FC,$9F,$E2,$7E,$DD,$FE,$F9,$7F,$C4,$F9,$7E,$CE,$FE
	.dc.b	$B8,$FE,$B5,$BF,$ED,$DF,$D5,$10,$FD,$75,$7E,$DD,$A7,$ED,$CB,$BF
	.dc.b	$ED,$DB,$C1,$FB,$7C,$3F,$6B,$18,$F0,$BF,$6B,$18,$25,$FF,$99,$BC
	.dc.b	$2F,$DB,$D7,$FB,$B3,$B3,$F6,$EF,$FB,$6F,$B3,$F3,$57,$FB,$88,$FC
	.dc.b	$D5,$FF,$9B,$BC,$7F,$F3,$0F,$85,$FB,$79,$FF,$39,$F8,$FF,$DE,$7C
	.dc.b	$3F,$DB,$C7,$F5,$3F,$5B,$F3,$0F,$F7,$BF,$03,$E1,$3E,$73,$E1,$65
	.dc.b	$E6,$7E,$03,$78,$3E,$03,$78,$3E,$06,$9F,$1E,$31,$C1,$F0,$23,$04
	.dc.b	$BE,$6D,$B7,$82,$F8,$7F,$F0,$87,$BF,$FE,$11,$F8,$7F,$F0,$97,$89
	.dc.b	$FE,$13,$F5,$7F,$C2,$9F,$01,$FE,$15,$F8,$7F,$F0,$B7,$AF,$FE,$17
	.dc.b	$F8,$0F,$F0,$C7,$C2,$7F,$86,$7D,$7F,$F0,$D7,$C3,$FF,$86,$FE,$06
	.dc.b	$71,$7F,$75,$7F,$D1,$9F,$C0,$3F,$E9,$B8,$72,$B9,$10,$E4,$56,$FC
	.dc.b	$FF,$E3,$DC,$FF,$0A,$BD,$A4,$AF,$8F,$7E,$F9,$D4,$62,$75,$18,$27
	.dc.b	$3F,$C9,$8C,$B6,$78,$12,$79,$F3,$FC,$9A,$F3,$E5,$E9,$7D,$19,$3D
	.dc.b	$9F,$99,$1F,$E1,$5E,$6F,$98,$F0,$CF,$DF,$97,$BB,$E1,$BE,$2F,$C3
	.dc.b	$7C,$6F,$88,$F8,$FF,$0F,$F3,$B0,$FF,$0F,$72,$7C,$E7,$E6,$10,$FF
	.dc.b	$24,$71,$7C,$EF,$88,$91,$FC,$33,$2F,$57,$17,$B7,$8B,$E0,$CF,$E4
	.dc.b	$B5,$F9,$13,$92,$46,$4F,$8A,$7F,$8B,$FB,$7E,$DF,$07,$BA,$3F,$CB
	.dc.b	$1C,$1E,$E7,$EE,$DF,$BB,$85,$C3,$EE,$9F,$84,$7C,$8E,$0F,$7B,$EE
	.dc.b	$21,$FE,$63,$E5,$FC,$5B,$EE,$3E,$07,$1F,$D1,$DB,$F0,$FB,$FC,$09
	.dc.b	$72,$FF,$65,$C9,$F1,$1F,$92,$24,$F8,$AF,$F1,$F1,$88,$F4,$9F,$71
	.dc.b	$C6,$1D,$EB,$9C,$F1,$6B,$C2,$B9,$B9,$AE,$3A,$7C,$36,$C7,$53,$44
	.dc.b	$39,$57,$4B,$F0,$4F,$0B,$FD,$2F,$C4,$7E,$EA,$FC,$5F,$0F,$FB,$A7
	.dc.b	$C9,$F0,$FF,$BA,$7F,$CC,$DD,$BF,$BA,$7F,$CD,$1D,$9F,$B4,$F8,$7F
	.dc.b	$DD,$47,$AF,$E1,$FF,$6F,$F0,$FF,$B8,$F8,$7F,$DC,$FC,$3F,$EE,$AA
	.dc.b	$1F,$8D,$EB,$FC,$5B,$FC,$74,$9C,$1F,$C5,$9F,$C5,$6F,$F9,$26,$CF
	.dc.b	$09,$D4,$70,$E2,$FD,$2F,$81,$FA,$F8,$FF,$52,$78,$DF,$8F,$AF,$F0
	.dc.b	$3F,$8D,$F8,$F7,$FC,$13,$E2,$FF,$17,$AF,$F4,$27,$8B,$FC,$5D,$FC
	.dc.b	$5C,$FF,$83,$BC,$5F,$D5,$47,$E2,$C9,$1F,$77,$5F,$E2,$FE,$CC,$9F
	.dc.b	$77,$85,$FA,$47,$CC,$7E,$90,$BB,$FB,$5B,$BF,$B7,$7F,$D6,$C7,$C2
	.dc.b	$DF,$D9,$97,$E2,$FC,$1F,$65,$7C,$0F,$07,$D8,$FE,$A7,$C1,$FD,$46
	.dc.b	$1F,$84,$F0,$7F,$4E,$FF,$B1,$FC,$4F,$E0,$CF,$E9,$CF,$D2,$D7,$CA
	.dc.b	$F1,$3F,$53,$97,$E6,$7D,$8F,$E6,$67,$F5,$47,$E8,$92,$7C,$F5,$EA
	.dc.b	$DF,$E0,$4E,$63,$E0,$37,$7F,$E4,$FB,$7B,$2F,$F8,$06,$DE,$CB,$FE
	.dc.b	$79,$2D,$BD,$B1,$FC,$A3,$6F,$6E,$3F,$AD,$DB,$DA,$FF,$91,$36,$F6
	.dc.b	$BF,$E4,$6F,$13,$D7,$9F,$B7,$0F,$79,$FC,$CB,$6F,$6B,$FC,$87,$C2
	.dc.b	$FD,$54,$FE,$C6,$3F,$5D,$7F,$AC,$6C,$65,$FD,$8F,$85,$FA,$FA,$E3
	.dc.b	$1F,$B7,$32,$70,$31,$5C,$DC,$91,$C1,$2B,$B3,$63,$FA,$4F,$85,$FB
	.dc.b	$B7,$F8,$0F,$85,$FB,$B7,$F8,$1F,$85,$FB,$B7,$F8,$77,$03,$C2,$FD
	.dc.b	$E4,$63,$C2,$FD,$E3,$FE,$E5,$EA,$FE,$E3,$E0,$FE,$F2,$3F,$8C,$78
	.dc.b	$1F,$BC,$7F,$8D,$75,$FE,$7D,$FE,$53,$DF,$F8,$2D,$BE,$F7,$67,$E6
	.dc.b	$E7,$F0,$3C,$19,$23,$82,$75,$7E,$8E,$BF,$0D,$B7,$E1,$67,$F0,$7E
	.dc.b	$07,$E1,$6B,$FC,$2B,$8F,$E1,$1F,$F0,$BF,$81,$FB,$BA,$F7,$F8,$1F
	.dc.b	$BB,$7F,$40,$7B,$4F,$E7,$58,$3A,$3F,$0B,$1F,$90,$EF,$F7,$3F,$CB
	.dc.b	$36,$F7,$78,$7F,$E0,$CF,$03,$F3,$BB,$FA,$B4,$F4,$37,$83,$E0,$46
	.dc.b	$30,$F6,$C6,$0C,$E7,$ED,$9F,$F0,$CF,$81,$ED,$8F,$D0,$57,$9B,$C8
	.dc.b	$FF,$11,$F8,$1F,$A0,$C7,$FB,$37,$81,$FA,$07,$F9,$0F,$81,$FA,$07
	.dc.b	$F9,$1F,$07,$B3,$3F,$F1,$DF,$05,$EE,$78,$1F,$A0,$D3,$EC,$9E,$6E
	.dc.b	$0F,$6B,$4F,$E6,$2E,$2F,$D8,$4E,$0F,$F0,$FF,$8D,$FD,$FA,$BF,$C3
	.dc.b	$44,$3F,$23,$BF,$A9,$FE,$C5,$E4,$7F,$86,$5F,$EC,$9E,$47,$A5,$FE
	.dc.b	$03,$BF,$F3,$0E,$AF,$F5,$FE,$DE,$BE,$33,$E9,$8F,$E2,$D5,$FB,$9F
	.dc.b	$0F,$F6,$44,$7F,$61,$AF,$F6,$E7,$F5,$DC,$5F,$AC,$23,$FB,$46,$C6
	.dc.b	$EB,$F6,$0D,$8C,$F0,$FE,$FA,$31,$5C,$48,$FF,$70,$7F,$1F,$D7,$FE
	.dc.b	$5C,$F0,$BF,$6D,$87,$F8,$17,$C2,$FD,$B3,$FE,$DC,$F0,$BF,$6A,$FF
	.dc.b	$B7,$BC,$2F,$DA,$E1,$FB,$AF,$0B,$F6,$AF,$F8,$2F,$C6,$FE,$E0,$FF
	.dc.b	$83,$7C,$6F,$DA,$5F,$F8,$47,$BF,$F9,$C6,$FF,$95,$E9,$FB,$5D,$3F
	.dc.b	$93,$69,$30,$FE,$A1,$D3,$FC,$37,$0F,$F7,$7F,$8D,$E8,$C3,$FB,$7F
	.dc.b	$57,$A1,$FE,$E3,$BF,$A1,$FE,$E7,$BF,$DB,$E3,$FF,$45,$F0,$7D,$B9
	.dc.b	$FE,$65,$9F,$F6,$A7,$FC,$81,$BF,$FA,$03,$DE,$93,$FD,$A8,$FC,$89
	.dc.b	$0F,$E9,$C5,$D3,$F8,$7E,$DE,$98,$FE,$3F,$B7,$E4,$DF,$F2,$79,$F9
	.dc.b	$39,$FF,$21,$74,$36,$FC,$A4,$FF,$8A,$BA,$FF,$28,$FF,$8B,$3A,$FD
	.dc.b	$54,$EB,$F5,$4F,$F0,$73,$FA,$5E,$FF,$E9,$2D,$7F,$D3,$3E,$8C,$9F
	.dc.b	$51,$E8,$26,$7E,$87,$D5,$BF,$D5,$69,$FE,$97,$C7,$F8,$37,$57,$FA
	.dc.b	$1E,$3D,$53,$FE,$01,$7F,$83,$1F,$E4,$DF,$03,$F8,$33,$FE,$67,$3F
	.dc.b	$5A,$7F,$87,$23,$F8,$3B,$FC,$8B,$BF,$F5,$9D,$DF,$C8,$31,$FF,$2C
	.dc.b	$65,$E9,$AF,$E0,$BA,$7E,$78,$99,$7E,$84,$9D,$DF,$AF,$9F,$D6,$9F
	.dc.b	$A2,$8C,$BF,$B2,$6B,$A9,$E7,$3C,$6F,$F5,$F4,$F2,$3F,$25,$9F,$FA
	.dc.b	$17,$BB,$F2,$58,$FF,$AA,$BB,$BF,$24,$FF,$82,$3A,$BF,$24,$FF,$82
	.dc.b	$BA,$BF,$53,$97,$F8,$EE,$3F,$CB,$07,$EA,$4F,$EB,$A7,$F9,$D7,$E1
	.dc.b	$3F,$22,$FF,$9B,$3A,$BF,$52,$FF,$23,$3E,$CE,$5F,$5D,$A6,$6B,$EE
	.dc.b	$3F,$E7,$DE,$DF,$E4,$AF,$F0,$23,$EC,$9F,$E1,$EF,$1F,$FA,$D7,$2B
	.dc.b	$C7,$FE,$B5,$7F,$96,$F1,$FF,$AD,$3F,$E1,$DB,$FE,$B4,$7F,$8E,$4F
	.dc.b	$D5,$1F,$93,$3F,$D0,$7D,$1E,$A8,$FF,$49,$67,$EA,$E4,$FE,$71,$DD
	.dc.b	$FA,$79,$F3,$93,$5F,$63,$70,$FF,$19,$F7,$7E,$AE,$3F,$81,$C7,$AA
	.dc.b	$DA,$7E,$AE,$3F,$24,$7F,$3E,$3F,$D4,$1E,$2F,$F5,$77,$F9,$E9,$F9
	.dc.b	$33,$FC,$7F,$D5,$F9,$27,$F8,$D1,$FE,$A7,$E9,$FC,$97,$8F,$FD,$C6
	.dc.b	$3F,$4B,$87,$A9,$BC,$5F,$C6,$DB,$AF,$F5,$7C,$BF,$53,$E3,$FF,$6D
	.dc.b	$F0,$FF,$C1,$1D,$7F,$EC,$EB,$F3,$1F,$E6,$73,$F4,$E7,$F8,$20,$9B
	.dc.b	$7F,$92,$DF,$DF,$DF,$F5,$27,$CE,$7F,$AC,$B4,$F3,$F2,$FF,$46,$EC
	.dc.b	$FC,$E7,$15,$AF,$D1,$46,$1F,$D3,$37,$6F,$D1,$C6,$88,$7F,$AF,$BB
	.dc.b	$3F,$46,$A1,$9B,$B3,$F4,$B1,$8E,$CF,$D2,$BF,$5B,$B3,$F4,$AF,$D8
	.dc.b	$EC,$FD,$2E,$7F,$EA,$AF,$0F,$FA,$6E,$5F,$EC,$CF,$17,$FA,$6B,$FE
	.dc.b	$CE,$EE,$FC,$65,$7F,$BD,$FA,$FF,$19,$CF,$F4,$6B,$F4,$79,$7D,$56
	.dc.b	$C6,$79,$3F,$19,$3F,$E8,$2E,$BF,$5B,$F8,$AE,$6F,$4F,$37,$E2,$F0
	.dc.b	$FF,$2C,$F7,$7E,$93,$3F,$F5,$FF,$77,$E9,$1F,$E3,$73,$FD,$94,$FE
	.dc.b	$38,$7D,$6E,$F3,$FB,$D3,$FD,$E5,$E9,$61,$FF,$74,$BE,$AE,$CF,$D2
	.dc.b	$57,$E8,$89,$3F,$BD,$6E,$7F,$E2,$2C,$BF,$1B,$3F,$A3,$7F,$C7,$2F
	.dc.b	$F8,$80,$FD,$19,$FD,$34,$FD,$19,$F8,$ED,$7F,$78,$FE,$3A,$BF,$C7
	.dc.b	$2F,$A4,$FC,$6C,$23,$D7,$4E,$BC,$57,$BA,$AC,$E6,$BD,$93,$FA,$42
	.dc.b	$1F,$53,$0F,$7B,$FE,$72,$C3,$D6,$FF,$68,$25,$FF,$68,$F1,$7F,$B3
	.dc.b	$BD,$BA,$5E,$2F,$F6,$88,$C7,$8B,$FD,$A3,$C5,$FE,$CF,$A7,$F9,$4B
	.dc.b	$1F,$EC,$F8,$3C,$4F,$A5,$87,$A3,$6F,$5D,$23,$1B,$7D,$7F,$09,$E0
	.dc.b	$FE,$AE,$BF,$A5,$63,$FD,$32,$FF,$A4,$F8,$3F,$AB,$8C,$61,$B9,$CD
	.dc.b	$7E,$B1,$AF,$2F,$07,$F2,$75,$FA,$B8,$FF,$43,$E1,$E6,$BF,$A3,$2F
	.dc.b	$03,$F5,$BC,$5F,$EA,$DB,$FF,$46,$F3,$7E,$55,$FD,$B5,$3C,$BE,$51
	.dc.b	$FE,$DE,$EF,$FD,$7E,$9F,$2F,$BF,$F5,$EF,$CD,$EF,$DF,$1F,$EF,$BA
	.dc.b	$6E,$47,$E2,$6F,$F8,$4E,$37,$B1,$C1,$BA,$7E,$17,$87,$53,$92,$46
	.dc.b	$4F,$E1,$93,$C1,$FA,$53,$C1,$D6,$3F,$91,$AE,$31,$A3,$8B,$8C,$FE
	.dc.b	$E3,$6B,$94,$EE,$B9,$3D,$2F,$C1,$7C,$23,$17,$C5,$EA,$6F,$D3,$F7
	.dc.b	$7B,$12,$39,$7A,$9F,$07,$D8,$F8,$5E,$AF,$C3,$F8,$4F,$89,$F0,$BF
	.dc.b	$17,$DC,$98,$78,$7D,$CF,$91,$F0,$3F,$27,$E0,$64,$F9,$7E,$07,$E6
	.dc.b	$7C,$0C,$8F,$E1,$BE,$0F,$E5,$1F,$29,$ED,$E3,$F5,$F2,$FC,$0E,$4F
	.dc.b	$64,$E7,$93,$D9,$CB,$F0,$A3,$07,$C0,$21,$82,$4F,$22,$57,$52,$47
	.dc.b	$8A,$32,$D8,$CC,$FF,$09,$F8,$7F,$91,$3E,$7E,$5F,$C4,$72,$FE,$27
	.dc.b	$C8,$FC,$57,$BF,$F2,$6B,$EB,$7C,$17,$C8,$7D,$DC,$9F,$8F,$E4,$F9
	.dc.b	$31,$F6,$F8,$FE,$4B,$F0,$F8,$FF,$29,$F0,$B2,$3E,$4F,$5F,$93,$8F
	.dc.b	$E3,$C7,$9C,$9C,$1F,$17,$8B,$D6,$DE,$2E,$1C,$7F,$2A,$F8,$CE,$6F
	.dc.b	$8C,$60,$93,$B2,$47,$E4,$78,$FE,$1E,$1F,$C4,$F8,$FE,$1C,$FE,$57
	.dc.b	$8B,$E1,$BF,$97,$E8,$FD,$77,$17,$C3,$BF,$E3,$5C,$5F,$0E,$BF,$39
	.dc.b	$CF,$F0,$A7,$F8,$F7,$17,$ED,$78,$BE,$15,$7E,$1F,$8B,$F7,$6F,$EC
	.dc.b	$BE,$1F,$F7,$44,$7F,$4C,$D7,$F1,$CD,$BF,$E5,$4D,$E2,$FE,$08,$D9
	.dc.b	$E8,$72,$21,$C8,$B3,$C9,$B3,$C9,$A7,$F0,$6E,$CF,$E2,$BF,$B4,$BE
	.dc.b	$0E,$FC,$3E,$3E,$1B,$F6,$57,$FB,$DE,$FD,$FC,$3F,$EC,$A3,$F8,$5E
	.dc.b	$0F,$88,$E7,$F0,$DD,$3E,$23,$AF,$C4,$76,$F6,$7B,$FC,$27,$83,$DA
	.dc.b	$FD,$96,$5F,$AF,$F1,$BE,$4F,$83,$F8,$12,$78,$3F,$80,$9F,$D5,$F2
	.dc.b	$FE,$2A,$31,$1F,$AC,$E5,$FC,$5B,$A6,$B9,$3F,$17,$E0,$FE,$21,$C5
	.dc.b	$FF,$73,$E6,$FC,$4D,$7F,$33,$F0,$3F,$13,$5F,$C0,$FC,$8F,$E2,$6F
	.dc.b	$E2,$4F,$E4,$1E,$47,$E2,$6B,$F9,$D7,$7F,$E2,$67,$F8,$57,$7F,$E2
	.dc.b	$67,$F9,$F7,$7F,$C1,$9F,$93,$E2,$7D,$23,$E0,$ED,$EC,$DF,$DD,$BF
	.dc.b	$9D,$B5,$ED,$3D,$65,$DF,$E3,$CE,$F1,$F8,$35,$C2,$BF,$8E,$78,$9E
	.dc.b	$E7,$FA,$3E,$FE,$E9,$FE,$F5,$E2,$7E,$9A,$7F,$91,$78,$9F,$A6,$7F
	.dc.b	$9F,$6F,$EF,$7F,$A7,$F8,$BF,$C2,$1F,$E5,$1B,$FB,$DF,$EA,$7E,$27
	.dc.b	$E3,$BD,$89,$1F,$C2,$0F,$49,$F4,$B8,$3E,$06,$DE,$82,$46,$B4,$F5
	.dc.b	$12,$33,$C1,$F0,$27,$E1,$53,$7F,$74,$FF,$8B,$F7,$F7,$5F,$F6,$FD
	.dc.b	$FD,$D8,$FF,$86,$77,$F7,$65,$FC,$90,$F7,$9F,$D3,$4F,$E0,$67,$F2
	.dc.b	$52,$1F,$C9,$4F,$D1,$E7,$EE,$7F,$DE,$39,$BC,$4F,$DD,$67,$FE,$41
	.dc.b	$6C,$66,$7F,$6B,$39,$6C,$65,$FD,$F1,$30,$FD,$F3,$72,$FD,$F3,$4F
	.dc.b	$E0,$E5,$AE,$8D,$AE,$95,$FE,$39,$E6,$FE,$07,$E2,$7F,$02,$AF,$DE
	.dc.b	$F8,$7F,$C0,$9F,$EC,$9E,$1F,$F0,$27,$F4,$27,$F0,$2F,$07,$F8,$0B
	.dc.b	$FC,$13,$C0,$FE,$02,$FF,$06,$EF,$FE,$02,$FF,$08,$EE,$FD,$03,$FA
	.dc.b	$23,$F0,$9A,$FD,$B7,$F0,$9D,$3F,$9B,$7F,$05,$BF,$E0,$37,$FC,$3C
	.dc.b	$60,$95,$C7,$7F,$C3,$47,$F1,$F3,$FC,$2D,$C9,$F8,$4A,$79,$1E,$DA
	.dc.b	$FF,$2D,$6D,$FD,$9E,$FF,$C0,$5D,$BF,$9F,$C7,$FC,$F1,$D7,$FB,$DB
	.dc.b	$FF,$30,$76,$FE,$7E,$7F,$CF,$9D,$7F,$CB,$3D,$AF,$E5,$BE,$F4,$CB
	.dc.b	$DB,$AF,$A2,$FC,$DB,$FA,$F8,$3D,$D1,$8A,$F5,$D6,$C9,$3A,$7F,$3D
	.dc.b	$1E,$C2,$1F,$E1,$9D,$FF,$C0,$4F,$F8,$D7,$CA,$F6,$6B,$FE,$77,$F2
	.dc.b	$BD,$96,$DF,$D9,$9F,$F9,$4F,$7F,$64,$B7,$FE,$31,$97,$F9,$07,$C7
	.dc.b	$7A,$5F,$D2,$8F,$E3,$3B,$79,$49,$3F,$B4,$6B,$F7,$0C,$10,$C1,$FC
	.dc.b	$58,$F5,$E9,$FD,$ED,$FF,$22,$9F,$DE,$CF,$61,$FC,$CC,$FF,$1C,$69
	.dc.b	$EC,$D3,$D7,$1F,$E6,$4C,$FD,$79,$7A,$AF,$FD,$5F,$97,$AA,$BF,$AD
	.dc.b	$E5,$FC,$CB,$B3,$F6,$A4,$8F,$D9,$B5,$FD,$53,$6B,$F6,$64,$D7,$F6
	.dc.b	$6D,$9F,$D9,$B7,$83,$F6,$6F,$EC,$63,$FB,$5F,$5F,$EC,$5F,$9D,$D7
	.dc.b	$F9,$5B,$FF,$43,$73,$7E,$C6,$7F,$2A,$7E,$C4,$FE,$E7,$E4,$7F,$71
	.dc.b	$7F,$BA,$F9,$1F,$DC,$5F,$EE,$FE,$47,$E5,$5F,$EF,$3C,$FF,$95,$5E
	.dc.b	$67,$F9,$8F,$57,$03,$DE,$FF,$03,$92,$BF,$C0,$2D,$7F,$08,$DD,$FE
	.dc.b	$6B,$75,$FF,$70,$D7,$3A,$E2,$43,$89,$6B,$8B,$67,$8D,$73,$85,$F9
	.dc.b	$A3,$0D,$8E,$06,$AF,$C0,$F6,$25,$F6,$6D,$F6,$78,$D7,$37,$27,$3B
	.dc.b	$EF,$7D,$2F,$91,$2F,$95,$7C,$CF,$23,$E0,$FB,$3F,$E3,$02,$47,$F8
	.dc.b	$A9,$AF,$EB,$5B,$5F,$96,$6D,$72,$9E,$04,$38,$16,$78,$36,$38,$3A
	.dc.b	$24,$69,$FD,$E7,$A7,$FE,$44,$F8,$1F,$F2,$2F,$BB,$FE,$DA,$FE,$01
	.dc.b	$55,$F8,$2C,$7C,$26,$7E,$03,$5F,$09,$BF,$81,$E1,$F1,$3C,$7E,$07
	.dc.b	$97,$C0,$F3,$F8,$5E,$9F,$05,$D7,$E0,$7B,$7C,$4C,$DB,$FA,$1F,$3B
	.dc.b	$9B,$E0,$C7,$DB,$6F,$4F,$E5,$9B,$3B,$9F,$80,$6C,$B3,$B6,$C6,$DF
	.dc.b	$87,$CF,$F1,$B9,$FE,$21,$1C,$12,$3A,$1F,$20,$FF,$31,$F3,$7C,$87
	.dc.b	$FD,$DF,$CB,$F2,$1B,$CB,$F2,$1C,$F2,$7C,$87,$7C,$7F,$21,$E3,$C5
	.dc.b	$F2,$1E,$7C,$3F,$21,$EB,$C1,$F1,$DF,$F7,$A6,$2E,$0F,$F3,$5E,$DF
	.dc.b	$95,$D7,$F2,$7B,$7E,$56,$FF,$34,$7E,$7A,$74,$E0,$FC,$FE,$FF,$A1
	.dc.b	$9E,$0D,$AE,$64,$9E,$13,$AA,$E7,$5B,$9C,$D6,$E3,$04,$9D,$93,$A7
	.dc.b	$F5,$A4,$AF,$33,$5F,$2C,$6D,$C9,$F8,$88,$D5,$F8,$DB,$7E,$1A,$E1
	.dc.b	$5C,$A7,$2F,$98,$97,$E5,$8C,$B6,$3B,$9C,$4B,$7D,$DB,$7D,$EB,$8D
	.dc.b	$FF,$12,$EA,$FE,$5C,$F8,$AF,$85,$F3,$23,$82,$5F,$38,$EA,$79,$A7
	.dc.b	$94,$66,$38,$B6,$78,$C6,$EF,$C6,$47,$D7,$7E,$48,$C4,$72,$38,$1E
	.dc.b	$0C,$3E,$13,$96,$D7,$66,$DF,$84,$93,$C8,$E1,$87,$A1,$C5,$71,$8C
	.dc.b	$57,$F3,$1C,$3F,$AA,$64,$F8,$2F,$E2,$84,$C3,$F5,$ED,$7F,$46,$DC
	.dc.b	$18,$FF,$1A,$C1,$7C,$EF,$89,$0E,$25,$AE,$2F,$F5,$08,$E0,$E8,$91
	.dc.b	$A9,$EB,$7E,$1B,$F0,$5F,$0A,$D3,$63,$33,$FC,$BF,$E2,$7F,$98,$7C
	.dc.b	$2F,$F3,$1F,$88,$FE,$65,$F1,$3F,$CC,$FE,$27,$F9,$A7,$9F,$FC,$D7
	.dc.b	$E1,$FF,$9B,$7C,$37,$F0,$FD,$7C,$5F,$C0,$3E,$8E,$04,$63,$D2,$D6
	.dc.b	$FF,$0E,$9E,$E6,$2F,$FB,$6C,$3D,$8C,$71,$A3,$FC,$D3,$3F,$D1,$31
	.dc.b	$EF,$F0,$7A,$97,$9B,$7D,$F9,$7F,$E7,$E9,$7B,$5B,$24,$73,$D3,$C9
	.dc.b	$18,$8F,$4C,$61,$D1,$E6,$8D,$63,$F0,$1B,$3B,$9E,$26,$CE,$39,$7A
	.dc.b	$E7,$71,$93,$71,$97,$77,$E8,$6C,$F0,$9D,$47,$09,$D3,$67,$51,$8B
	.dc.b	$F3,$4F,$22,$1B,$2B,$A7,$04,$70,$4B,$F7,$36,$3D,$5D,$1F,$E3,$BA
	.dc.b	$FF,$3F,$47,$AB,$D6,$FA,$C4,$70,$43,$D4,$E5,$A6,$6B,$F1,$B1,$96
	.dc.b	$C6,$6F,$F1,$93,$96,$C6,$4F,$C6,$C6,$5F,$C7,$FC,$1F,$E2,$7D,$7F
	.dc.b	$C5,$7A,$FF,$8B,$F8,$3F,$C6,$7C,$27,$E3,$7E,$13,$F1,$DF,$09,$F8
	.dc.b	$FF,$77,$F2,$1E,$EF,$E4,$7E,$1F,$F2,$5E,$EF,$E6,$3C,$E7,$BB,$7E
	.dc.b	$1F,$1E,$F6,$7E,$1F,$5F,$0D,$38,$FF,$2D,$C9,$E5,$EA,$FD,$03,$7A
	.dc.b	$3E,$5C,$FC,$42,$74,$7C,$26,$F4,$7C,$06,$9F,$00,$F8,$57,$D1,$B5
	.dc.b	$D2,$B9,$12,$B9,$36,$7E,$39,$F9,$4E,$7F,$8E,$FE,$5B,$9B,$E3,$BF
	.dc.b	$98,$E5,$F8,$EF,$E6,$B9,$3E,$3B,$F9,$CE,$3F,$8E,$FE,$93,$8B,$E3
	.dc.b	$B7,$87,$E3,$B9,$E0,$F8,$CF,$E9,$A5,$C1,$FE,$1F,$E9,$FC,$87,$57
	.dc.b	$E8,$09,$5F,$9C,$DF,$F2,$E4,$7F,$61,$BF,$E6,$78,$7F,$56,$DE,$1F
	.dc.b	$D4,$56,$E7,$2D,$8C,$D7,$28,$C5,$F8,$23,$F5,$11,$87,$C4,$6C,$95
	.dc.b	$D1,$B5,$D2,$30,$47,$04,$AF,$E2,$D8,$7C,$0A,$EB,$5B,$9F,$16,$1E
	.dc.b	$B6,$E1,$E3,$AF,$29,$E0,$8C,$B4,$E7,$87,$76,$C6,$AF,$B4,$60,$87
	.dc.b	$47,$2D,$9F,$E2,$B8,$79,$E3,$15,$CE,$7B,$46,$0E,$78,$7A,$5B,$87
	.dc.b	$A6,$31,$1B,$23,$82,$46,$C9,$1E,$D3,$6F,$C3,$BF,$24,$F0,$9D,$37
	.dc.b	$0F,$5B,$6F,$CE,$DC,$3D,$44,$AF,$39,$CB,$0F,$0D,$F2,$6D,$F2,$6D
	.dc.b	$7E,$6F,$DC,$FF,$04,$66,$F8,$0F,$E1,$84,$C7,$F5,$0D,$7F,$48,$DC
	.dc.b	$BF,$C0,$93,$C6,$76,$43,$65,$8D,$CF,$1C,$3C,$2D,$9F,$F7,$0D,$73
	.dc.b	$78,$92,$B8,$BF,$BB,$F8,$6F,$E4,$5F,$03,$FC,$8F,$E1,$BF,$92,$7C
	.dc.b	$17,$F2,$5F,$86,$FE,$4D,$A7,$F2,$7F,$87,$FE,$51,$F0,$FF,$CA,$7E
	.dc.b	$17,$F9,$57,$C3,$7F,$2B,$F5,$FF,$96,$7B,$FF,$CB,$7E,$1B,$F9,$77
	.dc.b	$0F,$F2,$FF,$86,$FF,$03,$FF,$00,$FF,$A8,$78,$BF,$B5,$B7,$A5,$B7
	.dc.b	$F8,$CE,$1F,$05,$C0,$95,$FD,$DA,$32,$D8,$CB,$D1,$BC,$BF,$E3,$19
	.dc.b	$DC,$E4,$E0,$43,$04,$8E,$13,$CB,$9F,$C8,$D9,$75,$2F,$AD,$F0,$9E
	.dc.b	$8D,$7A,$5F,$27,$2D,$72,$74,$9E,$71,$F0,$FE,$1F,$E4,$FC,$1C,$AF
	.dc.b	$84,$D7,$E5,$B6,$FE,$03,$9C,$FE,$6C,$F5,$F8,$7F,$A6,$47,$E6,$35
	.dc.b	$FA,$AD,$AF,$94,$DC,$3E,$B3,$67,$A1,$C8,$87,$22,$CF,$26,$CF,$26
	.dc.b	$9F,$07,$E1,$3F,$0F,$F0,$FF,$88,$F8,$5F,$C4,$FC,$3F,$E2,$BE,$1B
	.dc.b	$F1,$7F,$0F,$F8,$CF,$87,$FC,$6F,$C3,$FE,$3B,$83,$F1,$FF,$0B,$F9
	.dc.b	$0F,$85,$FC,$8F,$C2,$FE,$4B,$E1,$7F,$27,$F0,$BF,$94,$F8,$5F,$CA
	.dc.b	$FC,$2C,$E2,$0E,$6F,$93,$D3,$F7,$F9,$FE,$03,$67,$D2,$D7,$E0,$13
	.dc.b	$9F,$E0,$37,$9F,$E0,$34,$F9,$25,$9E,$53,$C0,$93,$C1,$B3,$C1,$B3
	.dc.b	$F9,$3E,$7F,$93,$3F,$A7,$E6,$F9,$2F,$EA,$79,$7E,$4B,$FA,$BE,$4F
	.dc.b	$92,$FE,$B7,$8F,$E4,$BF,$AF,$E2,$F9,$2F,$EC,$78,$7E,$4B,$FB,$3E
	.dc.b	$0F,$90,$FE,$F3,$83,$F9,$33,$F9,$BE,$AF,$D4,$6B,$FA,$76,$F5,$7E
	.dc.b	$7A,$74,$7E,$90,$A7,$E9,$27,$83,$96,$B9,$8E,$51,$8D,$FF,$60,$4E
	.dc.b	$0F,$D7,$36,$BA,$B6,$78,$4E,$89,$3A,$6C,$6A,$30,$47,$F8,$17,$57
	.dc.b	$EC,$67,$F8,$D3,$67,$F7,$8D,$9D,$C7,$EE,$8F,$C7,$F1,$FE,$3B,$0F
	.dc.b	$59,$31,$F8,$E4,$9F,$85,$19,$6C,$65,$F8,$8D,$C7,$DD,$3B,$9C,$9C
	.dc.b	$08,$60,$FE,$16,$79,$4A,$48,$FE,$1A,$E5,$AE,$AB,$D5,$C7,$FA,$6A
	.dc.b	$34,$93,$FA,$88,$CB,$63,$2F,$E8,$E3,$2D,$8C,$D7,$E8,$25,$C7,$FA
	.dc.b	$0A,$FE,$9B,$C7,$FA,$07,$FA,$7F,$1F,$E8,$1F,$EA,$3C,$7F,$A0,$7F
	.dc.b	$A9,$F1,$FD,$5C,$1B,$52,$4F,$BF,$7F,$BA,$FD,$8E,$BF,$DA,$44,$8F
	.dc.b	$DA,$37,$83,$F7,$6D,$BF,$5E,$BE,$9D,$BD,$73,$96,$DF,$22,$5F,$26
	.dc.b	$DF,$2A,$FE,$D1,$B7,$C1,$C3,$FB,$06,$DF,$05,$FE,$C5,$B7,$C1,$7F
	.dc.b	$B2,$6D,$F0,$5F,$EC,$DB,$7B,$B2,$FB,$D8,$FB,$9A,$7E,$68,$F8,$45
	.dc.b	$3F,$93,$6B,$FA,$6C,$FF,$1D,$7F,$90,$E0,$FC,$95,$E5,$B7,$FA,$9A
	.dc.b	$DC,$E6,$3F,$27,$5C,$5B,$19,$7F,$2D,$19,$7F,$81,$7C,$2F,$F8,$3B
	.dc.b	$E1,$7F,$9C,$C7,$FB,$E3,$F8,$05,$78,$BB,$3F,$DE,$A5,$F0,$D9,$AF
	.dc.b	$E8,$3C,$29,$7C,$36,$6F,$87,$C3,$5A,$D7,$C4,$DB,$3F,$C8,$A3,$97
	.dc.b	$C3,$E4,$97,$FD,$3B,$3F,$1C,$EA,$31,$97,$8B,$05,$F0,$9F,$63,$6F
	.dc.b	$8C,$EA,$78,$9E,$D9,$D5,$73,$6C,$64,$DD,$F8,$5E,$A6,$2B,$A3,$82
	.dc.b	$5F,$82,$B9,$D7,$18,$C4,$F1,$9C,$D7,$29,$CC,$F2,$BF,$23,$B2,$19
	.dc.b	$3C,$C5,$BE,$D1,$C2,$3C,$9D,$1F,$DD,$AD,$F0,$FF,$4C,$93,$EB,$6B
	.dc.b	$F8,$56,$E1,$F2,$9B,$7F,$45,$B3,$F8,$76,$E1,$F8,$56,$DF,$82,$7B
	.dc.b	$10,$EC,$5B,$FA,$FF,$0F,$F8,$7F,$57,$F1,$1E,$FF,$E2,$7E,$1F,$F1
	.dc.b	$5F,$01,$F8,$BF,$87,$FC,$67,$C0,$FE,$37,$E1,$FF,$1D,$F0,$5F,$8F
	.dc.b	$F8,$7F,$C8,$7C,$1F,$E4,$7E,$1F,$F2,$5F,$09,$F9,$3F,$87,$FC,$B7
	.dc.b	$F0,$0F,$FB,$EB,$89,$C5,$F9,$08,$7C,$36,$8F,$CC,$72,$7A,$EB,$D2
	.dc.b	$4A,$F8,$CF,$AF,$9F,$E3,$34,$F8,$C5,$E7,$F8,$CD,$AE,$95,$CA,$3A
	.dc.b	$57,$26,$9C,$8B,$5C,$9B,$4E,$6F,$95,$1F,$97,$E5,$F9,$4F,$E6,$79
	.dc.b	$3E,$53,$F9,$BE,$3F,$94,$FE,$77,$8B,$E5,$3F,$9F,$E1,$F9,$4F,$E8
	.dc.b	$78,$3E,$53,$FA,$6D,$FE,$56,$DF,$27,$7F,$91,$3F,$D0,$CF,$CB,$C7
	.dc.b	$E7,$89,$BF,$E5,$DB,$1A,$8C,$11,$C1,$37,$FC,$FB,$77,$FD,$54,$EE
	.dc.b	$72,$48,$CC,$EF,$A7,$F6,$24,$7F,$5B,$7E,$2A,$E1,$39,$9F,$D7,$57
	.dc.b	$29,$CC,$70,$BF,$04,$E4,$FE,$0B,$D3,$FB,$1B,$E5,$87,$59,$D4,$62
	.dc.b	$B8,$B6,$FC,$C7,$52,$9E,$27,$58,$78,$A3,$06,$DC,$5F,$79,$DD,$F8
	.dc.b	$2B,$4D,$7F,$AB,$9D,$A7,$83,$A9,$E2,$47,$04,$3C,$24,$8E,$2D,$BF
	.dc.b	$6B,$B8,$C9,$23,$31,$B3,$F8,$EE,$1F,$25,$E1,$1A,$9E,$11,$E7,$8C
	.dc.b	$11,$C1,$30,$F8,$71,$96,$C6,$70,$F8,$55,$D2,$72,$48,$CD,$79,$AF
	.dc.b	$FD,$43,$93,$E1,$FF,$85,$93,$1F,$D2,$35,$FE,$1B,$5F,$DA,$A7,$84
	.dc.b	$68,$86,$8F,$F7,$5B,$A7,$18,$23,$18,$FF,$17,$BE,$CD,$AE,$6D,$78
	.dc.b	$CF,$0B,$F1,$57,$06,$CF,$07,$CD,$F0,$FF,$C9,$BE,$1F,$F9,$3F,$C3
	.dc.b	$7F,$28,$F8,$6F,$E5,$3F,$0D,$FC,$AB,$E1,$BF,$95,$FC,$37,$F2,$CF
	.dc.b	$86,$FE,$5B,$E5,$FF,$2E,$F8,$6F,$E5,$FF,$0D,$FC,$C3,$E1,$BF,$98
	.dc.b	$FC,$37,$F3,$2F,$86,$FE,$67,$F0,$DF,$CD,$3E,$1A,$71,$FF,$09,$E7
	.dc.b	$F8,$1F,$0D,$FC,$E0,$91,$FC,$45,$AF,$B9,$BD,$1F,$CE,$DB,$CD,$F4
	.dc.b	$5B,$3F,$39,$B5,$DA,$3A,$10,$E8,$5A,$E8,$DA,$E8,$D3,$A1,$67,$A3
	.dc.b	$C8,$93,$C9,$B3,$C9,$A7,$22,$CF,$29,$E0,$49,$E1,$D3,$F9,$4F,$86
	.dc.b	$FE,$BA,$4B,$FE,$6C,$D7,$F0,$6D,$AF,$EC,$0D,$AF,$CB,$36,$7F,$26
	.dc.b	$DB,$FC,$93,$7A,$8D,$24,$FF,$8A,$BB,$BF,$CA,$2F,$E6,$9F,$F6,$A8
	.dc.b	$F5,$7F,$C9,$0F,$FA,$3A,$38,$7B,$58,$F6,$B3,$F0,$DA,$F6,$F7,$F0
	.dc.b	$DC,$3D,$CE,$3F,$0D,$CB,$DD,$E7,$F0,$DD,$3E,$1B,$AF,$C2,$76,$F8
	.dc.b	$69,$B7,$E3,$78,$BD,$2F,$C6,$21,$F1,$8B,$CB,$F1,$9B,$CB,$F1,$9A
	.dc.b	$7C,$62,$F2,$FC,$66,$D7,$67,$A1,$2B,$A3,$6B,$A3,$45,$38,$FE,$5D
	.dc.b	$7F,$99,$9F,$97,$C9,$F2,$BE,$0B,$E5,$F2,$FC,$9C,$3E,$97,$27,$C9
	.dc.b	$7E,$A7,$1F,$C9,$7E,$B7,$17,$C9,$7E,$C7,$0F,$C9,$7E,$D7,$07,$C8
	.dc.b	$7E,$D7,$07,$F9,$EF,$81,$D1,$FA,$A6,$CF,$E8,$5A,$FE,$69,$BB,$FE
	.dc.b	$66,$3F,$3C,$DE,$0F,$CF,$70,$FE,$8E,$31,$5C,$23,$04,$AE,$2D,$AE
	.dc.b	$31,$82,$38,$24,$F3,$8C,$B6,$33,$7E,$B2,$4F,$AD,$AF,$EE,$FA,$3F
	.dc.b	$A1,$C6,$30,$F3,$B7,$0F,$26,$1D,$49,$19,$BF,$0B,$70,$F0,$C7,$72
	.dc.b	$1F,$26,$35,$3C,$A3,$37,$E0,$6D,$F8,$2B,$EB,$1C,$09,$1A,$3C,$25
	.dc.b	$BF,$0B,$DC,$C5,$F7,$70,$4B,$F0,$36,$FC,$0F,$0B,$EE,$FF,$83,$AF
	.dc.b	$AD,$70,$25,$70,$9E,$47,$90,$B3,$C2,$35,$3C,$27,$4D,$9D,$46,$27
	.dc.b	$64,$70,$48,$FE,$6B,$83,$C0,$FF,$6F,$F1,$BE,$17,$F8,$A9,$2F,$F8
	.dc.b	$E3,$5F,$E3,$8D,$C7,$F6,$8D,$C7,$9B,$67,$F7,$75,$D5,$E6,$43,$99
	.dc.b	$6B,$9B,$6B,$9B,$4E,$7F,$0D,$FC,$AB,$E1,$BF,$95,$FC,$37,$F2,$CF
	.dc.b	$84,$FE,$5B,$F0,$9F,$EF,$5F,$E0,$1F,$F7,$FE,$1F,$E0,$C7,$FD,$FE
	.dc.b	$FF,$B8,$1F,$F7,$EB,$7E,$0F,$87,$C1,$F1,$F5,$39,$7C,$1F,$3F,$73
	.dc.b	$A7,$C1,$F5,$F7,$7B,$7C,$1C,$CB,$FB,$D6,$6F,$7B,$C4,$48,$F0,$35
	.dc.b	$F2,$37,$8D,$C6,$9F,$E4,$75,$CD,$E2,$43,$89,$67,$8B,$C2,$BA,$57
	.dc.b	$3E,$85,$F3,$AE,$24,$BE,$33,$AB,$E3,$3C,$CF,$91,$F0,$7F,$34,$F9
	.dc.b	$DE,$FF,$CF,$F8,$39,$3F,$37,$E0,$3E,$97,$C1,$FD,$3F,$81,$FA,$9F
	.dc.b	$07,$F5,$7B,$3E,$B7,$C1,$FB,$F0,$FF,$20,$FF,$00,$7F,$AB,$B8,$7D
	.dc.b	$0F,$FB,$5A,$3E,$27,$07,$9A,$3F,$C7,$D1,$FD,$21,$7C,$0F,$39,$D7
	.dc.b	$C0,$DB,$E6,$F7,$F0,$FF,$45,$87,$B4,$F8,$53,$F1,$F9,$3E,$35,$FC
	.dc.b	$46,$C6,$78,$FE,$2F,$27,$C7,$BF,$A5,$39,$69,$D6,$F9,$57,$0B,$E7
	.dc.b	$39,$AE,$0F,$88,$C1,$2F,$B3,$F9,$BE,$4F,$A4,$FE,$6F,$93,$DC,$FC
	.dc.b	$3F,$81,$B4,$D9,$F0,$3F,$A5,$E5,$CC,$3E,$07,$EB,$61,$FA,$5E,$4F
	.dc.b	$AC,$FE,$9F,$8F,$7F,$03,$C3,$E1,$38,$FC,$17,$2F,$82,$E7,$CD,$30
	.dc.b	$FD,$1F,$27,$BC,$8F,$C4,$DF,$D7,$CB,$F1,$23,$1C,$BF,$12,$73,$CB
	.dc.b	$F1,$2B,$55,$F1,$67,$2F,$C5,$AE,$04,$BF,$9D,$CB,$F1,$A7,$F8,$17
	.dc.b	$2F,$C0,$E5,$F8,$3C,$BF,$0E,$7F,$1B,$9F,$E4,$38,$FE,$1C,$7F,$09
	.dc.b	$E3,$F8,$6F,$F0,$BE,$3F,$86,$F0,$3F,$25,$C5,$F1,$1C,$6D,$F4,$78
	.dc.b	$7E,$24,$7E,$7B,$87,$F2,$5E,$57,$E7,$8F,$A6,$48,$E3,$CB,$F9,$67
	.dc.b	$F4,$B1,$96,$9A,$71,$C7,$FA,$56,$EB,$D6,$74,$43,$89,$27,$83,$96
	.dc.b	$B9,$8F,$D9,$FC,$17,$E2,$3E,$0B,$F1,$3F,$05,$FC,$02,$BF,$97,$77
	.dc.b	$FE,$33,$E0,$BF,$80,$C7,$AB,$E0,$BF,$1F,$F0,$5F,$90,$F8,$2F,$C8
	.dc.b	$FA,$5F,$92,$F8,$AF,$C9,$FC,$1F,$E5,$3E,$13,$F2,$BE,$24,$CB,$E1
	.dc.b	$BF,$21,$FC,$07,$7F,$E3,$1F,$C2,$78,$1F,$81,$7F,$17,$19,$E3,$FC
	.dc.b	$49,$38,$FF,$18,$49,$E6,$70,$3F,$27,$E0,$7E,$26,$7F,$81,$78,$1E
	.dc.b	$F7,$F8,$27,$17,$E2,$6B,$FA,$77,$8F,$F8,$89,$FE,$11,$DD,$F8,$87
	.dc.b	$F8,$57,$77,$BE,$BE,$EF,$77,$BD,$F6,$9F,$88,$FE,$01,$FF,$30,$E9
	.dc.b	$FC,$E5,$FF,$46,$F8,$DF,$C6,$3C,$0F,$D0,$93,$83,$E0,$5F,$C2,$E0
	.dc.b	$F8,$18,$7B,$6B,$51,$89,$F7,$9B,$3E,$01,$F2,$B8,$3E,$05,$7F,$23
	.dc.b	$E0,$FD,$44,$FE,$53,$C2,$F8,$1B,$7F,$33,$F0,$BE,$05,$FF,$29,$E0
	.dc.b	$F8,$0F,$F2,$BE,$0F,$D3,$E1,$FC,$FB,$83,$F4,$CF,$F4,$0E,$3F,$BC
	.dc.b	$D3,$F8,$49,$E4,$C7,$EC,$3E,$86,$ED,$EC,$8D,$61,$EF,$DF,$E1,$D6
	.dc.b	$9B,$19,$DF,$F8,$2D,$7F,$5C,$3F,$95,$1F,$B0,$F0,$FF,$82,$BF,$D6
	.dc.b	$CF,$D9,$EB,$EE,$7F,$BA,$EB,$EE,$7F,$BB,$EB,$EE,$CB,$FA,$B6,$BE
	.dc.b	$E7,$FA,$C6,$BE,$E6,$D7,$FC,$01,$F0,$7F,$7D,$1F,$B0,$26,$9F,$B9
	.dc.b	$9F,$D9,$1F,$D7,$FC,$1F,$DE,$B4,$FD,$C4,$6E,$78,$C6,$4E,$3B,$FF
	.dc.b	$43,$AE,$2D,$8E,$26,$E7,$8C,$66,$3F,$BA,$F8,$3F,$C0,$5F,$E2,$BE
	.dc.b	$07,$E8,$F0,$FE,$D1,$D1,$FC,$06,$7F,$C5,$DD,$1F,$D3,$BB,$FF,$80
	.dc.b	$C7,$F1,$CE,$EF,$E0,$2F,$F1,$EE,$DF,$E0,$2F,$F2,$0E,$CF,$EB,$3D
	.dc.b	$8F,$8A,$B3,$0F,$C2,$9F,$99,$26,$FF,$7F,$A3,$F3,$DC,$1F,$85,$9C
	.dc.b	$ED,$F8,$17,$F0,$E7,$E2,$DF,$C4,$76,$7E,$82,$3F,$C6,$3D,$9E,$D7
	.dc.b	$FC,$67,$C1,$FA,$0A,$FF,$03,$70,$7E,$0F,$0F,$F1,$6F,$47,$E0,$DF
	.dc.b	$F1,$7F,$47,$EF,$F0,$FD,$01,$ED,$3F,$7E,$34,$FF,$18,$ED,$FC,$02
	.dc.b	$3F,$DC,$6F,$F0,$0E,$5F,$44,$7A,$8F,$41,$2F,$DD,$5F,$06,$3D,$B1
	.dc.b	$8C,$B8,$CE,$67,$FC,$14,$7E,$84,$FE,$CA,$7E,$84,$FD,$F1,$FA,$13
	.dc.b	$FC,$6B,$D3,$EF,$EB,$67,$EF,$59,$FE,$87,$0F,$F1,$CF,$57,$E8,$5F
	.dc.b	$F1,$DF,$57,$E8,$5B,$8F,$DA,$CB,$C8,$48,$F3,$92,$3D,$D8,$7E,$E9
	.dc.b	$BC,$5F,$7A,$31,$A7,$F9,$7B,$8B,$ED,$3F,$E6,$0E,$2F,$E3,$37,$FE
	.dc.b	$4E,$EC,$FB,$4F,$F9,$27,$7F,$66,$FE,$BD,$FD,$53,$FD,$F0,$FE,$E7
	.dc.b	$E2,$FA,$5F,$F2,$86,$FF,$C6,$3B,$BF,$9A,$76,$FE,$EB,$83,$F6,$44
	.dc.b	$AF,$D6,$37,$5F,$D6,$70,$FE,$BA,$B8,$56,$89,$3A,$AE,$13,$A8,$C4
	.dc.b	$EA,$30,$4E,$3F,$DE,$B8,$ED,$FD,$C4,$FF,$9B,$BB,$7F,$39,$7F,$E8
	.dc.b	$4E,$8F,$DC,$4F,$F5,$A3,$FB,$C1,$FE,$12,$F0,$7F,$BC,$3F,$E1,$4F
	.dc.b	$07,$F3,$9A,$7F,$9C,$FC,$1F,$EB,$53,$F9,$BD,$7F,$AC,$76,$7E,$DE
	.dc.b	$30,$7E,$70,$FE,$F5,$93,$AD,$F0,$3F,$C0,$09,$3F,$D8,$DA,$F2,$9E
	.dc.b	$11,$8C,$7F,$52,$DC,$3F,$B4,$36,$BF,$56,$DA,$E3,$1D,$48,$75,$BF
	.dc.b	$1B,$8B,$EA,$DB,$FE,$E9,$F0,$3F,$C5,$8F,$E2,$FF,$03,$FC,$63,$E1
	.dc.b	$A5,$7F,$7B,$F8,$2F,$E3,$7F,$0D,$FC,$73,$E0,$FF,$8E,$FC,$37,$F1
	.dc.b	$EF,$84,$FE,$3F,$F0,$D3,$0F,$F0,$17,$C2,$FF,$22,$F8,$6F,$E4,$7F
	.dc.b	$0D,$FC,$93,$E1,$A5,$7F,$49,$ED,$FB,$E4,$7F,$9D,$F2,$7E,$18,$9D
	.dc.b	$BF,$85,$8C,$B5,$FE,$8A,$DE,$4F,$CA,$57,$39,$F8,$9C,$7E,$F7,$FC
	.dc.b	$2D,$F0,$BF,$CD,$BE,$17,$F9,$BF,$C2,$FF,$38,$F8,$5F,$7D,$7F,$86
	.dc.b	$FE,$17,$F9,$DF,$C2,$FB,$E3,$FC,$CB,$1F,$EF,$6C,$FF,$DA,$2F,$FB
	.dc.b	$DA,$3E,$97,$0F,$C1,$7F,$C4,$9C,$1F,$01,$FA,$DC,$1F,$8A,$DF,$D1
	.dc.b	$D5,$FA,$6E,$0F,$46,$5E,$AB,$F7,$C6,$27,$51,$82,$1F,$0A,$3D,$CF
	.dc.b	$C0,$3D,$E7,$F8,$D7,$6F,$7B,$FE,$37,$D7,$DE,$FE,$27,$4F,$7B,$FE
	.dc.b	$3B,$E4,$FF,$59,$C7,$BF,$0F,$F7,$A3,$EF,$C7,$FA,$0E,$BE,$FB,$FC
	.dc.b	$96,$BE,$F9,$FE,$8B,$AF,$EA,$3A,$FF,$2B,$5F,$97,$EC,$FC,$E4,$62
	.dc.b	$BF,$2F,$39,$3F,$30,$7E,$88,$BC,$9F,$A2,$74,$48,$E2,$4C,$FF,$57
	.dc.b	$1C,$2B,$B4,$62,$38,$36,$7E,$DF,$C4,$7C,$0F,$86,$FD,$2D,$7F,$97
	.dc.b	$7B,$7F,$4A,$FE,$BF,$B7,$FB,$F9,$FA,$57,$FC,$C5,$D7,$F2,$3E,$1B
	.dc.b	$F4,$91,$FB,$5E,$CF,$97,$F0,$DF,$33,$E1,$BE,$6F,$C3,$7C,$EF,$86
	.dc.b	$96,$1E,$5F,$F9,$A7,$9D,$F0,$DF,$50,$93,$F9,$46,$BF,$C3,$E3,$6E
	.dc.b	$48,$66,$37,$C3,$F9,$A6,$F1,$7D,$E6,$CB,$91,$7D,$23,$91,$D2,$F9
	.dc.b	$36,$F9,$3F,$A8,$F8,$6F,$C5,$FA,$FF,$8C,$F5,$BF,$1B,$EB,$7E,$3B
	.dc.b	$CE,$97,$FA,$EF,$84,$FC,$8F,$B5,$F9,$2F,$86,$FC,$9F,$AF,$F9,$4F
	.dc.b	$86,$FC,$AF,$B1,$30,$FF,$6C,$FF,$00,$3D,$FF,$95,$1F,$CC,$61,$EF
	.dc.b	$63,$89,$C5,$F3,$F9,$14,$F7,$F5,$39,$8C,$61,$F5,$79,$FF,$D3,$90
	.dc.b	$F6,$F7,$3F,$1E,$30,$F9,$23,$38,$76,$6C,$EA,$7C,$D1,$82,$3C,$48
	.dc.b	$68,$B8,$FA,$9F,$04,$70,$74,$48,$D5,$78,$63,$8E,$1E,$DA,$E1,$5A
	.dc.b	$B7,$C2,$F9,$BE,$0F,$EC,$61,$F3,$BE,$03,$EB,$BF,$8F,$E5,$F5,$FC
	.dc.b	$07,$D8,$8F,$A7,$F0,$FF,$61,$FE,$53,$C7,$F0,$3E,$07,$E0,$FC,$3F
	.dc.b	$C2,$E4,$F8,$7C,$9F,$13,$92,$61,$F3,$FD,$3F,$33,$E8,$3D,$E5,$E4
	.dc.b	$F6,$3E,$FE,$4D,$B4,$D9,$63,$7C,$7E,$FE,$4F,$83,$5C,$23,$15,$D7
	.dc.b	$93,$EB,$46,$23,$94,$62,$7F,$0D,$C9,$F1,$E7,$FA,$3F,$67,$DE,$E4
	.dc.b	$F8,$F8,$7F,$48,$E4,$F8,$F5,$FD,$33,$AB,$F0,$7C,$9F,$1E,$3F,$75
	.dc.b	$C9,$F1,$DF,$CE,$F1,$FC,$77,$F3,$FC,$5F,$19,$FC,$F7,$17,$F8,$5B
	.dc.b	$BF,$F6,$D9,$FE,$D2,$7F,$6E,$47,$04,$DF,$F6,$F1,$96,$C6,$63,$F8
	.dc.b	$14,$F4,$9C,$92,$33,$E0,$7F,$08,$E3,$FE,$0E,$DA,$E1,$3A,$3F,$83
	.dc.b	$1D,$3C,$0F,$E0,$AF,$EE,$B9,$7F,$82,$BF,$BC,$F8,$2F,$D3,$FC,$3F
	.dc.b	$F5,$4F,$03,$F5,$3F,$07,$FC,$16,$BF,$80,$FC,$3F,$EB,$3E,$1F,$F8
	.dc.b	$24,$7F,$23,$F0,$7F,$5F,$F0,$BF,$B0,$F8,$7F,$D8,$FC,$34,$AF,$C7
	.dc.b	$1E,$53,$E1,$17,$8F,$E1,$71,$7B,$23,$D2,$D7,$6D,$8D,$F0,$FC,$08
	.dc.b	$C7,$1F,$C7,$6D,$75,$8C,$4E,$C9,$C7,$F2,$78,$FE,$2D,$7F,$3D,$EC
	.dc.b	$F8,$AF,$EC,$4F,$E4,$3F,$0D,$FC,$3B,$F8,$07,$FD,$69,$91,$FF,$6A
	.dc.b	$4F,$F4,$DC,$FF,$81,$C7,$F6,$48,$FE,$7C,$BE,$0B,$97,$C3,$73,$F5
	.dc.b	$7A,$7A,$BD,$7E,$1B,$B7,$C3,$77,$F5,$E6,$5F,$2B,$4F,$D6,$F1,$7E
	.dc.b	$AC,$91,$FB,$2D,$7F,$54,$48,$CC,$F0,$9D,$46,$3C,$7F,$D7,$46,$09
	.dc.b	$3B,$24,$69,$BE,$47,$EF,$27,$73,$FC,$71,$FE,$8F,$E6,$7F,$1C,$7F
	.dc.b	$A3,$79,$1F,$B6,$C3,$FA,$07,$7F,$CB,$F7,$FE,$67,$C3,$FC,$DF,$77
	.dc.b	$E7,$72,$7C,$6C,$7E,$66,$1F,$D2,$F9,$5C,$9F,$22,$72,$7C,$EF,$67
	.dc.b	$EB,$7C,$07,$C8,$8F,$F4,$35,$38,$FF,$D1,$FC,$1E,$B2,$70,$FA,$F8
	.dc.b	$BD,$B7,$FC,$56,$31,$3A,$8C,$10,$FE,$2F,$1A,$37,$E7,$7F,$17,$8F
	.dc.b	$C6,$1F,$CF,$4F,$F2,$FF,$9D,$FB,$07,$D3,$7F,$8B,$E1,$F8,$71,$8C
	.dc.b	$BF,$15,$3F,$8F,$DF,$E2,$4F,$B3,$6F,$84,$FB,$B6,$F8,$2F,$1D,$7E
	.dc.b	$17,$A5,$FE,$1F,$EE,$FE,$10,$49,$FD,$7F,$7F,$E9,$1B,$19,$7F,$63
	.dc.b	$19,$6C,$67,$A3,$F6,$7D,$FF,$AA,$7F,$AF,$79,$3F,$AA,$7F,$B0,$79
	.dc.b	$3F,$DD,$3C,$0F,$D5,$61,$FA,$BF,$87,$FD,$67,$C0,$FE,$B7,$E0,$7F
	.dc.b	$5D,$F0,$FF,$AF,$F8,$4F,$D8,$7C,$3F,$E9,$70,$F8,$9E,$1F,$EC,$FE
	.dc.b	$16,$47,$E9,$78,$7C,$DB,$7F,$06,$E1,$F3,$77,$FF,$02,$6D,$7F,$05
	.dc.b	$27,$47,$F0,$82,$5F,$2A,$E1,$3F,$10,$FF,$18,$1F,$E9,$8F,$77,$F3
	.dc.b	$D8,$7F,$A6,$3D,$DF,$CF,$4F,$F0,$0B,$FE,$11,$F0,$FF,$E1,$6F,$E0
	.dc.b	$1F,$F4,$04,$FF,$90,$5F,$F4,$06,$3F,$E2,$75,$F0,$DB,$F8,$69,$5F
	.dc.b	$86,$97,$C3,$73,$3A,$7C,$37,$5F,$86,$95,$FC,$1F,$C7,$FC,$C3,$F9
	.dc.b	$63,$F3,$1B,$7F,$05,$7F,$73,$E3,$FE,$65,$CE,$9F,$C1,$63,$11,$C4
	.dc.b	$FE,$EB,$E3,$FE,$76,$BF,$D6,$FE,$8F,$F7,$39,$FC,$E9,$F9,$9E,$0F
	.dc.b	$81,$F0,$DF,$07,$E1,$BE,$17,$C6,$7C,$3E,$29,$7E,$DE,$7F,$63,$EF
	.dc.b	$E1,$F9,$1E,$1F,$C9,$E7,$93,$ED,$DF,$CB,$B7,$AF,$A3,$F7,$33,$E7
	.dc.b	$8C,$71,$7C,$26,$D7,$0E,$2F,$75,$7C,$03,$F8,$DF,$17,$B6,$3F,$D6
	.dc.b	$BC,$5E,$D9,$FE,$88,$7F,$0C,$E2,$F6,$BF,$57,$87,$DA,$FD,$7E,$0F
	.dc.b	$6A,$CB,$FC,$8F,$CB,$FD,$64,$FB,$71,$EF,$3E,$F6,$FE,$E7,$86,$5F
	.dc.b	$85,$DF,$FC,$25,$C3,$F9,$AE,$0F,$CB,$B7,$4F,$CC,$5F,$E7,$1B,$3A
	.dc.b	$8F,$CD,$1F,$BA,$2E,$5F,$EF,$0C,$7F,$75,$8F,$F9,$D7,$C3,$FD,$D3
	.dc.b	$FE,$77,$F0,$FF,$35,$BF,$E7,$30,$FC,$A4,$FE,$77,$AB,$F3,$DE,$5F
	.dc.b	$E6,$27,$D2,$7E,$7F,$AB,$19,$FC,$2F,$73,$F3,$33,$F9,$DE,$AF,$CC
	.dc.b	$BF,$EB,$6D,$3F,$49,$F0,$5F,$A9,$24,$7E,$8D,$AF,$6D,$DC,$69,$CE
	.dc.b	$3F,$41,$BA,$F4,$BE,$44,$39,$57,$17,$15,$C5,$B3,$C5,$DD,$74,$7F
	.dc.b	$78,$FE,$FF,$E1,$FF,$80,$7C,$3F,$F0,$1F,$83,$FE,$05,$EE,$FF,$A6
	.dc.b	$3C,$AF,$F5,$66,$1F,$D0,$9F,$F5,$5B,$7E,$07,$5D,$BB,$ED,$E1,$EC
	.dc.b	$71,$F6,$39,$7C,$3F,$3F,$85,$E9,$F0,$FD,$7E,$1B,$B7,$34,$D3,$FC
	.dc.b	$15,$CB,$ED,$7D,$27,$B0,$F4,$97,$9B,$E5,$F2,$7B,$5B,$19,$9E,$5C
	.dc.b	$DF,$16,$39,$91,$C1,$39,$BE,$43,$63,$B9,$B8,$C9,$FE,$C9,$E2,$76
	.dc.b	$FC,$7F,$73,$E5,$57,$C7,$F8,$5F,$95,$F0,$FF,$2F,$E1,$E4,$FC,$BE
	.dc.b	$6F,$92,$FC,$DF,$87,$F9,$0F,$CF,$F8,$7F,$A5,$F0,$FF,$22,$3E,$97
	.dc.b	$C3,$FC,$77,$F1,$9A,$B8,$FF,$DC,$3E,$D5,$9F,$4F,$0F,$AF,$8F,$E2
	.dc.b	$C7,$AC,$CF,$17,$B1,$A4,$8D,$CF,$1E,$3F,$97,$5B,$9C,$B6,$33,$5C
	.dc.b	$5B,$3A,$8C,$72,$7E,$37,$93,$E4,$47,$F4,$9E,$EF,$C8,$72,$7C,$88
	.dc.b	$FC,$87,$1F,$C8,$7D,$FC,$7F,$21,$F8,$3C,$7F,$21,$F8,$7C,$7F,$21
	.dc.b	$FC,$C7,$3F,$E6,$F8,$FE,$47,$07,$FB,$93,$8F,$E3,$57,$EE,$C6,$E8
	.dc.b	$7C,$3F,$E8,$C9,$3F,$A3,$6B,$F4,$9B,$5D,$5B,$C5,$F8,$76,$CF,$6A
	.dc.b	$EA,$F3,$21,$CC,$B5,$CD,$B5,$CD,$A7,$F7,$7F,$07,$F7,$6F,$ED,$EB
	.dc.b	$F7,$6F,$F7,$5F,$00,$FF,$8F,$3F,$80,$7F,$D0,$F2,$F0,$73,$3F,$C5
	.dc.b	$E7,$FD,$12,$FE,$3B,$C1,$FD,$EB,$C7,$E1,$3F,$7A,$FF,$05,$EF,$FD
	.dc.b	$F3,$FC,$1F,$BF,$9E,$AD,$7C,$3B,$B2,$FF,$48,$5F,$8F,$E0,$7F,$33
	.dc.b	$DB,$F8,$BE,$BF,$DE,$76,$7E,$28,$8E,$09,$5F,$BF,$6B,$F9,$DC,$3F
	.dc.b	$20,$49,$E3,$1F,$C1,$78,$BE,$3B,$FC,$B3,$AF,$E3,$BF,$AC,$EF,$FC
	.dc.b	$86,$BE,$98,$FC,$2E,$8F,$88,$F6,$ED,$EC,$8F,$F4,$0C,$FC,$DF,$88
	.dc.b	$CB,$C3,$E2,$2E,$BF,$07,$DB,$B7,$FE,$81,$AF,$A9,$F0,$D8,$F8,$4C
	.dc.b	$FC,$3E,$BE,$12,$57,$F2,$6E,$3F,$7E,$FF,$B5,$6E,$DF,$26,$7E,$21
	.dc.b	$0F,$89,$C7,$F0,$9B,$3C,$1F,$CA,$78,$BF,$B4,$7E,$21,$FC,$6C,$FE
	.dc.b	$85,$C5,$F1,$2F,$E0,$71,$7A,$BE,$1A,$57,$E3,$7E,$17,$DB,$C3,$EE
	.dc.b	$F8,$6F,$7F,$0F,$B6,$BF,$25,$C1,$EE,$69,$ED,$3D,$1B,$FB,$5B,$9F
	.dc.b	$B5,$F4,$98,$3F,$86,$97,$8F,$F0,$1B,$FF,$45,$EF,$FE,$6F,$3F,$D6
	.dc.b	$FC,$BF,$E6,$EF,$F2,$AF,$27,$F8,$83,$FC,$B3,$C9,$FA,$BC,$3E,$C8
	.dc.b	$FD,$27,$0F,$D8,$E1,$F6,$C7,$EA,$F8,$3D,$AF,$EB,$79,$BD,$8F,$EA
	.dc.b	$F8,$2F,$95,$FC,$D7,$7F,$E1,$DE,$2F,$F0,$46,$D7,$EF,$1A,$FF,$03
	.dc.b	$27,$17,$F0,$36,$F1,$7F,$03,$69,$FC,$36,$BA,$70,$FF,$0E,$8C,$DF
	.dc.b	$4B,$E4,$4B,$E4,$D9,$FE,$18,$7F,$9F,$7C,$FF,$F0,$25,$FE,$9A,$7F
	.dc.b	$C8,$FE,$37,$E6,$AB,$F3,$7F,$0D,$F9,$CF,$86,$FC,$EF,$7F,$E7,$BB
	.dc.b	$E5,$7F,$11,$F8,$6F,$D0,$FC,$37,$E8,$BE,$1B,$F4,$7E,$E7,$E7,$AB
	.dc.b	$F8,$97,$85,$FC,$11,$FB,$A7,$E0,$09,$AF,$E1,$5B,$C1,$F8,$58,$C1
	.dc.b	$1C,$12,$3D,$A7,$F8,$B3,$83,$F0,$73,$EC,$8F,$CF,$9F,$D4,$8F,$E5
	.dc.b	$1D,$3F,$83,$8F,$E0,$9C,$5F,$C6,$F6,$F6,$D7,$F2,$9E,$FF,$D0,$3F
	.dc.b	$BC,$3F,$08,$7E,$78,$F6,$9F,$CC,$F2,$F7,$57,$E6,$23,$F3,$5E,$17
	.dc.b	$ED,$F0,$F5,$47,$C2,$6C,$EE,$3E,$09,$9C,$FD,$D1,$FD,$6F,$C2,$FD
	.dc.b	$E5,$7F,$9F,$BC,$4F,$ED,$F7,$EE,$3F,$40,$7B,$8F,$E8,$DD,$BE,$FC
	.dc.b	$3F,$8F,$76,$FE,$82,$7F,$A3,$76,$FE,$81,$FE,$91,$DB,$FA,$0C,$7E
	.dc.b	$2F,$88,$EE,$FD,$0F,$8D,$E4,$DF,$D9,$9D,$F0,$FF,$89,$70,$7A,$F1
	.dc.b	$F6,$D6,$B3,$F6,$12,$7F,$9E,$F7,$7E,$F2,$3F,$C4,$7E,$1F,$DA,$C7
	.dc.b	$FC,$6D,$E1,$FE,$F3,$1F,$F0,$07,$87,$FB,$CB,$F6,$90,$FE,$C1,$C3
	.dc.b	$ED,$9F,$4F,$0F,$B5,$F5,$F0,$FB,$71,$FE,$81,$C1,$EC,$C3,$F6,$07
	.dc.b	$EE,$74,$FD,$85,$FE,$AF,$5F,$D8,$5F,$EB,$89,$5C,$0F,$DA,$D7,$08
	.dc.b	$CC,$F0,$9D,$57,$3A,$D4,$62,$B7,$18,$AE,$04,$D9,$E3,$7F,$60,$C3
	.dc.b	$FC,$45,$5F,$BB,$BF,$F7,$F7,$85,$FB,$BC,$7F,$C0,$DE,$17,$EE,$DF
	.dc.b	$EB,$87,$F8,$27,$C1,$FC,$FE,$1F,$7B,$C1,$FD,$DC,$FF,$83,$7C,$0F
	.dc.b	$DD,$BF,$9D,$21,$FE,$2F,$D7,$EF,$65,$F8,$18,$FC,$FF,$2F,$E0,$B6
	.dc.b	$FC,$05,$6A,$7F,$01,$18,$26,$DE,$BB,$FF,$06,$C3,$A7,$D9,$5F,$E2
	.dc.b	$FF,$1B,$FA,$C3,$FD,$EB,$BF,$F3,$8F,$F7,$CE,$FF,$5E,$1F,$DE,$FB
	.dc.b	$FF,$73,$3F,$E2,$3F,$1B,$FA,$C3,$FC,$EC,$FF,$15,$7C,$37,$F7,$A3
	.dc.b	$F7,$39,$F9,$76,$F5,$12,$7D,$A4,$3F,$6F,$B7,$B4,$FC,$DF,$81,$FB
	.dc.b	$6C,$BD,$7B,$7E,$DA,$BF,$C2,$87,$F1,$33,$FC,$6D,$AF,$AA,$FF,$C5
	.dc.b	$9E,$07,$F5,$59,$FF,$1F,$6B,$EA,$E3,$FD,$9E,$DE,$A7,$F6,$BD,$1E
	.dc.b	$9A,$F5,$9F,$DC,$4F,$2E,$DE,$B7,$D1,$3F,$B4,$DB,$D9,$1A,$7F,$73
	.dc.b	$D1,$FB,$39,$FF,$37,$77,$7E,$CE,$BF,$CC,$1D,$DF,$B3,$E0,$FF,$55
	.dc.b	$77,$7E,$CE,$7F,$B6,$9F,$B3,$BF,$E1,$3D,$DF,$B3,$AF,$F1,$77,$77
	.dc.b	$EC,$DF,$F1,$87,$77,$EC,$EF,$FC,$6B,$CD,$EB,$7F,$1D,$A7,$E8,$C9
	.dc.b	$E3,$7F,$64,$24,$7E,$DD,$AF,$EC,$9B,$97,$EB,$DB,$87,$ED,$37,$FD
	.dc.b	$5D,$7F,$65,$3F,$CE,$07,$F9,$BC,$87,$F9,$AC,$FE,$CA,$7F,$9C,$7C
	.dc.b	$6F,$EC,$AF,$EA,$C8,$7F,$9C,$7B,$7F,$56,$FF,$9C,$FB,$7F,$56,$B1
	.dc.b	$FE,$A3,$FC,$03,$FE,$55,$8C,$63,$FE,$6E,$8C,$79,$1F,$E0,$FC,$BE
	.dc.b	$D5,$7D,$73,$ED,$D7,$EC,$5C,$12,$FF,$98,$F7,$FE,$B6,$3F,$D4,$B5
	.dc.b	$FA,$D7,$DB,$3F,$EA,$BE,$FF,$D6,$F5,$B2,$7C,$2F,$F9,$81,$7C,$15
	.dc.b	$F7,$31,$F0,$B9,$F8,$7D,$7B,$BB,$F8,$7E,$1F,$0D,$C7,$E1,$79,$7C
	.dc.b	$2C,$E0,$FC,$B7,$07,$E4,$DF,$59,$FC,$1B,$C1,$FC,$AC,$66,$3D,$9C
	.dc.b	$9F,$ED,$9E,$AF,$CB,$3F,$EF,$BF,$07,$F2,$DC,$7F,$EC,$1F,$07,$F2
	.dc.b	$D7,$FE,$B7,$F0,$7E,$CB,$F3,$8F,$F6,$3F,$27,$B1,$FF,$56,$E5,$F5
	.dc.b	$39,$3D,$B1,$8C,$BE,$A7,$17,$BA,$30,$72,$F8,$6F,$8D,$F0,$DE,$F8
	.dc.b	$CE,$9F,$2F,$E1,$BE,$51,$FC,$3B,$83,$DB,$8F,$D9,$CF,$DC,$47,$04
	.dc.b	$BF,$F7,$E7,$07,$C0,$7F,$DA,$5C,$1F,$02,$1C,$1F,$03,$2F,$F7,$17
	.dc.b	$07,$C0,$C7,$D9,$87,$93,$83,$E0,$3F,$0E,$BF,$01,$CD,$F6,$E3,$5C
	.dc.b	$DF,$81,$E0,$F6,$C7,$1E,$4F,$5B,$CF,$83,$FC,$43,$AF,$ED,$A3,$F6
	.dc.b	$84,$3F,$68,$5C,$FF,$68,$DD,$FF,$B0,$F8,$1F,$B6,$8F,$EC,$84,$D1
	.dc.b	$D7,$FB,$88,$FF,$20,$1F,$B8,$3D,$4B,$CF,$FE,$93,$3F,$ED,$79,$FC
	.dc.b	$D7,$5F,$E6,$FE,$1B,$F3,$51,$F0,$FE,$1B,$F3,$DC,$DF,$9F,$EF,$FC
	.dc.b	$D4,$FC,$7E,$CF,$D1,$7A,$BF,$99,$8F,$91,$CC,$ED,$FB,$91,$F9,$AD
	.dc.b	$3F,$A2,$6D,$FB,$8E,$1F,$B9,$3F,$9D,$35,$1F,$D8,$9F,$E2,$47,$FA
	.dc.b	$E0,$F5,$EF,$F6,$F8,$9D,$1E,$A9,$FB,$5D,$BE,$A7,$C7,$D3,$F6,$5F
	.dc.b	$2F,$4F,$A9,$F3,$EF,$EA,$7F,$05,$E7,$7E,$F4,$F4,$BF,$AE,$3C,$C5
	.dc.b	$D7,$D7,$C5,$F9,$88,$CE,$DF,$E6,$AD,$93,$FE,$E6,$D7,$F2,$D6,$5D
	.dc.b	$9F,$97,$EC,$F6,$6F,$F9,$5E,$9F,$E4,$1C,$1F,$C3,$CF,$D8,$E9,$F3
	.dc.b	$BB,$BF,$60,$FD,$0E,$DF,$C9,$BF,$4B,$B7,$D4,$FB,$0F,$E1,$B7,$FA
	.dc.b	$9D,$3D,$31,$F4,$35,$F5,$46,$23,$DD,$AF,$A0,$F3,$BF,$BA,$DB,$F9
	.dc.b	$43,$FE,$CF,$3F,$CE,$1B,$FF,$95,$1F,$F1,$79,$F5,$08,$7D,$43,$F6
	.dc.b	$19,$FA,$A3,$FA,$0E,$7E,$A7,$FA,$1E,$7E,$AA,$FC,$6F,$07,$EB,$3B
	.dc.b	$3F,$49,$9F,$F9,$13,$3F,$D7,$57,$EB,$09,$87,$EB,$1B,$AF,$EB,$1F
	.dc.b	$D7,$1F,$AB,$9F,$F2,$76,$9F,$AB,$AF,$B5,$D1,$FE,$C6,$AF,$E8,$B7
	.dc.b	$FE,$4C,$7F,$8D,$FA,$87,$F2,$47,$E8,$BB,$3F,$25,$7F,$D7,$3B,$3F
	.dc.b	$57,$3F,$C3,$FB,$3F,$25,$3F,$D8,$39,$FF,$25,$3F,$AA,$EA,$FE,$9C
	.dc.b	$7E,$4A,$7F,$4F,$B7,$D5,$6E,$BF,$AB,$D3,$EE,$36,$FF,$AE,$F2,$79
	.dc.b	$5F,$E0,$79,$7F,$61,$C3,$F8,$4F,$0F,$F0,$4E,$07,$85,$FD,$83,$B7
	.dc.b	$FA,$FE,$1F,$63,$4F,$EE,$BE,$17,$E4,$9F,$EE,$FC,$FF,$67,$6F,$EE
	.dc.b	$DE,$17,$F5,$EB,$FE,$59,$CB,$F6,$6F,$FB,$EE,$9E,$BB,$F3,$EF,$F9
	.dc.b	$0E,$0F,$46,$FE,$D9,$CE,$3F,$8F,$AF,$56,$1F,$AD,$7F,$59,$E8,$7F
	.dc.b	$93,$67,$FB,$DE,$3F,$E4,$DA,$FE,$F5,$1F,$E4,$D7,$FD,$73,$E9,$FE
	.dc.b	$AE,$3D,$98,$7F,$83,$7C,$0F,$E0,$D5,$FE,$12,$F0,$3F,$9F,$57,$F4
	.dc.b	$5F,$13,$FA,$D3,$F9,$23,$F5,$87,$FA,$07,$5F,$49,$34,$F6,$12,$3F
	.dc.b	$5C,$7A,$CF,$EC,$7C,$9F,$AD,$F0,$7F,$81,$47,$EA,$B3,$FF,$25,$BF
	.dc.b	$AA,$3F,$96,$F7,$FA,$5F,$D5,$1F,$C0,$CF,$EA,$A7,$F5,$BE,$FF,$D5
	.dc.b	$4F,$F8,$A7,$BF,$E9,$3F,$E2,$BE,$0F,$D5,$57,$E8,$B8,$BF,$55,$DF
	.dc.b	$FA,$CC,$FF,$57,$C1,$FA,$92,$61,$FA,$9B,$FD,$59,$FA,$9A,$FE,$3C
	.dc.b	$7D,$68,$FF,$4C,$F5,$7E,$3E,$FF,$90,$57,$EA,$8F,$F1,$8F,$5F,$FA
	.dc.b	$D5,$FE,$93,$E1,$FF,$53,$7F,$BA,$75,$7F,$8B,$FB,$BF,$53,$1F,$D3
	.dc.b	$B0,$FD,$4E,$3F,$E4,$5E,$EF,$D4,$E9,$FE,$4A,$CB,$F5,$37,$FC,$80
	.dc.b	$FE,$6B,$5F,$8C,$7F,$9D,$72,$7D,$77,$EA,$9F,$DA,$39,$3D,$4F,$F9
	.dc.b	$1B,$1F,$D4,$4F,$F2,$0C,$3F,$9F,$61,$FC,$E2,$BF,$C3,$BF,$B6,$FE
	.dc.b	$A9,$BE,$17,$2E,$14,$BD,$EC,$36,$B8,$FC,$1E,$3D,$FC,$FC,$3E,$BE
	.dc.b	$1F,$7F,$0F,$C3,$E0,$66,$1E,$DC,$FF,$50,$FB,$88,$E0,$9B,$FB,$E3
	.dc.b	$FC,$3D,$CF,$F9,$17,$F9,$E7,$4F,$EA,$79,$3F,$8E,$5F,$F5,$7F,$1F
	.dc.b	$F8,$27,$4F,$A3,$8A,$61,$F0,$3E,$13,$D7,$C3,$E9,$8F,$87,$C1,$E9
	.dc.b	$7E,$2F,$C1,$7A,$1F,$F1,$77,$1B,$83,$F8,$24,$7A,$38,$3C,$DC,$DF
	.dc.b	$C7,$BA,$3F,$55,$3F,$61,$FE,$02,$FE,$AB,$C0,$FD,$4B,$FE,$CC,$3D
	.dc.b	$BA,$FF,$34,$A7,$0F,$B3,$2F,$F0,$4D,$7D,$63,$D4,$7D,$23,$ED,$F3
	.dc.b	$7D,$27,$EE,$F3,$7D,$8D,$7D,$58,$74,$97,$1F,$DE,$DB,$F5,$B1,$FA
	.dc.b	$68,$FD,$6B,$FA,$FE,$5F,$D6,$E7,$FA,$BB,$FD,$6C,$FE,$AC,$D7,$6F
	.dc.b	$F5,$EE,$0F,$EA,$95,$F5,$17,$8D,$FD,$56,$BF,$C4,$A7,$EB,$71,$FF
	.dc.b	$1F,$3F,$94,$E6,$FC,$79,$F9,$0E,$99,$1F,$98,$F7,$BF,$27,$E3,$FE
	.dc.b	$53,$93,$F2,$BE,$6F,$E4,$2B,$F3,$BD,$95,$FF,$69,$ED,$FC,$1C,$8F
	.dc.b	$A0,$87,$C1,$E0,$F7,$BE,$D3,$F2,$31,$FA,$2F,$86,$FD,$1F,$C3,$7E
	.dc.b	$93,$E1,$E5,$7E,$B3,$E1,$FF,$4F,$7F,$A8,$C3,$F5,$9E,$0F,$F5,$DA
	.dc.b	$FF,$3C,$2F,$8A,$CF,$C4,$6B,$E2,$37,$F0,$7C,$3E,$0F,$8F,$C2,$72
	.dc.b	$F8,$5E,$7F,$0B,$33,$FE,$F9,$EE,$39,$FE,$B7,$6C,$9F,$57,$77,$D8
	.dc.b	$EC,$FB,$B1,$EF,$E2,$F8,$6D,$3F,$9F,$BE,$8F,$82,$F7,$C7,$F0,$8E
	.dc.b	$6F,$5F,$C2,$FB,$3E,$0F,$DB,$F0,$BE,$EF,$86,$F7,$DF,$C0,$F8,$6F
	.dc.b	$6B,$FE,$AC,$E8,$73,$7C,$4F,$6F,$DD,$39,$F8,$5F,$8F,$F0,$BF,$23
	.dc.b	$E2,$24,$FB,$B9,$3D,$9C,$5E,$AE,$3F,$67,$6F,$D8,$8C,$72,$EA,$30
	.dc.b	$4E,$5F,$7D,$7E,$8E,$BF,$D2,$1C,$BE,$EC,$7F,$7F,$CB,$EE,$7D,$1C
	.dc.b	$9E,$E7,$D5,$C7,$EE,$7D,$9C,$7E,$E7,$EE,$F1,$7B,$5F,$81,$D5,$F8
	.dc.b	$2E,$2F,$6C,$7C,$3E,$1F,$6A,$E7,$FB,$4F,$ED,$B9,$5F,$0D,$F8,$A2
	.dc.b	$47,$CB,$6B,$F8,$C6,$F0,$7B,$67,$8C,$EC,$86,$CB,$1B,$9E,$3C,$7F
	.dc.b	$CB,$2B,$8D,$6C,$95,$B7,$8C,$EE,$BF,$4D,$1F,$7F,$E2,$3F,$4A,$FF
	.dc.b	$2A,$F0,$7F,$A5,$F8,$3F,$A6,$8F,$EA,$7C,$DF,$A6,$F8,$8F,$D3,$F8
	.dc.b	$5F,$D4,$FF,$80,$7F,$8C,$57,$FB,$7D,$FF,$1F,$BF,$AA,$F0,$B5,$EF
	.dc.b	$FE,$9A,$3F,$1D,$E1,$7E,$95,$FD,$7F,$85,$FC,$7B,$83,$F0,$FD,$BF
	.dc.b	$82,$E1,$FC,$33,$6B,$F4,$F1,$87,$FC,$21,$E9,$3C,$6F,$C3,$5F,$E2
	.dc.b	$61,$C3,$EE,$8C,$70,$FB,$6F,$D1,$EE,$FA,$CF,$67,$C1,$7B,$7E,$13
	.dc.b	$DD,$F0,$7E,$FF,$84,$F8,$1F,$01,$F0,$7E,$0B,$E1,$7C,$44,$CB,$FC
	.dc.b	$F5,$A7,$E6,$79,$BF,$DB,$51,$FA,$53,$D2,$4F,$1F,$F4,$6D,$EA,$FD
	.dc.b	$13,$FC,$82,$BE,$1B,$78,$FF,$01,$E3,$FB,$9F,$EA,$5E,$3F,$E8,$E7
	.dc.b	$F8,$CF,$8F,$FA,$37,$DA,$79,$27,$FB,$9E,$4E,$2F,$B5,$F0,$9E,$FA
	.dc.b	$D1,$FD,$4A,$5C,$3F,$7F,$E1,$7F,$01,$F1,$3F,$81,$F8,$8F,$81,$7F
	.dc.b	$81,$E1,$BE,$2F,$E9,$B8,$BD,$A4,$E2,$F6,$D7,$F1,$B2,$71,$7B,$9B
	.dc.b	$BF,$F8,$22,$5D,$9F,$C7,$E7,$F8,$F1,$0F,$E6,$FE,$87,$F1,$F7,$F8
	.dc.b	$D7,$A1,$F6,$DF,$5E,$2E,$0F,$81,$18,$5C,$3F,$06,$3F,$35,$C3,$EF
	.dc.b	$75,$C1,$EF,$78,$6F,$F0,$36,$FB,$CF,$FA,$97,$7F,$81,$E3,$FE,$D2
	.dc.b	$7F,$65,$5F,$B4,$E0,$FD,$93,$FB,$43,$F8,$57,$8F,$FB,$C6,$99,$D5
	.dc.b	$7C,$EB,$8C,$F3,$36,$59,$DD,$73,$6C,$73,$AE,$4F,$0B,$FD,$37,$81
	.dc.b	$FD,$8B,$C7,$FD,$36,$3F,$E1,$DF,$03,$F8,$1D,$FF,$95,$38,$FF,$80
	.dc.b	$7C,$3F,$F0,$1F,$81,$FD,$3C,$FC,$BF,$87,$FE,$09,$EE,$7E,$9E,$3E
	.dc.b	$77,$81,$FA,$77,$E8,$7C,$3F,$F0,$9D,$E4,$7E,$97,$83,$F0,$D9,$FE
	.dc.b	$FD,$B1,$FA,$13,$F4,$D1,$9E,$AF,$E9,$9E,$37,$F0,$19,$FC,$27,$17
	.dc.b	$F0,$1D,$FD,$71,$FE,$2C,$97,$8D,$FC,$06,$BF,$83,$61,$FE,$6D,$FE
	.dc.b	$01,$FF,$38,$5A,$7F,$A9,$47,$F8,$FA,$31,$C5,$FE,$5C,$8F,$F3,$84
	.dc.b	$60,$FF,$1C,$B8,$9E,$3F,$0D,$D7,$E1,$3B,$7C,$37,$7F,$85,$F0,$7B
	.dc.b	$B3,$1F,$85,$BF,$E9,$B1,$FD,$21,$33,$FE,$41,$3F,$A6,$3E,$19,$FE
	.dc.b	$B1,$E1,$79,$1F,$C1,$63,$E0,$79,$1F,$C1,$6F,$FA,$87,$91,$FC,$15
	.dc.b	$F6,$FC,$27,$C1,$F8,$3F,$85,$F0,$FF,$0F,$E1,$26,$3F,$AA,$E3,$71
	.dc.b	$FC,$7F,$86,$F9,$1E,$F7,$C9,$F8,$5F,$95,$F0,$B3,$0F,$5F,$47,$D7
	.dc.b	$23,$F6,$FB,$3E,$EB,$EB,$7F,$83,$F1,$FA,$A3,$ED,$F2,$7F,$1E,$7F
	.dc.b	$C2,$DE,$6F,$EF,$DF,$F1,$BF,$9B,$F6,$78,$BD,$51,$E7,$E1,$F4,$BF
	.dc.b	$57,$17,$07,$D5,$C7,$ED,$70,$7A,$EB,$DD,$BF,$AD,$F8,$1B,$7A,$DF
	.dc.b	$F0,$14,$7E,$CB,$8F,$F7,$A4,$BC,$12,$7F,$5C,$DC,$7F,$67,$1F,$B7
	.dc.b	$DB,$F6,$71,$8A,$E5,$3C,$09,$19,$AE,$55,$C0,$9E,$3F,$EF,$FB,$BF
	.dc.b	$45,$7F,$E5,$AF,$1B,$F4,$38,$7F,$B2,$3C,$0F,$DD,$CF,$FA,$A3,$8F
	.dc.b	$F4,$59,$7E,$6F,$E1,$7F,$45,$F0,$F2,$BF,$01,$E4,$FE,$97,$B3,$F3
	.dc.b	$51,$F8,$3F,$07,$F3,$4F,$E1,$7C,$0F,$CD,$3F,$18,$FD,$1E,$7F,$AD
	.dc.b	$23,$F8,$B6,$BF,$57,$97,$FB,$BB,$92,$19,$8D,$F1,$7E,$D1,$BA,$7E
	.dc.b	$CD,$B3,$E0,$38,$CE,$DB,$3B,$69,$FB,$8F,$0F,$F8,$16,$DF,$C0,$FC
	.dc.b	$19,$3F,$D2,$BF,$80,$7F,$D0,$D8,$7F,$95,$5F,$F4,$34,$67,$E1,$FF
	.dc.b	$DE,$D1,$FE,$6A,$8E,$1E,$E6,$3E,$07,$3F,$0F,$AF,$77,$7E,$EF,$0F
	.dc.b	$23,$8F,$87,$CB,$C8,$99,$7F,$25,$E3,$F8,$AF,$C3,$3D,$3D,$5F,$9C
	.dc.b	$E4,$F8,$2D,$8C,$CF,$0E,$0F,$85,$1C,$48,$E0,$9C,$9F,$16,$32,$D8
	.dc.b	$CC,$7C,$D3,$25,$7F,$93,$F2,$7C,$A8,$FE,$11,$C7,$F2,$9F,$D9,$71
	.dc.b	$7C,$AB,$47,$FB,$82,$30,$DF,$13,$E5,$FC,$1F,$CC,$E4,$F9,$53,$F4
	.dc.b	$FE,$17,$E7,$FC,$2F,$CA,$8F,$AD,$C5,$F2,$9F,$B1,$F0,$FF,$50,$FE
	.dc.b	$ED,$E2,$7A,$F4,$93,$E9,$6F,$0F,$C9,$8F,$6B,$78,$3E,$0B,$67,$7C
	.dc.b	$3F,$35,$BC,$5F,$4A,$31,$7D,$49,$5C,$5B,$1A,$8F,$E2,$BC,$5F,$26
	.dc.b	$7F,$49,$C5,$F2,$71,$FC,$84,$7E,$38,$FC,$7F,$0F,$C6,$7F,$1E,$7E
	.dc.b	$5F,$A3,$F2,$7C,$1F,$1A,$3F,$35,$CD,$F8,$F7,$F3,$9C,$DF,$15,$F8
	.dc.b	$E7,$F7,$3E,$57,$C3,$7E,$70,$91,$FA,$36,$BF,$71,$B9,$7E,$95,$BC
	.dc.b	$1F,$7B,$65,$F4,$BE,$44,$39,$16,$F9,$36,$B9,$4F,$18,$D9,$27,$6F
	.dc.b	$E3,$BE,$1F,$F6,$7F,$0F,$FB,$4F,$1B,$F6,$BE,$2F,$ED,$BE,$1F,$F6
	.dc.b	$FF,$0F,$FB,$8F,$03,$F2,$78,$FF,$47,$9F,$F4,$EE,$5F,$ED,$37,$FD
	.dc.b	$3A,$DF,$87,$FD,$DC,$FF,$05,$F0,$38,$CE,$FE,$0E,$EF,$F0,$70,$7C
	.dc.b	$3E,$30,$70,$A5,$F0,$FF,$C0,$27,$F4,$FE,$1F,$E6,$1D,$CF,$F9,$D1
	.dc.b	$FF,$2E,$B5,$F4,$7C,$0F,$5B,$E8,$D9,$D6,$7E,$B8,$C1,$2B,$9B,$5E
	.dc.b	$66,$08,$60,$91,$CD,$CB,$5C,$E5,$F0,$67,$C5,$F0,$FE,$3F,$6F,$C9
	.dc.b	$F0,$FE,$5F,$82,$98,$7B,$3E,$13,$D1,$F0,$FE,$9F,$85,$F5,$7C,$3F
	.dc.b	$AF,$E1,$7D,$9F,$0B,$ED,$F8,$5F,$77,$C2,$CC,$BE,$2F,$C2,$FC,$1F
	.dc.b	$6E,$47,$F9,$3F,$93,$E1,$F1,$F9,$63,$D2,$D7,$D2,$7D,$0E,$3F,$89
	.dc.b	$C9,$ED,$24,$ED,$B3,$B3,$99,$4E,$7C,$BF,$51,$CB,$5C,$DF,$91,$E1
	.dc.b	$3F,$91,$E4,$F8,$F1,$F8,$0E,$3F,$8E,$FE,$5F,$8F,$F0,$1F,$BF,$FF
	.dc.b	$96,$AF,$F3,$1C,$7F,$22,$BF,$D4,$73,$FE,$5B,$96,$7F,$B5,$8F,$F1
	.dc.b	$24,$FE,$51,$FF,$1F,$BF,$8E,$DF,$E6,$3F,$A5,$DF,$E5,$BF,$A7,$DF
	.dc.b	$F2,$DB,$7A,$F1,$F8,$6D,$C7,$E3,$12,$BD,$AD,$8D,$46,$08,$E0,$9B
	.dc.b	$7C,$56,$EF,$F5,$2B,$94,$E4,$91,$9A,$FD,$4E,$FF,$29,$FD,$F6,$FF
	.dc.b	$2B,$1F,$DF,$6F,$F2,$BE,$13,$F2,$15,$FA,$EE,$4F,$9B,$ED,$FC,$EF
	.dc.b	$84,$F9,$FF,$01,$F1,$31,$FE,$0D,$C5,$FA,$FE,$2F,$8B,$1F,$C2,$78
	.dc.b	$BE,$2F,$1B,$E1,$BE,$B9,$34,$FA,$AD,$7F,$CB,$11,$B7,$24,$33,$1B
	.dc.b	$AF,$B7,$1F,$66,$31,$C7,$F7,$5B,$19,$39,$4F,$0E,$45,$F7,$9E,$A6
	.dc.b	$AF,$93,$FC,$43,$AF,$F1,$FE,$E7,$E4,$3E,$1B,$F2,$3E,$EF,$E4,$BE
	.dc.b	$13,$F2,$7F,$09,$FB,$AC,$3F,$23,$F0,$DF,$96,$F6,$7F,$75,$1F,$C6
	.dc.b	$BB,$FF,$74,$FE,$5B,$E1,$BF,$37,$F0,$DF,$9C,$F8,$5F,$CE,$FC,$2C
	.dc.b	$AF,$F6,$47,$0F,$9F,$8F,$D4,$DE,$2F,$3C,$FA,$89,$3A,$E4,$F8,$7C
	.dc.b	$9F,$01,$B7,$C6,$BE,$31,$C2,$F9,$DF,$18,$F8,$86,$9C,$3B,$31,$1C
	.dc.b	$0F,$D7,$7C,$2F,$ED,$BD,$EF,$DB,$FB,$BF,$B8,$F8,$59,$7F,$B6,$F7
	.dc.b	$BF,$9D,$7F,$00,$FF,$89,$35,$3F,$E5,$09,$FF,$4E,$2F,$4F,$67,$0F
	.dc.b	$85,$E3,$F0,$B2,$BF,$C5,$FB,$7D,$FB,$7C,$07,$6F,$86,$EF,$EB,$4B
	.dc.b	$F9,$BC,$7F,$10,$8F,$C0,$D3,$E4,$37,$8F,$E4,$46,$23,$E1,$37,$2F
	.dc.b	$8A,$4E,$3F,$89,$3F,$18,$93,$B8,$CB,$63,$FB,$3E,$6F,$80,$F8,$BF
	.dc.b	$0B,$F2,$A7,$F8,$EF,$17,$CA,$7F,$53,$C5,$F2,$9F,$97,$ED,$7C,$CF
	.dc.b	$81,$F9,$BE,$BF,$C7,$9F,$8B,$A7,$F8,$5F,$37,$1F,$C9,$9F,$9E,$FD
	.dc.b	$4F,$83,$FA,$BF,$07,$27,$E3,$9F,$0F,$8B,$E3,$BF,$0B,$8B,$E2,$92
	.dc.b	$33,$C5,$F0,$E7,$E0,$CE,$5E,$44,$3E,$5E,$3F,$05,$F3,$9F,$13,$83
	.dc.b	$E1,$3F,$C2,$38,$3E,$15,$FE,$93,$83,$E1,$4F,$0E,$2F,$C5,$F0,$FC
	.dc.b	$38,$C6,$2E,$0F,$89,$1F,$00,$CE,$FF,$0F,$1F,$C5,$6F,$F0,$67,$F2
	.dc.b	$7B,$FF,$42,$C1,$F0,$DF,$9A,$24,$F6,$6B,$F9,$A6,$DF,$E5,$5B,$C3
	.dc.b	$F4,$DB,$3F,$82,$6E,$9F,$A4,$6D,$F8,$27,$B1,$0E,$C5,$BF,$E0,$FD
	.dc.b	$7F,$B3,$8F,$83,$F0,$DF,$B0,$F8,$5F,$D8,$FC,$27,$EC,$BE,$13,$F6
	.dc.b	$7F,$09,$FB,$4F,$84,$FD,$7E,$1F,$23,$BF,$F6,$FE,$DF,$F8,$C7,$F8
	.dc.b	$07,$FC,$93,$C3,$FE,$61,$7F,$C9,$2F,$E3,$7C,$0F,$D8,$5F,$D0,$EF
	.dc.b	$FD,$83,$FE,$E9,$E3,$5B,$E1,$7A,$12,$7E,$D4,$65,$B1,$95,$C5,$FC
	.dc.b	$F6,$75,$18,$C7,$C9,$5B,$AD,$36,$33,$3D,$08,$71,$3F,$CB,$35,$D2
	.dc.b	$B8,$C6,$9A,$EE,$FD,$77,$D6,$38,$FC,$2F,$C7,$F8,$5F,$91,$F0,$BF
	.dc.b	$27,$E0,$3E,$57,$AB,$F2,$FE,$03,$E6,$7C,$2F,$CD,$F8,$19,$97,$C2
	.dc.b	$F8,$1F,$A1,$F0,$BF,$47,$E0,$E4,$FC,$6F,$82,$FA,$9F,$0B,$F5,$7D
	.dc.b	$69,$3F,$ED,$2F,$89,$7B,$1F,$02,$3D,$1C,$DE,$DE,$7C,$73,$FB,$08
	.dc.b	$E0,$93,$F1,$39,$BD,$DC,$FF,$02,$3E,$24,$EA,$31,$1D,$A7,$77,$DE
	.dc.b	$39,$1C,$0A,$6C,$FC,$0F,$3F,$E3,$FE,$1B,$F2,$1F,$05,$F2,$67,$F0
	.dc.b	$BC,$FF,$93,$F8,$2F,$CA,$7C,$17,$FA,$47,$F7,$CF,$F7,$8F,$C3,$DB
	.dc.b	$FF,$22,$6D,$F9,$7E,$7F,$97,$1F,$8E,$E6,$F9,$6F,$E0,$6F,$FD,$F3
	.dc.b	$F0,$DC,$7E,$1B,$E5,$CF,$C8,$E6,$FF,$21,$F0,$FA,$F8,$FD,$FC,$9F
	.dc.b	$22,$BD,$6D,$8C,$E9,$F1,$6B,$81,$39,$3E,$29,$2F,$E2,$57,$07,$9D
	.dc.b	$71,$9F,$97,$C9,$F1,$67,$F4,$FC,$7F,$15,$FB,$BC,$7F,$15,$FD,$5F
	.dc.b	$17,$C3,$E2,$F8,$9F,$0F,$FB,$BE,$3F,$87,$5F,$84,$E3,$F8,$7B,$FF
	.dc.b	$8C,$38,$FE,$1C,$FF,$34,$4F,$F9,$A8,$F8,$87,$EE,$38,$7F,$D7,$91
	.dc.b	$AF,$5F,$E9,$12,$79,$1F,$E1,$38,$D3,$88,$D3,$F4,$CD,$11,$CD,$79
	.dc.b	$63,$07,$12,$1C,$63,$7C,$1E,$8A,$EF,$39,$37,$5C,$FE,$1F,$F7,$11
	.dc.b	$FC,$4B,$C1,$FC,$57,$C3,$FE,$E2,$3F,$8A,$F8,$3F,$B8,$7F,$8B,$F8
	.dc.b	$1F,$B8,$7F,$5A,$FE,$47,$E1,$FF,$25,$F0,$FF,$93,$F8,$7F,$E5,$5E
	.dc.b	$17,$EE,$2F,$E1,$FC,$3F,$E5,$CF,$92,$FE,$E0,$F8,$FE,$0F,$E6,$C8
	.dc.b	$E0,$87,$C7,$72,$D3,$3B,$7C,$78,$CB,$63,$3C,$1F,$15,$F0,$1C,$49
	.dc.b	$1B,$6C,$6E,$7F,$5D,$F0,$FF,$AF,$F8,$7F,$D8,$7C,$3F,$EC,$7E,$1F
	.dc.b	$F6,$5E,$B7,$EC,$FE,$1F,$F6,$9F,$0F,$FB,$5F,$81,$9A,$7E,$13,$E0
	.dc.b	$7F,$75,$FC,$03,$F8,$E8,$FC,$83,$FE,$45,$DF,$F2,$8D,$DF,$EE,$3F
	.dc.b	$E4,$79,$DF,$C4,$72,$F8,$19,$1F,$C8,$F8,$77,$EA,$F6,$24,$76,$6B
	.dc.b	$FC,$0E,$FF,$C9,$6E,$48,$66,$37,$8F,$91,$B8,$F6,$9D,$CE,$5B,$19
	.dc.b	$79,$18,$1D,$AB,$F0,$C7,$8F,$E2,$3E,$73,$F1,$3D,$6F,$9C,$FC,$6F
	.dc.b	$88,$F9,$CF,$C8,$F8,$6F,$97,$F1,$1F,$33,$E1,$BE,$6F,$C3,$7C,$EA
	.dc.b	$F9,$FF,$0F,$1F,$9F,$F0,$FF,$4B,$8F,$E9,$FC,$3C,$9F,$94,$79,$B6
	.dc.b	$F2,$F0,$FB,$63,$D6,$47,$04,$E1,$F8,$0D,$8E,$5C,$5F,$21,$A7,$C3
	.dc.b	$E2,$F8,$51,$96,$C6,$6F,$C3,$7D,$09,$1C,$8F,$90,$7F,$10,$EC,$FC
	.dc.b	$6F,$17,$C8,$8F,$BD,$C5,$F2,$1F,$C0,$71,$7C,$87,$F2,$3C,$3F,$21
	.dc.b	$FC,$99,$F1,$CF,$C3,$70,$7E,$63,$83,$E4,$47,$E6,$37,$F9,$0B,$D9
	.dc.b	$FD,$E6,$DF,$CA,$B9,$7F,$67,$B7,$EA,$EB,$F6,$67,$E0,$8F,$D9,$F0
	.dc.b	$FE,$C4,$91,$B8,$FD,$99,$B9,$E6,$D3,$9D,$77,$69,$D2,$79,$57,$48
	.dc.b	$E0,$74,$9F,$E7,$FD,$DF,$B4,$F8,$6F,$DE,$CF,$ED,$3E,$1B,$F7,$AF
	.dc.b	$ED,$BE,$1B,$F7,$AF,$F3,$EE,$CF,$DE,$BF,$CE,$FB,$3F,$7B,$F0,$DF
	.dc.b	$BE,$F8,$6F,$DF,$FC,$37,$F0,$0F,$86,$FE,$03,$F0,$DF,$B8,$C3,$F6
	.dc.b	$5E,$0F,$F4,$CE,$0F,$F2,$8C,$7E,$99,$BC,$1F,$E3,$98,$FE,$F5,$18
	.dc.b	$27,$0F,$F8,$15,$B1,$FD,$6D,$AF,$08,$D1,$1D,$35,$D1,$82,$18,$24
	.dc.b	$7F,$69,$3F,$CC,$AF,$9F,$DB,$F1,$7B,$7E,$3F,$86,$F2,$7A,$5E,$5F
	.dc.b	$6F,$CD,$F0,$DE,$7F,$4F,$D1,$F0,$D3,$2F,$47,$A9,$EB,$F6,$BD,$9F
	.dc.b	$0D,$ED,$F6,$FD,$DE,$A7,$BF,$E1,$A6,$1F,$B8,$E3,$FF,$34,$FB,$9F
	.dc.b	$0C,$91,$F0,$DA,$FC,$A6,$F1,$FA,$DB,$C9,$FD,$AE,$11,$89,$FA,$8D
	.dc.b	$AE,$F1,$D4,$87,$52,$CF,$56,$D7,$5F,$8E,$FB,$1C,$BF,$67,$E1,$FE
	.dc.b	$45,$7D,$9E,$4F,$90,$FD,$BF,$82,$FB,$DF,$0F,$F7,$FE,$0F,$FC,$89
	.dc.b	$FC,$02,$F8,$4B,$F0,$98,$F8,$8F,$93,$8F,$E0,$F9,$7E,$4B,$F8,$5E
	.dc.b	$4A,$FE,$E7,$CA,$F8,$1C,$9E,$DE,$3F,$83,$3E,$83,$E0,$97,$93,$E0
	.dc.b	$BE,$F2,$47,$BD,$BC,$9D,$1E,$11,$D1,$C9,$D0,$B3,$D2,$39,$12,$7E
	.dc.b	$44,$FE,$77,$8F,$E4,$3F,$9F,$ED,$FC,$7B,$FA,$1E,$DF,$D2,$71,$FC
	.dc.b	$88,$FD,$27,$17,$C8,$E1,$F8,$F1,$FE,$41,$E2,$70,$FC,$88,$FD,$7F
	.dc.b	$0F,$C6,$D7,$F4,$1C,$3F,$23,$CB,$FE,$3F,$1F,$AE,$3F,$49,$BF,$EA
	.dc.b	$DB,$7F,$AA,$6B,$FA,$A8,$FC,$71,$B2,$BB,$6B,$B3,$F8,$23,$B6,$9B
	.dc.b	$29,$BE,$06,$5F,$E5,$BA,$FF,$30,$3F,$E5,$55,$DF,$F8,$AF,$84,$FC
	.dc.b	$5F,$C2,$7E,$33,$E1,$3F,$6F,$8F,$F0,$BF,$07,$F6,$CF,$F1,$2F,$07
	.dc.b	$F8,$87,$83,$FB,$78,$F9,$5F,$09,$FB,$77,$E6,$7C,$27,$ED,$9F,$E3
	.dc.b	$3E,$0F,$F9,$AB,$67,$C2,$7E,$68,$93,$F4,$DA,$FF,$A6,$63,$6E,$48
	.dc.b	$66,$37,$BF,$B9,$BC,$3F,$E6,$5A,$E3,$5B,$38,$D6,$DB,$3B,$AE,$31
	.dc.b	$FE,$18,$AE,$35,$F2,$BB,$7F,$63,$E8,$FE,$CB,$D9,$FD,$9F,$C2,$7E
	.dc.b	$D3,$E1,$3F,$6B,$F0,$93,$2F,$C7,$7C,$27,$EE,$3E,$12,$47,$E4,$7D
	.dc.b	$4F,$DD,$FA,$9F,$EB,$FF,$03,$FC,$8D,$3F,$EB,$97,$FC,$88,$FF,$8A
	.dc.b	$9F,$F2,$23,$9F,$81,$E0,$7C,$DE,$1F,$4F,$6F,$E0,$9B,$C5,$EA,$E3
	.dc.b	$F7,$5F,$C0,$3D,$D3,$A2,$39,$AE,$4D,$AE,$4F,$03,$13,$E0,$C1,$E8
	.dc.b	$7A,$FE,$07,$E3,$57,$EB,$F8,$7D,$DF,$09,$EF,$F8,$1F,$81,$F0,$3F
	.dc.b	$C7,$38,$BE,$17,$C0,$FC,$3F,$84,$FE,$3F,$C7,$F1,$7E,$0B,$E3,$67
	.dc.b	$F1,$FE,$0B,$E4,$7B,$5F,$27,$E1,$24,$FF,$45,$E3,$F3,$F2,$7A,$A3
	.dc.b	$F0,$44,$EE,$FC,$13,$79,$BE,$2B,$4F,$71,$67,$94,$F0,$24,$F0,$6C
	.dc.b	$F0,$6C,$F0,$9D,$12,$7E,$E7,$27,$C7,$7F,$8A,$76,$7E,$31,$FF,$6B
	.dc.b	$73,$39,$3E,$44,$63,$8F,$E4,$3A,$E2,$F9,$0F,$E3,$4E,$3C,$1F,$21
	.dc.b	$E7,$C1,$F2,$1E,$BB,$FC,$87,$FA,$21,$F8,$FF,$83,$FC,$A1,$1F,$1B
	.dc.b	$5F,$CB,$37,$7F,$F7,$76,$68,$C6,$DF,$ED,$78,$C5,$72,$21,$C8,$B5
	.dc.b	$CA,$B8,$C7,$8C,$CD,$73,$8C,$EF,$FC,$83,$BF,$F6,$B5,$FC,$F7,$BF
	.dc.b	$F5,$9F,$07,$FB,$58,$FE,$81,$DF,$FB,$57,$F5,$DF,$07,$FB,$47,$F8
	.dc.b	$87,$81,$FB,$3F,$83,$FD,$AC,$7D,$2F,$83,$FD,$B7,$C1,$FE,$DF,$E0
	.dc.b	$FF,$63,$3F,$5B,$C3,$FF,$5F,$7A,$1F,$E6,$4E,$37,$C1,$FE,$F4,$95
	.dc.b	$F7,$1A,$FE,$FD,$BC,$9F,$29,$BC,$9F,$25,$B3,$FC,$0D,$BC,$7F,$E0
	.dc.b	$36,$DF,$82,$7B,$10,$EC,$5B,$FA,$FF,$07,$FE,$D5,$58,$27,$FD,$73
	.dc.b	$FC,$03,$FE,$54,$3F,$C8,$38,$97,$B7,$C6,$B8,$4E,$7D,$4C,$5F,$2F
	.dc.b	$6E,$CF,$1F,$7E,$E2,$6F,$B9,$8F,$84,$CF,$AB,$AF,$57,$7E,$AF,$0F
	.dc.b	$84,$E3,$ED,$F2,$F8,$49,$B7,$F9,$17,$47,$B9,$D8,$91,$ED,$E3,$F7
	.dc.b	$BE,$E8,$CA,$8C,$EF,$EF,$9C,$C6,$2B,$AD,$73,$E2,$F8,$31,$88,$E5
	.dc.b	$18,$AE,$2D,$8E,$07,$84,$B7,$DD,$B5,$CE,$FC,$17,$CA,$78,$1D,$8B
	.dc.b	$1A,$7F,$84,$F2,$7C,$EF,$5F,$E7,$FC,$07,$D0,$F8,$3F,$A3,$CD,$F4
	.dc.b	$BE,$07,$E9,$F3,$4C,$7C,$3F,$07,$F4,$DF,$1F,$37,$D3,$7C,$BC,$9F
	.dc.b	$6B,$E0,$BE,$9C,$7A,$BF,$80,$7F,$CA,$CD,$EC,$FD,$2F,$93,$F0,$63
	.dc.b	$D1,$C9,$EF,$6F,$27,$BF,$97,$5C,$9E,$B9,$CF,$2F,$C1,$BE,$13,$92
	.dc.b	$57,$02,$57,$C4,$6C,$F0,$6D,$F7,$9D,$12,$74,$FF,$71,$E5,$F9,$4F
	.dc.b	$D4,$E4,$F9,$4F,$D6,$E3,$F9,$4F,$D8,$E2,$F9,$4F,$B3,$8B,$E5,$3E
	.dc.b	$EE,$2F,$94,$FC,$0E,$2F,$94,$FE,$03,$9F,$F4,$FC,$5F,$2A,$3E,$3B
	.dc.b	$7B,$FF,$87,$77,$FF,$00,$24,$FE,$F7,$83,$F8,$43,$4F,$E1,$05,$DF
	.dc.b	$F8,$43,$79,$3F,$84,$3F,$C1,$89,$C5,$FC,$4D,$D5,$F6,$73,$7F,$A0
	.dc.b	$3F,$77,$F0,$BF,$BC,$F0,$24,$FE,$6B,$C0,$FD,$F9,$EC,$F8,$5F,$DF
	.dc.b	$C7,$E7,$BC,$1F,$DF,$BF,$89,$F8,$5F,$83,$F1,$5F,$BF,$8F,$C6,$7C
	.dc.b	$47,$EF,$9F,$C3,$F8,$3F,$8C,$5C,$0E,$27,$C3,$FC,$AF,$87,$95,$F9
	.dc.b	$56,$BF,$31,$BC,$5F,$C0,$29,$18,$E1,$FE,$3B,$18,$AE,$44,$39,$16
	.dc.b	$B9,$57,$18,$F2,$CF,$17,$64,$9D,$BF,$A9,$BF,$C0,$FC,$3F,$E0,$BE
	.dc.b	$1F,$F0,$7F,$0F,$F8,$4F,$85,$FC,$2F,$87,$F8,$6F,$53,$F0,$FE,$E7
	.dc.b	$E2,$3E,$13,$F1,$3F,$0F,$F8,$AF,$5B,$F1,$7F,$09,$F8,$CF,$5B,$F1
	.dc.b	$BF,$07,$F8,$EF,$87,$FE,$6B,$FC,$03,$FE,$52,$E3,$FE,$42,$FF,$95
	.dc.b	$3C,$36,$DF,$CF,$A3,$FC,$AB,$D1,$F3,$F9,$FD,$CD,$E6,$F8,$5D,$1F
	.dc.b	$22,$73,$D1,$F0,$72,$F7,$D7,$C1,$AD,$3C,$0F,$8F,$5C,$08,$E6,$FB
	.dc.b	$3C,$4F,$C0,$E1,$E4,$3F,$CC,$F7,$EA,$F5,$7D,$BF,$07,$EE,$F8,$8F
	.dc.b	$7F,$C2,$7C,$0F,$88,$F9,$38,$FF,$19,$E7,$F8,$7F,$0B,$F2,$63,$F8
	.dc.b	$E7,$2F,$C9,$7F,$9B,$E8,$F8,$5F,$93,$F1,$1F,$2B,$E1,$A5,$7F,$6B
	.dc.b	$E5,$F9,$2F,$9B,$E1,$BD,$6F,$A7,$8B,$D7,$C5,$F1,$09,$C9,$F1,$27
	.dc.b	$5C,$BF,$26,$31,$3A,$9F,$8E,$7C,$59,$D3,$83,$8F,$2F,$82,$7E,$EE
	.dc.b	$1F,$80,$F8,$8F,$C0,$F3,$7C,$6B,$F6,$F2,$FC,$67,$DF,$C9,$F1,$9F
	.dc.b	$E2,$7C,$9F,$19,$FE,$81,$D7,$F8,$E7,$FA,$17,$5F,$E3,$38,$FE,$2C
	.dc.b	$7C,$8E,$2F,$8A,$FF,$94,$38,$7F,$71,$E1,$FE,$4B,$E1,$E4,$FC,$E6
	.dc.b	$BF,$D2,$39,$91,$8E,$1F,$CB,$B7,$93,$F2,$CD,$AF,$C2,$B7,$05,$F2
	.dc.b	$9E,$C4,$3B,$4E,$E3,$51,$83,$F6,$C7,$F5,$FF,$13,$F6,$CF,$F6,$1F
	.dc.b	$0F,$F6,$CF,$F6,$3E,$DF,$DB,$3F,$88,$F8,$7F,$DB,$3F,$D9,$FC,$1F
	.dc.b	$D8,$FC,$3F,$EC,$BE,$1F,$F6,$7F,$0F,$FB,$6A,$FE,$D9,$E1,$FE,$D9
	.dc.b	$87,$E3,$71,$FD,$A7,$C3,$FE,$EB,$E1,$E4,$FE,$59,$AF,$C9,$6F,$17
	.dc.b	$EF,$9B,$C5,$FB,$D6,$CF,$EB,$1B,$C5,$F2,$1B,$C2,$69,$27,$F4,$B8
	.dc.b	$7F,$0C,$F8,$7F,$E1,$BE,$AF,$F0,$EF,$7B,$FC,$87,$97,$F8,$CA,$FF
	.dc.b	$C0,$6F,$F9,$38,$FF,$26,$BF,$E2,$45,$EE,$F0,$F5,$78,$FB,$BC,$BD
	.dc.b	$5E,$7E,$E7,$4F,$87,$EB,$F0,$FD,$BD,$EE,$FF,$0F,$34,$F9,$07,$F4
	.dc.b	$8D,$1E,$FF,$DF,$EF,$93,$F8,$3E,$6F,$C6,$B6,$BF,$1A,$D3,$F1,$A5
	.dc.b	$E6,$FC,$6D,$72,$8F,$C4,$12,$B9,$B6,$B9,$BF,$0A,$FE,$2F,$B1,$F1
	.dc.b	$BE,$0F,$E3,$FC,$3F,$C8,$F8,$3F,$93,$F0,$FF,$2B,$E1,$FE,$5F,$37
	.dc.b	$CC,$F8,$5F,$9B,$F0,$FF,$17,$4F,$F2,$6F,$EF,$1F,$E0,$BD,$FF,$CB
	.dc.b	$AB,$E1,$3E,$3C,$FC,$FE,$4F,$F3,$BC,$BE,$12,$9F,$DB,$F9,$3F,$D0
	.dc.b	$FC,$5E,$DC,$7D,$64,$E3,$F5,$B7,$1F,$8A,$7B,$CA,$E6,$BE,$3C,$7C
	.dc.b	$9F,$07,$F5,$F3,$96,$D7,$42,$4F,$26,$D7,$92,$3F,$11,$C9,$F2,$5F
	.dc.b	$C5,$71,$FC,$97,$F1,$9C,$5F,$25,$FC,$77,$0F,$C9,$7F,$21,$C1,$F2
	.dc.b	$5F,$F2,$BE,$DF,$27,$2F,$F4,$FB,$F2,$63,$F2,$DC,$FF,$8F,$7E,$7E
	.dc.b	$FF,$21,$FC,$7E,$9F,$C4,$3C,$0F,$D3,$47,$EC,$CF,$DB,$10,$C1,$37
	.dc.b	$FD,$B3,$96,$B9,$C7,$F7,$0E,$5A,$E4,$FD,$D1,$92,$99,$AF,$23,$FD
	.dc.b	$73,$C0,$FD,$BB,$FD,$7B,$BF,$F6,$F5,$FB,$8F,$86,$FD,$CF,$C3,$7E
	.dc.b	$1B,$E1,$BF,$69,$7F,$B7,$F0,$7F,$13,$F0,$DF,$B4,$8F,$DD,$78,$1F
	.dc.b	$B5,$7F,$27,$E7,$FE,$CD,$FE,$03,$E0,$FE,$37,$8F,$FA,$CE,$E5,$E7
	.dc.b	$FE,$50,$95,$F9,$C6,$BF,$59,$BC,$3F,$9E,$6E,$FF,$61,$B3,$FE,$48
	.dc.b	$AE,$AF,$32,$1C,$CB,$5C,$DB,$5C,$DF,$D6,$FB,$5F,$A8,$F8,$6F,$D4
	.dc.b	$FA,$3F,$AA,$F8,$6F,$D5,$F9,$9F,$AC,$F8,$6F,$D6,$FB,$1F,$AE,$F8
	.dc.b	$6F,$D7,$FA,$9F,$B0,$F6,$7F,$63,$F0,$5F,$B2,$F6,$BF,$67,$F0,$DF
	.dc.b	$B4,$F5,$7F,$DD,$7F,$17,$FE,$37,$E2,$FF,$78,$BF,$E3,$41,$F0,$3F
	.dc.b	$B7,$7F,$C4,$71,$AF,$80,$E2,$4E,$4F,$F6,$8C,$A3,$11,$FE,$B6,$8C
	.dc.b	$3A,$8C,$F2,$FF,$B9,$A7,$33,$B9,$CD,$77,$8E,$04,$34,$5A,$E5,$3C
	.dc.b	$23,$FC,$F7,$5D,$27,$B7,$C0,$FC,$D9,$FE,$37,$C9,$F3,$5F,$4F,$C3
	.dc.b	$FC,$D7,$F8,$FF,$17,$CD,$7F,$90,$F0,$FC,$D7,$DF,$F0,$5F,$2B,$E1
	.dc.b	$FE,$5F,$C1,$FC,$D9,$F8,$7F,$0F,$F3,$B8,$24,$7E,$EB,$6F,$96,$F9
	.dc.b	$B7,$F4,$77,$27,$D1,$C3,$F0,$6B,$DE,$4E,$1F,$7B,$F0,$63,$51,$88
	.dc.b	$D4,$60,$8F,$73,$81,$4D,$96,$39,$1F,$21,$FE,$83,$D7,$F9,$47,$FA
	.dc.b	$1F,$5F,$E2,$B8,$7E,$44,$7F,$1E,$E1,$F9,$0F,$F2,$0E,$1F,$90,$FE
	.dc.b	$0F,$87,$E4,$3F,$88,$E0,$F9,$0F,$E1,$F8,$3E,$3B,$F8,$9E,$0F,$DE
	.dc.b	$76,$7E,$C3,$83,$F7,$9D,$D7,$BB,$F6,$37,$FA,$F2,$70,$7E,$BD,$BC
	.dc.b	$3F,$AF,$69,$FB,$9A,$E4,$47,$F6,$A6,$4B,$7E,$07,$F6,$E7,$E8,$3E
	.dc.b	$13,$F5,$7F,$0F,$FA,$CF,$85,$FD,$BC,$FE,$93,$DC,$FD,$7F,$C2,$FE
	.dc.b	$C3,$E1,$7F,$63,$F0,$BF,$B2,$F8,$5F,$DB,$DF,$F6,$EF,$07,$F6,$EF
	.dc.b	$EB,$3D,$1F,$DB,$BF,$B8,$F8,$58,$FE,$2F,$D2,$FD,$DF,$C2,$FE,$F3
	.dc.b	$D3,$93,$FB,$46,$BF,$B3,$6F,$27,$EC,$5B,$C9,$FC,$09,$B3,$FA,$A6
	.dc.b	$F1,$FE,$55,$BE,$01,$A6,$30,$FE,$1B,$8C,$B0,$C5,$8E,$31,$FF,$15
	.dc.b	$4E,$A7,$2F,$FB,$D2,$1E,$9F,$28,$E5,$5C,$67,$FC,$7F,$18,$38,$7A
	.dc.b	$7B,$BE,$73,$FE,$FE,$F5,$F1,$7C,$7E,$16,$E1,$DF,$D4,$B6,$1F,$0B
	.dc.b	$85,$F0,$B7,$D5,$C7,$B3,$9F,$80,$D7,$AD,$32,$FA,$C7,$F2,$6E,$2F
	.dc.b	$8E,$FD,$52,$1F,$54,$BB,$FD,$56,$ED,$F5,$5A,$7D,$52,$E7,$F5,$5B
	.dc.b	$C9,$F5,$5B,$C9,$F5,$5F,$E6,$FC,$9F,$55,$F9,$15,$F4,$9A,$FD,$07
	.dc.b	$EA,$5F,$CF,$E3,$F8,$47,$C3,$E3,$F8,$9F,$01,$F1,$78,$FE,$37,$BF
	.dc.b	$F1,$F8,$FE,$47,$C2,$FC,$9F,$80,$F9,$5F,$0B,$F1,$74,$EF,$C7,$FD
	.dc.b	$3F,$7F,$CD,$C7,$A0,$87,$C5,$27,$17,$C5,$6E,$FF,$15,$A7,$C5,$2F
	.dc.b	$07,$C5,$6F,$17,$C5,$6F,$17,$C5,$7F,$AC,$61,$F1,$70,$FC,$4D,$FF
	.dc.b	$A3,$DB,$D3,$F0,$AB,$E2,$9F,$90,$2C,$FE,$3D,$AF,$FA,$19,$BC,$1F
	.dc.b	$19,$CE,$FF,$19,$DE,$DF,$19,$E3,$AF,$C7,$F8,$77,$B9,$FA,$1F,$82
	.dc.b	$FC,$A7,$C3,$7E,$57,$E1,$A6,$1E,$06,$BF,$71,$BB,$7B,$1B,$C5,$FA
	.dc.b	$B6,$CF,$EC,$5B,$B7,$E8,$DA,$9A,$62,$FF,$4D,$87,$E9,$FE,$13,$F5
	.dc.b	$7E,$77,$F9,$87,$0F,$D6,$3F,$E6,$17,$F2,$2F,$FA,$1D,$F9,$2B,$E0
	.dc.b	$F8,$FC,$1F,$EC,$7C,$2E,$7F,$07,$FB,$3F,$0F,$AF,$A3,$FB,$3C,$7E
	.dc.b	$7F,$87,$FB,$37,$9F,$A1,$E2,$3F,$D2,$59,$BD,$9F,$2F,$B9,$EA,$AF
	.dc.b	$90,$43,$E4,$17,$93,$E4,$37,$93,$E4,$34,$F9,$05,$E4,$F9,$0D,$7E
	.dc.b	$43,$79,$3E,$6E,$2C,$3E,$3E,$32,$BE,$1F,$B5,$F2,$B9,$BE,$5F,$C1
	.dc.b	$C9,$F8,$FE,$D7,$CE,$F8,$3F,$9F,$ED,$FD,$0F,$83,$95,$F3,$3E,$0F
	.dc.b	$E9,$FA,$3F,$53,$E0,$FE,$AF,$C1,$CA,$FF,$47,$F0,$7C,$57,$E9,$F0
	.dc.b	$7C,$57,$E1,$93,$93,$E1,$B7,$93,$E1,$B4,$F8,$65,$E4,$F8,$6D,$7E
	.dc.b	$1B,$79,$3B,$57,$43,$F9,$CE,$1F,$16,$7F,$0B,$C9,$F1,$5F,$C4,$F2
	.dc.b	$7F,$38,$FE,$01,$FF,$2E,$47,$E3,$B9,$3E,$33,$FC,$7A,$3F,$CB,$AF
	.dc.b	$E3,$B8,$F8,$7C,$27,$C6,$8F,$C9,$F1,$FE,$63,$8F,$E3,$C7,$E5,$38
	.dc.b	$BE,$43,$4F,$8F,$9F,$E2,$9F,$B8,$43,$D8,$5D,$7D,$9C,$1F,$19,$B1
	.dc.b	$F1,$9B,$C1,$F1,$9A,$FC,$66,$F0,$77,$9F,$D4,$DF,$C3,$9F,$8F,$BF
	.dc.b	$C3,$9F,$83,$1F,$2B,$7F,$8D,$C1,$F8,$79,$FE,$2F,$C1,$FE,$54,$CB
	.dc.b	$FD,$03,$B7,$C3,$C3,$E8,$ED,$F0,$AF,$F6,$FB,$7C,$26,$9F,$ED,$6C
	.dc.b	$3F,$CB,$2F,$F6,$95,$F0,$5F,$57,$E0,$A4,$F8,$1A,$FD,$56,$E1,$FD
	.dc.b	$E2,$91,$8E,$0F,$F0,$44,$62,$B9,$10,$E4,$5A,$E5,$5C,$63,$C8,$DC
	.dc.b	$D5,$CE,$7F,$0D,$87,$E2,$FE,$0F,$F1,$9F,$09,$F8,$DF,$84,$FD,$95
	.dc.b	$7C,$8F,$84,$FC,$87,$C1,$FF,$1E,$EC,$93,$FC,$6B,$B3,$F2,$9F,$07
	.dc.b	$F9,$5F,$83,$FC,$B7,$C1,$FE,$5F,$E0,$FF,$31,$F0,$7F,$B1,$C3,$F9
	.dc.b	$1F,$81,$FE,$1F,$F3,$BF,$9F,$C7,$CB,$F8,$3F,$CF,$7C,$1F,$E7,$FE
	.dc.b	$0E,$4F,$DA,$6B,$F9,$F6,$F4,$7C,$B6,$F4,$FC,$A6,$CF,$D7,$6F,$47
	.dc.b	$C4,$6F,$41,$A6,$30,$FD,$86,$3F,$B1,$F8,$3F,$D9,$7B,$1F,$B3,$F8
	.dc.b	$1F,$DA,$7B,$5F,$E2,$CF,$E0,$1F,$ED,$5A,$FF,$90,$5B,$AF,$F8,$BD
	.dc.b	$FF,$08,$3F,$E1,$47,$FC,$7B,$B7,$F8,$E9,$B3,$C7,$E0,$7A,$7A,$5D
	.dc.b	$7D,$BE,$DE,$97,$7F,$6E,$65,$F9,$5D,$FE,$4B,$F1,$C8,$7C,$72,$E7
	.dc.b	$F1,$DB,$97,$C7,$69,$F1,$CB,$C7,$F1,$DB,$C7,$F1,$DB,$C7,$F2,$1C
	.dc.b	$F1,$FC,$87,$F6,$38,$FC,$5F,$84,$F9,$11,$FB,$4E,$1F,$92,$FF,$83
	.dc.b	$38,$5E,$9F,$CB,$F8,$4F,$99,$F0,$92,$B4,$FC,$47,$87,$BB,$F1,$1E
	.dc.b	$5F,$09,$F4,$CF,$C7,$7A,$DF,$AC,$D3,$E2,$CF,$C3,$27,$17,$C3,$6F
	.dc.b	$17,$C3,$69,$F0,$CB,$C5,$F0,$DB,$C5,$F0,$DB,$C5,$F0,$DA,$7C,$43
	.dc.b	$E1,$90,$FC,$17,$0F,$C4,$7F,$9D,$70,$FC,$47,$F0,$DC,$1F,$0D,$F8
	.dc.b	$24,$3F,$15,$BF,$C3,$7F,$19,$BF,$C3,$7F,$1D,$BF,$C3,$7F,$D2,$FD
	.dc.b	$7F,$B0,$D1,$C4,$F8,$4F,$CD,$7C,$27,$E6,$FE,$06,$61,$F5,$5A,$FE
	.dc.b	$65,$BC,$5F,$99,$6F,$17,$E8,$5B,$3F,$A1,$6E,$3E,$06,$E2,$69,$8C
	.dc.b	$3F,$59,$8F,$EC,$6F,$F9,$DF,$6F,$EC,$9F,$C4,$7C,$27,$EC,$72,$91
	.dc.b	$F8,$BF,$83,$FD,$A7,$C1,$7E,$D7,$E0,$FF,$6D,$E0,$7E,$DF,$D9,$FD
	.dc.b	$65,$FE,$47,$E0,$FF,$56,$FE,$D7,$BF,$F1,$2F,$FA,$9F,$F8,$07,$FC
	.dc.b	$7D,$C4,$7F,$C7,$DC,$7F,$EB,$27,$FC,$7D,$3F,$E9,$06,$C6,$7E,$1F
	.dc.b	$FD,$63,$18,$DB,$FD,$8C,$DA,$ED,$3A,$23,$FE,$70,$8D,$5F,$86,$73
	.dc.b	$3D,$DB,$7D,$E3,$84,$62,$FB,$B6,$3D,$9F,$13,$E4,$F8,$BF,$2F,$C3
	.dc.b	$F9,$BE,$1F,$CF,$F0,$FE,$8F,$87,$F4,$FC,$3C,$C7,$C9,$F0,$7E,$C2
	.dc.b	$3E,$7F,$86,$F7,$FC,$27,$C0,$F2,$7E,$0F,$C2,$7C,$2F,$66,$5F,$E5
	.dc.b	$27,$E1,$ED,$F1,$BE,$1B,$E3,$FB,$52,$BE,$13,$5F,$9C,$DE,$4F,$A4
	.dc.b	$DE,$4F,$9A,$D9,$FB,$4D,$E5,$FB,$4D,$E6,$34,$C6,$1F,$16,$BE,$97
	.dc.b	$C2,$7D,$AF,$84,$F8,$B1,$F5,$78,$FE,$EF,$C3,$FD,$EF,$85,$F8,$73
	.dc.b	$F8,$3E,$4B,$B7,$E1,$F9,$3F,$07,$EC,$FC,$5A,$FB,$DC,$9F,$11,$FC
	.dc.b	$57,$27,$E4,$B8,$7F,$21,$C7,$F1,$67,$E1,$90,$F8,$65,$C7,$E1,$B7
	.dc.b	$8F,$E1,$B4,$F8,$65,$AF,$86,$DE,$3F,$86,$DE,$3F,$88,$E7,$8F,$E2
	.dc.b	$3F,$9F,$E3,$F8,$8F,$E5,$F9,$FF,$10,$FE,$67,$8B,$E2,$3F,$88,$21
	.dc.b	$F1,$4F,$C4,$1F,$9F,$DF,$E2,$3F,$AB,$DF,$E2,$6B,$FA,$FD,$FE,$27
	.dc.b	$57,$F9,$93,$6F,$DA,$3F,$9D,$7F,$4D,$7F,$A2,$21,$FA,$72,$D7,$E9
	.dc.b	$DC,$11,$C1,$37,$FD,$4C,$70,$6B,$B6,$C6,$E3,$F7,$67,$EE,$68,$FF
	.dc.b	$BA,$9F,$46,$FF,$EC,$38,$FD,$F7,$4E,$95,$2F,$8E,$0C,$3F,$CF,$F5
	.dc.b	$A9,$CC,$7F,$97,$F0,$74,$FE,$DE,$B4,$2F,$97,$C1,$F7,$C7,$B4,$E7
	.dc.b	$0F,$F2,$EB,$E7,$F8,$5D,$65,$DB,$E1,$3F,$7B,$3E,$BF,$84,$CB,$C7
	.dc.b	$E1,$2E,$7E,$7F,$84,$B9,$1F,$B3,$F0,$98,$F8,$4C,$FC,$26,$BE,$12
	.dc.b	$65,$F1,$DA,$FA,$B1,$F8,$51,$8D,$FD,$31,$89,$E0,$43,$86,$4A,$E1
	.dc.b	$B7,$C8,$6C,$FC,$16,$D7,$38,$E3,$1E,$67,$ED,$63,$EC,$CB,$DB,$F0
	.dc.b	$9E,$EF,$77,$DF,$EF,$FC,$0F,$77,$E0,$FC,$27,$C2,$F4,$FE,$1F,$C2
	.dc.b	$4C,$FF,$08,$FC,$6F,$7B,$E4,$FF,$00,$7F,$92,$B7,$34,$FF,$25,$F1
	.dc.b	$7C,$07,$FC,$9B,$ED,$FF,$7D,$E3,$70,$72,$F7,$B9,$FA,$9D,$3E,$16
	.dc.b	$6B,$F9,$4C,$B9,$46,$23,$E7,$64,$8C,$4E,$A7,$87,$37,$E5,$2B,$51
	.dc.b	$83,$F1,$CE,$1E,$75,$CA,$78,$36,$36,$DD,$3D,$ED,$A6,$1E,$BA,$FD
	.dc.b	$56,$5F,$1B,$E1,$7E,$3F,$C2,$FC,$8F,$57,$E8,$57,$8F,$2F,$93,$F0
	.dc.b	$7F,$46,$3E,$79,$0F,$C2,$F1,$7D,$0F,$83,$FA,$3B,$49,$E5,$AB,$E1
	.dc.b	$7E,$A9,$F9,$A8,$CF,$0F,$CE,$77,$C1,$E9,$23,$82,$1F,$10,$BC,$3F
	.dc.b	$35,$C1,$1C,$13,$0F,$93,$1C,$1A,$FD,$8E,$1F,$8C,$FE,$83,$87,$E3
	.dc.b	$3E,$AE,$0F,$8C,$FF,$13,$E7,$FC,$7E,$FF,$16,$3E,$19,$F1,$71,$FD
	.dc.b	$6F,$1B,$83,$E3,$4F,$C4,$CF,$E3,$3F,$88,$DB,$E2,$BF,$DF,$4F,$C1
	.dc.b	$FC,$27,$EF,$DF,$E4,$5D,$DF,$AE,$C3,$F6,$91,$FB,$7C,$7B,$F7,$64
	.dc.b	$91,$9C,$7F,$70,$D3,$F7,$05,$E4,$FD,$C5,$72,$8C,$5F,$67,$F8,$1E
	.dc.b	$1E,$37,$F3,$F8,$FE,$E9,$FE,$6D,$D9,$FB,$A7,$F0,$3F,$09,$FB,$A7
	.dc.b	$F9,$D7,$67,$ED,$BE,$07,$F6,$FF,$05,$FB,$8F,$83,$FD,$D5,$7E,$23
	.dc.b	$E0,$FF,$77,$F0,$5F,$BC,$F8,$39,$3F,$17,$83,$F7,$BF,$07,$FC,$03
	.dc.b	$E0,$FF,$80,$FC,$1F,$F2,$AF,$E0,$1F,$F2,$89,$3D,$BC,$13,$0F,$CF
	.dc.b	$46,$5B,$19,$7F,$94,$46,$5B,$19,$E5,$FE,$53,$39,$6C,$67,$9B,$F7
	.dc.b	$D7,$E5,$BF,$D1,$E3,$E5,$E6,$F3,$7C,$27,$9F,$E1,$3D,$1E,$AF,$A7
	.dc.b	$DA,$F5,$7C,$0F,$AF,$D6,$99,$7A,$3E,$03,$DD,$F0,$12,$3D,$7F,$09
	.dc.b	$F0,$7E,$12,$47,$C7,$E3,$7B,$BE,$38,$F6,$FB,$BF,$1F,$E0,$64,$7C
	.dc.b	$46,$BF,$09,$BC,$DF,$01,$BC,$BF,$DB,$B0,$46,$2B,$FB,$44,$61,$E8
	.dc.b	$43,$A1,$6B,$A3,$6B,$A5,$F9,$E7,$E7,$E1,$F2,$6B,$EC,$F1,$7D,$FF
	.dc.b	$82,$FC,$07,$C1,$7E,$07,$E1,$3F,$05,$F0,$7F,$22,$FE,$DF,$15,$CB
	.dc.b	$EC,$F0,$FC,$99,$FB,$EF,$C8,$7F,$03,$C1,$FE,$4B,$DF,$E4,$C7,$F1
	.dc.b	$CE,$BF,$C7,$6F,$EF,$EC,$57,$BE,$BD,$A4,$C7,$DA,$DE,$1F,$91,$C5
	.dc.b	$C6,$B8,$56,$89,$3A,$AF,$88,$7C,$CA,$E0,$71,$C3,$E3,$4F,$D3,$E2
	.dc.b	$F8,$D5,$F9,$AE,$2F,$8C,$FE,$83,$87,$E3,$3F,$A2,$E0,$F8,$97,$F6
	.dc.b	$F8,$2C,$FD,$AE,$0F,$8D,$A7,$C8,$8F,$D4,$F0,$7C,$7F,$5B,$FD,$CD
	.dc.b	$E9,$BB,$3F,$65,$F0,$9F,$B4,$BF,$D6,$90,$FD,$59,$70,$FD,$4B,$73
	.dc.b	$FC,$67,$7F,$EA,$E3,$F1,$6D,$E4,$F9,$ED,$E4,$FD,$95,$70,$AE,$47
	.dc.b	$03,$FA,$9F,$7F,$91,$FD,$99,$FC,$0B,$E1,$3F,$C6,$3F,$C0,$3F,$E6
	.dc.b	$5B,$FF,$27,$4F,$F0,$2E,$FF,$DA,$3F,$C2,$3B,$B7,$F1,$3C,$3E,$2B
	.dc.b	$F8,$87,$81,$CB,$E1,$B9,$FC,$57,$F1,$3F,$0A,$63,$FB,$63,$F8,$D6
	.dc.b	$8F,$87,$F0,$FC,$3F,$8B,$E1,$FC,$7F,$0F,$2F,$FB,$A5,$A3,$11,$FE
	.dc.b	$3D,$8C,$3A,$8C,$F3,$7F,$8F,$A7,$33,$B9,$CF,$27,$F9,$9A,$73,$5B
	.dc.b	$9C,$D7,$F9,$86,$73,$5B,$9C,$F3,$27,$35,$BC,$3D,$6F,$A3,$2F,$9B
	.dc.b	$F0,$FF,$3B,$DC,$F9,$FF,$0F,$F4,$3D,$7F,$A3,$EC,$7D,$2F,$87,$FA
	.dc.b	$7E,$C4,$CB,$E0,$FC,$3F,$D6,$23,$F1,$3D,$EF,$B3,$EC,$C8,$F9,$BA
	.dc.b	$FC,$B7,$E4,$ED,$E8,$DF,$CF,$C1,$E9,$E1,$C1,$2B,$E5,$10,$F9,$05
	.dc.b	$E3,$F8,$CD,$E3,$F8,$8D,$3E,$44,$F7,$C3,$C5,$1F,$5F,$1F,$86,$FD
	.dc.b	$CF,$87,$F8,$6F,$DE,$F7,$BF,$27,$F0,$FF,$E8,$4F,$5F,$FC,$B9,$AB
	.dc.b	$87,$E2,$5F,$E0,$78,$7E,$1B,$4F,$F1,$D5,$7F,$98,$35,$FF,$51,$AF
	.dc.b	$84,$E8,$7F,$B9,$F0,$FC,$95,$7C,$62,$1F,$00,$9B,$FC,$06,$DF,$C0
	.dc.b	$69,$F0,$0F,$85,$C1,$F3,$5B,$5F,$40,$9C,$1F,$3E,$38,$CE,$DF,$CE
	.dc.b	$70,$7C,$8C,$3F,$35,$C1,$F2,$27,$F4,$1C,$1F,$21,$FD,$17,$07,$C8
	.dc.b	$7E,$0E,$7F,$1A,$FF,$45,$BF,$C6,$7F,$49,$BD,$9F,$8B,$BF,$C4,$BF
	.dc.b	$93,$BF,$FB,$37,$C9,$79,$5F,$2B,$E0,$FF,$3B,$5F,$EA,$DB,$FF,$3C
	.dc.b	$65,$FE,$AC,$5F,$0F,$F6,$49,$3F,$51,$AF,$16,$CF,$16,$E5,$F6,$1B
	.dc.b	$3F,$ED,$3A,$EA,$F3,$21,$CC,$B5,$CD,$FC,$1F,$C2,$FA,$23,$CD,$87
	.dc.b	$E3,$7E,$17,$F6,$95,$F8,$8F,$87,$FD,$A3,$FC,$43,$AF,$F6,$8F,$E3
	.dc.b	$3E,$1B,$F2,$9F,$0F,$F9,$5F,$87,$FC,$B7,$C3,$FE,$5F,$E1,$FF,$95
	.dc.b	$F7,$FE,$CF,$0F,$E2,$DD,$FF,$2F,$8F,$F2,$7F,$0F,$F9,$DF,$87,$FC
	.dc.b	$F7,$C3,$FE,$7F,$E1,$FF,$41,$F0,$F3,$1F,$CD,$B5,$FA,$0D,$E4,$FA
	.dc.b	$0D,$E8,$FC,$FB,$67,$E4,$37,$C1,$74,$23,$0D,$9E,$E7,$EA,$B0,$F2
	.dc.b	$BF,$AC,$CB,$F6,$99,$FE,$D7,$D8,$FD,$B7,$C1,$7E,$DF,$D7,$FF,$7D
	.dc.b	$FF,$00,$BD,$9B,$F1,$18,$F8,$1C,$FC,$44,$D3,$FD,$61,$6F,$6B,$89
	.dc.b	$23,$3F,$11,$D0,$F9,$4F,$E0,$F6,$F9,$4F,$F0,$0E,$5F,$94,$FA,$8F
	.dc.b	$8E,$5E,$5F,$8E,$DE,$5F,$8E,$D3,$E3,$97,$97,$E3,$F2,$FC,$62,$46
	.dc.b	$79,$7E,$43,$FA,$3E,$5F,$8E,$FE,$97,$2F,$2A,$C5,$8A,$18,$B5,$7B
	.dc.b	$7F,$32,$7E,$2F,$C0,$FC,$CE,$3F,$9B,$ED,$7C,$B9,$F9,$3C,$5F,$2D
	.dc.b	$ED,$97,$5C,$BA,$71,$7C,$B9,$C9,$E0,$CF,$E8,$3F,$4B,$E0,$7E,$C9
	.dc.b	$F3,$AB,$F0,$FB,$7A,$B6,$FD,$F7,$17,$CE,$BF,$9A,$48,$F9,$AD,$CB
	.dc.b	$E6,$B4,$F9,$A5,$BF,$9A,$FC,$BE,$2F,$9A,$E7,$8B,$E6,$DF,$43,$F8
	.dc.b	$77,$17,$CE,$7F,$88,$65,$E5,$C7,$E0,$69,$F2,$B0,$FE,$87,$CF,$F9
	.dc.b	$4E,$2F,$95,$1F,$90,$E1,$F9,$4F,$C4,$D7,$E9,$70,$FC,$BC,$FE,$BC
	.dc.b	$FC,$7E,$1F,$95,$18,$AF,$9B,$3F,$24,$83,$C8,$FE,$52,$FF,$19,$EA
	.dc.b	$FD,$2F,$B3,$FA,$6E,$CF,$D3,$FC,$27,$EA,$3E,$16,$5F,$E9,$1A,$F9
	.dc.b	$78,$11,$8C,$FC,$51,$89,$E0,$43,$84,$EF,$87,$E2,$B6,$3F,$04,$DD
	.dc.b	$3F,$24,$F9,$23,$F6,$B8,$FA,$1F,$ED,$B9,$FF,$01,$7F,$73,$F0,$DF
	.dc.b	$C0,$5F,$DD,$FC,$47,$F0,$07,$F7,$BF,$0F,$FC,$13,$E2,$3F,$82,$FC
	.dc.b	$47,$F0,$6F,$87,$FE,$0F,$F0,$FF,$A3,$BF,$A5,$FC,$03,$5F,$81,$E1
	.dc.b	$7F,$01,$7F,$DC,$50,$F8,$7E,$1C,$7F,$4E,$1F,$0F,$8E,$3F,$D1,$43
	.dc.b	$E1,$F1,$87,$EC,$21,$F0,$F8,$E5,$47,$F8,$12,$7C,$7E,$EE,$63,$19
	.dc.b	$77,$8C,$12,$7F,$79,$19,$6E,$3E,$29,$D4,$74,$32,$78,$DC,$CF,$18
	.dc.b	$D1,$1C,$1E,$FC,$3C,$98,$F9,$72,$F3,$67,$E7,$F8,$7F,$47,$BF,$E9
	.dc.b	$F8,$0F,$57,$C3,$FA,$FE,$1E,$69,$E2,$E0,$F8,$3F,$0F,$EF,$3E,$07
	.dc.b	$C3,$C9,$D7,$13,$DA,$F8,$9F,$0F,$27,$EA,$FC,$1F,$8D,$DF,$0F,$C9
	.dc.b	$E2,$FC,$77,$83,$F9,$79,$FC,$A9,$F9,$7E,$4F,$CD,$F8,$3A,$24,$67
	.dc.b	$9F,$F3,$2D,$3F,$32,$5D,$FF,$32,$DD,$7E,$EB,$E2,$3D,$3C,$DF,$6F
	.dc.b	$4F,$8D,$1E,$CE,$5F,$8C,$FB,$B8,$FE,$2B,$F9,$4E,$3F,$88,$FC,$2E
	.dc.b	$3F,$88,$FE,$63,$8E,$ED,$F9,$3E,$2F,$89,$3F,$C0,$FC,$1F,$C9,$E7
	.dc.b	$F1,$63,$F3,$FC,$7F,$15,$F5,$71,$7C,$5C,$3E,$1C,$63,$8B,$E1,$BF
	.dc.b	$20,$91,$A6,$F1,$7C,$66,$9F,$18,$BC,$5F,$19,$FD,$7F,$1F,$E2,$9F
	.dc.b	$8A,$7D,$EC,$BE,$2B,$FB,$4E,$1F,$8A,$FE,$DB,$87,$E2,$E3,$FB,$1E
	.dc.b	$1F,$8B,$3F,$BA,$E1,$F8,$79,$FE,$CB,$86,$CF,$EC,$78,$7E,$2F,$89
	.dc.b	$FC,$9B,$6F,$D8,$F4,$7E,$BF,$E0,$FF,$61,$EC,$7E,$C7,$D8,$FD,$2E
	.dc.b	$7F,$AB,$CB,$F2,$7D,$98,$7F,$54,$60,$87,$EA,$63,$83,$4D,$97,$1F
	.dc.b	$E0,$33,$CD,$B1,$CB,$7F,$E1,$D7,$E4,$8F,$D1,$E3,$FB,$07,$F4,$BF
	.dc.b	$0B,$FB,$07,$F4,$FD,$BF,$C2,$7E,$17,$F8,$57,$C3,$FF,$0F,$FE,$01
	.dc.b	$FE,$9B,$5F,$D4,$5F,$F4,$C3,$FE,$04,$7F,$D3,$0F,$ED,$57,$C2,$71
	.dc.b	$F8,$5F,$E2,$9E,$0F,$3F,$86,$97,$FE,$0D,$E4,$F9,$98,$3E,$1B,$C1
	.dc.b	$F0,$DE,$1F,$86,$F1,$7C,$37,$8F,$E1,$A6,$3D,$DA,$F9,$5B,$C9,$FD
	.dc.b	$87,$14,$63,$9F,$FC,$21,$18,$AE,$44,$39,$16,$B9,$57,$18,$F8,$0F
	.dc.b	$91,$F8,$18,$7A,$1F,$85,$97,$C9,$F8,$6F,$95,$F0,$DF,$2F,$D3,$F9
	.dc.b	$9E,$DF,$CD,$F7,$FE,$77,$BB,$F3,$FD,$FF,$A1,$F0,$DF,$47,$DF,$9A
	.dc.b	$7C,$DF,$86,$FE,$85,$E8,$7F,$96,$63,$F9,$23,$5F,$DD,$1F,$57,$6F
	.dc.b	$23,$F6,$77,$E3,$F0,$DC,$BE,$1A,$4F,$26,$BF,$E0,$FB,$46,$38,$FF
	.dc.b	$C1,$91,$89,$E0,$43,$84,$EF,$97,$FC,$7D,$3B,$9C,$9B,$AE,$75,$B8
	.dc.b	$FF,$27,$CA,$FF,$05,$1F,$8D,$E7,$FC,$A6,$5F,$36,$3E,$0F,$C0,$7C
	.dc.b	$4F,$80,$F8,$BE,$BF,$CD,$9F,$CA,$71,$FC,$D7,$F2,$DC,$5F,$31,$FC
	.dc.b	$E7,$0F,$E7,$F8,$7E,$6F,$C3,$49,$FC,$B7,$AD,$F8,$8D,$3E,$5F,$0F
	.dc.b	$E3,$78,$FE,$5D,$7A,$F8,$FD,$FC,$7E,$BF,$23,$FB,$FB,$79,$3E,$16
	.dc.b	$1F,$02,$F8,$1F,$02,$74,$47,$35,$C9,$B5,$F1,$5F,$C8,$9F,$14,$FC
	.dc.b	$89,$FB,$3E,$AF,$C3,$F1,$FC,$58,$FC,$3F,$17,$C5,$7F,$79,$C5,$F1
	.dc.b	$5F,$DF,$71,$7C,$5C,$FF,$55,$C5,$F1,$67,$F1,$FC,$3F,$15,$FC,$7E
	.dc.b	$9F,$E1,$CF,$83,$FC,$AF,$C1,$FE,$5B,$E1,$7F,$4D,$5F,$AC,$26,$3F
	.dc.b	$AC,$7F,$5D,$C3,$FB,$4E,$EF,$D7,$62,$EF,$FD,$B5,$70,$AD,$3C,$0F
	.dc.b	$D9,$96,$78,$57,$2A,$E3,$7D,$EB,$F8,$D7,$81,$E5,$7F,$53,$97,$EB
	.dc.b	$33,$FD,$6F,$C2,$7E,$E2,$7F,$8E,$78,$1F,$B8,$7F,$5F,$D3,$FB,$87
	.dc.b	$F6,$3F,$09,$FB,$87,$F6,$7E,$CF,$ED,$BE,$13,$F6,$FF,$09,$FB,$8F
	.dc.b	$86,$95,$F8,$C3,$F3,$FF,$0B,$FB,$CF,$85,$FD,$EF,$C2,$FE,$FB,$E1
	.dc.b	$7F,$7F,$F0,$BF,$E4,$1F,$E0,$1F,$F2,$87,$27,$F9,$F5,$BD,$1F,$EA
	.dc.b	$B6,$E5,$FE,$B8,$6F,$41,$A4,$8E,$84,$3F,$C5,$CE,$5A,$64,$FF,$49
	.dc.b	$39,$7F,$83,$5F,$9D,$FE,$27,$8F,$AD,$FE,$2B,$9F,$A7,$DA,$F5,$78
	.dc.b	$5E,$BF,$84,$F6,$7C,$1F,$B7,$DB,$F7,$7C,$17,$BF,$D9,$F8,$1F,$0D
	.dc.b	$F0,$7E,$0E,$69,$E6,$F8,$3F,$88,$7F,$B0,$F8,$1F,$07,$F1,$FE,$0F
	.dc.b	$E4,$7B,$5F,$27,$E0,$FE,$5E,$1E,$C2,$1E,$C3,$DD,$CF,$F3,$5B,$CF
	.dc.b	$F3,$5A,$7C,$D2,$F3,$FC,$D6,$E5,$F1,$6B,$94,$7F,$36,$E7,$F1,$3F
	.dc.b	$3B,$1F,$33,$F4,$33,$F8,$CF,$D2,$F8,$4F,$C0,$7C,$17,$C6,$8F,$AB
	.dc.b	$C7,$F8,$3F,$84,$F8,$D1,$F6,$3E,$13,$E2,$BF,$5B,$8B,$F1,$1F,$09
	.dc.b	$23,$F2,$1B,$FE,$0F,$3F,$0C,$7D,$CD,$3C,$AF,$DC,$3E,$2E,$DE,$A7
	.dc.b	$E3,$18,$E0,$F8,$8D,$3E,$21,$6F,$E2,$37,$83,$E2,$4F,$08,$E6,$4A
	.dc.b	$FE,$9B,$C1,$E2,$7F,$3D,$87,$99,$FD,$06,$3F,$1A,$BF,$27,$C1,$F1
	.dc.b	$9F,$CC,$F3,$FE,$2D,$FC,$DE,$FF,$19,$FC,$CE,$FF,$1B,$0F,$D5,$6F
	.dc.b	$F1,$67,$F3,$BB,$FF,$AF,$FD,$FF,$F2,$07,$53,$B3,$F3,$95,$FD,$97
	.dc.b	$B7,$F4,$0F,$E9,$3E,$13,$F6,$DF,$09,$23,$F5,$6D,$7F,$52,$DC,$FF
	.dc.b	$74,$DE,$3F,$DD,$36,$7F,$74,$DE,$5F,$DD,$37,$94,$FF,$05,$C3,$C7
	.dc.b	$87,$F7,$6C,$7C,$EF,$F7,$8C,$BF,$D9,$5F,$B6,$7F,$8F,$23,$F8,$56
	.dc.b	$9F,$B3,$CF,$FD,$5D,$3F,$E3,$EC,$CF,$F8,$FA,$3F,$80,$F7,$72,$F8
	.dc.b	$5F,$DA,$C7,$0F,$85,$EB,$F0,$BF,$B4,$8F,$DB,$1F,$90,$BF,$F1,$F3
	.dc.b	$DB,$E0,$3C,$7E,$EF,$93,$D2,$F2,$FA,$9E,$DE,$29,$7F,$28,$87,$CA
	.dc.b	$2E,$FF,$29,$BC,$9F,$29,$A7,$CA,$2E,$FF,$29,$F7,$DF,$85,$F8,$38
	.dc.b	$FC,$6C,$BE,$3F,$BD,$F2,$3D,$DF,$93,$F0,$BF,$2B,$D3,$F9,$7F,$0B
	.dc.b	$F3,$3D,$DF,$9B,$F0,$BF,$3B,$DE,$F9,$FF,$0B,$F4,$3D,$FF,$A3,$F0
	.dc.b	$7F,$EF,$BF,$E0,$1F,$F1,$C7,$23,$91,$C4,$E4,$57,$F8,$F7,$C1,$FF
	.dc.b	$49,$3F,$E0,$19,$5F,$17,$87,$43,$93,$B7,$C2,$7B,$67,$79,$76,$CB
	.dc.b	$AF,$C2,$7C,$E9,$F9,$84,$C7,$E6,$36,$BE,$63,$4F,$98,$5A,$F9,$8D
	.dc.b	$F0,$FF,$2A,$FD,$06,$BE,$23,$DD,$9B,$0F,$36,$1E,$BB,$F8,$38,$FD
	.dc.b	$3B,$F8,$B3,$9B,$F8,$4D,$F8,$5F,$99,$F0,$BF,$56,$BF,$37,$CB,$F5
	.dc.b	$5F,$CE,$F2,$7D,$57,$F3,$FC,$7F,$55,$FD,$0F,$1F,$D5,$F8,$59,$1F
	.dc.b	$CA,$33,$EF,$C1,$F2,$78,$BE,$74,$F9,$23,$E6,$4F,$F8,$F9,$D7,$77
	.dc.b	$DF,$E1,$F9,$DB,$FC,$C2,$4F,$CC,$6D,$FC,$C6,$F0,$FC,$C6,$F1,$7C
	.dc.b	$A6,$F7,$74,$BF,$0E,$1F,$29,$F1,$F1,$FC,$9C,$7E,$86,$3F,$8C,$F8
	.dc.b	$6F,$C9,$72,$7C,$89,$FE,$03,$D5,$F9,$C7,$F8,$1F,$57,$E7,$1F,$E0
	.dc.b	$BD,$5F,$9C,$7E,$39,$0E,$BC,$0E,$2F,$93,$1F,$E4,$6E,$2F,$E0,$2E
	.dc.b	$BE,$0F,$F8,$0B,$F2,$4F,$C1,$FC,$27,$F0,$27,$1A,$FE,$8B,$E1,$BF
	.dc.b	$50,$48,$CE,$28,$C4,$78,$9B,$C7,$F8,$76,$F0,$FD,$66,$CF,$E4,$9B
	.dc.b	$7D,$E3,$AC,$78,$DF,$6E,$1E,$87,$D4,$7F,$28,$F0,$3F,$83,$3F,$CA
	.dc.b	$BB,$FF,$7D,$F0,$BF,$BF,$F8,$5F,$E0,$1F,$0B,$FC,$07,$E1,$7F,$81
	.dc.b	$7C,$2F,$F0,$3F,$85,$FE,$09,$E7,$7F,$06,$CB,$DB,$F0,$BF,$C1,$CF
	.dc.b	$EA,$98,$FE,$3B,$E1,$7F,$85,$7A,$3F,$C2,$FE,$17,$F8,$67,$B9,$FC
	.dc.b	$2B,$0F,$BF,$DF,$FE,$42,$1F,$0B,$9E,$7F,$D7,$46,$27,$E7,$B5,$E3
	.dc.b	$3B,$E7,$FE,$11,$39,$AE,$33,$9A,$FE,$31,$39,$AD,$CE,$79,$FF,$D9
	.dc.b	$6D,$9D,$57,$08,$C1,$1F,$1B,$5F,$59,$EE,$3F,$03,$8F,$C2,$7F,$07
	.dc.b	$9F,$5F,$53,$B7,$C2,$F7,$F6,$FC,$1E,$FF,$87,$D7,$F1,$7C,$2F,$8F
	.dc.b	$E1,$7C,$9E,$E7,$97,$E1,$7C,$DF,$03,$E7,$F8,$59,$B7,$F0,$3D,$BD
	.dc.b	$5F,$07,$F4,$63,$E4,$FC,$2F,$B5,$F7,$7C,$24,$7E,$6B,$5F,$AA,$DE
	.dc.b	$7F,$A6,$DE,$7F,$9E,$D9,$F9,$AD,$E7,$FF,$7A,$B7,$9C,$F8,$67,$C6
	.dc.b	$7C,$63,$EE,$E3,$FD,$47,$2F,$A5,$F0,$BF,$1A,$7F,$03,$CB,$F5,$7E
	.dc.b	$17,$EB,$7C,$2F,$C6,$9F,$C2,$FC,$2F,$C6,$7F,$0F,$C5,$F1,$9F,$C4
	.dc.b	$FC,$2F,$DD,$3F,$2E,$FF,$3E,$CB,$E3,$BE,$33,$F2,$BC,$3E,$77,$E1
	.dc.b	$9F,$18,$86,$09,$C3,$F1,$9C,$B5,$CF,$0F,$C8,$72,$D7,$27,$CA,$32
	.dc.b	$7E,$60,$FC,$A4,$FD,$B7,$F5,$3B,$7C,$77,$F4,$BB,$7C,$77,$F5,$BB
	.dc.b	$7C,$77,$F5,$3A,$FC,$77,$F6,$3A,$FC,$77,$F6,$7A,$FC,$77,$F6,$BA
	.dc.b	$FC,$77,$FC,$EB,$C1,$FC,$F7,$4F,$E9,$7D,$DF,$A9,$FE,$00,$FF,$1C
	.dc.b	$47,$F4,$EE,$FC,$7C,$57,$E9,$78,$7F,$67,$8F,$EA,$08,$7E,$E8,$AE
	.dc.b	$9C,$11,$C1,$2B,$F7,$31,$C1,$AE,$DF,$F0,$55,$FF,$91,$FC,$D6,$1E
	.dc.b	$79,$FD,$A5,$7F,$12,$F5,$FF,$68,$FF,$3C,$BF,$F2,$66,$1F,$C4,$BB
	.dc.b	$FF,$6B,$3F,$C4,$BB,$BE,$07,$C3,$FE,$D6,$38,$7C,$3F,$C3,$F8,$7F
	.dc.b	$89,$E7,$FE,$D2,$7F,$8F,$F7,$7E,$43,$B9,$A7,$F7,$69,$FE,$C2,$BE
	.dc.b	$1B,$E6,$7B,$BF,$37,$E1,$FE,$77,$C3,$FC,$FF,$4E,$61,$E7,$6B,$C9
	.dc.b	$BC,$3F,$E1,$7C,$51,$8E,$3F,$8A,$DA,$FF,$1E,$4E,$9E,$84,$3A,$16
	.dc.b	$BA,$5F,$96,$3E,$46,$3F,$83,$CB,$F0,$9F,$0B,$F8,$5F,$72,$57,$CD
	.dc.b	$F8,$4F,$C4,$7B,$BF,$89,$F8,$4F,$C5,$7B,$DF,$8B,$F8,$4F,$C6,$7C
	.dc.b	$27,$E3,$7E,$1F,$F9,$B7,$F0,$0F,$F9,$53,$8F,$FC,$8C,$FF,$88,$33
	.dc.b	$FF,$2A,$3F,$E5,$1F,$88,$7A,$FA,$F1,$3C,$95,$FD,$E2,$BF,$C7,$8F
	.dc.b	$DF,$E7,$F8,$5D,$1F,$53,$A2,$F4,$7C,$3A,$F8,$24,$E8,$F9,$93,$A2
	.dc.b	$47,$C5,$D3,$E6,$B4,$F9,$B1,$A7,$EF,$E1,$EC,$F8,$6F,$2C,$7F,$88
	.dc.b	$B2,$FF,$1B,$3F,$53,$97,$E4,$B9,$F8,$5F,$87,$F1,$1F,$26,$3D,$9F
	.dc.b	$0D,$F2,$5F,$77,$C2,$FC,$97,$E0,$7B,$BF,$2B,$E1,$24,$7F,$BC,$32
	.dc.b	$EF,$C3,$F9,$4E,$DF,$24,$FC,$53,$D5,$1F,$15,$BC,$5F,$17,$8B,$E1
	.dc.b	$C6,$38,$7D,$D1,$88,$F9,$1C,$5F,$11,$A7,$C4,$38,$1D,$38,$BC,$2F
	.dc.b	$E9,$31,$FC,$0F,$1F,$C2,$BF,$CE,$71,$7E,$13,$8B,$E1,$DF,$EE,$7A
	.dc.b	$BF,$0F,$C5,$F0,$E3,$E6,$71,$7C,$37,$E7,$71,$7C,$27,$E8,$71,$7E
	.dc.b	$D7,$83,$F6,$99,$FE,$A7,$BB,$F6,$75,$FC,$17,$BF,$F6,$6F,$E2,$FE
	.dc.b	$17,$F2,$A4,$70,$43,$EE,$39,$69,$9A,$FF,$46,$72,$FF,$91,$E3,$1C
	.dc.b	$7F,$9E,$AE,$15,$A6,$C7,$F1,$EB,$F1,$D7,$F0,$AC,$3C,$EF,$F0,$CF
	.dc.b	$03,$F5,$3F,$0B,$FA,$AE,$AF,$E7,$1E,$0F,$EC,$AB,$F4,$9F,$0B,$FB
	.dc.b	$27,$F9,$17,$83,$FB,$27,$F5,$1F,$0B,$FB,$2F,$85,$FD,$9F,$C3,$C9
	.dc.b	$FC,$57,$C3,$7E,$DB,$E1,$BF,$6F,$EE,$49,$FD,$87,$A9,$FB,$AF,$85
	.dc.b	$FD,$DF,$A7,$27,$F6,$8D,$7F,$44,$DE,$6F,$8A,$DE,$5F,$E0,$2D,$9F
	.dc.b	$E0,$2D,$F1,$1D,$1F,$D0,$27,$A9,$FC,$07,$0F,$E1,$78,$F9,$63,$F6
	.dc.b	$19,$7A,$5F,$E0,$DA,$7F,$91,$7F,$80,$7F,$CB,$11,$FE,$96,$6E,$9F
	.dc.b	$EC,$37,$FC,$8F,$91,$7C,$37,$1F,$73,$97,$AB,$CF,$E1,$BA,$7A,$DD
	.dc.b	$7E,$1B,$B7,$C3,$4D,$3E,$59,$FE,$8E,$C9,$F0,$DE,$3F,$7B,$C9,$F0
	.dc.b	$DE,$5F,$5F,$E6,$5F,$E2,$0F,$E4,$7E,$27,$E1,$9A,$67,$83,$E2,$C6
	.dc.b	$5B,$19,$E6,$FC,$93,$F9,$62,$71,$7E,$45,$D7,$37,$85,$F3,$63,$E5
	.dc.b	$7D,$19,$FF,$38,$E4,$F9,$51,$0F,$67,$C1,$FC,$DE,$5F,$9D,$F0,$9F
	.dc.b	$3F,$E1,$BE,$87,$C2,$FD,$1F,$86,$F8,$D8,$7C,$3F,$86,$FA,$87,$F6
	.dc.b	$BD,$D9,$FD,$7D,$3C,$55,$F5,$F5,$F3,$3F,$CB,$76,$F8,$17,$EE,$DF
	.dc.b	$D9,$3E,$EE,$EF,$D8,$3F,$0F,$8B,$64,$70,$7C,$23,$EB,$71,$FD,$47
	.dc.b	$EC,$72,$78,$DF,$C5,$9F,$6C,$F8,$47,$E2,$F8,$FE,$13,$F7,$F8,$BE
	.dc.b	$13,$F8,$1E,$1F,$84,$FE,$0F,$9F,$E0,$BF,$0C,$FE,$75,$DF,$F9,$97
	.dc.b	$F3,$67,$F5,$FE,$DF,$CF,$BF,$8B,$F8,$6F,$D3,$6F,$FA,$58,$FD,$79
	.dc.b	$FA,$62,$93,$7F,$D8,$B9,$6B,$9D,$FF,$66,$E5,$AE,$47,$47,$E9,$6F
	.dc.b	$C7,$1E,$62,$57,$F2,$2C,$7D,$4F,$EB,$7E,$1B,$F6,$5F,$09,$FB,$3F
	.dc.b	$87,$FF,$5C,$7C,$6F,$F9,$43,$0E,$BD,$AF,$0B,$F5,$F1,$FA,$9F,$07
	.dc.b	$F5,$EF,$FB,$E2,$9F,$11,$FA,$F8,$FD,$E7,$83,$FE,$35,$9F,$F2,$0F
	.dc.b	$02,$7D,$30,$F8,$7F,$0F,$C3,$C9,$E9,$F0,$FE,$4F,$87,$F2,$FA,$52
	.dc.b	$7C,$0D,$5C,$4C,$BB,$72,$B3,$FE,$A3,$18,$6D,$78,$AF,$95,$F1,$8C
	.dc.b	$1D,$88,$76,$AE,$53,$C1,$B1,$F0,$F0,$F9,$78,$FA,$E3,$81,$F0,$32
	.dc.b	$F9,$F9,$FD,$0C,$BE,$8F,$BD,$F4,$BD,$4F,$A7,$EF,$FD,$4F,$5B,$EA
	.dc.b	$FC,$07,$D6,$F7,$FE,$BF,$AD,$F6,$3E,$1F,$EC,$FB,$F3,$6F,$A8,$FD
	.dc.b	$1F,$5F,$EE,$FC,$3F,$DE,$F8,$7F,$BF,$F0,$FF,$EF,$7F,$E0,$1F,$F2
	.dc.b	$16,$1F,$53,$97,$1E,$CC,$8F,$B1,$D3,$F1,$7A,$97,$FE,$47,$6D,$6E
	.dc.b	$BD,$F1,$87,$8F,$4F,$D4,$AE,$04,$AE,$17,$D6,$31,$5C,$0F,$CF,$B8
	.dc.b	$7A,$F3,$31,$F5,$7C,$0F,$B2,$7F,$29,$A7,$C0,$7F,$2D,$B7,$CE,$7F
	.dc.b	$31,$CF,$F3,$9F,$CD,$73,$7C,$E7,$C1,$B7,$BB,$E1,$7E,$7C,$7E,$7F
	.dc.b	$8F,$E4,$7C,$3F,$C9,$F8,$69,$3F,$9D,$F8,$8F,$F3,$3B,$F9,$EC,$FC
	.dc.b	$4F,$A3,$4F,$33,$F1,$B8,$3E,$8E,$7F,$97,$DF,$E1,$F1,$FC,$7E,$3B
	.dc.b	$7F,$09,$B5,$F3,$63,$94,$62,$FE,$37,$1F,$CB,$8C,$1C,$38,$FE,$45
	.dc.b	$7B,$AB,$C3,$0E,$4F,$C3,$4F,$EF,$79,$7F,$82,$EB,$F2,$63,$E8,$F2
	.dc.b	$7C,$97,$E9,$F2,$7C,$97,$D9,$C9,$F2,$1F,$AF,$C9,$F2,$1F,$07,$1F
	.dc.b	$C8,$5C,$7F,$23,$83,$F7,$1A,$FE,$9A,$7F,$87,$76,$FE,$AF,$87,$F2
	.dc.b	$52,$EE,$FC,$D7,$C4,$7E,$CE,$FF,$72,$7E,$1B,$C0,$C6,$3F,$C0,$DB
	.dc.b	$7F,$C0,$A3,$13,$FB,$EC,$3A,$D6,$89,$1F,$1E,$FF,$7A,$F9,$08,$7E
	.dc.b	$53,$C0,$FD,$E3,$F3,$7E,$23,$F7,$8F,$CF,$F8,$8F,$DE,$3F,$CA,$7C
	.dc.b	$1F,$DE,$3F,$4F,$E2,$3F,$69,$EB,$7E,$EE,$3F,$3F,$E0,$5E,$1F,$8D
	.dc.b	$D4,$F8,$7F,$DD,$7C,$3F,$EE,$FE,$1E,$61,$F7,$7E,$1F,$F7,$DF,$0F
	.dc.b	$FB,$FF,$7E,$4F,$E0,$5A,$FE,$4D,$BC,$9F,$B3,$6F,$85,$FB,$96,$CF
	.dc.b	$ED,$1B,$CD,$FC,$11,$BA,$30,$FE,$1D,$8F,$92,$FF,$41,$97,$A1,$FC
	.dc.b	$96,$9F,$C4,$FD,$3F,$E2,$9E,$B7,$F1,$5F,$86,$FE,$2D,$EF,$7F,$17
	.dc.b	$F8,$7F,$E9,$58,$7F,$0E,$E2,$FF,$24,$F0,$7F,$40,$7F,$C9,$27,$F9
	.dc.b	$22,$3F,$76,$FF,$5B,$6F,$C3,$F2,$E2,$FE,$39,$28,$EA,$FE,$4C,$7A
	.dc.b	$7B,$BE,$DF,$0F,$70,$FE,$E7,$4F,$73,$1C,$BF,$9C,$F5,$2C,$FF,$94
	.dc.b	$2B,$C7,$F0,$1F,$1B,$9E,$47,$D3,$21,$F4,$CB,$AF,$D3,$6E,$7F,$4D
	.dc.b	$A7,$D3,$2E,$3F,$AC,$BF,$0F,$C1,$F8,$B2,$F1,$E7,$E4,$D3,$CB,$AF
	.dc.b	$D1,$C7,$F8,$27,$17,$D1,$7E,$4F,$C1,$FD,$17,$E5,$9E,$CF,$6F,$DB
	.dc.b	$F0,$7E,$EF,$83,$F7,$F1,$CA,$FE,$49,$DB,$F5,$72,$F0,$47,$D4,$D3
	.dc.b	$C8,$FD,$6D,$BE,$3E,$FF,$12,$3E,$11,$38,$BE,$13,$78,$BE,$13,$4F
	.dc.b	$84,$5E,$2F,$CA,$3F,$13,$B7,$FC,$BA,$FC,$6E,$56,$1F,$1A,$3F,$0F
	.dc.b	$C9,$F1,$9F,$C4,$F2,$7C,$67,$F9,$47,$4F,$E5,$9F,$E5,$5D,$3F,$7B
	.dc.b	$97,$E3,$47,$E5,$4F,$8C,$7E,$4B,$93,$E3,$3F,$92,$21,$FC,$8B,$8B
	.dc.b	$FB,$37,$57,$E9,$63,$F4,$E7,$F3,$7E,$BF,$D5,$BF,$CE,$7B,$7F,$1B
	.dc.b	$F0,$72,$3F,$3E,$D7,$F4,$CD,$E3,$FC,$FB,$78,$FF,$50,$D9,$FC,$DB
	.dc.b	$78,$FF,$9A,$B7,$8F,$F7,$93,$FC,$34,$F2,$10,$FE,$B9,$9F,$F8,$37
	.dc.b	$3F,$D0,$FC,$37,$E8,$BE,$13,$F4,$7F,$0D,$FB,$CB,$FD,$AF,$C3,$7E
	.dc.b	$9B,$E1,$7F,$4F,$F0,$BF,$A8,$F8,$7F,$D4,$FC,$24,$BF,$DD,$BF,$B6
	.dc.b	$D3,$F8,$07,$C3,$7E,$C3,$F8,$07,$FC,$B5,$C2,$7F,$CA,$B7,$FD,$F1
	.dc.b	$FF,$29,$BF,$E6,$A7,$FC,$A4,$E7,$D4,$E5,$EA,$73,$F7,$A4,$F1,$6B
	.dc.b	$FE,$1E,$AF,$ED,$6D,$8C,$F3,$7F,$89,$E7,$2D,$8C,$F3,$7F,$95,$A7
	.dc.b	$2D,$8C,$D7,$96,$FD,$98,$7B,$7D,$2F,$77,$AB,$EF,$F8,$6F,$81,$8F
	.dc.b	$F0,$ED,$7E,$17,$AD,$F0,$FD,$BF,$89,$F0,$DF,$17,$DC,$F8,$DE,$B7
	.dc.b	$C7,$F8,$6F,$91,$F0,$DF,$27,$E1,$BE,$57,$BD,$38,$BF,$72,$7F,$1E
	.dc.b	$D3,$E7,$7B,$1E,$18,$F9,$9F,$01,$E5,$7E,$77,$C0,$FD,$3F,$84,$F5
	.dc.b	$F5,$7D,$3E,$8B,$D3,$F0,$2F,$E9,$46,$08,$FC,$F2,$1F,$34,$BD,$5F
	.dc.b	$36,$FF,$01,$87,$86,$7E,$A6,$3E,$57,$EB,$65,$F1,$5F,$C1,$72,$7E
	.dc.b	$1F,$E0,$FE,$2C,$7F,$90,$3B,$7F,$C6,$67,$F9,$74,$7C,$1F,$F3,$C5
	.dc.b	$F0,$BF,$1E,$78,$7C,$26,$4F,$CA,$3F,$1C,$FC,$86,$3E,$3D,$7F,$09
	.dc.b	$9F,$9D,$FC,$36,$9F,$1E,$BD,$BD,$5F,$83,$69,$F1,$8B,$C3,$F2,$DB
	.dc.b	$C3,$F2,$E3,$11,$F1,$C9,$7F,$91,$77,$C1,$F2,$1F,$D2,$76,$FE,$A7
	.dc.b	$2F,$36,$3F,$95,$CF,$E4,$57,$E7,$F8,$7E,$43,$FA,$BE,$AF,$89,$F0
	.dc.b	$DF,$17,$E1,$BE,$37,$1F,$C7,$E2,$F8,$D7,$FE,$84,$E3,$71,$FC,$78
	.dc.b	$FC,$6F,$0F,$FB,$3F,$E1,$BF,$37,$1F,$B7,$F0,$3F,$3E,$FD,$5F,$86
	.dc.b	$FD,$1B,$FB,$BF,$0F,$EA,$7C,$34,$8F,$B4,$D7,$C0,$DE,$6F,$F4,$D6
	.dc.b	$68,$C7,$1F,$FB,$22,$31,$5C,$88,$72,$2D,$7E,$F6,$FC,$71,$ED,$C3
	.dc.b	$CE,$FF,$05,$C7,$D6,$FC,$1C,$BF,$15,$F0,$DF,$B5,$8F,$C8,$7C,$37
	.dc.b	$ED,$5F,$E2,$DE,$17,$ED,$5F,$E3,$1E,$0F,$E4,$7C,$FF,$DA,$47,$F1
	.dc.b	$1F,$07,$F1,$FC,$3F,$EF,$7F,$23,$F2,$C7,$E5,$FD,$3F,$CC,$7C,$37
	.dc.b	$E6,$7D,$49,$87,$E7,$FE,$13,$F3,$9E,$CC,$8F,$D1,$35,$FB,$0D,$E4
	.dc.b	$F9,$2D,$E6,$F9,$0D,$96,$6E,$64,$E6,$30,$FE,$AF,$0F,$1D,$7E,$17
	.dc.b	$1F,$3B,$FB,$3C,$FD,$6F,$ED,$74,$FD,$97,$C2,$7E,$CF,$DA,$FD,$A7
	.dc.b	$C3,$7E,$D7,$E1,$BF,$C6,$3F,$81,$FF,$91,$30,$F1,$AF,$87,$CF,$A9
	.dc.b	$27,$FD,$7F,$3F,$E4,$57,$FD,$B2,$DF,$82,$FF,$4A,$70,$FF,$AF,$27
	.dc.b	$F3,$5A,$7C,$D7,$8F,$C3,$7A,$B9,$3F,$1B,$C1,$ED,$9F,$C3,$F2,$7E
	.dc.b	$37,$C0,$FC,$7C,$7C,$6C,$7F,$1E,$FE,$34,$97,$F9,$69,$DF,$2F,$CB
	.dc.b	$6D,$70,$9D,$57,$63,$A7,$27,$8F,$0F,$CE,$63,$F1,$32,$F4,$57,$A8
	.dc.b	$F9,$87,$C2,$F8,$6F,$93,$F0,$5F,$2B,$E1,$BE,$64,$FC,$63,$E6,$FC
	.dc.b	$27,$C9,$7E,$3F,$17,$D0,$F8,$69,$1F,$BC,$DB,$E4,$6D,$FB,$0E,$1F
	.dc.b	$0A,$D9,$5F,$5B,$5F,$43,$F1,$76,$F9,$4F,$C4,$9F,$94,$FB,$F8,$7E
	.dc.b	$23,$71,$F8,$93,$A8,$E0,$7C,$FE,$1F,$92,$47,$F4,$47,$F4,$1B,$F0
	.dc.b	$DB,$8F,$FD,$09,$3F,$93,$3E,$39,$F2,$4F,$C3,$F0,$FC,$87,$F1,$7C
	.dc.b	$1F,$21,$FC,$59,$F2,$0F,$C7,$EF,$F2,$75,$F8,$51,$F9,$6D,$FE,$4B
	.dc.b	$FE,$43,$3F,$33,$F0,$DF,$9C,$F8,$6F,$CE,$FC,$37,$E9,$A7,$F2,$FF
	.dc.b	$0F,$FA,$0F,$87,$FD,$77,$17,$EC,$27,$F7,$64,$3F,$76,$5C,$BF,$76
	.dc.b	$DE,$2F,$DD,$B4,$FC,$B7,$7F,$42,$3F,$A8,$BF,$D7,$E1,$E3,$BF,$5F
	.dc.b	$0F,$C9,$CB,$D1,$7F,$AE,$D3,$F7,$2F,$F6,$FE,$DF,$DC,$BF,$B2,$F8
	.dc.b	$6F,$DC,$FC,$37,$EE,$BB,$7F,$77,$F0,$DF,$BC,$F8,$6F,$DE,$F0,$CC
	.dc.b	$3F,$A4,$F3,$BE,$1B,$F8,$0F,$C3,$7F,$02,$F8,$0F,$F5,$9F,$F0,$0F
	.dc.b	$F9,$10,$FF,$21,$DB,$8B,$CF,$1F,$E4,$4E,$23,$FE,$44,$9D,$73,$7E
	.dc.b	$5A,$73,$F0,$FD,$09,$3F,$9A,$BE,$71,$88,$E2,$7F,$24,$B4,$66,$77
	.dc.b	$3C,$6F,$B5,$6A,$30,$71,$AE,$13,$A6,$C7,$5F,$77,$DF,$E9,$7C,$0F
	.dc.b	$87,$F8,$37,$F0,$EB,$F8,$FE,$9F,$C9,$B5,$F8,$BE,$AF,$C6,$F8,$3F
	.dc.b	$8F,$EE,$FC,$8F,$87,$F9,$3E,$F7,$CA,$F8,$7F,$97,$EB,$7C,$CF,$87
	.dc.b	$F9,$BE,$BC,$D7,$E9,$EB,$F4,$9F,$4F,$2F,$CC,$F8,$7F,$A7,$EB,$F9
	.dc.b	$A7,$3C,$DF,$44,$FA,$1C,$FF,$53,$E0,$BE,$C9,$27,$6D,$7E,$6E,$1F
	.dc.b	$62,$31,$3C,$67,$64,$37,$49,$DF,$27,$D0,$7F,$0F,$CF,$E5,$AD,$F3
	.dc.b	$FA,$1F,$3E,$7E,$D7,$FA,$3E,$9F,$29,$F5,$F2,$7C,$A7,$DB,$F0,$7F
	.dc.b	$92,$F8,$6F,$95,$1F,$03,$8B,$E5,$BF,$0B,$87,$E5,$3F,$C0,$78,$3F
	.dc.b	$89,$63,$DF,$83,$F2,$F9,$F8,$EB,$CA,$7C,$AD,$3D,$2F,$BE,$7E,$13
	.dc.b	$F0,$36,$F7,$BF,$93,$6E,$1F,$26,$38,$6F,$F0,$DA,$6C,$B1,$B7,$E5
	.dc.b	$EF,$F1,$EF,$F5,$78,$79,$5F,$3C,$FE,$40,$87,$D2,$E7,$FC,$83,$F8
	.dc.b	$2E,$7F,$DB,$F0,$7C,$78,$FC,$2F,$2F,$C6,$7E,$CF,$07,$C6,$C3,$F7
	.dc.b	$3C,$1F,$E2,$2E,$CF,$E0,$55,$F8,$CF,$85,$FD,$03,$F8,$0D,$FF,$48
	.dc.b	$FE,$43,$E1,$FF,$A8,$BF,$8A,$FE,$01,$7C,$3F,$EB,$38,$BF,$7F,$3F
	.dc.b	$96,$EF,$C7,$4F,$ED,$63,$F1,$71,$82,$3B,$21,$A3,$F4,$77,$D9,$F1
	.dc.b	$1F,$A2,$EF,$FD,$BB,$FE,$14,$F0,$3F,$8C,$E9,$FB,$78,$FF,$0B,$F8
	.dc.b	$1E,$8E,$0F,$E5,$F8,$3E,$1F,$1B,$FF,$17,$8F,$F4,$CD,$FE,$BB,$C1
	.dc.b	$FD,$DB,$E2,$F8,$6D,$BF,$E3,$89,$FF,$3A,$BF,$B4,$F0,$73,$C0,$97
	.dc.b	$C2,$61,$B5,$FE,$7A,$5F,$0B,$92,$47,$C0,$6B,$F3,$DB,$A7,$A9,$B9
	.dc.b	$FA,$5B,$3E,$56,$E1,$E3,$6D,$F8,$27,$B1,$0E,$C5,$BE,$CD,$BE,$CD
	.dc.b	$3B,$16,$FB,$36,$BB,$3D,$09,$5D,$1B,$5D,$1A,$74,$38,$7B,$AF,$6B
	.dc.b	$EB,$7B,$3F,$2F,$E1,$7E,$67,$C2,$FC,$DF,$85,$F9,$D5,$F3,$FE,$13
	.dc.b	$E8,$7C,$2F,$D1,$F7,$A6,$7F,$8C,$6B,$F9,$B6,$EB,$F8,$66,$EB,$F8
	.dc.b	$56,$CF,$E0,$5B,$BF,$DF,$7B,$C7,$E6,$30,$F1,$BF,$9A,$CB,$CE,$FE
	.dc.b	$73,$4F,$C2,$6B,$F8,$5F,$82,$FC,$37,$B5,$F8,$7F,$85,$FC,$47,$AF
	.dc.b	$F8,$9F,$85,$FC,$57,$C2,$FE,$3B,$A7,$FC,$F1,$BF,$FB,$BB,$D7,$C7
	.dc.b	$C0,$4D,$BF,$1D,$AF,$C7,$7F,$57,$CD,$F1,$DF,$D6,$F3,$7C,$77,$F5
	.dc.b	$FC,$DE,$CE,$87,$2D,$24,$FC,$92,$1F,$24,$BC,$FF,$25,$BC,$FF,$25
	.dc.b	$EA,$7C,$53,$C2,$7C,$53,$CA,$7C,$53,$F8,$0E,$5F,$15,$FE,$07,$C7
	.dc.b	$F1,$5F,$E0,$BC,$5F,$15,$FE,$0F,$C3,$F1,$5F,$8D,$BF,$C3,$7F,$B0
	.dc.b	$F0,$B8,$3F,$D8,$7A,$FE,$5B,$4F,$CA,$13,$4F,$CA,$36,$FF,$28,$DE
	.dc.b	$1F,$CF,$C7,$E6,$1B,$19,$AE,$95,$C8,$95,$C9,$B5,$C1,$B3,$C2,$74
	.dc.b	$49,$D3,$63,$51,$82,$38,$27,$4F,$F0,$02,$4F,$ED,$5A,$FF,$69,$8D
	.dc.b	$B9,$21,$ED,$72,$D8,$E3,$7E,$F6,$DF,$A9,$B3,$E7,$6D,$F9,$9B,$7D
	.dc.b	$2F,$95,$78,$1A,$F9,$67,$64,$75,$7D,$DA,$77,$2D,$F7,$6D,$F7,$9E
	.dc.b	$11,$EB,$33,$5D,$5C,$13,$A7,$F8,$93,$D4,$CE,$1E,$62,$57,$46,$C6
	.dc.b	$DB,$87,$8E,$38,$CE,$C8,$ED,$B1,$B9,$E3,$87,$A6,$B4,$7F,$27,$BF
	.dc.b	$2D,$32,$FF,$7B,$5F,$FA,$39,$FF,$5A,$D7,$E8,$BE,$0B,$F8,$57,$C2
	.dc.b	$FF,$0B,$F8,$3F,$E1,$9F,$0B,$FC,$37,$E1,$3F,$87,$7C,$07,$F0,$FF
	.dc.b	$84,$98,$FF,$95,$B2,$46,$23,$F5,$AD,$CF,$C0,$DD,$FF,$D0,$95,$C2
	.dc.b	$B9,$4E,$56,$BE,$98,$F0,$C7,$7C,$7C,$AF,$87,$3F,$4B,$E3,$D7,$F9
	.dc.b	$27,$C2,$FF,$25,$F7,$BF,$93,$7C,$1F,$F2,$7F,$84,$FE,$51,$F0,$BF
	.dc.b	$CA,$7E,$17,$F9,$57,$C2,$FF,$2B,$F8,$5F,$E5,$9F,$0B,$FC,$B7,$CE
	.dc.b	$9A,$FF,$BC,$3A,$3E,$6F,$1B,$E0,$7F,$99,$7A,$3F,$E0,$0F,$E0,$1F
	.dc.b	$F6,$AD,$FF,$9B,$9F,$F5,$4F,$47,$FA,$F9,$FF,$54,$1F,$E7,$8E,$8F
	.dc.b	$F5,$33,$FE,$25,$8D,$7B,$DF,$91,$7E,$FF,$75,$AF,$C1,$F7,$7F,$00
	.dc.b	$B5,$F8,$7D,$5F,$17,$C0,$78,$FE,$03,$C9,$EB,$F9,$71,$F3,$E9,$F6
	.dc.b	$76,$F4,$7C,$27,$A7,$E0,$7D,$5E,$BF,$AF,$E1,$BD,$9F,$03,$ED,$F8
	.dc.b	$1F,$77,$B5,$EF,$F8,$19,$9F,$C5,$F4,$FD,$7F,$09,$DA,$3E,$16,$DE
	.dc.b	$27,$FD,$3B,$C8,$F8,$6F,$3C,$63,$93,$EF,$BF,$A4,$3F,$3B,$E5,$7E
	.dc.b	$86,$FE,$0F,$47,$C3,$9F,$C3,$F5,$BC,$BF,$85,$1D,$BA,$3E,$A6,$5E
	.dc.b	$18,$F1,$9F,$99,$EC,$FB,$1A,$FC,$28,$F2,$FC,$2F,$DB,$F8,$5F,$85
	.dc.b	$1E,$8E,$4F,$BD,$F0,$DF,$0A,$3D,$7C,$7F,$09,$FF,$49,$71,$31,$F8
	.dc.b	$5D,$FF,$81,$E1,$F1,$3F,$83,$D7,$E0,$DF,$E1,$F8,$FE,$0C,$FE,$02
	.dc.b	$3F,$89,$90,$F8,$9C,$7F,$CF,$F8,$BE,$04,$7C,$93,$E0,$1F,$3A,$FB
	.dc.b	$E1,$F8,$E3,$C6,$7D,$68,$FA,$7D,$5F,$99,$E0,$F8,$11,$F5,$BA,$3F
	.dc.b	$02,$FD,$8D,$FE,$03,$F9,$DD,$FE,$07,$81,$FA,$4E,$CF,$D1,$7C,$17
	.dc.b	$E8,$FE,$17,$FC,$C7,$FC,$03,$FE,$79,$C3,$F4,$9E,$0E,$3E,$13,$3F
	.dc.b	$13,$FA,$9F,$13,$F8,$17,$27,$F0,$1B,$FE,$0A,$43,$F8,$69,$79,$7F
	.dc.b	$86,$B7,$8F,$F8,$6F,$A7,$CB,$9B,$F8,$74,$F0,$3F,$69,$87,$81,$FF
	.dc.b	$77,$E9,$FE,$83,$7F,$2B,$A7,$A3,$E1,$FD,$51,$F9,$8D,$FF,$C6,$1C
	.dc.b	$1E,$CC,$7F,$51,$3F,$9C,$F1,$3F,$52,$F4,$F8,$9F,$D4,$3D,$BE,$1F
	.dc.b	$E1,$F8,$72,$3F,$50,$7F,$20,$D1,$EF,$7C,$8F,$85,$F9,$3F,$05,$F2
	.dc.b	$BE,$1B,$E5,$F9,$1F,$33,$E1,$FE,$6F,$C2,$7C,$EF,$84,$F8,$1C,$D3
	.dc.b	$4F,$88,$7E,$99,$C3,$F0,$FA,$3E,$4B,$7A,$3E,$4B,$F1,$2B,$AB,$F1
	.dc.b	$B0,$F0,$BF,$23,$2F,$2B,$F2,$B4,$FB,$FA,$FE,$03,$E1,$FF,$03,$F0
	.dc.b	$BF,$82,$F8,$7F,$C1,$FC,$37,$E1,$3E,$1E,$63,$F4,$BE,$1F,$FD,$53
	.dc.b	$FB,$AF,$F9,$70,$FF,$67,$C7,$FB,$F5,$7C,$47,$C5,$8F,$B3,$CB,$E4
	.dc.b	$5C,$4E,$24,$7C,$DE,$5F,$90,$E2,$74,$FF,$50,$C3,$FC,$88,$FC,$13
	.dc.b	$E4,$F8,$3F,$93,$7E,$31,$F2,$CC,$F6,$7E,$3A,$BE,$59,$30,$F9,$73
	.dc.b	$C2,$B9,$4E,$63,$F0,$1C,$BF,$03,$E2,$3C,$B1,$EA,$BF,$4D,$BA,$7F
	.dc.b	$C2,$33,$93,$D8,$7C,$23,$E9,$1E,$D9,$E3,$97,$C8,$8C,$1E,$FF,$82
	.dc.b	$FA,$AE,$2D,$3F,$E0,$6C,$BE,$2C,$63,$E1,$3E,$AF,$C4,$7D,$6F,$86
	.dc.b	$FD,$EF,$17,$D8,$F8,$89,$97,$E7,$30,$E7,$C7,$F5,$E3,$D1,$1E,$BE
	.dc.b	$3F,$B1,$9F,$6D,$FD,$AD,$EC,$F8,$97,$F4,$CF,$95,$C7,$F3,$F8,$7E
	.dc.b	$6D,$7C,$BC,$7E,$6C,$60,$95,$F5,$CE,$3D,$BF,$A2,$AF,$07,$27,$69
	.dc.b	$F2,$E4,$CB,$CB,$87,$D2,$8F,$2B,$73,$F6,$61,$F1,$B8,$BE,$76,$1E
	.dc.b	$E7,$FC,$63,$5C,$7A,$7F,$3B,$F0,$FF,$9E,$E4,$F9,$77,$F4,$6D,$C7
	.dc.b	$F3,$23,$1C,$5F,$31,$79,$9F,$E3,$ED,$7F,$83,$C7,$1F,$88,$FD,$4F
	.dc.b	$B7,$FC,$1E,$3A,$71,$7E,$B3,$E2,$3F,$77,$1E,$5F,$88,$FD,$7F,$C4
	.dc.b	$48,$F1,$35,$F5,$37,$8B,$DC,$DE,$2F,$C6,$B6,$7F,$54,$DE,$6F,$C5
	.dc.b	$DF,$82,$FE,$36,$1F,$BE,$C7,$CD,$87,$C7,$CB,$F8,$0E,$7F,$C2,$A3
	.dc.b	$F9,$AF,$83,$FC,$13,$E2,$3F,$85,$47,$F3,$8F,$07,$F8,$4B,$F6,$3C
	.dc.b	$1F,$E1,$3F,$11,$FC,$2A,$21,$F9,$3E,$57,$C4,$7F,$0E,$F8,$8F,$E1
	.dc.b	$FF,$11,$FC,$43,$CF,$FE,$23,$F0,$5F,$C2,$30,$FC,$BF,$C1,$FF,$8F
	.dc.b	$38,$FF,$46,$FF,$8F,$39,$FF,$6A,$BD,$4E,$13,$A8,$B3,$A6,$D7,$F9
	.dc.b	$79,$FF,$21,$35,$CF,$39,$CB,$5C,$98,$21,$82,$47,$EE,$DD,$5F,$0F
	.dc.b	$87,$E3,$7E,$EA,$FC,$86,$5F,$09,$FC,$96,$9F,$CC,$B5,$EF,$F0,$1E
	.dc.b	$0F,$5F,$C3,$EF,$F8,$BE,$1F,$C7,$EB,$F9,$3D,$FF,$2F,$C2,$79,$BE
	.dc.b	$1E,$69,$F4,$73,$F4,$FC,$3F,$AB,$E0,$7E,$84,$FE,$87,$9B,$DB,$F0
	.dc.b	$BE,$EF,$85,$F7,$FC,$2F,$C0,$F8,$79,$7E,$D6,$BF,$59,$BD,$5F,$59
	.dc.b	$BC,$DF,$59,$B3,$F4,$5B,$D5,$F3,$EF,$C1,$3F,$27,$1F,$8C,$FE,$EF
	.dc.b	$9F,$D0,$FC,$DD,$7E,$33,$F3,$F9,$BE,$3B,$F4,$7E,$1B,$E3,$3F,$4F
	.dc.b	$E0,$BE,$33,$F5,$7E,$1B,$EF,$FF,$00,$7F,$89,$77,$FE,$5E,$FF,$89
	.dc.b	$FC,$DF,$F5,$DF,$83,$F9,$8B,$F0,$1F,$77,$93,$E4,$BF,$C3,$F9,$3E
	.dc.b	$4B,$F8,$1E,$4F,$92,$FC,$72,$1F,$1C,$B8,$FC,$76,$F2,$7C,$76,$9F
	.dc.b	$1C,$BC,$7F,$95,$E4,$F9,$3E,$03,$97,$E4,$D7,$EA,$39,$7E,$4E,$3F
	.dc.b	$C9,$79,$7E,$4C,$FE,$B3,$97,$E4,$BF,$AE,$E5,$F9,$2F,$EC,$39,$7E
	.dc.b	$4B,$FB,$2E,$5F,$92,$FE,$D3,$97,$E4,$FC,$2C,$8F,$EA,$1C,$DF,$E3
	.dc.b	$B7,$F9,$97,$7F,$E7,$5F,$EA,$30,$F8,$6F,$A3,$EC,$7E,$92,$7F,$51
	.dc.b	$E1,$7E,$BB,$C3,$FD,$5C,$7F,$02,$E3,$BC,$3F,$BD,$8F,$D7,$46,$39
	.dc.b	$BF,$7D,$18,$21,$C4,$B3,$C5,$FE,$1B,$5F,$B5,$7F,$71,$3E,$47,$FB
	.dc.b	$A6,$5E,$77,$FA,$37,$89,$FC,$65,$CF,$C3,$FE,$2B,$E1,$FF,$6D,$1F
	.dc.b	$C0,$7C,$5F,$C6,$FC,$3F,$ED,$A3,$F8,$27,$8B,$FB,$67,$F8,$37,$87
	.dc.b	$FB,$67,$FA,$6F,$C0,$FF,$84,$33,$7C,$3F,$E5,$89,$3E,$46,$BF,$31
	.dc.b	$BB,$FE,$69,$BB,$73,$6C,$FF,$BF,$AB,$AB,$CC,$87,$32,$D7,$36,$D7
	.dc.b	$36,$9C,$EB,$AF,$7A,$FA,$CF,$32,$5F,$36,$DF,$36,$9C,$CE,$B5,$CE
	.dc.b	$FA,$DF,$4B,$E4,$4B,$E5,$E2,$31,$F9,$5F,$09,$FB,$EF,$85,$FD,$FF
	.dc.b	$BB,$FC,$03,$E1,$BF,$80,$D7,$F0,$2F,$7F,$F8,$1F,$C3,$FF,$04,$F8
	.dc.b	$1F,$E0,$BE,$AC,$E1,$FC,$1B,$5F,$CD,$B7,$6F,$E1,$2D,$DB,$ED,$36
	.dc.b	$7E,$83,$78,$7E,$35,$F8,$27,$F1,$F8,$F9,$1F,$C8,$E7,$E8,$7F,$27
	.dc.b	$AF,$F1,$8D,$BF,$8C,$FC,$0F,$F1,$AF,$87,$FE,$37,$EB,$7F,$1C,$F5
	.dc.b	$BF,$8E,$FC,$27,$F5,$CF,$E0,$0F,$F3,$DF,$0F,$F9,$31,$FF,$13,$B7
	.dc.b	$E2,$34,$7F,$71,$E6,$F9,$5A,$FE,$1F,$4F,$94,$FE,$93,$9B,$E5,$3F
	.dc.b	$A6,$E6,$F9,$4F,$EA,$39,$BE,$53,$F2,$08,$7C,$82,$E3,$F2,$1B,$CD
	.dc.b	$F2,$1A,$7C,$82,$F2,$FE,$37,$9B,$E5,$73,$B9,$FE,$55,$7F,$0E,$E7
	.dc.b	$F9,$4F,$F1,$0E,$7F,$94,$FF,$12,$E7,$F9,$4F,$EF,$B9,$FE,$53,$FC
	.dc.b	$03,$9B,$E5,$3F,$C0,$B9,$7E,$53,$F2,$FE,$1B,$E3,$BF,$E7,$9E,$6F
	.dc.b	$F4,$CB,$FB,$A3,$F3,$47,$E7,$36,$FC,$D3,$F9,$CD,$7F,$35,$BF,$E8
	.dc.b	$78,$3F,$35,$3C,$1B,$5C,$23,$95,$70,$75,$5C,$48,$E0,$93,$C6,$B9
	.dc.b	$D7,$18,$C5,$70,$6D,$70,$9D,$13,$A7,$F6,$24,$8F,$53,$5F,$2B,$F8
	.dc.b	$9E,$3E,$93,$C6,$76,$43,$65,$C3,$CB,$5C,$DB,$19,$C3,$9C,$E5,$B1
	.dc.b	$93,$D1,$19,$69,$9C,$3D,$A7,$CB,$E6,$FC,$EB,$8C,$3C,$D3,$E4,$26
	.dc.b	$1E,$26,$E1,$E0,$69,$D8,$B5,$CF,$0F,$13,$63,$C5,$7D,$AF,$A1,$2F
	.dc.b	$A3,$DF,$0F,$2C,$EE,$72,$4A,$F0,$57,$16,$C6,$6F,$C1,$1E,$4A,$E4
	.dc.b	$E8,$91,$AA,$E4,$DA,$E5,$5C,$23,$15,$F9,$CB,$F9,$39,$7F,$AB,$A1
	.dc.b	$28,$FF,$6D,$5F,$FB,$A9,$7C,$2F,$F1,$3F,$87,$FE,$29,$F0,$DF,$C5
	.dc.b	$7E,$1B,$F8,$B7,$C0,$7F,$17,$F8,$1F,$E3,$1F,$05,$34,$FD,$73,$5F
	.dc.b	$F6,$DE,$5F,$ED,$37,$C2,$DD,$3C,$53,$BD,$FF,$90,$37,$5F,$23,$6F
	.dc.b	$8E,$1E,$27,$F7,$99,$79,$9F,$DF,$69,$FC,$AF,$5F,$E5,$9F,$0F,$FC
	.dc.b	$B7,$E2,$3F,$97,$7C,$2F,$F2,$FF,$86,$FE,$61,$F0,$FF,$CC,$7E,$1F
	.dc.b	$F9,$97,$C3,$FF,$33,$F8,$3F,$E6,$9E,$AF,$F3,$5F,$42,$6D,$FE,$F9
	.dc.b	$E5,$7B,$9F,$CE,$7E,$07,$FD,$2B,$FC,$03,$FE,$A3,$E6,$5F,$D0,$8F
	.dc.b	$F5,$23,$7D,$4D,$FA,$BC,$3D,$DE,$3F,$11,$F9,$4F,$03,$EE,$13,$C1
	.dc.b	$C1,$F9,$8C,$53,$C7,$1E,$FD,$AC,$FC,$3D,$CB,$FF,$53,$9F,$EB,$3A
	.dc.b	$DC,$EA,$3C,$B3,$B8,$F6,$4F,$F2,$9D,$BD,$BE,$AF,$BB,$E2,$3D,$FE
	.dc.b	$AF,$C0,$F8,$8F,$83,$F1,$1F,$0B,$D7,$F8,$7F,$05,$F1,$3E,$1E,$67
	.dc.b	$F1,$3C,$9F,$DE,$EE,$F8,$4E,$F2,$E1,$F5,$3F,$27,$E1,$FE,$6F,$C3
	.dc.b	$7C,$EE,$9F,$93,$3E,$A3,$E7,$74,$7C,$A7,$F7,$7C,$FF,$21,$A4,$3B
	.dc.b	$72,$FC,$1E,$7F,$95,$18,$3E,$C7,$4F,$77,$B7,$47,$CB,$CB,$EF,$E7
	.dc.b	$E8,$D3,$CF,$3F,$3F,$5F,$8F,$7C,$36,$7C,$3F,$C8,$8C,$7C,$3F,$E2
	.dc.b	$3E,$1F,$F1,$3F,$0F,$F2,$27,$BF,$93,$FE,$A1,$E0,$FB,$6F,$FA,$87
	.dc.b	$D1,$FB,$19,$F8,$67,$ED,$6D,$F1,$AF,$C1,$C9,$F1,$9F,$66,$DF,$83
	.dc.b	$E4,$F8,$D1,$FC,$47,$93,$E3,$34,$F8,$DC,$3E,$CE,$1F,$C7,$C7,$AE
	.dc.b	$BA,$4E,$70,$F0,$4F,$6A,$63,$E5,$8F,$D1,$67,$F1,$1D,$70,$FC,$37
	.dc.b	$87,$0F,$C3,$7E,$47,$07,$C3,$7D,$3C,$1F,$0D,$FE,$36,$FC,$37,$F8
	.dc.b	$6F,$77,$EC,$1D,$FC,$1F,$ED,$38,$7F,$6B,$F0,$99,$DF,$EB,$E1,$FE
	.dc.b	$89,$BE,$BF,$0B,$9B,$F9,$FF,$0B,$67,$BF,$C2,$FF,$13,$E5,$FE,$33
	.dc.b	$1F,$C8,$08,$7F,$20,$F5,$35,$E3,$FF,$1F,$23,$83,$96,$4B,$ED,$68
	.dc.b	$61,$85,$95,$F8,$6F,$5E,$5F,$8D,$CF,$5B,$7F,$92,$F7,$E1,$D1,$FB
	.dc.b	$7A,$FC,$5F,$C1,$FE,$DD,$FC,$A7,$87,$FB,$77,$F1,$FE,$EF,$ED,$9F
	.dc.b	$B5,$E1,$7F,$81,$DF,$E7,$3C,$3E,$97,$DF,$C7,$FC,$E1,$BA,$7F,$38
	.dc.b	$69,$C0,$BC,$3F,$21,$F8,$C6,$38,$FE,$2B,$6F,$B3,$A2,$5F,$76,$DF
	.dc.b	$76,$9D,$CB,$7D,$E7,$81,$27,$83,$67,$83,$67,$84,$E8,$ED,$E1,$7F
	.dc.b	$85,$38,$9F,$07,$F5,$7D,$FF,$AD,$E0,$7D,$7F,$80,$FB,$1F,$07,$F6
	.dc.b	$7E,$0B,$ED,$7C,$1F,$DB,$F8,$3F,$D7,$ED,$FD,$F3,$D0,$FE,$7D,$C1
	.dc.b	$FB,$0D,$FF,$C8,$1B,$2F,$8C,$FF,$04,$C7,$AB,$FC,$1B,$3F,$0B,$FC
	.dc.b	$23,$5F,$C4,$6D,$F8,$9F,$86,$FC,$57,$C2,$7E,$2F,$E1,$BF,$19,$F0
	.dc.b	$BF,$90,$FE,$01,$FF,$48,$ED,$FD,$21,$FF,$48,$B7,$E1,$F5,$F0,$13
	.dc.b	$4F,$E7,$5C,$BF,$C5,$34,$F8,$51,$FC,$62,$3E,$13,$FC,$6B,$97,$E1
	.dc.b	$3F,$C7,$39,$7E,$13,$EF,$3F,$C3,$FC,$9F,$09,$FE,$24,$7F,$9D,$79
	.dc.b	$3E,$13,$C4,$F8,$47,$53,$E1,$1E,$13,$E1,$1F,$CA,$B0,$F8,$4F,$F2
	.dc.b	$CD,$FE,$13,$FC,$BB,$6F,$84,$FC,$3F,$87,$F8,$9A,$FB,$E3,$F9,$4E
	.dc.b	$ED,$BF,$49,$D1,$F8,$69,$FC,$4F,$07,$E1,$DA,$48,$D1,$3A,$BF,$3F
	.dc.b	$1F,$96,$6F,$47,$E4,$5B,$5C,$EB,$87,$57,$E9,$2B,$4D,$9D,$57,$0B
	.dc.b	$F0,$36,$F8,$C7,$EB,$48,$7E,$B7,$AB,$FC,$8F,$7C,$E3,$8B,$67,$8C
	.dc.b	$79,$E3,$15,$D4,$8E,$09,$1D,$B0,$F3,$B7,$0F,$3C,$EA,$3A,$1F,$E2
	.dc.b	$D7,$13,$C2,$32,$47,$04,$9E,$71,$D4,$F4,$96,$FD,$33,$B8,$C9,$CE
	.dc.b	$78,$B4,$CC,$7F,$0F,$C3,$E2,$C7,$3C,$3D,$ED,$C3,$CF,$7C,$AF,$B6
	.dc.b	$1E,$08,$C1,$2B,$AB,$63,$51,$82,$38,$25,$AF,$F2,$39,$7F,$9E,$AB
	.dc.b	$FC,$FA,$FF,$9E,$9F,$F0,$BA,$F8,$3F,$DF,$7C,$3F,$EF,$FE,$17,$F8
	.dc.b	$07,$C3,$FF,$01,$F8,$7F,$E0,$58,$FF,$03,$F8,$6F,$E0,$9F,$0D,$33
	.dc.b	$E6,$D7,$F4,$6D,$DF,$FC,$43,$A2,$31,$BF,$F0,$B6,$D7,$99,$EC,$F9
	.dc.b	$B0,$F1,$3E,$8C,$BC,$CF,$AB,$4F,$E2,$DA,$FF,$17,$F8,$7F,$E3,$1D
	.dc.b	$1F,$C6,$7E,$1F,$F8,$D7,$C3,$FF,$1B,$F8,$7F,$E3,$9F,$0F,$FC,$77
	.dc.b	$E1,$FF,$8F,$7C,$3F,$F1,$FF,$3A,$6B,$FC,$68,$F9,$3F,$0F,$FD,$CB
	.dc.b	$F8,$03,$FC,$F5,$1F,$EA,$87,$FC,$CA,$7F,$9A,$5F,$F1,$5A,$F8,$1E
	.dc.b	$1E,$FF,$1F,$82,$E5,$F1,$1C,$FE,$23,$E0,$F7,$BA,$BE,$36,$5F,$30
	.dc.b	$87,$CC,$2F,$5F,$CC,$EA,$E3,$3C,$27,$44,$9F,$F2,$8E,$8C,$3C,$33
	.dc.b	$FE,$58,$8F,$F1,$CB,$E8,$CF,$DD,$A7,$BF,$E0,$BE,$07,$B9,$F0,$7B
	.dc.b	$7F,$A2,$70,$7C,$3F,$88,$F8,$9F,$11,$F1,$7D,$79,$9F,$5F,$88,$F9
	.dc.b	$1F,$11,$23,$C0,$FE,$5B,$0F,$99,$EC,$76,$9F,$26,$9F,$3F,$E1,$3E
	.dc.b	$4C,$79,$FE,$17,$D2,$FC,$EF,$88,$FD,$6B,$FA,$4F,$2B,$F6,$2F,$ED
	.dc.b	$FA,$3F,$96,$F2,$2B,$8F,$37,$C4,$C3,$AD,$FD,$8C,$7E,$03,$F1,$78
	.dc.b	$FC,$AF,$EA,$35,$F8,$0F,$C9,$E1,$F8,$2F,$CB,$F8,$6F,$C3,$7C,$37
	.dc.b	$BE,$3F,$07,$C3,$7C,$D7,$AF,$FD,$AF,$1E,$B5,$F8,$8C,$BC,$2F,$8C
	.dc.b	$FA,$DD,$3F,$91,$D7,$D3,$D1,$FE,$BA,$E1,$F6,$57,$F0,$F2,$1F,$07
	.dc.b	$8F,$E1,$F0,$FE,$0B,$8B,$8C,$FE,$67,$0F,$83,$3F,$86,$E2,$F0,$BF
	.dc.b	$88,$CF,$CA,$FE,$EB,$4F,$83,$7F,$9E,$E2,$F8,$2F,$E8,$38,$BE,$0B
	.dc.b	$F7,$88,$7C,$23,$F5,$9C,$6D,$7F,$56,$DD,$BF,$5D,$F1,$1F,$AF,$F8
	.dc.b	$7F,$C9,$DF,$FA,$BB,$F8,$07,$FC,$FD,$7C,$7C,$1F,$F3,$2C,$3E,$13
	.dc.b	$F5,$95,$FA,$6C,$7F,$87,$34,$FE,$1C,$5B,$FD,$87,$83,$FA,$A2,$46
	.dc.b	$73,$FD,$43,$E1,$BF,$DD,$3E,$3F,$8A,$F2,$BF,$C0,$33,$F4,$BF,$AD
	.dc.b	$C7,$F5,$0F,$A7,$E1,$BC,$FF,$0F,$FA,$68,$FE,$0F,$E0,$7F,$0D,$F0
	.dc.b	$3D,$7F,$11,$EC,$F8,$99,$5F,$23,$81,$F1,$3F,$03,$E2,$64,$FC,$5F
	.dc.b	$89,$F8,$67,$C4,$F8,$0F,$8B,$F0,$DF,$1B,$C2,$F8,$FF,$0F,$F2,$3E
	.dc.b	$16,$61,$F3,$5A,$FF,$B1,$33,$46,$3B,$BF,$DB,$71,$89,$E0,$43,$84
	.dc.b	$EF,$8D,$0E,$95,$30,$F2,$4F,$DD,$CB,$D0,$FA,$C8,$78,$F5,$F7,$BE
	.dc.b	$5D,$BF,$03,$F1,$1F,$82,$F8,$6F,$C1,$FC,$4F,$E1,$3E,$27,$F0,$BF
	.dc.b	$0F,$F8,$6F,$89,$FC,$3F,$C4,$4C,$BE,$66,$5F,$EA,$EE,$7F,$F2,$FC
	.dc.b	$79,$DF,$F2,$FB,$E9,$DD,$CD,$F3,$A3,$FE,$00,$CB,$E1,$FE,$7C,$63
	.dc.b	$E1,$FE,$7B,$AF,$7F,$B1,$1C,$10,$FF,$79,$D7,$F7,$9E,$4F,$F5,$FC
	.dc.b	$EE,$72,$DE,$77,$37,$CE,$AD,$D7,$E9,$F9,$BD,$98,$F9,$23,$AE,$5E
	.dc.b	$97,$EF,$67,$F0,$7E,$23,$E1,$47,$EB,$36,$F8,$9F,$03,$F3,$63,$DF
	.dc.b	$F0,$5F,$35,$FD,$A7,$1F,$CD,$7F,$6D,$C7,$57,$A5,$FC,$81,$F9,$58
	.dc.b	$76,$7E,$4E,$3E,$3E,$0F,$C6,$E7,$F4,$F4,$F4,$47,$BA,$BE,$A6,$BF
	.dc.b	$31,$F8,$67,$C9,$8D,$6F,$F0,$A3,$24,$8C,$F1,$7C,$D2,$71,$7C,$BE
	.dc.b	$2F,$95,$18,$AF,$94,$7D,$B3,$F2,$C7,$E0,$31,$F4,$4F,$D1,$E3,$FC
	.dc.b	$2C,$FB,$0F,$C2,$9F,$2C,$FC,$3F,$17,$CB,$7E,$FF,$17,$CB,$7F,$61
	.dc.b	$C5,$F2,$73,$FD,$7F,$17,$CB,$E2,$F9,$5B,$35,$7C,$1F,$E9,$2F,$F6
	.dc.b	$BD,$5F,$9B,$F8,$8F,$D5,$E7,$FB,$BE,$FF,$CF,$7C,$2F,$EC,$23,$F1
	.dc.b	$FF,$11,$FB,$77,$F3,$DE,$05,$D3,$F6,$ED,$E2,$FE,$01,$3F,$BE,$3F
	.dc.b	$80,$67,$FC,$12,$74,$7E,$AE,$FB,$47,$D0,$C3,$C6,$FF,$0A,$C7,$CD
	.dc.b	$7F,$CF,$73,$F5,$71,$FD,$2D,$7F,$7B,$3F,$AE,$F8,$7F,$DE,$BF,$D0
	.dc.b	$FB,$BF,$71,$F1,$1F,$B9,$F3,$7F,$79,$3F,$C6,$7B,$FF,$D6,$7C,$6F
	.dc.b	$88,$FD,$F7,$C4,$4A,$FD,$CF,$C4,$7F,$01,$F8,$8F,$E0,$5F,$11,$FC
	.dc.b	$0F,$E2,$3F,$82,$7C,$47,$F0,$5F,$67,$F8,$DE,$0E,$5F,$C7,$FF,$00
	.dc.b	$FF,$5A,$27,$26,$A3,$2D,$3F,$6D,$1A,$9F,$E4,$B3,$A6,$CE,$8E,$0F
	.dc.b	$F2,$99,$E0,$F8,$63,$4F,$F1,$2A,$F3,$BF,$C5,$30,$F5,$BF,$C5,$B2
	.dc.b	$FE,$3D,$9F,$C1,$8F,$E1,$FA,$7F,$21,$D7,$BF,$5B,$37,$B7,$87,$FC
	.dc.b	$BD,$AF,$C0,$F8,$6C,$7C,$0E,$7D,$AD,$7C,$4E,$FE,$22,$6B,$FE,$00
	.dc.b	$C5,$F1,$1C,$FE,$03,$A7,$B1,$F4,$6B,$CC,$7C,$23,$E0,$7B,$1E,$C7
	.dc.b	$F9,$2F,$3F,$8F,$E1,$7C,$9D,$12,$78,$71,$7A,$63,$11,$F1,$CF,$94
	.dc.b	$E3,$87,$E8,$46,$67,$83,$CB,$4F,$9F,$3A,$3E,$FD,$F8,$DF,$E1,$78
	.dc.b	$7C,$5C,$7D,$57,$FC,$EB,$9B,$DA,$F3,$D3,$E5,$6B,$F2,$F8,$FE,$67
	.dc.b	$C3,$FC,$DF,$88,$F9,$77,$E1,$F8,$4F,$96,$F8,$F8,$BE,$5B,$FC,$B3
	.dc.b	$DB,$FC,$5E,$3F,$2E,$3D,$1C,$7E,$2D,$FF,$8F,$EB,$E5,$9F,$EA,$FB
	.dc.b	$FA,$78,$FF,$27,$C3,$F0,$DF,$C5,$CF,$C3,$E2,$F8,$31,$8C,$3E,$13
	.dc.b	$49,$3F,$30,$F9,$0F,$37,$F3,$15,$DD,$FE,$D9,$7E,$37,$E3,$1F,$80
	.dc.b	$3C,$18,$FA,$DF,$16,$5F,$1D,$FA,$9B,$FC,$77,$EB,$73,$7E,$57,$7F
	.dc.b	$8F,$1E,$9D,$FE,$3F,$99,$FC,$A3,$BF,$F8,$14,$FF,$78,$EB,$FD,$16
	.dc.b	$FF,$D1,$3B,$3F,$3D,$F0,$BF,$A8,$AF,$EE,$7D,$DF,$A1,$F8,$6F,$E0
	.dc.b	$51,$F8,$5F,$86,$FD,$21,$1C,$10,$F9,$4E,$5A,$67,$0F,$E1,$31,$97
	.dc.b	$FA,$AF,$83,$DE,$7F,$74,$7F,$83,$BC,$2F,$2B,$FC,$60,$F5,$10,$FF
	.dc.b	$04,$F8,$5F,$BB,$7F,$35,$F1,$1F,$B9,$F0,$FF,$75,$F0,$FF,$B7,$9F
	.dc.b	$CF,$7C,$27,$F0,$0F,$E0,$1F,$F3,$68,$F8,$6F,$E3,$FC,$9F,$D4,$23
	.dc.b	$FC,$DD,$5F,$AE,$6B,$FA,$96,$EB,$FD,$C9,$BD,$BF,$E1,$56,$CF,$F9
	.dc.b	$0D,$BE,$29,$A4,$F8,$7F,$21,$27,$C0,$D7,$C0,$DB,$F0,$36,$FC,$0D
	.dc.b	$9F,$03,$6F,$C0,$DB,$F0,$4F,$62,$1D,$8B,$7D,$9B,$7D,$9A,$76,$2D
	.dc.b	$F6,$79,$E1,$F2,$FE,$1F,$E8,$7C,$3F,$D1,$E9,$FA,$5F,$0F,$F4,$FE
	.dc.b	$0B,$EA,$7A,$7F,$57,$E0,$7E,$B7,$AB,$F5,$FE,$1F,$F1,$1B,$7F,$4C
	.dc.b	$EA,$BB,$7F,$47,$6E,$FF,$CE,$DB,$E4,$1A,$6B,$89,$87,$6B,$FB,$D9
	.dc.b	$78,$9F,$C0,$69,$E6,$7F,$05,$B7,$A9,$FC,$27,$07,$E2,$FC,$AF,$C6
	.dc.b	$7C,$27,$E3,$7D,$AF,$C7,$7C,$3F,$E3,$FE,$1F,$F2,$1E,$C7,$E4,$7E
	.dc.b	$1E,$65,$FE,$C6,$DD,$9F,$58,$FC,$76,$9E,$17,$F2,$1A,$F9,$5F,$C9
	.dc.b	$6D,$E9,$7F,$29,$BF,$B9,$FD,$87,$1F,$C7,$69,$F1,$CB,$C7,$F1,$DB
	.dc.b	$C7,$F1,$DC,$E7,$F1,$9F,$D1,$72,$7D,$57,$F4,$9C,$BF,$65,$FD,$37
	.dc.b	$37,$DD,$7F,$51,$CF,$FA,$FD,$FE,$34,$7E,$AF,$97,$E3,$3F,$AD,$E4
	.dc.b	$F8,$CF,$EB,$F8,$FE,$2B,$FA,$FE,$3F,$DA,$F3,$7E,$46,$3F,$2C,$4E
	.dc.b	$9F,$CB,$37,$87,$51,$82,$38,$24,$7E,$64,$E0,$58,$DB,$63,$7C,$3F
	.dc.b	$9D,$8D,$B6,$76,$D9,$DF,$07,$EB,$AB,$6D,$AD,$B6,$77,$D5,$FB,$52
	.dc.b	$57,$EA,$B0,$CB,$FA,$EA,$E0,$F9,$49,$87,$85,$B1,$A8,$C1,$1C,$12
	.dc.b	$BB,$CF,$99,$B8,$79,$9B,$87,$8A,$74,$78,$E3,$2F,$CB,$EA,$FE,$55
	.dc.b	$18,$C3,$C0,$DC,$3C,$15,$C4,$87,$42,$61,$E1,$72,$D7,$38,$78,$E3
	.dc.b	$04,$70,$48,$F2,$63,$70,$F1,$C7,$49,$C5,$F9,$23,$A9,$F2,$B0,$F4
	.dc.b	$11,$E0,$64,$B5,$D1,$E2,$DC,$3C,$91,$B3,$FC,$38,$7F,$45,$F8,$1F
	.dc.b	$EA,$9F,$03,$FD,$57,$D9,$FF,$2A,$FF,$00,$FF,$BD,$B0,$3F,$EF,$66
	.dc.b	$FC,$2E,$BE,$1F,$7F,$0F,$C3,$E1,$F8,$FC,$3C,$D3,$FC,$8F,$8A,$31
	.dc.b	$1F,$E4,$38,$C3,$A8,$CE,$BF,$EC,$E9,$CC,$EE,$73,$C1,$FE,$F3,$9C
	.dc.b	$D6,$EF,$C7,$7E,$0C,$7D,$79,$7B,$3E,$23,$DB,$3E,$F3,$FB,$76,$DF
	.dc.b	$DD,$77,$F8,$3F,$09,$F0,$BE,$23,$E1,$FC,$37,$C4,$F8,$7F,$8B,$F0
	.dc.b	$FF,$1B,$E1,$FE,$3F,$C3,$CC,$C7,$CE,$E5,$F9,$3E,$8F,$CC,$F6,$3E
	.dc.b	$6F,$B1,$F3,$BE,$07,$E7,$FA,$BF,$43,$D8,$FA,$3F,$01,$F4,$BD,$89
	.dc.b	$9F,$F9,$C3,$DD,$FF,$7B,$3F,$0C,$F9,$04,$30,$4E,$AF,$95,$DB,$F2
	.dc.b	$21,$B7,$DC,$8C,$36,$3B,$76,$F8,$1F,$AD,$87,$91,$FB,$79,$7A,$1F
	.dc.b	$BB,$A7,$B1,$FB,$FB,$7E,$2B,$7F,$F4,$FE,$9F,$E9,$E9,$F5,$7B,$D8
	.dc.b	$F6,$73,$3A,$F8,$7D,$FC,$3C,$BF,$97,$E3,$FF,$BD,$29,$EF,$FC,$B9
	.dc.b	$F8,$9F,$03,$F2,$DF,$C7,$F2,$F8,$3E,$1B,$D1,$1E,$A3,$F2,$7C,$3E
	.dc.b	$4F,$83,$F8,$3E,$37,$E2,$B9,$3F,$19,$CD,$82,$5F,$CD,$E7,$F9,$6D
	.dc.b	$9F,$BF,$CE,$D9,$8F,$C9,$BF,$CD,$65,$E2,$57,$C2,$99,$FC,$BC,$7F
	.dc.b	$49,$CD,$F2,$36,$F7,$47,$FA,$A6,$BE,$0B,$F8,$3F,$86,$F9,$BF,$0F
	.dc.b	$F3,$BE,$1B,$E7,$FC,$3F,$D0,$F8,$7F,$95,$7C,$79,$BF,$AD,$BF,$82
	.dc.b	$C7,$9B,$FB,$3E,$BF,$03,$F7,$0F,$96,$7E,$47,$5F,$56,$DF,$B1,$E0
	.dc.b	$F7,$3E,$BE,$1F,$8F,$5F,$91,$8F,$96,$FC,$73,$F2,$3C,$9F,$3D,$BC
	.dc.b	$9C,$38,$FE,$6C,$7B,$EF,$AB,$C3,$0F,$0B,$E9,$C7,$E4,$57,$F0,$7E
	.dc.b	$3F,$4C,$FA,$CF,$D4,$73,$7E,$4B,$8F,$E3,$C7,$EA,$F8,$BE,$3B,$FA
	.dc.b	$DE,$1F,$8F,$87,$D6,$8F,$8D,$C3,$F2,$39,$7F,$77,$1F,$C7,$3B,$BF
	.dc.b	$40,$FD,$5F,$87,$FC,$FF,$C3,$FE,$EE,$3F,$71,$F0,$7F,$A2,$F8,$7F
	.dc.b	$D1,$F2,$C9,$FD,$E7,$C3,$FE,$98,$8E,$08,$7E,$05,$CB,$4C,$E9,$E7
	.dc.b	$CF,$F2,$53,$D6,$7F,$11,$5E,$17,$C6,$7E,$E8,$FE,$19,$8F,$ED,$B2
	.dc.b	$F5,$E1,$FC,$C7,$3F,$DD,$57,$F1,$1F,$87,$FD,$DF,$83,$FB,$98,$FE
	.dc.b	$29,$D5,$FB,$97,$F9,$0D,$F8,$FF,$80,$7F,$D3,$D1,$8C,$FF,$4F,$18
	.dc.b	$8F,$E4,$CF,$FA,$58,$FF,$29,$BB,$3F,$8B,$AF,$77,$F9,$6C,$2B,$B7
	.dc.b	$C2,$E1,$F0,$7B,$F7,$0E,$FF,$0B,$6F,$9F,$C0,$DE,$26,$6F,$85,$C3
	.dc.b	$6B,$C5,$3A,$F8,$3C,$92,$33,$28,$C4,$79,$71,$F7,$C6,$33,$F8,$8D
	.dc.b	$8C,$CF,$28,$C3,$F0,$30,$F1,$9E,$4C,$7C,$B9,$79,$B3,$F3,$E9,$E8
	.dc.b	$D7,$D3,$F0,$7E,$AF,$77,$D7,$F0,$B3,$3F,$3F,$C0,$7B,$BE,$13,$DF
	.dc.b	$F0,$FF,$03,$E1,$65,$7F,$30,$AF,$AF,$F0,$FF,$13,$E1,$BE,$2F,$C3
	.dc.b	$78,$2B,$E0,$EF,$E4,$7E,$1F,$0F,$A1,$F8,$BC,$5F,$33,$DE,$FC,$CF
	.dc.b	$37,$E5,$FA,$5D,$F8,$E8,$FE,$48,$DC,$3F,$8B,$37,$AB,$F3,$B1,$FA
	.dc.b	$09,$DB,$F4,$B0,$F0,$3F,$44,$F9,$47,$90,$FA,$FD,$1F,$25,$FB,$3D
	.dc.b	$1E,$C7,$ED,$EF,$F8,$2E,$0F,$C1,$FB,$FF,$84,$F8,$5F,$C2,$F1,$FC
	.dc.b	$9B,$FC,$0F,$2F,$C8,$7C,$1A,$FE,$33,$97,$F8,$47,$63,$2F,$95,$5F
	.dc.b	$88,$E6,$F9,$4F,$09,$73,$7C,$B8,$C7,$37,$B3,$97,$F5,$9C,$DF,$2E
	.dc.b	$7E,$11,$F1,$8B,$9F,$C6,$7F,$83,$12,$7E,$87,$2F,$CB,$E6,$FA,$33
	.dc.b	$C3,$27,$2F,$7E,$6F,$DB,$E5,$F9,$1A,$F9,$27,$E0,$0F,$92,$7A,$CF
	.dc.b	$92,$7E,$93,$6F,$92,$FE,$9B,$8B,$E4,$BF,$A8,$E7,$FC,$8B,$FA,$AE
	.dc.b	$1F,$95,$EA,$39,$3F,$83,$F5,$7E,$72,$BF,$5F,$D9,$FB,$9F,$E0,$1F
	.dc.b	$F2,$B1,$FE,$55,$1F,$07,$9F,$84,$D7,$C3,$6F,$E1,$BF,$5B,$CF,$FB
	.dc.b	$BC,$7F,$60,$43,$E9,$78,$5A,$E5,$FD,$E4,$67,$A3,$F7,$2D,$E5,$FD
	.dc.b	$CB,$C0,$FE,$0F,$A3,$1F,$0C,$63,$3F,$46,$9E,$9F,$86,$F5,$4F,$EE
	.dc.b	$4F,$E1,$7E,$1F,$EE,$5F,$E1,$BE,$1F,$EE,$5F,$C6,$E3,$FD,$A7,$62
	.dc.b	$F8,$6F,$89,$F0,$DF,$17,$E1,$A5,$FE,$EC,$FF,$20,$D3,$E1,$BE,$57
	.dc.b	$C3,$7C,$BF,$86,$F9,$9F,$0D,$F3,$7E,$13,$E7,$7B,$DF,$3F,$CF,$FA
	.dc.b	$1E,$CC,$CF,$DF,$F0,$5F,$4C,$8E,$08,$7A,$1C,$B4,$CF,$47,$F5,$9D
	.dc.b	$53,$A8,$C3,$DE,$7E,$0D,$78,$DF,$9F,$87,$9D,$FA,$39,$7A,$DF,$A7
	.dc.b	$A7,$E1,$F5,$FC,$47,$C2,$FE,$27,$E1,$BF,$15,$EF,$7F,$43,$FE,$01
	.dc.b	$FF,$36,$EB,$FD,$AD,$FF,$2C,$63,$FB,$C5,$EA,$7F,$52,$8F,$EA,$7C
	.dc.b	$C9,$CC,$7F,$A0,$A7,$FC,$86,$FD,$0C,$BB,$7C,$3F,$7F,$33,$C1,$F0
	.dc.b	$FE,$9A,$FB,$BC,$1E,$3F,$73,$E0,$69,$F1,$78,$BC,$DF,$01,$E7,$F8
	.dc.b	$79,$87,$F5,$6D,$91,$88,$F4,$36,$FD,$8D,$F0,$FF,$CB,$35,$C2,$B9
	.dc.b	$4F,$67,$F3,$78,$78,$AB,$78,$FE,$12,$7F,$3F,$D3,$EA,$7C,$5A,$FB
	.dc.b	$9F,$D1,$EF,$F3,$7E,$0B,$E8,$47,$E9,$B9,$3E,$86,$3F,$47,$E0,$BE
	.dc.b	$97,$1F,$CD,$AF,$2B,$FE,$00,$3E,$5F,$27,$E8,$31,$ED,$3F,$57,$2F
	.dc.b	$13,$FA,$6C,$FC,$CF,$D9,$D3,$D4,$FE,$8B,$5F,$73,$F1,$8F,$82,$43
	.dc.b	$04,$E0,$F9,$2E,$5A,$E7,$4F,$97,$BF,$49,$FB,$13,$F2,$8F,$AD,$7E
	.dc.b	$48,$FD,$06,$1E,$87,$F2,$18,$FB,$1F,$D2,$65,$F2,$6F,$E4,$10,$FC
	.dc.b	$A6,$3F,$25,$FC,$16,$9F,$25,$FF,$5F,$F0,$7E,$51,$FC,$CF,$C1,$FE
	.dc.b	$00,$FE,$02,$FF,$18,$ED,$FE,$02,$FE,$2F,$E1,$FF,$49,$F0,$9F,$AF
	.dc.b	$8F,$D1,$7C,$3F,$E9,$FE,$17,$F6,$D9,$FF,$A8,$BC,$0F,$DC,$D7,$EF
	.dc.b	$23,$1E,$06,$A3,$F8,$8B,$79,$7F,$8C,$B6,$B8,$CF,$F0,$52,$70,$7E
	.dc.b	$CF,$0F,$DB,$63,$E3,$8F,$DB,$65,$FB,$9C,$FD,$37,$FC,$9B,$4F,$6B
	.dc.b	$FB,$9D,$7F,$80,$BF,$CD,$BB,$BF,$80,$BF,$A9,$F8,$2F,$E0,$5F,$07
	.dc.b	$F9,$28,$FC,$96,$FF,$BD,$E2,$FC,$96,$1F,$A8,$F8,$5F,$E1,$1F,$0B
	.dc.b	$FC,$27,$DC,$FE,$15,$F0,$B3,$1F,$F2,$4F,$A1,$FE,$5F,$E1,$FF,$3B
	.dc.b	$2F,$57,$FC,$A4,$FF,$98,$27,$FC,$12,$FF,$97,$DE,$3F,$09,$AF,$80
	.dc.b	$DF,$C2,$C9,$EA,$D7,$8B,$70,$EE,$DE,$AF,$F6,$BE,$08,$E9,$5D,$8F
	.dc.b	$3F,$B1,$E4,$F4,$BC,$B9,$F9,$BE,$2B,$D9,$5F,$C8,$F5,$F8,$0F,$F1
	.dc.b	$1D,$FF,$95,$70,$7B,$3E,$0B,$DB,$F0,$3E,$EF,$6B,$DF,$EB,$7C,$0F
	.dc.b	$73,$F2,$DD,$1F,$D9,$DF,$9B,$97,$E6,$32,$F9,$AF,$CA,$F7,$FE,$3F
	.dc.b	$C2,$7C,$8F,$80,$F9,$3F,$05,$E6,$AF,$49,$0F,$97,$D3,$F3,$BE,$1B
	.dc.b	$E1,$43,$3F,$EE,$DC,$D8,$6C,$7C,$C2,$1F,$30,$BD,$3F,$31,$FB,$B5
	.dc.b	$F6,$AF,$ED,$F3,$C9,$FC,$0F,$37,$99,$FC,$1E,$7E,$A7,$D8,$7C,$92
	.dc.b	$1F,$82,$F8,$5F,$92,$FE,$2B,$8B,$E4,$BF,$8C,$E1,$F9,$3D,$3F,$E2
	.dc.b	$78,$FE,$A1,$DB,$FC,$F3,$8F,$E4,$C7,$E3,$38,$FC,$0F,$E3,$B2,$F2
	.dc.b	$3F,$90,$CF,$D0,$FE,$4B,$4F,$63,$EE,$3E,$0F,$5F,$E2,$B8,$BE,$46
	.dc.b	$7F,$8B,$9C,$9F,$35,$C3,$F2,$49,$C3,$F2,$5F,$CF,$F0,$F6,$7C,$07
	.dc.b	$D9,$8F,$D3,$E3,$E6,$7F,$53,$97,$A9,$FD,$5E,$7E,$E7,$FB,$C6,$9F
	.dc.b	$AC,$E0,$F9,$38,$7E,$B3,$83,$E4,$BF,$AE,$E0,$F9,$27,$E4,$38,$BF
	.dc.b	$BD,$74,$7E,$D3,$E1,$7F,$3B,$5F,$9F,$3F,$B7,$75,$7E,$8D,$FD,$EF
	.dc.b	$67,$E9,$DF,$DF,$FC,$37,$EB,$1F,$DD,$F7,$7E,$FB,$E1,$A4,$7F,$04
	.dc.b	$6B,$FC,$01,$BB,$FF,$06,$6F,$37,$EE,$1E,$91,$FC,$2A,$BF,$84,$DF
	.dc.b	$86,$FF,$84,$E1,$E5,$7D,$04,$AF,$F0,$E7,$81,$ED,$8F,$F0,$46,$9F
	.dc.b	$B7,$7F,$DB,$FF,$C0,$1F,$E1,$DA,$66,$9F,$D4,$4B,$E0,$BF,$DD,$4F
	.dc.b	$F9,$C2,$3F,$28,$7F,$28,$E8,$FC,$A3,$FD,$40,$E3,$F0,$9C,$3E,$1A
	.dc.b	$47,$82,$79,$FC,$47,$4F,$85,$EB,$F0,$BD,$BE,$17,$BF,$AB,$E0,$F8
	.dc.b	$3F,$0F,$9F,$F1,$B2,$FC,$87,$35,$E9,$FF,$39,$37,$43,$4D,$6E,$BE
	.dc.b	$97,$F0,$B1,$F0,$3F,$13,$3F,$23,$F1,$B5,$F8,$38,$FA,$A3,$FD,$83
	.dc.b	$83,$6F,$8B,$BF,$C6,$F8,$4F,$8F,$EC,$FC,$8F,$81,$F9,$3F,$0B,$FC
	.dc.b	$8B,$03,$4F,$CA,$2F,$7B,$B7,$3F,$D4,$CF,$E9,$E7,$E2,$7E,$CF,$B5
	.dc.b	$F5,$3D,$5F,$AB,$F0,$1F,$5B,$E1,$7E,$4D,$7D,$EE,$7F,$B8,$FF,$B6
	.dc.b	$3F,$80,$7F,$CC,$3D,$9F,$8F,$8F,$CB,$10,$FC,$B1,$73,$FC,$B3,$63
	.dc.b	$83,$A3,$A3,$F3,$1F,$8B,$CD,$E3,$7F,$1F,$8F,$9D,$F4,$9F,$83,$3F
	.dc.b	$27,$A7,$C0,$AF,$C9,$74,$B7,$F8,$95,$F9,$5C,$7E,$24,$B2,$FF,$3C
	.dc.b	$C6,$29,$C5,$F4,$6F,$F3,$FC,$3F,$3D,$FE,$FF,$3F,$49,$FE,$56,$7E
	.dc.b	$83,$87,$C4,$FD,$03,$CC,$7E,$03,$87,$D4,$FE,$0B,$5F,$73,$F0,$0F
	.dc.b	$A0,$79,$F4,$F5,$E7,$F4,$23,$F3,$27,$CB,$29,$F2,$F3,$F9,$ED,$9F
	.dc.b	$9F,$18,$23,$D0,$FA,$0F,$C1,$DB,$C6,$FC,$3B,$F3,$BF,$17,$0F,$C5
	.dc.b	$5F,$E6,$B6,$F7,$ED,$F3,$F3,$FC,$3D,$FC,$EC,$FE,$6E,$7F,$53,$4F
	.dc.b	$9B,$8F,$F0,$AE,$2F,$CE,$BF,$33,$6F,$9C,$FF,$03,$CB,$ED,$72,$74
	.dc.b	$C3,$E1,$FA,$D6,$BF,$BB,$64,$F8,$6F,$E0,$B1,$FB,$13,$F8,$BF,$6F
	.dc.b	$F0,$67,$F8,$CF,$6F,$1C,$BF,$05,$F0,$17,$07,$C3,$7F,$08,$BF,$A4
	.dc.b	$76,$F8,$2D,$11,$CD,$FE,$2E,$31,$1F,$A7,$6F,$0F,$C3,$6E,$BF,$0F
	.dc.b	$BF,$C1,$87,$87,$2F,$E1,$78,$7F,$06,$EF,$FE,$16,$FF,$08,$EF,$F6
	.dc.b	$BF,$B3,$DF,$D1,$C1,$FC,$2A,$3F,$9A,$F7,$7A,$FE,$1B,$D9,$F0,$9F
	.dc.b	$3A,$7F,$DA,$9C,$1F,$A5,$E5,$F9,$F8,$7F,$AA,$38,$9F,$0D,$F0,$FD
	.dc.b	$69,$87,$BA,$5F,$01,$F1,$FE,$07,$E4,$7C,$17,$C9,$F8,$3F,$95,$F0
	.dc.b	$7F,$2F,$E0,$FE,$67,$99,$F3,$7D,$6F,$9D,$E7,$4D,$3A,$35,$FE,$1C
	.dc.b	$DE,$DF,$C5,$37,$B3,$F4,$6F,$39,$F2,$5F,$77,$CD,$8F,$8D,$F4,$67
	.dc.b	$E7,$7D,$5A,$FB,$1F,$E3,$9B,$FE,$07,$83,$F0,$5E,$97,$E1,$7F,$80
	.dc.b	$7F,$CE,$9A,$7F,$0C,$7F,$CE,$97,$FE,$80,$6B,$F3,$1A,$FF,$C0,$07
	.dc.b	$3C,$7F,$36,$FC,$1B,$7E,$97,$1F,$C8,$65,$DF,$D1,$F0,$7B,$BE,$9C
	.dc.b	$97,$FC,$63,$27,$BD,$EF,$9C,$E9,$E5,$F5,$7D,$1E,$F7,$A7,$E0,$E5
	.dc.b	$7E,$5B,$04,$62,$38,$9F,$E5,$F8,$CB,$B8,$D6,$3E,$16,$E1,$F0,$EF
	.dc.b	$E3,$61,$E3,$AF,$36,$3E,$97,$D1,$97,$B7,$1F,$A2,$7E,$2F,$5F,$87
	.dc.b	$8F,$F1,$ED,$BE,$8C,$F9,$FE,$13,$E9,$7B,$FF,$E8,$2D,$BE,$6C,$FD
	.dc.b	$38,$C1,$FE,$35,$EB,$FC,$CC,$75,$21,$FD,$67,$87,$C4,$FC,$0E,$D5
	.dc.b	$FF,$AA,$E3,$FA,$F7,$5F,$E0,$77,$F6,$69,$F3,$67,$E5,$70,$FC,$27
	.dc.b	$E4,$71,$5C,$79,$B7,$1F,$97,$3D,$DB,$C3,$F2,$E3,$E7,$1D,$0F,$AD
	.dc.b	$C5,$E0,$7E,$A6,$5F,$2F,$2F,$C9,$71,$7A,$AF,$F9,$76,$BF,$2A,$FF
	.dc.b	$91,$F1,$7E,$6D,$FD,$C7,$2F,$E6,$F8,$BE,$5C,$7E,$67,$8B,$E4,$B4
	.dc.b	$FE,$B7,$BF,$C4,$F8,$4F,$D1,$7B,$BF,$A3,$F8,$4F,$D2,$7B,$DF,$AB
	.dc.b	$BF,$F0,$27,$57,$E9,$FE,$03,$F8,$0C,$7E,$3F,$E1,$3F,$80,$BF,$C2
	.dc.b	$7C,$0F,$DD,$3F,$CE,$3C,$0B,$A7,$F7,$36,$F5,$9A,$62,$B9,$D7,$CB
	.dc.b	$BE,$EF,$CD,$C3,$F8,$9B,$FC,$5F,$C2,$F3,$BF,$A1,$D3,$D6,$FE,$8F
	.dc.b	$6F,$DF,$7C,$17,$EF,$FE,$13,$E2,$CF,$EE,$F8,$7E,$2B,$4F,$8B,$D6
	.dc.b	$F8,$4F,$E0,$DF,$09,$FC,$1F,$E1,$3F,$84,$7C,$27,$F0,$9F,$83,$FE
	.dc.b	$9D,$E0,$7F,$99,$FB,$FF,$C2,$AF,$F9,$97,$2F,$E9,$2F,$F9,$95,$FC
	.dc.b	$A2,$F5,$F8,$FB,$7C,$BE,$0F,$9F,$A5,$2F,$8B,$5E,$2D,$F0,$BF,$C3
	.dc.b	$F8,$23,$18,$F8,$9E,$95,$E2,$C3,$CF,$8F,$A3,$E2,$3D,$3F,$0B,$EA
	.dc.b	$AF,$61,$FC,$2F,$5F,$80,$FF,$0D,$DF,$F9,$47,$07,$C0,$F6,$FE,$0F
	.dc.b	$AD,$FE,$FA,$D8,$D2,$5F,$F9,$38,$F8,$AF,$FA,$BF,$8F,$E7,$47,$F1
	.dc.b	$9E,$3E,$EF,$EA,$F2,$F9,$BE,$DF,$96,$3E,$5E,$9E,$97,$E6,$FC,$0F
	.dc.b	$D2,$F8,$3F,$7C,$7C,$EF,$73,$EA,$91,$FA,$6D,$74,$60,$86,$0F,$AF
	.dc.b	$CD,$D5,$EE,$7D,$CE,$1F,$92,$FD,$E3,$CE,$7C,$93,$D8,$43,$F8,$37
	.dc.b	$07,$C9,$7F,$0B,$B7,$C6,$7F,$BC,$EF,$F2,$9F,$90,$3C,$34,$66,$37
	.dc.b	$3F,$2E,$BE,$4F,$3F,$F4,$BE,$0F,$CB,$D7,$E4,$B7,$F9,$4F,$E4,$B7
	.dc.b	$F9,$4F,$E5,$37,$F7,$3F,$00,$F8,$77,$F8,$ED,$BE,$43,$63,$3C,$FF
	.dc.b	$A0,$9C,$B6,$32,$7A,$74,$6F,$DA,$FF,$33,$87,$89,$FC,$DE,$3E,$67
	.dc.b	$F3,$B9,$7A,$B0,$FE,$5D,$9F,$BB,$2F,$EF,$1A,$7C,$B9,$FD,$6E,$BF
	.dc.b	$2D,$FF,$4E,$ED,$F2,$23,$11,$F9,$23,$F9,$1E,$AF,$67,$F7,$F1,$FB
	.dc.b	$8F,$83,$FD,$EF,$BF,$FA,$48,$FD,$CF,$B3,$FC,$03,$E0,$7F,$80,$FC
	.dc.b	$17,$EB,$27,$F5,$C7,$F0,$0F,$85,$FD,$9B,$F0,$7B,$AE,$BF,$69,$BA
	.dc.b	$7E,$E6,$7F,$7E,$43,$F7,$E7,$33,$A9,$FC,$2B,$B7,$C2,$FF,$0E,$C7
	.dc.b	$F8,$9E,$5E,$7E,$3F,$E0,$BA,$7F,$16,$D7,$DB,$7F,$C4,$B6,$FE,$33
	.dc.b	$F0,$9F,$BA,$8F,$C7,$EB,$F1,$DB,$A7,$E3,$DF,$96,$D7,$E5,$1F,$E7
	.dc.b	$B5,$F0,$BF,$E7,$1F,$E0,$0F,$F4,$E5,$1F,$F4,$3D,$7E,$6D,$7B,$9B
	.dc.b	$F8,$2E,$1F,$07,$C7,$E1,$39,$7C,$2F,$3F,$85,$E9,$F0,$BD,$7E,$07
	.dc.b	$B7,$B7,$DF,$DC,$9A,$FF,$9E,$35,$FF,$2D,$B6,$32,$FF,$81,$63,$2D
	.dc.b	$8C,$F2,$FF,$B2,$A5,$5D,$EB,$A6,$1E,$37,$B6,$5E,$77,$C1,$A7,$AD
	.dc.b	$F1,$6D,$F0,$FE,$27,$E2,$77,$FF,$48,$E2,$F8,$DE,$9F,$EF,$AF,$E2
	.dc.b	$37,$8F,$E4,$37,$53,$4F,$F7,$CE,$3E,$6A,$FD,$75,$F6,$7E,$16,$3F
	.dc.b	$4F,$D9,$F2,$47,$C1,$CF,$EB,$7C,$17,$D7,$F8,$4F,$64,$FC,$5D,$FE
	.dc.b	$6B,$F2,$F8,$3F,$2F,$DF,$23,$F2,$DC,$5F,$D4,$A3,$11,$F3,$67,$8B
	.dc.b	$CC,$FA,$9C,$BD,$DF,$AD,$8F,$E0,$1F,$B1,$CF,$E7,$7D,$27,$DB,$E3
	.dc.b	$F9,$0F,$DD,$E1,$F8,$4F,$E6,$38,$BE,$2B,$63,$2F,$E5,$78,$7E,$23
	.dc.b	$F8,$1E,$1E,$CF,$E0,$F0,$F1,$3F,$85,$C7,$CC,$FE,$1F,$2F,$C8,$BF
	.dc.b	$0C,$87,$ED,$89,$AF,$C3,$7E,$2E,$7F,$DC,$DA,$7F,$03,$2F,$1F,$E9
	.dc.b	$75,$F8,$F5,$F9,$1D,$FE,$9B,$F9,$3E,$0F,$AF,$7F,$8E,$3E,$39,$F9
	.dc.b	$FE,$2F,$C7,$3F,$1C,$FF,$08,$ED,$F8,$E7,$E2,$10,$FF,$63,$F0,$7C
	.dc.b	$86,$F0,$7E,$33,$24,$7E,$DF,$E1,$BF,$79,$F0,$DF,$BD,$F6,$FF,$7D
	.dc.b	$CD,$FB,$FF,$73,$F8,$07,$C0,$FE,$8F,$0F,$D7,$74,$FF,$BC,$FF,$80
	.dc.b	$7F,$D5,$F1,$FD,$6D,$FF,$56,$F0,$FE,$BF,$B3,$F6,$3C,$BE,$0E,$5F
	.dc.b	$E1,$1D,$98,$C3,$F8,$43,$63,$31,$FB,$D2,$1F,$BB,$2D,$F5,$7F,$DF
	.dc.b	$78,$78,$2F,$B3,$E4,$AF,$E2,$D8,$FA,$6F,$F8,$16,$5E,$DC,$3F,$76
	.dc.b	$7F,$0D,$D7,$E1,$BF,$C3,$F7,$F9,$CF,$CA,$3E,$71,$FD,$4B,$8B,$E7
	.dc.b	$37,$8B,$E7,$3F,$0F,$E1,$FF,$80,$69,$FE,$97,$C9,$AF,$9B,$E1,$7E
	.dc.b	$57,$81,$F2,$FE,$1F,$E6,$7C,$3F,$CD,$F7,$FE,$77,$C2,$CD,$3D,$FF
	.dc.b	$03,$F4,$7E,$17,$E9,$7C,$1F,$D3,$F8,$49,$5F,$11,$AF,$FA,$A7,$74
	.dc.b	$63,$7F,$1C,$FD,$98,$C6,$FF,$2A,$FB,$4F,$80,$F9,$F9,$78,$DE,$99
	.dc.b	$FA,$1F,$A7,$AF,$B1,$FA,$BB,$FE,$1F,$E1,$FF,$11,$C3,$F1,$09,$3F
	.dc.b	$D5,$B5,$46,$23,$FC,$9D,$D0,$9C,$C7,$FA,$81,$AE,$9F,$9D,$A7,$CD
	.dc.b	$8F,$C2,$FC,$2F,$E5,$FE,$17,$F3,$1E,$B7,$CD,$9F,$C4,$FC,$17,$E6
	.dc.b	$FD,$8F,$CE,$7C,$07,$C0,$9F,$91,$C5,$F9,$FF,$84,$91,$F9,$06,$BF
	.dc.b	$35,$BC,$BF,$33,$6F,$44,$63,$5F,$99,$5F,$98,$E6,$EE,$FE,$6B,$2F
	.dc.b	$D6,$67,$E5,$8F,$B1,$A7,$A5,$F5,$9F,$30,$FD,$0F,$0F,$CC,$7E,$57
	.dc.b	$0F,$C9,$7F,$C4,$FC,$3F,$19,$B1,$F3,$7C,$3F,$E9,$DA,$FC,$D8,$FC
	.dc.b	$BD,$FE,$F7,$8F,$F7,$DC,$7E,$1C,$3C,$67,$EC,$39,$7C,$EF,$A4,$FD
	.dc.b	$9F,$1F,$C9,$E2,$F8,$5A,$38,$FE,$15,$7F,$07,$E3,$BA,$FC,$68,$F9
	.dc.b	$84,$3E,$61,$6F,$9D,$75,$E3,$FD,$FE,$3F,$66,$BE,$D9,$E5,$3F,$82
	.dc.b	$F0,$7A,$6F,$C7,$A7,$B6,$7E,$49,$FC,$2B,$7F,$90,$FF,$0A,$6B,$FB
	.dc.b	$B3,$E6,$9F,$91,$7F,$BE,$7F,$00,$FF,$B1,$E3,$FC,$F8,$FF,$B1,$DF
	.dc.b	$D8,$75,$FE,$F5,$FE,$2F,$D7,$FC,$C9,$61,$F8,$7A,$FF,$66,$3F,$23
	.dc.b	$A7,$F5,$CF,$CA,$EB,$E0,$EF,$D8,$BB,$7F,$2D,$B7,$C2,$FE,$EB,$1F
	.dc.b	$DE,$1F,$C0,$C8,$7F,$03,$ED,$D7,$17,$EB,$5B,$5C,$A7,$FA,$3E,$3D
	.dc.b	$DF,$E0,$47,$F0,$BC,$FC,$B8,$F8,$7E,$0B,$F8,$14,$FF,$0F,$EE,$FE
	.dc.b	$04,$FB,$A7,$CD,$EF,$FA,$23,$E3,$9F,$99,$3F,$75,$C3,$F4,$34,$FC
	.dc.b	$A4,$7D,$03,$27,$F5,$AE,$BF,$CC,$E9,$FE,$92,$8F,$F2,$FB,$5F,$E1
	.dc.b	$2F,$F9,$CD,$F1,$1F,$5B,$E1,$7C,$54,$9F,$F5,$96,$BF,$4F,$E0,$33
	.dc.b	$0F,$81,$BF,$09,$8F,$84,$CF,$B9,$AF,$77,$7E,$F7,$0F,$7E,$6D,$F2
	.dc.b	$B8,$BE,$B4,$62,$3C,$7B,$78,$63,$1B,$7E,$0D,$B5,$F7,$6F,$B5,$F3
	.dc.b	$3F,$1F,$8F,$8D,$C6,$1E,$BC,$FD,$3A,$7A,$B6,$F5,$FC,$1F,$B3,$83
	.dc.b	$DB,$EE,$FB,$BE,$13,$F3,$64,$CB,$E5,$B5,$F9,$6D,$E3,$FA,$2E,$63
	.dc.b	$FD,$6B,$C7,$F3,$F8,$7F,$39,$87,$82,$7C,$B2,$CB,$E6,$FC,$17,$CC
	.dc.b	$9C,$FC,$27,$CC,$77,$F0,$9F,$31,$E3,$F0,$9F,$50,$8E,$08,$7C,$CC
	.dc.b	$11,$A8,$C3,$CD,$D4,$F7,$78,$57,$8D,$E5,$7F,$25,$F4,$ED,$F2,$5F
	.dc.b	$5E,$DF,$25,$F8,$E4,$3E,$39,$4F,$8B,$AF,$C9,$F6,$BF,$70,$FC,$98
	.dc.b	$FD,$D4,$F5,$7D,$55,$E1,$7C,$67,$CA,$E1,$FC,$AE,$3E,$9D,$FF,$85
	.dc.b	$67,$ED,$7F,$86,$69,$F0,$5B,$7F,$19,$AF,$C4,$31,$CB,$FC,$2E,$79
	.dc.b	$3F,$C3,$6B,$B3,$FC,$3E,$FE,$35,$FF,$19,$D7,$CC,$FF,$1B,$CB,$D4
	.dc.b	$FF,$1D,$CF,$E3,$65,$F1,$08,$7C,$43,$E3,$1F,$10,$F9,$5C,$3F,$8E
	.dc.b	$72,$7F,$2E,$F6,$FF,$73,$F0,$9F,$ED,$43,$FD,$A7,$CA,$7F,$DA,$73
	.dc.b	$F9,$0E,$5F,$D3,$3F,$85,$EA,$FD,$5B,$F8,$7E,$AE,$7F,$05,$FB,$08
	.dc.b	$FE,$77,$D7,$71,$FD,$B4,$7E,$6A,$31,$CE,$75,$92,$FA,$ED,$FC,$F7
	.dc.b	$1F,$0B,$FC,$FF,$2F,$DA,$E3,$FB,$8B,$F3,$BF,$C1,$1F,$59,$FE,$F0
	.dc.b	$C7,$FD,$E7,$1F,$AA,$E6,$FC,$43,$FD,$41,$B9,$FF,$1B,$6C,$FF,$23
	.dc.b	$DB,$FB,$FD,$BE,$13,$FD,$A8,$DC,$3F,$AA,$1F,$E7,$77,$0F,$F8,$C9
	.dc.b	$F9,$8F,$9F,$DA,$FA,$3F,$05,$F4,$BE,$0B,$E9,$FB,$DF,$53,$B3,$EA
	.dc.b	$FB,$DF,$5B,$DA,$FA,$FF,$03,$F6,$3E,$0F,$EC,$FC,$1C,$D3,$E1,$35
	.dc.b	$F8,$4D,$E4,$FF,$16,$6C,$8E,$53,$F2,$6F,$B3,$F2,$F1,$F1,$3F,$37
	.dc.b	$3F,$33,$F3,$F5,$FC,$4E,$DF,$8A,$C7,$F1,$7B,$FC,$42,$57,$C4,$1C
	.dc.b	$7F,$8D,$8C,$C7,$C5,$6F,$0F,$CE,$73,$1F,$DF,$E7,$83,$CD,$FC,$B5
	.dc.b	$78,$A3,$F0,$5C,$3E,$67,$F0,$5E,$FF,$E7,$BE,$0F,$F3,$FF,$03,$FA
	.dc.b	$0F,$83,$FD,$0F,$C1,$FC,$2B,$FE,$97,$FB,$FF,$FA,$84,$9F,$13,$FD
	.dc.b	$79,$B1,$FD,$19,$AF,$E3,$F9,$B9,$BF,$93,$BE,$EF,$E5,$71,$FD,$7E
	.dc.b	$5E,$5E,$6F,$A7,$A7,$A6,$BF,$31,$AF,$CA,$7D,$CF,$E2,$9F,$84,$D7
	.dc.b	$F1,$DB,$FC,$66,$9C,$1F,$F5,$5E,$FF,$22,$3E,$71,$93,$AC,$F8,$6F
	.dc.b	$C7,$1F,$28,$FD,$66,$1F,$25,$FD,$76,$FF,$25,$FD,$86,$FF,$26,$BF
	.dc.b	$CF,$1B,$7C,$46,$EF,$F2,$49,$5F,$25,$AF,$C9,$39,$9F,$24,$EE,$7C
	.dc.b	$93,$C6,$79,$4F,$DF,$69,$E9,$F1,$7F,$61,$9F,$C9,$9F,$79,$0F,$93
	.dc.b	$AF,$C7,$7F,$27,$1F,$2D,$B1,$F2,$23,$1A,$FC,$99,$E2,$7E,$47,$2F
	.dc.b	$F7,$43,$F9,$C7,$FD,$57,$97,$EB,$B8,$FF,$77,$3F,$A9,$3F,$8B,$74
	.dc.b	$FE,$F1,$FE,$31,$D3,$FB,$C7,$F1,$9F,$01,$FB,$7D,$3F,$85,$75,$59
	.dc.b	$F3,$B7,$3F,$E6,$CD,$CF,$FD,$73,$3E,$CE,$AF,$DD,$BF,$C8,$FA,$BF
	.dc.b	$76,$FF,$25,$EA,$F2,$BF,$C9,$F5,$FC,$F7,$9D,$F9,$FF,$88,$F9,$53
	.dc.b	$F0,$0F,$94,$7E,$38,$A7,$C8,$3F,$18,$3A,$FF,$D4,$B9,$3E,$07,$F8
	.dc.b	$14,$FC,$76,$EA,$9D,$4F,$CD,$9C,$C3,$DA,$FD,$97,$A7,$FB,$3F,$6B
	.dc.b	$F6,$9F,$03,$FB,$5F,$5F,$F6,$DE,$DF,$ED,$FE,$13,$F7,$1F,$01,$FB
	.dc.b	$9F,$6F,$F7,$5E,$F4,$DB,$F7,$4D,$7F,$80,$B7,$B3,$F3,$6F,$18,$FE
	.dc.b	$01,$5D,$5F,$E0,$58,$78,$5F,$E0,$99,$79,$5F,$E0,$DA,$7F,$09,$D7
	.dc.b	$F8,$56,$FE,$C2,$4F,$BC,$87,$FB,$E3,$7F,$86,$D8,$DB,$F3,$7B,$0F
	.dc.b	$CB,$38,$9F,$C9,$E7,$8B,$67,$E5,$D7,$CA,$8F,$E3,$5C,$1F,$C7,$7E
	.dc.b	$17,$F8,$F7,$C0,$7F,$1F,$F8,$4F,$E4,$1F,$0B,$F2,$AF,$E3,$90,$F8
	.dc.b	$E5,$E1,$F8,$ED,$EE,$F2,$F3,$7D,$DE,$3F,$99,$18,$D9,$1D,$6B,$F9
	.dc.b	$8E,$1E,$17,$F9,$76,$5E,$57,$CE,$7C,$C3,$F9,$96,$BF,$31,$F8,$27
	.dc.b	$5D,$BE,$3B,$F0,$4C,$9F,$1C,$F9,$87,$FA,$8F,$7E,$0E,$B3,$FC,$A4
	.dc.b	$F2,$8F,$45,$7C,$97,$FA,$3E,$FE,$57,$FA,$4E,$1E,$97,$D6,$7E,$20
	.dc.b	$F8,$87,$F0,$7C,$BE,$1B,$4C,$E1,$F1,$DB,$9F,$C6,$6F,$1F,$D4,$7F
	.dc.b	$A9,$6F,$F6,$1F,$EA,$9C,$1F,$91,$7F,$AD,$F0,$FE,$45,$FE,$B1,$AF
	.dc.b	$AD,$F8,$47,$C5,$3E,$19,$0C,$19,$E7,$FD,$DC,$71,$CB,$E2,$CF,$C7
	.dc.b	$C9,$F0,$7F,$D9,$FE,$13,$FD,$8B,$FC,$03,$FE,$FD,$BF,$DF,$73,$7E
	.dc.b	$9F,$0F,$ED,$FC,$FF,$AB,$7F,$D7,$75,$FF,$00,$79,$FF,$A5,$F4,$72
	.dc.b	$F8,$39,$5F,$EB,$6A,$FF,$7F,$F4,$FE,$39,$B1,$FE,$0E,$6E,$46,$98
	.dc.b	$D5,$87,$9B,$E0,$FC,$59,$7F,$5A,$CF,$CC,$FF,$5C,$D3,$D7,$97,$B4
	.dc.b	$FF,$01,$6D,$F9,$17,$E4,$1F,$8A,$29,$F3,$8C,$1F,$30,$D7,$03,$A7
	.dc.b	$D5,$F0,$BF,$C9,$35,$34,$95,$FE,$21,$8C,$B7,$E1,$7E,$7F,$A9,$F4
	.dc.b	$3D,$BF,$A3,$F0,$7F,$4B,$DD,$FA,$7F,$07,$F5,$3D,$D9,$8F,$CB,$F8
	.dc.b	$3F,$AE,$47,$04,$3E,$AB,$97,$EB,$C7,$27,$EC,$CF,$77,$EB,$5F,$8D
	.dc.b	$FB,$18,$FE,$13,$2F,$C2,$E9,$EC,$9F,$F3,$4F,$37,$D7,$8C,$47,$F9
	.dc.b	$07,$5F,$7B,$A6,$B9,$8D,$BF,$28,$FD,$E1,$F8,$37,$2D,$3E,$53,$89
	.dc.b	$FC,$35,$7C,$A9,$FC,$77,$B7,$F9,$AF,$84,$F6,$47,$E5,$BD,$DF,$81
	.dc.b	$CA,$EA,$FC,$96,$FF,$16,$31,$5F,$93,$E0,$C5,$7E,$5D,$B1,$93,$E7
	.dc.b	$4F,$0C,$FF,$31,$7D,$AF,$F3,$D8,$78,$AB,$5A,$B2,$FD,$7E,$7E,$9C
	.dc.b	$BF,$57,$A7,$CE,$7D,$E4,$3E,$56,$BF,$06,$FE,$5C,$7C,$93,$26,$F9
	.dc.b	$3F,$2C,$FC,$93,$24,$3E,$91,$D4,$FB,$15,$E3,$C7,$CF,$87,$CB,$9F
	.dc.b	$DD,$71,$7F,$06,$CF,$DD,$1F,$B9,$D7,$E2,$ED,$FB,$9D,$EE,$86,$98
	.dc.b	$9E,$33,$FB,$FB,$F8,$DB,$FF,$05,$DF,$E2,$E3,$F9,$23,$F8,$56,$BF
	.dc.b	$15,$F5,$1F,$04,$F5,$98,$3E,$01,$82,$1F,$1F,$6F,$8B,$87,$E6,$63
	.dc.b	$E1,$1C,$C8,$72,$3F,$8C,$7C,$2F,$F2,$6F,$80,$FD,$34,$7F,$22,$E3
	.dc.b	$FE,$53,$F0,$5F,$B8,$8F,$E4,$1F,$07,$FC,$B3,$E1,$3F,$96,$FC,$2C
	.dc.b	$9F,$E4,$8D,$7F,$95,$B7,$97,$F0,$EF,$18,$E6,$7F,$2B,$E8,$EE,$FE
	.dc.b	$E0,$FE,$5D,$D3,$E5,$7F,$98,$69,$F1,$DF,$82,$7A,$C9,$AF,$C1,$7E
	.dc.b	$43,$5D,$B9,$DB,$E3,$F2,$FF,$BD,$27,$F7,$9F,$0B,$FC,$84,$8F,$F1
	.dc.b	$F6,$BF,$C1,$BF,$80,$7F,$5A,$B4,$FD,$44,$3E,$27,$FC,$4B,$39,$E5
	.dc.b	$FF,$09,$CF,$F9,$56,$3F,$C5,$35,$C3,$E1,$F6,$D8,$CD,$70,$F7,$72
	.dc.b	$F2,$F7,$6C,$FF,$B7,$30,$7C,$36,$36,$43,$D8,$C3,$6B,$A9,$C6,$32
	.dc.b	$47,$33,$CA,$38,$D7,$2B,$E7,$8F,$4C,$E6,$1C,$31,$F7,$3F,$04,$F3
	.dc.b	$69,$E2,$CF,$FB,$AE,$BF,$32,$7F,$97,$35,$FE,$D8,$D7,$FB,$63,$74
	.dc.b	$FE,$D8,$FF,$B3,$B6,$FB,$53,$C8,$FB,$04,$3E,$E9,$4F,$F7,$E6,$1F
	.dc.b	$6B,$6F,$B3,$8F,$C6,$DF,$D7,$5E,$9C,$7E,$5E,$7F,$06,$3D,$DA,$FC
	.dc.b	$47,$E0,$6F,$F2,$5E,$19,$7C,$F6,$C6,$72,$FC,$3B,$C6,$37,$5D,$5E
	.dc.b	$37,$E2,$AF,$C6,$E3,$F6,$B2,$F9,$55,$FE,$13,$D3,$E2,$69,$F0,$A3
	.dc.b	$E4,$9E,$F2,$9F,$08,$87,$C7,$3F,$CF,$87,$CE,$38,$14,$D9,$C8,$E0
	.dc.b	$7C,$AD,$7C,$8F,$E0,$6F,$D1,$1F,$A0,$C7,$D9,$8F,$E9,$B2,$F8,$0F
	.dc.b	$C2,$BF,$92,$43,$E3,$95,$D3,$F3,$37,$E7,$7F,$A3,$8F,$C8,$5F,$81
	.dc.b	$FD,$26,$1E,$47,$E9,$CF,$E4,$1F,$6E,$DF,$8A,$6E,$5F,$25,$F8,$07
	.dc.b	$B8,$F8,$07,$CA,$AF,$E1,$39,$FD,$02,$47,$4D,$9E,$FF,$EC,$3D,$AF
	.dc.b	$D8,$F8,$9F,$B2,$F6,$FF,$67,$EF,$FE,$AB,$0F,$D4,$7B,$BF,$B6,$F8
	.dc.b	$6F,$DB,$FB,$F2,$7F,$58,$D7,$F1,$AD,$E3,$FE,$23,$3F,$BE,$AE,$95
	.dc.b	$F9,$2B,$F0,$3F,$B5,$C7,$C8,$FE,$DF,$3F,$88,$FD,$E3,$E0,$9F,$00
	.dc.b	$A4,$E0,$F9,$0E,$DB,$BF,$F5,$2D,$9F,$09,$FE,$EE,$6D,$7F,$24,$3F
	.dc.b	$91,$C6,$48,$E4,$FE,$38,$E1,$FE,$61,$F0,$BF,$C6,$BE,$17,$F8,$DF
	.dc.b	$C3,$7F,$1C,$F8,$6F,$E4,$1D,$CF,$86,$BF,$09,$8F,$85,$F8,$DD,$5F
	.dc.b	$CC,$B9,$AE,$9F,$CA,$5B,$C1,$FC,$A1,$E1,$1F,$C8,$B1,$61,$DB,$1F
	.dc.b	$04,$FF,$1F,$CF,$C8,$FF,$21,$D7,$F9,$7E,$DE,$98,$FE,$1E,$D7,$E0
	.dc.b	$9E,$FC,$BF,$9B,$C6,$63,$E0,$99,$DF,$E4,$BF,$D4,$63,$FA,$6F,$07
	.dc.b	$CF,$78,$57,$29,$E1,$18,$AE,$31,$E2,$3A,$5F,$A1,$FE,$2F,$BF,$B1
	.dc.b	$FE,$41,$A7,$CB,$7E,$16,$9F,$11,$FC,$89,$F9,$42,$18,$27,$37,$F6
	.dc.b	$4D,$B8,$3F,$E4,$DA,$F9,$73,$F3,$76,$F0,$4F,$F6,$AC,$3C,$8F,$F6
	.dc.b	$CC,$7E,$5D,$7B,$08,$7C,$32,$9E,$D3,$04,$30,$67,$5F,$90,$F0,$3E
	.dc.b	$99,$B2,$9C,$0A,$7F,$7A,$8F,$2E,$FF,$DF,$2F,$D2,$A5,$B3,$2F,$95
	.dc.b	$1F,$33,$4F,$87,$5F,$93,$8F,$93,$97,$C7,$6C,$65,$F9,$6F,$5B,$E9
	.dc.b	$C1,$F9,$E9,$F9,$47,$F7,$1C,$7C,$6F,$F8,$17,$87,$FC,$03,$9F,$CA
	.dc.b	$8F,$71,$F8,$E2,$DF,$E2,$5B,$BF,$E2,$1F,$82,$6C,$F9,$5C,$78,$9E
	.dc.b	$11,$F1,$CD,$19,$9E,$4E,$0F,$8B,$AB,$E1,$BF,$73,$ED,$FE,$F2,$7F
	.dc.b	$61,$F0,$7F,$BC,$F7,$7F,$7B,$F0,$7F,$BE,$F6,$3F,$7F,$F0,$D2,$FF
	.dc.b	$6C,$D7,$F3,$2D,$E3,$FC,$84,$F2,$9F,$DE,$5F,$F0,$8C,$3C,$11,$FA
	.dc.b	$0C,$7F,$85,$E5,$F1,$E3,$D9,$AF,$B9,$F6,$74,$7E,$B4,$8F,$E5,$39
	.dc.b	$BF,$CE,$7B,$3D,$AF,$E3,$93,$F0,$1B,$7F,$85,$6C,$EE,$7F,$82,$35
	.dc.b	$FF,$6A,$B5,$FF,$26,$CF,$17,$73,$CB,$E1,$BF,$91,$FC,$1F,$F2,$4F
	.dc.b	$7B,$F9,$2F,$C2,$7F,$26,$F8,$3F,$E4,$FF,$0B,$FC,$A3,$E1,$66,$5F
	.dc.b	$C7,$DA,$FF,$02,$6F,$27,$EE,$67,$94,$FF,$25,$BE,$CF,$F2,$7C,$7F
	.dc.b	$99,$E5,$FC,$D3,$3F,$69,$27,$CE,$43,$DA,$E5,$A6,$4F,$7E,$BF,$05
	.dc.b	$FD,$13,$F1,$76,$F9,$F3,$C6,$76,$D7,$6D,$7E,$91,$D4,$E7,$5F,$21
	.dc.b	$FE,$81,$AF,$F4,$CF,$86,$F7,$C7,$F2,$4F,$82,$FE,$9F,$F0,$D2,$3F
	.dc.b	$A3,$35,$FE,$52,$DD,$BF,$90,$4F,$27,$F9,$15,$7C,$8A,$FE,$5F,$A7
	.dc.b	$C8,$7F,$A8,$65,$E7,$E1,$FB,$D1,$F1,$08,$7B,$8F,$51,$EC,$32,$7C
	.dc.b	$03,$E7,$69,$F1,$9D,$C7,$47,$91,$C0,$E4,$60,$87,$C6,$3E,$F1,$F1
	.dc.b	$8F,$EC,$D7,$ED,$F0,$3E,$AE,$5F,$05,$FA,$F9,$DA,$FE,$02,$DB,$FE
	.dc.b	$4C,$FF,$7A,$AE,$53,$FC,$1A,$FE,$36,$7F,$D2,$73,$F8,$D3,$FD,$5B
	.dc.b	$3F,$8D,$B7,$E1,$63,$D2,$7C,$13,$D9,$97,$BD,$C1,$EC,$36,$7C,$9D
	.dc.b	$3E,$33,$C2,$7A,$4F,$18,$DB,$49,$1A,$FE,$01,$FF,$0F,$62,$D7,$F9
	.dc.b	$C4,$3E,$17,$FD,$6C,$BE,$1F,$F6,$B8,$FF,$AC,$21,$F0,$BF,$B6,$8F
	.dc.b	$DA,$10,$FD,$A1,$70,$FE,$D9,$CB,$FB,$6C,$7F,$4B,$CB,$DA,$FA,$DF
	.dc.b	$FA,$6F,$0F,$03,$DF,$2F,$1B,$FB,$73,$F0,$B9,$FC,$02,$6D,$F3,$89
	.dc.b	$1A,$7F,$C2,$7E,$07,$BB,$DF,$FE,$3B,$81,$A4,$AF,$F1,$04,$65,$B1
	.dc.b	$97,$FC,$8D,$BA,$75,$18,$6B,$FC,$89,$7C,$37,$CC,$C7,$E6,$FB,$FF
	.dc.b	$3B,$E1,$7E,$7F,$B9,$F4,$3E,$12,$63,$F3,$DA,$FC,$56,$F1,$FD,$98
	.dc.b	$FE,$F7,$92,$BA,$47,$D7,$C3,$C0,$FD,$9C,$BE,$EE,$9E,$B8,$F3,$35
	.dc.b	$F8,$C7,$C0,$21,$ED,$74,$F8,$1B,$A7,$C3,$7E,$7C,$64,$E2,$76,$29
	.dc.b	$D8,$E8,$60,$E0,$60,$F9,$93,$EA,$7F,$0D,$F0,$BF,$96,$F8,$6F,$CB
	.dc.b	$FC,$2F,$E6,$3E,$1A,$57,$E5,$DA,$FE,$35,$BA,$FC,$9C,$BE,$D5,$73
	.dc.b	$AF,$CF,$61,$DE,$B5,$AF,$DA,$9F,$D1,$67,$F0,$23,$EF,$C7,$AC,$F7
	.dc.b	$1E,$92,$90,$D1,$0F,$79,$C3,$A5,$C5,$9C,$BE,$64,$FC,$56,$BD,$4D
	.dc.b	$57,$89,$E6,$63,$5F,$57,$3F,$EF,$B3,$F7,$4F,$EB,$F4,$F8,$73,$FC
	.dc.b	$0B,$5B,$3F,$8C,$6E,$3F,$9F,$7E,$7C,$7F,$03,$DB,$F8,$3E,$1D,$A3
	.dc.b	$F2,$78,$FE,$39,$F4,$65,$E5,$C3,$D3,$1F,$0C,$F3,$91,$CB,$EC,$D7
	.dc.b	$E7,$1F,$8C,$8C,$C6,$E7,$A1,$27,$71,$E0,$7F,$7F,$F0,$3F,$C7,$3E
	.dc.b	$1B,$F6,$33,$FC,$6F,$93,$F6,$2F,$EB,$C8,$7E,$BC,$BC,$3F,$82,$E4
	.dc.b	$FD,$8E,$FF,$CA,$39,$3F,$5F,$5F,$C6,$79,$3B,$3F,$C9,$F2,$F8,$4F
	.dc.b	$D7,$3D,$19,$F9,$9C,$6B,$EC,$26,$BE,$B8,$FD,$8F,$27,$E1,$1F,$DF
	.dc.b	$FC,$37,$F1,$A2,$3F,$C6,$1A,$FF,$13,$6B,$FC,$4D,$AF,$F4,$07,$F9
	.dc.b	$1F,$C3,$7F,$40,$F8,$6F,$E8,$3F,$0D,$FD,$0B,$DF,$FF,$44,$5B,$87
	.dc.b	$F5,$1F,$C0,$3F,$C4,$63,$1C,$3F,$E8,$18,$C4,$EA,$38,$B6,$7F,$D2
	.dc.b	$4D,$78,$9F,$E9,$49,$E3,$3C,$CF,$14,$7F,$51,$BF,$33,$FD,$4F,$1F
	.dc.b	$EA,$F9,$7B,$23,$E2,$9F,$0C,$87,$C3,$29,$F0,$CA,$7C,$32,$9C,$86
	.dc.b	$7F,$47,$5F,$E8,$13,$DD,$B3,$C1,$AF,$72,$1D,$CA,$77,$3F,$A4,$57
	.dc.b	$B1,$FE,$C5,$7F,$01,$F8,$B2,$C5,$F0,$BF,$92,$9F,$E8,$3C,$76,$BF
	.dc.b	$AA,$36,$B7,$3F,$DE,$75,$FB,$AF,$F7,$BD,$BF,$BF,$E5,$F2,$A3,$D5
	.dc.b	$BB,$4F,$73,$63,$31,$82,$3F,$3B,$5F,$8D,$3B,$8F,$79,$F2,$CF,$84
	.dc.b	$43,$E1,$14,$D1,$82,$1F,$D7,$AB,$E3,$6F,$FC,$D3,$6F,$8C,$FE,$53
	.dc.b	$6F,$85,$6D,$69,$F0,$E3,$7B,$FF,$88,$2F,$9B,$FE,$22,$C3,$E1,$63
	.dc.b	$E8,$DF,$ED,$BE,$56,$BE,$46,$BE,$82,$38,$26,$3F,$0F,$3F,$8B,$A7
	.dc.b	$C8,$AE,$D3,$F1,$DE,$CD,$7B,$35,$EB,$BF,$E5,$1F,$D5,$E3,$FB,$1F
	.dc.b	$83,$FD,$7B,$FB,$3F,$87,$B3,$FA,$DF,$86,$B9,$FE,$2E,$AE,$1F,$B0
	.dc.b	$AF,$D9,$46,$08,$F6,$32,$59,$E4,$D9,$57,$4B,$EB,$87,$6C,$FE,$6D
	.dc.b	$FB,$B3,$F8,$4D,$3F,$1A,$43,$F5,$51,$FB,$1C,$1F,$0D,$FE,$D6,$6C
	.dc.b	$FF,$16,$3F,$89,$46,$48,$E4,$FE,$1C,$E1,$FF,$1D,$C7,$26,$C6,$5F
	.dc.b	$F6,$DC,$65,$BF,$0F,$F2,$3E,$07,$E4,$FC,$37,$CA,$F0,$BE,$5F,$C3
	.dc.b	$4B,$F8,$8D,$7F,$77,$82,$38,$47,$C9,$AE,$8F,$CB,$C3,$EA,$E3,$E3
	.dc.b	$8F,$69,$EB,$DF,$F8,$24,$66,$3D,$A6,$70,$F4,$3C,$0F,$69,$98,$E1
	.dc.b	$3C,$A3,$00,$FE,$43,$8F,$17,$E1,$CF,$0A,$EB,$1E,$03,$A9,$0D,$9D
	.dc.b	$0C,$1C,$CF,$E9,$17,$F0,$1F,$8B,$5F,$11,$F8,$FF,$09,$F9,$62,$3F
	.dc.b	$8F,$6B,$C0,$FC,$94,$73,$E7,$FE,$AF,$5F,$2A,$7C,$47,$90,$F1,$9E
	.dc.b	$93,$04,$3C,$C6,$4B,$8F,$C7,$78,$B7,$5F,$CB,$65,$F1,$49,$3E,$03
	.dc.b	$8E,$5F,$22,$B9,$47,$73,$B1,$FD,$8E,$FD,$5C,$5F,$DB,$72,$F8,$99
	.dc.b	$7F,$72,$CF,$E0,$B6,$FE,$21,$0F,$88,$47,$FB,$CE,$DF,$C0,$30,$F8
	.dc.b	$93,$EA,$CB,$C8,$4C,$FD,$31,$96,$BE,$7C,$FE,$06,$5E,$C9,$CC,$D3
	.dc.b	$E0,$1E,$C9,$F7,$B9,$3F,$63,$F0,$BF,$C4,$BE,$1B,$F5,$71,$FA,$82
	.dc.b	$1F,$A8,$E0,$FD,$56,$9F,$E1,$26,$ED,$FE,$11,$78,$CF,$EB,$AB,$F8
	.dc.b	$ED,$FB,$A3,$D1,$9F,$B5,$F2,$E3,$E4,$7C,$D1,$97,$E0,$1A,$7F,$83
	.dc.b	$6F,$FC,$4F,$E1,$7F,$89,$12,$7F,$60,$D7,$F8,$6B,$5F,$E1,$6D,$7F
	.dc.b	$C5,$0D,$7F,$C4,$4F,$F1,$DF,$5F,$F9,$AF,$C0,$7F,$36,$F8,$1F,$E6
	.dc.b	$FF,$0D,$2F,$F9,$83,$5F,$E1,$AE,$DF,$E6,$93,$FC,$FA,$BF,$9F,$DF
	.dc.b	$7C,$BF,$C1,$8D,$C8,$D2,$5F,$B0,$87,$B0,$A7,$B0,$FE,$11,$97,$C5
	.dc.b	$72,$7B,$8A,$7B,$8B,$5E,$26,$D7,$89,$B5,$E5,$BF,$87,$1F,$D3,$F2
	.dc.b	$FE,$B3,$F0,$72,$3F,$A9,$35,$FE,$5E,$FC,$37,$F9,$8E,$5F,$0D,$FE
	.dc.b	$B1,$7E,$E7,$D8,$43,$D8,$53,$D4,$53,$D4,$5C,$7D,$4D,$C7,$D4,$FF
	.dc.b	$2A,$36,$74,$21,$D0,$A6,$8C,$10,$C1,$23,$BB,$C1,$FF,$23,$63,$F1
	.dc.b	$A3,$FA,$7E,$3F,$0F,$83,$FC,$F3,$95,$E0,$FF,$3C,$36,$BF,$CE,$55
	.dc.b	$C6,$BF,$CF,$F7,$F8,$89,$FA,$64,$3D,$D9,$7C,$97,$C5,$97,$C0,$6C
	.dc.b	$7E,$8F,$4F,$C2,$47,$03,$91,$A2,$9E,$FC,$FD,$AD,$FE,$00,$7C,$3F
	.dc.b	$F8,$69,$7C,$3D,$F8,$7F,$D1,$69,$FE,$96,$E1,$BC,$1F,$E9,$56,$E2
	.dc.b	$C9,$8B,$0E,$98,$F5,$CB,$E0,$5F,$9C,$87,$9C,$A7,$9F,$5F,$80,$DD
	.dc.b	$7E,$07,$37,$F6,$BC,$BB,$FC,$07,$F8,$4E,$BC,$59,$A3,$13,$A9,$FE
	.dc.b	$FD,$39,$6C,$EA,$31,$1F,$E2,$19,$E0,$D8,$CB,$FE,$BA,$8C,$B6,$32
	.dc.b	$F9,$5A,$F9,$27,$B1,$DF,$E0,$7E,$87,$C1,$7D,$1F,$83,$FA,$5F,$0D
	.dc.b	$2F,$FC,$89,$82,$31,$1F,$E9,$A8,$4F,$28,$DD,$FD,$BC,$3B,$C7,$B8
	.dc.b	$87,$AD,$CB,$4C,$9E,$B7,$2D,$32,$7B,$5E,$3C,$7F,$EB,$BB,$F7,$E5
	.dc.b	$F1,$2B,$47,$03,$E1,$15,$D3,$83,$C4,$43,$C0,$53,$99,$F4,$EF,$E0
	.dc.b	$3F,$0F,$E1,$7F,$35,$F0,$D2,$3E,$4B,$5D,$1F,$33,$4F,$D0,$57,$C7
	.dc.b	$8F,$26,$1E,$A2,$65,$F0,$A3,$C2,$79,$08,$60,$87,$AC,$C9,$ED,$CB
	.dc.b	$E3,$47,$38,$D1,$C3,$2F,$86,$4A,$F0,$C7,$92,$78,$4E,$5F,$C5,$47
	.dc.b	$AC,$FC,$3E,$5E,$FE,$57,$2F,$E4,$F4,$C3,$6B,$67,$13,$F1,$99,$FC
	.dc.b	$37,$E8,$1E,$62,$63,$E1,$7C,$84,$36,$E5,$FB,$79,$FC,$07,$47,$C4
	.dc.b	$8D,$37,$3F,$7B,$C9,$AE,$7D,$9F,$E2,$DF,$09,$FA,$78,$FD,$2F,$07
	.dc.b	$E9,$34,$F7,$C7,$F0,$28,$C6,$7F,$B3,$E0,$E3,$3F,$AC,$C3,$DC,$F8
	.dc.b	$77,$F9,$B9,$78,$23,$EA,$35,$F8,$2D,$CF,$DD,$A7,$E8,$7E,$03,$F8
	.dc.b	$51,$2B,$F8,$7B,$5F,$E1,$4D,$7F,$84,$35,$FE,$10,$D7,$F8,$DB,$67
	.dc.b	$F9,$6B,$FC,$5F,$E1,$3F,$9C,$7C,$2F,$F3,$9F,$86,$97,$FC,$21,$AF
	.dc.b	$F0,$78,$FE,$7D,$3F,$CF,$EB,$A6,$26,$92,$BD,$24,$3D,$25,$3D,$25
	.dc.b	$3D,$26,$F1,$F8,$8E,$4F,$59,$71,$F5,$B6,$BC,$4D,$AF,$13,$E0,$21
	.dc.b	$E0,$3F,$9A,$5F,$C2,$7E,$01,$0F,$E4,$D8,$FC,$2D,$FF,$0B,$87,$F5
	.dc.b	$6C,$BE,$1D,$7F,$2E,$C3,$BB,$D4,$87,$B0,$F0,$94,$C9,$EC,$32,$53
	.dc.b	$33,$EE,$7D,$59,$7C,$47,$71,$97,$79,$7C,$27,$2D,$73,$5E,$67,$C4
	.dc.b	$43,$C0,$7F,$51,$CB,$E2,$3F,$08,$F8,$98,$FC,$2A,$65,$F1,$76,$FE
	.dc.b	$B1,$9F,$C4,$BF,$E9,$D8,$F7,$7A,$93,$2F,$43,$E1,$6B,$9C,$3D,$0E
	.dc.b	$5A,$E6,$7F,$49,$9F,$C3,$8F,$69,$F8,$12,$1F,$0A,$3C,$55,$D8,$95
	.dc.b	$D1,$FF,$08,$FC,$0F,$F8,$6F,$E0,$BF,$C3,$9F,$09,$30,$FE,$D6,$D7
	.dc.b	$FB,$54,$7F,$88,$67,$F4,$91,$E0,$3C,$67,$84,$C1,$EB,$C3,$C7,$87
	.dc.b	$84,$FB,$18,$FF,$63,$F8,$5F,$F0,$61,$2B,$FC,$22,$D7,$FC,$18,$D7
	.dc.b	$FC,$14,$D7,$FC,$14,$D7,$FC,$3A,$DC,$BF,$C5,$ED,$A7,$C3,$2A,$6F
	.dc.b	$FA,$0E,$3C,$46,$37,$FF,$15,$46,$21,$3B,$76,$D7,$FC,$27,$5F,$E4
	.dc.b	$7C,$3C,$04,$8F,$69,$0F,$69,$4F,$69,$4F,$69,$4F,$69,$71,$F6,$BF
	.dc.b	$EA,$2E,$AD,$BD,$08,$74,$29,$A3,$04,$30,$43,$B9,$C0,$A6,$CA,$7F
	.dc.b	$8E,$EB,$E1,$E3,$F2,$22,$9F,$E6,$1C,$7E,$56,$3F,$1E,$3E,$5C,$ED
	.dc.b	$F9,$27,$F8,$F3,$1F,$90,$F8,$71,$F7,$B7,$24,$EA,$3C,$93,$A6,$CE
	.dc.b	$9F,$90,$D7,$E3,$67,$F9,$37,$35,$CD,$F8,$55,$CD,$B5,$C5,$EB,$3B
	.dc.b	$7E,$19,$27,$FC,$FF,$97,$C2,$7F,$CA,$71,$72,$F8,$78,$FC,$18,$F8
	.dc.b	$67,$C7,$3F,$D2,$15,$F8,$B7,$CA,$43,$E1,$5F,$86,$FC,$D1,$E1,$C3
	.dc.b	$CC,$FA,$27,$E5,$E3,$EF,$8E,$27,$33,$D6,$53,$81,$4E,$91,$FE,$88
	.dc.b	$C5,$F0,$9E,$BF,$86,$95,$FA,$93,$F0,$0E,$1F,$6E,$3F,$E7,$49,$F6
	.dc.b	$47,$88,$87,$88,$A7,$88,$A7,$88,$FD,$56,$0D,$7F,$6F,$F0,$DF,$37
	.dc.b	$DB,$F9,$DF,$03,$F3,$FE,$1B,$E8,$7B,$FF,$47,$E0,$7E,$97,$C0,$FD
	.dc.b	$3F,$7B,$EA,$7C,$1F,$D5,$F8,$4F,$AD,$F0,$DF,$5F,$E1,$BE,$C7,$C2
	.dc.b	$FD,$9F,$86,$FB,$5E,$EF,$DB,$F8,$2F,$B9,$F0,$B3,$97,$E1,$73,$FF
	.dc.b	$A4,$A3,$11,$F0,$BA,$55,$CA,$37,$0C,$3B,$C6,$32,$F1,$BA,$D3,$CE
	.dc.b	$F0,$DB,$D6,$F2,$E0,$F7,$BD,$38,$BE,$1B,$DB,$93,$F2,$5C,$B2,$3F
	.dc.b	$92,$7C,$9B,$CA,$F9,$71,$B3,$CB,$CF,$F2,$DF,$3F,$3F,$CB,$7D,$3C
	.dc.b	$FF,$2D,$F5,$F3,$FC,$B7,$DB,$CF,$F2,$DF,$7F,$3F,$CB,$7E,$0F,$3F
	.dc.b	$CB,$7E,$1F,$37,$C8,$F0,$3F,$9B,$37,$C0,$FE,$6B,$D1,$F2,$6B,$E6
	.dc.b	$73,$FC,$87,$E7,$73,$FC,$87,$E8,$73,$FC,$87,$E9,$73,$FC,$87,$EA
	.dc.b	$73,$FC,$87,$EB,$73,$FC,$87,$EC,$73,$FC,$77,$ED,$73,$FF,$48,$E7
	.dc.b	$F9,$11,$FC,$7E,$3F,$24,$FC,$83,$E9,$1F,$20,$FA,$C7,$C8,$3E,$D1
	.dc.b	$F2,$0F,$BC,$7C,$83,$F0,$47,$C8,$3F,$0F,$9F,$C1,$DF,$FA,$D6,$BF
	.dc.b	$16,$7F,$98,$6F,$7A,$7F,$25,$1F,$8E,$9E,$1C,$1F,$D7,$B0,$F9,$18
	.dc.b	$7F,$63,$E0,$F9,$0F,$F6,$5E,$0F,$90,$FF,$67,$E0,$F9,$0F,$F6,$9E
	.dc.b	$0F,$90,$FF,$6B,$E0,$F9,$0F,$F6,$DE,$0F,$90,$FF,$6F,$DF,$E4,$74
	.dc.b	$7F,$6E,$F8,$5F,$E4,$9F,$0F,$FC,$9F,$F8,$07,$FD,$2F,$C3,$FC,$21
	.dc.b	$7C,$46,$7B,$F5,$F0,$BB,$F8,$6E,$1F,$0F,$C7,$E0,$F9,$7C,$27,$3F
	.dc.b	$85,$E9,$F0,$FD,$7C,$6E,$DF,$0D,$DF,$E1,$FC,$1F,$0D,$E1,$F8,$79
	.dc.b	$C3,$FB,$AE,$14,$62,$3C,$93,$C5,$D5,$F5,$78,$63,$E1,$79,$67,$E5
	.dc.b	$7A,$6B,$E9,$7B,$6F,$ED,$7C,$1C,$3F,$0F,$BB,$E2,$C7,$F0,$6E,$4F
	.dc.b	$E9,$47,$CA,$7F,$20,$7E,$DC,$FE,$1D,$E4,$FE,$DD,$FE,$21,$E4,$FE
	.dc.b	$DD,$FE,$25,$E4,$FE,$DD,$FE,$29,$E4,$FE,$DD,$FE,$2D,$E4,$FE,$DD
	.dc.b	$FE,$31,$E4,$FE,$DD,$FE,$35,$E4,$FE,$DD,$FE,$39,$E4,$7E,$3F,$CA
	.dc.b	$37,$C8,$FC,$7C,$FF,$80,$FC,$8F,$C7,$EF,$FE,$2E,$F2,$3F,$1E,$FF
	.dc.b	$8C,$3C,$8F,$C7,$BF,$E3,$2F,$23,$F1,$EF,$F8,$D3,$C8,$FC,$7B,$FE
	.dc.b	$36,$F2,$3F,$1E,$FF,$8E,$3C,$8F,$C7,$BF,$C6,$B2,$79,$5F,$B6,$F2
	.dc.b	$9B,$BC,$9F,$DB,$D7,$C8,$3F,$6E,$7C,$83,$F6,$E7,$C8,$3F,$6E,$7C
	.dc.b	$83,$F6,$E7,$C8,$3F,$6E,$7C,$83,$F6,$E7,$F1,$E3,$F6,$47,$F1,$E2
	.dc.b	$1F,$28,$BE,$57,$EF,$1E,$07,$F1,$E3,$F8,$11,$FC,$78,$FE,$0C,$7F
	.dc.b	$1E,$3F,$85,$1F,$C7,$8F,$E1,$C7,$F1,$E3,$F8,$91,$FC,$78,$FE,$2C
	.dc.b	$7F,$1E,$3F,$A9,$C7,$F1,$FB,$F9,$15,$F1,$E7,$F1,$D5,$FE,$69,$E8
	.dc.b	$FC,$73,$FE,$6B,$E8,$FC,$73,$FE,$6D,$E8,$FC,$73,$FE,$6F,$E8,$FC
	.dc.b	$73,$FE,$71,$E8,$FC,$73,$FE,$73,$E8,$FC,$73,$FE,$75,$E8,$FC,$73
	.dc.b	$FD,$0B,$2F,$8B,$BF,$F9,$DB,$A3,$F1,$D1,$F1,$A7,$FC,$F9,$CF,$F8
	.dc.b	$E8,$FF,$29,$73,$FE,$39,$FF,$2A,$73,$FE,$39,$FF,$2B,$73,$FE,$39
	.dc.b	$FF,$2C,$73,$FE,$39,$FF,$2D,$73,$FE,$39,$FF,$2E,$73,$FE,$39,$FD
	.dc.b	$89,$F8,$EF,$2B,$F8,$E3,$FD,$F3,$CA,$FD,$D3,$FB,$C3,$F8,$67,$95
	.dc.b	$FC,$69,$FE,$1D,$E5,$7F,$1A,$7F,$88,$79,$5F,$C6,$9F,$E2,$5E,$57
	.dc.b	$F1,$A7,$F8,$A7,$95,$FC,$69,$FE,$2D,$E5,$7F,$1A,$7F,$8C,$79,$3F
	.dc.b	$C5,$F8,$FF,$DF,$BE,$57,$F1,$A9,$C9,$FC,$77,$CA,$FE,$34,$FF,$2F
	.dc.b	$F2,$BF,$8D,$3F,$CC,$7C,$AF,$E3,$4F,$F3,$3F,$2B,$F8,$D3,$FC,$D7
	.dc.b	$CA,$FE,$34,$FF,$37,$F2,$BF,$8D,$3F,$CE,$7C,$AF,$82,$FF,$88,$78
	.dc.b	$FF,$8C,$BF,$C6,$CF,$F0,$8F,$27,$FA,$D7,$C8,$6A,$F7,$BF,$A2,$7C
	.dc.b	$24,$BF,$E6,$8D,$7F,$A3,$37,$83,$DE,$DF,$03,$A5,$A3,$14,$C3,$BC
	.dc.b	$61,$B1,$D8,$87,$62,$DF,$66,$DF,$6A,$E5,$7D,$E7,$27,$77,$0F,$32
	.dc.b	$57,$36,$D7,$37,$8D,$F9,$1C,$9C,$EF,$BE,$1F,$DA,$4F,$E8,$93,$F2
	.dc.b	$9F,$ED,$9F,$0F,$FD,$B7,$E1,$BF,$B7,$7C,$37,$F6,$FF,$87,$FF,$4B
	.dc.b	$7F,$00,$BE,$1A,$FC,$3E,$3E,$23,$3F,$03,$AF,$84,$DF,$C3,$F0,$F4
	.dc.b	$F8,$FC,$2F,$2F,$87,$E7,$F0,$FD,$3E,$1F,$F2,$5C,$DF,$4B,$C2,$BC
	.dc.b	$FF,$0D,$D3,$FE,$68,$E0,$5F,$38,$C6,$3D,$DD,$67,$E3,$78,$6B,$E7
	.dc.b	$79,$6F,$EC,$E0,$F7,$47,$F7,$BE,$2F,$84,$FF,$7D,$E4,$F9,$0F,$F7
	.dc.b	$FE,$5F,$93,$D3,$F9,$18,$D1,$FE,$03,$E6,$F9,$2F,$F8,$17,$9B,$B3
	.dc.b	$E0,$3D,$5C,$DF,$21,$F6,$73,$7C,$87,$DD,$CD,$F2,$1F,$81,$CD,$F2
	.dc.b	$1F,$85,$CD,$F1,$B9,$FF,$47,$CD,$74,$FD,$D5,$7F,$AB,$39,$97,$C6
	.dc.b	$31,$87,$57,$58,$F8,$5E,$19,$79,$5E,$59,$FE,$01,$FA,$5C,$3F,$21
	.dc.b	$FF,$19,$70,$FC,$87,$FC,$69,$C3,$F2,$1F,$F1,$B7,$0F,$C8,$E8,$FD
	.dc.b	$09,$23,$F3,$3B,$7E,$83,$A7,$F3,$C4,$E9,$FC,$F4,$FE,$73,$8B,$F3
	.dc.b	$D3,$98,$FC,$F1,$CA,$FA,$3B,$BE,$97,$CA,$B8,$5F,$28,$C1,$2B,$99
	.dc.b	$C6,$F9,$D7,$0B,$E7,$3A,$8C,$4F,$83,$D8,$FF,$05,$56,$B0,$ED,$7B
	.dc.b	$C3,$B4,$60,$93,$B2,$5F,$82,$77,$18,$AF,$71,$0F,$31,$70,$F1,$B6
	.dc.b	$78,$D7,$38,$F1,$12,$FC,$31,$97,$F8,$56,$1E,$AC,$7F,$82,$D7,$E7
	.dc.b	$4F,$D3,$1F,$9D,$3F,$06,$7E,$74,$FC,$39,$FB,$03,$F3,$A7,$ED,$0F
	.dc.b	$CE,$9F,$B8,$3F,$3A,$7E,$F7,$AB,$F7,$EF,$F9,$F7,$93,$F3,$87,$E7
	.dc.b	$F9,$3F,$37,$1B,$38,$9F,$9F,$EA,$FE,$04,$FE,$87,$AB,$F8,$13,$FA
	.dc.b	$3E,$AF,$E0,$4F,$E9,$7A,$BF,$6A,$FF,$A2,$FA,$7F,$72,$FF,$A3,$7A
	.dc.b	$BE,$0B,$F9,$A3,$F8,$11,$FD,$50,$FE,$71,$A6,$B8,$7F,$B5,$CF,$27
	.dc.b	$FB,$6D,$76,$5C,$3F,$D6,$F4,$63,$E4,$9C,$E3,$FD,$46,$59,$7A,$A7
	.dc.b	$3C,$09,$69,$F0,$27,$30,$D7,$F9,$36,$1F,$C0,$F7,$F9,$B5,$FE,$00
	.dc.b	$DF,$E6,$CF,$F8,$00,$FE,$11,$D5,$FC,$1E,$79,$F5,$7F,$07,$7A,$F5
	.dc.b	$7F,$07,$7B,$F5,$7F,$07,$7C,$3D,$5F,$C1,$F0,$FE,$25,$D5,$FC,$1E
	.dc.b	$7F,$C1,$FD,$5F,$C1,$DF,$F0,$8F,$57,$F0,$77,$FC,$27,$CF,$FC,$1E
	.dc.b	$FF,$34,$7E,$5F,$AB,$F4,$4F,$E8,$09,$3F,$A0,$6F,$37,$E8,$1E,$3C
	.dc.b	$9F,$9E,$6D,$72,$AE,$04,$AE,$13,$C8,$F0,$75,$7E,$96,$BF,$BA,$F5
	.dc.b	$7E,$FA,$30,$49,$EA,$D7,$96,$1E,$16,$D7,$86,$FB,$E1,$E1,$9E,$E7
	.dc.b	$22,$1C,$2F,$FC,$39,$87,$ED,$63,$FA,$89,$F9,$73,$F6,$A7,$E6,$CF
	.dc.b	$DA,$9F,$9F,$3F,$A8,$9F,$B5,$3F,$9F,$9F,$B5,$3F,$A1,$9F,$B5,$3F
	.dc.b	$A3,$9F,$AF,$3F,$0A,$7E,$1E,$7F,$14,$DA,$FE,$F4,$FF,$56,$38,$47
	.dc.b	$E2,$9E,$8F,$E2,$9F,$03,$F8,$A7,$C8,$FE,$29,$FE,$18,$FE,$73,$4F
	.dc.b	$C5,$4F,$F8,$E3,$4F,$C5,$3F,$E3,$AD,$3F,$14,$FF,$8F,$34,$FE,$EE
	.dc.b	$7E,$D5,$FF,$1F,$F5,$7E,$D5,$FF,$20,$F5,$7E,$D5,$FD,$A7,$57,$EC
	.dc.b	$DF,$DB,$75,$7E,$C5,$FD,$C7,$57,$EB,$DF,$DD,$74,$FE,$BB,$5F,$D8
	.dc.b	$F4,$FE,$BA,$7E,$1E,$FF,$C6,$CF,$8C,$7F,$87,$4F,$8C,$7E,$D4,$FE
	.dc.b	$37,$3F,$BA,$AE,$D1,$FC,$0E,$FC,$4F,$F0,$5C,$3C,$CF,$F0,$7C,$7D
	.dc.b	$4F,$F9,$2F,$2F,$73,$FE,$4D,$CF,$E1,$3F,$C6,$0F,$E3,$67,$FA,$54
	.dc.b	$FC,$89,$F8,$F3,$F2,$38,$FE,$57,$8F,$F3,$4D,$E3,$FC,$D4,$F0,$8C
	.dc.b	$71,$FE,$85,$B1,$CB,$5F,$CF,$91,$E0,$FF,$7E,$E3,$FD,$64,$6C,$C9
	.dc.b	$30,$F0,$CF,$16,$C6,$70,$F1,$C6,$08,$E0,$98,$79,$6B,$94,$62,$BF
	.dc.b	$B6,$5F,$89,$F8,$44,$C3,$D6,$DA,$E3,$7C,$E3,$C4,$DC,$3C,$2D,$C3
	.dc.b	$B4,$F4,$C3,$B6,$1D,$09,$87,$49,$F8,$58,$72,$6D,$F2,$AD,$C7,$80
	.dc.b	$CE,$1D,$CF,$E0,$FD,$7F,$D3,$CF,$F3,$D9,$FE,$7C,$F8,$1F,$F3,$EF
	.dc.b	$C3,$FF,$9F,$BE,$1B,$FC,$FF,$F0,$FF,$E8,$0F,$85,$FF,$40,$FC,$37
	.dc.b	$FA,$0B,$E1,$FF,$D0,$7F,$0B,$FE,$84,$F8,$76,$B3,$6F,$EA,$BF,$C0
	.dc.b	$38,$8C,$71,$7F,$AE,$61,$F0,$D9,$AD,$FC,$3D,$AD,$7C,$2D,$C3,$FD
	.dc.b	$65,$83,$D7,$C7,$32,$1F,$0B,$85,$EC,$D2,$5F,$88,$E7,$3B,$BE,$18
	.dc.b	$71,$CB,$96,$9C,$EF,$DC,$7F,$A4,$76,$F8,$4F,$FA,$4F,$83,$E3,$3F
	.dc.b	$E9,$5E,$2F,$94,$FF,$A5,$F9,$3E,$73,$FE,$99,$E6,$FA,$2F,$F8,$27
	.dc.b	$7F,$57,$C2,$FC,$78,$F9,$11,$87,$07,$C2,$8A,$D3,$31,$B7,$93,$5E
	.dc.b	$7C,$09,$E3,$39,$9F,$A9,$18,$9D,$47,$46,$BD,$1A,$F4,$8E,$04,$78
	.dc.b	$3C,$FD,$BF,$D1,$49,$FA,$D1,$B9,$E8,$D7,$A1,$C8,$86,$8A,$68,$C1
	.dc.b	$0C,$10,$F6,$9C,$BD,$B7,$B1,$88,$E0,$53,$81,$F3,$4C,$94,$CE,$1F
	.dc.b	$25,$C1,$C4,$86,$08,$EC,$98,$7C,$38,$FD,$EB,$57,$2F,$EA,$B0,$F8
	.dc.b	$91,$D2,$B8,$C6,$23,$AD,$7D,$47,$4D,$73,$1B,$9E,$31,$F0,$EF,$73
	.dc.b	$9A,$E1,$5A,$AF,$82,$D7,$E0,$C6,$89,$3C,$4E,$58,$FE,$00,$FB,$F3
	.dc.b	$F6,$0E,$E5,$3C,$A5,$AE,$CE,$8F,$0D,$76,$7D,$07,$03,$35,$E1,$79
	.dc.b	$46,$DC,$98,$8D,$3D,$3E,$17,$F8,$2B,$F4,$39,$BF,$8B,$73,$63,$3F
	.dc.b	$E3,$75,$EA,$AE,$64,$3E,$1D,$F4,$69,$DC,$87,$E5,$B4,$FC,$8B,$FC
	.dc.b	$73,$DD,$FE,$30,$7F,$19,$F8,$5F,$E3,$5E,$FF,$F1,$BF,$85,$FE,$39
	.dc.b	$F0,$3F,$C7,$7E,$17,$F8,$F7,$C1,$FF,$1F,$F8,$5F,$E4,$1F,$0D,$FC
	.dc.b	$87,$E1,$BF,$91,$7C,$3F,$F2,$3F,$73,$F9,$3F,$C3,$7F,$A3,$37,$F7
	.dc.b	$0E,$74,$3D,$ED,$F1,$70,$F7,$FE,$17,$47,$C4,$C3,$F9,$56,$1F,$CC
	.dc.b	$7A,$35,$18,$C5,$87,$83,$1F,$0E,$1E,$3A,$E3,$D2,$D3,$D1,$18,$DB
	.dc.b	$D8,$EB,$83,$E0,$3C,$38,$BE,$23,$CB,$93,$E4,$3D,$39,$BE,$5B,$FC
	.dc.b	$B7,$57,$C3,$FC,$23,$E1,$FC,$0F,$C4,$8F,$8B,$F0,$5F,$1B,$E1,$FE
	.dc.b	$3F,$C2,$7C,$8F,$03,$E4,$FC,$2F,$CA,$F8,$7F,$97,$F0,$FF,$33,$E1
	.dc.b	$BE,$6F,$BD,$F3,$BD,$5F,$9F,$EF,$FD,$0F,$67,$E8,$FC,$07,$F8,$0F
	.dc.b	$73,$9E,$2F,$91,$5C,$5F,$95,$87,$57,$E6,$65,$E1,$7E,$76,$9E,$57
	.dc.b	$E8,$6D,$E9,$7E,$97,$07,$B5,$FA,$9C,$3F,$05,$FA,$DC,$5F,$DD,$7D
	.dc.b	$A7,$97,$FD,$EB,$C0,$FC,$1C,$EE,$BF,$0F,$1F,$86,$BE,$91,$8D,$D8
	.dc.b	$F8,$63,$19,$79,$5D,$67,$E9,$78,$69,$ED,$79,$6B,$F0,$5E,$9B,$7C
	.dc.b	$57,$B6,$FF,$24,$C7,$95,$FE,$E2,$CD,$5C,$63,$17,$D5,$D6,$1E,$17
	.dc.b	$86,$3E,$57,$96,$5E,$97,$A6,$7E,$E7,$B6,$9F,$0D,$F0,$6D,$F2,$77
	.dc.b	$FF,$0E,$73,$FE,$43,$7F,$8F,$DA,$F2,$3F,$77,$1F,$A0,$F2,$3F,$76
	.dc.b	$FE,$8B,$C8,$FD,$DB,$FA,$4F,$23,$F7,$6F,$E9,$BC,$8F,$DD,$BF,$A8
	.dc.b	$F2,$3F,$56,$FE,$BF,$C5,$FD,$BB,$FB,$1F,$23,$F6,$EF,$F8,$A3,$C8
	.dc.b	$BD,$5F,$BE,$9D,$CF,$D9,$AE,$6F,$A2,$FB,$BE,$AC,$3C,$6F,$B3,$1F
	.dc.b	$3B,$EE,$CB,$D6,$FC,$0C,$FD,$EF,$C2,$D3,$E2,$70,$FF,$91,$37,$F9
	.dc.b	$13,$FC,$84,$FD,$C1,$FC,$12,$78,$CF,$F9,$2A,$BA,$BF,$E4,$CB,$F0
	.dc.b	$BF,$E4,$EC,$3C,$AF,$F9,$43,$1F,$4B,$FE,$52,$CB,$DA,$FE,$1F,$3F
	.dc.b	$82,$FE,$27,$4F,$8B,$BF,$F9,$23,$4F,$E2,$BB,$FF,$1C,$7F,$3D,$E4
	.dc.b	$7F,$1B,$7F,$8F,$79,$1F,$C6,$DF,$BF,$E4,$7F,$1A,$7F,$03,$E4,$7F
	.dc.b	$1A,$7F,$07,$E4,$7F,$1A,$7F,$0B,$E4,$7F,$1A,$7F,$0F,$E4,$7F,$1A
	.dc.b	$7F,$13,$CB,$F8,$BF,$0B,$F6,$B1,$FC,$DF,$C1,$FE,$36,$FE,$E8,$FD
	.dc.b	$E9,$FC,$C5,$FE,$36,$FF,$33,$F2,$3F,$8D,$BF,$CD,$7C,$8F,$E3,$6F
	.dc.b	$F3,$7F,$23,$F8,$DB,$FC,$E7,$C8,$FE,$36,$FF,$3B,$F2,$3F,$8D,$BF
	.dc.b	$CF,$7C,$8F,$E3,$6F,$FA,$53,$1F,$E3,$6E,$4F,$E3,$67,$13,$F8,$D9
	.dc.b	$D4,$FE,$36,$78,$4F,$E3,$67,$94,$FE,$36,$7A,$4F,$E3,$67,$B4,$F7
	.dc.b	$9F,$C4,$7C,$0F,$E2,$D9,$FC,$6E,$FF,$E3,$33,$F1,$4F,$E2,$DE,$0F
	.dc.b	$EE,$F9,$3F,$8F,$61,$FC,$62,$7B,$1F,$C6,$0F,$11,$FC,$60,$F3,$1F
	.dc.b	$C6,$0F,$51,$FC,$60,$F7,$1F,$C6,$0F,$84,$7F,$18,$3E,$37,$91,$FC
	.dc.b	$62,$7C,$3E,$47,$F1,$77,$C7,$E4,$7F,$17,$7C,$BE,$47,$F1,$77,$CF
	.dc.b	$E4,$7F,$17,$7D,$3E,$47,$F1,$77,$D7,$E4,$7B,$72,$FC,$36,$9F,$C5
	.dc.b	$E7,$DF,$BF,$EC,$36,$FF,$54,$73,$7E,$D2,$7F,$D6,$9E,$3F,$EC,$1F
	.dc.b	$F5,$B7,$8F,$FB,$07,$FD,$71,$E3,$FE,$C1,$FF,$5D,$78,$FF,$B0,$7F
	.dc.b	$D7,$9E,$3F,$EC,$1F,$F5,$F7,$8F,$FB,$07,$E7,$F8,$FF,$B0,$7E,$8F
	.dc.b	$8D,$FB,$0F,$17,$F8,$AF,$42,$35,$C3,$F9,$FB,$E4,$FF,$B0,$B0,$EC
	.dc.b	$FF,$B1,$31,$F1,$3F,$EC,$6C,$BC,$CF,$FB,$23,$3F,$53,$FE,$CA,$D3
	.dc.b	$DC,$FF,$B3,$35,$F8,$4F,$F1,$31,$D3,$F8,$BD,$1E,$DF,$F8,$A8,$93
	.dc.b	$FE,$1F,$6B,$FD,$8D,$BE,$47,$F5,$56,$F0,$7E,$41,$B3,$FB,$B6,$F1
	.dc.b	$7F,$8C,$9B,$7E,$09,$EC,$7F,$14,$21,$FD,$3B,$87,$73,$23,$67,$8C
	.dc.b	$B8,$78,$DB,$87,$8D,$A7,$8C,$B8,$78,$DB,$87,$8D,$E8,$7C,$6F,$27
	.dc.b	$FC,$11,$FC,$02,$F8,$2F,$F7,$AB,$67,$FC,$8E,$BD,$EC,$FC,$1E,$BE
	.dc.b	$03,$7F,$09,$C3,$E0,$B8,$FC,$2F,$2F,$84,$E7,$F0,$DD,$3E,$1B,$AF
	.dc.b	$C3,$F6,$DF,$BF,$C3,$78,$3D,$AF,$0F,$C3,$78,$BD,$CF,$C7,$71,$7C
	.dc.b	$8E,$77,$3F,$A2,$77,$3E,$6B,$E6,$FA,$31,$EE,$FA,$B3,$F1,$BE,$CD
	.dc.b	$7C,$EF,$BB,$7F,$5B,$F0,$38,$7D,$EF,$C2,$E3,$F8,$BC,$B2,$07,$CB
	.dc.b	$3F,$CE,$9C,$BF,$25,$FF,$18,$B7,$6F,$4C,$6A,$BA,$BA,$BF,$0B,$C3
	.dc.b	$0F,$2B,$CB,$1F,$4B,$D3,$2F,$6B,$DB,$3F,$82,$F8,$34,$F8,$AF,$8B
	.dc.b	$5F,$96,$FE,$44,$FB,$AE,$B3,$F2,$F3,$B5,$5F,$49,$CC,$7F,$5C,$BF
	.dc.b	$6F,$17,$DD,$7F,$0F,$C7,$F8,$17,$F1,$3C,$9F,$85,$7F,$17,$CB,$F8
	.dc.b	$97,$F1,$BC,$DF,$90,$E2,$F9,$91,$F2,$38,$FF,$90,$73,$FE,$67,$4F
	.dc.b	$D1,$4F,$E6,$9A,$4E,$5F,$D2,$46,$5B,$19,$8F,$D3,$19,$2B,$9D,$FF
	.dc.b	$51,$19,$6C,$67,$67,$C3,$7E,$98,$95,$FA,$EE,$9F,$DD,$3C,$71,$F4
	.dc.b	$63,$E2,$24,$78,$1B,$8F,$66,$CF,$09,$D1,$27,$4F,$AB,$AF,$F8,$F5
	.dc.b	$FE,$4A,$21,$EE,$31,$3E,$C6,$D7,$F0,$36,$CF,$07,$E4,$98,$31,$3C
	.dc.b	$5C,$90,$EC,$57,$4E,$08,$E0,$9E,$D7,$F8,$DE,$3F,$24,$4C,$FE,$EB
	.dc.b	$E1,$3D,$73,$C6,$FD,$11,$DD,$AF,$66,$C6,$EF,$D9,$39,$6C,$66,$3D
	.dc.b	$B3,$C0,$93,$F2,$FD,$AF,$F7,$FD,$6E,$BC,$75,$A7,$8D,$74,$8C,$46
	.dc.b	$A3,$04,$76,$43,$45,$3E,$01,$0F,$99,$87,$C7,$C3,$F6,$AD,$CB,$FD
	.dc.b	$F9,$3F,$CF,$0E,$07,$BC,$E8,$7C,$7A,$F1,$3E,$02,$1C,$30,$F7,$37
	.dc.b	$0F,$5B,$F4,$F1,$F8,$71,$88,$E8,$76,$3C,$07,$43,$25,$32,$75,$AF
	.dc.b	$1B,$E1,$C3,$D8,$FF,$A8,$4F,$E8,$18,$72,$78,$1D,$F0,$F3,$B9,$21
	.dc.b	$D8,$AF,$D9,$C3,$D3,$7E,$09,$CE,$1E,$B7,$81,$2B,$AB,$63,$51,$F9
	.dc.b	$BD,$3F,$81,$CF,$F6,$6D,$3F,$81,$BF,$4C,$EB,$7D,$DE,$4D,$AE,$A4
	.dc.b	$BE,$A7,$E2,$B0,$F3,$D7,$F0,$D2,$1F,$C3,$6B,$F8,$7E,$7F,$C6,$6B
	.dc.b	$F8,$45,$FF,$06,$23,$82,$4F,$F8,$63,$1F,$4D,$FF,$56,$C7,$D3,$87
	.dc.b	$3A,$EB,$97,$F1,$46,$D7,$F1,$3C,$39,$BE,$4A,$E0,$F9,$B4,$FE,$2A
	.dc.b	$E6,$F9,$34,$E5,$7D,$32,$F3,$B7,$0E,$B3,$FC,$16,$BF,$3F,$87,$E8
	.dc.b	$BB,$75,$72,$BA,$61,$C9,$B8,$72,$BF,$2C,$7F,$88,$B1,$F2,$BF,$2C
	.dc.b	$DD,$F3,$70,$49,$E2,$4C,$7C,$64,$8F,$CC,$1B,$7D,$ED,$3D,$E7,$62
	.dc.b	$19,$C3,$FA,$8D,$FF,$A8,$F3,$7B,$DF,$E3,$43,$FC,$6B,$F0,$DF,$E3
	.dc.b	$6F,$80,$FF,$1B,$FC,$37,$F4,$1C,$FF,$BC,$FF,$00,$FF,$42,$6F,$C1
	.dc.b	$E7,$5F,$F0,$BE,$8A,$7C,$26,$76,$4E,$7E,$17,$0E,$FE,$1A,$DB,$E1
	.dc.b	$EF,$26,$3E,$17,$3F,$0F,$AF,$87,$DF,$A7,$C3,$CC,$FC,$6E,$FF,$E5
	.dc.b	$0E,$AB,$BF,$83,$45,$70,$8F,$F1,$6F,$17,$B3,$1E,$B1,$8C,$3D,$76
	.dc.b	$CF,$C7,$39,$3A,$6B,$E8,$DB,$D3,$C1,$EA,$E1,$F5,$F1,$7B,$2F,$E5
	.dc.b	$9F,$E6,$0E,$4F,$9E,$FF,$98,$B9,$BE,$93,$FE,$40,$E8,$FF,$6C,$68
	.dc.b	$F8,$3F,$8A,$7C,$6F,$3B,$E3,$FC,$27,$C8,$F4,$3E,$4F,$C2,$FC,$AF
	.dc.b	$4B,$E5,$FC,$37,$CC,$F5,$3E,$6F,$C3,$FC,$EF,$5B,$E8,$FF,$00,$FF
	.dc.b	$92,$37,$75,$7F,$8D,$A1,$EC,$6B,$C7,$DF,$B3,$C3,$DA,$E3,$ED,$FC
	.dc.b	$EF,$17,$FC,$79,$E1,$5C,$53,$BC,$FE,$45,$27,$FD,$79,$18,$58,$F9
	.dc.b	$32,$F2,$F8,$1E,$7A,$E3,$CE,$DF,$D9,$18,$E1,$F8,$0E,$B8,$FE,$23
	.dc.b	$C3,$97,$E4,$3C,$B9,$FE,$5B,$F3,$FD,$5F,$BF,$18,$8F,$87,$97,$E0
	.dc.b	$BA,$3E,$7C,$78,$BA,$3E,$7B,$E4,$E8,$F9,$EF,$9B,$A3,$E7,$BE,$8E
	.dc.b	$8F,$5B,$EF,$E0,$F9,$8F,$C1,$E7,$F8,$CF,$80,$FD,$3F,$1F,$CA,$7F
	.dc.b	$A1,$10,$FE,$58,$7C,$9E,$5E,$3D,$1F,$66,$BA,$CF,$3E,$6F,$A1,$87
	.dc.b	$8A,$31,$CA,$CB,$CF,$18,$CF,$D6,$EB,$4F,$7B,$C3,$5F,$86,$F2,$DB
	.dc.b	$E4,$69,$FC,$1B,$83,$FC,$91,$D3,$F9,$A7,$3C,$3F,$2E,$FE,$A1,$F2
	.dc.b	$CF,$B0,$7C,$B3,$EE,$1F,$2C,$FC,$0F,$37,$E0,$DF,$67,$17,$E6,$70
	.dc.b	$FC,$C6,$FF,$01,$F8,$53,$F8,$DE,$9F,$CB,$BF,$26,$3F,$D1,$FD,$3F
	.dc.b	$97,$E4,$F9,$39,$7F,$1A,$C7,$F2,$B7,$DB,$A7,$F2,$8F,$E5,$E7,$C5
	.dc.b	$87,$F2,$4C,$FF,$27,$3F,$33,$A7,$F2,$6F,$CE,$E9,$FC,$9B,$F4,$3A
	.dc.b	$7F,$26,$FD,$2D,$FF,$21,$C0,$E9,$F1,$79,$7F,$B6,$3F,$6F,$EE,$7E
	.dc.b	$E3,$E2,$3F,$73,$EF,$7E,$EB,$E2,$3F,$77,$F0,$1F,$BC,$F8,$8F,$DE
	.dc.b	$FC,$17,$EF,$BE,$23,$F7,$FF,$09,$FC,$03,$E2,$3F,$80,$FC,$37,$F0
	.dc.b	$2F,$88,$FE,$07,$F1,$1F,$C1,$3D,$8F,$E6,$BE,$87,$FB,$37,$AF,$F9
	.dc.b	$EB,$FE,$F7,$E6,$FD,$DB,$5F,$C7,$37,$AF,$FD,$7E,$F1,$8F,$CD,$D7
	.dc.b	$67,$F8,$06,$1E,$27,$F8,$16,$5E,$67,$F8,$26,$9E,$A7,$F8,$36,$DE
	.dc.b	$E7,$F8,$47,$07,$C2,$7F,$85,$71,$7C,$67,$F8,$67,$27,$C9,$7F,$7C
	.dc.b	$7F,$43,$E7,$FC,$9B,$FD,$17,$C2,$FC,$DE,$DF,$D7,$BC,$3F,$CF,$BF
	.dc.b	$D8,$3C,$4F,$D1,$BF,$D8,$BC,$5F,$D3,$BF,$D9,$3C,$6F,$D6,$3F,$92
	.dc.b	$C3,$F8,$E7,$83,$FC,$A1,$FC,$91,$FB,$73,$F2,$5C,$9F,$D5,$DA,$7F
	.dc.b	$00,$F0,$FF,$26,$FF,$02,$F2,$7F,$39,$3F,$95,$F1,$FB,$BF,$D5,$B1
	.dc.b	$F1,$BF,$D6,$32,$F3,$BF,$D6,$B3,$F5,$BF,$D7,$34,$F7,$BF,$B3,$C7
	.dc.b	$F2,$47,$F2,$53,$E3,$F9,$3F,$DE,$F7,$F8,$EF,$F7,$DD,$F8,$CF,$F6
	.dc.b	$DB,$EA,$FF,$6F,$C3,$C2,$FF,$71,$C7,$CA,$FF,$73,$CB,$D2,$FF,$1E
	.dc.b	$CF,$DA,$FB,$CF,$8E,$7E,$FB,$5F,$E9,$DE,$C7,$F9,$33,$83,$F7,$33
	.dc.b	$F2,$CF,$E0,$7E,$4F,$CF,$7F,$70,$7F,$2F,$3F,$70,$7F,$33,$3F,$85
	.dc.b	$9F,$C3,$4F,$DC,$1F,$C4,$4F,$DC,$1F,$C5,$4F,$DC,$1F,$D4,$B4,$F8
	.dc.b	$7E,$57,$F7,$58,$F8,$ED,$7E,$59,$FC,$94,$FE,$3E,$7F,$29,$3F,$8F
	.dc.b	$9F,$CB,$4F,$E3,$E7,$F3,$13,$CA,$7F,$36,$CB,$E3,$ED,$FC,$EA,$3F
	.dc.b	$1A,$7F,$49,$3E,$3E,$1F,$DE,$72,$F8,$FE,$57,$F1,$D9,$CF,$0F,$F1
	.dc.b	$DE,$2F,$9F,$87,$39,$FE,$57,$3F,$C7,$0F,$F1,$0F,$93,$FC,$73,$1F
	.dc.b	$F2,$0F,$93,$FC,$71,$FF,$21,$F9,$3F,$C7,$1F,$F2,$2F,$93,$FC,$71
	.dc.b	$FE,$85,$87,$E2,$C7,$93,$FC,$75,$FE,$BB,$D3,$FB,$57,$F9,$DF,$91
	.dc.b	$FC,$75,$FF,$02,$79,$1F,$C7,$5F,$F0,$37,$91,$FC,$75,$FF,$04,$79
	.dc.b	$1F,$C7,$5F,$F0,$57,$91,$FC,$75,$FF,$06,$79,$1F,$C7,$5F,$F0,$77
	.dc.b	$91,$FC,$75,$FE,$33,$ED,$7F,$1E,$F6,$E4,$7F,$1C,$7F,$90,$C6,$E3
	.dc.b	$FA,$7C,$F3,$7F,$A8,$D7,$77,$FA,$9D,$F8,$DF,$EA,$B8,$79,$DF,$EA
	.dc.b	$F8,$FA,$DF,$EB,$39,$7B,$DF,$82,$7F,$18,$21,$F2,$3C,$BF,$E6,$3E
	.dc.b	$E7,$F3,$43,$F9,$AF,$C3,$7F,$36,$F7,$BF,$9B,$FC,$37,$F3,$8F,$80
	.dc.b	$FE,$73,$F0,$DF,$CE,$BE,$0B,$F9,$DF,$C3,$7F,$3C,$F8,$4F,$E7,$B9
	.dc.b	$FF,$3E,$F8,$8F,$E7,$FF,$0D,$FD,$03,$E2,$3F,$A0,$FC,$47,$F4,$2F
	.dc.b	$13,$FA,$1F,$B5,$F1,$B8,$BF,$9F,$F5,$5E,$EF,$E7,$F3,$B9,$FE,$8B
	.dc.b	$7C,$DF,$E8,$F8,$F7,$7F,$A4,$E7,$E3,$7F,$A5,$EB,$E7,$7F,$A6,$EF
	.dc.b	$EB,$7F,$A7,$F0,$FB,$DF,$EA,$3C,$7F,$0D,$FE,$A7,$C9,$FE,$10,$C3
	.dc.b	$FA,$BF,$C3,$FF,$5A,$3F,$AD,$FB,$9F,$D7,$3E,$1F,$FD,$23,$FC,$03
	.dc.b	$FE,$9B,$F1,$FF,$D3,$2B,$DF,$CF,$C0,$6B,$E0,$77,$F0,$5C,$3E,$0F
	.dc.b	$8F,$C4,$72,$F8,$3E,$7F,$11,$D3,$E1,$BA,$FC,$47,$6F,$88,$EF,$EF
	.dc.b	$FC,$7D,$FF,$B5,$F7,$5E,$6F,$F5,$2F,$0F,$F0,$9E,$E5,$79,$30,$E7
	.dc.b	$5F,$E6,$AA,$65,$E0,$8C,$69,$E4,$75,$B7,$A1,$E1,$C1,$EC,$79,$71
	.dc.b	$7C,$27,$E2,$1F,$DF,$39,$3E,$43,$FD,$FB,$97,$E6,$3F,$EB,$AC,$11
	.dc.b	$A8,$FF,$00,$CF,$27,$FC,$A1,$97,$CD,$7D,$B7,$E1,$73,$87,$95,$DE
	.dc.b	$3E,$97,$8E,$3E,$D7,$9E,$5F,$09,$EB,$9F,$CC,$7F,$70,$43,$F2,$A7
	.dc.b	$C7,$DF,$E9,$3F,$27,$83,$EB,$3F,$2F,$87,$ED,$3F,$37,$8B,$EF,$3F
	.dc.b	$3F,$8F,$F0,$4F,$D1,$E4,$FC,$4E,$DF,$26,$3F,$C4,$7C,$BF,$8E,$7E
	.dc.b	$41,$0F,$DD,$1F,$27,$9F,$F3,$0F,$F8,$A3,$A3,$F3,$8F,$F8,$AB,$A7
	.dc.b	$F4,$0F,$F8,$B3,$AB,$F4,$8F,$F8,$BB,$AF,$F5,$0F,$F8,$C3,$AF,$F5
	.dc.b	$8F,$F8,$CB,$B3,$F6,$0F,$F8,$D3,$B7,$F6,$8F,$F8,$DB,$BB,$F4,$3E
	.dc.b	$CC,$8F,$C6,$3F,$BE,$7F,$C7,$BE,$07,$F0,$27,$FC,$7F,$E0,$FF,$06
	.dc.b	$7F,$C8,$3E,$17,$F0,$A7,$FC,$87,$E1,$FF,$0E,$7F,$C8,$BE,$27,$F1
	.dc.b	$27,$FC,$8F,$E2,$FF,$16,$7F,$C9,$3E,$37,$F1,$A7,$F7,$24,$3E,$3F
	.dc.b	$AB,$FE,$F1,$E4,$F8,$FE,$37,$F9,$33,$97,$E3,$BF,$E4,$EE,$5F,$8E
	.dc.b	$FF,$94,$39,$7E,$3B,$FE,$52,$E5,$F8,$EF,$F9,$53,$97,$E3,$BF,$E5
	.dc.b	$6E,$5F,$8E,$FF,$96,$39,$7E,$3B,$FC,$73,$4F,$C6,$3F,$E5,$CF,$1F
	.dc.b	$F7,$1A,$BC,$17,$3F,$0F,$1F,$FC,$CF,$8F,$47,$FC,$D3,$97,$81,$FF
	.dc.b	$35,$E7,$E4,$7F,$CD,$BA,$7A,$1F,$F3,$7E,$BE,$C7,$FC,$E3,$B7,$C0
	.dc.b	$7F,$CE,$7B,$FC,$47,$FA,$2E,$FF,$E3,$8D,$3F,$6A,$FE,$47,$7F,$F3
	.dc.b	$9F,$8D,$FB,$57,$F8,$6C,$7F,$10,$3F,$36,$7E,$74,$FE,$0E,$7E,$84
	.dc.b	$FE,$16,$7E,$94,$FE,$1E,$7E,$A4,$FE,$29,$E1,$FE,$D6,$3A,$F0,$3C
	.dc.b	$67,$89,$FB,$7B,$FD,$A9,$2B,$E6,$3F,$B7,$3F,$8F,$47,$3C,$BF,$9A
	.dc.b	$E1,$FC,$0E,$BA,$DF,$D1,$BF,$21,$FC,$28,$F4,$1F,$C3,$8F,$61,$FC
	.dc.b	$53,$BB,$F7,$0F,$FA,$63,$BB,$F7,$0F,$FA,$6B,$C7,$FE,$39,$E8,$FF
	.dc.b	$85,$F4,$FB,$D1,$FC,$7F,$1F,$97,$5F,$C0,$0F,$E6,$87,$CA,$3F,$90
	.dc.b	$F8,$3F,$C2,$1F,$E4,$7E,$17,$F0,$C7,$F9,$2F,$87,$FC,$41,$FE,$4F
	.dc.b	$E2,$7F,$14,$7F,$94,$F8,$BF,$C6,$1F,$E5,$7E,$37,$E4,$39,$3F,$D5
	.dc.b	$DC,$5F,$D2,$A7,$E4,$1F,$C9,$0F,$E3,$C7,$D4,$3E,$B1,$FC,$B0,$F8
	.dc.b	$F9,$FC,$AF,$1F,$F9,$93,$F3,$3C,$8F,$E6,$CF,$CE,$E2,$F8,$EF,$FA
	.dc.b	$F3,$8B,$E3,$BF,$EB,$EE,$0F,$8F,$D7,$F9,$2C,$7F,$D8,$7E,$07,$EF
	.dc.b	$5F,$F6,$2F,$83,$FC,$05,$FF,$63,$F8,$5F,$C1,$5F,$F6,$4F,$87,$FC
	.dc.b	$25,$FF,$65,$F8,$9F,$C3,$5F,$F6,$6F,$8B,$FC,$45,$FF,$67,$F8,$DF
	.dc.b	$C5,$5F,$F6,$8F,$8F,$FC,$65,$FE,$A3,$E4,$7F,$1D,$69,$FE,$D8,$F1
	.dc.b	$FF,$6C,$FF,$5B,$F1,$FF,$6C,$FF,$5D,$F1,$FF,$6C,$FF,$5F,$DF,$F6
	.dc.b	$CF,$F6,$1F,$1F,$F6,$CF,$F6,$3F,$1F,$F6,$CF,$F6,$5F,$1F,$F6,$CF
	.dc.b	$F6,$7F,$1F,$F6,$CF,$F4,$13,$FD,$5B,$E2,$34,$F3,$73,$7E,$71,$BD
	.dc.b	$7F,$E1,$46,$D7,$F8,$19,$BE,$1F,$EB,$5B,$7C,$E7,$F9,$D9,$0F,$82
	.dc.b	$4B,$EC,$E5,$AE,$6F,$C0,$E5,$AE,$4F,$11,$92,$99,$BF,$23,$96,$B9
	.dc.b	$AF,$33,$6F,$D2,$7D,$7F,$13,$F6,$12,$F4,$BF,$C6,$E7,$F8,$E3,$DF
	.dc.b	$FF,$1C,$FA,$9F,$E3,$AF,$7F,$FC,$77,$EB,$7F,$8F,$3D,$FF,$F1,$EF
	.dc.b	$B1,$FE,$3E,$F6,$DA,$CE,$0F,$E2,$BF,$C0,$3F,$C8,$A1,$18,$F8,$0D
	.dc.b	$C6,$38,$B8,$F1,$2D,$F0,$3A,$C3,$9F,$BB,$6F,$74,$F8,$0C,$62,$C7
	.dc.b	$97,$BF,$87,$A7,$BF,$6D,$EF,$FE,$FF,$D8,$FE,$9F,$18,$C3,$D5,$97
	.dc.b	$6B,$E1,$8F,$1C,$B9,$67,$CF,$5E,$9B,$75,$C3,$E2,$63,$FE,$4B,$E0
	.dc.b	$F9,$0F,$F9,$37,$8B,$E6,$3F,$E4,$FE,$4F,$A0,$FF,$94,$79,$BE,$9B
	.dc.b	$F9,$A6,$BF,$E4,$0E,$6F,$AA,$FF,$95,$F9,$BE,$AB,$FE,$59,$E6,$FA
	.dc.b	$AF,$F9,$6F,$9B,$C8,$FB,$0F,$8D,$C4,$D7,$D5,$19,$E5,$FA,$5C,$1E
	.dc.b	$F8,$C5,$38,$BE,$57,$1F,$CB,$E5,$F9,$55,$FD,$EB,$8B,$EA,$C6,$A3
	.dc.b	$8D,$73,$7F,$00,$78,$B0,$F0,$B8,$C7,$F0,$59,$7A,$23,$FC,$BF,$9F
	.dc.b	$B1,$FF,$30,$E9,$F0,$1F,$F3,$1E,$BF,$11,$F8,$C7,$CB,$21,$FE,$20
	.dc.b	$E7,$FD,$2F,$3F,$ED,$09,$AF,$EC,$9B,$3F,$B0,$6E,$5F,$AE,$6D,$72
	.dc.b	$AE,$04,$AE,$0D,$AE,$0D,$9E,$13,$A2,$4E,$9B,$1A,$8C,$11,$C1,$3A
	.dc.b	$7F,$7B,$3F,$C0,$E3,$0D,$8C,$BF,$C5,$BA,$7F,$80,$C7,$EC,$A3,$F6
	.dc.b	$75,$FC,$29,$B5,$FC,$29,$B3,$B7,$F8,$B9,$83,$13,$C5,$C9,$0F,$E3
	.dc.b	$26,$CF,$03,$B7,$26,$DC,$12,$7C,$51,$96,$BC,$09,$1B,$9F,$34,$62
	.dc.b	$77,$39,$8F,$3C,$60,$93,$C2,$75,$3E,$69,$CB,$C2,$75,$18,$8F,$E2
	.dc.b	$FF,$07,$E5,$9C,$92,$38,$10,$E8,$E5,$A6,$4F,$8C,$43,$A1,$92,$9C
	.dc.b	$08,$6C,$F8,$0F,$57,$15,$EC,$6D,$7A,$5B,$3B,$7A,$98,$3D,$1D,$0B
	.dc.b	$CC,$43,$B9,$E5,$29,$E4,$35,$7E,$16,$DF,$85,$B7,$E1,$78,$9E,$13
	.dc.b	$91,$83,$67,$79,$E8,$F8,$2F,$CC,$EC,$91,$A2,$57,$86,$32,$D8,$CD
	.dc.b	$78,$DB,$3D,$E7,$44,$9D,$36,$35,$18,$23,$FD,$9A,$BC,$B5,$FB,$B9
	.dc.b	$FD,$AE,$5F,$C3,$1B,$97,$F0,$76,$DF,$F0,$46,$D7,$18,$FD,$F1,$0F
	.dc.b	$F0,$01,$6B,$A3,$6B,$A3,$4E,$87,$F8,$6F,$2F,$E0,$AE,$6B,$B3,$D0
	.dc.b	$95,$D1,$B5,$D1,$A7,$42,$CF,$29,$E0,$49,$E0,$D9,$E0,$D9,$F0,$D7
	.dc.b	$92,$79,$E3,$FC,$79,$B8,$7C,$48,$C5,$F1,$6E,$1D,$89,$5D,$9B,$87
	.dc.b	$95,$B3,$C2,$74,$49,$D3,$FE,$34,$C3,$FD,$35,$C6,$F8,$3F,$F0,$71
	.dc.b	$FE,$0F,$F6,$FF,$C2,$1A,$FF,$84,$7D,$CF,$F0,$97,$C2,$7F,$84,$FD
	.dc.b	$EF,$F0,$A7,$C2,$FF,$85,$7E,$03,$FC,$2D,$F0,$DF,$E1,$7F,$82,$FF
	.dc.b	$0C,$7C,$3F,$F8,$67,$E1,$3F,$C3,$5F,$11,$FE,$1D,$FE,$01,$7C,$3D
	.dc.b	$F2,$31,$EC,$67,$CA,$9C,$DF,$DF,$6D,$1A,$8C,$57,$3B,$EB,$1F,$E0
	.dc.b	$0C,$7C,$2F,$F8,$0B,$3F,$2B,$FE,$04,$D7,$D2,$FF,$81,$B7,$F6,$BF
	.dc.b	$E0,$8E,$1F,$82,$FF,$82,$B8,$FE,$2B,$FE,$0C,$E4,$FF,$21,$F3,$7F
	.dc.b	$B0,$F6,$7B,$1E,$D3,$DD,$E6,$7B,$FD,$8F,$81,$E7,$7C,$1F,$63,$E1
	.dc.b	$7A,$1F,$0F,$D8,$F8,$9E,$97,$C5,$F6,$3E,$37,$A9,$F1,$FD,$8F,$91
	.dc.b	$EB,$7C,$9F,$63,$E5,$7B,$1F,$2F,$D9,$F9,$9F,$0F,$F2,$39,$7F,$C2
	.dc.b	$9C,$7F,$21,$F9,$5D,$DA,$7E,$65,$72,$7E,$76,$1D,$9F,$A1,$97,$89
	.dc.b	$FA,$5A,$79,$9F,$A9,$B7,$A9,$FA,$DC,$1E,$E7,$EC,$71,$7C,$27,$ED
	.dc.b	$72,$7C,$57,$FC,$E3,$CB,$F2,$63,$11,$E8,$E5,$F9,$0F,$E0,$79,$7E
	.dc.b	$43,$F8,$3E,$5F,$90,$FE,$17,$97,$E4,$3F,$87,$E5,$F9,$0F,$E2,$79
	.dc.b	$7E,$43,$F8,$BE,$5F,$90,$FE,$37,$97,$F6,$5E,$5F,$C5,$8F,$93,$C9
	.dc.b	$F1,$5F,$97,$C9,$F1,$5F,$9B,$C9,$F1,$5F,$9F,$C9,$F1,$5F,$A3,$C9
	.dc.b	$F1,$5F,$A7,$C9,$F1,$5F,$AB,$C9,$F1,$5F,$AF,$C7,$F0,$B9,$3E,$24
	.dc.b	$7C,$83,$F4,$5C,$9F,$11,$FB,$BC,$9F,$11,$FB,$FC,$9F,$11,$FC,$0F
	.dc.b	$27,$C4,$7F,$07,$C9,$F1,$1F,$C2,$F2,$7C,$47,$F0,$FC,$9F,$11,$FC
	.dc.b	$4F,$27,$C4,$7E,$31,$F8,$83,$E5,$1F,$88,$3E,$71,$F8,$83,$E9,$1F
	.dc.b	$88,$3E,$B1,$F8,$83,$ED,$1F,$88,$3E,$F1,$F8,$83,$DE,$7C,$13,$E2
	.dc.b	$9F,$88,$BF,$F5,$8F,$33,$DB,$FE,$18,$7F,$AA,$73,$FF,$23,$79,$8F
	.dc.b	$33,$FD,$6F,$0F,$73,$5F,$0B,$BF,$7B,$87,$C2,$F1,$F8,$0E,$5F,$0B
	.dc.b	$CF,$E0,$BA,$7C,$2F,$5F,$86,$ED,$F0,$DD,$FE,$1F,$C1,$F0,$FE,$1F
	.dc.b	$13,$C5,$F0,$7F,$B3,$E2,$F1,$73,$7F,$AF,$F9,$7F,$68,$FE,$DB,$C6
	.dc.b	$DB,$FE,$9B,$E4,$FF,$35,$0C,$7B,$46,$73,$F1,$3B,$D7,$CC,$F1,$DF
	.dc.b	$D4,$F3,$E1,$F7,$3D,$78,$FE,$37,$27,$C7,$8F,$92,$7F,$2F,$E6,$FE
	.dc.b	$3B,$CF,$F3,$BB,$FF,$6B,$C1,$EA,$F6,$1C,$2E,$FF,$CA,$4E,$61,$C3
	.dc.b	$F9,$58,$C7,$7F,$E5,$5D,$63,$F9,$57,$87,$7F,$E5,$5E,$5D,$FF,$95
	.dc.b	$7A,$77,$FE,$55,$ED,$DF,$F3,$7B,$FF,$07,$D1,$78,$7E,$74,$EE,$7F
	.dc.b	$0F,$7C,$DF,$C4,$E1,$DD,$FC,$5E,$3E,$37,$F1,$B9,$79,$DF,$C7,$E7
	.dc.b	$EB,$7F,$23,$A7,$C0,$7F,$27,$B7,$C6,$7E,$46,$7F,$97,$3F,$01,$C1
	.dc.b	$97,$F0,$53,$C5,$FC,$25,$75,$7F,$0D,$7E,$17,$F1,$18,$79,$5F,$C5
	.dc.b	$63,$E9,$7F,$19,$97,$B5,$FC,$76,$7F,$05,$F8,$67,$F8,$97,$D9,$FF
	.dc.b	$5B,$70,$36,$D7,$0F,$DC,$BE,$4F,$DE,$C3,$B3,$F8,$0C,$7C,$4F,$E0
	.dc.b	$B2,$F3,$3F,$84,$CF,$D4,$FE,$1B,$4F,$73,$F8,$8D,$7E,$13,$F8,$AD
	.dc.b	$BF,$1F,$B7,$C6,$3F,$80,$70,$FE,$2E,$B8,$ED,$F8,$C9,$FE,$10,$7F
	.dc.b	$20,$3F,$86,$1F,$C8,$0F,$E2,$07,$F2,$03,$F8,$A1,$FC,$80,$FE,$30
	.dc.b	$7F,$20,$3F,$8E,$1F,$C8,$0F,$8D,$D7,$FD,$0E,$BF,$39,$E4,$7F,$43
	.dc.b	$7F,$3D,$E4,$7F,$43,$7F,$41,$E4,$7F,$43,$7F,$45,$E4,$7F,$43,$7F
	.dc.b	$49,$E4,$7F,$43,$7F,$4D,$E4,$7F,$43,$7F,$51,$E4,$7F,$43,$7F,$55
	.dc.b	$B7,$F4,$3E,$2F,$E3,$FD,$5F,$B6,$D9,$5F,$B3,$E2,$FE,$3F,$D3,$FC
	.dc.b	$7B,$AF,$F8,$ED,$7F,$0B,$F1,$FF,$8E,$BF,$C3,$7C,$7F,$E3,$AF,$F0
	.dc.b	$FF,$1F,$F8,$EB,$FC,$47,$C7,$FE,$3A,$FF,$13,$F1,$FF,$8E,$BF,$C5
	.dc.b	$79,$BF,$8B,$BF,$C6,$4F,$F0,$69,$0F,$2E,$5F,$CC,$3C,$7F,$E4,$91
	.dc.b	$FC,$77,$C7,$FE,$36,$E8,$75,$FF,$08,$F1,$FF,$8E,$C6,$63,$F8,$5F
	.dc.b	$8F,$FC,$7A,$31,$C0,$F1,$FF,$8F,$C6,$3C,$7F,$E3,$EE,$BC,$7F,$E3
	.dc.b	$EF,$F2,$D3,$E5,$72,$7F,$17,$E5,$FC,$8E,$5F,$C6,$3E,$0F,$FA,$51
	.dc.b	$FD,$2F,$E1,$3F,$A6,$7B,$BF,$D3,$7E,$17,$FA,$77,$BF,$FD,$3F,$E1
	.dc.b	$BF,$A8,$7C,$0B,$49,$A7,$F2,$7F,$E0,$1F,$E8,$31,$EB,$5F,$07,$B8
	.dc.b	$C7,$C1,$DA,$E1,$F0,$97,$A7,$A7,$03,$17,$C4,$67,$1F,$93,$0F,$86
	.dc.b	$C5,$6F,$E2,$2C,$BE,$22,$FB,$18,$F6,$66,$3E,$8C,$FB,$DA,$77,$39
	.dc.b	$3B,$DF,$4C,$3A,$E5,$DB,$3E,$FA,$78,$36,$F0,$EF,$E2,$D3,$E3,$E5
	.dc.b	$FD,$7F,$87,$E5,$BF,$D8,$78,$FE,$7B,$FD,$8F,$97,$E9,$BF,$D9,$79
	.dc.b	$BE,$BF,$07,$FB,$A2,$3E,$67,$C3,$FB,$BD,$BF,$7F,$C3,$FC,$0F,$77
	.dc.b	$E0,$FC,$3F,$C2,$F7,$FE,$1F,$C3,$FC,$4F,$81,$F8,$BF,$0F,$F1,$BE
	.dc.b	$0F,$E3,$FC,$3F,$CA,$FE,$01,$FF,$4B,$ED,$F0,$7B,$90,$F8,$7D,$7C
	.dc.b	$46,$FE,$1B,$87,$AF,$FD,$4E,$FF,$B3,$C5,$BF,$F4,$3E,$1C,$DF,$F0
	.dc.b	$1F,$87,$FD,$B5,$61,$E1,$C7,$C5,$F1,$1E,$4B,$F1,$73,$35,$F4,$C6
	.dc.b	$37,$F6,$BA,$E1,$F8,$2F,$0E,$3F,$8A,$F2,$E5,$FC,$2F,$34,$8F,$EB
	.dc.b	$1C,$1F,$8F,$8C,$47,$E6,$67,$8B,$F9,$BA,$EA,$FE,$76,$FC,$2F,$E7
	.dc.b	$F0,$F2,$BF,$A1,$C7,$D4,$FE,$F7,$6F,$9A,$DD,$7E,$5B,$F1,$78,$3E
	.dc.b	$5B,$FC,$37,$83,$FC,$1D,$1F,$2F,$87,$E8,$4F,$CB,$E1,$F9,$0F,$CD
	.dc.b	$E1,$EF,$7F,$B3,$CB,$C6,$FE,$D7,$3F,$3B,$FB,$7D,$3D,$6F,$EE,$75
	.dc.b	$F7,$BF,$BB,$DB,$E1,$BF,$BD,$DF,$F8,$4F,$98,$E8,$FC,$BE,$9F,$C5
	.dc.b	$BA,$7F,$36,$FF,$18,$EA,$FC,$FB,$FC,$6B,$AF,$F4,$6F,$F0,$5E,$CF
	.dc.b	$D3,$BF,$7B,$BB,$F5,$6F,$E0,$3C,$0F,$D7,$BF,$82,$F0,$BF,$7B,$E1
	.dc.b	$FE,$DA,$3F,$87,$78,$9F,$BC,$69,$F8,$8F,$17,$6F,$13,$F1,$7E,$2F
	.dc.b	$EE,$5F,$C6,$F8,$BF,$B9,$7F,$1F,$E2,$FE,$E5,$FF,$2B,$78,$DF,$C8
	.dc.b	$5F,$F2,$C7,$8F,$FC,$95,$FF,$2D,$79,$1F,$CA,$5F,$DC,$10,$7C,$43
	.dc.b	$D9,$FF,$3A,$F2,$FC,$8E,$8F,$EB,$9C,$DF,$21,$FE,$BD,$CD,$F2,$1F
	.dc.b	$EC,$1C,$DF,$21,$FE,$5D,$CD,$F2,$32,$FE,$56,$7A,$8F,$90,$7B,$8F
	.dc.b	$80,$7F,$33,$E3,$F9,$13,$FC,$DF,$8F,$E4,$36,$7E,$41,$FB,$09,$FE
	.dc.b	$40,$F3,$3F,$9E,$72,$FD,$67,$C2,$7F,$3F,$E1,$F9,$0F,$F5,$8E,$1F
	.dc.b	$90,$FF,$5A,$E1,$F9,$0F,$F5,$CE,$1F,$90,$FF,$5E,$DF,$F1,$7C,$3F
	.dc.b	$1E,$3F,$9F,$F4,$7E,$4B,$6F,$E6,$FD,$3F,$96,$7F,$9C,$F5,$7E,$69
	.dc.b	$FE,$77,$D7,$F9,$E7,$F9,$EF,$67,$E8,$9F,$E7,$FD,$BF,$A6,$7F,$A7
	.dc.b	$F8,$7F,$B4,$E3,$FE,$29,$E1,$FE,$D3,$4F,$DC,$78,$7F,$B4,$7F,$C0
	.dc.b	$44,$9F,$DA,$BF,$C7,$7C,$3F,$DA,$BF,$BF,$F0,$FF,$6A,$FF,$01,$F0
	.dc.b	$FC,$0F,$F5,$AC,$FC,$8F,$F5,$CD,$3D,$0F,$F5,$ED,$7D,$8F,$F6,$0D
	.dc.b	$BE,$03,$FD,$8B,$7F,$88,$FC,$63,$F8,$DF,$8B,$FD,$1F,$4F,$DC,$47
	.dc.b	$F2,$13,$F9,$5F,$95,$FD,$1D,$FE,$5B,$E5,$7F,$47,$7F,$97,$F9,$5F
	.dc.b	$D1,$DF,$E6,$3E,$57,$F4,$77,$F9,$9F,$95,$FC,$D9,$FE,$79,$C7,$F1
	.dc.b	$F1,$FE,$79,$C7,$F1,$DF,$8B,$3F,$1F,$97,$FD,$32,$FC,$73,$F9,$01
	.dc.b	$FC,$6C,$E6,$7F,$1B,$3B,$9F,$D0,$B8,$7E,$3F,$95,$FE,$05,$E2,$F8
	.dc.b	$EF,$F8,$1F,$8B,$E3,$BF,$E0,$9E,$2F,$8E,$FF,$82,$F8,$BE,$3B,$FC
	.dc.b	$FF,$0F,$E8,$3E,$57,$F1,$B7,$F7,$19,$FE,$D1,$FD,$E9,$FD,$DC,$FE
	.dc.b	$02,$7F,$7B,$3F,$82,$9F,$D5,$B3,$FD,$A4,$7F,$88,$FC,$AF,$D3,$3F
	.dc.b	$E2,$5F,$03,$F8,$DD,$7F,$89,$7C,$AF,$E2,$AF,$F8,$9F,$CA,$FD,$9E
	.dc.b	$1F,$E1,$4F,$12,$CF,$F8,$82,$77,$87,$F8,$BE,$F9,$BD,$67,$F8,$11
	.dc.b	$FB,$47,$FC,$69,$E1,$FE,$D1,$FF,$35,$78,$7F,$B4,$7F,$CD,$9E,$1F
	.dc.b	$ED,$1F,$F3,$77,$87,$FB,$47,$FC,$E1,$E1,$FE,$D3,$C8,$FD,$3F,$C1
	.dc.b	$7F,$81,$C9,$97,$F7,$D6,$BF,$D7,$DB,$5F,$DF,$DB,$B7,$29,$E0,$43
	.dc.b	$81,$67,$83,$63,$83,$FC,$F8,$91,$FC,$65,$E2,$47,$8B,$63,$8C,$60
	.dc.b	$8E,$09,$1F,$E7,$6F,$0F,$FE,$00,$3E,$13,$15,$EA,$69,$9E,$3F,$F1
	.dc.b	$3C,$7C,$2F,$17,$F8,$D4,$7F,$8A,$BD,$BF,$F1,$71,$FE,$2F,$E5,$FF
	.dc.b	$18,$7B,$9F,$E3,$1F,$83,$FF,$19,$7B,$DF,$E3,$3D,$BF,$C6,$9E,$FF
	.dc.b	$F8,$D7,$E1,$3F,$C6,$DF,$03,$FE,$37,$F8,$5F,$F1,$C7,$C1,$FF,$8E
	.dc.b	$7E,$1B,$FC,$75,$F0,$BF,$E3,$BF,$87,$FF,$1E,$7B,$1F,$E6,$6F,$E0
	.dc.b	$1F,$F5,$4F,$2F,$F9,$AB,$C0,$FC,$6F,$27,$F9,$BD,$D4,$70,$3F,$C7
	.dc.b	$17,$D1,$FF,$1D,$63,$E0,$7F,$C7,$99,$F9,$1F,$F1,$F6,$BE,$87,$FC
	.dc.b	$81,$BF,$B1,$FF,$21,$70,$FC,$07,$FC,$89,$C7,$F1,$1F,$F2,$37,$27
	.dc.b	$FA,$5B,$9C,$E6,$3F,$C5,$53,$C5,$FF,$16,$57,$57,$FC,$5D,$7E,$17
	.dc.b	$FC,$61,$87,$95,$FF,$19,$63,$E9,$7F,$C6,$99,$7B,$5F,$F1,$B6,$7F
	.dc.b	$05,$FF,$5A,$6D,$F2,$36,$F8,$D1,$FE,$54,$DB,$E4,$CE,$76,$6F,$F4
	.dc.b	$6B,$FC,$C1,$C1,$F5,$5F,$F3,$17,$0F,$D9,$7F,$CC,$9C,$5F,$75,$FF
	.dc.b	$33,$71,$FE,$05,$FF,$34,$72,$7E,$15,$FF,$35,$72,$FE,$25,$F8,$E4
	.dc.b	$3F,$C1,$FC,$5F,$9A,$DB,$F2,$71,$F9,$5D,$3F,$36,$FE,$8C,$FC,$C9
	.dc.b	$63,$6D,$8D,$F4,$FE,$9A,$36,$D9,$DB,$67,$7D,$3E,$1D,$DF,$0F,$F9
	.dc.b	$12,$61,$E3,$EA,$FD,$56,$3C,$AF,$D1,$87,$2A,$EF,$18,$AD,$CE,$6B
	.dc.b	$85,$68,$8E,$09,$1F,$03,$AB,$F6,$6F,$C2,$9F,$F6,$A4,$FC,$06,$C6
	.dc.b	$5F,$E1,$E6,$2B,$FD,$B4,$D2,$9F,$28,$C1,$A9,$E8,$F2,$21,$B2,$BA
	.dc.b	$70,$47,$03,$1F,$97,$1F,$EB,$56,$BF,$96,$C7,$F9,$8B,$E2,$21,$DA
	.dc.b	$FD,$8F,$56,$D7,$66,$92,$FD,$B3,$B6,$CE,$DB,$3B,$90,$F0,$DE,$A7
	.dc.b	$86,$B6,$49,$F9,$37,$CD,$B7,$CD,$F9,$4D,$7D,$67,$8C,$DC,$F3,$76
	.dc.b	$D7,$EE,$FA,$3F,$EF,$D7,$F7,$04,$3F,$A3,$9D,$2B,$C4,$F2,$3A,$90
	.dc.b	$C1,$0E,$15,$E5,$7D,$C6,$CC,$D7,$C4,$7E,$25,$E9,$E4,$53,$67,$13
	.dc.b	$04,$3E,$01,$C0,$A7,$BB,$9B,$F9,$DE,$1E,$F8,$F6,$9C,$70,$F6,$BD
	.dc.b	$8E,$4E,$1D,$61,$EB,$7A,$9F,$DC,$31,$FB,$6E,$09,$3C,$5B,$3C,$63
	.dc.b	$04,$70,$7E,$43,$3F,$CC,$EC,$DB,$F1,$35,$F9,$FD,$7F,$83,$E1,$EA
	.dc.b	$3A,$EB,$F7,$5F,$3B,$5C,$9F,$24,$A7,$D8,$CB,$D3,$5F,$C1,$27,$F8
	.dc.b	$59,$27,$F8,$A1,$0F,$E4,$4E,$5A,$7F,$7F,$C7,$D3,$87,$A1,$B0,$AF
	.dc.b	$53,$CF,$1F,$E4,$D7,$CD,$B1,$9B,$FE,$3C,$D3,$F8,$DE,$5F,$C7,$CF
	.dc.b	$E3,$5B,$EA,$3C,$64,$AE,$6D,$AD,$46,$08,$FF,$45,$D3,$F0,$D1,$FE
	.dc.b	$29,$24,$70,$24,$F8,$B0,$F4,$46,$31,$F9,$53,$E7,$6B,$FE,$32,$DF
	.dc.b	$F9,$7D,$70,$8E,$46,$F2,$F6,$46,$6B,$8E,$1E,$6A,$D9,$FE,$3E,$C3
	.dc.b	$FC,$C8,$FE,$E6,$35,$3C,$27,$C9,$7D,$2B,$B5,$F5,$8C,$47,$C8,$C3
	.dc.b	$FC,$6B,$9B,$DB,$FF,$13,$1F,$E2,$7F,$87,$FF,$14,$7B,$BF,$E2,$9F
	.dc.b	$87,$FF,$15,$7B,$FF,$E2,$BF,$87,$FF,$16,$7C,$0F,$F8,$B7,$E1,$FF
	.dc.b	$C5,$DF,$07,$FE,$2F,$F8,$7F,$F1,$87,$C2,$FF,$8C,$7E,$1F,$FC,$65
	.dc.b	$F0,$FF,$E3,$3F,$2F,$FC,$69,$F0,$BF,$E3,$5F,$37,$F3,$FD,$3F,$A3
	.dc.b	$E5,$FC,$57,$27,$E3,$3A,$7F,$17,$1F,$E1,$4A,$E2,$FF,$43,$C3,$AB
	.dc.b	$FD,$17,$2F,$0B,$FD,$1F,$4F,$2B,$FD,$27,$6F,$4B,$FD,$2F,$83,$DA
	.dc.b	$FF,$4D,$E2,$F8,$2F,$F4,$FE,$4F,$8A,$FF,$51,$E5,$FE,$39,$CD,$F2
	.dc.b	$BE,$17,$FC,$9C,$7F,$93,$FC,$FF,$F2,$87,$C2,$FF,$94,$7D,$1F,$E4
	.dc.b	$5C,$7E,$DE,$27,$C2,$FB,$F3,$4E,$61,$E9,$F0,$EF,$FE,$0B,$1F,$D3
	.dc.b	$6B,$FA,$7D,$6F,$E1,$B5,$1F,$B1,$97,$AF,$8C,$7B,$FC,$3D,$97,$AF
	.dc.b	$5C,$7B,$19,$F2,$75,$E5,$6F,$D9,$FA,$BB,$7E,$FF,$CB,$BA,$F7,$9D
	.dc.b	$BE,$49,$F3,$DF,$48,$C1,$E5,$C7,$C5,$97,$8F,$3F,$26,$BE,$5D,$BC
	.dc.b	$DE,$5F,$9F,$87,$D1,$9F,$CB,$CB,$FC,$05,$C7,$F3,$DF,$F0,$27,$2F
	.dc.b	$D2,$7F,$B9,$F3,$7D,$57,$27,$F8,$27,$9B,$EA,$BF,$E0,$BE,$6E,$AF
	.dc.b	$9B,$97,$EA,$E5,$E2,$8C,$70,$35,$F3,$C6,$37,$F5,$BA,$9F,$80,$FD
	.dc.b	$CD,$7E,$1B,$CB,$93,$E6,$3F,$E1,$BE,$4F,$97,$97,$E7,$B9,$BF,$05
	.dc.b	$CB,$F3,$2B,$91,$F7,$79,$7A,$CF,$E1,$F2,$F0,$CE,$AF,$EF,$BE,$8E
	.dc.b	$4F,$98,$FA,$B9,$3E,$63,$EC,$D3,$E6,$3F,$E8,$0E,$4F,$98,$FF,$A0
	.dc.b	$B9,$3E,$63,$FE,$6E,$3F,$6A,$E3,$2F,$CE,$78,$0E,$4F,$9B,$5F,$E8
	.dc.b	$AE,$4F,$9A,$F9,$75,$FC,$EB,$E3,$3E,$6F,$1F,$CC,$75,$C7,$EC,$C7
	.dc.b	$DD,$1F,$4B,$9F,$F2,$CF,$CB,$3E,$AF,$17,$CA,$CF,$F8,$41,$FC,$27
	.dc.b	$14,$7C,$BB,$FD,$77,$17,$CA,$DF,$F5,$B5,$DD,$FD,$7E,$5E,$3A,$FD
	.dc.b	$76,$3E,$73,$D5,$CB,$EC,$8F,$83,$AF,$C0,$7E,$1E,$DF,$17,$0F,$EC
	.dc.b	$DB,$FF,$08,$F2,$FF,$77,$F0,$9F,$B3,$3F,$69,$ED,$FE,$D7,$E1,$3F
	.dc.b	$6D,$EE,$FE,$DF,$E1,$3F,$71,$EF,$FE,$E7,$E1,$3F,$75,$F0,$3F,$BB
	.dc.b	$F8,$4F,$DF,$7F,$00,$BE,$1A,$FC,$36,$3E,$1F,$3F,$0F,$AF,$88,$DF
	.dc.b	$B3,$C3,$E2,$38,$FC,$44,$EB,$FD,$16,$68,$D4,$62,$B9,$3A,$C3,$B3
	.dc.b	$C3,$2F,$1E,$7E,$58,$FD,$4E,$BE,$97,$F5,$7B,$FB,$5F,$D6,$F0,$FC
	.dc.b	$17,$F5,$FC,$7F,$15,$FD,$8F,$27,$C9,$7F,$22,$7E,$F0,$FF,$6B,$F4
	.dc.b	$7D,$1E,$8F,$F0,$A7,$57,$D5,$7F,$94,$F5,$FE,$EE,$BF,$25,$E1,$7D
	.dc.b	$D9,$FB,$F1,$FB,$EF,$2B,$F9,$23,$F8,$F3,$F8,$0F,$81,$F8,$99,$FE
	.dc.b	$67,$E0,$FE,$3F,$C8,$FD,$CF,$85,$F8,$EB,$FC,$A1,$F9,$FB,$FC,$73
	.dc.b	$FC,$84,$EC,$7F,$09,$3F,$6C,$7E,$88,$FC,$77,$83,$FD,$0F,$3F,$54
	.dc.b	$7F,$64,$D3,$DC,$FF,$66,$D7,$F1,$CF,$C4,$3F,$6D,$E4,$7F,$AA,$7C
	.dc.b	$2C,$BF,$EA,$B9,$E2,$FF,$AB,$6B,$AB,$FC,$82,$BC,$3A,$FF,$50,$DF
	.dc.b	$E3,$47,$F4,$DF,$0F,$F4,$CF,$AC,$FD,$E7,$8B,$FB,$47,$F7,$DE,$4F
	.dc.b	$F1,$E7,$FD,$6C,$7F,$1F,$21,$FC,$0F,$87,$E6,$3F,$C7,$4F,$A0,$7F
	.dc.b	$07,$E0,$EF,$97,$F1,$C8,$FE,$15,$E4,$FF,$1C,$7F,$86,$79,$3F,$C7
	.dc.b	$1F,$E1,$DA,$7F,$1C,$E1,$FE,$2B,$E4,$FF,$1C,$7F,$8B,$F0,$FF,$1C
	.dc.b	$F2,$BF,$A3,$5F,$F1,$F3,$FA,$31,$FD,$38,$FE,$8C,$7F,$27,$3F,$A1
	.dc.b	$F1,$FE,$2E,$BF,$82,$C7,$90,$FB,$87,$F1,$FE,$6F,$C5,$BF,$18,$FC
	.dc.b	$59,$F8,$68,$FE,$7E,$FC,$47,$F9,$5F,$37,$E3,$37,$F8,$BB,$7F,$82
	.dc.b	$39,$BF,$16,$FF,$82,$B8,$3E,$28,$E3,$FA,$F1,$FE,$0E,$E4,$F3,$70
	.dc.b	$7E,$8B,$3F,$51,$FC,$D4,$F7,$1F,$AA,$3F,$59,$9F,$C6,$7F,$B5,$F9
	.dc.b	$3F,$C6,$1F,$D9,$9F,$B6,$3F,$B2,$F9,$3F,$C6,$1F,$EC,$FE,$4F,$F1
	.dc.b	$87,$FB,$4F,$93,$FC,$61,$FE,$D7,$E4,$FF,$06,$7F,$84,$1F,$C4,$3C
	.dc.b	$9F,$E2,$EF,$F1,$2F,$27,$F8,$BB,$FC,$53,$C9,$FE,$2C,$FF,$5F,$F2
	.dc.b	$7F,$8B,$3F,$C6,$3C,$9F,$E2,$EE,$7C,$9F,$E2,$EE,$FC,$9F,$E2,$EF
	.dc.b	$1F,$27,$C4,$FF,$55,$CF,$CC,$FF,$57,$D3,$D4,$FF,$59,$D7,$DC,$FF
	.dc.b	$5B,$DB,$F9,$CB,$F1,$0F,$EB,$DE,$27,$F1,$7E,$46,$BE,$69,$F4,$6B
	.dc.b	$EA,$E2,$FF,$22,$4E,$5B,$19,$E3,$FF,$62,$CE,$5B,$19,$AF,$F5,$84
	.dc.b	$E5,$B1,$9B,$F0,$B6,$75,$18,$9F,$90,$43,$CA,$5B,$F1,$B6,$FC,$2D
	.dc.b	$3B,$96,$7A,$3C,$89,$3C,$9B,$3C,$9A,$72,$2C,$F2,$9E,$1D,$8E,$4F
	.dc.b	$87,$EE,$FF,$60,$3F,$B0,$FC,$47,$F6,$2F,$7F,$FB,$1C,$FF,$64,$F8
	.dc.b	$0F,$EC,$BF,$11,$FD,$9B,$E0,$BF,$B3,$FC,$47,$F6,$8F,$84,$FE,$D3
	.dc.b	$F1,$1F,$DA,$BE,$1B,$FB,$5F,$C4,$7F,$6C,$F8,$8F,$F3,$57,$F0,$0B
	.dc.b	$8E,$F8,$F8,$F6,$3F,$A1,$47,$EE,$79,$BF,$DF,$75,$FE,$AD,$8F,$F2
	.dc.b	$63,$FE,$9B,$39,$3F,$DB,$AF,$BB,$FD,$C3,$1F,$1B,$FD,$CB,$3F,$3B
	.dc.b	$FD,$D3,$5F,$5B,$FD,$DB,$7F,$7B,$FD,$E3,$87,$E1,$BF,$DE,$B8,$FE
	.dc.b	$33,$FB,$A3,$F0,$EE,$1F,$D5,$71,$FC,$97,$FC,$03,$C7,$F2,$5F,$F0
	.dc.b	$1F,$1F,$C9,$7F,$C0,$BC,$7F,$25,$FF,$03,$F1,$FC,$97,$FC,$13,$C7
	.dc.b	$F2,$5F,$82,$7C,$83,$E1,$F1,$7C,$87,$FA,$29,$FE,$59,$E1,$F9,$0D
	.dc.b	$E3,$FA,$2F,$33,$F8,$05,$77,$7F,$81,$5F,$8D,$FE,$09,$87,$9D,$FE
	.dc.b	$0D,$8F,$AD,$FE,$11,$97,$BD,$F8,$27,$C7,$21,$FD,$AB,$9F,$F4,$E4
	.dc.b	$E8,$FD,$51,$23,$F4,$11,$96,$C6,$5F,$CD,$C6,$5B,$19,$E7,$FD,$21
	.dc.b	$39,$BF,$48,$D9,$FD,$23,$7A,$3F,$49,$3A,$24,$E9,$B1,$A8,$C1,$1C
	.dc.b	$1F,$B6,$E8,$FD,$64,$67,$0F,$35,$F7,$6D,$F4,$6C,$EF,$0F,$54,$E5
	.dc.b	$B1,$93,$E1,$90,$F8,$25,$AE,$D1,$C3,$0F,$2D,$F6,$6C,$7F,$40,$C3
	.dc.b	$CF,$18,$BE,$58,$78,$5A,$78,$4E,$E4,$C3,$BB,$A9,$E4,$E6,$79,$34
	.dc.b	$E5,$3C,$1C,$4F,$08,$C1,$FC,$5B,$1E,$2E,$2A,$92,$B9,$B6,$B9,$37
	.dc.b	$1F,$80,$4C,$7E,$23,$6B,$D0,$D7,$C6,$DB,$F1,$42,$DB,$BD,$9F,$E3
	.dc.b	$67,$F1,$CF,$87,$FE,$3B,$ED,$FF,$1E,$F8,$7F,$E3,$FE,$EF,$F2,$0F
	.dc.b	$87,$FE,$43,$EF,$FF,$22,$F8,$7F,$E4,$7F,$03,$FC,$93,$E1,$FF,$92
	.dc.b	$FC,$1F,$F2,$6F,$87,$FE,$4F,$F0,$BF,$CA,$3E,$1F,$F9,$4F,$C3,$FF
	.dc.b	$2A,$F6,$BF,$95,$FC,$0C,$E5,$FE,$4A,$E5,$FE,$4F,$3C,$5F,$E5,$37
	.dc.b	$D5,$FE,$57,$8F,$85,$FE,$5B,$9F,$95,$FE,$5F,$AF,$A5,$FE,$63,$BF
	.dc.b	$B5,$FE,$67,$C3,$F0,$5F,$E6,$BC,$7F,$11,$FE,$4D,$CB,$F2,$3D,$CF
	.dc.b	$E8,$27,$F4,$2F,$81,$FE,$73,$D0,$EB,$FE,$89,$4F,$E0,$1C,$61,$FD
	.dc.b	$1E,$AC,$E7,$DE,$E3,$7C,$2D,$EF,$E3,$0F,$B1,$0F,$6B,$18,$75,$F8
	.dc.b	$0B,$3F,$E3,$FA,$ED,$87,$3F,$80,$CF,$BD,$AF,$82,$DF,$C0,$70,$F8
	.dc.b	$2E,$3F,$05,$CB,$D7,$E7,$F0,$7F,$8E,$F2,$2E,$9D,$0F,$1C,$F0,$71
	.dc.b	$BF,$BF,$0E,$B1,$8D,$99,$F8,$A3,$1A,$FA,$3E,$0F,$D3,$C1,$EA,$E1
	.dc.b	$F5,$F1,$7B,$38,$FD,$BC,$7F,$3F,$1F,$ED,$3C,$DF,$49,$FA,$BA,$3D
	.dc.b	$AF,$86,$7C,$4F,$84,$F8,$BE,$E7,$C6,$F8,$5F,$8F,$EF,$7C,$8F,$86
	.dc.b	$F9,$3F,$01,$F3,$3C,$CF,$F7,$46,$7E,$C5,$F0,$39,$F8,$6D,$7C,$1E
	.dc.b	$FE,$1B,$87,$C2,$F1,$F8,$6E,$5D,$DF,$99,$CB,$F6,$6D,$F0,$FF,$03
	.dc.b	$9F,$F8,$C0,$57,$2A,$F0,$61,$D9,$F1,$65,$E5,$CF,$CD,$F0,$FE,$89
	.dc.b	$F9,$DB,$7B,$1F,$A1,$C1,$F0,$1F,$A5,$C5,$F1,$1F,$A9,$C7,$FE,$53
	.dc.b	$E4,$F9,$11,$83,$67,$D9,$E4,$F9,$0F,$DB,$E4,$F9,$0F,$DD,$E4,$F1
	.dc.b	$BF,$1F,$4F,$8C,$FC,$9E,$4F,$8C,$FC,$BE,$4F,$8C,$FC,$DE,$4F,$89
	.dc.b	$CD,$FC,$7D,$B1,$FC,$5F,$93,$E3,$72,$7C,$58,$72,$FC,$AE,$5F,$8D
	.dc.b	$18,$C3,$F6,$36,$E5,$F9,$13,$F6,$F9,$7E,$43,$B3,$F5,$BC,$BF,$25
	.dc.b	$C7,$7F,$EE,$79,$BE,$54,$63,$91,$CD,$F2,$E3,$1C,$DF,$E7,$DF,$59
	.dc.b	$D1,$F9,$59,$FF,$0D,$78,$DF,$9A,$7F,$49,$D5,$F9,$E7,$F4,$DD,$9F
	.dc.b	$A2,$7F,$51,$DD,$FB,$CE,$FF,$D4,$47,$87,$C0,$FD,$6B,$E3,$F0,$7F
	.dc.b	$64,$F9,$7C,$2F,$DB,$BF,$E2,$9F,$3F,$F9,$63,$4F,$DA,$78,$9F,$BF
	.dc.b	$7F,$6D,$E2,$FF,$03,$7F,$71,$E3,$7F,$07,$7F,$75,$E3,$F8,$DF,$29
	.dc.b	$FB,$DF,$1F,$F7,$0F,$EF,$FC,$7F,$DC,$3F,$C0,$7B,$7F,$70,$FF,$8E
	.dc.b	$F8,$FE,$27,$B3,$FD,$E3,$97,$E3,$CF,$EF,$0F,$F3,$2F,$2F,$C7,$7E
	.dc.b	$87,$2F,$C7,$7E,$97,$2F,$C7,$7E,$A7,$2F,$95,$FA,$DA,$FA,$5F,$B1
	.dc.b	$B7,$B5,$FB,$5B,$FC,$17,$EE,$70,$7F,$1F,$7E,$29,$FC,$67,$8B,$E6
	.dc.b	$F8,$1F,$CE,$3A,$5C,$BF,$4A,$31,$CD,$F5,$9D,$73,$FF,$1D,$C3,$E3
	.dc.b	$1F,$78,$F8,$C7,$E0,$8F,$8C,$7C,$03,$E1,$1F,$1C,$FE,$3B,$D1,$F8
	.dc.b	$F9,$F1,$74,$FE,$51,$F2,$75,$7E,$65,$F3,$76,$7E,$79,$FE,$1F,$DB
	.dc.b	$FA,$2C,$3F,$3B,$DF,$FA,$57,$F6,$C7,$F1,$03,$F6,$C7,$F2,$1F,$07
	.dc.b	$F8,$DB,$FE,$98,$F4,$7E,$33,$FE,$75,$72,$F4,$8F,$DB,$5F,$ED,$63
	.dc.b	$A9,$E0,$F0,$7C,$51,$E4,$8F,$E1,$47,$F1,$A3,$F8,$71,$FC,$68,$FE
	.dc.b	$24,$7F,$14,$3F,$5E,$7F,$17,$1D,$7F,$E4,$2E,$1F,$E3,$71,$A3,$F7
	.dc.b	$1A,$FC,$7E,$DF,$ED,$DB,$7C,$77,$FB,$86,$DF,$1D,$FE,$13,$B7,$99
	.dc.b	$FE,$17,$9F,$F0,$EC,$FD,$98,$FC,$AF,$13,$F8,$DB,$F3,$3C,$4F,$E3
	.dc.b	$3E,$1F,$F8,$EA,$3F,$CD,$DD,$EE,$1F,$8F,$87,$F4,$2E,$1F,$8F,$9F
	.dc.b	$F1,$8F,$13,$B3,$E0,$C3,$F9,$61,$FC,$64,$F8,$E7,$F1,$3D,$BD,$4F
	.dc.b	$F8,$1B,$7F,$73,$FE,$08,$E0,$FC,$45,$7F,$83,$79,$3F,$8B,$78,$3F
	.dc.b	$C5,$73,$FF,$09,$EA,$F0,$7F,$61,$1F,$9F,$F0,$7F,$60,$FE,$87,$C1
	.dc.b	$FD,$83,$FA,$3F,$07,$F6,$0F,$E9,$7C,$1F,$D8,$3F,$A7,$F0,$7F,$60
	.dc.b	$FE,$A7,$C1,$FD,$83,$FA,$BF,$07,$F8,$AB,$E2,$F1,$2E,$FF,$E2,$39
	.dc.b	$DB,$FE,$25,$AE,$6F,$F8,$9E,$FB,$BF,$E2,$9C,$3C,$6F,$F8,$AF,$1F
	.dc.b	$3B,$FE,$2D,$CB,$D6,$FF,$8B,$F3,$F7,$BF,$EF,$6D,$3E,$23,$FE,$F8
	.dc.b	$DB,$FD,$EB,$DB,$FE,$8E,$EE,$7C,$0F,$F8,$1C,$9C,$9F,$E0,$7F,$81
	.dc.b	$F4,$BF,$E6,$FE,$37,$16,$3A,$FF,$CE,$2D,$8C,$CF,$F9,$80,$87,$F3
	.dc.b	$72,$72,$7C,$57,$2D,$73,$A7,$F9,$42,$79,$C6,$0F,$15,$77,$23,$CA
	.dc.b	$78,$46,$23,$85,$F9,$9D,$1C,$8F,$71,$FE,$25,$C7,$E0,$BF,$C3,$1F
	.dc.b	$8F,$ED,$FF,$72,$F8,$2F,$F1,$61,$FE,$2D,$7F,$C5,$D9,$FF,$8B,$FE
	.dc.b	$0F,$FC,$61,$EF,$7F,$8C,$7E,$13,$FC,$65,$F0,$1F,$E3,$3F,$85,$FF
	.dc.b	$1A,$7C,$17,$F8,$D7,$E1,$BF,$C6,$DF,$09,$FE,$37,$F8,$7F,$F3,$1F
	.dc.b	$F0,$0B,$82,$FC,$3E,$3E,$23,$3E,$D7,$F1,$8E,$8F,$F3,$8D,$FF,$A8
	.dc.b	$6F,$FC,$B2,$AB,$8C,$60,$EC,$7F,$90,$F1,$F1,$3F,$E4,$5C,$FC,$CF
	.dc.b	$F9,$1F,$5F,$53,$FE,$49,$DF,$DC,$FC,$83,$F9,$8F,$17,$C4,$7F,$C9
	.dc.b	$DC,$7F,$21,$FF,$10,$39,$76,$71,$CF,$FC,$93,$C5,$F2,$23,$FC,$AD
	.dc.b	$C5,$F2,$1F,$F2,$C7,$17,$C8,$7F,$CB,$5C,$5F,$21,$FF,$2E,$71,$7C
	.dc.b	$87,$DE,$7F,$88,$38,$3E,$13,$F1,$38,$7E,$47,$3F,$FA,$72,$7E,$44
	.dc.b	$EF,$8B,$8E,$9F,$32,$3E,$39,$FD,$3B,$8B,$E3,$BF,$D4,$38,$BE,$3B
	.dc.b	$FD,$4B,$8B,$E3,$BF,$D5,$38,$BE,$3B,$FD,$5B,$93,$F1,$0F,$C6,$21
	.dc.b	$FD,$03,$8F,$FC,$7F,$CB,$F9,$2C,$BF,$CD,$AF,$E6,$DF,$D3,$13,$5F
	.dc.b	$CF,$D7,$E7,$39,$BF,$3C,$FE,$83,$A3,$F5,$35,$B8,$FD,$01,$9E,$8E
	.dc.b	$97,$CA,$30,$73,$AE,$31,$9A,$E5,$7D,$9B,$3A,$AE,$B1,$8B,$EB,$18
	.dc.b	$BF,$6B,$FA,$F2,$1F,$04,$FF,$33,$61,$F0,$63,$71,$E4,$25,$75,$7C
	.dc.b	$D5,$DB,$0F,$25,$79,$C9,$3E,$12,$46,$9B,$87,$A6,$30,$47,$04,$AE
	.dc.b	$CD,$9F,$09,$0F,$E1,$07,$B2,$FE,$CB,$FE,$F0,$E9,$FE,$5E,$E7,$0F
	.dc.b	$6C,$FF,$AA,$B3,$FC,$3B,$FE,$AC,$D3,$F1,$6F,$FA,$BB,$5F,$C7,$BF
	.dc.b	$EB,$0D,$BF,$26,$FF,$AC,$B7,$FC,$BB,$F9,$9C,$7F,$81,$F4,$FF,$09
	.dc.b	$7F,$44,$7F,$B7,$BE,$21,$E6,$7F,$BC,$9B,$19,$8F,$E1,$EF,$F0,$CE
	.dc.b	$3F,$D2,$EB,$FA,$98,$EE,$FF,$A5,$B1,$F1,$BF,$E9,$8C,$BC,$EF,$FA
	.dc.b	$6B,$3F,$5B,$FE,$9C,$D3,$DE,$FF,$A7,$B5,$FC,$06,$1F,$82,$3F,$0B
	.dc.b	$D9,$FD,$5F,$8F,$3C,$6E,$1F,$97,$5F,$AD,$3F,$DB,$DD,$7F,$C2,$DF
	.dc.b	$F2,$2F,$5F,$F0,$B7,$FC,$8F,$D7,$FC,$2D,$FF,$24,$F5,$FF,$0B,$7F
	.dc.b	$C9,$7D,$7F,$C2,$DF,$92,$7C,$13,$E4,$9F,$EE,$22,$F3,$7F,$32,$E3
	.dc.b	$F8,$B0,$CD,$D5,$FC,$EE,$3F,$8B,$9F,$CC,$4F,$E3,$67,$F3,$13,$F8
	.dc.b	$F9,$FC,$C4,$FE,$46,$7F,$31,$3D,$A7,$F2,$9E,$0F,$C4,$91,$F9,$07
	.dc.b	$EA,$E7,$C3,$F1,$1F,$D3,$0F,$E9,$BE,$EF,$E6,$BA,$E5,$7E,$67,$2F
	.dc.b	$D6,$3F,$AE,$C3,$F5,$38,$7E,$8E,$B7,$39,$3F,$52,$E2,$B8,$92,$B8
	.dc.b	$F5,$FE,$B7,$26,$3E,$3C,$3F,$AB,$E5,$FE,$CF,$C7,$96,$1C,$1B,$3D
	.dc.b	$8F,$34,$65,$A6,$4F,$26,$5E,$77,$2D,$73,$8F,$EE,$5B,$5E,$CE,$37
	.dc.b	$67,$F4,$5F,$88,$FE,$E6,$7F,$74,$F7,$FF,$BF,$DF,$D2,$3F,$9F,$1F
	.dc.b	$58,$FE,$84,$7D,$A3,$FA,$31,$F7,$8F,$E9,$47,$E0,$8F,$EA,$C7,$E2
	.dc.b	$31,$FC,$3E,$FF,$B0,$8F,$C7,$1F,$D6,$FB,$FF,$74,$FF,$5D,$EF,$FC
	.dc.b	$B3,$F9,$83,$F1,$87,$E7,$0F,$C6,$1F,$A0,$3F,$18,$7E,$90,$FC,$61
	.dc.b	$FE,$AA,$E6,$F6,$BF,$DA,$70,$F8,$2F,$F8,$33,$1F,$EF,$98,$FE,$71
	.dc.b	$CC,$6C,$FC,$A1,$FA,$1E,$75,$FE,$CF,$8D,$C1,$FE,$1D,$C3,$F7,$5C
	.dc.b	$FF,$B5,$9C,$E1,$C7,$9F,$F6,$D1,$8A,$FE,$01,$CF,$FB,$78,$C6,$4E
	.dc.b	$7F,$87,$1F,$E2,$4D,$BE,$3B,$FE,$26,$DF,$FC,$03,$97,$ED,$79,$FF
	.dc.b	$C0,$33,$FB,$98,$FD,$F9,$F5,$0F,$AC,$FF,$2D,$7E,$D3,$FC,$C5,$FB
	.dc.b	$CF,$F3,$57,$F0,$4F,$F3,$97,$F0,$CF,$F3,$D7,$FC,$6B,$C1,$F2,$23
	.dc.b	$FC,$73,$C1,$FA,$19,$FC,$BF,$2F,$E5,$67,$F5,$51,$F9,$D6,$EB,$F9
	.dc.b	$B6,$CE,$F8,$3F,$49,$5F,$9C,$39,$F2,$FE,$AA,$F8,$4E,$AF,$8C,$FF
	.dc.b	$99,$F9,$7F,$43,$7F,$E0,$63,$F8,$C5,$FF,$83,$1B,$87,$F9,$D1,$B5
	.dc.b	$FC,$75,$A5,$39,$1F,$E0,$D3,$26,$E7,$A3,$92,$1C,$08,$E9,$B3,$E1
	.dc.b	$77,$19,$24,$66,$37,$3E,$36,$9E,$59,$DC,$65,$B1,$99,$E5,$18,$8F
	.dc.b	$44,$60,$93,$C5,$B1,$C6,$3F,$A0,$7B,$2F,$63,$CF,$5A,$AE,$75,$C4
	.dc.b	$8F,$12,$1A,$29,$F2,$88,$74,$78,$BB,$21,$A3,$E0,$98,$38,$1D,$0A
	.dc.b	$64,$D9,$C6,$FE,$13,$67,$D4,$D9,$F7,$D7,$28,$C1,$D4,$E6,$64,$86
	.dc.b	$09,$5D,$9C,$B5,$CD,$78,$1C,$B5,$CD,$7C,$27,$C2,$7A,$E7,$FC,$CB
	.dc.b	$7F,$4E,$36,$79,$63,$B1,$2B,$A3,$6B,$A4,$F8,$1A,$F2,$3B,$4F,$02
	.dc.b	$46,$A7,$94,$F7,$9C,$B6,$7E,$4B,$FB,$FC,$BF,$6C,$47,$04,$AF,$DD
	.dc.b	$B6,$BF,$81,$36,$BF,$80,$46,$27,$64,$70,$48,$D9,$FC,$4E,$7F,$85
	.dc.b	$13,$2F,$E0,$EF,$F0,$C2,$19,$BF,$35,$F5,$AE,$73,$AA,$E2,$F4,$CB
	.dc.b	$F8,$7D,$6E,$35,$7C,$B0,$F2,$46,$0F,$61,$67,$83,$E1,$32,$57,$37
	.dc.b	$FE,$1C,$BE,$FA,$FF,$92,$6B,$FD,$CF,$C3,$FE,$F0,$3F,$DE,$3F,$01
	.dc.b	$FE,$F2,$F8,$8F,$F7,$9F,$C1,$7F,$BD,$3E,$23,$FD,$EB,$F0,$9F,$EF
	.dc.b	$6F,$88,$FF,$7B,$FC,$37,$FB,$E3,$E2,$3F,$DF,$3F,$11,$FE,$FA,$F6
	.dc.b	$FF,$DF,$7C,$9F,$EF,$CF,$37,$FD,$FB,$E7,$7F,$90,$F8,$FF,$AE,$7F
	.dc.b	$00,$FF,$60,$E2,$46,$3D,$0D,$74,$7F,$67,$A7,$4E,$2F,$FC,$9D,$18
	.dc.b	$47,$29,$E1,$5B,$E8,$57,$7C,$3F,$B6,$73,$B2,$F3,$63,$CA,$73,$A7
	.dc.b	$AE,$31,$97,$FB,$A7,$6F,$81,$18,$3F,$B6,$70,$7C,$57,$19,$FF,$98
	.dc.b	$B8,$BE,$54,$63,$37,$27,$CF,$8C,$73,$7D,$37,$EA,$9E,$CE,$8F,$C3
	.dc.b	$EE,$F4,$7D,$8D,$9E,$ED,$7E,$11,$F1,$BE,$13,$7E,$9F,$0F,$82,$E3
	.dc.b	$EA,$F2,$F8,$3E,$7E,$BF,$4F,$07,$AF,$B1,$DB,$E1,$3B,$FB,$5E,$0F
	.dc.b	$6F,$C3,$EE,$78,$BE,$27,$C7,$EF,$79,$3B,$FF,$DA,$9C,$47,$3C,$5E
	.dc.b	$6A,$E2,$FA,$30,$EA,$FA,$B2,$F0,$BE,$CD,$3C,$AF,$BB,$6F,$4B,$F0
	.dc.b	$38,$3D,$AF,$C2,$E2,$F8,$BC,$72,$3E,$57,$A7,$F8,$B8,$C4,$7C,$6E
	.dc.b	$AF,$C2,$C7,$CC,$AE,$73,$9D,$98,$78,$23,$18,$F9,$1D,$65,$E8,$78
	.dc.b	$67,$EC,$79,$69,$F0,$1E,$9A,$FC,$47,$F7,$1E,$4F,$FB,$CF,$27,$07
	.dc.b	$CD,$9F,$BB,$C3,$F4,$5F,$BF,$C5,$F5,$5F,$C0,$F1,$FD,$97,$F0,$7C
	.dc.b	$9F,$75,$FC,$2F,$2F,$E0,$5F,$C3,$F3,$7B,$DF,$C4,$F1,$FC,$4E,$7F
	.dc.b	$C7,$BF,$1B,$C0,$FD,$BB,$B9,$F8,$95,$CD,$F8,$D7,$DD,$F9,$18,$78
	.dc.b	$DF,$95,$8F,$9D,$F9,$99,$7A,$DF,$9D,$9F,$BD,$FA,$1A,$7F,$2A,$69
	.dc.b	$F8,$93,$E1,$9F,$89,$3E,$19,$F8,$93,$E1,$9F,$89,$3E,$19,$F8,$93
	.dc.b	$E1,$9F,$89,$3E,$19,$F8,$93,$E1,$90,$FF,$66,$6A,$F8,$9F,$D8,$92
	.dc.b	$3F,$63,$F0,$1F,$B4,$F8,$9F,$E2,$1F,$C0,$2F,$87,$BF,$15,$8F,$88
	.dc.b	$CF,$C5,$6B,$E1,$F7,$F1,$3C,$3E,$27,$8F,$C3,$F2,$F8,$2E,$7F,$07
	.dc.b	$D3,$E2,$7A,$F8,$DF,$AF,$E1,$FE,$03,$E8,$FF,$04,$30,$FE,$77,$83
	.dc.b	$FC,$BA,$DA,$E6,$FF,$16,$86,$1E,$08,$C6,$5E,$47,$5A,$7A,$1E,$1B
	.dc.b	$7B,$1E,$5C,$1F,$05,$F8,$67,$F1,$4E,$3F,$8C,$FC,$AE,$CF,$C6,$79
	.dc.b	$FF,$B3,$8F,$A0,$7E,$30,$FA,$87,$E3,$0F,$B0,$7E,$30,$FB,$87,$E3
	.dc.b	$0F,$C0,$1F,$8C,$3F,$08,$7E,$30,$FC,$41,$F8,$A3,$F7,$9E,$4F,$F1
	.dc.b	$63,$F6,$5D,$FF,$D5,$3C,$2F,$DB,$3F,$D5,$BC,$3F,$DD,$3F,$D6,$3C
	.dc.b	$4F,$DF,$3F,$D6,$BC,$5F,$E0,$4F,$F5,$CF,$1B,$F8,$33,$FD,$7B,$C7
	.dc.b	$FE,$16,$FF,$60,$F2,$3E,$0E,$7F,$D6,$B5,$FF,$66,$F7,$FF,$39,$AF
	.dc.b	$D7,$CF,$E4,$5E,$0F,$E5,$8F,$E2,$97,$DB,$6F,$EC,$D8,$F8,$9F,$ED
	.dc.b	$19,$79,$9F,$ED,$59,$FA,$9F,$ED,$9A,$7B,$9F,$ED,$DA,$FC,$27,$FC
	.dc.b	$25,$E8,$7F,$69,$F2,$7F,$8A,$78,$5F,$B5,$D7,$FB,$CF,$87,$FB,$97
	.dc.b	$FB,$DF,$89,$FB,$D7,$FB,$EF,$8B,$FC,$05,$FE,$FF,$E3,$7F,$05,$7F
	.dc.b	$C0,$3E,$3F,$F0,$97,$FC,$07,$E4,$7F,$0D,$7E,$0F,$1F,$F3,$1E,$5F
	.dc.b	$8A,$7C,$8F,$2B,$F9,$BC,$7E,$22,$78,$BF,$88,$7A,$BF,$88,$7C,$2F
	.dc.b	$E2,$1F,$2B,$F8,$87,$D2,$FE,$21,$F6,$BF,$88,$7F,$C2,$FA,$7F,$13
	.dc.b	$6C,$67,$7F,$F0,$B6,$BF,$0D,$FF,$0C,$6B,$F0,$DF,$F0,$D6,$BF,$0D
	.dc.b	$FF,$0E,$6B,$F0,$DF,$F0,$F6,$BF,$0D,$FF,$10,$6B,$F0,$DF,$D3,$63
	.dc.b	$F0,$CF,$EB,$67,$F8,$8B,$A7,$F5,$B9,$FF,$8B,$1F,$F0,$2F,$93,$FB
	.dc.b	$39,$FE,$05,$D7,$FB,$77,$F8,$27,$67,$EE,$DF,$E0,$DD,$BF,$BF,$7F
	.dc.b	$84,$77,$7F,$03,$7F,$85,$77,$FF,$08,$7F,$85,$1F,$C2,$7C,$3F,$E1
	.dc.b	$F9,$FF,$8F,$3C,$3F,$E2,$53,$FE,$14,$F0,$7F,$8B,$BF,$C3,$FC,$3E
	.dc.b	$54,$DF,$FC,$73,$CE,$F0,$FF,$89,$DF,$F1,$5F,$17,$F9,$23,$FC,$5F
	.dc.b	$C6,$FE,$50,$FF,$19,$F1,$FF,$96,$3F,$C6,$FC,$8F,$E6,$3C,$7F,$07
	.dc.b	$3F,$F2,$7F,$27,$C6,$D3,$FC,$99,$CB,$FE,$48,$C1,$E5,$7F,$9D,$35
	.dc.b	$7C,$0F,$F3,$A2,$4F,$F2,$7F,$85,$FE,$7B,$F0,$5F,$CF,$BE,$1B,$F9
	.dc.b	$FF,$C1,$FF,$40,$F8,$7F,$E8,$3F,$09,$FD,$0B,$E2,$3F,$A1,$FC,$2F
	.dc.b	$F9,$4F,$B3,$C1,$9F,$F8,$45,$7C,$2E,$7E,$1B,$5F,$0D,$BF,$87,$E1
	.dc.b	$F1,$1C,$7E,$1F,$F6,$9E,$7C,$CB,$FC,$9F,$92,$77,$18,$BE,$6E,$B1
	.dc.b	$F1,$65,$E3,$8F,$29,$FD,$0F,$4F,$4B,$FD,$17,$6F,$6B,$FD,$1F,$83
	.dc.b	$E0,$BF,$0C,$FE,$95,$C7,$F1,$9F,$97,$E1,$FD,$CD,$1F,$11,$F0,$49
	.dc.b	$3E,$8F,$88,$F8,$9E,$E7,$C5,$F8,$8F,$8D,$EF,$7C,$7F,$88,$F9,$1F
	.dc.b	$03,$F2,$7E,$23,$E5,$7C,$1F,$CB,$F8,$8F,$99,$F0,$BF,$37,$E2,$3E
	.dc.b	$77,$4F,$CF,$F8,$8F,$A1,$F1,$1F,$96,$F1,$BF,$D6,$FF,$09,$FE,$55
	.dc.b	$E2,$F9,$33,$B7,$E5,$DF,$37,$E6,$E3,$DD,$F9,$F9,$F8,$DF,$A3,$AF
	.dc.b	$9D,$FA,$7B,$FA,$DF,$AB,$C3,$F8,$1E,$2F,$83,$1F,$81,$E3,$FC,$97
	.dc.b	$8D,$FE,$C4,$E5,$F8,$F3,$F7,$27,$7E,$4F,$E9,$E1,$E7,$FF,$78,$C3
	.dc.b	$E4,$C7,$E0,$B9,$FE,$4B,$F8,$4E,$7F,$92,$FE,$1B,$9F,$E4,$BF,$88
	.dc.b	$E7,$F9,$2F,$E2,$B9,$FE,$4B,$F8,$CE,$7F,$92,$FE,$EC,$FF,$4E,$FB
	.dc.b	$DF,$EE,$DE,$7F,$95,$1F,$9B,$E7,$F9,$4F,$E7,$79,$FE,$53,$F9,$FE
	.dc.b	$7F,$94,$FE,$87,$9F,$E5,$3F,$A3,$E7,$F9,$4F,$E9,$79,$FE,$53,$FA
	.dc.b	$7E,$7F,$94,$FC,$52,$1F,$90,$3E,$2F,$93,$FB,$A7,$F2,$A4,$3F,$C4
	.dc.b	$9E,$4F,$E6,$9F,$F1,$37,$6F,$E7,$9F,$F1,$47,$77,$E8,$9F,$F1,$57
	.dc.b	$7F,$E9,$9F,$DC,$1F,$AA,$3F,$58,$7E,$CB,$C5,$FD,$C3,$FB,$4F,$17
	.dc.b	$F7,$0F,$F8,$D8,$87,$F0,$1F,$17,$F6,$EF,$F0,$3F,$17,$F6,$EF,$F0
	.dc.b	$5F,$17,$B3,$E0,$3F,$C7,$1E,$37,$F0,$D7,$FC,$75,$E3,$FF,$11,$7F
	.dc.b	$C7,$9E,$47,$F1,$57,$FC,$7D,$E4,$FF,$95,$37,$FD,$B7,$85,$F8,$BF
	.dc.b	$0B,$F9,$87,$91,$F9,$2F,$23,$F9,$9F,$27,$C9,$C7,$FC,$95,$3C,$3C
	.dc.b	$8C,$E3,$F2,$71,$E5,$6E,$5F,$95,$18,$E5,$F9,$4E,$B9,$7E,$53,$E1
	.dc.b	$F1,$FF,$91,$B7,$C9,$FE,$45,$A7,$CB,$3F,$8D,$1F,$2C,$FE,$3F,$CB
	.dc.b	$FC,$85,$FF,$5D,$78,$0E,$EF,$9F,$1F,$CB,$7A,$3E,$9B,$FC,$BF,$A7
	.dc.b	$AB,$FC,$C7,$0F,$E4,$4F,$CC,$3F,$91,$1F,$30,$FE,$44,$7C,$C3,$F9
	.dc.b	$11,$F3,$0F,$E4,$47,$C6,$3F,$1D,$E4,$FE,$EF,$C9,$FD,$CF,$17,$F2
	.dc.b	$37,$F3,$33,$F9,$B3,$F7,$07,$F0,$83,$F7,$07,$F0,$C3,$F7,$07,$F1
	.dc.b	$03,$F7,$07,$F1,$43,$F7,$07,$F1,$83,$F7,$07,$F1,$C3,$F7,$07,$C6
	.dc.b	$C7,$E4,$69,$F8,$1F,$0B,$F7,$2F,$D3,$F0,$BF,$72,$FD,$59,$FD,$CB
	.dc.b	$F5,$FC,$2F,$DC,$BF,$67,$C2,$FD,$CB,$F6,$FC,$2F,$DC,$BF,$77,$C2
	.dc.b	$FD,$CB,$F7,$FC,$2F,$82,$FE,$07,$4F,$E4,$7B,$7C,$7D,$7F,$77,$39
	.dc.b	$7F,$A3,$1F,$CA,$8F,$8E,$7F,$05,$3E,$39,$FC,$24,$F8,$E7,$F0,$D3
	.dc.b	$E3,$9F,$C4,$4F,$8E,$7F,$15,$3E,$39,$FC,$64,$F8,$EF,$E9,$78,$BF
	.dc.b	$8F,$7A,$5F,$C8,$A7,$FB,$C7,$89,$FC,$99,$FE,$F5,$E2,$FF,$2A,$7F
	.dc.b	$BE,$78,$DF,$CB,$9F,$EF,$DE,$3F,$F3,$27,$FC,$01,$E4,$7F,$36,$7F
	.dc.b	$C0,$5E,$4F,$F3,$A7,$FC,$09,$E5,$7E,$21,$FE,$7F,$D9,$FD,$11,$FE
	.dc.b	$3D,$E2,$7E,$E1,$FE,$39,$B7,$EC,$BC,$3F,$DF,$47,$E5,$4F,$E3,$A7
	.dc.b	$E6,$4F,$E3,$A7,$E7,$4F,$E3,$A7,$E8,$4F,$E1,$C7,$E9,$4F,$E3,$A7
	.dc.b	$EA,$4F,$D9,$9F,$AD,$3F,$66,$7F,$1C,$8F,$F2,$13,$83,$47,$F1,$D3
	.dc.b	$F7,$87,$F1,$D3,$FC,$33,$D7,$FB,$27,$FD,$B1,$E4,$7F,$1D,$7F,$DB
	.dc.b	$5E,$47,$F1,$D7,$FD,$B9,$E4,$7F,$1D,$7F,$DB,$DE,$47,$C0,$7F,$64
	.dc.b	$4E,$57,$91,$FE,$4E,$FE,$01,$FE,$D5,$2C,$FF,$91,$46,$1B,$E0,$FF
	.dc.b	$91,$5B,$23,$84,$73,$34,$57,$51,$C1,$D9,$FD,$E2,$7B,$4E,$E3,$3B
	.dc.b	$27,$31,$B9,$EA,$D3,$AB,$C6,$BB,$11,$C1,$23,$F1,$3E,$27,$F0,$08
	.dc.b	$FF,$37,$F4,$FF,$27,$9F,$AD,$5C,$5E,$C7,$92,$FC,$78,$79,$6F,$C6
	.dc.b	$D9,$F1,$9C,$2B,$94,$66,$39,$4F,$0A,$E6,$47,$04,$8E,$73,$C6,$BE
	.dc.b	$56,$5F,$6A,$7E,$57,$4F,$32,$3C,$30,$7A,$9D,$FD,$FF,$07,$C3,$78
	.dc.b	$7D,$AF,$17,$C3,$78,$FD,$CF,$27,$C3,$79,$7D,$EF,$37,$C3,$79,$FE
	.dc.b	$03,$D1,$F0,$DE,$9F,$82,$F5,$7C,$37,$AF,$E1,$3F,$A7,$F2,$7F,$07
	.dc.b	$F3,$BE,$77,$1F,$D0,$3C,$93,$C1,$F3,$5F,$47,$D1,$8F,$81,$F5,$67
	.dc.b	$E4,$7D,$9A,$FA,$1F,$76,$FE,$C7,$E0,$70,$FC,$DE,$2F,$87,$1F,$67
	.dc.b	$93,$FB,$E3,$F2,$0F,$8C,$EB,$5F,$D0,$CF,$C7,$9F,$95,$C7,$F1,$DF
	.dc.b	$99,$C7,$F1,$DF,$9D,$C7,$F1,$DF,$A1,$C7,$F1,$DF,$A5,$C7,$F1,$DF
	.dc.b	$A9,$C7,$F1,$DF,$C7,$F3,$FE,$4B,$93,$E3,$F3,$BC,$1F,$F4,$D5,$7C
	.dc.b	$83,$89,$F2,$0E,$A7,$C8,$3C,$27,$C8,$3C,$A7,$C8,$3D,$27,$C8,$3D
	.dc.b	$A7,$C8,$3F,$11,$AF,$C3,$BF,$E0,$DB,$FC,$8B,$FC,$94,$7E,$AD,$B5
	.dc.b	$F9,$F8,$C7,$3F,$E6,$E3,$F2,$51,$F9,$B7,$F4,$26,$08,$60,$9C,$1F
	.dc.b	$A8,$9E,$51,$FB,$1F,$86,$FF,$75,$BF,$C1,$BD,$1F,$CE,$BF,$9F,$8E
	.dc.b	$D5,$C6,$E9,$E4,$2C,$F9,$C9,$87,$A8,$98,$78,$2B,$94,$F9,$BA,$BF
	.dc.b	$61,$5F,$BB,$8F,$F7,$43,$5F,$E3,$46,$2B,$F9,$6B,$6F,$FD,$BE,$E8
	.dc.b	$87,$F0,$E3,$46,$A7,$A3,$93,$6E,$09,$3D,$A3,$2D,$7F,$60,$7F,$87
	.dc.b	$CF,$F4,$BE,$BF,$B6,$AD,$9E,$48,$EC,$4A,$EB,$19,$6C,$79,$6B,$B4
	.dc.b	$62,$3F,$DF,$6D,$7D,$66,$23,$67,$F5,$0D,$1C,$1E,$29,$E9,$7D,$EB
	.dc.b	$6D,$24,$68,$87,$57,$2D,$32,$78,$88,$77,$3C,$FB,$7E,$B9,$F8,$C4
	.dc.b	$CF,$FC,$E9,$D9,$F9,$98,$E2,$7C,$5B,$F2,$BD,$8E,$86,$B2,$F8,$4E
	.dc.b	$08,$7F,$4C,$CB,$E0,$BC,$F1,$F4,$BC,$6B,$AB,$EB,$21,$98,$DB,$D2
	.dc.b	$BB,$BC,$CF,$C5,$63,$ED,$C7,$F9,$A4,$7E,$08,$FE,$22,$6C,$EA,$73
	.dc.b	$BF,$0B,$DC,$F7,$1B,$30,$70,$3D,$99,$71,$C3,$CB,$1B,$30,$72,$C3
	.dc.b	$CE,$4A,$EF,$1F,$9A,$D7,$F7,$F8,$7F,$8A,$A3,$F8,$11,$FC,$20,$F0
	.dc.b	$90,$EA,$78,$EB,$A5,$F8,$9B,$1F,$E1,$9C,$BE,$8C,$7F,$00,$9F,$E0
	.dc.b	$AD,$9F,$E2,$B5,$FC,$09,$FE,$2B,$3F,$C2,$DC,$B5,$C9,$FC,$33,$0F
	.dc.b	$EF,$B8,$7F,$3C,$EC,$FF,$07,$4F,$A8,$DE,$1D,$DB,$5F,$C4,$6F,$A6
	.dc.b	$1E,$12,$38,$3C,$1B,$7F,$14,$9E,$15,$E1,$9F,$5C,$62,$FB,$5F,$2A
	.dc.b	$D1,$EA,$29,$E4,$D3,$F1,$17,$FD,$56,$3F,$10,$7F,$BD,$CE,$15,$D2
	.dc.b	$3C,$91,$CA,$BB,$4E,$63,$13,$A8,$FF,$12,$61,$E9,$8F,$39,$27,$AE
	.dc.b	$1E,$92,$46,$63,$E3,$B6,$FF,$6F,$1E,$67,$F3,$EF,$9C,$D9,$D2,$38
	.dc.b	$1E,$39,$EB,$19,$21,$FD,$56,$BE,$77,$C3,$7F,$8E,$49,$1F,$E2,$FF
	.dc.b	$73,$FC,$79,$EE,$FF,$8F,$7E,$23,$FC,$8B,$FC,$02,$F8,$0B,$F0,$38
	.dc.b	$F8,$8C,$FC,$1E,$BE,$23,$7F,$0B,$C3,$E2,$38,$FC,$3F,$2E,$5E,$7F
	.dc.b	$13,$D3,$E2,$7F,$C9,$5D,$3F,$A6,$E6,$FF,$3A,$74,$3B,$BF,$B2,$F2
	.dc.b	$FE,$3F,$A5,$5A,$E3,$63,$D2,$BF,$C9,$F1,$FE,$11,$6E,$9E,$37,$3B
	.dc.b	$79,$DD,$F0,$7A,$DE,$3C,$5F,$03,$8F,$E1,$47,$F9,$4F,$97,$FC,$C1
	.dc.b	$CD,$F2,$23,$FC,$D9,$CF,$FE,$0D,$CF,$FC,$B9,$1F,$E8,$26,$ED,$E8
	.dc.b	$87,$02,$73,$BF,$AF,$D2,$F9,$FF,$0F,$F4,$3C,$FF,$A3,$F0,$DF,$4B
	.dc.b	$D1,$FA,$7F,$0D,$F5,$3D,$3F,$AB,$F0,$DF,$5B,$D5,$FF,$3D,$FF,$00
	.dc.b	$FF,$A7,$77,$FF,$38,$AF,$5F,$3F,$13,$AF,$87,$DF,$C5,$7D,$2B,$FC
	.dc.b	$87,$0F,$E9,$DF,$A0,$FD,$4C,$13,$C6,$7A,$DF,$83,$0F,$0C,$78,$CF
	.dc.b	$A7,$97,$9D,$FA,$BA,$7A,$DF,$AF,$B7,$BD,$FB,$3C,$1F,$90,$E1,$FC
	.dc.b	$8F,$B5,$27,$FD,$71,$E3,$FF,$86,$B9,$3E,$4C,$7C,$C6,$F2,$FC,$DA
	.dc.b	$E5,$3D,$0F,$C2,$F2,$FC,$C7,$F0,$FC,$BF,$31,$FC,$4F,$2F,$CC,$7F
	.dc.b	$17,$CB,$EC,$7E,$5F,$07,$CA,$7E,$6F,$2F,$CA,$7F,$C7,$47,$CA,$3E
	.dc.b	$FF,$93,$FC,$A3,$9F,$FD,$AB,$E1,$6E,$7F,$D8,$FC,$EB,$E9,$18,$C3
	.dc.b	$C0,$FE,$2F,$1F,$23,$F8,$DC,$BD,$0F,$E3,$F3,$F6,$3F,$91,$D3,$F1
	.dc.b	$6F,$E9,$38,$7E,$53,$F9,$5D,$BF,$30,$63,$5F,$CB,$6F,$F2,$67,$81
	.dc.b	$F2,$4E,$87,$D8,$3E,$49,$F7,$0F,$92,$7E,$00,$F9,$27,$E4,$8F,$C3
	.dc.b	$1F,$88,$3E,$49,$F9,$6E,$7F,$C9,$3F,$E1,$9E,$8F,$C9,$57,$C8,$8D
	.dc.b	$1C,$0F,$93,$D1,$D1,$F9,$67,$E4,$4F,$D9,$74,$7E,$45,$FD,$A7,$47
	.dc.b	$E4,$5F,$DB,$74,$7E,$45,$FD,$C7,$47,$E2,$1F,$8C,$7F,$0C,$D3,$FD
	.dc.b	$15,$C8,$F8,$9F,$D7,$92,$7F,$5F,$E6,$FE,$CB,$E0,$FF,$67,$EE,$FE
	.dc.b	$D3,$E1,$3F,$6B,$EF,$FE,$DB,$E1,$7F,$6F,$F0,$3F,$B8,$F8,$6F,$DC
	.dc.b	$FC,$1F,$EE,$BE,$1F,$F7,$7F,$0B,$FB,$CF,$88,$FD,$EF,$C3,$FE,$FB
	.dc.b	$DA,$FD,$FF,$C4,$7F,$29,$E1,$FE,$1B,$E9,$7F,$28,$EB,$FB,$7C,$AA
	.dc.b	$E1,$5F,$C1,$30,$E8,$FF,$06,$CB,$C0,$FF,$08,$D3,$C8,$FF,$0A,$DB
	.dc.b	$D0,$FF,$0C,$E0,$F6,$3F,$C3,$B8,$BE,$03,$F0,$8F,$E2,$3C,$BF,$15
	.dc.b	$79,$3F,$B9,$E7,$FC,$87,$6F,$F3,$B7,$E8,$6F,$F9,$2F,$25,$D5,$F5
	.dc.b	$63,$1D,$7F,$65,$D7,$67,$DD,$78,$76,$FE,$05,$F8,$39,$FE,$15,$F8
	.dc.b	$7D,$DF,$C6,$63,$F6,$3D,$FF,$90,$9F,$ED,$E4,$D3,$FA,$9F,$B2,$F2
	.dc.b	$7F,$71,$3F,$33,$C9,$FD,$BB,$F3,$BC,$9F,$DB,$BE,$9F,$27,$F6,$EF
	.dc.b	$AF,$C9,$FD,$BB,$ED,$F2,$7F,$6E,$FB,$FC,$9F,$DB,$EB,$FD,$27,$C9
	.dc.b	$F8,$AF,$F1,$DE,$5F,$DB,$BF,$93,$21,$B3,$F6,$D5,$CD,$FD,$B1,$FC
	.dc.b	$14,$FD,$B1,$FC,$24,$FD,$B1,$FC,$34,$FD,$B1,$FC,$78,$FD,$B1,$FC
	.dc.b	$5B,$3F,$83,$B7,$C4,$D7,$FC,$89,$E0,$FC,$98,$FE,$8D,$C1,$F3,$70
	.dc.b	$F9,$F8,$73,$3F,$6A,$7F,$04,$3F,$6A,$7F,$08,$3F,$6A,$7F,$0C,$3F
	.dc.b	$6A,$7F,$38,$3D,$B3,$FD,$9D,$F8,$31,$F8,$7C,$7E,$2C,$7E,$E0,$91
	.dc.b	$F8,$BC,$BE,$6B,$F1,$63,$A9,$F1,$4F,$09,$F1,$4F,$29,$F1,$4F,$49
	.dc.b	$FC,$D4,$F8,$AF,$E0,$DF,$E2,$A7,$F6,$83,$FA,$0E,$1F,$DC,$E3,$F1
	.dc.b	$91,$FE,$02,$E0,$FC,$63,$C2,$39,$3F,$C7,$0F,$90,$FE,$69,$F9,$0F
	.dc.b	$E7,$9F,$90,$FE,$89,$F9,$0F,$E9,$9F,$90,$7F,$1C,$3F,$9D,$9F,$13
	.dc.b	$0F,$D8,$3F,$DB,$30,$FE,$FD,$D1,$FB,$7A,$FE,$12,$FF,$83,$3C,$8F
	.dc.b	$DE,$D7,$F5,$8E,$AF,$E0,$2F,$F5,$AE,$BF,$E0,$AF,$F5,$CE,$CF,$E1
	.dc.b	$2F,$F5,$EE,$DF,$E1,$AF,$F6,$0E,$EF,$E2,$2F,$F6,$2E,$FF,$E2,$AF
	.dc.b	$F6,$4F,$03,$F8,$E7,$93,$FE,$43,$F6,$BF,$8F,$CF,$F8,$73,$8B,$F9
	.dc.b	$1B,$FE,$1E,$F1,$7F,$93,$BF,$E2,$0F,$1B,$F9,$5B,$FE,$22,$F1,$FF
	.dc.b	$97,$BF,$E2,$4F,$23,$F9,$9B,$FE,$26,$F2,$7F,$9B,$BF,$E2,$8F,$2B
	.dc.b	$E0,$3F,$C6,$88,$7F,$1A,$E6,$FF,$12,$7C,$4F,$F8,$C0,$90,$1C,$FF
	.dc.b	$DB,$3E,$03,$FC,$85,$18,$E9,$FF,$2F,$C3,$DE,$FC,$34,$63,$82,$D7
	.dc.b	$F9,$22,$9F,$07,$8E,$24,$3E,$0F,$0B,$E1,$2F,$C2,$63,$E1,$73,$3A
	.dc.b	$F8,$6D,$FC,$27,$0F,$87,$E3,$F0,$FC,$BE,$23,$E7,$ED,$FD,$CC,$FC
	.dc.b	$D9,$FE,$4C,$EC,$D3,$EF,$E3,$F7,$57,$38,$C7,$17,$BB,$1F,$04,$60
	.dc.b	$F6,$67,$E5,$71,$C1,$EC,$DB,$D7,$BF,$B3,$83,$DB,$C3,$EE,$E2,$F7
	.dc.b	$F1,$FC,$0D,$3E,$8E,$5F,$E2,$CE,$4F,$F3,$17,$AA,$E6,$F9,$FF,$11
	.dc.b	$F1,$89,$5E,$0D,$DF,$13,$F2,$BC,$BF,$97,$ED,$FC,$CF,$73,$E6,$FB
	.dc.b	$BF,$3B,$E2,$7E,$97,$F0,$0F,$FA,$97,$2F,$4A,$F8,$1C,$FC,$56,$BE
	.dc.b	$0F,$7F,$15,$C3,$E1,$78,$FC,$57,$2F,$87,$F9,$FE,$07,$F0,$6E,$E7
	.dc.b	$C3,$7F,$92,$7A,$7F,$CE,$5E,$16,$F7,$EF,$8F,$37,$C3,$9F,$77,$C7
	.dc.b	$AF,$9B,$6F,$3F,$C3,$FA,$67,$E0,$71,$7B,$5F,$85,$C9,$F0,$5F,$89
	.dc.b	$CD,$F1,$9F,$CC,$1F,$93,$E8,$F9,$BD,$3F,$32,$3E,$54,$6F,$87,$F4
	.dc.b	$15,$CE,$78,$F0,$FE,$63,$0F,$04,$63,$A5,$97,$96,$37,$CD,$F3,$5B
	.dc.b	$AF,$CC,$7E,$7F,$37,$CC,$7E,$8F,$37,$CC,$7E,$9F,$37,$CC,$E5,$FF
	.dc.b	$25,$92,$3F,$D6,$7F,$0D,$FE,$25,$76,$7C,$C3,$99,$F3,$0E,$E7,$CC
	.dc.b	$3E,$FE,$1F,$31,$FC,$57,$47,$E1,$9F,$C6,$74,$FE,$29,$FC,$77,$57
	.dc.b	$ED,$B8,$7E,$36,$3F,$26,$7F,$86,$1E,$4E,$CF,$CD,$F1,$FF,$8E,$9F
	.dc.b	$CF,$47,$E5,$FB,$7F,$44,$FE,$67,$BF,$F4,$CF,$E6,$FC,$1F,$E0,$2F
	.dc.b	$E7,$7C,$8F,$E0,$2F,$E7,$FC,$8F,$E0,$2F,$E8,$7C,$8F,$E0,$2F,$E8
	.dc.b	$FC,$8F,$CE,$FC,$57,$F3,$2F,$88,$FF,$26,$6D,$FE,$F6,$E8,$FF,$77
	.dc.b	$5F,$D9,$BE,$2F,$DB,$C3,$AB,$F7,$71,$F0,$BF,$7F,$2F,$2B,$F8,$1C
	.dc.b	$FD,$2F,$E0,$F4,$F6,$BF,$85,$D7,$E0,$BF,$BD,$3F,$DB,$BB,$7F,$9D
	.dc.b	$0F,$8F,$A7,$F3,$48,$FD,$E3,$FC,$19,$FD,$F7,$8D,$FC,$29,$FE,$01
	.dc.b	$E3,$FF,$0E,$7F,$81,$79,$1F,$C4,$DF,$CA,$76,$FF,$18,$7F,$2D,$E4
	.dc.b	$FF,$1D,$7F,$31,$E5,$7F,$22,$7F,$AC,$1F,$EC,$1F,$07,$F7,$3E,$8F
	.dc.b	$E4,$63,$F9,$D7,$1F,$C6,$E7,$FE,$89,$C9,$F1,$9F,$E8,$DC,$9F,$19
	.dc.b	$FE,$91,$C9,$F1,$9F,$E9,$5C,$9F,$19,$FE,$99,$C9,$F1,$9F,$E9,$DC
	.dc.b	$9F,$19,$FC,$46,$5F,$18,$FE,$FC,$7E,$E4,$FD,$B9,$0F,$D9,$74,$7E
	.dc.b	$DE,$3F,$AC,$78,$5F,$B6,$7F,$AD,$78,$5F,$B5,$7F,$D8,$FD,$1F,$B5
	.dc.b	$6F,$77,$ED,$70,$FD,$CF,$85,$FB,$57,$F7,$7E,$17,$ED,$5F,$DE,$F8
	.dc.b	$5F,$17,$93,$E3,$C7,$C9,$F0,$BF,$68,$E8,$FD,$A1,$C8,$FD,$A1,$D8
	.dc.b	$FD,$A1,$FC,$18,$FE,$67,$E2,$F9,$FB,$7F,$B7,$EB,$EB,$7F,$B8,$ED
	.dc.b	$EF,$7F,$B9,$EF,$FB,$4B,$F8,$C7,$ED,$31,$FE,$DF,$E2,$65,$B1,$FC
	.dc.b	$98,$FD,$E9,$FC,$A8,$FE,$02,$7F,$2E,$3F,$82,$9E,$43,$F8,$49,$E8
	.dc.b	$DF,$E5,$ED,$F1,$DF,$9B,$B7,$C7,$7E,$7E,$DF,$1F,$CA,$FE,$89,$7F
	.dc.b	$DD,$0F,$C8,$9F,$1C,$FE,$53,$E4,$FF,$45,$7F,$95,$F9,$3F,$D1,$5F
	.dc.b	$E5,$BE,$4F,$F4,$57,$C8,$7F,$33,$F2,$3F,$8D,$3F,$CD,$7C,$8F,$E3
	.dc.b	$4F,$F3,$7F,$23,$F8,$D3,$F8,$9C,$FF,$C3,$DB,$FE,$CE,$3F,$20,$7E
	.dc.b	$DC,$FE,$7D,$C3,$FB,$37,$FA,$07,$8B,$FB,$37,$FA,$17,$8B,$FB,$37
	.dc.b	$FA,$27,$8B,$FB,$3C,$3F,$C5,$5E,$2F,$EC,$DF,$F1,$67,$8B,$FB,$37
	.dc.b	$FC,$23,$5F,$C6,$4F,$EE,$7E,$47,$F1,$9C,$7F,$8D,$C6,$4F,$DC,$9F
	.dc.b	$C8,$4F,$F0,$15,$75,$CB,$FC,$6B,$87,$85,$FF,$1B,$E3,$E5,$C7,$FC
	.dc.b	$79,$97,$A5,$FF,$1F,$63,$ED,$7F,$C8,$15,$FC,$53,$1F,$E7,$87,$EC
	.dc.b	$B9,$9D,$BF,$CA,$AB,$F5,$7A,$79,$AB,$FB,$0B,$63,$31,$EC,$21,$A2
	.dc.b	$BA,$70,$47,$04,$BF,$E7,$53,$D5,$B3,$FC,$67,$93,$F1,$B8,$79,$AB
	.dc.b	$52,$8C,$3E,$0A,$E2,$F7,$6D,$F7,$8E,$18,$7B,$E7,$24,$8C,$DF,$6B
	.dc.b	$EB,$1C,$CC,$9F,$03,$CB,$FF,$1E,$3D,$3E,$2B,$FC,$66,$47,$FC,$67
	.dc.b	$F1,$5F,$E3,$6F,$33,$FC,$6F,$EE,$FF,$8E,$3E,$27,$FC,$73,$F0,$3F
	.dc.b	$E3,$AF,$89,$FF,$1D,$FC,$1F,$F8,$F3,$E2,$7F,$C7,$BF,$0B,$FE,$3E
	.dc.b	$F8,$9F,$F1,$FF,$C3,$FF,$9C,$7F,$80,$7F,$CF,$DB,$7F,$96,$17,$C4
	.dc.b	$E7,$E2,$BF,$6D,$EA,$FF,$9D,$FA,$7E,$5F,$9B,$FC,$73,$A9,$D3,$FE
	.dc.b	$87,$E8,$64,$C7,$B6,$5D,$F3,$F0,$9F,$E4,$ED,$7C,$AF,$F9,$43,$7F
	.dc.b	$4B,$FE,$52,$E1,$F6,$BF,$E5,$4E,$3F,$82,$FF,$95,$B9,$7F,$CB,$3C
	.dc.b	$DF,$1E,$3F,$C5,$FC,$FF,$E3,$C7,$E6,$1F,$E5,$C7,$5B,$FF,$40,$9F
	.dc.b	$99,$3F,$E6,$0E,$7F,$98,$FF,$98,$B9,$FE,$63,$FE,$64,$E7,$F9,$8F
	.dc.b	$F9,$9B,$9F,$E6,$3F,$E6,$8E,$7F,$73,$F2,$F8,$3E,$53,$FE,$8F,$3F
	.dc.b	$29,$CF,$FD,$1F,$8F,$E3,$46,$CF,$A4,$7C,$63,$FC,$93,$C5,$F1,$9F
	.dc.b	$F2,$5F,$17,$C6,$7F,$C9,$BC,$5F,$19,$FF,$27,$F1,$7C,$67,$FC,$A3
	.dc.b	$C5,$F1,$9F,$84,$43,$FA,$27,$57,$E6,$DF,$CA,$9F,$97,$E5,$FC,$CF
	.dc.b	$1F,$E8,$5B,$CD,$F9,$F6,$CF,$09,$D5,$73,$9C,$91,$C1,$3A,$BF,$4C
	.dc.b	$EF,$D8,$FF,$1A,$DF,$ED,$5B,$5C,$6B,$64,$9F,$DC,$1E,$1C,$3C,$97
	.dc.b	$D6,$33,$5C,$5B,$19,$7E,$0D,$7E,$89,$FD,$1C,$FE,$E1,$B3,$FE,$46
	.dc.b	$6C,$ED,$FE,$0E,$60,$C4,$F1,$72,$43,$E6,$9B,$3C,$0E,$DC,$9B,$70
	.dc.b	$49,$F1,$46,$6F,$E2,$9F,$1A,$33,$1C,$89,$3C,$6B,$D1,$18,$9D,$CE
	.dc.b	$63,$D3,$18,$24,$F0,$9D,$4F,$A2,$72,$F0,$9F,$E1,$7D,$8F,$03,$D5
	.dc.b	$5A,$AE,$15,$CE,$31,$1F,$0C,$87,$08,$D1,$1E,$C4,$39,$94,$E2,$7C
	.dc.b	$0E,$CF,$E8,$27,$F4,$29,$EF,$1D,$4E,$05,$39,$90,$E7,$87,$C0,$6C
	.dc.b	$72,$7E,$55,$7C,$0B,$E5,$1A,$8F,$43,$CC,$C9,$0E,$27,$88,$A6,$CA
	.dc.b	$7E,$52,$FD,$7C,$5F,$DB,$63,$B1,$C7,$0F,$33,$D5,$AF,$23,$18,$7A
	.dc.b	$63,$FA,$AE,$3F,$5E,$74,$48,$E2,$49,$E0,$4C,$3D,$B3,$E2,$3D,$13
	.dc.b	$F9,$8C,$BF,$2B,$7E,$79,$F0,$D7,$06,$9E,$9B,$E8,$D8,$D3,$F1,$27
	.dc.b	$CD,$7F,$BB,$8F,$E2,$D5,$FB,$BD,$3F,$88,$DF,$F0,$59,$CB,$63,$31
	.dc.b	$FC,$0A,$3F,$BB,$61,$E6,$8C,$10,$E2,$59,$E2,$E0,$8E,$0F,$F0,$C7
	.dc.b	$6F,$F7,$CC,$3C,$D1,$9D,$7F,$8A,$E1,$E3,$6D,$F5,$6B,$D1,$B3,$FD
	.dc.b	$BA,$3F,$CC,$CD,$AF,$F4,$7C,$6C,$FE,$44,$47,$C3,$3C,$89,$19,$C7
	.dc.b	$E4,$46,$27,$E6,$61,$E9,$9C,$D6,$F0,$F4,$34,$F4,$14,$F1,$57,$81
	.dc.b	$C6,$1E,$7D,$3F,$33,$1C,$27,$44,$BE,$ED,$BE,$ED,$9F,$F1,$46,$DF
	.dc.b	$E0,$FE,$37,$C5,$7F,$86,$C9,$3F,$E1,$BF,$8A,$FF,$0F,$7C,$4F,$F8
	.dc.b	$7F,$E2,$7F,$C4,$1E,$F7,$F8,$87,$E2,$7F,$C4,$5F,$01,$FE,$23,$F8
	.dc.b	$9F,$F1,$27,$C1,$7F,$89,$7E,$27,$FC,$4D,$F0,$9F,$E2,$7F,$89,$FF
	.dc.b	$14,$7C,$37,$F8,$A7,$E2,$7F,$C5,$5F,$11,$FE,$2B,$F8,$9F,$F3,$B7
	.dc.b	$0F,$F8,$AD,$CB,$FE,$2D,$9E,$2F,$F8,$BE,$FA,$BF,$E3,$1C,$7C,$2F
	.dc.b	$F8,$CF,$3F,$2B,$FE,$35,$D7,$D2,$FF,$8D,$F7,$F6,$BE,$F3,$FC,$75
	.dc.b	$C5,$F0,$9F,$F6,$07,$1F,$F9,$88,$9D,$DF,$E4,$1F,$89,$FF,$21,$7C
	.dc.b	$4F,$F9,$0B,$B3,$F1,$9F,$BF,$FE,$3F,$04,$63,$E0,$F5,$E0,$7F,$8F
	.dc.b	$EB,$FB,$A8,$F8,$4C,$DF,$F9,$B6,$1E,$9E,$3B,$90,$F8,$5C,$2F,$56
	.dc.b	$FC,$36,$3D,$7C,$FC,$3E,$BC,$AD,$FC,$47,$CF,$D0,$79,$B6,$FF,$29
	.dc.b	$79,$5F,$BB,$8C,$1F,$E7,$FE,$57,$6F,$28,$E3,$C7,$F0,$31,$EF,$18
	.dc.b	$C3,$DD,$9F,$92,$30,$7B,$75,$F4,$EF,$EA,$E1,$F5,$F1,$7B,$3C,$BF
	.dc.b	$6F,$27,$BB,$2F,$9F,$8F,$F9,$BF,$9B,$FC,$E5,$CF,$23,$FD,$59,$C1
	.dc.b	$F5,$9F,$F3,$A4,$7D,$77,$FC,$ED,$CF,$CD,$F1,$F3,$FC,$CC,$7C,$11
	.dc.b	$8E,$7F,$A8,$DD,$7C,$EE,$77,$F6,$3F,$1A,$FE,$AF,$07,$C2,$8C,$5F
	.dc.b	$F4,$56,$F2,$FF,$A7,$7C,$AF,$DD,$C7,$F9,$1E,$53,$F5,$E3,$C5,$CF
	.dc.b	$F6,$DF,$27,$47,$E0,$32,$F1,$C7,$6C,$FE,$4B,$E9,$E2,$F9,$2F,$AF
	.dc.b	$8B,$E4,$BE,$7E,$1F,$F4,$94,$7F,$03,$E1,$F9,$53,$FA,$06,$F3,$7F
	.dc.b	$27,$AF,$D1,$71,$F2,$8F,$CF,$DF,$6B,$F7,$E1,$E2,$7C,$87,$CB,$E5
	.dc.b	$F4,$3E,$A3,$E7,$70,$FC,$7B,$FC,$54,$7D,$1D,$FE,$36,$7F,$EA,$93
	.dc.b	$FD,$89,$C1,$F1,$A7,$EB,$73,$71,$CF,$F1,$F1,$D4,$FA,$3B,$FC,$6C
	.dc.b	$7F,$6B,$BF,$95,$FD,$BE,$BE,$97,$D7,$5F,$86,$3F,$1C,$7E,$28,$83
	.dc.b	$CB,$FE,$AD,$93,$E2,$BF,$66,$49,$FC,$57,$C5,$7E,$DB,$DC,$FD,$BF
	.dc.b	$C3,$7E,$E3,$DE,$FD,$CF,$C3,$FE,$EB,$E0,$3F,$77,$F1,$1F,$BC,$F8
	.dc.b	$2F,$E0,$1E,$73,$E2,$6F,$C5,$63,$E1,$73,$F1,$5A,$F8,$7D,$FC,$57
	.dc.b	$EF,$78,$7F,$29,$E4,$FE,$EF,$9B,$ED,$79,$3A,$9F,$D0,$E8,$C3,$9C
	.dc.b	$63,$2E,$EE,$B4,$F2,$3F,$A5,$DB,$D0,$FE,$9F,$83,$D8,$FE,$A7,$8B
	.dc.b	$E0,$3F,$08,$FD,$67,$2F,$C5,$7F,$1C,$7C,$CD,$3F,$1D,$DB,$FB,$78
	.dc.b	$FC,$B7,$91,$FB,$77,$F5,$7E,$47,$ED,$9F,$D6,$F9,$3F,$C2,$5F,$C7
	.dc.b	$1F,$B5,$3F,$05,$7F,$CD,$3C,$2F,$C3,$47,$F2,$DF,$07,$F8,$D3,$FA
	.dc.b	$3E,$2F,$EA,$2D,$3F,$24,$7E,$38,$FD,$99,$F9,$83,$AF,$83,$FC,$EF
	.dc.b	$C1,$FC,$73,$FC,$F7,$C1,$FC,$73,$FC,$FF,$C1,$FC,$73,$FD,$04,$FD
	.dc.b	$51,$FD,$CF,$C4,$FD,$7B,$FB,$3B,$FE,$37,$BE,$1F,$E6,$4F,$C5,$D3
	.dc.b	$F6,$9B,$FC,$57,$FA,$74,$FF,$01,$3F,$8D,$6D,$F1,$5F,$E2,$7E,$27
	.dc.b	$EC,$9F,$E2,$BE,$37,$F0,$D7,$F8,$BF,$8F,$FC,$45,$FE,$2E,$7F,$B0
	.dc.b	$BC,$8F,$E2,$F8,$7F,$B6,$FC,$5F,$D9,$4F,$F1,$EE,$2F,$9B,$E4,$FF
	.dc.b	$22,$8F,$8A,$7F,$3F,$3E,$A9,$FC,$10,$FB,$27,$9C,$FB,$A7,$F1,$73
	.dc.b	$F6,$47,$F1,$7D,$FE,$14,$3C,$96,$FF,$D6,$AB,$F1,$67,$F1,$D7,$71
	.dc.b	$FC,$7A,$3F,$AD,$6F,$F1,$F8,$3F,$B9,$73,$7E,$2F,$4F,$EF,$3C,$3F
	.dc.b	$1D,$FE,$F7,$C3,$F1,$DF,$EF,$BC,$3F,$1D,$FE,$FF,$C3,$F1,$DF,$F0
	.dc.b	$0F,$93,$FC,$63,$3F,$E9,$BC,$5F,$EA,$9F,$67,$F1,$F5,$FC,$7A,$7E
	.dc.b	$79,$F8,$B3,$E9,$9F,$99,$3E,$B9,$FD,$44,$FB,$67,$C7,$7E,$FB,$FD
	.dc.b	$3B,$4F,$C5,$BF,$DD,$F8,$7E,$0F,$07,$C5,$F2,$BF,$8B,$F5,$7E,$C7
	.dc.b	$0F,$F0,$BB,$FE,$0A,$F2,$BF,$6F,$87,$F5,$8E,$CF,$DD,$BF,$D6,$BB
	.dc.b	$7F,$7E,$FF,$5C,$EE,$FE,$06,$FF,$61,$F0,$BF,$8B,$3F,$C3,$3C,$6F
	.dc.b	$E2,$CF,$F0,$EF,$1B,$F8,$B3,$FC,$43,$C6,$FE,$2D,$A7,$F9,$CF,$C6
	.dc.b	$FE,$2D,$C3,$FE,$55,$F0,$BF,$8B,$5F,$5F,$1B,$F8,$AB,$DF,$C6,$FE
	.dc.b	$2A,$F8,$7C,$6F,$1E,$3F,$E3,$2D,$BC,$EF,$F8,$D3,$7F,$5B,$FE,$36
	.dc.b	$E0,$F7,$BF,$E3,$8E,$1F,$89,$A7,$F8,$D7,$8B,$E4,$7B,$3F,$E2,$6F
	.dc.b	$05,$C8,$F8,$9F,$E6,$A4,$DB,$F9,$5F,$C5,$7F,$38,$EA,$FE,$73,$F1
	.dc.b	$1F,$CE,$BE,$23,$F9,$DE,$5F,$CF,$3E,$23,$F9,$EF,$C0,$7F,$3E,$F8
	.dc.b	$8F,$E7,$FF,$05,$FD,$03,$E2,$3F,$A0,$FC,$27,$F4,$2F,$88,$FE,$87
	.dc.b	$F0,$DF,$D1,$3E,$23,$FC,$89,$FC,$03,$FA,$1E,$1F,$E7,$FE,$97,$F9
	.dc.b	$1F,$C1,$3A,$8E,$07,$F4,$0B,$E8,$FF,$42,$C7,$C0,$FF,$44,$CF,$C8
	.dc.b	$FF,$46,$D7,$D0,$FF,$48,$DF,$D8,$FF,$4A,$E1,$F8,$0F,$C2,$3F,$A6
	.dc.b	$F2,$7C,$57,$FB,$44,$FC,$BD,$1F,$13,$EA,$24,$FF,$8D,$21,$F1,$5E
	.dc.b	$EF,$73,$DF,$F1,$3F,$03,$E2,$7E,$0F,$BF,$F0,$BE,$27,$E1,$FC,$0F
	.dc.b	$C4,$ED,$F8,$BF,$05,$F1,$BE,$27,$E3,$FC,$27,$C8,$F8,$9F,$93,$F0
	.dc.b	$DF,$2B,$E2,$7F,$27,$E3,$7D,$FF,$5B,$F8,$FB,$F9,$3E,$3F,$24,$F0
	.dc.b	$7C,$D7,$D1,$F4,$63,$E0,$7D,$59,$F9,$1F,$66,$BE,$87,$DD,$BF,$B1
	.dc.b	$F8,$1C,$3F,$73,$8B,$E1,$C7,$F9,$37,$8F,$E3,$BF,$E7,$8F,$53,$FD
	.dc.b	$CD,$18,$8D,$72,$F0,$F2,$BF,$DB,$D3,$F2,$A7,$A9,$F3,$79,$BE,$53
	.dc.b	$F3,$F9,$BE,$53,$F4,$79,$BE,$53,$F4,$F9,$BE,$53,$F5,$79,$BE,$53
	.dc.b	$F5,$F9,$BE,$53,$FE,$B9,$3F,$90,$FA,$BF,$E3,$97,$07,$AF,$97,$E5
	.dc.b	$3E,$DE,$5F,$94,$FB,$F9,$7E,$53,$F0,$79,$7E,$53,$F0,$F9,$7E,$53
	.dc.b	$F1,$79,$7E,$53,$F1,$F9,$7E,$53,$F8,$BE,$3F,$97,$C8,$E8,$FD,$E6
	.dc.b	$3F,$BB,$AF,$CB,$90,$FE,$39,$E4,$7E,$71,$FE,$3D,$D7,$FA,$07,$F9
	.dc.b	$05,$7E,$91,$FE,$45,$D9,$FA,$87,$F9,$27,$6F,$EB,$1F,$E4,$DD,$DF
	.dc.b	$B0,$7F,$94,$77,$FE,$D1,$FB,$C7,$F8,$4B,$AF,$EC,$B9,$5D,$BE,$89
	.dc.b	$E5,$18,$97,$93,$FC,$5A,$FC,$13,$98,$FE,$33,$87,$96,$31,$AB,$2F
	.dc.b	$54,$63,$3F,$73,$AD,$3E,$13,$C3,$5F,$DE,$3C,$BC,$0F,$E4,$FA,$78
	.dc.b	$B1,$F9,$8F,$F4,$58,$DE,$DF,$42,$F9,$BF,$4B,$0E,$EF,$D4,$C7,$C6
	.dc.b	$FD,$6C,$BD,$0F,$D8,$CF,$DA,$FD,$AD,$7E,$1B,$F1,$72,$FF,$2E,$F8
	.dc.b	$7F,$C8,$5F,$F5,$FF,$87,$FD,$39,$F8,$FE,$07,$F2,$9D,$FE,$1D,$7F
	.dc.b	$2E,$9F,$F3,$2F,$89,$FC,$C9,$FF,$33,$F8,$BF,$CD,$9F,$F3,$4F,$8D
	.dc.b	$FC,$E9,$FF,$35,$F8,$FF,$CF,$9F,$F3,$6F,$91,$FD,$09,$FF,$37,$F9
	.dc.b	$3F,$E7,$CE,$3F,$87,$1F,$1F,$93,$E4,$F3,$36,$FD,$E6,$DF,$BA,$E9
	.dc.b	$FD,$E7,$0F,$F2,$9F,$1F,$F9,$26,$3F,$BB,$3F,$85,$9F,$BB,$3F,$87
	.dc.b	$9F,$BB,$3F,$89,$9F,$BB,$3F,$8B,$9F,$BB,$3F,$8D,$9F,$BB,$3F,$8F
	.dc.b	$9F,$B8,$AF,$C5,$79,$1F,$C9,$23,$F1,$FE,$F3,$D3,$FE,$03,$1F,$C5
	.dc.b	$FC,$5F,$E0,$AF,$F1,$9F,$1B,$F8,$4B,$FC,$6F,$C7,$FE,$1A,$FF,$1D
	.dc.b	$F2,$3F,$88,$BF,$C7,$FC,$9F,$E2,$AF,$F2,$1F,$2B,$DE,$FF,$23,$D7
	.dc.b	$E1,$BF,$6B,$7F,$93,$C1,$F2,$38,$3F,$A7,$46,$B8,$7E,$65,$F2,$E1
	.dc.b	$FE,$4B,$8F,$67,$F9,$3E,$5E,$27,$F9,$4E,$7E,$67,$F9,$5E,$9E,$A7
	.dc.b	$F9,$6E,$BE,$E7,$F9,$7E,$DF,$09,$FF,$57,$ED,$FC,$88,$87,$FA,$34
	.dc.b	$FE,$44,$7C,$F3,$F9,$11,$F4,$CF,$E4,$47,$D7,$3F,$91,$1F,$6C,$FE
	.dc.b	$44,$7D,$F3,$F9,$11,$F8,$33,$F9,$11,$FD,$AF,$2F,$85,$B7,$F4,$7F
	.dc.b	$0F,$F9,$14,$FE,$44,$FE,$41,$5F,$BB,$7F,$BD,$1F,$E3,$DE,$5F,$CD
	.dc.b	$3F,$E3,$FE,$6F,$CF,$3F,$E4,$1E,$7F,$D1,$3F,$E4,$3E,$8F,$D3,$3F
	.dc.b	$E4,$5E,$9F,$D5,$3F,$E4,$7E,$AF,$D7,$3F,$E4,$9E,$BF,$DB,$3F,$E1
	.dc.b	$A3,$F6,$FE,$5E,$33,$FD,$AC,$7F,$86,$4F,$DA,$9F,$E1,$93,$F6,$A7
	.dc.b	$F8,$64,$FD,$A9,$FE,$19,$3F,$6A,$7F,$86,$4F,$DA,$9F,$E1,$93,$F6
	.dc.b	$A7,$C0,$7E,$13,$FC,$64,$FE,$E1,$E0,$FF,$86,$7E,$23,$FC,$62,$48
	.dc.b	$FF,$0C,$35,$FE,$F0,$D9,$FE,$E6,$DE,$9E,$53,$C0,$87,$02,$CF,$06
	.dc.b	$C7,$07,$FA,$39,$F1,$88,$60,$E4,$48,$E4,$D8,$E5,$18,$23,$82,$47
	.dc.b	$F2,$69,$F4,$34,$F1,$94,$E8,$5D,$FF,$C2,$BF,$13,$FE,$56,$24,$7F
	.dc.b	$8F,$7A,$FF,$CB,$3E,$6F,$F9,$6B,$D9,$FF,$2D,$FC,$07,$FA,$EF,$3F
	.dc.b	$EB,$1F,$C0,$3F,$D7,$70,$46,$3E,$07,$5A,$FF,$9D,$6B,$FC,$BF,$43
	.dc.b	$71,$9F,$81,$DC,$7F,$66,$97,$C2,$63,$C1,$43,$E0,$B0,$BE,$16,$FC
	.dc.b	$1E,$3E,$1B,$3F,$09,$F9,$5D,$3F,$C9,$1C,$3F,$E1,$FD,$3E,$96,$FE
	.dc.b	$AA,$E1,$1E,$58,$F6,$E1,$DB,$1E,$F9,$F8,$35,$F0,$ED,$E2,$E7,$F1
	.dc.b	$F0,$79,$35,$F8,$F9,$FF,$99,$F8,$BE,$5B,$F3,$4F,$F3,$57,$27,$D0
	.dc.b	$7F,$CE,$27,$F6,$07,$1A,$FF,$91,$9F,$F4,$5F,$27,$29,$FF,$28,$57
	.dc.b	$67,$E5,$9F,$26,$FC,$6F,$AD,$FA,$58,$FA,$23,$07,$CF,$CF,$DA,$E3
	.dc.b	$8F,$E7,$ED,$F3,$23,$1C,$6E,$3F,$9B,$1F,$E8,$33,$E6,$6B,$FD,$4B
	.dc.b	$8F,$E8,$47,$DB,$E3,$F9,$AF,$9B,$1F,$B7,$C7,$F3,$63,$AF,$1F,$CD
	.dc.b	$7B,$F1,$FC,$D7,$C3,$C7,$F3,$5F,$F4,$97,$27,$E5,$5F,$F4,$A6,$5F
	.dc.b	$19,$F9,$A4,$3E,$6E,$3F,$DE,$79,$BF,$4E,$D8,$FD,$3B,$5F,$D3,$B7
	.dc.b	$4F,$D8,$C7,$EA,$DB,$19,$E2,$FD,$9B,$67,$51,$8E,$F5,$F8,$38,$1F
	.dc.b	$0B,$FE,$4B,$AF,$D3,$3E,$4A,$DD,$F9,$1A,$F9,$1B,$8F,$89,$B5,$DF
	.dc.b	$0E,$44,$AE,$4F,$F7,$23,$FD,$A8,$FF,$17,$3F,$68,$7F,$07,$2E,$3F
	.dc.b	$0D,$B5,$E2,$6C,$F2,$78,$1C,$C8,$60,$93,$CD,$CB,$5C,$CF,$57,$38
	.dc.b	$FC,$67,$F9,$4E,$1F,$E6,$7F,$37,$F8,$5C,$7E,$2E,$74,$FF,$2C,$3C
	.dc.b	$A7,$BA,$FB,$12,$FB,$47,$83,$0F,$54,$E5,$B3,$A8,$C4,$F3,$8C,$12
	.dc.b	$76,$7C,$AE,$5F,$45,$70,$AF,$9A,$77,$39,$4F,$06,$90,$DC,$71,$7D
	.dc.b	$66,$2B,$C0,$EC,$C9,$0F,$9A,$FF,$9A,$72,$FE,$43,$1F,$E1,$EC,$BF
	.dc.b	$D3,$2F,$F5,$C3,$81,$E5,$3B,$5F,$95,$A7,$73,$AB,$97,$8D,$F9,$DC
	.dc.b	$9F,$CF,$F0,$F8,$78,$7A,$23,$99,$0E,$2E,$5B,$5D,$5E,$66,$CA,$68
	.dc.b	$FA,$87,$F1,$2C,$3D,$EF,$F3,$E7,$6C,$3D,$C7,$13,$A1,$83,$57,$EB
	.dc.b	$7F,$B5,$5F,$AA,$BC,$91,$B2,$4E,$C9,$7E,$B8,$E6,$D7,$F4,$78,$FE
	.dc.b	$63,$1F,$CE,$61,$FD,$7F,$2F,$CE,$C7,$F6,$EC,$FF,$C3,$17,$E7,$21
	.dc.b	$D3,$2F,$1E,$5E,$06,$E5,$DE,$3F,$8E,$63,$E9,$AF,$E0,$64,$3F,$86
	.dc.b	$57,$F0,$76,$99,$C3,$F8,$74,$65,$F2,$47,$ED,$33,$FE,$F9,$6C,$BD
	.dc.b	$31,$FC,$72,$79,$3F,$C5,$5A,$7F,$26,$24,$FF,$29,$26,$3F,$C7,$23
	.dc.b	$2D,$8C,$E1,$FE,$3B,$E9,$FF,$33,$C6,$2B,$AB,$5F,$59,$31,$F5,$BA
	.dc.b	$BE,$84,$BF,$EB,$34,$F2,$FF,$D5,$8F,$FB,$06,$BE,$16,$3E,$B7,$A3
	.dc.b	$70,$EB,$18,$3D,$37,$F3,$71,$F3,$E3,$DE,$7C,$53,$D1,$D9,$82,$1C
	.dc.b	$F0,$F1,$3F,$BB,$C3,$F3,$F8,$7E,$F2,$35,$7D,$5B,$7D,$6F,$E2,$57
	.dc.b	$98,$97,$FE,$59,$C3,$E5,$EA,$F8,$5F,$F1,$79,$31,$FF,$01,$FB,$7F
	.dc.b	$E3,$2F,$73,$FC,$67,$F1,$5F,$E3,$4F,$7B,$FC,$6B,$F1,$5F,$E3,$6F
	.dc.b	$80,$FF,$1B,$FC,$57,$F8,$E3,$E0,$BF,$C7,$3F,$15,$FE,$3D,$FE,$01
	.dc.b	$7C,$2D,$F8,$AC,$7C,$3E,$7E,$2B,$5F,$13,$FE,$40,$ED,$FE,$D1,$C5
	.dc.b	$FE,$47,$72,$FF,$86,$71,$57,$28,$C6,$1D,$F1,$F0,$C7,$F8,$73,$3F
	.dc.b	$2B,$FE,$1E,$D7,$D2,$FF,$88,$37,$F6,$BF,$E2,$2E,$1F,$F2,$8F,$17
	.dc.b	$C3,$8F,$EA,$5C,$7F,$DA,$B9,$7F,$27,$AB,$E2,$BD,$C4,$AF,$37,$BB
	.dc.b	$F0,$7D,$EF,$85,$EA,$FC,$3F,$82,$F8,$9E,$B7,$C5,$F4,$7E,$37,$AD
	.dc.b	$F1,$FE,$0B,$E4,$7A,$FF,$27,$E0,$BE,$57,$B1,$F2,$FE,$0B,$E6,$7B
	.dc.b	$3F,$37,$E1,$3E,$47,$07,$ED,$C7,$8F,$FB,$7F,$33,$31,$FE,$35,$E4
	.dc.b	$57,$28,$F9,$98,$76,$7E,$76,$5E,$27,$E8,$69,$E6,$7E,$96,$DE,$A7
	.dc.b	$EA,$70,$7B,$9F,$AD,$C5,$F8,$1E,$3F,$89,$1E,$1E,$4F,$95,$7F,$C3
	.dc.b	$34,$F9,$53,$F7,$67,$7E,$43,$C8,$D6,$1F,$2E,$BF,$05,$CD,$F2,$DF
	.dc.b	$C2,$73,$7C,$B7,$F0,$DC,$DF,$2D,$FC,$47,$37,$CB,$7F,$15,$CD,$F2
	.dc.b	$DF,$C6,$73,$7C,$B7,$FC,$C7,$E0,$FF,$4F,$7E,$5F,$96,$F3,$7E,$5E
	.dc.b	$DE,$59,$F9,$FC,$1F,$31,$E6,$7C,$C3,$B9,$F3,$0F,$19,$F3,$0F,$39
	.dc.b	$F3,$0F,$59,$F3,$0F,$79,$F3,$0F,$C6,$9F,$1C,$87,$F9,$89,$F9,$6F
	.dc.b	$CF,$33,$1E,$7E,$94,$B7,$F9,$95,$F8,$5D,$FE,$63,$F8,$7D,$FE,$63
	.dc.b	$F8,$9D,$FE,$63,$F8,$BD,$FE,$63,$F8,$DD,$FE,$63,$F8,$FD,$FE,$63
	.dc.b	$FE,$3B,$DF,$E6,$69,$FC,$C7,$6F,$97,$C1,$F8,$3E,$8F,$CA,$5F,$E5
	.dc.b	$FA,$3F,$28,$FE,$67,$A3,$F2,$8F,$E6,$FA,$3F,$28,$FE,$77,$9B,$F2
	.dc.b	$8F,$E7,$FA,$3F,$28,$FE,$87,$A3,$F2,$8F,$C4,$CB,$F2,$5D,$1F,$94
	.dc.b	$8F,$B7,$ED,$7F,$19,$23,$FE,$A9,$BF,$F1,$83,$FA,$E5,$F0,$5F,$ED
	.dc.b	$B5,$EE,$5F,$85,$C7,$C1,$E7,$E1,$B5,$F0,$9B,$F8,$7E,$1F,$0B,$C7
	.dc.b	$E1,$F9,$7C,$37,$3F,$88,$E9,$F0,$FD,$7E,$27,$B7,$C4,$7F,$26,$EC
	.dc.b	$70,$23,$5C,$5E,$7B,$E4,$FA,$71,$EC,$FA,$F3,$F1,$3E,$DD,$7C,$CF
	.dc.b	$BF,$7F,$53,$F0,$78,$7D,$DC,$5E,$F8,$F8,$A7,$F0,$DE,$3F,$95,$AF
	.dc.b	$E3,$B9,$3F,$77,$CB,$F3,$AB,$F3,$27,$D3,$3F,$3A,$7D,$73,$F4,$27
	.dc.b	$DB,$3F,$4A,$7D,$F3,$F5,$27,$E0,$CF,$D6,$9F,$87,$3F,$62,$7F,$1F
	.dc.b	$EB,$FC,$6B,$FC,$93,$DC,$FF,$0C,$B7,$C9,$CE,$1F,$A4,$AF,$C8,$F9
	.dc.b	$1F,$A7,$F2,$1D,$FF,$92,$8C,$77,$FE,$49,$D7,$7F,$E4,$9E,$1D,$FF
	.dc.b	$92,$79,$74,$7E,$49,$E9,$DF,$F9,$2D,$7F,$6C,$4F,$2B,$FA,$6F,$0F
	.dc.b	$C8,$8F,$E1,$AE,$EB,$8C,$7C,$87,$AB,$F2,$1F,$0B,$F2,$1F,$2B,$F2
	.dc.b	$1F,$4B,$F2,$1F,$76,$7F,$03,$87,$F6,$53,$FC,$88,$FF,$09,$1F,$C8
	.dc.b	$BC,$1F,$8F,$B7,$F1,$F8,$FD,$F1,$FC,$7C,$FE,$04,$7F,$1F,$3F,$83
	.dc.b	$1F,$C7,$CF,$E1,$47,$F1,$F3,$F8,$71,$FC,$7C,$FE,$24,$7F,$1F,$3F
	.dc.b	$AC,$79,$3F,$0F,$87,$F1,$90,$3F,$90,$65,$9A,$FE,$43,$97,$D9,$BF
	.dc.b	$E3,$AF,$DB,$F2,$7F,$8E,$BF,$77,$C9,$FE,$3A,$FD,$FF,$27,$F8,$EB
	.dc.b	$F8,$1F,$27,$F8,$EB,$F8,$3F,$27,$F8,$EB,$F8,$5F,$27,$F8,$EB,$FE
	.dc.b	$68,$D0,$FF,$2F,$F2,$BF,$C9,$BE,$D7,$FA,$0F,$87,$E6,$D7,$C8,$3F
	.dc.b	$30,$7C,$83,$F3,$87,$C8,$3F,$40,$7C,$83,$F4,$87,$C8,$3F,$50,$7C
	.dc.b	$83,$F5,$87,$C8,$3F,$8F,$1F,$8A,$3E,$47,$1F,$F8,$6B,$B3,$F7,$38
	.dc.b	$7E,$7D,$FD,$E4,$FF,$1F,$ED,$FE,$39,$DD,$FC,$0A,$3D,$B8,$7F,$A1
	.dc.b	$3A,$1D,$FF,$C2,$27,$27,$F2,$5F,$03,$F8,$6B,$8E,$7F,$E2,$FE,$0F
	.dc.b	$F1,$28,$C5,$BC,$2F,$E2,$F1,$8F,$0F,$F8,$DB,$AF,$13,$F9,$05,$7F
	.dc.b	$1F,$8F,$E6,$9E,$2F,$F2,$67,$F9,$B7,$8D,$FC,$A9,$FE,$71,$E3,$FF
	.dc.b	$2E,$7F,$9D,$79,$1F,$CC,$9F,$E7,$9E,$4F,$F3,$67,$F9,$F7,$95,$F0
	.dc.b	$1F,$84,$7F,$41,$3E,$3F,$0F,$F8,$67,$93,$F8,$FE,$6F,$8A,$FE,$84
	.dc.b	$4A,$FE,$91,$EE,$7F,$45,$F8,$9F,$E8,$DF,$13,$FD,$1F,$C6,$FE,$91
	.dc.b	$F1,$3F,$D2,$7E,$07,$F3,$98,$7C,$7F,$E0,$1F,$F0,$73,$7E,$0F,$3D
	.dc.b	$DF,$DE,$E7,$F8,$DF,$51,$7C,$26,$B8,$7F,$D7,$30,$F8,$AC,$63,$FE
	.dc.b	$46,$87,$C3,$63,$0E,$5F,$15,$65,$F1,$1F,$D4,$2F,$FC,$9D,$5F,$E4
	.dc.b	$D7,$FB,$27,$66,$E7,$85,$F1,$C3,$96,$3C,$F3,$E9,$EB,$75,$DB,$B6
	.dc.b	$BF,$07,$3F,$ED,$DC,$1F,$15,$FE,$E1,$C5,$F2,$5F,$96,$7F,$73,$E5
	.dc.b	$F9,$CB,$C5,$FF,$63,$F0,$BE,$2B,$D4,$49,$FE,$11,$C7,$F2,$7E,$2B
	.dc.b	$DD,$ED,$7B,$FD,$BF,$81,$EE,$7C,$1F,$77,$E1,$7C,$4F,$C3,$F7,$FE
	.dc.b	$27,$C4,$FC,$5F,$81,$F8,$DF,$13,$F1,$FE,$0F,$E4,$7C,$4F,$CB,$FE
	.dc.b	$01,$FF,$50,$F0,$78,$39,$FF,$DD,$70,$F8,$7F,$99,$3F,$A5,$F5,$BF
	.dc.b	$DA,$7E,$3F,$E7,$63,$F5,$15,$B7,$91,$FE,$7E,$7A,$BF,$87,$C3,$C4
	.dc.b	$EB,$C3,$FC,$26,$7E,$78,$C7,$3B,$6F,$64,$63,$83,$E0,$3A,$E2,$F8
	.dc.b	$8F,$0E,$4F,$C2,$72,$FC,$B8,$FC,$B7,$3F,$F8,$03,$1F,$A1,$1D,$A3
	.dc.b	$5C,$FF,$3E,$39,$1E,$2E,$7F,$9E,$F9,$39,$FE,$7B,$E6,$E7,$F9,$EF
	.dc.b	$A3,$9F,$E7,$BE,$AE,$7F,$9E,$FB,$F8,$FE,$1E,$3F,$E5,$F7,$E4,$47
	.dc.b	$F0,$93,$FA,$DF,$17,$F9,$45,$FC,$C9,$C7,$9B,$EC,$7C,$3F,$48,$FD
	.dc.b	$96,$1E,$07,$F6,$98,$F9,$1F,$DB,$65,$E8,$7F,$71,$9F,$B1,$FD,$D6
	.dc.b	$9F,$8E,$7E,$8E,$FF,$13,$A3,$F6,$BB,$FF,$0A,$EA,$FC,$DF,$07,$F9
	.dc.b	$DF,$A3,$F8,$66,$1F,$A0,$3F,$93,$F8,$FF,$C2,$9F,$F2,$A7,$8F,$FC
	.dc.b	$29,$FF,$2B,$78,$FF,$AC,$7E,$FF,$81,$FB,$07,$F0,$3E,$17,$ED,$1F
	.dc.b	$C1,$F8,$9F,$B8,$7F,$0B,$E3,$7F,$93,$7E,$1F,$F8,$69,$FC,$3B,$AB
	.dc.b	$F8,$1C,$FE,$2B,$C7,$FD,$F3,$F8,$CF,$1F,$F7,$CF,$E3,$BC,$7F,$DF
	.dc.b	$3F,$90,$E3,$FD,$F3,$FD,$03,$C7,$FD,$F3,$FD,$0B,$C7,$FD,$F3,$FD
	.dc.b	$13,$C9,$F8,$75,$FD,$3F,$97,$E3,$4F,$F0,$12,$1B,$3F,$CE,$DC,$7F
	.dc.b	$15,$FE,$9B,$C7,$F1,$5F,$E9,$FC,$7F,$15,$FE,$58,$79,$CF,$8A,$7A
	.dc.b	$CF,$8A,$7B,$CF,$8A,$7E,$20,$FE,$B9,$BF,$C6,$D3,$F7,$7C,$BF,$CD
	.dc.b	$78,$7E,$65,$7F,$3B,$E2,$FA,$0F,$F3,$DE,$3F,$A8,$F8,$8F,$E8,$1A
	.dc.b	$FC,$6B,$FC,$17,$07,$C6,$7F,$09,$C1,$F1,$9F,$C3,$70,$7C,$6B,$FF
	.dc.b	$4A,$70,$7E,$CF,$9B,$F1,$D1,$FC,$FD,$FD,$CB,$FB,$23,$F2,$A7,$EF
	.dc.b	$CF,$D9,$1F,$C0,$CF,$D9,$1F,$C2,$3C,$4F,$E1,$4F,$F5,$5F,$17,$F8
	.dc.b	$7B,$FC,$44,$FD,$97,$7F,$F2,$1F,$03,$E1,$EB,$FE,$37,$F0,$7F,$61
	.dc.b	$5F,$C7,$F9,$72,$FF,$1B,$AF,$D8,$BF,$C5,$BC,$6E,$BB,$7F,$5F,$CB
	.dc.b	$C2,$FF,$61,$CF,$F8,$CD,$7F,$67,$F1,$BF,$8C,$BF,$DA,$7C,$6F,$E3
	.dc.b	$2F,$F6,$BF,$1B,$E0,$BF,$C6,$0F,$E2,$C7,$F8,$93,$E1,$3F,$66,$E0
	.dc.b	$D1,$FD,$CF,$C1,$FD,$D3,$F2,$0E,$C7,$C8,$3C,$47,$C8,$3F,$BD,$61
	.dc.b	$E8,$DF,$FB,$9E,$7E,$C7,$FB,$AE,$9F,$02,$BF,$9F,$EB,$F1,$37,$F8
	.dc.b	$FB,$7F,$92,$F7,$C7,$8F,$FC,$72,$3F,$8F,$70,$FF,$17,$7F,$C0,$9E
	.dc.b	$3F,$F1,$77,$FC,$0D,$E3,$FF,$17,$7F,$C1,$1E,$3F,$F1,$77,$FC,$15
	.dc.b	$E3,$FF,$17,$7F,$C1,$9E,$3F,$B9,$FE,$71,$AF,$C2,$7F,$C2,$1C,$FF
	.dc.b	$B2,$8F,$F1,$CF,$47,$E4,$76,$FF,$69,$F4,$FE,$57,$8B,$FC,$8D,$D5
	.dc.b	$F9,$97,$FC,$91,$D7,$F9,$D7,$FC,$2A,$7F,$06,$C3,$FC,$41,$DF,$FC
	.dc.b	$29,$FF,$11,$78,$7F,$C3,$9F,$F1,$27,$89,$FE,$0F,$7F,$C2,$7E,$17
	.dc.b	$EC,$5F,$F0,$AF,$8B,$FC,$59,$D9,$FB,$A3,$99,$FB,$E3,$B9,$FC,$98
	.dc.b	$FE,$2C,$7F,$2A,$3F,$C2,$67,$F2,$E3,$FC,$26,$7F,$0E,$3F,$9A,$6F
	.dc.b	$FC,$DF,$6F,$F0,$9F,$92,$EE,$FD,$06,$1E,$5B,$FE,$0F,$9F,$F2,$D2
	.dc.b	$3E,$A6,$BA,$30,$43,$04,$C1,$A7,$69,$E3,$18,$9F,$F6,$F1,$C5,$C7
	.dc.b	$07,$F8,$E6,$39,$57,$17,$4D,$72,$72,$3C,$38,$FB,$5F,$F0,$97,$C5
	.dc.b	$7F,$87,$C8,$FF,$73,$F8,$9F,$F1,$17,$C5,$7F,$88,$FE,$17,$FC,$49
	.dc.b	$F0,$9F,$E2,$5F,$87,$FF,$13,$7C,$2F,$F8,$9F,$CF,$FF,$14,$7C,$2F
	.dc.b	$F8,$A7,$8B,$FC,$55,$F0,$BF,$E2,$BF,$88,$FF,$16,$7C,$37,$F8,$B7
	.dc.b	$E2,$7F,$C5,$DF,$0F,$FA,$7F,$5B,$FC,$53,$EB,$E3,$8B,$FC,$09,$5C
	.dc.b	$1F,$F0,$36,$1D,$1F,$F0,$46,$5E,$07,$FC,$15,$A7,$91,$FF,$06,$6D
	.dc.b	$E8,$7F,$C1,$DC,$1E,$C7,$DC,$7F,$84,$78,$FE,$0B,$FE,$53,$BF,$F1
	.dc.b	$FC,$63,$9F,$FC,$28,$FF,$91,$79,$39,$57,$F8,$FA,$FB,$3F,$E4,$0C
	.dc.b	$3C,$4F,$F9,$0B,$1F,$33,$FE,$44,$C7,$D4,$FF,$91,$B2,$F7,$3F,$E4
	.dc.b	$8C,$FE,$13,$FE,$21,$3F,$98,$1F,$E2,$2A,$FF,$27,$67,$FE,$27,$DB
	.dc.b	$E6,$4F,$F8,$AF,$7F,$A0,$FF,$8B,$78,$3E,$A3,$FE,$2F,$E1,$FB,$0F
	.dc.b	$F8,$C7,$8B,$EE,$3F,$E3,$3E,$3F,$C0,$3F,$E3,$5E,$4F,$80,$FC,$72
	.dc.b	$1F,$1C,$FC,$76,$5F,$96,$E9,$FD,$43,$7A,$7F,$3B,$D1,$F9,$58,$C5
	.dc.b	$7E,$60,$E3,$19,$23,$82,$74,$FE,$86,$78,$C7,$E7,$A3,$13,$C1,$B1
	.dc.b	$C2,$FC,$73,$96,$C6,$6F,$CA,$D3,$F7,$1D,$3F,$A7,$BF,$D8,$47,$FA
	.dc.b	$F9,$AF,$EB,$CC,$52,$BF,$D8,$71,$86,$BB,$21,$FB,$53,$46,$A7,$AB
	.dc.b	$93,$6E,$09,$3D,$E3,$83,$5F,$0C,$6A,$7C,$31,$CC,$91,$97,$8C,$6A
	.dc.b	$31,$1C,$A7,$A3,$E6,$31,$3B,$8C,$C7,$99,$AF,$66,$CF,$48,$E4,$73
	.dc.b	$97,$9F,$C5,$E8,$48,$F2,$90,$F0,$94,$E6,$71,$30,$53,$64,$36,$79
	.dc.b	$5E,$CE,$4E,$06,$A7,$D0,$E0,$87,$13,$B9,$4E,$47,$F3,$18,$F6,$B9
	.dc.b	$3A,$57,$C4,$24,$78,$A3,$2D,$7B,$CF,$89,$E4,$62,$78,$BE,$28,$D1
	.dc.b	$23,$31,$E4,$8C,$4F,$58,$E6,$48,$FE,$9D,$1E,$87,$71,$D6,$79,$BE
	.dc.b	$63,$13,$CE,$77,$1D,$E7,$6F,$9E,$75,$18,$8F,$03,$67,$8C,$75,$39
	.dc.b	$4F,$03,$EB,$9E,$59,$FD,$DC,$FF,$00,$23,$82,$47,$EF,$A3,$2D,$8C
	.dc.b	$CF,$F1,$39,$CB,$63,$2F,$F0,$F6,$E3,$FC,$48,$98,$FF,$0D,$B5,$7F
	.dc.b	$95,$4F,$E1,$87,$17,$C0,$47,$B3,$5D,$18,$21,$82,$4F,$81,$EA,$DB
	.dc.b	$F4,$C7,$F3,$56,$C6,$6F,$D3,$7E,$7A,$EB,$39,$BF,$13,$63,$33,$B7
	.dc.b	$C2,$D8,$D4,$60,$F1,$C6,$4F,$F0,$A6,$3F,$CA,$FC,$97,$C4,$FC,$6A
	.dc.b	$5F,$FC,$01,$8C,$36,$3E,$5B,$FE,$22,$97,$C4,$6B,$0F,$F4,$B4,$3E
	.dc.b	$2B,$16,$D1,$2F,$6F,$30,$FE,$01,$FE,$6F,$97,$F8,$B9,$7C,$4E,$97
	.dc.b	$C4,$DF,$7F,$1F,$13,$9F,$81,$D7,$C4,$EF,$E0,$F8,$7C,$4F,$1F,$85
	.dc.b	$E5,$F1,$3C,$FE,$1F,$A7,$C4,$CE,$3F,$6F,$2F,$FA,$BF,$75,$70,$9C
	.dc.b	$9E,$B8,$61,$DA,$33,$C5,$EB,$CF,$C7,$18,$A6,$BE,$88,$C6,$FE,$DE
	.dc.b	$0F,$77,$0F,$BE,$7E,$2F,$27,$C7,$7A,$BF,$35,$BC,$DA,$E7,$91,$F9
	.dc.b	$FD,$3E,$67,$C4,$FC,$92,$47,$8F,$CC,$F9,$9E,$77,$CD,$F5,$3E,$77
	.dc.b	$C4,$7C,$FF,$57,$E8,$7B,$BF,$47,$D6,$FA,$5F,$11,$F6,$3E,$0F,$FC
	.dc.b	$CD,$9F,$E2,$3A,$7F,$DA,$D0,$F8,$9D,$78,$9B,$F8,$7E,$1E,$D7,$1F
	.dc.b	$88,$FA,$58,$FE,$FF,$AF,$E7,$BF,$2B,$14,$F0,$8C,$5F,$47,$58,$F8
	.dc.b	$F2,$F2,$C7,$CF,$D3,$D2,$FD,$1D,$BD,$AF,$D3,$E0,$F8,$2F,$C3,$3E
	.dc.b	$B7,$1F,$C6,$7F,$71,$A2,$31,$E2,$7E,$55,$FC,$A7,$27,$2A,$F8,$47
	.dc.b	$E1,$6F,$BB,$8E,$5F,$CF,$63,$E4,$8C,$62,$CB,$D3,$1B,$DF,$E6,$E7
	.dc.b	$FE,$90,$7E,$64,$7C,$5E,$1F,$98,$FC,$7E,$1F,$98,$FF,$06,$3F,$C6
	.dc.b	$F3,$F2,$9D,$1F,$28,$E4,$7C,$A3,$E8,$6D,$E2,$BF,$D9,$61,$E7,$7F
	.dc.b	$69,$97,$B1,$FD,$B6,$9F,$05,$FD,$C6,$BF,$1A,$FF,$55,$B7,$F0,$DF
	.dc.b	$1F,$FC,$75,$B7,$CB,$CF,$E7,$90,$F8,$77,$CE,$FF,$65,$8F,$77,$F6
	.dc.b	$99,$78,$DF,$DB,$67,$E7,$7F,$22,$7E,$57,$83,$F0,$D8,$7C,$BE,$1F
	.dc.b	$C5,$3F,$37,$8B,$F9,$6D,$7E,$61,$CE,$DF,$3E,$B8,$BF,$46,$FF,$1D
	.dc.b	$7F,$BB,$E3,$FD,$23,$FB,$DE,$4F,$D4,$3F,$BF,$E5,$FD,$63,$FC,$07
	.dc.b	$9B,$F0,$EF,$BF,$0F,$C5,$3F,$B7,$E3,$7B,$7F,$AD,$24,$7E,$8F,$E2
	.dc.b	$7F,$61,$EE,$FE,$C6,$BF,$65,$E6,$7E,$CF,$DE,$FD,$A7,$C4,$7E,$D7
	.dc.b	$E0,$3F,$6D,$F1,$3F,$B7,$F8,$2F,$DC,$7C,$4F,$EE,$7E,$13,$F7,$5F
	.dc.b	$13,$FB,$BF,$86,$FD,$E7,$C4,$FE,$D3,$83,$F2,$BE,$7F,$F2,$57,$F2
	.dc.b	$FE,$7E,$9F,$DB,$57,$27,$F7,$18,$76,$7F,$75,$97,$89,$FD,$E6,$9E
	.dc.b	$67,$F7,$DB,$7A,$9F,$E0,$1C,$1E,$E7,$E0,$1F,$C0,$F8,$FE,$1B,$FB
	.dc.b	$4F,$0F,$F1,$CF,$F3,$37,$E6,$BF,$AC,$E6,$FA,$2F,$EB,$B9,$FE,$AB
	.dc.b	$FB,$0E,$8F,$B2,$FE,$CB,$A7,$EE,$BF,$B1,$EA,$FC,$0F,$6F,$F1,$CE
	.dc.b	$CF,$C2,$BF,$87,$9F,$E7,$9C,$5F,$D6,$BD,$6F,$F1,$0F,$87,$FC,$6A
	.dc.b	$7F,$1C,$7E,$C0,$FC,$91,$FB,$03,$B1,$FB,$03,$F1,$87,$F1,$93,$F1
	.dc.b	$87,$F1,$93,$F1,$87,$F1,$93,$F1,$87,$F1,$93,$F8,$5E,$BF,$C6,$7C
	.dc.b	$4F,$D7,$BF,$1B,$C7,$FE,$87,$3B,$F0,$3F,$B7,$70,$7C,$67,$FB,$87
	.dc.b	$07,$C6,$7F,$B9,$70,$7C,$67,$FB,$A7,$07,$C6,$7F,$49,$7E,$B7,$F5
	.dc.b	$0F,$BD,$FE,$31,$E0,$FE,$BA,$3F,$C0,$5C,$7F,$C4,$DF,$E3,$DE,$37
	.dc.b	$F1,$A7,$F9,$07,$8F,$FB,$97,$E2,$1F,$54,$F8,$87,$D9,$3E,$21,$F7
	.dc.b	$4F,$88,$7E,$04,$F7,$71,$7F,$12,$7F,$87,$6F,$F1,$1F,$D9,$46,$5F
	.dc.b	$E2,$27,$CC,$3F,$9A,$F8,$DF,$C4,$5F,$F0,$07,$8D,$FC,$45,$FF,$01
	.dc.b	$78,$DF,$C4,$5F,$6D,$7A,$5F,$59,$FC,$BC,$F7,$9F,$D7,$73,$F8,$9C
	.dc.b	$2D,$3E,$26,$BF,$CF,$F5,$F8,$8F,$F4,$1D,$7A,$3F,$D0,$F1,$F0,$5F
	.dc.b	$F8,$A7,$2F,$23,$FD,$3B,$2F,$C3,$9E,$A3,$D9,$5F,$CD,$9F,$82,$7E
	.dc.b	$A4,$FE,$25,$D1,$FA,$E8,$FE,$A0,$FF,$7F,$F1,$BF,$69,$3F,$D5,$73
	.dc.b	$FD,$C3,$FD,$5F,$AB,$F7,$8F,$F5,$9E,$BF,$E0,$0F,$F5,$BE,$CF,$E0
	.dc.b	$8F,$F5,$DE,$DF,$E1,$0F,$F5,$FE,$EF,$E1,$8F,$F0,$E3,$FC,$51,$DF
	.dc.b	$FC,$47,$4F,$F2,$27,$6E,$9F,$EC,$95,$C9,$FE,$CD,$7D,$9F,$ED,$18
	.dc.b	$78,$9F,$ED,$58,$F9,$9F,$ED,$99,$7A,$9F,$ED,$D9,$FB,$9F,$EE,$1A
	.dc.b	$7C,$3E,$0F,$F0,$EE,$DF,$D6,$FE,$07,$FD,$1B,$C0,$FE,$01,$7C,$4F
	.dc.b	$F9,$71,$BC,$5F,$E4,$46,$BF,$E0,$35,$F1,$5B,$F8,$8E,$1F,$11,$C7
	.dc.b	$DD,$E5,$F1,$1C,$FD,$FE,$9F,$11,$D7,$E0,$7B,$7C,$37,$7F,$83,$F0
	.dc.b	$70,$78,$7E,$13,$C5,$F1,$1E,$3F,$86,$F2,$7C,$47,$F1,$7F,$0A,$71
	.dc.b	$F9,$63,$4F,$9E,$B9,$3E,$9C,$3B,$3E,$BC,$BC,$4F,$B7,$4F,$33,$EF
	.dc.b	$DB,$D4,$FC,$1E,$0F,$73,$F0,$F8,$BE,$37,$24,$8F,$F2,$C7,$23,$E2
	.dc.b	$3E,$51,$23,$E5,$7B,$5F,$37,$DB,$F9,$DF,$11,$F3,$FE,$23,$FD,$17
	.dc.b	$FC,$03,$FE,$8A,$ED,$FA,$AF,$FA,$2B,$0F,$F3,$AB,$FE,$8A,$6F,$C1
	.dc.b	$70,$F8,$9E,$3F,$09,$CB,$E2,$79,$FC,$37,$4F,$89,$EB,$F1,$1F,$D9
	.dc.b	$7C,$6F,$F5,$79,$F9,$4C,$BA,$C6,$DE,$F5,$CD,$F0,$E1,$DD,$F0,$9E
	.dc.b	$4F,$88,$F4,$FC,$47,$AA,$3D,$87,$E2,$78,$3E,$03,$F0,$8F,$C6,$71
	.dc.b	$FC,$57,$FD,$9D,$C9,$F2,$5C,$90,$FC,$87,$27,$27,$F2,$55,$D9,$FF
	.dc.b	$2B,$70,$FC,$97,$FD,$70,$FC,$97,$38,$FC,$EE,$1F,$54,$70,$CF,$DC
	.dc.b	$F2,$D3,$E1,$3D,$35,$F9,$1C,$5F,$ED,$38,$FF,$55,$79,$FF,$31,$FA
	.dc.b	$BC,$3F,$41,$FA,$FC,$5F,$51,$FB,$3C,$7E,$17,$E4,$1F,$73,$8B,$E4
	.dc.b	$3F,$AF,$E4,$FC,$2B,$FB,$1E,$5F,$DA,$71,$7C,$7E,$56,$9F,$BA,$E4
	.dc.b	$F9,$17,$F2,$B9,$BF,$28,$FE,$DD,$FD,$E7,$3F,$E6,$A3,$F7,$5D,$1F
	.dc.b	$9E,$7F,$79,$D3,$FA,$27,$F7,$DD,$5F,$A6,$7F,$80,$75,$FE,$A9,$FE
	.dc.b	$05,$D9,$FA,$E7,$F8,$27,$6F,$EC,$9F,$E0,$DD,$DF,$C8,$75,$FF,$6A
	.dc.b	$7A,$1F,$BA,$9F,$E1,$9E,$0F,$EF,$9F,$E1,$DE,$17,$F0,$27,$F8,$87
	.dc.b	$87,$FC,$19,$FE,$25,$E2,$7F,$0A,$7F,$8A,$78,$BF,$C3,$9F,$E2,$DE
	.dc.b	$37,$F1,$27,$F8,$C7,$8F,$F0,$1F,$E3,$5C,$9F,$13,$D6,$70,$7C,$98
	.dc.b	$F8,$FC,$FF,$32,$BF,$E0,$07,$3F,$F2,$08,$C7,$93,$FC,$9F,$A3,$E5
	.dc.b	$46,$25,$D1,$F2,$E3,$1D,$1F,$2D,$D7,$47,$CB,$78,$74,$7C,$B7,$96
	.dc.b	$9F,$2D,$E9,$D1,$F2,$DE,$DD,$1F,$C8,$35,$FF,$5A,$7B,$AF,$2B,$E6
	.dc.b	$47,$F3,$5E,$8F,$98,$FF,$37,$E8,$F9,$8F,$F3,$9E,$8F,$98,$FF,$3B
	.dc.b	$E8,$F9,$8F,$F3,$DE,$8F,$98,$FF,$3F,$E8,$F9,$8F,$F4,$1E,$8F,$98
	.dc.b	$FE,$34,$F9,$A7,$F2,$2F,$2B,$F2,$DD,$7F,$95,$7F,$33,$1F,$9B,$3F
	.dc.b	$82,$1F,$9F,$3F,$8E,$9F,$C2,$8F,$D2,$9F,$BC,$F1,$FF,$6F,$C1,$F4
	.dc.b	$BC,$97,$91,$FB,$88,$C7,$91,$FB,$87,$5E,$47,$EE,$30,$76,$B9,$3F
	.dc.b	$CB,$FE,$47,$EE,$67,$F8,$4F,$91,$FB,$97,$F8,$5F,$91,$FB,$97,$F8
	.dc.b	$6F,$91,$FB,$97,$F8,$59,$FB,$93,$D2,$7A,$CF,$E2,$A7,$BC,$FE,$32
	.dc.b	$7F,$1B,$CB,$F1,$1B,$FF,$73,$8F,$C7,$BF,$21,$A7,$F2,$73,$FB,$87
	.dc.b	$8D,$FC,$11,$FE,$E5,$E3,$FF,$08,$7F,$90,$F9,$1F,$D2,$BC,$9F,$36
	.dc.b	$1F,$1F,$1F,$53,$F2,$72,$F7,$3F,$2F,$3F,$84,$FF,$43,$C3,$F1,$BD
	.dc.b	$9F,$8F,$8C,$1F,$32,$3F,$8F,$9F,$D1,$8F,$E3,$E7,$F4,$63,$F8,$F9
	.dc.b	$FD,$18,$F8,$E7,$98,$F8,$E7,$A8,$FC,$19,$EE,$3E,$03,$F1,$9F,$E8
	.dc.b	$5E,$3F,$EE,$1F,$DA,$F8,$FF,$B8,$E7,$FD,$D4,$7F,$75,$3F,$80,$9F
	.dc.b	$DE,$4F,$E0,$A7,$F7,$D3,$F8,$49,$FE,$01,$3F,$86,$9F,$E0,$53,$F8
	.dc.b	$89,$FE,$09,$3F,$8A,$9F,$E0,$D3,$F8,$C9,$FE,$11,$3F,$C7,$E4,$3F
	.dc.b	$90,$9A,$3F,$92,$9C,$8F,$E5,$27,$63,$F9,$69,$E2,$3F,$98,$9E,$63
	.dc.b	$F9,$A9,$EA,$3F,$9C,$9E,$E3,$FC,$69,$E3,$FC,$37,$E2,$81,$5F,$27
	.dc.b	$8D,$CE,$F8,$9F,$F0,$89,$2B,$FC,$1E,$D7,$FB,$6B,$72,$FD,$93,$71
	.dc.b	$FF,$0B,$B6,$7F,$B8,$B6,$FB,$B6,$FC,$13,$D8,$F9,$47,$E4,$0F,$45
	.dc.b	$FF,$30,$E5,$FE,$3D,$1A,$8F,$0B,$C4,$F2,$95,$CB,$82,$38,$26,$1E
	.dc.b	$88,$E0,$D7,$6D,$8D,$C7,$B0,$F9,$98,$FC,$BB,$FF,$74,$4F,$F1,$5E
	.dc.b	$5F,$F1,$D9,$FE,$3C,$F2,$7F,$C7,$BE,$EF,$F8,$FB,$E2,$3F,$C7,$FE
	.dc.b	$F7,$F9,$03,$DB,$FF,$20,$FB,$DF,$E4,$2F,$73,$FC,$87,$EF,$7F,$91
	.dc.b	$3D,$BF,$F2,$2F,$BD,$F6,$F8,$1B,$7F,$54,$FE,$01,$FE,$B1,$18,$E2
	.dc.b	$FF,$3E,$C3,$E0,$B8,$56,$FA,$6D,$6B,$E0,$BF,$A4,$CF,$F9,$12,$3F
	.dc.b	$C1,$9E,$76,$B1,$FE,$53,$8A,$F1,$D2,$8F,$AB,$3E,$13,$79,$6A,$3C
	.dc.b	$B8,$FF,$39,$CF,$D9,$5F,$E5,$CD,$7E,$03,$FE,$5E,$DF,$E2,$3F,$E6
	.dc.b	$0E,$1F,$90,$FF,$98,$B8,$FE,$63,$FE,$64,$E5,$F9,$EF,$D4,$F4,$3F
	.dc.b	$CD,$31,$88,$FF,$22,$BC,$1F,$F2,$3C,$F4,$7F,$C9,$35,$E0,$7F,$C9
	.dc.b	$77,$E4,$7F,$C9,$B8,$7A,$1F,$F2,$7E,$3E,$DE,$0F,$76,$BE,$FE,$1F
	.dc.b	$81,$C1,$F3,$27,$FC,$91,$C1,$FE,$AF,$E1,$CE,$3F,$9E,$9F,$AF,$7C
	.dc.b	$A3,$07,$D0,$8F,$C0,$BF,$E8,$2E,$2F,$C2,$BF,$E8,$4E,$3F,$C4,$BF
	.dc.b	$E8,$6E,$4F,$C6,$BF,$E8,$8E,$5F,$C8,$BF,$E8,$AE,$6F,$89,$CB,$F9
	.dc.b	$78,$F9,$07,$E7,$4F,$D0,$95,$FD,$0E,$9F,$BA,$8F,$D3,$99,$CB,$F4
	.dc.b	$C4,$70,$48,$E0,$F4,$6C,$F4,$E7,$FE,$07,$39,$E7,$7C,$07,$F9,$26
	.dc.b	$3F,$BC,$FC,$13,$2F,$C6,$4F,$F0,$BE,$C9,$5E,$4C,$BA,$E7,$E3,$6C
	.dc.b	$78,$F2,$EF,$9F,$86,$F8,$CF,$96,$7C,$7D,$7F,$BD,$9F,$D8,$4F,$EC
	.dc.b	$6F,$D6,$DA,$F4,$B6,$76,$FF,$0B,$30,$62,$78,$B9,$21,$F4,$4D,$9E
	.dc.b	$07,$6E,$4D,$B8,$24,$F8,$A3,$F7,$27,$E2,$A3,$FC,$41,$C3,$FD,$EF
	.dc.b	$1C,$CF,$F2,$53,$B9,$0E,$B5,$CD,$F0,$12,$BB,$DF,$BA,$31,$3B,$9C
	.dc.b	$C7,$BE,$30,$49,$FE,$35,$6F,$88,$EF,$1C,$EF,$8D,$FC,$1B,$E0,$F3
	.dc.b	$9D,$46,$23,$51,$82,$3B,$21,$A2,$9E,$F3,$FC,$FF,$C7,$F1,$5F,$F2
	.dc.b	$3F,$6F,$F1,$CD,$3F,$CE,$31,$FD,$3C,$E1,$87,$99,$B8,$78,$DB,$87
	.dc.b	$89,$F0,$10,$C9,$E2,$3F,$A8,$65,$F0,$DF,$97,$1C,$5D,$9C,$4F,$8E
	.dc.b	$71,$36,$5A,$F2,$3A,$38,$1E,$E3,$F0,$98,$7B,$73,$FF,$03,$4F,$E1
	.dc.b	$A7,$FB,$74,$71,$7C,$24,$30,$43,$C2,$76,$8F,$0B,$FC,$2E,$FD,$91
	.dc.b	$82,$61,$F1,$E7,$94,$63,$0F,$81,$18,$9D,$3F,$91,$E0,$FE,$0A,$FF
	.dc.b	$8B,$F2,$FF,$5A,$F0,$7F,$05,$8F,$D0,$4E,$9F,$CF,$1C,$EF,$AB,$E8
	.dc.b	$3D,$85,$34,$7C,$BB,$FE,$2B,$8F,$A9,$FE,$13,$97,$F0,$D6,$CF,$F0
	.dc.b	$DA,$FE,$17,$8F,$F0,$D8,$FE,$29,$3A,$CB,$F8,$99,$1C,$12,$BF,$C5
	.dc.b	$18,$FA,$6F,$FB,$A5,$FF,$72,$DF,$AE,$3E,$1B,$B3,$E1,$3B,$B8,$C7
	.dc.b	$B3,$FC,$7B,$7F,$E3,$93,$A2,$1F,$C8,$63,$86,$5E,$38,$CE,$3E,$B7
	.dc.b	$47,$E1,$8F,$D3,$E7,$FA,$9C,$3F,$AF,$47,$0C,$7C,$CD,$3C,$C7,$22
	.dc.b	$57,$53,$B6,$3E,$68,$CC,$F2,$69,$CA,$78,$46,$5B,$19,$8F,$E3,$F1
	.dc.b	$F2,$AB,$CB,$3C,$9F,$8A,$F9,$5A,$E7,$0F,$84,$DC,$7F,$CC,$98,$7F
	.dc.b	$9A,$F3,$7C,$1F,$F8,$7C,$95,$F9,$2F,$73,$FC,$45,$EE,$FF,$88,$FE
	.dc.b	$27,$FC,$49,$EF,$FF,$89,$79,$3F,$C4,$DF,$01,$FE,$27,$F8,$9F,$F1
	.dc.b	$47,$C1,$7F,$8A,$7E,$27,$FC,$55,$F0,$9F,$E2,$BF,$89,$FF,$16,$7C
	.dc.b	$37,$F8,$B7,$E2,$7F,$C5,$DF,$11,$FD,$93,$83,$F4,$9D,$1F,$BC,$EB
	.dc.b	$FD,$24,$7F,$8D,$7A,$3F,$C7,$1D,$7C,$23,$F4,$F8,$74,$7F,$53,$97
	.dc.b	$81,$FD,$5E,$9E,$47,$F5,$BB,$7A,$1F,$D7,$F0,$7B,$1F,$D8,$F1,$7C
	.dc.b	$07,$E1,$1F,$B4,$E5,$F8,$AF,$FA,$53,$CB,$FF,$2B,$FC,$4F,$F9,$1C
	.dc.b	$91,$FD,$67,$A9,$E7,$7F,$94,$17,$9D,$7D,$BC,$7C,$36,$7D,$DD,$7C
	.dc.b	$36,$FD,$FE,$1F,$0D,$C7,$E0,$79,$7C,$37,$3F,$83,$E9,$C1,$D7,$E1
	.dc.b	$3B,$7C,$37,$CB,$E1,$FF,$4A,$13,$C9,$FF,$38,$75,$EA,$3C,$D5,$C9
	.dc.b	$F4,$61,$D9,$F5,$65,$E2,$7D,$9A,$79,$9F,$76,$DE,$A7,$E0,$70,$7B
	.dc.b	$F8,$7E,$07,$1F,$C4,$9F,$B5,$E3,$B9,$7F,$D0,$9E,$47,$E4,$DD,$6F
	.dc.b	$FE,$85,$78,$D7,$CB,$C3,$E4,$3F,$37,$97,$E4,$3F,$3F,$97,$E4,$3F
	.dc.b	$47,$97,$E4,$3F,$4F,$97,$E4,$3F,$57,$97,$E4,$3F,$5F,$97,$E4,$67
	.dc.b	$FA,$FD,$7F,$8D,$78,$1F,$D3,$B9,$D5,$C2,$FF,$29,$1F,$EF,$D6,$F2
	.dc.b	$FC,$97,$3C,$BF,$25,$DF,$2F,$C9,$78,$F2,$FC,$97,$9F,$2F,$C9,$7F
	.dc.b	$15,$CB,$F1,$33,$FC,$7C,$7F,$98,$3B,$FF,$D3,$5C,$B9,$8F,$F4,$FB
	.dc.b	$F2,$1F,$2F,$2F,$C8,$7C,$FC,$BF,$21,$F4,$F2,$FC,$87,$D7,$CB,$F2
	.dc.b	$1F,$6F,$2F,$C8,$7D,$FC,$BF,$07,$95,$A7,$C5,$E4,$F9,$2F,$FA,$23
	.dc.b	$97,$E4,$DB,$6F,$F7,$C5,$71,$AF,$99,$87,$5C,$39,$DE,$F9,$59,$78
	.dc.b	$E7,$38,$FF,$C0,$1C,$FD,$11,$8A,$69,$ED,$8C,$6B,$F0,$5D,$6D,$F1
	.dc.b	$5F,$C7,$65,$FD,$C3,$7F,$C9,$E9,$F7,$FE,$1B,$F7,$04,$9F,$DE,$79
	.dc.b	$BF,$BB,$F8,$2F,$DE,$7C,$47,$EF,$7E,$0F,$F7,$DF,$13,$FB,$FF,$84
	.dc.b	$FF,$67,$FF,$00,$FF,$9C,$B5,$F4,$AF,$85,$CF,$C2,$EB,$E2,$77,$F0
	.dc.b	$DC,$3E,$27,$8F,$C3,$F2,$F8,$9E,$7F,$11,$FE,$8D,$EB,$7A,$1F,$C9
	.dc.b	$B5,$EB,$3B,$E4,$FF,$35,$6C,$C7,$A4,$63,$3F,$03,$AD,$7C,$FB,$7A
	.dc.b	$23,$D4,$7F,$1B,$E0,$F7,$3F,$C7,$78,$BE,$13,$F1,$0F,$E4,$1C,$BF
	.dc.b	$1D,$FE,$4C,$7E,$50,$F9,$C4,$3F,$53,$E0,$FE,$7D,$FD,$09,$FB,$D3
	.dc.b	$F8,$69,$FB,$D3,$F8,$89,$FB,$D3,$F8,$A9,$FB,$D3,$F8,$C9,$FB,$D3
	.dc.b	$F8,$E9,$FB,$D3,$FC,$5B,$E4,$FF,$7F,$F6,$FF,$C9,$51,$FE,$15,$34
	.dc.b	$7E,$F3,$CA,$FE,$48,$FE,$F0,$EC,$7E,$F0,$F1,$1F,$BC,$3C,$C7,$EF
	.dc.b	$0F,$51,$EC,$3F,$66,$7E,$F0,$F8,$5C,$1F,$92,$7E,$3F,$17,$EF,$1C
	.dc.b	$F8,$5F,$C0,$1F,$8E,$7F,$23,$3F,$97,$9D,$A7,$F4,$5E,$37,$F0,$D7
	.dc.b	$F4,$9E,$3F,$F1,$17,$F4,$DE,$47,$F1,$57,$F5,$1E,$4F,$B9,$FD,$56
	.dc.b	$5F,$D1,$70,$FE,$6F,$A7,$C7,$F6,$FF,$7B,$1F,$B5,$F2,$3F,$74,$FE
	.dc.b	$DF,$C1,$FD,$D6,$BF,$C9,$FC,$8F,$DC,$3F,$CA,$7C,$8F,$DB,$BF,$CA
	.dc.b	$FC,$7F,$DB,$3F,$CB,$7C,$7F,$DB,$3F,$CB,$FC,$7F,$DB,$3F,$07,$6F
	.dc.b	$8D,$E5,$FF,$47,$CF,$E2,$9F,$8D,$3E,$29,$F3,$8F,$A1,$1F,$C0,$9F
	.dc.b	$C5,$9F,$C8,$0F,$C5,$9F,$C8,$0F,$C5,$9F,$C8,$0F,$C5,$9F,$C8,$0F
	.dc.b	$C5,$9F,$D0,$34,$F8,$7C,$9F,$E9,$18,$FE,$DD,$7F,$E9,$4F,$2B,$F9
	.dc.b	$0C,$FF,$20,$9E,$37,$FC,$5F,$6F,$8F,$1F,$C7,$F6,$F8,$EF,$F2,$1D
	.dc.b	$BE,$3B,$FC,$8F,$6F,$8E,$FF,$25,$DB,$E3,$BF,$C9,$F6,$F8,$EF,$C3
	.dc.b	$C7,$E3,$E3,$FC,$6E,$3F,$6A,$43,$FA,$CF,$91,$FB,$A7,$FA,$DF,$57
	.dc.b	$EF,$9F,$EB,$BD,$7F,$C0,$9F,$F1,$97,$81,$FC,$6F,$83,$FD,$62,$FF
	.dc.b	$1B,$8F,$E1,$47,$F1,$0F,$17,$F8,$DB,$FC,$4B,$C5,$FE,$36,$FF,$14
	.dc.b	$F1,$7F,$8D,$BF,$E4,$03,$FA,$37,$8B,$FC,$6F,$0E,$BE,$2F,$F1,$B9
	.dc.b	$F0,$F8,$BF,$C6,$9F,$1F,$8B,$FC,$69,$F2,$F8,$BF,$C6,$9F,$E4,$27
	.dc.b	$A7,$87,$FA,$36,$DE,$D7,$FA,$46,$FF,$05,$FE,$95,$C1,$FC,$65,$FE
	.dc.b	$99,$E2,$7F,$44,$BF,$EB,$5C,$2F,$89,$FE,$7A,$48,$FE,$5F,$E5,$FF
	.dc.b	$40,$F8,$9F,$E8,$3E,$F7,$F4,$2F,$89,$FE,$87,$F0,$1F,$D1,$3E,$27
	.dc.b	$FA,$2F,$C1,$7F,$46,$F8,$9F,$E8,$FF,$09,$FD,$23,$E2,$7F,$A4,$FC
	.dc.b	$37,$F4,$AF,$89,$6D,$38,$BF,$B1,$FF,$00,$7F,$68,$8F,$F3,$32,$F8
	.dc.b	$AF,$9B,$1F,$94,$3F,$D3,$E7,$F4,$4E,$BE,$6F,$08,$B5,$D3,$B3,$C1
	.dc.b	$3F,$D2,$B1,$F2,$3F,$D3,$33,$F4,$3F,$D3,$B5,$F6,$3F,$D4,$37,$F8
	.dc.b	$0F,$F5,$2E,$1F,$88,$FC,$63,$FA,$AF,$27,$C9,$7F,$90,$EB,$FC,$D3
	.dc.b	$C7,$34,$9F,$15,$E6,$D3,$FE,$00,$C7,$EB,$8F,$F4,$DD,$7C,$58,$CF
	.dc.b	$C4,$EE,$3F,$C2,$52,$F8,$9C,$61,$DB,$DE,$B2,$F8,$9B,$F0,$18,$F8
	.dc.b	$9C,$FC,$16,$BE,$27,$7F,$09,$C3,$E2,$78,$FC,$37,$D0,$E8,$FF,$5C
	.dc.b	$78,$4E,$BC,$F0,$F9,$4F,$1D,$F2,$71,$C1,$F0,$32,$EF,$18,$DD,$AF
	.dc.b	$92,$31,$BF,$A3,$83,$D3,$C3,$EA,$E2,$F5,$FC,$37,$B3,$D9,$F6,$F3
	.dc.b	$7B,$BD,$99,$97,$F9,$2F,$4F,$9D,$18,$8F,$C5,$CF,$17,$C5,$7D,$5F
	.dc.b	$26,$3E,$17,$C6,$79,$F9,$BE,$43,$E9,$E6,$F9,$0F,$AF,$9B,$E4,$3E
	.dc.b	$DE,$6F,$91,$CF,$F9,$EE,$3F,$E5,$9C,$FF,$E4,$0F,$7F,$F0,$9D,$1F
	.dc.b	$2F,$0E,$07,$C5,$E7,$F9,$4F,$E8,$B9,$FE,$4B,$FA,$4E,$7F,$90,$FE
	.dc.b	$9B,$9F,$D0,$FE,$A3,$83,$D8,$FE,$AB,$87,$E0,$3F,$08,$FA,$5C,$7F
	.dc.b	$17,$DA,$7A,$7F,$C8,$BA,$FF,$35,$BF,$F8,$F9,$FF,$20,$F9,$1F,$A0
	.dc.b	$C3,$F9,$4F,$6F,$E9,$1F,$E5,$7D,$DF,$A8,$7F,$96,$F7,$FE,$B1,$FE
	.dc.b	$5F,$E0,$7E,$C1,$FE,$63,$E0,$FE,$D1,$FD,$FF,$3F,$E6,$3C,$2F,$CC
	.dc.b	$FC,$4F,$F9,$17,$D1,$FE,$55,$5F,$A0,$8D,$3F,$8B,$AE,$5B,$7F,$95
	.dc.b	$30,$EC,$FF,$95,$B1,$F1,$3F,$E5,$8C,$BC,$CF,$F9,$6B,$3F,$53,$FE
	.dc.b	$5C,$D3,$DC,$FF,$97,$B5,$F8,$78,$FC,$5C,$BF,$D3,$1D,$1F,$CA,$5F
	.dc.b	$E0,$24,$3F,$A6,$CF,$29,$FE,$9B,$7D,$9F,$E9,$F8,$78,$9F,$EA,$38
	.dc.b	$F9,$9F,$EA,$79,$7A,$9F,$EA,$B9,$FB,$9F,$E0,$E7,$C2,$C7,$FA,$DF
	.dc.b	$89,$FE,$1D,$7F,$92,$1F,$AF,$DF,$F8,$1C,$F1,$8F,$E8,$35,$D5,$FE
	.dc.b	$87,$7E,$17,$FA,$2E,$1E,$57,$FA,$3E,$3E,$97,$FA,$4E,$5E,$D7,$FA
	.dc.b	$5E,$7F,$19,$FE,$9B,$C1,$FD,$3F,$AF,$F9,$0F,$27,$EC,$F8,$3F,$9F
	.dc.b	$BF,$BA,$AF,$E0,$9E,$17,$EF,$9F,$E0,$DE,$1F,$F0,$27,$F8,$47,$89
	.dc.b	$FC,$73,$AB,$F4,$38,$7F,$48,$EB,$FD,$2B,$FD,$2B,$B3,$F5,$2F,$F4
	.dc.b	$CE,$DF,$E2,$F8,$7C,$63,$F8,$AF,$77,$F3,$53,$F6,$7C,$FA,$AE,$13
	.dc.b	$FB,$D3,$F8,$C5,$F6,$E6,$FE,$6F,$8F,$89,$FE,$73,$97,$EC,$AF,$FA
	.dc.b	$DF,$77,$EC,$9F,$EB,$BD,$DF,$B2,$7F,$AF,$F8,$7F,$0E,$FF,$A0,$F0
	.dc.b	$FC,$7F,$13,$F8,$CE,$9F,$B6,$3F,$94,$57,$1A,$FE,$8F,$87,$57,$FA
	.dc.b	$4E,$3E,$17,$C7,$5F,$C6,$4F,$E6,$19,$FA,$5F,$E9,$3A,$7C,$69,$FE
	.dc.b	$A3,$BF,$C1,$7F,$C2,$F3,$F1,$B3,$FE,$9B,$BF,$C6,$8F,$E2,$E7,$F7
	.dc.b	$6A,$E4,$F4,$3F,$93,$1E,$03,$C4,$7F,$17,$3E,$D9,$FC,$C0,$FB,$E7
	.dc.b	$F3,$43,$F0,$67,$C0,$3E,$11,$FE,$04,$9F,$C6,$76,$7E,$C6,$3F,$C7
	.dc.b	$7C,$BF,$91,$E5,$4B,$9B,$F2,$D1,$8E,$7F,$CD,$3A,$E8,$FC,$F3,$C3
	.dc.b	$A7,$F4,$4F,$2E,$AF,$D3,$3D,$3A,$FF,$54,$FF,$7C,$DF,$F6,$5C,$DF
	.dc.b	$E1,$AE,$DF,$E3,$19,$7F,$1A,$DB,$F6,$47,$F1,$73,$FD,$B7,$E2,$FF
	.dc.b	$89,$1F,$F6,$EF,$8D,$FE,$28,$7F,$DB,$FE,$3F,$F8,$B1,$FF,$70,$F9
	.dc.b	$1F,$E3,$07,$FD,$C7,$E4,$FF,$8D,$1F,$F7,$2F,$95,$FC,$D9,$FE,$25
	.dc.b	$9F,$F3,$97,$FB,$26,$7E,$37,$FC,$5A,$D8,$F4,$47,$F7,$58,$C6,$9F
	.dc.b	$E1,$46,$C7,$F3,$13,$44,$39,$96,$79,$B8,$23,$83,$FB,$17,$2F,$E3
	.dc.b	$0E,$BB,$7F,$89,$AA,$F6,$9A,$4B,$ED,$87,$95,$A7,$94,$EE,$E5,$A6
	.dc.b	$70,$F2,$F8,$DF,$D8,$23,$FC,$57,$F1,$1F,$E2,$82,$3F,$E1,$7F,$89
	.dc.b	$FF,$15,$FC,$47,$F8,$BB,$F8,$05,$F0,$B7,$DA,$C7,$C5,$67,$DC,$D7
	.dc.b	$C3,$6F,$DE,$E1,$F0,$DC,$7E,$03,$97,$C3,$73,$F8,$2E,$9F,$0D,$D7
	.dc.b	$D5,$FF,$04,$F0,$FF,$97,$8F,$F3,$07,$47,$E7,$3D,$0D,$47,$F8,$BB
	.dc.b	$85,$7C,$E3,$18,$F7,$75,$9F,$8D,$E1,$AF,$9D,$E5,$BF,$AD,$E9,$C3
	.dc.b	$F0,$1F,$84,$7F,$8E,$79,$3E,$2B,$FE,$5E,$6F,$37,$F9,$E9,$FF,$28
	.dc.b	$F2,$F1,$94,$B9,$FF,$C4,$9A,$FC,$A9,$F5,$72,$FC,$A7,$D9,$CB,$F2
	.dc.b	$9F,$77,$2F,$CA,$7E,$07,$2F,$CA,$7E,$17,$2F,$CA,$7E,$27,$2F,$CA
	.dc.b	$E5,$FF,$18,$E7,$FE,$33,$F2,$BF,$8D,$57,$92,$3C,$75,$C6,$30,$29
	.dc.b	$A3,$0F,$04,$EA,$73,$8F,$96,$31,$4C,$BD,$51,$8C,$FD,$CE,$B4,$FC
	.dc.b	$67,$07,$C7,$DB,$FD,$47,$BF,$CC,$DB,$FD,$0B,$C1,$FD,$17,$A7,$F4
	.dc.b	$58,$FE,$73,$6F,$D4,$3F,$A2,$6B,$9E,$AF,$D5,$39,$6B,$99,$E7,$8F
	.dc.b	$ED,$70,$FD,$43,$67,$FD,$B7,$E1,$7F,$A3,$BD,$5F,$F1,$9E,$3D,$AB
	.dc.b	$4D,$BE,$24,$BF,$D8,$36,$35,$18,$23,$82,$63,$E6,$AE,$AD,$AE,$4F
	.dc.b	$AB,$B3,$F7,$55,$F0,$2F,$FD,$86,$D7,$F8,$11,$89,$F8,$2D,$CB,$FD
	.dc.b	$92,$D2,$9F,$C2,$CC,$1A,$9E,$8E,$4E,$A4,$30,$47,$64,$9F,$0C,$6E
	.dc.b	$3F,$2A,$FE,$FF,$0F,$E5,$2F,$80,$86,$E7,$8B,$D5,$B3,$D6,$BD,$33
	.dc.b	$96,$C6,$63,$DA,$D7,$D9,$3C,$8F,$C2,$F6,$BD,$9F,$05,$F2,$9D,$34
	.dc.b	$95,$D2,$32,$D8,$CB,$D9,$AF,$B0,$F3,$1B,$21,$B3,$CF,$CF,$FE,$37
	.dc.b	$BF,$EB,$51,$C7,$0F,$53,$6F,$C2,$F1,$21,$9B,$F0,$BE,$C3,$B9,$90
	.dc.b	$2B,$E1,$3E,$63,$99,$C4,$A6,$4E,$46,$2B,$CA,$EC,$C9,$E9,$BF,$63
	.dc.b	$FD,$AC,$C9,$B2,$1D,$4B,$1D,$1C,$1E,$59,$F3,$C6,$49,$7F,$12,$7A
	.dc.b	$9C,$23,$37,$F1,$23,$13,$B7,$FD,$B5,$A7,$F0,$03,$F8,$0C,$66,$37
	.dc.b	$3C,$6B,$9C,$7C,$C3,$CF,$3B,$8C,$C7,$CD,$C7,$F7,$75,$FC,$4B,$4B
	.dc.b	$3F,$C5,$E3,$0F,$F0,$6B,$FE,$04,$D8,$CD,$FF,$08,$7F,$C6,$31,$C4
	.dc.b	$87,$F1,$F2,$E7,$FC,$32,$38,$D7,$4A,$E5,$5C,$34,$FE,$1B,$39,$25
	.dc.b	$70,$24,$F4,$C3,$D2,$F7,$26,$1F,$C9,$B4,$FB,$F1,$89,$DC,$F8,$E3
	.dc.b	$91,$ED,$2E,$3E,$47,$81,$FE,$1F,$C7,$CB,$87,$8A,$33,$3B,$AE,$91
	.dc.b	$8C,$7D,$2F,$46,$BF,$E0,$EC,$7F,$40,$E6,$FA,$57,$22,$63,$E8,$24
	.dc.b	$F9,$CF,$F2,$06,$3F,$CD,$39,$1F,$0D,$FE,$29,$24,$FF,$86,$7E,$1B
	.dc.b	$FC,$59,$EE,$7F,$8B,$7D,$DF,$F1,$77,$C5,$7F,$8B,$FD,$FF,$F1,$87
	.dc.b	$C5,$7F,$8C,$7E,$07,$FC,$65,$F1,$5F,$B6,$CB,$FB,$97,$8B,$FC,$EB
	.dc.b	$83,$FC,$1D,$1F,$DE,$61,$F1,$5B,$9C,$F2,$FF,$8A,$3E,$13,$1C,$7F
	.dc.b	$C0,$23,$FC,$57,$F1,$38,$C7,$97,$C3,$59,$FF,$17,$FC,$4D,$CD,$F1
	.dc.b	$1F,$90,$BF,$1B,$99,$D5,$EF,$0E,$19,$71,$D3,$96,$1E,$C3,$FC,$79
	.dc.b	$B7,$C0,$7F,$C7,$DC,$1F,$11,$FF,$20,$71,$7C,$87,$E5,$1F,$E4,$3E
	.dc.b	$5F,$9A,$B9,$FF,$B2,$EC,$F8,$9F,$41,$27,$F8,$FF,$17,$C4,$97,$9B
	.dc.b	$ED,$F6,$FD,$DE,$D7,$BF,$D1,$F8,$1E,$1F,$C1,$F4,$BE,$17,$B5,$F0
	.dc.b	$FD,$4F,$89,$ED,$7C,$5F,$5B,$E3,$7B,$5F,$1F,$D8,$F9,$1E,$D7,$C9
	.dc.b	$F6,$BE,$57,$77,$C9,$E2,$F1,$39,$E3,$F2,$F8,$1F,$57,$91,$87,$39
	.dc.b	$C9,$F7,$F2,$F0,$38,$F0,$3F,$CC,$BA,$F9,$63,$16,$DF,$D5,$18,$E1
	.dc.b	$F7,$3A,$E3,$F8,$4F,$0E,$5F,$90,$FE,$43,$9B,$F2,$5C,$FF,$39,$CF
	.dc.b	$07,$CB,$7F,$0D,$5C,$AB,$F2,$F7,$D9,$FC,$CE,$1E,$27,$F3,$78,$F9
	.dc.b	$DF,$CE,$E7,$EC,$7F,$3F,$AF,$C1,$7F,$43,$B7,$C7,$7F,$C8,$9E,$47
	.dc.b	$F0,$ED,$3F,$C6,$BB,$38,$BE,$8D,$7C,$2E,$3F,$AA,$FC,$4E,$4F,$B2
	.dc.b	$FC,$6E,$5F,$BA,$FC,$8E,$6F,$C0,$BF,$2B,$9F,$F0,$AF,$CC,$E8,$F7
	.dc.b	$BF,$3B,$93,$E3,$F2,$FC,$58,$C1,$0F,$DC,$F2,$FC,$5D,$BF,$25,$CB
	.dc.b	$F1,$5F,$CA,$72,$FC,$57,$F2,$DC,$BF,$15,$FC,$C7,$2F,$C5,$7F,$35
	.dc.b	$CB,$F1,$5F,$CE,$72,$FC,$57,$F9,$97,$99,$FE,$D1,$CF,$FC,$D1,$CB
	.dc.b	$F1,$5F,$C2,$D7,$1E,$8F,$C8,$61,$D5,$FC,$96,$3E,$17,$F2,$99,$79
	.dc.b	$5F,$CB,$67,$E9,$7F,$31,$A7,$B5,$FC,$D6,$BF,$86,$7F,$39,$BF,$E2
	.dc.b	$BD,$5F,$E2,$5F,$C0,$1F,$C6,$B5,$FF,$47,$37,$83,$F8,$CB,$6F,$7F
	.dc.b	$0F,$FC,$56,$3F,$9C,$D7,$0F,$83,$C7,$C3,$E7,$E1,$35,$F0,$FB,$F8
	.dc.b	$5E,$1F,$0F,$C7,$E1,$F9,$7A,$BC,$FE,$1F,$A7,$C5,$75,$F8,$8E,$DF
	.dc.b	$15,$FB,$8D,$BF,$5D,$ED,$7F,$81,$7A,$31,$1E,$89,$E0,$FE,$CF,$85
	.dc.b	$87,$58,$C6,$5E,$17,$5A,$79,$5E,$1B,$7A,$5E,$5C,$1E,$FE,$1F,$81
	.dc.b	$DD,$F0,$78,$FE,$17,$24,$BF,$F1,$DF,$7F,$E4,$5F,$DC,$1F,$91,$E6
	.dc.b	$FE,$41,$D3,$F5,$63,$FA,$47,$57,$D9,$7F,$A5,$71,$7D,$D7,$FA,$67
	.dc.b	$5F,$E0,$5F,$E9,$DD,$9F,$85,$7F,$89,$1F,$AD,$3F,$16,$7F,$10,$F1
	.dc.b	$BE,$2E,$3F,$D4,$B9,$BE,$44,$FF,$14,$F2,$7F,$2D,$AF,$F1,$FF,$03
	.dc.b	$F2,$53,$D0,$FE,$BF,$E3,$7E,$D9,$FE,$C3,$E3,$7E,$D9,$FE,$C7,$E3
	.dc.b	$7E,$D9,$FE,$CB,$E3,$7E,$D9,$FE,$CF,$E3,$7E,$D9,$F8,$64,$E3,$F8
	.dc.b	$DA,$FF,$19,$8F,$8D,$9F,$E6,$1F,$DF,$1F,$C9,$4F,$E0,$47,$F2,$93
	.dc.b	$F8,$31,$FC,$BB,$83,$E3,$79,$3F,$CC,$A3,$F8,$71,$FC,$CB,$C6,$F6
	.dc.b	$C3,$C7,$FE,$7B,$3F,$1C,$FF,$6E,$79,$3F,$C6,$F0,$FF,$65,$4F,$ED
	.dc.b	$E7,$F9,$09,$F3,$CF,$E4,$A7,$63,$F9,$49,$FC,$6C,$F2,$1E,$63,$F9
	.dc.b	$89,$EA,$3F,$9A,$9E,$E3,$F8,$A9,$F0,$8F,$F0,$67,$91,$FB,$76,$99
	.dc.b	$BF,$54,$7C,$DC,$BF,$B6,$69,$F4,$5F,$ED,$DA,$FD,$57,$FB,$86,$DF
	.dc.b	$65,$FE,$E5,$BF,$DD,$7F,$BA,$70,$7F,$36,$CB,$FC,$3F,$E4,$FF,$3A
	.dc.b	$7F,$C4,$3C,$BF,$E1,$5F,$6B,$FC,$B5,$C1,$F1,$E7,$FD,$DB,$BF,$C7
	.dc.b	$C7,$F9,$1C,$7C,$77,$F9,$3B,$F1,$DF,$E5,$6F,$C7,$3F,$A9,$9F,$1D
	.dc.b	$FE,$66,$FC,$77,$F9,$BB,$F1,$DF,$D5,$BF,$0E,$3F,$C2,$0F,$EC,$BC
	.dc.b	$AF,$E3,$BC,$3F,$B5,$8F,$E3,$AF,$EE,$5F,$E0,$DD,$BF,$BD,$7F,$7E
	.dc.b	$7F,$09,$F2,$BF,$8C,$BF,$C2,$FC,$AF,$E3,$2F,$F0,$DF,$2B,$F8,$CB
	.dc.b	$FC,$3F,$CA,$FE,$32,$FF,$11,$F2,$BF,$8C,$BF,$AF,$3F,$8C,$F7,$FF
	.dc.b	$8F,$7B,$8F,$ED,$B3,$FE,$67,$E5,$7F,$1A,$AF,$E6,$7E,$4F,$F1,$97
	.dc.b	$F9,$AF,$93,$FC,$65,$FE,$6F,$E4,$FF,$19,$7F,$9C,$F9,$3F,$C6,$5F
	.dc.b	$E7,$7E,$4F,$B5,$FE,$7B,$C3,$F0,$5F,$86,$7F,$27,$CB,$F1,$DD,$EF
	.dc.b	$29,$F1,$3F,$CF,$C9,$5F,$CB,$7E,$2B,$FA,$17,$B9,$FD,$0F,$DD,$FE
	.dc.b	$89,$EF,$7F,$45,$F8,$9F,$F2,$D7,$F0,$0B,$E0,$6F,$C1,$63,$E1,$F3
	.dc.b	$F0,$9A,$F8,$8D,$FC,$37,$0F,$89,$E3,$F1,$1C,$BE,$2B,$9F,$C5,$7F
	.dc.b	$8E,$79,$BF,$CD,$5A,$A3,$5C,$7F,$E3,$38,$5F,$38,$C6,$3D,$DD,$67
	.dc.b	$E3,$78,$6B,$E8,$DB,$D5,$1F,$D2,$F8,$3D,$CF,$F4,$DE,$2F,$EC,$5C
	.dc.b	$7F,$12,$3F,$B4,$F2,$7C,$AD,$3F,$87,$47,$8F,$C6,$F8,$67,$C4,$F1
	.dc.b	$BE,$2F,$B9,$F1,$BE,$27,$E3,$FC,$4F,$C8,$F7,$FE,$4F,$C4,$FC,$AF
	.dc.b	$81,$F9,$7F,$13,$F3,$3E,$0F,$E6,$FC,$4F,$FA,$1F,$F8,$07,$FD,$29
	.dc.b	$D5,$FE,$8D,$7F,$D2,$7B,$FF,$95,$57,$C3,$EF,$E2,$3F,$21,$2F,$11
	.dc.b	$EA,$7E,$ED,$FE,$4D,$F0,$5B,$9E,$57,$DF,$E2,$3C,$11,$E2,$3E,$86
	.dc.b	$9E,$67,$E9,$6D,$EA,$7E,$A7,$07,$B9,$FA,$DC,$5F,$09,$FB,$1C,$9F
	.dc.b	$19,$F9,$07,$EE,$CF,$37,$37,$CC,$5E,$B7,$FA,$BA,$30,$DE,$7F,$9D
	.dc.b	$3F,$83,$E7,$F9,$CF,$E1,$79,$FE,$73,$F8,$7E,$7F,$9C,$FE,$27,$9F
	.dc.b	$E7,$3F,$8B,$E7,$F7,$BF,$2F,$83,$E8,$70,$FC,$58,$FF,$59,$9F,$E8
	.dc.b	$3E,$4F,$95,$9F,$FA,$E2,$3E,$97,$2F,$07,$91,$F5,$4F,$94,$7B,$39
	.dc.b	$FF,$3B,$C9,$F2,$E3,$07,$C2,$E4,$F9,$8E,$39,$9C,$4E,$5F,$9D,$39
	.dc.b	$87,$2F,$CF,$8C,$72,$FC,$F7,$5A,$7F,$02,$F2,$3F,$80,$C7,$F8,$AF
	.dc.b	$A7,$F3,$DC,$FF,$E4,$4E,$AF,$E0,$33,$FE,$2F,$F2,$3F,$80,$BF,$E3
	.dc.b	$1F,$23,$F5,$4F,$EB,$0F,$DA,$79,$1F,$C0,$1F,$DB,$79,$1F,$C0,$1F
	.dc.b	$DC,$79,$1F,$C0,$1F,$DD,$79,$1F,$C0,$1F,$E5,$24,$3D,$9E,$47,$F0
	.dc.b	$07,$DD,$E4,$7E,$FD,$F8,$1E,$47,$EF,$DF,$85,$E4,$7E,$FD,$F8,$9E
	.dc.b	$47,$99,$FF,$30,$E9,$EA,$7F,$CC,$7A,$FB,$9F,$F3,$2E,$DF,$0B,$AB
	.dc.b	$FC,$D3,$C3,$F1,$CC,$4F,$EA,$F8,$BE,$4B,$C0,$FE,$10,$7C,$73,$F8
	.dc.b	$61,$F1,$CF,$E3,$1E,$47,$F2,$67,$E3,$9F,$C5,$8F,$8E,$7F,$1A,$3E
	.dc.b	$39,$FC,$7F,$5F,$85,$C7,$FA,$3D,$FF,$93,$4F,$EA,$1C,$E3,$FC,$A2
	.dc.b	$BE,$35,$7E,$A7,$7F,$8C,$FE,$AF,$7F,$8C,$FE,$B7,$7F,$8C,$FF,$A1
	.dc.b	$37,$F8,$CF,$FA,$1B,$7F,$8C,$FE,$17,$3F,$C4,$10,$7C,$13,$9B,$F9
	.dc.b	$CF,$37,$E4,$6B,$F7,$15,$FC,$08,$FD,$C1,$FC,$18,$FD,$C1,$FC,$28
	.dc.b	$FD,$C1,$FC,$38,$FD,$C1,$FC,$48,$FD,$C1,$FA,$D3,$F5,$E7,$E9,$FC
	.dc.b	$7F,$E4,$6F,$ED,$0F,$EE,$C7,$EA,$FB,$FF,$73,$E3,$FE,$F6,$9C,$1F
	.dc.b	$CB,$7C,$1F,$E4,$57,$D4,$FD,$77,$87,$FC,$21,$FD,$87,$89,$FC,$31
	.dc.b	$FD,$97,$8B,$FC,$41,$FD,$A7,$8D,$FC,$51,$FD,$B7,$8F,$FC,$61,$FD
	.dc.b	$99,$FC,$88,$F9,$3B,$FC,$87,$19,$FC,$87,$F7,$67,$C8,$3F,$80,$78
	.dc.b	$1F,$C0,$A3,$F9,$37,$85,$FC,$21,$FE,$51,$E2,$7F,$0D,$7F,$95,$78
	.dc.b	$BF,$C4,$9F,$7D,$FF,$45,$E2,$FE,$9A,$DF,$1F,$F8,$D3,$F2,$FC,$7F
	.dc.b	$E9,$7E,$3F,$F4,$AF,$2B,$FD,$C1,$3F,$C9,$0D,$1F,$D2,$8F,$A7,$87
	.dc.b	$69,$61,$D9,$FE,$37,$C5,$F2,$E7,$F9,$F7,$17,$CB,$7F,$A0,$71,$7C
	.dc.b	$B7,$FA,$17,$17,$CB,$7F,$A2,$71,$7C,$B7,$E3,$1F,$1F,$A3,$F2,$59
	.dc.b	$FE,$DF,$C2,$FD,$CE,$7F,$E2,$17,$FB,$EC,$7E,$E0,$FE,$06,$7E,$E0
	.dc.b	$FE,$0E,$7E,$E0,$FE,$3C,$7F,$2F,$F1,$FF,$8F,$3F,$CC,$7C,$7F,$E3
	.dc.b	$CF,$F3,$3F,$1F,$F8,$F3,$FC,$D7,$C7,$FE,$3C,$FF,$3B,$F4,$BF,$D6
	.dc.b	$EF,$F1,$E3,$F9,$19,$FB,$F3,$F8,$09,$FC,$78,$FE,$0A,$7F,$1E,$3F
	.dc.b	$84,$9F,$C7,$8F,$31,$E8,$3F,$DF,$FE,$5F,$F8,$71,$FF,$80,$3E,$5F
	.dc.b	$C0,$7F,$C3,$67,$EE,$74,$F8,$FA,$7D,$6F,$89,$FF,$08,$12,$7F,$C1
	.dc.b	$CD,$7F,$5C,$DD,$BE,$F7,$22,$31,$DD,$FE,$17,$6D,$7E,$3D,$B7,$DE
	.dc.b	$3A,$90,$EA,$7F,$8F,$FC,$8F,$E3,$F1,$FD,$89,$E3,$7D,$E7,$86,$1E
	.dc.b	$2B,$C9,$23,$38,$78,$DA,$78,$CB,$87,$8D,$F0,$93,$1F,$8B,$83,$E2
	.dc.b	$BF,$C6,$84,$9F,$F1,$97,$83,$FE,$37,$F8,$2F,$F1,$C7,$C4,$7F,$8E
	.dc.b	$7E,$0F,$FC,$75,$F1,$1F,$E3,$BF,$84,$FF,$1E,$7B,$3F,$E3,$DF,$84
	.dc.b	$FF,$1F,$7C,$37,$F8,$FF,$E1,$7F,$C8,$1F,$11,$FE,$41,$F8,$6F,$F2
	.dc.b	$17,$C3,$FF,$90,$FE,$1F,$F1,$7C,$FF,$AD,$F3,$FF,$C7,$9E,$D7,$F9
	.dc.b	$9F,$8F,$FC,$89,$5C,$1F,$F2,$36,$1D,$1F,$F2,$46,$5E,$07,$FC,$95
	.dc.b	$A7,$91,$FF,$26,$6D,$E8,$7F,$C9,$DC,$1E,$C7,$FC,$A1,$C5,$FE,$55
	.dc.b	$E3,$F8,$51,$FC,$C7,$93,$FD,$1F,$CB,$F1,$E3,$FC,$AD,$1A,$DF,$F0
	.dc.b	$F5,$F1,$E7,$FC,$B5,$CB,$F1,$DF,$F2,$E7,$2F,$C7,$7F,$CB,$DC,$BF
	.dc.b	$1D,$FF,$30,$72,$FC,$77,$FC,$C5,$CB,$F1,$DF,$F3,$27,$2F,$C7,$7F
	.dc.b	$14,$7F,$8E,$7C,$8F,$E3,$39,$FF,$6B,$8F,$8E,$70,$3E,$39,$D0,$F8
	.dc.b	$E7,$80,$F8,$E7,$90,$F8,$E7,$A0,$F8,$E7,$B0,$F8,$E7,$E1,$8F,$86
	.dc.b	$43,$FA,$09,$F9,$33,$F2,$2F,$E6,$DF,$CA,$EB,$F9,$F7,$3D,$3F,$9B
	.dc.b	$69,$C6,$78,$38,$E9,$FC,$E4,$60,$93,$C5,$B3,$C6,$3F,$D1,$DF,$11
	.dc.b	$FB,$0B,$3F,$AB,$9E,$91,$E3,$AE,$37,$DA,$FA,$12,$BA,$12,$34,$DA
	.dc.b	$E5,$3B,$9C,$9F,$B3,$EA,$FD,$8F,$5A,$E1,$27,$F7,$0D,$BF,$8C,$D9
	.dc.b	$DB,$4A,$7C,$F3,$06,$A7,$9B,$F3,$4C,$90,$C1,$0E,$45,$78,$B8,$3F
	.dc.b	$B7,$63,$F2,$CF,$E5,$4F,$33,$51,$C0,$95,$EA,$8E,$8D,$76,$D8,$DD
	.dc.b	$7B,$27,$2D,$8C,$C7,$C9,$C9,$EF,$F4,$AD,$5F,$32,$38,$25,$7C,$8B
	.dc.b	$E6,$FC,$79,$DB,$96,$B9,$9E,$2D,$34,$F9,$7A,$FF,$8C,$47,$F8,$D7
	.dc.b	$0F,$DA,$E7,$F7,$F0,$FE,$81,$3F,$D1,$4F,$90,$F8,$9B,$87,$81,$B8
	.dc.b	$78,$1A,$78,$08,$78,$C9,$7F,$11,$F9,$26,$8E,$26,$CA,$43,$E5,$1B
	.dc.b	$29,$B3,$F0,$98,$7B,$B4,$FE,$71,$3E,$E9,$FF,$3D,$1E,$D3,$89,$D4
	.dc.b	$F1,$9C,$CA,$67,$0F,$84,$F8,$23,$D5,$18,$8D,$9D,$08,$60,$8E,$C9
	.dc.b	$87,$E0,$F3,$FC,$C4,$F9,$4D,$C7,$CF,$79,$CF,$0C,$7E,$84,$62,$74
	.dc.b	$FF,$1D,$BF,$2C,$7F,$00,$8F,$E0,$53,$FC,$37,$0F,$E1,$73,$9B,$FE
	.dc.b	$06,$D2,$4E,$F2,$FE,$11,$18,$BF,$E4,$F8,$FA,$1F,$E2,$F5,$B7,$B9
	.dc.b	$FC,$B1,$C5,$F6,$6D,$73,$9F,$E6,$B3,$97,$81,$0E,$A5,$C3,$D3,$5C
	.dc.b	$8E,$B9,$7E,$07,$1F,$CE,$76,$7F,$78,$ED,$D6,$1C,$67,$D7,$78,$8E
	.dc.b	$A4,$C3,$AB,$FC,$8F,$6F,$E4,$AF,$13,$DA,$6E,$B9,$B9,$CB,$D4,$E4
	.dc.b	$DD,$7F,$63,$7F,$64,$7F,$67,$3F,$B4,$9B,$8E,$31,$E2,$8C,$67,$F1
	.dc.b	$E3,$04,$AF,$0C,$FC,$A9,$7C,$47,$F3,$AA,$FF,$24,$B6,$BF,$80,$4E
	.dc.b	$AB,$FA,$05,$7F,$A2,$F7,$FF,$4C,$C2,$31,$F1,$3C,$61,$8F,$F1,$EE
	.dc.b	$0C,$3F,$C8,$3E,$2A,$E0,$F8,$0B,$F1,$38,$F8,$2C,$FC,$4E,$BE,$13
	.dc.b	$7F,$13,$C3,$0E,$3F,$13,$CB,$E1,$F9,$FC,$4F,$6E,$2F,$EC,$FC,$1E
	.dc.b	$8E,$3F,$E3,$5E,$67,$FA,$03,$93,$FC,$CF,$D7,$C2,$B4,$7A,$33,$65
	.dc.b	$DA,$33,$C7,$FC,$E2,$7C,$71,$85,$B7,$A2,$31,$C1,$EC,$75,$C5,$F0
	.dc.b	$38,$FE,$0F,$27,$C2,$E5,$F8,$7B,$FA,$79,$A6,$1F,$8D,$F7,$7F,$99
	.dc.b	$68,$F8,$9F,$92,$49,$F1,$7A,$7F,$33,$E2,$7E,$6F,$9F,$F3,$BE,$0B
	.dc.b	$E7,$FA,$3F,$43,$E0,$FE,$9F,$F0,$0B,$E2,$6F,$C3,$63,$D9,$CF,$C3
	.dc.b	$EB,$DA,$DF,$C4,$70,$F6,$F8,$FC,$4F,$CF,$EF,$FD,$F7,$81,$FE,$CF
	.dc.b	$D1,$3B,$E3,$F9,$30,$C3,$A4,$63,$2F,$03,$AD,$3C,$BA,$F9,$E3,$E7
	.dc.b	$6F,$EB,$7E,$87,$0F,$BD,$F8,$27,$D3,$E4,$F8,$8F,$F0,$8E,$5F,$90
	.dc.b	$E4,$87,$E6,$7A,$3E,$6F,$37,$38,$C1,$F8,$3B,$F0,$38,$EF,$FE,$F3
	.dc.b	$8F,$96,$31,$B7,$CB,$7D,$47,$C0,$E4,$F9,$6F,$C2,$E4,$F9,$6F,$C4
	.dc.b	$E4,$F9,$6F,$E3,$BC,$2F,$CC,$BF,$C0,$89,$C9,$F9,$87,$E5,$F2,$7C
	.dc.b	$D7,$C3,$C9,$F3,$1F,$1F,$27,$CC,$7C,$BC,$9F,$31,$FC,$C7,$37,$E1
	.dc.b	$1F,$CD,$73,$FE,$21,$FC,$E7,$47,$ED,$B8,$FE,$5E,$7F,$C0,$79,$BF
	.dc.b	$C8,$3E,$47,$FB,$63,$9F,$E6,$E1,$F1,$5B,$5B,$8F,$AA,$67,$0E,$CE
	.dc.b	$F1,$F1,$3C,$72,$F3,$3E,$83,$E5,$1F,$87,$3E,$51,$F8,$CD,$7E,$15
	.dc.b	$FE,$47,$6F,$CA,$4F,$F2,$93,$E4,$F3,$FF,$9B,$1F,$92,$7D,$13,$91
	.dc.b	$D0,$F9,$27,$80,$F1,$1F,$7B,$A3,$F0,$15,$FC,$8F,$9B,$F0,$AF,$F2
	.dc.b	$5E,$8F,$C5,$3F,$C9,$F8,$3E,$4E,$1F,$BA,$E0,$F9,$3D,$5F,$E2,$DE
	.dc.b	$87,$B9,$FB,$42,$69,$FA,$8F,$2F,$F6,$FF,$13,$FB,$8B,$FD,$CF,$BD
	.dc.b	$FB,$AF,$88,$FD,$DF,$C0,$7E,$F3,$E2,$3F,$7B,$F0,$5F,$BE,$F8,$8F
	.dc.b	$DF,$FC,$27,$F0,$0F,$88,$FE,$03,$F0,$DF,$CD,$3F,$80,$7F,$CE,$BC
	.dc.b	$AE,$4F,$F2,$04,$3E,$27,$5F,$15,$FC,$EB,$D2,$FE,$55,$5A,$8E,$53
	.dc.b	$CF,$93,$A9,$FC,$03,$2F,$0B,$FC,$0B,$4F,$2B,$FC,$13,$6F,$4B,$FC
	.dc.b	$1B,$83,$DA,$FF,$08,$E2,$F8,$2F,$C3,$3F,$85,$F2,$FC,$67,$F7,$DC
	.dc.b	$FF,$39,$FE,$74,$FE,$65,$FC,$D9,$FC,$20,$FC,$F9,$FC,$30,$FD,$19
	.dc.b	$FC,$5B,$AF,$F9,$1B,$F8,$1E,$0F,$E3,$9E,$27,$EB,$5F,$E3,$DD,$DF
	.dc.b	$90,$E0,$FB,$5D,$DF,$90,$BF,$E0,$DC,$5F,$19,$FF,$48,$9F,$90,$3F
	.dc.b	$83,$F8,$DF,$C0,$9F,$E1,$3E,$3F,$F0,$67,$F8,$5F,$91,$E2,$7F,$20
	.dc.b	$79,$8F,$D7,$78,$3F,$90,$C7,$FA,$2F,$83,$F9,$07,$F8,$DC,$FF,$1E
	.dc.b	$D3,$F6,$4F,$C6,$31,$1F,$93,$7F,$6B,$1F,$DF,$7C,$4F,$DA,$BF,$DF
	.dc.b	$FC,$4F,$DA,$BF,$E0,$1F,$13,$F6,$AF,$F3,$FF,$1F,$F8,$73,$F1,$AF
	.dc.b	$F5,$4F,$C6,$7F,$5C,$7F,$26,$DF,$E3,$47,$F9,$CA,$3F,$6C,$FF,$1D
	.dc.b	$74,$70,$79,$3F,$18,$FE,$04,$7C,$63,$F9,$79,$F1,$8F,$E1,$47,$DD
	.dc.b	$3F,$87,$1F,$81,$3F,$89,$1F,$03,$1F,$84,$FF,$19,$E3,$FD,$AB,$F8
	.dc.b	$A3,$E3,$94,$FE,$3D,$E2,$FC,$F7,$FA,$31,$F8,$C3,$EA,$9F,$5E,$BF
	.dc.b	$B2,$72,$7D,$B7,$FB,$36,$FF,$7D,$FE,$D1,$C1,$F8,$37,$FB,$57,$0F
	.dc.b	$F4,$4B,$FE,$A1,$E4,$7F,$44,$9F,$91,$1F,$C7,$0F,$C6,$9F,$20,$FC
	.dc.b	$69,$FC,$70,$F9,$07,$F1,$C3,$E4,$1F,$C7,$1F,$2B,$E7,$CF,$FC,$1F
	.dc.b	$C5,$F8,$D8,$F6,$9F,$20,$F8,$27,$C4,$F2,$7F,$DB,$7A,$FC,$85,$E3
	.dc.b	$FF,$1C,$9F,$DB,$90,$FE,$CF,$E3,$FE,$F1,$FE,$D3,$D5,$FC,$01,$FE
	.dc.b	$D7,$D7,$FC,$11,$FE,$DB,$D9,$FC,$21,$FE,$DF,$DB,$FC,$31,$FE,$E3
	.dc.b	$DD,$FC,$41,$FE,$ED,$E0,$FF,$1C,$DB,$F8,$DF,$81,$FC,$6A,$7F,$C7
	.dc.b	$7D,$BF,$B5,$D9,$6F,$07,$F8,$DD,$7F,$90,$23,$93,$FE,$5C,$DD,$E1
	.dc.b	$7F,$2B,$8F,$F0,$EF,$87,$FC,$BD,$FF,$0F,$F8,$9F,$CC,$DF,$F1,$0F
	.dc.b	$8B,$FC,$DD,$FF,$11,$F8,$DF,$CE,$DF,$F1,$2F,$8F,$FE,$34,$E5,$FE
	.dc.b	$39,$D1,$FE,$2B,$F1,$FE,$66,$AF,$8A,$FE,$70,$4C,$3C,$DE,$CF,$F3
	.dc.b	$BF,$83,$FE,$79,$F0,$7F,$CF,$7E,$13,$F9,$F7,$C2,$7F,$3F,$F8,$5F
	.dc.b	$E8,$1F,$0B,$FD,$07,$E1,$BF,$A1,$7C,$37,$F4,$3F,$87,$FE,$89,$F0
	.dc.b	$FF,$D1,$7E,$23,$FA,$37,$C4,$7F,$47,$F8,$9F,$95,$C1,$FC,$7B,$0F
	.dc.b	$E8,$86,$39,$3F,$A5,$47,$FA,$0B,$8D,$5C,$A3,$E6,$61,$D9,$F9,$D9
	.dc.b	$78,$9F,$A1,$9F,$99,$FA,$5A,$FA,$9F,$A9,$BF,$B9,$FA,$DC,$3F,$D6
	.dc.b	$F8,$BE,$24,$7E,$0B,$8F,$E4,$FC,$4F,$F5,$F2,$47,$DD,$F8,$AF,$EC
	.dc.b	$5E,$DF,$F6,$3F,$73,$FD,$2F,$FC,$03,$FE,$9E,$E2,$F5,$F8,$68,$7B
	.dc.b	$DA,$F7,$F7,$F0,$1C,$3E,$07,$8F,$C4,$F2,$F8,$3E,$7F,$13,$D3,$E1
	.dc.b	$7A,$FC,$4F,$6F,$85,$F9,$1E,$87,$FB,$53,$3F,$F5,$3F,$C5,$1C,$F2
	.dc.b	$7F,$AB,$37,$5F,$28,$FF,$4A,$C7,$FA,$49,$B9,$78,$5C,$E9,$E5,$77
	.dc.b	$B7,$A5,$E3,$BF,$C0,$CF,$E1,$47,$E8,$78,$FF,$C0,$DC,$9F,$22,$3F
	.dc.b	$5B,$CB,$F3,$1F,$D3,$39,$E4,$F8,$FD,$EA,$F9,$B5,$FA,$BE,$6E,$CF
	.dc.b	$80,$F4,$F0,$7C,$A7,$D7,$CB,$F2,$9F,$6F,$2F,$CA,$7D,$FC,$BF,$05
	.dc.b	$FF,$3A,$71,$7C,$47,$F7,$1E,$7F,$CA,$C2,$46,$0F,$8F,$C9,$C7,$5F
	.dc.b	$BB,$7F,$26,$79,$F2,$7C,$97,$AF,$27,$C9,$7B,$F2,$7C,$97,$C3,$C9
	.dc.b	$EC,$7F,$88,$EF,$F0,$1F,$E2,$7C,$1F,$17,$93,$FC,$5B,$C7,$FB,$CE
	.dc.b	$8F,$CB,$EB,$F2,$1F,$CD,$5F,$E7,$0F,$E2,$5E,$37,$EE,$DF,$E2,$9E
	.dc.b	$37,$EE,$DF,$E2,$DE,$37,$EE,$DF,$E3,$1E,$27,$EE,$DF,$E3,$5E,$37
	.dc.b	$EE,$DF,$E3,$9E,$37,$EE,$DF,$E4,$73,$FB,$A6,$90,$FE,$57,$E3,$7E
	.dc.b	$E5,$FE,$5B,$E3,$7E,$E5,$FE,$5F,$E3,$7E,$E5,$FE,$63,$E3,$7E,$E5
	.dc.b	$FE,$67,$E3,$7E,$E5,$FE,$6B,$E3,$7E,$E5,$FE,$6F,$E3,$7C,$07,$F7
	.dc.b	$07,$F3,$1E,$0F,$8C,$D2,$1F,$CF,$BC,$6F,$E4,$0F,$C6,$3F,$82,$9F
	.dc.b	$18,$FE,$12,$7C,$63,$F8,$69,$F1,$8F,$E2,$27,$C6,$3F,$8A,$9F,$18
	.dc.b	$FE,$85,$9F,$EC,$BC,$67,$A1,$FD,$27,$C4,$FE,$91,$5A,$38,$1F,$18
	.dc.b	$E8,$7C,$63,$C0,$7C,$63,$C8,$7C,$63,$D0,$7C,$63,$D8,$7C,$63,$F9
	.dc.b	$F9,$F0,$B9,$7F,$10,$87,$C3,$BD,$7F,$C7,$C7,$ED,$88,$7E,$50,$FD
	.dc.b	$F1,$FB,$53,$F8,$5F,$4F,$F2,$2E,$CF,$CF,$5F,$E9,$BB,$7F,$44,$FE
	.dc.b	$A3,$BB,$F4,$CF,$EA,$BB,$FF,$8F,$61,$FC,$53,$C9,$FE,$3C,$FE,$C4
	.dc.b	$FF,$0C,$6C,$F2,$7F,$8F,$46,$6F,$F6,$7E,$1F,$ED,$74,$F6,$5F,$F3
	.dc.b	$1F,$1B,$F8,$15,$FF,$00,$F1,$FF,$83,$3F,$C0,$BC,$8F,$E1,$4F,$F0
	.dc.b	$4F,$1F,$D4,$FB,$0F,$71,$FE,$06,$F2,$7F,$8E,$BF,$13,$3F,$E3,$3E
	.dc.b	$2F,$F1,$D8,$FE,$3E,$67,$3F,$F5,$2F,$8D,$FB,$DA,$FE,$3F,$E3,$FF
	.dc.b	$01,$7F,$85,$F2,$7C,$8A,$F4,$F9,$0E,$5F,$93,$18,$E5,$F9,$2E,$B9
	.dc.b	$7E,$4E,$7F,$C7,$39,$7E,$4B,$FE,$1C,$F2,$7F,$8D,$BF,$2B,$A7,$F2
	.dc.b	$2F,$F2,$2E,$6F,$90,$FF,$47,$E5,$F8,$FB,$7F,$1E,$E5,$F8,$EF,$F2
	.dc.b	$0E,$5F,$8E,$FF,$22,$E5,$F8,$F7,$FC,$B3,$5F,$E8,$6F,$C7,$3F,$B4
	.dc.b	$1F,$0C,$FE,$FD,$C9,$FB,$67,$F7,$24,$AF,$E8,$7D,$9F,$B6,$8F,$F1
	.dc.b	$AF,$93,$FC,$6D,$FF,$1B,$F9,$3F,$C6,$DF,$F1,$CF,$93,$FC,$6D,$FF
	.dc.b	$1D,$F8,$FF,$C6,$DF,$F1,$EF,$93,$FC,$6D,$FF,$1F,$F9,$3F,$C6,$DF
	.dc.b	$F0,$A1,$FC,$67,$2F,$EF,$3E,$0F,$ED,$FC,$2F,$E3,$51,$FD,$5B,$C9
	.dc.b	$FE,$34,$FF,$58,$F2,$7F,$8D,$3F,$D6,$BC,$9F,$E3,$4F,$F5,$CF,$27
	.dc.b	$F8,$D3,$FD,$7B,$C9,$FE,$34,$FF,$60,$F2,$7F,$8D,$3F,$B4,$CF,$E1
	.dc.b	$F9,$1F,$B3,$5E,$37,$F0,$AE,$0F,$E6,$11,$E8,$25,$7A,$23,$2D,$8F
	.dc.b	$67,$37,$E9,$37,$FF,$2D,$12,$B7,$19,$6C,$F8,$27,$4F,$23,$11,$AB
	.dc.b	$F2,$46,$5A,$F9,$2B,$A3,$9A,$E6,$E4,$E6,$66,$BA,$41,$FF,$43,$78
	.dc.b	$4F,$8A,$FF,$17,$92,$BF,$C3,$7F,$13,$FE,$32,$F8,$AF,$F1,$9F,$B7
	.dc.b	$FE,$34,$F8,$2F,$F1,$AF,$C3,$FF,$8D,$BE,$23,$FC,$6F,$F1,$1F,$E3
	.dc.b	$8F,$80,$FF,$1C,$FC,$47,$F2,$3C,$DC,$5F,$E3,$AE,$77,$C3,$66,$31
	.dc.b	$F1,$17,$93,$F8,$7D,$3E,$1F,$18,$7D,$88,$7C,$56,$39,$D0,$F8,$8F
	.dc.b	$E8,$DE,$B7,$F1,$CB,$FD,$B7,$9D,$BA,$D4,$7F,$37,$D3,$C9,$8E,$72
	.dc.b	$D6,$9B,$D7,$D7,$7F,$E3,$ED,$FD,$EF,$F9,$03,$87,$E1,$BF,$E4,$2E
	.dc.b	$3F,$8E,$FF,$91,$39,$7F,$CB,$7C,$DF,$36,$3F,$C4,$FC,$FF,$48,$C6
	.dc.b	$BF,$E2,$27,$71,$C4,$FF,$26,$F3,$FD,$17,$FC,$9F,$CF,$F4,$5F,$F2
	.dc.b	$8F,$3F,$D1,$7F,$CA,$7C,$FE,$97,$DD,$CF,$F3,$38,$3D,$D1,$83,$E6
	.dc.b	$71,$7C,$5E,$6F,$93,$E4,$FF,$1E,$D7,$FC,$97,$CD,$9B,$E1,$D1,$F4
	.dc.b	$30,$E7,$1F,$7B,$2E,$CE,$B9,$FE,$FB,$E4,$3E,$51,$E8,$3E,$51,$EC
	.dc.b	$3E,$51,$F0,$0F,$84,$7F,$93,$B8,$7E,$3C,$FF,$3C,$E1,$FF,$70,$75
	.dc.b	$FE,$8B,$3F,$DB,$5F,$E8,$F2,$FD,$4C,$60,$97,$FA,$A9,$D5,$71,$8C
	.dc.b	$4E,$A7,$F5,$66,$BA,$FF,$56,$47,$F5,$C4,$34,$58,$E5,$7E,$2A,$EA
	.dc.b	$49,$EB,$3A,$BF,$DB,$47,$F8,$2F,$B3,$F8,$C5,$FF,$02,$8F,$D6,$35
	.dc.b	$F5,$18,$9E,$2F,$EC,$4F,$49,$9C,$30,$F0,$21,$A2,$CF,$57,$04,$70
	.dc.b	$49,$EF,$1D,$4E,$B1,$99,$EF,$1C,$67,$6F,$66,$B9,$8D,$CF,$18,$E7
	.dc.b	$3A,$8C,$47,$B5,$AF,$84,$C4,$6C,$93,$D1,$F8,$BB,$7F,$A5,$EB,$8D
	.dc.b	$78,$EB,$44,$7C,$2D,$7A,$B5,$F0,$10,$E4,$70,$21,$C0,$A6,$CF,$80
	.dc.b	$60,$E2,$6C,$86,$E3,$CE,$F0,$32,$70,$35,$5F,$01,$FF,$2A,$57,$BE
	.dc.b	$31,$1C,$0E,$47,$0A,$F6,$3C,$C8,$68,$B5,$E2,$24,$F4,$8E,$47,$AC
	.dc.b	$E0,$48,$DC,$F6,$8C,$4F,$92,$BC,$D3,$9A,$DD,$73,$6B,$E9,$6B,$E0
	.dc.b	$31,$5E,$37,$CE,$D7,$BC,$F5,$8C,$4F,$29,$DC,$E4,$FC,$05,$79,$23
	.dc.b	$F7,$98,$FE,$E6,$BF,$79,$1F,$BA,$8C,$5F,$F0,$B8,$C1,$27,$64,$8D
	.dc.b	$36,$FF,$86,$4F,$F0,$A2,$63,$FC,$16,$32,$D8,$F1,$9D,$6B,$FC,$07
	.dc.b	$8F,$F0,$87,$A9,$2B,$8E,$1E,$66,$D7,$16,$E1,$E3,$8E,$AD,$7C,$73
	.dc.b	$C5,$CD,$75,$6D,$7F,$8B,$2F,$C4,$F4,$9D,$4F,$23,$C0,$70,$21,$9B
	.dc.b	$F0,$61,$EF,$8C,$4F,$28,$C1,$30,$F2,$4F,$C9,$C3,$F9,$4E,$8F,$8A
	.dc.b	$FF,$01,$12,$7F,$AB,$7C,$4F,$F8,$17,$E2,$BF,$C0,$DE,$E7,$F8,$1F
	.dc.b	$DD,$FF,$04,$7B,$DF,$E0,$9F,$77,$FC,$15,$EF,$FF,$82,$FE,$27,$FC
	.dc.b	$19,$F0,$3F,$E0,$DF,$89,$FF,$07,$7C,$1F,$F8,$3F,$E2,$7F,$C2,$1F
	.dc.b	$0B,$FE,$11,$F8,$9F,$F0,$97,$C3,$FF,$86,$7F,$80,$7F,$9B,$73,$B9
	.dc.b	$FF,$D5,$F1,$87,$FC,$47,$3C,$79,$3F,$B6,$61,$D5,$FE,$DD,$97,$85
	.dc.b	$FE,$E1,$A7,$95,$FE,$E5,$B7,$A5,$FE,$E9,$C1,$ED,$7F,$BB,$71,$7C
	.dc.b	$17,$E1,$9F,$DE,$79,$7E,$33,$FE,$DC,$E2,$F2,$FC,$57,$A4,$91,$FE
	.dc.b	$04,$DB,$FD,$47,$BF,$AE,$3F,$CD,$F0,$F8,$9F,$81,$ED,$FC,$1F,$47
	.dc.b	$E1,$7B,$7F,$0F,$D3,$F8,$9E,$DF,$C5,$F5,$7E,$37,$B7,$F1,$FB,$FE
	.dc.b	$47,$87,$F2,$7D,$7F,$95,$ED,$FC,$BF,$67,$E6,$F0,$7E,$62,$DD,$79
	.dc.b	$8F,$35,$71,$7D,$18,$75,$7D,$58,$F8,$5F,$66,$7E,$57,$DD,$AF,$A5
	.dc.b	$F8,$1B,$FD,$BE,$0F,$B9,$E5,$FC,$49,$FF,$12,$1F,$0B,$93,$FC,$C9
	.dc.b	$E2,$FF,$7C,$E4,$46,$F9,$7E,$47,$8E,$9C,$DF,$CD,$C3,$B3,$E0,$3E
	.dc.b	$87,$37,$CA,$7E,$97,$37,$CA,$7E,$A7,$37,$CA,$7E,$B7,$27,$CA,$7E
	.dc.b	$C7,$37,$CA,$7E,$D7,$37,$ED,$F9,$3F,$D2,$99,$A3,$53,$EC,$D3,$CD
	.dc.b	$7C,$E3,$15,$FE,$74,$7E,$59,$E2,$3E,$59,$E6,$3E,$59,$EA,$3D,$87
	.dc.b	$2E,$0F,$96,$FE,$37,$7F,$99,$C1,$F2,$5F,$91,$DF,$F9,$49,$F8,$C7
	.dc.b	$E4,$EB,$F1,$EF,$53,$CB,$C7,$F6,$DF,$3F,$27,$DF,$7D,$3C,$BF,$83
	.dc.b	$7D,$7B,$7E,$1D,$F6,$F3,$7E,$2D,$FE,$27,$1F,$1B,$7F,$8B,$CB,$F6
	.dc.b	$B6,$6F,$F1,$63,$1C,$3F,$4F,$0F,$8F,$C5,$F5,$DF,$93,$C7,$F6,$DF
	.dc.b	$97,$C9,$EA,$7E,$21,$EE,$3E,$7F,$17,$C4,$E8,$FE,$13,$ED,$FF,$0D
	.dc.b	$24,$FE,$D7,$DB,$FE,$21,$EE,$7F,$11,$F8,$8F,$E2,$5F,$13,$FC,$4F
	.dc.b	$E2,$7F,$D5,$9F,$C0,$3F,$B9,$C3,$EF,$2F,$8A,$CF,$A1,$AF,$8A,$DF
	.dc.b	$C1,$F0,$F8,$AE,$3F,$0B,$CB,$E2,$B9,$FC,$3F,$ED,$33,$FC,$29,$FC
	.dc.b	$84,$C7,$47,$E8,$34,$4F,$19,$FF,$47,$C2,$FB,$46,$31,$F1,$3A,$CF
	.dc.b	$D3,$A7,$AB,$C2,$F6,$4F,$F1,$5D,$FE,$03,$FC,$5F,$87,$F9,$97,$17
	.dc.b	$C6,$8F,$CC,$F1,$FE,$DB,$93,$F1,$91,$F3,$8F,$C9,$1F,$47,$7F,$BD
	.dc.b	$E1,$7D,$57,$F0,$1C,$FF,$65,$FC,$17,$47,$DD,$7F,$09,$D3,$F8,$17
	.dc.b	$F0,$77,$FC,$48,$FD,$99,$FC,$58,$FD,$99,$FE,$AB,$F1,$FF,$BA,$F5
	.dc.b	$FF,$1C,$8C,$9F,$94,$38,$9F,$98,$3F,$7E,$7E,$C8,$FC,$79,$FB,$23
	.dc.b	$F1,$E7,$EC,$8F,$C7,$9F,$B2,$9F,$6B,$EF,$DF,$E1,$C7,$E7,$7A,$FE
	.dc.b	$34,$FF,$1D,$3E,$29,$F1,$CF,$E0,$7E,$37,$F1,$87,$F8,$2F,$8F,$FC
	.dc.b	$A5,$FD,$99,$FC,$0C,$FD,$99,$FC,$1C,$FD,$99,$EA,$3D,$87,$F0,$EE
	.dc.b	$EF,$D7,$E1,$FD,$A3,$C0,$FE,$31,$3F,$ED,$6F,$5B,$FD,$47,$97,$EC
	.dc.b	$E7,$73,$FD,$52,$FF,$69,$7F,$DC,$33,$FA,$8F,$F7,$2D,$7E,$C3,$FD
	.dc.b	$D3,$6F,$B8,$FF,$76,$DF,$F0,$0F,$F7,$8E,$0F,$E2,$F8,$FC,$1C,$FE
	.dc.b	$33,$FE,$BB,$E1,$F9,$35,$23,$E4,$1F,$18,$FE,$30,$7D,$17,$B3,$F5
	.dc.b	$67,$F1,$6F,$D9,$7F,$16,$FD,$D7,$F1,$6F,$C6,$7F,$89,$67,$F0,$3C
	.dc.b	$8F,$E9,$9A,$FC,$6A,$FF,$73,$78,$5F,$CF,$5F,$C8,$1F,$34,$FC,$A1
	.dc.b	$FC,$F3,$6F,$8A,$FF,$7C,$DB,$E2,$BF,$DF,$B6,$F8,$AF,$F8,$03,$6F
	.dc.b	$8A,$FF,$80,$B6,$F8,$BA,$7C,$48,$FC,$37,$91,$FD,$8F,$97,$F6,$53
	.dc.b	$FC,$A5,$FD,$AC,$FE,$DC,$FF,$07,$78,$7F,$C6,$1F,$F0,$87,$87,$FB
	.dc.b	$F7,$F8,$3F,$5F,$F1,$67,$F8,$4F,$87,$FC,$59,$FE,$17,$E1,$FF,$16
	.dc.b	$7F,$86,$F8,$7F,$C5,$9F,$D7,$1F,$E1,$B3,$F8,$B1,$83,$F9,$1F,$87
	.dc.b	$FC,$55,$FE,$4B,$E1,$7F,$14,$7F,$93,$F8,$5F,$C5,$1F,$E5,$3E,$17
	.dc.b	$F1,$47,$F8,$DF,$4F,$F1,$47,$FB,$3F,$85,$FC,$51,$FE,$D3,$E1,$7C
	.dc.b	$07,$F8,$99,$FE,$2D,$E0,$FD,$7F,$53,$C3,$7C,$57,$F3,$82,$57,$F2
	.dc.b	$8F,$89,$FE,$77,$EC,$FF,$3C,$F6,$BF,$9E,$FC,$17,$F3,$EF,$82,$FE
	.dc.b	$7F,$F0,$7F,$D0,$3E,$0F,$FA,$0F,$C2,$7F,$42,$F8,$4F,$E8,$7F,$0B
	.dc.b	$FD,$13,$E1,$5A,$CE,$FF,$F0,$BF,$F0,$0B,$E1,$F3,$3F,$E0,$58,$7C
	.dc.b	$3E,$2B,$7F,$11,$7A,$79,$F3,$2D,$F0,$FF,$E9,$26,$E3,$DE,$B7,$0B
	.dc.b	$E5,$2B,$E5,$8E,$1E,$99,$7A,$67,$FA,$5E,$9E,$D7,$FA,$6E,$DF,$05
	.dc.b	$FE,$9F,$C1,$F1,$5F,$EA,$3C,$5F,$25,$FE,$A7,$C9,$FD,$9B,$97,$E7
	.dc.b	$C7,$F6,$6E,$6F,$AB,$B3,$E2,$3C,$4D,$9F,$51,$E7,$70,$FC,$7F,$6F
	.dc.b	$7F,$11,$C3,$E2,$38,$FB,$DC,$BE,$23,$9F,$C0,$74,$F5,$7A,$FC,$0F
	.dc.b	$6F,$88,$EF,$F0,$7E,$0F,$88,$F0,$FC,$2F,$8B,$E2,$3C,$7F,$0F,$FD
	.dc.b	$C3,$BB,$FD,$4F,$E0,$7F,$94,$3D,$9F,$E9,$9C,$7F,$66,$B8,$3E,$8E
	.dc.b	$EF,$AD,$8F,$58,$C7,$2B,$4F,$14,$63,$6F,$33,$AE,$0F,$53,$C3,$8B
	.dc.b	$DC,$F2,$E4,$F8,$FC,$BF,$23,$9A,$4F,$F2,$93,$E4,$38,$7F,$1C,$F0
	.dc.b	$79,$1E,$4E,$6F,$92,$F9,$B9,$BE,$4B,$E8,$E6,$F9,$2F,$AB,$9B,$E4
	.dc.b	$BE,$CE,$6F,$92,$FB,$B9,$BE,$4B,$FC,$0F,$97,$FD,$A0,$FE,$47,$97
	.dc.b	$E4,$C7,$E9,$39,$7E,$4C,$FE,$8F,$93,$E4,$3F,$A5,$E3,$F8,$EF,$E9
	.dc.b	$F8,$FE,$3B,$FA,$9E,$3F,$8E,$FE,$AF,$8F,$E3,$BF,$AD,$E3,$F8,$0F
	.dc.b	$C2,$3E,$9F,$17,$F0,$6F,$1F,$F8,$2C,$7F,$8A,$BA,$3F,$2F,$AF,$F8
	.dc.b	$BF,$A7,$F8,$2C,$FF,$8B,$BC,$7F,$E0,$AF,$F8,$C3,$C7,$FE,$0A,$FF
	.dc.b	$8C,$BC,$7F,$E0,$AF,$F8,$D3,$C7,$FE,$0A,$FF,$8D,$BC,$7F,$CF,$F8
	.dc.b	$9F,$A0,$F1,$7F,$43,$E3,$7E,$8B,$C7,$FE,$57,$7F,$E4,$23,$F1,$2E
	.dc.b	$9F,$C5,$CF,$27,$A1,$FB,$83,$C0,$7E,$E0,$F2,$1F,$B8,$3D,$07,$EE
	.dc.b	$0F,$61,$FB,$83,$F9,$21,$F0,$C8,$7F,$B7,$4F,$E1,$5B,$FC,$87,$F7
	.dc.b	$86,$B6,$F9,$0F,$33,$FC,$CB,$AF,$C8,$7F,$CC,$FA,$FC,$87,$FC,$D3
	.dc.b	$AF,$C8,$7F,$CD,$7A,$FC,$87,$FC,$DB,$AF,$BD,$FE,$51,$97,$C3,$7F
	.dc.b	$95,$67,$FC,$8C,$C6,$5F,$CC,$27,$83,$FE,$69,$AE,$8F,$F9,$AE,$FC
	.dc.b	$0F,$F9,$B7,$0F,$23,$FE,$6F,$C7,$D0,$FF,$9C,$72,$FC,$1E,$5F,$E8
	.dc.b	$0D,$BF,$0F,$E1,$7F,$36,$D7,$E2,$7A,$AE,$EF,$E6,$7C,$7F,$B3,$CB
	.dc.b	$FA,$1B,$F9,$47,$FA,$2F,$27,$E6,$1F,$E8,$FC,$BF,$9C,$7F,$A4,$F3
	.dc.b	$7E,$81,$FE,$97,$CF,$FA,$47,$FA,$6F,$47,$EA,$1F,$E9,$FD,$3F,$AC
	.dc.b	$7F,$5D,$BF,$EC,$F6,$FE,$36,$7F,$4C,$EB,$FD,$9B,$FD,$23,$AF,$83
	.dc.b	$FD,$2A,$FA,$3F,$D3,$30,$F0,$3F,$D3,$B1,$F2,$3F,$D4,$32,$F4,$3F
	.dc.b	$D4,$B3,$F6,$3F,$D5,$34,$F8,$0F,$C4,$CF,$F6,$67,$F1,$9F,$0F,$FA
	.dc.b	$F6,$DF,$1E,$75,$3F,$D0,$4F,$E3,$27,$F6,$9F,$07,$F8,$CB,$FD,$AF
	.dc.b	$C1,$FE,$32,$FF,$6D,$E1,$FE,$32,$FF,$6F,$F0,$7F,$8C,$BF,$DC,$7C
	.dc.b	$1F,$E3,$2F,$F7,$3F,$07,$F8,$CB,$FE,$F9,$E3,$F8,$AF,$F6,$9F,$77
	.dc.b	$FC,$CF,$9F,$F7,$46,$C7,$F1,$DF,$81,$E3,$CD,$FE,$00,$C3,$AB,$FE
	.dc.b	$02,$C7,$C2,$FF,$81,$32,$F2,$BF,$E0,$6C,$FD,$2F,$F8,$23,$4F,$6B
	.dc.b	$FE,$0A,$D7,$FB,$0E,$5F,$D3,$3C,$4F,$8C,$FF,$40,$3E,$57,$4F,$ED
	.dc.b	$6B,$FD,$91,$E2,$7E,$5A,$7F,$73,$D7,$F9,$A9,$FC,$E4,$7F,$03,$3F
	.dc.b	$6A,$7F,$07,$3F,$6A,$7F,$0B,$3F,$6A,$7F,$0F,$3F,$6A,$7F,$5E,$F1
	.dc.b	$3F,$8E,$3F,$D8,$3C,$4F,$E3,$8F,$F9,$1C,$87,$EE,$0D,$1F,$C7,$0E
	.dc.b	$47,$F1,$C3,$FD,$DB,$E5,$7F,$84,$DF,$F7,$7F,$95,$FE,$13,$7F,$DE
	.dc.b	$3E,$57,$F8,$4D,$FF,$79,$F9,$5F,$E1,$37,$FD,$EB,$E5,$7F,$84,$DF
	.dc.b	$E2,$DC,$FF,$B7,$D7,$FC,$27,$F1,$1F,$E1,$22,$4F,$F7,$86,$BF,$E3
	.dc.b	$76,$EB,$FE,$3C,$6D,$7F,$8C,$1B,$3F,$E1,$96,$F5,$FF,$8E,$5B,$E5
	.dc.b	$9B,$E5,$FF,$19,$AF,$01,$0F,$01,$70,$F0,$37,$0F,$03,$4F,$01,$70
	.dc.b	$F0,$37,$0F,$03,$70,$F6,$F8,$BF,$C5,$E3,$FC,$61,$D7,$FE,$34,$23
	.dc.b	$FE,$1D,$E8,$FF,$2C,$AC,$99,$7F,$95,$A5,$F0,$5F,$E4,$D5,$F0,$57
	.dc.b	$DB,$C7,$C1,$67,$DD,$D7,$C4,$6F,$DF,$E1,$F1,$1C,$7E,$07,$97,$C4
	.dc.b	$73,$F1,$FA,$7C,$3F,$5F,$86,$FC,$6F,$17,$8F,$87,$F8,$EB,$FE,$49
	.dc.b	$E8,$D3,$E9,$AE,$4F,$AF,$0E,$CF,$B7,$2F,$13,$EF,$D3,$CC,$FC,$1D
	.dc.b	$BD,$4F,$C3,$E0,$F7,$71,$7B,$E3,$E3,$9F,$E7,$DE,$4F,$F2,$F9,$8E
	.dc.b	$4F,$F4,$1B,$B8,$F9,$93,$F1,$EB,$AC,$7A,$CF,$F3,$C7,$23,$0F,$1C
	.dc.b	$67,$A3,$FA,$76,$5E,$88,$C5,$33,$F6,$C6,$34,$F8,$2E,$B5,$F8,$AF
	.dc.b	$EC,$CE,$9E,$4F,$F1,$CC,$4D,$E9,$FF,$47,$CF,$6E,$2F,$A6,$FE,$17
	.dc.b	$8F,$EB,$BF,$87,$E4,$FB,$6F,$E2,$79,$7E,$FB,$F8,$BE,$6F,$C1,$BF
	.dc.b	$8D,$E7,$F8,$4F,$C4,$3F,$20,$7C,$9E,$7F,$F4,$EF,$2F,$EE,$23,$F3
	.dc.b	$44,$3F,$4E,$4E,$6F,$D3,$B6,$7F,$4E,$D3,$F4,$E5,$DB,$F4,$ED,$D7
	.dc.b	$F4,$EE,$78,$9F,$11,$FE,$36,$C3,$F4,$33,$FE,$AD,$F8,$6F,$D2,$C6
	.dc.b	$30,$FF,$01,$57,$F0,$0E,$B9,$7E,$7A,$F2,$93,$2F,$2B,$63,$CA,$DB
	.dc.b	$F2,$B7,$2F,$2B,$EE,$EB,$FD,$D4,$FA,$AB,$FD,$9E,$D7,$E0,$98,$AF
	.dc.b	$6B,$70,$FD,$EB,$A2,$1F,$18,$D1,$A9,$E8,$E4,$DB,$82,$4F,$68,$CB
	.dc.b	$5F,$DB,$1F,$EA,$9E,$3F,$C8,$1F,$8F,$EB,$FE,$63,$1E,$03,$F8,$8D
	.dc.b	$74,$BF,$3C,$78,$49,$19,$7C,$D7,$DE,$31,$1C,$A3,$04,$9E,$73,$AF
	.dc.b	$75,$E2,$FA,$27,$35,$BB,$EB,$18,$23,$82,$1D,$DC,$B4,$C9,$ED,$21
	.dc.b	$DC,$F4,$E1,$FD,$15,$FC,$E9,$A3,$DF,$3E,$37,$A1,$0D,$61,$F0,$1B
	.dc.b	$1B,$6D,$79,$30,$F7,$C7,$48,$F3,$BA,$3C,$E7,$94,$EC,$53,$46,$0E
	.dc.b	$47,$03,$EF,$9F,$C3,$B9,$CF,$E0,$5F,$E7,$7A,$7E,$E2,$39,$90,$EC
	.dc.b	$43,$A9,$DF,$0F,$73,$E5,$BF,$5C,$62,$36,$73,$29,$CD,$D1,$30,$F8
	.dc.b	$D1,$F9,$CE,$0F,$E0,$2F,$F8,$6A,$FF,$7D,$C1,$FC,$03,$7F,$EC,$5C
	.dc.b	$1F,$98,$AE,$98,$72,$7B,$93,$1E,$DA,$7C,$E8,$DB,$F1,$16,$7E,$AA
	.dc.b	$FD,$FE,$1F,$C1,$89,$3F,$C1,$9B,$7F,$C1,$9B,$3C,$27,$44,$9D,$36
	.dc.b	$35,$18,$23,$83,$39,$FC,$29,$FB,$47,$D9,$3F,$C0,$D8,$FF,$1E,$7F
	.dc.b	$95,$63,$73,$FE,$30,$DC,$7C,$AD,$9F,$13,$FC,$F5,$E2,$7F,$27,$8D
	.dc.b	$B4,$D9,$71,$F3,$CF,$35,$8F,$F5,$8B,$FE,$69,$BF,$F0,$ED,$7F,$1F
	.dc.b	$8F,$4C,$7B,$92,$BC,$EF,$5B,$F8,$58,$F5,$6C,$7F,$A6,$5A,$F1,$AE
	.dc.b	$95,$CA,$30,$7A,$0B,$8F,$A1,$FF,$1B,$6D,$FB,$F9,$E0,$4A,$E6,$49
	.dc.b	$E5,$19,$6C,$7A,$F8,$3F,$A1,$EA,$F8,$7F,$F1,$A1,$2B,$FC,$79,$EE
	.dc.b	$7F,$8D,$FE,$23,$FC,$71,$EF,$7F,$8E,$7E,$23,$FC,$75,$F0,$1F,$E3
	.dc.b	$BF,$88,$FC,$A6,$BF,$D7,$7F,$80,$7F,$B0,$E5,$FE,$13,$5F,$13,$AD
	.dc.b	$FF,$7B,$C2,$9F,$CD,$D7,$F0,$0B,$E5,$F1,$3B,$85,$6F,$E1,$70,$F1
	.dc.b	$F8,$9B,$83,$E1,$EF,$C4,$FF,$93,$FA,$FF,$BA,$69,$E0,$ED,$FD,$A6
	.dc.b	$7F,$E2,$A9,$DB,$E1,$9F,$2E,$1C,$E3,$07,$89,$63,$E0,$CF,$C3,$A7
	.dc.b	$8B,$B7,$C7,$BF,$93,$1F,$8D,$8F,$F9,$9F,$87,$E5,$3F,$E6,$9E,$3F
	.dc.b	$9C,$FD,$03,$FC,$D9,$CD,$F4,$D7,$47,$F7,$AD,$5F,$13,$F0,$09,$3E
	.dc.b	$6E,$3F,$99,$E6,$FC,$4F,$03,$E2,$F9,$DF,$1B,$E1,$3E,$3F,$A1,$F2
	.dc.b	$3E,$17,$E4,$FA,$5F,$2B,$E1,$BE,$5F,$A9,$F3,$3E,$1F,$E6,$FA,$DF
	.dc.b	$3B,$E2,$3E,$8F,$F0,$0B,$DA,$BF,$15,$F3,$F8,$BF,$80,$1F,$C0,$8F
	.dc.b	$F2,$8F,$7E,$9F,$35,$71,$7D,$1C,$AC,$3B,$C6,$32,$F1,$BA,$D3,$CE
	.dc.b	$F0,$DB,$D6,$F2,$E0,$F7,$BD,$38,$BF,$0B,$C7,$F1,$A3,$F2,$9C,$BF
	.dc.b	$CA,$F9,$BE,$63,$9F,$0F,$F3,$4F,$09,$E4,$79,$B9,$BE,$53,$E8,$E6
	.dc.b	$F9,$4F,$AB,$9B,$E5,$3E,$CE,$6F,$94,$FB,$B9,$BE,$53,$F0,$B9,$BE
	.dc.b	$23,$F1,$8F,$93,$97,$F1,$4F,$33,$E6,$47,$F9,$F7,$95,$CF,$F4,$AF
	.dc.b	$9C,$FE,$BF,$0E,$EF,$EC,$71,$F1,$BF,$B3,$CB,$CE,$FE,$D7,$3F,$5B
	.dc.b	$FB,$7D,$3F,$10,$FD,$4D,$FE,$27,$67,$E9,$37,$F9,$1E,$4D,$AF,$B3
	.dc.b	$1A,$7F,$4D,$5C,$9E,$87,$D5,$3E,$41,$F6,$4F,$90,$7D,$D3,$E4,$1F
	.dc.b	$81,$3E,$41,$F8,$6E,$7F,$80,$FE,$2B,$6F,$8B,$C3,$F1,$E3,$E4,$E3
	.dc.b	$FD,$C1,$D1,$F3,$CE,$47,$43,$F8,$B7,$3F,$E3,$9F,$E3,$1C,$FF,$8E
	.dc.b	$7F,$8D,$73,$FE,$39,$FE,$39,$CF,$F8,$E7,$F8,$F7,$3F,$E1,$DF,$C6
	.dc.b	$9F,$10,$FF,$0D,$F3,$FF,$9C,$B9,$9E,$DF,$EA,$89,$3F,$A4,$F2,$FF
	.dc.b	$5B,$F0,$7F,$AE,$F7,$BF,$5F,$F0,$9F,$B0,$F8,$0F,$D8,$FC,$2F,$EC
	.dc.b	$BE,$0B,$F6,$7F,$0D,$FB,$4E,$9F,$DA,$FC,$3F,$ED,$BE,$17,$F6,$FF
	.dc.b	$11,$FB,$8F,$87,$FD,$CF,$C4,$FE,$EB,$E2,$7F,$92,$F0,$FF,$A5,$CC
	.dc.b	$1F,$B6,$8E,$0F,$EE,$2B,$A3,$FB,$AC,$3C,$0F,$EF,$32,$F2,$3F,$BE
	.dc.b	$D3,$D0,$FF,$00,$DB,$D8,$FF,$02,$E0,$FE,$17,$C3,$F0,$A3,$F3,$DC
	.dc.b	$5F,$8E,$7F,$98,$77,$FE,$37,$C9,$FE,$90,$FC,$FE,$DF,$EA,$1D,$1F
	.dc.b	$4D,$FE,$A5,$D3,$F5,$DF,$EA,$9D,$5F,$6D,$FE,$AD,$D7,$F7,$DF,$EB
	.dc.b	$1D,$9F,$C4,$32,$FE,$B1,$DF,$F8,$E7,$F6,$1D,$FF,$8E,$EF,$FE,$B3
	.dc.b	$ED,$FF,$81,$E7,$FC,$3A,$7E,$38,$FD,$B7,$8F,$FB,$4C,$7F,$B2,$78
	.dc.b	$FF,$B4,$7F,$B3,$78,$FF,$B4,$7F,$B4,$78,$FF,$B4,$7F,$B5,$78,$FF
	.dc.b	$B4,$7F,$B6,$78,$FF,$AB,$7F,$B7,$78,$FF,$0F,$C9,$FE,$87,$9F,$E3
	.dc.b	$BC,$3F,$DA,$3F,$92,$25,$70,$7F,$8D,$9F,$B4,$3F,$36,$7E,$D0,$FC
	.dc.b	$F9,$FB,$43,$F4,$67,$ED,$0F,$D3,$9F,$A9,$7F,$C0,$1C,$1F,$1D,$FF
	.dc.b	$6F,$78,$FF,$C5,$A7,$E2,$9F,$D0,$E3,$F6,$CF,$F1,$F3,$E6,$9F,$B5
	.dc.b	$3E,$89,$D7,$83,$FB,$9F,$17,$D7,$7F,$BA,$F1,$FD,$B7,$FB,$BF,$27
	.dc.b	$DF,$7F,$BC,$F2,$FE,$0D,$FE,$F7,$CD,$F8,$77,$E2,$E3,$FC,$FB,$1F
	.dc.b	$90,$7E,$38,$FD,$C1,$F2,$23,$FC,$07,$B7,$D1,$7F,$C0,$BB,$FD,$57
	.dc.b	$FC,$0F,$C1,$F6,$5F,$F0,$4F,$0F,$DD,$7F,$C1,$7C,$5F,$81,$7F,$C1
	.dc.b	$BC,$7F,$85,$7F,$C4,$7C,$1F,$D8,$F2,$F8,$67,$C5,$CF,$E3,$92,$7F
	.dc.b	$8E,$46,$A3,$E7,$F2,$73,$8F,$F0,$B5,$F7,$7F,$C3,$18,$78,$DF,$F0
	.dc.b	$D6,$3E,$77,$FC,$39,$97,$AD,$FF,$0F,$67,$EF,$7F,$9D,$9F,$13,$1F
	.dc.b	$90,$FF,$85,$38,$FD,$3D,$3F,$B9,$9F,$C0,$BF,$BC,$9F,$DF,$1F,$E0
	.dc.b	$0F,$1F,$F8,$DB,$FE,$02,$F1,$FF,$8D,$BF,$E0,$4F,$1F,$F8,$DB,$FE
	.dc.b	$06,$C3,$F8,$DB,$FE,$08,$F1,$FF,$8D,$BF,$E0,$AF,$1F,$F8,$DB,$FE
	.dc.b	$0C,$F1,$FF,$8D,$F7,$BD,$3F,$F1,$FA,$E5,$4F,$0A,$FE,$D9,$7D,$1F
	.dc.b	$ED,$D8,$78,$1F,$EE,$18,$F9,$1F,$EE,$59,$7A,$1F,$EE,$99,$FB,$1F
	.dc.b	$EE,$DA,$7F,$3A,$7F,$9E,$1F,$E2,$0D,$7F,$8F,$79,$3F,$D0,$FE,$2B
	.dc.b	$FC,$56,$49,$FF,$12,$FB,$9F,$E2,$EF,$77,$FC,$BF,$FC,$03,$EA,$1F
	.dc.b	$15,$8F,$8A,$CF,$C0,$EB,$E2,$B7,$F0,$7C,$3E,$2B,$8F,$C2,$F2,$F8
	.dc.b	$AE,$7F,$0F,$D3,$E2,$BA,$FC,$4F,$F4,$0E,$0F,$DB,$78,$BF,$C7,$BD
	.dc.b	$9C,$78,$08,$FF,$3C,$F0,$AF,$95,$FF,$94,$61,$8F,$78,$C6,$7E,$37
	.dc.b	$5A,$F9,$DE,$1B,$FA,$DE,$5C,$3F,$01,$F8,$47,$C4,$3F,$C9,$3C,$9F
	.dc.b	$1D,$FC,$91,$FE,$86,$C5,$AF,$C8,$24,$79,$7E,$2B,$E5,$FB,$7F,$33
	.dc.b	$DC,$F9,$BF,$13,$F3,$BE,$27,$E7,$FB,$FF,$43,$E2,$7E,$8F,$C0,$FD
	.dc.b	$2F,$89,$FF,$48,$F9,$5F,$E9,$EE,$7F,$89,$AF,$FA,$5A,$3F,$D3,$EF
	.dc.b	$F9,$CD,$FF,$4E,$A8,$DF,$92,$97,$C2,$F4,$F8,$6F,$9F,$D5,$F4,$70
	.dc.b	$FF,$0D,$9F,$EA,$2F,$2F,$F3,$2F,$59,$F1,$D7,$93,$E1,$BC,$BF,$0D
	.dc.b	$E6,$C3,$D1,$7F,$57,$5F,$63,$F5,$F7,$F8,$0F,$D9,$E1,$F8,$8F,$DB
	.dc.b	$E3,$FC,$8F,$27,$CA,$8F,$56,$1F,$82,$E5,$FE,$ED,$C0,$E6,$FA,$35
	.dc.b	$F8,$39,$E1,$C5,$F0,$25,$87,$D2,$AF,$C4,$73,$7D,$27,$F1,$5C,$DF
	.dc.b	$49,$FC,$67,$37,$D2,$7F,$1D,$CD,$ED,$7E,$77,$0F,$C1,$7E,$87,$1F
	.dc.b	$CB,$7F,$82,$1F,$2B,$57,$B9,$E6,$8F,$F3,$67,$9D,$B7,$DA,$7D,$BA
	.dc.b	$E8,$E2,$BE,$C5,$F8,$63,$1A,$B1,$F3,$46,$32,$F5,$3A,$CF,$DC,$F0
	.dc.b	$D3,$E1,$3C,$B5,$F8,$FC,$1F,$36,$3F,$AA,$E7,$FC,$67,$A7,$F4,$1C
	.dc.b	$1F,$DB,$9F,$F5,$9F,$8D,$FA,$6A,$F1,$9F,$C2,$3A,$FF,$58,$F9,$BB
	.dc.b	$3F,$60,$FA,$3B,$7F,$68,$FA,$BB,$BF,$70,$FB,$3B,$FF,$7A,$FB,$BC
	.dc.b	$0F,$D0,$FC,$3F,$F2,$EF,$43,$F9,$6F,$85,$FE,$E4,$AF,$E0,$55,$C3
	.dc.b	$6F,$C0,$61,$D1,$FC,$16,$3E,$07,$F0,$99,$79,$1F,$C3,$67,$E8,$7F
	.dc.b	$11,$A7,$F0,$3C,$7F,$D9,$3E,$47,$C0,$7F,$81,$90,$FE,$07,$C1,$F1
	.dc.b	$DC,$57,$C7,$7F,$85,$1F,$EA,$B9,$E8,$FD,$AA,$F0,$70,$FF,$1E,$C3
	.dc.b	$C8,$FF,$20,$C7,$D0,$FF,$22,$CB,$D8,$FF,$24,$CF,$E0,$3F,$C9,$B4
	.dc.b	$FD,$DE,$2D,$3F,$23,$D8,$8F,$CA,$78,$7F,$BC,$7E,$71,$FC,$AF,$C4
	.dc.b	$FE,$6E,$FF,$2D,$F1,$7F,$9D,$BF,$CB,$FC,$6F,$E7,$EF,$F3,$1F,$1F
	.dc.b	$FA,$1B,$FC,$CF,$C8,$FC,$13,$FC,$D4,$FC,$37,$1F,$F8,$DB,$9F,$F1
	.dc.b	$7B,$7F,$84,$BA,$3F,$77,$DD,$FA,$5A,$FD,$C6,$DF,$94,$C3,$F9,$77
	.dc.b	$57,$E6,$27,$F7,$9D,$7F,$9C,$7F,$7D,$D9,$FA,$07,$F8,$07,$6F,$E9
	.dc.b	$1F,$E0,$5D,$DF,$A9,$7F,$91,$F6,$7F,$28,$C7,$F5,$CF,$F2,$88,$CF
	.dc.b	$81,$FB,$57,$FC,$9B,$E0,$FC,$9A,$FE,$AE,$7C,$DF,$0B,$D1,$C1,$FE
	.dc.b	$2C,$9E,$0F,$CF,$E8,$FE,$61,$87,$EF,$30,$F0,$F8,$5E,$3A,$FE,$BB
	.dc.b	$9F,$9D,$FE,$BF,$A7,$AD,$FE,$C3,$AF,$BD,$FE,$C7,$B7,$C3,$7F,$B2
	.dc.b	$EF,$F1,$F3,$FE,$49,$C1,$F3,$0C,$70,$F9,$B8,$3E,$59,$FC,$BC,$FE
	.dc.b	$46,$7F,$33,$3F,$91,$9F,$CD,$CF,$E4,$67,$F3,$B3,$F9,$19,$FD,$3E
	.dc.b	$FF,$A0,$BF,$D4,$7C,$7F,$E8,$AF,$F4,$73,$F9,$19,$F2,$2F,$F2,$6E
	.dc.b	$36,$FA,$7E,$47,$08,$FA,$75,$D1,$FA,$B7,$FD,$31,$FE,$43,$E4,$7F
	.dc.b	$4C,$7F,$91,$F9,$1F,$D3,$1F,$E4,$BE,$47,$F4,$C7,$F9,$3F,$91,$FD
	.dc.b	$31,$FE,$8C,$43,$F7,$3D,$EE,$9F,$CA,$C7,$F6,$17,$F8,$03,$FB,$93
	.dc.b	$F3,$87,$F8,$20,$FD,$01,$FE,$0C,$3F,$48,$7F,$84,$0F,$D4,$1F,$E1
	.dc.b	$43,$F5,$87,$F8,$67,$C1,$FD,$CB,$FE,$0B,$F0,$7F,$8F,$BF,$B7,$DB
	.dc.b	$FB,$CF,$8F,$FE,$4C,$3F,$90,$F9,$1F,$E2,$A9,$FD,$C1,$CC,$EA,$7F
	.dc.b	$2D,$3C,$27,$F3,$13,$CA,$7F,$35,$3D,$27,$F1,$33,$DA,$7B,$CF,$EA
	.dc.b	$58,$BB,$BE,$2F,$17,$F1,$F9,$FE,$43,$DD,$FC,$CD,$FD,$5D,$FA,$9A
	.dc.b	$7F,$85,$5C,$63,$FB,$26,$F0,$FE,$9E,$B9,$4E,$6B,$81,$0E,$64,$74
	.dc.b	$D9,$E5,$5D,$23,$F6,$DC,$BF,$B9,$AF,$F4,$E4,$7F,$8A,$67,$58,$7F
	.dc.b	$89,$5B,$87,$8B,$0E,$75,$E1,$BE,$E4,$8C,$C7,$84,$E8,$6D,$CC,$6E
	.dc.b	$32,$7A,$BC,$97,$9D,$F0,$7E,$27,$FC,$7E,$49,$FF,$1E,$7C,$4F,$F9
	.dc.b	$0B,$E0,$7F,$C8,$7E,$CF,$F9,$13,$DA,$FF,$22,$E9,$FE,$46,$F6,$FF
	.dc.b	$C8,$FE,$BF,$F9,$23,$DC,$FF,$24,$FB,$3F,$E4,$AF,$77,$FC,$97,$ED
	.dc.b	$FF,$93,$3D,$EF,$F2,$6F,$BB,$FE,$4E,$F7,$FF,$A4,$F3,$FE,$DB,$87
	.dc.b	$F7,$1D,$5F,$D1,$F9,$BF,$94,$D7,$09,$FF,$28,$61,$D1,$FF,$29,$65
	.dc.b	$E0,$7F,$CA,$9A,$79,$1F,$F2,$B6,$DE,$87,$FC,$B1,$C1,$EC,$7F,$CB
	.dc.b	$5C,$5F,$E6,$0E,$3F,$85,$1F,$CD,$B9,$3F,$C8,$3C,$67,$3A,$FF,$3C
	.dc.b	$78,$4F,$23,$FC,$C9,$C9,$F1,$DF,$F3,$35,$7C,$77,$FC,$D1,$C9,$F1
	.dc.b	$DF,$F3,$57,$27,$C7,$7F,$CD,$9C,$9F,$1D,$FF,$37,$72,$7C,$77,$F8
	.dc.b	$D7,$3F,$F9,$A5,$F8,$FE,$6E,$5F,$E9,$B3,$C5,$FE,$9F,$5D,$5F,$EA
	.dc.b	$37,$E1,$7F,$A9,$E1,$E5,$7F,$AA,$E3,$E9,$7F,$AB,$E5,$F8,$37,$F0
	.dc.b	$A7,$FA,$07,$3F,$8F,$D1,$F8,$C8,$FC,$7E,$BF,$91,$E7,$FC,$F4,$7E
	.dc.b	$54,$CE,$DF,$96,$23,$82,$47,$07,$F3,$CD,$9E,$98,$FE,$9E,$76,$D9
	.dc.b	$D5,$74,$7F,$DC,$D3,$FE,$E8,$9D,$13,$A7,$F6,$24,$8F,$34,$F2,$7A
	.dc.b	$C6,$5F,$1C,$F1,$7F,$83,$61,$F0,$6B,$D1,$1F,$DE,$5A,$FE,$E0,$C4
	.dc.b	$FF,$19,$6D,$7F,$7F,$69,$4F,$DC,$98,$35,$3D,$1C,$9D,$48,$60,$8E
	.dc.b	$C9,$3E,$18,$DC,$7C,$67,$FA,$15,$FF,$82,$FB,$3F,$A0,$C7,$8E,$78
	.dc.b	$DF,$96,$3A,$DF,$36,$C6,$67,$71,$C6,$75,$18,$8F,$5C,$60,$93,$C5
	.dc.b	$B1,$FE,$4B,$F1,$1C,$9E,$89,$CC,$F2,$AE,$04,$7C,$0D,$74,$60,$86
	.dc.b	$08,$7C,$F2,$1D,$DE,$4F,$DD,$F3,$BF,$C3,$D8,$7F,$19,$D3,$FA,$2C
	.dc.b	$F0,$3E,$19,$D8,$C4,$7C,$16,$DF,$A9,$B7,$D9,$F0,$1E,$6C,$7E,$0B
	.dc.b	$D0,$F3,$9A,$29,$93,$99,$E1,$34,$60,$DD,$78,$DF,$E1,$97,$FD,$62
	.dc.b	$31,$1C,$CE,$24,$30,$47,$64,$BF,$EE,$D5,$E8,$8C,$5F,$B6,$B8,$57
	.dc.b	$C1,$7A,$4E,$9F,$41,$A9,$E1,$1F,$97,$8F,$F6,$C6,$5F,$96,$7D,$75
	.dc.b	$CD,$E0,$72,$9E,$8E,$AB,$9C,$EE,$7F,$88,$5F,$9E,$3F,$7D,$5F,$C0
	.dc.b	$2B,$F8,$16,$1F,$C4,$9B,$9F,$F0,$FA,$DC,$E6,$BF,$85,$C6,$27,$F5
	.dc.b	$91,$FC,$4E,$B8,$BF,$C3,$8D,$B8,$C3,$F8,$91,$2B,$94,$EC,$F2,$1F
	.dc.b	$EE,$A7,$31,$B9,$E5,$18,$23,$82,$4F,$F2,$0C,$BF,$36,$7E,$72,$77
	.dc.b	$1E,$42,$1E,$02,$61,$EF,$6E,$1F,$0E,$FF,$B4,$9D,$A3,$27,$A6,$78
	.dc.b	$12,$78,$47,$68,$C3,$FE,$2C,$3F,$8A,$94,$E7,$87,$BD,$E1,$3C,$9C
	.dc.b	$DF,$A9,$B7,$FE,$47,$D3,$F9,$C7,$F0,$0B,$E0,$3F,$C7,$18,$7F,$AA
	.dc.b	$E3,$17,$E0,$C5,$18,$86,$4A,$E1,$F0,$5C,$A7,$FD,$C9,$4F,$73,$18
	.dc.b	$72,$F8,$3B,$2F,$84,$BE,$FE,$3E,$17,$3F,$03,$AF,$86,$DF,$C1,$F0
	.dc.b	$F8,$7E,$3E,$47,$2F,$88,$E7,$F0,$DD,$3E,$27,$AF,$C4,$78,$3B,$3F
	.dc.b	$CF,$9C,$78,$3D,$9C,$0A,$E3,$18,$C3,$AB,$AC,$BC,$2F,$0D,$3C,$AF
	.dc.b	$2D,$BD,$2F,$4E,$0F,$6F,$0F,$BB,$8F,$DF,$C9,$F0,$39,$E5,$87,$47
	.dc.b	$D4,$F8,$AF,$8A,$48,$F3,$F8,$BF,$23,$D2,$F9,$3E,$87,$CA,$F8,$9F
	.dc.b	$9B,$FC,$01,$FD,$AF,$BB,$EA,$60,$ED,$4B,$E2,$B7,$F0,$7C,$3E,$2B
	.dc.b	$8F,$C2,$F2,$F8,$AE,$7F,$0F,$D3,$E2,$BA,$FC,$4F,$6F,$73,$E5,$6B
	.dc.b	$FC,$48,$F9,$C6,$36,$F0,$ED,$E0,$9E,$33,$F8,$BD,$98,$76,$8C,$65
	.dc.b	$E2,$75,$A7,$99,$E1,$B7,$B9,$F8,$07,$DA,$E0,$F8,$8F,$C6,$3E,$EF
	.dc.b	$1F,$C9,$7E,$6F,$87,$F9,$D8,$C6,$BE,$6F,$01,$1F,$94,$E5,$E7,$7F
	.dc.b	$9E,$F1,$FF,$49,$8F,$82,$30,$7C,$DD,$BE,$84,$FD,$BE,$0F,$9F,$A7
	.dc.b	$A6,$B7,$1F,$4B,$5F,$74,$62,$DB,$7C,$38,$C6,$FF,$21,$D7,$07,$FA
	.dc.b	$AB,$CA,$FE,$2F,$ED,$FF,$AB,$30,$4E,$F8,$7E,$8C,$7C,$EB,$FA,$1C
	.dc.b	$9F,$75,$FA,$5C,$BF,$81,$7E,$A7,$37,$E2,$B7,$F9,$D1,$FA,$6E,$7F
	.dc.b	$C7,$BF,$A8,$E8,$F8,$CF,$CE,$21,$F3,$8F,$E0,$3C,$BF,$3B,$8F,$E8
	.dc.b	$47,$D5,$3E,$1F,$32,$BA,$C6,$2F,$C2,$FC,$5C,$3E,$56,$BE,$CE,$3F
	.dc.b	$92,$FB,$B8,$FE,$43,$F0,$38,$FE,$16,$7F,$4F,$83,$E2,$E3,$FC,$FF
	.dc.b	$8B,$F8,$9F,$1F,$F3,$FF,$2B,$F3,$93,$F9,$5E,$9F,$D0,$3F,$20,$FC
	.dc.b	$A9,$F2,$0F,$CA,$9F,$20,$FC,$A9,$F9,$D3,$F0,$D9,$7E,$37,$9F,$F1
	.dc.b	$6F,$E3,$FA,$3E,$24,$FC,$63,$F8,$17,$57,$F9,$AF,$37,$C5,$7E,$F0
	.dc.b	$93,$FA,$9F,$77,$F7,$FA,$FF,$00,$F8,$4F,$E0,$3E,$FF,$F0,$2F,$85
	.dc.b	$FE,$07,$F0,$3F,$C1,$3E,$1B,$F8,$2F,$C1,$FF,$06,$F8,$7F,$E0,$FF
	.dc.b	$0B,$FC,$23,$E2,$3F,$9A,$FF,$00,$FF,$08,$DB,$F2,$EB,$E2,$B3,$EC
	.dc.b	$FE,$F7,$AB,$FB,$DD,$7F,$45,$F4,$FF,$42,$FF,$40,$F5,$36,$FF,$49
	.dc.b	$C3,$93,$6B,$C0,$FF,$03,$CB,$C8,$FF,$05,$D3,$D0,$FF,$07,$DB,$D8
	.dc.b	$FF,$09,$E0,$F8,$0F,$F0,$BE,$2F,$E4,$3C,$7F,$1A,$3F,$59,$C9,$FC
	.dc.b	$B5,$FC,$A7,$93,$FD,$6B,$C5,$FA,$2D,$8F,$CA,$75,$7F,$03,$F1,$3F
	.dc.b	$26,$FF,$05,$F1,$3F,$26,$FE,$B3,$C4,$FC,$9F,$07,$E3,$7C,$4F,$C9
	.dc.b	$BF,$8F,$F1,$3F,$26,$FE,$47,$C4,$FC,$9E,$5F,$C1,$FC,$9F,$94,$FF
	.dc.b	$35,$E7,$FE,$AA,$FF,$54,$E7,$FC,$9B,$FC,$17,$C3,$FC,$9B,$FC,$1F
	.dc.b	$C9,$FD,$D3,$FC,$27,$C9,$FD,$D3,$FC,$2F,$C9,$FD,$D3,$F9,$33,$F7
	.dc.b	$47,$F2,$73,$F7,$47,$C0,$CF,$F6,$AF,$ED,$38,$FE,$47,$89,$FD,$4E
	.dc.b	$3F,$7B,$C3,$FC,$9E,$7F,$80,$C7,$EE,$B8,$FE,$3D,$FF,$82,$78,$FE
	.dc.b	$3B,$FB,$1E,$3F,$8F,$5F,$C9,$F8,$FE,$3B,$FC,$A7,$8F,$E3,$BF,$CA
	.dc.b	$F8,$FE,$3F,$27,$F4,$98,$FD,$C1,$FE,$13,$F2,$3F,$70,$FF,$4B,$E5
	.dc.b	$DF,$8B,$FC,$7E,$3E,$31,$FB,$83,$E3,$1F,$B8,$3E,$31,$FB,$83,$F9
	.dc.b	$91,$FB,$8C,$FF,$9F,$BF,$18,$FE,$2F,$E3,$7E,$23,$C9,$FE,$8F,$1F
	.dc.b	$B8,$F2,$BE,$4F,$17,$C5,$8F,$8F,$7F,$E2,$BB,$FE,$02,$F4,$8F,$E3
	.dc.b	$C7,$D5,$3F,$8F,$1F,$C2,$CF,$E8,$9B,$7C,$6D,$3F,$BC,$ED,$F1,$9F
	.dc.b	$EF,$7B,$7C,$0D,$7F,$A2,$76,$7F,$1C,$3F,$A2,$3F,$12,$38,$6D,$FD
	.dc.b	$3B,$1F,$E3,$47,$63,$C1,$A7,$E7,$5F,$23,$FA,$17,$F1,$07,$DF,$3F
	.dc.b	$10,$7F,$19,$3F,$10,$7C,$33,$FB,$27,$5F,$EC,$E3,$FA,$9B,$FB,$68
	.dc.b	$FD,$C1,$FE,$03,$F2,$3F,$8C,$BF,$E0,$5F,$23,$F8,$CB,$FE,$07,$F2
	.dc.b	$3F,$8C,$BF,$E0,$9F,$23,$F8,$CB,$FE,$0B,$F2,$3F,$8C,$F0,$FF,$1A
	.dc.b	$F2,$3F,$88,$CF,$F4,$EF,$03,$FC,$2A,$FF,$17,$3F,$63,$C2,$F1,$DE
	.dc.b	$47,$19,$FE,$2C,$75,$3F,$8B,$1E,$13,$F8,$B1,$E5,$3F,$8B,$1E,$93
	.dc.b	$F8,$B1,$ED,$3D,$E7,$F7,$FF,$13,$E2,$63,$FD,$57,$7F,$D7,$F4,$7F
	.dc.b	$82,$70,$7C,$5F,$F3,$42,$61,$FC,$AF,$E2,$BF,$9B,$F9,$9F,$CE,$3D
	.dc.b	$DF,$E7,$3F,$07,$FC,$EB,$E1,$3F,$9D,$FC,$07,$F3,$CF,$80,$FE,$7B
	.dc.b	$F0,$1F,$CF,$BE,$17,$F9,$FF,$C1,$7F,$40,$F8,$6F,$E8,$3F,$09,$FD
	.dc.b	$0B,$E1,$FF,$A1,$FC,$37,$F4,$4F,$88,$F9,$3C,$3F,$E3,$F3,$F9,$E3
	.dc.b	$A7,$F9,$F4,$F2,$7F,$A0,$5F,$67,$FA,$16,$3E,$27,$FA,$26,$7E,$67
	.dc.b	$FA,$36,$BE,$A7,$FA,$46,$FE,$E7,$E0,$1F,$D2,$F8,$BE,$1B,$FC,$F3
	.dc.b	$8D,$F1,$1F,$D6,$89,$1F,$D4,$7E,$27,$FD,$23,$E3,$FF,$A6,$7C,$4F
	.dc.b	$F2,$AA,$F7,$73,$EF,$6B,$DF,$DF,$C0,$70,$F8,$9E,$3F,$05,$CB,$E2
	.dc.b	$79,$FC,$27,$4F,$89,$EB,$F0,$DD,$BC,$2E,$FF,$0F,$F8,$EF,$5F,$FC
	.dc.b	$6F,$C5,$FC,$F4,$FF,$56,$70,$27,$84,$62,$FA,$3A,$C7,$C0,$F0,$CF
	.dc.b	$C8,$F2,$D7,$D0,$F4,$DF,$DF,$C1,$F0,$3E,$1F,$FB,$EF,$1F,$C4,$AF
	.dc.b	$EF,$BC,$9F,$90,$7F,$D0,$DC,$BF,$29,$FE,$F1,$1B,$E5,$FF,$3D,$CA
	.dc.b	$BA,$57,$AB,$0F,$92,$FB,$39,$7E,$4B,$EE,$E5,$F9,$2F,$C0,$E5,$F9
	.dc.b	$2F,$C2,$E5,$F9,$2F,$C4,$E5,$F9,$2F,$C7,$3F,$CD,$3C,$BF,$29,$FD
	.dc.b	$E1,$F2,$B9,$3E,$4B,$FE,$6E,$E5,$72,$FC,$A8,$C7,$2F,$CA,$75,$CB
	.dc.b	$F2,$9E,$1C,$BF,$29,$E5,$CB,$F2,$9F,$AF,$CB,$F2,$9F,$8A,$43,$F2
	.dc.b	$5C,$DF,$94,$E5,$F8,$EF,$F8,$B3,$C5,$FC,$D4,$FF,$8B,$3A,$3F,$3C
	.dc.b	$FF,$8B,$BA,$7F,$44,$FF,$8C,$3A,$BF,$4C,$FF,$8C,$BA,$FF,$54,$FF
	.dc.b	$8D,$3B,$3F,$5C,$FF,$8D,$BB,$7F,$64,$FF,$8E,$3B,$BF,$6F,$A7,$FA
	.dc.b	$CB,$E0,$BF,$C6,$BE,$5F,$F1,$F9,$DC,$FD,$CB,$E6,$FD,$EC,$3B,$BF
	.dc.b	$80,$C7,$C6,$FE,$0B,$2F,$3B,$F8,$4C,$FD,$6F,$E1,$B4,$F7,$BF,$88
	.dc.b	$D7,$E2,$79,$3F,$C7,$77,$FC,$86,$BF,$CC,$5C,$EF,$F9,$27,$E4,$3F
	.dc.b	$E4,$9F,$23,$F8,$F3,$FE,$4B,$F2,$3F,$8F,$3F,$E4,$DF,$23,$F8,$F3
	.dc.b	$FE,$4F,$F2,$3F,$8F,$3F,$E5,$1F,$23,$F8,$F3,$FE,$53,$F2,$3F,$8F
	.dc.b	$3F,$E5,$5F,$23,$F8,$F3,$FD,$14,$87,$CB,$34,$7F,$1E,$39,$1F,$C7
	.dc.b	$8E,$C7,$F1,$E3,$C4,$7F,$1E,$3C,$C7,$F1,$E3,$D4,$7F,$1E,$3D,$C7
	.dc.b	$F1,$E3,$FC,$E3,$97,$F1,$EF,$79,$E0,$FF,$45,$F2,$7F,$69,$BD,$BF
	.dc.b	$E3,$C7,$E5,$8F,$DA,$1F,$C0,$CF,$DA,$1F,$C1,$CF,$DA,$1F,$C2,$CF
	.dc.b	$DA,$1F,$C3,$CF,$DA,$1F,$C4,$CF,$DA,$1F,$AF,$3F,$69,$C3,$FE,$FE
	.dc.b	$F0,$BF,$68,$FF,$BA,$FC,$DF,$DC,$CF,$EC,$F0,$FD,$EB,$FB,$5F,$13
	.dc.b	$F8,$0B,$FB,$7F,$17,$F8,$2B,$FB,$9F,$1B,$F8,$4B,$FB,$BF,$1F,$F8
	.dc.b	$6B,$FB,$DF,$23,$F8,$8B,$FE,$96,$F2,$7E,$1B,$FC,$6B,$6F,$C6,$F9
	.dc.b	$5F,$C6,$E3,$F8,$F9,$FB,$7D,$7F,$8D,$D7,$27,$FA,$1D,$F6,$7F,$A2
	.dc.b	$E1,$E2,$7F,$A3,$E3,$E6,$7F,$A4,$E5,$EA,$7F,$A5,$E7,$EE,$7F,$A6
	.dc.b	$E9,$F0,$9F,$89,$97,$F9,$67,$C9,$FF,$1B,$3F,$2B,$80,$BA,$93,$FD
	.dc.b	$16,$3E,$89,$FC,$77,$C9,$FE,$39,$7F,$D6,$BC,$9F,$E3,$8F,$F5,$CF
	.dc.b	$27,$F8,$E3,$FD,$7B,$C9,$FE,$38,$FF,$60,$F2,$7F,$8E,$61,$FC,$9F
	.dc.b	$C9,$FE,$38,$FC,$62,$74,$7E,$41,$FD,$A1,$FD,$C4,$87,$FB,$1B,$C1
	.dc.b	$FD,$FE,$DF,$E1,$9F,$0B,$F8,$1B,$FE,$1B,$F0,$FF,$83,$BF,$E1,$DF
	.dc.b	$13,$F8,$5B,$FE,$1F,$F1,$7F,$87,$BF,$E2,$1F,$1B,$F8,$9B,$FE,$23
	.dc.b	$F1,$FF,$8B,$BF,$E2,$5F,$23,$F8,$EB,$FE,$D2,$CF,$F6,$C7,$FB,$63
	.dc.b	$C2,$FD,$EB,$FE,$DA,$F0,$FF,$80,$BF,$ED,$CF,$13,$F8,$2B,$FE,$DE
	.dc.b	$F1,$7F,$84,$BF,$EE,$0F,$1B,$F8,$6B,$FE,$E2,$F1,$FF,$88,$BF,$EE
	.dc.b	$4F,$23,$F8,$AB,$FB,$63,$E1,$F1,$FC,$7E,$CF,$B3,$F1,$3F,$E1,$02
	.dc.b	$4F,$F8,$01,$AF,$F5,$76,$F4,$7F,$85,$1B,$B7,$E1,$2B,$9D,$71,$21
	.dc.b	$C4,$B5,$C5,$B3,$C7,$B5,$7F,$C7,$A7,$FC,$74,$4A,$FE,$FB,$C1,$E0
	.dc.b	$8F,$C9,$57,$07,$99,$8A,$D6,$1E,$48,$F1,$12,$FC,$4D,$C3,$C5,$5C
	.dc.b	$2F,$A4,$E4,$E8,$7B,$31,$F9,$50,$F8,$9F,$F1,$C9,$27,$FC,$53,$E4
	.dc.b	$7F,$8F,$3E,$03,$FC,$7B,$EE,$7F,$8F,$BE,$03,$FC,$7F,$F1,$3F,$E4
	.dc.b	$0F,$81,$FD,$6E,$0D,$9B,$7C,$DF,$E0,$1F,$EC,$A3,$E0,$B7,$3A,$AF
	.dc.b	$DA,$4B,$E0,$F3,$7C,$7E,$0A,$D6,$BE,$12,$F2,$30,$FF,$58,$4F,$86
	.dc.b	$FC,$1F,$0B,$A9,$FF,$6C,$65,$F4,$AB,$F8,$31,$0F,$CE,$BE,$3B,$E9
	.dc.b	$F0,$5C,$6F,$97,$C2,$73,$CB,$A6,$7D,$74,$ED,$AF,$7C,$FE,$21,$FE
	.dc.b	$52,$DF,$E4,$3F,$E5,$4E,$1F,$98,$FC,$E3,$FC,$AF,$C9,$F4,$57,$37
	.dc.b	$D6,$8C,$6F,$FE,$6C,$7F,$CD,$FC,$BC,$AB,$FC,$69,$7D,$9F,$F1,$B6
	.dc.b	$1E,$27,$E2,$1F,$38,$F3,$9F,$C9,$2F,$E0,$70,$FD,$18,$C7,$17,$D1
	.dc.b	$69,$F4,$4B,$C5,$F4,$78,$3E,$34,$62,$BE,$7D,$7F,$0B,$8F,$9D,$7F
	.dc.b	$D6,$F9,$BF,$4C,$F8,$38,$F8,$CF,$9B,$1E,$8F,$3E,$2F,$9E,$F5,$E1
	.dc.b	$F4,$BB,$CB,$FC,$63,$8B,$93,$F2,$B3,$FE,$82,$E5,$FC,$CB,$FE,$84
	.dc.b	$E6,$F9,$2F,$D0,$21,$F4,$3C,$DF,$D5,$BF,$B4,$CF,$F6,$5A,$FE,$C0
	.dc.b	$9B,$6A,$7F,$5D,$3A,$8C,$4E,$A7,$F6,$26,$B2,$FD,$AC,$66,$B8,$4E
	.dc.b	$A3,$19,$BE,$1B,$F9,$9C,$FF,$59,$F6,$7F,$57,$1F,$BF,$EA,$93,$E7
	.dc.b	$C7,$AE,$5E,$22,$47,$56,$E5,$E0,$24,$F0,$9D,$12,$3F,$89,$73,$FF
	.dc.b	$08,$BF,$DF,$CF,$F0,$09,$FE,$2C,$DA,$FE,$22,$D9,$DB,$F1,$CC,$18
	.dc.b	$9E,$2E,$48,$7C,$83,$67,$81,$DB,$93,$6E,$09,$3E,$28,$FE,$25,$87
	.dc.b	$CA,$96,$3F,$2A,$7F,$9F,$9D,$88,$74,$AE,$4F,$72,$4F,$6B,$F6,$C6
	.dc.b	$27,$73,$98,$F7,$46,$09,$3F,$CA,$3D,$D7,$8F,$DE,$39,$DF,$1B,$F8
	.dc.b	$17,$C1,$E7,$3A,$8C,$46,$A3,$04,$76,$43,$45,$3D,$E7,$FB,$5B,$B7
	.dc.b	$FC,$52,$FF,$B1,$27,$F6,$FA,$79,$63,$81,$4E,$64,$39,$DF,$A5,$B3
	.dc.b	$E6,$6D,$F9,$56,$3F,$09,$F9,$51,$97,$67,$80,$C9,$6B,$B3,$E3,$36
	.dc.b	$78,$CF,$C2,$FA,$EC,$7D,$BB,$FF,$13,$9F,$E3,$73,$97,$F0,$67,$32
	.dc.b	$1E,$03,$A1,$4E,$E4,$3E,$26,$1E,$BB,$F2,$C6,$CE,$6E,$09,$8F,$C5
	.dc.b	$8C,$B5,$FC,$7E,$FF,$C0,$DF,$F5,$4E,$1F,$ED,$17,$FA,$F6,$DF,$E2
	.dc.b	$07,$C4,$77,$25,$F6,$CF,$E6,$47,$66,$BF,$C5,$B1,$F4,$63,$FC,$0E
	.dc.b	$BF,$82,$D7,$F0,$EA,$FE,$34,$DC,$3F,$8D,$56,$C9,$19,$D3,$F8,$7C
	.dc.b	$73,$CB,$D5,$7F,$E2,$59,$FF,$27,$5F,$F5,$2C,$B8,$4F,$F1,$1B,$E6
	.dc.b	$FF,$1D,$C3,$38,$F9,$23,$04,$3F,$C4,$FB,$FF,$17,$71,$87,$9A,$B8
	.dc.b	$DF,$49,$F5,$9A,$C7,$CF,$3D,$B4,$FC,$24,$FE,$8C,$FE,$B7,$3F,$D7
	.dc.b	$5E,$24,$BE,$CF,$33,$D5,$5E,$02,$3F,$E9,$4C,$3C,$CF,$1A,$F0,$BC
	.dc.b	$71,$F7,$39,$39,$99,$C7,$D5,$87,$F8,$76,$3C,$C4,$70,$79,$1E,$53
	.dc.b	$D2,$30,$4A,$F0,$CE,$AB,$E3,$5F,$F9,$C7,$37,$C2,$FF,$8B,$49,$5F
	.dc.b	$E0,$1F,$77,$FC,$61,$EF,$7F,$8C,$70,$FF,$19,$7C,$27,$F8,$CF,$E1
	.dc.b	$FF,$C6,$9F,$0B,$FE,$35,$F8,$8F,$F1,$B7,$C3,$7F,$8D,$FE,$27,$FC
	.dc.b	$71,$F0,$FF,$E3,$9F,$8A,$FF,$1D,$7C,$47,$F9,$03,$F8,$07,$FC,$91
	.dc.b	$BF,$F8,$1D,$7C,$4F,$E8,$A3,$F4,$BC,$9F,$8C,$3F,$A0,$BF,$E1,$89
	.dc.b	$DB,$63,$AB,$FE,$19,$BF,$0B,$FE,$1B,$C7,$CA,$FF,$87,$73,$F4,$BF
	.dc.b	$E1,$FD,$7D,$AF,$F8,$87,$7F,$82,$FC,$33,$FC,$49,$C5,$F1,$9F,$F0
	.dc.b	$71,$FE,$D3,$CB,$F9,$0B,$E5,$EA,$F7,$1E,$FF,$88,$F8,$1E,$D7,$C1
	.dc.b	$F3,$FE,$17,$B5,$F0,$FD,$1F,$89,$ED,$7C,$5F,$4F,$E3,$78,$7F,$1F
	.dc.b	$D4,$F9,$1E,$D7,$C9,$F5,$BE,$57,$B5,$F2,$FD,$8F,$99,$ED,$7C,$9E
	.dc.b	$1F,$DA,$1F,$E0,$F3,$1E,$27,$F9,$97,$8D,$3C,$67,$E6,$5F,$57,$E7
	.dc.b	$63,$E1,$7E,$86,$7E,$57,$E9,$6B,$E9,$7E,$A6,$FE,$D7,$EB,$70,$FC
	.dc.b	$17,$E2,$10,$F9,$3E,$2B,$3F,$93,$1F,$83,$E4,$F9,$2F,$DF,$E4,$F9
	.dc.b	$2F,$E0,$79,$3E,$4B,$F8,$3E,$4F,$92,$FE,$17,$93,$E4,$BF,$87,$E4
	.dc.b	$F9,$2F,$E2,$79,$3E,$4B,$F8,$BE,$4F,$92,$FF,$98,$7C,$8F,$E9,$BD
	.dc.b	$9F,$EC,$DF,$88,$74,$FC,$A8,$FF,$68,$6F,$F2,$AF,$91,$F2,$8E,$C7
	.dc.b	$CA,$3C,$47,$CA,$3C,$C7,$CA,$3D,$47,$CA,$3D,$C7,$CA,$3E,$11,$F2
	.dc.b	$7A,$3E,$D7,$07,$FA,$E3,$C7,$FF,$76,$57,$CA,$9C,$D6,$EF,$CF,$C3
	.dc.b	$F2,$A3,$F6,$1C,$3F,$29,$FD,$97,$0F,$CA,$7F,$69,$C3,$F2,$9F,$DB
	.dc.b	$70,$FC,$A7,$F7,$1C,$3F,$29,$FD,$D7,$0F,$CA,$7F,$CF,$9C,$3F,$29
	.dc.b	$71,$38,$BE,$5C,$7C,$F3,$F7,$9D,$3F,$93,$7F,$82,$74,$FE,$4D,$FE
	.dc.b	$0D,$D3,$F9,$37,$F8,$47,$4F,$E4,$DF,$E1,$5D,$3F,$93,$7F,$86,$74
	.dc.b	$FE,$4D,$FE,$1D,$D3,$F9,$37,$F1,$C7,$E4,$F9,$FE,$4F,$B7,$FC,$54
	.dc.b	$91,$FC,$3F,$CB,$FE,$31,$F1,$1F,$C6,$7E,$23,$FD,$B3,$FC,$02,$F7
	.dc.b	$EF,$C4,$E3,$E0,$73,$F1,$3A,$F8,$3D,$FC,$4F,$0F,$85,$E3,$F1,$3C
	.dc.b	$BE,$1F,$9F,$C4,$F4,$F8,$9E,$BE,$DF,$F9,$7F,$DB,$FF,$3B,$F9,$1F
	.dc.b	$ED,$DE,$04,$EF,$93,$FD,$77,$0C,$3A,$46,$32,$F0,$3A,$D3,$C8,$F0
	.dc.b	$DB,$D0,$F2,$E0,$F6,$F0,$FB,$E3,$F9,$87,$1F,$F3,$9E,$4F,$8B,$1F
	.dc.b	$CE,$79,$7F,$25,$AF,$E9,$3C,$AF,$DC,$CF,$CE,$3F,$9F,$79,$1F,$91
	.dc.b	$E0,$FE,$25,$E4,$7E,$45,$FE,$29,$E4,$7E,$45,$FE,$2D,$E4,$7E,$45
	.dc.b	$FE,$31,$E4,$7E,$45,$FE,$35,$E4,$7E,$45,$FE,$39,$E4,$7E,$45,$FF
	.dc.b	$09,$65,$FD,$28,$FE,$44,$7E,$3C,$FE,$A1,$E2,$FE,$3E,$3F,$92,$78
	.dc.b	$BF,$8F,$7F,$93,$78,$BF,$8F,$7F,$94,$78,$BF,$8F,$7F,$95,$78,$BF
	.dc.b	$8F,$7F,$96,$78,$BF,$8F,$7F,$97,$78,$BF,$8F,$7F,$67,$97,$C3,$F1
	.dc.b	$7F,$6F,$C7,$F1,$A7,$27,$F3,$5E,$3F,$8D,$E4,$FF,$0E,$E4,$F8,$CF
	.dc.b	$F1,$0E,$4F,$8C,$FF,$12,$E4,$F8,$CF,$F1,$4E,$4F,$8C,$FF,$16,$E4
	.dc.b	$F8,$CF,$F1,$8E,$4F,$8C,$FF,$6C,$D7,$FC,$1C,$FF,$1F,$F2,$FF,$8F
	.dc.b	$47,$F5,$DE,$5F,$8D,$3F,$C8,$39,$7E,$33,$FC,$8B,$97,$E3,$3F,$C9
	.dc.b	$39,$7E,$33,$FC,$9B,$97,$E3,$3F,$CA,$39,$7E,$33,$FC,$AB,$97,$E3
	.dc.b	$3F,$89,$CB,$F6,$9C,$D5,$FC,$57,$27,$C6,$9F,$E3,$71,$FC,$94,$FE
	.dc.b	$36,$7F,$29,$3F,$8D,$9F,$CB,$4F,$E3,$67,$F3,$13,$F8,$D9,$FC,$D4
	.dc.b	$FE,$36,$7F,$39,$3F,$8D,$9F,$C5,$0F,$C4,$47,$FA,$3D,$F8,$E7,$F1
	.dc.b	$CF,$67,$FC,$C7,$1F,$DB,$75,$F8,$F3,$FE,$05,$D7,$E3,$BF,$E0,$7D
	.dc.b	$7E,$3B,$FE,$09,$D7,$E3,$BF,$E0,$BD,$7E,$3B,$FE,$0D,$D7,$E3,$BF
	.dc.b	$E0,$FD,$7E,$3B,$FE,$11,$E6,$FD,$7F,$3F,$F1,$A8,$FF,$13,$74,$7E
	.dc.b	$DB,$5F,$F0,$07,$4F,$F1,$A9,$FF,$15,$79,$1F,$BE,$7F,$80,$1F,$C2
	.dc.b	$BC,$8F,$E3,$2F,$F0,$CF,$23,$F8,$CB,$FC,$3B,$C8,$FE,$32,$FF,$10
	.dc.b	$F2,$3F,$8C,$BF,$C4,$BC,$8F,$E3,$2F,$EC,$7C,$AF,$F7,$BB,$81,$9F
	.dc.b	$F1,$B9,$E3,$18,$E7,$3F,$D2,$3C,$86,$1E,$08,$D1,$9C,$7C,$AE,$31
	.dc.b	$53,$2F,$5C,$E6,$19,$FF,$38,$8F,$E5,$3E,$1F,$F1,$F7,$FC,$A9,$E1
	.dc.b	$FF,$1F,$7E,$59,$0F,$F0,$CD,$3E,$2B,$FA,$19,$23,$F9,$9F,$B7,$FD
	.dc.b	$1B,$E1,$FF,$A3,$FC,$3F,$F4,$8F,$7F,$FA,$4F,$C3,$FF,$4A,$F8,$1F
	.dc.b	$E9,$7F,$0F,$FD,$33,$E0,$FE,$0E,$5E,$BC,$FF,$A0,$F2,$7C,$4E,$2F
	.dc.b	$07,$C4,$6B,$9D,$39,$A7,$C2,$E6,$31,$F1,$37,$1F,$EE,$74,$F8,$7C
	.dc.b	$65,$EB,$F8,$9F,$AD,$97,$FB,$56,$FE,$D7,$27,$F5,$07,$EA,$C7,$81
	.dc.b	$E7,$3C,$B0,$E9,$1A,$CB,$C2,$D3,$C3,$9E,$F4,$E1,$1F,$04,$FE,$B7
	.dc.b	$B7,$C5,$7F,$AE,$F0,$7C,$97,$FA,$FF,$17,$CD,$7E,$79,$FD,$8B,$1F
	.dc.b	$A4,$FF,$93,$97,$C5,$79,$49,$1E,$3D,$BE,$D7,$C5,$7A,$7D,$BF,$57
	.dc.b	$BB,$EB,$F8,$9F,$67,$C4,$FB,$7D,$0F,$77,$C4,$FB,$FE,$07,$E0,$7C
	.dc.b	$4F,$C1,$F8,$3F,$85,$F1,$3F,$0F,$E1,$7E,$27,$C4,$FC,$5F,$87,$F9
	.dc.b	$BD,$3F,$CB,$F8,$BF,$DA,$67,$E7,$FD,$8D,$3E,$5E,$55,$F3,$8C,$63
	.dc.b	$DD,$D6,$7E,$37,$86,$BE,$77,$96,$FE,$B7,$A7,$0F,$BD,$ED,$C7,$F1
	.dc.b	$1F,$C8,$72,$FF,$7C,$7F,$0E,$E7,$B3,$F0,$1C,$DF,$A1,$8F,$C0,$57
	.dc.b	$3A,$FC,$C5,$F7,$7F,$35,$87,$8D,$FC,$E6,$3E,$77,$F3,$D9,$7A,$DF
	.dc.b	$D0,$67,$EF,$7F,$45,$A7,$C4,$7F,$75,$B7,$F9,$37,$4F,$F2,$67,$27
	.dc.b	$CC,$9F,$CE,$F0,$FD,$07,$E5,$71,$7D,$47,$E6,$71,$FD,$87,$E7,$72
	.dc.b	$7D,$C7,$E8,$5F,$E0,$1F,$A5,$CB,$F8,$47,$EA,$73,$7C,$07,$EB,$71
	.dc.b	$FC,$9F,$3F,$F4,$DA,$7F,$12,$E6,$FC,$9F,$37,$F8,$E5,$FF,$24,$F8
	.dc.b	$BF,$99,$C3,$F8,$A7,$4F,$E7,$5F,$E2,$DD,$5F,$A1,$7F,$8C,$75,$FE
	.dc.b	$95,$FE,$08,$7E,$A4,$FC,$37,$6F,$EC,$7C,$1F,$D9,$78,$3F,$B3,$F0
	.dc.b	$3F,$C9,$55,$0F,$C6,$3A,$7F,$1D,$3F,$B9,$AE,$91,$D8,$FF,$26,$F8
	.dc.b	$5F,$C5,$9F,$F2,$7F,$87,$FC,$69,$FF,$28,$F8,$9F,$C7,$9F,$F2,$9F
	.dc.b	$8B,$EE,$7F,$68,$43,$F9,$4E,$FF,$13,$C3,$FE,$39,$E2,$FF,$2B,$9F
	.dc.b	$DD,$9F,$CD,$EB,$A3,$FC,$E6,$FC,$0F,$F3,$BC,$3C,$8F,$F3,$DC,$7F
	.dc.b	$85,$BF,$A9,$F1,$7D,$8F,$B8,$FD,$66,$DF,$15,$FF,$34,$F7,$7C,$57
	.dc.b	$F1,$A7,$C6,$36,$71,$3F,$9A,$6B,$F1,$B9,$3F,$47,$B7,$F3,$69,$FE
	.dc.b	$B1,$8F,$92,$7F,$CF,$19,$7A,$1F,$F3,$D6,$7E,$C7,$FC,$F9,$A7,$C7
	.dc.b	$BF,$EA,$D8,$7F,$1A,$E6,$FC,$5B,$FC,$6A,$3F,$1E,$43,$FA,$37,$8B
	.dc.b	$FB,$1B,$F1,$F8,$BF,$B1,$7C,$BE,$2F,$EC,$5F,$3F,$8B,$FB,$17,$D3
	.dc.b	$E2,$FE,$C5,$F5,$F8,$BF,$B1,$7D,$BE,$2F,$EC,$5F,$F4,$97,$17,$EC
	.dc.b	$7E,$1B,$FD,$77,$A7,$C8,$3F,$AC,$78,$DF,$B1,$7F,$AD,$78,$DF,$B1
	.dc.b	$7F,$AE,$78,$DF,$B1,$7F,$AF,$78,$DF,$B1,$7F,$B0,$78,$DF,$B1,$7F
	.dc.b	$B1,$78,$DF,$B1,$7F,$B2,$78,$DF,$B1,$7F,$8B,$9F,$0F,$97,$E3,$F8
	.dc.b	$DF,$C5,$FB,$FF,$B2,$3B,$DF,$EB,$DF,$37,$EC,$E1,$DD,$FB,$78,$F8
	.dc.b	$DF,$BB,$97,$9D,$FB,$F9,$FA,$DF,$C0,$E9,$EF,$7F,$07,$AF,$F1,$3F
	.dc.b	$1B,$F8,$AE,$DF,$CF,$9F,$F0,$AF,$07,$CB,$DB,$FB,$C7,$0F,$CF,$7F
	.dc.b	$BD,$71,$7D,$37,$FB,$E7,$1F,$D7,$7F,$BF,$72,$7D,$B7,$FC,$01,$CB
	.dc.b	$F7,$DF,$F0,$17,$37,$E0,$DF,$F0,$27,$3F,$C3,$D3,$F9,$D9,$FE,$2E
	.dc.b	$E6,$FE,$2F,$E0,$7E,$C6,$3F,$22,$7E,$DF,$83,$FD,$A1,$E3,$7F,$17
	.dc.b	$7F,$DA,$5E,$37,$F1,$77,$FD,$A9,$E3,$7F,$17,$7F,$DA,$DE,$37,$F1
	.dc.b	$77,$F4,$B7,$FA,$7C,$FF,$C2,$DE,$17,$F1,$78,$FD,$69,$FE,$1F,$F1
	.dc.b	$BF,$65,$5F,$E1,$18,$FF,$1E,$7C,$37,$EB,$79,$BF,$C2,$D1,$C2,$7F
	.dc.b	$63,$E3,$FE,$FA,$3F,$C5,$1E,$3F,$F0,$27,$F8,$21,$FE,$0C,$3F,$8C
	.dc.b	$9F,$E0,$C3,$F8,$C9,$FE,$0C,$3F,$8C,$9F,$C4,$A7,$F9,$C3,$FC,$67
	.dc.b	$83,$F9,$EE,$1E,$32,$47,$A3,$3F,$F1,$B4,$62,$3F,$CC,$4D,$7D,$66
	.dc.b	$CF,$F6,$A9,$83,$66,$71,$FF,$34,$B8,$21,$B2,$5F,$C7,$25,$78,$A3
	.dc.b	$95,$78,$67,$31,$E0,$6D,$76,$8F,$F1,$47,$C4,$FF,$8E,$C8,$FF,$89
	.dc.b	$FF,$80,$3F,$C8,$0F,$F9,$BD,$7C,$57,$F9,$29,$7A,$B7,$E1,$31,$EB
	.dc.b	$E7,$E1,$35,$EC,$EF,$E1,$38,$7B,$7C,$7E,$13,$97,$BB,$CF,$E1,$3A
	.dc.b	$7B,$FD,$7E,$13,$F5,$9E,$87,$F1,$0E,$4F,$F5,$A1,$FC,$53,$8B,$CD
	.dc.b	$3C,$1F,$45,$F4,$7D,$58,$F8,$1F,$66,$7E,$47,$DD,$AF,$A1,$F8,$1B
	.dc.b	$FB,$78,$3D,$DC,$5E,$FE,$6F,$8B,$5F,$E7,$7E,$4F,$F1,$47,$2F,$FA
	.dc.b	$15,$CE,$DF,$D1,$9E,$13,$F2,$EF,$A3,$D8,$F9,$DC,$9F,$1D,$FA,$1C
	.dc.b	$9F,$1D,$FA,$5C,$9F,$1D,$FA,$9C,$9F,$1D,$FA,$DC,$9F,$1D,$FE,$BA
	.dc.b	$7F,$11,$E4,$F9,$1A,$FF,$70,$F2,$35,$3F,$6E,$B9,$3F,$76,$FB,$3F
	.dc.b	$7F,$0F,$13,$F8,$1C,$7C,$CF,$E0,$F2,$FC,$03,$F8,$7D,$FE,$3B,$F8
	.dc.b	$9C,$BE,$19,$FE,$97,$D7,$E3,$F3,$FE,$56,$3F,$31,$3F,$9E,$DF,$F2
	.dc.b	$B3,$98,$FC,$C4,$E9,$FD,$0B,$67,$71,$FA,$02,$1C,$E3,$44,$79,$93
	.dc.b	$83,$F4,$53,$CA,$38,$19,$AF,$14,$EA,$B9,$3F,$A9,$E8,$FE,$0B,$5F
	.dc.b	$EB,$A3,$FD,$77,$1F,$AD,$6D,$65,$A5,$3F,$81,$98,$35,$1C,$5F,$80
	.dc.b	$64,$EC,$E5,$E2,$68,$86,$08,$EC,$91,$E1,$8C,$B6,$33,$1E,$56,$BE
	.dc.b	$48,$D1,$1D,$35,$D1,$82,$18,$21,$EA,$78,$34,$FF,$71,$E2,$D7,$D7
	.dc.b	$5E,$29,$D1,$C8,$F8,$84,$37,$5D,$1C,$1E,$E3,$C6,$72,$21,$A8,$EC
	.dc.b	$D8,$E4,$F4,$38,$99,$29,$93,$C8,$53,$A9,$4F,$B5,$5E,$D7,$C8,$70
	.dc.b	$29,$C0,$A7,$73,$53,$E6,$7A,$10,$D9,$5D,$38,$23,$82,$4F,$BA,$32
	.dc.b	$FD,$49,$F3,$4E,$A7,$DE,$68,$AE,$5F,$2C,$F3,$8C,$47,$AD,$AF,$AC
	.dc.b	$C4,$6C,$8F,$46,$BD,$0C,$1F,$86,$AF,$D9,$C7,$ED,$89,$1F,$C0,$8F
	.dc.b	$E0,$E7,$EE,$CC,$63,$FB,$F7,$35,$FB,$D7,$47,$F0,$27,$04,$9E,$B1
	.dc.b	$96,$BC,$9B,$87,$F0,$48,$D1,$E0,$C7,$CE,$D3,$B9,$C4,$C5,$77,$6D
	.dc.b	$76,$BF,$41,$2B,$9C,$F6,$07,$F8,$1E,$FC,$4F,$53,$45,$24,$F3,$7C
	.dc.b	$6D,$BF,$14,$70,$8C,$5F,$8E,$FC,$33,$C9,$B3,$A8,$FF,$57,$5F,$FB
	.dc.b	$9B,$A1,$F0,$3F,$E0,$92,$4F,$F8,$3F,$E1,$3F,$C1,$9F,$07,$FE,$0D
	.dc.b	$F8,$3F,$F0,$77,$C1,$FF,$83,$FD,$CF,$F0,$87,$C1,$FB,$B0,$68,$D7
	.dc.b	$E0,$EE,$FE,$01,$7B,$FB,$AD,$5A,$B5,$F0,$D9,$C3,$8E,$0F,$81,$C5
	.dc.b	$EE,$1F,$0F,$8C,$3B,$7C,$1D,$97,$C4,$5F,$85,$C7,$C4,$E7,$E1,$F5
	.dc.b	$F1,$53,$3F,$AF,$BF,$93,$3F,$45,$70,$9C,$ED,$E4,$C7,$B6,$5D,$F4
	.dc.b	$F0,$6B,$E1,$DB,$C5,$BF,$8F,$83,$E3,$E5,$FE,$3B,$E2,$F9,$6F,$F8
	.dc.b	$F7,$93,$E7,$BF,$44,$FF,$20,$73,$FD,$45,$D2,$F0,$BE,$2F,$C4,$FB
	.dc.b	$C9,$3E,$63,$E3,$47,$C3,$F8,$4F,$89,$EA,$7C,$5F,$85,$F8,$DE,$AF
	.dc.b	$C7,$F8,$6F,$91,$EB,$7C,$9F,$87,$F9,$5E,$BF,$CB,$F8,$8F,$9D,$FC
	.dc.b	$03,$FE,$83,$1E,$B6,$3D,$9C,$FB,$1F,$37,$83,$F7,$1E,$1B,$BF,$31
	.dc.b	$E5,$F0,$18,$B0,$EB,$8F,$6A,$F0,$1B,$C1,$9F,$96,$31,$AF,$A5,$D6
	.dc.b	$FE,$D7,$87,$0F,$C1,$79,$71,$FC,$67,$F2,$FE,$07,$CA,$7F,$D6,$91
	.dc.b	$5C,$9F,$8F,$E5,$E1,$E0,$FE,$17,$3F,$A5,$3E,$5E,$6F,$9A,$F9,$F9
	.dc.b	$BE,$6B,$E9,$E6,$F9,$AF,$AF,$9B,$E6,$BE,$DE,$6F,$80,$FC,$5E,$3F
	.dc.b	$9A,$DE,$3F,$92,$FE,$F4,$FE,$0B,$B7,$EB,$A5,$CD,$F4,$E7,$97,$27
	.dc.b	$47,$D7,$CB,$FB,$3C,$7C,$31,$8D,$99,$F9,$A3,$1A,$7A,$9D,$6B,$EE
	.dc.b	$78,$6D,$FA,$FE,$1F,$8B,$B7,$C9,$E7,$F8,$BA,$7F,$01,$E2,$CE,$1F
	.dc.b	$2C,$FA,$67,$C9,$3E,$B9,$F2,$4F,$B6,$7C,$93,$EF,$9F,$24,$FC,$27
	.dc.b	$37,$E1,$9F,$6F,$0F,$E5,$78,$3E,$14,$63,$87,$F9,$CB,$F2,$B8,$FF
	.dc.b	$7D,$5F,$91,$8F,$8E,$7D,$13,$A4,$7D,$FB,$FC,$84,$7C,$7E,$6F,$C8
	.dc.b	$3F,$27,$9B,$F2,$0F,$CB,$E6,$FC,$83,$F3,$77,$F8,$3A,$FC,$38,$FC
	.dc.b	$79,$FC,$87,$27,$C5,$7E,$B8,$91,$F8,$3F,$6B,$F6,$3F,$15,$FB,$2F
	.dc.b	$77,$F6,$7F,$15,$FB,$4F,$7F,$F6,$BF,$15,$FB,$6F,$81,$FD,$BF,$C5
	.dc.b	$7E,$E3,$E0,$FF,$73,$F1,$5F,$BA,$F8,$5F,$DD,$FC,$57,$EF,$3E,$1F
	.dc.b	$F7,$BF,$15,$FB,$9E,$0F,$C2,$7A,$3F,$C9,$5F,$C3,$78,$FA,$7F,$2D
	.dc.b	$5C,$9F,$CC,$61,$D9,$FC,$D6,$5E,$27,$F3,$9A,$79,$9F,$CF,$6D,$EA
	.dc.b	$7F,$41,$C1,$EE,$7E,$01,$FA,$3E,$3F,$86,$FF,$22,$E6,$FD,$B7,$5F
	.dc.b	$EC,$FB,$BE,$6F,$8B,$FB,$4A,$FC,$A9,$F8,$C3,$F3,$27,$E3,$0F,$CE
	.dc.b	$9F,$8C,$3F,$45,$E2,$FE,$92,$BF,$7D,$E3,$7E,$A5,$FC,$3F,$83,$F8
	.dc.b	$C7,$F1,$3E,$0F,$C7,$7F,$BB,$61,$FC,$7F,$0F,$DD,$4E,$FC,$3F,$DB
	.dc.b	$5F,$37,$FA,$EE,$1D,$DF,$EB,$F8,$FE,$C7,$1F,$E1,$DE,$37,$EC,$5F
	.dc.b	$E2,$1E,$37,$EC,$5F,$E2,$5E,$37,$EC,$73,$FE,$4D,$AF,$F1,$DF,$1B
	.dc.b	$F6,$3B,$23,$E2,$9B,$9F,$8A,$7E,$5C,$FD,$81,$F9,$B3,$F6,$07,$E7
	.dc.b	$CF,$D8,$1F,$C1,$FC,$6F,$D2,$E9,$FC,$46,$3F,$9C,$9F,$C5,$CF,$82
	.dc.b	$7F,$85,$DF,$F5,$BF,$85,$FC,$5A,$34,$7E,$E3,$7F,$88,$FF,$7D,$DF
	.dc.b	$E2,$3F,$DF,$F7,$F8,$97,$FE,$04,$DF,$E2,$3F,$E0,$6D,$FE,$23,$FE
	.dc.b	$08,$DB,$DD,$BF,$E1,$70,$F8,$A7,$EC,$3C,$3F,$D9,$3F,$C8,$76,$FC
	.dc.b	$4F,$17,$EE,$2B,$F9,$F7,$89,$FB,$C7,$FA,$07,$8B,$FC,$01,$FE,$85
	.dc.b	$E3,$7F,$04,$7F,$83,$61,$FC,$EC,$FE,$14,$7F,$0C,$3F,$8A,$F8,$FF
	.dc.b	$C4,$74,$FF,$1A,$92,$3F,$81,$73,$7E,$1D,$FE,$09,$C5,$F1,$35,$FE
	.dc.b	$71,$1F,$C8,$0F,$A7,$87,$89,$FA,$EF,$99,$FB,$6F,$A9,$FE,$5C,$7F
	.dc.b	$13,$3E,$19,$F0,$4F,$89,$C7,$FC,$49,$FD,$81,$0F,$E5,$BE,$3F,$ED
	.dc.b	$5F,$E5,$FD,$3F,$B9,$7F,$98,$F5,$7E,$F5,$FE,$67,$D7,$FC,$05,$FE
	.dc.b	$6B,$D9,$FC,$15,$FE,$6F,$DB,$FC,$25,$FE,$73,$DD,$FC,$35,$FE,$1E
	.dc.b	$7F,$80,$7C,$0F,$E2,$5A,$7F,$71,$F0,$34,$FF,$3A,$AE,$4F,$F3,$CB
	.dc.b	$EC,$FF,$3E,$C3,$C4,$FF,$40,$C7,$CC,$FF,$42,$CB,$D4,$FF,$44,$CF
	.dc.b	$DC,$FF,$46,$D3,$F8,$8B,$FD,$23,$C0,$FF,$30,$78,$1F,$D7,$B7,$7C
	.dc.b	$4F,$F8,$98,$9C,$1F,$DE,$FE,$2B,$FC,$53,$EE,$FF,$95,$FE,$07,$FC
	.dc.b	$C3,$3F,$DD,$17,$C1,$E7,$DF,$D7,$C2,$6F,$E0,$78,$7C,$2F,$1F,$83
	.dc.b	$E5,$F0,$DC,$FE,$17,$A7,$C3,$F5,$F8,$6E,$DF,$11,$DF,$5F,$F3,$37
	.dc.b	$C3,$FF,$9D,$F7,$FF,$57,$C0,$DB,$FE,$35,$7F,$DC,$8F,$96,$FB,$39
	.dc.b	$C7,$C4,$EF,$3F,$33,$C7,$5F,$53,$CF,$7F,$81,$E1,$FC,$28,$F8,$87
	.dc.b	$F8,$7F,$8F,$E3,$BF,$CF,$4F,$F4,$26,$4F,$89,$F9,$04,$8F,$37,$B3
	.dc.b	$F2,$FE,$1F,$E6,$7C,$3F,$CD,$F7,$3E,$77,$C3,$FC,$FF,$7B,$E8,$7C
	.dc.b	$3F,$FA,$1F,$F8,$07,$FD,$27,$97,$F9,$4D,$7C,$16,$7E,$0F,$5F,$09
	.dc.b	$BF,$85,$E1,$F0,$DC,$7E,$1F,$FA,$07,$27,$EE,$7D,$0F,$F4,$3E,$7C
	.dc.b	$E7,$6F,$5B,$E6,$F7,$C7,$C5,$97,$8F,$E1,$FC,$B3,$F4,$76,$F4,$BF
	.dc.b	$4F,$83,$DA,$FD,$5E,$2F,$C5,$71,$FC,$38,$FC,$57,$27,$E5,$3D,$8F
	.dc.b	$F5,$8F,$37,$C9,$9F,$B9,$3B,$F2,$7F,$D3,$D3,$FE,$C2,$9E,$87,$E0
	.dc.b	$79,$FE,$53,$F8,$3E,$7F,$94,$FE,$17,$9F,$D0,$FC,$9D,$FE,$43,$F2
	.dc.b	$F9,$BE,$43,$F3,$79,$BE,$43,$FE,$7B,$F4,$7F,$26,$C3,$07,$C9,$38
	.dc.b	$1F,$24,$E8,$7C,$93,$C0,$7D,$9E,$2F,$92,$FE,$AF,$8B,$E4,$BF,$AD
	.dc.b	$E2,$F9,$2F,$EB,$F8,$BE,$4B,$F1,$08,$7E,$47,$9B,$FD,$2B,$D3,$F9
	.dc.b	$4E,$1F,$F2,$3B,$FE,$A2,$F1,$FF,$35,$5F,$B9,$E9,$FD,$E7,$5F,$E8
	.dc.b	$23,$F7,$9D,$9F,$A4,$7F,$7D,$DB,$FA,$87,$F8,$07,$77,$EB,$1F,$D7
	.dc.b	$1F,$B1,$F0,$BF,$66,$FF,$05,$E8,$FE,$43,$8F,$FB,$93,$D2,$FD,$E4
	.dc.b	$FE,$DF,$C3,$FE,$00,$FE,$E7,$C2,$E9,$BF,$DE,$CB,$C0,$FE,$03,$3F
	.dc.b	$23,$F8,$2D,$3D,$0F,$E1,$35,$FD,$BE,$1F,$E9,$3F,$23,$E0,$3F,$B7
	.dc.b	$21,$F1,$FC,$9F,$E4,$2E,$2B,$E3,$BF,$BE,$3F,$CE,$B3,$D1,$FE,$41
	.dc.b	$5E,$0D,$7F,$2F,$87,$91,$FC,$CE,$3E,$87,$F3,$79,$7B,$1F,$CE,$E7
	.dc.b	$F0,$30,$FE,$55,$A7,$C4,$7F,$D4,$DC,$5F,$C7,$77,$F9,$7A,$FE,$83
	.dc.b	$83,$F9,$3D,$7F,$1F,$F1,$BF,$95,$BF,$C8,$7C,$7F,$E5,$EF,$F2,$3F
	.dc.b	$23,$F9,$9B,$FC,$97,$C9,$FB,$EF,$F2,$7D,$7F,$07,$8F,$FA,$33,$97
	.dc.b	$F1,$11,$FE,$B5,$E6,$FD,$A5,$7F,$1C,$F0,$BF,$65,$3F,$B6,$7F,$27
	.dc.b	$1F,$95,$3F,$64,$7F,$01,$3F,$64,$7F,$05,$3F,$64,$7E,$84,$FD,$19
	.dc.b	$FE,$BE,$F2,$1E,$37,$F1,$08,$C7,$8F,$FC,$51,$D7,$91,$FC,$69,$FF
	.dc.b	$65,$8E,$2F,$E3,$6E,$8F,$D9,$9C,$8F,$D9,$9D,$EB,$F8,$DB,$E3,$3F
	.dc.b	$AE,$78,$9F,$C2,$1F,$EB,$DE,$2F,$F0,$C7,$FB,$07,$8D,$FC,$E5,$FE
	.dc.b	$76,$7E,$CB,$7F,$E2,$BA,$7C,$8C,$7F,$AB,$3F,$22,$7E,$51,$C0,$F9
	.dc.b	$07,$43,$E4,$1E,$03,$E4,$1E,$43,$CD,$2D,$51,$F0,$75,$FE,$86,$FF
	.dc.b	$0F,$3E,$67,$8B,$FC,$71,$F9,$DE,$2F,$F1,$C7,$FC,$2E,$7E,$43,$E0
	.dc.b	$3F,$D0,$4F,$C9,$36,$7F,$1B,$39,$9D,$70,$F1,$3E,$1F,$1F,$F9,$6B
	.dc.b	$E7,$3F,$8E,$1F,$82,$3F,$8E,$1F,$86,$3E,$0C,$7F,$3C,$F2,$7E,$36
	.dc.b	$9F,$22,$7F,$C7,$BC,$7F,$94,$7F,$DA,$0F,$EF,$1F,$DA,$9F,$9B,$3F
	.dc.b	$BC,$1F,$9F,$3F,$BE,$1F,$A3,$3F,$C0,$07,$E9,$CF,$F0,$21,$FA,$B3
	.dc.b	$F8,$91,$FC,$6C,$FE,$2C,$7F,$83,$7A,$BF,$8D,$BF,$8F,$C5,$E1,$FE
	.dc.b	$E7,$1F,$EA,$BD,$9F,$BD,$7F,$AB,$F6,$FF,$01,$BF,$F0,$2F,$77,$F0
	.dc.b	$57,$FC,$0F,$DF,$FC,$25,$FF,$04,$F8,$1F,$C3,$5F,$F0,$5F,$83,$FC
	.dc.b	$47,$0F,$F8,$03,$E1,$7C,$17,$F8,$BC,$9B,$E5,$BC,$AF,$C2,$F0,$B9
	.dc.b	$9F,$11,$FE,$0B,$26,$1F,$E0,$B6,$BF,$B3,$6E,$1F,$7A,$B9,$46,$3A
	.dc.b	$3A,$DF,$4B,$E4,$43,$91,$6F,$94,$6E,$32,$48,$FE,$FB,$C7,$FD,$B6
	.dc.b	$3F,$C7,$EF,$77,$F3,$D7,$70,$F1,$C7,$5C,$3C,$73,$9C,$3B,$61,$E0
	.dc.b	$8C,$12,$B8,$1D,$0F,$46,$5F,$25,$FD,$67,$97,$FE,$31,$23,$FE,$31
	.dc.b	$E3,$FF,$1A,$7B,$7F,$E3,$5D,$7F,$C6,$DE,$E7,$F8,$DF,$DC,$FF,$1C
	.dc.b	$7B,$BF,$E3,$9F,$77,$FC,$75,$EF,$7F,$8E,$FD,$EF,$F1,$E7,$BF,$FE
	.dc.b	$3D,$F8,$1F,$F1,$F7,$C0,$FE,$F3,$6F,$B9,$C9,$F6,$BF,$80,$7F,$CC
	.dc.b	$5C,$9F,$4D,$7C,$07,$E4,$67,$FC,$79,$D1,$FE,$65,$BF,$E0,$B7,$BA
	.dc.b	$E8,$F1,$8E,$E7,$F9,$2F,$1F,$1B,$FE,$4D,$CF,$CE,$FF,$93,$F5,$F5
	.dc.b	$BF,$E5,$1D,$FD,$EF,$F9,$4F,$87,$E1,$BF,$14,$FF,$2B,$72,$7C,$87
	.dc.b	$FC,$2D,$CF,$FC,$3B,$2F,$C1,$CE,$5F,$C2,$F2,$F2,$AF,$E8,$97,$D9
	.dc.b	$FE,$8D,$87,$89,$FE,$91,$8F,$99,$FE,$95,$97,$A9,$FE,$99,$9F,$B9
	.dc.b	$F9,$67,$F4,$7D,$FE,$5E,$FF,$37,$6F,$95,$3F,$D3,$3A,$3F,$B9,$3F
	.dc.b	$2A,$74,$D3,$91,$D7,$5F,$CE,$E1,$E1,$7F,$3F,$8F,$95,$FD,$0E,$5E
	.dc.b	$97,$F4,$79,$FB,$5F,$D2,$E9,$F0,$5F,$87,$87,$CA,$21,$F2,$BA,$3F
	.dc.b	$34,$FE,$A7,$2F,$D0,$BF,$9C,$CF,$F4,$AE,$4F,$CF,$15,$CD,$7E,$96
	.dc.b	$38,$70,$7E,$A2,$B9,$C6,$E3,$0F,$C9,$F8,$2F,$E4,$76,$F0,$7F,$91
	.dc.b	$CF,$C5,$87,$02,$BC,$F8,$F7,$AE,$99,$63,$2E,$F3,$92,$38,$24,$FA
	.dc.b	$A3,$2D,$8F,$3F,$57,$EE,$71,$F8,$33,$FA,$B6,$BE,$C3,$19,$FF,$13
	.dc.b	$6E,$1F,$EC,$F7,$44,$3E,$49,$A3,$53,$D1,$C9,$B7,$04,$9E,$D1,$96
	.dc.b	$BF,$EC,$8E,$4F,$E4,$18,$FF,$2F,$F6,$7F,$2B,$39,$3F,$96,$1E,$13
	.dc.b	$FD,$C9,$7D,$70,$F4,$47,$88,$91,$97,$CF,$87,$82,$31,$1C,$A3,$04
	.dc.b	$9E,$73,$AF,$41,$EA,$7A,$67,$35,$BB,$EB,$18,$23,$82,$1D,$DC,$B4
	.dc.b	$C9,$ED,$21,$DC,$F9,$F9,$7F,$1D,$7F,$CC,$ED,$CB,$F1,$91,$C0,$F8
	.dc.b	$35,$E5,$7B,$10,$D6,$3F,$09,$B1,$B6,$DF,$9B,$1F,$83,$1D,$23,$D2
	.dc.b	$E8,$F4,$9E,$73,$B1,$4D,$18,$39,$1C,$0F,$BC,$7B,$30,$FE,$E6,$F0
	.dc.b	$21,$CC,$87,$23,$A5,$FA,$DF,$15,$7A,$23,$11,$B3,$99,$4E,$6E,$89
	.dc.b	$7F,$0E,$3F,$39,$B7,$EE,$DF,$DF,$6F,$FC,$01,$FE,$5B,$5F,$C0,$F9
	.dc.b	$BF,$A7,$EF,$FD,$67,$83,$D7,$5D,$B1,$E8,$F8,$49,$7E,$0D,$3E,$6C
	.dc.b	$6D,$FE,$C9,$9F,$AE,$BF,$89,$D7,$F1,$AC,$3F,$82,$B6,$FF,$82,$11
	.dc.b	$C1,$27,$89,$2F,$F8,$94,$EE,$30,$FF,$2B,$CF,$E0,$3F,$EB,$66,$EB
	.dc.b	$FC,$4A,$FA,$5F,$26,$D7,$F2,$C8,$C5,$F4,$8F,$E4,$27,$F8,$43,$87
	.dc.b	$D3,$19,$BE,$8D,$C7,$37,$C8,$97,$D4,$FE,$81,$B7,$E8,$E7,$FC,$06
	.dc.b	$EA,$3B,$92,$FA,$D6,$5B,$1C,$CF,$E9,$C5,$72,$F1,$AF,$1B,$96,$B9
	.dc.b	$3E,$54,$FF,$6C,$C5,$C1,$FC,$71,$B3,$9C,$7D,$CF,$03,$A1,$4E,$91
	.dc.b	$F3,$1B,$1F,$2F,$87,$FA,$06,$8F,$82,$FC,$8D,$7F,$7F,$6D,$FF,$54
	.dc.b	$6C,$BD,$BB,$EE,$63,$E2,$73,$EF,$6B,$E2,$77,$F0,$1C,$3E,$27,$8F
	.dc.b	$C1,$72,$F8,$9E,$7F,$09,$D3,$E2,$7A,$FC,$37,$6F,$89,$EF,$F1,$1E
	.dc.b	$0E,$DF,$ED,$3D,$FF,$A9,$EB,$FF,$1E,$46,$B9,$BF,$B5,$72,$30,$E7
	.dc.b	$18,$CB,$BB,$AD,$3C,$6F,$0D,$BC,$EF,$2E,$0F,$5B,$D3,$8B,$DE,$F6
	.dc.b	$E4,$F8,$7C,$B2,$3F,$D9,$BC,$FF,$33,$E2,$7E,$39,$23,$CB,$EA,$FC
	.dc.b	$AF,$81,$F9,$7E,$87,$CE,$FE,$01,$FF,$62,$D7,$F0,$45,$E9,$67,$C4
	.dc.b	$D7,$A7,$BF,$83,$E1,$EA,$F1,$F8,$4E,$5E,$BF,$3F,$85,$E9,$EC,$F5
	.dc.b	$F3,$7E,$56,$BF,$E0,$FF,$23,$F1,$87,$E3,$7B,$3F,$6B,$E5,$7F,$A8
	.dc.b	$B8,$3F,$9C,$D7,$8F,$0E,$8F,$F8,$5E,$19,$78,$63,$1A,$79,$5D,$6D
	.dc.b	$E9,$78,$70,$7B,$F1,$F8,$31,$F6,$38,$BF,$1B,$C7,$F1,$E3,$F9,$2E
	.dc.b	$8E,$4F,$F3,$F7,$2F,$CD,$9F,$BF,$1B,$EF,$FF,$3F,$5F,$CD,$9F,$C1
	.dc.b	$F3,$77,$7C,$27,$A7,$9B,$E5,$BE,$BE,$6F,$96,$FB,$79,$BE,$5B,$EF
	.dc.b	$E6,$F9,$6F,$C1,$E6,$F9,$6F,$F1,$DE,$FF,$9C,$FC,$5E,$6F,$8E,$FD
	.dc.b	$BE,$5F,$8C,$FD,$DE,$5F,$8C,$FD,$FE,$5F,$8C,$FE,$07,$97,$D0,$FC
	.dc.b	$2D,$BD,$8F,$C4,$DF,$E0,$72,$FF,$A7,$F8,$BE,$2F,$99,$6B,$E2,$90
	.dc.b	$E0,$7C,$F3,$E2,$9F,$6E,$7F,$25,$87,$8B,$8F,$FD,$B5,$CE,$CF,$CF
	.dc.b	$18,$D3,$D6,$EB,$5F,$7B,$C3,$6F,$87,$97,$DA,$DF,$FC,$6F,$9F,$FB
	.dc.b	$BB,$A3,$F2,$11,$AB,$E1,$5F,$3C,$FC,$76,$1D,$AF,$F2,$98,$F8,$9F
	.dc.b	$CB,$65,$E7,$7F,$31,$A7,$C8,$C3,$FD,$81,$C1,$F2,$1F,$C2,$E3,$F0
	.dc.b	$DF,$F6,$26,$FF,$22,$1E,$4B,$77,$B5,$FB,$52,$4F,$E8,$7C,$DF,$DC
	.dc.b	$7B,$FF,$B9,$F7,$3F,$75,$F0,$FF,$BB,$F8,$1F,$DE,$7C,$47,$EF,$7E
	.dc.b	$0F,$F7,$DF,$13,$FB,$FF,$85,$FE,$5B,$E7,$BE,$2A,$FC,$26,$3E,$1F
	.dc.b	$3F,$0F,$AF,$89,$DF,$C4,$7F,$22,$F7,$2F,$1F,$FA,$27,$25,$70,$8C
	.dc.b	$61,$D1,$D6,$5E,$17,$F7,$5A,$79,$5F,$DE,$6D,$E9,$7F,$7D,$C1,$ED
	.dc.b	$7F,$80,$71,$7F,$17,$E3,$F8,$71,$FC,$5F,$93,$E5,$73,$7F,$A6,$B9
	.dc.b	$FF,$2B,$3F,$EC,$97,$F7,$AF,$E4,$4F,$CE,$1F,$91,$3F,$6A,$7F,$21
	.dc.b	$F2,$3F,$6A,$FF,$23,$F2,$3F,$6A,$FF,$25,$F2,$3F,$6A,$FF,$27,$F2
	.dc.b	$3F,$6A,$FF,$29,$F2,$3F,$6B,$DB,$FE,$85,$8F,$C8,$9F,$EB,$33,$F7
	.dc.b	$47,$ED,$CF,$C8,$1F,$B7,$3F,$81,$1F,$C7,$0F,$E0,$C7,$F1,$C3,$F8
	.dc.b	$51,$E8,$7F,$4F,$E0,$7F,$1C,$8F,$E3,$3E,$5F,$C0,$7F,$8D,$E7,$FC
	.dc.b	$77,$CA,$FE,$35,$18,$3F,$29,$47,$FD,$E1,$E5,$7E,$66,$31,$B7,$FB
	.dc.b	$1B,$C2,$FC,$F4,$7E,$82,$BC,$67,$F0,$B3,$CE,$7F,$0F,$3D,$67,$B4
	.dc.b	$F9,$07,$C1,$3E,$41,$FD,$AB,$C7,$F8,$F5,$FE,$10,$8C,$1F,$20,$FE
	.dc.b	$88,$7C,$8C,$FF,$81,$3D,$9F,$E0,$CF,$89,$FE,$17,$E2,$FE,$DE,$3F
	.dc.b	$5F,$E2,$7F,$5E,$7F,$04,$7E,$10,$F9,$27,$F6,$83,$E2,$9F,$8B,$3F
	.dc.b	$B7,$F9,$5F,$C7,$DF,$94,$4A,$FD,$FB,$C8,$E8,$FD,$67,$EC,$63,$FD
	.dc.b	$7F,$8F,$F1,$F3,$FC,$83,$8F,$F1,$EF,$F2,$2E,$3F,$C7,$BF,$C9,$38
	.dc.b	$FF,$1E,$FF,$26,$E3,$FC,$7E,$1F,$C7,$71,$F9,$3E,$5F,$F6,$C7,$DF
	.dc.b	$3C,$27,$FB,$75,$F4,$7F,$B8,$61,$E0,$7F,$B9,$63,$E4,$7C,$D7,$E8
	.dc.b	$3E,$4B,$F8,$FC,$4F,$E1,$B1,$72,$7E,$46,$7F,$8F,$D7,$F4,$3E,$3F
	.dc.b	$F8,$01,$B7,$F1,$E8,$FE,$F9,$CB,$FB,$CA,$FE,$83,$CD,$FC,$7A,$7F
	.dc.b	$C0,$1E,$37,$F1,$E7,$FC,$05,$E3,$7F,$1E,$7F,$C0,$9E,$37,$F1,$E7
	.dc.b	$FC,$0D,$E3,$7F,$1E,$EF,$FE,$25,$1F,$EA,$0F,$09,$E0,$FF,$17,$8C
	.dc.b	$78,$5F,$C7,$1F,$F5,$2F,$A9,$FE,$3F,$E3,$FD,$0B,$63,$F9,$BD,$BC
	.dc.b	$5F,$E4,$D1,$FC,$8F,$C6,$FE,$54,$FF,$25,$F1,$FC,$4F,$90,$FE,$42
	.dc.b	$7A,$0F,$E4,$27,$B0,$FE,$42,$7C,$03,$F9,$09,$FD,$04,$F8,$E4,$1C
	.dc.b	$8E,$CF,$DB,$E0,$F8,$9F,$E7,$E4,$AF,$E7,$1E,$EF,$F4,$2F,$83,$FE
	.dc.b	$87,$E5,$7F,$44,$F8,$9F,$E8,$B8,$FF,$46,$F8,$9F,$E8,$FF,$03,$FD
	.dc.b	$23,$E2,$7F,$A4,$FC,$1F,$F4,$AF,$89,$FE,$97,$F0,$BF,$D3,$3E,$27
	.dc.b	$FA,$6F,$C3,$FF,$4E,$F8,$9F,$99,$E0,$FF,$23,$E1,$FF,$21,$18,$E9
	.dc.b	$F3,$CF,$08,$FE,$99,$7D,$1F,$E9,$D8,$F8,$1F,$EA,$19,$F9,$1F,$EA
	.dc.b	$5A,$FA,$1F,$EA,$9B,$FB,$1F,$EA,$DC,$3F,$D8,$38,$BE,$14,$7E,$0F
	.dc.b	$8F,$FA,$4F,$F0,$0F,$FA,$7C,$91,$FE,$92,$5F,$15,$7C,$FC,$7B,$99
	.dc.b	$F7,$75,$F1,$3B,$F7,$F8,$7C,$4F,$1F,$81,$E5,$F1,$3C,$FE,$0F,$A7
	.dc.b	$C4,$F5,$F8,$5E,$DE,$0F,$7F,$86,$F8,$9C,$3F,$E0,$4F,$85,$FF,$33
	.dc.b	$B8,$F6,$DC,$6A,$E3,$3F,$CE,$7B,$BF,$66,$FF,$83,$25,$8F,$82,$35
	.dc.b	$DD,$FC,$F2,$5A,$79,$A7,$27,$1D,$BD,$6E,$32,$F0,$F0,$7C,$08,$C5
	.dc.b	$FF,$AD,$5B,$C7,$F1,$DC,$F2,$FC,$B6,$1F,$CA,$8F,$31,$81,$A7,$AB
	.dc.b	$0F,$7C,$7D,$39,$E8,$EB,$3F,$55,$FD,$38,$F1,$F1,$FD,$37,$CB,$C7
	.dc.b	$F4,$DF,$3F,$1F,$D3,$7D,$3C,$7F,$4D,$F5,$F1,$FD,$37,$FC,$3D,$E4
	.dc.b	$7F,$54,$E4,$FF,$72,$4F,$F0,$19,$F8,$3C,$9F,$49,$F2,$1F,$47,$93
	.dc.b	$E8,$38,$DF,$EC,$F2,$7C,$F8,$C7,$13,$93,$E7,$C6,$39,$3E,$73,$AE
	.dc.b	$5F,$C8,$BF,$43,$9B,$E2,$3F,$48,$F9,$5E,$37,$F0,$37,$F8,$BF,$47
	.dc.b	$E8,$3A,$3E,$66,$FF,$C1,$67,$F8,$19,$FA,$63,$FC,$79,$E3,$7F,$03
	.dc.b	$7F,$C7,$DE,$37,$F0,$37,$FC,$81,$E3,$7F,$03,$7F,$C8,$5E,$37,$F0
	.dc.b	$37,$FC,$89,$E3,$7F,$03,$7F,$C8,$DE,$37,$F0,$37,$F0,$FF,$13,$FE
	.dc.b	$4B,$69,$FE,$4B,$F1,$FF,$81,$3F,$CD,$FC,$7F,$E0,$4F,$F3,$9F,$1F
	.dc.b	$F8,$13,$FC,$EF,$C7,$FE,$04,$FF,$3D,$F1,$FF,$81,$3F,$CF,$FC,$7F
	.dc.b	$E0,$4F,$F4,$1F,$1F,$F8,$13,$FE,$59,$E3,$FE,$07,$C9,$F1,$FA,$DC
	.dc.b	$9F,$16,$BF,$83,$1B,$3E,$39,$CC,$F8,$E7,$F1,$03,$E3,$9F,$C5,$0F
	.dc.b	$8E,$7F,$18,$3E,$39,$FC,$70,$F8,$E7,$F2,$03,$E3,$9F,$0C,$FC,$89
	.dc.b	$FC,$A0,$87,$F5,$BE,$FF,$98,$FF,$5D,$E0,$FA,$0F,$F5,$FE,$1F,$A8
	.dc.b	$FF,$61,$E2,$FB,$0F,$F6,$3E,$3F,$B8,$FF,$65,$E4,$FC,$03,$FD,$9F
	.dc.b	$97,$FA,$33,$F0,$39,$BF,$13,$B7,$C8,$AF,$E1,$53,$FB,$99,$FD,$C7
	.dc.b	$5F,$F2,$BF,$1B,$F6,$F1,$8F,$17,$F6,$D5,$FA,$0F,$17,$F6,$CF,$E8
	.dc.b	$BC,$5F,$DB,$3F,$A4,$F1,$7F,$6C,$FE,$9B,$C5,$FD,$B3,$FA,$8E,$3F
	.dc.b	$DB,$3F,$0C,$F8,$B8,$FF,$4F,$70,$7E,$BB,$C4,$FD,$B4,$79,$BC,$4F
	.dc.b	$DA,$BE,$8F,$13,$F6,$8F,$AB,$C4,$FD,$9B,$EC,$F1,$3F,$66,$FB,$BC
	.dc.b	$4F,$D9,$E3,$FB,$AF,$13,$F6,$6F,$EF,$0F,$D9,$F1,$FC,$67,$F6,$C4
	.dc.b	$3F,$80,$F1,$7C,$6F,$1B,$F8,$DC,$7F,$2C,$BE,$F1,$FC,$DF,$1F,$1B
	.dc.b	$FC,$E7,$2F,$3B,$FC,$EF,$3F,$5B,$FC,$F7,$4F,$7B,$FC,$FF,$5F,$E9
	.dc.b	$5E,$37,$F4,$CC,$BF,$1A,$7C,$53,$F9,$09,$FC,$F2,$B9,$BF,$D2,$EF
	.dc.b	$BB,$FD,$14,$FE,$5A,$7F,$42,$C7,$CE,$FF,$51,$CB,$E2,$ED,$F8,$4D
	.dc.b	$BE,$2B,$F8,$69,$F8,$A7,$F1,$63,$F5,$C7,$EC,$0F,$F1,$C7,$8F,$F9
	.dc.b	$27,$F2,$87,$EB,$CF,$CC,$1F,$AF,$3F,$38,$7E,$BC,$FD,$01,$FA,$F3
	.dc.b	$F4,$87,$EB,$CF,$D4,$1F,$AF,$3F,$99,$F8,$FF,$C5,$5A,$66,$7F,$94
	.dc.b	$F6,$FE,$B9,$FE,$57,$DB,$FA,$E7,$F9,$6F,$6F,$EB,$9F,$E5,$FD,$BF
	.dc.b	$AE,$7F,$98,$F2,$7E,$B9,$FE,$67,$DB,$ED,$7F,$9A,$E7,$FC,$DF,$8F
	.dc.b	$FC,$77,$DB,$FE,$16,$D7,$FC,$28,$4D,$7F,$94,$9E,$62,$5F,$EC,$A3
	.dc.b	$51,$8A,$FF,$00,$46,$27,$53,$C9,$AE,$CF,$F0,$54,$6E,$78,$C7,$EB
	.dc.b	$9F,$F4,$83,$C5,$CB,$C7,$BF,$FC,$2F,$3C,$63,$99,$B2,$9B,$EE,$46
	.dc.b	$E7,$8C,$ED,$AE,$E3,$DB,$E0,$7F,$39,$A7,$C4,$7F,$8E,$49,$5F,$E1
	.dc.b	$6F,$89,$FF,$1E,$7C,$0F,$F8,$F7,$E0,$7F,$C7,$DE,$CF,$F8,$FF,$E0
	.dc.b	$7F,$DC,$57,$FD,$DF,$F8,$07,$FC,$56,$DF,$87,$CF,$07,$F8,$56,$5F
	.dc.b	$05,$8E,$1F,$F7,$3C,$3E,$0F,$18,$38,$92,$F8,$0C,$C6,$3E,$22,$CB
	.dc.b	$E0,$AF,$C4,$E3,$E1,$3F,$73,$8F,$FB,$9B,$C9,$FF,$09,$3E,$78,$DB
	.dc.b	$FE,$37,$BE,$4D,$C3,$AE,$5D,$B3,$EF,$AF,$83,$7F,$0E,$DF,$0F,$0F
	.dc.b	$F1,$EF,$0F,$C7,$7F,$C7,$FC,$7F,$E6,$5E,$4F,$9B,$1F,$E1,$AE,$5F
	.dc.b	$F3,$59,$8E,$3F,$E8,$6E,$E3,$89,$FE,$48,$E5,$FA,$2F,$F9,$2B,$97
	.dc.b	$C2,$FA,$73,$F2,$BE,$BD,$7D,$2F,$AC,$F7,$72,$7C,$07,$E1,$1F,$C7
	.dc.b	$31,$FE,$84,$4D,$7F,$D4,$BE,$7F,$CB,$9F,$8D,$CB,$F4,$2F,$E8,$F1
	.dc.b	$76,$9F,$C1,$63,$E2,$7F,$04,$7C,$DC,$3E,$33,$FE,$3A,$E0,$F8,$CF
	.dc.b	$F8,$F3,$83,$E0,$3F,$14,$FF,$33,$EF,$FB,$B7,$F3,$9C,$FF,$A0,$6F
	.dc.b	$4F,$E8,$63,$11,$F9,$F3,$24,$30,$4D,$BF,$72,$7F,$9A,$7D,$9F,$F6
	.dc.b	$C7,$49,$B5,$D2,$77,$9F,$EE,$1B,$7F,$B4,$6D,$FE,$C1,$B5,$C6,$B8
	.dc.b	$12,$B8,$36,$B8,$36,$78,$4F,$F0,$0E,$AF,$DC,$63,$FB,$08,$FD,$8D
	.dc.b	$FE,$F9,$B5,$F0,$9B,$3B,$7F,$87,$98,$31,$3C,$5C,$90,$FE,$1E,$6C
	.dc.b	$F0,$3B,$72,$6D,$C1,$27,$C5,$19,$F2,$3F,$B5,$3F,$CC,$4E,$84,$39
	.dc.b	$47,$07,$A9,$27,$A5,$7A,$E3,$13,$B9,$CC,$7B,$23,$04,$9E,$13,$FC
	.dc.b	$93,$9D,$F0,$BE,$09,$CD,$6E,$F9,$46,$08,$E0,$87,$C4,$8E,$0D,$8D
	.dc.b	$B4,$D9,$EF,$3D,$2F,$93,$0F,$8B,$9F,$F4,$C9,$D6,$3E,$C7,$91,$0C
	.dc.b	$DF,$81,$B7,$E0,$69,$E0,$3D,$C0,$FE,$67,$86,$1E,$85,$38,$98,$3C
	.dc.b	$E7,$32,$99,$AF,$33,$C0,$FC,$AE,$1E,$CC,$FA,$44,$38,$9B,$31,$3C
	.dc.b	$9B,$87,$B5,$FE,$CF,$97,$D8,$78,$9A,$C7,$E0,$34,$F3,$57,$37,$64
	.dc.b	$70,$7E,$FF,$BF,$FA,$6E,$BF,$C0,$2B,$9D,$6F,$0F,$13,$63,$37,$CA
	.dc.b	$72,$D8,$FE,$CB,$87,$9E,$3F,$7B,$5F,$C0,$08,$E0,$95,$FC,$01,$B5
	.dc.b	$FC,$74,$93,$FC,$6F,$4F,$E1,$B1,$88,$FF,$74,$61,$FC,$68,$F5,$3B
	.dc.b	$CB,$F8,$64,$EE,$B8,$CE,$4F,$1E,$1E,$12,$4F,$5D,$7F,$88,$CE,$F1
	.dc.b	$F2,$B6,$B6,$DA,$ED,$87,$82,$38,$69,$F8,$37,$BC,$E9,$A7,$12,$1F
	.dc.b	$CD,$E7,$93,$FE,$39,$C3,$C6,$E4,$E2,$5C,$3C,$AD,$C3,$CA,$DB,$F0
	.dc.b	$4B,$0F,$23,$C3,$1F,$8B,$5B,$3C,$6E,$09,$8F,$91,$FF,$21,$E3,$FE
	.dc.b	$E5,$EB,$7C,$57,$F8,$58,$97,$FE,$0F,$DF,$FC,$33,$F0,$9F,$E1,$AF
	.dc.b	$84,$FF,$0D,$FC,$3F,$F8,$73,$E1,$7F,$C3,$BF,$11,$FE,$1E,$F8,$6F
	.dc.b	$F0,$FF,$C4,$FF,$88,$3E,$1F,$FC,$43,$F1,$5F,$E2,$2F,$88,$FF,$12
	.dc.b	$FF,$00,$BE,$26,$FC,$36,$3E,$27,$3F,$11,$FE,$77,$E6,$73,$FF,$03
	.dc.b	$E8,$62,$AE,$1C,$BF,$C1,$EF,$6F,$F8,$D6,$58,$F8,$32,$F1,$4F,$F8
	.dc.b	$AF,$4F,$33,$FE,$2D,$DB,$D4,$FF,$8B,$F8,$3D,$CF,$F8,$C7,$8B,$E1
	.dc.b	$3F,$E3,$3E,$4F,$F1,$EF,$2F,$C8,$8F,$F2,$77,$37,$F9,$43,$A1,$E8
	.dc.b	$7A,$76,$7C,$4F,$C2,$25,$FA,$78,$F5,$E4,$7C,$6F,$43,$E3,$FA,$3F
	.dc.b	$23,$E0,$FE,$4F,$89,$F2,$BE,$13,$E5,$E9,$F3,$3E,$17,$E6,$FA,$BF
	.dc.b	$3B,$E1,$BE,$7F,$AF,$F4,$3E,$1F,$E8,$FB,$3F,$4B,$E2,$3E,$7F,$0F
	.dc.b	$84,$FF,$2E,$79,$7F,$B5,$D7,$30,$EF,$F9,$BE,$0F,$D2,$BE,$73,$C6
	.dc.b	$3E,$AE,$3E,$08,$C7,$33,$4F,$2C,$63,$6F,$4B,$AE,$0F,$6B,$C3,$87
	.dc.b	$E0,$BC,$B8,$FF,$13,$C9,$F2,$23,$C9,$CB,$F3,$79,$BE,$87,$2D,$9F
	.dc.b	$0C,$6F,$83,$D3,$AF,$F9,$B9,$B8,$F6,$AE,$67,$E2,$72,$F1,$B8,$DF
	.dc.b	$F2,$3A,$7A,$23,$18,$35,$F6,$C6,$36,$F8,$4E,$B7,$F8,$EF,$0E,$0F
	.dc.b	$F5,$27,$93,$FB,$CF,$2F,$F9,$66,$29,$DE,$DE,$8C,$39,$BE,$AC,$7B
	.dc.b	$BE,$CC,$BC,$8F,$BB,$3F,$4B,$F0,$35,$F7,$3F,$03,$4F,$AB,$C5,$F1
	.dc.b	$78,$BE,$44,$3A,$7F,$32,$FE,$AF,$8B,$4D,$38,$9F,$54,$FB,$FD,$3F
	.dc.b	$97,$7F,$03,$CB,$F9,$77,$F0,$7C,$BF,$97,$7F,$0B,$D3,$F9,$77,$F0
	.dc.b	$FD,$3F,$97,$7F,$13,$D3,$F9,$77,$FD,$91,$CD,$F1,$FC,$BC,$47,$D2
	.dc.b	$9E,$0F,$D4,$AE,$8F,$D6,$BF,$03,$F6,$30,$F2,$3F,$6B,$1F,$43,$F7
	.dc.b	$32,$FD,$93,$EE,$3E,$FE,$7F,$91,$D3,$E2,$F4,$7F,$14,$F6,$FF,$8E
	.dc.b	$92,$BF,$43,$F1,$3F,$EC,$4C,$FF,$CC,$9C,$5F,$AA,$5F,$09,$9F,$7B
	.dc.b	$5F,$13,$BF,$80,$E1,$F1,$3C,$7E,$0B,$97,$C4,$F3,$F8,$7E,$9F,$13
	.dc.b	$D7,$E1,$BB,$7C,$4F,$7F,$88,$FD,$CE,$FF,$E8,$9E,$7F,$F0,$49,$8E
	.dc.b	$DF,$E5,$9C,$2A,$E3,$C7,$FB,$B8,$63,$DA,$31,$9F,$89,$D6,$BE,$67
	.dc.b	$86,$FE,$A7,$97,$0F,$C1,$E2,$F8,$51,$FD,$0B,$8F,$E3,$3F,$AA,$E4
	.dc.b	$FD,$CB,$F8,$D3,$E7,$9F,$BA,$3E,$96,$FF,$99,$7E,$B1,$FC,$1C,$FC
	.dc.b	$79,$FC,$2C,$FC,$79,$FC,$3C,$FC,$79,$FC,$53,$C2,$FD,$5E,$FF,$CD
	.dc.b	$7C,$3F,$D7,$BF,$CD,$FC,$4F,$8D,$5F,$C9,$38,$BE,$4B,$FC,$73,$D2
	.dc.b	$FC,$8E,$3F,$C3,$3B,$FF,$22,$FF,$0E,$EF,$FC,$8B,$FC,$43,$93,$F2
	.dc.b	$2F,$F1,$2E,$FF,$C8,$BF,$C5,$31,$FC,$8B,$FC,$5B,$BF,$F2,$39,$7F
	.dc.b	$0E,$F0,$BF,$63,$DD,$FD,$8B,$C0,$F8,$ED,$9F,$8C,$7C,$83,$81,$FB
	.dc.b	$53,$E3,$9F,$C6,$3C,$3F,$E0,$BC,$1F,$D5,$7C,$4F,$E1,$2F,$F5,$7F
	.dc.b	$17,$F8,$6B,$FD,$67,$93,$E3,$DF,$F6,$DE,$4F,$8E,$FF,$6F,$E4,$FE
	.dc.b	$36,$FF,$9E,$4F,$E5,$1E,$2F,$E8,$2F,$FD,$95,$E3,$7F,$1D,$9F,$DE
	.dc.b	$9F,$D1,$AF,$AE,$9F,$DF,$B1,$F0,$CF,$F7,$BC,$BC,$AF,$F7,$DC,$FD
	.dc.b	$2F,$F3,$73,$F8,$89,$EF,$E0,$FE,$7F,$BF,$C8,$CB,$FB,$8C,$FF,$A3
	.dc.b	$7C,$5F,$DC,$C6,$23,$E4,$57,$F4,$6E,$0F,$E9,$F1,$CF,$5F,$EF,$5C
	.dc.b	$3F,$59,$FE,$F9,$C5,$F6,$9F,$EF,$DC,$7F,$79,$FF,$00,$72,$7E,$09
	.dc.b	$FF,$01,$72,$FE,$19,$FF,$02,$73,$7E,$29,$FF,$77,$72,$7F,$1D,$CB
	.dc.b	$FB,$4F,$2F,$CE,$E4,$FE,$DB,$1F,$C9,$8F,$CD,$1F,$9C,$3F,$A9,$F1
	.dc.b	$FC,$68,$FE,$AF,$1F,$A4,$3F,$00,$7C,$63,$FA,$CF,$91,$EE,$7E,$0E
	.dc.b	$1F,$D0,$B8,$BE,$2F,$4F,$ED,$B8,$BF,$B4,$BF,$B9,$AF,$E0,$C7,$EF
	.dc.b	$4F,$DF,$9F,$C2,$7C,$6F,$E3,$4F,$F0,$BF,$1B,$F8,$D3,$FC,$37,$C6
	.dc.b	$FE,$34,$FF,$0F,$F1,$BF,$8D,$3F,$C4,$7C,$6F,$E3,$4F,$EB,$CF,$F0
	.dc.b	$77,$93,$FE,$15,$70,$7F,$75,$E4,$FE,$34,$F1,$3F,$8D,$78,$DF,$C6
	.dc.b	$5D,$78,$DF,$C6,$5E,$1E,$37,$F1,$97,$97,$8D,$FC,$65,$E9,$E3,$7B
	.dc.b	$5E,$C7,$C1,$3F,$8C,$1F,$E1,$ED,$FF,$67,$8B,$E2,$3F,$A0,$12,$7F
	.dc.b	$96,$7C,$4F,$F4,$3F,$2B,$FA,$27,$BB,$FD,$17,$E1,$FF,$A3,$7B,$FF
	.dc.b	$D1,$FE,$1F,$FA,$47,$C0,$FF,$49,$F8,$7F,$D6,$74,$FF,$8C,$BF,$80
	.dc.b	$7F,$CB,$FA,$7F,$85,$57,$C4,$6B,$B7,$FC,$D1,$0F,$84,$C5,$3C,$04
	.dc.b	$BE,$17,$31,$8F,$89,$B2,$F8,$7B,$F1,$1F,$E7,$5C,$BA,$BA,$8D,$D7
	.dc.b	$0C,$38,$E5,$CB,$3E,$78,$7A,$F3,$FE,$9D,$B7,$BD,$FE,$A1,$C1,$F0
	.dc.b	$DF,$EA,$5C,$5F,$1D,$F9,$27,$F5,$5E,$5F,$98,$FF,$85,$FC,$AF,$7F
	.dc.b	$C4,$FA,$09,$1F,$E5,$5E,$07,$C5,$7B,$3D,$BF,$6F,$B9,$EE,$F7,$7D
	.dc.b	$FF,$13,$F0,$3D,$FF,$83,$F1,$3F,$0B,$E0,$7E,$1F,$C4,$FC,$4F,$83
	.dc.b	$F8,$BF,$13,$F1,$BE,$17,$E3,$FC,$4F,$CA,$FD,$F7,$FD,$3D,$C3,$F9
	.dc.b	$6F,$03,$FC,$D3,$EB,$FF,$AE,$BD,$DF,$F4,$D4,$FF,$9A,$6B,$6F,$DB
	.dc.b	$BE,$8F,$A7,$1F,$03,$EB,$CF,$C8,$FB,$75,$F4,$3E,$FD,$FD,$8F,$C1
	.dc.b	$E1,$F8,$0F,$C3,$E3,$F8,$AF,$E2,$79,$7F,$A5,$F2,$9C,$EF,$F8,$F7
	.dc.b	$F0,$75,$CA,$BF,$21,$7D,$9F,$C9,$61,$E2,$7F,$29,$8F,$99,$FC,$B6
	.dc.b	$5E,$A7,$F3,$19,$FB,$9F,$80,$7C,$3F,$88,$FE,$E3,$E1,$FF,$B8,$63
	.dc.b	$F2,$BC,$3F,$D2,$DF,$F5,$F7,$C2,$7C,$FA,$FA,$31,$F7,$F8,$FE,$AB
	.dc.b	$F8,$1E,$4F,$B2,$FE,$0F,$97,$EE,$BF,$85,$E6,$FC,$0B,$F8,$7E,$7F
	.dc.b	$C2,$BF,$89,$E8,$F8,$2F,$E2,$F8,$FE,$4F,$AF,$FE,$42,$F2,$FF,$80
	.dc.b	$4F,$E5,$48,$7F,$15,$F2,$FF,$34,$FF,$17,$EE,$FC,$F3,$FC,$67,$BF
	.dc.b	$F4,$4F,$F1,$BF,$03,$F4,$CF,$EF,$7C,$1F,$D5,$3F,$9B,$F0,$FF,$5C
	.dc.b	$FE,$77,$C5,$FD,$93,$F9,$FF,$1F,$F6,$CF,$EF,$FB,$D1,$A8,$FE,$11
	.dc.b	$3C,$9F,$E1,$55,$D9,$F0,$1F,$E4,$3F,$23,$F8,$9B,$FE,$45,$F2,$7F
	.dc.b	$8B,$BF,$E4,$7F,$2B,$F8,$DB,$FE,$49,$F2,$FF,$CC,$1B,$FE,$DB,$C3
	.dc.b	$FF,$30,$79,$1F,$C9,$8C,$6D,$FE,$52,$77,$7C,$4F,$E0,$27,$53,$F8
	.dc.b	$29,$E1,$3F,$84,$9E,$53,$F7,$D1,$E9,$9F,$E4,$D9,$FB,$5F,$E5,$1A
	.dc.b	$7C,$17,$F9,$56,$BF,$1B,$BB,$F9,$1B,$9B,$FE,$53,$5C,$5F,$E5,$77
	.dc.b	$D5,$FE,$5A,$7D,$6B,$FF,$3E,$70,$7D,$A7,$FC,$FD,$C3,$F7,$9F,$F4
	.dc.b	$07,$17,$E0,$9F,$F4,$16,$BF,$0B,$4F,$E6,$9C,$7F,$8A,$7F,$8E,$1F
	.dc.b	$67,$97,$F6,$79,$7F,$1E,$8F,$DD,$1F,$95,$3F,$7C,$7E,$64,$FE,$04
	.dc.b	$7E,$74,$FE,$0C,$7E,$84,$FE,$14,$7E,$94,$FE,$1C,$7E,$A4,$FE,$51
	.dc.b	$E2,$FE,$B5,$FE,$35,$07,$FA,$57,$7F,$F1,$BE,$FD,$47,$EE,$8F,$EA
	.dc.b	$9D,$DF,$B2,$7F,$AB,$77,$7E,$C9,$FE,$B1,$DD,$FB,$27,$FA,$D7,$77
	.dc.b	$EC,$9F,$EB,$9D,$DF,$B2,$7F,$AF,$77,$7E,$C9,$FE,$30,$7E,$CC,$FF
	.dc.b	$1E,$79,$1F,$15,$C9,$38,$7F,$B4,$78,$DF,$C6,$5F,$ED,$5E,$37,$F1
	.dc.b	$97,$FB,$67,$8D,$FC,$65,$FE,$DD,$E3,$7F,$19,$7F,$B8,$78,$DF,$C6
	.dc.b	$5F,$EE,$5E,$37,$F1,$97,$FB,$A7,$27,$C2,$5C,$9F,$1F,$B9,$E4,$7F
	.dc.b	$40,$D7,$FC,$CA,$7F,$1C,$3F,$8F,$F9,$1F,$D0,$AF,$9C,$7D,$33,$F9
	.dc.b	$39,$F5,$CF,$E5,$67,$DB,$3F,$97,$9F,$7C,$FE,$66,$7E,$0C,$FE,$6E
	.dc.b	$7E,$1C,$FF,$04,$6D,$F8,$C7,$F1,$C7,$CA,$3F,$27,$E0,$FE,$57,$07
	.dc.b	$57,$E5,$6F,$F3,$35,$FB,$73,$F8,$E1,$FB,$73,$F8,$E1,$FB,$73,$F8
	.dc.b	$E1,$FB,$73,$F8,$69,$FE,$19,$F2,$7F,$C0,$EF,$F7,$7F,$2B,$FC,$16
	.dc.b	$FF,$79,$F2,$FF,$C2,$2F,$F9,$0B,$DD,$FF,$59,$BF,$E2,$1F,$17,$F7
	.dc.b	$0F,$F1,$FF,$17,$F7,$0F,$F2,$1F,$17,$F7,$0F,$F2,$3E,$1F,$DC,$1F
	.dc.b	$EE,$2F,$17,$F7,$0F,$FB,$93,$C5,$FD,$C3,$FE,$E6,$F1,$7F,$70,$FF
	.dc.b	$BA,$3C,$5F,$89,$C1,$FB,$7F,$2D,$DB,$F7,$F9,$9F,$13,$FE,$03,$26
	.dc.b	$9F,$DF,$DA,$B9,$3F,$0E,$FF,$8D,$67,$8B,$B2,$1B,$CB,$B5,$72,$AE
	.dc.b	$3C,$AB,$E3,$7B,$3A,$56,$EF,$8C,$74,$BE,$35,$C1,$B5,$F4,$27,$95
	.dc.b	$BD,$4F,$F3,$94,$7D,$13,$96,$1D,$1D,$E1,$CE,$31,$87,$49,$D9,$E4
	.dc.b	$2D,$FA,$70,$F3,$46,$4E,$6E,$23,$51,$C2,$FD,$F9,$7D,$28,$F9,$5F
	.dc.b	$15,$FE,$33,$23,$FE,$2F,$F1,$FF,$C6,$DF,$03,$FE,$37,$F8,$2F,$F1
	.dc.b	$C7,$BB,$FE,$39,$F8,$2F,$F1,$D7,$C3,$7F,$8E,$FA,$BF,$C7,$9F,$09
	.dc.b	$FE,$3D,$E3,$FF,$1F,$7C,$0F,$F8,$FF,$97,$FC,$81,$F0,$BF,$E4,$1F
	.dc.b	$85,$FF,$21,$7C,$37,$C9,$E6,$FF,$20,$F0,$FE,$FC,$FE,$EC,$7F,$90
	.dc.b	$63,$83,$FE,$43,$AE,$8F,$F9,$17,$0F,$03,$FE,$47,$CB,$C8,$FF,$92
	.dc.b	$74,$F4,$3F,$E4,$BD,$BD,$8F,$F9,$37,$83,$FC,$A9,$C3,$F0,$A3,$EA
	.dc.b	$F1,$7F,$BD,$F8,$FE,$47,$2F,$E2,$63,$5B,$7E,$11,$E3,$5C,$CF,$F2
	.dc.b	$C7,$27,$C8,$7F,$CB,$5C,$9F,$21,$FF,$2E,$72,$7C,$87,$FC,$BD,$C9
	.dc.b	$F2,$1F,$F3,$07,$27,$C8,$7F,$CC,$5C,$9F,$21,$F8,$A7,$F2,$63,$FA
	.dc.b	$86,$5F,$29,$CB,$F9,$8E,$4F,$93,$1F,$90,$E4,$F9,$2F,$E4,$B9,$3E
	.dc.b	$4B,$F9,$4E,$4F,$92,$FE,$5B,$93,$E4,$BF,$98,$E4,$F9,$2F,$E6,$B9
	.dc.b	$3E,$0B,$F2,$0F,$F3,$C7,$0F,$C9,$DB,$F3,$BC,$F2,$7F,$31,$AF,$E7
	.dc.b	$DF,$CC,$9F,$97,$E7,$FC,$D7,$3F,$E9,$63,$75,$C7,$97,$F3,$B1,$C0
	.dc.b	$97,$CC,$93,$BA,$7C,$37,$F8,$6D,$B5,$DC,$87,$29,$E1,$87,$88,$91
	.dc.b	$AB,$EA,$F9,$0F,$D9,$73,$FE,$1A,$FF,$86,$C7,$F8,$0C,$FD,$B1,$71
	.dc.b	$F4,$B6,$36,$FF,$05,$31,$7C,$5A,$53,$99,$B2,$18,$24,$F6,$72,$D7
	.dc.b	$33,$E0,$7F,$96,$63,$F1,$4F,$E6,$2F,$80,$F1,$4F,$12,$4F,$16,$D7
	.dc.b	$96,$7A,$B6,$33,$1E,$96,$BD,$27,$91,$27,$E4,$77,$3C,$CF,$4C,$E6
	.dc.b	$B9,$35,$E5,$5D,$E3,$11,$E9,$21,$CA,$34,$47,$9B,$5E,$67,$9F,$D9
	.dc.b	$FF,$42,$BF,$AE,$D3,$FC,$31,$EA,$7F,$AB,$F2,$FF,$59,$4F,$EC,$27
	.dc.b	$B3,$D0,$87,$42,$1E,$43,$E3,$9A,$31,$1F,$19,$B8,$7B,$F2,$F8,$91
	.dc.b	$C0,$E6,$78,$0A,$64,$E3,$8F,$B5,$A7,$73,$81,$E0,$C7,$DE,$FF,$0D
	.dc.b	$26,$BF,$E8,$47,$FB,$79,$C8,$EF,$87,$9D,$E6,$43,$C4,$57,$F0,$37
	.dc.b	$EA,$78,$46,$88,$E0,$98,$7D,$C8,$E4,$D7,$F3,$7A,$7E,$65,$F2,$9A
	.dc.b	$C7,$E0,$34,$F8,$07,$69,$E5,$5D,$A7,$F2,$3A,$7F,$02,$AF,$E0,$19
	.dc.b	$D7,$F8,$31,$FC,$06,$BF,$89,$E5,$FC,$11,$B3,$B8,$C6,$1F,$C6,$E7
	.dc.b	$35,$C6,$3F,$89,$BF,$E0,$AE,$DF,$F0,$4D,$7F,$84,$32,$FE,$23,$3E
	.dc.b	$0B,$E2,$4B,$E8,$4C,$B7,$7E,$06,$E3,$E9,$DB,$F8,$B4,$63,$0F,$13
	.dc.b	$70,$E8,$DC,$3A,$4F,$C2,$36,$7E,$7C,$FF,$4F,$9F,$DE,$4F,$F1,$01
	.dc.b	$E7,$78,$B6,$3A,$11,$E9,$3E,$07,$FC,$77,$5E,$67,$35,$CB,$0F,$64
	.dc.b	$62,$F9,$3D,$4E,$7A,$FE,$6A,$7F,$B2,$47,$E6,$B1,$FF,$48,$4E,$5F
	.dc.b	$AC,$48,$F6,$9C,$C9,$3E,$06,$C7,$CB,$DF,$FC,$B9,$F0,$DF,$CF,$EB
	.dc.b	$FC,$8E,$D9,$CD,$7F,$09,$3F,$CB,$5E,$E6,$5F,$F2,$E7,$C3,$DD,$5F
	.dc.b	$C0,$3F,$C3,$47,$C4,$E5,$7C,$05,$F8,$9C,$7C,$16,$7E,$27,$5F,$09
	.dc.b	$BF,$89,$E1,$F0,$DC,$7E,$27,$97,$C4,$73,$F8,$9F,$07,$47,$E4,$B9
	.dc.b	$7F,$CC,$FD,$7F,$92,$E5,$FF,$3C,$75,$EF,$97,$D5,$C9,$E6,$CB,$A4
	.dc.b	$63,$8D,$AF,$86,$31,$BF,$95,$D7,$0F,$A5,$E1,$C7,$ED,$E4,$F7,$47
	.dc.b	$93,$97,$CB,$CD,$E6,$E7,$93,$FE,$B3,$D7,$E3,$FC,$57,$C5,$24,$78
	.dc.b	$FD,$7F,$91,$E7,$FC,$9F,$86,$F9,$5E,$8F,$CB,$F8,$7F,$99,$E9,$FC
	.dc.b	$DF,$88,$F9,$DE,$AF,$CF,$F8,$9F,$A9,$F0,$BF,$E9,$2D,$3F,$CF,$EB
	.dc.b	$DC,$CF,$93,$AF,$73,$7E,$D7,$CC,$ED,$FE,$29,$87,$08,$D3,$F2,$B0
	.dc.b	$5F,$6C,$3B,$F7,$78,$6B,$E6,$67,$E5,$7E,$76,$BE,$97,$E8,$6F,$ED
	.dc.b	$7E,$97,$0F,$E1,$F8,$BE,$1C,$7F,$A5,$BB,$5C,$9F,$E3,$58,$72,$FC
	.dc.b	$A7,$ED,$47,$0F,$03,$FD,$53,$5F,$ED,$DD,$3F,$DF,$0B,$1F,$05,$F5
	.dc.b	$CB,$C8,$F7,$CF,$D0,$F8,$74,$F6,$3F,$2B,$0F,$80,$FC,$3D,$BE,$2B
	.dc.b	$F1,$CF,$C7,$F8,$3F,$95,$9F,$EE,$31,$FC,$3F,$3F,$E4,$CF,$CB,$E1
	.dc.b	$E3,$1F,$EC,$0E,$1F,$9B,$3F,$3F,$87,$E6,$BF,$97,$E4,$FB,$EF,$E6
	.dc.b	$79,$7F,$06,$FE,$6F,$9B,$F0,$EF,$E7,$79,$FF,$16,$FE,$34,$F8,$E4
	.dc.b	$1E,$57,$FA,$23,$8F,$E7,$C6,$73,$F9,$7C,$9F,$56,$3E,$B9,$D4,$FB
	.dc.b	$67,$84,$FB,$E7,$94,$FC,$19,$E9,$3F,$15,$B7,$CA,$7F,$2B,$C1,$F0
	.dc.b	$A1,$1F,$7B,$66,$FF,$95,$8F,$DD,$F9,$9F,$E1,$B3,$FC,$39,$C3,$F3
	.dc.b	$E3,$84,$F3,$79,$61,$DF,$2F,$9D,$8F,$DC,$7F,$0F,$C5,$F8,$07,$F1
	.dc.b	$3C,$7F,$84,$7F,$17,$C9,$F8,$87,$F1,$BC,$BF,$0F,$1F,$CB,$6D,$F9
	.dc.b	$8E,$0F,$F5,$8E,$EF,$73,$F6,$64,$AF,$D5,$7B,$9F,$B6,$F8,$9F,$DB
	.dc.b	$FB,$DF,$B8,$F8,$5F,$DC,$FA,$3F,$BA,$F8,$6F,$DD,$FC,$37,$EF,$33
	.dc.b	$FD,$EF,$C3,$7E,$FB,$E1,$FF,$7F,$F0,$FF,$C0,$3E,$23,$F8,$0F,$C4
	.dc.b	$7F,$02,$F8,$9F,$E0,$7F,$13,$FC,$DF,$D2,$FE,$4D,$C5,$F3,$63,$4F
	.dc.b	$EF,$6B,$93,$FB,$FC,$3B,$3F,$C0,$72,$F1,$3F,$C0,$F4,$F3,$3F,$C1
	.dc.b	$76,$F5,$3F,$C1,$F8,$3D,$CF,$C0,$3F,$85,$71,$FC,$37,$F2,$3E,$C7
	.dc.b	$ED,$B9,$BF,$75,$D9,$67,$F3,$07,$F0,$3E,$8F,$A5,$1F,$BC,$EA,$FA
	.dc.b	$CF,$EF,$BA,$FE,$D3,$FC,$03,$B3,$EF,$3F,$8C,$F1,$DC,$FF,$C9,$7B
	.dc.b	$7F,$61,$E3,$FF,$24,$CF,$F8,$F7,$95,$FC,$91,$FD,$DF,$95,$FC,$92
	.dc.b	$BF,$6D,$E1,$7E,$37,$1E,$7E,$17,$E3,$73,$ED,$9F,$57,$F8,$8F,$93
	.dc.b	$FB,$8C,$BF,$8E,$79,$3F,$B8,$7F,$8F,$79,$3F,$B8,$7F,$90,$79,$3F
	.dc.b	$B8,$7F,$91,$79,$5F,$11,$FE,$43,$E5,$7E,$D9,$F9,$26,$39,$B5,$AF
	.dc.b	$F2,$58,$E2,$7E,$E0,$EA,$7E,$E0,$F0,$9F,$B8,$3C,$A7,$EE,$0F,$E6
	.dc.b	$59,$7F,$12,$7F,$A5,$F1,$7C,$77,$FA,$6F,$17,$C7,$5C,$3F,$C7,$9F
	.dc.b	$F0,$D1,$FE,$30,$3E,$59,$FC,$84,$F9,$E7,$F2,$B3,$F9,$01,$FC,$18
	.dc.b	$FE,$10,$7F,$33,$3F,$86,$1F,$CD,$CF,$E2,$07,$F3,$B3,$F8,$A1,$F8
	.dc.b	$73,$F1,$2F,$F1,$A7,$F9,$07,$91,$FC,$7F,$1F,$8F,$97,$CB,$AF,$E9
	.dc.b	$39,$FD,$17,$FA,$5E,$9F,$55,$FE,$9B,$AF,$D9,$7F,$A7,$ED,$F7,$5F
	.dc.b	$EA,$3A,$FC,$77,$F1,$B9,$FB,$4F,$8E,$7C,$13,$E3,$EF,$FE,$18,$7F
	.dc.b	$8E,$9F,$D1,$DC,$3F,$D3,$4E,$27,$F2,$83,$E3,$9F,$D5,$4F,$8F,$1F
	.dc.b	$C7,$1F,$8E,$FF,$1C,$7E,$3B,$FC,$71,$F8,$EF,$F1,$C7,$E3,$C7,$EC
	.dc.b	$5F,$E3,$A7,$F8,$67,$83,$F8,$E3,$FB,$A2,$1F,$E0,$EF,$23,$F7,$EF
	.dc.b	$F8,$43,$AF,$F8,$1B,$FE,$12,$EC,$FE,$0E,$FF,$85,$3B,$7F,$85,$BF
	.dc.b	$E1,$6E,$EF,$E1,$EF,$F8,$63,$BF,$F8,$9B,$FE,$1A,$F0,$3F,$8B,$BF
	.dc.b	$E1,$CF,$07,$F8,$E7,$C2,$C8,$FF,$0B,$46,$B6,$FE,$D9,$5C,$9F,$ED
	.dc.b	$D7,$D9,$FE,$E1,$87,$89,$FE,$E5,$8F,$99,$FE,$E9,$97,$A9,$FE,$ED
	.dc.b	$9F,$B9,$FE,$F1,$A7,$F1,$B7,$FB,$D7,$85,$FE,$17,$EA,$7F,$00,$BE
	.dc.b	$2F,$FC,$D0,$DD,$FF,$CA,$AD,$7F,$06,$7F,$87,$9F,$F5,$90,$F7,$B8
	.dc.b	$FC,$57,$2F,$80,$E7,$F1,$5D,$3E,$0B,$AF,$C5,$76,$F8,$4E,$FC,$FE
	.dc.b	$0F,$85,$F0,$F9,$9E,$2F,$86,$F1,$FC,$57,$93,$E2,$3F,$A3,$F0,$7F
	.dc.b	$9A,$3D,$0F,$E8,$FC,$7F,$CA,$3B,$75,$3E,$8B,$E4,$FA,$B1,$EC,$FB
	.dc.b	$33,$F1,$3E,$ED,$7C,$CF,$C0,$DF,$D4,$FC,$2E,$1F,$8B,$C5,$F1,$BE
	.dc.b	$23,$E3,$F2,$CA,$FF,$35,$72,$BE,$2B,$E5,$92,$3E,$5F,$C5,$7C,$EF
	.dc.b	$3F,$E7,$FB,$9F,$43,$DD,$FA,$3E,$F7,$FA,$67,$F8,$03,$FC,$1F,$87
	.dc.b	$FA,$19,$FF,$4C,$B7,$E0,$75,$F1,$5B,$F8,$3E,$1F,$15,$C7,$E1,$79
	.dc.b	$7C,$57,$3F,$87,$F9,$38,$FF,$20,$EC,$7C,$37,$68,$FC,$C1,$DE,$78
	.dc.b	$BE,$1B,$EA,$F8,$F1,$F0,$BE,$33,$CD,$F0,$FE,$97,$D6,$7E,$2B,$7F
	.dc.b	$7B,$F8,$CE,$1F,$C7,$F1,$7C,$58,$FC,$6E,$7F,$94,$E3,$FC,$B7,$27
	.dc.b	$CB,$9F,$C9,$C6,$FC,$77,$A1,$FD,$A6,$98,$7C,$DB,$FC,$CF,$37,$CD
	.dc.b	$7F,$37,$CD,$E5,$7C,$E7,$A4,$F0,$F3,$39,$7E,$64,$7D,$1E,$5F,$98
	.dc.b	$FD,$3E,$5F,$98,$FE,$F0,$FE,$4A,$43,$E5,$9D,$39,$3E,$5B,$DB,$93
	.dc.b	$E5,$BE,$0E,$4F,$96,$F8,$B9,$BF,$05,$CB,$E7,$8F,$8B,$BB,$5F,$64
	.dc.b	$63,$6F,$80,$EB,$7F,$8D,$C1,$F1,$78,$BE,$5F,$3F,$E5,$EB,$9B,$F9
	.dc.b	$B7,$F7,$FD,$3F,$BA,$7F,$80,$F8,$FF,$BA,$7F,$81,$F8,$FF,$BA,$7F
	.dc.b	$82,$F8,$FF,$BA,$7F,$83,$F8,$FF,$BA,$7F,$84,$F8,$FF,$BA,$7F,$85
	.dc.b	$F8,$FF,$C8,$DF,$BF,$E3,$FE,$E8,$78,$9F,$BA,$8F,$7F,$8D,$FB,$A7
	.dc.b	$E0,$F8,$DF,$BA,$7E,$1C,$FE,$E9,$F8,$BE,$37,$EE,$9F,$8F,$E3,$7E
	.dc.b	$E9,$F9,$3E,$37,$EE,$9F,$E3,$BE,$37,$C3,$F0,$BF,$17,$1F,$C7,$78
	.dc.b	$FE,$3B,$FC,$89,$D7,$2F,$E3,$EB,$94,$F4,$3E,$39,$FC,$24,$F8,$E7
	.dc.b	$F0,$D3,$E3,$9F,$C4,$4F,$8E,$7F,$15,$3E,$39,$FC,$64,$F8,$E7,$E3
	.dc.b	$4F,$8F,$97,$C9,$BF,$F5,$A7,$8F,$F1,$DF,$4F,$89,$FC,$F3,$6F,$91
	.dc.b	$18,$EB,$6F,$F2,$63,$18,$7C,$97,$5B,$FC,$97,$86,$FF,$25,$E5,$BF
	.dc.b	$C0,$7A,$6B,$F1,$35,$F9,$07,$F2,$2E,$8F,$C8,$BF,$D1,$DF,$DF,$BF
	.dc.b	$C0,$4F,$DA,$9F,$C1,$4F,$DA,$9F,$C7,$CF,$D0,$9F,$B5,$9F,$6F,$89
	.dc.b	$FB,$58,$FE,$24,$7E,$D4,$FE,$2C,$7E,$D4,$FE,$34,$7C,$98,$F8,$BE
	.dc.b	$3F,$FA,$D6,$3F,$91,$C7,$4F,$0B,$F6,$AF,$6F,$0B,$F6,$AF,$83,$C2
	.dc.b	$FD,$AB,$E2,$F0,$BC,$B7,$FC,$20,$FE,$3E,$7E,$DB,$C7,$FE,$3E,$FE
	.dc.b	$E3,$C7,$F8,$2F,$EE,$B4,$FE,$41,$EC,$7D,$48,$F8,$C7,$F1,$E2,$1F
	.dc.b	$BB,$3E,$41,$FB,$F3,$E4,$1F,$C1,$3C,$4F,$E0,$D8,$FC,$6F,$17,$FA
	.dc.b	$56,$5E,$8E,$AF,$F4,$77,$03,$5F,$6C,$63,$6F,$82,$EB,$7F,$8A,$FF
	.dc.b	$A5,$78,$3E,$46,$9F,$2E,$34,$FC,$F3,$F8,$E1,$F4,$CF,$E3,$87,$F5
	.dc.b	$9F,$23,$C6,$FF,$30,$3E,$41,$F8,$03,$E4,$1F,$84,$3E,$41,$FD,$A0
	.dc.b	$F8,$7A,$7E,$6F,$7F,$DA,$C7,$F1,$A3,$F6,$C7,$EF,$0F,$E3,$67,$F7
	.dc.b	$93,$F6,$A7,$F0,$43,$F6,$A7,$F0,$83,$F6,$A7,$F0,$C3,$F6,$A7,$F1
	.dc.b	$03,$F6,$A7,$F1,$43,$F6,$A7,$ED,$0F,$F0,$A7,$37,$ED,$6F,$FD,$79
	.dc.b	$DB,$FC,$72,$77,$7F,$B3,$EA,$FD,$AD,$7F,$59,$E9,$FD,$A3,$FD,$6F
	.dc.b	$0F,$DA,$3F,$D7,$7A,$7F,$68,$FF,$5F,$E9,$FD,$A3,$FD,$87,$A7,$F6
	.dc.b	$8F,$F1,$5C,$BE,$19,$F1,$4F,$F5,$9F,$5B,$4F,$E6,$18,$F9,$E5,$39
	.dc.b	$8C,$36,$3D,$A4,$3E,$01,$34,$FE,$23,$5D,$23,$51,$8A,$E7,$39,$9E
	.dc.b	$17,$F1,$A7,$24,$8E,$31,$96,$9E,$69,$E5,$5D,$63,$37,$E7,$7A,$9B
	.dc.b	$BF,$F3,$07,$1B,$E2,$BF,$C5,$C4,$9F,$F1,$77,$C4,$FF,$8C,$7E,$2B
	.dc.b	$FC,$65,$ED,$FF,$8C,$FD,$AF,$F1,$A7,$C1,$FF,$8D,$7D,$BF,$F1,$B7
	.dc.b	$C1,$FF,$8D,$FD,$CF,$F1,$C7,$C1,$FF,$8E,$7D,$DF,$F1,$D7,$C1,$FF
	.dc.b	$8E,$FD,$EF,$F1,$DF,$1F,$F3,$6F,$E0,$1F,$F7,$97,$07,$D0,$5F,$05
	.dc.b	$FB,$08,$FF,$6A,$1F,$E3,$7F,$80,$79,$9C,$BA,$53,$C6,$B0,$F3,$8F
	.dc.b	$03,$FE,$49,$CB,$C8,$FF,$92,$F4,$F4,$3F,$E4,$DD,$BD,$8F,$F9,$3F
	.dc.b	$83,$E0,$3F,$E5,$1E,$2F,$88,$FC,$63,$FC,$A9,$CB,$F2,$5F,$C6,$F3
	.dc.b	$7C,$D7,$24,$3F,$C5,$BC,$DC,$9F,$F1,$7D,$76,$7F,$C6,$37,$E2,$7F
	.dc.b	$C6,$78,$79,$9F,$F1,$AE,$3E,$A7,$FC,$6F,$97,$B9,$FF,$24,$EF,$F3
	.dc.b	$3A,$0F,$CC,$8F,$E3,$07,$CC,$E1,$FE,$86,$FC,$F9,$DE,$FF,$29,$E3
	.dc.b	$BF,$5A,$FE,$A3,$87,$85,$FE,$A7,$8F,$95,$FE,$AB,$97,$A5,$FE,$AF
	.dc.b	$9F,$B5,$FE,$B3,$A7,$C1,$7F,$AD,$EB,$F2,$5F,$EB,$BB,$FF,$AD,$7A
	.dc.b	$7F,$3B,$9F,$E8,$76,$FC,$DE,$BF,$9D,$BF,$D0,$46,$2B,$6F,$E7,$CF
	.dc.b	$D1,$B8,$9E,$15,$CE,$B8,$91,$C1,$2B,$91,$D7,$A7,$F6,$8D,$3A,$B8
	.dc.b	$9E,$51,$C1,$B5,$FB,$3F,$23,$FA,$F3,$FB,$59,$FF,$4E,$B5,$F7,$18
	.dc.b	$C3,$F9,$43,$FE,$95,$8D,$3E,$43,$57,$FE,$99,$78,$90,$D1,$63,$89
	.dc.b	$27,$AC,$72,$39,$46,$67,$A4,$72,$24,$65,$DC,$F2,$7B,$35,$DB,$63
	.dc.b	$84,$7A,$E3,$04,$9E,$F3,$A9,$F2,$1B,$8C,$CE,$E3,$F8,$0F,$C1,$BD
	.dc.b	$DF,$0D,$68,$93,$E6,$6B,$CC,$C1,$0C,$10,$F0,$90,$EA,$6C,$C1,$EE
	.dc.b	$7A,$B8,$BF,$53,$C0,$E8,$53,$57,$EB,$72,$73,$3B,$1C,$6F,$DC,$E4
	.dc.b	$DC,$F6,$70,$74,$AF,$1B,$DC,$86,$8B,$1C,$49,$5E,$38,$E4,$78,$8E
	.dc.b	$04,$8D,$CF,$68,$F2,$57,$9A,$73,$1C,$5A,$FA,$2B,$B3,$E2,$6B,$D5
	.dc.b	$B1,$C2,$3C,$67,$1A,$F0,$46,$63,$94,$62,$78,$9F,$36,$BC,$8F,$F0
	.dc.b	$32,$47,$EE,$72,$FE,$01,$7F,$C1,$AF,$F7,$51,$8A,$DC,$7F,$00,$21
	.dc.b	$FC,$0C,$99,$7F,$00,$8D,$C6,$27,$95,$F7,$8F,$E1,$B9,$7F,$04,$AD
	.dc.b	$DF,$1A,$D6,$1E,$32,$4E,$B0,$E7,$87,$8A,$F8,$36,$32,$F8,$4F,$59
	.dc.b	$2B,$9D,$F6,$BE,$93,$E4,$F0,$38,$E1,$D2,$39,$5F,$07,$B1,$CE,$B8
	.dc.b	$C6,$63,$71,$82,$5F,$72,$47,$29,$FE,$4D,$7F,$DF,$1F,$EB,$39,$7F
	.dc.b	$53,$E4,$34,$C7,$C1,$E7,$E1,$35,$EE,$EF,$DE,$E1,$F1,$5C,$7E,$03
	.dc.b	$97,$C5,$73,$E3,$E9,$F1,$5D,$7E,$0F,$B7,$81,$DF,$E1,$3C,$1F,$15
	.dc.b	$E1,$F8,$6F,$17,$C5,$78,$FE,$22,$73,$7F,$84,$BA,$7F,$C3,$93,$B8
	.dc.b	$C7,$57,$AF,$99,$8F,$59,$C9,$FE,$21,$CF,$C4,$E3,$9F,$FB,$CE,$DE
	.dc.b	$78,$C5,$B8,$3D,$91,$8E,$2F,$80,$EB,$93,$E2,$3C,$39,$BE,$43,$CB
	.dc.b	$9F,$FC,$52,$FF,$99,$30,$7C,$57,$CF,$24,$78,$3C,$6F,$F0,$27,$82
	.dc.b	$F7,$AF,$BB,$8F,$85,$CF,$A1,$AF,$85,$DF,$A5,$C3,$E1,$78,$FA,$9C
	.dc.b	$BE,$17,$9F,$8B,$D3,$E1,$7A,$F8,$DD,$BE,$17,$E7,$F0,$FE,$17,$8D
	.dc.b	$DD,$98,$F8,$DC,$5F,$EC,$95,$DE,$EF,$FC,$54,$B2,$EF,$5C,$34,$F1
	.dc.b	$BC,$B6,$F3,$BD,$38,$3D,$6F,$6E,$2F,$7B,$E0,$E4,$FC,$A7,$2F,$C6
	.dc.b	$8F,$F1,$8F,$37,$CC,$F2,$7F,$BF,$F0,$7C,$C9,$FB,$33,$BD,$FF,$CC
	.dc.b	$57,$F3,$30,$EB,$5E,$AD,$FD,$39,$78,$A3,$07,$FA,$5B,$76,$9E,$88
	.dc.b	$CF,$72,$5B,$7B,$A7,$30,$DF,$E1,$C6,$38,$3E,$8B,$F3,$38,$BF,$6A
	.dc.b	$FF,$8C,$29,$1A,$8E,$07,$D2,$E4,$FB,$8F,$D4,$E5,$FC,$03,$F5,$B9
	.dc.b	$BF,$08,$FD,$8E,$7F,$C4,$3F,$6B,$A3,$E1,$3F,$8F,$3E,$39,$0F,$CB
	.dc.b	$72,$7C,$9E,$1F,$D6,$3F,$4A,$3E,$A1,$CC,$FB,$07,$73,$EE,$1E,$33
	.dc.b	$F0,$07,$9C,$FC,$21,$EB,$3F,$1B,$A7,$BF,$93,$F1,$91,$F4,$74,$F8
	.dc.b	$B1,$FE,$3E,$D7,$E2,$CF,$E5,$3A,$3F,$38,$F3,$3E,$8E,$FF,$4E,$B9
	.dc.b	$F0,$FD,$7E,$1F,$B5,$18,$C3,$EE,$DF,$A4,$FB,$3C,$7F,$86,$9D,$72
	.dc.b	$7C,$19,$FC,$89,$F1,$F9,$FF,$C9,$F6,$F6,$3F,$72,$49,$FD,$4F,$C2
	.dc.b	$FE,$F3,$DA,$FD,$EF,$C3,$7E,$FB,$E1,$3F,$7F,$F0,$FF,$C0,$3E,$1B
	.dc.b	$F8,$0F,$C3,$7F,$A1,$7F,$80,$3F,$84,$E5,$F9,$05,$F0,$F9,$F8,$7D
	.dc.b	$7C,$1E,$FE,$1F,$87,$C5,$71,$F8,$8F,$DE,$E3,$FE,$98,$F4,$3F,$95
	.dc.b	$BF,$E9,$AF,$23,$4F,$F2,$3C,$D7,$CE,$31,$8F,$77,$59,$F9,$5F,$39
	.dc.b	$FC,$1B,$5F,$5B,$FC,$23,$7F,$7B,$FC,$2B,$87,$F8,$EF,$17,$C5,$8F
	.dc.b	$D3,$F1,$FE,$EF,$B3,$F1,$F1,$F9,$3E,$8F,$F6,$D7,$7F,$D1,$7E,$9C
	.dc.b	$7E,$44,$FA,$E7,$D9,$7E,$E7,$8D,$F7,$BC,$1F,$E2,$BD,$BF,$83,$7F
	.dc.b	$8B,$F7,$FE,$D2,$BF,$05,$E0,$FE,$46,$7F,$8D,$F8,$3F,$22,$7F,$BA
	.dc.b	$1F,$D4,$31,$FD,$C1,$F9,$13,$F3,$27,$E4,$5F,$C6,$5F,$78,$FE,$15
	.dc.b	$87,$8D,$FE,$19,$8F,$F0,$FA,$FC,$97,$81,$F8,$FC,$7F,$95,$F8,$1E
	.dc.b	$F7,$E1,$DF,$E4,$0F,$DA,$9F,$15,$CD,$7F,$1E,$8F,$DA,$9F,$C4,$BC
	.dc.b	$2F,$E0,$2F,$F1,$4F,$0F,$F8,$2B,$FC,$5B,$C4,$FE,$12,$FF,$18,$F1
	.dc.b	$7F,$86,$BF,$C6,$BC,$6F,$D6,$BF,$C7,$38,$BF,$A2,$3F,$E1,$08,$FF
	.dc.b	$6D,$7B,$1F,$EA,$AC,$BF,$59,$C3,$F2,$EB,$F9,$57,$17,$CF,$D7,$E8
	.dc.b	$C7,$F0,$33,$F9,$3F,$0F,$C5,$F1,$FF,$96,$71,$7C,$5A,$FE,$6B,$5F
	.dc.b	$7F,$1F,$C0,$BF,$EF,$2F,$23,$F8,$B2,$BF,$F7,$9F,$2F,$E2,$E3,$1E
	.dc.b	$4F,$F4,$57,$E5,$5F,$E3,$1F,$E4,$47,$F3,$8E,$4F,$A4,$FF,$41,$E5
	.dc.b	$F1,$3F,$60,$FE,$38,$7A,$0F,$54,$FE,$01,$FC,$16,$5F,$D9,$35,$F8
	.dc.b	$59,$7F,$47,$DB,$F8,$DC,$FF,$98,$BC,$8F,$E3,$77,$F3,$4F,$C5,$9F
	.dc.b	$44,$FE,$4E,$7D,$53,$F9,$59,$F6,$4F,$E5,$E7,$DD,$3F,$99,$9F,$81
	.dc.b	$3F,$07,$87,$E1,$5F,$83,$3F,$8C,$F2,$7F,$8C,$F4,$FE,$CE,$BF,$92
	.dc.b	$3F,$B6,$AF,$DC,$1F,$E0,$2F,$27,$F8,$C3,$FE,$04,$F2,$7F,$8C,$3F
	.dc.b	$E0,$6F,$27,$F8,$23,$FC,$18,$FE,$19,$E4,$FF,$16,$7F,$87,$79,$3F
	.dc.b	$C5,$9F,$E2,$1E,$4F,$EB,$F4,$FF,$1D,$79,$3F,$C5,$B4,$FD,$85,$7F
	.dc.b	$21,$AE,$0F,$F2,$3B,$E8,$FF,$25,$C3,$F9,$47,$8B,$E2,$8F,$EE,$79
	.dc.b	$79,$9F,$EE,$B9,$FA,$9F,$EE,$FA,$7F,$35,$7F,$9B,$9F,$E2,$0D,$7F
	.dc.b	$8B,$6C,$EC,$FF,$34,$79,$0F,$8A,$FE,$6C,$4B,$FE,$55,$F1,$3F,$CE
	.dc.b	$7D,$BF,$E7,$5E,$E7,$F3,$BF,$88,$FE,$79,$EF,$7F,$3D,$F8,$8F,$E7
	.dc.b	$DD,$1F,$CF,$FE,$23,$FA,$07,$C0,$FF,$41,$F8,$8F,$E8,$5F,$07,$FD
	.dc.b	$0F,$E2,$3F,$A2,$7C,$2F,$F4,$5F,$88,$FF,$26,$F6,$FF,$5E,$FE,$00
	.dc.b	$FE,$D3,$C5,$FE,$8A,$6C,$7F,$62,$E4,$57,$2A,$FE,$81,$87,$67,$FA
	.dc.b	$16,$5E,$27,$FA,$26,$9E,$67,$FA,$36,$DE,$A7,$FA,$47,$07,$B9,$FE
	.dc.b	$95,$C5,$FD,$67,$8F,$E2,$47,$F5,$9E,$4F,$DA,$78,$4D,$DF,$11,$EA
	.dc.b	$F0,$BF,$D2,$91,$8B,$FF,$3B,$35,$FF,$18,$BF,$E9,$67,$3E,$DF,$17
	.dc.b	$F8,$14,$3E,$23,$18,$77,$F7,$6C,$BE,$22,$FB,$F8,$F8,$8C,$FC,$0E
	.dc.b	$BE,$23,$7F,$07,$C3,$E2,$38,$FC,$2F,$2F,$88,$E7,$F0,$FF,$99,$E0
	.dc.b	$FF,$5B,$7B,$57,$83,$DD,$3B,$9F,$5D,$F3,$7D,$B8,$F7,$7D,$F9,$F8
	.dc.b	$DF,$83,$AF,$A3,$6F,$4F,$07,$AB,$87,$D7,$C5,$EC,$F8,$7F,$6F,$2C
	.dc.b	$C4,$3C,$8F,$CD,$3F,$1A,$34,$FC,$C7,$93,$F3,$0F,$A3,$7E,$07,$E9
	.dc.b	$E3,$F2,$1F,$AB,$C9,$F2,$1F,$AF,$C9,$F2,$1F,$B3,$C9,$F2,$1F,$B7
	.dc.b	$C9,$F2,$1F,$F5,$7F,$C3,$FF,$BC,$BD,$7F,$F4,$8C,$7E,$03,$93,$E4
	.dc.b	$3F,$80,$E4,$F9,$0F,$E0,$B9,$3E,$43,$F8,$4E,$4F,$90,$FE,$1B,$93
	.dc.b	$E4,$3F,$8A,$E4,$F9,$0F,$B4,$FC,$5F,$1F,$C2,$F0,$FF,$55,$C7,$F9
	.dc.b	$6F,$0F,$F7,$D5,$FE,$25,$EB,$FC,$DD,$7E,$AF,$B3,$F7,$D3,$FE,$29
	.dc.b	$F2,$7F,$7C,$FF,$8A,$FC,$9F,$DF,$3F,$E2,$DF,$27,$F7,$CF,$F8,$BF
	.dc.b	$C9,$FD,$F3,$FE,$31,$F2,$7F,$7C,$FF,$8C,$FC,$9F,$DF,$3F,$A2,$F0
	.dc.b	$BF,$92,$1F,$C9,$9C,$1F,$E3,$9F,$27,$F7,$8F,$F9,$1F,$C9,$FD,$E3
	.dc.b	$FE,$49,$F2,$7F,$78,$FF,$92,$FC,$9F,$DE,$3F,$E4,$DF,$27,$F7,$8F
	.dc.b	$F9,$3F,$C9,$FD,$E3,$FE,$51,$F2,$7F,$78,$FC,$32,$1F,$8D,$F2,$BF
	.dc.b	$92,$F2,$7C,$6D,$2C,$FF,$24,$3F,$82,$9F,$1C,$FE,$12,$7C,$73,$F8
	.dc.b	$69,$F1,$CF,$E2,$27,$C7,$3F,$8A,$9F,$1C,$FE,$32,$7C,$73,$E1,$1F
	.dc.b	$10,$FE,$44,$FC,$A7,$FC,$D5,$AF,$CE,$7F,$CD,$9B,$7D,$27,$FC,$DD
	.dc.b	$BF,$D6,$7F,$CE,$1C,$1F,$69,$FF,$39,$70,$FD,$E7,$FC,$E9,$C5,$F8
	.dc.b	$27,$FC,$ED,$C7,$F0,$5F,$F0,$A7,$2F,$E2,$9F,$8C,$7E,$C4,$FE,$3B
	.dc.b	$CD,$F9,$1D,$3F,$90,$F3,$FE,$F2,$7F,$62,$7E,$64,$FD,$89,$FC,$10
	.dc.b	$FD,$89,$FC,$20,$FD,$89,$FC,$30,$FD,$89,$FC,$40,$FD,$89,$FA,$D3
	.dc.b	$F6,$27,$FC,$01,$EF,$FD,$A3,$FC,$BB,$C0,$FD,$C3,$FC,$C3,$C1,$FD
	.dc.b	$E3,$FC,$CB,$C2,$FE,$00,$FF,$34,$F0,$FF,$82,$3F,$CD,$BC,$4F,$E1
	.dc.b	$0F,$F3,$8F,$17,$D8,$FE,$C0,$F8,$24,$3E,$2F,$91,$FD,$17,$5F,$89
	.dc.b	$1F,$1C,$E0,$7F,$42,$F2,$3F,$76,$FF,$44,$F0,$7F,$7E,$FF,$01,$3E
	.dc.b	$29,$FC,$14,$F8,$A7,$F0,$93,$E2,$9F,$C3,$4F,$8A,$7F,$4E,$F1,$BF
	.dc.b	$9D,$7B,$9F,$E5,$DE,$1F,$E5,$78,$A3,$15,$FE,$4E,$E0,$F9,$3E,$1F
	.dc.b	$F7,$3E,$1F,$92,$FF,$75,$E1,$F9,$2F,$F7,$7E,$1F,$92,$FF,$79,$E1
	.dc.b	$F9,$2F,$F7,$BE,$1F,$92,$FF,$7D,$E1,$F9,$2F,$F7,$FE,$2F,$8B,$AF
	.dc.b	$C7,$D3,$F9,$FF,$93,$FC,$FA,$1C,$9F,$91,$BF,$F0,$17,$67,$E5,$67
	.dc.b	$3C,$DF,$99,$9F,$F0,$37,$3F,$E7,$5F,$F0,$47,$47,$E8,$5F,$F0,$57
	.dc.b	$4F,$E9,$5F,$F0,$67,$57,$EA,$5F,$F0,$77,$5F,$EB,$5F,$F0,$87,$67
	.dc.b	$EC,$5F,$F0,$B7,$47,$F1,$AD,$FF,$CE,$BE,$A7,$F7,$48,$D1,$FE,$2E
	.dc.b	$AE,$4F,$F8,$C2,$FB,$3F,$E3,$2C,$3C,$4F,$F8,$D3,$1F,$33,$FE,$36
	.dc.b	$CB,$D4,$FF,$8E,$33,$F7,$3F,$E3,$AD,$3E,$16,$FF,$B5,$E0,$F9,$1E
	.dc.b	$63,$E2,$3F,$C0,$E4,$DF,$FC,$26,$D7,$FC,$26,$DE,$0F,$F1,$93,$71
	.dc.b	$FF,$0D,$B6,$7F,$C2,$AD,$E1,$FF,$11,$36,$FC,$13,$D8,$FE,$35,$7F
	.dc.b	$ED,$2B,$FF,$12,$9B,$A9,$3B,$3C,$65,$9F,$1B,$67,$C6,$D3,$C6,$5B
	.dc.b	$F1,$B6,$FC,$6F,$ED,$1F,$EE,$AF,$F8,$CB,$AB,$FC,$56,$7F,$8B,$38
	.dc.b	$BF,$C5,$BE,$D7,$F8,$BB,$DB,$FF,$18,$FF,$00,$BD,$CB,$F0,$58,$F7
	.dc.b	$73,$F0,$5A,$F7,$F7,$F0,$5C,$3E,$07,$8F,$C1,$72,$F8,$3E,$7E,$CF
	.dc.b	$4F,$83,$FC,$47,$17,$F8,$7B,$9B,$E3,$3F,$E4,$2E,$6D,$39,$7F,$73
	.dc.b	$C1,$FB,$8E,$62,$C7,$BD,$6C,$FF,$66,$43,$3F,$2C,$66,$B8,$EB,$EA
	.dc.b	$8C,$53,$7F,$7C,$63,$87,$E2,$3F,$E5,$CE,$3F,$90,$FF,$97,$B9,$7E
	.dc.b	$5B,$FB,$AF,$3B,$FC,$C7,$18,$8F,$A0,$7F,$99,$39,$BE,$7B,$FE,$66
	.dc.b	$E6,$F9,$F8,$78,$21,$C0,$E0,$F8,$39,$F9,$E7,$31,$F1,$35,$F6,$C6
	.dc.b	$0F,$83,$BF,$C4,$71,$BF,$C8,$E0,$FA,$71,$F0,$38,$7E,$9B,$FD,$C0
	.dc.b	$C1,$F1,$38,$BE,$D3,$F1,$B8,$FE,$F3,$F2,$39,$3F,$04,$FC,$AE,$5F
	.dc.b	$C3,$3F,$33,$9B,$DA,$FB,$CF,$94,$7C,$33,$E5,$10,$FC,$C7,$47,$E9
	.dc.b	$9F,$D2,$93,$83,$F4,$AD,$DF,$F4,$AD,$3F,$4A,$5D,$7F,$4A,$DE,$8F
	.dc.b	$D2,$D7,$22,$57,$27,$E8,$4F,$E2,$B9,$BF,$90,$5B,$DA,$73,$7E,$9A
	.dc.b	$7C,$D7,$CA,$FC,$2D,$BF,$0B,$71,$F0,$E3,$DC,$95,$C1,$B5,$C1,$B3
	.dc.b	$C2,$7D,$BD,$5F,$C8,$33,$FF,$43,$C4,$3E,$09,$8B,$F8,$2D,$AF,$E0
	.dc.b	$AD,$9E,$0F,$CA,$30,$62,$78,$B9,$21,$D8,$AE,$9C,$11,$C1,$27,$FD
	.dc.b	$BB,$7F,$DD,$3C,$BF,$DE,$1F,$BD,$8C,$D7,$17,$CE,$F6,$6B,$9B,$F4
	.dc.b	$B4,$F4,$46,$AF,$C1,$3A,$6C,$EA,$31,$3B,$3F,$94,$F8,$AF,$27,$C1
	.dc.b	$3A,$25,$71,$21,$F4,$6F,$AB,$6F,$A4,$F0,$8C,$35,$F9,$E6,$0E,$66
	.dc.b	$0F,$B1,$8F,$FB,$93,$1F,$8F,$C7,$FB,$8D,$3F,$47,$5B,$3C,$E7,$22
	.dc.b	$9D,$48,$6C,$F9,$06,$0D,$9E,$4C,$FE,$13,$A3,$86,$5E,$E6,$E5,$EE
	.dc.b	$7C,$06,$CE,$64,$30,$7F,$3F,$3F,$9A,$61,$EC,$7F,$B4,$E3,$BA,$ED
	.dc.b	$16,$BB,$3E,$73,$9C,$69,$B1,$F6,$B3,$E1,$9F,$CF,$9A,$79,$AB,$81
	.dc.b	$2B,$AC,$7B,$27,$83,$FC,$0C,$FE,$95,$BF,$F0,$26,$EF,$FC,$0D,$FE
	.dc.b	$F8,$71,$3B,$57,$47,$C2,$7A,$6F,$C1,$7D,$9F,$C4,$E7,$F4,$1F,$E0
	.dc.b	$53,$FC,$14,$8E,$09,$5F,$C1,$5B,$5F,$C5,$32,$FE,$12,$DA,$E3,$5B
	.dc.b	$38,$76,$B8,$3F,$1D,$BC,$AF,$EC,$B1,$C7,$1E,$B1,$FE,$0F,$C3,$18
	.dc.b	$73,$69,$23,$FB,$21,$E0,$70,$FF,$33,$6E,$FF,$C9,$5B,$97,$A2,$7F
	.dc.b	$06,$2F,$FC,$F5,$8F,$96,$3F,$AE,$E1,$FC,$22,$7C,$B1,$D4,$F3,$10
	.dc.b	$C1,$30,$F0,$39,$E0,$FE,$58,$EE,$32,$EF,$0F,$23,$4F,$61,$B7,$31
	.dc.b	$FE,$68,$8F,$CE,$E1,$E7,$8C,$47,$1B,$F3,$C7,$62,$57,$22,$4F,$63
	.dc.b	$D3,$7F,$E8,$7C,$DF,$07,$FE,$38,$24,$FF,$87,$3D,$CF,$F1,$DF,$C2
	.dc.b	$7F,$8F,$3E,$17,$FC,$7B,$EF,$FF,$8F,$BE,$1B,$FC,$7F,$F0,$3F,$E4
	.dc.b	$0F,$87,$FF,$20,$FC,$1F,$F4,$CD,$FF,$B1,$F0,$32,$FE,$9D,$18,$5F
	.dc.b	$09,$BE,$1F,$F5,$7C,$3E,$17,$1B,$FD,$98,$7C,$46,$31,$E9,$F0,$F6
	.dc.b	$5F,$13,$FA,$BB,$FF,$0C,$F6,$7E,$BD,$BD,$99,$D7,$C5,$5C,$70,$E5
	.dc.b	$97,$3C,$FA,$6B,$D7,$1F,$69,$FE,$40,$DF,$E0,$BF,$E4,$2E,$1F,$8A
	.dc.b	$FF,$91,$38,$FE,$4B,$F2,$CF,$F2,$3F,$2F,$CE,$7F,$DC,$3B,$BE,$27
	.dc.b	$D4,$48,$FF,$28,$47,$FA,$EF,$8B,$E2,$F9,$DE,$FF,$7F,$E0,$7A,$1F
	.dc.b	$07,$E0,$3E,$17,$A5,$F0,$FE,$07,$E2,$7A,$9F,$17,$E0,$BE,$37,$AD
	.dc.b	$F1,$FE,$0F,$E4,$7B,$1F,$27,$E1,$3E,$57,$B5,$F1,$77,$FE,$DD,$E3
	.dc.b	$7E,$45,$FF,$35,$F9,$3A,$E3,$F2,$DF,$27,$CF,$8F,$67,$D3,$9F,$89
	.dc.b	$F5,$EB,$E6,$7D,$BB,$FA,$9F,$7F,$0F,$DB,$E2,$FB,$9E,$DF,$C3,$9F
	.dc.b	$F2,$9F,$27,$C7,$7F,$82,$F8,$5F,$8D,$E6,$D7,$81,$F8,$D7,$8C,$FC
	.dc.b	$AC,$3E,$43,$F3,$39,$FE,$43,$F3,$B9,$FE,$43,$F4,$39,$FE,$43,$F4
	.dc.b	$B9,$FE,$43,$F5,$39,$FE,$43,$F5,$B9,$FE,$43,$FC,$3F,$CD,$FE,$A8
	.dc.b	$FF,$9A,$BC,$8F,$E4,$CE,$CF,$92,$73,$3E,$49,$DC,$F9,$27,$8C,$F9
	.dc.b	$27,$9C,$F9,$27,$AC,$F9,$27,$BC,$FC,$4F,$07,$C4,$E8,$FD,$6E,$DF
	.dc.b	$E4,$0F,$27,$C7,$C3,$F2,$6F,$F1,$F3,$B7,$F6,$17,$F4,$DE,$A7,$D2
	.dc.b	$DF,$E4,$BF,$53,$7F,$92,$FD,$6D,$FE,$4B,$F6,$37,$F9,$2F,$DA,$DF
	.dc.b	$E4,$BF,$73,$7F,$93,$9F,$F3,$68,$FE,$42,$7F,$4A,$3E,$41,$F8,$2E
	.dc.b	$8F,$C8,$3F,$84,$E8,$FC,$83,$F8,$6E,$8F,$C8,$3F,$88,$E8,$FC,$83
	.dc.b	$F8,$AE,$8F,$C8,$3F,$8C,$E8,$F7,$3F,$8E,$E0,$F8,$59,$7C,$58,$FD
	.dc.b	$BF,$C2,$FF,$0F,$23,$FE,$C5,$FE,$01,$FE,$94,$3E,$1B,$F8,$B2,$F3
	.dc.b	$EF,$C3,$63,$E0,$33,$F0,$DA,$F8,$2D,$FC,$3F,$0F,$84,$E3,$F1,$1C
	.dc.b	$BD,$7E,$7F,$13,$D3,$E1,$FA,$FC,$57,$6F,$89,$FF,$47,$FA,$BF,$C7
	.dc.b	$FC,$8F,$E6,$3C,$29,$DF,$1F,$9F,$0E,$6F,$A7,$2E,$EF,$AF,$4F,$1B
	.dc.b	$ED,$DB,$CE,$FB,$F8,$3D,$6F,$C1,$E2,$F7,$F1,$FC,$0F,$17,$E2,$CF
	.dc.b	$EE,$39,$7F,$21,$E5,$7E,$4A,$3F,$1E,$43,$E8,$1F,$99,$3E,$A1,$F9
	.dc.b	$D3,$EC,$1F,$A1,$3E,$E1,$FA,$53,$F0,$07,$EA,$4F,$C2,$1F,$AD,$3F
	.dc.b	$10,$7E,$2B,$7F,$EA,$BE,$3F,$ED,$23,$FD,$BF,$E4,$7E,$E2,$73,$5F
	.dc.b	$C7,$E7,$89,$F9,$93,$F6,$E7,$E7,$4F,$DB,$9F,$A1,$3F,$6E,$7E,$94
	.dc.b	$FD,$B9,$FA,$93,$F5,$71,$FD,$8F,$C9,$FE,$2F,$1F,$C7,$BC,$8F,$E4
	.dc.b	$1E,$4F,$ED,$BC,$AF,$E4,$13,$93,$E4,$5F,$F6,$BE,$EF,$CC,$BF,$DB
	.dc.b	$7B,$FF,$3A,$FF,$6F,$F0,$3F,$42,$FF,$71,$F0,$7F,$4A,$FF,$73,$F0
	.dc.b	$BF,$52,$FF,$75,$F0,$FF,$5A,$FF,$77,$F1,$3E,$13,$F1,$32,$FD,$A9
	.dc.b	$F2,$3C,$BF,$DC,$E8,$F1,$BF,$7B,$5F,$DF,$FC,$7F,$E0,$2F,$F8,$07
	.dc.b	$C8,$FE,$0A,$FF,$80,$FC,$9F,$E1,$2F,$F8,$17,$CA,$FE,$1A,$FF,$81
	.dc.b	$FC,$BF,$E7,$39,$FF,$42,$E5,$F8,$F3,$FC,$FE,$3F,$15,$C9,$F1,$F9
	.dc.b	$7F,$73,$AF,$C8,$AF,$8D,$5F,$8D,$7E,$33,$F8,$D7,$E3,$3F,$8D,$7E
	.dc.b	$33,$F8,$D7,$E3,$3F,$8D,$7E,$33,$F8,$D7,$E3,$3F,$8D,$EA,$FE,$CE
	.dc.b	$FC,$68,$FF,$11,$F4,$38,$3E,$3F,$07,$C6,$BF,$F1,$0F,$07,$C6,$7F
	.dc.b	$C4,$7C,$1F,$19,$FF,$12,$F0,$7C,$67,$FC,$4F,$C1,$F1,$9F,$F1,$4F
	.dc.b	$07,$C6,$CF,$F1,$9A,$7F,$42,$3F,$0F,$3F,$B1,$7F,$8D,$9F,$E0,$EE
	.dc.b	$AF,$DB,$CF,$F0,$89,$FD,$DC,$FE,$F4,$FE,$3D,$E5,$7F,$19,$7F,$90
	.dc.b	$79,$5F,$C6,$5F,$E4,$5E,$57,$F1,$97,$F9,$27,$95,$FC,$65,$FE,$4D
	.dc.b	$E5,$7F,$19,$7F,$94,$79,$5F,$C6,$5F,$F5,$19,$06,$5F,$C6,$BC,$8F
	.dc.b	$E3,$31,$8C,$CF,$F9,$F7,$C9,$79,$3F,$C7,$23,$46,$7C,$9F,$E3,$AE
	.dc.b	$3C,$B5,$3C,$AF,$E3,$F3,$98,$79,$5F,$C8,$23,$1E,$57,$F2,$07,$5E
	.dc.b	$57,$C6,$78,$67,$F2,$8F,$F2,$DF,$27,$F2,$08,$F3,$7C,$57,$F4,$42
	.dc.b	$3F,$D2,$7D,$CF,$E8,$FF,$11,$FD,$23,$D8,$FE,$93,$F1,$3F,$D2,$BE
	.dc.b	$27,$FC,$C1,$FC,$00,$F8,$AB,$F0,$58,$F8,$AC,$FC,$26,$BE,$2B,$7F
	.dc.b	$0D,$C3,$E2,$B8,$FC,$47,$2F,$8A,$FA,$1C,$3F,$B9,$3F,$2D,$C9,$F3
	.dc.b	$3A,$BE,$2F,$6E,$E7,$E7,$F7,$7F,$9F,$38,$18,$F5,$9C,$E1,$C6,$59
	.dc.b	$F8,$E7,$27,$FA,$CE,$1A,$FA,$63,$39,$25,$BF,$C0,$E0,$F8,$55,$FD
	.dc.b	$BF,$8B,$E3,$3F,$DC,$79,$3F,$B7,$F2,$FC,$C8,$FB,$5C,$DF,$47,$2E
	.dc.b	$9E,$5F,$CA,$24,$F8,$33,$7C,$57,$CE,$F6,$7E,$7F,$BB,$F4,$3E,$27
	.dc.b	$E8,$FA,$3F,$4B,$DF,$FA,$7F,$13,$F5,$3E,$07,$EA,$FC,$4F,$D6,$F8
	.dc.b	$3F,$AF,$F1,$3F,$E9,$7F,$E0,$1F,$F5,$0F,$57,$CD,$7F,$D4,$3C,$1F
	.dc.b	$01,$FF,$50,$3F,$E0,$BE,$DF,$F3,$83,$FE,$4B,$E2,$FA,$FD,$BF,$EA
	.dc.b	$CA,$DC,$FE,$6D,$E6,$D3,$C2,$7B,$71,$F2,$BE,$FC,$FD,$2F,$C1,$D7
	.dc.b	$DA,$FC,$3D,$FE,$0B,$F1,$78,$7F,$23,$C5,$F2,$23,$F5,$5C,$7F,$C6
	.dc.b	$DF,$F3,$9B,$5C,$E9,$BD,$3F,$11,$5F,$36,$7E,$87,$1F,$CD,$7E,$97
	.dc.b	$1F,$CD,$7E,$A7,$1F,$CD,$7E,$B7,$1F,$CD,$7E,$C7,$1F,$CD,$7E,$77
	.dc.b	$C3,$7C,$FF,$87,$F8,$FC,$BF,$E9,$A9,$FC,$E9,$FC,$1F,$97,$4F,$D1
	.dc.b	$AE,$4F,$43,$E4,$1E,$03,$E4,$1E,$43,$E4,$1E,$83,$E4,$1E,$C3,$E4
	.dc.b	$1F,$8D,$D3,$E1,$57,$E9,$76,$FC,$A7,$94,$E8,$FC,$C7,$07,$F8,$E1
	.dc.b	$FC,$DE,$1F,$9D,$3F,$C9,$DE,$37,$F0,$47,$FC,$A1,$E3,$7E,$95,$FC
	.dc.b	$D7,$7F,$EA,$DF,$AB,$E0,$7E,$C5,$FA,$FE,$0F,$ED,$DF,$B3,$E1,$7E
	.dc.b	$F5,$FC,$57,$C4,$7F,$92,$B1,$FC,$C3,$9C,$FF,$80,$57,$17,$F8,$15
	.dc.b	$F5,$7F,$82,$61,$E1,$7F,$83,$63,$E5,$C7,$F9,$66,$5E,$97,$F9,$76
	.dc.b	$5F,$C8,$1F,$E4,$47,$F9,$AF,$3F,$DD,$7B,$BF,$E9,$6D,$7E,$34,$E4
	.dc.b	$87,$F3,$7F,$07,$93,$D0,$FE,$0C,$7C,$63,$F7,$DE,$27,$F2,$87,$F8
	.dc.b	$07,$91,$FC,$A1,$FE,$05,$E4,$7F,$28,$7F,$82,$79,$1F,$CA,$1F,$E0
	.dc.b	$DE,$47,$E3,$1F,$DA,$1F,$8C,$F1,$BF,$93,$BA,$3F,$86,$71,$FC,$7E
	.dc.b	$6F,$E4,$9C,$9F,$1D,$FE,$4D,$C9,$F1,$DF,$E5,$1C,$7F,$1F,$93,$FC
	.dc.b	$F3,$C9,$F1,$DF,$F3,$DF,$27,$B9,$FF,$3E,$F0,$FE,$22,$FF,$9C,$71
	.dc.b	$7C,$6D,$1C,$7F,$EB,$EF,$03,$F6,$98,$7F,$39,$E7,$FC,$9D,$7F,$33
	.dc.b	$E8,$FD,$A3,$FB,$CF,$03,$F6,$8F,$EF,$BC,$0F,$DA,$3F,$C0,$3C,$0F
	.dc.b	$DA,$3F,$C0,$BC,$0F,$DA,$3F,$C1,$3C,$0F,$D5,$BF,$C5,$7B,$BF,$5E
	.dc.b	$FF,$6D,$EF,$FE,$3B,$D1,$FE,$BF,$8C,$1F,$C8,$FC,$0F,$DD,$BF,$C9
	.dc.b	$4F,$E3,$67,$F1,$0F,$17,$F8,$DB,$FC,$4B,$C5,$FE,$36,$FF,$14,$F1
	.dc.b	$7F,$8D,$BF,$C5,$BC,$5F,$E3,$6F,$F1,$8F,$17,$F8,$DB,$FD,$07,$8B
	.dc.b	$F8,$E7,$2F,$C7,$7E,$2E,$7F,$D2,$27,$F7,$47,$EF,$0F,$8C,$7F,$00
	.dc.b	$3E,$31,$FC,$10,$F8,$C7,$F0,$83,$E3,$1F,$C3,$0F,$8C,$7B,$4F,$E2
	.dc.b	$47,$C1,$3F,$A1,$1F,$8D,$F7,$3E,$CF,$47,$DE,$DB,$E5,$F3,$7F,$43
	.dc.b	$E0,$F9,$EF,$F7,$5E,$1F,$A6,$FF,$77,$E2,$FA,$EF,$F7,$9E,$3F,$B6
	.dc.b	$FF,$7B,$E4,$FB,$EF,$F7,$DE,$5F,$C1,$BF,$CE,$34,$F8,$C4,$3F,$1B
	.dc.b	$C1,$F9,$0F,$17,$F2,$51,$FE,$1E,$3F,$76,$7F,$8A,$FC,$8F,$EE,$EF
	.dc.b	$F8,$B7,$C9,$FE,$F6,$FF,$4A,$F1,$FF,$68,$FF,$4C,$F1,$FF,$68,$FF
	.dc.b	$4E,$F1,$FF,$68,$FF,$50,$F1,$FF,$68,$FF,$52,$F1,$FF,$8B,$BF,$E4
	.dc.b	$9D,$1C,$9F,$E1,$BF,$88,$FF,$72,$3F,$DC,$7C,$9F,$E3,$8F,$F8,$5F
	.dc.b	$C9,$FD,$B5,$FF,$85,$7C,$9F,$DB,$3F,$E1,$7F,$27,$F6,$CF,$F8,$67
	.dc.b	$C9,$FD,$B5,$FE,$C3,$C9,$FD,$B3,$FB,$2F,$27,$F6,$CF,$ED,$3C,$9F
	.dc.b	$8B,$DF,$FE,$50,$E6,$FF,$0A,$79,$4D,$BD,$39,$FF,$32,$3F,$C0,$8E
	.dc.b	$38,$3F,$C0,$0D,$C7,$D5,$3A,$21,$C4,$93,$C1,$CB,$5C,$C7,$26,$9F
	.dc.b	$C7,$0F,$F2,$21,$8C,$17,$E7,$8E,$73,$C6,$BA,$11,$C1,$21,$5C,$AF
	.dc.b	$D5,$18,$BE,$8F,$76,$DF,$38,$D4,$F0,$AE,$51,$FE,$82,$F0,$9D,$3F
	.dc.b	$E2,$DF,$89,$FF,$18,$12,$7F,$C5,$DF,$15,$FE,$33,$F8,$0F,$F1,$A7
	.dc.b	$C4,$7F,$8D,$7E,$0B,$FC,$6D,$F1,$1F,$E3,$7F,$7B,$FC,$71,$F1,$1F
	.dc.b	$E3,$9F,$80,$FF,$1D,$7C,$47,$F8,$EF,$E0,$BF,$C7,$9F,$11,$FE,$3D
	.dc.b	$F8,$4F,$F1,$F7,$C4,$7F,$8F,$FE,$1B,$FA,$37,$27,$F1,$DE,$6F,$F6
	.dc.b	$0F,$A3,$8E,$85,$70,$E4,$FF,$21,$63,$D1,$FF,$22,$67,$E0,$7F,$C8
	.dc.b	$DA,$F9,$1F,$F2,$46,$FE,$87,$FC,$95,$C3,$EC,$7F,$C9,$9C,$7F,$E5
	.dc.b	$1E,$4F,$85,$1F,$E4,$7E,$5F,$C8,$98,$DB,$FC,$98,$EE,$38,$9F,$E5
	.dc.b	$5E,$5F,$8C,$FF,$95,$F9,$7E,$33,$FE,$59,$E5,$F8,$CF,$F9,$6F,$97
	.dc.b	$E3,$3F,$E5,$DE,$5F,$8C,$FF,$97,$F9,$7E,$33,$FC,$78,$FC,$7E,$5F
	.dc.b	$19,$C3,$FE,$5F,$E2,$F8,$D1,$FE,$5F,$E2,$F8,$CF,$F9,$87,$8B,$E3
	.dc.b	$3F,$E6,$3E,$2F,$8C,$FF,$99,$78,$BE,$33,$FE,$67,$E2,$F8,$CF,$F9
	.dc.b	$A7,$8B,$E0,$3F,$14,$FF,$3A,$70,$FE,$DB,$A3,$F2,$1C,$9F,$94,$E9
	.dc.b	$FC,$E7,$3F,$E4,$AF,$F3,$15,$A2,$3F,$A6,$3F,$32,$58,$E1,$BF,$E7
	.dc.b	$5B,$5D,$3C,$2F,$E2,$F5,$FA,$86,$CE,$A3,$99,$E0,$3A,$61,$E5,$73
	.dc.b	$7D,$63,$13,$C4,$8E,$09,$87,$A2,$BF,$85,$75,$7E,$B1,$EE,$7F,$BC
	.dc.b	$0A,$7F,$09,$31,$5F,$0D,$B5,$FB,$C6,$D7,$16,$94,$E6,$49,$E4,$F5
	.dc.b	$21,$99,$E8,$E2,$36,$E6,$7A,$3F,$27,$B7,$D1,$87,$F4,$48,$C1,$FC
	.dc.b	$A0,$F1,$9E,$5A,$E6,$4B,$EF,$1D,$4F,$44,$66,$FB,$B6,$33,$3C,$23
	.dc.b	$11,$D6,$30,$49,$F9,$9E,$DB,$DF,$F6,$4E,$6B,$9D,$76,$8C,$47,$38
	.dc.b	$C1,$1D,$90,$D1,$4F,$79,$0E,$EF,$17,$E1,$7A,$7F,$E1,$07,$FA,$26
	.dc.b	$3F,$D2,$5E,$07,$C3,$3A,$90,$CE,$3F,$01,$B7,$E8,$6D,$F7,$68,$18
	.dc.b	$FC,$37,$B1,$A3,$B1,$E8,$3D,$85,$32,$70,$30,$71,$3C,$B8,$7B,$6F
	.dc.b	$F9,$B6,$BA,$98,$71,$24,$7A,$DC,$B5,$CE,$1F,$D6,$AF,$D1,$39,$24
	.dc.b	$7C,$19,$E2,$D7,$8E,$1F,$0D,$CB,$5F,$CB,$8C,$7F,$35,$4D,$7F,$82
	.dc.b	$DF,$38,$EF,$7C,$5B,$7F,$18,$EE,$57,$2F,$D0,$CF,$F8,$ED,$7F,$04
	.dc.b	$8F,$E1,$B5,$FC,$17,$0F,$E1,$8D,$CB,$6D,$9F,$E2,$31,$87,$FC,$68
	.dc.b	$FF,$8C,$F4,$D6,$3F,$C4,$AA,$E1,$DE,$FF,$8E,$13,$0E,$EF,$88,$FF
	.dc.b	$0F,$EB,$FC,$56,$F9,$46,$4E,$18,$F8,$5A,$79,$09,$1C,$1C,$B6,$BF
	.dc.b	$A9,$47,$F5,$49,$E5,$1E,$28,$DD,$F9,$E3,$04,$9D,$92,$7F,$9B,$DF
	.dc.b	$9E,$B8,$4E,$6F,$C0,$D8,$CD,$79,$9B,$3A,$8C,$61,$FE,$23,$D7,$F3
	.dc.b	$47,$E6,$EB,$64,$8F,$9A,$49,$F8,$D8,$F9,$DE,$67,$F8,$AF,$BB,$F6
	.dc.b	$58,$7F,$90,$3E,$23,$F9,$4E,$58,$C5,$18,$BF,$F1,$D3,$6B,$FC,$8F
	.dc.b	$F1,$16,$9E,$F6,$7D,$6F,$E6,$58,$7E,$6F,$1E,$F0,$F8,$AD,$71,$EF
	.dc.b	$E2,$78,$7C,$0F,$1F,$89,$E5,$F0,$7C,$FE,$27,$A7,$C2,$F5,$F8,$9E
	.dc.b	$DD,$3D,$FE,$27,$C1,$F1,$1E,$5E,$CF,$ED,$67,$F6,$CD,$D1,$BE,$1F
	.dc.b	$E2,$5D,$1F,$AE,$9F,$E2,$51,$E7,$3F,$CB,$6B,$1F,$0C,$70,$E2,$F5
	.dc.b	$E9,$E6,$8C,$1C,$36,$F5,$B8,$E6,$60,$E1,$F8,$33,$98,$F8,$8D,$C3
	.dc.b	$E6,$72,$7C,$DE,$6F,$9D,$CF,$2B,$ED,$1E,$6F,$89,$FA,$04,$7A,$61
	.dc.b	$F4,$FC,$EF,$A9,$E7,$FD,$5F,$6F,$EB,$78,$1F,$5F,$DB,$FB,$1E,$97
	.dc.b	$DD,$FE,$01,$7B,$97,$D5,$C7,$B9,$9F,$5F,$5E,$E6,$FD,$9E,$1E,$E7
	.dc.b	$D2,$E1,$FE,$06,$7F,$1C,$EC,$F9,$DD,$BA,$8F,$89,$A2,$F9,$C6,$31
	.dc.b	$EE,$EB,$3F,$26,$9E,$68,$F9,$1B,$7A,$9F,$95,$C1,$EE,$7E,$67,$17
	.dc.b	$E4,$F8,$FE,$24,$7D,$DE,$AF,$A7,$CB,$FE,$1F,$F2,$BF,$87,$52,$37
	.dc.b	$C1,$F8,$57,$95,$F4,$3E,$C7,$37,$CB,$7E,$D7,$37,$CB,$7E,$E7,$37
	.dc.b	$A1,$F5,$1F,$03,$9B,$E5,$3F,$0B,$9B,$E5,$3F,$13,$9B,$F8,$07,$B5
	.dc.b	$FA,$FF,$2F,$F8,$07,$97,$F2,$6B,$D5,$D1,$F2,$5E,$47,$C9,$3B,$1E
	.dc.b	$03,$7C,$BF,$25,$F8,$5C,$0E,$5F,$95,$18,$E5,$F9,$4E,$B9,$7E,$0B
	.dc.b	$C3,$8B,$E5,$F2,$7C,$87,$E3,$F6,$FE,$F9,$DF,$2F,$EB,$6F,$9B,$FA
	.dc.b	$FC,$3B,$BF,$B1,$C7,$C8,$FF,$60,$D3,$F2,$39,$7A,$63,$1C,$5F,$22
	.dc.b	$3E,$0F,$17,$C8,$7E,$1F,$17,$C8,$7F,$61,$C3,$F2,$1F,$98,$7E,$43
	.dc.b	$A7,$F1,$EF,$CA,$E9,$FC,$7B,$F3,$3A,$7E,$C3,$F1,$B9,$BE,$E3,$F2
	.dc.b	$39,$FF,$00,$FC,$AE,$8F,$D7,$3F,$B0,$3F,$79,$C5,$F1,$BC,$BF,$E5
	.dc.b	$FC,$0F,$6F,$F5,$C4,$AF,$D1,$FB,$9F,$B1,$EB,$FD,$97,$C4,$7E,$CF
	.dc.b	$0F,$DA,$7B,$DF,$B5,$F8,$4F,$DB,$7C,$07,$ED,$FE,$17,$F7,$1F,$05
	.dc.b	$FB,$9F,$86,$FD,$D7,$C2,$7E,$EF,$E1,$FF,$79,$F0,$DF,$CE,$7F,$80
	.dc.b	$7F,$CE,$DD,$DF,$B4,$E1,$FD,$BF,$8F,$FE,$0F,$E7,$FD,$D4,$7E,$9E
	.dc.b	$B7,$E1,$FA,$B0,$E9,$3F,$BA,$CB,$C0,$FE,$F3,$4F,$23,$FB,$ED,$BD
	.dc.b	$0F,$F0,$0E,$0F,$63,$FC,$0B,$8B,$E0,$3F,$08,$FE,$0B,$CB,$F1,$57
	.dc.b	$89,$F8,$F7,$E6,$90,$FC,$DF,$8B,$F4,$9F,$A8,$7F,$04,$3F,$6A,$7F
	.dc.b	$08,$3F,$6A,$7F,$12,$F1,$BF,$8F,$3F,$A8,$3F,$08,$7E,$B0,$FD,$7F
	.dc.b	$7F,$EC,$78,$BE,$44,$7F,$0B,$F4,$3F,$A7,$F1,$7C,$89,$FC,$D7,$85
	.dc.b	$F9,$57,$F3,$9E,$1F,$E6,$5F,$CD,$DF,$E7,$78,$FF,$0B,$97,$E8,$9F
	.dc.b	$EC,$FE,$2F,$E9,$9F,$ED,$3E,$37,$EA,$9F,$ED,$7E,$3F,$EB,$9F,$D8
	.dc.b	$5F,$C3,$DF,$E3,$3F,$B7,$3E,$36,$3F,$D7,$F7,$F8,$CF,$F6,$1F,$0B
	.dc.b	$F3,$6F,$F6,$3D,$BF,$3E,$FF,$65,$F0,$FF,$46,$FE,$D7,$C4,$FD,$3B
	.dc.b	$FB,$7F,$17,$F5,$6F,$EE,$78,$FE,$36,$9F,$EE,$AF,$1B,$FC,$33,$E1
	.dc.b	$FF,$81,$FC,$8C,$EC,$EF,$78,$DF,$C6,$E3,$F7,$A7,$F1,$B3,$F8,$FF
	.dc.b	$91,$F5,$9F,$E3,$67,$DA,$3F,$8D,$F2,$FC,$77,$FB,$CF,$2F,$C7,$7F
	.dc.b	$BD,$F2,$FC,$77,$FB,$41,$F1,$FC,$6F,$92,$FC,$72,$6F,$F1,$DF,$E8
	.dc.b	$07,$E3,$0F,$E4,$C7,$E3,$0F,$09,$E3,$7F,$8D,$73,$F9,$DF,$BF,$1F
	.dc.b	$8C,$3D,$A7,$E3,$0F,$E2,$A7,$F5,$5F,$63,$FD,$F7,$BF,$C8,$8F,$F0
	.dc.b	$54,$7F,$BF,$B4,$75,$7E,$37,$87,$F8,$6F,$57,$E3,$5F,$E1,$FB,$F8
	.dc.b	$5F,$C6,$E1,$E5,$7F,$1A,$FA,$72,$FE,$68,$FF,$5C,$3D,$EF,$C1,$7E
	.dc.b	$1C,$7C,$6E,$AF,$F3,$A6,$3F,$C6,$E7,$F6,$E4,$3F,$AA,$F9,$1F,$BC
	.dc.b	$7F,$AB,$F7,$7F,$00,$7F,$AC,$F7,$FF,$04,$7F,$AD,$F8,$1F,$C2,$1F
	.dc.b	$EB,$BE,$0F,$F0,$C7,$FA,$FF,$85,$FC,$41,$FE,$C3,$E1,$FF,$1B,$7F
	.dc.b	$61,$A7,$FA,$45,$FE,$37,$E0,$FF,$90,$1F,$F1,$97,$89,$FC,$69,$FC
	.dc.b	$86,$AB,$E9,$BF,$F5,$1C,$7C,$0F,$F5,$3C,$BC,$8F,$F5,$5C,$FD,$0F
	.dc.b	$F5,$7D,$3D,$8F,$F5,$9D,$7E,$03,$FD,$6F,$6F,$E3,$4F,$F5,$DF,$13
	.dc.b	$FC,$21,$CF,$FE,$72,$D9,$F1,$1F,$CD,$49,$BF,$A3,$E2,$7F,$9C,$77
	.dc.b	$7F,$39,$F6,$FF,$9D,$7C,$4F,$F3,$BF,$77,$F9,$E7,$C4,$FF,$3D,$F7
	.dc.b	$FF,$9F,$7C,$4F,$F3,$FF,$81,$FE,$81,$F1,$3F,$D0,$7E,$0F,$FA,$17
	.dc.b	$C4,$FF,$43,$F8,$5F,$E8,$9F,$13,$FD,$17,$E1,$FF,$CD,$5E,$8C,$E2
	.dc.b	$F9,$31,$A7,$E5,$D7,$27,$E6,$E1,$D9,$F9,$F9,$78,$9F,$A3,$A7,$99
	.dc.b	$FA,$7B,$7A,$9F,$AB,$C1,$EE,$7E,$01,$F6,$38,$FE,$1B,$FD,$0B,$97
	.dc.b	$E4,$7C,$4F,$F5,$B2,$47,$DE,$F8,$9F,$EB,$DE,$E7,$F5,$FF,$77,$FB
	.dc.b	$07,$BD,$FE,$99,$FE,$01,$FF,$4A,$F6,$FF,$BB,$9F,$F3,$8E,$1F,$E4
	.dc.b	$55,$F1,$3B,$F8,$1E,$1F,$13,$C7,$E0,$F9,$7C,$4F,$3F,$85,$E9,$F1
	.dc.b	$3D,$7E,$1F,$F2,$79,$FF,$1D,$F5,$7F,$74,$FF,$8C,$3E,$1F,$4F,$FA
	.dc.b	$17,$55,$F3,$8C,$63,$DD,$D6,$7E,$37,$86,$BE,$BF,$8A,$F6,$7A,$5E
	.dc.b	$E9,$FC,$FF,$17,$C2,$7E,$21,$FA,$2E,$5F,$8E,$FC,$B3,$F9,$97,$90
	.dc.b	$8C,$F1,$AF,$E7,$D7,$23,$FC,$09,$CB,$F3,$5F,$D6,$F2,$F8,$9F,$66
	.dc.b	$5F,$3F,$3F,$44,$7C,$DD,$3D,$8F,$CF,$D7,$E0,$3F,$47,$6F,$8A,$FC
	.dc.b	$73,$FC,$77,$EE,$FF,$98,$CF,$9B,$EE,$B8,$7F,$D6,$D3,$F2,$EB,$9E
	.dc.b	$DF,$77,$0E,$EF,$DF,$C7,$C6,$FE,$07,$2F,$95,$7F,$4F,$6F,$94,$FF
	.dc.b	$88,$78,$7E,$53,$FE,$23,$E8,$F8,$B3,$F1,$23,$E5,$68,$E2,$FD,$DC
	.dc.b	$7E,$64,$87,$F8,$AF,$CA,$FC,$F3,$FE,$2D,$EC,$FD,$13,$FE,$2F,$ED
	.dc.b	$FD,$33,$FE,$31,$EE,$FD,$53,$FE,$33,$EF,$FD,$73,$FE,$35,$F0,$3F
	.dc.b	$64,$FF,$8D,$FC,$1F,$DB,$3F,$E3,$9F,$0B,$F7,$4F,$F8,$5F,$C3,$FD
	.dc.b	$F3,$FC,$93,$C4,$FE,$04,$FF,$26,$F1,$7F,$83,$3F,$CA,$3C,$6F,$E1
	.dc.b	$4F,$F2,$AF,$1F,$F8,$73,$FC,$B3,$C8,$FE,$24,$FF,$2E,$F2,$7F,$8B
	.dc.b	$3F,$CC,$3C,$AF,$F2,$87,$17,$EE,$BA,$7F,$19,$3F,$90,$F2,$BF,$91
	.dc.b	$B4,$F8,$F1,$AE,$6F,$99,$3C,$A7,$A1,$FC,$20,$F0,$1F,$C3,$0F,$21
	.dc.b	$FC,$40,$F4,$1F,$C5,$0F,$61,$FC,$60,$F8,$07,$F1,$C3,$E2,$1F,$C8
	.dc.b	$0F,$F3,$2F,$C1,$B7,$F9,$31,$FE,$65,$DF,$E4,$BF,$E6,$7D,$FE,$4B
	.dc.b	$FE,$69,$DF,$E4,$BF,$E6,$BD,$FE,$4B,$FE,$6D,$DF,$E4,$BF,$E6,$FD
	.dc.b	$FE,$4B,$FE,$71,$DF,$E0,$3F,$E7,$3D,$3F,$91,$FA,$CF,$33,$F2,$5C
	.dc.b	$1F,$1F,$A7,$F7,$57,$FB,$F3,$F7,$11,$9F,$17,$F7,$11,$FE,$7D,$F1
	.dc.b	$7F,$70,$FF,$21,$F1,$7F,$70,$FF,$15,$F3,$3F,$90,$BF,$C5,$FC,$CF
	.dc.b	$E4,$2F,$F1,$9F,$33,$F9,$0B,$FC,$6F,$CC,$FD,$AF,$99,$FC,$83,$3F
	.dc.b	$F3,$39,$FC,$BE,$37,$3F,$CB,$EB,$9B,$FC,$C6,$FB,$BF,$CC,$F0,$F1
	.dc.b	$BE,$53,$F6,$C7,$F1,$03,$F8,$F9,$FD,$FB,$3F,$F4,$F7,$87,$FB,$68
	.dc.b	$FE,$3B,$E5,$FE,$3B,$AF,$F6,$CE,$5B,$1F,$1B,$C4,$FE,$E9,$7F,$1B
	.dc.b	$4F,$99,$BF,$C6,$7F,$A4,$6F,$E4,$7F,$A5,$67,$E8,$7F,$A6,$69,$EC
	.dc.b	$7F,$A7,$6B,$F0,$1F,$85,$90,$FE,$3C,$1F,$D0,$6F,$F2,$73,$D6,$8F
	.dc.b	$5B,$FC,$39,$7F,$D5,$F8,$3A,$56,$BC,$B6,$5E,$18,$C6,$7E,$57,$5A
	.dc.b	$7A,$5E,$1A,$FB,$5F,$F0,$A6,$DF,$05,$F8,$79,$7C,$53,$E5,$9F,$8D
	.dc.b	$3F,$90,$F1,$7F,$20,$7F,$2C,$7E,$EA,$3F,$C3,$DD,$FF,$C8,$38,$3F
	.dc.b	$2D,$E3,$7F,$20,$7F,$31,$E3,$7F,$20,$7F,$35,$E3,$7F,$20,$7F,$39
	.dc.b	$E3,$7F,$20,$7F,$3D,$E3,$7F,$20,$7F,$41,$E3,$7F,$20,$7F,$45,$E3
	.dc.b	$7F,$20,$7F,$85,$78,$DF,$B9,$D7,$F7,$87,$EE,$4F,$E0,$07,$EE,$4F
	.dc.b	$E0,$87,$EE,$4F,$E1,$07,$EE,$4F,$E1,$87,$EE,$4F,$E2,$07,$EE,$4F
	.dc.b	$E2,$87,$EE,$4F,$F0,$D9,$F1,$09,$DB,$FB,$97,$FD,$45,$CA,$F8,$AF
	.dc.b	$F0,$B1,$23,$FC,$0C,$D7,$FC,$2C,$DB,$FF,$0D,$37,$83,$F0,$4D,$9E
	.dc.b	$AD,$E9,$FE,$D0,$DB,$F0,$4F,$63,$E2,$DF,$C8,$F1,$FE,$56,$1F,$E5
	.dc.b	$7C,$F8,$61,$2F,$81,$E4,$2D,$F9,$1B,$5E,$46,$9E,$42,$E3,$E4,$6E
	.dc.b	$3E,$47,$FB,$6A,$F8,$AF,$F1,$D1,$23,$FC,$55,$E7,$FF,$8F,$7E,$2B
	.dc.b	$FC,$7D,$DB,$FE,$3F,$F7,$FF,$C8,$1F,$0D,$FE,$41,$F8,$1F,$F2,$17
	.dc.b	$07,$F9,$0F,$E0,$BF,$C8,$9F,$0D,$FB,$3C,$FF,$65,$BF,$EC,$3C,$F7
	.dc.b	$C1,$E7,$8B,$FA,$A4,$3E,$1B,$15,$BF,$81,$B8,$7F,$B7,$69,$F0,$D8
	.dc.b	$E5,$43,$E0,$FF,$A5,$E5,$FC,$03,$DA,$FF,$0F,$61,$F4,$B2,$F1,$57
	.dc.b	$1B,$E7,$3F,$C2,$6F,$C7,$96,$73,$D6,$9E,$CA,$FF,$27,$6D,$F0,$1F
	.dc.b	$F2,$87,$07,$C4,$7F,$CA,$5C,$5F,$21,$FF,$2A,$72,$7F,$99,$F9,$7E
	.dc.b	$74,$7F,$99,$79,$BF,$95,$73,$FD,$38,$FF,$2D,$46,$B5,$FF,$36,$57
	.dc.b	$D3,$9F,$F2,$F7,$3F,$D3,$7F,$CC,$1C,$FF,$4D,$FF,$31,$73,$FD,$37
	.dc.b	$FC,$C9,$CF,$EA,$7D,$DC,$9F,$4F,$87,$E1,$71,$7C,$3E,$7F,$89,$CB
	.dc.b	$F2,$AB,$F9,$3F,$95,$FC,$A7,$0F,$F2,$EF,$56,$9E,$67,$D5,$BE,$6E
	.dc.b	$3A,$7E,$AE,$5E,$18,$FF,$42,$67,$E5,$7F,$D0,$DA,$7A,$5F,$F4,$46
	.dc.b	$BE,$D7,$FD,$15,$B7,$C1,$7F,$D1,$9B,$FC,$5D,$7F,$9B,$70,$FC,$AE
	.dc.b	$AF,$D5,$57,$EA,$49,$C5,$FA,$96,$D7,$EA,$5A,$7E,$A4,$BA,$FE,$A5
	.dc.b	$BC,$7F,$A9,$AE,$44,$AE,$4F,$E6,$BE,$1B,$F7,$B4,$B7,$6F,$F2,$78
	.dc.b	$FD,$19,$F9,$B9,$F3,$D7,$38,$F1,$36,$BC,$4D,$C3,$C5,$87,$80,$95
	.dc.b	$C1,$B5,$C1,$B3,$C2,$7D,$DD,$BF,$C9,$EB,$DD,$1F,$B7,$6B,$F1,$4C
	.dc.b	$4F,$F0,$96,$E5,$FB,$D6,$94,$F9,$46,$0D,$4F,$47,$27,$52,$18,$23
	.dc.b	$B2,$4F,$86,$3F,$85,$63,$F3,$A7,$E6,$EF,$FB,$B8,$FE,$79,$3E,$12
	.dc.b	$1D,$A7,$9D,$F9,$E3,$BD,$F5,$6C,$66,$77,$1E,$D6,$BE,$C9,$E1,$1F
	.dc.b	$87,$F6,$3F,$A0,$BA,$9F,$09,$D0,$E5,$3C,$09,$5D,$DB,$1A,$8C,$11
	.dc.b	$C1,$0F,$13,$97,$E7,$F8,$CE,$5F,$E4,$31,$F9,$0B,$73,$3B,$FF,$7D
	.dc.b	$3E,$53,$99,$D4,$F4,$CF,$AD,$F9,$58,$F9,$9D,$90,$DB,$97,$F0,$59
	.dc.b	$7C,$48,$C4,$72,$CB,$D0,$F0,$6B,$93,$C4,$53,$B1,$4F,$2E,$5E,$F7
	.dc.b	$FD,$BA,$43,$FD,$70,$7F,$6C,$3B,$1D,$0F,$80,$60,$EA,$60,$EC,$43
	.dc.b	$F8,$35,$7A,$E7,$CF,$1B,$25,$FC,$76,$DF,$C0,$6C,$F0,$9F,$CF,$77
	.dc.b	$B4,$FC,$E7,$0F,$F0,$7E,$1F,$E1,$98,$7A,$4F,$15,$F5,$26,$3D,$5B
	.dc.b	$8F,$58,$C1,$FE,$01,$CB,$D1,$5F,$E0,$49,$FE,$23,$C3,$FC,$19,$B1
	.dc.b	$FC,$2C,$97,$FC,$2D,$B7,$FC,$2E,$30,$47,$19,$FA,$E7,$FD,$7F,$C1
	.dc.b	$FB,$57,$C0,$43,$F9,$6D,$F8,$2F,$93,$6F,$B1,$1F,$F1,$C7,$0F,$F1
	.dc.b	$98,$C5,$73,$CF,$F9,$35,$71,$24,$67,$2F,$73,$FE,$26,$DB,$F1,$0F
	.dc.b	$E9,$78,$3F,$AF,$64,$9C,$E1,$D2,$B9,$1F,$0B,$0E,$AF,$A8,$CD,$7A
	.dc.b	$9F,$F1,$DE,$5E,$79,$DD,$73,$9E,$04,$70,$4C,$7C,$F1,$FD,$53,$0F
	.dc.b	$1C,$7F,$74,$E0,$FD,$ED,$7F,$9D,$AB,$C3,$9F,$C4,$7A,$57,$29,$E9
	.dc.b	$1E,$93,$78,$7C,$CF,$43,$FC,$6D,$F0,$BF,$E2,$52,$4F,$F7,$BD,$DE
	.dc.b	$EF,$F8,$A7,$DE,$FF,$15,$69,$FE,$2B,$F7,$FF,$C5,$98,$FF,$8B,$7E
	.dc.b	$03,$FC,$5D,$F1,$5F,$E2,$FF,$82,$FF,$18,$7C,$57,$F8,$C7,$E1,$3F
	.dc.b	$C6,$5F,$15,$FE,$33,$F8,$6F,$F1,$A7,$C5,$7F,$8D,$7E,$23,$F1,$1C
	.dc.b	$3F,$C6,$7B,$7F,$49,$CD,$FE,$BA,$6C,$7F,$6B,$3C,$D3,$C9,$F4,$5F
	.dc.b	$67,$D5,$8F,$89,$F6,$67,$E6,$7D,$DA,$FA,$9F,$81,$BF,$B9,$F8,$5C
	.dc.b	$3F,$E4,$9E,$2F,$8B,$1F,$CB,$78,$DC,$9F,$EE,$5F,$E0,$1F,$EC,$DC
	.dc.b	$EE,$0F,$F6,$8C,$61,$B7,$FE,$5B,$8F,$F6,$04,$EE,$3F,$C1,$50,$F5
	.dc.b	$75,$F0,$1B,$F5,$B8,$7C,$0F,$1F,$5F,$97,$C1,$73,$F6,$3A,$7C,$1F
	.dc.b	$5F,$67,$B7,$C2,$77,$F6,$BC,$1F,$0B,$E1,$F6,$FE,$7F,$0F,$FA,$5F
	.dc.b	$87,$FC,$75,$DB,$F3,$BC,$8F,$F3,$7F,$99,$B9,$FF,$1F,$72,$31,$E9
	.dc.b	$18,$CF,$C0,$EB,$5F,$23,$C3,$7F,$43,$CB,$87,$D8,$F4,$E3,$F8,$9C
	.dc.b	$9F,$17,$E1,$7E,$34,$C3,$FE,$00,$79,$5F,$E1,$BD,$3E,$64,$7C,$83
	.dc.b	$47,$F6,$17,$E6,$F3,$FD,$2B,$F9,$73,$EA,$E7,$F9,$6F,$B3,$9F,$E5
	.dc.b	$BE,$EE,$7F,$96,$FC,$0E,$7F,$96,$FC,$2E,$7F,$96,$FC,$4E,$7F,$97
	.dc.b	$97,$ED,$E3,$FA,$A7,$3F,$F3,$C7,$C9,$3C,$1F,$35,$74,$7D,$17,$E0
	.dc.b	$7D,$58,$79,$1F,$66,$3E,$87,$DD,$97,$B1,$F8,$19,$FC,$97,$EB,$9F
	.dc.b	$13,$5F,$90,$FC,$6C,$FF,$21,$CC,$DB,$E5,$B7,$7F,$A7,$5F,$54,$F9
	.dc.b	$07,$D9,$3E,$41,$F7,$4F,$90,$7E,$04,$F9,$07,$E1,$4F,$90,$7E,$2B
	.dc.b	$97,$F1,$8F,$EA,$F9,$BF,$8E,$F0,$FC,$98,$C1,$F3,$4E,$07,$23,$EF
	.dc.b	$73,$7E,$41,$FC,$07,$37,$E4,$1F,$C1,$73,$7E,$41,$FC,$27,$37,$E4
	.dc.b	$1F,$C3,$73,$7B,$9F,$C4,$67,$F1,$71,$FD,$BC,$7E,$43,$C6,$FD,$DF
	.dc.b	$57,$EE,$49,$1F,$A4,$F3,$BF,$79,$F0,$9F,$BD,$F8,$5F,$DF,$7C,$2F
	.dc.b	$EF,$FE,$1B,$F8,$07,$C0,$FF,$B6,$FF,$80,$7F,$80,$63,$FE,$37,$5F
	.dc.b	$07,$9F,$57,$5F,$09,$BF,$88,$E1,$F0,$DC,$7E,$27,$97,$C4,$73,$F8
	.dc.b	$AF,$F0,$5F,$B7,$74,$FF,$5A,$68,$AE,$11,$8C,$3A,$3A,$CB,$C0,$F0
	.dc.b	$D3,$CD,$CB,$E8,$8F,$E0,$3B,$FB,$1F,$E0,$7C,$3F,$01,$F8,$47,$F0
	.dc.b	$6E,$4F,$8A,$BC,$1F,$C8,$BF,$94,$21,$F4,$0F,$E1,$BE,$17,$E6,$FC
	.dc.b	$2F,$CA,$F8,$4F,$0F,$F4,$11,$8F,$13,$F4,$8E,$BC,$5F,$D4,$3C,$3C
	.dc.b	$6F,$D6,$3C,$BC,$7F,$D8,$3D,$3C,$8F,$DB,$E7,$FE,$B9,$F2,$7F,$90
	.dc.b	$1F,$C8,$76,$FF,$0E,$47,$EE,$0F,$C8,$9F,$B8,$3F,$38,$7E,$E0,$FD
	.dc.b	$01,$FB,$83,$F4,$87,$EE,$0F,$D4,$1F,$B8,$3F,$58,$7E,$E0,$FE,$5B
	.dc.b	$E0,$FE,$45,$FE,$38,$7E,$44,$A7,$F1,$D3,$5E,$4F,$ED,$DF,$92,$74
	.dc.b	$C3,$E4,$EB,$F2,$38,$FF,$A9,$ED,$F2,$1F,$EA,$BB,$7C,$87,$FA,$BE
	.dc.b	$DF,$21,$FE,$B3,$B7,$C8,$7F,$AD,$ED,$F2,$33,$FF,$46,$FA,$FF,$B8
	.dc.b	$D3,$FB,$A1,$F1,$CF,$E6,$5E,$57,$F4,$67,$F6,$E7,$F0,$43,$F6,$E7
	.dc.b	$F0,$83,$F6,$E7,$F0,$C3,$F6,$E7,$F1,$03,$F6,$E7,$B8,$FC,$41,$F0
	.dc.b	$CF,$C5,$F1,$FE,$39,$C1,$36,$FC,$5E,$FF,$BF,$7F,$1B,$E2,$FF,$03
	.dc.b	$7F,$1F,$E3,$7F,$07,$7F,$23,$E3,$FF,$0B,$7F,$27,$E4,$7F,$0F,$7F
	.dc.b	$2B,$E4,$FB,$5F,$CB,$E7,$FC,$71,$FE,$80,$7F,$2E,$F6,$FF,$C9,$FC
	.dc.b	$DF,$8C,$8F,$C9,$69,$F8,$C7,$F2,$CF,$E3,$1F,$CD,$3F,$8C,$7F,$3C
	.dc.b	$FE,$31,$FD,$13,$F8,$C7,$EF,$BF,$81,$E1,$FE,$C0,$FE,$30,$FD,$7C
	.dc.b	$7E,$B1,$FF,$0C,$4F,$ED,$A3,$F5,$EF,$F8,$57,$C9,$FD,$E4,$FF,$6B
	.dc.b	$EB,$FE,$00,$FF,$6D,$EC,$FE,$08,$FF,$19,$C3,$F8,$E7,$6F,$F0,$A8
	.dc.b	$FE,$39,$DD,$FC,$39,$FE,$3D,$DF,$FC,$49,$FE,$41,$E0,$7F,$17,$7F
	.dc.b	$C8,$1E,$17,$FB,$FF,$C2,$78,$7F,$C7,$A7,$F8,$DF,$89,$FC,$89,$FE
	.dc.b	$3B,$E2,$FF,$26,$7F,$8F,$F8,$DE,$27,$F9,$0E,$BE,$6E,$2F,$E1,$7B
	.dc.b	$FA,$9F,$E1,$BC,$1E,$E7,$F8,$7F,$0F,$C3,$AF,$E7,$F7,$FB,$1E,$D6
	.dc.b	$AF,$8A,$FE,$74,$49,$FE,$5F,$ED,$7F,$3D,$F7,$7F,$9F,$7B,$DF,$CF
	.dc.b	$FE,$1B,$FA,$07,$C0,$7F,$41,$F8,$7F,$E8,$5F,$05,$FD,$0F,$E2,$3F
	.dc.b	$A2,$73,$FF,$45,$F8,$9F,$E8,$DF,$0A,$D6,$71,$FF,$78,$FE,$01,$7C
	.dc.b	$5E,$67,$FE,$00,$43,$E2,$31,$5B,$F8,$BF,$9C,$FF,$8F,$EB,$E7,$9F
	.dc.b	$D1,$BB,$77,$CF,$DF,$9D,$8B,$0C,$DA,$73,$8E,$1D,$E5,$E8,$9F,$EA
	.dc.b	$3A,$7B,$1F,$EA,$7B,$7C,$07,$FA,$AF,$07,$C4,$7F,$AB,$F1,$7C,$87
	.dc.b	$FA,$CF,$27,$F6,$7E,$5F,$9D,$1F,$D9,$F9,$BE,$07,$96,$F6,$7D,$1C
	.dc.b	$0F,$8A,$F2,$36,$FE,$4F,$16,$A7,$3F,$17,$AF,$73,$7E,$EF,$0F,$7B
	.dc.b	$8F,$BF,$CB,$D3,$E7,$F0,$1D,$35,$EB,$F0,$3D,$BE,$27,$BF,$C1,$F8
	.dc.b	$3E,$27,$C3,$F0,$BE,$2F,$89,$FA,$3C,$3F,$E1,$DF,$07,$F8,$26,$98
	.dc.b	$E3,$F9,$9C,$8B,$E3,$1E,$83,$38,$F6,$71,$DD,$F5,$B4,$F1,$C6,$2F
	.dc.b	$E2,$37,$7F,$53,$9E,$0F,$73,$BE,$2F,$84,$F1,$E4,$F9,$3C,$B2,$3F
	.dc.b	$45,$87,$CB,$7F,$1F,$19,$7F,$23,$3C,$9F,$C9,$D7,$67,$F2,$B7,$E2
	.dc.b	$7F,$2F,$87,$CB,$7D,$1C,$5F,$2D,$F5,$71,$7C,$B7,$D9,$C5,$F2,$DF
	.dc.b	$CC,$1F,$E6,$EE,$2F,$98,$7E,$67,$C7,$FE,$13,$1F,$0F,$87,$E5,$BF
	.dc.b	$A6,$E1,$F9,$4F,$EA,$38,$3E,$4B,$FA,$AE,$4F,$BE,$FC,$BE,$3F,$C1
	.dc.b	$BF,$37,$97,$F0,$EF,$CF,$E6,$FC,$5B,$F4,$78,$FE,$27,$3F,$D2,$E6
	.dc.b	$FE,$13,$D1,$F9,$AE,$8F,$F3,$8E,$FF,$C2,$F0,$FE,$12,$7E,$80,$FF
	.dc.b	$1C,$F8,$FF,$C2,$5F,$F1,$DF,$8F,$FC,$25,$FF,$1E,$F8,$FF,$C2,$5F
	.dc.b	$F1,$FF,$8F,$FC,$25,$FF,$20,$F8,$FF,$A2,$F1,$3F,$47,$E2,$FE,$93
	.dc.b	$C6,$FF,$23,$D7,$FB,$02,$FF,$75,$F0,$3F,$EC,$1F,$07,$F8,$04,$FE
	.dc.b	$2E,$F9,$3D,$0F,$DE,$9E,$03,$F7,$A7,$90,$FD,$E9,$E8,$3F,$7A,$7B
	.dc.b	$0F,$71,$FE,$61,$F0,$BE,$27,$5F,$D9,$E1,$FD,$EF,$45,$AF,$F2,$DC
	.dc.b	$69,$FE,$25,$C2,$BF,$E6,$11,$D4,$FF,$36,$78,$9F,$C3,$5F,$F3,$77
	.dc.b	$8B,$FC,$45,$FF,$38,$78,$DF,$C5,$5F,$F3,$97,$8F,$FC,$65,$FE,$57
	.dc.b	$C3,$F2,$1F,$D6,$70,$FE,$DF,$9D,$C3,$FE,$48,$E2,$C7,$17,$C8,$BF
	.dc.b	$E6,$7C,$5F,$21,$FF,$3F,$71,$FD,$57,$FD,$01,$5F,$65,$FF,$41,$72
	.dc.b	$7D,$D7,$FD,$09,$CB,$F8,$17,$FD,$0D,$CD,$EE,$7E,$41,$F1,$08,$7F
	.dc.b	$9B,$F9,$7F,$8F,$3F,$93,$21,$FB,$63,$F3,$07,$F0,$13,$F3,$87,$F0
	.dc.b	$53,$F4,$07,$F0,$93,$F8,$63,$FA,$6E,$7F,$EA,$5D,$BF,$AA,$7F,$AA
	.dc.b	$77,$7E,$B9,$FE,$AD,$DF,$F1,$FD,$47,$81,$FE,$47,$E2,$FE,$D7,$E4
	.dc.b	$FE,$EF,$3F,$EB,$3E,$17,$EF,$DF,$EB,$7E,$1F,$F0,$37,$FA,$EF,$89
	.dc.b	$FC,$1D,$FE,$12,$79,$8F,$E1,$A7,$A8,$FE,$22,$7B,$8F,$80,$7F,$18
	.dc.b	$3E,$1F,$83,$FB,$5E,$6F,$E3,$9A,$FF,$6D,$F0,$BF,$8E,$3F,$DB,$FC
	.dc.b	$8F,$E3,$8F,$F7,$1F,$23,$F8,$E3,$FD,$CF,$C8,$FE,$38,$FF,$75,$F2
	.dc.b	$3C,$CF,$F7,$7C,$FD,$4F,$F7,$9D,$3D,$CF,$F7,$BD,$7E,$13,$FC,$13
	.dc.b	$6F,$8D,$C3,$F2,$37,$FE,$3E,$7F,$7E,$E2,$F9,$DE,$37,$F6,$0E,$4F
	.dc.b	$A4,$FF,$62,$E5,$FA,$CF,$F6,$4E,$6F,$E3,$77,$FE,$06,$E6,$FE,$36
	.dc.b	$FF,$82,$3C,$8F,$E3,$6F,$F8,$2B,$C8,$FE,$36,$FC,$72,$73,$FF,$1B
	.dc.b	$F1,$FF,$8C,$47,$F7,$33,$F2,$3B,$FF,$1C,$F1,$FF,$2C,$FF,$1E,$EB
	.dc.b	$FC,$D3,$FC,$83,$B3,$F3,$CF,$F2,$2E,$DF,$D1,$3F,$C9,$3B,$BF,$4C
	.dc.b	$FF,$26,$EF,$FD,$53,$FE,$21,$F0,$3F,$8D,$65,$FC,$64,$87,$F8,$70
	.dc.b	$91,$FE,$21,$7F,$8D,$1F,$D9,$7C,$3F,$DF,$47,$F3,$2F,$13,$F8,$13
	.dc.b	$FC,$D3,$C5,$FE,$0C,$FF,$36,$F1,$BF,$85,$3F,$CE,$3C,$7F,$F0,$B6
	.dc.b	$1F,$EF,$9F,$27,$FC,$2C,$FC,$2E,$7F,$E3,$0F,$EC,$CF,$F6,$76,$BF
	.dc.b	$CC,$1F,$2C,$7F,$8B,$FC,$7F,$E1,$51,$EA,$C3,$FC,$6C,$49,$D5,$71
	.dc.b	$8C,$4F,$43,$84,$E9,$A6,$4E,$E5,$8E,$3E,$3F,$F1,$AA,$E9,$5C,$6B
	.dc.b	$73,$97,$9B,$5C,$DF,$A1,$B5,$C9,$F0,$72,$FE,$B7,$E1,$FF,$C6,$64
	.dc.b	$8F,$F0,$B7,$C4,$FF,$8D,$BE,$27,$FC,$95,$DD,$FE,$5B,$96,$8D,$3F
	.dc.b	$CA,$92,$F7,$B7,$F0,$DC,$3D,$FE,$3F,$0B,$CB,$E0,$39,$FC,$2F,$4F
	.dc.b	$81,$EB,$F0,$BD,$BE,$0B,$BF,$C2,$F8,$3E,$0F,$F8,$B7,$8F,$FC,$57
	.dc.b	$C5,$FF,$04,$F1,$78,$B8,$7C,$35,$C2,$3F,$C6,$3C,$2C,$7A,$C6,$33
	.dc.b	$F0,$BA,$D7,$CA,$F0,$DF,$D2,$F2,$E1,$F8,$2F,$C3,$3E,$29,$FE,$3B
	.dc.b	$E5,$F9,$0F,$E2,$FA,$7F,$CE,$B1,$8E,$4F,$F4,$23,$FE,$60,$E6,$E5
	.dc.b	$5E,$93,$FC,$F1,$A7,$CD,$AF,$99,$18,$C3,$C6,$EF,$05,$31,$F4,$CE
	.dc.b	$61,$97,$BA,$31,$9F,$C2,$75,$A7,$C7,$78,$6B,$F3,$9F,$F1,$C1,$FD
	.dc.b	$EB,$CC,$FA,$4F,$CC,$E2,$FA,$CF,$CE,$E3,$FB,$4F,$D0,$E4,$FB,$CF
	.dc.b	$D2,$E5,$FC,$13,$F5,$39,$BF,$17,$C1,$F3,$23,$FC,$C9,$CF,$F1,$5F
	.dc.b	$98,$43,$E6,$75,$7E,$6F,$8F,$F4,$AD,$E6,$FD,$24,$63,$AB,$F3,$DC
	.dc.b	$BF,$A3,$8C,$5F,$E9,$AF,$83,$FA,$FB,$E2,$47,$04,$AE,$51,$96,$C6
	.dc.b	$67,$9C,$60,$8E,$09,$7E,$44,$FF,$8A,$A7,$F8,$11,$FE,$C7,$9F,$DD
	.dc.b	$36,$76,$FE,$C4,$D9,$FC,$0C,$D4,$E5,$C9,$27,$9B,$B2,$19,$9E,$AF
	.dc.b	$EF,$63,$6E,$67,$A4,$60,$93,$C2,$7C,$27,$96,$33,$3C,$27,$53,$E0
	.dc.b	$9C,$BB,$8C,$BD,$CC,$4F,$12,$38,$21,$E7,$8E,$33,$D2,$1E,$DF,$F9
	.dc.b	$45,$E8,$6E,$B9,$B4,$E6,$7A,$8C,$10,$C1,$0F,$55,$78,$09,$1D,$23
	.dc.b	$47,$80,$B5,$E5,$6C,$F0,$7C,$27,$49,$E4,$D3,$27,$33,$07,$9A,$BD
	.dc.b	$D5,$D6,$32,$76,$9E,$6F,$12,$46,$DB,$5E,$37,$E0,$1C,$88,$6A,$79
	.dc.b	$CF,$12,$4F,$19,$F0,$1E,$88,$FA,$95,$E6,$9D,$4F,$49,$E3,$5D,$27
	.dc.b	$33,$DC,$E4,$E2,$77,$18,$9E,$04,$70,$43,$D6,$73,$2C,$72,$7F,$0F
	.dc.b	$E3,$FE,$F6,$BF,$6F,$1F,$C1,$EB,$F7,$04,$70,$43,$FB,$69,$71,$FE
	.dc.b	$1E,$E0,$8E,$09,$8F,$79,$E9,$3B,$EC,$FF,$6A,$5F,$8A,$38,$D6,$DE
	.dc.b	$CD,$72,$7F,$13,$AE,$77,$DA,$30,$70,$29,$C3,$1F,$E2,$F3,$C2,$BA
	.dc.b	$57,$AB,$0F,$1C,$E6,$F8,$CF,$5B,$E3,$1D,$5A,$F5,$C3,$D5,$3B,$6E
	.dc.b	$1D,$89,$3D,$1B,$5C,$AB,$85,$FF,$27,$8F,$EC,$3F,$0B,$FE,$14,$25
	.dc.b	$FF,$80,$3E,$13,$FC,$2F,$F0,$BF,$E1,$8F,$85,$FF,$0C,$FB,$BF,$E1
	.dc.b	$AF,$89,$FF,$0D,$FB,$FF,$E1,$CF,$89,$FF,$0E,$FC,$0F,$F8,$7B,$E2
	.dc.b	$59,$CD,$FF,$B9,$EC,$FE,$00,$FE,$EB,$3F,$DC,$D7,$C4,$F0,$8C,$73
	.dc.b	$5B,$E3,$4F,$89,$C7,$27,$6C,$7A,$7C,$36,$31,$47,$7F,$89,$C3,$E1
	.dc.b	$F8,$8B,$89,$B3,$8B,$CE,$3A,$CF,$7A,$70,$CF,$DD,$9F,$F8,$9B,$6F
	.dc.b	$84,$FF,$8A,$38,$3E,$33,$FE,$2A,$E2,$F9,$4F,$F8,$B3,$93,$FC,$87
	.dc.b	$CB,$F4,$23,$FC,$21,$CD,$F5,$78,$1F,$13,$E6,$24,$FC,$BC,$FE,$04
	.dc.b	$FF,$8E,$B8,$FE,$2F,$C4,$FB,$3B,$3D,$BE,$7F,$BB,$D9,$F7,$FA,$3F
	.dc.b	$03,$D9,$F8,$3E,$9F,$C2,$F6,$7E,$1F,$AB,$F1,$3C,$1F,$8B,$EB,$7C
	.dc.b	$6F,$67,$E4,$FF,$00,$BD,$9F,$8F,$C7,$F9,$7E,$27,$7E,$A7,$84,$F9
	.dc.b	$CF,$A7,$2C,$3B,$46,$7C,$1F,$C2,$67,$E3,$8C,$53,$5F,$44,$63,$7F
	.dc.b	$63,$AE,$1F,$80,$F0,$E3,$FB,$BE,$7F,$C8,$8F,$C7,$73,$FC,$D3,$1C
	.dc.b	$DF,$8C,$77,$1C,$4F,$0F,$37,$CA,$7C,$7C,$DF,$29,$F2,$F3,$7C,$A7
	.dc.b	$CF,$CD,$F2,$9F,$4F,$37,$CA,$7D,$7C,$DF,$15,$FD,$F7,$8F,$FE,$20
	.dc.b	$E7,$F9,$91,$A3,$E0,$F2,$FC,$67,$F4,$DC,$9F,$15,$FD,$47,$1F,$C4
	.dc.b	$7F,$55,$C7,$F1,$1F,$D6,$71,$FC,$47,$F5,$DC,$7E,$E7,$E1,$9F,$3B
	.dc.b	$87,$E2,$34,$87,$ED,$78,$FF,$4F,$3C,$F8,$FF,$67,$7D,$DF,$DA,$E1
	.dc.b	$E3,$7F,$6F,$8F,$9D,$FD,$CE,$5E,$B7,$F7,$79,$FE,$1B,$2F,$F3,$A6
	.dc.b	$9F,$09,$FE,$41,$BF,$EE,$36,$FF,$16,$DF,$F0,$AA,$D3,$FC,$13,$7F
	.dc.b	$9D,$7F,$C2,$EF,$E9,$3F,$C3,$78,$3E,$B3,$FC,$3F,$87,$ED,$3F,$C4
	.dc.b	$78,$BE,$F3,$FC,$4F,$8F,$DE,$FE,$4B,$8B,$F0,$FD,$1F,$B0,$E4,$FE
	.dc.b	$55,$AB,$DA,$FD,$29,$2F,$F3,$90,$F8,$AF,$D4,$FC,$4F,$EA,$BE,$27
	.dc.b	$F5,$7E,$F7,$EB,$3E,$27,$F5,$BF,$01,$FA,$EF,$89,$FD,$7F,$C0,$FE
	.dc.b	$C3,$E2,$7F,$63,$F0,$7F,$B2,$F8,$9F,$D9,$FC,$2F,$ED,$3E,$27,$F6
	.dc.b	$BF,$05,$FB,$4E,$1F,$E5,$9E,$97,$F1,$E3,$1D,$BF,$26,$78,$4F,$9A
	.dc.b	$FA,$3E,$8C,$7C,$0F,$AB,$3F,$23,$EC,$D7,$D0,$FB,$B7,$F6,$3F,$03
	.dc.b	$87,$F8,$3F,$17,$C3,$8F,$BB,$C9,$F8,$E7,$F2,$24,$D3,$F9,$EF,$3F
	.dc.b	$D0,$8F,$A4,$7E,$F4,$FD,$89,$FC,$04,$FD,$89,$FC,$1B,$C1,$FC,$77
	.dc.b	$47,$F2,$4F,$07,$F1,$CF,$F2,$6F,$07,$F1,$D8,$7E,$27,$C7,$F8,$73
	.dc.b	$FD,$9B,$C6,$FE,$33,$8F,$D8,$9D,$CF,$18,$FD,$D9,$FC,$53,$0E,$EF
	.dc.b	$F1,$CC,$7F,$41,$3F,$CD,$3C,$4F,$D2,$3F,$CD,$BC,$5F,$D4,$3F,$CE
	.dc.b	$3C,$6F,$D9,$67,$FC,$47,$0F,$D7,$FA,$EE,$5F,$C7,$47,$EC,$A7,$4F
	.dc.b	$EC,$CF,$E2,$C7,$C6,$3F,$7E,$7C,$63,$F8,$19,$F1,$8F,$E0,$E7,$F1
	.dc.b	$63,$D4,$7B,$0F,$EE,$5B,$7C,$07,$FB,$4F,$7F,$EB,$B1,$FF,$05,$F8
	.dc.b	$FF,$C5,$F7,$D7,$91,$FD,$EF,$0F,$9F,$87,$F3,$EE,$0F,$8B,$3F,$DF
	.dc.b	$F8,$3E,$2C,$FF,$42,$E0,$F8,$AF,$F4,$4E,$0F,$8A,$FF,$46,$E0,$F8
	.dc.b	$BA,$7C,$28,$FC,$3F,$07,$C5,$7F,$A5,$F8,$5F,$B4,$69,$FB,$73,$E2
	.dc.b	$F0,$7F,$82,$3C,$3F,$DE,$BF,$E0,$AF,$13,$F8,$0B,$FE,$0C,$F1,$7F
	.dc.b	$82,$BF,$E0,$EF,$1B,$F8,$4B,$FE,$10,$F1,$FF,$86,$BF,$C3,$F0,$FC
	.dc.b	$31,$FC,$EA,$3F,$8B,$1F,$CF,$5F,$F0,$E7,$91,$C3,$4F,$EE,$F3,$D1
	.dc.b	$FE,$F3,$7E,$07,$FB,$DE,$1E,$47,$FB,$EE,$3E,$87,$FB,$FE,$5F,$A5
	.dc.b	$CB,$F4,$F9,$7E,$16,$7E,$36,$9F,$17,$85,$CD,$FC,$56,$BF,$89,$F5
	.dc.b	$7E,$CF,$0F,$E5,$5D,$7F,$C5,$27,$F8,$BF,$91,$FC,$51,$FE,$33,$E4
	.dc.b	$7F,$14,$7F,$8D,$F9,$1F,$C5,$1F,$E3,$BE,$47,$F1,$47,$F8,$FF,$91
	.dc.b	$FC,$51,$FE,$43,$E4,$7F,$14,$7F,$BD,$F9,$3F,$C4,$DF,$EF,$BE,$4F
	.dc.b	$EC,$9F,$C7,$F9,$3F,$C5,$1F,$E8,$7E,$4F,$F1,$47,$FA,$2F,$87,$FC
	.dc.b	$51,$FE,$8F,$E4,$FF,$14,$7F,$A4,$F9,$3F,$C5,$1F,$E9,$7E,$4F,$F1
	.dc.b	$47,$FA,$6F,$93,$FC,$51,$F8,$64,$1C,$FF,$27,$99,$F1,$3F,$E2,$02
	.dc.b	$4F,$F3,$CF,$88,$FF,$11,$FC,$3F,$F8,$93,$E1,$FF,$C4,$BE,$E7,$F9
	.dc.b	$3B,$F8,$07,$FC,$B5,$6E,$BF,$F3,$0C,$3D,$ED,$7B,$FB,$F8,$8E,$1F
	.dc.b	$03,$C7,$E2,$39,$7C,$1F,$3F,$88,$E9,$F0,$BD,$7E,$23,$B7,$C3,$FE
	.dc.b	$43,$6F,$F1,$57,$AF,$FD,$9B,$C9,$FF,$32,$6E,$AE,$15,$FE,$2F,$86
	.dc.b	$1D,$63,$19,$78,$5D,$69,$E5,$78,$6D,$ED,$DF,$DD,$1F,$00,$FE,$BF
	.dc.b	$C5,$F1,$1F,$8C,$7F,$62,$E5,$F9,$2F,$CE,$EC,$F2,$FC,$47,$C7,$24
	.dc.b	$79,$7D,$BF,$95,$EE,$7C,$BF,$77,$E6,$7C,$47,$CD,$F7,$FE,$77,$C4
	.dc.b	$7C,$FF,$81,$FA,$1F,$11,$F4,$7E,$0F,$E9,$7C,$47,$FA,$03,$DF,$FF
	.dc.b	$47,$F7,$FF,$A4,$9F,$F4,$7E,$DF,$E5,$15,$F0,$DB,$F8,$7F,$CA,$7A
	.dc.b	$FF,$EA,$9E,$DF,$F3,$7F,$9F,$9C,$38,$5F,$5F,$87,$ED,$5E,$09,$F9
	.dc.b	$DA,$79,$1F,$A1,$B7,$A1,$FA,$5C,$1E,$C7,$EA,$71,$7C,$07,$E1,$1F
	.dc.b	$5F,$97,$E2,$BF,$24,$FF,$4E,$F7,$23,$3C,$1F,$8E,$7F,$25,$CD,$CE
	.dc.b	$BF,$25,$9F,$E1,$30,$F0,$46,$0E,$78,$F9,$5C,$6E,$C9,$9F,$AE,$73
	.dc.b	$0D,$3E,$04,$70,$E1,$F9,$F9,$7F,$AD,$5F,$9F,$19,$DF,$E5,$3F,$EC
	.dc.b	$1E,$0F,$CD,$3E,$0F,$03,$F3,$4F,$FA,$22,$37,$C1,$F6,$6B,$E8,$4F
	.dc.b	$D9,$C3,$C3,$3F,$AE,$C7,$CC,$FE,$C3,$3F,$63,$FB,$2D,$7E,$0B,$FB
	.dc.b	$4D,$BE,$33,$FB,$6D,$FE,$5E,$1F,$96,$C7,$FD,$B7,$E4,$FF,$05,$EA
	.dc.b	$FC,$DD,$FF,$94,$DF,$F5,$2F,$91,$FA,$1A,$FF,$2D,$F6,$7E,$95,$F8
	.dc.b	$E7,$EA,$4F,$D5,$9F,$B3,$F1,$BF,$80,$BF,$B5,$F1,$BF,$80,$BF,$B7
	.dc.b	$F1,$BF,$80,$BF,$B9,$F1,$BF,$80,$BF,$CA,$88,$7D,$4F,$1B,$F7,$EF
	.dc.b	$D6,$F1,$BF,$7C,$FD,$8F,$1B,$F7,$CF,$DA,$F1,$BF,$7C,$FD,$CF,$1B
	.dc.b	$CC,$FA,$0F,$F2,$DF,$8F,$FC,$7D,$FF,$2E,$F9,$1F,$01,$FD,$F1,$0F
	.dc.b	$CB,$77,$7F,$52,$E1,$F8,$FC,$9F,$EF,$7F,$27,$F9,$6D,$7F,$06,$3F
	.dc.b	$D0,$37,$D5,$FF,$41,$E1,$FC,$39,$FE,$31,$E4,$FF,$28,$7E,$41,$FC
	.dc.b	$58,$F9,$07,$F1,$A3,$E4,$1F,$C7,$F5,$F8,$7C,$BF,$CC,$B7,$FC,$7A
	.dc.b	$C9,$4D,$1D,$DF,$92,$7F,$45,$7C,$AB,$F9,$6E,$1F,$33,$0F,$E7,$BC
	.dc.b	$1F,$31,$FE,$7F,$C1,$F3,$1F,$E8,$3C,$1F,$31,$FE,$87,$9F,$E1,$5F
	.dc.b	$F4,$E7,$2F,$E2,$5F,$F4,$F7,$37,$C8,$9F,$F1,$9F,$17,$99,$F9,$BE
	.dc.b	$07,$F0,$17,$F8,$31,$0F,$DA,$F4,$FE,$73,$5F,$E0,$11,$FD,$07,$C7
	.dc.b	$FE,$54,$FF,$43,$F1,$FF,$95,$3F,$D1,$7C,$7F,$E5,$4F,$F4,$7F,$1F
	.dc.b	$F8,$EB,$FC,$9F,$C6,$FE,$59,$E1,$FE,$D3,$1F,$E4,$B3,$FD,$4B,$DD
	.dc.b	$F9,$B1,$88,$FE,$09,$E4,$70,$79,$1F,$B9,$3B,$1F,$CA,$0F,$11,$FC
	.dc.b	$A0,$F3,$1F,$CA,$0F,$51,$FC,$A0,$FE,$B9,$E4,$FF,$20,$7E,$17,$0F
	.dc.b	$F2,$37,$F7,$3E,$57,$EE,$DF,$E3,$3E,$57,$EF,$9F,$E5,$7C,$1F,$1D
	.dc.b	$FE,$5B,$E2,$FF,$06,$CB,$FB,$5F,$8D,$FC,$29,$FE,$DB,$E3,$FF,$0E
	.dc.b	$7F,$B7,$F9,$1F,$C4,$9F,$E4,$07,$F1,$6B,$FE,$40,$7F,$4F,$D7,$FA
	.dc.b	$6F,$7F,$E3,$4F,$C7,$7C,$37,$FA,$4A,$7E,$39,$FD,$38,$F8,$E7,$F2
	.dc.b	$DA,$EC,$FF,$42,$BF,$13,$FD,$13,$0F,$33,$FC,$F0,$F8,$E7,$F3,$8C
	.dc.b	$BD,$DE,$07,$F3,$DD,$3E,$13,$FC,$FF,$5F,$F6,$EF,$4F,$E4,$63,$F7
	.dc.b	$57,$FB,$7E,$0F,$F7,$97,$87,$FB,$77,$FD,$E9,$E1,$FE,$DE,$7F,$A2
	.dc.b	$F8,$7F,$B7,$7F,$A3,$F9,$1F,$C7,$9F,$E9,$3E,$47,$F1,$E7,$FA,$5F
	.dc.b	$91,$FC,$79,$FE,$9B,$E4,$7F,$1E,$7F,$DB,$67,$F1,$EC,$7F,$74,$FF
	.dc.b	$23,$3F,$7C,$7F,$00,$3F,$8F,$1F,$C1,$0F,$E3,$C7,$F0,$83,$F8,$F1
	.dc.b	$FC,$30,$FF,$12,$1F,$B3,$E2,$FE,$2D,$E1,$7E,$CD,$FE,$31,$E1,$7C
	.dc.b	$3F,$23,$F8,$CD,$7F,$1E,$F2,$DF,$11,$FE,$0F,$24,$7F,$83,$5A,$FF
	.dc.b	$65,$6E,$3F,$E1,$36,$E1,$FD,$9D,$B3,$FC,$4E,$BA,$BC,$C8,$73,$2D
	.dc.b	$73,$7F,$C7,$DE,$57,$EC,$9F,$E3,$9D,$5F,$C6,$FC,$6F,$F2,$24,$70
	.dc.b	$C7,$C0,$D3,$C1,$8F,$8B,$91,$97,$8A,$7C,$04,$CB,$C0,$DB,$E6,$D3
	.dc.b	$99,$D6,$B9,$DF,$C4,$F2,$FF,$D3,$18,$7E,$87,$0F,$F1,$B7,$99,$FE
	.dc.b	$39,$3F,$C7,$5E,$9F,$F8,$EF,$E0,$7F,$C7,$9F,$11,$FE,$3D,$F8,$3F
	.dc.b	$F1,$F7,$BF,$FE,$3F,$F8,$4F,$F2,$07,$C0,$FF,$90,$7E,$17,$FC,$85
	.dc.b	$F0,$7F,$E4,$3D,$7F,$C8,$9F,$09,$FE,$45,$F8,$6F,$F2,$37,$9B,$FE
	.dc.b	$47,$F8,$7F,$C8,$70,$FF,$45,$E5,$F9,$33,$F2,$F9,$7F,$A1,$C7,$F9
	.dc.b	$26,$B8,$BF,$E4,$BB,$EA,$FF,$93,$71,$F0,$BF,$E4,$FC,$FC,$AF,$F9
	.dc.b	$47,$5F,$4B,$FE,$53,$DF,$DA,$FF,$95,$78,$7E,$0B,$FE,$57,$E3,$F8
	.dc.b	$8F,$F1,$C3,$F0,$CE,$1F,$F1,$93,$C1,$FF,$1A,$47,$47,$FC,$6D,$3E
	.dc.b	$07,$FC,$71,$5E,$47,$FC,$75,$7E,$87,$FC,$79,$87,$B1,$FF,$1F,$63
	.dc.b	$F0,$1F,$F2,$06,$5F,$23,$9B,$FB,$5E,$9F,$E6,$99,$FF,$2B,$F4,$FC
	.dc.b	$D7,$FC,$E1,$B7,$D1,$7F,$CE,$5C,$1F,$55,$FF,$3A,$70,$FD,$97,$FC
	.dc.b	$ED,$C5,$F7,$5F,$F3,$C7,$1F,$E0,$5F,$F3,$D7,$27,$E1,$5F,$F3,$E7
	.dc.b	$2F,$C3,$E5,$FC,$5C,$7C,$7D,$3F,$22,$FE,$4F,$83,$F3,$93,$F9,$A2
	.dc.b	$6D,$F9,$A6,$CF,$E6,$8F,$CE,$57,$26,$D7,$2A,$E0,$48,$E1,$5C,$AB
	.dc.b	$83,$FE,$93,$F8,$7F,$F7,$24,$FF,$1E,$3F,$CD,$4F,$EC,$6F,$A4,$72
	.dc.b	$BE,$97,$C9,$B3,$A8,$C4,$EC,$8E,$09,$1B,$27,$47,$EC,$FC,$EF,$DC
	.dc.b	$D7,$F8,$3E,$BD,$CD,$AF,$E0,$4D,$AF,$F6,$FB,$C0,$87,$C7,$34,$6A
	.dc.b	$79,$BA,$6B,$B3,$13,$D9,$B3,$D9,$D1,$F8,$C3,$FC,$20,$43,$FC,$C4
	.dc.b	$7F,$26,$3C,$27,$FB,$D2,$7B,$92,$7B,$47,$48,$C4,$F7,$6C,$7A,$67
	.dc.b	$71,$88,$E9,$18,$24,$FF,$8D,$FD,$06,$BE,$AA,$E5,$3D,$6B,$71,$82
	.dc.b	$1D,$23,$44,$72,$7B,$08,$75,$38,$1A,$3F,$67,$D7,$FD,$01,$FF,$57
	.dc.b	$71,$7F,$1B,$8C,$6B,$FC,$EA,$79,$9F,$0E,$FC,$AF,$42,$1A,$CB,$E1
	.dc.b	$36,$36,$DC,$3D,$0F,$FB,$BB,$1F,$33,$93,$CC,$77,$8D,$B9,$3B,$9C
	.dc.b	$08,$6C,$FC,$66,$1E,$AD,$3F,$88,$F6,$7F,$8B,$73,$FE,$29,$5B,$39
	.dc.b	$94,$EC,$43,$AE,$5E,$C6,$E5,$FD,$87,$4F,$B7,$1C,$4C,$E5,$E8,$69
	.dc.b	$EC,$AE,$4E,$C8,$FE,$47,$6F,$E0,$3A,$FF,$4F,$87,$9A,$E0,$FF,$40
	.dc.b	$4F,$F0,$7B,$FC,$E1,$D8,$F5,$D7,$62,$65,$F4,$A3,$47,$CB,$CB,$D3
	.dc.b	$3F,$C0,$B2,$FE,$1C,$49,$FE,$48,$D7,$F8,$EB,$6F,$F8,$D5,$FF,$08
	.dc.b	$9F,$F5,$4C,$79,$5F,$C6,$9F,$B3,$70,$70,$9E,$5A,$FF,$18,$6C,$FF
	.dc.b	$16,$AE,$8F,$80,$F1,$9F,$E2,$0E,$DF,$F0,$3C,$F0,$C3,$D5,$3A,$38
	.dc.b	$13,$0F,$03,$96,$BF,$C1,$8F,$39,$1F,$F2,$6B,$C7,$0F,$33,$C5,$AE
	.dc.b	$4F,$45,$78,$F0,$F4,$E1,$E6,$9E,$06,$8B,$87,$A1,$E4,$DC,$3C,$A4
	.dc.b	$C3,$F7,$3C,$DF,$E1,$1D,$FF,$C7,$4F,$7C,$A6,$1D,$32,$F9,$38,$71
	.dc.b	$8F,$01,$F1,$32,$71,$7F,$76,$FE,$01,$7C,$4F,$E3,$B4,$C3,$6F,$FD
	.dc.b	$C9,$3A,$7E,$FC,$F1,$F8,$9D,$6C,$A7,$C4,$E1,$7B,$D7,$E2,$71,$F0
	.dc.b	$19,$F8,$9D,$7C,$16,$FE,$27,$87,$C2,$71,$F8,$9E,$5C,$3C,$FE,$27
	.dc.b	$A7,$C3,$F5,$F8,$9F,$0F,$3B,$BB,$F5,$7C,$9F,$E1,$2E,$EF,$D5,$73
	.dc.b	$7A,$0F,$F3,$64,$F1,$71,$CB,$E9,$C7,$B4,$63,$93,$E3,$B7,$5F,$2B
	.dc.b	$9D,$FD,$2E,$F8,$7D,$AF,$1E,$3F,$85,$C9,$F0,$F9,$7E,$27,$37,$C5
	.dc.b	$E7,$97,$FE,$15,$CD,$F1,$3F,$24,$91,$E3,$F6,$3E,$67,$A1,$F3,$7D
	.dc.b	$8F,$9D,$E9,$7C,$FF,$63,$E9,$FF,$00,$B0,$BE,$CE,$3E,$1F,$3E,$D6
	.dc.b	$BE,$23,$7E,$DF,$0F,$89,$E3,$EE,$72,$F8,$AF,$9B,$DB,$F9,$1F,$3A
	.dc.b	$E9,$F2,$35,$57,$08,$C6,$1D,$1D,$65,$E0,$78,$69,$E4,$79,$1E,$93
	.dc.b	$E8,$6F,$ED,$7E,$97,$0F,$E2,$F8,$BE,$1C,$3C,$9F,$E5,$1C,$BF,$22
	.dc.b	$3E,$57,$2E,$A7,$2F,$18,$E6,$7D,$DE,$5F,$92,$FD,$FE,$5F,$92,$F9
	.dc.b	$4E,$DE,$0F,$F9,$73,$81,$AF,$AE,$72,$3C,$04,$B7,$F8,$53,$A9,$CF
	.dc.b	$07,$C7,$8C,$57,$CA,$7F,$27,$E1,$7E,$6D,$FE,$55,$8F,$F5,$CE,$3F
	.dc.b	$A3,$1C,$38,$F9,$6F,$FD,$27,$06,$3E,$08,$C6,$7E,$77,$5A,$7B,$5E
	.dc.b	$18,$7D,$13,$CD,$C0,$E3,$FA,$51,$8E,$3F,$94,$EB,$8B,$E9,$F9,$3F
	.dc.b	$A6,$E1,$FE,$05,$1F,$47,$83,$F8,$73,$C3,$2F,$AF,$87,$47,$EC,$E3
	.dc.b	$E1,$7C,$78,$FE,$37,$A7,$F3,$0F,$E3,$FA,$7F,$30,$FE,$47,$A7,$F3
	.dc.b	$0F,$E4,$FA,$7F,$30,$FC,$AE,$9F,$CC,$34,$FC,$C7,$27,$C8,$D7,$EE
	.dc.b	$71,$FC,$77,$EF,$71,$FC,$77,$F0,$1C,$7F,$1D,$FC,$15,$F9,$DF,$E5
	.dc.b	$5B,$7A,$DF,$E5,$9B,$FB,$DF,$83,$3F,$C2,$F8,$3F,$68,$B8,$BE,$47
	.dc.b	$23,$DD,$FD,$C9,$27,$F5,$DE,$67,$EF,$3D,$EF,$DE,$FB,$FF,$BE,$F8
	.dc.b	$5F,$DF,$FC,$0F,$F0,$0F,$8A,$FE,$03,$F0,$DF,$C0,$BE,$2B,$F8,$1F
	.dc.b	$C2,$FF,$04,$F8,$AF,$E0,$BF,$0F,$FC,$1B,$E2,$BF,$9C,$FF,$00,$FF
	.dc.b	$B7,$F7,$FC,$B2,$F8,$BC,$FB,$7F,$CB,$27,$F3,$DE,$9F,$F2,$C7,$71
	.dc.b	$CA,$79,$F3,$F5,$9F,$E0,$D8,$F8,$5F,$E1,$19,$F9,$5F,$E1,$5A,$FA
	.dc.b	$5F,$E1,$9B,$FB,$5F,$E1,$DC,$3F,$05,$F8,$67,$F1,$1E,$4F,$8C,$FC
	.dc.b	$BF,$5B,$E7,$BF,$95,$3E,$81,$FC,$A4,$FE,$02,$7E,$7B,$CA,$FE,$02
	.dc.b	$FE,$83,$CA,$FE,$02,$FE,$44,$FE,$2E,$7E,$44,$FE,$36,$7E,$44,$FE
	.dc.b	$3E,$7E,$44,$FC,$69,$F8,$F3,$FC,$19,$A7,$EE,$5C,$78,$DF,$B7,$F0
	.dc.b	$7F,$25,$1F,$C2,$7A,$BF,$72,$F5,$3F,$82,$F8,$1F,$93,$75,$E0,$7E
	.dc.b	$4D,$E0,$7E,$4C,$FE,$EF,$E0,$7E,$4D,$FE,$F3,$E0,$7E,$4D,$FE,$F7
	.dc.b	$E0,$7E,$E3,$6F,$8F,$9F,$F2,$69,$F8,$E7,$F0,$2F,$2F,$FA,$7B,$FC
	.dc.b	$9C,$FE,$6A,$7F,$27,$3F,$9C,$9F,$1C,$FE,$37,$8F,$9F,$C8,$FE,$47
	.dc.b	$9F,$EB,$67,$DD,$E1,$FC,$09,$FE,$4F,$B7,$F1,$CC,$7F,$CC,$5E,$DF
	.dc.b	$F9,$E6,$FE,$3F,$07,$F3,$1A,$E1,$C1,$F1,$A7,$F9,$4F,$8F,$FC,$11
	.dc.b	$FE,$57,$E4,$7F,$08,$7F,$96,$F9,$3F,$71,$FB,$C7,$A8,$FC,$11,$F8
	.dc.b	$4C,$BF,$A5,$EF,$FC,$66,$3E,$39,$FD,$30,$F8,$B1,$F1,$DF,$8C,$71
	.dc.b	$CB,$FC,$59,$CB,$F8,$C9,$FE,$AB,$CB,$E1,$7F,$9B,$61,$E5,$7F,$9C
	.dc.b	$63,$E9,$7F,$9D,$65,$FC,$73,$1F,$E8,$BE,$4F,$F1,$C7,$FA,$57,$99
	.dc.b	$FD,$15,$FC,$69,$FD,$47,$77,$99,$F2,$F5,$F9,$13,$F9,$53,$FA,$7F
	.dc.b	$17,$D3,$E2,$FA,$B1,$F9,$D3,$EC,$9F,$6F,$3F,$E6,$6F,$C8,$3F,$A2
	.dc.b	$F4,$7E,$36,$7F,$C8,$1D,$1F,$8D,$7F,$A0,$CF,$E2,$BC,$AF,$E3,$BD
	.dc.b	$5F,$B6,$C3,$FA,$B3,$FE,$16,$F2,$BF,$79,$5F,$D8,$BB,$3F,$80,$3F
	.dc.b	$D9,$3B,$7F,$82,$3F,$D9,$BB,$BF,$84,$3F,$DA,$3B,$FF,$86,$3F,$DA
	.dc.b	$BC,$0F,$E2,$0F,$F1,$AE,$0F,$E2,$8F,$F8,$FB,$C4,$78,$7F,$C6,$F7
	.dc.b	$FF,$57,$FB,$9F,$C7,$AB,$5E,$27,$F1,$A7,$FC,$3D,$E3,$7F,$28,$7F
	.dc.b	$C4,$1E,$3F,$F2,$C7,$FC,$45,$E4,$7F,$30,$7F,$C4,$9E,$4F,$F3,$47
	.dc.b	$FC,$4D,$E5,$7F,$38,$7F,$C5,$1E,$5F,$C1,$7F,$8D,$10,$FE,$34,$7F
	.dc.b	$82,$F5,$7C,$5F,$F3,$62,$47,$AF,$DD,$FE,$73,$F1,$5F,$CE,$B6,$FE
	.dc.b	$77,$F1,$5F,$CF,$3E,$0B,$F9,$EF,$C5,$7F,$3E,$F8,$3F,$E7,$FF,$15
	.dc.b	$FD,$03,$E1,$7F,$A0,$FC,$57,$F4,$2F,$87,$FE,$87,$E4,$7F,$44,$F8
	.dc.b	$8F,$E8,$BF,$15,$F1,$78,$3F,$1F,$D4,$F2,$BF,$1F,$1F,$37,$B3,$6F
	.dc.b	$CF,$BE,$6F,$D1,$C7,$BB,$F4,$F3,$F1,$BF,$57,$5F,$3B,$F5,$F7,$F5
	.dc.b	$BF,$67,$87,$DE,$FC,$13,$EE,$72,$7C,$47,$E4,$E8,$E5,$F2,$C7,$F5
	.dc.b	$CF,$2B,$FA,$F9,$FD,$83,$E2,$BF,$B0,$FB,$5F,$E9,$AF,$0F,$FD,$3D
	.dc.b	$5F,$E3,$75,$EE,$E7,$DE,$D7,$BF,$BF,$80,$E1,$F1,$3C,$7E,$0B,$97
	.dc.b	$C4,$F3,$F8,$4E,$9F,$13,$D7,$E1,$BB,$7C,$17,$CB,$F4,$FF,$8F,$F1
	.dc.b	$7F,$AA,$BB,$7C,$F0,$F3,$7F,$CF,$5C,$DE,$BE,$EE,$3C,$9F,$E9,$8E
	.dc.b	$46,$7D,$A3,$1A,$F8,$9D,$6F,$E6,$78,$70,$FA,$9E,$5C,$7F,$07,$93
	.dc.b	$E1,$7C,$37,$C4,$9F,$D4,$73,$FC,$87,$F5,$5C,$3F,$E1,$AF,$57,$FC
	.dc.b	$81,$D3,$E8,$E1,$F3,$B6,$BF,$35,$5F,$3D,$FD,$97,$4F,$67,$D5,$97
	.dc.b	$CD,$7D,$9D,$3F,$35,$F7,$74,$FC,$D7,$E0,$74,$FC,$D7,$E1,$74,$FC
	.dc.b	$D7,$E2,$74,$FC,$DD,$3F,$B8,$47,$E6,$34,$FE,$02,$E8,$F9,$5C,$9F
	.dc.b	$25,$FC,$07,$27,$C9,$7F,$05,$C9,$F2,$5F,$C2,$72,$7C,$97,$F0,$DC
	.dc.b	$9F,$25,$FA,$9C,$9F,$25,$FF,$15,$F2,$7C,$97,$F2,$47,$FA,$D7,$93
	.dc.b	$E4,$F5,$7E,$67,$A7,$F0,$13,$FC,$06,$3F,$C6,$BE,$47,$E8,$23,$FC
	.dc.b	$6B,$D9,$FA,$47,$FC,$6F,$DB,$FA,$87,$FC,$73,$DD,$FA,$C7,$FC,$77
	.dc.b	$DF,$FB,$07,$FC,$7B,$E0,$7E,$D1,$FF,$1F,$F8,$3F,$B8,$7F,$41,$F1
	.dc.b	$3F,$C9,$7E,$1B,$FC,$7D,$AF,$93,$C2,$78,$2A,$E1,$39,$E8,$63,$D6
	.dc.b	$31,$97,$85,$D6,$7E,$57,$86,$9E,$97,$96,$BE,$D7,$A6,$DF,$05,$ED
	.dc.b	$BF,$C7,$E9,$FC,$7F,$0F,$E4,$FA,$AF,$17,$F0,$42,$1F,$E5,$CA,$E5
	.dc.b	$5F,$96,$C3,$B3,$F9,$8C,$7C,$4F,$E6,$B2,$F3,$3F,$9C,$CF,$D4,$FE
	.dc.b	$7B,$4F,$73,$FA,$0D,$7E,$1B,$FA,$2F,$0B,$FD,$23,$9F,$F9,$98,$FE
	.dc.b	$A5,$E2,$7F,$2D,$7F,$AA,$78,$BF,$CC,$5F,$EA,$DE,$37,$F3,$57,$FA
	.dc.b	$C7,$8F,$FC,$E5,$FE,$B5,$E4,$7F,$3D,$7F,$AE,$79,$3F,$D0,$5F,$EB
	.dc.b	$DE,$57,$E2,$1F,$83,$CB,$F5,$79,$7E,$34,$7E,$D0,$FD,$D9,$0F,$CB
	.dc.b	$9F,$B6,$3F,$36,$7E,$D8,$FE,$0E,$7E,$D8,$FE,$16,$7E,$D8,$FE,$1E
	.dc.b	$7E,$D8,$FE,$26,$7E,$D8,$FD,$79,$FB,$13,$E2,$1F,$D3,$3C,$AF,$E1
	.dc.b	$1D,$5F,$C7,$E3,$55,$F9,$1E,$AF,$E3,$F1,$CC,$FD,$B9,$DC,$FD,$B9
	.dc.b	$E3,$3F,$6E,$79,$CF,$DB,$9E,$B3,$F9,$F7,$77,$ED,$E7,$F3,$5D,$DF
	.dc.b	$B7,$7F,$39,$E5,$7C,$8C,$3F,$A6,$39,$E3,$FE,$3E,$FF,$1D,$C3,$FD
	.dc.b	$53,$E5,$7F,$1D,$7F,$A5,$79,$5F,$C7,$5F,$E9,$9E,$57,$F1,$D7,$FA
	.dc.b	$77,$95,$FC,$75,$FE,$A1,$E5,$7F,$1D,$7F,$A9,$79,$5F,$C7,$5F,$EA
	.dc.b	$9E,$0F,$F1,$D7,$FC,$81,$97,$F4,$93,$1A,$7E,$DA,$78,$6B,$F2,$6F
	.dc.b	$A3,$F2,$F0,$F0,$3F,$37,$1F,$23,$F3,$F2,$F4,$3F,$47,$3F,$63,$FB
	.dc.b	$FD,$3E,$03,$F1,$B4,$FC,$57,$8B,$FC,$71,$FC,$8E,$DF,$B4,$7F,$64
	.dc.b	$7E,$58,$FE,$F8,$7E,$68,$FF,$00,$1F,$9E,$3F,$C0,$87,$E8,$8F,$F0
	.dc.b	$41,$FA,$63,$FC,$18,$7E,$A8,$FF,$08,$1F,$AE,$3F,$B3,$78,$9F,$C6
	.dc.b	$FC,$2F,$EF,$CF,$ED,$A3,$FC,$5D,$DD,$FB,$A7,$FC,$61,$DF,$FB,$E7
	.dc.b	$FC,$65,$E0,$7F,$02,$7F,$C6,$9E,$0F,$F0,$67,$FC,$6D,$E1,$7F,$0A
	.dc.b	$7F,$C7,$18,$FF,$0E,$7F,$C7,$5E,$1F,$F1,$77,$F9,$E1,$FD,$07,$93
	.dc.b	$FB,$EF,$4F,$F2,$B8,$F3,$90,$FE,$66,$5D,$FF,$45,$AF,$AA,$31,$B7
	.dc.b	$28,$C4,$70,$24,$F0,$76,$39,$BF,$9D,$CF,$09,$D4,$F0,$AF,$ED,$31
	.dc.b	$FE,$16,$8E,$35,$C1,$B7,$E9,$78,$1D,$5C,$4F,$42,$57,$9E,$78,$DF
	.dc.b	$86,$31,$1C,$AF,$B3,$6F,$B7,$03,$1F,$93,$F1,$1F,$E3,$92,$57,$F8
	.dc.b	$97,$E2,$7F,$C7,$9F,$11,$FE,$3D,$F8,$4F,$F1,$F7,$B9,$FE,$3F,$F8
	.dc.b	$4F,$F2,$07,$BD,$FE,$41,$F8,$46,$33,$87,$FB,$07,$F0,$0E,$23,$1C
	.dc.b	$3F,$E3,$48,$7C,$2E,$6B,$7F,$13,$6B,$5F,$0D,$70,$FA,$18,$3E,$27
	.dc.b	$1C,$A8,$7C,$3E,$1F,$F6,$F6,$5F,$DB,$7D,$1F,$E9,$6D,$C3,$C6,$74
	.dc.b	$F8,$6E,$17,$C7,$1E,$59,$73,$D3,$A6,$BD,$70,$F8,$07,$F9,$2F,$7F
	.dc.b	$88,$FF,$93,$78,$7E,$43,$FE,$4F,$E3,$FF,$33,$72,$7C,$E8,$FF,$33
	.dc.b	$72,$FF,$25,$E8,$39,$DF,$F9,$C3,$C2,$79,$1F,$E5,$9E,$6F,$A6,$FF
	.dc.b	$96,$F9,$BE,$9B,$FE,$5D,$E6,$F3,$3E,$EE,$5F,$8D,$B7,$AE,$31,$CA
	.dc.b	$E1,$F8,$11,$8E,$3F,$95,$C9,$FC,$93,$CA,$FE,$9A,$7F,$4E,$F4,$FF
	.dc.b	$C8,$74,$DF,$EF,$D7,$CA,$8E,$BC,$DF,$29,$EF,$CD,$F2,$9F,$09,$F2
	.dc.b	$8F,$F4,$47,$37,$CA,$7F,$D1,$5C,$DF,$29,$FF,$46,$73,$7C,$37,$E4
	.dc.b	$9F,$E7,$BE,$4F,$95,$D7,$FA,$38,$FD,$53,$78,$BF,$47,$A7,$EA,$A7
	.dc.b	$27,$EA,$88,$7E,$9E,$78,$73,$7E,$CB,$A7,$F5,$B5,$C6,$F9,$C7,$FA
	.dc.b	$3B,$D4,$FF,$1F,$B7,$BB,$8E,$13,$1E,$B8,$71,$C7,$A4,$78,$8F,$01
	.dc.b	$AB,$EE,$47,$53,$FC,$3B,$B7,$F7,$51,$C4,$FD,$76,$5F,$C0,$1B,$7F
	.dc.b	$C3,$5B,$87,$EC,$DA,$53,$91,$FC,$28,$C9,$B9,$E8,$E4,$87,$02,$3A
	.dc.b	$6C,$F8,$5D,$C6,$4F,$F7,$17,$95,$FD,$C6,$3F,$9C,$1E,$02,$1C,$27
	.dc.b	$93,$D2,$BD,$0D,$3D,$55,$D6,$76,$D9,$DB,$67,$73,$A3,$F9,$97,$6B
	.dc.b	$C1,$F0,$5F,$09,$D3,$49,$5C,$9A,$FC,$78,$DB,$92,$19,$8D,$BE,$E3
	.dc.b	$D4,$64,$F5,$F3,$7E,$C3,$1F,$EA,$B1,$BC,$7D,$4E,$C8,$66,$FC,$2E
	.dc.b	$0F,$59,$E2,$35,$7E,$27,$FA,$4E,$3F,$05,$F1,$9A,$38,$1C,$8E,$86
	.dc.b	$8C,$1B,$33,$5E,$67,$D1,$87,$AD,$FB,$E7,$02,$1C,$C9,$1C,$5C,$B5
	.dc.b	$F8,$F3,$E7,$8C,$12,$FE,$0C,$F8,$C8,$72,$72,$DB,$F8,$4F,$F7,$DD
	.dc.b	$BF,$76,$DD,$BE,$8B,$B3,$E1,$9E,$6A,$E6,$DA,$E7,$7E,$37,$F0,$B8
	.dc.b	$79,$EB,$F8,$0C,$FF,$0A,$AF,$DE,$ED,$FB,$F8,$C1,$27,$64,$BF,$E1
	.dc.b	$95,$FC,$20,$9B,$7F,$0D,$8F,$E4,$19,$7A,$AC,$DB,$E5,$7F,$C3,$CF
	.dc.b	$09,$27,$C1,$5D,$AF,$A4,$EE,$72,$FF,$8B,$F6,$FE,$27,$18,$8F,$1E
	.dc.b	$3E,$BC,$3B,$C6,$0E,$CE,$2F,$AC,$66,$B9,$47,$F8,$A6,$3F,$A7,$57
	.dc.b	$F8,$B0,$E9,$87,$95,$B1,$E5,$BE,$91,$88,$E9,$18,$7F,$98,$61,$E8
	.dc.b	$8C,$61,$E0,$AE,$18,$79,$08,$E0,$97,$EA,$C3,$C4,$FF,$A3,$0F,$EE
	.dc.b	$1A,$7F,$AD,$30,$EA,$49,$DE,$1D,$EB,$91,$E9,$8C,$CF,$F8,$E3,$BF
	.dc.b	$FC,$51,$F0,$FF,$E1,$D2,$57,$F7,$AF,$89,$FF,$10,$7C,$47,$F8,$87
	.dc.b	$E2,$3F,$C4,$5E,$F7,$F8,$8F,$E2,$3F,$C4,$9F,$01,$FE,$25,$F8,$8F
	.dc.b	$F1,$37,$C1,$7F,$89,$F5,$FF,$14,$7C,$1F,$F8,$A7,$E2,$3F,$C5,$5F
	.dc.b	$0B,$FE,$2B,$F8,$8F,$F1,$67,$C3,$FF,$8C,$3F,$80,$5C,$FF,$E8,$5E
	.dc.b	$2F,$F0,$74,$61,$FF,$06,$CF,$17,$FC,$1F,$7D,$5F,$F0,$8E,$3E,$17
	.dc.b	$FC,$27,$9F,$95,$FF,$0A,$EB,$E9,$7F,$C2,$FB,$FB,$5F,$F0,$CF,$0F
	.dc.b	$F8,$FF,$8B,$E1,$C7,$E1,$8F,$F2,$3F,$1F,$FB,$47,$47,$C4,$FA,$49
	.dc.b	$3E,$8E,$CF,$67,$BB,$ED,$F4,$7D,$DE,$F7,$BF,$C0,$F8,$1E,$FF,$C1
	.dc.b	$F5,$3E,$17,$C0,$7C,$3F,$17,$E2,$7C,$0F,$C5,$F5,$FE,$37,$C1,$7C
	.dc.b	$7F,$67,$E4,$7C,$1F,$C9,$EC,$F9,$1C,$3E,$7F,$15,$C9,$F2,$63,$0A
	.dc.b	$B9,$47,$CE,$C3,$B3,$F4,$32,$F1,$3F,$4B,$4F,$33,$F5,$36,$F5,$3F
	.dc.b	$5B,$83,$DC,$FD,$8E,$2F,$B9,$C7,$F1,$23,$C3,$C9,$F2,$4C,$78,$FF
	.dc.b	$77,$6F,$3C,$70,$9E,$47,$E0,$B9,$7E,$4B,$F8,$4E,$5F,$92,$FE,$1B
	.dc.b	$97,$E4,$BF,$88,$E5,$F9,$2F,$E2,$B9,$7E,$4B,$F8,$CE,$5F,$92,$FF
	.dc.b	$AA,$BC,$7F,$E9,$4F,$FA,$7F,$97,$E3,$3B,$3E,$31,$CC,$F8,$C7,$73
	.dc.b	$E3,$1E,$33,$E3,$1E,$73,$E3,$1E,$B3,$DA,$7D,$4D,$BE,$1F,$37,$FA
	.dc.b	$83,$6F,$87,$3F,$ED,$53,$FD,$8D,$5F,$90,$7F,$4B,$B7,$07,$37,$F9
	.dc.b	$07,$B1,$F8,$0E,$1F,$AE,$FE,$0B,$8B,$ED,$BF,$84,$E3,$FB,$EF,$E1
	.dc.b	$B9,$3F,$06,$FE,$23,$97,$F0,$EF,$E2,$B9,$BF,$CC,$5B,$7C,$8E,$5F
	.dc.b	$F2,$D6,$BF,$22,$BF,$1F,$C1,$F4,$1F,$C8,$F0,$FD,$47,$F2,$7C,$5F
	.dc.b	$61,$FC,$AF,$1F,$DC,$7F,$2F,$C9,$F8,$07,$F3,$3C,$BE,$D7,$E4,$1F
	.dc.b	$08,$83,$C8,$FE,$93,$93,$DB,$FE,$1C,$49,$FE,$1B,$F0,$9F,$C4,$7C
	.dc.b	$CF,$E2,$5F,$07,$FC,$4F,$DE,$FF,$61,$FF,$00,$BE,$16,$FC,$0E,$3E
	.dc.b	$1B,$3F,$07,$AF,$87,$DF,$C2,$F0,$F8,$8E,$3F,$0F,$CB,$E2,$79,$FC
	.dc.b	$4F,$ED,$F8,$7F,$D5,$5E,$CF,$F6,$B3,$FB,$67,$27,$ED,$FB,$3F,$49
	.dc.b	$C0,$BE,$57,$FE,$B1,$86,$3D,$E3,$19,$F8,$DD,$6B,$E7,$78,$6F,$EC
	.dc.b	$E0,$F7,$47,$F2,$EE,$2F,$84,$FC,$43,$F9,$8F,$2F,$C7,$7E,$67,$B1
	.dc.b	$F9,$5D,$BF,$25,$1F,$BC,$3E,$8E,$DF,$BB,$8F,$F5,$DB,$7C,$9F,$E0
	.dc.b	$B3,$F9,$33,$F4,$87,$E4,$CF,$D4,$1F,$93,$3F,$58,$7E,$4C,$FD,$81
	.dc.b	$F9,$33,$F6,$87,$E4,$CF,$E8,$FE,$4F,$F2,$29,$F1,$3F,$BE,$8F,$C9
	.dc.b	$9F,$BA,$3F,$26,$7E,$E8,$FC,$99,$FB,$A3,$F2,$67,$EE,$8F,$C9,$9F
	.dc.b	$BA,$3F,$26,$7E,$EA,$3E,$07,$8F,$FB,$3F,$27,$F7,$33,$F2,$0F,$DC
	.dc.b	$F9,$39,$C7,$E4,$DF,$09,$E5,$87,$CD,$ED,$FD,$04,$FC,$BE,$EF,$D2
	.dc.b	$3F,$37,$BF,$F5,$0F,$CF,$F0,$3F,$58,$FD,$1F,$07,$F6,$0F,$D3,$F0
	.dc.b	$BE,$13,$FE,$CF,$F0,$FF,$71,$97,$F9,$17,$4F,$DC,$4E,$F4,$F8,$CF
	.dc.b	$37,$E3,$3D,$DF,$8C,$F8,$DF,$8C,$F9,$DF,$8C,$FE,$0F,$4F,$DC,$3E
	.dc.b	$F8,$FC,$41,$3C,$3F,$DB,$ED,$F1,$4F,$E8,$51,$FD,$63,$87,$F1,$35
	.dc.b	$F8,$DF,$27,$F8,$DB,$F8,$FF,$27,$F8,$DB,$F9,$1F,$27,$F8,$DB,$F9
	.dc.b	$3F,$27,$F8,$DB,$F9,$5F,$27,$F8,$DB,$F9,$7F,$27,$F8,$DF,$91,$FE
	.dc.b	$C9,$CF,$FA,$14,$6B,$83,$F4,$17,$C9,$FD,$16,$1D,$9F,$D2,$63,$E2
	.dc.b	$7F,$4D,$97,$99,$FD,$46,$7E,$A7,$F5,$5A,$7B,$B4,$FD,$46,$BF,$D9
	.dc.b	$F6,$FE,$33,$D5,$FB,$19,$FD,$63,$FE,$27,$F2,$7F,$6F,$BF,$F7,$CE
	.dc.b	$CF,$DD,$BF,$E0,$1F,$0B,$F8,$C3,$FC,$00,$FE,$11,$E4,$7F,$18,$7F
	.dc.b	$85,$79,$1F,$C6,$1F,$E1,$9E,$47,$F1,$87,$F8,$77,$91,$FC,$61,$FD
	.dc.b	$69,$FC,$63,$B8,$D3,$27,$87,$C7,$FE,$2E,$F8,$FC,$7F,$E2,$CF,$97
	.dc.b	$C7,$FE,$2C,$F9,$FC,$7F,$E2,$CF,$A7,$C7,$FE,$2C,$FF,$90,$BC,$9F
	.dc.b	$E6,$AF,$F9,$13,$B7,$E1,$3F,$C5,$88,$7E,$BF,$C5,$70,$3E,$2B,$F9
	.dc.b	$D9,$27,$F9,$57,$91,$FC,$FB,$DA,$FE,$7F,$ED,$FF,$40,$F8,$AF,$E8
	.dc.b	$3E,$EF,$F4,$2F,$8A,$FE,$87,$EF,$FF,$44,$F8,$AF,$E8,$BF,$03,$FD
	.dc.b	$1B,$E2,$BE,$1E,$9F,$E3,$1B,$AD,$F8,$AC,$F8,$1F,$E1,$F9,$7C,$26
	.dc.b	$3C,$5F,$F0,$04,$3E,$27,$18,$39,$12,$F8,$6F,$E3,$D7,$C1,$DC,$EA
	.dc.b	$ED,$76,$BF,$AF,$8F,$87,$2C,$C7,$A4,$FE,$99,$A7,$B5,$FE,$9D,$B7
	.dc.b	$C1,$7F,$A8,$6F,$F1,$5F,$EA,$5C,$3F,$25,$F9,$67,$F5,$5E,$4F,$9C
	.dc.b	$FD,$3C,$9F,$13,$E4,$24,$79,$37,$F8,$3F,$11,$E8,$F6,$BD,$3E,$DF
	.dc.b	$AB,$DC,$F5,$FC,$47,$B3,$DE,$F6,$FC,$47,$BB,$E0,$3D,$FF,$11,$F0
	.dc.b	$3E,$0B,$E0,$FC,$47,$C2,$F8,$4F,$87,$F1,$1F,$13,$E1,$BF,$B9,$70
	.dc.b	$FF,$99,$FD,$FF,$E4,$6F,$F8,$BB,$8B,$4F,$94,$FA,$5C,$8B,$E9,$19
	.dc.b	$ED,$FA,$B9,$78,$63,$14,$D3,$CD,$18,$DB,$D4,$EB,$83,$DC,$F0,$E2
	.dc.b	$F8,$4F,$2E,$4F,$8E,$FE,$37,$97,$F1,$FE,$AF,$FB,$9B,$2F,$9B,$3E
	.dc.b	$09,$DF,$6F,$E3,$EF,$9C,$FE,$57,$0E,$EF,$E5,$F1,$F1,$BF,$99,$CB
	.dc.b	$CE,$FE,$6F,$3F,$63,$F9,$DD,$3E,$0B,$F9,$FD,$BE,$3B,$FE,$8E,$DF
	.dc.b	$F9,$56,$7F,$EE,$F3,$FB,$9C,$7C,$4E,$1F,$A6,$FC,$6E,$2F,$AE,$FC
	.dc.b	$8E,$3F,$B6,$FC,$AE,$4F,$BE,$FC,$CE,$5F,$C1,$BF,$3B,$9B,$F0,$EF
	.dc.b	$D0,$E7,$FD,$0F,$17,$C2,$D3,$F5,$BC,$7F,$23,$C5,$FF,$37,$78,$DF
	.dc.b	$C1,$70,$FC,$C1,$0F,$E5,$51,$F9,$D7,$F9,$67,$5F,$E8,$5F,$E5,$DD
	.dc.b	$9F,$A5,$7F,$98,$76,$FE,$A5,$FE,$65,$DD,$FA,$D7,$F8,$21,$FB,$4E
	.dc.b	$FF,$DA,$F8,$5F,$B6,$F0,$7F,$6F,$E1,$7E,$F2,$BF,$1E,$E5,$FC,$5C
	.dc.b	$F1,$7F,$1A,$75,$E7,$FF,$23,$E1,$E1,$7F,$C9,$38,$F9,$5F,$F2,$5E
	.dc.b	$5E,$97,$FC,$9B,$9F,$B5,$FF,$27,$E9,$FB,$87,$FC,$A3,$E1,$7C,$7C
	.dc.b	$7F,$89,$6D,$F1,$DF,$E9,$D1,$BE,$0F,$97,$7C,$DF,$9B,$87,$77,$E7
	.dc.b	$E3,$E3,$7E,$8E,$5E,$77,$E9,$E7,$EB,$7F,$91,$E3,$EF,$3F,$96,$69
	.dc.b	$F0,$DF,$E5,$DA,$FC,$7F,$07,$F9,$2B,$9D,$BF,$97,$57,$17,$F9,$85
	.dc.b	$75,$3D,$78,$78,$5F,$6E,$3E,$57,$DF,$97,$A5,$F8,$39,$FB,$5F,$87
	.dc.b	$A7,$C1,$7E,$2E,$BF,$1F,$4F,$E6,$FC,$3F,$91,$9F,$E1,$2F,$E5,$1F
	.dc.b	$CB,$1F,$BF,$3F,$34,$7F,$03,$3F,$3C,$7F,$07,$3F,$44,$7F,$0C,$EC
	.dc.b	$FD,$9E,$9F,$D2,$FB,$3F,$66,$FF,$4D,$EC,$FD,$9B,$FB,$0A,$FE,$3D
	.dc.b	$E3,$FE,$D3,$8B,$FC,$F4,$68,$FE,$A3,$5C,$9F,$EA,$77,$D9,$FE,$AB
	.dc.b	$87,$F0,$4C,$3F,$AD,$77,$7F,$08,$7F,$AE,$77,$FF,$0C,$7F,$AF,$78
	.dc.b	$1F,$C4,$1F,$E2,$59,$FF,$15,$21,$FD,$17,$4F,$8D,$AF,$C8,$8F,$E4
	.dc.b	$07,$C6,$3F,$8C,$1F,$18,$FE,$30,$7C,$63,$F8,$C1,$F1,$8F,$E3,$07
	.dc.b	$F0,$B3,$F9,$9E,$7E,$D7,$F9,$DE,$9F,$05,$FE,$80,$7F,$3B,$F0,$FF
	.dc.b	$A1,$D3,$DF,$FF,$5A,$6D,$FD,$DB,$C2,$FE,$33,$7F,$DF,$FC,$2F,$E3
	.dc.b	$2F,$F8,$07,$C2,$FE,$32,$FF,$80,$FC,$2F,$E3,$2F,$F8,$17,$C2,$FE
	.dc.b	$32,$FF,$81,$FC,$2F,$E3,$2F,$F8,$27,$C2,$FC,$33,$FC,$FC,$F8,$FE
	.dc.b	$23,$A7,$F8,$CF,$4F,$E4,$77,$FF,$0F,$BF,$ED,$88,$FC,$B4,$7F,$88
	.dc.b	$BA,$BF,$34,$FF,$89,$3A,$FF,$3C,$FF,$89,$BB,$3F,$44,$FF,$8A,$3B
	.dc.b	$7F,$4C,$FF,$84,$32,$FD,$57,$17,$F8,$5B,$BB,$F5,$CF,$F8,$46,$3F
	.dc.b	$C8,$43,$A5,$1F,$E4,$6A,$7C,$33,$D0,$FE,$3F,$7F,$D9,$BC,$3F,$E4
	.dc.b	$6F,$F6,$8F,$13,$FC,$2A,$FF,$B7,$FC,$7F,$F1,$7B,$FE,$E1,$F2,$3F
	.dc.b	$C6,$6F,$FB,$8F,$C8,$FF,$1B,$BF,$EE,$5F,$27,$FC,$76,$FF,$B9,$FC
	.dc.b	$AF,$88,$FF,$3D,$DB,$F8,$EB,$FE,$5A,$E6,$FE,$D5,$F1,$1F,$E2,$82
	.dc.b	$47,$F7,$A6,$BF,$E3,$D6,$CF,$F8,$ED,$BC,$1F,$E3,$A6,$CF,$F8,$21
	.dc.b	$BC,$1F,$E3,$A6,$E4,$6F,$93,$FC,$6E,$3F,$C7,$D5,$E1,$21,$E1,$2E
	.dc.b	$1E,$16,$E1,$E1,$69,$E1,$2E,$1E,$16,$D7,$85,$B8,$7B,$B1,$FF,$06
	.dc.b	$7C,$47,$F9,$28,$93,$FE,$1F,$FE,$01,$FD,$1B,$FE,$55,$7F,$DB,$78
	.dc.b	$FD,$F7,$FE,$00,$B7,$DC,$FF,$2A,$C3,$E0,$EF,$BD,$8E,$5C,$FB,$FA
	.dc.b	$F8,$7D,$FC,$0F,$0F,$87,$E3,$F0,$7C,$BE,$1F,$9F,$C2,$FE,$2B,$7F
	.dc.b	$E3,$A7,$EE,$7D,$5C,$73,$7E,$4A,$3C,$D3,$C5,$F5,$5F,$57,$D9,$8F
	.dc.b	$85,$F7,$67,$E5,$7E,$06,$BE,$AD,$BD,$7C,$1E,$CF,$88,$F6,$FC,$37
	.dc.b	$BB,$8F,$E3,$E1,$FE,$83,$E4,$FE,$89,$CE,$73,$BF,$FA,$0D,$E1,$3F
	.dc.b	$36,$FA,$3F,$3F,$1F,$03,$E2,$3E,$97,$27,$C8,$7E,$A7,$27,$C8,$7E
	.dc.b	$B7,$27,$C8,$7E,$C7,$27,$C8,$7F,$CA,$BD,$3F,$D0,$DF,$F4,$77,$A0
	.dc.b	$E6,$F9,$3B,$FD,$BE,$6F,$91,$5F,$BD,$8F,$BD,$CD,$F2,$63,$07,$E0
	.dc.b	$F9,$BE,$53,$8E,$A7,$43,$A3,$E6,$4E,$61,$D1,$EE,$8C,$71,$7C,$27
	.dc.b	$5C,$7F,$19,$F9,$87,$F9,$BB,$93,$E6,$F5,$7E,$B0,$91,$F9,$C2,$1F
	.dc.b	$A9,$27,$37,$EA,$5B,$B7,$EA,$5A,$7E,$A4,$BA,$7A,$BE,$23,$FC,$21
	.dc.b	$A1,$EF,$3A,$E8,$FD,$63,$71,$CE,$3D,$C9,$8F,$76,$CF,$76,$E3,$FB
	.dc.b	$78,$F4,$75,$FF,$16,$7F,$D9,$A7,$94,$FF,$68,$C6,$DB,$87,$89,$B8
	.dc.b	$70,$78,$1E,$F3,$27,$49,$E4,$E4,$87,$02,$3A,$6C,$F7,$77,$1F,$1D
	.dc.b	$FE,$8B,$7F,$B3,$8F,$E8,$7D,$3F,$C1,$32,$FE,$5B,$5E,$23,$FD,$D3
	.dc.b	$3D,$49,$7D,$63,$BE,$1E,$B9,$CB,$67,$51,$89,$F6,$46,$09,$3B,$3F
	.dc.b	$DF,$3D,$4E,$9F,$4C,$EA,$B8,$D6,$DA,$43,$CF,$1B,$6C,$6D,$A6,$CF
	.dc.b	$71,$0E,$EF,$A7,$A3,$F6,$98,$3A,$3F,$69,$5F,$B8,$3F,$A1,$9C,$4F
	.dc.b	$86,$75,$39,$94,$C9,$E2,$21,$C6,$3E,$33,$9C,$7E,$13,$C0,$D6,$1E
	.dc.b	$D6,$E1,$ED,$69,$DE,$B9,$B8,$3B,$9F,$33,$0F,$6E,$1E,$C8,$FF,$3D
	.dc.b	$9E,$C3,$D2,$70,$3C,$46,$0E,$84,$30,$70,$C3,$D3,$5E,$28,$C9,$C8
	.dc.b	$8E,$5B,$87,$B5,$DC,$7E,$47,$7F,$DF,$47,$E6,$B7,$FF,$0E,$78,$5F
	.dc.b	$E1,$B8,$FE,$C9,$AF,$E6,$9E,$38,$73,$7A,$E9,$F1,$DC,$B5,$C8,$19
	.dc.b	$7A,$A7,$F8,$5E,$FF,$C1,$1B,$8F,$F0,$92,$57,$F0,$96,$E7,$FC,$26
	.dc.b	$B8,$61,$FC,$4A,$73,$7C,$AF,$81,$2F,$84,$FF,$7B,$C7,$F8,$F9,$C6
	.dc.b	$75,$5C,$6B,$71,$8B,$D4,$60,$93,$B2,$47,$F2,$8E,$0F,$E3,$95,$FC
	.dc.b	$9C,$FE,$6F,$8F,$A1,$B8,$FA,$2B,$AD,$F2,$9C,$9F,$EE,$83,$FC,$04
	.dc.b	$43,$F1,$A7,$52,$57,$56,$D7,$57,$9E,$1E,$8B,$F3,$79,$19,$9D,$9E
	.dc.b	$F3,$78,$78,$DC,$D7,$17,$35,$FD,$B1,$FE,$DF,$D5,$FB,$EA,$FF,$1F
	.dc.b	$C7,$07,$96,$3E,$52,$63,$E1,$6C,$F8,$4F,$66,$5F,$E8,$57,$FC,$69
	.dc.b	$F1,$1F,$E3,$02,$3F,$E3,$9F,$88,$FF,$19,$FC,$47,$F8,$D3,$2F,$F1
	.dc.b	$AF,$C4,$7F,$8D,$BD,$FF,$C9,$5F,$F7,$9F,$E0,$17,$C5,$63,$4F,$E7
	.dc.b	$50,$F8,$2C,$6D,$FE,$7F,$87,$C5,$63,$07,$12,$5F,$09,$98,$C7,$C5
	.dc.b	$59,$7C,$35,$F8,$AC,$7C,$46,$7E,$2B,$FA,$46,$5F,$B3,$3A,$F6,$7E
	.dc.b	$CD,$FF,$06,$75,$ED,$FD,$67,$07,$A2,$FA,$46,$37,$F4,$65,$E1,$D3
	.dc.b	$C5,$B7,$8F,$7F,$27,$07,$97,$87,$E3,$63,$FE,$63,$E3,$F9,$4F,$F9
	.dc.b	$97,$97,$FC,$BF,$CD,$F4,$23,$FC,$3D,$CF,$F3,$3A,$3E,$B7,$37,$C3
	.dc.b	$F8,$AF,$80,$4A,$F2,$1F,$23,$9D,$EC,$FC,$5E,$DF,$8D,$E7,$FC,$7F
	.dc.b	$67,$E4,$7A,$3F,$27,$D9,$F9,$5E,$07,$CB,$F6,$7E,$67,$A9,$F3,$7D
	.dc.b	$9F,$9D,$EB,$7D,$1F,$E0,$1F,$BF,$BF,$F8,$DD,$7B,$19,$F6,$BE,$7C
	.dc.b	$FF,$8E,$FC,$3F,$9D,$E1,$7F,$7B,$6F,$83,$FC,$1E,$7D,$3C,$CC,$58
	.dc.b	$77,$C7,$C1,$5E,$23,$7C,$AD,$3C,$F1,$8D,$BD,$6E,$B8,$3D,$EF,$0E
	.dc.b	$2F,$86,$F2,$E4,$FC,$57,$2F,$CB,$8F,$CD,$73,$7D,$19,$FC,$B3,$9E
	.dc.b	$1F,$CD,$BC,$27,$91,$E4,$E6,$F9,$CF,$9B,$9B,$E7,$3E,$8E,$6F,$9C
	.dc.b	$FA,$B9,$BE,$73,$EC,$E6,$F7,$BF,$13,$8F,$E6,$B7,$8F,$E6,$E7,$FE
	.dc.b	$25,$F2,$FF,$B2,$37,$C8,$CF,$3F,$FB,$AF,$0F,$97,$87,$3E,$6E,$AF
	.dc.b	$ED,$F2,$F0,$BF,$B9,$CF,$CA,$FE,$EF,$4F,$4B,$FB,$DD,$7D,$AF,$EF
	.dc.b	$F6,$FC,$73,$F5,$38,$7E,$36,$7F,$22,$3F,$DD,$1C,$35,$F9,$A7,$DA
	.dc.b	$E1,$F9,$4F,$F0,$98,$FA,$AF,$F0,$BE,$2F,$B2,$FF,$0D,$E3,$FB,$AF
	.dc.b	$F0,$FE,$4F,$C0,$BF,$C4,$79,$7F,$0C,$FE,$54,$FC,$51,$F2,$4F,$EE
	.dc.b	$27,$F3,$FE,$3F,$90,$E0,$FC,$E7,$2F,$07,$91,$F5,$4F,$C8,$CF,$82
	.dc.b	$7F,$8E,$D7,$91,$FE,$3F,$7E,$87,$F9,$0E,$1E,$C7,$F9,$1E,$3F,$01
	.dc.b	$FC,$5E,$1F,$0C,$F8,$E7,$F2,$2E,$37,$B5,$FB,$12,$47,$E1,$7C,$DF
	.dc.b	$DA,$7C,$2F,$ED,$7E,$2F,$F6,$DE,$EF,$ED,$FE,$1F,$F7,$1E,$FF,$EE
	.dc.b	$7E,$23,$F7,$5F,$03,$FB,$BF,$89,$FD,$E7,$C1,$FE,$F7,$E2,$BF,$7D
	.dc.b	$F0,$BF,$BF,$F8,$AF,$E0,$1F,$0F,$FC,$07,$E2,$BF,$99,$79,$1F,$BA
	.dc.b	$E2,$FE,$57,$EE,$E5,$FC,$C5,$71,$7F,$35,$87,$57,$F3,$99,$78,$5F
	.dc.b	$CF,$69,$E5,$7F,$41,$B7,$A5,$FD,$17,$07,$B5,$FD,$27,$17,$F1,$3E
	.dc.b	$3F,$87,$1F,$C4,$F9,$3F,$70,$FF,$23,$ED,$FC,$9B,$4F,$CB,$1F,$B8
	.dc.b	$3F,$34,$7F,$05,$3F,$70,$7F,$09,$3F,$70,$7F,$0D,$3F,$70,$7E,$08
	.dc.b	$FC,$21,$F8,$63,$F9,$27,$8B,$FC,$8D,$FC,$67,$47,$F2,$1D,$FE,$49
	.dc.b	$FB,$68,$FB,$DD,$DF,$8F,$9F,$F5,$07,$4F,$F2,$E7,$FB,$4F,$8B,$FB
	.dc.b	$67,$FB,$5F,$8B,$FB,$67,$FB,$6F,$8B,$FB,$67,$FB,$7F,$8B,$FB,$67
	.dc.b	$FB,$8F,$8B,$FB,$6C,$BF,$BB,$78,$BF,$13,$C9,$FE,$99,$1F,$90,$6B
	.dc.b	$FB,$62,$57,$F7,$A9,$E4,$F4,$8F,$90,$7F,$05,$3E,$41,$FC,$24,$F9
	.dc.b	$07,$F0,$D3,$E4,$1F,$C4,$4F,$E3,$C7,$C1,$F1,$BF,$8C,$6B,$F1,$6F
	.dc.b	$FD,$39,$E3,$FF,$A7,$BD,$CF,$F3,$6D,$FE,$D8,$FE,$38,$72,$3E,$9E
	.dc.b	$1F,$E1,$0F,$23,$F6,$D8,$FF,$83,$FC,$8F,$DB,$3F,$E1,$1F,$23,$F6
	.dc.b	$CF,$F8,$4F,$C8,$FD,$B3,$FE,$15,$F2,$3F,$6C,$FF,$19,$C3,$F6,$F8
	.dc.b	$BC,$BF,$EE,$BE,$5B,$A7,$E6,$36,$7F,$1E,$7F,$1D,$3F,$A3,$1F,$22
	.dc.b	$BE,$BB,$F2,$1F,$B6,$FC,$87,$EF,$BF,$21,$FC,$1B,$EF,$3F,$1E,$7F
	.dc.b	$1D,$3F,$C3,$67,$F1,$DB,$FF,$43,$6F,$F2,$A3,$F2,$71,$F3,$9F,$92
	.dc.b	$7E,$66,$7A,$BF,$CA,$CF,$CF,$9F,$CB,$CF,$D1,$9F,$CC,$CF,$D3,$9F
	.dc.b	$CD,$CF,$69,$EF,$3F,$1E,$7F,$1D,$3F,$19,$87,$F8,$73,$9F,$F8,$EC
	.dc.b	$7F,$5F,$EA,$FD,$DD,$7E,$8B,$AF,$F8,$EC,$FF,$63,$F2,$BF,$8E,$BF
	.dc.b	$D9,$7C,$AF,$E3,$AF,$F6,$7F,$2B,$F8,$EB,$FD,$A7,$CA,$FE,$3A,$FF
	.dc.b	$6B,$F2,$BF,$8E,$BF,$DB,$7C,$AF,$E3,$AF,$F6,$FF,$2B,$F6,$AF,$F9
	.dc.b	$1F,$83,$FC,$29,$1A,$8F,$F1,$04,$F2,$7F,$C4,$55,$D9,$FF,$12,$5F
	.dc.b	$89,$FF,$13,$61,$E6,$7F,$C5,$18,$FA,$9F,$F1,$56,$5E,$E7,$FC,$59
	.dc.b	$9F,$C2,$E0,$FE,$F5,$B7,$F1,$CD,$7F,$C4,$DF,$13,$FE,$33,$25,$FF
	.dc.b	$81,$FD,$CF,$F1,$B7,$BF,$FE,$67,$F4,$18,$DF,$80,$C7,$BF,$9F,$81
	.dc.b	$D7,$C4,$EF,$E0,$F8,$79,$5C,$7E,$13,$97,$C4,$F3,$F8,$6E,$9F,$0D
	.dc.b	$D7,$E1,$FF,$A0,$70,$FF,$8F,$F0,$F9,$2F,$F9,$17,$B3,$4F,$F8,$FB
	.dc.b	$81,$7C,$E3,$18,$F7,$75,$7E,$37,$86,$9E,$77,$96,$DE,$B7,$A7,$07
	.dc.b	$C0,$7E,$11,$FE,$29,$E3,$F8,$AF,$ED,$FC,$2F,$D1,$E6,$F8,$9F,$8C
	.dc.b	$49,$F3,$7C,$4F,$C9,$F8,$6F,$95,$EE,$FC,$BB,$F9,$9E,$F7,$CD,$F3
	.dc.b	$7E,$77,$BF,$F3,$FE,$27,$FD,$2D,$FC,$03,$FE,$A2,$CF,$FC,$BA,$FF
	.dc.b	$A8,$5B,$F0,$9A,$F8,$5D,$FC,$37,$0F,$87,$F9,$7E,$5F,$F8,$52,$CD
	.dc.b	$3F,$A2,$F5,$6E,$3A,$D7,$37,$A9,$E0,$F8,$7F,$1F,$C3,$F9,$63,$E9
	.dc.b	$6B,$E9,$7E,$A6,$FE,$D7,$EB,$70,$FC,$17,$E1,$9F,$67,$93,$E3,$3F
	.dc.b	$2C,$FF,$5B,$F1,$7C,$B7,$44,$3D,$3C,$FF,$43,$97,$A4,$60,$E3,$7E
	.dc.b	$17,$1E,$2A,$D8,$F9,$E7,$30,$CB,$D9,$1C,$38,$FE,$83,$75,$FA,$7C
	.dc.b	$7F,$16,$35,$BF,$C9,$7F,$2C,$7F,$27,$3F,$DD,$E7,$F0,$33,$EC,$F1
	.dc.b	$7D,$67,$ED,$F1,$F4,$7E,$71,$F7,$B8,$7C,$7C,$7F,$B0,$8F,$8B,$9F
	.dc.b	$A6,$31,$B7,$B7,$6F,$81,$18,$D9,$BF,$C6,$8C,$70,$7D,$07,$5C,$5F
	.dc.b	$C0,$FC,$E7,$4F,$E7,$A7,$FC,$12,$FF,$AD,$FC,$6F,$D2,$6B,$F2,$39
	.dc.b	$7F,$50,$FC,$AE,$BF,$D6,$3F,$33,$B3,$F6,$0F,$CE,$ED,$FD,$A3,$F4
	.dc.b	$3B,$BF,$70,$FD,$2E,$FF,$DE,$BF,$C0,$4F,$E5,$1F,$0D,$FE,$45,$E3
	.dc.b	$37,$89,$E0,$FF,$0A,$AF,$E0,$9E,$17,$F0,$E7,$F8,$37,$87,$FC,$49
	.dc.b	$FE,$11,$E2,$7F,$16,$7F,$85,$78,$BF,$C6,$9F,$E1,$9E,$37,$F1,$FE
	.dc.b	$1F,$7F,$27,$F8,$AB,$8F,$E2,$47,$FA,$8F,$93,$E4,$18,$BF,$F3,$73
	.dc.b	$F1,$DF,$F5,$87,$8F,$CD,$EA,$7F,$10,$3C,$27,$F1,$43,$CA,$7F,$18
	.dc.b	$3D,$27,$F1,$C3,$DA,$7F,$21,$E1,$F8,$F8,$7E,$EA,$3E,$3F,$1F,$F0
	.dc.b	$0C,$3F,$8A,$EF,$F1,$67,$F3,$FB,$FC,$57,$F4,$3B,$FC,$57,$F4,$7B
	.dc.b	$FC,$57,$F4,$BB,$FC,$57,$F4,$FB,$FA,$DF,$F7,$C6,$BE,$F7,$FD,$F5
	.dc.b	$B7,$F2,$6F,$41,$E4,$7F,$26,$E9,$FC,$86,$D6,$FF,$81,$9F,$B2,$F2
	.dc.b	$3F,$92,$BF,$CF,$79,$3F,$92,$BF,$CF,$FC,$8F,$E4,$AF,$F4,$1F,$23
	.dc.b	$F8,$9B,$FA,$83,$F7,$1E,$2F,$ED,$79,$7F,$4B,$E3,$BC,$6F,$DB,$47
	.dc.b	$C7,$AF,$E0,$BF,$0F,$FE,$E0,$D3,$FC,$3A,$D8,$FF,$54,$79,$3C,$27
	.dc.b	$FC,$43,$87,$39,$FD,$C9,$FC,$98,$FE,$0C,$7E,$E4,$F3,$1F,$C9,$8F
	.dc.b	$51,$FC,$98,$F7,$1F,$C9,$8F,$EB,$9A,$FF,$6F,$F0,$BE,$3C,$7F,$24
	.dc.b	$DF,$FB,$C6,$FF,$C9,$63,$F8,$CC,$6B,$BB,$F8,$F5,$7E,$FF,$1F,$EA
	.dc.b	$BE,$1F,$F0,$47,$FA,$BF,$89,$FC,$25,$FE,$B3,$E2,$FF,$0E,$7D,$BE
	.dc.b	$37,$F5,$0F,$27,$F8,$A6,$7F,$C8,$FC,$8F,$E3,$0F,$F2,$5F,$27,$F1
	.dc.b	$CF,$F1,$E3,$E5,$F3,$E3,$57,$0A,$B8,$F2,$7D,$4C,$3A,$BF,$5B,$1F
	.dc.b	$B7,$8F,$F3,$4E,$7F,$BE,$FF,$36,$CB,$F0,$6F,$F3,$8E,$2F,$C3,$BF
	.dc.b	$89,$3F,$17,$7F,$C7,$1F,$E9,$1D,$3F,$E4,$5E,$AF,$DB,$6B,$FE,$20
	.dc.b	$9F,$EF,$A7,$E6,$8F,$F0,$09,$FB,$63,$F8,$E1,$FB,$63,$F8,$E1,$FB
	.dc.b	$63,$F8,$E1,$FB,$63,$F8,$E1,$FB,$63,$F8,$E1,$FE,$17,$F0,$7F,$C3
	.dc.b	$77,$FE,$B9,$8F,$DB,$99,$3F,$78,$71,$7F,$9D,$78,$DF,$CA,$1F,$E7
	.dc.b	$9E,$1F,$F2,$C7,$F9,$F7,$8F,$FC,$C1,$FE,$81,$E4,$7F,$34,$7F,$A1
	.dc.b	$72,$7F,$1A,$3F,$CF,$3E,$4B,$C8,$F8,$6F,$F4,$0C,$BF,$D7,$1E,$4F
	.dc.b	$F2,$F2,$65,$E8,$BF,$F4,$0C,$62,$7D,$27,$F3,$72,$61,$ED,$6C,$F0
	.dc.b	$AE,$51,$FE,$FE,$38,$CE,$C9,$3B,$8C,$B5,$E6,$71,$2E,$BF,$D1,$2B
	.dc.b	$AB,$63,$89,$BA,$EA,$DA,$E9,$18,$9E,$13,$FE,$13,$7F,$33,$F1,$5F
	.dc.b	$E3,$22,$47,$F8,$CB,$E2,$7F,$C6,$BF,$15,$FE,$36,$F5,$BF,$C6,$FE
	.dc.b	$EF,$F8,$E3,$C7,$FF,$1C,$FB,$DF,$E3,$AF,$67,$FC,$77,$EF,$FF,$8F
	.dc.b	$3D,$BF,$F1,$EF,$C0,$7F,$8F,$BD,$DF,$F1,$FF,$C0,$FF,$90,$3D,$FF
	.dc.b	$F2,$0F,$C1,$7E,$AF,$CE,$FE,$29,$C7,$FE,$B1,$E2,$FF,$07,$9F,$E1
	.dc.b	$79,$E0,$FF,$91,$AF,$A3,$FE,$48,$C7,$C0,$FF,$92,$B3,$F2,$3F,$E4
	.dc.b	$CD,$7D,$0F,$F9,$3B,$7F,$63,$FE,$50,$E1,$FF,$29,$F1,$7C,$28,$FF
	.dc.b	$13,$71,$FF,$A2,$FC,$63,$9E,$5F,$E6,$2F,$09,$E4,$7F,$96,$B9,$7E
	.dc.b	$3B,$FE,$5C,$E5,$F8,$EF,$F9,$7B,$97,$E3,$BF,$E6,$0E,$5F,$8E,$FF
	.dc.b	$98,$B9,$7E,$3B,$FE,$64,$E5,$F8,$EF,$F8,$BC,$FE,$26,$7F,$8C,$F9
	.dc.b	$3F,$D6,$2F,$F9,$AB,$93,$E3,$BF,$E3,$FE,$4F,$8E,$FF,$90,$79,$3E
	.dc.b	$3B,$FE,$43,$E4,$F8,$EF,$F9,$17,$93,$E3,$BF,$E4,$7E,$4F,$8E,$FF
	.dc.b	$92,$79,$3E,$03,$F1,$8F,$F3,$CF,$0F,$FA,$C7,$A3,$F5,$51,$F9,$A6
	.dc.b	$F1,$FE,$4D,$BC,$FF,$96,$24,$FE,$64,$9B,$7E,$62,$B8,$74,$7E,$8A
	.dc.b	$F8,$D6,$9B,$5B,$BE,$6D,$9D,$57,$12,$3F,$AA,$8D,$12,$B9,$89,$F6
	.dc.b	$09,$84,$9F,$D6,$B6,$78,$3A,$3F,$7C,$66,$70,$E0,$87,$F0,$13,$46
	.dc.b	$A7,$AB,$A6,$BB,$31,$3E,$07,$C1,$1A,$24,$66,$3A,$9A,$70,$F2,$9F
	.dc.b	$2B,$D4,$93,$CA,$32,$D8,$F3,$4F,$27,$6D,$8D,$47,$79,$E3,$39,$23
	.dc.b	$F3,$3A,$FF,$DC,$33,$CA,$36,$47,$BB,$5E,$86,$08,$60,$87,$C0,$21
	.dc.b	$EF,$36,$60,$F0,$18,$34,$72,$29,$BA,$F6,$B6,$BD,$6F,$13,$99,$C4
	.dc.b	$B5,$EE,$9E,$31,$D0,$87,$03,$A5,$7B,$DA,$75,$8E,$2E,$CE,$2E,$09
	.dc.b	$5E,$48,$F1,$46,$27,$A4,$70,$9D,$3D,$85,$79,$E3,$31,$B9,$E2,$F8
	.dc.b	$49,$5D,$5F,$34,$F0,$72,$D8,$D4,$79,$23,$04,$9E,$B3,$A9,$F2,$1E
	.dc.b	$5A,$F2,$4F,$EF,$23,$F8,$83,$70,$FE,$13,$1F,$BF,$BF,$DE,$CE,$70
	.dc.b	$FE,$0F,$18,$9D,$4E,$5F,$E1,$26,$08,$76,$3F,$C0,$FE,$07,$FA,$11
	.dc.b	$D9,$92,$5F,$89,$B7,$E0,$AE,$D8,$FF,$11,$6E,$1E,$3A,$D9,$D2,$F9
	.dc.b	$12,$F9,$47,$92,$30,$FF,$82,$F0,$F2,$61,$D6,$35,$5C,$1E,$B1,$96
	.dc.b	$DF,$86,$F9,$CE,$5B,$19,$AE,$51,$8C,$3D,$24,$BF,$E5,$11,$FE,$29
	.dc.b	$C3,$C7,$FC,$03,$FC,$96,$3F,$DF,$ED,$BF,$E5,$D8,$A3,$16,$8C,$7C
	.dc.b	$27,$F3,$2B,$E3,$F0,$97,$0E,$38,$3D,$BC,$63,$C6,$1E,$E6,$32,$ED
	.dc.b	$F0,$B6,$5E,$F5,$F8,$6C,$7C,$06,$7E,$1F,$5F,$05,$BF,$88,$E1,$F0
	.dc.b	$9C,$7E,$27,$97,$C3,$73,$F8,$AE,$9F,$11,$38,$7E,$1F,$1F,$9A,$77
	.dc.b	$18,$3C,$D7,$D1,$C7,$0F,$C3,$CB,$C3,$18,$E0,$6B,$E6,$8C,$6F,$EB
	.dc.b	$E0,$F6,$70,$FB,$78,$BD,$DC,$7E,$FE,$4F,$81,$CD,$F0,$71,$87,$F5
	.dc.b	$FE,$7F,$AF,$F1,$7F,$14,$91,$E1,$F8,$0F,$91,$E9,$FC,$9F,$43,$E5
	.dc.b	$7C,$1F,$CD,$FE,$01,$73,$DF,$85,$C7,$83,$9F,$86,$D7,$AB,$BF,$87
	.dc.b	$E1,$E1,$F1,$F8,$8E,$5E,$C7,$3F,$89,$F9,$9C,$3F,$DE,$7B,$1E,$06
	.dc.b	$63,$E3,$EE,$BE,$51,$8C,$7B,$3A,$CF,$C4,$F0,$D7,$CC,$F2,$DF,$D6
	.dc.b	$FC,$FE,$1F,$7B,$F0,$4F,$A5,$C9,$F1,$1F,$F3,$76,$88,$C7,$8D,$F9
	.dc.b	$E7,$F1,$DC,$BC,$A9,$3F,$A2,$BE,$F1,$8C,$3F,$47,$AF,$CA,$8F,$6F
	.dc.b	$1F,$CA,$7D,$FC,$7F,$29,$F8,$3C,$7F,$29,$F8,$7C,$7F,$29,$F8,$BC
	.dc.b	$7F,$B5,$7F,$D1,$F8,$23,$51,$E1,$3F,$61,$09,$E9,$19,$F0,$12,$BF
	.dc.b	$14,$E6,$18,$79,$E3,$18,$FE,$17,$2F,$7E,$9F,$06,$7F,$39,$CD,$F1
	.dc.b	$9F,$9A,$43,$E6,$EF,$F3,$2B,$F7,$EE,$E3,$C9,$C1,$FE,$C1,$6E,$FD
	.dc.b	$AB,$E5,$9F,$7A,$3C,$DC,$3F,$82,$7D,$1C,$5F,$86,$7D,$5C,$7F,$8A
	.dc.b	$7D,$9C,$9F,$8F,$7F,$93,$C7,$FA,$DB,$7F,$95,$1F,$92,$9E,$1A,$FD
	.dc.b	$9E,$1F,$8F,$7D,$63,$16,$C3,$C5,$18,$C7,$CC,$EB,$2F,$56,$5F,$C6
	.dc.b	$F3,$FC,$54,$FF,$2A,$E4,$F8,$93,$FC,$7F,$6F,$F2,$BF,$3F,$F4,$2C
	.dc.b	$FF,$C5,$B3,$F5,$7E,$2B,$F6,$C7,$ED,$FD,$BF,$DC,$7C,$3F,$EE,$7E
	.dc.b	$1F,$F7,$5F,$11,$FB,$BF,$88,$FD,$E7,$C0,$7E,$F7,$E2,$3F,$7D,$F0
	.dc.b	$5F,$BF,$F8,$8F,$E0,$1F,$09,$FC,$B7,$81,$F1,$17,$E1,$B1,$F1,$59
	.dc.b	$F8,$8F,$DD,$F1,$7D,$8F,$47,$F9,$33,$F6,$BC,$6D,$3F,$C8,$B1,$5F
	.dc.b	$38,$C6,$3E,$07,$F8,$0E,$7E,$47,$F8,$1E,$BE,$87,$F8,$2E,$FE,$C7
	.dc.b	$F8,$3F,$0F,$C0,$7F,$84,$F1,$FC,$47,$F8,$5F,$27,$CA,$F8,$3F,$C7
	.dc.b	$C7,$E4,$FA,$7F,$78,$FF,$03,$3F,$23,$E1,$7F,$23,$7F,$9B,$79,$1F
	.dc.b	$9F,$7F,$72,$7F,$0F,$3F,$72,$7F,$13,$3F,$72,$7F,$17,$3F,$72,$7E
	.dc.b	$2C,$FD,$C9,$FD,$07,$83,$FC,$13,$E1,$FF,$22,$8C,$9F,$96,$38,$9F
	.dc.b	$9A,$3A,$9F,$9E,$3C,$27,$E8,$8F,$29,$FB,$9E,$EF,$C7,$F8,$5F,$DC
	.dc.b	$FB,$FF,$1F,$1F,$DB,$BB,$FF,$1F,$3F,$DD,$F9,$7E,$35,$FE,$44,$F8
	.dc.b	$D5,$F2,$1F,$DA,$D7,$27,$F3,$47,$C6,$3F,$98,$1E,$2D,$3F,$44,$F9
	.dc.b	$9F,$D3,$F8,$3F,$A9,$AF,$E4,$BE,$17,$EB,$9F,$E4,$FE,$1F,$ED,$6B
	.dc.b	$FC,$29,$7F,$DA,$BD,$7F,$F5,$34,$7E,$C3,$C9,$FE,$3E,$FE,$CB,$C9
	.dc.b	$FE,$3A,$FE,$D3,$C9,$FE,$36,$FE,$DB,$C9,$FE,$34,$FE,$E3,$C9,$FB
	.dc.b	$98,$FF,$10,$E4,$FC,$03,$F8,$23,$F0,$93,$F8,$67,$F8,$AF,$07,$C6
	.dc.b	$E1,$F9,$11,$F1,$8F,$90,$7C,$63,$F9,$19,$FD,$44,$F8,$C7,$D5,$3E
	.dc.b	$36,$DF,$D1,$BC,$9F,$E1,$2F,$F4,$8E,$1F,$8D,$3F,$CE,$B8,$7E,$33
	.dc.b	$FC,$F3,$87,$E3,$57,$F8,$E8,$91,$FD,$03,$83,$E2,$C7,$F5,$2E,$0F
	.dc.b	$8A,$FF,$54,$E0,$F8,$AF,$F5,$6E,$0F,$8A,$FF,$58,$E0,$F8,$BB,$FF
	.dc.b	$8F,$B8,$3E,$2B,$FE,$40,$E6,$FE,$70,$FC,$53,$E3,$79,$3F,$C5,$FA
	.dc.b	$7F,$65,$AF,$F1,$73,$F6,$A7,$F7,$EF,$23,$F8,$B3,$FE,$00,$F2,$3F
	.dc.b	$8B,$3F,$E0,$2F,$23,$F8,$B3,$FE,$04,$F2,$3F,$8B,$3F,$E0,$6F,$23
	.dc.b	$F8,$B7,$89,$FC,$2E,$3F,$AE,$74,$FF,$0E,$7F,$88,$78,$FF,$C5,$B5
	.dc.b	$FF,$2B,$F8,$FF,$C5,$9F,$F2,$CF,$81,$FC,$5A,$B8,$F8,$FC,$E7,$FA
	.dc.b	$06,$1F,$C8,$A7,$C0,$7F,$16,$3C,$87,$F1,$63,$D0,$7F,$16,$3D,$87
	.dc.b	$F1,$63,$E0,$1F,$08,$F8,$84,$E0,$FF,$3F,$C3,$E2,$BF,$99,$12,$7F
	.dc.b	$95,$7C,$57,$F3,$5F,$6B,$F9,$B7,$C4,$7F,$37,$F7,$3F,$9C,$7C,$3F
	.dc.b	$F3,$9F,$7B,$F9,$D7,$93,$FC,$EF,$DF,$FE,$79,$F0,$FF,$CF,$7E,$07
	.dc.b	$F9,$F7,$C3,$FF,$3F,$F8,$3F,$E8,$1F,$0F,$FD,$07,$E1,$7F,$A1,$7C
	.dc.b	$3F,$FA,$7B,$8C,$E7,$8B,$F9,$E5,$71,$7F,$9F,$61,$D5,$FE,$81,$97
	.dc.b	$85,$FE,$85,$A7,$95,$FE,$89,$B7,$A5,$FE,$8D,$C1,$ED,$7D,$E7,$F4
	.dc.b	$9E,$3F,$84,$B9,$3E,$3F,$C3,$FF,$57,$24,$7F,$9C,$3F,$80,$7F,$D2
	.dc.b	$3E,$79,$7B,$7F,$E8,$95,$EE,$5F,$77,$1E,$F6,$7D,$FD,$7C,$06,$FE
	.dc.b	$07,$87,$C4,$71,$F8,$3E,$5F,$11,$CF,$E1,$7A,$7C,$47,$5F,$87,$FC
	.dc.b	$87,$A9,$FB,$4E,$1F,$F5,$27,$91,$9E,$3F,$35,$F1,$7D,$18,$F5,$7D
	.dc.b	$59,$F8,$5F,$66,$BE,$57,$DD,$BF,$A5,$F8,$1C,$3E,$DE,$3F,$74,$7C
	.dc.b	$43,$F4,$BC,$9F,$C6,$7B,$8E,$7C,$7F,$CF,$BC,$27,$E5,$57,$C7,$BE
	.dc.b	$D1,$FE,$49,$E6,$FF,$59,$BF,$3F,$97,$E3,$BF,$47,$97,$E3,$BF,$4F
	.dc.b	$97,$E3,$BF,$57,$97,$E3,$BF,$5F,$97,$E3,$BF,$EE,$7E,$5F,$F5,$73
	.dc.b	$93,$E3,$9C,$4F,$8E,$75,$3E,$39,$E1,$3E,$39,$E5,$3E,$39,$F8,$5E
	.dc.b	$6F,$63,$EE,$3F,$10,$7C,$38,$FF,$6D,$EF,$F1,$A3,$FD,$59,$D7,$8E
	.dc.b	$9F,$C9,$F0,$FF,$06,$7F,$C5,$1E,$3F,$E6,$6B,$F8,$9F,$5F,$E7,$5F
	.dc.b	$E2,$BD,$9F,$A1,$7F,$8B,$E1,$FA,$57,$F8,$CF,$6F,$EA,$5F,$E3,$7D
	.dc.b	$DF,$AD,$7F,$8E,$F7,$FE,$C5,$FE,$3F,$E0,$7F,$1B,$DB,$FD,$D3,$E4
	.dc.b	$FE,$E2,$7F,$C6,$1E,$17,$EF,$1F,$F1,$97,$87,$FC,$01,$FF,$1A,$78
	.dc.b	$9F,$C1,$1F,$F1,$B7,$8B,$FC,$21,$FF,$1C,$78,$DF,$C3,$1F,$F1,$D7
	.dc.b	$8F,$FC,$41,$FF,$1E,$79,$1F,$01,$FD,$B1,$0F,$DB,$70,$7C,$77,$1C
	.dc.b	$BF,$1D,$FD,$E9,$FC,$E2,$7A,$3F,$CE,$AB,$C0,$FF,$3C,$BF,$23,$FC
	.dc.b	$FB,$0F,$43,$FD,$03,$1F,$63,$FD,$0B,$2F,$80,$FF,$44,$CF,$E2,$3F
	.dc.b	$EA,$1F,$77,$FC,$17,$B2,$1E,$42,$73,$18,$E8,$FE,$42,$FC,$C3,$F9
	.dc.b	$09,$F3,$0F,$E4,$27,$CC,$3F,$90,$9F,$30,$FE,$42,$7C,$C3,$F9,$09
	.dc.b	$F3,$0F,$E4,$27,$CC,$3E,$2E,$DF,$1D,$FE,$BB,$D1,$FB,$9F,$0B,$F2
	.dc.b	$D1,$E5,$7F,$70,$FE,$D7,$C2,$FC,$EB,$FB,$7E,$BF,$E3,$AF,$E8,$CF
	.dc.b	$DD,$78,$9F,$C3,$5F,$DE,$78,$BF,$C4,$5F,$DF,$78,$DF,$C5,$5F,$E0
	.dc.b	$1E,$3F,$F1,$97,$FC,$FC,$7F,$BE,$FC,$8C,$69,$FE,$A7,$F1,$95,$C6
	.dc.b	$31,$7D,$5D,$61,$E1,$78,$63,$FC,$77,$1F,$E1,$DE,$47,$F1,$09,$E9
	.dc.b	$E4,$7F,$1D,$9F,$DB,$9F,$E8,$DD,$BF,$8F,$79,$3F,$D1,$23,$04,$3F
	.dc.b	$8C,$1C,$4F,$E3,$53,$D7,$93,$F0,$5C,$9F,$D1,$F1,$F1,$63,$E4,$BF
	.dc.b	$E1,$87,$C6,$3F,$88,$1F,$18,$FE,$28,$7C,$63,$F8,$C1,$F1,$8F,$C8
	.dc.b	$7B,$7F,$8F,$69,$FE,$83,$37,$1F,$27,$6F,$91,$87,$93,$76,$FF,$26
	.dc.b	$31,$BF,$C9,$75,$BF,$A1,$E1,$9F,$B7,$2F,$CE,$E1,$FD,$81,$FE,$40
	.dc.b	$7C,$43,$E3,$65,$FD,$09,$FE,$3E,$7E,$87,$8F,$F8,$FD,$FE,$EE,$3F
	.dc.b	$C5,$FE,$2F,$F1,$FA,$FF,$13,$F8,$BF,$C7,$DF,$F1,$4F,$8B,$FC,$7D
	.dc.b	$FF,$15,$F8,$1F,$C7,$DF,$F1,$6F,$8B,$FC,$7D,$FF,$17,$F8,$BF,$C7
	.dc.b	$DF,$F1,$8F,$8B,$FC,$7D,$FF,$19,$F8,$BF,$C7,$F3,$FD,$BF,$C0,$3C
	.dc.b	$6F,$DE,$D7,$F8,$E3,$C0,$FE,$02,$FF,$8E,$BC,$1F,$E0,$AF,$F8,$F3
	.dc.b	$C2,$FE,$12,$FF,$8F,$BC,$3F,$E1,$AF,$F9,$03,$C4,$FE,$22,$FF,$90
	.dc.b	$BC,$5F,$E2,$AF,$F9,$13,$C6,$FE,$41,$A7,$ED,$0F,$E4,$1E,$03,$8B
	.dc.b	$EC,$7C,$47,$F8,$3C,$93,$FE,$15,$6B,$FD,$AD,$B7,$FA,$96,$F3,$F2
	.dc.b	$9E,$04,$38,$16,$78,$36,$38,$3F,$D1,$8F,$F5,$57,$7F,$E4,$65,$3F
	.dc.b	$A0,$F1,$FF,$8F,$F8,$2F,$1F,$F8,$FE,$1D,$89,$7D,$9B,$7D,$A3,$04
	.dc.b	$70,$48,$FC,$8E,$3E,$C6,$9E,$72,$9D,$0B,$9F,$CB,$7F,$33,$E2,$FF
	.dc.b	$8E,$48,$FF,$8F,$F9,$3F,$C7,$98,$FF,$8F,$7E,$03,$FC,$7D,$EB,$FF
	.dc.b	$8F,$FE,$07,$FC,$81,$EC,$FC,$1D,$3F,$7F,$FC,$03,$F0,$B2,$FE,$85
	.dc.b	$B5,$DD,$2F,$6F,$85,$7F,$46,$87,$C1,$62,$B7,$EE,$D9,$7C,$15,$F7
	.dc.b	$F1,$CD,$9F,$80,$FC,$BE,$BF,$E4,$5E,$2F,$9F,$B7,$87,$15,$70,$D7
	.dc.b	$CB,$7E,$8C,$7B,$65,$DF,$3F,$06,$BE,$1D,$BC,$57,$F0,$30,$FF,$32
	.dc.b	$70,$7C,$47,$FC,$CD,$C5,$F2,$1F,$94,$7F,$9A,$79,$3E,$6B,$FD,$E8
	.dc.b	$FA,$26,$48,$7F,$90,$F9,$39,$3F,$E4,$59,$EC,$FF,$91,$EB,$C4,$B7
	.dc.b	$F7,$61,$E7,$8F,$69,$F3,$72,$F6,$38,$E2,$FA,$7A,$FC,$B8,$C7,$07
	.dc.b	$E6,$0F,$8E,$EF,$87,$E5,$E1,$F3,$7C,$1F,$A1,$3C,$38,$F9,$3E,$39
	.dc.b	$FC,$1E,$1E,$08,$C1,$E3,$3B,$F0,$7C,$C7,$FD,$1D,$C9,$F8,$D7,$FD
	.dc.b	$21,$CB,$F9,$17,$FD,$25,$97,$FA,$3F,$87,$E5,$EF,$FD,$97,$8B,$FA
	.dc.b	$17,$37,$E7,$78,$3F,$D2,$DC,$FF,$A5,$6E,$3F,$A5,$6B,$FA,$56,$DF
	.dc.b	$EB,$E3,$F5,$2D,$8C,$ED,$FB,$16,$CE,$A3,$1E,$4A,$FC,$1C,$4F,$82
	.dc.b	$FE,$5D,$83,$C3,$FE,$5D,$8F,$D0,$9F,$36,$1C,$63,$CC,$D7,$CC,$DC
	.dc.b	$BC,$8D,$CB,$C3,$97,$22,$57,$27,$E2,$F6,$7F,$1A,$8F,$ED,$19,$7C
	.dc.b	$26,$DF,$76,$D7,$F6,$D6,$94,$FE,$1A,$60,$D4,$F3,$7B,$10,$D9,$5D
	.dc.b	$38,$23,$82,$6F,$F9,$08,$FF,$4D,$1F,$E4,$9E,$5F,$E7,$EF,$84,$87
	.dc.b	$5B,$F5,$36,$FD,$0F,$68,$CB,$63,$37,$E1,$9C,$B6,$33,$1F,$4F,$99
	.dc.b	$F0,$3D,$6F,$E2,$5E,$DB,$5C,$89,$7C,$89,$1A,$6B,$F2,$A3,$6E,$48
	.dc.b	$66,$36,$FD,$AC,$7E,$35,$7F,$9B,$F9,$FF,$95,$76,$FF,$52,$9C,$E1
	.dc.b	$D2,$F9,$1D,$0F,$92,$7C,$33,$07,$B4,$D9,$0D,$DF,$BD,$CE,$5F,$11
	.dc.b	$E4,$6C,$E4,$70,$9F,$13,$71,$F5,$BA,$21,$C8,$F7,$67,$EC,$CA,$47
	.dc.b	$F1,$58,$FC,$1B,$C3,$0F,$23,$D4,$F5,$9C,$CD,$5F,$89,$F7,$61,$E9
	.dc.b	$9E,$31,$AC,$3C,$AF,$12,$46,$9B,$5C,$DF,$CD,$F0,$FF,$01,$7F,$98
	.dc.b	$70,$FF,$00,$7F,$91,$D7,$F0,$3E,$4F,$EB,$DC,$5E,$A9,$E7,$8F,$82
	.dc.b	$BB,$13,$1E,$CD,$8D,$47,$E5,$77,$FE,$07,$5F,$C0,$48,$7F,$0A,$9F
	.dc.b	$E0,$AD,$33,$C5,$FC,$32,$3F,$8C,$E1,$F3,$A7,$FC,$DA,$7A,$33,$F6
	.dc.b	$3C,$49,$5C,$5E,$7C,$5E,$27,$2D,$8D,$47,$F6,$3C,$BF,$95,$93,$2F
	.dc.b	$E3,$91,$FC,$B4,$91,$9C,$7D,$35,$C8,$97,$D4,$95,$FC,$1C,$FF,$1B
	.dc.b	$90,$FF,$1D,$1E,$43,$AD,$F8,$2F,$B4,$7A,$8C,$DF,$82,$BA,$CE,$A7
	.dc.b	$D5,$5C,$49,$5B,$8E,$27,$39,$E1,$7E,$88,$C9,$E3,$8F,$F2,$46,$DF
	.dc.b	$46,$7F,$C5,$5C,$3F,$BB,$8E,$93,$CB,$1F,$1E,$3E,$0C,$3B,$35,$E8
	.dc.b	$DE,$3F,$EC,$D9,$BE,$0B,$FC,$5C,$4A,$FF,$01,$ED,$FE,$31,$F8,$3F
	.dc.b	$F1,$97,$C1,$FF,$8C,$FE,$13,$FC,$69,$F0,$9F,$E3,$5F,$85,$FF,$1B
	.dc.b	$7C,$2F,$F8,$DF,$E1,$BF,$C7,$1F,$0D,$FE,$39,$F8,$7F,$F1,$D7,$C3
	.dc.b	$FF,$8F,$7F,$80,$5F,$11,$7E,$27,$1F,$11,$FA,$2E,$1F,$D3,$F6,$FF
	.dc.b	$68,$E4,$FC,$7F,$5F,$F9,$2F,$15,$F1,$BF,$F0,$E4,$74,$7F,$C3,$79
	.dc.b	$78,$5F,$F0,$EE,$9E,$57,$FC,$3F,$B7,$A5,$FF,$10,$F0,$7B,$5F,$F1
	.dc.b	$1F,$17,$C1,$7E,$19,$FE,$26,$E5,$F8,$CF,$E6,$71,$7C,$4F,$B4,$91
	.dc.b	$E4,$F6,$FE,$07,$B7,$F0,$7D,$0F,$85,$ED,$FC,$3F,$4B,$E2,$7B,$7F
	.dc.b	$17,$D4,$F8,$DE,$57,$C7,$F5,$7E,$47,$B7,$F2,$7D,$7F,$95,$ED,$FC
	.dc.b	$BF,$67,$E6,$7B,$7F,$27,$83,$F6,$FD,$5F,$88,$F3,$F0,$FC,$99,$E1
	.dc.b	$3F,$22,$FA,$3F,$2B,$1F,$03,$F3,$33,$F2,$3F,$3B,$5F,$43,$F4,$37
	.dc.b	$F6,$3F,$4B,$87,$EF,$71,$7C,$28,$FE,$3F,$C9,$F1,$9F,$F7,$1F,$92
	.dc.b	$8C,$46,$B9,$78,$6D,$F7,$EF,$E4,$CF,$DE,$E5,$F9,$2F,$E0,$39,$7E
	.dc.b	$4B,$F8,$2E,$5F,$92,$FE,$13,$97,$E4,$BF,$86,$E5,$F9,$2F,$E2,$39
	.dc.b	$7E,$4B,$FE,$33,$3F,$83,$65,$F2,$5C,$1E,$AE,$4F,$92,$FB,$39,$3E
	.dc.b	$4B,$EE,$E4,$F9,$2F,$C0,$E4,$F9,$2F,$C2,$E4,$F9,$2F,$C4,$E4,$F9
	.dc.b	$2F,$C6,$E4,$F9,$2F,$E7,$38,$BE,$57,$8F,$FE,$5D,$E4,$F8,$FA,$FE
	.dc.b	$8E,$7E,$3F,$92,$B7,$37,$CD,$E6,$F9,$31,$F4,$B9,$BE,$4B,$F5,$39
	.dc.b	$BE,$4B,$F5,$B9,$BE,$4B,$F6,$39,$BE,$4B,$F6,$B9,$BE,$4B,$F7,$39
	.dc.b	$BE,$4B,$FD,$17,$CD,$FE,$31,$A7,$FB,$8F,$A3,$F2,$2F,$97,$5F,$C0
	.dc.b	$74,$7C,$AE,$0F,$DE,$74,$7C,$A7,$F7,$DD,$1F,$29,$FE,$01,$D1,$F2
	.dc.b	$9F,$E0,$5D,$1F,$29,$FE,$09,$D1,$F2,$9F,$E0,$DD,$1F,$2B,$6F,$C6
	.dc.b	$1F,$2C,$FC,$8F,$3B,$DB,$FE,$28,$48,$FE,$17,$E6,$7F,$17,$F8,$2F
	.dc.b	$F6,$A6,$DF,$E6,$3C,$7F,$DB,$8B,$E1,$33,$F0,$7A,$F8,$4D,$FC,$07
	.dc.b	$0F,$89,$E3,$F0,$5C,$BE,$27,$9F,$C2,$74,$F8,$9E,$BF,$0D,$DB,$E2
	.dc.b	$7B,$FC,$47,$FA,$23,$DC,$FF,$04,$70,$78,$23,$5C,$7F,$EB,$BE,$16
	.dc.b	$1C,$E3,$19,$77,$75,$A7,$8D,$E1,$B7,$9D,$E5,$C1,$EE,$E1,$F7,$C7
	.dc.b	$C1,$3E,$19,$FC,$BB,$93,$E3,$3F,$C8,$7C,$7F,$C8,$BF,$90,$E6,$FC
	.dc.b	$AF,$77,$DF,$5E,$3F,$E4,$B8,$3F,$9D,$78,$FF,$92,$7F,$9E,$78,$FF
	.dc.b	$92,$7F,$9F,$78,$FF,$92,$7F,$A0,$78,$FF,$92,$7F,$A1,$78,$FF,$92
	.dc.b	$7F,$A2,$78,$FF,$92,$7F,$E0,$07,$8F,$F9,$2D,$7F,$47,$ED,$FF,$B7
	.dc.b	$63,$5B,$79,$6F,$93,$E7,$C3,$B3,$E9,$C7,$C4,$FA,$F2,$F3,$3E,$DC
	.dc.b	$FD,$4F,$BF,$4F,$73,$F0,$75,$F8,$BE,$3F,$F1,$FE,$0F,$F5,$9E,$DF
	.dc.b	$1D,$CF,$07,$C8,$8F,$95,$5C,$9F,$99,$87,$67,$E7,$63,$E2,$7E,$86
	.dc.b	$5E,$67,$E9,$67,$EA,$7E,$A6,$9E,$E7,$EB,$6B,$F0,$9F,$E0,$FE,$8F
	.dc.b	$F6,$ED,$BF,$71,$1F,$C8,$9F,$DC,$46,$63,$EF,$74,$FE,$E6,$7E,$F7
	.dc.b	$91,$FB,$97,$F0,$1E,$47,$EE,$5F,$C1,$79,$1F,$B9,$7F,$09,$E4,$7E
	.dc.b	$E5,$FC,$37,$91,$FB,$97,$F1,$1E,$47,$EE,$7A,$3F,$8F,$67,$FB,$93
	.dc.b	$F2,$05,$72,$FF,$4E,$E3,$F9,$1C,$DF,$D3,$79,$3E,$43,$FD,$3F,$93
	.dc.b	$E4,$3F,$D4,$79,$3E,$43,$FD,$4F,$93,$E4,$3F,$D5,$79,$3E,$43,$FD
	.dc.b	$5F,$93,$E4,$2E,$4F,$88,$FF,$AC,$73,$F9,$07,$CC,$E7,$DF,$32,$F9
	.dc.b	$AC,$3B,$AC,$7C,$6B,$2F,$3A,$CF,$D6,$B4,$F7,$BF,$AE,$CB,$E2,$70
	.dc.b	$7E,$C7,$9F,$F1,$FA,$FF,$6F,$E9,$FD,$CF,$0F,$F8,$AD,$FF,$1A,$79
	.dc.b	$1F,$BF,$C3,$FD,$2B,$D3,$FC,$7F,$1F,$E7,$0F,$F1,$F8,$FE,$0E,$7F
	.dc.b	$11,$F1,$FF,$8F,$BF,$C4,$FC,$7F,$E3,$EF,$F1,$5F,$1F,$F8,$FB,$FC
	.dc.b	$5F,$C7,$FE,$3E,$FF,$19,$F1,$FF,$8F,$BF,$E4,$02,$1E,$2F,$1B,$F8
	.dc.b	$FB,$E4,$F1,$BF,$8F,$3E,$6F,$1B,$F8,$F3,$E8,$F1,$BF,$8F,$3E,$AF
	.dc.b	$1B,$F8,$F3,$EC,$F1,$BF,$8F,$3F,$CA,$4F,$81,$C5,$FE,$4D,$E1,$F8
	.dc.b	$BC,$3F,$C6,$78,$BE,$4F,$13,$E2,$7F,$A1,$93,$0F,$E6,$5E,$EF,$F4
	.dc.b	$6F,$80,$FE,$8F,$EA,$FF,$48,$F8,$0F,$E9,$3F,$0D,$FD,$2B,$E0,$BF
	.dc.b	$A5,$FC,$37,$F4,$CF,$84,$FE,$F1,$97,$FB,$13,$F8,$05,$BE,$38,$6C
	.dc.b	$FF,$7C,$97,$C3,$66,$BF,$B6,$43,$E1,$B1,$5B,$F8,$8B,$E0,$29,$F1
	.dc.b	$18,$5F,$13,$F4,$AF,$FC,$93,$8F,$F0,$27,$E9,$1F,$E0,$DC,$FC,$78
	.dc.b	$73,$AE,$58,$73,$C7,$A6,$5D,$73,$ED,$97,$BF,$2F,$EB,$1A,$FC,$37
	.dc.b	$FA,$D6,$FF,$1D,$FE,$B9,$C3,$F2,$DF,$9A,$7F,$5F,$E4,$FA,$0F,$FB
	.dc.b	$27,$77,$B1,$EB,$24,$79,$F1,$F8,$BF,$15,$EF,$F3,$7E,$07,$B7,$F0
	.dc.b	$7E,$27,$E1,$7C,$4F,$C3,$F7,$BE,$27,$C4,$FC,$5F,$80,$F8,$DF,$13
	.dc.b	$F1,$FE,$0B,$E4,$7C,$4F,$C9,$F8,$4F,$95,$F1,$3F,$2F,$E1,$BE,$57
	.dc.b	$77,$F9,$27,$8B,$E5,$9F,$E3,$4F,$11,$E6,$6E,$3C,$C7,$DD,$BE,$8E
	.dc.b	$3C,$0F,$B1,$97,$86,$31,$CE,$D7,$CD,$18,$DF,$D4,$EB,$87,$DC,$F0
	.dc.b	$E3,$F8,$4F,$2E,$5F,$8E,$FE,$47,$9F,$F8,$3E,$3F,$32,$3B,$C6,$B7
	.dc.b	$FC,$0C,$AB,$9D,$7E,$66,$FB,$BF,$9B,$C3,$C6,$FE,$77,$1F,$3B,$F9
	.dc.b	$FC,$BD,$8F,$E8,$73,$F8,$2F,$E8,$F5,$F8,$CF,$E4,$8F,$97,$AF,$F2
	.dc.b	$F7,$1A,$79,$38,$3E,$9C,$7C,$6E,$1F,$AE,$FC,$8E,$2F,$B6,$FC,$AE
	.dc.b	$3F,$BE,$FC,$CE,$4F,$C1,$BF,$3B,$97,$F0,$EF,$D0,$E6,$FD,$47,$17
	.dc.b	$C3,$E7,$FD,$8F,$2F,$F0,$6E,$8F,$CA,$ED,$FE,$46,$7F,$C9,$3E,$2F
	.dc.b	$E6,$F1,$FE,$5B,$D5,$F9,$F7,$F9,$7F,$5F,$E8,$DF,$E6,$3D,$9F,$A7
	.dc.b	$7F,$82,$65,$FA,$B7,$F0,$3D,$DF,$AF,$7F,$07,$E0,$7E,$EF,$C1,$FD
	.dc.b	$FF,$F0,$0F,$FB,$8E,$5A,$FE,$E3,$E1,$7F,$C9,$78,$FF,$9C,$DB,$3F
	.dc.b	$E8,$8A,$E1,$5F,$7A,$FF,$7B,$3F,$90,$F0,$FF,$7A,$FE,$4B,$C3,$FD
	.dc.b	$EB,$FE,$53,$F1,$3F,$8E,$BF,$E5,$5F,$17,$F9,$0B,$FE,$57,$F1,$BF
	.dc.b	$92,$BF,$E5,$9F,$1F,$E2,$BF,$BD,$20,$F8,$77,$C1,$7F,$52,$E6,$F9
	.dc.b	$BC,$16,$BF,$82,$9F,$D6,$2F,$AB,$FD,$6B,$0F,$0B,$FD,$73,$1F,$2B
	.dc.b	$FD,$7B,$2F,$4B,$FC,$C7,$3F,$6B,$FC,$CF,$4F,$82,$FF,$35,$D7,$F7
	.dc.b	$B9,$7F,$25,$DB,$F7,$C7,$F9,$D6,$7F,$6B,$3C,$23,$F4,$F5,$D1,$FD
	.dc.b	$4D,$F8,$1F,$F4,$5E,$1E,$47,$FD,$1B,$8F,$A1,$FF,$47,$E5,$EC,$7F
	.dc.b	$D2,$39,$FC,$8C,$BF,$BB,$E9,$FC,$93,$6F,$8F,$1F,$D1,$F9,$FF,$2C
	.dc.b	$D3,$F8,$01,$F9,$B3,$F8,$21,$F9,$F3,$F8,$41,$FA,$33,$F8,$61,$FA
	.dc.b	$73,$F8,$8F,$87,$FB,$8C,$3F,$81,$F8,$7F,$B8,$7F,$82,$F8,$7F,$B8
	.dc.b	$7E,$25,$7E,$E3,$E1,$7F,$86,$3F,$B8,$E2,$FE,$B3,$E1,$FE,$E2,$3F
	.dc.b	$B1,$78,$7F,$B8,$7F,$B2,$78,$7F,$B8,$7F,$B3,$78,$7F,$B8,$7F,$B4
	.dc.b	$78,$7F,$B8,$7F,$B5,$78,$7F,$B8,$7F,$B6,$78,$7F,$B8,$7F,$8D,$90
	.dc.b	$FE,$95,$C1,$F1,$BC,$4F,$EC,$CE,$E3,$89,$F1,$8E,$A7,$C6,$3C,$27
	.dc.b	$C6,$3C,$A7,$C6,$3F,$88,$1F,$CE,$38,$3E,$36,$BF,$DE,$38,$3E,$33
	.dc.b	$FE,$2E,$F2,$7F,$D3,$18,$7F,$1D,$E1,$F9,$27,$F1,$DC,$3F,$A0,$DF
	.dc.b	$27,$FC,$09,$87,$D3,$C3,$FC,$07,$C5,$FC,$BA,$FF,$9B,$F8,$DF,$CC
	.dc.b	$9F,$E7,$3E,$3F,$F3,$67,$F9,$DF,$91,$FC,$E9,$FE,$7B,$E4,$7C,$3F
	.dc.b	$27,$EB,$F4,$7E,$33,$67,$95,$FE,$23,$DB,$F6,$F8,$7F,$1F,$7F,$66
	.dc.b	$FE,$59,$FE,$F6,$7F,$01,$3F,$BF,$9F,$C1,$4F,$F0,$19,$FC,$24,$FE
	.dc.b	$BB,$AF,$F1,$F7,$FC,$49,$E3,$7F,$1F,$7F,$C4,$DE,$37,$F1,$F7,$FB
	.dc.b	$37,$8B,$FC,$62,$BF,$8E,$79,$3F,$E2,$0F,$1F,$F8,$EE,$6A,$D5,$FF
	.dc.b	$8A,$FC,$6F,$DB,$61,$FE,$2A,$F1,$BF,$6C,$FF,$8B,$3C,$6F,$DB,$3F
	.dc.b	$E6,$2F,$1B,$F6,$CF,$F9,$93,$C6,$FD,$B3,$FE,$66,$F1,$BF,$6C,$FF
	.dc.b	$9A,$3C,$6F,$DB,$70,$FC,$5C,$FF,$6E,$0F,$E3,$DE,$3D,$DB,$CD,$F1
	.dc.b	$5F,$CE,$89,$3E,$E6,$BF,$CF,$30,$E1,$BF,$B2,$B8,$CE,$C8,$6C,$FE
	.dc.b	$7D,$5C,$B0,$FF,$1D,$D7,$12,$46,$6F,$AB,$6B,$89,$FE,$9E,$BF,$CB
	.dc.b	$4E,$A3,$BD,$F4,$AE,$15,$E4,$9F,$13,$67,$51,$C2,$B9,$3A,$BF,$1B
	.dc.b	$93,$C2,$66,$FA,$C6,$A3,$E5,$9F,$D6,$FE,$2B,$FB,$19,$1F,$C3,$FC
	.dc.b	$57,$F6,$6F,$6B,$FB,$3F,$C1,$FF,$68,$F8,$4F,$ED,$3E,$F7,$F6,$AF
	.dc.b	$85,$FE,$D7,$F0,$1F,$DB,$3E,$1B,$FB,$6F,$C1,$7F,$6E,$E2,$FE,$DF
	.dc.b	$F0,$7F,$DC,$3E,$1F,$FB,$8F,$C2,$FF,$72,$F8,$8F,$E8,$FC,$0E,$6F
	.dc.b	$8E,$DF,$57,$F9,$1F,$3F,$EF,$B9,$FE,$DD,$F2,$AF,$CE,$E3,$D9,$FC
	.dc.b	$FE,$7E,$27,$F4,$3A,$F9,$9F,$D1,$EF,$EA,$7F,$4B,$C3,$EE,$7F,$4F
	.dc.b	$C7,$FE,$06,$E4,$F8,$91,$F3,$F9,$7F,$D8,$FC,$C7,$3B,$7E,$19,$E1
	.dc.b	$3C,$8F,$D8,$F3,$7C,$97,$F6,$7C,$DF,$25,$FD,$AF,$37,$C9,$7F,$6F
	.dc.b	$CD,$F2,$5F,$DC,$F3,$7C,$97,$F7,$7C,$DF,$25,$FF,$31,$78,$7F,$E6
	.dc.b	$87,$F2,$1C,$DF,$23,$9B,$F9,$44,$7F,$02,$E6,$F9,$0F,$E2,$79,$BE
	.dc.b	$43,$F8,$BE,$6F,$90,$FE,$37,$9B,$E4,$3F,$8F,$E6,$F9,$0F,$E4,$79
	.dc.b	$BE,$43,$F9,$3E,$6F,$82,$FC,$73,$F8,$8F,$1F,$C8,$E8,$FC,$99,$23
	.dc.b	$F3,$1D,$1F,$9F,$D7,$F3,$BC,$DF,$A0,$9F,$CD,$1A,$E7,$FC,$D4,$64
	.dc.b	$93,$C0,$93,$FA,$28,$FF,$1E,$78,$3F,$C7,$CF,$04,$74,$24,$72,$78
	.dc.b	$4F,$2B,$F2,$46,$08,$E0,$97,$E4,$9E,$0F,$EB,$FC,$4F,$E8,$8F,$F9
	.dc.b	$3E,$BF,$DB,$75,$FB,$F6,$C6,$5F,$DE,$98,$C7,$FD,$C2,$D2,$9F,$BE
	.dc.b	$30,$6A,$7A,$3C,$88,$6C,$AE,$9C,$11,$C1,$3C,$2F,$E8,$A7,$F4,$6A
	.dc.b	$F3,$4E,$DB,$3C,$1A,$4A,$F3,$CE,$DB,$3B,$6C,$EE,$7D,$B3,$96,$C6
	.dc.b	$5F,$F7,$D7,$A1,$EA,$8C,$47,$C2,$21,$9A,$EA,$47,$9B,$5F,$91,$1C
	.dc.b	$9E,$06,$88,$6A,$3D,$3C,$9F,$AD,$8F,$E2,$DE,$7F,$F5,$07,$C2,$7B
	.dc.b	$CE,$26,$30,$F6,$3C,$CC,$9E,$03,$04,$38,$5F,$C1,$F1,$98,$EE,$73
	.dc.b	$8F,$AD,$A7,$43,$C4,$73,$38,$10,$C9,$E5,$C7,$D9,$87,$F3,$23,$CF
	.dc.b	$1D,$0E,$47,$03,$93,$A6,$E1,$F2,$DF,$0E,$1E,$87,$81,$27,$83,$67
	.dc.b	$84,$60,$8E,$0C,$E5,$F9,$4D,$BE,$85,$78,$31,$F8,$4D,$3E,$15,$F2
	.dc.b	$8C,$C7,$77,$F0,$58,$79,$67,$F7,$71,$FC,$1B,$6F,$E0,$91,$8A,$FE
	.dc.b	$25,$39,$DB,$F7,$ED,$8F,$E0,$B9,$7F,$03,$7F,$AE,$65,$E7,$78,$90
	.dc.b	$FE,$23,$5C,$9C,$61,$E0,$AE,$58,$78,$67,$67,$6C,$3F,$96,$3C,$2F
	.dc.b	$93,$A9,$E1,$7C,$EB,$CD,$18,$AD,$CE,$6B,$FA,$4E,$BA,$C3,$C2,$4C
	.dc.b	$30,$49,$F0,$1E,$33,$FC,$71,$B7,$F2,$08,$D4,$78,$E7,$93,$C3,$1F
	.dc.b	$13,$C4,$D4,$F9,$B5,$FC,$99,$F9,$4A,$E3,$3A,$39,$38,$C7,$D5,$19
	.dc.b	$23,$FD,$2F,$0F,$E6,$7F,$0F,$FE,$4D,$9F,$F1,$53,$67,$FC,$7E,$DC
	.dc.b	$FF,$C9,$3D,$16,$BF,$C2,$D8,$3D,$BC,$69,$FE,$69,$F7,$2C,$BE,$2B
	.dc.b	$3F,$C0,$3F,$CE,$B4,$67,$FE,$5B,$87,$C5,$EB,$E0,$37,$F1,$3C,$3E
	.dc.b	$0B,$8F,$C4,$F2,$F8,$4E,$7F,$13,$D3,$E1,$BA,$FC,$4F,$6F,$88,$FE
	.dc.b	$BD,$C9,$FE,$69,$DF,$FB,$0B,$77,$F0,$4F,$08,$C1,$FE,$31,$DD,$87
	.dc.b	$68,$CF,$1A,$59,$F9,$27,$30,$D7,$D3,$18,$DF,$E0,$F0,$7C,$28,$F8
	.dc.b	$BC,$3F,$1B,$8F,$E3,$F2,$7C,$8E,$5F,$93,$CD,$2F,$F1,$7E,$27,$C7
	.dc.b	$F8,$9F,$96,$48,$F0,$FB,$FF,$3B,$E1,$BE,$7F,$C4,$FD,$0F,$4B,$E8
	.dc.b	$F8,$7F,$4B,$D3,$FA,$7F,$13,$F5,$3D,$5F,$AB,$F1,$3F,$5B,$D7,$FB
	.dc.b	$7F,$05,$F5,$B5,$F8,$2B,$D9,$CF,$95,$AF,$6B,$E8,$57,$C6,$F2,$FF
	.dc.b	$7C,$FF,$9E,$BB,$1E,$06,$E3,$9D,$F6,$C3,$BF,$83,$E1,$9F,$97,$A7
	.dc.b	$95,$F9,$BB,$7A,$5F,$9F,$C1,$ED,$7E,$8F,$17,$C1,$7E,$19,$F5,$39
	.dc.b	$7E,$33,$FE,$80,$F0,$3E,$D4,$63,$C2,$F7,$46,$27,$89,$E9,$E6,$F9
	.dc.b	$6F,$DE,$E6,$F9,$6F,$E0,$39,$BE,$5B,$F8,$2E,$6F,$96,$FE,$13,$9B
	.dc.b	$DA,$FC,$AE,$0F,$92,$FC,$CE,$5F,$F5,$0F,$93,$FB,$EF,$2F,$FD,$47
	.dc.b	$D0,$9D,$D7,$FA,$92,$3E,$93,$F5,$39,$3E,$43,$F3,$F9,$3E,$43,$F4
	.dc.b	$79,$3E,$43,$F4,$F9,$3E,$43,$F5,$79,$3E,$43,$F5,$FA,$7E,$26,$1F
	.dc.b	$8F,$8F,$C8,$9F,$A5,$E5,$CB,$4F,$8F,$CB,$F2,$BA,$3F,$E0,$04,$75
	.dc.b	$3E,$41,$F6,$0F,$90,$7D,$C3,$E4,$1F,$80,$3E,$41,$ED,$3E,$41,$F8
	.dc.b	$83,$E4,$1F,$E6,$8D,$FE,$43,$FE,$6A,$D7,$E4,$3B,$3E,$47,$07,$D2
	.dc.b	$7F,$29,$C3,$F5,$9F,$CB,$71,$7D,$A7,$F3,$1C,$7F,$79,$FC,$D7,$27
	.dc.b	$E0,$9F,$C2,$1F,$03,$5F,$8C,$FE,$7F,$5F,$E5,$9E,$53,$AF,$FD,$F1
	.dc.b	$A3,$DC,$FD,$69,$2F,$F5,$BE,$E7,$EC,$3A,$3F,$63,$EF,$7E,$CB,$E1
	.dc.b	$3F,$67,$F0,$1F,$B4,$F4,$7F,$6B,$F0,$3F,$B6,$F8,$5F,$DB,$FC,$1F
	.dc.b	$EE,$3E,$1B,$F7,$3F,$0B,$FB,$AF,$87,$FD,$DF,$07,$EF,$3E,$23,$F6
	.dc.b	$FC,$1F,$EE,$EF,$23,$F6,$FD,$DF,$67,$C8,$D4,$FF,$00,$BE,$4F,$F0
	.dc.b	$2C,$7B,$3F,$C1,$33,$F1,$3F,$C1,$B5,$F3,$3F,$C2,$37,$F5,$3F,$C2
	.dc.b	$B8,$7D,$CF,$C0,$3F,$86,$F2,$7C,$37,$F6,$DD,$EE,$8F,$C7,$F8,$FF
	.dc.b	$B5,$9F,$CB,$1F,$CC,$3C,$7F,$A9,$BF,$E4,$7B,$3E,$C3,$F9,$3E,$DF
	.dc.b	$B8,$FE,$57,$BB,$F0,$0F,$E5,$FB,$FF,$08,$FF,$02,$E2,$FD,$6E,$3F
	.dc.b	$AF,$F2,$3F,$8C,$3F,$C6,$F9,$3F,$A6,$BF,$27,$C5,$F9,$77,$C7,$C8
	.dc.b	$FD,$BB,$CF,$C3,$EB,$E1,$FF,$5F,$CF,$C2,$FF,$61,$D3,$CA,$FF,$63
	.dc.b	$D7,$D2,$FF,$65,$DB,$DA,$FF,$67,$DF,$FB,$5F,$8B,$F1,$38,$FE,$36
	.dc.b	$DF,$B6,$E2,$F8,$DA,$FE,$CF,$8F,$57,$FB,$E8,$FD,$B7,$8D,$FC,$09
	.dc.b	$FD,$C7,$0F,$F0,$67,$F7,$5E,$3F,$F0,$A7,$F7,$9E,$47,$F0,$E7,$E3
	.dc.b	$6B,$FC,$E7,$3F,$D7,$47,$C1,$7F,$69,$E2,$7F,$81,$3C,$8F,$F0,$C7
	.dc.b	$B1,$F2,$AB,$E7,$74,$71,$BF,$E3,$98,$75,$3F,$C0,$98,$F8,$5F,$F0
	.dc.b	$36,$5E,$57,$FC,$11,$9F,$A5,$FF,$05,$69,$ED,$7F,$C1,$9A,$FC,$5D
	.dc.b	$BF,$67,$E4,$7E,$DB,$4F,$89,$9F,$F4,$53,$F7,$47,$F8,$07,$8B,$E7
	.dc.b	$BF,$E0,$3E,$3F,$A6,$FF,$81,$79,$3E,$BB,$FE,$07,$E5,$FB,$6F,$F8
	.dc.b	$27,$9B,$EF,$AE,$57,$57,$E3,$32,$F8,$33,$FC,$68,$FF,$2B,$72,$FF
	.dc.b	$1A,$7F,$CB,$13,$FC,$69,$FC,$91,$FC,$88,$F8,$E7,$F2,$63,$FA,$01
	.dc.b	$FC,$A8,$FE,$80,$7F,$2E,$3F,$A0,$1F,$CC,$8F,$E6,$87,$F1,$A3,$F5
	.dc.b	$67,$F1,$93,$E2,$1F,$D9,$DF,$E3,$0F,$ED,$88,$7F,$53,$F2,$3F,$76
	.dc.b	$FF,$55,$ED,$FD,$FB,$FD,$5F,$BB,$F8,$1B,$FD,$67,$BF,$F8,$3B,$FD
	.dc.b	$6F,$C0,$FE,$16,$FF,$5D,$F0,$7F,$87,$BF,$D7,$F1,$FE,$26,$FE,$C4
	.dc.b	$FF,$07,$73,$FE,$CB,$4F,$E3,$11,$FE,$16,$F0,$FF,$8F,$BF,$E1,$8F
	.dc.b	$13,$F9,$1B,$FE,$1A,$F1,$7F,$93,$BF,$E1,$CF,$1B,$F9,$5B,$FE,$1E
	.dc.b	$F1,$FF,$97,$BF,$E2,$0F,$23,$F9,$9B,$FE,$22,$F0,$BF,$C4,$BC,$5F
	.dc.b	$C5,$F3,$FE,$11,$E3,$FE,$3A,$BF,$01,$B3,$E2,$3F,$C5,$64,$8F,$F1
	.dc.b	$5F,$C4,$FF,$9A,$3F,$80,$5E,$35,$F7,$31,$F1,$39,$F7,$B5,$F1,$3B
	.dc.b	$F8,$0E,$1F,$13,$C7,$E0,$B9,$7C,$4F,$3F,$84,$E9,$F1,$3D,$7C,$9E
	.dc.b	$DF,$13,$DF,$E1,$FE,$47,$9C,$E3,$FF,$5A,$F8,$19,$7F,$CB,$FC,$6B
	.dc.b	$E5,$18,$C7,$B3,$AC,$FC,$4F,$0D,$7C,$CF,$2D,$FD,$4F,$4E,$1F,$73
	.dc.b	$DB,$8F,$FC,$9F,$C9,$F1,$63,$FC,$3D,$CB,$F2,$FC,$76,$AF,$89,$F9
	.dc.b	$04,$AF,$3F,$C4,$FC,$BF,$6F,$E6,$7B,$9F,$37,$DD,$F9,$DE,$F7,$CF
	.dc.b	$F7,$FF,$D2,$FF,$C0,$3F,$E9,$3E,$7F,$85,$AA,$3F,$D2,$79,$FF,$A0
	.dc.b	$17,$C1,$70,$F8,$9E,$3F,$09,$CB,$E2,$79,$FC,$37,$4F,$89,$FE,$87
	.dc.b	$9F,$4F,$57,$F7,$9E,$47,$E5,$B4,$3F,$E5,$0D,$17,$CA,$7C,$18,$F6
	.dc.b	$7C,$59,$FA,$34,$F4,$FC,$3F,$AA,$7D,$87,$D8,$E0,$F8,$0F,$DA,$E2
	.dc.b	$F8,$8F,$C6,$3E,$EF,$2F,$C9,$7F,$CC,$9C,$DF,$35,$C9,$0F,$3E,$FF
	.dc.b	$1A,$1C,$DD,$27,$27,$F1,$F8,$5F,$8A,$33,$C7,$F6,$B1,$F3,$C6,$37
	.dc.b	$FA,$4F,$F2,$77,$E8,$B9,$D3,$E8,$3F,$4F,$8F,$E8,$3F,$57,$8F,$E8
	.dc.b	$3F,$EB,$2F,$85,$FE,$A1,$E6,$48,$C1,$F6,$F8,$FE,$73,$E1,$E3,$F9
	.dc.b	$CF,$8F,$8F,$E7,$3F,$81,$E3,$F9,$CF,$E9,$39,$BF,$16,$FE,$9B,$9F
	.dc.b	$F1,$EF,$EA,$3A,$3F,$79,$C9,$F3,$79,$FF,$D1,$BC,$BF,$36,$BF,$96
	.dc.b	$74,$FE,$7E,$7F,$C2,$0F,$E8,$A3,$F4,$87,$EC,$BC,$9F,$E0,$6F,$ED
	.dc.b	$3C,$9F,$E0,$6F,$ED,$BC,$9F,$E0,$6F,$EE,$3C,$9F,$E0,$6F,$EE,$BC
	.dc.b	$9F,$E0,$6F,$EF,$3C,$9F,$E0,$6F,$EF,$BC,$9F,$E0,$6F,$F0,$23,$07
	.dc.b	$F0,$2F,$23,$F8,$0B,$F6,$B5,$79,$1F,$C0,$A3,$1E,$47,$F0,$27,$5E
	.dc.b	$47,$F0,$27,$87,$91,$FC,$09,$E5,$E4,$7F,$02,$7A,$79,$1F,$05,$FE
	.dc.b	$02,$7F,$B4,$38,$BF,$81,$79,$5F,$CA,$E3,$04,$3F,$D4,$9E,$4F,$F2
	.dc.b	$BA,$E9,$CB,$F4,$30,$F0,$3F,$4B,$1F,$23,$F5,$32,$F4,$3F,$5B,$3F
	.dc.b	$63,$F6,$34,$F8,$2F,$DA,$DB,$F9,$36,$7F,$E4,$13,$FA,$99,$F2,$0F
	.dc.b	$8C,$F0,$F1,$FF,$51,$5D,$1F,$D5,$5F,$81,$FD,$66,$1E,$47,$F5,$D8
	.dc.b	$FA,$1F,$D8,$65,$EC,$7F,$65,$9F,$C7,$CF,$F1,$39,$FC,$3F,$87,$FF
	.dc.b	$48,$3F,$B3,$F2,$7F,$25,$7F,$B9,$7F,$2B,$1F,$97,$3F,$83,$1F,$9B
	.dc.b	$3F,$85,$1F,$9F,$3F,$91,$1F,$C4,$0F,$E4,$9E,$2F,$ED,$1F,$DD,$78
	.dc.b	$BF,$B4,$7F,$79,$E2,$FE,$D1,$FE,$F0,$7F,$02,$F2,$BF,$91,$E1,$FC
	.dc.b	$C3,$E1,$1E,$3E,$F4,$FE,$49,$87,$37,$F9,$36,$3D,$DF,$E5,$19,$78
	.dc.b	$F2,$FE,$14,$7F,$0C,$8F,$D7,$F9,$3F,$C4,$5F,$D8,$F9,$7F,$C5,$DF
	.dc.b	$D9,$F8,$3F,$12,$BE,$31,$FE,$69,$EA,$FD,$B4,$FE,$E3,$7D,$79,$7F
	.dc.b	$D4,$AF,$96,$FF,$C8,$F1,$EC,$FF,$25,$CB,$C4,$FF,$4C,$F2,$FF,$8E
	.dc.b	$DF,$F4,$DF,$1F,$F8,$6E,$9F,$C3,$CF,$E3,$87,$F4,$03,$F8,$E1,$FC
	.dc.b	$5C,$FE,$32,$7C,$9F,$2F,$F8,$E3,$F3,$34,$FE,$95,$E5,$FF,$1C,$7F
	.dc.b	$A6,$79,$7F,$C7,$1F,$E9,$DE,$5F,$81,$FE,$A1,$87,$F1,$CB,$FD,$E7
	.dc.b	$97,$FC,$71,$FD,$F7,$97,$FC,$71,$FE,$01,$E5,$FE,$21,$FE,$05,$C5
	.dc.b	$FD,$27,$AB,$F1,$FB,$7F,$97,$BC,$4F,$E3,$9C,$3F,$EE,$07,$F2,$CF
	.dc.b	$E6,$2F,$F8,$E1,$FE,$03,$3F,$8E,$1F,$E0,$73,$F8,$E1,$FE,$0B,$3F
	.dc.b	$8E,$1F,$E0,$F3,$F8,$E1,$FE,$13,$3F,$8E,$1F,$C5,$CF,$D9,$68,$8F
	.dc.b	$EF,$BF,$0F,$FE,$DE,$7F,$AD,$78,$DF,$C8,$AB,$F8,$E7,$8F,$FC,$99
	.dc.b	$FE,$3D,$E4,$7F,$2A,$7F,$90,$79,$3F,$CB,$9F,$E4,$5E,$57,$F3,$27
	.dc.b	$F9,$27,$97,$FC,$D9,$FE,$4D,$E6,$7F,$13,$7F,$C7,$E7,$F1,$7D,$3F
	.dc.b	$6E,$B7,$F9,$3D,$2F,$88,$FF,$0A,$92,$7F,$C2,$4D,$7F,$B4,$36,$BF
	.dc.b	$C1,$4D,$DB,$F2,$0D,$9F,$ED,$8D,$E1,$FF,$0F,$36,$FC,$13,$D8,$F9
	.dc.b	$35,$F2,$3C,$EF,$E8,$B7,$FE,$32,$E0,$E1,$84,$BE,$07,$90,$B8,$F9
	.dc.b	$1B,$87,$91,$A7,$90,$B7,$E4,$6E,$3E,$47,$E3,$70,$7E,$AF,$E2,$7F
	.dc.b	$C7,$24,$8F,$F1,$B7,$89,$FE,$3C,$F8,$2F,$F1,$EF,$C1,$FF,$8F,$BE
	.dc.b	$0F,$FC,$7F,$F0,$9F,$E4,$0F,$84,$FF,$20,$F0,$FF,$90,$BE,$13,$FC
	.dc.b	$87,$F0,$BF,$E4,$4F,$85,$FF,$22,$FC,$37,$F9,$1B,$E1,$BF,$CC,$DB
	.dc.b	$7F,$86,$BF,$80,$7F,$18,$FF,$9A,$79,$6F,$22,$7F,$D7,$DC,$BF,$8E
	.dc.b	$AD,$3F,$E4,$3F,$23,$F7,$37,$FE,$04,$E8,$EA,$F2,$BE,$37,$FE,$C9
	.dc.b	$7B,$BF,$E2,$7C,$7C,$AF,$F9,$57,$3F,$4B,$FE,$57,$D7,$DA,$FF,$96
	.dc.b	$77,$F8,$2F,$F9,$6F,$87,$E2,$BF,$E5,$DE,$3F,$F3,$1F,$27,$CB,$8F
	.dc.b	$F1,$A7,$2F,$F9,$C7,$9C,$E7,$5F,$F2,$C3,$C2,$79,$1F,$E6,$7E,$6F
	.dc.b	$A0,$FF,$9A,$79,$BE,$83,$FE,$6B,$E6,$FA,$0F,$F9,$B7,$9B,$E8,$3F
	.dc.b	$E6,$FE,$6F,$A0,$FC,$83,$E1,$72,$FC,$63,$FC,$EF,$C7,$F1,$B9,$BF
	.dc.b	$29,$3B,$5E,$8F,$F4,$E8,$FF,$2A,$73,$71,$7E,$67,$C4,$DA,$FB,$18
	.dc.b	$78,$27,$FC,$A9,$8F,$91,$FF,$2B,$65,$E8,$7F,$CB,$19,$FB,$1F,$F2
	.dc.b	$D6,$9F,$01,$FF,$2E,$6B,$F9,$37,$F2,$A7,$FA,$2F,$6F,$9D,$D7,$FC
	.dc.b	$03,$AF,$F5,$53,$FA,$92,$70,$FE,$A5,$BA,$7E,$A5,$A7,$EA,$4B,$AF
	.dc.b	$EA,$5B,$C5,$FA,$9A,$E4,$4A,$E4,$FE,$3F,$E2,$7F,$7B,$E1,$36,$F8
	.dc.b	$9C,$89,$F2,$D7,$1C,$7C,$0D,$CB,$C0,$DC,$BC,$19,$76,$25,$70,$6D
	.dc.b	$70,$6C,$F0,$9F,$66,$9F,$15,$FD,$89,$F0,$4F,$D9,$D7,$BD,$B3,$FC
	.dc.b	$1D,$BA,$70,$78,$10,$F9,$06,$8D,$4F,$37,$4D,$76,$62,$7B,$36,$7B
	.dc.b	$3F,$B7,$C7,$E5,$E7,$F2,$9F,$E7,$11,$DC,$F5,$DF,$52,$5F,$48,$EE
	.dc.b	$48,$CD,$F6,$6C,$7A,$27,$84,$62,$3A,$C6,$0F,$E5,$54,$F8,$0F,$44
	.dc.b	$F1,$BE,$95,$B6,$92,$34,$43,$AB,$96,$99,$3D,$C4,$3B,$9F,$03,$D0
	.dc.b	$72,$FF,$22,$D3,$F9,$24,$FF,$50,$3E,$05,$F8,$DE,$87,$23,$A1,$AB
	.dc.b	$F1,$38,$CB,$E1,$36,$79,$3F,$9B,$C7,$93,$0E,$47,$62,$9E,$C2,$1C
	.dc.b	$EB,$BB,$C6,$BE,$D6,$3E,$C8,$FE,$25,$CD,$FC,$46,$7F,$8A,$C7,$EF
	.dc.b	$4E,$47,$77,$CE,$F3,$C3,$CE,$E8,$F0,$99,$BF,$EA,$79,$FD,$B8,$F3
	.dc.b	$9B,$21,$B3,$B4,$F4,$70,$74,$3F,$1D,$BF,$EF,$DF,$F4,$D9,$FE,$9D
	.dc.b	$3F,$B0,$4F,$E6,$9E,$27,$63,$C2,$60,$87,$CA,$7C,$D7,$FB,$F8,$FE
	.dc.b	$04,$47,$04,$AF,$E0,$4D,$BF,$E2,$39,$FF,$06,$6D,$71,$AD,$9C,$30
	.dc.b	$F5,$47,$A7,$7F,$EE,$BC,$1F,$C7,$63,$FA,$34,$EA,$7F,$8C,$5F,$4C
	.dc.b	$7C,$15,$FE,$12,$8C,$61,$D9,$A7,$F8,$CF,$83,$F8,$B4,$FF,$30,$8C
	.dc.b	$BC,$48,$64,$B7,$E0,$7F,$C4,$3A,$FE,$25,$FF,$3F,$1F,$8A,$EE,$FF
	.dc.b	$67,$CE,$E3,$91,$1C,$13,$4F,$8D,$1D,$9A,$FF,$8A,$8F,$91,$7E,$67
	.dc.b	$94,$6A,$79,$46,$8E,$45,$0F,$F8,$65,$F3,$10,$CC,$7F,$88,$5E,$A7
	.dc.b	$18,$EB,$97,$CE,$8C,$B6,$33,$97,$CF,$91,$E9,$9F,$F1,$77,$C5,$7F
	.dc.b	$56,$24,$7D,$CF,$7F,$FA,$CF,$C0,$7F,$5A,$CF,$FA,$DF,$C5,$7F,$5C
	.dc.b	$F8,$2F,$EB,$BF,$15,$FD,$7B,$97,$FA,$FF,$C5,$7F,$60,$F8,$5F,$EC
	.dc.b	$3F,$15,$FD,$8B,$E1,$FF,$B1,$FC,$57,$F6,$4F,$89,$FE,$CB,$F1,$3F
	.dc.b	$CF,$78,$3F,$4F,$C9,$F8,$EE,$EC,$47,$F9,$1E,$78,$3F,$90,$BE,$8F
	.dc.b	$E4,$B1,$F0,$3F,$94,$CF,$C8,$FE,$5B,$5F,$43,$F9,$8D,$FD,$8F,$E6
	.dc.b	$B8,$7F,$BB,$F1,$7C,$28,$FF,$2B,$71,$FF,$96,$3A,$7F,$D6,$BD,$4F
	.dc.b	$8A,$FE,$FA,$4A,$FD,$1F,$A7,$FE,$00,$F8,$AF,$F1,$27,$8A,$F8,$4B
	.dc.b	$E8,$E3,$E1,$73,$E9,$EB,$E1,$B7,$EA,$F0,$F8,$7E,$3E,$BF,$2F,$88
	.dc.b	$E7,$EC,$F4,$F8,$9E,$BE,$D7,$C6,$EF,$F9,$1C,$5F,$2B,$C8,$FF,$1E
	.dc.b	$F9,$9A,$8F,$F1,$E7,$13,$0E,$71,$8C,$BB,$BA,$D3,$C6,$F0,$DB,$CE
	.dc.b	$F2,$E0,$F5,$BD,$38,$BE,$03,$F0,$8F,$E0,$BC,$BF,$15,$FC,$87,$37
	.dc.b	$C9,$72,$43,$FD,$09,$CD,$C9,$FD,$08,$F2,$1A,$FC,$B8,$F5,$F3,$7C
	.dc.b	$B7,$DB,$CD,$F2,$DF,$7F,$37,$CB,$7E,$0F,$37,$CB,$7E,$1F,$37,$CB
	.dc.b	$7E,$2F,$37,$CB,$E5,$FE,$D7,$1F,$C2,$8F,$E7,$26,$8F,$97,$C9,$F2
	.dc.b	$9F,$3E,$9E,$6E,$4F,$97,$18,$D5,$C9,$F3,$23,$1C,$9F,$31,$D7,$27
	.dc.b	$CC,$78,$72,$7C,$3E,$6F,$F4,$E7,$1F,$CB,$7F,$D3,$DC,$BF,$95,$DF
	.dc.b	$FC,$F6,$E7,$7B,$87,$D3,$3F,$D6,$1C,$FF,$94,$C7,$B6,$1F,$EB,$0D
	.dc.b	$59,$F8,$E3,$1A,$79,$DD,$6B,$EC,$78,$6D,$F0,$5E,$5C,$1F,$8F,$CF
	.dc.b	$E3,$9F,$26,$3F,$D7,$BE,$6B,$C9,$F9,$D3,$F7,$73,$70,$7D,$38,$F9
	.dc.b	$1C,$7F,$5D,$F9,$5C,$9F,$6D,$F9,$9C,$BF,$7D,$F9,$DC,$DF,$83,$7E
	.dc.b	$87,$3F,$BD,$FE,$AF,$C7,$FE,$74,$E6,$F8,$D1,$FE,$C5,$E6,$F9,$4F
	.dc.b	$8F,$E2,$BF,$72,$47,$F5,$3E,$7F,$EF,$3D,$CF,$DE,$FC,$37,$EF,$BD
	.dc.b	$EF,$DF,$FC,$37,$F0,$0F,$80,$FE,$03,$F0,$DF,$C0,$BE,$0B,$FD,$4D
	.dc.b	$FC,$03,$FE,$72,$E9,$FF,$6F,$AF,$84,$CF,$C4,$6B,$E1,$B7,$F1,$3C
	.dc.b	$3E,$23,$8F,$C5,$7F,$99,$38,$BF,$D6,$D8,$A3,$51,$8A,$E4,$EB,$0F
	.dc.b	$0E,$3E,$28,$F2,$1F,$C1,$34,$F4,$3F,$C1,$B6,$F6,$3F,$C2,$38,$3E
	.dc.b	$03,$F0,$8F,$E1,$7C,$7F,$15,$FD,$EF,$27,$CC,$F1,$FF,$DF,$4F,$CF
	.dc.b	$8F,$A2,$7F,$07,$3F,$72,$7F,$0B,$3F,$72,$7F,$0F,$3F,$72,$7F,$13
	.dc.b	$3F,$72,$7F,$17,$3F,$72,$7E,$C8,$FD,$A1,$F9,$0E,$0F,$E4,$6F,$F8
	.dc.b	$67,$C7,$FE,$46,$EA,$3F,$2E,$F2,$3F,$6D,$E1,$FE,$76,$7F,$B3,$DF
	.dc.b	$F1,$0F,$13,$F4,$51,$8F,$05,$E2,$FE,$9E,$31,$E3,$7E,$B1,$D7,$8F
	.dc.b	$FB,$97,$FD,$DF,$DB,$F1,$7C,$8F,$E4,$5A,$7C,$83,$F9,$16,$BF,$BC
	.dc.b	$7F,$70,$AB,$FD,$EF,$DF,$F9,$B9,$FC,$E9,$F2,$4F,$D0,$9F,$24,$FD
	.dc.b	$29,$F2,$4F,$D4,$9F,$24,$FD,$69,$F2,$4F,$E2,$DE,$2F,$ED,$36,$FF
	.dc.b	$80,$3E,$37,$FA,$36,$7F,$D6,$87,$C9,$BF,$85,$E4,$7F,$01,$7E,$27
	.dc.b	$0F,$C9,$9F,$DE,$70,$FC,$87,$F7,$DC,$1F,$1D,$FE,$01,$C1,$F1,$DF
	.dc.b	$E0,$5C,$1F,$1D,$FE,$81,$BB,$83,$E4,$71,$7E,$DB,$6F,$C6,$D7,$C8
	.dc.b	$D3,$F8,$FB,$FD,$28,$FC,$79,$FC,$A8,$FC,$79,$FC,$B8,$FC,$79,$FC
	.dc.b	$C8,$FC,$79,$FC,$DF,$3F,$54,$FF,$3B,$8F,$C3,$70,$FE,$21,$F8,$E7
	.dc.b	$F9,$8B,$C9,$F8,$D3,$F2,$23,$07,$C7,$3F,$8F,$6F,$F1,$DF,$E4,$1B
	.dc.b	$F6,$E3,$FE,$8B,$1E,$2C,$3F,$98,$BE,$67,$F9,$B6,$FF,$1F,$0F,$EC
	.dc.b	$7B,$FC,$75,$BF,$C7,$BF,$88,$7F,$1D,$3F,$C8,$3E,$57,$F2,$E7,$F8
	.dc.b	$F1,$FB,$A2,$1F,$D5,$FC,$4F,$DF,$BF,$D6,$79,$FF,$81,$BF,$D6,$FA
	.dc.b	$3F,$83,$BF,$D7,$7A,$7F,$85,$BF,$D8,$BB,$BF,$8F,$71,$FE,$A9,$FE
	.dc.b	$3D,$1F,$C4,$CF,$E3,$3E,$07,$F1,$E7,$F8,$DF,$81,$FC,$79,$FF,$0D
	.dc.b	$1F,$E0,$1F,$77,$FD,$20,$F5,$F0,$7F,$8F,$BD,$FC,$1F,$E3,$EF,$87
	.dc.b	$C1,$FE,$3E,$FF,$26,$3C,$BC,$9F,$D7,$F4,$F4,$BF,$D8,$75,$F6,$BF
	.dc.b	$D8,$F6,$F8,$2F,$F6,$5D,$FF,$8F,$3F,$D9,$FC,$0F,$C9,$6E,$F8,$AF
	.dc.b	$E7,$A4,$CB,$F9,$AF,$BB,$FD,$03,$DF,$FE,$83,$F0,$9F,$D0,$BE,$17
	.dc.b	$FA,$1F,$C1,$7F,$44,$F8,$6F,$E8,$BF,$09,$FD,$1B,$E1,$FF,$A3,$FB
	.dc.b	$DF,$D2,$3E,$23,$FA,$4F,$83,$FD,$2B,$E2,$7F,$A5,$FC,$47,$F4,$CF
	.dc.b	$8A,$F9,$1C,$1F,$C7,$CF,$91,$86,$39,$3F,$A7,$CF,$09,$FE,$93,$7D
	.dc.b	$1F,$E9,$78,$F8,$1F,$E9,$B9,$79,$1F,$E9,$FA,$7A,$1F,$EA,$3B,$7B
	.dc.b	$1F,$EA,$7C,$1F,$D7,$B8,$7E,$14,$7E,$0F,$8B,$F7,$1F,$C0,$3F,$C6
	.dc.b	$E3,$FD,$27,$AD,$F0,$D3,$99,$FF,$15,$B6,$74,$BE,$2F,$FD,$37,$3A
	.dc.b	$F2,$6D,$6B,$DC,$BE,$27,$F3,$AC,$1E,$EE,$30,$FF,$26,$43,$DE,$C6
	.dc.b	$7E,$2F,$46,$CB,$DF,$BF,$13,$8F,$81,$CF,$C4,$EB,$E0,$F7,$F1,$3C
	.dc.b	$3E,$17,$8F,$C4,$FD,$1E,$0F,$F2,$5F,$17,$F8,$8B,$5C,$70,$78,$AB
	.dc.b	$83,$F1,$2B,$D1,$87,$58,$C1,$E8,$CB,$C4,$E3,$7F,$93,$AF,$A3,$6F
	.dc.b	$4E,$5E,$AE,$0F,$5F,$0F,$B3,$8B,$DB,$C7,$EE,$E4,$F7,$ED,$34,$FE
	.dc.b	$0F,$ED,$7C,$98,$C4,$7D,$47,$83,$E1,$E5,$5F,$58,$C6,$3E,$17,$59
	.dc.b	$F9,$5E,$1A,$FA,$9F,$BF,$BF,$BD,$FC,$0F,$07,$C4,$7F,$07,$C3,$F2
	.dc.b	$38,$7F,$97,$47,$F9,$37,$4F,$E1,$DE,$6F,$D0,$8F,$27,$27,$D4,$7C
	.dc.b	$DC,$BF,$61,$F4,$73,$7D,$C7,$D5,$CB,$F2,$1F,$67,$2F,$C8,$7D,$DC
	.dc.b	$BF,$21,$F8,$1C,$BF,$13,$9B,$F2,$BA,$7F,$04,$E7,$FD,$0C,$7F,$8F
	.dc.b	$9F,$D2,$57,$E9,$8F,$F1,$CF,$8B,$FC,$0D,$FF,$1D,$F8,$BF,$C0,$DF
	.dc.b	$F1,$EF,$8B,$FC,$0D,$FF,$1F,$F8,$7F,$C0,$DF,$F2,$0F,$8B,$FC,$0D
	.dc.b	$FF,$21,$F8,$BF,$C0,$DF,$E5,$3C,$8F,$3B,$F9,$4E,$9F,$E2,$42,$6B
	.dc.b	$FC,$D7,$C6,$FE,$09,$7F,$CE,$7C,$6F,$E0,$8F,$F3,$BF,$1B,$F8,$23
	.dc.b	$FC,$F7,$C6,$FE,$08,$FF,$3F,$F1,$BF,$82,$3F,$D0,$7C,$6F,$E0,$8F
	.dc.b	$F4,$3F,$1B,$E0,$BF,$C0,$CF,$F2,$CF,$27,$C7,$69,$0F,$E9,$5E,$37
	.dc.b	$F2,$97,$E3,$9F,$C4,$0F,$8E,$7F,$14,$3E,$39,$FC,$60,$F8,$E7,$F1
	.dc.b	$C3,$E3,$9F,$C8,$0F,$8E,$7F,$9A,$B3,$FC,$66,$DF,$1B,$5F,$F2,$0B
	.dc.b	$67,$FD,$A5,$E3,$FF,$9B,$DE,$07,$23,$E4,$1D,$8F,$90,$78,$8F,$90
	.dc.b	$79,$8F,$90,$7A,$8F,$90,$7B,$8F,$90,$7F,$4A,$3E,$26,$FF,$8B,$7F
	.dc.b	$6A,$7F,$1A,$AF,$C9,$1F,$B7,$7F,$2A,$7E,$D7,$C5,$FD,$A3,$FC,$B3
	.dc.b	$C5,$FD,$A3,$FC,$BB,$C5,$FD,$A3,$FC,$C3,$C5,$FD,$A6,$1F,$CC,$3C
	.dc.b	$5F,$DA,$3F,$CC,$BC,$5F,$DA,$3F,$CD,$3C,$5F,$DA,$7C,$37,$FA,$F7
	.dc.b	$0F,$E2,$8D,$7F,$BC,$1F,$C1,$E3,$8E,$9F,$C7,$A3,$F6,$A7,$F0,$BC
	.dc.b	$3C,$31,$FC,$E7,$1F,$2E,$1F,$D0,$B2,$F4,$BF,$D1,$33,$F6,$BF,$D1
	.dc.b	$B4,$FE,$2B,$8F,$F1,$73,$F8,$B6,$1F,$D4,$8F,$8E,$FC,$93,$E3,$9F
	.dc.b	$D2,$CF,$DF,$9F,$C6,$8F,$E0,$67,$F1,$A3,$F8,$39,$FC,$68,$FE,$16
	.dc.b	$7A,$4F,$59,$FC,$68,$FE,$26,$7C,$13,$FD,$73,$E3,$7F,$92,$34,$F8
	.dc.b	$D7,$FE,$C4,$8F,$E3,$53,$FD,$8B,$C6,$FE,$34,$FF,$64,$F1,$BF,$8D
	.dc.b	$3F,$D9,$BC,$6F,$E3,$4F,$F6,$8F,$1B,$F8,$D3,$FD,$AB,$C6,$FE,$34
	.dc.b	$FF,$6C,$F1,$BF,$8D,$3F,$DB,$BC,$6F,$87,$E0,$7F,$72,$DB,$F6,$93
	.dc.b	$FB,$72,$6B,$FE,$D8,$E9,$FD,$A4,$7F,$8F,$FC,$6F,$E3,$4F,$F9,$07
	.dc.b	$C6,$FE,$34,$FF,$90,$FC,$6F,$E3,$4F,$F9,$17,$C6,$FE,$34,$FF,$91
	.dc.b	$FC,$6F,$E3,$4F,$F9,$27,$C6,$FE,$34,$FF,$86,$F4,$FE,$35,$5F,$EE
	.dc.b	$18,$FF,$27,$77,$FF,$18,$D3,$FD,$D7,$E3,$7F,$18,$7F,$DD,$BE,$37
	.dc.b	$F1,$87,$FD,$DF,$E3,$7F,$18,$7F,$DE,$3E,$37,$F1,$87,$FD,$E7,$E3
	.dc.b	$7F,$18,$7F,$DE,$BE,$37,$F1,$87,$F6,$19,$7C,$3F,$4B,$FC,$D5,$AF
	.dc.b	$92,$3F,$97,$F8,$DF,$C1,$E3,$11,$E9,$8C,$13,$6F,$55,$FF,$32,$9F
	.dc.b	$F1,$B9,$CA,$33,$CB,$FC,$49,$B1,$A3,$A3,$8B,$EF,$5C,$DB,$5C,$6B
	.dc.b	$A1,$0E,$25,$9F,$F7,$3E,$1E,$CF,$21,$F0,$FF,$E2,$22,$57,$F7,$FF
	.dc.b	$89,$FF,$12,$FC,$4F,$F8,$9B,$DC,$FF,$13,$FB,$BF,$E2,$8F,$7F,$FC
	.dc.b	$57,$FC,$02,$F6,$EF,$C3,$63,$DD,$CF,$C3,$EB,$E0,$B7,$F1,$1C,$3E
	.dc.b	$0F,$8F,$C4,$F2,$F8,$3F,$E1,$BE,$97,$FA,$BB,$D7,$FE,$65,$C6,$E7
	.dc.b	$FF,$25,$6E,$BE,$33,$FE,$25,$8F,$F5,$6B,$BE,$8F,$F7,$EC,$33,$F1
	.dc.b	$4E,$F5,$F3,$3C,$77,$F5,$F0,$7B,$63,$FC,$41,$C5,$F0,$5F,$F1,$17
	.dc.b	$27,$F9,$23,$97,$E3,$C7,$F9,$6B,$9B,$FC,$C2,$63,$6F,$F2,$DB,$B8
	.dc.b	$E2,$7F,$8A,$F9,$BE,$5B,$FE,$2D,$E6,$F0,$BE,$33,$D9,$CD,$F2,$9F
	.dc.b	$77,$37,$CA,$7E,$07,$37,$C2,$7E,$00,$E9,$71,$7C,$78,$FE,$26,$7F
	.dc.b	$38,$E6,$F9,$DE,$87,$D1,$8F,$95,$D1,$F5,$5F,$99,$CB,$F2,$1E,$B8
	.dc.b	$74,$E5,$F9,$31,$8C,$5C,$BE,$A8,$FF,$3E,$ED,$EE,$7F,$CF,$FB,$FE
	.dc.b	$31,$FC,$71,$FE,$4D,$E0,$F9,$7D,$5F,$B0,$EC,$FD,$54,$FE,$67,$A3
	.dc.b	$F3,$CF,$E7,$09,$C9,$FA,$39,$DC,$62,$77,$39,$EC,$FD,$55,$6A,$31
	.dc.b	$5D,$4F,$D8,$38,$79,$18,$AE,$64,$70,$4D,$BF,$66,$FE,$E7,$0F,$56
	.dc.b	$1F,$E9,$A8,$FF,$4D,$CF,$F0,$86,$E1,$FE,$E0,$6C,$ED,$F7,$98,$31
	.dc.b	$3C,$5C,$90,$FE,$10,$6C,$F0,$3B,$72,$6D,$C1,$27,$C5,$19,$6B,$C0
	.dc.b	$91,$B9,$F3,$46,$27,$73,$98,$F3,$C6,$09,$3C,$27,$53,$E6,$9C,$BC
	.dc.b	$27,$51,$88,$FF,$48,$FA,$BE,$59,$C9,$23,$81,$0E,$8E,$5A,$64,$F8
	.dc.b	$C4,$3A,$19,$29,$C0,$86,$CF,$80,$F5,$71,$5E,$C6,$D7,$A5,$B3,$B7
	.dc.b	$A9,$83,$D1,$E9,$AF,$31,$0E,$E7,$94,$A7,$90,$D5,$F8,$5B,$7E,$16
	.dc.b	$DF,$85,$E2,$78,$4E,$46,$0D,$9D,$E7,$A3,$E0,$BF,$33,$B2,$46,$89
	.dc.b	$5E,$18,$CB,$63,$35,$E3,$6C,$F7,$9D,$12,$74,$D8,$D4,$60,$88,$FC
	.dc.b	$39,$FB,$58,$FE,$01,$8F,$EE,$6B,$F6,$E4,$8C,$CE,$EB,$F8,$39,$1F
	.dc.b	$E0,$D3,$B8,$C4,$7F,$06,$8C,$3A,$9E,$71,$88,$D4,$7F,$16,$C7,$F8
	.dc.b	$2C,$6E,$B9,$E5,$FC,$44,$98,$79,$EF,$99,$27,$71,$96,$D7,$A4,$E8
	.dc.b	$43,$C2,$FC,$57,$F9,$DC,$A3,$17,$E4,$9D,$9C,$6F,$C6,$F5,$32,$47
	.dc.b	$38,$7B,$DB,$7E,$79,$E5,$87,$9D,$A7,$39,$FF,$0A,$FB,$3F,$DE,$3E
	.dc.b	$2B,$FC,$1A,$49,$FE,$FD,$F0,$BF,$E1,$09,$FF,$08,$FB,$9F,$E1,$2F
	.dc.b	$77,$FC,$27,$F1,$5F,$E1,$4F,$7F,$FC,$2B,$F1,$5F,$E1,$6F,$81,$FF
	.dc.b	$0B,$FC,$57,$F8,$63,$E0,$FF,$C3,$3F,$15,$F0,$F7,$70,$7E,$27,$89
	.dc.b	$7F,$CE,$67,$3C,$1F,$DB,$61,$F1,$5C,$6F,$84,$3E,$1B,$15,$BF,$89
	.dc.b	$B8,$F4,$F8,$8B,$66,$CE,$2F,$38,$EB,$0F,$39,$FE,$1F,$D3,$D6,$FF
	.dc.b	$88,$75,$F7,$BF,$E2,$3D,$FE,$1B,$FE,$25,$E1,$F8,$EF,$F8,$9F,$8F
	.dc.b	$FC,$73,$C9,$F3,$63,$FB,$D7,$2F,$E0,$A7,$D7,$F1,$3E,$62,$4F,$EB
	.dc.b	$78,$BE,$1F,$0F,$C6,$F1,$3D,$7F,$13,$EC,$F7,$7D,$BE,$97,$BB,$DD
	.dc.b	$F7,$FA,$9F,$03,$DD,$F8,$3E,$B7,$C2,$F7,$7E,$1F,$B1,$F1,$3C,$3F
	.dc.b	$8B,$EC,$FC,$6F,$77,$E3,$FB,$7F,$1B,$8B,$F6,$7D,$EE,$EC,$C7,$97
	.dc.b	$C0,$FA,$77,$CA,$30,$7D,$9C,$7B,$B8,$E4,$FB,$FA,$79,$23,$18,$B6
	.dc.b	$F4,$C6,$38,$3D,$AE,$B8,$BE,$0B,$C3,$93,$E3,$BF,$E2,$8B,$73,$7C
	.dc.b	$D8,$FC,$53,$AE,$6F,$97,$3A,$9E,$47,$8B,$9B,$E6,$3E,$4E,$6F,$98
	.dc.b	$F9,$B9,$BE,$63,$E8,$E6,$F9,$8F,$AB,$9B,$E6,$3E,$CE,$6F,$98,$FB
	.dc.b	$B9,$BF,$84,$F1,$7F,$24,$CD,$1A,$9F,$55,$72,$7D,$97,$D9,$F7,$61
	.dc.b	$E2,$7E,$06,$3E,$67,$E1,$65,$EA,$7E,$26,$7F,$A2,$DB,$E1,$72,$FE
	.dc.b	$42,$7F,$9D,$78,$DF,$EB,$2D,$7E,$24,$7C,$BE,$0D,$BF,$B3,$BF,$A2
	.dc.b	$FC,$FE,$1F,$AA,$FD,$1E,$2F,$B2,$FD,$3E,$3F,$BC,$FD,$5E,$4F,$C1
	.dc.b	$BF,$5F,$97,$F1,$0F,$D9,$E6,$FC,$6B,$FC,$7F,$5F,$E9,$5A,$7C,$68
	.dc.b	$FC,$3C,$F0,$E6,$FB,$77,$D1,$FB,$B8,$78,$1F,$BF,$8F,$91,$FC,$0E
	.dc.b	$5E,$87,$F0,$79,$FB,$1F,$C2,$E9,$F0,$39,$BF,$8A,$ED,$F8,$B7,$F8
	.dc.b	$BF,$BB,$FC,$30,$9A,$FE,$6B,$DD,$FE,$23,$FC,$01,$FC,$6A,$3F,$CF
	.dc.b	$6B,$DF,$CF,$97,$AF,$80,$DF,$C3,$F0,$F8,$2E,$3E,$C7,$2F,$83,$E7
	.dc.b	$F1,$1D,$3E,$17,$AF,$A7,$DB,$E1,$BB,$FC,$57,$ED,$F7,$FD,$17,$A1
	.dc.b	$FC,$80,$C7,$37,$F0,$AE,$15,$71,$AF,$DA,$47,$F8,$39,$F3,$E3,$E0
	.dc.b	$73,$9F,$91,$DE,$BE,$87,$8E,$FE,$C7,$9F,$0F,$C2,$EE,$FE,$63,$C7
	.dc.b	$F1,$A7,$F6,$DC,$9F,$2F,$BB,$F1,$FB,$7E,$D3,$9F,$E8,$47,$FB,$52
	.dc.b	$3F,$6F,$E0,$7E,$6D,$FD,$CF,$83,$F9,$F7,$F7,$7E,$17,$E8,$DF,$DE
	.dc.b	$F8,$7F,$A7,$7F,$7F,$E2,$7E,$AD,$FE,$03,$E2,$FE,$BD,$FE,$07,$E3
	.dc.b	$7C,$6D,$FF,$B4,$6D,$FC,$2F,$C7,$FC,$9D,$7F,$08,$F1,$BF,$21,$5F
	.dc.b	$D4,$BC,$6F,$C8,$3F,$D5,$3C,$6F,$C8,$3F,$D5,$BC,$6F,$C8,$3F,$D6
	.dc.b	$3C,$6F,$C8,$3F,$D6,$BC,$6F,$C8,$3F,$D7,$3C,$6F,$83,$CB,$F1,$4B
	.dc.b	$AF,$F1,$63,$F7,$67,$F2,$33,$F8,$B1,$FC,$9C,$FE,$2C,$7F,$2C,$C3
	.dc.b	$C6,$FF,$6B,$C7,$CE,$FF,$35,$3F,$8B,$1F,$AC,$3F,$8B,$1F,$E0,$8F
	.dc.b	$1B,$F8,$B6,$FF,$1F,$1F,$EE,$3C,$1F,$2D,$FE,$E7,$C3,$F3,$DF,$EE
	.dc.b	$BC,$5F,$4D,$FE,$EF,$C7,$F5,$DF,$EF,$3C,$9F,$6D,$FE,$F7,$CB,$FC
	.dc.b	$57,$3F,$E8,$7B,$7F,$3C,$F1,$FF,$8A,$3F,$0A,$3F,$9F,$70,$7C,$87
	.dc.b	$FC,$0D,$C3,$F3,$1F,$F0,$47,$17,$D0,$7F,$C1,$5C,$5F,$0B,$87,$FA
	.dc.b	$8C,$7F,$37,$3F,$A8,$F1,$FD,$C7,$FC,$21,$C9,$F8,$0D,$7F,$07,$CB
	.dc.b	$FE,$30,$E7,$FE,$25,$5F,$C5,$CF,$92,$7F,$1B,$3A,$1F,$C7,$CF,$E8
	.dc.b	$C7,$D3,$3E,$A9,$FC,$9C,$FB,$27,$F2,$B3,$FA,$91,$F7,$E3,$F4,$8F
	.dc.b	$C4,$3F,$9B,$F9,$1F,$E2,$2E,$9F,$D6,$4F,$F5,$C7,$F5,$F3,$FB,$CB
	.dc.b	$FE,$27,$D7,$FB,$58,$FE,$27,$D9,$FB,$97,$F8,$AF,$6F,$EF,$5F,$E2
	.dc.b	$FD,$DF,$C0,$5F,$E0,$67,$F0,$6F,$13,$F8,$43,$FC,$73,$C2,$FD,$43
	.dc.b	$FE,$37,$F0,$FF,$88,$3F,$DF,$3C,$0F,$E2,$AF,$F0,$EF,$13,$F8,$CB
	.dc.b	$FC,$43,$C3,$ED,$BF,$F1,$5C,$BC,$4F,$F1,$7C,$FC,$CF,$F1,$9D,$3D
	.dc.b	$4F,$F1,$BD,$7F,$87,$61,$FC,$C4,$FF,$05,$EF,$FC,$3F,$A9,$C3,$FE
	.dc.b	$4F,$F1,$1F,$11,$FC,$D8,$97,$FC,$8B,$E2,$BF,$9C,$FB,$1F,$CE,$BE
	.dc.b	$0F,$F9,$DF,$C2,$7F,$3C,$F6,$FF,$9E,$FC,$2F,$F3,$EF,$77,$FC,$9F
	.dc.b	$FC,$02,$F8,$7B,$F0,$18,$F8,$7C,$FC,$1E,$BE,$1F,$7F,$0B,$C3,$E1
	.dc.b	$F8,$FC,$3F,$F9,$63,$E1,$BF,$D3,$79,$A3,$5C,$7F,$E2,$98,$5F,$38
	.dc.b	$C6,$3D,$DD,$67,$E4,$D3,$CD,$1F,$CF,$B6,$F5,$3F,$D0,$38,$3D,$CF
	.dc.b	$F4,$2E,$2F,$EB,$3C,$7F,$12,$3F,$AC,$F2,$7F,$61,$D1,$F1,$1F,$00
	.dc.b	$93,$E5,$F7,$3E,$1F,$BB,$F1,$3A,$7E,$2F,$C3,$FC,$6F,$7F,$E3,$FC
	.dc.b	$3F,$C8,$F8,$1F,$93,$F0,$FF,$2B,$E0,$FE,$5F,$9D,$F3,$3E,$13,$E6
	.dc.b	$FC,$3F,$F9,$F3,$F8,$07,$FD,$0D,$E8,$FF,$A2,$5F,$F4,$2F,$07,$E3
	.dc.b	$CF,$F2,$BF,$9D,$BF,$87,$E1,$1C,$8F,$95,$87,$67,$E6,$65,$E2,$7E
	.dc.b	$76,$9E,$67,$E8,$6D,$EA,$7E,$97,$07,$B9,$F8,$07,$D5,$E3,$F8,$6B
	.dc.b	$C7,$FC,$8A,$F5,$5C,$DF,$2A,$38,$10,$79,$7F,$E8,$08,$73,$F7,$9C
	.dc.b	$F0,$FF,$AA,$67,$AC,$60,$EC,$B9,$BE,$84,$7E,$27,$9B,$E8,$3F,$8B
	.dc.b	$E6,$FA,$0F,$E3,$79,$BE,$83,$F8,$FE,$6F,$92,$FF,$9E,$79,$BE,$7B
	.dc.b	$FD,$94,$F9,$E6,$79,$76,$F8,$F6,$F7,$D7,$48,$C6,$7F,$EC,$07,$E8
	.dc.b	$9E,$33,$E8,$9E,$73,$E8,$9F,$A1,$CF,$F1,$8F,$E8,$F9,$FF,$63,$B7
	.dc.b	$D0,$DF,$E4,$70,$FF,$8D,$FC,$9F,$DF,$DF,$F8,$7B,$A7,$F3,$95,$FE
	.dc.b	$23,$EA,$FD,$FC,$FF,$88,$BC,$9F,$DF,$BF,$E2,$4F,$27,$F7,$EF,$F8
	.dc.b	$9B,$C9,$FD,$63,$FA,$E3,$F6,$9E,$4F,$EF,$9F,$DB,$79,$3F,$BE,$7F
	.dc.b	$71,$E4,$FE,$F9,$FF,$4F,$FA,$3F,$C9,$9F,$87,$E4,$FE,$F9,$F8,$BE
	.dc.b	$4F,$EF,$5F,$8F,$E4,$FE,$F5,$F9,$3E,$4F,$89,$FF,$26,$E7,$E6,$7F
	.dc.b	$C9,$FA,$7A,$9F,$F2,$8E,$BE,$E7,$FC,$A7,$B7,$C2,$EE,$FE,$49,$C3
	.dc.b	$F1,$CC,$4F,$E5,$F8,$BE,$4B,$C0,$FE,$0C,$7C,$73,$F8,$51,$F1,$CF
	.dc.b	$E1,$C7,$C7,$3F,$8C,$79,$3F,$C9,$1F,$8E,$7F,$18,$3E,$39,$FC,$77
	.dc.b	$5F,$85,$C7,$FA,$1D,$FF,$92,$7A,$B2,$BF,$91,$46,$B2,$FD,$15,$FC
	.dc.b	$7B,$FD,$47,$07,$C7,$7F,$55,$C1,$F1,$DF,$F3,$17,$07,$C7,$7F,$CC
	.dc.b	$9C,$1F,$1D,$FF,$33,$70,$7C,$77,$F0,$D9,$FE,$24,$87,$EF,$7C,$2F
	.dc.b	$D9,$73,$7E,$47,$4B,$3F,$C1,$0F,$CF,$F4,$7E,$61,$FD,$0F,$4F,$ED
	.dc.b	$1F,$D1,$F8,$3F,$B4,$7F,$4B,$E0,$FE,$D1,$FD,$3F,$83,$FA,$9C,$FF
	.dc.b	$82,$77,$7E,$B9,$65,$FC,$8F,$C0,$FE,$E7,$E0,$3C,$B7,$3F,$C9,$CB
	.dc.b	$F7,$07,$ED,$7C,$2F,$DD,$E3,$CA,$3F,$8B,$47,$F2,$E7,$F8,$5E,$5E
	.dc.b	$1C,$BF,$88,$67,$E5,$BF,$E2,$DA,$7A,$5F,$E3,$1A,$FB,$5F,$E3,$5B
	.dc.b	$7C,$17,$F8,$E6,$FF,$B8,$CB,$F9,$04,$7E,$E1,$FF,$0E,$9F,$C6,$B8
	.dc.b	$3E,$63,$4F,$99,$1C,$9F,$E6,$67,$F0,$53,$F8,$39,$F3,$0F,$E1,$67
	.dc.b	$CC,$3F,$87,$9F,$30,$FE,$86,$7F,$1F,$3F,$8F,$1F,$D2,$4F,$E9,$87
	.dc.b	$F1,$BD,$BE,$5A,$31,$B7,$D0,$DB,$E5,$EB,$FD,$56,$7F,$A8,$1F,$C8
	.dc.b	$3C,$7F,$EA,$0F,$F2,$2F,$1F,$FA,$83,$FC,$93,$C7,$FE,$A0,$FF,$26
	.dc.b	$F1,$FF,$A8,$3F,$CA,$39,$7F,$A1,$69,$FD,$17,$C5,$F8,$BC,$BF,$D3
	.dc.b	$37,$FD,$D5,$FE,$5A,$BF,$D8,$1E,$07,$F7,$C8,$FD,$C9,$FC,$10,$FD
	.dc.b	$C9,$FC,$20,$FD,$C9,$FC,$30,$FD,$C9,$FC,$7C,$FE,$73,$E2,$FF,$1F
	.dc.b	$7F,$9D,$F8,$BF,$C7,$DF,$E7,$BE,$2F,$F1,$F7,$FD,$A6,$7F,$20,$3F
	.dc.b	$72,$7F,$24,$3F,$80,$1F,$C0,$8F,$E3,$E7,$F0,$63,$F8,$F9,$E2,$3C
	.dc.b	$87,$FB,$B3,$CB,$FF,$0D,$BF,$EE,$EF,$2F,$FC,$36,$FF,$BC,$3C,$BF
	.dc.b	$80,$FF,$86,$8F,$DC,$F9,$5F,$DD,$35,$72,$BE,$23,$FC,$1C,$4A,$FE
	.dc.b	$C0,$D7,$F5,$0D,$DB,$F5,$15,$CA,$31,$CD,$FE,$15,$6D,$7E,$B1,$B3
	.dc.b	$B8,$EA,$43,$A9,$F1,$8F,$C6,$9F,$8F,$AE,$FD,$1F,$92,$8F,$E9,$27
	.dc.b	$18,$F1,$E1,$EB,$8C,$61,$E3,$69,$E3,$2D,$F8,$DF,$09,$2B,$9E,$1E
	.dc.b	$8B,$E9,$39,$BE,$8F,$0C,$7E,$64,$BC,$CF,$F1,$C1,$27,$FC,$15,$E5
	.dc.b	$FF,$8E,$FD,$BF,$F1,$E7,$B7,$FE,$3D,$F7,$3F,$C7,$DE,$E7,$F8,$FF
	.dc.b	$DD,$FF,$20,$7B,$BF,$E4,$1F,$7B,$FC,$85,$EF,$7F,$90,$FD,$FF,$F2
	.dc.b	$27,$C1,$FF,$91,$7E,$07,$FC,$8D,$E2,$FF,$91,$FE,$0B,$F2,$5C,$DF
	.dc.b	$22,$3F,$91,$3F,$D3,$78,$7F,$C3,$93,$C1,$FF,$0F,$57,$47,$FC,$41
	.dc.b	$87,$81,$FF,$11,$65,$E4,$7F,$C4,$9A,$7A,$1F,$F1,$36,$DE,$C7,$FC
	.dc.b	$51,$C1,$FE,$59,$E1,$F8,$51,$FC,$C7,$8B,$F8,$53,$FD,$2C,$FE,$62
	.dc.b	$EB,$4F,$B6,$F1,$8E,$67,$F8,$CF,$8B,$E4,$3F,$E3,$5E,$2F,$90,$FF
	.dc.b	$8D,$F8,$BE,$43,$FE,$39,$E2,$F9,$0F,$F8,$EF,$8B,$E4,$3F,$E3,$DE
	.dc.b	$2F,$90,$FE,$47,$2F,$90,$FE,$47,$97,$F9,$07,$3F,$F4,$DA,$F9,$B1
	.dc.b	$FE,$73,$E4,$F9,$55,$FD,$37,$93,$E5,$3F,$D3,$F9,$3E,$53,$FD,$47
	.dc.b	$93,$E5,$3F,$D4,$F9,$3E,$53,$FD,$57,$93,$E5,$3F,$D5,$F9,$3E,$13
	.dc.b	$F2,$4F,$F2,$A7,$1F,$CA,$E6,$FC,$8F,$3F,$E5,$C9,$3F,$9E,$E7,$FC
	.dc.b	$CF,$0F,$E7,$23,$13,$A7,$F3,$87,$E9,$4B,$C3,$FA,$59,$E1,$1F,$DA
	.dc.b	$7E,$0B,$FC,$75,$1F,$C0,$BC,$C7,$4F,$FA,$5E,$7A,$4F,$02,$71,$FE
	.dc.b	$A1,$B7,$DD,$B8,$77,$9D,$12,$74,$D8,$D4,$60,$F4,$67,$F0,$9F,$C6
	.dc.b	$9F,$B6,$3F,$1F,$1F,$BF,$6E,$7F,$BC,$6E,$7C,$1E,$04,$3E,$21,$A3
	.dc.b	$53,$CD,$D3,$5D,$98,$9E,$CD,$9E,$CF,$FA,$63,$D5,$FD,$AE,$1F,$8E
	.dc.b	$EB,$FE,$3D,$97,$ED,$EB,$F8,$A9,$E1,$3D,$B8,$F7,$26,$3D,$A3,$C2
	.dc.b	$48,$CE,$3E,$06,$C7,$AA,$78,$46,$23,$AC,$60,$FE,$43,$F1,$0F,$83
	.dc.b	$F5,$4F,$1B,$E9,$5B,$69,$23,$44,$3A,$B9,$69,$93,$DC,$43,$B9,$F0
	.dc.b	$3D,$B6,$9F,$22,$3F,$D9,$D9,$FC,$8D,$3F,$D8,$D3,$FD,$66,$79,$1F
	.dc.b	$0F,$0F,$33,$DC,$EA,$6B,$3F,$84,$D8,$DB,$E6,$26,$7F,$0D,$E9,$97
	.dc.b	$AD,$E0,$D7,$A9,$0D,$9E,$12,$9D,$4F,$16,$5E,$E7,$F8,$91,$EE,$E9
	.dc.b	$F8,$18,$7F,$BB,$31,$F7,$4F,$58,$F4,$B7,$1F,$4B,$E4,$3C,$11,$A7
	.dc.b	$C3,$8F,$AE,$30,$49,$E3,$3C,$F1,$F6,$3E,$83,$A1,$E9,$3C,$1B,$FF
	.dc.b	$06,$CF,$F9,$4F,$0F,$F0,$69,$FF,$76,$4F,$A1,$E4,$75,$3C,$B7,$E0
	.dc.b	$72,$4C,$3C,$98,$7F,$4A,$C7,$D1,$5F,$C7,$B1,$FE,$0C,$DB,$FE,$0D
	.dc.b	$7F,$C2,$09,$19,$6D,$FF,$18,$AE,$2D,$AE,$35,$B1,$87,$D2,$D7,$F2
	.dc.b	$13,$FE,$8A,$8E,$18,$74,$8F,$E5,$E4,$3F,$96,$D7,$3B,$DB,$63,$FC
	.dc.b	$49,$C1,$FC,$6D,$DC,$E8,$8F,$3E,$0F,$E5,$35,$C6,$36,$7F,$87,$B7
	.dc.b	$FC,$53,$FE,$2C,$24,$FF,$0D,$38,$D7,$48,$E2,$48,$CB,$DF,$3F,$8A
	.dc.b	$F9,$BC,$8A,$EE,$2C,$F0,$8C,$1F,$2A,$BB,$46,$4F,$14,$FF,$21,$E0
	.dc.b	$FC,$EE,$1E,$89,$D4,$F4,$25,$F3,$3C,$85,$3B,$E7,$F4,$5F,$F0,$73
	.dc.b	$FE,$60,$F8,$4F,$F0,$8D,$7F,$92,$1B,$5F,$DF,$B3,$46,$2F,$FD,$8B
	.dc.b	$EE,$DC,$3F,$CD,$3F,$0B,$7D,$7F,$F6,$55,$32,$FF,$32,$C3,$DF,$D7
	.dc.b	$1E,$FE,$03,$87,$C4,$71,$F8,$2E,$5F,$11,$CF,$E1,$3A,$7C,$47,$5F
	.dc.b	$86,$ED,$DB,$DF,$E1,$FF,$B7,$6F,$FE,$8D,$3F,$D1,$E7,$F7,$53,$F8
	.dc.b	$3F,$0B,$74,$F2,$9C,$C2,$FB,$C6,$31,$F1,$BA,$CF,$CE,$F0,$D7,$D6
	.dc.b	$F2,$DF,$E1,$72,$FC,$5E,$57,$1D,$E4,$C7,$2C,$C3,$FD,$BF,$BF,$97
	.dc.b	$E2,$3E,$51,$23,$CB,$EA,$7C,$DF,$88,$F9,$DE,$87,$CF,$F8,$2F,$A1
	.dc.b	$E9,$7D,$1E,$4F,$A5,$E9,$FD,$3F,$83,$FA,$DF,$C0,$2F,$88,$BF,$0D
	.dc.b	$8F,$6B,$3F,$0F,$AF,$6F,$7F,$11,$F3,$38,$7F,$4B,$D1,$F9,$37,$F8
	.dc.b	$B9,$F3,$32,$4F,$18,$C5,$F5,$75,$8F,$8B,$2F,$24,$7D,$0D,$3D,$0F
	.dc.b	$D2,$DB,$D8,$FD,$4E,$0F,$80,$FC,$23,$EB,$F1,$FC,$57,$F3,$7C,$08
	.dc.b	$C7,$89,$F9,$07,$F2,$9C,$9C,$AB,$F2,$9B,$FF,$96,$AD,$87,$82,$72
	.dc.b	$39,$7F,$11,$97,$9A,$33,$1F,$8B,$CF,$D7,$18,$E0,$F9,$CF,$C0,$3E
	.dc.b	$37,$17,$CE,$7E,$47,$17,$CE,$7E,$59,$F3,$8F,$EB,$86,$36,$FF,$66
	.dc.b	$D7,$CE,$8E,$27,$CE,$3A,$9D,$CF,$16,$FE,$49,$FD,$8E,$1E,$97,$F6
	.dc.b	$79,$7B,$DF,$DA,$E7,$F1,$5F,$DB,$ED,$F2,$B6,$F8,$DB,$FF,$9B,$7C
	.dc.b	$8B,$9F,$D2,$21,$F1,$38,$17,$CE,$FE,$76,$3D,$DF,$A1,$97,$8D,$FA
	.dc.b	$59,$F9,$DF,$F4,$6E,$BF,$2B,$1F,$99,$C3,$F8,$B7,$E7,$71,$7E,$3D
	.dc.b	$FA,$1C,$7F,$97,$DF,$FD,$2F,$1F,$2F,$8F,$FD,$FB,$E5,$7C,$F8,$FA
	.dc.b	$DC,$7F,$4D,$FB,$1C,$9F,$A4,$7F,$39,$C9,$E3,$7F,$05,$C4,$DF,$D1
	.dc.b	$18,$E0,$F6,$3A,$E1,$F8,$0F,$0E,$2F,$94,$F2,$E4,$F9,$4F,$E3,$78
	.dc.b	$9E,$E7,$ED,$89,$1F,$AD,$F3,$BF,$73,$EE,$FE,$EB,$DE,$FD,$DF,$C1
	.dc.b	$FE,$F3,$A7,$F7,$BF,$09,$FB,$EF,$81,$FD,$FF,$C2,$FF,$00,$F8,$3F
	.dc.b	$E0,$3F,$0D,$FC,$0B,$E1,$7F,$81,$FA,$FF,$C1,$3E,$1B,$F8,$2F,$C3
	.dc.b	$FF,$3B,$FE,$01,$FF,$43,$F7,$FF,$A6,$FD,$BF,$F0,$C7,$1F,$E8,$E7
	.dc.b	$3E,$27,$D6,$BE,$53,$FC,$0F,$1E,$CF,$F0,$5C,$FC,$4F,$F0,$7D,$7C
	.dc.b	$CF,$F0,$9D,$FD,$4F,$F0,$BE,$1F,$73,$FC,$37,$8F,$F8,$CF,$27,$C4
	.dc.b	$8F,$D7,$72,$FC,$AF,$07,$F9,$44,$7E,$4B,$83,$F8,$03,$FC,$1C,$FC
	.dc.b	$A7,$8F,$FC,$A1,$FE,$7F,$E3,$FF,$28,$7F,$A0,$F7,$FF,$28,$7F,$A1
	.dc.b	$F8,$FF,$C5,$DF,$D6,$5F,$EC,$BC,$2F,$C4,$3F,$B4,$F0,$FF,$18,$B9
	.dc.b	$3F,$B4,$78,$9F,$E2,$3F,$63,$F2,$91,$F9,$CF,$1B,$F3,$15,$FB,$CF
	.dc.b	$1F,$F3,$8F,$EF,$BC,$8F,$06,$7F,$AB,$D3,$C8,$FE,$B7,$5F,$43,$FA
	.dc.b	$FD,$BD,$8F,$B8,$F8,$17,$F0,$9F,$DA,$78,$7F,$1E,$BF,$93,$47,$C7
	.dc.b	$CF,$F7,$6F,$EE,$7C,$4F,$CC,$F1,$7F,$67,$F1,$7F,$3A,$FF,$22,$9F
	.dc.b	$8E,$7F,$39,$3E,$39,$FA,$83,$E3,$D7,$EE,$B8,$BE,$3B,$FB,$CE,$2F
	.dc.b	$8E,$FC,$69,$FE,$EB,$ED,$FF,$92,$2C,$DF,$37,$FC,$9B,$3F,$D3,$F8
	.dc.b	$BE,$4D,$FF,$06,$E2,$F9,$2F,$F0,$8E,$2F,$93,$C1,$FE,$10,$E2,$F9
	.dc.b	$2F,$F8,$4B,$83,$E4,$BF,$E1,$4E,$2F,$92,$FF,$85,$B9,$7F,$A3,$E7
	.dc.b	$F1,$4F,$DC,$1F,$DD,$AB,$E4,$BF,$30,$D6,$DF,$E0,$DE,$2F,$93,$3F
	.dc.b	$D3,$27,$F1,$E7,$F5,$D3,$E4,$9F,$C9,$FC,$9F,$E1,$D4,$F2,$3F,$90
	.dc.b	$63,$F8,$3A,$FE,$7A,$7F,$20,$3F,$A0,$9F,$C8,$0F,$E8,$DA,$FF,$75
	.dc.b	$7F,$90,$70,$7C,$87,$47,$F2,$03,$FA,$41,$FC,$80,$FA,$E7,$F1,$3F
	.dc.b	$2B,$C7,$97,$F1,$F8,$F3,$DF,$F4,$87,$D6,$7F,$3F,$E5,$F7,$E9,$FC
	.dc.b	$7B,$3F,$D9,$CF,$F9,$3B,$3F,$E3,$AF,$F8,$1F,$A7,$F7,$58,$7F,$5C
	.dc.b	$EA,$FE,$3B,$3F,$E0,$BF,$27,$F8,$EB,$FE,$0D,$F2,$7F,$8E,$BF,$E0
	.dc.b	$FF,$27,$F8,$EB,$FE,$11,$F2,$7F,$8E,$BF,$E1,$3F,$27,$F8,$EB,$FE
	.dc.b	$15,$F2,$7F,$8E,$BF,$E4,$2F,$23,$F8,$CB,$FD,$F7,$C7,$FF,$59,$3F
	.dc.b	$E4,$8D,$3F,$74,$FF,$1F,$38,$9F,$DF,$6B,$AB,$FD,$FE,$FC,$2F,$F8
	.dc.b	$07,$0F,$2B,$FE,$03,$C7,$D2,$FF,$81,$72,$F6,$BF,$E0,$7C,$FE,$0B
	.dc.b	$FE,$09,$D3,$E2,$F2,$FF,$62,$DB,$F8,$F6,$CF,$89,$FE,$6C,$4B,$F6
	.dc.b	$7B,$7F,$CE,$7D,$DF,$E7,$5E,$67,$F3,$BF,$7B,$F9,$E7,$C4,$FF,$3D
	.dc.b	$F8,$0F,$E7,$DF,$13,$FC,$FF,$E0,$BF,$A0,$7C,$4F,$F4,$1F,$84,$FE
	.dc.b	$85,$F1,$3F,$D0,$FE,$1B,$FA,$27,$C4,$FF,$45,$F8,$8F,$E8,$DC,$1F
	.dc.b	$C7,$4F,$98,$64,$F9,$D1,$C5,$FA,$15,$D5,$FA,$58,$78,$5F,$A9,$97
	.dc.b	$95,$FA,$DA,$7A,$5F,$B1,$B7,$B5,$F7,$9F,$6F,$87,$E1,$3F,$E9,$CE
	.dc.b	$2F,$E1,$44,$F4,$3F,$AE,$7C,$4F,$F5,$DF,$4B,$FA,$F7,$B5,$FD,$7F
	.dc.b	$DB,$FE,$C1,$EE,$7F,$A6,$FF,$80,$7F,$D3,$3E,$1F,$E0,$BC,$64,$7F
	.dc.b	$A6,$74,$FF,$45,$AF,$88,$E1,$F0,$1C,$7E,$23,$97,$C1,$73,$F8,$8E
	.dc.b	$9F,$09,$D7,$E2,$3E,$4E,$5F,$C8,$38,$BF,$D5,$5E,$D7,$A2,$34,$76
	.dc.b	$AE,$4F,$83,$0E,$CF,$8B,$2F,$46,$7E,$9F,$85,$F5,$78,$DE,$CA,$FD
	.dc.b	$07,$17,$C0,$7F,$45,$C9,$F1,$1F,$D2,$69,$FE,$01,$3F,$1E,$57,$27
	.dc.b	$EA,$79,$7E,$53,$FA,$BE,$5F,$94,$FE,$B7,$97,$C2,$FC,$6C,$BC,$AF
	.dc.b	$9C,$F9,$3C,$5F,$19,$F9,$7C,$5F,$19,$F9,$BC,$5F,$19,$FF,$1D,$71
	.dc.b	$7C,$57,$27,$C5,$38,$9F,$14,$EA,$77,$3F,$CE,$FC,$7F,$5D,$F2,$1F
	.dc.b	$6B,$7F,$8C,$FF,$9F,$39,$9C,$1F,$1E,$31,$C1,$F1,$DD,$70,$7C,$57
	.dc.b	$F2,$1C,$FF,$92,$E7,$FC,$46,$1F,$B7,$9F,$CB,$9D,$7C,$4F,$DA,$BF
	.dc.b	$C4,$FC,$4F,$DA,$BF,$C5,$7C,$4F,$DA,$BF,$C5,$FC,$4F,$DA,$BF,$C6
	.dc.b	$7C,$4F,$DA,$BF,$C6,$FC,$4F,$DA,$BF,$C7,$7C,$4F,$CD,$7C,$37,$F8
	.dc.b	$E3,$BD,$E7,$FF,$1E,$DB,$F3,$9C,$9F,$ED,$FC,$38,$1F,$B8,$3A,$1F
	.dc.b	$B8,$3C,$07,$EE,$0F,$21,$FB,$83,$D0,$7E,$E0,$F6,$1F,$B8,$3F,$98
	.dc.b	$6D,$FE,$4D,$F1,$FE,$37,$2A,$7F,$C9,$9C,$5F,$C7,$8C,$B8,$3E,$4E
	.dc.b	$1F,$32,$FE,$4C,$7D,$7E,$1F,$92,$FD,$9E,$1F,$92,$FD,$BE,$1F,$92
	.dc.b	$FD,$DE,$1F,$92,$FD,$FE,$1F,$92,$FE,$07,$87,$F2,$19,$7F,$A2,$89
	.dc.b	$A7,$9F,$81,$C5,$FD,$29,$B5,$CA,$BF,$03,$87,$67,$F0,$78,$F8,$9F
	.dc.b	$C2,$E5,$E6,$7F,$0F,$9F,$A9,$FC,$4E,$9E,$E7,$E0,$69,$F0,$B4,$F1
	.dc.b	$69,$F8,$D8,$FC,$C6,$DF,$B8,$7F,$6E,$7E,$58,$FD,$C1,$FC,$00,$FD
	.dc.b	$C1,$FC,$10,$FD,$C1,$FA,$33,$F4,$A7,$EA,$3B,$BF,$8E,$BF,$AC,$3F
	.dc.b	$57,$E0,$7F,$14,$7F,$5B,$E0,$FF,$18,$7F,$54,$7F,$45,$29,$FB,$83
	.dc.b	$67,$EE,$0E,$67,$5C,$7F,$2B,$E0,$FF,$08,$8F,$DC,$F7,$7F,$0C,$7F
	.dc.b	$6D,$DF,$FC,$41,$FD,$C7,$81,$FC,$51,$FD,$D7,$83,$FB,$5C,$3F,$67
	.dc.b	$D7,$F1,$67,$E2,$12,$3F,$6F,$3C,$5F,$DC,$D7,$57,$F7,$77,$E1,$7F
	.dc.b	$53,$87,$95,$F3,$DF,$A7,$7F,$C3,$74,$B5,$F8,$D5,$FB,$0D,$7E,$33
	.dc.b	$FC,$1A,$FE,$33,$5F,$8C,$ED,$F8,$CF,$EE,$27,$FA,$0C,$7F,$1C,$D7
	.dc.b	$C5,$87,$83,$1F,$33,$F8,$6C,$BD,$4F,$E2,$33,$F7,$3F,$8A,$D3,$E3
	.dc.b	$3F,$8C,$D7,$F6,$58,$FF,$6D,$21,$FC,$97,$A7,$FB,$93,$FC,$9F,$C2
	.dc.b	$FE,$EC,$FF,$29,$F0,$FF,$BD,$3F,$CA,$FC,$4F,$EF,$CF,$F2,$DF,$17
	.dc.b	$FC,$04,$FF,$2F,$F1,$BF,$C0,$CF,$F3,$1F,$1F,$FC,$14,$FF,$38,$F0
	.dc.b	$7F,$8D,$78,$3F,$C6,$75,$F9,$B1,$AE,$4F,$E5,$F7,$C9,$FE,$63,$87
	.dc.b	$67,$F9,$9E,$3E,$27,$F9,$AE,$5E,$67,$F9,$BE,$7E,$A7,$F9,$CE,$9E
	.dc.b	$E7,$F9,$DE,$BF,$C5,$3C,$0F,$E0,$3D,$3F,$ED,$1E,$8F,$E5,$5C,$1E
	.dc.b	$73,$FC,$0B,$B7,$F3,$26,$90,$D1,$30,$FE,$67,$1B,$8C,$6F,$EE,$78
	.dc.b	$4E,$9B,$19,$76,$30,$FF,$0B,$D7,$E3,$27,$55,$C6,$B7,$5A,$79,$9C
	.dc.b	$6F,$D7,$3B,$25,$72,$24,$EE,$FD,$13,$A8,$C1,$D2,$FB,$3F,$03,$E2
	.dc.b	$9E,$67,$E9,$BE,$27,$FC,$60,$49,$FF,$16,$FC,$4F,$F8,$CF,$E0,$BF
	.dc.b	$C6,$9E,$AF,$F8,$D7,$DB,$FF,$1B,$7A,$FF,$E3,$7F,$73,$FC,$71,$F1
	.dc.b	$3F,$E3,$9F,$6B,$FC,$75,$EE,$FF,$8E,$FD,$EF,$E9,$DA,$38,$7F,$C7
	.dc.b	$9E,$07,$FA,$BB,$93,$F9,$E2,$F7,$F7,$3A,$F8,$3B,$5A,$F8,$1F,$E8
	.dc.b	$AF,$FA,$BE,$7F,$A2,$E1,$FE,$87,$3F,$AE,$3D,$A7,$8C,$7D,$5C,$17
	.dc.b	$8E,$64,$78,$67,$8E,$1E,$4F,$80,$F5,$47,$F9,$43,$4F,$73,$FE,$52
	.dc.b	$DB,$E1,$3F,$E5,$4E,$0F,$8C,$FF,$95,$B8,$BE,$53,$F3,$0F,$F2,$CF
	.dc.b	$2F,$CF,$5C,$FF,$51,$FC,$14,$67,$7F,$F0,$3B,$FE,$65,$E6,$E7,$3F
	.dc.b	$E2,$CB,$EE,$FF,$8B,$B0,$F1,$BF,$E3,$0C,$7C,$EF,$F8,$CB,$2F,$5B
	.dc.b	$F4,$F4,$F7,$3F,$00,$F8,$DC,$3F,$17,$6F,$8F,$19,$DF,$E3,$F0,$FD
	.dc.b	$28,$FD,$07,$AA,$F1,$3F,$CD,$66,$6B,$73,$C6,$F9,$BE,$13,$B9,$9E
	.dc.b	$2F,$2F,$0F,$F8,$9B,$4F,$4B,$FE,$28,$D7,$DA,$FF,$8A,$B6,$F8,$2F
	.dc.b	$F8,$B3,$7F,$8A,$FF,$8B,$B8,$3E,$57,$07,$E7,$F2,$FF,$10,$71,$7C
	.dc.b	$FD,$7F,$4B,$7F,$A9,$6D,$7E,$B9,$B5,$FA,$C3,$F7,$3D,$7F,$B6,$9D
	.dc.b	$36,$35,$5C,$E3,$15,$C8,$8E,$09,$D7,$FC,$BB,$BB,$F7,$78,$74,$8B
	.dc.b	$87,$48,$C3,$FB,$DA,$E2,$D9,$DD,$F6,$6C,$FF,$26,$C7,$E0,$3F,$CA
	.dc.b	$2B,$F6,$B5,$FC,$19,$B1,$97,$F8,$F1,$8A,$FD,$DB,$4A,$7C,$53,$06
	.dc.b	$A7,$A3,$C8,$86,$CA,$E9,$C1,$1C,$13,$8B,$F1,$D1,$FE,$B9,$3F,$9E
	.dc.b	$1E,$02,$1D,$2B,$D4,$F3,$6C,$F4,$69,$2B,$D7,$3B,$6C,$ED,$B3,$B9
	.dc.b	$FF,$35,$79,$6E,$6F,$05,$68,$93,$F1,$AF,$93,$6F,$93,$F1,$DA,$FA
	.dc.b	$8F,$11,$B9,$E6,$ED,$AE,$E5,$87,$C6,$8F,$D7,$46,$39,$3F,$AF,$46
	.dc.b	$EF,$BB,$E8,$3B,$10,$EC,$53,$85,$F8,$1E,$06,$89,$8F,$C3,$7D,$26
	.dc.b	$4E,$27,$33,$89,$83,$46,$08,$72,$3D,$D8,$7B,$75,$FE,$83,$8F,$CC
	.dc.b	$9E,$26,$CE,$2E,$5E,$C4,$33,$87,$D2,$BF,$44,$EE,$73,$87,$BD,$E8
	.dc.b	$48,$D3,$6B,$E5,$B9,$F1,$7F,$C5,$0F,$F8,$15,$B3,$C0,$93,$C1,$B5
	.dc.b	$DD,$B5,$F1,$CE,$D1,$FD,$B2,$FC,$F5,$FC,$0A,$7F,$8A,$CF,$F0,$02
	.dc.b	$3F,$C1,$48,$68,$B8,$7F,$17,$8F,$E2,$51,$8D,$7F,$87,$BB,$24,$ED
	.dc.b	$D1,$FC,$54,$98,$7F,$11,$7F,$C6,$1A,$FF,$10,$BF,$1C,$72,$BE,$B7
	.dc.b	$CE,$30,$72,$C3,$D9,$19,$24,$FF,$80,$49,$3B,$AE,$B3,$A8,$C1,$E6
	.dc.b	$C3,$C8,$F6,$6C,$7F,$8D,$70,$F3,$4E,$F0,$F0,$E1,$E6,$24,$74,$39
	.dc.b	$46,$9C,$57,$F4,$BC,$FF,$2D,$5E,$6C,$3B,$BC,$9B,$5C,$23,$17,$CE
	.dc.b	$72,$47,$F9,$D7,$77,$F8,$A3,$E0,$FF,$BB,$12,$FF,$B7,$7C,$1F,$F7
	.dc.b	$9F,$6F,$FB,$D7,$B9,$FD,$EF,$E1,$3F,$BE,$7B,$DF,$DF,$7E,$17,$FB
	.dc.b	$F6,$5F,$DF,$FE,$1B,$FC,$01,$F0,$3F,$E0,$1F,$87,$FF,$01,$7C,$1F
	.dc.b	$F8,$0F,$E2,$3F,$C0,$9F,$0B,$FE,$05,$E9,$FF,$03,$7C,$37,$E2,$38
	.dc.b	$7F,$87,$F8,$88,$D4,$7F,$80,$6B,$93,$FE,$03,$C3,$B3,$FE,$05,$CB
	.dc.b	$C4,$FF,$81,$F4,$F3,$3F,$E0,$9D,$BD,$4F,$F8,$2F,$83,$DC,$FC,$03
	.dc.b	$FC,$1D,$C7,$F0,$DF,$F1,$6D,$7F,$93,$BE,$27,$FC,$38,$48,$FF,$0E
	.dc.b	$7F,$00,$FF,$5C,$EA,$74,$A7,$FC,$F9,$3F,$EF,$F9,$CC,$63,$D0,$FF
	.dc.b	$25,$4B,$E0,$EF,$85,$8F,$7B,$3E,$9E,$BD,$FD,$FA,$BC,$3E,$03,$8F
	.dc.b	$AF,$CB,$E0,$79,$FB,$3D,$3E,$0B,$E6,$EF,$FE,$31,$EC,$FE,$12,$7F
	.dc.b	$0A,$EC,$F9,$D1,$EA,$9E,$2F,$B2,$FA,$BE,$EC,$7C,$2F,$C0,$CF,$CA
	.dc.b	$FC,$2D,$7D,$2F,$C4,$DF,$DB,$C3,$EE,$F6,$BD,$F3,$F2,$4F,$BF,$CB
	.dc.b	$FE,$7E,$F2,$3F,$C5,$9A,$A3,$5C,$BF,$E8,$07,$8D,$7D,$0B,$EA,$F7
	.dc.b	$3E,$9F,$27,$C9,$7E,$AF,$27,$C9,$7E,$BF,$27,$C9,$7E,$CF,$27,$C9
	.dc.b	$7E,$DF,$27,$C9,$7F,$D4,$5E,$47,$EC,$9F,$F2,$BF,$97,$FE,$26,$E7
	.dc.b	$73,$7C,$89,$FE,$CF,$0E,$6F,$93,$18,$E6,$F9,$2E,$B9,$BE,$4B,$C3
	.dc.b	$9B,$E4,$BC,$B9,$BD,$AF,$4E,$1F,$82,$FC,$83,$FC,$C9,$C5,$F2,$7A
	.dc.b	$7F,$2D,$18,$21,$E4,$3E,$91,$F9,$8E,$7F,$C8,$DF,$83,$97,$F2,$BE
	.dc.b	$4B,$66,$7E,$69,$CC,$34,$F5,$C6,$35,$F7,$BA,$DB,$E2,$3C,$37,$FF
	.dc.b	$26,$F0,$FC,$BC,$BF,$C1,$9A,$FF,$B3,$F2,$F9,$6F,$D2,$3F,$24,$7D
	.dc.b	$63,$F2,$47,$DA,$3F,$24,$7D,$E3,$F2,$47,$E0,$8F,$C2,$1F,$53,$5F
	.dc.b	$82,$FC,$33,$E2,$E1,$F8,$D7,$FC,$FB,$B7,$EE,$6F,$E1,$7B,$7F,$B6
	.dc.b	$24,$FF,$00,$F8,$3F,$DC,$F1,$FE,$EB,$DD,$FD,$DF,$C2,$7E,$F3,$AF
	.dc.b	$F7,$BF,$0B,$FB,$EF,$80,$FF,$64,$7F,$00,$FF,$02,$CB,$FC,$5A,$BE
	.dc.b	$0B,$3E,$9E,$BE,$0F,$7F,$13,$C3,$E1,$78,$FC,$4F,$EF,$31,$FF,$1A
	.dc.b	$F8,$DF,$BC,$F1,$71,$CD,$F6,$B4,$5F,$1B,$FF,$59,$43,$1E,$D1,$8C
	.dc.b	$FC,$BA,$79,$BC,$7F,$44,$FF,$1E,$DF,$D8,$FF,$20,$E1,$F8,$0F,$F2
	.dc.b	$2E,$3F,$E3,$7C,$9F,$1A,$3F,$47,$CB,$F9,$57,$FD,$D5,$E0,$FE,$F3
	.dc.b	$1F,$CB,$C7,$D3,$3F,$3C,$7E,$54,$FB,$3B,$FE,$43,$83,$F7,$91,$FA
	.dc.b	$EF,$1B,$F7,$8F,$EC,$3C,$6F,$DE,$3F,$B2,$F1,$BF,$78,$FE,$D3,$C6
	.dc.b	$FD,$E3,$FD,$3B,$CE,$FE,$A1,$8F,$E8,$3A,$BF,$91,$CF,$EE,$BC,$1F
	.dc.b	$CC,$E9,$F8,$3F,$03,$F7,$93,$F9,$33,$F7,$87,$E4,$CF,$DE,$1F,$93
	.dc.b	$3F,$78,$7E,$4C,$FD,$E1,$ED,$3D,$E7,$C1,$BF,$C9,$3F,$BD,$3F,$A8
	.dc.b	$9F,$20,$FC,$BC,$FC,$88,$FD,$9C,$7E,$71,$FD,$AF,$81,$FB,$C8,$F0
	.dc.b	$47,$EE,$3C,$5F,$DD,$57,$AF,$C5,$FD,$CB,$ED,$F1,$7F,$72,$FB,$FC
	.dc.b	$5F,$83,$87,$F5,$AC,$FF,$73,$CD,$FE,$F6,$8F,$DC,$46,$BC,$3F,$E1
	.dc.b	$15,$CA,$7F,$B0,$E1,$D9,$FE,$C7,$8F,$89,$FE,$CB,$97,$F1,$1C,$BE
	.dc.b	$77,$8D,$FC,$83,$4F,$E1,$FE,$3F,$F2,$07,$E2,$9F,$1B,$C8,$FE,$8F
	.dc.b	$D1,$F8,$BC,$7E,$31,$F2,$F0,$FB,$17,$CD,$EB,$1F,$11,$FB,$98,$F8
	.dc.b	$F5,$FE,$41,$9F,$9D,$FE,$45,$A7,$AD,$F6,$DF,$F1,$53,$F0,$C7,$F4
	.dc.b	$6D,$FE,$23,$FE,$A8,$F5,$BF,$DF,$3C,$5F,$C7,$23,$F2,$47,$F1,$C3
	.dc.b	$F2,$C7,$F1,$C3,$F3,$47,$F1,$C3,$F3,$C7,$F1,$C3,$FA,$6F,$4F,$DD
	.dc.b	$7F,$8E,$1F,$D8,$CF,$F0,$99,$F0,$ED,$C9,$FD,$AE,$3F,$8E,$9F,$E1
	.dc.b	$4E,$AF,$DB,$70,$FF,$3C,$7F,$73,$1F,$BB,$3F,$C3,$BE,$3F,$F1,$C7
	.dc.b	$FC,$3F,$5F,$C7,$1F,$F1,$0F,$8F,$FC,$1D,$FE,$1B,$E0,$7F,$1A,$7F
	.dc.b	$87,$F8,$FF,$C6,$9F,$E2,$3E,$3F,$F1,$A7,$FC,$4D,$97,$EC,$7A,$FF
	.dc.b	$67,$C9,$FC,$C7,$2F,$D9,$CF,$F1,$BF,$1B,$F8,$D5,$FF,$28,$F1,$7F
	.dc.b	$8C,$BF,$CA,$BC,$5F,$E3,$2F,$F1,$EC,$3F,$8C,$BF,$DD,$3C,$5F,$E3
	.dc.b	$2F,$F7,$6F,$17,$F8,$CB,$FD,$E3,$9B,$E0,$BF,$C6,$0F,$F1,$EF,$17
	.dc.b	$F1,$9E,$37,$8C,$F8,$7F,$E7,$64,$AF,$E5,$BF,$13,$FC,$FB,$D9,$FE
	.dc.b	$7F,$ED,$7F,$40,$F6,$FF,$A0,$FC,$47,$F4,$2F,$77,$FA,$1F,$C4,$7F
	.dc.b	$44,$F7,$FF,$A2,$FC,$47,$F4,$6F,$81,$FE,$8F,$F1,$1F,$D2,$3E,$0F
	.dc.b	$FA,$4F,$C4,$7F,$93,$77,$FF,$7A,$FF,$00,$BE,$1B,$1D,$BF,$DE,$A1
	.dc.b	$F1,$1F,$32,$9C,$C9,$CC,$7D,$2E,$BF,$EC,$11,$CE,$3F,$A3,$E3,$E1
	.dc.b	$7F,$A4,$E7,$E5,$7F,$A5,$EB,$E9,$7F,$A6,$EF,$ED,$7F,$A7,$F0,$FC
	.dc.b	$17,$FA,$8F,$1F,$C5,$7E,$39,$FD,$53,$9B,$E5,$3F,$C8,$34,$F3,$7C
	.dc.b	$4F,$B3,$5F,$2B,$67,$D7,$AA,$31,$5E,$DF,$89,$B2,$F7,$6F,$C4,$E3
	.dc.b	$E2,$73,$F0,$1A,$F8,$9D,$FC,$17,$0F,$89,$E3,$F0,$5C,$BE,$27,$9F
	.dc.b	$C2,$F4,$F8,$9E,$BF,$0F,$DB,$E2,$7F,$B8,$F2,$7F,$AD,$FE,$23,$FC
	.dc.b	$3D,$E7,$FE,$6A,$3E,$6C,$F0,$E6,$FA,$07,$AB,$0E,$AE,$39,$7E,$B6
	.dc.b	$7E,$28,$C7,$23,$6F,$3C,$63,$83,$D6,$EB,$8B,$DE,$F0,$E4,$F8,$7C
	.dc.b	$BF,$13,$9A,$4F,$F0,$4C,$FE,$7B,$DA,$32,$FE,$3A,$7E,$5D,$74,$8F
	.dc.b	$C8,$DF,$81,$FC,$9E,$1E,$47,$F2,$B8,$FA,$5F,$CB,$E7,$EE,$7F,$33
	.dc.b	$A7,$C3,$7F,$37,$AF,$CA,$7F,$CD,$E7,$C3,$E0,$F9,$7E,$87,$CF,$8F
	.dc.b	$87,$C5,$F4,$DF,$8B,$C7,$F5,$DF,$8F,$C9,$F6,$DF,$93,$CB,$F7,$DF
	.dc.b	$97,$CD,$F8,$37,$E6,$F3,$FB,$9F,$9F,$C5,$F0,$BC,$0F,$D6,$72,$7F
	.dc.b	$03,$F2,$7F,$81,$57,$E6,$48,$7F,$8F,$3C,$9F,$CF,$3F,$E3,$EE,$CF
	.dc.b	$D1,$3F,$E4,$0E,$DF,$D3,$3F,$E4,$2E,$EF,$D5,$3F,$E4,$4E,$FF,$D7
	.dc.b	$3F,$E4,$6F,$03,$F6,$4F,$F9,$23,$C1,$FC,$EF,$87,$F9,$EF,$0F,$F7
	.dc.b	$DD,$FF,$BD,$73,$5F,$BA,$3F,$C8,$3E,$27,$F0,$C7,$FC,$87,$E2,$FF
	.dc.b	$10,$7F,$C8,$BE,$37,$F1,$47,$FC,$8F,$E3,$FF,$18,$7F,$C9,$3E,$47
	.dc.b	$F1,$C7,$FC,$97,$E4,$FF,$97,$38,$BF,$73,$CB,$FE,$51,$F2,$7E,$31
	.dc.b	$8F,$23,$FC,$D9,$CD,$F2,$30,$F8,$C7,$23,$F8,$29,$D8,$FE,$12,$78
	.dc.b	$8F,$E1,$A7,$98,$FE,$22,$7A,$8F,$E2,$A7,$B8,$F8,$07,$F2,$6E,$0F
	.dc.b	$8C,$FF,$B8,$FD,$6F,$E0,$FF,$10,$73,$1B,$38,$E3,$FE,$6B,$E0,$F9
	.dc.b	$11,$FE,$78,$E0,$F9,$0F,$F9,$EB,$83,$E4,$3F,$E7,$CE,$0F,$90,$FF
	.dc.b	$9F,$B8,$3E,$43,$FE,$80,$E0,$F9,$0F,$E2,$31,$F9,$0F,$EC,$FC,$AF
	.dc.b	$E3,$8F,$EC,$C8,$7E,$CC,$FD,$E1,$F9,$83,$F8,$01,$F9,$C3,$F8,$21
	.dc.b	$FA,$03,$F8,$4F,$8F,$FB,$39,$FE,$8D,$E3,$FE,$CD,$FE,$91,$E3,$FE
	.dc.b	$CD,$FE,$95,$E3,$FE,$CF,$9F,$FD,$6F,$82,$7F,$D8,$FE,$3F,$ED,$23
	.dc.b	$F8,$AD,$71,$E5,$FE,$97,$87,$57,$FA,$6E,$3E,$17,$FA,$7E,$5F,$C1
	.dc.b	$F1,$FE,$B1,$E4,$7F,$0B,$7F,$AD,$79,$3F,$C3,$DF,$EB,$9E,$57,$F1
	.dc.b	$AC,$FE,$1C,$7E,$D1,$F9,$1E,$17,$ED,$F1,$FE,$8F,$1F,$BA,$3E,$39
	.dc.b	$FD,$A3,$C4,$FE,$00,$FF,$6A,$F1,$7F,$82,$3F,$DB,$3C,$6F,$E1,$0F
	.dc.b	$F6,$EF,$1F,$F8,$63,$FD,$C3,$C8,$FE,$20,$FF,$72,$F2,$7F,$8A,$3F
	.dc.b	$E8,$D3,$F8,$C7,$9B,$FE,$44,$D3,$FC,$3D,$1F,$18,$FE,$32,$7F,$3C
	.dc.b	$E2,$F8,$CF,$F5,$CE,$2F,$8C,$FF,$5E,$E2,$F8,$DB,$7F,$76,$E2,$F8
	.dc.b	$CF,$F7,$8E,$2F,$8C,$FF,$7A,$E2,$F8,$D7,$FE,$06,$E2,$FE,$77,$EF
	.dc.b	$B8,$3F,$67,$E1,$7E,$D6,$BF,$C0,$BE,$17,$ED,$77,$FF,$65,$78,$7F
	.dc.b	$C6,$9F,$F6,$67,$95,$FC,$69,$FF,$67,$79,$5F,$C6,$9F,$F6,$87,$95
	.dc.b	$FC,$69,$FF,$69,$79,$5F,$C6,$9F,$F6,$A7,$95,$FC,$69,$FF,$6B,$79
	.dc.b	$5F,$C5,$B3,$FF,$0D,$F9,$5F,$C6,$BC,$1F,$F0,$AC,$7E,$CC,$FE,$40
	.dc.b	$7E,$F4,$FE,$48,$7F,$01,$3F,$94,$1F,$C1,$4F,$E5,$87,$F0,$93,$F9
	.dc.b	$81,$FC,$34,$FE,$68,$7F,$11,$3F,$89,$9F,$EF,$4D,$BF,$C2,$7C,$1F
	.dc.b	$1E,$3F,$C1,$FF,$13,$FE,$0C,$23,$FD,$D1,$AF,$F8,$D1,$B8,$FF,$8C
	.dc.b	$DB,$C9,$FE,$2E,$6C,$F4,$39,$10,$E4,$59,$E4,$FF,$80,$B1,$F8,$C7
	.dc.b	$2F,$0B,$FC,$79,$57,$23,$49,$5C,$F0,$F1,$34,$F1,$1E,$02,$61,$E0
	.dc.b	$71,$8F,$C3,$C3,$FC,$95,$E5,$7F,$8B,$89,$3F,$E1,$6F,$37,$FC,$63
	.dc.b	$F0,$1F,$E3,$47,$FC,$B1,$1F,$C2,$71,$3F,$E5,$48,$7C,$06,$FE,$13
	.dc.b	$87,$C1,$71,$F3,$F9,$7C,$17,$3F,$88,$E9,$F0,$7D,$7E,$23,$B7,$C2
	.dc.b	$77,$F8,$8F,$07,$C3,$7F,$3C,$E4,$FE,$2F,$E4,$FE,$CF,$87,$FC,$BF
	.dc.b	$D1,$AE,$2F,$F1,$87,$13,$1E,$71,$8C,$FB,$BA,$D7,$C6,$F0,$DF,$CE
	.dc.b	$F2,$E1,$F8,$1C,$5F,$06,$3E,$19,$F1,$4F,$F1,$DF,$27,$C8,$7F,$CC
	.dc.b	$1D,$08,$C6,$FF,$E7,$B7,$FC,$B7,$A7,$2A,$F9,$47,$C9,$3B,$9E,$13
	.dc.b	$D7,$C9,$F2,$5F,$6F,$27,$C9,$7D,$FC,$9F,$25,$F8,$3C,$9F,$25,$F8
	.dc.b	$7C,$9F,$27,$8F,$FC,$B9,$C9,$67,$EC,$4E,$DF,$1F,$33,$36,$1D,$67
	.dc.b	$30,$C7,$C5,$18,$CB,$CC,$EB,$3F,$C1,$3F,$53,$7F,$94,$FF,$98,$B9
	.dc.b	$BE,$1B,$F2,$88,$7C,$AE,$8F,$CF,$BF,$9A,$DB,$F3,$ED,$3F,$3F,$C1
	.dc.b	$F9,$D2,$46,$9B,$1A,$8C,$11,$C1,$35,$FD,$8B,$F8,$AF,$88,$FF,$19
	.dc.b	$EA,$7C,$77,$D2,$B7,$5E,$36,$C6,$78,$BF,$73,$3F,$B5,$9C,$E1,$E1
	.dc.b	$6C,$78,$67,$6F,$36,$F4,$FE,$DD,$CC,$7F,$B2,$67,$C2,$D8,$CB,$F1
	.dc.b	$8C,$65,$FB,$87,$04,$3F,$83,$9A,$35,$3D,$1D,$35,$D9,$89,$EE,$E0
	.dc.b	$91,$F8,$9C,$7E,$49,$FC,$B2,$3B,$1F,$C2,$A3,$A1,$27,$9B,$6B,$D1
	.dc.b	$3D,$27,$24,$8C,$CE,$E7,$D3,$39,$6C,$FF,$BB,$FD,$E7,$83,$E9,$8E
	.dc.b	$35,$BA,$F2,$35,$F3,$90,$E0,$74,$21,$C2,$39,$91,$C9,$C4,$FF,$6F
	.dc.b	$F5,$FF,$19,$7F,$D4,$A7,$F6,$83,$FA,$41,$BC,$3D,$8F,$53,$99,$E2
	.dc.b	$30,$43,$85,$7A,$5D,$1F,$43,$0F,$82,$FC,$BB,$F3,$3C,$4D,$90,$D1
	.dc.b	$E0,$29,$B2,$9E,$0B,$F6,$5F,$F9,$3B,$46,$9F,$CD,$AB,$DA,$75,$9F
	.dc.b	$6B,$E8,$BF,$1B,$4F,$04,$69,$E1,$8F,$E9,$B1,$F5,$CE,$67,$C6,$6A
	.dc.b	$B9,$B4,$E7,$5F,$15,$D9,$F9,$8D,$7F,$7C,$FF,$A3,$48,$7F,$20,$D3
	.dc.b	$F3,$4F,$33,$E1,$5F,$84,$95,$D5,$B7,$E2,$9F,$EE,$B8,$7A,$2B,$F7
	.dc.b	$F9,$7F,$07,$AF,$E1,$B9,$7F,$05,$9C,$B6,$33,$7F,$C1,$DB,$87,$F0
	.dc.b	$A2,$65,$EA,$9F,$E7,$7C,$DF,$EA,$57,$D7,$8F,$D4,$7F,$BC,$67,$CB
	.dc.b	$2E,$F1,$D4,$FE,$30,$E3,$0E,$CD,$26,$3C,$4F,$E6,$FB,$7F,$17,$71
	.dc.b	$5C,$9A,$FF,$1D,$6E,$1F,$06,$3B,$EB,$F8,$AF,$49,$CF,$FD,$5F,$2F
	.dc.b	$0E,$1C,$36,$F8,$98,$75,$21,$D4,$B8,$FF,$88,$72,$F2,$46,$27,$8B
	.dc.b	$B2,$4F,$8D,$B1,$A8,$C1,$F2,$6F,$FD,$ED,$3F,$E8,$62,$1F,$BB,$38
	.dc.b	$9F,$2C,$AF,$A9,$F1,$19,$8E,$B3,$E4,$AF,$F1,$B6,$CF,$88,$FF,$19
	.dc.b	$12,$FF,$C3,$FE,$EF,$F8,$D7,$E2,$7F,$C6,$DA,$7F,$8D,$FE,$1F,$FC
	.dc.b	$71,$F0,$1F,$E3,$9F,$80,$FF,$1D,$7C,$0F,$F8,$EF,$E1,$FF,$C7,$39
	.dc.b	$7F,$80,$3F,$80,$7F,$08,$DF,$88,$CF,$0F,$BE,$5D,$38,$E2,$FF,$20
	.dc.b	$F1,$A5,$F1,$19,$C5,$39,$F8,$6C,$3B,$F8,$8F,$F2,$67,$5F,$F4,$6A
	.dc.b	$FF,$25,$E7,$DE,$77,$8B,$0B,$8E,$33,$CE,$9A,$D3,$D9,$97,$F9,$0B
	.dc.b	$6F,$80,$FF,$91,$38,$3E,$23,$FE,$46,$E2,$F9,$0F,$CA,$3F,$C9,$3C
	.dc.b	$BF,$35,$73,$FF,$99,$F5,$7C,$47,$90,$93,$E8,$CB,$E1,$CF,$FA,$C5
	.dc.b	$BE,$8F,$AB,$E0,$7D,$7E,$07,$B3,$E0,$BD,$BE,$9F,$BB,$E0,$FD,$FE
	.dc.b	$A7,$C0,$F8,$4F,$83,$EA,$FC,$2F,$85,$F8,$7E,$B7,$C4,$F8,$6F,$8B
	.dc.b	$EB,$FC,$AE,$1F,$DE,$F0,$7D,$4F,$2B,$F8,$04,$7F,$19,$3C,$C7,$D4
	.dc.b	$E2,$57,$38,$CF,$81,$F6,$F1,$F0,$46,$29,$9F,$96,$31,$AF,$A5,$D6
	.dc.b	$FE,$D7,$87,$0F,$C1,$79,$71,$FD,$CF,$5F,$E4,$C7,$E4,$79,$7F,$27
	.dc.b	$CD,$F3,$DC,$F8,$1F,$91,$78,$4F,$23,$C9,$CF,$F3,$1F,$37,$3F,$CC
	.dc.b	$7D,$1C,$FF,$31,$F5,$73,$FC,$C7,$D9,$CF,$F3,$1F,$77,$3F,$CC,$7F
	.dc.b	$C8,$47,$EF,$32,$FF,$48,$B8,$3F,$51,$CB,$F2,$5F,$D5,$72,$7C,$87
	.dc.b	$F5,$9C,$7F,$1D,$FD,$77,$1F,$C7,$7F,$61,$C7,$F1,$DF,$D9,$71,$FC
	.dc.b	$77,$F6,$9C,$7F,$01,$F8,$C7,$D3,$E1,$F8,$FE,$CF,$E4,$1F,$8F,$D1
	.dc.b	$67,$EC,$72,$FE,$AA,$FA,$6D,$FB,$BC,$7C,$0F,$EF,$72,$F2,$3F,$BF
	.dc.b	$CF,$D0,$FF,$01,$D3,$D8,$FF,$03,$D7,$E0,$3F,$C1,$76,$FC,$96,$7F
	.dc.b	$E8,$AF,$2B,$F9,$CB,$F2,$75,$FE,$0F,$E4,$FF,$55,$BE,$13,$F4,$0F
	.dc.b	$C7,$61,$DA,$BF,$57,$8F,$89,$FD,$6E,$5E,$67,$F5,$F9,$FA,$9F,$D8
	.dc.b	$E9,$EE,$7F,$63,$87,$C2,$3E,$41,$0F,$DE,$F0,$7F,$1A,$FE,$01,$7C
	.dc.b	$47,$F1,$66,$CF,$E9,$1A,$FF,$A9,$97,$9B,$BF,$73,$87,$C4,$F1,$F7
	.dc.b	$B9,$7C,$47,$3F,$80,$E9,$F1,$1D,$7E,$0B,$B7,$C4,$77,$F8,$4F,$06
	.dc.b	$FE,$1F,$85,$F1,$7C,$47,$8F,$E1,$FC,$9F,$11,$FC,$AB,$CA,$FD,$EF
	.dc.b	$5F,$F5,$36,$F3,$7A,$E7,$87,$2F,$A7,$0E,$8F,$AF,$2F,$03,$ED,$D3
	.dc.b	$C8,$FB,$F6,$F4,$3F,$07,$83,$D8,$FC,$3E,$2F,$80,$FC,$5E,$4F,$91
	.dc.b	$CD,$20,$7C,$E3,$F8,$87,$83,$F9,$66,$9F,$54,$FC,$F9,$FB,$C3,$F8
	.dc.b	$49,$FB,$C3,$F8,$69,$FB,$C3,$F8,$89,$FB,$C3,$F8,$B7,$5F,$E2,$5F
	.dc.b	$E3,$DD,$9F,$96,$DF,$F9,$AF,$7F,$E5,$B8,$BE,$47,$C2,$7C,$B3,$13
	.dc.b	$AB,$FC,$A1,$FB,$C3,$F8,$39,$D8,$F0,$1F,$BC,$3C,$87,$EF,$0F,$41
	.dc.b	$FB,$C3,$D8,$7B,$8F,$EE,$1E,$17,$C3,$F2,$FF,$B5,$67,$FB,$C9,$F8
	.dc.b	$A7,$F0,$02,$61,$FD,$CE,$BF,$79,$1D,$23,$F7,$8F,$81,$FD,$E3,$E4
	.dc.b	$7F,$78,$FA,$1F,$DE,$3E,$C3,$DC,$7F,$80,$FC,$4F,$E4,$4F,$F4,$93
	.dc.b	$F7,$9E,$6F,$F9,$FB,$C3,$FD,$D6,$5F,$1E,$78,$1F,$BA,$3A,$5F,$67
	.dc.b	$FC,$1D,$E3,$FF,$33,$7F,$C2,$1E,$47,$F3,$77,$FC,$25,$E4,$FF,$3B
	.dc.b	$7F,$C2,$9E,$57,$B9,$F8,$E7,$EE,$6F,$FA,$23,$F8,$EF,$2F,$FA,$43
	.dc.b	$9C,$7F,$91,$3C,$0F,$A1,$3F,$E1,$BA,$EC,$FF,$87,$6F,$C4,$FF,$87
	.dc.b	$F0,$F3,$3F,$E2,$1C,$7F,$A4,$5F,$F8,$9F,$CB,$FE,$90,$FC,$03,$FA
	.dc.b	$19,$FD,$16,$7F,$A3,$BF,$20,$C6,$3F,$1E,$3F,$C5,$9A,$F2,$7F,$C5
	.dc.b	$D5,$D9,$F0,$57,$E7,$8F,$8E,$7E,$88,$F8,$E7,$E9,$8F,$8E,$7F,$65
	.dc.b	$3E,$03,$F1,$9F,$C5,$66,$EB,$FE,$41,$CD,$FB,$79,$FE,$38,$FF,$21
	.dc.b	$E7,$FD,$EC,$7F,$1F,$E8,$FE,$02,$FF,$21,$E9,$FE,$0A,$FF,$23,$EA
	.dc.b	$FE,$12,$FF,$25,$EB,$FE,$1A,$FF,$27,$C7,$F8,$8B,$FC,$A7,$B3,$F8
	.dc.b	$AB,$FC,$AF,$B7,$F6,$4F,$F8,$68,$FF,$24,$F8,$C8,$DE,$DF,$E7,$18
	.dc.b	$57,$48,$C5,$F8,$1D,$61,$E4,$78,$63,$E8,$79,$65,$EC,$7A,$67,$F0
	.dc.b	$1E,$DA,$7C,$57,$C1,$AF,$ED,$7A,$9A,$3E,$27,$FA,$41,$37,$FE,$91
	.dc.b	$EE,$7F,$4B,$F0,$7F,$A6,$75,$7F,$4D,$F7,$BF,$CC,$7F,$C0,$3F,$E7
	.dc.b	$2B,$69,$FE,$8D,$87,$C0,$6B,$E1,$B7,$F0,$5C,$3C,$3E,$3F,$07,$CB
	.dc.b	$E1,$F9,$FC,$2F,$4F,$88,$EB,$F0,$FF,$D2,$35,$F9,$9C,$8E,$1F,$E8
	.dc.b	$F1,$F3,$BB,$76,$FF,$90,$B7,$61,$D2,$31,$97,$81,$D6,$9E,$47,$86
	.dc.b	$DE,$CD,$FD,$B1,$EF,$3F,$B5,$71,$7C,$37,$E2,$9F,$DB,$79,$3E,$43
	.dc.b	$FD,$1F,$56,$6F,$89,$F8,$E4,$9F,$37,$C4,$FC,$AF,$83,$F9,$7E,$E7
	.dc.b	$CC,$F8,$8F,$9B,$F1,$1F,$3B,$CC,$F9,$FF,$11,$F4,$3E,$03,$E8,$FC
	.dc.b	$47,$D2,$F8,$2F,$A7,$F1,$1F,$E9,$6F,$33,$FD,$47,$DF,$FE,$BF,$5F
	.dc.b	$09,$9F,$85,$F9,$9C,$3F,$BD,$25,$7E,$57,$B3,$51,$FE,$6C,$BE,$2D
	.dc.b	$F3,$FB,$BF,$3F,$2F,$1B,$F4,$74,$F3,$BF,$4F,$6F,$5B,$F5,$78,$3D
	.dc.b	$EF,$D7,$E2,$FC,$5F,$1F,$C5,$8F,$C5,$E3,$FE,$B6,$E5,$F9,$51,$FC
	.dc.b	$9F,$CC,$C7,$36,$F7,$FC,$83,$CA,$7A,$1F,$83,$E6,$F9,$8F,$E1,$79
	.dc.b	$BE,$63,$F8,$7E,$6F,$98,$FE,$27,$9B,$E6,$3F,$8B,$E6,$F9,$9A,$7C
	.dc.b	$DE,$6F,$8B,$3F,$C0,$4F,$F2,$31,$FE,$C8,$F8,$3F,$F2,$4B,$93,$D6
	.dc.b	$7F,$B3,$39,$7F,$DA,$6F,$CD,$9E,$AF,$73,$E6,$1E,$33,$E6,$1E,$73
	.dc.b	$E6,$1E,$B3,$E6,$1E,$F3,$E0,$9E,$0E,$8F,$C9,$F0,$FC,$7C,$3E,$4C
	.dc.b	$FC,$B3,$F8,$BF,$4F,$E6,$B8,$3F,$8B,$3F,$E4,$BF,$1B,$F4,$13,$FE
	.dc.b	$49,$EB,$FD,$23,$D7,$93,$F8,$15,$FF,$01,$EC,$FE,$01,$3F,$B3,$F1
	.dc.b	$7F,$80,$3F,$B5,$F1,$7F,$80,$3F,$B7,$F1,$7F,$80,$3F,$B9,$F1,$7F
	.dc.b	$95,$BE,$9F,$17,$F7,$FC,$1F,$C0,$7C,$3C,$E7,$FC,$3E,$F8,$DF,$8F
	.dc.b	$0E,$AF,$97,$1F,$0B,$E7,$CB,$CA,$FA,$73,$FD,$FB,$FE,$67,$F1,$BF
	.dc.b	$8F,$BF,$E6,$9D,$3E,$13,$FB,$F2,$1F,$BF,$F1,$FF,$95,$38,$CF,$E4
	.dc.b	$BF,$C1,$8F,$F3,$84,$F4,$7F,$CE,$55,$E0,$7C,$47,$F1,$23,$E4,$1F
	.dc.b	$C5,$8F,$90,$7F,$1A,$3E,$41,$FC,$7F,$C7,$FE,$42,$FE,$3E,$BF,$C8
	.dc.b	$AF,$F2,$8B,$FE,$A4,$FC,$D3,$81,$C8,$FB,$5E,$3F,$F2,$57,$EE,$6D
	.dc.b	$FC,$95,$FB,$DE,$3F,$F2,$57,$F9,$71,$F8,$2B,$FE,$B9,$C5,$F8,$67
	.dc.b	$FA,$F7,$1F,$E2,$F9,$7F,$D2,$DC,$BF,$D4,$79,$7F,$77,$87,$E2,$7B
	.dc.b	$FF,$28,$B9,$BF,$2E,$FE,$6B,$1F,$E9,$9D,$FF,$BA,$8F,$C8,$77,$FE
	.dc.b	$E9,$FC,$97,$7F,$EE,$9F,$CA,$77,$FE,$E9,$FC,$B7,$7F,$EB,$9F,$D2
	.dc.b	$77,$7F,$1C,$6F,$8F,$FC,$93,$2F,$F1,$09,$37,$FF,$4A,$4F,$F2,$39
	.dc.b	$FE,$04,$FF,$23,$3F,$70,$7E,$8F,$C6,$FE,$46,$FE,$97,$C6,$FE,$46
	.dc.b	$FE,$9F,$C6,$FE,$46,$FE,$A7,$C6,$FE,$46,$FE,$AF,$C6,$FE,$34,$FC
	.dc.b	$53,$FA,$E7,$17,$C9,$3E,$3F,$17,$F4,$E7,$7A,$7F,$71,$AE,$6F,$F7
	.dc.b	$3B,$EE,$FF,$75,$C3,$C6,$FF,$77,$C7,$CE,$FF,$79,$CB,$F9,$EE,$5F
	.dc.b	$1B,$4F,$7B,$F1,$8F,$F6,$67,$8F,$FE,$A6,$7F,$A4,$C7,$FA,$A3,$47
	.dc.b	$8D,$FC,$8B,$0F,$E9,$95,$F2,$4F,$E4,$07,$C9,$3E,$A9,$FC,$80,$FB
	.dc.b	$27,$F2,$03,$EE,$9F,$C8,$0F,$C0,$9F,$C8,$0F,$E2,$5C,$9F,$DA,$DF
	.dc.b	$ED,$A7,$C8,$21,$FB,$7E,$6F,$CA,$3F,$C8,$0F,$DB,$4F,$F8,$D7,$BF
	.dc.b	$F6,$CF,$F8,$DF,$BF,$F6,$CF,$F1,$D9,$FD,$B1,$FC,$2C,$FD,$B1,$FC
	.dc.b	$3C,$FD,$B1,$FC,$4C,$FD,$79,$FC,$9F,$C4,$FE,$33,$9F,$C7,$BF,$E3
	.dc.b	$A4,$ED,$FD,$DD,$FF,$8A,$FC,$0F,$F0,$D6,$1F,$B6,$3F,$81,$9F,$B6
	.dc.b	$3F,$83,$9F,$B6,$3F,$C4,$5E,$0F,$A3,$0F,$A9,$E0,$B3,$F6,$C6,$32
	.dc.b	$FE,$3B,$7F,$D1,$EB,$F8,$DB,$74,$FF,$24,$4F,$96,$3F,$C6,$6D,$7F
	.dc.b	$9B,$67,$EC,$7F,$75,$19,$CF,$F8,$9C,$68,$E3,$3B,$DF,$F9,$E4,$60
	.dc.b	$86,$C9,$3C,$89,$8F,$C9,$AE,$57,$DE,$31,$5C,$AB,$6F,$92,$B6,$47
	.dc.b	$FD,$FF,$1E,$6F,$3F,$FD,$1D,$1F,$DD,$FE,$1B,$FC,$58,$47,$FC,$49
	.dc.b	$F0,$FF,$E2,$FF,$88,$FF,$18,$7B,$3F,$E3,$1F,$67,$FC,$65,$F0,$FF
	.dc.b	$E3,$3F,$6B,$FC,$69,$F0,$BF,$E3,$5F,$6F,$FC,$6D,$F0,$DF,$E3,$7F
	.dc.b	$73,$FC,$71,$F0,$FF,$E3,$9F,$77,$FC,$75,$F0,$1F,$E3,$BF,$7B,$F8
	.dc.b	$A7,$2F,$EB,$7B,$FF,$58,$DF,$0F,$3C,$7F,$E0,$DA,$E2,$FF,$8E,$70
	.dc.b	$EA,$FF,$8E,$F2,$F0,$BF,$E3,$DD,$3C,$AF,$F8,$FF,$6F,$4B,$FE,$41
	.dc.b	$E0,$F6,$BF,$E4,$3E,$2F,$F2,$77,$1F,$C3,$8F,$E5,$7C,$9F,$E8,$DF
	.dc.b	$20,$E7,$9F,$FC,$A4,$F0,$9E,$47,$F9,$37,$9B,$E4,$3F,$E4,$FE,$6F
	.dc.b	$90,$FF,$94,$79,$BE,$43,$FE,$53,$E6,$F9,$0F,$F9,$57,$9B,$E4,$3F
	.dc.b	$E5,$7E,$6F,$90,$FF,$16,$3F,$C6,$BD,$5F,$C5,$39,$7E,$3A,$D5,$CB
	.dc.b	$F2,$2B,$FA,$2F,$2F,$C8,$7F,$A3,$F2,$FC,$87,$FA,$4F,$2F,$C8,$7F
	.dc.b	$A5,$F2,$FC,$87,$FA,$6F,$2F,$C8,$7F,$A7,$F2,$FC,$17,$E3,$9F,$E6
	.dc.b	$DE,$3F,$D6,$F4,$FE,$4E,$3F,$3F,$CF,$F9,$A6,$9F,$99,$3F,$38,$E7
	.dc.b	$9B,$F3,$BB,$FE,$AA,$76,$D8,$CF,$57,$E9,$67,$91,$A7,$13,$BA,$E8
	.dc.b	$49,$DD,$72,$BF,$33,$63,$C6,$6A,$78,$57,$38,$EF,$D5,$FA,$C8,$F3
	.dc.b	$47,$F4,$96,$BE,$F3,$18,$30,$FF,$67,$46,$48,$7A,$8D,$BA,$70,$6A
	.dc.b	$7A,$B8,$23,$82,$4F,$78,$EF,$18,$9E,$8D,$8C,$CE,$DF,$F7,$14,$65
	.dc.b	$B1,$97,$CA,$70,$9E,$71,$98,$F3,$35,$F2,$B6,$7A,$47,$23,$AC,$F2
	.dc.b	$22,$1E,$43,$9F,$AC,$EA,$78,$9E,$32,$18,$21,$DC,$87,$23,$66,$0F
	.dc.b	$3B,$D5,$C5,$7A,$1E,$07,$42,$9A,$BF,$71,$27,$94,$68,$87,$63,$CF
	.dc.b	$7E,$D7,$07,$52,$9D,$63,$9B,$83,$9D,$79,$5C,$11,$C1,$2B,$C9,$1F
	.dc.b	$00,$E2,$E1,$F0,$46,$E7,$BB,$63,$EA,$57,$9E,$31,$1C,$2B,$93,$E7
	.dc.b	$3B,$1E,$37,$2F,$82,$BA,$C6,$23,$51,$82,$3C,$A7,$BB,$D4,$F0,$46
	.dc.b	$9F,$C0,$57,$96,$BF,$81,$92,$7F,$7F,$8F,$EF,$23,$15,$FC,$02,$FF
	.dc.b	$7C,$48,$CC,$7F,$0D,$AE,$2F,$F0,$52,$4E,$E3,$2D,$8E,$93,$CB,$2F
	.dc.b	$E1,$D0,$CB,$F8,$54,$E7,$0F,$26,$1C,$A3,$15,$D9,$B8,$75,$8F,$21
	.dc.b	$DA,$78,$11,$C1,$0D,$92,$FC,$31,$B8,$CE,$1E,$62,$61,$E4,$8E,$51
	.dc.b	$87,$C9,$3C,$63,$17,$E5,$6E,$1E,$82,$4F,$F4,$5C,$3F,$95,$71,$BE
	.dc.b	$03,$FC,$E5,$6C,$BF,$CE,$53,$9C,$BF,$C3,$B1,$8A,$DB,$A9,$B3,$FE
	.dc.b	$B1,$5F,$01,$8C,$DF,$C0,$2F,$6F,$0B,$DC,$BF,$11,$8F,$7B,$3F,$11
	.dc.b	$AF,$80,$DF,$C4,$70,$F8,$2E,$3F,$11,$CB,$E1,$39,$FC,$47,$4F,$85
	.dc.b	$EB,$F1,$1D,$BE,$1E,$72,$7A,$8F,$F6,$37,$02,$78,$46,$78,$BC,$F8
	.dc.b	$75,$8C,$53,$2F,$14,$63,$4F,$33,$AD,$BD,$4F,$0E,$0F,$7F,$0F,$C0
	.dc.b	$E2,$F8,$9C,$9F,$1A,$BC,$0F,$F8,$2D,$BC,$DA,$3F,$D4,$DD,$1F,$D8
	.dc.b	$74,$7C,$47,$CA,$25,$79,$BD,$5F,$9B,$E7,$FC,$EF,$80,$F9,$FD,$DF
	.dc.b	$43,$E0,$7E,$8F,$A7,$F5,$7F,$80,$7F,$D6,$58,$BC,$0F,$C1,$3F,$EE
	.dc.b	$B8,$7A,$BB,$F8,$3E,$1F,$0B,$C7,$C4,$E5,$F0,$DC,$FE,$1B,$A7,$C3
	.dc.b	$FD,$2C,$BF,$7D,$C7,$F3,$DF,$9D,$9A,$78,$46,$2F,$A3,$AC,$7C,$0F
	.dc.b	$0C,$FD,$0F,$A8,$FA,$BB,$7B,$9F,$AF,$C1,$F0,$9F,$88,$7D,$AE,$3F
	.dc.b	$8E,$FF,$03,$F0,$7F,$DA,$1A,$7C,$C8,$D1,$0F,$89,$E3,$EB,$97,$A4
	.dc.b	$60,$FC,$5E,$1E,$17,$1E,$07,$E8,$32,$F3,$46,$37,$F9,$FC,$7F,$3A
	.dc.b	$91,$F3,$A7,$E2,$F1,$FC,$E7,$E3,$F1,$FC,$E7,$E4,$F1,$FC,$E7,$F8
	.dc.b	$2E,$9F,$E2,$8B,$F9,$91,$A3,$E6,$1C,$8F,$98,$76,$3E,$61,$F5,$0F
	.dc.b	$3E,$3E,$5D,$59,$FB,$63,$1A,$7C,$27,$5A,$FC,$C7,$86,$FF,$98,$DF
	.dc.b	$E6,$B9,$CE,$E5,$F5,$0F,$8F,$97,$FB,$22,$FA,$C7,$EC,$39,$7F,$30
	.dc.b	$FE,$CB,$97,$F0,$2F,$A0,$F9,$9C,$1F,$1D,$F9,$DC,$1F,$1D,$FA,$1C
	.dc.b	$1F,$1D,$FE,$5F,$97,$F2,$D7,$EA,$72,$FE,$49,$FA,$DC,$BF,$92,$7E
	.dc.b	$C7,$2F,$E4,$9F,$B5,$CB,$E2,$7D,$39,$F9,$9F,$5E,$9F,$B1,$7F,$84
	.dc.b	$F3,$7F,$20,$E2,$F8,$51,$FC,$4B,$87,$E3,$3F,$AF,$F6,$FF,$68,$47
	.dc.b	$F4,$FE,$E7,$ED,$FE,$0B,$F7,$1F,$11,$FB,$9F,$83,$FD,$D7,$C0,$7E
	.dc.b	$EF,$E0,$FF,$79,$F0,$FF,$BD,$F6,$BF,$7D,$F0,$FF,$BF,$F8,$5F,$E0
	.dc.b	$1F,$0F,$FC,$07,$E1,$BF,$81,$7C,$3F,$F3,$7F,$E0,$1F,$E0,$3C,$1F
	.dc.b	$53,$D0,$FE,$52,$FD,$6F,$47,$F7,$8F,$E7,$A7,$7E,$0F,$E6,$FC,$66
	.dc.b	$1D,$6B,$F7,$F9,$78,$5F,$E0,$3A,$79,$5F,$E0,$7B,$7A,$5F,$E0,$BC
	.dc.b	$1E,$D7,$F8,$3F,$17,$C1,$7E,$19,$FC,$2B,$97,$E3,$3F,$BB,$3E,$61
	.dc.b	$FB,$9C,$FF,$26,$FE,$3D,$FC,$E7,$47,$D5,$8F,$AE,$7F,$0A,$3F,$1C
	.dc.b	$7E,$8C,$FC,$74,$7B,$7C,$2F,$C7,$3E,$FF,$0B,$F1,$CF,$C1,$F0,$BF
	.dc.b	$1C,$FE,$A7,$8B,$FC,$2B,$E1,$FF,$21,$C3,$F1,$E7,$ED,$4F,$C7,$9F
	.dc.b	$B5,$3F,$1E,$7E,$D4,$FC,$79,$E3,$3F,$1E,$79,$CF,$49,$FA,$AF,$17
	.dc.b	$F6,$9E,$27,$F1,$78,$F8,$4F,$EC,$1F,$8D,$E3,$FF,$74,$EE,$FC,$9B
	.dc.b	$63,$E3,$4F,$F2,$4D,$7E,$36,$5F,$C0,$FC,$0F,$CF,$3F,$C1,$7C,$1F
	.dc.b	$D1,$3F,$81,$F0,$BF,$4C,$FE,$0F,$C3,$FD,$53,$F8,$5F,$13,$F5,$CF
	.dc.b	$E1,$FC,$5F,$F0,$D6,$5F,$C6,$5C,$BF,$E9,$3D,$3F,$92,$D7,$28,$C7
	.dc.b	$5B,$0E,$F1,$8C,$7C,$6F,$F3,$CC,$BC,$EF,$F3,$EC,$FD,$6F,$B7,$5F
	.dc.b	$79,$F8,$83,$F8,$9E,$DF,$B4,$CF,$E2,$46,$0F,$96,$7F,$82,$BC,$5F
	.dc.b	$9E,$FF,$42,$3F,$8C,$1F,$57,$6F,$8C,$FD,$97,$E3,$3F,$75,$F8,$CF
	.dc.b	$E0,$5F,$8C,$FE,$15,$F8,$CF,$F3,$E3,$FC,$C3,$A7,$F3,$E7,$F8,$E9
	.dc.b	$FC,$F8,$FE,$42,$7F,$3E,$3F,$92,$9F,$9A,$C3,$E4,$BF,$CF,$9F,$EC
	.dc.b	$3E,$3F,$F3,$D7,$FB,$1F,$0F,$B3,$1F,$F1,$CE,$3F,$86,$E2,$FE,$CD
	.dc.b	$87,$E2,$B1,$73,$FE,$CB,$3F,$E0,$4F,$F8,$37,$C6,$FD,$C5,$7F,$61
	.dc.b	$E9,$FD,$E3,$FD,$8F,$AB,$F8,$03,$FD,$97,$AF,$F8,$23,$FD,$9F,$B3
	.dc.b	$F8,$43,$FD,$A7,$B7,$F8,$63,$FD,$AF,$BB,$F8,$C7,$0F,$F1,$4D,$3F
	.dc.b	$AC,$78,$1F,$C6,$1F,$D9,$57,$F8,$F7,$C0,$FE,$2F,$5F,$8A,$BE,$75
	.dc.b	$FC,$67,$0E,$EF,$F1,$BC,$7C,$6F,$F1,$DC,$BC,$EF,$F1,$FC,$FD,$6F
	.dc.b	$F2,$1D,$3D,$EF,$F2,$3D,$7E,$26,$5F,$98,$DB,$F1,$9A,$3E,$2B,$F9
	.dc.b	$81,$36,$F3,$7B,$1F,$CC,$FD,$AF,$E6,$9F,$0F,$FC,$D7,$DC,$FE,$6D
	.dc.b	$F0,$FF,$CD,$FD,$EF,$E7,$1F,$0F,$FC,$E7,$E0,$3F,$9D,$7C,$07,$F3
	.dc.b	$BF,$81,$FE,$79,$F0,$FF,$CF,$7E,$0F,$F9,$F7,$C3,$FF,$3F,$F8,$5F
	.dc.b	$E8,$1F,$0F,$FE,$AA,$E3,$39,$E2,$F9,$15,$C5,$F9,$58,$75,$7E,$66
	.dc.b	$5E,$17,$E7,$69,$E5,$7E,$86,$DE,$97,$E9,$70,$7B,$5F,$79,$F5,$78
	.dc.b	$FE,$13,$FC,$5B,$37,$C3,$FF,$55,$24,$7D,$BF,$6B,$FA,$C7,$B7,$FD
	.dc.b	$67,$E1,$FF,$D2,$1E,$AF,$FA,$3E,$BF,$CF,$8F,$FA,$3D,$BE,$FE,$BE
	.dc.b	$1F,$7F,$03,$C3,$E1,$F8,$FC,$1F,$2F,$87,$E7,$F0,$BD,$3E,$1F,$AF
	.dc.b	$C3,$FE,$3F,$C3,$FF,$05,$78,$BF,$8E,$D7,$FC,$83,$B2,$B8,$47,$8B
	.dc.b	$0E,$8F,$93,$2F,$03,$E6,$D3,$C8,$F9,$8F,$4F,$C3,$FB,$3E,$1F,$DD
	.dc.b	$1F,$9B,$E2,$FE,$F1,$C7,$F1,$23,$FB,$A7,$27,$E3,$8C,$6B,$FE,$4F
	.dc.b	$E5,$F9,$73,$F2,$4E,$47,$E9,$79,$7B,$3E,$03,$C4,$7B,$39,$3E,$3B
	.dc.b	$F3,$39,$3E,$3B,$F3,$B9,$3E,$3B,$F4,$39,$3E,$3B,$FE,$AA,$F7,$7F
	.dc.b	$B1,$98,$7E,$3C,$70,$7E,$4C,$F4,$7E,$5D,$78,$1F,$11,$E4,$CB,$FC
	.dc.b	$D5,$C9,$F7,$63,$E8,$EF,$F1,$9F,$F0,$B6,$FF,$19,$FE,$05,$B7,$C7
	.dc.b	$F8,$77,$C3,$FF,$4F,$E0,$FD,$A5,$7E,$4C,$87,$F8,$7F,$C8,$FC,$C3
	.dc.b	$FE,$21,$EB,$FC,$E3,$FE,$23,$EC,$FD,$03,$FE,$25,$ED,$FD,$23,$FE
	.dc.b	$27,$EE,$FD,$43,$FE,$29,$EF,$FD,$63,$FE,$2B,$F0,$3F,$60,$FF,$8B
	.dc.b	$7C,$1F,$E3,$5D,$FF,$E9,$6F,$47,$F6,$F3,$FC,$6F,$C3,$FD,$DB,$FC
	.dc.b	$77,$C4,$FD,$FB,$FC,$7F,$C5,$FE,$06,$FF,$21,$F1,$BF,$83,$BF,$C8
	.dc.b	$FC,$7F,$E1,$6F,$F2,$5F,$23,$F8,$7B,$FC,$9F,$C9,$FF,$22,$71,$FE
	.dc.b	$D3,$AF,$F1,$5E,$4F,$F1,$A3,$1B,$7C,$97,$77,$C4,$FD,$F9,$D4,$FE
	.dc.b	$06,$78,$4F,$E0,$E7,$94,$FE,$16,$7A,$4F,$E1,$E7,$B4,$FE,$26,$7C
	.dc.b	$13,$F9,$E6,$FF,$C6,$4F,$93,$5F,$9F,$E0,$F9,$AF,$E8,$78,$7E,$8B
	.dc.b	$FA,$3E,$2F,$AA,$FE,$97,$8F,$EC,$BF,$A7,$E4,$FB,$AF,$EA,$79,$7F
	.dc.b	$02,$FE,$AF,$5F,$83,$9B,$D5,$73,$FE,$2A,$3F,$8D,$75,$7E,$CB,$83
	.dc.b	$F1,$CF,$E4,$E3,$F2,$A7,$EC,$4F,$E2,$FD,$9F,$C6,$BC,$5F,$E3,$7E
	.dc.b	$57,$F0,$49,$FE,$1B,$E3,$FF,$08,$7F,$87,$F9,$1F,$C3,$1F,$E2,$3E
	.dc.b	$4F,$F1,$07,$F8,$9F,$95,$FC,$5D,$F3,$7B,$2F,$81,$44,$E1,$FF,$6A
	.dc.b	$4E,$63,$F7,$07,$F2,$AA,$E9,$3F,$B0,$BF,$03,$FB,$2C,$3C,$8F,$ED
	.dc.b	$31,$F4,$3F,$B6,$CB,$D9,$87,$B8,$F8,$07,$C2,$3F,$D1,$9E,$67,$C7
	.dc.b	$7E,$67,$07,$F1,$87,$E5,$9F,$BA,$21,$C0,$FE,$36,$7F,$A5,$7C,$CF
	.dc.b	$E3,$6F,$F3,$4D,$BE,$5B,$FD,$13,$CC,$FE,$19,$8F,$F0,$AF,$27,$F8
	.dc.b	$83,$FC,$33,$CA,$FE,$28,$FF,$0E,$F2,$FF,$8C,$3F,$CF,$CF,$8F,$C5
	.dc.b	$F2,$B0,$FF,$80,$3C,$5F,$25,$FF,$55,$1F,$24,$FE,$48,$7F,$43,$3F
	.dc.b	$94,$1F,$D0,$CF,$E5,$87,$F4,$33,$CC,$7F,$1F,$E5,$FF,$32,$79,$7F
	.dc.b	$C7,$EF,$F9,$0F,$2F,$F1,$F7,$F9,$F7,$97,$F8,$D7,$FA,$07,$3F,$F1
	.dc.b	$FF,$2B,$F6,$D8,$FF,$1C,$9F,$DD,$BF,$DC,$8F,$F5,$E7,$87,$FC,$09
	.dc.b	$FF,$5F,$78,$9F,$C1,$9F,$F6,$07,$8B,$FC,$29,$FF,$61,$78,$DF,$C3
	.dc.b	$9F,$F6,$27,$8F,$FC,$49,$FF,$63,$79,$1F,$C5,$9F,$F6,$47,$93,$FC
	.dc.b	$69,$FF,$66,$1F,$91,$E0,$FE,$3C,$FE,$F0,$FD,$F1,$FC,$94,$FE,$04
	.dc.b	$7F,$29,$3F,$83,$1F,$CB,$4F,$E1,$47,$F3,$13,$F8,$71,$FC,$D4,$FE
	.dc.b	$24,$7F,$39,$3F,$8B,$1F,$CF,$4F,$88,$FC,$67,$FD,$4F,$C1,$FD,$19
	.dc.b	$FF,$05,$EB,$FE,$0E,$23,$82,$1F,$E0,$27,$2D,$33,$AF,$F5,$88,$CB
	.dc.b	$63,$35,$F7,$A7,$2D,$8F,$88,$7E,$3B,$8B,$F2,$15,$F9,$28,$CB,$CD
	.dc.b	$F0,$11,$C1,$27,$A4,$65,$B1,$98,$FE,$39,$7F,$05,$D1,$8B,$F2,$B4
	.dc.b	$F4,$94,$F9,$3E,$67,$F8,$D0,$8F,$F8,$D3,$CB,$FF,$1B,$FB,$5F,$E3
	.dc.b	$8F,$63,$FC,$73,$F0,$1F,$E3,$AF,$6B,$FC,$77,$D1,$FE,$3C,$F6,$FF
	.dc.b	$C7,$BE,$E7,$C0,$DF,$FD,$8D,$FC,$01,$FB,$1D,$1C,$09,$CF,$BD,$B9
	.dc.b	$FE,$6D,$4F,$83,$C7,$96,$8F,$F6,$64,$FE,$4C,$DF,$26,$A1,$F0,$3F
	.dc.b	$99,$CB,$E8,$57,$D3,$3E,$8E,$5E,$49,$E1,$7C,$7E,$07,$96,$5C,$F3
	.dc.b	$E9,$87,$B0,$FF,$2B,$EB,$F0,$1F,$F2,$CE,$FF,$11,$FF,$2D,$F0,$FC
	.dc.b	$87,$E5,$1F,$E5,$EE,$4F,$9A,$FD,$03,$04,$3F,$CA,$BC,$9C,$5F,$F2
	.dc.b	$BC,$F5,$7F,$CB,$35,$E1,$7F,$CB,$77,$E5,$7E,$6E,$BF,$41,$F5,$1F
	.dc.b	$1F,$83,$E5,$67,$EF,$8C,$70,$7F,$23,$E0,$FA,$91,$F4,$B8,$3E,$56
	.dc.b	$1F,$E6,$A3,$FC,$65,$EA,$FC,$F8,$F9,$33,$CB,$83,$E4,$BD,$38,$3E
	.dc.b	$4B,$FE,$7E,$E2,$FC,$33,$FE,$80,$E3,$FC,$53,$FE,$82,$E4,$FC,$73
	.dc.b	$FE,$84,$E5,$FF,$43,$F1,$FC,$8D,$7F,$C9,$7C,$9F,$F0,$07,$9F,$F4
	.dc.b	$11,$FA,$3E,$24,$63,$83,$F4,$E4,$C7,$F4,$15,$B9,$D4,$62,$75,$5C
	.dc.b	$B9,$FF,$52,$D9,$FD,$4B,$5F,$D7,$B6,$FE,$4B,$F9,$7F,$83,$FF,$22
	.dc.b	$CF,$F8,$47,$E0,$99,$FE,$5E,$7A,$E5,$E2,$BE,$59,$78,$A3,$04,$CB
	.dc.b	$64,$C7,$C7,$19,$6C,$7C,$53,$F6,$0F,$F0,$A3,$FB,$15,$7A,$5B,$3F
	.dc.b	$C2,$5B,$3C,$1F,$E0,$C6,$0E,$73,$C5,$E8,$43,$65,$74,$E0,$8E,$09
	.dc.b	$E5,$7F,$B7,$23,$F2,$DB,$FE,$4F,$0F,$F1,$0D,$23,$13,$FE,$10,$E4
	.dc.b	$FE,$72,$F9,$4F,$E1,$D3,$D7,$1F,$64,$78,$DA,$F8,$9B,$1B,$C7,$E0
	.dc.b	$36,$7D,$F3,$A2,$4E,$9F,$E6,$1E,$8B,$D5,$ED,$39,$25,$72,$6B,$E6
	.dc.b	$AE,$6E,$48,$66,$36,$FB,$48,$76,$36,$7D,$CC,$FE,$53,$FC,$7C,$9B
	.dc.b	$FF,$B3,$72,$F8,$F1,$C4,$EC,$72,$30,$76,$3E,$35,$78,$9C,$90,$FE
	.dc.b	$95,$8F,$C1,$76,$73,$29,$D8,$87,$5C,$BE,$73,$83,$A3,$87,$58,$FD
	.dc.b	$6C,$7D,$B8,$FF,$3C,$26,$9F,$C5,$2B,$FB,$E9,$C3,$1F,$43,$E2,$30
	.dc.b	$79,$70,$EC,$F8,$30,$FA,$6F,$9F,$1F,$64,$EA,$39,$1B,$AE,$B1,$92
	.dc.b	$57,$29,$EE,$7B,$4F,$EB,$BA,$FF,$04,$96,$FF,$C1,$27,$F9,$9C,$FD
	.dc.b	$67,$B9,$27,$AD,$F8,$2F,$B4,$F8,$CF,$63,$F9,$4D,$7F,$85,$4F,$F0
	.dc.b	$79,$FE,$2F,$3F,$C1,$DB,$19,$CF,$F9,$23,$73,$E3,$3E,$43,$F6,$BD
	.dc.b	$9F,$DD,$37,$FD,$9E,$5F,$C7,$E7,$F8,$89,$0E,$A5,$C3,$AB,$6F,$AB
	.dc.b	$A3,$FC,$77,$87,$85,$FE,$68,$4A,$FE,$4A,$DC,$7D,$57,$D2,$39,$34
	.dc.b	$97,$F8,$58,$FD,$1B,$FC,$33,$0F,$4E,$DF,$8C,$8F,$1D,$70,$AE,$D8
	.dc.b	$F8,$F1,$F0,$93,$1F,$0D,$F0,$3D,$7E,$CF,$F9,$4E,$75,$5C,$6B,$77
	.dc.b	$D0,$F8,$45,$3D,$37,$CD,$F8,$18,$F9,$5F,$F6,$2E,$7F,$44,$FE,$EA
	.dc.b	$F1,$CB,$D1,$8F,$82,$3A,$B5,$EA,$6E,$FC,$F8,$FF,$8D,$BE,$0F,$FC
	.dc.b	$54,$49,$FF,$02,$FB,$7F,$E2,$DF,$73,$FC,$5D,$F1,$1F,$E2,$FF,$7B
	.dc.b	$FC,$61,$F1,$1F,$E3,$1F,$80,$FF,$19,$7C,$47,$F8,$CF,$E0,$BF,$C6
	.dc.b	$9F,$11,$FE,$35,$F8,$4F,$F1,$B7,$C4,$7F,$8D,$FE,$1B,$FC,$71,$CD
	.dc.b	$FE,$39,$F8,$7F,$EC,$9C,$1F,$E1,$7E,$CF,$D2,$3F,$E9,$23,$FC,$3A
	.dc.b	$7E,$A8,$FF,$0C,$C7,$27,$FC,$37,$5D,$9F,$F0,$EE,$1E,$27,$FC,$3F
	.dc.b	$97,$99,$FF,$10,$E9,$EA,$7F,$C4,$7B,$7B,$9F,$F1,$2F,$07,$F9,$37
	.dc.b	$87,$E2,$47,$8B,$97,$FC,$A9,$C7,$FE,$CD,$E9,$FC,$07,$F0,$0F,$F8
	.dc.b	$AA,$B4,$DE,$8F,$F0,$2B,$7A,$3F,$AE,$AF,$88,$FF,$2F,$AE,$FB,$EA
	.dc.b	$63,$E1,$33,$EA,$EB,$E1,$77,$EB,$70,$F8,$6E,$3E,$BF,$2F,$17,$9F
	.dc.b	$AF,$D3,$E1,$FA,$FB,$1D,$BE,$23,$E7,$F1,$FE,$90,$FC,$69,$FB,$F3
	.dc.b	$D2,$F0,$7F,$C9,$5C,$0A,$EB,$18,$C3,$C2,$EB,$2F,$2B,$C3,$4F,$4B
	.dc.b	$CB,$6F,$6B,$D3,$83,$E0,$F0,$FC,$2E,$29,$21,$C9,$F3,$A3,$31,$F3
	.dc.b	$9E,$2F,$33,$CF,$C7,$F2,$5F,$4F,$1F,$C9,$7D,$7C,$7F,$25,$F6,$F1
	.dc.b	$FC,$97,$DF,$C7,$F2,$5F,$83,$C7,$F2,$5F,$8D,$C9,$FD,$49,$F9,$27
	.dc.b	$E1,$7D,$47,$97,$B8,$FC,$1D,$73,$7F,$0B,$7D,$DF,$C3,$E1,$E3,$7F
	.dc.b	$13,$8F,$9D,$FC,$5E,$5E,$B7,$F1,$B9,$FE,$23,$7F,$83,$1F,$0C,$FA
	.dc.b	$FA,$FE,$57,$C8,$FC,$7B,$F2,$36,$FC,$9B,$C1,$FC,$AD,$7D,$4A,$FC
	.dc.b	$CF,$0F,$D8,$7F,$37,$C5,$F7,$1F,$CE,$F1,$FE,$01,$FC,$FF,$27,$E1
	.dc.b	$1F,$D0,$F2,$FE,$25,$FC,$59,$F2,$32,$FF,$46,$3F,$BB,$32,$7C,$63
	.dc.b	$F1,$F1,$CF,$2F,$D4,$D7,$77,$F5,$77,$E3,$7F,$5B,$87,$9D,$FD,$7E
	.dc.b	$3E,$B7,$DB,$87,$E6,$B9,$7E,$16,$5F,$12,$3F,$1D,$CB,$FB,$AF,$67
	.dc.b	$F6,$64,$8F,$D0,$F9,$BF,$B6,$F6,$BF,$6F,$F0,$7F,$B8,$F7,$3F,$73
	.dc.b	$F0,$7F,$ED,$8F,$DF,$DE,$FD,$F8,$4C,$7C,$2E,$7E,$17,$5F,$0D,$BF
	.dc.b	$86,$E1,$F0,$FC,$7E,$1F,$97,$C4,$73,$F8,$8F,$F0,$47,$A7,$FC,$93
	.dc.b	$A3,$FD,$1F,$B2,$77,$C9,$FE,$BF,$86,$1D,$23,$19,$78,$1D,$69,$E4
	.dc.b	$78,$6D,$E9,$DF,$D7,$1F,$C0,$38,$7D,$EF,$F0,$2E,$3F,$E2,$FC,$9F
	.dc.b	$16,$3F,$8F,$F2,$FE,$3D,$79,$6F,$27,$F9,$04,$61,$B1,$9E,$9F,$C9
	.dc.b	$1F,$C9,$3C,$0F,$B0,$FF,$26,$EC,$FB,$8F,$F2,$8E,$DF,$C0,$3F,$CA
	.dc.b	$BB,$BF,$08,$FF,$2C,$EF,$FC,$43,$FC,$BB,$C0,$FC,$97,$3F,$F3,$2F
	.dc.b	$2B,$F9,$0B,$FB,$9C,$FF,$D1,$DE,$D6,$2F,$F9,$AD,$70,$79,$6D,$FC
	.dc.b	$10,$FE,$40,$7F,$08,$3F,$90,$1F,$C3,$0F,$E4,$07,$F1,$03,$F9,$01
	.dc.b	$FC,$50,$FE,$2C,$7F,$46,$F0,$BF,$26,$FC,$58,$F9,$57,$F1,$63,$F7
	.dc.b	$64,$38,$1C,$B0,$FE,$9B,$E2,$7F,$05,$7F,$A7,$F8,$BF,$C2,$5F,$EA
	.dc.b	$3E,$37,$F0,$D7,$FA,$9F,$8F,$FC,$45,$FE,$AB,$E4,$7F,$15,$7F,$AB
	.dc.b	$F9,$3F,$C6,$5F,$F6,$EF,$97,$FC,$73,$2F,$F2,$19,$0F,$E4,$1E,$5E
	.dc.b	$F5,$FE,$BB,$7C,$DF,$EB,$F8,$77,$7F,$B0,$E3,$E3,$7F,$B1,$E5,$E7
	.dc.b	$7F,$B2,$E7,$EB,$7F,$B3,$E9,$FC,$7F,$2F,$DD,$1F,$0F,$4F,$E3,$6F
	.dc.b	$EE,$BC,$BF,$DE,$3F,$C8,$C9,$1F,$DC,$B7,$E5,$1F,$CA,$8F,$AA,$7F
	.dc.b	$2E,$3E,$C9,$FC,$C8,$FB,$A7,$F3,$63,$F0,$27,$F3,$A3,$F8,$B9,$FD
	.dc.b	$18,$FE,$CF,$C1,$F8,$B7,$FA,$2C,$FE,$3F,$7F,$95,$1F,$C8,$23,$E7
	.dc.b	$BC,$8E,$87,$F2,$C3,$F1,$C7,$F3,$03,$F1,$C7,$F3,$43,$F1,$C7,$F3
	.dc.b	$83,$F1,$C7,$F8,$4B,$9F,$F5,$AF,$E3,$4F,$C5,$79,$7F,$C7,$DF,$EF
	.dc.b	$BD,$5F,$BB,$AF,$F0,$BF,$5F,$F1,$F9,$FF,$00,$F9,$7F,$C7,$DF,$F0
	.dc.b	$1F,$97,$FC,$1D,$FE,$12,$7F,$11,$F2,$FF,$8F,$3F,$C4,$FC,$BF,$E3
	.dc.b	$CF,$F1,$5F,$2F,$F8,$F3,$FC,$5F,$CB,$FE,$3C,$FF,$90,$7D,$DF,$F5
	.dc.b	$37,$95,$FC,$76,$34,$7F,$1D,$F2,$BF,$8E,$3F,$E3,$38,$79,$5F,$C7
	.dc.b	$63,$1E,$57,$F1,$D7,$5E,$57,$9D,$FF,$1A,$EB,$EB,$7F,$C6,$FB,$7B
	.dc.b	$DF,$F1,$CE,$FF,$0F,$A7,$FC,$6F,$C5,$FB,$4C,$DF,$13,$FC,$FC,$97
	.dc.b	$FC,$C7,$A7,$FA,$17,$BF,$FD,$0F,$83,$FA,$27,$C0,$7F,$45,$F8,$9F
	.dc.b	$E8,$DF,$05,$FD,$1F,$E2,$7F,$A4,$7C,$1F,$F4,$9F,$89,$FE,$95,$F0
	.dc.b	$BF,$E2,$1D,$3F,$CA,$0D,$6F,$C3,$67,$B5,$C5,$FE,$27,$F2,$96,$F8
	.dc.b	$9D,$4E,$61,$F1,$1F,$43,$1F,$F2,$15,$FF,$22,$7F,$A5,$7A,$FC,$1F
	.dc.b	$E7,$75,$C1,$FF,$1A,$63,$E0,$73,$8E,$EF,$C3,$95,$BF,$45,$FF,$58
	.dc.b	$7D,$D7,$FD,$5F,$4F,$84,$FF,$59,$DB,$E3,$3F,$D6,$F8,$3E,$53,$FD
	.dc.b	$77,$8B,$FB,$77,$1F,$D0,$8F,$C1,$72,$7F,$28,$C7,$FD,$AA,$4F,$3B
	.dc.b	$C5,$F1,$3E,$3F,$3F,$C9,$ED,$79,$7D,$BF,$37,$B9,$E7,$F8,$8F,$47
	.dc.b	$BD,$E9,$F8,$8F,$57,$C0,$7A,$FE,$23,$D9,$F0,$5E,$DF,$88,$F7,$7C
	.dc.b	$27,$BF,$E2,$3E,$67,$4F,$F4,$DE,$9F,$F7,$3F,$89,$F9,$BE,$6F,$B7
	.dc.b	$5C,$1F,$97,$E0,$7D,$5C,$7A,$C6,$0F,$A3,$9F,$89,$C7,$6F,$DE,$DB
	.dc.b	$CF,$18,$E7,$70,$FB,$23,$1C,$7F,$01,$D7,$2F,$C4,$78,$73,$FD,$73
	.dc.b	$E5,$15,$C9,$D7,$9F,$E5,$3D,$F9,$FE,$53,$E1,$E7,$F9,$4F,$8F,$9F
	.dc.b	$E5,$3E,$5E,$7F,$94,$F9,$F9,$FE,$53,$E9,$E7,$F9,$4F,$E4,$8F,$E9
	.dc.b	$DC,$FF,$DD,$F9,$BE,$53,$F9,$1E,$2F,$CF,$F3,$7C,$BA,$FD,$3F,$2F
	.dc.b	$CA,$7F,$53,$CB,$F2,$5F,$D5,$F2,$FC,$87,$F5,$BC,$9F,$1D,$FD,$7F
	.dc.b	$27,$C7,$7F,$63,$C9,$F1,$DF,$D9,$F2,$7C,$57,$FD,$2B,$D3,$FB,$1D
	.dc.b	$BF,$82,$C7,$F8,$CF,$A7,$F2,$DB,$FF,$8E,$2B,$F8,$2C,$FF,$8D,$FC
	.dc.b	$4F,$E0,$AF,$F8,$E7,$C4,$FE,$0A,$FF,$8E,$FC,$4F,$E0,$AF,$F8,$F7
	.dc.b	$C4,$FE,$0A,$FD,$F3,$F5,$DE,$0F,$EB,$FC,$2F,$D8,$78,$9F,$B1,$F8
	.dc.b	$6F,$F2,$4E,$56,$FF,$0D,$3B,$7F,$11,$7F,$B7,$7A,$9F,$CB,$70,$F0
	.dc.b	$BF,$CB,$F1,$F2,$BF,$CC,$72,$F4,$BF,$CC,$F0,$F6,$BF,$CD,$73,$F8
	.dc.b	$3E,$07,$C5,$DB,$F9,$51,$89,$FF,$23,$EF,$F2,$1E,$07,$EF,$CF,$8C
	.dc.b	$7F,$03,$3E,$31,$FC,$1C,$F8,$C7,$F0,$B3,$E3,$1F,$E6,$3F,$03,$D8
	.dc.b	$FB,$8F,$D3,$6F,$F1,$9F,$C7,$1F,$C8,$BC,$1F,$DB,$F8,$B9,$8F,$E6
	.dc.b	$33,$F3,$9F,$A0,$75,$3B,$9F,$D4,$E7,$F8,$DB,$FD,$57,$C5,$FE,$36
	.dc.b	$FF,$57,$F1,$7F,$8D,$BF,$D6,$7C,$5F,$E3,$6F,$E1,$C8,$7E,$30,$FE
	.dc.b	$63,$CD,$F8,$FE,$6F,$D0,$67,$FC,$7B,$C5,$FD,$9D,$7F,$46,$F1,$7F
	.dc.b	$66,$FF,$48,$F1,$7F,$66,$FF,$4A,$F1,$7F,$66,$FF,$4C,$F1,$7F,$66
	.dc.b	$FF,$4E,$F1,$7F,$66,$FF,$50,$F1,$7F,$66,$FF,$6F,$E2,$FE,$7D,$C2
	.dc.b	$DA,$9F,$C8,$39,$7F,$A7,$78,$9F,$B3,$E0,$F3,$F8,$9F,$B3,$7D,$3E
	.dc.b	$27,$EC,$DF,$5F,$89,$FB,$37,$DB,$E2,$7E,$CD,$F7,$F8,$9F,$B3,$7E
	.dc.b	$0F,$89,$FB,$37,$E1,$E7,$FB,$3F,$49,$DD,$FD,$7E,$3E,$46,$B7,$83
	.dc.b	$F8,$D1,$FC,$88,$FE,$34,$7F,$26,$3F,$8D,$1F,$CA,$F2,$F2,$47,$F6
	.dc.b	$9C,$FD,$0F,$F6,$BD,$3D,$8F,$F3,$9A,$FE,$34,$7F,$33,$DB,$E2,$3F
	.dc.b	$E4,$28,$FE,$ED,$BF,$E3,$A3,$47,$03,$F8,$C9,$FC,$94,$FE,$79,$5E
	.dc.b	$07,$FB,$9D,$F9,$1F,$EE,$B8,$7C,$7B,$FE,$FD,$B7,$C7,$7F,$C0,$1B
	.dc.b	$7C,$77,$E1,$57,$F1,$A3,$FC,$D0,$7E,$CB,$BB,$F6,$9E,$33,$BF,$F6
	.dc.b	$9C,$3F,$C6,$6B,$F6,$87,$EF,$CF,$DA,$1F,$C0,$CF,$DA,$1F,$C1,$CF
	.dc.b	$DA,$1F,$C2,$CF,$D4,$1F,$E1,$C3,$F5,$87,$EB,$8F,$E2,$A7,$F5,$BF
	.dc.b	$1B,$F8,$CB,$4C,$9B,$3F,$C4,$67,$ED,$0F,$F2,$27,$67,$F8,$39,$FD
	.dc.b	$A1,$FC,$18,$FD,$A1,$FC,$28,$FD,$A1,$FC,$38,$FD,$A1,$FE,$3B,$3F
	.dc.b	$67,$3F,$EF,$0C,$3F,$C1,$D5,$FC,$AE,$7F,$C6,$AD,$9F,$DA,$B7,$1F
	.dc.b	$F1,$83,$67,$FC,$56,$DC,$7F,$C6,$ED,$C0,$D2,$4F,$C1,$3F,$8B,$70
	.dc.b	$E1,$F0,$99,$C3,$C2,$DC,$3C,$2D,$3C,$25,$C3,$C2,$DC,$3C,$2D,$C3
	.dc.b	$B7,$1F,$EB,$23,$FC,$87,$F1,$1F,$E3,$D2,$3F,$E1,$1F,$88,$FF,$21
	.dc.b	$7F,$00,$FF,$BF,$87,$B3,$8F,$5F,$3E,$D6,$BE,$13,$7E,$E7,$0F,$84
	.dc.b	$E3,$EF,$72,$F8,$4E,$7F,$01,$D3,$E1,$3A,$FC,$17,$6F,$84,$FD,$6F
	.dc.b	$2F,$F9,$67,$87,$FC,$B7,$D5,$FE,$06,$E7,$FF,$32,$71,$2F,$8D,$7F
	.dc.b	$8A,$A1,$8F,$68,$C6,$7E,$27,$5A,$F9,$9E,$1B,$FA,$9E,$5C,$3F,$03
	.dc.b	$8B,$E0,$E7,$FE,$61,$E3,$F8,$B5,$FE,$7A,$E4,$FF,$2B,$18,$E8,$FE
	.dc.b	$76,$EE,$3C,$95,$CD,$EA,$79,$F9,$7E,$43,$E9,$E5,$F9,$0F,$AF,$97
	.dc.b	$E4,$3E,$DE,$5F,$90,$FB,$F9,$7E,$43,$FD,$83,$2F,$F2,$B7,$4F,$CA
	.dc.b	$E3,$F8,$D3,$B3,$E3,$71,$FC,$55,$5F,$E3,$19,$71,$FC,$79,$CE,$BF
	.dc.b	$43,$8F,$E4,$46,$07,$32,$DC,$9F,$2E,$75,$39,$E4,$F8,$51,$8A,$FF
	.dc.b	$48,$3F,$EA,$4E,$3F,$95,$8F,$FA,$77,$8F,$F6,$B3,$F9,$92,$1F,$A6
	.dc.b	$26,$9F,$A6,$6C,$FE,$99,$A7,$E9,$8B,$8F,$E9,$9B,$D3,$FA,$66,$F4
	.dc.b	$FF,$16,$69,$E0,$BE,$C4,$BE,$CD,$8E,$CD,$BE,$CD,$BE,$CD,$BF,$D9
	.dc.b	$F8,$5E,$7B,$FF,$63,$5F,$95,$B7,$E5,$6D,$7F,$B2,$5E,$04,$3D,$C6
	.dc.b	$8D,$4F,$37,$4D,$76,$62,$7B,$36,$7B,$3A,$24,$BE,$25,$E8,$FF,$B1
	.dc.b	$DB,$3F,$C7,$4F,$11,$FE,$DF,$9F,$01,$2F,$BC,$74,$6C,$66,$FC,$2D
	.dc.b	$9F,$5C,$E8,$93,$A6,$C6,$A3,$FD,$E1,$B7,$F0,$48,$ED,$1B,$23,$B2
	.dc.b	$1E,$79,$E6,$D8,$DB,$4D,$9E,$E2,$1F,$24,$C9,$F4,$3C,$3F,$F1,$73
	.dc.b	$FE,$A2,$D3,$4F,$32,$1C,$8F,$39,$C8,$A6,$8E,$26,$F0,$F8,$CF,$3C
	.dc.b	$3E,$04,$62,$35,$87,$B1,$B8,$7B,$1A,$77,$9E,$0F,$62,$1F,$82,$C3
	.dc.b	$D6,$FE,$F9,$FF,$7B,$CF,$38,$DC,$F2,$7D,$27,$63,$D4,$60,$DD,$7D
	.dc.b	$0B,$F4,$BE,$78,$CD,$75,$78,$12,$BB,$12,$78,$46,$71,$FC,$AB,$FC
	.dc.b	$51,$BA,$7F,$09,$8F,$21,$0E,$E7,$C3,$9E,$75,$D6,$FC,$4F,$D0,$C3
	.dc.b	$CF,$1F,$DE,$EF,$F7,$D7,$FC,$27,$2F,$E0,$E4,$7F,$83,$B6,$FF,$83
	.dc.b	$B7,$0F,$E0,$ED,$BE,$8A,$BF,$98,$E3,$EA,$8C,$5F,$ED,$73,$DD,$76
	.dc.b	$24,$76,$6E,$1D,$9B,$87,$09,$D1,$27,$4D,$8F,$F1,$47,$6F,$F8,$2E
	.dc.b	$31,$5E,$F3,$99,$5C,$BF,$CB,$09,$97,$C6,$6E,$3F,$E6,$CE,$5F,$F4
	.dc.b	$9C,$71,$C7,$C2,$4C,$7C,$CD,$8D,$46,$0F,$E5,$AF,$88,$EA,$5B,$F0
	.dc.b	$12,$FE,$74,$72,$21,$FE,$36,$2C,$6D,$E5,$1C,$1E,$B8,$7B,$A7,$4F
	.dc.b	$C0,$C7,$E3,$70,$BE,$17,$FC,$5A,$4B,$FF,$0D,$7C,$2F,$F8,$C3,$D9
	.dc.b	$FF,$18,$FB,$5F,$E3,$2F,$87,$FF,$19,$FB,$9F,$E3,$4F,$88,$FF,$1A
	.dc.b	$FB,$DF,$83,$C7,$FB,$7F,$F0,$0B,$E2,$B1,$C0,$E1,$FC,$2C,$BE,$07
	.dc.b	$31,$8F,$88,$B8,$7F,$7C,$A7,$C1,$E3,$91,$0F,$88,$C2,$F8,$5B,$F1
	.dc.b	$18,$F8,$7F,$ED,$5A,$F8,$72,$E9,$1A,$7F,$35,$19,$C7,$AD,$F4,$C7
	.dc.b	$AE,$5D,$B3,$EF,$AF,$83,$6F,$83,$97,$F9,$07,$83,$E2,$BF,$E4,$3E
	.dc.b	$2F,$92,$FF,$91,$79,$3F,$CB,$7C,$BF,$3E,$3F,$C2,$5C,$DF,$27,$A3
	.dc.b	$F3,$5C,$6F,$88,$F6,$92,$BC,$FC,$BF,$2B,$D8,$F8,$3E,$8F,$C2,$F0
	.dc.b	$FE,$1F,$C0,$7C,$4F,$07,$E2,$FC,$0F,$C6,$F8,$4F,$8F,$F0,$7F,$23
	.dc.b	$E1,$7E,$4F,$91,$F2,$BE,$1B,$E5,$F8,$DF,$3F,$F8,$03,$EE,$70,$7F
	.dc.b	$8A,$97,$C4,$7C,$D8,$F1,$9F,$C0,$B4,$FE,$E4,$65,$E1,$1C,$DE,$A7
	.dc.b	$0F,$03,$ED,$E1,$E2,$8C,$72,$33,$F3,$C6,$35,$F5,$BA,$C7,$DE,$F0
	.dc.b	$E0,$F8,$6F,$2E,$2F,$90,$FE,$53,$93,$E7,$4F,$E4,$9C,$F8,$3F,$81
	.dc.b	$CB,$E8,$57,$13,$C7,$C9,$F3,$5F,$2F,$27,$CD,$7C,$FC,$9F,$35,$F4
	.dc.b	$F2,$7C,$D7,$D7,$C9,$F3,$5F,$6F,$27,$C3,$7E,$17,$27,$CC,$7F,$D5
	.dc.b	$53,$FD,$73,$3F,$CA,$3F,$D7,$23,$7C,$9C,$5F,$D7,$DF,$57,$F6,$33
	.dc.b	$E1,$7F,$67,$87,$95,$FD,$AE,$3E,$97,$F6,$F9,$7B,$5F,$DC,$E7,$FA
	.dc.b	$CD,$FE,$2F,$2F,$C7,$9F,$F7,$EF,$2F,$EC,$B8,$33,$1F,$B7,$78,$3F
	.dc.b	$20,$FA,$87,$C8,$3E,$C1,$F2,$0F,$B8,$7C,$83,$F0,$3C,$9F,$83,$7F
	.dc.b	$85,$F2,$FE,$21,$FC,$91,$F9,$5D,$FF,$DC,$1B,$7C,$7C,$7F,$9C,$4F
	.dc.b	$FB,$4F,$87,$E6,$3C,$0F,$A2,$7E,$3A,$FB,$70,$FE,$07,$1F,$C7,$CF
	.dc.b	$F1,$AE,$5F,$C7,$BF,$C7,$39,$7F,$1E,$FF,$1E,$E5,$FC,$7B,$FC,$83
	.dc.b	$7F,$87,$BF,$C5,$8F,$8F,$9B,$E2,$3F,$58,$48,$FC,$0F,$C4,$FE,$BF
	.dc.b	$CB,$FD,$87,$C4,$FE,$C7,$DD,$FD,$97,$C4,$FE,$CF,$DF,$FD,$A7,$C4
	.dc.b	$FE,$D7,$E0,$7F,$6D,$F1,$3F,$B7,$F8,$3F,$DC,$7C,$4F,$EE,$7E,$17
	.dc.b	$F7,$5F,$13,$FB,$BF,$88,$FD,$A7,$4F,$F1,$FF,$63,$FB,$C7,$17,$EE
	.dc.b	$4F,$D8,$FA,$5B,$7F,$2B,$7C,$DF,$CB,$E3,$DD,$FC,$CE,$7E,$37,$F3
	.dc.b	$7A,$F9,$DF,$CE,$EF,$EB,$7F,$3F,$C3,$EF,$7E,$09,$FA,$2E,$4F,$88
	.dc.b	$BC,$7F,$E4,$7E,$0F,$CC,$F1,$FF,$6B,$3F,$3C,$FC,$71,$F4,$CF,$C7
	.dc.b	$1F,$5C,$FC,$71,$F6,$CF,$C7,$1F,$D5,$BB,$BF,$03,$E0,$7F,$54,$F0
	.dc.b	$3F,$1D,$C5,$FD,$0F,$AF,$F9,$0B,$7C,$2F,$93,$5F,$E1,$EE,$4F,$93
	.dc.b	$DB,$FE,$1F,$CB,$E4,$9F,$B6,$3F,$7A,$74,$9E,$CF,$F0,$43,$F3,$E7
	.dc.b	$92,$FF,$89,$78,$3F,$90,$7F,$8A,$78,$3F,$90,$7F,$8B,$78,$3F,$90
	.dc.b	$7F,$8C,$78,$3F,$17,$83,$F8,$EE,$1F,$B3,$3F,$95,$BF,$26,$38,$63
	.dc.b	$FD,$CA,$FA,$3F,$DD,$30,$F0,$3F,$DD,$B1,$FD,$1E,$3F,$A5,$C7,$D0
	.dc.b	$7E,$D0,$FD,$59,$FA,$D3,$F9,$F9,$FC,$6B,$C0,$FF,$14,$BF,$EE,$0D
	.dc.b	$BF,$67,$3F,$36,$3E,$21,$FD,$FF,$7F,$A6,$FF,$80,$78,$3C,$38,$F8
	.dc.b	$E3,$ED,$D7,$C4,$7E,$FB,$F1,$1F,$C1,$BF,$11,$F8,$24,$3F,$19,$9F
	.dc.b	$F1,$B7,$FA,$21,$98,$E0,$FF,$22,$3F,$8C,$9F,$C9,$8F,$E3,$27,$F2
	.dc.b	$A3,$F8,$C9,$FC,$B8,$FB,$67,$F3,$23,$F8,$71,$F1,$9F,$80,$FE,$14
	.dc.b	$3F,$EF,$7F,$0B,$F8,$D4,$37,$47,$F4,$18,$F9,$A7,$CF,$9F,$EB,$1E
	.dc.b	$3F,$F4,$17,$FA,$D4,$FF,$41,$7F,$AE,$78,$FF,$D0,$5F,$EB,$DE,$3F
	.dc.b	$F4,$17,$FB,$07,$8F,$FD,$05,$FE,$C5,$E3,$FF,$3A,$C7,$E2,$93,$A3
	.dc.b	$FC,$DD,$C1,$FC,$6E,$3F,$70,$43,$FC,$0D,$E3,$FE,$F5,$FF,$04,$75
	.dc.b	$FF,$01,$7F,$C1,$5D,$9F,$C1,$5F,$F0,$67,$6F,$F0,$97,$FC,$1D,$DD
	.dc.b	$FC,$35,$FF,$08,$77,$FF,$11,$7F,$C2,$5E,$07,$F1,$57,$FC,$9F,$E0
	.dc.b	$FF,$8F,$B9,$BF,$C9,$AE,$77,$FD,$D5,$71,$7F,$79,$7D,$5F,$DF,$61
	.dc.b	$E1,$7F,$80,$63,$E5,$7F,$81,$65,$E9,$7F,$82,$67,$ED,$7F,$83,$69
	.dc.b	$FC,$61,$FE,$11,$E0,$FE,$CB,$99,$BB,$E2,$7F,$C5,$84,$E0,$FF,$00
	.dc.b	$FC,$4F,$F8,$BF,$DA,$FF,$18,$7B,$9F,$E3,$1F,$77,$FC,$C1,$EC,$FF
	.dc.b	$96,$6F,$F0,$FC,$BF,$EA,$E8,$7C,$3E,$BD,$ED,$FC,$3F,$0F,$80,$E3
	.dc.b	$F0,$FC,$BE,$0B,$9F,$C3,$F4,$F8,$4E,$BF,$0F,$DB,$E1,$BE,$56,$DF
	.dc.b	$91,$D5,$1A,$8F,$F1,$C4,$2B,$9C,$63,$0E,$EE,$B2,$F1,$BC,$34,$F5
	.dc.b	$EB,$EC,$DB,$DD,$3F,$E2,$2E,$0F,$84,$FC,$43,$FC,$4B,$C7,$F1,$DF
	.dc.b	$DC,$9E,$5F,$87,$F8,$A4,$7C,$BF,$11,$F2,$3D,$AF,$93,$ED,$FC,$AF
	.dc.b	$73,$E5,$FC,$3F,$CC,$F7,$BE,$6F,$C3,$FC,$EF,$80,$F9,$FC,$9F,$E8
	.dc.b	$FF,$E0,$1F,$F4,$D6,$9F,$E5,$77,$FD,$30,$DF,$84,$D7,$C2,$EF,$E1
	.dc.b	$BF,$A2,$CF,$EE,$7D,$2F,$DD,$F8,$7F,$BA,$F5,$F2,$B1,$FF,$4D,$57
	.dc.b	$5C,$3C,$18,$F8,$7E,$1B,$C5,$F0,$DE,$4A,$FA,$3A,$FA,$1F,$A7,$BF
	.dc.b	$B1,$FA,$BC,$3F,$01,$FA,$FC,$7F,$8A,$E4,$F8,$D1,$F8,$A3,$E5,$1F
	.dc.b	$E3,$CE,$17,$2E,$BC,$8F,$F5,$24,$62,$7E,$69,$D0,$FC,$1F,$37,$CD
	.dc.b	$7F,$0B,$CD,$F3,$5F,$C3,$F3,$7C,$D7,$F1,$35,$EC,$7E,$66,$FF,$01
	.dc.b	$F8,$47,$CF,$E6,$F8,$FC,$DF,$28,$F9,$7F,$00,$F4,$7E,$85,$70,$3D
	.dc.b	$9C,$FF,$37,$97,$E5,$47,$D9,$E5,$F9,$4E,$B3,$72,$FC,$B8,$C7,$2F
	.dc.b	$CB,$75,$CB,$F2,$DE,$1C,$BF,$2D,$FD,$1F,$27,$CC,$F4,$BF,$92,$F5
	.dc.b	$7E,$5F,$9B,$FC,$14,$FF,$AA,$FC,$9F,$CE,$DF,$FA,$AF,$B3,$F4,$2F
	.dc.b	$8F,$B7,$F4,$AF,$E9,$CF,$D7,$F8,$FF,$BB,$7F,$63,$E3,$FE,$ED,$FD
	.dc.b	$9F,$8F,$FB,$B7,$F6,$BE,$3F,$EE,$DF,$F5,$77,$BF,$FE,$16,$3F,$C2
	.dc.b	$FC,$7F,$BC,$8D,$9F,$BC,$F1,$FF,$76,$FC,$5F,$1F,$F7,$6F,$C7,$F1
	.dc.b	$FF,$76,$FC,$9F,$1F,$CE,$FA,$4F,$F4,$DF,$93,$FC,$69,$FF,$4E,$F9
	.dc.b	$5F,$05,$FD,$D9,$0F,$DD,$F6,$7F,$26,$E3,$F9,$1C,$29,$FE,$C5,$E5
	.dc.b	$FF,$29,$7F,$82,$1F,$ED,$3B,$EC,$FF,$B5,$70,$FE,$18,$FF,$16,$F2
	.dc.b	$BF,$92,$3F,$28,$FE,$28,$7C,$A3,$F8,$C1,$F2,$8F,$84,$7C,$43,$F0
	.dc.b	$DE,$5F,$F2,$4C,$9B,$FC,$99,$FC,$57,$0F,$D1,$7F,$19,$C5,$F5,$5F
	.dc.b	$C7,$71,$FD,$97,$F2,$1C,$9E,$47,$F2,$59,$7A,$1F,$51,$FC,$8C,$F9
	.dc.b	$27,$C0,$E7,$FC,$5B,$F2,$7A,$1E,$CF,$F4,$3E,$AF,$DC,$71,$5A,$FD
	.dc.b	$B9,$F9,$83,$F6,$E7,$F0,$63,$F6,$E7,$F0,$A3,$F6,$E7,$F0,$E3,$F6
	.dc.b	$E7,$EA,$4F,$DB,$9F,$C8,$4F,$D8,$5A,$3E,$C6,$DF,$C7,$7C,$4F,$DC
	.dc.b	$F9,$9F,$E4,$2A,$FF,$77,$78,$7F,$B8,$C3,$F6,$FE,$63,$CC,$FE,$05
	.dc.b	$18,$F2,$3F,$83,$3F,$B7,$E4,$FE,$14,$FE,$E7,$C9,$FE,$1C,$FE,$EF
	.dc.b	$CA,$FE,$24,$FE,$F7,$CB,$FE,$84,$FC,$2C,$FF,$8D,$F0,$7E,$DF,$87
	.dc.b	$E3,$4F,$F0,$38,$F9,$2F,$F6,$1F,$33,$8B,$CC,$F9,$07,$73,$E4,$1E
	.dc.b	$33,$E4,$1E,$73,$E4,$1E,$B3,$E4,$1F,$03,$4F,$85,$7F,$12,$3F,$40
	.dc.b	$FF,$C0,$0D,$FE,$47,$17,$F8,$72,$7F,$92,$1F,$41,$F9,$FB,$BC,$AF
	.dc.b	$E3,$F3,$F3,$3C,$AF,$E3,$F8,$7F,$3A,$F2,$BF,$8F,$3F,$CF,$3C,$AF
	.dc.b	$E3,$CF,$F3,$EF,$2B,$F8,$F3,$FD,$03,$CA,$FE,$3D,$87,$C5,$3F,$8F
	.dc.b	$9F,$ED,$CF,$0F,$F6,$EF,$ED,$9F,$DF,$3F,$B5,$3F,$34,$7E,$70,$FE
	.dc.b	$0C,$7E,$80,$FE,$3C,$7E,$90,$FD,$A9,$FA,$83,$F6,$A7,$EB,$0F,$DA
	.dc.b	$9F,$B0,$3F,$C2,$7E,$07,$ED,$79,$BF,$69,$7F,$D6,$3C,$0F,$DA,$3F
	.dc.b	$D6,$BC,$0F,$DA,$3F,$D7,$3C,$0F,$DA,$3F,$D7,$BC,$0F,$DA,$3F,$D8
	.dc.b	$3C,$0F,$DA,$3F,$EC,$3F,$07,$F8,$7B,$FE,$C5,$F0,$BD,$EF,$ED,$34
	.dc.b	$FE,$A3,$E1,$FF,$17,$7F,$8E,$F2,$BE,$1F,$FC,$18,$48,$FF,$06,$35
	.dc.b	$FD,$1B,$6B,$F5,$91,$FE,$83,$9E,$4F,$02,$1C,$0B,$1C,$3C,$3F,$F0
	.dc.b	$75,$73,$E8,$FC,$3B,$6B,$A3,$FD,$1D,$FE,$D3,$E4,$FF,$47,$F7,$B1
	.dc.b	$87,$3C,$7C,$35,$DC,$91,$9B,$E9,$8F,$82,$7C,$78,$F8,$1A,$78,$0B
	.dc.b	$1B,$79,$C6,$DA,$6C,$F9,$7F,$11,$FE,$35,$23,$FE,$35,$F1,$7F,$C7
	.dc.b	$1F,$03,$FE,$39,$F8,$1F,$F1,$D7,$B5,$FE,$3B,$F1,$3F,$C7,$9E,$DF
	.dc.b	$F8,$F7,$7F,$F1,$F7,$B9,$FE,$3F,$F8,$1F,$F2,$07,$BD,$FE,$41,$F8
	.dc.b	$0F,$F2,$17,$AF,$FE,$43,$F8,$1F,$F2,$27,$C0,$FC,$4E,$2F,$E3,$F3
	.dc.b	$FE,$15,$F2,$71,$C5,$F5,$EB,$83,$FE,$49,$9E,$8F,$F9,$2F,$1F,$03
	.dc.b	$FE,$4D,$CF,$C8,$FF,$93,$F5,$F4,$3F,$E5,$1D,$FD,$8F,$F9,$4F,$87
	.dc.b	$FC,$AD,$C5,$F0,$A3,$F9,$5F,$1F,$FA,$47,$9C,$E7,$9F,$ED,$3C,$27
	.dc.b	$91,$FE,$5D,$E5,$F8,$EF,$F9,$7F,$97,$E3,$BF,$E6,$1E,$5F,$8E,$FF
	.dc.b	$98,$F9,$7E,$3B,$FE,$65,$E5,$F8,$EF,$F9,$9F,$97,$E3,$BF,$B6,$3F
	.dc.b	$8E,$F2,$FC,$8E,$5F,$E3,$B1,$FE,$8F,$F0,$B6,$FF,$44,$AE,$6F,$F4
	.dc.b	$6B,$EE,$FF,$48,$C3,$C6,$FF,$4A,$C7,$CE,$FF,$4C,$CB,$D6,$FF,$4E
	.dc.b	$CF,$F0,$AF,$E1,$CF,$F3,$E6,$9F,$27,$D2,$FE,$D7,$1F,$8F,$DB,$F3
	.dc.b	$8F,$E5,$B7,$FD,$0C,$62,$3F,$2B,$D1,$F9,$98,$C1,$38,$3F,$34,$D9
	.dc.b	$E9,$18,$23,$82,$4F,$38,$FE,$D7,$FC,$03,$FE,$23,$9F,$F1,$DE,$96
	.dc.b	$FF,$7D,$D6,$8C,$5F,$0A,$FD,$9E,$3E,$88,$EF,$8F,$59,$EF,$39,$6F
	.dc.b	$5F,$02,$3F,$05,$AF,$F1,$C7,$C8,$F9,$67,$F7,$4D,$9F,$E0,$ED,$9D
	.dc.b	$BF,$BD,$30,$62,$78,$B9,$21,$FC,$04,$D9,$E0,$76,$E4,$DB,$82,$4F
	.dc.b	$8A,$3E,$21,$F2,$88,$7F,$A2,$32,$F5,$E5,$F7,$63,$B9,$0E,$B3,$CD
	.dc.b	$F0,$12,$BB,$DF,$BA,$31,$3B,$9C,$C7,$BE,$30,$49,$FB,$DE,$A7,$F9
	.dc.b	$D6,$B7,$5E,$FA,$D3,$C6,$BB,$4E,$A3,$CC,$78,$23,$33,$C2,$39,$46
	.dc.b	$08,$F3,$21,$F3,$7C,$5F,$ED,$79,$7E,$DE,$3F,$72,$7F,$B7,$79,$7F
	.dc.b	$8F,$E1,$E1,$8F,$8B,$7D,$1E,$E7,$43,$CB,$1B,$7C,$C6,$4E,$27,$8B
	.dc.b	$1F,$85,$8F,$A6,$38,$B5,$D9,$D8,$86,$AB,$C4,$D3,$B1,$E0,$3D,$DC
	.dc.b	$BF,$C4,$AF,$FD,$ED,$1F,$C5,$CF,$EB,$C7,$13,$A9,$E0,$30,$43,$04
	.dc.b	$3B,$5F,$AA,$BC,$71,$99,$F8,$44,$BF,$64,$79,$4E,$85,$3F,$3B,$A7
	.dc.b	$E5,$B2,$FF,$37,$ED,$FD,$DE,$7E,$AD,$AB,$51,$C4,$91,$DF,$0F,$1C
	.dc.b	$63,$0F,$0C,$E6,$3E,$96,$5F,$C2,$27,$F8,$34,$FF,$09,$23,$82,$57
	.dc.b	$F0,$6D,$7F,$90,$4E,$A3,$19,$73,$7F,$C0,$F8,$F9,$A5,$97,$AF,$2F
	.dc.b	$DA,$1C,$0F,$E2,$A7,$F1,$32,$1F,$CC,$B1,$F0,$B8,$C7,$C3,$18,$8F
	.dc.b	$E3,$06,$71,$F1,$34,$EA,$73,$C7,$C1,$87,$A6,$34,$FF,$B0,$B2,$FC
	.dc.b	$47,$63,$5F,$D1,$E3,$DF,$2F,$0E,$3C,$A3,$19,$78,$32,$EC,$74,$8F
	.dc.b	$37,$07,$F5,$2C,$FC,$13,$A2,$61,$D1,$E1,$87,$52,$3F,$99,$8F,$F4
	.dc.b	$53,$F8,$4C,$BC,$CE,$71,$F1,$4F,$38,$EB,$18,$23,$F1,$23,$46,$F0
	.dc.b	$F3,$6A,$F8,$3F,$EA,$34,$CB,$F2,$51,$86,$C7,$E5,$A5,$EE,$65,$7B
	.dc.b	$B7,$6C,$7B,$D9,$F8,$9D,$7C,$06,$FE,$27,$87,$C1,$71,$F8,$9E,$5F
	.dc.b	$09,$CF,$E2,$7A,$7C,$37,$5F,$89,$ED,$F1,$1D,$F8,$7F,$DF,$9C,$1F
	.dc.b	$91,$11,$B8,$F3,$57,$37,$D1,$87,$77,$D5,$97,$8D,$F6,$69,$E7,$7D
	.dc.b	$DB,$7A,$DF,$81,$C1,$EF,$E1,$F8,$1C,$72,$43,$93,$E6,$FC,$4F,$C5
	.dc.b	$24,$7C,$9F,$57,$E4,$7C,$47,$C9,$F4,$3E,$57,$C4,$7C,$BF,$4B,$E8
	.dc.b	$7F,$00,$FF,$A2,$B0,$76,$7F,$A1,$61,$EA,$6B,$C4,$DF,$AB,$C3,$E1
	.dc.b	$F8,$FA,$FC,$BE,$23,$9F,$B3,$D3,$E2,$7E,$56,$9F,$DA,$B4,$FF,$09
	.dc.b	$F9,$98,$8E,$FE,$37,$B3,$C6,$6E,$C3,$9D,$7F,$18,$86,$5E,$08,$C6
	.dc.b	$9E,$47,$5B,$7B,$37,$F6,$E5,$EF,$9F,$C6,$70,$FC,$37,$F1,$DC,$7F
	.dc.b	$8D,$E4,$F9,$31,$FD,$FF,$97,$FC,$59,$CD,$F3,$A3,$F2,$91,$AE,$1F
	.dc.b	$F5,$57,$81,$FE,$BB,$6E,$1D,$6B,$FC,$11,$8F,$85,$F7,$9F,$A1,$DF
	.dc.b	$E8,$57,$CF,$D3,$E7,$4F,$BF,$97,$E7,$3F,$07,$97,$E7,$3F,$0F,$97
	.dc.b	$E7,$3F,$A7,$3F,$87,$F8,$DF,$37,$0F,$B5,$C9,$F2,$A7,$C3,$CB,$F3
	.dc.b	$B9,$7E,$54,$63,$47,$2F,$CB,$8C,$72,$FC,$B7,$E9,$72,$FA,$F2,$F8
	.dc.b	$3C,$1E,$F7,$E0,$CF,$E4,$4F,$8E,$43,$F3,$1E,$7F,$F9,$63,$93,$E5
	.dc.b	$5D,$9E,$07,$F8,$BF,$8F,$A4,$7F,$97,$B0,$F0,$3F,$87,$CF,$E4,$B7
	.dc.b	$8F,$E4,$DF,$AF,$8F,$E4,$BE,$DE,$3F,$92,$FB,$F8,$FE,$4B,$FD,$28
	.dc.b	$FE,$45,$C7,$FD,$FF,$93,$FA,$86,$5F,$93,$9F,$A0,$7E,$4F,$8B,$E4
	.dc.b	$F0,$7E,$9F,$8B,$E4,$3F,$A9,$E2,$F9,$15,$F3,$76,$F4,$BF,$C8,$71
	.dc.b	$6F,$EE,$8C,$70,$7E,$4D,$F9,$18,$7E,$3B,$87,$F2,$7C,$5F,$07,$DB
	.dc.b	$FD,$C9,$2B,$F4,$DE,$67,$EF,$3E,$13,$F7,$BF,$09,$FB,$EF,$7B,$F7
	.dc.b	$FF,$0B,$FC,$03,$E0,$3F,$80,$FC,$37,$F0,$2F,$82,$FE,$07,$F0,$FF
	.dc.b	$C1,$3E,$13,$F8,$2F,$AF,$FC,$D7,$C4,$7C,$2D,$F8,$8C,$7C,$3E,$7E
	.dc.b	$23,$F9,$D7,$B7,$FD,$D7,$9D,$D7,$FB,$DC,$55,$C3,$97,$FC,$27,$0C
	.dc.b	$7B,$65,$E0,$9F,$E0,$1A,$79,$1F,$E0,$5B,$7A,$1F,$E0,$9C,$1E,$C7
	.dc.b	$F8,$37,$17,$C0,$7F,$84,$72,$7F,$1E,$E5,$F8,$D1,$F9,$FE,$6F,$96
	.dc.b	$FD,$3E,$8F,$C9,$B4,$FC,$9F,$07,$E7,$1F,$F4,$17,$95,$FC,$A9,$FE
	.dc.b	$AD,$BF,$F0,$E7,$F8,$B7,$8B,$F9,$2A,$FE,$55,$E2,$FE,$49,$FE,$59
	.dc.b	$E2,$FE,$49,$FE,$5D,$E2,$FE,$49,$FE,$17,$7F,$D4,$0C,$72,$FE,$7E
	.dc.b	$78,$3F,$A1,$AE,$8F,$E8,$EF,$C0,$FE,$8E,$BF,$6E,$79,$8F,$E4,$27
	.dc.b	$F6,$DC,$BF,$A2,$BE,$E8,$FD,$93,$FC,$84,$FE,$9F,$E4,$7F,$21,$70
	.dc.b	$4C,$7F,$90,$F9,$3C,$A3,$E2,$9D,$8F,$F5,$46,$9F,$15,$FE,$1D,$A7
	.dc.b	$99,$FE,$45,$97,$A9,$FE,$49,$9F,$B9,$FE,$4D,$A7,$C2,$7F,$50,$7F
	.dc.b	$1E,$3F,$90,$1F,$D1,$7C,$9F,$E3,$D9,$7F,$59,$F1,$BF,$6B,$AF,$F0
	.dc.b	$7F,$1B,$F6,$AF,$F0,$9F,$1B,$F6,$B5,$FD,$7E,$3F,$6B,$E4,$FF,$99
	.dc.b	$7C,$6F,$DA,$3F,$E6,$7F,$1B,$F6,$99,$FC,$1A,$F8,$FC,$5F,$17,$0F
	.dc.b	$DB,$4F,$EE,$27,$FD,$EB,$C5,$F1,$67,$FA,$3F,$1F,$E2,$6B,$E2,$9F
	.dc.b	$58,$F8,$A7,$F5,$A3,$FA,$11,$FD,$78,$FE,$84,$7F,$11,$3F,$89,$9F
	.dc.b	$CE,$76,$FE,$81,$7F,$EB,$38,$FE,$7F,$EC,$FF,$8F,$1F,$E3,$C6,$CE
	.dc.b	$37,$F1,$DE,$A7,$F4,$3D,$BE,$3B,$FD,$5F,$6F,$8F,$7F,$CC,$4F,$8F
	.dc.b	$E4,$7F,$1A,$9F,$C6,$1F,$DB,$0F,$C6,$1F,$AB,$3F,$13,$A7,$F9,$67
	.dc.b	$C9,$74,$7E,$D3,$2F,$F0,$83,$FB,$7A,$FD,$C9,$FE,$01,$F2,$3F,$8D
	.dc.b	$3F,$E0,$3F,$23,$F8,$D3,$FE,$05,$F2,$3F,$8D,$3F,$E0,$7F,$23,$F8
	.dc.b	$D3,$FE,$09,$F2,$3F,$86,$BF,$C4,$3C,$5F,$E2,$4F,$F6,$4E,$DF,$E2
	.dc.b	$AF,$F8,$5F,$7F,$F5,$4F,$97,$FC,$76,$3F,$8B,$F9,$3F,$C8,$5F,$E3
	.dc.b	$3E,$3F,$F1,$77,$FB,$57,$8F,$FC,$5D,$FE,$D9,$E3,$FF,$17,$7F,$B7
	.dc.b	$78,$FF,$C5,$DF,$EE,$1E,$3F,$F1,$77,$FB,$97,$8F,$F0,$5F,$E2,$C7
	.dc.b	$F8,$87,$8F,$F8,$BF,$7B,$67,$C4,$FF,$35,$25,$7F,$2C,$F6,$FF,$9C
	.dc.b	$7B,$BF,$CE,$7E,$0B,$F9,$D7,$C1,$FF,$3B,$F8,$0F,$E7,$9F,$09,$FC
	.dc.b	$F7,$D1,$FE,$7D,$F0,$BF,$CF,$F9,$BF,$A0,$7C,$37,$F4,$1F,$84,$FE
	.dc.b	$85,$F0,$FF,$D0,$FE,$1B,$FA,$27,$91,$F1,$78,$3F,$C1,$BE,$B7,$F4
	.dc.b	$1E,$DF,$E8,$1E,$B6,$A7,$F9,$FD,$F2,$7F,$A0,$E3,$D9,$FE,$87,$9F
	.dc.b	$89,$FE,$8B,$AF,$99,$FE,$8F,$BF,$A9,$FE,$93,$C3,$EE,$7E,$01,$FD
	.dc.b	$33,$93,$E1,$BF,$E1,$03,$FD,$21,$CA,$F8,$7F,$EB,$84,$8F,$EA,$7F
	.dc.b	$11,$FE,$93,$FE,$01,$FF,$68,$1F,$E9,$98,$FF,$74,$BF,$E9,$46,$FB
	.dc.b	$BB,$F6,$F8,$7B,$DC,$7E,$23,$97,$C0,$73,$F8,$8E,$9F,$05,$D7,$E2
	.dc.b	$3B,$7C,$27,$7F,$88,$F9,$5E,$97,$FA,$D3,$5F,$DC,$E9,$8E,$AF,$F3
	.dc.b	$36,$EB,$E3,$5E,$4C,$7A,$BE,$6C,$FC,$2F,$A3,$5F,$2B,$EA,$C7,$DD
	.dc.b	$BF,$BF,$E1,$7E,$07,$C2,$FC,$1A,$FF,$00,$71,$7C,$57,$F4,$DC,$7F
	.dc.b	$92,$E4,$F9,$51,$FD,$EA,$35,$C1,$FE,$00,$8F,$9F,$5C,$DF,$93,$7D
	.dc.b	$DF,$92,$78,$CF,$9B,$C9,$F2,$5F,$9F,$C9,$F2,$5F,$A3,$C9,$F2,$5F
	.dc.b	$A7,$C9,$F2,$5F,$F0,$5F,$67,$F4,$57,$F7,$8E,$5F,$F3,$0F,$22,$79
	.dc.b	$46,$2B,$B3,$AB,$F1,$3C,$30,$FB,$CF,$A0,$E7,$B7,$C8,$7F,$0A,$7C
	.dc.b	$83,$E2,$10,$FC,$97,$8D,$FB,$9E,$7F,$CA,$ED,$F1,$E3,$13,$F9,$A3
	.dc.b	$F8,$A7,$8B,$FB,$97,$FC,$4F,$E2,$FE,$E5,$FF,$14,$F8,$BF,$B9,$7F
	.dc.b	$C5,$7E,$2F,$EE,$5F,$F1,$6F,$8B,$FB,$97,$FC,$5F,$E2,$FE,$E5,$FF
	.dc.b	$18,$F8,$BF,$E3,$8F,$0A,$47,$F2,$13,$FD,$ED,$E2,$A3,$F7,$51,$FC
	.dc.b	$9F,$C5,$FD,$D3,$FC,$A7,$C5,$FD,$D3,$FC,$AE,$7F,$74,$FF,$2D,$F1
	.dc.b	$7F,$74,$FF,$2F,$F1,$7F,$74,$FF,$31,$F1,$7F,$74,$FF,$33,$F1,$BE
	.dc.b	$2F,$8D,$FE,$49,$E5,$FC,$81,$89,$FD,$DB,$FE,$55,$9E,$3C,$BF,$2F
	.dc.b	$C0,$61,$DA,$31,$8F,$89,$D6,$5E,$67,$86,$7E,$A7,$96,$9E,$E7,$A6
	.dc.b	$BF,$09,$ED,$DB,$F2,$35,$FE,$43,$7F,$B7,$7F,$87,$F0,$7C,$F7,$FA
	.dc.b	$77,$0F,$D3,$7F,$A8,$71,$7D,$77,$FA,$97,$1F,$DB,$7F,$AA,$72,$7D
	.dc.b	$F7,$FA,$B7,$2F,$E0,$DF,$C2,$9F,$87,$CF,$E0,$71,$FF,$20,$F5,$9C
	.dc.b	$1F,$E7,$1F,$07,$F6,$BC,$D7,$0F,$E3,$E7,$E6,$0F,$DA,$9F,$C7,$BA
	.dc.b	$BF,$82,$BF,$B5,$3F,$44,$7F,$0B,$3F,$6A,$7F,$0F,$3F,$6A,$7F,$13
	.dc.b	$3F,$6A,$7F,$28,$F1,$FE,$27,$5B,$87,$FC,$29,$A7,$F1,$F3,$E5,$D6
	.dc.b	$CE,$27,$EE,$3C,$7F,$E3,$AF,$EE,$BC,$7F,$E3,$AF,$EF,$3C,$7F,$E3
	.dc.b	$AF,$EF,$BC,$7F,$4B,$FB,$63,$DA,$7F,$02,$3F,$6C,$7F,$1C,$3F,$6D
	.dc.b	$BF,$C8,$7F,$A3,$ED,$F2,$1A,$7C,$83,$91,$D3,$1F,$DD,$78,$9F,$D1
	.dc.b	$63,$F8,$7F,$8F,$FC,$21,$F9,$27,$9C,$F9,$27,$AC,$F9,$27,$BC,$F8
	.dc.b	$33,$D3,$6F,$93,$AF,$C8,$E0,$FF,$1F,$4F,$C8,$AF,$E3,$B5,$FD,$11
	.dc.b	$E4,$7F,$1D,$3E,$B1,$FC,$74,$FB,$47,$92,$3D,$39,$FC,$87,$D7,$C1
	.dc.b	$FC,$76,$7D,$E7,$F8,$9F,$C7,$F8,$9C,$3F,$E1,$3D,$7F,$6B,$3F,$DC
	.dc.b	$48,$7E,$EC,$FF,$5F,$F8,$5F,$99,$C3,$F6,$DC,$BF,$9D,$7F,$71,$CD
	.dc.b	$FA,$17,$F7,$5C,$FF,$A5,$7F,$79,$D1,$FA,$97,$F7,$DD,$3F,$AD,$7F
	.dc.b	$80,$70,$FE,$CD,$FE,$59,$E1,$7F,$84,$F3,$FE,$37,$D7,$FB,$9E,$3F
	.dc.b	$F6,$A7,$6F,$EF,$5F,$F6,$B7,$77,$F0,$17,$FD,$B1,$3F,$C1,$5F,$F6
	.dc.b	$D7,$67,$F0,$97,$FD,$B9,$DF,$FC,$35,$FF,$6F,$78,$1F,$C4,$9F,$F7
	.dc.b	$07,$85,$FC,$5F,$97,$F9,$EE,$7F,$D0,$BA,$DF,$0D,$FE,$0F,$24,$7F
	.dc.b	$83,$9A,$FF,$58,$7F,$C2,$9F,$11,$ED,$C3,$FC,$AB,$3F,$CD,$ED,$BF
	.dc.b	$BD,$BA,$EE,$7E,$09,$B9,$D1,$23,$53,$C2,$75,$18,$9D,$46,$0F,$F1
	.dc.b	$1D,$F9,$B8,$3F,$43,$8D,$C7,$FC,$2E,$FA,$BB,$BF,$59,$93,$37,$C4
	.dc.b	$7F,$8B,$C9,$97,$F8,$BB,$E1,$FF,$C6,$5F,$01,$FE,$33,$F8,$1F,$83
	.dc.b	$3F,$E3,$2F,$81,$FF,$1B,$7C,$17,$F8,$DF,$CA,$4C,$DF,$FB,$67,$F0
	.dc.b	$0F,$F8,$4E,$11,$8F,$88,$DC,$63,$87,$8E,$EB,$7C,$16,$B0,$E7,$F1
	.dc.b	$16,$F7,$4F,$84,$C6,$2C,$79,$7C,$46,$1E,$9F,$0F,$FB,$AF,$2B,$1E
	.dc.b	$0F,$EE,$EF,$FC,$AB,$E8,$EF,$25,$DC,$71,$96,$74,$D6,$DB,$DF,$86
	.dc.b	$DF,$0F,$3F,$F2,$BF,$0F,$F9,$6F,$8B,$E4,$C7,$F8,$8B,$8F,$E6,$BF
	.dc.b	$3C,$FF,$2E,$F3,$7D,$27,$F2,$CD,$E3,$FE,$3A,$FF,$99,$79,$B8,$CF
	.dc.b	$33,$F0,$67,$D5,$3C,$27,$F9,$A3,$97,$CA,$FD,$5C,$BE,$4E,$7E,$A8
	.dc.b	$C1,$F2,$78,$5B,$7C,$08,$CF,$2F,$CC,$6C,$FC,$67,$FC,$82,$6B,$93
	.dc.b	$E6,$4F,$F4,$FE,$5F,$EA,$35,$BA,$FC,$01,$DF,$8F,$B3,$8A,$FC,$16
	.dc.b	$1E,$58,$C5,$31,$F5,$C7,$F9,$33,$3F,$C9,$BF,$04,$FF,$27,$F5,$7E
	.dc.b	$65,$FF,$44,$F3,$FC,$97,$E7,$90,$FE,$09,$D5,$FA,$97,$F1,$5D,$3F
	.dc.b	$C1,$9F,$E0,$84,$3F,$82,$1F,$B3,$3F,$C2,$25,$E2,$FE,$43,$D9,$FB
	.dc.b	$92,$46,$78,$7F,$84,$BC,$09,$3C,$1D,$1F,$C2,$0C,$C7,$08,$D4,$60
	.dc.b	$D3,$82,$47,$F0,$28,$F4,$47,$F0,$66,$C6,$5F,$82,$60,$F8,$37,$F0
	.dc.b	$F6,$FE,$10,$DE,$CF,$E2,$73,$E1,$9F,$E3,$0D,$8C,$DF,$8A,$3C,$B7
	.dc.b	$E9,$23,$82,$57,$29,$E0,$D9,$DB,$67,$75,$DA,$72,$FF,$99,$E3,$D5
	.dc.b	$3E,$86,$E3,$F6,$DF,$B2,$7C,$02,$DF,$C5,$6D,$FC,$58,$C1,$FE,$0E
	.dc.b	$CF,$F7,$93,$FA,$E3,$F2,$47,$E9,$27,$F8,$23,$FA,$6E,$8F,$E1,$0F
	.dc.b	$EA,$3A,$7F,$86,$3F,$AA,$EA,$FE,$28,$FE,$A7,$93,$F4,$F1,$A8,$FD
	.dc.b	$0C,$F2,$7F,$47,$5D,$9F,$D2,$DF,$F1,$17,$FD,$5F,$D5,$E6,$7F,$46
	.dc.b	$7C,$2D,$BE,$E6,$9F,$0F,$4F,$81,$D9,$FD,$1A,$31,$87,$C1,$25,$FE
	.dc.b	$10,$E4,$7F,$10,$3F,$D7,$BD,$7F,$B2,$7E,$19,$E4,$3F,$83,$F2,$7E
	.dc.b	$EF,$0F,$E7,$71,$F0,$CF,$E1,$7C,$FF,$CC,$76,$F8,$71,$FE,$F2,$3F
	.dc.b	$86,$9F,$C0,$9F,$89,$3F,$D3,$B6,$F8,$8F,$F5,$0D,$BE,$23,$FD,$4B
	.dc.b	$2F,$88,$FF,$91,$B6,$F8,$8F,$DB,$C3,$D6,$7E,$00,$F7,$1F,$86,$D7
	.dc.b	$FB,$4F,$33,$E2,$3F,$A3,$12,$7F,$91,$FC,$47,$F4,$9F,$67,$FA,$57
	.dc.b	$C4,$FF,$00,$9F,$D2,$74,$5C,$BF,$3B,$D1,$F9,$99,$FC,$DC,$E7,$5F
	.dc.b	$CD,$B6,$32,$7F,$04,$E8,$FD,$11,$2B,$FC,$21,$A3,$1F,$15,$79,$0F
	.dc.b	$E6,$FA,$79,$DF,$F7,$1D,$78,$6F,$96,$1D,$DB,$3D,$E7,$44,$9D,$36
	.dc.b	$35,$18,$23,$83,$FA,$4C,$7E,$D5,$FF,$79,$E0,$F6,$FF,$B8,$12,$3F
	.dc.b	$A6,$FC,$3F,$F7,$3F,$77,$FB,$A7,$C3,$FF,$6A,$DB,$FC,$FF,$D1,$F8
	.dc.b	$F7,$FD,$03,$DD,$FD,$11,$FF,$07,$61,$FB,$13,$FA,$BF,$7F,$F4,$E8
	.dc.b	$FF,$0A,$F4,$FE,$3E,$7F,$1B,$1F,$EA,$3E,$2F,$F3,$E9,$FE,$14,$B7
	.dc.b	$89,$BA,$FE,$A7,$7C,$EB,$FC,$27,$87,$E7,$5F,$ED,$B3,$FC,$0C,$FE
	.dc.b	$E4,$7F,$07,$C3,$D2,$FE,$A5,$FF,$13,$E5,$EF,$8F,$82,$7F,$6E,$D7
	.dc.b	$E2,$3F,$21,$B3,$FE,$2E,$7F,$C3,$1A,$F1,$9F,$EC,$F5,$F1,$EF,$FC
	.dc.b	$5B,$D8,$D7,$E4,$46,$35,$F9,$0F,$A2,$FF,$86,$9F,$20,$FD,$B1,$F0
	.dc.b	$0F,$E3,$07,$C8,$7E,$3B,$FE,$31,$F0,$7F,$C0,$33,$F2,$23,$67,$13
	.dc.b	$F8,$D9,$D4,$EE,$7F,$7E,$ED,$F1,$BF,$DC,$A3,$CE,$FF,$74,$BF,$63
	.dc.b	$FE,$17,$C7,$E0,$69,$FE,$50,$CF,$E2,$E5,$FE,$46,$D3,$FC,$9F,$D3
	.dc.b	$FE,$62,$E3,$FC,$C6,$3F,$B4,$3F,$2C,$52,$61,$FB,$6E,$4F,$CD,$C6
	.dc.b	$30,$FD,$AB,$C2,$74,$D8,$CB,$B3,$F7,$27,$F7,$FB,$F1,$57,$F9,$8A
	.dc.b	$3F,$CC,$7C,$BF,$DA,$9B,$87,$E9,$DB,$3B,$7F,$C1,$26,$0C,$4F,$17
	.dc.b	$24,$3F,$C1,$46,$CF,$03,$B7,$26,$DC,$12,$7C,$51,$96,$BC,$09,$1B
	.dc.b	$9F,$34,$62,$77,$39,$8F,$3C,$60,$93,$C2,$75,$3E,$69,$CB,$C2,$75
	.dc.b	$18,$8F,$F4,$3F,$3F,$96,$72,$48,$E0,$43,$A3,$96,$99,$3E,$31,$0E
	.dc.b	$86,$4A,$70,$21,$B3,$E0,$3D,$5C,$57,$B1,$B5,$E9,$6C,$ED,$EA,$60
	.dc.b	$F4,$75,$FF,$3D,$7A,$9E,$32,$9E,$23,$89,$0C,$9C,$CD,$18,$35,$5E
	.dc.b	$56,$D7,$95,$B5,$E5,$78,$12,$78,$3E,$2A,$F3,$46,$09,$5C,$E7,$83
	.dc.b	$63,$35,$D5,$B3,$ED,$6B,$E8,$6B,$EB,$21,$E6,$38,$4F,$53,$F0,$D5
	.dc.b	$E4,$7F,$80,$E5,$FB,$89,$FD,$B5,$7F,$60,$6C,$EB,$2F,$DD,$D6,$C8
	.dc.b	$E0,$93,$FB,$E8,$CB,$63,$35,$D6,$7F,$CF,$39,$7F,$02,$AF,$E1,$06
	.dc.b	$C9,$3B,$AE,$37,$E3,$9C,$B6,$75,$18,$C7,$F8,$BE,$1E,$29,$C9,$2B
	.dc.b	$FB,$96,$1E,$1B,$F7,$DF,$0C,$3B,$B6,$3A,$1E,$6B,$E5,$19,$6C,$66
	.dc.b	$FC,$87,$F7,$FC,$3D,$FE,$47,$FB,$E3,$DF,$FF,$77,$12,$BF,$D9,$6F
	.dc.b	$FB,$C7,$E1,$FF,$DE,$5F,$03,$FE,$F3,$F8,$7F,$F7,$A7,$C1,$FF,$BD
	.dc.b	$7E,$1F,$FD,$F9,$FC,$02,$F8,$6B,$F1,$18,$F8,$8C,$F8,$1A,$F8,$3D
	.dc.b	$FA,$3C,$3D,$BE,$3E,$9F,$2F,$6F,$9F,$AB,$D3,$DB,$BC,$9F,$EE,$0D
	.dc.b	$95,$C2,$31,$87,$47,$59,$78,$1E,$1A,$79,$1E,$5B,$7A,$5F,$F7,$4F
	.dc.b	$07,$B5,$FF,$75,$F1,$7C,$17,$E1,$90,$75,$B9,$90,$E9,$6B,$98,$FE
	.dc.b	$86,$D7,$F9,$FB,$5F,$E9,$ED,$79,$C7,$29,$FE,$65,$39,$9E,$CD,$9D
	.dc.b	$47,$69,$D9,$1C,$10,$F0,$B9,$69,$93,$C6,$E5,$F2,$ED,$F7,$A3,$11
	.dc.b	$E6,$AF,$11,$23,$31,$E6,$6D,$79,$A3,$B9,$0E,$E7,$84,$E9,$1C,$27
	.dc.b	$A4,$7C,$EE,$CF,$E3,$71,$88,$E4,$70,$39,$1E,$12,$D7,$85,$EE,$70
	.dc.b	$34,$57,$4F,$33,$11,$DD,$E2,$43,$FD,$C6,$E5,$E0,$64,$EA,$58,$EA
	.dc.b	$E0,$D9,$0C,$11,$D9,$23,$B4,$7A,$27,$E1,$D7,$C7,$24,$7C,$09,$DD
	.dc.b	$71,$9C,$CF,$B1,$AF,$96,$77,$D5,$FC,$1F,$B3,$C9,$7C,$6B,$A5,$F0
	.dc.b	$73,$7D,$89,$19,$8F,$54,$7E,$87,$0F,$3C,$72,$21,$E3,$29,$E3,$9E
	.dc.b	$55,$E1,$8D,$10,$FC,$A9,$FC,$16,$37,$5E,$28,$E6,$79,$A3,$24,$7F
	.dc.b	$49,$5F,$8C,$F6,$FF,$5E,$48,$FD,$CB,$5F,$DC,$B5,$FD,$CB,$5F,$D2
	.dc.b	$35,$FD,$BB,$6F,$F8,$2B,$6C,$D3,$1D,$BF,$CB,$1C,$57,$88,$87,$88
	.dc.b	$A7,$88,$A7,$88,$A7,$88,$B5,$E2,$6D,$78,$9E,$B7,$F0,$DC,$D7,$89
	.dc.b	$F0,$10,$F0,$14,$F0,$14,$F0,$16,$BC,$0D,$AF,$03,$67,$E0,$BF,$CF
	.dc.b	$DF,$03,$D8,$87,$62,$9D,$8B,$3D,$9B,$3D,$9B,$3D,$1F,$3D,$7B,$DF
	.dc.b	$11,$F0,$08,$60,$87,$23,$25,$32,$7F,$25,$BF,$B4,$E8,$F4,$10,$F4
	.dc.b	$16,$FD,$0D,$BF,$43,$6F,$D0,$F0,$24,$FC,$CB,$F5,$47,$8C,$87,$13
	.dc.b	$C4,$70,$AE,$AD,$AE,$B5,$E2,$8C,$1F,$86,$F6,$FD,$6F,$B8,$C9,$63
	.dc.b	$81,$2B,$AB,$6B,$AC,$60,$8F,$F6,$A9,$FF,$44,$9F,$D7,$BD,$9F,$EC
	.dc.b	$24,$7F,$B5,$B5,$FE,$D6,$D7,$FB,$5B,$5F,$EA,$CD,$7A,$1C,$88,$72
	.dc.b	$29,$C8,$A7,$23,$FB,$DF,$B7,$FE,$20,$8E,$97,$FE,$00,$6D,$9A,$4A
	.dc.b	$E6,$78,$8F,$01,$0F,$01,$4F,$01,$AC,$7F,$B0,$C7,$38,$D1,$CC,$E2
	.dc.b	$43,$89,$4D,$18,$21,$82,$57,$AD,$F3,$B5,$EE,$D7,$FC,$33,$7F,$0D
	.dc.b	$DD,$FC,$16,$D7,$9D,$F1,$57,$17,$C2,$D7,$C3,$3C,$5E,$ED,$78,$B9
	.dc.b	$BF,$87,$1D,$88,$70,$AF,$03,$DA,$79,$36,$32,$F8,$1B,$5F,$E3,$AB
	.dc.b	$F6,$4F,$33,$A9,$E2,$9D,$D7,$48,$E4,$68,$EC,$7E,$3A,$FC,$EF,$32
	.dc.b	$1C,$CE,$07,$23,$27,$22,$99,$3F,$C9,$3E,$DF,$99,$E4,$6B,$0F,$98
	.dc.b	$4A,$F3,$36,$FD,$24,$9F,$37,$85,$FE,$3D,$F2,$7F,$CF,$A4,$8F,$F3
	.dc.b	$93,$57,$86,$E0,$E1,$1A,$21,$A7,$2D,$3F,$CB,$73,$CE,$79,$46,$0E
	.dc.b	$2E,$30,$FF,$36,$CE,$EB,$9C,$E7,$14,$E6,$F9,$1D,$EF,$8B,$D2,$F8
	.dc.b	$91,$DD,$F3,$AE,$55,$C2,$79,$5F,$67,$91,$C2,$F9,$4E,$EF,$94,$EA
	.dc.b	$31,$3A,$AE,$53,$FE,$8C,$E0,$FF,$36,$7B,$9F,$EA,$C2,$47,$FA,$85
	.dc.b	$FF,$57,$FB,$BF,$C9,$F1,$F8,$FF,$11,$FC,$A6,$BF,$D2,$71,$85,$1B
	.dc.b	$C7,$FD,$2E,$FF,$A6,$3D,$FC,$4E,$87,$C0,$E2,$7F,$D3,$BE,$DD,$97
	.dc.b	$C2,$DF,$83,$C7,$C3,$67,$E1,$75,$F0,$FB,$F8,$7E,$1F,$11,$C7,$E2
	.dc.b	$39,$7C,$1F,$3F,$84,$BA,$7A,$38,$7C,$AA,$B8,$B9,$3C,$D8,$76,$71
	.dc.b	$BF,$FA,$CB,$3F,$1C,$60,$6B,$E9,$DB,$D5,$BF,$AF,$83,$D9,$C3,$ED
	.dc.b	$C7,$DD,$87,$CE,$3E,$81,$0F,$AC,$7F,$AC,$B9,$3E,$AB,$F8,$46,$F1
	.dc.b	$FE,$4F,$93,$9C,$F8,$EB,$BB,$E5,$C3,$C6,$F9,$F2,$F9,$8F,$A7,$93
	.dc.b	$E6,$3E,$BE,$4F,$98,$FB,$79,$3E,$63,$EF,$E4,$F9,$8D,$3E,$17,$17
	.dc.b	$CA,$7F,$61,$97,$FB,$63,$8B,$E5,$46,$07,$17,$CA,$71,$C5,$F2,$5F
	.dc.b	$99,$C5,$F2,$5F,$9D,$C5,$F2,$5F,$A0,$7C,$93,$E2,$1F,$27,$C3,$FF
	.dc.b	$72,$74,$7E,$8A,$3F,$50,$47,$04,$E8,$FD,$2F,$27,$EA,$A7,$F6,$3D
	.dc.b	$3F,$B4,$9C,$F3,$FE,$AE,$30,$49,$FD,$84,$E9,$EC,$DB,$ED,$3B,$9E
	.dc.b	$A7,$07,$17,$E1,$25,$F7,$9D,$10,$FC,$2D,$7F,$07,$7F,$DF,$B5,$FE
	.dc.b	$FA,$23,$FB,$73,$F8,$A9,$A3,$11,$C5,$FE,$4E,$64,$FD,$F3,$97,$89
	.dc.b	$A2,$18,$23,$B2,$5F,$F1,$08,$E8,$43,$8C,$6A,$7C,$F1,$E0,$9E,$0E
	.dc.b	$9A,$F6,$6C,$F1,$8F,$5C,$60,$93,$D2,$3C,$07,$F3,$6E,$DF,$F7,$EC
	.dc.b	$66,$3B,$35,$EC,$DA,$ED,$1E,$58,$C5,$76,$8F,$0C,$60,$8E,$08,$71
	.dc.b	$29,$E4,$9E,$AD,$9E,$8E,$C8,$66,$7A,$B8,$3F,$AF,$D7,$B1,$EA,$7B
	.dc.b	$0A,$7B,$0B,$5E,$07,$47,$33,$45,$74,$E0,$F1,$1C,$AB,$CE,$E0,$E0
	.dc.b	$E1,$EA,$64,$87,$A3,$AF,$2C,$3B,$16,$7C,$24,$9E,$91,$A2,$46,$6F
	.dc.b	$E2,$C7,$63,$33,$C2,$75,$1E,$99,$DB,$E3,$3F,$72,$FE,$D3,$0F,$D9
	.dc.b	$47,$EE,$09,$8F,$ED,$9D,$9F,$B5,$71,$5F,$C0,$23,$13,$CA,$73,$7F
	.dc.b	$C1,$49,$7D,$EB,$99,$DA,$B8,$11,$CD,$72,$9F,$F4,$E6,$5F,$C8,$2F
	.dc.b	$9D,$78,$0D,$DF,$1C,$3C,$91,$D4,$C9,$1C,$E1,$D6,$39,$46,$27,$95
	.dc.b	$76,$9E,$53,$97,$94,$66,$B9,$47,$C9,$BF,$36,$1E,$FB,$E0,$4B,$E1
	.dc.b	$87,$69,$D4,$62,$FA,$C6,$08,$F9,$A3,$55,$CE,$73,$3E,$69,$FF,$11
	.dc.b	$DB,$AF,$FC,$07,$F0,$FF,$E1,$12,$57,$F7,$D7,$FC,$29,$F0,$FF,$E1
	.dc.b	$5F,$87,$FF,$0B,$7C,$47,$F8,$5F,$DB,$FF,$0C,$7C,$37,$F8,$67,$DD
	.dc.b	$FF,$0E,$7F,$00,$FF,$93,$B2,$FE,$C6,$BD,$FC,$F2,$6B,$E0,$77,$F1
	.dc.b	$1C,$3E,$0F,$8F,$C4,$F2,$F8,$7E,$7F,$13,$D3,$E2,$6E,$9F,$E1,$4D
	.dc.b	$15,$C2,$31,$87,$47,$59,$78,$1E,$1A,$79,$BA,$7D,$11,$FE,$1B,$DF
	.dc.b	$D8,$FF,$87,$78,$7E,$03,$FE,$1F,$E3,$F8,$AF,$F7,$BE,$4F,$C8,$F2
	.dc.b	$FF,$90,$35,$75,$FC,$22,$57,$9D,$AF,$C4,$6F,$5F,$F9,$3A,$78,$CE
	.dc.b	$CF,$77,$C2,$7C,$17,$E0,$70,$D7,$E6,$B6,$B9,$37,$5F,$23,$67,$FC
	.dc.b	$AF,$68,$C7,$84,$8C,$4F,$62,$1D,$8B,$7D,$9B,$7D,$9F,$39,$E9,$3D
	.dc.b	$C5,$8F,$76,$3F,$02,$7E,$43,$F1,$3A,$BF,$29,$5E,$6F,$85,$FC,$59
	.dc.b	$1F,$C0,$3F,$8E,$F8,$8F,$C7,$FC,$3F,$E4,$3E,$1F,$F2,$3F,$0F,$F9
	.dc.b	$2F,$6B,$F2,$7C,$1F,$94,$F8,$7F,$CA,$FB,$9F,$96,$F8,$7F,$F3,$97
	.dc.b	$03,$DE,$BF,$0F,$8F,$80,$CF,$C3,$EB,$E0,$B7,$F0,$FF,$EA,$1E,$8F
	.dc.b	$F5,$AB,$9E,$4F,$F3,$3E,$4B,$E5,$18,$C7,$BE,$5E,$18,$FC,$D6,$9E
	.dc.b	$57,$E4,$9F,$A8,$DF,$D6,$FE,$7F,$87,$DE,$FE,$87,$8F,$E2,$3F,$AE
	.dc.b	$E4,$FC,$9B,$F2,$4F,$8E,$43,$47,$13,$F5,$5C,$9F,$25,$FD,$67,$27
	.dc.b	$C9,$7C,$7A,$FE,$57,$8F,$E4,$C7,$EC,$78,$FD,$6F,$B4,$F8,$3C,$7F
	.dc.b	$21,$F8,$7C,$7F,$21,$FF,$5D,$F1,$FC,$87,$27,$C8,$E2,$F8,$EF,$F9
	.dc.b	$AB,$17,$17,$68,$FE,$17,$8F,$89,$FE,$1B,$97,$DE,$7F,$82,$70,$7C
	.dc.b	$87,$F8,$8F,$07,$C8,$7F,$89,$F0,$7C,$8C,$7F,$61,$C1,$F2,$34,$FD
	.dc.b	$17,$37,$E5,$6F,$F4,$ED,$8C,$C7,$E6,$F9,$FF,$32,$D3,$67,$E9,$0F
	.dc.b	$D7,$F4,$FE,$BB,$A7,$F5,$B1,$FA,$B2,$63,$FA,$B6,$D7,$EA,$DA,$72
	.dc.b	$3A,$5F,$ED,$EF,$A1,$23,$37,$DA,$FA,$12,$FC,$F3,$FB,$42,$47,$ED
	.dc.b	$1F,$55,$FB,$9D,$74,$FE,$D2,$31,$3B,$9E,$35,$D2,$30,$47,$04,$8F
	.dc.b	$8A,$7B,$27,$D4,$FA,$30,$F5,$39,$C3,$D5,$5F,$0E,$FD,$53,$F0,$CD
	.dc.b	$57,$08,$CD,$F4,$9C,$9D,$4F,$6D,$FF,$2E,$7D,$27,$92,$7D,$0D,$BF
	.dc.b	$8F,$87,$C1,$AE,$15,$AB,$F8,$75,$ED,$6D,$F1,$23,$82,$53,$A7,$F8
	.dc.b	$04,$FF,$66,$E7,$FE,$08,$FE,$87,$A3,$F8,$9B,$FA,$C3,$F4,$C7,$EA
	.dc.b	$0F,$D3,$74,$FF,$13,$C8,$FF,$13,$8F,$D5,$72,$FE,$B2,$59,$A5,$9A
	.dc.b	$7F,$5F,$1F,$C6,$B0,$F1,$C7,$F5,$5C,$7C,$EF,$F5,$7C,$BF,$49,$9F
	.dc.b	$C4,$C7,$E1,$37,$0F,$82,$DC,$7F,$19,$A7,$CB,$DB,$89,$2B,$EA,$B5
	.dc.b	$FF,$04,$E1,$E0,$C3,$FB,$76,$5E,$47,$D5,$87,$A7,$3F,$8F,$7F,$C2
	.dc.b	$79,$FF,$89,$3F,$C2,$F9,$FF,$89,$3F,$E4,$1A,$FE,$24,$E4,$F9,$A7
	.dc.b	$13,$F8,$E1,$F4,$CF,$E4,$07,$F3,$83,$FC,$25,$CB,$FC,$45,$FC,$5C
	.dc.b	$7D,$EC,$FD,$5B,$7E,$1F,$6F,$C8,$6D,$FA,$B8,$7C,$27,$F5,$22,$61
	.dc.b	$FD,$29,$FE,$AB,$F0,$FF,$D5,$BE,$1B,$FA,$BF,$C3,$FF,$58,$F8,$7F
	.dc.b	$EB,$3D,$FF,$A1,$E2,$FD,$31,$31,$FD,$36,$5F,$A4,$D3,$F5,$71,$89
	.dc.b	$D6,$DF,$A7,$AD,$73,$F3,$AD,$12,$75,$5C,$2F,$9C,$62,$75,$7F,$DE
	.dc.b	$9F,$E7,$58,$FA,$1F,$EC,$B9,$FB,$1F,$71,$06,$DF,$17,$E1,$3F,$DF
	.dc.b	$51,$F9,$6B,$46,$30,$FC,$C3,$63,$33,$FF,$00,$67,$2D,$8C,$BF,$A1
	.dc.b	$3F,$12,$E1,$C4,$70,$74,$78,$67,$B1,$1E,$8D,$77,$1C,$5E,$73,$D5
	.dc.b	$EF,$3B,$72,$D7,$E9,$C3,$E1,$BF,$C2,$B3,$99,$F1,$1E,$02,$1E,$09
	.dc.b	$EC,$E8,$C4,$71,$AF,$24,$78,$8E,$27,$88,$A6,$4E,$27,$F0,$7F,$37
	.dc.b	$FC,$AB,$1C,$A3,$B1,$C4,$D9,$0C,$10,$E8,$53,$89,$82,$18,$23,$DC
	.dc.b	$95,$EE,$6D,$7E,$0F,$3F,$C9,$CB,$E1,$BE,$4D,$FB,$0E,$71,$C4,$95
	.dc.b	$F1,$5B,$1B,$BE,$AD,$8C,$BE,$EC,$3E,$1C,$62,$3D,$93,$C1,$F8,$47
	.dc.b	$B1,$C3,$A8,$E0,$E5,$AF,$9A,$BC,$87,$B7,$6F,$F2,$BE,$1E,$C9,$EA
	.dc.b	$6C,$F3,$4F,$06,$90,$DD,$F8,$9F,$BD,$87,$AA,$78,$46,$B0,$F1,$D6
	.dc.b	$E3,$B1,$99,$DB,$EB,$6B,$EB,$F8,$4F,$F3,$6E,$7E,$78,$D1,$E3,$29
	.dc.b	$CC,$C1,$F0,$E7,$9B,$F9,$BC,$3E,$7F,$C3,$7E,$FC,$91,$FB,$D6,$BF
	.dc.b	$A3,$6B,$FB,$26,$BF,$C0,$1A,$FF,$0D,$6E,$1F,$D8,$DB,$89,$A6,$3E
	.dc.b	$1B,$FD,$13,$EB,$7F,$A2,$A3,$37,$E3,$21,$E3,$29,$E3,$29,$E3,$29
	.dc.b	$E3,$2D,$F8,$DB,$7E,$37,$F6,$9C,$FC,$5E,$C4,$3B,$14,$D1,$82,$18
	.dc.b	$21,$E1,$38,$14,$D9,$4D,$CF,$95,$DB,$5F,$E4,$3A,$7F,$AB,$BE,$17
	.dc.b	$E3,$47,$AB,$0F,$8E,$F0,$C3,$D8,$D8,$CB,$E0,$BE,$8F,$82,$76,$FB
	.dc.b	$B3,$FC,$2B,$D4,$87,$32,$9C,$E3,$8B,$CE,$BB,$B4,$D9,$FD,$1B,$0F
	.dc.b	$53,$FE,$BE,$8D,$BF,$0C,$97,$E6,$6D,$76,$6C,$F0,$9D,$12,$74,$FC
	.dc.b	$BB,$F5,$D7,$A4,$87,$A6,$B9,$57,$02,$38,$24,$7E,$23,$3F,$CB,$BE
	.dc.b	$93,$DA,$79,$0B,$7E,$39,$E3,$1E,$03,$E2,$0B,$FE,$43,$F0,$BF,$DC
	.dc.b	$49,$1F,$DB,$DA,$FF,$55,$6B,$FD,$91,$AF,$F7,$26,$BF,$DF,$DB,$5F
	.dc.b	$E5,$56,$E0,$69,$AF,$E0,$1F,$EE,$F1,$FE,$5E,$AF,$F2,$EC,$7F,$97
	.dc.b	$9F,$F7,$AC,$F0,$7C,$C4,$3C,$C5,$3C,$C5,$3C,$C5,$3C,$C5,$AF,$33
	.dc.b	$6B,$CC,$FE,$D8,$D9,$D8,$87,$62,$9D,$8A,$70,$34,$43,$45,$34,$60
	.dc.b	$86,$09,$1E,$67,$AA,$AF,$93,$CD,$FE,$6C,$7E,$44,$FF,$BC,$EF,$E4
	.dc.b	$3C,$2F,$DA,$D8,$CB,$E3,$BE,$CD,$9D,$B6,$7F,$0B,$87,$E2,$A3,$E4
	.dc.b	$47,$27,$D3,$5C,$9B,$5C,$9B,$3B,$7D,$B1,$96,$C6,$67,$DD,$1F,$E4
	.dc.b	$19,$FC,$34,$7C,$13,$18,$74,$9F,$19,$0F,$71,$4F,$51,$4E,$07,$E9
	.dc.b	$75,$AE,$8E,$24,$30,$43,$64,$3D,$27,$E3,$71,$7B,$7E,$68,$FC,$EF
	.dc.b	$C3,$5C,$3E,$34,$70,$3B,$1C,$DC,$93,$1F,$8C,$DC,$BE,$34,$60,$FB
	.dc.b	$B9,$FF,$90,$31,$7B,$BD,$DB,$87,$FB,$2E,$BE,$5C,$62,$3F,$D9,$9A
	.dc.b	$7E,$2E,$B8,$12,$38,$46,$5A,$7E,$E2,$34,$F2,$30,$6B,$1F,$CC,$46
	.dc.b	$AF,$C3,$F0,$DE,$5F,$7E,$5E,$F1,$46,$23,$A6,$1F,$A1,$8C,$5F,$A9
	.dc.b	$B8,$FE,$8D,$B5,$FC,$31,$B5,$C6,$3B,$47,$C5,$C3,$C9,$5E,$D9,$EF
	.dc.b	$87,$C5,$C3,$E3,$E3,$A9,$F6,$9A,$71,$8F,$BA,$3F,$DF,$1D,$2E,$CF
	.dc.b	$DB,$7C,$37,$D3,$24,$FB,$DA,$F8,$3E,$07,$EC,$7C,$37,$D9,$F8,$3F
	.dc.b	$B5,$F0,$DF,$6F,$E1,$7E,$E7,$C3,$7D,$DF,$87,$F5,$63,$F2,$7D,$8F
	.dc.b	$C0,$7B,$3F,$81,$F8,$5F,$C1,$7B,$7F,$83,$F8,$5F,$C2,$7B,$BF,$85
	.dc.b	$F8,$5F,$CA,$7A,$5F,$CC,$F2,$F8,$33,$B7,$EA,$5F,$37,$EB,$63,$F8
	.dc.b	$DC,$BC,$31,$F2,$33,$F2,$BF,$6F,$5F,$4B,$F7,$77,$F6,$BF,$7F,$87
	.dc.b	$E0,$F3,$FE,$53,$8F,$FD,$81,$CB,$F2,$6B,$FA,$2C,$6B,$A7,$E5,$C7
	.dc.b	$C9,$21,$F2,$4F,$C5,$72,$FC,$97,$F1,$9C,$BF,$25,$FC,$77,$2F,$C9
	.dc.b	$7F,$21,$D5,$EB,$7D,$A7,$E4,$F9,$BE,$4B,$F9,$5E,$6F,$92,$FE,$5F
	.dc.b	$9B,$E4,$E5,$88,$FC,$4C,$F0,$7F,$17,$5F,$51,$FA,$C7,$80,$FD,$07
	.dc.b	$0F,$C8,$7F,$23,$C3,$F2,$1F,$C9,$F0,$FC,$87,$DC,$7E,$5B,$6F,$C4
	.dc.b	$EF,$F1,$FA,$3F,$25,$3F,$95,$3F,$59,$D3,$F9,$6E,$AF,$CE,$6B,$F9
	.dc.b	$FE,$3F,$CE,$D6,$88,$71,$23,$A7,$F4,$BD,$7F,$AC,$AF,$DB,$C7,$EB
	.dc.b	$08,$68,$B3,$FB,$76,$D7,$ED,$DA,$7E,$DC,$B8,$7E,$DD,$E4,$49,$F1
	.dc.b	$12,$7D,$44,$8F,$DA,$C6,$5B,$19,$9F,$8B,$8F,$C7,$BF,$87,$5A,$24
	.dc.b	$FC,$AB,$E4,$DB,$E5,$5C,$23,$15,$C2,$30,$4C,$3E,$31,$23,$DB,$1F
	.dc.b	$21,$F6,$57,$C3,$69,$F0,$CB,$5F,$0E,$FE,$54,$62,$B9,$47,$99,$F0
	.dc.b	$E3,$F7,$C9,$3F,$72,$7E,$D6,$3F,$06,$73,$3F,$7B,$0F,$6C,$62,$BA
	.dc.b	$36,$3F,$C2,$1C,$3F,$A7,$AF,$E4,$FC,$FF,$C1,$5F,$E5,$3D,$1F,$C2
	.dc.b	$5F,$E1,$67,$E9,$CF,$E8,$9D,$5F,$C3,$F1,$FE,$A7,$D5,$B7,$FA,$AD
	.dc.b	$73,$7F,$AB,$DF,$77,$FA,$4D,$FA,$F1,$F3,$B4,$FA,$59,$79,$DF,$E6
	.dc.b	$F9,$FB,$DD,$6B,$F6,$09,$AF,$C8,$8C,$B5,$F8,$87,$43,$F7,$07,$F4
	.dc.b	$6E,$AF,$DE,$3F,$BE,$3F,$93,$9F,$C1,$CF,$EB,$9A,$7F,$2F,$7E,$0E
	.dc.b	$3F,$C1,$F2,$FE,$E7,$CB,$FC,$1D,$FE,$EB,$CB,$D1,$FE,$EF,$87,$F2
	.dc.b	$E7,$E9,$DF,$9B,$4F,$76,$7E,$A7,$FB,$9E,$9F,$C0,$F7,$7B,$FF,$CF
	.dc.b	$89,$7F,$49,$AF,$F6,$E8,$FF,$44,$3F,$5B,$E0,$3F,$A2,$FC,$2F,$F4
	.dc.b	$6F,$82,$FE,$8F,$F0,$BF,$D2,$3E,$13,$FA,$4F,$C2,$FE,$77,$0F,$CA
	.dc.b	$74,$5D,$FF,$2E,$4E,$3F,$CF,$36,$7F,$3C,$DE,$0F,$CF,$4E,$89,$3A
	.dc.b	$4A,$1A,$B0,$F2,$4F,$E3,$B2,$F4,$3F,$90,$D3,$FA,$FE,$BF,$D8,$3C
	.dc.b	$3F,$83,$9F,$F9,$3B,$1F,$8F,$7F,$DE,$BC,$3E,$F9,$F6,$F4,$3C,$33
	.dc.b	$0F,$15,$23,$11,$C2,$30,$E2,$79,$36,$32,$F3,$8C,$B6,$32,$F8,$23
	.dc.b	$2D,$8C,$BD,$E7,$C2,$F9,$1B,$3C,$23,$47,$86,$33,$3C,$A3,$81,$92
	.dc.b	$1E,$38,$E7,$D1,$FC,$43,$6F,$AB,$7F,$4B,$C8,$FF,$05,$D7,$42,$5F
	.dc.b	$23,$D0,$E1,$E8,$64,$F2,$98,$21,$C0,$D1,$E6,$35,$9A,$78,$C6,$CE
	.dc.b	$26,$08,$71,$34,$43,$35,$F1,$DD,$18,$C7,$E3,$46,$48,$E0,$97,$ED
	.dc.b	$9F,$B1,$8F,$C7,$7F,$34,$7E,$22,$FE,$3B,$EC,$BE,$75,$C6,$3C,$0D
	.dc.b	$7C,$64,$8D,$BC,$6F,$C2,$FF,$B6,$E3,$2F,$C2,$38,$46,$9C,$46,$9F
	.dc.b	$84,$6A,$7C,$11,$98,$F1,$9D,$E3,$F0,$7C,$1F,$E6,$EA,$FA,$53,$FC
	.dc.b	$FA,$78,$C7,$78,$EB,$5C,$DE,$C7,$32,$1E,$48,$FD,$2E,$3D,$27,$81
	.dc.b	$23,$A4,$F9,$0E,$15,$E0,$23,$D0,$F0,$C7,$07,$F7,$3E,$DF,$F9,$8B
	.dc.b	$DE,$FF,$31,$FB,$FF,$52,$79,$9C,$4E,$C7,$2B,$EA,$E8,$98,$F9,$F0
	.dc.b	$7C,$2F,$EB,$89,$3F,$C0,$5A,$FE,$FD,$AF,$EF,$5A,$FF,$0B,$6B,$FC
	.dc.b	$41,$B9,$7F,$22,$6C,$F8,$27,$B1,$0E,$C7,$F1,$1C,$BF,$D0,$27,$FA
	.dc.b	$0B,$E0,$7F,$D0,$75,$C8,$93,$C4,$F2,$14,$F2,$14,$F2,$16,$FC,$8D
	.dc.b	$BF,$23,$FE,$F1,$36,$79,$08,$76,$29,$D8,$A7,$03,$44,$34,$53,$46
	.dc.b	$08,$60,$91,$E6,$7A,$BF,$D3,$2F,$E2,$C7,$F8,$63,$E1,$7E,$2B,$EB
	.dc.b	$BF,$8A,$F0,$BF,$6B,$E1,$BE,$8D,$8C,$BE,$28,$CB,$63,$E2,$E3,$F8
	.dc.b	$9A,$EB,$3C,$5A,$ED,$AF,$AA,$7A,$B6,$33,$5D,$A3,$58,$7B,$1F,$F0
	.dc.b	$06,$1E,$C7,$D5,$5C,$2F,$9D,$F1,$23,$82,$46,$CF,$CD,$E1,$E9,$AF
	.dc.b	$8C,$43,$A9,$4F,$61,$4F,$61,$4F,$51,$FD,$9F,$E0,$FF,$DA,$FE,$F7
	.dc.b	$E5,$E3,$FC,$97,$5E,$17,$86,$1E,$68,$C4,$78,$09,$7F,$B5,$F8,$1F
	.dc.b	$EC,$24,$8F,$EE,$8D,$7F,$B9,$35,$FE,$E0,$D7,$FB,$F3,$5F,$F0,$1B
	.dc.b	$6B,$FC,$2A,$DC,$4D,$31,$F0,$9F,$E0,$CE,$67,$2F,$F8,$43,$F8,$07
	.dc.b	$FC,$3B,$8F,$F8,$82,$72,$FF,$88,$63,$2F,$F8,$E6,$FA,$3E,$A2,$1E
	.dc.b	$A2,$9E,$A2,$9E,$A2,$9E,$A2,$E3,$EA,$79,$E5,$F2,$5C,$E5,$EC,$6D
	.dc.b	$78,$9F,$01,$0F,$01,$4F,$01,$4F,$01,$6B,$C0,$D9,$EC,$DC,$BE,$46
	.dc.b	$5F,$E2,$F6,$C7,$C5,$F8,$6B,$7C,$A3,$C2,$43,$C2,$53,$C2,$57,$83
	.dc.b	$A2,$1A,$2B,$A7,$07,$AB,$1F,$8C,$FA,$71,$F8,$8F,$5C,$7D,$6D,$8C
	.dc.b	$BD,$23,$2D,$8C,$BD,$AB,$C0,$7F,$98,$B0,$F7,$D1,$F8,$39,$31,$CE
	.dc.b	$1C,$AF,$8D,$F6,$C3,$A3,$63,$33,$B8,$C1,$1F,$D3,$63,$EE,$AF,$51
	.dc.b	$F8,$22,$1F,$04,$A6,$8C,$10,$C1,$FE,$5F,$C1,$F0,$9E,$AB,$93,$FE
	.dc.b	$CA,$BF,$21,$1E,$A7,$C7,$29,$EC,$7F,$CF,$37,$FE,$0B,$B7,$C2,$FF
	.dc.b	$99,$9B,$5F,$E7,$76,$BF,$E7,$56,$BE,$58,$E0,$47,$83,$5E,$0D,$78
	.dc.b	$1A,$21,$A2,$9A,$30,$43,$E2,$7C,$2F,$DC,$F8,$6F,$BB,$F0,$DF,$7A
	.dc.b	$BB,$57,$F0,$96,$BF,$C1,$DB,$A3,$0F,$5C,$EC,$F0,$94,$F0,$14,$F2
	.dc.b	$10,$D6,$3F,$2E,$BA,$CF,$49,$F0,$1D,$0A,$74,$29,$CC,$A7,$72,$18
	.dc.b	$25,$7A,$67,$B9,$23,$EC,$63,$FA,$0F,$86,$F8,$B1,$FC,$A7,$83,$9D
	.dc.b	$F1,$8F,$23,$5F,$23,$5F,$23,$5E,$38,$7B,$63,$11,$E4,$8C,$13,$0F
	.dc.b	$C7,$F6,$70,$7C,$D3,$DA,$74,$F8,$9A,$E6,$36,$F8,$A3,$2D,$8F,$CC
	.dc.b	$63,$EF,$8F,$71,$FD,$17,$3F,$C2,$3D,$67,$9B,$E3,$30,$43,$C0,$5C
	.dc.b	$FE,$E3,$93,$C0,$53,$91,$C0,$86,$8A,$68,$FD,$E7,$C2,$7E,$65,$F3
	.dc.b	$E5,$F9,$3D,$7C,$F1,$C0,$F9,$97,$E8,$69,$E1,$29,$E1,$BF,$DE,$7C
	.dc.b	$37,$F2,$02,$47,$6C,$D1,$88,$72,$7F,$6C,$C7,$F4,$93,$98,$C5,$70
	.dc.b	$AE,$53,$9D,$BF,$BB,$36,$3F,$BA,$36,$F9,$C7,$87,$E1,$BF,$99,$F8
	.dc.b	$32,$3C,$8D,$7C,$77,$FE,$01,$8C,$63,$E2,$6C,$66,$7A,$1E,$42,$1C
	.dc.b	$8F,$19,$89,$F5,$E1,$E5,$C7,$D9,$3B,$C7,$CE,$D8,$CD,$70,$8E,$BE
	.dc.b	$4B,$1F,$F0,$B7,$C1,$FF,$4E,$24,$FC,$F6,$BC,$DF,$EA,$7F,$09,$FD
	.dc.b	$53,$E1,$7F,$AA,$FC,$1F,$F5,$6F,$82,$FE,$AF,$F0,$7F,$D6,$3E,$17
	.dc.b	$FA,$CF,$C2,$FF,$5A,$F8,$3F,$EF,$7F,$C0,$2D,$3F,$AE,$F2,$63,$8B
	.dc.b	$3F,$0D,$AF,$86,$FE,$C3,$D5,$C3,$D6,$E3,$EE,$E7,$1F,$C8,$E4,$C3
	.dc.b	$94,$7C,$C3,$E2,$47,$77,$F1,$59,$F9,$1F,$C6,$6B,$E8,$7F,$1D,$BF
	.dc.b	$B1,$FC,$87,$0F,$C0,$7E,$21,$07,$2F,$CB,$7F,$C7,$9C,$9F,$23,$9B
	.dc.b	$F2,$2D,$E9,$F9,$F3,$C6,$7F,$2B,$CD,$F2,$E3,$F3,$DC,$DF,$2D,$FD
	.dc.b	$07,$37,$CB,$7F,$45,$CD,$F2,$DF,$82,$7C,$A3,$E1,$F2,$FC,$A7,$E2
	.dc.b	$F2,$FC,$A7,$E3,$F2,$7C,$A7,$27,$CA,$E3,$F9,$2F,$E2,$79,$BE,$BB
	.dc.b	$DC,$FC,$66,$1E,$37,$F1,$D8,$F9,$DF,$C8,$65,$EB,$7F,$25,$9F,$C1
	.dc.b	$7E,$29,$F8,$1D,$FE,$46,$9F,$98,$D7,$F2,$D3,$F9,$9E,$9F,$CF,$11
	.dc.b	$FD,$16,$FF,$B8,$69,$FA,$63,$81,$8A,$E9,$1B,$8F,$D2,$74,$7E,$C9
	.dc.b	$FF,$39,$36,$7F,$B6,$37,$A7,$EE,$4F,$03,$F7,$E4,$30,$78,$09,$87
	.dc.b	$A8,$97,$E1,$6C,$79,$DB,$7E,$68,$C4,$F9,$F1,$F8,$B7,$FB,$A8,$E4
	.dc.b	$49,$D9,$2B,$91,$27,$84,$65,$B1,$9C,$7E,$64,$FA,$2B,$F8,$A3,$EC
	.dc.b	$C3,$F8,$C3,$EF,$6B,$98,$DC,$60,$8E,$09,$1E,$19,$EF,$3F,$02,$FC
	.dc.b	$64,$8C,$D7,$C1,$C7,$DD,$87,$9F,$0F,$B9,$7B,$6D,$F0,$7B,$18,$37
	.dc.b	$B7,$E7,$31,$FE,$8D,$CF,$FC,$1E,$7F,$42,$7F,$27,$E9,$FE,$1C,$FF
	.dc.b	$0C,$E9,$FE,$1C,$FF,$0A,$E2,$E5,$AF,$FA,$F7,$1F,$06,$9E,$5A,$F9
	.dc.b	$47,$88,$AE,$72,$FC,$EE,$DF,$73,$4E,$3A,$FA,$EF,$6D,$AF,$8D,$3A
	.dc.b	$37,$87,$8A,$98,$35,$F6,$47,$F2,$0E,$4F,$E0,$0F,$F2,$2E,$4F,$E0
	.dc.b	$0E,$4F,$E0,$07,$F0,$83,$F9,$46,$3E,$07,$E5,$1E,$4C,$7C,$4E,$33
	.dc.b	$FD,$CE,$0F,$3F,$F9,$E9,$27,$F9,$AB,$5F,$E6,$6F,$F4,$2F,$7B,$FA
	.dc.b	$1F,$C3,$7F,$44,$F8,$6F,$E8,$BF,$03,$FD,$1B,$E1,$BF,$A3,$FC,$1F
	.dc.b	$F4,$8F,$86,$FE,$93,$F0,$BF,$D2,$BE,$03,$F6,$DD,$53,$5F,$CB,$73
	.dc.b	$FE,$43,$97,$F7,$71,$83,$F7,$E5,$72,$FF,$46,$AE,$AF,$F4,$CC,$3C
	.dc.b	$2F,$F4,$EC,$BC,$AF,$F5,$0D,$3D,$2F,$F5,$2D,$BD,$DC,$7F,$8E,$E0
	.dc.b	$FE,$A1,$C5,$F1,$BE,$13,$FA,$95,$FF,$53,$D3,$EC,$46,$23,$FB,$24
	.dc.b	$61,$C3,$5E,$2D,$79,$71,$A7,$94,$7E,$12,$30,$F2,$6B,$C9,$AF,$28
	.dc.b	$E3,$3D,$63,$C7,$3A,$6C,$EA,$31,$1D,$E3,$04,$7C,$0D,$7C,$1F,$05
	.dc.b	$FE,$02,$F8,$5F,$F0,$1F,$C2,$7F,$81,$3E,$17,$FC,$0B,$7B,$BF,$21
	.dc.b	$C8,$87,$A5,$D3,$E4,$32,$70,$34,$60,$D1,$81,$E4,$BC,$EC,$C7,$12
	.dc.b	$18,$21,$EF,$3A,$14,$E8,$70,$30,$5C,$FE,$63,$83,$9B,$82,$67,$F3
	.dc.b	$1C,$77,$7F,$52,$DF,$F8,$CE,$9F,$E2,$AF,$86,$F9,$18,$7F,$88,$B3
	.dc.b	$F8,$6F,$52,$61,$D2,$32,$D8,$F4,$B5,$F4,$9F,$1B,$3F,$88,$F2,$CB
	.dc.b	$C8,$4A,$E6,$D8,$E6,$D7,$99,$C6,$36,$47,$6F,$FB,$E7,$9F,$F0,$91
	.dc.b	$FE,$3C,$C3,$DA,$FF,$8E,$A7,$83,$D3,$0F,$43,$F3,$6F,$C0,$D9,$DB
	.dc.b	$F8,$3C,$3D,$75,$C2,$3E,$8D,$70,$6D,$70,$69,$E6,$21,$E1,$8E,$0F
	.dc.b	$C5,$DF,$FC,$9D,$9F,$FA,$4F,$F8,$05,$D1,$F9,$EA,$FF,$53,$C3,$E1
	.dc.b	$FF,$3F,$1D,$0F,$41,$4F,$09,$DF,$1F,$23,$F8,$FC,$FD,$75,$F9,$A8
	.dc.b	$FE,$0E,$D7,$F8,$3B,$5F,$E2,$0D,$7F,$87,$B5,$FE,$96,$DC,$CD,$24
	.dc.b	$FF,$06,$6C,$F8,$23,$04,$70,$43,$C6,$E5,$A7,$A3,$E0,$BE,$4E,$3E
	.dc.b	$4F,$83,$B3,$E4,$F8,$7B,$8F,$93,$E1,$6E,$5E,$2A,$E6,$43,$DF,$97
	.dc.b	$76,$9E,$32,$9B,$3E,$1B,$B6,$9E,$97,$E7,$99,$7E,$3B,$6B,$E3,$B6
	.dc.b	$7C,$93,$D8,$87,$43,$C0,$72,$3C,$86,$4D,$99,$37,$3D,$9F,$90,$7E
	.dc.b	$33,$E1,$FE,$1F,$C3,$FC,$4F,$84,$F8,$B8,$78,$C9,$5D,$8E,$87,$6B
	.dc.b	$E8,$F6,$C3,$D2,$F8,$CF,$44,$FC,$27,$94,$F0,$79,$4F,$07,$A9,$C4
	.dc.b	$A4,$8E,$0E,$5A,$E6,$39,$46,$08,$FE,$4F,$5F,$C0,$E3,$F0,$2B,$F5
	.dc.b	$B8,$7C,$07,$CB,$5C,$89,$87,$C2,$8F,$73,$5F,$C1,$BE,$D7,$27,$27
	.dc.b	$4F,$9E,$39,$B9,$21,$C0,$87,$EC,$39,$FF,$2F,$AF,$E6,$35,$FC,$CE
	.dc.b	$BE,$7A,$D9,$D7,$0E,$EF,$C3,$6E,$9F,$09,$FD,$51,$FC,$0B,$E0,$FF
	.dc.b	$68,$49,$FE,$06,$D7,$F8,$1B,$5F,$E0,$6D,$7F,$81,$B5,$FE,$06,$D9
	.dc.b	$FE,$06,$DC,$8D,$31,$F0,$DF,$B4,$F8,$6F,$DA,$FC,$17,$ED,$BE,$1B
	.dc.b	$F6,$F5,$BC,$7C,$A4,$3C,$A5,$3C,$A5,$3C,$A5,$3C,$A5,$C7,$CA,$8F
	.dc.b	$EF,$BB,$70,$7A,$90,$EA,$53,$46,$08,$60,$87,$80,$E0,$53,$65,$37
	.dc.b	$1E,$47,$6D,$7F,$99,$FC,$37,$F1,$AF,$86,$F8,$F1,$FD,$1B,$E1,$BE
	.dc.b	$3B,$ED,$D7,$E7,$B7,$1F,$94,$F4,$8C,$B6,$32,$FB,$B6,$FC,$33,$C4
	.dc.b	$86,$CA,$7A,$B1,$F3,$3E,$AA,$EA,$E5,$AE,$6B,$BB,$FD,$33,$1F,$63
	.dc.b	$FD,$83,$4F,$E6,$F8,$FB,$23,$C0,$43,$A5,$72,$7E,$51,$30,$F1,$B7
	.dc.b	$0F,$7E,$3E,$C8,$E4,$78,$CF,$59,$4C,$9C,$6B,$B5,$74,$25,$7B,$27
	.dc.b	$B6,$DF,$9B,$7F,$BC,$FC,$17,$E6,$DF,$EF,$77,$CA,$B8,$5F,$28,$C1
	.dc.b	$27,$E7,$E1,$FD,$E3,$E1,$BF,$AC,$92,$3F,$C0,$AD,$5D,$7F,$E0,$B8
	.dc.b	$C3,$63,$64,$37,$D3,$FD,$F2,$76,$D8,$D4,$F1,$76,$7F,$80,$27,$6D
	.dc.b	$3F,$55,$1F,$E1,$A7,$FD,$9F,$2F,$84,$FE,$E6,$BE,$1A,$92,$78,$C7
	.dc.b	$FA,$70,$D3,$83,$9C,$F8,$32,$FF,$07,$E1,$73,$34,$95,$EA,$BF,$63
	.dc.b	$EA,$AE,$2E,$4F,$50,$F1,$9E,$17,$8F,$E1,$BD,$24,$9F,$25,$7F,$8E
	.dc.b	$23,$11,$D6,$D1,$8D,$3E,$25,$7F,$8F,$BE,$0B,$E1,$FC,$2F,$C4,$F8
	.dc.b	$4F,$8B,$F0,$BF,$1B,$E1,$BE,$3F,$BB,$F2,$3D,$8F,$93,$ED,$FC,$AD
	.dc.b	$FE,$5F,$C1,$7C,$CF,$7F,$E6,$F2,$FC,$EF,$83,$F9,$FF,$03,$F4,$3E
	.dc.b	$13,$E8,$FC,$1E,$B9,$BD,$3C,$CC,$79,$C6,$33,$EE,$EB,$5F,$1B,$C3
	.dc.b	$7F,$3B,$CB,$87,$D6,$F4,$E3,$F8,$18,$FC,$98,$FE,$5D,$CF,$F3,$1F
	.dc.b	$C2,$F3,$E9,$A7,$91,$C7,$37,$CD,$8C,$1D,$4F,$4F,$37,$C9,$7D,$7C
	.dc.b	$DF,$25,$F6,$F3,$7C,$97,$DF,$CD,$F2,$5F,$83,$CD,$F2,$5F,$87,$CD
	.dc.b	$F2,$5F,$8B,$CB,$F2,$5F,$8F,$C9,$F2,$5F,$BD,$C9,$F2,$1F,$C0,$72
	.dc.b	$7C,$87,$F0,$5C,$9F,$21,$FC,$27,$27,$C8,$7F,$0D,$C9,$F2,$1F,$C4
	.dc.b	$72,$7C,$0E,$4F,$C4,$C7,$C8,$C3,$F3,$3C,$FF,$B3,$24,$E0,$93,$F9
	.dc.b	$A8,$CB,$63,$3C,$7F,$A0,$8F,$D1,$74,$FE,$A1,$FD,$14,$66,$B8,$B6
	.dc.b	$B8,$D6,$EB,$8F,$4F,$EE,$5F,$D0,$75,$5C,$3F,$3F,$7F,$BA,$9C,$9F
	.dc.b	$90,$9F,$DA,$BC,$8F,$0D,$F5,$AF,$0B,$6B,$C3,$7C,$27,$2F,$F7,$9A
	.dc.b	$FD,$C6,$3F,$BE,$7F,$D3,$31,$FE,$91,$23,$FE,$A0,$6B,$FE,$A0,$6B
	.dc.b	$C0,$D1,$0D,$14,$D1,$82,$18,$21,$FC,$6F,$0F,$89,$76,$3E,$11,$0F
	.dc.b	$8A,$7C,$33,$99,$EB,$21,$FC,$58,$EA,$53,$38,$FD,$D7,$2D,$73,$8F
	.dc.b	$D9,$EB,$FE,$99,$1E,$B3,$38,$7C,$26,$99,$3B,$12,$BA,$36,$BF,$08
	.dc.b	$F8,$1A,$F2,$3D,$67,$F3,$A3,$F8,$8F,$5F,$F1,$67,$F8,$9F,$5F,$F1
	.dc.b	$67,$F6,$67,$F5,$73,$A5,$F9,$5F,$03,$6F,$C6,$4B,$F0,$CE,$8F,$B5
	.dc.b	$87,$C0,$7C,$73,$BA,$E5,$5C,$23,$13,$F0,$5A,$F0,$8D,$12,$78,$9E
	.dc.b	$3C,$7F,$B0,$4F,$B8,$FE,$16,$7B,$8F,$E1,$67,$B8,$FC,$01,$EE,$3C
	.dc.b	$13,$D5,$B3,$D9,$A7,$B3,$C3,$F9,$CF,$13,$CD,$1B,$69,$C0,$A4,$38
	.dc.b	$1F,$AB,$B5,$FE,$E7,$5F,$DD,$E9,$F9,$AA,$FE,$FA,$DA,$DB,$FE,$13
	.dc.b	$8E,$67,$90,$87,$D3,$3C,$07,$E1,$72,$FA,$95,$FB,$C8,$FE,$07,$87
	.dc.b	$EF,$F0,$FE,$03,$18,$21,$FA,$79,$E0,$D8,$FE,$3E,$7A,$4F,$D8,$3F
	.dc.b	$AE,$8F,$5B,$FB,$16,$BC,$4F,$E5,$04,$C3,$F9,$13,$72,$FE,$28,$DB
	.dc.b	$F3,$34,$FF,$17,$76,$7F,$5F,$AE,$24,$AE,$2D,$AE,$2D,$AF,$6D,$FB
	.dc.b	$09,$87,$C2,$8F,$34,$FD,$6E,$6F,$EB,$11,$FE,$11,$E6,$FE,$B0,$FF
	.dc.b	$84,$F9,$BF,$AC,$61,$CA,$3C,$C7,$2C,$BC,$64,$CB,$C6,$D8,$F5,$F0
	.dc.b	$F8,$6B,$96,$5E,$79,$E5,$7C,$08,$FA,$B3,$E0,$7F,$8F,$73,$F3,$DF
	.dc.b	$F4,$6C,$FC,$EF,$F4,$8C,$FC,$F5,$C2,$FC,$F3,$E0,$CF,$C9,$1E,$03
	.dc.b	$B1,$FE,$3A,$D3,$E8,$53,$E1,$7F,$CA,$24,$9F,$F1,$0B,$5F,$F1,$C3
	.dc.b	$6F,$FA,$CF,$C2,$FF,$96,$3E,$1B,$FA,$74,$7F,$8E,$BF,$80,$7F,$A8
	.dc.b	$5F,$F9,$2D,$6F,$AC,$7F,$D9,$8A,$1F,$0D,$9A,$DC,$3E,$1B,$15,$BF
	.dc.b	$86,$B2,$F8,$7B,$EF,$63,$DF,$CF,$C3,$FE,$C7,$4F,$CF,$67,$FB,$43
	.dc.b	$F3,$8F,$E8,$35,$FC,$E9,$E9,$9E,$53,$FE,$5A,$C5,$87,$83,$1F,$0E
	.dc.b	$5E,$2D,$3C,$7A,$F9,$35,$F2,$E3,$F1,$4F,$F2,$5E,$FF,$27,$5F,$F3
	.dc.b	$D7,$07,$F7,$5E,$1F,$D1,$6A,$F8,$1F,$93,$87,$C6,$8C,$38,$C7,$E4
	.dc.b	$4E,$A3,$E3,$B6,$33,$1B,$9E,$2F,$D4,$8C,$B6,$32,$FF,$1D,$6B,$E1
	.dc.b	$9E,$91,$88,$EE,$75,$8C,$91,$C1,$0E,$CE,$5F,$16,$1F,$52,$BF,$05
	.dc.b	$F0,$FF,$73,$E0,$FE,$EF,$C3,$FD,$EF,$85,$FB,$F8,$F1,$C7,$CC,$D7
	.dc.b	$CC,$77,$32,$53,$27,$88,$C1,$EA,$30,$3B,$5D,$FC,$23,$99,$A2,$90
	.dc.b	$D1,$0E,$86,$4A,$67,$2F,$A2,$E5,$AE,$74,$F9,$D1,$88,$E2,$47,$FA
	.dc.b	$5B,$F9,$5C,$FE,$43,$F9,$7C,$FE,$43,$F9,$9C,$7E,$33,$C6,$FA,$92
	.dc.b	$33,$7D,$C9,$19,$9F,$8F,$1F,$88,$C7,$E2,$C7,$3A,$E3,$5B,$8F,$8C
	.dc.b	$7A,$E3,$24,$79,$C6,$DF,$84,$D7,$E3,$69,$F8,$58,$63,$EE,$9F,$E1
	.dc.b	$98,$FB,$9F,$E1,$D7,$D5,$F0,$9E,$53,$BB,$97,$BB,$F9,$3C,$3D,$78
	.dc.b	$FC,$08,$D4,$FA,$4E,$24,$3A,$C6,$DC,$1D,$0A,$7E,$E7,$E1,$FF,$8B
	.dc.b	$FC,$3F,$F1,$8F,$7F,$F3,$B3,$FC,$37,$E0,$7F,$3A,$F0,$3D,$07,$6D
	.dc.b	$3E,$23,$83,$B1,$F9,$8C,$BD,$30,$F8,$4F,$DB,$92,$7F,$81,$35,$FE
	.dc.b	$00,$D7,$F8,$03,$5F,$E2,$0D,$7F,$99,$B7,$2F,$E7,$0D,$9F,$04,$F6
	.dc.b	$21,$D8,$FE,$25,$F0,$FF,$CE,$3E,$1F,$F9,$CF,$C3,$7F,$3C,$FE,$01
	.dc.b	$FF,$15,$D0,$77,$E3,$24,$BE,$1B,$FA,$0E,$7E,$12,$5F,$80,$F5,$94
	.dc.b	$F5,$94,$F5,$96,$BD,$6F,$87,$4F,$97,$3D,$23,$D4,$43,$A9,$4E,$A5
	.dc.b	$3A,$96,$7A,$B7,$4F,$8E,$D9,$EC,$F4,$3F,$99,$70,$7C,$77,$FA,$96
	.dc.b	$7F,$1F,$93,$E4,$65,$F0,$A7,$D5,$87,$47,$91,$E4,$29,$E4,$2E,$3E
	.dc.b	$46,$C7,$17,$81,$EC,$CB,$E1,$BD,$B2,$F3,$BA,$3A,$16,$BA,$36,$BA
	.dc.b	$3A,$24,$69,$B1,$A8,$C1,$1C,$0C,$BE,$24,$FF,$83,$72,$F8,$8F,$E2
	.dc.b	$74,$FC,$46,$9F,$87,$8F,$EF,$58,$FA,$1E,$ED,$C7,$B3,$71,$EE,$FF
	.dc.b	$BA,$72,$F8,$35,$F5,$CF,$84,$43,$04,$3E,$5C,$73,$23,$FA,$23,$CD
	.dc.b	$A7,$F8,$8B,$0F,$EA,$11,$83,$FC,$4E,$BE,$17,$F4,$DF,$07,$FA,$7F
	.dc.b	$83,$FC,$05,$77,$3E,$76,$1D,$9C,$1E,$E3,$F6,$78,$FF,$1E,$C1,$F0
	.dc.b	$7F,$B1,$24,$F8,$74,$FD,$7C,$62,$3F,$67,$9F,$9E,$31,$AF,$CE,$7F
	.dc.b	$75,$F0,$BF,$BB,$F8,$6F,$DE,$7C,$37,$EF,$7B,$26,$3D,$73,$7B,$DF
	.dc.b	$C0,$7E,$03,$F8,$17,$C3,$7F,$03,$F8,$39,$7B,$F8,$6F,$E0,$DF,$0B
	.dc.b	$FC,$1F,$E1,$BF,$84,$7C,$37,$F0,$9F,$86,$FE,$15,$EB,$FF,$0B,$BE
	.dc.b	$38,$7A,$0F,$97,$8F,$67,$C5,$9F,$F1,$1D,$3C,$B1,$BE,$5F,$E1,$FB
	.dc.b	$FA,$A3,$D1,$C3,$EE,$7D,$5C,$7F,$17,$C9,$FF,$4D,$7C,$67,$FB,$72
	.dc.b	$79,$37,$9B,$F8,$D3,$79,$FF,$CC,$2D,$9F,$F3,$2B,$72,$34,$9E,$C7
	.dc.b	$86,$7C,$DC,$CC,$FC,$D1,$E0,$F6,$BF,$CC,$F6,$F8,$7B,$EF,$7C,$1A
	.dc.b	$E3,$D3,$AF,$87,$DF,$C0,$7C,$99,$FC,$16,$0F,$87,$FA,$11,$8F,$82
	.dc.b	$EB,$F0,$FD,$BE,$13,$84,$FC,$0C,$3A,$3F,$7F,$2F,$07,$1F,$C2,$D3
	.dc.b	$CB,$AF,$9B,$6F,$4C,$FE,$E3,$7F,$73,$F2,$38,$7D,$67,$F9,$9F,$E0
	.dc.b	$BE,$4B,$8C,$BC,$ED,$7E,$8B,$67,$94,$68,$FC,$97,$85,$F9,$77,$F2
	.dc.b	$9E,$17,$8D,$F8,$39,$7C,$87,$D4,$43,$EE,$F1,$FC,$87,$F2,$C7,$C8
	.dc.b	$E2,$F8,$EF,$E0,$B8,$BE,$3B,$F3,$38,$7E,$3E,$5F,$84,$E1,$F8,$F3
	.dc.b	$F5,$0E,$D1,$F1,$DF,$19,$F5,$37,$F8,$FC,$DF,$C2,$78,$3E,$3B,$F1
	.dc.b	$27,$DC,$39,$FF,$DD,$DC,$DF,$C0,$23,$F3,$24,$3F,$17,$D1,$FB,$CA
	.dc.b	$FD,$F4,$67,$1F,$DF,$3F,$84,$F8,$7F,$C6,$F2,$7E,$A2,$7F,$97,$F5
	.dc.b	$5B,$FE,$39,$D7,$FB,$DC,$3F,$17,$F0,$FF,$B4,$BF,$E6,$BD,$DF,$B8
	.dc.b	$9F,$E0,$9E,$27,$E6,$3E,$1F,$F7,$B1,$FC,$1F,$C3,$D3,$FA,$0F,$0F
	.dc.b	$96,$7F,$CA,$70,$FC,$FE,$5D,$EB,$F4,$B9,$F8,$DF,$E5,$7A,$5F,$07
	.dc.b	$F5,$BB,$7E,$9B,$7F,$81,$5F,$CC,$38,$3F,$1C,$D3,$E1,$67,$F1,$1F
	.dc.b	$8E,$60,$D1,$82,$70,$7F,$52,$EC,$FC,$74,$7E,$E3,$B3,$C5,$C7,$FE
	.dc.b	$6A,$8F,$39,$FC,$50,$FC,$0D,$FB,$5F,$E5,$1B,$FE,$22,$3F,$A4,$71
	.dc.b	$FE,$3B,$1F,$EB,$FE,$27,$F2,$3B,$FE,$5D,$C5,$B7,$F9,$09,$C8,$EA
	.dc.b	$7F,$4E,$F0,$FF,$91,$3F,$CD,$78,$7C,$BA,$7A,$30,$F5,$4F,$BF,$87
	.dc.b	$FA,$2B,$FE,$37,$E6,$FE,$43,$DF,$FA,$FD,$7F,$25,$3F,$B2,$3F,$27
	.dc.b	$7F,$B2,$70,$47,$AB,$FC,$73,$A3,$F3,$AF,$F2,$4E,$9F,$E0,$EF,$F1
	.dc.b	$4F,$03,$F8,$5D,$FF,$69,$F0,$7F,$53,$87,$EA,$EB,$F6,$67,$F3,$2E
	.dc.b	$FF,$8D,$5F,$CC,$3B,$FF,$67,$3F,$CD,$BB,$FE,$36,$DF,$CC,$3B,$F9
	.dc.b	$DF,$5A,$F8,$F3,$FC,$D3,$3F,$1C,$7F,$2F,$3F,$84,$9F,$D1,$8F,$E1
	.dc.b	$A7,$C0,$9F,$87,$BF,$C6,$8C,$1F,$D0,$37,$F9,$0D,$33,$7F,$D0,$A3
	.dc.b	$2F,$F9,$8B,$A3,$EA,$5F,$F8,$27,$87,$F8,$C5,$FF,$50,$F0,$7F,$8C
	.dc.b	$3F,$C2,$C9,$BF,$F5,$2F,$07,$F8,$C4,$7F,$4A,$F1,$3F,$A3,$F0,$FC
	.dc.b	$68,$FE,$A7,$C3,$F1,$AF,$FA,$CF,$0F,$C6,$F1,$1C,$5F,$1A,$FF,$C2
	.dc.b	$FC,$7F,$51,$FE,$B7,$C7,$F1,$B0,$FE,$C3,$C9,$F7,$38,$BF,$CF,$7C
	.dc.b	$BF,$C6,$2F,$E0,$1F,$84,$E5,$FE,$2F,$DB,$FB,$4A,$FC,$7C,$FE,$B3
	.dc.b	$93,$F6,$58,$FF,$14,$8C,$1F,$B2,$7F,$B7,$F7,$7F,$16,$8F,$CE,$10
	.dc.b	$FE,$D3,$D1,$FA,$07,$FB,$5F,$4F,$F0,$A7,$F4,$C7,$F6,$EE,$DF,$E2
	.dc.b	$37,$FE,$4C,$ED,$FE,$2D,$3F,$E1,$BE,$DF,$D8,$3F,$C5,$4F,$D8,$9C
	.dc.b	$A7,$F8,$FB,$FC,$88,$FD,$F1,$FC,$00,$F1,$9E,$57,$F9,$5B,$FD,$CB
	.dc.b	$BF,$F8,$B3,$FB,$1E,$FF,$E7,$4F,$F3,$D8,$FF,$05,$FC,$3F,$FA,$A6
	.dc.b	$21,$FE,$0C,$6B,$A2,$ED,$FE,$0E,$8C,$C6,$CC,$F3,$23,$33,$B9,$CC
	.dc.b	$62,$75,$18,$21,$E2,$9E,$0D,$8D,$B6,$79,$11,$E2,$7A,$E7,$8B,$67
	.dc.b	$8B,$9B,$65,$FE,$20,$F8,$6F,$F1,$0F,$C3,$FF,$88,$BE,$1F,$FC,$47
	.dc.b	$EE,$FF,$89,$33,$ED,$8F,$84,$EE,$7C,$53,$04,$3C,$0F,$06,$9C,$0F
	.dc.b	$90,$70,$F2,$DC,$5C,$E3,$67,$33,$04,$38,$9A,$21,$9D,$7E,$0B,$75
	.dc.b	$F8,$2D,$BF,$23,$4F,$24,$F0,$7F,$D1,$7A,$7C,$57,$FC,$7B,$9F,$F9
	.dc.b	$1B,$DE,$FF,$23,$FC,$37,$C5,$9E,$38,$F6,$7C,$A7,$53,$D4,$60,$98
	.dc.b	$77,$24,$71,$7F,$03,$9F,$C4,$9D,$CE,$49,$97,$AA,$32,$DA,$E6,$D9
	.dc.b	$DC,$F9,$A3,$04,$7F,$3B,$BF,$E1,$1F,$F2,$DE,$FF,$84,$7F,$CB,$B9
	.dc.b	$7C,$07,$FC,$BF,$5C,$5F,$6E,$1E,$07,$B1,$0C,$E1,$E2,$58,$FB,$67
	.dc.b	$84,$6A,$BA,$3F,$70,$F6,$9C,$CC,$47,$06,$D7,$56,$CF,$EC,$A7,$F7
	.dc.b	$BC,$2E,$65,$26,$1F,$C0,$E7,$51,$8F,$80,$FD,$23,$FE,$7E,$E1,$FC
	.dc.b	$03,$E7,$3C,$87,$DF,$3C,$26,$0F,$49,$F0,$38,$7C,$37,$E1,$F8,$1F
	.dc.b	$06,$38,$84,$62,$BC,$AD,$8E,$11,$FE,$D1,$76,$D8,$D4,$62,$3F,$D6
	.dc.b	$CD,$C7,$DA,$F2,$8C,$CF,$02,$3C,$C8,$68,$B3,$CA,$7A,$57,$A6,$B8
	.dc.b	$D7,$83,$E0,$BC,$67,$93,$E0,$F3,$87,$1A,$B7,$C6,$FA,$B5,$F0,$12
	.dc.b	$78,$3A,$3E,$29,$70,$F6,$46,$A3,$8F,$63,$9F,$E7,$7C,$27,$C6,$24
	.dc.b	$FC,$B6,$BF,$21,$B5,$EF,$7E,$6F,$C3,$7C,$EF,$86,$F9,$FF,$05,$F4
	.dc.b	$3E,$13,$E8,$FC,$27,$D2,$F8,$6F,$A7,$F0,$DF,$53,$E0,$3E,$AF,$C3
	.dc.b	$7D,$6F,$86,$FA,$FF,$0D,$F6,$3D,$EF,$B3,$F0,$DF,$1F,$87,$EF,$7C
	.dc.b	$37,$DC,$F8,$5D,$C7,$E0,$70,$E6,$FE,$0F,$2E,$EF,$E1,$74,$F1,$BF
	.dc.b	$87,$DB,$CE,$FE,$27,$83,$D6,$FC,$12,$1F,$D9,$CF,$D2,$F1,$7C,$2E
	.dc.b	$7F,$93,$1A,$E4,$F8,$8D,$8C,$BF,$24,$FC,$B7,$27,$C7,$7F,$31,$C9
	.dc.b	$F1,$DF,$CD,$72,$7C,$77,$F3,$9C,$9F,$1D,$FC,$F7,$27,$C7,$7F,$41
	.dc.b	$C9,$F1,$DF,$D1,$71,$FC,$77,$F3,$FC,$7F,$1D,$D9,$F1,$CE,$67,$C7
	.dc.b	$3B,$9F,$1C,$F1,$9F,$1C,$F3,$9F,$1C,$F5,$9E,$F3,$E2,$9F,$E7,$DE
	.dc.b	$5F,$CC,$6B,$F9,$12,$6B,$82,$4F,$E6,$0F,$CE,$38,$E7,$FC,$CB,$5F
	.dc.b	$D6,$C6,$E7,$9B,$67,$5D,$1F,$AA,$8C,$12,$BF,$8F,$75,$7E,$CA,$3F
	.dc.b	$84,$3B,$8E,$F1,$B7,$B9,$82,$18,$24,$F1,$6C,$7E,$C5,$AF,$E2,$5F
	.dc.b	$E1,$CF,$F8,$55,$FE,$28,$D8,$FE,$6C,$47,$FC,$5F,$1A,$7F,$81,$1B
	.dc.b	$33,$1B,$7F,$C5,$11,$B7,$E6,$1B,$33,$1B,$70,$43,$A9,$0F,$E8,$73
	.dc.b	$F1,$DF,$E3,$47,$F2,$23,$81,$D8,$E1,$3E,$E6,$BE,$33,$0E,$C8,$E0
	.dc.b	$86,$C9,$3F,$E5,$5F,$03,$FB,$34,$62,$7E,$09,$0F,$69,$4F,$2B,$D2
	.dc.b	$7E,$04,$79,$1A,$F8,$A3,$A3,$C8,$87,$25,$D3,$FC,$52,$3F,$A8,$75
	.dc.b	$FF,$3E,$75,$5D,$2B,$C1,$1E,$33,$C0,$53,$C1,$1B,$8F,$0D,$FB,$A3
	.dc.b	$04,$9E,$13,$D1,$F2,$9D,$0A,$74,$36,$FB,$1F,$29,$F8,$CB,$F5,$3F
	.dc.b	$D9,$2F,$D4,$F3,$36,$74,$9F,$03,$82,$1B,$3F,$AB,$5F,$A5,$F0,$DF
	.dc.b	$A5,$E5,$3E,$37,$81,$E7,$33,$1B,$7A,$9F,$57,$0F,$C9,$65,$FD,$BD
	.dc.b	$F3,$C7,$13,$75,$E6,$7A,$B5,$D1,$F1,$EF,$CC,$FF,$10,$3F,$80,$5F
	.dc.b	$EE,$C8,$FE,$ED,$AE,$E7,$F8,$1C,$62,$FF,$89,$C7,$F0,$6B,$F3,$E5
	.dc.b	$FE,$1B,$74,$FF,$16,$BE,$84,$3A,$4F,$F0,$8A,$E6,$D9,$FF,$01,$75
	.dc.b	$FF,$79,$AF,$39,$C6,$B7,$87,$68,$C9,$D1,$C5,$FA,$23,$85,$FF,$2D
	.dc.b	$CF,$F0,$77,$FE,$1B,$CF,$77,$E1,$BE,$73,$D9,$E2,$62,$FD,$8F,$F8
	.dc.b	$3B,$4B,$3B,$8F,$11,$C2,$BC,$4E,$09,$1B,$25,$79,$63,$FA,$E9,$FB
	.dc.b	$03,$94,$F9,$4E,$73,$C2,$3C,$84,$8C,$CF,$CA,$AF,$F2,$8F,$F0,$0F
	.dc.b	$F9,$7A,$97,$FE,$44,$8C,$36,$BF,$CA,$11,$87,$89,$82,$4F,$F9,$B2
	.dc.b	$78,$36,$3B,$FC,$3F,$2D,$B9,$FC,$37,$4F,$86,$EB,$F0,$DD,$BD,$1E
	.dc.b	$FE,$F7,$83,$D3,$F0,$FB,$DE,$2F,$17,$C7,$F0,$3F,$8E,$DB,$F9,$B9
	.dc.b	$FD,$70,$F0,$1F,$E7,$DE,$07,$A9,$EA,$F6,$38,$CE,$B9,$15,$FE,$AF
	.dc.b	$8F,$87,$C5,$F1,$B2,$F1,$46,$E5,$A7,$9E,$31,$B7,$AD,$D7,$07,$BD
	.dc.b	$E1,$C5,$F1,$9E,$A7,$97,$1F,$F4,$9B,$FD,$9F,$07,$AB,$88,$FF,$7E
	.dc.b	$46,$1C,$1E,$D8,$AD,$33,$9F,$B2,$32,$D8,$CC,$F3,$7B,$46,$5B,$19
	.dc.b	$7C,$33,$CA,$31,$1F,$8A,$8C,$11,$EA,$D7,$9B,$5E,$67,$F7,$EE,$AF
	.dc.b	$B5,$5F,$4B,$DD,$FC,$7F,$A9,$F9,$0F,$86,$FC,$8F,$AF,$F9,$2C,$79
	.dc.b	$63,$E1,$3B,$90,$EE,$53,$B9,$82,$18,$33,$D1,$FE,$D3,$CF,$DB,$1F
	.dc.b	$2C,$D9,$DC,$A6,$4E,$46,$4A,$64,$C1,$0C,$13,$2F,$A0,$F0,$6B,$B6
	.dc.b	$BF,$9E,$3F,$C2,$67,$E2,$9F,$CC,$FC,$27,$ED,$BE,$0F,$FD,$8D,$FC
	.dc.b	$02,$F8,$4B,$EE,$FC,$9C,$3E,$21,$D7,$1F,$7E,$3D,$88,$E0,$99,$FD
	.dc.b	$09,$F4,$47,$E2,$73,$F9,$13,$9A,$E2,$D9,$E3,$5E,$03,$8C,$64,$8E
	.dc.b	$67,$8B,$67,$CB,$0E,$0F,$C3,$C3,$4F,$81,$5F,$BB,$CF,$E0,$37,$2F
	.dc.b	$80,$FD,$8C,$3A,$B6,$FC,$AD,$C7,$C6,$FE,$1F,$2F,$6B,$E3,$8C,$BF
	.dc.b	$5E,$78,$B6,$32,$FC,$D6,$BF,$00,$F7,$60,$F8,$4F,$AF,$EF,$FF,$2E
	.dc.b	$E2,$FD,$0D,$7F,$0F,$E1,$FD,$0B,$FC,$47,$0E,$79,$FA,$A3,$C0,$79
	.dc.b	$48,$7A,$7B,$0A,$19,$3E,$13,$F8,$2B,$67,$F8,$4B,$5F,$E0,$AD,$7F
	.dc.b	$8B,$CF,$29,$E2,$48,$E2,$D7,$8B,$5E,$33,$CD,$F2,$CF,$37,$89,$0E
	.dc.b	$25,$38,$CF,$37,$8C,$FC,$B7,$F0,$5F,$03,$FA,$23,$F4,$7F,$09,$FA
	.dc.b	$4F,$84,$FD,$2F,$C3,$FE,$9B,$0E,$57,$F1,$70,$F2,$36,$33,$8F,$96
	.dc.b	$3C,$64,$3C,$67,$94,$F1,$C7,$C1,$E1,$67,$89,$D1,$D4,$B5,$D5,$E6
	.dc.b	$75,$38,$14,$E0,$57,$83,$A3,$89,$0C,$10,$E2,$64,$F8,$1C,$1F,$C2
	.dc.b	$0F,$E1,$3F,$0B,$F2,$23,$F5,$9F,$0B,$F2,$1F,$D7,$63,$F1,$32,$F8
	.dc.b	$F8,$78,$A7,$A9,$DD,$C1,$23,$64,$9F,$8D,$8E,$E3,$15,$F1,$1A,$FB
	.dc.b	$DA,$FA,$6B,$A5,$78,$C9,$1C,$5F,$36,$3E,$D8,$FE,$0D,$8F,$B5,$FE
	.dc.b	$55,$6C,$7D,$D1,$8B,$E6,$F8,$5A,$F8,$67,$C6,$FA,$63,$FD,$71,$8D
	.dc.b	$8F,$11,$0E,$45,$39,$47,$47,$81,$A3,$E0,$79,$1F,$9B,$7F,$3A,$7F
	.dc.b	$1F,$F8,$6F,$CE,$3F,$C8,$7E,$1F,$F3,$8F,$53,$99,$D8,$C1,$F5,$89
	.dc.b	$A7,$E7,$75,$FC,$BC,$7E,$8F,$E0,$3F,$82,$91,$FE,$0A,$D7,$F8,$2B
	.dc.b	$5F,$DE,$B5,$FD,$AB,$5E,$87,$22,$1C,$8A,$72,$29,$C8,$FE,$2B,$F0
	.dc.b	$7F,$D7,$BE,$1B,$FA,$FF,$C3,$FF,$60,$F8,$2F,$EC,$3F,$0B,$2F,$FA
	.dc.b	$26,$1E,$1C,$BF,$90,$B7,$80,$D2,$4F,$80,$F3,$9E,$51,$9F,$C6,$AE
	.dc.b	$71,$B6,$BB,$3D,$45,$3D,$45,$3A,$9A,$21,$A2,$9A,$30,$43,$04,$7C
	.dc.b	$2F,$F2,$EC,$FF,$C0,$7F,$0F,$F1,$A3,$FB,$36,$5F,$15,$FD,$0E,$5F
	.dc.b	$E0,$8C,$7E,$0C,$79,$70,$E6,$F7,$6B,$DF,$2F,$92,$FC,$16,$BF,$1F
	.dc.b	$2F,$8D,$18,$8F,$5C,$F2,$24,$66,$3B,$35,$ED,$5E,$47,$8C,$66,$78
	.dc.b	$CE,$67,$8B,$FD,$73,$2F,$83,$1F,$E0,$5C,$FF,$0C,$FF,$81,$F1,$F8
	.dc.b	$5E,$8B,$C6,$E3,$3E,$BC,$3A,$3E,$23,$04,$3F,$07,$C7,$C1,$F9,$19
	.dc.b	$F9,$49,$3C,$B3,$F3,$CF,$13,$D7,$19,$9D,$C7,$E2,$FF,$80,$7F,$D0
	.dc.b	$58,$F9,$63,$FA,$B7,$07,$E8,$5F,$F4,$D4,$FF,$7D,$6F,$C3,$7F,$A3
	.dc.b	$2B,$86,$2A,$7A,$DF,$7E,$FC,$47,$A4,$F1,$98,$3E,$01,$F6,$F3,$FE
	.dc.b	$EF,$3F,$E3,$38,$FF,$47,$B5,$FF,$45,$B5,$FF,$03,$B5,$E3,$E2,$FF
	.dc.b	$A5,$23,$11,$FE,$15,$6D,$F9,$E7,$44,$9E,$ED,$7B,$B5,$EE,$D7,$B9
	.dc.b	$C0,$87,$9B,$D5,$F8,$DB,$7B,$3E,$02,$CF,$8F,$E0,$2E,$1F,$27,$E0
	.dc.b	$6E,$4F,$81,$BF,$05,$8D,$3C,$18,$FA,$DB,$87,$AE,$3C,$27,$98,$A7
	.dc.b	$98,$F5,$98,$39,$1F,$E3,$AE,$7F,$E2,$18,$7C,$62,$57,$66,$CF,$1A
	.dc.b	$F0,$3D,$08,$78,$4A,$78,$4C,$10,$C1,$0F,$19,$C0,$FB,$1A,$FC,$4F
	.dc.b	$82,$F8,$BF,$07,$F1,$BE,$0F,$E3,$E9,$2F,$F2,$F9,$79,$C8,$F9,$F2
	.dc.b	$F4,$BD,$72,$F4,$BE,$33,$C3,$9F,$C2,$AE,$31,$BA,$E2,$EE,$B8,$BE
	.dc.b	$96,$BE,$A3,$51,$C1,$D3,$5D,$18,$CF,$DC,$FE,$C3,$2F,$73,$FB,$2C
	.dc.b	$7D,$CF,$FA,$D6,$B9,$5F,$80,$8E,$09,$87,$AA,$7E,$A6,$3E,$D7,$E2
	.dc.b	$46,$5F,$2B,$5F,$29,$2B,$AC,$65,$B1,$C8,$FD,$BF,$C2,$7E,$6B,$83
	.dc.b	$F3,$7C,$1F,$9C,$E0,$FC,$EF,$C2,$7E,$72,$FE,$B9,$C4,$F9,$D8,$F5
	.dc.b	$CF,$D3,$19,$CF,$F2,$BF,$0B,$FB,$D2,$47,$F0,$66,$BF,$C1,$9A,$FF
	.dc.b	$06,$6B,$FC,$39,$AF,$F0,$56,$E9,$FC,$6D,$B4,$69,$8F,$85,$FD,$BF
	.dc.b	$C1,$7E,$E3,$E1,$7F,$73,$F0,$BF,$BA,$F8,$6F,$DE,$FA,$BF,$CB,$73
	.dc.b	$FF,$76,$2F,$6F,$F7,$F9,$74,$9F,$51,$0F,$51,$4F,$51,$4F,$51,$BD
	.dc.b	$7E,$2B,$93,$D8,$5D,$7D,$8D,$D7,$D8,$DA,$F1,$3E,$02,$1E,$02,$9E
	.dc.b	$02,$9D,$8B,$2C,$3F,$C3,$71,$FE,$C6,$ED,$FE,$8F,$0F,$73,$E2,$FC
	.dc.b	$2F,$C6,$F7,$BE,$45,$EF,$BB,$FD,$73,$7F,$27,$0E,$3F,$0B,$F2,$2B
	.dc.b	$D0,$43,$C3,$A7,$B1,$BA,$7B,$35,$F8,$B1,$83,$81,$ED,$21,$ED,$29
	.dc.b	$ED,$2E,$BF,$29,$E4,$43,$35,$E1,$77,$8F,$C2,$7F,$03,$AF,$C0,$7D
	.dc.b	$C7,$FB,$A6,$1A,$7C,$28,$FC,$3E,$9F,$8A,$CF,$E0,$47,$73,$CC,$43
	.dc.b	$CC,$5C,$7C,$CF,$93,$3F,$D2,$E7,$F0,$A3,$07,$C2,$38,$67,$E9,$6C
	.dc.b	$FE,$11,$D9,$E1,$3B,$92,$BB,$BC,$B8,$FF,$4A,$FF,$82,$B8,$BF,$4A
	.dc.b	$FC,$09,$FC,$37,$C0,$7E,$9B,$E1,$7F,$4F,$C5,$EE,$AE,$E4,$3C,$67
	.dc.b	$9C,$F0,$65,$F7,$F3,$FF,$34,$6E,$F8,$25,$42,$65,$F9,$58,$C4,$62
	.dc.b	$78,$38,$3E,$A4,$56,$99,$9E,$4D,$9E,$2F,$48,$CB,$63,$2F,$79,$E9
	.dc.b	$3A,$8C,$1C,$E3,$33,$C2,$3F,$47,$18,$23,$B2,$1D,$FC,$3F,$E2,$3F
	.dc.b	$0B,$FC,$4B,$E1,$7F,$89,$FC,$2F,$F1,$4F,$73,$F8,$AF,$BD,$FC,$5B
	.dc.b	$E1,$7F,$8B,$E5,$D3,$2F,$43,$5F,$29,$83,$B9,$82,$1E,$63,$C7,$AF
	.dc.b	$D4,$8E,$51,$93,$91,$0E,$27,$CE,$30,$6C,$CE,$BF,$1D,$C1,$0E,$06
	.dc.b	$B5,$F9,$0E,$0D,$47,$17,$F6,$1A,$FF,$36,$F7,$FF,$9B,$FC,$2F,$C6
	.dc.b	$9F,$CB,$5B,$4F,$8F,$18,$CF,$E3,$BF,$08,$EB,$3E,$0C,$7B,$11,$C1
	.dc.b	$32,$F8,$73,$EF,$CF,$E3,$3C,$CF,$41,$2B,$94,$F0,$6C,$6D,$AF,$C8
	.dc.b	$38,$57,$77,$4D,$7F,$0C,$7B,$B2,$F8,$2F,$C0,$D7,$F0,$CF,$C2,$D7
	.dc.b	$F0,$CF,$C4,$C3,$C5,$7C,$E3,$E8,$46,$08,$FE,$0B,$5F,$8B,$18,$8F
	.dc.b	$36,$3E,$87,$89,$E6,$8D,$38,$37,$5D,$DF,$9F,$F0,$7F,$9E,$7E,$BE
	.dc.b	$FF,$DC,$3E,$17,$FC,$DB,$FC,$02,$F8,$6B,$86,$3E,$03,$EC,$E1,$D3
	.dc.b	$2E,$EF,$98,$F1,$10,$FF,$59,$8C,$D8,$3E,$17,$F7,$6D,$9F,$E0,$CD
	.dc.b	$7F,$82,$35,$FE,$26,$D7,$F8,$BC,$F4,$9E,$64,$8E,$6D,$79,$B5,$E6
	.dc.b	$D7,$9C,$F5,$7D,$0D,$9E,$D1,$D0,$87,$9B,$DF,$F9,$3F,$0B,$F2,$BE
	.dc.b	$17,$E5,$FC,$2F,$CC,$F8,$5F,$9B,$F0,$BF,$3B,$E1,$7E,$7E,$9E,$1C
	.dc.b	$BB,$B7,$3E,$F1,$E8,$3B,$9E,$82,$9E,$88,$F7,$74,$39,$71,$3A,$3A
	.dc.b	$96,$BA,$BC,$CE,$C4,$30,$43,$99,$92,$99,$3A,$99,$29,$97,$BB,$F3
	.dc.b	$BE,$17,$E3,$BF,$D7,$77,$6B,$F2,$63,$1F,$0B,$F9,$4F,$85,$F9,$11
	.dc.b	$BC,$FE,$35,$FD,$52,$38,$24,$7B,$E3,$F1,$99,$FC,$58,$F2,$9E,$A2
	.dc.b	$18,$26,$5E,$86,$CF,$67,$A5,$76,$7E,$04,$EC,$93,$B8,$CA,$CB,$E0
	.dc.b	$C3,$CB,$79,$3F,$92,$D3,$E1,$4E,$63,$F2,$BA,$7C,$38,$F9,$1B,$7E
	.dc.b	$29,$F1,$90,$EF,$9F,$95,$FA,$59,$7C,$2D,$BD,$FA,$79,$AB,$55,$8A
	.dc.b	$F4,$3C,$23,$2F,$C2,$6C,$F1,$8F,$9D,$C7,$FC,$2B,$E0,$3F,$85,$FC
	.dc.b	$1F,$E8,$A7,$EF,$71,$FE,$89,$FC,$07,$C0,$7A,$5F,$39,$E9,$3C,$05
	.dc.b	$3C,$98,$FF,$85,$78,$BE,$D6,$5F,$C3,$3E,$0F,$F8,$59,$2B,$F8,$5B
	.dc.b	$5F,$E1,$6D,$7F,$85,$B5,$FE,$02,$D7,$F8,$0B,$63,$B4,$74,$21,$D0
	.dc.b	$A7,$43,$F8,$DF,$C2,$7F,$2C,$F8,$4F,$E5,$BF,$0B,$FE,$2D,$E5,$FB
	.dc.b	$F5,$FC,$97,$F8,$07,$FC,$5B,$18,$C1,$1F,$CA,$67,$FC,$84,$BE,$17
	.dc.b	$97,$37,$F2,$28,$7C,$2E,$2B,$A7,$B5,$FC,$F3,$5F,$57,$29,$A4,$9F
	.dc.b	$49,$F0,$8F,$6E,$FF,$3A,$78,$46,$8F,$8A,$5C,$FE,$2B,$F0,$C8,$7C
	.dc.b	$32,$9D,$4A,$72,$2B,$C9,$E0,$43,$81,$4E,$07,$F4,$4D,$FE,$73,$FC
	.dc.b	$2F,$6F,$85,$3F,$2A,$2E,$6F,$82,$BE,$FE,$35,$CF,$BF,$F1,$EF,$D9
	.dc.b	$9F,$C0,$7C,$AD,$7C,$58,$F9,$DF,$29,$0F,$66,$9F,$27,$5F,$9D,$1C
	.dc.b	$9A,$F2,$AE,$8F,$1A,$E8,$FB,$8C,$46,$9C,$12,$36,$48,$D3,$FD,$C3
	.dc.b	$5F,$8B,$3F,$C2,$CF,$EB,$BA,$FC,$47,$E2,$E9,$F1,$1F,$C7,$E7,$F1
	.dc.b	$1F,$E6,$D8,$FA,$1F,$1B,$5F,$16,$3E,$03,$E4,$65,$F0,$63,$E5,$CE
	.dc.b	$F2,$F3,$36,$B8,$34,$93,$C0,$FF,$45,$6A,$F7,$FE,$DF,$C0,$FD,$CF
	.dc.b	$85,$FB,$BF,$0B,$F7,$B2,$FB,$FC,$9F,$80,$F8,$4F,$C0,$F2,$7D,$DC
	.dc.b	$FE,$91,$D4,$A4,$3F,$31,$B7,$F0,$ED,$1F,$0B,$FE,$92,$6C,$F9,$23
	.dc.b	$04,$70,$43,$FD,$3D,$1C,$1B,$1B,$69,$B3,$FD,$63,$1B,$6C,$6D,$A6
	.dc.b	$CF,$E0,$4E,$5F,$E0,$9F,$0B,$FA,$0F,$85,$FD,$0F,$BD,$FA,$2F,$7B
	.dc.b	$F4,$7F,$0B,$FA,$4F,$7F,$F4,$BF,$0B,$FA,$6F,$85,$FD,$3E,$7C,$F3
	.dc.b	$FE,$2C,$DE,$1F,$E2,$AD,$D7,$CD,$3E,$23,$C8,$70,$3F,$D9,$BB,$7C
	.dc.b	$B9,$E0,$48,$DD,$74,$6D,$74,$79,$1F,$18,$87,$C2,$29,$E1,$29,$C0
	.dc.b	$D1,$0F,$E5,$1C,$BF,$C1,$BF,$80,$7F,$99,$C7,$E9,$6B,$F9,$8E,$DF
	.dc.b	$1E,$3F,$DE,$52,$F8,$0F,$91,$18,$CB,$97,$C2,$FC,$88,$F4,$E7,$ED
	.dc.b	$7C,$78,$FB,$5B,$9F,$B6,$31,$7F,$16,$78,$90,$DD,$73,$7D,$47,$B4
	.dc.b	$F8,$06,$27,$B3,$6B,$B1,$27,$EC,$67,$F0,$7C,$1F,$C2,$D7,$F4,$EE
	.dc.b	$1F,$89,$9F,$BA,$3E,$26,$8C,$FD,$F1,$FD,$53,$1F,$23,$E2,$21,$E5
	.dc.b	$26,$7E,$47,$2F,$C0,$F2,$AE,$9E,$59,$F8,$07,$A6,$BA,$BC,$A3,$4F
	.dc.b	$C4,$36,$72,$8F,$CA,$72,$7E,$86,$3E,$CF,$C3,$7D,$FF,$86,$FD,$0C
	.dc.b	$7D,$CF,$86,$FD,$0B,$FD,$BB,$8F,$ED,$3D,$48,$78,$CE,$E7,$E5,$34
	.dc.b	$FF,$09,$EA,$F8,$64,$C2,$6C,$9D,$46,$4F,$F4,$DB,$87,$11,$B9,$D4
	.dc.b	$69,$AF,$FA,$B6,$78,$3D,$9A,$EC,$E3,$3C,$DB,$1A,$79,$4F,$49,$D4
	.dc.b	$78,$A7,$51,$88,$ED,$3B,$9C,$91,$C1,$9F,$86,$FD,$77,$B1,$FA,$FE
	.dc.b	$0F,$D8,$7C,$2F,$EC,$7D,$EF,$D9,$7C,$0F,$EC,$FE,$13,$F6,$9F,$05
	.dc.b	$FB,$5C,$B9,$E9,$E8,$6B,$E5,$6B,$E4,$34,$43,$47,$F1,$2C,$7E,$23
	.dc.b	$82,$1C,$4A,$68,$C1,$0C,$10,$F9,$27,$73,$46,$0E,$65,$33,$AF,$CB
	.dc.b	$79,$3F,$85,$D3,$F8,$A7,$C2,$7F,$15,$F8,$4F,$E2,$DF,$0B,$F1,$EF
	.dc.b	$F8,$9E,$BF,$1D,$FE,$2B,$9F,$C7,$7C,$47,$4E,$0F,$A8,$E0,$F1,$19
	.dc.b	$D3,$E5,$F0,$FE,$69,$F9,$71,$93,$86,$9E,$77,$89,$1C,$12,$BB,$C7
	.dc.b	$C0,$8C,$4F,$48,$E1,$3A,$7F,$31,$A7,$C0,$7F,$97,$E7,$F0,$1F,$79
	.dc.b	$0F,$E2,$7B,$7E,$19,$FE,$6B,$7D,$70,$EF,$1E,$8C,$3B,$3E,$B3,$F0
	.dc.b	$18,$FB,$DD,$57,$19,$DC,$79,$A3,$04,$76,$43,$DD,$1C,$2B,$F7,$DF
	.dc.b	$0B,$FA,$09,$FE,$93,$BF,$E8,$1F,$E9,$7F,$0B,$FA,$07,$FA,$6F,$0F
	.dc.b	$F9,$63,$AD,$F0,$BF,$6E,$7A,$E7,$E1,$69,$E1,$3F,$31,$A7,$FA,$92
	.dc.b	$9F,$0B,$CC,$93,$FC,$53,$D5,$FF,$47,$4E,$A3,$1E,$07,$F1,$8D,$BF
	.dc.b	$AF,$5F,$09,$FE,$C9,$5B,$6C,$66,$F9,$B6,$F9,$56,$E7,$35,$D8,$5F
	.dc.b	$67,$0D,$7A,$98,$9E,$55,$D2,$FB,$4F,$29,$CB,$63,$35,$CE,$FC,$33
	.dc.b	$D0,$8E,$09,$7D,$63,$C1,$5C,$AF,$AD,$F3,$AD,$57,$26,$DF,$6F,$89
	.dc.b	$FF,$7F,$4F,$97,$E1,$BE,$E1,$1F,$63,$5F,$9A,$D9,$FA,$AD,$AF,$96
	.dc.b	$D9,$F8,$1E,$CF,$E1,$FD,$EF,$C4,$7C,$37,$E2,$7E,$0B,$F1,$5F,$0D
	.dc.b	$F8,$BF,$7F,$F1,$9F,$0D,$F8,$DF,$83,$FC,$77,$C2,$FE,$3F,$E1,$7F
	.dc.b	$21,$F0,$DF,$91,$F8,$6F,$C9,$7C,$37,$E4,$FE,$1B,$F2,$9E,$F7,$E5
	.dc.b	$7E,$0F,$97,$17,$D5,$D3,$B3,$F5,$F6,$F1,$3F,$67,$83,$CC,$FD,$BE
	.dc.b	$2F,$57,$3F,$F8,$AF,$93,$FB,$37,$3F,$C8,$8C,$57,$C6,$21,$F1,$8B
	.dc.b	$CF,$F1,$9B,$CF,$F1,$9A,$7D,$12,$CF,$F8,$E7,$0F,$1C,$7E,$73,$1F
	.dc.b	$3B,$F9,$EC,$BD,$6F,$E8,$33,$F7,$BF,$A2,$E3,$F9,$0F,$E5,$78,$BE
	.dc.b	$43,$F9,$7E,$1F,$90,$FE,$67,$83,$E4,$3F,$9B,$E4,$FA,$6F,$E7,$79
	.dc.b	$7E,$BB,$F9,$FE,$6F,$B6,$FE,$87,$9F,$F8,$26,$DF,$23,$A1,$C9,$F2
	.dc.b	$35,$FC,$8F,$14,$C3,$F4,$47,$E6,$88,$7E,$5C,$C9,$78,$BF,$38,$E3
	.dc.b	$8F,$F4,$6D,$9E,$93,$A2,$61,$FA,$42,$74,$FE,$9C,$95,$E0,$BE,$75
	.dc.b	$C2,$73,$3E,$1A,$E2,$EE,$FB,$C6,$23,$53,$EF,$7D,$27,$F8,$B2,$3F
	.dc.b	$8A,$36,$3F,$C3,$51,$88,$FD,$B9,$93,$FC,$46,$E5,$E6,$68,$86,$09
	.dc.b	$5F,$C5,$5E,$47,$F8,$DA,$39,$46,$63,$8B,$5E,$31,$A2,$4F,$98,$91
	.dc.b	$E4,$6B,$D5,$AF,$80,$87,$68,$E6,$F1,$21,$C4,$BD,$5F,$D1,$67,$9C
	.dc.b	$71,$6B,$E6,$6D,$7A,$49,$5D,$A7,$A1,$0F,$19,$C4,$F0,$98,$8E,$33
	.dc.b	$D2,$79,$4E,$5A,$F8,$CE,$25,$3B,$14,$E1,$1F,$E6,$DA,$F6,$47,$03
	.dc.b	$C4,$70,$29,$C0,$D5,$79,$1B,$5E,$27,$67,$72,$18,$24,$71,$7B,$35
	.dc.b	$EC,$62,$3A,$46,$E7,$CF,$39,$AF,$88,$47,$C6,$D7,$46,$08,$60,$87
	.dc.b	$80,$87,$A2,$7C,$8D,$AF,$19,$27,$F1,$F5,$E4,$C3,$F6,$F3,$FB,$16
	.dc.b	$CF,$EE,$1B,$5F,$B7,$3F,$83,$12,$FF,$7C,$E5,$AF,$EE,$8C,$57,$5C
	.dc.b	$3F,$7F,$7D,$5B,$3A,$8C,$3D,$67,$83,$FC,$6B,$9B,$FA,$44,$62,$FB
	.dc.b	$E1,$E6,$AF,$01,$1C,$13,$0E,$F1,$96,$C6,$63,$C1,$87,$8A,$BA,$D6
	.dc.b	$89,$3A,$3D,$07,$F3,$0C,$3C,$B5,$C7,$0F,$31,$1C,$12,$78,$9E,$33
	.dc.b	$A1,$6F,$C2,$E3,$0F,$34,$60,$95,$D6,$79,$DF,$C6,$F8,$6F,$F2,$71
	.dc.b	$23,$FC,$5C,$D7,$FC,$7E,$DB,$FF,$23,$B7,$1F,$F1,$EB,$67,$FC,$64
	.dc.b	$FF,$96,$7E,$13,$FD,$C1,$4A,$FF,$26,$7F,$00,$7F,$B7,$E3,$17,$FE
	.dc.b	$42,$8F,$F2,$1B,$FE,$42,$5E,$FF,$2A,$E3,$F0,$96,$B8,$7B,$F7,$56
	.dc.b	$0F,$85,$C2,$F8,$7B,$F0,$F8,$F7,$33,$F0,$DA,$F8,$6D,$FC,$37,$0F
	.dc.b	$86,$E3,$EF,$72,$F7,$B9,$E9,$D3,$6E,$BC,$1D,$B8,$BB,$F0,$78,$38
	.dc.b	$7E,$11,$F3,$08,$39,$BF,$C3,$DC,$4F,$86,$FF,$78,$36,$7F,$DE,$ED
	.dc.b	$44,$3E,$44,$70,$23,$C1,$AE,$CD,$99,$21,$9A,$F9,$31,$C2,$3B,$46
	.dc.b	$1D,$98,$9F,$23,$67,$83,$F7,$9F,$EB,$3C,$BF,$7A,$BE,$AF,$1F,$D4
	.dc.b	$F8,$3F,$AB,$F0,$5F,$5B,$E1,$BE,$BF,$C3,$7D,$8F,$86,$FB,$3F,$03
	.dc.b	$F7,$7F,$80,$7F,$0F,$AB,$4E,$9B,$78,$A3,$C0,$79,$8A,$7B,$5D,$3F
	.dc.b	$08,$C6,$DE,$87,$1E,$0B,$D7,$E3,$1E,$D3,$A1,$C4,$C1,$C8,$CF,$07
	.dc.b	$C7,$70,$43,$64,$3A,$90,$E8,$57,$93,$B3,$27,$EF,$FB,$BF,$57,$C1
	.dc.b	$F1,$7C,$1F,$95,$3F,$A9,$DF,$E4,$63,$F2,$A3,$FD,$5B,$AF,$CA,$D7
	.dc.b	$E4,$E9,$F0,$73,$EF,$5E,$1D,$7E,$3C,$62,$3C,$2D,$3D,$D8,$7C,$78
	.dc.b	$CC,$F0,$25,$75,$6C,$7C,$66,$BF,$14,$F0,$46,$C8,$ED,$FC,$5E,$9F
	.dc.b	$E5,$29,$69,$F0,$6B,$F8,$47,$07,$E2,$9F,$7E,$1F,$05,$FC,$45,$FF
	.dc.b	$B7,$73,$F8,$4F,$82,$3B,$10,$F2,$1E,$CC,$7F,$C1,$DB,$6E,$33,$9F
	.dc.b	$C3,$CF,$D7,$1F,$5E,$31,$5E,$A8,$DC,$E9,$E4,$6D,$C3,$FC,$8F,$BF
	.dc.b	$F8,$BB,$F6,$3E,$1F,$F4,$EF,$DA,$F8,$6F,$D2,$7C,$37,$E9,$E3,$EE
	.dc.b	$FC,$37,$E9,$FD,$8F,$E4,$39,$78,$67,$B3,$5F,$56,$FF,$67,$46,$EF
	.dc.b	$80,$FE,$11,$5F,$1D,$AF,$FB,$E2,$34,$EA,$78,$CE,$9B,$19,$9D,$BF
	.dc.b	$C3,$DA,$FF,$0E,$6B,$DC,$E1,$3C,$E3,$31,$E1,$8C,$11,$D9,$0E,$CE
	.dc.b	$5A,$7F,$2C,$D7,$F8,$67,$C3,$7F,$0D,$E0,$FE,$1D,$F0,$DF,$C3,$FE
	.dc.b	$13,$F8,$87,$BD,$FC,$47,$E0,$BF,$89,$7C,$27,$F1,$3F,$86,$FE,$29
	.dc.b	$A7,$5D,$3D,$4D,$7D,$47,$80,$86,$8F,$36,$FF,$47,$6F,$6E,$FF,$22
	.dc.b	$76,$72,$38,$18,$3E,$11,$EF,$32,$43,$81,$0D,$9C,$5E,$CE,$CC,$95
	.dc.b	$CB,$FD,$E3,$7F,$8E,$FF,$0A,$F8,$5F,$8E,$FF,$0C,$F8,$6F,$8E,$FF
	.dc.b	$0E,$F8,$5F,$8E,$FF,$10,$F8,$5F,$8E,$F9,$CF,$26,$1D,$9F,$09,$8C
	.dc.b	$7E,$19,$33,$FD,$8C,$68,$93,$C4,$93,$B8,$EC,$69,$C3,$F0,$5A,$FB
	.dc.b	$DA,$F8,$67,$ED,$3F,$B3,$3F,$94,$F1,$FF,$53,$CB,$E0,$C7,$F0,$C3
	.dc.b	$F4,$99,$7C,$27,$18,$7E,$B3,$2F,$87,$18,$C5,$97,$C4,$8F,$26,$5E
	.dc.b	$0B,$F5,$B6,$BD,$3A,$7E,$1E,$30,$4A,$F7,$D7,$06,$D7,$B1,$F8,$53
	.dc.b	$B7,$B9,$A7,$F7,$1F,$C0,$2F,$86,$FF,$16,$2F,$57,$F4,$D5,$FE,$12
	.dc.b	$F0,$7F,$BE,$4B,$D7,$FD,$44,$E7,$C0,$FF,$0B,$E6,$A7,$B1,$FA,$BA
	.dc.b	$D0,$9D,$7B,$3F,$AD,$8F,$EA,$FA,$79,$DF,$AC,$D7,$EB,$1F,$E2,$4D
	.dc.b	$BF,$8D,$60,$F8,$2F,$59,$27,$F8,$B4,$65,$B1,$97,$FC,$45,$19,$6C
	.dc.b	$65,$FF,$11,$C6,$5B,$19,$7D,$F1,$96,$C6,$5F,$74,$65,$B1,$9C,$3E
	.dc.b	$33,$F4,$BD,$BF,$A3,$F0,$5F,$4B,$DD,$FA,$7F,$05,$30,$F9,$38,$7C
	.dc.b	$DF,$7F,$EB,$FB,$FF,$63,$E1,$A5,$6B,$4F,$05,$F6,$8F,$82,$43,$DE
	.dc.b	$53,$DA,$53,$D6,$7C,$FF,$0E,$BC,$08,$70,$29,$A3,$04,$30,$43,$E1
	.dc.b	$9D,$CA,$72,$29,$C4,$A7,$18,$F6,$3F,$6B,$4F,$CC,$FC,$1F,$C9,$8F
	.dc.b	$85,$F0,$DF,$9C,$F8,$4F,$CE,$FC,$37,$C8,$9F,$8D,$F0,$BF,$A0,$F8
	.dc.b	$6F,$D0,$F1,$FD,$5D,$7E,$49,$1C,$12,$BE,$21,$E4,$26,$9E,$47,$D0
	.dc.b	$D7,$D1,$3C,$DF,$81,$19,$78,$C6,$5F,$85,$19,$24,$66,$3A,$AD,$3E
	.dc.b	$0C,$7E,$07,$8B,$F7,$7A,$7C,$18,$FC,$26,$7F,$05,$FE,$2F,$6C,$FE
	.dc.b	$14,$7E,$27,$2F,$73,$E2,$9F,$03,$EA,$8F,$D5,$72,$7E,$C9,$DC,$79
	.dc.b	$1A,$EC,$F7,$CF,$27,$DF,$1A,$69,$AF,$86,$FD,$1B,$F9,$BE,$5F,$D1
	.dc.b	$3E,$8E,$5F,$E2,$FC,$BF,$C6,$3E,$1B,$F4,$33,$EC,$E5,$FB,$6F,$83
	.dc.b	$1E,$DC,$3F,$29,$F6,$E9,$FE,$23,$F7,$3F,$C3,$F5,$FC,$55,$B5,$FE
	.dc.b	$D4,$8F,$F0,$F4,$60,$8E,$08,$7F,$89,$5C,$B4,$C9,$FE,$B1,$72,$D3
	.dc.b	$27,$FA,$C9,$CB,$4F,$0C,$7F,$B8,$63,$C9,$1F,$D4,$17,$BB,$93,$5E
	.dc.b	$F6,$FD,$FE,$1E,$4F,$1F,$7B,$97,$C3,$73,$C7,$C2,$7F,$3A,$C7,$C2
	.dc.b	$D3,$27,$C5,$2E,$5E,$25,$9F,$CB,$73,$3C,$9E,$07,$C3,$21,$F0,$4A
	.dc.b	$7B,$CA,$78,$4A,$72,$38,$10,$E0,$53,$81,$F3,$FE,$1B,$E4,$7C,$37
	.dc.b	$C9,$F8,$6F,$95,$F0,$DF,$1A,$BF,$96,$DB,$5F,$9D,$F0,$DF,$3F,$E1
	.dc.b	$BE,$3D,$68,$F6,$90,$F2,$E9,$F0,$DF,$19,$E4,$D3,$E2,$3A,$C3,$E2
	.dc.b	$4E,$E3,$CE,$70,$29,$0F,$80,$64,$A6,$5E,$4E,$08,$E0,$91,$F9,$9D
	.dc.b	$3E,$1C,$FC,$0C,$FE,$1B,$F0,$B8,$3F,$10,$FC,$4C,$BE,$13,$F1,$B6
	.dc.b	$FC,$43,$EB,$3E,$11,$F4,$4F,$0E,$1F,$9E,$CB,$E0,$C6,$72,$F1,$BC
	.dc.b	$0F,$63,$87,$A1,$0E,$13,$F2,$08,$FE,$90,$FB,$2F,$E8,$5F,$B7,$C3
	.dc.b	$5F,$D3,$7B,$1F,$A2,$7E,$FF,$0F,$E8,$DF,$B8,$77,$3C,$75,$D9,$F1
	.dc.b	$61,$FC,$1F,$0F,$D9,$12,$3F,$12,$D7,$F6,$8D,$F2,$9E,$4F,$EB,$63
	.dc.b	$0D,$9E,$24,$38,$96,$78,$BF,$C5,$67,$8B,$4D,$4F,$37,$89,$27,$8C
	.dc.b	$AF,$FC,$65,$3C,$0E,$46,$EB,$95,$70,$9D,$CE,$4D,$C6,$6B,$C6,$75
	.dc.b	$2C,$F0,$8D,$4F,$79,$D2,$1F,$0C,$F1,$FF,$89,$7C,$0F,$F1,$F2,$4F
	.dc.b	$ED,$9A,$FF,$21,$6D,$F9,$DB,$8F,$29,$E0,$43,$F9,$47,$C3,$7F,$2B
	.dc.b	$F8,$2F,$E5,$9F,$07,$FC,$B7,$E1,$BF,$97,$7C,$37,$96,$FF,$CF,$BF
	.dc.b	$C0,$3F,$49,$BF,$0F,$FC,$CF,$A7,$5A,$FC,$EA,$FE,$47,$5F,$E3,$85
	.dc.b	$F0,$DA,$D7,$FC,$CF,$0F,$86,$C5,$7F,$99,$61,$F0,$D8,$E3,$43,$E1
	.dc.b	$B0,$BE,$1A,$FA,$B8,$F8,$0C,$EF,$AD,$3D,$79,$FF,$1E,$E1,$F7,$BF
	.dc.b	$C8,$38,$FF,$A6,$72,$7C,$88,$FE,$C5,$CB,$F0,$79,$FE,$A4,$60,$F9
	.dc.b	$5C,$7F,$3D,$B8,$FC,$F7,$EA,$18,$38,$18,$24,$71,$6C,$7F,$51,$BF
	.dc.b	$A9,$1F,$CD,$39,$BE,$A3,$FC,$DB,$9B,$D8,$FC,$DD,$3E,$93,$F3,$FA
	.dc.b	$7E,$2F,$37,$C6,$E6,$F8,$FC,$DF,$23,$97,$E4,$F2,$7C,$8C,$3E,$B7
	.dc.b	$1F,$C8,$7E,$C6,$DF,$21,$FE,$DD,$C7,$F2,$1F,$EE,$1C,$7F,$21,$F7
	.dc.b	$73,$7E,$B7,$93,$E4,$74,$7E,$AB,$09,$5F,$A8,$6D,$7E,$8A,$3F,$58
	.dc.b	$DE,$4F,$D3,$C6,$3A,$7F,$55,$5C,$1B,$1C,$C9,$5C,$67,$4F,$EC,$89
	.dc.b	$1F,$B1,$9E,$31,$B8,$E3,$18,$24,$70,$8F,$13,$6B,$CA,$4A,$FE,$21
	.dc.b	$CD,$F8,$69,$F6,$57,$F8,$6D,$FE,$6C,$FF,$88,$5A,$E4,$FF,$14,$38
	.dc.b	$7F,$8D,$19,$38,$1A,$21,$82,$1B,$21,$C8,$87,$F1,$E3,$25,$32,$FA
	.dc.b	$1B,$7F,$C8,$1D,$C6,$5E,$04,$70,$48,$E1,$1C,$A3,$D8,$FF,$2E,$E2
	.dc.b	$FE,$94,$FA,$8E,$B3,$C1,$E8,$D9,$E8,$72,$2C,$71,$3B,$14,$E2,$EE
	.dc.b	$39,$11,$C1,$23,$83,$D1,$F9,$67,$42,$9D,$0D,$C7,$37,$FD,$47,$1C
	.dc.b	$5F,$5B,$D5,$E2,$E5,$E8,$68,$E6,$ED,$C1,$C2,$3B,$36,$34,$F2,$21
	.dc.b	$E5,$34,$74,$3F,$0F,$3F,$41,$C9,$0E,$E4,$39,$C7,$9D,$E8,$64,$A6
	.dc.b	$A7,$CA,$E6,$38,$12,$33,$1B,$8F,$4B,$F8,$3F,$5B,$F6,$EF,$ED,$63
	.dc.b	$F5,$D8,$62,$3F,$62,$64,$86,$09,$8F,$EC,$DF,$DD,$CF,$26,$9C,$23
	.dc.b	$51,$FC,$04,$97,$FC,$02,$75,$5E,$1A,$FE,$27,$8F,$EF,$6B,$55,$E0
	.dc.b	$AD,$11,$EE,$64,$B5,$E0,$9E,$2D,$BF,$31,$E2,$2C,$6A,$BA,$46,$6B
	.dc.b	$CA,$74,$8F,$8D,$7E,$59,$E2,$49,$DB,$C5,$F2,$36,$FC,$51,$AB,$F0
	.dc.b	$3D,$49,$3B,$BF,$1C,$EA,$31,$3A,$8F,$8B,$7F,$C8,$FA,$58,$FF,$78
	.dc.b	$25,$7F,$81,$5A,$FF,$83,$9B,$8F,$F8,$2D,$B8,$FF,$82,$1B,$3F,$E0
	.dc.b	$77,$FC,$07,$F0,$DF,$E0,$4F,$86,$FF,$02,$FC,$37,$F8,$1B,$E1,$BF
	.dc.b	$C0,$FF,$0D,$FE,$08,$F8,$6F,$F0,$4F,$C3,$7F,$82,$BE,$1B,$FC,$17
	.dc.b	$F0,$DF,$E0,$CF,$86,$FF,$08,$7F,$00,$BE,$1A,$FC,$36,$3E,$1B,$3F
	.dc.b	$0D,$AF,$86,$DF,$AF,$D3,$93,$FC,$41,$B7,$81,$FF,$11,$70,$79,$1F
	.dc.b	$F1,$27,$17,$A1,$F8,$44,$3F,$C8,$7D,$9F,$E2,$6E,$0F,$34,$7F,$8A
	.dc.b	$DA,$FF,$8D,$1B,$CF,$FE,$31,$6F,$3F,$19,$E0,$43,$81,$67,$83,$63
	.dc.b	$83,$A2,$46,$9B,$1A,$8C,$11,$C1,$23,$FC,$79,$3E,$46,$9D,$CA,$74
	.dc.b	$2C,$F3,$6C,$F3,$EB,$FF,$21,$D7,$C2,$EA,$FE,$E1,$C4,$F8,$3F,$B2
	.dc.b	$4C,$3F,$CD,$D8,$FF,$A0,$A7,$51,$8C,$7F,$CE,$58,$A7,$32,$AE,$17
	.dc.b	$CE,$31,$7C,$2B,$75,$A9,$E8,$2F,$A5,$F1,$9D,$4F,$23,$54,$F7,$7F
	.dc.b	$23,$F0,$7F,$92,$F7,$BF,$27,$EF,$7E,$53,$6F,$CA,$FB,$BF,$96,$F8
	.dc.b	$6F,$F4,$8E,$2F,$84,$FC,$CF,$C3,$7E,$6B,$E1,$7F,$37,$F0,$BF,$9C
	.dc.b	$F8,$6F,$CE,$FC,$2F,$E7,$BE,$17,$FC,$DD,$FC,$03,$FE,$7F,$E2,$F2
	.dc.b	$72,$FC,$3D,$7B,$C7,$B2,$3E,$36,$FE,$48,$C6,$CE,$1F,$4C,$63,$8F
	.dc.b	$E0,$73,$7C,$7E,$6F,$F2,$2C,$FC,$07,$2D,$39,$B8,$C7,$E4,$C6,$08
	.dc.b	$79,$B0,$FA,$35,$D6,$33,$CD,$F3,$8F,$C0,$61,$E4,$8E,$18,$FA,$5E
	.dc.b	$59,$7B,$9E,$99,$FC,$CE,$3F,$97,$1E,$EE,$2F,$96,$FC,$0E,$1F,$96
	.dc.b	$FC,$2E,$0F,$96,$FD,$53,$E2,$F2,$7D,$97,$E3,$F2,$FD,$D7,$E4,$F3
	.dc.b	$7F,$10,$D7,$E0,$70,$7E,$4B,$9F,$F9,$57,$47,$EA,$78,$FF,$43,$87
	.dc.b	$EA,$5A,$FE,$A5,$BC,$5F,$A9,$6F,$1F,$EA,$7C,$95,$F3,$6D,$F3,$6B
	.dc.b	$CD,$B7,$CD,$B5,$CE,$B9,$12,$B9,$35,$E4,$DA,$E4,$E7,$0F,$6C,$FE
	.dc.b	$44,$FE,$06,$7F,$B4,$70,$F6,$B7,$0F,$2B,$6B,$83,$C0,$FE,$32,$64
	.dc.b	$E9,$3C,$9C,$90,$E0,$47,$4D,$9E,$EE,$E3,$24,$8C,$C6,$E7,$C7,$39
	.dc.b	$6C,$EA,$31,$3E,$48,$C1,$27,$64,$8D,$35,$FD,$B4,$F2,$6C,$FD,$8F
	.dc.b	$2F,$FB,$7C,$78,$CE,$24,$37,$5C,$DF,$8C,$60,$D1,$82,$1F,$06,$7C
	.dc.b	$CE,$8C,$47,$36,$99,$3C,$47,$52,$1D,$0C,$96,$BD,$8D,$9E,$EF,$03
	.dc.b	$A8,$AF,$7C,$F0,$8C,$9C,$6B,$C8,$F6,$32,$43,$04,$8E,$4D,$AF,$0B
	.dc.b	$CC,$F6,$46,$89,$1A,$8E,$91,$89,$47,$FA,$1A,$74,$F5,$8F,$2C,$EE
	.dc.b	$3A,$CE,$A3,$11,$F0,$4E,$06,$A7,$8C,$EC,$8E,$09,$1B,$21,$CE,$7F
	.dc.b	$84,$D7,$CA,$AF,$DE,$CF,$F7,$02,$4F,$F0,$A6,$C6,$A3,$04,$70,$4A
	.dc.b	$FE,$01,$8F,$F0,$7A,$FE,$0A,$43,$F8,$29,$71,$FE,$0A,$D8,$E0,$E8
	.dc.b	$91,$A7,$18,$FF,$0A,$73,$3C,$27,$44,$70,$48,$F5,$D7,$99,$A7,$84
	.dc.b	$A7,$42,$CF,$36,$CF,$3C,$3E,$15,$7F,$1D,$E4,$C6,$3F,$0E,$FA,$56
	.dc.b	$9B,$19,$7B,$46,$5B,$19,$C3,$C1,$39,$6C,$67,$0E,$F8,$7A,$EB,$51
	.dc.b	$F2,$B1,$63,$FE,$20,$E7,$FF,$0B,$65,$FD,$F5,$AF,$F8,$09,$B9,$2F
	.dc.b	$86,$1F,$E0,$82,$57,$16,$BC,$5B,$3C,$63,$FC,$49,$5D,$2B,$9C,$71
	.dc.b	$3C,$5E,$5F,$F8,$53,$DD,$FF,$0A,$FC,$07,$F8,$5B,$E0,$FF,$C2,$FF
	.dc.b	$0D,$FE,$18,$F7,$FF,$C3,$3F,$03,$FE,$1A,$F8,$0F,$F0,$DF,$C3,$7F
	.dc.b	$87,$3E,$13,$FC,$3B,$F0,$DF,$E1,$EF,$85,$FF,$0F,$FC,$37,$F8,$83
	.dc.b	$E1,$BF,$C4,$3E,$CF,$F8,$8B,$E1,$7A,$71,$7C,$3D,$BC,$0F,$C5,$E0
	.dc.b	$F2,$3E,$63,$D0,$7B,$CF,$DF,$F4,$7C,$E7,$DF,$ED,$FF,$44,$E6,$34
	.dc.b	$91,$FD,$5A,$32,$D8,$CB,$FD,$02,$32,$D8,$CB,$FD,$9E,$32,$D8,$CB
	.dc.b	$F7,$DA,$F9,$0E,$C4,$3A,$1E,$3E,$1F,$F2,$2F,$C2,$FF,$91,$BD,$DF
	.dc.b	$F2,$3F,$C2,$CA,$FF,$22,$7A,$BF,$75,$FF,$24,$AF,$85,$FE,$43,$18
	.dc.b	$F0,$6F,$6F,$E0,$FB,$96,$F8,$5C,$F7,$FF,$9D,$21,$F0,$BF,$E5,$6D
	.dc.b	$FD,$36,$E0,$F7,$E7,$ED,$3D,$47,$C1,$29,$93,$FA,$47,$C0,$D7,$47
	.dc.b	$C0,$21,$82,$1B,$21,$F2,$4D,$98,$9F,$25,$79,$AB,$CB,$3B,$39,$94
	.dc.b	$E0,$EA,$7F,$95,$AE,$0F,$87,$3F,$E8,$BC,$5F,$0B,$8B,$F9,$3F,$0B
	.dc.b	$65,$F0,$B7,$E1,$B1,$F0,$D9,$E2,$D6,$DB,$DF,$E6,$69,$F1,$8F,$21
	.dc.b	$33,$F5,$B9,$6B,$F1,$EB,$FB,$8E,$DE,$92,$6F,$F0,$49,$3C,$63,$2D
	.dc.b	$8C,$BF,$28,$F0,$46,$9C,$57,$C4,$DF,$DD,$BF,$BF,$7F,$81,$B7,$BE
	.dc.b	$FF,$27,$AF,$BD,$FC,$AE,$BE,$F7,$F8,$F6,$7E,$07,$E9,$E5,$EC,$6E
	.dc.b	$7E,$1D,$3D,$B5,$B9,$FA,$ED,$7C,$ED,$AE,$51,$A3,$DC,$43,$D7,$4E
	.dc.b	$6F,$B1,$F0,$DF,$AD,$9F,$E6,$5C,$DF,$AD,$7F,$57,$C7,$F7,$79,$A4
	.dc.b	$7E,$C3,$97,$F0,$1E,$D7,$DA,$8E,$BA,$78,$F3,$F9,$51,$F9,$ED,$7F
	.dc.b	$C4,$FC,$CF,$86,$FF,$2B,$B6,$7F,$CC,$ED,$7F,$C9,$ED,$7F,$CE,$8D
	.dc.b	$7F,$6D,$1C,$88,$F2,$6B,$C9,$AF,$26,$BC,$8E,$04,$38,$14,$F0,$FB
	.dc.b	$9F,$9F,$F8,$6F,$D0,$7B,$DF,$A1,$F8,$69,$5F,$C5,$3C,$FF,$D2,$7C
	.dc.b	$37,$E9,$7E,$1B,$F5,$1F,$C0,$2F,$87,$BF,$0F,$FA,$9D,$7C,$78,$FE
	.dc.b	$BD,$BD,$3F,$C5,$9B,$BF,$C1,$9D,$1F,$E3,$EE,$0F,$9B,$3A,$24,$7A
	.dc.b	$4F,$71,$EC,$21,$F0,$CD,$1D,$0C,$9C,$0D,$18,$34,$60,$87,$F2,$9E
	.dc.b	$0F,$8C,$FF,$BE,$72,$FE,$5E,$DF,$87,$F8,$7F,$0F,$FC,$27,$7F,$8B
	.dc.b	$F0,$FF,$1E,$B8,$7C,$27,$C8,$F6,$FE,$34,$79,$B6,$F8,$8D,$DB,$E2
	.dc.b	$46,$23,$CC,$DD,$BE,$2C,$6F,$6F,$8A,$EA,$B8,$92,$33,$1E,$C3,$DE
	.dc.b	$7C,$32,$18,$25,$78,$5E,$0F,$F4,$ED,$BE,$1C,$3C,$7F,$EA,$5B,$FC
	.dc.b	$48,$C7,$67,$F9,$E2,$D2,$E0,$F8,$F5,$F8,$D3,$E6,$6D,$F1,$DE,$22
	.dc.b	$B8,$ED,$F2,$63,$FB,$3E,$7E,$C7,$D3,$3E,$76,$E9,$F0,$5F,$C6,$78
	.dc.b	$18,$AE,$11,$EE,$9D,$3F,$26,$33,$5C,$E3,$51,$89,$DC,$7C,$4C,$9E
	.dc.b	$E7,$E9,$3E,$13,$F5,$73,$FD,$F3,$A3,$F5,$6F,$B3,$1F,$87,$EF,$7E
	.dc.b	$B2,$3F,$19,$EE,$7E,$B3,$A3,$F5,$71,$EE,$3E,$F1,$0F,$31,$FB,$DE
	.dc.b	$4F,$F5,$CE,$6F,$7F,$FD,$D2,$D9,$FF,$7C,$B5,$FF,$70,$B5,$FF,$06
	.dc.b	$E3,$F5,$A7,$89,$23,$8B,$5E,$2D,$78,$CF,$37,$F0,$0D,$7E,$D3,$67
	.dc.b	$B4,$74,$21,$D0,$A7,$4F,$84,$FE,$1F,$F0,$9F,$C4,$3E,$1F,$F8,$8F
	.dc.b	$C2,$FF,$12,$F8,$7F,$E2,$7F,$0D,$FC,$53,$E1,$FF,$8A,$FC,$3F,$F7
	.dc.b	$DF,$E0,$1F,$F1,$4E,$7F,$45,$FF,$34,$3E,$4D,$BD,$4F,$A1,$AE,$4F
	.dc.b	$19,$93,$CF,$BF,$D0,$8D,$ED,$EC,$9F,$51,$0F,$8B,$1A,$DF,$E1,$11
	.dc.b	$C1,$23,$B9,$E1,$39,$1E,$12,$9F,$A5,$B7,$C3,$7B,$FE,$1B,$E3,$4F
	.dc.b	$FB,$A3,$BF,$FC,$8D,$4F,$6B,$E4,$4E,$7C,$0F,$DB,$4B,$E1,$BE,$47
	.dc.b	$B9,$F2,$7E,$1B,$E5,$5F,$CB,$F5,$3F,$A0,$EB,$E7,$8F,$1E,$FF,$05
	.dc.b	$F7,$9E,$A7,$E2,$70,$7C,$78,$C7,$07,$C1,$6D,$73,$78,$9D,$23,$4D
	.dc.b	$8D,$3F,$0C,$D9,$0F,$5F,$07,$C3,$7F,$9C,$73,$7E,$0F,$7F,$83,$1F
	.dc.b	$83,$D7,$DE,$FC,$33,$E6,$69,$F0,$1C,$66,$DB,$FC,$21,$1F,$D0,$F2
	.dc.b	$F1,$BF,$5C,$F1,$14,$F3,$1F,$67,$4F,$84,$FA,$CF,$82,$43,$04,$39
	.dc.b	$D7,$CD,$79,$9D,$48,$E5,$BA,$7E,$3B,$A3,$F5,$31,$EA,$C1,$F0,$DF
	.dc.b	$AA,$8F,$EB,$9D,$1F,$AA,$7F,$AF,$73,$7E,$A9,$E1,$F0,$DF,$AB,$CB
	.dc.b	$F0,$31,$E7,$3C,$B9,$7D,$66,$E9,$FE,$98,$D1,$E1,$FE,$3B,$E1,$BF
	.dc.b	$68,$4A,$F2,$35,$FD,$B3,$79,$FF,$32,$DE,$7E,$53,$C0,$87,$02,$CF
	.dc.b	$06,$C7,$07,$44,$8D,$36,$35,$18,$23,$82,$47,$EC,$67,$C8,$D3,$B9
	.dc.b	$4E,$85,$9E,$6D,$9E,$79,$7E,$36,$BF,$D7,$FF,$01,$FE,$E1,$8F,$C8
	.dc.b	$7C,$37,$F1,$B2,$5F,$DB,$6B,$FC,$75,$B8,$7F,$00,$6E,$1F,$BE,$6C
	.dc.b	$FF,$09,$6D,$7E,$77,$E1,$BF,$94,$7C,$37,$F2,$9F,$86,$FE,$55,$F0
	.dc.b	$DF,$CA,$FE,$1B,$F9,$67,$B3,$FC,$B7,$E0,$FF,$97,$7C,$37,$F2,$FF
	.dc.b	$7F,$FD,$59,$FC,$03,$FE,$DB,$E0,$FF,$6B,$3F,$ED,$BD,$FF,$56,$B9
	.dc.b	$37,$EF,$F0,$F8,$6E,$3F,$0D,$CB,$E1,$BB,$E1,$F9,$9D,$BC,$6F,$E6
	.dc.b	$F8,$3C,$EF,$E7,$78,$BD,$7D,$EE,$5F,$97,$E7,$7C,$1A,$F9,$04,$3E
	.dc.b	$41,$77,$F9,$0D,$E7,$E0,$E8,$91,$A6,$C6,$A3,$F9,$7E,$1E,$49,$FE
	.dc.b	$A1,$8F,$A1,$FE,$A5,$97,$B1,$FE,$A9,$C7,$F2,$9F,$DB,$71,$7C,$AC
	.dc.b	$7E,$5F,$C1,$FC,$CE,$2F,$8F,$5F,$2B,$87,$E3,$BF,$33,$83,$A6,$BF
	.dc.b	$D0,$33,$F0,$3F,$D0,$B4,$F2,$3E,$6A,$F7,$1F,$83,$03,$A7,$F2,$5C
	.dc.b	$2E,$AF,$CF,$F5,$7E,$5E,$BF,$3E,$D7,$F3,$ED,$BF,$CF,$B7,$6F,$CF
	.dc.b	$DF,$3A,$E2,$D7,$8B,$6B,$8B,$67,$8D,$F8,$A3,$15,$FA,$B2,$5F,$76
	.dc.b	$C6,$A3,$04,$70,$49,$E8,$49,$F2,$BF,$8A,$7F,$78,$7F,$8A,$23,$D0
	.dc.b	$D8,$FF,$07,$BF,$C3,$CC,$18,$9F,$8C,$E0,$FF,$0A,$B8,$78,$10,$EE
	.dc.b	$57,$4E,$08,$E0,$93,$E4,$8E,$0D,$76,$D8,$DC,$FA,$27,$2D,$8C,$C7
	.dc.b	$A6,$78,$12,$74,$D8,$D4,$60,$8E,$0F,$F3,$99,$D8,$C1,$F0,$A3,$93
	.dc.b	$63,$8B,$4D,$9F,$24,$87,$B9,$DB,$83,$B1,$C4,$86,$67,$DA,$E0,$E4
	.dc.b	$76,$29,$93,$99,$C6,$7C,$ED,$9F,$3E,$BF,$D8,$23,$67,$23,$47,$53
	.dc.b	$24,$AF,$1B,$A3,$C0,$E1,$E4,$43,$85,$78,$9C,$46,$DC,$D7,$C4,$73
	.dc.b	$3F,$AF,$AF,$35,$6C,$91,$CE,$78,$9E,$78,$CD,$73,$8C,$9E,$69,$E1
	.dc.b	$3A,$8E,$11,$82,$4F,$19,$D4,$F1,$8C,$4F,$97,$AF,$F7,$F3,$FB,$22
	.dc.b	$1F,$C1,$A7,$F6,$CD,$33,$8F,$F0,$88,$CB,$63,$38,$F4,$78,$12,$34
	.dc.b	$D8,$D4,$60,$FF,$03,$E5,$FC,$92,$BA,$CE,$89,$5F,$C4,$23,$2D,$3C
	.dc.b	$25,$3B,$16,$BB,$4F,$2C,$7F,$8B,$57,$26,$D7,$26,$BE,$AB,$F2,$39
	.dc.b	$C7,$F8,$E3,$6B,$94,$62,$3C,$35,$CB,$0F,$0C,$78,$F0,$F1,$56,$E7
	.dc.b	$35,$B9,$CD,$F4,$25,$FF,$4A,$C3,$FC,$35,$CE,$F8,$6F,$F0,$29,$27
	.dc.b	$FC,$06,$D7,$FD,$CD,$1B,$72,$43,$31,$BC,$7F,$C1,$AD,$C3,$FA,$CE
	.dc.b	$2B,$E5,$7D,$3E,$1F,$FC,$2B,$F0,$FF,$E1,$6D,$7F,$C2,$FF,$0F,$FE
	.dc.b	$18,$F8,$7F,$F0,$CF,$B5,$FE,$1A,$F7,$BF,$C3,$7F,$0D,$FE,$1C,$F8
	.dc.b	$6F,$F0,$EF,$B5,$FE,$1E,$F8,$2F,$F0,$FF,$C3,$7F,$88,$3E,$1B,$FB
	.dc.b	$4E,$FF,$EF,$FE,$87,$B9,$8E,$67,$2F,$9A,$5F,$0D,$98,$C7,$C0,$F8
	.dc.b	$B0,$FF,$13,$F0,$F9,$9F,$F1,$4F,$1F,$A9,$FF,$15,$F2,$FC,$15,$CF
	.dc.b	$F0,$7A,$3F,$AC,$3F,$E2,$FF,$84,$FE,$79,$A1,$A4,$AF,$EA,$D1,$96
	.dc.b	$C6,$5F,$EA,$F1,$96,$C6,$5F,$83,$3D,$DE,$24,$36,$53,$73,$CD,$B3
	.dc.b	$CD,$F8,$CF,$F5,$3E,$97,$37,$C7,$9F,$7D,$3E,$1B,$18,$7F,$93,$7E
	.dc.b	$16,$CB,$E1,$6F,$C3,$63,$E1,$A6,$3F,$EB,$EA,$7C,$37,$0F,$86,$E3
	.dc.b	$E6,$72,$F5,$39,$F0,$79,$70,$F0,$B7,$5F,$81,$1E,$E2,$63,$FC,$B8
	.dc.b	$F7,$3E,$C3,$E1,$14,$E0,$7C,$23,$27,$62,$C7,$67,$47,$03,$91,$82
	.dc.b	$18,$21,$CC,$E0,$7E,$23,$83,$E6,$BF,$0F,$DA,$F9,$3B,$FC,$AD,$FE
	.dc.b	$5F,$C3,$7C,$CF,$7F,$E6,$FB,$7F,$3B,$83,$E4,$63,$F2,$FE,$1B,$E4
	.dc.b	$3F,$37,$6F,$8D,$B7,$C2,$8F,$11,$E6,$29,$F0,$DF,$91,$1C,$09,$1F
	.dc.b	$0A,$39,$6D,$F0,$E3,$DA,$D7,$99,$1C,$12,$38,$92,$7A,$CF,$F8,$D3
	.dc.b	$6F,$86,$FE,$EF,$6F,$86,$FE,$1F,$5F,$86,$FE,$27,$4F,$86,$FE,$2F
	.dc.b	$3F,$86,$FF,$03,$CF,$E1,$BF,$C1,$71,$EE,$FC,$1C,$FC,$91,$FA,$FC
	.dc.b	$FE,$0C,$FB,$E3,$6F,$BA,$3D,$0E,$48,$7D,$C9,$E6,$E0,$ED,$1F,$03
	.dc.b	$E1,$BF,$4B,$CB,$FA,$6F,$77,$F4,$FF,$0D,$FA,$6B,$FD,$27,$37,$E9
	.dc.b	$97,$81,$FC,$69,$FE,$1F,$FC,$03,$FC,$81,$CF,$81,$FC,$CA,$3F,$89
	.dc.b	$C7,$EB,$7A,$3F,$6B,$0F,$84,$FD,$5C,$F9,$4F,$66,$FF,$4B,$6F,$F0
	.dc.b	$AD,$3E,$1F,$FC,$A8,$DA,$FF,$36,$35,$FF,$35,$B5,$FE,$26,$D7,$C9
	.dc.b	$1C,$88,$F2,$6B,$C9,$AF,$26,$BC,$8E,$04,$38,$14,$F0,$FC,$2F,$C0
	.dc.b	$F8,$5F,$83,$F0,$FF,$0B,$E1,$BE,$1F,$C3,$FC,$4F,$87,$F8,$BF,$0F
	.dc.b	$F1,$BD,$6F,$8F,$F0,$1F,$23,$E0,$3E,$4F,$C3,$7C,$AD,$BC,$7B,$7D
	.dc.b	$26,$F5,$B8,$3E,$04,$E4,$F3,$70,$FC,$CA,$DC,$EA,$B8,$46,$09,$5E
	.dc.b	$B3,$E0,$1E,$E2,$1F,$14,$D1,$DC,$C9,$C0,$D1,$83,$F9,$CF,$5F,$C5
	.dc.b	$7F,$DA,$3C,$FF,$8D,$AF,$57,$B7,$F8,$8F,$86,$FC,$4F,$BB,$F8,$AF
	.dc.b	$86,$F8,$F8,$70,$EB,$FC,$87,$17,$E3,$BE,$0F,$E4,$4E,$4F,$CB,$F0
	.dc.b	$7C,$97,$CF,$9F,$C0,$6E,$FF,$0C,$9B,$FF,$7C,$E0,$F9,$13,$A8,$F8
	.dc.b	$AF,$AE,$B9,$36,$B8,$BB,$21,$99,$E2,$FC,$43,$FB,$1F,$07,$C3,$7F
	.dc.b	$56,$7E,$BB,$7F,$84,$FA,$EF,$EF,$6D,$F0,$E3,$18,$34,$F8,$B1,$CA
	.dc.b	$BF,$09,$A7,$C4,$8C,$53,$4F,$89,$1E,$53,$CE,$79,$4F,$C0,$E9,$F0
	.dc.b	$9D,$10,$D1,$F8,$12,$DF,$C4,$6D,$7C,$47,$81,$F8,$52,$1C,$4B,$1C
	.dc.b	$5F,$E1,$9D,$3F,$C2,$BE,$1B,$F5,$71,$F0,$3A,$3F,$56,$FC,$2E,$7F
	.dc.b	$D6,$3F,$13,$9F,$F5,$6F,$C6,$E7,$FE,$27,$F0,$9F,$02,$3C,$BC,$7F
	.dc.b	$51,$F0,$E9,$FC,$37,$5F,$F5,$3F,$C0,$7F,$21,$9F,$F7,$CB,$5F,$F1
	.dc.b	$0B,$5F,$F2,$9B,$5F,$F1,$AB,$76,$34,$9C,$BF,$EF,$16,$CF,$F2,$D3
	.dc.b	$8C,$64,$8E,$67,$9C,$62,$3C,$AD,$7C,$AD,$7C,$BF,$09,$FE,$28,$A5
	.dc.b	$D3,$1F,$0D,$9F,$85,$D7,$C3,$6F,$E1,$B8,$7A,$DC,$7D,$7E,$5F,$0B
	.dc.b	$CF,$D9,$E9,$B7,$7D,$BD,$27,$9F,$5F,$6B,$E6,$31,$1F,$AF,$9F,$92
	.dc.b	$F1,$DF,$DB,$3A,$21,$C4,$87,$B8,$C9,$4C,$9D,$0C,$94,$C9,$E6,$2B
	.dc.b	$E0,$6D,$37,$F9,$1F,$0B,$F2,$7D,$BF,$95,$F0,$BF,$2F,$DC,$F9,$9F
	.dc.b	$0B,$F1,$F1,$FD,$C6,$3F,$E9,$4F,$1B,$F8,$F2,$F8,$5F,$A5,$F0,$BF
	.dc.b	$4F,$E1,$7E,$4D,$F9,$8F,$21,$EB,$21,$E9,$3D,$DD,$1F,$96,$E0,$F9
	.dc.b	$18,$7C,$E9,$D9,$C8,$E0,$43,$04,$E0,$F8,$0F,$06,$BB,$6B,$B8,$EC
	.dc.b	$E5,$F8,$CF,$C0,$DF,$E3,$3F,$0B,$6F,$8C,$FC,$4D,$7E,$33,$FC,$5E
	.dc.b	$BE,$36,$1F,$AA,$D3,$E2,$CF,$CA,$CF,$E2,$BF,$52,$7D,$E4,$47,$E7
	.dc.b	$38,$7C,$91,$98,$E0,$7C,$DC,$FD,$64,$C3,$D7,$1E,$D9,$E0,$FC,$38
	.dc.b	$CC,$EF,$E1,$7F,$51,$3F,$7B,$5F,$D4,$BF,$80,$E8,$FD,$4B,$F8,$2F
	.dc.b	$85,$FF,$4B,$FF,$00,$FF,$B1,$A3,$F0,$BC,$FF,$AA,$7F,$0F,$A7,$B9
	.dc.b	$F0,$69,$E2,$7D,$67,$BB,$14,$FF,$B8,$AF,$FD,$9B,$3F,$C1,$5A,$FF
	.dc.b	$8D,$F3,$46,$23,$FD,$72,$D7,$97,$C2,$FF,$11,$7F,$D9,$B1,$96,$C6
	.dc.b	$67,$9B,$67,$51,$89,$F1,$4E,$48,$E0,$87,$96,$78,$92,$37,$19,$F8
	.dc.b	$5F,$93,$E9,$C8,$FE,$C9,$5F,$ED,$5E,$6F,$F0,$53,$FE,$C5,$8C,$FC
	.dc.b	$17,$D0,$F7,$BE,$8F,$C2,$FD,$2F,$67,$E9,$FB,$DF,$53,$E1,$7E,$AF
	.dc.b	$B7,$F5,$B6,$F0,$E5,$F1,$9A,$FC,$63,$D3,$C3,$F4,$9F,$8A,$7B,$0E
	.dc.b	$05,$38,$1E,$F8,$E6,$F0,$30,$6C,$C1,$C8,$C9,$0C,$10,$EA,$53,$89
	.dc.b	$F8,$5D,$FF,$22,$7E,$4B,$E1,$7F,$27,$F0,$5F,$94,$F8,$0F,$CA,$EF
	.dc.b	$F9,$6F,$85,$F8,$F8,$7C,$EF,$83,$FC,$CF,$C2,$FE,$6B,$E1,$3F,$37
	.dc.b	$ED,$FC,$7A,$F4,$74,$7E,$05,$F3,$1E,$43,$E0,$EF,$F1,$E3,$51,$E7
	.dc.b	$6B,$B3,$D2,$47,$2D,$E6,$FC,$33,$E0,$6B,$E0,$35,$C1,$F3,$A7,$89
	.dc.b	$25,$C1,$F1,$2B,$F1,$7C,$1F,$11,$FC,$6E,$FF,$11,$FC,$7E,$DF,$11
	.dc.b	$FF,$17,$5B,$6F,$8B,$1F,$94,$D7,$E2,$BF,$96,$CF,$C5,$5E,$68,$F8
	.dc.b	$0F,$ED,$B4,$F8,$8E,$63,$E1,$69,$E7,$AE,$24,$8F,$80,$D7,$E0,$46
	.dc.b	$DC,$9F,$17,$E1,$3F,$50,$FA,$3A,$6F,$4F,$EA,$3E,$13,$F9,$4C,$FF
	.dc.b	$92,$7D,$1F,$F5,$F4,$FE,$B3,$AF,$2F,$F9,$DE,$3F,$D7,$F3,$F0,$BA
	.dc.b	$BF,$56,$FE,$D3,$A7,$F5,$6F,$C1,$3C,$67,$9B,$5F,$AF,$B7,$E0,$EF
	.dc.b	$FD,$8D,$3F,$E1,$D6,$BF,$C6,$5A,$F9,$B5,$FF,$63,$7C,$27,$F1,$39
	.dc.b	$CC,$FF,$42,$9C,$BC,$E7,$44,$70,$43,$B3,$96,$99,$3C,$13,$D9,$4F
	.dc.b	$67,$64,$3F,$96,$ED,$F2,$3B,$BC,$95,$FE,$D9,$D3,$E5,$39,$AE,$6B
	.dc.b	$D9,$C6,$1F,$EE,$36,$C3,$E1,$2F,$C2,$63,$DA,$CF,$C2,$6B,$DB,$DF
	.dc.b	$BF,$C3,$DC,$E3,$C1,$ED,$D7,$D6,$D7,$CA,$FC,$1E,$1F,$A8,$E6,$3C
	.dc.b	$AD,$DB,$E0,$47,$23,$E3,$1C,$8A,$70,$2B,$C1,$D1,$F1,$CA,$73,$30
	.dc.b	$71,$2B,$FD,$6B,$87,$E1,$FB,$9F,$13,$E1,$3E,$2F,$0C,$BF,$F5,$47
	.dc.b	$0F,$C7,$6F,$07,$CA,$F7,$FE,$5F,$C1,$7C,$CF,$84,$F9,$BF,$07,$F1
	.dc.b	$EF,$8E,$DE,$FC,$7F,$01,$B7,$A6,$7C,$87,$9F,$7F,$8E,$FB,$A7,$64
	.dc.b	$AE,$44,$8D,$35,$F6,$C6,$DC,$9F,$23,$A1,$C1,$F0,$23,$DB,$BF,$BD
	.dc.b	$FD,$C6,$FF,$86,$E6,$FC,$2C,$7C,$2D,$FD,$EF,$C4,$DB,$DE,$FC,$6E
	.dc.b	$2F,$C2,$BE,$03,$C7,$A7,$E0,$FC,$CF,$AB,$10,$F8,$25,$DF,$E2,$13
	.dc.b	$6F,$54,$72,$3A,$1F,$74,$87,$17,$6E,$0E,$A7,$E9,$3E,$13,$F4,$31
	.dc.b	$23,$EE,$BF,$A4,$F8,$4F,$D2,$FC,$27,$E9,$BE,$13,$F4,$95,$F8,$2E
	.dc.b	$AF,$D2,$3F,$84,$E9,$FD,$6F,$F0,$0F,$FA,$F7,$B7,$F8,$E6,$DF,$4E
	.dc.b	$BC,$5A,$FA,$E3,$FD,$2D,$1F,$C0,$2B,$FC,$BE,$FF,$8B,$5A,$FF,$15
	.dc.b	$6B,$FC,$4D,$AF,$1F,$7F,$FC,$55,$18,$A3,$49,$1A,$24,$FF,$B0,$63
	.dc.b	$2D,$8C,$BE,$36,$BE,$36,$BE,$3E,$8F,$83,$F0,$92,$3F,$CD,$38,$BC
	.dc.b	$0F,$47,$AD,$F1,$8F,$8F,$F0,$9F,$23,$E1,$3E,$4F,$C2,$7C,$AF,$63
	.dc.b	$E5,$FC,$27,$CC,$E8,$99,$F4,$CB,$E1,$6D,$E2,$8F,$84,$7A,$48,$7A
	.dc.b	$4F,$F2,$F4,$F0,$72,$7A,$CA,$64,$E0,$43,$45,$34,$60,$86,$08,$79
	.dc.b	$0E,$05,$36,$53,$E5,$1E,$D3,$F0,$9F,$03,$F8,$CF,$84,$F9,$11,$9E
	.dc.b	$2F,$BB,$B7,$E4,$3E,$0B,$F2,$3F,$09,$F2,$AB,$5C,$8D,$BE,$5C,$63
	.dc.b	$3F,$96,$B8,$BE,$B3,$E3,$26,$5F,$8A,$D7,$E7,$36,$3E,$6F,$27,$F8
	.dc.b	$C7,$1F,$C4,$C7,$BA,$74,$D8,$CB,$F0,$23,$2D,$8C,$BF,$22,$32,$D8
	.dc.b	$CB,$E4,$7F,$39,$3F,$B0,$DB,$E3,$4F,$C0,$D7,$E3,$BE,$DD,$FF,$51
	.dc.b	$B7,$C6,$8F,$8B,$C3,$F8,$F7,$E3,$F0,$FC,$4D,$7E,$1E,$9F,$AA,$DB
	.dc.b	$F7,$BB,$7C,$2D,$BC,$73,$E6,$C3,$EF,$6D,$F0,$EB,$44,$8F,$C6,$9E
	.dc.b	$97,$0F,$C5,$6B,$F0,$DA,$FC,$1F,$83,$FD,$63,$EB,$3F,$84,$72,$FF
	.dc.b	$10,$F8,$3F,$E2,$3F,$09,$FA,$C9,$FC,$0F,$37,$F1,$4F,$84,$FE,$2B
	.dc.b	$F0,$7F,$C5,$BE,$0F,$FC,$07,$FC,$03,$FE,$C9,$BE,$9B,$23,$FD,$7F
	.dc.b	$C3,$F0,$B8,$3D,$E4,$DF,$F8,$2F,$07,$F9,$C6,$BF,$CC,$38,$7F,$90
	.dc.b	$0F,$F1,$43,$87,$FD,$90,$D7,$FC,$60,$D7,$7E,$EF,$FB,$0E,$31,$1F
	.dc.b	$E2,$46,$E3,$E5,$9D,$12,$7C,$06,$DC,$3E,$06,$BE,$06,$BE,$03,$8C
	.dc.b	$75,$6C,$F8,$0F,$F6,$27,$37,$B3,$DC,$F9,$27,$CA,$F8,$0F,$97,$EB
	.dc.b	$FC,$CF,$84,$F9,$BE,$C4,$C7,$F7,$18,$BE,$13,$E8,$FB,$3F,$4B,$E1
	.dc.b	$3E,$9F,$C0,$FD,$4D,$FC,$B8,$7A,$CF,$1B,$8D,$FC,$D0,$E0,$FA,$53
	.dc.b	$9C,$BD,$0D,$24,$F2,$21,$F1,$4A,$7C,$52,$9F,$14,$D1,$0D,$14,$FB
	.dc.b	$FF,$09,$F8,$CF,$82,$FC,$6F,$C1,$7C,$59,$F8,$3C,$3F,$15,$F8,$7B
	.dc.b	$FC,$57,$F2,$38,$FC,$57,$E3,$ED,$F1,$5F,$93,$B7,$C5,$7C,$A7,$C5
	.dc.b	$3C,$45,$3C,$BA,$7C,$47,$7A,$FC,$87,$D2,$43,$DC,$53,$91,$82,$18
	.dc.b	$23,$CD,$CB,$5C,$DF,$D2,$7C,$2B,$4F,$8D,$1F,$84,$CF,$E3,$3F,$0B
	.dc.b	$3F,$8C,$FC,$4C,$FE,$37,$37,$F1,$4C,$3E,$34,$FC,$8D,$3E,$33,$F2
	.dc.b	$B4,$F8,$CF,$C4,$CF,$C8,$4A,$F2,$47,$E2,$7B,$FD,$31,$88,$F8,$53
	.dc.b	$BD,$FF,$13,$1C,$E7,$4F,$08,$CB,$C2,$32,$FC,$2F,$80,$FD,$3B,$FA
	.dc.b	$2E,$5F,$E2,$3F,$09,$FA,$88,$FD,$2F,$37,$F1,$4F,$83,$FE,$2B,$F0
	.dc.b	$9F,$C5,$BD,$BF,$E2,$FF,$07,$FE,$9A,$FE,$01,$FF,$65,$F6,$7F,$0E
	.dc.b	$EA,$F7,$E5,$E1,$DF,$C6,$FE,$67,$A5,$C5,$FE,$BB,$C3,$FD,$9F,$3F
	.dc.b	$EC,$16,$B9,$E2,$FE,$16,$FF,$AD,$23,$2D,$8C,$88,$CB,$4C,$B8,$23
	.dc.b	$82,$47,$F1,$08,$CB,$63,$2F,$F2,$B7,$A3,$92,$19,$8F,$F6,$EF,$2F
	.dc.b	$F5,$87,$FD,$89,$97,$BF,$DC,$F9,$1E,$E7,$C9,$F8,$1F,$95,$F0,$9F
	.dc.b	$2F,$D6,$F9,$9F,$09,$33,$FF,$07,$64,$F5,$FE,$87,$BB,$F4,$7E,$13
	.dc.b	$E9,$7C,$0F,$D3,$E0,$F3,$E1,$F2,$DA,$FC,$B3,$FC,$62,$7C,$93,$D2
	.dc.b	$6C,$A6,$CF,$63,$C9,$E0,$60,$D9,$83,$91,$92,$18,$21,$D4,$A7,$13
	.dc.b	$F0,$9F,$03,$F8,$DF,$84,$FC,$77,$C1,$7C,$89,$F8,$7B,$FE,$47,$E1
	.dc.b	$3F,$25,$F0,$7F,$1E,$7F,$B4,$EC,$DF,$F2,$DB,$FC,$89,$F9,$7B,$7E
	.dc.b	$67,$E0,$BE,$44,$7A,$B6,$FB,$AF,$A0,$F8,$5D,$1F,$98,$7C,$C4,$8D
	.dc.b	$BE,$A6,$BE,$D2,$1B,$C7,$F1,$AF,$53,$11,$F3,$4E,$2E,$0F,$24,$7E
	.dc.b	$67,$7F,$89,$1F,$84,$DB,$E2,$3F,$03,$6F,$88,$FE,$23,$5F,$89,$CF
	.dc.b	$F9,$3D,$BE,$1C,$FE,$33,$8B,$F1,$4F,$E3,$B5,$F8,$6F,$97,$4F,$19
	.dc.b	$34,$3F,$EA,$83,$07,$A5,$F6,$C7,$C3,$6B,$F0,$A3,$6F,$B1,$E0,$47
	.dc.b	$07,$EC,$FE,$03,$F8,$77,$C1,$FF,$0F,$E8,$93,$FA,$4F,$83,$FD,$4B
	.dc.b	$FA,$6E,$7F,$E2,$9F,$06,$8F,$F3,$B7,$8D,$FE,$C3,$EC,$C6,$68,$FF
	.dc.b	$5F,$F0,$7F,$93,$9F,$EF,$55,$C3,$D9,$FD,$6B,$FB,$2A,$F3,$3F,$17
	.dc.b	$6F,$86,$FE,$B2,$3A,$7B,$DF,$C5,$2C,$D2,$4F,$F8,$7A,$32,$D8,$CB
	.dc.b	$FC,$3E,$32,$D8,$CB,$ED,$3E,$9C,$69,$C4,$F8,$9E,$84,$39,$14,$DC
	.dc.b	$71,$7E,$2B,$5E,$B3,$EB,$F8,$3F,$06,$3F,$1F,$D9,$B9,$F0,$F7,$AE
	.dc.b	$1E,$4F,$81,$B9,$FC,$FF,$7E,$E9,$FA,$BA,$FF,$2F,$AF,$81,$CF,$C1
	.dc.b	$EB,$E0,$B7,$F0,$7C,$3E,$0F,$8F,$C1,$F2,$F8,$3E,$7C,$7F,$13,$3F
	.dc.b	$4B,$5F,$49,$F5,$73,$F6,$6B,$F1,$63,$E1,$90,$F8,$73,$C6,$35,$18
	.dc.b	$3E,$41,$D8,$A7,$63,$91,$92,$99,$39,$18,$7A,$BF,$43,$E0,$FE,$2F
	.dc.b	$17,$C6,$F8,$3F,$8F,$C3,$F2,$38,$7E,$4F,$C1,$FC,$AE,$2F,$8F,$8F
	.dc.b	$E7,$B8,$BE,$7F,$F0,$0F,$FA,$2F,$9B,$FD,$52,$FF,$A2,$B9,$BE,$EF
	.dc.b	$17,$C9,$BF,$D2,$F0,$7B,$F8,$7E,$29,$1F,$DF,$36,$3D,$FC,$3E,$87
	.dc.b	$D8,$7B,$CF,$69,$0C,$10,$D9,$1E,$44,$8E,$D1,$F1,$5A,$FE,$5B,$87
	.dc.b	$E2,$BF,$BC,$E0,$F8,$AF,$EF,$B7,$F8,$AF,$F0,$0D,$BE,$2B,$F8,$ED
	.dc.b	$BE,$2F,$C5,$7C,$6F,$87,$FC,$5C,$FF,$06,$E4,$FC,$5B,$EE,$CF,$C2
	.dc.b	$FD,$BF,$27,$FA,$16,$BE,$7D,$BE,$41,$27,$83,$5F,$53,$5F,$88,$7C
	.dc.b	$1A,$EA,$DE,$AF,$B3,$E9,$7E,$9B,$AF,$ED,$D7,$DC,$F6,$BF,$53,$E9
	.dc.b	$FD,$EF,$7F,$EF,$FB,$5F,$A7,$CB,$F8,$FF,$77,$E9,$DF,$E4,$5D,$5F
	.dc.b	$82,$F8,$4F,$C2,$FC,$07,$BA,$7C,$A7,$A0,$FE,$35,$D9,$FE,$5C,$7F
	.dc.b	$85,$B5,$FE,$52,$D7,$FA,$1B,$78,$4D,$27,$C2,$7F,$8D,$9B,$3F,$EB
	.dc.b	$E6,$BF,$E2,$D6,$BE,$08,$E0,$47,$83,$5E,$0F,$5F,$81,$FD,$09,$FA
	.dc.b	$2F,$81,$FD,$1F,$B7,$FA,$4F,$83,$FD,$2F,$C2,$7E,$9B,$E1,$3F,$4F
	.dc.b	$F0,$7F,$A8,$F8,$3F,$D4,$FC,$1F,$EA,$BE,$0F,$F5,$7F,$07,$FC,$6B
	.dc.b	$F8,$07,$FC,$BF,$BF,$C0,$DF,$C4,$F9,$CF,$29,$0F,$F3,$E1,$B3,$06
	.dc.b	$CF,$49,$83,$E6,$BD,$1A,$64,$E4,$70,$21,$CC,$87,$03,$25,$32,$72
	.dc.b	$3E,$EE,$FF,$21,$FC,$3F,$93,$FD,$9A,$BF,$AE,$C3,$95,$39,$F5,$BE
	.dc.b	$27,$AD,$F3,$2B,$89,$F8,$4D,$FE,$6B,$8E,$56,$2F,$5B,$E6,$7C,$27
	.dc.b	$CD,$F7,$3E,$77,$C2,$7C,$FC,$3F,$8D,$EF,$F0,$DF,$41,$EB,$3D,$DE
	.dc.b	$37,$ED,$3A,$BF,$13,$19,$36,$4C,$FF,$2B,$5C,$DB,$1F,$35,$AF,$C8
	.dc.b	$38,$4F,$46,$CF,$18,$C4,$FE,$83,$87,$E5,$57,$C7,$E0,$F9,$53,$F8
	.dc.b	$9E,$0F,$8F,$3F,$0E,$DC,$FF,$93,$8C,$73,$FE,$3F,$7F,$8B,$1F,$27
	.dc.b	$81,$BF,$C6,$8C,$1F,$18,$F5,$61,$E4,$3F,$DA,$BB,$7C,$5D,$BD,$04
	.dc.b	$9D,$B5,$F8,$1A,$7C,$27,$F0,$ED,$7E,$63,$5F,$86,$72,$3F,$C2,$87
	.dc.b	$E2,$7D,$8F,$D4,$7B,$9F,$B3,$8F,$81,$D7,$FB,$37,$D9,$F0,$3F,$AD
	.dc.b	$F7,$7F,$5D,$F0,$3F,$B3,$9F,$8D,$D9,$FB,$37,$E4,$75,$FE,$CF,$E1
	.dc.b	$3E,$04,$7A,$37,$F8,$1B,$7F,$95,$78,$5F,$05,$FE,$C6,$6D,$79,$8F
	.dc.b	$F2,$DB,$87,$FD,$7C,$D7,$FC,$56,$D7,$F4,$91,$CC,$8F,$3E,$0F,$CE
	.dc.b	$C6,$23,$9B,$5E,$6D,$79,$9C,$67,$B3,$E4,$8D,$34,$D7,$C2,$7F,$13
	.dc.b	$F8,$3F,$E2,$9F,$09,$FC,$57,$E1,$3F,$8B,$7C,$27,$F1,$7F,$84,$FE
	.dc.b	$31,$F0,$9F,$E3,$2F,$E0,$1F,$F3,$16,$5F,$E7,$A7,$FC,$C3,$DF,$FE
	.dc.b	$64,$7F,$CC,$31,$F9,$93,$5F,$0B,$C7,$D6,$E5,$EB,$73,$F8,$2F,$E4
	.dc.b	$9C,$1E,$6E,$2F,$83,$87,$AE,$3E,$91,$9E,$2F,$6C,$EA,$7E,$29,$B2
	.dc.b	$18,$21,$F2,$0C,$94,$C9,$F2,$8A,$7C,$72,$BD,$DE,$07,$DC,$E2,$F9
	.dc.b	$4F,$EB,$78,$BE,$53,$FC,$5F,$47,$AD,$F2,$E3,$1F,$05,$F2,$FD,$EF
	.dc.b	$99,$F0,$FF,$37,$D7,$F9,$DE,$FF,$C7,$BE,$5C,$1F,$47,$DF,$F8,$F1
	.dc.b	$EA,$3D,$07,$AF,$C5,$FF,$06,$F0,$7C,$62,$70,$FC,$92,$4F,$13,$DC
	.dc.b	$7B,$0E,$64,$32,$5E,$1F,$84,$D9,$EA,$F3,$8F,$E6,$BC,$3F,$12,$3F
	.dc.b	$05,$1F,$CE,$38,$7E,$2C,$63,$87,$E2,$BF,$C5,$79,$FF,$1D,$C3,$F1
	.dc.b	$63,$E5,$70,$7C,$57,$E6,$6F,$F1,$5E,$5B,$FC,$57,$F8,$F6,$BE,$E7
	.dc.b	$F0,$5B,$7C,$48,$C4,$7B,$CF,$69,$34,$F6,$B7,$6F,$6B,$F8,$26,$BC
	.dc.b	$63,$64,$8C,$C7,$C7,$D1,$F0,$BF,$A9,$F6,$3F,$5F,$3F,$85,$F8,$5F
	.dc.b	$D7,$BF,$87,$F8,$4F,$D7,$BF,$89,$EA,$FD,$87,$C0,$7E,$C7,$5F,$D9
	.dc.b	$75,$FE,$CF,$E1,$3F,$69,$F0,$9F,$AB,$C3,$E0,$9F,$4F,$6F,$4D,$FF
	.dc.b	$94,$36,$7C,$2F,$F1,$9A,$FD,$D3,$78,$13,$A9,$FE,$28,$D7,$F9,$AC
	.dc.b	$6A,$78,$4E,$67,$FD,$92,$D7,$F8,$24,$68,$8F,$46,$BD,$23,$83,$E4
	.dc.b	$34,$43,$51,$C1,$EE,$FA,$72,$FF,$4C,$69,$FE,$F6,$7F,$D3,$12,$B4
	.dc.b	$3E,$0F,$FD,$9A,$FF,$BE,$E7,$FC,$1C,$BE,$0B,$3F,$05,$27,$3F,$05
	.dc.b	$C3,$E0,$B8,$FC,$17,$2F,$27,$9F,$B3,$D3,$E0,$BA,$FA,$7F,$CA,$78
	.dc.b	$3D,$19,$7C,$43,$D9,$C5,$F5,$63,$31,$F0,$4F,$79,$F0,$4F,$79,$4E
	.dc.b	$46,$88,$68,$A6,$8C,$10,$C1,$0F,$01,$93,$F9,$F7,$57,$C6,$F8,$4F
	.dc.b	$8F,$C3,$F2,$3E,$0F,$E4,$F0,$FC,$AF,$6F,$E5,$FB,$FF,$33,$DC,$F9
	.dc.b	$BF,$07,$F1,$B3,$F9,$BC,$3F,$1D,$F9,$FE,$B7,$D2,$F8,$3F,$8D,$1E
	.dc.b	$D3,$C6,$7B,$B8,$3E,$33,$9E,$9F,$C9,$37,$AF,$F1,$31,$88,$F8,$27
	.dc.b	$22,$39,$7D,$CD,$7E,$13,$5D,$C7,$42,$71,$7D,$69,$FD,$3F,$17,$C5
	.dc.b	$AF,$C4,$F0,$FC,$57,$F1,$7C,$1F,$15,$FC,$6E,$FF,$15,$FC,$7E,$DF
	.dc.b	$15,$FE,$4F,$D4,$DF,$E3,$46,$37,$F8,$CF,$F4,$DD,$3C,$5A,$F9,$20
	.dc.b	$7F,$A2,$35,$F5,$12,$35,$1F,$12,$31,$3C,$89,$1F,$2D,$AE,$E7,$8B
	.dc.b	$F2,$30,$FD,$7B,$FA,$D5,$EB,$F8,$A3,$FA,$0F,$F0,$07,$FA,$83,$0F
	.dc.b	$D8,$FC,$2E,$7E,$17,$FD,$07,$5F,$B5,$AF,$F4,$3C,$62,$BF,$6F,$DD
	.dc.b	$FB,$1A,$F8,$BD,$DF,$B1,$7F,$D0,$96,$F8,$2E,$9E,$07,$C2,$C1,$2F
	.dc.b	$85,$FD,$AD,$FE,$FB,$BB,$F6,$AF,$CB,$3E,$01,$FB,$4F,$4B,$95,$7F
	.dc.b	$A3,$E3,$FC,$6C,$D7,$F9,$1B,$5F,$81,$7E,$0F,$83,$FE,$59,$39,$9E
	.dc.b	$71,$82,$3D,$1A,$F4,$6B,$D2,$39,$3F,$1E,$32,$F4,$39,$10,$E4,$7A
	.dc.b	$3E,$17,$E2,$BF,$3F,$E1,$2D,$FC,$6F,$85,$B8,$3E,$16,$FC,$2C,$C3
	.dc.b	$E7,$BA,$F8,$5D,$FC,$2F,$0F,$85,$E3,$F0,$BC,$BD,$4E,$7E,$DF,$4F
	.dc.b	$6F,$AF,$0F,$A3,$3F,$51,$E7,$F1,$9C,$7E,$B8,$F4,$90,$F8,$53,$C7
	.dc.b	$87,$E7,$47,$BC,$C9,$4E,$46,$0E,$46,$08,$78,$8C,$94,$E0,$43,$F0
	.dc.b	$7F,$09,$F1,$B8,$FE,$3F,$C2,$7C,$8E,$2F,$93,$F0,$9F,$2B,$DE,$F9
	.dc.b	$7F,$09,$F1,$F1,$FC,$BF,$1F,$C7,$7F,$33,$C5,$F1,$DF,$CD,$F0,$FC
	.dc.b	$77,$F3,$BB,$7B,$F9,$BF,$03,$BF,$C0,$9F,$DF,$46,$09,$3B,$24,$7C
	.dc.b	$33,$D8,$7C,$09,$E4,$FB,$4C,$91,$CB,$B9,$EA,$F3,$8F,$D6,$EF,$F0
	.dc.b	$E3,$F6,$7B,$7C,$37,$F6,$BA,$FC,$37,$F6,$FA,$7C,$37,$F8,$2E,$9F
	.dc.b	$0D,$FE,$0F,$A7,$C3,$7F,$84,$E9,$F0,$DF,$E1,$7A,$7B,$1F,$C2,$F0
	.dc.b	$FE,$21,$F4,$6B,$E6,$7E,$DE,$BE,$A7,$35,$C5,$F8,$11,$92,$4F,$18
	.dc.b	$CB,$5F,$85,$3F,$BC,$E8,$FD,$4F,$EF,$FF,$EA,$4B,$FD,$55,$7F,$14
	.dc.b	$E9,$FD,$43,$FC,$8B,$A6,$F8,$1F,$C7,$6F,$FD,$53,$3F,$E1,$57,$FD
	.dc.b	$53,$19,$F0,$FA,$FC,$2F,$6F,$7B,$BF,$C2,$F8,$3E,$17,$F5,$78,$FD
	.dc.b	$3D,$FE,$07,$07,$C3,$D9,$A3,$E0,$FF,$C7,$CD,$BF,$F6,$83,$5F,$F1
	.dc.b	$EB,$5F,$E3,$6D,$7F,$96,$35,$F4,$47,$42,$3D,$1A,$F4,$6B,$D1,$AF
	.dc.b	$46,$BD,$0E,$47,$9B,$E1,$7E,$7F,$6F,$D0,$F8,$3F,$A3,$F0,$BF,$4B
	.dc.b	$D4,$97,$FD,$8B,$8B,$E7,$47,$D6,$F8,$4F,$AF,$F0,$9F,$63,$E1,$3E
	.dc.b	$CF,$C2,$7D,$AF,$84,$FB,$7F,$09,$F7,$3D,$4F,$E4,$BF,$C0,$3F,$D7
	.dc.b	$B8,$FE,$5F,$17,$99,$BB,$7F,$70,$E4,$FA,$1C,$9F,$16,$73,$3F,$04
	.dc.b	$F5,$94,$87,$C3,$32,$53,$27,$C5,$32,$53,$27,$C7,$2B,$E0,$79,$10
	.dc.b	$FB,$3C,$9F,$29,$F8,$DE,$7F,$F6,$7C,$7F,$CC,$F4,$7F,$CC,$D3,$A3
	.dc.b	$FC,$D4,$FF,$99,$DF,$F5,$EB,$FE,$67,$75,$1C,$7D,$2F,$A1,$18,$C1
	.dc.b	$C3,$F3,$3D,$2F,$A3,$39,$DF,$E7,$FC,$27,$D1,$8E,$79,$7D,$1C,$3E
	.dc.b	$61,$36,$FF,$07,$6D,$F4,$1F,$A9,$19,$AF,$8E,$7C,$53,$E1,$14,$F9
	.dc.b	$46,$34,$F9,$0E,$08,$6C,$8E,$9B,$1D,$DF,$D4,$ED,$F3,$5F,$8B,$F0
	.dc.b	$3F,$8A,$F7,$7F,$17,$EB,$7E,$33,$DD,$FC,$6F,$0F,$C6,$BF,$F7,$3E
	.dc.b	$3F,$19,$F8,$3C,$1F,$19,$CF,$07,$C6,$7E,$67,$1F,$E5,$BE,$23,$F8
	.dc.b	$9E,$FF,$08,$F8,$BD,$BE,$AE,$0F,$41,$32,$F4,$61,$F0,$5A,$ED,$AF
	.dc.b	$BA,$BA,$3F,$32,$79,$47,$E8,$F8,$9F,$0B,$FA,$8F,$77,$F6,$D7,$F8
	.dc.b	$1E,$EF,$DA,$BE,$EE,$DB,$DD,$FB,$68,$FC,$37,$77,$ED,$5F,$C0,$FC
	.dc.b	$2F,$EC,$BD,$9F,$DA,$C7,$E1,$3E,$17,$F6,$BF,$0B,$F0,$23,$C9,$C1
	.dc.b	$FC,$6F,$B3,$F9,$16,$EF,$85,$FF,$6F,$36,$BF,$CA,$C7,$E8,$63,$24
	.dc.b	$72,$7E,$81,$C3,$FE,$E2,$8E,$4D,$8C,$BF,$E4,$28,$CB,$63,$2F,$73
	.dc.b	$AF,$17,$EB,$E3,$31,$DC,$CC,$F2,$78,$1E,$1E,$1F,$E2,$BF,$07,$FC
	.dc.b	$5B,$E1,$7F,$8B,$FC,$27,$F1,$8F,$84,$97,$FA,$A7,$F8,$DF,$C2,$FF
	.dc.b	$1C,$F8,$4F,$F2,$0F,$A9,$FE,$75,$3F,$CE,$3E,$3F,$F9,$A5,$FF,$36
	.dc.b	$5F,$EF,$1F,$F3,$53,$5E,$3E,$A7,$F2,$6D,$BD,$3B,$FA,$EB,$F9,$27
	.dc.b	$67,$A1,$F8,$47,$B2,$77,$C1,$ED,$8F,$80,$7C,$B3,$44,$30,$43,$99
	.dc.b	$4E,$C4,$39,$9A,$21,$9A,$FE,$3D,$C1,$F0,$BD,$EF,$95,$3E,$FF,$57
	.dc.b	$E2,$FB,$DF,$1B,$D6,$F8,$FF,$05,$F2,$AB,$F8,$AF,$07,$CA,$F6,$BE
	.dc.b	$5F,$C1,$7C,$CF,$6F,$E6,$FC,$1F,$CE,$E3,$F9,$FF,$07,$F1,$F1,$F6
	.dc.b	$1E,$AE,$BF,$CB,$F1,$FC,$42,$46,$8F,$47,$1F,$C7,$E3,$F9,$D1,$F0
	.dc.b	$4F,$79,$A9,$E2,$E6,$36,$FC,$23,$67,$F3,$CE,$3F,$84,$FF,$30,$EA
	.dc.b	$FC,$2F,$1F,$C2,$8F,$C1,$71,$7C,$27,$F9,$AF,$47,$E2,$DF,$C3,$70
	.dc.b	$FC,$27,$F1,$1C,$1F,$09,$FC,$41,$F8,$BD,$BE,$13,$E5,$D7,$F5,$7B
	.dc.b	$7C,$28,$F8,$07,$9B,$5F,$4B,$ED,$32,$E1,$D4,$FC,$52,$6D,$F0,$A3
	.dc.b	$E0,$4E,$C9,$3A,$8F,$D9,$EE,$F7,$7F,$5B,$1F,$A1,$F8,$3F,$D6,$BF
	.dc.b	$AB,$21,$FD,$8B,$AF,$F5,$7F,$05,$FA,$CF,$82,$FD,$6C,$FE,$BB,$2F
	.dc.b	$D7,$FC,$27,$EC,$3E,$0F,$F6,$3F,$09,$FA,$AA,$FE,$BF,$B7,$D8,$DF
	.dc.b	$FC,$75,$B3,$E1,$3F,$6E,$4A,$FD,$A3,$57,$95,$FE,$E9,$9E,$11,$A2
	.dc.b	$1A,$F0,$BF,$7D,$5C,$6B,$83,$63,$33,$C2,$72,$D8,$CB,$CC,$68,$BF
	.dc.b	$0C,$66,$1F,$C0,$3F,$E6,$73,$FC,$C2,$E3,$1F,$F6,$9C,$3E,$16,$FC
	.dc.b	$2E,$3E,$17,$3F,$0B,$AF,$85,$99,$7F,$84,$30,$FF,$2D,$6A,$D5,$3E
	.dc.b	$49,$7B,$DD,$BC,$8E,$FF,$09,$E0,$F5,$3C,$3E,$A7,$8B,$E0,$7C,$79
	.dc.b	$7C,$3C,$BD,$BC,$7E,$CF,$8E,$7A,$9D,$B9,$7D,$BE,$B7,$A8,$97,$E5
	.dc.b	$6B,$F0,$F7,$46,$38,$3D,$13,$E7,$8C,$65,$D6,$B7,$3C,$C8,$73,$9D
	.dc.b	$BC,$4C,$D7,$4B,$F2,$C6,$2F,$A7,$C2,$FF,$42,$F0,$7E,$8F,$C2,$FD
	.dc.b	$2F,$77,$E9,$FC,$2F,$D4,$F7,$BE,$5E,$3F,$6F,$F8,$07,$FC,$C9,$B3
	.dc.b	$A3,$E9,$CE,$7E,$17,$ED,$F4,$F0,$E0,$F1,$61,$FE,$4C,$AE,$5F,$09
	.dc.b	$F4,$5F,$B5,$D1,$BE,$1F,$99,$5F,$E4,$D9,$F9,$1C,$3E,$EE,$2F,$5C
	.dc.b	$39,$7E,$9B,$F0,$4F,$88,$7B,$CC,$65,$F0,$F6,$F9,$71,$A2,$46,$79
	.dc.b	$3B,$4E,$A3,$13,$CE,$FC,$31,$88,$E3,$97,$D7,$BE,$27,$E7,$31,$F9
	.dc.b	$D3,$F9,$7E,$3F,$9C,$FE,$67,$8B,$E7,$3F,$E6,$AD,$3E,$6F,$0F,$CC
	.dc.b	$CB,$E4,$F0,$FC,$CA,$D7,$C2,$FC,$77,$F6,$37,$FE,$75,$AF,$77,$07
	.dc.b	$CD,$9F,$B1,$F0,$BE,$37,$D5,$C1,$F0,$B8,$7E,$4F,$0F,$EE,$7C,$2F
	.dc.b	$D0,$36,$BF,$80,$1F,$A2,$8C,$F4,$FE,$FB,$2F,$E0,$B3,$FB,$F8,$C7
	.dc.b	$1F,$F0,$69,$C9,$1C,$12,$7A,$1D,$AF,$C0,$F3,$BF,$D9,$BF,$BF,$F8
	.dc.b	$AF,$E0,$FE,$1F,$F0,$8F,$0F,$F8,$3C,$FF,$03,$F0,$BF,$83,$BF,$2F
	.dc.b	$E1,$BE,$36,$2E,$17,$C3,$7F,$10,$F8,$5C,$C3,$E1,$FF,$83,$61,$FC
	.dc.b	$37,$C4,$CF,$B1,$FC,$1A,$3F,$88,$78,$9E,$47,$CC,$7F,$2E,$E4,$F9
	.dc.b	$7C,$9F,$0A,$3D,$A6,$89,$C9,$F0,$DC,$B7,$5F,$82,$D9,$E5,$18,$E4
	.dc.b	$F9,$0D,$AE,$31,$89,$D9,$39,$3C,$D7,$FA,$1E,$4F,$93,$7F,$CA,$78
	.dc.b	$FE,$4B,$FC,$57,$8F,$E4,$BF,$C5,$FA,$BF,$25,$97,$F1,$3E,$2F,$C9
	.dc.b	$4F,$F1,$BF,$1B,$F9,$5B,$FB,$A3,$F9,$A7,$8D,$FC,$AD,$FE,$6D,$5F
	.dc.b	$88,$D3,$F8,$5E,$FF,$90,$E8,$FD,$74,$FE,$A6,$BF,$2B,$E0,$7E,$A9
	.dc.b	$A4,$3F,$5E,$7E,$F2,$38,$38,$E1,$FD,$AC,$63,$83,$F6,$C4,$7F,$90
	.dc.b	$CF,$F9,$77,$C5,$FE,$42,$FF,$4B,$F1,$7F,$90,$BF,$D3,$7C,$5F,$C2
	.dc.b	$FC,$3F,$C2,$8F,$E9,$9C,$3F,$88,$F0,$7E,$1E,$5F,$C7,$BC,$1F,$C5
	.dc.b	$F8,$BF,$B1,$8F,$D9,$DF,$EF,$7C,$1F,$D9,$3E,$7E,$0F,$D9,$78,$FE
	.dc.b	$7D,$E4,$FB,$49,$C1,$EC,$6D,$7C,$86,$F0,$FC,$77,$A3,$71,$F8,$11
	.dc.b	$A3,$FB,$47,$8D,$FC,$51,$FE,$A9,$E3,$7F,$14,$7F,$AB,$78,$DF,$C5
	.dc.b	$1F,$E8,$1C,$3F,$07,$2F,$EF,$7C,$3F,$06,$7F,$B9,$78,$FF,$AA,$E1
	.dc.b	$F8,$31,$F8,$33,$FB,$07,$8F,$FA,$EE,$1F,$83,$AF,$93,$1F,$E7,$7C
	.dc.b	$FF,$A8,$E8,$FC,$36,$1F,$C0,$FB,$FF,$34,$D8,$CB,$FA,$AE,$3F,$D0
	.dc.b	$46,$30,$FD,$43,$FC,$6B,$8F,$FC,$21,$DF,$FA,$0B,$FE,$E5,$DF,$FA
	.dc.b	$0B,$FE,$89,$DF,$FA,$07,$FA,$BF,$8D,$FC,$13,$C6,$FE,$0D,$7F,$CD
	.dc.b	$7C,$5F,$E0,$AF,$F7,$BF,$17,$F8,$2B,$FA,$03,$F8,$29,$FA,$03,$D0
	.dc.b	$7F,$85,$37,$FE,$5F,$E3,$3E,$1F,$FD,$9F,$3F,$CE,$9B,$C5,$FC,$5D
	.dc.b	$B3,$B9,$FF,$64,$B5,$FF,$6D,$CE,$DC,$CF,$02,$38,$21,$C2,$79,$3F
	.dc.b	$DF,$A7,$93,$99,$E2,$D3,$89,$B9,$E2,$FF,$78,$8F,$3F,$C3,$FF,$2C
	.dc.b	$EE,$93,$FD,$77,$E1,$FF,$97,$FC,$3C,$8F,$EC,$0F,$F3,$2F,$3B,$F9
	.dc.b	$9F,$C3,$7F,$34,$F7,$BF,$9A,$FC,$37,$F3,$6F,$5E,$61,$FD,$9F,$DD
	.dc.b	$FE,$72,$7F,$3A,$D7,$CF,$1E,$57,$FC,$33,$E8,$BC,$EF,$34,$FA,$8A
	.dc.b	$7B,$48,$71,$34,$7D,$52,$F0,$FC,$67,$47,$03,$99,$82,$18,$21,$CC
	.dc.b	$E0,$7F,$95,$FA,$BE,$3B,$FE,$2F,$E0,$FE,$AB,$F0,$5F,$D6,$BF,$80
	.dc.b	$7F,$C6,$73,$FE,$6D,$7F,$DF,$2F,$F9,$C9,$FF,$7C,$BF,$E3,$7E,$0E
	.dc.b	$1F,$05,$F2,$E3,$FC,$CB,$D7,$FE,$45,$AF,$F3,$C4,$7F,$84,$77,$F9
	.dc.b	$AF,$5F,$87,$F0,$7C,$3F,$CD,$8F,$83,$D1,$F9,$1E,$0F,$99,$1E,$B3
	.dc.b	$DE,$53,$DF,$BF,$C7,$6C,$FE,$21,$D1,$23,$4D,$8D,$46,$08,$E0,$9C
	.dc.b	$3F,$80,$9E,$FE,$07,$FA,$3A,$7F,$C8,$B1,$F2,$CF,$F1,$76,$FF,$2D
	.dc.b	$F9,$97,$F3,$7D,$BF,$C9,$47,$C6,$F0,$FF,$CF,$DB,$38,$3E,$54,$FD
	.dc.b	$1F,$13,$E3,$F0,$FF,$97,$B8,$7E,$B7,$0F,$C8,$AF,$F1,$E6,$FE,$57
	.dc.b	$F0,$1C,$1F,$23,$5F,$5C,$E5,$B3,$F3,$A3,$1B,$FB,$A3,$04,$3F,$13
	.dc.b	$3C,$1B,$1F,$17,$B3,$F1,$1F,$09,$F8,$9F,$87,$FF,$0F,$F7,$7E,$C6
	.dc.b	$FF,$CF,$9D,$D7,$6F,$C2,$7C,$2F,$EC,$A7,$E1,$1F,$8D,$F8,$7F,$D9
	.dc.b	$B8,$D1,$C6,$F8,$6F,$CB,$7C,$3F,$ED,$2B,$F6,$C7,$C2,$3D,$1C,$3F
	.dc.b	$06,$9F,$0F,$FC,$81,$B3,$F9,$E3,$F3,$91,$92,$39,$C3,$C3,$18,$8F
	.dc.b	$E4,$73,$C9,$B1,$97,$F8,$F4,$65,$B1,$97,$EF,$B5,$F0,$9D,$E3,$83
	.dc.b	$A3,$11,$E9,$F8,$7F,$DB,$1F,$B7,$F6,$FF,$71,$F0,$DF,$B9,$F8,$69
	.dc.b	$7F,$27,$07,$C3,$7E,$F7,$E1,$BF,$7D,$F0,$DF,$BF,$F4,$BF,$CB,$3E
	.dc.b	$9F,$FB,$1B,$0F,$F7,$63,$FE,$EA,$1B,$E8,$DF,$A7,$FC,$1F,$7F,$57
	.dc.b	$07,$C6,$BF,$F7,$2D,$7D,$19,$CE,$9E,$C9,$F7,$9F,$20,$F6,$90,$C1
	.dc.b	$0F,$92,$64,$A6,$4F,$8C,$53,$A9,$A2,$1A,$3E,$AF,$07,$CA,$7E,$33
	.dc.b	$F2,$9F,$C8,$62,$F8,$6F,$8B,$EA,$7F,$25,$DF,$E3,$FC,$37,$CA,$BD
	.dc.b	$FB,$3F,$2B,$C5,$F9,$7E,$64,$9E,$7F,$09,$F2,$1F,$A9,$B7,$C4,$3F
	.dc.b	$A4,$F9,$DE,$58,$87,$A8,$BC,$3F,$35,$F7,$35,$F6,$9B,$23,$97,$8F
	.dc.b	$07,$D1,$8C,$46,$A3,$04,$AF,$AF,$C1,$F1,$67,$FA,$13,$F1,$5F,$C2
	.dc.b	$EF,$F1,$1F,$C7,$6D,$F1,$1F,$C8,$72,$7E,$33,$6F,$89,$1F,$33,$5F
	.dc.b	$88,$FE,$3B,$5F,$88,$FF,$4F,$CB,$E2,$3F,$8A,$D7,$D3,$18,$8F,$83
	.dc.b	$A7,$9D,$FC,$21,$A2,$9E,$D8,$E0,$FC,$6D,$7E,$09,$27,$E2,$70,$BD
	.dc.b	$AF,$D1,$E1,$FA,$DB,$FC,$1F,$AB,$FA,$C7,$FB,$27,$45,$EF,$FE,$BD
	.dc.b	$D3,$FA,$AF,$67,$F5,$95,$F8,$AF,$84,$FD,$6F,$C2,$FE,$A6,$3F,$1B
	.dc.b	$D5,$5F,$80,$7A,$37,$FB,$D1,$FC,$E7,$67,$C0,$FF,$8E,$1B,$5F,$C5
	.dc.b	$5A,$FF,$8D,$9A,$FF,$8B,$DA,$85,$23,$31,$86,$C6,$67,$C0,$49,$F0
	.dc.b	$35,$F0,$35,$F0,$35,$F0,$47,$37,$89,$0F,$47,$C0,$FF,$12,$F8,$19
	.dc.b	$1F,$B2,$F8,$5F,$E2,$BF,$05,$FC,$5B,$E1,$64,$FF,$45,$FE,$01,$7C
	.dc.b	$27,$FB,$C9,$7C,$35,$F8,$5C,$7C,$36,$77,$D7,$C3,$6F,$E1,$A6,$5C
	.dc.b	$FE,$1B,$F9,$1F,$17,$B7,$8F,$E9,$3E,$B9,$F8,$CD,$E3,$F6,$47,$C2
	.dc.b	$3D,$87,$C2,$29,$F0,$B1,$F9,$A4,$9E,$4D,$8D,$9D,$8B,$3D,$9F,$B9
	.dc.b	$D9,$F1,$9F,$E6,$FC,$7F,$CE,$B8,$FE,$3C,$7C,$5F,$87,$F8,$9E,$C7
	.dc.b	$C5,$F5,$FE,$37,$14,$AF,$95,$C3,$F2,$7E,$0B,$E2,$C7,$CD,$F5,$BE
	.dc.b	$67,$AD,$F3,$78,$BE,$2C,$FD,$9E,$2F,$8A,$F9,$B8,$7D,$4F,$C3,$21
	.dc.b	$F0,$CA,$7C,$3E,$2F,$82,$47,$93,$5D,$9F,$0C,$FC,$F7,$0F,$C1,$EF
	.dc.b	$79,$9F,$37,$8F,$E1,$57,$F3,$AE,$BF,$C2,$71,$FC,$28,$FC,$47,$17
	.dc.b	$C1,$58,$72,$7F,$4B,$D3,$F8,$A7,$E2,$9F,$8F,$E1,$F8,$AF,$0C,$BF
	.dc.b	$3B,$2E,$1F,$8F,$3F,$8B,$3F,$A5,$EB,$F8,$BF,$0F,$E0,$47,$B8,$87
	.dc.b	$BB,$9F,$EF,$3A,$31,$C1,$F3,$1F,$94,$43,$45,$3E,$31,$63,$9B,$83
	.dc.b	$F2,$5D,$BF,$A7,$F8,$2F,$D4,$7C,$37,$EA,$7E,$1B,$F5,$5F,$0D,$FA
	.dc.b	$EB,$FF,$01,$F8,$3F,$AD,$F0,$7F,$5D,$EF,$FE,$B6,$7F,$65,$E1,$7E
	.dc.b	$B5,$FD,$A7,$C3,$7E,$B1,$F8,$C7,$CC,$F0,$BF,$6D,$F0,$9F,$8A,$8F
	.dc.b	$C3,$76,$FF,$95,$FF,$80,$7F,$DA,$F1,$FC,$59,$BC,$26,$92,$3F,$DB
	.dc.b	$6D,$7F,$C7,$0D,$7F,$DB,$2D,$7F,$DA,$4D,$7F,$8A,$B5,$F0,$1D,$08
	.dc.b	$74,$3B,$C7,$F2,$B7,$DA,$F9,$7E,$17,$1D,$5F,$8B,$8C,$4F,$F9,$6E
	.dc.b	$BF,$CF,$CD,$F8,$4D,$F5,$78,$BE,$12,$CF,$C1,$6D,$FC,$1F,$87,$B8
	.dc.b	$1B,$F0,$98,$F8,$49,$87,$7F,$84,$DF,$C2,$48,$F1,$7C,$27,$27,$DA
	.dc.b	$7F,$8F,$BB,$BE,$1B,$83,$E3,$94,$C9,$F0,$C8,$7C,$E3,$47,$CA,$32
	.dc.b	$72,$29,$C8,$E0,$60,$E0,$60,$87,$D8,$F8,$4F,$83,$C3,$23,$E6,$7A
	.dc.b	$FF,$13,$DE,$F8,$BF,$09,$F1,$BE,$13,$E3,$EF,$2F,$E9,$7B,$1F,$2B
	.dc.b	$DF,$F9,$7E,$CF,$CC,$E0,$F9,$BE,$FF,$CE,$F8,$4F,$A3,$FC,$03,$FE
	.dc.b	$65,$EB,$FB,$9C,$FF,$8D,$EC,$FC,$C6,$9F,$03,$A7,$F2,$CD,$9F,$C6
	.dc.b	$46,$23,$DC,$D7,$DA,$7C,$33,$E3,$18,$3E,$19,$82,$1B,$21,$A3,$F3
	.dc.b	$7C,$9F,$15,$FC,$AC,$FF,$BE,$2F,$D3,$C7,$F1,$E7,$F2,$9C,$5F,$1D
	.dc.b	$FC,$B7,$0F,$C7,$78,$F0,$FC,$77,$F3,$5C,$1F,$1D,$FC,$E6,$FF,$1D
	.dc.b	$FF,$5C,$6D,$F1,$FA,$FC,$BB,$7B,$F7,$F8,$F5,$F0,$B6,$F8,$A4,$7E
	.dc.b	$41,$0D,$14,$F8,$93,$F2,$5F,$86,$7C,$13,$F8,$43,$F6,$3E,$03,$EC
	.dc.b	$D7,$EA,$3E,$1F,$F5,$3D,$DF,$AA,$EE,$98,$7A,$3E,$1B,$EF,$FC,$27
	.dc.b	$EB,$23,$F7,$1D,$FF,$AC,$7F,$7F,$B3,$E1,$BF,$59,$1F,$3B,$C0,$FD
	.dc.b	$AF,$81,$FA,$D8,$FD,$2E,$BF,$EF,$3B,$7C,$37,$FB,$46,$7F,$C6,$71
	.dc.b	$87,$FD,$A0,$D7,$F8,$71,$A8,$E1,$C0,$9D,$4F,$F1,$38,$C4,$EA,$73
	.dc.b	$3F,$E3,$26,$BC,$CE,$F1,$B2,$38,$21,$E2,$9E,$AE,$0F,$04,$F4,$8C
	.dc.b	$C6,$E3,$03,$E1,$BF,$57,$F0,$DF,$AC,$F6,$BF,$5B,$E2,$7E,$BB,$E1
	.dc.b	$7F,$5F,$F0,$BF,$B0,$F8,$5F,$D8,$FC,$2C,$CF,$F1,$BE,$BF,$ED,$3D
	.dc.b	$6F,$DA,$FB,$DF,$B6,$F8,$4F,$DB,$FB,$FF,$B8,$F5,$7F,$94,$F4,$FF
	.dc.b	$91,$B1,$FC,$AF,$27,$B9,$FE,$40,$7C,$33,$D0,$E5,$F6,$B5,$F9,$4D
	.dc.b	$78,$1A,$3E,$21,$A2,$9A,$30,$76,$33,$C9,$F2,$1E,$46,$4F,$95,$C7
	.dc.b	$F2,$1F,$D6,$F1,$FB,$3E,$0F,$DB,$EA,$7B,$BE,$0F,$E4,$57,$EC,$B8
	.dc.b	$7E,$43,$FD,$5B,$3F,$F4,$03,$FE,$83,$F2,$BF,$AB,$43,$E1,$7E,$5C
	.dc.b	$E7,$C5,$FF,$3C,$CB,$D5,$F9,$7F,$07,$F3,$6B,$47,$FA,$32,$5F,$07
	.dc.b	$FC,$A3,$87,$E7,$77,$7E,$85,$F9,$1C,$5E,$FB,$F8,$67,$C0,$21,$82
	.dc.b	$1F,$24,$8F,$16,$F7,$7E,$3A,$7A,$B6,$3E,$6B,$F9,$2E,$3F,$95,$1F
	.dc.b	$BA,$E3,$F9,$4F,$AA,$FF,$E0,$07,$17,$CB,$8F,$E3,$7D,$5F,$2B,$87
	.dc.b	$FB,$E6,$2E,$2F,$C6,$FB,$FF,$8E,$E3,$F9,$38,$7C,$FE,$2F,$8E,$FF
	.dc.b	$24,$E6,$FC,$9F,$0F,$C5,$8F,$D6,$F0,$FC,$57,$CF,$C3,$EA,$23,$82
	.dc.b	$1F,$85,$3E,$39,$F0,$9D,$B9,$3E,$4E,$EE,$27,$C2,$FE,$93,$E1,$3F
	.dc.b	$65,$5F,$CE,$3C,$0F,$D9,$3F,$CE,$BB,$FF,$66,$FC,$5B,$7C,$27,$EB
	.dc.b	$3E,$0F,$FB,$2F,$83,$FB,$3A,$FE,$83,$E0,$FE,$CD,$FE,$87,$DF,$FB
	.dc.b	$47,$97,$BF,$F0,$5F,$CF,$70,$34,$FC,$7F,$C2,$FE,$E0,$93,$E1,$6B
	.dc.b	$F9,$B6,$F4,$FE,$69,$BD,$FC,$A7,$81,$0E,$05,$9E,$0D,$8E,$0E,$89
	.dc.b	$1F,$A6,$F8,$4F,$E1,$7F,$09,$FC,$33,$E1,$3F,$86,$FC,$27,$F0,$EF
	.dc.b	$84,$FE,$1F,$F0,$9F,$C4,$3E,$13,$FC,$73,$FC,$03,$FC,$EB,$2F,$F2
	.dc.b	$23,$FE,$F2,$CF,$F2,$0F,$F7,$5D,$3F,$DC,$EF,$F9,$1A,$3F,$C9,$AB
	.dc.b	$D4,$E7,$EA,$74,$F6,$FA,$FC,$2F,$6F,$57,$E0,$E1,$FB,$3E,$2F,$F2
	.dc.b	$B7,$C5,$7D,$88,$C1,$EF,$29,$EE,$CF,$E0,$E9,$F1,$23,$06,$9C,$12
	.dc.b	$36,$48,$D3,$7C,$0F,$C8,$5F,$91,$B3,$E1,$6B,$FC,$57,$93,$E6,$BF
	.dc.b	$8F,$E3,$F9,$AF,$C6,$F8,$2F,$9A,$FC,$8F,$5B,$E6,$FA,$DF,$3B,$09
	.dc.b	$3F,$33,$E0,$7E,$8F,$C0,$FD,$2F,$85,$FA,$7F,$05,$F5,$3E,$17,$EA
	.dc.b	$FC,$1F,$97,$C6,$70,$7C,$8F,$1B,$D5,$18,$CF,$D0,$D7,$E2,$E9,$EA
	.dc.b	$E3,$F8,$91,$F0,$89,$19,$E3,$E7,$87,$72,$4F,$47,$C5,$C7,$F0,$A3
	.dc.b	$F9,$C7,$1F,$C2,$7F,$07,$C7,$F0,$9F,$C2,$F1,$FC,$27,$F1,$78,$FC
	.dc.b	$27,$F1,$3C,$5F,$09,$FC,$7F,$0F,$C2,$7F,$23,$C1,$F0,$9F,$10,$1C
	.dc.b	$9F,$89,$B7,$53,$E1,$7F,$3A,$4A,$FC,$DB,$5F,$CC,$B7,$83,$F4,$2D
	.dc.b	$E0,$FC,$FB,$65,$7D,$A3,$0D,$8E,$A4,$3A,$96,$BA,$BF,$AC,$F8,$4F
	.dc.b	$D8,$FC,$27,$EC,$BE,$17,$FD,$57,$FC,$01,$FE,$55,$AF,$EB,$1E,$16
	.dc.b	$3E,$17,$F7,$11,$FB,$43,$7E,$3F,$EE,$BE,$1F,$F7,$7F,$0D,$FB,$CF
	.dc.b	$13,$F6,$75,$FB,$AF,$0F,$F6,$6F,$EF,$3C,$2F,$4F,$1F,$C7,$8F,$99
	.dc.b	$C9,$F2,$F5,$7C,$37,$90,$93,$FD,$4E,$32,$D8,$CB,$FE,$D0,$8C,$B6
	.dc.b	$33,$C5,$FE,$D1,$9C,$B6,$33,$CB,$F7,$79,$15,$A8,$C3,$6F,$E9,$B6
	.dc.b	$77,$5C,$A7,$2D,$8E,$7D,$DF,$3B,$E1,$BE,$7F,$A9,$F4,$3D,$FF,$A3
	.dc.b	$EF,$CC,$7D,$DF,$0D,$F5,$3C,$FF,$AB,$F0,$DF,$5B,$D6,$FA,$FF,$01
	.dc.b	$F6,$3D,$FF,$B3,$EB,$FD,$AF,$5F,$ED,$FC,$37,$DC,$F8,$6F,$17,$27
	.dc.b	$C4,$E7,$F9,$DC,$B8,$26,$FF,$2C,$87,$C9,$2F,$3F,$C7,$6F,$3F,$C4
	.dc.b	$9E,$31,$B2,$4E,$DB,$3B,$7F,$24,$7D,$BF,$67,$E1,$BF,$77,$93,$E1
	.dc.b	$3F,$6F,$93,$F2,$BE,$F7,$FC,$00,$FE,$01,$FF,$2C,$4F,$DF,$E4,$C7
	.dc.b	$C3,$FC,$48,$FF,$0C,$4B,$E1,$7E,$2C,$7E,$27,$8B,$97,$C3,$FC,$58
	.dc.b	$F5,$9F,$60,$FC,$9F,$A7,$E7,$7D,$26,$4B,$DD,$F8,$07,$8B,$5C,$F1
	.dc.b	$FE,$14,$9C,$5F,$1B,$8B,$E2,$13,$8B,$B3,$6B,$B3,$6F,$F8,$6F,$95
	.dc.b	$FB,$6E,$3F,$8F,$5F,$9F,$E3,$F8,$F7,$F9,$DE,$3F,$8F,$3F,$0B,$DE
	.dc.b	$FC,$4B,$F1,$38,$BE,$32,$A6,$7F,$A1,$E2,$F9,$13,$FA,$FE,$8F,$C6
	.dc.b	$3F,$33,$87,$E4,$3E,$ED,$FF,$A2,$3F,$C5,$AB,$FC,$D6,$F5,$BE,$9F
	.dc.b	$0B,$F6,$49,$1F,$59,$AE,$CE,$4E,$31,$F2,$C6,$0E,$04,$38,$70,$2F
	.dc.b	$C3,$3C,$23,$15,$CE,$FC,$77,$C6,$B5,$7C,$27,$33,$D2,$30,$FF,$00
	.dc.b	$F0,$7F,$74,$FE,$3B,$E1,$7F,$29,$F0,$BF,$95,$F8,$7F,$CB,$7C,$37
	.dc.b	$F2,$CF,$0F,$F7,$57,$F9,$4F,$86,$FD,$D3,$F9,$6F,$86,$FD,$D3,$F9
	.dc.b	$8E,$AF,$CF,$7C,$37,$EE,$A3,$F3,$7E,$E7,$A5,$F8,$47,$F5,$29,$FB
	.dc.b	$1E,$E7,$E9,$49,$1F,$A2,$6B,$EB,$6F,$0F,$C0,$6F,$0F,$29,$E0,$43
	.dc.b	$81,$67,$83,$63,$83,$A2,$47,$CC,$F8,$6F,$DC,$7B,$BF,$B9,$F8,$6F
	.dc.b	$DD,$7A,$DF,$BB,$F8,$6F,$DE,$7C,$37,$F6,$DC,$7E,$DF,$C3,$7F,$91
	.dc.b	$73,$FE,$68,$FF,$91,$63,$3E,$BE,$F4,$FB,$D3,$FE,$45,$6F,$AF,$9B
	.dc.b	$E3,$EB,$DC,$3F,$B5,$D8,$C7,$6D,$F5,$EC,$3E,$1A,$FB,$1F,$1A,$FF
	.dc.b	$3D,$C7,$FA,$0C,$39,$7C,$17,$02,$4F,$FA,$A7,$04,$62,$3D,$AD,$D7
	.dc.b	$D2,$DD,$3D,$ED,$9F,$3B,$6F,$BC,$75,$8C,$5F,$68,$C1,$0F,$97,$F0
	.dc.b	$5E,$CF,$86,$FA,$51,$E5,$F8,$3F,$A4,$FF,$28,$E3,$F8,$3F,$0D,$F0
	.dc.b	$B2,$F8,$7C,$7F,$13,$8E,$5F,$B7,$8F,$E3,$E9,$F2,$38,$FE,$4F,$1F
	.dc.b	$CA,$F8,$4F,$2F,$07,$C7,$AF,$36,$1F,$08,$F3,$71,$7C,$26,$F1,$7C
	.dc.b	$26,$9F,$08,$BA,$FC,$27,$91,$27,$93,$67,$93,$4E,$45,$9E,$53,$F8
	.dc.b	$AE,$2F,$8D,$3F,$D2,$B9,$FF,$07,$C5,$F1,$A3,$F8,$27,$17,$C6,$7E
	.dc.b	$F6,$1F,$19,$FC,$07,$07,$C6,$7F,$37,$C1,$F1,$9F,$C2,$70,$7C,$67
	.dc.b	$F0,$DB,$FC,$67,$FA,$89,$F8,$E3,$F5,$C5,$3F,$5B,$AF,$EC,$35,$FD
	.dc.b	$9C,$60,$D3,$82,$46,$C9,$1A,$6E,$3F,$CA,$79,$3F,$3C,$D9,$F0,$B5
	.dc.b	$FE,$49,$D3,$FB,$C7,$F3,$DF,$0D,$FB,$C7,$F4,$1F,$0B,$FB,$C7,$F4
	.dc.b	$47,$EB,$3E,$1B,$F5,$BF,$0D,$FA,$EF,$4F,$F5,$FF,$0D,$FB,$0F,$86
	.dc.b	$FD,$9F,$F0,$0F,$F9,$4F,$42,$F8,$7C,$FC,$3F,$ED,$33,$F2,$F9,$EE
	.dc.b	$0F,$DA,$5F,$F8,$D4,$9E,$8F,$42,$57,$F9,$EF,$8F,$FA,$34,$EA,$31
	.dc.b	$C7,$FE,$43,$8C,$47,$F4,$E6,$CF,$0A,$E7,$39,$6C,$66,$39,$9F,$E7
	.dc.b	$A9,$E3,$5D,$6B,$DF,$E8,$7C,$0F,$86,$F8,$3F,$0D,$F0,$BD,$9F,$87
	.dc.b	$EC,$FC,$4F,$43,$E2,$FC,$37,$C6,$F7,$E6,$BD,$B8,$FE,$3F,$C3,$7C
	.dc.b	$AF,$86,$F9,$7F,$0D,$F3,$3D,$CF,$9B,$F0,$3F,$3B,$E1,$BC,$FC,$7E
	.dc.b	$3C,$B9,$72,$7F,$91,$FD,$4F,$85,$F0,$DF,$50,$93,$A6,$BF,$15,$BC
	.dc.b	$FF,$31,$B7,$CA,$78,$10,$E0,$59,$E0,$D8,$E0,$E8,$91,$F6,$39,$BF
	.dc.b	$0F,$F0,$5F,$22,$7D,$3F,$0B,$F2,$1F,$5F,$27,$F3,$EF,$E0,$1F,$F2
	.dc.b	$4E,$1F,$80,$E4,$F9,$35,$FE,$0A,$97,$C1,$FC,$98,$F8,$7C,$7C,$7E
	.dc.b	$1B,$E4,$C7,$CF,$A7,$C2,$75,$F8,$4F,$7C,$FD,$FE,$0F,$C9,$98,$E8
	.dc.b	$FC,$DB,$78,$CD,$24,$FC,$13,$91,$67,$8B,$67,$8B,$67,$8B,$F9,$7E
	.dc.b	$1F,$8A,$FD,$CE,$CF,$D6,$F0,$FC,$58,$FC,$E7,$0F,$C5,$F8,$5F,$87
	.dc.b	$C5,$F1,$67,$DD,$C3,$F0,$DF,$D1,$70,$FC,$37,$F0,$FC,$FF,$37,$83
	.dc.b	$E1,$47,$8C,$FE,$37,$BF,$EE,$BB,$FF,$1F,$1F,$A3,$6C,$66,$3F,$23
	.dc.b	$18,$E0,$FD,$3C,$6C,$C9,$0F,$D8,$17,$2F,$D8,$38,$23,$82,$73,$7F
	.dc.b	$AF,$F8,$DF,$0B,$FB,$1A,$FB,$9F,$0B,$FB,$17,$F4,$7E,$07,$E2,$7E
	.dc.b	$17,$F6,$31,$F8,$1F,$85,$FC,$67,$C2,$FF,$18,$F0,$7F,$1D,$F0,$BF
	.dc.b	$B0,$AF,$D5,$F8,$3F,$91,$F0,$3F,$61,$1E,$7D,$BF,$60,$F9,$FE,$1D
	.dc.b	$3E,$5F,$86,$FC,$C1,$27,$F2,$0D,$7F,$30,$DE,$1F,$F3,$07,$0A,$31
	.dc.b	$87,$3A,$E2,$43,$89,$6B,$8C,$F0,$87,$22,$FA,$46,$EF,$A3,$FA,$9E
	.dc.b	$2F,$D7,$7C,$37,$EB,$FC,$9F,$D8,$7A,$9F,$B1,$F8,$6F,$D9,$7B,$9F
	.dc.b	$B3,$F8,$69,$8F,$EA,$FD,$CF,$DB,$7C,$37,$F8,$D3,$F8,$07,$FC,$85
	.dc.b	$3E,$27,$FC,$85,$C3,$FE,$7D,$5F,$09,$BF,$7F,$87,$BF,$E9,$E3,$F4
	.dc.b	$72,$7C,$7E,$6F,$9B,$97,$3F,$82,$EC,$47,$F9,$56,$5F,$E4,$56,$C6
	.dc.b	$5F,$F4,$7C,$65,$B1,$9C,$FC,$8D,$E5,$4F,$73,$99,$0E,$25,$9E,$2D
	.dc.b	$8E,$38,$AF,$C1,$5F,$7B,$97,$E7,$DF,$F1,$3E,$5F,$91,$F0,$1F,$3E
	.dc.b	$3F,$8B,$71,$FC,$E7,$F8,$F7,$1D,$C3,$F0,$FC,$5F,$3E,$7E,$86,$7F
	.dc.b	$47,$E1,$3E,$97,$C3,$FC,$C9,$F8,$BF,$0B,$ED,$7D,$87,$FA,$4C,$B8
	.dc.b	$7A,$9F,$69,$1C,$13,$7F,$6E,$FE,$C6,$F0,$75,$CB,$E0,$36,$75,$7D
	.dc.b	$AB,$6F,$C7,$30,$4A,$E2,$E5,$FC,$D7,$07,$C8,$8F,$E8,$1C,$1F,$21
	.dc.b	$FE,$79,$CF,$F9,$67,$F9,$F6,$FF,$1F,$0F,$E1,$0F,$E5,$A3,$F0,$E7
	.dc.b	$C8,$3F,$13,$B7,$C8,$7F,$17,$C9,$F9,$BD,$BD,$5A,$FF,$A2,$7A,$BF
	.dc.b	$57,$5F,$B2,$26,$FF,$B2,$7F,$5C,$7E,$CA,$37,$19,$8D,$E5,$FB,$28
	.dc.b	$C1,$1C,$12,$39,$72,$FE,$96,$B9,$CE,$77,$FD,$3D,$78,$1F,$C8,$7C
	.dc.b	$2F,$EC,$FE,$1F,$F7,$93,$FD,$87,$AF,$F6,$DF,$0D,$FB,$7F,$87,$FD
	.dc.b	$C7,$C3,$FE,$E7,$E1,$FF,$B8,$77,$CC,$3F,$5B,$F0,$FF,$BD,$F8,$7F
	.dc.b	$DF,$78,$3F,$B9,$9F,$D8,$F8,$3E,$5E,$2F,$DC,$78,$7F,$86,$F8,$7F
	.dc.b	$E0,$84,$AF,$DB,$B5,$FD,$AB,$79,$7F,$83,$37,$93,$94,$F0,$21,$C0
	.dc.b	$B3,$C1,$B1,$C1,$D1,$23,$F8,$97,$C3,$FF,$17,$F8,$7F,$E3,$1F,$0F
	.dc.b	$FE,$CC,$F0,$3F,$CB,$47,$F9,$66,$FF,$C6,$CB,$E1,$B5,$E9,$49,$CF
	.dc.b	$B7,$C7,$E1,$79,$7B,$9C,$FD,$3E,$9F,$0B,$D7,$DC,$ED,$F0,$BD,$FE
	.dc.b	$0B,$DD,$97,$F2,$7E,$3F,$F6,$B7,$96,$73,$E1,$FF,$0B,$6C,$66,$7D
	.dc.b	$E7,$12,$CE,$DB,$1B,$7E,$17,$2F,$C5,$E5,$F9,$11,$83,$91,$6B,$93
	.dc.b	$78,$9E,$A7,$C9,$8F,$E6,$3C,$9F,$27,$DC,$F9,$5F,$0B,$F2,$FD,$5F
	.dc.b	$8D,$4C,$99,$FF,$26,$C9,$CD,$F4,$7D,$DF,$95,$7F,$43,$97,$EA,$7C
	.dc.b	$2F,$CA,$8F,$A7,$F0,$5F,$5F,$E1,$7E,$C7,$C1,$FC,$4E,$4F,$92,$5A
	.dc.b	$F8,$2D,$7D,$9C,$9F,$0F,$6F,$82,$4D,$BE,$21,$27,$E1,$C6,$5B,$19
	.dc.b	$E4,$F8,$B3,$97,$FC,$2B,$C9,$F1,$6B,$FA,$8F,$27,$C5,$7F,$17,$C7
	.dc.b	$F1,$5F,$C6,$F0,$FC,$6C,$3E,$2C,$7F,$7A,$E2,$71,$7C,$48,$F9,$59
	.dc.b	$FC,$9E,$2F,$8B,$1F,$96,$E8,$FC,$D7,$17,$C5,$8F,$CC,$F0,$FC,$1F
	.dc.b	$67,$FB,$5F,$07,$F8,$AB,$9F,$F2,$55,$F9,$E6,$BD,$8F,$13,$8D,$7C
	.dc.b	$B3,$C6,$78,$10,$E0,$E5,$BA,$AB,$96,$9F,$81,$6C,$F6,$39,$5F,$6A
	.dc.b	$E1,$18,$AE,$0F,$F8,$03,$BF,$FB,$BB,$E0,$FE,$01,$FF,$43,$CF,$F7
	.dc.b	$27,$FD,$0B,$1F,$BA,$8F,$DF,$7C,$47,$EF,$FC,$0E,$38,$FB,$2D,$F0
	.dc.b	$F8,$BE,$5F,$09,$FC,$17,$C2,$C6,$5F,$E2,$EB,$7C,$3E,$37,$47,$F9
	.dc.b	$5A,$7F,$84,$F8,$59,$87,$C2,$7F,$03,$9F,$E1,$BE,$17,$C2,$BF,$91
	.dc.b	$AF,$C1,$E2,$F6,$F1,$FD,$8C,$5F,$09,$CC,$98,$FF,$B2,$E3,$2D,$8C
	.dc.b	$BF,$15,$B7,$EF,$6E,$FF,$1D,$B3,$ED,$6E,$3E,$C6,$DF,$69,$E8,$77
	.dc.b	$3C,$5F,$09,$EF,$F8,$4F,$81,$F0,$9F,$07,$CA,$F8,$5E,$5F,$C3,$F7
	.dc.b	$FE,$27,$C2,$4C,$7D,$5E,$BF,$C7,$F8,$0F,$91,$F0,$7F,$27,$E0,$7E
	.dc.b	$57,$AD,$F2,$FE,$0F,$E8,$7F,$00,$FF,$94,$38,$CF,$F9,$43,$93,$D9
	.dc.b	$9F,$94,$FF,$10,$4F,$AD,$C7,$2F,$C3,$6C,$F0,$7D,$A4,$8D,$36,$35
	.dc.b	$18,$23,$82,$47,$CB,$3A,$94,$E4,$58,$FC,$97,$1F,$C9,$8F,$B9,$C5
	.dc.b	$F2,$5F,$F2,$AC,$FE,$CB,$0F,$F3,$5C,$7E,$07,$83,$E5,$3F,$A3,$E0
	.dc.b	$F8,$BE,$AF,$C6,$F8,$2F,$8F,$F0,$3F,$2E,$B9,$E9,$FE,$58,$C5,$F0
	.dc.b	$BF,$33,$E1,$3E,$6F,$C2,$FC,$EF,$85,$FC,$0F,$67,$E5,$3A,$BD,$BC
	.dc.b	$7F,$2F,$3F,$5B,$78,$7E,$1C,$7B,$74,$DF,$67,$E2,$67,$35,$B9,$CE
	.dc.b	$1D,$AF,$E3,$CE,$5E,$44,$70,$7E,$C3,$93,$E4,$47,$F0,$3E,$4F,$90
	.dc.b	$FF,$14,$DF,$E3,$BF,$C1,$F9,$3E,$3E,$5F,$C0,$79,$3E,$3C,$FC,$43
	.dc.b	$FD,$57,$4E,$3F,$93,$1F,$AF,$E7,$FC,$8F,$1F,$C7,$8E,$1A,$7E,$4B
	.dc.b	$8B,$E4,$47,$AB,$83,$F5,$1D,$BF,$CC,$B2,$7C,$27,$E6,$89,$5C,$A3
	.dc.b	$2D,$8C,$BE,$B6,$EB,$F8,$36,$F1,$7D,$B6,$CF,$A9,$B9,$7E,$99,$B7
	.dc.b	$DA,$7A,$1D,$CF,$17,$C2,$7E,$BB,$E1,$3F,$5F,$F0,$9F,$B0,$F8,$4F
	.dc.b	$DD,$D7,$9F,$E1,$3F,$67,$F0,$9F,$B4,$F8,$4F,$E4,$1E,$1F,$EE,$EB
	.dc.b	$F8,$B7,$85,$FB,$8F,$73,$F7,$71,$FC,$67,$C2,$FD,$DB,$FB,$CE,$0F
	.dc.b	$2F,$07,$E3,$E3,$E9,$FA,$3F,$C0,$08,$FC,$76,$BF,$7D,$BC,$7F,$A7
	.dc.b	$6F,$1F,$E7,$DB,$3D,$0E,$44,$39,$16,$79,$36,$79,$3F,$5F,$E0,$FF
	.dc.b	$89,$FB,$7F,$D0,$FF,$80,$7F,$BD,$CF,$F6,$47,$FC,$98,$DF,$4F,$5E
	.dc.b	$DE,$FD,$2E,$1F,$09,$C7,$DB,$E5,$E9,$F3,$F6,$FA,$7C,$27,$5F,$52
	.dc.b	$67,$F9,$0D,$3F,$B3,$BD,$FD,$FF,$6F,$27,$F2,$4E,$3F,$94,$A9,$F0
	.dc.b	$9E,$72,$5F,$36,$BE,$96,$F2,$7F,$80,$A7,$8C,$EC,$86,$CB,$1B,$9E
	.dc.b	$3C,$9F,$EE,$6A,$E3,$5B,$25,$6D,$E3,$3F,$C5,$7C,$0F,$9B,$1E,$F8
	.dc.b	$F9,$DF,$01,$F3,$FE,$13,$E5,$CF,$C3,$E3,$F9,$6F,$C5,$E3,$F9,$6F
	.dc.b	$C7,$F8,$1F,$96,$FC,$9E,$1F,$B1,$F0,$9F,$67,$E0,$BE,$D7,$17,$A2
	.dc.b	$9C,$3F,$27,$87,$E5,$F0,$FA,$B8,$7D,$15,$EA,$E0,$F7,$57,$1B,$E1
	.dc.b	$3E,$C3,$9D,$F1,$AD,$C6,$67,$6F,$33,$75,$CE,$B8,$46,$0E,$0E,$27
	.dc.b	$8D,$FF,$47,$E1,$F9,$35,$F9,$3C,$3E,$4B,$F8,$9E,$0F,$92,$FE,$5F
	.dc.b	$83,$E4,$BF,$99,$E0,$F9,$2F,$E6,$F8,$3E,$4B,$F9,$DE,$0F,$92,$FE
	.dc.b	$4F,$7F,$92,$F9,$36,$FC,$8F,$8C,$E3,$FC,$AB,$67,$F5,$87,$EC,$CF
	.dc.b	$D8,$93,$6F,$D8,$B7,$5F,$D8,$B4,$E1,$1A,$24,$69,$B1,$A8,$C1,$1F
	.dc.b	$E7,$FD,$7F,$B9,$F8,$4F,$DD,$7C,$1F,$EE,$FE,$0B,$F7,$9D,$BF,$E3
	.dc.b	$FE,$0F,$F2,$7E,$1F,$84,$7F,$C9,$EF,$F9,$89,$FF,$25,$BF,$B5,$EC
	.dc.b	$E1,$F0,$3F,$BB,$8F,$E8,$F3,$FE,$4A,$7F,$75,$D7,$FB,$C7,$F8,$47
	.dc.b	$BF,$DF,$E0,$FE,$3F,$0F,$EE,$AF,$FC,$97,$5C,$BE,$07,$C8,$47,$04
	.dc.b	$3C,$8E,$5A,$67,$87,$FC,$FB,$3D,$23,$24,$8C,$CF,$16,$C7,$1E,$13
	.dc.b	$F0,$E7,$51,$8E,$A5,$F5,$BE,$37,$D6,$B5,$39,$6D,$F2,$9E,$87,$72
	.dc.b	$9C,$4B,$5D,$09,$3C,$A3,$85,$F8,$AB,$A3,$63,$D3,$E7,$7F,$A9,$B0
	.dc.b	$FF,$0B,$5F,$93,$D8,$FC,$01,$2B,$EC,$B5,$FC,$03,$71,$F9,$6D,$9E
	.dc.b	$53,$C0,$87,$02,$CF,$06,$C7,$07,$47,$E3,$7D,$8F,$C9,$7C,$0F,$E4
	.dc.b	$FD,$9F,$CA,$7B,$5F,$95,$F8,$1F,$CB,$7B,$7F,$97,$F6,$FF,$31,$E8
	.dc.b	$7E,$67,$E0,$7F,$35,$E8,$FE,$6F,$DB,$FC,$E7,$C0,$FE,$77,$D2,$FF
	.dc.b	$4E,$7F,$00,$FF,$9B,$4F,$F3,$67,$07,$F2,$27,$FC,$57,$C5,$E7,$8F
	.dc.b	$2F,$2F,$C9,$E4,$F8,$11,$FA,$56,$F3,$7E,$09,$BC,$DF,$3A,$74,$7C
	.dc.b	$02,$9F,$02,$FB,$13,$9B,$E2,$36,$36,$DB,$ED,$1C,$DF,$C8,$F3,$7C
	.dc.b	$88,$FD,$8F,$2F,$C8,$7F,$67,$C9,$F2,$1F,$DA,$F1,$FC,$87,$F6,$FC
	.dc.b	$5F,$21,$FD,$CF,$0F,$C8,$7F,$77,$C1,$F2,$1F,$93,$97,$CA,$F8,$6F
	.dc.b	$7E,$FF,$B9,$E8,$FC,$8F,$07,$E7,$EF,$F3,$3C,$1F,$97,$7F,$39,$87
	.dc.b	$E5,$5B,$3B,$8C,$70,$7E,$7A,$78,$92,$33,$5C,$E7,$51,$89,$D5,$73
	.dc.b	$25,$F9,$09,$1F,$A8,$8C,$B6,$32,$F9,$1B,$5F,$BC,$E1,$FD,$54,$63
	.dc.b	$0F,$11,$2B,$C3,$19,$69,$D4,$87,$52,$D7,$AA,$FC,$71,$92,$46,$67
	.dc.b	$A6,$1E,$F2,$57,$09,$D5,$7F,$25,$C3,$CE,$F6,$C3,$CD,$7C,$30,$E7
	.dc.b	$87,$19,$CB,$6B,$D2,$78,$A3,$33,$C0,$91,$D4,$DE,$1E,$27,$81,$89
	.dc.b	$E2,$FF,$14,$E7,$AE,$B0,$F2,$3C,$DA,$E7,$0F,$24,$62,$FB,$12,$F9
	.dc.b	$B6,$BC,$04,$9E,$13,$A2,$46,$6B,$FA,$36,$2C,$3F,$9F,$71,$BE,$0F
	.dc.b	$F8,$49,$2F,$FD,$4B,$19,$6C,$65,$64,$C7,$9C,$E6,$31,$3F,$C4,$63
	.dc.b	$13,$A8,$E6,$43,$9B,$96,$9C,$4D,$4F,$4C,$97,$E5,$9D,$C6,$2F,$BC
	.dc.b	$7F,$21,$F8,$0F,$E4,$BF,$07,$FC,$9B,$E0,$7F,$93,$FC,$1F,$F2,$8F
	.dc.b	$82,$FE,$53,$F0,$7F,$CA,$BD,$BF,$E5,$7E,$9F,$F2,$CF,$82,$FE,$5B
	.dc.b	$F0,$5F,$CB,$BE,$0B,$F9,$7F,$C1,$7F,$30,$F8,$2F,$E6,$3F,$05,$FC
	.dc.b	$CB,$E0,$BF,$99,$F9,$3E,$2E,$4F,$27,$1F,$E1,$4F,$E6,$9E,$9F,$F3
	.dc.b	$82,$3F,$41,$AF,$B3,$99,$18,$E7,$F8,$51,$89,$E0,$43,$84,$EF,$A3
	.dc.b	$EF,$B6,$3C,$26,$EB,$9C,$66,$FE,$6B,$FD,$33,$E0,$BF,$C0,$F5,$F3
	.dc.b	$BF,$80,$7F,$D4,$F7,$F4,$DB,$87,$F8,$25,$FF,$54,$C6,$7D,$2E,$2E
	.dc.b	$FE,$0E,$D6,$FD,$CB,$87,$F9,$E3,$0F,$F0,$8B,$FE,$F6,$5E,$A6,$B4
	.dc.b	$43,$E0,$F0,$BE,$0E,$FB,$B8,$F5,$B3,$EB,$4C,$BA,$7C,$0F,$BF,$8F
	.dc.b	$E5,$43,$C6,$FD,$33,$E3,$F7,$7A,$91,$F2,$35,$F8,$8D,$E2,$F8,$8D
	.dc.b	$E1,$F8,$8D,$9E,$87,$22,$1C,$8B,$3C,$9B,$3C,$9F,$89,$F0,$7F,$02
	.dc.b	$24,$7C,$7F,$6F,$E1,$F2,$FC,$4E,$5F,$8B,$F0,$3F,$1B,$CB,$F8,$FC
	.dc.b	$DF,$23,$E0,$7E,$4F,$C1,$FC,$8C,$BE,$87,$27,$CC,$F8,$2F,$9B,$F0
	.dc.b	$3F,$3B,$93,$D5,$D5,$F4,$B8,$FF,$DA,$FD,$3F,$E9,$B2,$3A,$6E,$3E
	.dc.b	$A3,$E0,$9E,$F2,$72,$7B,$DB,$C7,$EF,$69,$C2,$34,$48,$D3,$63,$F1
	.dc.b	$5C,$9F,$11,$FC,$C4,$FE,$1B,$93,$E2,$C7,$E3,$78,$FE,$2B,$F8,$FE
	.dc.b	$2F,$8A,$FE,$47,$87,$E2,$BF,$10,$87,$E8,$F7,$F8,$AF,$E3,$0F,$76
	.dc.b	$BF,$15,$FE,$00,$FE,$B1,$FD,$86,$3F,$B4,$69,$FB,$43,$F7,$26,$23
	.dc.b	$83,$82,$67,$FB,$87,$5C,$DF,$9A,$8C,$46,$AB,$B3,$4E,$D1,$A3,$F8
	.dc.b	$5F,$67,$EA,$BE,$07,$F5,$DF,$C0,$3F,$EA,$8C,$4F,$FA,$A2,$BF,$99
	.dc.b	$3F,$EA,$8B,$FE,$62,$BE,$0B,$F6,$7D,$FC,$BE,$1B,$9F,$C1,$FE,$DB
	.dc.b	$C1,$FD,$BF,$83,$DB,$D5,$FD,$B6,$5F,$BF,$F0,$7F,$6C,$F4,$F5,$BF
	.dc.b	$6D,$E2,$7F,$12,$F0,$FF,$6B,$C1,$E4,$F7,$BC,$E4,$BF,$F4,$EF,$0F
	.dc.b	$9A,$31,$1E,$66,$C6,$63,$7C,$9F,$EA,$1E,$25,$6A,$30,$D9,$5F,$85
	.dc.b	$E2,$62,$B8,$B8,$8D,$3E,$13,$11,$A7,$04,$BF,$A6,$FF,$21,$BF,$CD
	.dc.b	$47,$F2,$FE,$AF,$CC,$CF,$F3,$3E,$AF,$9E,$F4,$E2,$FB,$3F,$07,$F6
	.dc.b	$BD,$FF,$B7,$EC,$FD,$CF,$7F,$EE,$FC,$1F,$DE,$F8,$0F,$BF,$F0,$7F
	.dc.b	$80,$F8,$1F,$77,$0F,$A3,$8F,$E7,$F1,$3E,$0F,$F0,$A4,$DF,$D6,$D7
	.dc.b	$F0,$6D,$E2,$FC,$3B,$78,$FE,$73,$67,$E4,$37,$0F,$C7,$B6,$FC,$13
	.dc.b	$D8,$87,$63,$F9,$19,$F1,$C8,$7F,$57,$E3,$F8,$F8,$7F,$51,$8F,$E9
	.dc.b	$9E,$67,$E5,$5F,$A7,$C5,$F1,$DF,$AB,$F0,$5F,$1D,$FA,$FF,$07,$F1
	.dc.b	$DF,$B3,$BF,$C7,$7C,$78,$7B,$9F,$DF,$E5,$E8,$7C,$C4,$8D,$37,$5F
	.dc.b	$86,$F2,$6B,$93,$E2,$99,$29,$9D,$7E,$3B,$96,$B9,$AF,$15,$78,$0F
	.dc.b	$EF,$1E,$3F,$E3,$2F,$FB,$C1,$FC,$96,$BF,$B8,$79,$FF,$D2,$5F,$8E
	.dc.b	$7E,$34,$F8,$E7,$E6,$B9,$BF,$83,$E9,$F1,$E3,$F8,$1E,$9F,$1D,$FE
	.dc.b	$0B,$A7,$C7,$7C,$9C,$1F,$BB,$E6,$FD,$6C,$7E,$93,$A7,$F2,$3B,$7E
	.dc.b	$97,$6F,$DB,$4F,$E9,$75,$DB,$4F,$D8,$57,$02,$3C,$C8,$6F,$7F,$D9
	.dc.b	$D7,$42,$57,$F9,$93,$C4,$FE,$45,$5F,$B6,$F8,$CF,$E4,$7F,$19,$FC
	.dc.b	$93,$E2,$7F,$92,$F8,$DF,$B0,$AF,$E4,$3F,$0D,$FE,$15,$F8,$2F,$F1
	.dc.b	$3C,$7F,$95,$1F,$F7,$9E,$9F,$CA,$B7,$FD,$9C,$FF,$02,$ED,$E1,$F0
	.dc.b	$DC,$7E,$1B,$97,$C3,$7A,$AB,$FC,$29,$67,$F8,$8E,$2F,$86,$F0,$12
	.dc.b	$3F,$88,$46,$5B,$19,$7C,$8D,$E2,$FF,$36,$4E,$E7,$27,$02,$18,$24
	.dc.b	$70,$9E,$5C,$5E,$5D,$17,$CE,$B5,$7C,$9F,$E3,$AF,$F7,$43,$F6,$EF
	.dc.b	$F1,$D7,$F2,$27,$F7,$53,$FC,$85,$E6,$7F,$54,$DB,$E7,$57,$CF,$F8
	.dc.b	$6F,$A1,$F0,$D2,$FE,$0F,$C3,$7D,$3F,$86,$FA,$9F,$0D,$F5,$7E,$1B
	.dc.b	$CF,$5F,$2B,$7F,$C9,$34,$F3,$6F,$EB,$CF,$D4,$D9,$E3,$BF,$B9,$B3
	.dc.b	$AA,$E4,$7C,$6A,$E0,$F4,$33,$BF,$C4,$74,$62,$36,$E2,$FB,$3F,$C7
	.dc.b	$8F,$90,$49,$F9,$0F,$F9,$D7,$C1,$FD,$C3,$FE,$77,$F1,$3E,$43,$F8
	.dc.b	$4C,$3F,$2D,$F0,$DF,$1E,$3F,$2D,$BF,$E6,$7E,$1B,$F3,$5F,$0D,$F2
	.dc.b	$27,$F1,$7B,$7B,$B6,$F8,$FA,$F9,$27,$CD,$97,$C1,$8C,$6B,$ED,$26
	.dc.b	$DE,$C6,$D7,$BC,$93,$C2,$74,$48,$CD,$75,$25,$F8,$E3,$E2,$5F,$FA
	.dc.b	$4E,$BF,$8E,$47,$FA,$2F,$C8,$FE,$83,$1F,$E9,$5F,$3B,$FC,$7F,$3F
	.dc.b	$B8,$3E,$21,$FA,$8D,$BE,$25,$FE,$93,$6F,$89,$3F,$AC,$D7,$E2,$3F
	.dc.b	$BD,$D7,$C4,$FE,$D8,$FD,$AF,$C3,$7F,$03,$23,$FB,$46,$BF,$C1,$5B
	.dc.b	$7F,$B2,$6E,$5F,$91,$6C,$F4,$39,$10,$E4,$59,$E4,$D9,$E4,$FF,$BF
	.dc.b	$3C,$1F,$E3,$AF,$F8,$E3,$C1,$FE,$3A,$FF,$85,$3C,$0F,$E3,$AF,$EC
	.dc.b	$0F,$E1,$7F,$0D,$FC,$77,$E1,$BF,$61,$1F,$C3,$BE,$1B,$F9,$07,$A7
	.dc.b	$FC,$87,$AF,$F6,$13,$FC,$4B,$AF,$F9,$26,$7E,$AE,$7F,$E0,$BB,$FE
	.dc.b	$2B,$F8,$07,$FD,$93,$7F,$E2,$C6,$C9,$A4,$DF,$FD,$84,$D9,$FF,$17
	.dc.b	$35,$70,$27,$F8,$A4,$E6,$31,$3D,$E7,$99,$1E,$6D,$79,$BC,$8C,$47
	.dc.b	$2A,$F3,$47,$93,$9B,$CB,$D7,$2F,$CB,$CF,$E8,$EF,$F4,$FB,$FE,$AF
	.dc.b	$86,$F5,$FB,$7E,$CF,$6F,$DB,$F0,$BE,$EF,$87,$F7,$FB,$9F,$03,$E1
	.dc.b	$7E,$0F,$C2,$FC,$2F,$5F,$E1,$F8,$DE,$4E,$2F,$37,$07,$FA,$43,$95
	.dc.b	$F0,$9F,$20,$91,$F2,$1A,$FB,$8F,$A4,$E3,$9B,$DD,$3C,$A7,$81,$0E
	.dc.b	$1C,$DF,$5E,$B8,$5F,$D0,$6C,$F4,$38,$5F,$48,$CD,$71,$9F,$B7,$CD
	.dc.b	$F2,$5E,$B8,$39,$BF,$03,$F0,$FF,$2E,$7F,$03,$CB,$F2,$DF,$C1,$FC
	.dc.b	$2F,$CB,$7F,$0B,$F0,$FF,$88,$F8,$5F,$97,$1F,$88,$E2,$FE,$D8,$FA
	.dc.b	$BF,$80,$7F,$D4,$19,$B8,$B3,$C1,$F7,$E7,$FD,$45,$C3,$ED,$BF,$85
	.dc.b	$EF,$FF,$A5,$30,$F8,$D1,$E8,$E2,$F6,$C6,$30,$F8,$13,$F0,$AF,$93
	.dc.b	$B3,$07,$C7,$30,$48,$E2,$D8,$E3,$18,$23,$82,$47,$EB,$78,$7E,$64
	.dc.b	$FD,$CE,$1F,$97,$3F,$AA,$EB,$FC,$AB,$FA,$CE,$BF,$D7,$F0,$FC,$B8
	.dc.b	$FA,$7C,$3F,$2D,$FD,$8F,$3F,$E5,$5F,$F5,$06,$FF,$2B,$83,$E4,$EB
	.dc.b	$FE,$05,$8F,$93,$3E,$DD,$3D,$7C,$1F,$26,$BF,$35,$3F,$B5,$3F,$66
	.dc.b	$43,$F0,$6E,$B9,$7F,$63,$19,$AF,$CB,$CE,$78,$3F,$80,$3F,$A4,$9E
	.dc.b	$26,$DC,$4F,$08,$C1,$FE,$BE,$C3,$FD,$BD,$1F,$81,$F5,$BD,$73,$9F
	.dc.b	$83,$FE,$27,$DD,$8C,$98,$3E,$1B,$0B,$E1,$AF,$C3,$7F,$00,$C7,$E3
	.dc.b	$FB,$9F,$C0,$1F,$E3,$7D,$FC,$3E,$0F,$8F,$C1,$FF,$00,$9F,$E3,$FE
	.dc.b	$0F,$F0,$07,$CF,$A7,$C2,$7F,$A9,$3E,$2F,$67,$C2,$48,$FA,$AD,$7F
	.dc.b	$40,$DE,$1F,$CF,$37,$8B,$EC,$36,$7E,$23,$6B,$B4,$74,$21,$D0,$B5
	.dc.b	$D1,$FD,$E9,$FC,$E3,$AF,$F2,$77,$FC,$DB,$87,$F7,$93,$FC,$03,$C0
	.dc.b	$FD,$E3,$FB,$5F,$86,$F9,$7E,$F7,$CC,$F8,$6F,$9B,$F0,$7F,$3B,$E1
	.dc.b	$BE,$7F,$C2,$7D,$0F,$86,$FA,$3F,$0D,$F4,$BE,$07,$D1,$D4,$E3,$FF
	.dc.b	$0D,$FF,$00,$FF,$9E,$38,$FF,$9E,$37,$8B,$FC,$B8,$DE,$0F,$F8,$02
	.dc.b	$DE,$43,$49,$1F,$83,$6B,$FE,$79,$AE,$AE,$48,$66,$BB,$B8,$3F,$96
	.dc.b	$CF,$26,$99,$9E,$6F,$CC,$3F,$93,$F2,$7C,$C7,$FB,$67,$97,$F3,$27
	.dc.b	$F9,$DF,$27,$CC,$7D,$F8,$F9,$FE,$17,$E6,$5F,$8C,$F7,$FC,$2F,$AF
	.dc.b	$E1,$7E,$6C,$66,$BE,$0F,$C0,$FB,$FE,$0B,$E0,$7C,$2F,$97,$7F,$5D
	.dc.b	$FF,$A6,$7C,$7F,$A0,$FC,$A3,$45,$3A,$70,$7B,$F5,$F6,$C7,$C0,$32
	.dc.b	$70,$CB,$E8,$B6,$78,$5F,$82,$38,$12,$B9,$92,$34,$FF,$5E,$F3,$BE
	.dc.b	$45,$FF,$84,$78,$7F,$92,$DF,$F5,$2F,$3B,$F9,$2B,$FD,$53,$CE,$FB
	.dc.b	$D5,$F2,$63,$F9,$0F,$67,$E0,$78,$7E,$44,$7F,$24,$EA,$FC,$27,$0F
	.dc.b	$C7,$8F,$7F,$07,$A5,$FF,$36,$F3,$FC,$67,$F5,$44,$DB,$F5,$4D,$DF
	.dc.b	$F6,$AF,$EB,$9A,$E4,$FD,$B9,$92,$99,$CF,$F7,$2D,$AF,$05,$76,$25
	.dc.b	$74,$7F,$B1,$7B,$DF,$C7,$3C,$AF,$E3,$BE,$57,$F2,$2C,$3F,$A0,$F9
	.dc.b	$5F,$C8,$3C,$AF,$E3,$F1,$FD,$13,$C3,$FD,$CB,$F7,$BE,$0B,$F7,$2F
	.dc.b	$F4,$8F,$0B,$F5,$1E,$DF,$EE,$63,$FA,$5F,$85,$FB,$97,$D5,$C1,$E8
	.dc.b	$EB,$FC,$5B,$EA,$DF,$CF,$3E,$E6,$C6,$78,$3D,$44,$9D,$36,$35,$18
	.dc.b	$23,$82,$70,$FC,$69,$E0,$D9,$DB,$67,$FC,$B5,$D5,$F8,$CF,$37,$FC
	.dc.b	$35,$5F,$E4,$43,$FB,$81,$FE,$1A,$3F,$B8,$1F,$1C,$FE,$D7,$E3,$7E
	.dc.b	$D1,$FF,$0B,$78,$DF,$C7,$DF,$C6,$1F,$DC,$3C,$6F,$E3,$EF,$F7,$2F
	.dc.b	$1B,$F8,$FF,$8D,$FC,$7A,$3F,$1C,$7E,$A8,$87,$E9,$74,$FD,$19,$1F
	.dc.b	$DA,$19,$2E,$BF,$A9,$6C,$6F,$4F,$D9,$C7,$49,$D4,$62,$39,$3F,$11
	.dc.b	$FE,$84,$43,$FC,$7F,$E3,$7F,$17,$AF,$F3,$CF,$8D,$FC,$5D,$FE,$67
	.dc.b	$5F,$AF,$AF,$F1,$C7,$7F,$EB,$DF,$E1,$3E,$2F,$F1,$67,$F9,$4F,$77
	.dc.b	$EB,$DF,$E5,$7D,$DF,$09,$FE,$CF,$E2,$7F,$08,$7D,$64,$C3,$D1,$A7
	.dc.b	$AA,$75,$18,$CB,$E0,$46,$09,$5D,$6B,$83,$FE,$85,$3D,$CF,$F8,$C5
	.dc.b	$FF,$4C,$E9,$EF,$7F,$D4,$5A,$7B,$DF,$E9,$F8,$7B,$EB,$FC,$A5,$A7
	.dc.b	$BE,$7F,$A2,$E9,$EF,$9F,$EB,$3A,$7B,$DF,$EB,$7A,$7B,$F9,$7F,$4F
	.dc.b	$CB,$F9,$52,$6F,$F8,$6C,$7F,$5B,$18,$6C,$65,$FE,$0C,$DB,$FD,$6C
	.dc.b	$EE,$7F,$CD,$5D,$FF,$C1,$5F,$F2,$1F,$77,$F0,$57,$FC,$DD,$DD,$FC
	.dc.b	$15,$FF,$23,$F9,$BF,$E2,$6E,$EF,$E0,$B1,$FD,$83,$BB,$F8,$2B,$FE
	.dc.b	$75,$EE,$FE,$0A,$FE,$7C,$FE,$0A,$7E,$7C,$F3,$9F,$D3,$F4,$FF,$28
	.dc.b	$73,$FF,$09,$8F,$F3,$F3,$75,$FF,$3E,$37,$9F,$73,$CC,$D9,$67,$6D
	.dc.b	$8D,$BA,$24,$69,$B1,$A8,$C1,$FD,$3E,$3F,$07,$1F,$ED,$6F,$17,$E3
	.dc.b	$4F,$FA,$C3,$C5,$FE,$9E,$FF,$AC,$BC,$5F,$8D,$7F,$C1,$0F,$E9,$E7
	.dc.b	$F9,$C3,$B3,$F8,$23,$FE,$9B,$F1,$3F,$95,$3F,$E9,$DF,$07,$DC,$FF
	.dc.b	$97,$73,$F8,$9C,$BF,$CA,$BB,$5E,$DF,$F8,$64,$95,$FD,$D1,$AF,$F8
	.dc.b	$51,$B9,$FF,$8A,$1B,$B7,$F8,$A1,$B3,$FE,$28,$6F,$67,$F8,$A1,$B7
	.dc.b	$E0,$9E,$C4,$3B,$1F,$E1,$DB,$FD,$43,$FE,$1E,$3F,$52,$7F,$2A,$3F
	.dc.b	$C9,$07,$F2,$A3,$FC,$53,$5F,$A9,$BF,$F4,$97,$67,$EA,$6B,$FC,$7F
	.dc.b	$D9,$FA,$99,$FF,$1F,$F6,$7E,$A5,$FF,$20,$F6,$79,$B3,$FF,$23,$97
	.dc.b	$2F,$36,$1F,$05,$B5,$F0,$5A,$7C,$12,$E5,$F0,$5B,$9F,$C1,$7D,$E4
	.dc.b	$AE,$8D,$AE,$8D,$3F,$D7,$87,$E2,$85,$7F,$86,$5F,$8A,$7F,$86,$4F
	.dc.b	$E9,$C7,$F8,$64,$FF,$17,$BF,$15,$FF,$50,$67,$F1,$67,$FC,$C3,$9F
	.dc.b	$C5,$9F,$F3,$0E,$7F,$15,$FF,$31,$E7,$F0,$DF,$F4,$8F,$63,$6F,$C7
	.dc.b	$E5,$FE,$A4,$6C,$66,$7F,$54,$7E,$4C,$B3,$B6,$C6,$DF,$D5,$9B,$2C
	.dc.b	$6D,$B1,$BC,$FF,$5F,$97,$EC,$4F,$F6,$97,$67,$F1,$5A,$FF,$80,$1D
	.dc.b	$9F,$B0,$9F,$F6,$B7,$67,$F1,$59,$5F,$FC,$01,$EC,$FD,$8C,$62,$BF
	.dc.b	$DB,$9E,$07,$F9,$FF,$B3,$F8,$BD,$A7,$FD,$C1,$D9,$FC,$62,$F8,$1F
	.dc.b	$EC,$DF,$03,$FB,$BB,$E8,$DF,$F6,$7D,$BF,$E9,$0F,$85,$4C,$27,$0F
	.dc.b	$FA,$42,$31,$18,$E6,$46,$23,$91,$67,$83,$67,$6D,$9E,$0F,$FA,$79
	.dc.b	$B3,$C2,$3B,$B5,$F0,$46,$A7,$9C,$69,$AE,$8E,$C4,$35,$1C,$48,$F9
	.dc.b	$7E,$0B,$FD,$55,$F0,$9F,$EA,$BF,$85,$FF,$56,$7C,$2F,$FA,$B7,$DF
	.dc.b	$FF,$57,$7C,$27,$FA,$BF,$E1,$3F,$D6,$1F,$09,$32,$FF,$45,$FC,$27
	.dc.b	$FA,$CF,$E1,$3F,$DE,$DE,$7F,$D5,$9F,$F4,$D2,$F3,$B3,$E5,$6B,$D4
	.dc.b	$DF,$97,$FE,$BB,$E2,$F2,$E0,$A7,$B1,$F3,$A3,$D6,$7B,$4F,$80,$60
	.dc.b	$86,$08,$71,$29,$F4,$4E,$46,$48,$64,$E6,$63,$97,$EA,$37,$95,$CB
	.dc.b	$F2,$A3,$FD,$5B,$CB,$F0,$BC,$EF,$87,$E6,$FC,$4F,$84,$F9,$55,$FE
	.dc.b	$B2,$E3,$F9,$4F,$E6,$32,$70,$BE,$13,$E5,$FC,$27,$CC,$EF,$F9,$BB
	.dc.b	$FC,$EF,$7B,$E5,$63,$F3,$BE,$0B,$E5,$3F,$81,$E2,$F9,$2F,$AC,$F8
	.dc.b	$39,$FA,$9C,$1E,$BE,$3F,$A2,$EB,$8F,$F1,$AE,$4F,$8E,$70,$24,$F3
	.dc.b	$24,$70,$8C,$B6,$3F,$DF,$7C,$7F,$1E,$7F,$DE,$3C,$7F,$1D,$FC,$3F
	.dc.b	$17,$C7,$7F,$DA,$FD,$3F,$1B,$C2,$FC,$66,$0C,$5C,$7F,$90,$E2,$F8
	.dc.b	$F8,$7E,$43,$87,$E3,$BF,$92,$E0,$F8,$EF,$CF,$E3,$70,$FC,$88,$FC
	.dc.b	$57,$0F,$C8,$CF,$D7,$C3,$F1,$27,$E1,$1F,$84,$3D,$AE,$5A,$7C,$68
	.dc.b	$E2,$DD,$3E,$11,$27,$84,$FE,$03,$DB,$FD,$4F,$BF,$FB,$3B,$FF,$7B
	.dc.b	$F7,$FE,$CD,$FF,$7C,$F7,$7E,$D1,$FD,$6F,$C2,$7E,$C3,$E0,$FF,$8D
	.dc.b	$71,$FC,$6D,$1E,$07,$ED,$6F,$FE,$00,$F8,$1F,$B6,$F0,$3F,$67,$1F
	.dc.b	$B6,$E0,$FD,$9B,$FC,$93,$0E,$5C,$1F,$B4,$9F,$89,$C4,$F6,$3F,$7E
	.dc.b	$4B,$FD,$FB,$5F,$9C,$DB,$FB,$33,$C6,$76,$43,$65,$8D,$CF,$1E,$5F
	.dc.b	$E6,$FC,$7F,$A6,$8C,$57,$F7,$4B,$E8,$F2,$25,$F8,$3E,$0F,$F8,$A9
	.dc.b	$FC,$5B,$E0,$FF,$8B,$FC,$27,$C8,$CB,$AF,$F0,$03,$E1,$B1,$8F,$F8
	.dc.b	$02,$1F,$0F,$8D,$50,$F4,$F0,$BE,$06,$98,$F5,$33,$EA,$6B,$83,$7E
	.dc.b	$A7,$0F,$87,$E3,$F0,$FC,$BE,$1F,$DB,$A7,$BF,$6F,$F3,$81,$E3,$8D
	.dc.b	$3F,$5F,$83,$D3,$18,$8F,$E1,$9A,$A3,$1C,$9F,$AA,$8C,$4F,$13,$A1
	.dc.b	$6F,$DB,$5C,$2F,$AC,$E6,$B8,$BD,$A7,$4F,$23,$06,$DC,$D7,$48,$E5
	.dc.b	$E9,$7C,$8F,$82,$FA,$91,$FD,$23,$93,$E5,$F2,$7C,$CF,$63,$E6,$F2
	.dc.b	$4A,$F1,$7B,$BF,$43,$E0,$BE,$6C,$79,$78,$FE,$6B,$E7,$E3,$F9,$AF
	.dc.b	$8C,$87,$98,$7C,$03,$5F,$7C,$67,$8B,$E4,$70,$7A,$B8,$7E,$0C,$62
	.dc.b	$79,$B5,$E1,$C5,$CA,$30,$47,$04,$8F,$91,$19,$6C,$67,$8B,$E4,$47
	.dc.b	$DC,$C3,$E4,$3F,$DB,$BA,$7F,$36,$FD,$5E,$1F,$90,$FD,$7E,$0F,$90
	.dc.b	$FE,$13,$83,$E4,$3F,$6F,$9B,$F2,$8F,$C7,$21,$F8,$AE,$5F,$81,$B7
	.dc.b	$E3,$77,$F8,$D3,$FB,$1E,$BF,$CB,$F1,$7E,$CB,$83,$F7,$33,$FB,$BE
	.dc.b	$0F,$DC,$CE,$5B,$5C,$09,$3B,$3A,$17,$83,$F4,$4F,$EF,$27,$8B,$9A
	.dc.b	$E8,$FF,$03,$EB,$FD,$FB,$F9,$7F,$6B,$F6,$BF,$05,$FB,$6F,$6F,$F6
	.dc.b	$FD,$DF,$B8,$F8,$2F,$DB,$DF,$E7,$7B,$FF,$70,$FE,$7F,$DC,$FE,$01
	.dc.b	$FC,$03,$FE,$44,$E0,$FE,$A6,$FF,$A1,$27,$F4,$9E,$07,$EE,$DF,$D5
	.dc.b	$43,$E2,$3D,$B1,$F8,$EE,$BF,$4B,$EB,$3D,$E5,$E2,$F7,$E1,$F2,$A3
	.dc.b	$1C,$3E,$E9,$CF,$17,$2A,$D1,$1C,$12,$BA,$C6,$5B,$19,$9E,$71,$F2
	.dc.b	$4F,$F1,$CF,$17,$ED,$1F,$ED,$1E,$07,$ED,$1F,$E0,$5C,$1F,$8C,$C3
	.dc.b	$F8,$66,$FF,$2A,$7F,$88,$F8,$5F,$B4,$7F,$89,$E7,$FB,$47,$F8,$AE
	.dc.b	$FF,$B4,$7F,$4E,$79,$0F,$87,$CD,$F8,$ED,$3F,$43,$8F,$E9,$5B,$AF
	.dc.b	$E9,$7B,$BF,$66,$DE,$8F,$CB,$E1,$FB,$48,$C3,$4F,$D6,$92,$B9,$B6
	.dc.b	$B9,$B7,$83,$F5,$EF,$F8,$27,$BB,$F5,$EF,$F2,$EE,$EF,$D7,$BF,$CC
	.dc.b	$3B,$BE,$26,$3F,$D3,$3B,$BF,$5F,$3F,$CD,$3B,$BF,$5E,$FF,$36,$EF
	.dc.b	$F8,$9B,$7F,$35,$EF,$F4,$CF,$F1,$5E,$0F,$85,$B7,$93,$7F,$39,$1C
	.dc.b	$13,$83,$8E,$1F,$1E,$31,$7F,$06,$B8,$56,$B4,$F8,$B3,$F4,$A7,$FA
	.dc.b	$EF,$07,$C0,$C3,$FB,$CF,$07,$C0,$7F,$A5,$70,$7C,$07,$FA,$67,$07
	.dc.b	$C0,$C3,$FB,$47,$07,$C0,$9F,$EA,$1C,$1F,$01,$FE,$A5,$C1,$F0,$34
	.dc.b	$FE,$E3,$C1,$E3,$C3,$FA,$57,$07,$E0,$78,$BF,$40,$4E,$FF,$D5,$E3
	.dc.b	$F9,$BE,$0F,$CC,$92,$75,$7F,$C1,$6B,$78,$FE,$7E,$73,$3F,$AC,$9F
	.dc.b	$E0,$EF,$F9,$17,$A3,$FB,$DF,$7F,$F0,$88,$FE,$C1,$DD,$FC,$21,$FE
	.dc.b	$C5,$DD,$F9,$FC,$FF,$43,$1F,$D9,$7A,$3F,$C0,$7D,$DF,$C2,$23,$F8
	.dc.b	$4E,$BF,$C1,$5F,$D0,$9F,$C1,$BB,$7F,$82,$EF,$F9,$FE,$E7,$C3,$FF
	.dc.b	$10,$25,$7F,$0C,$6B,$FE,$89,$8D,$B9,$21,$98,$DE,$5F,$C6,$1B,$C3
	.dc.b	$FE,$2C,$AE,$35,$B3,$8D,$6D,$B3,$BA,$E3,$0C,$56,$F8,$4F,$E4,$FF
	.dc.b	$09,$FC,$A3,$E2,$3F,$94,$FC,$2F,$F2,$AF,$84,$FE,$57,$F1,$1F,$CB
	.dc.b	$3E,$1F,$F9,$6F,$99,$FC,$BB,$DE,$9A,$FF,$2A,$F7,$BF,$98,$FC,$37
	.dc.b	$F3,$2F,$5E,$4F,$F2,$EF,$86,$F1,$F0,$F8,$B8,$BE,$2F,$2F,$CB,$73
	.dc.b	$C7,$EE,$E5,$F6,$61,$EF,$23,$82,$72,$FB,$E3,$E0,$9E,$F9,$E5,$19
	.dc.b	$8E,$2D,$9E,$31,$BA,$F0,$9E,$2F,$63,$E2,$C7,$F4,$6F,$7F,$E2,$BF
	.dc.b	$D2,$3D,$FF,$E9,$FE,$BF,$C5,$8F,$E9,$7C,$5F,$11,$FC,$07,$17,$F5
	.dc.b	$4E,$2F,$8B,$1F,$D4,$3D,$7F,$F2,$B7,$F0,$0F,$FB,$CE,$3F,$0D,$C5
	.dc.b	$E2,$7F,$0B,$C3,$F1,$5F,$6F,$17,$C9,$6F,$17,$B9,$F5,$18,$E0,$E4
	.dc.b	$FC,$03,$25,$9E,$4D,$9E,$51,$8E,$1F,$8A,$DE,$2F,$C9,$71,$7C,$4B
	.dc.b	$FC,$E7,$17,$C4,$7F,$3D,$C5,$F1,$1F,$ED,$BC,$BF,$88,$7F,$B7,$F0
	.dc.b	$FC,$5C,$BF,$BA,$F0,$FC,$49,$FD,$37,$0F,$C4,$CF,$FB,$DF,$07,$AA
	.dc.b	$FF,$07,$C1,$FD,$3F,$85,$EC,$7C,$A2,$63,$FE,$65,$8C,$B6,$32,$FF
	.dc.b	$BE,$E3,$2D,$8C,$D7,$FB,$6A,$72,$D8,$CE,$1F,$E9,$E9,$CB,$63,$35
	.dc.b	$FA,$F9,$CB,$63,$3C,$09,$CB,$FB,$4E,$CF,$C1,$FB,$1F,$B3,$BF,$F0
	.dc.b	$AF,$67,$EC,$DF,$F0,$BF,$57,$E2,$3E,$1F,$F1,$3F,$0F,$F8,$AF,$87
	.dc.b	$FC,$5F,$C2,$7E,$33,$E1,$FF,$1B,$F0,$BF,$B2,$C7,$F8,$97,$7F,$E4
	.dc.b	$3E,$1F,$F2,$3F,$0F,$E4,$E1,$F8,$7C,$BF,$8F,$BF,$C2,$77,$FD,$1E
	.dc.b	$5F,$D4,$12,$3F,$17,$CD,$F8,$6E,$FF,$C4,$92,$B8,$4E,$AB,$87,$37
	.dc.b	$E2,$E3,$04,$7E,$13,$FE,$34,$F1,$7E,$13,$FC,$A7,$C5,$F8,$4F,$F2
	.dc.b	$BE,$5F,$D8,$3F,$CB,$7C,$2F,$C6,$61,$FC,$9F,$BB,$F6,$13,$FE,$3E
	.dc.b	$F1,$3E,$16,$BF,$E4,$4E,$2F,$C6,$CF,$E9,$CF,$D7,$FA,$DE,$6D,$FD
	.dc.b	$51,$93,$E1,$96,$BE,$06,$FE,$ED,$FD,$84,$AE,$35,$C0,$FE,$85,$E2
	.dc.b	$7E,$A6,$FF,$CA,$BE,$27,$EA,$5F,$F2,$BF,$89,$FA,$9A,$FE,$91,$E2
	.dc.b	$7E,$A5,$FE,$16,$7E,$A8,$FE,$60,$7B,$4F,$E1,$87,$E8,$77,$F6,$BE
	.dc.b	$1A,$FD,$4E,$9E,$DD,$3D,$19,$F9,$A7,$D7,$5E,$DB,$D3,$4F,$AD,$3B
	.dc.b	$8C,$C7,$B0,$FB,$07,$F2,$73,$FC,$DB,$C5,$FE,$74,$CB,$D9,$7F,$E7
	.dc.b	$4C,$BD,$8F,$F9,$DB,$83,$EC,$DF,$F3,$9F,$13,$F8,$24,$FF,$9E,$BC
	.dc.b	$4F,$E0,$93,$FD,$2C,$F2,$70,$7F,$1A,$E6,$FE,$09,$E1,$FE,$6C,$93
	.dc.b	$FC,$04,$87,$EF,$CB,$7F,$BD,$6E,$BF,$BB,$69,$FB,$92,$DF,$EE,$6B
	.dc.b	$AC,$73,$3F,$BF,$78,$7F,$BA,$7F,$72,$43,$FC,$05,$E3,$7F,$6C,$7F
	.dc.b	$8B,$78,$5F,$BA,$7F,$8C,$78,$3F,$BA,$7F,$8D,$78,$1F,$BA,$7F,$D2
	.dc.b	$FD,$1F,$9A,$CF,$F8,$97,$7F,$EE,$A7,$CB,$DD,$FB,$AE,$8F,$E2,$DF
	.dc.b	$0F,$FD,$38,$93,$FC,$95,$AF,$F2,$36,$F0,$7F,$21,$6E,$1C,$A7,$81
	.dc.b	$0E,$05,$9E,$0D,$8E,$0E,$89,$1A,$6C,$6A,$30,$47,$04,$8F,$E5,$93
	.dc.b	$E4,$69,$DC,$A7,$42,$CF,$36,$CF,$3E,$0F,$E6,$54,$99,$C3,$FD,$37
	.dc.b	$E1,$FF,$BA,$92,$3F,$A6,$35,$FE,$94,$DC,$3F,$A4,$37,$0F,$E8,$CD
	.dc.b	$9F,$E8,$8D,$C3,$FA,$13,$72,$34,$C7,$C3,$FF,$81,$BE,$1F,$FC,$0F
	.dc.b	$F0,$FF,$E0,$8F,$87,$FF,$04,$FA,$9F,$E0,$AF,$53,$FC,$17,$EE,$7F
	.dc.b	$83,$3E,$0B,$D9,$AB,$76,$FE,$FD,$DE,$5B,$DD,$DD,$6A,$FF,$C9,$12
	.dc.b	$F8,$2C,$E1,$C6,$1E,$F6,$2B,$7F,$05,$72,$EB,$EB,$DB,$7B,$D7,$DC
	.dc.b	$C7,$BD,$EF,$C7,$E8,$F5,$FC,$87,$E7,$90,$F9,$E5,$DB,$E7,$B7,$5E
	.dc.b	$0E,$89,$1A,$6C,$6A,$30,$47,$04,$8F,$AE,$75,$3F,$BE,$73,$FD,$37
	.dc.b	$FC,$1D,$CD,$F4,$DF,$F0,$87,$2F,$D3,$7F,$C2,$5C,$9F,$4D,$FC,$76
	.dc.b	$1F,$37,$7F,$A3,$C7,$F1,$78,$FE,$37,$1F,$C7,$E3,$F9,$1C,$7F,$27
	.dc.b	$8B,$E4,$65,$CB,$7F,$A5,$C3,$F2,$61,$D5,$FE,$25,$EB,$FD,$36,$BF
	.dc.b	$9E,$9F,$D3,$35,$FD,$33,$76,$FD,$33,$70,$FD,$37,$5A,$F9,$B6,$F9
	.dc.b	$B5,$E6,$DB,$E6,$DA,$E7,$5C,$89,$5C,$9A,$F2,$6D,$72,$73,$87,$B6
	.dc.b	$78,$10,$E0,$59,$E0,$DA,$F0,$B4,$FE,$06,$4C,$3E,$03,$70,$F4,$37
	.dc.b	$0F,$1D,$72,$8E,$E4,$BE,$ED,$8F,$E4,$38,$79,$E3,$07,$33,$C5,$87
	.dc.b	$80,$98,$78,$1B,$87,$82,$78,$1C,$A3,$2F,$29,$E0,$49,$E1,$3A,$9E
	.dc.b	$13,$97,$84,$FF,$42,$AF,$36,$1E,$A2,$61,$EF,$6D,$78,$9B,$87,$56
	.dc.b	$E1,$D6,$7B,$E1,$E2,$8C,$4E,$A3,$04,$9F,$B5,$83,$93,$F8,$77,$C1
	.dc.b	$7F,$09,$24,$FE,$84,$FE,$2A,$E1,$FE,$3D,$97,$07,$59,$7F,$1E,$9D
	.dc.b	$63,$FB,$1E,$74,$62,$F8,$5F,$23,$A9,$CA,$76,$D9,$E0,$E8,$E1,$7D
	.dc.b	$DD,$10,$CC,$BD,$FF,$E4,$9F,$03,$FC,$97,$2F,$E4,$DF,$0D,$FC,$9F
	.dc.b	$E1,$BF,$94,$7C,$1F,$F2,$9F,$86,$FE,$55,$F0,$9F,$CA,$FE,$1B,$F9
	.dc.b	$67,$C2,$FF,$2D,$F8,$5F,$E5,$DF,$0B,$FC,$C7,$F8,$03,$FA,$0F,$07
	.dc.b	$91,$70,$E7,$E1,$B5,$EC,$FA,$A3,$D1,$5F,$C2,$B8,$3F,$A3,$5B,$D1
	.dc.b	$E8,$49,$F9,$B1,$96,$C6,$5F,$03,$79,$FF,$90,$4E,$E7,$27,$02,$18
	.dc.b	$24,$70,$9E,$5C,$FE,$46,$CF,$B2,$BA,$56,$E3,$9B,$5E,$75,$C5,$CB
	.dc.b	$5C,$9C,$EB,$AE,$0B,$EB,$5B,$BE,$84,$70,$4B,$E4,$7B,$3A,$3F,$38
	.dc.b	$F9,$FD,$FF,$AC,$47,$E5,$B5,$FB,$0D,$AF,$92,$DA,$FA,$AD,$9E,$87
	.dc.b	$22,$1C,$8B,$3C,$9B,$3C,$9F,$B7,$F0,$7F,$89,$F6,$BF,$15,$EA,$7E
	.dc.b	$2F,$E0,$3F,$19,$ED,$7E,$37,$E0,$3F,$1D,$ED,$FE,$3F,$DB,$FC,$87
	.dc.b	$C0,$7E,$47,$D7,$FC,$97,$C2,$FE,$4F,$E0,$7F,$29,$F0,$3F,$95,$F8
	.dc.b	$0F,$CB,$7C,$37,$E5,$FE,$07,$D5,$D1,$F3,$F9,$BF,$C3,$3C,$FF,$22
	.dc.b	$31,$CB,$E8,$3D,$E7,$B4,$9C,$FE,$D6,$F3,$FB,$5A,$70,$8D,$12,$34
	.dc.b	$D8,$D4,$60,$8E,$09,$3F,$85,$E7,$F9,$11,$FA,$4E,$6F,$90,$FE,$9B
	.dc.b	$97,$E4,$3F,$A8,$E4,$F9,$0F,$EA,$B8,$FE,$43,$FA,$CE,$2F,$90,$FE
	.dc.b	$BB,$87,$E4,$3F,$B0,$E0,$F9,$0F,$C3,$CF,$E2,$BF,$B9,$3F,$22,$7E
	.dc.b	$71,$FC,$D7,$4F,$E7,$3A,$7F,$37,$3A,$DF,$F3,$4D,$DF,$F3,$57,$C2
	.dc.b	$3F,$4E,$64,$8E,$5E,$D5,$D2,$31,$5D,$1B,$5D,$27,$44,$BF,$21,$2F
	.dc.b	$BD,$70,$7B,$CE,$DC,$B5,$F0,$CF,$E2,$3A,$BB,$3C,$E3,$27,$07,$18
	.dc.b	$79,$6B,$93,$63,$38,$72,$AE,$04,$39,$92,$78,$39,$6B,$98,$E4,$FD
	.dc.b	$5C,$3D,$84,$BF,$0C,$EA,$B8,$91,$C1,$2B,$E1,$DF,$99,$A7,$86,$35
	.dc.b	$7D,$67,$4D,$9D,$1C,$0F,$31,$F8,$8E,$1F,$D8,$61,$E4,$8D,$61,$E3
	.dc.b	$6E,$1D,$A7,$57,$D6,$79,$B6,$B8,$36,$B8,$61,$E4,$BE,$24,$70,$7D
	.dc.b	$9C,$3F,$81,$7F,$00,$7F,$A5,$71,$FE,$DE,$DC,$7F,$CD,$8D,$C3,$FD
	.dc.b	$3C,$D9,$FE,$C2,$DC,$4D,$24,$FF,$A6,$5B,$3F,$E0,$A6,$BE,$08,$D1
	.dc.b	$1D,$35,$D1,$83,$97,$C1,$75,$F8,$7E,$DF,$07,$DF,$E1,$FC,$1F,$09
	.dc.b	$E1,$F3,$BC,$5F,$09,$E3,$F8,$7F,$27,$C2,$F9,$7E,$1F,$CD,$EA,$79
	.dc.b	$FE,$1B,$D1,$F0,$3E,$9F,$86,$F5,$7C,$17,$AF,$E1,$BC,$9C,$DF,$1B
	.dc.b	$93,$F4,$DC,$9F,$13,$E0,$FE,$01,$1F,$80,$D7,$E7,$B6,$7E,$7B,$67
	.dc.b	$E7,$B6,$7A,$1C,$88,$72,$2C,$F2,$6C,$F2,$7E,$87,$C3,$7D,$32,$38
	.dc.b	$21,$F8,$27,$2D,$33,$3F,$84,$F8,$4F,$27,$4F,$E0,$30,$BD,$26,$92
	.dc.b	$FF,$DA,$67,$E2,$3E,$1B,$F0,$84,$7F,$08,$D7,$F2,$6D,$C3,$F2,$6D
	.dc.b	$CB,$F2,$6D,$9E,$87,$22,$1C,$8B,$3C,$9B,$3C,$9F,$F2,$EF,$F0,$0F
	.dc.b	$FB,$E5,$BF,$0B,$FE,$F1,$8F,$F7,$AC,$6B,$E1,$31,$F0,$99,$F8,$4D
	.dc.b	$7C,$3E,$FE,$17,$87,$C2,$7E,$87,$87,$97,$C2,$73,$F8,$4E,$9E,$47
	.dc.b	$5F,$84,$ED,$F0,$9D,$FE,$13,$D9,$CF,$FA,$CE,$5F,$C6,$B7,$87,$D2
	.dc.b	$7C,$03,$DC,$4E,$5F,$73,$79,$7D,$CD,$38,$46,$89,$1A,$6C,$6A,$30
	.dc.b	$47,$04,$97,$27,$C9,$8F,$F3,$CA,$E5,$F9,$5F,$0D,$F2,$F9,$BE,$4D
	.dc.b	$7D,$0E,$5F,$92,$FD,$2E,$4F,$92,$FF,$07,$E3,$F9,$2F,$D6,$E2,$F9
	.dc.b	$2F,$D8,$E1,$F9,$2F,$DA,$E0,$F9,$3D,$0E,$1F,$92,$FE,$56,$7F,$4A
	.dc.b	$7E,$B0,$FD,$51,$38,$7F,$54,$DE,$1F,$D5,$34,$E1,$1A,$24,$69,$B1
	.dc.b	$A8,$C1,$1C,$13,$87,$F9,$4F,$4F,$EC,$2B,$F9,$8B,$6B,$F9,$8B,$6B
	.dc.b	$F9,$8B,$6B,$8C,$7E,$DA,$7B,$47,$ED,$63,$DE,$D7,$D4,$DA,$F3,$B9
	.dc.b	$F2,$BF,$A2,$F5,$7E,$E6,$B7,$8F,$C9,$6C,$66,$F8,$CE,$5B,$19,$C3
	.dc.b	$C7,$39,$6C,$7A,$61,$8D,$25,$FA,$CF,$79,$71,$F7,$BE,$23,$DB,$3D
	.dc.b	$5F,$98,$D8,$F7,$5F,$DA,$C7,$DB,$5C,$30,$F3,$3C,$1B,$5C,$23,$06
	.dc.b	$C8,$78,$4F,$97,$C9,$FA,$1C,$98,$27,$FD,$8F,$3F,$A3,$7F,$87,$F1
	.dc.b	$FE,$8D,$FE,$23,$D5,$FC,$2D,$FE,$27,$D5,$FC,$2D,$FE,$2B,$D5,$FC
	.dc.b	$2D,$F0,$E3,$F3,$9B,$8F,$CC,$7F,$97,$E5,$E9,$26,$BF,$0C,$9A,$7B
	.dc.b	$9B,$3C,$8F,$79,$ED,$BE,$CD,$BF,$65,$FF,$00,$F8,$7F,$E4,$DD,$1F
	.dc.b	$BE,$AF,$E4,$C7,$B0,$FD,$F1,$EC,$3F,$7C,$7B,$0F,$DF,$1E,$19,$F3
	.dc.b	$E3,$E5,$3F,$B4,$6F,$FE,$99,$DD,$F0,$DF,$CD,$49,$7F,$CE,$9A,$FF
	.dc.b	$12,$6E,$7F,$52,$3E,$9C,$63,$3F,$5D,$74,$9E,$44,$39,$16,$B9,$63
	.dc.b	$FC,$82,$F9,$4F,$43,$57,$D2,$3F,$37,$1F,$DA,$C9,$5F,$BE,$7F,$39
	.dc.b	$A7,$E8,$9B,$C3,$FB,$DC,$5F,$0D,$FD,$4F,$E1,$BF,$AA,$7C,$37,$F5
	.dc.b	$5F,$86,$FE,$AD,$F0,$DF,$8F,$C5,$9F,$9B,$E1,$31,$1E,$A2,$70,$7F
	.dc.b	$90,$B4,$F2,$D7,$8E,$7C,$BB,$FF,$10,$7C,$64,$32,$7E,$D0,$BC,$1F
	.dc.b	$B4,$9E,$44,$9E,$53,$AD,$FD,$CD,$BF,$0C,$EA,$B8,$4F,$56,$DF,$59
	.dc.b	$D1,$27,$4D,$8D,$46,$08,$E0,$FE,$13,$87,$FB,$8F,$14,$7C,$9B,$F9
	.dc.b	$0D,$BF,$CF,$CE,$AF,$F7,$95,$A6,$CE,$AB,$84,$74,$AE,$0F,$D4,$9D
	.dc.b	$3C,$0D,$4F,$01,$3C,$2B,$B4,$F1,$36,$71,$9D,$9C,$67,$85,$76,$9E
	.dc.b	$13,$A7,$84,$E9,$E6,$69,$C4,$6E,$3F,$BD,$F1,$FE,$F9,$FE,$32,$7E
	.dc.b	$F8,$FE,$FF,$CD,$FC,$65,$FF,$00,$F3,$7F,$19,$7F,$C0,$7C,$DF,$C6
	.dc.b	$5E,$DF,$0D,$E7,$F8,$6F,$47,$AB,$E9,$F7,$FD,$5E,$FF,$AF,$5F,$66
	.dc.b	$B3,$1F,$E3,$B1,$E6,$24,$66,$3D,$4D,$DF,$E4,$93,$7F,$88,$D9,$F8
	.dc.b	$2D,$9F,$80,$DA,$EB,$3F,$22,$7A,$57,$23,$F8,$4E,$FF,$16,$7F,$C4
	.dc.b	$3D,$9F,$CF,$5F,$8A,$7F,$22,$3E,$29,$FC,$88,$F8,$A7,$F2,$23,$E2
	.dc.b	$9F,$5B,$4F,$8A,$FD,$8C,$FE,$2B,$F6,$B1,$F1,$BF,$6B,$87,$F7,$8F
	.dc.b	$ED,$F3,$FD,$4F,$1F,$EC,$1B,$C7,$FB,$09,$FD,$2C,$E7,$4F,$D5,$36
	.dc.b	$33,$3B,$D7,$F5,$53,$96,$C6,$63,$04,$70,$49,$F2,$57,$68,$FF,$47
	.dc.b	$E1,$E8,$9F,$E0,$86,$8F,$09,$6F,$BB,$6F,$B3,$6B,$B3,$C8,$98,$7C
	.dc.b	$07,$44,$33,$87,$8D,$FE,$D3,$5D,$1C,$9E,$22,$DF,$89,$B1,$BA,$ED
	.dc.b	$18,$BF,$1C,$60,$97,$DE,$74,$78,$E3,$2F,$F2,$DC,$3E,$24,$63,$0F
	.dc.b	$03,$70,$F0,$5F,$38,$C6,$1E,$08,$C1,$30,$F5,$4E,$9F,$04,$66,$FA
	.dc.b	$B6,$FE,$16,$1F,$C9,$FD,$FF,$E1,$C4,$BF,$E1,$6D,$7F,$85,$37,$0F
	.dc.b	$EA,$4D,$BF,$E5,$6D,$9F,$E4,$CD,$C3,$F9,$0B,$72,$34,$C7,$C3,$7F
	.dc.b	$1F,$F8,$0F,$E4,$1F,$0D,$FC,$87,$E0,$7F,$91,$7C,$37,$F2,$3F,$82
	.dc.b	$FE,$49,$F0,$DF,$C9,$7E,$0F,$F9,$37,$C3,$7F,$27,$F8,$4F,$E5,$1F
	.dc.b	$0D,$FC,$A7,$E1,$7F,$95,$7C,$37,$F3,$3F,$E0,$0F,$F0,$47,$1F,$F2
	.dc.b	$A5,$F0,$D9,$F8,$6F,$24,$FF,$8C,$25,$F0,$DC,$49,$1F,$CD,$E3,$2D
	.dc.b	$8C,$BD,$5B,$D1,$FE,$0E,$9D,$CE,$4E,$04,$30,$48,$E1,$3C,$BA,$3F
	.dc.b	$C6,$F5,$CA,$73,$5C,$5A,$F1,$9D,$D7,$18,$FE,$CB,$5C,$67,$31,$C5
	.dc.b	$AF,$13,$73,$C7,$A1,$5C,$67,$27,$12,$9C,$63,$73,$F1,$3A,$5E,$C7
	.dc.b	$AB,$E1,$BE,$A1,$2F,$E6,$35,$FA,$CD,$CB,$E5,$37,$2E,$53,$C0,$87
	.dc.b	$02,$CF,$06,$C7,$07,$44,$8F,$AD,$F0,$DF,$87,$F8,$6F,$C4,$7C,$37
	.dc.b	$E2,$7E,$1B,$F1,$5F,$0D,$F8,$BF,$86,$FC,$67,$C3,$7E,$37,$E1,$BF
	.dc.b	$1D,$F0,$DF,$8F,$F8,$6F,$C8,$7C,$37,$E4,$7E,$1B,$F2,$5E,$DF,$E4
	.dc.b	$FC,$AF,$CA,$7C,$0F,$E5,$7D,$EF,$67,$1F,$D6,$E5,$FF,$21,$73,$FC
	.dc.b	$88,$C1,$EA,$29,$E9,$E6,$F6,$73,$7B,$E3,$06,$9C,$12,$36,$48,$D3
	.dc.b	$7A,$7E,$FD,$F9,$1B,$3E,$16,$BF,$AB,$E7,$F9,$0F,$E9,$79,$BE,$43
	.dc.b	$FA,$7E,$5F,$90,$FE,$A7,$93,$E4,$3F,$AB,$E3,$F9,$0F,$EB,$78,$BE
	.dc.b	$43,$FA,$FE,$1F,$90,$FE,$C7,$83,$E4,$3E,$3E,$A6,$BF,$23,$A7,$F3
	.dc.b	$7C,$1F,$A4,$27,$06,$09,$7F,$98,$8C,$B6,$33,$C1,$F9,$A8,$FC,$F3
	.dc.b	$5F,$CF,$61,$FA,$38,$C5,$72,$6D,$72,$9D,$13,$AB,$F5,$8D,$BE,$AD
	.dc.b	$7A,$C6,$88,$E9,$AB,$8B,$F0,$D7,$E0,$8F,$51,$1C,$12,$BB,$61,$E7
	.dc.b	$6E,$1E,$7C,$3A,$CE,$5B,$19,$BF,$11,$0F,$39,$67,$83,$63,$83,$A3
	.dc.b	$F8,$FE,$1E,$8B,$F0,$C7,$9D,$B8,$79,$F0,$F0,$CE,$5B,$19,$8E,$C6
	.dc.b	$4A,$E5,$F0,$61,$E4,$6E,$1E,$48,$C1,$1F,$3E,$1E,$5B,$EF,$87,$90
	.dc.b	$8E,$09,$7E,$08,$CB,$63,$2F,$80,$F3,$17,$0F,$36,$1E,$18,$CD,$FD
	.dc.b	$CF,$E0,$17,$A3,$FE,$8D,$6C,$FF,$A2,$5A,$FF,$A1,$9B,$5F,$ED,$E6
	.dc.b	$E5,$FF,$00,$1B,$3F,$EA,$86,$E0,$69,$2F,$C0,$49,$F0,$35,$F0,$36
	.dc.b	$FC,$0D,$BF,$03,$E9,$F7,$FD,$5E,$97,$AF,$E1,$7D,$9E,$CF,$B7,$E1
	.dc.b	$7D,$DE,$D7,$BF,$E1,$7E,$07,$B7,$F0,$7E,$1F,$E1,$7C,$37,$C3,$E5
	.dc.b	$F8,$9F,$0D,$F1,$7E,$1F,$E3,$7C,$3F,$C7,$F8,$7F,$91,$F0,$DE,$3E
	.dc.b	$5F,$81,$C7,$FC,$1D,$F8,$5F,$0D,$F3,$48,$FC,$D6,$BF,$35,$BC,$DF
	.dc.b	$35,$BC,$DF,$35,$B3,$D0,$E4,$43,$91,$67,$93,$67,$93,$F7,$7E,$1B
	.dc.b	$F0,$1F,$0D,$F8,$1F,$86,$FF,$67,$7F,$00,$AA,$F6,$63,$E1,$F3,$F0
	.dc.b	$BA,$F8,$7D,$FC,$37,$0F,$87,$E3,$F0,$D3,$1F,$F6,$96,$5F,$E3,$66
	.dc.b	$FC,$2F,$6F,$85,$EF,$F0,$BE,$CD,$7F,$0D,$C7,$F2,$56,$0F,$85,$F2
	.dc.b	$93,$1E,$AD,$7F,$DB,$71,$B7,$24,$33,$1B,$E4,$72,$7B,$EF,$BC,$E6
	.dc.b	$31,$7C,$EF,$89,$CE,$F8,$AB,$E2,$E6,$B8,$DF,$38,$FD,$27,$27,$CE
	.dc.b	$F8,$5F,$A1,$1F,$18,$F9,$E7,$C8,$E3,$FA,$7F,$0B,$F5,$3E,$17,$E7
	.dc.b	$CF,$CC,$E2,$F9,$CF,$C0,$E2,$FB,$3F,$0B,$F6,$BE,$17,$E7,$4F,$D2
	.dc.b	$F8,$5F,$16,$FE,$68,$FE,$7D,$E0,$F9,$9F,$41,$EB,$E2,$F8,$1C,$5E
	.dc.b	$EA,$E0,$D8,$CF,$17,$1A,$F8,$A4,$9D,$71,$78,$2B,$73,$D6,$30,$47
	.dc.b	$C3,$7D,$AF,$A5,$FF,$0F,$E2,$F9,$15,$FC,$57,$AB,$F2,$EF,$C8,$3F
	.dc.b	$03,$C5,$F2,$1F,$C8,$F0,$FC,$87,$F0,$BC,$3F,$23,$0F,$57,$0F,$C8
	.dc.b	$9F,$CB,$F0,$7C,$87,$F1,$7C,$1E,$57,$E3,$9F,$B1,$3F,$5E,$4A,$FD
	.dc.b	$7B,$74,$FD,$7B,$4E,$11,$A2,$46,$9B,$1A,$8C,$11,$C1,$35,$FD,$3B
	.dc.b	$FB,$D3,$EC,$7C,$2F,$EF,$1F,$D8,$FC,$2F,$EF,$3E,$17,$F7,$90,$C3
	.dc.b	$EE,$79,$1F,$E9,$A6,$FC,$2F,$EF,$A7,$F7,$3D,$DC,$27,$7F,$0B,$74
	.dc.b	$FD,$46,$1F,$E1,$73,$FC,$4F,$7F,$C2,$6B,$AF,$BB,$FC,$33,$E1,$BF
	.dc.b	$86,$F8,$1C,$6B,$F5,$16,$F8,$6F,$E0,$71,$EB,$E0,$F6,$BF,$41,$B8
	.dc.b	$FC,$66,$E5,$F1,$5B,$3F,$17,$D4,$78,$1F,$C3,$5B,$1D,$5A,$FD,$59
	.dc.b	$E2,$48,$CD,$F7,$6D,$F7,$6D,$7E,$F0,$FE,$D1,$E2,$FE,$F1,$FE,$26
	.dc.b	$7E,$56,$BF,$51,$D3,$F3,$63,$F8,$BF,$07,$E5,$67,$F5,$9D,$1F,$95
	.dc.b	$C7,$F8,$8F,$47,$C9,$BF,$E5,$27,$F2,$CF,$0F,$E3,$BE,$9C,$FD,$7E
	.dc.b	$AF,$F8,$E3,$BF,$F5,$FB,$7E,$AE,$BF,$60,$7E,$BA,$32,$4E,$FF,$CF
	.dc.b	$69,$FB,$89,$D4,$63,$3F,$D9,$92,$39,$9F,$BF,$71,$7F,$DF,$FB,$FF
	.dc.b	$6F,$3F,$E0,$9E,$FF,$DB,$BF,$D1,$27,$FA,$63,$FD,$1B,$C2,$FE,$3C
	.dc.b	$FF,$48,$F1,$7F,$A6,$3F,$D2,$B7,$F8,$8F,$F1,$5F,$0F,$F8,$F3,$F0
	.dc.b	$CF,$E3,$C7,$A0,$FE,$B5,$B7,$C3,$D3,$D5,$1E,$8D,$7D,$CF,$A4,$C6
	.dc.b	$BE,$AD,$BE,$24,$73,$35,$A7,$CB,$24,$6A,$32,$FD,$0D,$DB,$FC,$2C
	.dc.b	$3F,$C5,$1B,$FC,$2B,$FE,$AF,$87,$C2,$7F,$AC,$EF,$F0,$9F,$EB,$7B
	.dc.b	$FC,$27,$F9,$F1,$F0,$CF,$ED,$BB,$7C,$27,$FB,$0F,$07,$F7,$AF,$03
	.dc.b	$F8,$C7,$07,$F8,$FF,$9B,$F4,$44,$AF,$D1,$35,$FC,$53,$74,$FD,$03
	.dc.b	$78,$3F,$40,$FF,$81,$BC,$0F,$E1,$D5,$FD,$C3,$C0,$FE,$1C,$FF,$01
	.dc.b	$AF,$E0,$A7,$EF,$F6,$75,$3B,$FF,$45,$97,$F4,$BE,$8D,$4B,$BE,$FA
	.dc.b	$D8,$F6,$33,$E0,$FE,$8F,$4F,$EF,$FE,$07,$0F,$03,$F4,$79,$FA,$27
	.dc.b	$F0,$3E,$2F,$F1,$6F,$7F,$A9,$23,$CC,$7F,$12,$70,$FF,$25,$E0,$E0
	.dc.b	$EB,$83,$D3,$3A,$BF,$E5,$2D,$AD,$D7,$13,$A1,$C7,$64,$EE,$32,$6E
	.dc.b	$FA,$BB,$8D,$46,$23,$52,$F8,$6F,$8E,$7C,$8F,$80,$F9,$3F,$0D,$F2
	.dc.b	$BE,$07,$E5,$FC,$37,$CC,$F8,$2F,$9B,$F0,$DF,$3B,$E0,$E6,$7E,$5F
	.dc.b	$86,$FA,$3F,$09,$F4,$BE,$1B,$E9,$FC,$2F,$D4,$F8,$6F,$B5,$FC,$03
	.dc.b	$FE,$7C,$E1,$F5,$E3,$E7,$3E,$79,$FE,$23,$87,$C3,$70,$24,$7C,$B8
	.dc.b	$CB,$63,$2F,$C3,$8C,$B6,$33,$CD,$FE,$13,$9C,$B6,$33,$CD,$FE,$3B
	.dc.b	$9C,$B6,$33,$5E,$56,$F2,$AB,$73,$93,$EC,$72,$FC,$E7,$D1,$F0,$DF
	.dc.b	$0B,$E1,$BE,$74,$7A,$FE,$1B,$E2,$FC,$37,$C6,$F8,$6F,$9B,$3F,$63
	.dc.b	$92,$ED,$FD,$C3,$8F,$E7,$4F,$C3,$E0,$F9,$BF,$0D,$F3,$BE,$1B,$CB
	.dc.b	$BF,$C2,$F2,$BF,$DE,$6F,$9B,$87,$D1,$C5,$A2,$63,$F2,$C8,$7C,$92
	.dc.b	$E7,$F1,$DB,$C7,$F1,$5A,$7C,$13,$81,$23,$83,$63,$FC,$03,$C7,$F1
	.dc.b	$63,$EC,$71,$FC,$57,$F4,$1D,$7F,$88,$E3,$F8,$B1,$F8,$1E,$2F,$87
	.dc.b	$87,$F8,$2B,$8A,$CF,$F8,$13,$8B,$E2,$BF,$83,$E2,$F8,$AF,$E2,$78
	.dc.b	$7D,$EF,$F8,$16,$3F,$26,$FE,$9B,$BB,$F5,$31,$8D,$BF,$6F,$3F,$97
	.dc.b	$34,$57,$2E,$08,$E0,$98,$FE,$F9,$B1,$FB,$A6,$F0,$7E,$DD,$BC,$2F
	.dc.b	$86,$FD,$4F,$81,$FB,$7B,$FE,$09,$DF,$FA,$CF,$86,$FD,$BC,$7F,$07
	.dc.b	$EF,$FD,$BB,$FA,$7F,$86,$FD,$8F,$C3,$7F,$26,$F0,$3F,$67,$F0,$DF
	.dc.b	$B4,$F8,$6F,$DB,$DF,$F0,$FF,$03,$F6,$FF,$0D,$F0,$B9,$3F,$D5,$9C
	.dc.b	$9F,$0F,$93,$D4,$4B,$F5,$37,$93,$D4,$D9,$E1,$3A,$24,$E9,$B1,$A8
	.dc.b	$C1,$1C,$12,$BE,$AF,$67,$E2,$79,$7F,$CA,$7E,$37,$F1,$AC,$3F,$12
	.dc.b	$7F,$29,$F1,$BF,$8D,$3F,$CA,$FC,$6F,$C3,$E5,$FD,$33,$AF,$E2,$4F
	.dc.b	$F2,$FE,$BF,$C4,$E3,$FE,$5E,$F1,$7F,$8D,$4F,$F1,$1D,$FE,$23,$FC
	.dc.b	$63,$0F,$D5,$BF,$A2,$6B,$F9,$F3,$F4,$95,$FA,$22,$6F,$FA,$98,$D4
	.dc.b	$F0,$9C,$C7,$09,$D4,$62,$75,$18,$24,$EC,$91,$FD,$A3,$5F,$8D,$3F
	.dc.b	$C7,$7C,$0F,$D7,$57,$F4,$4F,$03,$F5,$CF,$F2,$DF,$13,$F8,$93,$FD
	.dc.b	$47,$5F,$8F,$C4,$7F,$9F,$C7,$F5,$5F,$1B,$F9,$EC,$FE,$BA,$3F,$99
	.dc.b	$78,$1F,$AE,$F1,$3F,$88,$F8,$9F,$B0,$DF,$D5,$E1,$7F,$13,$24,$67
	.dc.b	$6F,$5F,$07,$BB,$1F,$5D,$69,$F9,$CD,$7E,$4B,$FC,$77,$2F,$EB,$D1
	.dc.b	$FC,$C8,$FE,$62,$7F,$49,$F1,$BF,$98,$BF,$D1,$FC,$5F,$E1,$AF,$F4
	.dc.b	$9F,$17,$DF,$BF,$C0,$E6,$FC,$1E,$7F,$D1,$71,$F7,$6B,$FD,$2F,$7F
	.dc.b	$77,$89,$F9,$FE,$BF,$C2,$E7,$FC,$0B,$0C,$3F,$C1,$DA,$E4,$FD,$0C
	.dc.b	$70,$BF,$E0,$78,$7E,$7E,$78,$1F,$D9,$7C,$5F,$E0,$6F,$F8,$17,$B3
	.dc.b	$F4,$0F,$F6,$1F,$13,$F8,$1B,$FE,$00,$F1,$3F,$81,$BF,$D9,$7C,$4F
	.dc.b	$E0,$6F,$F7,$93,$FB,$67,$85,$FC,$0D,$FF,$08,$77,$7E,$81,$FE,$D9
	.dc.b	$DD,$EC,$E9,$FE,$97,$E8,$FF,$7C,$24,$FF,$75,$6B,$FD,$CD,$BB,$7F
	.dc.b	$71,$6F,$07,$F6,$F6,$CF,$F8,$05,$BD,$3F,$DB,$5B,$D3,$FE,$A8,$8C
	.dc.b	$4F,$72,$1D,$CB,$7D,$DB,$7D,$DA,$77,$2C,$F4,$79,$12,$79,$36,$79
	.dc.b	$34,$E4,$59,$E5,$87,$F7,$0F,$63,$FC,$42,$49,$FF,$0D,$35,$FF,$0C
	.dc.b	$36,$FF,$C2,$CD,$BF,$F0,$A3,$67,$FC,$4C,$DB,$FF,$09,$37,$0F,$F4
	.dc.b	$4C,$F6,$21,$D8,$FF,$15,$7B,$FF,$E3,$4F,$67,$FC,$DD,$FC,$03,$FE
	.dc.b	$FE,$9F,$F2,$62,$C3,$3F,$01,$AF,$87,$DF,$A9,$C3,$DB,$E3,$F0,$3C
	.dc.b	$BE,$07,$9F,$C1,$74,$F5,$3A,$FB,$BD,$BE,$1F,$BF,$C1,$F8,$3E,$1F
	.dc.b	$D1,$C8,$E1,$FE,$59,$C9,$EA,$7D,$04,$E6,$F8,$AF,$B1,$AE,$79,$BE
	.dc.b	$3B,$96,$B9,$3E,$49,$92,$99,$E3,$F9,$6E,$5A,$E6,$9C,$DF,$26,$3F
	.dc.b	$CF,$DC,$BF,$27,$06,$1E,$BC,$13,$F2,$EB,$7C,$BF,$2D,$E3,$C9,$F2
	.dc.b	$DE,$7C,$7F,$2D,$EB,$C5,$F2,$DE,$FC,$3F,$2D,$F0,$F0,$7C,$B7,$D1
	.dc.b	$B7,$C9,$E7,$FD,$BC,$7E,$7A,$3F,$54,$4E,$8F,$D6,$92,$34,$D7,$F3
	.dc.b	$91,$B7,$24,$33,$1B,$DF,$F4,$E4,$DF,$F4,$ED,$9F,$D3,$B7,$7F,$D8
	.dc.b	$36,$FB,$4F,$43,$F7,$B8,$7A,$9C,$12,$78,$61,$E1,$C3,$B9,$27,$BB
	.dc.b	$70,$EE,$DC,$3B,$DF,$2A,$F6,$35,$ED,$3B,$BE,$D1,$D0,$97,$D1,$B7
	.dc.b	$D2,$B8,$47,$E7,$2B,$8D,$6C,$95,$B9,$E3,$3A,$8C,$4E,$A3,$04,$9F
	.dc.b	$3D,$76,$7C,$E7,$23,$25,$33,$3C,$DF,$21,$A7,$0E,$AB,$C2,$4A,$F0
	.dc.b	$46,$5B,$7F,$1C,$97,$E3,$6C,$F8,$5B,$7D,$9F,$9D,$83,$C3,$FD,$47
	.dc.b	$03,$E0,$BF,$8A,$93,$0F,$D7,$B5,$FE,$2E,$DB,$FD,$8C,$7F,$19,$9E
	.dc.b	$4F,$02,$1C,$0B,$1C,$30,$FD,$BD,$73,$C7,$C6,$DA,$E8,$D9,$DC,$7F
	.dc.b	$02,$F8,$7F,$E5,$7F,$07,$FC,$B3,$E1,$FF,$96,$FC,$27,$F2,$EF,$84
	.dc.b	$FE,$5F,$F0,$9F,$CC,$3E,$1F,$F9,$06,$3F,$EF,$BF,$E0,$1F,$4B,$7E
	.dc.b	$1B,$3B,$F9,$65,$F0,$D8,$E0,$FE,$1D,$0F,$86,$C6,$1F,$CE,$61,$F0
	.dc.b	$D8,$E1,$43,$E1,$B0,$BE,$1A,$FA,$58,$F7,$BD,$DC,$FF,$2B,$27,$4F
	.dc.b	$D7,$7B,$7A,$3C,$88,$FF,$3B,$E0,$F4,$46,$23,$DC,$D8,$CC,$6F,$7F
	.dc.b	$5C,$E5,$B1,$93,$B9,$75,$F5,$CE,$E3,$27,$63,$5A,$23,$53,$C2,$73
	.dc.b	$F0,$BF,$08,$F8,$7E,$F7,$C4,$F8,$1F,$8B,$E9,$FC,$6F,$81,$F8,$FE
	.dc.b	$D7,$C8,$F8,$4F,$93,$ED,$FC,$AF,$80,$F9,$7E,$B7,$CC,$F8,$3F,$A1
	.dc.b	$CC,$F5,$EF,$C0,$E3,$E0,$F3,$F0,$5E,$4E,$5F,$37,$1F,$FA,$35,$AF
	.dc.b	$AC,$9C,$FF,$21,$BC,$FF,$19,$A7,$C2,$8D,$12,$34,$D8,$D4,$60,$8E
	.dc.b	$09,$CD,$F1,$DE,$AB,$37,$C1,$7B,$BE,$0F,$DF,$F0,$7F,$26,$FE,$B5
	.dc.b	$3E,$0B,$E1,$FC,$37,$C4,$F8,$4F,$8B,$F0,$9F,$2A,$FE,$BF,$2F,$CA
	.dc.b	$7E,$CF,$27,$CA,$F4,$3E,$5F,$C2,$7C,$CF,$84,$F9,$BF,$0D,$F2,$7C
	.dc.b	$2F,$D5,$F8,$7F,$95,$E3,$F4,$72,$FB,$DB,$CB,$EE,$6E,$9E,$EE,$6F
	.dc.b	$9F,$5B,$6D,$6D,$B3,$BC,$3C,$0F,$E3,$9A,$FC,$6E,$3F,$8A,$FC,$2E
	.dc.b	$5F,$8A,$FE,$53,$C1,$FC,$2F,$2F,$C4,$8F,$8F,$C9,$F1,$1F,$CC,$F6
	.dc.b	$FE,$35,$FC,$DF,$6F,$E3,$5F,$9B,$C7,$F1,$1F,$9F,$C5,$F1,$1F,$21
	.dc.b	$0F,$DC,$C7,$F8,$57,$37,$C2,$FE,$58,$93,$F6,$1A,$B8,$3F,$93,$4F
	.dc.b	$08,$D1,$0D,$70,$75,$AE,$35,$C1,$B1,$99,$E1,$39,$6C,$65,$E6,$7F
	.dc.b	$94,$EB,$C0,$E3,$87,$F2,$57,$FB,$98,$FD,$D9,$FC,$03,$C3,$FD,$8F
	.dc.b	$C3,$7E,$EA,$3F,$81,$F8,$7F,$BA,$7F,$82,$F8,$5F,$B6,$F5,$BF,$6F
	.dc.b	$EF,$FE,$E3,$DF,$FD,$CF,$C3,$7E,$EB,$C6,$FD,$CE,$1F,$9C,$F1,$7F
	.dc.b	$72,$FA,$0F,$36,$BF,$92,$E2,$FF,$21,$7F,$00,$7F,$91,$78,$AD,$7F
	.dc.b	$33,$8C,$3F,$E6,$F6,$F0,$9A,$48,$E0,$43,$F9,$13,$96,$99,$3E,$33
	.dc.b	$96,$99,$E3,$FE,$C1,$19,$7F,$A7,$79,$5F,$2E,$7F,$AA,$71,$FE,$ED
	.dc.b	$FE,$AD,$E5,$7F,$23,$7F,$96,$F9,$5F,$C8,$DF,$CB,$9F,$BA,$3F,$2E
	.dc.b	$7E,$E8,$FE,$07,$B7,$CA,$7E,$D7,$33,$7F,$89,$A7,$97,$5F,$8F,$5F
	.dc.b	$E8,$5F,$03,$F7,$9B,$79,$E7,$33,$EC,$3E,$09,$0C,$13,$6F,$7B,$D4
	.dc.b	$FA,$05,$BF,$A0,$DA,$EC,$E8,$E9,$3C,$9F,$E3,$BE,$47,$F2,$17,$F8
	.dc.b	$FF,$91,$F1,$E7,$F4,$FD,$DF,$BD,$9F,$7F,$77,$EF,$5F,$F0,$07,$77
	.dc.b	$EF,$5F,$F0,$17,$77,$EF,$5F,$8B,$CB,$FB,$D7,$E3,$F7,$7E,$41,$F6
	.dc.b	$6D,$F0,$B7,$F8,$B1,$E9,$21,$E8,$8F,$8F,$9F,$A6,$38,$9E,$D2,$EF
	.dc.b	$F0,$9E,$06,$63,$81,$27,$99,$23,$4F,$F0,$DD,$BE,$24,$7D,$ED,$BE
	.dc.b	$24,$FF,$11,$DB,$E2,$5F,$F0,$CD,$BE,$23,$FE,$1F,$DB,$E2,$3F,$E2
	.dc.b	$1D,$BE,$23,$FC,$4B,$6F,$88,$FF,$14,$DB,$E2,$3E,$5C,$FD,$9B,$7E
	.dc.b	$23,$E0,$7F,$60,$48,$FD,$4B,$5F,$D3,$B6,$7F,$4A,$DD,$3F,$76,$D9
	.dc.b	$FD,$C3,$77,$FD,$C3,$78,$0D,$35,$E1,$BA,$3F,$5D,$8F,$F8,$F7,$A3
	.dc.b	$F8,$BE,$3F,$E1,$8E,$8F,$D7,$3F,$E4,$5E,$8F,$D7,$3F,$A8,$ED,$FE
	.dc.b	$2E,$FF,$6E,$EC,$FE,$2E,$FE,$B8,$FE,$E3,$E5,$FF,$83,$DF,$E0,$FB
	.dc.b	$78,$DF,$D6,$71,$7F,$16,$F8,$7F,$E2,$84,$8F,$E1,$AD,$7F,$85,$B7
	.dc.b	$4F,$E1,$2D,$E1,$E5,$3C,$08,$70,$2C,$F0,$6C,$70,$74,$48,$FE,$26
	.dc.b	$FF,$3C,$3F,$89,$9F,$DA,$4F,$8C,$7F,$3C,$3F,$89,$9F,$CF,$0F,$E2
	.dc.b	$67,$F9,$B7,$C9,$FF,$06,$3F,$E6,$FF,$27,$F9,$E3,$F1,$8F,$E7,$84
	.dc.b	$3C,$55,$E5,$E4,$FE,$23,$87,$BE,$7F,$9C,$37,$2F,$E7,$0D,$E4,$34
	.dc.b	$93,$F0,$0E,$25,$9E,$2D,$9E,$2D,$9F,$F2,$5F,$5F,$EB,$33,$FE,$E7
	.dc.b	$D7,$FA,$CA,$FE,$CD,$D7,$FA,$CC,$FF,$BF,$75,$FE,$B1,$FF,$43,$F5
	.dc.b	$FE,$B1,$FF,$44,$F5,$FE,$B1,$FF,$02,$70,$FE,$B1,$FF,$03,$75,$FF
	.dc.b	$34,$C7,$C4,$7E,$20,$F4,$9F,$18,$B9,$FD,$D7,$1B,$FB,$1B,$BF,$B1
	.dc.b	$BB,$FC,$46,$9D,$8A,$72,$ED,$FF,$15,$BF,$D2,$3C,$6F,$F1,$5B,$FE
	.dc.b	$A9,$F1,$BF,$C5,$6F,$F9,$43,$CB,$FF,$27,$3F,$E5,$2F,$2F,$FC,$9C
	.dc.b	$FE,$0C,$FE,$AA,$7E,$0C,$FF,$15,$9F,$E5,$8F,$2F,$FC,$9C,$FE,$A0
	.dc.b	$F2,$73,$7F,$52,$F8,$2F,$EF,$44,$8F,$EE,$6D,$7F,$B8,$B7,$6F,$ED
	.dc.b	$ED,$E0,$FF,$02,$B6,$7F,$C0,$4D,$DF,$FC,$04,$DD,$CD,$35,$E0,$FF
	.dc.b	$0B,$AF,$D3,$61,$FC,$2D,$FF,$38,$F9,$5F,$E2,$17,$FD,$9D,$E5,$7F
	.dc.b	$88,$5F,$F6,$87,$95,$FE,$21,$7F,$CE,$FE,$57,$F8,$85,$FF,$3C,$F9
	.dc.b	$5F,$A6,$7F,$BB,$F8,$1F,$C2,$DF,$EF,$3E,$47,$8D,$FC,$26,$DF,$E1
	.dc.b	$9F,$87,$FF,$0F,$12,$3F,$C2,$ED,$7F,$C2,$AD,$D3,$FC,$26,$DE,$0F
	.dc.b	$F1,$4B,$67,$FC,$4C,$DA,$ED,$1D,$08,$74,$2D,$74,$7F,$C4,$FF,$07
	.dc.b	$FE,$31,$F8,$79,$1F,$E2,$AF,$84,$FF,$1A,$7C,$3F,$F8,$D7,$E1,$7F
	.dc.b	$C6,$DF,$0F,$FE,$37,$F8,$6F,$F1,$C7,$C3,$FF,$8E,$7E,$1B,$FC,$75
	.dc.b	$F0,$DF,$E3,$BF,$86,$FF,$1E,$7C,$37,$F8,$F7,$E1,$BF,$C7,$DF,$0D
	.dc.b	$FE,$3F,$F8,$6F,$47,$7B,$97,$FC,$8D,$C3,$EC,$3C,$FC,$DF,$01,$A6
	.dc.b	$4F,$61,$39,$7D,$8D,$8E,$0D,$38,$46,$89,$1A,$6C,$6A,$30,$47,$FC
	.dc.b	$A9,$F0,$C7,$FD,$E1,$FB,$A2,$71,$1F,$E7,$38,$7C,$3E,$2F,$52,$E9
	.dc.b	$53,$E1,$F5,$87,$59,$CF,$C3,$FD,$1B,$FF,$2E,$F3,$7D,$17,$FC,$BF
	.dc.b	$CB,$C1,$E9,$8B,$E1,$F1,$87,$6F,$87,$FA,$95,$FE,$66,$E4,$FA,$8F
	.dc.b	$AB,$E1,$F7,$87,$FB,$6E,$3E,$7D,$FE,$31,$FF,$53,$9F,$14,$8E,$76
	.dc.b	$F9,$2D,$D7,$E4,$36,$77,$C5,$F7,$DE,$AD,$7E,$B4,$6C,$91,$9A,$EA
	.dc.b	$DA,$EA,$DA,$FA,$A7,$D4,$8F,$5F,$07,$B3,$83,$DB,$C1,$F2,$EB,$FD
	.dc.b	$07,$C1,$F2,$DF,$F4,$2F,$07,$C2,$E0,$F8,$7C,$1F,$22,$7F,$D1,$3C
	.dc.b	$1F,$21,$FB,$DC,$1F,$23,$CE,$FC,$DF,$1F,$E6,$B8,$7E,$EF,$C3,$FC
	.dc.b	$C2,$5F,$E2,$1A,$FE,$1D,$BC,$5F,$86,$9E,$33,$B2,$1B,$2C,$6F,$8D
	.dc.b	$5C,$B8,$BF,$15,$18,$AD,$C6,$27,$51,$F8,$87,$FD,$65,$E0,$7F,$02
	.dc.b	$8F,$F5,$3F,$81,$FE,$B5,$F0,$3F,$81,$C7,$FA,$BB,$C0,$FC,$3F,$C2
	.dc.b	$7E,$23,$E0,$7F,$13,$F0,$3F,$C0,$EB,$FD,$67,$E0,$7E,$33,$D1,$FC
	.dc.b	$6F,$C0,$FE,$3B,$CB,$FE,$07,$5F,$EB,$8F,$03,$C7,$C7,$F0,$23,$F1
	.dc.b	$EF,$E9,$BE,$07,$F2,$A4,$8F,$D8,$B5,$FD,$83,$79,$3E,$23,$78,$FE
	.dc.b	$1B,$67,$F4,$ED,$AE,$D1,$D0,$87,$42,$D7,$47,$F9,$49,$FE,$D0,$E1
	.dc.b	$F9,$11,$FE,$CA,$E1,$F8,$F3,$FE,$D0,$F1,$FF,$94,$BF,$ED,$6F,$1F
	.dc.b	$F9,$4B,$FE,$D8,$F1,$FF,$94,$BF,$ED,$AF,$1F,$F9,$4B,$FE,$DC,$F1
	.dc.b	$FE,$3B,$FE,$DA,$DB,$D1,$D2,$DB,$F2,$0D,$9F,$3E,$FF,$01,$A7,$C0
	.dc.b	$2E,$FF,$01,$F7,$13,$6F,$73,$67,$93,$4E,$45,$9E,$53,$C0,$93,$C1
	.dc.b	$FF,$76,$F8,$DF,$C3,$A3,$F6,$B5,$FC,$AB,$C4,$FD,$AB,$FE,$FA,$DF
	.dc.b	$E3,$BF,$EF,$8D,$FE,$3B,$FC,$B7,$C1,$FD,$AB,$F1,$CF,$E3,$87,$C7
	.dc.b	$3F,$8E,$1F,$CD,$BB,$BE,$3E,$BF,$18,$B1,$F0,$B5,$F9,$0F,$AD,$BA
	.dc.b	$7A,$A3,$DE,$DD,$7D,$FA,$7B,$49,$3A,$6C,$6A,$30,$47,$03,$C6,$6D
	.dc.b	$F1,$67,$FA,$6F,$8D,$FC,$65,$FE,$9F,$E3,$7F,$19,$C7,$F8,$35,$BC
	.dc.b	$6F,$E3,$51,$8F,$1B,$F8,$D5,$FF,$56,$F1,$BF,$8D,$4F,$0F,$1B,$F8
	.dc.b	$D3,$FC,$27,$C6,$FE,$35,$5F,$0F,$83,$F1,$7E,$67,$F4,$42,$47,$F6
	.dc.b	$06,$BF,$D7,$DB,$5F,$C4,$DB,$C1,$FC,$49,$B3,$FD,$59,$BC,$1F,$D8
	.dc.b	$9B,$C2,$69,$AD,$9E,$2F,$EB,$B0,$FE,$CD,$0E,$DF,$E2,$D8,$FF,$15
	.dc.b	$21,$FD,$CF,$C4,$FE,$2C,$FF,$78,$F0,$FF,$8B,$2D,$3E,$5F,$87,$FC
	.dc.b	$5E,$3F,$BE,$F8,$5F,$C5,$DF,$F0,$04,$FD,$0F,$23,$FC,$35,$1F,$0C
	.dc.b	$87,$F1,$8D,$BF,$13,$A7,$E0,$9B,$1F,$81,$6E,$FF,$82,$69,$38,$3F
	.dc.b	$2B,$3F,$93,$6E,$1F,$91,$9D,$4F,$EE,$B6,$F8,$AF,$F1,$32,$1E,$DC
	.dc.b	$3E,$3F,$85,$FC,$52,$FF,$13,$DB,$FA,$E7,$F1,$7D,$BF,$AE,$7F,$1B
	.dc.b	$E1,$7F,$14,$C3,$91,$FE,$19,$C5,$E1,$7F,$16,$8C,$EB,$F9,$BF,$0B
	.dc.b	$F8,$BE,$1F,$96,$F0,$BD,$7E,$83,$D7,$F8,$9C,$3E,$CE,$0B,$97,$E4
	.dc.b	$48,$7C,$12,$E1,$F0,$67,$84,$7E,$AF,$4F,$D4,$BF,$E0,$13,$F5,$27
	.dc.b	$F0,$D3,$F5,$27,$F0,$D3,$FC,$56,$7F,$33,$3F,$C5,$67,$F8,$F3,$C3
	.dc.b	$FE,$1A,$FB,$23,$DE,$FF,$0E,$31,$9F,$F8,$A3,$AF,$F5,$31,$E7,$3F
	.dc.b	$05,$7E,$6D,$7E,$F3,$78,$FE,$CC,$EF,$2F,$63,$EE,$CF,$E1,$B9,$7F
	.dc.b	$3F,$E1,$7F,$05,$8F,$F2,$77,$85,$FC,$15,$FF,$28,$78,$5F,$C1,$6B
	.dc.b	$F4,$BE,$17,$F0,$47,$F4,$FE,$17,$F0,$4A,$FF,$2B,$78,$5F,$C1,$27
	.dc.b	$F5,$7E,$17,$F0,$47,$FC,$13,$BF,$B3,$4F,$2D,$FE,$77,$C2,$FD,$AB
	.dc.b	$FA,$08,$FE,$00,$DA,$FE,$00,$D3,$F8,$01,$78,$3F,$80,$37,$7F,$E0
	.dc.b	$0D,$BE,$AF,$F0,$D3,$87,$17,$E7,$9F,$F0,$A9,$F9,$EC,$3F,$B8,$3F
	.dc.b	$C0,$1F,$E4,$BE,$47,$F8,$5A,$3F,$CE,$FE,$47,$F8,$59,$FE,$E0,$7F
	.dc.b	$85,$8F,$F3,$DF,$91,$FE,$16,$7F,$CF,$BE,$47,$E6,$B4,$FE,$1F,$D5
	.dc.b	$E4,$EA,$FC,$D5,$7F,$9C,$3D,$9F,$F3,$49,$1F,$F4,$2B,$5F,$F4,$23
	.dc.b	$73,$FF,$1C,$36,$FF,$C6,$ED,$9E,$87,$22,$1C,$8B,$3C,$9B,$3C,$9F
	.dc.b	$F4,$AF,$BF,$FE,$7C,$F6,$BF,$CF,$BE,$FC,$9F,$E3,$75,$FE,$8F,$FE
	.dc.b	$01,$7C,$27,$FA,$21,$7A,$54,$C7,$A5,$9F,$87,$D7,$C1,$6F,$DD,$99
	.dc.b	$74,$3F,$D3,$D8,$3E,$0B,$A7,$C1,$75,$F7,$FB,$7C,$17,$BF,$B7,$FD
	.dc.b	$47,$C9,$FE,$69,$C1,$F0,$1E,$32,$65,$D6,$93,$A8,$C3,$5F,$E2,$6D
	.dc.b	$E5,$FF,$36,$37,$8B,$E0,$5F,$4A,$DB,$6B,$6D,$9D,$CF,$68,$C3,$EE
	.dc.b	$6D,$F9,$1F,$95,$E5,$7C,$BF,$80,$F9,$9F,$05,$F3,$7E,$1F,$E7,$73
	.dc.b	$7C,$DC,$7D,$3C,$DF,$39,$F5,$F2,$FD,$3F,$87,$F9,$B1,$EC,$E5,$FA
	.dc.b	$DC,$BF,$5F,$E0,$7E,$64,$FC,$2E,$4F,$1F,$1F,$C3,$EC,$FC,$BF,$17
	.dc.b	$96,$BD,$44,$3E,$09,$79,$3E,$0B,$79,$3E,$0B,$A2,$46,$9B,$1A,$8C
	.dc.b	$11,$C1,$23,$E4,$1D,$4F,$C3,$F1,$FC,$77,$ED,$F1,$7C,$77,$EE,$F0
	.dc.b	$FC,$77,$EE,$F0,$FC,$77,$EF,$F0,$FC,$77,$F0,$7C,$1F,$23,$2F,$C0
	.dc.b	$F0,$7C,$79,$FC,$3F,$37,$E4,$9F,$16,$BF,$15,$FE,$F5,$CD,$F9,$5F
	.dc.b	$10,$D2,$47,$EC,$0F,$DB,$9F,$B5,$26,$1F,$B5,$6E,$FF,$B5,$69,$C2
	.dc.b	$34,$48,$D3,$FE,$BD,$ED,$FD,$D4,$FE,$86,$DD,$9F,$BB,$8F,$CF,$FC
	.dc.b	$37,$EE,$7E,$0B,$F7,$5F,$0D,$FB,$BF,$83,$57,$EA,$7D,$8A,$FE,$A3
	.dc.b	$C0,$BA,$7F,$B4,$65,$BE,$36,$E2,$A1,$F0,$F9,$C3,$70,$F8,$7F,$DF
	.dc.b	$C7,$7F,$87,$D5,$F1,$F8,$6F,$93,$3F,$38,$F9,$51,$FC,$09,$BC,$3F
	.dc.b	$41,$BC,$7C,$34,$F8,$D1,$F2,$74,$F8,$E4,$70,$4B,$E8,$FD,$06,$CF
	.dc.b	$25,$6E,$3F,$9F,$5F,$E1,$C8,$F9,$EF,$D1,$E3,$FD,$CB,$FB,$82,$1F
	.dc.b	$C4,$3C,$5F,$E4,$CF,$CD,$3F,$26,$43,$F8,$AF,$37,$C6,$7F,$8B,$F3
	.dc.b	$7E,$4D,$F3,$5F,$C1,$7F,$DC,$7D,$DF,$C6,$5F,$D4,$9F,$B0,$2E,$7F
	.dc.b	$AF,$6E,$9F,$AF,$D3,$F3,$93,$96,$C6,$6B,$A4,$62,$B8,$91,$C1,$23
	.dc.b	$64,$ED,$FD,$C4,$FF,$30,$ED,$FD,$C3,$FC,$CB,$B7,$F7,$18,$FF,$32
	.dc.b	$ED,$FD,$C3,$FC,$D3,$B7,$F7,$15,$FC,$74,$F8,$D1,$FB,$87,$F9,$EF
	.dc.b	$81,$FC,$79,$FE,$7B,$E0,$7F,$1E,$F1,$3F,$A4,$16,$3F,$77,$A7,$9A
	.dc.b	$31,$1F,$1F,$5F,$54,$63,$0F,$89,$5B,$AE,$33,$FC,$AB,$5F,$6D,$7F
	.dc.b	$4A,$D7,$DA,$FF,$4C,$D7,$DB,$B7,$F6,$BD,$BD,$AF,$F6,$DD,$BD,$B5
	.dc.b	$FD,$4B,$C0,$FE,$29,$3F,$DC,$78,$BF,$8A,$69,$F8,$23,$F9,$D1,$E2
	.dc.b	$3F,$73,$E0,$FF,$3A,$E2,$FC,$26,$1F,$99,$6D,$FE,$87,$83,$F3,$AD
	.dc.b	$8C,$DF,$E8,$DB,$3A,$8C,$73,$FE,$B6,$B8,$36,$38,$73,$FE,$93,$05
	.dc.b	$7F,$04,$C3,$F4,$B5,$FA,$7E,$7F,$F0,$57,$83,$FC,$4E,$3F,$BA,$F8
	.dc.b	$1F,$C4,$DF,$E2,$47,$F8,$28,$FE,$29,$9F,$F0,$6E,$DF,$E1,$B1,$FC
	.dc.b	$5F,$B7,$F4,$D9,$FF,$15,$ED,$D7,$6F,$F0,$AE,$0F,$EB,$B8,$7F,$91
	.dc.b	$6B,$F8,$D7,$C1,$F3,$26,$3F,$C8,$5A,$FF,$A6,$27,$86,$FF,$C7,$9B
	.dc.b	$1F,$C7,$27,$75,$D2,$39,$10,$E5,$1A,$AE,$4D,$8E,$46,$E7,$F9,$77
	.dc.b	$C1,$FB,$CF,$81,$7F,$07,$CD,$F8,$5E,$F7,$C3,$F5,$7E,$27,$BD,$F1
	.dc.b	$7D,$6F,$8D,$EF,$7C,$7F,$81,$F9,$1E,$FF,$C9,$F6,$3E,$57,$BD,$F2
	.dc.b	$FE,$03,$E6,$7A,$9F,$37,$DE,$F2,$74,$FF,$3B,$E4,$FE,$87,$CD,$FE
	.dc.b	$06,$FE,$01,$FB,$BC,$DF,$E2,$F6,$BF,$EB,$E6,$F3,$6A,$D1,$8E,$7F
	.dc.b	$E8,$ED,$AF,$E9,$6D,$9D,$CF,$F4,$48,$C1,$1C,$12,$7C,$4D,$7C,$4D
	.dc.b	$AF,$14,$F1,$9D,$9F,$D7,$F9,$FE,$73,$EB,$F7,$FE,$73,$ED,$F8,$6F
	.dc.b	$9C,$FF,$65,$E5,$F9,$CF,$C1,$F8,$0F,$37,$C1,$79,$FE,$07,$E7,$4F
	.dc.b	$F6,$BE,$4F,$57,$C0,$FC,$E8,$F9,$1F,$0D,$E5,$DF,$E1,$9F,$3B,$CE
	.dc.b	$B3,$E6,$3D,$65,$E0,$F5,$CF,$18,$F9,$2D,$E1,$F9,$31,$B8,$E6,$F2
	.dc.b	$24,$F2,$74,$72,$2C,$F2,$77,$3F,$A1,$E0,$F8,$F1,$F7,$F8,$3E,$3B
	.dc.b	$F8,$1E,$0F,$8E,$FE,$A7,$83,$E3,$BF,$85,$E0,$F8,$EF,$E2,$77,$F8
	.dc.b	$EF,$EB,$F7,$F8,$EF,$F8,$4F,$97,$F0,$5B,$FB,$78,$BF,$1F,$C7,$FB
	.dc.b	$CE,$4F,$CB,$F7,$FE,$D3,$87,$F3,$18,$FE,$AF,$87,$F5,$ED,$3F,$5E
	.dc.b	$7E,$D0,$C4,$70,$70,$4E,$2F,$DE,$36,$79,$B4,$E7,$1A,$3F,$1F,$F0
	.dc.b	$FF,$99,$F8,$7F,$DC,$CF,$F8,$B7,$0F,$DC,$BF,$A9,$F8,$2F,$CF,$7C
	.dc.b	$3F,$E7,$FE,$0F,$F7,$33,$FE,$33,$EF,$FD,$CB,$FA,$FF,$87,$FD,$CB
	.dc.b	$FE,$37,$EE,$FD,$37,$C2,$7E,$E6,$3D,$79,$7F,$B8,$F8,$7E,$34,$63
	.dc.b	$87,$E3,$91,$C1,$0F,$67,$0F,$C3,$E0,$F6,$70,$FC,$06,$CF,$C0,$8C
	.dc.b	$11,$FF,$1F,$F5,$FE,$1A,$7F,$C9,$18,$FE,$19,$FF,$25,$78,$BF,$86
	.dc.b	$7F,$C9,$9D,$5F,$86,$7F,$C9,$DE,$1F,$F1,$9C,$3F,$CA,$9E,$1F,$E1
	.dc.b	$A7,$FC,$A5,$E1,$FF,$19,$9F,$C3,$1F,$E5,$9F,$0F,$D8,$B7,$FE,$31
	.dc.b	$AF,$E2,$49,$1F,$9E,$E9,$FE,$08,$FE,$78,$FD,$11,$77,$FD,$16,$3F
	.dc.b	$A7,$8C,$4F,$08,$C1,$2B,$F5,$EF,$F3,$A9,$FE,$1C,$7F,$0D,$3F,$CD
	.dc.b	$DE,$37,$F1,$DF,$03,$E1,$57,$F4,$DD,$BE,$13,$FC,$9B,$C0,$FD,$3D
	.dc.b	$FF,$2D,$F0,$3E,$14,$FF,$53,$F0,$3F,$4F,$3F,$D5,$FC,$0F,$D3,$BE
	.dc.b	$D9,$FE,$18,$FB,$33,$F3,$36,$3C,$E4,$70,$4D,$7E,$0C,$65,$B1,$9D
	.dc.b	$7E,$14,$E5,$B1,$98,$FE,$29,$C8,$E5,$FC,$24,$7F,$42,$F0,$7F,$08
	.dc.b	$FF,$44,$DF,$E1,$DF,$FA,$23,$7F,$86,$FF,$46,$DF,$E1,$E9,$FC,$F3
	.dc.b	$7F,$87,$3F,$E8,$ED,$FE,$1C,$FF,$A5,$F7,$F8,$73,$F6,$32,$F8,$7B
	.dc.b	$FE,$75,$FC,$C3,$63,$F3,$84,$70,$4D,$BF,$87,$13,$9F,$F1,$F8,$7E
	.dc.b	$92,$75,$18,$BE,$B5,$FC,$57,$81,$E0,$7F,$05,$C7,$F8,$21,$0F,$F0
	.dc.b	$87,$47,$F6,$1E,$FF,$E0,$B1,$FD,$BF,$BB,$F8,$2B,$FE,$15,$EE,$FE
	.dc.b	$0A,$FF,$71,$EE,$FE,$0A,$FF,$86,$7B,$BF,$82,$BF,$A5,$3F,$7D,$D3
	.dc.b	$FD,$EF,$E1,$FF,$B7,$12,$3F,$BE,$35,$FE,$F8,$DE,$1F,$ED,$6D,$DF
	.dc.b	$FB,$DB,$67,$FB,$DB,$78,$7F,$C0,$2D,$DC,$D3,$1F,$0F,$FD,$FF,$D1
	.dc.b	$FF,$00,$7A,$3F,$E0,$1F,$86,$FF,$01,$7A,$3F,$EC,$6F,$DA,$5E,$C5
	.dc.b	$F6,$31,$EF,$E7,$D9,$D7,$C0,$EF,$DA,$E1,$ED,$71,$F8,$79,$C7,$FE
	.dc.b	$E0,$E5,$FF,$0B,$68,$F8,$1E,$C7,$96,$7F,$CD,$BE,$97,$B9,$F8,$A4
	.dc.b	$3E,$29,$6B,$E2,$B7,$9F,$83,$A2,$46,$9B,$1A,$8C,$11,$C1,$23,$E6
	.dc.b	$9D,$4F,$F7,$9D,$BE,$03,$E3,$7B,$FF,$2A,$7F,$C6,$DC,$9F,$27,$1F
	.dc.b	$95,$C5,$F2,$FE,$1F,$E4,$D7,$C3,$CB,$FD,$F7,$93,$87,$E8,$7C,$07
	.dc.b	$CB,$AD,$1E,$0F,$87,$F9,$6B,$6F,$97,$C1,$F5,$FE,$1F,$EC,$7C,$27
	.dc.b	$CB,$AF,$E6,$9B,$7A,$FB,$BF,$0B,$1E,$B6,$BF,$10,$F7,$95,$CE,$1F
	.dc.b	$16,$32,$D8,$CE,$BD,$7B,$BF,$1F,$5D,$5A,$F3,$6D,$71,$DB,$E5,$4F
	.dc.b	$BA,$D1,$FE,$E2,$A7,$0F,$CD,$AF,$81,$87,$C8,$C7,$E6,$C7,$1B,$70
	.dc.b	$7C,$E8,$D9,$F6,$38,$3E,$6B,$BE,$6F,$CF,$70,$7C,$CD,$7E,$5C,$F3
	.dc.b	$DF,$E5,$BE,$BE,$EF,$E3,$79,$3E,$0B,$F4,$C4,$9F,$13,$5F,$0F,$07
	.dc.b	$B2,$31,$87,$E8,$9B,$B7,$E0,$9B,$5F,$8E,$6D,$71,$8E,$A4,$3A,$DF
	.dc.b	$8D,$C5,$F5,$7F,$D1,$1D,$FF,$BF,$7E,$37,$C3,$FE,$FD,$FF,$46,$77
	.dc.b	$7E,$FF,$E0,$FF,$7D,$1F,$2F,$E1,$CF,$D5,$FE,$01,$AF,$CF,$EF,$B9
	.dc.b	$7F,$A4,$65,$A7,$F0,$2A,$FA,$9D,$FA,$DF,$9C,$AB,$FD,$9B,$5E,$AE
	.dc.b	$FF,$E0,$6F,$AA,$BE,$23,$FC,$10,$FC,$8B,$F8,$FE,$1F,$CF,$63,$F9
	.dc.b	$3E,$9F,$C6,$CE,$6B,$EC,$10,$FC,$95,$70,$8D,$47,$FA,$AF,$83,$E3
	.dc.b	$3F,$EA,$DE,$9F,$8C,$FF,$AB,$F8,$3F,$1E,$FD,$0F,$07,$F1,$D9,$7F
	.dc.b	$1C,$E8,$FE,$47,$3F,$8D,$3E,$09,$F9,$5E,$CF,$C7,$65,$FC,$03,$5F
	.dc.b	$7C,$F9,$0F,$76,$FF,$AC,$7D,$E4,$70,$4D,$BD,$F1,$96,$C6,$76,$F7
	.dc.b	$4E,$5B,$19,$8F,$E0,$5C,$BF,$B2,$7F,$17,$DF,$FA,$FD,$3F,$D7,$7B
	.dc.b	$FF,$1B,$9F,$D7,$77,$FF,$1B,$7F,$61,$E0,$FF,$43,$7F,$65,$E0,$7F
	.dc.b	$19,$7F,$2B,$D9,$F0,$1F,$E5,$3C,$1F,$C5,$DF,$DC,$78,$1F,$C5,$FC
	.dc.b	$4F,$EC,$D1,$F5,$F3,$FB,$9B,$FA,$38,$3D,$A4,$70,$49,$E1,$C5,$FC
	.dc.b	$7F,$83,$D9,$7F,$D4,$38,$3D,$93,$FE,$E4,$E0,$F6,$5F,$EA,$B8,$3D
	.dc.b	$8F,$EB,$38,$3D,$8F,$EB,$B8,$3D,$98,$FF,$5A,$E0,$F6,$70,$FF,$18
	.dc.b	$E1,$F6,$5F,$ED,$38,$7D,$9E,$0F,$E6,$A3,$F6,$FD,$DF,$BC,$E0,$FD
	.dc.b	$C4,$60,$FD,$01,$FC,$04,$B8,$7F,$01,$74,$4C,$BF,$7C,$DD,$FB,$46
	.dc.b	$08,$E0,$91,$D4,$FD,$01,$FC,$C7,$89,$E2,$7F,$80,$7B,$FF,$43,$7F
	.dc.b	$EF,$CE,$9F,$E0,$D7,$FC,$BF,$A7,$FC,$0D,$DD,$FB,$D8,$FE,$03,$D9
	.dc.b	$FC,$19,$FE,$2B,$DD,$9D,$7F,$80,$1F,$C1,$63,$F8,$CF,$76,$6B,$F8
	.dc.b	$1E,$5F,$E0,$6E,$CF,$D2,$BE,$AA,$F4,$76,$38,$FF,$85,$DB,$E1,$79
	.dc.b	$12,$BF,$8C,$35,$FE,$67,$1B,$72,$43,$31,$BB,$FE,$38,$DE,$2E,$ED
	.dc.b	$9F,$7B,$73,$FE,$38,$DB,$ED,$3E,$0F,$85,$F7,$7C,$2F,$BF,$D0,$F8
	.dc.b	$1E,$C7,$C1,$F4,$7E,$17,$B3,$30,$FE,$61,$E8,$FC,$5F,$85,$F8,$DE
	.dc.b	$C7,$C7,$F8,$5F,$91,$EF,$FC,$9F,$4E,$61,$FC,$DD,$F9,$9E,$FF,$8F
	.dc.b	$87,$FC,$E7,$C3,$E4,$E4,$F6,$DF,$C1,$6B,$F0,$8C,$72,$7C,$06,$9F
	.dc.b	$00,$BC,$9F,$01,$B5,$D6,$78,$47,$32,$57,$62,$46,$9F,$E9,$5C,$9F
	.dc.b	$83,$F6,$7F,$09,$EF,$FE,$17,$D5,$F8,$D8,$7F,$53,$E3,$F8,$CF,$DC
	.dc.b	$7F,$15,$ED,$7C,$68,$FE,$B1,$EF,$FE,$47,$B1,$A7,$C6,$9F,$C1,$F0
	.dc.b	$DD,$7F,$AD,$70,$FC,$79,$FE,$67,$BF,$AD,$F8,$07,$AA,$FD,$04,$73
	.dc.b	$5E,$C6,$C6,$78,$3E,$34,$71,$32,$47,$33,$C5,$B3,$C6,$BC,$51,$8A
	.dc.b	$FD,$7F,$07,$C7,$7F,$BB,$6F,$F1,$DF,$EF,$1C,$FF,$1B,$1F,$EE,$DB
	.dc.b	$DC,$BF,$53,$BF,$C7,$BF,$D2,$EF,$F1,$E7,$E2,$1F,$AE,$DA,$CF,$EF
	.dc.b	$B6,$F8,$CF,$F9,$38,$FF,$00,$EE,$F7,$FE,$81,$23,$FC,$7D,$19,$6C
	.dc.b	$65,$FF,$13,$C6,$5B,$19,$CF,$EC,$C7,$68,$C6,$1F,$E0,$CA,$E1,$5A
	.dc.b	$DF,$FD,$1D,$5A,$BE,$07,$80,$87,$62,$CF,$1A,$7B,$FF,$B8,$9F,$F1
	.dc.b	$17,$5F,$EE,$1F,$5F,$BF,$F8,$DF,$85,$FC,$77,$C0,$7F,$1B,$ED,$FD
	.dc.b	$C5,$7C,$0F,$85,$FD,$C3,$FE,$2B,$EB,$FC,$A7,$C2,$FE,$57,$E0,$BF
	.dc.b	$2D,$F0,$BF,$97,$F8,$3F,$47,$17,$AB,$0F,$C9,$F7,$FD,$0E,$2F,$C3
	.dc.b	$76,$FE,$10,$91,$9E,$5F,$C4,$36,$BF,$1D,$3A,$7F,$6F,$1F,$C5,$FB
	.dc.b	$7D,$D1,$FC,$9B,$8F,$DC,$FF,$28,$EC,$FC,$43,$FC,$FB,$B3,$DC,$FF
	.dc.b	$91,$7C,$2F,$C5,$67,$FE,$43,$F0,$7F,$8A,$CF,$F1,$32,$1F,$B0,$3F
	.dc.b	$C9,$9D,$7F,$8A,$F1,$BF,$9E,$47,$9B,$4F,$46,$FF,$9F,$26,$DE,$99
	.dc.b	$FD,$54,$62,$77,$39,$C7,$E2,$C6,$36,$FD,$35,$FF,$42,$F0,$3F,$4C
	.dc.b	$FB,$4F,$E1,$84,$3F,$86,$1F,$E6,$1F,$03,$DB,$C3,$FD,$43,$C0,$FD
	.dc.b	$34,$FB,$31,$FF,$31,$78,$1F,$A6,$D3,$FC,$BD,$B7,$B7,$C6,$FA,$77
	.dc.b	$F4,$08,$7A,$B6,$F4,$65,$A8,$C6,$DE,$C2,$3F,$E7,$3D,$BD,$79,$FF
	.dc.b	$9C,$76,$F5,$BF,$E7,$3E,$3F,$AF,$A7,$F9,$BF,$8E,$E6,$E0,$F5,$E5
	.dc.b	$FD,$E3,$83,$D6,$FD,$52,$63,$EC,$7E,$B9,$E8,$E1,$FA,$F8,$FE,$56
	.dc.b	$3F,$54,$49,$FD,$57,$83,$F9,$AF,$07,$F6,$BC,$1F,$B7,$8C,$12,$BF
	.dc.b	$63,$19,$69,$FB,$83,$A3,$96,$99,$AF,$DC,$78,$3F,$9A,$7F,$D2,$9C
	.dc.b	$FF,$9A,$7F,$79,$E0,$FE,$DE,$BF,$89,$F8,$1F,$9A,$7F,$C0,$BD,$3F
	.dc.b	$D1,$FB,$FF,$6F,$1F,$E0,$8E,$FF,$DB,$BF,$C6,$BB,$BF,$6E,$FF,$1C
	.dc.b	$ED,$F5,$3F,$C8,$F8,$1C,$0F,$85,$FE,$A4,$49,$FE,$46,$D7,$F9,$0B
	.dc.b	$75,$FE,$24,$DE,$0F,$E3,$CD,$9F,$E5,$ED,$BE,$ED,$BF,$04,$F6,$21
	.dc.b	$D8,$FE,$67,$F0,$9F,$D9,$7E,$13,$FB,$37,$C2,$FF,$67,$F7,$7F,$B4
	.dc.b	$7B,$FF,$DA,$7D,$3F,$ED,$5E,$EF,$F6,$BF,$85,$FE,$D9,$EA,$7F,$8F
	.dc.b	$7F,$80,$1E,$FD,$F5,$B1,$EF,$CD,$FF,$DF,$D4,$F8,$1E,$1E,$BF,$1F
	.dc.b	$84,$F4,$72,$7F,$B3,$BC,$2F,$C6,$F2,$FA,$B1,$F4,$13,$A3,$E2,$BE
	.dc.b	$C6,$B9,$E8,$F8,$EE,$5A,$E4,$F9,$26,$4A,$67,$A3,$E5,$B9,$6B,$9A
	.dc.b	$74,$7C,$98,$F6,$FB,$1F,$0F,$DF,$F8,$95,$F2,$67,$E0,$FC,$37,$C7
	.dc.b	$F8,$3F,$93,$1F,$13,$93,$E4,$3F,$1B,$8F,$E4,$2D,$7F,$AE,$71,$FF
	.dc.b	$26,$8C,$6D,$FE,$EC,$87,$C3,$7D,$3F,$84,$F9,$75,$FE,$CC,$D3,$E5
	.dc.b	$61,$F1,$89,$87,$C4,$6C,$FC,$26,$F0,$7C,$06,$D7,$2A,$E0,$4A,$E0
	.dc.b	$DA,$E0,$D9,$E1,$3A,$24,$FE,$9B,$83,$E5,$47,$B0,$43,$27,$07,$CD
	.dc.b	$9F,$C5,$DF,$CD,$78,$56,$F7,$F9,$D1,$FE,$10,$E7,$FC,$B6,$FF,$3A
	.dc.b	$37,$8B,$7F,$9F,$AF,$CE,$F4,$7F,$39,$C1,$F2,$EF,$F0,$5A,$7E,$4F
	.dc.b	$8F,$F9,$0E,$2F,$84,$FD,$11,$2B,$F3,$ED,$7B,$B7,$1F,$B6,$DE,$1F
	.dc.b	$B4,$D9,$F8,$35,$D5,$E6,$43,$99,$6B,$9B,$6B,$9B,$FB,$0F,$86,$FD
	.dc.b	$B7,$C2,$FE,$FA,$3F,$C6,$7E,$07,$EF,$9F,$DA,$FC,$2F,$EF,$9F,$F1
	.dc.b	$BF,$77,$EF,$DF,$DC,$FC,$1F,$EF,$FE,$1B,$F8,$07,$C2,$7F,$AF,$7F
	.dc.b	$80,$7F,$D8,$1A,$FE,$5B,$C1,$FD,$EE,$1F,$C0,$3C,$0F,$67,$57,$DB
	.dc.b	$E2,$FD,$F7,$37,$E1,$7C,$0F,$CB,$DF,$EB,$5B,$8F,$E4,$1B,$C7,$B9
	.dc.b	$FA,$71,$87,$F1,$7C,$DF,$B4,$7F,$C9,$FE,$3F,$BD,$FF,$26,$F1,$FB
	.dc.b	$DF,$F2,$7F,$8B,$EF,$7F,$CA,$3C,$7F,$B4,$7F,$0B,$D7,$F8,$7C,$FF
	.dc.b	$63,$D7,$F8,$6D,$3F,$61,$C3,$EE,$BF,$C5,$F0,$FA,$BA,$BF,$8C,$EF
	.dc.b	$E5,$E2,$FD,$1D,$79,$A3,$1C,$1F,$A4,$9C,$91,$C1,$23,$67,$0E,$FF
	.dc.b	$D3,$E9,$FC,$13,$BF,$F4,$98,$7E,$72,$FF,$49,$3F,$C2,$3B,$FF,$47
	.dc.b	$3F,$E6,$EE,$FF,$D1,$BF,$CD,$3C,$4F,$63,$F9,$0E,$0F,$E7,$9C,$DF
	.dc.b	$A2,$8F,$E7,$3E,$0F,$B6,$3E,$94,$7D,$73,$E9,$90,$C1,$31,$FB,$4E
	.dc.b	$5A,$FF,$2B,$D3,$D7,$C5,$FC,$6F,$5F,$5D,$7F,$20,$D7,$D7,$3F,$C7
	.dc.b	$F5,$F5,$CF,$F9,$DF,$5F,$5E,$3F,$BC,$D7,$D7,$BF,$F0,$9D,$BD,$73
	.dc.b	$FD,$4B,$C1,$FD,$E3,$E6,$E1,$FD,$DF,$4F,$E5,$BC,$1F,$D8,$4F,$EB
	.dc.b	$89,$9F,$EB,$9B,$AF,$EB,$9A,$7E,$E2,$B9,$11,$FD,$A1,$92,$E5,$FA
	.dc.b	$36,$DB,$C1,$FD,$BC,$7F,$AA,$3C,$7F,$EE,$8F,$FA,$AB,$C7,$FE,$E1
	.dc.b	$E1,$7E,$67,$3F,$F4,$C7,$5F,$ED,$EB,$F8,$EF,$85,$FB,$77,$FB,$8F
	.dc.b	$83,$FB,$77,$FB,$C7,$67,$E6,$6F,$FA,$6F,$67,$9E,$7F,$C3,$5E,$0F
	.dc.b	$F3,$24,$7F,$2E,$70,$DD,$FF,$81,$11,$DD,$7F,$AC,$E7,$6D,$9D,$4F
	.dc.b	$13,$7C,$3D,$A1,$5C,$23,$0D,$9F,$E8,$AD,$9D,$9D,$2B,$B4,$F4,$37
	.dc.b	$19,$AE,$6D,$8E,$7F,$15,$FC,$6F,$DA,$FE,$39,$ED,$7F,$1D,$F8,$6F
	.dc.b	$E3,$DE,$9F,$F1,$FF,$86,$FE,$41,$ED,$CC,$7B,$65,$E8,$F6,$FF,$92
	.dc.b	$7C,$17,$F2,$5F,$56,$56,$B8,$1F,$0D,$FC,$A7,$E0,$7F,$95,$7A,$DF
	.dc.b	$CA,$FE,$07,$C9,$9F,$B3,$0F,$A3,$E0,$7F,$9C,$31,$7C,$0F,$F3,$12
	.dc.b	$4F,$56,$BF,$5D,$BC,$DF,$4D,$BC,$DF,$31,$B3,$F3,$1B,$87,$D2,$6D
	.dc.b	$F8,$27,$B1,$0E,$C7,$B7,$E1,$BE,$53,$EF,$F8,$2F,$E8,$DF,$0D,$FD
	.dc.b	$1F,$E0,$BF,$A4,$7C,$17,$C9,$AF,$C6,$72,$FF,$4B,$E5,$F9,$31,$F1
	.dc.b	$F9,$3E,$4B,$FD,$FB,$F8,$05,$BF,$C9,$8F,$99,$C7,$F2,$5F,$27,$0F
	.dc.b	$BD,$F9,$3E,$87,$BD,$CD,$57,$D2,$49,$E0,$D8,$D4,$60,$8E,$09,$1F
	.dc.b	$24,$EA,$53,$91,$63,$8B,$63,$F3,$BB,$FC,$88,$FD,$76,$FF,$21,$FC
	.dc.b	$46,$DF,$21,$FC,$56,$BF,$27,$0F,$D7,$6B,$F2,$27,$F6,$DA,$FC,$89
	.dc.b	$F8,$F1,$FE,$2C,$D6,$BF,$BB,$E5,$FC,$8B,$F9,$5E,$5F,$BF,$C7,$FD
	.dc.b	$DA,$3F,$52,$7E,$67,$93,$F2,$BC,$9F,$B1,$23,$82,$46,$C8,$7E,$FC
	.dc.b	$98,$FE,$D9,$BC,$1F,$B4,$77,$19,$3F,$85,$76,$FF,$12,$8F,$C1,$7C
	.dc.b	$1F,$ED,$9F,$D6,$76,$7E,$D9,$FD,$77,$5E,$EF,$E0,$E2,$D9,$F0,$79
	.dc.b	$CF,$F8,$84,$3E,$0F,$18,$FC,$7F,$87,$B8,$71,$F8,$4B,$6F,$84,$BF
	.dc.b	$0F,$FE,$23,$F0,$33,$F0,$BA,$F8,$7D,$FC,$37,$C4,$D3,$F9,$6E,$77
	.dc.b	$83,$DD,$BF,$C1,$9F,$86,$7D,$02,$CF,$F0,$36,$F2,$FD,$07,$E3,$CF
	.dc.b	$1A,$E7,$1C,$5B,$5C,$5B,$4E,$5F,$C8,$47,$F3,$7F,$1B,$F9,$05,$FF
	.dc.b	$17,$F1,$BF,$90,$3F,$90,$21,$FD,$0F,$B3,$F2,$15,$FC,$FF,$B3,$F2
	.dc.b	$19,$FF,$38,$DF,$F6,$D7,$FD,$27,$C2,$FD,$B3,$EA,$D7,$F1,$C7,$F0
	.dc.b	$F7,$F5,$2D,$7F,$52,$DC,$3F,$52,$DB,$FD,$8B,$67,$F5,$84,$70,$4D
	.dc.b	$7F,$8D,$47,$46,$D7,$46,$D7,$45,$E0,$FC,$68,$FE,$A7,$E0,$FE,$D6
	.dc.b	$FF,$B2,$78,$3F,$B5,$7F,$8D,$9F,$1B,$0F,$DA,$BF,$D0,$7C,$6F,$8A
	.dc.b	$FF,$5A,$D7,$F1,$78,$7C,$6A,$FE,$D1,$A7,$C6,$7C,$9D,$FF,$1B,$4F
	.dc.b	$2E,$1E,$D6,$EB,$ED,$9F,$DC,$CF,$CA,$9C,$92,$33,$1F,$05,$B5,$DA
	.dc.b	$3F,$9D,$1F,$D5,$FC,$7F,$F3,$F6,$BE,$F9,$FE,$F9,$AF,$BD,$FE,$FD
	.dc.b	$E1,$FF,$13,$D3,$FB,$BF,$0F,$E1,$1F,$EF,$3B,$7C,$0D,$7F,$65,$1E
	.dc.b	$F3,$FB,$EE,$DE,$F7,$E8,$6B,$EF,$9F,$D3,$47,$E7,$49,$AF,$E7,$5B
	.dc.b	$1F,$9D,$6E,$DF,$A5,$9F,$C7,$1F,$A6,$38,$12,$78,$36,$3F,$8E,$72
	.dc.b	$7E,$FB,$C1,$FE,$1D,$3F,$E2,$1F,$07,$F8,$73,$FE,$23,$F0,$7F,$81
	.dc.b	$F8,$3F,$C1,$3C,$1F,$E0,$BE,$0F,$F0,$6F,$07,$F8,$3F,$81,$FC,$1B
	.dc.b	$0F,$F0,$17,$57,$F0,$67,$FC,$13,$E0,$BB,$FF,$83,$C7,$9F,$83,$F8
	.dc.b	$07,$8E,$F1,$7F,$8B,$FB,$9F,$C4,$89,$5C,$9A,$FF,$67,$6F,$57,$F2
	.dc.b	$2C,$7F,$D2,$04,$8C,$F2,$F4,$AE,$44,$39,$16,$B9,$4F,$08,$FE,$98
	.dc.b	$DC,$7F,$97,$36,$FB,$4F,$1B,$EC,$F4,$25,$F4,$74,$74,$2D,$F4,$6D
	.dc.b	$74,$79,$63,$F1,$5F,$EB,$DE,$DF,$F3,$62,$3F,$39,$AF,$F8,$2D,$B7
	.dc.b	$F1,$5B,$87,$F4,$46,$CF,$43,$91,$0E,$45,$9E,$4D,$9E,$4F,$DE,$F8
	.dc.b	$0F,$E9,$3E,$97,$F4,$AF,$80,$FE,$97,$E9,$FF,$4C,$F8,$0F,$E9,$BF
	.dc.b	$0D,$FD,$3B,$D4,$FE,$9F,$EE,$7F,$52,$F3,$1F,$03,$7D,$DC,$7B,$B9
	.dc.b	$F5,$B5,$F0,$DB,$F5,$F8,$7C,$0F,$1F,$86,$F5,$F4,$FF,$63,$E7,$FE
	.dc.b	$63,$D3,$FE,$E9,$E6,$F5,$93,$A3,$D6,$DE,$8F,$5B,$4F,$88,$5B,$E8
	.dc.b	$FC,$13,$33,$C8,$93,$C9,$B1,$A8,$C1,$1F,$F0,$C7,$47,$C8,$7F,$4F
	.dc.b	$CF,$F2,$1F,$D4,$F3,$7C,$87,$F5,$7C,$BF,$21,$FF,$3A,$F2,$7C,$86
	.dc.b	$F1,$FC,$87,$3C,$5F,$21,$DF,$0F,$C8,$78,$F0,$7C,$8D,$FE,$37,$4F
	.dc.b	$E3,$F8,$3F,$29,$5F,$9A,$DF,$F3,$BC,$1F,$9B,$8C,$1F,$9B,$2C,$F0
	.dc.b	$8D,$12,$78,$92,$78,$46,$5B,$19,$9E,$4D,$BF,$D2,$56,$E7,$35,$C0
	.dc.b	$8E,$09,$3C,$2B,$97,$57,$EE,$49,$5F,$B0,$7F,$86,$E1,$F0,$6B,$84
	.dc.b	$F4,$6D,$F4,$AE,$33,$F0,$27,$24,$8C,$E1,$E7,$8C,$4F,$72,$1C,$67
	.dc.b	$78,$79,$A3,$35,$C6,$7B,$E1,$E5,$77,$3E,$56,$E1,$E5,$8E,$37,$CD
	.dc.b	$EC,$60,$E0,$58,$E0,$E8,$97,$E1,$75,$87,$A9,$CD,$7F,$20,$C3,$C9
	.dc.b	$1A,$C3,$C7,$5B,$BE,$53,$C0,$F0,$97,$0F,$0C,$68,$98,$7A,$5B,$3C
	.dc.b	$1B,$3C,$27,$47,$BF,$0F,$CD,$6A,$F8,$2F,$E1,$C4,$C3,$F5,$2D,$58
	.dc.b	$7F,$9B,$27,$84,$68,$86,$AF,$FD,$F9,$1A,$8C,$0C,$11,$9A,$E9,$7D
	.dc.b	$DB,$5D,$1A,$F2,$AE,$33,$BB,$EE,$D9,$ED,$6F,$82,$FE,$55,$F0,$DF
	.dc.b	$CA,$FB,$BF,$96,$7C,$1F,$F2,$DF,$84,$FE,$5D,$F0,$9F,$CB,$FE,$0F
	.dc.b	$F9,$87,$C3,$7F,$31,$F8,$4F,$E6,$5F,$09,$FC,$CF,$E1,$BF,$9A,$7C
	.dc.b	$2F,$FA,$1B,$F8,$07,$FD,$2D,$C1,$E5,$7D,$C8,$C7,$A7,$BF,$73,$CD
	.dc.b	$CB,$EA,$9F,$F6,$B9,$F1,$69,$F0,$1D,$49,$1C,$F9,$FF,$D3,$B3,$A8
	.dc.b	$C7,$3F,$7B,$4E,$67,$C9,$7C,$9A,$70,$3C,$0E,$6F,$AB,$6F,$A5,$73
	.dc.b	$9C,$D7,$1B,$EF,$7B,$8E,$66,$6B,$85,$F2,$9D,$57,$0A,$D4,$62,$B7
	.dc.b	$1C,$8F,$39,$2F,$C2,$DB,$E8,$F2,$21,$93,$94,$70,$24,$70,$87,$47
	.dc.b	$F3,$DA,$7A,$9F,$74,$93,$EB,$6B,$F7,$DB,$7F,$51,$B7,$F2,$1B,$3F
	.dc.b	$75,$B8,$FD,$66,$DF,$82,$7B,$10,$EC,$7C,$BF,$80,$FC,$87,$AB,$F9
	.dc.b	$1F,$73,$F2,$5E,$E7,$E4,$FD,$FF,$CA,$7B,$BF,$95,$F5,$3F,$2D,$F0
	.dc.b	$9F,$97,$F7,$BF,$31,$F0,$9F,$99,$F7,$FF,$35,$F0,$5F,$9B,$F8,$2F
	.dc.b	$CE,$7C,$37,$E7,$7E,$0F,$F3,$DF,$0D,$E6,$E4,$FF,$5C,$F2,$79,$F9
	.dc.b	$3C,$A4,$E7,$F8,$4F,$A5,$AE,$79,$FE,$23,$96,$B9,$3E,$31,$92,$99
	.dc.b	$E7,$F9,$0E,$5A,$E6,$9C,$FF,$1A,$3F,$25,$CD,$F1,$9F,$CA,$72,$FC
	.dc.b	$67,$F2,$DC,$9F,$19,$FC,$C7,$1F,$C6,$7F,$35,$C5,$F1,$9F,$CE,$70
	.dc.b	$FC,$67,$F3,$DC,$1F,$19,$F2,$67,$F1,$7A,$3F,$19,$D3,$F9,$D2,$4F
	.dc.b	$E5,$B8,$3F,$26,$FE,$5B,$83,$F2,$91,$8E,$0F,$D3,$4F,$E6,$DB,$19
	.dc.b	$C3,$F4,$F1,$8A,$E2,$4A,$E8,$48,$D3,$67,$89,$2F,$CB,$1D,$E3,$0F
	.dc.b	$9E,$BA,$57,$22,$4E,$A7,$F6,$F8,$7C,$07,$81,$DE,$FA,$E1,$E9,$6E
	.dc.b	$1D,$2F,$94,$E6,$B8,$E1,$DA,$F7,$1C,$CC,$D7,$0B,$E5,$3A,$AE,$15
	.dc.b	$A8,$C5,$79,$F0,$F3,$4F,$BE,$72,$47,$81,$0D,$17,$0F,$33,$82,$38
	.dc.b	$26,$1E,$48,$E0,$D7,$6D,$AF,$1C,$F0,$9F,$71,$A2,$B9,$C3,$DB,$19
	.dc.b	$6C,$67,$0F,$89,$3C,$2F,$C1,$39,$C3,$DD,$3C,$6B,$77,$D3,$0F,$F3
	.dc.b	$07,$F0,$03,$1F,$EB,$97,$FD,$CE,$30,$FF,$43,$6B,$98,$DE,$3A,$B5
	.dc.b	$6A,$30,$D9,$FE,$B4,$DA,$FF,$04,$36,$B8,$CF,$43,$65,$9D,$B6,$37
	.dc.b	$1E,$58,$C4,$F3,$23,$82,$46,$CF,$67,$C2,$70,$F8,$4E,$3F,$09,$CB
	.dc.b	$E1,$39,$FC,$27,$4F,$84,$EB,$F0,$FD,$BD,$0E,$FF,$0B,$E0,$F8,$7F
	.dc.b	$0F,$AB,$E2,$F7,$BC,$7E,$F7,$93,$E1,$BC,$BE,$FF,$9B,$E1,$BC,$7C
	.dc.b	$5F,$2C,$FF,$03,$F2,$FC,$AF,$80,$F5,$92,$3E,$93,$5F,$AC,$DE,$7F
	.dc.b	$9E,$DE,$6F,$9E,$D9,$F9,$ED,$AE,$D1,$D0,$87,$42,$D7,$47,$F0,$5F
	.dc.b	$0D,$F8,$6E,$9F,$C4,$B7,$D8,$FB,$AD,$9F,$BA,$DB,$FC,$2B,$7A,$8D
	.dc.b	$24,$FC,$28,$CB,$63,$2F,$F7,$CA,$FC,$5F,$C0,$FA,$E9,$95,$25,$73
	.dc.b	$F8,$6F,$C2,$91,$FC,$E3,$5F,$D0,$37,$3F,$CC,$B7,$4F,$CC,$B6,$7F
	.dc.b	$32,$DC,$FF,$42,$DC,$0D,$35,$FC,$03,$FE,$D9,$AF,$F3,$02,$F8,$3F
	.dc.b	$CE,$2F,$86,$BF,$09,$8F,$86,$CF,$C2,$EB,$E1,$B7,$F0,$DC,$3E,$17
	.dc.b	$8F,$C2,$F2,$F8,$5F,$D1,$F2,$74,$E5,$EB,$F0,$BD,$BE,$17,$BF,$C2
	.dc.b	$F9,$78,$BE,$31,$F1,$08,$7C,$42,$F3,$FC,$46,$F3,$FC,$46,$9F,$10
	.dc.b	$BC,$FF,$11,$BC,$FF,$11,$BC,$FE,$17,$B1,$F1,$3E,$1F,$E1,$F4,$7C
	.dc.b	$4E,$7F,$87,$3F,$E5,$5E,$87,$3F,$C4,$8C,$73,$7C,$47,$5C,$BF,$11
	.dc.b	$E1,$C9,$F1,$1E,$5C,$7F,$11,$E9,$C5,$F1,$1F,$29,$F9,$3E,$9F,$C5
	.dc.b	$BF,$A5,$27,$17,$E9,$5B,$C5,$FB,$07,$F5,$2D,$72,$7E,$C8,$C9,$4C
	.dc.b	$F1,$7E,$D1,$BC,$5F,$9C,$77,$1F,$AA,$7F,$48,$43,$F4,$87,$EC,$0C
	.dc.b	$75,$FE,$D9,$B9,$7E,$D9,$A6,$CE,$33,$D5,$B5,$FB,$26,$E3,$F0,$B1
	.dc.b	$F8,$04,$C7,$DD,$8F,$F4,$C9,$D4,$7C,$19,$D5,$7B,$DB,$87,$BD,$B8
	.dc.b	$7B,$E7,$E0,$DF,$27,$85,$F3,$23,$82,$5F,$A7,$0F,$81,$1E,$AE,$BF
	.dc.b	$E1,$31,$82,$65,$F0,$E7,$47,$C4,$AE,$51,$99,$DC,$F6,$C7,$CD,$97
	.dc.b	$A3,$2F,$C2,$D7,$C1,$AF,$AF,$5A,$CB,$CF,$8F,$BD,$B1,$99,$E9,$7E
	.dc.b	$FA,$D4,$60,$E7,$19,$9E,$12,$E0,$FE,$19,$8F,$F7,$37,$F8,$65,$FE
	.dc.b	$9F,$87,$F5,$2F,$F1,$0E,$1F,$D4,$3F,$C4,$B8,$7F,$50,$FF,$14,$E1
	.dc.b	$FD,$43,$FC,$5B,$87,$F5,$0F,$D4,$3C,$64,$3F,$0A,$5C,$7E,$23,$75
	.dc.b	$F7,$EB,$E9,$24,$64,$F7,$1F,$C0,$4F,$DF,$71,$7F,$01,$7F,$80,$75
	.dc.b	$FF,$23,$7F,$81,$75,$FF,$23,$7F,$82,$75,$FF,$23,$7F,$83,$75,$FF
	.dc.b	$23,$7F,$C3,$5D,$7E,$07,$D4,$7E,$EF,$91,$CB,$FE,$EF,$E3,$7C,$37
	.dc.b	$F3,$42,$57,$F1,$16,$BF,$CE,$9B,$BF,$D1,$6E,$FC,$DB,$2E,$5F,$B1
	.dc.b	$18,$6C,$75,$21,$D4,$B5,$D5,$FE,$E6,$7E,$5F,$9B,$F6,$D9,$7E,$5A
	.dc.b	$BF,$33,$3F,$BD,$8C,$4E,$B2,$FC,$E1,$27,$4D,$AF,$C6,$7C,$37,$F5
	.dc.b	$4F,$86,$FE,$AB,$F0,$DF,$D5,$BE,$1B,$D1,$B7,$C6,$C7,$C5,$D7,$FA
	.dc.b	$EC,$B9,$67,$DE,$7A,$93,$3F,$29,$23,$4D,$AE,$53,$B9,$C9,$23,$33
	.dc.b	$B9,$F0,$90,$FD,$D9,$34,$F8,$33,$C2,$B9,$4E,$E7,$2D,$BF,$C1,$7C
	.dc.b	$37,$F7,$42,$47,$F5,$16,$BF,$DE,$9B,$B7,$EC,$5B,$C1,$CA,$78,$10
	.dc.b	$E0,$59,$E0,$D8,$E0,$E8,$91,$FB,$DE,$EF,$EC,$B1,$FB,$D3,$F4,$A7
	.dc.b	$F6,$5E,$CF,$DE,$CF,$F0,$3E,$CF,$DE,$BF,$C1,$7B,$3F,$7A,$FF,$07
	.dc.b	$EC,$FD,$EB,$FE,$7B,$F8,$3F,$F4,$1E,$08,$C7,$07,$F9,$C6,$5F,$0D
	.dc.b	$7E,$1B,$DF,$7F,$9D,$E3,$6F,$FE,$41,$6C,$7B,$4D,$47,$0C,$3F,$8D
	.dc.b	$36,$77,$3F,$03,$7F,$87,$97,$C6,$8C,$1C,$8E,$97,$C9,$B5,$CA,$78
	.dc.b	$12,$B9,$9F,$BF,$DF,$E5,$D7,$F8,$D3,$C0,$FE,$4F,$5F,$CA,$F8,$3E
	.dc.b	$5D,$FF,$05,$E0,$F9,$6F,$F0,$7E,$0F,$96,$FF,$09,$E0,$F9,$6B,$1F
	.dc.b	$E0,$1C,$1F,$33,$1F,$97,$19,$7E,$87,$07,$CB,$8F,$4E,$FF,$97,$E4
	.dc.b	$91,$F9,$53,$F4,$8E,$1F,$CF,$71,$7E,$C2,$31,$C7,$F9,$C9,$C9,$C0
	.dc.b	$E5,$C5,$FA,$97,$2D,$8E,$04,$AE,$93,$C3,$93,$F5,$2F,$06,$CE,$E3
	.dc.b	$B1,$9B,$F1,$D7,$28,$C1,$CE,$32,$F0,$9E,$57,$E3,$AE,$11,$8B,$EF
	.dc.b	$18,$25,$71,$6D,$71,$8C,$11,$C1,$27,$9C,$65,$B1,$9B,$F2,$B6,$7D
	.dc.b	$04,$3D,$64,$BF,$34,$F0,$AE,$64,$70,$49,$E5,$3C,$1B,$7E,$AA,$E1
	.dc.b	$18,$3C,$D5,$FC,$8B,$37,$AD,$FE,$AF,$6C,$FF,$B3,$1A,$FF,$8F,$9B
	.dc.b	$87,$F6,$08,$FC,$AC,$F1,$23,$C5,$AF,$16,$C7,$1C,$55,$D3,$1F,$D0
	.dc.b	$46,$2B,$71,$89,$D5,$F8,$0E,$C4,$3B,$1E,$9F,$7F,$D9,$EF,$FB,$7E
	.dc.b	$1B,$DD,$F0,$1E,$FB,$F8,$1F,$01,$F0,$7E,$1B,$E1,$7C,$0F,$C3,$F8
	.dc.b	$6F,$89,$F0,$5F,$17,$E1,$BE,$37,$C1,$FC,$7F,$86,$F9,$1F,$09,$F2
	.dc.b	$7E,$1B,$E5,$7C,$2F,$A3,$8B,$CC,$7C,$9F,$86,$F9,$C4,$7E,$73,$5D
	.dc.b	$18,$21,$82,$73,$FC,$9E,$04,$F1,$8C,$37,$97,$E4,$36,$BF,$85,$B7
	.dc.b	$9F,$E8,$B6,$F9,$CF,$82,$F9,$B6,$F9,$B4,$E6,$5B,$E6,$DA,$E6,$FA
	.dc.b	$27,$8B,$4E,$25,$9E,$2F,$FA,$33,$F8,$05,$DD,$FC,$8E,$30,$DC,$3F
	.dc.b	$15,$3A,$9F,$E5,$11,$89,$D5,$FE,$32,$B5,$5C,$8D,$C6,$67,$71,$FE
	.dc.b	$2F,$9D,$D7,$38,$D9,$99,$DD,$AB,$77,$CE,$B7,$39,$AD,$D6,$A7,$91
	.dc.b	$A7,$13,$B8,$C4,$F0,$38,$FC,$37,$2F,$86,$E7,$F0,$DD,$3E,$1B,$AF
	.dc.b	$C3,$76,$F8,$6E,$FF,$0D,$E0,$F8,$6F,$0F,$C3,$78,$BE,$1B,$C7,$F0
	.dc.b	$DE,$4F,$86,$F2,$FC,$37,$9B,$E1,$BC,$FF,$0D,$E8,$F8,$6F,$17,$1F
	.dc.b	$90,$87,$CB,$27,$3F,$C7,$6F,$3F,$C3,$69,$F0,$0B,$CD,$EF,$6F,$3F
	.dc.b	$C2,$27,$3F,$82,$32,$DB,$EC,$7E,$0B,$9F,$E2,$4F,$CB,$E6,$F8,$8F
	.dc.b	$CD,$E5,$F8,$8F,$CF,$E4,$F8,$8F,$D1,$E3,$F8,$8F,$D3,$E2,$F8,$8F
	.dc.b	$D5,$E1,$F8,$8F,$D7,$E0,$F8,$8F,$C2,$CF,$C7,$BF,$E7,$9F,$D4,$C7
	.dc.b	$E7,$7A,$3F,$47,$18,$24,$EC,$93,$FA,$48,$CB,$63,$3C,$1F,$A6,$9C
	.dc.b	$B6,$33,$5D,$9B,$D3,$FB,$D2,$5F,$ED,$5A,$F9,$5B,$7D,$2B,$94,$F8
	.dc.b	$49,$7D,$DB,$1A,$8C,$11,$C1,$2B,$AD,$F9,$2B,$C4,$43,$C4,$5B,$F1
	.dc.b	$4F,$08,$EE,$DB,$EC,$DB,$ED,$3C,$6F,$B3,$D0,$97,$D1,$D1,$D0,$B7
	.dc.b	$D1,$B5,$D1,$E3,$5D,$2B,$94,$60,$E4,$5A,$E4,$DA,$E5,$3C,$09,$5C
	.dc.b	$C9,$3C,$A3,$2D,$8C,$D7,$49,$CB,$63,$35,$D9,$B7,$F1,$49,$7E,$C7
	.dc.b	$FA,$17,$0B,$E0,$BF,$8A,$92,$7F,$8A,$35,$FE,$68,$DC,$7F,$98,$B6
	.dc.b	$FF,$97,$36,$7F,$95,$37,$1F,$E4,$CD,$BF,$04,$F6,$21,$D8,$FE,$49
	.dc.b	$E5,$FF,$28,$F7,$7F,$94,$FC,$27,$F3,$0F,$E0,$07,$B1,$7E,$13,$1E
	.dc.b	$CE,$7E,$13,$5E,$D6,$FE,$13,$87,$B7,$C7,$E1,$39,$7B,$9C,$FE,$13
	.dc.b	$A7,$BB,$D7,$97,$B7,$C3,$F9,$F9,$7C,$1F,$0D,$39,$7F,$9E,$F1,$23
	.dc.b	$10,$E8,$FF,$0F,$4E,$63,$73,$A9,$CB,$70,$FF,$10,$D6,$EB,$94,$62
	.dc.b	$75,$0E,$85,$6E,$F9,$CF,$62,$1C,$8E,$11,$BB,$E9,$1A,$9F,$3F,$C3
	.dc.b	$FC,$DF,$87,$F9,$DF,$0F,$2B,$DF,$F0,$FF,$44,$FA,$5F,$0F,$23,$E1
	.dc.b	$FC,$3F,$D5,$F8,$3F,$EC,$BF,$C0,$3F,$EC,$4C,$FF,$C8,$0F,$A6,$BF
	.dc.b	$B2,$BF,$EC,$46,$FC,$47,$0F,$84,$E3,$F0,$9F,$03,$87,$DF,$C5,$8B
	.dc.b	$E3,$35,$F7,$4B,$04,$E7,$1F,$0B,$78,$7F,$C9,$F3,$AB,$E6,$43,$9B
	.dc.b	$96,$CF,$17,$E9,$46,$6B,$AC,$71,$6B,$C0,$C5,$78,$3E,$13,$E2,$FC
	.dc.b	$27,$CD,$9F,$4F,$23,$8B,$E4,$FC,$27,$CA,$F2,$3E,$5F,$C2,$7C,$CF
	.dc.b	$84,$FD,$17,$27,$CE,$F8,$4F,$9F,$97,$D0,$F8,$4F,$A3,$F0,$9F,$4B
	.dc.b	$9B,$E6,$4F,$88,$F4,$F1,$FA,$F9,$3E,$43,$79,$3D,$9C,$7F,$1A,$30
	.dc.b	$4A,$E5,$5C,$1B,$5C,$1E,$A6,$27,$83,$F0,$C9,$C9,$F8,$4E,$4F,$87
	.dc.b	$1F,$AC,$E2,$F8,$93,$F8,$8E,$2F,$85,$3F,$88,$E1,$F8,$4F,$E4,$0F
	.dc.b	$C5,$EF,$F0,$9F,$C6,$EF,$F0,$9F,$C6,$9F,$90,$E6,$F1,$E1,$F5,$8F
	.dc.b	$D4,$C7,$D2,$E6,$FC,$87,$2F,$E9,$E3,$F5,$31,$82,$3F,$A9,$E0,$FD
	.dc.b	$5B,$C4,$C1,$C0,$BC,$1F,$9A,$6D,$75,$8C,$47,$E0,$BA,$FF,$64,$FE
	.dc.b	$9B,$E1,$3F,$57,$F0,$9F,$B1,$8F,$D9,$9F,$C4,$BA,$FF,$64,$FE,$63
	.dc.b	$E1,$3F,$64,$FE,$BB,$E1,$3F,$67,$8F,$ED,$3E,$1B,$F6,$BF,$0B,$FB
	.dc.b	$6F,$85,$FD,$BF,$A9,$EC,$C3,$D4,$4E,$1F,$8A,$4E,$0F,$7B,$63,$DC
	.dc.b	$6E,$32,$47,$31,$BC,$7D,$ED,$E0,$F8,$44,$E1,$F1,$47,$F1,$4A,$FE
	.dc.b	$5F,$E1,$FE,$13,$0F,$E4,$FD,$1F,$0E,$7F,$53,$E1,$FC,$2B,$FE,$63
	.dc.b	$C3,$FA,$69,$FE,$6F,$E1,$FE,$15,$FD,$EE,$9F,$0D,$FC,$29,$FC,$03
	.dc.b	$9B,$F4,$9A,$F8,$E3,$F3,$44,$3F,$33,$9F,$E9,$5C,$63,$F8,$B8,$C4
	.dc.b	$68,$FD,$19,$6F,$F5,$0D,$ED,$FD,$44,$62,$B9,$4F,$07,$F4,$CF,$F2
	.dc.b	$8F,$0F,$E1,$3F,$CF,$35,$FE,$0E,$FF,$2C,$F0,$FF,$83,$BF,$CB,$BC
	.dc.b	$3F,$84,$FF,$1C,$EC,$F8,$58,$7F,$34,$EC,$FD,$34,$FF,$20,$F0,$7F
	.dc.b	$83,$BE,$5F,$07,$F4,$FB,$7B,$2F,$CB,$AF,$9E,$30,$4A,$F4,$36,$35
	.dc.b	$18,$23,$82,$6D,$F0,$89,$7E,$09,$DC,$3C,$1F,$E0,$F1,$FC,$A3,$6F
	.dc.b	$87,$7F,$D0,$7C,$1F,$E0,$F3,$FC,$EB,$C1,$FE,$0E,$FF,$2E,$F0,$FF
	.dc.b	$94,$3F,$D4,$8F,$E5,$07,$F5,$4F,$0F,$F9,$43,$FC,$1C,$FB,$BA,$7A
	.dc.b	$F9,$3F,$09,$1F,$9F,$3F,$3B,$1F,$9F,$C7,$F3,$AD,$9D,$ED,$F9,$FE
	.dc.b	$AF,$CE,$CE,$5B,$7C,$89,$7F,$DF,$BC,$3F,$E0,$71,$FD,$53,$C3,$FE
	.dc.b	$06,$FF,$02,$3F,$42,$7F,$03,$21,$FC,$0C,$FD,$09,$FC,$0C,$FE,$E3
	.dc.b	$3F,$A1,$7F,$AF,$F7,$FF,$03,$7F,$BA,$F7,$79,$25,$E3,$FF,$B5,$89
	.dc.b	$EE,$FF,$B8,$9B,$5F,$EE,$06,$BF,$DD,$1B,$A3,$44,$FF,$86,$E7,$31
	.dc.b	$8B,$EF,$7D,$49,$5D,$5A,$F5,$6D,$F5,$8E,$91,$E5,$F7,$7C,$DF,$0B
	.dc.b	$E7,$F7,$BD,$1F,$0B,$E9,$F7,$FD,$5F,$0B,$EB,$F8,$0F,$67,$C2,$FB
	.dc.b	$7E,$07,$DD,$F0,$BE,$FF,$82,$F8,$1F,$0B,$F0,$7E,$0F,$E1,$7C,$2F
	.dc.b	$C3,$F8,$4F,$89,$F0,$BE,$4E,$6F,$67,$3F,$BB,$97,$CA,$7A,$C8,$7A
	.dc.b	$CB,$CD,$EB,$E7,$F8,$71,$8A,$E7,$CF,$D6,$39,$9C,$49,$5C,$6F,$C1
	.dc.b	$7C,$A7,$37,$C9,$E8,$7D,$3E,$7F,$8E,$FD,$BF,$5B,$EF,$7C,$27,$C7
	.dc.b	$8F,$BD,$CD,$F1,$DF,$C0,$7C,$27,$C7,$7F,$05,$F0,$9F,$79,$FF,$1A
	.dc.b	$FF,$00,$FF,$9E,$1F,$C3,$71,$FC,$97,$F0,$7B,$7B,$E5,$F0,$9F,$2E
	.dc.b	$7F,$19,$C5,$EF,$DF,$D5,$7E,$DE,$1C,$70,$FB,$88,$E0,$97,$EE,$AE
	.dc.b	$3C,$3F,$06,$3E,$19,$F2,$AF,$A3,$BB,$FD,$07,$0F,$C4,$8F,$CE,$70
	.dc.b	$FC,$47,$E1,$90,$FD,$07,$4F,$E9,$78,$3E,$24,$7E,$8F,$83,$E2,$63
	.dc.b	$EF,$F8,$5F,$C5,$CF,$EA,$79,$B3,$A7,$E8,$F8,$7E,$1C,$FC,$22,$1E
	.dc.b	$8D,$BD,$6F,$FA,$16,$1F,$0B,$CA,$30,$D9,$4E,$A7,$E8,$D7,$7A,$D4
	.dc.b	$E6,$7E,$CC,$60,$8F,$03,$A9,$6B,$AC,$6A,$38,$99,$9E,$98,$FD,$F9
	.dc.b	$D3,$67,$51,$8B,$F2,$47,$1F,$85,$F2,$FC,$2F,$EE,$E3,$F8,$87,$7F
	.dc.b	$A3,$E1,$7F,$77,$1D,$FE,$17,$F7,$6F,$87,$E1,$7F,$76,$FF,$17,$EE
	.dc.b	$FD,$E3,$E5,$F8,$5F,$DD,$BE,$7E,$7F,$DD,$BE,$42,$1F,$45,$FC,$7B
	.dc.b	$F8,$7E,$8F,$C7,$B9,$6E,$FF,$8C,$8C,$1B,$21,$FC,$9F,$A3,$F1,$D1
	.dc.b	$FB,$9F,$13,$F8,$BC,$FE,$36,$3F,$79,$E2,$7E,$C6,$3F,$97,$78,$9F
	.dc.b	$8E,$9F,$E0,$1E,$1F,$B2,$7F,$99,$77,$FE,$BD,$FE,$69,$E1,$FE,$B5
	.dc.b	$FE,$6D,$E1,$F9,$B9,$BF,$49,$87,$9C,$8E,$09,$3F,$A7,$6F,$4F,$E4
	.dc.b	$89,$BF,$C4,$9F,$79,$2B,$9B,$C4,$93,$FD,$3B,$6F,$83,$7F,$CA,$3B
	.dc.b	$FF,$5D,$5F,$D2,$FC,$2F,$E2,$4F,$EB,$8F,$E2,$27,$F5,$1F,$13,$FA
	.dc.b	$A7,$7F,$E6,$BB,$FF,$5B,$3F,$D4,$FB,$FF,$3B,$E1,$FC,$1D,$7F,$93
	.dc.b	$78,$7F,$06,$BC,$7B,$FA,$67,$CF,$BF,$B3,$9F,$4D,$AF,$B3,$3A,$AE
	.dc.b	$1B,$FC,$47,$F8,$4C,$FF,$65,$F1,$7F,$B4,$70,$7B,$2F,$F9,$B7,$07
	.dc.b	$B1,$FE,$71,$E2,$7F,$09,$BF,$61,$FC,$B0,$FD,$56,$FE,$C9,$FE,$7D
	.dc.b	$BF,$B2,$7F,$82,$C7,$B1,$F2,$E7,$FA,$A7,$F6,$64,$8F,$D8,$36,$35
	.dc.b	$18,$23,$82,$47,$F0,$03,$81,$63,$6D,$8D,$E9,$FC,$12,$3F,$53,$3A
	.dc.b	$8F,$E2,$7E,$0F,$EF,$27,$FC,$15,$D5,$FA,$07,$FA,$4F,$7F,$EF,$1F
	.dc.b	$E9,$7D,$FF,$BC,$7F,$A6,$F7,$FE,$F1,$FF,$08,$77,$FE,$87,$2F,$E9
	.dc.b	$9D,$FF,$BC,$9F,$EA,$7C,$7F,$BC,$7D,$3A,$79,$5F,$EA,$FF,$0B,$FC
	.dc.b	$88,$8F,$F1,$96,$BA,$30,$43,$04,$BF,$EB,$91,$C1,$B1,$B6,$BB,$DB
	.dc.b	$FA,$04,$ED,$B1,$FC,$A5,$FE,$6B,$F0,$BF,$CD,$BC,$CF,$E6,$FE,$67
	.dc.b	$F3,$8F,$80,$97,$FC,$BF,$D3,$FE,$77,$F0,$1F,$CF,$3E,$03,$F9,$EF
	.dc.b	$A7,$FC,$FB,$E0,$3F,$A2,$7F,$00,$FF,$77,$C1,$BF,$F8,$46,$1C,$1A
	.dc.b	$F8,$3D,$FB,$9C,$3E,$0B,$CD,$87,$F5,$1E,$3B,$7F,$D3,$9B,$E8,$9A
	.dc.b	$49,$F8,$5C,$9E,$E7,$E1,$1C,$CC,$94,$CD,$75,$72,$D7,$39,$7F,$AB
	.dc.b	$0F,$F4,$43,$FD,$47,$8F,$DF,$F0,$5F,$03,$DB,$F8,$3F,$05,$F0,$BD
	.dc.b	$CF,$8F,$7F,$D6,$F0,$7C,$17,$C6,$F8,$2F,$8F,$F0,$5F,$22,$BE,$4E
	.dc.b	$9F,$15,$E1,$C3,$F3,$3D,$FF,$74,$7A,$38,$7E,$13,$E8,$6B,$9E,$1D
	.dc.b	$F2,$7E,$10,$9C,$5F,$15,$B3,$F0,$DB,$8F,$C1,$6D,$F4,$6D,$FD,$EF
	.dc.b	$82,$FB,$FF,$03,$F8,$9C,$3D,$DC,$7F,$0D,$F8,$1C,$5F,$0D,$FE,$F1
	.dc.b	$D5,$F8,$6E,$1F,$85,$1F,$17,$83,$E1,$3F,$1F,$5F,$84,$FE,$2B,$6F
	.dc.b	$F3,$E7,$13,$87,$E2,$4F,$C1,$DB,$F2,$3F,$01,$23,$E8,$35,$E4,$7E
	.dc.b	$4A,$91,$9E,$0F,$CB,$37,$0E,$ED,$AE,$0D,$C5,$7C,$A7,$B9,$0E,$F1
	.dc.b	$A9,$E1,$3B,$9C,$B6,$FF,$CB,$17,$FB,$49,$FC,$47,$C3,$7E,$D1,$FF
	.dc.b	$10,$77,$7E,$D1,$FF,$31,$F7,$7E,$C3,$E1,$BF,$69,$1F,$E2,$5E,$EF
	.dc.b	$DA,$3F,$91,$F8,$6F,$DA,$FC,$1F,$ED,$BE,$1B,$F6,$FF,$09,$FB,$1A
	.dc.b	$F4,$6B,$F0,$9F,$98,$43,$F1,$85,$7F,$06,$DD,$7F,$18,$47,$F1,$8D
	.dc.b	$E0,$FC,$63,$7A,$BF,$18,$FF,$95,$4F,$81,$3F,$E7,$EE,$FF,$80,$FF
	.dc.b	$A0,$37,$F8,$18,$7F,$9F,$77,$FD,$6C,$FF,$A1,$3C,$3F,$E2,$6F,$F9
	.dc.b	$17,$C3,$FC,$5B,$F3,$B3,$FC,$5D,$FF,$13,$E6,$F8,$13,$EC,$24,$7B
	.dc.b	$1B,$9F,$9F,$9B,$F5,$6D,$8F,$80,$7B,$89,$7E,$E6,$EB,$EE,$7F,$21
	.dc.b	$DB,$FA,$87,$F8,$51,$FA,$83,$F8,$37,$83,$FC,$29,$FE,$11,$E0,$FB
	.dc.b	$B2,$FE,$55,$D7,$FA,$89,$FC,$4F,$5F,$BA,$7F,$97,$65,$EE,$CF,$F4
	.dc.b	$FA,$FE,$9A,$FD,$67,$93,$5F,$5B,$4F,$5D,$7A,$49,$AF,$A5,$B3,$C1
	.dc.b	$B3,$C3,$AB,$F9,$37,$7F,$F0,$38,$FC,$EF,$7F,$F0,$37,$F3,$FD,$FF
	.dc.b	$C0,$DF,$F5,$DF,$7F,$F0,$37,$F9,$37,$83,$FC,$91,$FE,$50,$7F,$24
	.dc.b	$3F,$95,$78,$7F,$D2,$1F,$E5,$9E,$17,$8F,$2F,$D5,$6F,$FB,$89,$FC
	.dc.b	$DB,$5F,$D7,$F8,$5F,$B2,$23,$A3,$F7,$A5,$D7,$F7,$0D,$AE,$AE,$8E
	.dc.b	$66,$67,$FD,$3F,$5F,$9A,$E1,$FE,$C1,$E1,$7E,$69,$FF,$6A,$75,$FE
	.dc.b	$6A,$7F,$D6,$9D,$7F,$B9,$C3,$FD,$39,$D7,$F9,$A7,$FD,$BF,$D7,$FD
	.dc.b	$B3,$C0,$FD,$CC,$7F,$4B,$EF,$FD,$CB,$FD,$37,$C3,$F5,$BF,$C4,$3F
	.dc.b	$6A,$7A,$BF,$EB,$76,$C7,$FA,$B5,$AE,$E3,$44,$74,$D7,$46,$08,$60
	.dc.b	$9A,$B4,$47,$F8,$02,$BB,$CE,$63,$15,$D6,$31,$C2,$AD,$DF,$28,$C5
	.dc.b	$6E,$31,$2F,$6B,$DD,$EB,$7B,$FD,$0F,$81,$F0,$3F,$07,$D6,$F8,$5F
	.dc.b	$03,$F0,$FD,$7F,$89,$F0,$3F,$17,$DD,$F8,$DF,$03,$F1,$FD,$4F,$91
	.dc.b	$F0,$3F,$27,$DA,$F9,$5F,$03,$F2,$FD,$E9,$CD,$F0,$B9,$BD,$3C,$7E
	.dc.b	$88,$F8,$1F,$03,$F4,$48,$FD,$16,$BF,$5D,$B3,$C9,$BC,$BE,$E6,$CF
	.dc.b	$43,$91,$0E,$45,$9E,$4D,$9E,$4F,$E0,$79,$7F,$09,$EF,$FC,$78,$FC
	.dc.b	$27,$27,$E1,$FE,$17,$E3,$C7,$E1,$F9,$3E,$3B,$F8,$9E,$2F,$8E,$FE
	.dc.b	$2F,$87,$E3,$BF,$8D,$F8,$5F,$8E,$FE,$3F,$E0,$FF,$06,$FF,$64,$A7
	.dc.b	$C1,$7B,$3B,$3F,$01,$C1,$78,$4D,$25,$FA,$CF,$84,$7C,$02,$70,$FC
	.dc.b	$06,$F0,$FC,$06,$9C,$23,$44,$8D,$36,$3F,$31,$C3,$F1,$9F,$D4,$75
	.dc.b	$7E,$31,$FD,$57,$07,$C6,$7F,$61,$C1,$F1,$9F,$D9,$6B,$F1,$9F,$DA
	.dc.b	$70,$7C,$67,$F6,$DC,$1F,$19,$FD,$C7,$07,$C6,$CF,$F4,$77,$F1,$A7
	.dc.b	$E0,$6F,$F1,$9F,$C8,$6B,$FA,$2E,$CF,$D2,$47,$EB,$0C,$94,$FD,$AC
	.dc.b	$6D,$CC,$FE,$B1,$BC,$1F,$99,$8C,$47,$03,$96,$1F,$9E,$BE,$F3,$A8
	.dc.b	$C4,$FF,$00,$7D,$3F,$07,$FB,$48,$C6,$1F,$BB,$C5,$F0,$BE,$F8,$C7
	.dc.b	$C2,$7E,$DE,$B6,$7F,$A2,$29,$F0,$BF,$B9,$8C,$E4,$97,$C2,$F3,$9E
	.dc.b	$1F,$0B,$6B,$5F,$0B,$6D,$F0,$B7,$E1,$7F,$7D,$87,$49,$FE,$C1,$C2
	.dc.b	$ED,$FD,$FC,$FC,$3C,$7E,$06,$DF,$2E,$31,$5F,$24,$9C,$5F,$19,$B1
	.dc.b	$F3,$A3,$04,$70,$43,$9F,$4F,$C9,$7F,$9E,$74,$FE,$45,$FE,$7D,$E2
	.dc.b	$7E,$47,$97,$F8,$8F,$89,$F9,$17,$F8,$9E,$BF,$91,$7F,$8A,$F4,$FE
	.dc.b	$45,$FE,$2F,$E0,$FE,$45,$FE,$33,$BF,$B9,$FB,$78,$FB,$3B,$7F,$62
	.dc.b	$FE,$AB,$83,$F2,$4D,$DF,$F6,$EF,$E5,$9A,$E6,$78,$D7,$EB,$6B,$8B
	.dc.b	$6B,$6D,$9F,$EB,$1B,$7E,$C6,$FE,$09,$FC,$58,$FD,$89,$FC,$54,$87
	.dc.b	$F6,$1F,$13,$F9,$E3,$FD,$7F,$C4,$FE,$78,$FF,$61,$F1,$3F,$B2,$77
	.dc.b	$7C,$1E,$4F,$E9,$7D,$DF,$85,$F0,$7C,$F1,$FC,$10,$97,$FC,$25,$BB
	.dc.b	$7A,$9B,$B7,$A9,$BA,$F2,$AE,$04,$AE,$0D,$AE,$0D,$9F,$E4,$5A,$FC
	.dc.b	$1D,$BF,$A6,$F2,$FE,$05,$FE,$9F,$E0,$FE,$07,$2F,$EC,$7E,$0F,$E0
	.dc.b	$5F,$EC,$BC,$BF,$81,$7F,$B3,$F2,$FE,$05,$FE,$D3,$C1,$F0,$74,$FE
	.dc.b	$ED,$C1,$F0,$5F,$B5,$AF,$E5,$5F,$CE,$F4,$5D,$3F,$0A,$4C,$3F,$44
	.dc.b	$D9,$FE,$0F,$18,$23,$82,$70,$7E,$A9,$BC,$1F,$A6,$7F,$A3,$78,$3F
	.dc.b	$A1,$BF,$E4,$7E,$0F,$EE,$7C,$1F,$E0,$D5,$FD,$2F,$C1,$FE,$08,$FF
	.dc.b	$4D,$F0,$7F,$7D,$E0,$FE,$FF,$C1,$FE,$01,$E0,$FE,$FE,$BF,$C4,$5D
	.dc.b	$DF,$C0,$FD,$7F,$D0,$70,$7F,$83,$3B,$FD,$CF,$F1,$73,$F8,$41,$2B
	.dc.b	$F8,$33,$5F,$EB,$8D,$DB,$F9,$CB,$4E,$51,$C0,$87,$02,$C7,$06,$C7
	.dc.b	$07,$44,$8F,$E8,$BF,$0B,$FC,$6F,$E0,$3F,$8E,$7C,$2F,$F1,$DF,$43
	.dc.b	$F8,$F7,$A1,$FC,$7F,$D7,$FE,$41,$E8,$FF,$21,$F8,$4F,$E4,$5E,$94
	.dc.b	$D3,$F9,$5F,$F0,$0F,$F7,$67,$FD,$B2,$BA,$FF,$95,$2F,$62,$FB,$D2
	.dc.b	$B8,$BE,$F7,$F9,$C7,$0D,$BF,$E6,$CD,$F0,$FF,$C8,$53,$EB,$8F,$77
	.dc.b	$17,$C1,$8C,$1C,$8F,$86,$53,$35,$D1,$CB,$5C,$D7,$69,$E8,$4A,$FE
	.dc.b	$A5,$C5,$F2,$6B,$E0,$FA,$7F,$25,$FE,$85,$C1,$F2,$5F,$E8,$9B,$FC
	.dc.b	$2F,$83,$F8,$7E,$F7,$C4,$F7,$BE,$2D,$7C,$6D,$7E,$3F,$C1,$FC,$58
	.dc.b	$F9,$7B,$DD,$BC,$51,$EB,$6D,$7A,$5B,$BF,$A7,$7F,$8A,$4A,$E0,$DA
	.dc.b	$E0,$D9,$E1,$3A,$24,$E9,$B1,$A8,$C1,$1C,$0D,$FE,$34,$FE,$13,$7F
	.dc.b	$8C,$FF,$8E,$F7,$F8,$CF,$F8,$F7,$7F,$8C,$FE,$33,$6F,$8C,$FE,$38
	.dc.b	$F8,$C7,$E4,$35,$F8,$99,$FF,$9F,$F8,$5B,$62,$BC,$F3,$F9,$2F,$7E
	.dc.b	$47,$E6,$1A,$FE,$61,$BB,$FF,$A3,$2D,$18,$AE,$CD,$AF,$CF,$4E,$9E
	.dc.b	$84,$3A,$16,$BA,$4F,$1D,$D5,$C6,$B6,$7E,$AB,$E0,$FF,$53,$F0,$7F
	.dc.b	$B6,$8F,$F0,$77,$4F,$EB,$3E,$07,$F5,$BF,$07,$FA,$EF,$82,$FD,$B5
	.dc.b	$7F,$84,$FA,$7F,$6C,$FE,$D3,$E0,$7F,$67,$F0,$3F,$B4,$EB,$93,$FB
	.dc.b	$AF,$86,$FD,$BF,$C1,$79,$77,$F2,$71,$7E,$33,$8F,$F6,$44,$BF,$D4
	.dc.b	$F5,$FE,$02,$31,$1F,$AA,$6C,$F1,$8D,$C7,$F8,$AF,$AF,$F0,$B3,$FE
	.dc.b	$58,$E1,$FC,$2B,$FE,$5A,$EA,$FC,$2B,$FE,$5C,$EA,$F6,$E1,$FE,$83
	.dc.b	$E0,$F6,$CF,$D7,$8F,$C2,$9F,$C4,$B6,$FC,$17,$37,$CE,$E8,$BA,$F9
	.dc.b	$EB,$C6,$43,$D0,$5B,$F4,$38,$23,$82,$6B,$FA,$9D,$7D,$6F,$E4,$FA
	.dc.b	$FD,$6F,$F9,$67,$5F,$5B,$FE,$5B,$D7,$F3,$F8,$7F,$28,$EB,$FC,$FB
	.dc.b	$FC,$A8,$F5,$E3,$F9,$3E,$BF,$5F,$0F,$E7,$A7,$F9,$49,$E7,$C3,$F7
	.dc.b	$47,$D0,$2B,$93,$F4,$65,$DB,$F7,$11,$B8,$C9,$FC,$8B,$4F,$5B,$F9
	.dc.b	$ED,$3D,$6F,$F9,$B7,$4F,$5B,$FE,$6F,$D3,$D7,$9F,$EA,$74,$F5,$BF
	.dc.b	$56,$7F,$A3,$E5,$FC,$79,$F4,$9F,$C7,$8B,$1F,$31,$FD,$6F,$5F,$E4
	.dc.b	$C8,$FE,$AF,$3F,$D9,$B4,$D9,$6B,$F5,$B3,$C2,$3F,$5B,$5F,$B2,$75
	.dc.b	$A7,$E7,$AF,$9C,$69,$FE,$19,$D7,$FB,$28,$FD,$89,$0F,$EB,$FD,$5F
	.dc.b	$B2,$7F,$B1,$F4,$FE,$5B,$0F,$F5,$9F,$27,$E5,$B0,$FE,$2D,$D3,$FB
	.dc.b	$2B,$FE,$D3,$CB,$FB,$27,$FB,$4F,$47,$EC,$9F,$2E,$BF,$DA,$BE,$1A
	.dc.b	$47,$F7,$06,$BF,$DC,$1B,$97,$F1,$96,$E3,$FC,$95,$B3,$FC,$61,$BB
	.dc.b	$FF,$12,$6E,$E6,$92,$7F,$BF,$FC,$1F,$F7,$CF,$86,$FF,$53,$7F,$00
	.dc.b	$FF,$B0,$49,$ED,$63,$E1,$73,$ED,$EB,$E0,$F7,$F0,$5C,$3D,$CE,$3F
	.dc.b	$05,$CB,$E0,$B9,$F8,$3D,$3E,$0B,$AF,$BF,$DB,$E0,$BC,$5C,$2E,$67
	.dc.b	$32,$3F,$C2,$1C,$2F,$4F,$CA,$48,$FF,$60,$C6,$5B,$19,$7A,$4F,$3E
	.dc.b	$75,$6E,$72,$76,$AE,$17,$E2,$8C,$1D,$08,$70,$AE,$73,$C2,$FB,$5F
	.dc.b	$5B,$E5,$5A,$BE,$31,$89,$D5,$70,$BF,$F1,$6F,$2F,$D3,$BF,$6F,$C2
	.dc.b	$FD,$4F,$4F,$EA,$F3,$7D,$39,$F8,$3E,$FF,$D3,$7E,$1F,$27,$DA,$F8
	.dc.b	$5F,$A7,$1F,$1B,$8F,$E9,$BF,$23,$E0,$FE,$FF,$C1,$FD,$38,$F2,$ED
	.dc.b	$E3,$69,$F0,$8F,$81,$BF,$A1,$C7,$07,$AF,$83,$E1,$46,$2B,$93,$C2
	.dc.b	$33,$C3,$F0,$E3,$07,$67,$15,$CE,$31,$7D,$E7,$37,$D2,$BF,$BE,$70
	.dc.b	$FC,$7B,$F2,$F0,$FC,$77,$F0,$BC,$1F,$1D,$F4,$F0,$7C,$77,$D7,$C1
	.dc.b	$F1,$DF,$C5,$F3,$7E,$65,$FC,$6F,$37,$E9,$38,$3E,$3C,$7E,$43,$93
	.dc.b	$D7,$BF,$97,$B7,$F5,$D3,$FB,$B2,$1F,$BB,$2E,$DF,$BB,$6E,$FF,$BB
	.dc.b	$69,$FB,$B2,$F0,$7E,$ED,$B5,$D9,$E8,$4A,$E8,$DA,$E8,$FD,$8F,$83
	.dc.b	$FD,$63,$F9,$9F,$E0,$17,$C2,$FF,$00,$F8,$59,$3F,$B1,$C3,$F3,$34
	.dc.b	$F8,$4E,$31,$8F,$84,$FE,$07,$5B,$CB,$F5,$D6,$F8,$4D,$C6,$3E,$13
	.dc.b	$F8,$35,$6C,$FF,$46,$53,$E1,$37,$18,$F8,$4B,$C0,$A7,$C2,$7F,$0B
	.dc.b	$AF,$E0,$3E,$0E,$A5,$F0,$9F,$C3,$36,$F8,$64,$D3,$E1,$B6,$FE,$71
	.dc.b	$F1,$CA,$E7,$1F,$A1,$19,$6C,$67,$8F,$E9,$37,$C9,$F0,$5F,$62,$5F
	.dc.b	$F9,$33,$BB,$F2,$D5,$FC,$07,$8F,$E7,$CF,$EB,$3B,$7F,$2D,$3F,$C1
	.dc.b	$7C,$7F,$CB,$3F,$C1,$FB,$7E,$4D,$7E,$CB,$B7,$F2,$CF,$ED,$3B,$7E
	.dc.b	$23,$FB,$6E,$1F,$DE,$3E,$9D,$BC,$FA,$FE,$A4,$91,$FB,$72,$6F,$FA
	.dc.b	$E6,$F5,$FE,$C4,$9E,$07,$E8,$09,$1F,$B6,$6B,$FB,$C2,$69,$FB,$67
	.dc.b	$FD,$9B,$E4,$FF,$4E,$F0,$7F,$6D,$1F,$04,$FE,$77,$E4,$FF,$52,$F0
	.dc.b	$BE,$0E,$7F,$CB,$38,$3F,$6D,$5F,$04,$FE,$38,$7E,$D8,$FE,$38,$7F
	.dc.b	$44,$F0,$FE,$0E,$5F,$C2,$E3,$C7,$5F,$03,$5F,$71,$35,$F6,$CF,$AC
	.dc.b	$87,$03,$E1,$EB,$EE,$BE,$84,$BE,$95,$CA,$BF,$DD,$3C,$BF,$80,$9F
	.dc.b	$E8,$BA,$FC,$0B,$FE,$85,$C9,$F8,$09,$FE,$93,$AF,$C0,$C3,$FB,$0E
	.dc.b	$BF,$02,$FF,$A4,$71,$FE,$01,$FE,$95,$C7,$FC,$56,$BE,$01,$FC,$54
	.dc.b	$F4,$C7,$E5,$5F,$E1,$3C,$DF,$96,$7F,$3F,$8F,$E5,$B7,$FC,$EB,$78
	.dc.b	$FF,$0F,$3A,$24,$E9,$B1,$A8,$C1,$1C,$6E,$F0,$7F,$6D,$E0,$FE,$83
	.dc.b	$2F,$D1,$61,$FC,$0F,$0F,$DF,$78,$3F,$BB,$F0,$7F,$86,$CF,$F1,$7F
	.dc.b	$07,$F7,$DE,$0F,$E8,$B0,$FE,$2B,$E0,$FF,$01,$F0,$7F,$81,$78,$3F
	.dc.b	$A2,$9F,$ED,$DD,$BF,$C1,$7C,$0F,$81,$AF,$F0,$BF,$86,$FE,$10,$4D
	.dc.b	$3C,$5E,$12,$31,$1F,$C9,$DB,$87,$F2,$46,$F2,$7F,$20,$6C,$FF,$1C
	.dc.b	$6D,$F7,$8E,$B1,$8B,$ED,$18,$21,$FC,$D7,$E1,$7F,$8E,$5F,$F1,$DF
	.dc.b	$63,$F8,$F7,$C2,$FF,$1F,$F7,$3F,$90,$7C,$2F,$F2,$1F,$6A,$63,$EA
	.dc.b	$F8,$5F,$E4,$9E,$DC,$8F,$6F,$C2,$FF,$27,$F6,$3F,$96,$FB,$7F,$D1
	.dc.b	$27,$FC,$38,$B8,$72,$79,$F8,$64,$F9,$B8,$BE,$13,$78,$3E,$13,$4F
	.dc.b	$84,$5A,$F8,$4F,$CB,$31,$C5,$F0,$9C,$12,$BA,$B6,$BA,$B4,$F3,$60
	.dc.b	$F6,$3D,$7E,$FF,$C7,$9F,$C6,$4B,$DE,$F9,$11,$F7,$B8,$BE,$3B,$FD
	.dc.b	$8F,$87,$FA,$57,$0F,$C7,$8F,$EC,$DC,$3F,$D3,$78,$7E,$44,$7E,$1B
	.dc.b	$87,$E4,$FC,$2F,$CA,$F7,$FE,$2C,$F9,$78,$BD,$FC,$3E,$A2,$70,$FA
	.dc.b	$9B,$3F,$18,$E1,$C5,$F0,$E3,$35,$CA,$B8,$12,$B8,$4E,$AB,$84,$E6
	.dc.b	$B8,$56,$A3,$FC,$C5,$C1,$F1,$E3,$E2,$F0,$FC,$77,$F0,$3C,$3F,$19
	.dc.b	$FF,$03,$70,$FC,$6C,$BF,$C0,$7C,$3F,$1A,$7F,$4F,$C1,$F1,$E7,$FC
	.dc.b	$0F,$C1,$F1,$A7,$FC,$1D,$C1,$F1,$9F,$A7,$BF,$C6,$AF,$F3,$2E,$8F
	.dc.b	$85,$FC,$99,$2B,$EC,$B5,$FC,$AB,$67,$8B,$76,$F6,$B6,$7F,$0A,$DD
	.dc.b	$D7,$DE,$7A,$90,$EA,$5B,$EA,$FF,$04,$E8,$FD,$9B,$F8,$CF,$81,$FD
	.dc.b	$9B,$F8,$EF,$77,$F6,$6F,$E4,$2F,$F5,$BF,$0B,$FA,$EF,$82,$FD,$7F
	.dc.b	$C2,$FE,$C3,$E0,$FF,$63,$F0,$7F,$B2,$F8,$5F,$D9,$E3,$FB,$03,$F3
	.dc.b	$3B,$7A,$72,$F4,$47,$E1,$BB,$7E,$57,$0F,$DF,$EA,$FC,$0C,$FD,$F2
	.dc.b	$5F,$E0,$89,$1A,$6D,$72,$9D,$CE,$49,$1F,$E2,$DE,$0F,$83,$3F,$E4
	.dc.b	$9F,$03,$E0,$BF,$86,$3F,$94,$F7,$FE,$1F,$1F,$F2,$1F,$4F,$E1,$E7
	.dc.b	$F5,$67,$E2,$0F,$E2,$67,$EA,$CF,$DB,$9F,$04,$FE,$27,$97,$94,$F3
	.dc.b	$BF,$C2,$31,$C6,$5E,$68,$C1,$0F,$8C,$5E,$4F,$D3,$4F,$F3,$4D,$3F
	.dc.b	$45,$3F,$C8,$3B,$3F,$45,$5F,$94,$EC,$FD,$13,$FE,$62,$EC,$F6,$57
	.dc.b	$F9,$63,$3F,$64,$FF,$99,$BB,$3F,$45,$AF,$F9,$8B,$B3,$F4,$53,$FE
	.dc.b	$6A,$EC,$FD,$16,$1F,$CB,$BB,$3D,$3B,$48,$F3,$61,$F6,$E3,$11,$EA
	.dc.b	$25,$FA,$9D,$9F,$C8,$BB,$FF,$90,$BF,$D6,$B8,$FF,$9E,$ED,$EA,$C3
	.dc.b	$FC,$EB,$B7,$AB,$8B,$FC,$D9,$BF,$AA,$7F,$CF,$3B,$FA,$B1,$FF,$3A
	.dc.b	$EF,$EA,$9F,$F3,$EE,$FE,$AD,$7F,$73,$BF,$D5,$E1,$FC,$AC,$7E,$A3
	.dc.b	$0F,$D5,$91,$C1,$2F,$F6,$F1,$FA,$F6,$BB,$6C,$6E,$3F,$76,$6C,$A6
	.dc.b	$CA,$EF,$3F,$DF,$F6,$FE,$66,$7F,$BE,$D7,$F6,$0E,$DF,$CC,$DF,$F1
	.dc.b	$2E,$DF,$DC,$D7,$F6,$2C,$7F,$72,$FF,$4A,$EC,$FD,$CB,$FD,$33,$AF
	.dc.b	$F7,$2F,$F4,$EE,$AF,$DC,$BF,$B8,$21,$E5,$D3,$C6,$FF,$54,$F8,$4F
	.dc.b	$EE,$24,$70,$43,$FB,$9B,$96,$99,$CB,$F9,$4C,$65,$B1,$9D,$3F,$9F
	.dc.b	$CE,$5B,$19,$3F,$B4,$C6,$5F,$ED,$7F,$0B,$FE,$04,$F4,$3F,$C0,$BE
	.dc.b	$BF,$F8,$1B,$DB,$FF,$59,$FF,$00,$BD,$EB,$EC,$E3,$DD,$CF,$BD,$AF
	.dc.b	$86,$DF,$B7,$C3,$E1,$B8,$FB,$9C,$BE,$0F,$9F,$BB,$D3,$E1,$BA,$FA
	.dc.b	$7E,$3E,$7F,$E6,$FC,$B7,$AC,$D2,$73,$7A,$E7,$6D,$9D,$B6,$76,$FC
	.dc.b	$A2,$74,$7C,$86,$F4,$7C,$67,$FC,$47,$D1,$F0,$3E,$0F,$E0,$FC,$17
	.dc.b	$C3,$9F,$F0,$6F,$47,$C4,$F8,$2F,$85,$1F,$E5,$BE,$97,$37,$C8,$F8
	.dc.b	$6F,$93,$F0,$7F,$0E,$B5,$CB,$F0,$DE,$1C,$BF,$0D,$E5,$C9,$E8,$E0
	.dc.b	$F3,$13,$83,$CC,$D9,$FB,$67,$BC,$87,$B0,$C9,$78,$FE,$3B,$78,$FE
	.dc.b	$3B,$78,$FC,$17,$FE,$08,$E3,$F8,$53,$ED,$E2,$F8,$4F,$F9,$23,$8B
	.dc.b	$E1,$3F,$07,$A7,$F0,$6F,$C3,$E0,$F8,$2F,$C5,$E7,$FC,$1B,$F0,$08
	.dc.b	$35,$64,$8F,$F2,$A7,$1B,$7F,$8D,$1E,$73,$E8,$7C,$37,$E5,$88,$E0
	.dc.b	$87,$5A,$E9,$C0,$B5,$70,$AD,$1E,$07,$18,$79,$23,$07,$47,$15,$CE
	.dc.b	$77,$7D,$AF,$AD,$EE,$31,$3A,$8E,$67,$82,$B9,$3F,$8B,$F8,$6F,$DA
	.dc.b	$3F,$E5,$CE,$EF,$DA,$BF,$8F,$EE,$FD,$9B,$F9,$1F,$86,$FD,$BF,$C2
	.dc.b	$7E,$CE,$3F,$29,$DD,$F2,$DE,$DF,$C0,$0F,$85,$C6,$FE,$A8,$7C,$2E
	.dc.b	$37,$F8,$F0,$F8,$5F,$DC,$E1,$FE,$B9,$87,$C2,$FE,$EA,$3D,$C7,$B3
	.dc.b	$0F,$CB,$70,$7E,$41,$BD,$7F,$90,$6F,$5F,$E4,$2B,$F2,$F1,$87,$F3
	.dc.b	$6D,$7F,$00,$47,$FC,$97,$D7,$F1,$A3,$FD,$37,$C5,$F9,$69,$FA,$BC
	.dc.b	$BF,$C6,$67,$F6,$DE,$2F,$EE,$1F,$A1,$1F,$B8,$7F,$75,$E2,$FF,$1A
	.dc.b	$7F,$30,$7E,$E0,$FC,$BD,$74,$F1,$7F,$26,$F6,$F1,$7D,$5A,$FB,$F8
	.dc.b	$7F,$6E,$4B,$F6,$B6,$35,$18,$23,$82,$6D,$F1,$E7,$83,$67,$6D,$9D
	.dc.b	$ED,$E2,$86,$DF,$BA,$AF,$7F,$7F,$C5,$C3,$FD,$0B,$DF,$FB,$A9,$F8
	.dc.b	$7D,$FF,$BA,$7E,$2F,$7F,$EE,$9F,$8F,$DF,$FB,$A7,$E4,$F8,$1F,$C6
	.dc.b	$DF,$97,$E0,$7C,$4C,$7C,$FA,$F8,$C9,$3F,$72,$32,$D3,$D7,$1A,$7E
	.dc.b	$F4,$7F,$13,$7F,$97,$78,$BF,$CD,$1F,$E6,$64,$3E,$DF,$8B,$F7,$78
	.dc.b	$BF,$03,$E2,$FF,$33,$9F,$E6,$DE,$2F,$F3,$37,$F0,$3E,$2F,$A6,$BF
	.dc.b	$0F,$A7,$A6,$7F,$88,$E9,$F7,$3A,$3F,$77,$1F,$88,$F1,$7F,$75,$CD
	.dc.b	$FB,$88,$C1,$FA,$27,$18,$7E,$4A,$77,$5C,$A3,$15,$B8,$C6,$DF,$C1
	.dc.b	$A3,$15,$C0,$E3,$E2,$FE,$7E,$FF,$D5,$7D,$9F,$C1,$2B,$F3,$7E,$2F
	.dc.b	$F0,$47,$F3,$BE,$2F,$F0,$47,$F3,$A7,$F0,$43,$F7,$5E,$1F,$EF,$D6
	.dc.b	$8F,$07,$F8,$1C,$7E,$FF,$C2,$FE,$02,$FF,$0A,$F0,$BF,$80,$3F,$00
	.dc.b	$99,$FF,$0B,$F8,$5E,$98,$73,$6E,$06,$92,$BF,$86,$35,$FE,$2E,$DA
	.dc.b	$E9,$1A,$21,$A2,$BA,$70,$47,$04,$DF,$F8,$EC,$F5,$6C,$F2,$6B,$C5
	.dc.b	$B3,$C6,$3E,$3D,$7A,$DE,$06,$2B,$A3,$4C,$DF,$C3,$6C,$F1,$7B,$1D
	.dc.b	$3E,$17,$E3,$91,$FE,$60,$D7,$F9,$DB,$6F,$F9,$D3,$70,$FE,$72,$D9
	.dc.b	$E8,$72,$21,$C8,$B3,$C9,$B3,$C9,$FE,$81,$F0,$BF,$6B,$E1,$7E,$DF
	.dc.b	$C2,$FD,$CF,$6F,$EE,$FB,$FF,$7B,$0F,$BF,$ED,$FE,$03,$E1,$7F,$03
	.dc.b	$ED,$FE,$0B,$E0,$3F,$07,$EE,$7E,$13,$E1,$7F,$0B,$EE,$FE,$1B,$DD
	.dc.b	$FC,$3F,$BB,$F8,$AF,$E0,$1F,$EB,$5C,$5E,$72,$71,$F9,$DB,$B7,$9D
	.dc.b	$A7,$C0,$2F,$37,$E7,$79,$7E,$23,$63,$DE,$6E,$BA,$12,$BA,$36,$35
	.dc.b	$18,$23,$83,$F0,$BC,$BF,$1D,$FE,$D1,$C9,$F1,$DF,$ED,$5C,$7F,$1D
	.dc.b	$FE,$D9,$C7,$F1,$DF,$ED,$DC,$5F,$1D,$FE,$E1,$C3,$F1,$DF,$EE,$5C
	.dc.b	$1F,$1D,$FE,$E9,$BF,$C7,$7D,$BC,$CD,$FF,$27,$BF,$E4,$67,$F1,$ED
	.dc.b	$DF,$F3,$8D,$DF,$F3,$E4,$9F,$CD,$1C,$2F,$F3,$15,$C1,$B3,$BA,$E5
	.dc.b	$5C,$2B,$5C,$FE,$2B,$DC,$E6,$F8,$46,$2F,$89,$1C,$12,$B9,$12,$BF
	.dc.b	$56,$D7,$CC,$DB,$F0,$C6,$88,$68,$AE,$9C,$11,$C1,$2F,$CF,$3D,$5B
	.dc.b	$3C,$9A,$F1,$6C,$F1,$8F,$80,$4B,$F7,$57,$38,$C5,$FB,$DE,$86,$23
	.dc.b	$53,$C2,$32,$D7,$C0,$6C,$CC,$F3,$73,$1B,$AF,$23,$6F,$CE,$49,$E4
	.dc.b	$70,$AE,$55,$C1,$B3,$BA,$E5,$5C,$2B,$57,$E8,$BD,$CE,$6F,$84,$62
	.dc.b	$F8,$8E,$1F,$EA,$B5,$FE,$85,$C3,$FC,$E3,$18,$8F,$F3,$53,$63,$33
	.dc.b	$FC,$24,$93,$C6,$FF,$D0,$75,$B6,$CE,$A7,$9B,$6B,$9C,$71,$3F,$B3
	.dc.b	$D7,$16,$CE,$EB,$9C,$6E,$3D,$46,$5D,$99,$97,$C0,$7F,$1F,$F8,$1F
	.dc.b	$E4,$1F,$03,$FC,$87,$E1,$BF,$91,$7C,$17,$F2,$3F,$82,$FE,$49,$F0
	.dc.b	$5F,$C9,$7E,$0B,$F9,$37,$AF,$FC,$9F,$E1,$BF,$94,$7C,$1F,$F2,$9F
	.dc.b	$86,$FE,$55,$F0,$9F,$CA,$FE,$1B,$F9,$67,$C2,$FF,$2D,$F5,$FC,$9C
	.dc.b	$9E,$6E,$4F,$27,$BB,$FC,$C4,$8E,$08,$7D,$07,$2D,$33,$CF,$F5,$A3
	.dc.b	$2D,$8C,$F3,$FD,$29,$CB,$63,$27,$CE,$8C,$BF,$43,$E1,$7F,$A2,$79
	.dc.b	$FF,$D1,$7D,$DF,$F0,$7F,$87,$FE,$7F,$AF,$F5,$62,$F8,$2C,$FB,$DA
	.dc.b	$F6,$B7,$F0,$5C,$3E,$12,$61,$F5,$F0,$7B,$FD,$3D,$CE,$BE,$E7,$6F
	.dc.b	$83,$EF,$EE,$F8,$3E,$0F,$CB,$CB,$F9,$7E,$3B,$E3,$7F,$9C,$B4,$F5
	.dc.b	$CE,$5B,$19,$7E,$41,$39,$BE,$33,$79,$BE,$23,$4F,$88,$72,$BE,$E4
	.dc.b	$7D,$5F,$03,$F1,$3E,$0F,$E3,$47,$B7,$DD,$F8,$FF,$05,$F2,$3E,$0F
	.dc.b	$E4,$FC,$1F,$CA,$F8,$2F,$97,$F0,$5F,$33,$E0,$BE,$6F,$37,$C3,$CB
	.dc.b	$E3,$7C,$2F,$C3,$7F,$3D,$FC,$03,$FE,$CD,$6F,$C3,$78,$DF,$47,$27
	.dc.b	$C1,$6F,$27,$C1,$E4,$F7,$93,$93,$DE,$D9,$F7,$9C,$39,$3A,$46,$6F
	.dc.b	$A3,$6F,$A5,$F2,$25,$FF,$97,$BA,$9C,$BF,$16,$3F,$65,$E0,$7E,$2B
	.dc.b	$97,$E2,$C7,$E0,$F9,$3E,$2B,$F8,$5E,$3F,$8A,$FE,$1F,$8B,$E2,$BF
	.dc.b	$89,$EB,$FC,$9F,$17,$C5,$8F,$F7,$F7,$17,$C5,$6F,$17,$AB,$64,$7F
	.dc.b	$9C,$A7,$FD,$E0,$FC,$CF,$84,$FF,$35,$C6,$23,$9C,$60,$8E,$C8,$68
	.dc.b	$BC,$2A,$EF,$C4,$BE,$93,$9A,$F1,$5F,$47,$91,$23,$55,$C6,$75,$18
	.dc.b	$AE,$77,$D9,$B7,$FB,$93,$F9,$07,$85,$EC,$F8,$4F,$6F,$AD,$FB,$AF
	.dc.b	$13,$F6,$F5,$FB,$3F,$0F,$F6,$EF,$ED,$7C,$2F,$DB,$BF,$33,$D7,$F8
	.dc.b	$BE,$F7,$ED,$A3,$F2,$7F,$09,$FB,$67,$F7,$9E,$17,$9B,$6F,$27,$07
	.dc.b	$D1,$9F,$C6,$F0,$7E,$27,$B7,$1C,$5F,$8C,$BF,$C4,$B6,$B7,$18,$ED
	.dc.b	$FC,$7C,$60,$93,$B2,$46,$9B,$E4,$7F,$1A,$9F,$C8,$9F,$B5,$3F,$89
	.dc.b	$70,$7C,$27,$F8,$DF,$07,$C2,$7F,$8E,$F0,$7C,$2C,$FE,$9F,$77,$F1
	.dc.b	$89,$FC,$91,$FC,$6F,$C6,$FC,$94,$EB,$C5,$FE,$31,$3F,$C2,$48,$7E
	.dc.b	$B4,$B1,$F0,$DB,$AF,$BF,$6F,$8B,$B7,$AC,$8F,$C8,$C7,$F5,$B5,$FC
	.dc.b	$AF,$C5,$FE,$22,$FF,$34,$EF,$FD,$6B,$FC,$DB,$BF,$F5,$AF,$F3,$8E
	.dc.b	$FF,$D6,$BF,$CE,$BB,$FF,$5A,$FF,$3C,$EF,$FD,$6B,$FC,$DF,$C6,$FD
	.dc.b	$2F,$C2,$FE,$9B,$C0,$F2,$EB,$2B,$D2,$7D,$4D,$BC,$EE,$39,$FE,$AC
	.dc.b	$EE,$3E,$11,$FC,$63,$7F,$5B,$FD,$33,$7F,$5F,$0F,$D0,$E0,$F5,$BF
	.dc.b	$4B,$83,$D6,$FD,$4E,$0F,$5B,$F5,$B8,$3D,$6F,$D8,$E0,$F5,$E3,$FD
	.dc.b	$63,$83,$D6,$FF,$2A,$3C,$DE,$07,$EA,$B6,$FD,$84,$7E,$F3,$8B,$F3
	.dc.b	$6D,$3F,$72,$58,$DB,$FB,$98,$DB,$4D,$96,$37,$87,$EE,$FA,$7F,$51
	.dc.b	$3A,$8C,$3D,$78,$5D,$3F,$9B,$8F,$E3,$FD,$FF,$BC,$7F,$36,$7F,$6D
	.dc.b	$F2,$7F,$B9,$3F,$DB,$FC,$9F,$CD,$67,$FC,$53,$91,$D5,$FC,$83,$C0
	.dc.b	$FD,$ED,$FE,$6C,$FD,$E9,$FD,$DF,$AB,$F3,$6F,$A7,$83,$E3,$0F,$80
	.dc.b	$FE,$52,$48,$FE,$AE,$D7,$F9,$AB,$77,$FE,$34,$DC,$BF,$8D,$36,$7F
	.dc.b	$9B,$37,$87,$F9,$53,$6F,$C1,$3D,$88,$76,$3F,$B3,$FC,$2F,$F3,$CD
	.dc.b	$3F,$9E,$E1,$FC,$FB,$E0,$7F,$9F,$FC,$2F,$F4,$0F,$82,$FE,$83,$F0
	.dc.b	$BF,$D0,$BE,$0F,$FA,$1F,$C2,$CC,$FF,$B8,$FC,$27,$F4,$6F,$85,$FE
	.dc.b	$8F,$F0,$BF,$D3,$25,$C5,$FC,$E3,$F8,$07,$FD,$D6,$D8,$D3,$9E,$3F
	.dc.b	$57,$1E,$38,$3F,$A3,$DF,$F9,$5A,$FF,$CB,$8D,$7A,$C6,$BE,$17,$84
	.dc.b	$2D,$5B,$C3,$C5,$5A,$9C,$B6,$75,$7C,$EF,$BC,$E4,$8E,$08,$78,$8F
	.dc.b	$04,$F2,$6C,$70,$75,$5E,$07,$8D,$F9,$1F,$6F,$C2,$F8,$FE,$17,$C9
	.dc.b	$F0,$BF,$56,$7E,$0F,$C2,$F9,$FE,$17,$EA,$C7,$F8,$8B,$8B,$EB,$3F
	.dc.b	$1B,$E1,$7D,$9F,$0B,$F5,$63,$FC,$4F,$C3,$EF,$F8,$5F,$81,$F0,$BF
	.dc.b	$07,$E1,$7C,$DB,$F8,$EF,$CF,$C5,$EB,$E1,$F4,$46,$38,$B7,$C7,$F1
	.dc.b	$E3,$07,$B1,$C5,$F3,$23,$82,$47,$C6,$38,$14,$E6,$48,$E2,$E5,$FC
	.dc.b	$67,$1F,$C7,$8F,$C2,$F1,$7C,$77,$D3,$C5,$F1,$DF,$F2,$3F,$47,$C6
	.dc.b	$7F,$C4,$BC,$5F,$81,$E2,$F8,$F1,$FE,$4C,$E6,$FC,$DB,$F9,$0E,$1F
	.dc.b	$8E,$FE,$4B,$83,$E0,$BE,$3D,$7F,$6E,$FE,$C3,$5F,$DD,$B6,$32,$FE
	.dc.b	$CC,$91,$A6,$C6,$A3,$04,$70,$48,$FE,$06,$75,$29,$C8,$B1,$FE,$25
	.dc.b	$ED,$FE,$01,$1F,$E6,$89,$FD,$07,$C2,$FF,$00,$8F,$F3,$5F,$5F,$E8
	.dc.b	$FE,$17,$F4,$9F,$0B,$FB,$F9,$FF,$25,$F6,$FF,$91,$3B,$64,$7E,$D3
	.dc.b	$E1,$7F,$54,$C3,$F6,$FC,$FF,$BC,$7C,$CF,$F1,$86,$F8,$5F,$C6,$1F
	.dc.b	$44,$65,$B1,$9D,$BE,$14,$7B,$35,$F8,$55,$B9,$CB,$63,$35,$C5,$BD
	.dc.b	$FF,$C9,$F7,$F8,$78,$7F,$99,$A7,$F1,$18,$7F,$92,$FC,$0F,$E4,$53
	.dc.b	$F8,$83,$F5,$C7,$FA,$43,$4F,$89,$3F,$E5,$1D,$3F,$5D,$3F,$E7,$0E
	.dc.b	$DF,$84,$FE,$A8,$FE,$1D,$A7,$C3,$7F,$39,$8F,$E8,$A3,$F5,$B7,$F9
	.dc.b	$E6,$C6,$7B,$3F,$64,$4E,$CF,$C8,$CE,$AF,$F2,$44,$70,$49,$E1,$1F
	.dc.b	$AE,$8C,$76,$7E,$A6,$3F,$7F,$D9,$F0,$B0,$FE,$63,$D9,$FA,$99,$FF
	.dc.b	$45,$F6,$7E,$A7,$3F,$F3,$87,$67,$EA,$67,$F9,$AF,$77,$C2,$9F,$F3
	.dc.b	$A6,$DF,$C3,$E7,$F9,$CF,$77,$F0,$F7,$C5,$B7,$91,$F8,$67,$A5,$C5
	.dc.b	$7C,$36,$E7,$F0,$F7,$F8,$31,$8A,$E1,$18,$23,$FC,$B0,$FE,$93,$E0
	.dc.b	$7F,$2C,$7F,$D9,$3E,$0F,$F4,$1D,$BD,$D8,$FF,$A2,$F6,$F7,$65,$FE
	.dc.b	$8B,$DB,$DD,$87,$F2,$4D,$BD,$D3,$FE,$8F,$DB,$DD,$3F,$CA,$36,$F7
	.dc.b	$3E,$CE,$1F,$A7,$CF,$F9,$38,$FC,$D1,$23,$F3,$0D,$8D,$46,$08,$E0
	.dc.b	$9C,$FF,$8B,$D3,$F4,$31,$8A,$FD,$6C,$FE,$7D,$FF,$73,$F8,$5F,$E0
	.dc.b	$07,$FB,$5F,$6F,$F0,$17,$FD,$8F,$DB,$FC,$05,$FE,$DF,$D9,$F9,$ED
	.dc.b	$7F,$D5,$9D,$3F,$C0,$27,$FD,$9B,$D3,$FD,$A7,$B3,$F3,$FA,$FF,$7D
	.dc.b	$9F,$E0,$27,$88,$FE,$F3,$F0,$BF,$DB,$C8,$E0,$87,$F5,$07,$2D,$33
	.dc.b	$1F,$E0,$B8,$CB,$63,$38,$7F,$82,$27,$2D,$8C,$9F,$DD,$A3,$2F,$F7
	.dc.b	$8F,$53,$FC,$05,$E5,$FF,$80,$FD,$EF,$F0,$27,$C0,$7F,$81,$FF,$80
	.dc.b	$5E,$B5,$F7,$26,$3F,$EE,$AD,$3F,$DA,$B4,$F3,$F8,$7B,$9C,$7D,$09
	.dc.b	$7B,$DF,$FD,$C7,$7F,$E5,$08,$C1,$FE,$19,$7F,$D5,$8F,$F8,$21,$7A
	.dc.b	$BE,$3F,$73,$C9,$EB,$7A,$BB,$3F,$C3,$FC,$77,$C6,$34,$98,$FC,$29
	.dc.b	$DB,$67,$6D,$9D,$BF,$44,$9B,$7C,$F6,$E7,$F3,$5F,$F0,$3F,$27,$C7
	.dc.b	$F4,$7E,$54,$7B,$31,$6A,$F7,$7E,$67,$C1,$7C,$AA,$F7,$71,$FC,$FF
	.dc.b	$73,$E4,$C7,$1B,$71,$7C,$A8,$F8,$9F,$0D,$F5,$7E,$13,$EB,$72,$7C
	.dc.b	$59,$E1,$EF,$7A,$B8,$7D,$04,$E1,$F4,$36,$7F,$08,$7C,$12,$1E,$E3
	.dc.b	$25,$DF,$E4,$B7,$8F,$E4,$B7,$8F,$C1,$7F,$E5,$AE,$3F,$89,$3F,$E5
	.dc.b	$AE,$AF,$C6,$71,$FC,$28,$F7,$71,$7C,$3C,$FF,$17,$C5,$F0,$A7,$F0
	.dc.b	$5C,$3F,$06,$7F,$1F,$C3,$F0,$67,$E3,$70,$7C,$17,$F1,$1C,$1E,$27
	.dc.b	$F1,$9F,$0D,$F9,$B2,$3F,$35,$AF,$E3,$DB,$7F,$9C,$6E,$DF,$98,$6C
	.dc.b	$F4,$39,$10,$E4,$59,$E4,$D9,$E4,$FD,$DF,$7F,$F6,$2F,$DF,$F8,$6F
	.dc.b	$D8,$BF,$81,$EC,$FD,$97,$C0,$7E,$CF,$E1,$BF,$69,$DD,$FE,$7B,$EB
	.dc.b	$7C,$1D,$F8,$3C,$7C,$1F,$EE,$7E,$07,$5F,$05,$BF,$86,$E1,$F0,$7F
	.dc.b	$AA,$DB,$C8,$7A,$CF,$B4,$7E,$0B,$8B,$F0,$DD,$FF,$80,$6F,$7F,$E0
	.dc.b	$1B,$DF,$F3,$E7,$5C,$BF,$8E,$25,$FE,$2A,$B6,$FE,$C7,$BF,$E0,$BF
	.dc.b	$E7,$0E,$FF,$82,$FF,$01,$E5,$FC,$25,$FF,$08,$EE,$F8,$33,$FA,$5F
	.dc.b	$17,$F0,$93,$FC,$33,$BB,$E0,$4F,$EA,$7C,$1F,$D2,$BF,$AB,$F1,$7E
	.dc.b	$03,$EE,$26,$9E,$46,$DF,$A3,$5F,$5C,$7A,$63,$1B,$FC,$16,$EF,$F0
	.dc.b	$63,$13,$B2,$57,$23,$FD,$6B,$BF,$E9,$23,$E0,$1F,$BF,$F0,$BF,$48
	.dc.b	$FF,$04,$3F,$4A,$7F,$03,$F0,$BF,$49,$97,$F3,$0D,$7E,$0C,$FF,$29
	.dc.b	$F0,$BF,$48,$FF,$29,$F0,$BF,$49,$5F,$CD,$BC,$2F,$36,$B2,$BE,$F6
	.dc.b	$DE,$A6,$9E,$8D,$BD,$7B,$7A,$5A,$4A,$E5,$3F,$C8,$8F,$E3,$1D,$9F
	.dc.b	$C8,$9F,$E7,$5B,$7B,$32,$FE,$8F,$C7,$F7,$27,$FA,$06,$FE,$CC,$FE
	.dc.b	$D4,$7F,$21,$DF,$FA,$96,$FE,$CC,$3F,$A7,$EF,$EC,$BF,$E9,$7B,$FE
	.dc.b	$47,$C3,$78,$7F,$B6,$E1,$FE,$62,$D8,$CE,$7F,$A0,$9C,$B6,$32,$FE
	.dc.b	$E8,$98,$FE,$E9,$B8,$7E,$E9,$A7,$EA,$63,$44,$8D,$3F,$D9,$BC,$3F
	.dc.b	$CF,$BF,$DA,$3C,$3F,$CF,$BF,$DA,$BA,$3F,$7F,$7F,$D0,$7C,$2F,$CF
	.dc.b	$BF,$EE,$AE,$3F,$DF,$BF,$D3,$57,$81,$FC,$02,$3F,$A4,$77,$FF,$00
	.dc.b	$75,$DD,$FC,$01,$FE,$99,$E2,$FC,$3E,$0F,$E1,$5F,$0D,$27,$A3,$5F
	.dc.b	$03,$74,$6E,$97,$0B,$45,$EE,$B5,$39,$8C,$57,$0A,$DB,$E0,$21,$E0
	.dc.b	$2D,$F8,$1E,$13,$CA,$BA,$46,$AF,$BD,$BE,$1B,$F9,$37,$A3,$FC,$9F
	.dc.b	$DF,$FE,$51,$E9,$7F,$29,$BF,$E5,$5E,$D7,$F2,$BF,$7F,$F9,$67,$A7
	.dc.b	$FC,$B7,$DA,$FE,$5D,$ED,$7F,$2F,$F5,$3F,$98,$7A,$9F,$CC,$7E,$0F
	.dc.b	$F9,$97,$B5,$F0,$F8,$5C,$9E,$7E,$37,$F0,$0E,$67,$3B,$FC,$48,$C5
	.dc.b	$6F,$93,$E1,$D7,$C1,$36,$7C,$E9,$DD,$71,$8F,$05,$77,$9C,$C6,$CC
	.dc.b	$CE,$DF,$F0,$A4,$6A,$FC,$B1,$A3,$81,$A3,$0E,$9C,$10,$F2,$B9,$6C
	.dc.b	$75,$23,$C9,$AF,$1A,$F8,$1C,$99,$AD,$7B,$FB,$F8,$1F,$AD,$7E,$8F
	.dc.b	$80,$E5,$F0,$3C,$FE,$07,$A7,$C3,$7D,$6A,$FD,$27,$1F,$D6,$7D,$FF
	.dc.b	$0D,$E1,$F8,$3F,$14,$43,$C5,$C1,$F1,$5F,$49,$92,$9A,$3E,$29,$82
	.dc.b	$6F,$F4,$5E,$0D,$78,$6D,$F1,$63,$11,$C2,$7C,$11,$82,$38,$24,$7E
	.dc.b	$9F,$7F,$8C,$FD,$AD,$FE,$33,$FC,$0E,$7E,$76,$DF,$1A,$3E,$FE,$BF
	.dc.b	$19,$F8,$A4,$3F,$86,$72,$7D,$5D,$3E,$34,$7A,$38,$7E,$BB,$E7,$3F
	.dc.b	$62,$7E,$BC,$A7,$E5,$F3,$FC,$D7,$6F,$ED,$78,$FF,$37,$18,$9E,$1B
	.dc.b	$7E,$E1,$E1,$A7,$EE,$2F,$81,$23,$77,$CA,$75,$1F,$9A,$F8,$6F,$DE
	.dc.b	$BF,$C8,$FB,$7F,$7A,$FF,$25,$ED,$FD,$EB,$FC,$9F,$B3,$F7,$AF,$F2
	.dc.b	$9E,$BF,$CA,$7C,$27,$EF,$63,$F9,$67,$67,$EF,$5F,$E5,$DD,$7F,$BE
	.dc.b	$7F,$53,$F0,$DE,$57,$D2,$7C,$2E,$0F,$73,$8E,$0F,$74,$60,$9B,$7A
	.dc.b	$A7,$5B,$FB,$5E,$51,$99,$DC,$60,$95,$D2,$74,$48,$CD,$7F,$21,$7F
	.dc.b	$A3,$78,$1F,$C8,$5F,$88,$7E,$1C,$FE,$95,$E0,$7F,$21,$7F,$A6,$78
	.dc.b	$1F,$C8,$5F,$C3,$63,$FC,$57,$4F,$D6,$47,$F5,$3F,$03,$E2,$3F,$85
	.dc.b	$CF,$F3,$D9,$F8,$DF,$34,$7E,$A2,$FF,$3A,$D9,$FC,$EC,$F0,$8C,$11
	.dc.b	$C1,$39,$7F,$27,$5C,$A3,$13,$FB,$0A,$E7,$5A,$7A,$E9,$F0,$E7,$F8
	.dc.b	$AD,$7F,$0E,$7F,$91,$F8,$1F,$C3,$9F,$E3,$3D,$5F,$A9,$7F,$8D,$F5
	.dc.b	$7E,$A5,$FE,$E5,$DD,$FC,$3D,$FD,$49,$FC,$38,$FE,$5B,$DD,$E3,$7D
	.dc.b	$59,$F8,$DA,$67,$1F,$4B,$FA,$F3,$11,$B8,$CC,$7C,$D8,$C1,$1E,$59
	.dc.b	$7B,$74,$FE,$85,$5E,$D9,$FE,$67,$9F,$B6,$7F,$A3,$67,$ED,$7F,$A4
	.dc.b	$67,$ED,$7F,$A9,$67,$ED,$7F,$AA,$76,$7F,$05,$C3,$DA,$79,$C8,$7E
	.dc.b	$6B,$AF,$F4,$93,$F9,$8C,$BF,$02,$4C,$FF,$3C,$D9,$DE,$DF,$87,$9C
	.dc.b	$92,$33,$7F,$A0,$D5,$D7,$FC,$02,$3F,$C4,$9D,$7F,$C0,$1F,$F1,$37
	.dc.b	$5F,$F0,$07,$FC,$51,$D7,$FC,$01,$FF,$15,$75,$FF,$00,$7F,$7E,$43
	.dc.b	$F3,$87,$F0,$02,$1F,$D7,$3A,$7C,$7A,$FF,$7C,$F8,$4F,$EB,$84,$9F
	.dc.b	$ED,$AD,$7F,$B1,$B7,$7F,$EE,$CD,$DF,$FC,$04,$D9,$FE,$FC,$DA,$ED
	.dc.b	$1D,$08,$74,$2D,$74,$7F,$BE,$FC,$37,$F6,$FF,$85,$FE,$E1,$F0,$BF
	.dc.b	$DC,$7D,$6F,$EE,$5E,$C7,$F7,$3F,$85,$FE,$E9,$EE,$FF,$75,$F8,$5F
	.dc.b	$EF,$FF,$C0,$2F,$46,$FB,$58,$F8,$4C,$FC,$06,$BD,$BD,$FA,$7C,$3E
	.dc.b	$03,$8F,$B7,$E9,$EE,$73,$F4,$F8,$09,$CD,$FE,$49,$E7,$FF,$13,$6E
	.dc.b	$9C,$CF,$F8,$32,$30,$EF,$A7,$FC,$97,$18,$8D,$4F,$0A,$E6,$DA,$DB
	.dc.b	$67,$FC,$AF,$5C,$6B,$71,$D4,$CC,$EF,$A5,$5B,$BE,$75,$B9,$CD,$7F
	.dc.b	$87,$7A,$7E,$4F,$AB,$F4,$2B,$FC,$5F,$D1,$F4,$1E,$18,$FF,$99,$DF
	.dc.b	$F1,$9F,$37,$D0,$F7,$FE,$8F,$37,$D0,$9D,$F2,$FD,$07,$8F,$C3,$7D
	.dc.b	$6F,$80,$FA,$11,$D3,$E0,$3D,$19,$79,$E3,$D6,$78,$36,$C9,$1C,$10
	.dc.b	$F9,$65,$E1,$F8,$AE,$08,$E0,$9C,$3F,$12,$32,$D8,$CC,$7F,$96,$78
	.dc.b	$3E,$2B,$FE,$52,$ED,$FC,$A3,$FE,$54,$ED,$FC,$8F,$8D,$F9,$48,$F8
	.dc.b	$FC,$1F,$11,$F9,$3D,$9F,$93,$7E,$5F,$67,$E6,$38,$3E,$24,$7C,$EE
	.dc.b	$AF,$CD,$F0,$7B,$E3,$F5,$04,$D7,$F5,$0D,$DB,$F6,$6F,$EB,$1A,$E4
	.dc.b	$FD,$A9,$92,$99,$E7,$FD,$BB,$7A,$7F,$42,$EE,$33,$5E,$02,$57,$E7
	.dc.b	$3E,$03,$F7,$15,$F8,$6F,$6B,$F6,$7E,$FF,$ED,$3D,$FF,$DC,$4F,$E2
	.dc.b	$BE,$1B,$F7,$0F,$E3,$3E,$03,$FC,$EF,$FC,$03,$FE,$C6,$8F,$F3,$AB
	.dc.b	$FE,$C6,$7F,$30,$BE,$1B,$F7,$51,$F9,$4F,$0F,$8F,$C3,$7E,$EA,$3D
	.dc.b	$1C,$1E,$A7,$D9,$E1,$FE,$F4,$8E,$9A,$E8,$C1,$0C,$1F,$17,$AF,$DE
	.dc.b	$FF,$A2,$3A,$FF,$06,$FF,$A2,$BA,$FD,$EF,$FA,$33,$C1,$F7,$BF,$C1
	.dc.b	$7C,$1F,$D3,$3F,$C1,$FC,$1F,$D3,$3F,$C2,$7C,$5F,$E2,$AF,$E0,$CF
	.dc.b	$D8,$F8,$BE,$6E,$8F,$CA,$F4,$FE,$EA,$7F,$41,$D3,$F9,$92,$3F,$AA
	.dc.b	$3F,$3A,$59,$E3,$E1,$7E,$BE,$B9,$4E,$9B,$3A,$F0,$BF,$24,$D9,$FE
	.dc.b	$6E,$7F,$04,$F1,$7F,$84,$3F,$EA,$AF,$17,$F8,$43,$FC,$8F,$C4,$FD
	.dc.b	$23,$FC,$97,$C4,$FD,$23,$FC,$9F,$C4,$FD,$23,$FC,$A7,$C4,$FD,$23
	.dc.b	$FC,$43,$C5,$FE,$0E,$FF,$2D,$F0,$FC,$7B,$FB,$B6,$F3,$B7,$7F,$61
	.dc.b	$37,$DD,$FC,$19,$D4,$62,$B8,$CE,$48,$E0,$95,$F1,$FA,$5D,$1F,$C1
	.dc.b	$2B,$F0,$47,$F2,$83,$F8,$21,$FC,$A3,$6F,$83,$C3,$FD,$23,$7F,$82
	.dc.b	$FF,$4A,$DF,$E0,$BF,$D3,$37,$F8,$37,$FD,$2F,$7F,$83,$3F,$D4,$37
	.dc.b	$F4,$67,$F9,$D8,$FD,$9D,$7E,$F1,$BA,$FE,$E3,$7F,$E0,$71,$8A,$E1
	.dc.b	$18,$25,$71,$6D,$71,$8C,$11,$C1,$27,$9C,$67,$67,$87,$FA,$6C,$3F
	.dc.b	$B3,$F8,$7F,$C1,$2B,$FA,$5F,$85,$FA,$69,$FE,$D7,$E1,$7E,$99,$FE
	.dc.b	$DB,$E1,$7F,$04,$F8,$7F,$E0,$BF,$09,$FC,$1B,$C3,$FE,$03,$8F,$F6
	.dc.b	$FF,$0F,$F8,$0B,$FA,$63,$F7,$18,$38,$10,$E0,$78,$8F,$84,$FE,$22
	.dc.b	$49,$FE,$1E,$D7,$F8,$9B,$6F,$A4,$73,$9E,$4F,$02,$1C,$0B,$1C,$2B
	.dc.b	$C3,$5C,$EF,$CB,$18,$AD,$C6,$2F,$AC,$73,$25,$F3,$87,$C3,$FF,$2A
	.dc.b	$F4,$FF,$95,$FC,$1F,$F2,$CF,$7F,$F9,$6F,$A9,$FC,$BB,$D4,$FE,$5F
	.dc.b	$9F,$F3,$0F,$7F,$F9,$8F,$AB,$FC,$CB,$DF,$FE,$67,$EB,$7F,$34,$F6
	.dc.b	$7F,$9A,$FB,$FF,$CD,$BD,$AF,$E6,$FE,$FF,$F3,$8F,$80,$F5,$72,$7C
	.dc.b	$3E,$6B,$CD,$EB,$8F,$3E,$DE,$B9,$CF,$2F,$B3,$9B,$D5,$18,$25,$7C
	.dc.b	$23,$A1,$6F,$A4,$F0,$8C,$57,$3B,$ED,$1F,$E1,$FF,$81,$FC,$0B,$FD
	.dc.b	$2B,$9A,$DF,$D3,$9F,$C2,$4E,$FE,$0F,$37,$F5,$25,$F0,$38,$AF,$C5
	.dc.b	$43,$E0,$7E,$5C,$77,$F8,$5D,$68,$A7,$C1,$61,$7C,$2D,$F8,$5C,$7C
	.dc.b	$2F,$CE,$C7,$D1,$F0,$BB,$BE,$1F,$0B,$F1,$27,$E5,$9F,$10,$B9,$7C
	.dc.b	$4D,$BE,$0C,$63,$8B,$E2,$3B,$E1,$F9,$0E,$78,$FA,$46,$09,$5C,$9B
	.dc.b	$3C,$63,$07,$C5,$7F,$20,$7C,$FB,$F7,$71,$7B,$F8,$BE,$07,$17,$C6
	.dc.b	$A6,$1F,$47,$8B,$E2,$71,$7C,$5E,$2F,$89,$5F,$67,$87,$E2,$3C,$34
	.dc.b	$FA,$9C,$3F,$16,$3E,$F7,$07,$AA,$7E,$C7,$BD,$F3,$C9,$3F,$83,$6B
	.dc.b	$CC,$FB,$6E,$33,$FC,$23,$67,$9C,$F1,$21,$C7,$7F,$BD,$3B,$8E,$66
	.dc.b	$AB,$9B,$6B,$9B,$FE,$1E,$F0,$BF,$C4,$27,$EE,$1F,$F0,$FF,$83,$FB
	.dc.b	$87,$FC,$43,$E0,$7E,$27,$D4,$FD,$C4,$7F,$89,$3C,$0F,$DC,$3F,$E2
	.dc.b	$6E,$FF,$DC,$3F,$A1,$F8,$5F,$DC,$3F,$E2,$AE,$DF,$DC,$3E,$33,$CE
	.dc.b	$7E,$0C,$FD,$5F,$47,$CB,$7F,$00,$4D,$FF,$1C,$DE,$9F,$C6,$C6,$08
	.dc.b	$E0,$93,$CE,$33,$AF,$E2,$2B,$FC,$7F,$E2,$7F,$17,$7F,$C7,$FE,$27
	.dc.b	$F1,$77,$FC,$83,$C1,$F0,$1F,$F2,$2E,$BF,$01,$FF,$23,$EB,$F0,$1F
	.dc.b	$F2,$4F,$83,$F0,$1F,$F2,$5E,$BF,$01,$F7,$13,$8B,$F3,$BA,$7A,$35
	.dc.b	$F3,$6D,$E8,$AF,$86,$43,$86,$7F,$A2,$7E,$14,$FF,$20,$ED,$FE,$1F
	.dc.b	$1F,$E5,$AE,$FF,$E0,$18,$7E,$92,$3F,$8A,$76,$FE,$91,$F6,$9F,$C2
	.dc.b	$4F,$E3,$1D,$7F,$A3,$7F,$8D,$77,$FF,$06,$7F,$CC,$7B,$7F,$0B,$EA
	.dc.b	$FD,$14,$7E,$4F,$27,$57,$AB,$49,$3E,$9D,$FD,$FB,$FB,$63,$D6,$7A
	.dc.b	$8D,$6F,$ED,$DB,$E2,$47,$F3,$1C,$7D,$B7,$FD,$47,$7F,$6C,$FF,$29
	.dc.b	$3D,$A7,$F3,$1E,$EF,$E0,$13,$FB,$F8,$F6,$9F,$CD,$7B,$7F,$7E,$FA
	.dc.b	$7B,$7F,$7F,$3F,$CD,$76,$F6,$BF,$CD,$F6,$FC,$56,$5F,$E1,$0C,$FF
	.dc.b	$63,$87,$E6,$DA,$FC,$3E,$DF,$DA,$46,$2F,$F6,$73,$93,$85,$73,$25
	.dc.b	$70,$6C,$6A,$39,$ED,$FD,$C2,$3F,$D2,$BD,$3F,$B9,$7F,$D3,$3D,$BF
	.dc.b	$B9,$7F,$D3,$7D,$BF,$B9,$7F,$D3,$7E,$27,$F7,$07,$FD,$3B,$E2,$7F
	.dc.b	$70,$7F,$D3,$FE,$27,$F4,$7E,$FF,$CD,$6B,$FD,$FF,$AF,$F3,$4F,$A8
	.dc.b	$9C,$DF,$5F,$5F,$B9,$1F,$7B,$9F,$F0,$38,$FD,$EE,$1F,$B6,$DA,$DD
	.dc.b	$F1,$AF,$EE,$BB,$FE,$D7,$0F,$F0,$77,$7F,$ED,$70,$FF,$5A,$F7,$FE
	.dc.b	$D5,$FF,$5B,$E1,$FB,$5A,$FF,$0A,$77,$FE,$D5,$FF,$0B,$77,$FE,$D5
	.dc.b	$FF,$0C,$76,$7B,$2B,$FC,$3D,$B7,$B1,$F1,$61,$F9,$23,$F2,$87,$9F
	.dc.b	$4F,$5D,$FB,$63,$5B,$7B,$23,$11,$FD,$93,$9B,$F2,$F7,$FE,$2B,$EC
	.dc.b	$FC,$BC,$FE,$9F,$C2,$FE,$AC,$FE,$5C,$FD,$99,$FA,$BF,$0B,$FA,$AB
	.dc.b	$F9,$73,$F6,$67,$F5,$53,$F6,$67,$F7,$83,$FC,$21,$7F,$DE,$1F,$2F
	.dc.b	$17,$CE,$7D,$78,$F9,$89,$8F,$C0,$8F,$71,$C3,$3F,$7C,$7F,$6F,$8F
	.dc.b	$F0,$D1,$FD,$BC,$FD,$81,$FC,$07,$C1,$FE,$18,$FD,$43,$F8,$61,$FC
	.dc.b	$17,$C1,$FE,$18,$FF,$07,$F0,$7F,$86,$3F,$CA,$0F,$51,$FC,$2F,$2F
	.dc.b	$53,$F8,$97,$F4,$6F,$EA,$49,$8F,$E4,$9F,$E9,$6D,$DB,$F6,$18,$FF
	.dc.b	$60,$8C,$57,$26,$F6,$7E,$BC,$97,$FB,$BD,$3F,$5C,$FF,$67,$3F,$CF
	.dc.b	$73,$F9,$47,$FB,$39,$FD,$20,$FE,$CE,$7F,$90,$7C,$2F,$CA,$61,$FE
	.dc.b	$6B,$E4,$FC,$A3,$FE,$6D,$E4,$FC,$A3,$FE,$3F,$E8,$FC,$A5,$FF,$9F
	.dc.b	$79,$3F,$28,$FB,$27,$D4,$7F,$91,$7D,$FF,$F1,$E1,$23,$FC,$A0,$D7
	.dc.b	$F9,$A3,$72,$FF,$28,$36,$79,$4F,$02,$1C,$0B,$3C,$1B,$1C,$1D,$12
	.dc.b	$34,$D8,$D4,$60,$8E,$09,$1F,$E5,$C9,$F2,$34,$EE,$53,$A1,$67,$9B
	.dc.b	$67,$9E,$DF,$E6,$1A,$FF,$30,$7B,$12,$BF,$CE,$8D,$7F,$AC,$36,$FF
	.dc.b	$CE,$8D,$BF,$F3,$93,$67,$FC,$E2,$DB,$FF,$37,$37,$73,$49,$3F,$E8
	.dc.b	$3F,$7F,$FC,$E5,$F0,$BF,$E7,$3F,$80,$FF,$3A,$7C,$2F,$F9,$E3,$F6
	.dc.b	$A7,$C1,$5F,$6F,$1F,$05,$9F,$73,$5F,$05,$BF,$47,$87,$C2,$F1,$F8
	.dc.b	$3E,$5F,$05,$CF,$E0,$FA,$7C,$1F,$5F,$83,$F1,$F1,$FC,$03,$E1,$90
	.dc.b	$C1,$3A,$3E,$1B,$96,$B9,$E8,$F8,$AE,$5A,$E4,$F8,$E6,$4A,$67,$A3
	.dc.b	$E4,$B9,$7F,$D3,$1C,$DF,$1A,$3F,$D5,$5C,$BF,$19,$FD,$8D,$7F,$B6
	.dc.b	$39,$3E,$3C,$60,$FF,$77,$2E,$3F,$93,$19,$C1,$2E,$2F,$97,$39,$87
	.dc.b	$0F,$CC,$8C,$70,$7C,$C7,$5B,$FC,$C7,$86,$DF,$33,$6F,$C9,$E5,$F9
	.dc.b	$19,$FC,$CB,$5F,$CC,$B6,$37,$D0,$E9,$FD,$0C,$E7,$6F,$CD,$CE,$6F
	.dc.b	$9B,$5E,$6D,$BE,$77,$CA,$31,$5C,$A3,$04,$AE,$6D,$79,$B6,$B9,$C6
	.dc.b	$88,$68,$AE,$9C,$11,$C1,$27,$C5,$1C,$1A,$ED,$B1,$B8,$FD,$E1,$2F
	.dc.b	$E2,$36,$FD,$AD,$BF,$45,$F2,$24,$64,$F0,$14,$F0,$16,$FC,$17,$D8
	.dc.b	$97,$D9,$B7,$DA,$B8,$5F,$69,$CC,$76,$6B,$D9,$B7,$DA,$FA,$12,$FA
	.dc.b	$4E,$AF,$A4,$E6,$FA,$35,$E8,$DB,$E9,$7C,$A3,$15,$CA,$30,$76,$D1
	.dc.b	$F0,$DF,$C4,$49,$87,$E8,$DA,$FF,$B3,$E3,$6E,$48,$66,$37,$82,$FF
	.dc.b	$6D,$18,$6C,$F0,$9D,$3F,$B8,$9D,$34,$D4,$F0,$7C,$5F,$09,$FC,$9F
	.dc.b	$E1,$3F,$94,$7C,$17,$F2,$9F,$82,$FE,$55,$F0,$DF,$CA,$FE,$0F,$F9
	.dc.b	$67,$C1,$FF,$2D,$F8,$36,$5F,$C0,$FF,$80,$5F,$09,$88,$FF,$76,$43
	.dc.b	$E1,$31,$C1,$E6,$E4,$4B,$E1,$33,$C3,$D2,$73,$F0,$F8,$C5,$1C,$3D
	.dc.b	$2C,$3C,$BD,$9B,$8B,$D6,$BE,$CF,$C2,$AF,$E2,$FC,$FA,$F7,$A6,$5E
	.dc.b	$96,$BE,$86,$ED,$E7,$C5,$18,$D7,$CD,$18,$AE,$44,$39,$16,$B9,$57
	.dc.b	$18,$F5,$CF,$17,$64,$9D,$BF,$C8,$21,$ED,$7B,$BD,$E9,$3F,$47,$0F
	.dc.b	$83,$EF,$7C,$2F,$4F,$E1,$FB,$D2,$BE,$B7,$A9,$F1,$BD,$EF,$8F,$EF
	.dc.b	$7C,$8F,$85,$F9,$3E,$B7,$CA,$F8,$5F,$9B,$FC,$03,$FE,$70,$C7,$CF
	.dc.b	$C5,$E9,$E4,$4F,$D4,$9F,$E8,$8D,$5C,$7F,$7E,$30,$D9,$EA,$49,$E2
	.dc.b	$D7,$8F,$07,$E0,$EB,$8B,$67,$55,$CE,$B8,$90,$E2,$7E,$22,$B8,$B6
	.dc.b	$78,$D7,$38,$E1,$5F,$D4,$3C,$57,$89,$F0,$BE,$07,$E1,$FC,$07,$E3
	.dc.b	$B0,$F9,$F8,$7E,$87,$8F,$E3,$FC,$37,$C8,$F8,$2F,$C9,$F2,$7C,$AF
	.dc.b	$82,$F9,$7F,$0D,$F3,$F0,$FD,$47,$1F,$CF,$7E,$4F,$C1,$F9,$1F,$31
	.dc.b	$E4,$CF,$CC,$FB,$38,$7E,$89,$B2,$BB,$38,$9B,$E2,$F8,$31,$A8,$C3
	.dc.b	$C6,$32,$47,$04,$8E,$8F,$F3,$BE,$1F,$8B,$1F,$80,$E0,$F8,$AF,$E4
	.dc.b	$CF,$E0,$9D,$3F,$93,$7F,$0B,$BF,$C5,$7F,$26,$7E,$23,$5F,$8A,$FF
	.dc.b	$0B,$E6,$FC,$9B,$FC,$37,$9B,$F2,$7A,$F8,$E3,$F4,$DA,$FE,$CB,$9B
	.dc.b	$F2,$9D,$9F,$AB,$E7,$FD,$4D,$68,$FC,$C9,$FB,$3A,$E0,$D3,$26,$F7
	.dc.b	$FE,$2B,$7D,$09,$1B,$8C,$BF,$E1,$FF,$33,$F8,$8F,$6F,$ED,$23,$F9
	.dc.b	$0F,$6F,$EA,$BE,$1B,$F5,$7F,$0D,$FA,$CF,$43,$F6,$95,$FC,$9B,$C0
	.dc.b	$FD,$7F,$C3,$7E,$D6,$3F,$96,$78,$1F,$B2,$F8,$6F,$D9,$FC,$2F,$ED
	.dc.b	$3E,$17,$F6,$B5,$E4,$3D,$1C,$7F,$07,$8B,$DB,$18,$E2,$D7,$1F,$C4
	.dc.b	$24,$E9,$B3,$F0,$9B,$C7,$EE,$8E,$47,$C1,$BE,$CE,$EB,$FD,$0F,$1F
	.dc.b	$C5,$A3,$FA,$27,$0F,$C3,$BF,$E5,$55,$FA,$89,$FE,$8B,$E0,$FC,$29
	.dc.b	$FE,$5D,$BF,$EA,$27,$FA,$67,$89,$FC,$59,$FC,$21,$FA,$83,$F0,$87
	.dc.b	$F1,$03,$DB,$7F,$A6,$AF,$E0,$91,$F9,$E3,$F3,$3D,$DF,$9D,$6E,$7F
	.dc.b	$8B,$9D,$92,$B9,$12,$7F,$57,$19,$6C,$66,$7F,$59,$3F,$C2,$7C,$4F
	.dc.b	$E1,$15,$FC,$93,$C4,$FE,$10,$FE,$90,$FE,$10,$7E,$90,$FE,$53,$E2
	.dc.b	$7C,$17,$F7,$1D,$BF,$A4,$7F,$B5,$F6,$FE,$91,$FE,$DB,$DB,$FA,$47
	.dc.b	$D6,$43,$E0,$13,$1F,$46,$5E,$6C,$FD,$1A,$7C,$3A,$D4,$62,$7E,$23
	.dc.b	$63,$A5,$72,$E8,$FF,$1D,$78,$3F,$C0,$A7,$F8,$77,$83,$FC,$09,$FF
	.dc.b	$00,$DF,$DF,$BF,$ED,$5C,$5F,$C9,$A7,$DE,$7F,$02,$3D,$E7,$F4,$AF
	.dc.b	$0F,$F9,$33,$FE,$0A,$F0,$FF,$93,$3F,$C6,$FC,$1F,$2B,$F9,$12,$69
	.dc.b	$F9,$7C,$3F,$33,$CF,$F9,$68,$C6,$5F,$9F,$25,$70,$9D,$13,$A3,$F1
	.dc.b	$71,$96,$D7,$28,$FF,$52,$F8,$3F,$C0,$1F,$CF,$90,$FE,$61,$E0,$7F
	.dc.b	$00,$7F,$B2,$F0,$7E,$7F,$1F,$F0,$4F,$4F,$F0,$09,$FC,$F9,$FC,$00
	.dc.b	$FE,$D7,$DB,$FC,$01,$FE,$DB,$D9,$FC,$01,$FD,$96,$7F,$E2,$55,$F6
	.dc.b	$5F,$E0,$B7,$FC,$64,$62,$2B,$FE,$35,$8D,$36,$32,$FF,$8F,$23,$2D
	.dc.b	$8C,$BC,$CD,$C6,$9C,$1D,$9C,$12,$38,$B6,$38,$C6,$08,$E0,$87,$37
	.dc.b	$2D,$32,$7B,$A3,$A3,$92,$19,$8E,$2D,$38,$9F,$04,$A7,$BC,$A7,$B0
	.dc.b	$E0,$74,$38,$14,$E0,$53,$81,$A3,$89,$0C,$10,$E2,$64,$A6,$4E,$66
	.dc.b	$08,$60,$91,$F0,$9E,$46,$23,$E0,$BC,$E3,$4D,$8C,$BD,$23,$2D,$8C
	.dc.b	$BC,$CF,$8B,$1A,$70,$7C,$17,$04,$8E,$2D,$8E,$31,$82,$38,$21,$CD
	.dc.b	$CB,$4C,$9E,$E8,$E8,$E4,$86,$63,$8B,$4E,$27,$C1,$29,$EF,$29,$EC
	.dc.b	$38,$1D,$0E,$05,$38,$14,$FD,$9F,$85,$FA,$02,$3F,$A8,$6B,$FA,$76
	.dc.b	$BF,$9B,$6B,$FB,$76,$BF,$B5,$6B,$D8,$E8,$43,$A1,$4E,$85,$3A,$14
	.dc.b	$E8,$53,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05
	.dc.b	$38,$1A,$21,$A2,$9A,$30,$43,$04,$9F,$F0,$A3,$E4,$6B,$DD,$AF,$46
	.dc.b	$BC,$DA,$F3,$8F,$A3,$18,$23,$82,$47,$CB,$8C,$B6,$32,$F8,$DA,$F8
	.dc.b	$DA,$F8,$DA,$F8,$DA,$F8,$CE,$84,$3A,$14,$E8,$53,$A1,$4E,$85,$3A
	.dc.b	$1C,$88,$72,$29,$C8,$A7,$22,$9C,$8E,$04,$38,$14,$E0,$53,$80,$FC
	.dc.b	$A2,$00

;--------------------------------------------------------------------------------
;	FMOVE.P <mem>,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmoveptoreg_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+FPSP040+FPSP060,-(sp)
	peamsg	'FMOVE.P <EA>,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.P <ea>,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fmoveptoreg_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a0,a2			;source handle,...
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;<mem>,FPn
	fmove.l	#0,fpcr
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fmove.p	([a2]),fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)		;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.P <mem>='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmoveptoreg_data_compressed::
	.dc.b	$00,$01,$49,$54,$10,$99,$BD,$47,$CD,$6A,$31,$FA,$F5,$F6,$1F,$81
	.dc.b	$E7,$DB,$79,$9B,$F7,$1F,$93,$F1,$F7,$F3,$AB,$37,$C0,$3F,$10,$ED
	.dc.b	$F0,$4F,$DD,$BC,$1F,$0A,$FA,$FF,$17,$C3,$B8,$BC,$9F,$12,$FA,$4B
	.dc.b	$F7,$7E,$7F,$8C,$E5,$BF,$7F,$8D,$7A,$1E,$B5,$F1,$9E,$DF,$8E,$7E
	.dc.b	$C7,$EF,$F9,$17,$D0,$7C,$1F,$93,$CE,$5E,$9F,$95,$7E,$17,$F1,$7E
	.dc.b	$61,$F1,$17,$8F,$E4,$7C,$D7,$C6,$C3,$E0,$7C,$E7,$C2,$ED,$7C,$EF
	.dc.b	$C1,$9F,$89,$F3,$DE,$D7,$E3,$7D,$03,$E4,$7E,$97,$D0,$BF,$31,$BF
	.dc.b	$A4,$FD,$5E,$37,$E5,$FF,$5F,$E9,$9F,$13,$7C,$4F,$B5,$F4,$EF,$D1
	.dc.b	$BE,$E7,$D4,$BE,$CF,$EF,$7D,$5B,$F5,$DF,$C0,$78,$DF,$82,$E3,$FB
	.dc.b	$1F,$5D,$97,$ED,$7D,$87,$C0,$7E,$E7,$D8,$FC,$47,$EF,$7D,$9B,$DE
	.dc.b	$FC,$57,$DA,$3F,$1B,$FC,$67,$DB,$3E,$33,$F1,$DF,$6E,$CB,$F2,$1F
	.dc.b	$70,$FC,$E3,$F2,$FF,$18,$FC,$57,$D5,$B2,$31,$FA,$96,$A7,$CD,$1B
	.dc.b	$FD,$9F,$1F,$B9,$E7,$F7,$EF,$CE,$7D,$FF,$E8,$69,$F8,$D3,$E4,$7B
	.dc.b	$AF,$9A,$BF,$49,$8F,$27,$3F,$6B,$AF,$B7,$DF,$9B,$C3,$BB,$8F,$C1
	.dc.b	$72,$F8,$0E,$7F,$19,$D3,$E1,$BF,$59,$F8,$0F,$5E,$7F,$56,$EF,$CA
	.dc.b	$3C,$3F,$17,$E2,$EA,$F1,$FE,$91,$E4,$FD,$4B,$CB,$FA,$EF,$ED,$7F
	.dc.b	$06,$7B,$9F,$B7,$FC,$1F,$E2,$FB,$8F,$C2,$7E,$76,$DF,$B1,$FC,$29
	.dc.b	$F0,$9F,$BD,$FC,$2F,$F2,$D6,$FC,$31,$EE,$FF,$01,$FC,$3B,$E2,$63
	.dc.b	$FB,$9E,$87,$C8,$7F,$05,$FC,$3D,$F8,$97,$F0,$7F,$C4,$7F,$80,$DF
	.dc.b	$EF,$FF,$12,$7E,$51,$FC,$2E,$9F,$97,$FF,$0D,$FC,$57,$7B,$BF,$16
	.dc.b	$7E,$67,$FC,$47,$F1,$8F,$CE,$E7,$FC,$27,$F1,$9F,$8C,$FF,$0B,$FC
	.dc.b	$6B,$AB,$FC,$37,$F1,$CF,$D4,$3F,$C3,$FC,$6F,$D1,$BF,$C4,$7F,$1D
	.dc.b	$FC,$43,$7E,$5B,$F8,$F7,$E3,$FF,$1B,$57,$E4,$1F,$03,$D4,$FC,$2F
	.dc.b	$CD,$FE,$13,$F7,$1F,$C2,$FE,$F7,$F9,$4F,$C9,$3F,$3D,$9F,$F1,$DC
	.dc.b	$3F,$30,$FE,$56,$3E,$8E,$07,$C6,$DF,$9B,$E5,$F5,$9C,$FB,$7A,$78
	.dc.b	$FD,$7E,$B7,$B7,$C7,$FF,$33,$FC,$91,$E3,$78,$4F,$17,$47,$8F,$D2
	.dc.b	$F2,$78,$1F,$CE,$7E,$0B,$F8,$DF,$07,$F3,$17,$F5,$CF,$F3,$3E,$0F
	.dc.b	$9E,$FF,$35,$FC,$97,$7C,$CF,$C9,$9E,$F7,$F4,$37,$F9,$1C,$7F,$39
	.dc.b	$CB,$E1,$78,$1F,$0F,$E6,$FE,$27,$E9,$3F,$17,$CC,$FE,$95,$F9,$47
	.dc.b	$F3,$0D,$3E,$27,$EB,$9F,$27,$E8,$FE,$57,$ED,$FF,$2F,$E7,$3F,$7F
	.dc.b	$C1,$FB,$53,$EB,$F6,$9F,$E7,$78,$7F,$01,$7F,$96,$62,$F8,$AE,$3F
	.dc.b	$8C,$F2,$FA,$AE,$7E,$B7,$4F,$A2,$FE,$B9,$F9,$5F,$E8,$3F,$2C,$FB
	.dc.b	$3C,$C7,$F8,$1E,$9F,$D5,$FE,$5F,$F0,$F2,$FC,$B7,$F9,$C2,$FC,$BB
	.dc.b	$F1,$1F,$97,$FC,$C8,$2F,$94,$FE,$D5,$E0,$3E,$37,$DB,$F2,$FE,$EF
	.dc.b	$83,$F7,$FB,$7F,$03,$D3,$FE,$E1,$F9,$83,$BF,$E1,$EF,$F1,$3D,$CF
	.dc.b	$8B,$F5,$1F,$1B,$F4,$8F,$8F,$F4,$DF,$23,$E6,$BE,$4F,$D2,$7C,$AE
	.dc.b	$1F,$EF,$5F,$99,$7F,$72,$F0,$7E,$57,$7F,$CE,$F9,$CF,$E5,$BE,$5F
	.dc.b	$E7,$CF,$E2,$9D,$0B,$FE,$71,$D7,$FD,$E5,$57,$F8,$17,$E4,$BF,$88
	.dc.b	$CF,$F7,$D7,$E4,$D7,$F1,$BF,$27,$FC,$15,$F9,$A7,$CA,$3F,$88,$D7
	.dc.b	$F8,$0F,$F3,$5F,$EC,$DA,$7E,$3F,$CF,$F1,$FC,$DF,$93,$E6,$FC,$BF
	.dc.b	$37,$FE,$13,$B7,$CA,$FA,$3C,$0F,$4F,$CD,$FA,$BE,$6F,$D7,$F2,$1E
	.dc.b	$CF,$88,$F6,$FC,$DF,$BB,$8F,$DF,$F3,$5C,$FE,$1B,$E0,$FC,$D7,$C2
	.dc.b	$D3,$E1,$FC,$D7,$C4,$F9,$AF,$F1,$17,$E6,$DF,$D7,$FA,$4F,$F7,$3E
	.dc.b	$7F,$F0,$E7,$E6,$EA,$FF,$14,$2C,$BF,$C5,$5F,$9C,$7F,$12,$C0,$BE
	.dc.b	$2B,$FC,$5D,$F9,$CB,$C4,$FF,$18,$7C,$0B,$AB,$FC,$65,$F9,$D7,$ED
	.dc.b	$B1,$7E,$79,$F5,$B8,$FF,$90,$C3,$F3,$EF,$E6,$ED,$E3,$FF,$1D,$79
	.dc.b	$7F,$A5,$E2,$FF,$19,$7E,$83,$FD,$F6,$9F,$A1,$78,$67,$FC,$6D,$FA
	.dc.b	$1F,$DC,$97,$E8,$9F,$D3,$38,$7F,$A6,$FE,$8D,$FC,$D2,$5F,$A4,$7F
	.dc.b	$29,$F4,$7F,$C0,$50,$FD,$2B,$F5,$BF,$A5,$FF,$46,$CF,$FC,$85,$EB
	.dc.b	$7F,$36,$A7,$E9,$BF,$A9,$FD,$39,$E0,$7F,$94,$7F,$4F,$FE,$11,$4F
	.dc.b	$D4,$3F,$03,$8F,$F9,$2F,$D8,$7C,$4F,$F9,$5F,$83,$FA,$6C,$FF,$93
	.dc.b	$FF,$54,$FE,$0F,$87,$F5,$CF,$27,$FC,$B9,$FA,$AF,$F9,$1A,$1F,$AB
	.dc.b	$BC,$CF,$F3,$06,$CF,$90,$FF,$31,$7E,$B1,$D2,$DF,$AD,$3E,$53,$FC
	.dc.b	$CD,$FA,$DF,$D7,$DB,$FC,$BD,$FA,$E7,$E1,$6F,$5E,$C7,$F9,$AF,$F5
	.dc.b	$EF,$B5,$C3,$FC,$33,$D0,$FD,$3F,$EC,$3F,$56,$DF,$B1,$FF,$8D,$72
	.dc.b	$FF,$34,$7C,$2F,$F0,$B7,$FC,$D5,$FB,$2F,$D9,$FD,$9B,$F8,$AF,$EC
	.dc.b	$FF,$DD,$27,$FC,$DD,$FB,$47,$D5,$FA,$CF,$C8,$7E,$D5,$E9,$FD,$B3
	.dc.b	$FA,$F6,$CF,$A5,$FF,$3F,$7E,$DB,$F9,$7C,$7F,$CE,$DF,$B7,$FF,$1B
	.dc.b	$EA,$FC,$E4,$7F,$9E,$7F,$71,$FE,$8B,$FB,$97,$D6,$8F,$F3,$E7,$D9
	.dc.b	$73,$7F,$CF,$DF,$BA,$7F,$62,$E5,$FE,$69,$F8,$F7,$FA,$2F,$F7,$6F
	.dc.b	$E5,$3F,$BC,$7F,$2C,$FA,$9F,$F0,$8E,$CC,$BF,$D2,$1F,$BD,$FF,$A0
	.dc.b	$73,$FF,$44,$7C,$2F,$CF,$7F,$D1,$5F,$BE,$BE,$43,$FD,$2D,$FB,$F7
	.dc.b	$F8,$62,$7F,$D1,$DF,$BF,$FD,$3D,$CF,$58,$7E,$6B,$FE,$A2,$F9,$4F
	.dc.b	$F5,$1A,$FC,$C3,$FA,$0A,$FC,$C7,$F8,$FD,$70,$FC,$5B,$DD,$2F,$CC
	.dc.b	$FE,$C6,$FF,$E9,$3A,$EB,$F8,$8F,$F9,$9E,$BB,$FD,$F7,$6F,$CD,$9F
	.dc.b	$03,$E0,$FC,$DF,$F8,$16,$1E,$4F,$CE,$3F,$C3,$F8,$3F,$39,$E8,$BF
	.dc.b	$3A,$FE,$66,$BF,$3B,$FB,$0F,$F9,$5F,$EB,$BD,$7F,$9E,$FB,$E1,$F9
	.dc.b	$FF,$F8,$01,$FF,$2F,$7C,$C7,$C0,$FD,$07,$F9,$9C,$3F,$42,$FF,$0C
	.dc.b	$AF,$D0,$FF,$D7,$2B,$F4,$4F,$F1,$EB,$FE,$04,$F1,$BE,$4F,$5F,$F0
	.dc.b	$98,$7E,$91,$FE,$CD,$7F,$D8,$BF,$4B,$F3,$BF,$49,$FF,$4F,$43,$F4
	.dc.b	$BF,$6F,$E1,$DF,$76,$1A,$FF,$81,$D7,$E9,$DF,$D8,$F9,$FE,$CF,$E9
	.dc.b	$FF,$E1,$47,$ED,$FE,$A1,$FE,$42,$AF,$E0,$7F,$91,$FD,$CF,$D4,$BE
	.dc.b	$DF,$E8,$BC,$FE,$A3,$FC,$53,$87,$E0,$BF,$54,$FA,$DF,$A2,$7F,$5A
	.dc.b	$C7,$F9,$87,$47,$E1,$3F,$43,$FD,$64,$78,$BF,$0F,$FC,$3F,$E8,$5F
	.dc.b	$EA,$18,$3F,$B7,$6F,$E6,$DF,$8C,$F1,$F1,$FA,$0F,$EE,$65,$FA,$07
	.dc.b	$E2,$BF,$40,$FF,$36,$C1,$FD,$63,$FE,$76,$EC,$FC,$AF,$EA,$FF,$E3
	.dc.b	$68,$7E,$B1,$FE,$5E,$7F,$D0,$3F,$7F,$F9,$AF,$D0,$3F,$C1,$DC,$7F
	.dc.b	$9E,$7F,$4C,$FE,$83,$F4,$0F,$C7,$D6,$A3,$F4,$7F,$AD,$7F,$B4,$E0
	.dc.b	$F8,$BF,$41,$FF,$34,$C1,$FD,$29,$FE,$EE,$5F,$A0,$FE,$34,$FE,$9C
	.dc.b	$FF,$BA,$BE,$87,$F5,$BF,$AD,$F8,$0E,$1F,$A1,$7E,$8B,$1F,$D9,$FE
	.dc.b	$85,$E1,$FC,$97,$F5,$BF,$A1,$7F,$47,$C5,$FA,$17,$F3,$5B,$FD,$CF
	.dc.b	$E0,$DF,$D0,$1F,$DD,$FE,$85,$F0,$29,$FA,$EF,$EE,$BF,$43,$FA,$1F
	.dc.b	$AF,$7D,$DF,$D1,$3F,$DA,$D4,$FD,$13,$F8,$EA,$FD,$13,$F4,$DF,$A1
	.dc.b	$FF,$AD,$61,$E9,$FE,$A7,$F4,$2F,$D2,$FE,$83,$FE,$8A,$93,$FA,$AF
	.dc.b	$D0,$71,$EF,$7E,$C7,$F4,$0F,$D4,$FE,$C1,$FE,$98,$BF,$E9,$DF,$A6
	.dc.b	$7F,$10,$FD,$0B,$F8,$C4,$3F,$40,$FD,$7F,$E4,$9F,$C8,$61,$FA,$07
	.dc.b	$CF,$FC,$FF,$1F,$49,$FA,$AF,$D8,$7F,$A8,$53,$F3,$FF,$D3,$2F,$CF
	.dc.b	$FD,$5F,$B1,$7E,$27,$F2,$AF,$D4,$4F,$FA,$7B,$F1,$8F,$E4,$3F,$9E
	.dc.b	$7F,$1E,$8F,$F7,$1F,$E7,$DF,$C9,$3F,$3B,$FE,$27,$F9,$47,$E0,$BF
	.dc.b	$22,$73,$FF,$28,$FC,$E3,$E4,$FE,$69,$FE,$8E,$C7,$FD,$FD,$F9,$27
	.dc.b	$F2,$CF,$CC,$FF,$81,$7E,$67,$F8,$DF,$CC,$7F,$85,$FE,$5B,$FC,$3A
	.dc.b	$BF,$86,$FE,$59,$EF,$FC,$AF,$FB,$87,$9B,$FC,$DB,$F3,$0F,$B9,$F9
	.dc.b	$4F,$F0,$5F,$C7,$7F,$41,$CD,$FA,$3F,$D9,$3E,$B7,$E4,$FF,$CD,$3F
	.dc.b	$27,$FE,$4B,$FB,$2B,$DC,$FE,$7B,$F9,$47,$C4,$FD,$9B,$F2,$1C,$AF
	.dc.b	$CA,$3F,$54,$7F,$20,$FD,$9F,$FD,$CF,$07,$FA,$1F,$E5,$3F,$CC,$3F
	.dc.b	$29,$FD,$AF,$C9,$7C,$EF,$CA,$7F,$9F,$7E,$D1,$FC,$27,$F3,$0F,$C2
	.dc.b	$1F,$CA,$3F,$20,$FD,$DF,$E5,$5F,$9B,$FB,$3F,$E5,$3F,$95,$7F,$30
	.dc.b	$FC,$A7,$FA,$7F,$ED,$7F,$CD,$4F,$CA,$FE,$55,$F9,$BF,$CA,$BC,$67
	.dc.b	$F0,$F3,$F8,$9F,$E5,$BF,$91,$1F,$94,$FF,$5A,$7F,$A8,$7E,$57,$FC
	.dc.b	$EB,$F6,$CF,$E7,$27,$F1,$1F,$DB,$7F,$AE,$FE,$41,$FE,$17,$F1,$1F
	.dc.b	$65,$FE,$E8,$6F,$E6,$3F,$D9,$3F,$75,$FE,$BF,$F8,$FF,$EA,$3F,$43
	.dc.b	$FE,$41,$F8,$F7,$F0,$7F,$C7,$BF,$45,$F8,$F7,$E5,$FF,$77,$FE,$48
	.dc.b	$7F,$28,$FC,$7F,$FA,$11,$E6,$CF,$FB,$8B,$FD,$94,$FC,$87,$E7,$7F
	.dc.b	$D4,$3F,$17,$FE,$E2,$7F,$9B,$7C,$43,$FA,$6F,$C7,$FF,$49,$FB,$CF
	.dc.b	$F5,$4F,$C7,$FF,$8E,$D3,$F1,$FF,$C4,$AF,$C7,$FF,$8F,$FE,$F7,$FD
	.dc.b	$4C,$FC,$0F,$E4,$1F,$C7,$3F,$7C,$FC,$DE,$DF,$E0,$4F,$C8,$7F,$85
	.dc.b	$7E,$43,$FE,$02,$FD,$F7,$FC,$97,$91,$FD,$37,$EF,$DF,$8D,$BF,$F0
	.dc.b	$5B,$FD,$BF,$F1,$8F,$E8,$5F,$91,$7F,$6F,$FD,$FF,$58,$3F,$23,$7D
	.dc.b	$6F,$F8,$43,$F2,$2F,$E2,$7F,$A0,$7F,$7A,$FE,$00,$FD,$57,$F0,$0F
	.dc.b	$F8,$B7,$F1,$EF,$F0,$A7,$E7,$4D,$D1,$FE,$15,$F2,$F0,$BF,$22,$7D
	.dc.b	$1E,$3F,$22,$7C,$26,$BF,$64,$E5,$8F,$2F,$C8,$DF,$2F,$CB,$8D,$FA
	.dc.b	$1F,$4D,$33,$7E,$08,$75,$78,$3F,$23,$FF,$10,$57,$F8,$5F,$E3,$3C
	.dc.b	$7F,$91,$3F,$14,$F2,$FE,$45,$FE,$EC,$A7,$EC,$AF,$CC,$3D,$3F,$92
	.dc.b	$7F,$96,$65,$FB,$37,$ED,$76,$F7,$FE,$43,$ED,$7E,$0F,$EC,$FF,$E8
	.dc.b	$BA,$FF,$0C,$74,$7C,$2F,$C4,$FF,$DA,$F1,$FE,$17,$FD,$7B,$E2,$FE
	.dc.b	$D0,$FD,$2F,$E3,$FE,$25,$F3,$7F,$6A,$FE,$95,$7F,$E1,$8E,$FF,$97
	.dc.b	$F0,$1F,$2B,$5F,$9F,$FB,$5F,$CA,$7E,$8F,$E4,$DF,$EE,$0B,$7E,$D9
	.dc.b	$97,$FC,$2F,$3F,$53,$F3,$CF,$93,$0F,$C9,$78,$7E,$DD,$BF,$C9,$9F
	.dc.b	$69,$F6,$BF,$25,$7D,$57,$DC,$FD,$BF,$E7,$E7,$F8,$0F,$C9,$7E,$7F
	.dc.b	$ED,$6E,$7F,$C0,$FC,$6F,$F8,$03,$67,$ED,$7F,$ED,$55,$FB,$5F,$F8
	.dc.b	$9D,$7E,$B3,$F4,$FF,$26,$FD,$A4,$3F,$70,$E9,$A7,$E3,$BF,$27,$F9
	.dc.b	$0F,$E4,$3F,$27,$F2,$BF,$92,$FD,$AD,$E0,$FE,$4B,$F2,$8E,$B3,$F9
	.dc.b	$6F,$CF,$3F,$D4,$99,$7F,$87,$7F,$2D,$FC,$BF,$ED,$7F,$36,$0F,$F9
	.dc.b	$B9,$7E,$E3,$FA,$1F,$D7,$7F,$D1,$90,$FC,$B3,$EB,$9F,$ED,$75,$FB
	.dc.b	$67,$C1,$3F,$D8,$AF,$F8,$9B,$F3,$5F,$D2,$FE,$5C,$FD,$6E,$FE,$87
	.dc.b	$FA,$8F,$C2,$FC,$5F,$B9,$70,$3F,$61,$F9,$7B,$F5,$FF,$D6,$BF,$8F
	.dc.b	$FD,$6F,$FD,$4D,$9F,$F8,$8F,$F5,$DF,$D8,$D7,$C2,$E4,$FD,$AF,$E5
	.dc.b	$FF,$ED,$2A,$7E,$5D,$FE,$7E,$7F,$80,$FD,$A7,$EE,$3F,$2D,$FD,$07
	.dc.b	$EE,$7F,$A3,$C3,$F7,$BF,$96,$FF,$06,$7F,$7F,$F9,$6F,$FB,$E7,$0F
	.dc.b	$CD,$7E,$CB,$FC,$02,$3F,$D8,$90,$FC,$B1,$E9,$FF,$04,$FC,$B3,$E0
	.dc.b	$61,$FC,$23,$F2,$C7,$C7,$FF,$08,$FC,$B3,$F6,$9F,$B7,$7D,$1F,$CB
	.dc.b	$3F,$DA,$19,$BF,$6E,$EB,$F8,$E7,$E7,$FF,$74,$FC,$3F,$ED,$FF,$B7
	.dc.b	$FD,$A7,$FC,$75,$7E,$1F,$85,$FE,$27,$F9,$67,$C1,$FD,$BF,$F7,$12
	.dc.b	$FD,$BF,$F7,$1F,$97,$7C,$AF,$CB,$7F,$98,$CB,$F6,$FF,$F6,$7B,$FE
	.dc.b	$D5,$F0,$3F,$8E,$FE,$DF,$F7,$3F,$2C,$FD,$BF,$E3,$2F,$C4,$3F,$90
	.dc.b	$FE,$57,$F8,$BD,$BF,$80,$FE,$DD,$FC,$6B,$AF,$F9,$3F,$E5,$9F,$E4
	.dc.b	$9D,$1F,$B7,$7F,$83,$97,$E5,$BF,$EC,$F5,$FB,$6B,$F0,$8F,$E5,$BF
	.dc.b	$51,$FE,$6A,$97,$E5,$BF,$EC,$F5,$F9,$6F,$D0,$CF,$F9,$AF,$9E,$FA
	.dc.b	$6F,$E6,$BF,$96,$FE,$C7,$F2,$DF,$A7,$83,$83,$F9,$9A,$FC,$B3,$F3
	.dc.b	$FF,$BA,$FD,$CF,$CB,$7F,$05,$2F,$CB,$7F,$0C,$BF,$2D,$FE,$25,$F9
	.dc.b	$6F,$D2,$FC,$B7,$F1,$DA,$FE,$47,$F2,$DF,$DD,$DB,$F2,$DF,$E4,$BF
	.dc.b	$96,$FD,$4D,$7F,$D4,$72,$FC,$B3,$FD,$CC,$FF,$88,$7C,$2F,$E9,$BF
	.dc.b	$95,$7F,$53,$FC,$A5,$F1,$5F,$D4,$3F,$29,$FD,$7F,$E5,$3F,$BB,$F1
	.dc.b	$3F,$91,$7E,$ED,$F4,$72,$7E,$53,$FD,$9D,$7E,$EF,$F0,$3F,$2A,$FA
	.dc.b	$9B,$3F,$1E,$FE,$B7,$F9,$4F,$F4,$CF,$CA,$7E,$77,$D8,$FF,$0A,$FC
	.dc.b	$9F,$FD,$85,$93,$F2,$7F,$A0,$FF,$88,$BE,$4B,$FB,$1F,$EF,$1F,$D4
	.dc.b	$7C,$6F,$EC,$FF,$93,$FF,$0A,$FD,$1B,$F5,$5F,$BC,$FC,$0F,$DA,$3F
	.dc.b	$53,$B7,$F2,$9F,$DE,$BF,$86,$7E,$99,$FB,$7F,$D9,$3F,$A9,$6A,$FD
	.dc.b	$ED,$F3,$BF,$DC,$3F,$66,$FD,$F7,$EF,$9F,$9E,$FD,$F7,$F9,$DF,$07
	.dc.b	$F7,$6F,$D8,$BE,$5F,$EF,$DF,$CA,$B2,$7E,$CD,$EF,$AF,$EF,$7F,$9A
	.dc.b	$7F,$0F,$7F,$BE,$EF,$FC,$D3,$DA,$FE,$57,$6F,$D9,$7F,$A7,$7E,$FE
	.dc.b	$FC,$2F,$FC,$03,$DF,$FC,$D2,$9F,$99,$FD,$1F,$CC,$FF,$41,$F9,$9F
	.dc.b	$F4,$A9,$7F,$00,$7F,$78,$FC,$CD,$CB,$FE,$0A,$FC,$CF,$FA,$57,$E6
	.dc.b	$7F,$9D,$DF,$FB,$2F,$F0,$0F,$F4,$BF,$CC,$BE,$66,$4F,$DE,$3F,$BE
	.dc.b	$7E,$65,$FC,$33,$97,$FC,$2B,$F9,$97,$F3,$8F,$C8,$3F,$8A,$FE,$63
	.dc.b	$FC,$33,$F3,$1F,$E7,$1F,$98,$7F,$7F,$FC,$C3,$FA,$8F,$E5,$DF,$C9
	.dc.b	$7F,$67,$FE,$51,$BF,$EA,$73,$FF,$0F,$FE,$5D,$FE,$09,$3F,$52,$7E
	.dc.b	$08,$FD,$B3,$FE,$18,$FD,$9F,$FC,$69,$F9,$4B,$F4,$6F,$9C,$DF,$DA
	.dc.b	$7F,$C5,$4E,$7F,$35,$BF,$2D,$FE,$0D,$87,$ED,$1F,$CA,$1F,$F1,$2F
	.dc.b	$E2,$7A,$FC,$2B,$7E,$77,$0F,$C2,$38,$FE,$1B,$CB,$F0,$3E,$7F,$9B
	.dc.b	$74,$FB,$BE,$F9,$BD,$AE,$FF,$5D,$E0,$FD,$2F,$C3,$FB,$DF,$8B,$E0
	.dc.b	$37,$DF,$E6,$FD,$FB,$FD,$BD,$C5,$FE,$3D,$FD,$CF,$D3,$F9,$4F,$F2
	.dc.b	$1E,$AF,$F1,$BA,$E8,$F6,$FE,$6D,$F9,$69,$FF,$1D,$AD,$7E,$0F,$EF
	.dc.b	$FF,$E7,$3C,$7F,$C8,$3F,$35,$F0,$BE,$53,$E1,$FE,$8F,$F1,$38,$BE
	.dc.b	$3F,$E6,$EF,$B7,$F8,$FE,$97,$C8,$F4,$3E,$4F,$E1,$DF,$2B,$93,$E6
	.dc.b	$FD,$57,$F9,$3B,$5F,$E4,$E6,$3D,$BC,$FE,$73,$D3,$C4,$DF,$E7,$5C
	.dc.b	$3F,$1C,$E3,$F3,$9C,$BE,$7B,$C3,$E4,$FD,$8F,$DB,$7F,$CE,$F9,$9F
	.dc.b	$ED,$0B,$F3,$A7,$E2,$3F,$7B,$DA,$74,$7E,$02,$BF,$C7,$83,$F6,$DF
	.dc.b	$94,$BF,$3B,$79,$9F,$86,$78,$F8,$1F,$E6,$4E,$4F,$F1,$1D,$7F,$42
	.dc.b	$FC,$0F,$DF,$FA,$9F,$C0,$FC,$AB,$F1,$BF,$B6,$FD,$8B,$7D,$AF,$C4
	.dc.b	$FA,$FF,$8B,$FA,$7F,$C6,$FC,$57,$E3,$FE,$97,$F2,$3E,$6B,$E4,$FE
	.dc.b	$77,$F2,$BE,$37,$9C,$FC,$CF,$88,$F9,$BF,$05,$F9,$AF,$DF,$7F,$61
	.dc.b	$C4,$7F,$90,$3F,$E4,$5F,$4B,$F3,$FF,$A0,$7E,$0E,$0F,$F0,$0E,$5F
	.dc.b	$F3,$FF,$E7,$FF,$D2,$DF,$F4,$0F,$E7,$EF,$D7,$7F,$4F,$FA,$0F,$F9
	.dc.b	$13,$07,$E8,$2E,$2F,$17,$9F,$E3,$FC,$67,$C9,$F9,$87,$97,$F2,$9F
	.dc.b	$D7,$FE,$82,$FD,$6F,$D1,$E3,$7A,$75,$F5,$47,$AF,$8B,$D9,$F8,$6F
	.dc.b	$B7,$CF,$F7,$7E,$6D,$EF,$FD,$0F,$F7,$3F,$B7,$3E,$2B,$E1,$7D,$D7
	.dc.b	$C3,$FB,$4F,$89,$FB,$AF,$C5,$FA,$AF,$DF,$FB,$3F,$93,$3F,$C0,$9D
	.dc.b	$0F,$D1,$7F,$82,$FE,$DF,$F8,$AF,$92,$FF,$3F,$4B,$F6,$FF,$E0,$5F
	.dc.b	$B7,$FF,$3C,$8F,$D1,$7C,$A7,$F0,$AF,$E0,$0F,$C5,$FE,$8F,$F4,$BF
	.dc.b	$70,$F8,$DF,$05,$FC,$D7,$C5,$FF,$55,$FE,$91,$F5,$3F,$48,$FF,$23
	.dc.b	$64,$FD,$23,$C7,$9B,$C2,$FE,$2B,$FA,$47,$ED,$F1,$FF,$59,$FF,$00
	.dc.b	$FF,$A8,$F0,$7E,$E0,$F8,$4F,$E3,$7F,$A5,$7F,$06,$FD,$27,$E7,$61
	.dc.b	$FE,$BA,$FD,$2B,$F3,$38,$3E,$2D,$EA,$7E,$27,$ED,$3F,$15,$F3,$1F
	.dc.b	$8B,$CF,$F9,$2F,$E3,$7F,$C8,$CF,$C0,$1E,$8F,$D2,$BC,$98,$FF,$88
	.dc.b	$38,$B7,$F0,$FC,$3C,$5E,$3F,$47,$CB,$E8,$F9,$FD,$1F,$4F,$9C,$EB
	.dc.b	$F4,$5D,$BE,$8B,$BF,$D1,$78,$3E,$7B,$C3,$F2,$3F,$CD,$BF,$4A,$FC
	.dc.b	$A6,$FF,$E2,$EF,$9F,$F2,$FC,$FF,$9B,$E7,$BC,$FF,$3D,$E8,$F9,$EF
	.dc.b	$4F,$CF,$7F,$3E,$E4,$7A,$9F,$D0,$3F,$49,$FE,$29,$FB,$7F,$F0,$6F
	.dc.b	$07,$FD,$BF,$E3,$FF,$4F,$D9,$FB,$7F,$F8,$ED,$FF,$54,$F0,$FF,$48
	.dc.b	$FD,$27,$1F,$C0,$3F,$D1,$7F,$4A,$FE,$23,$FA,$57,$F9,$DE,$DF,$A5
	.dc.b	$3E,$53,$FA,$7F,$C4,$BC,$0F,$EA,$3F,$B8,$7F,$69,$DF,$FD,$DF,$FA
	.dc.b	$4B,$C1,$FE,$AB,$F4,$FF,$EA,$AC,$FF,$D4,$1F,$9D,$FF,$58,$FD,$21
	.dc.b	$E1,$5E,$AF,$EB,$7F,$A4,$7E,$57,$57,$C1,$FF,$5E,$FD,$C3,$F8,$FF
	.dc.b	$E9,$1F,$EF,$AC,$DF,$B8,$7F,$4C,$FD,$C3,$7C,$3F,$F0,$01,$FE,$CF
	.dc.b	$FA,$57,$C1,$FD,$BF,$F8,$35,$BC,$DF,$E0,$96,$87,$ED,$FA,$FD,$29
	.dc.b	$F2,$7F,$DB,$BF,$4A,$FE,$84,$7F,$30,$3D,$9F,$A5,$7D,$4C,$39,$3E
	.dc.b	$2C,$3F,$DE,$DC,$FF,$DD,$BF,$6E,$D4,$3F,$4A,$FE,$73,$D5,$EB,$FD
	.dc.b	$27,$F7,$7F,$A4,$FB,$7F,$49,$FE,$CF,$FA,$4F,$B3,$4F,$F6,$2D,$99
	.dc.b	$DB,$FD,$8F,$F0,$51,$0D,$44,$FB,$9A,$C3,$FA,$71,$4E,$8D,$69,$5C
	.dc.b	$3C,$0E,$0E,$08,$EC,$86,$8F,$0F,$DD,$7F,$55,$F0,$3F,$C5,$DF,$82
	.dc.b	$FF,$8B,$E3,$FA,$C7,$B5,$8F,$0F,$DD,$1B,$7E,$04,$EF,$EE,$7F,$DC
	.dc.b	$51,$83,$D9,$8F,$C5,$FA,$AF,$F3,$EE,$5F,$1F,$F0,$2F,$F2,$75,$67
	.dc.b	$9B,$72,$AD,$AA,$DB,$99,$DC,$EB,$F0,$3F,$D1,$5F,$3A,$F9,$B8,$57
	.dc.b	$FC,$63,$E2,$FF,$92,$EF,$C7,$A7,$F7,$5F,$7B,$DD,$5F,$E0,$1D,$3F
	.dc.b	$BE,$E3,$EB,$8C,$4E,$AB,$FB,$E7,$A5,$C2,$7E,$E1,$E5,$3D,$DF,$39
	.dc.b	$DF,$5F,$39,$F8,$2D,$BD,$79,$FB,$AB,$EE,$BE,$BD,$38,$47,$A8,$E8
	.dc.b	$7A,$B3,$F7,$65,$FE,$0D,$F7,$BC,$9A,$FF,$90,$E7,$D1,$39,$6C,$66
	.dc.b	$9C,$1F,$E6,$D7,$D3,$8F,$D0,$CB,$CB,$9F,$A2,$7D,$78,$7C,$B7,$CF
	.dc.b	$1F,$9C,$E0,$FA,$31,$E6,$32,$7E,$95,$CD,$FE,$0B,$AF,$FC,$33,$3F
	.dc.b	$41,$B7,$F3,$DF,$F0,$E9,$9F,$B6,$FD,$74,$7D,$39,$F3,$E1,$F2,$F7
	.dc.b	$F8,$F1,$83,$CE,$78,$CF,$D6,$E1,$F3,$6B,$D5,$18,$AF,$56,$FF,$43
	.dc.b	$1F,$F1,$6E,$1E,$1F,$95,$F1,$37,$0F,$16,$1E,$E3,$F2,$D1,$E2,$71
	.dc.b	$A7,$CA,$7E,$D6,$5E,$17,$C9,$C7,$F9,$5D,$7E,$2E,$9F,$E8,$AD,$7F
	.dc.b	$B6,$65,$F0,$D7,$DF,$FC,$3F,$8D,$F8,$8F,$E4,$71,$EF,$A7,$85,$F6
	.dc.b	$1F,$0B,$3F,$C7,$38,$3F,$14,$78,$8F,$C6,$E5,$E2,$7F,$99,$61,$FE
	.dc.b	$F4,$C7,$C5,$F7,$FF,$4F,$3F,$A8,$F8,$6B,$A6,$BE,$F7,$E9,$1F,$12
	.dc.b	$BE,$7E,$BE,$FE,$0F,$33,$63,$F7,$1C,$1F,$E5,$CF,$D5,$7F,$01,$F7
	.dc.b	$DF,$81,$6C,$FD,$0C,$3B,$EF,$C9,$FD,$2E,$5E,$17,$EA,$6D,$EB,$7C
	.dc.b	$6F,$B7,$1F,$F3,$4F,$DF,$7E,$34,$FC,$77,$DF,$5C,$7B,$69,$FC,$27
	.dc.b	$1E,$98,$F8,$27,$D6,$FC,$8D,$7C,$5C,$3F,$74,$FC,$F7,$0F,$F9,$DA
	.dc.b	$BE,$D3,$F9,$BF,$BC,$FC,$E3,$F2,$71,$ED,$B7,$2E,$2F,$CE,$5F,$C5
	.dc.b	$CB,$F9,$13,$FA,$FF,$33,$FD,$C9,$A7,$EE,$78,$3F,$62,$D8,$CC,$FE
	.dc.b	$2F,$EF,$3F,$54,$7E,$AF,$EC,$EC,$7B,$B0,$F9,$39,$78,$5F,$C4,$65
	.dc.b	$DE,$7F,$9C,$1F,$95,$FB,$AE,$B3,$FB,$DC,$6F,$1F,$FA,$52,$6E,$CF
	.dc.b	$BC,$FD,$D7,$DE,$7E,$ED,$B8,$F8,$78,$BF,$88,$69,$E1,$CF,$F1,$07
	.dc.b	$E5,$3C,$6E,$EF,$FB,$AB,$E4,$2F,$17,$F8,$77,$0F,$2C,$FF,$0F,$9F
	.dc.b	$0F,$BF,$FC,$24,$FE,$15,$F7,$96,$BC,$1C,$3F,$3A,$FF,$45,$9F,$8A
	.dc.b	$7F,$69,$87,$6F,$91,$EE,$D8,$EF,$C3,$FE,$B2,$F9,$9F,$E2,$C7,$F1
	.dc.b	$BF,$E0,$04,$E0,$EF,$8F,$CC,$3F,$81,$E5,$E1,$CB,$FA,$D6,$5F,$2A
	.dc.b	$DF,$75,$DE,$FF,$02,$7F,$89,$63,$38,$7C,$08,$6D,$FD,$C9,$FE,$7B
	.dc.b	$1F,$CD,$7C,$AF,$EA,$D3,$FE,$3E,$8F,$E5,$4E,$FC,$5F,$E8,$55,$FC
	.dc.b	$2A,$7F,$3B,$97,$C3,$8F,$EB,$9F,$82,$3E,$61,$F7,$3F,$82,$8C,$7C
	.dc.b	$75,$9C,$E1,$F5,$27,$2F,$C8,$D3,$FE,$00,$3F,$CF,$DD,$E1,$E8,$3D
	.dc.b	$DF,$75,$F2,$F1,$FD,$34,$7E,$5A,$BC,$BC,$1F,$24,$FF,$19,$61,$E5
	.dc.b	$F8,$9F,$11,$E3,$FB,$9B,$5E,$1D,$3E,$B4,$FD,$76,$C7,$D6,$C3,$F2
	.dc.b	$3A,$FF,$28,$F0,$BF,$6B,$3F,$E0,$DC,$7E,$17,$DC,$FB,$FE,$E7,$E0
	.dc.b	$36,$7F,$27,$E8,$7F,$C0,$1F,$98,$FE,$EA,$7F,$76,$F9,$9B,$3D,$FC
	.dc.b	$7F,$E1,$B9,$76,$D3,$FC,$15,$D6,$E4,$70,$FE,$17,$4F,$C3,$37,$1E
	.dc.b	$FE,$93,$E3,$3F,$C0,$67,$F8,$13,$E6,$6C,$F7,$F0,$FF,$5D,$AF,$F5
	.dc.b	$5C,$FB,$70,$FE,$47,$E0,$FF,$24,$DB,$F5,$FE,$01,$FE,$75,$7F,$C1
	.dc.b	$AD,$DF,$EC,$71,$A3,$FB,$BC,$61,$D7,$67,$E3,$A3,$F9,$51,$FB,$0E
	.dc.b	$9F,$B4,$FE,$77,$EE,$BF,$44,$7E,$8F,$EE,$AE,$3E,$2F,$63,$FC,$65
	.dc.b	$89,$BB,$E1,$FC,$3F,$57,$E6,$4F,$B5,$8F,$E2,$4F,$9B,$CC,$FB,$AF
	.dc.b	$D8,$1F,$B1,$FB,$AB,$97,$C0,$F6,$7B,$9E,$0F,$6A,$C7,$CB,$3F,$27
	.dc.b	$87,$EF,$79,$3E,$C7,$DC,$FE,$EC,$FD,$E7,$CF,$5A,$FD,$17,$DC,$FF
	.dc.b	$07,$F9,$4F,$E1,$0D,$8E,$98,$7C,$53,$F8,$47,$85,$FB,$1E,$9F,$D7
	.dc.b	$72,$FE,$BD,$B5,$FA,$DF,$A8,$FE,$24,$7F,$13,$FA,$8B,$1D,$38,$FE
	.dc.b	$19,$F1,$7A,$3F,$53,$E4,$BA,$7F,$74,$7E,$EF,$96,$DF,$E0,$AB,$ED
	.dc.b	$75,$FD,$F6,$F5,$7F,$81,$F7,$FF,$46,$B5,$FF,$48,$7D,$EF,$F0,$B3
	.dc.b	$E9,$96,$7F,$D5,$5C,$BF,$9F,$FA,$9F,$E5,$35,$FC,$73,$1C,$FE,$78
	.dc.b	$FC,$1B,$3E,$76,$B0,$E1,$5F,$E4,$F3,$73,$A7,$8F,$E0,$FF,$E4,$FB
	.dc.b	$7D,$DD,$8C,$3F,$E7,$BF,$13,$FD,$1E,$7F,$99,$4B,$5F,$E6,$3C,$BF
	.dc.b	$AA,$C6,$5F,$EA,$AF,$E8,$0F,$25,$BE,$B3,$33,$FD,$9A,$1F,$53,$E3
	.dc.b	$3C,$9F,$59,$70,$FF,$2F,$61,$E0,$BF,$F2,$DB,$FA,$9F,$37,$F8,$33
	.dc.b	$E5,$F1,$BC,$27,$6A,$FE,$89,$1C,$76,$65,$FE,$76,$F1,$FC,$BF,$09
	.dc.b	$F0,$FE,$07,$E2,$3F,$07,$C9,$F5,$9D,$31,$E6,$E1,$A7,$F6,$2B,$FE
	.dc.b	$E5,$A7,$F5,$63,$FD,$71,$3E,$FF,$6B,$E8,$1F,$47,$D9,$B7,$EE,$F2
	.dc.b	$3E,$4F,$93,$F2,$9B,$1E,$A7,$FC,$1F,$B7,$F2,$2D,$35,$3F,$E0,$08
	.dc.b	$C3,$E3,$F8,$DF,$BC,$7D,$FF,$66,$C7,$CC,$C3,$AE,$9E,$7B,$FF,$37
	.dc.b	$B6,$33,$1F,$81,$7F,$C3,$FD,$2F,$C4,$DE,$2E,$B0,$FC,$AF,$C7,$7E
	.dc.b	$34,$FC,$77,$C7,$58,$FB,$7D,$FF,$86,$38,$67,$FE,$E3,$71,$1F,$61
	.dc.b	$FC,$37,$C7,$62,$7C,$DF,$1D,$F9,$5F,$8D,$FC,$B3,$6F,$FD,$AC,$3F
	.dc.b	$0D,$FE,$84,$FF,$89,$07,$E1,$BF,$95,$7F,$D5,$71,$AB,$FF,$06,$C6
	.dc.b	$9F,$EF,$5E,$A7,$F7,$BF,$52,$CF,$F7,$BF,$31,$1F,$9B,$FA,$EF,$E7
	.dc.b	$9D,$FF,$8E,$6F,$6F,$FB,$3F,$11,$FB,$A8,$FE,$84,$FF,$BF,$7E,$43
	.dc.b	$F6,$CE,$B9,$D1,$8B,$FC,$CF,$BD,$AA,$E1,$9F,$EF,$DE,$15,$9B,$E3
	.dc.b	$39,$E7,$7E,$F6,$F8,$BD,$5F,$5D,$1F,$03,$E8,$26,$1E,$8C,$A6,$9F
	.dc.b	$AA,$9F,$4E,$6F,$8E,$E5,$EF,$78,$5B,$87,$97,$05,$6E,$3C,$9F,$B1
	.dc.b	$BE,$47,$84,$61,$F2,$7C,$0F,$CA,$24,$78,$EF,$F9,$1C,$78,$BE,$33
	.dc.b	$E3,$7E,$7D,$FD,$C3,$8B,$F8,$C4,$62,$BE,$23,$F3,$A1,$FA,$E7,$E6
	.dc.b	$B0,$F3,$66,$AD,$4E,$6F,$F9,$90,$DD,$8F,$F8,$06,$3E,$2B,$5F,$E7
	.dc.b	$6D,$78,$35,$F8,$7F,$1D,$F5,$CF,$4F,$9D,$72,$F1,$C5,$9F,$0E,$4F
	.dc.b	$D9,$FF,$CC,$31,$E4,$39,$D7,$E0,$F1,$F2,$F9,$BF,$94,$CB,$FC,$0D
	.dc.b	$3F,$26,$BC,$73,$93,$ED,$FC,$7F,$E2,$4F,$97,$E6,$5A,$F8,$B1,$6B
	.dc.b	$E1,$E1,$F0,$BC,$9C,$70,$7F,$53,$CB,$FB,$05,$7C,$2F,$17,$C4,$4C
	.dc.b	$7C,$51,$FD,$6D,$FC,$E7,$AF,$FE,$88,$7D,$54,$F1,$33,$10,$FC,$64
	.dc.b	$EE,$BC,$74,$F1,$75,$18,$F9,$6F,$84,$4B,$F8,$5A,$7D,$A8,$FD,$9E
	.dc.b	$1F,$33,$3F,$57,$6F,$93,$E4,$3F,$1A,$DC,$3C,$BF,$2D,$72,$F0,$BF
	.dc.b	$92,$F9,$4F,$A0,$49,$EF,$AF,$FB,$8E,$BE,$96,$5F,$E8,$27,$D3,$D7
	.dc.b	$F0,$BE,$4B,$F3,$6D,$BF,$87,$F2,$D6,$BD,$EF,$C8,$F9,$3F,$49,$27
	.dc.b	$E6,$E1,$FE,$8C,$AF,$D4,$35,$FD,$E3,$5F,$E2,$AD,$7F,$C1,$D3,$CD
	.dc.b	$F3,$B5,$E6,$FC,$8F,$92,$95,$E0,$6C,$FB,$9A,$F8,$9A,$F8,$9A,$F8
	.dc.b	$9A,$FF,$87,$FE,$4F,$EA,$92,$7D,$AD,$8F,$96,$D7,$DE,$D7,$DE,$D7
	.dc.b	$DE,$FE,$03,$E4,$E4,$FC,$86,$CF,$D6,$F8,$9F,$CA,$9F,$CD,$A7,$FA
	.dc.b	$5F,$26,$1F,$C9,$C5,$9F,$DA,$7E,$79,$FE,$6D,$9C,$9F,$9C,$F1,$7F
	.dc.b	$C7,$BE,$EF,$E9,$DB,$1F,$C2,$3E,$5E,$DF,$F0,$89,$AF,$F9,$07,$0F
	.dc.b	$F2,$2F,$83,$FE,$46,$8C,$5F,$F5,$FF,$4F,$FD,$29,$39,$E0,$5F,$0A
	.dc.b	$FC,$0B,$6F,$F0,$2D,$59,$FA,$5F,$F0,$E5,$72,$AF,$F8,$03,$C8,$F9
	.dc.b	$7F,$3E,$DF,$C9,$3C,$DB,$39,$AF,$EC,$3B,$FF,$05,$9F,$5E,$1C,$F1
	.dc.b	$EB,$9F,$8E,$FE,$69,$FE,$FD,$9D,$69,$EB,$9F,$34,$62,$7F,$AB,$69
	.dc.b	$E7,$AD,$47,$1F,$4F,$F7,$A4,$7F,$4A,$FF,$00,$BF,$B1,$D9,$FE,$70
	.dc.b	$9F,$0E,$3F,$CB,$71,$F3,$4F,$48,$C5,$7F,$73,$C7,$E3,$4E,$4F,$E4
	.dc.b	$5E,$4F,$C2,$3E,$1F,$8F,$71,$F8,$67,$D2,$3D,$D7,$F0,$78,$7E,$CC
	.dc.b	$67,$1F,$8A,$E8,$FF,$03,$1F,$03,$E2,$BE,$1D,$7D,$BE,$1F,$EA,$BE
	.dc.b	$2F,$D4,$F6,$FF,$93,$71,$7F,$35,$8C,$61,$F4,$AF,$8B,$F8,$9B,$B8
	.dc.b	$F8,$09,$C3,$F8,$FD,$FE,$75,$79,$63,$E2,$4F,$F0,$EB,$F9,$93,$FC
	.dc.b	$49,$B1,$FA,$B8,$C1,$1C,$10,$E4,$E5,$FF,$08,$E5,$FD,$83,$2F,$F5
	.dc.b	$03,$67,$FD,$52,$FF,$2B,$8C,$F8,$3E,$D9,$FD,$27,$95,$EC,$C7,$EF
	.dc.b	$6D,$FE,$8D,$C7,$F5,$27,$F9,$E7,$E3,$DE,$FE,$F0,$FF,$11,$46,$1F
	.dc.b	$A3,$3F,$E0,$3C,$DD,$DF,$3F,$BB,$E8,$36,$BE,$FD,$7E,$B9,$F1,$9F
	.dc.b	$6D,$FF,$28,$E3,$E3,$E6,$FC,$E4,$7F,$92,$31,$F0,$3F,$4A,$BC,$79
	.dc.b	$7E,$DE,$7F,$D5,$71,$FD,$B1,$B3,$FD,$FD,$AF,$F8,$A5,$AF,$E3,$79
	.dc.b	$BE,$1E,$BF,$A2,$AF,$F2,$B9,$FE,$68,$71,$3E,$3B,$AA,$FC,$B7,$F5
	.dc.b	$3B,$BF,$20,$7E,$47,$E5,$AE,$3F,$0A,$FE,$7E,$FE,$08,$F8,$4D,$8F
	.dc.b	$A5,$8F,$B4,$FA,$4F,$86,$7E,$8F,$23,$81,$DF,$F9,$FF,$95,$FD,$03
	.dc.b	$6B,$C1,$BF,$CC,$7D,$87,$80,$F7,$F0,$7E,$0F,$7F,$9F,$A7,$F2,$D6
	.dc.b	$CF,$E7,$CF,$A2,$7E,$0B,$3F,$85,$A7,$F9,$F2,$7E,$BC,$7C,$F8,$C1
	.dc.b	$FA,$2D,$FD,$C7,$F0,$7A,$F0,$F0,$7F,$9F,$6F,$8D,$F6,$3E,$D5,$F2
	.dc.b	$AF,$BD,$EE,$F7,$23,$83,$FC,$5B,$F8,$87,$FA,$F3,$1F,$E8,$CD,$8F
	.dc.b	$7F,$9F,$FE,$8B,$8C,$61,$F8,$FB,$E6,$FE,$25,$AF,$EC,$BE,$47,$F8
	.dc.b	$61,$FC,$37,$E4,$AE,$1F,$76,$FF,$D9,$7A,$F8,$3D,$1F,$E3,$99,$7F
	.dc.b	$4D,$9C,$D7,$EB,$32,$C5,$7E,$6D,$B1,$F7,$B8,$DF,$23,$FC,$6F,$BF
	.dc.b	$F8,$E3,$67,$C1,$7C,$35,$FD,$5E,$57,$3F,$B5,$95,$BF,$01,$F6,$73
	.dc.b	$F1,$DF,$EB,$FD,$7F,$B4,$49,$FD,$63,$FE,$B8,$9F,$8F,$27,$E4,$CF
	.dc.b	$FB,$3B,$0F,$99,$8C,$8F,$F5,$14,$F0,$E1,$63,$9C,$3F,$CD,$7A,$BC
	.dc.b	$3F,$E7,$07,$F3,$9F,$76,$E3,$EE,$3D,$19,$7C,$C6,$92,$FF,$D5,$EF
	.dc.b	$FA,$E7,$2F,$6C,$7E,$DF,$1F,$84,$FF,$19,$F9,$5F,$E8,$DF,$29,$FD
	.dc.b	$1D,$B7,$DE,$79,$9F,$ED,$AB,$F8,$F1,$EF,$3C,$1B,$FD,$F6,$9F,$3A
	.dc.b	$78,$9F,$1F,$1F,$AC,$FF,$AE,$E9,$1F,$34,$93,$F9,$A6,$CF,$FA,$66
	.dc.b	$77,$39,$24,$66,$76,$F5,$7F,$DE,$BA,$3E,$57,$FB,$27,$C2,$FF,$65
	.dc.b	$6D,$FB,$F1,$FE,$8B,$7F,$1A,$1D,$3F,$2B,$29,$E9,$FC,$AC,$FF,$04
	.dc.b	$FF,$71,$CF,$FC,$39,$E2,$7F,$56,$F1,$7F,$B9,$7C,$8F,$F7,$37,$FC
	.dc.b	$77,$87,$CA,$C7,$A6,$7F,$E0,$13,$F4,$D8,$CC,$FE,$5E,$7F,$46,$30
	.dc.b	$FC,$37,$F2,$84,$8F,$87,$D7,$EA,$7F,$23,$5F,$E5,$9B,$FC,$CE,$0B
	.dc.b	$E1,$87,$3B,$FE,$13,$1F,$E4,$D8,$1F,$EE,$1F,$DA,$7F,$D7,$0B,$45
	.dc.b	$6B,$C2,$53,$E4,$BF,$C8,$AF,$0A,$DD,$F3,$9F,$87,$9F,$C1,$C3,$DE
	.dc.b	$7D,$07,$DF,$C1,$FC,$6B,$D1,$B3,$F9,$D7,$FC,$41,$8B,$E4,$BB,$9E
	.dc.b	$0F,$92,$B5,$F7,$4F,$E7,$F8,$7D,$DD,$FE,$E6,$1F,$C5,$4E,$B9,$FE
	.dc.b	$7D,$C5,$7F,$2C,$7F,$81,$11,$F9,$E7,$43,$E7,$9F,$23,$83,$E9,$4F
	.dc.b	$FB,$DF,$81,$5C,$63,$F3,$8D,$7E,$79,$F3,$FE,$4B,$E4,$9F,$2B,$E4
	.dc.b	$AE,$1F,$E0,$C3,$F2,$BE,$FF,$F9,$24,$FF,$25,$7B,$F6,$3F,$C2,$FD
	.dc.b	$EF,$89,$76,$7F,$A0,$63,$0D,$87,$6D,$8F,$E7,$1E,$9F,$F2,$12,$5F
	.dc.b	$F4,$E6,$E9,$FB,$56,$BF,$BD,$6B,$FC,$4D,$AF,$F0,$B7,$F9,$67,$A7
	.dc.b	$27,$FA,$CB,$67,$F8,$03,$5F,$E1,$0D,$7F,$8F,$35,$FE,$2C,$D5,$F8
	.dc.b	$8F,$E5,$A7,$E6,$C7,$CE,$8C,$38,$37,$1F,$A1,$7F,$43,$1A,$78,$B5
	.dc.b	$E5,$1B,$78,$BC,$DA,$75,$8E,$0F,$48,$E1,$18,$8D,$9D,$BD,$3C,$92
	.dc.b	$7A,$4A,$31,$3D,$5A,$F1,$9D,$C6,$23,$E0,$35,$F6,$FB,$9E,$22,$4F
	.dc.b	$66,$C7,$A5,$AF,$C2,$6B,$F2,$DA,$F4,$8F,$3F,$A7,$EE,$23,$82,$47
	.dc.b	$C5,$8C,$B6,$32,$FD,$08,$CB,$63,$E1,$FA,$9F,$28,$8E,$09,$1F,$46
	.dc.b	$32,$D8,$CB,$F7,$23,$2D,$8F,$9F,$EA,$7D,$62,$38,$24,$7D,$D8,$CB
	.dc.b	$63,$2F,$E2,$23,$2D,$8F,$B7,$EA,$7E,$08,$8E,$09,$1F,$89,$8C,$B6
	.dc.b	$32,$FE,$52,$32,$D8,$FC,$3F,$A7,$F8,$E2,$38,$24,$7E,$56,$32,$D8
	.dc.b	$CB,$FA,$08,$CB,$63,$F2,$7E,$A7,$E6,$88,$E0,$91,$FA,$18,$CB,$63
	.dc.b	$2F,$EB,$23,$2D,$8F,$CF,$FA,$3F,$A6,$23,$82,$47,$EB,$63,$2D,$8C
	.dc.b	$BF,$B8,$8C,$B6,$3F,$57,$E8,$FE,$C8,$8E,$09,$1F,$B9,$8C,$B6,$32
	.dc.b	$FF,$04,$8C,$B6,$3F,$6F,$E8,$7E,$F8,$8E,$09,$1F,$C1,$63,$2D,$8C
	.dc.b	$BF,$C4,$23,$2D,$8F,$E0,$7E,$8F,$F0,$A2,$38,$24,$7F,$11,$8C,$B6
	.dc.b	$32,$FF,$1C,$8C,$B6,$3F,$92,$7C,$B3,$E5,$9F,$05,$FC,$96,$30,$D8
	.dc.b	$CC,$F0,$6D,$7F,$30,$F2,$7F,$C0,$BF,$29,$FE,$02,$F0,$BF,$C0,$71
	.dc.b	$8B,$FE,$D0,$7F,$82,$FC,$77,$C9,$EE,$31,$77,$06,$7E,$29,$F0,$67
	.dc.b	$E3,$23,$FC,$A3,$D4,$FE,$5E,$47,$04,$AF,$E6,$D1,$96,$C6,$5F,$E8
	.dc.b	$31,$96,$C7,$F3,$4F,$53,$F9,$D9,$1C,$12,$3F,$A1,$46,$5B,$19,$7F
	.dc.b	$A6,$C6,$5B,$1F,$D0,$3B,$7F,$A3,$91,$C1,$23,$FA,$74,$65,$B1,$97
	.dc.b	$FA,$CC,$65,$B1,$F5,$FE,$0F,$ED,$11,$C1,$23,$FB,$4E,$3F,$83,$9D
	.dc.b	$46,$37,$46,$23,$FC,$37,$3B,$AE,$51,$8A,$DC,$7E,$1F,$E0,$7F,$14
	.dc.b	$47,$07,$F7,$5D,$FF,$C9,$BF,$35,$F8,$16,$C7,$E2,$2F,$FC,$AD,$F2
	.dc.b	$B9,$E1,$5F,$E1,$F2,$FC,$D6,$3D,$6F,$D0,$D7,$FB,$EF,$85,$FD,$D0
	.dc.b	$98,$FF,$70,$6C,$7F,$7B,$6B,$C2,$34,$47,$4D,$7F,$C1,$3E,$BF,$F8
	.dc.b	$00,$91,$FD,$F1,$FF,$05,$2C,$BF,$3D,$3F,$E6,$E7,$FC,$F5,$F0,$F8
	.dc.b	$9D,$1F,$06,$7F,$C3,$F5,$D1,$FF,$08,$35,$FF,$10,$C7,$F9,$E9,$BE
	.dc.b	$F7,$80,$91,$FE,$19,$3A,$C3,$9B,$FC,$F2,$F8,$E3,$E8,$1F,$EA,$6B
	.dc.b	$F2,$61,$E5,$BE,$75,$FE,$2D,$AF,$03,$FE,$5E,$6B,$FE,$8E,$5E,$87
	.dc.b	$C2,$23,$E3,$7E,$2D,$76,$8F,$0F,$9E,$C3,$17,$E7,$9E,$D7,$CA,$3C
	.dc.b	$4D,$7E,$1F,$3F,$D0,$24,$FC,$17,$BB,$F3,$4F,$87,$97,$FA,$69,$F8
	.dc.b	$06,$71,$FA,$93,$CA,$BE,$03,$5F,$A3,$1D,$1F,$88,$D7,$33,$F4,$63
	.dc.b	$11,$F1,$DA,$FC,$76,$BF,$4C,$F9,$DE,$C7,$E2,$89,$3F,$6D,$F0,$5F
	.dc.b	$A7,$A1,$F1,$5F,$22,$30,$FE,$36,$78,$DF,$4B,$E4,$7E,$17,$E4,$BF
	.dc.b	$2E,$47,$07,$A4,$FC,$2E,$1D,$9F,$8C,$62,$FF,$35,$3C,$6B,$F0,$CD
	.dc.b	$7F,$2F,$F1,$BF,$A4,$25,$7E,$55,$F5,$BF,$89,$BE,$D3,$F8,$13,$E3
	.dc.b	$47,$73,$ED,$4F,$28,$FC,$A3,$5F,$E3,$A7,$F9,$EF,$8F,$F2,$8F,$F6
	.dc.b	$23,$07,$80,$FF,$70,$D7,$83,$2F,$D2,$DF,$E1,$8F,$26,$3F,$BA,$BE
	.dc.b	$77,$DA,$7F,$D0,$D1,$88,$FE,$4E,$78,$7D,$E9,$3A,$6C,$7F,$74,$F8
	.dc.b	$7E,$87,$FB,$F7,$1B,$5F,$C0,$ED,$7C,$2F,$F9,$CB,$5F,$E6,$BE,$87
	.dc.b	$BC,$95,$E4,$7E,$16,$1D,$27,$E6,$4B,$8B,$FB,$EE,$7F,$C1,$31,$5F
	.dc.b	$BF,$0E,$78,$7F,$82,$9A,$FF,$4A,$F8,$AF,$9C,$4C,$7E,$1B,$E0,$7D
	.dc.b	$55,$ED,$69,$ED,$8E,$13,$FE,$16,$6B,$FD,$5B,$E2,$BE,$D1,$27,$E7
	.dc.b	$BD,$63,$F0,$C7,$F8,$2A,$BB,$38,$C3,$D7,$5F,$2E,$B9,$4F,$F8,$91
	.dc.b	$AF,$F6,$4A,$E8,$D9,$FB,$B1,$88,$FE,$C6,$D4,$78,$63,$64,$7E,$F3
	.dc.b	$5F,$EE,$3F,$0F,$F9,$52,$47,$E2,$9F,$66,$5F,$C5,$73,$65,$F6,$63
	.dc.b	$0F,$E4,$E7,$8D,$FF,$8F,$9A,$FF,$7E,$F4,$3F,$44,$4C,$3F,$2E,$F6
	.dc.b	$7F,$9C,$61,$D9,$FA,$B3,$F8,$68,$C3,$F2,$E7,$93,$FE,$4C,$6B,$FE
	.dc.b	$F4,$F9,$67,$9F,$FE,$EB,$6D,$FF,$BA,$DB,$3F,$EF,$76,$BF,$3C,$F0
	.dc.b	$63,$EF,$9F,$2B,$FD,$AA,$77,$3F,$40,$38,$7E,$9B,$5C,$D7,$8F,$CD
	.dc.b	$97,$E3,$6C,$FD,$63,$C3,$1F,$0B,$3F,$9E,$FF,$75,$B5,$70,$AF,$80
	.dc.b	$D7,$C9,$0F,$33,$E1,$12,$BE,$13,$67,$E1,$35,$F9,$4D,$79,$B5,$F6
	.dc.b	$BF,$E7,$7F,$8A,$93,$FE,$77,$6C,$FF,$8E,$DA,$FF,$8D,$DA,$FE,$63
	.dc.b	$83,$FD,$09,$18,$8F,$F3,$9F,$86,$8C,$47,$CE,$F3,$7E,$E1,$2F,$EE
	.dc.b	$36,$3E,$E3,$5F,$C2,$35,$F8,$AD,$7E,$BB,$F8,$7F,$36,$4F,$E1,$DB
	.dc.b	$3F,$87,$6B,$F8,$F6,$BF,$41,$AF,$E0,$9A,$FE,$1B,$CD,$FC,$A1,$27
	.dc.b	$F2,$8D,$8F,$CA,$35,$FC,$E3,$5F,$BA,$D7,$F1,$EF,$E7,$FC,$D9,$3F
	.dc.b	$9F,$6C,$FE,$7D,$AF,$E9,$DA,$FE,$21,$AF,$E6,$9A,$FE,$7B,$CD,$FD
	.dc.b	$61,$27,$F5,$8D,$8F,$D6,$35,$FD,$A3,$5F,$CA,$B5,$FD,$3B,$FB,$7F
	.dc.b	$2E,$4F,$ED,$DB,$3F,$AF,$3F,$8F,$FE,$22,$F6,$3F,$CB,$AD,$AF,$F2
	.dc.b	$D7,$C4,$BE,$13,$35,$FB,$9C,$39,$E1,$FC,$1D,$AF,$EA,$3E,$17,$F8
	.dc.b	$41,$30,$FE,$10,$FF,$96,$23,$F8,$06,$3D,$DF,$E8,$2D,$E5,$FF,$2E
	.dc.b	$E5,$E1,$BF,$F3,$64,$E5,$EC,$74,$25,$7F,$13,$BF,$E3,$19,$49,$FE
	.dc.b	$30,$D9,$F8,$42,$BA,$3F,$13,$2F,$F3,$FE,$1E,$18,$F2,$C7,$F1,$9A
	.dc.b	$E5,$1F,$C9,$DA,$FB,$3E,$03,$DC,$FC,$E6,$BD,$DB,$66,$D7,$CE,$C3
	.dc.b	$E8,$DD,$BF,$8B,$5C,$09,$3E,$A8,$E8,$DB,$F9,$91,$F7,$DA,$FD,$0B
	.dc.b	$EF,$3B,$6C,$7C,$49,$DB,$F0,$E7,$44,$75,$3C,$1E,$DE,$6F,$B0,$93
	.dc.b	$E3,$7A,$E7,$FE,$98,$B7,$C9,$7B,$63,$0F,$E4,$A7,$8D,$F8,$DA,$FA
	.dc.b	$FE,$5F,$E4,$93,$0F,$80,$F6,$7F,$1B,$5D,$5A,$7E,$2A,$38,$4F,$BD
	.dc.b	$AF,$C7,$F9,$6F,$AA,$49,$F9,$8F,$D8,$AF,$D3,$9D,$67,$F5,$6F,$EA
	.dc.b	$A3,$84,$7C,$B6,$BF,$4E,$39,$4F,$D7,$6C,$7D,$16,$BB,$8D,$11,$D3
	.dc.b	$5D,$1F,$67,$C5,$FC,$69,$23,$F0,$8D,$8D,$46,$08,$E0,$87,$E2,$63
	.dc.b	$83,$FC,$27,$D2,$7C,$67,$E7,$1B,$5F,$93,$6C,$FE,$B9,$AF,$38,$FF
	.dc.b	$73,$43,$D4,$FE,$23,$7F,$EE,$E8,$C4,$EA,$FB,$11,$EC,$F8,$7E,$02
	.dc.b	$47,$85,$B8,$78,$70,$E8,$76,$9F,$0B,$DE,$77,$3E,$D6,$BF,$E1,$5B
	.dc.b	$79,$3E,$F2,$5F,$BD,$F8,$57,$DA,$3F,$8B,$E3,$6F,$FC,$BB,$7C,$A7
	.dc.b	$81,$23,$E3,$5F,$28,$F9,$57,$67,$D5,$18,$23,$82,$1F,$56,$38,$36
	.dc.b	$36,$FD,$9F,$53,$EC,$12,$3E,$C3,$DF,$74,$7C,$1E,$CF,$54,$62,$BF
	.dc.b	$CC,$F8,$FF,$83,$6B,$ED,$DF,$39,$FC,$03,$5F,$C4,$5F,$E1,$C9,$3F
	.dc.b	$87,$6C,$7E,$1D,$AF,$08,$D1,$1D,$35,$FC,$A7,$BD,$F9,$32,$47,$E4
	.dc.b	$DF,$E6,$3F,$88,$BD,$EF,$56,$5E,$BB,$4E,$6F,$F2,$B8,$69,$FF,$2B
	.dc.b	$46,$67,$ED,$E1,$D9,$FC,$F3,$5F,$ED,$D1,$B6,$E1,$E2,$25,$7F,$39
	.dc.b	$9D,$4F,$69,$CB,$FD,$5A,$32,$FF,$A5,$9A,$FA,$23,$6F,$69,$E9,$EF
	.dc.b	$7C,$02,$38,$24,$75,$8C,$B6,$32,$FA,$A3,$2D,$8F,$47,$BD,$F3,$08
	.dc.b	$E0,$91,$E9,$8C,$B6,$32,$FC,$68,$CB,$63,$E2,$78,$DF,$60,$8E,$09
	.dc.b	$1F,$16,$32,$D8,$CB,$F4,$A3,$2D,$8F,$F0,$0F,$88,$F1,$19,$FF,$78
	.dc.b	$8C,$36,$33,$3C,$1B,$5F,$E2,$BA,$DC,$E6,$B8,$11,$C1,$27,$85,$72
	.dc.b	$7F,$D4,$55,$C9,$FC,$07,$BD,$F9,$82,$57,$E1,$DB,$1A,$8C,$11,$C1
	.dc.b	$0F,$C9,$C7,$07,$F8,$E7,$B6,$F8,$DF,$DE,$7C,$4F,$EF,$78,$D3,$9C
	.dc.b	$3F,$DC,$B3,$98,$75,$F0,$9F,$E0,$71,$BC,$15,$FE,$E1,$F8,$AF,$E1
	.dc.b	$1F,$13,$FE,$E3,$E7,$B8,$7F,$B6,$26,$DF,$F1,$0F,$43,$FA,$0A,$BF
	.dc.b	$E2,$1F,$0A,$F8,$67,$BD,$C2,$30,$D8,$FE,$5D,$5C,$71,$FF,$7E,$D6
	.dc.b	$9F,$F7,$FD,$FF,$01,$E8,$79,$BF,$C1,$89,$3F,$B5,$6D,$FE,$CD,$AF
	.dc.b	$EF,$9A,$FF,$06,$6B,$FB,$77,$F8,$8F,$9B,$27,$F8,$03,$67,$F7,$CD
	.dc.b	$7F,$84,$B5,$FE,$22,$D7,$F8,$13,$79,$1F,$15,$FC,$19,$BD,$7F,$9B
	.dc.b	$8C,$11,$C1,$2F,$F4,$71,$96,$C6,$6F,$F3,$D3,$FC,$43,$17,$9B,$FC
	.dc.b	$B0,$97,$FC,$6D,$B5,$FC,$65,$AF,$F2,$46,$BF,$CB,$1A,$FF,$1D,$7F
	.dc.b	$9B,$FC,$1C,$9F,$E4,$CD,$9F,$E4,$8D,$7F,$97,$B5,$FE,$6E,$D7,$F9
	.dc.b	$43,$5F,$E7,$FF,$05,$FD,$10,$93,$FC,$CD,$B1,$FC,$C5,$AF,$F3,$C6
	.dc.b	$BF,$D1,$1A,$FF,$35,$7F,$CF,$FF,$38,$F8,$3A,$4A,$FF,$3C,$46,$5B
	.dc.b	$19,$AF,$F4,$AC,$E5,$B1,$97,$FD,$75,$19,$6C,$65,$12,$1F,$07,$E4
	.dc.b	$25,$78,$9B,$3E,$06,$BD,$9A,$F4,$13,$D2,$79,$12,$3D,$3F,$07,$F0
	.dc.b	$89,$1F,$01,$B1,$A8,$C1,$1C,$10,$F7,$C7,$07,$E0,$FC,$1F,$CE,$24
	.dc.b	$FC,$C6,$C7,$CA,$6B,$C2,$34,$47,$4D,$7E,$6F,$C1,$FD,$A2,$47,$D8
	.dc.b	$6C,$7D,$66,$BC,$23,$44,$74,$D7,$EC,$FC,$1F,$E1,$89,$1F,$84,$6C
	.dc.b	$7E,$09,$AF,$08,$D1,$1D,$35,$FC,$2F,$C1,$FE,$48,$91,$F9,$06,$C7
	.dc.b	$E3,$9A,$F0,$8D,$11,$D3,$5F,$C1,$1F,$DA,$4F,$F6,$EF,$85,$F9,$08
	.dc.b	$FF,$6A,$36,$3F,$DD,$ED,$78,$B5,$FC,$63,$5F,$CC,$B5,$ED,$1D,$1E
	.dc.b	$E4,$70,$78,$3E,$0F,$C6,$47,$04,$8F,$1C,$65,$B1,$97,$CB,$19,$6C
	.dc.b	$7B,$3E,$0F,$DE,$47,$04,$8F,$7C,$65,$B1,$97,$E0,$C6,$5B,$1F,$23
	.dc.b	$DF,$F9,$64,$70,$48,$F9,$71,$96,$C6,$5F,$9B,$19,$6C,$7D,$4F,$83
	.dc.b	$FA,$E4,$70,$48,$FA,$F1,$96,$C6,$5F,$B3,$19,$6C,$7E,$03,$E0,$FF
	.dc.b	$06,$47,$04,$8F,$C1,$C6,$5B,$19,$7F,$0B,$19,$6C,$7E,$33,$E0,$FF
	.dc.b	$1E,$47,$04,$8F,$C7,$C6,$5B,$19,$7F,$23,$19,$6C,$7E,$63,$E0,$FF
	.dc.b	$36,$47,$04,$8F,$CD,$C6,$5B,$19,$7F,$3B,$19,$6C,$7E,$A1,$FE,$6B
	.dc.b	$D1,$FE,$7A,$27,$4F,$FA,$4E,$3F,$97,$46,$27,$84,$7F,$7C,$6C,$66
	.dc.b	$3A,$35,$FD,$AB,$5F,$E8,$7D,$5F,$E0,$6F,$37,$FC,$71,$08,$FF,$04
	.dc.b	$F9,$B8,$75,$7F,$E4,$03,$8D,$A7,$8F,$93,$FE,$46,$E4,$FF,$1C,$B6
	.dc.b	$BF,$C5,$A7,$F9,$13,$E3,$73,$E4,$AA,$FA,$F2,$B7,$9B,$0F,$F2,$8F
	.dc.b	$3E,$63,$FC,$74,$7C,$8F,$84,$F8,$CD,$AF,$87,$18,$8F,$F1,$1B,$74
	.dc.b	$34,$93,$C0,$93,$F2,$A3,$2D,$8F,$9F,$B7,$32,$38,$24,$7D,$28,$CB
	.dc.b	$63,$2F,$D4,$8C,$B6,$3E,$DE,$BE,$72,$38,$24,$7D,$E8,$CB,$63,$2F
	.dc.b	$E0,$23,$2D,$8F,$C3,$FA,$1F,$0C,$8E,$09,$1F,$8A,$8C,$B6,$32,$FE
	.dc.b	$32,$32,$D8,$FF,$24,$F9,$FF,$89,$C4,$DA,$C3,$F8,$8D,$9F,$2F,$F8
	.dc.b	$4F,$67,$F0,$AD,$C3,$F0,$7B,$7F,$A4,$BB,$3F,$12,$7E,$2B,$E3,$EC
	.dc.b	$FE,$23,$4F,$F3,$0E,$5F,$7C,$91,$FA,$66,$C7,$E9,$9A,$FE,$99,$AF
	.dc.b	$EC,$1A,$FE,$D9,$FC,$57,$C2,$49,$FD,$8B,$67,$F6,$2D,$7F,$62,$D7
	.dc.b	$F7,$6D,$7F,$80,$B5,$FF,$3C,$FC,$27,$E5,$C9,$3F,$BE,$6C,$7E,$F9
	.dc.b	$AF,$EF,$9A,$FF,$08,$6B,$FC,$39,$FD,$17,$C2,$49,$FE,$12,$D9,$FE
	.dc.b	$12,$D7,$F8,$4B,$5F,$E2,$6D,$7F,$8C,$B5,$FF,$48,$FC,$27,$EB,$C9
	.dc.b	$3F,$C5,$9B,$1F,$C5,$9A,$FF,$16,$6B,$FC,$81,$AF,$F2,$67,$F7,$5F
	.dc.b	$07,$27,$F9,$0B,$67,$F9,$0B,$5F,$E4,$2D,$7F,$95,$B5,$FE,$62,$D5
	.dc.b	$F8,$7B,$E1,$3F,$85,$36,$BF,$98,$B6,$7F,$9D,$35,$FF,$11,$B5,$FF
	.dc.b	$33,$35,$FF,$39,$B5,$EF,$3D,$89,$3D,$9B,$1D,$BE,$13,$C6,$49,$F1
	.dc.b	$B6,$35,$18,$23,$82,$1E,$68,$E0,$FA,$3E,$0F,$DE,$49,$F7,$B6,$3D
	.dc.b	$ED,$78,$46,$88,$E9,$AF,$C4,$F6,$BE,$59,$23,$E5,$B6,$3E,$5B,$5E
	.dc.b	$11,$A2,$3A,$6B,$F4,$3A,$FE,$B9,$23,$EB,$B6,$3E,$BB,$5E,$11,$A2
	.dc.b	$3A,$6B,$F7,$39,$3F,$06,$48,$FC,$1B,$63,$F0,$6D,$78,$46,$88,$E9
	.dc.b	$AF,$E2,$3E,$07,$F1,$E4,$8F,$C7,$B6,$3F,$1E,$D7,$84,$68,$8E,$9A
	.dc.b	$FF,$9D,$BE,$6D,$F0,$9F,$D9,$5B,$3F,$E0,$C6,$CF,$F8,$55,$AF,$19
	.dc.b	$E0,$49,$E0,$D8,$E0,$D7,$84,$75,$F8,$4F,$01,$1C,$12,$3C,$11,$96
	.dc.b	$C6,$5F,$54,$65,$B1,$E9,$F8,$4F,$61,$1C,$12,$3D,$91,$96,$C6,$5F
	.dc.b	$8D,$19,$6C,$7C,$5F,$7B,$E4,$11,$C1,$23,$E4,$46,$5B,$19,$7E,$94
	.dc.b	$65,$B1,$F4,$7E,$13,$EA,$11,$C1,$23,$EA,$46,$5B,$19,$7E,$F4,$65
	.dc.b	$B1,$F7,$7E,$0F,$F0,$04,$70,$48,$FC,$04,$65,$B1,$97,$F1,$51,$96
	.dc.b	$C7,$F9,$6F,$C6,$78,$0F,$87,$FF,$01,$75,$7F,$80,$DB,$5F,$E4,$BB
	.dc.b	$2F,$CA,$1E,$5F,$F9,$1B,$87,$35,$FE,$40,$AD,$4E,$5F,$F0,$2F,$E7
	.dc.b	$3F,$EC,$DF,$7B,$7E,$12,$3F,$C9,$53,$AB,$FF,$2A,$C6,$1E,$9F,$BC
	.dc.b	$7F,$85,$A7,$FC,$E8,$1C,$47,$FB,$7A,$31,$E0,$FF,$94,$1A,$70,$9D
	.dc.b	$C8,$FF,$71,$38,$6C,$F1,$8C,$47,$C2,$6C,$F4,$8F,$04,$60,$DA,$E6
	.dc.b	$7C,$AF,$EE,$C9,$5F,$BB,$6C,$FE,$ED,$AF,$F0,$36,$BE,$73,$A3,$87
	.dc.b	$D5,$3D,$A7,$F8,$37,$C4,$7F,$09,$23,$82,$4F,$F0,$98,$CB,$63,$2F
	.dc.b	$F1,$38,$CB,$63,$F8,$97,$A1,$FC,$54,$8E,$09,$1F,$C5,$63,$2D,$8C
	.dc.b	$BF,$C7,$E3,$2D,$8F,$F5,$4F,$CF,$BD,$1F,$F4,$33,$63,$31,$82,$38
	.dc.b	$24,$FF,$AE,$67,$83,$67,$6D,$9D,$BF,$37,$97,$F4,$BE,$FF,$F9,$94
	.dc.b	$FF,$33,$78,$F6,$BF,$49,$4C,$3A,$F0,$FE,$73,$83,$F9,$AB,$FA,$49
	.dc.b	$D6,$5F,$9C,$ED,$FC,$EF,$AF,$F9,$EF,$95,$B9,$7E,$99,$FD,$0F,$7F
	.dc.b	$FA,$E3,$23,$7D,$8C,$3F,$EC,$4A,$7C,$3B,$D8,$FD,$9E,$BF,$B4,$9C
	.dc.b	$FC,$C6,$1F,$D0,$1C,$27,$1B,$7E,$FA,$D5,$8F,$86,$7D,$E3,$C2,$C3
	.dc.b	$71,$FF,$61,$CE,$BB,$73,$5F,$BA,$F8,$9F,$57,$8D,$EB,$6D,$FF,$0D
	.dc.b	$3D,$3E,$2F,$89,$A7,$FB,$38,$F6,$FB,$99,$D1,$3E,$2E,$EF,$1F,$6D
	.dc.b	$BF,$1E,$A8,$F6,$EF,$F5,$FC,$5F,$C5,$92,$BF,$1E,$D8,$FC,$43,$5F
	.dc.b	$D8,$1C,$63,$64,$76,$D7,$F6,$DE,$17,$F1,$DF,$77,$ED,$1F,$6F,$E2
	.dc.b	$AD,$7C,$1D,$1F,$2D,$ED,$E1,$F7,$36,$7E,$6F,$B5,$F8,$E7,$F8,$C7
	.dc.b	$27,$FB,$F7,$BB,$EB,$4F,$FB,$A2,$77,$1F,$CA,$BC,$BF,$C5,$74,$7E
	.dc.b	$2D,$B3,$E1,$E7,$F9,$87,$CD,$E3,$B1,$E0,$F6,$FA,$E3,$D9,$B1,$FE
	.dc.b	$BB,$F0,$5F,$0F,$FB,$62,$47,$EE,$9B,$C7,$FB,$36,$BF,$32,$78,$CE
	.dc.b	$C9,$1B,$6B,$F3,$3B,$FF,$2B,$1F,$D5,$79,$3F,$A5,$76,$7E,$9F,$A7
	.dc.b	$F5,$0F,$E9,$79,$3F,$05,$5F,$8B,$5E,$37,$E2,$FA,$7F,$18,$FC,$AF
	.dc.b	$32,$57,$D0,$EA,$BA,$3B,$FF,$7D,$DB,$FB,$5E,$9F,$DB,$36,$FF,$39
	.dc.b	$FA,$FF,$F6,$87,$F3,$4D,$BF,$EC,$CF,$DB,$7F,$B6,$B9,$BF,$ED,$7D
	.dc.b	$4E,$57,$C5,$FD,$EF,$8B,$FB,$ED,$C7,$FD,$D9,$D9,$FD,$F1,$FC,$BE
	.dc.b	$5F,$D7,$63,$0E,$1A,$FE,$71,$AF,$F4,$D6,$BC,$7D,$0F,$E3,$4E,$5A
	.dc.b	$E6,$7B,$BF,$CD,$E3,$2D,$8C,$CF,$48,$C1,$FA,$CE,$5F,$F0,$17,$77
	.dc.b	$F1,$F3,$F9,$07,$65,$BF,$97,$F0,$1F,$C1,$77,$76,$58,$C3,$FA,$2F
	.dc.b	$94,$95,$FA,$8F,$72,$E7,$F6,$63,$FC,$01,$9F,$F8,$47,$BF,$F9,$77
	.dc.b	$4F,$F2,$F6,$CF,$DA,$E2,$F0,$1E,$1F,$81,$B1,$FB,$4F,$03,$F6,$3F
	.dc.b	$29,$FB,$26,$C7,$D4,$F6,$7E,$1F,$19,$A4,$8F,$8A,$DD,$3E,$B3,$5F
	.dc.b	$AC,$D7,$EB,$35,$69,$FE,$2B,$ED,$FE,$90,$7F,$8D,$3E,$3F,$FC,$5F
	.dc.b	$FA,$17,$E4,$5F,$EA,$D1,$FD,$27,$93,$FC,$53,$D7,$FD,$1D,$FF,$22
	.dc.b	$65,$F9,$8F,$5B,$E8,$78,$1F,$C3,$BE,$02,$4F,$F1,$4F,$8E,$BA,$FC
	.dc.b	$D3,$C7,$A7,$F7,$0D,$3C,$3D,$3E,$3B,$FA,$33,$96,$C7,$8F,$C2,$FB
	.dc.b	$C7,$DF,$EE,$B3,$FC,$8B,$E6,$BF,$8F,$FC,$17,$F2,$06,$C7,$E8,$BE
	.dc.b	$1F,$E7,$91,$FC,$8B,$73,$FD,$13,$5F,$D1,$35,$FD,$13,$5F,$D5,$3F
	.dc.b	$A0,$E3,$67,$F9,$2A,$FE,$24,$DA,$69,$F4,$3A,$7E,$04,$7C,$4E,$EF
	.dc.b	$A1,$E1,$7E,$6F,$D7,$FC,$E3,$71,$FE,$D7,$FA,$E3,$E0,$FF,$B9,$36
	.dc.b	$33,$F1,$F8,$9F,$EF,$3A,$F9,$3B,$FB,$BE,$6E,$8B,$BA,$6C,$7E,$07
	.dc.b	$DB,$FD,$61,$FA,$DF,$12,$E5,$E2,$F9,$2F,$0F,$C2,$F8,$9B,$1F,$E6
	.dc.b	$9E,$EF,$F0,$F7,$E2,$BF,$E7,$7E,$3F,$EC,$ED,$7F,$CB,$0D,$7F,$CE
	.dc.b	$0D,$4D,$26,$BF,$C2,$9B,$5F,$C3,$9B,$3F,$C7,$9E,$93,$F5,$3B,$BD
	.dc.b	$E7,$C0,$EB,$B7,$FC,$D3,$C7,$FE,$15,$E2,$7F,$0B,$6C,$7D,$0F,$0F
	.dc.b	$E7,$9F,$43,$E3,$EC,$7B,$CF,$CC,$F7,$FC,$CE,$8F,$9A,$D8,$FE,$83
	.dc.b	$E3,$7F,$17,$3F,$8C,$78,$76,$3E,$EF,$CA,$7D,$CF,$88,$FB,$AD,$8E
	.dc.b	$DC,$3F,$1B,$DC,$F8,$44,$8F,$84,$D8,$F8,$4D,$7E,$8B,$5F,$A2,$D7
	.dc.b	$A4,$7F,$46,$E1,$FC,$DF,$77,$E1,$8F,$C3,$F4,$59,$FE,$B7,$E3,$7F
	.dc.b	$31,$F1,$3F,$99,$36,$3F,$2F,$F1,$1F,$96,$3F,$2F,$E2,$D8,$F5,$E3
	.dc.b	$FA,$8E,$CF,$C9,$F2,$FE,$51,$B3,$FD,$A3,$C6,$FE,$7C,$7F,$90,$FC
	.dc.b	$97,$C9,$E2,$7F,$4B,$F1,$BF,$A4,$F8,$AF,$D2,$B6,$3E,$4F,$0F,$E3
	.dc.b	$3D,$4F,$C5,$12,$3F,$14,$D8,$FC,$53,$5F,$CC,$B5,$FC,$CB,$5F,$CC
	.dc.b	$BF,$C0,$FC,$5F,$F0,$CC,$7F,$00,$9F,$E3,$BD,$BF,$B0,$E9,$FD,$8B
	.dc.b	$FE,$77,$CF,$FB,$9F,$83,$FE,$75,$6D,$7F,$9D,$1F,$DF,$5F,$FB,$AD
	.dc.b	$CA,$F5,$1E,$4B,$E3,$7F,$82,$FC,$6F,$F0,$66,$E1,$F4,$34,$77,$7F
	.dc.b	$02,$E8,$FE,$06,$D9,$FF,$39,$66,$F2,$FE,$87,$8F,$F4,$5B,$3F,$C5
	.dc.b	$3E,$2F,$F8,$99,$FE,$85,$FC,$03,$FD,$F9,$1F,$E5,$4D,$53,$FE,$70
	.dc.b	$FC,$57,$FC,$41,$C1,$FE,$6A,$6B,$FE,$DB,$6B,$FE,$F3,$6A,$69,$2B
	.dc.b	$FC,$1A,$D7,$FC,$22,$D8,$FF,$10,$BD,$3A,$FF,$9B,$F7,$7F,$22,$3F
	.dc.b	$91,$F5,$5B,$F8,$3E,$5F,$E3,$CF,$C8,$7C,$A5,$8F,$F4,$EE,$8F,$8A
	.dc.b	$F8,$1F,$15,$F0,$5B,$3F,$9E,$F0,$FF,$2E,$FF,$95,$5F,$F1,$8F,$A8
	.dc.b	$8F,$F0,$8C,$6B,$6F,$C4,$1F,$DE,$9D,$C7,$E0,$FE,$6F,$F4,$5E,$6F
	.dc.b	$E8,$DB,$1E,$4F,$8A,$FA,$47,$D3,$F8,$BB,$1F,$A2,$E2,$FE,$F7,$8F
	.dc.b	$C4,$F4,$7E,$49,$27,$E4,$B6,$35,$18,$23,$82,$1F,$6E,$3F,$B2,$65
	.dc.b	$F9,$8C,$FD,$D3,$EB,$F5,$7F,$A9,$CF,$E3,$6F,$E2,$47,$E8,$FD,$1F
	.dc.b	$DF,$1F,$BF,$F3,$AC,$FD,$0F,$89,$FC,$77,$C4,$FE,$3D,$B1,$F3,$FB
	.dc.b	$3A,$F3,$76,$6C,$7E,$73,$D1,$FE,$15,$F2,$7F,$C2,$DB,$1F,$73,$E2
	.dc.b	$7F,$34,$7E,$6F,$E2,$AC,$7C,$9F,$43,$F1,$E4,$7F,$1E,$DE,$0F,$C7
	.dc.b	$B5,$FC,$F3,$5F,$CF,$35,$FC,$F3,$F2,$3A,$FE,$29,$F1,$B9,$ED,$7E
	.dc.b	$D3,$E5,$3F,$91,$1F,$C8,$FD,$3B,$1F,$94,$F3,$BF,$65,$F0,$DF,$B3
	.dc.b	$6C,$7E,$4F,$B3,$E8,$F2,$FD,$26,$C7,$F0,$0F,$53,$F9,$77,$A3,$FC
	.dc.b	$BD,$B1,$FA,$0F,$85,$FD,$F1,$FB,$FF,$85,$B1,$FC,$BB,$F1,$5F,$CD
	.dc.b	$70,$FF,$B9,$9A,$88,$E1,$BE,$B7,$FA,$1A,$31,$06,$92,$34,$49,$FF
	.dc.b	$47,$C6,$5F,$EB,$5D,$AE,$CF,$C7,$9F,$90,$E7,$B3,$FC,$7B,$D6,$FE
	.dc.b	$98,$7F,$4D,$F4,$6C,$7E,$FB,$E1,$7F,$8E,$7C,$37,$F1,$D6,$C7,$EF
	.dc.b	$7B,$3F,$37,$CF,$F9,$C6,$C7,$F2,$AF,$53,$FA,$C7,$9D,$FD,$65,$B1
	.dc.b	$FC,$2B,$D1,$FE,$50,$7E,$9B,$D6,$FF,$5F,$FE,$B9,$FE,$C1,$9F,$F7
	.dc.b	$EF,$9D,$FE,$D6,$7E,$77,$77,$C7,$23,$F0,$9B,$CB,$FE,$C0,$E4,$46
	.dc.b	$23,$EA,$35,$FA,$87,$CF,$70,$FD,$67,$F7,$7D,$DF,$B8,$3F,$73,$D3
	.dc.b	$6F,$EE,$79,$1F,$BE,$5F,$8D,$FF,$42,$C3,$F8,$F3,$63,$3E,$97,$BD
	.dc.b	$FC,$21,$C2,$70,$7F,$5D,$D3,$FC,$D5,$AF,$C4,$F8,$5C,$FB,$29,$CD
	.dc.b	$FE,$36,$31,$3F,$E6,$2E,$B7,$7F,$F1,$0E,$AF,$E2,$2D,$BF,$49,$E2
	.dc.b	$D1,$F2,$3E,$28,$F5,$FA,$56,$73,$1F,$E3,$43,$27,$F8,$DB,$2F,$27
	.dc.b	$C2,$F9,$B4,$FF,$37,$4E,$5F,$F5,$6B,$E7,$73,$B7,$FA,$9C,$FD,$07
	.dc.b	$17,$E8,$C9,$3F,$9E,$6C,$7D,$D6,$BF,$AF,$6B,$F7,$39,$11,$88,$FD
	.dc.b	$8B,$FC,$DB,$B7,$F9,$99,$FC,$D3,$A2,$DF,$C4,$AF,$B3,$EA,$7C,$13
	.dc.b	$E1,$7A,$96,$7D,$9F,$03,$EF,$F8,$2F,$80,$D8,$FE,$C5,$DB,$FE,$92
	.dc.b	$FD,$13,$EF,$E3,$FD,$7B,$A5,$CF,$FE,$06,$BF,$04,$2B,$7D,$5F,$8D
	.dc.b	$E4,$7A,$FF,$4F,$E4,$BE,$A3,$70,$F9,$DF,$07,$F4,$4F,$A5,$95,$8F
	.dc.b	$F2,$1F,$4F,$8B,$4F,$D0,$9F,$80,$3F,$AB,$98,$74,$FE,$99,$AF,$F1
	.dc.b	$B6,$BC,$63,$9B,$93,$99,$C4,$A6,$5D,$BC,$4E,$31,$D0,$EF,$A7,$C3
	.dc.b	$3E,$51,$F2,$DB,$3F,$5C,$FD,$8F,$C2,$7E,$5C,$FC,$C7,$AD,$63,$F2
	.dc.b	$5F,$05,$F9,$5F,$83,$FC,$B3,$63,$FC,$EB,$C1,$FE,$1D,$7E,$9A,$E5
	.dc.b	$FB,$5D,$BF,$6D,$FF,$6D,$CF,$F0,$0F,$67,$F4,$DE,$BF,$E9,$DB,$1F
	.dc.b	$A1,$F8,$1F,$D2,$1F,$A5,$F8,$2B,$1F,$53,$93,$E2,$F0,$FE,$C9,$BC
	.dc.b	$06,$92,$7E,$4C,$65,$B1,$99,$F9,$6D,$7E,$E3,$5F,$EA,$9C,$3F,$8F
	.dc.b	$EA,$FC,$B1,$F9,$7E,$5B,$7F,$37,$99,$EC,$FF,$03,$3F,$82,$7C,$75
	.dc.b	$9F,$DF,$7C,$0F,$F0,$1F,$81,$FE,$04,$D8,$F9,$BA,$7E,$8F,$A3,$F4
	.dc.b	$8D,$8E,$DF,$1F,$FC,$3F,$DA,$FE,$20,$D8,$FE,$1D,$AF,$F9,$1F,$CB
	.dc.b	$FF,$16,$F9,$08,$C4,$FF,$23,$7F,$D0,$F1,$9F,$5B,$F9,$6F,$17,$F1
	.dc.b	$EC,$3E,$69,$2F,$E5,$B6,$3F,$26,$D7,$F4,$8D,$7F,$48,$D7,$F4,$8F
	.dc.b	$E7,$B8,$BF,$75,$DB,$FC,$04,$FE,$05,$CF,$6F,$E6,$7B,$7F,$CB,$4F
	.dc.b	$E5,$DE,$CD,$8F,$7F,$BD,$FD,$03,$57,$BF,$63,$0F,$F4,$8C,$BF,$84
	.dc.b	$76,$7F,$0F,$E7,$FE,$20,$DC,$3E,$D7,$B7,$FC,$EF,$D8,$FE,$78,$D8
	.dc.b	$F9,$BE,$F7,$73,$C1,$EA,$D8,$FE,$13,$EB,$FE,$50,$8F,$E4,$9B,$C7
	.dc.b	$FB,$A6,$BF,$C2,$DA,$FF,$0B,$6B,$FC,$2D,$F9,$1E,$23,$B7,$F9,$41
	.dc.b	$FC,$A7,$96,$DF,$E4,$FD,$AF,$EA,$87,$F5,$5B,$B1,$F8,$6F,$77,$E4
	.dc.b	$7B,$DF,$25,$B1,$FB,$EE,$CF,$E6,$9C,$FF,$CD,$5B,$1F,$C2,$FF,$5F
	.dc.b	$FF,$09,$6B,$FE,$97,$5F,$1D,$FD,$95,$CC,$9C,$BF,$96,$F8,$0F,$B0
	.dc.b	$7D,$9F,$80,$B1,$FE,$F4,$FC,$65,$F1,$9F,$C0,$1B,$1F,$C9,$1B,$8F
	.dc.b	$F2,$C6,$BF,$CE,$1A,$FF,$86,$9A,$FF,$93,$DA,$9A,$49,$F0,$12,$7F
	.dc.b	$9C,$F4,$FF,$51,$ED,$FE,$B0,$7F,$59,$E7,$B3,$F1,$7E,$3B,$E1,$1F
	.dc.b	$0F,$E3,$AC,$7E,$DB,$DA,$FC,$C7,$C0,$FE,$65,$B1,$FC,$BB,$B3,$FB
	.dc.b	$27,$3F,$F6,$56,$C7,$D1,$F8,$1F,$9D,$EF,$FC,$F6,$C7,$F0,$2F,$80
	.dc.b	$FD,$21,$FA,$5F,$81,$B1,$F3,$7E,$33,$E6,$11,$F9,$8D,$E3,$F9,$8D
	.dc.b	$7E,$63,$5F,$98,$D7,$E6,$3F,$96,$EC,$FE,$F8,$7F,$AF,$BD,$BF,$F5
	.dc.b	$FF,$1B,$F4,$5F,$F6,$F6,$1F,$D2,$AB,$FA,$3F,$4F,$F8,$8B,$1F,$C6
	.dc.b	$73,$7E,$2F,$E2,$7F,$18,$FE,$FF,$E0,$BF,$80,$7C,$15,$C7,$F9,$0B
	.dc.b	$FB,$FC,$BF,$1F,$87,$F8,$2D,$73,$F8,$EB,$F9,$EC,$6B,$5F,$2B,$A8
	.dc.b	$FC,$F7,$C7,$7E,$6F,$E0,$BF,$38,$D8,$FE,$5F,$F0,$5F,$C4,$4F,$F2
	.dc.b	$8F,$E1,$1F,$E5,$DF,$4F,$F9,$84,$7F,$2C,$EF,$7B,$1F,$9B,$25,$FE
	.dc.b	$6D,$BA,$FE,$6D,$AF,$E6,$DA,$FE,$6D,$AF,$E6,$DF,$E5,$99,$BB,$BD
	.dc.b	$D2,$E5,$FA,$11,$87,$E9,$74,$DA,$E1,$87,$EF,$7E,$13,$F7,$47,$EE
	.dc.b	$FE,$12,$C7,$F9,$6F,$6F,$F0,$16,$4F,$5F,$FA,$7C,$7F,$80,$E7,$5F
	.dc.b	$03,$F1,$3E,$07,$EA,$77,$F7,$E9,$F0,$36,$BF,$86,$7C,$27,$F0,$9E
	.dc.b	$EF,$E1,$4D,$8F,$07,$AF,$F3,$8F,$9F,$EF,$D8,$FF,$65,$69,$FE,$36
	.dc.b	$FD,$D5,$EA,$FF,$82,$1B,$3F,$E2,$C6,$EF,$FE,$34,$6B,$FE,$40,$6B
	.dc.b	$FE,$A2,$6B,$FE,$D7,$6A,$69,$27,$C0,$49,$FE,$13,$C4,$F0,$3E,$81
	.dc.b	$F4,$7A,$EC,$FF,$2D,$F8,$BF,$E5,$47,$F2,$BF,$85,$B1,$F4,$7D,$FF
	.dc.b	$C7,$FC,$07,$E4,$1B,$1F,$08,$FC,$EF,$81,$F7,$7A,$7E,$F3,$63,$F9
	.dc.b	$D7,$C3,$7F,$37,$F8,$6F,$E7,$0D,$8F,$BD,$F0,$3F,$9C,$3F,$3B,$F0
	.dc.b	$36,$3B,$71,$7C,$3F,$8B,$F9,$C4,$8F,$9C,$D8,$F9,$CD,$7E,$73,$5F
	.dc.b	$9C,$D7,$A4,$7C,$8E,$EF,$CB,$1F,$97,$E8,$B1,$FD,$4F,$E1,$FF,$A8
	.dc.b	$1F,$D4,$7E,$16,$C7,$E5,$FD,$EF,$DA,$FB,$DF,$B6,$6C,$7A,$BC,$4F
	.dc.b	$E2,$9D,$DF,$A3,$E6,$FD,$23,$67,$FB,$07,$8D,$FE,$C1,$FB,$1F,$F1
	.dc.b	$5E,$DF,$D8,$63,$1F,$0D,$FC,$0B,$DD,$97,$FA,$7F,$7A,$DF,$FA,$EF
	.dc.b	$C0,$FF,$4E,$3F,$91,$F3,$3F,$2C,$4A,$FC,$B3,$63,$51,$82,$38,$21
	.dc.b	$F9,$98,$FE,$07,$DB,$FB,$D3,$F7,$DD,$16,$BF,$D0,$9A,$3E,$53,$D9
	.dc.b	$F1,$DE,$D6,$CF,$EF,$FD,$DF,$E3,$A7,$F1,$EF,$76,$C7,$C9,$F1,$1D
	.dc.b	$DF,$C2,$FA,$7F,$86,$36,$7B,$FC,$8F,$C9,$3E,$57,$C4,$D8,$FC,$5F
	.dc.b	$D1,$3F,$31,$F9,$AD,$D3,$FE,$00,$B9,$E6,$B3,$E9,$AF,$C9,$F1,$BF
	.dc.b	$1C,$FF,$61,$C7,$FA,$A9,$B1,$FE,$B2,$6B,$FE,$BE,$6A,$23,$86,$F9
	.dc.b	$9F,$E9,$28,$C4,$77,$8C,$12,$7C,$0D,$8F,$F6,$5F,$67,$F6,$1E,$FF
	.dc.b	$E6,$07,$F3,$1E,$BB,$7F,$53,$E3,$3F,$1D,$E6,$7E,$3D,$B1,$F1,$BF
	.dc.b	$38,$FE,$BB,$E7,$23,$D1,$BA,$73,$18,$7F,$AE,$67,$FC,$DE,$31,$F0
	.dc.b	$FF,$E6,$5F,$E0,$07,$0F,$F4,$57,$E6,$57,$FB,$CD,$CD,$7F,$BC,$9C
	.dc.b	$FB,$7E,$19,$F1,$76,$3E,$33,$F4,$C7,$E9,$FE,$2A,$E1,$FD,$27,$E0
	.dc.b	$B8,$6D,$FE,$96,$FC,$37,$FC,$97,$EF,$BD,$C5,$6A,$FE,$A4,$AB,$FC
	.dc.b	$A3,$9F,$DF,$F2,$FE,$A1,$2B,$F0,$0D,$8F,$C0,$35,$FC,$03,$5F,$C0
	.dc.b	$1C,$E3,$89,$1F,$B3,$B3,$C3,$F8,$C7,$C7,$EE,$B5,$FA,$CF,$8C,$FE
	.dc.b	$0D,$F1,$9F,$C1,$DB,$1F,$CA,$33,$F5,$72,$7B,$72,$46,$27,$5A,$7D
	.dc.b	$8D,$FE,$35,$FD,$7F,$03,$EB,$74,$7D,$76,$DF,$EF,$3E,$33,$F8,$B1
	.dc.b	$FC,$5F,$E5,$6C,$7E,$27,$E0,$BF,$57,$D4,$F8,$2B,$18,$7F,$D8,$7E
	.dc.b	$13,$E2,$FF,$40,$4B,$FD,$63,$78,$3F,$58,$D7,$F5,$8D,$7E,$6C,$F3
	.dc.b	$9E,$24,$8F,$9B,$DF,$F9,$13,$F2,$5D,$76,$7F,$8B,$FA,$DF,$CC,$7E
	.dc.b	$33,$F9,$93,$63,$F8,$4F,$8C,$E8,$FE,$0A,$7F,$06,$E1,$B3,$FC,$FF
	.dc.b	$A9,$DD,$F9,$FE,$9F,$D0,$36,$7F,$92,$FC,$6F,$F3,$F3,$FA,$07,$C6
	.dc.b	$58,$FA,$BF,$07,$F6,$7B,$3F,$29,$F0,$97,$B5,$36,$3F,$DD,$BD,$F8
	.dc.b	$BF,$C7,$F2,$7E,$42,$30,$E0,$FC,$75,$7E,$B5,$DC,$69,$B1,$97,$93
	.dc.b	$5F,$E3,$51,$B7,$33,$D5,$EA,$D8,$D4,$70,$9E,$4E,$8F,$D0,$F7,$FF
	.dc.b	$05,$3F,$83,$76,$5B,$FE,$97,$F1,$9F,$D9,$7E,$2F,$FB,$33,$63,$F3
	.dc.b	$7F,$0B,$F1,$4F,$8D,$F0,$B6,$3E,$A7,$81,$FC,$47,$A3,$F8,$93,$63
	.dc.b	$FA,$BF,$C5,$FF,$71,$3F,$B9,$7B,$16,$3F,$03,$E1,$FF,$44,$EE,$FE
	.dc.b	$8A,$D8,$FE,$8B,$C2,$F7,$FD,$AD,$8F,$85,$18,$23,$82,$70,$FD,$59
	.dc.b	$E0,$D9,$DB,$67,$6F,$F0,$CF,$41,$D6,$EF,$FE,$5C,$7F,$2F,$EA,$B7
	.dc.b	$FD,$47,$CC,$FE,$59,$F3,$FF,$C5,$23,$FC,$5E,$FF,$55,$C3,$F8,$43
	.dc.b	$9C,$7F,$9E,$07,$53,$FD,$1F,$8F,$F7,$E7,$F0,$0E,$3B,$1F,$A3,$EC
	.dc.b	$FE,$81,$CB,$FD,$05,$B1,$E3,$3F,$B2,$5F,$F6,$5F,$93,$F0,$E2,$6F
	.dc.b	$C5,$E1,$FD,$17,$C3,$7E,$8F,$E1,$EE,$1F,$C7,$1F,$D1,$FA,$BF,$48
	.dc.b	$91,$F9,$36,$E1,$F3,$DA,$FE,$89,$AF,$EA,$5A,$FE,$81,$FD,$16,$DF
	.dc.b	$43,$AF,$FA,$E1,$FD,$77,$A6,$DF,$B6,$FE,$41,$FE,$99,$3E,$5E,$9F
	.dc.b	$E0,$3F,$8A,$F7,$FC,$5F,$C0,$7F,$7D,$F0,$F2,$FF,$9B,$F4,$5C,$3E
	.dc.b	$9E,$7F,$80,$EC,$FF,$58,$7A,$5F,$E0,$DB,$FE,$D5,$18,$AF,$F5,$BE
	.dc.b	$7F,$E1,$1F,$67,$E6,$9F,$3B,$E2,$AD,$FF,$80,$78,$BF,$9C,$F1,$7F
	.dc.b	$3A,$6C,$7F,$88,$BF,$1E,$77,$FE,$E9,$B1,$FC,$05,$B9,$7F,$0F,$6B
	.dc.b	$FC,$51,$AF,$F6,$46,$BF,$E2,$56,$A6,$92,$7C,$04,$9F,$E3,$9A,$B5
	.dc.b	$FF,$6A,$47,$D1,$8C,$1F,$1B,$97,$E3,$BF,$13,$07,$C6,$7E,$43,$E3
	.dc.b	$3F,$22,$DB,$FE,$83,$E5,$FF,$91,$0F,$F2,$2F,$99,$63,$FC,$FD,$97
	.dc.b	$FB,$63,$BF,$FC,$31,$97,$8F,$9F,$EA,$F5,$FD,$67,$C1,$F1,$9F,$9E
	.dc.b	$3F,$3F,$F2,$57,$1F,$EE,$BF,$17,$FC,$FF,$E3,$3F,$A0,$36,$3E,$97
	.dc.b	$93,$F4,$48,$FD,$16,$F0,$FD,$16,$BF,$45,$AF,$D1,$6B,$F4,$5F,$CE
	.dc.b	$EA,$ED,$FC,$6F,$4F,$E3,$9B,$7F,$9D,$EA,$FF,$2E,$7C,$9F,$EE,$0F
	.dc.b	$DC,$DD,$9F,$F0,$7F,$C5,$7F,$59,$F8,$AF,$EB,$4D,$8F,$93,$DB,$F9
	.dc.b	$A3,$F3,$7C,$F6,$3B,$6A,$F8,$EF,$E0,$BF,$1D,$FC,$19,$B3,$FD,$DB
	.dc.b	$D1,$FF,$3C,$9F,$E7,$AF,$46,$C7,$E6,$7C,$5F,$CC,$11,$FC,$C3,$70
	.dc.b	$FC,$C3,$5F,$CC,$35,$FC,$C3,$5F,$CC,$3F,$23,$B7,$F6,$7C,$FF,$B4
	.dc.b	$6D,$7A,$BE,$3B,$F8,$E9,$FC,$7B,$E3,$6C,$7F,$9C,$79,$FF,$C5,$3D
	.dc.b	$1F,$E2,$A6,$C7,$E3,$BA,$7F,$7E,$7F,$00,$E5,$B1,$F1,$BE,$37,$F9
	.dc.b	$4F,$C9,$7F,$2A,$6C,$7F,$90,$3E,$33,$FC,$14,$7F,$82,$FE,$36,$C7
	.dc.b	$FA,$7B,$F1,$EF,$C0,$3F,$DE,$DB,$C3,$FE,$09,$6B,$FE,$0E,$6B,$FE
	.dc.b	$5E,$6B,$FE,$9F,$6A,$69,$3B,$3F,$C0,$6D,$9F,$04,$7E,$4B,$AF,$F8
	.dc.b	$EF,$37,$F1,$E6,$CF,$E1,$3E,$0B,$FA,$39,$FD,$23,$E2,$2C,$7F,$97
	.dc.b	$BD,$0F,$F1,$47,$C6,$FF,$8A,$5B,$1F,$B8,$EB,$FE,$52,$7F,$2A,$E6
	.dc.b	$B1,$F9,$DF,$1B,$FD,$44,$FF,$66,$7F,$D4,$3B,$7F,$63,$72,$FF,$3C
	.dc.b	$73,$F1,$BF,$EA,$7F,$03,$FC,$C6,$7F,$99,$3C,$1B,$7F,$3F,$9B,$E0
	.dc.b	$91,$FA,$0D,$C3,$E8,$35,$FA,$0D,$7E,$83,$5F,$A0,$FE,$EB,$AB,$FA
	.dc.b	$2F,$37,$F4,$66,$D7,$B7,$E4,$BF,$BF,$E7,$FE,$83,$8F,$EE,$6B,$E2
	.dc.b	$33,$3F,$EE,$4F,$7B,$FD,$CB,$ED,$DB,$FF,$32,$BF,$E8,$FF,$2F,$F9
	.dc.b	$97,$67,$F5,$43,$FA,$AF,$35,$AF,$99,$F1,$1E,$0F,$80,$F0,$B6,$3F
	.dc.b	$D3,$5F,$87,$FF,$B1,$38,$7F,$C2,$3F,$08,$8C,$7C,$67,$FB,$16,$31
	.dc.b	$F2,$1F,$E6,$79,$CD,$FE,$4F,$3F,$C6,$91,$FC,$E3,$70,$FC,$E3,$5F
	.dc.b	$CE,$35,$E5,$1C,$08,$FF,$9B,$BC,$D5,$BC,$5F,$ED,$EB,$C6,$B1,$87
	.dc.b	$FD,$B9,$C9,$FE,$71,$F0,$D9,$7F,$AD,$FA,$DF,$AC,$3C,$5F,$F5,$D6
	.dc.b	$9F,$A9,$AD,$4E,$75,$FE,$A3,$E1,$7D,$1F,$0F,$FD,$BB,$0B,$FF,$1C
	.dc.b	$F1,$7F,$8E,$7A,$F5,$7F,$5F,$EB,$7F,$33,$B7,$F5,$86,$E1,$FE,$43
	.dc.b	$5F,$1F,$E8,$8C,$1F,$E0,$C9,$7C,$47,$E2,$BE,$13,$F1,$6D,$9F,$16
	.dc.b	$5F,$EF,$4F,$C1,$FF,$52,$9A,$7E,$47,$44,$67,$1F,$F6,$D7,$C9,$BD
	.dc.b	$27,$A5,$E3,$B4,$E6,$30,$EE,$69,$C3,$1F,$F0,$26,$BD,$FB,$CD,$3F
	.dc.b	$41,$1E,$83,$E1,$78,$5F,$EE,$BC,$FF,$C0,$FE,$05,$F4,$FE,$7E,$9F
	.dc.b	$39,$FF,$69,$37,$8B,$E7,$78,$1E,$7C,$FB,$F8,$37,$4F,$F7,$36,$9E
	.dc.b	$48,$F9,$9F,$0D,$FB,$43,$F6,$BE,$FD,$9F,$F6,$9F,$C3,$FE,$1F,$E3
	.dc.b	$9F,$0F,$63,$0F,$E2,$3F,$61,$EB,$B7,$FB,$05,$AF,$FB,$19,$AF,$E1
	.dc.b	$C3,$87,$F1,$4D,$73,$AF,$FA,$EA,$30,$47,$04,$C3,$FD,$89,$1C,$F5
	.dc.b	$F9,$FD,$1F,$04,$F8,$5D,$36,$BF,$98,$7C,$27,$F1,$4F,$87,$FE,$2A
	.dc.b	$D8,$FA,$99,$3C,$27,$C9,$7E,$8E,$3F,$D2,$F1,$8F,$0F,$F5,$92,$F5
	.dc.b	$2C,$7E,$87,$8F,$E9,$71,$FD,$36,$E3,$FD,$07,$E2,$3F,$92,$9F,$C9
	.dc.b	$BC,$8B,$1E,$2E,$B7,$EE,$5F,$95,$8C,$3F,$8F,$D7,$F6,$2E,$5B,$E0
	.dc.b	$7D,$5E,$0F,$94,$4A,$F8,$ED,$9F,$B0,$D7,$EC,$35,$FB,$1C,$88,$C4
	.dc.b	$7D,$87,$FD,$2F,$D4,$E2,$FC,$81,$F9,$18,$B7,$FD,$7F,$C9,$FE,$86
	.dc.b	$7F,$44,$F5,$2C,$7B,$FD,$DF,$F4,$76,$3F,$C3,$1F,$F3,$7B,$4F,$D3
	.dc.b	$38,$D7,$F8,$66,$FF,$9F,$DF,$F4,$0D,$BF,$ED,$DE,$AF,$F5,$2F,$6F
	.dc.b	$FA,$9B,$63,$EF,$F3,$FD,$EF,$67,$ED,$9F,$C9,$FD,$8F,$E7,$FA,$7E
	.dc.b	$0F,$A5,$3F,$A1,$E8,$FC,$B1,$30,$FC,$9B,$6B,$F4,$8D,$7F,$48,$D7
	.dc.b	$E6,$B5,$FD,$1B,$F3,$38,$3F,$75,$C1,$FB,$B6,$D7,$F8,$26,$7F,$B6
	.dc.b	$F4,$FF,$6E,$6C,$7B,$3C,$07,$C5,$7E,$EC,$FD,$E7,$BF,$67,$F3,$F9
	.dc.b	$3C,$7F,$E9,$F1,$FB,$78,$FD,$C4,$61,$C1,$BF,$49,$38,$8C,$36,$78
	.dc.b	$47,$36,$BD,$27,$8B,$D2,$32,$FF,$4D,$6B,$D0,$D4,$70,$7A,$35,$FE
	.dc.b	$1D,$D8,$F9,$1F,$E6,$FF,$54,$F8,$7F,$E7,$95,$FA,$C9,$CC,$7F,$27
	.dc.b	$E5,$FE,$B1,$CD,$FD,$65,$B9,$FE,$CF,$F2,$4F,$E7,$38,$7F,$60,$7F
	.dc.b	$90,$F5,$7A,$83,$A7,$0E,$A7,$FB,$07,$0F,$F3,$53,$F9,$B7,$15,$8F
	.dc.b	$E9,$38,$78,$30,$FF,$79,$F7,$23,$18,$F4,$F7,$3F,$A4,$65,$D6,$FF
	.dc.b	$AE,$BE,$1C,$3F,$B2,$B9,$F1,$BF,$A9,$F4,$7F,$45,$3F,$A4,$7F,$00
	.dc.b	$7F,$88,$FD,$3F,$EE,$D1,$FD,$C4,$F6,$5F,$F7,$FF,$67,$F9,$27,$A5
	.dc.b	$FE,$3C,$9C,$E4,$9C,$D7,$C7,$F2,$7C,$A7,$9B,$C8,$B1,$F9,$0E,$97
	.dc.b	$CA,$7D,$9C,$3F,$00,$D7,$F2,$8D,$8D,$CF,$E3,$63,$04,$70,$49,$FC
	.dc.b	$64,$65,$B1,$F9,$5F,$8D,$E8,$47,$04,$8F,$CA,$46,$5B,$19,$7F,$39
	.dc.b	$19,$6C,$7E,$87,$E4,$FD,$04,$70,$48,$FD,$04,$65,$B1,$97,$F5,$11
	.dc.b	$96,$C7,$EB,$7E,$57,$E2,$11,$C1,$23,$F5,$91,$96,$C6,$5F,$DA,$46
	.dc.b	$5B,$1F,$B9,$F8,$EF,$A0,$47,$04,$8F,$DC,$46,$5B,$19,$7F,$80,$46
	.dc.b	$5B,$1F,$C1,$7A,$BE,$E1,$1C,$12,$3F,$82,$46,$5B,$19,$7F,$86,$46
	.dc.b	$5B,$1F,$C4,$7D,$7F,$C4,$11,$C1,$23,$F8,$84,$65,$B1,$97,$F8,$C4
	.dc.b	$65,$B1,$FC,$77,$CE,$FC,$A1,$1C,$12,$3F,$8E,$46,$5B,$19,$7F,$92
	.dc.b	$46,$5B,$1F,$CA,$7E,$33,$F4,$04,$70,$48,$FE,$51,$19,$6C,$65,$FE
	.dc.b	$61,$19,$6C,$7F,$35,$F8,$3F,$D6,$11,$C1,$23,$F9,$A4,$65,$B1,$97
	.dc.b	$F9,$E4,$65,$B1,$FD,$07,$E0,$FF,$70,$47,$04,$8F,$E8,$11,$96,$C6
	.dc.b	$5F,$E9,$11,$96,$C7,$F4,$DF,$5B,$F8,$21,$1C,$12,$3F,$A6,$46,$5B
	.dc.b	$19,$7F,$AA,$46,$5B,$1F,$D6,$7D,$6F,$E2,$04,$70,$48,$FE,$B1,$19
	.dc.b	$6C,$65,$FE,$C1,$19,$6C,$7F,$65,$F9,$2F,$E3,$84,$70,$48,$FE,$C9
	.dc.b	$19,$6C,$65,$FE,$D9,$19,$6C,$7F,$71,$F2,$3F,$94,$11,$C1,$23,$FB
	.dc.b	$84,$65,$B1,$97,$FB,$C4,$65,$B1,$FD,$F7,$C8,$FE,$68,$47,$04,$8F
	.dc.b	$EF,$91,$96,$C6,$5F,$F0,$24,$65,$B1,$FE,$09,$F9,$2F,$E8,$04,$70
	.dc.b	$48,$FF,$04,$46,$5B,$19,$7F,$C2,$11,$96,$C7,$F8,$5F,$F2,$3F,$F2
	.dc.b	$7C,$7F,$87,$DA,$FF,$A5,$DA,$FF,$B7,$DA,$9A,$47,$FA,$C3,$63,$B4
	.dc.b	$60,$8F,$87,$BB,$64,$70,$48,$FF,$17,$CF,$F9,$89,$F3,$4E,$DC,$B5
	.dc.b	$CC,$F1,$7D,$9F,$25,$E4,$24,$7C,$16,$C7,$66,$BC,$23,$44,$74,$D7
	.dc.b	$E4,$7C,$97,$C0,$24,$7C,$D6,$C7,$A9,$AF,$08,$D1,$1D,$35,$FF,$64
	.dc.b	$53,$C0,$FF,$29,$78,$1F,$E4,$88,$C7,$CA,$FF,$86,$BC,$DF,$F0,$DB
	.dc.b	$FE,$3E,$CD,$F2,$7F,$E1,$6B,$7E,$EA,$F8,$8F,$F0,$CE,$7F,$6E,$B5
	.dc.b	$39,$D3,$FC,$CB,$87,$27,$FD,$69,$FC,$02,$EF,$E3,$39,$C1,$E4,$FF
	.dc.b	$90,$EB,$57,$FE,$CC,$DB,$D5,$E7,$FF,$8C,$3C,$2F,$F1,$8B,$65,$83
	.dc.b	$6F,$F2,$F6,$4F,$91,$F5,$7C,$27,$F9,$DE,$FC,$B1,$8F,$17,$C6,$EB
	.dc.b	$CC,$B3,$FE,$DC,$C1,$7F,$E4,$5C,$7F,$C7,$1C,$9F,$ED,$D9,$FF,$39
	.dc.b	$F9,$28,$C5,$FF,$90,$57,$A3,$FE,$78,$AD,$C6,$AD,$BF,$D7,$E2,$FF
	.dc.b	$22,$B7,$0F,$85,$CE,$F2,$32,$FC,$6B,$F9,$33,$F0,$AB,$F0,$1E,$46
	.dc.b	$8D,$F9,$16,$79,$F9,$1F,$E5,$A3,$FC,$B7,$E4,$58,$F9,$7A,$FC,$AF
	.dc.b	$95,$F9,$3F,$05,$F6,$1B,$3F,$79,$F9,$BE,$66,$63,$E4,$71,$A3,$CF
	.dc.b	$E5,$FF,$9A,$3C,$BF,$F3,$4B,$6B,$B9,$FC,$27,$E5,$7F,$0A,$7E,$3B
	.dc.b	$E8,$5F,$09,$89,$FC,$6C,$DE,$2F,$82,$7E,$0A,$BF,$DD,$93,$A3,$E4
	.dc.b	$F9,$DF,$E7,$8F,$3B,$FC,$F2,$D8,$F0,$D7,$F1,$7D,$3F,$0F,$ED,$7E
	.dc.b	$59,$B3,$F9,$C7,$F1,$7E,$DE,$63,$DE,$FF,$9F,$3B,$3F,$05,$E9,$7D
	.dc.b	$F9,$FC,$13,$70,$FC,$4F,$AF,$6B,$75,$E2,$E9,$FD,$C7,$A5,$F7,$8F
	.dc.b	$BF,$E9,$59,$FC,$1F,$A5,$F8,$2F,$4B,$F0,$6D,$8F,$C3,$FA,$5F,$E9
	.dc.b	$63,$FD,$2F,$E9,$58,$F7,$E9,$FC,$CF,$E5,$BF,$66,$7C,$FF,$16,$CF
	.dc.b	$EE,$A2,$CF,$CE,$1C,$AF,$67,$3E,$77,$E8,$BE,$C3,$FD,$49,$87,$FA
	.dc.b	$85,$FE,$22,$DF,$5F,$E0,$46,$30,$FD,$0F,$E9,$BF,$BA,$AF,$E0,$0F
	.dc.b	$E9,$1B,$1F,$A4,$8C,$3F,$80,$E2,$43,$D7,$D1,$BF,$5E,$DF,$4F,$57
	.dc.b	$E9,$5F,$06,$DF,$F0,$79,$FA,$4F,$F1,$5F,$17,$F8,$5F,$C2,$7F,$18
	.dc.b	$3F,$29,$D5,$6B,$F8,$FC,$5C,$FE,$37,$D1,$36,$C6,$FF,$CC,$72,$78
	.dc.b	$DE,$03,$C3,$E1,$5C,$BE,$AF,$A6,$F9,$9F,$E4,$E7,$F3,$0C,$9E,$D6
	.dc.b	$2B,$F9,$7C,$DD,$3C,$3E,$B5,$D3,$C1,$E1,$FB,$8F,$7F,$4D,$C3,$BF
	.dc.b	$CB,$FF,$3F,$FC,$29,$E8,$FF,$36,$8C,$36,$7F,$92,$4B,$D1,$C6,$7F
	.dc.b	$B2,$E6,$4E,$6B,$F9,$AE,$7F,$85,$DB,$F0,$3A,$FE,$09,$B3,$F4,$39
	.dc.b	$7E,$F6,$FF,$7D,$B1,$F8,$4D,$7F,$05,$F2,$7F,$83,$6C,$7E,$23,$E6
	.dc.b	$BE,$D9,$F7,$3E,$6A,$C7,$CC,$AF,$77,$A7,$FD,$49,$B1,$FD,$61,$F8
	.dc.b	$5E,$86,$63,$DE,$FE,$47,$E2,$3F,$0F,$F2,$BF,$88,$6C,$FD,$8C,$7F
	.dc.b	$AE,$9F,$5E,$AC,$7F,$64,$8B,$C5,$F8,$6F,$4E,$F0,$FE,$17,$E6,$BF
	.dc.b	$27,$F0,$5F,$94,$6E,$1F,$84,$DB,$FB,$49,$F8,$3F,$52,$C7,$F7,$08
	.dc.b	$B8,$FE,$4B,$CF,$B8,$FE,$47,$E5,$3F,$3F,$F2,$9F,$A0,$6E,$1F,$90
	.dc.b	$DF,$F0,$5E,$77,$F7,$66,$C7,$F7,$C7,$F0,$DE,$76,$63,$F0,$6F,$F5
	.dc.b	$8F,$59,$1F,$AE,$5F,$2B,$63,$0F,$F8,$3F,$93,$FC,$06,$7E,$76,$AC
	.dc.b	$FF,$82,$A2,$EB,$FC,$47,$E8,$3F,$83,$3F,$C4,$FC,$44,$F8,$7E,$0F
	.dc.b	$C1,$F0,$5E,$16,$E1,$FA,$CE,$4F,$F0,$89,$FE,$1A,$C7,$F8,$44,$7F
	.dc.b	$86,$23,$2D,$3F,$32,$7F,$83,$B1,$4B,$CE,$DC,$7F,$13,$D5,$F2,$BF
	.dc.b	$04,$F8,$5F,$2D,$70,$FD,$EF,$47,$F8,$88,$F1,$7C,$2D,$8F,$F1,$4C
	.dc.b	$5C,$3E,$07,$99,$79,$FF,$C7,$59,$3E,$03,$E7,$1F,$3F,$E0,$2E,$3F
	.dc.b	$C2,$BA,$3F,$C5,$E7,$BF,$CC,$B1,$FE,$36,$8B,$97,$CD,$F3,$2E,$5D
	.dc.b	$BE,$27,$ED,$1F,$6F,$E0,$6E,$1F,$C5,$BB,$7F,$C7,$67,$CB,$F2,$2C
	.dc.b	$7F,$90,$A2,$E3,$F6,$7C,$7B,$8F,$AB,$E5,$7F,$0C,$7E,$1F,$E5,$6E
	.dc.b	$1F,$D5,$BF,$26,$FC,$5B,$F4,$B0,$FF,$1F,$C6,$1F,$C2,$78,$FF,$E5
	.dc.b	$48,$C6,$1F,$92,$9D,$D6,$4F,$C0,$78,$B5,$FF,$27,$72,$3E,$57,$F3
	.dc.b	$07,$E6,$7E,$5A,$CF,$8F,$C4,$EC,$7E,$27,$C2,$B1,$E5,$C5,$9F,$E8
	.dc.b	$23,$0D,$8F,$CD,$78,$1C,$1F,$C4,$7C,$37,$E9,$FE,$1B,$F5,$0D,$BF
	.dc.b	$85,$E1,$FF,$9C,$FF,$0B,$F6,$C7,$E4,$F1,$FF,$2E,$61,$FC,$C6,$72
	.dc.b	$FC,$6A,$E1,$9F,$C5,$6D,$FC,$0F,$91,$FD,$B1,$FB,$7F,$83,$B1,$F4
	.dc.b	$7C,$7F,$98,$7F,$A0,$F0,$F5,$71,$FC,$D7,$DF,$1A,$9F,$7B,$BC,$31
	.dc.b	$3F,$A5,$F8,$2F,$E0,$9F,$11,$FC,$15,$B7,$F7,$FC,$7F,$B4,$77,$2B
	.dc.b	$F2,$9B,$5F,$D3,$C7,$8F,$8C,$BF,$4B,$EF,$FF,$7E,$FE,$01,$FF,$09
	.dc.b	$D7,$F9,$BB,$A1,$39,$F8,$0F,$F1,$A4,$FD,$8C,$3E,$77,$83,$F8,$B6
	.dc.b	$E7,$F6,$9F,$A5,$DD,$98,$FF,$43,$72,$7F,$94,$2F,$FC,$29,$9F,$F8
	.dc.b	$77,$E0,$8D,$3F,$BC,$D7,$F8,$FD,$FF,$11,$4F,$E2,$3C,$8F,$CD,$1F
	.dc.b	$3F,$0B,$5F,$8A,$8B,$3F,$07,$C0,$BB,$FF,$AF,$70,$7B,$5F,$04,$F8
	.dc.b	$5F,$1F,$71,$FD,$67,$95,$F8,$BE,$DF,$D4,$36,$3F,$2E,$FE,$3F,$AB
	.dc.b	$31,$EE,$7F,$D2,$B8,$7F,$A4,$7D,$DF,$07,$C6,$FC,$FF,$8E,$FA,$0F
	.dc.b	$CC,$EF,$79,$5F,$B5,$3F,$15,$85,$C7,$F4,$B1,$78,$3E,$E7,$7D,$E0
	.dc.b	$F6,$7C,$0F,$DB,$F8,$1F,$B8,$DC,$3F,$83,$74,$FF,$01,$3F,$DB,$1B
	.dc.b	$7C,$28,$FD,$8C,$65,$A7,$FA,$F6,$1C,$59,$CB,$EC,$FC,$6F,$E2,$4F
	.dc.b	$C5,$7B,$F6,$FF,$8A,$74,$7F,$B9,$73,$E3,$1F,$BE,$8C,$61,$FC,$4D
	.dc.b	$D7,$15,$9F,$F6,$65,$7F,$AE,$75,$FB,$1E,$F7,$E5,$FD,$EF,$CC,$36
	.dc.b	$BF,$91,$74,$F7,$E0,$FE,$3C,$D8,$FE,$1A,$F8,$F0,$CC,$76,$7E,$BF
	.dc.b	$81,$FA,$3F,$7B,$F4,$8D,$9F,$E5,$DE,$67,$F2,$93,$E4,$E5,$63,$F8
	.dc.b	$CC,$5E,$1F,$F1,$F7,$12,$F1,$AF,$E4,$7D,$EF,$D8,$7C,$67,$EC,$5B
	.dc.b	$8F,$F3,$BF,$3B,$F9,$B1,$F5,$A2,$C7,$F2,$68,$B8,$F6,$AB,$7F,$9D
	.dc.b	$F8,$BF,$DE,$7C,$67,$EF,$5B,$87,$F4,$7F,$3B,$FA,$11,$F8,$29,$B1
	.dc.b	$FC,$CA,$2E,$1E,$AF,$81,$B8,$7E,$A5,$FE,$11,$F0,$1F,$C2,$5B,$7F
	.dc.b	$D4,$FC,$3F,$E9,$C7,$F4,$DA,$FF,$7E,$7C,$C7,$F8,$51,$FE,$49,$E0
	.dc.b	$A7,$75,$FD,$09,$E0,$FF,$BB,$AB,$FD,$37,$5C,$27,$FE,$00,$FE,$9D
	.dc.b	$FE,$1E,$E2,$FE,$B5,$E4,$23,$FD,$43,$18,$75,$87,$FC,$01,$72,$FF
	.dc.b	$45,$F0,$FF,$B2,$1E,$0D,$54,$CB,$FC,$35,$5F,$D5,$AB,$6D,$3F,$B8
	.dc.b	$4E,$FE,$2F,$37,$FE,$B1,$C3,$FC,$52,$F4,$AF,$17,$C0,$FF,$AB,$A7
	.dc.b	$37,$E7,$8F,$E0,$9E,$27,$F7,$73,$C9,$1E,$3F,$7F,$17,$FD,$A2,$6E
	.dc.b	$1F,$E8,$DC,$DF,$15,$8D,$FE,$16,$9D,$FE,$0B,$C0,$DC,$7F,$C1,$5E
	.dc.b	$9F,$6F,$9C,$E3,$1F,$DD,$23,$18,$7F,$82,$5D,$7C,$15,$9F,$83,$48
	.dc.b	$FA,$BA,$FB,$4F,$75,$DA,$FF,$0C,$7A,$DF,$E1,$33,$DB,$A3,$DE,$C4
	.dc.b	$FF,$81,$26,$ED,$ED,$F8,$7B,$BF,$DD,$CF,$E5,$1F,$2F,$E0,$AE,$1F
	.dc.b	$E2,$5F,$63,$D9,$7F,$E2,$16,$C7,$F8,$41,$F8,$1F,$01,$98,$F5,$BF
	.dc.b	$8E,$F8,$8F,$AD,$F0,$7F,$5D,$B2,$CF,$EE,$17,$F2,$8F,$F6,$E4,$7F
	.dc.b	$87,$DB,$1F,$57,$E0,$7F,$1C,$A3,$3A,$FF,$81,$DB,$5F,$E6,$B6,$CF
	.dc.b	$E5,$38,$1F,$0D,$CA,$73,$DB,$FA,$1E,$2F,$C6,$6F,$F8,$D6,$E3,$E7
	.dc.b	$EA,$ED,$F2,$7C,$4F,$C6,$E9,$FE,$74,$9F,$44,$65,$A7,$E7,$A3,$FC
	.dc.b	$16,$E3,$4F,$CF,$70,$7E,$6F,$E1,$7F,$38,$D9,$F0,$1E,$BF,$8B,$F4
	.dc.b	$7B,$BE,$96,$C7,$C7,$7F,$25,$F0,$F9,$8F,$C6,$BF,$C0,$3E,$2F,$F5
	.dc.b	$1F,$11,$FA,$96,$CF,$6C,$BE,$96,$1D,$16,$7F,$37,$1C,$3F,$3E,$6D
	.dc.b	$7C,$58,$FE,$E6,$FD,$98,$FE,$1F,$C9,$FB,$7E,$3F,$DC,$36,$7C,$18
	.dc.b	$7D,$BF,$67,$EA,$7B,$1F,$55,$B3,$F8,$37,$F7,$7F,$87,$7F,$91,$A3
	.dc.b	$EF,$46,$3E,$27,$F8,$EF,$C5,$7F,$05,$E0,$FE,$0C,$DA,$EF,$D0,$C7
	.dc.b	$FC,$F5,$F7,$BF,$E1,$CC,$FF,$25,$EF,$63,$7F,$C5,$55,$9C,$E1,$F9
	.dc.b	$9B,$E1,$1F,$93,$9F,$C6,$7B,$9E,$88,$F9,$35,$F9,$38,$FE,$4D,$D6
	.dc.b	$FD,$3B,$FD,$1D,$D5,$FC,$C9,$FF,$3D,$36,$BF,$D1,$CE,$77,$FF,$25
	.dc.b	$EB,$E9,$9F,$16,$9F,$9F,$D7,$CD,$9F,$8E,$72,$FA,$06,$BE,$8B,$69
	.dc.b	$E8,$AF,$D8,$6B,$6F,$D2,$35,$F4,$7C,$5F,$92,$BF,$75,$1E,$BF,$8C
	.dc.b	$C5,$6F,$7F,$F4,$4E,$FE,$D9,$F4,$EB,$E6,$D0,$DC,$F0,$E7,$DE,$FD
	.dc.b	$B3,$FB,$76,$E7,$FE,$B7,$CF,$E5,$7C,$67,$81,$F9,$9F,$17,$74,$FD
	.dc.b	$A6,$BF,$C3,$A7,$37,$F8,$2C,$7B,$BF,$1A,$7E,$41,$F1,$B8,$7E,$64
	.dc.b	$67,$4F,$3B,$FC,$56,$BC,$8D,$3C,$98,$7F,$BA,$5F,$D6,$61,$F7,$FE
	.dc.b	$2B,$C9,$F1,$5E,$57,$ED,$FC,$27,$E3,$78,$5D,$2A,$F8,$D3,$98,$C6
	.dc.b	$1E,$38,$FD,$1F,$C4,$7F,$1B,$6C,$78,$A3,$E1,$1F,$0F,$E2,$2C,$7E
	.dc.b	$8B,$E1,$3E,$91,$E1,$EF,$7A,$58,$9F,$81,$3F,$C9,$CF,$E3,$FE,$7D
	.dc.b	$8F,$7F,$C3,$FC,$FF,$87,$FA,$0D,$8F,$D7,$FC,$2F,$DF,$F8,$7F,$C0
	.dc.b	$36,$3E,$63,$FB,$CF,$87,$FE,$68,$D8,$F9,$71,$F6,$FE,$1B,$EE,$36
	.dc.b	$3F,$77,$5F,$17,$E0,$FA,$75,$7D,$93,$FC,$27,$C4,$8C,$EF,$FA,$C8
	.dc.b	$FF,$15,$5F,$F4,$FB,$FE,$1B,$F0,$1F,$B6,$E8,$E1,$3F,$69,$DE,$3F
	.dc.b	$8B,$72,$D9,$FD,$AF,$C3,$FE,$77,$E1,$FF,$3C,$DB,$FC,$53,$FC,$6B
	.dc.b	$E0,$BF,$A9,$36,$3F,$13,$1F,$99,$3F,$35,$F0,$16,$3F,$8C,$D7,$3C
	.dc.b	$3D,$C7,$F6,$3B,$C7,$8F,$FB,$D9,$FE,$BC,$7F,$88,$BE,$31,$F1,$58
	.dc.b	$C3,$F2,$FD,$1F,$A7,$E7,$FD,$43,$63,$B9,$FA,$5F,$CA,$5E,$47,$F0
	.dc.b	$3D,$3C,$3E,$1B,$3F,$E2,$35,$AF,$0B,$F5,$1F,$13,$FC,$E3,$2F,$EE
	.dc.b	$0D,$CB,$FC,$05,$CC,$7F,$C0,$9F,$2C,$F9,$5F,$F5,$0E,$9F,$E0,$68
	.dc.b	$C7,$97,$FC,$39,$D7,$A3,$6B,$FD,$91,$7F,$E0,$DB,$69,$E3,$7E,$3B
	.dc.b	$78,$FE,$34,$FB,$78,$3D,$91,$FD,$0F,$6F,$E6,$DE,$FF,$F3,$8E,$EF
	.dc.b	$F0,$3F,$57,$FB,$FB,$0F,$E2,$51,$FE,$0E,$D7,$D1,$CB,$E3,$F3,$FC
	.dc.b	$8D,$AF,$EB,$FA,$78,$7C,$BF,$33,$76,$FF,$11,$72,$FC,$AD,$FC,$75
	.dc.b	$FC,$BB,$37,$B5,$E3,$3C,$9E,$B5,$CB,$E5,$9F,$2B,$89,$E8,$7C,$5F
	.dc.b	$8D,$F8,$CD,$9F,$1E,$5D,$B4,$FF,$1A,$6B,$E3,$7F,$C5,$DA,$F8,$F8
	.dc.b	$FF,$A5,$8D,$7F,$DC,$53,$FD,$1A,$31,$B7,$C7,$E5,$F9,$0F,$FA,$17
	.dc.b	$8B,$EE,$79,$7F,$5C,$FB,$1E,$7D,$D7,$C7,$87,$D3,$3F,$98,$67,$E1
	.dc.b	$CB,$F7,$91,$E1,$E8,$FA,$9C,$9F,$55,$B7,$F0,$A9,$97,$E2,$FC,$7F
	.dc.b	$C3,$1F,$87,$F2,$AD,$78,$2B,$F0,$BC,$3F,$0D,$F8,$31,$E3,$7C,$3C
	.dc.b	$1F,$DA,$23,$F0,$3C,$5F,$85,$E7,$FC,$33,$F8,$7E,$4F,$EA,$95,$FD
	.dc.b	$22,$FF,$A9,$28,$CF,$07,$F6,$89,$FC,$BE,$3E,$1E,$4F,$A5,$DD,$FA
	.dc.b	$DC,$FC,$5D,$7F,$DE,$7E,$2F,$FC,$B4,$7F,$A7,$BE,$93,$FD,$5F,$8F
	.dc.b	$F7,$83,$FC,$CD,$1F,$99,$E6,$F1,$9E,$4E,$0B,$87,$C1,$C3,$FC,$37
	.dc.b	$97,$83,$9B,$E8,$47,$F8,$1F,$E3,$7F,$CD,$07,$82,$BF,$CE,$CF,$E9
	.dc.b	$78,$FD,$FE,$47,$C0,$6C,$F5,$E9,$FC,$FF,$07,$CD,$E0,$F6,$6F,$F4
	.dc.b	$3A,$7F,$B9,$7C,$5F,$FB,$7F,$0F,$DD,$36,$7F,$D0,$D9,$FF,$A0,$A3
	.dc.b	$15,$F9,$3F,$1B,$E7,$ED,$F4,$1B,$87,$87,$0F,$CE,$E5,$E1,$AF,$6E
	.dc.b	$7D,$B0,$E8,$7C,$BF,$88,$B8,$7E,$FB,$E3,$BE,$D9,$F7,$3C,$5B,$1D
	.dc.b	$72,$FB,$37,$F9,$88,$FA,$F8,$7A,$BE,$5B,$D0,$74,$AF,$AD,$AF,$FB
	.dc.b	$7B,$C6,$FF,$79,$F5,$BA,$3F,$13,$C3,$F8,$A6,$E5,$DF,$9B,$EB,$3D
	.dc.b	$B1,$FF,$56,$62,$F8,$DF,$F2,$1B,$67,$E6,$E7,$89,$FF,$2E,$6A,$FD
	.dc.b	$3B,$FD,$FF,$87,$F8,$FD,$FF,$7F,$B4,$FF,$17,$38,$7D,$E7,$D9,$3F
	.dc.b	$A1,$F1,$7C,$C7,$C7,$C5,$E2,$3C,$18,$FD,$E7,$C3,$5F,$EF,$A8,$FF
	.dc.b	$25,$B6,$7C,$59,$78,$32,$FD,$36,$7F,$C5,$4F,$83,$3E,$1E,$8F,$F3
	.dc.b	$9D,$7F,$BC,$A3,$FB,$2B,$63,$FC,$FD,$EB,$BB,$3F,$DE,$19,$7E,$43
	.dc.b	$E3,$3C,$7F,$1D,$EE,$3D,$FC,$16,$7E,$87,$47,$E2,$5F,$A1,$9B,$3E
	.dc.b	$6F,$83,$E3,$7E,$8E,$3F,$49,$F7,$F0,$7C,$A8,$F9,$6D,$BE,$BC,$5F
	.dc.b	$89,$BF,$67,$3F,$E8,$BE,$37,$ED,$9C,$EB,$F4,$EF,$CA,$C7,$EA,$9F
	.dc.b	$59,$B1,$D2,$BE,$AB,$E8,$C9,$F1,$9F,$86,$3F,$CA,$5B,$BE,$27,$15
	.dc.b	$FD,$03,$97,$6F,$F9,$71,$B8,$7F,$9B,$57,$C6,$FE,$1B,$0F,$AF,$C1
	.dc.b	$FC,$6C,$F9,$15,$EC,$3F,$2F,$3F,$87,$7E,$53,$F8,$9F,$23,$FC,$05
	.dc.b	$E1,$7F,$7D,$F0,$3F,$BF,$37,$2F,$67,$3F,$F7,$AC,$3A,$1F,$94,$7F
	.dc.b	$8D,$E3,$FC,$EF,$6F,$2D,$7E,$9F,$1F,$F0,$77,$81,$FE,$09,$3F,$C1
	.dc.b	$5D,$76,$7A,$F1,$FF,$82,$FC,$1F,$E2,$F9,$74,$E5,$FC,$0F,$5F,$F6
	.dc.b	$7C,$FA,$EB,$ED,$D3,$FC,$35,$DB,$FE,$15,$3F,$C2,$DE,$0D,$D3,$AD
	.dc.b	$7F,$80,$1E,$7B,$FC,$7C,$BF,$91,$F1,$7C,$5C,$FF,$C3,$57,$FE,$1D
	.dc.b	$3F,$62,$57,$2F,$F2,$E9,$D3,$FC,$81,$AF,$03,$46,$A3,$8B,$B3,$15
	.dc.b	$FE,$3F,$7B,$35,$E2,$78,$7E,$05,$EA,$76,$A4,$62,$7B,$35,$E1,$3B
	.dc.b	$8C,$47,$F8,$91,$AF,$F8,$51,$AF,$F8,$F1,$B3,$DA,$79,$12,$7B,$BF
	.dc.b	$0B,$47,$DD,$BB,$7F,$C8,$DE,$07,$F9,$46,$FF,$C0,$F3,$94,$7F,$94
	.dc.b	$5C,$37,$B7,$8E,$DE,$7C,$7F,$CC,$D9,$FB,$F8,$BF,$B5,$B7,$93,$FC
	.dc.b	$A9,$F9,$53,$C6,$FF,$2B,$56,$AF,$89,$32,$FF,$2B,$FA,$BA,$C3,$FC
	.dc.b	$CB,$A3,$C0,$F0,$D7,$A1,$BC,$1E,$E6,$CF,$EF,$A7,$FC,$A3,$A7,$8D
	.dc.b	$F9,$98,$F7,$C9,$8F,$9B,$3F,$05,$68,$FF,$34,$DF,$17,$E9,$5F,$FA
	.dc.b	$17,$3D,$E7,$F3,$BE,$37,$C3,$CB,$2F,$E5,$F0,$FC,$FA,$D4,$79,$30
	.dc.b	$F1,$67,$F2,$E3,$FD,$3B,$BF,$C7,$72,$78,$F5,$F9,$0F,$C6,$3F,$8B
	.dc.b	$65,$F1,$39,$7F,$AE,$37,$BC,$DF,$8E,$F9,$35,$EA,$3D,$B9,$76,$58
	.dc.b	$F8,$5E,$C6,$89,$8F,$F8,$A3,$6F,$8F,$97,$F6,$BF,$8E,$F0,$12,$BF
	.dc.b	$94,$78,$5E,$0E,$8F,$CC,$47,$2D,$BE,$06,$3F,$4A,$3F,$23,$97,$86
	.dc.b	$3F,$C6,$BE,$CF,$D4,$9F,$F4,$EB,$6C,$FD,$8E,$17,$E0,$6F,$0B,$CB
	.dc.b	$C4,$F8,$EF,$89,$F1,$F3,$2F,$E7,$91,$E5,$E7,$FB,$D9,$F9,$B4,$F9
	.dc.b	$E7,$96,$BE,$2F,$47,$87,$DC,$91,$FE,$C7,$7C,$DB,$FE,$FE,$BF,$D5
	.dc.b	$99,$BE,$3B,$EA,$12,$FF,$A6,$7C,$7F,$87,$9F,$E9,$67,$E1,$E3,$FC
	.dc.b	$A5,$7E,$D3,$1F,$16,$5D,$67,$F5,$B7,$E0,$3C,$3E,$EF,$C0,$F8,$EF
	.dc.b	$C1,$13,$0F,$EB,$5F,$1F,$DD,$FA,$47,$E3,$6F,$BB,$F9,$2A,$E6,$F8
	.dc.b	$63,$F7,$71,$F3,$FC,$1F,$E6,$99,$7F,$B2,$7F,$71,$FE,$19,$6E,$7F
	.dc.b	$E1,$DD,$69,$7C,$87,$E5,$DB,$5F,$A6,$C3,$F8,$8D,$6D,$F3,$61,$F4
	.dc.b	$B2,$FC,$39,$E6,$7E,$1F,$C0,$FD,$02,$5F,$9C,$FC,$5C,$FE,$5B,$07
	.dc.b	$C8,$78,$7C,$89,$5E,$11,$97,$9E,$7C,$37,$E0,$9F,$8C,$7E,$5B,$9F
	.dc.b	$EF,$C4,$8F,$01,$FA,$A7,$F0,$7F,$01,$E1,$E9,$F8,$04,$9F,$06,$1F
	.dc.b	$17,$E4,$3B,$C7,$63,$F6,$DA,$FE,$03,$1E,$71,$FE,$BF,$3F,$6F,$87
	.dc.b	$FC,$01,$F6,$BE,$8F,$32,$31,$C1,$FC,$CB,$E0,$2D,$FB,$B8,$1F,$21
	.dc.b	$F4,$FE,$42,$4F,$92,$72,$FF,$17,$3F,$0E,$79,$38,$3E,$91,$D5,$F2
	.dc.b	$64,$C3,$BC,$7F,$0C,$7C,$3C,$AC,$9E,$DF,$93,$87,$F0,$DF,$21,$F8
	.dc.b	$5F,$90,$98,$F8,$E3,$E2,$BE,$13,$E3,$65,$FC,$1F,$3F,$E9,$FF,$09
	.dc.b	$E0,$83,$4C,$7C,$27,$7D,$9F,$1F,$F9,$32,$61,$F0,$7E,$43,$BF,$83
	.dc.b	$FB,$38,$EF,$CD,$FB,$EC,$39,$C7,$C0,$7C,$1F,$0D,$D7,$E4,$3F,$3D
	.dc.b	$F0,$B2,$BE,$64,$74,$EC,$FE,$1E,$7F,$4D,$DB,$D5,$3F,$C8,$AF,$9B
	.dc.b	$F3,$1F,$93,$C9,$FE,$9F,$8F,$F8,$01,$E9,$FF,$7E,$F4,$BF,$B0,$92
	.dc.b	$3F,$39,$1F,$E0,$2F,$47,$C1,$E3,$FF,$3F,$C7,$BF,$3F,$ED,$7E,$43
	.dc.b	$E2,$7C,$7C,$C7,$FC,$07,$E0,$7E,$EB,$99,$5F,$8A,$CF,$FC,$1B,$E9
	.dc.b	$FF,$70,$25,$F8,$3D,$7F,$E6,$0F,$E2,$7C,$3F,$DD,$3F,$28,$FD,$B6
	.dc.b	$7F,$C1,$7E,$3F,$EA,$7C,$7C,$9F,$0F,$17,$F0,$4D,$15,$E1,$C7,$F9
	.dc.b	$65,$77,$F4,$FF,$C0,$27,$F8,$0B,$D0,$F7,$C7,$E6,$38,$7D,$F8,$3D
	.dc.b	$7F,$C1,$12,$7B,$9F,$A5,$F5,$FB,$7A,$3F,$E0,$CF,$4A,$4F,$B6,$3A
	.dc.b	$F0,$7B,$23,$FA,$29,$EC,$C1,$EA,$7E,$3C,$91,$DB,$85,$FA,$47,$F9
	.dc.b	$17,$9B,$FC,$0D,$F7,$DF,$CA,$BC,$AF,$F0,$47,$13,$D5,$FF,$0E,$9F
	.dc.b	$E2,$09,$D7,$2F,$F1,$5C,$BF,$48,$4E,$2F,$BB,$93,$D5,$FD,$72,$F5
	.dc.b	$29,$88,$CF,$95,$FE,$0B,$97,$C7,$D3,$11,$FE,$0A,$BF,$CB,$E1,$E0
	.dc.b	$F2,$3F,$4C,$FC,$23,$FC,$37,$F1,$DF,$C0,$3F,$12,$F1,$47,$CE,$E0
	.dc.b	$46,$21,$C3,$FC,$6A,$BC,$FF,$21,$E3,$24,$FF,$28,$D3,$F8,$F3,$FC
	.dc.b	$E4,$F0,$73,$7F,$27,$7F,$99,$E7,$E6,$3F,$C0,$8E,$3C,$BF,$E0,$C4
	.dc.b	$7C,$9E,$0F,$F5,$AC,$BF,$AC,$E1,$FC,$6E,$7C,$1E,$77,$C5,$3E,$37
	.dc.b	$A5,$F0,$FD,$CF,$F0,$0E,$32,$BC,$31,$FE,$04,$3E,$17,$24,$8F,$E4
	.dc.b	$F1,$F5,$3C,$EF,$A1,$E5,$C9,$F7,$47,$CF,$F8,$9F,$F0,$F7,$17,$D4
	.dc.b	$9F,$EB,$DC,$7F,$2B,$9B,$FC,$23,$52,$7E,$75,$38,$FE,$91,$FC,$96
	.dc.b	$7F,$07,$C9,$F7,$FE,$42,$5F,$CD,$8F,$BD,$F1,$3C,$E3,$FC,$AB,$5F
	.dc.b	$E1,$E8,$91,$DB,$83,$EF,$1F,$CC,$63,$F1,$FC,$1F,$8B,$24,$FD,$9D
	.dc.b	$FF,$15,$F1,$5F,$E2,$82,$47,$B3,$95,$EB,$62,$3F,$0B,$D9,$FE,$8C
	.dc.b	$AF,$E7,$C7,$F4,$CA,$FE,$77,$3F,$AD,$E6,$69,$F9,$B2,$47,$E2,$73
	.dc.b	$FC,$D7,$C3,$FC,$CD,$A4,$FE,$22,$3F,$C7,$67,$E6,$77,$91,$FD,$7A
	.dc.b	$3F,$57,$97,$E9,$D9,$1F,$95,$8F,$D3,$7C,$47,$F9,$10,$91,$FE,$06
	.dc.b	$FC,$B1,$F1,$58,$8F,$69,$FE,$62,$9F,$EB,$47,$F6,$A9,$FF,$71,$75
	.dc.b	$FE,$52,$3F,$6D,$F2,$32,$BF,$49,$1E,$07,$F4,$2F,$5E,$EF,$F2,$D5
	.dc.b	$F8,$27,$C4,$76,$3F,$BC,$CF,$77,$FB,$A7,$A9,$FE,$F1,$EA,$FF,$1D
	.dc.b	$BF,$EE,$56,$F8,$DF,$E9,$18,$D7,$27,$FB,$2A,$98,$77,$7F,$D3,$B8
	.dc.b	$78,$FC,$2D,$4E,$72,$F2,$E3,$EF,$25,$FF,$87,$E7,$96,$9E,$2F,$73
	.dc.b	$C1,$B4,$AF,$F1,$5E,$9F,$E0,$C7,$F8,$25,$F5,$CB,$FD,$A7,$9F,$86
	.dc.b	$7F,$DB,$F3,$96,$DF,$81,$FA,$67,$AB,$2F,$A5,$1F,$13,$E4,$7E,$07
	.dc.b	$C8,$CA,$F0,$C7,$FA,$BA,$CF,$08,$F5,$57,$DB,$89,$1F,$B6,$E1,$FE
	.dc.b	$4C,$61,$FC,$1E,$0C,$DF,$23,$F3,$FE,$3A,$57,$C0,$8F,$06,$7F,$87
	.dc.b	$24,$7C,$98,$AF,$B3,$3F,$F0,$53,$67,$F5,$5E,$87,$E0,$3E,$3B,$EE
	.dc.b	$7C,$74,$AF,$9B,$1E,$DC,$39,$F1,$7F,$A0,$31,$E6,$D8,$FC,$C1,$FD
	.dc.b	$07,$82,$5F,$8C,$FE,$C5,$A7,$FB,$DD,$FF,$44,$72,$7F,$A0,$A1,$F1
	.dc.b	$DF,$8D,$3F,$1D,$F2,$1F,$37,$7F,$D2,$12,$FF,$A6,$7E,$0F,$FD,$75
	.dc.b	$FC,$44,$65,$F8,$0F,$F9,$4A,$3F,$D7,$99,$FE,$73,$E4,$3F,$35,$F0
	.dc.b	$12,$FF,$17,$1F,$6F,$7F,$D9,$12,$3B,$F9,$75,$F9,$91,$6F,$FC,$BD
	.dc.b	$1F,$EA,$3C,$1D,$FF,$AC,$D7,$F5,$04,$9F,$CC,$FB,$BF,$89,$E3,$FF
	.dc.b	$33,$FE,$EF,$F8,$09,$FE,$D0,$D7,$ED,$47,$F0,$53,$7B,$48,$FD,$44
	.dc.b	$7E,$EB,$E4,$3F,$6C,$48,$FD,$3F,$BB,$F9,$7E,$1F,$E1,$C4,$8F,$97
	.dc.b	$E2,$57,$F1,$11,$76,$F6,$F8,$2D,$DD,$9F,$C2,$3D,$DF,$E0,$9E,$E4
	.dc.b	$C7,$F6,$B1,$FA,$5E,$7F,$6F,$34,$9F,$DF,$47,$F1,$F3,$E0,$F3,$C8
	.dc.b	$C9,$F8,$8F,$73,$F8,$87,$B3,$23,$F8,$0C,$7E,$C6,$FF,$94,$12,$3F
	.dc.b	$C2,$1E,$2F,$FA,$F9,$FE,$19,$19,$7F,$48,$FF,$A5,$A3,$F8,$54,$7E
	.dc.b	$CB,$F0,$CF,$F1,$7B,$FE,$C6,$7F,$DB,$2F,$E6,$0C,$7C,$87,$F2,$1D
	.dc.b	$FF,$DC,$33,$C2,$7F,$91,$E5,$FC,$EC,$8F,$93,$C4,$AF,$ED,$62,$F0
	.dc.b	$FC,$EC,$7F,$5B,$DD,$FE,$E6,$C3,$FD,$B7,$7F,$EC,$A5,$F2,$1E,$19
	.dc.b	$D3,$67,$D1,$F2,$1F,$C5,$7B,$7F,$A6,$12,$3E,$2F,$5D,$7F,$83,$45
	.dc.b	$CB,$C5,$DF,$FE,$B7,$C9,$F0,$FD,$7C,$74,$3E,$43,$0D,$C7,$CD,$5F
	.dc.b	$C9,$7C,$1F,$EB,$84,$8F,$AB,$D9,$5F,$E2,$D1,$72,$F8,$5D,$1E,$1F
	.dc.b	$8F,$EB,$F1,$F2,$FE,$14,$7F,$31,$F0,$7F,$B4,$12,$3F,$03,$CB,$5F
	.dc.b	$E4,$51,$70,$F9,$DC,$1E,$DF,$8F,$F4,$FA,$B2,$FE,$74,$7F,$3D,$F0
	.dc.b	$FF,$BA,$12,$3F,$1B,$85,$7F,$97,$45,$C3,$ED,$3F,$27,$E3,$FE,$2F
	.dc.b	$C7,$CA,$FB,$51,$FD,$27,$C1,$FF,$73,$F8,$CA,$A7,$F8,$0E,$25,$7F
	.dc.b	$9A,$E5,$F1,$DC,$0C,$C7,$E6,$5F,$A6,$47,$F1,$1F,$1F,$FD,$5F,$C5
	.dc.b	$EF,$F1,$B2,$7F,$B8,$47,$F8,$44,$F1,$FC,$74,$8C,$9F,$E9,$BE,$9F
	.dc.b	$F6,$2B,$F9,$38,$FC,$09,$F8,$2A,$FE,$C9,$E3,$7F,$86,$C9,$3F,$E8
	.dc.b	$3E,$CF,$CC,$BF,$DF,$E3,$2F,$F5,$77,$FC,$19,$C1,$FC,$42,$7F,$23
	.dc.b	$3E,$1F,$8D,$FC,$7F,$C7,$48,$FD,$24,$7F,$73,$F2,$7F,$C5,$44,$8F
	.dc.b	$07,$0D,$7F,$B2,$45,$C3,$B6,$FF,$9F,$F8,$EF,$CD,$92,$BF,$61,$EA
	.dc.b	$7E,$B5,$7A,$FF,$E8,$9F,$1D,$0F,$29,$FB,$07,$F5,$09,$CC,$62,$BD
	.dc.b	$B1,$FE,$29,$79,$57,$26,$CF,$2F,$85,$DC,$F5,$BF,$F4,$23,$FE,$18
	.dc.b	$AE,$3C,$1F,$B4,$F8,$EF,$D9,$92,$7C,$38,$F9,$A5,$FA,$8B,$E4,$7D
	.dc.b	$98,$7F,$1E,$DF,$FA,$DE,$1E,$EF,$43,$C2,$D3,$ED,$72,$62,$B8,$65
	.dc.b	$EB,$AC,$1E,$8D,$99,$FF,$BC,$F0,$FF,$72,$67,$FE,$C3,$9F,$ED,$33
	.dc.b	$E4,$F8,$AF,$E1,$D3,$FD,$F6,$BF,$98,$4F,$A3,$E3,$BC,$FE,$7F,$5E
	.dc.b	$07,$C5,$D3,$C9,$12,$FF,$DA,$52,$F6,$B8,$19,$8F,$F1,$8E,$3F,$71
	.dc.b	$F8,$E7,$A1,$F9,$F9,$F9,$9B,$AF,$CB,$78,$4F,$F4,$53,$F0,$BE,$97
	.dc.b	$C5,$23,$FE,$E3,$FC,$2F,$D9,$C5,$F3,$B8,$7F,$47,$3F,$6E,$B6,$FF
	.dc.b	$A1,$B1,$5F,$E2,$E3,$C9,$F1,$7F,$18,$8F,$A3,$83,$EE,$FC,$9F,$DC
	.dc.b	$F4,$BE,$B1,$27,$C5,$55,$FC,$54,$5C,$FF,$BF,$E1,$FE,$F0,$9F,$8B
	.dc.b	$F1,$9F,$53,$E3,$A4,$77,$DB,$FD,$D0,$FC,$1D,$BF,$A1,$3E,$9E,$17
	.dc.b	$C3,$FD,$73,$F1,$1E,$B4,$9F,$AB,$F0,$3F,$6C,$C3,$F5,$39,$BF,$06
	.dc.b	$47,$C1,$CB,$F9,$8F,$4F,$F2,$44,$8F,$9D,$F0,$35,$FD,$34,$5E,$0F
	.dc.b	$6F,$C4,$7E,$B3,$97,$F4,$77,$F9,$02,$5F,$B6,$7F,$4B,$EB,$7E,$7C
	.dc.b	$91,$F6,$FE,$32,$BF,$B3,$8B,$87,$CA,$F8,$AE,$DC,$4F,$8D,$FC,$E9
	.dc.b	$2F,$E5,$61,$FB,$2F,$5B,$F5,$84,$8F,$C4,$7C,$15,$7F,$80,$45,$C7
	.dc.b	$EB,$F1,$F6,$CB,$F5,$24,$AF,$AD,$F0,$9F,$BE,$F6,$7F,$70,$48,$FC
	.dc.b	$A7,$C5,$D7,$F8,$64,$5C,$3F,$C8,$3F,$85,$FF,$B5,$3C,$2F,$D0,$69
	.dc.b	$FB,$62,$5F,$E0,$F3,$FE,$17,$EC,$FF,$05,$24,$7F,$26,$C3,$9F,$C6
	.dc.b	$62,$3F,$8D,$4D,$9E,$FF,$19,$F2,$3A,$BF,$81,$EB,$2F,$F2,$11,$FC
	.dc.b	$63,$D2,$FE,$24,$48,$ED,$A7,$F3,$0D,$7F,$C5,$71,$FC,$B6,$74,$FF
	.dc.b	$29,$7F,$8B,$47,$F9,$BF,$1F,$E2,$7F,$B9,$7F,$B7,$AB,$F8,$CB,$FE
	.dc.b	$E3,$BF,$E5,$4D,$8C,$9F,$CA,$BD,$2F,$E4,$84,$7C,$67,$F4,$5B,$7C
	.dc.b	$26,$63,$F9,$D4,$DC,$3E,$7F,$B9,$F9,$5C,$3F,$CA,$BE,$4F,$93,$6F
	.dc.b	$F1,$94,$7F,$BB,$E3,$CF,$C1,$FC,$EB,$DC,$FE,$61,$F5,$EF,$5E,$9F
	.dc.b	$CD,$E2,$69,$E4,$B7,$EC,$3F,$D4,$0F,$EA,$71,$FD,$7A,$7C,$5F,$13
	.dc.b	$E0,$F8,$F9,$3F,$C3,$F3,$FF,$70,$BF,$CA,$0F,$EC,$BE,$EF,$F4,$52
	.dc.b	$3D,$F8,$BF,$DB,$CF,$C4,$36,$5F,$2B,$FC,$1A,$FF,$5D,$9F,$AB,$F0
	.dc.b	$FF,$0F,$E3,$BE,$0F,$C5,$4A,$F1,$47,$F5,$AF,$81,$FE,$A8,$48,$F0
	.dc.b	$7C,$2D,$7F,$B4,$45,$E6,$FC,$EF,$C2,$FC,$FE,$9F,$9A,$4A,$F7,$74
	.dc.b	$7F,$66,$F8,$2F,$EB,$9C,$8F,$16,$9F,$D8,$E2,$57,$CD,$97,$87,$C0
	.dc.b	$CC,$32,$FF,$64,$FC,$8B,$E5,$7F,$CF,$F3,$FE,$8F,$8C,$7B,$DF,$E7
	.dc.b	$D6,$C7,$F5,$67,$87,$B9,$F8,$4F,$8C,$FC,$47,$C6,$4A,$FC,$8F,$A4
	.dc.b	$EA,$FF,$79,$67,$F9,$EF,$8E,$FC,$77,$C6,$CB,$DD,$27,$F2,$5F,$1B
	.dc.b	$F9,$02,$4F,$E0,$7A,$BF,$C1,$3F,$09,$F1,$3C,$99,$3F,$6E,$3F,$C0
	.dc.b	$C7,$C8,$F2,$24,$64,$EF,$F1,$BF,$9C,$23,$F8,$DE,$CF,$F0,$AF,$C5
	.dc.b	$7F,$83,$C9,$1F,$99,$FC,$37,$FC,$8A,$FF,$88,$E3,$2D,$35,$EE,$FE
	.dc.b	$AF,$BB,$F5,$24,$8F,$CD,$7C,$77,$F8,$8B,$E3,$3F,$C3,$84,$8F,$CB
	.dc.b	$71,$2B,$11,$FE,$31,$9B,$C7,$FE,$28,$A5,$FF,$8E,$3A,$3E,$E7,$C2
	.dc.b	$7D,$9E,$BF,$EF,$FE,$5F,$E0,$1B,$E8,$7F,$92,$AB,$79,$FE,$B7,$E3
	.dc.b	$7F,$81,$79,$D2,$3E,$AC,$7F,$59,$F6,$7F,$C8,$24,$BC,$78,$19,$F0
	.dc.b	$3F,$D7,$13,$F8,$CC,$FF,$C5,$ED,$AF,$F2,$23,$E2,$C3,$FC,$A5,$F1
	.dc.b	$9F,$E3,$D2,$3F,$3F,$C5,$AF,$F9,$46,$2F,$0A,$FF,$CB,$FC,$DF,$A0
	.dc.b	$AF,$E4,$3B,$7E,$77,$97,$E1,$12,$3F,$35,$C0,$F8,$0F,$3F,$17,$B4
	.dc.b	$93,$F9,$AD,$7F,$C1,$B9,$FA,$AB,$E1,$7B,$D2,$B5,$1F,$17,$6F,$83
	.dc.b	$E2,$C9,$FF,$13,$61,$FB,$97,$F8,$C7,$E6,$3F,$E6,$37,$F9,$27,$C1
	.dc.b	$BE,$23,$FC,$B4,$D8,$FE,$31,$3B,$7F,$75,$C3,$FE,$82,$51,$9D,$3E
	.dc.b	$24,$7F,$A1,$6B,$FC,$59,$18,$E3,$FF,$2C,$FC,$4F,$F9,$BA,$F8,$61
	.dc.b	$F5,$3D,$CF,$B3,$E1,$CC,$BD,$31,$FE,$93,$A7,$C5,$F1,$24,$FE,$97
	.dc.b	$CD,$AF,$98,$FF,$4B,$CF,$07,$D7,$3F,$E7,$7E,$77,$8B,$F8,$DF,$17
	.dc.b	$F1,$24,$AF,$93,$EC,$78,$7E,$33,$E2,$FC,$1C,$9F,$8B,$1E,$B3,$E4
	.dc.b	$F2,$48,$C9,$F8,$4E,$A7,$4F,$E9,$49,$1F,$84,$E8,$FC,$CF,$AD,$F9
	.dc.b	$F2,$47,$E0,$FD,$4F,$D1,$70,$C8,$FC,$33,$FA,$5D,$BF,$4D,$E4,$7E
	.dc.b	$8B,$C7,$95,$F8,$28,$F9,$5E,$BF,$DE,$F8,$19,$3F,$7A,$3E,$A1,$F8
	.dc.b	$2F,$81,$91,$93,$F2,$9E,$27,$E0,$7B,$3F,$D3,$7F,$A9,$FE,$15,$BF
	.dc.b	$1D,$FB,$29,$FB,$1F,$1F,$FE,$A5,$FD,$DF,$FC,$59,$87,$FB,$25,$AF
	.dc.b	$E2,$A3,$F0,$87,$FB,$0B,$C1,$7E,$B6,$E2,$C3,$6B,$F2,$B5,$F8,$BF
	.dc.b	$13,$F2,$91,$3B,$FF,$34,$E0,$FC,$BC,$57,$F0,$AD,$72,$66,$38,$BC
	.dc.b	$E3,$51,$88,$FD,$13,$5E,$67,$18,$D9,$C3,$C3,$91,$D5,$AF,$96,$7F
	.dc.b	$D2,$4F,$A6,$32,$D3,$81,$0E,$0E,$9F,$1F,$C7,$F9,$88,$FA,$DA,$F7
	.dc.b	$6B,$F1,$DA,$F2,$38,$10,$F3,$7C,$7F,$C2,$23,$F1,$DA,$E8,$C1,$0C
	.dc.b	$10,$FA,$CF,$07,$EC,$78,$DF,$38,$8F,$D3,$6B,$F1,$DA,$F0,$34,$43
	.dc.b	$45,$3F,$09,$F1,$DF,$68,$8F,$DF,$6B,$F4,$DA,$F0,$34,$43,$45,$3F
	.dc.b	$21,$F1,$DF,$86,$23,$F8,$B6,$BF,$7D,$AF,$03,$44,$34,$53,$F3,$9F
	.dc.b	$1D,$F9,$22,$3F,$97,$6B,$F8,$B6,$BC,$0D,$10,$D1,$4F,$D4,$78,$5F
	.dc.b	$9E,$23,$FA,$36,$BF,$97,$6B,$C0,$D1,$0D,$14,$FD,$A7,$83,$FA,$A2
	.dc.b	$3F,$AF,$6B,$FA,$36,$BC,$0D,$10,$D1,$4F,$E0,$1F,$1B,$FB,$62,$3F
	.dc.b	$BB,$6B,$FA,$F6,$BC,$0D,$10,$D1,$4F,$E1,$9F,$1B,$FC,$08,$8F,$F0
	.dc.b	$76,$BF,$BB,$6B,$C0,$D1,$0D,$14,$FE,$31,$D7,$FC,$38,$8F,$F1,$36
	.dc.b	$BF,$C1,$DA,$F0,$34,$43,$45,$3F,$92,$75,$7F,$1A,$23,$FC,$7D,$AF
	.dc.b	$F1,$36,$BC,$0D,$10,$D1,$4F,$E6,$1F,$19,$FC,$98,$8F,$F2,$B6,$BF
	.dc.b	$C7,$DA,$F0,$34,$43,$45,$3F,$9E,$7C,$6F,$F3,$22,$3F,$CD,$DA,$FF
	.dc.b	$2B,$6B,$C0,$D1,$0D,$14,$FE,$93,$F9,$A7,$F4,$34,$D2,$3F,$D1,$DA
	.dc.b	$FF,$4B,$6B,$FD,$3D,$AF,$43,$81,$C7,$E0,$63,$C5,$AF,$16,$BE,$46
	.dc.b	$BF,$E3,$18,$46,$20,$7A,$1C,$3E,$AF,$8D,$F4,$11,$F4,$35,$F4,$35
	.dc.b	$F8,$6D,$7C,$0D,$7B,$CA,$3E,$27,$C7,$49,$F8,$8D,$7E,$23,$5F,$9E
	.dc.b	$D7,$D8,$D7,$AB,$5F,$A7,$F1,$BF,$44,$8F,$D1,$6B,$F4,$5A,$FD,$C6
	.dc.b	$BF,$25,$AF,$A9,$FF,$02,$5B,$C1,$FF,$24,$4B,$C1,$FF,$0F,$F9,$FF
	.dc.b	$E2,$1F,$4B,$E7,$DE,$E3,$15,$F5,$2D,$E6,$E2,$1F,$20,$F6,$BF,$CD
	.dc.b	$5E,$1F,$F9,$A2,$73,$ED,$63,$3F,$F3,$75,$FF,$B9,$AB,$4F,$1A,$FB
	.dc.b	$F1,$8F,$92,$E1,$1F,$82,$94,$62,$3C,$2D,$8F,$F6,$5C,$FF,$8E,$BC
	.dc.b	$9F,$4E,$BF,$05,$BA,$1B,$E2,$63,$C4,$EF,$5F,$1F,$C3,$B7,$FE,$80
	.dc.b	$D3,$E2,$BF,$E3,$2C,$7F,$D0,$A3,$F5,$87,$5F,$F8,$FD,$75,$D3,$11
	.dc.b	$E9,$F0,$9E,$67,$CE,$F2,$DF,$A2,$F2,$E3,$FD,$0F,$47,$FA,$4A,$73
	.dc.b	$87,$D1,$C3,$8D,$6B,$5F,$2F,$07,$A3,$79,$3F,$32,$72,$D1,$B7,$A7
	.dc.b	$0F,$34,$66,$3D,$9A,$FA,$23,$81,$FC,$2B,$E3,$FF,$84,$11,$FE,$10
	.dc.b	$D7,$F8,$43,$5F,$E2,$6D,$7F,$80,$35,$FD,$8B,$F9,$8E,$07,$91,$F6
	.dc.b	$3C,$89,$5E,$F3,$F2,$91,$F9,$3F,$8D,$FF,$7C,$C7,$99,$7C,$07,$E1
	.dc.b	$DB,$19,$F8,$2C,$7C,$14,$8E,$0F,$E2,$F6,$FC,$77,$86,$F8,$BF,$E4
	.dc.b	$C4,$AF,$E4,$CD,$7F,$93,$35,$FE,$62,$D7,$F8,$F3,$5F,$E1,$EF,$BF
	.dc.b	$8F,$F2,$BE,$6C,$9F,$77,$57,$EE,$67,$DD,$5E,$BF,$8A,$95,$EF,$8F
	.dc.b	$77,$C2,$C8,$F8,$4F,$C1,$F8,$AF,$CD,$FA,$1F,$A2,$24,$F6,$E3,$7A
	.dc.b	$1F,$2F,$97,$E4,$12,$7E,$67,$C7,$7C,$BE,$3F,$99,$F1,$B2,$7E,$84
	.dc.b	$7C,$8F,$89,$FE,$98,$48,$FE,$98,$D7,$FA,$63,$5F,$EA,$ED,$7F,$A2
	.dc.b	$35,$FE,$6A,$FE,$17,$D3,$FD,$DF,$A7,$27,$D5,$1F,$C5,$7E,$61,$E8
	.dc.b	$FF,$07,$6C,$FE,$69,$AF,$FB,$B4,$61,$FE,$BA,$3F,$A9,$BF,$8B,$F4
	.dc.b	$7F,$14,$7E,$2F,$97,$F8,$2F,$A9,$FC,$30,$93,$F2,$FC,$1F,$11,$E1
	.dc.b	$F0,$24,$78,$63,$C9,$EA,$48,$F2,$3E,$7F,$53,$F8,$F7,$E6,$DF,$C8
	.dc.b	$27,$FD,$7C,$D7,$FD,$84,$D7,$FD,$8E,$D7,$FB,$48,$70,$FF,$6C,$6B
	.dc.b	$9F,$3F,$FD,$91,$18,$23,$82,$47,$FB,$47,$2F,$E9,$98,$FF,$2C,$C2
	.dc.b	$4F,$E2,$E3,$E9,$78,$9F,$44,$91,$F4,$7B,$7E,$A7,$77,$D5,$EA,$93
	.dc.b	$F5,$63,$FA,$0F,$90,$FD,$7D,$CB,$FC,$9D,$B5,$FC,$0E,$75,$39,$EB
	.dc.b	$FE,$7F,$C1,$FC,$E8,$95,$FC,$87,$6F,$EC,$9C,$DF,$EF,$5F,$B7,$FE
	.dc.b	$7D,$3F,$C9,$1F,$E2,$2F,$F2,$36,$BF,$CA,$3C,$3F,$EA,$F1,$C2,$9F
	.dc.b	$C0,$3F,$EF,$8F,$31,$5F,$DB,$23,$0D,$BF,$EA,$B1,$F8,$E7,$FB,$A4
	.dc.b	$7D,$8E,$A7,$A1,$F9,$9F,$36,$57,$E6,$63,$F3,$9E,$6C,$8F,$CE,$3F
	.dc.b	$9F,$F3,$3F,$B4,$FC,$73,$E1,$5F,$0F,$FD,$F2,$31,$F1,$56,$4D,$9F
	.dc.b	$EF,$DD,$CF,$8A,$FD,$57,$B5,$33,$FE,$87,$39,$7F,$A6,$7D,$3B,$F0
	.dc.b	$37,$C5,$7F,$7A,$8C,$36,$33,$3F,$D4,$FE,$B1,$FB,$AF,$F7,$58,$EB
	.dc.b	$97,$E5,$08,$FE,$89,$AF,$E8,$9A,$FE,$89,$AF,$E8,$B1,$F3,$3B,$39
	.dc.b	$1F,$9B,$E5,$FF,$20,$3F,$CC,$1A,$FD,$43,$FC,$AA,$E3,$9B,$EE,$C6
	.dc.b	$3E,$0B,$FC,$49,$F4,$6F,$8A,$FF,$0B,$2F,$8A,$A6,$23,$3E,$F4,$BF
	.dc.b	$ED,$B5,$FD,$F1,$FE,$28,$BE,$2B,$FC,$2B,$93,$E1,$BF,$13,$F3,$72
	.dc.b	$FD,$11,$E2,$F8,$FF,$09,$23,$AF,$C7,$79,$3E,$33,$CB,$F1,$D2,$7C
	.dc.b	$31,$FC,$43,$5F,$E0,$64,$8F,$E2,$2D,$7F,$88,$B5,$FE,$22,$D7,$E8
	.dc.b	$35,$FE,$20,$FF,$95,$32,$7C,$D7,$E8,$89,$3F,$47,$DB,$F9,$7F,$2B
	.dc.b	$F2,$BE,$52,$4F,$C6,$8F,$9B,$D5,$23,$E4,$BF,$43,$E4,$BE,$3F,$BD
	.dc.b	$FA,$EF,$6E,$57,$DD,$8F,$AF,$F1,$DF,$58,$91,$F4,$BE,$3B,$EC,$FB
	.dc.b	$5F,$6B,$E5,$64,$FD,$68,$FF,$29,$1F,$97,$7F,$97,$9F,$E9,$87,$0E
	.dc.b	$1A,$FF,$34,$6B,$BF,$8E,$E2,$E5,$AE,$4B,$1C,$1F,$F3,$1C,$69,$A6
	.dc.b	$9C,$47,$6F,$3F,$F8,$AF,$C1,$7F,$10,$25,$7E,$7F,$E1,$BF,$CF,$9F
	.dc.b	$92,$3E,$67,$FC,$DC,$BE,$66,$98,$8C,$FC,$CC,$BF,$F2,$2D,$23,$8F
	.dc.b	$C3,$FE,$63,$E2,$FF,$8F,$FC,$5C,$BF,$D7,$47,$83,$2E,$E4,$8E,$71
	.dc.b	$E2,$9F,$1F,$C5,$C9,$EF,$1F,$4B,$E0,$7D,$E4,$8F,$8E,$D7,$E6,$EE
	.dc.b	$8C,$47,$D5,$6B,$F5,$5A,$FD,$B7,$F3,$15,$FC,$E0,$91,$FC,$1F,$4F
	.dc.b	$93,$E4,$7C,$8F,$0E,$4F,$C4,$8F,$97,$E2,$48,$F8,$EF,$CE,$EE,$FD
	.dc.b	$EF,$0F,$F4,$4E,$09,$5F,$C4,$E3,$FD,$61,$D5,$FE,$18,$F9,$0F,$F5
	.dc.b	$73,$7D,$BF,$EF,$ED,$AF,$F7,$4F,$27,$E5,$2B,$FD,$68,$FF,$4F,$6F
	.dc.b	$C5,$7D,$E3,$EF,$F3,$FE,$73,$E1,$FF,$12,$4B,$FC,$8B,$5F,$96,$D7
	.dc.b	$F3,$CD,$7F,$3C,$D7,$F4,$8F,$F0,$1E,$7F,$EC,$3F,$1F,$27,$F9,$84
	.dc.b	$7B,$BE,$0B,$FD,$7D,$9F,$F9,$52,$7F,$D2,$2D,$7F,$C7,$0F,$F8,$09
	.dc.b	$BF,$19,$C3,$E3,$24,$70,$8F,$E5,$3F,$21,$FD,$C0,$91,$FC,$FF,$E4
	.dc.b	$3E,$6F,$C6,$78,$3E,$32,$4F,$82,$3C,$7D,$72,$3C,$6F,$9B,$E3,$3F
	.dc.b	$DA,$9F,$9C,$BA,$BF,$68,$DA,$FD,$DB,$5F,$DF,$B5,$FE,$12,$D7,$FC
	.dc.b	$72,$D4,$1C,$36,$3B,$91,$EE,$D7,$F8,$AF,$13,$F5,$9F,$F7,$B7,$B1
	.dc.b	$FD,$6A,$31,$E4,$7F,$BA,$27,$FD,$E9,$5F,$D8,$67,$87,$4F,$E4,$BE
	.dc.b	$37,$EA,$92,$7E,$AF,$CB,$7D,$8F,$81,$FB,$3E,$4C,$9F,$B3,$1F,$EE
	.dc.b	$EC,$5E,$5F,$72,$4F,$FA,$82,$3F,$CB,$DA,$7F,$6E,$87,$E4,$FF,$DB
	.dc.b	$5F,$F6,$9B,$FD,$B5,$BE,$E7,$F9,$83,$BF,$84,$EE,$7F,$CF,$CF,$F6
	.dc.b	$D8,$CF,$B9,$BD,$BF,$CC,$F3,$96,$FB,$7F,$73,$CE,$FB,$64,$C3,$ED
	.dc.b	$B5,$FB,$6D,$7E,$DB,$5F,$B6,$D7,$ED,$F2,$23,$F9,$FF,$17,$D5,$F3
	.dc.b	$FE,$97,$A7,$2F,$E2,$47,$F9,$F3,$6F,$6F,$CE,$7A,$67,$F0,$BA,$23
	.dc.b	$11,$F2,$1B,$1F,$E2,$2D,$BE,$2F,$C2,$E3,$E1,$65,$FC,$C8,$FA,$DE
	.dc.b	$BF,$E0,$C9,$1F,$57,$D7,$F1,$70,$FC,$8D,$7A,$D4,$AF,$AB,$1E,$0A
	.dc.b	$91,$F6,$1F,$1F,$89,$FA,$8F,$33,$F4,$E4,$9F,$D3,$B5,$FD,$3B,$5F
	.dc.b	$D3,$B5,$FD,$3B,$5F,$9A,$FC,$CE,$64,$FE,$77,$69,$5F,$8F,$8F,$5B
	.dc.b	$FA,$EF,$B1,$79,$3F,$29,$B1,$FB,$2D,$FF,$29,$18,$8F,$F6,$52,$F1
	.dc.b	$78,$78,$BF,$43,$79,$5C,$5F,$F5,$AD,$27,$F4,$9F,$05,$FB,$12,$57
	.dc.b	$E8,$F8,$7C,$AA,$BF,$BB,$3C,$CF,$47,$9D,$E7,$3D,$1E,$97,$A7,$82
	.dc.b	$4E,$CE,$4F,$B7,$CC,$FC,$3F,$E7,$5F,$B4,$9F,$F0,$8B,$5F,$F0,$AB
	.dc.b	$5F,$F0,$FB,$5F,$F6,$7B,$5F,$E2,$ED,$4D,$27,$81,$FE,$19,$6C,$78
	.dc.b	$23,$EC,$73,$FF,$13,$F8,$19,$3F,$C1,$63,$E1,$E6,$F7,$3E,$B1,$27
	.dc.b	$DD,$0F,$6F,$ED,$7B,$9F,$6F,$C0,$95,$A8,$FE,$29,$F0,$3F,$C8,$49
	.dc.b	$1F,$C4,$FE,$03,$BF,$99,$F8,$3F,$66,$4F,$68,$FC,$37,$B3,$23,$C2
	.dc.b	$FE,$27,$D9,$F9,$BD,$DE,$F2,$4F,$CE,$6B,$F3,$9A,$FC,$E6,$BF,$39
	.dc.b	$AF,$CE,$7F,$3D,$F0,$3F,$CE,$7E,$06,$4F,$F2,$B8,$FC,$4F,$D9,$BE
	.dc.b	$5B,$F8,$2B,$67,$F0,$8D,$7F,$DB,$C7,$F5,$05,$ED,$F0,$F9,$6E,$3F
	.dc.b	$29,$2B,$8C,$7F,$3D,$F4,$7F,$A5,$12,$3F,$9E,$7A,$3E,$2F,$77,$C3
	.dc.b	$F2,$B2,$7C,$31,$E4,$F7,$64,$79,$1F,$3F,$BD,$F9,$6D,$BF,$10,$49
	.dc.b	$FC,$BB,$5F,$CB,$B5,$FC,$BB,$5F,$CB,$B5,$FC,$BB,$FC,$1F,$D5,$FE
	.dc.b	$07,$EA,$3A,$1E,$CE,$1B,$7F,$D5,$EB,$E6,$FC,$8F,$CC,$24,$7C,$CF
	.dc.b	$92,$F9,$FF,$25,$F4,$3E,$4A,$4F,$D0,$8F,$E7,$59,$77,$F7,$7B,$12
	.dc.b	$7F,$C0,$3E,$CF,$F7,$7C,$DE,$E7,$DC,$D7,$ED,$DC,$BF,$B7,$1F,$7A
	.dc.b	$64,$7D,$E7,$F0,$3E,$EF,$F4,$6F,$CE,$B9,$CF,$FA,$41,$AF,$FA,$51
	.dc.b	$AF,$FA,$79,$AF,$F8,$50,$38,$7F,$C2,$ED,$73,$3F,$E9,$78,$C1,$1C
	.dc.b	$12,$3F,$D4,$79,$7D,$1A,$FA,$1B,$49,$F8,$7E,$9F,$F8,$92,$B5,$18
	.dc.b	$3F,$C4,$FF,$04,$F2,$FF,$B1,$B6,$3F,$91,$35,$FC,$F8,$7F,$00,$EB
	.dc.b	$C5,$E3,$C9,$23,$07,$F3,$1B,$FC,$37,$1F,$E1,$49,$1E,$38,$FC,$57
	.dc.b	$17,$F3,$AC,$3F,$BB,$4F,$8B,$D3,$F2,$9E,$6F,$0B,$CF,$E9,$C9,$F3
	.dc.b	$BE,$AF,$37,$EC,$7C,$1F,$CB,$24,$FD,$96,$BF,$65,$AF,$D9,$6B,$F6
	.dc.b	$78,$D1,$88,$FB,$2F,$F8,$53,$99,$D1,$FA,$1C,$E5,$7B,$67,$2F,$D1
	.dc.b	$3E,$57,$C0,$FD,$0F,$42,$47,$D0,$8F,$A7,$F0,$F2,$3E,$9B,$F5,$BE
	.dc.b	$07,$F4,$77,$E4,$9E,$D9,$7E,$C4,$FD,$97,$C2,$7F,$64,$FB,$8F,$EE
	.dc.b	$5E,$E7,$F9,$1D,$BD,$3F,$95,$6D,$FF,$9A,$9F,$EE,$CF,$F6,$37,$FB
	.dc.b	$93,$F8,$48,$FC,$39,$FE,$2A,$8C,$C7,$E1,$B7,$FE,$4C,$FE,$8F,$CB
	.dc.b	$FC,$E1,$1F,$D5,$35,$FD,$53,$5E,$06,$88,$68,$FD,$0E,$4F,$85,$FE
	.dc.b	$21,$E2,$C9,$FA,$11,$F0,$7E,$2F,$CD,$F1,$72,$7D,$91,$E9,$F8,$B9
	.dc.b	$1E,$F7,$D9,$F1,$7F,$3B,$CA,$FE,$38,$49,$FB,$9E,$57,$CD,$F6,$BE
	.dc.b	$11,$23,$E4,$7B,$3F,$17,$D8,$F8,$DE,$CC,$9F,$99,$1F,$E5,$2F,$65
	.dc.b	$F1,$5F,$C1,$E6,$46,$1C,$1F,$E9,$E7,$0E,$1A,$F0,$8D,$BB,$6A,$8D
	.dc.b	$BF,$C5,$5A,$EC,$E2,$E1,$E1,$1C,$A3,$7E,$27,$F4,$1F,$43,$FA,$07
	.dc.b	$24,$BF,$D0,$47,$F8,$D3,$F6,$5F,$F7,$84,$39,$FF,$DE,$1D,$1F,$91
	.dc.b	$8F,$F3,$9C,$EA,$7F,$DE,$53,$F9,$5C,$3F,$1C,$7E,$3F,$2F,$D0,$FA
	.dc.b	$DF,$D3,$C9,$87,$EB,$FD,$5F,$06,$7D,$09,$1F,$E3,$DC,$BF,$BB,$B6
	.dc.b	$7C,$3E,$B4,$8C,$1F,$E4,$3D,$5B,$7C,$12,$47,$C6,$6B,$F4,$DA,$FD
	.dc.b	$A6,$BF,$3D,$AF,$DB,$7F,$85,$4F,$F6,$9F,$98,$93,$FC,$26,$3E,$57
	.dc.b	$89,$F1,$BC,$09,$3E,$B8,$F9,$3E,$04,$8F,$73,$F3,$3B,$3F,$B4,$FE
	.dc.b	$EB,$F8,$DF,$57,$FD,$B8,$DE,$3F,$E8,$CD,$B7,$1A,$2C,$7F,$19,$E0
	.dc.b	$7A,$3F,$60,$91,$F3,$7C,$EF,$B7,$E7,$7D,$CF,$2A,$4F,$D1,$8F,$D0
	.dc.b	$7C,$1F,$E2,$C9,$1F,$90,$6B,$F9,$D6,$BF,$A4,$6B,$F9,$96,$BF,$A5
	.dc.b	$7F,$D8,$F8,$FE,$9B,$4F,$7F,$BD,$FE,$18,$83,$7A,$7F,$D1,$3F,$7B
	.dc.b	$FE,$E3,$F2,$7F,$DD,$8F,$FB,$A7,$1F,$EC,$2D,$8C,$FB,$BF,$EC,$A8
	.dc.b	$FF,$72,$48,$38,$6F,$B7,$BF,$6F,$E9,$F5,$F8,$C9,$3F,$D8,$F3,$7B
	.dc.b	$FE,$4F,$63,$B1,$27,$FD,$2D,$8F,$F7,$27,$CB,$EC,$78,$FD,$89,$1A
	.dc.b	$8F,$F6,$FF,$EF,$4F,$53,$F8,$1B,$67,$F8,$3B,$5F,$E1,$ED,$7F,$8A
	.dc.b	$35,$FF,$0E,$35,$F8,$0D,$4D,$24,$78,$09,$1F,$E1,$AC,$9F,$01,$F5
	.dc.b	$7C,$A9,$3F,$0E,$3E,$AF,$C6,$7C,$FF,$8D,$93,$F0,$A3,$E9,$7C,$6C
	.dc.b	$8F,$8A,$FD,$5F,$8C,$FD,$0F,$9D,$F8,$12,$4F,$CF,$F2,$3F,$03,$A7
	.dc.b	$DB,$24,$7C,$EF,$6B,$EF,$7B,$5F,$7F,$DA,$93,$F4,$A3,$E6,$E3,$F3
	.dc.b	$09,$1F,$31,$AF,$CC,$6B,$F3,$1A,$FC,$C6,$BF,$31,$FC,$DF,$A3,$F9
	.dc.b	$9F,$5A,$4F,$E1,$E3,$FD,$C3,$9F,$FB,$82,$3F,$C6,$EB,$C0,$FF,$29
	.dc.b	$76,$BC,$7F,$F2,$A4,$E6,$BF,$DC,$0F,$19,$D4,$7E,$6E,$1F,$11,$CA
	.dc.b	$B8,$B6,$74,$D7,$FD,$99,$F1,$1F,$C4,$3D,$9F,$D8,$12,$3F,$3D,$EC
	.dc.b	$FA,$34,$E1,$89,$A6,$2B,$33,$AF,$66,$5F,$A2,$3F,$3B,$D7,$F9,$C2
	.dc.b	$47,$E7,$1A,$FE,$71,$AF,$E7,$1A,$FE,$71,$AF,$E7,$1F,$E1,$73,$FC
	.dc.b	$27,$E2,$64,$FE,$DE,$3E,$8F,$8D,$EA,$F0,$A4,$FC,$C8,$F6,$F8,$52
	.dc.b	$3E,$7B,$F0,$3B,$3F,$9A,$F1,$7F,$15,$24,$FF,$03,$E0,$FF,$45,$78
	.dc.b	$FF,$52,$1F,$37,$FD,$B0,$9E,$A7,$CA,$8F,$97,$E7,$FC,$CF,$2A,$47
	.dc.b	$DE,$8F,$AD,$F9,$13,$E1,$FF,$C0,$AD,$9F,$F0,$4B,$5F,$F0,$8B,$5F
	.dc.b	$F0,$B3,$5F,$F4,$F3,$5F,$E4,$4D,$4D,$24,$78,$09,$1F,$C0,$F9,$FF
	.dc.b	$9B,$7C,$74,$8F,$E4,$91,$F0,$BA,$7F,$D5,$99,$BD,$FF,$C8,$2F,$7E
	.dc.b	$98,$8C,$FB,$F3,$0F,$E6,$DD,$BF,$2E,$D3,$CB,$DF,$FE,$CF,$F1,$DF
	.dc.b	$D1,$C9,$7F,$CD,$7E,$3B,$C4,$FF,$3F,$BF,$F5,$23,$D3,$3F,$2F,$B7
	.dc.b	$E4,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$92,$F8,$1C,$E3,$EA,$D7,$D4,$24
	.dc.b	$7D,$46,$BF,$51,$AF,$D4,$6B,$F5,$1A,$FD,$47,$FC,$8B,$C8,$FD,$73
	.dc.b	$F1,$9D,$1F,$C0,$23,$1E,$5F,$FB,$FA,$7F,$A2,$4E,$5D,$E7,$F4,$BE
	.dc.b	$2F,$E8,$F3,$CC,$BD,$67,$E5,$A3,$EA,$FC,$5C,$8C,$EF,$FE,$EA,$E4
	.dc.b	$FA,$5F,$17,$FE,$FE,$C9,$E7,$78,$49,$7F,$EC,$78,$79,$BF,$4A,$1F
	.dc.b	$97,$7F,$79,$D7,$FD,$E0,$D8,$FE,$E8,$D7,$7F,$1B,$FC,$5E,$3F,$BB
	.dc.b	$0F,$6F,$06,$7D,$CF,$D3,$F5,$7E,$98,$97,$FA,$66,$BF,$A6,$6B,$FA
	.dc.b	$66,$BF,$A6,$6B,$FA,$67,$E9,$E4,$F4,$BE,$87,$AB,$2B,$E0,$47,$BF
	.dc.b	$1F,$F7,$2C,$7F,$BD,$17,$C6,$FA,$CF,$67,$C6,$FB,$7E,$36,$57,$D4
	.dc.b	$A4,$7C,$1F,$8D,$FA,$1E,$D7,$DF,$25,$7C,$FF,$6B,$C1,$9F,$C7,$24
	.dc.b	$73,$9F,$95,$3F,$2F,$D9,$93,$DE,$3F,$A2,$7E,$7B,$FE,$4A,$8F,$F4
	.dc.b	$DB,$5F,$F5,$03,$5F,$F1,$50,$70,$FF,$8B,$5A,$E7,$BB,$FD,$3B,$18
	.dc.b	$23,$82,$47,$FA,$92,$39,$C7,$E8,$6B,$F3,$DA,$4A,$FC,$74,$7E,$37
	.dc.b	$F3,$0F,$F7,$2C,$1F,$F7,$2A,$E9,$FE,$69,$18,$6F,$95,$FD,$8E,$3F
	.dc.b	$DC,$B1,$AF,$17,$07,$E4,$7A,$3F,$41,$C5,$FA,$F2,$61,$F9,$FE,$1F
	.dc.b	$17,$AB,$CC,$91,$FA,$EC,$7F,$BD,$3E,$3B,$F9,$31,$FD,$E2,$35,$E5
	.dc.b	$F8,$CF,$27,$9F,$F5,$3D,$AF,$92,$49,$F8,$CD,$7E,$BB,$5F,$AE,$D7
	.dc.b	$EB,$F0,$A3,$11,$F5,$DF,$CE,$F2,$FF,$0B,$F8,$29,$3F,$BB,$8F,$9D
	.dc.b	$EF,$7C,$7A,$93,$ED,$8F,$95,$EF,$48,$F8,$0F,$CD,$F7,$BF,$84,$7C
	.dc.b	$77,$F1,$72,$4F,$F0,$7F,$73,$ED,$7B,$5F,$4C,$91,$F2,$7D,$BF,$AD
	.dc.b	$ED,$FD,$7F,$6E,$4F,$CD,$8F,$CC,$F2,$7E,$38,$91,$F8,$B6,$BF,$9C
	.dc.b	$6B,$F9,$C6,$BF,$19,$AF,$E6,$DF,$E0,$3F,$05,$FC,$BF,$E1,$24,$FF
	.dc.b	$1F,$8F,$E3,$DD,$EF,$8C,$FF,$17,$3F,$ED,$A9,$FC,$01,$FE,$91,$71
	.dc.b	$D9,$FE,$07,$78,$47,$F3,$45,$F1,$9C,$89,$19,$F8,$CF,$E6,$5E,$5F
	.dc.b	$F4,$12,$47,$F3,$1F,$2B,$C7,$ED,$F8,$89,$1D,$7D,$CF,$2F,$B9,$E6
	.dc.b	$F3,$64,$F8,$63,$FC,$8D,$F1,$7F,$E1,$B8,$FD,$F9,$FE,$CC,$F0,$AB
	.dc.b	$FC,$0A,$34,$D8,$CB,$FE,$AF,$8C,$B6,$32,$F3,$37,$1A,$70,$76,$70
	.dc.b	$48,$FE,$67,$E7,$FF,$7E,$F2,$9F,$AE,$3D,$6C,$37,$0F,$EA,$95,$F5
	.dc.b	$B0,$FA,$BF,$19,$27,$E8,$47,$D8,$F8,$B9,$1F,$4D,$FB,$7F,$19,$FE
	.dc.b	$6B,$CB,$BF,$B5,$D8,$95,$FE,$CD,$F6,$FF,$BB,$66,$F6,$FF,$0B,$AF
	.dc.b	$E1,$09,$5F,$7E,$FF,$0F,$3F,$88,$F6,$A4,$FE,$0E,$3E,$9F,$47,$C8
	.dc.b	$D4,$D2,$4F,$CA,$6B,$F2,$5A,$FD,$A8,$E2,$EC,$86,$CF,$BB,$9F,$CC
	.dc.b	$BF,$97,$32,$7D,$DC,$1F,$EE,$BA,$D4,$60,$FF,$77,$7E,$64,$F2,$BF
	.dc.b	$D2,$0F,$FB,$8F,$AF,$FC,$CC,$DB,$FE,$96,$EE,$BF,$C2,$AB,$C4,$E2
	.dc.b	$48,$CF,$77,$D3,$E0,$72,$7E,$00,$93,$E3,$CF,$F0,$BC,$9F,$E1,$3C
	.dc.b	$7F,$BA,$E7,$E2,$F5,$3C,$A7,$9B,$D1,$F3,$FA,$3E,$8F,$36,$65,$E4
	.dc.b	$8F,$D2,$F9,$3F,$4C,$91,$F9,$A6,$BF,$99,$6B,$F4,$5A,$FE,$A5,$AF
	.dc.b	$E9,$DF,$F2,$0E,$AE,$9F,$CD,$FB,$D2,$BD,$93,$97,$E8,$0D,$1E,$F7
	.dc.b	$D0,$24,$7C,$CF,$77,$E9,$FB,$BF,$53,$DC,$93,$F4,$23,$E6,$7C,$0F
	.dc.b	$EA,$89,$1E,$0F,$80,$FF,$1E,$7E,$D1,$FD,$C0,$FE,$D9,$D3,$FE,$04
	.dc.b	$6C,$83,$BF,$6B,$FD,$A4,$FF,$70,$9E,$11,$F8,$67,$FD,$AD,$FB,$77
	.dc.b	$EE,$6B,$F8,$7B,$5F,$E2,$6D,$7F,$C2,$6D,$7C,$0D,$4D,$27,$85,$FC
	.dc.b	$49,$B1,$FC,$61,$AF,$F2,$0E,$9F,$E1,$BF,$07,$FC,$33,$C9,$93,$F4
	.dc.b	$A3,$E1,$FC,$67,$B0,$91,$FE,$6E,$CD,$F1,$9F,$03,$E3,$3E,$0F,$C6
	.dc.b	$4A,$D4,$7E,$6F,$CD,$FE,$32,$48,$FB,$FE,$67,$D0,$F6,$FE,$4F,$B9
	.dc.b	$27,$B4,$7C,$CF,$72,$47,$85,$F9,$FE,$E7,$CC,$EE,$F8,$04,$9F,$80
	.dc.b	$D7,$E7,$B5,$F9,$ED,$7E,$7B,$5F,$9E,$FE,$63,$D1,$FE,$63,$EB,$49
	.dc.b	$FC,$BC,$7F,$A0,$FF,$35,$7C,$67,$F9,$2B,$C8,$FF,$74,$46,$2B,$FD
	.dc.b	$2A,$D7,$FB,$49,$F8,$A3,$F1,$3C,$DF,$8A,$F8,$C9,$1F,$3E,$3F,$82
	.dc.b	$7C,$97,$F4,$22,$47,$EA,$3D,$AF,$16,$7E,$1E,$CF,$F2,$DE,$3F,$DD
	.dc.b	$1B,$79,$BF,$CD,$4F,$F7,$67,$C6,$FF,$76,$7C,$F7,$B5,$19,$77,$ED
	.dc.b	$7E,$73,$8F,$F2,$04,$8F,$C8,$35,$FC,$FB,$5F,$CF,$B5,$FC,$FB,$5F
	.dc.b	$CF,$BF,$C2,$23,$FB,$0E,$B2,$7F,$83,$C7,$D1,$F1,$FC,$D7,$FE,$E4
	.dc.b	$9F,$5C,$63,$C1,$CF,$83,$AE,$D9,$5F,$52,$3F,$B1,$7B,$9F,$E0,$4F
	.dc.b	$63,$FD,$A8,$DD,$FF,$A2,$36,$9C,$28,$FF,$47,$57,$E0,$3D,$4F,$1F
	.dc.b	$A3,$2F,$F0,$31,$E6,$F4,$64,$7E,$11,$F4,$F4,$7C,$FF,$D0,$3F,$C1
	.dc.b	$72,$69,$1F,$F0,$7B,$5F,$F0,$AB,$5F,$F0,$DB,$5F,$F0,$FB,$5F,$5F
	.dc.b	$3F,$F8,$3F,$27,$C0,$D3,$11,$FE,$DF,$BF,$F6,$D1,$3D,$EF,$C2,$2F
	.dc.b	$7A,$FB,$D8,$F7,$A6,$5F,$EF,$4A,$4F,$CE,$F8,$1F,$01,$27,$FA,$E6
	.dc.b	$6F,$80,$F0,$7B,$9D,$BD,$69,$5C,$E3,$C3,$EE,$48,$EC,$F9,$3D,$CF
	.dc.b	$95,$E3,$FC,$62,$4F,$C6,$6B,$F1,$9A,$FC,$66,$BF,$EC,$4E,$24,$62
	.dc.b	$3F,$A1,$46,$1F,$F0,$6E,$4F,$81,$F9,$95,$27,$DB,$1F,$2F,$E2,$7E
	.dc.b	$49,$23,$E3,$7C,$4F,$CC,$F8,$1F,$9B,$F1,$32,$7E,$54,$7E,$5B,$CD
	.dc.b	$FB,$04,$8F,$93,$E6,$FD,$7F,$6F,$EA,$FB,$72,$7E,$94,$7D,$8F,$6A
	.dc.b	$47,$D5,$7E,$DF,$B7,$F9,$1F,$53,$F1,$A4,$9F,$C6,$B5,$FC,$6B,$5F
	.dc.b	$C6,$B5,$F9,$0D,$7E,$3B,$FE,$6A,$F0,$FF,$51,$E9,$7E,$47,$D7,$95
	.dc.b	$F8,$28,$FF,$79,$F8,$AF,$1B,$FD,$EB,$FB,$57,$EB,$EB,$FB,$97,$A7
	.dc.b	$97,$FB,$A4,$65,$FE,$5A,$D8,$D4,$23,$FD,$B3,$18,$7F,$DB,$11,$C3
	.dc.b	$E2,$3F,$90,$41,$BF,$11,$FC,$1B,$D9,$FD,$41,$2B,$F3,$3E,$CF,$A3
	.dc.b	$1F,$3E,$9C,$AA,$57,$A2,$72,$D3,$D3,$5E,$02,$39,$3D,$1B,$7F,$05
	.dc.b	$F7,$BF,$80,$92,$3F,$80,$B5,$FE,$02,$D7,$F8,$0B,$5F,$CE,$35,$FC
	.dc.b	$DB,$F5,$7B,$13,$FC,$15,$93,$FB,$68,$F7,$3F,$43,$CA,$F8,$E4,$7D
	.dc.b	$1A,$BC,$4F,$97,$E2,$7C,$CE,$E9,$5A,$8F,$E6,$3F,$21,$FC,$4C,$91
	.dc.b	$FC,$13,$83,$BA,$FC,$E3,$FB,$57,$1F,$D9,$8F,$AE,$7F,$57,$8C,$F0
	.dc.b	$7E,$23,$9B,$ED,$92,$7C,$BE,$67,$F6,$3F,$A7,$7D,$3F,$F9,$DE,$5D
	.dc.b	$1F,$8D,$9C,$F9,$1F,$01,$B3,$FE,$85,$78,$F9,$3E,$6F,$CB,$1E,$2F
	.dc.b	$8F,$DB,$97,$F6,$E7,$2F,$E0,$74,$FE,$95,$DB,$FD,$4B,$E3,$E4,$E7
	.dc.b	$AB,$F2,$4F,$1C,$7F,$D9,$F8,$72,$8F,$F6,$CB,$5F,$F6,$EB,$5F,$F2
	.dc.b	$C0,$70,$FF,$96,$DA,$E6,$3A,$91,$FF,$70,$C6,$89,$1E,$2F,$1E,$4E
	.dc.b	$9A,$F9,$1A,$F9,$1A,$F9,$25,$18,$8F,$23,$5F,$47,$C5,$7C,$02,$3E
	.dc.b	$46,$BF,$09,$AF,$C2,$6B,$DD,$AF,$C1,$7E,$5F,$C5,$48,$F7,$B5,$F9
	.dc.b	$AD,$7E,$6B,$5F,$53,$5F,$98,$DE,$37,$EF,$CF,$1F,$E9,$D7,$D4,$9C
	.dc.b	$C6,$23,$FD,$1D,$18,$70,$D7,$8B,$5E,$53,$D5,$DB,$5D,$B5,$DC,$F7
	.dc.b	$7F,$97,$46,$5F,$0B,$5C,$C6,$F8,$7F,$4F,$E9,$FE,$8F,$95,$EA,$53
	.dc.b	$11,$FC,$1B,$1F,$83,$F1,$7F,$E2,$EF,$94,$99,$7F,$21,$9C,$BF,$9F
	.dc.b	$F1,$3F,$CA,$DC,$AF,$D8,$3F,$9C,$F6,$FF,$3F,$8C,$5F,$F8,$E2,$BF
	.dc.b	$CA,$DC,$6C,$BF,$4A,$4C,$7F,$32,$D7,$F4,$ED,$7F,$4E,$D7,$F2,$8D
	.dc.b	$7F,$4C,$FE,$C3,$D4,$91,$FA,$46,$BF,$B2,$6B,$FB,$26,$BF,$9F,$6B
	.dc.b	$FB,$16,$BF,$B5,$D3,$F7,$A4,$7F,$62,$D7,$F7,$ED,$7F,$7E,$D7,$F5
	.dc.b	$8D,$7F,$7C,$FD,$F5,$E5,$FF,$B2,$C9,$F1,$59,$6C,$7E,$15,$AF,$E5
	.dc.b	$5A,$FE,$69,$AF,$F0,$E7,$A7,$C7,$73,$24,$73,$6B,$CD,$AF,$36,$BC
	.dc.b	$DA,$FF,$BC,$61,$1E,$7F,$6A,$4F,$9D,$AF,$9D,$AF,$9D,$AF,$9D,$AF
	.dc.b	$56,$BF,$1B,$DF,$F8,$84,$7E,$23,$5F,$88,$D7,$E2,$35,$F8,$8D,$7D
	.dc.b	$4F,$CF,$F8,$A9,$1F,$3D,$AF,$CF,$6B,$F3,$DA,$FC,$F6,$BF,$15,$AF
	.dc.b	$DE,$F8,$8F,$B8,$47,$EE,$35,$FB,$8D,$7E,$E3,$5F,$B8,$D7,$E9,$3F
	.dc.b	$87,$F0,$24,$7E,$1D,$AF,$E1,$DA,$FE,$1D,$AF,$E1,$DA,$FD,$D6,$AF
	.dc.b	$CA,$7F,$D4,$64,$7F,$34,$D7,$F8,$D3,$5F,$E5,$4D,$7F,$97,$B5,$FE
	.dc.b	$C6,$F4,$F3,$79,$92,$39,$B5,$E6,$D7,$9B,$5E,$6D,$47,$A5,$F3,$F9
	.dc.b	$72,$3C,$ED,$7C,$ED,$7C,$ED,$7C,$ED,$7A,$CF,$FA,$3D,$E6,$7C,$2F
	.dc.b	$57,$E2,$91,$C1,$0F,$8A,$E5,$A6,$4F,$8E,$E5,$A7,$CE,$F4,$7E,$89
	.dc.b	$1C,$10,$FA,$2E,$5A,$64,$FA,$6E,$5A,$7D,$AF,$0F,$EE,$91,$C1,$0F
	.dc.b	$BA,$E5,$A6,$4F,$BE,$E5,$A7,$E1,$BC,$5F,$C4,$91,$C1,$0F,$C4,$B9
	.dc.b	$69,$93,$F1,$6E,$5A,$7E,$4B,$9B,$F2,$A4,$70,$43,$F2,$AE,$5A,$64
	.dc.b	$FC,$BB,$96,$9F,$DD,$FF,$28,$7A,$DF,$CE,$DB,$19,$8C,$11,$C1,$0F
	.dc.b	$F0,$74,$70,$6C,$6D,$A6,$CF,$F2,$4F,$81,$FE,$66,$F7,$7F,$C0,$04
	.dc.b	$8F,$F3,$17,$82,$F7,$FF,$D0,$BD,$DF,$DE,$70,$FE,$F7,$E0,$7F,$84
	.dc.b	$FC,$0F,$EF,$4D,$F0,$3F,$4F,$5B,$C7,$FC,$C9,$2F,$63,$89,$23,$3D
	.dc.b	$9F,$E5,$CC,$3F,$D7,$BB,$7F,$85,$E4,$D2,$78,$9F,$1C,$91,$F2,$DA
	.dc.b	$FC,$46,$BF,$E3,$68,$E2,$FD,$1B,$FA,$04,$70,$43,$E6,$B9,$69,$93
	.dc.b	$A3,$96,$9F,$76,$FE,$E1,$1C,$10,$FB,$2E,$5A,$64,$F4,$39,$69,$F8
	.dc.b	$9F,$6F,$F1,$04,$70,$43,$F0,$AE,$5A,$64,$F8,$8E,$5A,$7F,$96,$78
	.dc.b	$3F,$2B,$F2,$5F,$EA,$7D,$9E,$05,$31,$1F,$86,$D5,$87,$E1,$89,$87
	.dc.b	$E0,$BC,$BF,$C4,$FB,$92,$3F,$0A,$FE,$33,$CC,$FC,$17,$17,$FA,$7B
	.dc.b	$7F,$F4,$D9,$2F,$F4,$C4,$BF,$D5,$35,$FD,$1B,$5F,$C0,$35,$E8,$FE
	.dc.b	$BE,$A4,$7E,$CD,$AF,$EB,$1A,$FE,$25,$AF,$95,$AF,$F4,$69,$FF,$50
	.dc.b	$BC,$CF,$DF,$7C,$8F,$EF,$48,$E0,$87,$EE,$9C,$B4,$C9,$F9,$97,$2D
	.dc.b	$3F,$85,$79,$FF,$C2,$48,$E0,$87,$F0,$67,$2D,$32,$7E,$95,$CB,$4F
	.dc.b	$E2,$DE,$5F,$F1,$52,$38,$21,$FC,$49,$CB,$4C,$9F,$B1,$72,$D3,$F9
	.dc.b	$17,$37,$F2,$12,$38,$21,$FC,$79,$CB,$4C,$9F,$BD,$72,$D3,$FD,$49
	.dc.b	$F9,$43,$E4,$FF,$CD,$0D,$8C,$C6,$08,$E0,$87,$FA,$EE,$38,$36,$36
	.dc.b	$D3,$67,$FB,$12,$1F,$0B,$E0,$24,$79,$1A,$F5,$6B,$C5,$AF,$FC,$00
	.dc.b	$8E,$6F,$12,$1E,$5F,$88,$F6,$91,$F8,$2D,$74,$60,$86,$08,$7A,$5E
	.dc.b	$0F,$AF,$E2,$3E,$49,$1F,$9A,$D7,$E3,$35,$E0,$68,$86,$8A,$7C,$7F
	.dc.b	$8C,$FA,$A4,$7E,$CB,$5F,$A4,$D7,$81,$A2,$1A,$29,$F4,$FE,$37,$F0
	.dc.b	$24,$7F,$0A,$D7,$EF,$35,$E0,$68,$86,$8A,$7D,$FF,$8D,$FC,$69,$1F
	.dc.b	$C8,$B5,$FC,$53,$5E,$06,$88,$68,$A7,$F9,$48,$FF,$29,$FB,$CF,$6F
	.dc.b	$FD,$84,$D8,$D1,$B7,$0E,$03,$87,$6D,$78,$47,$42,$3C,$4E,$2E,$1E
	.dc.b	$91,$C5,$F2,$C7,$92,$24,$71,$7C,$2D,$72,$F2,$6B,$E4,$6B,$EB,$6B
	.dc.b	$F0,$9F,$83,$F0,$B2,$3B,$35,$F2,$B5,$F7,$B5,$F8,$CD,$7E,$6B,$5F
	.dc.b	$8F,$F0,$9F,$38,$8F,$AD,$AF,$C2,$6B,$F3,$1A,$FD,$36,$BF,$69,$FB
	.dc.b	$3F,$07,$23,$E3,$35,$F9,$AD,$7E,$BB,$5F,$BC,$D7,$F0,$AD,$7E,$FF
	.dc.b	$C0,$7E,$18,$8F,$D3,$6B,$F6,$9A,$FE,$11,$AF,$E2,$DA,$FE,$49,$FC
	.dc.b	$27,$7F,$F9,$6B,$D4,$BE,$07,$FB,$F2,$BF,$C1,$4D,$7F,$2B,$3B,$73
	.dc.b	$3C,$08,$E0,$87,$09,$E4,$FF,$96,$9A,$FF,$AD,$A7,$A3,$F9,$4F,$77
	.dc.b	$F4,$A4,$7F,$24,$D7,$46,$08,$60,$87,$E9,$9E,$0F,$EA,$3D,$AF,$D8
	.dc.b	$91,$FC,$F3,$5F,$D2,$B5,$E0,$68,$86,$8A,$7E,$FF,$C7,$78,$2F,$77
	.dc.b	$FC,$5C,$49,$FF,$6A,$FB,$FF,$E7,$2F,$7B,$F8,$32,$F0,$BF,$84,$7E
	.dc.b	$7A,$EE,$FF,$16,$DF,$FB,$72,$1E,$DE,$21,$18,$7F,$DC,$3E,$67,$37
	.dc.b	$85,$6E,$5E,$75,$24,$F0,$C3,$C3,$E0,$F5,$FC,$93,$FD,$2D,$3F,$C5
	.dc.b	$9A,$FF,$1D,$6B,$FE,$F0,$6A,$69,$38,$FF,$8A,$46,$5B,$19,$8F,$E3
	.dc.b	$0D,$7F,$15,$EE,$7E,$58,$8F,$E5,$9A,$E8,$C1,$0C,$10,$FC,$93,$C1
	.dc.b	$F6,$79,$FE,$22,$3E,$26,$BF,$CC,$E7,$FD,$3C,$48,$CC,$72,$38,$10
	.dc.b	$E1,$E7,$FB,$88,$FB,$9A,$E8,$C1,$0C,$10,$FA,$4F,$07,$EA,$79,$FF
	.dc.b	$28,$8F,$CA,$6B,$E9,$6B,$C0,$D1,$0D,$14,$FC,$07,$97,$F5,$88,$FD
	.dc.b	$66,$BF,$15,$AF,$03,$44,$34,$53,$F1,$9C,$7F,$82,$23,$F8,$27,$AF
	.dc.b	$D6,$FE,$6E,$3F,$19,$3F,$DB,$63,$18,$A3,$1A,$FE,$57,$DF,$B7,$F9
	.dc.b	$5C,$3A,$BF,$6D,$AF,$E7,$63,$F2,$64,$7F,$26,$D7,$F0,$4D,$78,$1A
	.dc.b	$21,$A2,$9F,$A9,$F7,$FF,$3E,$47,$F3,$EF,$E8,$B9,$3C,$F1,$FA,$99
	.dc.b	$B8,$9B,$38,$AE,$55,$F9,$56,$BF,$B7,$8E,$8F,$EB,$1A,$FE,$6D,$AE
	.dc.b	$CD,$10,$D1,$4D,$1F,$E1,$5F,$C8,$FF,$D9,$AF,$F8,$3D,$AE,$7D,$3F
	.dc.b	$DF,$C6,$23,$8E,$9F,$62,$BD,$70,$11,$F7,$E5,$F0,$1F,$11,$E5,$3F
	.dc.b	$E2,$D6,$BC,$5A,$9A,$7F,$50,$75,$DF,$E4,$23,$EF,$AF,$23,$E3,$6B
	.dc.b	$E4,$D7,$C1,$7E,$3A,$E3,$3F,$0D,$AF,$F8,$F2,$DC,$5F,$08,$91,$F1
	.dc.b	$DF,$8B,$5F,$0A,$3B,$1F,$CA,$F1,$F8,$11,$F0,$67,$8C,$FD,$16,$BD
	.dc.b	$E3,$9B,$F4,$DA,$FD,$76,$BB,$34,$43,$45,$34,$7D,$FF,$80,$FB,$64
	.dc.b	$7F,$00,$D7,$46,$08,$60,$87,$C5,$78,$3F,$1F,$E0,$3F,$0E,$47,$F1
	.dc.b	$8F,$5C,$FF,$09,$3F,$83,$3F,$4F,$F9,$13,$E0,$3F,$0D,$5C,$6F,$F2
	.dc.b	$CD,$7E,$A7,$C0,$FE,$50,$8F,$E6,$5F,$EA,$B5,$CE,$3F,$05,$B7,$E3
	.dc.b	$F0,$F0,$46,$0F,$CC,$4E,$DF,$B2,$7E,$46,$BA,$BF,$A7,$6B,$FE,$80
	.dc.b	$8F,$EC,$69,$A4,$7C,$78,$78,$1E,$BA,$AF,$E0,$CF,$C2,$9F,$49,$0F
	.dc.b	$F4,$15,$74,$8E,$71,$86,$BF,$E8,$F8,$F5,$3E,$58,$CB,$4E,$04,$38
	.dc.b	$3A,$7F,$D4,$10,$F8,$1F,$96,$48,$F1,$BE,$0C,$FF,$9E,$4F,$C9,$6D
	.dc.b	$7F,$22,$FD,$B1,$E6,$FD,$EB,$E5,$87,$9D,$AF,$CB,$F8,$1F,$B2,$47
	.dc.b	$E0,$BD,$F5,$F2,$E1,$F2,$27,$ED,$0C,$3B,$B8,$C9,$7C,$E3,$E3,$35
	.dc.b	$FB,$5F,$03,$F8,$82,$3F,$41,$F0,$3F,$86,$BF,$03,$F8,$6B,$EE,$FE
	.dc.b	$22,$D5,$CA,$3E,$A3,$5F,$C4,$7C,$0F,$E5,$88,$FD,$E7,$BB,$F8,$BB
	.dc.b	$EF,$A7,$F6,$9C,$3B,$BF,$42,$B8,$D7,$E0,$5A,$FE,$57,$CB,$FD,$19
	.dc.b	$1F,$C5,$B5,$FC,$43,$5E,$06,$88,$68,$A7,$E8,$7E,$07,$F5,$E4,$7F
	.dc.b	$2E,$D7,$F2,$8D,$78,$1A,$21,$A2,$9F,$AD,$F2,$FF,$76,$47,$F4,$6D
	.dc.b	$7F,$40,$D7,$81,$A2,$1A,$29,$FB,$9F,$2F,$F8,$39,$1F,$D7,$B5,$FD
	.dc.b	$63,$5E,$06,$88,$68,$A7,$F0,$5F,$81,$FE,$26,$47,$F7,$6D,$7F,$70
	.dc.b	$D7,$81,$A2,$1A,$29,$FE,$B3,$E6,$FF,$60,$91,$FF,$60,$B5,$FF,$5D
	.dc.b	$35,$E0,$68,$86,$8A,$7F,$7B,$3F,$BE,$7B,$1F,$E3,$97,$01,$C3,$B6
	.dc.b	$BC,$38,$7F,$D0,$F1,$88,$E1,$18,$23,$D0,$E0,$53,$65,$8E,$44,$F0
	.dc.b	$FA,$92,$34,$F8,$3D,$AF,$E2,$5F,$23,$27,$F8,$3D,$27,$FC,$6E,$AB
	.dc.b	$94,$F8,$E3,$0F,$43,$C5,$F0,$1E,$C2,$38,$3C,$D5,$DD,$FE,$0D,$7D
	.dc.b	$A3,$BC,$F0,$9E,$CD,$7E,$6D,$3D,$FF,$92,$48,$F8,$4F,$B5,$FE,$43
	.dc.b	$C4,$C3,$C1,$C5,$FE,$6F,$BF,$75,$7C,$0B,$E4,$FB,$5A,$F8,$3D,$EF
	.dc.b	$AE,$47,$E8,$3D,$DF,$59,$E0,$BF,$63,$F3,$23,$83,$F2,$5A,$FB,$32
	.dc.b	$E4,$D3,$E3,$90,$FA,$0E,$9F,$C6,$35,$FC,$9B,$5F,$C9,$3F,$8E,$F3
	.dc.b	$64,$7E,$1D,$AF,$43,$F0,$31,$E8,$7F,$9D,$78,$5F,$82,$9D,$C7,$D2
	.dc.b	$6B,$F9,$DF,$2F,$F3,$64,$7F,$28,$FC,$5C,$FE,$85,$F5,$7F,$9C,$5F
	.dc.b	$E3,$A7,$85,$7E,$35,$AF,$EB,$3F,$44,$7B,$FF,$EE,$E6,$C7,$E8,$DA
	.dc.b	$FE,$9D,$AF,$CB,$3F,$C3,$3A,$FE,$01,$B8,$FF,$73,$C3,$FD,$2B,$5C
	.dc.b	$27,$F7,$CD,$7F,$CF,$4D,$4D,$3F,$80,$3E,$3A,$F3,$11,$FB,$58,$FC
	.dc.b	$27,$E7,$9F,$96,$AF,$0C,$EE,$3D,$4D,$7E,$17,$BD,$F0,$48,$FC,$17
	.dc.b	$E2,$77,$FF,$AB,$BC,$46,$D4,$F6,$C7,$BA,$78,$D7,$C7,$6B,$F3,$FD
	.dc.b	$EF,$9C,$47,$E7,$3D,$9F,$F0,$6D,$F6,$6D,$F6,$7F,$C2,$39,$AB,$94
	.dc.b	$7D,$56,$BF,$77,$DE,$FB,$84,$7E,$E3,$DD,$FF,$28,$DF,$76,$DF,$77
	.dc.b	$FC,$AB,$6A,$E5,$1F,$83,$6B,$F8,$BF,$7B,$F1,$44,$7F,$14,$F7,$D7
	.dc.b	$FD,$E9,$87,$76,$E1,$DD,$FF,$7C,$60,$AE,$57,$F9,$16,$BF,$99,$CF
	.dc.b	$D4,$D3,$F3,$0E,$1F,$C7,$B5,$FC,$7B,$5F,$C7,$B6,$7F,$C5,$B1,$FD
	.dc.b	$43,$F4,$47,$BF,$FD,$16,$11,$87,$FA,$43,$5F,$E9,$8D,$7C,$87,$CA
	.dc.b	$AF,$77,$4F,$F5,$FC,$DE,$9F,$FB,$EA,$F8,$5F,$F6,$36,$BF,$ED,$F6
	.dc.b	$BE,$2C,$3C,$90,$F7,$E9,$27,$CA,$FD,$1C,$3E,$1C,$7E,$2B,$1F,$CE
	.dc.b	$61,$E1,$AE,$13,$EA,$6B,$F0,$FD,$FE,$E4,$7E,$0B,$D5,$F9,$A7,$53
	.dc.b	$F3,$73,$ED,$8E,$11,$F1,$9A,$FC,$FF,$7F,$D6,$47,$E6,$BD,$5F,$AE
	.dc.b	$75,$3F,$33,$3F,$26,$38,$47,$D2,$6B,$F6,$FE,$37,$E3,$91,$FB,$2D
	.dc.b	$7E,$CB,$5E,$06,$88,$68,$A7,$F8,$67,$DF,$FB,$7E,$67,$F8,$2D,$B1
	.dc.b	$FE,$0F,$6B,$FE,$1C,$6B,$FE,$21,$F7,$DE,$FF,$E9,$75,$4A,$73,$C9
	.dc.b	$FE,$27,$E4,$56,$A7,$35,$FE,$CF,$79,$FA,$BB,$9E,$78,$F8,$2A,$EB
	.dc.b	$FE,$C5,$7C,$CD,$7C,$CD,$7D,$CB,$D2,$F7,$36,$38,$C6,$08,$E0,$87
	.dc.b	$47,$2D,$32,$76,$7F,$3B,$F0,$12,$3F,$5E,$D7,$F5,$ED,$7F,$70,$D7
	.dc.b	$F5,$4D,$7F,$54,$D7,$F7,$7E,$67,$EA,$88,$FE,$F1,$AF,$EF,$1A,$FF
	.dc.b	$05,$6B,$FB,$76,$BF,$B7,$7F,$6B,$F1,$72,$3F,$83,$B5,$FE,$0E,$D7
	.dc.b	$F8,$83,$5F,$E0,$4D,$7F,$81,$37,$0F,$E1,$3E,$A7,$D2,$24,$FD,$A7
	.dc.b	$FC,$B9,$A7,$E3,$3E,$2B,$F0,$53,$F8,$BE,$04,$62,$75,$3F,$89,$9F
	.dc.b	$EF,$9B,$FE,$3A,$18,$78,$1E,$C4,$3F,$0F,$87,$F1,$4F,$3E,$47,$F2
	.dc.b	$66,$BF,$C9,$9A,$FF,$2F,$6B,$FC,$85,$AF,$F2,$16,$BF,$CC,$BC,$FF
	.dc.b	$E4,$24,$7F,$99,$B5,$FE,$66,$D7,$F9,$E3,$5F,$E5,$CD,$7F,$97,$3F
	.dc.b	$CB,$3C,$F9,$1F,$CF,$9A,$FF,$3E,$6B,$FD,$1D,$AF,$F3,$96,$BF,$CE
	.dc.b	$5A,$FF,$4A,$F3,$FF,$9C,$91,$FE,$96,$D7,$FA,$5B,$5F,$EA,$8D,$7F
	.dc.b	$A3,$35,$FE,$8C,$FF,$44,$F3,$E4,$7F,$56,$6B,$FD,$59,$AF,$F5,$F6
	.dc.b	$BF,$D4,$5A,$FF,$51,$6A,$FD,$BF,$FD,$5C,$4E,$4F,$EC,$CD,$8F,$F5
	.dc.b	$63,$5F,$F5,$83,$5F,$F5,$EB,$5F,$F6,$13,$D3,$93,$99,$23,$9B,$5E
	.dc.b	$6D,$79,$B5,$E6,$D7,$F9,$5C,$7F,$70,$7C,$BE,$7F,$A0,$8E,$08,$7A
	.dc.b	$1C,$B4,$C9,$EA,$72,$D3,$E0,$FB,$DF,$10,$8E,$08,$7C,$47,$2D,$32
	.dc.b	$7C,$67,$2D,$3E,$6F,$BD,$F4,$08,$E0,$87,$D0,$72,$D3,$27,$D2,$72
	.dc.b	$D3,$EC,$FA,$3F,$70,$8E,$08,$7D,$C7,$2D,$32,$7D,$E7,$2D,$3F,$0B
	.dc.b	$EF,$7E,$20,$8E,$08,$7E,$21,$CB,$4C,$9F,$8A,$72,$D3,$F2,$3F,$0D
	.dc.b	$F9,$42,$38,$21,$F9,$47,$2D,$32,$7E,$59,$CB,$4F,$CE,$FA,$5F,$A0
	.dc.b	$23,$82,$1F,$A0,$72,$D3,$27,$E8,$9C,$B4,$FD,$4F,$BF,$FA,$C2,$38
	.dc.b	$21,$FA,$C7,$2D,$32,$7E,$B9,$CB,$4F,$F0,$69,$FE,$11,$E6,$7B,$D8
	.dc.b	$F8,$3F,$E8,$1D,$3F,$E1,$48,$C7,$0F,$F8,$3E,$B7,$7F,$E1,$6B,$DC
	.dc.b	$3E,$4B,$73,$FE,$1A,$FD,$0D,$CD,$E0,$F8,$5F,$E9,$92,$9C,$BF,$E2
	.dc.b	$08,$CC,$62,$11,$8E,$7E,$59,$7A,$A9,$CF,$8B,$F0,$35,$EC,$DF,$33
	.dc.b	$C3,$87,$8B,$9F,$FB,$1D,$7F,$18,$EE,$FE,$36,$47,$E3,$9F,$23,$95
	.dc.b	$F2,$BF,$DB,$23,$D5,$D9,$73,$39,$AD,$CF,$F8,$F3,$BF,$C1,$87,$58
	.dc.b	$FE,$4A,$D7,$EA,$7B,$1F,$2B,$97,$FC,$8A,$DE,$33,$7E,$0B,$E7,$D3
	.dc.b	$1E,$EF,$C8,$E1,$F8,$D9,$FC,$3D,$3C,$B3,$F3,$E7,$D9,$AF,$C7,$DB
	.dc.b	$E1,$CF,$13,$F9,$4E,$B9,$35,$F0,$BF,$3F,$6F,$F2,$CB,$E5,$F6,$A5
	.dc.b	$7E,$17,$1F,$C4,$4F,$9B,$4F,$2F,$0F,$DE,$D7,$C8,$7D,$78,$CD,$7F
	.dc.b	$44,$AF,$05,$F9,$CF,$19,$4F,$EA,$8F,$91,$FE,$A2,$FF,$54,$23,$F6
	.dc.b	$8F,$F3,$6C,$FC,$6E,$1F,$14,$7D,$5A,$F8,$ED,$37,$AF,$DF,$D5,$87
	.dc.b	$49,$FE,$BA,$D7,$FC,$F9,$FB,$13,$D9,$FB,$4D,$8F,$88,$7F,$9F,$5C
	.dc.b	$2E,$9F,$D8,$C3,$E2,$FE,$46,$3F,$1A,$BE,$07,$17,$DA,$7E,$B1,$F8
	.dc.b	$18,$F8,$4E,$32,$F3,$BF,$1D,$AF,$0D,$7E,$7B,$E3,$F7,$3C,$FF,$0D
	.dc.b	$23,$CF,$C6,$CD,$5E,$49,$F1,$F0,$F9,$5E,$99,$7D,$B3,$C7,$C5,$FA
	.dc.b	$FE,$0C,$4F,$D9,$CF,$F2,$EF,$EB,$9A,$FE,$50,$FB,$07,$B4,$AE,$9F
	.dc.b	$AA,$62,$BF,$DF,$11,$98,$E0,$7E,$05,$C1,$23,$FC,$CF,$8F,$EA,$37
	.dc.b	$FA,$3E,$E7,$D8,$3E,$CC,$FB,$EF,$DA,$7A,$5F,$D6,$79,$7E,$C8,$C9
	.dc.b	$32,$FE,$3F,$8B,$E6,$7F,$51,$93,$AF,$F0,$A4,$BF,$9F,$0F,$97,$F0
	.dc.b	$61,$CB,$2E,$CF,$F2,$6C,$AF,$37,$C7,$C3,$C0,$FF,$17,$CB,$F5,$9F
	.dc.b	$01,$F1,$09,$3F,$29,$F8,$F8,$7F,$17,$24,$7E,$84,$FE,$6D,$78,$5D
	.dc.b	$1E,$37,$5A,$FE,$AF,$1F,$04,$7F,$B4,$F4,$7B,$FF,$A2,$F8,$29,$3E
	.dc.b	$E8,$F6,$D7,$A7,$5E,$93,$F9,$CE,$86,$7F,$C1,$76,$F7,$CF,$E0,$0F
	.dc.b	$CD,$3F,$33,$6F,$D7,$92,$3E,$5F,$C0,$76,$78,$61,$D8,$F9,$4F,$E4
	.dc.b	$EB,$AB,$F8,$88,$FB,$31,$E1,$8F,$C8,$BF,$8E,$23,$F9,$13,$E1,$EF
	.dc.b	$F1,$F7,$EB,$1F,$23,$3F,$8A,$FE,$4E,$B8,$B4,$FC,$EF,$CC,$FE,$68
	.dc.b	$8E,$0F,$D1,$FE,$8C,$F9,$7F,$EA,$1E,$D4,$9F,$B1,$39,$7E,$AD,$72
	.dc.b	$7F,$44,$D7,$F8,$BF,$B5,$FC,$5B,$E6,$E4,$7E,$4A,$3F,$A6,$D7,$38
	.dc.b	$FA,$97,$FD,$2C,$FE,$DD,$19,$D7,$F3,$F8,$7F,$3B,$25,$7F,$37,$9D
	.dc.b	$66,$8F,$E4,$5F,$09,$FC,$98,$8F,$E8,$BD,$AF,$0E,$9D,$E3,$F9,$0E
	.dc.b	$5E,$08,$EB,$1F,$C6,$F1,$B9,$7F,$83,$B2,$EF,$3D,$D7,$CB,$FF,$0C
	.dc.b	$84,$61,$F9,$B1,$98,$C4,$7F,$49,$7E,$16,$FF,$E1,$6C,$7E,$84,$FD
	.dc.b	$D3,$B6,$3F,$E0,$5A,$DE,$9F,$E0,$9C,$BF,$C0,$86,$67,$FD,$E5,$CF
	.dc.b	$FC,$16,$FF,$80,$FC,$2F,$F4,$83,$FA,$4F,$0F,$C2,$8F,$82,$43,$FA
	.dc.b	$0E,$3E,$E7,$D7,$7F,$0B,$1F,$04,$7D,$18,$FE,$23,$EC,$FF,$54,$24
	.dc.b	$7F,$10,$CB,$B3,$F7,$8E,$DB,$7C,$BA,$FA,$11,$FC,$E6,$7B,$3F,$00
	.dc.b	$EC,$7E,$5A,$3E,$51,$1C,$10,$F9,$8E,$5A,$64,$FF,$18,$E1,$F9,$16
	.dc.b	$C7,$F3,$8F,$82,$FF,$13,$7F,$00,$FF,$8D,$BC,$0F,$F2,$CB,$73,$FC
	.dc.b	$77,$89,$FD,$66,$BD,$B9,$FA,$F1,$F3,$9F,$C3,$FE,$07,$DA,$FC,$CC
	.dc.b	$AC,$FF,$95,$3E,$2B,$FC,$41,$93,$1F,$09,$2F,$FB,$3C,$3E,$1B,$BE
	.dc.b	$7F,$E7,$0D,$7E,$B1,$F5,$FB,$9A,$76,$9D,$69,$71,$FF,$35,$67,$FA
	.dc.b	$EF,$80,$FC,$A1,$23,$F8,$1B,$F2,$79,$BD,$91,$F9,$33,$F0,$67,$F3
	.dc.b	$BC,$7B,$BF,$C1,$AF,$F2,$4D,$35,$7F,$E8,$6A,$F9,$7F,$15,$F3,$BA
	.dc.b	$25,$7B,$A3,$DA,$FE,$04,$F9,$07,$F8,$5B,$3F,$53,$E2,$32,$7F,$3D
	.dc.b	$BF,$97,$F0,$DF,$68,$91,$F2,$BE,$37,$AB,$C7,$8B,$DA,$7F,$A1,$AF
	.dc.b	$AE,$FF,$33,$4F,$F2,$F6,$3F,$E8,$1F,$DC,$3F,$D1,$F8,$7F,$9E,$9A
	.dc.b	$FF,$A8,$9A,$FF,$13,$6B,$E7,$D7,$E5,$93,$83,$FD,$21,$8F,$C8,$9F
	.dc.b	$9D,$8F,$EF,$9F,$85,$5F,$87,$7F,$B5,$5B,$E0,$7F,$88,$36,$3F,$C3
	.dc.b	$AD,$7F,$CA,$CD,$7D,$11,$C0,$8F,$06,$BC,$1A,$F0,$3A,$FC,$0F,$80
	.dc.b	$8E,$08,$78,$1C,$B4,$C9,$F2,$5C,$B4,$F4,$FC,$0F,$B0,$8E,$08,$7B
	.dc.b	$1C,$B4,$C9,$F5,$5C,$B4,$FF,$3C,$73,$39,$9F,$37,$FE,$44,$8C,$36
	.dc.b	$33,$3C,$1A,$FC,$CF,$2B,$F3,$9B,$7F,$04,$97,$C4,$E2,$BF,$79,$18
	.dc.b	$7E,$CD,$7F,$AF,$F4,$FF,$57,$47,$FA,$17,$1F,$27,$C7,$F1,$6C,$70
	.dc.b	$35,$7F,$C6,$5F,$07,$17,$9C,$8F,$E0,$BE,$13,$94,$FE,$DA,$3F,$A6
	.dc.b	$57,$C1,$E2,$F1,$54,$9E,$BC,$7C,$0C,$3C,$2D,$E1,$FC,$52,$47,$C5
	.dc.b	$7F,$27,$F0,$9E,$D2,$47,$A3,$E2,$2B,$FC,$2F,$54,$E3,$97,$F9,$7F
	.dc.b	$9B,$E8,$EE,$FA,$44,$BF,$A4,$FE,$83,$E1,$7E,$51,$23,$E2,$FA,$B5
	.dc.b	$ED,$16,$FF,$9D,$FC,$3F,$C4,$F8,$3F,$BC,$4A,$FB,$CF,$EB,$3E,$1B
	.dc.b	$EB,$12,$3E,$8F,$C2,$57,$D5,$16,$DD,$3F,$E1,$3E,$F7,$E1,$98,$BF
	.dc.b	$B3,$F0,$BF,$8D,$23,$F8,$D7,$F7,$7F,$29,$F8,$1F,$43,$84,$FD,$86
	.dc.b	$E3,$8E,$35,$7E,$2C,$FC,$84,$3F,$3D,$FF,$78,$BE,$13,$1F,$09,$BC
	.dc.b	$3F,$21,$83,$E2,$BF,$44,$D8,$FD,$3B,$5F,$F1,$71,$FD,$89,$C3,$FC
	.dc.b	$8F,$E4,$3F,$40,$BC,$C7,$83,$86,$C7,$F1,$A9,$FD,$11,$9F,$06,$46
	.dc.b	$4F,$2F,$8F,$E4,$23,$E4,$7F,$90,$FC,$07,$EA,$C9,$1F,$D5,$7C,$CA
	.dc.b	$FE,$3E,$2E,$BF,$33,$1F,$F1,$9B,$F3,$BF,$48,$FF,$73,$57,$FB,$B0
	.dc.b	$3F,$EE,$CF,$47,$5F,$0D,$F0,$9E,$1F,$0D,$7A,$3E,$2F,$E8,$AF,$53
	.dc.b	$E3,$F5,$77,$CA,$D7,$CA,$D3,$E5,$C7,$FB,$C2,$F8,$3F,$EF,$4F,$2B
	.dc.b	$6F,$F9,$76,$7F,$C0,$4F,$FB,$E3,$D4,$CB,$F3,$E7,$63,$95,$E8,$79
	.dc.b	$67,$FC,$1D,$18,$7F,$DE,$6D,$5A,$FE,$5E,$B9,$7A,$1C,$67,$E1,$FC
	.dc.b	$2F,$83,$87,$FD,$F3,$1F,$9F,$D7,$F0,$B1,$88,$6D,$F3,$7E,$1B,$85
	.dc.b	$C9,$E0,$7F,$9E,$6B,$85,$78,$73,$FE,$4B,$3C,$B0,$FC,$87,$9D,$EB
	.dc.b	$CF,$DC,$DA,$37,$CA,$F4,$62,$CF,$D0,$4C,$7C,$27,$DE,$F7,$F0,$F7
	.dc.b	$3C,$91,$8C,$FF,$55,$FA,$9F,$FB,$BE,$7D,$39,$FC,$FF,$27,$C4,$48
	.dc.b	$F9,$BE,$37,$8B,$1E,$5B,$7C,$6C,$BE,$54,$4A,$F8,$79,$7A,$A3,$F0
	.dc.b	$2D,$8F,$3E,$69,$FB,$FB,$7F,$02,$9F,$B7,$E1,$FC,$33,$E2,$7C,$8F
	.dc.b	$C2,$F8,$8F,$B0,$49,$E7,$F1,$15,$F8,$07,$E8,$23,$19,$7D,$3F,$91
	.dc.b	$EF,$E3,$7C,$F2,$57,$E2,$BC,$4F,$9F,$87,$E1,$49,$1F,$3B,$B3,$FD
	.dc.b	$F3,$E0,$7F,$AD,$FE,$6D,$ED,$FE,$22,$F7,$7E,$2E,$1F,$9C,$DF,$88
	.dc.b	$F7,$FC,$8F,$DF,$F9,$19,$3F,$9A,$8F,$0B,$EC,$21,$FA,$D9,$65,$E1
	.dc.b	$C7,$F2,$F1,$F9,$B2,$4F,$DC,$E3,$7C,$87,$EC,$BE,$4B,$F9,$37,$8B
	.dc.b	$FB,$8A,$FF,$68,$46,$1C,$35,$FD,$EB,$5F,$DF,$EF,$F0,$97,$E2,$FF
	.dc.b	$09,$FD,$16,$FF,$CB,$A3,$11,$FC,$4A,$31,$9F,$F0,$EC,$FD,$73,$CA
	.dc.b	$3F,$9A,$7C,$97,$F4,$9F,$2A,$4F,$F3,$E8,$FE,$99,$F2,$5F,$A7,$24
	.dc.b	$7F,$BA,$7C,$6F,$E0,$1F,$AF,$7F,$1F,$EF,$ED,$C5,$F1,$EB,$CA,$53
	.dc.b	$C8,$5D,$BD,$9C,$DE,$85,$F1,$1F,$C5,$9B,$3F,$83,$8C,$7C,$47,$F1
	.dc.b	$8C,$7F,$98,$CF,$F1,$2C,$7F,$D5,$F1,$FC,$2F,$2F,$2F,$91,$FC,$06
	.dc.b	$37,$1F,$D9,$B1,$F0,$EF,$F8,$BC,$3C,$DD,$3F,$8A,$D3,$C9,$C3,$F2
	.dc.b	$1F,$37,$C4,$F9,$7E,$2A,$5F,$E5,$E3,$F0,$1F,$9D,$7F,$B0,$61,$CE
	.dc.b	$8C,$36,$3F,$97,$39,$F2,$3F,$94,$69,$E4,$9F,$92,$31,$FB,$B9,$F9
	.dc.b	$74,$F9,$F3,$97,$E0,$F8,$DF,$14,$91,$FC,$43,$7F,$E0,$4F,$F2,$AC
	.dc.b	$BC,$9E,$77,$8B,$4F,$E8,$F3,$F2,$CF,$4B,$8E,$0F,$F0,$A5,$7A,$FD
	.dc.b	$2F,$3F,$0F,$C5,$CF,$E5,$61,$E3,$3D,$3A,$FC,$AE,$6F,$C2,$70,$35
	.dc.b	$78,$BF,$6C,$97,$E7,$EB,$F9,$7F,$17,$E5,$24,$7E,$47,$9F,$F5,$7C
	.dc.b	$0F,$8B,$F1,$E7,$F6,$B0,$FF,$08,$E9,$F2,$1B,$AF,$C8,$8F,$D5,$6B
	.dc.b	$E3,$F9,$3F,$C6,$13,$3F,$91,$BF,$DC,$BF,$8C,$48,$F1,$1F,$1B,$E2
	.dc.b	$FC,$2F,$C1,$3F,$3E,$5E,$6F,$B0,$87,$3F,$DF,$C7,$E2,$7C,$9F,$E7
	.dc.b	$7E,$4E,$57,$D9,$8F,$C6,$FC,$6F,$D7,$24,$78,$B9,$FF,$1D,$EE,$78
	.dc.b	$B7,$F8,$13,$F0,$CF,$D8,$6D,$F5,$D6,$5F,$4F,$E4,$FF,$59,$F2,$52
	.dc.b	$7F,$11,$1F,$AF,$F9,$3F,$D5,$7C,$9C,$9F,$83,$7F,$6E,$7C,$1C,$5F
	.dc.b	$82,$CB,$C3,$5F,$0F,$83,$F3,$8F,$C0,$C3,$F7,$DE,$5C,$8F,$E8,$11
	.dc.b	$FB,$D8,$FD,$C7,$83,$27,$BF,$6F,$F8,$5D,$FF,$18,$5F,$4E,$4F,$9B
	.dc.b	$1F,$43,$1F,$C8,$3E,$1D,$BF,$86,$11,$FE,$9D,$A7,$F8,$57,$C2,$FF
	.dc.b	$00,$92,$3F,$C7,$97,$FC,$49,$FD,$21,$F6,$38,$3E,$53,$DF,$6F,$80
	.dc.b	$7E,$34,$FF,$1E,$F2,$FF,$1B,$E1,$93,$FD,$76,$3F,$8D,$61,$FC,$54
	.dc.b	$91,$EE,$F1,$7E,$5B,$FA,$AB,$EA,$FE,$27,$5F,$81,$E2,$FF,$23,$24
	.dc.b	$7F,$80,$1F,$E4,$DE,$57,$F2,$12,$47,$D8,$FD,$93,$F8,$B3,$EB,$ED
	.dc.b	$FC,$33,$F9,$7E,$DF,$D3,$BD,$F6,$F7,$EF,$FE,$50,$8F,$EE,$18,$FF
	.dc.b	$8F,$4F,$F1,$F6,$3F,$E4,$CC,$7F,$1B,$3D,$FC,$5F,$F1,$94,$7F,$0B
	.dc.b	$F0,$7F,$45,$D2,$8D,$D7,$CE,$D7,$EE,$69,$E2,$8F,$9C,$78,$F5,$FE
	.dc.b	$99,$B7,$F9,$93,$D2,$FF,$3A,$69,$E2,$AF,$F1,$BC,$E7,$3E,$F9,$3C
	.dc.b	$5F,$19,$2B,$E2,$78,$9F,$41,$EE,$4D,$3F,$33,$87,$9B,$2F,$F3,$06
	.dc.b	$1F,$C4,$6B,$FC,$7A,$D7,$FC,$4B,$3B,$C7,$C2,$FC,$22,$47,$D3,$F0
	.dc.b	$7C,$1C,$3F,$9A,$3F,$37,$8F,$83,$83,$FC,$F3,$7C,$EF,$F8,$53,$5F
	.dc.b	$F1,$2C,$7D,$1F,$03,$E7,$92,$3F,$01,$DF,$DB,$87,$E6,$63,$DB,$93
	.dc.b	$F3,$79,$77,$BF,$9B,$8F,$D1,$8F,$E2,$DA,$BE,$07,$ED,$FC,$8C,$BF
	.dc.b	$C6,$47,$57,$F1,$7D,$BF,$B2,$26,$DF,$EC,$43,$FA,$7B,$FE,$18,$69
	.dc.b	$FA,$98,$FE,$2B,$6F,$91,$FC,$4F,$C8,$49,$FC,$D4,$77,$7F,$30,$7F
	.dc.b	$1E,$E7,$8F,$F3,$3B,$EA,$FC,$8D,$3D,$9A,$FF,$23,$E8,$FC,$97,$C8
	.dc.b	$7E,$5B,$E4,$25,$FE,$9E,$3B,$6B,$FC,$70,$F5,$C7,$E4,$CE,$D8,$7E
	.dc.b	$5F,$1E,$F1,$FA,$D3,$FB,$26,$1F,$9F,$F9,$0F,$D1,$FC,$84,$9F,$DA
	.dc.b	$47,$6A,$FD,$FC,$71,$C3,$BB,$F7,$8F,$AA,$7B,$9F,$EB,$7B,$BE,$43
	.dc.b	$F5,$E4,$9F,$E0,$1F,$21,$D5,$FE,$18,$7A,$6B,$AD,$FA,$0F,$4F,$27
	.dc.b	$F8,$4F,$CA,$FF,$0D,$72,$FF,$81,$F9,$66,$1F,$D4,$7C,$E4,$F5,$8F
	.dc.b	$D4,$71,$FE,$30,$FE,$4F,$8F,$5E,$1F,$A3,$C9,$FE,$47,$CB,$FC,$4D
	.dc.b	$CD,$FE,$13,$E6,$98,$F5,$8E,$9C,$FF,$97,$BF,$81,$8F,$E2,$F1,$FA
	.dc.b	$07,$4B,$FF,$09,$63,$FE,$31,$E7,$FF,$0F,$73,$C9,$F4,$47,$4E,$8E
	.dc.b	$1C,$5F,$0B,$0E,$75,$D8,$FB,$C7,$F9,$D3,$6F,$F3,$5B,$73,$FF,$3F
	.dc.b	$3F,$7C,$C1,$FE,$81,$E2,$FF,$3D,$38,$21,$FE,$3A,$29,$D8,$C1,$FE
	.dc.b	$07,$3F,$C0,$FE,$67,$FB,$C1,$E8,$D7,$FD,$0E,$D7,$84,$23,$11,$FD
	.dc.b	$25,$AF,$4F,$43,$D3,$3B,$6C,$EA,$31,$3E,$09,$C9,$1F,$37,$E4,$82
	.dc.b	$00

;--------------------------------------------------------------------------------
;	FMOVE.S FPn,Dr
;	FMOVE.S FPn,<mem>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovesregto_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.S FPN,<EA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.S FPn,<ea>',13,10
;------------------------------------------------
;	d1	actual result
;	d4	actual status
;	d6	fpcr=(XRN..DRP)<<4
;	d7	0=failed,1=successful
;	a0	expected result,expected status,...
;	a1	expected result
;	a4	expected status
;	a5	source handle,...
;	fp5	source
;------------------------------------------------
	lea.l	push_decompressed,a3
;decompress data
	move.l	a3,-(sp)
	pea.l	fmovesregto_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a3,a5			;source handle,...
@@:
	add.l	d0,(a3)+		;source handle
	tst.l	(a3)
	bpl	@b
	addq.l	#4,a3			;-1
	movea.l	a3,a0			;expected result,expected status,...
;
55:
	move.l	#0,d6		;fpcr=(XRN..DRP)<<4
66:
;FPn,Dr
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	#-1,d1			;fill=-1
					;source
	fmove.s	fp5,d1			;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.S FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',Dr @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
;FPn,<mem>
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	#-1,-(sp)		;fill=-1
					;source
	fmove.s	fp5,(sp)		;EXECUTE
	move.l	(sp)+,d1		;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.S FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',<mem> @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected
;
	add.w	#1<<4,d6		;rprm++
	cmp.w	#176,d6		;fpcr=(XRN..DRP)<<4
	bls	66b
;
	addq.l	#4,a5			;source handle
	tst.l	(a5)			;source handle
	bpl	55b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmovesregto_data_compressed::
	.dc.b	$00,$02,$6C,$B8,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$13,$9F,$EE,$EB,$FC,$7A,$11,$88,$FE,$6B,$18,$70,$D7,$8B
	.dc.b	$5F,$C3,$B5,$E9,$1C,$DF,$13,$63,$51,$DA,$30,$47,$BB,$E6,$FA,$DF
	.dc.b	$F7,$94,$BF,$22,$F5,$D7,$F4,$A9,$CD,$6F,$D1,$F8,$15,$AA,$56,$9F
	.dc.b	$CE,$BD,$23,$4E,$63,$FA,$0B,$9B,$F9,$51,$DE,$BD,$6F,$27,$BC,$6A
	.dc.b	$3B,$C6,$9E,$6E,$A2,$9F,$1E,$7B,$38,$BF,$53,$EF,$3A,$47,$96,$BC
	.dc.b	$51,$93,$83,$E3,$70,$EA,$79,$DF,$BA,$FC,$93,$CA,$31,$5F,$67,$E1
	.dc.b	$BC,$73,$FD,$D9,$F0,$3F,$D8,$6B,$D4,$E5,$E6,$E6,$7C,$2D,$BF,$C7
	.dc.b	$D7,$48,$C1,$FC,$39,$C1,$D8,$B5,$D3,$0E,$07,$88,$AE,$4F,$1C,$FD
	.dc.b	$0B,$F9,$91,$89,$E0,$D3,$84,$79,$A3,$2D,$8C,$E1,$F1,$30,$F9,$33
	.dc.b	$F4,$A7,$2D,$3B,$9C,$4B,$3C,$6F,$EA,$35,$FC,$0B,$5F,$A5,$1C,$EB
	.dc.b	$BB,$F7,$23,$6E,$5E,$6D,$72,$ED,$E6,$D7,$E2,$4F,$76,$C7,$17,$67
	.dc.b	$E1,$CE,$AF,$1B,$F9,$EF,$B0,$E1,$1C,$9D,$5F,$CD,$74,$60,$F8,$97
	.dc.b	$EC,$9E,$C7,$A0,$A7,$88,$FB,$4E,$DA,$79,$0D,$9E,$12,$D7,$9A,$38
	.dc.b	$3F,$72,$FE,$03,$99,$E4,$74,$3C,$45,$3C,$4F,$1A,$F7,$BC,$63,$6E
	.dc.b	$63,$DD,$3D,$AF,$E1,$5F,$89,$FC,$3C,$E8,$E6,$EA,$FC,$77,$CD,$F5
	.dc.b	$C6,$5E,$EF,$27,$33,$F3,$1E,$31,$97,$D4,$7C,$52,$CF,$A5,$F6,$C6
	.dc.b	$E3,$C1,$1A,$8F,$03,$A9,$FB,$6F,$82,$FD,$D1,$E1,$8F,$2B,$A7,$93
	.dc.b	$97,$95,$FC,$38,$D5,$F8,$09,$7E,$32,$46,$9B,$3C,$49,$5F,$02,$74
	.dc.b	$48,$CC,$79,$1A,$F8,$8E,$0E,$1F,$8D,$5E,$29,$E5,$1A,$21,$CC,$A7
	.dc.b	$33,$04,$30,$43,$AD,$7A,$9E,$26,$23,$BB,$C4,$C5,$79,$6B,$EC,$46
	.dc.b	$0E,$2E,$09,$5F,$62,$34,$73,$8C,$D7,$D3,$8C,$9E,$58,$CD,$FD,$C2
	.dc.b	$4F,$E5,$8E,$11,$A7,$11,$A7,$F3,$84,$9F,$C3,$3A,$21,$97,$F0,$B1
	.dc.b	$C5,$DC,$66,$7E,$95,$FC,$1B,$F6,$47,$86,$75,$1F,$2E,$BD,$CE,$89
	.dc.b	$1E,$86,$DF,$79,$F0,$CF,$91,$F0,$90,$CB,$D8,$FA,$AF,$C7,$9E,$71
	.dc.b	$C6,$36,$71,$AF,$4B,$F5,$A7,$73,$CE,$FD,$8F,$94,$97,$E5,$BE,$37
	.dc.b	$EC,$9C,$9D,$E3,$31,$E3,$8C,$9D,$5F,$03,$87,$C9,$7F,$1D,$E1,$5D
	.dc.b	$E3,$11,$F8,$42,$57,$CF,$78,$D7,$C2,$73,$3E,$C7,$37,$EC,$BE,$95
	.dc.b	$BB,$E5,$5F,$18,$8F,$A9,$E7,$1E,$77,$37,$E4,$9E,$26,$CA,$E4,$F1
	.dc.b	$47,$09,$F0,$46,$27,$83,$4E,$11,$DE,$32,$D8,$CD,$7A,$AF,$D9,$3F
	.dc.b	$46,$72,$D3,$B9,$C4,$B3,$C6,$FE,$9B,$5F,$B8,$D7,$E9,$C7,$3A,$EE
	.dc.b	$FD,$B8,$DB,$97,$9B,$5C,$BB,$7F,$C6,$8F,$F8,$CB,$0F,$F1,$65,$FF
	.dc.b	$89,$9B,$19,$8F,$F1,$CC,$7F,$8C,$23,$FC,$85,$1B,$32,$57,$DA,$78
	.dc.b	$67,$D2,$F8,$5F,$B7,$7E,$D6,$C6,$DA,$EF,$0F,$B6,$F4,$31,$5F,$5D
	.dc.b	$C4,$FE,$38,$F8,$2E,$8F,$A1,$87,$C8,$8D,$1F,$10,$A7,$98,$A7,$88
	.dc.b	$B5,$E2,$8E,$51,$F8,$B6,$C7,$E1,$AF,$D8,$EE,$7C,$67,$2B,$F4,$57
	.dc.b	$07,$F0,$F7,$E7,$BE,$31,$AA,$E2,$77,$3F,$AE,$47,$FA,$67,$EE,$9F
	.dc.b	$1F,$FE,$5E,$6D,$FF,$6F,$6C,$EE,$74,$7F,$80,$E3,$33,$FE,$95,$6A
	.dc.b	$A9,$1D,$30,$FF,$18,$3A,$8C,$D7,$4A,$F0,$36,$FC,$11,$83,$FD,$03
	.dc.b	$5C,$68,$F5,$7F,$C6,$2E,$5F,$F6,$7C,$FF,$AD,$DB,$3F,$EA,$B6,$BF
	.dc.b	$ED,$08,$D4,$7F,$A0,$0C,$9F,$ED,$27,$31,$FE,$92,$8C,$9F,$ED,$F8
	.dc.b	$C9,$21,$FC,$03,$D5,$B3,$CD,$AF,$16,$BE,$68,$E0,$47,$83,$5E,$0D
	.dc.b	$78,$1F,$EE,$83,$FD,$C8,$43,$FD,$DE,$53,$FD,$DE,$53,$81,$A2,$1A
	.dc.b	$29,$F7,$4F,$BC,$A3,$E4,$B5,$F9,$D1,$F0,$9C,$90,$CC,$6D,$E9,$1D
	.dc.b	$1D,$18,$79,$3C,$0C,$BC,$23,$93,$63,$C9,$1C,$4F,$21,$5C,$C6,$DF
	.dc.b	$A2,$F6,$8D,$46,$1E,$31,$B7,$24,$70,$43,$E3,$7B,$6F,$CE,$5F,$AB
	.dc.b	$BF,$47,$F9,$59,$7E,$CA,$B5,$2F,$D1,$F3,$A2,$73,$78,$77,$37,$0E
	.dc.b	$B3,$6F,$96,$3F,$AE,$BE,$2D,$AF,$1D,$EE,$F8,$36,$75,$7C,$9F,$D7
	.dc.b	$E1,$E3,$BE,$F1,$AA,$E0,$EB,$0F,$2B,$CB,$0E,$AD,$38,$14,$E1,$5E
	.dc.b	$47,$18,$FC,$58,$C4,$78,$BF,$6B,$6D,$F8,$5A,$FE,$37,$18,$9F,$C5
	.dc.b	$C6,$27,$F8,$A7,$E0,$3F,$32,$B5,$56,$79,$66,$9C,$CF,$F1,$67,$A5
	.dc.b	$6E,$3A,$CE,$EB,$8D,$70,$9F,$34,$78,$1E,$4E,$8F,$CE,$3F,$C9,$E3
	.dc.b	$83,$63,$2F,$53,$71,$A7,$07,$E9,$5C,$12,$38,$B6,$38,$C6,$0F,$3E
	.dc.b	$FF,$CF,$67,$D1,$87,$CA,$8C,$4E,$B7,$FE,$7D,$87,$83,$0F,$89,$5A
	.dc.b	$7E,$1E,$1F,$CD,$A7,$51,$F5,$E7,$4F,$D1,$6B,$CA,$BA,$D7,$8A,$BE
	.dc.b	$95,$6A,$B9,$56,$AF,$C8,$FD,$38,$E8,$F2,$7E,$99,$06,$3F,$81,$70
	.dc.b	$53,$D2,$7D,$73,$F1,$C6,$8F,$C1,$1C,$DE,$0C,$32,$72,$3B,$98,$34
	.dc.b	$74,$34,$E3,$1F,$CC,$C7,$73,$51,$9C,$3F,$29,$1F,$70,$87,$E1,$0A
	.dc.b	$7E,$0C,$A7,$F2,$CC,$7F,$16,$F2,$38,$10,$E0,$53,$FB,$AF,$ED,$7F
	.dc.b	$CB,$1F,$41,$FC,$51,$C5,$FA,$23,$F4,$71,$8A,$ED,$1B,$C3,$F1,$CF
	.dc.b	$EB,$A7,$57,$FA,$2B,$FF,$09,$C6,$27,$55,$C1,$EA,$D7,$A3,$F7,$DC
	.dc.b	$B5,$CB,$F2,$63,$A4,$6A,$30,$73,$7D,$ED,$7F,$8F,$C7,$07,$99,$F6
	.dc.b	$CD,$C7,$E9,$23,$F6,$84,$C7,$F7,$2D,$9F,$DF,$B5,$FE,$B8,$D7,$FB
	.dc.b	$0B,$5F,$ED,$CF,$47,$99,$0E,$65,$39,$94,$E6,$53,$99,$4F,$F1,$B3
	.dc.b	$F8,$4F,$DA,$FF,$86,$12,$7F,$88,$B6,$7F,$8B,$B5,$FE,$E8,$D7,$FB
	.dc.b	$CB,$5F,$F0,$33,$D1,$E6,$43,$99,$4E,$65,$39,$94,$E6,$53,$FC,$9C
	.dc.b	$FE,$73,$E1,$7F,$92,$12,$7F,$94,$B6,$7F,$97,$B5,$FF,$06,$35,$FF
	.dc.b	$08,$B5,$FF,$0F,$3D,$1E,$64,$39,$94,$E6,$53,$99,$4E,$65,$3F,$CC
	.dc.b	$CF,$ED,$3F,$6B,$FE,$78,$49,$FE,$82,$D9,$FE,$8E,$D7,$FC,$48,$D7
	.dc.b	$FC,$52,$D7,$FC,$6C,$F4,$79,$90,$E6,$53,$99,$4E,$65,$39,$94,$FF
	.dc.b	$3F,$3F,$C3,$3F,$6B,$FE,$A8,$49,$FE,$B2,$D9,$FE,$BE,$D7,$FC,$78
	.dc.b	$D7,$FC,$82,$D7,$FC,$9C,$F4,$7F,$BA,$BF,$C1,$F4,$3F,$E5,$3D,$7F
	.dc.b	$C4,$EE,$49,$3C,$0F,$E7,$9A,$7F,$A8,$23,$31,$FD,$E9,$B3,$CA,$38
	.dc.b	$19,$2B,$97,$04,$30,$E5,$87,$FA,$7A,$3A,$B4,$EA,$68,$FF,$16,$BE
	.dc.b	$B7,$47,$52,$9D,$8A,$43,$89,$E6,$31,$3E,$F7,$B9,$D2,$7C,$11,$98
	.dc.b	$EC,$72,$3F,$AC,$94,$DD,$7D,$07,$B1,$93,$BB,$97,$E2,$CF,$0A,$F9
	.dc.b	$F1,$E9,$38,$12,$BC,$4E,$5A,$F4,$8F,$92,$EE,$BB,$3D,$4F,$BA,$E1
	.dc.b	$E0,$79,$A7,$9B,$C6,$79,$C6,$09,$3C,$67,$E1,$57,$57,$E2,$CF,$C7
	.dc.b	$8C,$46,$5A,$F8,$0F,$93,$1D,$2B,$E0,$C7,$D1,$34,$71,$29,$C6,$79
	.dc.b	$BE,$73,$DB,$1C,$9C,$1D,$AB,$D1,$5D,$E3,$E0,$46,$09,$3F,$01,$B1
	.dc.b	$F8,$43,$DE,$78,$8E,$B5,$E2,$AE,$91,$DE,$BC,$51,$89,$D5,$74,$AD
	.dc.b	$1D,$67,$4F,$56,$BE,$A3,$E1,$14,$F8,$47,$8C,$1F,$54,$B5,$FC,$45
	.dc.b	$FE,$1E,$43,$F8,$79,$4F,$E1,$E5,$3F,$87,$94,$FE,$1E,$7F,$A0,$BF
	.dc.b	$6B,$FF,$0D,$47,$F0,$F6,$C6,$75,$FF,$6F,$3F,$A0,$8C,$BF,$E8,$46
	.dc.b	$BF,$EB,$A6,$BE,$AA,$5A,$73,$3F,$EA,$E6,$CE,$9B,$3B,$AF,$0D,$7E
	.dc.b	$8E,$72,$48,$CD,$7E,$72,$31,$1F,$E7,$AB,$F6,$3A,$BF,$C3,$4E,$A3
	.dc.b	$F7,$ED,$BF,$74,$6E,$B8,$BC,$E3,$E2,$35,$F8,$4E,$DC,$BD,$0C,$3A
	.dc.b	$78,$3F,$1D,$CB,$5C,$BF,$26,$BE,$14,$62,$3E,$43,$C1,$D5,$7A,$5F
	.dc.b	$8D,$19,$76,$F1,$BF,$9C,$E5,$AE,$6B,$BB,$E9,$31,$5D,$67,$E7,$47
	.dc.b	$19,$DD,$7C,$88,$D9,$9A,$E1,$3F,$42,$30,$43,$64,$3C,$04,$9E,$8D
	.dc.b	$7D,$AE,$DC,$90,$CB,$BA,$F3,$BB,$8C,$9B,$8C,$C6,$E7,$CC,$EE,$33
	.dc.b	$3B,$9C,$CE,$EF,$F0,$AD,$9F,$C0,$9B,$BF,$F5,$F1,$31,$FE,$72,$D8
	.dc.b	$FD,$54,$7E,$CC,$C9,$5C,$93,$1F,$D7,$46,$62,$99,$3F,$06,$E5,$A6
	.dc.b	$4F,$D8,$39,$69,$9C,$7F,$65,$19,$6C,$67,$1F,$DD,$4E,$5B,$19,$BF
	.dc.b	$63,$E4,$3C,$B7,$5F,$19,$4F,$0B,$6F,$BC,$5A,$EB,$36,$BA,$CF,$B2
	.dc.b	$31,$7E,$39,$3F,$2A,$7E,$49,$A3,$FA,$54,$72,$7E,$09,$E6,$3C,$A4
	.dc.b	$32,$77,$31,$7E,$66,$DF,$99,$B3,$D6,$7E,$4D,$F8,$67,$C4,$D7,$77
	.dc.b	$E7,$70,$7C,$43,$27,$A8,$F1,$1E,$42,$4F,$8D,$E0,$62,$FE,$AB,$E0
	.dc.b	$39,$94,$D5,$F9,$E3,$D5,$18,$AF,$98,$79,$5E,$27,$8A,$BC,$04,$9E
	.dc.b	$F1,$96,$BC,$CF,$C3,$4F,$46,$9E,$D3,$57,$F0,$AB,$A5,$FD,$C9,$FB
	.dc.b	$35,$C5,$FB,$07,$53,$D9,$3C,$23,$B5,$73,$BF,$1C,$FD,$4A,$F5,$9E
	.dc.b	$02,$BC,$5E,$E7,$94,$FE,$EB,$7E,$37,$FB,$B9,$1C,$38,$3E,$09,$C2
	.dc.b	$F0,$ED,$AE,$4B,$1C,$9D,$18,$8D,$BD,$09,$1F,$C3,$B7,$FE,$BC,$DC
	.dc.b	$47,$EA,$63,$35,$FE,$01,$6D,$7F,$7B,$6B,$FD,$ED,$AF,$F4,$E6,$BF
	.dc.b	$DE,$5B,$7F,$DE,$5B,$87,$F7,$F6,$E0,$69,$30,$F2,$DF,$EA,$23,$11
	.dc.b	$E7,$33,$1F,$96,$8C,$BB,$8C,$DF,$E1,$A7,$33,$B9,$CD,$FE,$00,$F2
	.dc.b	$38,$BF,$8C,$F7,$24,$F5,$6B,$CD,$B7,$DD,$E5,$87,$CB,$6C,$FC,$B6
	.dc.b	$BF,$2D,$AF,$CB,$6B,$F2,$DB,$3F,$2D,$B7,$E8,$7C,$2F,$AE,$FE,$64
	.dc.b	$69,$AE,$9F,$6D,$7A,$C8,$7A,$CB,$5E,$B9,$E1,$7E,$B6,$CF,$AD,$AF
	.dc.b	$AD,$AF,$AD,$AF,$AD,$B3,$EB,$6D,$FA,$E7,$C2,$FA,$F0,$FB,$35,$E3
	.dc.b	$9C,$E1,$EA,$7D,$04,$8D,$36,$B9,$5F,$C3,$BE,$71,$88,$F0,$DF,$C6
	.dc.b	$AF,$F1,$4F,$9C,$1C,$BF,$E5,$B7,$0D,$5F,$AF,$58,$93,$CA,$78,$B6
	.dc.b	$72,$FF,$97,$E3,$9B,$B9,$E9,$3A,$CB,$FC,$C5,$5A,$AE,$0F,$46,$CE
	.dc.b	$F1,$FF,$36,$4E,$EB,$8B,$B6,$CE,$F0,$F7,$4E,$6B,$8B,$63,$33,$BB
	.dc.b	$FE,$1F,$1F,$C4,$1F,$F3,$D1,$82,$1F,$C6,$0A,$7F,$1E,$78,$11,$C9
	.dc.b	$FC,$95,$E9,$EC,$7F,$1F,$8F,$F5,$0B,$53,$49,$A7,$F2,$26,$C7,$F2
	.dc.b	$A6,$BF,$CB,$9A,$FF,$33,$79,$3C,$48,$71,$29,$C4,$A7,$12,$9F,$D0
	.dc.b	$1F,$4B,$83,$FD,$73,$FB,$5F,$A4,$93,$FC,$E1,$B1,$FD,$01,$AF,$F4
	.dc.b	$46,$BF,$D2,$5A,$FF,$4E,$7A,$3C,$C8,$73,$29,$CC,$A7,$32,$9C,$CA
	.dc.b	$7A,$0F,$E9,$BF,$B5,$FD,$12,$4E,$09,$1F,$D6,$23,$2D,$8C,$BF,$D7
	.dc.b	$A3,$2D,$8E,$AC,$39,$14,$E0,$53,$FB,$54,$7E,$44,$8E,$4F,$EE,$2E
	.dc.b	$1F,$ED,$7F,$B5,$FE,$24,$93,$FD,$A1,$B1,$A8,$C1,$1C,$10,$FE,$EF
	.dc.b	$1C,$1A,$73,$21,$E0,$29,$A3,$04,$30,$43,$FC,$0C,$FE,$B4,$FD,$77
	.dc.b	$97,$FA,$12,$4F,$F8,$01,$B3,$FE,$08,$6B,$FE,$0C,$6B,$FE,$11,$6B
	.dc.b	$FE,$16,$7A,$3C,$C8,$73,$29,$CC,$A7,$32,$9C,$CA,$03,$FC,$1F,$F6
	.dc.b	$BF,$DD,$12,$7F,$C3,$2D,$9F,$F1,$0B,$5F,$F1,$2B,$5F,$F1,$53,$5F
	.dc.b	$F1,$7B,$D1,$E6,$43,$99,$4E,$65,$39,$94,$E6,$53,$D4,$FF,$1F,$FD
	.dc.b	$AF,$F8,$91,$27,$FC,$62,$D9,$FF,$1C,$B5,$FF,$1E,$B5,$FF,$21,$35
	.dc.b	$FF,$23,$BD,$1E,$64,$39,$94,$E6,$53,$99,$4E,$65,$3E,$93,$FC,$DF
	.dc.b	$F6,$BF,$E5,$44,$9F,$F2,$4B,$67,$FC,$A2,$D7,$FC,$AA,$D7,$FC,$B4
	.dc.b	$D7,$FC,$BE,$F4,$79,$90,$E6,$53,$99,$4E,$65,$39,$94,$FC,$53,$FD
	.dc.b	$3F,$E4,$3F,$A1,$12,$7F,$CC,$2D,$9F,$F3,$4B,$5F,$F3,$6B,$5F,$F3
	.dc.b	$93,$5F,$F3,$BB,$D1,$E6,$43,$99,$4E,$65,$39,$94,$E6,$53,$F4,$4F
	.dc.b	$F6,$7F,$DA,$FF,$CF,$5C,$4E,$24,$61,$B7,$FA,$9E,$6F,$F4,$5D,$6E
	.dc.b	$B4,$49,$FF,$44,$C6,$5E,$18,$7E,$C6,$30,$FF,$A4,$5A,$F4,$BE,$EE
	.dc.b	$5B,$79,$BE,$D3,$96,$C6,$6F,$C3,$5C,$AB,$06,$E3,$C6,$AF,$C2,$F9
	.dc.b	$67,$83,$6F,$AB,$6B,$AC,$78,$A3,$AD,$F6,$61,$B3,$C0,$1C,$BE,$1A
	.dc.b	$EE,$D9,$EE,$DA,$EE,$47,$9D,$72,$BF,$4C,$6C,$C9,$E1,$35,$7D,$A3
	.dc.b	$E8,$CE,$2F,$D5,$5E,$78,$C5,$F7,$8F,$69,$CC,$87,$73,$25,$33,$5E
	.dc.b	$43,$BB,$8B,$F1,$12,$FF,$DD,$17,$E0,$9F,$8E,$74,$AE,$55,$BB,$F8
	.dc.b	$55,$BA,$E0,$D7,$B5,$F4,$6C,$FA,$DF,$71,$83,$B9,$83,$D4,$4B,$F0
	.dc.b	$CF,$38,$C5,$F8,$DF,$9E,$62,$3E,$5B,$F4,$EF,$C8,$E6,$FF,$23,$7D
	.dc.b	$1F,$24,$6C,$F2,$1E,$C2,$9E,$79,$E7,$3C,$63,$F1,$86,$8E,$C6,$0A
	.dc.b	$73,$29,$CC,$E0,$F3,$8E,$B7,$ED,$6D,$FE,$52,$7C,$15,$CD,$AF,$33
	.dc.b	$A9,$FD,$BC,$FC,$E9,$FD,$C6,$3F,$CD,$04,$BD,$44,$8C,$38,$38,$45
	.dc.b	$69,$63,$6E,$DA,$FE,$E7,$99,$CD,$FE,$EE,$9F,$E2,$F1,$88,$FD,$13
	.dc.b	$67,$F1,$2D,$7E,$E3,$6F,$F0,$6D,$C7,$F3,$4D,$BE,$B7,$DF,$0F,$E0
	.dc.b	$F3,$96,$C6,$70,$FE,$35,$39,$6C,$65,$FE,$41,$19,$6C,$65,$F7,$35
	.dc.b	$FE,$15,$1B,$73,$1E,$E3,$58,$7F,$39,$9D,$4F,$09,$CE,$1F,$D0,$A7
	.dc.b	$35,$B9,$CE,$1F,$ED,$5B,$F8,$B1,$8A,$DD,$F1,$8C,$10,$F1,$12,$7E
	.dc.b	$7B,$96,$B9,$7A,$B6,$FE,$8B,$B8,$C9,$C0,$86,$09,$1C,$23,$94,$FD
	.dc.b	$87,$94,$66,$78,$B5,$E3,$3B,$9E,$37,$F7,$9B,$87,$DC,$6D,$7D,$A6
	.dc.b	$BF,$61,$AF,$D4,$3B,$D7,$56,$D7,$56,$D7,$56,$9D,$4A,$75,$29,$D4
	.dc.b	$F4,$C7,$36,$C7,$37,$FA,$81,$FD,$70,$87,$F0,$9B,$FE,$FC,$E0,$FE
	.dc.b	$E4,$6B,$1A,$FF,$73,$34,$7F,$69,$3F,$8F,$DF,$F8,$31,$C9,$E0,$38
	.dc.b	$1D,$DD,$38,$3A,$9D,$A2,$BD,$4D,$15,$CB,$D4,$B3,$EA,$7B,$5F,$F0
	.dc.b	$B8,$C4,$6A,$BC,$35,$6B,$51,$E4,$9F,$11,$E1,$34,$7C,$33,$04,$3C
	.dc.b	$47,$94,$C5,$78,$1B,$1E,$06,$CF,$80,$8F,$CA,$6B,$F0,$0F,$95,$5E
	.dc.b	$27,$15,$EF,$8C,$46,$5A,$F6,$3A,$1F,$52,$3C,$8F,$43,$47,$4A,$F3
	.dc.b	$3E,$C2,$47,$03,$C0,$F0,$3C,$05,$39,$CF,$89,$E3,$3D,$27,$93,$F3
	.dc.b	$E3,$67,$13,$75,$FD,$36,$B9,$57,$08,$C4,$EE,$30,$FB,$8F,$54,$7C
	.dc.b	$A9,$E3,$5D,$E7,$E0,$1D,$8F,$88,$F9,$4D,$57,$48,$FC,$CC,$F2,$AE
	.dc.b	$91,$E1,$AD,$9C,$63,$EC,$C7,$C4,$3B,$1E,$43,$CC,$73,$29,$CD,$F6
	.dc.b	$BC,$0E,$55,$E3,$6C,$F8,$27,$FC,$FD,$1F,$DA,$4F,$71,$FD,$CC,$C3
	.dc.b	$A7,$FB,$61,$82,$18,$21,$FD,$BC,$C9,$FE,$79,$FD,$AF,$FB,$66,$3F
	.dc.b	$AB,$6C,$7E,$79,$AF,$E0,$1B,$7F,$80,$6D,$FC,$86,$DF,$EC,$9B,$5F
	.dc.b	$E7,$E6,$BF,$E8,$76,$BE,$3B,$FE,$E8,$ED,$AE,$DA,$ED,$FF,$03,$3B
	.dc.b	$6B,$B6,$BB,$7D,$C7,$27,$FD,$15,$1E,$26,$C6,$5F,$F4,$04,$65,$B1
	.dc.b	$9B,$FF,$59,$4E,$5B,$19,$B4,$E5,$B1,$98,$C1,$1C,$10,$F4,$D7,$91
	.dc.b	$AF,$89,$B5,$E0,$6D,$76,$6D,$75,$69,$D4,$F8,$66,$08,$60,$95,$F8
	.dc.b	$37,$83,$5D,$B5,$DB,$F8,$77,$6D,$76,$D7,$75,$F8,$B7,$6D,$76,$D7
	.dc.b	$75,$F8,$F7,$6D,$76,$D7,$77,$F9,$37,$6D,$76,$D7,$75,$F8,$B6,$BF
	.dc.b	$8B,$6B,$F8,$B3,$F1,$38,$FE,$DF,$03,$49,$1F,$E6,$38,$CB,$63,$2F
	.dc.b	$F9,$A6,$32,$D8,$CB,$FE,$70,$8C,$B6,$3C,$0C,$3B,$14,$E8,$53,$91
	.dc.b	$4F,$F4,$04,$7F,$0F,$7F,$2E,$64,$18,$3F,$CF,$9F,$B5,$F8,$9F,$F4
	.dc.b	$43,$5D,$B5,$34,$93,$FC,$75,$B1,$D2,$30,$47,$07,$80,$E0,$43,$04
	.dc.b	$3F,$D2,$51,$EB,$23,$B3,$25,$3F,$D3,$6F,$C1,$F8,$3F,$41,$27,$F9
	.dc.b	$93,$67,$FD,$38,$D7,$FD,$40,$D7,$FD,$4A,$D7,$FD,$58,$F4,$79,$90
	.dc.b	$E6,$53,$99,$4E,$65,$39,$94,$FA,$AF,$D9,$FD,$AF,$FD,$C1,$9B,$34
	.dc.b	$7D,$78,$C3,$A8,$CD,$FF,$AE,$67,$2F,$F9,$0A,$78,$4E,$AF,$FC,$81
	.dc.b	$18,$39,$16,$78,$1C,$E7,$84,$FF,$B3,$6B,$AB,$FE,$CA,$6E,$99,$9D
	.dc.b	$CF,$E4,$27,$6E,$6F,$B3,$E2,$BE,$D1,$88,$D4,$78,$9E,$D5,$D1,$57
	.dc.b	$47,$B4,$7B,$6F,$8C,$70,$BE,$8E,$9F,$7D,$F8,$09,$1B,$7C,$0F,$22
	.dc.b	$1C,$5C,$B4,$E0,$49,$DB,$E6,$AF,$45,$F9,$E3,$17,$E2,$78,$9C,$23
	.dc.b	$E5,$BD,$4C,$4F,$23,$DC,$E4,$87,$4B,$F5,$DF,$0B,$F8,$33,$93,$E8
	.dc.b	$1E,$03,$8D,$F8,$1D,$1C,$6B,$9B,$AB,$F1,$57,$DA,$BD,$4F,$33,$B3
	.dc.b	$87,$2D,$79,$47,$C1,$70,$70,$33,$5F,$51,$FB,$73,$AB,$F9,$47,$9C
	.dc.b	$B7,$E4,$AF,$3D,$7A,$8E,$67,$D5,$29,$E4,$29,$AA,$ED,$3C,$E3,$E3
	.dc.b	$98,$BF,$43,$C0,$F4,$1D,$C9,$7E,$48,$DC,$62,$7A,$57,$3B,$F3,$DE
	.dc.b	$DF,$81,$5F,$8D,$9D,$3E,$93,$95,$FD,$CA,$F8,$57,$E7,$9F,$61,$AB
	.dc.b	$FF,$29,$57,$F3,$B2,$38,$24,$7F,$46,$21,$FA,$F3,$FA,$4C,$70,$72
	.dc.b	$62,$38,$B8,$8D,$3F,$BF,$D5,$A4,$8F,$C0,$C6,$23,$F2,$CD,$8C,$C6
	.dc.b	$DF,$7C,$65,$B1,$9C,$3E,$74,$E5,$B1,$9C,$3F,$2F,$39,$6C,$66,$FF
	.dc.b	$91,$4E,$5B,$19,$C3,$F8,$AC,$E5,$B1,$9C,$3F,$87,$4E,$5B,$19,$7E
	.dc.b	$13,$5F,$E2,$51,$B7,$31,$F0,$8D,$3F,$C1,$A3,$4F,$08,$CE,$1F,$D0
	.dc.b	$A7,$33,$B9,$CE,$1F,$D2,$27,$35,$B9,$CE,$09,$CD,$6E,$73,$18,$9D
	.dc.b	$46,$08,$79,$C9,$7F,$61,$AF,$8D,$F6,$B9,$3D,$4E,$1F,$B6,$DB,$F3
	.dc.b	$3C,$9B,$1C,$0E,$47,$53,$07,$22,$BC,$1C,$1B,$23,$F8,$68,$F0,$5F
	.dc.b	$C1,$7C,$0D,$9E,$6F,$E2,$23,$75,$F1,$1D,$B6,$76,$FE,$2A,$37,$7F
	.dc.b	$2C,$FC,$71,$1F,$97,$3C,$23,$51,$F9,$57,$4F,$67,$F4,$51,$FB,$46
	.dc.b	$BE,$73,$F7,$53,$E3,$6C,$6A,$31,$1C,$67,$F5,$CD,$7F,$75,$3B,$23
	.dc.b	$A8,$EC,$6C,$FC,$A7,$8B,$EE,$83,$49,$1E,$D8,$CB,$63,$FC,$08,$FF
	.dc.b	$85,$C8,$7F,$7E,$3F,$BC,$18,$CB,$6F,$F8,$68,$FF,$07,$1A,$30,$47
	.dc.b	$78,$F0,$64,$7F,$1E,$C3,$FD,$FA,$FF,$1E,$3F,$CC,$F1,$FC,$6D,$A6
	.dc.b	$CA,$7C,$E2,$9C,$43,$C5,$A7,$12,$9C,$4A,$71,$78,$11,$F6,$1E,$83
	.dc.b	$FC,$12,$7F,$8A,$0F,$8E,$66,$7F,$C8,$AE,$4F,$86,$67,$1F,$F2,$5B
	.dc.b	$99,$F9,$6E,$6B,$F9,$EB,$D8,$E2,$7C,$D8,$DB,$4F,$29,$4E,$07,$33
	.dc.b	$B1,$A3,$A3,$C9,$A6,$6B,$D2,$F8,$0C,$9D,$0E,$CE,$9E,$C7,$F8,$FC
	.dc.b	$F8,$66,$AB,$DE,$E4,$E9,$1D,$A7,$C4,$D8,$F1,$4F,$09,$CB,$63,$35
	.dc.b	$FD,$9E,$72,$D8,$C9,$E2,$3D,$04,$AE,$CD,$AF,$2B,$C8,$E1,$1E,$D7
	.dc.b	$99,$E8,$8D,$B8,$AE,$0F,$B4,$F8,$66,$0A,$72,$29,$E1,$AF,$BF,$5E
	.dc.b	$43,$C1,$19,$3E,$C9,$4E,$C7,$CD,$8E,$35,$E4,$8D,$57,$85,$F3,$1E
	.dc.b	$12,$9C,$CF,$4B,$B7,$89,$B2,$9B,$38,$D7,$98,$93,$E1,$7D,$4D,$3C
	.dc.b	$91,$FB,$53,$F6,$84,$3F,$68,$53,$F6,$85,$38,$1A,$21,$A2,$C7,$FA
	.dc.b	$DF,$F6,$BF,$F0,$E6,$3F,$88,$6C,$FE,$31,$AF,$E0,$5B,$3E,$A6,$D7
	.dc.b	$A5,$B8,$7C,$E6,$D7,$FA,$D5,$AF,$FA,$99,$AF,$FA,$CD,$AF,$FA,$99
	.dc.b	$AF,$FA,$55,$B5,$E9,$BF,$F1,$14,$F5,$6B,$D5,$AF,$56,$BD,$5A,$F5
	.dc.b	$7F,$C6,$13,$EC,$3B,$D9,$C4,$68,$E3,$5E,$37,$FC,$75,$1A,$7F,$C7
	.dc.b	$15,$E5,$72,$71,$32,$F9,$DC,$90,$FA,$86,$CF,$51,$C1,$AC,$7F,$CC
	.dc.b	$71,$F6,$11,$E9,$8F,$EE,$31,$98,$C4,$7F,$67,$8C,$11,$D9,$0D,$14
	.dc.b	$FF,$36,$BE,$FC,$7C,$53,$F2,$E7,$FB,$4B,$5F,$4C,$6A,$BE,$43,$6F
	.dc.b	$C6,$DB,$FF,$38,$36,$BF,$CE,$AD,$7A,$D7,$D1,$7A,$DF,$DD,$7E,$5C
	.dc.b	$EC,$91,$97,$6F,$94,$C1,$0F,$53,$C1,$A6,$67,$C0,$DB,$F4,$B6,$FE
	.dc.b	$23,$6F,$D2,$F9,$5F,$0D,$7C,$28,$D4,$7A,$1B,$7E,$19,$D4,$78,$8F
	.dc.b	$41,$5C,$92,$FE,$ED,$76,$7C,$B3,$FE,$A9,$BF,$9A,$FE,$22,$7C,$AE
	.dc.b	$5A,$FE,$2A,$36,$E6,$3E,$EB,$97,$6F,$63,$0E,$9C,$12,$39,$BA,$BF
	.dc.b	$CE,$FF,$00,$F9,$E7,$FD,$60,$DA,$FF,$5A,$B5,$34,$95,$FC,$27,$4F
	.dc.b	$F0,$2B,$FB,$96,$D1,$A4,$8F,$F0,$5C,$65,$B1,$97,$F3,$ED,$7F,$5E
	.dc.b	$D7,$E6,$1D,$88,$74,$29,$FA,$82,$9F,$B8,$29,$C8,$E0,$43,$FD,$B7
	.dc.b	$FB,$67,$95,$B1,$EE,$8C,$47,$FB,$91,$AF,$FB,$9D,$AF,$F0,$46,$A6
	.dc.b	$92,$3A,$12,$3F,$C5,$F1,$A7,$C0,$43,$25,$3C,$E5,$3C,$85,$3C,$85
	.dc.b	$3C,$87,$D8,$F6,$1E,$1B,$F4,$A7,$EB,$5F,$F0,$07,$66,$CB,$DD,$3F
	.dc.b	$5A,$D4,$E7,$0E,$58,$62,$F7,$56,$73,$8F,$F2,$8C,$7A,$D6,$A3,$18
	.dc.b	$73,$C7,$C7,$18,$9D,$2A,$E3,$7D,$23,$17,$C7,$0F,$06,$1C,$A7,$38
	.dc.b	$72,$C3,$8C,$E5,$B5,$C3,$1E,$2D,$BF,$16,$3F,$1E,$31,$1D,$23,$0E
	.dc.b	$AF,$A6,$3E,$38,$C5,$F4,$7D,$99,$FD,$32,$3F,$E7,$0C,$BF,$12,$48
	.dc.b	$CC,$7F,$50,$24,$65,$E1,$1A,$74,$7D,$97,$F0,$0C,$3B,$14,$D9,$F7
	.dc.b	$CC,$1C,$A3,$F7,$EE,$4E,$24,$38,$BA,$37,$FB,$67,$EF,$E7,$F0,$ED
	.dc.b	$9F,$BE,$FF,$03,$FD,$03,$33,$FD,$B0,$B3,$FB,$3D,$3F,$27,$5C,$63
	.dc.b	$EF,$BC,$5E,$73,$C5,$A7,$1C,$53,$C6,$B9,$D7,$F7,$AB,$EE,$CB,$F9
	.dc.b	$35,$DD,$FE,$FE,$D7,$9C,$77,$7E,$A4,$64,$91,$98,$E2,$D7,$89,$FC
	.dc.b	$2A,$36,$D8,$D4,$79,$EF,$7C,$9E,$5A,$F8,$35,$FE,$12,$8C,$61,$EC
	.dc.b	$94,$62,$B8,$C7,$C4,$AF,$9B,$39,$BF,$E6,$8F,$D0,$F9,$59,$5F,$2B
	.dc.b	$F2,$AA,$F8,$04,$63,$1F,$04,$DC,$7D,$B3,$70,$FE,$81,$37,$1F,$F1
	.dc.b	$4C,$FE,$42,$7F,$A2,$36,$33,$87,$D2,$6B,$F3,$9A,$FC,$C3,$C6,$6C
	.dc.b	$A6,$CA,$6E,$3B,$38,$3C,$E7,$17,$27,$E4,$BF,$2A,$F4,$41,$A4,$7F
	.dc.b	$1F,$1B,$69,$F9,$28,$FC,$DF,$B3,$23,$F2,$BF,$5B,$5F,$99,$17,$3F
	.dc.b	$B5,$37,$3F,$C3,$CD,$CB,$FB,$54,$DC,$BF,$CA,$73,$EC,$26,$1F,$0D
	.dc.b	$AF,$C3,$6B,$F2,$5A,$F6,$38,$10,$E0,$7E,$B3,$EB,$FC,$09,$A4,$7F
	.dc.b	$63,$1F,$BA,$8C,$47,$E2,$7E,$A7,$F7,$A7,$EF,$BF,$06,$B1,$FB,$88
	.dc.b	$AF,$E3,$E2,$E7,$F9,$D9,$B9,$FE,$AA,$6E,$7F,$E0,$D9,$FB,$A4,$BF
	.dc.b	$C2,$35,$FC,$23,$5F,$C5,$35,$FD,$AB,$63,$A4,$72,$3F,$86,$FE,$0F
	.dc.b	$E0,$CF,$FD,$06,$D8,$34,$8F,$B4,$E2,$7F,$16,$3F,$8F,$9F,$C8,$3B
	.dc.b	$6B,$92,$1F,$B1,$72,$FE,$CC,$B9,$7F,$01,$8B,$97,$F0,$E9,$FC,$F9
	.dc.b	$0F,$D3,$14,$FD,$31,$4F,$D6,$14,$E4,$70,$21,$C0,$FE,$77,$C2,$EB
	.dc.b	$91,$E8,$BF,$E7,$33,$99,$F2,$E5,$FC,$59,$DD,$7A,$CE,$6E,$2B,$F8
	.dc.b	$E4,$7F,$4A,$AE,$4F,$5B,$FE,$7B,$18,$8F,$E8,$AF,$8F,$E5,$78,$59
	.dc.b	$BF,$B5,$61,$95,$E0,$72,$D2,$1D,$27,$93,$E6,$9E,$2E,$67,$F8,$E5
	.dc.b	$78,$63,$FA,$5B,$5F,$E8,$ED,$7F,$A6,$B5,$F5,$D7,$67,$6D,$76,$D7
	.dc.b	$73,$E0,$8C,$63,$F3,$63,$04,$9F,$9B,$FB,$57,$A3,$2F,$AB,$68,$C6
	.dc.b	$5E,$26,$E1,$F5,$DB,$8F,$DF,$6D,$71,$C3,$EC,$EB,$F4,$0F,$A3,$CD
	.dc.b	$63,$24,$3E,$63,$97,$E7,$45,$CB,$CB,$37,$2F,$8B,$3E,$F2,$57,$CE
	.dc.b	$BF,$ED,$98,$FE,$4E,$73,$3F,$29,$AF,$CA,$6B,$D0,$E4,$F0,$C3,$E2
	.dc.b	$CF,$87,$2F,$CC,$34,$FD,$01,$6F,$FB,$8D,$FB,$6B,$8B,$70,$FB,$B5
	.dc.b	$F9,$4E,$29,$3F,$A0,$E3,$AF,$E3,$E2,$DF,$D6,$9B,$7F,$86,$9B,$8F
	.dc.b	$EA,$E6,$E1,$FB,$39,$FC,$21,$30,$F8,$CD,$7F,$1A,$D7,$F1,$AD,$7B
	.dc.b	$1C,$08,$70,$3F,$6F,$C7,$E0,$CB,$FA,$24,$7E,$EA,$7F,$63,$39,$24
	.dc.b	$64,$E2,$7F,$00,$3F,$6E,$7E,$E3,$82,$B9,$21,$FB,$37,$2F,$ED,$4B
	.dc.b	$9F,$E7,$22,$E7,$FA,$C9,$FD,$31,$0F,$C4,$14,$FD,$79,$4F,$D7,$94
	.dc.b	$E4,$70,$21,$C0,$FE,$47,$DF,$FC,$66,$37,$1F,$C5,$23,$1A,$7F,$21
	.dc.b	$8F,$E3,$91,$8C,$FF,$91,$C6,$F3,$F4,$D7,$F8,$F2,$B5,$1F,$E7,$EC
	.dc.b	$7F,$96,$C7,$4A,$63,$9C,$7F,$CC,$31,$83,$FC,$E6,$6A,$29,$BC,$11
	.dc.b	$B3,$FC,$EE,$53,$FC,$E2,$71,$33,$1C,$5D,$5F,$F9,$E1,$B5,$E2,$9E
	.dc.b	$93,$97,$81,$FE,$92,$9D,$1F,$E4,$53,$C0,$7F,$9C,$8A,$68,$FF,$3B
	.dc.b	$98,$AF,$4C,$73,$6C,$73,$8C,$3B,$23,$83,$D4,$FC,$B6,$E1,$F1,$A3
	.dc.b	$E4,$99,$2B,$97,$E3,$99,$2B,$C3,$82,$46,$B8,$6B,$F3,$0F,$E8,$18
	.dc.b	$7C,$96,$E3,$F6,$94,$F0,$9D,$55,$AD,$4D,$C7,$FD,$91,$3F,$2E,$FF
	.dc.b	$A2,$9F,$79,$C1,$1F,$96,$68,$8E,$5F,$F4,$F9,$C4,$DB,$97,$EB,$1E
	.dc.b	$88,$FF,$6C,$70,$F8,$E3,$F1,$AD,$7F,$A4,$1F,$06,$7E,$6C,$64,$91
	.dc.b	$F3,$38,$3E,$89,$F4,$B8,$6C,$64,$87,$CB,$72,$FC,$D8,$B9,$7C,$D9
	.dc.b	$B9,$7C,$D9,$F8,$C4,$AF,$86,$D7,$E0,$9F,$EB,$93,$FD,$6E,$43,$04
	.dc.b	$78,$3F,$11,$FC,$4E,$DE,$0C,$BF,$DF,$8D,$9F,$F5,$E1,$EA,$C3,$F0
	.dc.b	$51,$98,$FE,$9A,$7A,$CF,$C7,$BF,$94,$E0,$91,$F9,$2E,$1A,$FE,$36
	.dc.b	$2D,$7E,$36,$6E,$3F,$8D,$9B,$87,$F0,$2B,$56,$1F,$F4,$79,$F8,$69
	.dc.b	$FC,$61,$1C,$10,$F8,$EE,$5A,$64,$FC,$3B,$96,$9F,$AA,$E1,$F0,$3F
	.dc.b	$D3,$0F,$51,$93,$F5,$F1,$FB,$48,$D9,$23,$F6,$67,$EE,$0F,$DC,$F0
	.dc.b	$57,$24,$3F,$64,$E5,$FD,$A1,$71,$FD,$A4,$5C,$7F,$69,$3F,$AD,$21
	.dc.b	$FA,$A2,$9F,$88,$29,$FA,$62,$9C,$8E,$04,$38,$1F,$CD,$25,$0F,$3F
	.dc.b	$FC,$D9,$C3,$E3,$6D,$7A,$EB,$71,$88,$F5,$9C,$EB,$95,$7F,$9A,$6B
	.dc.b	$F8,$84,$FE,$05,$E3,$19,$9F,$34,$A7,$33,$FD,$7E,$BF,$CF,$DA,$F1
	.dc.b	$BE,$B2,$8C,$4F,$FA,$0F,$6B,$3F,$E3,$7C,$3C,$93,$C9,$B1,$97,$C1
	.dc.b	$7D,$2B,$8C,$F5,$9D,$B5,$DD,$73,$6C,$7F,$A5,$E7,$8D,$C9,$FD,$01
	.dc.b	$E8,$6B,$4C,$D7,$37,$FB,$3B,$5D,$1F,$A9,$70,$4A,$F5,$47,$06,$BC
	.dc.b	$0C,$3E,$96,$CF,$BB,$6F,$34,$7A,$A7,$F8,$1E,$3F,$4A,$B5,$3F,$09
	.dc.b	$AE,$DB,$1B,$8F,$8F,$5F,$29,$87,$CA,$CA,$BD,$CB,$3F,$80,$8B,$97
	.dc.b	$DE,$9B,$8F,$FA,$FE,$6E,$3F,$E1,$C9,$F6,$9F,$32,$7F,$D7,$0F,$C7
	.dc.b	$8D,$3C,$CF,$9E,$E2,$37,$1C,$8C,$C7,$F8,$26,$33,$1C,$4E,$91,$93
	.dc.b	$FC,$5B,$7F,$9C,$E2,$FC,$C4,$63,$0F,$D1,$C6,$23,$FC,$55,$7F,$ED
	.dc.b	$D8,$C5,$FE,$61,$AF,$E5,$9A,$FE,$39,$EA,$7E,$4E,$FC,$19,$71,$CF
	.dc.b	$93,$2B,$93,$96,$90,$E8,$E5,$A6,$4F,$C6,$57,$E2,$5F,$5C,$7C,$FC
	.dc.b	$13,$9C,$3D,$93,$98,$C4,$EA,$30,$4A,$FA,$73,$A2,$46,$63,$E0,$67
	.dc.b	$EB,$69,$E9,$C7,$E8,$46,$5B,$19,$CB,$E8,$CE,$5B,$1F,$34,$FA,$07
	.dc.b	$D1,$CE,$B9,$21,$F4,$5C,$BF,$4C,$B5,$F8,$18,$B3,$F8,$09,$F8,$44
	.dc.b	$3D,$E5,$3F,$2D,$87,$E2,$C8,$E4,$FA,$25,$3A,$1C,$8F,$E3,$7E,$07
	.dc.b	$E1,$1F,$E4,$2D,$73,$97,$93,$32,$9D,$4F,$F1,$A6,$E3,$FC,$D9,$FE
	.dc.b	$65,$9F,$F9,$AE,$7F,$A3,$36,$FF,$09,$1F,$C6,$23,$04,$7F,$54,$D7
	.dc.b	$F6,$CD,$7F,$90,$B5,$FE,$3C,$75,$21,$FB,$32,$9F,$BF,$29,$FC,$A8
	.dc.b	$A7,$23,$81,$0F,$E6,$3E,$8F,$A1,$34,$91,$EF,$6C,$7F,$A2,$DA,$F2
	.dc.b	$8D,$11,$D3,$C5,$87,$12,$9E,$22,$9F,$D1,$A3,$E7,$3E,$43,$27,$F8
	.dc.b	$A5,$C9,$1F,$A9,$F4,$DE,$B2,$4E,$09,$1F,$EA,$38,$CB,$63,$2F,$F5
	.dc.b	$48,$CB,$63,$AB,$0E,$45,$38,$14,$FE,$BF,$1E,$A2,$39,$3F,$C8,$4E
	.dc.b	$1F,$F5,$B7,$ED,$9F,$DE,$E7,$FD,$9C,$D8,$FF,$00,$39,$D6,$CF,$F5
	.dc.b	$BC,$BF,$C3,$33,$BA,$FF,$09,$35,$FE,$FA,$F1,$73,$3C,$75,$4E,$EB
	.dc.b	$9C,$E6,$38,$57,$28,$C1,$FA,$08,$ED,$1F,$ED,$E6,$BD,$23,$44,$3D
	.dc.b	$E5,$7F,$C0,$71,$C0,$91,$98,$FF,$04,$C7,$96,$BA,$6D,$FB,$A9,$F5
	.dc.b	$D7,$C8,$11,$98,$F4,$46,$63,$17,$EE,$8C,$6B,$F3,$1B,$5D,$0F,$9F
	.dc.b	$5C,$2F,$D9,$D7,$DC,$99,$7D,$47,$15,$4C,$45,$7E,$8C,$69,$B1,$97
	.dc.b	$E7,$47,$C6,$23,$F1,$AD,$18,$8F,$8B,$18,$75,$19,$7F,$D2,$19,$7F
	.dc.b	$91,$E3,$11,$F5,$E7,$83,$CA,$38,$65,$F8,$2E,$BF,$81,$7C,$6B,$F2
	.dc.b	$31,$8B,$E3,$1F,$85,$8C,$5F,$D3,$6C,$70,$CF,$DD,$87,$57,$F3,$7C
	.dc.b	$7F,$64,$99,$62,$A1,$F9,$88,$CB,$4C,$9F,$95,$72,$D3,$27,$BD,$FC
	.dc.b	$89,$1F,$98,$D7,$E5,$B5,$F9,$38,$A3,$11,$CC,$E2,$43,$F5,$BC,$7F
	.dc.b	$16,$3F,$28,$D8,$FD,$CB,$5D,$DF,$AB,$0F,$D8,$CE,$67,$F6,$0F,$CF
	.dc.b	$EA,$FC,$F1,$31,$FE,$06,$E2,$69,$88,$AF,$F0,$08,$D3,$63,$2F,$EF
	.dc.b	$63,$F6,$A4,$7F,$20,$D7,$F1,$ED,$7E,$2B,$5F,$C6,$35,$FE,$04,$EE
	.dc.b	$BF,$86,$F5,$7A,$E7,$F8,$DB,$5F,$E2,$AD,$7F,$88,$36,$38,$E1,$E8
	.dc.b	$6C,$FF,$1C,$3F,$81,$12,$FF,$93,$B8,$29,$86,$BC,$4A,$7F,$1F,$0E
	.dc.b	$23,$F9,$2C,$7F,$1A,$21,$FB,$52,$9F,$B4,$29,$F8,$A2,$9F,$B1,$29
	.dc.b	$FC,$9C,$DB,$FC,$F3,$7F,$F6,$36,$DF,$E5,$18,$C5,$FE,$C6,$BF,$A3
	.dc.b	$D6,$B4,$FF,$56,$D6,$AB,$D1,$BF,$F3,$EA,$D5,$F0,$9F,$DE,$C3,$A7
	.dc.b	$99,$32,$FE,$A3,$19,$AB,$3A,$6C,$E1,$87,$F5,$29,$E0,$D9,$F9,$6D
	.dc.b	$36,$7F,$16,$8D,$B6,$3F,$B2,$C6,$23,$FA,$73,$5F,$64,$68,$8F,$F4
	.dc.b	$C8,$D6,$5F,$D1,$27,$33,$FD,$32,$72,$78,$31,$FA,$1D,$3F,$12,$BE
	.dc.b	$6B,$6B,$E1,$B5,$F7,$35,$FE,$DF,$48,$C5,$74,$C7,$E5,$E7,$F6,$09
	.dc.b	$96,$30,$87,$C9,$8C,$B4,$C9,$D9,$CB,$4C,$9F,$19,$F9,$24,$7E,$23
	.dc.b	$5F,$84,$D7,$E9,$35,$E4,$70,$21,$F7,$70,$FC,$4C,$7E,$49,$B1,$F8
	.dc.b	$86,$BB,$AF,$4D,$FE,$76,$73,$3F,$21,$FC,$9E,$FF,$9C,$26,$1F,$92
	.dc.b	$71,$34,$C4,$57,$F2,$11,$A6,$C6,$5F,$AB,$1F,$8C,$23,$F8,$76,$BF
	.dc.b	$85,$6B,$F9,$56,$BF,$92,$6B,$FA,$D7,$75,$FB,$9D,$FD,$73,$FA,$A6
	.dc.b	$BF,$A5,$6B,$F8,$86,$C7,$1C,$3D,$0D,$9F,$D6,$1F,$C0,$09,$7F,$B6
	.dc.b	$70,$53,$0D,$78,$94,$FC,$B9,$FB,$42,$99,$3F,$5E,$7E,$C8,$87,$EB
	.dc.b	$4A,$7E,$AC,$A7,$EE,$CA,$72,$38,$10,$FF,$2B,$FE,$87,$FC,$7B,$1F
	.dc.b	$E3,$3B,$7F,$95,$23,$13,$FE,$64,$6B,$FE,$64,$6B,$CB,$0F,$E3,$6E
	.dc.b	$DF,$F0,$CD,$F4,$DB,$DD,$3F,$E6,$68,$3F,$07,$1F,$F3,$5C,$6C,$91
	.dc.b	$FE,$74,$76,$7F,$9D,$0C,$B4,$D1,$CF,$0F,$F3,$6C,$6A,$3F,$CB,$6F
	.dc.b	$FB,$2E,$BC,$71,$FE,$36,$8C,$11,$FF,$42,$B5,$FF,$33,$B5,$E6,$68
	.dc.b	$87,$FA,$20,$FE,$92,$FC,$16,$9F,$E8,$E2,$9B,$BF,$8E,$E5,$AE,$5E
	.dc.b	$9B,$71,$24,$E2,$61,$FE,$92,$3F,$96,$5F,$FA,$39,$FE,$9D,$1A,$7F
	.dc.b	$D2,$A6,$AF,$E9,$D7,$48,$E4,$F7,$7F,$D0,$E7,$F3,$C2,$3F,$EA,$A6
	.dc.b	$9A,$70,$47,$04,$3F,$D1,$A7,$03,$D1,$B7,$F9,$81,$F2,$1F,$CE,$F0
	.dc.b	$F4,$63,$FE,$B0,$8C,$1E,$63,$E8,$4F,$1A,$F9,$CD,$8C,$D7,$D5,$DB
	.dc.b	$E9,$13,$1C,$4C,$3E,$54,$65,$A6,$4F,$92,$E5,$A6,$4F,$F5,$A1,$FE
	.dc.b	$B7,$23,$F1,$C8,$FC,$C6,$BF,$11,$AF,$FA,$28,$E2,$6C,$FC,$76,$9F
	.dc.b	$1E,$3F,$18,$D8,$CD,$79,$F1,$F9,$13,$99,$FC,$33,$5F,$C1,$BF,$47
	.dc.b	$6F,$CB,$13,$0F,$CE,$38,$9A,$62,$2B,$F9,$18,$D3,$63,$2F,$E4,$23
	.dc.b	$F1,$04,$7F,$1A,$D7,$F0,$AD,$7E,$13,$5F,$9A,$D7,$F3,$CE,$EB,$F0
	.dc.b	$DB,$7A,$E7,$F6,$8D,$7F,$4C,$D7,$F4,$6D,$8E,$38,$7A,$1B,$3F,$AD
	.dc.b	$3F,$74,$4B,$FE,$00,$E0,$A6,$1A,$F1,$29,$FB,$23,$F8,$C6,$9F,$C9
	.dc.b	$1E,$47,$E9,$8F,$D7,$10,$FD,$99,$4D,$18,$21,$82,$1F,$8B,$38,$09
	.dc.b	$6A,$FD,$B1,$87,$F9,$8A,$BF,$A2,$C6,$23,$F6,$D1,$E7,$39,$4F,$18
	.dc.b	$E1,$19,$AE,$B5,$C6,$7E,$0C,$61,$DC,$FC,$18,$C4,$7F,$C0,$19,$F1
	.dc.b	$E9,$CE,$FB,$36,$33,$8F,$19,$CE,$16,$78,$B6,$D7,$C5,$B7,$98,$C1
	.dc.b	$27,$11,$96,$1F,$06,$BA,$B6,$BA,$BF,$14,$F1,$99,$AC,$38,$3F,$01
	.dc.b	$3C,$9B,$3C,$9E,$07,$38,$D3,$FD,$A9,$AF,$33,$11,$C5,$FA,$B1,$F7
	.dc.b	$34,$F9,$93,$9C,$7E,$94,$62,$FE,$23,$6F,$E4,$35,$FE,$F3,$3C,$2B
	.dc.b	$E1,$C7,$65,$3D,$A3,$E7,$13,$0F,$A4,$E0,$A6,$1A,$F6,$C7,$F1,$8E
	.dc.b	$CC,$63,$F9,$67,$FC,$00,$D7,$89,$F2,$8F,$9E,$E5,$FA,$C6,$2F,$E9
	.dc.b	$BD,$8F,$9C,$E2,$38,$47,$33,$27,$12,$18,$3F,$21,$D1,$FA,$77,$CF
	.dc.b	$3F,$AA,$8C,$47,$C7,$3F,$34,$E3,$1F,$D2,$46,$27,$E5,$3D,$1A,$FC
	.dc.b	$29,$E1,$7F,$B9,$7E,$E3,$72,$EA,$7C,$68,$CE,$06,$92,$34,$70,$AF
	.dc.b	$5B,$CB,$1B,$10,$FD,$14,$FE,$3D,$FC,$A3,$5F,$11,$F9,$97,$15,$E1
	.dc.b	$AF,$AB,$18,$23,$82,$4F,$16,$BF,$07,$2F,$E0,$D3,$EA,$6C,$FC,$C6
	.dc.b	$CF,$D4,$6B,$C7,$1F,$53,$FA,$6A,$FC,$CC,$7D,$A8,$FA,$A4,$C3,$0C
	.dc.b	$3E,$91,$92,$99,$3A,$19,$29,$93,$E3,$9F,$08,$87,$C7,$29,$F3,$8A
	.dc.b	$7F,$08,$9F,$CB,$11,$C9,$D0,$FE,$45,$D1,$F8,$6C,$7F,$01,$1F,$CA
	.dc.b	$63,$13,$FC,$F9,$B1,$9B,$F9,$AD,$7D,$99,$FF,$3C,$9E,$99,$7E,$9C
	.dc.b	$95,$FB,$06,$BF,$C6,$5A,$FF,$29,$6B,$FC,$89,$AF,$F1,$D7,$F6,$4C
	.dc.b	$3F,$76,$53,$F9,$21,$4F,$E6,$85,$3F,$96,$94,$FE,$50,$51,$D9,$F9
	.dc.b	$76,$C7,$F0,$48,$C1,$1C,$12,$3F,$9D,$C6,$5B,$19,$7F,$A1,$47,$2E
	.dc.b	$4F,$F5,$7A,$69,$FE,$9E,$8F,$E4,$11,$97,$FB,$03,$5F,$EB,$CD,$77
	.dc.b	$1C,$5F,$EE,$71,$96,$C7,$67,$9B,$BB,$FF,$00,$CE,$DB,$1F,$04,$93
	.dc.b	$EE,$6B,$EC,$6B,$FE,$03,$9F,$03,$FE,$BE,$6B,$F3,$A3,$6F,$F8,$7E
	.dc.b	$32,$78,$75,$FA,$7C,$9F,$C7,$23,$01,$C5,$7C,$A6,$D7,$C6,$6B,$F1
	.dc.b	$9A,$FC,$C7,$A7,$E1,$AF,$05,$FB,$07,$CC,$8C,$45,$33,$5F,$52,$32
	.dc.b	$D8,$CB,$F4,$23,$2D,$8C,$E5,$F3,$A7,$2D,$8C,$E5,$F9,$29,$CB,$63
	.dc.b	$39,$27,$2D,$8C,$C6,$0F,$66,$1E,$98,$F5,$37,$0F,$43,$5F,$33,$70
	.dc.b	$F2,$37,$0F,$D2,$FE,$1E,$FC,$A6,$5F,$E5,$A3,$11,$F9,$47,$5F,$95
	.dc.b	$58,$DC,$FE,$52,$76,$F1,$8C,$E1,$F9,$29,$E3,$39,$9E,$44,$78,$55
	.dc.b	$7A,$47,$2B,$F3,$B7,$35,$F4,$AD,$CF,$1C,$3C,$0E,$E7,$51,$88,$D5
	.dc.b	$F4,$7D,$98,$63,$07,$C6,$FE,$F0,$98,$B9,$3F,$6F,$18,$8A,$7E,$F1
	.dc.b	$FD,$96,$1F,$C4,$9D,$47,$F0,$27,$9C,$F0,$88,$7E,$BE,$3F,$8B,$11
	.dc.b	$FE,$26,$D7,$F8,$8B,$5E,$06,$88,$68,$A7,$F1,$2F,$C4,$BF,$65,$1F
	.dc.b	$C6,$1B,$1F,$C5,$1A,$FF,$14,$6B,$C7,$8F,$FC,$69,$9F,$F0,$EA,$E5
	.dc.b	$18,$C3,$F9,$8F,$E2,$7F,$98,$BF,$E4,$2F,$E7,$8C,$3F,$C7,$1C,$6F
	.dc.b	$FA,$46,$EF,$FC,$91,$B3,$CA,$7F,$7C,$D9,$D4,$62,$3A,$3C,$CC,$63
	.dc.b	$FC,$92,$BC,$2D,$8C,$DF,$A1,$E0,$74,$BE,$AD,$BF,$13,$5E,$EF,$F3
	.dc.b	$AF,$C5,$3F,$47,$1F,$CE,$1B,$7F,$CD,$1F,$C9,$DF,$F5,$67,$2D,$DF
	.dc.b	$F3,$D1,$8D,$8F,$06,$C6,$88,$78,$46,$F5,$98,$7F,$52,$AE,$87,$32
	.dc.b	$C7,$28,$D3,$C9,$AF,$27,$D8,$F8,$67,$18,$7F,$50,$FA,$FF,$EC,$64
	.dc.b	$AF,$D4,$1F,$D8,$9C,$3F,$D8,$1D,$70,$58,$FE,$CB,$1A,$78,$1C,$8F
	.dc.b	$EC,$6E,$5F,$EA,$0F,$F7,$D2,$3F,$DE,$9A,$FF,$78,$6B,$FD,$75,$AF
	.dc.b	$23,$81,$0F,$EE,$BF,$8A,$FF,$59,$8F,$EF,$ED,$8F,$EF,$6D,$77,$BF
	.dc.b	$E8,$F3,$57,$08,$C3,$5F,$F0,$0F,$E2,$DF,$94,$D7,$FC,$32,$FF,$44
	.dc.b	$31,$47,$1B,$7F,$4A,$6E,$FC,$A3,$F3,$73,$B7,$46,$27,$99,$C3,$1F
	.dc.b	$F0,$AB,$5F,$F0,$7B,$74,$FF,$05,$B7,$0F,$2D,$73,$3A,$9E,$1C,$3C
	.dc.b	$4F,$F8,$73,$F1,$7F,$CD,$13,$7F,$C6,$63,$FE,$1C,$8C,$46,$CF,$E7
	.dc.b	$CF,$E7,$CC,$C7,$F8,$A9,$E5,$BE,$E6,$4E,$5E,$27,$6C,$BF,$C4,$75
	.dc.b	$CA,$7C,$13,$AA,$E3,$18,$8B,$5C,$1E,$58,$FF,$93,$2F,$F6,$9F,$80
	.dc.b	$FF,$90,$09,$7F,$95,$BF,$F1,$8C,$62,$3F,$C7,$AE,$B8,$2C,$7F,$90
	.dc.b	$A3,$4F,$F9,$19,$FE,$93,$A9,$CC,$7F,$8B,$5F,$F2,$D1,$1F,$F2,$BB
	.dc.b	$5F,$F2,$AB,$5F,$F1,$D3,$5E,$47,$02,$1F,$E5,$2F,$C6,$3F,$C6,$D1
	.dc.b	$FE,$5C,$6C,$7F,$96,$1A,$EF,$7F,$CF,$4E,$DB,$1B,$7F,$45,$F1,$FF
	.dc.b	$9A,$3F,$C3,$E5,$DB,$F2,$ED,$7F,$CC,$6D,$9F,$F3,$4B,$5F,$F3,$1B
	.dc.b	$72,$FF,$2F,$B7,$4F,$F4,$0B,$72,$34,$97,$FE,$32,$7C,$6D,$3F,$53
	.dc.b	$F8,$CF,$F3,$12,$6D,$F8,$2A,$46,$23,$FC,$F2,$EB,$7B,$1F,$E1,$78
	.dc.b	$D6,$5F,$E8,$27,$F9,$D1,$87,$FC,$FA,$F1,$75,$1C,$48,$74,$9E,$AC
	.dc.b	$8D,$B9,$69,$FE,$90,$F9,$1F,$F4,$A9,$1F,$CB,$65,$FE,$A6,$8C,$47
	.dc.b	$FA,$49,$D6,$F6,$3F,$D2,$F1,$A7,$A1,$CB,$14,$66,$3F,$D0,$4F,$FA
	.dc.b	$BC,$8F,$FA,$B1,$AF,$FA,$A9,$AF,$FA,$3D,$AF,$23,$81,$0F,$F5,$3F
	.dc.b	$E3,$5F,$E8,$B8,$FF,$58,$B6,$3F,$D5,$AD,$77,$B7,$F3,$39,$DB,$63
	.dc.b	$6F,$F3,$7F,$C6,$FF,$97,$9F,$E7,$42,$EB,$FC,$B1,$AF,$FA,$E1,$B3
	.dc.b	$FE,$BE,$6B,$FE,$B8,$6E,$FF,$EB,$46,$E7,$FE,$D6,$6E,$46,$92,$FF
	.dc.b	$D0,$EF,$8D,$A7,$F3,$EF,$C7,$3F,$C0,$C4,$D7,$F2,$CD,$7F,$D9,$AE
	.dc.b	$76,$B1,$FE,$6B,$8D,$65,$FE,$D6,$7F,$C1,$C6,$1F,$F6,$8B,$C5,$D4
	.dc.b	$71,$21,$D2,$7A,$B2,$36,$E5,$A7,$FB,$93,$EE,$7F,$DD,$24,$7F,$18
	.dc.b	$D7,$FD,$CA,$E7,$7B,$1F,$EE,$A8,$D3,$CC,$E2,$64,$FF,$7D,$BF,$F0
	.dc.b	$04,$8E,$08,$7F,$BF,$1C,$B4,$C9,$FE,$E3,$72,$D3,$FD,$F7,$F8,$EF
	.dc.b	$FB,$76,$11,$8D,$BF,$C0,$51,$88,$FF,$7F,$CE,$DD,$4F,$B1,$CB,$C1
	.dc.b	$C9,$D3,$6F,$F7,$EF,$E3,$DF,$E0,$43,$FD,$92,$5D,$3F,$C0,$2D,$7C
	.dc.b	$11,$C1,$FF,$68,$C7,$06,$9C,$0A,$70,$DD,$BA,$7A,$5F,$8E,$73,$18
	.dc.b	$85,$F7,$E7,$E5,$7C,$70,$ED,$18,$BF,$13,$63,$37,$D6,$B8,$47,$2F
	.dc.b	$C7,$FF,$CC,$84,$DB,$F9,$A3,$5D,$1F,$EC,$77,$FC,$D0,$66,$BA,$3C
	.dc.b	$77,$DC,$C9,$CB,$C4,$EA,$53,$4F,$BB,$3F,$8B,$5D,$A7,$53,$C9,$F1
	.dc.b	$D7,$0C,$D7,$F3,$0F,$C2,$57,$BD,$AF,$5C,$76,$FF,$2F,$8C,$CF,$29
	.dc.b	$CC,$FC,$E9,$C9,$D5,$E0,$76,$9E,$11,$A9,$E3,$3E,$13,$67,$33,$C6
	.dc.b	$73,$3C,$AF,$07,$F3,$8D,$7F,$42,$D7,$81,$EE,$70,$48,$EB,$1A,$39
	.dc.b	$B8,$8D,$47,$A6,$3F,$27,$F9,$03,$5F,$F3,$74,$7A,$E3,$11,$E9,$9F
	.dc.b	$C8,$C6,$1F,$71,$88,$E2,$F5,$A4,$EE,$3E,$95,$75,$AF,$F3,$E7,$E4
	.dc.b	$3F,$E7,$41,$A7,$F9,$C1,$C3,$46,$BF,$A2,$8D,$C5,$33,$7F,$9F,$8C
	.dc.b	$B6,$32,$FE,$0E,$32,$D8,$CE,$5F,$83,$9C,$B6,$33,$87,$EE,$A7,$2D
	.dc.b	$8C,$E4,$9C,$B6,$33,$18,$3D,$AF,$A7,$0F,$53,$70,$F4,$3F,$C0,$3F
	.dc.b	$22,$FF,$6D,$13,$5F,$F1,$53,$5D,$02,$EB,$86,$57,$F0,$39,$E5,$38
	.dc.b	$8F,$D3,$47,$37,$53,$D8,$E8,$E1,$E1,$9F,$EA,$67,$FD,$D8,$71,$9E
	.dc.b	$73,$C1,$B5,$C3,$3F,$E4,$7F,$85,$7E,$A4,$93,$FE,$23,$6B,$FC,$81
	.dc.b	$CF,$1D,$8F,$C7,$46,$9F,$E3,$CF,$FA,$00,$C3,$FA,$C7,$F4,$44,$7F
	.dc.b	$67,$AA,$31,$1F,$B4,$6B,$FA,$F6,$BC,$CE,$2E,$32,$FE,$15,$F9,$1F
	.dc.b	$E7,$DF,$E1,$91,$FC,$16,$31,$87,$C0,$AF,$D3,$46,$2F,$C5,$18,$DF
	.dc.b	$FD,$C1,$7C,$5B,$19,$7F,$A1,$E1,$FE,$DC,$3F,$1E,$5D,$7F,$DB,$2D
	.dc.b	$76,$7E,$59,$DB,$4D,$94,$DE,$DC,$1B,$5F,$D2,$E7,$8C,$E6,$31,$3C
	.dc.b	$5B,$19,$7F,$99,$4F,$8D,$FE,$A9,$9F,$E5,$F5,$FE,$BA,$D8,$CC,$7F
	.dc.b	$A1,$E3,$11,$FC,$ED,$B3,$FC,$E5,$BA,$7F,$6E,$7F,$45,$AF,$4B,$FC
	.dc.b	$95,$6D,$B5,$B6,$E8,$BE,$38,$78,$A7,$91,$1F,$2E,$15,$F2,$47,$29
	.dc.b	$F1,$D7,$38,$F5,$19,$21,$E5,$AE,$58,$7F,$53,$BF,$E9,$84,$3F,$D0
	.dc.b	$C5,$3F,$C0,$66,$6E,$BF,$CC,$4D,$1F,$E0,$13,$FD,$C6,$60,$FE,$9E
	.dc.b	$7F,$45,$21,$FA,$42,$9F,$D5,$8A,$7F,$51,$29,$C8,$E0,$43,$FC,$41
	.dc.b	$E9,$7F,$4A,$D3,$CB,$18,$7F,$40,$FF,$6E,$8D,$3E,$73,$FA,$B3,$87
	.dc.b	$FB,$D1,$B8,$E2,$E9,$FD,$6C,$3C,$BF,$C1,$B7,$1F,$BE,$DC,$7C,$B3
	.dc.b	$A2,$4E,$9F,$C0,$DF,$F8,$56,$B9,$4F,$03,$44,$AC,$32,$7F,$C7,$93
	.dc.b	$C1,$B5,$D3,$1F,$F0,$CC,$E6,$7B,$4F,$8A,$BB,$5F,$39,$CC,$78,$0C
	.dc.b	$BD,$1F,$6F,$F0,$07,$F6,$AD,$7E,$8B,$FD,$5B,$03,$9A,$F2,$B9,$30
	.dc.b	$FB,$DC,$6D,$FD,$82,$78,$BF,$DD,$A7,$8B,$99,$DD,$78,$23,$53,$CD
	.dc.b	$D1,$8C,$BE,$0D,$78,$1B,$19,$D9,$87,$83,$0E,$93,$D5,$B8,$78,$9A
	.dc.b	$F7,$6D,$FC,$EF,$A6,$FC,$71,$27,$FD,$F3,$68,$C4,$7E,$61,$D6,$D6
	.dc.b	$3F,$27,$1A,$7E,$EB,$FE,$96,$8D,$44,$3F,$CA,$73,$D5,$FF,$20,$B5
	.dc.b	$FF,$2A,$B5,$DC,$FF,$8E,$A3,$11,$FE,$9F,$6B,$FE,$A4,$6B,$F1,$A3
	.dc.b	$81,$1E,$ED,$7F,$67,$FB,$FF,$E9,$A3,$F6,$0D,$8F,$D7,$B5,$DE,$FF
	.dc.b	$E4,$57,$DE,$69,$C3,$FA,$59,$E7,$18,$C3,$F3,$BF,$BF,$7E,$C8,$9B
	.dc.b	$37,$FE,$0B,$18,$8A,$7E,$D9,$FD,$36,$DF,$82,$75,$8F,$EB,$1E,$75
	.dc.b	$C2,$A4,$E5,$E4,$77,$70,$E9,$FD,$E4,$F2,$72,$43,$31,$CA,$78,$5F
	.dc.b	$EE,$3F,$7D,$FC,$C6,$1F,$96,$7F,$5C,$63,$43,$8D,$FF,$54,$DD,$FF
	.dc.b	$92,$B7,$2F,$E0,$ED,$7F,$88,$37,$0F,$05,$7E,$96,$FA,$BD,$0C,$4F
	.dc.b	$84,$E1,$87,$96,$32,$6C,$F0,$96,$FF,$8B,$FD,$CF,$F3,$32,$3F,$A6
	.dc.b	$6B,$FC,$F1,$CF,$05,$8F,$E6,$F1,$A7,$99,$C4,$C9,$FA,$27,$F3,$C4
	.dc.b	$70,$43,$F4,$AE,$5A,$64,$FD,$6B,$96,$9F,$D2,$3F,$7C,$FE,$87,$1F
	.dc.b	$D3,$5B,$19,$DF,$F2,$D8,$7F,$0E,$AD,$46,$32,$46,$23,$DD,$5F,$D7
	.dc.b	$2F,$FA,$7F,$EF,$7F,$D3,$89,$BF,$E3,$E7,$FA,$44,$62,$3F,$B1,$BA
	.dc.b	$DE,$C7,$F5,$48,$D4,$FF,$48,$7F,$43,$89,$CC,$74,$9E,$07,$7D,$BF
	.dc.b	$A5,$CE,$9B,$1A,$BE,$EC,$AE,$CE,$5A,$FF,$78,$FD,$EB,$F3,$47,$F4
	.dc.b	$92,$ED,$F9,$16,$BB,$3F,$A3,$3B,$69,$B2,$9B,$E0,$E0,$DA,$FE,$CF
	.dc.b	$3C,$67,$31,$89,$E2,$D8,$CB,$FD,$D6,$7C,$6F,$F7,$0F,$C2,$7F,$C1
	.dc.b	$84,$7F,$9F,$B5,$FF,$0A,$B9,$E0,$B1,$FE,$10,$8D,$3C,$CE,$26,$4F
	.dc.b	$E7,$6F,$F3,$72,$38,$21,$FC,$F9,$CB,$4C,$9F,$D1,$9C,$B4,$FF,$10
	.dc.b	$FE,$F3,$FE,$1C,$8F,$F1,$33,$63,$3B,$FE,$5E,$72,$D8,$CB,$EC,$7F
	.dc.b	$C6,$73,$FE,$26,$FD,$E3,$FC,$4A,$4C,$FF,$15,$C5,$FD,$C2,$31,$1F
	.dc.b	$E3,$A7,$5A,$D8,$FF,$15,$C6,$AF,$FC,$3E,$FE,$60,$C3,$FE,$29,$78
	.dc.b	$BA,$8E,$24,$3A,$4F,$56,$46,$DC,$B4,$FF,$16,$FE,$EF,$FC,$BC,$FF
	.dc.b	$0F,$17,$6F,$E4,$CD,$7F,$CA,$6D,$8E,$31,$FE,$1D,$8E,$2D,$38,$94
	.dc.b	$E2,$58,$E3,$B6,$23,$2D,$8C,$9D,$A3,$2D,$3F,$C8,$BF,$67,$FE,$61
	.dc.b	$23,$FE,$14,$6B,$FE,$6A,$73,$B5,$8F,$F3,$2C,$69,$E6,$71,$32,$7F
	.dc.b	$84,$1F,$F0,$61,$1C,$10,$FF,$09,$B9,$69,$93,$FC,$36,$E5,$A7,$F9
	.dc.b	$DB,$E9,$BF,$CE,$31,$FE,$7B,$6C,$67,$6F,$E5,$93,$96,$C6,$5F,$63
	.dc.b	$FE,$88,$9F,$F3,$DF,$C7,$FF,$9E,$89,$9F,$E5,$B9,$11,$88,$FF,$47
	.dc.b	$BA,$D6,$C7,$FA,$02,$35,$7F,$E7,$47,$F9,$69,$87,$FC,$FC,$F1,$75
	.dc.b	$1C,$48,$74,$9E,$AC,$8D,$B9,$69,$FE,$82,$FA,$3F,$F0,$21,$FE,$73
	.dc.b	$2E,$9F,$DE,$DA,$FF,$AA,$1B,$1C,$63,$FC,$E5,$1C,$5A,$71,$29,$C4
	.dc.b	$B1,$C7,$5C,$46,$5B,$19,$3B,$46,$5A,$7F,$A6,$BE,$87,$FD,$6C,$47
	.dc.b	$FC,$D2,$D7,$FD,$7E,$E7,$6B,$1F,$EB,$A8,$D3,$CC,$E2,$64,$FF,$32
	.dc.b	$BF,$E6,$12,$38,$21,$FE,$68,$72,$D3,$27,$F9,$C1,$CB,$4F,$F6,$5F
	.dc.b	$D0,$FF,$B1,$A3,$FD,$A0,$D8,$CE,$DF,$E0,$19,$CB,$63,$2F,$B1,$FF
	.dc.b	$6E,$CF,$FB,$43,$E8,$3F,$D9,$E4,$CF,$F2,$CD,$7F,$DC,$6E,$74,$B1
	.dc.b	$FE,$D4,$8D,$57,$FB,$21,$FF,$00,$98,$7F,$DA,$4F,$17,$51,$C4,$87
	.dc.b	$49,$EA,$C8,$DB,$96,$9F,$ED,$6F,$9E,$FF,$2C,$1F,$EC,$72,$E7,$FE
	.dc.b	$4F,$6B,$FE,$FC,$6C,$71,$8F,$F6,$34,$71,$69,$C4,$A7,$12,$C7,$1D
	.dc.b	$31,$19,$6C,$64,$ED,$19,$69,$FE,$EE,$F9,$D3,$0C,$14,$FF,$7D,$3F
	.dc.b	$EB,$77,$26,$FA,$7F,$DE,$71,$B8,$D9,$C6,$32,$FF,$B2,$23,$51,$88
	.dc.b	$E5,$CD,$FE,$D8,$9C,$B6,$38,$1C,$DC,$3F,$EC,$28,$FF,$60,$11,$FF
	.dc.b	$63,$B5,$D1,$82,$18,$21,$FE,$D0,$78,$3F,$12,$53,$FE,$EC,$F9,$D6
	.dc.b	$DF,$E5,$FB,$F3,$F1,$A7,$33,$EC,$84,$EA,$73,$87,$CA,$3B,$5F,$5B
	.dc.b	$E9,$1E,$FB,$E9,$18,$8E,$95,$C9,$61,$E6,$A7,$CE,$35,$95,$DA,$73
	.dc.b	$18,$8E,$38,$A7,$3B,$7A,$6F,$95,$D3,$86,$1D,$E3,$38,$75,$7F,$CE
	.dc.b	$26,$1E,$B7,$E0,$AE,$47,$3B,$B3,$C6,$3A,$E1,$E3,$BE,$58,$77,$9F
	.dc.b	$EB,$51,$89,$D4,$70,$C3,$B5,$F3,$C3,$AB,$E2,$BE,$13,$A8,$C5,$74
	.dc.b	$39,$7C,$D7,$FB,$4C,$6B,$FE,$C9,$70,$D1,$B7,$DF,$8D,$C5,$33,$5F
	.dc.b	$46,$32,$D8,$CB,$F1,$A3,$2D,$8C,$E7,$F7,$67,$2D,$8C,$E7,$F6,$67
	.dc.b	$2D,$8C,$E2,$9C,$B6,$33,$18,$3D,$AF,$A7,$0F,$53,$70,$F4,$3F,$9D
	.dc.b	$F9,$7F,$D0,$92,$1C,$99,$7F,$35,$19,$71,$19,$61,$F9,$58,$E5,$18
	.dc.b	$8F,$C4,$C7,$F0,$D7,$F8,$BE,$08,$C4,$7E,$65,$B1,$98,$E0,$D7,$F8
	.dc.b	$B3,$5F,$E3,$CD,$7F,$91,$35,$EF,$1D,$08,$F4,$6B,$FB,$FF,$95,$FD
	.dc.b	$A6,$7F,$B3,$9F,$E0,$B1,$8C,$FC,$D5,$FB,$D8,$C5,$F8,$6F,$F7,$51
	.dc.b	$8B,$E3,$C1,$FE,$ED,$6B,$FC,$37,$E4,$BF,$36,$4D,$BF,$C7,$CD,$74
	.dc.b	$7E,$95,$FF,$72,$99,$C7,$F8,$23,$C7,$83,$73,$27,$2F,$13,$A9,$4D
	.dc.b	$3F,$BE,$9E,$B1,$98,$E2,$F8,$A7,$78,$FF,$17,$F8,$9F,$D6,$EF,$FC
	.dc.b	$9D,$FC,$A9,$8B,$FE,$30,$E3,$7F,$CD,$B6,$78,$47,$ED,$A7,$6E,$8C
	.dc.b	$3F,$C4,$A7,$A3,$63,$35,$E3,$78,$1D,$2B,$AB,$67,$AB,$5E,$4D,$9E
	.dc.b	$4D,$DB,$F9,$65,$FE,$C3,$6F,$E8,$E4,$BF,$D3,$9F,$CD,$5C,$3F,$D2
	.dc.b	$5D,$6F,$63,$FA,$24,$69,$E0,$72,$3F,$97,$B9,$7F,$36,$FE,$90,$8F
	.dc.b	$E9,$1A,$FE,$91,$AF,$E9,$1A,$F2,$38,$10,$FE,$BD,$F1,$DF,$9A,$8F
	.dc.b	$EC,$6D,$8F,$EB,$4D,$77,$B7,$F6,$A7,$D2,$61,$F7,$34,$FE,$B1,$FB
	.dc.b	$AF,$F3,$82,$6D,$FE,$9B,$6B,$FD,$BD,$CE,$D6,$3F,$A2,$46,$A7,$9C
	.dc.b	$71,$32,$43,$84,$F1,$64,$70,$72,$D2,$1F,$D7,$A7,$C2,$E3,$5F,$EE
	.dc.b	$9F,$BA,$7E,$41,$BB,$7E,$43,$6F,$EE,$13,$96,$C6,$74,$4E,$5B,$19
	.dc.b	$8C,$1F,$CF,$0A,$E4,$FE,$82,$E5,$A6,$77,$FF,$0B,$46,$5B,$19,$C3
	.dc.b	$FC,$11,$39,$6C,$65,$FE,$FB,$1F,$D3,$3F,$0F,$FF,$11,$92,$3F,$29
	.dc.b	$8F,$F8,$56,$31,$1F,$E2,$57,$5B,$D8,$FF,$10,$46,$9F,$F0,$C3,$FE
	.dc.b	$51,$7F,$C1,$6E,$5F,$E6,$EF,$F4,$02,$3F,$D0,$1A,$FF,$40,$6B,$FD
	.dc.b	$01,$AF,$23,$81,$0F,$F1,$D7,$EE,$7F,$CD,$A3,$FC,$7E,$D8,$FF,$1B
	.dc.b	$35,$DE,$FF,$9D,$7C,$E6,$9C,$5F,$C3,$8C,$47,$F9,$1F,$F7,$2F,$F0
	.dc.b	$A9,$37,$FF,$7C,$35,$D1,$FE,$1F,$7F,$C7,$DC,$18,$89,$5F,$E3,$F9
	.dc.b	$E5,$38,$8F,$F1,$9C,$73,$75,$3D,$8E,$84,$3B,$CF,$1B,$FF,$27,$57
	.dc.b	$1A,$E7,$5C,$4F,$F3,$07,$EE,$3F,$97,$3F,$C1,$E5,$D7,$F3,$8D,$76
	.dc.b	$7F,$85,$9D,$B4,$D9,$4D,$ED,$C1,$B5,$FE,$5D,$9E,$33,$98,$C4,$F1
	.dc.b	$6C,$65,$FF,$2A,$CF,$8D,$FF,$33,$7D,$DF,$F9,$F0,$8F,$E6,$B8,$7F
	.dc.b	$CC,$D1,$88,$FF,$3F,$3A,$DE,$C7,$F9,$E6,$34,$F4,$39,$70,$A3,$31
	.dc.b	$FE,$10,$7F,$C3,$24,$7F,$C3,$2D,$7F,$C3,$2D,$7F,$C3,$2D,$79,$1C
	.dc.b	$08,$7F,$A4,$FF,$70,$FF,$07,$C7,$FA,$61,$B1,$FE,$8F,$6B,$BD,$FF
	.dc.b	$2F,$87,$C0,$8D,$46,$2F,$D5,$18,$8F,$F5,$07,$D9,$7F,$9B,$89,$AB
	.dc.b	$83,$2F,$F9,$DE,$32,$E2,$32,$CA,$FF,$4C,$4F,$29,$C4,$7F,$A3,$23
	.dc.b	$8B,$A9,$EC,$D7,$7C,$1F,$EA,$59,$E5,$39,$21,$CE,$7C,$87,$27,$0F
	.dc.b	$FA,$EF,$7F,$E6,$47,$F9,$A8,$A7,$F3,$72,$9B,$3F,$CE,$26,$CA,$6C
	.dc.b	$A6,$F7,$E0,$D9,$FF,$5C,$3C,$63,$31,$88,$E2,$D8,$CB,$FE,$B0,$8F
	.dc.b	$1B,$FE,$C1,$FC,$2B,$FD,$AE,$47,$F2,$6D,$7F,$DB,$0E,$77,$B1,$FE
	.dc.b	$D3,$8D,$3C,$CE,$26,$4F,$F3,$7B,$FE,$72,$23,$82,$1F,$E7,$27,$2D
	.dc.b	$32,$7F,$9D,$1C,$B4,$FF,$72,$7E,$E1,$FE,$68,$8F,$F7,$8B,$63,$3A
	.dc.b	$7E,$56,$72,$D8,$CB,$EC,$7F,$DE,$D3,$FE,$72,$FD,$BF,$FD,$8A,$4C
	.dc.b	$BF,$05,$A7,$FB,$B6,$31,$1F,$EF,$D7,$5A,$D8,$FF,$69,$46,$AF,$FD
	.dc.b	$E0,$FF,$36,$6C,$68,$87,$19,$E4,$C8,$DB,$96,$90,$FF,$76,$CF,$89
	.dc.b	$C5,$37,$FF,$7E,$7E,$DD,$FE,$09,$3F,$D8,$25,$D7,$FC,$1C,$D5,$52
	.dc.b	$B9,$70,$2B,$53,$9B,$EB,$7F,$EC,$E9,$EA,$D3,$A9,$4E,$B2,$BE,$B3
	.dc.b	$D2,$31,$5D,$37,$DD,$6A,$F9,$C6,$27,$C3,$17,$0F,$33,$DE,$FC,$D1
	.dc.b	$8A,$D6,$1F,$EE,$2F,$C6,$3C,$75,$A6,$BE,$2A,$47,$F8,$47,$1F,$86
	.dc.b	$E9,$B3,$C9,$AF,$8E,$38,$62,$AE,$73,$B8,$C4,$F3,$6C,$66,$77,$C1
	.dc.b	$CE,$31,$5F,$ED,$E2,$3F,$ED,$E6,$BF,$ED,$E6,$BF,$ED,$E6,$BF,$ED
	.dc.b	$E6,$BF,$EE,$C7,$37,$F8,$0F,$DB,$7F,$DA,$B1,$E5,$B4,$62,$BC,$CD
	.dc.b	$7E,$93,$5F,$A2,$D7,$E6,$3F,$EE,$0F,$B6,$6D,$27,$EB,$C6,$23,$EE
	.dc.b	$3F,$A9,$8F,$A8,$E5,$DC,$71,$72,$DC,$3F,$33,$3C,$5B,$19,$CB,$EA
	.dc.b	$4F,$3A,$E3,$3E,$03,$F8,$0C,$E9,$E1,$82,$B8,$5F,$49,$E1,$BF,$16
	.dc.b	$61,$E6,$72,$D2,$1E,$03,$D7,$1C,$1F,$C9,$7E,$D7,$FE,$81,$1B,$FF
	.dc.b	$A1,$5C,$34,$6F,$FA,$38,$DC,$53,$33,$FA,$48,$CB,$63,$2F,$81,$E9
	.dc.b	$3C,$5E,$4D,$8E,$0D,$78,$37,$0F,$E1,$35,$E6,$AE,$35,$DD,$AF,$77
	.dc.b	$F7,$5F,$82,$FE,$3C,$91,$FE,$28,$6B,$A3,$F4,$CF,$EB,$B3,$3A,$8F
	.dc.b	$E0,$2F,$2D,$F8,$44,$3F,$80,$CF,$07,$F8,$09,$FB,$79,$AD,$33,$1B
	.dc.b	$7F,$83,$B5,$E7,$1C,$23,$11,$FC,$81,$AF,$F2,$E6,$BD,$4E,$71,$C4
	.dc.b	$FD,$87,$ED,$4E,$4F,$F1,$84,$61,$B3,$FC,$1E,$75,$1E,$D8,$FE,$35
	.dc.b	$18,$7E,$19,$B8,$E2,$FF,$29,$8D,$3F,$C8,$7F,$69,$FC,$EF,$07,$F1
	.dc.b	$96,$E0,$7F,$59,$19,$DF,$F4,$B1,$A2,$3F,$AA,$9E,$0E,$88,$7F,$39
	.dc.b	$8E,$71,$A8,$C1,$AA,$FE,$77,$3A,$3A,$C6,$8D,$EF,$67,$71,$C5,$F6
	.dc.b	$3E,$39,$F5,$FE,$D3,$FE,$ED,$3F,$47,$3F,$D0,$5C,$6B,$FE,$F3,$6B
	.dc.b	$FD,$3D,$B1,$FD,$39,$AF,$18,$FD,$5E,$FF,$DA,$A7,$AB,$63,$39,$7F
	.dc.b	$45,$9C,$B6,$33,$92,$72,$D8,$CC,$60,$F0,$D7,$5E,$0F,$34,$FA,$4D
	.dc.b	$9F,$C2,$B0,$FE,$CB,$F8,$AF,$F3,$D2,$3F,$E9,$66,$BF,$DB,$5C,$F0
	.dc.b	$58,$FE,$AD,$1A,$79,$9C,$4C,$9F,$A8,$7F,$48,$47,$04,$3F,$56,$E5
	.dc.b	$A6,$4F,$DB,$B9,$69,$FD,$9F,$F6,$8F,$CC,$E1,$FE,$18,$9F,$EB,$11
	.dc.b	$8C,$FF,$B9,$46,$38,$3F,$D3,$4D,$7D,$97,$FE,$0B,$AE,$1A,$2B,$53
	.dc.b	$FE,$0C,$FD,$9F,$F4,$64,$DF,$F3,$CD,$7F,$C2,$8E,$78,$2C,$7F,$4A
	.dc.b	$8D,$4F,$38,$E2,$64,$87,$09,$E2,$C8,$E0,$E5,$A4,$3F,$C3,$53,$E1
	.dc.b	$71,$9F,$F8,$F3,$F6,$6F,$E9,$7B,$FF,$8A,$5F,$C6,$98,$9F,$F1,$4B
	.dc.b	$8D,$FF,$26,$D9,$E1,$1F,$8B,$9F,$F1,$44,$F1,$6C,$66,$BC,$0F,$23
	.dc.b	$1B,$7F,$89,$6B,$84,$EA,$7A,$DF,$66,$DF,$91,$E3,$A2,$F8,$E1,$CE
	.dc.b	$73,$18,$9D,$46,$0F,$11,$FE,$49,$FC,$8B,$FC,$3C,$4B,$FF,$7B,$B5
	.dc.b	$D1,$FE,$35,$7F,$A8,$19,$7F,$C9,$EF,$1E,$0D,$C4,$3F,$A7,$C7,$F4
	.dc.b	$52,$3F,$D3,$1A,$FF,$4C,$6B,$C0,$D1,$0D,$14,$FF,$2F,$FE,$CB,$F9
	.dc.b	$88,$FF,$1C,$36,$3F,$C9,$8D,$7F,$CD,$0D,$78,$F0,$FF,$BF,$67,$8B
	.dc.b	$FE,$FF,$FD,$93,$FA,$09,$37,$FE,$78,$D7,$FC,$E4,$E7,$82,$C7,$F8
	.dc.b	$A2,$35,$7F,$E7,$47,$F9,$F9,$87,$FC,$E0,$F1,$75,$1C,$48,$74,$9E
	.dc.b	$AC,$8D,$B9,$69,$FE,$82,$FD,$8F,$F0,$47,$F2,$A2,$E7,$F8,$A6,$BF
	.dc.b	$E8,$B6,$C7,$18,$FE,$67,$1C,$5A,$71,$29,$C4,$B1,$C7,$7C,$46,$5B
	.dc.b	$19,$3B,$46,$5A,$7F,$A6,$BE,$9B,$FC,$E4,$45,$AF,$FA,$86,$31,$14
	.dc.b	$FF,$45,$3F,$E2,$53,$2F,$33,$27,$23,$07,$13,$FC,$3A,$FF,$86,$C8
	.dc.b	$FF,$88,$9A,$E8,$C1,$0C,$10,$FF,$17,$BC,$1F,$F5,$AF,$D4,$7F,$2E
	.dc.b	$8F,$F4,$73,$63,$FD,$44,$D7,$FD,$7C,$D7,$8E,$9F,$E2,$57,$D2,$61
	.dc.b	$FF,$62,$7E,$C3,$FC,$04,$FE,$D2,$5C,$3F,$84,$B5,$D9,$FD,$C1,$DB
	.dc.b	$4D,$94,$DE,$3C,$1B,$5F,$ED,$09,$E3,$39,$8C,$4F,$16,$C6,$5F,$F6
	.dc.b	$84,$F8,$DF,$F6,$CF,$EB,$DF,$E3,$02,$69,$FE,$2B,$6B,$FE,$E1,$73
	.dc.b	$AD,$8F,$F4,$BC,$6A,$79,$C7,$13,$24,$38,$4F,$16,$47,$07,$2D,$21
	.dc.b	$FE,$E7,$9F,$0B,$8E,$47,$27,$FB,$87,$F0,$9F,$F6,$51,$27,$F2,$07
	.dc.b	$FB,$BD,$C3,$FE,$FB,$75,$BD,$8F,$F7,$24,$69,$FF,$7E,$BF,$E8,$13
	.dc.b	$0F,$F9,$BD,$FF,$39,$11,$FF,$3C,$B5,$FF,$3C,$B5,$FF,$41,$B5,$E4
	.dc.b	$70,$20,$25,$7B,$B9,$1F,$AD,$FF,$82,$2B,$FD,$D0,$DA,$FF,$7E,$35
	.dc.b	$E4,$2B,$94,$66,$3A,$E0,$BE,$BB,$FE,$7E,$7E,$41,$D1,$C5,$F8,$FF
	.dc.b	$58,$FE,$C0,$7F,$92,$0B,$C1,$FD,$A1,$AF,$8A,$38,$3F,$E5,$08,$E0
	.dc.b	$D3,$81,$4E,$18,$79,$63,$9D,$FA,$30,$46,$23,$DD,$3B,$AE,$51,$B9
	.dc.b	$D5,$F8,$B8,$3A,$E1,$CA,$FC,$31,$8A,$FF,$31,$FE,$AB,$FE,$94,$26
	.dc.b	$FF,$E8,$F6,$BC,$9C,$F0,$58,$FF,$7D,$C6,$B2,$F4,$3F,$E9,$23,$0F
	.dc.b	$A3,$14,$72,$8D,$CF,$22,$1D,$6B,$B3,$23,$6E,$5A,$7C,$7F,$B4,$FC
	.dc.b	$1B,$4B,$9F,$E3,$63,$FD,$BC,$6A,$3C,$8D,$8D,$47,$0D,$3F,$11,$3C
	.dc.b	$23,$18,$A3,$84,$EA,$73,$1B,$7E,$8B,$FE,$DB,$23,$FE,$E2,$6B,$FE
	.dc.b	$E2,$6B,$FE,$EC,$6B,$FE,$E1,$6B,$FE,$E1,$73,$5F,$A9,$FB,$4F,$F3
	.dc.b	$7C,$79,$2D,$18,$AF,$4B,$5F,$C2,$B5,$FC,$83,$5F,$B0,$FF,$9E,$BF
	.dc.b	$50,$FF,$21,$1F,$EA,$A2,$ED,$FE,$4A,$6B,$F9,$06,$CF,$E6,$DA,$FE
	.dc.b	$0D,$B8,$7F,$01,$6E,$3F,$B7,$6E,$06,$92,$FF,$D7,$AF,$8D,$A7,$F9
	.dc.b	$73,$2F,$F7,$79,$0F,$F7,$51,$4F,$D2,$99,$CE,$A2,$93,$1F,$DA,$3F
	.dc.b	$EE,$E3,$93,$87,$83,$63,$83,$FA,$18,$DB,$5F,$19,$C1,$C3,$C3,$1F
	.dc.b	$E4,$8F,$73,$4F,$F2,$7F,$C6,$BE,$F1,$27,$FA,$1C,$65,$FC,$BC,$65
	.dc.b	$FE,$06,$FF,$31,$E3,$B1,$A2,$18,$72,$C3,$F5,$B9,$27,$F5,$51,$FC
	.dc.b	$EE,$30,$E1,$AF,$EB,$DA,$F2,$8E,$2F,$F3,$56,$C6,$A3,$F9,$EB,$5E
	.dc.b	$D1,$A2,$3D,$DF,$D0,$7E,$97,$FE,$D0,$6F,$17,$F3,$CD,$7B,$57,$EE
	.dc.b	$AB,$4D,$9D,$5F,$87,$1F,$EA,$95,$AA,$FE,$56,$FF,$B7,$3F,$15,$FF
	.dc.b	$52,$0D,$7F,$D5,$6E,$1A,$38,$FF,$99,$C6,$E2,$99,$AF,$E8,$11,$96
	.dc.b	$C6,$5F,$E5,$51,$96,$C6,$6F,$FB,$54,$E5,$B1,$9C,$BF,$B1,$4E,$5B
	.dc.b	$19,$C5,$39,$6C,$66,$30,$7B,$5F,$4E,$1E,$A6,$E1,$E8,$7F,$A8,$FE
	.dc.b	$61,$FA,$BD,$A9,$F9,$AC,$7F,$BA,$47,$EB,$F7,$CC,$CB,$FE,$C5,$3C
	.dc.b	$5B,$09,$D4,$7F,$4D,$9D,$36,$7B,$64,$9E,$D3,$CA,$78,$E7,$FE,$11
	.dc.b	$9E,$2D,$8E,$D3,$B2,$1D,$5F,$5D,$F2,$C3,$D7,$1E,$0B,$E7,$18,$9D
	.dc.b	$60,$FB,$8F,$E6,$C4,$C3,$F7,$35,$FE,$17,$8C,$47,$F7,$A7,$5C,$16
	.dc.b	$3F,$A9,$C6,$9F,$F0,$43,$FB,$D6,$9A,$3F,$68,$FE,$B8,$8E,$08,$7E
	.dc.b	$DD,$CB,$4C,$9F,$C0,$9C,$B4,$FE,$CB,$F7,$0E,$0F,$D5,$CF,$F5,$38
	.dc.b	$C4,$7A,$67,$FC,$2B,$18,$7C,$E6,$23,$8B,$D1,$AE,$5F,$F0,$EF,$E6
	.dc.b	$DF,$B1,$DE,$EF,$FB,$3A,$FF,$15,$4E,$5B,$19,$7F,$C2,$91,$96,$C6
	.dc.b	$77,$FF,$2A,$4E,$5B,$19,$CF,$FC,$99,$39,$6C,$67,$44,$E5,$B1,$98
	.dc.b	$C1,$FA,$F2,$B9,$3F,$68,$E5,$A6,$70,$F7,$BF,$01,$B8,$7F,$92,$7F
	.dc.b	$3A,$FC,$8D,$9A,$7E,$8A,$3F,$C6,$8E,$38,$3F,$4E,$DD,$FF,$C9,$2E
	.dc.b	$62,$9A,$3F,$4B,$3C,$9C,$1A,$37,$36,$3C,$91,$A6,$C6,$5E,$51,$73
	.dc.b	$FF,$42,$57,$9D,$B1,$FB,$CF,$CB,$3F,$C3,$C4,$AF,$D4,$E0,$8C,$47
	.dc.b	$F9,$89,$D7,$05,$8F,$F1,$DC,$69,$FF,$38,$3F,$A9,$7F,$CB,$0E,$5F
	.dc.b	$EA,$6F,$F5,$62,$3F,$D7,$DA,$FF,$5F,$6B,$FD,$A1,$AF,$23,$81,$0F
	.dc.b	$F3,$BF,$E7,$DF,$A3,$7C,$27,$F9,$41,$C6,$1F,$EA,$79,$FF,$1F,$46
	.dc.b	$36,$FF,$4F,$46,$37,$FE,$B6,$D7,$E0,$5F,$67,$FA,$FF,$E8,$5F,$95
	.dc.b	$3F,$A0,$17,$7F,$CD,$35,$FF,$45,$B6,$7F,$D2,$CD,$7F,$CF,$CD,$C7
	.dc.b	$FD,$78,$DC,$3F,$D6,$4D,$CC,$D2,$5F,$F5,$37,$C6,$D3,$F5,$9F,$A3
	.dc.b	$7E,$70,$9A,$FE,$4B,$0F,$F6,$54,$62,$3F,$D4,$2E,$B7,$B1,$FD,$12
	.dc.b	$35,$97,$FA,$BD,$FC,$76,$5F,$EC,$77,$31,$FE,$99,$79,$3B,$9E,$44
	.dc.b	$3A,$D7,$66,$46,$DC,$B4,$FF,$5F,$7D,$E7,$F9,$DC,$8F,$E6,$9A,$FF
	.dc.b	$AE,$9C,$EF,$63,$FD,$27,$1A,$79,$9C,$4C,$9F,$E3,$37,$FC,$5A,$47
	.dc.b	$04,$3F,$C6,$CE,$5A,$64,$FF,$20,$B9,$69,$FE,$A4,$FB,$CF,$C9,$6A
	.dc.b	$9F,$F4,$94,$61,$BB,$7F,$8B,$9A,$FA,$EB,$FD,$AF,$3B,$7A,$47,$F8
	.dc.b	$D3,$EE,$3F,$95,$1F,$E1,$52,$ED,$FC,$BD,$AF,$FB,$81,$B3,$FE,$EB
	.dc.b	$6B,$FE,$D3,$6E,$CD,$99,$27,$31,$86,$C6,$6F,$FC,$51,$1E,$36,$9E
	.dc.b	$32,$9E,$32,$DF,$FB,$F7,$ED,$3F,$9A,$93,$5F,$CD,$61,$E2,$8C,$47
	.dc.b	$FB,$E9,$D6,$D6,$3F,$C3,$B1,$AA,$E9,$1C,$B1,$4E,$67,$FD,$E4,$F2
	.dc.b	$77,$3C,$C8,$F2,$AA,$F0,$8E,$E4,$8C,$E5,$F2,$6D,$F5,$FF,$ED,$12
	.dc.b	$5F,$F3,$46,$A2,$4F,$18,$DE,$F9,$94,$E5,$FF,$78,$CF,$27,$8E,$5F
	.dc.b	$62,$3F,$9C,$4E,$E7,$93,$5F,$F4,$CC,$7F,$A3,$88,$FF,$A5,$DA,$FF
	.dc.b	$A5,$DA,$F0,$34,$43,$45,$3C,$76,$FB,$0F,$D1,$CF,$FB,$CD,$B3,$ED
	.dc.b	$3D,$B7,$F9,$08,$CC,$7A,$23,$1B,$7F,$A5,$2B,$93,$67,$96,$1F,$46
	.dc.b	$FF,$63,$F5,$9F,$E0,$F3,$FC,$F4,$5D,$BF,$C2,$8D,$7B,$64,$8C,$5F
	.dc.b	$A9,$AF,$9A,$35,$3F,$E8,$79,$EA,$D3,$A9,$4E,$AF,$2B,$FC,$33,$72
	.dc.b	$FD,$7B,$6B,$AF,$3F,$0B,$FF,$15,$7D,$4F,$F8,$68,$9B,$FE,$75,$AF
	.dc.b	$A9,$CE,$F6,$3F,$D0,$31,$AC,$FE,$AB,$FC,$FB,$03,$98,$F7,$3C,$9D
	.dc.b	$CF,$22,$1D,$6B,$B3,$23,$6E,$5A,$7E,$10,$FC,$01,$F2,$CB,$9F,$EA
	.dc.b	$5F,$CC,$E7,$FC,$49,$CB,$5F,$BE,$D7,$EC,$9C,$30,$FF,$79,$CF,$38
	.dc.b	$DC,$62,$79,$6F,$C2,$72,$7F,$BC,$08,$7F,$BE,$4A,$7F,$BE,$4A,$7E
	.dc.b	$66,$7F,$DF,$2F,$8C,$C1,$A3,$FD,$F2,$5A,$FE,$5D,$F1,$1F,$E8,$2D
	.dc.b	$FC,$53,$F9,$F8,$C3,$F1,$CF,$7B,$87,$E4,$18,$8F,$E9,$EF,$C6,$8D
	.dc.b	$4F,$C1,$8E,$93,$FE,$25,$97,$D1,$7F,$AD,$1B,$97,$FB,$4D,$B9,$F8
	.dc.b	$E7,$44,$9D,$3F,$E0,$69,$E3,$1B,$32,$49,$C3,$2F,$F8,$BC,$F0,$6D
	.dc.b	$73,$3B,$4E,$E3,$F8,$2C,$F9,$1D,$C7,$63,$58,$7C,$CA,$E5,$1E,$BF
	.dc.b	$9E,$FF,$4A,$59,$FF,$80,$0D,$D9,$BF,$D1,$9C,$9C,$0B,$1A,$C7,$D3
	.dc.b	$18,$9E,$13,$FE,$A3,$AE,$91,$88,$CB,$6B,$8C,$73,$7D,$B5,$CD,$CC
	.dc.b	$70,$69,$C2,$79,$3C,$23,$96,$3F,$8C,$AF,$6F,$CD,$7E,$0C,$95,$FE
	.dc.b	$EC,$C1,$18,$8F,$C8,$3A,$DA,$C7,$E4,$23,$4F,$D8,$7F,$6B,$3C,$A3
	.dc.b	$47,$EF,$03,$97,$F7,$ED,$7F,$A0,$35,$FE,$BC,$D7,$FB,$23,$5E,$71
	.dc.b	$D5,$A7,$78,$D1,$23,$4D,$8F,$D4,$FC,$D7,$E8,$67,$F4,$CD,$8C,$ED
	.dc.b	$FE,$F7,$7D,$C6,$1F,$86,$FE,$D6,$38,$36,$38,$3F,$96,$F9,$6F,$D6
	.dc.b	$13,$5F,$F0,$DE,$08,$C4,$7E,$FD,$D6,$F6,$3F,$65,$1A,$AF,$D3,$BF
	.dc.b	$AF,$30,$F7,$8D,$9D,$A3,$3A,$7F,$0E,$7A,$CF,$29,$D9,$DC,$F0,$55
	.dc.b	$8F,$DA,$FC,$97,$EA,$0F,$D3,$97,$6F,$CF,$35,$FE,$30,$D8,$E3,$1F
	.dc.b	$94,$E2,$FC,$1C,$F2,$8C,$64,$8C,$70,$72,$9E,$A6,$A3,$37,$D2,$35
	.dc.b	$3B,$6C,$64,$F0,$C6,$5F,$E1,$FC,$5F,$C9,$48,$FE,$8D,$AF,$F3,$17
	.dc.b	$3C,$16,$3F,$98,$C6,$9E,$67,$13,$27,$E4,$DF,$C8,$91,$C1,$0F,$D1
	.dc.b	$B9,$69,$93,$F4,$AE,$5A,$7F,$42,$C3,$F9,$D4,$7F,$3F,$6C,$67,$6F
	.dc.b	$C9,$ED,$FD,$3A,$B5,$18,$CF,$F6,$B1,$88,$F7,$57,$F5,$1B,$FE,$8D
	.dc.b	$F1,$5F,$D1,$49,$B7,$E6,$9A,$FF,$57,$73,$B5,$8F,$E9,$51,$AA,$FE
	.dc.b	$7E,$FF,$47,$30,$FF,$3C,$78,$BA,$8E,$24,$3A,$4F,$56,$46,$DC,$B4
	.dc.b	$FE,$9B,$AF,$F4,$03,$F9,$F9,$4F,$E6,$C5,$3F,$B8,$15,$E2,$FE,$58
	.dc.b	$E2,$53,$89,$4E,$25,$78,$EB,$88,$CB,$63,$27,$68,$CB,$4F,$EC,$BF
	.dc.b	$05,$FD,$EC,$8F,$F3,$96,$BF,$E0,$77,$3B,$58,$FF,$03,$C6,$9E,$67
	.dc.b	$13,$27,$F2,$97,$F9,$39,$1C,$10,$FE,$72,$E5,$A6,$4F,$E7,$6E,$5A
	.dc.b	$7F,$86,$3E,$0B,$FC,$21,$1F,$E1,$56,$C6,$74,$FC,$A6,$8A,$D4,$61
	.dc.b	$AF,$B6,$7F,$C4,$D5,$FE,$1B,$F7,$BF,$C3,$44,$CF,$F9,$7B,$5F,$F1
	.dc.b	$63,$9D,$2C,$7F,$87,$E3,$55,$FE,$14,$7F,$C3,$86,$1F,$F0,$8B,$C5
	.dc.b	$D4,$71,$21,$D2,$7A,$B2,$36,$E5,$A7,$F8,$8B,$DA,$FF,$0A,$9F,$E1
	.dc.b	$42,$E9,$FE,$0B,$6B,$FE,$45,$6C,$71,$8F,$E5,$71,$C5,$A7,$12,$9C
	.dc.b	$4B,$1C,$75,$C4,$65,$B1,$93,$B4,$65,$A7,$F8,$EB,$D6,$FF,$28,$11
	.dc.b	$FF,$07,$35,$FF,$2E,$39,$DA,$C7,$F9,$72,$34,$F3,$38,$99,$3F,$BF
	.dc.b	$3F,$DF,$08,$E0,$87,$F8,$39,$CB,$4C,$9F,$E1,$07,$2D,$3F,$CD,$BE
	.dc.b	$B7,$F9,$96,$3F,$CD,$4D,$8C,$EB,$F9,$69,$CB,$63,$2F,$B1,$FF,$3D
	.dc.b	$4F,$F9,$BF,$D2,$FF,$28,$9F,$E5,$02,$E1,$FE,$47,$6B,$FE,$81,$6C
	.dc.b	$FF,$99,$9A,$FF,$98,$9B,$C9,$FE,$8A,$6F,$27,$F8,$C1,$BC,$46,$92
	.dc.b	$FF,$B8,$BE,$36,$9F,$E5,$9D,$BF,$D0,$84,$3F,$C3,$25,$3F,$D2,$C6
	.dc.b	$77,$AF,$FA,$24,$D6,$3F,$E7,$E7,$FD,$0C,$61,$FF,$3D,$BC,$5D,$3C
	.dc.b	$48,$74,$8E,$AC,$8D,$B9,$69,$FE,$8C,$F2,$BF,$D4,$44,$7F,$CC,$0D
	.dc.b	$7F,$D5,$CE,$76,$B1,$FE,$AE,$8D,$3C,$CE,$26,$4F,$F2,$83,$FE,$4E
	.dc.b	$23,$82,$1F,$E6,$07,$2D,$32,$7F,$98,$9C,$B4,$FF,$5F,$F9,$5F,$EB
	.dc.b	$78,$FF,$5E,$36,$33,$B7,$F2,$89,$CB,$63,$2F,$B1,$FF,$66,$4F,$FB
	.dc.b	$07,$C6,$FF,$51,$9F,$EA,$22,$E5,$FE,$99,$6B,$FE,$D3,$6C,$FF,$AE
	.dc.b	$1A,$FF,$AD,$1B,$B7,$FB,$71,$BB,$7F,$A0,$9B,$B1,$A4,$BF,$F2,$1B
	.dc.b	$E3,$69,$FE,$AB,$F0,$BF,$DA,$C4,$D7,$FC,$D6,$D7,$FD,$D0,$E7,$6B
	.dc.b	$1F,$ED,$B8,$D6,$5F,$ED,$07,$FD,$B0,$61,$FF,$66,$BC,$5D,$47,$12
	.dc.b	$1D,$27,$AB,$23,$6E,$5A,$7F,$B7,$8F,$F7,$A1,$0F,$F5,$91,$4F,$F8
	.dc.b	$00,$67,$6A,$FF,$C0,$03,$47,$33,$89,$93,$FD,$44,$7F,$A8,$08,$60
	.dc.b	$87,$FA,$C8,$C9,$4C,$9F,$EB,$43,$25,$04,$19,$B3,$7E,$EC,$DA,$1E
	.dc.b	$AA,$E1,$5B,$9C,$CE,$67,$AC,$E6,$38,$C7,$BC,$E3,$5D,$A3,$31,$E2
	.dc.b	$8F,$F6,$7B,$E5,$39,$BB,$8F,$F0,$6C,$64,$E1,$5E,$27,$6F,$FA,$D7
	.dc.b	$01,$E0,$70,$DE,$BC,$E5,$63,$75,$CA,$78,$C6,$4D,$4D,$9D,$47,$07
	.dc.b	$C7,$3C,$1C,$CE,$DC,$CE,$F1,$F9,$55,$B9,$CD,$6E,$73,$5B,$C7,$EA
	.dc.b	$DE,$E7,$35,$B9,$CD,$6F,$0F,$7E,$1E,$23,$CF,$1F,$0F,$41,$FE,$BD
	.dc.b	$9A,$E2,$7E,$5B,$AE,$4E,$0E,$63,$E8,$B9,$76,$E7,$1F,$C3,$39,$8D
	.dc.b	$B9,$C3,$F1,$EE,$67,$6E,$6F,$C7,$7E,$08,$F0,$B6,$FB,$B5,$EA,$DA
	.dc.b	$E6,$DB,$EF,$5E,$F3,$DF,$5C,$8F,$4B,$F8,$7E,$4F,$C7,$5F,$E7,$DB
	.dc.b	$5F,$A6,$7F,$57,$AE,$EF,$F3,$15,$B6,$D6,$DB,$8A,$F8,$C7,$E6,$1E
	.dc.b	$B1,$98,$DC,$60,$DB,$FA,$07,$0F,$E8,$CE,$C7,$E9,$63,$93,$63,$4E
	.dc.b	$8F,$D7,$46,$9B,$1A,$8C,$47,$F0,$76,$BF,$C2,$9A,$F3,$35,$1C,$4F
	.dc.b	$5F,$27,$EF,$AF,$E2,$4F,$EC,$E3,$17,$C6,$7F,$86,$46,$2F,$C3,$18
	.dc.b	$DB,$FD,$C1,$18,$AF,$E2,$F5,$BC,$3B,$D6,$B8,$3F,$19,$5F,$C7,$1F
	.dc.b	$BF,$8F,$F0,$87,$38,$7F,$5B,$CF,$F2,$11,$83,$79,$7F,$1F,$71,$1A
	.dc.b	$3F,$50,$EA,$FB,$46,$0D,$69,$FC,$46,$75,$7D,$E3,$13,$AC,$3F,$8C
	.dc.b	$D6,$8E,$F1,$A9,$DB,$67,$6F,$E9,$DF,$B0,$43,$F0,$95,$F8,$07,$F9
	.dc.b	$E9,$93,$F0,$C5,$3F,$06,$53,$F0,$05,$3F,$A4,$FC,$7F,$ED,$1F,$E6
	.dc.b	$AF,$F0,$89,$FE,$7A,$E6,$3F,$9E,$39,$77,$1F,$52,$31,$1A,$8C,$1F
	.dc.b	$2E,$39,$63,$FD,$55,$DB,$63,$67,$F0,$B8,$DE,$27,$6D,$8D,$9E,$62
	.dc.b	$D7,$91,$61,$97,$A1,$A8,$F3,$3A,$78,$3F,$03,$3F,$D6,$DF,$E9,$08
	.dc.b	$E0,$93,$FB,$78,$CB,$63,$2F,$F5,$D8,$CB,$4E,$A4,$39,$14,$D1,$82
	.dc.b	$18,$21,$FD,$B5,$FC,$21,$F8,$5F,$C7,$FF,$76,$49,$FE,$0E,$D9,$FE
	.dc.b	$1C,$D7,$FB,$93,$5F,$EF,$0D,$7F,$C1,$CF,$47,$99,$0E,$65,$39,$94
	.dc.b	$E6,$53,$99,$4F,$F0,$DB,$F9,$DF,$DD,$BF,$89,$92,$7F,$8F,$B6,$7F
	.dc.b	$93,$35,$FF,$05,$35,$FF,$08,$35,$FF,$13,$3D,$1E,$64,$39,$94,$E6
	.dc.b	$53,$99,$4E,$65,$3F,$C6,$6F,$ED,$7E,$47,$F9,$59,$27,$F9,$BB,$67
	.dc.b	$F9,$F3,$5F,$F1,$13,$5F,$F1,$43,$5F,$F1,$F3,$D1,$E6,$43,$99,$4E
	.dc.b	$65,$39,$94,$E6,$53,$FC,$96,$FF,$0D,$FD,$DB,$FA,$19,$27,$FA,$7B
	.dc.b	$67,$FA,$B3,$5F,$F1,$D3,$5F,$F2,$03,$5F,$F2,$B3,$D1,$E6,$43,$99
	.dc.b	$4E,$65,$39,$94,$E6,$53,$FC,$C6,$FF,$25,$F9,$2F,$EB,$64,$9F,$EC
	.dc.b	$ED,$9F,$ED,$CD,$7F,$CA,$4D,$7F,$CB,$0D,$7F,$CD,$CF,$47,$99,$0E
	.dc.b	$65,$39,$94,$E6,$53,$99,$4F,$F3,$DB,$FC,$F7,$E4,$BF,$B9,$92,$7F
	.dc.b	$BF,$B6,$7F,$C0,$CD,$7F,$CD,$4D,$7F,$CE,$0D,$7F,$D0,$CF,$47,$99
	.dc.b	$0E,$65,$39,$94,$E6,$53,$99,$4F,$F4,$9B,$FD,$57,$F7,$6F,$F0,$59
	.dc.b	$27,$FC,$2E,$D9,$FF,$0F,$35,$FF,$41,$35,$FF,$44,$35,$FF,$4F,$3D
	.dc.b	$1E,$64,$39,$94,$E6,$53,$99,$4E,$65,$3F,$D5,$6F,$F6,$DF,$07,$FC
	.dc.b	$46,$49,$FF,$17,$B6,$7F,$C6,$CD,$7F,$D3,$4D,$7F,$D4,$0D,$7F,$D6
	.dc.b	$CF,$47,$99,$0E,$65,$39,$94,$E6,$53,$99,$4F,$F6,$1B,$FE,$05,$FC
	.dc.b	$B7,$FC,$76,$49,$FF,$23,$B6,$7F,$C9,$CD,$7F,$D6,$4D,$7F,$D7,$0D
	.dc.b	$7F,$D9,$CF,$47,$99,$0E,$65,$39,$94,$E6,$53,$99,$4F,$F6,$DB,$FE
	.dc.b	$1D,$FD,$DB,$FC,$A6,$49,$FF,$2F,$B6,$7F,$CC,$CD,$7F,$D9,$4D,$7F
	.dc.b	$DA,$0D,$7F,$DC,$CF,$47,$99,$0E,$65,$39,$94,$E6,$53,$99,$4F,$F7
	.dc.b	$9B,$FE,$35,$F9,$5F,$F3,$59,$27,$FC,$EE,$D9,$FF,$3F,$35,$FF,$71
	.dc.b	$35,$FF,$74,$35,$FF,$7F,$3D,$1E,$64,$39,$94,$E6,$53,$99,$4E,$65
	.dc.b	$01,$FF,$27,$7E,$ED,$FE,$84,$24,$FF,$A4,$1B,$3F,$E9,$76,$BF,$EF
	.dc.b	$B6,$BF,$F0,$05,$AF,$60,$F5,$7A,$10,$E8,$53,$A1,$4E,$85,$3A,$3E
	.dc.b	$87,$07,$B3,$E5,$BF,$D3,$A4,$9F,$F5,$4B,$63,$FD,$60,$D7,$CC,$71
	.dc.b	$8D,$91,$DB,$5D,$88,$FF,$3B,$11,$EE,$43,$B9,$4E,$2E,$DC,$90,$C9
	.dc.b	$FE,$C0,$F9,$6F,$F5,$B9,$27,$04,$8F,$F6,$54,$65,$B1,$97,$CF,$19
	.dc.b	$6C,$75,$61,$C8,$A7,$02,$9F,$1A,$3F,$D2,$84,$72,$7D,$F7,$0F,$FB
	.dc.b	$77,$F1,$DF,$F6,$79,$27,$FD,$B6,$D8,$D4,$60,$8E,$08,$7D,$08,$E0
	.dc.b	$D3,$99,$0F,$01,$4D,$18,$21,$82,$1F,$82,$7F,$D6,$07,$FA,$C7,$E5
	.dc.b	$FF,$DC,$E4,$9F,$F7,$9B,$67,$FD,$F4,$D7,$F0,$AD,$7F,$1A,$D7,$F4
	.dc.b	$8F,$47,$99,$0E,$65,$39,$94,$E6,$53,$99,$4F,$DE,$BF,$EC,$9F,$CD
	.dc.b	$FF,$DF,$E4,$9F,$CC,$CF,$F0,$28,$C3,$86,$BF,$9E,$6B,$FA,$66,$BF
	.dc.b	$BD,$7A,$BD,$03,$E1,$70,$D7,$B1,$82,$18,$21,$E1,$32,$7E,$C3,$9D
	.dc.b	$3D,$DB,$1E,$C6,$BB,$8D,$11,$D3,$5D,$18,$21,$83,$C0,$77,$21,$DC
	.dc.b	$A6,$8C,$10,$C1,$0F,$E1,$F1,$FE,$FD,$3F,$DF,$DC,$FE,$88,$FE,$26
	.dc.b	$D7,$F9,$53,$5F,$E7,$2D,$4D,$24,$FD,$36,$C7,$E0,$1A,$FF,$1A,$78
	.dc.b	$3B,$21,$B2,$9B,$29,$E6,$29,$E6,$29,$E6,$3F,$09,$F8,$8F,$9C,$93
	.dc.b	$82,$47,$E5,$23,$2D,$8C,$BF,$CC,$23,$2D,$8E,$AC,$39,$14,$E0,$53
	.dc.b	$F9,$FC,$7E,$5D,$34,$8E,$CF,$EB,$4E,$FE,$67,$E8,$12,$7F,$40,$D9
	.dc.b	$FD,$6B,$5F,$E8,$ED,$7F,$A7,$B5,$FE,$CC,$F4,$79,$90,$E6,$53,$99
	.dc.b	$4E,$65,$39,$94,$FE,$FE,$FB,$FE,$67,$F1,$04,$9F,$DC,$36,$7F,$82
	.dc.b	$B5,$FE,$BE,$D7,$FB,$3B,$5F,$EF,$CF,$47,$99,$0E,$65,$39,$94,$E6
	.dc.b	$53,$99,$4F,$F0,$BB,$F5,$FF,$76,$FD,$01,$27,$F8,$83,$67,$F8,$EB
	.dc.b	$5F,$EE,$ED,$7F,$BF,$B5,$FF,$0B,$3D,$1E,$64,$39,$94,$E6,$53,$99
	.dc.b	$4E,$65,$3F,$C5,$EF,$E3,$FE,$6B,$F7,$04,$9F,$E5,$0D,$9F,$E6,$AD
	.dc.b	$7F,$C1,$ED,$7F,$C2,$ED,$7F,$C5,$CF,$47,$99,$0E,$65,$39,$94,$E6
	.dc.b	$53,$99,$4F,$F2,$3B,$FA,$7F,$CA,$3F,$88,$12,$7F,$A0,$36,$7F,$A6
	.dc.b	$B5,$FF,$13,$B5,$FF,$17,$B5,$FF,$23,$3D,$1E,$64,$39,$94,$E6,$53
	.dc.b	$99,$4E,$65,$3F,$CB,$EF,$EF,$FF,$76,$FE,$50,$49,$FE,$B0,$D9,$FE
	.dc.b	$CA,$D7,$FC,$7E,$D7,$FC,$8E,$D7,$FC,$BC,$F4,$79,$90,$E6,$53,$99
	.dc.b	$4E,$65,$39,$94,$FF,$3B,$BF,$C5,$FE,$73,$FA,$01,$27,$FB,$83,$67
	.dc.b	$FB,$EB,$5F,$F2,$BB,$5F,$F2,$FB,$5F,$F3,$B3,$D1,$E6,$43,$99,$4E
	.dc.b	$65,$39,$94,$E6,$53,$FD,$1E,$FF,$2F,$FD,$DB,$FA,$C1,$27,$FC,$10
	.dc.b	$D9,$FF,$0A,$B5,$FF,$37,$B5,$FF,$3B,$B5,$FF,$47,$3D,$1E,$64,$39
	.dc.b	$94,$E6,$53,$99,$4E,$65,$3F,$D4,$EF,$F4,$7F,$4F,$FB,$81,$27,$FC
	.dc.b	$40,$D9,$FF,$16,$B5,$FF,$43,$B5,$FF,$47,$B5,$FF,$53,$3D,$1E,$64
	.dc.b	$39,$94,$E6,$53,$99,$4E,$65,$3F,$D7,$EF,$F5,$FF,$4F,$FC,$10,$49
	.dc.b	$FF,$1C,$36,$7F,$C8,$AD,$7F,$D3,$ED,$7F,$D4,$ED,$7F,$D7,$CF,$F8
	.dc.b	$4F,$F1,$1F,$F6,$07,$95,$FE,$A6,$6C,$FF,$AB,$9A,$FF,$B1,$5A,$FF
	.dc.b	$B5,$1A,$9A,$47,$FD,$B3,$1D,$5A,$75,$29,$D4,$A7,$52,$9D,$4F,$F2
	.dc.b	$C4,$7F,$80,$C8,$E0,$87,$F9,$A1,$CB,$4C,$9E,$33,$07,$F9,$BA,$9E
	.dc.b	$4C,$3F,$D9,$2D,$69,$98,$FF,$69,$C6,$5B,$19,$7F,$DD,$91,$96,$C6
	.dc.b	$5F,$F7,$EC,$65,$B1,$97,$FC,$D2,$D8,$D4,$62,$3C,$EF,$8A,$3E,$39
	.dc.b	$FE,$6E,$AF,$4B,$82,$42,$32,$D3,$91,$4F,$4C,$70,$8D,$92,$36,$F2
	.dc.b	$61,$C8,$A7,$23,$E7,$47,$FA,$44,$FF,$32,$10,$C1,$1D,$90,$13,$DC
	.dc.b	$7E,$53,$FE,$8E,$F3,$3E,$84,$62,$BF,$CE,$8D,$8C,$CF,$FA,$52,$79
	.dc.b	$D4,$AD,$61,$F3,$21,$1C,$E7,$71,$E1,$AF,$05,$6E,$B5,$3A,$79,$C7
	.dc.b	$59,$F0,$4E,$5E,$31,$9A,$F3,$C6,$27,$B4,$60,$5F,$A9,$C3,$5F,$2C
	.dc.b	$7A,$27,$93,$63,$75,$E7,$9F,$1C,$F8,$6B,$A1,$27,$55,$C1,$D1,$88
	.dc.b	$F0,$4F,$AE,$7C,$B3,$92,$1E,$D7,$CB,$7F,$12,$7D,$B5,$DE,$B5,$1C
	.dc.b	$E3,$0E,$A3,$C7,$5C,$AF,$DB,$7C,$EB,$DD,$7E,$47,$CF,$71,$FC,$4F
	.dc.b	$93,$FB,$09,$FD,$B3,$67,$F0,$0D,$7F,$0A,$D9,$E2,$7C,$C7,$09,$A4
	.dc.b	$74,$4C,$3E,$2C,$F4,$8F,$C4,$FE,$1D,$67,$E4,$B6,$FF,$89,$B6,$8D
	.dc.b	$25,$79,$2B,$8B,$6B,$8B,$6B,$8B,$6B,$8B,$E9,$3C,$67,$3F,$53,$F9
	.dc.b	$AE,$1F,$C5,$9B,$1F,$EF,$C6,$E0,$69,$23,$F8,$54,$F1,$69,$C4,$B3
	.dc.b	$C5,$B3,$C7,$3F,$88,$F3,$8A,$7E,$CA,$BA,$38,$21,$E9,$34,$77,$32
	.dc.b	$7B,$EB,$C0,$E2,$39,$BC,$4F,$A9,$EA,$7E,$19,$4A,$30,$DC,$7F,$13
	.dc.b	$3A,$9C,$92,$33,$3B,$7A,$98,$8D,$38,$24,$FD,$D3,$F9,$0F,$9B,$F8
	.dc.b	$76,$99,$0E,$5A,$65,$C1,$1C,$10,$F8,$87,$42,$9C,$8A,$F5,$C6,$47
	.dc.b	$85,$AF,$06,$E1,$FD,$7A,$3F,$0E,$43,$F0,$25,$8D,$BF,$38,$FF,$02
	.dc.b	$E9,$FC,$ED,$FE,$01,$8F,$F8,$52,$31,$1F,$C4,$4D,$EB,$C5,$C1,$FD
	.dc.b	$6E,$36,$FF,$2F,$37,$3F,$CA,$1F,$E4,$C6,$0E,$E5,$39,$D7,$16,$9F
	.dc.b	$CD,$8D,$9D,$EB,$C4,$EE,$38,$34,$E0,$79,$8C,$98,$6B,$E4,$34,$64
	.dc.b	$E2,$6A,$BD,$AE,$4E,$06,$67,$C6,$E6,$36,$E8,$E6,$66,$36,$E4,$91
	.dc.b	$F4,$27,$A5,$7C,$98,$E6,$66,$79,$9C,$2B,$9C,$66,$31,$3C,$67,$27
	.dc.b	$D9,$AE,$B3,$F3,$0F,$90,$70,$30,$79,$E7,$91,$1F,$A7,$1C,$9F,$92
	.dc.b	$7C,$B7,$0E,$5A,$FA,$2B,$CC,$7B,$2B,$CB,$5D,$67,$EF,$35,$E4,$71
	.dc.b	$8D,$BE,$23,$11,$B7,$D0,$7A,$8F,$61,$D0,$C4,$F8,$63,$F4,$13,$D2
	.dc.b	$7D,$73,$E6,$8C,$47,$88,$E6,$58,$E4,$F4,$8F,$9C,$F9,$0C,$4F,$63
	.dc.b	$CC,$F9,$2B,$E2,$B4,$E0,$78,$A7,$9B,$E0,$33,$1C,$2B,$C3,$1E,$D3
	.dc.b	$85,$7C,$B9,$D5,$7B,$5B,$19,$9E,$E7,$AE,$7D,$EF,$63,$25,$3C,$47
	.dc.b	$9C,$B5,$E2,$AE,$B1,$F5,$4C,$CF,$85,$E6,$71,$9F,$4B,$E1,$AF,$8B
	.dc.b	$1C,$E3,$6D,$79,$C7,$A5,$A7,$78,$E6,$F9,$DA,$F7,$9F,$6B,$67,$BC
	.dc.b	$79,$63,$99,$FC,$48,$A7,$E9,$A7,$99,$E3,$2C,$71,$6C,$F9,$5B,$3E
	.dc.b	$59,$F0,$C7,$16,$9C,$4B,$3C,$5B,$3C,$63,$CE,$64,$AE,$4F,$C7,$94
	.dc.b	$FC,$19,$4F,$AA,$7E,$1C,$B5,$87,$DD,$3E,$07,$E1,$57,$A9,$B1,$97
	.dc.b	$75,$E2,$72,$D7,$27,$90,$C9,$4C,$9E,$63,$25,$32,$79,$8F,$C1,$C7
	.dc.b	$C7,$8E,$D1,$FC,$00,$E0,$53,$F9,$F9,$FD,$9C,$FF,$6A,$1F,$EA,$72
	.dc.b	$18,$21,$FE,$D6,$32,$53,$27,$FB,$58,$FE,$BC,$42,$10,$84,$21,$08
	.dc.b	$43,$FD,$34,$7F,$39,$21,$FD,$C8,$A6,$8F,$F3,$C1,$FC,$68,$C9,$FE
	.dc.b	$A5,$37,$55,$D9,$0E,$47,$F6,$D3,$89,$FE,$B3,$38,$94,$E0,$73,$38
	.dc.b	$B5,$86,$CC,$94,$C9,$FE,$AB,$3A,$94,$EA,$6C,$E4,$F0,$AF,$F6,$AC
	.dc.b	$79,$8E,$07,$A4,$E2,$60,$F3,$47,$1A,$FA,$F3,$CA,$7E,$14,$E5,$E9
	.dc.b	$19,$7A,$1F,$18,$D1,$F4,$CE,$EE,$DE,$E7,$A4,$F1,$15,$E9,$1C,$A3
	.dc.b	$A1,$EE,$35,$3E,$62,$3F,$74,$EE,$E1,$F1,$1C,$8E,$E7,$C4,$8E,$CF
	.dc.b	$76,$BD,$8E,$87,$19,$F5,$3E,$33,$9B,$86,$CF,$FB,$E2,$79,$C6,$0E
	.dc.b	$66,$27,$DE,$F7,$3D,$26,$8C,$1F,$86,$39,$1A,$3D,$A7,$39,$F0,$3C
	.dc.b	$0F,$61,$4E,$73,$B7,$F1,$B3,$E9,$24,$FC,$98,$FB,$A6,$CE,$E7,$33
	.dc.b	$E5,$16,$38,$BE,$93,$CE,$7A,$5C,$BE,$93,$A9,$C6,$3D,$2F,$03,$D2
	.dc.b	$53,$D2,$6C,$F4,$BE,$77,$71,$D9,$FB,$11,$D2,$31,$1A,$8E,$0F,$43
	.dc.b	$B1,$E8,$9F,$80,$FA,$63,$6E,$4F,$11,$FA,$62,$9B,$3F,$4E,$FC,$B8
	.dc.b	$E7,$1F,$A2,$3B,$CF,$9E,$33,$1F,$8D,$8F,$3B,$E5,$6B,$E1,$38,$CF
	.dc.b	$57,$C2,$75,$8D,$9D,$5C,$9D,$63,$D7,$19,$8E,$67,$CC,$3D,$25,$3E
	.dc.b	$C1,$D6,$38,$9E,$17,$89,$F0,$CA,$7C,$33,$89,$E1,$3A,$9B,$8F,$3C
	.dc.b	$73,$6C,$78,$4F,$B6,$53,$EA,$CF,$C8,$7F,$26,$78,$0F,$31,$D8,$F1
	.dc.b	$1C,$8F,$04,$FA,$9E,$87,$DA,$8E,$2F,$E5,$DA,$F1,$3F,$90,$BE,$27
	.dc.b	$D8,$6C,$A6,$DE,$2F,$E4,$41,$1A,$AE,$B9,$36,$E1,$47,$08,$C3,$63
	.dc.b	$F5,$71,$87,$89,$82,$1E,$08,$E0,$D3,$65,$3C,$87,$03,$FD,$F1,$1E
	.dc.b	$3C,$38,$C6,$0F,$4B,$87,$FD,$F9,$1B,$7A,$9B,$3E,$0C,$6C,$8E,$8E
	.dc.b	$27,$43,$CB,$93,$29,$18,$6B,$E8,$11,$B8,$F7,$B6,$38,$46,$0E,$35
	.dc.b	$F5,$27,$53,$CD,$B1,$99,$DC,$FD,$C9,$CB,$63,$2F,$89,$B9,$FD,$BA
	.dc.b	$FC,$0B,$67,$F0,$AD,$4D,$23,$F8,$68,$E2,$D3,$89,$4E,$25,$8E,$2F
	.dc.b	$E0,$1E,$78,$58,$FC,$69,$F8,$28,$FC,$93,$F2,$DB,$3E,$09,$E0,$49
	.dc.b	$E0,$72,$38,$16,$38,$3A,$8E,$0F,$D7,$D7,$EE,$6B,$F9,$46,$C7,$E5
	.dc.b	$1A,$EF,$04,$ED,$B1,$B6,$BB,$8D,$11,$D3,$5D,$1C,$63,$A1,$1C,$10
	.dc.b	$F1,$39,$69,$9B,$FD,$ED,$7E,$12,$0D,$24,$FE,$63,$4F,$F1,$DC,$7E
	.dc.b	$41,$B1,$9D,$3F,$CF,$15,$FE,$53,$AD,$47,$F0,$C6,$C7,$69,$E0,$B5
	.dc.b	$FF,$32,$CE,$89,$3A,$8C,$BC,$23,$2F,$F9,$A6,$32,$FF,$9A,$E3,$33
	.dc.b	$FE,$65,$9C,$BE,$59,$E4,$75,$7F,$CF,$71,$A7,$FC,$C1,$3E,$88,$FF
	.dc.b	$3C,$99,$BF,$83,$5E,$1A,$E2,$7F,$A1,$2B,$C2,$FF,$9A,$6B,$8B,$99
	.dc.b	$F0,$9C,$58,$79,$5F,$51,$E5,$72,$79,$1E,$2F,$B4,$F0,$9F,$EA,$03
	.dc.b	$F9,$CC,$7F,$A7,$C8,$FF,$A9,$8F,$F5,$21,$B3,$EC,$98,$3F,$D5,$44
	.dc.b	$3F,$D5,$59,$7F,$AD,$08,$FF,$AC,$5A,$E8,$C1,$0C,$10,$F2,$9F,$B3
	.dc.b	$2B,$FD,$2D,$FE,$49,$34,$FE,$3E,$FF,$54,$6B,$FD,$61,$AF,$03,$FA
	.dc.b	$A3,$82,$1F,$CC,$9E,$4D,$79,$35,$E4,$7F,$60,$7B,$C7,$FB,$24,$91
	.dc.b	$FE,$CC,$6B,$E9,$6B,$FD,$7C,$C1,$B2,$18,$23,$B2,$3F,$E5,$26,$BF
	.dc.b	$D8,$5A,$FF,$22,$6B,$FC,$75,$A7,$F6,$77,$FB,$EF,$C1,$7E,$BD,$F9
	.dc.b	$CD,$7F,$BD,$35,$FE,$FE,$D7,$FC,$06,$D7,$9D,$7F,$BB,$23,$10,$69
	.dc.b	$23,$44,$8F,$BD,$F0,$7F,$EF,$32,$3F,$EF,$66,$BA,$30,$43,$04,$3F
	.dc.b	$C2,$4F,$06,$C7,$32,$3E,$06,$BA,$30,$43,$04,$3F,$C4,$31,$EE,$3D
	.dc.b	$FF,$8A,$7E,$5E,$3F,$C5,$2D,$4D,$23,$B6,$BF,$9F,$6B,$FE,$2B,$6B
	.dc.b	$FE,$30,$79,$4F,$12,$47,$16,$BC,$5A,$F1,$6B,$FE,$3C,$8F,$4C,$62
	.dc.b	$3F,$C8,$3F,$A8,$FA,$48,$FB,$5A,$FE,$D5,$AF,$F9,$05,$AF,$F9,$19
	.dc.b	$AF,$F9,$29,$E9,$1C,$C9,$1C,$DA,$F3,$6B,$CD,$AF,$36,$BE,$87,$E9
	.dc.b	$7E,$A3,$F4,$49,$18,$21,$FC,$35,$CB,$4C,$9F,$E5,$B7,$2D,$3A,$C4
	.dc.b	$79,$35,$E0,$D7,$FC,$CF,$1F,$A4,$24,$66,$3F,$CD,$B1,$87,$F9,$07
	.dc.b	$E7,$9F,$89,$23,$F8,$D6,$BA,$30,$43,$04,$3F,$CE,$8F,$06,$CF,$32
	.dc.b	$3E,$06,$BA,$30,$43,$04,$3F,$D0,$71,$FC,$00,$FE,$03,$F1,$1F,$A1
	.dc.b	$24,$7E,$95,$AF,$F3,$D6,$BF,$E8,$96,$BF,$E8,$E6,$BF,$E9,$27,$A4
	.dc.b	$F3,$24,$73,$6B,$CD,$AF,$36,$BC,$DA,$FF,$A8,$23,$F8,$CF,$EA,$3F
	.dc.b	$B9,$24,$7E,$F5,$AF,$F5,$56,$BF,$EA,$16,$BF,$EA,$66,$BF,$EA,$A7
	.dc.b	$A4,$F3,$24,$73,$6B,$CD,$AF,$36,$BC,$DA,$FF,$AE,$23,$F9,$8F,$EA
	.dc.b	$3F,$C4,$49,$1F,$C5,$5A,$FF,$6D,$6B,$FE,$B9,$6B,$FE,$BE,$6B,$FE
	.dc.b	$C2,$7A,$4F,$32,$47,$36,$BC,$DA,$F3,$6B,$CD,$AF,$FB,$42,$3F,$A4
	.dc.b	$F0,$FF,$29,$24,$7F,$2D,$6B,$FE,$05,$6B,$FE,$D1,$6B,$FE,$D6,$6B
	.dc.b	$FE,$DA,$7A,$4F,$32,$47,$36,$BC,$DA,$F3,$6B,$CD,$AF,$FB,$A2,$3F
	.dc.b	$B0,$FE,$A3,$FD,$04,$91,$FD,$15,$AF,$F8,$75,$AF,$FB,$A5,$AF,$FB
	.dc.b	$B9,$AF,$FB,$C9,$E9,$3C,$C9,$1C,$DA,$F3,$6B,$CD,$AF,$36,$BF,$F0
	.dc.b	$02,$3F,$BC,$FB,$3F,$D6,$49,$1F,$D7,$5A,$FF,$8D,$5A,$FF,$C0,$16
	.dc.b	$A2,$07,$2E,$1E,$D3,$D4,$91,$D5,$AF,$56,$BD,$0E,$8E,$1E,$8D,$7C
	.dc.b	$51,$FE,$14,$FD,$47,$FB,$A1,$23,$FB,$C3,$5F,$F2,$83,$5E,$C1,$C3
	.dc.b	$E1,$6B,$EE,$7A,$CF,$42,$47,$46,$BD,$23,$C9,$18,$8E,$8D,$7A,$35
	.dc.b	$F9,$9F,$9B,$FF,$83,$48,$FF,$84,$5A,$FF,$9A,$5A,$FA,$9A,$FC,$96
	.dc.b	$BF,$59,$E9,$1C,$C9,$1C,$DA,$F3,$6B,$CD,$AF,$36,$BE,$87,$FC,$4D
	.dc.b	$FA,$8F,$F8,$94,$91,$82,$1F,$E8,$17,$2D,$32,$7D,$47,$2D,$3A,$C4
	.dc.b	$79,$35,$E0,$D7,$F1,$51,$FE,$58,$24,$66,$3F,$33,$18,$7F,$D2,$9F
	.dc.b	$A8,$FF,$8F,$48,$FF,$90,$5A,$E8,$C1,$0C,$10,$FC,$6B,$C1,$B3,$CC
	.dc.b	$8F,$81,$AE,$8C,$10,$C1,$0F,$D3,$47,$F9,$D0,$FF,$3A,$FC,$8F,$F9
	.dc.b	$54,$91,$FE,$59,$6B,$FE,$B1,$6B,$FA,$26,$BF,$B1,$6B,$FB,$E7,$A4
	.dc.b	$F3,$24,$73,$6B,$CD,$AF,$36,$BC,$DA,$FF,$0B,$8F,$F4,$6F,$C8,$FF
	.dc.b	$9B,$49,$1F,$E7,$16,$BF,$EC,$96,$BF,$BA,$6B,$FC,$25,$AF,$F1,$67
	.dc.b	$A4,$F3,$24,$73,$6B,$CD,$AF,$36,$BC,$DA,$FF,$23,$8F,$F5,$2F,$EA
	.dc.b	$3F,$E8,$52,$47,$FA,$25,$AF,$FB,$85,$AF,$F1,$26,$BF,$C8,$5A,$FF
	.dc.b	$2E,$7A,$4F,$32,$47,$36,$BC,$DA,$F3,$6B,$CD,$AF,$F3,$B8,$FF,$5E
	.dc.b	$FE,$A3,$FE,$9D,$24,$7F,$A8,$5A,$FF,$BE,$5A,$FF,$2A,$6B,$FC,$E5
	.dc.b	$AF,$F4,$67,$A4,$F3,$24,$73,$6B,$CD,$AF,$36,$BC,$DA,$FF,$53,$8F
	.dc.b	$F6,$AF,$EA,$3F,$EB,$52,$47,$FA,$E5,$AF,$F4,$00,$E1,$FE,$86,$D7
	.dc.b	$FA,$93,$5F,$EB,$EF,$59,$E8,$48,$E9,$1E,$28,$C4,$74,$6B,$D1,$AF
	.dc.b	$46,$BF,$DB,$7F,$36,$FF,$67,$11,$FF,$69,$35,$F5,$35,$FE,$B6,$D7
	.dc.b	$FB,$53,$5F,$EE,$EF,$48,$E6,$48,$E6,$D7,$9B,$5E,$6D,$79,$B5,$F4
	.dc.b	$3F,$EE,$7F,$96,$FF,$73,$12,$30,$43,$E9,$39,$69,$93,$FB,$AB,$96
	.dc.b	$9D,$62,$3C,$9A,$F0,$6B,$FE,$0B,$84,$62,$3F,$C3,$31,$87,$91,$0E
	.dc.b	$47,$E1,$FF,$51,$FF,$80,$04,$7F,$14,$1C,$3B,$32,$E1,$FF,$07,$B5
	.dc.b	$FF,$0F,$35,$FF,$15,$BE,$37,$73,$C8,$8F,$8D,$AE,$8C,$10,$C1,$0E
	.dc.b	$EF,$A3,$F5,$1F,$4C,$1A,$47,$44,$3F,$50,$E5,$A6,$4F,$F1,$5B,$96
	.dc.b	$C7,$72,$3C,$9A,$E8,$C1,$0C,$10,$FF,$22,$47,$C5,$6C,$7D,$2F,$CF
	.dc.b	$BD,$64,$8C,$10,$FE,$00,$E5,$A6,$4F,$F2,$1B,$96,$9D,$62,$3C,$9A
	.dc.b	$F0,$6B,$FE,$59,$8F,$33,$FE,$66,$6B,$FE,$69,$6A,$69,$23,$F1,$BF
	.dc.b	$A8,$FD,$62,$4E,$08,$7F,$1B,$72,$D3,$27,$F9,$79,$CB,$4E,$B1,$1E
	.dc.b	$4D,$78,$35,$FF,$3A,$47,$A0,$91,$98,$FF,$41,$C6,$1F,$E5,$DF,$37
	.dc.b	$F8,$E2,$3F,$A8,$6B,$A3,$04,$30,$43,$FC,$EE,$F0,$6C,$F3,$23,$E0
	.dc.b	$6B,$A3,$04,$30,$43,$FD,$21,$1F,$50,$FA,$BF,$A8,$FE,$98,$91,$FC
	.dc.b	$01,$AF,$F4,$B6,$BF,$E8,$C6,$BF,$E9,$B6,$BF,$EA,$87,$A4,$F3,$24
	.dc.b	$73,$6B,$CD,$AF,$36,$BC,$DA,$FF,$AD,$63,$F1,$BF,$A4,$FE,$F8,$91
	.dc.b	$FC,$61,$AF,$F6,$36,$BF,$EA,$46,$BF,$EB,$36,$BF,$EC,$07,$A4,$F3
	.dc.b	$24,$73,$6B,$CD,$AF,$36,$BC,$DA,$FF,$B3,$63,$F4,$BF,$A8,$FF,$16
	.dc.b	$24,$7F,$30,$6B,$FD,$ED,$AF,$FA,$F1,$AF,$FB,$2D,$AF,$FB,$61,$E9
	.dc.b	$3C,$C9,$1C,$DA,$F3,$6B,$CD,$AF,$36,$BF,$EE,$58,$F4,$FB,$01,$EC
	.dc.b	$FF,$BA,$3C,$D7,$94,$AD,$56,$EF,$FD,$DD,$5A,$9C,$CF,$1A,$E0,$F5
	.dc.b	$BC,$4F,$26,$C7,$4B,$E7,$5C,$23,$B4,$2F,$B4,$73,$8D,$DF,$36,$C7
	.dc.b	$37,$85,$73,$8E,$E7,$BA,$36,$D7,$73,$EA,$73,$1C,$2E,$46,$08,$70
	.dc.b	$8F,$23,$5E,$97,$E8,$7D,$6D,$79,$C7,$9B,$F5,$1F,$EB,$E4,$7F,$BC
	.dc.b	$B5,$FF,$22,$35,$10,$AE,$5C,$07,$0E,$DE,$FD,$9E,$8A,$FB,$F0,$8F
	.dc.b	$95,$18,$6E,$3C,$DE,$0D,$BF,$7B,$5F,$C1,$57,$37,$E0,$C6,$F0,$CD
	.dc.b	$72,$AD,$56,$0D,$BC,$67,$06,$62,$61,$F8,$89,$D3,$6B,$A1,$E8,$8E
	.dc.b	$B1,$99,$F3,$C7,$08,$FA,$53,$B2,$4E,$B2,$F9,$73,$F5,$FC,$A7,$95
	.dc.b	$23,$C5,$18,$8F,$83,$3C,$33,$F9,$AD,$CF,$EA,$56,$EF,$8D,$F3,$7F
	.dc.b	$09,$5B,$CB,$FB,$BE,$1F,$5A,$F8,$57,$8E,$F8,$4F,$68,$FC,$0B,$5F
	.dc.b	$4D,$7F,$7C,$25,$FE,$38,$91,$F8,$E6,$BC,$1A,$FE,$32,$38,$BB,$3B
	.dc.b	$E1,$F8,$F9,$F4,$12,$1E,$57,$C4,$77,$5E,$CA,$E1,$18,$AF,$CE,$D6
	.dc.b	$E2,$1F,$7E,$3A,$47,$E3,$B2,$FC,$51,$F7,$C8,$7D,$E3,$27,$E3,$0C
	.dc.b	$E9,$C5,$D1,$92,$39,$34,$CA,$FC,$83,$D2,$31,$1F,$BB,$8E,$AE,$9E
	.dc.b	$A6,$B1,$FD,$7B,$FA,$1F,$90,$7C,$7C,$AF,$CE,$46,$23,$F5,$6D,$8C
	.dc.b	$C3,$E4,$32,$F1,$8F,$E1,$93,$B6,$C6,$6F,$F7,$F3,$E0,$AE,$33,$2B
	.dc.b	$AB,$D2,$31,$1C,$A2,$D7,$27,$C5,$1C,$08,$6C,$F3,$CF,$0C,$FF,$8A
	.dc.b	$5F,$96,$F9,$5F,$7B,$FE,$87,$E8,$FF,$B8,$E7,$F8,$AE,$DF,$A0,$6B
	.dc.b	$FA,$0C,$13,$B6,$D7,$EA,$67,$84,$7F,$23,$6C,$EC,$E1,$B7,$EE,$63
	.dc.b	$F8,$C9,$AA,$FD,$C4,$FF,$A5,$08,$FF,$AB,$9A,$FF,$BE,$5A,$FF,$08
	.dc.b	$6B,$FC,$D1,$AF,$F3,$96,$DF,$F7,$07,$C9,$1E,$03,$F4,$67,$C4,$30
	.dc.b	$7E,$C1,$DB,$FA,$E3,$46,$4D,$9F,$DD,$5F,$E9,$1A,$DF,$91,$FA,$58
	.dc.b	$FC,$A6,$C9,$A7,$E1,$EB,$EA,$B8,$CB,$F1,$95,$C5,$B5,$C5,$FC,$2D
	.dc.b	$7F,$74,$8C,$47,$84,$DD,$73,$D3,$11,$2B,$BB,$A3,$9E,$3F,$1E,$BC
	.dc.b	$51,$B9,$ED,$1E,$76,$C7,$13,$53,$C5,$E7,$87,$E2,$B8,$7F,$AE,$E3
	.dc.b	$FD,$22,$FD,$11,$FD,$8A,$3F,$A4,$5F,$D7,$8D,$99,$26,$DF,$D9,$27
	.dc.b	$77,$D6,$3F,$B1,$C7,$5A,$E5,$1B,$AE,$6E,$9E,$7B,$FD,$97,$31,$4D
	.dc.b	$1C,$48,$64,$D4,$7E,$07,$D2,$FC,$DD,$FE,$88,$FD,$1B,$F9,$E3,$3B
	.dc.b	$7E,$B6,$3F,$5A,$D8,$D6,$5F,$91,$9E,$53,$98,$FC,$A4,$FE,$94,$8F
	.dc.b	$E9,$5A,$FE,$95,$AF,$E9,$5A,$FE,$95,$AF,$F8,$82,$51,$AC,$7F,$50
	.dc.b	$F9,$27,$DC,$7E,$8C,$FC,$83,$97,$F8,$6B,$EA,$DB,$71,$0C,$9E,$A3
	.dc.b	$94,$FE,$D7,$4F,$D2,$90,$FC,$F9,$F8,$F7,$4E,$08,$7E,$C4,$D6,$9F
	.dc.b	$C6,$1E,$46,$4E,$1B,$56,$57,$F2,$47,$C1,$18,$8F,$D6,$47,$27,$4F
	.dc.b	$E7,$31,$FE,$94,$EE,$FC,$AF,$7A,$FE,$09,$F2,$7F,$A0,$3F,$3C,$4D
	.dc.b	$7F,$34,$E5,$FC,$CB,$9D,$78,$C6,$8C,$91,$C9,$A6,$63,$FC,$FA,$7A
	.dc.b	$4E,$23,$F8,$34,$75,$75,$3D,$4D,$61,$FC,$9E,$7F,$6F,$C5,$FC,$DB
	.dc.b	$8F,$F9,$E9,$FC,$F9,$FE,$72,$67,$7F,$CE,$46,$70,$FE,$55,$39,$9F
	.dc.b	$E4,$D3,$99,$FE,$51,$3F,$CF,$C8,$FF,$3F,$6B,$FC,$FD,$AF,$F3,$F6
	.dc.b	$BF,$CF,$DA,$FE,$6D,$CD,$FF,$41,$7C,$F1,$E4,$3F,$23,$AF,$F5,$E8
	.dc.b	$CC,$79,$0D,$6D,$58,$7F,$6A,$8E,$71,$FD,$07,$E5,$3F,$8E,$9F,$C6
	.dc.b	$C9,$9F,$F1,$77,$2F,$F1,$67,$3A,$71,$8D,$19,$23,$93,$4C,$BF,$EF
	.dc.b	$73,$D2,$71,$1F,$D0,$E3,$AB,$A9,$EA,$6B,$1F,$EC,$93,$FC,$D3,$E5
	.dc.b	$7F,$AA,$93,$7F,$E9,$EF,$EB,$38,$0E,$63,$FB,$03,$BE,$0B,$1C,$08
	.dc.b	$61,$CB,$31,$FF,$09,$CF,$16,$C7,$06,$BC,$5A,$43,$FB,$04,$71,$7F
	.dc.b	$B1,$7B,$1F,$E0,$CF,$63,$FC,$26,$7F,$85,$1F,$F0,$79,$9E,$1F,$E6
	.dc.b	$51,$99,$FF,$00,$4E,$67,$FB,$E4,$E6,$7F,$BF,$4F,$F8,$54,$8F,$F8
	.dc.b	$55,$AF,$F8,$55,$AF,$F8,$55,$AF,$F8,$55,$AF,$F3,$17,$39,$FF,$85
	.dc.b	$DF,$3C,$79,$0F,$CB,$E4,$8C,$C7,$90,$D6,$D5,$87,$F9,$02,$39,$C7
	.dc.b	$F8,$5F,$5F,$EE,$67,$F7,$12,$1F,$DB,$4C,$9F,$DB,$0C,$ED,$C5,$D1
	.dc.b	$92,$39,$34,$CA,$FF,$28,$BD,$23,$11,$FE,$19,$8E,$AE,$9E,$A6,$B0
	.dc.b	$FF,$1D,$3F,$E0,$AF,$8B,$FF,$15,$93,$5F,$F1,$2B,$F9,$73,$0F,$F8
	.dc.b	$D5,$D6,$D6,$36,$43,$0E,$59,$7F,$E6,$89,$E2,$D8,$E0,$D7,$8B,$48
	.dc.b	$7F,$8D,$63,$8B,$FE,$37,$F9,$6F,$F3,$0F,$CB,$7F,$9A,$0F,$F3,$4B
	.dc.b	$FE,$64,$33,$BF,$F8,$1E,$33,$3F,$E5,$59,$CC,$FF,$94,$67,$33,$FE
	.dc.b	$53,$9F,$F3,$51,$1F,$F3,$53,$5F,$F3,$53,$5F,$F3,$53,$5F,$F3,$53
	.dc.b	$5F,$F0,$33,$9C,$BF,$CD,$8F,$9E,$3C,$87,$E5,$9C,$3E,$33,$3A,$56
	.dc.b	$1F,$E9,$48,$E5,$1F,$E6,$BF,$97,$FF,$23,$9F,$E4,$52,$69,$FE,$41
	.dc.b	$72,$FF,$90,$1C,$EB,$C6,$34,$64,$8E,$4D,$32,$FF,$D4,$B3,$D2,$71
	.dc.b	$1F,$E6,$D8,$EA,$EA,$7A,$9A,$C3,$FD,$1D,$3F,$E5,$EE,$BF,$F3,$F9
	.dc.b	$35,$FF,$3C,$BF,$CA,$CC,$3F,$E8,$97,$5B,$58,$D9,$0C,$39,$66,$1F
	.dc.b	$EB,$C9,$E2,$D8,$E0,$D7,$8B,$48,$7F,$A2,$63,$8B,$FE,$8B,$F9,$9F
	.dc.b	$F5,$AF,$CC,$FF,$AF,$0F,$F5,$EB,$FE,$B8,$33,$BF,$F9,$6E,$33,$3F
	.dc.b	$EA,$99,$CC,$FF,$A9,$67,$33,$FE,$A7,$9F,$F5,$F1,$1F,$F5,$F3,$5F
	.dc.b	$F5,$F3,$5F,$F5,$F3,$5F,$F5,$F3,$5F,$F2,$D3,$9C,$BF,$D8,$0F,$9E
	.dc.b	$3C,$87,$F2,$A7,$0F,$8C,$CE,$B5,$87,$FB,$A2,$39,$47,$FA,$FF,$E6
	.dc.b	$BF,$D3,$67,$FA,$64,$9A,$FF,$A5,$5C,$BF,$E9,$47,$3B,$71,$8D,$19
	.dc.b	$23,$93,$4C,$BF,$F7,$CC,$F4,$9C,$47,$FB,$06,$3A,$BA,$9E,$A6,$B0
	.dc.b	$FF,$71,$4F,$FA,$CB,$E6,$FF,$DA,$64,$D7,$FD,$9A,$FF,$7F,$30,$FF
	.dc.b	$B7,$5D,$6D,$63,$64,$30,$E5,$98,$0F,$F6,$B4,$F2,$71,$7F,$EF,$58
	.dc.b	$EF,$18,$E9,$72,$A7,$84,$F2,$BF,$03,$B9,$D4,$79,$30,$F0,$4B,$06
	.dc.b	$1F,$E9,$E8,$C9,$C7,$71,$FE,$B5,$29,$CA,$78,$F6,$70,$9F,$33,$5F
	.dc.b	$F7,$94,$F0,$7A,$3D,$1D,$9C,$67,$9B,$6B,$FD,$87,$3E,$13,$84,$7A
	.dc.b	$63,$D0,$FA,$27,$8B,$E1,$AF,$13,$F3,$DA,$F0,$3D,$0E,$08,$F0,$6B
	.dc.b	$CC,$C1,$0C,$1E,$19,$E8,$FB,$0C,$47,$D0,$DB,$E4,$4E,$A7,$D5,$3F
	.dc.b	$1E,$BD,$31,$A8,$E0,$F3,$30,$7C,$57,$84,$FC,$2A,$E1,$F3,$8C,$A5
	.dc.b	$FD,$48,$C4,$68,$CC,$78,$A3,$FD,$8F,$67,$71,$C6,$77,$87,$E0,$27
	.dc.b	$33,$C6,$59,$75,$64,$71,$7C,$11,$88,$E1,$16,$BA,$3C,$A3,$81,$0D
	.dc.b	$9E,$39,$E1,$8F,$FB,$DA,$FC,$77,$CA,$FB,$DF,$E5,$3D,$F6,$EE,$0F
	.dc.b	$9D,$18,$8A,$6A,$78,$3B,$6C,$EA,$24,$60,$87,$DC,$9E,$8F,$28,$EB
	.dc.b	$32,$30,$43,$84,$77,$6C,$76,$73,$1C,$5C,$9C,$6F,$F6,$75,$E0,$8C
	.dc.b	$9F,$8E,$F9,$EF,$C4,$7C,$F7,$EC,$E3,$F2,$2D,$AF,$DD,$CF,$E3,$4D
	.dc.b	$EF,$FB,$08,$FD,$FC,$EF,$46,$CB,$E1,$87,$E8,$2B,$52,$6C,$EA,$7F
	.dc.b	$62,$FE,$BC,$8F,$EE,$5A,$E8,$C1,$0C,$10,$FE,$33,$3E,$37,$47,$F1
	.dc.b	$48,$F0,$34,$FF,$6F,$C6,$DF,$D8,$46,$9F,$90,$69,$C5,$FC,$68,$FD
	.dc.b	$EC,$7F,$3B,$DF,$F7,$1E,$17,$E4,$B1,$F8,$AD,$C0,$D3,$F5,$E5,$DB
	.dc.b	$F5,$B3,$C1,$B3,$C1,$FD,$26,$1F,$C2,$2B,$84,$78,$0D,$BD,$1A,$43
	.dc.b	$5C,$1B,$64,$FF,$38,$9F,$14,$E2,$3B,$4F,$17,$4F,$8B,$2F,$E8,$3F
	.dc.b	$3F,$FA,$CE,$0F,$E3,$0F,$E4,$63,$3C,$1F,$AB,$BF,$E3,$F3,$B7,$FA
	.dc.b	$3B,$5C,$C6,$E7,$8B,$D0,$FD,$23,$D2,$B1,$12,$79,$BA,$CF,$F2,$B7
	.dc.b	$D6,$F9,$D6,$AB,$A5,$C8,$C1,$0F,$1B,$97,$CB,$1D,$DC,$9F,$D8,$3E
	.dc.b	$07,$F9,$A7,$C1,$7E,$80,$FD,$0B,$FC,$DC,$CF,$0F,$EA,$F2,$FD,$4C
	.dc.b	$EA,$BF,$A3,$D6,$AB,$FA,$14,$E6,$7F,$A2,$4F,$F5,$52,$3F,$D7,$DA
	.dc.b	$FF,$5C,$6B,$FA,$46,$BF,$E0,$2D,$91,$88,$FE,$CC,$EB,$2F,$EF,$AF
	.dc.b	$A6,$3C,$A7,$EA,$1C,$3E,$43,$3B,$D6,$1F,$14,$C9,$FE,$06,$F7,$BF
	.dc.b	$48,$4D,$BF,$3E,$FF,$52,$30,$FF,$63,$75,$BD,$8D,$90,$C3,$96,$4F
	.dc.b	$77,$71,$88,$FE,$E3,$3D,$9D,$46,$70,$FF,$0E,$CE,$67,$75,$FA,$1D
	.dc.b	$3F,$90,$F8,$DF,$8D,$6E,$DF,$99,$6B,$FD,$EE,$35,$A1,$A7,$E9,$E7
	.dc.b	$77,$FE,$2A,$8C,$D7,$E6,$EB,$83,$D5,$AF,$53,$51,$C1,$C1,$9D,$E5
	.dc.b	$75,$8E,$53,$E1,$AE,$B5,$AA,$FF,$0A,$CE,$49,$5E,$3F,$57,$FC,$25
	.dc.b	$EB,$7F,$3E,$3F,$9F,$BF,$E1,$53,$3C,$3F,$9F,$C5,$3A,$C3,$FC,$4B
	.dc.b	$5A,$AF,$F1,$04,$E6,$7F,$C4,$53,$FE,$33,$23,$FE,$3D,$6B,$FE,$3A
	.dc.b	$6B,$FD,$09,$AF,$E7,$DA,$FF,$90,$DC,$EB,$FE,$5A,$7D,$11,$E4,$3F
	.dc.b	$A2,$B8,$7C,$66,$77,$AC,$3F,$CC,$31,$CA,$3F,$C7,$5E,$97,$F4,$12
	.dc.b	$6D,$FC,$F1,$FF,$17,$98,$76,$64,$FC,$76,$48,$D5,$FF,$9B,$9E,$7C
	.dc.b	$1C,$2A,$4E,$5E,$51,$88,$FF,$27,$CF,$26,$CF,$68,$C6,$3F,$E7,$C7
	.dc.b	$AC,$FF,$9F,$3C,$EF,$CC,$9F,$87,$26,$BF,$95,$72,$FE,$55,$CE,$FF
	.dc.b	$E5,$B7,$31,$4D,$1C,$48,$64,$D4,$49,$F1,$47,$03,$0F,$F9,$AA,$7A
	.dc.b	$BA,$8C,$DF,$FA,$5B,$2F,$F3,$47,$57,$F8,$50,$FF,$0A,$BF,$E6,$B3
	.dc.b	$3B,$FE,$5E,$33,$7F,$E7,$A9,$CC,$FF,$9D,$67,$33,$FE,$77,$9F,$F4
	.dc.b	$41,$1F,$F4,$93,$5F,$F4,$7B,$5F,$F0,$BB,$5F,$E6,$8D,$7F,$D3,$0E
	.dc.b	$6B,$D5,$1E,$32,$6B,$E1,$70,$53,$0D,$7F,$D6,$B3,$EF,$78,$9F,$EB
	.dc.b	$8F,$A0,$FF,$0B,$13,$5F,$F0,$8B,$FE,$84,$30,$FF,$A5,$5D,$6D,$63
	.dc.b	$64,$30,$E5,$93,$DD,$DC,$62,$3F,$D4,$73,$D9,$D4,$67,$0F,$F6,$7C
	.dc.b	$E6,$77,$5F,$A6,$FA,$1F,$E5,$A7,$F1,$B2,$69,$FC,$A5,$CB,$FC,$A5
	.dc.b	$CE,$DF,$EA,$F7,$31,$4D,$1C,$48,$64,$D4,$4A,$FF,$70,$CF,$82,$71
	.dc.b	$1F,$EB,$E8,$EA,$EA,$3A,$CF,$FA,$FF,$E8,$FF,$D7,$9F,$47,$FE,$68
	.dc.b	$3F,$CD,$2F,$FA,$FC,$CE,$FF,$CA,$E3,$37,$FE,$CE,$9C,$CF,$FB,$26
	.dc.b	$73,$3D,$4F,$F7,$24,$FF,$B7,$48,$E0,$87,$FB,$91,$CB,$4C,$9F,$E6
	.dc.b	$E7,$2D,$3C,$44,$3C,$1B,$78,$58,$0F,$F7,$0C,$7B,$5F,$F3,$9B,$5D
	.dc.b	$9A,$3F,$CE,$F9,$7F,$9A,$C8,$7F,$99,$8F,$F6,$C9,$83,$FD,$D4,$6B
	.dc.b	$5A,$EC,$86,$1C,$B2,$3C,$DA,$B9,$7F,$D3,$F3,$CE,$73,$18,$8E,$71
	.dc.b	$8D,$FF,$DD,$F1,$D6,$30,$F9,$5A,$F8,$EF,$FE,$00,$7D,$27,$F8,$18
	.dc.b	$FE,$E8,$4D,$BF,$C0,$4E,$5F,$F0,$13,$9D,$B8,$C6,$8C,$91,$C9,$A6
	.dc.b	$5F,$9E,$7A,$4E,$21,$97,$C1,$C5,$5D,$E7,$85,$F7,$38,$56,$E7,$FC
	.dc.b	$3F,$F4,$DF,$7E,$A4,$F3,$FA,$6F,$F6,$29,$FE,$C6,$9D,$4F,$C9,$6B
	.dc.b	$E2,$D3,$FD,$9F,$3C,$38,$3F,$C1,$95,$AB,$F3,$36,$B7,$3F,$E1,$1A
	.dc.b	$ED,$7D,$AF,$6F,$63,$FD,$B7,$75,$A6,$72,$46,$5B,$19,$8C,$11,$C1
	.dc.b	$0F,$5C,$70,$6C,$6D,$F8,$B8,$78,$1B,$3B,$7E,$21,$FE,$D8,$9F,$74
	.dc.b	$66,$3E,$33,$6B,$DB,$BF,$2A,$D9,$2B,$E1,$5F,$BA,$7B,$FD,$3F,$FB
	.dc.b	$5C,$9B,$FF,$B4,$93,$FE,$14,$72,$F0,$34,$64,$A6,$4D,$6F,$6F,$F4
	.dc.b	$D3,$C9,$B3,$C4,$D3,$EC,$8F,$1C,$E3,$1F,$CA,$47,$19,$D5,$7F,$8A
	.dc.b	$3D,$2F,$F3,$79,$FE,$5D,$26,$FF,$E6,$B7,$2F,$F9,$AD,$CF,$07,$D0
	.dc.b	$73,$14,$D1,$C4,$86,$4D,$44,$AF,$E0,$53,$E0,$9C,$47,$D6,$8E,$AE
	.dc.b	$A3,$AC,$FE,$03,$EA,$7E,$EF,$D4,$B8,$7F,$CD,$4D,$75,$7F,$6B,$81
	.dc.b	$5C,$27,$67,$F9,$C5,$FC,$39,$97,$F1,$B5,$CE,$BF,$8C,$4E,$9F,$D0
	.dc.b	$3F,$9F,$23,$82,$1F,$9C,$72,$D3,$27,$EF,$2B,$CA,$E0,$FF,$69,$47
	.dc.b	$77,$D6,$7B,$36,$AF,$90,$F7,$47,$96,$33,$1F,$A0,$8E,$4E,$E3,$F5
	.dc.b	$3F,$55,$F9,$8D,$29,$F9,$A9,$FE,$6D,$19,$6C,$7F,$9C,$23,$11,$FA
	.dc.b	$C6,$ED,$FC,$8D,$BC,$0C,$51,$FA,$6E,$0E,$F8,$4C,$3A,$BE,$28,$C4
	.dc.b	$75,$8F,$23,$5D,$3C,$08,$64,$F3,$C7,$2C,$3C,$93,$C9,$AF,$29,$FE
	.dc.b	$0B,$F5,$7F,$ED,$23,$FD,$80,$4D,$FF,$D9,$CE,$5F,$F6,$73,$9D,$FF
	.dc.b	$89,$B9,$8A,$68,$E2,$43,$26,$A2,$4F,$F6,$59,$F0,$4E,$23,$CA,$F1
	.dc.b	$74,$F5,$9F,$E4,$7E,$AF,$F2,$0F,$5B,$F3,$15,$FC,$81,$B8,$FF,$57
	.dc.b	$9F,$E4,$A6,$F8,$7F,$DA,$53,$AD,$BE,$C4,$EA,$FF,$55,$39,$D8,$D9
	.dc.b	$D4,$FF,$3A,$7F,$A1,$11,$FE,$8C,$D7,$46,$08,$60,$87,$E9,$5E,$0F
	.dc.b	$F8,$25,$EE,$D1,$BF,$C2,$7F,$80,$46,$9F,$E8,$2F,$F8,$B6,$35,$10
	.dc.b	$E4,$71,$7F,$A6,$FD,$63,$46,$98,$71,$5F,$E1,$B9,$D4,$66,$64,$E6
	.dc.b	$36,$FF,$52,$9D,$B9,$9D,$B9,$9D,$E9,$FE,$0D,$AD,$CE,$6B,$73,$9A
	.dc.b	$DE,$7F,$AF,$BD,$CE,$6B,$73,$9A,$DE,$6B,$DC,$E6,$B7,$39,$AD,$D6
	.dc.b	$A3,$13,$FD,$CB,$EB,$7F,$7F,$06,$93,$7F,$F7,$BB,$5E,$07,$EC,$F0
	.dc.b	$FE,$0D,$1F,$C1,$72,$FF,$23,$3B,$6C,$EE,$BB,$3D,$0C,$4E,$E7,$FB
	.dc.b	$FB,$CD,$B1,$A2,$1E,$2E,$0F,$1B,$23,$BB,$97,$FC,$B1,$F5,$DF,$E0
	.dc.b	$CF,$AF,$FD,$11,$FA,$37,$FC,$1E,$67,$87,$FA,$A4,$FF,$97,$E7,$4F
	.dc.b	$F5,$39,$D4,$FF,$86,$A7,$33,$FE,$1C,$9F,$F1,$51,$1F,$F1,$73,$5F
	.dc.b	$F1,$3B,$5F,$E9,$4D,$7F,$C9,$98,$A3,$11,$FE,$8D,$8C,$3D,$0F,$21
	.dc.b	$0F,$11,$FA,$C3,$E0,$99,$29,$AD,$F6,$C3,$27,$07,$FC,$6F,$ED,$7E
	.dc.b	$AF,$5B,$BF,$E4,$A3,$F1,$86,$6B,$FD,$0C,$D7,$FC,$72,$EA,$38,$1F
	.dc.b	$B4,$8D,$B5,$DB,$A3,$BC,$C8,$DC,$78,$8C,$3C,$23,$0E,$8E,$AE,$B8
	.dc.b	$3F,$D0,$15,$E3,$AC,$6F,$FD,$7F,$EC,$3F,$46,$4D,$BF,$30,$FE,$4E
	.dc.b	$FF,$D1,$0E,$63,$FC,$B6,$EF,$7B,$1C,$08,$61,$CB,$32,$FF,$56,$4F
	.dc.b	$16,$C7,$06,$BC,$5A,$43,$FC,$CB,$1C,$5F,$F3,$3F,$D8,$FF,$9A,$BE
	.dc.b	$CB,$FA,$01,$FD,$05,$FF,$36,$99,$E1,$FC,$FC,$FF,$9B,$A7,$4F,$F8
	.dc.b	$C6,$75,$3F,$E7,$89,$CC,$FF,$9E,$A7,$FD,$10,$47,$FD,$18,$D7,$FD
	.dc.b	$0A,$D7,$FC,$50,$D7,$F3,$ED,$7F,$3C,$E7,$1F,$F5,$B3,$E8,$8F,$21
	.dc.b	$F9,$D3,$DE,$60,$F2,$1A,$DA,$B0,$FF,$6D,$47,$38,$FF,$49,$FD,$9F
	.dc.b	$E5,$EB,$FD,$88,$FF,$27,$26,$BF,$95,$8C,$B7,$5F,$F1,$34,$EA,$39
	.dc.b	$99,$E0,$47,$59,$DC,$FF,$3A,$8E,$BC,$1C,$0E,$31,$32,$FF,$78,$4F
	.dc.b	$1B,$E7,$3D,$63,$C1,$3A,$8C,$10,$E4,$FF,$AA,$A3,$97,$BD,$FD,$00
	.dc.b	$9B,$FF,$35,$7F,$3D,$8F,$F9,$05,$CC,$7F,$AE,$1D,$F0,$58,$E0,$43
	.dc.b	$0E,$59,$63,$FD,$77,$3C,$9C,$67,$FE,$F3,$8E,$F1,$8B,$F1,$57,$6B
	.dc.b	$E7,$3C,$A3,$C0,$EF,$25,$FA,$97,$DA,$FF,$B3,$3E,$D7,$FC,$44,$7F
	.dc.b	$88,$DF,$F6,$79,$9E,$1F,$D1,$E6,$CF,$FC,$C5,$5B,$C7,$FD,$29,$7B
	.dc.b	$BF,$F6,$DC,$E6,$7F,$DB,$B3,$FE,$EB,$23,$FE,$EF,$6B,$FE,$E8,$6B
	.dc.b	$FE,$8B,$6B,$FD,$09,$AF,$F4,$17,$37,$E5,$C3,$DB,$3E,$42,$6F,$E2
	.dc.b	$70,$53,$0D,$7E,$45,$7C,$27,$89,$F2,$FE,$DB,$F2,$07,$F8,$24,$9B
	.dc.b	$FE,$75,$CB,$FE,$14,$73,$C1,$FE,$FE,$73,$14,$D1,$C4,$86,$4D,$44
	.dc.b	$9F,$14,$70,$30,$B1,$F8,$55,$DA,$37,$3A,$9F,$8D,$F6,$FF,$E1,$E2
	.dc.b	$6F,$FE,$13,$7F,$0C,$61,$FC,$B5,$75,$DB,$FD,$45,$3C,$38,$33,$7D
	.dc.b	$2F,$91,$0C,$D7,$05,$38,$D5,$7D,$F0,$FC,$C3,$D5,$B5,$CD,$AF,$18
	.dc.b	$F3,$B9,$8F,$34,$F4,$8F,$1D,$77,$F8,$1F,$5D,$7E,$97,$E0,$BF,$D0
	.dc.b	$47,$FA,$0E,$33,$1C,$9B,$97,$E3,$67,$7C,$3F,$A4,$CF,$FD,$89,$5A
	.dc.b	$6D,$FF,$BD,$EB,$75,$E1,$8F,$F7,$D5,$76,$7F,$C3,$4D,$78,$9E,$2C
	.dc.b	$91,$96,$C7,$16,$BB,$8D,$11,$D3,$5D,$18,$3C,$64,$3C,$DC,$1E,$76
	.dc.b	$1F,$C2,$E3,$DD,$3E,$29,$CC,$FC,$73,$91,$B3,$FA,$A6,$3F,$B5,$F9
	.dc.b	$EF,$E7,$07,$F9,$C4,$9B,$7F,$45,$72,$FF,$9F,$1C,$F0,$7E,$1D,$CC
	.dc.b	$53,$47,$12,$19,$35,$12,$BF,$92,$4F,$82,$71,$1F,$88,$8E,$AE,$A3
	.dc.b	$AC,$FE,$5F,$E8,$7F,$D0,$A4,$DF,$FC,$F4,$FF,$2D,$30,$EC,$C9,$FA
	.dc.b	$0A,$FE,$69,$1A,$6C,$6B,$7E,$0C,$8C,$BC,$E3,$11,$F9,$E9,$E4,$D9
	.dc.b	$EC,$D7,$78,$FF,$25,$9F,$45,$7E,$C7,$E9,$3E,$DF,$D2,$7F,$B8,$CF
	.dc.b	$F7,$23,$F7,$8C,$EF,$F9,$FD,$93,$AC,$BF,$57,$8F,$F3,$1B,$DD,$FE
	.dc.b	$F6,$B5,$5D,$CF,$CC,$4F,$E6,$C8,$E0,$87,$E8,$5C,$B4,$C9,$FB,$AB
	.dc.b	$46,$A3,$0F,$A5,$F2,$F0,$A7,$71,$86,$EF,$E5,$88,$7F,$69,$AE,$51
	.dc.b	$88,$FE,$1B,$1D,$5F,$64,$7F,$1B,$DB,$FC,$28,$7F,$B3,$48,$7F,$88
	.dc.b	$4C,$9F,$ED,$83,$3B,$71,$74,$64,$8E,$4D,$32,$7F,$BF,$3D,$23,$11
	.dc.b	$FC,$8A,$3A,$BA,$7A,$9A,$CF,$FB,$43,$FE,$2B,$FA,$7F,$F7,$09,$37
	.dc.b	$FF,$6B,$3F,$E0,$93,$0F,$F3,$97,$5B,$58,$D9,$0C,$39,$65,$FF,$85
	.dc.b	$67,$8B,$63,$83,$5E,$2D,$21,$FC,$FA,$38,$BF,$D0,$3E,$AB,$F8,$AF
	.dc.b	$D5,$38,$3F,$30,$D7,$FA,$84,$7F,$1A,$36,$E8,$FC,$F3,$99,$FC,$34
	.dc.b	$F1,$7F,$C5,$8D,$7C,$07,$E2,$E3,$F9,$B1,$1F,$E6,$CD,$7F,$9E,$35
	.dc.b	$FC,$D3,$5F,$F6,$FB,$5F,$F6,$F3,$9C,$BF,$C3,$CF,$96,$3D,$67,$F2
	.dc.b	$ED,$FF,$3D,$19,$8F,$F2,$13,$EB,$D7,$71,$0C,$9E,$93,$96,$EF,$AC
	.dc.b	$FF,$37,$0D,$A4,$FF,$93,$63,$11,$FD,$ED,$DB,$FE,$7F,$73,$1B,$8E
	.dc.b	$2F,$F7,$68,$DE,$FC,$67,$39,$7F,$8E,$E7,$8D,$6E,$7A,$10,$E0,$71
	.dc.b	$A9,$1D,$E3,$2F,$86,$32,$F3,$8C,$BD,$63,$FC,$29,$F1,$4D,$E4,$FF
	.dc.b	$9A,$E3,$11,$A3,$31,$FE,$10,$8F,$F3,$71,$A7,$84,$6B,$0F,$F2,$B4
	.dc.b	$E6,$78,$4F,$F3,$89,$CB,$BD,$3F,$8C,$56,$E7,$33,$E1,$E0,$F1,$32
	.dc.b	$FB,$B9,$69,$0E,$EE,$5A,$66,$7A,$BE,$89,$FF,$1B,$7D,$8F,$F8,$03
	.dc.b	$EC,$7F,$41,$5F,$E6,$F6,$E3,$FC,$B6,$7F,$C0,$A6,$F8,$7F,$A4,$4E
	.dc.b	$B6,$4E,$B1,$FE,$65,$39,$35,$19,$8F,$F1,$13,$FE,$21,$23,$FE,$21
	.dc.b	$6B,$A3,$04,$30,$43,$FA,$43,$C1,$FF,$53,$3D,$DA,$7E,$89,$C3,$FE
	.dc.b	$4C,$7D,$1B,$FC,$48,$FE,$95,$1A,$C7,$FD,$24,$FF,$AF,$67,$FC,$BD
	.dc.b	$F6,$7F,$ED,$73,$F3,$73,$FE,$97,$6E,$5F,$D5,$5D,$6F,$FE,$E7,$73
	.dc.b	$91,$A6,$B7,$EB,$3F,$A8,$AE,$4D,$AE,$0F,$32,$1C,$0E,$33,$32,$FF
	.dc.b	$64,$5F,$69,$E4,$F7,$8C,$C6,$09,$3F,$E7,$BF,$B6,$FC,$B6,$1F,$DB
	.dc.b	$DB,$81,$A7,$FB,$4C,$BB,$FE,$7E,$34,$D8,$D3,$48,$77,$3F,$51,$3D
	.dc.b	$DC,$E1,$FE,$E1,$9E,$BB,$EE,$64,$E5,$E9,$18,$8F,$F4,$4C,$F9,$27
	.dc.b	$B4,$FF,$A5,$FE,$37,$FC,$C3,$F1,$DF,$9B,$3F,$38,$FF,$99,$0C,$F0
	.dc.b	$FF,$89,$E3,$38,$7F,$83,$E7,$33,$FE,$A1,$9C,$CF,$53,$FC,$F9,$3F
	.dc.b	$E7,$F2,$38,$21,$FE,$85,$72,$D3,$27,$F8,$95,$CB,$4F,$11,$0F,$06
	.dc.b	$FE,$16,$03,$FD,$57,$1E,$D7,$FA,$2B,$5D,$9A,$3F,$A4,$FD,$C7,$F4
	.dc.b	$7D,$7F,$D7,$CF,$E6,$89,$B2,$32,$D9,$F0,$70,$7F,$BE,$EB,$73,$AA
	.dc.b	$95,$AD,$BF,$C7,$95,$A8,$C6,$E8,$F1,$46,$1A,$FF,$B1,$9E,$0D,$BF
	.dc.b	$13,$6F,$CB,$12,$39,$3A,$C3,$CC,$DC,$3C,$6F,$EC,$6F,$DF,$7F,$ED
	.dc.b	$FF,$B9,$FD,$19,$37,$FD,$03,$FD,$3F,$7D,$46,$08,$61,$CB,$32,$F7
	.dc.b	$CF,$39,$C4,$7F,$B8,$A3,$B4,$F1,$8D,$BD,$1A,$E5,$E0,$E5,$AE,$72
	.dc.b	$65,$FE,$6B,$FB,$AF,$F6,$1F,$DD,$FF,$3D,$3F,$9F,$3F,$EC,$63,$3C
	.dc.b	$3F,$E8,$98,$CE,$1F,$E7,$19,$CC,$FE,$3E,$3F,$DF,$33,$A7,$17,$E0
	.dc.b	$D9,$7F,$EE,$32,$4F,$FB,$8D,$AF,$FB,$A5,$AF,$FA,$1D,$AF,$78,$E6
	.dc.b	$F1,$21,$C4,$F3,$10,$F1,$E1,$F0,$5F,$F1,$3C,$FB,$DD,$1F,$92,$7F
	.dc.b	$A1,$34,$C8,$AF,$64,$71,$E0,$EB,$6F,$92,$FC,$BD,$7C,$17,$FA,$19
	.dc.b	$38,$3F,$21,$3B,$6C,$65,$E6,$F2,$C7,$FD,$44,$D9,$E1,$3A,$3F,$A4
	.dc.b	$46,$48,$E0,$9B,$EC,$E0,$CC,$BF,$63,$3C,$27,$11,$EF,$BF,$0D,$72
	.dc.b	$8F,$DB,$FD,$E7,$F4,$62,$6D,$FD,$0D,$FD,$24,$6A,$30,$43,$E3,$C6
	.dc.b	$A3,$B6,$28,$E7,$3A,$38,$F0,$56,$57,$57,$51,$88,$E7,$3C,$9F,$14
	.dc.b	$67,$3F,$CA,$4E,$67,$75,$FD,$52,$DF,$7D,$F9,$D9,$FE,$2B,$F7,$DF
	.dc.b	$E2,$33,$FC,$49,$19,$8F,$4B,$70,$FD,$04,$EF,$87,$FD,$E5,$39,$D3
	.dc.b	$FD,$B8,$DC,$15,$BA,$FC,$5C,$E4,$ED,$87,$F3,$DA,$F2,$B8,$6B,$E3
	.dc.b	$6B,$FE,$EF,$8E,$2F,$80,$E0,$E2,$37,$1F,$AD,$8C,$11,$D9,$0F,$29
	.dc.b	$FE,$F3,$7C,$CF,$FA,$49,$AE,$8D,$EF,$5F,$E7,$F1,$A6,$9F,$E9,$AF
	.dc.b	$BF,$FC,$F1,$FE,$2A,$26,$DF,$A8,$72,$FF,$89,$DC,$EF,$FB,$27,$31
	.dc.b	$4D,$1C,$48,$64,$D4,$49,$FE,$A7,$3E,$09,$C4,$7E,$5A,$3A,$BA,$8E
	.dc.b	$B3,$FC,$0B,$E5,$BF,$C5,$24,$DB,$FC,$4C,$FE,$7F,$6F,$E4,$8E,$63
	.dc.b	$81,$A3,$F2,$4E,$6F,$FB,$2B,$CF,$83,$73,$27,$2F,$18,$C4,$7E,$B2
	.dc.b	$79,$36,$7A,$9D,$CD,$D7,$91,$FD,$17,$E0,$3F,$95,$FC,$07,$FD,$04
	.dc.b	$7F,$A0,$DF,$CC,$19,$DF,$FD,$F1,$7F,$DE,$A7,$4F,$EA,$6F,$FA,$15
	.dc.b	$6E,$BF,$93,$D6,$AB,$F9,$4C,$FE,$4C,$8F,$E7,$5A,$FE,$75,$AF,$ED
	.dc.b	$77,$46,$23,$FA,$33,$5F,$D4,$39,$BF,$1C,$79,$78,$55,$CA,$30,$DE
	.dc.b	$0F,$2C,$62,$FE,$0B,$4F,$F7,$8B,$87,$67,$08,$FE,$8B,$F8,$17,$F3
	.dc.b	$93,$FD,$0A,$4D,$FF,$A1,$B9,$7F,$D0,$4E,$77,$E3,$1A,$32,$47,$26
	.dc.b	$99,$5F,$E2,$C9,$E9,$38,$8F,$E5,$F1,$D5,$D4,$F5,$35,$97,$F7,$99
	.dc.b	$FE,$A1,$F8,$1F,$FA,$10,$9B,$7F,$A0,$5F,$CE,$69,$F2,$9C,$C7,$F6
	.dc.b	$47,$7C,$16,$38,$10,$C3,$96,$63,$FE,$44,$9E,$2D,$8E,$0D,$78,$B4
	.dc.b	$87,$F4,$C8,$E2,$FF,$4E,$F9,$9F,$E6,$1F,$35,$FE,$DD,$3F,$DB,$CF
	.dc.b	$F3,$33,$3C,$3F,$9C,$9F,$EA,$B3,$A7,$F0,$B3,$A9,$FF,$00,$CE,$67
	.dc.b	$FC,$07,$3F,$CB,$48,$FF,$37,$6B,$FC,$DD,$AF,$E6,$DA,$FF,$87,$DA
	.dc.b	$FF,$41,$73,$9F,$F8,$B5,$F4,$47,$90,$6D,$E2,$70,$53,$0D,$7F,$CE
	.dc.b	$31,$C9,$C9,$FE,$74,$FC,$13,$FC,$20,$7F,$B7,$09,$B7,$F8,$69,$CB
	.dc.b	$FE,$D9,$73,$BF,$F8,$91,$CC,$53,$47,$12,$19,$35,$12,$7C,$51,$C0
	.dc.b	$C3,$FE,$08,$9E,$AE,$A3,$38,$7F,$94,$7F,$05,$FF,$6D,$93,$5F,$F6
	.dc.b	$C3,$F9,$8E,$23,$98,$FF,$1D,$BB,$DE,$C7,$02,$18,$72,$CC,$3F,$D3
	.dc.b	$73,$C5,$B1,$C1,$AF,$16,$90,$FF,$11,$C7,$17,$FC,$4B,$F8,$3F,$F8
	.dc.b	$1F,$F0,$77,$07,$E5,$72,$FF,$37,$46,$23,$FC,$A7,$3F,$E0,$D3,$83
	.dc.b	$BE,$2F,$E3,$F3,$BB,$FF,$58,$D6,$DF,$E4,$31,$97,$FC,$6A,$DB,$FF
	.dc.b	$3E,$4E,$E7,$FC,$20,$FF,$82,$C8,$FF,$85,$5A,$E8,$C1,$0C,$10,$FE
	.dc.b	$78,$F0,$7C,$A7,$F3,$47,$CD,$C1,$EE,$8C,$10,$C3,$96,$1C,$8C,$9F
	.dc.b	$ED,$0F,$9E,$FF,$36,$0D,$A4,$65,$AF,$F9,$5A,$37,$1F,$E7,$A7,$83
	.dc.b	$FE,$82,$6C,$71,$E0,$C4,$EC,$E7,$87,$FA,$96,$B7,$5D,$67,$55,$C1
	.dc.b	$91,$CC,$86,$DC,$BC,$63,$B9,$8A,$ED,$3E,$0C,$BF,$AF,$DF,$FA,$2F
	.dc.b	$F0,$86,$D3,$1F,$F7,$9C,$62,$3F,$D3,$AD,$8C,$C7,$F9,$9E,$31,$1F
	.dc.b	$E7,$A6,$E1,$FE,$C1,$6D,$FF,$6E,$7F,$60,$63,$80,$D2,$5F,$53,$C9
	.dc.b	$7D,$5B,$7D,$5B,$7D,$5E,$07,$A0,$CD,$F6,$6D,$F8,$9A,$43,$87,$07
	.dc.b	$FA,$F3,$F0,$AF,$F3,$B7,$E1,$5F,$A1,$C3,$FC,$0E,$DC,$13,$FE,$7B
	.dc.b	$37,$C3,$FA,$F8,$4E,$E7,$59,$FF,$83,$A3,$91,$97,$FC,$23,$1F,$E7
	.dc.b	$52,$3F,$E7,$F6,$BF,$E7,$F6,$BF,$E2,$36,$BF,$EE,$76,$BF,$E8,$97
	.dc.b	$39,$FF,$BD,$5F,$1C,$7B,$CF,$E9,$20,$FF,$5A,$BA,$71,$1C,$5F,$39
	.dc.b	$C0,$D9,$83,$A1,$D7,$7F,$F7,$3F,$E1,$7F,$ED,$B3,$F4,$32,$CB,$FC
	.dc.b	$67,$18,$6E,$FF,$EE,$88,$E0,$FE,$6F,$24,$6D,$B1,$B2,$1E,$03,$84
	.dc.b	$72,$E0,$C4,$4C,$7D,$B7,$E0,$9E,$06,$63,$FD,$AF,$5D,$CE,$4F,$0B
	.dc.b	$F0,$DE,$F0,$F6,$47,$92,$FE,$65,$7F,$B8,$3E,$8B,$F4,$44,$DF,$F2
	.dc.b	$EF,$F5,$53,$0F,$0B,$E7,$1B,$DF,$19,$FF,$96,$A3,$77,$E3,$74,$DB
	.dc.b	$EA,$43,$8E,$29,$F0,$46,$67,$83,$A8,$CB,$4D,$47,$AD,$FF,$75,$7E
	.dc.b	$19,$FE,$DD,$FC,$37,$F4,$C7,$E9,$DF,$F7,$01,$9E,$1F,$D3,$46,$6B
	.dc.b	$FC,$E3,$39,$9F,$2C,$E6,$7C,$F3,$FE,$D9,$23,$FE,$E3,$6B,$FE,$E3
	.dc.b	$6B,$FE,$87,$6B,$F4,$B0,$46,$23,$FD,$DC,$EB,$3F,$B2,$FA,$23,$CA
	.dc.b	$7E,$7E,$32,$E0,$F2,$9A,$DE,$B0,$FC,$34,$F3,$8F,$A3,$F8,$77,$F5
	.dc.b	$1D,$AE,$FF,$9D,$8F,$CE,$19,$BF,$C7,$35,$61,$F8,$E8,$DC,$F1,$15
	.dc.b	$D0,$EA,$CA,$E6,$E9,$B1,$97,$AE,$DF,$EA,$6A,$EB,$3C,$E7,$8D,$6E
	.dc.b	$FC,$4F,$38,$E3,$92,$FD,$15,$AA,$E1,$87,$57,$84,$7E,$A3,$F0,$FF
	.dc.b	$E8,$C4,$DB,$F9,$DB,$FE,$39,$30,$FE,$69,$D7,$05,$8D,$90,$C3,$96
	.dc.b	$65,$FB,$99,$E2,$D8,$E0,$D7,$8B,$48,$7B,$A3,$8B,$F2,$3F,$2A,$FD
	.dc.b	$FB,$FA,$EE,$0F,$E9,$07,$F4,$B8,$CC,$7F,$10,$6B,$F2,$23,$67,$F8
	.dc.b	$E6,$33,$87,$F2,$76,$E5,$FE,$DA,$9D,$9F,$C1,$23,$27,$4A,$FE,$78
	.dc.b	$D4,$D2,$3B,$6B,$E4,$6B,$E8,$6B,$EE,$6B,$F0,$4F,$6D,$79,$DB,$5E
	.dc.b	$07,$FA,$76,$DE,$77,$35,$E8,$69,$A3,$E3,$18,$3F,$DF,$4E,$DF,$DA
	.dc.b	$4F,$B2,$7F,$C2,$33,$FE,$A4,$C7,$FC,$0E,$D7,$E3,$47,$FB,$6A,$31
	.dc.b	$5F,$4E,$38,$92,$59,$70,$79,$C6,$5C,$46,$59,$3C,$DD,$9C,$9C,$3A
	.dc.b	$7F,$11,$3D,$E7,$18,$7D,$F8,$ED,$3C,$E7,$EF,$4F,$EB,$30,$FB,$6F
	.dc.b	$F6,$9C,$3F,$82,$E7,$FE,$41,$6E,$DC,$23,$F5,$F5,$C1,$D5,$71,$3A
	.dc.b	$57,$0C,$3F,$2D,$19,$9D,$9C,$A7,$53,$FA,$19,$D4,$F7,$B9,$3C,$A3
	.dc.b	$89,$87,$83,$CE,$78,$C6,$E7,$F3,$9F,$89,$7E,$1B,$F1,$3F,$D5,$1F
	.dc.b	$AB,$7F,$12,$67,$5F,$D4,$46,$6B,$F0,$F3,$99,$FD,$34,$E6,$7A,$9F
	.dc.b	$9F,$9F,$D1,$11,$C1,$0F,$D1,$39,$69,$93,$F4,$8E,$5A,$78,$88,$78
	.dc.b	$34,$F0,$B0,$FE,$09,$1E,$C7,$F3,$B9,$23,$11,$B3,$F8,$83,$FA,$77
	.dc.b	$F8,$97,$D4,$FE,$88,$9B,$7E,$89,$FD,$01,$87,$F8,$43,$AD,$AC,$6C
	.dc.b	$86,$1C,$B2,$7F,$8D,$4F,$16,$C7,$06,$BC,$5A,$43,$CE,$F7,$9F,$E3
	.dc.b	$5F,$8A,$7E,$48,$FC,$E1,$36,$FC,$83,$97,$F3,$CE,$75,$E3,$1A,$32
	.dc.b	$47,$26,$99,$7F,$CC,$E7,$A4,$E2,$3F,$9E,$47,$57,$53,$D4,$D6,$DF
	.dc.b	$CD,$67,$F5,$3B,$FF,$1C,$FC,$57,$FA,$01,$FD,$05,$FE,$3E,$67,$83
	.dc.b	$F9,$F4,$67,$0F,$E3,$B3,$99,$FE,$79,$39,$9F,$E7,$D3,$FC,$DC,$8F
	.dc.b	$F3,$76,$BF,$CD,$DA,$FF,$37,$6B,$FE,$54,$DB,$FC,$E4,$F1,$21,$C4
	.dc.b	$F1,$90,$F0,$9F,$94,$8F,$5B,$93,$FA,$E3,$F9,$3C,$3F,$AC,$3A,$8E
	.dc.b	$27,$2D,$FF,$AC,$7E,$2D,$FC,$E0,$9B,$7F,$38,$7F,$9B,$18,$76,$64
	.dc.b	$FE,$59,$3F,$DC,$63,$4F,$F6,$D7,$9E,$DC,$26,$4E,$5E,$51,$88,$E5
	.dc.b	$1D,$4E,$F1,$9C,$3F,$B9,$57,$6A,$DC,$FF,$71,$FA,$CF,$E5,$27,$F3
	.dc.b	$32,$6D,$FC,$9D,$CB,$FC,$D5,$CE,$FC,$63,$46,$48,$E4,$D3,$2B,$FC
	.dc.b	$15,$3D,$27,$11,$FE,$13,$8E,$AE,$A7,$A9,$AC,$3F,$C1,$93,$FD,$03
	.dc.b	$F1,$8F,$EE,$7F,$8C,$7F,$85,$CF,$F0,$C3,$FD,$D8,$CF,$07,$F8,$56
	.dc.b	$33,$87,$F7,$49,$CC,$FF,$84,$E7,$33,$FE,$15,$9F,$F0,$71,$1F,$F0
	.dc.b	$73,$5F,$F0,$73,$5F,$F0,$73,$5F,$CC,$35,$FF,$0E,$B9,$BF,$F1,$33
	.dc.b	$E7,$8F,$21,$FC,$A9,$C3,$E3,$33,$AD,$61,$FE,$35,$8E,$51,$FE,$2D
	.dc.b	$FC,$67,$FC,$1A,$4D,$BF,$C1,$AF,$F8,$2C,$C3,$B3,$27,$E5,$AF,$FC
	.dc.b	$75,$1A,$BF,$F2,$13,$CF,$5E,$15,$27,$2F,$28,$C4,$7F,$93,$E7,$93
	.dc.b	$67,$B6,$FF,$E3,$E9,$CD,$F8,$27,$95,$7F,$94,$7F,$1A,$FE,$FE,$7F
	.dc.b	$82,$49,$B7,$F7,$D7,$2F,$F8,$2D,$CE,$FF,$E4,$F7,$31,$4D,$1C,$48
	.dc.b	$64,$D4,$49,$F1,$47,$03,$0F,$F9,$AE,$7A,$BA,$8C,$DF,$F9,$8F,$EB
	.dc.b	$FF,$C9,$3F,$61,$FE,$6D,$3F,$CD,$CF,$F9,$30,$CF,$07,$F9,$AE,$33
	.dc.b	$7F,$E4,$A9,$CC,$FF,$9A,$67,$33,$FE,$6B,$9F,$F3,$21,$1F,$F3,$23
	.dc.b	$5F,$F3,$23,$5F,$F3,$23,$5F,$E5,$CD,$7F,$CE,$6E,$6B,$D5,$1E,$32
	.dc.b	$6D,$E1,$70,$53,$0D,$7F,$D1,$53,$EF,$78,$9F,$E8,$DF,$C6,$FF,$CC
	.dc.b	$64,$D3,$FC,$C6,$FF,$98,$4C,$3F,$E8,$97,$5B,$D8,$D9,$0C,$39,$64
	.dc.b	$F7,$77,$18,$8F,$F5,$24,$F6,$75,$19,$C3,$FD,$39,$39,$9D,$D7,$E6
	.dc.b	$B9,$1F,$8E,$7F,$95,$0F,$F2,$F1,$35,$FF,$29,$39,$7F,$CC,$0E,$77
	.dc.b	$FF,$52,$39,$8A,$68,$E2,$43,$26,$A2,$5F,$FA,$DE,$7C,$13,$88,$FF
	.dc.b	$60,$47,$57,$51,$D6,$7F,$D6,$9F,$8E,$FF,$A7,$BF,$1E,$FF,$61,$1F
	.dc.b	$EC,$37,$FD,$42,$67,$83,$FD,$81,$19,$BF,$F4,$FC,$E6,$7F,$D7,$D3
	.dc.b	$99,$EA,$7F,$AD,$E7,$FD,$74,$47,$04,$3F,$D7,$4E,$5A,$64,$FF,$5E
	.dc.b	$39,$69,$E2,$21,$E0,$DF,$C2,$C3,$FD,$B9,$1E,$C7,$FC,$B2,$D7,$47
	.dc.b	$FB,$9D,$FF,$03,$71,$FF,$B5,$3E,$D3,$FD,$72,$4D,$7F,$D7,$2F,$FA
	.dc.b	$DC,$C3,$FE,$DF,$75,$C1,$63,$64,$30,$E5,$97,$FE,$F1,$9E,$2D,$8E
	.dc.b	$0D,$78,$B4,$87,$FB,$FE,$38,$BF,$F0,$07,$F1,$FF,$F5,$49,$FE,$B3
	.dc.b	$26,$BF,$EA,$77,$2F,$FA,$D5,$CE,$BF,$EF,$97,$31,$4D,$1C,$48,$64
	.dc.b	$D4,$49,$1E,$38,$E2,$65,$C4,$71,$77,$C2,$C9,$C0,$C3,$CB,$5A,$9C
	.dc.b	$FE,$41,$FE,$F8,$FC,$85,$C3,$2B,$58,$27,$33,$E4,$96,$3D,$E9,$79
	.dc.b	$9E,$2F,$5A,$E9,$1C,$30,$F2,$E1,$D2,$36,$71,$C3,$A5,$F1,$BF,$04
	.dc.b	$77,$AE,$15,$CF,$0F,$0B,$C7,$0F,$6B,$E3,$3E,$5C,$69,$C3,$5F,$43
	.dc.b	$5F,$F3,$66,$1C,$5E,$2D,$7A,$9C,$9C,$47,$18,$D9,$C5,$C3,$C0,$EF
	.dc.b	$1E,$5B,$F7,$C7,$A6,$7C,$51,$8E,$0F,$83,$7C,$6B,$E0,$5F,$18,$C5
	.dc.b	$6E,$7E,$59,$C2,$F9,$CE,$8F,$8B,$79,$FC,$89,$BC,$BF,$C0,$C6,$23
	.dc.b	$46,$63,$E8,$47,$0E,$0C,$CC,$9E,$2E,$F1,$FC,$35,$6E,$73,$3B,$79
	.dc.b	$BB,$C3,$F3,$33,$B9,$CC,$EE,$B9,$BB,$C5,$5B,$9C,$CE,$EF,$9B,$B9
	.dc.b	$D4,$62,$35,$5C,$9D,$1F,$D6,$7E,$E1,$C0,$E0,$F9,$91,$88,$A6,$B2
	.dc.b	$FD,$4C,$70,$9C,$9A,$99,$3A,$7A,$36,$32,$FE,$76,$7B,$39,$21,$99
	.dc.b	$F0,$35,$EE,$EB,$2F,$DB,$4F,$28,$D1,$9C,$7F,$83,$C7,$9E,$3C,$95
	.dc.b	$E2,$AE,$2F,$9A,$3F,$61,$F9,$1F,$E4,$72,$7E,$49,$FB,$4B,$FE,$08
	.dc.b	$DC,$7F,$89,$D6,$CE,$11,$9A,$E3,$C3,$FC,$1A,$FF,$86,$56,$F0,$FD
	.dc.b	$25,$EC,$E1,$19,$8F,$DD,$BF,$B2,$23,$FB,$16,$BA,$30,$43,$04,$3F
	.dc.b	$81,$BC,$1F,$E4,$6F,$76,$8D,$FF,$5D,$19,$8F,$8A,$7E,$ED,$FE,$CB
	.dc.b	$81,$D5,$FB,$A7,$47,$33,$F9,$8F,$E4,$BF,$B1,$6E,$DF,$9F,$8C,$47
	.dc.b	$F2,$17,$F3,$76,$73,$5F,$CD,$5B,$9F,$F4,$46,$CE,$EF,$FC,$6D,$18
	.dc.b	$23,$DC,$C9,$4F,$E1,$8E,$B8,$2C,$67,$0F,$3B,$88,$CB,$23,$27,$B2
	.dc.b	$3F,$8B,$7E,$4D,$F9,$18,$FE,$78,$FE,$5F,$1F,$E9,$CD,$C7,$FA,$EB
	.dc.b	$B8,$CE,$DF,$95,$6B,$D0,$FD,$2B,$AE,$0C,$19,$89,$7F,$D8,$2B,$A5
	.dc.b	$F6,$9E,$8D,$77,$3E,$67,$39,$AB,$8B,$63,$C3,$5C,$6B,$B4,$70,$AD
	.dc.b	$9E,$DA,$E1,$1E,$69,$FE,$E7,$F7,$7F,$AF,$FB,$CF,$D4,$1F,$A9,$7F
	.dc.b	$64,$67,$87,$F5,$58,$FF,$80,$A7,$55,$FD,$36,$B5,$5F,$D1,$67,$33
	.dc.b	$FD,$1E,$7F,$AD,$91,$FE,$B4,$D7,$FB,$4B,$5F,$ED,$2D,$7F,$C0,$D8
	.dc.b	$23,$11,$FD,$C9,$D6,$7F,$E0,$97,$D3,$1E,$53,$F4,$47,$F7,$D7,$2F
	.dc.b	$94,$D7,$05,$61,$F1,$8C,$9F,$E1,$AF,$A7,$FE,$EA,$47,$FB,$A3,$5F
	.dc.b	$F0,$1B,$5F,$F0,$1B,$5F,$5B,$5F,$F0,$5B,$FE,$03,$FD,$47,$FC,$09
	.dc.b	$06,$91,$D1,$0F,$F0,$63,$96,$99,$3F,$C3,$0E,$5B,$1D,$C8,$F2,$6B
	.dc.b	$A3,$04,$30,$43,$FC,$6D,$08,$E3,$18,$FC,$5B,$FC,$19,$2F,$33,$F9
	.dc.b	$D3,$74,$FC,$6A,$9D,$9C,$27,$67,$07,$35,$FE,$35,$9E,$55,$FE,$1C
	.dc.b	$6C,$7F,$95,$5E,$8E,$E7,$E7,$11,$F9,$2D,$7E,$3B,$5F,$51,$A2,$1A
	.dc.b	$29,$A3,$C4,$F9,$9A,$7E,$4B,$2F,$F3,$64,$F9,$23,$24,$3D,$13,$FE
	.dc.b	$A2,$8D,$46,$0F,$51,$31,$EF,$8F,$B7,$EA,$BE,$5D,$9B,$1F,$04,$D6
	.dc.b	$3E,$68,$CD,$F2,$31,$11,$F1,$CF,$17,$E2,$BF,$20,$8E,$08,$7E,$05
	.dc.b	$CB,$4C,$9F,$E8,$1C,$3F,$01,$1A,$8C,$5F,$E2,$DF,$21,$E2,$3D,$F8
	.dc.b	$61,$FF,$41,$57,$27,$18,$FC,$EA,$FC,$A5,$63,$6F,$D6,$47,$5A,$E8
	.dc.b	$7B,$A7,$F1,$58,$FE,$48,$F9,$67,$E5,$8F,$CD,$99,$3E,$F9,$98,$FC
	.dc.b	$A3,$97,$ED,$B9,$7A,$9F,$B8,$7F,$80,$11,$C1,$0F,$DC,$B9,$69,$93
	.dc.b	$F6,$2E,$5A,$78,$88,$78,$23,$C2,$C3,$F4,$D1,$EC,$7F,$2D,$5F,$C4
	.dc.b	$63,$11,$B3,$E7,$BF,$95,$3E,$53,$8C,$BE,$31,$F1,$F2,$F8,$67,$C4
	.dc.b	$73,$3F,$3A,$33,$3F,$42,$72,$F9,$23,$2F,$B9,$F9,$04,$7E,$2B,$5D
	.dc.b	$18,$21,$82,$1F,$91,$9F,$4B,$FE,$E6,$CB,$FD,$B3,$3E,$08,$F7,$BE
	.dc.b	$AC,$3D,$11,$0F,$31,$EB,$70,$F0,$2E,$5F,$C0,$2F,$F2,$71,$FE,$E9
	.dc.b	$FD,$47,$FD,$89,$39,$9F,$C7,$1F,$DC,$23,$39,$1B,$8F,$FB,$AA,$7F
	.dc.b	$2E,$6E,$FF,$D7,$53,$A8,$FE,$5E,$7F,$9A,$5E,$67,$F4,$E2,$47,$FB
	.dc.b	$4D,$AF,$FB,$5D,$AF,$FB,$85,$AF,$FB,$D5,$AF,$FB,$ED,$B6,$BF,$44
	.dc.b	$F9,$08,$7E,$3A,$7C,$6E,$0F,$F6,$AA,$66,$18,$74,$7C,$93,$9B,$9F
	.dc.b	$43,$FD,$F1,$3C,$7E,$E7,$E2,$1F,$16,$33,$46,$C6,$B1,$E5,$30,$ED
	.dc.b	$19,$7C,$13,$C8,$F8,$CF,$33,$E3,$14,$F8,$C5,$3E,$31,$4F,$75,$7B
	.dc.b	$E3,$99,$23,$9B,$5E,$6D,$79,$BE,$37,$D0,$31,$FA,$53,$D7,$0C,$C6
	.dc.b	$27,$C8,$DA,$F6,$4E,$B1,$FB,$31,$D6,$BE,$0C,$FD,$3C,$FE,$5E,$9F
	.dc.b	$91,$3F,$24,$FC,$E3,$39,$FD,$98,$CD,$71,$31,$11,$FB,$93,$C5,$F9
	.dc.b	$6F,$CE,$23,$82,$1F,$39,$CB,$4C,$9F,$41,$CB,$4F,$11,$E0,$3D,$95
	.dc.b	$87,$E5,$61,$F9,$A9,$F2,$47,$9A,$75,$8F,$CB,$9C,$CF,$0C,$BF,$35
	.dc.b	$1F,$9F,$87,$C7,$7E,$4B,$E3,$FF,$70,$7E,$E5,$FC,$A9,$9C,$BF,$41
	.dc.b	$19,$9F,$D1,$4E,$5F,$D1,$46,$5F,$D2,$47,$E5,$08,$FE,$51,$AF,$E5
	.dc.b	$1A,$FE,$51,$AF,$E5,$0F,$39,$C4,$87,$13,$C6,$43,$C3,$8D,$7E,$11
	.dc.b	$C8,$F2,$9E,$A7,$58,$FE,$4E,$7E,$0C,$E4,$FE,$4D,$3F,$C2,$6F,$F8
	.dc.b	$F1,$FC,$95,$FE,$3E,$65,$FE,$0C,$66,$3F,$81,$C6,$5F,$E1,$11,$97
	.dc.b	$F8,$7C,$7F,$21,$23,$FC,$85,$AF,$F2,$16,$BF,$C8,$5A,$FF,$21,$6B
	.dc.b	$FB,$C7,$35,$FB,$27,$CF,$1E,$43,$F9,$13,$87,$C6,$65,$AC,$3E,$19
	.dc.b	$93,$FA,$56,$7F,$D0,$3F,$2A,$FE,$7C,$7F,$3F,$7F,$A1,$99,$BF,$E9
	.dc.b	$91,$99,$FE,$4F,$39,$9F,$E6,$33,$99,$FE,$67,$3F,$D1,$48,$FF,$45
	.dc.b	$6B,$FD,$15,$AF,$F4,$56,$BF,$D1,$5A,$FF,$20,$73,$7F,$C5,$9F,$3C
	.dc.b	$79,$0F,$E8,$CE,$1F,$19,$9B,$AC,$3E,$19,$93,$E2,$EB,$EE,$DB,$FB
	.dc.b	$65,$1B,$1F,$08,$D6,$3F,$02,$33,$5F,$0A,$72,$FD,$08,$CB,$D8,$F8
	.dc.b	$51,$F1,$48,$E0,$87,$C5,$72,$D3,$27,$C7,$72,$D3,$C4,$43,$DB,$75
	.dc.b	$FE,$21,$7F,$CF,$36,$FF,$04,$37,$6F,$F4,$25,$FC,$C3,$E0,$57,$1C
	.dc.b	$BE,$94,$7D,$0D,$BF,$0B,$7F,$20,$F9,$2F,$E2,$0C,$E7,$F8,$68,$CD
	.dc.b	$FE,$22,$72,$FE,$5A,$32,$FE,$62,$3F,$14,$47,$F1,$4D,$7F,$14,$D7
	.dc.b	$F1,$4D,$7F,$14,$D7,$DE,$EE,$FF,$18,$F9,$E3,$C9,$97,$95,$87,$8C
	.dc.b	$C9,$48,$7C,$C8,$F8,$11,$FA,$4D,$7F,$53,$B7,$E3,$0F,$C6,$BF,$AC
	.dc.b	$33,$8F,$EA,$A3,$37,$C4,$C4,$47,$F7,$73,$C5,$FD,$63,$FA,$F2,$38
	.dc.b	$21,$FA,$F7,$2D,$32,$7E,$C5,$CB,$4F,$11,$E0,$3D,$95,$87,$F1,$71
	.dc.b	$E3,$7C,$B1,$9C,$3F,$71,$39,$9D,$CF,$F8,$EB,$0F,$E2,$37,$FB,$03
	.dc.b	$FC,$B8,$FF,$95,$8C,$BF,$C3,$8C,$CF,$F8,$EA,$32,$FF,$1E,$8C,$BD
	.dc.b	$4F,$F2,$B4,$7F,$96,$48,$E0,$87,$F9,$65,$CB,$4C,$9F,$E5,$B7,$2D
	.dc.b	$3C,$44,$3C,$13,$E1,$61,$FE,$7B,$8F,$63,$FE,$5D,$6B,$A3,$FC,$08
	.dc.b	$FF,$97,$1E,$1C,$5E,$FE,$4F,$8A,$7C,$67,$E1,$19,$BF,$99,$19,$9F
	.dc.b	$9D,$39,$7C,$51,$97,$A9,$F0,$63,$E3,$91,$C1,$0F,$84,$E5,$A6,$4F
	.dc.b	$F3,$F5,$23,$51,$87,$CE,$F9,$0F,$2D,$D7,$C6,$43,$E1,$C7,$C7,$38
	.dc.b	$B9,$C7,$FC,$B3,$5D,$EB,$F1,$9A,$7E,$17,$7F,$C5,$9F,$8C,$7F,$10
	.dc.b	$67,$2F,$C9,$46,$70,$E2,$62,$23,$F6,$67,$8B,$F8,$77,$F1,$A4,$70
	.dc.b	$43,$F1,$0E,$5A,$64,$F8,$EE,$5A,$78,$8F,$01,$EC,$AC,$3F,$86,$8F
	.dc.b	$1B,$E5,$8C,$E1,$F1,$27,$33,$B9,$FD,$87,$DA,$7E,$A7,$ED,$7F,$5E
	.dc.b	$7E,$C1,$FD,$61,$9C,$3F,$6D,$19,$AE,$26,$22,$3F,$9D,$9E,$2F,$EA
	.dc.b	$DF,$D8,$91,$C1,$0F,$D6,$39,$69,$93,$F1,$6E,$5A,$78,$8F,$01,$EC
	.dc.b	$AC,$3F,$AA,$8F,$1B,$E5,$8C,$E1,$F8,$89,$CC,$EE,$7F,$8A,$65,$FC
	.dc.b	$37,$7F,$E2,$67,$F1,$47,$F8,$81,$9C,$3F,$8D,$46,$6B,$89,$88,$8F
	.dc.b	$F0,$19,$E2,$FF,$0F,$7F,$8A,$91,$C1,$0F,$E2,$0E,$5A,$64,$FD,$7B
	.dc.b	$96,$9E,$23,$C0,$7B,$2B,$0F,$F0,$E8,$F1,$BE,$58,$CE,$1F,$AC,$9C
	.dc.b	$CE,$E7,$FD,$E0,$FF,$A8,$3F,$45,$AC,$0D,$76,$D7,$FD,$F0,$D7,$66
	.dc.b	$5C,$11,$C1,$0E,$8E,$5B,$3C,$48,$E0,$87,$96,$11,$A8,$C3,$5F,$5C
	.dc.b	$69,$E2,$62,$3D,$5F,$77,$E9,$24,$60,$87,$C1,$72,$D3,$27,$C7,$72
	.dc.b	$D3,$AC,$47,$93,$5E,$0D,$7E,$8C,$79,$DC,$9C,$08,$60,$87,$D8,$FD
	.dc.b	$47,$E9,$11,$FA,$CD,$7E,$D3,$5F,$BC,$D7,$91,$C0,$87,$09,$E6,$47
	.dc.b	$04,$39,$B9,$69,$93,$CC,$E5,$A7,$93,$17,$3F,$E4,$B2,$72,$B0,$53
	.dc.b	$A2,$E0,$3F,$DF,$33,$A8,$CC,$62,$7A,$BF,$A0,$BE,$B5,$D3,$0F,$06
	.dc.b	$1D,$EF,$B1,$B7,$18,$74,$BE,$33,$D3,$0F,$14,$62,$F9,$E1,$E3,$AE
	.dc.b	$4D,$AD,$E1,$DE,$B7,$5C,$E7,$F6,$B5,$E5,$7E,$BC,$65,$F8,$11,$97
	.dc.b	$F0,$51,$97,$A1,$C2,$7E,$71,$1C,$10,$E2,$E5,$F2,$CF,$57,$E5,$47
	.dc.b	$98,$69,$EC,$AF,$46,$1F,$06,$B1,$5B,$AB,$10,$EB,$5C,$30,$F8,$53
	.dc.b	$C3,$2E,$3A,$CB,$DE,$1F,$A5,$C7,$E7,$56,$67,$CB,$5C,$5D,$C6,$67
	.dc.b	$EA,$4E,$67,$D7,$3F,$14,$8F,$C2,$6B,$F4,$B1,$FC,$AC,$62,$3E,$8C
	.dc.b	$61,$E0,$43,$85,$FE,$ED,$DB,$5F,$C0,$4F,$8E,$FD,$71,$EE,$FC,$6F
	.dc.b	$19,$6A,$3D,$66,$4A,$7A,$8B,$3F,$91,$9F,$A9,$87,$D7,$3E,$99,$89
	.dc.b	$87,$CC,$72,$FD,$47,$F2,$67,$17,$46,$4E,$27,$C5,$7F,$74,$47,$F2
	.dc.b	$ED,$7F,$2A,$D7,$81,$A2,$1A,$29,$FB,$97,$C6,$FB,$32,$F0,$D7,$EC
	.dc.b	$A7,$33,$E9,$9C,$DF,$EE,$E7,$33,$CE,$7E,$2E,$1F,$1B,$29,$3F,$06
	.dc.b	$32,$FC,$38,$CB,$E3,$8C,$BE,$58,$CE,$1E,$39,$CC,$FE,$1A,$FF,$01
	.dc.b	$9F,$F1,$A6,$9F,$C7,$27,$E3,$92,$7E,$3B,$5F,$9C,$D7,$A3,$5F,$73
	.dc.b	$E4,$AF,$14,$7D,$BC,$B1,$96,$A3,$DC,$64,$A7,$B4,$B3,$F3,$67,$F9
	.dc.b	$17,$EA,$3F,$E0,$38,$FA,$18,$7E,$38,$CD,$7F,$82,$63,$F9,$CC,$EE
	.dc.b	$FE,$95,$6E,$BF,$90,$CE,$4D,$63,$F1,$67,$52,$BE,$6F,$5B,$5F,$F8
	.dc.b	$44,$91,$FE,$15,$6B,$FE,$19,$6B,$FE,$1D,$6B,$FE,$20,$6B,$FE,$24
	.dc.b	$6C,$F8,$23,$FA,$2C,$79,$DF,$F5,$E1,$E5,$9F,$1B,$E5,$8D,$3C,$0D
	.dc.b	$CF,$37,$E3,$99,$26,$78,$E2,$CF,$1C,$9F,$8E,$7F,$5F,$8D,$38,$11
	.dc.b	$B7,$C9,$7F,$E3,$E9,$DC,$60,$FE,$D1,$3B,$BF,$74,$EA,$7F,$B6,$57
	.dc.b	$17,$E3,$1E,$12,$88,$F9,$8E,$4F,$94,$53,$99,$4F,$8E,$71,$39,$BC
	.dc.b	$9C,$1C,$8A,$72,$29,$C9,$F6,$CF,$AA,$3E,$27,$5E,$31,$D4,$7C,$7A
	.dc.b	$FA,$33,$9B,$F7,$4F,$C5,$3C,$37,$F8,$19,$FA,$FA,$FD,$8D,$A5,$7D
	.dc.b	$18,$CB,$F4,$E3,$2F,$38,$CB,$D6,$33,$5F,$8D,$9C,$CF,$B5,$F8,$44
	.dc.b	$7E,$7B,$5F,$88,$D7,$81,$A2,$1A,$2C,$77,$7D,$07,$E2,$76,$F2,$C6
	.dc.b	$63,$D0,$64,$D9,$82,$3F,$A5,$8F,$CD,$F3,$FE,$73,$A2,$57,$E5,$63
	.dc.b	$2F,$E5,$E3,$2F,$CF,$8C,$BF,$46,$33,$5F,$B1,$9C,$CF,$E0,$5F,$C3
	.dc.b	$11,$FC,$9B,$5F,$C4,$35,$E0,$68,$86,$8B,$1D,$DF,$41,$FA,$DF,$CB
	.dc.b	$3C,$B1,$98,$F4,$19,$36,$60,$8F,$F0,$98,$FE,$4B,$BF,$EF,$F8,$3F
	.dc.b	$80,$98,$A8,$7E,$ED,$CB,$FC,$02,$3F,$85,$1C,$5D,$19,$38,$9F,$B0
	.dc.b	$7F,$90,$3F,$E3,$63,$F6,$AE,$08,$E0,$87,$EF,$9C,$B4,$C9,$D4,$EF
	.dc.b	$5F,$CD,$23,$CA,$63,$7F,$E4,$6F,$A2,$7E,$0B,$9D,$BF,$94,$C7,$F1
	.dc.b	$38,$E2,$FF,$44,$D7,$FC,$F2,$FF,$1A,$7F,$D0,$B8,$7F,$52,$6C,$6E
	.dc.b	$3F,$9A,$C7,$0C,$3F,$A8,$BA,$9E,$5C,$5F,$E7,$9C,$7F,$A4,$CE,$DB
	.dc.b	$7C,$A7,$58,$AB,$96,$1D,$2B,$B9,$FD,$36,$3C,$F8,$7F,$28,$6B,$D0
	.dc.b	$FE,$A0,$E2,$BA,$4F,$12,$4E,$9A,$FF,$55,$9E,$04,$8F,$0C,$7C,$37
	.dc.b	$C7,$8D,$88,$7C,$53,$4E,$1F,$99,$1C,$9F,$95,$1C,$F8,$3A,$70,$C9
	.dc.b	$F9,$D1,$97,$E8,$46,$5E,$71,$97,$AC,$67,$1F,$91,$39,$9F,$8D,$7F
	.dc.b	$CE,$B1,$F8,$73,$F0,$49,$3F,$06,$75,$8F,$FB,$0A,$73,$3F,$DD,$E3
	.dc.b	$0E,$E3,$2F,$67,$77,$E2,$8F,$59,$F8,$9E,$1F,$44,$66,$3D,$66,$4D
	.dc.b	$98,$23,$F8,$38,$FA,$5C,$3F,$4E,$E5,$7E,$56,$32,$FE,$5E,$32,$FD
	.dc.b	$28,$CB,$F5,$23,$35,$F8,$F9,$CC,$FC,$87,$F0,$C4,$7E,$33,$5F,$89
	.dc.b	$87,$FA,$3F,$1F,$C7,$CE,$67,$91,$C0,$87,$07,$C7,$5E,$18,$FD,$77
	.dc.b	$0E,$36,$FD,$3D,$5B,$FC,$DC,$DC,$7F,$CB,$97,$F0,$4E,$27,$ED,$F8
	.dc.b	$3F,$31,$C3,$23,$04,$3F,$2A,$E5,$FC,$BC,$7E,$7C,$C6,$5F,$B3,$9C
	.dc.b	$4F,$13,$F1,$B3,$FA,$B2,$3F,$8B,$6B,$F1,$9A,$F0,$34,$43,$45,$9F
	.dc.b	$E0,$51,$E0,$8F,$D5,$F0,$D9,$CC,$7A,$DF,$E1,$11,$9C,$FF,$87,$D5
	.dc.b	$CB,$F8,$1C,$FF,$9B,$B3,$FF,$25,$65,$FE,$87,$C7,$F8,$F6,$99,$7F
	.dc.b	$CA,$73,$A8,$FE,$2F,$3B,$9F,$F2,$0C,$FF,$94,$CD,$4F,$F0,$73,$94
	.dc.b	$43,$FC,$C5,$1F,$E6,$92,$3F,$E6,$96,$BF,$E6,$96,$BC,$0D,$10,$D1
	.dc.b	$4F,$F3,$FB,$E3,$7D,$27,$F5,$53,$CE,$60,$F2,$18,$E0,$B9,$FF,$A3
	.dc.b	$27,$94,$FC,$39,$F8,$9D,$1F,$17,$A6,$4F,$80,$46,$9C,$46,$9F,$0C
	.dc.b	$65,$F1,$C6,$6F,$C1,$1C,$8E,$93,$EE,$23,$F0,$DA,$FF,$A2,$EC,$F1
	.dc.b	$30,$53,$FC,$DE,$EA,$38,$C6,$0D,$1F,$EA,$08,$D3,$DC,$87,$A8,$F2
	.dc.b	$9F,$07,$F1,$DF,$65,$7A,$AA,$ED,$CE,$95,$8D,$58,$76,$9E,$87,$D6
	.dc.b	$9F,$C9,$71,$7E,$4F,$8E,$4F,$CF,$8C,$BF,$46,$32,$FD,$E8,$CB,$F8
	.dc.b	$08,$CE,$5F,$3A,$73,$3F,$05,$FC,$59,$1F,$C8,$35,$F9,$6D,$78,$1A
	.dc.b	$21,$A2,$C7,$77,$D0,$7E,$1F,$8F,$CB,$19,$8F,$41,$93,$66,$08,$FE
	.dc.b	$1A,$3F,$6D,$D3,$FB,$7E,$A9,$5F,$93,$8C,$BF,$95,$8C,$BF,$A2,$8C
	.dc.b	$BF,$A4,$8C,$D7,$E4,$A7,$33,$F8,$57,$F5,$E4,$7F,$68,$D7,$F1,$ED
	.dc.b	$78,$1A,$21,$A2,$C7,$77,$D0,$7E,$AF,$F3,$0F,$2C,$66,$3D,$06,$4D
	.dc.b	$98,$23,$FA,$A8,$FE,$35,$9F,$F1,$BE,$29,$5F,$B7,$8C,$BF,$B9,$8C
	.dc.b	$BF,$C1,$A3,$2F,$F0,$88,$CD,$7E,$DA,$73,$3F,$A9,$7F,$89,$91,$FE
	.dc.b	$30,$D7,$F6,$6D,$78,$1A,$21,$A2,$C7,$77,$D0,$7F,$0F,$E6,$F2,$C6
	.dc.b	$63,$D0,$64,$D9,$82,$3F,$C3,$A3,$F8,$3C,$21,$BB,$F4,$0C,$CC,$30
	.dc.b	$53,$F8,$7B,$5A,$57,$0F,$03,$83,$82,$3B,$21,$A3,$95,$C7,$8B,$5F
	.dc.b	$24,$7F,$0A,$72,$43,$31,$B7,$D7,$1B,$72,$6C,$CF,$C7,$3F,$32,$78
	.dc.b	$5F,$1B,$1F,$8B,$6A,$D7,$E4,$B9,$CB,$E4,$46,$31,$FE,$67,$7C,$23
	.dc.b	$A6,$1B,$C3,$96,$3D,$AF,$78,$72,$86,$3B,$C3,$A5,$7D,$7A,$D6,$19
	.dc.b	$BE,$53,$9C,$3A,$E1,$93,$55,$27,$BC,$E9,$B8,$76,$C7,$D1,$87,$46
	.dc.b	$DF,$3A,$DC,$F9,$9B,$1D,$5F,$15,$73,$C7,$D9,$8E,$70,$FC,$7F,$E9
	.dc.b	$1F,$86,$25,$FE,$2D,$AF,$E3,$DA,$FE,$4D,$AF,$E5,$9A,$FE,$71,$E9
	.dc.b	$9F,$32,$47,$36,$BC,$DA,$F3,$6B,$CD,$AF,$F6,$E8,$FC,$53,$FA,$6F
	.dc.b	$C1,$7F,$3E,$47,$F4,$8D,$7F,$50,$D7,$F5,$8D,$79,$1C,$08,$70,$9E
	.dc.b	$64,$70,$43,$9B,$96,$99,$3C,$CE,$5A,$79,$3A,$83,$BB,$F3,$95,$88
	.dc.b	$78,$18,$76,$D7,$85,$7E,$6A,$78,$56,$6A,$4E,$6B,$BC,$F1,$33,$1F
	.dc.b	$9D,$AF,$05,$66,$38,$57,$67,$F8,$CC,$F6,$7E,$94,$74,$6B,$B7,$FC
	.dc.b	$4B,$1B,$78,$C6,$DA,$ED,$46,$DE,$31,$B6,$BB,$8D,$47,$08,$D3,$5F
	.dc.b	$46,$3F,$CB,$A7,$C4,$D8,$F1,$35,$FA,$73,$F0,$2F,$E1,$3F,$16,$37
	.dc.b	$9F,$57,$35,$C6,$AC,$E6,$B9,$10,$E5,$5E,$48,$D4,$62,$78,$DF,$9A
	.dc.b	$73,$7D,$6B,$93,$5E,$77,$C0,$F2,$D7,$13,$F1,$4E,$4F,$03,$97,$FD
	.dc.b	$19,$3C,$C9,$5F,$26,$FF,$1D,$32,$FC,$33,$97,$F0,$B1,$97,$D4,$6A
	.dc.b	$FC,$2E,$8C,$47,$98,$F0,$CF,$B8,$E9,$1C,$8F,$69,$C4,$C5,$F9,$AB
	.dc.b	$AC,$68,$CC,$F2,$9F,$B9,$5E,$59,$FC,$4B,$FC,$08,$FF,$56,$38,$3F
	.dc.b	$32,$E4,$FC,$1B,$F9,$67,$04,$76,$43,$45,$8F,$0B,$E9,$7F,$BF,$FE
	.dc.b	$A5,$FD,$ED,$CC,$7A,$5A,$69,$FF,$01,$38,$AF,$8D,$7F,$02,$35,$3F
	.dc.b	$E0,$87,$51,$99,$F9,$95,$DD,$E3,$7F,$42,$75,$39,$78,$12,$3E,$AC
	.dc.b	$71,$34,$F1,$33,$7E,$6A,$4E,$A7,$E4,$DF,$A6,$39,$56,$EF,$B5,$6A
	.dc.b	$F8,$CF,$3A,$E5,$3F,$8D,$BF,$7C,$FF,$80,$67,$2F,$CD,$AE,$2F,$C3
	.dc.b	$6B,$C5,$F8,$6F,$53,$C5,$5E,$B3,$C5,$3E,$68,$E5,$7E,$69,$F0,$47
	.dc.b	$13,$2F,$13,$2F,$86,$7F,$C6,$64,$7F,$C7,$0D,$7F,$C7,$8D,$7F,$C8
	.dc.b	$0D,$7F,$C8,$6D,$7F,$C9,$2D,$BF,$87,$1F,$E1,$A8,$F4,$76,$F8,$A2
	.dc.b	$19,$3D,$0E,$1F,$F0,$E4,$7B,$DF,$4C,$66,$7A,$C6,$8E,$27,$33,$DC
	.dc.b	$7C,$33,$89,$93,$96,$1F,$49,$E1,$38,$AF,$04,$60,$E0,$43,$C0,$76
	.dc.b	$11,$C5,$F0,$1C,$4D,$99,$38,$35,$C3,$5E,$D5,$FA,$A8,$C4,$72,$3A
	.dc.b	$47,$C2,$25,$7A,$E3,$CB,$3C,$27,$35,$FE,$7E,$8C,$56,$E7,$B5,$78
	.dc.b	$23,$04,$AE,$11,$97,$E2,$BD,$8F,$1C,$6D,$FA,$66,$4F,$41,$27,$BB
	.dc.b	$97,$CA,$F0,$72,$61,$F8,$E7,$69,$F0,$C6,$A7,$F5,$B5,$EC,$9F,$3D
	.dc.b	$75,$9F,$23,$C8,$C4,$72,$8E,$31,$D4,$FB,$26,$81,$F7,$8F,$84,$79
	.dc.b	$DE,$2F,$B5,$EF,$3F,$90,$8C,$4F,$E8,$9E,$C7,$E2,$4D,$57,$AE,$3C
	.dc.b	$87,$A9,$C4,$79,$1A,$F8,$CF,$D8,$9E,$69,$FD,$BB,$4F,$DB,$94,$D9
	.dc.b	$C4,$F6,$CF,$C8,$69,$B2,$9B,$3D,$E6,$23,$BB,$98,$E4,$FC,$49,$F3
	.dc.b	$CF,$F1,$27,$DF,$59,$25,$7F,$AB,$6B,$E2,$D6,$A3,$11,$FC,$13,$8F
	.dc.b	$FA,$09,$FE,$20,$83,$4C,$E1,$FC,$A1,$C3,$FE,$1D,$8D,$B9,$26,$5F
	.dc.b	$CE,$67,$71,$4F,$ED,$F5,$F1,$A3,$FB,$AB,$FD,$31,$B5,$D2,$7F,$93
	.dc.b	$C6,$08,$F2,$9E,$0D,$9E,$11,$FC,$B2,$77,$3C,$8D,$3E,$C7,$C8,$7F
	.dc.b	$4F,$D3,$FA,$A4,$66,$7F,$98,$3F,$D9,$A3,$33,$EF,$38,$CF,$F3,$58
	.dc.b	$CC,$62,$3E,$73,$FD,$42,$33,$1D,$23,$93,$C4,$FE,$99,$3E,$17,$8C
	.dc.b	$79,$5C,$BC,$1C,$9C,$4F,$06,$5F,$E1,$29,$E4,$FF,$85,$2B,$C3,$3C
	.dc.b	$09,$19,$7A,$D7,$2C,$BE,$6B,$E1,$8C,$61,$F3,$9E,$33,$FE,$04,$BE
	.dc.b	$B7,$E1,$BF,$33,$D6,$B8,$E1,$E4,$BE,$11,$E0,$3E,$25,$76,$BE,$73
	.dc.b	$E2,$9C,$9E,$F2,$1E,$2C,$3C,$98,$77,$8C,$BE,$3B,$E9,$1F,$DE,$6F
	.dc.b	$AD,$70,$BF,$1C,$FB,$CE,$E7,$59,$FF,$0C,$1C,$30,$FC,$24,$73,$3D
	.dc.b	$46,$CC,$1F,$02,$FF,$04,$E0,$EB,$87,$95,$F3,$DF,$89,$CB,$F0,$5F
	.dc.b	$EF,$77,$E2,$7B,$CF,$17,$27,$53,$E2,$5F,$99,$F6,$18,$BF,$1B,$89
	.dc.b	$E4,$4B,$F1,$47,$43,$D5,$3B,$8C,$CE,$F0,$F9,$47,$FB,$EA,$B8,$1E
	.dc.b	$A7,$38,$7A,$63,$99,$F0,$23,$84,$F3,$C3,$D1,$3F,$78,$FE,$5B,$5F
	.dc.b	$E2,$E2,$3F,$E2,$E6,$BF,$E2,$E6,$BF,$E2,$E6,$BF,$E2,$E6,$BD,$0F
	.dc.b	$91,$3F,$89,$7E,$5E,$5E,$FB,$F2,$E1,$F0,$0F,$CC,$C6,$6F,$D9,$5F
	.dc.b	$BB,$8C,$3D,$27,$6F,$BE,$76,$F6,$30,$6E,$91,$FD,$3E,$FC,$84,$AF
	.dc.b	$14,$EA,$BF,$AD,$4E,$63,$CE,$4A,$E5,$5C,$1F,$0C,$FF,$01,$BF,$3D
	.dc.b	$FA,$1A,$F9,$1A,$FD,$C3,$D0,$58,$F0,$4F,$E5,$67,$71,$89,$F2,$D7
	.dc.b	$85,$F7,$57,$92,$FC,$D1,$83,$6E,$1F,$21,$F1,$8A,$78,$67,$A3,$FC
	.dc.b	$9E,$FC,$31,$DD,$B3,$D9,$B5,$E7,$AE,$87,$BC,$A7,$59,$F9,$D3,$C2
	.dc.b	$3F,$80,$35,$F0,$1C,$2F,$E0,$46,$63,$F0,$86,$6F,$C8,$DA,$F2,$36
	.dc.b	$FC,$8F,$85,$FA,$F3,$F8,$A8,$C1,$F1,$EB,$B3,$F5,$2F,$E8,$3F,$19
	.dc.b	$AF,$B4,$D4,$FC,$87,$CA,$62,$FE,$83,$89,$FC,$33,$E5,$3E,$DF,$E8
	.dc.b	$BF,$E3,$C6,$15,$AE,$1A,$E4,$A1,$AD,$32,$E0,$8E,$08,$74,$75,$87
	.dc.b	$02,$46,$5A,$F9,$23,$F8,$B3,$FE,$54,$6B,$E3,$8D,$BE,$96,$BE,$D7
	.dc.b	$D3,$FA,$9F,$A0,$91,$82,$1F,$01,$CB,$4C,$9F,$19,$CB,$4E,$B1,$1E
	.dc.b	$4D,$78,$35,$FA,$31,$EA,$24,$66,$3E,$B4,$61,$FA,$7F,$AA,$7D,$02
	.dc.b	$3F,$51,$AE,$8C,$10,$C1,$0F,$BE,$F0,$6C,$F3,$23,$E0,$6B,$A3,$04
	.dc.b	$30,$43,$F1,$B1,$F5,$8F,$AF,$F8,$4F,$E3,$30,$64,$96,$B0,$FC,$B4
	.dc.b	$EA,$31,$18,$95,$73,$6D,$FE,$6A,$F8,$57,$38,$FF,$4D,$C6,$23,$BC
	.dc.b	$62,$78,$BF,$A4,$9F,$2D,$75,$BF,$0C,$ED,$AA,$76,$F8,$6B,$9B,$E7
	.dc.b	$9D,$CF,$86,$FD,$57,$E9,$3B,$3E,$7A,$E9,$7F,$16,$31,$5D,$A7,$CD
	.dc.b	$3D,$9F,$24,$EA,$38,$3E,$E8,$E7,$5D,$DF,$35,$F9,$2B,$75,$E4,$9F
	.dc.b	$55,$79,$23,$95,$FA,$67,$E4,$4F,$37,$C7,$7C,$CE,$91,$98,$F8,$55
	.dc.b	$E5,$9C,$C7,$63,$C7,$19,$3D,$53,$C6,$3D,$B1,$E2,$BF,$3C,$78,$8E
	.dc.b	$47,$8A,$39,$1B,$9E,$51,$F3,$2B,$CD,$3D,$A3,$89,$E6,$7B,$47,$9A
	.dc.b	$3E,$7C,$FC,$39,$D1,$F1,$E3,$27,$A9,$E2,$75,$72,$79,$5C,$9E,$57
	.dc.b	$C2,$7B,$9E,$AF,$B4,$E2,$60,$F5,$3D,$5F,$11,$9B,$FA,$CF,$E0,$0E
	.dc.b	$A7,$08,$F0,$9C,$1E,$C7,$1B,$F9,$AF,$13,$C2,$79,$48,$64,$E4,$75
	.dc.b	$8E,$55,$F4,$2B,$C6,$4A,$FC,$C5,$71,$39,$C6,$67,$84,$F8,$A3,$13
	.dc.b	$EA,$79,$5F,$D5,$8F,$0C,$62,$FE,$24,$7B,$0E,$B5,$CC,$F0,$CF,$EA
	.dc.b	$E3,$E0,$57,$37,$81,$EB,$38,$DF,$92,$3F,$13,$7F,$66,$7E,$29,$B2
	.dc.b	$57,$57,$9B,$E5,$8E,$11,$E8,$BF,$25,$78,$63,$E2,$1C,$A3,$D1,$7E
	.dc.b	$1A,$FC,$E9,$C5,$FE,$E6,$49,$FE,$EE,$D7,$FB,$DB,$5F,$EF,$ED,$7F
	.dc.b	$C0,$6D,$7F,$C0,$CD,$BF,$54,$79,$1F,$E2,$5A,$78,$63,$EA,$5F,$F2
	.dc.b	$A8,$C4,$7C,$49,$DB,$F0,$A7,$E8,$92,$BF,$C2,$95,$C2,$78,$B5,$E1
	.dc.b	$1A,$69,$E0,$70,$F9,$67,$A3,$FC,$12,$79,$39,$9D,$D7,$B0,$95,$C2
	.dc.b	$7F,$CC,$33,$F9,$23,$F8,$F5,$72,$AD,$C7,$2A,$E1,$5E,$83,$CA,$7C
	.dc.b	$17,$E2,$38,$34,$78,$CF,$51,$FE,$80,$8F,$03,$E9,$38,$1D,$2B,$C5
	.dc.b	$3E,$76,$CF,$9A,$76,$75,$3D,$67,$63,$F8,$9C,$70,$7B,$1B,$3C,$B5
	.dc.b	$EE,$78,$46,$DE,$27,$68,$D3,$D8,$F8,$13,$E3,$7A,$1E,$93,$85,$7A
	.dc.b	$5B,$1B,$9E,$71,$E2,$9F,$9A,$F1,$39,$9E,$DA,$F2,$3F,$CC,$DA,$FF
	.dc.b	$29,$6B,$FC,$6D,$B3,$F9,$48,$E6,$FA,$A3,$FB,$AE,$3F,$D8,$B5,$FE
	.dc.b	$A6,$7F,$54,$7F,$B2,$99,$C7,$FA,$74,$66,$FF,$A3,$CE,$67,$FB,$FC
	.dc.b	$E6,$7A,$9F,$D3,$1E,$B1,$FD,$5E,$7A,$CE,$63,$A9,$97,$A9,$97,$A9
	.dc.b	$97,$A9,$86,$3F,$DE,$23,$AB,$83,$FC,$0B,$3E,$57,$47,$53,$FB,$FB
	.dc.b	$97,$FC,$17,$1E,$D7,$33,$E1,$78,$9D,$6F,$FC,$57,$5C,$5F,$F1,$6D
	.dc.b	$71,$9F,$0D,$71,$7A,$9E,$13,$2F,$B1,$EA,$78,$1E,$D8,$7F,$91,$27
	.dc.b	$DB,$39,$21,$D0,$E4,$74,$33,$5E,$47,$E6,$9D,$31,$61,$E9,$BE,$D3
	.dc.b	$E5,$AF,$6E,$1E,$97,$94,$71,$AE,$17,$FF,$00,$63,$15,$BB,$E7,$18
	.dc.b	$8F,$2D,$F2,$8F,$2C,$FF,$98,$2F,$9E,$1F,$26,$BC,$37,$C2,$B9,$DF
	.dc.b	$0C,$3D,$33,$DD,$F1,$46,$70,$E6,$FA,$EB,$B5,$6F,$0E,$D1,$88,$ED
	.dc.b	$7C,$AB,$E5,$13,$0F,$43,$F3,$CE,$D3,$CD,$ED,$3E,$48,$E7,$7D,$A3
	.dc.b	$13,$F7,$8E,$47,$5B,$F1,$39,$BE,$E7,$C2,$7E,$05,$7E,$09,$E1,$5D
	.dc.b	$1E,$73,$C2,$B9,$C6,$AB,$84,$F3,$8C,$1F,$84,$3D,$77,$E8,$7D,$78
	.dc.b	$7C,$A7,$ED,$1E,$5B,$FA,$6F,$F6,$CB,$FE,$EA,$47,$FB,$AB,$5F,$EE
	.dc.b	$AD,$7F,$BA,$B5,$E4,$70,$21,$E3,$3F,$86,$E3,$5F,$16,$29,$FE,$9C
	.dc.b	$7B,$AB,$CB,$8F,$E8,$6F,$85,$FC,$69,$FC,$FC,$62,$7E,$45,$F3,$AF
	.dc.b	$EC,$F7,$CE,$3F,$09,$3C,$E7,$C3,$39,$3C,$2F,$CB,$3B,$4F,$3A,$ED
	.dc.b	$7E,$43,$C3,$3C,$8E,$17,$D5,$E3,$7D,$5F,$19,$A9,$E5,$7D,$63,$91
	.dc.b	$F6,$6F,$BC,$7A,$27,$BC,$F4,$39,$47,$17,$07,$13,$99,$E5,$9E,$75
	.dc.b	$FC,$E9,$B1,$FC,$E1,$AF,$D6,$38,$9D,$DC,$BF,$CC,$9A,$FF,$24,$6D
	.dc.b	$FC,$13,$E9,$CF,$C1,$8C,$DF,$AA,$73,$5E,$AA,$FA,$2F,$92,$B7,$7E
	.dc.b	$DA,$DC,$FB,$4F,$79,$ED,$3C,$EF,$49,$E3,$5D,$27,$37,$E1,$3E,$8C
	.dc.b	$79,$63,$37,$F3,$CF,$04,$EA,$3C,$C7,$DA,$3C,$C7,$09,$F9,$8F,$FB
	.dc.b	$E7,$F0,$93,$08,$61,$AE,$1A,$ED,$AE,$CA,$E1,$34,$8E,$88,$74,$75
	.dc.b	$1C,$89,$19,$6B,$E4,$8F,$E5,$4F,$9E,$32,$F9,$DA,$F9,$DA,$FC,$0F
	.dc.b	$D5,$7C,$C4,$7D,$6D,$7D,$ED,$7E,$1B,$5F,$8E,$D7,$E5,$3D,$23,$99
	.dc.b	$23,$9B,$5E,$6D,$79,$B5,$E6,$D7,$D0,$FD,$0F,$D5,$BE,$71,$23,$04
	.dc.b	$3E,$BB,$96,$99,$3E,$EB,$96,$9D,$62,$3C,$9A,$F0,$6B,$F8,$88,$FA
	.dc.b	$69,$A4,$8D,$C7,$E3,$E3,$7F,$AB,$FE,$1C,$91,$F8,$C6,$BF,$90,$6B
	.dc.b	$F9,$46,$BF,$98,$6B,$F9,$B7,$A4,$F3,$24,$73,$6B,$CD,$AF,$36,$BC
	.dc.b	$DA,$FE,$A2,$3D,$FF,$AA,$FE,$7C,$91,$FA,$46,$BF,$A8,$6B,$FA,$C6
	.dc.b	$BF,$B0,$6B,$FB,$37,$A4,$F3,$24,$73,$6B,$CD,$AF,$36,$BC,$DA,$FF
	.dc.b	$00,$8F,$AF,$E9,$FE,$DC,$91,$FB,$C6,$BF,$C0,$1A,$FF,$04,$6B,$FC
	.dc.b	$21,$AF,$F0,$B7,$A4,$F3,$24,$73,$6B,$CD,$AF,$36,$BC,$DA,$FF,$18
	.dc.b	$8F,$C7,$FA,$7F,$C3,$C9,$1F,$C5,$1A,$FF,$18,$6B,$FC,$71,$AF,$F2
	.dc.b	$06,$BF,$C8,$DE,$93,$CC,$91,$CD,$AF,$36,$BC,$DA,$F3,$6B,$FC,$C2
	.dc.b	$3F,$4F,$FA,$C7,$F2,$72,$47,$F2,$C6,$BF,$CC,$1A,$FF,$34,$6B,$FC
	.dc.b	$E1,$AF,$F3,$B7,$A4,$F3,$24,$73,$6B,$CD,$AF,$36,$BC,$DA,$FF,$48
	.dc.b	$8F,$DF,$FE,$99,$FC,$FC,$91,$FD,$11,$AF,$F4,$86,$BF,$D3,$1A,$FF
	.dc.b	$50,$6B,$FD,$4D,$E9,$3C,$C9,$1C,$DA,$F3,$6B,$CD,$AF,$36,$BF,$D8
	.dc.b	$23,$F8,$BF,$EB,$3F,$D5,$C9,$1F,$D7,$1A,$FF,$60,$6B,$FD,$91,$AF
	.dc.b	$F6,$86,$BF,$DA,$DE,$9F,$A4,$07,$D6,$7F,$6F,$E3,$70,$AB,$55,$AB
	.dc.b	$FE,$EB,$5A,$9C,$DF,$1B,$E1,$7D,$7F,$3F,$C5,$6B,$0E,$35,$AA,$E9
	.dc.b	$0C,$3A,$47,$97,$0E,$2F,$39,$E3,$7C,$DE,$17,$CE,$B7,$5D,$2F,$9A
	.dc.b	$C3,$33,$2F,$9C,$F7,$6B,$B6,$C6,$DE,$F8,$78,$6F,$AD,$E5,$EF,$57
	.dc.b	$0F,$74,$FC,$2C,$BF,$C3,$35,$F2,$79,$5C,$D3,$2F,$6C,$62,$38,$47
	.dc.b	$F8,$8A,$7E,$85,$72,$C3,$EB,$5E,$EB,$8D,$F3,$7C,$F5,$BC,$BF,$CC
	.dc.b	$87,$FA,$A7,$0F,$15,$6D,$EB,$5D,$F0,$FC,$15,$F1,$C3,$B4,$FC,$26
	.dc.b	$D7,$08,$F3,$9F,$80,$8C,$D7,$9A,$FB,$3D,$27,$B6,$1E,$4A,$DB,$E5
	.dc.b	$9D,$4F,$43,$DB,$5D,$4E,$73,$EC,$C3,$CD,$18,$BE,$71,$F0,$4C,$C7
	.dc.b	$13,$A5,$7C,$63,$C7,$3E,$23,$8B,$E6,$9E,$2F,$88,$FC,$9B,$D5,$E3
	.dc.b	$3E,$13,$CB,$19,$8F,$01,$C4,$F0,$1E,$2C,$3E,$9C,$EA,$32,$76,$3A
	.dc.b	$B9,$8F,$A1,$19,$31,$87,$CF,$9E,$8F,$CA,$8F,$04,$F6,$78,$10,$FB
	.dc.b	$93,$E9,$38,$E1,$F5,$5C,$46,$E7,$C2,$F0,$8F,$0C,$FD,$38,$F3,$9E
	.dc.b	$0A,$F1,$3E,$C3,$81,$98,$E0,$F2,$C3,$E8,$BE,$27,$EA,$57,$CC,$8F
	.dc.b	$A8,$76,$39,$9D,$6F,$CA,$7B,$DC,$B8,$BF,$1B,$92,$38,$25,$FE,$4C
	.dc.b	$98,$7D,$BA,$F4,$61,$E6,$7E,$25,$7D,$17,$D9,$3B,$7C,$31,$97,$CC
	.dc.b	$F6,$9F,$6E,$1F,$42,$BF,$1F,$5C,$2F,$E2,$BE,$09,$F8,$4F,$C5,$7E
	.dc.b	$AB,$D2,$3E,$C4,$7D,$49,$F1,$61,$EE,$9F,$5D,$7C,$A3,$E8,$57,$47
	.dc.b	$D7,$1F,$06,$FD,$D1,$F6,$4F,$F5,$BC,$7F,$58,$BF,$EA,$C4,$8F,$EA
	.dc.b	$CD,$7F,$AB,$35,$FE,$AC,$D7,$FA,$B3,$6F,$EB,$4F,$E2,$F9,$BE,$25
	.dc.b	$79,$CF,$D2,$38,$72,$D7,$D3,$5F,$EC,$C8,$C4,$7C,$58,$C3,$A3,$A1
	.dc.b	$FC,$0E,$37,$5E,$29,$DC,$F1,$37,$1F,$CA,$27,$78,$FE,$22,$B7,$3C
	.dc.b	$6B,$73,$FC,$86,$B7,$7E,$D9,$F1,$1E,$A2,$9E,$58,$DD,$FF,$0F,$6C
	.dc.b	$7E,$89,$AF,$F4,$86,$BF,$18,$E6,$7A,$5C,$BE,$F3,$37,$E7,$8C,$BE
	.dc.b	$D8,$CD,$FB,$54,$EA,$3D,$91,$AB,$F6,$3F,$1C,$F6,$14,$F6,$1B,$BF
	.dc.b	$63,$4F,$61,$4F,$61,$4F,$61,$CC,$F6,$19,$3D,$86,$6F,$D8,$E5,$F6
	.dc.b	$39,$AF,$65,$FA,$ED,$7F,$6A,$7E,$05,$7A,$AF,$EE,$36,$32,$FB,$0F
	.dc.b	$F0,$AF,$E2,$47,$FB,$96,$5F,$D7,$8F,$EC,$B1,$98,$FE,$E0,$D7,$FC
	.dc.b	$53,$1B,$3F,$B4,$C6,$63,$FC,$56,$D7,$FC,$2B,$1B,$3F,$C6,$91,$9A
	.dc.b	$FF,$1E,$B6,$BF,$C7,$D3,$BA,$E6,$7F,$83,$67,$58,$7F,$84,$E7,$C6
	.dc.b	$78,$A7,$47,$B1,$FE,$E6,$7B,$1C,$9F,$DF,$9E,$CF,$C4,$32,$74,$63
	.dc.b	$E2,$3E,$19,$D5,$CB,$C4,$C9,$D5,$F0,$9E,$57,$27,$0B,$FF,$2C,$47
	.dc.b	$96,$32,$7F,$93,$1F,$2C,$7C,$68,$F2,$99,$8E,$A7,$CC,$72,$7D,$07
	.dc.b	$2F,$A4,$D1,$E5,$7C,$56,$BF,$A4,$7C,$E8,$F0,$C6,$67,$AD,$7A,$27
	.dc.b	$AC,$7C,$13,$A9,$D2,$FE,$B3,$E1,$21,$E5,$8F,$0D,$F8,$23,$E6,$1E
	.dc.b	$D8,$E6,$F0,$AF,$04,$70,$8F,$1D,$FE,$22,$3C,$73,$F1,$E7,$E5,$9D
	.dc.b	$CE,$87,$39,$F6,$3F,$0C,$C9,$EB,$35,$7E,$A7,$91,$92,$4F,$91,$E2
	.dc.b	$F8,$5F,$49,$C1,$D3,$C2,$7C,$2F,$C7,$AF,$C1,$3C,$27,$64,$9D,$C7
	.dc.b	$92,$7E,$AD,$FC,$98,$F5,$1D,$E7,$6F,$7B,$F5,$5F,$3A,$E1,$3E,$1B
	.dc.b	$FC,$B3,$E1,$BF,$99,$1F,$3E,$3E,$5D,$78,$EF,$9D,$76,$8F,$81,$3F
	.dc.b	$0A,$38,$BF,$3E,$FF,$19,$7F,$5A,$78,$12,$F9,$9D,$27,$CB,$5C,$E3
	.dc.b	$0F,$E7,$8F,$E3,$51,$FC,$3D,$A5,$70,$FC,$23,$27,$F1,$57,$24,$78
	.dc.b	$10,$D1,$4F,$88,$FC,$FC,$6C,$7E,$4A,$3F,$94,$46,$23,$E3,$D7,$E3
	.dc.b	$A3,$11,$96,$BF,$14,$FE,$79,$3F,$01,$FC,$33,$D1,$D1,$D1,$C3,$A3
	.dc.b	$C4,$72,$29,$FC,$52,$3B,$BB,$3F,$B3,$9D,$0F,$21,$4E,$EF,$0C,$3E
	.dc.b	$EC,$7F,$1F,$8C,$57,$DA,$AF,$1D,$68,$91,$9A,$F2,$1E,$93,$F9,$9D
	.dc.b	$72,$AF,$99,$1E,$13,$CD,$19,$76,$66,$BA,$CF,$82,$BB,$4F,$53,$DA
	.dc.b	$7B,$CF,$69,$4F,$69,$CD,$E3,$5D,$E3,$E4,$19,$9F,$2B,$C4,$F2,$99
	.dc.b	$3C,$A6,$6B,$CA,$FC,$03,$A4,$F5,$7F,$99,$B5,$FE,$9C,$D7,$E7,$1D
	.dc.b	$4F,$19,$D6,$BC,$EE,$4E,$65,$39,$94,$E6,$71,$AF,$5D,$7F,$72,$C7
	.dc.b	$F5,$04,$9F,$D3,$35,$FE,$FD,$3F,$E1,$27,$24,$33,$1B,$7F,$C1,$11
	.dc.b	$B7,$27,$F8,$17,$F5,$AF,$DA,$91,$FD,$D3,$5D,$18,$21,$82,$1F,$E1
	.dc.b	$17,$83,$67,$99,$1F,$03,$5D,$18,$21,$82,$1F,$E2,$18,$F5,$36,$3F
	.dc.b	$88,$6D,$FC,$38,$91,$82,$1F,$E2,$37,$2D,$32,$7F,$8A,$5C,$B4,$EB
	.dc.b	$11,$E4,$D7,$83,$5F,$F1,$B4,$79,$9F,$F1,$D3,$5F,$F2,$03,$53,$49
	.dc.b	$1F,$CA,$FF,$5B,$FE,$52,$49,$C1,$0F,$F1,$F3,$96,$99,$3F,$C8,$8E
	.dc.b	$5A,$75,$88,$F2,$6B,$C1,$AF,$F9,$46,$3D,$04,$8C,$C7,$F9,$5E,$30
	.dc.b	$FF,$94,$7F,$5C,$FE,$82,$47,$FA,$33,$5D,$18,$21,$82,$1F,$E5,$D7
	.dc.b	$83,$67,$99,$1F,$03,$5D,$18,$21,$82,$1F,$E6,$D8,$FA,$87,$D5,$F2
	.dc.b	$FF,$CC,$8F,$F6,$E3,$FB,$7B,$FE,$67,$33,$B0,$93,$FE,$7D,$CD,$A2
	.dc.b	$F8,$CF,$F9,$DE,$73,$7C,$AF,$8C,$A7,$8D,$70,$8F,$F4,$14,$70,$79
	.dc.b	$5F,$5A,$F1,$1F,$E8,$99,$E2,$E2,$7C,$92,$BF,$54,$72,$BF,$75,$6E
	.dc.b	$3C,$55,$B7,$BD,$79,$E7,$B4,$F8,$5F,$4D,$F9,$6B,$CF,$3C,$5B,$7D
	.dc.b	$AB,$73,$99,$E7,$7D,$5F,$3D,$7C,$17,$27,$9E,$BC,$D7,$EE,$79,$D7
	.dc.b	$B2,$7F,$D4,$4F,$F8,$48,$8F,$F8,$5D,$AE,$8C,$10,$C1,$0F,$F5,$73
	.dc.b	$C1,$F1,$9E,$E7,$C1,$AE,$22,$1F,$42,$32,$D3,$E5,$90,$E4,$7F,$B3
	.dc.b	$5F,$53,$C8,$FA,$6F,$F6,$F3,$2F,$33,$89,$FD,$C2,$BE,$93,$C5,$AE
	.dc.b	$E7,$E7,$E3,$F7,$37,$F9,$79,$7E,$15,$B8,$7F,$1B,$C3,$F2,$F5,$BC
	.dc.b	$7E,$7D,$6A,$3F,$0A,$66,$3E,$DC,$67,$1F,$8D,$39,$7E,$24,$65,$F1
	.dc.b	$9F,$2D,$F0,$99,$7E,$54,$71,$7E,$1C,$65,$F8,$B1,$F7,$48,$FE,$09
	.dc.b	$AF,$DA,$6B,$FE,$ED,$BF,$E0,$11,$88,$FF,$79,$46,$1E,$47,$A8,$87
	.dc.b	$A3,$1F,$4B,$0F,$39,$B7,$0F,$98,$E2,$69,$FD,$5F,$EA,$6F,$C0,$DC
	.dc.b	$6F,$CE,$73,$18,$FB,$EB,$26,$CC,$C3,$F7,$F7,$CE,$B5,$18,$BE,$4D
	.dc.b	$8E,$CD,$AE,$B7,$23,$04,$3F,$82,$DF,$58,$D4,$62,$FD,$35,$E3,$8C
	.dc.b	$5F,$89,$B5,$9D,$7F,$93,$E1,$E7,$BE,$71,$8C,$3C,$F7,$CA,$7F,$88
	.dc.b	$FB,$DF,$C8,$3C,$54,$C2,$C6,$5C,$07,$0E,$DA,$F0,$DF,$F9,$E4,$F0
	.dc.b	$8E,$11,$87,$9C,$F0,$73,$09,$E1,$39,$24,$71,$9F,$03,$5F,$03,$FD
	.dc.b	$09,$AF,$23,$F9,$A3,$83,$C2,$E1,$F0,$BB,$39,$38,$36,$E2,$3F,$DD
	.dc.b	$91,$CC,$F1,$C7,$0D,$FF,$D7,$24,$8F,$F5,$F3,$5F,$88,$1C,$3F,$1D
	.dc.b	$AF,$D0,$6B,$F5,$5E,$A7,$C1,$79,$34,$E9,$1F,$2E,$39,$46,$49,$19
	.dc.b	$8E,$91,$8A,$F1,$36,$38,$CE,$DF,$C3,$C7,$69,$DC,$FC,$18,$F9,$87
	.dc.b	$89,$C9,$27,$9B,$5F,$0B,$C4,$86,$5E,$F1,$D0,$FE,$5E,$75,$7F,$BA
	.dc.b	$B5,$FE,$E2,$FF,$92,$FE,$35,$E7,$7C,$6A,$FE,$D3,$3B,$CF,$FC,$1E
	.dc.b	$DE,$2F,$A9,$C5,$F9,$9C,$38,$CE,$48,$78,$A3,$58,$75,$C3,$FB,$EC
	.dc.b	$E9,$EF,$3A,$8C,$46,$9E,$58,$FE,$BA,$FB,$C7,$B8,$93,$CA,$75,$7E
	.dc.b	$66,$CF,$C1,$AE,$35,$FE,$24,$7D,$37,$E1,$3C,$53,$FE,$17,$8D,$CF
	.dc.b	$F0,$17,$75,$D2,$B7,$3C,$EF,$DD,$3F,$C2,$2B,$6F,$18,$F2,$36,$FC
	.dc.b	$91,$C2,$13,$C1,$E5,$1C,$1B,$3C,$09,$7E,$98,$F7,$C7,$A6,$7A,$57
	.dc.b	$39,$EB,$1C,$DB,$3C,$CE,$07,$E1,$63,$53,$F0,$A7,$C3,$3A,$AF,$04
	.dc.b	$7A,$E3,$89,$97,$E3,$1E,$4B,$F8,$D1,$E8,$8C,$9C,$63,$31,$C6,$3E
	.dc.b	$34,$FC,$E9,$D1,$88,$E2,$79,$1E,$27,$68,$F2,$C6,$63,$8B,$F2,$A3
	.dc.b	$2F,$C2,$9F,$4C,$F1,$8E,$A6,$5E,$A7,$85,$87,$52,$1F,$32,$7A,$D7
	.dc.b	$C9,$8E,$53,$C1,$D1,$F1,$5F,$2B,$83,$51,$C4,$97,$F5,$09,$7F,$66
	.dc.b	$BD,$15,$C2,$7F,$02,$49,$FB,$84,$3D,$F3,$B7,$0F,$BA,$7E,$31,$F0
	.dc.b	$AB,$C3,$3F,$66,$38,$3C,$9D,$98,$9E,$4F,$AC,$C9,$F1,$A7,$9D,$FC
	.dc.b	$6B,$F3,$5F,$E0,$6B,$67,$C8,$38,$47,$C0,$9F,$C2,$46,$27,$DB,$18
	.dc.b	$8E,$07,$84,$95,$EA,$9E,$D1,$8A,$F0,$DF,$CA,$8F,$0D,$6E,$31,$1B
	.dc.b	$AF,$A1,$1F,$26,$FF,$BE,$7C,$7F,$DF,$26,$1F,$AF,$7F,$64,$62,$BF
	.dc.b	$9A,$38,$8D,$1F,$C0,$DD,$4F,$48,$FB,$F3,$F3,$EF,$C5,$3E,$FC,$3E
	.dc.b	$5C,$7D,$7A,$F5,$E1,$E5,$8F,$C7,$1F,$1E,$BE,$85,$FA,$A3,$DF,$5D
	.dc.b	$AB,$EC,$DF,$F0,$92,$47,$49,$DD,$7E,$22,$BF,$33,$5A,$8C,$4F,$73
	.dc.b	$C8,$F7,$7D,$27,$90,$FA,$F1,$C6,$FD,$97,$F9,$97,$CF,$5F,$06,$75
	.dc.b	$5C,$5F,$82,$FE,$46,$FC,$77,$F8,$78,$D3,$CD,$F1,$D7,$38,$D4,$70
	.dc.b	$8F,$CF,$3E,$E7,$47,$E3,$9E,$6E,$9F,$C2,$3A,$BF,$3D,$7C,$AA,$EB
	.dc.b	$5F,$97,$AF,$C8,$DF,$C7,$78,$47,$92,$B8,$47,$78,$F2,$3C,$2B,$EB
	.dc.b	$4F,$9E,$FB,$B6,$B9,$35,$F8,$87,$D3,$29,$F4,$27,$CA,$DA,$F6,$5F
	.dc.b	$96,$BC,$77,$DA,$B9,$1F,$16,$32,$F8,$0C,$9F,$11,$CB,$E0,$9F,$AE
	.dc.b	$F1,$8F,$33,$4E,$31,$E1,$8C,$C7,$36,$C6,$63,$77,$F1,$1B,$7F,$92
	.dc.b	$7D,$97,$E9,$69,$E7,$29,$E5,$9F,$C3,$36,$7E,$1C,$75,$8D,$9E,$72
	.dc.b	$9E,$53,$C3,$3F,$9A,$9E,$27,$8E,$BE,$0C,$EC,$E2,$FA,$DB,$5D,$E3
	.dc.b	$66,$4A,$E4,$F3,$C7,$8E,$33,$3D,$E7,$33,$DE,$73,$5E,$5A,$FC,$61
	.dc.b	$B3,$CE,$E6,$FF,$90,$CF,$86,$F9,$B6,$B7,$7F,$A4,$79,$DF,$1B,$FE
	.dc.b	$25,$3F,$9A,$79,$DF,$A6,$B8,$CE,$67,$6F,$E9,$DD,$C7,$94,$EB,$1D
	.dc.b	$E3,$35,$DE,$3F,$8A,$B5,$FE,$1E,$75,$32,$71,$8E,$EF,$E8,$67,$AB
	.dc.b	$B3,$89,$E3,$8D,$DF,$F4,$56,$DF,$F2,$59,$D9,$CE,$3C,$AF,$73,$C2
	.dc.b	$75,$9F,$E3,$6D,$9F,$D9,$C7,$58,$D9,$E1,$29,$D6,$7C,$35,$FC,$B6
	.dc.b	$79,$CE,$CF,$1B,$CD,$B5,$E7,$8D,$C7,$33,$AB,$B3,$25,$8F,$29,$C4
	.dc.b	$B3,$E1,$6B,$FC,$72,$7F,$54,$FC,$C7,$C9,$3D,$27,$94,$FF,$1F,$9C
	.dc.b	$C7,$2A,$FE,$13,$1F,$9D,$8F,$C8,$9F,$89,$9E,$91,$E9,$3D,$AF,$8C
	.dc.b	$E8,$E6,$BC,$11,$9B,$FE,$09,$1F,$A4,$9F,$AC,$72,$9F,$53,$E1,$7F
	.dc.b	$2E,$F9,$9E,$B1,$F1,$5B,$5E,$CA,$EA,$FA,$23,$F3,$57,$F8,$4B,$FC
	.dc.b	$04,$70,$3C,$8F,$F1,$5B,$F2,$3A,$BE,$73,$E2,$8F,$65,$FC,$7B,$FE
	.dc.b	$4A,$FC,$57,$B4,$79,$EB,$8D,$7C,$8B,$F4,$46,$2B,$75,$D2,$BC,$87
	.dc.b	$C2,$7F,$85,$5F,$AE,$BF,$5D,$7F,$32,$B4,$FF,$05,$7B,$4F,$C3,$AF
	.dc.b	$63,$C5,$10,$61,$FE,$27,$96,$12,$77,$3F,$E2,$99,$61,$FE,$27,$7F
	.dc.b	$C5,$B7,$FE,$2B,$3F,$C5,$AE,$AB,$B1,$DA,$35,$1C,$0D,$47,$8E,$BD
	.dc.b	$57,$F1,$30,$F8,$F3,$EA,$AF,$A5,$18,$8F,$CF,$CE,$EF,$E0,$47,$FA
	.dc.b	$04,$F7,$D7,$67,$C6,$D7,$E1,$35,$DE,$1E,$87,$2D,$72,$58,$E0,$FF
	.dc.b	$AF,$CC,$9C,$5C,$92,$BE,$69,$92,$C7,$91,$AF,$91,$AF,$A9,$AA,$39
	.dc.b	$C7,$33,$EC,$47,$16,$BC,$5A,$F1,$79,$C7,$28,$F1,$47,$04

;--------------------------------------------------------------------------------
;	FMOVE.S Dr,FPn
;	FMOVE.S <mem>,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovestoreg_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.S <EA>,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.S <ea>,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d2	source
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source,...
;	a4	expected result handle,expected status,...
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fmovestoreg_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a0,a2			;source,...
@@:
	addq.l	#4,a0			;source
	cmpi.l	#$FF800000,(a0)		;-Inf
	bne	@b
	addq.l	#4,a0			;0
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;Dr,FPn
	fmove.l	#0,fpcr
	move.l	(a2),d2			;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fmove.s	d2,fp7			;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)		;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.S Dr='
	puthex8	d2			;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
;<mem>,FPn
	fmove.l	#0,fpcr
	move.l	(a2),d2			;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	d2,-(sp)		;source
	fmove.s	(sp)+,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)		;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.S <mem>='
	puthex8	d2			;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source,...
	cmpi.l	#$FF800000,(a2)		;source,.... -Inf
	bne	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmovestoreg_data_compressed::
	.dc.b	$00,$00,$EC,$5C,$7F,$00,$BF,$03,$44,$87,$E0,$7A,$76,$F0,$FE,$01
	.dc.b	$E3,$3C,$03,$B9,$A6,$FF,$78,$7E,$44,$FD,$29,$F9,$A3,$F6,$4F,$16
	.dc.b	$3E,$27,$A2,$FD,$93,$2D,$3C,$1F,$B2,$60,$DF,$EC,$9A,$38,$FE,$C9
	.dc.b	$C0,$E7,$FB,$27,$23,$AF,$EC,$9D,$07,$D6,$CD,$11,$E2,$FD,$93,$C2
	.dc.b	$79,$3F,$64,$F1,$9E,$6F,$D9,$3C,$A7,$A3,$F6,$4F,$39,$EA,$FD,$93
	.dc.b	$D2,$7B,$3F,$64,$F5,$9E,$EF,$D9,$3D,$A7,$C0,$FD,$93,$DE,$7C,$2F
	.dc.b	$D9,$3E,$09,$F1,$3F,$64,$F8,$67,$C6,$FD,$93,$E2,$9F,$23,$F6,$4F
	.dc.b	$8E,$7C,$AF,$D9,$3E,$49,$F3,$3F,$64,$F9,$FE,$09,$F9,$A7,$E0,$3B
	.dc.b	$BF,$0A,$FD,$D8,$D7,$4F,$DF,$FD,$93,$EF,$1F,$85,$F0,$0F,$D6,$3E
	.dc.b	$CF,$EC,$9F,$60,$FB,$7F,$B2,$3B,$5F,$99,$BF,$49,$4F,$DE,$3F,$03
	.dc.b	$FB,$27,$E1,$3C,$73,$F8,$6F,$C1,$6C,$7E,$63,$E9,$EE,$AF,$88,$AF
	.dc.b	$0F,$2B,$F3,$FF,$AF,$DE,$3F,$D0,$65,$EF,$C3,$17,$D8,$71,$7E,$92
	.dc.b	$7F,$3D,$79,$9E,$27,$86,$3F,$2F,$FB,$27,$E9,$F3,$3F,$A9,$9B,$8F
	.dc.b	$E9,$A6,$F1,$FE,$C2,$BF,$3A,$7E,$83,$C3,$FC,$F9,$FA,$2F,$D9,$3F
	.dc.b	$42,$7E,$93,$F6,$4F,$DC,$5F,$EE,$1F,$DD,$70,$58,$6D,$78,$31,$A6
	.dc.b	$F8,$1B,$DB,$C5,$E6,$BA,$70,$7E,$F7,$2F,$E0,$B7,$DE,$7F,$6F,$1F
	.dc.b	$45,$3F,$B8,$FD,$93,$F8,$6E,$1F,$C3,$1F,$E0,$F2,$7E,$13,$77,$FE
	.dc.b	$15,$1F,$BE,$3F,$80,$FE,$C9,$FC,$00,$FB,$93,$FC,$59,$7C,$6F,$E2
	.dc.b	$7F,$51,$43,$E3,$9F,$81,$C8,$7C,$7F,$F1,$75,$F1,$DC,$57,$C7,$70
	.dc.b	$AE,$7F,$1D,$BF,$8D,$D7,$C6,$E6,$DF,$1B,$CF,$E3,$39,$7C,$66,$7E
	.dc.b	$33,$B7,$C5,$F5,$F8,$BE,$1F,$17,$E1,$F8,$AF,$07,$C5,$73,$F8,$AF
	.dc.b	$27,$C4,$F8,$FE,$27,$B7,$C4,$F9,$FE,$23,$CD,$E0,$F8,$7B,$3F,$27
	.dc.b	$FC,$03,$E7,$F0,$9F,$BF,$FC,$35,$FB,$F7,$E6,$3F,$7C,$D7,$EF,$79
	.dc.b	$FD,$EB,$F3,$7F,$BC,$71,$FD,$DB,$87,$EE,$BF,$9E,$FD,$CF,$A7,$EE
	.dc.b	$3C,$FF,$70,$FD,$0F,$B5,$DF,$E3,$7B,$7C,$5F,$E9,$3E,$27,$C5,$F0
	.dc.b	$FE,$1F,$86,$FD,$3F,$C2,$79,$7E,$0B,$C9,$EF,$FE,$AB,$DC,$F4,$7A
	.dc.b	$9E,$7E,$7F,$D6,$FE,$DD,$EB,$FD,$B3,$D5,$FB,$5F,$EC,$3F,$69,$F7
	.dc.b	$7E,$CF,$ED,$FD,$9B,$F6,$7F,$B2,$FC,$1F,$87,$F8,$1F,$B1,$FE,$DB
	.dc.b	$F6,$1F,$89,$FA,$FF,$C3,$FD,$7B,$F7,$3F,$3D,$F1,$FF,$5A,$F8,$DF
	.dc.b	$AC,$FE,$F3,$F5,$7F,$95,$FA,$B7,$C9,$FD,$57,$F7,$FF,$A9,$FC,$DF
	.dc.b	$D4,$7E,$67,$EA,$1F,$C0,$BF,$4E,$FA,$1F,$A6,$7C,$FF,$8D,$FE,$0B
	.dc.b	$FA,$3F,$D1,$FD,$0F,$EA,$7E,$7F,$FC,$23,$F3,$AF,$A9,$F9,$B7,$D7
	.dc.b	$F9,$7F,$E1,$7F,$97,$7D,$7F,$8D,$FB,$5F,$93,$7F,$0E,$FC,$87,$ED
	.dc.b	$7E,$3B,$F7,$7E,$67,$F8,$8F,$E2,$9F,$77,$F1,$0F,$C0,$7E,$19,$FC
	.dc.b	$53,$F0,$7F,$C0,$7E,$07,$F8,$3F,$A0,$FE,$2F,$F3,$3F,$83,$F9,$6F
	.dc.b	$C3,$7C,$97,$F1,$AF,$8F,$FC,$37,$C6,$7E,$27,$E2,$7F,$BD,$7E,$E0
	.dc.b	$F8,$7F,$D1,$2F,$84,$FD,$72,$F8,$1D,$FB,$DA,$F6,$F3,$EC,$72,$F5
	.dc.b	$78,$FA,$5C,$3C,$FE,$BE,$67,$4F,$27,$9F,$8D,$E0,$F0,$FA,$F8,$1E
	.dc.b	$0E,$DF,$1F,$57,$83,$9F,$C7,$C9,$FE,$0F,$FD,$C1,$C3,$FB,$77,$FA
	.dc.b	$56,$CF,$9E,$47,$F8,$20,$61,$FC,$1D,$FF,$06,$FB,$5C,$24,$FF,$83
	.dc.b	$8F,$F0,$8F,$EE,$1F,$E1,$03,$FC,$27,$FB,$87,$F8,$48,$FF,$0A,$FB
	.dc.b	$9F,$E1,$43,$FC,$2F,$FB,$87,$F8,$58,$FF,$0C,$FB,$BF,$E1,$83,$FC
	.dc.b	$37,$FB,$87,$F8,$68,$FF,$0E,$FE,$E1,$FE,$1C,$3F,$C3,$FD,$1F,$E1
	.dc.b	$E3,$FC,$43,$FB,$87,$F8,$80,$FF,$11,$FE,$E1,$FE,$22,$3F,$C4,$BF
	.dc.b	$03,$FE,$24,$3F,$C4,$FF,$B8,$7F,$89,$8F,$F1,$4F,$C1,$7F,$8A,$0F
	.dc.b	$F1,$5F,$EE,$1F,$E2,$A3,$FC,$5B,$F8,$0F,$F8,$B0,$FF,$17,$FC,$27
	.dc.b	$F8,$B8,$FF,$18,$FE,$E1,$FE,$30,$3F,$C6,$7C,$9F,$E3,$23,$FC,$6B
	.dc.b	$CF,$FE,$34,$3F,$C6,$FF,$B8,$7F,$8D,$8F,$F1,$CF,$C3,$7F,$8E,$0F
	.dc.b	$F1,$DF,$EE,$1F,$E3,$A3,$FC,$7B,$FB,$87,$F8,$F0,$FF,$1F,$FC,$47
	.dc.b	$F8,$F8,$FF,$20,$FE,$E1,$FE,$40,$3F,$C8,$7F,$B8,$7F,$90,$8F,$F2
	.dc.b	$2F,$C5,$7F,$91,$0F,$F2,$3F,$EE,$1F,$E4,$63,$FC,$93,$E5,$7F,$92
	.dc.b	$0F,$F2,$5F,$EE,$1F,$E4,$AF,$D7,$FF,$C3,$87,$F9,$3B,$CA,$FF,$2C
	.dc.b	$1F,$E2,$43,$FC,$9E,$7F,$88,$FE,$E3,$FC,$CD,$F6,$DF,$E6,$18,$FE
	.dc.b	$DB,$E4,$7F,$81,$9A,$FF,$9A,$FF,$2D,$FF,$01,$3F,$E6,$CF,$A4,$FF
	.dc.b	$2C,$FD,$27,$F9,$6A,$3F,$CB,$A7,$F9,$79,$FF,$2F,$C7,$F9,$CA,$7F
	.dc.b	$C1,$1E,$6F,$F9,$88,$FF,$3A,$CF,$F9,$DB,$0F,$F3,$95,$F9,$4E,$98
	.dc.b	$69,$FF,$39,$4F,$57,$8F,$E2,$9F,$E6,$AF,$77,$FC,$DA,$7F,$9B,$3F
	.dc.b	$15,$FF,$37,$9F,$E6,$EF,$C5,$BF,$CE,$27,$F9,$C3,$F1,$7F,$F3,$99
	.dc.b	$FE,$72,$FC,$63,$FC,$EA,$7F,$9D,$3E,$0B,$FC,$EE,$7F,$9D,$BF,$19
	.dc.b	$FF,$3C,$9F,$E7,$8F,$C6,$BF,$CF,$67,$F9,$EB,$F1,$BF,$F3,$E9,$FE
	.dc.b	$7C,$FC,$73,$FC,$FE,$7F,$9F,$BD,$6F,$F4,$09,$FE,$80,$FC,$77,$FD
	.dc.b	$06,$7F,$A0,$BF,$1E,$FF,$42,$9F,$E8,$4F,$C7,$FF,$D0,$E7,$FA,$1B
	.dc.b	$F2,$0F,$F4,$49,$FE,$88,$FC,$87,$FD,$16,$7F,$A2,$BE,$33,$FD,$1A
	.dc.b	$7F,$A3,$3F,$22,$FF,$47,$9F,$E8,$EF,$C8,$FF,$D2,$27,$FA,$43,$F2
	.dc.b	$4F,$F4,$99,$FE,$92,$FC,$97,$FD,$2A,$7F,$A5,$3E,$47,$FD,$2E,$7F
	.dc.b	$A5,$BF,$26,$FF,$4C,$9F,$E9,$8F,$C9,$FF,$D3,$67,$FA,$6B,$F2,$8F
	.dc.b	$F4,$E9,$FE,$9C,$FC,$A7,$FD,$3E,$7F,$A7,$BE,$5B,$FD,$42,$7F,$A8
	.dc.b	$3F,$2A,$FF,$51,$9F,$EA,$2F,$CA,$FF,$D4,$A7,$FA,$93,$F2,$CF,$F5
	.dc.b	$39,$7D,$5F,$F1,$27,$E5,$BF,$EA,$A3,$F9,$0B,$FC,$95,$FF,$55,$9F
	.dc.b	$C8,$4F,$E5,$7F,$1B,$FE,$28,$F2,$7F,$D6,$23,$E3,$7F,$C4,$8F,$2F
	.dc.b	$64,$FE,$80,$FF,$14,$4E,$3E,$3B,$FC,$69,$1F,$EB,$81,$F1,$F8,$F9
	.dc.b	$0E,$7E,$9F,$FB,$26,$7F,$99,$0F,$86,$EF,$F2,$39,$F9,$2E,$7F,$25
	.dc.b	$E2,$F9,$3E,$1F,$2B,$DB,$E5,$BF,$C7,$1E,$B7,$E9,$FE,$5E,$FC,$C7
	.dc.b	$F9,$16,$1F,$33,$FE,$D7,$7D,$9C,$5F,$E4,$67,$FC,$87,$CD,$FD,$25
	.dc.b	$FF,$27,$FA,$DF,$E6,$AF,$43,$AF,$99,$CF,$C7,$E5,$E1,$F8,$3B,$FB
	.dc.b	$61,$D7,$E6,$FC,$7F,$39,$E1,$F9,$DF,$06,$BE,$6F,$9E,$F2,$7C,$FF
	.dc.b	$8F,$E8,$3D,$3D,$FE,$7F,$A1,$F3,$7D,$17,$B3,$E8,$FD,$5F,$03,$E9
	.dc.b	$FA,$4F,$67,$D2,$FC,$0F,$A6,$F6,$78,$DE,$FF,$A7,$F8,$7F,$51,$EF
	.dc.b	$FA,$9F,$85,$E4,$FC,$6F,$AA,$F8,$5F,$57,$F1,$7E,$B3,$E4,$F9,$9F
	.dc.b	$17,$E4,$7E,$47,$CA,$7C,$CF,$97,$F9,$1E,$27,$CB,$F9,$AF,$9F,$F3
	.dc.b	$BF,$2F,$E7,$FE,$77,$A3,$F4,$BE,$93,$E7,$7D,$37,$D1,$FA,$9F,$AB
	.dc.b	$EA,$7D,$1F,$AD,$FA,$9F,$5D,$F6,$3E,$BF,$EA,$7A,$FF,$5F,$EC,$3E
	.dc.b	$DF,$D8,$FD,$7F,$B2,$FB,$5E,$D7,$DE,$FB,$3F,$B5,$F6,$9F,$77,$ED
	.dc.b	$7F,$03,$EE,$FD,$DF,$B6,$FC,$07,$DB,$FE,$13,$EE,$3F,$01,$F0,$1F
	.dc.b	$83,$FB,$9F,$C3,$FD,$D7,$E0,$FE,$EF,$F0,$DF,$07,$F8,$AF,$BC,$FC
	.dc.b	$37,$DE,$FE,$27,$EF,$BF,$D0,$9F,$97,$3E,$12,$07,$FD,$8F,$ED,$FF
	.dc.b	$A4,$17,$D4,$7F,$A5,$57,$E0,$1C,$17,$C4,$F4,$FC,$07,$5F,$81,$67
	.dc.b	$F0,$3E,$FF,$19,$C7,$F0,$4E,$1F,$82,$F8,$BF,$06,$E9,$F1,$FC,$FF
	.dc.b	$07,$F2,$FE,$11,$DF,$F0,$9E,$DF,$2B,$E8,$FC,$2F,$C5,$F8,$67,$87
	.dc.b	$F0,$DF,$5F,$E1,$9E,$5F,$C3,$BC,$9F,$87,$FB,$BF,$10,$F4,$7D,$B7
	.dc.b	$9F,$E2,$FD,$FF,$97,$FC,$0E,$D2,$8C,$FE,$25,$1F,$8D,$1F,$EC,$03
	.dc.b	$FD,$87,$2F,$CC,$39,$1F,$EC,$93,$E6,$4F,$CD,$7F,$D9,$F1,$BF,$BB
	.dc.b	$F9,$F0,$FB,$CA,$7F,$72,$71,$5C,$63,$FD,$BB,$5D,$E3,$89,$FE,$CA
	.dc.b	$9F,$47,$CD,$F7,$87,$CB,$D7,$EF,$7E,$63,$EF,$3D,$D7,$F8,$1F,$9C
	.dc.b	$F8,$46,$17,$E6,$5F,$14,$C1,$FD,$D8,$FF,$04,$0F,$BF,$91,$C1,$FF
	.dc.b	$73,$FE,$67,$F3,$4F,$E2,$C7,$E3,$B0,$3F,$90,$F8,$0B,$4D,$BF,$29
	.dc.b	$3F,$16,$EE,$DD,$F5,$E2,$F6,$3C,$58,$74,$C7,$F3,$77,$EE,$E1,$C6
	.dc.b	$4C,$2E,$D8,$B7,$E6,$9F,$83,$3E,$10,$F5,$DF,$8E,$3E,$29,$F9,$AF
	.dc.b	$E2,$4F,$91,$F4,$5F,$8B,$2F,$07,$C5,$E7,$47,$C7,$8F,$91,$F9,$B7
	.dc.b	$E4,$4F,$D8,$F1,$9F,$F1,$B1,$F4,$CF,$AE,$BF,$37,$FC,$C1,$F5,$FF
	.dc.b	$38,$FC,$D1,$F6,$7F,$39,$FC,$E1,$F6,$FF,$3A,$FC,$F1,$F7,$7E,$9F
	.dc.b	$F4,$07,$DF,$FC,$EF,$F4,$47,$E0,$7F,$3C,$FD,$21,$F8,$3F,$CF,$7F
	.dc.b	$4C,$7E,$17,$F3,$EF,$D4,$1F,$87,$F6,$7F,$54,$7E,$27,$F3,$FF,$D6
	.dc.b	$1F,$8B,$FD,$03,$F5,$C7,$E3,$7F,$41,$FD,$81,$F8,$FF,$D0,$BF,$64
	.dc.b	$3F,$36,$9F,$26,$8F,$CA,$7D,$87,$ED,$CF,$CB,$7E,$87,$FB,$93,$F3
	.dc.b	$15,$FB,$B3,$F3,$5F,$A2,$7E,$F4,$FC,$E7,$E8,$BF,$BF,$3F,$3D,$FA
	.dc.b	$37,$F0,$13,$F4,$1F,$AB,$BE,$99,$9B,$F2,$07,$EB,$FF,$67,$BE,$F7
	.dc.b	$7D,$BF,$33,$F8,$51,$F9,$78,$FD,$E3,$9F,$1B,$F9,$6B,$F8,$19,$FE
	.dc.b	$13,$E4,$FF,$9A,$3E,$07,$FC,$D2,$FF,$32,$6B,$FC,$D5,$A9,$A4,$76
	.dc.b	$D7,$9B,$5E,$C6,$8F,$F3,$BF,$E8,$4F,$65,$F0,$9E,$28,$C3,$63,$33
	.dc.b	$C1,$AF,$F2,$99,$DB,$99,$E0,$47,$04,$38,$4F,$27,$FA,$8C,$F2,$7A
	.dc.b	$7E,$2B,$97,$FA,$B4,$23,$10,$8C,$57,$A6,$73,$18,$9F,$7C,$60,$93
	.dc.b	$B2,$46,$9A,$FB,$7E,$F3,$99,$23,$E3,$36,$3E,$5B,$5E,$11,$A2,$3A
	.dc.b	$6B,$F2,$7F,$03,$F3,$92,$3E,$93,$63,$EB,$B5,$E1,$1A,$23,$A6,$BF
	.dc.b	$57,$CE,$F8,$64,$8F,$BC,$D8,$FC,$1B,$5E,$11,$A2,$3A,$6B,$FE,$08
	.dc.b	$5F,$BF,$BF,$43,$FB,$11,$86,$C6,$63,$FD,$8E,$64,$AE,$67,$FB,$A4
	.dc.b	$65,$B1,$97,$C0,$D7,$FC,$35,$1B,$73,$3E,$02,$47,$E7,$7F,$5C,$FC
	.dc.b	$29,$1C,$12,$7F,$4B,$19,$6C,$65,$FD,$04,$65,$B1,$FA,$9F,$CA,$7F
	.dc.b	$22,$47,$04,$8F,$D8,$C6,$5B,$19,$7F,$59,$19,$6C,$7E,$D7,$F7,$4F
	.dc.b	$CE,$91,$C1,$23,$F7,$B1,$96,$C6,$5F,$DC,$46,$5B,$1F,$C0,$7E,$5F
	.dc.b	$F5,$24,$70,$48,$FE,$13,$19,$6C,$65,$FE,$09,$19,$6C,$7F,$0D,$FE
	.dc.b	$01,$FD,$A9,$1C,$12,$3F,$8A,$C6,$5B,$19,$7F,$88,$46,$5B,$1F,$CA
	.dc.b	$BE,$4B,$F8,$DB,$4F,$E0,$EE,$23,$FC,$80,$7F,$96,$A3,$24,$38,$11
	.dc.b	$DC,$71,$7F,$C9,$0D,$7F,$CD,$2D,$7F,$CD,$8D,$7F,$CC,$FF,$11,$E2
	.dc.b	$24,$F9,$DB,$1D,$9A,$F6,$6B,$D9,$AF,$67,$DB,$ED,$49,$F8,$4D,$9F
	.dc.b	$4B,$5F,$4B,$5F,$4B,$5F,$4B,$5F,$F3,$FF,$E2,$DF,$28,$93,$F3,$DB
	.dc.b	$1F,$19,$AF,$C6,$6B,$F1,$9A,$FC,$67,$EA,$FE,$1B,$27,$ED,$36,$7E
	.dc.b	$8B,$5F,$A2,$D7,$E8,$B5,$FA,$2D,$7F,$D2,$FF,$03,$F8,$22,$4F,$E1
	.dc.b	$DB,$1F,$79,$AF,$DE,$6B,$F7,$9A,$FD,$E7,$F1,$BF,$BD,$49,$FC,$93
	.dc.b	$67,$F1,$2D,$7F,$12,$D7,$F1,$2D,$7F,$12,$D5,$F2,$6E,$5F,$EC,$2D
	.dc.b	$AF,$F0,$4B,$67,$F3,$AD,$7F,$D3,$CD,$7F,$D7,$ED,$7F,$D8,$4D,$7B
	.dc.b	$CF,$62,$4F,$66,$C7,$F3,$6F,$98,$73,$7F,$40,$6D,$7F,$3E,$6C,$EE
	.dc.b	$74,$49,$D3,$63,$51,$82,$38,$3D,$BC,$1F,$08,$91,$F0,$9B,$1A,$8C
	.dc.b	$11,$C1,$0F,$8B,$1C,$1F,$8F,$C1,$F3,$89,$3F,$39,$B1,$F3,$9A,$F0
	.dc.b	$8D,$11,$D3,$5F,$F6,$87,$E9,$1F,$C1,$49,$1F,$E1,$C6,$BF,$E4,$46
	.dc.b	$C7,$F1,$86,$BC,$A3,$81,$0E,$1C,$1C,$C8,$E0,$91,$CE,$32,$D8,$CB
	.dc.b	$E6,$8C,$B6,$3C,$9C,$1E,$72,$38,$24,$79,$E3,$2D,$8C,$BF,$0A,$32
	.dc.b	$D8,$F8,$1C,$1F,$0C,$8E,$09,$1F,$0E,$32,$D8,$CB,$F3,$A3,$2D,$8F
	.dc.b	$99,$C1,$F3,$C8,$E0,$91,$F3,$E3,$2D,$8C,$BF,$6A,$32,$D8,$FB,$1C
	.dc.b	$1F,$6C,$8E,$09,$1F,$6E,$32,$D8,$CB,$F8,$68,$CB,$63,$F0,$9E,$A7
	.dc.b	$E1,$C8,$E0,$91,$F8,$78,$CB,$63,$2F,$E4,$A3,$2D,$8F,$C8,$79,$3F
	.dc.b	$93,$23,$82,$47,$E4,$E3,$2D,$8C,$BF,$9E,$8C,$B6,$3F,$39,$FA,$A7
	.dc.b	$E7,$C8,$E0,$91,$F9,$F8,$CB,$63,$2F,$EA,$A3,$2D,$8F,$D4,$78,$FF
	.dc.b	$AB,$23,$82,$47,$EA,$E3,$2D,$8C,$BF,$B6,$8C,$B6,$3F,$69,$E1,$7E
	.dc.b	$DC,$8E,$09,$1F,$B7,$8C,$B6,$32,$FF,$02,$8C,$B6,$3F,$80,$7E,$D5
	.dc.b	$FC,$0C,$8E,$09,$1F,$C0,$E3,$2D,$8C,$BF,$C3,$A3,$2D,$8F,$E1,$9F
	.dc.b	$75,$FC,$3C,$8E,$09,$1F,$C3,$E3,$2D,$8C,$BF,$C6,$A3,$2D,$8F,$E3
	.dc.b	$1F,$94,$7F,$1B,$23,$82,$47,$F1,$B8,$CB,$63,$2F,$F2,$68,$CB,$63
	.dc.b	$F9,$27,$DA,$FF,$27,$23,$82,$47,$F2,$78,$CB,$63,$2F,$F3,$28,$CB
	.dc.b	$63,$F9,$87,$E6,$9F,$CC,$C8,$E0,$91,$FC,$CE,$32,$D8,$CB,$FC,$FA
	.dc.b	$32,$D8,$FE,$79,$F6,$1F,$CF,$C8,$E0,$91,$FC,$FE,$32,$D8,$CB,$FD
	.dc.b	$2A,$32,$D8,$FF,$2E,$BF,$EF,$CF,$C1,$FF,$DC,$F0,$69,$1D,$99,$2B
	.dc.b	$98,$FF,$5B,$B5,$FF,$5A,$B5,$FF,$61,$3F,$EF,$4F,$9D,$78,$14,$95
	.dc.b	$E4,$8C,$B6,$33,$5F,$E7,$49,$CB,$63,$28,$82,$36,$E0,$C4,$70,$77
	.dc.b	$E4,$38,$57,$F1,$A3,$0D,$8C,$CF,$FA,$1A,$72,$D8,$CB,$DE,$32,$D8
	.dc.b	$CB,$E0,$6B,$E9,$8D,$B9,$9F,$01,$1F,$07,$0F,$05,$28,$C3,$67,$D5
	.dc.b	$18,$23,$82,$4F,$46,$BE,$98,$DB,$EF,$E1,$EA,$49,$F8,$CD,$8D,$46
	.dc.b	$08,$E0,$87,$C5,$8E,$0F,$C7,$E1,$F4,$92,$7E,$93,$63,$E8,$35,$E1
	.dc.b	$1A,$23,$A6,$BF,$4F,$87,$E2,$92,$3E,$F3,$63,$EE,$35,$E1,$1A,$23
	.dc.b	$A6,$BF,$7F,$87,$E8,$92,$3F,$14,$D8,$FC,$43,$5E,$11,$A2,$3A,$6B
	.dc.b	$F9,$9F,$3D,$F4,$5F,$81,$6C,$FE,$F1,$B3,$FE,$DA,$6B,$C6,$78,$12
	.dc.b	$78,$36,$38,$35,$E1,$1F,$C2,$3F,$4B,$7C,$67,$E4,$5B,$19,$8C,$11
	.dc.b	$C1,$27,$FD,$DB,$3C,$1B,$3B,$6C,$ED,$FD,$C4,$BE,$C3,$C0,$49,$F2
	.dc.b	$36,$7A,$B5,$E2,$D7,$F8,$2C,$F3,$9E,$24,$8F,$2F,$91,$ED,$24,$7C
	.dc.b	$16,$C6,$A3,$04,$70,$43,$D3,$1C,$1F,$E2,$5F,$40,$FD,$9B,$F7,$4D
	.dc.b	$AF,$E6,$AD,$9F,$E3,$01,$C3,$CA,$78,$92,$78,$B6,$38,$CF,$07,$F9
	.dc.b	$2B,$5F,$E5,$0D,$7B,$7E,$81,$E4,$25,$79,$1B,$1D,$1A,$FB,$1A,$F6
	.dc.b	$6B,$D9,$FE,$31,$F4,$4F,$06,$92,$BF,$AA,$C6,$5B,$19,$AF,$87,$39
	.dc.b	$6C,$65,$FE,$7D,$19,$6C,$65,$FE,$8B,$0F,$03,$C6,$4A,$F0,$B6,$7B
	.dc.b	$B5,$EA,$D7,$9B,$5F,$EA,$53,$D6,$7D,$7E,$07,$C0,$23,$82,$4F,$B2
	.dc.b	$32,$D8,$CB,$EC,$8C,$B6,$3E,$3F,$81,$F3,$08,$E0,$91,$F2,$23,$2D
	.dc.b	$8C,$BF,$22,$32,$D8,$FA,$7E,$07,$D8,$23,$82,$47,$D4,$8C,$B6,$32
	.dc.b	$FD,$48,$CB,$63,$F3,$A3,$C6,$79,$9F,$D9,$23,$0D,$8C,$C7,$06,$CF
	.dc.b	$F7,$D8,$D9,$99,$E0,$47,$04,$8E,$13,$C9,$FF,$04,$C7,$27,$F1,$FD
	.dc.b	$1F,$98,$24,$7E,$59,$B1,$A8,$C1,$1C,$10,$FC,$AC,$70,$7F,$2F,$D1
	.dc.b	$FA,$42,$4F,$E8,$9B,$1F,$A0,$6B,$C2,$34,$47,$4D,$7F,$47,$CF,$FB
	.dc.b	$02,$47,$EB,$9B,$1F,$AC,$6B,$C2,$34,$47,$4D,$7F,$5F,$E0,$7E,$F0
	.dc.b	$91,$FB,$A6,$C7,$EE,$1A,$F0,$8D,$11,$D3,$5F,$DD,$FD,$77,$F0,$82
	.dc.b	$47,$F0,$66,$C7,$F0,$46,$BC,$23,$44,$74,$D7,$FC,$89,$F4,$6F,$37
	.dc.b	$FC,$58,$D9,$FF,$33,$B6,$7F,$99,$B5,$E3,$3C,$09,$3C,$1B,$1C,$1A
	.dc.b	$F0,$8E,$BF,$91,$78,$08,$E0,$91,$E0,$8C,$B6,$32,$FA,$A3,$2D,$8F
	.dc.b	$4F,$C6,$7B,$08,$E0,$91,$EC,$8C,$B6,$32,$FC,$68,$CB,$63,$E2,$F8
	.dc.b	$9F,$20,$8E,$09,$1F,$22,$32,$D8,$CB,$F4,$A3,$2D,$8F,$A3,$FB,$77
	.dc.b	$D4,$23,$82,$47,$D4,$8C,$B6,$32,$FD,$E8,$CB,$63,$EE,$FD,$D7,$E0
	.dc.b	$08,$E0,$91,$F8,$08,$CB,$63,$2F,$E2,$A3,$2D,$8F,$C4,$FE,$97,$F8
	.dc.b	$C2,$38,$24,$7E,$32,$32,$D8,$CB,$F9,$68,$CB,$63,$FD,$77,$FA,$77
	.dc.b	$FA,$85,$B1,$FE,$85,$8C,$3F,$E0,$A6,$BF,$EA,$36,$BF,$EA,$86,$BF
	.dc.b	$EA,$B6,$A6,$91,$EC,$48,$F1,$FA,$FE,$12,$38,$24,$78,$E3,$2D,$8C
	.dc.b	$BE,$58,$CB,$63,$DF,$EB,$FB,$48,$E0,$91,$EF,$8C,$B6,$32,$FC,$18
	.dc.b	$CB,$63,$E5,$FE,$B5,$F2,$48,$E0,$91,$F2,$E3,$2D,$8C,$BF,$36,$32
	.dc.b	$D8,$FA,$FF,$84,$7D,$52,$38,$24,$7D,$78,$CB,$63,$2F,$D9,$8C,$B6
	.dc.b	$3F,$07,$E5,$FE,$04,$8E,$09,$1F,$83,$8C,$B6,$32,$FE,$16,$32,$D8
	.dc.b	$FC,$7F,$DB,$FE,$34,$8E,$09,$1F,$8F,$8C,$B6,$32,$FE,$46,$32,$D8
	.dc.b	$FC,$DF,$A7,$F9,$92,$38,$24,$7E,$6E,$32,$D8,$CB,$F9,$D8,$CB,$63
	.dc.b	$FC,$D6,$BE,$AF,$FC,$8C,$88,$E1,$BE,$0F,$FB,$BE,$31,$1C,$63,$04
	.dc.b	$9E,$B3,$A2,$46,$67,$FC,$F7,$19,$F4,$FA,$12,$B8,$B6,$76,$D7,$CC
	.dc.b	$79,$9C,$38,$94,$62,$35,$3E,$4F,$0B,$D4,$47,$04,$8F,$2C,$65,$B1
	.dc.b	$97,$C2,$7C,$98,$D3,$8A,$FF,$30,$3E,$EF,$07,$E4,$12,$BE,$2B,$63
	.dc.b	$51,$82,$38,$21,$D6,$38,$3D,$FC,$1F,$A8,$49,$FA,$2D,$8F,$9E,$D7
	.dc.b	$84,$68,$8E,$9A,$FA,$FC,$1F,$C0,$12,$3E,$EB,$63,$ED,$B5,$E1,$1A
	.dc.b	$23,$A6,$BF,$1F,$CD,$FC,$61,$23,$F1,$2D,$8F,$C3,$B5,$E1,$1A,$23
	.dc.b	$A6,$BF,$4F,$CD,$FC,$C1,$23,$F2,$AD,$8F,$C9,$B5,$E1,$1A,$23,$A6
	.dc.b	$BF,$7F,$D2,$FD,$21,$23,$F4,$2D,$8F,$CF,$B5,$E1,$1A,$23,$A6,$BF
	.dc.b	$8B,$FD,$27,$F6,$04,$8F,$D6,$B6,$3F,$56,$D7,$84,$68,$8E,$9A,$FE
	.dc.b	$5F,$CE,$FD,$E1,$23,$F7,$2D,$8F,$DB,$B5,$E1,$1A,$23,$A6,$BF,$A6
	.dc.b	$FA,$C7,$D3,$A7,$F8,$8C,$61,$B1,$99,$FF,$7F,$CE,$5B,$19,$7A,$35
	.dc.b	$FE,$07,$1B,$73,$3D,$08,$F2,$6B,$E5,$F4,$39,$3F,$C8,$65,$18,$84
	.dc.b	$62,$7D,$33,$98,$C4,$FA,$23,$04,$9D,$92,$34,$D7,$E3,$7E,$AD,$CC
	.dc.b	$91,$F1,$9B,$1F,$11,$AF,$08,$D1,$1D,$35,$FA,$5F,$86,$F9,$C9,$1F
	.dc.b	$49,$B1,$F4,$1A,$F0,$8D,$11,$D3,$5F,$BD,$F9,$B7,$C3,$24,$7D,$E6
	.dc.b	$C7,$DC,$6B,$C2,$34,$47,$4D,$7F,$15,$E9,$FC,$F2,$47,$E2,$9B,$1F
	.dc.b	$88,$6B,$C2,$34,$47,$4D,$7F,$2D,$E9,$7D,$B2,$47,$E5,$9B,$1F,$94
	.dc.b	$6B,$C2,$34,$47,$4D,$7F,$A3,$FD,$7B,$E1,$3F,$18,$D9,$FE,$E6,$D9
	.dc.b	$FE,$9A,$1C,$3C,$A7,$89,$27,$8B,$63,$8C,$F0,$7F,$B3,$B5,$FE,$FC
	.dc.b	$D7,$B7,$BD,$E4,$25,$79,$1B,$1D,$1A,$FB,$1A,$F6,$6B,$D9,$F7,$FD
	.dc.b	$1C,$9F,$7B,$67,$CE,$D7,$E3,$B5,$F4,$B5,$F4,$B5,$F8,$DE,$B7,$CC
	.dc.b	$24,$FC,$C6,$C7,$C4,$6B,$F5,$1A,$FC,$66,$BF,$19,$FA,$FF,$83,$49
	.dc.b	$FA,$ED,$9F,$9E,$D7,$EF,$B5,$FA,$2D,$7E,$8B,$5F,$BD,$EB,$7E,$10
	.dc.b	$93,$F8,$46,$C7,$DC,$6B,$F8,$C6,$BF,$79,$AF,$DE,$7F,$1F,$F0,$92
	.dc.b	$7F,$1E,$D9,$FC,$3B,$5F,$CB,$B5,$FC,$4B,$5F,$C4,$B5,$FC,$B7,$C0
	.dc.b	$FE,$70,$93,$F9,$C6,$C7,$E5,$1A,$FE,$91,$AF,$E5,$9A,$FE,$59,$FD
	.dc.b	$3F,$83,$27,$F4,$ED,$9F,$CF,$B5,$FD,$7B,$5F,$D0,$B5,$FD,$0B,$57
	.dc.b	$EA,$3F,$E5,$C2,$78,$5F,$D0,$5B,$5F,$E6,$86,$CF,$ED,$9A,$FF,$91
	.dc.b	$1A,$FF,$95,$9F,$F3,$54,$3D,$77,$8D,$FD,$42,$30,$D8,$CC,$7F,$02
	.dc.b	$32,$57,$35,$FE,$52,$8C,$B6,$32,$FF,$99,$A3,$2D,$8C,$BF,$EE,$76
	.dc.b	$BF,$DD,$5B,$1C,$23,$5A,$79,$49,$1E,$36,$C7,$85,$AF,$76,$BC,$A3
	.dc.b	$81,$1F,$95,$DD,$F3,$C9,$1F,$3D,$B1,$A8,$C1,$1C,$10,$FA,$51,$C1
	.dc.b	$F8,$1F,$4F,$F3,$49,$3F,$2D,$B1,$F2,$5A,$F0,$8D,$11,$D3,$5F,$99
	.dc.b	$E3,$7D,$92,$47,$D7,$6C,$7D,$56,$BC,$23,$44,$74,$D7,$F1,$9E,$AF
	.dc.b	$E4,$C9,$1F,$93,$6C,$7E,$4D,$AF,$08,$D1,$1D,$35,$FC,$26,$DF,$91
	.dc.b	$24,$7E,$3D,$B1,$F8,$D6,$BC,$23,$44,$74,$D7,$F2,$5F,$C0,$3F,$E4
	.dc.b	$18,$FC,$E3,$63,$F3,$4D,$77,$1F,$A5,$8C,$47,$FB,$56,$11,$88,$FF
	.dc.b	$26,$36,$32,$71,$21,$E0,$2C,$7F,$6E,$DB,$FD,$C5,$1F,$EE,$46,$C6
	.dc.b	$7C,$6F,$F2,$14,$62,$3F,$D8,$2D,$7F,$DA,$6D,$7F,$DD,$4D,$4D,$24
	.dc.b	$74,$24,$FC,$4F,$23,$E0,$11,$C1,$2B,$E3,$C6,$5B,$19,$7E,$7C,$65
	.dc.b	$B1,$F4,$3C,$7F,$98,$47,$04,$8F,$A7,$19,$6C,$65,$FB,$71,$96,$C7
	.dc.b	$C9,$EE,$F8,$E4,$70,$48,$FA,$11,$96,$C6,$5F,$9B,$19,$6C,$7E,$23
	.dc.b	$C6,$FC,$21,$1C,$12,$3F,$17,$19,$6C,$65,$FC,$9C,$65,$B1,$F9,$4D
	.dc.b	$FF,$20,$47,$04,$8F,$CB,$C6,$5B,$19,$7F,$3F,$19,$6C,$7E,$37,$CF
	.dc.b	$FC,$59,$1C,$12,$3F,$29,$19,$6C,$65,$FC,$8C,$65,$B1,$FA,$CF,$63
	.dc.b	$F5,$04,$70,$48,$FD,$7C,$65,$B1,$97,$F6,$F1,$96,$C7,$EE,$3D,$8F
	.dc.b	$DA,$11,$C1,$23,$F7,$71,$96,$C6,$5F,$E0,$71,$96,$C7,$F9,$81,$7D
	.dc.b	$93,$DD,$FF,$59,$36,$74,$D9,$FF,$33,$B5,$10,$E7,$3C,$89,$3C,$9B
	.dc.b	$1C,$9A,$F1,$38,$C7,$F1,$8F,$2B,$F8,$A1,$1C,$12,$3F,$8D,$C6,$5B
	.dc.b	$19,$7F,$91,$C6,$5B,$1F,$C9,$3E,$0F,$F9,$01,$1C,$12,$3F,$93,$C6
	.dc.b	$5B,$19,$7F,$97,$C6,$5B,$1F,$23,$BF,$E5,$91,$C1,$23,$E5,$C6,$5B
	.dc.b	$19,$7E,$26,$09,$D4,$61,$FE,$71,$ED,$CA,$FE,$7E,$D9,$FE,$7E,$D7
	.dc.b	$FA,$2B,$5F,$E7,$DC,$88,$C4,$7F,$3E,$6C,$FF,$42,$F2,$FF,$A3,$11
	.dc.b	$FE,$9A,$D9,$FE,$9A,$D7,$FA,$9B,$5E,$ED,$7F,$A5,$BF,$90,$F4,$24
	.dc.b	$FE,$41,$B3,$F9,$06,$BF,$11,$AF,$E4,$71,$46,$23,$F2,$B1,$89,$E8
	.dc.b	$7F,$8F,$FB,$FF,$A5,$26,$93,$9B,$FC,$90,$D8,$E1,$18,$23,$82,$57
	.dc.b	$F9,$1E,$79,$FB,$3C,$89,$3E,$16,$CE,$DA,$FF,$92,$E5,$18,$8F,$EE
	.dc.b	$91,$89,$E4,$48,$FF,$6D,$FD,$97,$F7,$74,$D2,$40,$FE,$04,$E1,$A6
	.dc.b	$63,$9B,$5F,$DB,$46,$DC,$C7,$03,$E6,$79,$9F,$20,$8E,$09,$1F,$16
	.dc.b	$32,$D8,$CB,$F0,$63,$2D,$8F,$B1,$F3,$1F,$50,$8E,$09,$1F,$46,$32
	.dc.b	$D8,$CB,$F3,$63,$2D,$8F,$87,$F0,$9E,$F8,$FE,$39,$68,$C4,$7F,$20
	.dc.b	$8C,$4F,$02,$47,$D2,$9D,$BF,$49,$B3,$C2,$35,$3C,$1F,$9B,$3A,$3F
	.dc.b	$23,$ED,$C9,$FC,$C3,$67,$F1,$CD,$7F,$10,$D7,$F0,$ED,$7F,$01,$8A
	.dc.b	$31,$3F,$A1,$F3,$FF,$3E,$47,$F4,$CD,$AF,$CE,$35,$FC,$B3,$5F,$CA
	.dc.b	$B5,$EE,$FC,$0E,$49,$3F,$90,$6C,$FE,$61,$AF,$E3,$9A,$FD,$26,$BF
	.dc.b	$45,$B1,$FB,$9F,$C2,$FF,$6E,$47,$F7,$CD,$9F,$DA,$35,$FD,$73,$5F
	.dc.b	$D6,$B5,$F8,$EF,$F0,$2E,$99,$3F,$C2,$5B,$3F,$BF,$6B,$FB,$96,$BF
	.dc.b	$B8,$6B,$F4,$9A,$BF,$53,$7B,$FF,$88,$6D,$7F,$77,$6C,$FF,$67,$38
	.dc.b	$4E,$C9,$3B,$6C,$6E,$78,$BF,$DB,$DA,$FF,$79,$6B,$FD,$F1,$B3,$FC
	.dc.b	$97,$D1,$FE,$46,$47,$F9,$53,$67,$F9,$03,$5F,$E3,$4D,$7F,$8C,$B5
	.dc.b	$FC,$9B,$FC,$BB,$EF,$64,$FF,$35,$6C,$FF,$2B,$6B,$FC,$95,$AF,$F2
	.dc.b	$46,$BF,$9E,$6B,$F3,$78,$3E,$61,$27,$E6,$36,$3E,$33,$5F,$8C,$D7
	.dc.b	$E3,$35,$F8,$CF,$F0,$48,$7D,$CF,$FA,$A6,$7F,$C3,$91,$87,$51,$97
	.dc.b	$FC,$51,$19,$3E,$E1,$5E,$31,$B7,$FD,$68,$D7,$F9,$D3,$63,$84,$74
	.dc.b	$8C,$11,$F0,$FA,$52,$7B,$CA,$31,$3E,$06,$BC,$DA,$F8,$9A,$F8,$9A
	.dc.b	$FE,$6F,$D4,$FC,$D1,$2B,$F3,$4D,$8F,$CA,$35,$FC,$A3,$5F,$CA,$35
	.dc.b	$FC,$A3,$F2,$BE,$32,$4F,$A9,$B3,$F2,$1A,$FC,$36,$BF,$29,$AF,$CA
	.dc.b	$6C,$7D,$AF,$77,$EB,$91,$F8,$ED,$9F,$AA,$D7,$E8,$35,$FA,$ED,$7E
	.dc.b	$BB,$FD,$12,$FE,$D4,$7D,$13,$04,$9F,$F3,$B4,$68,$91,$93,$F8,$13
	.dc.b	$88,$DC,$64,$FC,$17,$77,$E4,$09,$1F,$80,$6C,$7E,$39,$AF,$08,$D1
	.dc.b	$1D,$35,$FC,$77,$C2,$FE,$70,$91,$F8,$C6,$C7,$E6,$9A,$F0,$8D,$11
	.dc.b	$D3,$5F,$F4,$25,$3E,$83,$E4,$12,$7E,$2B,$67,$E1,$B5,$E2,$D7,$FD
	.dc.b	$2D,$3C,$E7,$89,$23,$F5,$FE,$8F,$ED,$89,$1F,$A6,$6C,$6A,$30,$47
	.dc.b	$04,$3F,$65,$1C,$1F,$DA,$7A,$1F,$C0,$89,$3F,$B2,$6C,$7F,$00,$6B
	.dc.b	$C2,$34,$47,$4D,$7E,$3C,$7E,$34,$91,$F8,$A6,$C7,$E2,$1A,$F0,$8D
	.dc.b	$11,$D3,$5F,$E1,$9E,$87,$F1,$A2,$47,$F0,$A6,$C7,$F1,$86,$BC,$23
	.dc.b	$44,$74,$D7,$F8,$C7,$A1,$FC,$98,$91,$FC,$59,$B1,$FC,$91,$AF,$08
	.dc.b	$D1,$1D,$35,$FC,$5F,$57,$EC,$49,$1F,$AE,$6C,$7E,$B1,$AF,$08,$D1
	.dc.b	$1D,$35,$FE,$61,$E7,$FF,$3E,$24,$7F,$2E,$6C,$7F,$3C,$6B,$C2,$34
	.dc.b	$47,$4D,$7F,$A3,$FC,$A7,$F9,$E6,$3F,$A3,$B6,$3F,$A7,$B5,$DC,$7F
	.dc.b	$55,$8C,$47,$FB,$E2,$11,$88,$FF,$42,$36,$32,$71,$21,$E0,$2C,$7F
	.dc.b	$84,$FE,$D9,$E6,$FF,$98,$1B,$3F,$E1,$79,$46,$1E,$13,$B2,$4E,$E7
	.dc.b	$52,$21,$CA,$76,$F2,$9E,$11,$88,$F8,$FE,$87,$C1,$24,$FC,$86,$C7
	.dc.b	$C8,$6B,$C2,$34,$47,$4D,$7E,$9F,$A1,$F3,$49,$1F,$51,$B1,$F5,$1A
	.dc.b	$F0,$8D,$11,$D3,$5F,$79,$F2,$3E,$3B,$E6,$12,$3E,$33,$63,$E9,$35
	.dc.b	$E3,$82,$31,$1F,$2A,$31,$3C,$89,$1F,$8A,$C3,$F1,$04,$8F,$C7,$B6
	.dc.b	$35,$18,$23,$82,$1F,$98,$8E,$0F,$E6,$BD,$0F,$CA,$12,$7F,$36,$D8
	.dc.b	$FC,$DB,$5E,$11,$A2,$3A,$6B,$F0,$3C,$AF,$C9,$12,$3F,$18,$D8,$FC
	.dc.b	$C3,$5E,$11,$A2,$3A,$6B,$FB,$2F,$43,$F5,$84,$8F,$D9,$B6,$3F,$66
	.dc.b	$D7,$84,$68,$8E,$9A,$FE,$FB,$D0,$FD,$C1,$23,$F7,$ED,$8F,$DF,$B5
	.dc.b	$E1,$1A,$23,$A6,$BF,$8C,$FB,$67,$C4,$7F,$B5,$9B,$3F,$B7,$6C,$FF
	.dc.b	$A3,$DA,$F1,$9E,$04,$9E,$0D,$8E,$0D,$78,$47,$F1,$8F,$17,$F8,$A1
	.dc.b	$1C,$12,$3F,$8D,$C6,$5B,$19,$7F,$91,$C6,$5B,$1F,$C9,$3D,$1F,$E4
	.dc.b	$04,$70,$48,$FE,$4F,$19,$6C,$65,$FE,$5F,$19,$6C,$7C,$5F,$1B,$E4
	.dc.b	$11,$C1,$23,$E4,$46,$5B,$19,$7E,$94,$65,$B1,$FC,$F3,$D0,$FE,$70
	.dc.b	$47,$04,$8F,$E7,$F1,$96,$C6,$5F,$E8,$F1,$96,$C7,$F4,$8D,$3F,$A2
	.dc.b	$11,$C1,$23,$FA,$5C,$65,$B1,$97,$FA,$9C,$65,$B1,$F8,$9F,$77,$F1
	.dc.b	$84,$70,$48,$FC,$64,$65,$B1,$97,$F2,$D1,$96,$C7,$FA,$9B,$E5,$5E
	.dc.b	$87,$FA,$C5,$B1,$98,$C1,$1C,$12,$7F,$D6,$B3,$C1,$B3,$B6,$CE,$DF
	.dc.b	$F5,$BC,$BD,$0F,$01,$27,$C8,$D9,$EA,$D7,$8B,$5F,$EE,$53,$CE,$78
	.dc.b	$92,$3F,$DF,$DF,$AA,$3C,$34,$FF,$6B,$8C,$36,$33,$3C,$1A,$FF,$4A
	.dc.b	$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7F,$B8,$CF,$27,$E1,$FA,$3F,$44
	.dc.b	$95,$F5,$5B,$1F,$39,$AF,$08,$D1,$1D,$35,$F9,$FE,$47,$DD,$24,$7E
	.dc.b	$05,$B1,$F6,$9A,$F0,$8D,$11,$D3,$5F,$89,$FB,$A7,$B5,$FF,$07,$5A
	.dc.b	$31,$08,$C4,$FD,$19,$CC,$62,$7E,$BC,$60,$93,$B2,$46,$9A,$FE,$33
	.dc.b	$D0,$FC,$D1,$23,$F3,$CD,$8F,$CB,$B5,$E1,$1A,$23,$A6,$BF,$98,$F4
	.dc.b	$3F,$4C,$48,$FD,$53,$63,$F4,$6D,$78,$46,$88,$E9,$AF,$E3,$7B,$FE
	.dc.b	$19,$23,$F2,$CD,$8F,$CD,$B5,$E1,$1A,$23,$A6,$BF,$B0,$E2,$FD,$F1
	.dc.b	$23,$F8,$13,$63,$F7,$6D,$78,$46,$88,$E9,$AF,$EF,$3D,$0F,$E1,$44
	.dc.b	$8F,$E1,$CD,$8F,$E0,$ED,$78,$46,$88,$E9,$AF,$F8,$EB,$E0,$9F,$0B
	.dc.b	$F8,$C6,$CF,$F9,$95,$B3,$FE,$40,$0E,$1E,$53,$C4,$93,$C5,$B1,$C6
	.dc.b	$78,$3F,$E6,$86,$BF,$E6,$B6,$CB,$81,$E8,$7F,$2B,$24,$7F,$2F,$6D
	.dc.b	$7F,$28,$6B,$FC,$91,$AF,$F1,$C6,$BF,$C6,$27,$AC,$FF,$39,$F4,$3F
	.dc.b	$9C,$11,$C1,$27,$F9,$AC,$65,$B1,$97,$F9,$9C,$65,$B1,$F2,$FB,$3E
	.dc.b	$71,$1C,$12,$3E,$34,$65,$B1,$97,$EC,$46,$5B,$1F,$57,$E4,$BF,$D2
	.dc.b	$6D,$8F,$F5,$61,$FE,$78,$8C,$B4,$C9,$FE,$9A,$72,$D3,$27,$FA,$7D
	.dc.b	$CB,$4C,$87,$2D,$32,$E0,$8B,$CB,$FE,$3D,$1F,$EA,$B6,$C6,$67,$D5
	.dc.b	$18,$23,$82,$47,$AE,$32,$D8,$CB,$F9,$FF,$95,$FC,$F1,$23,$F3,$CD
	.dc.b	$8F,$CB,$35,$E1,$1A,$23,$A6,$BF,$1F,$D0,$F4,$92,$3E,$93,$63,$E8
	.dc.b	$35,$E1,$1A,$23,$A6,$BF,$4F,$9F,$E2,$92,$3E,$F3,$63,$EE,$35,$E1
	.dc.b	$1A,$23,$A6,$BF,$EB,$CF,$B8,$74,$7F,$4E,$6C,$FF,$B9,$DB,$3F,$C0
	.dc.b	$DA,$F1,$9E,$04,$9E,$0D,$8E,$0D,$78,$47,$E5,$FD,$1F,$C0,$91,$C1
	.dc.b	$23,$F3,$11,$96,$C6,$5F,$CC,$46,$5B,$1F,$A3,$F4,$7F,$1A,$47,$04
	.dc.b	$8F,$D2,$46,$5B,$19,$7F,$49,$19,$6C,$7C,$5F,$8E,$F9,$04,$70,$48
	.dc.b	$F9,$11,$96,$C6,$5F,$A5,$19,$6C,$7E,$EF,$D1,$FD,$29,$1C,$12,$3F
	.dc.b	$79,$19,$6C,$65,$FD,$E4,$65,$B1,$FC,$1F,$D1,$FD,$89,$1C,$12,$3F
	.dc.b	$84,$46,$5B,$19,$7F,$84,$46,$5B,$1F,$89,$E1,$FC,$61,$1C,$12,$3F
	.dc.b	$19,$19,$6C,$65,$FC,$B4,$65,$B1,$FC,$7F,$D1,$FE,$12,$47,$04,$8F
	.dc.b	$E4,$11,$96,$C6,$5F,$E4,$11,$96,$C7,$F2,$BF,$47,$F8,$A9,$1C,$12
	.dc.b	$3F,$96,$46,$5B,$19,$7F,$96,$46,$5B,$1F,$DB,$57,$DC,$3E,$1B,$FC
	.dc.b	$7E,$D9,$D3,$67,$FB,$A3,$51,$0E,$73,$C8,$93,$C9,$B1,$C9,$AF,$13
	.dc.b	$8C,$7F,$4B,$EC,$FE,$6A,$47,$04,$8F,$E9,$91,$96,$C6,$5F,$E9,$91
	.dc.b	$96,$C7,$F5,$9F,$AD,$FF,$45,$B4,$FE,$8A,$E2,$3F,$AD,$35,$FE,$B0
	.dc.b	$D7,$FA,$B3,$5F,$EA,$8D,$E6,$34,$93,$D8,$87,$CE,$D7,$E8,$91,$C1
	.dc.b	$2B,$E8,$C6,$5B,$19,$7E,$56,$29,$D4,$61,$F6,$FA,$52,$BE,$13,$67
	.dc.b	$DE,$D7,$DE,$D7,$DE,$D7,$DE,$DC,$3D,$FE,$97,$CA,$25,$7C,$F6,$D7
	.dc.b	$CC,$6B,$F3,$1A,$FC,$C6,$BF,$31,$FC,$AF,$C2,$C9,$FC,$AB,$67,$F2
	.dc.b	$AD,$7E,$1B,$5F,$CB,$62,$8C,$47,$E6,$A3,$13,$D0,$FC,$37,$A3,$F8
	.dc.b	$42,$38,$24,$FE,$1A,$32,$D8,$CB,$F8,$88,$CB,$63,$F2,$5E,$8F,$E4
	.dc.b	$08,$E0,$91,$F9,$28,$CB,$63,$2F,$E5,$23,$2D,$8F,$DF,$7E,$68,$C7
	.dc.b	$FD,$76,$D8,$CC,$60,$8E,$09,$3F,$E7,$39,$E0,$D9,$DB,$67,$6F,$E9
	.dc.b	$BD,$2F,$D5,$92,$3F,$60,$D8,$FD,$6B,$5E,$11,$A2,$3A,$6B,$FB,$2E
	.dc.b	$EF,$DB,$92,$3F,$78,$D8,$FD,$CB,$5E,$11,$A2,$3A,$6B,$FC,$4E,$DF
	.dc.b	$19,$F2,$09,$3F,$31,$B3,$F1,$5A,$F1,$6B,$FC,$96,$79,$CF,$12,$47
	.dc.b	$F0,$EF,$47,$F8,$89,$23,$F8,$B3,$63,$51,$82,$38,$21,$FC,$5A,$38
	.dc.b	$3F,$C6,$3D,$1F,$E3,$A4,$9F,$E4,$4D,$8F,$E3,$ED,$78,$46,$88,$E9
	.dc.b	$AF,$C7,$F4,$7F,$1A,$48,$FC,$8B,$63,$F1,$4D,$78,$46,$88,$E9,$AF
	.dc.b	$F3,$08,$FE,$6A,$48,$FE,$74,$D8,$FE,$6E,$D7,$84,$68,$8E,$9A,$FF
	.dc.b	$51,$E1,$79,$E9,$FE,$CD,$18,$6C,$66,$7F,$DA,$D3,$96,$C6,$5E,$8D
	.dc.b	$7F,$B3,$46,$DC,$CF,$42,$3C,$9A,$FF,$50,$F4,$7F,$AA,$92,$3F,$AD
	.dc.b	$36,$3F,$AB,$B5,$E1,$1A,$23,$A6,$BF,$07,$F2,$7F,$33,$FD,$F2,$91
	.dc.b	$88,$46,$27,$E2,$CE,$63,$13,$F1,$23,$04,$9D,$92,$34,$D7,$E9,$78
	.dc.b	$1C,$C9,$1F,$49,$B1,$F4,$1A,$F0,$8D,$11,$D3,$5F,$BD,$E9,$F9,$C9
	.dc.b	$1F,$79,$B1,$F7,$1A,$F0,$8D,$11,$D3,$5F,$C5,$7A,$3F,$0C,$91,$F8
	.dc.b	$A6,$C7,$E2,$1A,$F0,$8D,$11,$D3,$5F,$CB,$7C,$27,$CF,$24,$7E,$59
	.dc.b	$B1,$F9,$46,$BC,$23,$44,$74,$D7,$F4,$5E,$87,$DB,$24,$7E,$89,$B1
	.dc.b	$FA,$06,$BC,$23,$44,$74,$D7,$F5,$DF,$B3,$FE,$1C,$91,$FA,$E6,$C7
	.dc.b	$EB,$1A,$F0,$8D,$11,$D3,$5F,$DD,$7A,$3F,$93,$24,$7E,$E9,$B1,$FB
	.dc.b	$86,$BC,$23,$44,$74,$D7,$F8,$37,$89,$F9,$F2,$47,$F0,$66,$C7,$F0
	.dc.b	$46,$BC,$23,$44,$74,$D7,$F8,$97,$A3,$FA,$B2,$47,$F1,$26,$C7,$F1
	.dc.b	$06,$BC,$23,$44,$74,$D7,$F8,$F7,$B9,$FB,$72,$47,$F1,$E6,$C7,$F1
	.dc.b	$C6,$BC,$23,$44,$74,$D7,$F9,$57,$A5,$FC,$0C,$91,$FC,$A9,$B1,$FC
	.dc.b	$A1,$AF,$08,$D1,$1D,$35,$FE,$6D,$FA,$47,$F0,$F2,$47,$F3,$66,$C7
	.dc.b	$F3,$46,$BC,$23,$44,$74,$D7,$FA,$17,$A3,$FC,$6C,$91,$FD,$09,$B1
	.dc.b	$FD,$01,$AF,$08,$D1,$1D,$35,$FE,$9D,$F2,$9F,$C9,$C9,$1F,$D3,$9B
	.dc.b	$1F,$D3,$1A,$F0,$8D,$11,$D3,$5F,$EB,$5E,$4F,$F3,$32,$47,$F5,$A6
	.dc.b	$C7,$F5,$86,$BC,$23,$44,$74,$D7,$FB,$37,$B1,$FC,$FC,$91,$FD,$99
	.dc.b	$B1,$FD,$91,$AF,$08,$D1,$1D,$35,$FF,$61,$7E,$6D,$FE,$CC,$8F,$F6
	.dc.b	$B3,$63,$FD,$8C,$1C,$3C,$0F,$EA,$CE,$1F,$F6,$A3,$5F,$F6,$BB,$5E
	.dc.b	$47,$02,$47,$81,$A7,$82,$1F,$11,$D8,$93,$E6,$6C,$ED,$AF,$B1,$AF
	.dc.b	$91,$AF,$91,$F4,$FA,$92,$7E,$0B,$67,$C4,$D7,$E3,$B5,$F7,$B5,$F7
	.dc.b	$B5,$F9,$1F,$29,$F1,$89,$3F,$39,$B1,$EF,$6B,$F5,$1A,$FC,$C6,$BF
	.dc.b	$31,$FA,$3E,$A4,$9F,$B2,$D9,$F9,$4D,$7E,$FB,$5F,$AE,$D7,$EB,$B5
	.dc.b	$FC,$07,$9B,$F7,$89,$3F,$86,$6C,$7D,$76,$BF,$8C,$6B,$F8,$46,$BF
	.dc.b	$84,$7F,$13,$EA,$49,$FC,$8B,$67,$F0,$4D,$7F,$2E,$D7,$F1,$ED,$7F
	.dc.b	$1E,$D7,$F3,$1F,$03,$F9,$62,$4F,$E7,$9B,$1F,$8F,$6B,$FA,$46,$BF
	.dc.b	$9C,$6B,$F9,$C7,$F4,$3E,$A4,$9F,$D4,$B6,$7F,$34,$D7,$F5,$ED,$7F
	.dc.b	$4E,$D7,$F4,$ED,$7F,$61,$F2,$9F,$AE,$24,$FE,$D9,$B1,$FA,$76,$BF
	.dc.b	$BC,$6B,$FB,$46,$BF,$B4,$7F,$73,$EA,$C9,$FE,$02,$D9,$FD,$93,$5F
	.dc.b	$E0,$ED,$7F,$7E,$D7,$F7,$ED,$7F,$84,$7C,$87,$F0,$62,$4F,$F0,$E6
	.dc.b	$C7,$EF,$DA,$FF,$14,$6B,$FC,$31,$AF,$F0,$C7,$F8,$8F,$A3,$27,$F8
	.dc.b	$CB,$67,$F8,$53,$5F,$E3,$ED,$7F,$8B,$B5,$FE,$2E,$D7,$F9,$07,$B9
	.dc.b	$FC,$78,$93,$FC,$99,$B1,$FC,$5D,$AF,$F2,$C6,$BF,$C9,$1A,$FF,$24
	.dc.b	$7F,$94,$FA,$F2,$7F,$98,$B6,$7F,$91,$35,$FE,$6E,$D7,$F9,$7B,$5F
	.dc.b	$E5,$ED,$7F,$9C,$7C,$57,$F3,$62,$4F,$F3,$E6,$C7,$F2,$F6,$BF,$D1
	.dc.b	$1A,$FF,$3C,$6B,$FC,$F1,$FE,$83,$EC,$49,$FE,$92,$D9,$FE,$74,$D7
	.dc.b	$FA,$7B,$5F,$E8,$ED,$7F,$A3,$B5,$FE,$A1,$F2,$1F,$D3,$89,$3F,$D5
	.dc.b	$9B,$1F,$D1,$DA,$FF,$5C,$6B,$FD,$51,$AF,$F5,$47,$FA,$CF,$B3,$27
	.dc.b	$FB,$0B,$67,$FA,$93,$5F,$EC,$ED,$7F,$AF,$B5,$FE,$BE,$D7,$FB,$47
	.dc.b	$AD,$FD,$98,$93,$FD,$B9,$B1,$FD,$7D,$AF,$F7,$46,$BF,$DB,$1A,$FF
	.dc.b	$6C,$7F,$B8,$FB,$32,$7F,$BC,$B6,$7F,$B5,$35,$FE,$FE,$D7,$FB,$BB
	.dc.b	$5F,$EE,$ED,$7F,$C0,$1F,$0B,$FD,$F8,$93,$FE,$06,$6C,$7F,$77,$6B
	.dc.b	$FE,$0C,$6B,$FE,$04,$6B,$FE,$04,$7F,$DB,$DF,$9B,$7F,$C0,$19,$FF
	.dc.b	$07,$CA,$30,$FF,$84,$9A,$88,$6E,$39,$90,$E7,$1C,$23,$6D,$79,$35
	.dc.b	$E5,$1C,$0E,$4F,$87,$E3,$BC,$E4,$8E,$AD,$8D,$46,$08,$E0,$87,$A0
	.dc.b	$F7,$47,$17,$13,$FF,$00,$27,$D3,$E6,$FC,$62,$38,$24,$FC,$58,$CB
	.dc.b	$63,$2F,$48,$CB,$DB,$E5,$7E,$89,$27,$E3,$B6,$3E,$83,$5E,$11,$A2
	.dc.b	$3A,$6B,$EA,$F6,$3E,$E9,$23,$E9,$B6,$3E,$E3,$5E,$11,$A2,$3A,$6B
	.dc.b	$F1,$BE,$0B,$F1,$24,$8F,$BE,$D8,$FC,$43,$5E,$11,$A2,$3A,$6B,$F4
	.dc.b	$BD,$BF,$CA,$92,$3F,$16,$D8,$FC,$A3,$5E,$11,$A2,$3A,$6B,$F7,$BE
	.dc.b	$33,$F4,$24,$8F,$CB,$B6,$3F,$40,$D7,$84,$68,$8E,$9A,$FE,$2B,$DC
	.dc.b	$FD,$69,$23,$F4,$6D,$8F,$D6,$35,$E1,$1A,$23,$A6,$BF,$96,$F9,$8F
	.dc.b	$DC,$92,$3F,$5E,$D8,$FD,$C3,$5E,$11,$A2,$3A,$6B,$FA,$2F,$77,$F8
	.dc.b	$29,$23,$F7,$6D,$8F,$E0,$8D,$78,$46,$88,$E9,$AF,$EB,$BD,$CF,$E2
	.dc.b	$24,$8F,$E0,$ED,$8F,$E2,$0D,$78,$46,$88,$E9,$AF,$EE,$BD,$DF,$E3
	.dc.b	$A4,$8F,$E2,$6D,$8F,$E3,$8D,$78,$46,$88,$E9,$AF,$F0,$6F,$89,$FE
	.dc.b	$52,$48,$FE,$3E,$D8,$FE,$50,$D7,$84,$68,$8E,$9A,$FF,$12,$F7,$7F
	.dc.b	$9A,$92,$3F,$95,$B6,$3F,$9A,$35,$E1,$1A,$23,$A6,$BF,$C7,$BE,$47
	.dc.b	$FA,$09,$23,$F9,$BB,$63,$FA,$03,$5E,$11,$A2,$3A,$6B,$FC,$AB,$DE
	.dc.b	$FE,$9A,$48,$FE,$86,$D8,$FE,$98,$D7,$84,$68,$8E,$9A,$FF,$36,$F9
	.dc.b	$6F,$EB,$24,$8F,$E9,$ED,$8F,$EB,$0D,$78,$46,$88,$E9,$AF,$F4,$2F
	.dc.b	$7F,$FB,$29,$23,$FA,$DB,$63,$FB,$23,$5E,$11,$A2,$3A,$6B,$FD,$3B
	.dc.b	$E1,$7F,$B8,$92,$3F,$B3,$B6,$3F,$B8,$35,$E1,$1A,$23,$A6,$BF,$D6
	.dc.b	$BD,$FF,$EF,$A4,$8F,$EE,$6D,$8F,$EF,$8D,$78,$46,$88,$E9,$AF,$F6
	.dc.b	$9F,$CD,$9F,$20,$9F,$F0,$4C,$61,$B1,$99,$FE,$E0,$27,$4E,$1A,$F5
	.dc.b	$8E,$0D,$9E,$64,$78,$46,$9F,$F0,$9C,$F2,$6C,$65,$FF,$0C,$43,$E0
	.dc.b	$B1,$1E,$36,$C6,$6B,$D7,$18,$23,$82,$4F,$A6,$32,$D8,$CB,$F1,$FE
	.dc.b	$27,$A9,$23,$E3,$B6,$3E,$13,$5E,$11,$A2,$3A,$6B,$F4,$FE,$2F,$D2
	.dc.b	$48,$FA,$6D,$8F,$9C,$D7,$84,$68,$8E,$9A,$FD,$FF,$84,$F8,$A4,$8F
	.dc.b	$BE,$D8,$FB,$4D,$78,$46,$88,$E9,$AF,$E2,$FE,$13,$E8,$92,$3F,$16
	.dc.b	$D8,$FC,$33,$5E,$11,$A2,$3A,$6B,$F9,$7F,$80,$FB,$A4,$8F,$CB,$B6
	.dc.b	$3F,$15,$92,$3F,$31,$3C,$4F,$F3,$01,$67,$71,$C5,$C1,$0F,$C9,$FC
	.dc.b	$07,$E3,$09,$3F,$A6,$6C,$78,$1E,$EF,$08,$EE,$D3,$4F,$12,$3F,$A9
	.dc.b	$8F,$CC,$FC,$04,$9F,$D9,$B6,$7F,$5C,$F6,$7F,$46,$D7,$F6,$8D,$B3
	.dc.b	$4E,$51,$FC,$05,$AF,$EC,$3E,$03,$F5,$24,$AF,$E0,$6D,$8F,$9A,$F5
	.dc.b	$24,$7B,$A3,$F6,$B1,$B7,$F8,$73,$5F,$DE,$FC,$07,$ED,$89,$3F,$C4
	.dc.b	$1B,$1F,$69,$F5,$12,$3B,$47,$F1,$18,$DB,$FC,$6D,$AF,$F0,$AF,$80
	.dc.b	$FE,$06,$49,$FE,$3A,$D8,$FC,$3B,$F1,$C9,$1D,$B0,$46,$E7,$F9,$43
	.dc.b	$5F,$E2,$FF,$01,$FC,$40,$95,$FC,$A9,$B1,$F9,$47,$EA,$12,$3B,$3A
	.dc.b	$8F,$E6,$8D,$7F,$D6,$BF,$92,$3C,$47,$B9,$FE,$BE,$8C,$36,$33,$3F
	.dc.b	$CD,$A7,$2D,$8C,$BF,$EA,$98,$CB,$63,$2F,$81,$AF,$FB,$1E,$36,$E6
	.dc.b	$7C,$04,$7C,$1E,$2F,$F4,$C6,$F2,$F9,$E3,$11,$E1,$6B,$EE,$6B,$DD
	.dc.b	$AA,$91,$86,$CF,$52,$4F,$80,$87,$FB,$1F,$F3,$66,$DF,$EE,$68,$46
	.dc.b	$1F,$F7,$4B,$5F,$EB,$ED,$79,$47,$FB,$A3,$3B,$1F,$94,$9E,$11,$88
	.dc.b	$FE,$B9,$1C,$5F,$F7,$7B,$5F,$F7,$93,$5F,$1C,$F9,$61,$B5,$24,$F9
	.dc.b	$9A,$F6,$7C,$8F,$66,$BD,$9A,$FC,$07,$8B,$F1,$5A,$F9,$B5,$F1,$11
	.dc.b	$F8,$AD,$7D,$AF,$49,$FD,$3C,$7C,$38,$C1,$B7,$E9,$35,$F8,$7A,$FB
	.dc.b	$C8,$FD,$26,$BF,$29,$EC,$47,$E8,$3A,$7E,$F3,$5F,$9F,$F0,$5F,$2C
	.dc.b	$8F,$DE,$6B,$F5,$9F,$51,$1F,$B8,$E9,$FC,$53,$5F,$B7,$F0,$5F,$5C
	.dc.b	$8F,$E2,$9A,$FE,$09,$F8,$C4,$7F,$10,$E9,$FC,$B3,$5F,$C3,$FC,$17
	.dc.b	$E0,$C8,$FE,$59,$AF,$E3,$9F,$A4,$47,$F2,$8E,$9F,$D1,$35,$13,$F3
	.dc.b	$67,$C0,$7E,$95,$B1,$FE,$40,$0E,$1F,$F2,$1B,$5E,$67,$38,$E4,$47
	.dc.b	$94,$6D,$DB,$5E,$06,$A3,$83,$DB,$E0,$3C,$84,$79,$B5,$D1,$82,$18
	.dc.b	$21,$E6,$3D,$6F,$17,$12,$F8,$0F,$84,$48,$F4,$B5,$F8,$2D,$78,$B5
	.dc.b	$F6,$B5,$F9,$CF,$CD,$F8,$09,$1F,$11,$AF,$CC,$6B,$E4,$6B,$F2,$1A
	.dc.b	$FD,$96,$BF,$5B,$E0,$3E,$D1,$1F,$A2,$D7,$EC,$B5,$F8,$2D,$7E,$AB
	.dc.b	$5F,$C3,$3F,$85,$F8,$09,$1F,$71,$AF,$E1,$1A,$FC,$C6,$BF,$80,$6B
	.dc.b	$F9,$16,$BF,$8E,$F8,$0F,$C9,$11,$FC,$4B,$5F,$C8,$B5,$FB,$2D,$7F
	.dc.b	$1A,$D7,$F3,$CF,$E7,$7E,$02,$47,$E5,$1A,$FE,$71,$AF,$E1,$1A,$FE
	.dc.b	$61,$AF,$EA,$5A,$FE,$9B,$E0,$3F,$54,$47,$F4,$2D,$7F,$52,$D7,$F2
	.dc.b	$2D,$7F,$4A,$D7,$F6,$CF,$ED,$7E,$02,$47,$EB,$1A,$FE,$D1,$AF,$E7
	.dc.b	$1A,$FE,$C1,$AF,$F0,$16,$BF,$BE,$F8,$0F,$E0,$44,$7F,$72,$D7,$F8
	.dc.b	$0B,$5F,$D4,$B5,$FD,$EB,$5F,$E1,$CF,$F0,$DF,$80,$91,$FC,$11,$AF
	.dc.b	$F0,$C6,$BF,$B4,$6B,$FC,$21,$AF,$F1,$96,$BF,$C5,$BE,$03,$F8,$D1
	.dc.b	$1F,$E2,$2D,$7F,$8C,$B5,$FE,$02,$D7,$F8,$AB,$5F,$E4,$CF,$F2,$5F
	.dc.b	$80,$91,$FC,$71,$AF,$F2,$46,$BF,$C3,$1A,$FF,$20,$6B,$FC,$C5,$AF
	.dc.b	$F2,$EF,$80,$FE,$64,$47,$F9,$4B,$5F,$E6,$2D,$7F,$8C,$B5,$FE,$5A
	.dc.b	$D7,$F9,$F3,$FC,$F7,$E0,$24,$7F,$34,$6B,$FC,$F1,$AF,$F2,$46,$BF
	.dc.b	$CE,$1A,$FF,$49,$6B,$FD,$1B,$E0,$3F,$A5,$11,$FE,$82,$D7,$FA,$4B
	.dc.b	$5F,$E6,$2D,$7F,$A2,$B5,$FE,$AC,$FF,$55,$F8,$09,$1F,$D3,$1A,$FF
	.dc.b	$54,$6B,$FC,$F1,$AF,$F5,$06,$BF,$D8,$5A,$FF,$5E,$F8,$0F,$EC,$44
	.dc.b	$7F,$AC,$B5,$FE,$C2,$D7,$FA,$4B,$5F,$EB,$AD,$7F,$B7,$3F,$E5,$1F
	.dc.b	$2D,$E0,$52,$4F,$F6,$F8,$CB,$63,$2F,$F5,$91,$1A,$70,$D7,$FA,$F3
	.dc.b	$5F,$EE,$AD,$7F,$C0,$2D,$79,$C7,$97,$E0,$24,$78,$DA,$F6,$6B,$D9
	.dc.b	$AF,$66,$BD,$9A,$FA,$3D,$FF,$84,$47,$E0,$35,$F5,$B5,$F5,$B5,$F5
	.dc.b	$B5,$F5,$BF,$37,$C7,$91,$F2,$DA,$FC,$66,$BF,$19,$AF,$C6,$6B,$F1
	.dc.b	$9A,$FD,$0F,$1B,$ED,$11,$FB,$0D,$7E,$9B,$5F,$A6,$D7,$E9,$B5,$FA
	.dc.b	$6F,$E1,$7D,$F9,$1F,$83,$6B,$F7,$9A,$FD,$E6,$BF,$79,$AF,$DE,$6B
	.dc.b	$F8,$8F,$7F,$F2,$44,$7F,$20,$D7,$F1,$6D,$7F,$16,$D7,$F1,$6D,$7F
	.dc.b	$16,$FE,$77,$DF,$91,$F9,$B6,$BF,$96,$6B,$F9,$66,$BF,$96,$6B,$F9
	.dc.b	$66,$BF,$A0,$F7,$FF,$54,$47,$F5,$0D,$7F,$46,$D7,$F4,$6D,$7F,$46
	.dc.b	$D7,$F4,$6F,$ED,$7D,$F9,$1F,$B3,$6B,$FA,$E6,$BF,$AE,$6B,$FA,$E6
	.dc.b	$BF,$AE,$6B,$FB,$8F,$7F,$F8,$11,$1F,$E0,$0D,$7F,$76,$D7,$F7,$6D
	.dc.b	$7F,$76,$D7,$F7,$6F,$F0,$DF,$7E,$47,$F0,$B6,$BF,$C1,$9A,$FF,$06
	.dc.b	$6B,$FC,$19,$AF,$F0,$66,$BF,$C4,$3D,$FF,$E3,$44,$7F,$8C,$35,$FE
	.dc.b	$26,$D7,$F8,$9B,$5F,$E2,$6D,$7F,$89,$BF,$C9,$7D,$F9,$1F,$C8,$DA
	.dc.b	$FF,$1E,$6B,$FC,$79,$AF,$F1,$E6,$BF,$C7,$9A,$FF,$28,$F7,$FF,$99
	.dc.b	$11,$FE,$60,$D7,$F9,$5B,$5F,$E5,$6D,$7F,$95,$B5,$FE,$56,$FF,$3D
	.dc.b	$F7,$E4,$7F,$3B,$6B,$FC,$D9,$AF,$F3,$66,$BF,$CD,$9A,$FF,$36,$6B
	.dc.b	$FD,$03,$DF,$FE,$94,$47,$FA,$43,$5F,$E8,$6D,$7F,$A1,$B5,$FE,$86
	.dc.b	$D7,$FA,$1B,$FD,$57,$DF,$91,$FD,$4D,$AF,$F4,$E6,$BF,$D3,$9A,$FF
	.dc.b	$4E,$6B,$FD,$39,$AF,$F5,$8F,$7F,$FB,$11,$1F,$EC,$0D,$7F,$AD,$B5
	.dc.b	$FE,$B6,$D7,$FA,$DB,$5F,$EB,$6F,$F6,$DF,$7E,$47,$F6,$B6,$BF,$D9
	.dc.b	$9A,$FF,$66,$6B,$FD,$99,$AF,$F6,$66,$BF,$DC,$3D,$FF,$EF,$44,$7F
	.dc.b	$BC,$35,$FE,$E6,$D7,$FB,$9B,$5F,$EE,$6D,$7F,$B9,$BF,$EE,$F5,$F9
	.dc.b	$B3,$D6,$C3,$63,$FD,$E8,$D7,$FC,$08,$D7,$FD,$BC,$D7,$FD,$EA,$D7
	.dc.b	$FE,$00,$B5,$EF,$1D,$88,$F6,$6B,$DB,$E0,$3C,$64,$7C,$6D,$74,$60
	.dc.b	$86,$08,$79,$9E,$0F,$A3,$E0,$3D,$E4,$7D,$ED,$7D,$ED,$78,$1A,$21
	.dc.b	$A2,$9F,$13,$B3,$E5,$91,$F9,$6D,$7E,$5B,$5E,$06,$88,$68,$A7,$D0
	.dc.b	$F8,$0F,$AE,$47,$EB,$B5,$FA,$ED,$78,$1A,$21,$A2,$9F,$73,$E0,$3F
	.dc.b	$06,$47,$F0,$6D,$7F,$06,$D7,$81,$A2,$1A,$29,$F8,$8F,$88,$FC,$79
	.dc.b	$1F,$C7,$B5,$FC,$7B,$5E,$06,$88,$68,$A7,$E5,$3E,$23,$F3,$64,$7F
	.dc.b	$36,$D7,$F3,$6D,$78,$1A,$21,$A2,$9F,$A0,$F8,$8F,$D3,$91,$FD,$3B
	.dc.b	$5F,$D3,$B5,$E0,$68,$86,$8A,$7E,$B3,$E2,$3F,$66,$47,$F6,$6D,$7F
	.dc.b	$66,$D7,$81,$A2,$1A,$29,$FB,$8F,$84,$FD,$F9,$1F,$DF,$B5,$FD,$FB
	.dc.b	$5E,$06,$88,$68,$A7,$F0,$4F,$84,$FE,$16,$47,$F8,$5B,$5F,$E1,$6D
	.dc.b	$78,$1A,$21,$A2,$9F,$C4,$3E,$13,$F8,$B9,$1F,$E2,$ED,$7F,$8B,$B5
	.dc.b	$E0,$68,$86,$8A,$7F,$1C,$F8,$1F,$E4,$64,$7F,$91,$B5,$FE,$46,$D7
	.dc.b	$81,$A2,$1A,$29,$FC,$A3,$E0,$7F,$97,$91,$FE,$5E,$D7,$F9,$7B,$5E
	.dc.b	$06,$88,$68,$A7,$F3,$4F,$80,$FE,$76,$47,$F9,$DB,$5F,$E7,$6D,$78
	.dc.b	$1A,$21,$A2,$9F,$D0,$3E,$03,$FA,$39,$1F,$E8,$ED,$7F,$A3,$B5,$E0
	.dc.b	$68,$86,$8A,$7F,$4C,$F8,$0F,$EA,$64,$7F,$A9,$B5,$FE,$A6,$D7,$81
	.dc.b	$A2,$1A,$29,$FD,$63,$E0,$3F,$AF,$91,$FE,$BE,$D7,$FA,$FB,$5E,$06
	.dc.b	$88,$68,$A7,$F6,$4F,$80,$FE,$D6,$47,$FB,$5B,$5F,$ED,$6D,$78,$1A
	.dc.b	$21,$A2,$9F,$DC,$3E,$23,$FB,$B9,$1F,$EE,$ED,$7F,$BB,$B5,$E0,$68
	.dc.b	$86,$8A,$08,$3F,$27,$FF,$04,$34,$FF,$05,$38,$23,$82,$1C,$8E,$6E
	.dc.b	$9C,$1D,$DC,$38,$7A,$7C,$2F,$62,$47,$03,$C8,$E1,$E0,$D7,$C1,$08
	.dc.b	$C4,$78,$9A,$F8,$5F,$5F,$C4,$48,$ED,$28,$C4,$79,$5A,$F6,$6B,$EE
	.dc.b	$6B,$ED,$6B,$F2,$FE,$13,$E4,$91,$EE,$D7,$E1,$35,$F5,$B5,$F9,$6D
	.dc.b	$7E,$53,$F5,$3E,$12,$47,$A9,$AF,$CD,$6B,$F1,$9A,$FD,$66,$BF,$55
	.dc.b	$AB,$F5,$E7,$A5,$FE,$4B,$6C,$7D,$56,$BF,$79,$AF,$E0,$DA,$FE,$39
	.dc.b	$AF,$E4,$5A,$F7,$8E,$C4,$7B,$35,$FC,$47,$C1,$7E,$50,$8F,$DA,$6B
	.dc.b	$A3,$04,$30,$43,$F1,$2F,$07,$F3,$5F,$02,$F6,$3F,$CD,$4D,$8F,$CA
	.dc.b	$35,$FC,$EB,$5E,$31,$C0,$8F,$06,$BC,$1A,$F0,$3F,$27,$F0,$1F,$B0
	.dc.b	$23,$82,$1F,$A9,$72,$D3,$27,$E9,$1C,$B4,$F4,$BE,$C2,$38,$21,$EC
	.dc.b	$72,$D3,$27,$C6,$72,$D3,$E2,$D7,$C8,$23,$82,$1F,$21,$CB,$4C,$9F
	.dc.b	$49,$CB,$4F,$DB,$FB,$5F,$C5,$08,$E0,$87,$F0,$D7,$2D,$32,$7F,$08
	.dc.b	$72,$D3,$F8,$87,$EB,$CE,$0F,$F5,$53,$63,$31,$82,$38,$21,$FC,$82
	.dc.b	$38,$36,$36,$D3,$67,$F2,$98,$72,$78,$09,$1E,$46,$BD,$5A,$F1,$6B
	.dc.b	$FC,$F2,$39,$BC,$48,$7F,$34,$FC,$E1,$EC,$7F,$B1,$DB,$1F,$CC,$5A
	.dc.b	$EE,$34,$47,$4D,$74,$60,$86,$0F,$89,$DD,$F3,$08,$FD,$06,$BA,$30
	.dc.b	$43,$04,$3E,$43,$C1,$F9,$5E,$17,$D8,$23,$F7,$1A,$FD,$56,$BC,$0D
	.dc.b	$10,$D1,$4F,$EB,$D4,$F4,$BE,$41,$23,$E4,$35,$F9,$CD,$78,$B5,$FE
	.dc.b	$E3,$1C,$DE,$24,$3F,$B4,$FE,$4E,$F3,$7F,$DF,$8D,$8F,$EC,$CD,$77
	.dc.b	$1A,$23,$A6,$BA,$30,$43,$07,$83,$D1,$EE,$47,$BB,$5D,$18,$21,$82
	.dc.b	$1F,$E0,$59,$47,$18,$F9,$73,$F9,$12,$3F,$91,$6B,$F9,$76,$BC,$0D
	.dc.b	$10,$D1,$4F,$07,$B1,$F2,$08,$FC,$86,$BF,$39,$AF,$03,$44,$34,$53
	.dc.b	$D9,$C9,$F5,$08,$FD,$46,$BF,$69,$AF,$03,$44,$34,$53,$FC,$59,$A3
	.dc.b	$C8,$FF,$12,$47,$EE,$1B,$19,$8F,$F1,$43,$5F,$F1,$8B,$5E,$71,$C0
	.dc.b	$8F,$06,$BC,$1A,$FF,$8F,$7F,$5E,$7C,$0F,$89,$B1,$FE,$37,$6B,$FE
	.dc.b	$40,$6B,$C6,$38,$11,$E0,$D7,$83,$5E,$07,$5E,$5F,$01,$1C,$10,$F0
	.dc.b	$39,$69,$93,$D4,$E5,$A7,$C8,$F1,$BE,$59,$1C,$10,$F9,$6E,$5A,$64
	.dc.b	$FA,$EE,$5A,$7C,$5E,$BF,$90,$47,$04,$3E,$43,$96,$99,$3E,$93,$96
	.dc.b	$9F,$E6,$7F,$63,$FA,$7B,$FE,$73,$6B,$FE,$7E,$6B,$FE,$81,$6A,$69
	.dc.b	$39,$BF,$1E,$D8,$ED,$18,$23,$F9,$0F,$0B,$F2,$64,$70,$43,$F2,$6E
	.dc.b	$5A,$64,$FC,$FB,$96,$9E,$DF,$1B,$C6,$47,$04,$3F,$D1,$F4,$8D,$46
	.dc.b	$1A,$FC,$56,$BF,$15,$F7,$F3,$48,$F9,$8D,$7A,$B5,$F9,$CD,$7E,$73
	.dc.b	$5F,$9C,$D5,$F9,$CB,$1F,$DD,$B6,$3F,$D4,$0D,$7F,$D5,$2D,$7F,$D5
	.dc.b	$CD,$7F,$D7,$2D,$7F,$D7,$8D,$7B,$C7,$62,$3D,$9A,$FC,$8F,$3F,$EF
	.dc.b	$91,$FC,$63,$5D,$18,$21,$82,$1F,$91,$78,$3F,$EC,$5F,$D7,$BF,$AB
	.dc.b	$37,$93,$F8,$DC,$62,$3F,$DA,$6D,$7F,$DA,$CD,$4D,$24,$72,$24,$7F
	.dc.b	$B2,$63,$2F,$CC,$F8,$7F,$9E,$47,$04,$3E,$7B,$96,$99,$3E,$8B,$96
	.dc.b	$9D,$FD,$BF,$59,$1C,$10,$FF,$71,$CA,$35,$18,$6B,$F1,$DA,$FC,$47
	.dc.b	$E3,$F3,$48,$F7,$35,$EA,$D7,$E8,$35,$F9,$AD,$7E,$6B,$5F,$CA,$FC
	.dc.b	$1F,$E5,$08,$FE,$51,$AF,$E5,$1A,$FE,$51,$AF,$E5,$1A,$FE,$51,$FE
	.dc.b	$B3,$FA,$F7,$FA,$9E,$04,$70,$DF,$0B,$64,$76,$41,$1C,$1B,$1B,$74
	.dc.b	$60,$A5,$33,$1C,$49,$CB,$E6,$24,$70,$8F,$EB,$6F,$89,$AF,$A6,$36
	.dc.b	$F4,$6B,$E7,$8D,$BE,$88,$CB,$EC,$FC,$ED,$F0,$5F,$DF,$9B,$1E,$86
	.dc.b	$BE,$C6,$BF,$0D,$AF,$38,$E4,$47,$93,$5E,$4D,$7C,$DE,$27,$D6,$23
	.dc.b	$F0,$5A,$FC,$86,$BC,$0D,$10,$D1,$4F,$85,$C1,$F8,$22,$3F,$35,$AF
	.dc.b	$D4,$6B,$C0,$D1,$0D,$14,$F8,$5A,$7C,$92,$3F,$25,$AF,$C9,$6B,$C0
	.dc.b	$D1,$0D,$14,$FC,$87,$E5,$BF,$EA,$C7,$F1,$AD,$73,$E2,$7F,$8D,$A3
	.dc.b	$11,$F9,$86,$BF,$A0,$6A,$69,$23,$91,$23,$F3,$31,$AE,$5E,$C4,$8C
	.dc.b	$B5,$F2,$B5,$F4,$B5,$F2,$35,$F2,$3F,$91,$F7,$24,$7E,$45,$AF,$E4
	.dc.b	$5A,$FE,$45,$AF,$E6,$DA,$FE,$FB,$34,$62,$3E,$07,$89,$F1,$C8,$FB
	.dc.b	$9A,$FC,$E6,$BF,$49,$AF,$CD,$6B,$F3,$5F,$A5,$CB,$23,$E4,$B5,$FB
	.dc.b	$2D,$7E,$EB,$5F,$B0,$D7,$EC,$35,$7E,$76,$F6,$3F,$CE,$AD,$8F,$E1
	.dc.b	$ED,$7F,$89,$B5,$FE,$34,$D7,$F8,$EB,$5F,$E4,$AD,$7B,$C7,$62,$3D
	.dc.b	$9A,$FF,$20,$E9,$FF,$57,$BF,$C8,$1A,$E7,$C3,$FF,$45,$C6,$23,$F9
	.dc.b	$63,$5F,$E5,$ED,$7F,$9B,$B5,$34,$91,$D0,$91,$E1,$DB,$B9,$1C,$10
	.dc.b	$F4,$39,$69,$93,$CA,$E5,$A7,$CE,$F5,$7E,$89,$1C,$10,$FA,$2E,$5A
	.dc.b	$64,$FA,$6E,$5A,$7C,$9F,$13,$E3,$91,$C1,$0F,$A0,$E5,$A6,$4F,$9A
	.dc.b	$E5,$A7,$F5,$4F,$D7,$BF,$AD,$13,$9F,$FD,$7E,$D8,$D4,$60,$8E,$08
	.dc.b	$7F,$60,$8E,$0F,$E4,$38,$BF,$2A,$47,$F2,$AD,$74,$60,$86,$08,$7E
	.dc.b	$61,$E0,$FF,$73,$87,$85,$E8,$24,$7B,$5A,$F9,$5A,$F1,$6B,$FD,$FA
	.dc.b	$39,$BC,$48,$7B,$79,$7E,$29,$1F,$94,$D7,$46,$08,$60,$87,$A5,$E0
	.dc.b	$FF,$83,$7C,$D7,$9D,$FE,$F0,$6C,$7F,$82,$9A,$FF,$83,$9A,$F1,$8E
	.dc.b	$04,$78,$35,$E0,$D7,$81,$FE,$16,$FD,$7B,$FD,$6A,$4F,$13,$FE,$00
	.dc.b	$B6,$35,$18,$23,$82,$1F,$E2,$28,$E0,$DC,$B9,$91,$F0,$35,$D1,$82
	.dc.b	$18,$21,$FE,$33,$8F,$EB,$67,$F5,$CB,$79,$DF,$24,$93,$F2,$5A,$FC
	.dc.b	$96,$BF,$45,$AF,$F9,$02,$39,$BC,$48,$7B,$78,$7E,$29,$1F,$94,$D7
	.dc.b	$46,$08,$60,$87,$AD,$E0,$FB,$79,$7E,$89,$1F,$AC,$D7,$E7,$35,$E0
	.dc.b	$68,$86,$8A,$7E,$3F,$E2,$BF,$1C,$47,$F1,$CD,$7F,$1C,$D7,$81,$A2
	.dc.b	$1A,$29,$FE,$67,$FD,$7B,$FD,$54,$4F,$13,$FC,$C7,$08,$C3,$FE,$66
	.dc.b	$6B,$FE,$6A,$6B,$CE,$38,$C6,$23,$FC,$EA,$FF,$45,$27,$2E,$89,$1D
	.dc.b	$5A,$F5,$6B,$E7,$6B,$D1,$F3,$FE,$7C,$E4,$A4,$9F,$F3,$EC,$65,$B1
	.dc.b	$97,$FD,$07,$19,$6C,$65,$FF,$46,$46,$5B,$19,$7D,$BE,$1F,$C0,$23
	.dc.b	$F1,$DA,$FC,$76,$BF,$45,$AF,$23,$81,$0F,$9D,$1F,$30,$8F,$D3,$6B
	.dc.b	$A3,$04,$30,$43,$F0,$0F,$07,$DB,$E3,$FC,$A2,$3F,$21,$AF,$C6,$6B
	.dc.b	$C0,$D1,$0D,$14,$FF,$5A,$7C,$87,$FA,$78,$9E,$1F,$E3,$9B,$1F,$EB
	.dc.b	$26,$BF,$EB,$66,$BC,$A3,$81,$1E,$1C,$FC,$C8,$E0,$87,$37,$2D,$32
	.dc.b	$79,$9C,$B4,$FC,$4F,$C2,$7E,$38,$8E,$08,$7E,$29,$CB,$4C,$9F,$8A
	.dc.b	$72,$D3,$E0,$78,$7F,$0C,$8E,$08,$7C,$37,$2D,$32,$7C,$E7,$2D,$3F
	.dc.b	$DB,$BF,$74,$E7,$FD,$EB,$63,$31,$82,$38,$21,$FE,$E8,$8E,$0D,$8D
	.dc.b	$B4,$D9,$FE,$F3,$FC,$F9,$F0,$5F,$C3,$1B,$1F,$EF,$06,$BF,$EF,$46
	.dc.b	$BC,$63,$81,$1E,$0D,$78,$35,$E0,$7C,$0F,$17,$DE,$47,$04,$3D,$8E
	.dc.b	$5A,$64,$10,$5A,$38,$46,$4C,$BF,$3B,$9A,$47,$D2,$6B,$F2,$DA,$F1
	.dc.b	$9F,$F3,$4B,$B2,$1B,$29,$B8,$F9,$84,$7E,$63,$5F,$98,$D7,$EA,$35
	.dc.b	$F6,$5F,$81,$D9,$C8,$87,$E0,$7C,$3F,$C4,$91,$FC,$6B,$5D,$18,$21
	.dc.b	$82,$1E,$F7,$83,$F2,$7E,$F1,$CD,$FD,$49,$B1,$F3,$4F,$A6,$E1,$C3
	.dc.b	$5E,$51,$C4,$8F,$11,$1B,$7E,$E3,$5E,$B1,$C1,$EA,$FE,$67,$DD,$91
	.dc.b	$F9,$96,$BF,$99,$6B,$FA,$16,$BF,$39,$AF,$E3,$2D,$18,$8F,$2F,$87
	.dc.b	$F0,$48,$FA,$6D,$18,$8F,$82,$D7,$E5,$35,$F6,$B5,$F9,$2F,$CD,$E5
	.dc.b	$91,$D5,$AF,$CC,$6B,$F5,$5A,$FC,$86,$BF,$51,$AB,$F3,$E7,$B9,$FE
	.dc.b	$0C,$6C,$7D,$36,$BF,$75,$AF,$EA,$DA,$FE,$C5,$AF,$ED,$1A,$F7,$8E
	.dc.b	$C4,$7B,$35,$FC,$57,$DF,$3C,$5F,$F0,$FB,$63,$F0,$CD,$77,$1A,$23
	.dc.b	$A6,$BA,$30,$43,$07,$83,$9B,$B9,$1E,$ED,$74,$60,$86,$08,$7F,$0F
	.dc.b	$94,$71,$8F,$91,$CF,$F3,$C8,$FC,$F6,$BF,$3D,$AF,$03,$44,$34,$53
	.dc.b	$C1,$E2,$7C,$82,$3F,$21,$AF,$CE,$6B,$C0,$D1,$0D,$14,$F6,$73,$7D
	.dc.b	$42,$3F,$51,$AF,$DA,$6B,$C0,$D1,$0D,$14,$FC,$65,$7E,$4C,$8F,$E4
	.dc.b	$DA,$FE,$4D,$AF,$03,$44,$34,$53,$FA,$17,$C6,$BC,$4F,$F3,$5B,$63
	.dc.b	$F8,$93,$5F,$E3,$8D,$78,$C7,$02,$3C,$1A,$F0,$6B,$C0,$EB,$CD,$E0
	.dc.b	$23,$82,$1E,$07,$2D,$32,$7A,$9C,$B4,$FE,$55,$E0,$BE,$43,$FD,$0E
	.dc.b	$D8,$CC,$60,$8E,$08,$7F,$3B,$8E,$0D,$8D,$B4,$D9,$F3,$23,$E5,$91
	.dc.b	$F9,$6D,$7E,$5B,$5E,$06,$88,$68,$A7,$D8,$E6,$FA,$E4,$7E,$BB,$5F
	.dc.b	$AE,$D7,$81,$A2,$1A,$29,$FE,$05,$86,$1F,$20,$91,$F3,$1A,$FC,$56
	.dc.b	$BC,$5A,$FF,$84,$E3,$9B,$C4,$87,$F6,$7F,$BF,$78,$3F,$EB,$C6,$C7
	.dc.b	$F6,$36,$BB,$8D,$11,$D3,$5D,$18,$21,$83,$C1,$C9,$DC,$8F,$76,$BA
	.dc.b	$30,$43,$04,$3F,$C5,$93,$FD,$81,$D1,$F2,$FE,$37,$F2,$24,$7F,$2A
	.dc.b	$D7,$F1,$CD,$78,$1A,$21,$A2,$9E,$0F,$07,$E4,$11,$F9,$0D,$7E,$73
	.dc.b	$5E,$06,$88,$68,$A7,$F9,$1F,$F0,$0F,$E9,$CD,$E4,$FF,$78,$C6,$23
	.dc.b	$FC,$A0,$D7,$FC,$A6,$D4,$D2,$47,$22,$47,$F8,$86,$32,$FF,$88,$FF
	.dc.b	$36,$6B,$FF,$00,$1B,$19,$8C,$11,$C1,$0F,$F1,$A4,$70,$6C,$6D,$A6
	.dc.b	$CF,$81,$E1,$7B,$88,$F9,$5A,$FF,$8F,$E9,$18,$8E,$26,$C8,$6D,$D3
	.dc.b	$F2,$39,$3E,$59,$1F,$84,$D7,$B3,$5F,$A8,$D7,$91,$C0,$87,$C0,$F8
	.dc.b	$9F,$90,$47,$E6,$35,$D1,$82,$18,$21,$FE,$89,$C1,$1C,$63,$FD,$19
	.dc.b	$F8,$17,$FA,$C1,$BE,$1F,$F9,$6D,$46,$5F,$F4,$B3,$5F,$F4,$CB,$5D
	.dc.b	$C7,$18,$C1,$1C,$12,$3F,$CD,$11,$AE,$5E,$04,$8C,$B5,$FF,$3B,$47
	.dc.b	$F5,$B7,$C7,$19,$7D,$6D,$72,$72,$3F,$29,$F1,$1F,$93,$23,$82,$1F
	.dc.b	$90,$72,$D3,$27,$CB,$72,$D3,$D7,$E1,$7C,$12,$38,$21,$EA,$72,$D3
	.dc.b	$27,$D4,$72,$D3,$E3,$F2,$7C,$D2,$38,$21,$F1,$9C,$B4,$C9,$F8,$07
	.dc.b	$2D,$3F,$D3,$5F,$9F,$7E,$05,$34,$8F,$E1,$5A,$F0,$30,$43,$04,$3F
	.dc.b	$D5,$CF,$5C,$BF,$D4,$91,$F9,$4F,$0F,$F2,$11,$82,$38,$24,$7F,$AB
	.dc.b	$E3,$2D,$8C,$BF,$EB,$B8,$E5,$CD,$D0,$91,$C5,$AE,$DA,$FF,$BB,$E3
	.dc.b	$FA,$CB,$63,$2F,$FB,$DE,$32,$D8,$F8,$9B,$48,$F9,$6D,$7E,$0B,$5F
	.dc.b	$F7,$E5,$A3,$10,$24,$65,$E6,$43,$99,$F0,$FC,$0F,$92,$47,$04,$3E
	.dc.b	$23,$96,$99,$3E,$03,$96,$9F,$3F,$93,$EA,$91,$C1,$0F,$A0,$E5,$A6
	.dc.b	$4F,$98,$E5,$A7,$E4,$3E,$63,$F1,$84,$70,$43,$F1,$2E,$5A,$64,$F9
	.dc.b	$6E,$5A,$7F,$BD,$7F,$04,$FF,$4D,$B7,$B7,$F8,$F4,$62,$3E,$81,$1C
	.dc.b	$38,$0E,$1D,$47,$22,$3D,$48,$71,$38,$BA,$71,$1F,$D4,$9F,$37,$27
	.dc.b	$90,$8E,$08,$70,$72,$D3,$27,$9E,$3A,$39,$3F,$63,$F3,$1F,$B9,$23
	.dc.b	$FC,$01,$AE,$8C,$10,$C1,$0F,$CC,$BC,$1F,$07,$7F,$CC,$23,$F4,$5A
	.dc.b	$FC,$06,$BC,$0D,$10,$D1,$4F,$A5,$F8,$2F,$F4,$F2,$73,$7F,$57,$6C
	.dc.b	$7C,$F0,$E1,$FA,$4D,$79,$C7,$12,$3C,$7F,$3E,$7C,$CF,$F6,$36,$C6
	.dc.b	$63,$04,$70,$43,$F0,$11,$C1,$B1,$B6,$9B,$3F,$1D,$0F,$37,$C0,$48
	.dc.b	$F2,$35,$EA,$D7,$8B,$5F,$CD,$C7,$37,$89,$0F,$2F,$81,$ED,$23,$F0
	.dc.b	$5A,$E8,$C1,$0C,$10,$F4,$BC,$1F,$5F,$C4,$FC,$92,$3F,$35,$AF,$C6
	.dc.b	$6B,$C0,$D1,$0D,$14,$F8,$FF,$31,$F5,$48,$FD,$96,$BF,$49,$AF,$03
	.dc.b	$44,$34,$53,$E9,$F9,$FF,$81,$23,$F8,$56,$BF,$79,$AF,$03,$44,$34
	.dc.b	$53,$EF,$FC,$C7,$E3,$48,$FE,$45,$AF,$E2,$9A,$F0,$34,$43,$45,$3F
	.dc.b	$8C,$7C,$D7,$F9,$A5,$BE,$17,$F8,$F6,$31,$1F,$C8,$1A,$FF,$51,$6A
	.dc.b	$69,$23,$91,$23,$F6,$11,$96,$FC,$4F,$72,$38,$21,$FC,$96,$11,$A8
	.dc.b	$C3,$5F,$5B,$5F,$43,$EB,$F9,$59,$1E,$26,$BD,$5A,$FC,$46,$BF,$05
	.dc.b	$AF,$C1,$6B,$F0,$7D,$CF,$90,$47,$DE,$D7,$D4,$D7,$E8,$B5,$F9,$CD
	.dc.b	$7E,$73,$F4,$FC,$B9,$1F,$29,$AF,$C5,$6B,$F7,$1A,$FD,$96,$BF,$65
	.dc.b	$AF,$D9,$F4,$7F,$00,$47,$EB,$B5,$FA,$4D,$7F,$12,$D7,$F0,$CD,$7F
	.dc.b	$0C,$FE,$2F,$E4,$E4,$7E,$09,$AF,$DD,$6B,$F9,$46,$BF,$91,$6B,$F9
	.dc.b	$16,$AF,$CF,$BF,$CA,$84,$7F,$D2,$CD,$7F,$A0,$35,$FE,$E0,$D7,$FB
	.dc.b	$D3,$5F,$EF,$EF,$4E,$6E,$64,$8E,$6D,$79,$B5,$E6,$D7,$9B,$5F,$F1
	.dc.b	$D4,$23,$CF,$F3,$F2,$7C,$ED,$7C,$ED,$7C,$ED,$7C,$ED,$7A,$B5,$F8
	.dc.b	$DE,$E7,$C4,$23,$F1,$1A,$FC,$46,$BF,$11,$AF,$C4,$6B,$EA,$7E,$7F
	.dc.b	$CF,$C8,$F9,$ED,$7E,$7B,$5F,$9E,$D7,$E7,$B5,$F8,$AD,$7E,$F7,$6F
	.dc.b	$DC,$23,$F7,$1A,$FD,$C6,$BF,$71,$AF,$DC,$6B,$F4,$9F,$C3,$FC,$9C
	.dc.b	$8F,$C3,$B5,$FC,$3B,$5F,$C3,$B5,$FC,$3B,$5F,$BA,$D7,$F2,$DD,$9F
	.dc.b	$94,$23,$F9,$46,$BF,$94,$6B,$F9,$46,$BF,$94,$6B,$F8,$A7,$FD,$E9
	.dc.b	$F9,$CB,$C1,$A4,$9F,$F1,$04,$65,$B1,$97,$FC,$A5,$19,$6C,$65,$FF
	.dc.b	$2F,$46,$5B,$19,$7F,$CC,$B0,$F8,$8F,$19,$23,$C2,$D7,$BB,$5E,$AD
	.dc.b	$79,$B5,$FF,$50,$47,$57,$D7,$F2,$3F,$00,$8E,$08,$7B,$1C,$B4,$C9
	.dc.b	$EC,$72,$D3,$E3,$F2,$FC,$C2,$38,$21,$F2,$1C,$B4,$C9,$F2,$1C,$B4
	.dc.b	$FA,$7E,$27,$D8,$23,$82,$1F,$51,$CB,$4C,$9F,$51,$CB,$4F,$BF,$DF
	.dc.b	$F8,$42,$38,$21,$F8,$07,$2D,$32,$7E,$01,$CB,$4F,$C5,$F7,$FE,$40
	.dc.b	$8E,$08,$7E,$31,$CB,$4C,$9F,$8C,$72,$D3,$F2,$FF,$3D,$F9,$C2,$38
	.dc.b	$21,$F9,$87,$2D,$32,$7E,$61,$CB,$4F,$D1,$FA,$7F,$A8,$23,$82,$1F
	.dc.b	$A4,$72,$D3,$27,$E9,$1C,$B4,$FD,$7F,$97,$FB,$42,$38,$21,$FB,$07
	.dc.b	$2D,$32,$7E,$C1,$CB,$4F,$DD,$FC,$EF,$F0,$02,$38,$21,$FB,$C7,$2D
	.dc.b	$32,$7E,$F1,$CB,$4F,$F5,$37,$E7,$CF,$88,$FF,$54,$42,$30,$EA,$33
	.dc.b	$18,$8F,$F6,$E3,$5F,$F7,$33,$5F,$F7,$5B,$51,$D6,$39,$91,$EC,$43
	.dc.b	$AF,$CE,$F8,$08,$FA,$1A,$E8,$C1,$0C,$10,$F2,$BC,$1F,$3F,$A3,$EC
	.dc.b	$23,$F1,$1A,$FB,$9A,$F0,$34,$43,$45,$3E,$1F,$CE,$FC,$82,$3F,$41
	.dc.b	$AF,$CA,$6B,$C0,$D1,$0D,$14,$F9,$FE,$9F,$D4,$23,$F7,$1A,$FD,$66
	.dc.b	$BC,$0D,$10,$D1,$4F,$B7,$F2,$9F,$80,$23,$F8,$86,$BF,$82,$6B,$C0
	.dc.b	$D1,$0D,$14,$FC,$3F,$97,$F8,$C2,$3F,$94,$6B,$F8,$E6,$BC,$0D,$10
	.dc.b	$D1,$4F,$C7,$9F,$90,$FA,$27,$C5,$FE,$79,$B1,$A3,$6E,$1C,$07,$0E
	.dc.b	$DA,$F0,$8E,$84,$78,$9C,$5C,$3D,$23,$8B,$E5,$8F,$24,$48,$E2,$F8
	.dc.b	$5A,$E5,$E4,$D7,$C8,$D7,$D6,$D7,$E1,$3F,$07,$DF,$91,$D9,$AF,$95
	.dc.b	$AF,$BD,$AF,$C6,$6B,$F3,$5A,$FC,$7F,$47,$E7,$11,$F5,$B5,$F8,$4D
	.dc.b	$7E,$63,$5F,$A6,$D7,$ED,$3F,$67,$D2,$91,$F1,$9A,$FC,$D6,$BF,$5D
	.dc.b	$AF,$DE,$6B,$F8,$56,$BF,$7F,$E4,$BF,$0C,$47,$E9,$B5,$FB,$4D,$7F
	.dc.b	$08,$D7,$F1,$6D,$7F,$24,$FE,$47,$DD,$91,$F7,$9A,$FE,$15,$AF,$E3
	.dc.b	$DA,$FE,$59,$AF,$E7,$5A,$FE,$5F,$E8,$7F,$3C,$47,$F1,$6D,$7F,$24
	.dc.b	$D7,$F3,$8D,$7F,$46,$D7,$F5,$4F,$F4,$6F,$C3,$9C,$B4,$93,$FA,$08
	.dc.b	$CB,$63,$2F,$EB,$23,$2D,$8C,$BF,$B3,$8C,$B6,$32,$FE,$F3,$F3,$EF
	.dc.b	$F2,$B3,$7C,$0F,$EB,$31,$88,$FE,$9E,$D4,$D2,$47,$02,$47,$EE,$23
	.dc.b	$2D,$8E,$BE,$04,$8D,$35,$FE,$0F,$3F,$E5,$96,$C6,$5F,$E1,$D1,$96
	.dc.b	$C6,$5F,$63,$5F,$4F,$7F,$B0,$8F,$95,$AF,$81,$AF,$66,$BF,$1D,$AF
	.dc.b	$43,$DD,$DF,$F2,$08,$E0,$87,$B1,$CB,$4C,$9E,$C7,$2D,$3E,$57,$77
	.dc.b	$D4,$23,$82,$1F,$21,$CB,$4C,$9F,$21,$CB,$4F,$AD,$E9,$7E,$00,$8E
	.dc.b	$08,$7D,$47,$2D,$32,$7D,$47,$2D,$3F,$05,$E8,$FE,$30,$8E,$08,$7E
	.dc.b	$01,$CB,$4C,$9F,$80,$72,$D3,$FA,$2F,$E7,$CF,$3F,$FC,$62,$D8,$CC
	.dc.b	$60,$8E,$08,$7F,$56,$8E,$0D,$8D,$B4,$D9,$FD,$7E,$1E,$7F,$80,$91
	.dc.b	$E4,$6B,$D5,$AF,$16,$BF,$DB,$A3,$9B,$C4,$87,$F6,$AF,$CF,$5E,$6F
	.dc.b	$F9,$39,$B1,$FD,$99,$AE,$E3,$44,$74,$D7,$46,$08,$60,$FE,$F3,$F8
	.dc.b	$9B,$AF,$FC,$BA,$D8,$FE,$EA,$D7,$71,$A2,$3A,$6B,$A3,$04,$30,$7B
	.dc.b	$F9,$BD,$C4,$7D,$CD,$74,$60,$86,$08,$7F,$85,$6D,$1C,$63,$1E,$57
	.dc.b	$CB,$23,$F2,$DA,$FD,$16,$BC,$0D,$10,$D1,$4F,$07,$3F,$D7,$23,$F5
	.dc.b	$DA,$FD,$D6,$BC,$0D,$10,$D1,$4F,$67,$3F,$E0,$C8,$FE,$0D,$AF,$E2
	.dc.b	$5A,$F0,$34,$43,$45,$3E,$47,$95,$F8,$F2,$3F,$8F,$6B,$F9,$56,$BC
	.dc.b	$0D,$10,$D1,$4F,$F2,$47,$84,$F8,$6F,$F5,$4B,$63,$FC,$82,$D7,$FC
	.dc.b	$96,$D7,$8C,$70,$23,$C1,$AF,$06,$BC,$0E,$BE,$3F,$80,$8E,$08,$78
	.dc.b	$1C,$B4,$C9,$EA,$72,$D3,$FC,$B3,$F9,$07,$E9,$5F,$F3,$33,$5F,$F3
	.dc.b	$6B,$5F,$F5,$EB,$53,$49,$AF,$FB,$39,$B1,$DA,$30,$47,$FC,$D3,$F8
	.dc.b	$FB,$EA,$7F,$DA,$AD,$8C,$C6,$08,$E0,$87,$F9,$EA,$38,$36,$36,$D3
	.dc.b	$67,$FA,$0E,$1D,$FE,$02,$47,$91,$AF,$56,$BC,$5A,$FF,$A4,$A3,$9B
	.dc.b	$C4,$87,$97,$CF,$F6,$91,$F8,$2D,$74,$60,$86,$08,$7A,$5E,$0F,$AF
	.dc.b	$E4,$FE,$49,$1F,$9A,$D7,$E3,$35,$E0,$68,$86,$8A,$7C,$7E,$1F,$AA
	.dc.b	$47,$EC,$B5,$FA,$4D,$78,$1A,$21,$A2,$9F,$4F,$87,$F0,$24,$7F,$0A
	.dc.b	$D7,$EF,$35,$E0,$68,$86,$8A,$7D,$FE,$1F,$C6,$91,$FC,$8B,$5F,$C5
	.dc.b	$35,$E0,$68,$86,$8A,$7E,$2F,$87,$F3,$24,$7F,$3A,$D7,$F2,$CD,$78
	.dc.b	$1A,$21,$A2,$9F,$97,$E1,$FD,$29,$1F,$D4,$B5,$FD,$13,$5E,$06,$88
	.dc.b	$68,$A7,$FB,$97,$F1,$FF,$E3,$0F,$FB,$B1,$A8,$38,$6F,$0F,$FB,$89
	.dc.b	$A7,$FB,$92,$34,$48,$D3,$63,$51,$83,$3E,$04,$8E,$4D,$7F,$DE,$F3
	.dc.b	$FE,$7F,$24,$66,$3D,$4D,$7D,$4D,$7D,$51,$FE,$75,$7D,$DD,$FE,$42
	.dc.b	$3E,$96,$BE,$06,$BF,$1D,$AF,$23,$81,$F0,$7B,$FD,$E4,$70,$43,$D6
	.dc.b	$E5,$A6,$4F,$A8,$E5,$A7,$CD,$EF,$F9,$64,$70,$43,$E3,$B9,$69,$93
	.dc.b	$F0,$0E,$5A,$7D,$9F,$43,$EB,$91,$C1,$0F,$A6,$E5,$A6,$4F,$C6,$39
	.dc.b	$69,$F8,$5F,$03,$F0,$64,$70,$43,$EF,$B9,$69,$93,$F3,$0E,$5A,$7E
	.dc.b	$47,$E2,$FF,$1E,$47,$04,$3F,$16,$E5,$A6,$4F,$D2,$39,$69,$F9,$DF
	.dc.b	$47,$F3,$64,$70,$43,$F2,$EE,$5A,$64,$FD,$83,$96,$9F,$A9,$F7,$BF
	.dc.b	$4E,$47,$04,$3F,$46,$E5,$A6,$4F,$DE,$39,$69,$FB,$B2,$0F,$C5,$DC
	.dc.b	$3F,$BB,$6C,$68,$FE,$04,$E1,$C3,$5E,$6D,$7B,$35,$E3,$1D,$48,$F5
	.dc.b	$3C,$4E,$1E,$AF,$F0,$DC,$5F,$0F,$49,$3C,$63,$2D,$8C,$BE,$98,$CB
	.dc.b	$63,$2F,$C2,$8C,$B6,$32,$FC,$7F,$C7,$DF,$41,$FC,$7D,$B1,$E9,$6B
	.dc.b	$F1,$DA,$FC,$E6,$BC,$E3,$91,$1E,$4D,$79,$35,$FA,$FF,$8F,$FF,$19
	.dc.b	$7F,$9A,$35,$34,$9C,$1F,$CD,$5B,$1F,$3D,$AF,$DD,$6B,$DA,$38,$70
	.dc.b	$48,$E0,$D7,$83,$5F,$2B,$5F,$C7,$4A,$31,$1E,$86,$BE,$8E,$0F,$31
	.dc.b	$1F,$33,$5F,$33,$5F,$86,$D7,$BB,$5F,$8A,$FC,$1C,$A4,$7C,$16,$BF
	.dc.b	$05,$AF,$CE,$6B,$EA,$6B,$F4,$1A,$FD,$0E,$0F,$9C,$47,$E7,$35,$F9
	.dc.b	$CD,$7E,$DB,$5F,$8E,$D7,$EE,$BF,$67,$82,$47,$D9,$6B,$F6,$5A,$FE
	.dc.b	$19,$AF,$D2,$6B,$F8,$86,$BF,$88,$E0,$FC,$31,$1F,$C3,$35,$FC,$33
	.dc.b	$5F,$C9,$B5,$FB,$ED,$7F,$2A,$FE,$47,$82,$47,$E4,$5A,$FE,$45,$AF
	.dc.b	$E7,$9A,$FE,$29,$AF,$E8,$1A,$FE,$83,$E0,$FF,$3C,$47,$F3,$CD,$7F
	.dc.b	$3C,$D7,$F5,$6D,$7F,$2E,$D7,$F5,$AF,$EA,$7C,$09,$1F,$A9,$6B,$FA
	.dc.b	$96,$BF,$B6,$6B,$FA,$26,$BF,$B8,$6B,$FB,$8F,$AD,$FD,$B1,$1F,$DB
	.dc.b	$35,$FD,$B3,$5F,$E0,$6D,$7F,$5E,$D7,$F8,$2B,$FE,$35,$FC,$6B,$FD
	.dc.b	$70,$4F,$07,$04,$9F,$E6,$51,$96,$C6,$5F,$E8,$51,$96,$C7,$F9,$07
	.dc.b	$F1,$F7,$B1,$49,$3F,$CF,$63,$2D,$8C,$BF,$D3,$63,$2D,$8C,$BF,$D5
	.dc.b	$A3,$2D,$8C,$BF,$D6,$E1,$F0,$9E,$32,$47,$85,$AF,$76,$BD,$5A,$F3
	.dc.b	$6B,$FE,$61,$8E,$AF,$AF,$9B,$E0,$11,$C1,$0F,$63,$96,$99,$3D,$8E
	.dc.b	$5A,$7C,$7F,$1B,$E6,$11,$C1,$0F,$90,$E5,$A6,$4F,$90,$E5,$A7,$D3
	.dc.b	$F8,$4F,$B0,$47,$04,$3E,$A3,$96,$99,$3E,$A3,$96,$9F,$8A,$F6,$9F
	.dc.b	$90,$D2,$47,$E2,$A3,$2D,$8C,$BF,$E1,$4D,$13,$A8,$C3,$5F,$F0,$D4
	.dc.b	$6A,$78,$C6,$27,$4F,$FA,$66,$74,$FE,$6B,$CF,$7D,$D7,$E8,$1B,$19
	.dc.b	$8C,$11,$C1,$0F,$34,$70,$6C,$6D,$A6,$C7,$DE,$40,$00

;--------------------------------------------------------------------------------
;	FMOVE.W FPn,Dr
;	FMOVE.W FPn,<mem>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovewregto_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.W FPN,<EA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.W FPn,<ea>',13,10
;------------------------------------------------
;	d1	actual result
;	d3	fill=$00000000,$FFFFFFFF
;	d4	actual status
;	d6	fpcr=(XRN..DRP)<<4
;	d7	0=failed,1=successful
;	a0	expected result,expected status,...
;	a1	expected result
;	a4	expected status
;	a5	source handle,...
;	fp5	source
;------------------------------------------------
	lea.l	push_decompressed,a3
;decompress data
	move.l	a3,-(sp)
	pea.l	fmovewregto_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a3,a5			;source handle,...
@@:
	add.l	d0,(a3)+		;source handle
	tst.l	(a3)
	bpl	@b
	addq.l	#4,a3			;-1
	movea.l	a3,a0			;expected result,expected status,...
;
55:
	move.l	#0,d6		;fpcr=(XRN..DRP)<<4
66:
;FPn,Dr=fill
	moveq.l	#$00000000,d3		;fill=$00000000,$FFFFFFFF
33:
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	d3,d1			;fill
					;source
	fmove.w	fp5,d1			;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.W FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',Dr='
	puthex8	d3			;fill
	putmsg	' @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result,expected status,...
;
	not.l	d3			;fill
	bne	33b
;FPn,<mem>=fill
	moveq.l	#$00000000,d3		;fill=$00000000,$FFFFFFFF
33:
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	d3,-(sp)		;fill
					;source
	fmove.w	fp5,(sp)		;EXECUTE
	move.l	(sp)+,d1		;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.W FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',<mem>='
	puthex8	d3			;fill
	putmsg	' @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result,expected status,...
;
	not.l	d3			;fill
	bne	33b
;
	add.w	#1<<4,d6		;rprm++
	cmp.w	#176,d6		;fpcr=(XRN..DRP)<<4
	bls	66b
;
	addq.l	#4,a5			;source handle
	tst.l	(a5)			;source handle
	bpl	55b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmovewregto_data_compressed::
	.dc.b	$00,$09,$68,$58,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$E6,$3C,$FF,$7B,$E0,$4E,$8C,$18,$8F,$C1,$9F,$B0
	.dc.b	$71,$3C,$E3,$07,$43,$F9,$4C,$F4,$8F,$14,$F4,$39,$CF,$37,$FA,$1C
	.dc.b	$73,$9F,$04,$C3,$C1,$1C,$9E,$B3,$97,$FA,$5C,$FA,$A3,$9C,$EA,$7C
	.dc.b	$86,$A3,$DC,$F2,$7C,$73,$C6,$3E,$39,$B3,$89,$E2,$9F,$24,$6E,$3D
	.dc.b	$87,$03,$E1,$9A,$9E,$EF,$85,$EE,$7C,$A9,$F3,$46,$DF,$73,$B3,$E8
	.dc.b	$CF,$96,$3E,$34,$FF,$56,$3E,$BB,$C2,$7E,$61,$C2,$3B,$3E,$B3,$B4
	.dc.b	$79,$63,$C6,$7C,$88,$D3,$D2,$35,$1C,$E3,$4F,$B9,$D1,$F0,$9F,$87
	.dc.b	$1E,$67,$E4,$CF,$E0,$E3,$9C,$69,$F8,$33,$CA,$3E,$A1,$DC,$F9,$93
	.dc.b	$F8,$48,$D4,$7C,$53,$91,$E8,$3F,$0B,$3F,$0E,$7E,$B4,$74,$8F,$23
	.dc.b	$F0,$A7,$D0,$E6,$3C,$6F,$27,$CE,$E4,$EC,$7E,$2D,$ED,$1F,$93,$8F
	.dc.b	$A3,$1D,$A3,$CF,$1F,$50,$F3,$9F,$24,$EC,$F8,$9F,$C0,$47,$17,$27
	.dc.b	$C4,$38,$CF,$C3,$72,$7C,$33,$CC,$75,$3C,$D1,$EA,$7A,$BE,$E7,$33
	.dc.b	$E1,$7D,$31,$97,$F4,$11,$99,$FC,$94,$E5,$FB,$91,$93,$F1,$F3,$E1
	.dc.b	$38,$9E,$93,$CB,$1E,$97,$C2,$64,$F2,$9F,$38,$E2,$7B,$4C,$9D,$5F
	.dc.b	$0B,$C4,$F4,$9C,$5C,$BC,$4C,$9C,$63,$F6,$AF,$D5,$3C,$AF,$EB,$5F
	.dc.b	$9A,$64,$F0,$C7,$A5,$F2,$99,$3A,$B9,$3C,$2E,$5F,$92,$F9,$63,$C3
	.dc.b	$1F,$96,$8E,$B1,$E1,$7E,$29,$E9,$8E,$2F,$A6,$7F,$23,$3F,$BA,$8F
	.dc.b	$D3,$BC,$5F,$2C,$F5,$7C,$B1,$98,$F2,$99,$3A,$9E,$97,$CA,$7E,$75
	.dc.b	$CB,$C4,$C9,$E1,$9F,$CC,$4F,$A0,$F8,$11,$F8,$49,$E9,$3E,$89,$FD
	.dc.b	$F3,$C1,$F3,$CF,$73,$81,$DE,$38,$47,$E8,$A3,$11,$F9,$D8,$F6,$9F
	.dc.b	$C4,$23,$CA,$F6,$8F,$CC,$46,$23,$91,$B8,$EA,$FE,$08,$C3,$D6,$7E
	.dc.b	$0C,$62,$7F,$40,$7F,$24,$8F,$6C,$EE,$3D,$33,$B3,$F2,$D3,$C9,$FE
	.dc.b	$24,$62,$36,$E0,$FC,$03,$83,$F8,$5C,$7D,$77,$11,$F9,$68,$C4,$78
	.dc.b	$CF,$8A,$79,$A3,$F0,$A7,$8C,$EA,$ED,$C1,$F6,$1C,$3E,$33,$07,$23
	.dc.b	$11,$E0,$71,$1E,$63,$ED,$1C,$8C,$1F,$87,$9F,$A6,$7D,$08,$FD,$09
	.dc.b	$E2,$9E,$91,$F8,$57,$F2,$F3,$D2,$7B,$9F,$05,$C4,$ED,$FB,$D1,$F6
	.dc.b	$CF,$3B,$A8,$F0,$9F,$5D,$F3,$47,$37,$CC,$FE,$DE,$3C,$CF,$E3,$8E
	.dc.b	$71,$A7,$F6,$91,$CD,$F2,$9F,$A3,$9F,$5C,$78,$27,$EA,$3E,$73,$AC
	.dc.b	$7E,$16,$3E,$3C,$FE,$64,$F1,$4F,$53,$81,$4E,$8F,$E2,$E3,$D6,$F8
	.dc.b	$5F,$80,$F2,$7C,$31,$C2,$3F,$06,$7A,$0F,$D9,$C7,$B6,$7F,$3A,$F0
	.dc.b	$8E,$47,$73,$84,$7D,$68,$E0,$7E,$4A,$35,$3E,$88,$E5,$1F,$21,$B3
	.dc.b	$DA,$37,$3E,$57,$F2,$2F,$17,$27,$88,$F4,$B9,$6B,$C6,$3E,$8C,$6C
	.dc.b	$E7,$19,$9F,$8B,$1E,$33,$A9,$4E,$67,$19,$E6,$E4,$E6,$66,$39,$C7
	.dc.b	$AE,$38,$9B,$78,$99,$78,$99,$8E,$27,$C9,$78,$C6,$CE,$31,$98,$E3
	.dc.b	$19,$8E,$31,$FA,$B6,$BF,$98,$3F,$18,$77,$7E,$BB,$B6,$BB,$3F,$93
	.dc.b	$94,$FD,$63,$F9,$67,$2D,$72,$7E,$AC,$EF,$1E,$58,$EA,$6E,$3F,$80
	.dc.b	$9D,$4A,$E5,$FD,$83,$63,$F4,$D1,$D4,$F0,$94,$F0,$9F,$A1,$8E,$EE
	.dc.b	$4E,$A5,$3A,$96,$3A,$BC,$4F,$4C,$7C,$F8,$FA,$31,$F1,$CC,$94,$F1
	.dc.b	$9C,$4A,$7E,$B0,$D9,$EF,$3C,$26,$CF,$49,$B3,$8B,$FA,$D7,$8C,$7C
	.dc.b	$33,$FA,$1B,$C4,$F6,$94,$EA,$73,$3A,$C7,$37,$AC,$7F,$28,$78,$9B
	.dc.b	$2C,$6D,$E3,$1B,$7A,$C6,$5D,$C6,$63,$E5,$C7,$33,$71,$93,$D2,$E6
	.dc.b	$36,$73,$77,$3F,$C9,$A7,$73,$93,$6F,$33,$F9,$0C,$F3,$76,$65,$FE
	.dc.b	$2E,$D9,$FE,$11,$1F,$AA,$8F,$BC,$7E,$A0,$F9,$C7,$49,$F0,$3D,$27
	.dc.b	$C8,$E4,$E8,$FF,$0D,$7F,$23,$3F,$4E,$7C,$F3,$C2,$3C,$6D,$3C,$71
	.dc.b	$F8,$E8,$FC,$33,$C2,$3F,$A0,$47,$F0,$F9,$EA,$FB,$DF,$14,$73,$3C
	.dc.b	$51,$F8,$83,$F3,$D1,$E9,$8E,$71,$FC,$0C,$D1,$E7,$8D,$1F,$CD,$8F
	.dc.b	$8A,$78,$27,$E1,$BE,$09,$E7,$1F,$BF,$9E,$67,$94,$F7,$C6,$A7,$9C
	.dc.b	$6A,$79,$BA,$8F,$03,$FA,$D3,$F2,$33,$ED,$7E,$11,$C6,$3E,$8B,$83
	.dc.b	$F3,$B1,$B7,$DC,$61,$F1,$3D,$4F,$CC,$BB,$9F,$54,$6E,$7C,$6F,$D4
	.dc.b	$7E,$2B,$FD,$22,$36,$F9,$A7,$AB,$B9,$F8,$C7,$23,$F1,$AF,$BC,$E4
	.dc.b	$FC,$47,$C0,$60,$F3,$1F,$A4,$8E,$B1,$C9,$FC,$4B,$F9,$98,$FB,$EF
	.dc.b	$81,$E4,$78,$CF,$01,$FD,$39,$F0,$3E,$97,$CC,$F2,$36,$7E,$9C,$F1
	.dc.b	$CF,$E5,$63,$71,$F8,$97,$C0,$F5,$36,$FC,$87,$07,$8E,$7D,$31,$F8
	.dc.b	$D3,$ED,$C7,$D8,$3C,$87,$A9,$F8,$11,$F9,$E8,$F8,$E7,$48,$CB,$F8
	.dc.b	$33,$89,$93,$D9,$1E,$07,$F1,$4F,$E0,$63,$E1,$C7,$E1,$23,$E7,$BC
	.dc.b	$1F,$03,$D0,$F8,$87,$AD,$CB,$C6,$3E,$2B,$FA,$38,$F0,$3E,$83,$8C
	.dc.b	$7A,$1F,$80,$71,$3E,$01,$E1,$3D,$86,$4F,$61,$F4,$4F,$AF,$3F,$C2
	.dc.b	$E3,$A3,$F9,$68,$E0,$78,$0E,$87,$80,$FC,$74,$7A,$23,$F6,$B3,$E8
	.dc.b	$9F,$81,$1D,$5E,$2F,$57,$89,$E8,$3C,$91,$C5,$F0,$C7,$17,$31,$F2
	.dc.b	$1E,$27,$B0,$F9,$0E,$63,$F2,$11,$D4,$C9,$F1,$0E,$27,$B4,$EA,$65
	.dc.b	$E2,$F5,$7D,$B1,$D6,$7F,$2D,$3C,$4E,$A7,$C5,$9E,$2E,$5E,$2F,$58
	.dc.b	$E3,$1E,$D8,$E2,$FC,$53,$89,$93,$8B,$D6,$38,$9F,$BB,$7D,$26,$5F
	.dc.b	$D6,$BE,$17,$F4,$F1,$D5,$E2,$7E,$AE,$7F,$39,$3F,$B2,$8E,$33,$E1
	.dc.b	$78,$9F,$A0,$8F,$93,$19,$9E,$31,$F9,$D3,$DD,$3C,$8F,$33,$EE,$79
	.dc.b	$4F,$B9,$CB,$EE,$3C,$CF,$13,$CD,$3F,$AE,$9F,$7C,$E9,$FC,$DC,$F2
	.dc.b	$3C,$EF,$AC,$CB,$C8,$FC,$C1,$A7,$D0,$7C,$93,$67,$F0,$09,$F0,$3E
	.dc.b	$93,$A3,$F9,$77,$E1,$C7,$86,$3E,$1B,$FA,$C3,$DE,$7D,$08,$E8,$F8
	.dc.b	$A3,$71,$FA,$68,$D9,$F0,$5D,$9F,$9B,$9F,$8D,$1A,$8F,$39,$FC,$1E
	.dc.b	$3E,$A4,$FA,$1D,$C7,$A2,$3F,$35,$3D,$63,$83,$E9,$8E,$0F,$C5,$3D
	.dc.b	$51,$C4,$FB,$A7,$80,$FD,$FB,$C4,$F6,$47,$23,$DB,$1F,$62,$3D,$11
	.dc.b	$F2,$67,$F0,$51,$F5,$67,$DC,$F8,$5F,$88,$78,$4F,$AD,$3C,$63,$E6
	.dc.b	$47,$A9,$E4,$FE,$E5,$E4,$F8,$4E,$53,$F8,$99,$E5,$3F,$0E,$7D,$6F
	.dc.b	$27,$F5,$51,$EA,$7F,$04,$7E,$22,$79,$47,$CC,$8F,$88,$7F,$1A,$9F
	.dc.b	$88,$FB,$23,$94,$FD,$48,$E5,$3F,$12,$79,$1E,$48,$E6,$7E,$3A,$3C
	.dc.b	$93,$F7,$9F,$9E,$48,$ED,$3F,$44,$91,$98,$F8,$91,$E2,$8F,$8B,$1E
	.dc.b	$C7,$EF,$BE,$43,$E3,$BE,$57,$F0,$31,$E5,$21,$E8,$23,$A6,$BF,$61
	.dc.b	$B3,$F2,$23,$44,$34,$57,$4E,$08,$E0,$93,$E5,$8C,$B6,$33,$1E,$A2
	.dc.b	$1E,$D2,$47,$C0,$23,$F9,$03,$25,$72,$FE,$2C,$C9,$5C,$BF,$2C,$E0
	.dc.b	$43,$45,$34,$60,$86,$09,$1F,$87,$8E,$AD,$8E,$4D,$78,$B5,$E3,$1F
	.dc.b	$7A,$30,$47,$04,$9F,$C2,$10,$FC,$4C,$FE,$32,$3F,$14,$FE,$92,$32
	.dc.b	$FE,$31,$AE,$63,$73,$F9,$78,$E4,$43,$84,$68,$FD,$3B,$83,$47,$32
	.dc.b	$C7,$37,$44,$34,$57,$4E,$08,$E0,$91,$E0,$8C,$B6,$32,$FE,$C0,$C1
	.dc.b	$0C,$12,$3F,$9B,$9E,$17,$4E,$0E,$AE,$DF,$E3,$06,$63,$BB,$63,$8B
	.dc.b	$A2,$1F,$C5,$8D,$3C,$DC,$1A,$3F,$89,$39,$69,$C0,$86,$CF,$E1,$C6
	.dc.b	$0E,$64,$38,$9B,$8F,$E7,$2E,$CC,$9B,$39,$9B,$8F,$E6,$EE,$63,$8B
	.dc.b	$98,$FE,$64,$F2,$21,$C8,$E0,$F2,$73,$1C,$9C,$C7,$27,$81,$A7,$83
	.dc.b	$82,$47,$17,$51,$C6,$31,$1A,$8E,$31,$82,$38,$21,$C4,$B1,$FC,$EA
	.dc.b	$37,$19,$24,$65,$DC,$FF,$2C,$8F,$E4,$51,$FC,$6C,$8F,$F1,$A7,$F8
	.dc.b	$19,$FB,$98,$FD,$8C,$FE,$2A,$7E,$61,$D6,$7B,$CF,$58,$FE,$4E,$73
	.dc.b	$3F,$85,$C7,$33,$F6,$87,$E2,$5F,$84,$F2,$8F,$E7,$6F,$92,$3F,$91
	.dc.b	$9F,$C4,$23,$F8,$54,$72,$8F,$C3,$1C,$8F,$21,$23,$C9,$3F,$CE,$23
	.dc.b	$9B,$FC,$5E,$39,$C7,$F0,$29,$FD,$2B,$F8,$03,$94,$79,$1F,$61,$FC
	.dc.b	$92,$3D,$8F,$17,$D9,$3F,$52,$38,$9F,$40,$E2,$7A,$E7,$E6,$CF,$82
	.dc.b	$3F,$0E,$70,$3E,$59,$E8,$7B,$BF,$C0,$DE,$E7,$DE,$8F,$34,$6D,$F6
	.dc.b	$BB,$3E,$3C,$F9,$63,$E2,$CF,$CF,$3F,$02,$F0,$9F,$E1,$A7,$08,$EC
	.dc.b	$FA,$CE,$D1,$EA,$8F,$19,$F6,$E3,$4F,$48,$D4,$73,$8D,$3E,$E7,$47
	.dc.b	$DD,$7F,$1F,$1E,$67,$E4,$CF,$D9,$8E,$71,$A7,$E0,$CF,$28,$FA,$87
	.dc.b	$73,$E6,$4F,$E0,$E3,$51,$F1,$4E,$47,$A0,$FC,$CC,$FC,$39,$FD,$14
	.dc.b	$7E,$16,$3C,$8F,$C2,$9F,$43,$98,$F1,$BC,$9F,$3B,$93,$B1,$F8,$B7
	.dc.b	$B4,$7E,$0E,$3F,$39,$1D,$A3,$CF,$1F,$50,$F3,$9F,$24,$EC,$F8,$9F
	.dc.b	$C5,$C7,$17,$27,$C4,$38,$CF,$C3,$72,$7C,$33,$CC,$75,$3C,$D1,$EA
	.dc.b	$7A,$BE,$E7,$33,$E1,$7D,$31,$97,$F4,$11,$99,$FC,$94,$E5,$FB,$91
	.dc.b	$93,$F1,$F3,$E1,$38,$9E,$93,$CB,$1E,$97,$C2,$64,$F2,$9F,$38,$E2
	.dc.b	$7B,$4C,$9D,$5F,$0B,$C4,$F4,$9C,$5C,$BC,$4C,$9C,$63,$F6,$AF,$D5
	.dc.b	$3C,$AF,$EB,$5F,$9A,$64,$F0,$C7,$A5,$F2,$99,$3A,$B9,$3C,$2E,$5F
	.dc.b	$92,$F9,$63,$C3,$1F,$96,$8E,$B1,$E1,$7E,$29,$E9,$8E,$2F,$A6,$7F
	.dc.b	$23,$3F,$BA,$8F,$D3,$BC,$5F,$2C,$F5,$7C,$B1,$98,$F2,$99,$3A,$9E
	.dc.b	$97,$CA,$7E,$75,$CB,$C4,$C9,$E1,$9F,$CC,$4F,$A0,$F8,$11,$F8,$49
	.dc.b	$E9,$3E,$89,$FD,$F3,$C1,$F3,$CF,$73,$81,$DE,$38,$47,$E8,$A3,$11
	.dc.b	$F9,$D8,$F6,$9F,$C4,$23,$CA,$F6,$8F,$CC,$46,$23,$91,$B8,$EA,$FE
	.dc.b	$08,$C3,$D6,$7E,$0C,$62,$7F,$40,$7F,$24,$8F,$6C,$EE,$3D,$33,$B3
	.dc.b	$F2,$D3,$C9,$FE,$24,$62,$36,$E0,$FC,$03,$83,$F8,$5C,$7D,$77,$11
	.dc.b	$F9,$68,$C4,$78,$CF,$8A,$79,$A3,$F0,$A7,$8C,$EA,$ED,$C1,$F6,$1C
	.dc.b	$3E,$33,$07,$23,$11,$E0,$71,$1E,$63,$ED,$1C,$8C,$1F,$87,$9F,$A6
	.dc.b	$7D,$08,$09,$4C,$84,$C7,$73,$96,$0F,$A1,$7C,$07,$0A,$CC,$62,$FB
	.dc.b	$D4,$38,$57,$7A,$ED,$18,$3C,$B3,$BA,$EB,$3E,$48,$C5,$78,$AB,$D7
	.dc.b	$5D,$DF,$11,$88,$F0,$CF,$0A,$E5,$3E,$4A,$E0,$F4,$9F,$8B,$3E,$49
	.dc.b	$F6,$CF,$90,$E1,$5E,$FA,$E1,$1E,$D3,$DA,$F2,$AD,$C7,$13,$C8,$F8
	.dc.b	$9F,$34,$71,$78,$9E,$A3,$E2,$1F,$32,$7C,$0F,$8A,$38,$57,$C1,$AF
	.dc.b	$45,$7C,$0A,$F3,$C7,$77,$A9,$E8,$76,$F9,$4F,$AA,$EE,$7E,$2C,$77
	.dc.b	$3C,$8F,$C4,$7C,$91,$B9,$FA,$91,$EA,$8E,$AE,$9F,$55,$7A,$27,$F0
	.dc.b	$51,$F1,$A7,$53,$F3,$1F,$34,$66,$3E,$D3,$CA,$3E,$85,$7C,$18,$D4
	.dc.b	$7E,$DE,$7F,$4A,$FE,$A0,$C3,$FA,$57,$FA,$34,$7E,$95,$FE,$3A,$7F
	.dc.b	$9B,$A7,$F2,$31,$DE,$3C,$67,$E3,$4F,$91,$1A,$9F,$44,$7C,$79,$F8
	.dc.b	$31,$E8,$33,$1F,$80,$3E,$5D,$7D,$77,$F0,$B3,$DD,$E4,$FB,$1C,$CF
	.dc.b	$27,$E7,$46,$5F,$B2,$F1,$7B,$3E,$27,$F0,$6E,$5E,$27,$AC,$FA,$27
	.dc.b	$DF,$3D,$E7,$E4,$A3,$D4,$E4,$F2,$9C,$5C,$BE,$D3,$8D,$7E,$1E,$33
	.dc.b	$1F,$92,$38,$9D,$4F,$2B,$ED,$8F,$B9,$1D,$4F,$B2,$7F,$4F,$3E,$CC
	.dc.b	$FF,$43,$9F,$E4,$51,$A3,$FA,$89,$FC,$A9,$FC,$1C,$7F,$63,$BF,$EE
	.dc.b	$B1,$88,$FE,$4E,$73,$9F,$C4,$1F,$DB,$E7,$F9,$94,$71,$9F,$EB,$D3
	.dc.b	$FD,$6C,$EC,$FE,$35,$FF,$05,$47,$33,$81,$F8,$F9,$FE,$AD,$3F,$CE
	.dc.b	$27,$D8,$70,$7F,$A6,$4F,$03,$91,$E6,$72,$D8,$FE,$BD,$1F,$09,$F4
	.dc.b	$3E,$D7,$E3,$BC,$1A,$70,$3D,$E7,$F8,$19,$F5,$4F,$F7,$07,$84,$79
	.dc.b	$4E,$13,$E5,$8C,$CE,$E7,$33,$E5,$7F,$B1,$9D,$A3,$E8,$CF,$63,$FB
	.dc.b	$5C,$70,$3C,$A7,$09,$EC,$E4,$F1,$19,$7B,$3F,$DA,$9D,$B9,$39,$99
	.dc.b	$7F,$CE,$55,$FB,$98,$FC,$E1,$FA,$88,$FD,$64,$66,$3F,$38,$7E,$61
	.dc.b	$FC,$E3,$F9,$83,$89,$FA,$C3,$F3,$87,$EE,$0C,$BF,$9C,$72,$7E,$E2
	.dc.b	$7F,$41,$7F,$C9,$23,$F4,$07,$13,$C2,$79,$4C,$9C,$4E,$AF,$17,$AC
	.dc.b	$FA,$5F,$D8,$1C,$4C,$C7,$F4,$C7,$A9,$E9,$8F,$6B,$E1,$38,$99,$8F
	.dc.b	$E3,$0F,$84,$E3,$1E,$D3,$8B,$93,$8B,$98,$FB,$F1,$F5,$4F,$B6,$7D
	.dc.b	$53,$EB,$DF,$EC,$2F,$EE,$B6,$3F,$C2,$B7,$E2,$BF,$71,$FD,$62,$3D
	.dc.b	$CF,$88,$CD,$FF,$56,$7C,$51,$FC,$EA,$3E,$13,$E6,$7E,$54,$71,$72
	.dc.b	$F5,$8F,$84,$71,$BF,$EC,$55,$FD,$74,$FF,$01,$47,$D6,$9F,$0C,$66
	.dc.b	$3F,$A0,$9C,$67,$F9,$A4,$F9,$23,$F9,$99,$F3,$2B,$C6,$7B,$CF,$19
	.dc.b	$EB,$32,$F3,$AF,$C5,$9E,$27,$EF,$D7,$99,$FE,$D3,$3C,$1E,$E7,$C1
	.dc.b	$9E,$15,$F0,$6B,$84,$FF,$3C,$AF,$14,$62,$3F,$49,$18,$8F,$5B,$D6
	.dc.b	$BF,$A0,$D7,$C9,$3D,$CE,$23,$D4,$F8,$1E,$4F,$AA,$BE,$39,$E6,$9E
	.dc.b	$53,$D6,$3F,$64,$7C,$79,$E4,$FC,$33,$F4,$AE,$1E,$51,$E0,$7A,$C7
	.dc.b	$8C,$FA,$D5,$FD,$A6,$3F,$45,$5E,$09,$E4,$6E,$3E,$5B,$87,$C6,$7C
	.dc.b	$B3,$A9,$88,$E4,$E0,$EB,$5E,$3A,$F0,$47,$EA,$67,$71,$87,$F1,$71
	.dc.b	$E0,$39,$1B,$9E,$53,$FC,$69,$FC,$3D,$7B,$23,$07,$D9,$9F,$D3,$CF
	.dc.b	$D0,$7E,$39,$D2,$BB,$BF,$81,$3E,$61,$F3,$CD,$47,$67,$C9,$3E,$78
	.dc.b	$C4,$71,$3E,$8D,$78,$DF,$D3,$46,$9F,$C5,$9D,$8F,$95,$3C,$CF,$43
	.dc.b	$FB,$23,$99,$E0,$7D,$A7,$E4,$A7,$EA,$47,$CE,$7C,$0F,$E9,$23,$4F
	.dc.b	$A9,$EB,$5F,$96,$8F,$B1,$5F,$31,$F3,$CE,$A3,$F2,$51,$D4,$F1,$1A
	.dc.b	$8E,$A7,$A4,$A7,$9A,$BF,$3C,$E5,$AE,$63,$E9,$B9,$6B,$97,$E9,$39
	.dc.b	$6B,$97,$E4,$B6,$BF,$16,$DA,$F6,$4F,$C3,$AF,$0B,$E9,$6B,$93,$C2
	.dc.b	$64,$A6,$4F,$09,$92,$99,$3C,$66,$4A,$66,$BC,$AE,$5A,$E6,$7C,$F1
	.dc.b	$FC,$56,$BE,$5C,$EE,$72,$7E,$6D,$F0,$CE,$EB,$D6,$D8,$F4,$B5,$DC
	.dc.b	$F1,$7C,$B1,$96,$C6,$63,$CA,$64,$AE,$4F,$1C,$7E,$6E,$32,$D8,$CB
	.dc.b	$F9,$A3,$F4,$D3,$FC,$35,$A7,$E4,$9F,$CC,$CF,$79,$E3,$1B,$9F,$D4
	.dc.b	$9C,$CA,$E5,$FD,$F9,$E1,$79,$C7,$9D,$B1,$93,$F3,$67,$D1,$72,$F5
	.dc.b	$39,$94,$FD,$A3,$DD,$A7,$A4,$E7,$3E,$98,$E6,$6D,$FD,$84,$7E,$AC
	.dc.b	$F8,$25,$3D,$67,$33,$27,$F2,$18,$E6,$E4,$D9,$C4,$F3,$BD,$E7,$F7
	.dc.b	$31,$C4,$F2,$BF,$0D,$A7,$C3,$3B,$CF,$F2,$47,$A9,$C4,$CC,$7F,$1B
	.dc.b	$7B,$9E,$32,$9E,$33,$E2,$96,$3E,$0B,$CC,$EA,$F7,$78,$99,$2B,$97
	.dc.b	$99,$92,$B9,$3A,$96,$3E,$0B,$67,$F9,$3C,$F1,$8D,$CF,$AE,$39,$B6
	.dc.b	$33,$5F,$C8,$EB,$8C,$73,$AF,$E5,$93,$FB,$67,$FC,$C3,$7F,$E8,$27
	.dc.b	$FD,$42,$FF,$A2,$5F,$F2,$23,$5F,$F2,$21,$FE,$46,$7F,$C8,$49,$57
	.dc.b	$D7,$8F,$A3,$3E,$37,$C3,$1F,$C0,$2F,$51,$F3,$1F,$51,$F0,$8B,$1E
	.dc.b	$A6,$C7,$23,$F8,$3B,$C9,$FA,$2E,$A7,$F9,$94,$7F,$06,$3F,$10,$77
	.dc.b	$7C,$86,$AF,$F8,$1C,$7E,$A5,$F9,$71,$F1,$DF,$D3,$47,$E4,$8D,$4F
	.dc.b	$9E,$79,$9F,$CE,$EB,$F4,$33,$F8,$B9,$D4,$F3,$7D,$11,$F6,$A3,$8B
	.dc.b	$FC,$A6,$78,$BF,$81,$9E,$31,$F2,$63,$D6,$7C,$09,$E2,$FF,$05,$3C
	.dc.b	$D1,$FA,$D3,$E5,$46,$AB,$D7,$3F,$5A,$3C,$8F,$43,$C8,$E1,$F7,$CF
	.dc.b	$28,$C3,$D6,$3E,$A1,$FC,$E0,$E5,$3F,$2A,$3D,$11,$C8,$D9,$F5,$4C
	.dc.b	$4F,$D3,$AF,$E5,$71,$FB,$73,$C7,$3C,$8F,$04,$ED,$F8,$F5,$F9,$F8
	.dc.b	$C4,$FC,$23,$94,$78,$DF,$C4,$9E,$6A,$F9,$AE,$0F,$1C,$FD,$B7,$91
	.dc.b	$F6,$EB,$E1,$1E,$37,$AC,$F2,$7E,$11,$C8,$FB,$87,$23,$EE,$47,$27
	.dc.b	$C0,$FB,$27,$C1,$1F,$56,$3B,$3E,$13,$E4,$D7,$E6,$23,$84,$76,$38
	.dc.b	$1D,$0F,$2B,$DD,$F2,$BC,$1F,$C4,$1C,$0C,$9C,$0E,$F5,$F5,$1F,$15
	.dc.b	$7C,$E7,$81,$F6,$E7,$A5,$7E,$3A,$3F,$61,$3F,$88,$7A,$3C,$5C,$CF
	.dc.b	$B9,$F8,$A7,$81,$FB,$B5,$FB,$D9,$F8,$B3,$C0,$F7,$1E,$47,$2F,$E9
	.dc.b	$5F,$03,$93,$E3,$1C,$4F,$79,$93,$E1,$99,$3D,$87,$A1,$EA,$FE,$50
	.dc.b	$F0,$9E,$87,$D8,$F8,$67,$D9,$5F,$95,$9E,$2F,$E4,$E7,$C2,$71,$8C
	.dc.b	$D7,$E2,$E3,$C2,$79,$67,$E6,$3D,$4F,$0B,$D5,$F6,$99,$3A,$B9,$7D
	.dc.b	$A6,$4E,$A7,$84,$E2,$7B,$4E,$31,$F4,$5C,$C7,$A5,$E2,$F5,$10,$89
	.dc.b	$55,$C5,$FF,$51,$70,$7F,$4B,$8F,$E9,$6A,$E3,$C2,$39,$1C,$27,$94
	.dc.b	$75,$7F,$A9,$C7,$FB,$C5,$FE,$B4,$7F,$58,$2B,$DA,$39,$3E,$09,$EF
	.dc.b	$1F,$D7,$E3,$4F,$57,$51,$63,$DA,$65,$F4,$47,$43,$DB,$1E,$97,$C8
	.dc.b	$7F,$42,$BF,$94,$7E,$EE,$3F,$1C,$7C,$EB,$FC,$14,$66,$3F,$05,$1C
	.dc.b	$4C,$9F,$B5,$7E,$D1,$F7,$8F,$B4,$7D,$67,$ED,$39,$8F,$0C,$71,$7E
	.dc.b	$D3,$D6,$7F,$84,$4F,$1A,$FE,$95,$39,$8F,$DB,$1F,$C7,$6B,$8B,$FC
	.dc.b	$B6,$B8,$C7,$A6,$3C,$AF,$E1,$9F,$92,$7D,$DA,$F9,$2F,$CF,$9F,$92
	.dc.b	$FE,$15,$F9,$2F,$03,$F1,$11,$C1,$FF,$0C,$C7,$E1,$9F,$A9,$1F,$6A
	.dc.b	$FF,$05,$39,$9F,$04,$E6,$3E,$FC,$7D,$73,$F2,$51,$F7,$DD,$3D,$DF
	.dc.b	$F1,$2C,$FE,$32,$7C,$EF,$E4,$A3,$89,$F9,$93,$D0,$FE,$61,$ED,$3F
	.dc.b	$90,$8F,$CA,$4F,$F2,$AB,$F9,$D3,$FC,$32,$BE,$DD,$FC,$B3,$E7,$BF
	.dc.b	$D3,$CF,$B9,$3F,$84,$AF,$E5,$35,$F6,$27,$EA,$1F,$60,$FE,$73,$5C
	.dc.b	$EB,$F0,$AF,$F0,$E8,$EF,$1F,$85,$32,$79,$27,$F8,$BD,$7D,$E8,$D4
	.dc.b	$77,$9F,$25,$73,$3F,$1F,$5E,$49,$FE,$6B,$3F,$84,$AE,$8F,$E8,$67
	.dc.b	$A1,$B3,$F9,$5C,$78,$A3,$D7,$3F,$16,$BA,$C7,$E4,$E7,$4F,$89,$EA
	.dc.b	$FC,$6A,$EA,$EE,$BD,$11,$F4,$63,$AC,$7B,$9F,$90,$7F,$45,$AF,$91
	.dc.b	$1C,$23,$F7,$27,$53,$67,$F2,$F8,$DC,$7C,$07,$EC,$CF,$79,$F3,$1E
	.dc.b	$03,$E1,$1C,$8F,$A6,$FC,$37,$FA,$0D,$7B,$A3,$91,$23,$C7,$3F,$46
	.dc.b	$7F,$09,$1E,$FA,$F1,$47,$18,$DB,$F1,$A7,$85,$7C,$69,$FC,$3C,$FD
	.dc.b	$EA,$E2,$7C,$18,$FE,$2A,$7A,$4F,$84,$FD,$58,$F0,$C7,$29,$F5,$3F
	.dc.b	$D3,$A7,$93,$F8,$33,$95,$78,$9F,$84,$FD,$E7,$91,$EB,$9F,$C2,$CF
	.dc.b	$29,$F9,$B3,$EC,$39,$D7,$B4,$8E,$08,$7D,$22,$57,$CB,$7D,$A6,$0E
	.dc.b	$85,$32,$ED,$F5,$19,$2B,$99,$F2,$C7,$03,$9C,$66,$7E,$A4,$64,$87
	.dc.b	$23,$64,$33,$3D,$1D,$18,$8E,$6E,$09,$1B,$24,$69,$B5,$F2,$E3,$8D
	.dc.b	$FF,$19,$AF,$E6,$87,$F1,$E3,$FB,$45,$FF,$22,$3F,$B3,$BF,$C8,$4F
	.dc.b	$E5,$35,$FC,$BE,$34,$7F,$3C,$8C,$BF,$DA,$5F,$F6,$AB,$FC,$DD,$FE
	.dc.b	$7F,$1C,$5C,$9F,$CB,$23,$F9,$B9,$C5,$E5,$1F,$CB,$63,$94,$FA,$4F
	.dc.b	$E6,$D1,$99,$FC,$0C,$FD,$18,$F8,$04,$3F,$3E,$47,$F1,$ED,$8F,$D9
	.dc.b	$BB,$8C,$92,$33,$1B,$7F,$40,$D7,$FB,$73,$6F,$F8,$3C,$F1,$9D,$B5
	.dc.b	$DB,$63,$71,$DA,$31,$7F,$B7,$69,$FB,$08,$E2,$6C,$87,$62,$9A,$30
	.dc.b	$43,$04,$9F,$0B,$96,$B9,$9F,$13,$D2,$32,$48,$CC,$71,$6B,$C6,$7D
	.dc.b	$13,$B6,$CE,$A3,$13,$FC,$0C,$ED,$19,$24,$F0,$24,$6C,$F0,$38,$7F
	.dc.b	$86,$90,$FE,$81,$1C,$9E,$07,$94,$B3,$E2,$78,$1F,$D0,$63,$83,$83
	.dc.b	$F9,$D9,$C1,$CB,$DA,$7C,$4D,$34,$E1,$E8,$D7,$99,$D6,$7F,$93,$3C
	.dc.b	$4C,$4F,$F3,$27,$04,$39,$9D,$88,$70,$9F,$E8,$4E,$CE,$E5,$9E,$EF
	.dc.b	$23,$81,$FC,$E2,$7C,$0E,$4E,$0E,$1E,$66,$48,$60,$91,$D9,$B1,$C6
	.dc.b	$30,$47,$04,$9F,$4C,$F4,$9D,$46,$23,$9B,$63,$31,$FC,$75,$FF,$2C
	.dc.b	$3F,$C5,$8F,$F2,$B1,$FC,$48,$C1,$FE,$52,$3F,$8A,$57,$F9,$1E,$FF
	.dc.b	$CA,$31,$88,$FE,$34,$FF,$16,$7F,$C9,$D5,$FE,$94,$78,$3F,$E5,$48
	.dc.b	$FE,$5F,$1F,$C5,$DE,$E7,$F9,$4A,$3F,$8E,$9F,$C8,$9F,$29,$D4,$FE
	.dc.b	$64,$F5,$8F,$0C,$7A,$9E,$91,$FC,$85,$FE,$25,$7F,$94,$9F,$90,$FC
	.dc.b	$B7,$EE,$4F,$C7,$8C,$4F,$F7,$F8,$C1,$27,$64,$8D,$35,$FF,$17,$18
	.dc.b	$21,$82,$3F,$E2,$47,$2D,$72,$FF,$84,$1B,$7F,$E0,$97,$71,$93,$81
	.dc.b	$0C,$12,$38,$47,$2A,$F9,$AF,$D4,$AF,$EB,$35,$FD,$86,$FF,$D0,$4F
	.dc.b	$E0,$E3,$E7,$1F,$D4,$DF,$9D,$3F,$79,$FF,$4A,$C6,$5F,$F4,$1C,$65
	.dc.b	$FF,$4A,$C6,$5F,$C0,$C6,$5F,$EC,$71,$99,$FE,$B7,$39,$7F,$AF,$47
	.dc.b	$F7,$29,$FE,$EA,$79,$4F,$09,$F8,$63,$CA,$7E,$18,$C9,$F9,$B3,$F1
	.dc.b	$F1,$F9,$A6,$C7,$E5,$63,$C3,$1C,$4F,$49,$D4,$F6,$99,$3D,$26,$4F
	.dc.b	$69,$97,$D3,$1F,$E0,$53,$E5,$C6,$E3,$BB,$EC,$3F,$3C,$53,$D6,$53
	.dc.b	$C2,$53,$B9,$5E,$67,$DA,$79,$B5,$E6,$D7,$9B,$5E,$6D,$39,$C7,$E2
	.dc.b	$27,$F5,$F3,$C2,$30,$7D,$93,$EB,$4F,$16,$9A,$8F,$80,$E1,$FE,$14
	.dc.b	$F1,$3E,$33,$F6,$0F,$E0,$86,$8E,$2E,$9F,$AA,$7D,$D7,$E6,$B4,$E5
	.dc.b	$1F,$7E,$33,$1D,$0F,$C0,$BF,$A2,$9F,$E1,$31,$C2,$3F,$08,$79,$DD
	.dc.b	$9F,$91,$3C,$27,$37,$AC,$F7,$8E,$51,$D6,$7F,$38,$FF,$1D,$8F,$31
	.dc.b	$CA,$78,$36,$3C,$27,$BD,$FC,$79,$C0,$A7,$84,$EE,$79,$4C,$C7,$46
	.dc.b	$C7,$47,$F8,$03,$F5,$5E,$91,$F9,$87,$27,$D1,$33,$1E,$97,$C0,$79
	.dc.b	$DF,$73,$FA,$17,$51,$E1,$3A,$4F,$CE,$9E,$07,$9E,$33,$1F,$18,$FE
	.dc.b	$2D,$19,$9E,$67,$A6,$39,$47,$B2,$3C,$2F,$E8,$DF,$B0,$7D,$43,$84
	.dc.b	$75,$3F,$37,$1F,$A0,$8F,$84,$FB,$0E,$87,$86,$3C,$6F,$EB,$23,$73
	.dc.b	$ED,$7C,$A7,$89,$FD,$13,$F6,$5F,$6C,$7C,$59,$E6,$FE,$D2,$79,$9E
	.dc.b	$67,$51,$E6,$9F,$D9,$C7,$D3,$3C,$CF,$47,$E0,$4F,$73,$C6,$FA,$67
	.dc.b	$94,$7C,$48,$FC,$39,$EB,$3A,$4F,$AE,$7C,$B1,$FB,$F3,$33,$E5,$7B
	.dc.b	$9F,$86,$3F,$4A,$73,$29,$ED,$38,$BF,$14,$F7,$95,$C9,$E7,$7F,$82
	.dc.b	$3D,$5F,$B4,$6D,$EA,$79,$4A,$7D,$B3,$99,$E3,$78,$BD,$4F,$E1,$67
	.dc.b	$53,$E1,$9E,$73,$9C,$FD,$06,$C7,$C9,$7F,$32,$75,$38,$9E,$37,$BB
	.dc.b	$93,$67,$57,$2D,$38,$9E,$93,$BB,$F6,$E7,$E9,$C6,$63,$8B,$DC,$E3
	.dc.b	$3E,$78,$CC,$6C,$CB,$E9,$7F,$3A,$79,$4A,$78,$CF,$39,$D4,$A7,$32
	.dc.b	$C7,$77,$AC,$79,$E7,$F1,$ED,$3E,$DB,$FB,$89,$FC,$C1,$FA,$69,$FA
	.dc.b	$47,$49,$F3,$46,$9F,$C3,$C7,$D9,$3F,$89,$CF,$C4,$9F,$14,$71,$9F
	.dc.b	$03,$F7,$A7,$C1,$1F,$A4,$8F,$81,$3D,$5F,$C3,$9F,$04,$F8,$67,$C1
	.dc.b	$29,$F0,$4A,$F8,$1B,$1E,$83,$B3,$CE,$3B,$3F,$70,$F7,$9C,$23,$CE
	.dc.b	$FB,$27,$E4,$BA,$9E,$33,$F7,$1F,$01,$88,$FA,$26,$E3,$C2,$FA,$63
	.dc.b	$4F,$C5,$35,$3F,$26,$7D,$0D,$7C,$71,$FC,$39,$E6,$FC,$19,$E6,$FA
	.dc.b	$4E,$73,$E8,$3E,$79,$CD,$CC,$FE,$0A,$3F,$C8,$56,$BF,$E9,$8F,$FB
	.dc.b	$46,$BF,$D5,$B5,$FE,$71,$AF,$F3,$0C,$FF,$97,$27,$FC,$C0,$D7,$FC
	.dc.b	$C1,$46,$57,$FA,$0C,$C3,$FE,$E6,$29,$B9,$F1,$47,$1A,$1C,$23,$FD
	.dc.b	$53,$1F,$44,$F8,$F1,$EF,$3F,$A6,$3E,$D9,$FE,$72,$D7,$F9,$3C,$7C
	.dc.b	$43,$83,$A3,$F2,$CF,$26,$99,$3F,$09,$1C,$EC,$FB,$0F,$74,$79,$EF
	.dc.b	$FD,$40,$FF,$B1,$0F,$F4,$B1,$F2,$1E,$0E,$EB,$DF,$3F,$12,$FE,$94
	.dc.b	$FF,$AA,$23,$E9,$4F,$FA,$2E,$3E,$69,$C5,$FF,$57,$C7,$FB,$EE,$7F
	.dc.b	$DA,$31,$F6,$2B,$BD,$7C,$F7,$EF,$9E,$69,$DC,$F5,$9F,$69,$D6,$BC
	.dc.b	$95,$B3,$FD,$6F,$3B,$8F,$83,$3F,$74,$FF,$7C,$C7,$85,$F3,$9E,$17
	.dc.b	$DE,$FC,$78,$F4,$4F,$48,$F8,$C7,$01,$5F,$97,$AE,$31,$DC,$FC,$B9
	.dc.b	$F3,$8F,$C9,$9C,$63,$84,$7C,$68,$FC,$89,$F5,$5F,$33,$F0,$0F,$33
	.dc.b	$B7,$D6,$FB,$DF,$88,$7A,$4E,$71,$F2,$9F,$49,$CC,$F4,$CF,$D0,$AF
	.dc.b	$79,$F7,$23,$6F,$CA,$76,$E9,$F6,$4F,$EB,$C9,$3F,$72,$3A,$D7,$C4
	.dc.b	$78,$9E,$48,$F9,$8F,$C4,$3C,$93,$F1,$2B,$EE,$9D,$EB,$E1,$BE,$38
	.dc.b	$FC,$9C,$FC,$79,$F8,$4F,$C6,$9E,$2F,$77,$D8,$F7,$79,$57,$E2,$5C
	.dc.b	$D7,$AA,$7E,$B4,$FC,$57,$91,$F0,$1F,$74,$79,$E3,$E8,$1F,$50,$F9
	.dc.b	$27,$D4,$3F,$08,$7E,$A4,$E4,$76,$AF,$8D,$1E,$C3,$EC,$47,$9C,$E4
	.dc.b	$7C,$A3,$B1,$93,$E0,$1E,$23,$F2,$31,$F2,$1E,$31,$D6,$7E,$8B,$EF
	.dc.b	$8F,$A3,$1E,$F7,$AC,$66,$BE,$39,$F4,$EB,$E7,$CF,$DE,$7F,$28,$F8
	.dc.b	$1C,$CF,$C5,$8F,$82,$7D,$73,$E2,$9E,$83,$2F,$C0,$3C,$0F,$A2,$BE
	.dc.b	$B3,$F2,$0F,$3C,$FC,$88,$CB,$EF,$8E,$87,$C7,$3C,$87,$80,$F4,$57
	.dc.b	$E1,$9F,$7C,$FD,$C3,$E1,$3F,$AB,$9F,$E8,$2F,$F1,$73,$FC,$0A,$7F
	.dc.b	$6B,$9F,$EC,$77,$FD,$C9,$A7,$F4,$43,$F9,$B1,$FE,$18,$8F,$F0,$7B
	.dc.b	$A3,$FB,$6C,$7F,$9B,$E7,$FA,$1C,$FF,$46,$7F,$A1,$3F,$D6,$E3,$F9
	.dc.b	$FC,$FF,$98,$4F,$AB,$1F,$96,$9F,$84,$FB,$8F,$C2,$1F,$91,$73,$7F
	.dc.b	$56,$3E,$31,$98,$FB,$37,$F0,$A3,$07,$C9,$9E,$07,$E2,$EB,$F5,$33
	.dc.b	$F2,$6B,$EA,$47,$E1,$EB,$84,$F3,$7F,$04,$7E,$64,$F3,$1F,$82,$7F
	.dc.b	$2A,$73,$9F,$C6,$C7,$88,$E1,$5E,$29,$FC,$0C,$FA,$0F,$0C,$FC,$37
	.dc.b	$07,$27,$C3,$3F,$8E,$7F,$0E,$FA,$DF,$C7,$57,$CC,$7F,$0A,$7E,$DE
	.dc.b	$79,$4F,$E5,$EB,$CD,$5F,$8B,$AF,$C7,$9E,$68,$F8,$13,$B8,$E5,$3B
	.dc.b	$7D,$91,$E3,$7E,$19,$F5,$0F,$3C,$F6,$7E,$D9,$83,$99,$88,$EC,$FE
	.dc.b	$9A,$BC,$F3,$F3,$23,$67,$CF,$AF,$B2,$E0,$F0,$57,$28,$F8,$51,$EC
	.dc.b	$8F,$31,$F4,$63,$EF,$47,$BE,$BE,$8B,$F2,$0F,$14,$7B,$DF,$41,$83
	.dc.b	$D1,$5E,$57,$07,$94,$FA,$67,$F0,$38,$EF,$5E,$29,$F9,$A7,$18,$F3
	.dc.b	$CF,$CC,$8F,$90,$F3,$AF,$D0,$C7,$09,$F8,$33,$E0,$3E,$0C,$73,$7C
	.dc.b	$07,$E2,$CF,$CD,$CF,$B9,$F8,$91,$D9,$E3,$3D,$0F,$1B,$F0,$1B,$5F
	.dc.b	$36,$3A,$9F,$1E,$33,$3E,$18,$FC,$AB,$F9,$E7,$C3,$5F,$72,$7B,$47
	.dc.b	$23,$C6,$7D,$BA,$FC,$E4,$FC,$03,$D1,$1F,$14,$D5,$75,$7F,$69,$7F
	.dc.b	$E7,$C2,$5F,$F1,$78,$FF,$3C,$9F,$E5,$43,$F8,$B1,$FE,$55,$3F,$D0
	.dc.b	$D5,$FE,$7F,$3F,$CB,$D1,$FC,$8A,$3B,$BF,$6A,$BE,$45,$F9,$AB,$F6
	.dc.b	$F3,$F8,$78,$F8,$D5,$F0,$A3,$83,$C8,$FA,$87,$E4,$67,$E2,$3F,$88
	.dc.b	$38,$1A,$9F,$C4,$47,$C2,$8F,$A8,$7C,$93,$E2,$1F,$83,$8F,$23,$F4
	.dc.b	$27,$C9,$5F,$E5,$4A,$FE,$8B,$3F,$E2,$C3,$FA,$29,$FE,$A5,$9F,$91
	.dc.b	$1F,$0A,$3E,$45,$FF,$B7,$E3,$11,$FE,$7C,$33,$3F,$D1,$DF,$F6,$43
	.dc.b	$CA,$3F,$A1,$C2,$7F,$A1,$BF,$E4,$A7,$FA,$94,$7D,$13,$FD,$5B,$5F
	.dc.b	$9F,$9F,$88,$7F,$0E,$BF,$6B,$F2,$6F,$DA,$7E,$31,$F9,$27,$E5,$4F
	.dc.b	$8C,$7C,$83,$27,$E4,$0F,$91,$3F,$8B,$9F,$9F,$5F,$50,$FC,$91,$4F
	.dc.b	$BF,$3F,$C0,$AB,$C0,$F9,$6B,$91,$F8,$62,$BF,$81,$3C,$87,$88,$FB
	.dc.b	$31,$FA,$A3,$9D,$7F,$17,$8F,$C1,$4F,$39,$FC,$5C,$7A,$1A,$71,$7E
	.dc.b	$3C,$F1,$7B,$9E,$58,$DB,$F1,$23,$F2,$EE,$4E,$E7,$9C,$F8,$D1,$D6
	.dc.b	$BE,$1B,$F4,$23,$BD,$7D,$79,$CC,$7C,$53,$2F,$E6,$6B,$B9,$E5,$8E
	.dc.b	$A7,$8E,$7F,$85,$3E,$33,$D3,$1B,$7E,$39,$BA,$F8,$F3,$E1,$AF,$AB
	.dc.b	$3F,$CA,$5A,$FF,$1E,$3C,$6F,$DB,$7F,$4C,$79,$DF,$4B,$C4,$C9,$4F
	.dc.b	$C8,$1B,$7A,$D7,$E5,$0F,$C4,$D7,$C5,$8E,$73,$E1,$6C,$EE,$BE,$E4
	.dc.b	$FD,$03,$EE,$16,$BE,$85,$76,$9F,$C5,$CE,$8F,$8A,$FC,$17,$CE,$77
	.dc.b	$8E,$13,$F8,$08,$FC,$81,$FA,$33,$E7,$1F,$48,$CD,$7D,$6A,$E9,$1F
	.dc.b	$00,$F8,$51,$EE,$7E,$11,$E0,$7D,$8F,$F0,$CB,$FD,$ED,$FF,$91,$9F
	.dc.b	$E1,$4F,$F0,$47,$FC,$95,$7F,$CB,$E7,$FD,$81,$1F,$CE,$A7,$F8,$24
	.dc.b	$7F,$37,$8F,$E1,$2F,$F9,$52,$32,$FF,$0D,$9F,$96,$FC,$48,$FC,$85
	.dc.b	$FE,$32,$FF,$15,$3F,$05,$A7,$BE,$77,$5F,$8C,$6C,$FE,$22,$3E,$E4
	.dc.b	$7C,$C8,$FA,$C7,$E4,$CD,$1F,$10,$F9,$07,$D3,$3E,$35,$7D,$A9,$FB
	.dc.b	$B1,$87,$EB,$3F,$50,$FC,$F9,$F5,$23,$D4,$7D,$4A,$FA,$51,$F5,$23
	.dc.b	$E0,$BF,$6C,$E4,$7B,$DE,$06,$9E,$A6,$A3,$E1,$57,$E9,$63,$D3,$1F
	.dc.b	$71,$FC,$81,$D5,$F9,$33,$E4,$9F,$8E,$7E,$79,$F7,$4F,$C5,$8C,$57
	.dc.b	$BA,$B8,$BF,$3E,$3C,$11,$EC,$9F,$90,$78,$AB,$9B,$F0,$EB,$81,$F5
	.dc.b	$1F,$94,$7D,$43,$47,$D4,$3E,$31,$FC,$74,$F8,$33,$E2,$74,$78,$1F
	.dc.b	$11,$F0,$63,$EB,$C7,$8D,$F5,$4F,$CC,$8E,$87,$CF,$AF,$86,$7E,$1D
	.dc.b	$FC,$A3,$EE,$8F,$A5,$1B,$8C,$3C,$AB,$E9,$D7,$B6,$3C,$07,$B4,$FB
	.dc.b	$15,$F4,$6B,$8C,$7D,$63,$DB,$1E,$57,$07,$79,$E9,$5E,$E9,$F9,$B3
	.dc.b	$D2,$7E,$6B,$F9,$B9,$D4,$7C,$0A,$F6,$CF,$FB,$5E,$FF,$71,$1F,$BF
	.dc.b	$9F,$F0,$24,$FF,$B4,$63,$F6,$F3,$FC,$0A,$3F,$79,$23,$FD,$8C,$8F
	.dc.b	$F6,$53,$FE,$0B,$76,$E9,$CC,$F4,$8E,$F1,$FC,$0D,$E5,$1F,$E2,$09
	.dc.b	$ED,$3F,$36,$FE,$45,$7C,$27,$E9,$DF,$D0,$9F,$81,$18,$36,$E0,$E5
	.dc.b	$3F,$19,$F4,$4A,$7B,$C7,$AC,$CB,$FE,$33,$BC,$4E,$E7,$F9,$45,$7F
	.dc.b	$BC,$DF,$F1,$69,$87,$FD,$E2,$74,$7F,$C5,$F3,$F0,$5D,$1F,$E6,$03
	.dc.b	$FC,$6C,$43,$FC,$BE,$FD,$57,$6F,$F8,$F1,$CC,$7F,$33,$3D,$0F,$F2
	.dc.b	$D8,$EC,$FF,$36,$8F,$A8,$FF,$9B,$09,$7F,$75,$B1,$F5,$23,$FA,$1B
	.dc.b	$D6,$3E,$24,$7E,$22,$31,$1F,$72,$7C,$CF,$77,$C3,$1E,$49,$E1,$3F
	.dc.b	$18,$D4,$F0,$9D,$1E,$69,$F2,$11,$F2,$1F,$3E,$7E,$34,$7F,$54,$7A
	.dc.b	$47,$29,$F0,$C7,$A2,$3A,$1C,$8F,$AA,$7E,$55,$F9,$11,$E3,$78,$3C
	.dc.b	$9F,$74,$7C,$07,$6D,$9F,$6C,$79,$23,$F0,$0F,$EA,$E7,$A3,$C8,$F3
	.dc.b	$9D,$9C,$BD,$0C,$9B,$3A,$47,$9E,$7B,$C7,$EA,$4F,$F3,$B4,$7E,$5D
	.dc.b	$FD,$99,$FA,$03,$F9,$3D,$7F,$25,$7F,$9F,$57,$F1,$F7,$FA,$35,$FF
	.dc.b	$11,$AF,$E5,$35,$F9,$F9,$FE,$26,$FE,$9E,$3F,$6F,$5F,$A4,$AF,$99
	.dc.b	$5F,$9B,$3F,$2A,$7A,$8F,$61,$EA,$3C,$F3,$F4,$9A,$71,$BF,$B8,$FF
	.dc.b	$0C,$8F,$80,$7D,$A7,$D9,$5F,$D5,$9F,$E5,$D7,$FD,$16,$31,$1F,$CE
	.dc.b	$A7,$F8,$43,$C2,$3E,$03,$FC,$C1,$FE,$4A,$FF,$38,$38,$1F,$C2,$4F
	.dc.b	$83,$7F,$C3,$1F,$EA,$A6,$27,$F8,$63,$FD,$4E,$3F,$89,$46,$2B,$E8
	.dc.b	$4F,$F5,$EA,$F2,$BF,$41,$F0,$C7,$F3,$88,$F5,$D7,$D3,$7E,$B9,$FD
	.dc.b	$A0,$F2,$9E,$2A,$FA,$F1,$E0,$8F,$BE,$FF,$46,$9F,$89,$5F,$02,$7F
	.dc.b	$A6,$1E,$DA,$E2,$FC,$53,$B1,$E5,$8F,$EA,$F1,$A8,$F8,$07,$67,$FA
	.dc.b	$74,$6A,$3C,$E7,$CB,$39,$47,$D2,$78,$C7,$C2,$7E,$93,$C5,$F8,$CF
	.dc.b	$9E,$7F,$9F,$47,$BC,$EA,$FC,$77,$E1,$9F,$44,$F5,$1E,$13,$AC,$F8
	.dc.b	$4F,$51,$1E,$75,$F2,$6B,$C9,$1F,$18,$F0,$1E,$43,$E2,$9D,$0F,$94
	.dc.b	$72,$3C,$35,$E5,$7E,$09,$A3,$91,$E5,$3B,$1E,$93,$E2,$C7,$C1,$9F
	.dc.b	$C8,$46,$6B,$D0,$F2,$AF,$DB,$47,$F4,$27,$EF,$C7,$F5,$A3,$FA,$4D
	.dc.b	$7F,$9E,$E7,$F8,$AC,$7F,$09,$8F,$E9,$97,$FB,$08,$FD,$C9,$FB,$73
	.dc.b	$FC,$F5,$1F,$C5,$A3,$FC,$04,$FF,$25,$8F,$DE,$9F,$C4,$CF,$E7,$73
	.dc.b	$FC,$19,$FE,$26,$7B,$0F,$92,$7C,$0A,$FD,$2B,$FE,$1B,$3F,$18,$7E
	.dc.b	$8A,$7F,$4B,$7F,$55,$F8,$95,$FA,$D8,$FC,$F1,$FA,$79,$F8,$8E,$DE
	.dc.b	$F5,$F8,$A7,$E5,$1F,$5C,$F2,$1F,$4D,$D1,$E4,$3E,$5B,$98,$FA,$71
	.dc.b	$97,$C1,$1E,$9A,$FC,$4C,$F2,$9F,$2C,$7A,$A7,$F0,$B1,$E5,$7A,$57
	.dc.b	$AC,$FD,$59,$F9,$77,$CC,$FE,$01,$ED,$3F,$09,$F8,$E7,$E8,$AB,$C9
	.dc.b	$1D,$E3,$DB,$1F,$93,$7F,$30,$7E,$70,$E6,$7C,$6A,$F6,$9E,$89,$FD
	.dc.b	$A4,$68,$F3,$9E,$03,$F3,$A7,$39,$F8,$E7,$77,$EF,$4F,$3A,$F6,$47
	.dc.b	$D6,$3E,$19,$0F,$3C,$7C,$59,$E4,$7C,$B9,$F8,$B3,$FA,$B3,$C9,$1A
	.dc.b	$9F,$4C,$EA,$7C,$91,$F5,$A7,$CA,$F5,$24,$78,$67,$E7,$4E,$A3,$07
	.dc.b	$D2,$21,$E9,$3C,$E6,$8C,$47,$29,$F0,$12,$BE,$A4,$7F,$85,$DF,$D6
	.dc.b	$3F,$D7,$67,$FC,$4A,$FF,$83,$4C,$1B,$31,$1F,$B1,$BF,$F0,$19,$FB
	.dc.b	$93,$FD,$E3,$3F,$E5,$08,$FF,$06,$DF,$AE,$3C,$E4,$AF,$34,$F9,$23
	.dc.b	$DA,$7C,$CB,$F8,$AE,$CC,$D7,$3A,$E3,$5F,$86,$8C,$4E,$AB,$E0,$4E
	.dc.b	$5B,$5F,$82,$9E,$C7,$F1,$63,$FD,$12,$7F,$81,$27,$F8,$AC,$7F,$18
	.dc.b	$BF,$F2,$AC,$FF,$83,$5F,$83,$19,$3F,$C8,$D1,$FC,$68,$91,$F1,$1F
	.dc.b	$F0,$60,$9F,$F4,$84,$FB,$8F,$59,$E5,$BF,$5C,$FD,$C3,$D9,$3F,$1D
	.dc.b	$C1,$0F,$59,$F1,$CC,$47,$27,$33,$D9,$FC,$D4,$64,$97,$F1,$A3,$2D
	.dc.b	$7F,$37,$5E,$17,$99,$E4,$7E,$71,$99,$E6,$D9,$E5,$3F,$C8,$8F,$2B
	.dc.b	$89,$FE,$13,$3C,$89,$5F,$0E,$74,$7E,$EE,$BC,$B3,$D4,$87,$07,$2D
	.dc.b	$7D,$2D,$7C,$47,$A8,$F1,$BC,$A7,$C0,$D8,$E7,$18,$8E,$55,$E8,$8C
	.dc.b	$12,$76,$43,$C2,$E5,$A6,$4F,$1C,$7C,$89,$EB,$5E,$F9,$D9,$0D,$BF
	.dc.b	$12,$BF,$88,$46,$27,$D0,$EE,$BC,$CF,$76,$B9,$AE,$CF,$19,$F0,$D7
	.dc.b	$4A,$DC,$E6,$3D,$4D,$3A,$57,$29,$D9,$1C,$12,$7A,$3F,$E0,$48,$FF
	.dc.b	$3F,$4F,$E9,$6B,$FC,$FC,$FE,$D4,$FF,$63,$96,$FF,$D6,$CD,$BF,$F4
	.dc.b	$1C,$FE,$E1,$FF,$3C,$BF,$C5,$DF,$E3,$71,$FE,$A1,$7F,$D5,$95,$F3
	.dc.b	$DF,$5B,$E7,$7C,$A4,$8F,$F0,$24,$FC,$46,$C7,$0B,$F6,$46,$0F,$87
	.dc.b	$1C,$5F,$EC,$D1,$A6,$C6,$5F,$EB,$67,$CB,$3E,$A3,$97,$FB,$49,$92
	.dc.b	$57,$9D,$F0,$1C,$0A,$6A,$7C,$0F,$68,$E0,$F8,$8D,$B8,$75,$1C,$1F
	.dc.b	$0C,$65,$B1,$97,$FB,$7C,$F9,$DA,$74,$8D,$4F,$F6,$B8,$C4,$EE,$31
	.dc.b	$1E,$32,$1E,$28,$D4,$FF,$6E,$8E,$24,$36,$79,$23,$67,$F8,$75,$D1
	.dc.b	$CC,$A6,$CA,$7A,$0E,$65,$38,$18,$9F,$3B,$82,$1C,$CD,$4F,$FB,$95
	.dc.b	$47,$F9,$A1,$FC,$91,$FD,$2E,$7F,$2A,$FF,$BC,$4F,$E9,$2F,$F4,$D3
	.dc.b	$FD,$B0,$FF,$99,$1F,$EE,$75,$FD,$E8,$FC,$F1,$0C,$1F,$DC,$AB,$F8
	.dc.b	$34,$FB,$AB,$D2,$F9,$CF,$49,$EF,$07,$A2,$3D,$51,$E9,$8F,$EA,$17
	.dc.b	$FD,$8A,$FF,$5F,$7E,$07,$CA,$49,$F7,$47,$96,$31,$1F,$E7,$F3,$FC
	.dc.b	$10,$E1,$F8,$2F,$43,$13,$FE,$0F,$7E,$24,$7F,$BE,$5B,$3F,$C0,$63
	.dc.b	$FB,$D3,$FE,$91,$8E,$51,$E2,$8F,$33,$63,$A9,$CA,$38,$47,$23,$FC
	.dc.b	$0A,$7D,$13,$81,$0F,$51,$FD,$F9,$F1,$3B,$32,$7F,$A8,$5E,$AE,$4F
	.dc.b	$79,$DC,$FE,$3E,$7F,$8D,$0E,$C7,$FA,$7C,$F8,$E7,$D6,$30,$78,$5C
	.dc.b	$4A,$7F,$C1,$D3,$FD,$DD,$FE,$30,$7E,$99,$FF,$36,$1F,$83,$21,$F8
	.dc.b	$32,$57,$E2,$1B,$7F,$B3,$7F,$3F,$3F,$EB,$17,$F1,$07,$F3,$52,$1E
	.dc.b	$B3,$E1,$D7,$F8,$B1,$B1,$FE,$2A,$8F,$BC,$FD,$88,$CB,$FA,$27,$E0
	.dc.b	$98,$3F,$04,$ED,$F9,$67,$CD,$8F,$F1,$93,$F4,$4F,$B6,$7C,$68,$FD
	.dc.b	$B4,$7C,$78,$FF,$36,$1F,$E0,$52,$57,$F4,$77,$F9,$E9,$FA,$BA,$F8
	.dc.b	$4F,$EA,$A3,$FA,$DD,$FF,$8F,$E3,$FD,$C8,$7F,$8D,$4F,$F6,$CD,$7B
	.dc.b	$9F,$C0,$3F,$D7,$1F,$8C,$7A,$8F,$79,$4F,$BC,$7A,$8A,$1F,$53,$DE
	.dc.b	$7D,$0F,$F4,$53,$EB,$9F,$80,$3E,$DD,$FF,$27,$73,$5F,$EC,$CB,$F6
	.dc.b	$C7,$E1,$2B,$F0,$44,$8C,$C7,$23,$A4,$66,$BA,$47,$C5,$9D,$CE,$63
	.dc.b	$71,$82,$4F,$08,$F0,$46,$1D,$C6,$5F,$05,$79,$E3,$15,$D6,$31,$3C
	.dc.b	$08,$E0,$87,$FB,$74,$AF,$FB,$56,$79,$D7,$86,$73,$3F,$ED,$46,$D7
	.dc.b	$FB,$0E,$B6,$49,$EF,$3D,$48,$F5,$32,$57,$9D,$7E,$E2,$3F,$CA,$B1
	.dc.b	$FE,$11,$7E,$AB,$C9,$F1,$36,$BC,$51,$D0,$C9,$D6,$3F,$C5,$AF,$E3
	.dc.b	$AF,$F2,$4F,$E9,$EF,$F9,$13,$F8,$D3,$F5,$07,$F1,$47,$FD,$88,$FF
	.dc.b	$14,$8E,$2F,$E5,$9F,$D7,$9F,$C4,$63,$F8,$23,$F9,$47,$FD,$25,$1F
	.dc.b	$93,$3F,$08,$FC,$79,$F8,$31,$FB,$2A,$FC,$CB,$FE,$1B,$9F,$E3,$13
	.dc.b	$A8,$FC,$F5,$FF,$5E,$7E,$29,$EF,$21,$F0,$89,$5F,$1A,$79,$57,$92
	.dc.b	$7B,$1F,$BD,$7C,$91,$F3,$6B,$EA,$4F,$F0,$47,$C0,$FA,$8F,$A1,$5D
	.dc.b	$67,$84,$7A,$E7,$6F,$E7,$2B,$AB,$E6,$AE,$0D,$9E,$F1,$F0,$8F,$F1
	.dc.b	$94,$7E,$82,$BA,$BF,$09,$FF,$40,$BE,$A3,$E1,$9F,$02,$79,$9E,$27
	.dc.b	$91,$DC,$F3,$47,$29,$EE,$7E,$AD,$EF,$5F,$1E,$33,$5F,$A5,$9C,$D7
	.dc.b	$F4,$F3,$F9,$39,$E2,$9F,$6B,$E3,$3B,$C7,$E7,$5F,$94,$FD,$47,$91
	.dc.b	$FE,$14,$8F,$F4,$3B,$FD,$84,$FF,$0B,$57,$F5,$18,$FF,$0B,$47,$F3
	.dc.b	$9B,$FF,$0B,$3F,$CB,$A7,$FD,$0F,$5F,$DC,$A3,$F3,$11,$FA,$07,$F3
	.dc.b	$C7,$F8,$D8,$FE,$9D,$1F,$DF,$1F,$E9,$11,$F9,$C3,$E5,$C7,$C7,$BF
	.dc.b	$F2,$F9,$F0,$E7,$E6,$1F,$56,$3D,$CF,$E0,$4F,$C3,$C7,$28,$CC,$7D
	.dc.b	$43,$FB,$59,$FD,$26,$BD,$6F,$ED,$EB,$FB,$04,$7E,$CA,$3F,$A1,$BF
	.dc.b	$D2,$E3,$F6,$F1,$EA,$9F,$EC,$0F,$F5,$EF,$B9,$7C,$77,$FA,$C6,$BE
	.dc.b	$64,$61,$F9,$D3,$FE,$44,$8C,$3F,$EB,$73,$13,$FA,$48,$E7,$43,$FA
	.dc.b	$9C,$A3,$8B,$DA,$7F,$CC,$D1,$E2,$8F,$24,$78,$DF,$F0,$F4,$70,$6C
	.dc.b	$70,$3C,$06,$DF,$1D,$7A,$1F,$04,$6D,$EB,$5E,$59,$F1,$1B,$AF,$6C
	.dc.b	$7F,$79,$9E,$27,$07,$D7,$1E,$38,$EE,$7C,$33,$67,$F7,$F7,$E1,$C7
	.dc.b	$0A,$F8,$73,$C2,$3B,$4F,$C1,$8E,$13,$EB,$7D,$35,$E5,$8F,$54,$6A
	.dc.b	$BD,$4F,$C4,$9D,$1F,$E3,$3A,$E8,$FE,$9F,$16,$3D,$5F,$D5,$D7,$F9
	.dc.b	$2D,$FF,$0C,$DF,$F4,$6A,$FA,$CF,$69,$F7,$9F,$20,$F8,$17,$D9,$E5
	.dc.b	$7D,$4F,$F7,$F1,$FE,$8A,$3F,$CB,$15,$D2,$BE,$23,$E8,$9E,$78,$7D
	.dc.b	$C9,$F0,$47,$C2,$3F,$27,$7F,$5A,$3F,$07,$1F,$56,$BC,$8F,$96,$3E
	.dc.b	$6D,$73,$7B,$C7,$28,$F9,$73,$F1,$6F,$D0,$FE,$27,$0F,$87,$5F,$E3
	.dc.b	$A8,$F4,$1D,$9F,$A0,$7A,$0F,$11,$83,$C5,$7E,$67,$EC,$3E,$27,$27
	.dc.b	$EA,$8E,$27,$CE,$3A,$BE,$A7,$27,$C3,$38,$9F,$90,$8F,$54,$7C,$89
	.dc.b	$FB,$B3,$FC,$A6,$BE,$93,$FE,$BE,$AF,$5B,$E7,$3B,$D7,$9E,$39,$BF
	.dc.b	$BA,$AE,$73,$FC,$C0,$FD,$03,$FC,$C3,$0F,$CF,$D7,$C5,$9F,$CF,$57
	.dc.b	$37,$F0,$AF,$96,$FF,$0B,$3F,$CE,$5F,$C2,$CF,$EF,$A7,$F0,$35,$FA
	.dc.b	$F8,$FC,$14,$7C,$37,$E5,$D7,$77,$31,$DD,$F8,$0F,$78,$FB,$95,$F0
	.dc.b	$A7,$F6,$93,$DB,$0F,$DA,$C7,$23,$EA,$1D,$8F,$8C,$7B,$0C,$D7,$C6
	.dc.b	$3F,$14,$76,$3D,$8F,$CF,$3D,$93,$F8,$17,$E7,$9C,$8F,$8A,$FC,$FA
	.dc.b	$F8,$B1,$F3,$9F,$8A,$FF,$30,$8F,$44,$FE,$A5,$F8,$8F,$8E,$3F,$66
	.dc.b	$F8,$DE,$47,$8C,$E5,$1E,$37,$D8,$7E,$2C,$F1,$C7,$B1,$FE,$87,$87
	.dc.b	$E9,$1F,$EA,$B1,$F8,$A7,$F3,$07,$2C,$3F,$82,$3F,$4A,$BD,$93,$E3
	.dc.b	$7F,$97,$57,$E2,$70,$FE,$41,$7F,$7C,$FD,$F5,$FC,$53,$CF,$7F,$7E
	.dc.b	$33,$1E,$73,$27,$D6,$3E,$D1,$93,$FD,$78,$FE,$59,$D3,$C8,$F3,$BF
	.dc.b	$D2,$2F,$E6,$BF,$C1,$E7,$F9,$65,$7D,$28,$D4,$FC,$97,$BD,$72,$BF
	.dc.b	$E5,$F3,$F9,$19,$F9,$38,$7E,$47,$0F,$91,$1F,$BB,$3F,$06,$79,$EF
	.dc.b	$CC,$FC,$C3,$AE,$1E,$18,$CC,$71,$3E,$84,$7F,$1F,$9F,$8C,$FF,$1E
	.dc.b	$8F,$C8,$C6,$0F,$5B,$F8,$33,$D6,$E5,$FE,$38,$7E,$84,$F1,$4F,$C7
	.dc.b	$9F,$11,$F6,$0F,$D4,$B9,$8E,$8F,$98,$FB,$91,$E3,$8F,$C2,$4F,$E3
	.dc.b	$5F,$C7,$5F,$E1,$2F,$AB,$F4,$A3,$F8,$FD,$FD,$63,$A5,$F8,$0E,$91
	.dc.b	$EA,$3E,$EB,$8C,$3E,$9D,$7E,$4A,$7E,$9E,$1F,$B6,$7F,$81,$C7,$DB
	.dc.b	$BF,$54,$62,$7E,$24,$FE,$BA,$3E,$59,$DA,$3E,$5B,$F1,$CF,$A8,$7D
	.dc.b	$83,$E1,$DF,$C7,$8F,$DC,$4F,$F1,$79,$F4,$1F,$3A,$7D,$18,$7F,$54
	.dc.b	$8F,$39,$FB,$A3,$C8,$78,$4F,$AF,$5F,$8C,$BF,$9E,$F9,$0C,$57,$C7
	.dc.b	$71,$3F,$BE,$3F,$83,$9D,$AF,$E9,$3E,$87,$E9,$3F,$96,$AF,$A5,$5E
	.dc.b	$B8,$FD,$31,$E2,$3E,$29,$EB,$AE,$EF,$E0,$0E,$E7,$E0,$2B,$FC,$20
	.dc.b	$F0,$3E,$4D,$7D,$29,$F4,$9F,$22,$3F,$82,$1F,$20,$DD,$7C,$8A,$FD
	.dc.b	$84,$75,$3F,$BD,$1D,$6B,$E2,$BE,$73,$F8,$49,$E7,$30,$75,$77,$1F
	.dc.b	$81,$8F,$51,$FD,$B6,$BF,$8E,$47,$AA,$39,$9E,$48,$F8,$2F,$3A,$FA
	.dc.b	$4F,$6A,$FC,$DC,$FB,$63,$F8,$C6,$1F,$98,$8F,$D3,$47,$D6,$BF,$AA
	.dc.b	$FE,$08,$F4,$3E,$57,$B4,$73,$AF,$AF,$3F,$6A,$7F,$83,$9F,$E0,$C8
	.dc.b	$F6,$C7,$E4,$DF,$1C,$7D,$C8,$E9,$18,$8F,$D9,$DF,$F0,$F7,$F2,$53
	.dc.b	$A8,$FC,$24,$68,$FC,$54,$7E,$B6,$3E,$B3,$FB,$38,$FA,$C7,$F4,$C3
	.dc.b	$EB,$47,$C6,$8E,$87,$A6,$3F,$58,$7C,$13,$E7,$1F,$62,$FE,$43,$F8
	.dc.b	$E3,$E4,$47,$D5,$7E,$29,$ED,$7E,$04,$FA,$23,$F3,$CF,$F1,$28,$F4
	.dc.b	$5F,$DA,$AE,$B5,$F8,$13,$F5,$13,$D6,$FE,$34,$72,$3F,$97,$03,$FC
	.dc.b	$FC,$60,$FF,$10,$C7,$FA,$45,$FF,$6F,$57,$FA,$6E,$3F,$CC,$71,$B3
	.dc.b	$FC,$18,$ED,$FE,$A8,$75,$76,$FF,$87,$0D,$BF,$EA,$57,$FC,$3E,$EA
	.dc.b	$31,$07,$94,$6D,$F1,$3F,$E2,$67,$66,$A3,$FD,$60,$E8,$AF,$13,$FD
	.dc.b	$4C,$7B,$0F,$F6,$6C,$78,$5D,$3E,$B7,$E2,$39,$79,$B9,$7D,$2F,$8D
	.dc.b	$E2,$F7,$7E,$71,$EA,$8F,$1B,$CA,$34,$E9,$EC,$FF,$B8,$23,$D8,$D3
	.dc.b	$CF,$1D,$DE,$47,$43,$F1,$67,$E7,$CF,$95,$F7,$5F,$DC,$0F,$F5,$B1
	.dc.b	$FE,$0B,$3F,$80,$61,$F8,$F8,$FD,$99,$C0,$FF,$0F,$C7,$07,$FD,$AE
	.dc.b	$F0,$9F,$D1,$5F,$F2,$58,$ED,$3F,$EE,$27,$B4,$FE,$A0,$FC,$CC,$7F
	.dc.b	$46,$8D,$BF,$EC,$E7,$71,$FA,$23,$AD,$7F,$A9,$A3,$74,$8F,$35,$7B
	.dc.b	$2B,$B5,$FF,$66,$7F,$D1,$57,$C2,$B7,$5E,$59,$DB,$FC,$21,$DB,$F0
	.dc.b	$0F,$51,$F0,$1D,$BF,$C7,$9D,$BF,$E1,$89,$F8,$CF,$8E,$3D,$47,$C0
	.dc.b	$3F,$CC,$B1,$E3,$A3,$F0,$27,$FA,$5D,$FE,$E1,$FA,$73,$A7,$C8,$7C
	.dc.b	$73,$C9,$5D,$A3,$E0,$47,$FA,$12,$BB,$1D,$5F,$F1,$E5,$FD,$F7,$F8
	.dc.b	$CC,$F5,$7D,$47,$C0,$3D,$47,$92,$38,$5F,$DF,$8F,$1B,$F4,$E7,$94
	.dc.b	$7C,$DB,$F2,$3F,$34,$F2,$1C,$2B,$E9,$BF,$81,$8F,$9A,$FF,$2A,$7E
	.dc.b	$3B,$B9,$EA,$EE,$3A,$BF,$4C,$EA,$7A,$8F,$8C,$7A,$8F,$82,$FE,$32
	.dc.b	$3F,$C2,$97,$F1,$A7,$E5,$9E,$F9,$D3,$F9,$73,$C5,$1F,$2E,$FF,$AD
	.dc.b	$CF,$DD,$3F,$06,$FF,$B4,$A3,$F0,$73,$F0,$CE,$93,$B7,$CF,$3B,$7E
	.dc.b	$81,$B3,$CF,$1B,$7F,$DD,$D3,$E7,$7D,$93,$E2,$AF,$C0,$C6,$8F,$39
	.dc.b	$EC,$39,$BA,$8F,$D5,$47,$FB,$CE,$7A,$3E,$18,$E8,$FD,$93,$A1,$F3
	.dc.b	$5F,$C0,$47,$C3,$BF,$EA,$D5,$D5,$FE,$42,$FE,$36,$7D,$D5,$F5,$0E
	.dc.b	$D3,$C1,$ED,$1E,$47,$F7,$AF,$03,$BB,$EB,$7C,$AF,$E0,$4F,$2B,$B8
	.dc.b	$FC,$6B,$E6,$BF,$C6,$BF,$E0,$E9,$FA,$32,$99,$5E,$78,$E2,$7F,$2A
	.dc.b	$8E,$2F,$F8,$0D,$FB,$EF,$17,$CE,$F0,$8F,$DE,$C7,$07,$FC,$45,$1F
	.dc.b	$12,$FE,$7D,$7E,$6C,$FC,$2C,$FC,$97,$EA,$1F,$68,$FB,$F3,$F1,$DF
	.dc.b	$81,$7F,$1E,$7D,$17,$EA,$BF,$C8,$D7,$53,$F0,$27,$CE,$BF,$13,$D4
	.dc.b	$FA,$67,$F8,$9A,$BC,$4F,$B2,$BA,$3F,$96,$3A,$5F,$D2,$7F,$71,$1F
	.dc.b	$10,$DB,$F7,$CD,$D7,$A5,$D9,$F4,$A3,$73,$F7,$A3,$DE,$7C,$23,$F2
	.dc.b	$13,$F3,$6F,$F6,$91,$8B,$FE,$0B,$5F,$7E,$FF,$82,$CF,$63,$F0,$B7
	.dc.b	$F5,$EB,$EF,$1C,$1F,$BD,$3E,$93,$ED,$C6,$4F,$B6,$E5,$E3,$7F,$A1
	.dc.b	$AF,$90,$7C,$D3,$D1,$7F,$93,$9E,$53,$E8,$8C,$3C,$67,$E7,$57,$82
	.dc.b	$31,$5F,$E9,$09,$FF,$23,$1F,$1E,$74,$FE,$8E,$FE,$DB,$F0,$EB,$89
	.dc.b	$F8,$EB,$F8,$91,$F8,$E3,$CC,$78,$2F,$D6,$F3,$3D,$35,$F1,$EB,$F1
	.dc.b	$75,$CA,$7E,$5D,$FD,$8A,$FA,$8F,$E7,$E3,$CE,$69,$F3,$BF,$70,$F0
	.dc.b	$57,$49,$FA,$31,$ED,$BF,$1B,$B7,$E1,$3F,$E1,$D7,$F3,$F5,$D2,$3F
	.dc.b	$32,$74,$3F,$2D,$7E,$73,$F4,$D7,$F0,$EF,$BC,$FE,$90,$FD,$5B,$F2
	.dc.b	$A7,$EA,$C7,$C3,$3A,$1F,$9B,$7A,$46,$CF,$8D,$1E,$07,$AC,$7E,$94
	.dc.b	$FE,$21,$1F,$0E,$FD,$AF,$A1,$F5,$98,$7F,$4E,$7D,$2B,$F9,$D1,$F8
	.dc.b	$A7,$E7,$47,$C3,$BF,$64,$FE,$5A,$3E,$54,$7C,$D8,$C4,$78,$63,$07
	.dc.b	$F8,$EE,$7F,$A7,$57,$C6,$9F,$E0,$87,$88,$F8,$2F,$ED,$5F,$44,$F8
	.dc.b	$0E,$55,$FC,$20,$F5,$BD,$AB,$FB,$DC,$FE,$54,$F9,$93,$A7,$E4,$3E
	.dc.b	$6A,$F9,$17,$FC,$BE,$7D,$66,$AF,$F5,$51,$F4,$23,$EA,$3F,$40,$FA
	.dc.b	$97,$F2,$27,$F5,$73,$FE,$BC,$BF,$E6,$6F,$CD,$9F,$C5,$3E,$F8,$F2
	.dc.b	$3D,$1F,$25,$74,$9D,$DF,$E0,$AB,$F8,$09,$F8,$A3,$A5,$7D,$B8,$D4
	.dc.b	$7C,$AA,$FD,$3D,$7F,$0B,$BF,$8C,$FF,$18,$AF,$C3,$9F,$4E,$BE,$41
	.dc.b	$FA,$39,$FF,$41,$DF,$F8,$90,$FF,$12,$C7,$F7,$BA,$FF,$4B,$47,$F8
	.dc.b	$A1,$FE,$67,$1F,$CA,$CF,$F5,$EC,$FF,$A9,$1F,$E1,$27,$FA,$5D,$FF
	.dc.b	$49,$1F,$EB,$13,$4F,$13,$E8,$14,$F0,$46,$DF,$F5,$09,$A3,$FC,$7E
	.dc.b	$68,$FF,$5B,$C7,$A8,$EE,$E2,$3D,$31,$98,$F3,$99,$8E,$CF,$FA,$A5
	.dc.b	$FF,$25,$CF,$C2,$8F,$DE,$9C,$0D,$3F,$1E,$7D,$93,$DA,$7C,$EF,$C6
	.dc.b	$7C,$E7,$23,$E2,$9E,$B9,$EC,$7E,$B1,$EC,$7C,$92,$9F,$DF,$E3,$B3
	.dc.b	$F6,$8F,$E0,$15,$FE,$C6,$3F,$D3,$91,$FE,$12,$71,$7F,$EE,$37,$E5
	.dc.b	$C9,$FF,$38,$1F,$A2,$9F,$CF,$47,$EB,$23,$8C,$73,$8F,$F8,$01,$1C
	.dc.b	$1F,$AA,$F2,$8F,$DA,$BF,$C8,$23,$F9,$83,$CA,$3C,$51,$CA,$6B,$E6
	.dc.b	$AF,$24,$F2,$9F,$DA,$57,$F0,$7A,$FD,$63,$CE,$BD,$51,$CD,$AF,$38
	.dc.b	$F6,$57,$EB,$AB,$F7,$92,$3C,$33,$F2,$26,$CF,$92,$7A,$4F,$2A,$FD
	.dc.b	$94,$F2,$9D,$D7,$C5,$7E,$64,$71,$6C,$FB,$27,$E9,$C7,$87,$CB,$FA
	.dc.b	$6F,$F9,$05,$A6,$1D,$1F,$A8,$0F,$08,$FF,$28,$3C,$23,$83,$FF,$00
	.dc.b	$1F,$D4,$BB,$8E,$31,$F8,$28,$EC,$E2,$3F,$CB,$4F,$33,$F3,$6F,$07
	.dc.b	$99,$D8,$F2,$47,$53,$83,$D4,$C9,$F9,$08,$F3,$B9,$3D,$0F,$81,$C1
	.dc.b	$FB,$38,$E6,$F8,$23,$4F,$88,$F2,$C6,$9F,$B8,$7A,$E3,$CC,$E8,$EA
	.dc.b	$73,$72,$F7,$32,$72,$8F,$94,$D3,$D5,$1F,$0D,$C1,$EA,$79,$BF,$8D
	.dc.b	$7D,$EF,$9C,$F8,$8F,$A9,$F3,$F3,$FD,$53,$F0,$67,$D4,$3C,$2F,$E5
	.dc.b	$DD,$3F,$BA,$74,$FE,$15,$CB,$FC,$A1,$CC,$77,$8E,$47,$AA,$3B,$BF
	.dc.b	$8F,$7B,$A7,$9B,$E6,$34,$77,$39,$1E,$33,$31,$DD,$E4,$68,$F2,$1A
	.dc.b	$79,$BF,$8E,$8F,$1B,$E4,$8E,$51,$E2,$8F,$49,$98,$E2,$FF,$39,$8E
	.dc.b	$2F,$8E,$38,$BE,$13,$D4,$71,$36,$7D,$77,$CE,$F9,$8D,$1F,$CA,$3C
	.dc.b	$8F,$D5,$1F,$B8,$3F,$52,$7C,$83,$2E,$CF,$DD,$39,$7F,$12,$F1,$72
	.dc.b	$7E,$F5,$E2,$FF,$60,$7A,$B9,$38,$9D,$4E,$51,$F3,$E3,$8C,$7E,$89
	.dc.b	$E3,$19,$8E,$27,$53,$C2,$E5,$FD,$FC,$79,$5E,$B1,$E6,$72,$78,$4F
	.dc.b	$73,$C5,$FE,$D5,$1C,$5C,$C7,$17,$27,$13,$C4,$F1,$7E,$94,$71,$73
	.dc.b	$1F,$C8,$F9,$FF,$8D,$1F,$CA,$0F,$E3,$27,$E6,$0C,$C7,$F5,$E7,$2F
	.dc.b	$F2,$97,$2F,$F8,$31,$CC,$7E,$5D,$CC,$7F,$0A,$73,$1E,$18,$FE,$03
	.dc.b	$1E,$17,$11,$C5,$C9,$E1,$32,$F1,$7F,$95,$C7,$17,$94,$71,$72,$71
	.dc.b	$32,$F1,$7F,$C2,$D1,$C5,$ED,$1C,$5C,$9C,$4C,$BC,$5F,$CF,$47,$17
	.dc.b	$E1,$1F,$14,$F4,$98,$3C,$27,$F5,$6F,$23,$FA,$51,$5F,$EA,$CF,$F0
	.dc.b	$83,$2F,$F8,$39,$FE,$9A,$65,$FE,$52,$F1,$79,$4F,$F0,$78,$E3,$1F
	.dc.b	$E4,$07,$8B,$FD,$85,$CB,$FC,$F9,$CC,$71,$8E,$A7,$84,$F2,$9C,$4C
	.dc.b	$9D,$4C,$BC,$5F,$EB,$F1,$E5,$79,$46,$5F,$F2,$34,$65,$E5,$19,$7F
	.dc.b	$85,$46,$5F,$31,$E1,$3A,$9E,$53,$E0,$C7,$99,$C9,$E5,$32,$7F,$81
	.dc.b	$BF,$80,$7F,$BD,$15,$FF,$03,$3F,$CE,$0C,$BF,$E3,$E7,$FB,$E9,$97
	.dc.b	$FA,$CB,$C5,$E5,$1F,$CD,$E3,$8C,$7F,$9C,$1E,$2F,$F8,$45,$CB,$FD
	.dc.b	$B9,$CC,$71,$8E,$A7,$84,$F2,$9C,$4C,$9D,$4C,$BC,$5F,$F0,$7C,$79
	.dc.b	$5E,$51,$97,$FC,$ED,$19,$79,$46,$5F,$E7,$51,$97,$CC,$78,$4E,$A7
	.dc.b	$94,$F8,$31,$E6,$72,$79,$4C,$9F,$45,$FF,$2D,$47,$F8,$5F,$46,$88
	.dc.b	$FE,$F9,$18,$53,$0F,$F2,$0D,$6D,$E6,$7F,$9D,$A3,$8D,$66,$3B,$3F
	.dc.b	$3E,$3C,$26,$CF,$F3,$DD,$77,$7B,$57,$F9,$B4,$FF,$2F,$3C,$AB,$C6
	.dc.b	$F4,$3B,$9E,$57,$CA,$ED,$F3,$47,$F9,$DE,$37,$1D,$A3,$FC,$BE,$76
	.dc.b	$7D,$6E,$DF,$2D,$7A,$DE,$B3,$EA,$9E,$8E,$8F,$80,$FA,$4F,$09,$FE
	.dc.b	$6C,$8F,$49,$F1,$CE,$07,$8E,$BE,$9C,$78,$CF,$F4,$74,$70,$7C,$87
	.dc.b	$FA,$62,$BC,$8F,$7A,$F8,$07,$C5,$9E,$A7,$D2,$8F,$97,$1E,$97,$FC
	.dc.b	$FE,$FE,$05,$E8,$FA,$9F,$8A,$7A,$AB,$FD,$1D,$1D,$63,$73,$E9,$AF
	.dc.b	$D6,$3F,$B6,$3F,$74,$7E,$F8,$FD,$91,$98,$FD,$B3,$93,$F7,$46,$5E
	.dc.b	$67,$E6,$5C,$BF,$AE,$7A,$B9,$7B,$39,$3A,$99,$3F,$44,$64,$EA,$79
	.dc.b	$9E,$AF,$ED,$9E,$AE,$5E,$AE,$4E,$A6,$4E,$A6,$4E,$A7,$B9,$EA,$FE
	.dc.b	$F9,$EA,$E5,$EA,$E4,$EA,$64,$EA,$64,$EA,$7D,$07,$D3,$5F,$59,$FC
	.dc.b	$74,$7D,$67,$F0,$4F,$FA,$4A,$3F,$D9,$11,$F8,$B3,$FD,$76,$FF,$BC
	.dc.b	$A3,$EE,$1F,$93,$3F,$35,$1F,$83,$9F,$CE,$C7,$E9,$E3,$A9,$CC,$ED
	.dc.b	$1F,$E7,$99,$E9,$1E,$2A,$FC,$BB,$E6,$AF,$D2,$3C,$9F,$F7,$83,$E5
	.dc.b	$7F,$66,$FF,$97,$DF,$D0,$3F,$B8,$7F,$36,$FE,$80,$FE,$2E,$78,$CE
	.dc.b	$4F,$37,$C4,$71,$9F,$E4,$0F,$18,$FD,$D3,$C5,$FF,$5A,$C7,$C5,$79
	.dc.b	$47,$F1,$03,$95,$7C,$48,$F1,$C7,$AD,$FD,$DB,$F1,$0F,$9F,$1F,$ED
	.dc.b	$87,$C7,$1A,$AF,$D7,$3F,$12,$3D,$6F,$C4,$3C,$73,$F3,$4E,$51,$E3
	.dc.b	$8E,$4F,$CD,$7E,$B3,$C9,$F1,$BF,$11,$E6,$FC,$D7,$F8,$0C,$FD,$67
	.dc.b	$F8,$89,$E3,$3E,$29,$E3,$3F,$8B,$9F,$58,$FC,$49,$F8,$78,$FE,$34
	.dc.b	$FD,$63,$F2,$8F,$8D,$FC,$6B,$E3,$79,$BF,$11,$FC,$8C,$7C,$47,$D6
	.dc.b	$78,$8F,$59,$EA,$3C,$51,$C6,$BF,$49,$1E,$A3,$89,$FC,$7A,$3E,$F3
	.dc.b	$EA,$3E,$E9,$F8,$E3,$E5,$9C,$4F,$BA,$FA,$63,$F5,$71,$E1,$38,$9F
	.dc.b	$2A,$3F,$0B,$3F,$D4,$9F,$C2,$47,$C0,$38,$1E,$97,$D1,$5F,$DC,$67
	.dc.b	$E0,$4F,$80,$F8,$07,$13,$D0,$71,$9F,$C1,$3F,$21,$FC,$13,$C4,$F8
	.dc.b	$07,$D9,$3F,$AB,$1F,$21,$E3,$5F,$22,$3E,$C1,$F8,$13,$89,$F4,$0F
	.dc.b	$09,$F6,$23,$8D,$7A,$5F,$0C,$FE,$B9,$F8,$2F,$C8,$9E,$27,$8A,$3E
	.dc.b	$09,$F8,$88,$FD,$AC,$7E,$20,$F8,$2F,$17,$D2,$7D,$07,$FB,$5B,$F9
	.dc.b	$18,$FE,$31,$5F,$45,$FB,$15,$FE,$09,$9F,$A2,$7A,$4F,$C0,$1F,$24
	.dc.b	$FC,$05,$7A,$67,$89,$FE,$0B,$8F,$09,$F4,$4E,$31,$F2,$4F,$4B,$F2
	.dc.b	$5F,$2C,$7E,$42,$38,$9E,$13,$F1,$07,$E4,$0F,$A2,$7E,$62,$38,$BF
	.dc.b	$65,$FE,$28,$F8,$4E,$27,$A4,$E2,$7E,$05,$E2,$FD,$93,$C2,$7D,$18
	.dc.b	$FB,$27,$D1,$9F,$E3,$93,$F0,$67,$E8,$C7,$C9,$7E,$CC,$7F,$3E,$9E
	.dc.b	$27,$D0,$3F,$96,$47,$E2,$0F,$CC,$1E,$08,$E0,$FC,$89,$FC,$43,$F4
	.dc.b	$1F,$F0,$FB,$C6,$3F,$00,$7C,$07,$E4,$3F,$80,$AF,$E3,$11,$F0,$23
	.dc.b	$D0,$F1,$9F,$43,$E1,$7F,$94,$BE,$1A,$F8,$15,$F8,$78,$F7,$C7,$DF
	.dc.b	$AF,$04,$7C,$73,$84,$7B,$DF,$CB,$1D,$DE,$15,$FC,$DD,$FE,$5D,$3F
	.dc.b	$4A,$7E,$1B,$D0,$F8,$6F,$18,$FE,$90,$FE,$D4,$EF,$5F,$C1,$27,$DA
	.dc.b	$73,$3E,$2D,$77,$9F,$E6,$75,$F8,$F9,$ED,$3F,$02,$7B,$4F,$F3,$17
	.dc.b	$D8,$7E,$2A,$7F,$35,$3F,$82,$3E,$9B,$DD,$F9,$6F,$49,$F6,$C2,$64
	.dc.b	$78,$63,$C8,$FF,$02,$78,$98,$7F,$13,$5E,$D8,$EE,$73,$9F,$ED,$53
	.dc.b	$F1,$1F,$E0,$11,$98,$EC,$7F,$6E,$3C,$71,$E2,$3F,$51,$1D,$DF,$97
	.dc.b	$1F,$C5,$A3,$BC,$6C,$FD,$35,$7B,$5F,$96,$72,$9E,$CE,$0D,$CF,$57
	.dc.b	$E5,$3F,$9C,$7F,$28,$F6,$7D,$71,$F4,$67,$F0,$27,$CD,$7F,$12,$76
	.dc.b	$3E,$41,$88,$FC,$21,$D9,$FC,$C9,$FB,$08,$E4,$FC,$78,$FE,$0D,$18
	.dc.b	$8F,$04,$FE,$0E,$31,$3F,$C9,$0F,$C1,$BF,$AD,$9F,$EB,$67,$59,$FA
	.dc.b	$13,$F0,$CF,$1C,$F2,$7F,$82,$C7,$53,$6F,$C6,$9F,$61,$F8,$03,$E9
	.dc.b	$57,$F0,$DA,$E5,$3F,$B7,$9F,$B6,$FC,$83,$CD,$1F,$22,$BA,$C7,$EB
	.dc.b	$1F,$A5,$1E,$98,$FE,$58,$FF,$34,$7F,$96,$47,$08,$FD,$4C,$F9,$27
	.dc.b	$F8,$0B,$F2,$5F,$E1,$F3,$B3,$F3,$31,$B8,$FA,$4F,$F0,$37,$FA,$04
	.dc.b	$6E,$3E,$C4,$7E,$60,$FC,$71,$E7,$79,$4F,$C0,$3F,$20,$FF,$64,$3A
	.dc.b	$47,$A0,$E3,$1E,$87,$E8,$CF,$79,$F4,$47,$DA,$8C,$47,$E0,$0F,$97
	.dc.b	$5F,$CA,$2B,$E2,$57,$F4,$7A,$DB,$F3,$9F,$B4,$F8,$1F,$82,$F8,$0C
	.dc.b	$57,$B9,$FC,$7C,$7F,$AC,$27,$F8,$5D,$7F,$2D,$9F,$E2,$27,$5A,$FB
	.dc.b	$A7,$EE,$9E,$87,$DF,$3F,$16,$FA,$0F,$C9,$BF,$CC,$27,$E4,$57,$E0
	.dc.b	$0F,$4C,$70,$3F,$6C,$7C,$F3,$AD,$7C,$18,$FE,$3C,$FE,$45,$E3,$1F
	.dc.b	$91,$38,$CF,$CF,$7E,$D1,$E5,$35,$5F,$5E,$BF,$8D,$1C,$E3,$E2,$C7
	.dc.b	$D4,$9F,$CC,$46,$23,$EA,$3D,$AB,$E6,$BD,$AB,$CC,$7C,$9A,$F7,$D7
	.dc.b	$E2,$2B,$CF,$1F,$88,$8F,$F0,$B9,$DE,$3E,$CD,$7F,$89,$6B,$CB,$3D
	.dc.b	$CF,$B1,$1F,$9E,$3F,$2A,$7F,$3A,$47,$F3,$37,$F5,$E5,$7F,$D1,$4D
	.dc.b	$34,$7F,$70,$39,$9F,$C2,$1E,$6E,$8A,$73,$6B,$D8,$E0,$ED,$C9,$E5
	.dc.b	$79,$19,$7F,$AF,$1E,$33,$A3,$A7,$93,$E8,$7B,$3F,$E3,$D7,$CE,$7F
	.dc.b	$49,$7B,$BE,$A7,$4E,$0D,$3D,$1F,$82,$E8,$F2,$1D,$DE,$4F,$43,$2F
	.dc.b	$43,$0E,$5E,$8F,$73,$DE,$65,$F6,$39,$7C,$2E,$5F,$19,$D5,$FD,$73
	.dc.b	$F9,$A5,$AF,$FB,$C8,$FF,$62,$9F,$EF,$23,$F0,$CF,$E2,$CF,$DB,$BB
	.dc.b	$8F,$F0,$05,$AF,$F0,$F3,$C6,$BF,$AC,$9C,$6B,$FC,$2B,$1C,$55,$71
	.dc.b	$9F,$1C,$7E,$3E,$31,$3E,$29,$FD,$39,$F8,$88,$FF,$36,$C7,$37,$73
	.dc.b	$C6,$BC,$75,$E2,$38,$9F,$EE,$C8,$FF,$0B,$D7,$19,$F4,$C7,$19,$F8
	.dc.b	$A7,$17,$F8,$C1,$EA,$3F,$DD,$B3,$F2,$1E,$0F,$9D,$DC,$79,$E7,$FC
	.dc.b	$31,$3D,$E3,$DF,$3F,$41,$EC,$FF,$19,$78,$3E,$63,$B9,$F4,$27,$CE
	.dc.b	$F1,$8F,$E0,$27,$B9,$D1,$C4,$FF,$78,$CF,$9D,$E0,$7F,$60,$8E,$0F
	.dc.b	$FB,$9E,$3C,$C7,$DB,$8E,$B1,$EC,$7E,$53,$E2,$AF,$B5,$16,$BC,$B3
	.dc.b	$E7,$8F,$96,$72,$7D,$CF,$23,$E1,$CF,$26,$CF,$29,$FC,$3B,$F7,$23
	.dc.b	$94,$7F,$32,$79,$57,$A2,$3E,$9C,$7C,$C7,$47,$C4,$3C,$CF,$F1,$E3
	.dc.b	$B9,$F5,$1F,$44,$7D,$B7,$CE,$71,$8D,$9C,$A3,$E3,$47,$CB,$3E,$31
	.dc.b	$B7,$C8,$FD,$03,$FB,$19,$D0,$FA,$07,$DE,$3D,$F1,$E3,$6C,$78,$DF
	.dc.b	$59,$C0,$FB,$A7,$A0,$F8,$31,$F1,$AB,$FA,$34,$78,$CE,$CF,$EA,$1F
	.dc.b	$23,$FD,$39,$FE,$E7,$1E,$E3,$E2,$CF,$BA,$3D,$27,$8A,$7F,$66,$F4
	.dc.b	$AF,$B1,$1F,$BA,$AF,$A4,$F8,$A7,$D0,$FD,$FA,$FC,$A4,$7D,$09,$DD
	.dc.b	$7F,$7D,$8F,$CD,$56,$DF,$71,$E8,$8F,$EC,$66,$DF,$BB,$3F,$55,$F8
	.dc.b	$67,$D5,$3E,$84,$FA,$1F,$92,$7C,$E3,$67,$B9,$F4,$1B,$3C,$4E,$CF
	.dc.b	$11,$FD,$F4,$FC,$09,$EE,$9F,$C7,$4F,$CA,$8E,$B3,$F3,$DF,$E6,$95
	.dc.b	$F8,$67,$E1,$4F,$D9,$AF,$63,$F4,$E7,$E3,$CF,$E1,$23,$E3,$9F,$E0
	.dc.b	$57,$E3,$CF,$82,$77,$5F,$4E,$BA,$1B,$AF,$8E,$F9,$5F,$59,$FE,$1F
	.dc.b	$8F,$83,$5F,$C6,$A3,$EF,$4F,$E5,$67,$A4,$7C,$D3,$F0,$93,$B7,$E6
	.dc.b	$C7,$E3,$DF,$69,$B7,$AB,$E8,$9F,$DA,$57,$A2,$BF,$80,$4E,$DF,$87
	.dc.b	$1F,$1E,$3E,$14,$FB,$1F,$CC,$C6,$6B,$CB,$1F,$4E,$7A,$BF,$8A,$3C
	.dc.b	$66,$5F,$C4,$3F,$09,$FE,$D2,$FF,$97,$8F,$33,$A7,$8C,$7E,$16,$BE
	.dc.b	$84,$75,$3C,$A7,$EC,$4F,$86,$FE,$98,$F6,$BF,$8F,$7D,$AF,$CE,$AF
	.dc.b	$D8,$4F,$5A,$F4,$CF,$E4,$4F,$0D,$7C,$F9,$FC,$A9,$F3,$E3,$C2,$75
	.dc.b	$9F,$C0,$1F,$8B,$7C,$D1,$F1,$8E,$D1,$ED,$8E,$31,$88,$FC,$34,$78
	.dc.b	$0F,$D0,$1F,$AC,$9F,$11,$F0,$9F,$31,$C9,$CC,$7A,$A7,$CC,$78,$63
	.dc.b	$31,$E1,$8C,$4F,$63,$F1,$A7,$D2,$8F,$1B,$D4,$C5,$7E,$26,$3F,$A8
	.dc.b	$C7,$D1,$9F,$8B,$5F,$46,$73,$3F,$14,$FC,$49,$F4,$4F,$DB,$C7,$92
	.dc.b	$3B,$CF,$92,$7F,$53,$5D,$CF,$E6,$D1,$F9,$C8,$C9,$F3,$AB,$EF,$C7
	.dc.b	$CB,$9E,$15,$FE,$00,$AD,$9F,$22,$7B,$47,$E4,$A3,$FC,$F0,$6A,$3E
	.dc.b	$4D,$74,$3C,$D5,$D2,$BF,$B2,$D7,$AC,$FC,$E1,$FE,$FB,$3F,$88,$99
	.dc.b	$3E,$E1,$D6,$7E,$34,$FF,$88,$63,$F3,$33,$FC,$06,$7C,$11,$F5,$6B
	.dc.b	$E0,$4F,$E3,$63,$D5,$3A,$7A,$47,$13,$E8,$CF,$C3,$8F,$8D,$18,$9F
	.dc.b	$E7,$AF,$E6,$2B,$F7,$91,$CE,$3B,$57,$C2,$78,$57,$E0,$8E,$75,$F5
	.dc.b	$2B,$F5,$B1,$EA,$8F,$D6,$C7,$C8,$8D,$CF,$D3,$3A,$C7,$C2,$7F,$97
	.dc.b	$C7,$E1,$8E,$47,$E1,$8F,$01,$F0,$1F,$1B,$F0,$27,$BB,$F1,$E7,$C7
	.dc.b	$5F,$94,$AF,$89,$5F,$3E,$30,$F0,$AE,$47,$73,$F9,$B5,$72,$3F,$61
	.dc.b	$3F,$0E,$7E,$C4,$FF,$AA,$5F,$F4,$B9,$FE,$A4,$3F,$D0,$C7,$FA,$2A
	.dc.b	$BF,$C2,$07,$F8,$46,$3F,$CC,$EF,$FA,$4E,$34,$72,$8F,$F5,$9B,$C9
	.dc.b	$FF,$58,$B9,$7B,$C7,$17,$27,$FB,$2E,$3C,$51,$98,$1E,$73,$47,$67
	.dc.b	$4E,$4F,$E9,$C7,$37,$2F,$F1,$13,$FC,$C4,$65,$F8,$0F,$88,$E2,$7F
	.dc.b	$9D,$9E,$27,$F2,$E7,$89,$97,$8B,$E6,$3B,$1F,$E5,$73,$E3,$BA,$7C
	.dc.b	$C7,$23,$27,$C8,$3F,$60,$7F,$19,$3F,$DE,$33,$D4,$FF,$32,$16,$B3
	.dc.b	$1F,$E9,$78,$FC,$34,$7E,$00,$FC,$B4,$FF,$9B,$63,$83,$FE,$5F,$69
	.dc.b	$FE,$F8,$3F,$D6,$47,$27,$F9,$A4,$72,$9F,$F3,$04,$F2,$9F,$0C,$7F
	.dc.b	$95,$A3,$94,$FA,$27,$F5,$15,$FB,$67,$F4,$27,$FB,$02,$BF,$D0,$8F
	.dc.b	$2A,$F0,$C7,$AE,$38,$BF,$12,$3F,$90,$9E,$18,$E5,$1E,$13,$83,$FC
	.dc.b	$79,$FD,$99,$FD,$E6,$38,$1F,$01,$D9,$FE,$7B,$FC,$8F,$A1,$F1,$8F
	.dc.b	$5C,$9F,$F4,$34,$6C,$F4,$C6,$5F,$F4,$F4,$70,$36,$FF,$A6,$0F,$CD
	.dc.b	$C3,$F2,$3F,$51,$F5,$0E,$E7,$0A,$FF,$76,$4F,$0A,$F2,$BD,$8F,$F6
	.dc.b	$B3,$BC,$3F,$25,$7E,$39,$FF,$44,$A3,$15,$FE,$EF,$AF,$34,$7F,$1D
	.dc.b	$7A,$CF,$FA,$41,$CC,$7F,$AD,$6F,$CF,$3F,$EB,$B7,$CF,$1C,$63,$E1
	.dc.b	$CF,$C9,$7C,$B1,$97,$C6,$F8,$AF,$C9,$3D,$A7,$E6,$1A,$79,$3F,$4E
	.dc.b	$7F,$99,$57,$8A,$3F,$D4,$AF,$8A,$3F,$D9,$53,$EA,$BF,$75,$78,$A7
	.dc.b	$2F,$ED,$CF,$F3,$9C,$D7,$C6,$53,$CB,$1F,$25,$C3,$F3,$5F,$B6,$F0
	.dc.b	$7E,$8C,$71,$78,$C5,$38,$3E,$03,$C4,$E1,$FB,$CF,$77,$C8,$F0,$3F
	.dc.b	$DA,$D1,$83,$A1,$F8,$E8,$E6,$FF,$39,$34,$FA,$1D,$3E,$63,$C3,$1E
	.dc.b	$97,$F1,$C6,$DF,$13,$D0,$DB,$E3,$8F,$0B,$A8,$E8,$7E,$49,$E8,$C3
	.dc.b	$DA,$FA,$1E,$71,$E6,$7D,$91,$CD,$F1,$9C,$DF,$03,$E5,$79,$39,$8E
	.dc.b	$4F,$C2,$8E,$4F,$89,$F9,$2F,$AC,$EE,$F2,$95,$FD,$59,$FD,$11,$F8
	.dc.b	$93,$FE,$00,$6F,$FC,$44,$FE,$56,$7F,$06,$3F,$64,$FA,$27,$9D,$78
	.dc.b	$23,$99,$F9,$B7,$F9,$D3,$FC,$12,$7A,$3F,$D8,$27,$8C,$75,$9F,$D5
	.dc.b	$47,$47,$F2,$0F,$E7,$A7,$A1,$E6,$3F,$8A,$CF,$99,$FD,$04,$74,$9E
	.dc.b	$75,$EF,$8E,$53,$FC,$6A,$7F,$C4,$15,$E4,$84,$F9,$1E,$75,$FC,$36
	.dc.b	$7E,$0C,$79,$23,$A1,$F8,$29,$FE,$21,$5D,$27,$CA,$7B,$E7,$93,$D2
	.dc.b	$15,$D1,$FE,$10,$7C,$BA,$F3,$CF,$F1,$78,$E8,$FF,$0E,$9F,$80,$FD
	.dc.b	$68,$F7,$9C,$6B,$C8,$7E,$DA,$3E,$41,$C4,$F2,$47,$89,$F2,$4F,$8E
	.dc.b	$FF,$87,$57,$2A,$FE,$8C,$78,$AF,$EB,$3C,$6F,$E4,$CF,$B9,$F2,$47
	.dc.b	$F1,$47,$F9,$01,$FC,$F4,$FB,$67,$F4,$93,$84,$F4,$BF,$6C,$79,$9F
	.dc.b	$23,$FC,$C8,$E4,$7D,$B9,$FC,$7B,$CA,$BF,$A2,$45,$3C,$D1,$CD,$F8
	.dc.b	$E7,$DB,$3F,$8C,$D7,$DA,$79,$9E,$D7,$C9,$1E,$33,$E3,$C7,$23,$EC
	.dc.b	$DF,$F3,$F8,$F6,$47,$1A,$F1,$B4,$F1,$C7,$C8,$7F,$06,$73,$9F,$E8
	.dc.b	$B1,$ED,$3E,$CD,$78,$9E,$17,$FA,$9A,$F9,$D3,$F5,$2B,$8D,$7E,$26
	.dc.b	$7D,$75,$F9,$FA,$FC,$F4,$FD,$63,$FA,$AB,$C5,$F9,$A7,$13,$C3,$7E
	.dc.b	$A9,$FA,$B1,$E2,$AF,$43,$C2,$BE,$53,$C2,$3E,$9C,$F0,$9F,$DB,$47
	.dc.b	$A2,$BF,$56,$FD,$27,$F3,$15,$F2,$AB,$D1,$1F,$8D,$8F,$B5,$3C,$4F
	.dc.b	$D0,$1C,$4F,$CB,$9F,$01,$FD,$99,$F2,$0F,$B2,$F1,$3F,$05,$7F,$66
	.dc.b	$FF,$C2,$B3,$F8,$99,$FC,$55,$FC,$F8,$F8,$F7,$E8,$BE,$0F,$F8,$5A
	.dc.b	$7C,$F5,$F7,$AF,$DF,$7C,$23,$CF,$7C,$23,$F9,$31,$E6,$9F,$AA,$FB
	.dc.b	$8E,$13,$E6,$7F,$9C,$1F,$14,$E1,$7F,$15,$FA,$91,$ED,$3F,$2A,$F0
	.dc.b	$9F,$6C,$7E,$4C,$F6,$C7,$C4,$3B,$BC,$1F,$DB,$C7,$0B,$F6,$9F,$0E
	.dc.b	$3F,$3F,$7F,$C2,$2B,$E9,$3F,$C7,$67,$73,$D4,$D9,$FB,$B3,$FA,$49
	.dc.b	$F9,$F3,$D8,$7F,$17,$3E,$C1,$F9,$73,$89,$FC,$30,$E2,$7E,$5C,$E3
	.dc.b	$5F,$A7,$3E,$5B,$F1,$0F,$6C,$FE,$9A,$3F,$2F,$1C,$2F,$F3,$CF,$EF
	.dc.b	$CF,$04,$7C,$89,$F7,$1F,$66,$3F,$0D,$7F,$BD,$BF,$89,$5B,$9F,$73
	.dc.b	$FC,$F0,$EE,$7C,$B7,$EC,$CF,$E1,$EF,$FA,$83,$FC,$BD,$F1,$C7,$C2
	.dc.b	$3F,$08,$79,$E3,$F5,$AF,$9C,$CC,$F8,$5F,$8C,$F8,$67,$F5,$AF,$86
	.dc.b	$3E,$73,$F1,$8F,$CB,$BD,$67,$F0,$AF,$C2,$9F,$CF,$3E,$8A,$F7,$C7
	.dc.b	$F8,$9A,$BF,$85,$4F,$D4,$7D,$F1,$C2,$3B,$BF,$97,$8E,$E7,$8C,$FB
	.dc.b	$0E,$4F,$87,$5F,$BB,$9E,$2F,$C7,$3F,$94,$9F,$9F,$BF,$B9,$3E,$E3
	.dc.b	$C1,$5D,$27,$85,$FE,$56,$FD,$47,$E5,$67,$C5,$5F,$41,$F8,$8F,$8A
	.dc.b	$3F,$12,$7F,$1F,$7B,$3F,$BA,$3C,$B3,$F1,$1C,$4F,$86,$31,$1F,$88
	.dc.b	$7F,$92,$4F,$58,$FC,$EC,$FC,$A3,$EC,$57,$F1,$C8,$FE,$58,$FD,$47
	.dc.b	$EF,$CF,$B8,$F4,$1F,$26,$3E,$2C,$70,$7F,$2C,$FB,$E3,$EC,$1F,$C6
	.dc.b	$DF,$E0,$CF,$38,$FC,$6C,$7E,$44,$FD,$51,$F8,$57,$9B,$F8,$53,$DF
	.dc.b	$1E,$28,$F9,$E7,$F4,$68,$FC,$24,$7C,$87,$E1,$1F,$99,$3A,$CF,$D2
	.dc.b	$9F,$0B,$F4,$9F,$5C,$78,$6B,$E4,$C6,$27,$F4,$07,$C9,$AF,$F0,$D5
	.dc.b	$FD,$FB,$FA,$A7,$C1,$3C,$F5,$DA,$7E,$45,$EE,$FE,$C5,$6C,$FF,$04
	.dc.b	$57,$F3,$37,$F9,$2D,$7C,$78,$FA,$91,$EA,$8C,$47,$C6,$8F,$AE,$78
	.dc.b	$1D,$CF,$E8,$A7,$E9,$C7,$F4,$69,$FA,$66,$DE,$4F,$53,$15,$F0,$8F
	.dc.b	$44,$FD,$17,$F1,$F3,$FE,$4E,$3D,$13,$FC,$BE,$BE,$74,$78,$AB,$F2
	.dc.b	$51,$88,$EE,$F9,$67,$B9,$F5,$DC,$3F,$12,$3F,$A4,$3D,$67,$E0,$D7
	.dc.b	$96,$7F,$8E,$1F,$3C,$FC,$29,$FB,$DA,$F9,$F5,$F8,$B3,$A5,$FE,$F2
	.dc.b	$7F,$45,$1F,$81,$3C,$57,$E8,$3B,$B8,$8F,$C2,$46,$23,$F3,$55,$E8
	.dc.b	$9F,$88,$7A,$23,$FA,$54,$7C,$1B,$F4,$4F,$96,$BE,$65,$7E,$0E,$3F
	.dc.b	$B5,$1F,$1A,$3C,$51,$FE,$A7,$7F,$58,$FE,$18,$F9,$27,$89,$DB,$FA
	.dc.b	$23,$11,$E8,$77,$5F,$A1,$AF,$F3,$1C,$7F,$3C,$AF,$CE,$D7,$CC,$3E
	.dc.b	$D1,$C6,$3A,$57,$18,$D1,$E7,$34,$7C,$2B,$F9,$C7,$E2,$9F,$AD,$7F
	.dc.b	$8A,$3F,$BD,$D7,$D5,$8F,$83,$1F,$1A,$3F,$92,$9F,$BE,$8F,$AF,$5F
	.dc.b	$2E,$3E,$A4,$FE,$CD,$1F,$D7,$8F,$E5,$46,$0F,$F6,$54,$FF,$B8,$0F
	.dc.b	$F3,$49,$FC,$51,$D9,$FE,$31,$75,$1C,$63,$AC,$6D,$FF,$13,$1D,$1D
	.dc.b	$3F,$D5,$4D,$BF,$E2,$B3,$FC,$EE,$EC,$D9,$97,$51,$88,$FE,$74,$78
	.dc.b	$63,$CE,$69,$E2,$FF,$1A,$75,$1E,$28,$F3,$1E,$08,$E6,$79,$4F,$8C
	.dc.b	$7A,$8C,$9F,$09,$E3,$1E,$37,$B4,$71,$72,$F8,$A3,$FD,$4E,$D3,$D1
	.dc.b	$1E,$77,$C9,$1A,$69,$A8,$FA,$71,$DC,$D4,$7B,$0F,$ED,$A7,$E1,$FF
	.dc.b	$24,$FB,$95,$FE,$CA,$6E,$0B,$47,$D4,$94,$E6,$3F,$29,$3F,$9E,$BF
	.dc.b	$F7,$54,$72,$8F,$F6,$0B,$4E,$53,$C6,$3F,$43,$3F,$97,$BF,$F5,$BD
	.dc.b	$71,$9E,$D1,$FE,$E6,$8E,$31,$E4,$38,$C7,$E6,$9F,$43,$FE,$FF,$3F
	.dc.b	$1F,$1C,$E3,$D7,$1C,$CF,$CA,$5F,$B2,$3A,$1F,$C3,$4F,$D7,$C7,$F0
	.dc.b	$D9,$E4,$7A,$DE,$37,$EB,$9E,$07,$C3,$3C,$71,$EC,$3F,$DF,$EF,$B1
	.dc.b	$FF,$68,$4F,$B0,$F9,$71,$FC,$0E,$FE,$5C,$F0,$8F,$4B,$C0,$F0,$BE
	.dc.b	$87,$FC,$19,$1F,$2E,$FF,$D1,$B5,$E8,$9F,$80,$7D,$3A,$F9,$2F,$F1
	.dc.b	$EA,$EE,$FF,$B8,$E3,$CD,$5F,$16,$BC,$D1,$C2,$3F,$8D,$3F,$EF,$68
	.dc.b	$ED,$3E,$53,$B4,$7C,$43,$D8,$7C,$49,$F2,$C7,$B6,$7E,$63,$C0,$F5
	.dc.b	$9F,$76,$7B,$3B,$3D,$67,$E1,$5F,$25,$FF,$82,$1E,$CF,$F1,$97,$E4
	.dc.b	$D7,$E7,$E7,$89,$F4,$A3,$89,$FD,$89,$FC,$51,$F0,$A7,$F4,$91,$4F
	.dc.b	$A7,$1E,$17,$DD,$7F,$46,$7C,$17,$F1,$2F,$E5,$5F,$D8,$8F,$BF,$5B
	.dc.b	$8F,$CB,$9E,$D3,$E5,$1D,$4F,$86,$7D,$7B,$F6,$BC,$0F,$B0,$FB,$AF
	.dc.b	$F1,$D1,$C4,$FE,$EC,$71,$8F,$96,$F0,$AF,$A2,$EE,$FF,$A3,$3B,$3E
	.dc.b	$1C,$F5,$3F,$7A,$FC,$A8,$FC,$34,$7C,$AB,$FF,$0F,$D7,$C1,$BF,$1D
	.dc.b	$FC,$FA,$F8,$EF,$DD,$3A,$CE,$E3,$E6,$C6,$CF,$83,$7F,$90,$76,$7C
	.dc.b	$77,$E6,$BF,$89,$3D,$EE,$DF,$D9,$1E,$F3,$D1,$5F,$B9,$7A,$BF,$89
	.dc.b	$9F,$8B,$3F,$87,$9F,$A6,$78,$CE,$B3,$B7,$C6,$75,$37,$1F,$12,$7D
	.dc.b	$F1,$F2,$CF,$1C,$75,$3D,$2F,$D1,$9F,$F1,$34,$7A,$4F,$D3,$C7,$57
	.dc.b	$6F,$8D,$FA,$E6,$E3,$F2,$31,$F1,$5F,$A3,$1E,$38,$F7,$CF,$E7,$DF
	.dc.b	$1C,$FE,$7A,$3E,$F5,$7F,$00,$7D,$35,$D6,$BF,$20,$FA,$6B,$A9,$F2
	.dc.b	$E3,$E2,$CF,$59,$F4,$9B,$7E,$EB,$E3,$8D,$DF,$E2,$1F,$44,$EA,$BA
	.dc.b	$57,$B9,$FA,$13,$E8,$9F,$C9,$BF,$42,$36,$7C,$A3,$73,$E2,$7D,$D7
	.dc.b	$FA,$09,$F9,$51,$BA,$F4,$3F,$2A,$3E,$83,$EE,$8F,$E6,$11,$E8,$3F
	.dc.b	$A3,$1E,$89,$F9,$0F,$E1,$2B,$E0,$46,$23,$D1,$3F,$C2,$E3,$13,$ED
	.dc.b	$8F,$0C,$FC,$EA,$F3,$BF,$98,$8F,$0C,$FA,$DF,$82,$FE,$10,$F9,$86
	.dc.b	$5E,$D1,$FC,$45,$FD,$B3,$E8,$3D,$AF,$CF,$3A,$CF,$C1,$9F,$4C,$F1
	.dc.b	$9F,$D2,$CF,$92,$3E,$84,$70,$7F,$CC,$F1,$F4,$0E,$46,$E7,$CC,$FD
	.dc.b	$68,$F7,$4F,$AA,$3E,$75,$7F,$54,$AF,$0D,$7E,$F5,$F8,$D7,$F8,$78
	.dc.b	$F4,$CF,$CA,$AE,$F7,$CC,$F6,$47,$38,$F1,$1D,$A3,$E1,$CF,$F8,$94
	.dc.b	$FB,$47,$BC,$FB,$51,$C2,$39,$9F,$CC,$A3,$FA,$23,$F3,$9F,$34,$7E
	.dc.b	$2E,$BC,$D7,$EA,$9F,$DF,$1F,$02,$B8,$38,$39,$BF,$B2,$7C,$47,$E3
	.dc.b	$CF,$31,$F0,$0C,$C7,$DF,$8E,$11,$E2,$3B,$38,$7B,$3F,$E8,$87,$F9
	.dc.b	$91,$F9,$47,$4F,$E9,$0E,$37,$FA,$79,$FC,$6D,$FC,$EA,$F4,$D6,$CF
	.dc.b	$F4,$FC,$7E,$05,$E2,$7E,$66,$BE,$94,$FD,$57,$EC,$46,$8F,$C5,$BF
	.dc.b	$D9,$9F,$2D,$7B,$63,$C5,$5F,$89,$9F,$8F,$18,$AF,$C9,$46,$0F,$03
	.dc.b	$F7,$9F,$F8,$02,$64,$EE,$FD,$B3,$FA,$F1,$EC,$AF,$7C,$79,$5F,$CC
	.dc.b	$C7,$63,$E5,$47,$F3,$D8,$F3,$3E,$08,$FD,$39,$EF,$7E,$BD,$7C,$98
	.dc.b	$F9,$F3,$DA,$7E,$4C,$7F,$02,$7C,$D1,$E0,$AF,$93,$3E,$5A,$F8,$4F
	.dc.b	$A0,$E6,$E2,$FF,$D8,$F7,$F0,$27,$55,$F1,$CE,$6F,$03,$D8,$FF,$73
	.dc.b	$3E,$AC,$79,$27,$F3,$75,$FA,$39,$CD,$7F,$78,$3F,$99,$D7,$F8,$54
	.dc.b	$FF,$59,$47,$FA,$51,$FE,$58,$7F,$AF,$A3,$84,$7F,$A9,$CD,$47,$13
	.dc.b	$F8,$71,$4F,$F3,$EC,$7F,$B4,$E3,$FD,$70,$FF,$83,$8F,$F0,$4C,$69
	.dc.b	$E8,$FF,$A9,$E3,$FD,$7C,$77,$8E,$51,$93,$CE,$7F,$B1,$A3,$C6,$F5
	.dc.b	$8E,$D1,$C5,$EB,$1F,$E4,$F7,$2F,$17,$AB,$97,$DE,$E5,$E2,$FC,$38
	.dc.b	$F2,$B9,$3C,$47,$94,$F9,$27,$F9,$3A,$3A,$BC,$4F,$49,$EE,$7C,$37
	.dc.b	$FC,$46,$3F,$6D,$0F,$C9,$7F,$D4,$C7,$FA,$E4,$FF,$57,$9F,$84,$AF
	.dc.b	$C3,$47,$E9,$63,$72,$15,$FE,$AB,$8E,$33,$FA,$BA,$FC,$55,$7E,$80
	.dc.b	$E1,$1E,$17,$FD,$3D,$3E,$17,$FD,$3A,$F8,$5F,$D5,$1F,$E5,$27,$C2
	.dc.b	$F8,$2F,$FD,$B1,$1E,$09,$FD,$79,$FA,$53,$F6,$CF,$0B,$F3,$57,$69
	.dc.b	$F3,$3F,$D0,$0F,$E5,$07,$43,$CA,$7B,$E3,$8B,$F1,$E3,$F8,$47,$0F
	.dc.b	$E4,$93,$F1,$63,$FD,$24,$D6,$46,$8F,$F6,$3B,$88,$E4,$6D,$FF,$7A
	.dc.b	$35,$F9,$2F,$E5,$23,$BC,$74,$76,$F3,$38,$3D,$CE,$4E,$8F,$CA,$47
	.dc.b	$78,$E7,$16,$3B,$1E,$47,$A1,$98,$F4,$BE,$E3,$C4,$F2,$7C,$E7,$8A
	.dc.b	$39,$BA,$7B,$39,$8E,$6F,$03,$CA,$78,$0F,$8C,$E9,$F5,$1F,$C2,$E3
	.dc.b	$9B,$E5,$8F,$79,$E3,$7F,$CB,$D1,$A7,$9C,$6A,$3D,$D1,$EB,$39,$1E
	.dc.b	$93,$D0,$78,$CF,$BB,$1E,$31,$C3,$F8,$33,$E6,$1E,$23,$EC,$19,$8F
	.dc.b	$C1,$B9,$8E,$4E,$63,$FD,$B4,$E6,$39,$47,$CE,$8E,$AF,$DE,$7A,$B9
	.dc.b	$3A,$99,$8E,$AE,$4E,$A6,$5E,$AF,$F1,$78,$EA,$FA,$A3,$8B,$93,$A9
	.dc.b	$97,$8B,$FC,$F6,$38,$BE,$E8,$E2,$E4,$E2,$65,$E2,$FF,$A4,$E3,$8B
	.dc.b	$F0,$A3,$8B,$93,$89,$97,$EF,$F0,$FE,$CC,$FC,$E1,$87,$EB,$39,$8F
	.dc.b	$D9,$B9,$8E,$31,$F9,$A8,$E2,$FE,$A9,$FD,$81,$C5,$FD,$F4,$23,$94
	.dc.b	$6C,$E4,$76,$39,$47,$89,$C9,$CA,$38,$BE,$03,$FA,$43,$5F,$3C,$71
	.dc.b	$72,$75,$3D,$EE,$5F,$ED,$D1,$C5,$ED,$19,$7F,$D8,$11,$97,$F9,$AB
	.dc.b	$EE,$7C,$A6,$0F,$4C,$7C,$67,$C2,$7A,$4C,$9F,$C4,$3F,$7C,$FE,$45
	.dc.b	$83,$2F,$F6,$74,$61,$52,$3F,$90,$4E,$E3,$07,$F8,$D6,$38,$3C,$0E
	.dc.b	$15,$FC,$D6,$B9,$CF,$26,$CF,$F8,$F7,$F7,$CE,$2F,$81,$E6,$6D,$3D
	.dc.b	$6B,$A1,$FD,$62,$B8,$CF,$82,$3F,$A9,$35,$F1,$3E,$83,$DC,$F1,$32
	.dc.b	$F9,$9D,$C7,$84,$F0,$3F,$D7,$4E,$07,$68,$E0,$F8,$23,$84,$7B,$63
	.dc.b	$AB,$EA,$7C,$73,$ED,$3A,$DF,$D3,$8E,$AF,$CF,$9E,$AF,$D1,$AE,$B1
	.dc.b	$EA,$9C,$DF,$DA,$7D,$27,$F5,$DA,$E1,$3F,$5E,$3C,$75,$F6,$0E,$6F
	.dc.b	$CC,$3E,$91,$E8,$BF,$A6,$F3,$BF,$AD,$5C,$E3,$E1,$3F,$46,$FC,$D1
	.dc.b	$B8,$FA,$0F,$57,$EA,$3E,$73,$F6,$F7,$FD,$10,$FE,$D4,$7F,$79,$3F
	.dc.b	$B1,$19,$8F,$E7,$8E,$63,$FB,$CB,$98,$E5,$1F,$D8,$23,$8B,$FA,$18
	.dc.b	$E2,$E4,$E2,$65,$E2,$FF,$91,$63,$8B,$E2,$8E,$2E,$4E,$26,$5E,$2F
	.dc.b	$F9,$AE,$38,$BE,$68,$E2,$E4,$E2,$65,$E2,$FC,$53,$F1,$E6,$4F,$DD
	.dc.b	$1E,$93,$E4,$9E,$93,$27,$94,$C9,$E5,$33,$3F,$8D,$8F,$B1,$5F,$9A
	.dc.b	$9F,$D0,$C7,$FA,$C7,$F7,$DF,$D6,$9F,$BE,$3F,$68,$7E,$62,$7A,$D7
	.dc.b	$3B,$FC,$65,$FE,$05,$D3,$FE,$4E,$AE,$6F,$ED,$1E,$77,$F8,$BA,$E5
	.dc.b	$3F,$9A,$3F,$2D,$1E,$37,$F5,$73,$E3,$7F,$1C,$F6,$7F,$47,$1F,$C4
	.dc.b	$EB,$C7,$1C,$A3,$F5,$02,$7C,$71,$FC,$21,$F1,$C6,$A7,$F8,$69,$CA
	.dc.b	$7E,$2C,$72,$9E,$35,$FC,$95,$E3,$3F,$9F,$AE,$2F,$AC,$E2,$7E,$8E
	.dc.b	$38,$DF,$F3,$07,$8C,$7F,$0F,$3D,$77,$EC,$AF,$DB,$C6,$63,$D4,$FC
	.dc.b	$43,$E6,$1F,$11,$F9,$95,$FA,$FB,$FD,$AC,$FC,$33,$C4,$71,$3E,$11
	.dc.b	$C6,$7E,$13,$FC,$20,$F5,$1F,$C7,$DF,$A8,$FF,$02,$BF,$A8,$FE,$1E
	.dc.b	$3C,$27,$D3,$78,$CF,$F1,$48,$F9,$51,$E1,$8F,$94,$F8,$4F,$94,$78
	.dc.b	$4F,$E5,$C7,$E3,$63,$C3,$1F,$9B,$9F,$0C,$FE,$36,$7C,$27,$C9,$78
	.dc.b	$3E,$07,$D0,$7C,$83,$D0,$FC,$07,$F8,$ED,$FE,$05,$F8,$11,$C6,$BC
	.dc.b	$11,$C4,$F0,$47,$18,$FC,$DC,$7A,$0F,$A0,$7D,$9A,$FC,$0C,$FD,$A8
	.dc.b	$F0,$9F,$00,$F0,$9F,$22,$FF,$B4,$C7,$85,$FE,$E0,$F1,$AF,$93,$3C
	.dc.b	$8F,$C3,$5F,$E4,$9F,$11,$F9,$23,$C4,$7D,$1A,$FC,$93,$F4,$4E,$27
	.dc.b	$D1,$9F,$C4,$BC,$5F,$EB,$0F,$03,$E5,$9F,$8A,$3C,$4E,$AB,$F8,$79
	.dc.b	$F8,$38,$F1,$9F,$30,$FC,$99,$CA,$3D,$6F,$CC,$3E,$23,$F3,$0F,$1B
	.dc.b	$C4,$F5,$9C,$67,$C8,$F1,$3D,$67,$CC,$3C,$31,$F9,$37,$C2,$FD,$C7
	.dc.b	$D4,$71,$7D,$55,$F2,$DF,$09,$C6,$3C,$2F,$13,$D2,$FF,$0D,$AF,$C9
	.dc.b	$9F,$0A,$7F,$77,$3F,$16,$FF,$06,$7F,$02,$7B,$D7,$D1,$BF,$E6,$CF
	.dc.b	$BE,$38,$3F,$66,$FF,$14,$F9,$E7,$F9,$A9,$FC,$04,$E3,$3E,$73,$8C
	.dc.b	$7C,$79,$E2,$F9,$DF,$CD,$4F,$17,$FB,$CC,$F1,$8F,$F3,$7D,$7A,$2B
	.dc.b	$F1,$13,$E8,$7E,$73,$F1,$A7,$F3,$31,$F3,$9F,$CC,$CF,$EE,$23,$C0
	.dc.b	$7D,$E8,$F0,$5F,$C6,$7B,$C7,$BE,$7E,$34,$7B,$AF,$F1,$97,$DD,$FF
	.dc.b	$08,$D7,$D4,$9E,$EF,$ED,$CE,$E7,$E1,$27,$F1,$91,$F1,$8F,$3C,$71
	.dc.b	$9F,$F2,$F5,$79,$EB,$FA,$04,$7D,$CB,$FE,$0D,$1F,$16,$BF,$B0,$1E
	.dc.b	$73,$E2,$DF,$9E,$BF,$8A,$47,$F0,$49,$FA,$67,$E2,$67,$F3,$31,$E6
	.dc.b	$8F,$98,$79,$A7,$E9,$CF,$9A,$38,$BE,$E9,$FC,$C9,$C5,$F0,$4F,$E4
	.dc.b	$63,$F9,$19,$F4,$DF,$6B,$F1,$67,$DB,$1F,$82,$3E,$6C,$FC,$57,$E6
	.dc.b	$C7,$81,$E3,$3E,$08,$E2,$FD,$B8,$E2,$7A,$0E,$33,$F7,$23,$F2,$57
	.dc.b	$F7,$1F,$F3,$1C,$7E,$09,$F3,$9F,$4E,$7F,$88,$57,$EB,$8E,$F1,$F6
	.dc.b	$E7,$83,$F3,$5E,$D1,$F3,$67,$F3,$EF,$03,$E6,$C7,$09,$FE,$3C,$F0
	.dc.b	$3E,$64,$77,$7F,$BC,$4F,$E0,$67,$F9,$B3,$EC,$7F,$02,$7B,$0F,$C0
	.dc.b	$9E,$C7,$E1,$DF,$E7,$63,$75,$E5,$BF,$85,$1F,$E0,$2A,$FB,$0F,$AC
	.dc.b	$DD,$F9,$0F,$C3,$D7,$AD,$FB,$F7,$F9,$47,$F0,$C7,$AE,$BE,$BC,$FA
	.dc.b	$DF,$F1,$21,$EB,$AF,$E4,$B3,$C1,$EF,$1C,$23,$B9,$F0,$9F,$C3,$1F
	.dc.b	$9C,$8F,$92,$70,$7E,$BC,$70,$7B,$9C,$0F,$6C,$70,$36,$F7,$3E,$BC
	.dc.b	$77,$8F,$C6,$47,$77,$F9,$24,$7B,$63,$8D,$FC,$A7,$FA,$89,$BA,$FD
	.dc.b	$B5,$76,$8D,$BF,$58,$FA,$17,$F2,$63,$EE,$46,$DF,$EE,$17,$EB,$BF
	.dc.b	$F2,$94,$76,$3D,$73,$E4,$3E,$13,$C0,$F5,$9F,$09,$F5,$BC,$2F,$CA
	.dc.b	$EC,$F2,$C7,$F6,$CA,$FB,$AF,$E4,$2F,$C3,$7F,$CC,$E3,$47,$E3,$EB
	.dc.b	$E0,$57,$DB,$8F,$54,$FF,$1B,$8F,$99,$1F,$A9,$9F,$CA,$BF,$B0,$BF
	.dc.b	$DF,$47,$8A,$3F,$A7,$1D,$5F,$44,$F5,$7D,$07,$BE,$BE,$95,$FF,$69
	.dc.b	$9F,$E1,$4F,$A2,$BF,$93,$1E,$83,$E8,$9E,$DA,$FF,$1F,$CF,$E7,$DF
	.dc.b	$93,$5F,$42,$BA,$47,$F0,$D7,$A5,$7E,$C4,$E9,$1F,$56,$3E,$E1,$AB
	.dc.b	$FD,$84,$FE,$06,$74,$F8,$5F,$F1,$81,$F9,$3A,$F5,$C7,$F7,$58,$F0
	.dc.b	$1E,$63,$4F,$3A,$F2,$9F,$BF,$7E,$89,$FC,$BE,$FE,$8D,$FF,$12,$AF
	.dc.b	$A3,$3F,$BF,$36,$7F,$2D,$9D,$9F,$82,$7F,$6F,$1F,$96,$3E,$8D,$FD
	.dc.b	$D7,$F1,$D1,$F7,$5F,$C1,$9B,$8F,$C4,$47,$DC,$3E,$89,$D2,$FE,$35
	.dc.b	$7D,$19,$F8,$33,$FD,$FA,$3E,$DD,$7C,$13,$F0,$D1,$FD,$44,$FC,$35
	.dc.b	$7D,$08,$D1,$F8,$58,$F8,$C7,$CB,$8F,$8D,$1E,$A8,$F8,$CF,$AA,$FE
	.dc.b	$33,$EA,$9F,$2B,$F4,$27,$EF,$C7,$D0,$7F,$35,$5F,$C1,$67,$C1,$1A
	.dc.b	$3C,$06,$8F,$DB,$DF,$C9,$3F,$1C,$E8,$FA,$91,$E5,$7E,$A3,$E5,$9F
	.dc.b	$A9,$7F,$E1,$3B,$F2,$CF,$C4,$8F,$E6,$D5,$F1,$0F,$C7,$1F,$DC,$E3
	.dc.b	$55,$F7,$A3,$47,$D3,$8F,$93,$7E,$57,$F9,$C5,$77,$73,$3F,$22,$7D
	.dc.b	$2F,$D7,$8F,$CB,$9F,$9C,$3B,$9E,$97,$BC,$F3,$8D,$1F,$22,$3F,$0C
	.dc.b	$FE,$99,$F2,$3A,$9F,$23,$FC,$F6,$FF,$00,$E6,$7C,$8F,$A2,$FF,$12
	.dc.b	$E6,$3E,$5D,$7A,$2F,$FC,$41,$7F,$84,$8F,$CB,$CF,$E3,$4F,$CB,$CF
	.dc.b	$9D,$F9,$55,$E7,$74,$F9,$E3,$51,$FE,$0B,$8D,$47,$D4,$8F,$D7,$BF
	.dc.b	$D9,$67,$EC,$CE,$AF,$EA,$57,$81,$F4,$BE,$09,$F8,$2D,$3D,$F3,$ED
	.dc.b	$AE,$8F,$C1,$3E,$9C,$6C,$F8,$2F,$DA,$6B,$C6,$BF,$A8,$3C,$6F,$E2
	.dc.b	$4F,$F2,$A3,$EA,$DF,$C6,$BF,$65,$7E,$5D,$F0,$BF,$56,$7F,$22,$7C
	.dc.b	$39,$F4,$B4,$F4,$9E,$59,$E2,$FF,$58,$AF,$BA,$E6,$FC,$B7,$E7,$3F
	.dc.b	$9B,$9E,$BA,$E3,$3F,$C4,$E3,$FA,$49,$FC,$4D,$F5,$1F,$62,$FD,$55
	.dc.b	$F4,$27,$D4,$F6,$9F,$11,$EA,$AF,$EA,$F3,$F5,$CF,$CE,$47,$DC,$9E
	.dc.b	$0F,$B4,$F9,$A7,$23,$F3,$67,$AE,$3D,$27,$AE,$BF,$C1,$87,$91,$EA
	.dc.b	$7E,$36,$7F,$07,$3D,$6B,$DE,$79,$2B,$E6,$B6,$3E,$69,$F0,$EF,$E5
	.dc.b	$BC,$6B,$CF,$3B,$8F,$39,$9B,$F2,$BE,$98,$FC,$E9,$B3,$C3,$3F,$C8
	.dc.b	$6F,$C2,$D9,$F0,$CF,$F2,$CA,$FA,$77,$FA,$0B,$DD,$7F,$24,$37,$7C
	.dc.b	$5F,$1D,$F1,$AF,$4B,$E1,$9F,$49,$B7,$C7,$3C,$CE,$31,$FC,$A0,$F9
	.dc.b	$31,$C6,$FD,$B3,$EF,$BF,$A2,$FE,$1A,$7F,$3B,$7F,$CF,$E3,$F9,$AC
	.dc.b	$FD,$53,$D4,$FA,$0F,$04,$F8,$9F,$DD,$57,$EC,$9E,$C7,$F8,$0E,$79
	.dc.b	$47,$E6,$6F,$D7,$3F,$C9,$9F,$39,$ED,$32,$FE,$3E,$3D,$35,$EB,$AF
	.dc.b	$ED,$2F,$F1,$C9,$FE,$A0,$6E,$3D,$AF,$F8,$64,$FD,$D9,$EF,$AF,$D1
	.dc.b	$5F,$BE,$3F,$CD,$17,$FA,$43,$BC,$F8,$DB,$1F,$02,$3C,$B1,$F9,$73
	.dc.b	$E0,$5F,$96,$7E,$01,$DE,$3F,$3B,$7F,$09,$FE,$11,$1F,$A6,$3E,$39
	.dc.b	$FE,$2E,$8D,$3F,$A9,$7D,$F5,$FE,$AD,$8F,$F3,$21,$EF,$3B,$9E,$F3
	.dc.b	$F8,$01,$E1,$8F,$E8,$51,$C6,$BF,$CC,$B1,$DA,$7C,$27,$C6,$39,$47
	.dc.b	$E1,$1F,$EF,$07,$C6,$8E,$11,$EF,$8F,$41,$DE,$BF,$96,$3F,$51,$F9
	.dc.b	$93,$F5,$27,$FC,$8F,$3C,$0E,$F5,$F5,$1E,$E7,$BA,$B8,$BE,$E9,$E2
	.dc.b	$F9,$DA,$FF,$56,$BF,$B8,$F1,$BF,$3D,$FD,$B7,$C4,$FF,$8C,$E3,$FC
	.dc.b	$52,$7F,$7D,$9F,$E8,$6F,$92,$FE,$CB,$ED,$7F,$8D,$3C,$2F,$F2,$4F
	.dc.b	$EA,$0D,$1F,$64,$F4,$57,$77,$FC,$50,$77,$3E,$35,$F9,$9F,$C7,$57
	.dc.b	$E4,$1F,$BC,$70,$AE,$EF,$D3,$3B,$9F,$68,$FC,$09,$E9,$9F,$C0,$CF
	.dc.b	$F9,$92,$38,$19,$7F,$68,$77,$3E,$7B,$DE,$BF,$10,$7F,$A5,$67,$CD
	.dc.b	$1E,$43,$E9,$47,$B5,$FF,$0E,$DF,$D2,$BF,$C7,$C7,$99,$FC,$43,$C2
	.dc.b	$FF,$CA,$0F,$03,$F0,$31,$C1,$F3,$47,$D0,$7F,$11,$1F,$80,$3F,$2C
	.dc.b	$FC,$B3,$E9,$1C,$0F,$A4,$70,$AF,$33,$F8,$73,$BB,$ED,$7C,$CF,$E1
	.dc.b	$DF,$B3,$3E,$67,$81,$DC,$F9,$73,$E3,$9F,$ED,$D7,$F1,$AB,$E5,$D7
	.dc.b	$B4,$E0,$7E,$54,$EE,$02,$8C,$44,$8C,$99,$72,$70,$8A,$ED,$DB,$83
	.dc.b	$47,$08,$E7,$1C,$1E,$91,$63,$8B,$4E,$11,$C9,$EC,$75,$7F,$D4,$EF
	.dc.b	$58,$E9,$1B,$8F,$73,$D9,$D3,$F0,$1E,$51,$C2,$35,$1C,$4F,$63,$B7
	.dc.b	$C8,$7C,$57,$47,$07,$BC,$7A,$9F,$21,$EA,$8E,$A6,$DF,$F7,$D9,$A3
	.dc.b	$A3,$A8,$F9,$51,$A3,$E0,$9A,$7C,$6F,$B2,$36,$76,$3C,$B1,$83,$C9
	.dc.b	$1C,$8F,$8B,$1A,$7E,$1B,$EF,$7C,$86,$9F,$23,$A7,$A3,$F8,$C3,$F1
	.dc.b	$7C,$9F,$8A,$B5,$FE,$2A,$93,$98,$DD,$7E,$32,$B7,$5A,$3F,$83,$DF
	.dc.b	$E3,$A3,$FC,$CB,$7C,$67,$F2,$AD,$35,$7E,$28,$E3,$7D,$63,$FC,$A9
	.dc.b	$3C,$67,$C3,$2A,$F0,$CF,$29,$F1,$1F,$C0,$E7,$FD,$57,$1C,$63,$D7
	.dc.b	$1F,$AB,$9E,$2F,$78,$E2,$FF,$04,$38,$9B,$AF,$89,$5E,$98,$E0,$FF
	.dc.b	$88,$9F,$F1,$79,$F9,$A7,$C0,$6D,$F8,$13,$EE,$7E,$25,$FE,$AA,$FE
	.dc.b	$8C,$71,$3C,$13,$F2,$1B,$5E,$F3,$C1,$3F,$E2,$F9,$F6,$CF,$C5,$BF
	.dc.b	$3B,$C0,$F3,$C7,$85,$F6,$9B,$7B,$36,$BB,$46,$E3,$FD,$7F,$1E,$C7
	.dc.b	$A1,$EC,$8F,$A0,$E8,$F4,$BF,$62,$BE,$7C,$6C,$F9,$C6,$CF,$14,$6E
	.dc.b	$BE,$FC,$7C,$08,$F1,$4F,$E3,$23,$AC,$7C,$08,$EB,$1B,$8F,$1C,$6E
	.dc.b	$3A,$C6,$E7,$AC,$7B,$E3,$AC,$7E,$42,$3A,$D7,$D9,$7A,$1E,$8A,$E8
	.dc.b	$7E,$8A,$BD,$07,$D2,$8D,$9F,$DC,$63,$75,$F5,$23,$E7,$1F,$52,$7C
	.dc.b	$F1,$E1,$8F,$3D,$78,$5F,$34,$E8,$FF,$52,$C7,$D3,$8E,$67,$CD,$3C
	.dc.b	$13,$CE,$BC,$D5,$F0,$4F,$34,$F4,$3D,$93,$B3,$CD,$3F,$32,$34,$74
	.dc.b	$9F,$0B,$F1,$8F,$C1,$9F,$02,$3F,$BD,$BF,$75,$F0,$D7,$AD,$FD,$CC
	.dc.b	$FD,$03,$9C,$FC,$03,$99,$E0,$8E,$67,$CA,$8F,$CB,$BE,$53,$9B,$F1
	.dc.b	$6F,$F5,$97,$FD,$96,$FE,$35,$6E,$3F,$88,$56,$CF,$E9,$11,$FB,$2B
	.dc.b	$FB,$71,$F7,$2B,$F4,$6E,$CF,$E4,$91,$FB,$73,$D0,$7D,$23,$E1,$57
	.dc.b	$F8,$1D,$FD,$FC,$76,$8F,$24,$7C,$93,$81,$EF,$78,$3F,$B6,$AF,$92
	.dc.b	$FC,$47,$DE,$FD,$89,$F9,$31,$A8,$E6,$E9,$FB,$17,$F1,$1F,$44,$6C
	.dc.b	$F8,$93,$FC,$8A,$36,$F9,$DD,$D7,$E1,$0F,$D4,$CF,$53,$EA,$BF,$89
	.dc.b	$9F,$AA,$E8,$F9,$F3,$E7,$AF,$13,$F3,$A7,$F2,$97,$F3,$A3,$51,$E1
	.dc.b	$75,$1F,$89,$AE,$71,$E6,$9F,$A6,$E8,$F8,$D1,$AB,$F2,$CF,$C6,$8F
	.dc.b	$53,$FA,$28,$F5,$47,$D0,$9F,$CC,$3A,$AF,$24,$77,$3F,$A2,$1F,$12
	.dc.b	$BD,$E7,$39,$F0,$1C,$EB,$F3,$11,$F8,$B9,$FE,$00,$7B,$9F,$49,$FB
	.dc.b	$53,$4F,$EE,$E3,$D3,$7F,$22,$74,$7C,$89,$E6,$79,$4F,$E8,$47,$95
	.dc.b	$E6,$FE,$50,$FF,$45,$C7,$F6,$93,$A5,$FE,$9A,$FE,$C5,$7E,$0E,$FC
	.dc.b	$2F,$D3,$3C,$33,$F4,$DD,$4F,$BC,$FC,$13,$E0,$BF,$CE,$1E,$03,$FB
	.dc.b	$85,$F8,$1F,$D1,$5F,$E4,$AF,$FA,$ED,$FF,$BB,$1F,$25,$FC,$D3,$BD
	.dc.b	$73,$3F,$5A,$FC,$A8,$D9,$F6,$2B,$D3,$19,$9E,$F5,$EE,$AE,$F3,$FE
	.dc.b	$9F,$78,$CF,$CC,$AE,$2F,$E0,$0F,$C7,$1F,$AE,$AF,$BE,$F8,$CF,$DC
	.dc.b	$3F,$89,$79,$3F,$16,$BF,$24,$7E,$22,$BC,$D1,$9B,$F8,$95,$FC,$32
	.dc.b	$FE,$25,$FF,$27,$9C,$C7,$EE,$6B,$D5,$39,$3E,$B5,$FA,$AF,$F1,$53
	.dc.b	$D6,$73,$1F,$09,$F9,$93,$F0,$AF,$F0,$91,$A9,$F1,$57,$68,$FE,$45
	.dc.b	$1F,$0D,$FC,$1B,$FB,$07,$EA,$BF,$EC,$33,$F5,$6F,$BD,$F8,$67,$C7
	.dc.b	$32,$FE,$1E,$FF,$67,$19,$7F,$53,$1F,$9F,$8F,$59,$99,$FC,$7C,$66
	.dc.b	$3F,$3C,$78,$67,$8D,$FE,$5C,$F7,$47,$E4,$A7,$DD,$3F,$2D,$F8,$67
	.dc.b	$CB,$3F,$11,$1F,$09,$C9,$F0,$9F,$D7,$9F,$84,$8F,$AF,$19,$7F,$A6
	.dc.b	$3F,$1A,$3F,$10,$FC,$63,$FD,$86,$FC,$27,$D4,$FE,$A0,$F5,$57,$BA
	.dc.b	$BD,$53,$E5,$9F,$A5,$1F,$7D,$F8,$31,$D2,$BE,$0C,$7D,$59,$F8,$2F
	.dc.b	$EA,$0E,$27,$E0,$DF,$C9,$BF,$36,$7F,$2E,$7D,$5A,$FC,$FC,$79,$23
	.dc.b	$27,$A2,$BE,$CC,$66,$FF,$2F,$5F,$A3,$8F,$B2,$7C,$77,$F1,$8F,$88
	.dc.b	$F9,$D7,$E2,$79,$5F,$D6,$9E,$51,$EE,$9F,$80,$FC,$79,$F9,$D7,$F4
	.dc.b	$A3,$F9,$71,$D5,$FB,$D3,$D6,$77,$5E,$2A,$FB,$71,$83,$EF,$BC,$8C
	.dc.b	$CF,$82,$BD,$F3,$B8,$E8,$66,$BE,$3C,$7D,$37,$ED,$57,$D3,$8F,$61
	.dc.b	$FA,$57,$E2,$BC,$DF,$D3,$9F,$09,$FC,$9D,$7A,$A3,$31,$B7,$8D,$78
	.dc.b	$63,$F1,$D7,$FE,$A3,$3F,$D2,$33,$FA,$18,$E7,$7F,$32,$3E,$55,$71
	.dc.b	$3E,$CC,$F9,$63,$F1,$B5,$CD,$FF,$4B,$BF,$C2,$23,$99,$B8,$E6,$7E
	.dc.b	$34,$F3,$CF,$58,$FC,$D4,$7C,$C3,$47,$A8,$A7,$DF,$29,$F7,$E7,$C4
	.dc.b	$FF,$AC,$8F,$C6,$47,$D8,$7A,$1F,$8C,$3C,$27,$2B,$FB,$33,$C9,$FF
	.dc.b	$56,$57,$96,$BD,$67,$E8,$4F,$DC,$39,$AF,$C2,$4F,$EF,$A3,$11,$E0
	.dc.b	$8F,$AC,$FA,$CF,$11,$F7,$AB,$C1,$7F,$D7,$E3,$C8,$FE,$22,$B9,$D7
	.dc.b	$59,$CD,$FD,$C7,$FD,$F8,$FE,$05,$F8,$C7,$62,$C7,$89,$FB,$27,$09
	.dc.b	$EB,$5F,$E5,$48,$E0,$7D,$82,$CF,$C6,$78,$1E,$29,$FC,$2D,$7C,$56
	.dc.b	$BF,$C0,$EB,$F4,$37,$EE,$8F,$ED,$D5,$C5,$FD,$09,$C4,$A7,$13,$DC
	.dc.b	$FF,$AE,$A3,$F3,$2F,$F1,$93,$E7,$4F,$E6,$6F,$F6,$77,$EC,$9F,$F1
	.dc.b	$15,$F3,$8F,$E4,$31,$FD,$DA,$3F,$2D,$3C,$23,$C9,$1F,$80,$38,$1F
	.dc.b	$96,$8F,$A5,$1F,$12,$BE,$B3,$C0,$FE,$1C,$F8,$CF,$82,$FC,$83,$84
	.dc.b	$FE,$29,$E0,$7A,$9F,$1C,$70,$8F,$80,$7E,$B8,$FD,$C1,$DA,$36,$73
	.dc.b	$7A,$D7,$D2,$3F,$A3,$C7,$F5,$39,$FB,$D5,$F8,$3A,$FB,$CD,$3E,$13
	.dc.b	$CA,$7E,$04,$F0,$3D,$8F,$79,$E1,$1F,$0A,$78,$4F,$77,$E8,$D7,$89
	.dc.b	$F4,$47,$DF,$3E,$74,$66,$3F,$CB,$55,$F3,$8E,$B3,$DE,$BC,$35,$ED
	.dc.b	$AF,$E5,$53,$DD,$FE,$03,$1F,$CC,$9F,$9B,$5C,$EB,$8D,$7B,$27,$CA
	.dc.b	$72,$3C,$AF,$E0,$0F,$2B,$93,$CB,$1F,$0E,$3E,$8D,$78,$E7,$FB,$EB
	.dc.b	$6B,$FA,$94,$70,$3B,$9A,$9F,$CF,$CF,$C1,$9F,$34,$74,$3C,$D3,$EC
	.dc.b	$3F,$8C,$9F,$38,$F3,$CF,$1A,$FF,$1F,$4F,$A2,$7B,$36,$BC,$67,$67
	.dc.b	$E9,$47,$F1,$A9,$F6,$46,$23,$F2,$8D,$9E,$0F,$C5,$8E,$07,$B2,$3A
	.dc.b	$CF,$29,$D4,$F2,$6C,$FF,$9B,$23,$15,$E9,$79,$57,$F7,$39,$EF,$39
	.dc.b	$9F,$EC,$47,$C0,$3E,$A4,$69,$B3,$E1,$9E,$15,$FD,$C4,$FF,$0B,$B8
	.dc.b	$3F,$83,$4F,$CD,$3F,$9A,$1F,$B7,$AF,$C8,$C7,$DB,$AF,$21,$D6,$7B
	.dc.b	$9F,$E2,$7B,$D5,$FF,$51,$BF,$24,$FD,$23,$D0,$53,$F0,$87,$84,$FE
	.dc.b	$6F,$5F,$96,$9F,$B0,$F8,$23,$B4,$7E,$21,$F0,$9C,$4F,$DD,$C7,$17
	.dc.b	$EA,$99,$9F,$DE,$9E,$C9,$E5,$1E,$38,$FA,$87,$A0,$FB,$68,$42,$F1
	.dc.b	$12,$32,$65,$C9,$C2,$2B,$B7,$6E,$0D,$1C,$23,$9C,$70,$7A,$45,$8E
	.dc.b	$2D,$38,$47,$27,$FA,$E1,$CE,$36,$D3,$71,$D1,$F7,$35,$F3,$3F,$01
	.dc.b	$F3,$39,$3C,$AF,$33,$C8,$ED,$F2,$1F,$15,$F2,$3D,$A3,$E0,$3F,$16
	.dc.b	$38,$3C,$CF,$1B,$C1,$87,$17,$6F,$B8,$FA,$2F,$96,$3A,$BF,$5C,$DC
	.dc.b	$78,$4F,$83,$1E,$67,$E3,$46,$9F,$03,$EE,$7B,$1E,$17,$AC,$7D,$57
	.dc.b	$6E,$1E,$47,$C7,$33,$1F,$2A,$3E,$9B,$FA,$B3,$F2,$55,$F8,$BF,$DF
	.dc.b	$BF,$1B,$68,$F6,$C6,$1F,$C9,$BF,$8C,$8D,$5A,$3F,$67,$7F,$E7,$F7
	.dc.b	$84,$7E,$59,$A7,$1A,$E7,$1F,$A4,$AF,$D0,$DF,$63,$F4,$4F,$68,$C4
	.dc.b	$23,$C3,$5C,$E7,$C4,$73,$AF,$DD,$57,$0A,$E5,$3E,$78,$EB,$3D,$1B
	.dc.b	$1C,$8F,$E0,$67,$F0,$D8,$F8,$D1,$A8,$E9,$1C,$63,$C8,$FB,$23,$8D
	.dc.b	$FC,$39,$E1,$1E,$17,$6D,$7C,$37,$F2,$9F,$F4,$7B,$F2,$A3,$B9,$B8
	.dc.b	$E3,$1B,$7F,$D3,$6E,$E3,$B5,$FC,$69,$FE,$0A,$FC,$88,$F8,$C7,$B6
	.dc.b	$3F,$67,$3C,$4F,$AE,$F0,$BF,$63,$FC,$B2,$BE,$BD,$76,$9F,$C3,$CF
	.dc.b	$A8,$E1,$1F,$2D,$E1,$1F,$52,$38,$1F,$85,$8E,$07,$E1,$9F,$2D,$FC
	.dc.b	$98,$F8,$F1,$F0,$AF,$F0,$07,$C1,$8F,$F7,$9B,$E9,$79,$9F,$32,$39
	.dc.b	$BF,$30,$FA,$71,$EE,$7C,$11,$F6,$1D,$1E,$07,$CF,$1E,$07,$9C,$7D
	.dc.b	$03,$99,$E0,$3E,$AB,$A7,$CA,$68,$FB,$27,$C6,$8F,$53,$FE,$ED,$7D
	.dc.b	$4E,$8F,$29,$F2,$A3,$EC,$3F,$84,$7E,$23,$CD,$F0,$3C,$CF,$05,$FE
	.dc.b	$0D,$E6,$7F,$81,$0F,$EB,$B1,$F5,$0E,$6E,$A3,$FA,$B5,$FD,$9A,$F2
	.dc.b	$C7,$39,$F9,$D5,$F2,$5F,$88,$7C,$99,$F9,$CE,$8F,$55,$FD,$DA,$FD
	.dc.b	$AC,$EA,$7E,$4D,$73,$7E,$49,$CC,$F9,$D3,$F9,$B8,$D4,$7B,$CD,$1F
	.dc.b	$24,$F0,$1E,$F3,$F9,$09,$FA,$03,$E2,$DF,$EE,$E7,$A4,$79,$8F,$C1
	.dc.b	$BE,$67,$C3,$1F,$A4,$7A,$47,$E9,$2B,$E2,$CF,$33,$51,$F2,$AB,$EC
	.dc.b	$9F,$3C,$FC,$6D,$7D,$97,$D7,$1F,$CB,$AF,$FB,$8B,$FC,$D6,$7D,$07
	.dc.b	$B9,$FA,$4F,$C8,$3D,$07,$D2,$7D,$0F,$C3,$34,$7A,$2B,$EE,$3F,$0A
	.dc.b	$BF,$87,$4F,$C2,$8D,$47,$A2,$FD,$AE,$AB,$DA,$7F,$4D,$AF,$85,$5C
	.dc.b	$EB,$F5,$15,$EC,$AF,$EE,$4F,$F4,$D9,$F5,$D7,$98,$F5,$9E,$13,$E4
	.dc.b	$5F,$E1,$EB,$F8,$7C,$FA,$E7,$51,$E1,$3D,$71,$F1,$5F,$A5,$5F,$C4
	.dc.b	$E3,$D7,$5C,$E3,$F8,$94,$73,$9F,$5B,$FE,$66,$8F,$80,$FA,$CF,$AC
	.dc.b	$7C,$E3,$F2,$B5,$E9,$9E,$F7,$FB,$C8,$FB,$55,$FE,$17,$7E,$D5,$7E
	.dc.b	$E6,$3F,$02,$FE,$BE,$BD,$51,$E0,$8F,$53,$F3,$CF,$54,$FC,$48,$FC
	.dc.b	$AC,$FC,$48,$F9,$2F,$AA,$3F,$CA,$8F,$D0,$7F,$B0,$9F,$41,$E0,$7A
	.dc.b	$27,$F8,$4C,$FA,$1F,$AD,$3E,$7A,$FD,$B4,$FD,$08,$F8,$D7,$F2,$EB
	.dc.b	$F8,$4C,$7E,$C2,$7C,$CF,$E4,$EF,$BC,$FA,$EF,$E3,$47,$96,$7F,$B5
	.dc.b	$9E,$59,$FA,$B5,$EC,$7E,$0C,$FC,$6B,$F0,$47,$CA,$AF,$04,$E5,$F8
	.dc.b	$B1,$CA,$3D,$CF,$E4,$CF,$F1,$C1,$EE,$3F,$69,$7E,$E8,$F9,$51,$FE
	.dc.b	$64,$BF,$F3,$E3,$B9,$FD,$19,$F9,$EA,$FC,$21,$FC,$4E,$7F,$09,$3E
	.dc.b	$F3,$E6,$D7,$BE,$3F,$07,$1E,$09,$FE,$1A,$FB,$9F,$9A,$7A,$9F,$9A
	.dc.b	$FE,$EA,$3E,$6B,$F9,$29,$F9,$B5,$E3,$8F,$CF,$9C,$6B,$E3,$3C,$67
	.dc.b	$E3,$3F,$56,$3E,$31,$E7,$3E,$35,$FF,$12,$9F,$F1,$1C,$FE,$52,$BB
	.dc.b	$5F,$8A,$3E,$04,$F8,$9F,$D3,$9E,$23,$B5,$FA,$A7,$F1,$51,$F9,$68
	.dc.b	$FA,$6F,$E5,$8F,$C8,$1F,$12,$BF,$0F,$1A,$8F,$C3,$9A,$3F,$74,$FC
	.dc.b	$97,$A5,$FA,$5F,$C6,$47,$86,$FE,$EC,$FF,$B5,$A3,$E9,$4F,$2A,$EF
	.dc.b	$3C,$63,$B5,$FE,$4E,$73,$5F,$BD,$BF,$8B,$3F,$80,$9F,$8B,$1F,$3A
	.dc.b	$BE,$B5,$7C,$E9,$CC,$7D,$E3,$E2,$C7,$D6,$AF,$8A,$FE,$46,$BF,$DA
	.dc.b	$13,$F0,$23,$CF,$3E,$37,$F1,$A7,$C5,$7F,$1B,$5F,$96,$9F,$A7,$1F
	.dc.b	$5C,$FD,$F3,$F3,$E7,$9D,$7E,$9A,$3F,$37,$3A,$7C,$15,$CD,$FE,$30
	.dc.b	$7A,$4F,$6D,$7C,$D3,$E9,$1E,$93,$FD,$3D,$3F,$D1,$A3,$F1,$47,$92
	.dc.b	$BE,$33,$F8,$A9,$F8,$CE,$4E,$C6,$5E,$D5,$F4,$9F,$DA,$47,$F3,$73
	.dc.b	$CA,$7E,$09,$FA,$CF,$E0,$9F,$9D,$3F,$82,$8F,$F3,$BD,$FF,$A6,$8F
	.dc.b	$E1,$B5,$FC,$0E,$BD,$71,$F8,$78,$F8,$53,$FC,$40,$F2,$DF,$C7,$3E
	.dc.b	$8B,$E3,$34,$78,$EB,$CE,$FC,$23,$CE,$FD,$23,$F6,$4F,$D5,$9E,$EF
	.dc.b	$EE,$4E,$F7,$FC,$2C,$FC,$89,$FB,$17,$9C,$70,$AF,$C5,$B6,$7E,$04
	.dc.b	$FF,$97,$EF,$E0,$3F,$C1,$E7,$E0,$1F,$A1,$BF,$23,$F9,$88,$FD,$1D
	.dc.b	$7E,$1D,$FD,$F1,$F8,$29,$F9,$A7,$E0,$A7,$C0,$D8,$F8,$87,$AC,$F8
	.dc.b	$87,$BC,$CD,$FA,$1B,$5F,$CD,$2B,$99,$F8,$18,$E0,$7C,$B3,$E7,$D7
	.dc.b	$A1,$FC,$51,$E8,$7E,$6B,$E8,$3F,$18,$FA,$23,$E0,$1D,$9F,$C4,$C7
	.dc.b	$18,$F8,$51,$F9,$D0,$7E,$7E,$FF,$13,$3F,$C6,$63,$89,$E3,$7E,$A9
	.dc.b	$EC,$7F,$1A,$FE,$48,$FF,$3A,$CF,$86,$3F,$0F,$7F,$C3,$67,$D7,$1C
	.dc.b	$DF,$7D,$7E,$55,$E9,$5E,$A6,$CF,$8A,$3A,$47,$EF,$AF,$F1,$4F,$E9
	.dc.b	$63,$F0,$33,$B8,$E2,$F8,$A7,$F1,$D1,$E0,$3D,$07,$F9,$B8,$F8,$C7
	.dc.b	$9A,$7B,$BF,$C2,$8E,$67,$86,$7F,$8C,$4F,$E6,$9F,$E0,$31,$F1,$8F
	.dc.b	$B8,$7E,$1D,$F8,$0D,$AF,$80,$EA,$FF,$10,$DA,$FC,$43,$E0,$3F,$01
	.dc.b	$1F,$B7,$8F,$05,$7F,$0C,$3E,$04,$F1,$7F,$30,$DB,$FC,$83,$EB,$31
	.dc.b	$1C,$2B,$EB,$47,$A5,$AF,$C7,$3F,$60,$7D,$E3,$C8,$FF,$08,$79,$C7
	.dc.b	$CF,$3E,$1B,$C6,$3F,$C2,$6F,$E8,$1E,$AF,$C5,$9F,$9F,$1C,$CC,$9F
	.dc.b	$D3,$2F,$E3,$B8,$BF,$8B,$5E,$03,$A3,$F7,$9F,$9C,$7D,$E8,$FE,$74
	.dc.b	$71,$BF,$8A,$DB,$FC,$84,$F3,$3E,$7D,$7E,$39,$FC,$29,$F1,$5E,$D3
	.dc.b	$F6,$0E,$55,$FD,$72,$F9,$C6,$23,$E2,$1F,$5C,$A7,$EE,$4E,$07,$D3
	.dc.b	$9E,$A7,$BA,$39,$57,$BA,$7E,$6B,$EE,$3E,$A9,$F9,$B7,$F1,$F1,$E5
	.dc.b	$9E,$71,$F0,$1B,$7F,$84,$9F,$A1,$18,$9F,$11,$DA,$39,$36,$3F,$99
	.dc.b	$C7,$DC,$AF,$C6,$3E,$66,$C7,$39,$F7,$D7,$ED,$9F,$F1,$B9,$D8,$F8
	.dc.b	$93,$ED,$6B,$C9,$F0,$1F,$C4,$1F,$7D,$7D,$CA,$ED,$1C,$67,$E6,$1E
	.dc.b	$07,$F8,$E3,$E0,$30,$FF,$8A,$5F,$11,$C1,$E5,$5E,$77,$F8,$05,$FD
	.dc.b	$CA,$EF,$1E,$5A,$F3,$DF,$CA,$7C,$F7,$F3,$6B,$F6,$F1,$D6,$B8,$36
	.dc.b	$B8,$3C,$A7,$FC,$1B,$1F,$BF,$9F,$E3,$C7,$A8,$F2,$94,$E7,$5E,$77
	.dc.b	$F3,$35,$FC,$59,$E2,$7B,$DF,$83,$3F,$E3,$36,$D7,$F6,$58,$FF,$18
	.dc.b	$3F,$A1,$8F,$C5,$BF,$93,$3E,$0C,$7A,$23,$D5,$1E,$83,$85,$7E,$F1
	.dc.b	$F8,$2F,$96,$78,$36,$38,$4F,$9E,$BF,$5E,$7F,$64,$9E,$4F,$AA,$BD
	.dc.b	$0F,$D4,$9F,$B3,$3C,$E3,$55,$EB,$3C,$2F,$C7,$AF,$6D,$77,$3D,$A7
	.dc.b	$E9,$4F,$B2,$74,$8E,$13,$E6,$BF,$C1,$4E,$8F,$92,$7F,$76,$AF,$6D
	.dc.b	$70,$3F,$C4,$31,$DE,$FD,$71,$FD,$C9,$30,$B1,$FF,$00,$1B,$19,$64
	.dc.b	$23,$FB,$DB,$98,$FF,$50,$9C,$9E,$CF,$08,$D3,$D2,$3F,$C0,$6E,$1E
	.dc.b	$8F,$67,$8B,$D0,$F3,$3E,$13,$8B,$E1,$78,$BA,$78,$BE,$A3,$B4,$7A
	.dc.b	$1A,$77,$8E,$2E,$CE,$91,$B6,$9B,$6B,$A8,$F1,$BE,$F3,$45,$34,$FC
	.dc.b	$27,$CA,$7C,$A7,$C2,$E0,$F8,$CE,$CC,$9F,$55,$D1,$F6,$DF,$0B,$F1
	.dc.b	$DD,$1D,$1E,$67,$C0,$3E,$33,$D2,$3B,$1E,$F8,$F4,$1F,$6E,$34,$FC
	.dc.b	$A7,$BB,$FA,$07,$F2,$57,$F5,$BF,$7F,$FC,$65,$A3,$DD,$18,$7F,$C8
	.dc.b	$6F,$E2,$63,$52,$8F,$CF,$BF,$C5,$6B,$F2,$B1,$C6,$3F,$22,$D3,$95
	.dc.b	$74,$8F,$D5,$4F,$F1,$13,$FD,$75,$3C,$63,$C3,$1F,$A2,$AF,$0C,$F2
	.dc.b	$9F,$11,$CA,$3F,$80,$CF,$FB,$36,$79,$4F,$9E,$39,$57,$47,$C9,$1C
	.dc.b	$63,$F8,$24,$7A,$8D,$C7,$A9,$B3,$E1,$3E,$0C,$7F,$1E,$8E,$02,$BE
	.dc.b	$3C,$77,$7D,$31,$FB,$77,$D7,$1E,$13,$EA,$C7,$28,$E3,$1E,$98,$E9
	.dc.b	$5E,$97,$C7,$1C,$6F,$FC,$F7,$7F,$00,$FA,$13,$DE,$BE,$85,$7F,$14
	.dc.b	$AD,$CF,$BA,$3C,$D3,$D6,$3B,$C7,$CE,$38,$3F,$E7,$8A,$EC,$FC,$43
	.dc.b	$B4,$6D,$FE,$A6,$7D,$D3,$EA,$1D,$DF,$65,$FC,$48,$FA,$71,$D9,$FF
	.dc.b	$65,$CF,$67,$C6,$FD,$18,$F8,$11,$E3,$3A,$47,$C0,$79,$9E,$29,$FC
	.dc.b	$29,$F9,$F7,$E7,$57,$E8,$1F,$51,$C0,$FE,$D2,$7F,$37,$3E,$2D,$7E
	.dc.b	$0D,$F1,$9F,$02,$BF,$07,$7F,$D3,$E3,$73,$C2,$7D,$0F,$B6,$3F,$D7
	.dc.b	$B1,$ED,$BF,$85,$1F,$72,$74,$FF,$A1,$E7,$57,$F5,$6B,$F1,$35,$A7
	.dc.b	$F6,$B7,$FE,$DF,$AD,$CF,$E2,$6F,$A1,$F0,$EF,$F3,$EF,$9A,$3E,$9C
	.dc.b	$75,$3F,$0F,$1B,$8F,$CA,$BB,$9F,$9C,$F4,$3E,$39,$F7,$EF,$E7,$3B
	.dc.b	$3E,$69,$B7,$FB,$1B,$BA,$FE,$28,$FD,$A3,$67,$E0,$2B,$EF,$BF,$84
	.dc.b	$34,$7D,$23,$EC,$4F,$43,$EF,$CF,$48,$FC,$DC,$FD,$37,$EC,$3D,$1F
	.dc.b	$A6,$74,$9F,$A6,$F8,$63,$E7,$4F,$F3,$B7,$F9,$34,$7E,$29,$FC,$24
	.dc.b	$F0,$3E,$AC,$7C,$9B,$FD,$7B,$E2,$9F,$9E,$7D,$57,$B1,$B3,$F3,$86
	.dc.b	$EF,$E7,$BA,$3E,$7D,$7B,$9F,$14,$FB,$5F,$87,$7F,$E1,$E7,$E7,$1A
	.dc.b	$8F,$0D,$7E,$52,$3C,$07,$CD,$3C,$07,$D2,$AF,$5B,$E0,$AF,$C4,$D7
	.dc.b	$A8,$F0,$47,$DD,$8F,$04,$F9,$63,$F0,$47,$D7,$34,$7F,$55,$BF,$8B
	.dc.b	$1F,$C8,$2F,$DE,$F8,$0F,$B9,$5E,$08,$FC,$E9,$E9,$AF,$25,$FE,$58
	.dc.b	$F2,$1E,$E7,$EF,$BF,$68,$FF,$0B,$1F,$69,$F2,$D7,$D9,$AF,$D0,$C7
	.dc.b	$4B,$F2,$C7,$F0,$78,$F5,$1F,$4E,$7C,$B5,$E1,$BF,$2C,$F8,$0F,$9A
	.dc.b	$7E,$38,$F9,$A7,$29,$F7,$57,$C8,$3D,$D1,$F9,$37,$F6,$95,$FC,$B6
	.dc.b	$3C,$87,$E1,$A7,$E4,$3F,$5E,$7D,$31,$E4,$BF,$4D,$7E,$5A,$79,$DF
	.dc.b	$F1,$8B,$FE,$3D,$3F,$22,$3E,$8C,$F2,$3C,$91,$ED,$9F,$23,$F9,$9B
	.dc.b	$F4,$C7,$DB,$3C,$93,$F1,$E3,$E6,$4F,$A5,$F2,$4E,$A3,$C8,$73,$9F
	.dc.b	$04,$73,$3F,$29,$3E,$E7,$E4,$46,$9F,$F4,$29,$FC,$9C,$F2,$D7,$E1
	.dc.b	$67,$BC,$72,$9F,$21,$F0,$AB,$F8,$14,$71,$3E,$AC,$FA,$2B,$E3,$47
	.dc.b	$EA,$8F,$C3,$D7,$1A,$FC,$65,$FF,$31,$9F,$09,$F8,$13,$EB,$9E,$18
	.dc.b	$F7,$D7,$96,$BE,$6D,$7D,$69,$FD,$5D,$7E,$36,$72,$FC,$78,$CD,$FD
	.dc.b	$08,$F8,$2F,$F4,$28,$FF,$38,$BF,$CC,$9F,$B9,$1F,$85,$7E,$94,$F4
	.dc.b	$8F,$24,$74,$8F,$98,$FF,$A8,$AB,$D9,$3F,$12,$7E,$03,$F9,$C3,$E4
	.dc.b	$57,$D5,$3E,$45,$7C,$C7,$C8,$FE,$BC,$F1,$C7,$DA,$7C,$6F,$DF,$3F
	.dc.b	$4D,$1C,$E7,$BC,$7C,$A3,$2F,$DA,$9F,$C4,$1C,$A3,$E1,$C7,$F3,$77
	.dc.b	$FC,$9A,$7D,$43,$F9,$29,$DC,$F4,$BC,$EF,$EA,$46,$6B,$D2,$F2,$3E
	.dc.b	$15,$78,$27,$F0,$B3,$FC,$7E,$3D,$AF,$C2,$7F,$0A,$7D,$4A,$FE,$A9
	.dc.b	$1E,$D8,$E4,$7C,$78,$F4,$57,$F8,$A0,$EF,$5E,$47,$DB,$3F,$C6,$9F
	.dc.b	$25,$73,$9D,$4F,$D9,$3C,$13,$FD,$06,$3F,$7B,$1F,$1C,$D1,$F5,$5F
	.dc.b	$E3,$31,$F1,$8E,$F7,$F8,$2A,$ED,$1F,$9D,$7F,$5A,$FD,$4A,$FC,$EB
	.dc.b	$9B,$FC,$73,$98,$FC,$71,$E1,$AF,$F7,$0B,$F2,$27,$EF,$D7,$4A,$FA
	.dc.b	$B3,$D1,$F2,$3B,$3E,$49,$E9,$7E,$F1,$FD,$38,$F4,$BF,$09,$FD,$CB
	.dc.b	$C1,$F9,$33,$F8,$17,$D0,$7C,$73,$F5,$17,$F7,$A3,$F9,$AC,$7D,$43
	.dc.b	$67,$48,$FA,$27,$CA,$7C,$53,$F3,$9E,$EF,$AD,$CC,$6D,$FE,$00,$78
	.dc.b	$4F,$D3,$3E,$18,$FE,$65,$1E,$13,$E1,$C7,$1A,$FE,$09,$3F,$16,$3F
	.dc.b	$07,$7F,$39,$D5,$FE,$F9,$B3,$F8,$36,$9D,$E7,$F8,$A4,$7F,$6B,$BE
	.dc.b	$E7,$CE,$3C,$E7,$C1,$3E,$1B,$EF,$7F,$35,$7F,$88,$78,$9E,$93,$E0
	.dc.b	$9C,$DF,$4C,$F9,$E3,$17,$E7,$6D,$79,$DB,$7E,$77,$F9,$34,$79,$DE
	.dc.b	$A7,$73,$CE,$FC,$57,$CF,$5D,$DD,$5F,$CA,$AF,$87,$5F,$89,$38,$4F
	.dc.b	$E4,$AB,$85,$7B,$63,$BB,$5E,$F5,$CA,$BF,$81,$9F,$97,$8E,$F7,$ED
	.dc.b	$6C,$7B,$5B,$3E,$D7,$F9,$BC,$7B,$5E,$A7,$C2,$7C,$D1,$DE,$3F,$9B
	.dc.b	$57,$F1,$A7,$E9,$D7,$E7,$DF,$A7,$1E,$73,$E0,$3C,$EB,$CE,$FE,$61
	.dc.b	$FD,$CD,$7F,$03,$AE,$53,$F0,$1A,$FC,$EB,$F8,$B1,$8B,$EB,$3E,$CB
	.dc.b	$F3,$BD,$CF,$3C,$FF,$22,$9F,$EB,$57,$EB,$9F,$4D,$7E,$A5,$F8,$27
	.dc.b	$37,$CE,$E9,$F3,$9D,$9D,$1E,$B7,$E5,$4F,$E4,$23,$FD,$AD,$7F,$EB
	.dc.b	$B9,$FB,$F1,$ED,$7C,$13,$EA,$3E,$45,$FF,$68,$AF,$7C,$EA,$3E,$E4
	.dc.b	$EA,$3E,$E3,$6F,$EE,$5F,$F2,$59,$F1,$D7,$03,$E4,$15,$E8,$7F,$29
	.dc.b	$8F,$39,$FE,$3A,$9F,$8D,$5F,$22,$3E,$61,$F0,$E7,$A9,$C5,$FF,$4A
	.dc.b	$5F,$BE,$B8,$DF,$F5,$63,$C3,$5E,$86,$C7,$A2,$BF,$9A,$CF,$CE,$3F
	.dc.b	$39,$7E,$D7,$E8,$35,$FA,$AF,$96,$3E,$99,$E5,$9F,$EA,$A7,$E3,$67
	.dc.b	$DE,$F7,$9F,$CD,$1C,$4E,$93,$F6,$E3,$8C,$F8,$2B,$CE,$FF,$6D,$3F
	.dc.b	$B4,$4F,$73,$F9,$EB,$B3,$B9,$4F,$63,$F9,$47,$8D,$78,$67,$AC,$7C
	.dc.b	$06,$CF,$8C,$EB,$1F,$02,$3E,$63,$D8,$E0,$58,$E0,$7F,$40,$7D,$17
	.dc.b	$F0,$DF,$1C,$7A,$4D,$3C,$8C,$D7,$F6,$93,$E3,$D7,$82,$32,$D7,$F0
	.dc.b	$77,$D5,$A7,$80,$F8,$46,$0E,$C5,$AF,$3D,$7F,$87,$DF,$8D,$7C,$EB
	.dc.b	$E9,$D7,$06,$CF,$0B,$FE,$FE,$7F,$81,$9F,$C5,$CF,$57,$DD,$7D,$5B
	.dc.b	$3C,$1F,$F1,$B3,$FC,$1C,$F6,$57,$68,$DD,$7F,$7F,$6C,$7F,$64,$7F
	.dc.b	$52,$7F,$9F,$63,$F7,$EF,$F0,$CA,$FC,$6B,$C0,$FE,$8E,$70,$7F,$C8
	.dc.b	$13,$FA,$88,$FE,$F1,$1E,$73,$F0,$D3,$F7,$63,$FC,$B4,$7B,$A7,$CD
	.dc.b	$3A,$9F,$AB,$3C,$AB,$DC,$FF,$86,$A3,$E5,$3F,$8C,$3E,$53,$ED,$9E
	.dc.b	$4F,$C9,$7F,$05,$3F,$16,$7A,$4E,$E3,$E3,$C7,$33,$DC,$7F,$5D,$3D
	.dc.b	$CF,$F8,$82,$7E,$19,$DA,$BD,$31,$CA,$7F,$26,$7E,$0A,$BF,$13,$3A
	.dc.b	$80,$84,$62,$24,$64,$CB,$93,$84,$57,$6E,$DC,$1A,$38,$47,$38,$E0
	.dc.b	$F4,$8B,$1C,$5A,$70,$8E,$4F,$63,$AB,$FE,$F3,$7A,$C7,$48,$DC,$7B
	.dc.b	$9E,$CE,$9F,$80,$F2,$8E,$11,$A8,$E2,$7B,$1D,$BE,$43,$E2,$BA,$38
	.dc.b	$3D,$E3,$D4,$F9,$0F,$54,$75,$36,$FC,$A7,$AB,$B3,$AB,$B8,$F9,$71
	.dc.b	$B8,$A7,$33,$C5,$18,$3C,$31,$F1,$CE,$6F,$A5,$F8,$51,$B7,$CB,$19
	.dc.b	$7E,$04,$65,$E2,$79,$4E,$27,$C9,$3F,$AE,$D7,$FB,$7F,$F8,$03,$F1
	.dc.b	$4A,$FF,$15,$49,$CC,$7E,$45,$FC,$54,$EE,$74,$7E,$62,$BF,$49,$1D
	.dc.b	$5F,$CC,$BB,$8E,$51,$8A,$AF,$68,$E6,$7E,$E1,$FD,$28,$9E,$8D,$9F
	.dc.b	$3D,$71,$3C,$AF,$9E,$BC,$93,$CA,$B8,$9E,$38,$F7,$47,$8D,$AF,$37
	.dc.b	$8C,$7F,$15,$9F,$D7,$D7,$9E,$7D,$33,$F0,$E3,$C3,$1E,$AA,$FE,$1F
	.dc.b	$7F,$EC,$E9,$E1,$3D,$AB,$85,$7E,$DD,$E5,$1E,$0B,$FE,$36,$ED,$F7
	.dc.b	$CF,$9E,$BB,$9F,$D2,$5F,$A3,$1E,$67,$D9,$3E,$68,$F9,$F1,$C6,$3B
	.dc.b	$1F,$C7,$5F,$D9,$4F,$98,$F9,$8F,$96,$7E,$64,$FD,$5A,$F8,$93,$C0
	.dc.b	$EF,$1E,$73,$E5,$BD,$DE,$13,$EC,$AE,$D3,$EB,$3B,$4F,$C9,$7F,$96
	.dc.b	$9F,$EC,$48,$DB,$F8,$18,$EA,$D8,$F8,$4F,$DF,$9F,$84,$EE,$7E,$11
	.dc.b	$C2,$BE,$7B,$E4,$3E,$14,$69,$DD,$7C,$E7,$E3,$DF,$DB,$9F,$7C,$FC
	.dc.b	$67,$F1,$13,$CE,$75,$1E,$18,$F8,$93,$CD,$D4,$7F,$B0,$A3,$DB,$5F
	.dc.b	$84,$8E,$67,$E2,$8D,$1E,$03,$E4,$1A,$3E,$39,$FE,$B8,$8E,$77,$F8
	.dc.b	$08,$F8,$D5,$EA,$79,$9E,$59,$F5,$3F,$5C,$F0,$1F,$B5,$BF,$E8,$17
	.dc.b	$FA,$AB,$FC,$9C,$FD,$13,$EC,$D7,$D2,$9F,$C2,$C7,$CB,$BF,$82,$E9
	.dc.b	$F1,$3D,$5F,$F6,$44,$FA,$1F,$AD,$5F,$90,$9F,$19,$F8,$33,$E1,$1E
	.dc.b	$13,$E1,$57,$F6,$F7,$F4,$67,$E5,$4F,$71,$F2,$0F,$71,$F2,$1F,$E3
	.dc.b	$97,$EE,$9F,$C4,$C7,$F0,$7B,$F8,$B5,$F2,$AF,$EC,$BE,$97,$F9,$3B
	.dc.b	$DE,$78,$3D,$CE,$07,$A8,$FA,$C7,$A9,$FC,$03,$E4,$7D,$4F,$E6,$CE
	.dc.b	$95,$F4,$23,$71,$D0,$F5,$47,$F5,$17,$F9,$71,$C2,$7F,$C0,$D5,$EA
	.dc.b	$AF,$D0,$CF,$8D,$F8,$33,$F4,$A3,$C9,$5F,$89,$9D,$C7,$69,$FF,$09
	.dc.b	$9B,$7E,$69,$F4,$CF,$54,$FF,$87,$E7,$F0,$73,$FC,$14,$FD,$6C,$FE
	.dc.b	$0D,$F9,$6F,$DD,$39,$C7,$B8,$F0,$9F,$10,$E7,$5F,$87,$3F,$59,$1E
	.dc.b	$BA,$F8,$4F,$F4,$73,$E1,$57,$81,$FC,$3C,$79,$4F,$E2,$8F,$C0,$8F
	.dc.b	$C8,$9F,$C4,$EB,$C1,$3A,$9F,$2B,$F1,$4F,$CB,$3F,$16,$FF,$14,$F8
	.dc.b	$2B,$F4,$2F,$80,$FF,$26,$C7,$F0,$6A,$FC,$3C,$7C,$93,$EE,$3E,$09
	.dc.b	$FE,$2A,$7C,$93,$9C,$7C,$E3,$99,$FB,$C8,$E6,$7C,$9A,$F2,$BC,$EB
	.dc.b	$C1,$1F,$8B,$3D,$55,$EF,$7E,$23,$E0,$BF,$9D,$7F,$5C,$F9,$D1,$F8
	.dc.b	$23,$F4,$2F,$EF,$E3,$F1,$D3,$E0,$39,$D7,$82,$BF,$27,$5F,$A5,$BF
	.dc.b	$BB,$5A,$8F,$53,$E5,$9F,$A7,$7F,$C3,$63,$E9,$57,$B9,$F4,$C6,$67
	.dc.b	$D1,$5F,$6E,$3F,$0A,$FD,$58,$FC,$C5,$71,$7A,$9F,$C5,$23,$F2,$51
	.dc.b	$99,$E2,$FC,$98,$F2,$CF,$A9,$FB,$11,$EA,$7E,$99,$FC,$14,$FD,$99
	.dc.b	$F1,$4F,$81,$5F,$15,$D9,$E0,$7E,$05,$78,$1F,$89,$1C,$CF,$CE,$1C
	.dc.b	$CF,$0B,$CD,$FF,$39,$CF,$FA,$4A,$3F,$8D,$9D,$23,$ED,$1F,$0E,$BE
	.dc.b	$C9,$F8,$DA,$FB,$2F,$E0,$E3,$CA,$7F,$18,$3F,$42,$FC,$83,$F7,$2E
	.dc.b	$6B,$E4,$3D,$9F,$90,$FA,$5F,$E9,$51,$F2,$5F,$CB,$3D,$E3,$D3,$5F
	.dc.b	$E5,$4B,$FF,$52,$47,$BE,$B7,$7F,$7E,$7F,$19,$7F,$BD,$AF,$D2,$1C
	.dc.b	$9F,$F6,$65,$78,$1D,$47,$EB,$AF,$E6,$BF,$46,$3C,$91,$F8,$C7,$F0
	.dc.b	$91,$ED,$8F,$94,$FE,$EC,$ED,$1F,$ED,$47,$CD,$1C,$63,$E2,$CF,$F2
	.dc.b	$E7,$E7,$DF,$EB,$2B,$94,$7F,$BB,$27,$51,$F8,$88,$FA,$8F,$E2,$1F
	.dc.b	$8C,$6E,$BF,$8F,$CF,$B6,$3F,$82,$CF,$B4,$FB,$31,$D1,$FC,$4C,$74
	.dc.b	$7F,$12,$F4,$BF,$F7,$B4,$7A,$A3,$EF,$3F,$C0,$23,$F8,$EB,$9B,$FC
	.dc.b	$B3,$DA,$FC,$67,$E4,$2B,$E5,$9C,$A3,$F2,$07,$B6,$7E,$13,$F6,$23
	.dc.b	$D8,$7F,$0D,$74,$7E,$EC,$FC,$64,$FD,$87,$EA,$47,$69,$F1,$C7,$D4
	.dc.b	$8F,$F7,$F4,$F8,$DF,$C0,$DF,$ED,$E7,$57,$F8,$E9,$F8,$93,$F8,$E7
	.dc.b	$F0,$2F,$C9,$3E,$75,$FE,$98,$FC,$95,$7F,$04,$9E,$B7,$FB,$37,$D7
	.dc.b	$7F,$21,$EB,$1F,$85,$9F,$99,$1F,$C5,$23,$F1,$B1,$F0,$E7,$E9,$CF
	.dc.b	$03,$E5,$5F,$EB,$27,$F3,$11,$E0,$3F,$0B,$7F,$32,$3F,$04,$7D,$73
	.dc.b	$E4,$9F,$5E,$3E,$73,$F8,$48,$E1,$3F,$16,$7C,$11,$F3,$1B,$3C,$8F
	.dc.b	$92,$7F,$8C,$0E,$C7,$E7,$23,$9D,$FC,$36,$CF,$E3,$E7,$FD,$AB,$3F
	.dc.b	$56,$3F,$33,$5E,$1A,$D9,$F2,$4B,$1F,$52,$7F,$1F,$5F,$DB,$E7,$EA
	.dc.b	$4F,$CB,$3F,$09,$1E,$77,$D2,$FE,$C2,$FE,$6C,$F4,$9F,$DA,$CF,$E3
	.dc.b	$CF,$51,$4F,$A8,$7A,$0F,$01,$FB,$58,$F3,$3B,$3E,$8C,$78,$1E,$27
	.dc.b	$FA,$16,$7F,$21,$3E,$18,$FC,$F9,$FC,$B9,$FA,$06,$5F,$D4,$9F,$5C
	.dc.b	$ED,$7F,$1D,$B7,$F3,$DB,$7F,$3D,$F0,$1F,$5D,$F2,$3E,$6A,$ED,$5F
	.dc.b	$9E,$8F,$64,$FC,$F3,$F0,$F1,$98,$FE,$13,$3F,$01,$F4,$1F,$3D,$E7
	.dc.b	$5F,$B7,$9F,$D5,$3F,$7E,$3E,$BD,$7D,$0A,$E2,$7D,$02,$9F,$22,$BF
	.dc.b	$AE,$3F,$40,$F0,$C7,$DF,$3F,$45,$3E,$39,$FE,$A7,$3F,$7C,$FB,$07
	.dc.b	$E5,$8F,$E2,$11,$E6,$7A,$47,$F0,$19,$FE,$73,$5F,$8E,$9E,$07,$C7
	.dc.b	$3E,$6D,$77,$8F,$DC,$C7,$E0,$2F,$F1,$AE,$AB,$CC,$7E,$7A,$7E,$B4
	.dc.b	$7B,$8E,$07,$9A,$3E,$AD,$FD,$D8,$D3,$4F,$6B,$F1,$67,$FB,$43,$F7
	.dc.b	$6F,$D8,$FA,$27,$84,$F9,$9F,$DF,$5F,$9A,$7A,$CF,$92,$7A,$C7,$BC
	.dc.b	$CD,$FE,$2E,$3E,$9C,$75,$9E,$07,$52,$CF,$EC,$A3,$E9,$1F,$92,$3F
	.dc.b	$D3,$C7,$17,$F8,$CB,$F1,$DF,$C2,$4F,$88,$F3,$C7,$C5,$BF,$6C,$FC
	.dc.b	$6A,$EE,$FE,$08,$FC,$C5,$FD,$57,$C5,$5F,$80,$3B,$1F,$B0,$BE,$D5
	.dc.b	$FC,$E6,$BE,$24,$62,$BC,$B5,$F0,$AF,$A4,$77,$3F,$0E,$7E,$9A,$FF
	.dc.b	$3D,$3F,$AA,$8F,$5C,$7C,$57,$9D,$75,$AF,$E9,$75,$FE,$27,$AE,$2F
	.dc.b	$E7,$AB,$89,$EA,$7E,$A5,$71,$8F,$79,$F8,$EA,$FD,$6C,$FB,$6B,$C1
	.dc.b	$1E,$89,$FC,$91,$F6,$E3,$F1,$33,$F5,$23,$E6,$1E,$68,$E2,$F5,$8F
	.dc.b	$61,$FD,$69,$F2,$9F,$0D,$F0,$3E,$7A,$FC,$8B,$67,$E7,$4F,$DB,$3E
	.dc.b	$99,$FC,$74,$F9,$D1,$DD,$F8,$47,$73,$AB,$DD,$E1,$3F,$0E,$3C,$67
	.dc.b	$BA,$BE,$6C,$FF,$68,$8C,$1D,$67,$F8,$04,$74,$8C,$9F,$72,$FE,$45
	.dc.b	$7C,$C9,$EF,$7F,$12,$BB,$CE,$6F,$FB,$B5,$7F,$85,$A7,$CD,$1F,$CA
	.dc.b	$4E,$F1,$CA,$3E,$21,$F4,$CF,$F2,$CD,$7E,$0A,$76,$71,$7F,$C1,$53
	.dc.b	$C6,$7A,$C7,$F2,$F3,$F8,$C3,$F6,$AB,$D3,$3F,$60,$E1,$1E,$E7,$E6
	.dc.b	$18,$8F,$54,$F8,$4E,$F5,$E1,$AF,$F0,$9C,$77,$7F,$41,$5F,$CA,$A3
	.dc.b	$67,$18,$DC,$65,$AF,$F8,$28,$FC,$85,$7E,$F0,$8F,$E1,$2B,$E5,$9F
	.dc.b	$D2,$9E,$51,$FD,$99,$10,$4F,$FA,$DD,$CA,$8D,$44,$8C,$1C,$1C,$9C
	.dc.b	$22,$C7,$38,$EB,$1A,$75,$1C,$5D,$C7,$08,$DC,$77,$8E,$0F,$28,$E0
	.dc.b	$E4,$F2,$3C,$1F,$63,$E9,$7C,$0F,$A0,$E6,$65,$DB,$C5,$F5,$1C,$23
	.dc.b	$E2,$C7,$07,$CC,$F1,$8D,$9E,$A7,$CA,$F8,$A3,$67,$B9,$E9,$1A,$78
	.dc.b	$9F,$41,$F1,$30,$F8,$EF,$85,$F7,$BE,$17,$47,$C5,$74,$FB,$1D,$1B
	.dc.b	$79,$9E,$D3,$9C,$79,$5F,$54,$72,$7C,$8F,$28,$F8,$4F,$A6,$3F,$AA
	.dc.b	$47,$F9,$E7,$9F,$F1,$94,$BF,$93,$69,$CC,$7E,$41,$FC,$4C,$EE,$74
	.dc.b	$7F,$0F,$9F,$D2,$DF,$E5,$2B,$8C,$FD,$46,$9C,$A7,$A4,$71,$3F,$52
	.dc.b	$78,$67,$F4,$35,$DE,$3C,$51,$89,$AF,$48,$E5,$7F,$E7,$DB,$F0,$C6
	.dc.b	$AF,$D1,$1E,$18,$F4,$9F,$C7,$AB,$F8,$54,$FC,$37,$D4,$77,$8F,$53
	.dc.b	$67,$91,$FC,$49,$FE,$76,$7F,$B5,$1F,$0C,$7A,$E1,$3E,$27,$F6,$EF
	.dc.b	$92,$3D,$31,$FE,$7D,$9E,$27,$C8,$77,$7F,$22,$7D,$11,$E0,$AF,$43
	.dc.b	$EF,$9F,$B7,$5C,$63,$E3,$57,$0A,$FE,$63,$5F,$B2,$AE,$F5,$ED,$3B
	.dc.b	$D7,$C6,$7F,$A5,$5F,$F2,$A7,$E2,$9E,$CA,$EE,$F1,$3B,$9C,$0F,$9A
	.dc.b	$F7,$7E,$64,$F9,$CF,$8C,$FC,$D7,$E1,$BF,$8A,$9F,$CC,$3C,$23,$E6
	.dc.b	$5F,$CB,$8E,$E7,$F3,$F3,$B9,$E6,$7F,$15,$1C,$2B,$F1,$33,$DE,$3E
	.dc.b	$24,$7E,$6A,$3C,$91,$E5,$BF,$BE,$F9,$23,$FD,$FD,$1F,$21,$F8,$51
	.dc.b	$EA,$6C,$7E,$16,$BE,$7C,$78,$EB,$E3,$BA,$3E,$7C,$EE,$3C,$55,$E7
	.dc.b	$3F,$23,$3B,$3D,$F5,$FD,$66,$FF,$41,$1F,$16,$FF,$00,$FB,$4F,$A4
	.dc.b	$7E,$1E,$FE,$93,$E8,$BF,$73,$E2,$7D,$CF,$FB,$02,$7D,$D7,$F3,$E3
	.dc.b	$A1,$F3,$CE,$87,$B6,$7C,$F1,$B8,$F8,$71,$D2,$3F,$DE,$93,$D2,$BE
	.dc.b	$9D,$6A,$BE,$99,$FC,$12,$74,$FC,$DA,$FD,$43,$AB,$F8,$35,$EB,$8F
	.dc.b	$91,$5F,$61,$D4,$7E,$B4,$F8,$C7,$E5,$CF,$5D,$7C,$B7,$47,$39,$F9
	.dc.b	$6F,$94,$F9,$71,$F7,$1F,$C3,$CF,$81,$F9,$4F,$81,$FC,$D3,$FD,$AE
	.dc.b	$3F,$6E,$FD,$D3,$E8,$C7,$DC,$BF,$F0,$75,$FB,$2B,$F3,$37,$E2,$7E
	.dc.b	$0D,$78,$AB,$F0,$F3,$F1,$8F,$81,$19,$BF,$94,$FC,$03,$F1,$47,$C0
	.dc.b	$9E,$EF,$C0,$AE,$EF,$C5,$3E,$01,$F7,$DF,$81,$3E,$93,$E0,$47,$A4
	.dc.b	$FF,$0D,$9C,$E7,$F3,$F3,$F8,$33,$D3,$1C,$E3,$D2,$F9,$63,$51,$FC
	.dc.b	$7E,$3D,$57,$F0,$E3,$99,$F8,$89,$E6,$7D,$99,$F9,$F5,$FD,$69,$FC
	.dc.b	$A3,$F3,$9F,$80,$79,$8D,$9F,$D3,$4D,$BE,$C7,$67,$F6,$A7,$ED,$4F
	.dc.b	$EA,$A7,$E3,$CF,$CC,$3E,$3B,$E3,$7F,$83,$9E,$D8,$F4,$1E,$D7,$E1
	.dc.b	$4F,$D6,$7F,$20,$7D,$19,$F6,$CF,$D6,$8F,$39,$FD,$3C,$E9,$3F,$A5
	.dc.b	$8E,$93,$E1,$9F,$DD,$1A,$BF,$C2,$BE,$0B,$F9,$97,$FB,$8B,$F2,$BF
	.dc.b	$D0,$67,$47,$DB,$74,$77,$35,$7D,$CF,$DE,$C7,$73,$F4,$67,$DA,$39
	.dc.b	$47,$D3,$7E,$3C,$FB,$5F,$8E,$7C,$23,$E6,$4F,$C2,$7E,$C1,$A7,$E8
	.dc.b	$9F,$88,$9F,$1C,$F3,$9F,$82,$73,$9D,$5F,$F2,$FA,$FA,$11,$CA,$BE
	.dc.b	$0C,$E6,$FE,$2D,$79,$EB,$C5,$7F,$83,$8F,$C7,$3F,$C0,$AB,$F6,$D7
	.dc.b	$F4,$8F,$15,$7C,$49,$F1,$47,$6A,$FE,$48,$FF,$91,$63,$F2,$47,$33
	.dc.b	$F0,$4F,$37,$E0,$9E,$C9,$F4,$4F,$D3,$AF,$44,$FC,$E8,$FC,$D3,$F3
	.dc.b	$8E,$33,$E3,$38,$CF,$E1,$1F,$B5,$5E,$39,$FC,$71,$E8,$AE,$C7,$A0
	.dc.b	$FF,$3C,$3E,$87,$F8,$25,$7C,$27,$E9,$C7,$82,$7A,$57,$F2,$2A,$FE
	.dc.b	$31,$3F,$86,$3F,$1E,$7E,$7A,$34,$FE,$2C,$FE,$1C,$7D,$13,$C1,$1C
	.dc.b	$9C,$D7,$FB,$1C,$FD,$BB,$FB,$C8,$F2,$1F,$19,$ED,$5F,$BB,$3F,$86
	.dc.b	$CF,$B8,$FE,$75,$1F,$8F,$8C,$CF,$C7,$7E,$E9,$EC,$39,$4F,$E9,$A3
	.dc.b	$53,$F0,$63,$E3,$BF,$32,$BE,$8B,$F0,$5F,$A3,$1A,$7E,$0C,$7D,$19
	.dc.b	$E6,$FC,$13,$95,$79,$1E,$4F,$92,$B9,$3E,$47,$F3,$4F,$E2,$E3,$95
	.dc.b	$76,$3C,$6F,$CA,$BF,$F2,$D9,$FE,$41,$7F,$28,$7E,$D2,$79,$57,$CF
	.dc.b	$3F,$46,$7E,$0A,$3F,$32,$FE,$09,$EF,$1E,$38,$FC,$AD,$7E,$B1,$E5
	.dc.b	$3F,$70,$FC,$01,$FC,$C9,$CB,$F0,$1F,$9C,$FC,$03,$E7,$57,$F9,$21
	.dc.b	$F9,$06,$4F,$4C,$FF,$02,$9F,$C2,$C6,$2B,$F6,$91,$F8,$08,$FD,$74
	.dc.b	$7C,$D7,$F1,$B7,$F8,$F7,$C8,$D7,$F5,$C7,$B8,$FC,$81,$F0,$9F,$25
	.dc.b	$7C,$E8,$E2,$7B,$8E,$35,$EE,$8F,$A9,$5E,$19,$D9,$EC,$3C,$67,$18
	.dc.b	$F9,$57,$F6,$23,$FC,$C6,$7B,$6B,$F3,$8F,$CD,$39,$1D,$27,$C8,$F8
	.dc.b	$A7,$93,$FE,$3E,$39,$5F,$E4,$AB,$EF,$47,$E0,$DF,$BC,$62,$BE,$73
	.dc.b	$F2,$DF,$E2,$17,$EB,$9F,$7D,$7C,$99,$FC,$13,$67,$F1,$B7,$E2,$9F
	.dc.b	$F0,$8C,$E6,$FB,$B4,$F9,$B5,$D6,$7E,$1D,$FF,$A2,$6F,$F3,$57,$E5
	.dc.b	$AF,$BF,$3F,$0E,$BE,$09,$E2,$7E,$64,$78,$9E,$47,$CC,$9E,$47,$CC
	.dc.b	$7F,$28,$FE,$B8,$F2,$BC,$9F,$8B,$5E,$B9,$FF,$10,$CF,$D2,$3E,$19
	.dc.b	$BB,$F5,$D7,$E7,$A3,$F6,$87,$88,$FA,$8F,$8A,$3E,$A1,$F0,$DF,$CE
	.dc.b	$99,$8F,$D1,$3F,$8D,$9F,$D9,$4F,$F2,$48,$DB,$C5,$FD,$DB,$67,$FB
	.dc.b	$D1,$FC,$A0,$FC,$E9,$F8,$17,$CF,$1F,$05,$AF,$EF,$A7,$E5,$4F,$F9
	.dc.b	$DE,$BE,$9D,$75,$8F,$93,$5B,$3E,$4B,$F1,$6F,$ED,$BC,$4F,$A5,$3C
	.dc.b	$63,$F4,$F3,$F1,$E7,$F3,$97,$F8,$33,$C3,$3F,$EC,$4B,$F0,$CF,$CF
	.dc.b	$AF,$82,$D7,$E0,$B4,$F8,$27,$F4,$37,$B3,$F6,$0F,$35,$7F,$24,$BF
	.dc.b	$23,$FC,$CE,$BE,$1C,$7C,$FA,$EB,$5C,$1F,$7D,$FE,$4A,$FC,$E2,$20
	.dc.b	$BF,$4D,$71,$BE,$AF,$F0,$F9,$F5,$B9,$7F,$BC,$4E,$A3,$EC,$46,$AF
	.dc.b	$C0,$F8,$AF,$E4,$3F,$98,$9F,$C8,$4F,$EC,$9F,$88,$7D,$8B,$F6,$3F
	.dc.b	$AB,$AE,$31,$FD,$BE,$3E,$45,$7E,$14,$F0,$4E,$9A,$7C,$CA,$FB,$E7
	.dc.b	$D7,$9F,$BE,$78,$1F,$BF,$1C,$CF,$96,$FA,$AB,$C1,$3F,$CE,$A7,$A5
	.dc.b	$7D,$58,$E4,$FE,$89,$F9,$11,$FC,$5E,$3F,$BE,$47,$E7,$AB,$F8,$DB
	.dc.b	$ED,$9E,$67,$C7,$2D,$78,$27,$F8,$FC,$60,$F6,$9E,$79,$E4,$D7,$FC
	.dc.b	$42,$7C,$67,$F4,$53,$F5,$E7,$C3,$3F,$3C,$F3,$9F,$94,$3D,$D3,$CE
	.dc.b	$3E,$2B,$C8,$F0,$94,$F7,$1F,$E0,$68,$F7,$57,$F2,$47,$B1,$F8,$68
	.dc.b	$F9,$B1,$F5,$DE,$37,$E5,$8E,$15,$E0,$BF,$C1,$DF,$28,$F4,$9F,$E1
	.dc.b	$EA,$E1,$7F,$CA,$63,$84,$FB,$0E,$17,$D0,$FD,$54,$FA,$AF,$AD,$70
	.dc.b	$AF,$31,$F3,$27,$DC,$FC,$58,$FA,$71,$F1,$6B,$F0,$13,$C0,$EF,$1E
	.dc.b	$38,$EE,$FC,$03,$ED,$14,$E5,$1F,$21,$E0,$FF,$84,$9F,$33,$EB,$9F
	.dc.b	$31,$F0,$A7,$C2,$D8,$E1,$3D,$4F,$4B,$E4,$AF,$E7,$AF,$E6,$0D,$9E
	.dc.b	$F7,$E1,$D7,$43,$E4,$47,$96,$3F,$BC,$5F,$F5,$1A,$FF,$50,$D7,$DB
	.dc.b	$AE,$47,$9D,$E5,$1E,$C8,$FF,$0E,$1F,$E0,$23,$FA,$5D,$7F,$1E,$7B
	.dc.b	$1F,$D3,$AB,$B1,$FE,$AA,$9F,$5D,$FC,$2B,$F2,$D7,$33,$FA,$E9,$FC
	.dc.b	$F4,$FE,$BB,$7D,$5D,$1F,$CF,$2B,$E9,$BF,$76,$7F,$6E,$E8,$FE,$68
	.dc.b	$0F,$E5,$30,$9C,$44,$8C,$38,$31,$1F,$CC,$1C,$1C,$E3,$6F,$17,$F9
	.dc.b	$B4,$69,$CC,$73,$7F,$DD,$8D,$8D,$47,$08,$C1,$D8,$E6,$FF,$41,$70
	.dc.b	$F4,$7D,$71,$DC,$E0,$F0,$8A,$6C,$FE,$93,$1B,$8E,$6F,$B5,$C1,$EB
	.dc.b	$8E,$11,$C8,$D4,$79,$8D,$3C,$1F,$64,$79,$DC,$1E,$63,$81,$A7,$A3
	.dc.b	$F1,$5F,$84,$FD,$03,$CA,$F3,$78,$9C,$CD,$1F,$05,$D3,$83,$EB,$3D
	.dc.b	$1F,$B1,$1E,$B8,$DC,$60,$F1,$3F,$01,$F0,$BE,$67,$FB,$53,$FE,$9E
	.dc.b	$FE,$01,$FC,$4D,$AF,$F1,$B1,$87,$FB,$C3,$F8,$88,$D4,$A3,$F9,$0D
	.dc.b	$FF,$9A,$DE,$11,$FD,$F9,$A7,$1A,$E7,$1F,$C3,$2F,$F8,$C5,$F8,$5F
	.dc.b	$CA,$57,$29,$F0,$C7,$E8,$67,$C3,$3C,$E7,$C4,$73,$8F,$E2,$53,$C0
	.dc.b	$3E,$88,$C3,$CA,$3F,$4C,$F4,$8E,$55,$FB,$C7,$DD,$3E,$18,$F8,$AD
	.dc.b	$8C,$BE,$E7,$89,$FC,$3C,$CC,$F8,$5B,$3E,$59,$E5,$5F,$29,$FF,$2E
	.dc.b	$BF,$06,$76,$76,$8D,$BD,$AF,$E8,$47,$8E,$3D,$B5,$D6,$3E,$0B,$F4
	.dc.b	$23,$F9,$A4,$FF,$2E,$9F,$9E,$71,$8E,$F5,$F3,$4F,$24,$7D,$C9,$FD
	.dc.b	$75,$FD,$78,$EF,$5F,$7D,$F5,$9F,$32,$FC,$0F,$F3,$43,$EF,$BF,$C5
	.dc.b	$EF,$DC,$7E,$1A,$38,$4F,$DC,$3C,$EF,$F3,$B3,$CE,$7B,$1F,$03,$63
	.dc.b	$D1,$1F,$16,$3E,$59,$ED,$8E,$CF,$D7,$8E,$AF,$90,$EB,$5F,$5E,$77
	.dc.b	$1F,$42,$77,$3F,$8E,$7E,$A9,$F0,$A7,$C2,$EA,$BF,$15,$1E,$EA,$F8
	.dc.b	$A7,$84,$F8,$95,$CE,$35,$1E,$B3,$E1,$BF,$49,$FC,$85,$FD,$2A,$F9
	.dc.b	$51,$ED,$3A,$9E,$73,$AB,$E2,$74,$75,$3C,$47,$F7,$53,$E3,$9F,$36
	.dc.b	$FD,$8F,$89,$F9,$8F,$E0,$DF,$F5,$AD,$7A,$2F,$F5,$93,$D1,$FE,$CD
	.dc.b	$3F,$A3,$9D,$57,$E5,$67,$F6,$13,$F1,$CD,$57,$9A,$39,$BE,$B3,$CC
	.dc.b	$78,$63,$F8,$D3,$B9,$E8,$7E,$41,$D9,$F2,$23,$CE,$ED,$FD,$63,$F8
	.dc.b	$97,$DA,$ED,$EA,$FC,$37,$F4,$AF,$C3,$7D,$17,$FB,$28,$D5,$FE,$21
	.dc.b	$E9,$3A,$7E,$6D,$FE,$35,$FA,$E6,$8F,$AF,$1F,$09,$E6,$6A,$3E,$3B
	.dc.b	$AA,$F0,$47,$38,$FB,$87,$C7,$7E,$8C,$6A,$FF,$8C,$CE,$8F,$01,$F1
	.dc.b	$8F,$D6,$47,$E3,$5F,$D1,$C7,$09,$FA,$71,$C1,$EA,$FB,$23,$47,$D2
	.dc.b	$AF,$DD,$47,$9C,$D3,$FB,$89,$E6,$7B,$23,$C4,$FF,$4E,$AF,$14,$7F
	.dc.b	$0E,$9F,$C9,$BF,$8D,$AF,$F1,$EC,$FB,$67,$A4,$78,$6F,$F5,$15,$EC
	.dc.b	$9F,$C0,$BF,$30,$F8,$57,$E1,$3D,$6F,$E3,$4F,$E8,$07,$C2,$8E,$A7
	.dc.b	$E2,$23,$85,$78,$A7,$DB,$3E,$67,$E9,$46,$5E,$87,$F3,$43,$A4,$6D
	.dc.b	$F9,$91,$B7,$F2,$47,$88,$F8,$51,$F9,$27,$E1,$1F,$E5,$83,$AC,$FF
	.dc.b	$35,$8F,$E2,$A6,$E3,$C6,$6C,$EA,$78,$CE,$B5,$F9,$99,$F6,$C6,$CF
	.dc.b	$95,$3F,$A4,$8F,$C8,$1F,$6A,$BF,$60,$FC,$A9,$EA,$F8,$EB,$F8,$0D
	.dc.b	$7C,$9A,$D9,$EF,$AE,$A7,$BE,$3D,$33,$E3,$9D,$CF,$8E,$BF,$0D,$5F
	.dc.b	$71,$F1,$CF,$E0,$63,$F7,$73,$D0,$F7,$3D,$1F,$D6,$47,$47,$FC,$E1
	.dc.b	$3D,$27,$DD,$3F,$9C,$7D,$07,$E7,$27,$F2,$F1,$FE,$57,$7D,$15,$EE
	.dc.b	$3F,$A1,$BF,$A1,$3E,$AC,$7F,$38,$7F,$25,$5F,$26,$BA,$3F,$26,$3D
	.dc.b	$0F,$C9,$7F,$8C,$57,$47,$C3,$3F,$E8,$33,$E1,$57,$9E,$75,$5C,$CF
	.dc.b	$65,$79,$8E,$67,$F1,$B3,$CC,$7D,$89,$F3,$47,$CC,$AF,$83,$5F,$AC
	.dc.b	$8E,$75,$F3,$23,$E0,$3C,$CF,$2D,$FE,$EA,$BE,$84,$7E,$C2,$FE,$84
	.dc.b	$7F,$A4,$AF,$FC,$4D,$5F,$E1,$5A,$FC,$99,$F5,$A7,$E2,$D7,$E1,$EB
	.dc.b	$C2,$7C,$59,$F9,$71,$F8,$19,$FE,$61,$1F,$40,$FD,$7B,$FA,$37,$F5
	.dc.b	$E7,$E2,$E3,$ED,$3E,$B3,$E2,$9F,$AE,$3F,$0A,$EA,$7C,$07,$C5,$BF
	.dc.b	$9F,$1E,$53,$47,$CA,$3E,$A4,$FC,$59,$FA,$91,$A8,$F2,$CF,$E7,$EB
	.dc.b	$DF,$1E,$57,$EF,$1F,$94,$3F,$89,$C7,$82,$35,$7F,$3E,$3E,$9C,$7A
	.dc.b	$67,$E7,$57,$FB,$9E,$3D,$C7,$C8,$8F,$BA,$7E,$A2,$FE,$6C,$FA,$63
	.dc.b	$BC,$FA,$63,$DD,$1E,$48,$FC,$61,$CE,$BC,$8F,$C8,$8D,$3E,$0A,$F9
	.dc.b	$0F,$33,$D3,$1F,$9A,$3B,$BE,$9B,$FF,$75,$47,$E2,$A7,$F5,$15,$F9
	.dc.b	$3B,$F9,$57,$F8,$83,$F1,$D5,$F2,$9F,$15,$FC,$57,$EE,$9F,$C9,$0C
	.dc.b	$C7,$84,$FE,$69,$7F,$C8,$A3,$31,$F4,$4C,$D7,$E6,$AB,$CB,$7F,$42
	.dc.b	$BA,$57,$D0,$3F,$DF,$EF,$C1,$AF,$C1,$BE,$D3,$FC,$E5,$5F,$A1,$9F
	.dc.b	$74,$FE,$6E,$7F,$6A,$7F,$04,$7E,$A4,$78,$8F,$71,$DC,$FF,$69,$DF
	.dc.b	$F9,$36,$7D,$E7,$E8,$0F,$59,$F6,$1F,$A6,$F8,$CF,$F6,$39,$99,$F3
	.dc.b	$BE,$CA,$F7,$BD,$27,$E3,$57,$A0,$FC,$59,$F1,$CF,$D8,$3A,$9E,$33
	.dc.b	$F8,$73,$F0,$47,$D9,$8F,$BB,$3F,$4C,$E3,$5F,$3E,$7E,$39,$FE,$85
	.dc.b	$38,$3F,$1D,$F9,$D1,$F8,$A7,$E9,$DF,$D5,$7E,$14,$7C,$73,$EB,$5F
	.dc.b	$F8,$2D,$F9,$93,$F8,$79,$F3,$1F,$AB,$79,$5F,$AE,$3C,$67,$FA,$2C
	.dc.b	$F1,$CF,$C2,$7E,$53,$E3,$7C,$33,$CA,$FD,$91,$F4,$0F,$4C,$F8,$8F
	.dc.b	$85,$3F,$C1,$4F,$25,$7D,$89,$FA,$87,$89,$FC,$03,$E8,$3E,$09,$F5
	.dc.b	$EB,$F9,$8C,$F8,$6B,$F1,$37,$F5,$2F,$A5,$F9,$8E,$13,$F8,$EA,$E0
	.dc.b	$7E,$C5,$FD,$EC,$FF,$24,$BF,$C3,$D7,$13,$C4,$58,$EA,$7C,$97,$E0
	.dc.b	$3C,$67,$E3,$3F,$8C,$AE,$F1,$F0,$A7,$FA,$24,$78,$6F,$E1,$DF,$82
	.dc.b	$F8,$57,$A2,$B3,$7D,$2B,$E7,$D7,$F5,$0A,$FC,$54,$FE,$66,$BF,$8C
	.dc.b	$B6,$3F,$2A,$D7,$F4,$D3,$C5,$FC,$04,$65,$F9,$E7,$C0,$8F,$1B,$6B
	.dc.b	$C2,$7F,$33,$8F,$81,$3F,$7C,$FD,$45,$7D,$F7,$E0,$57,$F1,$C8,$F8
	.dc.b	$11,$89,$FD,$3B,$C2,$7E,$2B,$9A,$FD,$24,$FF,$50,$8F,$AF,$1F,$C2
	.dc.b	$9F,$D6,$9F,$76,$7E,$43,$C4,$FC,$D3,$FD,$F1,$C5,$7C,$63,$EA,$BE
	.dc.b	$18,$F3,$3F,$D2,$E7,$F9,$AB,$F1,$63,$F0,$51,$F5,$CF,$C9,$BE,$08
	.dc.b	$FE,$F2,$FF,$84,$CF,$EC,$A7,$B1,$ED,$5F,$02,$3F,$5B,$1F,$25,$FE
	.dc.b	$4F,$1F,$D6,$9E,$27,$92,$38,$CF,$C4,$3E,$15,$7C,$F8,$F0,$CF,$F2
	.dc.b	$F3,$8D,$77,$7F,$6B,$1F,$22,$7F,$1A,$F8,$63,$E3,$3D,$8F,$93,$3F
	.dc.b	$79,$E0,$7C,$A8,$FC,$6D,$FE,$C6,$FE,$24,$FD,$39,$FC,$43,$6B,$F5
	.dc.b	$33,$F8,$A8,$FE,$9B,$1D,$6F,$D5,$5F,$12,$7B,$9E,$38,$FC,$ED,$F7
	.dc.b	$8F,$D2,$1D,$6F,$C3,$3E,$C8,$E3,$5F,$32,$7E,$DC,$FE,$02,$3B,$3E
	.dc.b	$07,$E2,$B6,$FF,$21,$7C,$E3,$15,$F4,$2B,$EE,$1C,$8A,$7C,$A3,$07
	.dc.b	$D9,$2D,$7D,$33,$E0,$3C,$1E,$F7,$F3,$63,$F9,$C1,$F8,$D3,$8C,$FE
	.dc.b	$05,$FF,$08,$C7,$C5,$7F,$7F,$5E,$D7,$D6,$75,$9F,$F1,$0D,$FA,$E3
	.dc.b	$07,$E1,$A3,$A3,$FD,$0A,$7E,$7B,$FC,$D6,$15,$FE,$FF,$BF,$B6,$FD
	.dc.b	$13,$E9,$D7,$5A,$F8,$07,$C3,$3C,$66,$27,$E0,$B6,$BE,$63,$4F,$69
	.dc.b	$F0,$2B,$F2,$71,$A9,$F9,$4D,$3D,$31,$FD,$96,$FF,$C2,$97,$F0,$EB
	.dc.b	$B9,$C2,$FE,$05,$74,$7B,$9E,$68,$D3,$F0,$CF,$9A,$7D,$33,$D8,$F4
	.dc.b	$BF,$0D,$77,$7C,$95,$D6,$BF,$C3,$27,$58,$DB,$CE,$B9,$4E,$CF,$54
	.dc.b	$7F,$3E,$AE,$8F,$A1,$FE,$34,$F1,$AF,$BB,$3F,$89,$8F,$E7,$60,$FB
	.dc.b	$2F,$DA,$9C,$47,$FA,$F4,$A6,$63,$EF,$B8,$39,$28,$C3,$D1,$A6,$E3
	.dc.b	$6F,$81,$C4,$71,$7B,$BC,$5B,$1C,$5F,$14,$53,$83,$E8,$78,$3E,$93
	.dc.b	$11,$DD,$F1,$9D,$A3,$E0,$3D,$8F,$1C,$6D,$C1,$D6,$36,$F5,$3D,$6F
	.dc.b	$C4,$7A,$9B,$7C,$AF,$C0,$3A,$BB,$7A,$9F,$29,$EA,$EC,$F9,$2E,$DF
	.dc.b	$63,$EE,$74,$7C,$67,$4E,$0F,$BA,$F8,$9E,$A7,$C4,$76,$E1,$E8,$E8
	.dc.b	$F6,$BF,$3E,$3E,$EB,$B7,$0F,$23,$FB,$25,$7F,$9A,$BA,$3F,$16,$AF
	.dc.b	$C7,$18,$7F,$84,$BF,$8B,$8D,$4A,$3F,$84,$47,$EC,$23,$17,$F9,$98
	.dc.b	$DC,$72,$8C,$55,$7B,$47,$33,$F4,$4F,$F9,$A2,$B3,$1B,$AF,$0C,$78
	.dc.b	$EB,$CB,$3C,$27,$D9,$5D,$A3,$F5,$35,$D8,$E0,$F7,$3F,$99,$47,$FB
	.dc.b	$66,$3B,$C7,$AA,$13,$E0,$7F,$5E,$F8,$A3,$C1,$5F,$A0,$9F,$69,$FB
	.dc.b	$0A,$E6,$78,$5E,$31,$E8,$9F,$0C,$7F,$AE,$A7,$6F,$A5,$F3,$9F,$3A
	.dc.b	$7F,$93,$BA,$8F,$2C,$7C,$B7,$84,$F9,$0F,$F5,$BC,$EE,$BD,$53,$F0
	.dc.b	$E7,$57,$FB,$19,$FC,$14,$FA,$4F,$95,$3F,$46,$FB,$47,$82,$3E,$53
	.dc.b	$FC,$FE,$BF,$CB,$77,$DE,$B7,$7F,$1E,$FD,$F7,$F7,$4E,$AF,$CF,$8F
	.dc.b	$2C,$7A,$A3,$71,$F7,$23,$C1,$1E,$33,$EB,$3B,$3C,$47,$F9,$AA,$BC
	.dc.b	$51,$A8,$F4,$47,$CB,$8F,$E7,$0E,$8F,$9E,$FC,$43,$F8,$44,$FC,$49
	.dc.b	$F0,$47,$86,$76,$F9,$AB,$F0,$53,$F5,$0F,$94,$7A,$23,$DD,$1F,$10
	.dc.b	$FC,$14,$7A,$0F,$C9,$1F,$12,$7F,$10,$FD,$79,$EC,$7D,$C7,$47,$E1
	.dc.b	$CF,$C6,$BF,$51,$FE,$B0,$FB,$9D,$9F,$08,$DC,$7C,$3B,$FB,$31,$AB
	.dc.b	$FA,$6F,$C8,$AF,$88,$7F,$05,$7D,$B1,$FB,$E8,$F6,$9F,$3D,$FC,$04
	.dc.b	$7B,$4F,$CF,$1D,$6F,$E1,$4F,$9E,$7F,$B5,$D7,$8A,$BE,$45,$FC,$27
	.dc.b	$EF,$1E,$C8,$D4,$F8,$67,$53,$E7,$AF,$8F,$5F,$9D,$BF,$EA,$55,$F7
	.dc.b	$EB,$E0,$D7,$B1,$FD,$C9,$F7,$9F,$5C,$7D,$E8,$F5,$1F,$99,$3C,$A7
	.dc.b	$D1,$3E,$B1,$AB,$FC,$C1,$E5,$74,$7A,$9F,$7B,$F9,$73,$E2,$4F,$C9
	.dc.b	$AF,$C2,$C7,$BC,$D1,$E0,$3E,$D1,$E0,$34,$7E,$28,$D4,$FD,$A3,$E9
	.dc.b	$BA,$3E,$99,$A8,$FA,$75,$F8,$38,$EF,$3E,$97,$B9,$E9,$9F,$F1,$3D
	.dc.b	$FD,$27,$FA,$4C,$7B,$8F,$A4,$7C,$37,$F4,$B3,$FA,$BB,$FC,$39,$FA
	.dc.b	$B9,$FA,$6F,$D8,$3F,$63,$3F,$62,$3E,$2D,$FD,$37,$D4,$7D,$3A,$F3
	.dc.b	$1E,$F9,$F9,$2F,$CE,$8F,$88,$EA,$3D,$F5,$F8,$19,$FD,$C1,$F8,$1A
	.dc.b	$FC,$5D,$7C,$83,$94,$7C,$87,$94,$FD,$73,$F3,$55,$F1,$27,$E4,$BF
	.dc.b	$11,$F9,$27,$C4,$3D,$33,$F8,$68,$D3,$E9,$9E,$73,$FE,$49,$39,$BF
	.dc.b	$D7,$1F,$9D,$1E,$59,$F0,$47,$B9,$E6,$7B,$A7,$E6,$1F,$B9,$8F,$71
	.dc.b	$CA,$3E,$39,$C9,$F7,$3C,$A3,$DD,$5F,$85,$9E,$D1,$E3,$7D,$B5,$E3
	.dc.b	$7D,$06,$A3,$EF,$3F,$97,$8F,$96,$FD,$9A,$ED,$5F,$76,$72,$7C,$5A
	.dc.b	$EC,$FC,$58,$F1,$57,$9A,$73,$5E,$F8,$FC,$84,$7F,$27,$7F,$3C,$FA
	.dc.b	$8F,$C0,$57,$E4,$EB,$F8,$24,$F9,$6B,$E1,$3E,$5A,$EA,$71,$3E,$5B
	.dc.b	$FC,$21,$FC,$D9,$FC,$21,$FC,$D9,$E2,$BF,$E2,$4F,$F2,$89,$F7,$BA
	.dc.b	$3F,$06,$FE,$98,$FD,$03,$E8,$AF,$89,$1E,$4B,$F8,$97,$F3,$2B,$F6
	.dc.b	$87,$D0,$3E,$23,$97,$F2,$11,$F3,$27,$B4,$78,$E7,$DB,$5D,$A3,$91
	.dc.b	$F0,$63,$95,$7D,$D8,$FF,$57,$1F,$76,$3E,$03,$F2,$4E,$31,$E6,$8C
	.dc.b	$C7,$98,$E2,$79,$9C,$BE,$6A,$F7,$1E,$67,$F2,$E7,$E3,$9F,$CB,$BD
	.dc.b	$9D,$47,$D2,$AF,$DC,$C6,$A7,$B9,$EB,$9E,$F5,$EB,$7F,$CC,$B3,$CA
	.dc.b	$BF,$07,$5E,$79,$C9,$F8,$BA,$FE,$03,$3F,$1E,$7A,$C7,$BE,$3C,$33
	.dc.b	$FB,$A9,$CD,$75,$7F,$8D,$9F,$81,$AF,$A9,$3D,$67,$F0,$B3,$F2,$CE
	.dc.b	$31,$93,$EB,$BC,$63,$E3,$C6,$4F,$AE,$FC,$17,$D3,$7F,$92,$9F,$97
	.dc.b	$5F,$AB,$9F,$D1,$D7,$E4,$23,$DB,$3D,$4F,$9D,$3D,$4F,$93,$3D,$6F
	.dc.b	$E3,$CF,$D9,$3F,$19,$3E,$23,$E1,$1E,$23,$F9,$EB,$A7,$99,$A8,$E7
	.dc.b	$5F,$55,$B5,$EB,$BF,$F7,$6D,$FA,$A3,$F8,$03,$FC,$EA,$FF,$04,$FF
	.dc.b	$1D,$7A,$3F,$40,$F9,$8F,$8C,$FE,$48,$70,$7F,$1C,$F0,$7F,$64,$7C
	.dc.b	$33,$C0,$FF,$03,$7F,$7C,$FB,$1C,$D7,$E1,$63,$A5,$FC,$F7,$BC,$7B
	.dc.b	$1F,$8B,$7E,$C7,$77,$EC,$AF,$E2,$D7,$E9,$3F,$29,$7C,$4F,$E4,$6F
	.dc.b	$E7,$5F,$C8,$9F,$7E,$B9,$9E,$78,$F8,$D1,$C6,$3F,$03,$1F,$4E,$38
	.dc.b	$C7,$AC,$E2,$FA,$CE,$0F,$F2,$28,$E6,$79,$4F,$C1,$1F,$08,$F9,$0F
	.dc.b	$26,$DF,$BE,$79,$C7,$F2,$B6,$BD,$0F,$A2,$7D,$88,$E0,$F7,$AF,$C3
	.dc.b	$36,$FF,$28,$DB,$F9,$AF,$FA,$42,$BF,$DA,$27,$F8,$66,$3F,$57,$1F
	.dc.b	$84,$BF,$0D,$FD,$08,$F0,$9F,$C0,$EB,$C2,$E0,$F7,$D7,$B5,$E4,$78
	.dc.b	$6B,$EA,$D7,$F0,$58,$FC,$A4,$7D,$53,$BD,$FE,$C4,$FE,$03,$1D,$CF
	.dc.b	$41,$FE,$DD,$BF,$45,$FE,$B6,$B9,$4F,$78,$FB,$11,$D2,$3F,$73,$3A
	.dc.b	$7D,$17,$F2,$1B,$7F,$21,$B7,$F2,$1F,$F4,$D5,$7C,$87,$FC,$27,$1E
	.dc.b	$13,$BB,$E1,$7F,$3A,$D3,$F4,$AE,$1E,$B7,$E3,$BF,$E1,$F1,$F2,$0C
	.dc.b	$CF,$E5,$1F,$C5,$19,$8D,$BF,$A2,$3B,$1F,$DA,$A3,$F8,$B4,$FE,$69
	.dc.b	$F0,$4F,$09,$FA,$F3,$F1,$5F,$E5,$B5,$E3,$BF,$A3,$1A,$3A,$D7,$E0
	.dc.b	$CE,$47,$F4,$99,$FC,$1B,$89,$F9,$4D,$BF,$E1,$4D,$BF,$C1,$CF,$F4
	.dc.b	$3B,$F7,$D7,$9A,$3C,$35,$F5,$4E,$11,$F0,$CF,$C7,$4F,$C1,$38,$9F
	.dc.b	$A9,$3C,$F3,$FD,$12,$FD,$93,$F9,$E9,$E1,$5D,$23,$C0,$66,$3F,$37
	.dc.b	$3E,$CA,$F1,$9E,$C2,$CF,$B1,$B7,$EC,$8F,$A3,$5F,$9E,$7F,$0A,$F9
	.dc.b	$8F,$DE,$5F,$85,$FC,$29,$E9,$34,$FC,$FA,$EB,$3C,$A3,$84,$FD,$FB
	.dc.b	$F6,$CF,$E4,$CF,$59,$63,$A4,$F0,$3F,$B7,$CF,$09,$FE,$A3,$74,$FC
	.dc.b	$45,$70,$BF,$0C,$FF,$1F,$BF,$23,$F1,$DF,$AD,$1A,$9E,$13,$98,$FE
	.dc.b	$B1,$1F,$BA,$12,$84,$FF,$BD,$E7,$E0,$C7,$F0,$38,$E4,$7C,$F2,$D7
	.dc.b	$CF,$BF,$45,$7C,$F9,$F8,$27,$B0,$F9,$15,$E5,$8F,$91,$18,$39,$94
	.dc.b	$F9,$0F,$D6,$8E,$17,$F2,$27,$47,$53,$BD,$FC,$8A,$E1,$1D,$0F,$91
	.dc.b	$5E,$B7,$AD,$7C,$A6,$C6,$5E,$93,$F8,$43,$9B,$A7,$FA,$B1,$EF,$3E
	.dc.b	$7B,$97,$F7,$6F,$9A,$BC,$95,$DE,$3A,$CF,$46,$C7,$B5,$F5,$C7,$53
	.dc.b	$53,$CA,$BC,$CD,$8F,$E5,$B3,$DD,$E1,$5D,$E7,$51,$FC,$68,$FE,$78
	.dc.b	$7F,$2C,$3F,$9E,$1F,$CB,$27,$C7,$5D,$A3,$9B,$5F,$F0,$AC,$FE,$06
	.dc.b	$7C,$11,$86,$D7,$F8,$3E,$3E,$2C,$74,$7C,$EF,$B9,$E0,$FF,$35,$47
	.dc.b	$FB,$64,$AA,$73,$10,$CB,$F9,$A8,$C4,$7F,$A6,$8E,$2F,$E7,$A3,$2F
	.dc.b	$38,$D9,$97,$F4,$6E,$1D,$3E,$07,$FD,$A0,$F8,$0D,$47,$85,$A7,$96
	.dc.b	$36,$F8,$4D,$BD,$9F,$1C,$79,$23,$51,$E9,$3C,$11,$D5,$D9,$D6,$36
	.dc.b	$F8,$CE,$6F,$57,$C5,$1D,$63,$D7,$1B,$32,$ED,$F0,$3E,$83,$6F,$89
	.dc.b	$D3,$C4,$FA,$C7,$88,$D3,$F3,$1C,$3C,$E3,$0F,$DB,$39,$C6,$9D,$C6
	.dc.b	$8F,$81,$1E,$A3,$E0,$47,$90,$F2,$BF,$82,$79,$9F,$D5,$9F,$F1,$2B
	.dc.b	$F8,$89,$FE,$40,$FE,$1C,$C5,$76,$9F,$E4,$AE,$63,$F2,$D3,$F9,$B9
	.dc.b	$FF,$3F,$47,$03,$93,$F9,$88,$EB,$1F,$92,$6B,$FA,$49,$FF,$36,$1F
	.dc.b	$8B,$79,$1F,$76,$38,$CF,$E7,$5E,$31,$E2,$9E,$33,$F9,$C3,$C0,$F9
	.dc.b	$DB,$1F,$6E,$78,$C7,$EC,$AB,$F8,$1C,$FE,$46,$73,$3F,$02,$38,$57
	.dc.b	$38,$FE,$22,$7F,$02,$3C,$06,$E3,$D1,$1F,$26,$79,$9F,$C4,$67,$81
	.dc.b	$A9,$F8,$EF,$13,$DD,$3D,$E7,$84,$7F,$04,$3D,$A7,$81,$EB,$1D,$CE
	.dc.b	$07,$F1,$57,$EB,$1F,$D7,$63,$84,$6A,$7D,$A7,$8A,$7B,$9C,$23,$EA
	.dc.b	$4F,$CA,$9F,$41,$F6,$E3,$BC,$7F,$42,$3C,$C7,$E1,$0E,$11,$F3,$63
	.dc.b	$CC,$F7,$7C,$C7,$DA,$8E,$B1,$B3,$CA,$7F,$6C,$7A,$CF,$AD,$DC,$7E
	.dc.b	$21,$F9,$2F,$57,$C9,$1E,$B7,$E0,$C7,$84,$EA,$7E,$4C,$FF,$5A,$47
	.dc.b	$57,$C4,$7C,$08,$EA,$F8,$E3,$A4,$78,$8F,$8C,$6A,$3A,$57,$FB,$45
	.dc.b	$FC,$E9,$D2,$34,$FE,$6C,$F0,$D7,$D0,$8F,$C6,$46,$DE,$2E,$E7,$A4
	.dc.b	$7C,$E8,$F8,$4E,$8F,$CF,$1F,$90,$7D,$67,$37,$81,$A3,$E6,$CF,$86
	.dc.b	$BF,$1B,$3F,$18,$F0,$4F,$FB,$29,$EC,$FE,$64,$B1,$F8,$A9,$FB,$93
	.dc.b	$E0,$3C,$B3,$E0,$3F,$56,$78,$0F,$AA,$FC,$F7,$9C,$7E,$1A,$3E,$2C
	.dc.b	$69,$F5,$3F,$19,$F2,$BF,$C4,$EB,$FD,$E1,$1C,$8F,$54,$7B,$EB,$EC
	.dc.b	$C7,$96,$3E,$C9,$E0,$3F,$13,$5E,$A8,$E4,$7E,$84,$FC,$74,$7F,$82
	.dc.b	$CF,$E5,$A7,$B6,$7F,$8C,$BC,$CD,$3F,$CC,$2B,$CB,$1B,$AF,$D2,$D7
	.dc.b	$E3,$5F,$04,$66,$3D,$2F,$C8,$AF,$4C,$79,$23,$F1,$67,$B6,$78,$BF
	.dc.b	$14,$F6,$9D,$CF,$6C,$7C,$C7,$F3,$47,$DB,$34,$7A,$23,$FC,$09,$1F
	.dc.b	$D1,$A3,$E9,$3F,$71,$EC,$79,$63,$C3,$3F,$8A,$9F,$90,$7B,$4F,$01
	.dc.b	$F4,$67,$C0,$72,$8F,$C5,$46,$8F,$74,$7C,$B7,$FA,$E4,$7C,$B7,$F8
	.dc.b	$74,$79,$0F,$79,$CE,$7F,$15,$5F,$A8,$8C,$4F,$CC,$AE,$F3,$97,$D1
	.dc.b	$3F,$30,$F6,$4F,$8D,$FA,$8F,$B0,$FD,$FB,$FA,$53,$27,$18,$EA,$E4
	.dc.b	$F0,$C7,$13,$F3,$2F,$F5,$89,$FC,$2B,$A3,$95,$7A,$9E,$55,$E6,$7E
	.dc.b	$24,$FF,$BA,$EB,$E6,$4F,$E4,$A7,$ED,$C7,$C8,$AF,$B7,$1F,$D3,$0F
	.dc.b	$C0,$CE,$DF,$90,$7E,$1A,$BE,$24,$FF,$49,$AF,$44,$FE,$3D,$F8,$8F
	.dc.b	$9C,$F1,$9F,$EE,$23,$EA,$1F,$7A,$BF,$AC,$BD,$4F,$5C,$F5,$7D,$E6
	.dc.b	$4F,$DE,$4F,$DE,$7C,$F1,$FA,$88,$FA,$31,$A7,$F2,$CE,$9F,$BD,$1F
	.dc.b	$26,$34,$7C,$18,$F8,$F1,$CD,$FA,$71,$FB,$C3,$D8,$7C,$AA,$F9,$D1
	.dc.b	$F9,$48,$C1,$EF,$3C,$0F,$C6,$7F,$13,$1F,$2A,$3F,$A9,$19,$AF,$C2
	.dc.b	$CF,$90,$FE,$D4,$78,$0F,$A6,$7E,$CE,$3F,$C9,$07,$B8,$F4,$9E,$E7
	.dc.b	$E4,$1F,$8B,$7E,$41,$A3,$C1,$5F,$0E,$35,$1C,$CF,$63,$E0,$7F,$10
	.dc.b	$78,$1F,$90,$EA,$7E,$DC,$79,$6B,$F3,$31,$F9,$58,$F8,$A7,$FB,$B0
	.dc.b	$FC,$1B,$F8,$F3,$8C,$F6,$9E,$35,$F8,$3A,$FC,$59,$C4,$F9,$A7,$E1
	.dc.b	$8F,$04,$FB,$8F,$24,$7F,$85,$A3,$31,$ED,$7F,$6A,$74,$AF,$F2,$AD
	.dc.b	$7F,$91,$27,$89,$FC,$F8,$F2,$1D,$67,$89,$EC,$9E,$2F,$F6,$F8,$E2
	.dc.b	$7F,$1F,$9F,$C0,$3F,$D7,$63,$85,$7F,$BF,$0F,$EE,$07,$BD,$CC,$7F
	.dc.b	$BE,$5F,$9E,$78,$0F,$A9,$3F,$92,$71,$5F,$2A,$31,$1F,$89,$8C,$3F
	.dc.b	$95,$9F,$C1,$3F,$25,$F6,$9F,$15,$E1,$5F,$26,$33,$1E,$23,$2F,$C5
	.dc.b	$AE,$F5,$F7,$E7,$CA,$77,$9F,$86,$7E,$D9,$FC,$79,$CE,$38,$1F,$82
	.dc.b	$8D,$9E,$37,$35,$E3,$AE,$EF,$C7,$AF,$CB,$57,$07,$A4,$66,$BE,$0C
	.dc.b	$7E,$26,$7E,$0B,$FD,$2D,$E0,$FE,$C2,$3B,$C7,$B6,$3D,$47,$CE,$9F
	.dc.b	$89,$1C,$27,$FC,$90,$D8,$FD,$64,$FF,$09,$9E,$73,$F9,$C3,$F8,$0C
	.dc.b	$7D,$17,$83,$FB,$83,$84,$FA,$4E,$EF,$86,$3F,$28,$70,$7F,$3A,$7C
	.dc.b	$07,$AB,$ED,$38,$3E,$13,$F2,$4F,$D6,$3E,$23,$CC,$F8,$AF,$F4,$87
	.dc.b	$15,$F4,$1E,$B1,$93,$AB,$FA,$57,$F0,$AF,$E8,$5F,$E7,$AF,$07,$EB
	.dc.b	$9C,$1F,$BD,$3E,$F7,$E8,$47,$90,$D1,$ED,$7D,$35,$F8,$78,$C9,$FC
	.dc.b	$B8,$EF,$1A,$9F,$93,$1E,$48,$F3,$4F,$D4,$7F,$29,$1F,$41,$F3,$47
	.dc.b	$58,$E1,$1E,$63,$33,$F9,$38,$FC,$29,$A3,$CD,$1F,$E5,$69,$E4,$77
	.dc.b	$7F,$A6,$BF,$E1,$97,$A1,$E7,$3E,$EC,$7B,$23,$F1,$AD,$9F,$C6,$C7
	.dc.b	$18,$EF,$3F,$E3,$38,$EF,$3F,$2E,$3F,$00,$78,$DF,$44,$FC,$D3,$D0
	.dc.b	$7E,$85,$E1,$1F,$19,$AF,$A1,$FF,$04,$C6,$E3,$E1,$9E,$87,$9C,$FA
	.dc.b	$1B,$3E,$83,$8C,$64,$F8,$91,$EA,$3F,$C7,$2F,$AA,$7A,$9F,$8B,$8F
	.dc.b	$EC,$D1,$C6,$7C,$B1,$C4,$FF,$07,$3F,$80,$8E,$B1,$F8,$53,$8C,$7F
	.dc.b	$30,$9E,$A7,$BC,$A7,$D4,$3E,$41,$F9,$49,$FE,$7F,$3F,$E9,$93,$FA
	.dc.b	$99,$FA,$F3,$DE,$7F,$5E,$9F,$F0,$63,$FE,$27,$AF,$E8,$75,$D0,$FA
	.dc.b	$A7,$D9,$9F,$41,$EC,$9F,$41,$FC,$A0,$FA,$11,$F8,$73,$8B,$F2,$23
	.dc.b	$8C,$7A,$A3,$C0,$7E,$91,$F0,$1E,$F3,$91,$4C,$9F,$7C,$FF,$00,$3E
	.dc.b	$07,$2F,$CF,$AF,$EF,$53,$E0,$9E,$67,$1A,$F2,$D7,$D4,$8F,$C3,$57
	.dc.b	$D4,$3E,$D3,$E7,$8F,$79,$FB,$37,$9B,$DE,$BE,$5B,$E3,$8F,$90,$7D
	.dc.b	$58,$FE,$D9,$1F,$46,$3F,$C4,$8F,$D6,$3F,$8C,$1F,$58,$FE,$52,$73
	.dc.b	$3E,$53,$E6,$8F,$A5,$3E,$68,$F1,$CF,$9A,$3F,$64,$F1,$3D,$D1,$F0
	.dc.b	$23,$E2,$BC,$CF,$73,$FE,$40,$8F,$B8,$F5,$8F,$9B,$1C,$E3,$8C,$73
	.dc.b	$9F,$83,$1C,$67,$E3,$4F,$08,$E4,$FC,$43,$81,$F0,$8F,$EF,$71,$D6
	.dc.b	$79,$9F,$A3,$8F,$6C,$F2,$9F,$87,$1F,$B1,$7C,$F3,$B8,$F0,$35,$F0
	.dc.b	$1F,$8B,$29,$F8,$33,$11,$D1,$AF,$C8,$6C,$FC,$07,$F9,$AB,$F3,$8F
	.dc.b	$24,$6E,$3F,$6C,$76,$7F,$0C,$FB,$EB,$FA,$C3,$EC,$AF,$34,$F8,$63
	.dc.b	$8B,$4F,$D3,$9F,$C1,$8F,$C5,$3F,$E6,$86,$9F,$DE,$4F,$E6,$87,$F7
	.dc.b	$9A,$F3,$4F,$68,$F3,$1F,$E2,$90,$26,$15,$C9,$A7,$0E,$0C,$3C,$1C
	.dc.b	$18,$34,$F4,$32,$72,$72,$F1,$72,$C3,$4E,$4D,$B8,$7B,$1E,$67,$B3
	.dc.b	$D0,$F1,$9E,$16,$BC,$CC,$1B,$3C,$0E,$8F,$09,$D1,$EC,$E4,$E4,$F8
	.dc.b	$8F,$86,$F0,$34,$79,$DF,$03,$CD,$E0,$74,$74,$71,$39,$3E,$A7,$83
	.dc.b	$E2,$3E,$1B,$F2,$1D,$3E,$47,$BB,$1C,$BD,$8F,$7B,$F0,$4E,$67,$23
	.dc.b	$F1,$67,$DC,$9F,$E5,$AF,$CC,$7F,$B0,$3F,$92,$3F,$0E,$FF,$AC,$9F
	.dc.b	$C3,$3C,$1E,$53,$FE,$8E,$78,$1F,$92,$8E,$0F,$F5,$13,$B4,$6D,$FE
	.dc.b	$50,$7F,$B4,$8E,$AF,$DB,$83,$F7,$9F,$2C,$70,$7A,$BF,$D5,$4E,$A7
	.dc.b	$E2,$9F,$DD,$9F,$BF,$9F,$F7,$CC,$76,$7F,$27,$1D,$A3,$89,$FC,$72
	.dc.b	$3F,$68,$FB,$9D,$CF,$8C,$F5,$3F,$E5,$C8,$F5,$1E,$F3,$E2,$9E,$38
	.dc.b	$FD,$C4,$F8,$E3,$E3,$47,$39,$EB,$1E,$27,$F8,$BC,$7F,$2A,$8F,$93
	.dc.b	$1A,$7A,$9E,$73,$FD,$B0,$7B,$5C,$BF,$30,$F1,$1F,$58,$F3,$9F,$24
	.dc.b	$E6,$F8,$0F,$EB,$AF,$33,$91,$A7,$F8,$EB,$FC,$48,$4F,$CB,$7E,$44
	.dc.b	$78,$9F,$F3,$99,$E1,$8F,$B2,$FF,$8B,$0F,$81,$1D,$23,$E3,$47,$49
	.dc.b	$F8,$D3,$CE,$3F,$D6,$67,$38,$FF,$7A,$3E,$73,$EF,$9E,$03,$F1,$0E
	.dc.b	$9F,$AC,$7B,$23,$E0,$1F,$80,$79,$1F,$58,$F2,$9F,$D0,$9D,$3C,$8E
	.dc.b	$CF,$E0,$E7,$DA,$7E,$9E,$3B,$47,$B6,$7E,$E4,$7C,$63,$E6,$9A,$3B
	.dc.b	$1F,$70,$EC,$7C,$03,$D7,$1F,$C6,$E7,$E8,$47,$A0,$F9,$07,$A0,$FE
	.dc.b	$E9,$1F,$71,$E6,$7F,$9D,$4F,$94,$F3,$9F,$9F,$1E,$43,$DB,$1F,$EC
	.dc.b	$57,$B4,$F8,$1C,$9F,$C3,$8C,$CF,$E9,$DF,$C8,$9E,$89,$FE,$8D,$3C
	.dc.b	$63,$F2,$47,$C1,$3C,$F1,$F3,$9C,$9D,$8F,$F1,$71,$D8,$F7,$47,$F0
	.dc.b	$09,$F9,$B1,$93,$F3,$07,$CC,$3F,$8E,$BF,$83,$7A,$9F,$36,$7E,$D4
	.dc.b	$66,$3F,$2C,$72,$3C,$A7,$D5,$38,$9F,$2C,$F7,$47,$F6,$E9,$F7,$47
	.dc.b	$23,$DC,$F9,$E3,$EA,$47,$28,$FE,$3F,$1E,$23,$EA,$47,$F0,$27,$F4
	.dc.b	$4F,$77,$47,$CC,$9F,$B3,$1F,$A2,$3D,$91,$F6,$5F,$C0,$CF,$68,$C4
	.dc.b	$78,$67,$F8,$EC,$62,$3D,$07,$D1,$9F,$F3,$74,$F6,$8F,$D3,$1F,$16
	.dc.b	$3D,$93,$F6,$63,$89,$E2,$3F,$0C,$78,$A7,$BB,$83,$C4,$7E,$22,$3D
	.dc.b	$A7,$CB,$8F,$04,$7E,$78,$F0,$3F,$CB,$8F,$C0,$BD,$4E,$51,$EB,$9F
	.dc.b	$DA,$47,$AE,$7F,$CE,$67,$48,$FA,$E7,$C5,$8F,$21,$EA,$3E,$84,$F3
	.dc.b	$9F,$EB,$07,$E7,$E3,$F9,$A1,$F8,$E8,$FB,$2F,$EF,$CF,$A8,$72,$3A
	.dc.b	$CF,$91,$F8,$C7,$E7,$8F,$51,$D0,$FE,$93,$1D,$1F,$CF,$1F,$8C,$9F
	.dc.b	$85,$1F,$8C,$8F,$44,$FA,$DE,$87,$CE,$3A,$1F,$D1,$0D,$CF,$D4,$76
	.dc.b	$60,$F9,$67,$A1,$F8,$2F,$EC,$9F,$0C,$FE,$06,$7D,$93,$EF,$39,$B8
	.dc.b	$3C,$87,$E3,$1F,$CF,$1F,$C4,$4F,$86,$E5,$F9,$47,$E4,$0F,$80,$7F
	.dc.b	$1A,$8F,$D1,$3C,$8F,$9F,$1D,$A7,$EF,$C6,$4E,$47,$E0,$E7,$FD,$57
	.dc.b	$1F,$83,$3E,$E3,$FD,$1A,$3F,$80,$1F,$0C,$F9,$33,$FD,$02,$7E,$1C
	.dc.b	$7E,$50,$F2,$4F,$E6,$5F,$01,$F9,$93,$D2,$7F,$08,$7C,$F1,$CA,$33
	.dc.b	$1F,$A2,$32,$7F,$80,$DC,$BF,$93,$30,$FA,$DF,$F4,$B3,$E0,$9F,$CE
	.dc.b	$CF,$DD,$7E,$64,$7E,$28,$E1,$3F,$EB,$27,$84,$FC,$28,$FE,$B4,$7E
	.dc.b	$36,$7E,$D4,$7F,$8F,$67,$91,$FD,$5C,$A7,$E0,$63,$BB,$F9,$C9,$F8
	.dc.b	$8F,$DC,$3E,$21,$FD,$E4,$E3,$1F,$66,$3E,$AC,$79,$0F,$C8,$14,$FA
	.dc.b	$07,$38,$F7,$BF,$98,$3F,$23,$3F,$35,$EC,$78,$E7,$E6,$9F,$CB,$A7
	.dc.b	$C0,$FC,$A8,$E9,$3F,$0D,$FC,$49,$FA,$B8,$F5,$BE,$88,$F9,$C7,$E4
	.dc.b	$1E,$27,$A0,$A7,$A0,$E0,$79,$E7,$E3,$BD,$67,$CA,$FE,$94,$FE,$A0
	.dc.b	$70,$3F,$27,$3F,$D6,$67,$EA,$C7,$F7,$23,$53,$E6,$7E,$01,$FD,$94
	.dc.b	$FE,$1A,$FE,$A9,$F2,$9F,$96,$38,$BE,$27,$F2,$E7,$9E,$7D,$CE,$8F
	.dc.b	$8A,$7E,$9E,$3C,$0F,$AC,$F0,$1F,$C4,$67,$CE,$F8,$4E,$47,$E1,$0E
	.dc.b	$07,$F7,$23,$F9,$74,$6D,$F5,$9B,$3C,$0F,$9D,$F1,$C7,$DC,$8F,$CF
	.dc.b	$47,$B8,$FC,$19,$DA,$3F,$4F,$1F,$AC,$8F,$E2,$73,$EE,$7B,$CE,$A7
	.dc.b	$F1,$13,$FA,$39,$CB,$FC,$91,$FE,$13,$3E,$E9,$F8,$71,$E6,$3C,$B3
	.dc.b	$F9,$57,$BC,$7E,$81,$F3,$1F,$AD,$3A,$9F,$A1,$3F,$81,$9F,$CD,$5F
	.dc.b	$43,$F8,$63,$67,$23,$F6,$91,$EB,$7E,$03,$FC,$85,$FC,$B1,$F4,$A7
	.dc.b	$FB,$EB,$EA,$8F,$33,$D4,$E2,$78,$CF,$E1,$73,$F5,$9E,$0F,$92,$7F
	.dc.b	$D7,$B1,$D6,$7C,$0F,$EE,$A7,$F9,$03,$F1,$1E,$11,$F3,$63,$B9,$EB
	.dc.b	$9F,$B9,$3D,$E3,$F0,$AF,$CE,$8F,$A0,$FE,$0E,$3F,$0D,$1E,$27,$E2
	.dc.b	$C7,$81,$CC,$FB,$A7,$51,$E0,$3D,$07,$F3,$E9,$E9,$1F,$C4,$23,$F2
	.dc.b	$86,$CE,$B1,$FC,$DC,$FC,$3B,$FB,$53,$EA,$9D,$0F,$E8,$87,$C1,$76
	.dc.b	$F4,$3F,$10,$79,$9F,$8A,$FE,$CD,$FC,$43,$CC,$87,$88,$FC,$33,$D0
	.dc.b	$F8,$D3,$EF,$3E,$74,$76,$9F,$C0,$BF,$32,$7D,$F3,$FC,$8E,$3C,$51
	.dc.b	$CC,$CB,$FC,$7D,$F9,$93,$FA,$78,$FE,$34,$FC,$C7,$B4,$FE,$09,$FC
	.dc.b	$D9,$FC,$E2,$7D,$AF,$C1,$9F,$B6,$FC,$53,$9C,$FF,$30,$9F,$83,$1F
	.dc.b	$30,$87,$E2,$1F,$1C,$77,$8F,$E0,$27,$88,$FC,$D1,$FB,$28,$FE,$84
	.dc.b	$FC,$A3,$EE,$1C,$DF,$B8,$F8,$8F,$B8,$68,$E8,$7D,$C3,$E3,$9B,$3D
	.dc.b	$91,$F4,$5E,$27,$08,$F3,$BE,$43,$F6,$07,$43,$F1,$0E,$9F,$D2,$47
	.dc.b	$DA,$9F,$8F,$1E,$03,$F9,$C9,$FD,$22,$7F,$C9,$8F,$C6,$3A,$1F,$87
	.dc.b	$3E,$93,$F7,$5F,$A0,$72,$7F,$12,$7D,$63,$FC,$50,$FE,$1C,$D9,$FC
	.dc.b	$78,$F6,$C7,$90,$FC,$23,$CE,$3F,$C1,$E7,$F9,$2E,$7E,$24,$7C,$68
	.dc.b	$EC,$7E,$39,$FE,$10,$FD,$27,$91,$E0,$3F,$76,$78,$23,$9C,$79,$A3
	.dc.b	$FB,$0B,$EA,$3D,$CF,$C7,$3D,$07,$85,$CC,$FE,$92,$38,$C7,$8D,$FE
	.dc.b	$D0,$7C,$B8,$F7,$4F,$F4,$38,$F8,$87,$53,$CA,$75,$3F,$09,$3E,$36
	.dc.b	$BF,$11,$FE,$56,$F8,$A3,$F3,$51,$FD,$14,$FC,$A1,$FB,$E3,$89,$F0
	.dc.b	$E7,$F8,$B4,$F5,$8F,$8E,$7F,$6D,$3D,$66,$E7,$F7,$D3,$F5,$A3,$E1
	.dc.b	$CF,$16,$BE,$57,$DB,$1B,$7D,$B1,$B6,$BF,$1D,$FE,$8E,$7F,$84,$8F
	.dc.b	$69,$E8,$3E,$34,$7A,$23,$84,$7E,$D5,$FE,$F6,$FE,$D4,$FC,$C9,$F0
	.dc.b	$8F,$CC,$9C,$E3,$C8,$73,$7D,$C7,$F1,$78,$FC,$53,$EC,$32,$7F,$8D
	.dc.b	$CF,$D2,$47,$F8,$D9,$EE,$47,$EF,$BF,$D1,$4E,$27,$E1,$0F,$E0,$8E
	.dc.b	$9F,$E0,$8F,$92,$7E,$79,$F7,$23,$F7,$F1,$B9,$FD,$F4,$7B,$E3,$93
	.dc.b	$FC,$0A,$3D,$2F,$CE,$3C,$27,$F1,$E8,$E5,$1A,$7E,$14,$7C,$E9,$F0
	.dc.b	$BF,$C3,$E3,$FB,$6B,$F5,$4F,$BB,$1E,$58,$FE,$74,$78,$A3,$E2,$C7
	.dc.b	$F8,$35,$F8,$8F,$EC,$27,$EC,$C7,$F8,$49,$FE,$6E,$FC,$48,$D3,$E8
	.dc.b	$3E,$F4,$7F,$7D,$7F,$2A,$79,$CF,$F0,$B9,$FA,$F8,$E2,$FE,$6C,$FF
	.dc.b	$27,$38,$9F,$9D,$1C,$0F,$C3,$3F,$6C,$F5,$3E,$39,$FE,$1E,$FF,$96
	.dc.b	$23,$FD,$81,$18,$8F,$11,$F8,$53,$9C,$FC,$19,$E7,$1F,$0A,$78,$47
	.dc.b	$F0,$98,$F2,$C6,$27,$E6,$1F,$CE,$4F,$41,$FC,$26,$3E,$0C,$FA,$9F
	.dc.b	$01,$D6,$7F,$82,$3F,$91,$3C,$A7,$E4,$4D,$1F,$A3,$9E,$8F,$8A,$38
	.dc.b	$BF,$D2,$8F,$E5,$0F,$95,$E9,$3E,$17,$A3,$EE,$3F,$1A,$FF,$11,$3F
	.dc.b	$86,$BB,$7F,$25,$3E,$C7,$9B,$F0,$1E,$F3,$E2,$7F,$38,$73,$3F,$68
	.dc.b	$7B,$0E,$6F,$95,$FD,$5C,$73,$7F,$96,$1E,$27,$11,$FD,$06,$3E,$01
	.dc.b	$F0,$CF,$D0,$47,$D1,$8E,$87,$EC,$23,$F1,$CF,$77,$CA,$7F,$0E,$3B
	.dc.b	$1C,$8F,$59,$D4,$F2,$99,$3C,$A7,$B2,$7E,$33,$FA,$03,$F6,$07,$E0
	.dc.b	$27,$D3,$3F,$AA,$9F,$D8,$C7,$CB,$9E,$A7,$D9,$3A,$9F,$C0,$CF,$90
	.dc.b	$75,$10,$84,$FF,$BF,$DE,$93,$FC,$3D,$ED,$1F,$0E,$3F,$4C,$FD,$77
	.dc.b	$F8,$0B,$FB,$C7,$F1,$E7,$C4,$3C,$8F,$D1,$9F,$D5,$BC,$63,$99,$F1
	.dc.b	$DE,$CF,$23,$E3,$4F,$C2,$7F,$68,$FF,$B8,$4F,$3B,$F5,$A3,$F3,$67
	.dc.b	$29,$FC,$B4,$EA,$7F,$6F,$3E,$79,$CB,$F7,$63,$D6,$FF,$90,$9E,$87
	.dc.b	$DB,$3F,$03,$1F,$C2,$67,$A4,$79,$CF,$A1,$1F,$BF,$3D,$EF,$13,$99
	.dc.b	$F1,$67,$CD,$3C,$A7,$C7,$1E,$A3,$F9,$C4,$7F,$36,$9F,$97,$1F,$83
	.dc.b	$8F,$F3,$13,$E7,$8F,$96,$7D,$13,$CF,$1F,$C5,$4E,$C4,$8F,$CF,$9D
	.dc.b	$A3,$F9,$31,$EB,$9F,$E8,$E7,$AE,$7E,$84,$FC,$79,$EE,$7D,$63,$F7
	.dc.b	$A7,$E7,$DF,$F0,$14,$F2,$9F,$B5,$3F,$CD,$E3,$E2,$BE,$73,$E2,$9C
	.dc.b	$CF,$CE,$9F,$93,$9F,$03,$FE,$67,$7D,$0F,$F2,$48,$F9,$2F,$E2,$E3
	.dc.b	$2F,$F1,$89,$F8,$11,$F3,$63,$F8,$71,$F6,$23,$CB,$3F,$B4,$8F,$EB
	.dc.b	$13,$F3,$5F,$89,$3F,$4D,$F9,$B1,$F8,$B8,$EB,$1F,$7C,$FC,$7C,$71
	.dc.b	$9F,$2C,$7B,$CC,$C7,$F8,$71,$F8,$27,$8C,$F9,$4F,$F8,$BD,$FE,$08
	.dc.b	$60,$FE,$D0,$7F,$82,$27,$F6,$F1,$F0,$23,$F9,$3B,$FA,$F7,$E4,$1F
	.dc.b	$02,$7E,$A3,$CC,$FE,$0E,$7C,$08,$F2,$3F,$78,$F2,$1F,$92,$3F,$66
	.dc.b	$FF,$A3,$1D,$47,$47,$47,$DF,$9F,$C4,$CF,$8E,$7E,$1C,$FD,$F8,$FD
	.dc.b	$19,$F7,$A3,$31,$E0,$8C,$9C,$67,$E7,$C7,$EA,$E3,$E0,$3F,$E6,$53
	.dc.b	$EB,$9F,$E3,$F3,$E0,$47,$D9,$7E,$7C,$F9,$E7,$AC,$7D,$93,$89,$FE
	.dc.b	$90,$38,$9F,$A6,$8F,$29,$C6,$3C,$31,$F4,$DF,$54,$FD,$48,$E2,$7E
	.dc.b	$5D,$F4,$BF,$06,$38,$CF,$F3,$E7,$89,$FB,$68,$E2,$7E,$6D,$C9,$F9
	.dc.b	$B7,$FA,$A1,$FA,$93,$F5,$0E,$63,$F8,$C3,$C2,$3D,$D1,$FD,$E9,$E2
	.dc.b	$7D,$E3,$89,$FA,$B3,$C2,$7E,$30,$E3,$3E,$D7,$8C,$7E,$05,$E2,$7D
	.dc.b	$53,$89,$F5,$5F,$93,$1F,$C7,$67,$E8,$CF,$E9,$4F,$A2,$0F,$F4,$49
	.dc.b	$FD,$6A,$3F,$AC,$C7,$FB,$24,$F6,$47,$19,$F6,$1F,$02,$3F,$10,$6A
	.dc.b	$3E,$81,$E4,$3E,$C3,$A3,$E4,$4F,$F2,$97,$84,$7C,$53,$FA,$3B,$FA
	.dc.b	$67,$CE,$7C,$53,$E0,$1F,$12,$3F,$20,$FE,$99,$F8,$11,$F8,$07,$FC
	.dc.b	$24,$F8,$63,$E8,$1C,$4F,$C8,$47,$E2,$1F,$D2,$47,$17,$FC,$22,$7C
	.dc.b	$18,$FA,$93,$F0,$63,$EB,$CF,$AA,$3F,$31,$1F,$05,$FC,$84,$78,$63
	.dc.b	$D3,$1C,$5E,$91,$C6,$7D,$31,$C5,$F4,$BD,$A3,$E4,$9F,$C9,$4E,$0F
	.dc.b	$A9,$FC,$E3,$ED,$3F,$48,$E9,$FD,$81,$93,$F4,$0F,$95,$FB,$26,$4F
	.dc.b	$A0,$7E,$84,$FA,$AF,$FA,$2D,$FF,$20,$9E,$A7,$93,$F2,$8F,$51,$F2
	.dc.b	$8E,$47,$F3,$67,$94,$FF,$58,$8C,$9F,$80,$3F,$C5,$CF,$F8,$B6,$3F
	.dc.b	$C5,$87,$AE,$3D,$27,$DC,$8F,$98,$F2,$7F,$46,$FF,$17,$3D,$87,$E9
	.dc.b	$E7,$F0,$11,$F6,$23,$27,$E6,$0F,$A0,$75,$8F,$CC,$4F,$BA,$7F,$69
	.dc.b	$3F,$10,$FF,$27,$9F,$10,$F1,$4F,$63,$EE,$BD,$A3,$F0,$AF,$63,$F0
	.dc.b	$B1,$D9,$C1,$CA,$7E,$EC,$FB,$A3,$F5,$07,$E8,$4F,$73,$93,$CA,$E4
	.dc.b	$FC,$0B,$99,$FD,$4C,$7C,$F3,$FA,$B4,$FD,$48,$FE,$9B,$1E,$23,$B9
	.dc.b	$FD,$80,$F5,$3F,$96,$7F,$04,$7C,$78,$FC,$B1,$F1,$E3,$BC,$FF,$0A
	.dc.b	$7D,$D3,$C0,$87,$E4,$A3,$CC,$FE,$76,$79,$3E,$E9,$E2,$FC,$58,$CB
	.dc.b	$EE,$8C,$BE,$89,$FC,$1B,$99,$F2,$BE,$D8,$F8,$EF,$B5,$F0,$BF,$AA
	.dc.b	$8E,$AF,$F2,$93,$F2,$B3,$FC,$A1,$C1,$F8,$13,$C0,$FE,$05,$F4,$47
	.dc.b	$81,$F7,$9A,$7E,$EC,$7A,$27,$C6,$7A,$27,$AB,$FE,$8F,$7F,$9D,$4F
	.dc.b	$FA,$0E,$7E,$BC,$74,$9E,$31,$FC,$60,$E8,$7D,$23,$C9,$1F,$0E,$7F
	.dc.b	$C5,$91,$88,$FD,$69,$F9,$48,$F4,$3F,$3C,$F9,$11,$F9,$C9,$F8,$F1
	.dc.b	$C0,$D1,$C6,$7A,$9F,$B9,$7F,$86,$9F,$C1,$CF,$C5,$9F,$C1,$CF,$88
	.dc.b	$FF,$20,$38,$9F,$49,$E3,$1F,$B9,$3C,$2E,$27,$DD,$3F,$30,$F4,$C7
	.dc.b	$43,$D2,$7D,$47,$AB,$83,$F0,$07,$E6,$DC,$BF,$36,$3A,$C7,$94,$FE
	.dc.b	$B1,$3F,$5E,$7C,$47,$58,$CC,$FC,$D8,$C9,$F4,$63,$27,$C5,$7F,$8B
	.dc.b	$3E,$93,$C2,$FC,$C2,$1F,$26,$7E,$0B,$C2,$3E,$33,$F9,$67,$F9,$49
	.dc.b	$87,$F8,$F1,$F5,$5F,$09,$98,$FE,$2B,$19,$8F,$E0,$87,$F9,$55,$FE
	.dc.b	$11,$1A,$7F,$71,$3F,$8E,$7F,$70,$7C,$B3,$F3,$2F,$FA,$BA,$7B,$C6
	.dc.b	$4F,$86,$68,$FC,$EB,$FE,$77,$7F,$59,$3F,$9C,$8F,$8D,$1F,$C6,$0F
	.dc.b	$E6,$31,$FD,$2D,$FB,$C7,$D7,$3E,$94,$7A,$8F,$F7,$EC,$73,$7E,$04
	.dc.b	$F9,$5D,$1F,$EB,$07,$EF,$CF,$9C,$F9,$91,$FE,$27,$38,$BF,$99,$9F
	.dc.b	$CA,$4F,$C3,$8F,$C5,$9E,$33,$F0,$13,$D8,$FF,$69,$3C,$63,$BB,$F4
	.dc.b	$4F,$6B,$FB,$E7,$F0,$33,$E9,$7F,$2D,$3F,$31,$EC,$7E,$CC,$FC,$34
	.dc.b	$F8,$DC,$4F,$17,$B1,$FE,$43,$9F,$C3,$BE,$F3,$F0,$E7,$DE,$7F,$57
	.dc.b	$1C,$63,$C7,$3F,$C0,$CE,$CF,$DC,$21,$F2,$0F,$C1,$CF,$E2,$E7,$F5
	.dc.b	$11,$F9,$93,$D7,$1F,$15,$FF,$5A,$C7,$E3,$DF,$BA,$78,$E3,$F4,$AE
	.dc.b	$09,$1F,$4A,$3C,$F3,$F3,$27,$CE,$7E,$10,$F9,$B1,$E3,$8F,$DB,$47
	.dc.b	$8C,$F8,$B3,$FE,$13,$9F,$C0,$47,$C5,$8F,$21,$DE,$3C,$6F,$C5,$8F
	.dc.b	$1C,$FC,$59,$F1,$C7,$E9,$8F,$AE,$7F,$9D,$27,$B3,$F1,$A3,$F2,$F3
	.dc.b	$E9,$8E,$67,$A4,$FB,$11,$E4,$79,$9E,$93,$E3,$3F,$85,$7E,$09,$F9
	.dc.b	$43,$2F,$CF,$7F,$DB,$33,$EC,$9F,$03,$ED,$9E,$53,$E2,$39,$1F,$8E
	.dc.b	$7D,$AE,$23,$E1,$1C,$63,$B3,$FD,$6E,$3F,$DB,$AD,$7F,$D3,$E7,$90
	.dc.b	$FA,$B1,$F0,$9F,$90,$7A,$8F,$E1,$07,$F7,$33,$FD,$CC,$FD,$B3,$F8
	.dc.b	$3C,$7A,$E3,$E4,$3F,$39,$FE,$F8,$FF,$09,$3C,$F3,$F0,$0F,$B1,$3F
	.dc.b	$71,$C4,$F8,$8F,$81,$19,$71,$3C,$9E,$F3,$E5,$7F,$D6,$51,$DE,$39
	.dc.b	$C7,$C7,$7B,$CF,$CE,$9E,$53,$93,$9C,$7E,$D9,$FB,$E7,$ED,$8F,$1B
	.dc.b	$C6,$7A,$C7,$F0,$07,$AC,$7D,$27,$FC,$33,$1F,$EF,$A7,$6F,$E4,$DF
	.dc.b	$43,$FC,$C9,$CB,$F8,$D3,$BC,$73,$79,$9F,$D1,$A3,$FC,$8C,$FD,$B3
	.dc.b	$F2,$E7,$E7,$63,$F9,$73,$F2,$23,$AB,$FC,$F4,$D9,$F5,$9F,$E4,$4F
	.dc.b	$E4,$4F,$61,$F1,$23,$F4,$4F,$E1,$4F,$E2,$27,$39,$FC,$03,$F4,$A7
	.dc.b	$9C,$7C,$E8,$E6,$7F,$36,$3F,$23,$3E,$67,$E9,$10,$FC,$8B,$FE,$33
	.dc.b	$3F,$20,$7F,$44,$7F,$30,$73,$3D,$67,$85,$DB,$FC,$2E,$7E,$3B,$F3
	.dc.b	$0F,$0C,$7F,$46,$79,$9E,$63,$C2,$FF,$67,$9F,$C6,$47,$CB,$21,$E1
	.dc.b	$7E,$0B,$F6,$CC,$9C,$CF,$C0,$47,$E3,$5F,$94,$7D,$F8,$FD,$D3,$E4
	.dc.b	$3F,$9A,$CF,$A2,$3E,$54,$72,$9F,$A6,$FD,$73,$E9,$C7,$F0,$18,$F9
	.dc.b	$07,$B2,$33,$3C,$CF,$CB,$BD,$CF,$E2,$AE,$5F,$49,$DC,$FC,$64,$7C
	.dc.b	$73,$EC,$3E,$D3,$FA,$91,$F8,$B7,$E8,$BF,$AA,$9F,$E0,$CF,$C7,$9F
	.dc.b	$D0,$C7,$CC,$8F,$C0,$1E,$63,$FC,$2E,$7E,$2C,$A7,$F9,$70,$F6,$1F
	.dc.b	$26,$3D,$6F,$E0,$8F,$A4,$74,$34,$7E,$6C,$F5,$9F,$38,$FF,$03,$1F
	.dc.b	$00,$F5,$9F,$8B,$9E,$EF,$F0,$39,$FA,$6E,$8E,$E7,$C3,$35,$3E,$48
	.dc.b	$FC,$29,$EF,$8D,$3F,$11,$FC,$79,$EF,$3F,$98,$3F,$E5,$27,$99,$E5
	.dc.b	$3F,$61,$1F,$28,$FD,$79,$FB,$E8,$E9,$1F,$64,$FA,$93,$EA,$8C,$C7
	.dc.b	$C4,$3E,$A4,$7E,$55,$FC,$61,$93,$CA,$7F,$42,$79,$4F,$E1,$E3,$E7
	.dc.b	$1E,$48,$E5,$3F,$1A,$3D,$27,$68,$FC,$CC,$64,$F1,$C7,$18,$CC,$FA
	.dc.b	$CE,$D1,$E2,$9F,$CF,$4F,$B0,$FE,$69,$1F,$16,$39,$9A,$7E,$64,$0E
	.dc.b	$07,$F2,$D7,$FB,$3C,$FF,$2E,$7F,$97,$1F,$90,$9E,$4F,$E4,$67,$F4
	.dc.b	$2F,$E5,$CF,$AF,$3F,$ED,$B8,$ED,$1F,$89,$7C,$AF,$C6,$9F,$2C,$FF
	.dc.b	$99,$63,$F6,$51,$FD,$DD,$FE,$90,$7A,$E3,$4F,$E7,$63,$F1,$4F,$86
	.dc.b	$7C,$6F,$E4,$E3,$2F,$D9,$35,$1F,$C6,$E7,$F0,$F1,$D4,$FD,$09,$F5
	.dc.b	$4F,$D2,$9F,$C1,$0C,$9F,$98,$7C,$73,$F9,$C7,$9C,$7B,$5F,$AE,$FE
	.dc.b	$44,$F8,$53,$FC,$24,$8F,$EC,$4F,$92,$7C,$F8,$FF,$2F,$BE,$92,$1F
	.dc.b	$85,$3C,$B3,$F1,$DF,$F3,$D9,$EB,$3F,$2A,$73,$3A,$4F,$F8,$A9,$F5
	.dc.b	$1E,$C3,$CC,$7E,$E8,$F8,$E7,$95,$FE,$56,$7C,$E7,$F3,$67,$23,$D2
	.dc.b	$7E,$09,$EC,$F2,$3E,$51,$A7,$F0,$AF,$E2,$DF,$86,$E9,$EE,$62,$3D
	.dc.b	$D1,$F9,$43,$94,$7E,$69,$FC,$4C,$FE,$66,$7F,$D0,$31,$E6,$3D,$AF
	.dc.b	$E1,$1C,$1F,$31,$FC,$E9,$23,$E8,$C6,$89,$19,$7E,$5C,$FF,$8D,$5F
	.dc.b	$C0,$1F,$2C,$FC,$94,$FE,$11,$F8,$D1,$F2,$67,$94,$7C,$78,$E4,$E9
	.dc.b	$F8,$13,$FA,$B7,$DD,$1F,$A1,$8E,$8F,$CD,$9F,$8A,$7E,$4D,$FE,$7C
	.dc.b	$70,$73,$1F,$E9,$28,$F5,$BF,$9A,$3F,$95,$4F,$F5,$88,$FE,$EA,$68
	.dc.b	$FC,$51,$E7,$7F,$76,$F9,$CF,$C3,$3F,$AD,$39,$3F,$CA,$8E,$47,$D8
	.dc.b	$3F,$46,$7D,$83,$FC,$7B,$3F,$60,$FE,$0A,$F8,$23,$F0,$F1,$FD,$0C
	.dc.b	$CC,$7E,$78,$E8,$7E,$2C,$F1,$90,$FE,$71,$1F,$46,$7F,$6B,$1F,$81
	.dc.b	$7E,$14,$7F,$38,$7F,$98,$47,$DB,$3F,$60,$7C,$B7,$F1,$C7,$CB,$9F
	.dc.b	$1C,$FE,$62,$7C,$11,$F4,$23,$FD,$0F,$3D,$9F,$E7,$B3,$FB,$77,$C8
	.dc.b	$FB,$DE,$C7,$8D,$D3,$F8,$47,$F2,$2F,$D2,$7F,$05,$1E,$48,$FD,$6C
	.dc.b	$7F,$99,$67,$F8,$B4,$7D,$47,$FA,$84,$FC,$29,$F6,$9F,$50,$FD,$94
	.dc.b	$F7,$7F,$A2,$9E,$07,$FA,$43,$FE,$AC,$3E,$F4,$FE,$72,$3E,$BC,$FE
	.dc.b	$AA,$3E,$2C,$73,$32,$7F,$06,$9F,$A4,$7C,$F8,$E5,$1F,$08,$F8,$F1
	.dc.b	$FA,$E3,$F0,$0F,$CF,$7C,$E7,$D0,$9F,$85,$1F,$E9,$29,$FF,$26,$CF
	.dc.b	$EA,$63,$E1,$4F,$DE,$8E,$71,$F9,$92,$BC,$CF,$CA,$C7,$F8,$3D,$FD
	.dc.b	$69,$F1,$E7,$EF,$4F,$F9,$D4,$FB,$71,$F9,$48,$FC,$C4,$FC,$F7,$E1
	.dc.b	$C7,$F7,$49,$F7,$3F,$CF,$27,$D2,$FF,$9E,$4F,$88,$FE,$25,$D1,$E4
	.dc.b	$8F,$C6,$BA,$3F,$14,$79,$63,$F7,$B1,$EC,$8F,$8E,$49,$FE,$41,$1F
	.dc.b	$4E,$3F,$6E,$FA,$8F,$9D,$1E,$E8,$F2,$1C,$DE,$31,$FC,$6A,$3F,$26
	.dc.b	$79,$5F,$A5,$3F,$E4,$33,$C6,$F5,$3F,$1D,$03,$FC,$6E,$7F,$BC,$8F
	.dc.b	$F1,$04,$7F,$23,$7D,$91,$FA,$57,$F0,$8F,$37,$E6,$1A,$7F,$12,$FF
	.dc.b	$17,$3E,$04,$73,$7F,$84,$18,$3D,$07,$84,$FE,$92,$FF,$30,$7F,$2E
	.dc.b	$7C,$17,$94,$FD,$E8,$FA,$46,$4F,$D4,$1F,$C7,$CF,$D5,$90,$F8,$07
	.dc.b	$F7,$A3,$FC,$24,$7F,$0F,$3C,$67,$FA,$2E,$3F,$14,$7E,$0D,$D1,$CC
	.dc.b	$FF,$14,$9C,$8F,$C1,$1C,$8F,$E7,$B1,$F9,$27,$F9,$2C,$FD,$27,$89
	.dc.b	$F4,$0F,$49,$FA,$97,$D8,$F7,$3D,$06,$CF,$E8,$CF,$F0,$67,$99,$F8
	.dc.b	$43,$27,$C3,$3F,$85,$47,$F1,$47,$F3,$27,$D4,$7F,$20,$FC,$13,$2F
	.dc.b	$CF,$7C,$67,$CA,$3F,$B3,$1F,$B6,$3F,$C3,$87,$D8,$8F,$C3,$3F,$00
	.dc.b	$FD,$F1,$EB,$38,$9F,$24,$E2,$7E,$91,$FE,$34,$7E,$E1,$FA,$27,$AA
	.dc.b	$7F,$86,$47,$EA,$1F,$DD,$C7,$A4,$F9,$31,$FC,$88,$FD,$43,$E5,$7A
	.dc.b	$BF,$99,$3F,$7E,$79,$67,$FB,$34,$FE,$06,$3E,$E9,$97,$FA,$A1,$97
	.dc.b	$D3,$1F,$7C,$F9,$6F,$F4,$58,$FD,$9C,$FE,$06,$3E,$D9,$F5,$DF,$31
	.dc.b	$FD,$3D,$F3,$1F,$21,$C9,$F7,$DE,$26,$8F,$D6,$1F,$80,$3F,$14,$FD
	.dc.b	$29,$F1,$C7,$F4,$A9,$F4,$47,$F8,$DC,$F4,$3F,$2A,$3D,$13,$FC,$05
	.dc.b	$FD,$E4,$F2,$3F,$9F,$9F,$D8,$23,$F0,$B1,$F9,$23,$FC,$6A,$FF,$42
	.dc.b	$7F,$B6,$9E,$B7,$EE,$3F,$30,$FC,$64,$7F,$8B,$5F,$E3,$67,$F8,$90
	.dc.b	$FC,$81,$FE,$F6,$3F,$D5,$47,$FA,$AE,$3F,$B9,$B8,$3F,$D6,$4D,$76
	.dc.b	$E9,$AE,$5E,$4D,$7B,$14,$E6,$65,$E4,$F1,$6B,$E6,$36,$71,$39,$9E
	.dc.b	$03,$06,$0F,$09,$83,$89,$E9,$7A,$BE,$D6,$BF,$B3,$6B,$C8,$F2,$3C
	.dc.b	$DE,$67,$17,$07,$07,$C4,$77,$7E,$01,$97,$E8,$B4,$C3,$E1,$76,$74
	.dc.b	$2B,$EB,$78,$19,$7E,$53,$5F,$39,$B7,$8B,$F1,$9E,$2E,$4E,$25,$38
	.dc.b	$9B,$7E,$E4,$7E,$EA,$3F,$A3,$C7,$F8,$A1,$FE,$87,$D1,$FD,$88,$FE
	.dc.b	$8B,$2A,$FE,$CD,$1F,$92,$3F,$3B,$3F,$B0,$3F,$8B,$1F,$DA,$E7,$83
	.dc.b	$D0,$C4,$FE,$22,$7A,$BD,$23,$F3,$EF,$6A,$FD,$B3,$FB,$18,$FD,$B1
	.dc.b	$FA,$19,$F3,$3D,$03,$E5,$9E,$31,$FD,$44,$E2,$7F,$B2,$0D,$9E,$44
	.dc.b	$78,$D5,$7A,$A7,$A4,$FF,$17,$9F,$F6,$23,$F2,$5E,$13,$F1,$E3,$9C
	.dc.b	$76,$39,$D7,$57,$D1,$5F,$29,$E7,$3F,$11,$F9,$07,$07,$9C,$64,$F1
	.dc.b	$C7,$F4,$A3,$FA,$11,$FE,$48,$3E,$31,$ED,$8F,$2C,$F8,$63,$C6,$7C
	.dc.b	$88,$F8,$EF,$F3,$B7,$F8,$4D,$7F,$AE,$A3,$99,$E3,$8E,$6F,$B4,$F5
	.dc.b	$BC,$9F,$19,$FC,$60,$FA,$15,$F6,$1F,$0B,$EA,$7F,$9B,$9C,$0F,$EF
	.dc.b	$87,$C9,$9F,$99,$3E,$09,$E1,$1F,$CC,$6B,$FB,$0C,$FF,$8F,$27,$B4
	.dc.b	$FD,$9A,$FC,$69,$EF,$9F,$B5,$1F,$A3,$8F,$29,$F1,$CF,$05,$7C,$78
	.dc.b	$FC,$14,$FB,$67,$F9,$E1,$F2,$8E,$27,$C5,$38,$9F,$14,$F7,$BC,$5F
	.dc.b	$D4,$47,$F4,$E7,$EF,$BE,$E7,$8D,$7B,$AB,$F4,$53,$FE,$17,$9F,$F0
	.dc.b	$8C,$FB,$A7,$CC,$7E,$A6,$BD,$AF,$F7,$49,$FB,$35,$D6,$BF,$15,$1F
	.dc.b	$11,$FC,$94,$7C,$3A,$FC,$E4,$FC,$27,$F2,$2F,$E6,$0E,$0F,$AD,$FC
	.dc.b	$C9,$FB,$33,$F1,$EF,$07,$EF,$9C,$1F,$29,$FE,$8A,$7E,$4C,$79,$5E
	.dc.b	$C7,$E7,$E3,$B4,$79,$5F,$D3,$47,$67,$F0,$8F,$53,$F4,$87,$D6,$8F
	.dc.b	$E2,$6F,$6A,$FE,$24,$F6,$3F,$33,$3E,$43,$FD,$58,$76,$7E,$13,$D9
	.dc.b	$F2,$3B,$7F,$D2,$AF,$C8,$AF,$A9,$1F,$81,$AF,$C4,$CE,$E3,$FC,$68
	.dc.b	$ED,$FC,$84,$FD,$47,$A9,$EA,$3F,$0F,$5F,$E7,$27,$ED,$BE,$A7,$E6
	.dc.b	$9C,$0F,$D5,$9F,$05,$FF,$43,$9F,$E5,$78,$F2,$9F,$8D,$8F,$E4,$D5
	.dc.b	$F8,$DA,$FD,$D4,$EA,$7E,$61,$D0,$F1,$9F,$D2,$A3,$E3,$46,$E3,$C5
	.dc.b	$3F,$B5,$7F,$C3,$E7,$EF,$A3,$4F,$D0,$3D,$C7,$E2,$8F,$D6,$1F,$40
	.dc.b	$FD,$19,$E2,$8E,$8F,$FA,$CA,$7F,$80,$3F,$28,$FB,$87,$EC,$1F,$A0
	.dc.b	$FF,$A4,$AB,$E1,$BF,$B0,$7E,$54,$FF,$07,$8F,$4C,$7E,$C0,$FE,$1A
	.dc.b	$FF,$2F,$AF,$4B,$D4,$F4,$CF,$E5,$5F,$D1,$1E,$98,$FB,$8F,$BC,$F9
	.dc.b	$47,$BE,$3F,$2A,$EC,$F4,$BF,$88,$7A,$CF,$EE,$DF,$CD,$57,$EB,$67
	.dc.b	$D1,$1D,$5F,$8A,$FF,$95,$A7,$D4,$FF,$45,$38,$CF,$E6,$DE,$11,$A7
	.dc.b	$E6,$1E,$F3,$E6,$1F,$8E,$36,$7E,$21,$F8,$AF,$DD,$37,$3E,$47,$E3
	.dc.b	$1F,$01,$F8,$CF,$DE,$AF,$97,$1E,$F8,$D4,$FB,$EB,$F4,$D1,$FC,$B6
	.dc.b	$BE,$5B,$D2,$BD,$EF,$FA,$FC,$FF,$05,$CF,$F5,$AA,$FF,$29,$47,$F3
	.dc.b	$78,$E7,$3F,$C3,$E3,$D8,$73,$9F,$84,$FC,$93,$DC,$7C,$99,$E3,$1F
	.dc.b	$88,$3F,$98,$C7,$F4,$D3,$EB,$C7,$F1,$D3,$67,$D7,$8D,$D7,$F3,$07
	.dc.b	$DD,$1F,$AA,$9F,$93,$1A,$8F,$E4,$2E,$A3,$F1,$B3,$F2,$23,$FA,$04
	.dc.b	$78,$27,$F8,$CB,$A3,$99,$FC,$4E,$36,$FC,$E8,$F3,$1E,$17,$E7,$1F
	.dc.b	$92,$8D,$D7,$CE,$8F,$F5,$73,$5F,$99,$3C,$63,$E6,$1F,$81,$3E,$E1
	.dc.b	$B3,$FB,$04,$7F,$A6,$9F,$F4,$F3,$FE,$00,$AF,$63,$F2,$CF,$34,$FB
	.dc.b	$1F,$C4,$1E,$C8,$FE,$34,$FF,$86,$63,$53,$FA,$C9,$E6,$7C,$F8,$F7
	.dc.b	$CF,$C5,$3D,$E7,$E0,$A7,$CF,$5F,$82,$8F,$C6,$1F,$E5,$89,$FC,$63
	.dc.b	$F8,$37,$E0,$47,$E0,$E3,$E7,$1E,$99,$F8,$11,$FC,$E6,$3F,$BC,$9E
	.dc.b	$93,$F8,$E3,$DD,$F9,$AD,$7E,$69,$E7,$8F,$9B,$1F,$85,$9F,$E3,$B3
	.dc.b	$FE,$5E,$7F,$1E,$7E,$F5,$FC,$E1,$F9,$A3,$F7,$A7,$ED,$CF,$83,$3F
	.dc.b	$5E,$7E,$F9,$E1,$7E,$BC,$7E,$B2,$7F,$C7,$8F,$D7,$3D,$47,$F2,$83
	.dc.b	$B1,$FA,$93,$EF,$4F,$FA,$C2,$7F,$C9,$91,$FC,$48,$FE,$7A,$71,$3E
	.dc.b	$D1,$FD,$B2,$7E,$69,$FA,$97,$E6,$D7,$DB,$8F,$BF,$5F,$D7,$67,$C6
	.dc.b	$7F,$16,$39,$9F,$DC,$23,$99,$F1,$6B,$EF,$C7,$A0,$FB,$E7,$F5,$E3
	.dc.b	$CD,$5F,$C4,$23,$71,$D0,$FE,$98,$7F,$74,$9F,$34,$FE,$52,$3F,$CB
	.dc.b	$87,$F1,$57,$F2,$71,$FB,$C3,$F2,$71,$F9,$C3,$E9,$D7,$DD,$7F,$78
	.dc.b	$7C,$33,$F0,$4F,$E3,$E7,$B4,$7F,$7F,$3F,$BA,$1A,$9F,$1B,$F8,$E3
	.dc.b	$DC,$7E,$38,$A7,$E3,$8F,$15,$7E,$39,$F9,$F1,$E7,$3E,$7B,$FD,$A5
	.dc.b	$EA,$FD,$93,$DE,$75,$3F,$C6,$C7,$DE,$3E,$A1,$A3,$FD,$92,$78,$2B
	.dc.b	$D8,$F9,$CF,$9C,$7E,$76,$3E,$CB,$CC,$FC,$E9,$A9,$FF,$0D,$46,$AB
	.dc.b	$EC,$57,$E7,$2B,$DF,$5F,$9C,$7F,$79,$5F,$9C,$7F,$C9,$AF,$EA,$9E
	.dc.b	$D1,$F8,$07,$D0,$FE,$92,$BE,$DC,$FE,$CA,$3C,$33,$F4,$63,$F1,$D5
	.dc.b	$E1,$8F,$33,$FD,$A2,$7F,$8B,$47,$E6,$0F,$59,$DA,$3F,$84,$4F,$8E
	.dc.b	$BD,$73,$FC,$36,$7F,$6C,$6A,$7F,$C4,$71,$F8,$A3,$EC,$BE,$58,$F0
	.dc.b	$1F,$A7,$8D,$4F,$E2,$9F,$D0,$4F,$DF,$7F,$D5,$27,$FA,$ED,$F1,$3F
	.dc.b	$D1,$4F,$11,$F3,$27,$C4,$FF,$82,$4E,$A7,$DB,$3D,$87,$E5,$8F,$C0
	.dc.b	$47,$E2,$9F,$C2,$9F,$9B,$3F,$93,$9C,$8F,$C7,$19,$7F,$CA,$EF,$C9
	.dc.b	$AF,$AC,$E6,$3E,$1C,$7E,$6E,$BD,$D3,$FE,$31,$AF,$92,$FE,$E6,$BE
	.dc.b	$BB,$EC,$3E,$B9,$FE,$5A,$AF,$AE,$F8,$CF,$39,$F2,$23,$E7,$3F,$D1
	.dc.b	$4F,$C1,$1F,$8A,$9F,$C1,$3F,$08,$F6,$CF,$95,$F8,$C7,$A4,$F9,$4E
	.dc.b	$5F,$82,$77,$3F,$82,$9F,$E9,$73,$F3,$33,$FA,$17,$EB,$57,$F2,$77
	.dc.b	$F8,$51,$F4,$23,$DF,$19,$8F,$E4,$C7,$A1,$FC,$79,$97,$F8,$F9,$FC
	.dc.b	$90,$EB,$1F,$87,$78,$9D,$4F,$80,$7C,$B3,$E0,$3F,$61,$F9,$47,$E8
	.dc.b	$0F,$80,$7F,$55,$9F,$80,$FC,$E3,$EC,$1F,$8F,$3E,$C1,$DA,$7C,$CF
	.dc.b	$F0,$73,$D4,$79,$A7,$EA,$BF,$D8,$A7,$ED,$9D,$9F,$11,$F3,$4F,$F0
	.dc.b	$89,$FA,$23,$FB,$01,$E8,$7F,$1E,$ED,$FE,$FE,$FE,$75,$EB,$5F,$C7
	.dc.b	$23,$EC,$9E,$A3,$FB,$2C,$FF,$83,$1F,$D7,$3F,$65,$FF,$0A,$47,$E2
	.dc.b	$1F,$8B,$3D,$9F,$A4,$68,$F9,$C6,$A3,$F3,$AE,$A3,$FB,$F3,$A3,$D1
	.dc.b	$1A,$9F,$A9,$3F,$0C,$F6,$CF,$9C,$FF,$01,$1F,$0C,$FC,$79,$F3,$8F
	.dc.b	$15,$7D,$28,$FE,$47,$1F,$83,$8F,$C5,$CF,$E1,$E3,$F1,$6F,$F0,$C7
	.dc.b	$8B,$97,$ED,$1C,$4F,$B4,$7E,$80,$E2,$FC,$79,$F6,$CF,$86,$BF,$8F
	.dc.b	$D7,$F5,$7A,$FC,$F9,$D6,$BC,$A7,$D6,$7F,$89,$47,$E0,$8F,$80,$7F
	.dc.b	$46,$3E,$00,$3F,$B4,$47,$18,$EA,$7E,$3A,$3F,$C4,$47,$FA,$F4,$F6
	.dc.b	$B9,$7F,$C0,$87,$F1,$99,$FB,$CF,$D4,$3F,$3C,$7F,$06,$7F,$14,$FE
	.dc.b	$8E,$7F,$D5,$67,$E0,$E7,$E5,$3F,$09,$F7,$4F,$C2,$9F,$EE,$75,$F4
	.dc.b	$A7,$C5,$1F,$DD,$CF,$35,$71,$8C,$C7,$F7,$68,$F7,$3F,$AF,$8E,$AF
	.dc.b	$F2,$5A,$FE,$99,$3F,$89,$7D,$71,$F2,$8F,$61,$E2,$3E,$D9,$FA,$E3
	.dc.b	$F8,$4C,$76,$AF,$34,$7D,$F3,$F0,$71,$97,$8B,$F0,$6B,$CE,$AB,$FD
	.dc.b	$FF,$3D,$CF,$A2,$73,$8F,$C0,$C7,$A0,$F9,$B1,$F8,$18,$F9,$B1,$FA
	.dc.b	$59,$FC,$73,$FE,$35,$9F,$41,$F9,$23,$F4,$53,$F2,$5E,$D1,$F4,$5F
	.dc.b	$9A,$F1,$8F,$44,$68,$FE,$90,$7E,$7C,$F3,$9D,$8F,$83,$1E,$79,$FC
	.dc.b	$BC,$FE,$D8,$F3,$CF,$E2,$23,$F0,$73,$DC,$F5,$47,$73,$F8,$79,$E8
	.dc.b	$74,$F1,$8F,$97,$1F,$2A,$3D,$27,$F8,$30,$E4,$7E,$7A,$BF,$C4,$6E
	.dc.b	$0F,$CE,$9F,$C9,$E3,$F0,$75,$E6,$8F,$E4,$35,$FE,$31,$7F,$59,$1F
	.dc.b	$65,$FE,$7B,$5F,$82,$7E,$14,$7D,$87,$F7,$87,$E4,$E7,$CE,$FF,$31
	.dc.b	$AF,$8D,$1F,$B4,$8F,$21,$F9,$FA,$F3,$BD,$27,$EB,$46,$2B,$EA,$CE
	.dc.b	$4F,$AA,$FE,$45,$FA,$AE,$89,$3F,$3A,$3E,$D4,$7E,$0D,$E4,$7F,$51
	.dc.b	$8F,$2B,$E9,$3C,$A7,$EB,$DC,$CF,$C1,$9E,$B3,$F5,$4F,$F0,$2C,$FE
	.dc.b	$62,$3F,$04,$FF,$AB,$CF,$C8,$47,$F1,$93,$FC,$0C,$7C,$E7,$EE,$4F
	.dc.b	$BD,$F4,$4F,$69,$FA,$A7,$F2,$03,$F8,$FC,$79,$1F,$D8,$BC,$08,$F9
	.dc.b	$4F,$82,$FE,$31,$FF,$15,$9F,$04,$FB,$33,$ED,$8E,$31,$A3,$FB,$7C
	.dc.b	$FF,$0B,$7F,$AC,$1F,$C2,$C3,$FB,$FA,$FE,$D5,$1C,$27,$E2,$1F,$90
	.dc.b	$3E,$23,$FC,$5C,$F3,$4F,$DB,$3E,$B1,$F0,$5F,$F0,$BC,$FF,$00,$7F
	.dc.b	$40,$7D,$A3,$F4,$07,$F3,$93,$CB,$1F,$9D,$7F,$B3,$9F,$40,$F9,$27
	.dc.b	$C0,$3E,$49,$E2,$9F,$82,$F0,$3E,$B1,$F9,$07,$E8,$3F,$90,$76,$FF
	.dc.b	$7F,$3A,$19,$3A,$10,$FB,$67,$8E,$3F,$5C,$FF,$81,$63,$EB,$3F,$00
	.dc.b	$FC,$6C,$FF,$0E,$8F,$94,$FF,$B6,$27,$E3,$57,$E2,$6B,$B1,$C4,$FC
	.dc.b	$09,$F2,$23,$E6,$3E,$73,$F9,$A3,$F4,$2B,$D7,$1F,$D1,$8F,$84,$7D
	.dc.b	$E7,$FB,$83,$FC,$60,$EB,$3F,$E7,$C9,$FA,$11,$F9,$F3,$E4,$9B,$7E
	.dc.b	$F4,$FB,$A3,$C4,$76,$8F,$86,$FE,$26,$BE,$03,$F7,$27,$F1,$2F,$E6
	.dc.b	$4F,$96,$76,$38,$47,$F0,$F7,$F0,$C7,$F2,$29,$FC,$D5,$7F,$85,$63
	.dc.b	$F2,$47,$C2,$AF,$23,$F0,$4F,$2C,$7C,$D7,$F4,$87,$38,$FC,$EB,$E5
	.dc.b	$AF,$A9,$1F,$8A,$9E,$13,$F4,$EB,$EB,$BF,$97,$8F,$13,$EB,$3F,$9E
	.dc.b	$BF,$2E,$7D,$F1,$FE,$1A,$3E,$61,$F4,$CF,$CA,$1F,$90,$9F,$B4,$FE
	.dc.b	$08,$F4,$9D,$EB,$FC,$0E,$F7,$AF,$2B,$F4,$8E,$B3,$F4,$9F,$F2,$69
	.dc.b	$F7,$8F,$F5,$29,$F1,$48,$7E,$16,$7F,$80,$BC,$23,$EF,$BF,$91,$AF
	.dc.b	$C6,$BF,$E5,$BA,$F0,$BC,$A7,$F7,$11,$88,$FE,$1D,$1F,$C3,$6B,$FA
	.dc.b	$B4,$7E,$E8,$FB,$07,$53,$E0,$1C,$EB,$E4,$BE,$37,$F7,$33,$F1,$A3
	.dc.b	$ED,$9F,$D4,$CF,$8D,$3F,$D6,$9F,$F2,$03,$F9,$2A,$FA,$F1,$FC,$40
	.dc.b	$FE,$D8,$7E,$06,$3C,$B3,$FB,$C7,$E7,$C7,$E5,$0F,$E9,$C7,$E5,$0F
	.dc.b	$01,$CA,$7A,$BF,$8D,$8F,$01,$C9,$C3,$F9,$D9,$FD,$DC,$7B,$E3,$E8
	.dc.b	$BF,$B0,$7F,$C2,$2F,$C5,$3D,$E7,$F5,$D3,$F1,$15,$FA,$07,$E7,$BF
	.dc.b	$21,$E5,$1F,$CC,$6B,$F2,$8F,$E7,$A3,$F8,$7C,$7D,$7A,$FE,$03,$1F
	.dc.b	$C2,$63,$15,$CE,$3C,$07,$EE,$0F,$CD,$1F,$C1,$5F,$AA,$78,$E7,$9C
	.dc.b	$7E,$36,$BE,$6C,$FF,$68,$8F,$AE,$7F,$78,$9F,$A1,$3F,$CB,$67,$C1
	.dc.b	$18,$9F,$F1,$71,$F0,$CF,$DA,$47,$ED,$DE,$67,$D5,$3C,$E7,$ED,$0F
	.dc.b	$D0,$9F,$1A,$7E,$7B,$F9,$33,$47,$FB,$3A,$7D,$27,$E6,$5F,$EF,$4F
	.dc.b	$8D,$FD,$61,$B9,$F3,$9F,$E3,$F7,$F7,$27,$E1,$67,$F0,$4F,$E2,$27
	.dc.b	$F8,$DC,$62,$3C,$F5,$F4,$23,$EB,$47,$CF,$9F,$DE,$3F,$70,$F8,$73
	.dc.b	$F7,$27,$E6,$C7,$F8,$E4,$FB,$A7,$C4,$9F,$C7,$BF,$B9,$31,$5F,$BC
	.dc.b	$8F,$DC,$C6,$0F,$D9,$9B,$3D,$53,$FA,$D3,$FD,$BF,$3F,$CA,$AB,$F4
	.dc.b	$33,$ED,$8C,$4F,$B6,$BE,$41,$FC,$AA,$79,$9F,$81,$9F,$90,$FE,$44
	.dc.b	$F1,$1F,$A0,$8E,$31,$F7,$E3,$89,$F7,$E3,$2F,$C0,$32,$FC,$03,$DA
	.dc.b	$E2,$3E,$01,$F6,$1F,$90,$7F,$14,$AF,$98,$FE,$02,$3C,$AF,$C3,$9F
	.dc.b	$B3,$5F,$0C,$FC,$C1,$F8,$C7,$C7,$1C,$CF,$C8,$BF,$8C,$9F,$04,$7F
	.dc.b	$82,$4F,$CE,$1F,$CE,$CF,$84,$79,$CF,$A7,$1E,$A3,$C2,$FC,$93,$D2
	.dc.b	$7E,$68,$FF,$6F,$47,$69,$FA,$4F,$EC,$63,$A1,$F3,$5F,$6C,$FE,$7E
	.dc.b	$3F,$64,$7F,$A2,$8F,$F1,$79,$B3,$BD,$7E,$5E,$BF,$23,$1F,$82,$3F
	.dc.b	$59,$19,$7E,$51,$E6,$3F,$AB,$D7,$F8,$51,$FE,$5D,$1F,$E5,$B7,$DF
	.dc.b	$1E,$99,$F9,$A7,$63,$F9,$2C,$FD,$69,$FF,$70,$CF,$E1,$6B,$E2,$C7
	.dc.b	$E1,$63,$DA,$73,$7F,$52,$7D,$47,$0E,$CF,$80,$7F,$90,$CF,$80,$7F
	.dc.b	$4C,$AF,$E7,$51,$F8,$A7,$F6,$F1,$F4,$4F,$F7,$14,$74,$9F,$C9,$34
	.dc.b	$FE,$4F,$1F,$80,$3F,$A1,$4F,$B4,$F7,$BF,$20,$FC,$35,$7B,$DF,$F7
	.dc.b	$41,$FC,$C8,$FF,$19,$4F,$CB,$7F,$B4,$1F,$D1,$8E,$C7,$F2,$19,$EC
	.dc.b	$FC,$23,$F8,$D3,$EA,$7E,$A9,$FE,$16,$3F,$34,$69,$E2,$FC,$73,$CA
	.dc.b	$7C,$B7,$93,$F9,$AA,$FA,$67,$C2,$8F,$8F,$5F,$85,$9F,$B2,$7E,$E6
	.dc.b	$77,$5F,$36,$BB,$C7,$F0,$93,$33,$F8,$68,$E6,$FF,$08,$3F,$A2,$9C
	.dc.b	$E7,$F8,$9C,$7E,$DE,$3F,$CE,$AF,$DA,$79,$3D,$A3,$F1,$47,$08,$FD
	.dc.b	$71,$F8,$13,$FC,$07,$3E,$B9,$E1,$1D,$DF,$F0,$09,$DD,$FA,$C7,$CC
	.dc.b	$3F,$B4,$1F,$56,$BE,$E3,$A3,$F9,$F1,$F3,$63,$F3,$EF,$CD,$7B,$BF
	.dc.b	$94,$72,$FD,$53,$F1,$0F,$F0,$07,$F3,$73,$FE,$29,$7E,$29,$FD,$36
	.dc.b	$BF,$10,$F8,$23,$FC,$83,$5E,$47,$FA,$8C,$7F,$6C,$3D,$47,$EE,$0F
	.dc.b	$E0,$A7,$43,$F2,$67,$E2,$4F,$D2,$3F,$EE,$EA,$FC,$2C,$7F,$81,$4F
	.dc.b	$3B,$FE,$17,$7D,$C6,$0F,$A2,$FB,$9F,$E3,$AF,$28,$FE,$31,$1F,$E8
	.dc.b	$B3,$F4,$27,$E6,$0D,$C7,$C8,$7C,$D3,$F9,$17,$F5,$73,$F8,$DA,$FC
	.dc.b	$04,$7D,$1A,$FD,$9B,$FC,$90,$FC,$F9,$F3,$CF,$CC,$9F,$2C,$F8,$15
	.dc.b	$D5,$FB,$87,$C9,$3F,$49,$3D,$AB,$F1,$51,$EF,$36,$FB,$1B,$1F,$CE
	.dc.b	$63,$F4,$31,$F8,$1A,$F2,$BF,$04,$FB,$53,$FC,$3A,$BE,$D3,$F7,$A3
	.dc.b	$C4,$7D,$13,$D9,$3F,$35,$FE,$32,$FC,$D8,$FE,$31,$1F,$97,$3E,$A9
	.dc.b	$F0,$0F,$BE,$7E,$70,$FA,$07,$28,$FA,$0F,$F8,$92,$BD,$2F,$E1,$CF
	.dc.b	$C5,$9F,$60,$FC,$3C,$FF,$BA,$27,$8C,$7F,$26,$3E,$34,$FE,$59,$F1
	.dc.b	$9F,$6A,$7F,$73,$3F,$B2,$8F,$C4,$BF,$19,$F9,$67,$E9,$23,$FC,$B0
	.dc.b	$7B,$9F,$98,$7F,$25,$3F,$12,$78,$1F,$E2,$8D,$3E,$21,$F8,$33,$C4
	.dc.b	$7F,$03,$8F,$E1,$CF,$43,$ED,$BF,$E0,$78,$FB,$87,$E0,$CE,$B1,$F3
	.dc.b	$6B,$F1,$55,$FB,$23,$FC,$1B,$1F,$9A,$3F,$90,$1F,$E0,$93,$EE,$1F
	.dc.b	$83,$3F,$B9,$9A,$3E,$51,$D4,$C1,$FD,$14,$F6,$1F,$69,$FD,$C5,$7D
	.dc.b	$28,$FF,$11,$57,$9D,$FE,$52,$70,$3F,$06,$7F,$05,$3F,$B5,$1F,$7C
	.dc.b	$FD,$0C,$7D,$57,$F5,$C7,$F3,$43,$F9,$6C,$7C,$27,$E5,$9E,$37,$F0
	.dc.b	$8F,$73,$F1,$8F,$8D,$F9,$D3,$FB,$B9,$EC,$FE,$BA,$3E,$01,$FE,$22
	.dc.b	$3C,$87,$C8,$3F,$0A,$7E,$F8,$FC,$09,$F9,$F3,$F2,$E7,$EB,$8F,$C0
	.dc.b	$9A,$3C,$E7,$C2,$3F,$39,$5F,$6A,$BF,$65,$3F,$6A,$7D,$35,$E6,$9F
	.dc.b	$81,$1E,$28,$FF,$37,$99,$7E,$79,$EA,$9F,$D5,$BF,$81,$7F,$84,$1F
	.dc.b	$87,$9F,$C6,$CF,$19,$F8,$B1,$FA,$68,$FC,$AC,$7C,$29,$E3,$3E,$63
	.dc.b	$8C,$EA,$7B,$47,$E9,$DF,$DB,$9F,$EE,$B3,$EA,$1F,$8E,$3F,$25,$3F
	.dc.b	$98,$AF,$85,$1F,$D1,$4F,$83,$5D,$5E,$07,$C1,$7F,$71,$3E,$B8,$FD
	.dc.b	$C1,$FE,$1C,$7F,$CD,$91,$FA,$B8,$CC,$FF,$20,$7B,$1D,$27,$93,$F7
	.dc.b	$4F,$E2,$C7,$A0,$FC,$EC,$FC,$77,$E2,$57,$BE,$BD,$31,$FC,$7C,$FD
	.dc.b	$91,$F8,$B3,$E2,$CF,$8D,$ED,$5F,$AA,$7B,$3F,$E7,$BA,$FC,$C4,$76
	.dc.b	$9E,$07,$F3,$39,$E1,$1F,$91,$3F,$53,$1F,$1D,$FD,$C1,$FC,$E9,$FE
	.dc.b	$D4,$FF,$13,$3E,$51,$FC,$88,$FE,$38,$7C,$03,$A1,$FA,$18,$F8,$B1
	.dc.b	$E1,$9F,$D7,$CE,$A7,$F0,$AF,$F5,$E3,$F8,$4B,$F4,$A7,$FB,$75,$7C
	.dc.b	$69,$FF,$3F,$4F,$E7,$DF,$A5,$5F,$8D,$7E,$B9,$FC,$E8,$F3,$9F,$C1
	.dc.b	$0F,$F7,$79,$FA,$D3,$31,$FB,$B7,$F1,$B1,$F9,$67,$F8,$49,$FA,$63
	.dc.b	$F6,$73,$E3,$7F,$A7,$1F,$90,$36,$7E,$90,$FE,$70,$60,$FA,$95,$EC
	.dc.b	$7D,$F3,$FA,$17,$94,$FF,$A6,$5E,$47,$F8,$0C,$FC,$33,$E1,$7F,$84
	.dc.b	$1A,$3F,$52,$FF,$AA,$6B,$E3,$47,$D3,$9F,$8C,$FE,$29,$FA,$6F,$F0
	.dc.b	$E3,$F8,$99,$FC,$F8,$FC,$91,$F3,$8F,$D0,$9C,$CF,$B0,$68,$FC,$49
	.dc.b	$F2,$8F,$A6,$7E,$26,$3F,$07,$5F,$E5,$B7,$FB,$84,$66,$7C,$B3,$FE
	.dc.b	$4F,$8F,$31,$FC,$0C,$FC,$44,$7E,$A5,$F9,$31,$FE,$0D,$3F,$3E,$7E
	.dc.b	$61,$FB,$D3,$E0,$7F,$30,$7E,$08,$FC,$C1,$F0,$4F,$D1,$9F,$5C,$FD
	.dc.b	$8C,$7D,$77,$DD,$1F,$95,$7F,$AB,$57,$D3,$7F,$C4,$71,$F9,$C7,$EB
	.dc.b	$9E,$69,$FE,$C1,$1F,$B7,$7F,$82,$47,$63,$99,$B3,$F7,$B5,$F8,$57
	.dc.b	$CB,$5F,$85,$7F,$16,$70,$3C,$B1,$B7,$F8,$8C,$FE,$F6,$BD,$11,$FB
	.dc.b	$D3,$F7,$2F,$F0,$47,$E1,$1F,$74,$F4,$9D,$23,$E7,$1F,$0E,$3F,$A3
	.dc.b	$C7,$E1,$4F,$E3,$47,$E2,$9F,$E6,$CF,$AC,$FB,$31,$F5,$9F,$0C,$FE
	.dc.b	$49,$FD,$9B,$FD,$62,$7E,$C4,$7D,$77,$CC,$E8,$FD,$A1,$4F,$EF,$87
	.dc.b	$E1,$CF,$5B,$F4,$DF,$F0,$53,$FE,$99,$AF,$B0,$FE,$98,$D4,$7F,$6E
	.dc.b	$3F,$D8,$51,$A9,$FF,$80,$35,$E9,$9F,$24,$14,$FE,$60,$FD,$94,$7F
	.dc.b	$0B,$7E,$C1,$CE,$38,$4F,$F7,$43,$FB,$4C,$66,$3F,$D8,$CF,$07,$E5
	.dc.b	$1F,$C4,$8F,$29,$83,$E1,$1F,$E0,$73,$F8,$29,$FC,$12,$3F,$99,$3F
	.dc.b	$14,$D9,$F5,$E7,$CC,$FF,$AC,$A7,$C5,$3E,$58,$FE,$35,$18,$8F,$ED
	.dc.b	$73,$FB,$87,$F6,$12,$A4,$FC,$5A,$F8,$95,$FB,$D3,$C6,$E1,$FE,$09
	.dc.b	$1F,$E8,$FA,$FE,$7E,$F1,$8D,$9C,$4E,$F5,$DA,$7E,$8C,$68,$E6,$FF
	.dc.b	$8B,$0F,$24,$F3,$8E,$2F,$E4,$67,$F1,$E7,$F8,$FE,$3E,$9C,$FC,$D8
	.dc.b	$CC,$FF,$87,$E3,$E1,$99,$7E,$13,$FE,$51,$AF,$B0,$FE,$A6,$BF,$80
	.dc.b	$CF,$F1,$98,$FF,$24,$4F,$CB,$78,$1F,$44,$E8,$7F,$30,$AF,$CA,$BF
	.dc.b	$72,$3F,$20,$F2,$36,$7E,$C8,$FE,$0E,$73,$3C,$47,$F0,$CA,$FC,$D4
	.dc.b	$FF,$32,$8F,$CD,$3F,$2E,$7E,$A4,$7F,$28,$8D,$3F,$28,$F9,$F3,$F2
	.dc.b	$9F,$C6,$47,$C6,$7F,$B0,$CF,$C2,$8F,$E0,$F1,$FC,$50,$F9,$27,$90
	.dc.b	$FF,$29,$57,$F8,$6A,$7E,$24,$7F,$A7,$AB,$F8,$84,$E6,$3D,$E7,$E1
	.dc.b	$8F,$D5,$C7,$E5,$1F,$EA,$B5,$FE,$C6,$AE,$6F,$DD,$8F,$0B,$E4,$3F
	.dc.b	$9A,$4F,$E0,$23,$C8,$7E,$BC,$F9,$C6,$0F,$D4,$9F,$38,$FF,$4A,$CF
	.dc.b	$6A,$FC,$33,$F0,$5F,$F1,$AC,$7F,$7E,$3E,$A5,$78,$63,$F9,$C3,$F7
	.dc.b	$A3,$FB,$93,$E9,$7D,$E7,$F8,$40,$F0,$3F,$C6,$DC,$9F,$C0,$0F,$DE
	.dc.b	$1F,$DC,$8F,$BB,$1F,$A1,$7F,$98,$9F,$61,$FC,$CB,$F3,$CF,$E9,$91
	.dc.b	$E1,$7E,$03,$F7,$5F,$F0,$A1,$FC,$B2,$3F,$7C,$FD,$39,$54,$8F,$09
	.dc.b	$F9,$88,$FC,$CB,$F6,$0F,$E5,$26,$0A,$53,$F1,$E5,$32,$71,$74,$65
	.dc.b	$D9,$83,$BC,$78,$E3,$11,$C2,$3C,$CE,$0E,$87,$8E,$3A,$1E,$C8,$B1
	.dc.b	$D1,$F8,$2F,$B2,$37,$1D,$23,$91,$B3,$D2,$70,$3B,$3B,$79,$1B,$8F
	.dc.b	$23,$D6,$34,$FC,$B3,$47,$63,$D8,$F9,$9F,$E0,$EF,$C1,$74,$78,$1D
	.dc.b	$3F,$00,$F0,$9C,$8F,$A1,$1F,$56,$39,$46,$8E,$46,$8F,$19,$93,$D0
	.dc.b	$F8,$EB,$F1,$84,$AF,$F4,$CB,$FC,$70,$FE,$93,$1F,$89,$71,$1F,$C5
	.dc.b	$A7,$F8,$04,$7F,$16,$7F,$B9,$9F,$80,$7F,$8B,$9F,$C6,$E7,$EE,$BF
	.dc.b	$D1,$E3,$91,$FD,$12,$3B,$3F,$96,$8F,$C9,$1F,$D3,$4E,$C7,$49,$FD
	.dc.b	$54,$78,$2B,$F0,$8F,$E7,$8E,$B3,$EA,$AB,$5E,$09,$ED,$1F,$D5,$DF
	.dc.b	$D9,$D7,$69,$FF,$4F,$C7,$EE,$27,$CA,$FF,$31,$AF,$2C,$FF,$38,$9E
	.dc.b	$53,$E3,$6C,$FB,$E3,$FB,$A1,$CE,$3C,$6E,$0F,$9A,$7F,$79,$78,$CF
	.dc.b	$F7,$38,$CD,$79,$5F,$AC,$73,$39,$57,$C7,$7E,$AD,$72,$7F,$B9,$47
	.dc.b	$D3,$8F,$E9,$13,$F2,$27,$FA,$61,$F1,$A3,$DF,$3C,$E3,$CA,$F8,$5F
	.dc.b	$F1,$FC,$D7,$C3,$5F,$4A,$77,$1E,$2A,$EE,$78,$DE,$2F,$E1,$27,$E7
	.dc.b	$C7,$E0,$E3,$E2,$CF,$CE,$8F,$89,$1C,$88,$FD,$87,$93,$F1,$0E,$47
	.dc.b	$90,$F9,$C7,$E0,$1E,$2F,$D7,$6C,$7D,$73,$F3,$EF,$C2,$9F,$C4,$4F
	.dc.b	$E0,$63,$D4,$FF,$8F,$5E,$31,$EA,$9F,$C8,$4F,$85,$F5,$1F,$1E,$BA
	.dc.b	$D7,$F5,$F8,$DC,$7F,$A3,$5C,$CF,$13,$E5,$1E,$F9,$EE,$FD,$CA,$FD
	.dc.b	$1C,$FA,$63,$F0,$8F,$D2,$7E,$CC,$FA,$2B,$A9,$FD,$95,$FE,$17,$3F
	.dc.b	$E1,$19,$FE,$95,$1F,$31,$FC,$23,$F5,$1F,$19,$F2,$4F,$21,$0F,$C5
	.dc.b	$9F,$84,$7E,$53,$F4,$8F,$94,$FA,$9F,$C1,$B5,$E0,$F7,$7D,$2E,$67
	.dc.b	$D3,$1F,$DE,$AB,$F8,$CB,$F0,$0F,$E1,$47,$9A,$BF,$84,$3F,$C5,$63
	.dc.b	$F5,$8F,$F6,$49,$FC,$D4,$FD,$03,$E4,$47,$E1,$CF,$63,$CA,$37,$1D
	.dc.b	$CF,$E0,$87,$79,$EA,$FF,$8D,$E3,$F0,$0F,$ED,$DF,$BC,$F1,$9F,$D7
	.dc.b	$3D,$CF,$98,$47,$C1,$5F,$4E,$3E,$29,$F8,$53,$E2,$9E,$6A,$F0,$CF
	.dc.b	$EF,$E3,$07,$E9,$5F,$CA,$9E,$E7,$E9,$9A,$3E,$99,$4E,$EE,$CF,$E6
	.dc.b	$93,$F8,$B8,$FF,$51,$9F,$B0,$AF,$6B,$FB,$29,$E1,$5F,$C9,$EB,$ED
	.dc.b	$4F,$09,$FD,$0B,$F4,$CF,$C8,$C7,$F2,$47,$F2,$27,$BC,$F8,$D1,$F1
	.dc.b	$23,$9C,$78,$4E,$47,$F2,$87,$E4,$47,$DD,$8F,$C2,$9F,$20,$F9,$D3
	.dc.b	$F2,$0F,$D5,$95,$FC,$41,$F2,$5F,$E3,$CF,$23,$C6,$EE,$BE,$FC,$78
	.dc.b	$CF,$C1,$1E,$C3,$FA,$B1,$F3,$0E,$47,$D2,$3C,$67,$EF,$4F,$09,$F5
	.dc.b	$0F,$D6,$CF,$AD,$FC,$89,$EB,$7E,$23,$F6,$63,$F4,$EF,$88,$F8,$95
	.dc.b	$F3,$5F,$11,$C8,$F5,$9E,$18,$F5,$CF,$D6,$8E,$47,$EA,$67,$E6,$BF
	.dc.b	$9B,$AF,$E4,$F3,$F8,$69,$FA,$AF,$F8,$3D,$FE,$1F,$1F,$E3,$19,$FD
	.dc.b	$34,$72,$7F,$10,$7C,$49,$E4,$FA,$CF,$C4,$4F,$F0,$07,$67,$A4,$F9
	.dc.b	$51,$F0,$9F,$96,$7C,$97,$F9,$33,$C1,$FD,$8B,$C2,$3E,$D4,$FE,$76
	.dc.b	$3E,$D4,$FF,$25,$7E,$49,$F6,$9F,$A4,$72,$3C,$EF,$80,$FE,$8C,$7E
	.dc.b	$AD,$DC,$FC,$63,$F4,$91,$ED,$3F,$D3,$F1,$F3,$CF,$03,$F3,$E7,$EC
	.dc.b	$47,$28,$FE,$58,$F8,$0F,$80,$49,$F5,$CF,$09,$FA,$C7,$08,$F0,$BF
	.dc.b	$8C,$33,$3F,$D6,$27,$E1,$3E,$39,$FB,$0F,$F0,$D9,$E0,$7C,$B8,$F6
	.dc.b	$3F,$DE,$0F,$8C,$74,$3F,$93,$1F,$8D,$9F,$ED,$73,$EB,$3C,$0F,$9E
	.dc.b	$3A,$1E,$73,$A1,$F3,$6B,$F3,$B3,$F1,$A3,$F1,$73,$C6,$3F,$8F,$C7
	.dc.b	$E2,$EB,$DF,$1F,$82,$33,$1F,$89,$7E,$54,$7C,$88,$FB,$07,$A4,$FD
	.dc.b	$73,$FD,$FC,$FE,$06,$F1,$7A,$9C,$5F,$C1,$47,$B5,$F8,$27,$F1,$C7
	.dc.b	$F9,$2B,$EC,$7E,$C3,$F1,$67,$D5,$3D,$A7,$F1,$CF,$69,$FE,$29,$3F
	.dc.b	$DB,$A7,$E4,$C7,$E2,$DF,$B0,$FF,$A8,$1F,$B0,$7C,$19,$E8,$7D,$87
	.dc.b	$FC,$88,$66,$7A,$4F,$A0,$FC,$D9,$EC,$3E,$21,$FC,$20,$F6,$1F,$3C
	.dc.b	$F6,$4F,$C3,$9F,$6C,$68,$F9,$06,$8F,$5C,$FF,$80,$5E,$87,$82,$38
	.dc.b	$BF,$C6,$1F,$86,$F1,$3A,$38,$3F,$8C,$CF,$F0,$77,$EF,$99,$9F,$E2
	.dc.b	$0E,$23,$E1,$47,$E8,$1E,$07,$C6,$9F,$C9,$4F,$C6,$8F,$C1,$4F,$C6
	.dc.b	$9F,$C2,$9F,$64,$FC,$49,$F9,$A9,$FA,$AF,$EA,$1F,$03,$D2,$3C,$07
	.dc.b	$F2,$67,$EC,$4F,$F5,$47,$E2,$9F,$D8,$CF,$E3,$06,$27,$F1,$67,$C0
	.dc.b	$9E,$55,$FE,$6B,$3D,$13,$B3,$EB,$4E,$1F,$93,$5F,$0A,$BF,$9C,$BF
	.dc.b	$09,$F7,$47,$D4,$9F,$CE,$3F,$00,$FE,$5D,$3F,$7C,$F8,$B3,$F6,$E3
	.dc.b	$CE,$76,$3E,$03,$D8,$F3,$BF,$D9,$CF,$8D,$1F,$46,$7E,$49,$F9,$59
	.dc.b	$E8,$FE,$E2,$7E,$81,$F9,$53,$E8,$1F,$99,$7C,$EF,$E3,$2B,$F8,$4D
	.dc.b	$7F,$1C,$3E,$74,$7A,$8F,$8C,$7E,$5A,$7E,$33,$F9,$18,$F8,$CF,$EB
	.dc.b	$DF,$8D,$5E,$47,$F6,$D1,$EF,$9F,$4C,$7C,$58,$F3,$C7,$AD,$FB,$8F
	.dc.b	$AD,$E0,$7E,$C6,$BC,$EF,$E1,$0F,$F0,$29,$FA,$59,$F9,$EF,$27,$C2
	.dc.b	$F9,$CF,$9E,$79,$EB,$E7,$BF,$D7,$0F,$9E,$FD,$E7,$EB,$9C,$6B,$D2
	.dc.b	$FA,$27,$D4,$F9,$8F,$C1,$1B,$3D,$07,$F7,$89,$EE,$FF,$5B,$3F,$A2
	.dc.b	$9F,$58,$E3,$3F,$59,$F0,$D7,$C0,$3F,$35,$3F,$D2,$23,$F8,$89,$E6
	.dc.b	$9E,$07,$C9,$7E,$21,$E7,$3E,$DB,$FC,$06,$3F,$96,$9E,$F7,$84,$F7
	.dc.b	$9F,$C9,$47,$F1,$03,$AD,$7C,$C6,$C7,$F1,$13,$B9,$F8,$93,$E2,$1F
	.dc.b	$A7,$3C,$EF,$67,$E2,$51,$57,$F9,$AA,$7F,$82,$9F,$42,$39,$9F,$10
	.dc.b	$E2,$70,$3F,$CC,$27,$CD,$3E,$DC,$FE,$4D,$F3,$47,$A0,$FE,$0D,$1E
	.dc.b	$19,$F0,$57,$EB,$EB,$F5,$07,$A1,$FC,$5C,$7C,$58,$FC,$45,$7C,$57
	.dc.b	$47,$62,$BD,$9F,$31,$FE,$4A,$3F,$BB,$9F,$82,$7D,$EF,$80,$FC,$21
	.dc.b	$FB,$73,$DF,$1E,$6A,$FB,$15,$F4,$E7,$BC,$7F,$2A,$3D,$8F,$C4,$3B
	.dc.b	$9F,$10,$FD,$31,$E7,$3B,$9F,$34,$FF,$19,$C7,$EF,$1D,$D7,$F0,$23
	.dc.b	$E5,$4F,$78,$F6,$C7,$F8,$4C,$FE,$DF,$3F,$B4,$7F,$9A,$CF,$F5,$08
	.dc.b	$F9,$D3,$FA,$8A,$F9,$8F,$D9,$AF,$98,$F9,$1F,$CE,$1E,$0A,$FE,$50
	.dc.b	$7B,$A3,$E6,$47,$81,$FC,$FD,$7F,$16,$7E,$04,$78,$A7,$8C,$7A,$27
	.dc.b	$F9,$13,$FC,$32,$7B,$3F,$B6,$3F,$2D,$23,$F3,$11,$F9,$98,$87,$FB
	.dc.b	$FC,$C3,$B8,$CC,$6A,$2C,$73,$8E,$11,$A7,$8C,$65,$C4,$70,$8F,$0C
	.dc.b	$74,$8E,$0F,$F9,$70,$C1,$E4,$8E,$D1,$CA,$36,$F9,$5A,$6E,$3A,$BE
	.dc.b	$73,$6F,$8D,$EA,$7C,$17,$AB,$83,$C8,$ED,$E4,$6D,$EA,$FC,$63,$AB
	.dc.b	$EA,$51,$EF,$7E,$29,$A3,$CE,$73,$32,$7C,$68,$F0,$3D,$4D,$47,$C1
	.dc.b	$75,$1D,$E3,$B1,$98,$F4,$C6,$5F,$88,$76,$32,$78,$A3,$F7,$2F,$F3
	.dc.b	$A3,$F6,$71,$FC,$B8,$FE,$25,$1F,$DA,$BF,$06,$FC,$2D,$7D,$16,$D1
	.dc.b	$FB,$91,$FC,$95,$FD,$84,$74,$3F,$84,$47,$EA,$9F,$E5,$C7,$E9,$DF
	.dc.b	$EC,$31,$DA,$BF,$1C,$F6,$3A,$CF,$78,$FD,$A1,$FC,$31,$FE,$09,$5E
	.dc.b	$C7,$AB,$5F,$EC,$B3,$D6,$7C,$D1,$FA,$4A,$AF,$78,$EA,$7C,$22,$BF
	.dc.b	$C3,$1B,$5F,$B5,$8E,$47,$EB,$6B,$B3,$F1,$4C,$BE,$A3,$E2,$BF,$04
	.dc.b	$F4,$D7,$C5,$7F,$B3,$1F,$24,$FC,$6B,$FB,$88,$EA,$7C,$58,$FD,$B4
	.dc.b	$7D,$1A,$F2,$C7,$F9,$21,$AF,$F8,$95,$B5,$FD,$B6,$3D,$27,$C2,$53
	.dc.b	$F7,$E3,$D2,$76,$3E,$94,$FE,$1D,$F8,$8D,$9F,$89,$1F,$C6,$8F,$14
	.dc.b	$7E,$21,$ED,$3D,$5F,$69,$F1,$CF,$A5,$1C,$DE,$27,$E2,$9F,$C0,$4F
	.dc.b	$CA,$3E,$F1,$FE,$0B,$AF,$8B,$3E,$83,$99,$C6,$7E,$E1,$D5,$F8,$8D
	.dc.b	$9F,$89,$1E,$43,$CF,$1F,$94,$7A,$1F,$71,$F9,$75,$FC,$E6,$7F,$C8
	.dc.b	$C7,$C2,$2C,$FA,$8E,$87,$BD,$F9,$F3,$D6,$3E,$1D,$7E,$1E,$3A,$C7
	.dc.b	$F9,$10,$F9,$87,$A0,$F6,$1F,$48,$FB,$67,$5A,$F9,$D5,$F7,$23,$E7
	.dc.b	$1E,$99,$F8,$13,$E9,$3F,$B1,$47,$9E,$7D,$2F,$73,$D3,$5F,$39,$F3
	.dc.b	$9F,$85,$7E,$A4,$FE,$6E,$7F,$B1,$CF,$E6,$DE,$67,$CE,$3C,$8F,$CE
	.dc.b	$AF,$C2,$BE,$63,$E1,$BF,$0A,$3F,$C5,$D1,$FE,$59,$3F,$8D,$1F,$AB
	.dc.b	$7F,$8C,$3F,$BD,$8F,$99,$1F,$06,$BE,$64,$7C,$09,$F9,$93,$E4,$AF
	.dc.b	$85,$1F,$DD,$0C,$BF,$B1,$7A,$4F,$9A,$B9,$C7,$D2,$AF,$C0,$57,$CA
	.dc.b	$8F,$91,$1D,$5E,$67,$F0,$73,$99,$F5,$CF,$80,$F9,$EB,$E0,$47,$F8
	.dc.b	$BE,$3F,$3F,$3F,$E2,$B8,$F8,$13,$FD,$C6,$7E,$5C,$7E,$06,$3E,$0C
	.dc.b	$FF,$1E,$8F,$C0,$1E,$73,$EF,$3E,$79,$FA,$8D,$3E,$1C,$7D,$53,$CD
	.dc.b	$5F,$01,$FC,$9C,$F0,$7E,$43,$F5,$8F,$F0,$AB,$F8,$83,$DF,$1F,$C6
	.dc.b	$5F,$DA,$9C,$A7,$BB,$F7,$4F,$E7,$6F,$E2,$CF,$E8,$EF,$23,$F6,$67
	.dc.b	$E1,$AB,$EE,$57,$86,$3F,$21,$3E,$B7,$E8,$9E,$BA,$F1,$3F,$10,$F1
	.dc.b	$1A,$3C,$47,$F3,$43,$E7,$D7,$F1,$B7,$E7,$3F,$83,$7F,$44,$FD,$87
	.dc.b	$F2,$B5,$F6,$1F,$E7,$F1,$FC,$68,$FB,$15,$F8,$D8,$FE,$7F,$1F,$61
	.dc.b	$FC,$1C,$7F,$3D,$8F,$6C,$7E,$0E,$3E,$23,$D2,$3F,$8E,$BF,$C0,$AB
	.dc.b	$F6,$D5,$FA,$18,$F2,$CF,$27,$CB,$5C,$A7,$F7,$33,$E5,$3D,$D5,$CA
	.dc.b	$7F,$15,$5F,$3C,$FC,$51,$F9,$E7,$E7,$57,$D1,$8F,$8E,$78,$CF,$B0
	.dc.b	$7F,$08,$3F,$02,$FE,$94,$E0,$7C,$73,$C7,$5E,$C8,$FC,$51,$F1,$9F
	.dc.b	$A0,$F2,$9F,$0B,$F1,$63,$D4,$F9,$A7,$F2,$CF,$E0,$4F,$C4,$BF,$AF
	.dc.b	$8F,$89,$1F,$90,$78,$BF,$84,$8F,$E2,$35,$F0,$E3,$EA,$9F,$D9,$0E
	.dc.b	$D5,$F9,$A9,$FD,$09,$ED,$8F,$ED,$4F,$F1,$67,$D4,$7E,$B0,$FA,$A7
	.dc.b	$23,$F2,$4F,$28,$F6,$C7,$FB,$EA,$7C,$47,$EC,$E7,$B4,$7D,$CA,$F4
	.dc.b	$C6,$23,$FC,$2E,$F8,$4F,$EA,$31,$F2,$A3,$F4,$51,$B7,$F4,$0F,$D3
	.dc.b	$8F,$CE,$47,$27,$F3,$2F,$F1,$B3,$D7,$3F,$AC,$7F,$0C,$7C,$93,$EA
	.dc.b	$47,$DB,$9E,$C7,$F7,$98,$FC,$2B,$F9,$77,$EE,$1F,$C3,$0E,$27,$F5
	.dc.b	$F3,$A1,$F0,$CE,$35,$F9,$68,$FB,$91,$F2,$DF,$C6,$C7,$E7,$E7,$D7
	.dc.b	$3F,$E4,$63,$89,$F7,$2B,$F2,$2F,$7A,$F3,$D7,$F3,$17,$E2,$57,$E9
	.dc.b	$A7,$8D,$7D,$3A,$FA,$F1,$F8,$B9,$F8,$11,$88,$F0,$1F,$20,$FC,$9B
	.dc.b	$FB,$F7,$D0,$7C,$0A,$FE,$9A,$FC,$33,$E0,$1C,$4F,$D6,$1F,$6A,$3B
	.dc.b	$47,$F7,$28,$FD,$99,$F2,$0F,$B5,$3E,$97,$94,$7D,$17,$91,$DC,$F5
	.dc.b	$1F,$0E,$7A,$9F,$CD,$1E,$2F,$D1,$3D,$4F,$73,$F1,$6F,$D1,$9F,$D7
	.dc.b	$C7,$C2,$9F,$C9,$BF,$A3,$3F,$02,$7E,$DC,$F6,$9F,$8C,$7E,$9B,$FC
	.dc.b	$84,$F2,$1F,$93,$79,$D7,$E4,$DF,$C1,$C7,$F0,$A7,$2F,$D0,$3F,$74
	.dc.b	$79,$DF,$90,$7D,$E7,$F8,$D1,$EE,$3F,$0D,$3F,$A2,$7C,$D1,$F2,$5F
	.dc.b	$35,$7C,$67,$CD,$1F,$69,$DC,$7C,$77,$C3,$1F,$DE,$AB,$F7,$B1,$FC
	.dc.b	$10,$F2,$4F,$E4,$4F,$94,$F9,$27,$D6,$FC,$67,$DE,$73,$7F,$2A,$73
	.dc.b	$7E,$49,$F9,$E3,$C8,$FF,$5C,$3E,$91,$F8,$23,$CE,$7C,$0A,$FB,$31
	.dc.b	$F1,$A7,$C7,$3F,$35,$FF,$28,$9E,$DA,$FE,$20,$E2,$39,$1F,$9E,$3F
	.dc.b	$8C,$57,$E9,$9F,$B5,$1B,$9F,$D6,$46,$27,$FB,$34,$62,$07,$F7,$8A
	.dc.b	$F2,$4F,$F2,$2A,$FE,$DD,$1F,$18,$FA,$C7,$C6,$72,$77,$9F,$8C,$FD
	.dc.b	$98,$F8,$D1,$FE,$28,$3C,$53,$FC,$A5,$FB,$E6,$0F,$92,$60,$F2,$C7
	.dc.b	$F3,$B7,$F5,$87,$E8,$67,$C0,$FE,$80,$FA,$55,$F9,$98,$FD,$5C,$FD
	.dc.b	$8A,$FE,$32,$FF,$29,$AF,$CA,$57,$C3,$78,$9F,$9D,$7D,$2F,$F7,$73
	.dc.b	$C3,$1C,$4F,$89,$5F,$5E,$BF,$39,$3C,$8F,$C6,$47,$88,$F8,$91,$E2
	.dc.b	$7F,$81,$4F,$29,$FB,$A7,$E0,$E7,$EB,$BF,$30,$F0,$CF,$E6,$E3,$EB
	.dc.b	$47,$C9,$AF,$80,$FE,$92,$7E,$AC,$F2,$8F,$8D,$5F,$2D,$F9,$A7,$F9
	.dc.b	$00,$E4,$7C,$18,$F5,$D7,$EB,$23,$89,$B3,$89,$FC,$9A,$7F,$50,$FC
	.dc.b	$C3,$D0,$7E,$8A,$3E,$74,$78,$63,$F2,$15,$F0,$8F,$B6,$FE,$A4,$F9
	.dc.b	$51,$F9,$08,$F9,$51,$F0,$8F,$E0,$A7,$E3,$4E,$C7,$80,$E2,$FD,$5A
	.dc.b	$FD,$8B,$F2,$5F,$D4,$47,$03,$F1,$55,$EF,$7C,$55,$E7,$3D,$11,$98
	.dc.b	$FC,$B1,$D8,$FB,$31,$F8,$F9,$EF,$08,$FD,$DC,$7D,$38,$F7,$4F,$39
	.dc.b	$FD,$64,$7B,$5F,$C8,$9F,$E3,$27,$B4,$7E,$5C,$E1,$5F,$12,$3E,$9C
	.dc.b	$FF,$31,$7F,$90,$9E,$1A,$FB,$6F,$F1,$C3,$DE,$53,$DE,$7D,$D3,$EA
	.dc.b	$9C,$5F,$EC,$11,$F9,$7A,$F1,$4F,$13,$E1,$C7,$C7,$79,$47,$99,$FC
	.dc.b	$A1,$F8,$E3,$F4,$A6,$0F,$4D,$7F,$24,$9F,$95,$1F,$CD,$E3,$F0,$AF
	.dc.b	$D0,$9E,$87,$C6,$8F,$F2,$54,$78,$0F,$24,$FC,$C7,$EF,$1E,$E3,$F3
	.dc.b	$11,$FB,$4A,$F6,$BD,$A7,$84,$FD,$03,$F1,$8F,$B5,$E2,$7A,$DF,$CB
	.dc.b	$3E,$B3,$81,$FA,$83,$67,$BA,$7F,$72,$F9,$0F,$61,$E8,$9F,$B8,$FD
	.dc.b	$29,$F7,$3F,$EA,$FA,$F8,$35,$F7,$23,$F6,$07,$79,$FC,$11,$F2,$9F
	.dc.b	$E1,$91,$FC,$C6,$3D,$B3,$CA,$3F,$2F,$1D,$8F,$F3,$CC,$F8,$CF,$64
	.dc.b	$ED,$FE,$A4,$7F,$87,$A3,$FB,$1C,$66,$3B,$9F,$CF,$23,$E5,$9F,$24
	.dc.b	$F2,$D7,$F1,$6A,$F2,$3F,$02,$B9,$4F,$A4,$91,$BA,$E8,$ED,$F3,$4F
	.dc.b	$A5,$FF,$21,$1F,$C1,$4F,$A6,$7E,$C0,$DD,$7F,$33,$7F,$85,$9F,$CF
	.dc.b	$CF,$A9,$5F,$19,$DC,$7C,$63,$84,$7E,$F1,$F4,$C7,$C6,$3F,$B6,$9F
	.dc.b	$DB,$91,$FE,$C4,$53,$A6,$1F,$E2,$C2,$B8,$8E,$31,$C0,$C9,$C0,$FF
	.dc.b	$24,$98,$3C,$2F,$08,$D1,$B8,$D9,$FE,$EB,$3C,$D1,$E5,$8D,$47,$8C
	.dc.b	$D1,$EA,$8D,$3C,$E3,$0D,$3D,$91,$E1,$7C,$87,$B1,$F5,$BE,$43,$6F
	.dc.b	$47,$CE,$E9,$E2,$76,$8E,$0E,$A3,$CF,$1F,$1D,$E5,$1D,$DC,$BD,$8F
	.dc.b	$94,$7B,$0F,$AF,$1E,$A7,$A9,$F1,$0D,$9F,$EB,$93,$A1,$EF,$3E,$7C
	.dc.b	$78,$1F,$B9,$1C,$DF,$A9,$3F,$CA,$5F,$EA,$D1,$F9,$EA,$FC,$E3,$FD
	.dc.b	$E8,$F1,$4A,$3F,$0D,$F2,$DF,$8C,$8F,$F4,$0B,$63,$FC,$76,$FE,$2A
	.dc.b	$38,$52,$3F,$2C,$FE,$A0,$FC,$B1,$FC,$20,$FD,$21,$4F,$EB,$AF,$5B
	.dc.b	$FC,$8B,$DE,$3A,$1F,$99,$9E,$F1,$DA,$FD,$55,$FC,$56,$7E,$05,$7F
	.dc.b	$50,$3B,$3F,$CB,$AB,$B4,$FA,$A7,$FC,$83,$75,$F0,$4F,$57,$FC,$2D
	.dc.b	$3F,$CE,$0F,$ED,$8F,$F2,$A3,$AC,$FF,$2A,$9E,$B1,$E8,$3F,$BA,$47
	.dc.b	$57,$B9,$F1,$E7,$D1,$1F,$D0,$CF,$E1,$D1,$D0,$F9,$35,$EC,$7F,$C3
	.dc.b	$15,$D8,$F3,$BF,$16,$7E,$AD,$74,$3C,$F5,$FC,$36,$7A,$5F,$B5,$FE
	.dc.b	$CA,$FE,$0A,$BF,$03,$1F,$12,$BE,$9B,$E8,$BF,$C0,$4F,$69,$F8,$87
	.dc.b	$F1,$4A,$E9,$7F,$D5,$9B,$3F,$E2,$5B,$F8,$27,$CF,$9F,$C6,$BD,$4F
	.dc.b	$51,$F1,$6B,$AB,$F8,$BA,$E6,$FF,$88,$AB,$F3,$6F,$D6,$3A,$9F,$46
	.dc.b	$BA,$BF,$48,$FA,$87,$AE,$BE,$31,$E9,$2B,$DD,$B5,$F8,$E7,$F2,$91
	.dc.b	$E9,$AF,$E2,$77,$F4,$A3,$F1,$33,$C1,$F2,$1F,$D5,$AB,$93,$FE,$0F
	.dc.b	$8F,$AB,$5F,$89,$8F,$A1,$1D,$5F,$39,$E9,$AF,$83,$5F,$E2,$48,$FB
	.dc.b	$73,$DD,$FB,$6F,$FA,$1A,$7F,$0D,$1F,$34,$F2,$C7,$CC,$9F,$1C,$79
	.dc.b	$4F,$3D,$FF,$9B,$8F,$C8,$3F,$B6,$3C,$A7,$CD,$BF,$CE,$3E,$5A,$F1
	.dc.b	$4F,$E0,$5B,$3F,$81,$9E,$C7,$9A,$3E,$5B,$CD,$FC,$7B,$EF,$3F,$07
	.dc.b	$7F,$01,$FE,$36,$7F,$66,$3E,$54,$7D,$93,$9C,$7C,$08,$E4,$7C,$03
	.dc.b	$67,$DD,$8F,$84,$FC,$09,$FC,$5B,$E5,$3F,$17,$3C,$EB,$F8,$8C,$79
	.dc.b	$2B,$DA,$FE,$4C,$FF,$1B,$57,$BE,$7E,$53,$FB,$D8,$FF,$1C,$57,$38
	.dc.b	$F9,$33,$FC,$72,$B9,$DF,$DD,$8F,$A2,$7E,$E4,$B3,$F8,$7B,$FE,$1B
	.dc.b	$5F,$22,$7C,$B1,$EF,$3E,$7B,$F4,$E3,$DF,$1E,$D3,$33,$E7,$3F,$80
	.dc.b	$BD,$23,$F2,$57,$FC,$02,$3F,$68,$7C,$AB,$FC,$4D,$FB,$AB,$F3,$73
	.dc.b	$FC,$62,$FE,$05,$7F,$0D,$8F,$F0,$BD,$FC,$17,$F3,$B7,$F0,$6F,$CA
	.dc.b	$F9,$E3,$B1,$F5,$4F,$9A,$7C,$F3,$F6,$51,$F0,$6F,$E5,$9F,$C5,$63
	.dc.b	$89,$EF,$9F,$34,$FE,$1E,$33,$1D,$DF,$E0,$C7,$48,$F8,$2F,$DF,$7C
	.dc.b	$B3,$F1,$E3,$C7,$7F,$5E,$3D,$F3,$FB,$19,$FD,$3D,$FF,$05,$7F,$88
	.dc.b	$47,$27,$D4,$7F,$92,$0F,$E9,$97,$F7,$9F,$C3,$9C,$8F,$DD,$1C,$8F
	.dc.b	$4D,$78,$DF,$F2,$5B,$FD,$2A,$39,$5F,$D1,$9F,$61,$FB,$93,$E8,$CF
	.dc.b	$C7,$7F,$06,$79,$23,$FC,$10,$FB,$8F,$F7,$05,$FD,$93,$F0,$87,$ED
	.dc.b	$5F,$EE,$15,$E0,$BF,$CD,$C7,$03,$D4,$7E,$D0,$F5,$1E,$0A,$F9,$51
	.dc.b	$F8,$A7,$A4,$7F,$60,$38,$CF,$E7,$9E,$2F,$DC,$7F,$78,$7C,$43,$D4
	.dc.b	$7F,$07,$3E,$E9,$FC,$3E,$7E,$EB,$FA,$78,$E5,$3F,$9C,$BF,$E0,$71
	.dc.b	$F8,$53,$33,$F9,$33,$E5,$DF,$DD,$7F,$91,$5F,$DD,$BF,$5B,$F8,$F8
	.dc.b	$F7,$1F,$0C,$FD,$71,$F4,$E7,$E5,$C7,$C8,$8F,$97,$7F,$8C,$8E,$91
	.dc.b	$C6,$FF,$AE,$9F,$0D,$F0,$9F,$06,$FC,$2F,$C4,$3C,$27,$F8,$B5,$FC
	.dc.b	$23,$F0,$CF,$11,$F9,$F3,$91,$FC,$C0,$FE,$9F,$3E,$27,$93,$F9,$F7
	.dc.b	$D7,$7F,$0E,$BF,$79,$1C,$8F,$5C,$FD,$D3,$DC,$FD,$D3,$89,$F9,$E3
	.dc.b	$D6,$7A,$0F,$F2,$4C,$7E,$36,$3C,$51,$C6,$3E,$1D,$FF,$02,$8F,$E2
	.dc.b	$E7,$13,$F5,$AF,$CB,$7C,$87,$17,$FB,$DB,$E1,$8F,$4B,$C5,$FE,$65
	.dc.b	$1C,$63,$0F,$F3,$38,$F0,$C7,$EB,$0F,$0B,$F8,$87,$F8,$94,$7E,$30
	.dc.b	$FD,$C3,$FC,$06,$3F,$CA,$CF,$E7,$A3,$D5,$5F,$A2,$8F,$4D,$7C,$93
	.dc.b	$E0,$BF,$DD,$4F,$92,$7E,$B4,$F9,$35,$FE,$76,$7F,$2E,$7B,$6B,$F8
	.dc.b	$04,$79,$4F,$CF,$3E,$58,$FE,$2B,$5F,$BF,$7F,$08,$7E,$A9,$C3,$FA
	.dc.b	$98,$E1,$7F,$90,$3F,$55,$1F,$D0,$0F,$61,$F1,$67,$F2,$27,$27,$F7
	.dc.b	$73,$C9,$FC,$85,$7D,$F7,$FB,$99,$CA,$7F,$4F,$7F,$E0,$08,$C1,$E6
	.dc.b	$9F,$65,$7C,$EB,$FE,$37,$3F,$08,$F8,$D3,$F7,$8F,$8C,$FE,$8E,$FF
	.dc.b	$35,$5F,$18,$F2,$47,$E6,$AF,$F7,$95,$C2,$BD,$53,$F4,$2B,$FB,$5C
	.dc.b	$FE,$12,$7C,$F3,$F9,$03,$ED,$D7,$2A,$D9,$F9,$28,$F2,$BE,$17,$E5
	.dc.b	$1F,$7A,$BF,$80,$1E,$98,$F2,$C7,$E9,$9F,$CB,$47,$96,$3E,$29,$F5
	.dc.b	$CE,$6F,$95,$F9,$11,$C9,$F2,$9F,$C2,$63,$93,$E1,$8F,$CA,$CF,$F0
	.dc.b	$19,$E5,$3D,$4F,$EA,$67,$F8,$0E,$3F,$30,$FD,$A3,$91,$F8,$37,$94
	.dc.b	$FC,$19,$F6,$BF,$06,$79,$DF,$E8,$E7,$F4,$F1,$CA,$FE,$6B,$F9,$53
	.dc.b	$DE,$7E,$30,$FE,$B4,$7D,$93,$D9,$5F,$42,$FF,$C1,$51,$E3,$7F,$18
	.dc.b	$F5,$3E,$29,$C8,$F9,$CF,$B1,$F0,$1B,$3E,$11,$F8,$B3,$F6,$07,$F0
	.dc.b	$48,$F9,$D5,$FC,$32,$79,$3F,$38,$F9,$73,$E9,$3F,$30,$FC,$47,$91
	.dc.b	$E3,$3E,$BD,$FE,$C1,$DD,$7D,$6A,$F5,$D7,$49,$FC,$B1,$C5,$F1,$47
	.dc.b	$E4,$E3,$F0,$B3,$F4,$CE,$4F,$D9,$9E,$47,$E4,$5E,$51,$FE,$11,$3F
	.dc.b	$22,$7C,$D9,$FE,$DF,$1E,$B3,$E2,$1F,$9E,$3D,$67,$DE,$7F,$A2,$BF
	.dc.b	$C7,$67,$BC,$FE,$0E,$38,$C7,$C3,$BF,$C2,$4F,$86,$BE,$5B,$E1,$AF
	.dc.b	$EB,$A7,$D4,$9F,$DB,$D7,$A6,$3C,$15,$FC,$B2,$3F,$1A,$78,$23,$CA
	.dc.b	$78,$2B,$E1,$BF,$D0,$8F,$CF,$4F,$C9,$7E,$89,$F8,$B3,$E4,$47,$CB
	.dc.b	$70,$71,$BF,$4B,$FE,$40,$3F,$14,$7C,$7B,$F3,$CF,$C1,$9F,$E3,$73
	.dc.b	$FD,$26,$3F,$A9,$1D,$67,$F4,$77,$F8,$B8,$C4,$FE,$68,$F7,$9F,$C4
	.dc.b	$E3,$DF,$1A,$8F,$8C,$7E,$A8,$FC,$39,$C0,$F4,$5F,$EE,$0F,$D4,$D7
	.dc.b	$82,$7E,$83,$FC,$AA,$7E,$4D,$FA,$0F,$4C,$7B,$23,$E6,$B4,$F8,$71
	.dc.b	$2B,$B1,$F8,$A9,$F1,$9F,$4C,$F9,$77,$FD,$01,$FF,$20,$9F,$E1,$83
	.dc.b	$E8,$C7,$99,$F8,$47,$1A,$F4,$3F,$9A,$3E,$D9,$FC,$18,$FD,$1C,$71
	.dc.b	$7E,$29,$F5,$8F,$1D,$7B,$5F,$F2,$85,$7D,$19,$E1,$1F,$6C,$F9,$91
	.dc.b	$F9,$17,$ED,$1E,$09,$FD,$D3,$E5,$3C,$F5,$F1,$6B,$F4,$B5,$FB,$89
	.dc.b	$F3,$1F,$90,$9F,$5B,$4F,$7D,$79,$5F,$88,$7F,$18,$9F,$CF,$4F,$F7
	.dc.b	$48,$F2,$DF,$E3,$DF,$5B,$FC,$60,$FC,$7D,$FB,$23,$E9,$CF,$96,$3F
	.dc.b	$55,$3F,$B3,$7F,$49,$3E,$B3,$E4,$D7,$82,$3D,$73,$B9,$F5,$9B,$3E
	.dc.b	$41,$F3,$E7,$E4,$5F,$AD,$FF,$03,$47,$AE,$7F,$34,$F0,$8E,$D5,$E4
	.dc.b	$8F,$9C,$7F,$9E,$CE,$C7,$A8,$F8,$27,$E6,$0F,$82,$7E,$16,$BE,$33
	.dc.b	$E3,$AF,$E2,$97,$F0,$63,$F5,$91,$EB,$9F,$B7,$39,$9F,$35,$7C,$9B
	.dc.b	$F5,$C7,$63,$F5,$CF,$E2,$63,$F4,$0F,$E1,$E7,$93,$F9,$FA,$FF,$18
	.dc.b	$57,$AE,$73,$7F,$B2,$8F,$EA,$31,$FA,$BB,$F8,$55,$EA,$BF,$8B,$5F
	.dc.b	$DD,$41,$FE,$13,$3F,$C2,$26,$0F,$F7,$DC,$8C,$30,$55,$8D,$46,$0D
	.dc.b	$46,$DE,$71,$5F,$0C,$71,$8B,$1C,$0D,$47,$86,$38,$47,$53,$CC,$F8
	.dc.b	$23,$B1,$A3,$84,$72,$3D,$31,$B7,$B1,$C1,$F4,$BC,$8D,$BE,$47,$E0
	.dc.b	$C7,$58,$DC,$72,$3E,$5B,$D5,$D9,$EA,$76,$E0,$F4,$BD,$5F,$A2,$76
	.dc.b	$76,$E1,$E8,$E8,$F9,$67,$B4,$E6,$7C,$53,$47,$FC,$01,$3C,$C7,$DF
	.dc.b	$3C,$07,$77,$E4,$3F,$00,$E4,$68,$EE,$7C,$53,$F4,$87,$F3,$C3,$F5
	.dc.b	$71,$F8,$FB,$FD,$AC,$7F,$3A,$FC,$1B,$F1,$96,$9F,$AB,$18,$7F,$26
	.dc.b	$FE,$36,$35,$28,$E9,$7F,$AB,$AF,$D9,$CF,$82,$3F,$A4,$BF,$E0,$23
	.dc.b	$FC,$25,$3D,$63,$F2,$AF,$78,$E8,$7E,$72,$BB,$C7,$69,$FD,$9D,$76
	.dc.b	$AF,$7D,$E5,$EC,$7B,$4F,$D5,$DF,$B6,$BB,$C6,$2B,$DA,$F7,$7D,$F1
	.dc.b	$E8,$AF,$ED,$47,$52,$B9,$9F,$DE,$BF,$96,$AF,$F4,$DD,$FB,$63,$AC
	.dc.b	$A7,$D6,$F7,$3E,$54,$73,$7F,$C6,$95,$DD,$EB,$1F,$E4,$AB,$FB,$35
	.dc.b	$FD,$66,$FB,$57,$C8,$3B,$4F,$81,$F7,$1D,$AB,$91,$F5,$DF,$C3,$1F
	.dc.b	$29,$F1,$57,$F8,$F5,$F6,$4F,$F0,$EA,$EE,$7D,$73,$F0,$8F,$C8,$8E
	.dc.b	$A7,$D6,$AE,$B5,$FC,$79,$FE,$63,$19,$BF,$E0,$71,$E8,$3F,$8F,$5F
	.dc.b	$67,$D2,$F6,$7D,$47,$F8,$32,$79,$1E,$8B,$FF,$22,$DF,$F3,$99,$FC
	.dc.b	$1B,$F1,$DF,$C0,$4F,$E0,$9F,$05,$75,$3E,$83,$F4,$A3,$A3,$FE,$1E
	.dc.b	$AF,$E7,$8E,$63,$F3,$0F,$DE,$AF,$C7,$45,$9F,$A5,$3D,$A3,$9D,$FF
	.dc.b	$81,$A3,$FB,$0D,$7F,$4D,$7F,$51,$3D,$1F,$B9,$1D,$1B,$5D,$23,$F4
	.dc.b	$B3,$D2,$7F,$C2,$E7,$23,$E3,$1F,$0A,$7F,$CF,$57,$F3,$63,$E0,$9F
	.dc.b	$93,$7E,$0D,$7E,$06,$BE,$DC,$F1,$3E,$DD,$7E,$48,$E6,$FC,$C8,$F8
	.dc.b	$0F,$E4,$CF,$B2,$F9,$5F,$F0,$8C,$71,$79,$C7,$E4,$EF,$F0,$75,$FA
	.dc.b	$A9,$E8,$7C,$03,$A1,$F2,$E7,$CB,$3F,$B4,$6C,$7E,$A0,$F8,$31,$97
	.dc.b	$F4,$67,$D5,$8F,$C1,$4F,$ED,$23,$F3,$91,$E7,$7F,$C3,$55,$E7,$AF
	.dc.b	$8A,$FE,$38,$FD,$24,$79,$AF,$F4,$8F,$CB,$AE,$4F,$F8,$74,$FD,$73
	.dc.b	$E6,$7F,$14,$79,$0F,$BA,$F9,$27,$EF,$1F,$AD,$8F,$59,$ED,$3F,$56
	.dc.b	$7D,$5A,$F2,$BF,$97,$3F,$56,$79,$4F,$D5,$9E,$59,$F1,$5F,$DF,$8E
	.dc.b	$A7,$90,$EB,$3F,$25,$FD,$B1,$F2,$67,$EF,$DF,$EA,$5F,$7D,$7E,$6E
	.dc.b	$7D,$E7,$47,$B3,$C8,$F7,$C7,$27,$C8,$6E,$FE,$73,$FE,$82,$8F,$DF
	.dc.b	$9E,$88,$CC,$78,$CF,$C3,$57,$8E,$75,$1F,$0C,$FF,$3A,$1F,$92,$7F
	.dc.b	$20,$7F,$1E,$7E,$39,$FC,$A9,$FA,$4F,$E4,$2B,$F5,$F1,$D2,$FF,$74
	.dc.b	$7F,$90,$27,$CC,$7D,$38,$E8,$F9,$6B,$F0,$AF,$E3,$A3,$93,$FB,$83
	.dc.b	$E3,$9F,$40,$FB,$D5,$FA,$37,$C9,$1F,$1D,$E4,$7C,$73,$DC,$7F,$0C
	.dc.b	$BF,$EC,$D3,$FC,$C4,$E4,$7F,$0B,$AF,$1C,$FE,$DA,$BC,$6F,$C5,$7E
	.dc.b	$7B,$D6,$FF,$8C,$4F,$F0,$9A,$FD,$44,$F1,$34,$7A,$DF,$05,$FF,$28
	.dc.b	$7F,$02,$78,$9F,$F0,$7C,$7E,$50,$F5,$9F,$04,$FC,$94,$7F,$3B,$AF
	.dc.b	$C9,$36,$3E,$2C,$FE,$C9,$F8,$A7,$E0,$8F,$C5,$9C,$E3,$C3,$1E,$33
	.dc.b	$F1,$31,$F1,$CF,$F3,$4D,$FD,$87,$F2,$A7,$33,$CB,$3E,$17,$CB,$3C
	.dc.b	$DF,$A3,$5F,$83,$7F,$00,$7B,$4F,$8C,$78,$E3,$C3,$1B,$8F,$13,$FC
	.dc.b	$8A,$3D,$93,$F5,$E7,$91,$FC,$84,$F9,$E7,$D7,$3F,$21,$5F,$69,$FF
	.dc.b	$17,$9F,$C0,$4F,$E7,$E7,$AA,$7D,$15,$F9,$A8,$FE,$90,$79,$4F,$C5
	.dc.b	$DF,$DF,$9F,$E2,$71,$FD,$C8,$D4,$F1,$7F,$33,$7F,$14,$F9,$EF,$EB
	.dc.b	$1F,$53,$E2,$3E,$1B,$B9,$F8,$71,$E2,$8C,$3E,$23,$D4,$FE,$41,$F8
	.dc.b	$47,$88,$FC,$84,$7F,$37,$BF,$E2,$6F,$F5,$68,$FC,$2D,$FE,$81,$E3
	.dc.b	$5D,$AF,$D4,$FE,$62,$3F,$77,$1F,$90,$AF,$A6,$7E,$62,$3D,$47,$27
	.dc.b	$E5,$BE,$2B,$FE,$1A,$7D,$37,$FC,$A8,$78,$8F,$B7,$5F,$C3,$1F,$B6
	.dc.b	$7F,$50,$8E,$2E,$0E,$31,$F2,$8F,$84,$71,$7E,$11,$F8,$3A,$F4,$3E
	.dc.b	$A3,$F4,$E7,$D3,$AF,$54,$F2,$7F,$A7,$47,$23,$C5,$5F,$C0,$9C,$1E
	.dc.b	$28,$E2,$7E,$12,$FF,$A8,$DF,$CA,$8F,$A5,$5F,$06,$FF,$C6,$57,$F0
	.dc.b	$4F,$92,$FA,$4F,$A4,$FC,$98,$F0,$46,$23,$FD,$0D,$5F,$A5,$7D,$53
	.dc.b	$FC,$DA,$FF,$A0,$47,$F6,$B7,$F3,$0F,$F3,$9B,$F6,$DF,$38,$F6,$9F
	.dc.b	$A8,$38,$1F,$8E,$8F,$65,$7F,$03,$AE,$67,$69,$F8,$75,$FD,$7C,$FC
	.dc.b	$43,$FC,$00,$F6,$D7,$80,$FC,$4D,$79,$E3,$F8,$AC,$F8,$DF,$F0,$33
	.dc.b	$C9,$FE,$23,$5F,$97,$8F,$8D,$5C,$AB,$FA,$15,$7E,$AA,$3F,$22,$FE
	.dc.b	$04,$FB,$E7,$8E,$7E,$F3,$FB,$97,$F9,$94,$F8,$E3,$F8,$21,$F8,$73
	.dc.b	$F8,$E9,$63,$F7,$6F,$39,$F8,$B3,$FE,$57,$AF,$8B,$3E,$C9,$F1,$BF
	.dc.b	$3A,$FF,$26,$F1,$3E,$A1,$E3,$8F,$63,$FC,$BC,$F6,$1E,$17,$F8,$2B
	.dc.b	$FC,$39,$DC,$FC,$57,$E7,$47,$E1,$2B,$E2,$BF,$B5,$39,$1F,$CD,$8F
	.dc.b	$E1,$D1,$C9,$FD,$9B,$C9,$F3,$4F,$DF,$BF,$AC,$7D,$F9,$FC,$EC,$FE
	.dc.b	$20,$FC,$E4,$7C,$E9,$FC,$B4,$72,$3E,$21,$E2,$BF,$D3,$5F,$EF,$A3
	.dc.b	$FC,$49,$3F,$29,$FA,$67,$12,$9F,$D1,$A3,$F8,$EB,$FD,$61,$FA,$4F
	.dc.b	$E6,$4F,$9B,$1E,$27,$F1,$67,$B8,$FA,$07,$8C,$FD,$81,$FA,$33,$C6
	.dc.b	$73,$7E,$15,$FC,$78,$FC,$E9,$C8,$FF,$16,$9F,$7C,$FC,$89,$E0,$37
	.dc.b	$1E,$33,$DC,$FD,$E3,$89,$FA,$28,$F9,$AF,$88,$F9,$AF,$E4,$1F,$A1
	.dc.b	$7F,$C8,$E3,$11,$FB,$47,$FC,$42,$72,$8F,$CF,$D7,$E1,$1F,$D5,$CF
	.dc.b	$E1,$CF,$C8,$1F,$87,$3C,$6F,$DF,$8F,$AC,$F2,$8D,$BC,$9F,$AC,$7A
	.dc.b	$CF,$AD,$1F,$C8,$67,$FA,$94,$F9,$E7,$EA,$CF,$F0,$D7,$E7,$1F,$86
	.dc.b	$3C,$33,$EA,$8F,$0B,$F4,$67,$F1,$D1,$C6,$7E,$29,$D1,$F0,$3F,$EA
	.dc.b	$E7,$F9,$5C,$F1,$3E,$3C,$FA,$EB,$F5,$55,$F6,$CF,$49,$FD,$46,$3F
	.dc.b	$9C,$9E,$98,$FC,$1C,$7C,$97,$89,$FD,$09,$FC,$5D,$FF,$00,$7F,$05
	.dc.b	$3E,$8B,$FC,$BC,$7A,$2B,$9B,$F0,$0F,$41,$F9,$39,$F3,$3F,$99,$8F
	.dc.b	$C5,$3F,$20,$FD,$2D,$7F,$8A,$CF,$F1,$64,$7E,$3A,$3E,$E1,$E2,$BF
	.dc.b	$82,$7F,$16,$7F,$21,$7F,$39,$F6,$1D,$DF,$9D,$7F,$06,$3E,$B1,$F1
	.dc.b	$8F,$C8,$1F,$B3,$3F,$2B,$3F,$D9,$9F,$A8,$62,$7F,$98,$3F,$87,$BF
	.dc.b	$43,$FC,$50,$F0,$9F,$A8,$8F,$0C,$F3,$BF,$8F,$3E,$1B,$FE,$2B,$3F
	.dc.b	$8E,$35,$1E,$69,$F8,$53,$F9,$78,$E1,$3F,$06,$3E,$EB,$C2,$3F,$87
	.dc.b	$CF,$08,$FE,$49,$1C,$0F,$E1,$07,$F7,$4A,$F2,$9E,$C7,$E7,$CF,$17
	.dc.b	$F1,$97,$EE,$7F,$9D,$1E,$EB,$FC,$23,$F1,$6B,$F6,$D7,$E0,$AE,$15
	.dc.b	$FD,$96,$FA,$1F,$82,$BF,$C5,$1F,$52,$77,$5F,$C0,$4F,$D0,$47,$D2
	.dc.b	$3F,$86,$57,$DB,$7F,$12,$7D,$13,$F2,$53,$D6,$35,$1F,$BD,$AF,$0D
	.dc.b	$FF,$32,$AF,$D0,$4F,$C3,$8F,$CE,$D7,$08,$F8,$67,$E5,$CF,$C0,$1D
	.dc.b	$CF,$98,$7F,$0A,$36,$7B,$0F,$29,$F2,$0F,$2C,$7D,$03,$E3,$BF,$B9
	.dc.b	$AF,$2C,$7F,$02,$AF,$2C,$FF,$A1,$0E,$87,$AC,$F2,$1F,$A6,$7E,$4C
	.dc.b	$7E,$8E,$3C,$95,$F8,$9A,$EB,$1F,$C1,$6F,$F8,$B5,$7F,$0E,$8F,$B9
	.dc.b	$1F,$AE,$7F,$DD,$A0,$FF,$35,$18,$30,$7F,$94,$8A,$6E,$6B,$C0,$C1
	.dc.b	$83,$81,$FE,$C3,$3B,$48,$DB,$93,$AC,$66,$39,$46,$8E,$2D,$61,$E6
	.dc.b	$8E,$0F,$07,$CB,$1D,$E1,$1E,$53,$AB,$FE,$A9,$3D,$6F,$08,$EA,$E8
	.dc.b	$EB,$1C,$4F,$87,$1D,$63,$11,$D6,$3C,$0E,$0F,$49,$E2,$3B,$9E,$77
	.dc.b	$6E,$0F,$A2,$F5,$70,$79,$9D,$BF,$25,$D3,$F2,$CF,$88,$E9,$FB,$8E
	.dc.b	$9F,$BC,$69,$F7,$3E,$C7,$E0,$BE,$57,$2F,$E5,$8F,$D1,$D7,$EA,$CF
	.dc.b	$C9,$3F,$98,$7F,$8E,$7E,$0D,$F8,$A5,$7F,$5A,$30,$FE,$45,$FC,$5C
	.dc.b	$6A,$51,$FB,$F9,$FE,$45,$1F,$A5,$3C,$53,$E0,$6C,$7F,$24,$9E,$B5
	.dc.b	$F8,$17,$BC,$F4,$3F,$37,$5D,$E3,$B4,$FF,$27,$8E,$87,$B6,$BF,$97
	.dc.b	$DF,$EB,$AD,$3E,$77,$D4,$7E,$AD,$F7,$D7,$86,$31,$5A,$7C,$77,$E0
	.dc.b	$9F,$E0,$F7,$67,$FC,$37,$5E,$07,$E6,$3E,$D7,$9C,$F5,$7C,$71,$D5
	.dc.b	$EE,$7C,$A8,$FE,$28,$F4,$3F,$8A,$1D,$0A,$7C,$A3,$FC,$21,$5F,$2A
	.dc.b	$3D,$17,$F1,$E7,$E4,$D7,$D9,$9E,$B7,$FE,$10,$6C,$FF,$77,$7F,$98
	.dc.b	$CF,$DA,$8F,$F0,$0D,$F5,$7E,$A1,$EA,$29,$EA,$7E,$25,$75,$AE,$67
	.dc.b	$D7,$AF,$8D,$5E,$99,$FF,$37,$57,$53,$E7,$C7,$A0,$F8,$95,$C8,$FA
	.dc.b	$27,$A1,$FE,$FC,$E5,$F9,$13,$D1,$FA,$31,$F1,$2B,$E8,$47,$A6,$BC
	.dc.b	$EF,$C1,$3F,$B5,$1C,$A7,$B5,$FF,$6E,$8F,$80,$73,$8F,$B9,$1F,$3E
	.dc.b	$3F,$0F,$1C,$2B,$F9,$D4,$7F,$42,$7E,$0C,$FE,$5E,$3E,$C9,$E6,$3E
	.dc.b	$6B,$CD,$F9,$93,$CD,$FE,$E5,$1F,$E3,$C7,$E0,$9E,$8B,$FA,$0F,$E3
	.dc.b	$0F,$34,$FC,$17,$EB,$9F,$05,$E9,$1F,$E2,$43,$F5,$13,$D4,$F8,$07
	.dc.b	$EA,$67,$CB,$5F,$C1,$5F,$C0,$1F,$5E,$FE,$5B,$E4,$3F,$C0,$F1,$F8
	.dc.b	$B9,$FD,$BB,$4F,$D3,$47,$E9,$2F,$E5,$C7,$D6,$9F,$E0,$11,$E6,$9F
	.dc.b	$F1,$A1,$FA,$3B,$EF,$3E,$6A,$FE,$C3,$7F,$8B,$AF,$29,$F5,$2F,$F0
	.dc.b	$D7,$F1,$A3,$CB,$1F,$95,$7F,$5B,$7F,$56,$7F,$15,$1E,$59,$F9,$55
	.dc.b	$FC,$8A,$FF,$59,$1D,$27,$F1,$47,$47,$F2,$E6,$5F,$89,$7F,$2E,$BC
	.dc.b	$D1,$F9,$33,$E5,$4F,$BD,$FF,$25,$9E,$F7,$0F,$FA,$9E,$38,$CF,$E1
	.dc.b	$E3,$EE,$4F,$78,$F8,$91,$FE,$0C,$3F,$13,$5F,$E3,$CB,$FF,$31,$DF
	.dc.b	$28,$FA,$07,$B2,$7F,$18,$7B,$2B,$F0,$91,$F9,$C8,$FF,$42,$1D,$0F
	.dc.b	$F0,$9C,$F3,$7F,$1F,$3F,$6A,$FC,$95,$C6,$FF,$1E,$F8,$EB,$FD,$0B
	.dc.b	$7C,$4F,$9C,$FF,$02,$3C,$67,$CA,$8E,$0F,$C6,$3F,$86,$57,$F7,$CB
	.dc.b	$F9,$D1,$FA,$DA,$FE,$1A,$FE,$0C,$E7,$1F,$19,$FD,$B1,$F8,$2A,$FC
	.dc.b	$B3,$FB,$63,$F0,$47,$E8,$8F,$6B,$F1,$DF,$C0,$1F,$C2,$6B,$DE,$FC
	.dc.b	$53,$AD,$FE,$32,$BF,$A8,$CF,$E3,$27,$F0,$2F,$09,$FD,$A4,$7E,$06
	.dc.b	$30,$FC,$63,$D8,$7E,$78,$FE,$0E,$E9,$F8,$93,$F9,$B8,$FC,$4C,$7F
	.dc.b	$3C,$38,$9F,$91,$3E,$21,$EA,$9F,$D3,$4F,$E0,$23,$D4,$7E,$02,$3F
	.dc.b	$12,$F2,$8F,$C9,$1F,$7D,$F9,$B5,$F7,$DF,$51,$83,$9D,$FD,$F7,$E0
	.dc.b	$99,$AF,$82,$FD,$CB,$F6,$3F,$A1,$3C,$15,$F9,$87,$F0,$C7,$A4,$FC
	.dc.b	$C3,$F8,$67,$FC,$29,$7F,$0E,$7F,$C1,$F1,$FC,$98,$FB,$97,$F9,$03
	.dc.b	$F8,$6D,$7A,$63,$C4,$79,$4F,$A4,$7E,$12,$3F,$1A,$FD,$23,$47,$D2
	.dc.b	$8F,$C5,$DF,$48,$F8,$27,$A8,$FC,$59,$EA,$BF,$A1,$1F,$09,$F4,$47
	.dc.b	$43,$E1,$1E,$13,$F0,$47,$F0,$23,$47,$F9,$0A,$7C,$37,$F8,$29,$FF
	.dc.b	$20,$CF,$86,$FE,$04,$E6,$BF,$AB,$9C,$1F,$A3,$3F,$89,$3C,$07,$E2
	.dc.b	$4F,$80,$7D,$13,$F6,$C7,$F1,$1B,$F5,$4F,$38,$FE,$AC,$7F,$4C,$9F
	.dc.b	$E8,$2F,$C2,$AF,$8D,$5C,$27,$F9,$B3,$89,$FA,$37,$FA,$A7,$EC,$9F
	.dc.b	$8A,$3C,$4F,$EF,$5F,$CC,$1F,$D2,$8F,$B6,$7F,$7C,$3C,$53,$FB,$37
	.dc.b	$F4,$71,$F4,$9E,$0F,$D2,$7A,$DF,$F2,$3B,$FC,$D4,$70,$8F,$C8,$3F
	.dc.b	$46,$BD,$27,$CD,$7F,$86,$57,$A6,$BF,$13,$1F,$64,$F2,$3D,$1F,$9B
	.dc.b	$1E,$37,$F1,$93,$D0,$FC,$DB,$EA,$3E,$19,$E3,$3F,$9A,$9F,$7C,$F8
	.dc.b	$D1,$CD,$FD,$2D,$7D,$57,$C2,$6C,$E4,$7E,$9E,$BC,$F3,$FB,$F8,$F8
	.dc.b	$67,$F0,$6B,$FE,$7A,$F8,$0F,$C3,$C7,$F0,$6A,$FE,$50,$FF,$36,$8C
	.dc.b	$C7,$EE,$EB,$EF,$47,$BA,$FF,$B5,$BF,$88,$8F,$88,$FC,$7A,$E4,$7C
	.dc.b	$B7,$F4,$6F,$94,$F7,$3F,$4E,$79,$BE,$E3,$C2,$7E,$45,$FE,$95,$5F
	.dc.b	$E0,$C8,$F8,$47,$D2,$7E,$EC,$FF,$87,$27,$E9,$4F,$94,$D9,$E5,$9E
	.dc.b	$4F,$6A,$FD,$7C,$FE,$B2,$B9,$3E,$47,$D7,$5F,$E2,$FA,$FD,$71,$CC
	.dc.b	$F9,$91,$EF,$72,$FB,$E3,$CA,$7E,$EA,$7C,$B3,$EF,$9E,$07,$D3,$7F
	.dc.b	$88,$4F,$9A,$32,$FD,$39,$FC,$C5,$FD,$4A,$FE,$23,$5A,$7E,$44,$FE
	.dc.b	$4A,$FE,$AC,$F8,$E7,$F8,$11,$FA,$AB,$FC,$74,$7B,$AB,$F8,$0D,$7C
	.dc.b	$4A,$F9,$47,$C7,$3F,$2F,$5E,$C7,$FB,$B4,$FB,$2B,$93,$FB,$98,$FA
	.dc.b	$C7,$C6,$9F,$95,$3E,$FA,$F5,$4F,$E0,$CF,$EF,$07,$82,$3D,$91,$E3
	.dc.b	$8F,$91,$3C,$8F,$89,$5F,$65,$F9,$CF,$D9,$7E,$51,$B8,$F8,$A7,$57
	.dc.b	$13,$D6,$BF,$2A,$7E,$0A,$7C,$67,$82,$BE,$2B,$F0,$8F,$63,$87,$F6
	.dc.b	$A7,$F1,$E3,$F5,$07,$28,$F8,$AF,$F0,$43,$E2,$9F,$BB,$3F,$4F,$5F
	.dc.b	$B4,$3E,$75,$FC,$69,$FE,$8F,$3F,$4C,$EF,$7F,$7A,$B8,$47,$DE,$3C
	.dc.b	$35,$FE,$0A,$8F,$87,$7E,$79,$F0,$9F,$C5,$EB,$F8,$5C,$F2,$8F,$F1
	.dc.b	$3C,$FE,$40,$FD,$9B,$F0,$1F,$BC,$FE,$8C,$F5,$1F,$A2,$8E,$4F,$C3
	.dc.b	$3F,$8E,$C7,$E1,$E3,$C6,$F2,$BF,$CB,$4F,$27,$F0,$D1,$E1,$3E,$F4
	.dc.b	$7F,$38,$9F,$B2,$FD,$33,$83,$F8,$47,$FB,$8C,$FD,$BB,$F8,$51,$E9
	.dc.b	$3E,$91,$E9,$3E,$11,$F4,$8F,$D9,$9F,$2C,$F4,$D7,$E3,$5E,$57,$F2
	.dc.b	$1D,$9E,$23,$F1,$E7,$F3,$49,$FE,$D5,$3C,$63,$53,$F9,$1B,$F8,$71
	.dc.b	$E4,$BF,$86,$FC,$93,$F4,$D7,$EA,$78,$9F,$04,$F4,$9F,$44,$F1,$9C
	.dc.b	$63,$E1,$3F,$E3,$43,$D2,$7B,$EB,$FC,$C9,$5F,$32,$3F,$37,$5F,$D2
	.dc.b	$9F,$C1,$DF,$F0,$D7,$F9,$6D,$F9,$AF,$F2,$51,$E0,$AE,$0F,$86,$FF
	.dc.b	$D9,$F5,$FB,$C8,$F4,$57,$03,$F3,$D3,$FB,$A8,$FD,$43,$F1,$23,$E7
	.dc.b	$1E,$6B,$FC,$D4,$FD,$9B,$EE,$78,$27,$A4,$FE,$62,$7E,$39,$EC,$3F
	.dc.b	$CC,$67,$D5,$3F,$C3,$75,$F0,$5F,$D2,$CF,$29,$FC,$8D,$79,$A7,$F2
	.dc.b	$51,$E2,$BF,$85,$1F,$12,$FF,$43,$5E,$48,$EF,$3F,$C0,$AF,$D7,$7E
	.dc.b	$59,$FB,$33,$FA,$08,$EE,$7E,$86,$3F,$8C,$3D,$EF,$F5,$B5,$E5,$9F
	.dc.b	$CF,$C7,$C9,$BF,$C3,$47,$48,$FD,$23,$FC,$6E,$BF,$61,$1E,$68,$F2
	.dc.b	$BF,$C4,$23,$C9,$5D,$67,$81,$F9,$59,$E0,$FF,$8E,$AF,$F9,$4C,$7D
	.dc.b	$E3,$E5,$47,$EB,$67,$94,$EA,$3F,$A4,$D7,$E8,$DF,$31,$FA,$88,$F8
	.dc.b	$CF,$E5,$63,$E1,$3F,$5C,$1F,$D8,$AB,$BC,$FD,$28,$FE,$4C,$7F,$1E
	.dc.b	$BF,$A4,$F9,$E7,$F1,$8F,$5A,$FC,$E3,$C0,$FC,$61,$F1,$0F,$E0,$47
	.dc.b	$F9,$C6,$BF,$34,$FF,$32,$3F,$00,$7E,$84,$E0,$7E,$A0,$FB,$31,$F4
	.dc.b	$1F,$DA,$DF,$C4,$AF,$A1,$3F,$80,$3C,$75,$DA,$BF,$26,$7C,$03,$E0
	.dc.b	$C0,$F8,$75,$98,$FF,$20,$14,$FF,$5A,$95,$C4,$71,$84,$6D,$A5,$38
	.dc.b	$9F,$E2,$A3,$C3,$1E,$07,$A1,$B8,$F2,$47,$18,$E6,$7A,$5F,$04,$77
	.dc.b	$34,$70,$8E,$47,$AE,$36,$F6,$38,$3E,$97,$91,$B7,$C8,$F9,$CF,$89
	.dc.b	$1D,$5D,$9E,$38,$DB,$F2,$0D,$BD,$5F,$01,$F0,$0F,$59,$B3,$EA,$47
	.dc.b	$57,$67,$F9,$24,$D1,$F5,$A3,$A3,$E5,$7D,$AF,$9C,$E6,$E9,$F2,$3A
	.dc.b	$79,$9F,$50,$CB,$EA,$61,$C9,$F9,$AF,$E5,$A3,$F9,$54,$7E,$4D,$B1
	.dc.b	$FB,$D3,$FA,$4F,$E0,$DF,$8F,$AF,$F5,$F3,$6F,$F6,$2F,$E3,$61,$18
	.dc.b	$BF,$E1,$35,$DD,$AF,$EB,$67,$F6,$C7,$E7,$5F,$EE,$55,$D6,$7F,$28
	.dc.b	$F7,$83,$DE,$BB,$4F,$E9,$9F,$E0,$27,$F2,$49,$F0,$46,$4F,$E5,$0F
	.dc.b	$B0,$FE,$13,$5E,$C6,$CF,$B1,$FD,$54,$7B,$1F,$0C,$FA,$A3,$D0,$7A
	.dc.b	$A7,$C3,$3F,$BC,$3C,$27,$FB,$BA,$BD,$95,$D6,$15,$D6,$7D,$91,$DA
	.dc.b	$7D,$07,$D5,$3C,$15,$DA,$FE,$B9,$FD,$2A,$3F,$DA,$35,$DA,$BE,$39
	.dc.b	$EC,$29,$F2,$DE,$2F,$F2,$13,$F8,$A9,$F8,$19,$CB,$FE,$0B,$7F,$51
	.dc.b	$7F,$0E,$3D,$4F,$C9,$7D,$77,$ED,$7C,$15,$CA,$7F,$AE,$47,$6A,$F4
	.dc.b	$34,$F1,$1C,$8F,$E3,$D3,$EB,$3D,$0F,$D4,$AF,$C0,$BC,$4F,$35,$FC
	.dc.b	$63,$EF,$D7,$86,$7F,$B0,$DF,$A6,$BE,$1B,$E9,$3E,$93,$D4,$FA,$47
	.dc.b	$57,$D2,$FF,$62,$9F,$4B,$4F,$4C,$75,$BF,$44,$7C,$23,$A3,$E3,$7F
	.dc.b	$AF,$57,$F5,$27,$D3,$7F,$80,$7E,$93,$FE,$05,$9E,$71,$F1,$5F,$C3
	.dc.b	$9C,$EF,$FC,$95,$1F,$E3,$C3,$99,$E9,$AF,$82,$7F,$45,$9E,$77,$F0
	.dc.b	$27,$F0,$97,$F2,$4F,$C8,$1E,$3B,$F7,$CF,$F7,$49,$EA,$7D,$73,$F0
	.dc.b	$77,$FC,$E6,$FE,$BC,$74,$BE,$4F,$C6,$9F,$81,$5F,$30,$E8,$57,$CD
	.dc.b	$1F,$B4,$9F,$E0,$6F,$CB,$3A,$1E,$D3,$E5,$9F,$06,$79,$36,$7E,$0B
	.dc.b	$F8,$C3,$A4,$FA,$1F,$D2,$9F,$98,$7E,$13,$F4,$9F,$4B,$F0,$27,$CF
	.dc.b	$3D,$23,$E0,$CF,$99,$FA,$C7,$E0,$1E,$77,$F2,$A7,$E0,$9C,$67,$FC
	.dc.b	$43,$1F,$69,$E7,$3E,$5B,$FB,$EF,$28,$F9,$47,$B4,$F8,$27,$F9,$78
	.dc.b	$FF,$5E,$C7,$DE,$AF,$DA,$47,$E2,$67,$93,$EB,$8F,$1B,$C8,$F1,$9F
	.dc.b	$E8,$D9,$E4,$FC,$63,$94,$FC,$4B,$F9,$F1,$E1,$3F,$C8,$07,$F2,$23
	.dc.b	$F0,$C7,$EC,$2B,$F0,$CF,$DB,$3E,$84,$F3,$8F,$89,$7F,$E2,$59,$F7
	.dc.b	$3F,$15,$E5,$3F,$3E,$BF,$35,$1C,$8F,$C2,$4E,$CE,$4F,$96,$3F,$D5
	.dc.b	$B7,$EC,$3E,$6D,$78,$67,$E1,$57,$07,$E1,$9F,$EF,$E9,$E0,$FD,$D9
	.dc.b	$E1,$3D,$E3,$E0,$57,$E1,$9F,$99,$5C,$1F,$CC,$9F,$C2,$4F,$A4,$7B
	.dc.b	$8F,$F5,$29,$CA,$FF,$46,$FF,$2E,$BF,$81,$1F,$72,$BF,$A5,$47,$DC
	.dc.b	$7E,$9C,$FA,$0F,$F2,$49,$FD,$C6,$7F,$DE,$B5,$E3,$8F,$90,$78,$E7
	.dc.b	$8D,$FC,$13,$E1,$4E,$E7,$F0,$27,$CE,$3C,$E7,$53,$FB,$11,$63,$FA
	.dc.b	$4D,$7A,$AB,$F9,$83,$E9,$3F,$9D,$CF,$85,$FC,$0B,$F0,$4F,$88,$7D
	.dc.b	$13,$F7,$27,$E2,$CF,$CA,$47,$C2,$BF,$A7,$1E,$48,$FE,$EF,$3F,$E0
	.dc.b	$58,$F2,$4F,$F2,$D7,$9D,$FF,$18,$7E,$1D,$7D,$9B,$FB,$91,$C2,$BA
	.dc.b	$BE,$33,$EA,$3F,$B7,$7F,$08,$7C,$53,$FD,$0C,$71,$30,$7E,$D8,$F0
	.dc.b	$9F,$A3,$3E,$69,$F0,$CF,$C9,$9F,$0C,$FC,$D5,$FD,$A7,$89,$FC,$40
	.dc.b	$FC,$6D,$7A,$5F,$85,$1F,$E0,$99,$FC,$33,$FC,$46,$FE,$E5,$7C,$9A
	.dc.b	$F2,$9F,$BA,$3D,$07,$8E,$FF,$65,$3C,$2B,$EC,$D7,$7A,$F7,$BF,$41
	.dc.b	$FC,$1B,$6B,$F0,$71,$FC,$CC,$F9,$27,$DB,$AF,$D9,$BE,$29,$FF,$17
	.dc.b	$5F,$F7,$C3,$D3,$1F,$AC,$3E,$0C,$7F,$02,$7E,$4B,$E3,$7F,$42,$F8
	.dc.b	$63,$8D,$7A,$A7,$8D,$FE,$4D,$FF,$00,$9E,$F3,$E4,$CF,$84,$FC,$0B
	.dc.b	$F6,$4F,$4D,$FE,$25,$F0,$CF,$F2,$CA,$FC,$4D,$F8,$6F,$8D,$7F,$98
	.dc.b	$0F,$09,$F9,$AA,$F9,$67,$17,$C3,$1F,$DD,$1F,$89,$1F,$81,$7C,$27
	.dc.b	$E2,$67,$ED,$3E,$93,$E9,$C7,$E0,$67,$C3,$3F,$C8,$A7,$C3,$7F,$41
	.dc.b	$E0,$7C,$8A,$E0,$F8,$0F,$F3,$51,$E0,$3F,$20,$7E,$A1,$FD,$F1,$FC
	.dc.b	$F9,$C9,$F1,$0F,$DD,$9D,$6B,$E6,$3F,$E6,$73,$C6,$7E,$A8,$FE,$93
	.dc.b	$3F,$8F,$7E,$AD,$7D,$D7,$F2,$CF,$CD,$79,$1E,$57,$F0,$87,$C4,$3F
	.dc.b	$C6,$CF,$63,$F4,$F5,$F3,$DF,$8F,$5E,$47,$C2,$6C,$F7,$1F,$99,$9F
	.dc.b	$E3,$6F,$D9,$3C,$B3,$CD,$F7,$4F,$D9,$79,$9E,$58,$FC,$85,$7E,$92
	.dc.b	$3F,$C0,$27,$E8,$CF,$85,$5E,$F3,$F4,$75,$CD,$F5,$C7,$33,$C5,$3C
	.dc.b	$DF,$EC,$F3,$FC,$55,$F7,$9F,$48,$F9,$35,$F7,$9F,$2D,$7B,$2B,$F0
	.dc.b	$8F,$63,$11,$E4,$9F,$73,$CA,$BE,$44,$EE,$7F,$04,$FD,$27,$F3,$B3
	.dc.b	$F2,$5C,$9F,$0A,$7F,$6B,$1F,$A0,$3F,$C6,$47,$DC,$9F,$0B,$EA,$3E
	.dc.b	$3D,$7E,$C6,$3E,$F5,$79,$1F,$8F,$3C,$8F,$6C,$78,$1F,$8F,$5F,$D0
	.dc.b	$DF,$95,$5F,$02,$79,$1F,$D2,$6B,$F1,$D3,$CA,$BF,$B9,$BC,$9F,$F5
	.dc.b	$9B,$F8,$73,$F4,$07,$F3,$83,$F9,$1C,$FC,$F9,$F8,$D3,$CA,$BC,$15
	.dc.b	$F1,$63,$0F,$CF,$3E,$C4,$FB,$2F,$FC,$99,$1B,$9F,$BC,$FE,$58,$F8
	.dc.b	$A7,$F0,$5B,$FA,$2F,$8E,$7F,$D1,$4F,$F8,$FC,$F4,$57,$CA,$7A,$1F
	.dc.b	$93,$3E,$81,$FE,$0D,$3F,$01,$5F,$93,$AF,$E0,$2F,$A2,$37,$7F,$B9
	.dc.b	$3B,$BF,$AE,$8E,$0F,$68,$FB,$53,$E6,$8F,$CC,$9E,$13,$D8,$78,$6B
	.dc.b	$E1,$3F,$E6,$43,$DE,$7F,$10,$7D,$4E,$1F,$C9,$3F,$B6,$3E,$5C,$7D
	.dc.b	$D3,$EA,$1E,$AA,$F9,$71,$F6,$8F,$1D,$79,$9E,$47,$CB,$8F,$CC,$BF
	.dc.b	$E0,$5A,$FC,$BB,$E8,$9F,$85,$3A,$8F,$E0,$A7,$E5,$67,$F0,$71,$E0
	.dc.b	$7D,$31,$E0,$3E,$D1,$E0,$3E,$15,$7A,$67,$F6,$2F,$A6,$3E,$41,$E9
	.dc.b	$AF,$0C,$FF,$42,$9E,$0F,$D2,$9F,$8D,$1F,$B3,$8F,$CD,$C7,$E9,$0F
	.dc.b	$C9,$1E,$F7,$93,$FD,$12,$BD,$EF,$73,$F1,$73,$F6,$EF,$FC,$99,$1F
	.dc.b	$E1,$B3,$D8,$7D,$78,$FC,$05,$F9,$A3,$FA,$84,$FE,$21,$FE,$E1,$5D
	.dc.b	$8F,$8B,$3B,$8F,$EB,$27,$E5,$0F,$99,$5F,$0E,$3E,$C1,$FC,$94,$E1
	.dc.b	$3F,$CB,$63,$E1,$CF,$03,$F2,$87,$D0,$9E,$0F,$B0,$FA,$55,$EC,$7F
	.dc.b	$C4,$F5,$F6,$9F,$A2,$F9,$9E,$13,$EB,$7C,$D3,$FB,$A7,$E8,$9F,$C2
	.dc.b	$CF,$E4,$15,$FC,$5E,$7F,$82,$C7,$D1,$3E,$2D,$FF,$19,$BF,$89,$5F
	.dc.b	$2A,$7B,$CF,$E3,$DE,$E7,$E9,$CF,$A3,$3F,$A4,$AF,$6C,$7D,$F3,$F7
	.dc.b	$F7,$ED,$9F,$E4,$27,$B6,$7E,$5D,$78,$6F,$CD,$7F,$86,$3F,$2A,$7F
	.dc.b	$20,$8F,$CA,$BF,$30,$FC,$01,$F9,$B9,$FA,$73,$FC,$C2,$38,$4F,$E5
	.dc.b	$5F,$3C,$7E,$55,$F8,$B1,$FC,$3D,$FE,$0B,$7E,$79,$FD,$DC,$77,$BF
	.dc.b	$75,$79,$CE,$11,$FB,$DA,$E1,$1E,$68,$FA,$77,$F4,$1F,$09,$F5,$4F
	.dc.b	$C7,$D7,$F1,$27,$FD,$76,$70,$3E,$14,$7B,$1F,$E1,$E7,$B0,$FC,$71
	.dc.b	$B3,$F9,$39,$F6,$27,$F5,$B5,$F8,$7A,$FF,$1C,$4F,$D5,$3F,$86,$1B
	.dc.b	$3F,$07,$3C,$0F,$71,$FC,$2E,$FF,$09,$02,$40,$C5,$E6,$3F,$DA,$25
	.dc.b	$3F,$D6,$E5,$0D,$8E,$51,$83,$51,$B7,$A3,$DE,$31,$1E,$03,$B3,$CA
	.dc.b	$38,$1A,$8F,$F3,$A9,$D0,$DC,$71,$7C,$E7,$9E,$3B,$3D,$9F,$34,$6E
	.dc.b	$3F,$D2,$47,$99,$F4,$BE,$03,$A9,$EB,$3A,$9E,$A7,$6F,$8C,$F9,$0F
	.dc.b	$A5,$C1,$D5,$F5,$47,$8A,$3A,$1E,$23,$D1,$1D,$1F,$5B,$CD,$F0,$9F
	.dc.b	$48,$CB,$EB,$61,$E5,$7B,$BE,$A8,$F6,$3F,$32,$3B,$3E,$17,$B3,$E3
	.dc.b	$39,$1F,$24,$FC,$71,$FE,$21,$7F,$18,$FE,$E4,$CD,$2B,$F1,$07,$E2
	.dc.b	$DC,$4F,$E1,$A3,$07,$E4,$CF,$C6,$47,$19,$DC,$FF,$52,$8F,$E3,$B1
	.dc.b	$DC,$FD,$11,$FD,$3C,$AF,$F6,$89,$EB,$3D,$E7,$F4,$13,$D6,$7B,$9E
	.dc.b	$19,$FD,$3C,$FF,$38,$8F,$D3,$CF,$82,$BF,$92,$B6,$BF,$0E,$F9,$EB
	.dc.b	$D4,$76,$AF,$C8,$9E,$E7,$46,$4F,$69,$FC,$6A,$7E,$5D,$76,$9F,$E5
	.dc.b	$73,$FC,$10,$F6,$BF,$E2,$53,$DB,$3E,$08,$57,$B5,$FD,$C3,$EF,$8F
	.dc.b	$6D,$7F,$38,$7E,$65,$76,$AF,$E1,$67,$63,$FC,$1C,$F6,$7D,$D5,$D5
	.dc.b	$F6,$47,$9E,$7F,$95,$CF,$F5,$F9,$E5,$59,$9F,$69,$F2,$2B,$F0,$B3
	.dc.b	$F1,$E3,$EB,$3E,$A8,$F6,$3F,$1D,$E4,$FD,$53,$F0,$55,$F4,$E7,$D6
	.dc.b	$78,$A7,$E3,$4F,$58,$F8,$D3,$E9,$8E,$91,$F1,$1F,$EF,$CD,$7F,$AA
	.dc.b	$D7,$80,$F9,$0F,$D9,$39,$D7,$DE,$7C,$C7,$D0,$7F,$19,$3E,$43,$E0
	.dc.b	$96,$3D,$53,$D5,$F0,$C7,$E1,$DE,$AF,$C4,$AF,$44,$7C,$53,$F1,$27
	.dc.b	$EC,$CE,$71,$E9,$8F,$EB,$EF,$79,$F4,$CF,$13,$AC,$7A,$9F,$CA,$9D
	.dc.b	$6B,$C8,$FC,$28,$E9,$1F,$DC,$63,$C1,$3D,$2B,$F3,$B3,$F8,$73,$F0
	.dc.b	$B3,$F0,$CF,$3C,$7F,$82,$5F,$79,$FA,$13,$D3,$5E,$77,$27,$9C,$FC
	.dc.b	$8C,$FC,$F7,$CC,$75,$7C,$F3,$F0,$23,$A5,$79,$67,$FC,$85,$1D,$27
	.dc.b	$CE,$FE,$14,$EA,$7C,$27,$F5,$15,$E3,$8F,$84,$D8,$ED,$1F,$08,$E9
	.dc.b	$3F,$9C,$7F,$35,$3F,$66,$3F,$37,$1F,$36,$3E,$5B,$FE,$0B,$9F,$88
	.dc.b	$F6,$3D,$27,$43,$FB,$A4,$FC,$D7,$F1,$E7,$EC,$A7,$9D,$7F,$85,$CF
	.dc.b	$F0,$84,$7F,$83,$4F,$F5,$4C,$F1,$7E,$49,$C6,$7D,$91,$F7,$6B,$C6
	.dc.b	$7B,$8A,$E4,$F7,$3F,$C6,$9F,$BF,$3C,$CF,$49,$EB,$3E,$31,$63,$9B
	.dc.b	$F0,$63,$99,$F7,$63,$F3,$E7,$D2,$8E,$6F,$8E,$3E,$DC,$73,$6C,$66
	.dc.b	$3E,$39,$F7,$63,$CA,$FF,$00,$9F,$7B,$E1,$3F,$C8,$D1,$ED,$69,$93
	.dc.b	$F4,$E7,$E1,$9F,$E9,$2F,$E7,$1F,$09,$C5,$FA,$07,$F5,$59,$FE,$2C
	.dc.b	$FF,$09,$3F,$37,$3F,$85,$7F,$21,$3F,$C2,$1F,$7B,$CD,$FF,$28,$D7
	.dc.b	$DD,$7F,$B7,$C7,$C8,$9E,$4F,$90,$FD,$19,$C6,$7F,$07,$3C,$8F,$F3
	.dc.b	$B4,$7E,$0E,$BE,$7B,$E1,$8F,$AF,$1F,$2D,$F5,$BF,$CF,$4F,$5B,$FB
	.dc.b	$58,$E4,$7E,$7A,$3E,$C4,$72,$8F,$13,$E8,$3E,$6C,$FF,$9B,$63,$8C
	.dc.b	$7C,$23,$E6,$1A,$3F,$CD,$67,$13,$C6,$FF,$3E,$7F,$5F,$5F,$C6,$E3
	.dc.b	$E1,$3F,$A4,$8E,$27,$F5,$47,$0F,$F7,$59,$FF,$23,$BE,$B3,$F1,$07
	.dc.b	$EB,$8F,$C4,$1F,$50,$FC,$21,$F2,$CF,$EE,$D5,$F1,$0F,$E1,$AF,$90
	.dc.b	$FC,$41,$FC,$5E,$7C,$2F,$F7,$08,$E1,$5F,$8A,$8E,$AF,$C4,$3D,$73
	.dc.b	$F9,$8A,$FC,$14,$FA,$E3,$0F,$DC,$3D,$27,$EF,$8F,$49,$F4,$CF,$4C
	.dc.b	$7E,$10,$FC,$2C,$F7,$8F,$34,$F4,$7D,$B1,$FE,$07,$7F,$38,$DA,$FF
	.dc.b	$18,$47,$81,$FC,$5B,$F6,$8F,$AE,$7F,$78,$8E,$07,$F8,$B8,$AF,$F8
	.dc.b	$40,$E4,$FE,$2C,$FC,$A4,$7F,$08,$AF,$82,$FC,$89,$FC,$31,$F6,$EB
	.dc.b	$FC,$1C,$FF,$02,$3E,$95,$7F,$02,$7F,$99,$47,$B5,$FA,$51,$E2,$AF
	.dc.b	$D4,$C7,$D2,$3C,$67,$E5,$63,$31,$E2,$78,$4F,$DC,$9F,$EE,$D3,$F8
	.dc.b	$CA,$FC,$19,$FC,$9E,$3F,$18,$78,$8F,$C6,$47,$39,$FF,$0D,$47,$C7
	.dc.b	$6C,$73,$AF,$E0,$11,$F9,$F9,$FA,$E7,$23,$F3,$F5,$F1,$5F,$A3,$5C
	.dc.b	$5F,$89,$1E,$37,$F0,$C7,$8C,$FC,$34,$78,$E7,$FC,$33,$1D,$4F,$9C
	.dc.b	$FF,$85,$DF,$E3,$6F,$8D,$FE,$84,$7E,$AC,$FA,$27,$C0,$9F,$E0,$A7
	.dc.b	$DB,$7C,$F5,$F2,$A7,$A3,$F9,$38,$E7,$3F,$2A,$3F,$A3,$1F,$85,$8F
	.dc.b	$C7,$1F,$B3,$3E,$DC,$FC,$83,$F0,$6F,$F3,$A8,$FC,$14,$F8,$9F,$97
	.dc.b	$1F,$B6,$3B,$BF,$B3,$AF,$85,$1F,$E1,$63,$F9,$D4,$79,$5F,$D6,$CF
	.dc.b	$96,$BF,$99,$C7,$D4,$3E,$59,$EE,$7C,$87,$B9,$FC,$6C,$FE,$F9,$EA
	.dc.b	$60,$E5,$3F,$1D,$FD,$9C,$F2,$76,$7C,$E3,$ED,$C7,$B5,$FB,$F3,$F2
	.dc.b	$1F,$CD,$4F,$E5,$E7,$F9,$E4,$FD,$19,$E7,$1F,$85,$3F,$77,$1F,$89
	.dc.b	$8F,$94,$7D,$B7,$F6,$A7,$C7,$7D,$D3,$F8,$87,$F0,$53,$CE,$7E,$C1
	.dc.b	$CE,$3F,$12,$7D,$13,$E5,$BC,$DF,$7C,$F3,$9F,$9C,$FE,$3E,$36,$7B
	.dc.b	$9F,$A2,$FD,$A3,$F8,$29,$F8,$47,$94,$FE,$00,$FC,$19,$EA,$8E,$51
	.dc.b	$D5,$FD,$24,$FD,$17,$91,$F8,$33,$DB,$3F,$34,$FB,$4E,$27,$C6,$F9
	.dc.b	$8F,$D9,$BF,$46,$3F,$AD,$4F,$C5,$9F,$AB,$3F,$85,$7E,$E4,$7D,$28
	.dc.b	$E1,$1F,$C1,$5F,$5B,$E8,$3E,$51,$E8,$9F,$73,$F5,$CF,$D3,$D7,$F4
	.dc.b	$17,$FB,$61,$F4,$A3,$E1,$BF,$AC,$7F,$3C,$F3,$3E,$59,$FD,$0A,$7F
	.dc.b	$24,$FB,$DF,$91,$5F,$91,$8F,$C3,$9F,$94,$9F,$9C,$FC,$29,$FE,$64
	.dc.b	$FF,$28,$7F,$6E,$F8,$CF,$44,$7E,$1C,$F1,$CF,$58,$FE,$00,$78,$E7
	.dc.b	$E1,$BE,$08,$F9,$11,$F2,$DF,$9D,$3F,$59,$FB,$93,$C4,$FB,$8F,$E3
	.dc.b	$9F,$96,$F7,$8F,$09,$EA,$8F,$0C,$7E,$AA,$7E,$1B,$C6,$7E,$59,$E9
	.dc.b	$9F,$DE,$1F,$95,$3E,$11,$FC,$B4,$FF,$3B,$9A,$3F,$10,$7C,$33,$E5
	.dc.b	$1F,$68,$FB,$35,$FD,$E9,$E0,$7D,$73,$E1,$47,$EC,$2B,$F1,$33,$FD
	.dc.b	$6A,$7F,$C1,$07,$E3,$A7,$53,$F1,$DF,$CE,$47,$F1,$27,$F6,$8F,$E8
	.dc.b	$A7,$FD,$5C,$FE,$46,$3D,$F3,$F6,$9F,$8F,$1E,$73,$FC,$30,$7B,$CF
	.dc.b	$C0,$47,$73,$F4,$EF,$A4,$F8,$15,$F6,$9E,$0F,$CF,$8F,$BF,$1A,$9F
	.dc.b	$4B,$FC,$69,$F4,$BF,$3C,$FA,$EF,$F8,$46,$3E,$BB,$FA,$68,$F9,$AF
	.dc.b	$D7,$7E,$21,$FB,$43,$C2,$F1,$8F,$0B,$EE,$38,$4F,$E3,$A7,$E4,$1F
	.dc.b	$A8,$7D,$0F,$E3,$A3,$C1,$1E,$D8,$FD,$51,$F9,$08,$FC,$75,$7E,$DE
	.dc.b	$3E,$49,$A8,$F9,$F5,$F2,$4F,$D8,$BC,$23,$F0,$A7,$EF,$9F,$34,$7D
	.dc.b	$58,$FC,$3D,$7E,$AE,$BB,$3F,$E0,$98,$F9,$87,$C1,$3C,$07,$ED,$27
	.dc.b	$F1,$64,$3C,$4F,$A4,$FE,$14,$76,$AF,$23,$EA,$9F,$D8,$9F,$E4,$A7
	.dc.b	$E7,$C7,$ED,$4F,$8A,$7D,$C7,$FB,$59,$FD,$10,$F9,$47,$E4,$4F,$24
	.dc.b	$F8,$A3,$E2,$C7,$D8,$3F,$94,$1C,$0F,$D2,$9E,$D9,$FD,$C3,$C8,$E1
	.dc.b	$5F,$A7,$7F,$C4,$27,$E3,$6B,$F2,$F3,$FC,$16,$3D,$C7,$E7,$DF,$95
	.dc.b	$3F,$B5,$8F,$69,$F8,$43,$C3,$5E,$F7,$F0,$CF,$E3,$5F,$31,$F4,$9D
	.dc.b	$3C,$0E,$47,$E7,$0F,$6C,$7E,$D1,$D1,$FA,$18,$FC,$0D,$7F,$30,$AF
	.dc.b	$C9,$D6,$EB,$B1,$D2,$B6,$FC,$2A,$FB,$CF,$C1,$7F,$1E,$FD,$E3,$DA
	.dc.b	$7E,$BE,$7F,$17,$3F,$E0,$D8,$FC,$49,$FA,$E9,$FC,$93,$C0,$FE,$3F
	.dc.b	$3C,$23,$F6,$B1,$F7,$4E,$0F,$F6,$54,$18,$A7,$2C,$30,$53,$2F,$FB
	.dc.b	$09,$C3,$98,$E6,$EE,$38,$95,$E0,$F5,$78,$47,$91,$D4,$71,$3C,$91
	.dc.b	$E6,$72,$79,$9E,$EF,$88,$E1,$1D,$9F,$F4,$E1,$F0,$23,$6F,$84,$DB
	.dc.b	$D5,$E4,$75,$7C,$8E,$0F,$7B,$B7,$91,$B3,$DE,$78,$0F,$73,$B7,$0F
	.dc.b	$89,$D1,$F3,$1D,$3E,$03,$A4,$7C,$07,$C3,$1F,$15,$EE,$7D,$B7,$83
	.dc.b	$D0,$F9,$8F,$47,$6F,$85,$D9,$E6,$7A,$47,$D4,$8E,$A4,$39,$C6,$9F
	.dc.b	$DC,$C7,$F3,$A3,$FA,$4B,$FD,$C9,$FD,$1A,$AF,$8B,$F8,$37,$E3,$9F
	.dc.b	$BB,$5F,$94,$8C,$55,$AF,$C9,$57,$E4,$A7,$B5,$7F,$0A,$AF,$F0,$61
	.dc.b	$DA,$7F,$89,$D7,$E8,$0F,$CF,$BE,$06,$C7,$C7,$55,$E0,$69,$E0,$8F
	.dc.b	$25,$7E,$C2,$BF,$AB,$4F,$F0,$53,$E0,$C6,$67,$E0,$C7,$F7,$3A,$F5
	.dc.b	$CF,$EB,$9E,$C7,$EA,$EB,$D9,$1C,$4F,$64,$FF,$05,$8E,$EF,$CB,$3B
	.dc.b	$C7,$FB,$2E,$7A,$BF,$1E,$3E,$75,$57,$DE,$7B,$A3,$B1,$FC,$4A,$3E
	.dc.b	$5B,$4C,$C7,$F5,$38,$FF,$48,$4F,$68,$F8,$A7,$62,$9D,$9E,$6F,$19
	.dc.b	$E7,$3F,$D3,$23,$F8,$DD,$7F,$79,$3E,$A3,$E8,$9F,$8C,$F3,$3E,$91
	.dc.b	$E9,$7F,$AB,$3F,$D3,$CE,$73,$FD,$7E,$7F,$02,$75,$7F,$8D,$57,$A4
	.dc.b	$FA,$33,$D1,$F3,$C7,$96,$78,$BC,$CF,$ED,$73,$FD,$7D,$E6,$FE,$2A
	.dc.b	$BC,$EF,$C3,$AF,$B9,$3F,$88,$9F,$3C,$FE,$4C,$F3,$9C,$CF,$E6,$D5
	.dc.b	$E7,$7F,$98,$D7,$9D,$F8,$47,$E2,$27,$E5,$BF,$06,$7C,$EF,$33,$FA
	.dc.b	$C9,$FE,$62,$9E,$4F,$94,$F8,$13,$C9,$F2,$4F,$2A,$FE,$CF,$1F,$D7
	.dc.b	$0F,$F2,$15,$72,$9F,$D9,$47,$2A,$F6,$C7,$A6,$3C,$27,$F9,$4E,$BF
	.dc.b	$68,$7C,$87,$EF,$4F,$E5,$E7,$9C,$7B,$63,$9C,$FD,$18,$F8,$AD,$7A
	.dc.b	$CF,$FA,$6E,$BC,$93,$F9,$F9,$FA,$55,$ED,$AF,$24,$7A,$CF,$D5,$1F
	.dc.b	$DA,$4F,$DB,$1F,$6E,$3C,$93,$CE,$7F,$6B,$1E,$49,$F9,$C7,$33,$DC
	.dc.b	$7E,$D0,$FD,$79,$C9,$FC,$2C,$7B,$A3,$F4,$D1,$E5,$9F,$64,$73,$9F
	.dc.b	$F1,$23,$E5,$7D,$CF,$94,$FD,$8B,$CD,$F7,$CF,$96,$7E,$C9,$F4,$23
	.dc.b	$E4,$9E,$8A,$F8,$EF,$EE,$0E,$27,$F8,$54,$FD,$C4,$F3,$7F,$C3,$87
	.dc.b	$B4,$FE,$00,$7B,$5F,$8E,$F3,$AF,$8E,$F3,$AF,$D6,$1F,$AA,$3C,$AF
	.dc.b	$38,$FE,$02,$7F,$42,$7D,$B1,$F4,$4F,$D5,$3C,$CF,$D0,$3C,$DF,$29
	.dc.b	$F1,$DF,$F6,$75,$7B,$8F,$D9,$C7,$F0,$93,$91,$FA,$08,$E4,$FF,$3B
	.dc.b	$39,$1E,$33,$DE,$79,$2B,$ED,$C7,$96,$3A,$1F,$BC,$7D,$AF,$27,$F8
	.dc.b	$6B,$E8,$3F,$7A,$7D,$B9,$FE,$26,$FE,$84,$F2,$D7,$E6,$27,$ED,$47
	.dc.b	$E8,$1F,$6B,$C4,$FE,$0A,$7E,$32,$BE,$C3,$EB,$9F,$E9,$2F,$E5,$8F
	.dc.b	$24,$7F,$8A,$E3,$E8,$CF,$F6,$3A,$F1,$D7,$33,$C7,$1F,$96,$3C,$6F
	.dc.b	$E5,$DF,$1B,$FB,$C3,$EC,$9F,$83,$7E,$CB,$F1,$4F,$AF,$3E,$43,$F3
	.dc.b	$E5,$8E,$6F,$E0,$E7,$C7,$3F,$96,$39,$1E,$D9,$FE,$DF,$3C,$9A,$72
	.dc.b	$3D,$91,$F1,$8E,$51,$FC,$7D,$F1,$C7,$07,$FB,$83,$4F,$EC,$2F,$CF
	.dc.b	$3F,$95,$1F,$5C,$F1,$9F,$C3,$23,$D4,$F2,$AF,$EF,$73,$CA,$3E,$35
	.dc.b	$7B,$2B,$F4,$F1,$C8,$FA,$CF,$29,$FC,$4C,$7E,$5C,$FC,$49,$C8,$F6
	.dc.b	$47,$23,$D8,$FC,$5A,$F0,$9F,$C3,$DF,$E0,$47,$AE,$3F,$85,$1C,$67
	.dc.b	$F4,$71,$EB,$7E,$14,$F1,$7F,$A2,$D7,$EB,$67,$FA,$11,$F8,$63,$F9
	.dc.b	$2C,$7E,$18,$FC,$3C,$6E,$3E,$45,$79,$9E,$D5,$E6,$3F,$32,$FF,$59
	.dc.b	$3C,$D5,$FA,$FA,$F8,$75,$FA,$59,$FE,$15,$3E,$87,$F9,$D9,$F1,$4F
	.dc.b	$AE,$7E,$6C,$F9,$D1,$F8,$7A,$F0,$47,$A6,$7D,$AF,$80,$FC,$D1,$FB
	.dc.b	$E9,$F4,$3F,$5A,$3E,$2C,$FA,$23,$C3,$3F,$99,$7E,$AC,$FE,$8A,$38
	.dc.b	$9E,$19,$FF,$18,$1C,$5F,$E7,$F3,$E9,$7E,$AC,$F1,$3D,$CF,$18,$FE
	.dc.b	$6E,$7A,$4F,$F6,$ED,$7C,$F7,$E0,$47,$82,$7F,$CC,$F1,$F2,$0F,$E8
	.dc.b	$AF,$C0,$7C,$0F,$CF,$AF,$C0,$47,$E9,$9F,$8E,$FC,$A8,$FD,$74,$7F
	.dc.b	$24,$8F,$8F,$5F,$5A,$7F,$84,$47,$73,$F8,$D1,$FC,$B6,$BF,$88,$D7
	.dc.b	$1A,$FE,$6E,$7E,$9E,$3E,$54,$FD,$A3,$F7,$35,$F0,$8F,$5B,$FC,$72
	.dc.b	$BF,$CC,$8F,$D9,$AE,$71,$FC,$7E,$7F,$C2,$55,$D0,$F9,$B5,$F7,$63
	.dc.b	$A3,$FB,$99,$FA,$AF,$F7,$C3,$FA,$13,$FD,$82,$BA,$BE,$08,$F9,$27
	.dc.b	$C2,$7E,$4B,$FB,$37,$9B,$FD,$32,$3E,$B3,$EC,$3F,$7B,$5E,$5A,$FA
	.dc.b	$55,$F8,$57,$DF,$1F,$C1,$63,$D9,$3C,$27,$F8,$FB,$F1,$E7,$F7,$8E
	.dc.b	$67,$E7,$3F,$B8,$3E,$D1,$EF,$3F,$16,$79,$4F,$94,$6E,$3E,$BC,$7C
	.dc.b	$18,$C4,$FC,$83,$F0,$0F,$F0,$E3,$E1,$D7,$C6,$7E,$53,$F0,$2B,$CB
	.dc.b	$1F,$84,$9E,$27,$E4,$E3,$E1,$47,$CB,$AF,$9D,$18,$AE,$93,$EA,$AE
	.dc.b	$2F,$E8,$8F,$49,$FC,$41,$FC,$09,$F1,$A7,$F0,$33,$C2,$7E,$B1,$F0
	.dc.b	$A3,$DE,$FF,$5C,$8E,$F1,$87,$F9,$D4,$FE,$D6,$7F,$56,$7E,$9E,$BE
	.dc.b	$FD,$6E,$3F,$81,$9F,$8F,$3E,$A4,$7C,$B7,$CC,$72,$3C,$73,$FE,$2A
	.dc.b	$76,$78,$CF,$A2,$62,$BC,$6F,$53,$E6,$4F,$E4,$9F,$E5,$87,$D4,$38
	.dc.b	$9F,$E3,$B3,$A1,$E2,$38,$BF,$91,$7F,$71,$3F,$2A,$3F,$17,$1B,$3F
	.dc.b	$4A,$72,$3F,$88,$C7,$E1,$63,$C5,$1F,$C9,$A3,$E9,$1F,$9E,$7F,$0A
	.dc.b	$7D,$83,$89,$F0,$CF,$E8,$F1,$F3,$23,$C4,$FC,$98,$D9,$EB,$9F,$86
	.dc.b	$7D,$D7,$89,$F7,$67,$8B,$E7,$3F,$64,$71,$3E,$BC,$71,$7F,$A0,$4F
	.dc.b	$F2,$2A,$FB,$91,$FA,$FA,$FC,$13,$C4,$C1,$F4,$CF,$F1,$41,$FA,$F7
	.dc.b	$EE,$3B,$8F,$C9,$CF,$EF,$23,$13,$F1,$5F,$CF,$47,$CA,$9F,$E8,$E7
	.dc.b	$E8,$E3,$ED,$9F,$4A,$BE,$0C,$EE,$BF,$84,$9F,$D6,$27,$E7,$9F,$83
	.dc.b	$3E,$D4,$FE,$D6,$3D,$71,$E8,$7E,$15,$78,$1F,$B3,$1F,$B5,$39,$C7
	.dc.b	$8D,$FC,$24,$7F,$80,$DF,$E8,$47,$94,$F9,$B5,$EF,$70,$7C,$8A,$EF
	.dc.b	$3E,$89,$FA,$0F,$D8,$8F,$6C,$7F,$04,$8F,$1C,$7E,$D9,$FC,$84,$7E
	.dc.b	$65,$FE,$98,$6A,$7E,$3C,$FB,$DF,$AD,$1E,$77,$FA,$61,$EC,$78,$3F
	.dc.b	$18,$FC,$B1,$F8,$83,$E3,$4F,$CF,$7F,$4E,$7E,$D4,$F7,$BE,$73,$F5
	.dc.b	$2F,$E6,$0E,$2F,$9C,$F8,$93,$E5,$3D,$2F,$08,$F0,$4F,$9E,$BF,$A2
	.dc.b	$4F,$53,$F4,$93,$E5,$AE,$51,$E5,$AF,$9B,$3E,$63,$F1,$E7,$DB,$3F
	.dc.b	$A1,$1F,$85,$3C,$CF,$73,$F1,$4E,$AB,$F0,$0F,$AE,$BB,$D7,$AC,$91
	.dc.b	$F2,$DF,$54,$FF,$42,$8F,$C0,$BF,$A2,$8D,$C7,$E8,$1F,$63,$FB,$78
	.dc.b	$E1,$3D,$A3,$FC,$44,$7F,$28,$3F,$D6,$07,$B1,$FC,$23,$F4,$CF,$C7
	.dc.b	$4F,$C3,$21,$DA,$3F,$1B,$1E,$78,$F9,$73,$E7,$AF,$85,$1F,$94,$9E
	.dc.b	$93,$E9,$78,$1E,$79,$FD,$59,$C2,$7F,$8D,$47,$08,$FD,$F4,$FF,$0C
	.dc.b	$39,$47,$C1,$3D,$A7,$F3,$A7,$E7,$BF,$74,$FC,$9B,$E5,$7F,$5F,$1F
	.dc.b	$E7,$27,$2F,$EF,$2B,$E4,$47,$52,$4F,$86,$7C,$62,$3F,$D2,$25,$A1
	.dc.b	$9A,$D4,$40,$D7,$0F,$08,$C4,$74,$3F,$C3,$25,$39,$C6,$DD,$9D,$0E
	.dc.b	$D1,$E2,$70,$76,$37,$1C,$5E,$EF,$8E,$38,$38,$8D,$C7,$37,$07,$B1
	.dc.b	$DB,$EB,$8F,$1B,$EE,$8E,$0E,$4F,$2B,$F1,$5C,$1E,$27,$6F,$9C,$EC
	.dc.b	$ED,$C1,$DD,$DB,$F1,$DF,$1B,$A3,$E6,$BA,$79,$1F,$5D,$AC,$3E,$6B
	.dc.b	$B7,$E4,$9B,$29,$B3,$FD,$F6,$7A,$8D,$47,$C6,$7C,$D1,$F1,$23,$E7
	.dc.b	$3D,$DC,$3D,$DF,$C8,$D7,$E3,$E7,$F8,$83,$63,$F6,$F1,$F9,$1F,$C1
	.dc.b	$BF,$1B,$2B,$FC,$24,$61,$D4,$FE,$46,$74,$A7,$F9,$4D,$FE,$EA,$BF
	.dc.b	$45,$3F,$A4,$3F,$BB,$5F,$92,$3F,$35,$5D,$6B,$F1,$CF,$7A,$E8,$77
	.dc.b	$9F,$34,$FF,$38,$8E,$95,$E0,$7F,$81,$1D,$8C,$C7,$F6,$1A,$EC,$FA
	.dc.b	$A1,$7E,$A9,$FD,$53,$67,$B4,$FF,$35,$3F,$81,$16,$3F,$92,$BD,$A3
	.dc.b	$D0,$F5,$8F,$95,$1E,$C8,$FD,$F4,$77,$7D,$87,$C9,$8E,$F1,$FC,$40
	.dc.b	$EB,$1D,$EF,$E5,$B4,$FF,$10,$57,$B1,$F8,$E7,$B2,$3D,$CF,$81,$F9
	.dc.b	$35,$FC,$C6,$3F,$B3,$9F,$CA,$AF,$F9,$9B,$FD,$3E,$7F,$B9,$BF,$DA
	.dc.b	$6B,$B1,$F5,$4F,$8F,$1D,$23,$91,$D0,$E5,$5F,$D4,$A3,$D3,$7E,$D7
	.dc.b	$FC,$43,$5D,$2B,$CC,$7A,$8A,$FD,$DB,$F2,$D7,$E6,$67,$FC,$1A,$7D
	.dc.b	$13,$FD,$95,$3C,$DF,$84,$73,$2B,$CD,$F3,$47,$37,$FC,$20,$7C,$D9
	.dc.b	$FE,$39,$1F,$9E,$8F,$B9,$5F,$A4,$7C,$A7,$DA,$AE,$8F,$F6,$0B,$F9
	.dc.b	$CD,$3E,$29,$4F,$29,$D1,$FC,$33,$E6,$3A,$3F,$84,$9F,$A7,$1F,$92
	.dc.b	$AF,$94,$FE,$7C,$F4,$4F,$E7,$DF,$C0,$3D,$23,$F2,$11,$F6,$63,$A1
	.dc.b	$F5,$5E,$51,$D2,$FD,$6F,$F6,$63,$E1,$47,$E3,$9F,$CE,$CF,$C4,$8F
	.dc.b	$9D,$1F,$A9,$8F,$82,$F1,$AF,$F2,$64,$FD,$06,$BF,$0E,$3A,$DF,$D9
	.dc.b	$7C,$F1,$F8,$2A,$F3,$9F,$C3,$5F,$3D,$7C,$49,$FD,$39,$E9,$9F,$85
	.dc.b	$3F,$A3,$7E,$14,$FE,$92,$3F,$07,$1D,$1F,$E1,$37,$F0,$27,$FB,$AC
	.dc.b	$7F,$82,$4F,$E5,$2F,$F1,$87,$E6,$CF,$C1,$BF,$D1,$C7,$EE,$E3,$E0
	.dc.b	$CF,$28,$FF,$13,$1F,$98,$3E,$BC,$7F,$08,$7E,$64,$7E,$35,$FC,$11
	.dc.b	$F5,$E3,$9B,$E2,$3C,$95,$FC,$E9,$FF,$1E,$CF,$77,$9C,$FC,$07,$EA
	.dc.b	$C7,$ED,$DF,$DA,$C7,$CB,$9F,$81,$1D,$4F,$39,$FE,$D1,$AF,$A9,$5F
	.dc.b	$26,$FE,$F4,$FE,$71,$FB,$CF,$90,$F7,$3F,$4E,$FD,$D5,$CA,$7C,$67
	.dc.b	$C7,$39,$57,$A6,$FF,$A3,$47,$33,$C4,$7D,$03,$E3,$9C,$A7,$D3,$1F
	.dc.b	$0C,$E3,$1E,$A3,$F8,$D1,$C6,$3F,$8D,$3F,$AE,$AF,$9B,$3C,$63,$F8
	.dc.b	$61,$F1,$8F,$C2,$C7,$E3,$2F,$F4,$31,$E2,$AF,$C0,$9E,$27,$E0,$DF
	.dc.b	$E5,$4F,$E4,$25,$7F,$9C,$9E,$33,$F8,$59,$F7,$CF,$86,$7D,$33,$EE
	.dc.b	$96,$FD,$AF,$DE,$BF,$C9,$3E,$B3,$8C,$7A,$DF,$C7,$98,$3F,$DB,$75
	.dc.b	$F4,$CF,$4C,$FE,$8E,$7F,$97,$D7,$ED,$AF,$EE,$DF,$CB,$3F,$D6,$97
	.dc.b	$C5,$F0,$1F,$1E,$B8,$BF,$64,$FF,$4C,$1F,$32,$7F,$7E,$FC,$13,$F3
	.dc.b	$E7,$EF,$EF,$E9,$BF,$29,$FE,$2D,$3F,$36,$3F,$06,$7E,$F6,$3E,$0B
	.dc.b	$F2,$4F,$11,$FC,$3C,$F1,$1F,$CA,$6F,$E9,$4F,$F3,$E7,$F7,$91,$CC
	.dc.b	$FB,$87,$B2,$BF,$A3,$3E,$A3,$E5,$9D,$8F,$E3,$51,$CA,$3F,$0B,$7F
	.dc.b	$8E,$9F,$24,$72,$7C,$57,$EC,$3E,$83,$E7,$3F,$A2,$3E,$C7,$E2,$1F
	.dc.b	$64,$FE,$AF,$1F,$AD,$70,$7E,$78,$F5,$4F,$85,$E2,$79,$E3,$ED,$BF
	.dc.b	$8C,$BF,$C3,$47,$38,$FE,$AA,$7E,$34,$F8,$67,$AA,$FF,$08,$7E,$DD
	.dc.b	$FB,$8E,$1F,$CC,$57,$E1,$5E,$07,$D2,$9F,$E6,$F3,$F4,$AF,$E0,$CF
	.dc.b	$CA,$3F,$D3,$F5,$C5,$F4,$D7,$C9,$38,$CF,$C9,$3F,$0F,$1F,$84,$8F
	.dc.b	$A2,$FE,$B0,$E3,$5E,$5B,$F7,$BE,$1A,$E3,$3F,$26,$B8,$BF,$94,$3F
	.dc.b	$A3,$C6,$9F,$B2,$FF,$29,$3D,$31,$F3,$2F,$D1,$3C,$0F,$AA,$70,$3F
	.dc.b	$85,$CF,$39,$FF,$7B,$4F,$07,$F9,$6C,$FD,$28,$C5,$7F,$39,$9F,$84
	.dc.b	$FF,$0C,$AF,$85,$5F,$E2,$23,$F5,$93,$FD,$A9,$F1,$1C,$CC,$CF,$F3
	.dc.b	$B3,$A1,$F1,$5F,$D8,$4F,$B4,$FE,$72,$F0,$3F,$98,$3F,$C8,$E3,$FA
	.dc.b	$C4,$F7,$3D,$27,$9A,$7F,$BB,$CF,$99,$F1,$1F,$96,$3E,$B1,$E6,$8F
	.dc.b	$E0,$AE,$4F,$A4,$F4,$9F,$34,$7E,$89,$FB,$4F,$C1,$38,$47,$8E,$3D
	.dc.b	$B1,$F8,$C3,$D5,$5F,$C8,$2B,$E9,$4F,$CB,$7E,$04,$F3,$AF,$C1,$9C
	.dc.b	$DF,$EB,$07,$A9,$CC,$FC,$B8,$CC,$7C,$1B,$FE,$31,$1F,$8D,$9F,$8E
	.dc.b	$6E,$FF,$89,$47,$94,$F4,$C6,$6F,$F1,$AF,$EB,$E3,$F8,$8D,$FC,$63
	.dc.b	$C2,$FD,$49,$FA,$EF,$23,$07,$E4,$A3,$F5,$31,$EE,$8F,$54,$F9,$1D
	.dc.b	$98,$8F,$F0,$55,$7C,$78,$E5,$3F,$1D,$FC,$D9,$F1,$DE,$51,$B8,$E5
	.dc.b	$3F,$C6,$DF,$BD,$3F,$69,$FE,$59,$5F,$C4,$4F,$64,$72,$8F,$E6,$4F
	.dc.b	$F2,$D3,$0F,$F0,$C3,$FA,$A4,$7B,$1F,$19,$E0,$3F,$8C,$9F,$C0,$8F
	.dc.b	$9C,$FC,$8A,$F3,$3F,$5E,$BE,$7C,$FF,$1A,$3F,$0E,$78,$8E,$35,$E2
	.dc.b	$7A,$1E,$27,$E2,$5F,$F6,$28,$FE,$76,$78,$8F,$82,$72,$9F,$9A,$F5
	.dc.b	$39,$1F,$A8,$7E,$B5,$7C,$47,$71,$F0,$4F,$E6,$D7,$FC,$FD,$EA,$72
	.dc.b	$AF,$C8,$BF,$DE,$0F,$59,$C6,$3F,$86,$BF,$98,$BF,$A8,$FF,$17,$9F
	.dc.b	$53,$D8,$FD,$BD,$FE,$CD,$F5,$5F,$DD,$8C,$47,$86,$3E,$5C,$FF,$95
	.dc.b	$EF,$B9,$FA,$09,$F8,$AF,$A6,$B8,$C7,$F5,$53,$F8,$99,$F7,$1F,$85
	.dc.b	$5D,$EF,$F7,$B3,$C8,$F2,$B8,$79,$1F,$C1,$E3,$F8,$94,$F2,$8F,$A8
	.dc.b	$F8,$2B,$CB,$3E,$89,$F0,$DF,$C2,$9C,$CF,$EB,$AB,$F1,$55,$F6,$4E
	.dc.b	$E7,$E7,$2F,$F8,$AB,$F9,$3A,$F7,$47,$95,$E7,$1F,$AC,$3F,$00,$7B
	.dc.b	$E3,$F4,$15,$F8,$08,$FC,$79,$F4,$CF,$E4,$B1,$EF,$BF,$F2,$29,$FE
	.dc.b	$46,$AE,$B3,$E1,$3E,$29,$F7,$CF,$EE,$93,$F9,$87,$FC,$24,$75,$BF
	.dc.b	$E4,$13,$F5,$E3,$13,$FC,$56,$FF,$94,$C7,$A2,$FE,$33,$F8,$E3,$CF
	.dc.b	$5E,$2B,$FD,$75,$70,$AE,$93,$9B,$F7,$CF,$DD,$BF,$31,$F8,$B7,$81
	.dc.b	$FD,$44,$F8,$8F,$E8,$0F,$29,$FC,$6A,$FB,$BF,$36,$FB,$BE,$53,$85
	.dc.b	$F8,$5F,$86,$7A,$CF,$86,$E5,$FC,$71,$F4,$0F,$29,$F8,$03,$F5,$07
	.dc.b	$03,$B4,$FE,$35,$B7,$F5,$DF,$F1,$49,$E0,$AF,$2B,$FC,$E6,$BF,$8A
	.dc.b	$CF,$C4,$8F,$CB,$BC,$5F,$E8,$6F,$EE,$2B,$FC,$17,$1D,$8F,$2C,$FA
	.dc.b	$5F,$34,$7F,$27,$AF,$AF,$1D,$CF,$CD,$4E,$E3,$F1,$71,$E6,$3E,$0C
	.dc.b	$FE,$12,$FE,$91,$F5,$EB,$F8,$7C,$75,$9F,$84,$75,$3F,$88,$1E,$58
	.dc.b	$DD,$FE,$46,$3F,$C8,$47,$67,$F1,$CF,$F4,$DA,$FB,$F5,$FC,$1A,$FC
	.dc.b	$B3,$F0,$8F,$E2,$67,$C6,$3E,$09,$DE,$7F,$78,$FF,$03,$7F,$96,$D7
	.dc.b	$CD,$BF,$DA,$CF,$03,$D0,$7E,$C2,$BB,$DF,$F5,$7A,$E1,$3F,$AC,$AF
	.dc.b	$B3,$3F,$C1,$23,$F9,$B5,$70,$AF,$E0,$73,$FC,$46,$BB,$9C,$23,$FA
	.dc.b	$65,$7E,$6E,$B8,$5F,$F1,$89,$EF,$3F,$91,$AE,$13,$F1,$27,$F0,$97
	.dc.b	$E5,$AF,$C6,$4F,$D6,$7E,$25,$7E,$39,$08,$8A,$32,$FF,$A9,$8A,$70
	.dc.b	$78,$46,$23,$46,$0E,$46,$63,$4F,$F9,$FA,$3B,$B8,$8E,$51,$E2,$71
	.dc.b	$1C,$0F,$F1,$89,$CC,$F1,$C6,$DC,$1C,$DE,$AE,$0F,$7B,$B7,$CE,$78
	.dc.b	$DE,$CF,$A0,$F8,$2E,$DF,$F1,$B1,$EB,$7A,$BF,$01,$D3,$F1,$4F,$44
	.dc.b	$7C,$37,$D6,$74,$8D,$3E,$23,$E6,$3D,$1E,$E7,$CB,$8E,$2F,$86,$3C
	.dc.b	$F1,$F2,$1E,$71,$C0,$F6,$1E,$23,$E3,$BA,$61,$A3,$A9,$DC,$F9,$07
	.dc.b	$E0,$23,$D4,$FC,$E9,$FC,$F3,$FC,$F8,$FD,$74,$FF,$65,$AF,$CD,$CF
	.dc.b	$F7,$CF,$C1,$BF,$16,$A3,$F0,$D1,$87,$F2,$6F,$E2,$E3,$54,$8F,$CF
	.dc.b	$D7,$F0,$F3,$F3,$97,$FA,$A6,$CF,$EC,$DF,$C6,$D7,$28,$EA,$7E,$81
	.dc.b	$F1,$47,$E8,$2B,$C5,$3D,$CF,$CF,$CF,$8A,$3C,$31,$FA,$A9,$EE,$FE
	.dc.b	$CD,$B3,$FC,$95,$E7,$1E,$01,$1E,$08,$C3,$D6,$3B,$DF,$C2,$9F,$4B
	.dc.b	$FB,$0F,$9C,$FB,$D2,$7F,$56,$7E,$60,$FA,$AE,$CF,$BE,$6C,$FD,$B3
	.dc.b	$B9,$FC,$74,$7F,$02,$8E,$A7,$47,$F6,$0F,$47,$F8,$1B,$A7,$99,$A8
	.dc.b	$E6,$F7,$3F,$3C,$53,$89,$B7,$D0,$F9,$A3,$C0,$FB,$23,$9C,$77,$8F
	.dc.b	$61,$98,$E2,$F6,$8E,$2F,$BA,$38,$BD,$4F,$39,$C4,$F6,$99,$3A,$9C
	.dc.b	$5E,$AF,$F0,$87,$E3,$3C,$8D,$1C,$8F,$11,$93,$EC,$3F,$13,$C4,$FB
	.dc.b	$27,$F2,$23,$F8,$79,$F3,$63,$F9,$9B,$A7,$F9,$63,$A8,$E5,$1F,$72
	.dc.b	$39,$3F,$C9,$5E,$49,$E4,$FF,$38,$79,$47,$82,$39,$1D,$8E,$4F,$67
	.dc.b	$EC,$C7,$67,$D1,$19,$7E,$21,$DC,$E4,$69,$EC,$F2,$32,$7B,$1E,$4F
	.dc.b	$F4,$78,$EC,$E6,$38,$BD,$A3,$2F,$AE,$38,$9F,$D6,$23,$31,$FD,$61
	.dc.b	$F8,$67,$C2,$70,$7B,$A3,$F9,$FF,$93,$FA,$C3,$FA,$A9,$87,$F3,$6E
	.dc.b	$63,$FA,$F3,$98,$E3,$1F,$AE,$8E,$2F,$F5,$77,$E9,$9C,$5F,$ED,$2F
	.dc.b	$17,$47,$13,$A9,$C6,$3C,$2E,$4E,$26,$5F,$0B,$FA,$E8,$F0,$BE,$A8
	.dc.b	$E2,$E4,$EA,$65,$E2,$FD,$A8,$E2,$FB,$A3,$8B,$93,$89,$97,$8B,$FD
	.dc.b	$FE,$38,$BE,$E8,$E2,$E4,$E2,$65,$FE,$DF,$E1,$7F,$41,$7F,$C2,$2F
	.dc.b	$F3,$23,$FA,$16,$1F,$E1,$67,$FD,$25,$E1,$EE,$7F,$C3,$2F,$F3,$83
	.dc.b	$74,$76,$FF,$3F,$AF,$F0,$A4,$6E,$3F,$97,$9B,$9F,$E5,$EF,$F8,$82
	.dc.b	$BB,$27,$C8,$F0,$7F,$D4,$86,$8C,$BE,$98,$E6,$F9,$8F,$13,$FE,$2E
	.dc.b	$9F,$14,$7A,$63,$6F,$A4,$DB,$C5,$FE,$93,$55,$F6,$1E,$98,$FE,$9E
	.dc.b	$F9,$CE,$25,$38,$94,$E3,$5E,$A7,$6F,$A9,$CC,$79,$2B,$E1,$9E,$88
	.dc.b	$F0,$BF,$39,$F2,$C7,$B5,$C9,$F3,$A3,$B3,$A8,$FB,$8F,$88,$F9,$8F
	.dc.b	$C2,$9F,$EB,$E7,$F9,$20,$FC,$3B,$FD,$39,$CC,$7F,$94,$5C,$C7,$E3
	.dc.b	$23,$FB,$0C,$71,$7F,$C9,$4F,$F2,$63,$8B,$FE,$5C,$78,$BA,$38,$9D
	.dc.b	$4E,$31,$E1,$72,$71,$32,$F8,$5F,$EC,$31,$E1,$7D,$51,$C5,$C9,$D4
	.dc.b	$CB,$C5,$FE,$5B,$1C,$5F,$74,$71,$72,$71,$32,$F1,$7F,$CE,$51,$C5
	.dc.b	$F7,$47,$17,$27,$13,$2F,$A7,$2F,$F0,$B4,$7E,$25,$FF,$19,$CF,$EB
	.dc.b	$8F,$B6,$E9,$FC,$44,$7E,$39,$CD,$F6,$8C,$1F,$7A,$BF,$2F,$E1,$FE
	.dc.b	$B8,$FF,$18,$9F,$BA,$38,$BF,$C1,$23,$F0,$C7,$EB,$5F,$CF,$D7,$EC
	.dc.b	$63,$27,$E6,$23,$2F,$E7,$0C,$9E,$57,$F5,$F3,$E1,$8F,$D5,$BE,$18
	.dc.b	$FC,$94,$71,$3D,$27,$84,$FD,$89,$F9,$A3,$F3,$87,$C7,$AC,$3E,$33
	.dc.b	$66,$63,$D8,$F4,$34,$7E,$D8,$FF,$5A,$CF,$95,$FD,$CB,$E5,$8F,$6C
	.dc.b	$78,$2B,$E8,$DF,$D6,$9F,$4C,$E4,$F8,$31,$98,$E2,$FD,$98,$CD,$FF
	.dc.b	$2F,$9F,$8B,$1F,$C2,$AF,$EA,$BF,$CA,$1F,$AB,$1F,$06,$BF,$0B,$3F
	.dc.b	$74,$F7,$DF,$D5,$7D,$C7,$D1,$9F,$0D,$FE,$16,$B8,$CF,$CD,$3E,$10
	.dc.b	$BF,$85,$1F,$E8,$47,$E1,$47,$63,$EB,$47,$27,$E5,$BF,$E8,$17,$E8
	.dc.b	$57,$9E,$BE,$AC,$78,$CE,$47,$C7,$3C,$F3,$F1,$DF,$B3,$5E,$29,$FA
	.dc.b	$B1,$F0,$EB,$E7,$D7,$CB,$8E,$33,$F2,$E7,$DE,$7E,$02,$7C,$D3,$E1
	.dc.b	$3D,$F7,$F1,$E3,$27,$86,$3E,$1D,$F8,$5F,$88,$78,$6B,$A9,$ED,$7F
	.dc.b	$9F,$1D,$67,$E1,$57,$5B,$F8,$31,$D4,$CC,$FC,$A8,$C9,$E9,$7E,$09
	.dc.b	$F8,$33,$F4,$E7,$84,$F9,$31,$F0,$4F,$4C,$71,$8F,$69,$D4,$F2,$C7
	.dc.b	$86,$3F,$95,$47,$D1,$8E,$2F,$F7,$1A,$FC,$9B,$FA,$EB,$FC,$9B,$FC
	.dc.b	$22,$3E,$8D,$7E,$2C,$F4,$C7,$18,$F8,$AF,$ED,$4F,$0C,$79,$6B,$E0
	.dc.b	$C7,$86,$38,$DF,$FB,$E6,$7F,$15,$7F,$44,$FC,$8D,$7B,$0E,$07,$43
	.dc.b	$C0,$7C,$4A,$F2,$39,$AF,$D4,$BE,$37,$F9,$65,$FB,$AF,$F9,$8D,$FB
	.dc.b	$AB,$F8,$AC,$FB,$8F,$C5,$5F,$99,$C1,$F8,$1B,$E6,$FC,$07,$F8,$2C
	.dc.b	$7B,$C4,$C3,$F1,$67,$CD,$3E,$E9,$F3,$67,$93,$FC,$18,$F6,$9F,$06
	.dc.b	$FF,$01,$3E,$39,$DC,$FE,$88,$DC,$FC,$A7,$F1,$47,$F1,$C3,$F1,$47
	.dc.b	$39,$F8,$F5,$EA,$8F,$95,$5F,$01,$F9,$4F,$C0,$7E,$51,$FC,$44,$FD
	.dc.b	$39,$DA,$BF,$84,$3D,$9F,$BE,$7A,$1E,$71,$E8,$8F,$C8,$CF,$D6,$3E
	.dc.b	$4C,$FA,$AB,$E4,$3F,$22,$7E,$BB,$F2,$0F,$AF,$5F,$D7,$9F,$14,$FF
	.dc.b	$08,$7E,$61,$EC,$8F,$B0,$FA,$67,$CB,$3F,$0A,$30,$F8,$2F,$F2,$51
	.dc.b	$EB,$8C,$3D,$8F,$B6,$78,$8E,$C7,$C6,$3F,$97,$4F,$6A,$F9,$97,$F3
	.dc.b	$E7,$E4,$9F,$83,$3E,$B1,$E8,$33,$3F,$01,$C9,$DC,$FB,$53,$D1,$F4
	.dc.b	$9F,$E6,$D3,$F8,$14,$F7,$9F,$C2,$47,$77,$EB,$BD,$DC,$9D,$CC,$BD
	.dc.b	$DF,$54,$FC,$A8,$F6,$D7,$F0,$C9,$FB,$D5,$DE,$3E,$3C,$F7,$7E,$A3
	.dc.b	$CD,$CC,$F7,$73,$1F,$DC,$AB,$F2,$15,$F8,$43,$F0,$D1,$FA,$57,$F8
	.dc.b	$44,$F2,$13,$27,$F1,$11,$D2,$BF,$AE,$4F,$F0,$18,$DB,$F8,$78,$F7
	.dc.b	$CF,$86,$3F,$2D,$3F,$AF,$8F,$BB,$5E,$39,$E4,$77,$BF,$E1,$53,$C4
	.dc.b	$FE,$76,$7C,$13,$F0,$E6,$4F,$B3,$1F,$2D,$FE,$D7,$5F,$2C,$FC,$A3
	.dc.b	$F3,$8F,$E1,$53,$F9,$47,$EC,$47,$C1,$9F,$14,$7E,$94,$ED,$7F,$61
	.dc.b	$FC,$A1,$D8,$FC,$31,$FC,$B2,$FE,$C3,$D8,$FC,$F5,$FF,$32,$9F,$C2
	.dc.b	$DF,$C8,$9F,$9A,$7B,$23,$E6,$DF,$C8,$8F,$DF,$9F,$20,$FB,$0F,$C8
	.dc.b	$3E,$84,$FD,$53,$FC,$66,$E6,$3A,$D7,$86,$7E,$24,$FE,$7E,$7E,$21
	.dc.b	$FE,$0F,$70,$FB,$4E,$AF,$C4,$7F,$52,$7C,$48,$F9,$11,$E2,$8F,$54
	.dc.b	$78,$A3,$D5,$7F,$12,$7B,$47,$D7,$7F,$C0,$13,$C9,$F3,$9F,$C1,$CF
	.dc.b	$94,$7E,$06,$BD,$4F,$D0,$BF,$D3,$BF,$B8,$3A,$9F,$58,$EA,$72,$7F
	.dc.b	$4C,$F9,$8F,$F1,$C5,$7C,$79,$FC,$8C,$FD,$29,$FA,$07,$F4,$5B,$FD
	.dc.b	$4B,$6B,$83,$F8,$79,$E0,$FF,$10,$BF,$BD,$7D,$1F,$89,$1D,$27,$F9
	.dc.b	$CD,$7E,$DA,$3E,$EB,$FC,$CD,$F8,$07,$C3,$3E,$9B,$F3,$EB,$E1,$C7
	.dc.b	$C0,$BF,$D6,$4F,$D0,$8F,$88,$7E,$E9,$FB,$55,$FB,$63,$CE,$7E,$51
	.dc.b	$CD,$FE,$2E,$7E,$2C,$73,$7F,$0E,$7F,$95,$8F,$9D,$3F,$E2,$57,$E7
	.dc.b	$4F,$C0,$7E,$23,$F0,$1F,$89,$3D,$9F,$A5,$7F,$32,$BE,$1C,$F7,$9F
	.dc.b	$C9,$1E,$97,$E0,$3F,$0C,$C1,$F8,$83,$F7,$67,$E1,$6B,$F8,$0B,$CA
	.dc.b	$FF,$C9,$8F,$EB,$0F,$1C,$F4,$78,$9F,$CA,$9F,$D0,$3F,$92,$3F,$B5
	.dc.b	$9F,$10,$A6,$DF,$E6,$2F,$F1,$63,$F8,$7B,$FC,$AD,$DD,$F2,$8E,$B1
	.dc.b	$C0,$F8,$C7,$68,$D4,$74,$7B,$C6,$8F,$E8,$A7,$F8,$9A,$39,$BE,$13
	.dc.b	$47,$A4,$FA,$87,$03,$FB,$B3,$E5,$7C,$07,$F8,$21,$E7,$1E,$08,$F8
	.dc.b	$07,$A8,$D4,$79,$5F,$2C,$7F,$1D,$74,$7F,$8E,$63,$9B,$A7,$C0,$F8
	.dc.b	$DC,$BF,$19,$CB,$E9,$72,$FC,$57,$2F,$D1,$3E,$84,$79,$CF,$13,$EC
	.dc.b	$3F,$56,$7D,$E3,$F1,$2F,$ED,$5F,$6D,$FE,$8E,$7F,$09,$1F,$DB,$CF
	.dc.b	$F2,$97,$87,$FE,$53,$96,$1F,$CE,$E3,$0F,$F9,$61,$FC,$3C,$6A,$51
	.dc.b	$FD,$C1,$FD,$3D,$7E,$8E,$FF,$77,$1E,$43,$F5,$27,$92,$FF,$93,$47
	.dc.b	$9C,$E5,$1D,$CF,$F3,$3D,$F9,$2B,$FB,$D5,$F9,$27,$C2,$7F,$91,$27
	.dc.b	$C9,$1C,$A3,$C0,$7F,$3E,$8F,$E0,$B7,$E0,$8F,$1D,$EC,$FE,$D4,$FC
	.dc.b	$53,$C4,$27,$C5,$78,$7B,$C7,$86,$BE,$34,$FF,$2A,$8F,$04,$7F,$07
	.dc.b	$9F,$A5,$1F,$BD,$3B,$1E,$67,$F8,$D4,$6C,$FF,$2F,$DF,$D2,$AF,$83
	.dc.b	$5F,$12,$7C,$AD,$3E,$9C,$78,$6B,$9C,$7C,$F3,$EA,$C7,$E0,$9F,$EC
	.dc.b	$06,$0F,$21,$83,$DD,$3F,$83,$7B,$57,$E0,$A3,$DF,$3F,$CF,$6B,$FA
	.dc.b	$9D,$FE,$00,$F9,$77,$CC,$F7,$57,$F3,$38,$FA,$D7,$ED,$BE,$F5,$F7
	.dc.b	$0F,$BD,$1F,$55,$FF,$08,$9F,$C9,$E7,$AD,$FD,$C8,$F5,$47,$E2,$EB
	.dc.b	$F8,$B4,$FE,$5E,$7D,$71,$F2,$9F,$B9,$7F,$29,$A7,$CA,$8F,$64,$7F
	.dc.b	$0E,$9F,$97,$5E,$3A,$FD,$89,$F6,$9F,$1C,$FD,$98,$FD,$69,$F5,$EB
	.dc.b	$C1,$5F,$88,$8F,$6C,$7D,$79,$F8,$8D,$8F,$E9,$6D,$AF,$D5,$DF,$C7
	.dc.b	$9F,$8D,$7C,$DF,$AB,$5B,$7E,$31,$F2,$0F,$E0,$D5,$E6,$7F,$08,$7C
	.dc.b	$48,$FA,$35,$F7,$63,$E1,$1F,$E5,$8A,$FB,$AF,$43,$E0,$9F,$76,$BA
	.dc.b	$1F,$71,$F8,$15,$D1,$FD,$31,$D0,$FE,$AE,$7E,$42,$7E,$D3,$FE,$04
	.dc.b	$31,$7F,$2D,$C5,$7D,$DA,$FE,$05,$1E,$93,$E1,$CF,$DB,$AF,$E0,$73
	.dc.b	$F0,$E7,$E0,$9F,$0C,$FB,$05,$3F,$2D,$7F,$34,$FC,$8B,$E8,$7F,$01
	.dc.b	$3F,$0A,$39,$9F,$09,$F4,$BF,$9E,$3E,$CC,$F9,$9F,$2B,$F8,$2A,$F3
	.dc.b	$1F,$AA,$8F,$9C,$FE,$3C,$FC,$1C,$7E,$48,$FE,$59,$1F,$46,$7C,$4F
	.dc.b	$F0,$AA,$F4,$3F,$B6,$3D,$13,$F0,$5F,$79,$E7,$AF,$98,$F9,$EF,$F1
	.dc.b	$AF,$D8,$3E,$59,$FB,$33,$F1,$B3,$E3,$3F,$85,$5F,$C0,$9F,$C0,$C7
	.dc.b	$79,$F9,$67,$3B,$F9,$6F,$E8,$23,$93,$F7,$8F,$C9,$9E,$33,$DD,$7F
	.dc.b	$0D,$F9,$67,$F1,$F3,$F0,$13,$D1,$F2,$99,$3E,$7C,$FE,$45,$F0,$1F
	.dc.b	$39,$F9,$6F,$DA,$BF,$96,$FA,$0F,$E0,$B1,$F5,$DF,$84,$7F,$06,$9F
	.dc.b	$CF,$3A,$8E,$CF,$88,$FC,$ED,$FE,$5D,$F0,$47,$C8,$BF,$A3,$5F,$62
	.dc.b	$7A,$1F,$62,$7F,$11,$19,$BF,$DF,$4F,$86,$7E,$74,$7D,$18,$FC,$D3
	.dc.b	$F9,$29,$E1,$5F,$8E,$9F,$D8,$4F,$F2,$37,$E2,$9F,$7A,$7F,$80,$1F
	.dc.b	$9E,$9F,$A1,$7F,$CE,$E7,$AC,$7A,$1F,$B8,$7E,$3C,$F9,$A7,$F2,$A3
	.dc.b	$EA,$99,$3E,$25,$7F,$62,$9F,$89,$1F,$18,$FE,$E7,$7F,$CE,$2B,$FB
	.dc.b	$0C,$76,$3E,$29,$DA,$BD,$57,$EE,$8F,$AB,$1F,$C0,$2B,$F1,$B7,$DA
	.dc.b	$3C,$13,$DA,$7E,$6B,$D6,$3E,$69,$C5,$FC,$55,$75,$3E,$2C,$FD,$0B
	.dc.b	$FD,$4D,$FE,$82,$39,$D7,$E0,$CF,$98,$7D,$B3,$99,$DE,$7E,$A5,$7C
	.dc.b	$F9,$FC,$9C,$FE,$DA,$FF,$26,$7D,$B7,$81,$EE,$AF,$E1,$0F,$CF,$3E
	.dc.b	$DD,$7E,$8D,$FC,$35,$7C,$27,$E0,$1F,$92,$3F,$61,$7F,$9A,$BF,$F0
	.dc.b	$EC,$7A,$0F,$9D,$7F,$49,$F6,$5F,$59,$F8,$D1,$F6,$4F,$F0,$F9,$FB
	.dc.b	$AA,$F8,$8F,$C0,$38,$9F,$83,$3C,$A6,$63,$C3,$3D,$6B,$DB,$1F,$62
	.dc.b	$7A,$D7,$C7,$BF,$D5,$47,$F0,$8B,$FA,$CF,$CF,$3E,$23,$87,$D6,$7F
	.dc.b	$81,$8F,$86,$7A,$6B,$FB,$EC,$FE,$46,$3F,$BC,$9E,$83,$F2,$27,$E6
	.dc.b	$AF,$E2,$4F,$A2,$3D,$8F,$C4,$3D,$07,$49,$FA,$51,$EA,$7B,$3F,$A6
	.dc.b	$3A,$9F,$10,$F4,$9F,$80,$AF,$C8,$C7,$29,$E9,$3F,$6E,$FF,$3D,$7E
	.dc.b	$43,$FC,$71,$1D,$27,$C9,$19,$AF,$E3,$97,$D2,$B9,$1F,$12,$BD,$35
	.dc.b	$FD,$4A,$3E,$B3,$D2,$3F,$4E,$FD,$63,$F0,$D1,$FE,$B9,$AF,$F1,$63
	.dc.b	$FC,$A0,$FF,$47,$BF,$E9,$E7,$F9,$5B,$FE,$33,$7E,$79,$FE,$D4,$AF
	.dc.b	$F4,$FB,$D0,$F5,$9A,$3F,$CC,$B1,$CE,$3F,$D9,$B1,$DD,$FF,$5F,$1C
	.dc.b	$9E,$11,$DD,$F2,$9F,$AA,$3C,$07,$FB,$08,$E4,$7F,$B5,$A3,$93,$FE
	.dc.b	$47,$8E,$4F,$B4,$E4,$79,$DC,$C7,$89,$F3,$47,$58,$FF,$77,$99,$8F
	.dc.b	$3C,$64,$F1,$9A,$78,$BF,$ED,$17,$E6,$C7,$18,$F3,$1C,$4F,$90,$7D
	.dc.b	$08,$EA,$FE,$6E,$BE,$CB,$FD,$12,$BF,$09,$3F,$D6,$A3,$F2,$53,$FC
	.dc.b	$B9,$FC,$91,$FA,$D7,$FA,$57,$87,$FE,$53,$3F,$D6,$87,$F9,$24,$FE
	.dc.b	$19,$5F,$87,$3F,$8F,$D7,$E8,$2B,$F8,$B9,$FA,$88,$FC,$53,$FB,$0A
	.dc.b	$FD,$63,$F9,$98,$CA,$15,$E1,$7F,$D4,$4F,$86,$BB,$5F,$F0,$19,$FD
	.dc.b	$C1,$D6,$BC,$8F,$F2,$B3,$E1,$19,$9F,$D2,$BE,$F8,$F0,$3E,$1A,$F0
	.dc.b	$3F,$E8,$87,$C1,$3F,$E3,$F3,$FC,$EC,$7F,$B5,$CF,$F3,$61,$FE,$D5
	.dc.b	$E7,$FF,$3D,$1F,$ED,$B3,$FC,$E0,$66,$7F,$C6,$4E,$5F,$F3,$4B,$FE
	.dc.b	$DD,$8E,$B1,$FE,$C1,$7A,$9F,$34,$CB,$FE,$9C,$72,$7F,$94,$E3,$C3
	.dc.b	$1F,$EC,$97,$C3,$1F,$55,$CC,$7F,$A9,$1C,$C7,$B6,$3F,$C8,$CF,$B5
	.dc.b	$FF,$3E,$BE,$57,$27,$94,$CB,$D5,$F6,$9E,$13,$DA,$E5,$FE,$D5,$1F
	.dc.b	$05,$F8,$D1,$ED,$7A,$9E,$D3,$AB,$E5,$7F,$4C,$1F,$B2,$FD,$A6,$B8
	.dc.b	$7F,$7A,$EC,$FE,$04,$E9,$FE,$30,$E8,$FE,$1E,$64,$D3,$FC,$85,$AF
	.dc.b	$33,$C6,$ED,$DB,$CD,$EE,$79,$1E,$26,$5E,$26,$4E,$2F,$91,$E2,$7A
	.dc.b	$CC,$BC,$9C,$9F,$04,$C9,$E2,$32,$7E,$B5,$F0,$9D,$5F,$09,$D4,$F8
	.dc.b	$0E,$5F,$E8,$07,$BC,$EC,$69,$F1,$3C,$8C,$9E,$F7,$B6,$1F,$E1,$27
	.dc.b	$F1,$91,$FA,$97,$F0,$0F,$F8,$B5,$CB,$FA,$17,$31,$C6,$3F,$20,$71
	.dc.b	$8F,$F6,$3F,$3F,$E0,$A5,$87,$E1,$65,$39,$8B,$5F,$86,$AD,$CF,$BA
	.dc.b	$3F,$C6,$15,$FA,$E9,$E7,$1F,$04,$FD,$84,$79,$2B,$B4,$7F,$83,$63
	.dc.b	$F6,$6F,$FA,$5D,$E5,$1F,$46,$3F,$36,$7F,$A7,$EB,$C9,$3F,$9B,$7C
	.dc.b	$C7,$E5,$6B,$CB,$3F,$B1,$3E,$31,$F0,$8F,$8D,$3F,$A5,$8F,$8D,$1F
	.dc.b	$A9,$3E,$31,$E3,$3F,$5A,$FB,$23,$CF,$1F,$E7,$D9,$F1,$C7,$C6,$8F
	.dc.b	$E6,$93,$E2,$7C,$A7,$8D,$EE,$73,$7D,$A7,$CA,$3B,$D7,$C1,$7C,$66
	.dc.b	$4F,$59,$E7,$3D,$67,$C5,$3F,$9A,$C7,$DC,$9F,$8D,$18,$9F,$33,$6B
	.dc.b	$ED,$57,$FA,$8E,$BE,$8C,$FD,$98,$F8,$11,$F6,$4F,$E0,$87,$E4,$23
	.dc.b	$EE,$BF,$2E,$BE,$EB,$99,$E2,$FE,$42,$7C,$4F,$91,$F8,$F3,$E4,$6D
	.dc.b	$79,$23,$F5,$D5,$F2,$1F,$89,$3F,$00,$F8,$91,$97,$D1,$1F,$C4,$5F
	.dc.b	$B4,$7C,$88,$FA,$0F,$CC,$7F,$2E,$FC,$47,$E4,$55,$9F,$24,$A7,$F1
	.dc.b	$B3,$F0,$CF,$AC,$F4,$8F,$23,$F2,$DF,$B6,$79,$0F,$D7,$47,$C7,$3E
	.dc.b	$5B,$F9,$53,$E5,$9F,$6C,$F8,$AF,$C7,$7C,$53,$E4,$9F,$1C,$F3,$3C
	.dc.b	$71,$F1,$A7,$99,$F1,$9F,$AA,$F7,$9F,$AA,$FC,$67,$E5,$4F,$E7,$9F
	.dc.b	$8B,$5F,$9E,$3F,$37,$3F,$9C,$9F,$84,$FE,$EC,$FC,$DC,$FC,$17,$E8
	.dc.b	$57,$8E,$3B,$C7,$8E,$7E,$0B,$F9,$B3,$CF,$1E,$07,$F0,$F1,$F4,$EB
	.dc.b	$F7,$D5,$FC,$52,$3C,$27,$E3,$DF,$C0,$4F,$58,$F8,$6F,$D1,$7F,$65
	.dc.b	$1F,$0A,$3C,$15,$E4,$3C,$0F,$D0,$9F,$03,$FC,$86,$7C,$11,$F3,$E3
	.dc.b	$E0,$1F,$8B,$8F,$CB,$9D,$67,$F0,$91,$EC,$8C,$CF,$C0,$9C,$9E,$18
	.dc.b	$FF,$16,$D7,$CF,$AF,$DB,$1E,$17,$F4,$8F,$F4,$C9,$FA,$EF,$D0,$9E
	.dc.b	$B5,$F0,$0F,$B4,$E2,$3C,$E7,$E5,$9E,$EF,$CE,$AF,$BA,$F8,$27,$E6
	.dc.b	$BF,$87,$9F,$9B,$5F,$74,$F0,$C7,$C1,$3C,$91,$F0,$4F,$C0,$1C,$0E
	.dc.b	$EF,$CD,$AE,$EF,$F0,$03,$EA,$D7,$C5,$9E,$31,$FA,$83,$F2,$CF,$57
	.dc.b	$EE,$47,$99,$FB,$71,$DD,$FF,$33,$4F,$CB,$7B,$1F,$99,$9F,$B4,$7E
	.dc.b	$25,$EC,$7E,$24,$F8,$E7,$D9,$3D,$AF,$95,$F0,$BF,$C7,$5F,$61,$F6
	.dc.b	$0E,$C7,$EE,$CF,$DC,$BF,$BB,$AF,$E8,$D1,$F3,$E3,$C6,$FC,$F3,$C7
	.dc.b	$1F,$3E,$BE,$04,$FD,$C9,$F6,$9F,$64,$F0,$1C,$CF,$04,$FF,$1C,$AF
	.dc.b	$9F,$1F,$C2,$9F,$C5,$BF,$C2,$27,$D8,$75,$70,$FE,$2D,$FC,$71,$E0
	.dc.b	$3D,$E7,$CC,$3C,$07,$F6,$53,$C8,$64,$FB,$CF,$D7,$9F,$D0,$D7,$D7
	.dc.b	$9E,$E7,$D0,$8F,$D2,$3D,$CF,$93,$3F,$B4,$7E,$69,$DC,$F1,$9E,$D3
	.dc.b	$F5,$07,$E3,$1F,$53,$F9,$73,$BB,$F8,$13,$FC,$2A,$F3,$3B,$D7,$C8
	.dc.b	$9F,$E5,$F1,$F3,$2B,$FC,$51,$39,$7F,$82,$C7,$E5,$AB,$E5,$BF,$20
	.dc.b	$EE,$7D,$D7,$BD,$7F,$10,$8E,$E7,$DA,$3E,$3C,$7E,$35,$C9,$F2,$CF
	.dc.b	$73,$F2,$E7,$F4,$EF,$BE,$3E,$54,$7C,$DA,$FD,$14,$7C,$53,$E4,$D7
	.dc.b	$84,$FC,$33,$E1,$3A,$D7,$E5,$1F,$D6,$9F,$94,$AF,$E1,$91,$F2,$4F
	.dc.b	$F3,$F4,$F3,$8F,$8B,$3C,$E3,$E4,$BF,$35,$EE,$FA,$DC,$9F,$9F,$9F
	.dc.b	$93,$1F,$5C,$EB,$3F,$A2,$7E,$65,$7B,$2B,$F5,$71,$F2,$0F,$AD,$5F
	.dc.b	$AF,$AF,$ED,$07,$D6,$8F,$DF,$47,$A4,$F7,$4F,$E5,$23,$DC,$7E,$45
	.dc.b	$F3,$BF,$24,$FC,$FB,$EF,$8E,$D5,$FD,$58,$FD,$6C,$FA,$23,$D6,$FF
	.dc.b	$31,$3F,$85,$BF,$94,$70,$7F,$84,$5E,$6F,$E6,$9F,$82,$F8,$EB,$E6
	.dc.b	$BF,$25,$F6,$57,$C9,$9E,$93,$FC,$64,$FC,$89,$F6,$9F,$C8,$CF,$F3
	.dc.b	$48,$FC,$69,$ED,$3E,$51,$F5,$DE,$EE,$4F,$C2,$9F,$9F,$8E,$F3,$FB
	.dc.b	$88,$EE,$FC,$B8,$EF,$1E,$97,$DA,$FE,$4A,$7C,$53,$F3,$8F,$B7,$1D
	.dc.b	$8F,$AE,$7E,$A6,$7F,$40,$FF,$32,$3E,$FB,$FC,$15,$FA,$B3,$D0,$F4
	.dc.b	$CF,$43,$F1,$EF,$E6,$A3,$A4,$FE,$0A,$3F,$56,$FD,$08,$FA,$8F,$57
	.dc.b	$E2,$CF,$E0,$63,$11,$E7,$8F,$A4,$FE,$C8,$FC,$59,$F1,$27,$AC,$FE
	.dc.b	$F6,$7E,$04,$7E,$5A,$BF,$9A,$9F,$9D,$71,$5F,$BD,$AF,$CC,$3F,$44
	.dc.b	$EC,$FC,$E7,$CF,$3F,$C5,$4E,$6F,$F3,$43,$ED,$57,$C2,$7F,$26,$72
	.dc.b	$9F,$9A,$7F,$43,$9F,$11,$F9,$D8,$EC,$E1,$FC,$B4,$7C,$97,$EA,$1F
	.dc.b	$60,$FC,$83,$F2,$E3,$84,$7E,$04,$FE,$7C,$7F,$BC,$63,$FC,$E3,$08
	.dc.b	$84,$3F,$B0,$C7,$FB,$3D,$E2,$7E,$58,$A6,$9E,$0F,$E7,$CE,$D1,$87
	.dc.b	$FC,$FE,$6C,$C3,$FE,$FE,$2B,$B3,$F8,$51,$E3,$74,$F6,$74,$E8,$F0
	.dc.b	$1E,$53,$B9,$93,$D3,$1F,$E7,$F7,$D2,$64,$FF,$49,$B9,$78,$C7,$FB
	.dc.b	$CD,$F1,$1E,$E8,$D4,$78,$8D,$47,$8A,$32,$FA,$E3,$2F,$8A,$3E,$8B
	.dc.b	$93,$FB,$29,$D8,$FA,$4F,$27,$E2,$BF,$4D,$FC,$89,$FD,$06,$BF,$8A
	.dc.b	$9F,$B3,$8F,$E0,$31,$FB,$63,$F8,$23,$FA,$AA,$FC,$A4,$2B,$FD,$DD
	.dc.b	$1F,$E1,$B7,$FD,$B4,$7F,$99,$4F,$C9,$C7,$E7,$63,$F7,$D5,$E0,$8E
	.dc.b	$B3,$FA,$37,$67,$F8,$22,$77,$1F,$B0,$3F,$CC,$87,$68,$FC,$84,$27
	.dc.b	$C9,$3C,$A7,$F9,$4C,$F2,$70,$FE,$F2,$3F,$51,$5F,$A1,$9F,$35,$7E
	.dc.b	$F2,$3F,$49,$3E,$09,$FF,$71,$BE,$D8,$E1,$1E,$79,$E1,$1F,$C2,$5F
	.dc.b	$E3,$A6,$23,$F9,$1C,$FB,$E7,$AC,$7F,$9C,$63,$D8,$FC,$A8,$EA,$FA
	.dc.b	$A7,$93,$FD,$06,$3E,$BB,$FC,$82,$3E,$D4,$7D,$77,$E7,$47,$C4,$3B
	.dc.b	$1E,$A3,$E4,$BF,$22,$3B,$D7,$C5,$9F,$E1,$27,$A0,$FC,$45,$7F,$18
	.dc.b	$7E,$03,$F2,$A7,$F8,$1C,$FF,$B5,$A3,$E4,$4F,$F9,$75,$F9,$13,$F1
	.dc.b	$4F,$4C,$7F,$35,$AF,$4C,$FC,$59,$D1,$EB,$7F,$05,$3E,$A3,$A1,$EA
	.dc.b	$7F,$29,$1E,$A9,$FA,$31,$E5,$9F,$51,$FD,$D5,$F5,$3F,$35,$D3,$FE
	.dc.b	$FC,$3F,$D2,$EF,$AA,$BF,$1C,$FE,$86,$3C,$0F,$E3,$27,$BC,$7C,$83
	.dc.b	$EB,$57,$A5,$D3,$F0,$2B,$F5,$75,$FA,$AA,$FF,$2C,$BF,$82,$AF,$C7
	.dc.b	$D7,$AC,$E1,$3F,$72,$7F,$9B,$CF,$ED,$CF,$DE,$9E,$D3,$FA,$1B,$E9
	.dc.b	$8F,$C1,$47,$E3,$67,$E6,$CF,$CF,$8F,$8A,$7F,$52,$9E,$CF,$E9,$23
	.dc.b	$E2,$D7,$DD,$8F,$A4,$FC,$48,$FB,$B1,$CA,$3F,$61,$5E,$93,$E1,$16
	.dc.b	$7C,$13,$F3,$CF,$2C,$FF,$0B,$7F,$2C,$79,$CE,$95,$FA,$28,$FE,$FD
	.dc.b	$3C,$9B,$3F,$96,$3E,$09,$E6,$9F,$82,$FC,$CA,$F8,$2E,$DF,$23,$5C
	.dc.b	$BF,$2D,$FD,$BB,$FE,$92,$AF,$AC,$FE,$76,$38,$BB,$78,$BF,$2E,$3F
	.dc.b	$1F,$5F,$AD,$8F,$ED,$C7,$C0,$AF,$C0,$BF,$00,$F3,$1F,$02,$7E,$8C
	.dc.b	$7E,$9A,$3C,$C6,$4F,$C8,$4F,$D5,$7C,$C7,$33,$CD,$5E,$48,$FA,$AD
	.dc.b	$9F,$98,$75,$9F,$51,$FD,$FE,$7A,$1F,$55,$FD,$89,$F5,$4F,$CF,$D7
	.dc.b	$D3,$9E,$51,$F0,$4F,$D0,$D7,$E0,$8F,$E0,$F1,$FE,$0A,$9F,$D1,$40
	.dc.b	$F8,$8F,$EB,$E7,$B5,$78,$AB,$E3,$C7,$88,$F8,$E7,$88,$F8,$A5,$3D
	.dc.b	$67,$E3,$E7,$BD,$76,$7F,$33,$19,$7F,$50,$78,$5E,$B5,$F8,$E7,$9B
	.dc.b	$5F,$74,$FE,$1A,$3E,$31,$FA,$58,$FE,$0F,$19,$AF,$4C,$E6,$7D,$B1
	.dc.b	$F2,$23,$EA,$C7,$DD,$3F,$0F,$5E,$83,$F8,$41,$E5,$7E,$E9,$F5,$E7
	.dc.b	$E6,$47,$A8,$F6,$CF,$9C,$EB,$3F,$14,$F4,$D7,$D4,$7F,$90,$9F,$A9
	.dc.b	$9E,$D3,$FB,$F9,$F5,$1F,$5E,$3C,$67,$6A,$F5,$1F,$8C,$AF,$EF,$E7
	.dc.b	$D4,$AF,$B5,$5C,$6B,$F5,$71,$D6,$3C,$53,$FA,$57,$4F,$E6,$E3,$4F
	.dc.b	$C5,$3F,$91,$4F,$F2,$47,$FB,$21,$EA,$3F,$46,$7E,$1C,$F8,$E7,$DF
	.dc.b	$3D,$CF,$A5,$F3,$9F,$7A,$BD,$71,$FB,$28,$E9,$3F,$7A,$7F,$86,$C7
	.dc.b	$48,$F6,$D7,$E3,$63,$F0,$2F,$EB,$EB,$94,$7F,$38,$3E,$89,$F6,$4E
	.dc.b	$B5,$F8,$63,$AD,$7C,$08,$F4,$1E,$C7,$F1,$4F,$E5,$A3,$CE,$FF,$04
	.dc.b	$6B,$F8,$A9,$FC,$23,$E4,$3C,$07,$CF,$AF,$43,$F9,$43,$E1,$9E,$63
	.dc.b	$F1,$15,$F8,$A3,$DF,$3F,$82,$8E,$87,$9A,$3E,$34,$7E,$2A,$BF,$C2
	.dc.b	$15,$F0,$EB,$CC,$76,$AE,$93,$F9,$A8,$FA,$A7,$48,$EF,$5F,$88,$9E
	.dc.b	$73,$F8,$33,$F0,$27,$CD,$9E,$6F,$CC,$AF,$C5,$BF,$83,$8F,$98,$74
	.dc.b	$7F,$1D,$3D,$1F,$7B,$F9,$B7,$A3,$6B,$F6,$87,$8C,$FF,$09,$D7,$E6
	.dc.b	$9C,$CF,$47,$2F,$9C,$FF,$11,$1F,$06,$7A,$47,$D6,$8F,$A4,$7D,$57
	.dc.b	$E7,$38,$7F,$40,$F2,$3C,$AF,$E3,$8F,$0B,$EF,$8F,$65,$7F,$47,$8F
	.dc.b	$F5,$8C,$FF,$A9,$CF,$F4,$91,$FE,$8E,$7E,$91,$FE,$C0,$AF,$92,$FE
	.dc.b	$B4,$FF,$58,$46,$DC,$1F,$CD,$8A,$7F,$8B,$23,$FD,$29,$1F,$C6,$5E
	.dc.b	$71,$FA,$A3,$4F,$89,$FF,$4E,$BF,$AE,$8F,$F6,$29,$F2,$A3,$D0,$69
	.dc.b	$FE,$48,$68,$F2,$3F,$B6,$8E,$F1,$97,$C6,$FF,$AD,$DE,$CF,$C2,$73
	.dc.b	$1D,$63,$BB,$F1,$E3,$2F,$C5,$78,$BF,$ED,$88,$F0,$9E,$F8,$CC,$71
	.dc.b	$7D,$CF,$18,$F4,$9E,$18,$CC,$FE,$AA,$3F,$8B,$4F,$E0,$DF,$C8,$1F
	.dc.b	$E2,$07,$F8,$25,$7F,$08,$3E,$F9,$FD,$BE,$7F,$B8,$D7,$F3,$97,$FB
	.dc.b	$9D,$1F,$F7,$A9,$FE,$60,$3A,$1E,$1A,$FD,$95,$7E,$12,$31,$1F,$93
	.dc.b	$9F,$CC,$3F,$C4,$C1,$FE,$C9,$8F,$0D,$58,$FE,$B1,$3E,$17,$FC,$F9
	.dc.b	$3E,$0A,$FE,$45,$3F,$C2,$4F,$C0,$47,$ED,$E7,$F3,$91,$89,$F2,$1C
	.dc.b	$1F,$8E,$75,$8F,$74,$7F,$83,$EB,$C5,$18,$9D,$3E,$4A,$D3,$FC,$99
	.dc.b	$F5,$BF,$94,$6B,$FE,$A2,$29,$67,$FD,$E5,$18,$8F,$03,$4C,$3F,$E2
	.dc.b	$F3,$F2,$EE,$E3,$FC,$48,$EE,$36,$FC,$C8,$EB,$1F,$99,$8D,$3B,$8C
	.dc.b	$BE,$73,$EE,$9C,$8E,$E7,$13,$27,$17,$4F,$13,$FD,$E2,$71,$7A,$BF
	.dc.b	$A1,$8F,$0B,$DA,$32,$F9,$DF,$9A,$FB,$CE,$C7,$03,$51,$5F,$95,$1E
	.dc.b	$27,$B1,$E6,$3B,$39,$78,$99,$3C,$EF,$F9,$35,$FA,$5C,$DF,$36,$0E
	.dc.b	$23,$E9,$BE,$77,$07,$F1,$26,$BF,$79,$D3,$4F,$E1,$4E,$9F,$C2,$BA
	.dc.b	$69,$A3,$81,$E1,$8F,$33,$83,$F8,$64,$73,$7F,$22,$F3,$69,$CC,$EA
	.dc.b	$73,$8F,$5B,$93,$66,$4E,$8F,$F9,$C6,$3B,$47,$EE,$1E,$CD,$39,$14
	.dc.b	$E4,$79,$CF,$BC,$F2,$7E,$69,$FC,$C9,$FA,$06,$4E,$AE,$8F,$9C,$73
	.dc.b	$3D,$B1,$F1,$1F,$D7,$DF,$D9,$3F,$58,$7E,$A9,$3D,$0D,$3F,$B4,$74
	.dc.b	$65,$FE,$80,$F2,$7F,$66,$E5,$FE,$A2,$E5,$D4,$78,$DF,$E0,$71,$D9
	.dc.b	$F2,$9C,$8F,$EA,$2E,$5D,$99,$8E,$4F,$F2,$68,$D3,$4D,$47,$37,$C6
	.dc.b	$7A,$CC,$9E,$53,$FA,$94,$57,$C7,$1D,$9F,$8A,$7C,$03,$91,$EC,$7B
	.dc.b	$3C,$8C,$9D,$DE,$51,$E6,$8E,$67,$13,$F8,$F7,$07,$EB,$4A,$FF,$1E
	.dc.b	$7E,$F9,$97,$F9,$3B,$F6,$4C,$BF,$DC,$DE,$2F,$29,$F5,$47,$18,$FE
	.dc.b	$62,$F1,$7F,$BF,$B9,$7F,$9B,$B9,$8E,$31,$D4,$F0,$9E,$53,$89,$93
	.dc.b	$A9,$97,$8B,$FE,$14,$8F,$2B,$CA,$32,$FF,$80,$E3,$2F,$28,$CB,$FC
	.dc.b	$8E,$32,$F9,$8F,$09,$D4,$F2,$9F,$06,$3C,$CE,$4F,$29,$93,$FA,$77
	.dc.b	$07,$F1,$D2,$BF,$D3,$9F,$D7,$99,$7F,$AB,$BF,$A9,$32,$FF,$88,$DE
	.dc.b	$2F,$28,$FC,$54,$71,$8F,$EC,$2F,$17,$FC,$5E,$E5,$FE,$CE,$E6,$38
	.dc.b	$C7,$53,$C2,$79,$4E,$26,$4E,$A6,$5E,$2F,$F9,$12,$3C,$AF,$28,$CB
	.dc.b	$FE,$33,$8C,$BC,$A3,$2F,$F5,$38,$CB,$E6,$3C,$27,$53,$CA,$7C,$18
	.dc.b	$F3,$39,$3C,$A6,$4F,$EF,$DC,$1F,$D3,$4A,$FF,$7E,$7F,$8F,$99,$7F
	.dc.b	$C0,$EF,$F1,$93,$2F,$F9,$4D,$E2,$F2,$8F,$DD,$47,$18,$FF,$08,$BC
	.dc.b	$5F,$F2,$FB,$97,$FC,$2E,$E6,$38,$C7,$53,$C2,$79,$4E,$26,$4E,$A6
	.dc.b	$5E,$2F,$F9,$D2,$3C,$AF,$28,$CB,$FE,$63,$8C,$BC,$A3,$2F,$F8,$0E
	.dc.b	$32,$F9,$8F,$09,$D4,$F2,$9F,$06,$3C,$CE,$4F,$29,$93,$FC,$5D,$BF
	.dc.b	$F6,$AC,$BF,$A1,$C8,$46,$71,$B1,$FE,$60,$6B,$9A,$FF,$2C,$CB,$4F
	.dc.b	$F2,$CB,$63,$FC,$BB,$7F,$D4,$23,$11,$FE,$46,$8C,$4F,$F9,$2A,$3F
	.dc.b	$CA,$27,$47,$99,$BB,$F2,$BE,$3A,$F0,$5F,$8E,$BC,$E7,$82,$BF,$CB
	.dc.b	$53,$EB,$3F,$AA,$5F,$8E,$BC,$0F,$F3,$68,$F8,$11,$F0,$DE,$0F,$73
	.dc.b	$73,$FE,$5E,$8F,$5C,$FC,$A9,$F2,$D7,$86,$7D,$ED,$BE,$2F,$B4,$F3
	.dc.b	$4F,$78,$F7,$4F,$09,$FF,$39,$BE,$E9,$F8,$51,$F4,$1F,$9E,$70,$3C
	.dc.b	$C7,$C3,$8F,$5B,$F1,$23,$9B,$A3,$9B,$E8,$3F,$D0,$2F,$8D,$F4,$1F
	.dc.b	$E7,$AA,$FA,$F1,$D6,$BE,$AC,$E8,$F4,$BE,$E7,$C1,$3F,$16,$BE,$5C
	.dc.b	$73,$BF,$C5,$DF,$8A,$7F,$C1,$0F,$D1,$3F,$D0,$EF,$F7,$53,$FD,$28
	.dc.b	$FF,$A1,$1D,$BE,$D9,$F8,$87,$E7,$9F,$F6,$EC,$68,$FF,$0A,$1C,$DC
	.dc.b	$9F,$ED,$D7,$2F,$53,$31,$FE,$EE,$8C,$BD,$63,$93,$97,$CE,$71,$32
	.dc.b	$FF,$B8,$63,$8B,$A8,$CB,$FE,$81,$8C,$BF,$EA,$87,$DE,$FA,$CE,$C6
	.dc.b	$63,$B1,$E7,$72,$7E,$55,$CB,$C4,$F1,$3E,$67,$D4,$F1,$72,$75,$32
	.dc.b	$71,$AF,$B5,$1F,$8F,$8F,$B5,$3E,$67,$F7,$27,$E0,$2B,$F6,$33,$B8
	.dc.b	$FC,$B3,$F9,$53,$F3,$2F,$E9,$DF,$CA,$3F,$B9,$3E,$D8,$F1,$1E,$2F
	.dc.b	$48,$FF,$74,$C6,$2F,$FC,$A0,$DA,$FE,$18,$F0,$9E,$78,$7A,$9F,$01
	.dc.b	$EA,$9F,$CC,$57,$EA,$0F,$CB,$C7,$A2,$3F,$2C,$7A,$1F,$D5,$9C,$5F
	.dc.b	$63,$E4,$BF,$33,$E9,$7F,$2C,$D9,$F9,$17,$E5,$8F,$91,$5F,$A1,$7C
	.dc.b	$A7,$84,$F8,$86,$9F,$9D,$3F,$C7,$30,$FD,$5C,$EA,$3D,$53,$F4,$63
	.dc.b	$E7,$3E,$67,$C8,$F7,$7C,$8F,$FB,$DA,$BE,$53,$FE,$B0,$7C,$D3,$D2
	.dc.b	$7C,$A7,$49,$F8,$AF,$F1,$C7,$CA,$7D,$C7,$F0,$B1,$C5,$F9,$27,$B4
	.dc.b	$CC,$1F,$94,$F9,$67,$EE,$46,$4F,$9D,$3F,$2A,$FF,$41,$1E,$18,$FB
	.dc.b	$2F,$94,$F8,$95,$F6,$70,$FE,$85,$1F,$66,$3D,$2F,$F2,$27,$D2,$E4
	.dc.b	$E2,$79,$1F,$E5,$97,$E4,$8F,$23,$F2,$22,$DF,$92,$7A,$5F,$91,$FE
	.dc.b	$67,$3F,$4E,$FF,$29,$1F,$10,$F4,$1F,$10,$F6,$47,$DB,$78,$BE,$8B
	.dc.b	$FD,$2B,$F1,$EF,$F1,$F1,$E3,$9F,$CA,$1F,$A2,$C3,$F2,$B7,$F0,$A3
	.dc.b	$E9,$3F,$D8,$EF,$E3,$4F,$E6,$23,$FB,$44,$79,$E3,$89,$93,$F2,$33
	.dc.b	$E7,$3F,$1B,$1F,$1A,$FC,$66,$AF,$F1,$71,$DB,$0F,$CD,$1F,$C2,$F0
	.dc.b	$F1,$5F,$C0,$3F,$1E,$7C,$03,$E9,$9F,$32,$7E,$24,$79,$9F,$89,$7F
	.dc.b	$B2,$BF,$AD,$5F,$88,$9F,$E0,$B3,$F5,$6B,$E8,$9F,$0E,$7E,$8B,$E2
	.dc.b	$7B,$1E,$27,$B3,$E2,$78,$E1,$F0,$F0,$F5,$3F,$81,$C3,$EA,$E1,$F0
	.dc.b	$E7,$F8,$BE,$1F,$0C,$F1,$57,$CC,$8F,$87,$3F,$C5,$0F,$8C,$7E,$AE
	.dc.b	$FC,$87,$82,$37,$1F,$BD,$BE,$F1,$FB,$9B,$FA,$2F,$E9,$F0,$FC,$15
	.dc.b	$7F,$01,$BF,$81,$1F,$34,$FB,$4F,$58,$F0,$61,$E6,$78,$3E,$F7,$CD
	.dc.b	$3F,$A5,$AF,$01,$EE,$3C,$0F,$F1,$97,$F4,$55,$F2,$A3,$BD,$7C,$1A
	.dc.b	$EF,$1E,$67,$F6,$6F,$E5,$A3,$E6,$5F,$DD,$3D,$B3,$F7,$27,$85,$F6
	.dc.b	$AF,$CA,$DF,$C8,$7B,$4F,$EF,$4F,$0D,$7B,$2B,$E5,$BF,$8A,$3F,$58
	.dc.b	$FA,$DE,$D1,$EB,$7F,$23,$3D,$8F,$8C,$FA,$23,$F1,$71,$89,$F2,$1F
	.dc.b	$86,$8F,$EB,$F8,$7D,$8A,$FE,$13,$3D,$DF,$B5,$3E,$C7,$BC,$71,$8F
	.dc.b	$11,$F9,$5A,$F1,$3B,$31,$1B,$3F,$09,$3E,$C3,$FA,$06,$1F,$93,$9F
	.dc.b	$1B,$F3,$67,$C5,$3F,$C7,$27,$A1,$FC,$5C,$FC,$46,$1F,$36,$3E,$51
	.dc.b	$F6,$A3,$F8,$13,$DD,$FE,$74,$7F,$50,$7D,$AE,$9F,$D3,$1E,$49,$FC
	.dc.b	$49,$FD,$26,$7F,$2E,$7E,$24,$F0,$47,$CF,$9F,$1D,$FE,$1A,$3C,$11
	.dc.b	$F3,$1F,$B3,$5F,$09,$EF,$1E,$4C,$3F,$94,$3F,$DE,$4F,$7D,$F8,$2F
	.dc.b	$E6,$C7,$D0,$3E,$14,$78,$E3,$C1,$5D,$27,$C0,$FB,$AF,$C0,$7E,$99
	.dc.b	$F7,$1E,$43,$F7,$F1,$F9,$5A,$F6,$CF,$E2,$27,$F5,$F3,$FC,$BA,$39
	.dc.b	$3E,$9C,$3E,$D3,$F2,$E3,$2F,$8A,$7E,$D3,$E9,$AF,$64,$FE,$16,$7C
	.dc.b	$33,$FA,$F3,$F0,$6F,$CA,$7F,$95,$3E,$B3,$FA,$F9,$97,$EE,$3D,$A3
	.dc.b	$EC,$5F,$F0,$78,$FE,$94,$7E,$1A,$7D,$D1,$F2,$27,$CF,$1D,$A7,$91
	.dc.b	$D8,$F9,$8E,$4F,$EB,$A7,$AE,$7B,$47,$F7,$CA,$FE,$35,$3D,$8F,$5C
	.dc.b	$72,$7C,$E7,$2A,$FA,$B1,$F8,$78,$F9,$91,$F9,$28,$F8,$B1,$FB,$C3
	.dc.b	$F0,$B5,$E0,$7F,$AA,$E1,$FB,$B7,$FA,$81,$CA,$7E,$F3,$FC,$9D,$EC
	.dc.b	$FD,$53,$C4,$F6,$7F,$36,$76,$AF,$E2,$AF,$6C,$3F,$6D,$3F,$8A,$3E
	.dc.b	$A4,$75,$3C,$27,$53,$F3,$87,$C7,$9E,$6F,$F2,$28,$FD,$56,$1F,$9B
	.dc.b	$8F,$8F,$3F,$51,$F8,$E7,$27,$E3,$BD,$8F,$A8,$66,$7D,$53,$CE,$3C
	.dc.b	$11,$FC,$16,$78,$C7,$F9,$48,$F8,$D1,$EA,$7F,$3C,$FC,$67,$AD,$FF
	.dc.b	$82,$9F,$F0,$31,$FB,$2C,$3F,$6B,$5F,$90,$3D,$B7,$FD,$EE,$BE,$A5
	.dc.b	$77,$3E,$23,$CD,$EE,$7A,$CF,$7B,$98,$FC,$5C,$7E,$94,$F5,$9F,$8D
	.dc.b	$3F,$61,$7E,$77,$8E,$1F,$9D,$8C,$57,$F0,$E3,$A6,$1F,$D5,$2B,$A6
	.dc.b	$1E,$9A,$F8,$C7,$E8,$5F,$A5,$5C,$6F,$A5,$79,$E3,$0F,$E1,$EB,$8B
	.dc.b	$F1,$B0,$FD,$FC,$7F,$0A,$9F,$86,$7F,$89,$A7,$CE,$7E,$02,$3C,$F5
	.dc.b	$FC,$32,$FC,$93,$EF,$6B,$E8,$9F,$31,$C2,$3F,$51,$1F,$93,$3E,$81
	.dc.b	$F6,$A7,$FB,$24,$FF,$5C,$3F,$8E,$DF,$CE,$9F,$D7,$1F,$30,$F7,$4F
	.dc.b	$8E,$7B,$C6,$2B,$BB,$FE,$38,$3F,$08,$FF,$1D,$7E,$16,$1F,$B7,$7F
	.dc.b	$CB,$18,$7C,$2B,$FF,$05,$DF,$E4,$73,$70,$FF,$AF,$1B,$8A,$7F,$AD
	.dc.b	$E0,$9C,$C6,$4F,$F4,$C3,$CA,$3F,$DB,$AF,$48,$FF,$59,$3C,$23,$FB
	.dc.b	$2B,$CA,$3B,$47,$FA,$96,$39,$29,$CB,$E1,$AF,$31,$FE,$D0,$9E,$55
	.dc.b	$E2,$8F,$F4,$A4,$78,$63,$91,$9A,$E7,$5D,$EB,$FC,$FB,$5E,$09,$EE
	.dc.b	$F1,$6C,$77,$36,$76,$3C,$6F,$49,$F3,$3F,$16,$BC,$91,$F1,$27,$71
	.dc.b	$F3,$2B,$D3,$3E,$33,$51,$F2,$27,$C1,$5C,$E7,$47,$B0,$F5,$4E,$E3
	.dc.b	$C5,$5D,$27,$E4,$1D,$27,$83,$E6,$7C,$55,$EC,$8D,$57,$C0,$39,$C6
	.dc.b	$63,$8D,$7C,$5A,$F9,$87,$2A,$F3,$4F,$80,$FC,$E1,$FA,$43,$F6,$26
	.dc.b	$30,$FD,$A1,$FB,$C7,$18,$7E,$5A,$BF,$B2,$C7,$EB,$67,$73,$FA,$58
	.dc.b	$C3,$B7,$FA,$69,$8A,$FF,$1E,$B8,$3F,$90,$9E,$39,$F0,$47,$F3,$67
	.dc.b	$AB,$FA,$E7,$F5,$0F,$28,$FE,$80,$75,$AF,$EA,$87,$99,$EB,$3C,$A3
	.dc.b	$FB,$1C,$62,$BC,$CF,$23,$15,$D6,$BF,$9B,$1E,$0A,$EB,$3E,$98,$C3
	.dc.b	$FD,$2A,$3C,$11,$88,$FE,$7F,$5F,$CA,$1F,$73,$F1,$DE,$F3,$D1,$F1
	.dc.b	$4F,$17,$F8,$C1,$FC,$94,$FE,$21,$1E,$A8,$D4,$7D,$AA,$F9,$15,$EC
	.dc.b	$9F,$4B,$FD,$E2,$7C,$C7,$A5,$C3,$E6,$3B,$1E,$69,$FA,$11,$A8,$F1
	.dc.b	$9F,$84,$8F,$8D,$1E,$D9,$F9,$D5,$FD,$AD,$F2,$9F,$71,$FE,$53,$5D
	.dc.b	$5F,$54,$F5,$8F,$54,$6A,$79,$BF,$32,$75,$1C,$E7,$F0,$07,$B0,$F9
	.dc.b	$31,$A3,$E7,$47,$89,$A6,$6B,$DB,$5F,$E4,$B9,$E9,$5F,$50,$E8,$FC
	.dc.b	$A8,$D5,$72,$6D,$72,$7F,$C7,$D3,$C8,$F0,$46,$4F,$ED,$73,$C6,$76
	.dc.b	$D8,$C9,$B8,$FE,$B8,$7C,$63,$B9,$4E,$2E,$EB,$F0,$B1,$F4,$67,$DC
	.dc.b	$EE,$7A,$B4,$EA,$6C,$EA,$53,$A9,$B9,$EA,$FD,$33,$65,$36,$71,$8E
	.dc.b	$ED,$3B,$9B,$72,$D3,$31,$F9,$57,$AB,$5E,$A6,$DF,$19,$98,$E6,$D8
	.dc.b	$E6,$FE,$11,$E6,$D7,$9B,$B3,$99,$98,$E6,$D8,$E7,$5F,$83,$78,$C6
	.dc.b	$4E,$31,$99,$E2,$D9,$E3,$5F,$A7,$3F,$6C,$59,$FD,$5D,$7E,$CA,$BF
	.dc.b	$42,$F9,$4F,$D4,$57,$CB,$7B,$D7,$18,$FE,$15,$3D,$E7,$27,$39,$EE
	.dc.b	$DA,$EE,$FE,$EE,$B8,$CF,$38,$FD,$0B,$5F,$E0,$C6,$CC,$9C,$E3,$8B
	.dc.b	$4C,$9B,$38,$CF,$F1,$63,$C2,$58,$E3,$1B,$3F,$87,$C6,$5E,$6D,$79
	.dc.b	$C7,$F0,$79,$F6,$C7,$13,$F8,$59,$FC,$18,$D9,$4E,$B1,$C6,$7C,$6D
	.dc.b	$3C,$6F,$78,$D9,$C4,$EA,$7F,$07,$3A,$94,$D9,$4D,$D7,$F0,$49,$FD
	.dc.b	$64,$7D,$C7,$F3,$8D,$3F,$01,$1F,$82,$7F,$1C,$64,$AE,$4F,$C0,$39
	.dc.b	$69,$9A,$FB,$51,$E4,$8F,$14,$FE,$2E,$38,$4E,$9F,$3B,$FC,$2A,$35
	.dc.b	$3C,$E7,$EF,$1F,$01,$E7,$1E,$A7,$51,$F4,$DF,$E4,$93,$CE,$74,$FD
	.dc.b	$8A,$F9,$D5,$E0,$9F,$DB,$D7,$AA,$B5,$1F,$56,$7F,$06,$FF,$93,$CF
	.dc.b	$F1,$81,$FE,$21,$3F,$CC,$8F,$F8,$DA,$BF,$C9,$31,$A8,$FF,$30,$E1
	.dc.b	$FE,$2B,$9C,$CF,$F8,$B2,$3F,$C7,$EF,$29,$FF,$15,$4E,$8F,$F1,$A4
	.dc.b	$77,$3F,$CA,$B1,$FE,$38,$7F,$C6,$6F,$68,$FF,$19,$E1,$FE,$5D,$9F
	.dc.b	$41,$DE,$3F,$C7,$A6,$AB,$FC,$B9,$1E,$7A,$FF,$1E,$BD,$9F,$33,$97
	.dc.b	$FC,$D5,$3C,$5F,$11,$88,$F3,$1F,$E6,$27,$8B,$E1,$32,$FC,$77,$CB
	.dc.b	$1E,$E8,$F5,$47,$C3,$AF,$54,$7F,$98,$23,$93,$A7,$CD,$3F,$73,$0F
	.dc.b	$C0,$C7,$DC,$C3,$F1,$D1,$88,$FE,$56,$FD,$F9,$FD,$9D,$7E,$BE,$35
	.dc.b	$1F,$CB,$2B,$89,$FD,$A6,$BF,$96,$D7,$E5,$0F,$CF,$47,$E3,$DF,$D5
	.dc.b	$BF,$8A,$7A,$3C,$5F,$E4,$E7,$E5,$DC,$3F,$A4,$3A,$1F,$95,$8F,$3C
	.dc.b	$78,$A7,$FB,$94,$EC,$E4,$E2,$BF,$46,$FC,$39,$EB,$18,$AF,$9E,$7E
	.dc.b	$22,$BE,$FB,$F5,$EB,$F2,$0E,$63,$EE,$3F,$38,$E2,$FD,$F3,$E9,$3C
	.dc.b	$4C,$BF,$E8,$F3,$27,$DF,$3E,$D4,$7D,$F7,$C2,$64,$FF,$60,$47,$E3
	.dc.b	$DF,$F4,$C3,$E1,$78,$BE,$57,$2F,$85,$C9,$E5,$3A,$9F,$AC,$7C,$B7
	.dc.b	$FE,$D5,$8C,$BF,$8E,$7C,$31,$F8,$E7,$38,$7E,$9A,$31,$1F,$11,$F0
	.dc.b	$3C,$2B,$D8,$E4,$F6,$5F,$F1,$87,$F0,$B1,$F9,$07,$F9,$C4,$7E,$42
	.dc.b	$7F,$1B,$1F,$71,$FE,$73,$1F,$66,$3F,$17,$3F,$7F,$0F,$C7,$47,$27
	.dc.b	$EF,$98,$9F,$04,$7F,$3A,$8C,$1F,$97,$71,$3E,$39,$D9,$FA,$5A,$FE
	.dc.b	$AA,$7F,$23,$9F,$E6,$55,$C6,$75,$1F,$C9,$8F,$2C,$FA,$27,$BB,$FB
	.dc.b	$38,$FC,$EC,$FB,$DF,$D1,$D7,$E5,$CF,$61,$E3,$AE,$C7,$E8,$6B,$A4
	.dc.b	$65,$E9,$3F,$1D,$FE,$93,$3E,$07,$F5,$31,$E0,$8F,$8B,$3C,$67,$33
	.dc.b	$C5,$FE,$8F,$19,$3F,$81,$47,$17,$CB,$1F,$12,$3D,$EF,$82,$38,$3F
	.dc.b	$69,$F6,$3E,$83,$84,$7D,$78,$E6,$FC,$93,$11,$B7,$15,$F5,$2B,$E1
	.dc.b	$4F,$73,$9C,$66,$7C,$D1,$C5,$FA,$EF,$C6,$AF,$1B,$C6,$3A,$D7,$CE
	.dc.b	$78,$C6,$5E,$2F,$C6,$9E,$2F,$C6,$3C,$33,$C6,$7E,$51,$F9,$E8,$C9
	.dc.b	$E1,$3F,$14,$FF,$28,$70,$FC,$A7,$A3,$E4,$72,$FC,$47,$C8,$F1,$7C
	.dc.b	$87,$47,$F9,$44,$FD,$87,$C6,$7E,$7A,$3C,$67,$07,$C0,$79,$CF,$03
	.dc.b	$97,$C0,$F9,$1F,$87,$1D,$4F,$86,$F1,$33,$1F,$81,$3C,$95,$F9,$F8
	.dc.b	$FE,$71,$3C,$1F,$21,$E0,$7A,$4F,$0A,$FC,$BB,$E8,$8C,$CF,$A0,$E2
	.dc.b	$78,$0F,$D5,$C7,$18,$FC,$6C,$64,$F0,$C6,$4F,$C3,$3F,$2E,$3A,$BF
	.dc.b	$00,$F6,$19,$8F,$63,$E1,$AF,$E0,$2F,$57,$F8,$B3,$C4,$F2,$9C,$4F
	.dc.b	$29,$C6,$BF,$37,$3F,$36,$34,$79,$8D,$1F,$94,$74,$FD,$E7,$C6,$F7
	.dc.b	$9F,$1B,$C9,$F1,$9F,$A5,$7F,$06,$E9,$E4,$79,$0F,$E7,$51,$FD,$37
	.dc.b	$06,$1F,$E3,$98,$FE,$92,$FF,$8E,$1F,$EA,$93,$C0,$FE,$A0,$7F,$41
	.dc.b	$3F,$A8,$1F,$D5,$0F,$EC,$87,$F4,$C3,$FC,$52,$60,$E1,$3F,$D4,$CF
	.dc.b	$0C,$7F,$63,$7C,$78,$7F,$8C,$67,$FC,$5A,$FF,$50,$7F,$B1,$9F,$DA
	.dc.b	$E3,$8C,$FF,$58,$8C,$9E,$43,$D0,$78,$4C,$C7,$86,$93,$E1,$7E,$14
	.dc.b	$FA,$E7,$FB,$53,$E8,$8F,$EC,$D5,$E3,$9F,$3D,$7F,$8E,$E7,$EF,$D7
	.dc.b	$E3,$0F,$A2,$7D,$F3,$ED,$61,$A7,$F1,$26,$B0,$FB,$71,$FC,$3F,$0F
	.dc.b	$D5,$47,$DB,$3F,$5C,$6A,$FF,$84,$BF,$81,$8D,$CF,$58,$E0,$FE,$BA
	.dc.b	$BF,$65,$5F,$8C,$9E,$0F,$E9,$CF,$E2,$31,$E3,$9F,$E9,$D5,$FA,$8A
	.dc.b	$DC,$FE,$60,$FC,$7C,$F7,$23,$F9,$B3,$E1,$47,$A2,$39,$D7,$85,$F7
	.dc.b	$3E,$13,$F4,$31,$BA,$FE,$9F,$5F,$99,$8F,$0B,$F2,$8E,$C7,$CA,$78
	.dc.b	$4F,$F0,$B3,$B1,$F1,$AB,$BC,$FA,$6B,$E3,$BC,$AB,$E3,$C7,$F0,$C8
	.dc.b	$FD,$74,$FA,$1F,$13,$C9,$F1,$1F,$08,$E4,$FC,$27,$04,$8E,$ED,$AF
	.dc.b	$80,$FC,$69,$FA,$CF,$F0,$F7,$C0,$FE,$26,$BE,$09,$23,$C4,$D9,$F7
	.dc.b	$10,$FE,$7C,$E5,$B1,$C0,$8E,$EB,$EF,$46,$23,$66,$4F,$96,$60,$91
	.dc.b	$C1,$D5,$7C,$66,$BE,$99,$E9,$3B,$9E,$27,$47,$04,$70,$49,$E8,$F0
	.dc.b	$6C,$6D,$B1,$B7,$C0,$F8,$AB,$F5,$0F,$18,$CB,$63,$33,$CE,$72,$D8
	.dc.b	$CC,$F5,$6C,$F8,$08,$70,$8D,$57,$DA,$9D,$CE,$4D,$B8,$24,$6C,$95
	.dc.b	$FA,$29,$E0,$7F,$0B,$C3,$F9,$0C,$7F,$13,$C3,$F8,$CC,$FF,$20,$9F
	.dc.b	$E3,$2F,$F2,$08,$FE,$33,$87,$F4,$17,$73,$CA,$31,$5F,$CC,$4F,$F5
	.dc.b	$3D,$7F,$40,$39,$1F,$CF,$27,$A9,$FC,$8A,$7F,$A3,$BC,$E3,$83,$98
	.dc.b	$FF,$45,$B9,$8F,$E6,$8E,$5F,$E7,$95,$FC,$D2,$7F,$9E,$4E,$67,$F9
	.dc.b	$E3,$C4,$F0,$CF,$A8,$E1,$5F,$0C,$F5,$57,$F5,$C7,$F9,$FD,$7F,$AE
	.dc.b	$0F,$E8,$D8,$7E,$96,$3E,$F6,$1F,$42,$BF,$85,$57,$D6,$8C,$3F,$CA
	.dc.b	$8C,$4F,$D5,$AF,$B2,$D9,$FF,$4C,$E1,$E7,$C3,$FB,$66,$1F,$09,$F3
	.dc.b	$BF,$40,$FE,$95,$87,$CD,$8D,$BF,$12,$3F,$AB,$5F,$B2,$3F,$AB,$4E
	.dc.b	$9F,$8A,$F4,$7E,$8B,$F1,$67,$89,$DA,$3F,$B6,$BE,$13,$EC,$C7,$18
	.dc.b	$FA,$AF,$18,$FA,$6D,$9F,$9D,$1C,$DF,$C3,$CF,$07,$CC,$FE,$4E,$7C
	.dc.b	$D5,$FD,$C6,$BC,$67,$80,$A7,$C5,$2C,$FC,$17,$C3,$5D,$27,$E6,$3F
	.dc.b	$72,$7F,$C2,$07,$B0,$F8,$2E,$EB,$D9,$3F,$88,$8C,$57,$6A,$F9,$2D
	.dc.b	$9F,$33,$E0,$3D,$C7,$E2,$C4,$7E,$2E,$3F,$DD,$D8,$7E,$2E,$7F,$80
	.dc.b	$D7,$E0,$C8,$FE,$32,$BF,$77,$3F,$88,$8C,$36,$3F,$1C,$6B,$0F,$C8
	.dc.b	$BF,$D1,$6B,$9B,$A9,$E0,$E6,$B9,$D7,$08,$FE,$B7,$5C,$1D,$1C,$63
	.dc.b	$F2,$F8,$7C,$63,$EB,$9F,$C2,$23,$E5,$D7,$C2,$7F,$23,$5F,$45,$FF
	.dc.b	$13,$1F,$99,$70,$FF,$08,$3E,$A1,$F8,$58,$E4,$FE,$16,$37,$87,$EF
	.dc.b	$CF,$A2,$FE,$3A,$7F,$0C,$7D,$32,$B9,$7F,$10,$64,$AE,$67,$F1,$4D
	.dc.b	$7E,$FC,$6D,$CB,$FA,$33,$89,$5C,$9F,$09,$F0,$99,$79,$C7,$59,$F4
	.dc.b	$CE,$5B,$3C,$63,$AB,$4E,$A7,$C5,$7A,$B9,$3A,$BC,$EB,$EA,$46,$E3
	.dc.b	$9B,$4E,$73,$F9,$F8,$F9,$EF,$EA,$1F,$CC,$C7,$EA,$2B,$F7,$33,$B9
	.dc.b	$F0,$CF,$33,$C2,$78,$E3,$D6,$E5,$A7,$18,$F1,$B6,$3C,$75,$F8,$C7
	.dc.b	$F0,$D3,$E2,$7F,$12,$7E,$1A,$78,$3E,$03,$B4,$FC,$08,$C9,$E0,$7A
	.dc.b	$47,$16,$D7,$B1,$B5,$FB,$99,$FD,$C4,$73,$32,$FE,$72,$3C,$8F,$E6
	.dc.b	$67,$91,$F9,$69,$FE,$45,$3F,$4E,$7D,$67,$43,$FC,$86,$43,$FC,$80
	.dc.b	$7F,$94,$5F,$F2,$0B,$FE,$59,$BF,$F2,$0B,$FB,$68,$FD,$CD,$7E,$DE
	.dc.b	$30,$7E,$DD,$FD,$6E,$0A,$47,$F6,$9A,$FB,$B3,$9A,$F2,$9F,$7E,$3E
	.dc.b	$13,$4F,$80,$7A,$4F,$95,$3E,$F3,$F9,$3C,$7E,$9E,$3E,$A6,$1F,$4B
	.dc.b	$0F,$95,$5B,$AF,$C7,$C7,$F0,$B8,$C4,$7E,$32,$78,$57,$90,$FC,$0C
	.dc.b	$FC,$E9,$FA,$33,$FC,$0E,$BA,$C6,$1F,$F0,$FD,$FF,$2F,$BF,$EE,$73
	.dc.b	$F2,$08,$FC,$76,$BF,$C8,$08,$7C,$C2,$9F,$EA,$1B,$FA,$0D,$8E,$2E
	.dc.b	$C8,$64,$EC,$43,$9C,$FF,$A5,$1C,$1C,$1C,$3A,$C3,$F0,$2F,$33,$15
	.dc.b	$FE,$93,$70,$71,$8D,$CF,$28,$C4,$EE,$31,$1C,$08,$E0,$93,$C8,$93
	.dc.b	$C5,$B5,$FE,$33,$AF,$F4,$B4,$E4,$91,$98,$FC,$D4,$62,$3F,$D9,$75
	.dc.b	$DA,$31,$1F,$EB,$EA,$EA,$E9,$AE,$63,$75,$E0,$8E,$D5,$FE,$B6,$6D
	.dc.b	$7F,$AB,$EB,$FD,$5A,$48,$FD,$6C,$ED,$ED,$1D,$2B,$CF,$1C,$9F,$DB
	.dc.b	$47,$17,$F6,$C4,$3F,$D7,$EE,$5B,$1D,$AB,$C4,$D8,$E6,$D7,$F8,$09
	.dc.b	$C0,$E9,$5F,$99,$7F,$BF,$57,$E4,$23,$E4,$4F,$96,$78,$C6,$CF,$EF
	.dc.b	$EF,$E0,$23,$53,$E5,$9D,$57,$F8,$06,$BF,$02,$FC,$E7,$E6,$57,$C4
	.dc.b	$38,$9E,$7A,$E5,$3F,$EB,$88,$FF,$0D,$9F,$A2,$9E,$73,$F2,$EB,$E7
	.dc.b	$4F,$D9,$39,$CF,$BA,$B9,$3F,$B6,$3F,$46,$72,$8D,$BF,$14,$93,$EA
	.dc.b	$AF,$89,$5E,$99,$FB,$35,$E9,$21,$E1,$2B,$EE,$70,$47,$FC,$73,$5F
	.dc.b	$C9,$5F,$E3,$F1,$FC,$36,$BF,$8F,$9F,$DE,$EF,$F8,$8C,$FF,$25,$8F
	.dc.b	$E2,$CD,$9F,$E5,$6F,$F1,$A6,$BF,$D2,$5B,$87,$F1,$B8,$FE,$40,$64
	.dc.b	$FE,$62,$7F,$33,$3F,$C4,$91,$E3,$3F,$95,$9F,$E5,$5A,$F1,$C6,$EF
	.dc.b	$E3,$4F,$D5,$7F,$12,$FC,$28,$F8,$EF,$C1,$69,$F1,$48,$60,$91,$F7
	.dc.b	$1C,$B5,$CE,$1F,$A0,$7E,$F4,$FF,$BB,$E3,$F2,$13,$E2,$8F,$03,$C4
	.dc.b	$91,$C6,$7E,$FB,$F8,$67,$F4,$73,$EB,$8F,$C0,$4F,$09,$FD,$69,$E0
	.dc.b	$9F,$F7,$D4,$F8,$0D,$9D,$67,$E2,$CF,$E4,$1F,$23,$EC,$8D,$1E,$18
	.dc.b	$FD,$6C,$7E,$56,$7E,$54,$78,$63,$F6,$73,$B3,$C1,$87,$E8,$DE,$75
	.dc.b	$F0,$E3,$F3,$26,$67,$D2,$7C,$08,$D1,$EF,$8D,$47,$68,$CB,$E9,$9F
	.dc.b	$6B,$CA,$BD,$AF,$F4,$89,$FD,$5C,$62,$3F,$83,$47,$AA,$7C,$AF,$C8
	.dc.b	$AF,$2B,$C2,$3D,$91,$88,$E1,$3D,$E7,$31,$E7,$38,$1E,$68,$F1,$3C
	.dc.b	$1F,$53,$DC,$E0,$66,$7F,$87,$BC,$63,$CE,$FE,$22,$7C,$F3,$FE,$F6
	.dc.b	$3C,$91,$99,$F2,$3C,$5F,$B0,$E5,$F0,$C7,$18,$F0,$99,$7C,$2F,$96
	.dc.b	$3E,$74,$7D,$23,$27,$17,$31,$F0,$0F,$F1,$89,$FE,$30,$9F,$F5,$5B
	.dc.b	$FB,$78,$FD,$FE,$1F,$C0,$5F,$E1,$E7,$F0,$93,$F8,$09,$FC,$24,$A7
	.dc.b	$F0,$53,$F8,$19,$B2,$9F,$C1,$DF,$F3,$E3,$96,$B9,$C3,$FC,$65,$3F
	.dc.b	$C2,$CF,$F1,$93,$FE,$60,$8E,$67,$84,$FC,$A4,$7C,$37,$ED,$BF,$1C
	.dc.b	$FE,$8C,$47,$F6,$57,$F1,$5D,$9F,$AD,$3E,$AC,$73,$7E,$E9,$93,$EA
	.dc.b	$BF,$D3,$9F,$9A,$FE,$14,$EC,$FF,$45,$AF,$6B,$E7,$6B,$F1,$DA,$7C
	.dc.b	$F9,$FE,$6F,$1F,$0E,$7E,$A3,$F4,$8C,$3F,$E5,$D9,$FF,$46,$4F,$F9
	.dc.b	$62,$3F,$D3,$75,$F1,$E7,$AB,$FE,$82,$AF,$F5,$23,$FC,$D1,$F9,$58
	.dc.b	$7E,$C5,$FC,$99,$EF,$BF,$C5,$DF,$BE,$3E,$39,$FC,$B0,$F8,$F1,$F2
	.dc.b	$DF,$8E,$64,$F8,$E7,$F6,$B3,$8C,$FF,$15,$8E,$B1,$93,$8C,$66,$7E
	.dc.b	$5B,$E5,$3F,$91,$C7,$59,$E2,$FF,$32,$3E,$9C,$FE,$05,$FB,$EF,$A5
	.dc.b	$EB,$87,$ED,$A3,$11,$F0,$0E,$8F,$82,$3F,$84,$C7,$43,$E2,$46,$4F
	.dc.b	$DD,$57,$90,$F4,$47,$80,$F6,$1D,$5C,$BF,$98,$AF,$90,$E6,$BC,$8F
	.dc.b	$09,$E9,$18,$8F,$71,$F2,$9C,$47,$F4,$23,$71,$F5,$2B,$DB,$18,$8F
	.dc.b	$A6,$72,$3E,$19,$B3,$C1,$3F,$CE,$1D,$C7,$58,$C5,$7D,$08,$C3,$FC
	.dc.b	$46,$BE,$21,$FA,$9A,$F6,$1D,$4F,$01,$83,$AD,$7F,$54,$8F,$E2,$D3
	.dc.b	$F1,$2B,$D0,$F7,$9D,$4F,$EB,$EB,$E7,$1D,$2B,$71,$E2,$AE,$55,$E8
	.dc.b	$8C,$1F,$11,$C5,$72,$AF,$5C,$F8,$8D,$57,$B6,$75,$5F,$CA,$27,$F6
	.dc.b	$57,$FA,$B9,$FD,$E3,$FC,$69,$FE,$3C,$D7,$F7,$35,$FE,$37,$7F,$97
	.dc.b	$C7,$ED,$56,$1F,$ED,$07,$F8,$01,$FC,$26,$39,$36,$BF,$82,$D7,$F0
	.dc.b	$88,$C4,$79,$1F,$A1,$87,$B2,$3E,$44,$7E,$45,$F6,$9E,$F7,$F0,$33
	.dc.b	$EF,$74,$62,$34,$FC,$FC,$3E,$4B,$A3,$13,$C6,$BF,$71,$3A,$6C,$66
	.dc.b	$3F,$5B,$3F,$9B,$3C,$EF,$DE,$AF,$DD,$47,$5A,$F5,$46,$1E,$6F,$91
	.dc.b	$CC,$7E,$D8,$FA,$F5,$FD,$C9,$C1,$CE,$BB,$38,$9E,$75,$F1,$9A,$F7
	.dc.b	$9F,$3C,$FE,$A0,$FC,$CD,$75,$7F,$84,$9E,$09,$FE,$F2,$F5,$9E,$2F
	.dc.b	$E8,$67,$F3,$87,$1A,$FE,$0A,$78,$23,$85,$7F,$0E,$AF,$94,$72,$9F
	.dc.b	$CF,$CF,$73,$F8,$B1,$E2,$3D,$D5,$D6,$7F,$83,$9F,$02,$35,$3D,$DB
	.dc.b	$1D,$4E,$F5,$E1,$AF,$B0,$D3,$E8,$47,$3A,$F4,$BC,$67,$73,$FC,$0E
	.dc.b	$7C,$F5,$E4,$8F,$8A,$F8,$1E,$87,$D1,$AE,$4F,$6A,$F5,$4F,$E1,$9A
	.dc.b	$FD,$28,$E7,$5F,$CB,$B0,$FE,$C3,$5F,$EA,$69,$57,$F6,$23,$F6,$31
	.dc.b	$97,$07,$EC,$0F,$DD,$C7,$EB,$8F,$F5,$2E,$1F,$B7,$AF,$DB,$57,$EC
	.dc.b	$6B,$75,$FB,$17,$94,$FF,$7D,$8E,$32,$61,$E4,$9F,$E0,$35,$F5,$A7
	.dc.b	$DA,$7E,$1A,$7D,$B8,$65,$B5,$F0,$DA,$F8,$2B,$CC,$7F,$80,$CF,$E1
	.dc.b	$2F,$9A,$BF,$89,$3B,$7E,$13,$E2,$7F,$C1,$8F,$C8,$BF,$F0,$7D,$7F
	.dc.b	$8D,$4F,$F0,$7C,$EA,$7D,$27,$CC,$8F,$F6,$19,$F0,$E3,$E3,$4F,$C3
	.dc.b	$9F,$F2,$4B,$5F,$F1,$E9,$E5,$9F,$89,$1F,$39,$F9,$27,$49,$F4,$9D
	.dc.b	$2B,$DE,$F8,$E7,$CF,$5F,$CC,$AB,$ED,$3E,$0A,$DC,$7A,$E3,$DD,$1F
	.dc.b	$CE,$E7,$C8,$7E,$2A,$7C,$07,$92,$3E,$63,$D2,$7D,$33,$C2,$7E,$7B
	.dc.b	$67,$4F,$37,$FC,$BB,$5E,$79,$F0,$1E,$4A,$F4,$4F,$C9,$3E,$09,$C0
	.dc.b	$F4,$3F,$40,$FB,$4F,$99,$FC,$DC,$FC,$53,$EE,$D7,$CB,$8F,$85,$18
	.dc.b	$AF,$5D,$77,$9C,$BC,$DF,$C2,$3E,$0A,$F9,$B1,$FD,$72,$3E,$25,$7A
	.dc.b	$E3,$F0,$51,$E7,$7D,$71,$F6,$23,$D6,$7E,$28,$E9,$5F,$0A,$3C,$87
	.dc.b	$39,$F6,$CF,$37,$C0,$7B,$9E,$6F,$DA,$3C,$2F,$57,$C2,$F9,$8E,$EE
	.dc.b	$8E,$07,$67,$83,$D8,$F8,$27,$23,$C0,$7C,$67,$C8,$FF,$72,$8F,$73
	.dc.b	$98,$E6,$E5,$F1,$3F,$EA,$7C,$3F,$99,$1F,$CA,$A3,$F9,$19,$FD,$44
	.dc.b	$B1,$FD,$11,$FE,$59,$87,$F7,$2C,$3F,$19,$1D,$A7,$F5,$47,$98,$FC
	.dc.b	$AE,$1F,$8D,$7D,$93,$F8,$37,$38,$7D,$0A,$F8,$B1,$F0,$CF,$8B,$1F
	.dc.b	$8B,$8F,$C2,$4F,$13,$C3,$3A,$8F,$F1,$7B,$FE,$09,$7F,$9E,$1E,$A7
	.dc.b	$FA,$03,$EA,$9F,$E8,$E7,$F8,$E0,$AF,$F8,$58,$E2,$6C,$FE,$88,$7F
	.dc.b	$20,$7F,$93,$47,$F6,$2C,$3E,$54,$EC,$E3,$3D,$5E,$33,$BC,$3F,$96
	.dc.b	$34,$FE,$58,$53,$FA,$71,$67,$BB,$FE,$19,$32,$57,$2F,$F8,$1D,$AF
	.dc.b	$F8,$06,$7E,$2B,$FE,$28,$7C,$6F,$17,$35,$FE,$1C,$8F,$8F,$3E,$99
	.dc.b	$F5,$9F,$14,$FC,$31,$EF,$2B,$B6,$9B,$8F,$C5,$46,$5B,$19,$7B,$B6
	.dc.b	$7E,$5C,$EE,$72,$7D,$D9,$EA,$D3,$99,$4D,$9E,$F7,$D2,$E6,$7A,$C7
	.dc.b	$7A,$FA,$73,$F0,$CF,$F0,$F1,$4F,$F0,$51,$F4,$E3,$E3,$C7,$F8,$4C
	.dc.b	$E3,$1F,$46,$33,$5F,$DB,$1A,$7E,$24,$B1,$CC,$EE,$71,$3F,$74,$7E
	.dc.b	$74,$A7,$F3,$18,$FD,$24,$7F,$02,$8F,$D8,$9F,$C2,$0F,$D7,$E1,$FC
	.dc.b	$BA,$7F,$82,$C7,$EC,$27,$F5,$73,$97,$F8,$2C,$70,$9F,$E4,$47,$68
	.dc.b	$FE,$A5,$1F,$96,$3E,$25,$FF,$0A,$BF,$83,$1F,$8F,$3D,$65,$3E,$49
	.dc.b	$C1,$FF,$17,$CF,$D4,$8F,$AC,$FD,$C6,$9F,$70,$A7,$DC,$9F,$CB,$1F
	.dc.b	$69,$FB,$93,$E2,$8F,$F2,$09,$E5,$8F,$41,$F3,$4F,$19,$5D,$3F,$AC
	.dc.b	$31,$1E,$69,$F2,$4F,$82,$3B,$1F,$E5,$88,$FD,$14,$F9,$E7,$4F,$80
	.dc.b	$F8,$4F,$6A,$FD,$AB,$E6,$AE,$D3,$C5,$FC,$63,$C6,$35,$5F,$C6,$DF
	.dc.b	$81,$1F,$11,$A7,$81,$F5,$1E,$09,$EA,$FC,$97,$AB,$A8,$EB,$3E,$A8
	.dc.b	$D5,$7E,$65,$FB,$73,$A8,$FD,$A3,$E0,$AE,$71,$EC,$7E,$23,$A7,$E3
	.dc.b	$BF,$CB,$A7,$EB,$9E,$D8,$DD,$7E,$CE,$31,$3B,$7D,$D3,$CA,$7F,$8C
	.dc.b	$C7,$DB,$C3,$FA,$1B,$FA,$58,$FE,$EA,$D3,$F5,$31,$FE,$0A,$7F,$B2
	.dc.b	$1F,$D5,$DE,$26,$5F,$D5,$BF,$D3,$F0,$FF,$16,$9F,$D6,$E3,$89,$FA
	.dc.b	$F2,$C7,$F5,$DC,$3E,$F5,$7E,$FA,$7F,$40,$FD,$9B,$F5,$3F,$A1,$9F
	.dc.b	$B1,$87,$C9,$3E,$03,$88,$E5,$5F,$94,$79,$3F,$19,$FF,$3E,$BF,$E9
	.dc.b	$08,$57,$F8,$76,$7F,$7F,$5F,$BC,$7E,$2D,$FB,$CF,$F2,$43,$FE,$33
	.dc.b	$BF,$E0,$CF,$F0,$D8,$CC,$FF,$0A,$7E,$3E,$1D,$67,$4F,$C5,$AF,$F1
	.dc.b	$CC,$7F,$A4,$5F,$E3,$47,$EC,$0F,$CA,$3E,$A7,$F2,$57,$F8,$27,$F0
	.dc.b	$A6,$2F,$E1,$C7,$E3,$A7,$E1,$CF,$27,$F3,$18,$7E,$4E,$BF,$31,$3D
	.dc.b	$E3,$F3,$13,$F8,$D8,$E8,$FE,$29,$E5,$3F,$97,$7E,$64,$7D,$5A,$F0
	.dc.b	$47,$E8,$9F,$33,$F9,$43,$71,$FC,$1A,$37,$1E,$17,$75,$F6,$E3,$95
	.dc.b	$7B,$0F,$3C,$7E,$22,$33,$5E,$58,$F4,$C7,$C4,$9F,$C3,$4E,$6B,$A9
	.dc.b	$E0,$AE,$55,$E1,$AE,$D5,$FD,$46,$3A,$BE,$D8,$F0,$D7,$F0,$6A,$F9
	.dc.b	$53,$BA,$E4,$F7,$AE,$71,$FD,$64,$CC,$73,$38,$4F,$88,$95,$D4,$FD
	.dc.b	$0C,$7C,$7A,$F0,$BF,$83,$7D,$AF,$3A,$ED,$5C,$A3,$47,$E0,$9F,$8B
	.dc.b	$1F,$4C,$FB,$B1,$C5,$FF,$70,$47,$E9,$27,$FD,$C7,$38,$AF,$CD,$3F
	.dc.b	$E5,$BC,$3F,$3E,$FF,$64,$BF,$DE,$E1,$FD,$9D,$FE,$10,$FF,$04,$BF
	.dc.b	$F2,$D4,$7F,$9D,$CF,$E0,$87,$F0,$A8,$E0,$E2,$BA,$BC,$AB,$FD,$79
	.dc.b	$87,$B5,$FC,$DE,$1F,$1A,$7F,$A2,$61,$F1,$A7,$31,$F9,$03,$2F,$D0
	.dc.b	$9F,$9D,$19,$3F,$9D,$C6,$4F,$92,$FE,$09,$FA,$46,$4E,$AF,$CA,$72
	.dc.b	$FD,$28,$EA,$E4,$F0,$99,$8E,$AF,$E1,$1E,$2F,$F2,$27,$AC,$FD,$E7
	.dc.b	$31,$FC,$75,$C9,$F7,$67,$F3,$B3,$F0,$4F,$DB,$4F,$84,$F6,$9E,$19
	.dc.b	$E2,$FB,$4F,$29,$E1,$8E,$27,$85,$C9,$E1,$73,$1E,$17,$F3,$F1,$C6
	.dc.b	$3F,$85,$47,$18,$F0,$99,$3F,$6D,$1C,$4F,$D4,$10,$FC,$A3,$DD,$E6
	.dc.b	$6C,$CB,$FC,$51,$CB,$EB,$3B,$9C,$4E,$E6,$6B,$F7,$2F,$C3,$9F,$5B
	.dc.b	$FD,$8B,$0F,$CF,$BF,$CD,$30,$FE,$76,$FF,$B5,$4F,$ED,$07,$F4,$2B
	.dc.b	$FD,$24,$FF,$3D,$9F,$CC,$CF,$E6,$E3,$04,$AF,$D1,$C7,$F9,$2E,$BF
	.dc.b	$D7,$B1,$F1,$EF,$CB,$3F,$1C,$FD,$C4,$67,$0F,$D0,$42,$7F,$AD,$A1
	.dc.b	$3F,$D3,$5D,$4A,$A4,$E6,$BF,$80,$47,$1A,$FE,$01,$3F,$C0,$9F,$ED
	.dc.b	$15,$F1,$1F,$83,$87,$C4,$7F,$DA,$86,$30,$FE,$01,$7F,$DB,$0F,$69
	.dc.b	$FD,$B1,$E5,$3F,$DB,$5F,$89,$1B,$9F,$E0,$D3,$FE,$0D,$8F,$39,$FC
	.dc.b	$3E,$3F,$84,$5F,$F7,$77,$EC,$1F,$56,$7A,$47,$CE,$8F,$E2,$AE,$DF
	.dc.b	$B0,$7D,$DA,$F3,$D7,$F7,$69,$E1,$3D,$A3,$D2,$F6,$3D,$F5,$DA,$3D
	.dc.b	$11,$F6,$CF,$4D,$7F,$1C,$9F,$54,$FF,$82,$5F,$91,$5F,$E8,$9A,$F1
	.dc.b	$4F,$B9,$FE,$40,$EE,$BC,$AF,$E3,$A7,$CB,$5F,$8E,$3A,$BF,$E1,$19
	.dc.b	$DD,$74,$AE,$47,$57,$F2,$93,$CA,$7E,$8C,$FC,$13,$D7,$3F,$21,$F8
	.dc.b	$2F,$C8,$3E,$F3,$CE,$BC,$2F,$D5,$7C,$93,$F0,$49,$5F,$46,$3E,$0C
	.dc.b	$FE,$0E,$BE,$C5,$7E,$10,$8E,$08,$7D,$07,$2D,$33,$5F,$2D,$F2,$CF
	.dc.b	$C0,$24,$7C,$57,$D1,$3E,$47,$DB,$5E,$4A,$FB,$D5,$CE,$3E,$F4,$FC
	.dc.b	$52,$1F,$4C,$F8,$F3,$F8,$73,$F0,$33,$EF,$AF,$A7,$1E,$48,$F1,$4F
	.dc.b	$90,$F9,$13,$F4,$9E,$87,$69,$FB,$53,$C0,$91,$C1,$FB,$44,$7C,$75
	.dc.b	$F4,$A3,$C5,$1F,$01,$AF,$86,$7F,$1F,$1D,$1B,$1D,$4E,$75,$E5,$8F
	.dc.b	$1C,$64,$F3,$3E,$D7,$07,$A8,$8F,$03,$A9,$0F,$C9,$BD,$A7,$D4,$F4
	.dc.b	$AF,$A2,$F2,$9E,$C7,$18,$FC,$8B,$6B,$EE,$3F,$4A,$7E,$EC,$7D,$29
	.dc.b	$E7,$3E,$33,$C3,$3D,$E7,$F1,$B5,$F8,$16,$C7,$E5,$E7,$F2,$04,$9D
	.dc.b	$3F,$92,$8D,$46,$23,$51,$82,$3F,$8E,$6B,$FA,$DA,$F2,$BE,$E3,$CB
	.dc.b	$1F,$06,$32,$4A,$F2,$CF,$32,$1C,$DC,$B4,$F3,$1A,$8F,$D0,$39,$3C
	.dc.b	$E6,$63,$E4,$1C,$A3,$64,$73,$1D,$0E,$55,$F0,$49,$5E,$EA,$E0,$FE
	.dc.b	$4C,$FA,$A5,$9F,$B0,$43,$F9,$61,$0F,$E2,$86,$4A,$64,$FD,$91,$63
	.dc.b	$F4,$51,$D8,$FD,$04,$7C,$C3,$F0,$51,$F8,$A7,$07,$D1,$70,$4A,$FC
	.dc.b	$6C,$75,$7F,$48,$ED,$CB,$E4,$21,$E4,$2C,$79,$1F,$51,$92,$39,$7B
	.dc.b	$57,$E9,$2B,$F0,$55,$F4,$CF,$A3,$3F,$60,$FA,$31,$F3,$E3,$33,$F4
	.dc.b	$A7,$EE,$4F,$CF,$7A,$BF,$B1,$21,$FA,$37,$EE,$57,$D8,$8F,$92,$FC
	.dc.b	$39,$FE,$16,$7F,$09,$9F,$93,$1F,$95,$7F,$CD,$D8,$7F,$87,$1F,$F1
	.dc.b	$A1,$FE,$1C,$AF,$F1,$E1,$FE,$5E,$7F,$C7,$87,$F8,$E0,$FF,$33,$1F
	.dc.b	$E5,$23,$27,$13,$FC,$79,$87,$F8,$92,$7B,$46,$9F,$F6,$63,$C8,$FF
	.dc.b	$22,$99,$8F,$F1,$3C,$FF,$88,$63,$FC,$3F,$7F,$E2,$17,$FC,$52,$7C
	.dc.b	$2C,$3E,$5C,$FC,$D3,$E5,$47,$DC,$8F,$C1,$E1,$F0,$63,$EA,$3F,$26
	.dc.b	$BE,$33,$F2,$6B,$FB,$6E,$1F,$E9,$D9,$FF,$29,$D7,$9E,$3D,$67,$FA
	.dc.b	$E8,$F5,$9B,$9F,$F3,$0C,$6F,$0F,$84,$47,$E8,$B5,$FF,$42,$4F,$48
	.dc.b	$FF,$44,$CF,$FA,$76,$BC,$33,$D6,$BE,$93,$5F,$A0,$77,$3A,$90,$EB
	.dc.b	$5D,$E7,$FC,$B1,$19,$6B,$A3,$15,$F8,$27,$EF,$57,$BE,$BF,$DF,$07
	.dc.b	$FA,$B2,$B8,$92,$B8,$57,$DC,$AE,$13,$FE,$A3,$9F,$8B,$1E,$93,$E4
	.dc.b	$D7,$38,$E3,$1F,$0E,$7F,$D3,$75,$F9,$68,$DC,$FF,$AE,$1E,$55,$E4
	.dc.b	$8C,$3C,$E7,$F2,$67,$B0,$A7,$BC,$95,$E9,$72,$D7,$E6,$1A,$9F,$2C
	.dc.b	$FF,$9D,$67,$FC,$F0,$FF,$9E,$E3,$F3,$33,$F9,$87,$F3,$4F,$FA,$02
	.dc.b	$7F,$35,$5F,$E8,$68,$FC,$9E,$1F,$9A,$7F,$81,$D7,$EF,$23,$F8,$CC
	.dc.b	$7F,$12,$7F,$36,$7F,$0F,$8F,$CF,$C6,$6B,$F8,$A3,$FE,$91,$C1,$5F
	.dc.b	$81,$AF,$4C,$FC,$33,$DB,$5E,$FB,$F9,$6F,$D7,$8F,$C7,$E1,$F2,$DD
	.dc.b	$D7,$27,$E2,$1F,$21,$C4,$F3,$9E,$15,$D2,$7F,$38,$7E,$0A,$32,$47
	.dc.b	$EC,$90,$FF,$72,$D7,$80,$EC,$43,$04,$9E,$CF,$FB,$2E,$3B,$1C,$EB
	.dc.b	$F5,$75,$DE,$3A,$D7,$0A,$FF,$5F,$3F,$8E,$8C,$CE,$EB,$8C,$73,$3A
	.dc.b	$9A,$31,$5F,$21,$AF,$ED,$AB,$B3,$F9,$BA,$FC,$C4,$F1,$25,$76,$69
	.dc.b	$D9,$C1,$1C,$12,$7C,$0F,$06,$C6,$DF,$E7,$58,$7F,$21,$9F,$F2,$2C
	.dc.b	$7F,$70,$C3,$F1,$F3,$FD,$7E,$3F,$95,$57,$E3,$1F,$ED,$8F,$F4,$EA
	.dc.b	$E3,$18,$7F,$BE,$47,$F2,$CA,$FD,$1C,$FF,$B6,$EB,$F2,$73,$FC,$ED
	.dc.b	$FC,$AC,$7E,$4F,$0F,$8A,$FF,$43,$9F,$C9,$3F,$27,$0F,$A5,$87,$BE
	.dc.b	$30,$FC,$EA,$FB,$15,$EF,$9F,$E4,$B7,$EF,$7F,$CA,$91,$E9,$6D,$7A
	.dc.b	$0F,$5C,$FF,$47,$9F,$EE,$D3,$FD,$E8,$FD,$AB,$ED,$3F,$DB,$55,$E8
	.dc.b	$3F,$8F,$D7,$E7,$A7,$E8,$BE,$78,$F2,$B6,$FF,$03,$5E,$79,$CD,$70
	.dc.b	$21,$EE,$AF,$B8,$D9,$E1,$5C,$A7,$A4,$7C,$8A,$E7,$5B,$9C,$9E,$57
	.dc.b	$07,$68,$D9,$2B,$A4,$65,$AF,$02,$46,$EB,$C1,$3A,$8C,$57,$76,$D7
	.dc.b	$ED,$4E,$95,$C0,$93,$BA,$E4,$48,$CC,$F1,$3C,$31,$99,$E1,$3A,$9F
	.dc.b	$DD,$CE,$5E,$15,$E2,$73,$3B,$7F,$68,$D7,$F2,$55,$D0,$8F,$2A,$E8
	.dc.b	$F2,$8F,$E7,$F4,$9F,$F0,$DC,$7F,$46,$91,$FA,$49,$D4,$0C,$4F,$EA
	.dc.b	$27,$FA,$54,$71,$7F,$A3,$BF,$D0,$2F,$F6,$D1,$FD,$06,$3F,$C8,$E7
	.dc.b	$F9,$51,$FF,$13,$3B,$C3,$FA,$4C,$62,$3F,$A7,$61,$F9,$AB,$FD,$6C
	.dc.b	$FD,$49,$F7,$BE,$D3,$E0,$4F,$C5,$C3,$F3,$F1,$89,$DB,$F2,$9A,$FC
	.dc.b	$A3,$44,$3A,$10,$E0,$64,$A6,$4E,$45,$8F,$0B,$6B,$FC,$52,$F1,$36
	.dc.b	$72,$21,$82,$47,$27,$2D,$73,$1D,$1E,$D5,$FE,$14,$72,$D9,$E0,$49
	.dc.b	$E1,$3D,$1A,$F0,$AF,$F0,$43,$FE,$2E,$8C,$CF,$1A,$E9,$18,$23,$82
	.dc.b	$57,$2A,$FF,$18,$4E,$A3,$15,$FE,$2B,$6C,$7F,$80,$E7,$FB,$54,$FF
	.dc.b	$19,$8E,$13,$F8,$FA,$FE,$F4,$FF,$2D,$8F,$CF,$D7,$E5,$23,$F3,$6F
	.dc.b	$F9,$D0,$FC,$CB,$F9,$B9,$FF,$12,$9F,$9D,$8F,$F3,$E5,$7E,$96,$3F
	.dc.b	$C4,$75,$FE,$1B,$AF,$C9,$9F,$CC,$EB,$F9,$A9,$FE,$B4,$C3,$F3,$E4
	.dc.b	$C3,$F3,$71,$FA,$49,$E2,$FF,$A2,$67,$FA,$7C,$7C,$57,$FA,$7C,$7F
	.dc.b	$86,$0F,$F0,$45,$7E,$81,$FE,$17,$7F,$9D,$8F,$DC,$D7,$BB,$0F,$F1
	.dc.b	$C4,$FF,$90,$23,$DC,$66,$7F,$71,$5F,$E3,$F8,$F9,$EF,$EB,$E7,$FA
	.dc.b	$44,$7B,$63,$FD,$6B,$87,$C1,$9C,$C7,$BE,$3D,$B1,$FD,$4B,$0F,$EB
	.dc.b	$53,$C0,$E9,$5F,$D6,$2B,$93,$63,$27,$CA,$3F,$D9,$AF,$FB,$2D,$F9
	.dc.b	$15,$E0,$73,$1F,$C2,$5C,$BC,$DA,$73,$3F,$C0,$AF,$F1,$07,$FC,$15
	.dc.b	$1F,$DB,$5F,$31,$C0,$AF,$AE,$7F,$CC,$11,$FD,$B4,$95,$EE,$3F,$D6
	.dc.b	$8F,$C3,$9F,$F0,$94,$7F,$75,$AF,$F6,$01,$F8,$09,$F4,$9E,$27,$E7
	.dc.b	$47,$28,$C9,$FE,$17,$3A,$47,$A5,$C1,$A8,$EC,$FF,$80,$23,$E1,$4F
	.dc.b	$A5,$FE,$58,$F9,$2B,$E8,$CF,$C5,$3F,$9A,$4F,$CF,$9E,$0F,$9E,$3E
	.dc.b	$4B,$93,$D7,$1F,$4E,$33,$1F,$4E,$38,$99,$8E,$B1,$93,$89,$E6,$2B
	.dc.b	$E3,$AF,$D4,$4F,$9E,$3F,$0A,$7D,$49,$DB,$4F,$80,$71,$AF,$C9,$C6
	.dc.b	$AB,$E6,$4F,$98,$FA,$07,$CC,$8F,$63,$FC,$3A,$FF,$5D,$87,$F7,$18
	.dc.b	$FE,$49,$87,$F1,$AA,$FE,$20,$7E,$AC,$FD,$6C,$7F,$AA,$CF,$E1,$35
	.dc.b	$FB,$47,$FB,$5D,$7E,$D2,$7C,$98,$7F,$96,$8F,$3C,$FA,$8F,$61,$F8
	.dc.b	$E7,$F2,$38,$7E,$3D,$E1,$87,$C5,$9F,$E0,$D7,$FC,$96,$3C,$D7,$FB
	.dc.b	$F9,$FF,$2C,$61,$EC,$3F,$80,$D7,$F1,$D9,$FE,$03,$3F,$A1,$C3,$E3
	.dc.b	$BF,$60,$F8,$E7,$E3,$F0,$FC,$4B,$E6,$8F,$26,$1F,$AB,$78,$9F,$6A
	.dc.b	$BF,$30,$FE,$5A,$7D,$91,$F9,$43,$35,$D2,$78,$BF,$12,$77,$3F,$93
	.dc.b	$7C,$87,$E8,$8F,$B8,$53,$33,$DC,$F2,$C7,$EB,$A7,$E4,$BE,$09,$FC
	.dc.b	$6C,$F8,$2B,$E6,$1F,$9A,$9F,$CB,$4F,$D5,$9D,$47,$18,$F9,$15,$D2
	.dc.b	$3F,$17,$1E,$8A,$FD,$15,$7C,$39,$CB,$E4,$9F,$C9,$47,$E9,$5E,$D5
	.dc.b	$CE,$31,$1C,$DF,$1B,$4F,$D0,$57,$A2,$3F,$1A,$FD,$39,$E5,$3F,$68
	.dc.b	$F8,$F5,$F1,$63,$F1,$A4,$AF,$A8,$79,$9C,$3F,$46,$3D,$4F,$91,$F5
	.dc.b	$4F,$E1,$4F,$97,$3E,$07,$EA,$47,$07,$C4,$7D,$68,$FC,$01,$B8,$F2
	.dc.b	$4F,$88,$ED,$1E,$2A,$F9,$6F,$E1,$67,$C3,$5E,$C8,$E7,$3F,$04,$95
	.dc.b	$F8,$16,$CF,$D0,$3C,$75,$E5,$7C,$F5,$D4,$F0,$57,$D0,$9E,$F3,$E9
	.dc.b	$3D,$F5,$D1,$F8,$47,$E1,$1C,$3E,$53,$73,$E3,$23,$E3,$3C,$31,$F6
	.dc.b	$DF,$C1,$BC,$9F,$01,$D6,$38,$BE,$08,$FC,$C3,$F8,$48,$FC,$B4,$62
	.dc.b	$3F,$71,$5E,$C7,$53,$F7,$CF,$A0,$FC,$48,$F8,$51,$88,$F2,$BD,$0F
	.dc.b	$C7,$9E,$27,$DC,$79,$4F,$65,$7D,$C8,$FC,$54,$F4,$6D,$7B,$EB,$F3
	.dc.b	$87,$9A,$3C,$15,$C6,$7F,$02,$6D,$FA,$67,$D9,$35,$1F,$5E,$BD,$D5
	.dc.b	$CE,$3B,$9E,$97,$E6,$BE,$97,$E6,$D7,$C7,$AF,$DF,$CF,$C5,$38,$96
	.dc.b	$3C,$6F,$39,$F4,$CF,$08,$FC,$C9,$EA,$AF,$CA,$BD,$E3,$D1,$18,$9F
	.dc.b	$8E,$D7,$EE,$4F,$CE,$72,$7E,$28,$A7,$CF,$36,$7B,$0C,$D7,$AE,$7F
	.dc.b	$5C,$75,$8F,$AF,$19,$3F,$1D,$5E,$48,$EC,$74,$3D,$B3,$C9,$C9,$CA
	.dc.b	$BC,$CE,$E7,$EA,$CF,$DF,$AF,$8F,$5F,$24,$FA,$C7,$C5,$AE,$67,$D2
	.dc.b	$79,$BD,$4F,$61,$D6,$7E,$ED,$7C,$F8,$FC,$C5,$7E,$16,$72,$79,$E3
	.dc.b	$E3,$1C,$CF,$5C,$7C,$08,$F3,$57,$90,$FB,$E7,$DA,$2C,$F6,$7F,$01
	.dc.b	$5E,$EA,$F0,$47,$AE,$7B,$9E,$B3,$D2,$78,$67,$84,$72,$AF,$3D,$78
	.dc.b	$CE,$71,$E0,$AF,$D8,$BF,$25,$FD,$CB,$E5,$8F,$9A,$D3,$D4,$E2,$3D
	.dc.b	$11,$C5,$FD,$24,$65,$EB,$1C,$5E,$A7,$07,$0F,$07,$A9,$1D,$9C,$8E
	.dc.b	$04,$30,$4A,$FD,$BB,$C1,$AF,$53,$15,$F9,$77,$E5,$1F,$62,$7F,$55
	.dc.b	$1F,$62,$36,$7E,$2A,$7F,$34,$FD,$26,$D7,$E7,$EB,$8B,$F3,$8F,$85
	.dc.b	$1E,$23,$F1,$93,$E7,$AF,$14,$72,$AF,$CA,$CF,$92,$3E,$DC,$7C,$D7
	.dc.b	$89,$CE,$7F,$25,$5E,$1A,$FD,$AD,$72,$AE,$91,$F8,$63,$A5,$7E,$12
	.dc.b	$7C,$E7,$E1,$CF,$C7,$57,$39,$FD,$24,$60,$E9,$3F,$16,$3E,$61,$2B
	.dc.b	$CB,$1E,$C3,$4E,$23,$E5,$46,$0F,$4C,$F4,$AF,$90,$FE,$06,$7A,$4E
	.dc.b	$9F,$55,$7B,$1F,$B0,$7C,$03,$31,$D6,$BF,$41,$5E,$D3,$F4,$15,$F2
	.dc.b	$63,$31,$F3,$67,$B5,$74,$3E,$0C,$70,$8F,$64,$F9,$A7,$F0,$A7,$E0
	.dc.b	$E3,$C1,$3F,$30,$E9,$3D,$9F,$CF,$D7,$C3,$AE,$13,$CD,$F7,$BB,$7F
	.dc.b	$3D,$1E,$79,$E8,$72,$AF,$3C,$7E,$1A,$31,$5C,$6B,$F1,$C7,$D0,$7B
	.dc.b	$4F,$85,$FB,$D1,$FC,$26,$BF,$08,$FC,$B3,$E1,$19,$AF,$C1,$3F,$3A
	.dc.b	$79,$CF,$88,$E3,$1F,$22,$3F,$43,$1F,$22,$7E,$94,$62,$7C,$91,$83
	.dc.b	$F0,$15,$E5,$7F,$08,$66,$3E,$7C,$FA,$1F,$CD,$C7,$91,$F3,$98,$7E
	.dc.b	$C3,$D9,$FA,$67,$CA,$7E,$6D,$79,$63,$31,$F0,$6B,$E6,$3A,$39,$47
	.dc.b	$CE,$39,$3F,$38,$FB,$6F,$9D,$FB,$6E,$A3,$ED,$47,$59,$F8,$EF,$CD
	.dc.b	$9F,$3B,$C9,$F1,$BF,$2D,$F8,$B1,$F3,$DF,$81,$3E,$69,$F1,$57,$C5
	.dc.b	$8F,$D3,$19,$3C,$31,$CC,$F9,$73,$F3,$8E,$EF,$8E,$7E,$B3,$E3,$8F
	.dc.b	$80,$FA,$5F,$0D,$7A,$A3,$E8,$4F,$03,$C5,$1F,$A4,$8E,$13,$EE,$9C
	.dc.b	$9F,$8A,$3F,$31,$5E,$69,$F1,$4F,$C0,$9C,$D7,$D7,$8F,$1C,$FA,$CF
	.dc.b	$9E,$76,$9F,$14,$7B,$4D,$C7,$43,$E0,$9F,$9B,$7E,$C5,$7E,$46,$31
	.dc.b	$1F,$11,$F6,$BE,$33,$EE,$57,$8E,$3F,$0C,$7C,$72,$D7,$BA,$7A,$CF
	.dc.b	$D9,$7E,$11,$E5,$9D,$B5,$FB,$6D,$AF,$89,$1E,$23,$D2,$7E,$81,$FB
	.dc.b	$F3,$F4,$DB,$3F,$58,$87,$82,$7D,$AF,$C4,$7A,$1F,$28,$F9,$91,$F0
	.dc.b	$DE,$D3,$F6,$DC,$BE,$BA,$F9,$F1,$D0,$F5,$94,$F5,$96,$7D,$67,$9A
	.dc.b	$BC,$91,$DA,$7F,$09,$1F,$3A,$3F,$62,$74,$34,$FC,$C8,$FC,$2C,$68
	.dc.b	$EE,$6A,$3E,$D4,$F7,$9F,$3C,$E6,$7E,$14,$E5,$FB,$D1,$F8,$B6,$BF
	.dc.b	$02,$7B,$BF,$7C,$EF,$3F,$AF,$AF,$C2,$46,$1F,$A8,$7C,$97,$F2,$47
	.dc.b	$CF,$2C,$7C,$F3,$93,$EB,$7E,$3C,$71,$8F,$AF,$1B,$8E,$B3,$F2,$27
	.dc.b	$D0,$7E,$1A,$79,$BE,$77,$9C,$6E,$3D,$95,$F3,$DD,$57,$AA,$B7,$3F
	.dc.b	$3C,$D5,$7E,$75,$F9,$B3,$E5,$9F,$95,$5E,$18,$F2,$CF,$18,$F2,$9F
	.dc.b	$04,$F5,$C7,$E2,$E3,$E7,$C7,$C4,$8C,$3C,$E3,$F3,$75,$F6,$AB,$DA
	.dc.b	$F8,$E3,$ED,$D7,$8C,$91,$EF,$AE,$8F,$DA,$38,$3F,$8D,$AE,$75,$FB
	.dc.b	$1A,$F8,$44,$9F,$A8,$D7,$E3,$98,$21,$82,$57,$67,$F0,$93,$EA,$3D
	.dc.b	$EF,$C2,$8F,$24,$78,$63,$9C,$71,$AF,$A7,$3F,$42,$3D,$93,$C6,$3E
	.dc.b	$A9,$C5,$F0,$B6,$39,$57,$A5,$B3,$DA,$74,$47,$F7,$13,$B8,$C4,$7E
	.dc.b	$A4,$F9,$11,$F7,$A3,$C8,$7D,$F8,$FC,$69,$F8,$0A,$FC,$C4,$7F,$A7
	.dc.b	$40,$89,$09,$CC,$C9,$CC,$57,$80,$9E,$13,$C0,$CB,$CE,$3B,$47,$81
	.dc.b	$E6,$ED,$AF,$46,$BB,$7B,$4F,$82,$31,$3E,$07,$C8,$D9,$E4,$E6,$37
	.dc.b	$3C,$5F,$2B,$E7,$69,$DE,$7C,$2E,$9F,$4B,$A3,$BC,$7C,$38,$D9,$F1
	.dc.b	$8D,$98,$3D,$31,$D5,$AF,$27,$66,$1F,$39,$B9,$F5,$3E,$38,$F0,$BF
	.dc.b	$25,$E6,$7C,$09,$F6,$3E,$33,$71,$1D,$3C,$67,$EB,$46,$A3,$C2,$7D
	.dc.b	$37,$C8,$F8,$4C,$9A,$79,$BE,$B7,$93,$F1,$DC,$BF,$62,$40

;--------------------------------------------------------------------------------
;	FMOVE.W Dr,FPn
;	FMOVE.W <mem>,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovewtoreg_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.W <EA>,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.W <ea>,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d2	source
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source,...
;	a4	expected result handle,expected status,...
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fmovewtoreg_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a0,a2			;source,...
@@:
	addq.l	#4,a0			;source
	tst.l	(a0)
	bne	@b
	addq.l	#4,a0			;0
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;Dr,FPn
	fmove.l	#0,fpcr
	moveq.l	#0,d2
	move.w	(a2),d2			;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fmove.w	d2,fp7			;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)		;fpcr
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.W Dr='
	puthex4	d2			;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
;<mem>,FPn
	fmove.l	#0,fpcr
	moveq.l	#0,d2
	move.w	(a2),d2			;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.w	d2,-(sp)		;source
	fmove.w	(sp)+,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)		;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.W <mem>='
	puthex4	d2			;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source
	tst.l	(a2)			;source
	bne	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmovewtoreg_data_compressed::
	.dc.b	$00,$00,$64,$08,$10,$8D,$51,$67,$1F,$7D,$9F,$BF,$D7,$E0,$1B,$FC
	.dc.b	$07,$87,$E0,$5C,$7F,$03,$E5,$F8,$27,$3F,$C1,$7A,$7E,$FD,$D7,$F7
	.dc.b	$FE,$DF,$C0,$1D,$FF,$80,$7C,$3C,$1E,$2E,$1F,$1F,$0F,$93,$87,$CB
	.dc.b	$C3,$E6,$E1,$F3,$F0,$FA,$38,$7D,$3C,$3E,$AE,$1F,$5F,$0F,$B3,$87
	.dc.b	$DB,$C3,$EE,$E1,$F7,$F0,$FC,$0E,$1F,$85,$F1,$9F,$04,$70,$E3,$87
	.dc.b	$3C,$3A,$E1,$DF,$0F,$0E,$1E,$3C,$3C,$B8,$79,$F0,$F4,$E1,$EB,$C3
	.dc.b	$DB,$87,$BF,$0F,$83,$87,$C3,$C3,$F4,$FE,$37,$E9,$1E,$2E,$3F,$AE
	.dc.b	$FD,$46,$F0,$6F,$83,$87,$07,$1E,$0E,$5C,$1C,$F8,$3A,$70,$75,$E0
	.dc.b	$ED,$C1,$DF,$83,$C1,$C1,$E1,$E0,$F1,$70,$7E,$17,$E3,$BF,$08,$78
	.dc.b	$F8,$0F,$8D,$BC,$1B,$E0,$E1,$C1,$C7,$83,$97,$07,$3E,$0E,$9C,$1D
	.dc.b	$78,$3B,$70,$77,$E0,$F0,$70,$78,$78,$3C,$5C,$1F,$97,$F8,$FF,$CB
	.dc.b	$1E,$33,$C4,$37,$D6,$FB,$DF,$86,$FC,$77,$E5,$BF,$3D,$FA,$6F,$D7
	.dc.b	$7E,$DB,$F7,$DF,$C1,$BF,$87,$7F,$D4,$67,$FA,$2E,$4B,$C9,$8E,$0C
	.dc.b	$F0,$6B,$83,$77,$C3,$83,$8F,$07,$2E,$0E,$7C,$1D,$38,$3A,$F0,$76
	.dc.b	$E0,$EF,$C1,$E0,$E0,$FD,$C6,$5F,$B1,$E0,$BC,$18,$E0,$CF,$06,$B8
	.dc.b	$37,$C1,$C3,$0E,$3C,$1C,$B8,$39,$F0,$74,$E0,$EB,$C1,$DB,$83,$BF
	.dc.b	$07,$83,$83,$F8,$46,$3F,$C0,$38,$2F,$06,$38,$33,$C1,$AE,$0D,$F0
	.dc.b	$70,$E0,$E3,$8F,$2E,$0E,$7C,$1D,$38,$3A,$F0,$76,$E0,$EF,$C1,$E0
	.dc.b	$E0,$FE,$31,$87,$F0,$FE,$0B,$C1,$8E,$0C,$F0,$6B,$83,$7C,$1C,$38
	.dc.b	$38,$F0,$72,$CB,$9F,$07,$4E,$0E,$BC,$1D,$B8,$3B,$F0,$78,$38,$3F
	.dc.b	$94,$7E,$89,$FC,$9C,$FD,$21,$FA,$33,$F2,$5C,$7F,$CB,$96,$FC,$37
	.dc.b	$E3,$BF,$2D,$F9,$EF,$D3,$7E,$BB,$F6,$DF,$BE,$FE,$0D,$FC,$3B,$F8
	.dc.b	$B7,$FE,$75,$FA,$2F,$F3,$93,$F6,$C7,$ED,$4F,$1F,$01,$F1,$BF,$D0
	.dc.b	$96,$DC,$B6,$E7,$B7,$4D,$BA,$ED,$DB,$6E,$FB,$78,$36,$F0,$ED,$E2
	.dc.b	$DB,$C7,$B7,$F4,$DF,$D1,$BF,$A6,$1F,$C2,$4F,$E1,$07,$90,$F1,$EE
	.dc.b	$7C,$6F,$F5,$65,$AF,$3D,$7A,$6B,$D7,$5E,$DA,$F7,$D7,$C1,$AF,$87
	.dc.b	$5F,$16,$BE,$3D,$7F,$B0,$7E,$8F,$FD,$7C,$FE,$38,$7F,$1B,$3E,$39
	.dc.b	$E4,$3C,$7B,$1F,$1B,$FD,$A9,$69,$D3,$4E,$BA,$76,$D3,$BE,$9E,$0D
	.dc.b	$3C,$3A,$78,$B4,$F1,$E9,$FD,$CD,$FE,$CB,$CB,$79,$71,$CB,$9E,$4D
	.dc.b	$71,$EF,$8B,$87,$07,$1E,$0E,$5C,$1C,$F8,$3A,$70,$75,$E0,$ED,$C1
	.dc.b	$DF,$7F,$07,$07,$F8,$0C,$FE,$E9,$C1,$78,$31,$C1,$9E,$0D,$70,$6F
	.dc.b	$83,$87,$07,$1E,$0E,$5C,$1C,$F8,$3A,$70,$75,$E0,$ED,$C1,$DF,$83
	.dc.b	$C1,$C1,$FE,$11,$E2,$9E,$DB,$E3,$9D,$99,$94,$E6,$30,$D8,$E3,$3C
	.dc.b	$1B,$3A,$8C,$47,$F8,$7E,$78,$36,$32,$FF,$89,$63,$2D,$8C,$BF,$E2
	.dc.b	$A8,$CB,$63,$2B,$E0,$9B,$7A,$63,$0D,$8C,$C7,$B5,$AF,$C3,$6B,$FE
	.dc.b	$35,$6B,$FE,$39,$6B,$FE,$3C,$6B,$D8,$E9,$1D,$C8,$FC,$BF,$83,$7E
	.dc.b	$AD,$E5,$6C,$66,$30,$47,$04,$3E,$C4,$70,$6C,$6D,$A6,$CF,$F2,$77
	.dc.b	$C3,$3D,$CF,$88,$D8,$FB,$CD,$7F,$08,$D7,$8C,$70,$23,$C1,$AF,$06
	.dc.b	$BC,$0F,$C7,$78,$AF,$D4,$DF,$9C,$7E,$36,$30,$D8,$CC,$F0,$6B,$FE
	.dc.b	$67,$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7F,$CE,$53,$C9,$CF,$E8,$9E
	.dc.b	$12,$3E,$16,$BA,$30,$43,$04,$3A,$CA,$38,$C7,$FA,$07,$C7,$7E,$54
	.dc.b	$FC,$0F,$F5,$D1,$86,$C6,$67,$F6,$D3,$96,$C6,$5F,$F4,$54,$65,$B1
	.dc.b	$97,$C0,$D7,$FD,$27,$1B,$73,$3E,$02,$3F,$C0,$7E,$51,$F9,$3F,$EA
	.dc.b	$DB,$19,$8C,$11,$C1,$0F,$E1,$D1,$C1,$B1,$B6,$9B,$3F,$D4,$9C,$4F
	.dc.b	$8A,$FD,$D3,$63,$F8,$A3,$5F,$E3,$4D,$78,$C7,$02,$3C,$1A,$F0,$6B
	.dc.b	$C0,$EB,$97,$80,$8E,$08,$78,$1C,$B4,$C9,$EA,$72,$D3,$ED,$FE,$98
	.dc.b	$FC,$77,$F0,$2D,$8C,$C6,$08,$E0,$87,$E0,$23,$83,$63,$6D,$36,$7F
	.dc.b	$B4,$7D,$37,$E8,$EF,$DD,$BF,$9F,$46,$1B,$19,$9F,$E8,$D3,$96,$C6
	.dc.b	$5F,$F6,$D4,$65,$B1,$97,$C0,$D7,$FD,$C7,$1B,$73,$3E,$02,$3F,$D5
	.dc.b	$7D,$3F,$F7,$2F,$B3,$FD,$4D,$B1,$9F,$25,$39,$6C,$66,$30,$47,$04
	.dc.b	$3F,$DE,$53,$D1,$B3,$C9,$B3,$D7,$E1,$64,$78,$5A,$F0,$6B,$FE,$FF
	.dc.b	$AF,$F7,$EC,$62,$02,$32,$F3,$21,$CC,$FC,$3E,$FF,$86,$23,$82,$1F
	.dc.b	$83,$72,$D3,$27,$F1,$1D,$3F,$64,$D8,$DB,$FB,$0F,$D8,$FF,$5E,$47
	.dc.b	$F5,$ED,$7F,$5E,$D7,$81,$A2,$1A,$29,$F0,$0F,$83,$F3,$7F,$C4,$DB
	.dc.b	$C7,$FD,$DE,$31,$06,$92,$34,$48,$FF,$0B,$46,$5B,$19,$7F,$C3,$D0
	.dc.b	$F1,$3B,$12,$38,$B5,$DB,$5F,$41,$E3,$70,$FA,$E3,$A3,$C8,$87,$ED
	.dc.b	$7F,$36,$FD,$E9,$1F,$E0,$2D,$74,$60,$86,$08,$7D,$97,$83,$FA,$C9
	.dc.b	$FD,$B9,$1F,$DA,$B5,$FD,$9B,$5E,$06,$88,$68,$A7,$F0,$7F,$D2,$9D
	.dc.b	$5F,$C1,$9B,$1F,$C1,$1A,$FF,$02,$6B,$C6,$38,$11,$E0,$D7,$83,$5E
	.dc.b	$07,$5F,$82,$F0,$11,$C1,$0F,$03,$96,$99,$3D,$4E,$5A,$7A,$7B,$7D
	.dc.b	$84,$70,$43,$D8,$E5,$A6,$4F,$8C,$E5,$A7,$C5,$F0,$3E,$41,$1C,$10
	.dc.b	$F9,$0E,$5A,$64,$FA,$4E,$5A,$7D,$1E,$9F,$A8,$47,$04,$3E,$A3,$96
	.dc.b	$99,$3E,$F3,$96,$9F,$77,$E9,$3F,$00,$47,$04,$3F,$00,$E5,$A6,$4F
	.dc.b	$C5,$39,$69,$F8,$9F,$B0,$FC,$61,$1C,$10,$FC,$63,$96,$99,$3F,$2C
	.dc.b	$E5,$A7,$E5,$7E,$F3,$F3,$04,$70,$43,$F3,$0E,$5A,$64,$FD,$13,$96
	.dc.b	$9F,$A1,$FC,$23,$F4,$84,$70,$43,$F4,$8E,$5A,$64,$FD,$73,$96,$9F
	.dc.b	$AD,$FC,$4F,$F6,$04,$70,$43,$F6,$0E,$5A,$64,$FD,$D3,$96,$9F,$B9
	.dc.b	$F1,$BF,$78,$47,$04,$3F,$78,$E5,$A6,$4F,$E0,$CE,$5A,$7F,$A0,$3F
	.dc.b	$3D,$FF,$66,$37,$2F,$F4,$04,$62,$3F,$CF,$8D,$7F,$CE,$AD,$7F,$B3
	.dc.b	$EC,$8C,$47,$F1,$46,$C6,$67,$A9,$27,$C0,$43,$F8,$B7,$77,$F1,$B2
	.dc.b	$3F,$C6,$DA,$E8,$C1,$0C,$10,$FE,$4C,$F0,$7F,$94,$73,$7F,$27,$23
	.dc.b	$FC,$9D,$AF,$F2,$76,$BC,$0D,$10,$D1,$4F,$E6,$9C,$7F,$CC,$C8,$FF
	.dc.b	$33,$6B,$FC,$CD,$AF,$03,$44,$34,$53,$FA,$07,$E8,$5F,$CF,$C8,$FF
	.dc.b	$3F,$6B,$FC,$FD,$AF,$03,$44,$34,$53,$FA,$67,$E9,$7F,$D2,$C8,$FF
	.dc.b	$4B,$6B,$FD,$2D,$AF,$03,$44,$34,$53,$FA,$C7,$EA,$5F,$D5,$C8,$FF
	.dc.b	$57,$6B,$FD,$5D,$AF,$03,$44,$34,$53,$FB,$27,$EB,$9F,$D8,$C8,$FF
	.dc.b	$63,$6B,$FD,$8D,$AF,$03,$44,$34,$53,$FB,$87,$EC,$FF,$DB,$C8,$FF
	.dc.b	$6F,$6B,$FD,$BD,$AF,$03,$44,$34,$53,$FB,$E7,$EE,$7F,$DE,$C8,$FF
	.dc.b	$7B,$6B,$FD,$ED,$AF,$03,$44,$34,$53,$FC,$11,$FB,$F7,$F8,$1C,$8F
	.dc.b	$F8,$1D,$AF,$F8,$1D,$AF,$03,$44,$34,$53,$FD,$35,$83,$D1,$FF,$4C
	.dc.b	$B6,$3F,$D2,$ED,$7F,$D2,$AD,$78,$C7,$02,$3C,$1A,$F0,$6B,$C0,$EB
	.dc.b	$E1,$78,$08,$E0,$87,$81,$CB,$4C,$9E,$A7,$2D,$3D,$3E,$17,$B0,$8E
	.dc.b	$08,$7B,$1C,$B4,$C9,$F1,$9C,$B4,$F8,$BE,$47,$C8,$23,$82,$1F,$21
	.dc.b	$CB,$4C,$9F,$49,$CB,$4F,$A3,$F5,$1F,$50,$8E,$08,$7D,$47,$2D,$32
	.dc.b	$7D,$E7,$2D,$3E,$EF,$C2,$7E,$00,$8E,$08,$7E,$01,$CB,$4C,$9F,$8A
	.dc.b	$72,$D3,$F1,$3F,$71,$F8,$C2,$38,$21,$F8,$C7,$2D,$32,$7E,$59,$CB
	.dc.b	$4F,$CA,$FE,$07,$F9,$82,$38,$21,$F9,$87,$2D,$32,$7E,$89,$CB,$4F
	.dc.b	$D0,$FE,$1D,$FA,$42,$38,$21,$FA,$47,$2D,$32,$7E,$B9,$CB,$4F,$D6
	.dc.b	$FE,$31,$FB,$02,$38,$21,$FB,$07,$2D,$32,$7E,$E9,$CB,$4F,$DC,$FC
	.dc.b	$0F,$EF,$08,$E0,$87,$EF,$1C,$B4,$C9,$FC,$19,$CB,$4F,$E0,$BF,$93
	.dc.b	$FF,$08,$23,$82,$1F,$C2,$1C,$B4,$C9,$FC,$49,$CB,$4F,$E2,$3F,$99
	.dc.b	$7F,$14,$23,$82,$1F,$C5,$1C,$B4,$C9,$FC,$79,$CB,$4F,$E3,$BF,$9E
	.dc.b	$7F,$20,$23,$82,$1F,$C8,$1C,$B4,$C9,$FC,$A9,$CB,$4F,$E5,$3E,$BF
	.dc.b	$F2,$C2,$38,$21,$FC,$B1,$CB,$4C,$9F,$CD,$9C,$B4,$FE,$6B,$E9,$FF
	.dc.b	$38,$23,$82,$1F,$CE,$1C,$B4,$C9,$FD,$09,$CB,$4F,$E8,$3E,$77,$F4
	.dc.b	$42,$38,$21,$FD,$11,$CB,$4C,$9F,$D3,$9C,$B4,$FE,$9B,$EC,$FF,$50
	.dc.b	$23,$82,$1F,$D4,$1C,$B4,$C9,$FD,$69,$CB,$4F,$EB,$3E,$7F,$F5,$C2
	.dc.b	$38,$21,$FD,$71,$CB,$4C,$9F,$D9,$9C,$B4,$FE,$CB,$FB,$8F,$F6,$82
	.dc.b	$38,$21,$FD,$A1,$CB,$4C,$9F,$DC,$9C,$B4,$10,$BE,$BB,$F3,$A7,$3E
	.dc.b	$67,$53,$98,$C4,$70,$9D,$B8,$9D,$36,$32,$F3,$8C,$B6,$32,$F6,$6B
	.dc.b	$D6,$7B,$A9,$E2,$EE,$33,$18,$8F,$27,$46,$2B,$FC,$19,$18,$3C,$91
	.dc.b	$92,$38,$21,$E5,$72,$D3,$27,$C5,$FB,$47,$97,$E6,$84,$61,$F4,$B5
	.dc.b	$F9,$2D,$79,$47,$18,$C4,$7D,$56,$BF,$61,$AF,$68,$E0,$47,$BB,$FE
	.dc.b	$30,$FD,$EE,$47,$F8,$C1,$AF,$F8,$C1,$AF,$F8,$E1,$AF,$E1,$BD,$87
	.dc.b	$6A,$73,$29,$CC,$61,$7E,$A7,$F0,$DB,$3F,$69,$AF,$E2,$9A,$FE,$3D
	.dc.b	$AF,$E5,$5A,$F8,$E3,$A1,$1E,$8D,$7A,$3A,$8C,$92,$32,$D7,$C1,$1E
	.dc.b	$B7,$24,$33,$1B,$7C,$91,$B7,$27,$83,$E8,$3C,$64,$70,$43,$BB,$96
	.dc.b	$99,$3A,$B9,$69,$EC,$FA,$AF,$79,$1C,$10,$F5,$B9,$69,$93,$D2,$E5
	.dc.b	$A7,$C8,$FB,$1F,$96,$47,$04,$3E,$3B,$96,$99,$3E,$2B,$96,$9F,$53
	.dc.b	$EE,$3E,$B9,$1C,$10,$FA,$6E,$5A,$64,$FA,$2E,$5A,$7E,$03,$EF,$FF
	.dc.b	$06,$47,$04,$3E,$FB,$96,$99,$3E,$EB,$96,$9F,$8C,$FC,$27,$F1,$E4
	.dc.b	$70,$43,$F1,$6E,$5A,$64,$FC,$4B,$96,$9F,$98,$FC,$4F,$F3,$64,$70
	.dc.b	$43,$F2,$EE,$5A,$64,$FC,$AB,$96,$9F,$A4,$F7,$7F,$4E,$47,$04,$3F
	.dc.b	$46,$E5,$A6,$4F,$D0,$B9,$69,$FB,$0F,$C9,$BF,$66,$47,$04,$3F,$5E
	.dc.b	$E5,$A6,$4F,$D6,$B9,$69,$FB,$CF,$CA,$BF,$7E,$47,$04,$3F,$76,$E5
	.dc.b	$A6,$4F,$DC,$B9,$69,$FC,$23,$F3,$4F,$E1,$64,$70,$43,$F8,$3B,$96
	.dc.b	$99,$3F,$82,$B9,$69,$FC,$53,$E1,$BF,$8B,$91,$C1,$0F,$E2,$6E,$5A
	.dc.b	$64,$FE,$22,$E5,$A7,$F7,$4F,$55,$FA,$C7,$F5,$36,$C7,$F6,$A6,$BB
	.dc.b	$8D,$11,$D3,$5D,$18,$21,$83,$F9,$7F,$E8,$FF,$CC,$88,$E0,$87,$F2
	.dc.b	$E7,$2D,$32,$7F,$2C,$72,$D3,$F9,$DF,$EA,$1F,$CF,$88,$E0,$87,$F3
	.dc.b	$A7,$2D,$32,$7F,$38,$72,$D3,$FA,$3F,$CD,$7F,$4A,$23,$82,$1F,$D1
	.dc.b	$9C,$B4,$C9,$FD,$11,$CB,$4F,$EA,$7F,$17,$FD,$58,$8E,$08,$7F,$52
	.dc.b	$72,$D3,$27,$F5,$07,$2D,$3F,$AF,$FE,$CB,$FD,$88,$8E,$08,$7F,$5E
	.dc.b	$72,$D3,$27,$F5,$C7,$2D,$3F,$B5,$FE,$DF,$FD,$B8,$8E,$08,$7F,$6A
	.dc.b	$72,$D3,$27,$F6,$87,$2D,$3F,$BB,$FC,$5F,$F7,$A2,$38,$21,$FD,$D9
	.dc.b	$CB,$4C,$9F,$DD,$1C,$B4,$FF,$01,$FC,$47,$F8,$18,$8E,$08,$7F,$80
	.dc.b	$9C,$B4,$C9,$FE,$00,$72,$D3,$FC,$35,$F3,$CD,$7F,$C3,$CD,$8C,$C6
	.dc.b	$08,$E0,$87,$F8,$86,$38,$36,$36,$D3,$67,$F8,$96,$1F,$11,$E0,$24
	.dc.b	$79,$1A,$F5,$6B,$C5,$AF,$F8,$FE,$39,$BC,$48,$79,$7E,$23,$DA,$47
	.dc.b	$E0,$B5,$D1,$82,$18,$21,$E9,$78,$3E,$BF,$88,$F9,$24,$7E,$6B,$5F
	.dc.b	$8C,$D7,$81,$A2,$1A,$29,$F1,$FE,$0B,$EA,$91,$FB,$2D,$7E,$93,$5E
	.dc.b	$06,$88,$68,$A7,$D3,$F8,$8F,$C0,$91,$FC,$2B,$5F,$BC,$D7,$81,$A2
	.dc.b	$1A,$29,$F7,$FC,$3F,$C6,$91,$FC,$8B,$5F,$C5,$35,$E0,$68,$86,$8A
	.dc.b	$7E,$2F,$E2,$3F,$32,$47,$F3,$AD,$7F,$2C,$D7,$81,$A2,$1A,$29,$F9
	.dc.b	$7F,$88,$FD,$29,$1F,$D4,$B5,$FD,$13,$5E,$06,$88,$68,$A7,$E8,$FE
	.dc.b	$23,$F6,$24,$7F,$6A,$D7,$F5,$CD,$78,$1A,$21,$A2,$9F,$AF,$F8,$8F
	.dc.b	$DE,$91,$FE,$02,$D7,$F7,$4D,$78,$1A,$21,$A2,$9F,$BB,$F8,$7F,$E1
	.dc.b	$24,$7F,$86,$B5,$FE,$0C,$D7,$81,$A2,$1A,$29,$FC,$1F,$E1,$FF,$8A
	.dc.b	$91,$FE,$32,$D7,$F8,$93,$5E,$06,$88,$68,$A7,$F1,$3F,$87,$FE,$42
	.dc.b	$47,$F9,$2B,$5F,$E3,$CD,$78,$1A,$21,$A2,$9F,$C7,$FE,$1F,$F9,$69
	.dc.b	$1F,$E6,$2D,$7F,$95,$35,$E0,$68,$86,$8A,$7F,$2B,$F2,$FF,$9C,$91
	.dc.b	$FE,$7A,$D7,$F9,$B3,$5E,$06,$88,$68,$A7,$F3,$7F,$87,$FE,$8A,$47
	.dc.b	$FA,$4B,$5F,$E8,$4D,$78,$1A,$21,$A2,$9F,$D0,$FE,$1F,$FA,$89,$1F
	.dc.b	$EA,$AD,$7F,$A7,$35,$E0,$68,$86,$8A,$7F,$4F,$F8,$7F,$EB,$A4,$7F
	.dc.b	$B0,$B5,$FE,$B4,$D7,$81,$A2,$1A,$29,$FD,$6F,$E1,$FF,$B4,$91,$FE
	.dc.b	$DA,$D7,$FB,$33,$5E,$06,$88,$68,$A7,$F6,$7F,$87,$FE,$EA,$47,$FB
	.dc.b	$CB,$5F,$EE,$4D,$78,$1A,$21,$A2,$82,$79,$EF,$9E,$7A,$79,$98,$27
	.dc.b	$51,$88,$E3,$3C,$1C,$4E,$9B,$19,$79,$C6,$5B,$19,$7C,$11,$96,$C6
	.dc.b	$5E,$F3,$E1,$7C,$91,$96,$C6,$67,$D9,$EB,$5A,$34,$95,$E6,$6B,$E4
	.dc.b	$6B,$E2,$6B,$F0,$1A,$F8,$BE,$23,$B1,$1F,$82,$D7,$DE,$D7,$DA,$D7
	.dc.b	$E5,$B5,$E8,$7C,$AF,$88,$F5,$11,$C1,$0F,$96,$E5,$A6,$4F,$96,$E5
	.dc.b	$A7,$D6,$F8,$8F,$8C,$47,$04,$3E,$BB,$96,$99,$3E,$BB,$96,$9F,$82
	.dc.b	$F8,$8F,$A4,$47,$04,$3F,$06,$E5,$A6,$4F,$C1,$B9,$69,$F8,$EF,$88
	.dc.b	$FB,$C4,$70,$43,$F1,$EE,$5A,$64,$FC,$7B,$96,$9F,$9A,$F8,$8F,$C5
	.dc.b	$11,$C1,$0F,$CD,$B9,$69,$93,$F3,$6E,$5A,$7E,$9B,$E2,$3F,$2C,$47
	.dc.b	$04,$3F,$4E,$E5,$A6,$4F,$D3,$B9,$69,$FB,$2F,$88,$FD,$11,$1C,$10
	.dc.b	$FD,$9B,$96,$99,$3F,$66,$E5,$A7,$EF,$BE,$23,$F5,$C4,$70,$43,$F7
	.dc.b	$EE,$5A,$64,$FD,$FB,$96,$9F,$C2,$BE,$23,$F7,$44,$70,$43,$F8,$5B
	.dc.b	$96,$99,$3F,$85,$B9,$69,$FC,$5B,$E2,$3F,$83,$11,$C1,$0F,$E2,$EE
	.dc.b	$5A,$64,$FE,$2E,$E5,$A7,$F2,$2F,$88,$FE,$24,$47,$04,$3F,$91,$B9
	.dc.b	$69,$93,$F9,$1B,$96,$9F,$CB,$BE,$23,$F8,$F1,$1C,$10,$FE,$5E,$E5
	.dc.b	$A6,$4F,$E5,$EE,$5A,$7F,$3A,$F8,$8F,$E5,$44,$70,$43,$F9,$DB,$96
	.dc.b	$99,$3F,$9D,$B9,$69,$FD,$2F,$E1,$1F,$10,$ED,$FE,$7D,$18,$6C,$66
	.dc.b	$78,$35,$FE,$AF,$3B,$73,$3C,$08,$E0,$87,$09,$E4,$FF,$61,$9E,$4F
	.dc.b	$F5,$FF,$84,$78,$9F,$D4,$61,$18,$7F,$AA,$35,$E1,$1B,$8C,$47,$F6
	.dc.b	$C6,$BF,$DB,$DA,$FF,$75,$6B,$DA,$39,$11,$EE,$F8,$FE,$A1,$EE,$7F
	.dc.b	$BD,$27,$FB,$D4,$63,$CF,$FE,$D0,$D8,$FE,$F4,$D7,$FB,$FB,$5E,$33
	.dc.b	$CD,$FF,$04,$46,$5B,$19,$51,$96,$C6,$63,$04,$F4,$30,$4A,$F2,$B6
	.dc.b	$BC,$AD,$7C,$AD,$7D,$AD,$7C,$AF,$9F,$1F,$C5,$F9,$37,$33,$49,$7F
	.dc.b	$8A,$6D,$FE,$0D,$AF,$E0,$E7,$8C,$EC,$91,$B7,$FC,$3F,$1F,$1F,$B7
	.dc.b	$FC,$58,$DE,$DF,$CE,$CE,$89,$5F,$E2,$29,$D1,$23,$31,$FE,$22,$8C
	.dc.b	$3B,$8F,$F1,$6F,$C4,$F5,$23,$82,$47,$F8,$BA,$32,$D8,$CB,$FE,$2E
	.dc.b	$8C,$B6,$3F,$C7,$3F,$13,$E9,$23,$82,$47,$F8,$EA,$32,$D8,$CB,$FE
	.dc.b	$3A,$8C,$B6,$3F,$C8,$BF,$13,$F1,$48,$E0,$91,$FE,$46,$8C,$B6,$32
	.dc.b	$FF,$91,$A3,$2D,$8F,$F2,$8F,$C4,$7D,$12,$38,$24,$7F,$94,$A3,$2D
	.dc.b	$8C,$BF,$E5,$28,$CB,$63,$FC,$BF,$EC,$FF,$9A,$DB,$1F,$81,$8C,$3F
	.dc.b	$E6,$06,$BF,$E5,$E6,$BF,$E7,$26,$BF,$E6,$F6,$E6,$69,$23,$B1,$27
	.dc.b	$C7,$3E,$12,$38,$24,$78,$E3,$2D,$8C,$BE,$58,$CB,$63,$DF,$F0,$FE
	.dc.b	$D2,$38,$24,$7B,$E3,$2D,$8C,$BF,$06,$32,$D8,$F9,$7F,$0F,$F2,$48
	.dc.b	$E0,$91,$F2,$E3,$2D,$8C,$BF,$36,$32,$D8,$FA,$FF,$0F,$F5,$48,$E0
	.dc.b	$91,$F5,$E3,$2D,$8C,$BF,$66,$32,$D8,$FC,$1F,$C3,$FE,$04,$8E,$09
	.dc.b	$1F,$83,$8C,$B6,$32,$FE,$16,$32,$D8,$FC,$7F,$C3,$FE,$34,$8E,$09
	.dc.b	$1F,$8F,$8C,$B6,$32,$FE,$46,$32,$D8,$FC,$DF,$C3,$FE,$64,$8E,$09
	.dc.b	$1F,$9B,$8C,$B6,$32,$FE,$76,$32,$D8,$FD,$3F,$C3,$FE,$94,$8E,$09
	.dc.b	$1F,$A7,$8C,$B6,$32,$FE,$A6,$32,$D8,$FD,$9F,$3F,$EC,$48,$E0,$91
	.dc.b	$FB,$38,$CB,$63,$2F,$ED,$63,$2D,$8F,$DF,$FC,$3F,$EF,$48,$E0,$91
	.dc.b	$FB,$F8,$CB,$63,$2F,$F0,$18,$CB,$63,$F8,$5F,$C3,$FF,$09,$23,$82
	.dc.b	$47,$F0,$B8,$CB,$63,$2F,$F0,$D8,$CB,$63,$F8,$BF,$C3,$FF,$15,$23
	.dc.b	$82,$47,$F1,$78,$CB,$63,$2F,$F1,$98,$CB,$63,$F9,$1F,$C3,$FF,$21
	.dc.b	$23,$82,$47,$F2,$38,$CB,$63,$2F,$F2,$58,$CB,$63,$F9,$7F,$C3,$7F
	.dc.b	$2D,$23,$82,$47,$F2,$F8,$CB,$63,$2F,$F3,$18,$CB,$63,$F9,$DF,$C3
	.dc.b	$7F,$39,$23,$82,$47,$F3,$B8,$CB,$63,$2F,$F3,$D8,$CB,$63,$FA,$3F
	.dc.b	$1F,$F4,$52,$38,$24,$7F,$47,$8C,$B6,$32,$FF,$49,$8C,$B6,$3F,$A9
	.dc.b	$FC,$37,$F5,$12,$38,$24,$7F,$53,$8C,$B6,$32,$FF,$55,$8C,$B6,$3F
	.dc.b	$AF,$FC,$37,$F5,$D2,$38,$24,$7F,$5F,$8C,$B6,$32,$FF,$61,$8C,$B6
	.dc.b	$3F,$B5,$FC,$37,$F6,$92,$38,$24,$7F,$6B,$8C,$B6,$32,$FF,$6D,$8C
	.dc.b	$B6,$3F,$BB,$FC,$37,$F7,$52,$38,$24,$7F,$77,$8C,$B6,$32,$FF,$79
	.dc.b	$8C,$B6,$3F,$C0,$7E,$07,$F8,$04,$8E,$09,$1F,$E0,$38,$CB,$63,$2F
	.dc.b	$F8,$16,$32,$D8,$30,$BD,$0F,$9F,$4F,$F9,$45,$AE,$E7,$71,$87,$13
	.dc.b	$C6,$78,$10,$E1,$08,$D4,$61,$B1,$CA,$7A,$3E,$58,$CB,$63,$2F,$8A
	.dc.b	$7C,$DF,$0F,$1E,$92,$8C,$4F,$56,$BC,$DA,$FB,$1A,$F9,$E3,$94,$F4
	.dc.b	$9F,$89,$F0,$D2,$3B,$B6,$BD,$4D,$7D,$0D,$7E,$4B,$5E,$CD,$BF,$A5
	.dc.b	$F0,$DF,$44,$95,$EC,$6D,$7C,$76,$BF,$15,$AF,$D6,$6B,$EB,$7E,$E7
	.dc.b	$C3,$49,$F8,$ED,$9F,$A4,$D7,$E8,$35,$FC,$0B,$5F,$8C,$DA,$FC,$57
	.dc.b	$C3,$7E,$24,$95,$F5,$1B,$5F,$7D,$AF,$DD,$6B,$F8,$E6,$BF,$4D,$FC
	.dc.b	$A7,$C3,$49,$FB,$ED,$9F,$C5,$35,$FC,$43,$5F,$CC,$B5,$FB,$CD,$AF
	.dc.b	$D1,$78,$DF,$A1,$25,$7E,$31,$B5,$F9,$76,$BF,$95,$6B,$FA,$66,$BF
	.dc.b	$8B,$7F,$59,$E4,$C9,$FC,$BB,$67,$F4,$4D,$7F,$40,$D7,$F6,$2D,$7F
	.dc.b	$2C,$DA,$FD,$D7,$C3,$7E,$E4,$95,$FA,$46,$D7,$EB,$DA,$FE,$B5,$AF
	.dc.b	$EF,$9A,$FE,$8D,$FE,$09,$F0,$D2,$7F,$5E,$D9,$FD,$D3,$5F,$DC,$35
	.dc.b	$FE,$12,$D7,$F5,$CD,$AF,$E2,$5F,$0D,$FC,$44,$95,$FB,$C6,$D7,$F0
	.dc.b	$76,$BF,$C1,$5A,$FF,$16,$6B,$FB,$B7,$F8,$E7,$C3,$49,$FE,$0E,$D9
	.dc.b	$FE,$24,$D7,$F8,$83,$5F,$E4,$2D,$7F,$83,$36,$BF,$95,$7C,$37,$F2
	.dc.b	$92,$57,$F1,$46,$D7,$F1,$F6,$BF,$C7,$5A,$FF,$2E,$6B,$FC,$4D,$FE
	.dc.b	$69,$E8,$49,$FE,$3E,$D9,$FE,$54,$D7,$F9,$43,$5F,$E7,$2D,$7F,$8F
	.dc.b	$36,$BF,$A1,$7C,$37,$F4,$12,$57,$F2,$C6,$D7,$F3,$76,$BF,$CD,$5A
	.dc.b	$FF,$46,$6B,$FC,$AD,$FE,$99,$F0,$D2,$7F,$9B,$B6,$7F,$A1,$35,$FE
	.dc.b	$80,$D7,$FA,$8B,$5F,$E6,$CD,$AF,$EB,$5F,$0D,$FD,$64,$95,$FD,$11
	.dc.b	$B5,$FD,$3D,$AF,$F4,$D6,$BF,$D7,$9A,$FF,$43,$7F,$B2,$BF,$D8,$E1
	.dc.b	$F8,$93,$E1,$73,$3F,$DB,$67,$31,$88,$FF,$7F,$4E,$E7,$FB,$29,$A2
	.dc.b	$B9,$8E,$CD,$7A,$35,$E7,$3D,$5F,$EF,$31,$96,$C6,$67,$B4,$74,$8F
	.dc.b	$EF,$DF,$0F,$27,$FB,$4B,$67,$FB,$A3,$5F,$EE,$4D,$7F,$C0,$6D,$7F
	.dc.b	$B4,$37,$0F,$F0,$67,$C3,$FF,$82,$C9,$5F,$DD,$9B,$5F,$E0,$16,$BF
	.dc.b	$DF,$DA,$FF,$84,$1A,$FF,$75,$7F,$C2,$D7,$27,$FC,$02,$D9,$FF,$06
	.dc.b	$35,$FF,$05,$35,$FF,$0D,$B5,$FF,$00,$36,$9F,$AB,$BE,$23,$E7,$46
	.dc.b	$1B,$19,$BF,$F0,$E3,$6F,$FC,$4A,$D7,$FC,$38,$D7,$FC,$62,$D7,$FC
	.dc.b	$4C,$DA,$ED,$3D,$2B,$B9,$1E,$FF,$11,$E2,$23,$82,$5F,$8A,$32,$D8
	.dc.b	$CB,$E4,$8C,$B6,$3D,$7F,$11,$EE,$23,$82,$47,$BA,$32,$D8,$CB,$F0
	.dc.b	$23,$2D,$8F,$8F,$F1,$1F,$28,$8E,$09,$1F,$2A,$32,$D8,$CB,$F3,$23
	.dc.b	$2D,$8F,$A7,$F1,$1F,$58,$8E,$09,$1F,$5A,$32,$D8,$CB,$F6,$23,$2D
	.dc.b	$8F,$BF,$F1,$1F,$82,$23,$82,$47,$E0,$A3,$2D,$8C,$BF,$84,$8C,$B6
	.dc.b	$3F,$17,$F1,$1F,$8E,$23,$82,$47,$E3,$A3,$2D,$8C,$BF,$90,$8C,$B6
	.dc.b	$3F,$2F,$F1,$1F,$9A,$23,$82,$47,$E6,$A3,$2D,$8C,$BF,$9C,$8C,$B6
	.dc.b	$3F,$47,$F1,$1F,$A6,$23,$82,$47,$E9,$A3,$2D,$8C,$BF,$A8,$8C,$B6
	.dc.b	$3F,$5F,$F1,$1F,$B2,$23,$82,$47,$EC,$A3,$2D,$8C,$BF,$B4,$8C,$B6
	.dc.b	$3F,$77,$F1,$1F,$BE,$23,$82,$47,$EF,$A3,$2D,$8C,$BF,$C0,$23,$2D
	.dc.b	$8F,$F5,$2F,$C0,$BF,$47,$79,$3F,$CE,$E3,$0D,$8C,$CF,$06,$D7,$FA
	.dc.b	$D6,$B7,$39,$AE,$04,$70,$49,$E1,$5C,$9F,$F6,$25,$72,$7F,$D7,$3C
	.dc.b	$1F,$EA,$26,$F9,$1F,$D3,$D4,$65,$FF,$63,$35,$FF,$67,$35,$FF,$68
	.dc.b	$B5,$E3,$1C,$E3,$04,$70,$4B,$F8,$1E,$57,$C3,$23,$82,$57,$B2,$32
	.dc.b	$D8,$CB,$EE,$C1,$3A,$8C,$3F,$3F,$CB,$95,$F3,$DB,$3F,$21,$AF,$36
	.dc.b	$BF,$EE,$EC,$7E,$E4,$62,$11,$89,$E8,$7F,$48,$F8,$9F,$E9,$64,$70
	.dc.b	$49,$FE,$97,$19,$6C,$65,$FE,$9B,$19,$6C,$7F,$54,$F8,$9F,$EA,$E4
	.dc.b	$70,$48,$FE,$AF,$19,$6C,$65,$FE,$B3,$19,$6C,$7F,$60,$F8,$9F,$EC
	.dc.b	$64,$70,$48,$FE,$C7,$19,$6C,$65,$FE,$CB,$19,$6C,$7F,$6C,$F8,$9F
	.dc.b	$ED,$E4,$70,$48,$FE,$DF,$19,$6C,$65,$FE,$E3,$19,$6C,$7F,$78,$F8
	.dc.b	$9F,$EF,$64,$70,$48,$FE,$F7,$19,$6C,$65,$FE,$FB,$19,$6C,$7F,$81
	.dc.b	$3E,$1B,$FC,$0E,$47,$04,$8F,$F0,$3C,$65,$B1,$97,$FC,$13,$19,$6C
	.dc.b	$7F,$84,$3E,$1B,$FC,$26,$47,$04,$8F,$F0,$9C,$65,$B1,$97,$FC,$2B
	.dc.b	$19,$6C,$7F,$87,$3E,$1B,$FC,$3E,$47,$04,$8F,$F0,$FC,$65,$B1,$97
	.dc.b	$FC,$43,$19,$6C,$7F,$8A,$3E,$1B,$FC,$56,$47,$04,$8F,$F1,$5C,$65
	.dc.b	$B1,$97,$FC,$5B,$19,$6C,$7F,$8D,$3E,$1B,$FC,$6E,$47,$04,$8F,$F1
	.dc.b	$BC,$65,$B1,$97,$FC,$73,$19,$6C,$18,$5E,$F7,$EB,$19,$9C,$CE,$63
	.dc.b	$0E,$27,$6F,$F9,$3A,$32,$F4,$8C,$CF,$28,$46,$1E,$53,$C1,$EC,$73
	.dc.b	$24,$F1,$9E,$11,$D6,$78,$3D,$1E,$B3,$B6,$C7,$69,$F3,$12,$74,$D8
	.dc.b	$D4,$60,$8E,$08,$78,$7E,$1F,$E0,$12,$3C,$0D,$8F,$6B,$5E,$11,$A2
	.dc.b	$3A,$6B,$ED,$F8,$7F,$98,$48,$F6,$36,$3E,$4B,$5E,$11,$A2,$3A,$6B
	.dc.b	$F2,$7E,$1F,$EC,$12,$3E,$43,$63,$EA,$B5,$E1,$1A,$23,$A6,$BF,$57
	.dc.b	$E1,$FF,$08,$48,$FA,$8D,$8F,$C0,$B5,$E1,$1A,$23,$A6,$BF,$81,$F8
	.dc.b	$7F,$C8,$12,$3F,$00,$D8,$FC,$6B,$5E,$11,$A2,$3A,$6B,$F8,$DF,$87
	.dc.b	$FC,$E1,$23,$F1,$8D,$8F,$CC,$B5,$E1,$1A,$23,$A6,$BF,$99,$F8,$7F
	.dc.b	$D4,$12,$3F,$30,$D8,$FD,$2B,$5E,$11,$A2,$3A,$6B,$FA,$5F,$87,$FD
	.dc.b	$A1,$23,$F4,$8D,$8F,$D8,$B5,$E1,$1A,$23,$A6,$BF,$B1,$F8,$7F,$E0
	.dc.b	$04,$8F,$D8,$36,$3F,$7A,$D7,$84,$68,$8E,$9A,$FE,$F7,$E1,$FF,$86
	.dc.b	$12,$3F,$78,$D8,$FE,$12,$D7,$84,$68,$8E,$9A,$FF,$09,$F8,$7F,$E3
	.dc.b	$04,$8F,$E1,$0D,$8F,$E2,$AD,$78,$46,$88,$E9,$AF,$F1,$5F,$87,$FE
	.dc.b	$48,$48,$FE,$28,$D8,$FE,$42,$D7,$84,$68,$8E,$9A,$FF,$21,$F8,$7F
	.dc.b	$E6,$04,$8F,$E4,$0D,$8F,$E5,$AD,$78,$46,$88,$E9,$AF,$F2,$DF,$87
	.dc.b	$FE,$78,$48,$FE,$58,$D8,$FE,$72,$D7,$84,$68,$8E,$9A,$FF,$39,$F8
	.dc.b	$7F,$E9,$04,$8F,$E7,$0D,$8F,$E8,$AD,$78,$46,$88,$E9,$AF,$F4,$5F
	.dc.b	$87,$FE,$A8,$48,$FE,$88,$D8,$FE,$A2,$D7,$84,$68,$8E,$9A,$FF,$51
	.dc.b	$F8,$7F,$EC,$04,$8F,$EA,$0D,$8F,$EB,$AD,$78,$46,$88,$E9,$AF,$FA
	.dc.b	$53,$E0,$BF,$D1,$C4,$8F,$ED,$51,$FE,$9A,$8C,$3F,$EA,$06,$BC,$63
	.dc.b	$7E,$3A,$73,$3F,$E9,$E6,$A9,$7E,$58,$F3,$F1,$3F,$E0,$29,$CC,$62
	.dc.b	$3F,$C1,$31,$83,$FB,$F1,$5E,$13,$CE,$32,$D8,$CC,$F3,$7F,$C2,$11
	.dc.b	$97,$AB,$5C,$C6,$DF,$44,$65,$B1,$FE,$13,$F1,$A5,$7F,$80,$DB,$3F
	.dc.b	$E1,$06,$BF,$E0,$B6,$BF,$E1,$F6,$BF,$E0,$D6,$F5,$7F,$89,$BC,$7F
	.dc.b	$F1,$01,$2B,$FC,$20,$DA,$FF,$0E,$B5,$FF,$0C,$35,$FF,$16,$35,$FF
	.dc.b	$0D,$3F,$E2,$BF,$2A,$4F,$F8,$6D,$B3,$FE,$28,$6B,$FE,$23,$6B,$FE
	.dc.b	$37,$6B,$FE,$25,$6D,$3C,$2F,$C0,$57,$F8,$B5,$B5,$FE,$3D,$6B,$FE
	.dc.b	$3B,$6B,$FE,$47,$6B,$FE,$3C,$6C,$F5,$9F,$AC,$D7,$B1,$88,$D9,$1C
	.dc.b	$79,$51,$F2,$36,$BC,$8D,$7C,$8D,$7C,$8D,$7C,$8D,$AF,$81,$F0,$FE
	.dc.b	$12,$57,$C1,$6D,$7C,$16,$BF,$05,$AF,$C1,$6B,$F0,$5F,$67,$C3,$C9
	.dc.b	$F9,$8D,$9F,$98,$D7,$E6,$35,$F9,$8D,$7E,$63,$6B,$EC,$7C,$3F,$C9
	.dc.b	$25,$7D,$96,$D7,$D9,$6B,$F6,$5A,$FD,$96,$BF,$65,$FA,$9F,$0F,$27
	.dc.b	$F0,$8D,$9F,$C2,$35,$FC,$23,$5F,$C2,$35,$FC,$23,$6B,$F2,$1C,$1F
	.dc.b	$81,$25,$7E,$45,$B5,$F9,$16,$BF,$91,$6B,$F9,$16,$BF,$91,$7F,$19
	.dc.b	$F0,$F2,$7F,$38,$D9,$FC,$E3,$5F,$CE,$35,$FC,$E3,$5F,$CE,$36,$BF
	.dc.b	$51,$F0,$FF,$99,$25,$7E,$A5,$B5,$FA,$96,$BF,$A9,$6B,$FA,$96,$BF
	.dc.b	$A9,$7F,$8E,$79,$FD,$A7,$FD,$50,$D9,$FF,$59,$B5,$FF,$5C,$B5,$FF
	.dc.b	$5E,$35,$34,$9B,$2B,$BB,$6B,$BB,$6B,$BB,$5E,$ED,$7F,$D9,$3C,$95
	.dc.b	$FF,$66,$35,$34,$95,$FC,$E6,$32,$D8,$CD,$7F,$58,$8C,$11,$C1,$FD
	.dc.b	$73,$DB,$E0,$47,$04,$AF,$F6,$85,$27,$51,$86,$BF,$ED,$E8,$D4,$F1
	.dc.b	$8C,$4F,$9F,$EB,$9E,$8D,$25,$7F,$74,$8C,$B6,$33,$5F,$EE,$19,$CB
	.dc.b	$63,$2F,$FB,$BA,$32,$D8,$CB,$F7,$31,$7C,$47,$D8,$25,$FD,$86,$CF
	.dc.b	$D8,$6B,$F6,$1A,$F3,$6B,$F8,$69,$EB,$3F,$EF,$7F,$94,$7D,$73,$DA
	.dc.b	$FE,$E1,$18,$6C,$66,$78,$36,$C4,$15,$C6,$77,$5C,$88,$E0,$93,$C4
	.dc.b	$E3,$5D,$5E,$95,$CD,$FE,$B1,$87,$F4,$D2,$57,$F5,$D6,$C6,$A3,$04
	.dc.b	$70,$43,$FB,$04,$70,$7F,$B1,$7C,$57,$F5,$92,$4F,$F6,$96,$C7,$F6
	.dc.b	$96,$BC,$23,$44,$74,$D7,$F5,$B1,$FB,$42,$47,$ED,$1B,$1F,$B4,$6B
	.dc.b	$C2,$34,$47,$4D,$7F,$1D,$EA,$7E,$6C,$91,$F9,$96,$C7,$E5,$DA,$F0
	.dc.b	$8D,$11,$D3,$5F,$AD,$F5,$EF,$33,$F7,$CD,$9F,$F2,$CB,$67,$EE,$07
	.dc.b	$0F,$29,$E2,$49,$E2,$D8,$E3,$3C,$1F,$C7,$35,$FC,$8B,$5E,$DE,$0F
	.dc.b	$90,$95,$E4,$6C,$74,$6B,$EC,$6B,$D9,$AF,$67,$F9,$CF,$E9,$0F,$7A
	.dc.b	$92,$BF,$9D,$C6,$5B,$19,$AF,$E7,$B3,$96,$C6,$5F,$E7,$F1,$96,$C6
	.dc.b	$5F,$E8,$30,$F7,$FC,$64,$AF,$0B,$67,$BB,$5E,$AD,$79,$B5,$FE,$97
	.dc.b	$3D,$67,$FD,$27,$F6,$CF,$57,$F9,$E3,$63,$31,$82,$38,$25,$7E,$0A
	.dc.b	$78,$36,$76,$D9,$DB,$FC,$0F,$E0,$5D,$BF,$D2,$DB,$3F,$EA,$66,$CF
	.dc.b	$E3,$9A,$F1,$9E,$04,$9E,$0D,$8E,$0D,$78,$47,$FA,$CB,$EE,$1E,$EF
	.dc.b	$F5,$B6,$C6,$63,$04,$70,$49,$FD,$0C,$F0,$6C,$ED,$B3,$B7,$F9,$17
	.dc.b	$DC,$BC,$7F,$ED,$4D,$9F,$F6,$2B,$67,$F5,$AD,$78,$CF,$02,$4F,$06
	.dc.b	$C7,$06,$BC,$23,$F9,$97,$D0,$88,$00

;--------------------------------------------------------------------------------
;	FMOVE.X FPn,<mem>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovexregto_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.X FPN,<EA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.X FPn,<ea>',13,10
;------------------------------------------------
;	d1-d3	actual result
;	d4	actual status
;	d6	fpcr=(XRN..DRP)<<4
;	d7	0=failed,1=successful
;	a0	expected result handle,expected status,...
;	a1-a3	expected result
;	a4	expected status
;	a5	source handle,...
;	fp5	source
;------------------------------------------------
	lea.l	push_decompressed,a3
;decompress data
	move.l	a3,-(sp)
	pea.l	fmovexregto_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a3,a5			;source handle,...
@@:
	add.l	d0,(a3)+		;source handle
	tst.l	(a3)
	bpl	@b
	addq.l	#4,a3			;-1
	movea.l	a3,a0			;expected result handle,expected status,...
@@:
	add.l	d0,(a3)+		;expected result handle
	addq.l	#4,a3			;expected status
	tst.l	(a3)
	bpl	@b
;	addq.l	#4,a3			;-1
;
55:
	move.l	#0,d6		;fpcr=(XRN..DRP)<<4
66:
;FPn,<mem>
	fmove.l	#0,fpcr
	fmove.x	([a5]),fp5		;source
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
	move.l	#-1,-(sp)		;fill=-1
	move.l	#-1,-(sp)
	move.l	#-1,-(sp)
					;source
	fmove.x	fp5,(sp)		;EXECUTE
	movem.l	(sp)+,d1-d3		;actual result
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
;
	movem.l	([a0]),a1-a3		;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1-d4/a1-a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_packed
	lea.l	32(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.X FPn='
	puthex24	([a5]),(4,[a5]),(8,[a5])	;source
	putmsg	',<mem> @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d4/a1-a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_packed
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a0			;expected
;
	add.w	#1<<4,d6		;rprm++
	cmp.w	#176,d6		;fpcr=(XRN..DRP)<<4
	bls	66b
;
	addq.l	#4,a5			;source handle,...
	tst.l	(a5)			;source handle,...
	bpl	55b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmovexregto_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$BA,$42,$10,$84,$21,$08,$42,$7A,$9C,$48,$F4,$6B
	.dc.b	$DD,$AF,$89,$AF,$E4,$1A,$FF,$32,$7C,$9E,$A4,$8F,$3B,$5F,$53,$5F
	.dc.b	$6B,$5F,$CD,$B5,$FE,$7A,$D7,$FA,$9F,$A9,$F0,$48,$FC,$46,$BF,$1D
	.dc.b	$AF,$CA,$6B,$FA,$86,$BF,$D2,$9F,$99,$EA,$48,$F9,$ED,$7E,$93,$5F
	.dc.b	$AA,$D7,$F6,$6D,$7F,$AA,$B5,$FE,$D7,$EA,$7D,$92,$3F,$71,$AF,$DF
	.dc.b	$6B,$F8,$26,$BF,$E9,$83,$81,$0F,$C3,$7A,$9F,$85,$23,$F8,$86,$BA
	.dc.b	$30,$43,$04,$3F,$22,$F0,$7F,$27,$EA,$7E,$44,$8F,$E5,$1A,$FE,$5D
	.dc.b	$AF,$03,$44,$34,$53,$F3,$FE,$A7,$E7,$48,$FE,$81,$AF,$E8,$DA,$F0
	.dc.b	$34,$43,$45,$3F,$57,$EA,$7E,$A4,$8F,$EB,$1A,$FE,$BD,$AF,$03,$44
	.dc.b	$34,$53,$F6,$FE,$A7,$ED,$48,$FE,$E1,$AF,$EE,$DA,$F0,$34,$43,$45
	.dc.b	$3F,$81,$FA,$9F,$C0,$48,$FF,$04,$6B,$FC,$1D,$AF,$03,$44,$34,$53
	.dc.b	$F8,$7F,$A9,$FC,$34,$8F,$F1,$06,$BF,$C4,$DA,$F0,$34,$43,$45,$3F
	.dc.b	$8D,$FA,$9F,$C6,$48,$FF,$1C,$6B,$FC,$7D,$AF,$03,$44,$34,$53,$F9
	.dc.b	$3F,$A9,$FC,$94,$8F,$F2,$86,$BF,$CA,$DA,$F0,$34,$43,$45,$3F,$99
	.dc.b	$FA,$9F,$CC,$48,$FF,$34,$6B,$FC,$DD,$AF,$03,$44,$34,$53,$F9,$FF
	.dc.b	$A9,$FC,$F4,$8F,$F4,$06,$BF,$D0,$DA,$F0,$34,$43,$45,$3F,$A5,$FA
	.dc.b	$9F,$D2,$48,$FF,$4C,$6B,$FD,$3D,$AF,$03,$44,$34,$53,$FA,$BF,$A9
	.dc.b	$FD,$54,$8F,$F5,$86,$BF,$D6,$DA,$F0,$34,$43,$45,$3F,$B1,$FA,$9F
	.dc.b	$D8,$48,$FF,$64,$6B,$FD,$9D,$AF,$03,$44,$34,$53,$FB,$7F,$A9,$FD
	.dc.b	$B4,$8F,$F7,$06,$BF,$DC,$DA,$F0,$34,$43,$45,$3F,$BD,$FA,$9F,$DE
	.dc.b	$48,$FF,$7C,$6B,$FD,$FD,$AF,$03,$44,$34,$53,$FC,$87,$F7,$5F,$EB
	.dc.b	$B6,$FC,$6F,$F8,$22,$31,$1F,$EA,$86,$BF,$EB,$E6,$A6,$92,$39,$12
	.dc.b	$3F,$C2,$91,$96,$FC,$5F,$72,$38,$21,$FE,$1C,$8F,$F4,$3B,$E8,$8D
	.dc.b	$B9,$6B,$93,$89,$EE,$F8,$4F,$61,$1F,$1B,$5E,$CD,$78,$1A,$21,$A2
	.dc.b	$9F,$2B,$D7,$F9,$04,$7D,$ED,$7D,$4D,$78,$1A,$21,$A2,$9F,$5B,$D7
	.dc.b	$FA,$84,$7E,$5B,$5F,$8C,$D7,$81,$A2,$1A,$29,$F8,$2F,$5F,$F0,$04
	.dc.b	$7E,$BB,$5F,$A4,$D7,$81,$A2,$1A,$29,$F8,$EF,$5F,$F1,$84,$7F,$06
	.dc.b	$D7,$EF,$35,$E0,$68,$86,$8A,$7E,$6B,$D7,$FC,$C1,$1F,$C7,$B5,$FC
	.dc.b	$53,$5E,$06,$88,$68,$A7,$E9,$BD,$7F,$D2,$11,$FC,$DB,$5F,$CB,$35
	.dc.b	$E0,$68,$86,$8A,$7E,$CB,$D7,$FD,$81,$1F,$D3,$B5,$FD,$13,$5E,$06
	.dc.b	$88,$68,$A7,$EF,$BD,$7F,$DE,$11,$FD,$9B,$5F,$D7,$35,$E0,$68,$86
	.dc.b	$8A,$7F,$0A,$F5,$FF,$84,$11,$FD,$FB,$5F,$DD,$35,$E0,$68,$86,$8A
	.dc.b	$7F,$16,$F5,$FF,$8A,$11,$FE,$16,$D7,$F8,$33,$5E,$06,$88,$68,$A7
	.dc.b	$F2,$2F,$5F,$F9,$01,$1F,$E2,$ED,$7F,$89,$35,$E0,$68,$86,$8A,$7F
	.dc.b	$2E,$F5,$FF,$96,$11,$FE,$46,$D7,$F8,$F3,$5E,$06,$88,$68,$A7,$F3
	.dc.b	$AF,$5F,$F9,$C1,$1F,$E5,$ED,$7F,$95,$35,$E0,$68,$86,$8A,$7F,$46
	.dc.b	$F5,$FF,$A2,$11,$FE,$76,$D7,$F9,$B3,$5E,$06,$88,$68,$A7,$F5,$2F
	.dc.b	$5F,$FA,$81,$1F,$E8,$ED,$7F,$A1,$35,$E0,$68,$86,$8A,$7F,$5E,$F5
	.dc.b	$FF,$AE,$11,$FE,$A6,$D7,$FA,$73,$5E,$06,$88,$68,$A7,$F6,$AF,$5F
	.dc.b	$FB,$41,$1F,$EB,$ED,$7F,$AD,$35,$E0,$68,$86,$8A,$7F,$76,$F5,$FF
	.dc.b	$BA,$11,$FE,$D6,$D7,$FB,$33,$5E,$06,$88,$68,$A0,$83,$EE,$9F,$13
	.dc.b	$88,$91,$87,$07,$17,$0E,$1A,$F0,$8D,$BB,$6B,$DD,$B1,$C2,$3C,$6D
	.dc.b	$78,$1C,$9C,$3C,$4C,$7C,$4F,$22,$38,$21,$E5,$72,$D3,$27,$B9,$CB
	.dc.b	$4F,$07,$B3,$E6,$23,$82,$1F,$05,$CB,$4C,$9F,$29,$CB,$4F,$67,$B3
	.dc.b	$F0,$88,$E0,$87,$CD,$72,$D3,$27,$D6,$72,$D3,$E4,$7B,$3F,$38,$8E
	.dc.b	$08,$7D,$97,$2D,$32,$7E,$09,$CB,$4F,$A9,$EC,$FD,$A2,$38,$21,$F8
	.dc.b	$57,$2D,$32,$7E,$39,$CB,$4F,$C0,$7B,$3F,$86,$23,$82,$1F,$91,$72
	.dc.b	$D3,$27,$E6,$9C,$B4,$FC,$67,$B3,$F9,$22,$38,$21,$F9,$D7,$2D,$32
	.dc.b	$7E,$99,$CB,$4F,$CC,$7B,$3F,$9E,$23,$82,$1F,$A9,$72,$D3,$27,$EC
	.dc.b	$9C,$B4,$FD,$27,$B3,$FA,$A2,$38,$21,$FB,$57,$2D,$32,$7E,$F9,$CB
	.dc.b	$4F,$D8,$7C,$37,$ED,$88,$E0,$87,$F0,$17,$2D,$32,$7F,$0A,$72,$D3
	.dc.b	$F7,$9F,$0D,$FC,$08,$8E,$08,$7F,$0D,$72,$D3,$27,$F1,$67,$2D,$3F
	.dc.b	$84,$7C,$37,$F0,$E2,$38,$21,$FC,$65,$CB,$4C,$9F,$C8,$9C,$B4,$FE
	.dc.b	$29,$F0,$DF,$C6,$88,$E0,$87,$F2,$57,$2D,$32,$7F,$2E,$72,$D3,$F9
	.dc.b	$07,$C3,$7F,$26,$23,$82,$1F,$CC,$5C,$B4,$C9,$FC,$E9,$CB,$4F,$E5
	.dc.b	$9F,$0D,$FC,$C8,$8E,$08,$7F,$3D,$72,$D3,$27,$F4,$67,$2D,$3F,$9C
	.dc.b	$7C,$37,$F3,$E2,$38,$21,$FD,$25,$CB,$4C,$9F,$D4,$9C,$B4,$FE,$89
	.dc.b	$F0,$DF,$D2,$88,$E0,$87,$F5,$57,$2D,$32,$7F,$5E,$72,$D3,$FA,$87
	.dc.b	$C3,$7F,$56,$23,$82,$1F,$D8,$5C,$B4,$C9,$FD,$A9,$CB,$4F,$EB,$9F
	.dc.b	$0D,$FD,$88,$8E,$08,$7F,$6D,$72,$D3,$27,$F7,$67,$2D,$3F,$B4,$7C
	.dc.b	$37,$F6,$E2,$38,$21,$FD,$E5,$CB,$4C,$9F,$E0,$27,$2D,$3F,$BE,$FD
	.dc.b	$D7,$F7,$C4,$D2,$3F,$E0,$06,$BC,$0C,$10,$C1,$0F,$F0,$7B,$CF,$E1
	.dc.b	$39,$12,$3C,$2D,$76,$D7,$FC,$35,$08,$C4,$7F,$88,$23,$0F,$22,$1E
	.dc.b	$0F,$84,$F4,$11,$F7,$B5,$D1,$82,$18,$21,$E2,$78,$3E,$4F,$84,$F8
	.dc.b	$84,$7E,$5B,$5F,$82,$D7,$81,$A2,$1A,$29,$F0,$3E,$13,$E8,$11,$FA
	.dc.b	$ED,$7E,$6B,$5E,$06,$88,$68,$A7,$CC,$F8,$4F,$B8,$47,$F0,$6D,$7E
	.dc.b	$CB,$5E,$06,$88,$68,$A7,$D8,$F8,$4F,$C4,$11,$FC,$7B,$5F,$C2,$B5
	.dc.b	$E0,$68,$86,$8A,$7E,$13,$E1,$3F,$28,$47,$F3,$6D,$7F,$22,$D7,$81
	.dc.b	$A2,$1A,$29,$F9,$0F,$84,$FD,$01,$1F,$D3,$B5,$FC,$EB,$5E,$06,$88
	.dc.b	$68,$A7,$E7,$3E,$13,$F5,$84,$7F,$66,$D7,$F5,$2D,$78,$1A,$21,$A2
	.dc.b	$9F,$A8,$F8,$4F,$DC,$11,$FD,$FB,$5F,$DA,$B5,$E0,$68,$86,$8A,$7E
	.dc.b	$D3,$E1,$3F,$82,$11,$FE,$16,$D7,$F8,$0B,$5E,$06,$88,$68,$A7,$F0
	.dc.b	$0F,$84,$FE,$20,$47,$F8,$BB,$5F,$E1,$AD,$78,$1A,$21,$A2,$9F,$C3
	.dc.b	$3E,$13,$F8,$E1,$1F,$E4,$6D,$7F,$8C,$B5,$E0,$68,$86,$8A,$7F,$18
	.dc.b	$F8,$4F,$E5,$04,$7F,$97,$B5,$FE,$4A,$D7,$81,$A2,$1A,$29,$FC,$93
	.dc.b	$E1,$3F,$9A,$11,$FE,$76,$D7,$F9,$8B,$5E,$06,$88,$68,$A7,$F3,$0F
	.dc.b	$84,$FE,$80,$47,$FA,$3B,$5F,$E7,$AD,$78,$1A,$21,$A2,$9F,$CF,$3E
	.dc.b	$13,$FA,$61,$1F,$EA,$6D,$7F,$A4,$B5,$E0,$68,$86,$8A,$7F,$48,$F8
	.dc.b	$4F,$EB,$04,$7F,$AF,$B5,$FE,$AA,$D7,$81,$A2,$1A,$29,$FD,$53,$E1
	.dc.b	$3F,$B2,$11,$FE,$D6,$D7,$FB,$0B,$5E,$06,$88,$68,$A7,$F6,$0F,$84
	.dc.b	$FE,$E0,$47,$FB,$BB,$5F,$ED,$AD,$78,$1A,$21,$A2,$9F,$DB,$3E,$13
	.dc.b	$FB,$E1,$1F,$F0,$1B,$5F,$EF,$2D,$78,$1A,$21,$A2,$82,$0F,$39,$F0
	.dc.b	$98,$89,$18,$70,$71,$70,$E1,$AF,$08,$DB,$B6,$BD,$DB,$1C,$23,$C6
	.dc.b	$D7,$81,$C9,$C3,$C4,$C7,$C1,$F2,$23,$82,$1E,$57,$2D,$32,$7B,$9C
	.dc.b	$B4,$F0,$7C,$1F,$98,$8E,$08,$7C,$17,$2D,$32,$7C,$A7,$2D,$3D,$9F
	.dc.b	$07,$F0,$88,$E0,$87,$CD,$72,$D3,$27,$D6,$72,$D3,$E4,$7C,$1F,$CE
	.dc.b	$23,$82,$1F,$65,$CB,$4C,$9F,$82,$72,$D3,$EA,$7C,$1F,$DA,$23,$82
	.dc.b	$1F,$85,$72,$D3,$27,$E3,$9C,$B4,$FC,$07,$C1,$FE,$18,$8E,$08,$7E
	.dc.b	$45,$CB,$4C,$9F,$9A,$72,$D3,$F1,$9F,$07,$F9,$22,$38,$21,$F9,$D7
	.dc.b	$2D,$32,$7E,$99,$CB,$4F,$CC,$7C,$1F,$E7,$88,$E0,$87,$EA,$5C,$B4
	.dc.b	$C9,$FB,$27,$2D,$3F,$49,$F0,$7F,$AA,$23,$82,$1F,$B5,$72,$D3,$27
	.dc.b	$EF,$9C,$B4,$FD,$87,$C1,$FE,$D8,$8E,$08,$7F,$01,$72,$D3,$27,$F0
	.dc.b	$A7,$2D,$3F,$79,$F0,$7F,$C0,$88,$E0,$87,$F0,$D7,$2D,$32,$7F,$16
	.dc.b	$72,$D3,$F8,$47,$C1,$FF,$0E,$23,$82,$1F,$C6,$5C,$B4,$C9,$FC,$89
	.dc.b	$CB,$4F,$E2,$9F,$07,$FC,$68,$8E,$08,$7F,$25,$72,$D3,$27,$F2,$E7
	.dc.b	$2D,$3F,$90,$7C,$1F,$F2,$62,$38,$21,$FC,$C5,$CB,$4C,$9F,$CE,$9C
	.dc.b	$B4,$FE,$59,$F0,$7F,$CC,$88,$E0,$87,$F3,$D7,$2D,$32,$7F,$46,$72
	.dc.b	$D3,$F9,$C7,$C1,$FF,$3E,$23,$82,$1F,$D2,$5C,$B4,$C9,$FD,$49,$CB
	.dc.b	$4F,$E8,$9F,$07,$FD,$28,$8E,$08,$7F,$55,$72,$D3,$27,$F5,$E7,$2D
	.dc.b	$3F,$A8,$7C,$1F,$F5,$62,$38,$21,$FD,$85,$CB,$4C,$9F,$DA,$9C,$B4
	.dc.b	$FE,$B9,$F0,$7F,$D8,$88,$E0,$87,$F6,$D7,$2D,$32,$7F,$76,$72,$D3
	.dc.b	$FB,$47,$C1,$FF,$6E,$23,$82,$1F,$DE,$5C,$B4,$C9,$FE,$02,$72,$D3
	.dc.b	$FB,$EF,$DD,$7F,$82,$5B,$D1,$FD,$FA,$31,$1F,$E0,$86,$BF,$E0,$D6
	.dc.b	$BF,$E1,$26,$BF,$E1,$56,$A6,$92,$3B,$12,$3C,$7F,$07,$E1,$23,$82
	.dc.b	$1E,$37,$2D,$32,$79,$5C,$B4,$F7,$FC,$1F,$B4,$8E,$08,$7B,$DC,$B4
	.dc.b	$C9,$F0,$5C,$B4,$F9,$7F,$07,$F2,$48,$E0,$87,$CB,$72,$D3,$27,$CD
	.dc.b	$72,$D3,$EB,$FC,$1F,$D5,$23,$82,$1F,$5D,$CB,$4C,$9F,$65,$CB,$4F
	.dc.b	$C1,$FC,$1F,$E0,$48,$E0,$87,$E0,$DC,$B4,$C9,$F8,$57,$2D,$3F,$1F
	.dc.b	$F0,$7F,$8D,$23,$82,$1F,$8F,$72,$D3,$27,$E4,$5C,$B4,$FC,$DF,$C1
	.dc.b	$FE,$64,$8E,$08,$7E,$6D,$CB,$4C,$9F,$9D,$72,$D3,$F4,$FF,$07,$FA
	.dc.b	$52,$38,$21,$FA,$77,$2D,$32,$7E,$A5,$CB,$4F,$D9,$FC,$1F,$EC,$48
	.dc.b	$E0,$87,$EC,$DC,$B4,$C9,$FB,$57,$2D,$3F,$7F,$F0,$7F,$BD,$23,$82
	.dc.b	$1F,$BF,$72,$D3,$27,$F0,$17,$2D,$3F,$85,$FC,$1F,$F0,$92,$38,$21
	.dc.b	$FC,$2D,$CB,$4C,$9F,$C3,$5C,$B4,$FE,$2F,$F0,$7F,$C5,$48,$E0,$87
	.dc.b	$F1,$77,$2D,$32,$7F,$19,$72,$D3,$F9,$1F,$C1,$FF,$21,$23,$82,$1F
	.dc.b	$C8,$DC,$B4,$C9,$FC,$95,$CB,$4F,$E5,$FF,$07,$FC,$B4,$8E,$08,$7F
	.dc.b	$2F,$72,$D3,$27,$F3,$17,$2D,$3F,$9D,$FC,$1F,$F3,$92,$38,$21,$FC
	.dc.b	$ED,$CB,$4C,$9F,$CF,$5C,$B4,$FE,$8F,$F0,$7F,$D1,$48,$E0,$87,$F4
	.dc.b	$77,$2D,$32,$7F,$49,$72,$D3,$FA,$9F,$C1,$FF,$51,$23,$82,$1F,$D4
	.dc.b	$DC,$B4,$C9,$FD,$55,$CB,$4F,$EB,$FF,$07,$FD,$74,$8E,$08,$7F,$5F
	.dc.b	$72,$D3,$27,$F6,$17,$2D,$3F,$B5,$FC,$1F,$F6,$92,$38,$21,$FD,$AD
	.dc.b	$CB,$4C,$9F,$DB,$5C,$B4,$FE,$EF,$F0,$7F,$DD,$48,$E0,$87,$F7,$77
	.dc.b	$2D,$32,$7F,$79,$72,$D3,$FD,$FF,$F7,$5F,$EF,$E4,$D2,$23,$65,$21
	.dc.b	$97,$6E,$88,$68,$EA,$53,$81,$4C,$9E,$4F,$84,$C4,$1A,$48,$D1,$0E
	.dc.b	$D1,$B7,$07,$99,$C3,$A8,$E4,$D3,$E1,$FC,$1F,$80,$8E,$08,$76,$72
	.dc.b	$D3,$27,$77,$2D,$3E,$7F,$C1,$FB,$08,$E0,$87,$A9,$CB,$4C,$9E,$B7
	.dc.b	$2D,$3E,$DF,$C1,$FC,$82,$38,$21,$F1,$9C,$B4,$C9,$F1,$DC,$B4,$FC
	.dc.b	$3F,$C1,$FD,$42,$38,$21,$F4,$9C,$B4,$C9,$F4,$DC,$B4,$FC,$9F,$C1
	.dc.b	$FE,$00,$8E,$08,$7D,$E7,$2D,$32,$7D,$F7,$2D,$3F,$3F,$F0,$7F,$8C
	.dc.b	$23,$82,$1F,$8A,$72,$D3,$27,$E2,$DC,$B4,$FD,$5F,$C1,$FE,$60,$8E
	.dc.b	$08,$7E,$59,$CB,$4C,$9F,$97,$72,$D3,$F6,$FF,$07,$FA,$42,$38,$21
	.dc.b	$FA,$27,$2D,$32,$7E,$8D,$CB,$4F,$E0,$7F,$07,$FB,$02,$38,$21,$FA
	.dc.b	$E7,$2D,$32,$7E,$BD,$CB,$4F,$E1,$FF,$07,$FB,$C2,$38,$21,$FB,$A7
	.dc.b	$2D,$32,$7E,$ED,$CB,$4F,$E3,$7F,$07,$FC,$20,$8E,$08,$7F,$06,$72
	.dc.b	$D3,$27,$F0,$77,$2D,$3F,$93,$FC,$1F,$F1,$42,$38,$21,$FC,$49,$CB
	.dc.b	$4C,$9F,$C4,$DC,$B4,$FE,$67,$F0,$7F,$C8,$08,$E0,$87,$F1,$E7,$2D
	.dc.b	$32,$7F,$1F,$72,$D3,$F9,$FF,$C1,$FF,$2C,$23,$82,$1F,$CA,$9C,$B4
	.dc.b	$C9,$FC,$AD,$CB,$4F,$E9,$7F,$07,$FC,$E0,$8E,$08,$7F,$36,$72,$D3
	.dc.b	$27,$F3,$77,$2D,$3F,$AB,$FC,$1F,$F4,$42,$38,$21,$FD,$09,$CB,$4C
	.dc.b	$9F,$D0,$DC,$B4,$FE,$C7,$F0,$7F,$D4,$08,$E0,$87,$F4,$E7,$2D,$32
	.dc.b	$7F,$4F,$72,$D3,$FB,$7F,$C1,$FF,$5C,$23,$82,$1F,$D6,$9C,$B4,$C9
	.dc.b	$FD,$6D,$CB,$4F,$EF,$7F,$07,$FD,$A0,$8E,$08,$7F,$66,$72,$D3,$27
	.dc.b	$F6,$77,$2D,$3F,$C0,$FF,$07,$FD,$D0,$8E,$08,$7F,$72,$72,$D3,$27
	.dc.b	$F7,$37,$2D,$3F,$C0,$9F,$74,$F8,$4F,$F0,$23,$63,$31,$82,$38,$21
	.dc.b	$FE,$0F,$8E,$0D,$8D,$B4,$D9,$FE,$1A,$87,$C2,$78,$09,$1E,$46,$BD
	.dc.b	$5A,$F1,$6B,$FE,$25,$8E,$6F,$12,$1E,$5F,$84,$F6,$91,$F8,$2D,$74
	.dc.b	$60,$86,$08,$7A,$5E,$0F,$AF,$E1,$3E,$49,$1F,$9A,$D7,$E3,$35,$E0
	.dc.b	$68,$86,$8A,$7C,$7F,$84,$FA,$A4,$7E,$CB,$5F,$A4,$D7,$81,$A2,$1A
	.dc.b	$29,$F4,$FE,$13,$F0,$24,$7F,$0A,$D7,$EF,$35,$E0,$68,$86,$8A,$7D
	.dc.b	$FF,$84,$FC,$69,$1F,$C8,$B5,$FC,$53,$5E,$06,$88,$68,$A7,$E2,$FE
	.dc.b	$13,$F3,$24,$7F,$3A,$D7,$F2,$CD,$78,$1A,$21,$A2,$9F,$97,$F8,$4F
	.dc.b	$D2,$91,$FD,$4B,$5F,$D1,$35,$E0,$68,$86,$8A,$7E,$8F,$E1,$3F,$62
	.dc.b	$47,$F6,$AD,$7F,$5C,$D7,$81,$A2,$1A,$29,$FA,$FF,$84,$FD,$E9,$1F
	.dc.b	$E0,$2D,$7F,$74,$D7,$81,$A2,$1A,$29,$FB,$BF,$84,$FE,$12,$47,$F8
	.dc.b	$6B,$5F,$E0,$CD,$78,$1A,$21,$A2,$9F,$C1,$FE,$13,$F8,$A9,$1F,$E3
	.dc.b	$2D,$7F,$89,$35,$E0,$68,$86,$8A,$7F,$13,$F8,$4F,$E4,$24,$7F,$92
	.dc.b	$B5,$FE,$3C,$D7,$81,$A2,$1A,$29,$FC,$7F,$E1,$3F,$96,$91,$FE,$62
	.dc.b	$D7,$F9,$53,$5E,$06,$88,$68,$A7,$F2,$BF,$84,$FE,$72,$47,$F9,$EB
	.dc.b	$5F,$E6,$CD,$78,$1A,$21,$A2,$9F,$CD,$FE,$13,$FA,$29,$1F,$E9,$2D
	.dc.b	$7F,$A1,$35,$E0,$68,$86,$8A,$7F,$43,$F8,$4F,$EA,$24,$7F,$AA,$B5
	.dc.b	$FE,$9C,$D7,$81,$A2,$1A,$29,$FD,$3F,$E1,$3F,$AE,$91,$FE,$C2,$D7
	.dc.b	$FA,$D3,$5E,$06,$88,$68,$A7,$F5,$BF,$84,$FE,$D2,$47,$FB,$6B,$5F
	.dc.b	$EC,$CD,$78,$1A,$21,$A2,$9F,$D9,$FE,$13,$FB,$A9,$1F,$EF,$2D,$7F
	.dc.b	$B9,$35,$E0,$68,$86,$8A,$08,$7F,$80,$FE,$0E,$B3,$A5,$18,$8D,$C6
	.dc.b	$1C,$1C,$4A,$E5,$E0,$D7,$A4,$70,$78,$1A,$71,$1B,$8E,$F1,$83,$3F
	.dc.b	$07,$23,$BB,$5E,$ED,$7D,$4D,$7D,$2D,$7E,$14,$A3,$11,$FE,$25,$F8
	.dc.b	$3F,$19,$1F,$6B,$5F,$6B,$5F,$90,$D7,$E3,$B5,$EE,$FB,$BE,$0E,$47
	.dc.b	$C8,$6B,$F2,$1A,$FD,$36,$BF,$49,$AF,$A9,$AF,$F8,$F7,$E0,$FE,$59
	.dc.b	$1F,$AA,$D7,$EA,$B5,$FC,$03,$5F,$BE,$D7,$E3,$BF,$5B,$E0,$E4,$7E
	.dc.b	$01,$AF,$E0,$1A,$FE,$2D,$AF,$E2,$9A,$FD,$26,$BF,$E5,$5F,$83,$FC
	.dc.b	$19,$1F,$C6,$B5,$FC,$6B,$5F,$CC,$35,$FC,$BB,$5F,$BE,$FE,$3B,$E0
	.dc.b	$E4,$7E,$61,$AF,$E6,$1A,$FE,$8D,$AF,$E8,$9A,$FE,$29,$AF,$F9,$B7
	.dc.b	$E0,$FF,$36,$47,$F4,$AD,$7F,$4A,$D7,$F6,$0D,$7F,$5E,$D7,$F2,$EF
	.dc.b	$E9,$BE,$0E,$47,$EC,$1A,$FE,$C1,$AF,$EE,$DA,$FE,$E9,$AF,$E8,$9A
	.dc.b	$FF,$A1,$7E,$0F,$F6,$64,$7F,$7A,$D7,$F7,$AD,$7F,$84,$35,$FE,$0E
	.dc.b	$D7,$F5,$EF,$EF,$BE,$0E,$47,$F0,$86,$BF,$C2,$1A,$FF,$13,$6B,$FC
	.dc.b	$49,$AF,$EE,$9A,$FF,$A7,$7E,$0F,$F8,$59,$1F,$E2,$AD,$7F,$8A,$B5
	.dc.b	$FE,$40,$D7,$F8,$FB,$5F,$E0,$EF,$F1,$6F,$83,$91,$FC,$81,$AF,$F2
	.dc.b	$06,$BF,$CA,$DA,$FF,$2A,$6B,$FC,$49,$AF,$FA,$D7,$E0,$FF,$91,$91
	.dc.b	$FE,$5A,$D7,$F9,$6B,$5F,$E7,$0D,$7F,$9B,$B5,$FE,$3E,$FF,$2E,$F8
	.dc.b	$39,$1F,$CE,$1A,$FF,$38,$6B,$FD,$0D,$AF,$F4,$26,$BF,$CA,$9A,$FF
	.dc.b	$B3,$7E,$0F,$F9,$D9,$1F,$E8,$AD,$7F,$A2,$B5,$FE,$A0,$D7,$FA,$7B
	.dc.b	$5F,$E6,$EF,$F4,$6F,$83,$91,$FD,$41,$AF,$F5,$06,$BF,$D6,$DA,$FF
	.dc.b	$5A,$6B,$FD,$09,$AF,$FB,$97,$E0,$FF,$A9,$91,$FE,$BA,$D7,$FA,$EB
	.dc.b	$5F,$ED,$0D,$7F,$B3,$B5,$FE,$9E,$FF,$5E,$F8,$39,$1F,$DA,$1A,$FF
	.dc.b	$68,$6B,$FD,$CD,$AF,$F7,$26,$BF,$D6,$9A,$FF,$BF,$7E,$0F,$FB,$59
	.dc.b	$1F,$EE,$AD,$7F,$BA,$B5,$FF,$00,$35,$FE,$FE,$D7,$FB,$3B,$FD,$E7
	.dc.b	$EE,$BF,$BD,$41,$A4,$74,$43,$FC,$0E,$E5,$A6,$4F,$F0,$73,$96,$FC
	.dc.b	$1F,$72,$3C,$9A,$E8,$C1,$0C,$10,$FF,$0E,$C2,$38,$C6,$3E,$0F,$D8
	.dc.b	$47,$CE,$D7,$CA,$D7,$81,$A2,$1A,$29,$E0,$F8,$3F,$90,$47,$E1,$B5
	.dc.b	$F8,$2D,$78,$1A,$21,$A2,$9E,$CF,$83,$FA,$84,$7E,$7B,$5F,$9A,$D7
	.dc.b	$81,$A2,$1A,$29,$F2,$3E,$0F,$F0,$04,$7E,$DB,$5F,$B2,$D7,$81,$A2
	.dc.b	$1A,$29,$F5,$3E,$0F,$F1,$84,$7F,$0E,$D7,$F0,$AD,$78,$1A,$21,$A2
	.dc.b	$9F,$80,$F8,$3F,$CC,$11,$FC,$9B,$5F,$C8,$B5,$E0,$68,$86,$8A,$7E
	.dc.b	$33,$E0,$FF,$48,$47,$F3,$ED,$7F,$3A,$D7,$81,$A2,$1A,$29,$F9,$8F
	.dc.b	$83,$FD,$81,$1F,$D5,$B5,$FD,$4B,$5E,$06,$88,$68,$A7,$E9,$3E,$0F
	.dc.b	$F7,$84,$7F,$6E,$D7,$F6,$AD,$78,$1A,$21,$A2,$9F,$B0,$F8,$3F,$E1
	.dc.b	$04,$7F,$81,$B5,$FE,$02,$D7,$81,$A2,$1A,$29,$FB,$CF,$83,$FE,$28
	.dc.b	$47,$F8,$7B,$5F,$E1,$AD,$78,$1A,$21,$A2,$9F,$C2,$3E,$0F,$F9,$01
	.dc.b	$1F,$E3,$6D,$7F,$8C,$B5,$E0,$68,$86,$8A,$7F,$14,$F8,$3F,$E5,$84
	.dc.b	$7F,$93,$B5,$FE,$4A,$D7,$81,$A2,$1A,$29,$FC,$83,$E0,$FF,$9C,$11
	.dc.b	$FE,$66,$D7,$F9,$8B,$5E,$06,$88,$68,$A7,$F2,$CF,$83,$FE,$88,$47
	.dc.b	$F9,$FB,$5F,$E7,$AD,$78,$1A,$21,$A2,$9F,$CE,$3E,$0F,$FA,$81,$1F
	.dc.b	$E9,$6D,$7F,$A4,$B5,$E0,$68,$86,$8A,$7F,$44,$F8,$3F,$EB,$84,$7F
	.dc.b	$AB,$B5,$FE,$AA,$D7,$81,$A2,$1A,$29,$FD,$43,$E0,$FF,$B4,$11,$FE
	.dc.b	$C6,$D7,$FB,$0B,$5E,$06,$88,$68,$A7,$F5,$CF,$83,$FE,$E8,$47,$FB
	.dc.b	$7B,$5F,$ED,$AD,$78,$1A,$21,$A2,$9F,$DA,$3E,$0F,$FC,$00,$47,$FB
	.dc.b	$DB,$5F,$EF,$2D,$78,$1A,$21,$A2,$82,$0F,$BA,$79,$D8,$89,$18,$70
	.dc.b	$71,$70,$E1,$AF,$08,$DB,$B6,$BD,$DB,$1C,$23,$C6,$D7,$81,$C9,$C3
	.dc.b	$C4,$C7,$C2,$72,$23,$82,$1E,$57,$2D,$32,$7B,$9C,$B4,$F0,$7C,$1F
	.dc.b	$98,$8E,$08,$7C,$17,$2D,$32,$7C,$A7,$2D,$3D,$9F,$07,$F0,$88,$E0
	.dc.b	$87,$CD,$72,$D3,$27,$D6,$72,$D3,$E4,$7C,$1F,$CE,$23,$82,$1F,$65
	.dc.b	$CB,$4C,$9F,$82,$72,$D3,$EA,$7C,$1F,$DA,$23,$82,$1F,$85,$72,$D3
	.dc.b	$27,$E3,$9C,$B4,$FC,$07,$C1,$FE,$18,$8E,$08,$7E,$45,$CB,$4C,$9F
	.dc.b	$9A,$72,$D3,$F1,$9F,$07,$F9,$22,$38,$21,$F9,$D7,$2D,$32,$7E,$99
	.dc.b	$CB,$4F,$CC,$7C,$1F,$E7,$88,$E0,$87,$EA,$5C,$B4,$C9,$FB,$27,$2D
	.dc.b	$3F,$49,$F0,$7F,$AA,$23,$82,$1F,$B5,$72,$D3,$27,$EF,$9C,$B4,$FD
	.dc.b	$87,$C1,$FE,$D8,$8E,$08,$7F,$01,$72,$D3,$27,$F0,$A7,$2D,$3F,$79
	.dc.b	$F0,$7F,$C0,$88,$E0,$87,$F0,$D7,$2D,$32,$7F,$16,$72,$D3,$F8,$47
	.dc.b	$C1,$FF,$0E,$23,$82,$1F,$C6,$5C,$B4,$C9,$FC,$89,$CB,$4F,$E2,$9F
	.dc.b	$07,$FC,$68,$8E,$08,$7F,$25,$72,$D3,$27,$F2,$E7,$2D,$3F,$90,$7C
	.dc.b	$1F,$F2,$62,$38,$21,$FC,$C5,$CB,$4C,$9F,$CE,$9C,$B4,$FE,$59,$F0
	.dc.b	$7F,$CC,$88,$E0,$87,$F3,$D7,$2D,$32,$7F,$46,$72,$D3,$F9,$C7,$C1
	.dc.b	$FF,$3E,$23,$82,$1F,$D2,$5C,$B4,$C9,$FD,$49,$CB,$4F,$E8,$9F,$07
	.dc.b	$FD,$28,$8E,$08,$7F,$55,$72,$D3,$27,$F5,$E7,$2D,$3F,$A8,$7C,$1F
	.dc.b	$F5,$62,$38,$21,$FD,$85,$CB,$4C,$9F,$DA,$9C,$B4,$FE,$B9,$F0,$7F
	.dc.b	$D8,$88,$E0,$87,$F6,$D7,$2D,$32,$7F,$76,$72,$D3,$FB,$47,$C1,$FF
	.dc.b	$6E,$23,$82,$1F,$DE,$5C,$B4,$C9,$FE,$02,$72,$D3,$FB,$EF,$DD,$3E
	.dc.b	$13,$FB,$EB,$63,$31,$82,$38,$21,$FE,$0C,$8E,$0D,$8D,$B4,$D9,$FE
	.dc.b	$15,$87,$C2,$78,$09,$1E,$46,$BD,$5A,$F1,$6B,$FE,$22,$8E,$6F,$12
	.dc.b	$1E,$5F,$84,$F6,$91,$F8,$2D,$74,$60,$86,$08,$7A,$5E,$0F,$AF,$E1
	.dc.b	$3E,$49,$1F,$9A,$D7,$E3,$35,$E0,$68,$86,$8A,$7C,$7F,$84,$FA,$A4
	.dc.b	$7E,$CB,$5F,$A4,$D7,$81,$A2,$1A,$29,$F4,$FE,$13,$F0,$24,$7F,$0A
	.dc.b	$D7,$EF,$35,$E0,$68,$86,$8A,$7D,$FF,$84,$FC,$69,$1F,$C8,$B5,$FC
	.dc.b	$53,$5E,$06,$88,$68,$A7,$E2,$FE,$13,$F3,$24,$7F,$3A,$D7,$F2,$CD
	.dc.b	$78,$1A,$21,$A2,$9F,$97,$F8,$4F,$D2,$91,$FD,$4B,$5F,$D1,$35,$E0
	.dc.b	$68,$86,$8A,$7E,$8F,$E1,$3F,$62,$47,$F6,$AD,$7F,$5C,$D7,$81,$A2
	.dc.b	$1A,$29,$FA,$FF,$84,$FD,$E9,$1F,$E0,$2D,$7F,$74,$D7,$81,$A2,$1A
	.dc.b	$29,$FB,$BF,$84,$FE,$12,$47,$F8,$6B,$5F,$E0,$CD,$78,$1A,$21,$A2
	.dc.b	$9F,$C1,$FE,$13,$F8,$A9,$1F,$E3,$2D,$7F,$89,$35,$E0,$68,$86,$8A
	.dc.b	$7F,$13,$F8,$4F,$E4,$24,$7F,$92,$B5,$FE,$3C,$D7,$81,$A2,$1A,$29
	.dc.b	$FC,$7F,$E1,$3F,$96,$91,$FE,$62,$D7,$F9,$53,$5E,$06,$88,$68,$A7
	.dc.b	$F2,$BF,$84,$FE,$72,$47,$F9,$EB,$5F,$E6,$CD,$78,$1A,$21,$A2,$9F
	.dc.b	$CD,$FE,$13,$FA,$29,$1F,$E9,$2D,$7F,$A1,$35,$E0,$68,$86,$8A,$7F
	.dc.b	$43,$F8,$4F,$EA,$24,$7F,$AA,$B5,$FE,$9C,$D7,$81,$A2,$1A,$29,$FD
	.dc.b	$3F,$E1,$3F,$AE,$91,$FE,$C2,$D7,$FA,$D3,$5E,$06,$88,$68,$A7,$F5
	.dc.b	$BF,$84,$FE,$D2,$47,$FB,$6B,$5F,$EC,$CD,$78,$1A,$21,$A2,$9F,$D9
	.dc.b	$FE,$13,$FB,$A9,$1F,$EF,$2D,$7F,$B9,$35,$E0,$68,$86,$8A,$08,$3E
	.dc.b	$0F,$FD,$FE,$D3,$FE,$00,$B8,$23,$82,$1C,$8E,$6E,$9C,$1D,$DC,$38
	.dc.b	$7A,$1D,$88,$F0,$3C,$8E,$1E,$0D,$7C,$0F,$9D,$C9,$C5,$C3,$A7,$83
	.dc.b	$E6,$F8,$3F,$69,$1E,$0D,$7C,$CD,$7B,$B5,$E4,$70,$21,$ED,$F8,$3F
	.dc.b	$92,$47,$C8,$D7,$46,$08,$60,$87,$B9,$E0,$FC,$0F,$83,$FA,$A4,$7E
	.dc.b	$03,$5F,$9C,$D7,$81,$A2,$1A,$29,$F3,$3E,$0F,$F0,$24,$7E,$63,$5F
	.dc.b	$B4,$D7,$81,$A2,$1A,$29,$F6,$3E,$0F,$F1,$A4,$7E,$C3,$5F,$C3,$35
	.dc.b	$E0,$68,$86,$8A,$7E,$13,$E0,$FF,$32,$47,$F0,$8D,$7F,$24,$D7,$81
	.dc.b	$A2,$1A,$29,$F9,$0F,$83,$FD,$29,$1F,$C8,$35,$FC,$F3,$5E,$06,$88
	.dc.b	$68,$A7,$E7,$3E,$0F,$F6,$24,$7F,$38,$D7,$F5,$4D,$78,$1A,$21,$A2
	.dc.b	$9F,$A8,$F8,$3F,$DE,$91,$FD,$43,$5F,$DB,$35,$E0,$68,$86,$8A,$7E
	.dc.b	$D3,$E0,$FF,$84,$91,$FD,$A3,$5F,$E0,$4D,$78,$1A,$21,$A2,$9F,$C0
	.dc.b	$3E,$0F,$F8,$A9,$1F,$E0,$0D,$7F,$87,$35,$E0,$68,$86,$8A,$7F,$0C
	.dc.b	$F8,$3F,$E4,$24,$7F,$86,$35,$FE,$34,$D7,$81,$A2,$1A,$29,$FC,$63
	.dc.b	$E0,$FF,$96,$91,$FE,$30,$D7,$F9,$33,$5E,$06,$88,$68,$A7,$F2,$4F
	.dc.b	$83,$FE,$72,$47,$F9,$23,$5F,$E6,$4D,$78,$1A,$21,$A2,$9F,$CC,$3E
	.dc.b	$0F,$FA,$29,$1F,$E6,$0D,$7F,$9F,$35,$E0,$68,$86,$8A,$7F,$3C,$F8
	.dc.b	$3F,$EA,$24,$7F,$9E,$35,$FE,$94,$D7,$81,$A2,$1A,$29,$FD,$23,$E0
	.dc.b	$FF,$AE,$91,$FE,$90,$D7,$FA,$B3,$5E,$06,$88,$68,$A7,$F5,$4F,$83
	.dc.b	$FE,$D2,$47,$FA,$A3,$5F,$EC,$4D,$78,$1A,$21,$A2,$9F,$D8,$3E,$0F
	.dc.b	$FB,$A9,$1F,$EC,$0D,$7F,$B7,$35,$E0,$68,$86,$8A,$7F,$6C,$F8,$3F
	.dc.b	$F0,$09,$1F,$ED,$8D,$7F,$BD,$35,$E0,$68,$86,$8A,$7F,$81,$FE,$E9
	.dc.b	$F0,$9F,$DF,$E1,$18,$7F,$C0,$4D,$7F,$C1,$6D,$79,$47,$18,$C4,$7F
	.dc.b	$87,$1A,$FF,$87,$DA,$F6,$8E,$04,$7B,$BC,$FE,$0E,$47,$89,$AF,$89
	.dc.b	$AF,$AD,$AF,$85,$AF,$85,$AF,$BB,$E0,$FD,$04,$7D,$ED,$7D,$ED,$7E
	.dc.b	$43,$5F,$73,$5F,$73,$F0,$FE,$0E,$47,$CA,$6B,$F2,$9A,$FD,$36,$BF
	.dc.b	$25,$AF,$C9,$6B,$F5,$BE,$0F,$E8,$11,$FA,$ED,$7E,$BB,$5F,$C0,$35
	.dc.b	$FA,$CD,$7E,$B3,$F6,$FE,$0E,$47,$E0,$9A,$FE,$09,$AF,$E2,$DA,$FE
	.dc.b	$05,$AF,$E0,$5A,$FE,$3B,$E0,$FF,$10,$47,$F1,$ED,$7F,$1E,$D7,$F3
	.dc.b	$0D,$7F,$1C,$D7,$F1,$CF,$E4,$FE,$0E,$47,$E6,$9A,$FE,$69,$AF,$E8
	.dc.b	$DA,$FE,$65,$AF,$E6,$5A,$FE,$9B,$E0,$FF,$40,$47,$F4,$ED,$7F,$4E
	.dc.b	$D7,$F6,$0D,$7F,$4C,$D7,$F4,$CF,$EA,$FE,$0E,$47,$EC,$9A,$FE,$C9
	.dc.b	$AF,$EE,$DA,$FE,$C5,$AF,$EC,$5A,$FE,$FB,$E0,$FF,$70,$47,$F7,$ED
	.dc.b	$7F,$7E,$D7,$F8,$43,$5F,$DF,$35,$FD,$F3,$FC,$0F,$E0,$E4,$7F,$0A
	.dc.b	$6B,$FC,$29,$AF,$F1,$36,$BF,$C2,$5A,$FF,$09,$6B,$FC,$5B,$E0,$FF
	.dc.b	$88,$11,$FE,$2E,$D7,$F8,$BB,$5F,$E4,$0D,$7F,$8B,$35,$FE,$2C,$FF
	.dc.b	$1B,$F8,$39,$1F,$C8,$9A,$FF,$22,$6B,$FC,$AD,$AF,$F2,$16,$BF,$C8
	.dc.b	$5A,$FF,$2E,$F8,$3F,$E5,$04,$7F,$97,$B5,$FE,$5E,$D7,$F9,$C3,$5F
	.dc.b	$E5,$CD,$7F,$97,$3F,$CC,$FE,$0E,$47,$F3,$A6,$BF,$CE,$9A,$FF,$43
	.dc.b	$6B,$FC,$E5,$AF,$F3,$96,$BF,$D1,$BE,$0F,$FA,$01,$1F,$E8,$ED,$7F
	.dc.b	$A3,$B5,$FE,$A0,$D7,$FA,$33,$5F,$E8,$CF,$F4,$BF,$83,$91,$FD,$49
	.dc.b	$AF,$F5,$26,$BF,$D6,$DA,$FF,$51,$6B,$FD,$45,$AF,$F5,$EF,$83,$FE
	.dc.b	$B0,$47,$FA,$FB,$5F,$EB,$ED,$7F,$B4,$35,$FE,$BC,$D7,$FA,$F3,$FD
	.dc.b	$8F,$E0,$E4,$7F,$6A,$6B,$FD,$A9,$AF,$F7,$36,$BF,$DA,$5A,$FF,$69
	.dc.b	$6B,$FD,$DB,$E0,$FF,$B8,$11,$FE,$EE,$D7,$FB,$BB,$5F,$F0,$03,$5F
	.dc.b	$EE,$CD,$7F,$BB,$3F,$E0,$3F,$BA,$7C,$25,$24,$88,$48,$DB,$83,$11
	.dc.b	$C1,$C0,$8D,$B8,$6B,$C5,$AF,$23,$89,$C6,$3A,$3B,$F8,$4F,$41,$1D
	.dc.b	$CA,$31,$1D,$DA,$F1,$36,$E1,$F6,$35,$F8,$10,$F8,$3F,$8C,$48,$EE
	.dc.b	$D7,$D8,$D7,$D8,$D7,$E4,$35,$EA,$FD,$1F,$83,$91,$EA,$6B,$F1,$DA
	.dc.b	$FC,$76,$BF,$4D,$AF,$A1,$AF,$D1,$F8,$3F,$BC,$47,$E3,$B5,$FA,$8D
	.dc.b	$7E,$A3,$5F,$C0,$35,$F8,$AF,$E2,$7E,$0E,$47,$D2,$6B,$F7,$DA,$FD
	.dc.b	$F6,$BF,$8B,$6B,$F4,$1A,$FE,$27,$E0,$FF,$2C,$47,$EF,$B5,$FC,$63
	.dc.b	$5F,$C6,$35,$FC,$C3,$5F,$BA,$FE,$87,$E0,$E4,$7E,$29,$AF,$E5,$DA
	.dc.b	$FE,$5D,$AF,$E8,$DA,$FE,$21,$AF,$E8,$7E,$0F,$F5,$C4,$7F,$2E,$D7
	.dc.b	$F4,$8D,$7F,$48,$D7,$F6,$0D,$7F,$2A,$FE,$E7,$E0,$E4,$7E,$89,$AF
	.dc.b	$EB,$DA,$FE,$BD,$AF,$EE,$DA,$FE,$81,$AF,$EE,$7E,$0F,$F8,$31,$1F
	.dc.b	$D7,$B5,$FD,$E3,$5F,$DE,$35,$FE,$10,$D7,$F5,$AF,$F1,$1F,$83,$91
	.dc.b	$FB,$A6,$BF,$C1,$DA,$FF,$07,$6B,$FC,$4D,$AF,$EE,$1A,$FF,$11,$F8
	.dc.b	$3F,$E3,$C4,$7F,$83,$B5,$FE,$28,$D7,$F8,$A3,$5F,$E4,$0D,$7F,$82
	.dc.b	$BF,$CA,$7E,$0E,$47,$F1,$26,$BF,$C7,$DA,$FF,$1F,$6B,$FC,$AD,$AF
	.dc.b	$F1,$06,$BF,$CA,$7E,$0F,$F9,$B1,$1F,$E3,$ED,$7F,$96,$35,$FE,$58
	.dc.b	$D7,$F9,$C3,$5F,$E3,$AF,$F4,$1F,$83,$91,$FC,$A9,$AF,$F3,$76,$BF
	.dc.b	$CD,$DA,$FF,$43,$6B,$FC,$A1,$AF,$F4,$1F,$83,$FE,$9C,$47,$F9,$BB
	.dc.b	$5F,$E8,$8D,$7F,$A2,$35,$FE,$A0,$D7,$F9,$AB,$FD,$67,$E0,$E4,$7F
	.dc.b	$42,$6B,$FD,$3D,$AF,$F4,$F6,$BF,$D6,$DA,$FF,$40,$6B,$FD,$67,$E0
	.dc.b	$FF,$B3,$11,$FE,$9E,$D7,$FA,$E3,$5F,$EB,$8D,$7F,$B4,$35,$FE,$9A
	.dc.b	$FF,$71,$F8,$39,$1F,$D6,$9A,$FF,$67,$6B,$FD,$9D,$AF,$F7,$36,$BF
	.dc.b	$D6,$1A,$FF,$71,$F8,$3F,$EF,$C4,$7F,$B3,$B5,$FE,$E8,$D7,$FB,$A3
	.dc.b	$5F,$F0,$03,$5F,$EC,$AF,$F7,$65,$F7,$55,$A6,$5C,$11,$C1,$0F,$EF
	.dc.b	$EE,$5A,$64,$FF,$05,$BC,$BE,$13,$A1,$23,$8B,$5D,$B5,$FF,$0B,$42
	.dc.b	$31,$1F,$E1,$C8,$C3,$C8,$87,$83,$E1,$3D,$44,$7C,$ED,$74,$60,$86
	.dc.b	$08,$78,$9E,$0F,$93,$E1,$3E,$31,$1F,$86,$D7,$E0,$B5,$E0,$68,$86
	.dc.b	$8A,$7C,$0F,$84,$FA,$44,$7E,$7B,$5F,$9A,$D7,$81,$A2,$1A,$29,$F3
	.dc.b	$3E,$13,$EF,$11,$FB,$6D,$7E,$CB,$5E,$06,$88,$68,$A7,$D8,$F8,$4F
	.dc.b	$C5,$11,$FC,$3B,$5F,$C2,$B5,$E0,$68,$86,$8A,$7E,$13,$E1,$3F,$2C
	.dc.b	$47,$F2,$6D,$7F,$22,$D7,$81,$A2,$1A,$29,$F9,$0F,$84,$FD,$11,$1F
	.dc.b	$CF,$B5,$FC,$EB,$5E,$06,$88,$68,$A7,$E7,$3E,$13,$F5,$C4,$7F,$56
	.dc.b	$D7,$F5,$2D,$78,$1A,$21,$A2,$9F,$A8,$F8,$4F,$DD,$11,$FD,$BB,$5F
	.dc.b	$DA,$B5,$E0,$68,$86,$8A,$7E,$D3,$E1,$3F,$83,$11,$FE,$06,$D7,$F8
	.dc.b	$0B,$5E,$06,$88,$68,$A7,$F0,$0F,$84,$FE,$24,$47,$F8,$7B,$5F,$E1
	.dc.b	$AD,$78,$1A,$21,$A2,$9F,$C3,$3E,$13,$F8,$F1,$1F,$E3,$6D,$7F,$8C
	.dc.b	$B5,$E0,$68,$86,$8A,$7F,$18,$F8,$4F,$E5,$44,$7F,$93,$B5,$FE,$4A
	.dc.b	$D7,$81,$A2,$1A,$29,$FC,$93,$E1,$3F,$9B,$11,$FE,$66,$D7,$F9,$8B
	.dc.b	$5E,$06,$88,$68,$A7,$F3,$0F,$84,$FE,$84,$47,$F9,$FB,$5F,$E7,$AD
	.dc.b	$78,$1A,$21,$A2,$9F,$CF,$3E,$13,$FA,$71,$1F,$E9,$6D,$7F,$A4,$B5
	.dc.b	$E0,$68,$86,$8A,$7F,$48,$F8,$4F,$EB,$44,$7F,$AB,$B5,$FE,$AA,$D7
	.dc.b	$81,$A2,$1A,$29,$FD,$53,$E1,$3F,$B3,$11,$FE,$C6,$D7,$FB,$0B,$5E
	.dc.b	$06,$88,$68,$A7,$F6,$0F,$84,$FE,$E4,$47,$FB,$7B,$5F,$ED,$AD,$78
	.dc.b	$1A,$21,$A2,$9F,$DB,$3E,$13,$FB,$F1,$1F,$EF,$6D,$7F,$BC,$B5,$E0
	.dc.b	$68,$86,$8A,$08,$7F,$BF,$7E,$0F,$FC,$21,$E7,$1A,$48,$D9,$C2,$34
	.dc.b	$E0,$E8,$E1,$C3,$5E,$2D,$79,$47,$57,$6D,$77,$F0,$7E,$42,$3D,$C4
	.dc.b	$A3,$31,$E0,$6B,$C8,$E1,$1B,$6C,$69,$F6,$3F,$0B,$E0,$E4,$70,$6B
	.dc.b	$EC,$6B,$EC,$6B,$F1,$5A,$FC,$46,$BE,$8F,$83,$F9,$E4,$7C,$AD,$7E
	.dc.b	$4B,$5F,$92,$D7,$E9,$35,$FA,$2F,$DA,$F8,$39,$1F,$01,$AF,$D4,$6B
	.dc.b	$F5,$1A,$FD,$D6,$BF,$71,$AF,$D0,$F8,$3F,$C3,$91,$F9,$AD,$7F,$02
	.dc.b	$D7,$F0,$2D,$7F,$14,$D7,$F1,$2F,$E4,$BE,$0E,$47,$D8,$6B,$F8,$C6
	.dc.b	$BF,$8C,$6B,$F9,$56,$BF,$94,$6B,$F8,$8F,$83,$FC,$F9,$1F,$C2,$B5
	.dc.b	$FC,$CB,$5F,$CC,$B5,$FD,$13,$5F,$D0,$BF,$AA,$F8,$39,$1F,$90,$6B
	.dc.b	$FA,$46,$BF,$A4,$6B,$FA,$D6,$BF,$AC,$6B,$FA,$0F,$83,$FD,$B9,$1F
	.dc.b	$CE,$B5,$FD,$8B,$5F,$D8,$B5,$FD,$D3,$5F,$DC,$BF,$C0,$BE,$0E,$47
	.dc.b	$EA,$1A,$FE,$F1,$AF,$EF,$1A,$FF,$05,$6B,$FC,$11,$AF,$EE,$3E,$0F
	.dc.b	$F8,$79,$1F,$DA,$B5,$FE,$12,$D7,$F8,$4B,$5F,$E2,$4D,$7F,$88,$BF
	.dc.b	$C6,$BE,$0E,$47,$F0,$06,$BF,$C5,$1A,$FF,$14,$6B,$FC,$75,$AF,$F1
	.dc.b	$C6,$BF,$C4,$3E,$0F,$F9,$39,$1F,$E1,$AD,$7F,$90,$B5,$FE,$42,$D7
	.dc.b	$F9,$53,$5F,$E5,$2F,$F3,$2F,$83,$91,$FC,$61,$AF,$F2,$C6,$BF,$CB
	.dc.b	$1A,$FF,$35,$6B,$FC,$D1,$AF,$F2,$8F,$83,$FE,$7E,$47,$F9,$2B,$5F
	.dc.b	$E7,$2D,$7F,$9C,$B5,$FE,$84,$D7,$FA,$0B,$FD,$2B,$E0,$E4,$7F,$30
	.dc.b	$6B,$FD,$11,$AF,$F4,$46,$BF,$D3,$5A,$FF,$4C,$6B,$FD,$03,$E0,$FF
	.dc.b	$AB,$91,$FE,$7A,$D7,$FA,$8B,$5F,$EA,$2D,$7F,$AD,$35,$FE,$B2,$FF
	.dc.b	$62,$F8,$39,$1F,$D2,$1A,$FF,$5C,$6B,$FD,$71,$AF,$F6,$56,$BF,$D9
	.dc.b	$1A,$FF,$58,$F8,$3F,$ED,$E4,$7F,$AA,$B5,$FE,$D2,$D7,$FB,$4B,$5F
	.dc.b	$EE,$4D,$7F,$B8,$BF,$DE,$BE,$0E,$47,$F6,$06,$BF,$DD,$1A,$FF,$74
	.dc.b	$6B,$FD,$F5,$AF,$F7,$C6,$AF,$BA,$7C,$27,$F7,$38,$C3,$63,$31,$FD
	.dc.b	$DD,$AF,$F8,$1D,$AF,$F8,$31,$AF,$F8,$51,$AF,$F8,$5D,$AF,$63,$A4
	.dc.b	$77,$23,$DF,$A7,$C4,$47,$04,$3C,$4E,$5A,$64,$F2,$39,$69,$EB,$F8
	.dc.b	$3F,$71,$1C,$10,$F7,$39,$69,$93,$E0,$39,$69,$F1,$FE,$0F,$E5,$11
	.dc.b	$C1,$0F,$94,$E5,$A6,$4F,$98,$E5,$A7,$D3,$F8,$3F,$AC,$47,$04,$3E
	.dc.b	$B3,$96,$99,$3E,$C3,$96,$9F,$7F,$E0,$FF,$04,$47,$04,$3F,$04,$E5
	.dc.b	$A6,$4F,$C2,$39,$69,$F8,$BF,$83,$FC,$71,$1C,$10,$FC,$73,$96,$99
	.dc.b	$3F,$20,$E5,$A7,$E5,$FE,$0F,$F3,$44,$70,$43,$F3,$4E,$5A,$64,$FC
	.dc.b	$E3,$96,$9F,$A3,$F8,$3F,$D3,$11,$C1,$0F,$D3,$39,$69,$93,$F5,$0E
	.dc.b	$5A,$7E,$BF,$E0,$FF,$64,$47,$04,$3F,$64,$E5,$A6,$4F,$DA,$39,$69
	.dc.b	$FB,$BF,$83,$FD,$F1,$1C,$10,$FD,$F3,$96,$99,$3F,$80,$39,$69,$FC
	.dc.b	$1F,$E0,$FF,$85,$11,$C1,$0F,$E1,$4E,$5A,$64,$FE,$18,$E5,$A7,$F1
	.dc.b	$3F,$83,$FE,$2C,$47,$04,$3F,$8B,$39,$69,$93,$F8,$C3,$96,$9F,$C7
	.dc.b	$FE,$0F,$F9,$11,$1C,$10,$FE,$44,$E5,$A6,$4F,$E4,$8E,$5A,$7F,$2B
	.dc.b	$F8,$3F,$E5,$C4,$70,$43,$F9,$73,$96,$99,$3F,$98,$39,$69,$FC,$DF
	.dc.b	$E0,$FF,$9D,$11,$C1,$0F,$E7,$4E,$5A,$64,$FE,$78,$E5,$A7,$F4,$3F
	.dc.b	$83,$FE,$8C,$47,$04,$3F,$A3,$39,$69,$93,$FA,$43,$96,$9F,$D3,$FE
	.dc.b	$0F,$FA,$91,$1C,$10,$FE,$A4,$E5,$A6,$4F,$EA,$8E,$5A,$7F,$5B,$F8
	.dc.b	$3F,$EB,$C4,$70,$43,$FA,$F3,$96,$99,$3F,$B0,$39,$69,$FD,$9F,$E0
	.dc.b	$FF,$B5,$11,$C1,$0F,$ED,$4E,$5A,$64,$FE,$D8,$E5,$A7,$F7,$3F,$83
	.dc.b	$FE,$EC,$47,$04,$3F,$BB,$39,$69,$93,$FB,$C3,$96,$9F,$EF,$5F,$BA
	.dc.b	$FF,$02,$26,$91,$FF,$04,$35,$E0,$60,$86,$08,$08,$57,$AF,$C2,$74
	.dc.b	$24,$78,$DA,$F0,$38,$38,$71,$08,$C4,$76,$8C,$3C,$C8,$73,$3D,$DF
	.dc.b	$07,$EC,$23,$F0,$DA,$EE,$11,$88,$F0,$B5,$F0,$35,$F7,$3F,$23,$E0
	.dc.b	$E4,$7C,$F6,$BD,$5A,$FB,$1A,$FA,$DA,$FC,$96,$BF,$5F,$E0,$FE,$A9
	.dc.b	$1F,$B8,$D7,$D4,$D7,$E4,$B5,$F9,$0D,$7E,$B3,$F8,$0F,$83,$91,$F8
	.dc.b	$76,$BF,$15,$AF,$D4,$6B,$F4,$DA,$FE,$05,$AF,$E3,$FE,$0F,$F1,$A4
	.dc.b	$7F,$28,$D7,$E9,$35,$FC,$0B,$5F,$C0,$35,$FC,$73,$F9,$8F,$83,$91
	.dc.b	$F9,$F6,$BF,$75,$AF,$E3,$1A,$FE,$2D,$AF,$E6,$5A,$FE,$9F,$E0,$FF
	.dc.b	$4A,$47,$F5,$8D,$7F,$14,$D7,$F3,$2D,$7F,$30,$D7,$F4,$CF,$EC,$3E
	.dc.b	$0E,$47,$ED,$DA,$FE,$55,$AF,$E9,$1A,$FE,$8D,$AF,$EC,$5A,$FE,$FF
	.dc.b	$E0,$FF,$7A,$47,$F8,$23,$5F,$D1,$35,$FD,$8B,$5F,$D8,$35,$FD,$F3
	.dc.b	$FC,$23,$E0,$E4,$7F,$0F,$6B,$FA,$D6,$BF,$BC,$6B,$FB,$B6,$BF,$C2
	.dc.b	$5A,$FF,$17,$F8,$3F,$E2,$A4,$7F,$8E,$35,$FD,$D3,$5F,$E1,$2D,$7F
	.dc.b	$84,$35,$FE,$2C,$FF,$20,$F8,$39,$1F,$C9,$DA,$FF,$05,$6B,$FC,$51
	.dc.b	$AF,$F1,$36,$BF,$C8,$5A,$FF,$2F,$F8,$3F,$E5,$A4,$7F,$9A,$35,$FE
	.dc.b	$24,$D7,$F9,$0B,$5F,$E4,$0D,$7F,$97,$3F,$CE,$3E,$0E,$47,$F3,$F6
	.dc.b	$BF,$C7,$5A,$FF,$2C,$6B,$FC,$AD,$AF,$F3,$96,$BF,$D1,$FE,$0F,$FA
	.dc.b	$29,$1F,$E9,$8D,$7F,$95,$35,$FE,$72,$D7,$F9,$C3,$5F,$E8,$CF,$F5
	.dc.b	$0F,$83,$91,$FD,$5D,$AF,$F3,$56,$BF,$D1,$1A,$FF,$43,$6B,$FD,$45
	.dc.b	$AF,$F5,$FF,$83,$FE,$BA,$47,$FB,$23,$5F,$E8,$4D,$7F,$A8,$B5,$FE
	.dc.b	$A0,$D7,$FA,$F3,$FD,$A3,$E0,$E4,$7F,$6F,$6B,$FD,$35,$AF,$F5,$C6
	.dc.b	$BF,$D6,$DA,$FF,$69,$6B,$FD,$DF,$E0,$FF,$BA,$91,$FE,$F8,$D7,$FA
	.dc.b	$D3,$5F,$ED,$2D,$7F,$B4,$35,$FE,$EC,$FF,$80,$3E,$0E,$47,$F8,$1D
	.dc.b	$AF,$F6,$56,$BF,$DD,$1A,$FF,$73,$6B,$FE,$01,$6A,$FB,$A7,$C2,$7F
	.dc.b	$75,$8F,$F0,$44,$61,$C3,$5F,$EF,$4D,$7F,$C0,$CD,$7F,$C1,$ED,$79
	.dc.b	$C7,$57,$C0,$D8,$D4,$60,$EF,$19,$35,$F0,$92,$3C,$6D,$7C,$6D,$7C
	.dc.b	$6D,$7C,$6D,$7D,$AD,$7E,$17,$C2,$78,$C8,$FC,$06,$BF,$01,$AF,$C0
	.dc.b	$6B,$F0,$1A,$FC,$A7,$DD,$F0,$92,$3E,$5B,$5F,$96,$D7,$E5,$B5,$F9
	.dc.b	$6D,$7E,$AB,$5F,$B5,$F0,$9F,$2C,$8F,$D8,$6B,$F6,$1A,$FD,$86,$BF
	.dc.b	$61,$AF,$E0,$9F,$AD,$F0,$92,$3F,$06,$D7,$F0,$6D,$7F,$06,$D7,$F0
	.dc.b	$6D,$7F,$1A,$D7,$F2,$5F,$09,$F8,$32,$3F,$90,$6B,$F9,$06,$BF,$90
	.dc.b	$6B,$F9,$06,$BF,$9A,$7F,$1D,$F0,$92,$3F,$36,$D7,$F3,$6D,$7F,$36
	.dc.b	$D7,$F3,$6D,$7F,$4A,$D7,$F5,$5F,$09,$F9,$B2,$3F,$A8,$6B,$FA,$86
	.dc.b	$BF,$A8,$6B,$FA,$86,$BF,$B2,$7F,$4D,$F0,$92,$3F,$66,$D7,$F6,$6D
	.dc.b	$7F,$66,$D7,$F6,$6D,$7F,$7A,$D7,$F8,$17,$C2,$7E,$CC,$8F,$F0,$06
	.dc.b	$BF,$C0,$1A,$FF,$00,$6B,$FC,$01,$AF,$F0,$A7,$F7,$DF,$09,$23,$F8
	.dc.b	$5B,$5F,$E1,$6D,$7F,$85,$B5,$FE,$16,$D7,$F8,$AB,$5F,$E3,$5F,$09
	.dc.b	$FC,$2C,$8F,$F1,$86,$BF,$C6,$1A,$FF,$18,$6B,$FC,$61,$AF,$F2,$27
	.dc.b	$F8,$B7,$C2,$48,$FE,$46,$D7,$F9,$1B,$5F,$E4,$6D,$7F,$91,$B5,$FE
	.dc.b	$5A,$D7,$F9,$97,$C2,$7F,$23,$23,$FC,$C1,$AF,$F3,$06,$BF,$CC,$1A
	.dc.b	$FF,$30,$6B,$FC,$E9,$FE,$5D,$F0,$92,$3F,$9D,$B5,$FE,$76,$D7,$F9
	.dc.b	$DB,$5F,$E7,$6D,$7F,$A2,$B5,$FE,$95,$F0,$9F,$CE,$C8,$FF,$48,$6B
	.dc.b	$FD,$21,$AF,$F4,$86,$BF,$D2,$1A,$FF,$52,$7F,$A3,$7C,$24,$8F,$EA
	.dc.b	$6D,$7F,$A9,$B5,$FE,$A6,$D7,$FA,$9B,$5F,$EB,$AD,$7F,$B1,$7C,$27
	.dc.b	$F5,$32,$3F,$D8,$1A,$FF,$60,$6B,$FD,$81,$AF,$F6,$06,$BF,$DA,$9F
	.dc.b	$EB,$DF,$09,$23,$FB,$5B,$5F,$ED,$6D,$7F,$B5,$B5,$FE,$D6,$D7,$FB
	.dc.b	$AB,$5F,$EF,$5F,$09,$FD,$AC,$8F,$F7,$86,$BF,$DE,$1A,$FF,$78,$6B
	.dc.b	$FD,$E1,$AF,$F8,$09,$FE,$F1,$CC,$F8,$3A,$49,$FF,$04,$46,$5B,$19
	.dc.b	$7F,$E0,$04,$65,$B1,$94,$41,$1B,$70,$62,$38,$3B,$87,$C1,$F9,$89
	.dc.b	$1E,$46,$BE,$26,$BD,$23,$8B,$B2,$1B,$29,$BF,$83,$F8,$64,$7E,$0B
	.dc.b	$5F,$7B,$5E,$06,$88,$68,$A7,$8F,$E0,$FE,$79,$1F,$9A,$D7,$E5,$B5
	.dc.b	$E0,$68,$86,$8A,$7B,$FE,$0F,$ED,$91,$FB,$2D,$7E,$BB,$5E,$06,$88
	.dc.b	$68,$A7,$CB,$F8,$3F,$C3,$91,$FC,$2B,$5F,$C1,$B5,$E0,$68,$86,$8A
	.dc.b	$7D,$7F,$83,$FC,$99,$1F,$C8,$B5,$FC,$7B,$5E,$06,$88,$68,$A7,$E0
	.dc.b	$FE,$0F,$F3,$E4,$7F,$3A,$D7,$F3,$6D,$78,$1A,$21,$A2,$9F,$8F,$F8
	.dc.b	$3F,$D5,$91,$FD,$4B,$5F,$D3,$B5,$E0,$68,$86,$8A,$7E,$6F,$E0,$FF
	.dc.b	$6E,$47,$F6,$AD,$7F,$66,$D7,$81,$A2,$1A,$29,$FA,$7F,$83,$FE,$06
	.dc.b	$47,$F8,$0B,$5F,$DF,$B5,$E0,$68,$86,$8A,$7E,$CF,$E0,$FF,$87,$91
	.dc.b	$FE,$1A,$D7,$F8,$5B,$5E,$06,$88,$68,$A7,$EF,$FE,$0F,$F8,$D9,$1F
	.dc.b	$E3,$2D,$7F,$8B,$B5,$E0,$68,$86,$8A,$7F,$0B,$F8,$3F,$E4,$E4,$7F
	.dc.b	$92,$B5,$FE,$46,$D7,$81,$A2,$1A,$29,$FC,$5F,$E0,$FF,$99,$91,$FE
	.dc.b	$62,$D7,$F9,$7B,$5E,$06,$88,$68,$A7,$F2,$3F,$83,$FE,$7E,$47,$F9
	.dc.b	$EB,$5F,$E7,$6D,$78,$1A,$21,$A2,$9F,$CB,$FE,$0F,$FA,$59,$1F,$E9
	.dc.b	$2D,$7F,$A3,$B5,$E0,$68,$86,$8A,$7F,$3B,$F8,$3F,$EA,$E4,$7F,$AA
	.dc.b	$B5,$FE,$A6,$D7,$81,$A2,$1A,$29,$FD,$1F,$E0,$FF,$B1,$91,$FE,$C2
	.dc.b	$D7,$FA,$FB,$5E,$06,$88,$68,$A7,$F5,$3F,$83,$FE,$DE,$47,$FB,$6B
	.dc.b	$5F,$ED,$6D,$78,$1A,$21,$A2,$9F,$D7,$FE,$0F,$FB,$D9,$1F,$EF,$2D
	.dc.b	$7F,$BB,$B5,$E0,$68,$86,$8A,$7F,$76,$FB,$AF,$F7,$F2,$69,$1F,$EF
	.dc.b	$67,$F7,$F7,$0E,$0E,$05,$29,$97,$6F,$26,$BE,$1F,$84,$E1,$06,$92
	.dc.b	$39,$35,$E4,$D7,$91,$A2,$1A,$3C,$5F,$07,$D8,$8E,$08,$79,$5C,$B4
	.dc.b	$C9,$F1,$DC,$B4,$F7,$7C,$1F,$A8,$8E,$08,$7C,$17,$2D,$32,$7D,$37
	.dc.b	$2D,$3E,$57,$C1,$FC,$62,$38,$21,$F3,$5C,$B4,$C9,$F7,$DC,$B4,$FA
	.dc.b	$DF,$07,$F4,$88,$E0,$87,$D9,$72,$D3,$27,$E2,$DC,$B4,$FC,$17,$C1
	.dc.b	$FD,$E2,$38,$21,$F8,$57,$2D,$32,$7E,$5D,$CB,$4F,$C7,$7C,$1F,$E2
	.dc.b	$88,$E0,$87,$E4,$5C,$B4,$C9,$FA,$37,$2D,$3F,$35,$F0,$7F,$96,$23
	.dc.b	$82,$1F,$9D,$72,$D3,$27,$EB,$DC,$B4,$FD,$37,$C1,$FE,$88,$8E,$08
	.dc.b	$7E,$A5,$CB,$4C,$9F,$BB,$72,$D3,$F6,$5F,$07,$FA,$E2,$38,$21,$FB
	.dc.b	$57,$2D,$32,$7F,$07,$72,$D3,$F7,$DF,$07,$FB,$A2,$38,$21,$FC,$05
	.dc.b	$CB,$4C,$9F,$C4,$DC,$B4,$FE,$15,$F0,$7F,$C1,$88,$E0,$87,$F0,$D7
	.dc.b	$2D,$32,$7F,$1F,$72,$D3,$F8,$B7,$C1,$FF,$12,$23,$82,$1F,$C6,$5C
	.dc.b	$B4,$C9,$FC,$AD,$CB,$4F,$E4,$5F,$07,$FC,$78,$8E,$08,$7F,$25,$72
	.dc.b	$D3,$27,$F3,$77,$2D,$3F,$97,$7C,$1F,$F2,$A2,$38,$21,$FC,$C5,$CB
	.dc.b	$4C,$9F,$D0,$DC,$B4,$FE,$75,$F0,$7F,$CD,$88,$E0,$87,$F3,$D7,$2D
	.dc.b	$32,$7F,$4F,$72,$D3,$FA,$37,$C1,$FF,$42,$23,$82,$1F,$D2,$5C,$B4
	.dc.b	$C9,$FD,$6D,$CB,$4F,$EA,$5F,$07,$FD,$38,$8E,$08,$7F,$55,$72,$D3
	.dc.b	$27,$F6,$77,$2D,$3F,$AF,$7C,$1F,$F5,$A2,$38,$21,$FD,$85,$CB,$4C
	.dc.b	$9F,$DC,$DC,$B4,$FE,$D5,$F0,$7F,$D9,$88,$E0,$87,$F6,$D7,$2D,$32
	.dc.b	$7F,$7F,$72,$D3,$FB,$B7,$C1,$FF,$72,$23,$82,$1F,$DE,$5C,$B4,$C9
	.dc.b	$FE,$0B,$72,$D3,$FC,$0D,$F7,$4F,$83,$FF,$03,$36,$33,$18,$23,$82
	.dc.b	$1F,$E1,$28,$E0,$D8,$DB,$4D,$82,$08,$79,$DE,$22,$47,$99,$AF,$76
	.dc.b	$BC,$8E,$4E,$1E,$B1,$D1,$E4,$43,$D1,$F0,$7F,$00,$8F,$C4,$6B,$A3
	.dc.b	$04,$30,$43,$C5,$08,$E3,$18,$F8,$3F,$9A,$47,$E8,$B5,$F9,$4D,$78
	.dc.b	$1A,$21,$A2,$9E,$0F,$83,$FB,$24,$7E,$EB,$5F,$AC,$D7,$81,$A2,$1A
	.dc.b	$29,$EC,$F8,$3F,$C2,$91,$FC,$4B,$5F,$C1,$35,$E0,$68,$86,$8A,$7C
	.dc.b	$8F,$83,$FC,$89,$1F,$CA,$B5,$FC,$73,$5E,$06,$88,$68,$A7,$D4,$F8
	.dc.b	$3F,$CE,$91,$FD,$0B,$5F,$CD,$35,$E0,$68,$86,$8A,$7E,$03,$E0,$FF
	.dc.b	$52,$47,$F5,$AD,$7F,$4C,$D7,$81,$A2,$1A,$29,$F8,$CF,$83,$FD,$A9
	.dc.b	$1F,$DC,$B5,$FD,$93,$5E,$06,$88,$68,$A7,$E6,$3E,$0F,$F8,$09,$1F
	.dc.b	$E0,$AD,$7F,$7C,$D7,$81,$A2,$1A,$29,$FA,$4F,$83,$FE,$1A,$47,$F8
	.dc.b	$8B,$5F,$E1,$4D,$78,$1A,$21,$A2,$9F,$B0,$F8,$3F,$E3,$24,$7F,$8E
	.dc.b	$B5,$FE,$2C,$D7,$81,$A2,$1A,$29,$FB,$CF,$83,$FE,$4A,$47,$F9,$4B
	.dc.b	$5F,$E4,$4D,$78,$1A,$21,$A2,$9F,$C2,$3E,$0F,$F9,$89,$1F,$E6,$AD
	.dc.b	$7F,$97,$35,$E0,$68,$86,$8A,$7F,$14,$F8,$3F,$E7,$A4,$7F,$A0,$B5
	.dc.b	$FE,$74,$D7,$81,$A2,$1A,$29,$FC,$83,$E0,$FF,$A4,$91,$FE,$9A,$D7
	.dc.b	$FA,$33,$5E,$06,$88,$68,$A7,$F2,$CF,$83,$FE,$AA,$47,$FA,$CB,$5F
	.dc.b	$EA,$4D,$78,$1A,$21,$A2,$9F,$CE,$3E,$0F,$FB,$09,$1F,$EC,$AD,$7F
	.dc.b	$AF,$35,$E0,$68,$86,$8A,$7F,$44,$F8,$3F,$ED,$A4,$7F,$B8,$B5,$FE
	.dc.b	$D4,$D7,$81,$A2,$1A,$29,$FD,$43,$E0,$FF,$BC,$91,$FE,$FA,$D7,$FB
	.dc.b	$B3,$5E,$06,$88,$68,$A7,$F6,$DF,$BA,$7C,$27,$F6,$F8,$FE,$F3,$18
	.dc.b	$70,$1C,$3B,$6B,$CE,$36,$ED,$B1,$A8,$EB,$18,$23,$DA,$34,$FF,$82
	.dc.b	$DA,$F8,$7E,$13,$99,$1E,$0D,$7C,$8D,$7D,$CD,$7D,$AD,$7B,$3E,$6F
	.dc.b	$84,$91,$E3,$6B,$EF,$6B,$F2,$5A,$FC,$86,$BE,$96,$BF,$27,$E1,$3E
	.dc.b	$19,$1F,$80,$D7,$E6,$35,$FA,$CD,$7E,$AB,$5F,$8C,$FC,$EF,$84,$91
	.dc.b	$F2,$DA,$FD,$76,$BF,$81,$6B,$F8,$06,$BF,$45,$AF,$E0,$7E,$13,$ED
	.dc.b	$91,$FB,$0D,$7F,$08,$D7,$F1,$CD,$7F,$1A,$D7,$EF,$3F,$86,$F8,$49
	.dc.b	$1F,$83,$6B,$F8,$F6,$BF,$99,$6B,$F9,$86,$BF,$89,$6B,$F9,$9F,$84
	.dc.b	$FC,$99,$1F,$C8,$35,$FC,$E3,$5F,$D3,$35,$FD,$2B,$5F,$CB,$3F,$9E
	.dc.b	$F8,$49,$1F,$9B,$6B,$FA,$76,$BF,$B1,$6B,$FB,$06,$BF,$A1,$6B,$FB
	.dc.b	$1F,$84,$FD,$59,$1F,$D4,$35,$FD,$A3,$5F,$DF,$35,$FD,$EB,$5F,$D7
	.dc.b	$3F,$B6,$F8,$49,$1F,$B3,$6B,$FB,$F6,$BF,$C2,$5A,$FF,$08,$6B,$FB
	.dc.b	$96,$BF,$C2,$7E,$13,$F8,$19,$1F,$E0,$0D,$7F,$86,$35,$FE,$2C,$D7
	.dc.b	$F8,$AB,$5F,$E0,$CF,$F0,$EF,$84,$91,$FC,$2D,$AF,$F1,$76,$BF,$C8
	.dc.b	$5A,$FF,$20,$6B,$FC,$45,$AF,$F2,$1F,$84,$FE,$36,$47,$F8,$C3,$5F
	.dc.b	$E4,$8D,$7F,$97,$35,$FE,$5A,$D7,$F8,$F3,$FC,$9B,$E1,$24,$7F,$23
	.dc.b	$6B,$FC,$BD,$AF,$F3,$96,$BF,$CE,$1A,$FF,$29,$6B,$FC,$E7,$E1,$3F
	.dc.b	$99,$91,$FE,$60,$D7,$F9,$E3,$5F,$E8,$CD,$7F,$A2,$B5,$FE,$6C,$FF
	.dc.b	$3E,$F8,$49,$1F,$CE,$DA,$FF,$47,$6B,$FD,$45,$AF,$F5,$06,$BF,$D0
	.dc.b	$5A,$FF,$51,$F8,$4F,$E9,$64,$7F,$A4,$35,$FE,$A8,$D7,$FA,$F3,$5F
	.dc.b	$EB,$AD,$7F,$A7,$3F,$D5,$BE,$12,$47,$F5,$36,$BF,$D7,$DA,$FF,$69
	.dc.b	$6B,$FD,$A1,$AF,$F5,$96,$BF,$DA,$7E,$13,$FB,$19,$1F,$EC,$0D,$7F
	.dc.b	$B6,$35,$FE,$EC,$D7,$FB,$AB,$5F,$EC,$CF,$F6,$EF,$84,$91,$FD,$AD
	.dc.b	$AF,$F7,$76,$BF,$E0,$16,$BF,$E0,$06,$BF,$DC,$5A,$BE,$0B,$FB,$EB
	.dc.b	$4F,$EF,$EE,$08,$E0,$87,$F8,$29,$CB,$4C,$9F,$E1,$17,$97,$C1,$F4
	.dc.b	$24,$71,$6B,$B6,$A2,$1E,$77,$2E,$21,$18,$8E,$11,$87,$CC,$FB,$08
	.dc.b	$E0,$87,$9D,$CB,$4C,$9E,$23,$E4,$3A,$70,$FC,$97,$E0,$7C,$1F,$CA
	.dc.b	$23,$F1,$DA,$E8,$C1,$0C,$10,$EA,$F0,$7B,$FC,$1F,$D6,$23,$F4,$DA
	.dc.b	$FD,$16,$BC,$0D,$10,$D1,$4F,$5F,$C1,$FE,$08,$8F,$DF,$6B,$F7,$5A
	.dc.b	$F0,$34,$43,$45,$3E,$3F,$C1,$FE,$38,$8F,$E2,$DA,$FE,$25,$AF,$03
	.dc.b	$44,$34,$53,$E9,$FC,$1F,$E6,$88,$FE,$5D,$AF,$E5,$5A,$F0,$34,$43
	.dc.b	$45,$3E,$FF,$C1,$FE,$98,$8F,$E8,$DA,$FE,$85,$AF,$03,$44,$34,$53
	.dc.b	$F1,$7F,$07,$FB,$22,$3F,$AF,$6B,$FA,$D6,$BC,$0D,$10,$D1,$4F,$CB
	.dc.b	$FC,$1F,$EF,$88,$FE,$ED,$AF,$EE,$5A,$F0,$34,$43,$45,$3F,$47,$F0
	.dc.b	$7F,$C2,$88,$FF,$07,$6B,$FC,$15,$AF,$03,$44,$34,$53,$F5,$FF,$07
	.dc.b	$FC,$58,$8F,$F1,$36,$BF,$C4,$5A,$F0,$34,$43,$45,$3F,$77,$F0,$7F
	.dc.b	$C8,$88,$FF,$1F,$6B,$FC,$75,$AF,$03,$44,$34,$53,$F8,$3F,$C1,$FF
	.dc.b	$2E,$23,$FC,$AD,$AF,$F2,$96,$BC,$0D,$10,$D1,$4F,$E2,$7F,$07,$FC
	.dc.b	$E8,$8F,$F3,$76,$BF,$CD,$5A,$F0,$34,$43,$45,$3F,$8F,$FC,$1F,$F4
	.dc.b	$62,$3F,$D0,$DA,$FF,$41,$6B,$C0,$D1,$0D,$14,$FE,$57,$F0,$7F,$D4
	.dc.b	$88,$FF,$4F,$6B,$FD,$35,$AF,$03,$44,$34,$53,$F9,$BF,$C1,$FF,$5E
	.dc.b	$23,$FD,$6D,$AF,$F5,$96,$BC,$0D,$10,$D1,$4F,$E8,$7F,$07,$FD,$A8
	.dc.b	$8F,$F6,$76,$BF,$D9,$5A,$F0,$34,$43,$45,$3F,$A7,$FC,$1F,$F7,$62
	.dc.b	$3F,$DC,$DA,$FF,$71,$6B,$C0,$D1,$0D,$14,$FE,$B7,$F0,$7F,$E0,$22
	.dc.b	$3F,$DF,$DA,$FF,$7D,$6B,$C0,$D1,$0D,$14,$FE,$D5,$F7,$4F,$84,$FE
	.dc.b	$D9,$1F,$DF,$A3,$0E,$03,$87,$6D,$79,$C6,$DD,$B6,$35,$1D,$63,$04
	.dc.b	$7B,$46,$9F,$F0,$83,$5F,$0F,$C2,$73,$23,$C1,$AF,$91,$AF,$B9,$AF
	.dc.b	$B5,$AF,$67,$CD,$F0,$72,$3C,$6D,$7D,$ED,$7E,$4B,$5F,$90,$D7,$D2
	.dc.b	$D7,$E4,$FC,$1F,$C3,$23,$F0,$1A,$FC,$C6,$BF,$59,$AF,$D5,$6B,$F1
	.dc.b	$9F,$9D,$F0,$72,$3E,$5B,$5F,$AE,$D7,$F0,$2D,$7F,$00,$D7,$E8,$B5
	.dc.b	$FC,$0F,$C1,$FD,$B2,$3F,$61,$AF,$E1,$1A,$FE,$39,$AF,$E3,$5A,$FD
	.dc.b	$E7,$F0,$DF,$07,$23,$F0,$6D,$7F,$1E,$D7,$F3,$2D,$7F,$30,$D7,$F1
	.dc.b	$2D,$7F,$33,$F0,$7F,$93,$23,$F9,$06,$BF,$9C,$6B,$FA,$66,$BF,$A5
	.dc.b	$6B,$F9,$67,$F3,$DF,$07,$23,$F3,$6D,$7F,$4E,$D7,$F6,$2D,$7F,$60
	.dc.b	$D7,$F4,$2D,$7F,$63,$F0,$7F,$AB,$23,$FA,$86,$BF,$B4,$6B,$FB,$E6
	.dc.b	$BF,$BD,$6B,$FA,$E7,$F6,$DF,$07,$23,$F6,$6D,$7F,$7E,$D7,$F8,$4B
	.dc.b	$5F,$E1,$0D,$7F,$72,$D7,$F8,$4F,$C1,$FF,$03,$23,$FC,$01,$AF,$F0
	.dc.b	$C6,$BF,$C5,$9A,$FF,$15,$6B,$FC,$19,$FE,$1D,$F0,$72,$3F,$85,$B5
	.dc.b	$FE,$2E,$D7,$F9,$0B,$5F,$E4,$0D,$7F,$88,$B5,$FE,$43,$F0,$7F,$C6
	.dc.b	$C8,$FF,$18,$6B,$FC,$91,$AF,$F2,$E6,$BF,$CB,$5A,$FF,$1E,$7F,$93
	.dc.b	$7C,$1C,$8F,$E4,$6D,$7F,$97,$B5,$FE,$72,$D7,$F9,$C3,$5F,$E5,$2D
	.dc.b	$7F,$9C,$FC,$1F,$F3,$32,$3F,$CC,$1A,$FF,$3C,$6B,$FD,$19,$AF,$F4
	.dc.b	$56,$BF,$CD,$9F,$E7,$DF,$07,$23,$F9,$DB,$5F,$E8,$ED,$7F,$A8,$B5
	.dc.b	$FE,$A0,$D7,$FA,$0B,$5F,$EA,$3F,$07,$FD,$2C,$8F,$F4,$86,$BF,$D5
	.dc.b	$1A,$FF,$5E,$6B,$FD,$75,$AF,$F4,$E7,$FA,$B7,$C1,$C8,$FE,$A6,$D7
	.dc.b	$FA,$FB,$5F,$ED,$2D,$7F,$B4,$35,$FE,$B2,$D7,$FB,$4F,$C1,$FF,$63
	.dc.b	$23,$FD,$81,$AF,$F6,$C6,$BF,$DD,$9A,$FF,$75,$6B,$FD,$99,$FE,$DD
	.dc.b	$F0,$72,$3F,$B5,$B5,$FE,$EE,$D7,$FC,$02,$D7,$FC,$00,$D7,$FB,$8B
	.dc.b	$57,$DD,$3E,$13,$FB,$F3,$63,$FC,$0C,$D7,$FC,$16,$D7,$FD,$F0,$D4
	.dc.b	$47,$0D,$8E,$A4,$7A,$B5,$EA,$F8,$BE,$0E,$47,$89,$AF,$89,$AF,$89
	.dc.b	$AF,$88,$F6,$B8,$7C,$8D,$7E,$0F,$C1,$FC,$02,$3F,$01,$AF,$C0,$6B
	.dc.b	$F0,$1A,$F7,$A4,$62,$3E,$13,$F3,$3E,$0E,$47,$CC,$6B,$F3,$1A,$FC
	.dc.b	$C6,$BD,$5A,$FC,$D6,$BF,$6B,$E0,$FE,$C9,$1F,$B2,$D7,$EC,$B5,$FB
	.dc.b	$2D,$7D,$4D,$7E,$D3,$F8,$4F,$83,$91,$F8,$46,$BF,$84,$6B,$F8,$46
	.dc.b	$BF,$15,$AF,$E1,$5A,$FE,$4B,$E0,$FF,$22,$47,$F2,$2D,$7F,$22,$D7
	.dc.b	$F2,$2D,$7E,$93,$5F,$C9,$3F,$9C,$F8,$39,$1F,$9C,$6B,$F9,$C6,$BF
	.dc.b	$9C,$6B,$F7,$5A,$FE,$75,$AF,$EA,$BE,$0F,$F5,$24,$7F,$52,$D7,$F5
	.dc.b	$2D,$7F,$52,$D7,$F1,$4D,$7F,$54,$FE,$D3,$E0,$E4,$7E,$D1,$AF,$ED
	.dc.b	$1A,$FE,$D1,$AF,$E5,$5A,$FE,$D5,$AF,$F0,$2F,$83,$FE,$02,$47,$F8
	.dc.b	$0B,$5F,$E0,$2D,$7F,$80,$B5,$FD,$13,$5F,$E0,$4F,$F0,$CF,$83,$91
	.dc.b	$FC,$31,$AF,$F0,$C6,$BF,$C3,$1A,$FE,$B5,$AF,$F0,$D6,$BF,$C6,$BE
	.dc.b	$0F,$F8,$C9,$1F,$E3,$2D,$7F,$8C,$B5,$FE,$32,$D7,$F7,$4D,$7F,$8D
	.dc.b	$3F,$C9,$3E,$0E,$47,$F2,$46,$BF,$C9,$1A,$FF,$24,$6B,$FC,$15,$AF
	.dc.b	$F2,$56,$BF,$CC,$BE,$0F,$F9,$89,$1F,$E6,$2D,$7F,$98,$B5,$FE,$62
	.dc.b	$D7,$F8,$93,$5F,$E6,$4F,$F3,$CF,$83,$91,$FC,$F1,$AF,$F3,$C6,$BF
	.dc.b	$CF,$1A,$FF,$1D,$6B,$FC,$F5,$AF,$F4,$AF,$83,$FE,$92,$47,$FA,$4B
	.dc.b	$5F,$E9,$2D,$7F,$A4,$B5,$FE,$54,$D7,$FA,$53,$FD,$53,$E0,$E4,$7F
	.dc.b	$54,$6B,$FD,$51,$AF,$F5,$46,$BF,$CD,$5A,$FF,$55,$6B,$FD,$8B,$E0
	.dc.b	$FF,$B0,$91,$FE,$C2,$D7,$FB,$0B,$5F,$EC,$2D,$7F,$A1,$35,$FE,$C4
	.dc.b	$FF,$6C,$F8,$39,$1F,$DB,$1A,$FF,$6C,$6B,$FD,$B1,$AF,$F4,$D6,$BF
	.dc.b	$DB,$5A,$FF,$7A,$F8,$3F,$EF,$24,$7F,$BC,$B5,$FE,$F2,$D7,$FB,$CB
	.dc.b	$5F,$EB,$4D,$7F,$BD,$3F,$F0,$07,$EE,$BF,$B6,$44,$0D,$7F,$B8,$35
	.dc.b	$FF,$01,$B5,$D9,$CC,$87,$17,$44,$72,$77,$73,$F0,$9E,$32,$47,$16
	.dc.b	$BB,$6B,$E8,$84,$62,$3C,$ED,$7C,$EF,$BF,$E1,$24,$79,$5A,$F8,$DA
	.dc.b	$F5,$6B,$F0,$9A,$FC,$26,$BF,$23,$E1,$3E,$61,$1F,$84,$D7,$E0,$35
	.dc.b	$F5,$35,$F9,$ED,$7E,$7B,$F5,$FE,$12,$47,$CD,$6B,$F2,$DA,$FC,$56
	.dc.b	$BF,$69,$AF,$DA,$6B,$F8,$0F,$84,$FC,$21,$1F,$B4,$D7,$EC,$35,$FA
	.dc.b	$4D,$7F,$0E,$D7,$F0,$EF,$E3,$FE,$12,$47,$E1,$5A,$FE,$0D,$AF,$DD
	.dc.b	$6B,$F9,$26,$BF,$92,$6B,$F9,$8F,$84,$FC,$E1,$1F,$C9,$35,$FC,$83
	.dc.b	$5F,$C5,$35,$FC,$FB,$5F,$CF,$BF,$A7,$F8,$49,$1F,$9D,$6B,$F9,$B6
	.dc.b	$BF,$95,$6B,$FA,$A6,$BF,$AA,$6B,$FB,$0F,$84,$FD,$A1,$1F,$D5,$35
	.dc.b	$FD,$43,$5F,$D1,$35,$FD,$BB,$5F,$DB,$BF,$BF,$F8,$49,$1F,$B5,$6B
	.dc.b	$FB,$36,$BF,$AD,$6B,$FC,$09,$AF,$F0,$26,$BF,$C2,$3E,$13,$F8,$61
	.dc.b	$1F,$E0,$4D,$7F,$80,$35,$FD,$D3,$5F,$E1,$ED,$7F,$87,$BF,$C5,$FE
	.dc.b	$12,$47,$F0,$D6,$BF,$C2,$DA,$FF,$05,$6B,$FC,$69,$AF,$F1,$A6,$BF
	.dc.b	$C8,$3E,$13,$F9,$21,$1F,$E3,$4D,$7F,$8C,$35,$FE,$24,$D7,$F9,$3B
	.dc.b	$5F,$E4,$EF,$F2,$FF,$84,$91,$FC,$95,$AF,$F2,$36,$BF,$C7,$5A,$FF
	.dc.b	$32,$6B,$FC,$C9,$AF,$F3,$8F,$84,$FE,$78,$47,$F9,$93,$5F,$E6,$0D
	.dc.b	$7F,$95,$35,$FE,$7E,$D7,$F9,$FB,$FD,$1F,$E1,$24,$7F,$3D,$6B,$FC
	.dc.b	$ED,$AF,$F3,$56,$BF,$D2,$9A,$FF,$4A,$6B,$FD,$43,$E1,$3F,$AA,$11
	.dc.b	$FE,$94,$D7,$FA,$43,$5F,$E8,$4D,$7F,$AB,$B5,$FE,$AE,$FF,$5F,$F8
	.dc.b	$49,$1F,$D5,$5A,$FF,$53,$6B,$FD,$35,$AF,$F6,$26,$BF,$D8,$9A,$FF
	.dc.b	$68,$F8,$4F,$ED,$84,$7F,$B1,$35,$FE,$C0,$D7,$FA,$D3,$5F,$ED,$ED
	.dc.b	$7F,$B7,$BF,$DD,$FE,$12,$47,$F6,$D6,$BF,$DA,$DA,$FF,$65,$6B,$FD
	.dc.b	$E9,$AF,$F7,$A6,$BF,$E0,$0F,$84,$FF,$02,$11,$FE,$F4,$D7,$FB,$C3
	.dc.b	$5F,$EE,$4D,$7F,$C0,$ED,$7F,$C0,$EF,$F8,$27,$CB,$7C,$25,$24,$FF
	.dc.b	$83,$E3,$2D,$8C,$BF,$E1,$58,$CB,$63,$2F,$F8,$6A,$32,$D8,$CA,$24
	.dc.b	$3E,$0F,$C8,$48,$F1,$35,$F0,$35,$EC,$D7,$A0,$8E,$8F,$22,$1E,$9F
	.dc.b	$83,$F8,$44,$7E,$03,$5D,$18,$21,$82,$1E,$F7,$83,$F0,$7E,$0F,$E7
	.dc.b	$11,$F9,$8D,$7E,$53,$5E,$06,$88,$68,$A7,$CD,$F8,$3F,$B4,$47,$EC
	.dc.b	$35,$FA,$CD,$78,$1A,$21,$A2,$9F,$67,$E0,$FF,$0C,$47,$F0,$8D,$7F
	.dc.b	$04,$D7,$81,$A2,$1A,$29,$F8,$5F,$83,$FC,$91,$1F,$C8,$35,$FC,$73
	.dc.b	$5E,$06,$88,$68,$A7,$E4,$7E,$0F,$F3,$C4,$7F,$38,$D7,$F3,$4D,$78
	.dc.b	$1A,$21,$A2,$9F,$9D,$F8,$3F,$D5,$11,$FD,$43,$5F,$D3,$35,$E0,$68
	.dc.b	$86,$8A,$7E,$A7,$E0,$FF,$6C,$47,$F6,$8D,$7F,$64,$D7,$81,$A2,$1A
	.dc.b	$29,$FB,$5F,$83,$FE,$04,$47,$F8,$03,$5F,$DF,$35,$E0,$68,$86,$8A
	.dc.b	$7F,$01,$F8,$3F,$E1,$C4,$7F,$86,$35,$FE,$14,$D7,$81,$A2,$1A,$29
	.dc.b	$FC,$37,$E0,$FF,$8D,$11,$FE,$30,$D7,$F8,$B3,$5E,$06,$88,$68,$A7
	.dc.b	$F1,$9F,$83,$FE,$4C,$47,$F9,$23,$5F,$E4,$4D,$78,$1A,$21,$A2,$9F
	.dc.b	$C9,$7E,$0F,$F9,$91,$1F,$E6,$0D,$7F,$97,$35,$E0,$68,$86,$8A,$7F
	.dc.b	$31,$F8,$3F,$E7,$C4,$7F,$9E,$35,$FE,$74,$D7,$81,$A2,$1A,$29,$FC
	.dc.b	$F7,$E0,$FF,$A5,$11,$FE,$90,$D7,$FA,$33,$5E,$06,$88,$68,$A7,$F4
	.dc.b	$9F,$83,$FE,$AC,$47,$FA,$A3,$5F,$EA,$4D,$78,$1A,$21,$A2,$9F,$D5
	.dc.b	$7E,$0F,$FB,$11,$1F,$EC,$0D,$7F,$AF,$35,$E0,$68,$86,$8A,$7F,$61
	.dc.b	$F8,$3F,$ED,$C4,$7F,$B6,$35,$FE,$D4,$D7,$81,$A2,$1A,$29,$FD,$B7
	.dc.b	$E0,$FF,$BD,$11,$FE,$F0,$D7,$FB,$B3,$5E,$06,$88,$68,$A7,$F6,$A1
	.dc.b	$F7,$4F,$84,$C4,$7F,$6E,$8C,$38,$37,$15,$A6,$63,$6F,$26,$BD,$DA
	.dc.b	$F5,$8E,$51,$88,$FF,$04,$35,$EA,$73,$8D,$FC,$24,$8F,$03,$5F,$03
	.dc.b	$5F,$03,$5F,$63,$5E,$ED,$7E,$1F,$4F,$90,$8F,$B5,$AF,$B5,$AF,$B5
	.dc.b	$AF,$C9,$6B,$EC,$7D,$FF,$07,$23,$E4,$35,$F9,$0D,$7E,$43,$5F,$A8
	.dc.b	$D7,$E3,$B5,$FB,$7F,$07,$F3,$08,$FD,$56,$BF,$55,$AF,$D5,$6B,$F8
	.dc.b	$16,$BF,$51,$FA,$FF,$07,$23,$F0,$0D,$7F,$00,$D7,$F0,$0D,$7F,$18
	.dc.b	$D7,$EF,$B5,$FC,$9F,$C1,$FE,$10,$8F,$E3,$5A,$FE,$35,$AF,$E3,$5A
	.dc.b	$FE,$65,$AF,$E3,$1F,$C7,$FC,$1C,$8F,$CC,$35,$FC,$C3,$5F,$CC,$35
	.dc.b	$FD,$23,$5F,$CB,$B5,$FD,$5F,$C1,$FE,$70,$8F,$E9,$5A,$FE,$95,$AF
	.dc.b	$E9,$5A,$FE,$C5,$AF,$E9,$1F,$D3,$FC,$1C,$8F,$D8,$35,$FD,$83,$5F
	.dc.b	$D8,$35,$FD,$E3,$5F,$D7,$B5,$FE,$07,$F0,$7F,$B4,$23,$FB,$D6,$BF
	.dc.b	$BD,$6B,$FB,$D6,$BF,$C2,$5A,$FE,$F1,$FD,$FF,$C1,$C8,$FE,$10,$D7
	.dc.b	$F8,$43,$5F,$E1,$0D,$7F,$8A,$35,$FE,$0E,$D7,$F8,$DF,$C1,$FF,$0C
	.dc.b	$23,$FC,$55,$AF,$F1,$56,$BF,$C5,$5A,$FF,$21,$6B,$FC,$51,$FE,$2F
	.dc.b	$F0,$72,$3F,$90,$35,$FE,$40,$D7,$F9,$03,$5F,$E5,$8D,$7F,$8F,$B5
	.dc.b	$FE,$67,$F0,$7F,$C9,$08,$FF,$2D,$6B,$FC,$B5,$AF,$F2,$D6,$BF,$CE
	.dc.b	$5A,$FF,$2C,$7F,$97,$FC,$1C,$8F,$E7,$0D,$7F,$9C,$35,$FE,$70,$D7
	.dc.b	$FA,$23,$5F,$E6,$ED,$7F,$A5,$FC,$1F,$F3,$C2,$3F,$D1,$5A,$FF,$45
	.dc.b	$6B,$FD,$15,$AF,$F5,$16,$BF,$D1,$1F,$E8,$FF,$07,$23,$FA,$83,$5F
	.dc.b	$EA,$0D,$7F,$A8,$35,$FE,$B8,$D7,$FA,$7B,$5F,$EC,$7F,$07,$FD,$50
	.dc.b	$8F,$F5,$D6,$BF,$D7,$5A,$FF,$5D,$6B,$FD,$A5,$AF,$F5,$C7,$FA,$FF
	.dc.b	$C1,$C8,$FE,$D0,$D7,$FB,$43,$5F,$ED,$0D,$7F,$BA,$35,$FE,$CE,$D7
	.dc.b	$FB,$DF,$C1,$FF,$6C,$23,$FD,$D5,$AF,$F7,$56,$BF,$DD,$5A,$FF,$80
	.dc.b	$5A,$FF,$74,$7F,$C0,$1F,$75,$FE,$01,$83,$48,$E8,$87,$F8,$25,$CB
	.dc.b	$4C,$9F,$E1,$07,$2D,$F8,$3E,$E4,$79,$35,$D1,$82,$18,$21,$FE,$1F
	.dc.b	$84,$71,$8C,$7C,$1F,$B0,$8F,$9D,$AF,$95,$AF,$03,$44,$34,$53,$C1
	.dc.b	$F0,$7F,$20,$8F,$C3,$6B,$F0,$5A,$F0,$34,$43,$45,$3D,$9F,$07,$F5
	.dc.b	$08,$FC,$F6,$BF,$35,$AF,$03,$44,$34,$53,$E4,$7C,$1F,$E0,$08,$FD
	.dc.b	$B6,$BF,$65,$AF,$03,$44,$34,$53,$EA,$7C,$1F,$E3,$08,$FE,$1D,$AF
	.dc.b	$E1,$5A,$F0,$34,$43,$45,$3F,$01,$F0,$7F,$98,$23,$F9,$36,$BF,$91
	.dc.b	$6B,$C0,$D1,$0D,$14,$FC,$67,$C1,$FE,$90,$8F,$E7,$DA,$FE,$75,$AF
	.dc.b	$03,$44,$34,$53,$F3,$1F,$07,$FB,$02,$3F,$AB,$6B,$FA,$96,$BC,$0D
	.dc.b	$10,$D1,$4F,$D2,$7C,$1F,$EF,$08,$FE,$DD,$AF,$ED,$5A,$F0,$34,$43
	.dc.b	$45,$3F,$61,$F0,$7F,$C2,$08,$FF,$03,$6B,$FC,$05,$AF,$03,$44,$34
	.dc.b	$53,$F7,$9F,$07,$FC,$50,$8F,$F0,$F6,$BF,$C3,$5A,$F0,$34,$43,$45
	.dc.b	$3F,$84,$7C,$1F,$F2,$02,$3F,$C6,$DA,$FF,$19,$6B,$C0,$D1,$0D,$14
	.dc.b	$FE,$29,$F0,$7F,$CB,$08,$FF,$27,$6B,$FC,$95,$AF,$03,$44,$34,$53
	.dc.b	$F9,$07,$C1,$FF,$38,$23,$FC,$CD,$AF,$F3,$16,$BC,$0D,$10,$D1,$4F
	.dc.b	$E5,$9F,$07,$FD,$10,$8F,$F3,$F6,$BF,$CF,$5A,$F0,$34,$43,$45,$3F
	.dc.b	$9C,$7C,$1F,$F5,$02,$3F,$D2,$DA,$FF,$49,$6B,$C0,$D1,$0D,$14,$FE
	.dc.b	$89,$F0,$7F,$D7,$08,$FF,$57,$6B,$FD,$55,$AF,$03,$44,$34,$53,$FA
	.dc.b	$87,$C1,$FF,$68,$23,$FD,$8D,$AF,$F6,$16,$BC,$0D,$10,$D1,$4F,$EB
	.dc.b	$9F,$07,$FD,$D0,$8F,$F6,$F6,$BF,$DB,$5A,$F0,$34,$43,$45,$3F,$B4
	.dc.b	$7C,$1F,$F8,$00,$8F,$F7,$B6,$BF,$DE,$5A,$F0,$34,$43,$45,$04,$1F
	.dc.b	$74,$F8,$4C,$44,$8C,$38,$38,$B8,$70,$D7,$84,$6D,$DB,$5E,$ED,$8E
	.dc.b	$11,$E3,$6B,$C0,$E4,$E1,$E2,$63,$E1,$39,$11,$C1,$0F,$2B,$96,$99
	.dc.b	$3D,$CE,$5A,$78,$3E,$13,$CC,$47,$04,$3E,$0B,$96,$99,$3E,$53,$96
	.dc.b	$9E,$CF,$84,$F8,$44,$70,$43,$E6,$B9,$69,$93,$EB,$39,$69,$F2,$3E
	.dc.b	$13,$E7,$11,$C1,$0F,$B2,$E5,$A6,$4F,$C1,$39,$69,$F5,$3E,$13,$ED
	.dc.b	$11,$C1,$0F,$C2,$B9,$69,$93,$F1,$CE,$5A,$7E,$03,$E1,$3F,$0C,$47
	.dc.b	$04,$3F,$22,$E5,$A6,$4F,$CD,$39,$69,$F8,$CF,$84,$FC,$91,$1C,$10
	.dc.b	$FC,$EB,$96,$99,$3F,$4C,$E5,$A7,$E6,$3E,$13,$F3,$C4,$70,$43,$F5
	.dc.b	$2E,$5A,$64,$FD,$93,$96,$9F,$A4,$F8,$4F,$D5,$11,$C1,$0F,$DA,$B9
	.dc.b	$69,$93,$F7,$CE,$5A,$7E,$C3,$E1,$3F,$6C,$47,$04,$3F,$80,$B9,$69
	.dc.b	$93,$F8,$53,$96,$9F,$BC,$F8,$4F,$E0,$44,$70,$43,$F8,$6B,$96,$99
	.dc.b	$3F,$8B,$39,$69,$FC,$23,$E1,$3F,$87,$11,$C1,$0F,$E3,$2E,$5A,$64
	.dc.b	$FE,$44,$E5,$A7,$F1,$4F,$84,$FE,$34,$47,$04,$3F,$92,$B9,$69,$93
	.dc.b	$F9,$73,$96,$9F,$C8,$3E,$13,$F9,$31,$1C,$10,$FE,$62,$E5,$A6,$4F
	.dc.b	$E7,$4E,$5A,$7F,$2C,$F8,$4F,$E6,$44,$70,$43,$F9,$EB,$96,$99,$3F
	.dc.b	$A3,$39,$69,$FC,$E3,$E1,$3F,$9F,$11,$C1,$0F,$E9,$2E,$5A,$64,$FE
	.dc.b	$A4,$E5,$A7,$F4,$4F,$84,$FE,$94,$47,$04,$3F,$AA,$B9,$69,$93,$FA
	.dc.b	$F3,$96,$9F,$D4,$3E,$13,$FA,$B1,$1C,$10,$FE,$C2,$E5,$A6,$4F,$ED
	.dc.b	$4E,$5A,$7F,$5C,$F8,$4F,$EC,$44,$70,$43,$FB,$6B,$96,$99,$3F,$BB
	.dc.b	$39,$69,$FD,$A3,$E1,$3F,$B7,$11,$C1,$0F,$EF,$2E,$5A,$64,$FF,$01
	.dc.b	$39,$69,$FD,$F7,$95,$F0,$7F,$DF,$5B,$19,$8C,$11,$C1,$0F,$F0,$64
	.dc.b	$70,$6C,$6D,$A6,$CF,$F0,$AC,$3E,$0F,$C0,$48,$F2,$35,$EA,$D7,$8B
	.dc.b	$5F,$F1,$14,$73,$78,$90,$F2,$FC,$1F,$B4,$8F,$C1,$6B,$A3,$04,$30
	.dc.b	$43,$D2,$F0,$7D,$7F,$07,$F2,$48,$FC,$D6,$BF,$19,$AF,$03,$44,$34
	.dc.b	$53,$E3,$FC,$1F,$D5,$23,$F6,$5A,$FD,$26,$BC,$0D,$10,$D1,$4F,$A7
	.dc.b	$F0,$7F,$81,$23,$F8,$56,$BF,$79,$AF,$03,$44,$34,$53,$EF,$FC,$1F
	.dc.b	$E3,$48,$FE,$45,$AF,$E2,$9A,$F0,$34,$43,$45,$3F,$17,$F0,$7F,$99
	.dc.b	$23,$F9,$D6,$BF,$96,$6B,$C0,$D1,$0D,$14,$FC,$BF,$C1,$FE,$94,$8F
	.dc.b	$EA,$5A,$FE,$89,$AF,$03,$44,$34,$53,$F4,$7F,$07,$FB,$12,$3F,$B5
	.dc.b	$6B,$FA,$E6,$BC,$0D,$10,$D1,$4F,$D7,$FC,$1F,$EF,$48,$FF,$01,$6B
	.dc.b	$FB,$A6,$BC,$0D,$10,$D1,$4F,$DD,$FC,$1F,$F0,$92,$3F,$C3,$5A,$FF
	.dc.b	$06,$6B,$C0,$D1,$0D,$14,$FE,$0F,$F0,$7F,$C5,$48,$FF,$19,$6B,$FC
	.dc.b	$49,$AF,$03,$44,$34,$53,$F8,$9F,$C1,$FF,$21,$23,$FC,$95,$AF,$F1
	.dc.b	$E6,$BC,$0D,$10,$D1,$4F,$E3,$FF,$07,$FC,$B4,$8F,$F3,$16,$BF,$CA
	.dc.b	$9A,$F0,$34,$43,$45,$3F,$95,$FC,$1F,$F3,$92,$3F,$CF,$5A,$FF,$36
	.dc.b	$6B,$C0,$D1,$0D,$14,$FE,$6F,$F0,$7F,$D1,$48,$FF,$49,$6B,$FD,$09
	.dc.b	$AF,$03,$44,$34,$53,$FA,$1F,$C1,$FF,$51,$23,$FD,$55,$AF,$F4,$E6
	.dc.b	$BC,$0D,$10,$D1,$4F,$E9,$FF,$07,$FD,$74,$8F,$F6,$16,$BF,$D6,$9A
	.dc.b	$F0,$34,$43,$45,$3F,$AD,$FC,$1F,$F6,$92,$3F,$DB,$5A,$FF,$66,$6B
	.dc.b	$C0,$D1,$0D,$14,$FE,$CF,$F0,$7F,$DD,$48,$FF,$79,$6B,$FD,$C9,$AF
	.dc.b	$03,$44,$34,$50,$41,$F7,$5F,$EF,$F6,$9F,$F0,$05,$C1,$1C,$10,$E4
	.dc.b	$73,$74,$E0,$EE,$E1,$C3,$D3,$E0,$FB,$12,$38,$1E,$47,$0F,$06,$BE
	.dc.b	$08,$46,$23,$C4,$D7,$C2,$FA,$FE,$0E,$47,$69,$46,$23,$CA,$D7,$B3
	.dc.b	$5F,$73,$5F,$6B,$5F,$97,$F0,$7F,$24,$8F,$76,$BF,$09,$AF,$AD,$AF
	.dc.b	$CB,$6B,$F2,$9F,$A9,$F0,$72,$3D,$4D,$7E,$6B,$5F,$8C,$D7,$EB,$35
	.dc.b	$FA,$AD,$7F,$07,$F0,$7F,$81,$23,$F1,$DA,$FD,$A6,$BF,$4D,$AF,$E0
	.dc.b	$DA,$FE,$09,$FC,$67,$C1,$C8,$FA,$4D,$7F,$0A,$D7,$EF,$35,$FC,$73
	.dc.b	$5F,$C6,$B5,$FC,$DF,$C1,$FE,$64,$8F,$DF,$6B,$F9,$26,$BF,$8B,$6B
	.dc.b	$F9,$B6,$BF,$9A,$7F,$49,$F0,$72,$3F,$14,$D7,$F3,$AD,$7F,$2C,$D7
	.dc.b	$F4,$CD,$7F,$4A,$D7,$F6,$7F,$07,$FB,$12,$3F,$97,$6B,$FA,$A6,$BF
	.dc.b	$A3,$6B,$FB,$36,$BF,$B2,$7F,$79,$F0,$72,$3F,$44,$D7,$F6,$AD,$7F
	.dc.b	$5C,$D7,$F7,$CD,$7F,$7A,$D7,$F8,$5F,$C1,$FF,$09,$23,$FA,$F6,$BF
	.dc.b	$C0,$9A,$FE,$ED,$AF,$F0,$B6,$BF,$C2,$9F,$E2,$9F,$07,$23,$F7,$4D
	.dc.b	$7F,$86,$B5,$FE,$0C,$D7,$F8,$B3,$5F,$E2,$AD,$7F,$91,$FC,$1F,$F2
	.dc.b	$12,$3F,$C1,$DA,$FF,$1A,$6B,$FC,$4D,$AF,$F2,$36,$BF,$C8,$9F,$E5
	.dc.b	$9F,$07,$23,$F8,$93,$5F,$E4,$AD,$7F,$8F,$35,$FE,$5C,$D7,$F9,$6B
	.dc.b	$5F,$E7,$7F,$07,$FC,$E4,$8F,$F1,$F6,$BF,$CC,$9A,$FF,$2B,$6B,$FC
	.dc.b	$ED,$AF,$F3,$A7,$FA,$27,$C1,$C8,$FE,$54,$D7,$F9,$EB,$5F,$E6,$CD
	.dc.b	$7F,$A3,$35,$FE,$8A,$D7,$FA,$9F,$C1,$FF,$51,$23,$FC,$DD,$AF,$F4
	.dc.b	$A6,$BF,$D0,$DA,$FF,$53,$6B,$FD,$49,$FE,$B9,$F0,$72,$3F,$A1,$35
	.dc.b	$FE,$AA,$D7,$FA,$73,$5F,$EB,$CD,$7F,$AE,$B5,$FE,$D7,$F0,$7F,$DA
	.dc.b	$48,$FF,$4F,$6B,$FD,$89,$AF,$F5,$B6,$BF,$DA,$DA,$FF,$6A,$7F,$BA
	.dc.b	$7C,$1C,$8F,$EB,$4D,$7F,$B6,$B5,$FE,$CC,$D7,$FB,$B3,$5F,$EE,$AD
	.dc.b	$7F,$C0,$7F,$07,$FE,$01,$23,$FD,$9D,$AF,$F7,$A6,$BF,$DC,$DA,$FF
	.dc.b	$80,$DA,$FF,$80,$9F,$EF,$AB,$EE,$9F,$09,$86,$C7,$F8,$09,$AF,$F8
	.dc.b	$2D,$AF,$F8,$41,$AF,$F8,$6D,$AF,$F8,$79,$AF,$78,$EC,$47,$B3,$5E
	.dc.b	$DF,$07,$E3,$23,$E3,$6B,$A3,$04,$30,$43,$CC,$F0,$7D,$1E,$7F,$BC
	.dc.b	$8F,$BD,$AF,$BD,$AF,$03,$44,$34,$53,$E2,$7C,$1F,$CB,$23,$F2,$DA
	.dc.b	$FC,$B6,$BC,$0D,$10,$D1,$4F,$A1,$F0,$7F,$5C,$8F,$D7,$6B,$F5,$DA
	.dc.b	$F0,$34,$43,$45,$3E,$E7,$C1,$FE,$0C,$8F,$E0,$DA,$FE,$0D,$AF,$03
	.dc.b	$44,$34,$53,$F1,$1F,$07,$F8,$F2,$3F,$8F,$6B,$F8,$F6,$BC,$0D,$10
	.dc.b	$D1,$4F,$CA,$7C,$1F,$E6,$C8,$FE,$6D,$AF,$E6,$DA,$F0,$34,$43,$45
	.dc.b	$3F,$41,$F0,$7F,$A7,$23,$FA,$76,$BF,$A7,$6B,$C0,$D1,$0D,$14,$FD
	.dc.b	$67,$C1,$FE,$CC,$8F,$EC,$DA,$FE,$CD,$AF,$03,$44,$34,$53,$F7,$1F
	.dc.b	$07,$FB,$F2,$3F,$BF,$6B,$FB,$F6,$BC,$0D,$10,$D1,$4F,$E0,$9F,$07
	.dc.b	$FC,$2C,$8F,$F0,$B6,$BF,$C2,$DA,$F0,$34,$43,$45,$3F,$88,$7C,$1F
	.dc.b	$F1,$72,$3F,$C5,$DA,$FF,$17,$6B,$C0,$D1,$0D,$14,$FE,$39,$F0,$7F
	.dc.b	$C8,$C8,$FF,$23,$6B,$FC,$8D,$AF,$03,$44,$34,$53,$F9,$47,$C1,$FF
	.dc.b	$2F,$23,$FC,$BD,$AF,$F2,$F6,$BC,$0D,$10,$D1,$4F,$E6,$9F,$07,$FC
	.dc.b	$EC,$8F,$F3,$B6,$BF,$CE,$DA,$F0,$34,$43,$45,$3F,$A0,$7C,$1F,$F4
	.dc.b	$72,$3F,$D1,$DA,$FF,$47,$6B,$C0,$D1,$0D,$14,$FE,$99,$F0,$7F,$D4
	.dc.b	$C8,$FF,$53,$6B,$FD,$4D,$AF,$03,$44,$34,$53,$FA,$C7,$C1,$FF,$5F
	.dc.b	$23,$FD,$7D,$AF,$F5,$F6,$BC,$0D,$10,$D1,$4F,$EC,$9F,$07,$FD,$AC
	.dc.b	$8F,$F6,$B6,$BF,$DA,$DA,$F0,$34,$43,$45,$3F,$B8,$7C,$1F,$F7,$72
	.dc.b	$3F,$DD,$DA,$FF,$77,$6B,$C0,$D1,$0D,$14,$10,$7D,$D3,$E1,$3F,$C1
	.dc.b	$2D,$8D,$1B,$70,$E0,$E3,$1C,$08,$E4,$C9,$63,$8B,$D1,$AF,$58,$E2
	.dc.b	$F5,$8C,$BE,$85,$F0,$9E,$62,$47,$56,$BC,$5A,$F7,$6B,$EB,$6B,$D5
	.dc.b	$F8,$3F,$09,$23,$D0,$D7,$C8,$D7,$D4,$D7,$E3,$35,$F4,$35,$F9,$FF
	.dc.b	$09,$F3,$88,$FC,$56,$BF,$05,$AF,$C7,$6B,$F4,$DA,$FC,$57,$EC,$FC
	.dc.b	$24,$8F,$A0,$D7,$E6,$35,$FA,$4D,$7E,$F3,$5F,$A0,$D7,$F0,$FF,$09
	.dc.b	$F8,$62,$3F,$75,$AF,$D9,$6B,$F7,$DA,$FE,$2D,$AF,$DD,$7F,$23,$F0
	.dc.b	$92,$3F,$10,$D7,$F0,$8D,$7F,$14,$D7,$F2,$CD,$7F,$10,$D7,$F3,$FF
	.dc.b	$09,$F9,$E2,$3F,$95,$6B,$F9,$16,$BF,$97,$6B,$FA,$36,$BF,$95,$7F
	.dc.b	$53,$F0,$92,$3F,$40,$D7,$F3,$8D,$7F,$44,$D7,$F5,$CD,$7F,$40,$D7
	.dc.b	$F6,$FF,$09,$FB,$62,$3F,$AD,$6B,$FA,$96,$BF,$AF,$6B,$FB,$B6,$BF
	.dc.b	$AD,$7F,$80,$FC,$24,$8F,$DC,$35,$FD,$A3,$5F,$DD,$35,$FE,$0C,$D7
	.dc.b	$F7,$0D,$7F,$87,$FC,$27,$F0,$E2,$3F,$C1,$5A,$FF,$01,$6B,$FC,$1D
	.dc.b	$AF,$F1,$36,$BF,$C1,$5F,$E3,$3F,$09,$23,$F8,$83,$5F,$E1,$8D,$7F
	.dc.b	$89,$35,$FE,$3C,$D7,$F8,$83,$5F,$E4,$FF,$09,$FC,$98,$8F,$F1,$D6
	.dc.b	$BF,$C6,$5A,$FF,$1F,$6B,$FC,$AD,$AF,$F1,$D7,$F9,$8F,$C2,$48,$FE
	.dc.b	$50,$D7,$F9,$23,$5F,$E5,$4D,$7F,$9B,$35,$FE,$50,$D7,$F9,$FF,$C2
	.dc.b	$7F,$3E,$23,$FC,$D5,$AF,$F3,$16,$BF,$CD,$DA,$FF,$43,$6B,$FC,$D5
	.dc.b	$FE,$93,$F0,$92,$3F,$A0,$35,$FE,$78,$D7,$FA,$13,$5F,$E9,$CD,$7F
	.dc.b	$A0,$35,$FE,$AF,$F0,$9F,$D5,$88,$FF,$4D,$6B,$FD,$25,$AF,$F4,$F6
	.dc.b	$BF,$D6,$DA,$FF,$4D,$7F,$B0,$FC,$24,$8F,$EB,$0D,$7F,$AA,$35,$FE
	.dc.b	$B4,$D7,$FB,$33,$5F,$EB,$0D,$7F,$B7,$FC,$27,$F6,$E2,$3F,$D9,$5A
	.dc.b	$FF,$61,$6B,$FD,$9D,$AF,$F7,$36,$BF,$D9,$5F,$EF,$3F,$09,$23,$FB
	.dc.b	$83,$5F,$ED,$8D,$7F,$B9,$35,$FE,$FC,$D7,$FB,$83,$57,$C0,$FF,$7D
	.dc.b	$7F,$C1,$6D,$72,$65,$C1,$1C,$10,$FF,$03,$39,$69,$93,$FC,$1C,$BE
	.dc.b	$13,$A1,$23,$8B,$5D,$B5,$FF,$0C,$C7,$17,$64,$36,$53,$7F,$07,$E9
	.dc.b	$23,$E6,$6B,$E4,$6B,$C0,$D1,$0D,$14,$F1,$C7,$C5,$23,$F0,$9A,$FC
	.dc.b	$06,$BC,$0D,$10,$D1,$4F,$7F,$C1,$FD,$12,$3F,$39,$AF,$CC,$6B,$C0
	.dc.b	$D1,$0D,$14,$F9,$7F,$07,$F7,$48,$FD,$A6,$BF,$61,$AF,$03,$44,$34
	.dc.b	$53,$EB,$FC,$1F,$E2,$48,$FE,$19,$AF,$E1,$1A,$F0,$34,$43,$45,$3F
	.dc.b	$07,$F0,$7F,$95,$23,$F9,$26,$BF,$90,$6B,$C0,$D1,$0D,$14,$FC,$7F
	.dc.b	$C1,$FE,$84,$8F,$E7,$9A,$FE,$71,$AF,$03,$44,$34,$53,$F3,$7F,$07
	.dc.b	$FA,$D2,$3F,$AA,$6B,$FA,$86,$BC,$0D,$10,$D1,$4F,$D3,$FC,$1F,$EE
	.dc.b	$48,$FE,$D9,$AF,$ED,$1A,$F0,$34,$43,$45,$3F,$67,$F0,$7F,$C1,$48
	.dc.b	$FF,$02,$6B,$FC,$01,$AF,$03,$44,$34,$53,$F7,$FF,$07,$FC,$44,$8F
	.dc.b	$F0,$E6,$BF,$C3,$1A,$F0,$34,$43,$45,$3F,$85,$FC,$1F,$F1,$D2,$3F
	.dc.b	$C6,$9A,$FF,$18,$6B,$C0,$D1,$0D,$14,$FE,$2F,$F0,$7F,$CA,$48,$FF
	.dc.b	$26,$6B,$FC,$91,$AF,$03,$44,$34,$53,$F9,$1F,$C1,$FF,$35,$23,$FC
	.dc.b	$C9,$AF,$F3,$06,$BC,$0D,$10,$D1,$4F,$E5,$FF,$07,$FD,$04,$8F,$F3
	.dc.b	$E6,$BF,$CF,$1A,$F0,$34,$43,$45,$3F,$9D,$FC,$1F,$F4,$D2,$3F,$D2
	.dc.b	$9A,$FF,$48,$6B,$C0,$D1,$0D,$14,$FE,$8F,$F0,$7F,$EE,$5F,$7E,$C7
	.dc.b	$FB,$A1,$A9,$A4,$7F,$DD,$9C,$69,$E5,$18,$6B,$FD,$A1,$AF,$F6,$C6
	.dc.b	$C7,$18,$F0,$1C,$4A,$71,$29,$FD,$D3,$E0,$F8,$91,$CC,$23,$11,$E9
	.dc.b	$3C,$EE,$2B,$FB,$3B,$63,$27,$27,$0F,$F8,$03,$E0,$FC,$E4,$7B,$B5
	.dc.b	$EC,$6C,$C9,$0C,$9F,$E0,$C7,$84,$FF,$78,$6C,$7C,$CF,$83,$F8,$A4
	.dc.b	$7D,$AD,$79,$DA,$31,$5F,$E0,$26,$CF,$33,$89,$0F,$1F,$DD,$7A,$25
	.dc.b	$10,$46,$5C,$35,$DB,$5D,$C7,$DE,$8C,$11,$C1,$23,$ED,$C6,$5B,$F0
	.dc.b	$7C,$48,$E0,$87,$A1,$CB,$4C,$CF,$A5,$F4,$C6,$A3,$12,$8F,$37,$C1
	.dc.b	$CA,$F1,$B5,$F8,$AD,$AE,$2D,$8F,$86,$D7,$E1,$B5,$F9,$3F,$07,$FE
	.dc.b	$49,$B3,$49,$1E,$73,$E5,$B8,$71,$1C,$1D,$1D,$0A,$E5,$DC,$71,$7D
	.dc.b	$EF,$87,$E0,$E4,$70,$6B,$D9,$AF,$C6,$6B,$F8,$06,$BF,$80,$6B,$F8
	.dc.b	$6F,$83,$FC,$01,$1F,$BA,$D7,$F1,$ED,$C3,$E9,$B6,$3F,$1A,$D7,$F1
	.dc.b	$AF,$E2,$FE,$0E,$47,$E2,$1A,$FE,$69,$B3,$F7,$9B,$1F,$98,$6B,$F9
	.dc.b	$86,$BF,$9E,$F8,$3F,$CC,$11,$FC,$AB,$5F,$D3,$B6,$7F,$16,$D8,$FD
	.dc.b	$2B,$5F,$D2,$BF,$A3,$F8,$39,$1F,$A0,$6B,$FB,$26,$CF,$E5,$9B,$1F
	.dc.b	$B0,$6B,$FB,$06,$BF,$B6,$F8,$3F,$D8,$11,$FD,$6B,$5F,$DF,$B6,$7F
	.dc.b	$46,$D8,$FD,$EB,$5F,$DE,$BF,$BB,$F8,$39,$1F,$B8,$6B,$FC,$29,$B3
	.dc.b	$FA,$E6,$C7,$F0,$86,$BF,$C2,$1A,$FF,$0E,$F8,$3F,$E1,$04,$7F,$82
	.dc.b	$B5,$FE,$2E,$D9,$FD,$DB,$63,$F8,$AB,$5F,$E2,$AF,$F1,$3F,$83,$FD
	.dc.b	$91,$23,$F5,$2D,$7F,$5E,$D7,$F8,$3B,$5F,$E4,$2D,$7F,$90,$BF,$BD
	.dc.b	$F8,$39,$1F,$B4,$6B,$FB,$A6,$BF,$C4,$9A,$FF,$2C,$6B,$FC,$B1,$AF
	.dc.b	$F3,$2F,$83,$FE,$14,$47,$F8,$0B,$5F,$E0,$ED,$7F,$8F,$B5,$FE,$72
	.dc.b	$D7,$F9,$CB,$FC,$57,$E0,$E4,$7F,$0C,$6B,$FC,$49,$AF,$F2,$A6,$BF
	.dc.b	$D1,$1A,$FF,$44,$6B,$FD,$2B,$E0,$FF,$91,$11,$FE,$32,$D7,$F8,$FB
	.dc.b	$5F,$E6,$ED,$7F,$A8,$B5,$FE,$A2,$FF,$2D,$F8,$39,$1F,$C9,$1A,$FF
	.dc.b	$2A,$6B,$FD,$09,$AF,$F5,$C6,$BF,$D7,$1A,$FF,$62,$F8,$3F,$E7,$44
	.dc.b	$7F,$98,$B5,$FE,$6E,$D7,$FA,$7B,$5F,$ED,$2D,$7F,$B4,$BF,$D1,$7E
	.dc.b	$0E,$47,$F3,$C6,$BF,$D0,$9A,$FF,$5A,$6B,$FD,$D1,$AF,$F7,$46,$BF
	.dc.b	$DE,$BE,$0F,$FA,$91,$1F,$E9,$2D,$7F,$A7,$B5,$FE,$CE,$D7,$FC,$02
	.dc.b	$D7,$FC,$02,$FF,$54,$5F,$74,$F8,$4C,$36,$3F,$AC,$B5,$FE,$C0,$D7
	.dc.b	$FB,$C3,$5F,$F0,$93,$5F,$F0,$AB,$5E,$F1,$D8,$8F,$66,$BD,$BE,$0F
	.dc.b	$C6,$47,$C6,$D7,$46,$08,$60,$87,$99,$E0,$FA,$3E,$0F,$DE,$47,$DE
	.dc.b	$D7,$DE,$D7,$81,$A2,$1A,$29,$F1,$3E,$0F,$E5,$91,$F9,$6D,$7E,$5B
	.dc.b	$5E,$06,$88,$68,$A7,$D0,$F8,$3F,$AE,$47,$EB,$B5,$FA,$ED,$78,$1A
	.dc.b	$21,$A2,$9F,$73,$E0,$FF,$06,$47,$F0,$6D,$7F,$06,$D7,$81,$A2,$1A
	.dc.b	$29,$F8,$8F,$83,$FC,$79,$1F,$C7,$B5,$FC,$7B,$5E,$06,$88,$68,$A7
	.dc.b	$E5,$3E,$0F,$F3,$64,$7F,$36,$D7,$F3,$6D,$78,$1A,$21,$A2,$9F,$A0
	.dc.b	$F8,$3F,$D3,$91,$FD,$3B,$5F,$D3,$B5,$E0,$68,$86,$8A,$7E,$B3,$E0
	.dc.b	$FF,$66,$47,$F6,$6D,$7F,$66,$D7,$81,$A2,$1A,$29,$FB,$8F,$83,$FD
	.dc.b	$F9,$1F,$DF,$B5,$FD,$FB,$5E,$06,$88,$68,$A7,$F0,$4F,$83,$FE,$16
	.dc.b	$47,$F8,$5B,$5F,$E1,$6D,$78,$1A,$21,$A2,$9F,$C4,$3E,$0F,$F8,$B9
	.dc.b	$1F,$E2,$ED,$7F,$8B,$B5,$E0,$68,$86,$8A,$7F,$1C,$F8,$3F,$E4,$64
	.dc.b	$7F,$91,$B5,$FE,$46,$D7,$81,$A2,$1A,$29,$FC,$A3,$E0,$FF,$97,$91
	.dc.b	$FE,$5E,$D7,$F9,$7B,$5E,$06,$88,$68,$A7,$F3,$4F,$83,$FE,$76,$47
	.dc.b	$F9,$DB,$5F,$E7,$6D,$78,$1A,$21,$A2,$9F,$D0,$3E,$0F,$FA,$39,$1F
	.dc.b	$E8,$ED,$7F,$A3,$B5,$E0,$68,$86,$8A,$7F,$4C,$F8,$3F,$EA,$64,$7F
	.dc.b	$A9,$B5,$FE,$A6,$D7,$81,$A2,$1A,$29,$FD,$63,$E0,$FF,$AF,$91,$FE
	.dc.b	$BE,$D7,$FA,$FB,$5E,$06,$88,$68,$A7,$F6,$4F,$83,$FE,$D6,$47,$FB
	.dc.b	$5B,$5F,$ED,$6D,$78,$1A,$21,$A2,$9F,$DC,$3E,$0F,$FB,$B9,$1F,$EE
	.dc.b	$ED,$7F,$BB,$B5,$E0,$68,$86,$8A,$7F,$BC,$FE,$EB,$FC,$0C,$9A,$47
	.dc.b	$FC,$14,$D7,$FD,$BE,$D7,$FD,$CC,$D7,$A1,$C0,$E3,$F0,$91,$E2,$D7
	.dc.b	$8B,$5F,$23,$51,$21,$19,$8C,$35,$F5,$7C,$27,$A0,$8F,$A1,$AF,$A1
	.dc.b	$AF,$C3,$6B,$E0,$39,$9C,$48,$7C,$0F,$4F,$E2,$91,$F8,$AD,$74,$60
	.dc.b	$86,$08,$7D,$37,$83,$F5,$7E,$13,$E8,$91,$FA,$2D,$7E,$8B,$5E,$06
	.dc.b	$88,$68,$A7,$E0,$7E,$13,$EE,$91,$FB,$AD,$7E,$EB,$5E,$06,$88,$68
	.dc.b	$A7,$E3,$7E,$13,$F1,$24,$7F,$12,$D7,$F1,$2D,$78,$1A,$21,$A2,$9F
	.dc.b	$99,$F8,$4F,$CA,$91,$FC,$AB,$5F,$CA,$B5,$E0,$68,$86,$8A,$7E,$97
	.dc.b	$E1,$3F,$42,$47,$F4,$2D,$7F,$42,$D7,$81,$A2,$1A,$29,$FB,$1F,$84
	.dc.b	$FD,$69,$1F,$D6,$B5,$FD,$6B,$5E,$06,$88,$68,$A7,$EF,$7E,$13,$F7
	.dc.b	$24,$7F,$72,$D7,$F7,$2D,$78,$1A,$21,$A2,$9F,$C2,$7E,$13,$F8,$29
	.dc.b	$1F,$E0,$AD,$7F,$82,$B5,$E0,$68,$86,$8A,$7F,$15,$F8,$4F,$E2,$24
	.dc.b	$7F,$88,$B5,$FE,$22,$D7,$81,$A2,$1A,$29,$FC,$87,$E1,$3F,$8E,$91
	.dc.b	$FE,$3A,$D7,$F8,$EB,$5E,$06,$88,$68,$A7,$F2,$DF,$84,$FE,$52,$47
	.dc.b	$F9,$4B,$5F,$E5,$2D,$78,$1A,$21,$A2,$9F,$CE,$7E,$13,$F9,$A9,$1F
	.dc.b	$E6,$AD,$7F,$9A,$B5,$E0,$68,$86,$8A,$7F,$45,$F8,$4F,$E8,$24,$7F
	.dc.b	$A0,$B5,$FE,$82,$D7,$81,$A2,$1A,$29,$FD,$47,$E1,$3F,$A6,$91,$FE
	.dc.b	$9A,$D7,$FA,$6B,$5E,$06,$88,$68,$A7,$F5,$DF,$84,$FE,$B2,$47,$FA
	.dc.b	$CB,$5F,$EB,$2D,$78,$1A,$21,$A2,$9F,$DA,$7E,$13,$FB,$29,$1F,$EC
	.dc.b	$AD,$7F,$B2,$B5,$E0,$68,$86,$8A,$7F,$75,$F8,$4F,$EE,$24,$7F,$B8
	.dc.b	$B5,$FE,$E2,$D7,$81,$A2,$1A,$29,$FE,$01,$F8,$4F,$EF,$A4,$7F,$BE
	.dc.b	$B5,$FE,$FA,$D7,$81,$A2,$1A,$29,$FD,$A4,$79,$4F,$67,$11,$FD,$B6
	.dc.b	$30,$E0,$DC,$56,$99,$8D,$BC,$9A,$F7,$6B,$D6,$39,$46,$23,$FC,$16
	.dc.b	$D7,$A9,$CE,$37,$F0,$92,$3C,$0D,$7C,$0D,$7C,$0D,$7D,$8D,$7B,$B5
	.dc.b	$F8,$7F,$09,$E4,$23,$ED,$6B,$ED,$6B,$ED,$6B,$F2,$5A,$FB,$1F,$7F
	.dc.b	$C1,$C8,$F9,$0D,$7E,$43,$5F,$90,$D7,$EA,$35,$F8,$ED,$7E,$DF,$C1
	.dc.b	$FC,$C2,$3F,$55,$AF,$D5,$6B,$F5,$5A,$FE,$05,$AF,$D4,$7E,$BF,$C1
	.dc.b	$C8,$FC,$03,$5F,$C0,$35,$FC,$03,$5F,$C6,$35,$FB,$ED,$7F,$27,$F0
	.dc.b	$7F,$84,$23,$F8,$D6,$BF,$8D,$6B,$F8,$D6,$BF,$99,$6B,$F8,$C7,$F1
	.dc.b	$FF,$07,$23,$F3,$0D,$7F,$30,$D7,$F3,$0D,$7F,$48,$D7,$F2,$ED,$7F
	.dc.b	$57,$F0,$7F,$9C,$23,$FA,$56,$BF,$A5,$6B,$FA,$56,$BF,$B1,$6B,$FA
	.dc.b	$47,$F4,$FF,$07,$23,$F6,$0D,$7F,$60,$D7,$F6,$0D,$7F,$78,$D7,$F5
	.dc.b	$ED,$7F,$81,$FC,$1F,$ED,$08,$FE,$F5,$AF,$EF,$5A,$FE,$F5,$AF,$F0
	.dc.b	$96,$BF,$BC,$7F,$7F,$F0,$72,$3F,$84,$35,$FE,$10,$D7,$F8,$43,$5F
	.dc.b	$E2,$8D,$7F,$83,$B5,$FE,$37,$F0,$7F,$C3,$08,$FF,$15,$6B,$FC,$55
	.dc.b	$AF,$F1,$56,$BF,$C8,$5A,$FF,$14,$7F,$8B,$FC,$1C,$8F,$E4,$0D,$7F
	.dc.b	$90,$35,$FE,$40,$D7,$F9,$63,$5F,$E3,$ED,$7F,$99,$FC,$1F,$F2,$42
	.dc.b	$3F,$CB,$5A,$FF,$2D,$6B,$FC,$B5,$AF,$F3,$96,$BF,$CB,$1F,$E5,$FF
	.dc.b	$07,$23,$F9,$C3,$5F,$E7,$0D,$7F,$9C,$35,$FE,$88,$D7,$F9,$BB,$5F
	.dc.b	$E9,$7F,$07,$FC,$F0,$8F,$F4,$56,$BF,$D1,$5A,$FF,$45,$6B,$FD,$45
	.dc.b	$AF,$F4,$47,$FA,$3F,$C1,$C8,$FE,$A0,$D7,$FA,$83,$5F,$EA,$0D,$7F
	.dc.b	$AE,$35,$FE,$9E,$D7,$FB,$1F,$C1,$FF,$54,$23,$FD,$75,$AF,$F5,$D6
	.dc.b	$BF,$D7,$5A,$FF,$69,$6B,$FD,$71,$FE,$BF,$F0,$72,$3F,$B4,$35,$FE
	.dc.b	$D0,$D7,$FB,$43,$5F,$EE,$8D,$7F,$B3,$B5,$FE,$F7,$F0,$7F,$DB,$08
	.dc.b	$FF,$75,$6B,$FD,$D5,$AF,$F7,$56,$BF,$E0,$16,$BF,$DD,$1F,$F0,$07
	.dc.b	$DD,$3E,$0E,$92,$7F,$C0,$D1,$96,$C6,$5F,$F0,$6C,$65,$B1,$97,$FC
	.dc.b	$29,$19,$6C,$65,$FF,$80,$10,$F8,$3F,$19,$23,$C2,$D7,$BB,$5E,$AD
	.dc.b	$79,$B5,$08,$EC,$FB,$3E,$0F,$E0,$91,$C1,$0F,$6B,$96,$99,$3D,$AE
	.dc.b	$5A,$7C,$8E,$CF,$9A,$47,$04,$3E,$4B,$96,$99,$3E,$4B,$96,$9F,$53
	.dc.b	$E0,$FE,$C9,$1C,$10,$FA,$AE,$5A,$64,$FA,$AE,$5A,$7E,$03,$E0,$FF
	.dc.b	$0A,$47,$04,$3F,$02,$E5,$A6,$4F,$C0,$B9,$69,$F8,$CF,$83,$FC,$89
	.dc.b	$1C,$10,$FC,$6B,$96,$99,$3F,$1A,$E5,$A7,$E6,$3E,$0F,$F3,$A4,$70
	.dc.b	$43,$F3,$2E,$5A,$64,$FC,$CB,$96,$9F,$A4,$F8,$3F,$D4,$91,$C1,$0F
	.dc.b	$D2,$B9,$69,$93,$F4,$AE,$5A,$7E,$C3,$E0,$FF,$6A,$47,$04,$3F,$62
	.dc.b	$E5,$A6,$4F,$D8,$B9,$69,$FB,$CF,$83,$FE,$02,$47,$04,$3F,$7A,$E5
	.dc.b	$A6,$4F,$DE,$B9,$69,$FC,$23,$E0,$FF,$86,$91,$C1,$0F,$E1,$2E,$5A
	.dc.b	$64,$FE,$12,$E5,$A7,$F1,$4F,$83,$FE,$32,$47,$04,$3F,$8A,$B9,$69
	.dc.b	$93,$F8,$AB,$96,$9F,$C8,$3E,$0F,$F9,$29,$1C,$10,$FE,$42,$E5,$A6
	.dc.b	$4F,$E4,$2E,$5A,$7F,$2C,$F8,$3F,$E6,$24,$70,$43,$F9,$6B,$96,$99
	.dc.b	$3F,$96,$B9,$69,$FC,$E3,$E0,$FF,$9E,$91,$C1,$0F,$E7,$2E,$5A,$64
	.dc.b	$FE,$72,$E5,$A7,$F4,$4F,$83,$FE,$92,$47,$04,$3F,$A2,$B9,$69,$93
	.dc.b	$FA,$2B,$96,$9F,$D4,$3E,$0F,$FA,$A9,$1C,$10,$FE,$A2,$E5,$A6,$4F
	.dc.b	$EA,$2E,$5A,$7F,$5C,$F8,$3F,$EC,$24,$70,$43,$FA,$EB,$96,$99,$3F
	.dc.b	$AE,$B9,$69,$FD,$A3,$E0,$FF,$B6,$91,$C1,$0F,$ED,$2E,$5A,$64,$FE
	.dc.b	$D2,$E5,$A7,$F7,$4F,$83,$FE,$F2,$47,$04,$3F,$BA,$B9,$69,$93,$FB
	.dc.b	$AB,$96,$9F,$DA,$08,$3E,$EA,$B8,$6B,$B6,$BF,$DC,$1A,$F1,$32,$E0
	.dc.b	$8E,$08,$74,$72,$DF,$83,$E2,$47,$04,$3C,$B0,$8D,$46,$1A,$FA,$5A
	.dc.b	$FA,$5F,$2F,$C1,$C8,$F1,$35,$EA,$D7,$E1,$35,$F8,$4D,$7E,$13,$5F
	.dc.b	$95,$F0,$7F,$08,$8F,$BD,$AF,$A9,$AF,$CF,$6B,$F3,$DA,$FC,$F7,$E6
	.dc.b	$FC,$1C,$8F,$94,$D7,$E2,$B5,$FB,$4D,$7E,$D3,$5F,$B4,$D7,$F0,$5F
	.dc.b	$07,$F6,$88,$FD,$76,$BF,$49,$AF,$E1,$DA,$FE,$1D,$AF,$E1,$DF,$C2
	.dc.b	$FC,$1C,$8F,$C1,$35,$FB,$AD,$7F,$24,$D7,$F2,$4D,$7F,$24,$D7,$F3
	.dc.b	$5F,$07,$F9,$22,$3F,$8F,$6B,$F8,$A6,$BF,$9F,$6B,$F9,$F6,$BF,$9F
	.dc.b	$7F,$3B,$F0,$72,$3F,$34,$D7,$F2,$AD,$7F,$54,$D7,$F5,$4D,$7F,$54
	.dc.b	$D7,$F6,$5F,$07,$FA,$A2,$3F,$A7,$6B,$FA,$26,$BF,$B7,$6B,$FB,$76
	.dc.b	$BF,$B7,$7F,$6B,$F0,$72,$3F,$64,$D7,$F5,$AD,$7F,$81,$35,$FE,$04
	.dc.b	$D7,$F8,$13,$5F,$E1,$5F,$07,$FC,$08,$8F,$EF,$DA,$FE,$E9,$AF,$F0
	.dc.b	$F6,$BF,$C3,$DA,$FF,$0F,$7F,$86,$FC,$1C,$8F,$E1,$4D,$7F,$82,$B5
	.dc.b	$FE,$34,$D7,$F8,$D3,$5F,$E3,$4D,$7F,$91,$7C,$1F,$F1,$A2,$3F,$C5
	.dc.b	$DA,$FF,$12,$6B,$FC,$9D,$AF,$F2,$76,$BF,$C9,$DF,$E4,$BF,$07,$23
	.dc.b	$F9,$13,$5F,$E3,$AD,$7F,$99,$35,$FE,$64,$D7,$F9,$93,$5F,$E7,$5F
	.dc.b	$07,$FC,$C8,$8F,$F2,$F6,$BF,$CA,$9A,$FF,$3F,$6B,$FC,$FD,$AF,$F3
	.dc.b	$F7,$F9,$EF,$C1,$C8,$FE,$74,$D7,$F9,$AB,$5F,$E9,$4D,$7F,$A5,$35
	.dc.b	$FE,$94,$D7,$FA,$97,$C1,$FF,$4A,$23,$FD,$1D,$AF,$F4,$26,$BF,$D5
	.dc.b	$DA,$FF,$57,$6B,$FD,$5D,$FE,$AB,$F0,$72,$3F,$A9,$35,$FE,$9A,$D7
	.dc.b	$FB,$13,$5F,$EC,$4D,$7F,$B1,$35,$FE,$D5,$F0,$7F,$D8,$88,$FF,$5F
	.dc.b	$6B,$FD,$69,$AF,$F6,$F6,$BF,$DB,$DA,$FF,$6F,$7F,$B6,$FC,$1C,$8F
	.dc.b	$ED,$4D,$7F,$B2,$B5,$FE,$F4,$D7,$FB,$D3,$5F,$EF,$4D,$7F,$C0,$5F
	.dc.b	$07,$FD,$E8,$8F,$F7,$76,$BF,$DC,$9A,$FF,$81,$DA,$FF,$81,$DA,$FF
	.dc.b	$81,$DF,$F0,$4F,$DD,$3E,$12,$92,$7F,$C1,$F1,$96,$C6,$5F,$F0,$AC
	.dc.b	$65,$B1,$97,$FC,$37,$19,$6C,$65,$FF,$10,$43,$E1,$3C,$64,$8F,$0B
	.dc.b	$5E,$ED,$7A,$B5,$E6,$D4,$23,$B3,$EC,$F8,$4F,$82,$47,$04,$3D,$AE
	.dc.b	$5A,$64,$F6,$B9,$69,$F2,$3E,$13,$E6,$91,$C1,$0F,$92,$E5,$A6,$4F
	.dc.b	$92,$E5,$A7,$D4,$F8,$4F,$B2,$47,$04,$3E,$AB,$96,$99,$3E,$AB,$96
	.dc.b	$9F,$80,$F8,$4F,$C2,$91,$C1,$0F,$C0,$B9,$69,$93,$F0,$2E,$5A,$7E
	.dc.b	$33,$E1,$3F,$22,$47,$04,$3F,$1A,$E5,$A6,$4F,$C6,$B9,$69,$F9,$8F
	.dc.b	$84,$FC,$E9,$1C,$10,$FC,$CB,$96,$99,$3F,$32,$E5,$A7,$E9,$3E,$13
	.dc.b	$F5,$24,$70,$43,$F4,$AE,$5A,$64,$FD,$2B,$96,$9F,$B0,$F8,$4F,$DA
	.dc.b	$91,$C1,$0F,$D8,$B9,$69,$93,$F6,$2E,$5A,$7E,$F3,$E1,$3F,$80,$91
	.dc.b	$C1,$0F,$DE,$B9,$69,$93,$F7,$AE,$5A,$7F,$08,$F8,$4F,$E1,$A4,$70
	.dc.b	$43,$F8,$4B,$96,$99,$3F,$84,$B9,$69,$FC,$53,$E1,$3F,$8C,$91,$C1
	.dc.b	$0F,$E2,$AE,$5A,$64,$FE,$2A,$E5,$A7,$F2,$0F,$84,$FE,$4A,$47,$04
	.dc.b	$3F,$90,$B9,$69,$93,$F9,$0B,$96,$9F,$CB,$3E,$13,$F9,$89,$1C,$10
	.dc.b	$FE,$5A,$E5,$A6,$4F,$E5,$AE,$5A,$7F,$38,$F8,$4F,$E7,$A4,$70,$43
	.dc.b	$F9,$CB,$96,$99,$3F,$9C,$B9,$69,$FD,$13,$E1,$3F,$A4,$91,$C1,$0F
	.dc.b	$E8,$AE,$5A,$64,$FE,$8A,$E5,$A7,$F5,$0F,$84,$FE,$AA,$47,$04,$3F
	.dc.b	$A8,$B9,$69,$93,$FA,$8B,$96,$9F,$D7,$3E,$13,$FB,$09,$1C,$10,$FE
	.dc.b	$BA,$E5,$A6,$4F,$EB,$AE,$5A,$7F,$68,$F8,$4F,$ED,$A4,$70,$43,$FB
	.dc.b	$4B,$96,$99,$3F,$B4,$B9,$69,$FD,$D3,$E1,$3F,$BC,$91,$C1,$0F,$EE
	.dc.b	$AE,$5A,$64,$FE,$EA,$E5,$A7,$F6,$82,$0E,$47,$C2,$62,$35,$18,$70
	.dc.b	$6E,$2B,$4C,$C6,$DE,$4D,$7B,$B5,$EB,$1C,$A3,$11,$FD,$F1,$AF,$53
	.dc.b	$9C,$6F,$E1,$24,$78,$1A,$F8,$1A,$F8,$1A,$FB,$1A,$F7,$6B,$F0,$FE
	.dc.b	$0F,$C8,$47,$DA,$D7,$DA,$D7,$DA,$D7,$E4,$B5,$F6,$3E,$FF,$83,$91
	.dc.b	$F2,$1A,$FC,$86,$BF,$21,$AF,$D4,$6B,$F1,$DA,$FD,$BF,$83,$F9,$84
	.dc.b	$7E,$AB,$5F,$AA,$D7,$EA,$B5,$FC,$0B,$5F,$A8,$FD,$7F,$83,$91,$F8
	.dc.b	$06,$BF,$80,$6B,$F8,$06,$BF,$8C,$6B,$F7,$DA,$FE,$4F,$E0,$FF,$08
	.dc.b	$47,$F1,$AD,$7F,$1A,$D7,$F1,$AD,$7F,$32,$D7,$F1,$8F,$E3,$FE,$0E
	.dc.b	$47,$E6,$1A,$FE,$61,$AF,$E6,$1A,$FE,$91,$AF,$E5,$DA,$FE,$AF,$E0
	.dc.b	$FF,$38,$47,$F4,$AD,$7F,$4A,$D7,$F4,$AD,$7F,$62,$D7,$F4,$8F,$E9
	.dc.b	$FE,$0E,$47,$EC,$1A,$FE,$C1,$AF,$EC,$1A,$FE,$F1,$AF,$EB,$DA,$FF
	.dc.b	$03,$F8,$3F,$DA,$11,$FD,$EB,$5F,$DE,$B5,$FD,$EB,$5F,$E1,$2D,$7F
	.dc.b	$78,$FE,$FF,$E0,$E4,$7F,$08,$6B,$FC,$21,$AF,$F0,$86,$BF,$C5,$1A
	.dc.b	$FF,$07,$6B,$FC,$6F,$E0,$FF,$86,$11,$FE,$2A,$D7,$F8,$AB,$5F,$E2
	.dc.b	$AD,$7F,$90,$B5,$FE,$28,$FF,$17,$F8,$39,$1F,$C8,$1A,$FF,$20,$6B
	.dc.b	$FC,$81,$AF,$F2,$C6,$BF,$C7,$DA,$FF,$33,$F8,$3F,$E4,$84,$7F,$96
	.dc.b	$B5,$FE,$5A,$D7,$F9,$6B,$5F,$E7,$2D,$7F,$96,$3F,$CB,$FE,$0E,$47
	.dc.b	$F3,$86,$BF,$CE,$1A,$FF,$38,$6B,$FD,$11,$AF,$F3,$76,$BF,$D2,$FE
	.dc.b	$0F,$F9,$E1,$1F,$E8,$AD,$7F,$A2,$B5,$FE,$8A,$D7,$FA,$8B,$5F,$E8
	.dc.b	$8F,$F4,$7F,$83,$91,$FD,$41,$AF,$F5,$06,$BF,$D4,$1A,$FF,$5C,$6B
	.dc.b	$FD,$3D,$AF,$F6,$3F,$83,$FE,$A8,$47,$FA,$EB,$5F,$EB,$AD,$7F,$AE
	.dc.b	$B5,$FE,$D2,$D7,$FA,$E3,$FD,$7F,$E0,$E4,$7F,$68,$6B,$FD,$A1,$AF
	.dc.b	$F6,$86,$BF,$DD,$1A,$FF,$67,$6B,$FD,$EF,$E0,$FF,$B6,$11,$FE,$EA
	.dc.b	$D7,$FB,$AB,$5F,$EE,$AD,$7F,$C0,$2D,$7F,$BA,$3F,$E0,$0F,$BA,$FF
	.dc.b	$00,$C1,$A4,$74,$43,$FC,$12,$E5,$A6,$4F,$F0,$83,$96,$FC,$1F,$72
	.dc.b	$3C,$9A,$E8,$C1,$0C,$10,$FF,$0F,$C2,$38,$C6,$3E,$0F,$D8,$47,$CE
	.dc.b	$D7,$CA,$D7,$81,$A2,$1A,$29,$E0,$F8,$3F,$90,$47,$E1,$B5,$F8,$2D
	.dc.b	$78,$1A,$21,$A2,$9E,$CF,$83,$FA,$84,$7E,$7B,$5F,$9A,$D7,$81,$A2
	.dc.b	$1A,$29,$F2,$3E,$0F,$F0,$04,$7E,$DB,$5F,$B2,$D7,$81,$A2,$1A,$29
	.dc.b	$F5,$3E,$0F,$F1,$84,$7F,$0E,$D7,$F0,$AD,$78,$1A,$21,$A2,$9F,$80
	.dc.b	$F8,$3F,$CC,$11,$FC,$9B,$5F,$C8,$B5,$E0,$68,$86,$8A,$7E,$33,$E0
	.dc.b	$FF,$48,$47,$F3,$ED,$7F,$3A,$D7,$81,$A2,$1A,$29,$F9,$8F,$83,$FD
	.dc.b	$81,$1F,$D5,$B5,$FD,$4B,$5E,$06,$88,$68,$A7,$E9,$3E,$0F,$F7,$84
	.dc.b	$7F,$6E,$D7,$F6,$AD,$78,$1A,$21,$A2,$9F,$B0,$F8,$3F,$E1,$04,$7F
	.dc.b	$81,$B5,$FE,$02,$D7,$81,$A2,$1A,$29,$FB,$CF,$83,$FE,$28,$47,$F8
	.dc.b	$7B,$5F,$E1,$AD,$78,$1A,$21,$A2,$9F,$C2,$3E,$0F,$F9,$01,$1F,$E3
	.dc.b	$6D,$7F,$8C,$B5,$E0,$68,$86,$8A,$7F,$14,$F8,$3F,$E5,$84,$7F,$93
	.dc.b	$B5,$FE,$4A,$D7,$81,$A2,$1A,$29,$FC,$83,$E0,$FF,$9C,$11,$FE,$66
	.dc.b	$D7,$F9,$8B,$5E,$06,$88,$68,$A7,$F2,$CF,$83,$FE,$88,$47,$F9,$FB
	.dc.b	$5F,$E7,$AD,$78,$1A,$21,$A2,$9F,$CE,$3E,$0F,$FA,$81,$1F,$E9,$6D
	.dc.b	$7F,$A4,$B5,$E0,$68,$86,$8A,$7F,$44,$F8,$3F,$EB,$84,$7F,$AB,$B5
	.dc.b	$FE,$AA,$D7,$81,$A2,$1A,$29,$FD,$43,$E0,$FF,$B4,$11,$FE,$C6,$D7
	.dc.b	$FB,$0B,$5E,$06,$88,$68,$A7,$F5,$CF,$83,$FE,$E8,$47,$FB,$7B,$5F
	.dc.b	$ED,$AD,$78,$1A,$21,$A2,$9F,$DA,$3E,$0F,$FC,$00,$47,$FB,$DB,$5F
	.dc.b	$EF,$2D,$78,$1A,$21,$A2,$82,$0F,$BA,$7C,$26,$22,$46,$1C,$1C,$5C
	.dc.b	$38,$6B,$C2,$36,$ED,$AF,$76,$C7,$08,$F1,$B5,$E0,$72,$70,$F1,$31
	.dc.b	$F0,$9C,$88,$E0,$87,$95,$CB,$4C,$9E,$E7,$2D,$3C,$1F,$07,$E6,$23
	.dc.b	$82,$1F,$05,$CB,$4C,$9F,$29,$CB,$4F,$67,$A5,$F0,$88,$E0,$87,$CD
	.dc.b	$72,$D3,$27,$D6,$72,$D3,$E4,$7C,$1F,$CE,$23,$82,$1F,$65,$CB,$4C
	.dc.b	$9F,$82,$72,$D3,$EA,$7C,$1F,$DA,$23,$82,$1F,$85,$72,$D3,$27,$E3
	.dc.b	$9C,$B4,$FC,$07,$C1,$FE,$18,$8E,$08,$7E,$45,$CB,$4C,$9F,$9A,$72
	.dc.b	$D3,$F1,$9F,$07,$F9,$22,$38,$21,$F9,$D7,$2D,$32,$7E,$99,$CB,$4F
	.dc.b	$CC,$7C,$1F,$E7,$88,$E0,$87,$EA,$5C,$B4,$C9,$FB,$27,$2D,$3F,$49
	.dc.b	$F0,$7F,$AA,$23,$82,$1F,$B5,$72,$D3,$27,$EF,$9C,$B4,$FD,$87,$C1
	.dc.b	$FE,$D8,$8E,$08,$7F,$01,$72,$D3,$27,$F0,$A7,$2D,$3F,$79,$F0,$7F
	.dc.b	$C0,$88,$E0,$87,$F0,$D7,$2D,$32,$7F,$16,$72,$D3,$F8,$47,$C1,$FF
	.dc.b	$0E,$23,$82,$1F,$C6,$5C,$B4,$C9,$FC,$89,$CB,$4F,$E2,$9F,$07,$FC
	.dc.b	$68,$8E,$08,$7F,$25,$72,$D3,$27,$F2,$E7,$2D,$3F,$90,$7C,$1F,$F2
	.dc.b	$62,$38,$21,$FC,$C5,$CB,$4C,$9F,$CE,$9C,$B4,$FE,$59,$F0,$7F,$CC
	.dc.b	$88,$E0,$87,$F3,$D7,$2D,$32,$7F,$46,$72,$D3,$F9,$C7,$C1,$FF,$3E
	.dc.b	$23,$82,$1F,$D2,$5C,$B4,$C9,$FD,$49,$CB,$4F,$E8,$9F,$07,$FD,$28
	.dc.b	$8E,$08,$7F,$55,$72,$D3,$27,$F5,$E7,$2D,$3F,$A8,$7C,$1F,$F5,$62
	.dc.b	$38,$21,$FD,$85,$CB,$4C,$9F,$DA,$9C,$B4,$FE,$B9,$F0,$7F,$D8,$88
	.dc.b	$E0,$87,$F6,$D7,$2D,$32,$7F,$76,$72,$D3,$FB,$47,$C1,$FF,$6E,$23
	.dc.b	$82,$1F,$DE,$5C,$B4,$C9,$FE,$02,$72,$D3,$FB,$EF,$DD,$3E,$13,$FB
	.dc.b	$EB,$63,$31,$82,$38,$21,$FE,$0C,$8E,$0D,$8D,$B4,$D9,$FE,$15,$87
	.dc.b	$C2,$78,$09,$1E,$46,$BD,$5A,$F1,$6B,$FE,$22,$8E,$6F,$12,$1E,$5F
	.dc.b	$84,$F6,$91,$F8,$2D,$74,$60,$86,$08,$7A,$5E,$0F,$AF,$E1,$3E,$49
	.dc.b	$1F,$9A,$D7,$E3,$35,$E0,$68,$86,$8A,$7C,$7F,$84,$FA,$A4,$7E,$CB
	.dc.b	$5F,$A4,$D7,$81,$A2,$1A,$29,$F4,$FE,$13,$F0,$24,$7F,$0A,$D7,$EF
	.dc.b	$35,$E0,$68,$86,$8A,$7D,$FF,$84,$FC,$69,$1F,$C8,$B5,$FC,$53,$5E
	.dc.b	$06,$88,$68,$A7,$E2,$FE,$13,$F3,$24,$7F,$3A,$D7,$F2,$CD,$78,$1A
	.dc.b	$21,$A2,$9F,$97,$F8,$4F,$D2,$91,$FD,$4B,$5F,$D1,$35,$E0,$68,$86
	.dc.b	$8A,$7E,$8F,$E1,$3F,$62,$47,$F6,$AD,$7F,$5C,$D7,$81,$A2,$1A,$29
	.dc.b	$FA,$FF,$84,$FD,$E9,$1F,$E0,$2D,$7F,$74,$D7,$81,$A2,$1A,$29,$FB
	.dc.b	$BF,$84,$FE,$12,$47,$F8,$6B,$5F,$E0,$CD,$78,$1A,$21,$A2,$9F,$C1
	.dc.b	$FE,$13,$F8,$A9,$1F,$E3,$2D,$7F,$89,$35,$E0,$68,$86,$8A,$7F,$13
	.dc.b	$F8,$4F,$E4,$24,$7F,$92,$B5,$FE,$3C,$D7,$81,$A2,$1A,$29,$FC,$7F
	.dc.b	$E1,$3F,$96,$91,$FE,$62,$D7,$F9,$53,$5E,$06,$88,$68,$A7,$F2,$BF
	.dc.b	$84,$FE,$72,$47,$F9,$EB,$5F,$E6,$CD,$78,$1A,$21,$A2,$9F,$CD,$FE
	.dc.b	$13,$FA,$29,$1F,$E9,$2D,$7F,$A1,$35,$E0,$68,$86,$8A,$7F,$43,$F8
	.dc.b	$4F,$EA,$24,$7F,$AA,$B5,$FE,$9C,$D7,$81,$A2,$1A,$29,$FD,$3F,$E1
	.dc.b	$3F,$AE,$91,$FE,$C2,$D7,$FA,$D3,$5E,$06,$88,$68,$A7,$F5,$BF,$84
	.dc.b	$FE,$D2,$47,$FB,$6B,$5F,$EC,$CD,$78,$1A,$21,$A2,$9F,$D9,$FE,$13
	.dc.b	$FB,$A9,$1F,$EF,$2D,$7F,$B9,$35,$E0,$68,$86,$8A,$08,$3E,$03,$FD
	.dc.b	$FE,$D3,$FE,$00,$B8,$23,$82,$1C,$8E,$6E,$9C,$1D,$DC,$38,$7A,$7C
	.dc.b	$1F,$62,$47,$03,$C8,$E1,$E0,$D7,$C1,$08,$C4,$78,$9A,$F8,$5F,$5F
	.dc.b	$C1,$C8,$ED,$28,$C4,$79,$5A,$F6,$6B,$EE,$6B,$ED,$6B,$F2,$FE,$0F
	.dc.b	$E4,$91,$EE,$D7,$E1,$35,$F5,$B5,$F9,$6D,$7E,$53,$F5,$26,$47,$A9
	.dc.b	$AF,$CD,$6B,$F1,$9A,$FD,$66,$BF,$55,$AF,$E0,$FE,$0F,$F0,$24,$7E
	.dc.b	$3B,$5F,$B4,$D7,$E9,$B5,$FC,$1B,$5F,$C1,$3F,$8C,$F8,$39,$1F,$49
	.dc.b	$AF,$E1,$5A,$FD,$E6,$BF,$8E,$6B,$F8,$D6,$BF,$9B,$F8,$3F,$CC,$91
	.dc.b	$FB,$ED,$7F,$24,$D7,$F1,$6D,$7F,$36,$D7,$F3,$4F,$E9,$3E,$0E,$47
	.dc.b	$E2,$9A,$FE,$75,$AF,$E5,$9A,$FE,$99,$AF,$E9,$5A,$FE,$CF,$E0,$FF
	.dc.b	$62,$47,$F2,$ED,$7F,$54,$D7,$F4,$6D,$7F,$66,$D7,$F6,$4F,$EF,$3E
	.dc.b	$0E,$47,$E8,$9A,$FE,$D5,$AF,$EB,$9A,$FE,$F9,$AF,$EF,$5A,$FF,$0B
	.dc.b	$F8,$3F,$E1,$24,$7F,$5E,$D7,$F8,$13,$5F,$DD,$B5,$FE,$16,$D7,$F8
	.dc.b	$53,$FC,$53,$E0,$E4,$7E,$E9,$AF,$F0,$D6,$BF,$C1,$9A,$FF,$16,$6B
	.dc.b	$FC,$55,$AF,$F2,$3F,$83,$FE,$42,$47,$F8,$3B,$5F,$E3,$4D,$7F,$89
	.dc.b	$B5,$FE,$46,$D7,$F9,$13,$FC,$B3,$E0,$E4,$7F,$12,$6B,$FC,$95,$AF
	.dc.b	$F1,$E6,$BF,$CB,$9A,$FF,$2D,$6B,$FC,$EF,$E0,$FF,$9C,$91,$FE,$3E
	.dc.b	$D7,$F9,$93,$5F,$E5,$6D,$7F,$9D,$B5,$FE,$74,$FF,$44,$F8,$39,$1F
	.dc.b	$CA,$9A,$FF,$3D,$6B,$FC,$D9,$AF,$F4,$66,$BF,$D1,$5A,$FF,$53,$F8
	.dc.b	$3F,$EA,$24,$7F,$9B,$B5,$FE,$94,$D7,$FA,$1B,$5F,$EA,$6D,$7F,$A9
	.dc.b	$3F,$D7,$3E,$0E,$47,$F4,$26,$BF,$D5,$5A,$FF,$4E,$6B,$FD,$79,$AF
	.dc.b	$F5,$D6,$BF,$DA,$FE,$0F,$FB,$49,$1F,$E9,$ED,$7F,$B1,$35,$FE,$B6
	.dc.b	$D7,$FB,$5B,$5F,$ED,$4F,$F7,$4F,$83,$91,$FD,$69,$AF,$F6,$D6,$BF
	.dc.b	$D9,$9A,$FF,$76,$6B,$FD,$D5,$AF,$F8,$0F,$E0,$FF,$C0,$24,$7F,$B3
	.dc.b	$B5,$FE,$F4,$D7,$FB,$9B,$5F,$F0,$1B,$5F,$F0,$13,$FD,$F5,$7D,$D3
	.dc.b	$E1,$30,$D8,$FF,$01,$35,$FF,$05,$B5,$FF,$08,$35,$FF,$0D,$B5,$FF
	.dc.b	$0F,$35,$EF,$1D,$88,$F6,$6B,$DB,$E0,$FC,$64,$7C,$6D,$74,$60,$86
	.dc.b	$08,$79,$9E,$0F,$A3,$E0,$FD,$E4,$7D,$ED,$7D,$ED,$78,$1A,$21,$A2
	.dc.b	$9F,$13,$E0,$FE,$59,$1F,$96,$D7,$E5,$B5,$E0,$68,$86,$8A,$7D,$0F
	.dc.b	$83,$FA,$E4,$7E,$BB,$5F,$AE,$D7,$81,$A2,$1A,$29,$F7,$3E,$0F,$F0
	.dc.b	$64,$7F,$06,$D7,$F0,$6D,$78,$1A,$21,$A2,$9F,$88,$F8,$3F,$C7,$91
	.dc.b	$FC,$7B,$5F,$C7,$B5,$E0,$68,$86,$8A,$7E,$53,$E0,$FF,$36,$47,$F3
	.dc.b	$6D,$7F,$36,$D7,$81,$A2,$1A,$29,$FA,$0F,$83,$FD,$39,$1F,$D3,$B5
	.dc.b	$FD,$3B,$5E,$06,$88,$68,$A7,$EB,$3E,$0F,$F6,$64,$7F,$66,$D7,$F6
	.dc.b	$6D,$78,$1A,$21,$A2,$9F,$B8,$F8,$3F,$DF,$91,$FD,$FB,$5F,$DF,$B5
	.dc.b	$E0,$68,$86,$8A,$7F,$04,$F8,$3F,$E1,$64,$7F,$85,$B5,$FE,$16,$D7
	.dc.b	$81,$A2,$1A,$29,$FC,$43,$E0,$FF,$8B,$91,$FE,$2E,$D7,$F8,$BB,$5E
	.dc.b	$06,$88,$68,$A7,$F1,$CF,$83,$FE,$46,$47,$F9,$1B,$5F,$E4,$6D,$78
	.dc.b	$1A,$21,$A2,$9F,$CA,$3E,$0F,$F9,$79,$1F,$E5,$ED,$7F,$97,$B5,$E0
	.dc.b	$68,$86,$8A,$7F,$34,$F8,$3F,$E7,$64,$7F,$9D,$B5,$FE,$76,$D7,$81
	.dc.b	$A2,$1A,$29,$FD,$03,$E0,$FF,$A3,$91,$FE,$8E,$D7,$FA,$3B,$5E,$06
	.dc.b	$88,$68,$A7,$F4,$CF,$83,$FE,$A6,$47,$FA,$9B,$5F,$EA,$6D,$78,$1A
	.dc.b	$21,$A2,$9F,$D6,$3E,$0F,$FA,$F9,$1F,$EB,$ED,$7F,$AF,$B5,$E0,$68
	.dc.b	$86,$8A,$7F,$64,$F8,$3F,$ED,$64,$7F,$B5,$B5,$FE,$D6,$D7,$81,$A2
	.dc.b	$1A,$29,$FD,$C3,$E0,$FF,$BB,$91,$FE,$EE,$D7,$FB,$BB,$5E,$06,$88
	.dc.b	$68,$A0,$83,$EE,$9F,$09,$FE,$09,$6C,$68,$DB,$87,$07,$18,$E0,$47
	.dc.b	$26,$4B,$1C,$5E,$8D,$7A,$C7,$17,$AC,$65,$F4,$2F,$84,$F3,$12,$3A
	.dc.b	$B5,$E2,$D7,$BB,$5F,$5B,$5E,$AF,$C1,$F8,$39,$1E,$86,$BE,$46,$BE
	.dc.b	$A6,$BF,$19,$AF,$A1,$AF,$CF,$F8,$3F,$9C,$47,$E2,$B5,$F8,$2D,$7E
	.dc.b	$3B,$5F,$A6,$D7,$E2,$BF,$67,$E0,$E4,$7D,$06,$BF,$31,$AF,$D2,$6B
	.dc.b	$F7,$9A,$FD,$06,$BF,$87,$F8,$3F,$C3,$11,$FB,$AD,$7E,$CB,$5F,$BE
	.dc.b	$D7,$F1,$6D,$7E,$EB,$F9,$1F,$83,$91,$F8,$86,$BF,$84,$6B,$F8,$A6
	.dc.b	$BF,$96,$6B,$F8,$86,$BF,$9F,$F8,$3F,$CF,$11,$FC,$AB,$5F,$C8,$B5
	.dc.b	$FC,$BB,$5F,$D1,$B5,$FC,$AB,$FA,$9F,$83,$91,$FA,$06,$BF,$9C,$6B
	.dc.b	$FA,$26,$BF,$AE,$6B,$FA,$06,$BF,$B7,$F8,$3F,$DB,$11,$FD,$6B,$5F
	.dc.b	$D4,$B5,$FD,$7B,$5F,$DD,$B5,$FD,$6B,$FC,$07,$E0,$E4,$7E,$E1,$AF
	.dc.b	$ED,$1A,$FE,$E9,$AF,$F0,$66,$BF,$B8,$6B,$FC,$3F,$E0,$FF,$87,$11
	.dc.b	$FE,$0A,$D7,$F8,$0B,$5F,$E0,$ED,$7F,$89,$B5,$FE,$0A,$FF,$19,$F8
	.dc.b	$39,$1F,$C4,$1A,$FF,$0C,$6B,$FC,$49,$AF,$F1,$E6,$BF,$C4,$1A,$FF
	.dc.b	$27,$F8,$3F,$E4,$C4,$7F,$8E,$B5,$FE,$32,$D7,$F8,$FB,$5F,$E5,$6D
	.dc.b	$7F,$8E,$BF,$CC,$7E,$0E,$47,$F2,$86,$BF,$C9,$1A,$FF,$2A,$6B,$FC
	.dc.b	$D9,$AF,$F2,$86,$BF,$CF,$FE,$0F,$F9,$F1,$1F,$E6,$AD,$7F,$98,$B5
	.dc.b	$FE,$6E,$D7,$FA,$1B,$5F,$E6,$AF,$F4,$9F,$83,$91,$FD,$01,$AF,$F3
	.dc.b	$C6,$BF,$D0,$9A,$FF,$4E,$6B,$FD,$01,$AF,$F5,$7F,$83,$FE,$AC,$47
	.dc.b	$FA,$6B,$5F,$E9,$2D,$7F,$A7,$B5,$FE,$B6,$D7,$FA,$6B,$FD,$87,$E0
	.dc.b	$E4,$7F,$58,$6B,$FD,$51,$AF,$F5,$A6,$BF,$D9,$9A,$FF,$58,$6B,$FD
	.dc.b	$BF,$E0,$FF,$B7,$11,$FE,$CA,$D7,$FB,$0B,$5F,$EC,$ED,$7F,$B9,$B5
	.dc.b	$FE,$CA,$FF,$79,$F8,$39,$1F,$DC,$1A,$FF,$6C,$6B,$FD,$C9,$AF,$F7
	.dc.b	$E6,$BF,$DC,$1A,$BE,$EB,$FB,$EB,$FE,$0B,$6B,$93,$2E,$08,$E0,$87
	.dc.b	$F8,$19,$CB,$4C,$9F,$E0,$E5,$F0,$9D,$09,$1C,$5A,$ED,$AF,$F8,$66
	.dc.b	$38,$BB,$21,$B2,$9B,$F8,$4F,$49,$1F,$33,$5F,$23,$5E,$06,$88,$68
	.dc.b	$A7,$8F,$E1,$3E,$29,$1F,$84,$D7,$E0,$35,$E0,$68,$86,$8A,$7B,$FE
	.dc.b	$13,$E8,$91,$F9,$CD,$7E,$63,$5E,$06,$88,$68,$A7,$CB,$F8,$4F,$BA
	.dc.b	$47,$ED,$35,$FB,$0D,$78,$1A,$21,$A2,$9F,$5F,$E1,$3F,$12,$47,$F0
	.dc.b	$CD,$7F,$08,$D7,$81,$A2,$1A,$29,$F8,$3F,$84,$FC,$A9,$1F,$C9,$35
	.dc.b	$FC,$83,$5E,$06,$88,$68,$A7,$E3,$FE,$13,$F4,$24,$7F,$3C,$D7,$F3
	.dc.b	$8D,$78,$1A,$21,$A2,$9F,$9B,$F8,$4F,$D6,$91,$FD,$53,$5F,$D4,$35
	.dc.b	$E0,$68,$86,$8A,$7E,$9F,$E1,$3F,$72,$47,$F6,$CD,$7F,$68,$D7,$81
	.dc.b	$A2,$1A,$29,$FB,$3F,$84,$FE,$0A,$47,$F8,$13,$5F,$E0,$0D,$78,$1A
	.dc.b	$21,$A2,$9F,$BF,$F8,$4F,$E2,$24,$7F,$87,$35,$FE,$18,$D7,$81,$A2
	.dc.b	$1A,$29,$FC,$2F,$E1,$3F,$8E,$91,$FE,$34,$D7,$F8,$C3,$5E,$06,$88
	.dc.b	$68,$A7,$F1,$7F,$84,$FE,$52,$47,$F9,$33,$5F,$E4,$8D,$78,$1A,$21
	.dc.b	$A2,$9F,$C8,$FE,$13,$F9,$A9,$1F,$E6,$4D,$7F,$98,$35,$E0,$68,$86
	.dc.b	$8A,$7F,$2F,$F8,$4F,$E8,$24,$7F,$9F,$35,$FE,$78,$D7,$81,$A2,$1A
	.dc.b	$29,$FC,$EF,$E1,$3F,$A6,$91,$FE,$94,$D7,$FA,$43,$5E,$06,$88,$68
	.dc.b	$A7,$F4,$7F,$84,$FE,$B2,$47,$FA,$B3,$5F,$EA,$8D,$78,$1A,$21,$A2
	.dc.b	$9F,$D4,$FE,$13,$FB,$29,$1F,$EC,$4D,$7F,$B0,$35,$E0,$68,$86,$8A
	.dc.b	$7F,$5F,$F8,$4F,$EE,$24,$7F,$B7,$35,$FE,$D8,$D7,$81,$A2,$1A,$29
	.dc.b	$FD,$AF,$E1,$3F,$BE,$91,$FE,$F4,$D7,$FB,$C3,$5E,$06,$88,$68,$A0
	.dc.b	$83,$C9,$7C,$1F,$F8,$41,$B1,$A3,$6E,$1C,$1C,$63,$81,$1C,$99,$2C
	.dc.b	$71,$7A,$35,$EB,$1C,$5E,$B1,$97,$D0,$BE,$0F,$CC,$48,$EA,$D7,$8B
	.dc.b	$5E,$ED,$7D,$6D,$7A,$BF,$07,$E0,$E4,$7A,$1A,$F9,$1A,$FA,$9A,$FC
	.dc.b	$66,$BE,$86,$BF,$3F,$E0,$FE,$71,$1F,$8A,$D7,$E0,$B5,$F8,$ED,$7E
	.dc.b	$9B,$5F,$8A,$FD,$9F,$83,$91,$F4,$1A,$FC,$C6,$BF,$49,$AF,$DE,$6B
	.dc.b	$F4,$1A,$FE,$1F,$E0,$FF,$0C,$47,$EE,$B5,$FB,$2D,$7E,$FB,$5F,$C5
	.dc.b	$B5,$FB,$AF,$E4,$7E,$0E,$47,$E2,$1A,$FE,$11,$AF,$E2,$9A,$FE,$59
	.dc.b	$AF,$E2,$1A,$FE,$7F,$E0,$FF,$3C,$47,$F2,$AD,$7F,$22,$D7,$F2,$ED
	.dc.b	$7F,$46,$D7,$F2,$AF,$EA,$7E,$0E,$47,$E8,$1A,$FE,$71,$AF,$E8,$9A
	.dc.b	$FE,$B9,$AF,$E8,$1A,$FE,$DF,$E0,$FF,$6C,$47,$F5,$AD,$7F,$52,$D7
	.dc.b	$F5,$ED,$7F,$76,$D7,$F5,$AF,$F0,$1F,$83,$91,$FB,$86,$BF,$B4,$6B
	.dc.b	$FB,$A6,$BF,$C1,$9A,$FE,$E1,$AF,$F0,$FF,$83,$FE,$1C,$47,$F8,$2B
	.dc.b	$5F,$E0,$2D,$7F,$83,$B5,$FE,$26,$D7,$F8,$2B,$FC,$67,$E0,$E4,$7F
	.dc.b	$10,$6B,$FC,$31,$AF,$F1,$26,$BF,$C7,$9A,$FF,$10,$6B,$FC,$9F,$E0
	.dc.b	$FF,$93,$11,$FE,$3A,$D7,$F8,$CB,$5F,$E3,$ED,$7F,$95,$B5,$FE,$3A
	.dc.b	$FF,$31,$F8,$39,$1F,$CA,$1A,$FF,$24,$6B,$FC,$A9,$AF,$F3,$66,$BF
	.dc.b	$CA,$1A,$FF,$3F,$F8,$3F,$E7,$C4,$7F,$9A,$B5,$FE,$62,$D7,$F9,$BB
	.dc.b	$5F,$E8,$6D,$7F,$9A,$BF,$D2,$7E,$0E,$47,$F4,$06,$BF,$CF,$1A,$FF
	.dc.b	$42,$6B,$FD,$39,$AF,$F4,$06,$BF,$D5,$FE,$0F,$FA,$B1,$1F,$E9,$AD
	.dc.b	$7F,$A4,$B5,$FE,$9E,$D7,$FA,$DB,$5F,$E9,$AF,$F6,$1F,$83,$91,$FD
	.dc.b	$61,$AF,$F5,$46,$BF,$D6,$9A,$FF,$66,$6B,$FD,$61,$AF,$F6,$FF,$83
	.dc.b	$FE,$DC,$47,$FB,$2B,$5F,$EC,$2D,$7F,$B3,$B5,$FE,$E6,$D7,$FB,$2B
	.dc.b	$FD,$E7,$E0,$E4,$7F,$70,$6B,$FD,$B1,$AF,$F7,$26,$BF,$DF,$9A,$FF
	.dc.b	$70,$6A,$FB,$A7,$C2,$7F,$7E,$8C,$36,$33,$1F,$E0,$26,$BF,$E0,$86
	.dc.b	$BF,$E0,$D6,$BF,$E1,$46,$BF,$E1,$76,$BD,$8E,$91,$DC,$8F,$7F,$83
	.dc.b	$F1,$11,$C1,$0F,$13,$96,$99,$3C,$8E,$5A,$7A,$FE,$0F,$DC,$47,$04
	.dc.b	$3D,$CE,$5A,$64,$F8,$0E,$5A,$7C,$7F,$83,$F9,$44,$70,$43,$E5,$39
	.dc.b	$69,$93,$E6,$39,$69,$F4,$FB,$3E,$B1,$1C,$10,$FA,$CE,$5A,$64,$FB
	.dc.b	$0E,$5A,$7D,$FF,$83,$FC,$11,$1C,$10,$FC,$13,$96,$99,$3F,$08,$E5
	.dc.b	$A7,$E2,$FE,$0F,$F1,$C4,$70,$43,$F1,$CE,$5A,$64,$FC,$83,$96,$9F
	.dc.b	$97,$F8,$3F,$CD,$11,$C1,$0F,$CD,$39,$69,$93,$F3,$8E,$5A,$7E,$8F
	.dc.b	$E0,$FF,$4C,$47,$04,$3F,$4C,$E5,$A6,$4F,$D4,$39,$69,$FA,$FF,$83
	.dc.b	$FD,$91,$1C,$10,$FD,$93,$96,$99,$3F,$68,$E5,$A7,$EE,$FE,$0F,$F7
	.dc.b	$C4,$70,$43,$F7,$CE,$5A,$64,$FE,$00,$E5,$A7,$F0,$7F,$83,$FE,$14
	.dc.b	$47,$04,$3F,$85,$39,$69,$93,$F8,$63,$96,$9F,$C4,$FE,$0F,$F8,$B1
	.dc.b	$1C,$10,$FE,$2C,$E5,$A6,$4F,$E3,$0E,$5A,$7F,$1F,$F8,$3F,$E4,$44
	.dc.b	$70,$43,$F9,$13,$96,$99,$3F,$92,$39,$69,$FC,$AF,$E0,$FF,$97,$11
	.dc.b	$C1,$0F,$E5,$CE,$5A,$64,$FE,$60,$E5,$A7,$F3,$7F,$83,$FE,$74,$47
	.dc.b	$04,$3F,$9D,$39,$69,$93,$F9,$E3,$96,$9F,$D0,$FE,$0F,$FA,$31,$1C
	.dc.b	$10,$FE,$8C,$E5,$A6,$4F,$E9,$0E,$5A,$7F,$4F,$F8,$3F,$EA,$44,$70
	.dc.b	$43,$FA,$93,$96,$99,$3F,$AA,$39,$69,$FD,$6F,$E0,$FF,$AF,$11,$C1
	.dc.b	$0F,$EB,$CE,$5A,$64,$FE,$C0,$E5,$A7,$F6,$7F,$83,$FE,$D4,$47,$04
	.dc.b	$3F,$B5,$39,$69,$93,$FB,$63,$96,$9F,$DC,$FE,$0F,$FB,$B1,$1C,$10
	.dc.b	$FE,$EC,$E5,$A6,$4F,$EF,$0E,$5A,$7F,$C0,$0F,$BA,$FF,$02,$26,$91
	.dc.b	$FF,$04,$35,$E0,$60,$86,$08,$08,$57,$AF,$C2,$74,$24,$78,$DA,$F0
	.dc.b	$38,$38,$71,$08,$C4,$76,$8C,$3C,$C8,$73,$3D,$DF,$07,$EC,$23,$F0
	.dc.b	$DA,$EE,$11,$88,$F0,$B5,$F0,$35,$F7,$3F,$23,$E0,$E4,$7C,$F6,$BD
	.dc.b	$5A,$FB,$1A,$FA,$DA,$FC,$96,$BF,$5F,$E0,$FE,$A9,$1F,$B8,$D7,$D4
	.dc.b	$D7,$E4,$B5,$F9,$0D,$7E,$B3,$F8,$0F,$83,$91,$F8,$76,$BF,$15,$AF
	.dc.b	$D4,$6B,$F4,$DA,$FE,$05,$AF,$E3,$FE,$0F,$F1,$A4,$7F,$28,$D7,$E9
	.dc.b	$35,$FC,$0B,$5F,$C0,$35,$FC,$73,$F9,$8F,$83,$91,$F9,$F6,$BF,$75
	.dc.b	$AF,$E3,$1A,$FE,$2D,$AF,$E6,$5A,$FE,$9F,$E0,$FF,$4A,$47,$F5,$8D
	.dc.b	$7F,$14,$D7,$F3,$2D,$7F,$30,$D7,$F4,$CF,$EC,$3E,$0E,$47,$ED,$DA
	.dc.b	$FE,$55,$AF,$E9,$1A,$FE,$8D,$AF,$EC,$5A,$FE,$FF,$E0,$FF,$7A,$47
	.dc.b	$F8,$23,$5F,$D1,$35,$FD,$8B,$5F,$D8,$35,$FD,$F3,$FC,$23,$E0,$E4
	.dc.b	$7F,$0F,$6B,$FA,$D6,$BF,$BC,$6B,$FB,$B6,$BF,$C2,$5A,$FF,$17,$F8
	.dc.b	$3F,$E2,$A4,$7F,$8E,$35,$FD,$D3,$5F,$E1,$2D,$7F,$84,$35,$FE,$2C
	.dc.b	$FF,$20,$F8,$39,$1F,$C9,$DA,$FF,$05,$6B,$FC,$51,$AF,$F1,$36,$BF
	.dc.b	$C8,$5A,$FF,$2F,$F8,$3F,$E5,$A4,$7F,$9A,$35,$FE,$24,$D7,$F9,$0B
	.dc.b	$5F,$E4,$0D,$7F,$97,$3F,$CE,$3E,$0E,$47,$F3,$F6,$BF,$C7,$5A,$FF
	.dc.b	$2C,$6B,$FC,$AD,$AF,$F3,$96,$BF,$D1,$FE,$0F,$FA,$29,$1F,$E9,$8D
	.dc.b	$7F,$95,$35,$FE,$72,$D7,$F9,$C3,$5F,$E8,$CF,$F5,$0F,$83,$91,$FD
	.dc.b	$5D,$AF,$F3,$56,$BF,$D1,$1A,$FF,$43,$6B,$FD,$45,$AF,$F5,$FF,$83
	.dc.b	$FE,$BA,$47,$FB,$23,$5F,$E8,$4D,$7F,$A8,$B5,$FE,$A0,$D7,$FA,$F3
	.dc.b	$FD,$A3,$E0,$E4,$7F,$6F,$6B,$FD,$35,$AF,$F5,$C6,$BF,$D6,$DA,$FF
	.dc.b	$69,$6B,$FD,$DF,$E0,$FF,$BA,$91,$FE,$F8,$D7,$FA,$D3,$5F,$ED,$2D
	.dc.b	$7F,$B4,$35,$FE,$EC,$FF,$80,$3E,$0E,$47,$F8,$1D,$AF,$F6,$56,$BF
	.dc.b	$DD,$1A,$FF,$73,$6B,$FE,$01,$6A,$FB,$A7,$C2,$7F,$75,$8F,$F0,$44
	.dc.b	$61,$C3,$5F,$EF,$4D,$7F,$C0,$CD,$7F,$C1,$ED,$79,$C7,$57,$C0,$D8
	.dc.b	$D4,$60,$EF,$19,$35,$F0,$92,$3C,$6D,$7C,$6D,$7C,$6D,$7C,$6D,$7D
	.dc.b	$AD,$7E,$17,$C2,$78,$C8,$FC,$06,$BF,$01,$AF,$C0,$6B,$F0,$1A,$FC
	.dc.b	$A7,$DD,$F0,$92,$3E,$5B,$5F,$96,$D7,$E5,$B5,$F9,$6D,$7E,$AB,$5F
	.dc.b	$B5,$F0,$9F,$2C,$8F,$D8,$6B,$F6,$1A,$FD,$86,$BF,$61,$AF,$E0,$9F
	.dc.b	$AD,$F0,$92,$3F,$06,$D7,$F0,$6D,$7F,$06,$D7,$F0,$6D,$7F,$1A,$D7
	.dc.b	$F2,$5F,$09,$F8,$32,$3F,$90,$6B,$F9,$06,$BF,$90,$6B,$F9,$06,$BF
	.dc.b	$9A,$7F,$1D,$F0,$92,$3F,$36,$D7,$F3,$6D,$7F,$36,$D7,$F3,$6D,$7F
	.dc.b	$4A,$D7,$F5,$5F,$09,$F9,$B2,$3F,$A8,$6B,$FA,$86,$BF,$A8,$6B,$FA
	.dc.b	$86,$BF,$B2,$7F,$4D,$F0,$92,$3F,$66,$D7,$F6,$6D,$7F,$66,$D7,$F6
	.dc.b	$6D,$7F,$7A,$D7,$F8,$17,$C2,$7E,$CC,$8F,$F0,$06,$BF,$C0,$1A,$FF
	.dc.b	$00,$6B,$FC,$01,$AF,$F0,$A7,$F7,$DF,$09,$23,$F8,$5B,$5F,$E1,$6D
	.dc.b	$7F,$85,$B5,$FE,$16,$D7,$F8,$AB,$5F,$E3,$5F,$09,$FC,$2C,$8F,$F1
	.dc.b	$86,$BF,$C6,$1A,$FF,$18,$6B,$FC,$61,$AF,$F2,$27,$F8,$B7,$C2,$48
	.dc.b	$FE,$46,$D7,$F9,$1B,$5F,$E4,$6D,$7F,$91,$B5,$FE,$5A,$D7,$F9,$97
	.dc.b	$C2,$7F,$23,$23,$FC,$C1,$AF,$F3,$06,$BF,$CC,$1A,$FF,$30,$6B,$FC
	.dc.b	$E9,$FE,$5D,$F0,$92,$3F,$9D,$B5,$FE,$76,$D7,$F9,$DB,$5F,$E7,$6D
	.dc.b	$7F,$A2,$B5,$FE,$95,$F0,$9F,$CE,$C8,$FF,$48,$6B,$FD,$21,$AF,$F4
	.dc.b	$86,$BF,$D2,$1A,$FF,$52,$7F,$A3,$7C,$24,$8F,$EA,$6D,$7F,$A9,$B5
	.dc.b	$FE,$A6,$D7,$FA,$9B,$5F,$EB,$AD,$7F,$B1,$7C,$27,$F5,$32,$3F,$D8
	.dc.b	$1A,$FF,$60,$6B,$FD,$81,$AF,$F6,$06,$BF,$DA,$9F,$EB,$DF,$09,$23
	.dc.b	$FB,$5B,$5F,$ED,$6D,$7F,$B5,$B5,$FE,$D6,$D7,$FB,$AB,$5F,$EF,$5F
	.dc.b	$09,$FD,$AC,$8F,$F7,$86,$BF,$DE,$1A,$FF,$78,$6B,$FD,$E1,$AF,$F8
	.dc.b	$09,$FE,$F1,$C6,$F8,$3A,$49,$FF,$04,$46,$5B,$19,$7F,$E0,$04,$65
	.dc.b	$B1,$94,$41,$1B,$70,$62,$38,$3B,$87,$C1,$F9,$89,$1E,$46,$BE,$26
	.dc.b	$BD,$23,$8B,$B2,$1B,$29,$BF,$83,$F8,$64,$7E,$0B,$5F,$7B,$5E,$06
	.dc.b	$88,$68,$A7,$8F,$E0,$FE,$79,$1F,$9A,$D7,$E5,$B5,$E0,$68,$86,$8A
	.dc.b	$7B,$FE,$0F,$ED,$91,$FB,$2D,$7E,$BB,$5E,$06,$88,$68,$A7,$CB,$F8
	.dc.b	$3F,$C3,$91,$FC,$2B,$5F,$C1,$B5,$E0,$68,$86,$8A,$7D,$7F,$83,$FC
	.dc.b	$99,$1F,$C8,$B5,$FC,$7B,$5E,$06,$88,$68,$A7,$E0,$FE,$0F,$F3,$E4
	.dc.b	$7F,$3A,$D7,$F3,$6D,$78,$1A,$21,$A2,$9F,$8F,$F8,$3F,$D5,$91,$FD
	.dc.b	$4B,$5F,$D3,$B5,$E0,$68,$86,$8A,$7E,$6F,$E0,$FF,$6E,$47,$F6,$AD
	.dc.b	$7F,$66,$D7,$81,$A2,$1A,$29,$FA,$7F,$83,$FE,$06,$47,$F8,$0B,$5F
	.dc.b	$DF,$B5,$E0,$68,$86,$8A,$7E,$CF,$E0,$FF,$87,$91,$FE,$1A,$D7,$F8
	.dc.b	$5B,$5E,$06,$88,$68,$A7,$EF,$FE,$0F,$F8,$D9,$1F,$E3,$2D,$7F,$8B
	.dc.b	$B5,$E0,$68,$86,$8A,$7F,$0B,$F8,$3F,$E4,$E4,$7F,$92,$B5,$FE,$46
	.dc.b	$D7,$81,$A2,$1A,$29,$FC,$5F,$E0,$FF,$99,$91,$FE,$62,$D7,$F9,$7B
	.dc.b	$5E,$06,$88,$68,$A7,$F2,$3F,$83,$FE,$7E,$47,$F9,$EB,$5F,$E7,$6D
	.dc.b	$78,$1A,$21,$A2,$9F,$CB,$FE,$0F,$FA,$59,$1F,$E9,$2D,$7F,$A3,$B5
	.dc.b	$E0,$68,$86,$8A,$7F,$3B,$F8,$3F,$EA,$E4,$7F,$AA,$B5,$FE,$A6,$D7
	.dc.b	$81,$A2,$1A,$29,$FD,$1F,$E0,$FF,$B1,$91,$FE,$C2,$D7,$FA,$FB,$5E
	.dc.b	$06,$88,$68,$A7,$F5,$3F,$83,$FE,$DE,$47,$FB,$6B,$5F,$ED,$6D,$78
	.dc.b	$1A,$21,$A2,$9F,$D7,$FE,$0F,$FB,$D9,$1F,$EF,$2D,$7F,$BB,$B5,$E0
	.dc.b	$68,$86,$8A,$7F,$76,$FB,$AF,$F7,$F2,$69,$1F,$EF,$67,$F7,$F7,$0E
	.dc.b	$0E,$05,$29,$97,$6F,$26,$BE,$1F,$84,$E1,$06,$92,$39,$35,$E4,$D7
	.dc.b	$91,$A2,$1A,$3C,$5F,$07,$D8,$8E,$08,$79,$5C,$B4,$C9,$F1,$DC,$B4
	.dc.b	$F7,$7C,$1F,$A8,$8E,$08,$7C,$17,$2D,$32,$7D,$37,$2D,$3E,$57,$C1
	.dc.b	$FC,$62,$38,$21,$F3,$5C,$B4,$C9,$F7,$DC,$B4,$FA,$DF,$07,$F4,$88
	.dc.b	$E0,$87,$D9,$72,$D3,$27,$E2,$DC,$B4,$FC,$17,$C1,$FD,$E2,$38,$21
	.dc.b	$F8,$57,$2D,$32,$7E,$5D,$CB,$4F,$C7,$7C,$1F,$E2,$88,$E0,$87,$E4
	.dc.b	$5C,$B4,$C9,$FA,$37,$2D,$3F,$35,$F0,$7F,$96,$23,$82,$1F,$9D,$72
	.dc.b	$D3,$27,$EB,$DC,$B4,$FD,$37,$C1,$FE,$88,$8E,$08,$7E,$A5,$CB,$4C
	.dc.b	$9F,$BB,$72,$D3,$F6,$5F,$07,$FA,$E2,$38,$21,$FB,$57,$2D,$32,$7F
	.dc.b	$07,$72,$D3,$F7,$DF,$07,$FB,$A2,$38,$21,$FC,$05,$CB,$4C,$9F,$C4
	.dc.b	$DC,$B4,$FE,$15,$F0,$7F,$C1,$88,$E0,$87,$F0,$D7,$2D,$32,$7F,$1F
	.dc.b	$72,$D3,$F8,$B7,$C1,$FF,$12,$23,$82,$1F,$C6,$5C,$B4,$C9,$FC,$AD
	.dc.b	$CB,$4F,$E4,$5F,$07,$FC,$78,$8E,$08,$7F,$25,$72,$D3,$27,$F3,$77
	.dc.b	$2D,$3F,$97,$7C,$1F,$F2,$A2,$38,$21,$FC,$C5,$CB,$4C,$9F,$D0,$DC
	.dc.b	$B4,$FE,$75,$F0,$7F,$CD,$88,$E0,$87,$F3,$D7,$2D,$32,$7F,$4F,$72
	.dc.b	$D3,$FA,$37,$C1,$FF,$42,$23,$82,$1F,$D2,$5C,$B4,$C9,$FD,$6D,$CB
	.dc.b	$4F,$EA,$5F,$07,$FD,$38,$8E,$08,$7F,$55,$72,$D3,$27,$F6,$77,$2D
	.dc.b	$3F,$AF,$7C,$1F,$F5,$A2,$38,$21,$FD,$85,$CB,$4C,$9F,$DC,$DC,$B4
	.dc.b	$FE,$D5,$F0,$7F,$D9,$88,$E0,$87,$F6,$D7,$2D,$32,$7F,$7F,$72,$D3
	.dc.b	$FB,$B7,$C1,$FF,$72,$23,$82,$1F,$DE,$5C,$B4,$C9,$FE,$0B,$72,$D3
	.dc.b	$FC,$0D,$F7,$4F,$83,$FF,$03,$36,$33,$18,$23,$82,$1F,$E1,$28,$E0
	.dc.b	$D8,$DB,$4D,$82,$08,$7C,$1F,$88,$91,$E6,$6B,$DD,$AF,$23,$93,$87
	.dc.b	$AC,$74,$79,$10,$F4,$7C,$1F,$C0,$23,$F1,$1A,$E8,$C1,$0C,$10,$F1
	.dc.b	$42,$38,$C6,$3E,$0F,$E6,$91,$FA,$2D,$7E,$53,$5E,$06,$88,$68,$A7
	.dc.b	$83,$D1,$FB,$24,$7E,$EB,$5F,$AC,$D7,$81,$A2,$1A,$29,$EC,$F8,$3F
	.dc.b	$C2,$91,$FC,$4B,$5F,$C1,$35,$E0,$68,$86,$8A,$7C,$8F,$83,$FC,$89
	.dc.b	$1F,$CA,$B5,$FC,$73,$5E,$06,$88,$68,$A7,$D4,$F8,$3F,$CE,$91,$FD
	.dc.b	$0B,$5F,$CD,$35,$E0,$68,$86,$8A,$7E,$03,$E0,$FF,$52,$47,$F5,$AD
	.dc.b	$7F,$4C,$D7,$81,$A2,$1A,$29,$F8,$CF,$83,$FD,$A9,$1F,$DC,$B5,$FD
	.dc.b	$93,$5E,$06,$88,$68,$A7,$E6,$3E,$0F,$F8,$09,$1F,$E0,$AD,$7F,$7C
	.dc.b	$D7,$81,$A2,$1A,$29,$FA,$4F,$83,$FE,$1A,$47,$F8,$8B,$5F,$E1,$4D
	.dc.b	$78,$1A,$21,$A2,$9F,$B0,$F8,$3F,$E3,$24,$7F,$8E,$B5,$FE,$2C,$D7
	.dc.b	$81,$A2,$1A,$29,$FB,$CF,$83,$FE,$4A,$47,$F9,$4B,$5F,$E4,$4D,$78
	.dc.b	$1A,$21,$A2,$9F,$C2,$3E,$0F,$F9,$89,$1F,$E6,$AD,$7F,$97,$35,$E0
	.dc.b	$68,$86,$8A,$7F,$14,$F8,$3F,$E7,$A4,$7F,$A0,$B5,$FE,$74,$D7,$81
	.dc.b	$A2,$1A,$29,$FC,$83,$E0,$FF,$A4,$91,$FE,$9A,$D7,$FA,$33,$5E,$06
	.dc.b	$88,$68,$A7,$F2,$CF,$83,$FE,$AA,$47,$FA,$CB,$5F,$EA,$4D,$78,$1A
	.dc.b	$21,$A2,$9F,$CE,$3E,$0F,$FB,$09,$1F,$EC,$AD,$7F,$AF,$35,$E0,$68
	.dc.b	$86,$8A,$7F,$44,$F8,$3F,$ED,$A4,$7F,$B8,$B5,$FE,$D4,$D7,$81,$A2
	.dc.b	$1A,$29,$FD,$43,$E0,$FF,$BC,$91,$FE,$FA,$D7,$FB,$B3,$5E,$06,$88
	.dc.b	$68,$A7,$F6,$DF,$BA,$7C,$27,$F6,$F8,$FE,$F3,$18,$70,$1C,$3B,$6B
	.dc.b	$CE,$36,$ED,$B1,$A8,$EB,$18,$23,$DA,$34,$FF,$82,$DA,$F8,$7E,$13
	.dc.b	$99,$1E,$0D,$7C,$8D,$7D,$CD,$7D,$AD,$7B,$3E,$6F,$84,$91,$E3,$6B
	.dc.b	$EF,$6B,$F2,$5A,$FC,$86,$BE,$96,$BF,$27,$E1,$3E,$19,$1F,$80,$D7
	.dc.b	$E6,$35,$FA,$CD,$7E,$AB,$5F,$8C,$FC,$EF,$84,$91,$F2,$DA,$FD,$76
	.dc.b	$BF,$81,$6B,$F8,$06,$BF,$45,$AF,$E0,$7E,$13,$ED,$91,$FB,$0D,$7F
	.dc.b	$08,$D7,$F1,$CD,$7F,$1A,$D7,$EF,$3F,$86,$F8,$49,$1F,$83,$6B,$F8
	.dc.b	$F6,$BF,$99,$6B,$F9,$86,$BF,$89,$6B,$F9,$9F,$84,$FC,$99,$1F,$C8
	.dc.b	$35,$FC,$E3,$5F,$D3,$35,$FD,$2B,$5F,$CB,$3F,$9E,$F8,$49,$1F,$9B
	.dc.b	$6B,$FA,$76,$BF,$B1,$6B,$FB,$06,$BF,$A1,$6B,$FB,$1F,$84,$FD,$59
	.dc.b	$1F,$D4,$35,$FD,$A3,$5F,$DF,$35,$FD,$EB,$5F,$D7,$3F,$B6,$F8,$49
	.dc.b	$1F,$B3,$6B,$FB,$F6,$BF,$C2,$5A,$FF,$08,$6B,$FB,$96,$BF,$C2,$7E
	.dc.b	$13,$F8,$19,$1F,$E0,$0D,$7F,$86,$35,$FE,$2C,$D7,$F8,$AB,$5F,$E0
	.dc.b	$CF,$F0,$EF,$84,$91,$FC,$2D,$AF,$F1,$76,$BF,$C8,$5A,$FF,$20,$6B
	.dc.b	$FC,$45,$AF,$F2,$1F,$84,$FE,$36,$47,$F8,$C3,$5F,$E4,$8D,$7F,$97
	.dc.b	$35,$FE,$5A,$D7,$F8,$F3,$FC,$9B,$E1,$24,$7F,$23,$6B,$FC,$BD,$AF
	.dc.b	$F3,$96,$BF,$CE,$1A,$FF,$29,$6B,$FC,$E7,$E1,$3F,$99,$91,$FE,$60
	.dc.b	$D7,$F9,$E3,$5F,$E8,$CD,$7F,$A2,$B5,$FE,$6C,$FF,$3E,$F8,$49,$1F
	.dc.b	$CE,$DA,$FF,$47,$6B,$FD,$45,$AF,$F5,$06,$BF,$D0,$5A,$FF,$51,$F8
	.dc.b	$4F,$E9,$64,$7F,$A4,$35,$FE,$A8,$D7,$FA,$F3,$5F,$EB,$AD,$7F,$A7
	.dc.b	$3F,$D5,$BE,$12,$47,$F5,$36,$BF,$D7,$DA,$FF,$69,$6B,$FD,$A1,$AF
	.dc.b	$F5,$96,$BF,$DA,$7E,$13,$FB,$19,$1F,$EC,$0D,$7F,$B6,$35,$FE,$EC
	.dc.b	$D7,$FB,$AB,$5F,$EC,$CF,$F6,$EF,$84,$91,$FD,$AD,$AF,$F7,$76,$BF
	.dc.b	$E0,$16,$BF,$E0,$06,$BF,$DC,$5A,$BD,$FF,$EF,$AD,$3F,$BF,$B8,$23
	.dc.b	$82,$1F,$E0,$A7,$2D,$32,$7F,$84,$5E,$5F,$07,$D0,$91,$C5,$AE,$DA
	.dc.b	$88,$79,$DC,$B8,$84,$62,$38,$46,$1F,$37,$C1,$FB,$08,$E0,$87,$9D
	.dc.b	$CB,$4C,$9E,$23,$E4,$3A,$71,$08,$F9,$3F,$07,$27,$E3,$35,$F8,$8D
	.dc.b	$78,$B5,$F6,$35,$F5,$B5,$FB,$1F,$07,$F5,$88,$FD,$36,$BF,$45,$AF
	.dc.b	$99,$AF,$C9,$6B,$F2,$1F,$C0,$D4,$8F,$BC,$D7,$EE,$35,$F8,$2D,$7E
	.dc.b	$A3,$5F,$A6,$D7,$F2,$1F,$07,$F8,$E2,$3F,$8B,$6B,$F8,$96,$BF,$39
	.dc.b	$AF,$E0,$5A,$FE,$01,$FC,$CF,$C1,$C8,$FC,$B3,$5F,$CA,$35,$FB,$2D
	.dc.b	$7F,$18,$D7,$F1,$6D,$7F,$51,$F0,$7F,$A6,$23,$FA,$36,$BF,$A1,$6B
	.dc.b	$F8,$66,$BF,$99,$6B,$F9,$87,$F6,$3F,$07,$23,$F5,$CD,$7F,$58,$D7
	.dc.b	$F2,$2D,$7F,$48,$D7,$F4,$6D,$7F,$80,$7C,$1F,$EF,$88,$FE,$ED,$AF
	.dc.b	$EE,$5A,$FE,$79,$AF,$EC,$5A,$FE,$C1,$FE,$13,$F0,$72,$3F,$83,$35
	.dc.b	$FE,$08,$D7,$F5,$2D,$7F,$78,$D7,$F7,$6D,$7F,$8C,$7C,$1F,$F1,$62
	.dc.b	$3F,$C4,$DA,$FF,$11,$6B,$FB,$66,$BF,$C2,$5A,$FF,$08,$7F,$90,$FC
	.dc.b	$1C,$8F,$E3,$CD,$7F,$8E,$35,$FE,$02,$D7,$F8,$A3,$5F,$E2,$6D,$7F
	.dc.b	$98,$7C,$1F,$F2,$E2,$3F,$CA,$DA,$FF,$29,$6B,$FC,$39,$AF,$F2,$16
	.dc.b	$BF,$C8,$1F,$E7,$3F,$07,$23,$F9,$B3,$5F,$E6,$8D,$7F,$8C,$B5,$FE
	.dc.b	$58,$D7,$F9,$5B,$5F,$E9,$1F,$07,$FD,$18,$8F,$F4,$36,$BF,$D0,$5A
	.dc.b	$FF,$26,$6B,$FC,$E5,$AF,$F3,$87,$FA,$8F,$C1,$C8,$FE,$9C,$D7,$FA
	.dc.b	$63,$5F,$E6,$2D,$7F,$A2,$35,$FE,$86,$D7,$FB,$07,$C1,$FF,$5E,$23
	.dc.b	$FD,$6D,$AF,$F5,$96,$BF,$CF,$9A,$FF,$51,$6B,$FD,$41,$FE,$D3,$F0
	.dc.b	$72,$3F,$B3,$35,$FE,$C8,$D7,$FA,$4B,$5F,$EB,$8D,$7F,$AD,$B5,$FE
	.dc.b	$F1,$F0,$7F,$DD,$88,$FF,$73,$6B,$FD,$C5,$AF,$F5,$66,$BF,$DA,$5A
	.dc.b	$FF,$68,$7F,$C0,$3F,$07,$23,$FB,$F3,$5F,$EF,$8D,$7F,$B0,$B5,$FE
	.dc.b	$E8,$D7,$FB,$9B,$57,$DD,$3E,$13,$FB,$64,$7F,$7E,$8C,$38,$37,$15
	.dc.b	$A6,$63,$6F,$26,$BF,$DD,$DA,$F5,$8E,$51,$88,$FF,$07,$35,$EA,$73
	.dc.b	$8D,$FC,$24,$8F,$03,$5F,$03,$5F,$03,$5F,$63,$5E,$ED,$7E,$1F,$C1
	.dc.b	$F9,$08,$FB,$5A,$FB,$5A,$FB,$5A,$FC,$96,$BE,$C7,$DF,$F0,$72,$3E
	.dc.b	$43,$5F,$90,$D7,$E4,$35,$FA,$8D,$7E,$3B,$5F,$B7,$F0,$7F,$30,$8F
	.dc.b	$D5,$6B,$F5,$5A,$FD,$56,$BF,$81,$6B,$F5,$1F,$AF,$F0,$72,$3F,$00
	.dc.b	$D7,$F0,$0D,$7F,$00,$D7,$F1,$8D,$7E,$FB,$5F,$C9,$FC,$1F,$E1,$08
	.dc.b	$FE,$35,$AF,$E3,$5A,$FE,$35,$AF,$E6,$5A,$FE,$31,$FC,$7F,$C1,$C8
	.dc.b	$FC,$C3,$5F,$CC,$35,$FC,$C3,$5F,$D2,$35,$FC,$BB,$5F,$D5,$FC,$1F
	.dc.b	$E7,$08,$FE,$95,$AF,$E9,$5A,$FE,$95,$AF,$EC,$5A,$FE,$91,$FD,$3F
	.dc.b	$C1,$C8,$FD,$83,$5F,$D8,$35,$FD,$83,$5F,$DE,$35,$FD,$7B,$5F,$E0
	.dc.b	$7F,$07,$FB,$42,$3F,$BD,$6B,$FB,$D6,$BF,$BD,$6B,$FC,$25,$AF,$EF
	.dc.b	$1F,$DF,$FC,$1C,$8F,$E1,$0D,$7F,$84,$35,$FE,$10,$D7,$F8,$A3,$5F
	.dc.b	$E0,$ED,$7F,$8D,$FC,$1F,$F0,$C2,$3F,$C5,$5A,$FF,$15,$6B,$FC,$55
	.dc.b	$AF,$F2,$16,$BF,$C5,$1F,$E2,$FF,$07,$23,$F9,$03,$5F,$E4,$0D,$7F
	.dc.b	$90,$35,$FE,$58,$D7,$F8,$FB,$5F,$E6,$7F,$07,$FC,$90,$8F,$F2,$D6
	.dc.b	$BF,$CB,$5A,$FF,$2D,$6B,$FC,$E5,$AF,$F2,$C7,$F9,$7F,$C1,$C8,$FE
	.dc.b	$70,$D7,$F9,$C3,$5F,$E7,$0D,$7F,$A2,$35,$FE,$6E,$D7,$FA,$5F,$C1
	.dc.b	$FF,$3C,$23,$FD,$15,$AF,$F4,$56,$BF,$D1,$5A,$FF,$51,$6B,$FD,$11
	.dc.b	$FE,$8F,$F0,$72,$3F,$A8,$35,$FE,$A0,$D7,$FA,$83,$5F,$EB,$8D,$7F
	.dc.b	$A7,$B5,$FE,$C7,$F0,$7F,$D5,$08,$FF,$5D,$6B,$FD,$75,$AF,$F5,$D6
	.dc.b	$BF,$DA,$5A,$FF,$5C,$7F,$AF,$FC,$1C,$8F,$ED,$0D,$7F,$B4,$35,$FE
	.dc.b	$D0,$D7,$FB,$A3,$5F,$EC,$ED,$7F,$BD,$FC,$1F,$F6,$C2,$3F,$DD,$5A
	.dc.b	$FF,$75,$6B,$FD,$D5,$AF,$F8,$05,$AF,$F7,$47,$FC,$01,$F7,$4F,$84
	.dc.b	$A4,$9F,$F0,$34,$65,$B1,$97,$FC,$1B,$19,$6C,$65,$FF,$7E,$46,$5B
	.dc.b	$19,$44,$87,$C1,$F9,$09,$1E,$26,$BE,$06,$BD,$9A,$F4,$11,$D1,$E4
	.dc.b	$43,$D3,$F0,$7F,$08,$8F,$C0,$6B,$A3,$04,$30,$43,$DE,$F0,$7E,$0F
	.dc.b	$C1,$FC,$E2,$3F,$31,$AF,$CA,$6B,$C0,$D1,$0D,$14,$F9,$BF,$07,$F6
	.dc.b	$88,$FD,$86,$BF,$59,$AF,$03,$44,$34,$53,$EC,$FC,$1F,$E1,$88,$FE
	.dc.b	$11,$AF,$E0,$9A,$F0,$34,$43,$45,$3F,$0B,$F0,$7F,$92,$23,$F9,$06
	.dc.b	$BF,$8E,$6B,$C0,$D1,$0D,$14,$FC,$8F,$C1,$FE,$78,$8F,$E7,$1A,$FE
	.dc.b	$69,$AF,$03,$44,$34,$53,$F3,$BF,$07,$FA,$A2,$3F,$A8,$6B,$FA,$66
	.dc.b	$BC,$0D,$10,$D1,$4F,$D4,$FC,$1F,$ED,$88,$FE,$D1,$AF,$EC,$9A,$F0
	.dc.b	$34,$43,$45,$3F,$6B,$F0,$7F,$C0,$88,$FF,$00,$6B,$FB,$E6,$BC,$0D
	.dc.b	$10,$D1,$4F,$E0,$3F,$07,$FC,$38,$8F,$F0,$C6,$BF,$C2,$9A,$F0,$34
	.dc.b	$43,$45,$3F,$86,$FC,$1F,$F1,$A2,$3F,$C6,$1A,$FF,$16,$6B,$C0,$D1
	.dc.b	$0D,$14,$FE,$33,$F0,$7F,$C9,$88,$FF,$24,$6B,$FC,$89,$AF,$03,$44
	.dc.b	$34,$53,$F9,$2F,$C1,$FF,$32,$23,$FC,$C1,$AF,$F2,$E6,$BC,$0D,$10
	.dc.b	$D1,$4F,$E6,$3F,$07,$FC,$F8,$8F,$F3,$C6,$BF,$CE,$9A,$F0,$34,$43
	.dc.b	$45,$3F,$9E,$FC,$1F,$F4,$A2,$3F,$D2,$1A,$FF,$46,$6B,$C0,$D1,$0D
	.dc.b	$14,$FE,$93,$F0,$7F,$D5,$88,$FF,$54,$6B,$FD,$49,$AF,$03,$44,$34
	.dc.b	$53,$FA,$AF,$C1,$FF,$62,$23,$FD,$81,$AF,$F5,$E6,$BC,$0D,$10,$D1
	.dc.b	$4F,$EC,$3F,$07,$FD,$B8,$8F,$F6,$C6,$BF,$DA,$9A,$F0,$34,$43,$45
	.dc.b	$3F,$B6,$FC,$1F,$F7,$A2,$3F,$DE,$1A,$FF,$76,$6B,$C0,$D1,$0D,$14
	.dc.b	$FE,$D4,$3E,$EA,$B8,$6B,$FD,$BD,$AF,$F7,$C6,$BC,$4C,$B8,$23,$82
	.dc.b	$1D,$1C,$B7,$E1,$38,$91,$C1,$0F,$2C,$23,$51,$86,$BE,$96,$BE,$97
	.dc.b	$CB,$F0,$92,$3C,$4D,$7A,$B5,$F8,$4D,$7E,$13,$5F,$84,$D7,$E5,$7C
	.dc.b	$27,$C2,$23,$EF,$6B,$EA,$6B,$F3,$DA,$FC,$F6,$BF,$3D,$F9,$BF,$09
	.dc.b	$23,$E5,$35,$F8,$AD,$7E,$D3,$5F,$B4,$D7,$ED,$35,$FC,$17,$C2,$7D
	.dc.b	$A2,$3F,$5D,$AF,$D2,$6B,$F8,$76,$BF,$87,$6B,$F8,$77,$F0,$BF,$09
	.dc.b	$23,$F0,$4D,$7E,$EB,$5F,$C9,$35,$FC,$93,$5F,$C9,$35,$FC,$D7,$C2
	.dc.b	$7E,$48,$8F,$E3,$DA,$FE,$29,$AF,$E7,$DA,$FE,$7D,$AF,$E7,$DF,$CE
	.dc.b	$FC,$24,$8F,$CD,$35,$FC,$AB,$5F,$D5,$35,$FD,$53,$5F,$D5,$35,$FD
	.dc.b	$97,$C2,$7E,$A8,$8F,$E9,$DA,$FE,$89,$AF,$ED,$DA,$FE,$DD,$AF,$ED
	.dc.b	$DF,$DA,$FC,$24,$8F,$D9,$35,$FD,$6B,$5F,$E0,$4D,$7F,$81,$35,$FE
	.dc.b	$04,$D7,$F8,$57,$C2,$7F,$02,$23,$FB,$F6,$BF,$BA,$6B,$FC,$3D,$AF
	.dc.b	$F0,$F6,$BF,$C3,$DF,$E1,$BF,$09,$23,$F8,$53,$5F,$E0,$AD,$7F,$8D
	.dc.b	$35,$FE,$34,$D7,$F8,$D3,$5F,$E4,$5F,$09,$FC,$68,$8F,$F1,$76,$BF
	.dc.b	$C4,$9A,$FF,$27,$6B,$FC,$9D,$AF,$F2,$77,$F9,$2F,$C2,$48,$FE,$44
	.dc.b	$D7,$F8,$EB,$5F,$E6,$4D,$7F,$99,$35,$FE,$64,$D7,$F9,$D7,$C2,$7F
	.dc.b	$32,$23,$FC,$BD,$AF,$F2,$A6,$BF,$CF,$DA,$FF,$3F,$6B,$FC,$FD,$FE
	.dc.b	$7B,$F0,$92,$3F,$9D,$35,$FE,$6A,$D7,$FA,$53,$5F,$E9,$4D,$7F,$A5
	.dc.b	$35,$FE,$A5,$F0,$9F,$D2,$88,$FF,$47,$6B,$FD,$09,$AF,$F5,$76,$BF
	.dc.b	$D5,$DA,$FF,$57,$7F,$AA,$FC,$24,$8F,$EA,$4D,$7F,$A6,$B5,$FE,$C4
	.dc.b	$D7,$FB,$13,$5F,$EC,$4D,$7F,$B5,$7C,$27,$F6,$22,$3F,$D7,$DA,$FF
	.dc.b	$5A,$6B,$FD,$BD,$AF,$F6,$F6,$BF,$DB,$DF,$ED,$BF,$09,$23,$FB,$53
	.dc.b	$5F,$EC,$AD,$7F,$BD,$35,$FE,$F4,$D7,$FB,$D3,$5F,$F0,$17,$C2,$7F
	.dc.b	$7A,$23,$FD,$DD,$AF,$F7,$26,$BF,$E0,$76,$BF,$E0,$76,$BF,$E0,$77
	.dc.b	$FC,$13,$E4,$3E,$0E,$92,$7F,$C1,$F1,$96,$C6,$5F,$F0,$AC,$65,$B1
	.dc.b	$97,$FC,$37,$19,$6C,$65,$12,$1F,$07,$E4,$24,$78,$9A,$F8,$1A,$F6
	.dc.b	$6B,$D0,$47,$47,$91,$0F,$4F,$C1,$FC,$22,$3F,$01,$AE,$8C,$10,$C1
	.dc.b	$0F,$7B,$C1,$F8,$3F,$07,$F3,$88,$FC,$C6,$BF,$29,$AF,$03,$44,$34
	.dc.b	$53,$E6,$FC,$1F,$DA,$23,$F6,$1A,$FD,$66,$BC,$0D,$10,$D1,$4F,$B3
	.dc.b	$F0,$7F,$86,$23,$F8,$46,$BF,$82,$6B,$C0,$D1,$0D,$14,$FC,$2F,$C1
	.dc.b	$FE,$48,$8F,$E4,$1A,$FE,$39,$AF,$03,$44,$34,$53,$F2,$3F,$07,$F9
	.dc.b	$E2,$3F,$9C,$6B,$F9,$A6,$BC,$0D,$10,$D1,$4F,$CE,$FC,$1F,$EA,$88
	.dc.b	$FE,$A1,$AF,$E9,$9A,$F0,$34,$43,$45,$3F,$53,$F0,$7F,$B6,$23,$FB
	.dc.b	$46,$BF,$B2,$6B,$C0,$D1,$0D,$14,$FD,$AF,$C1,$FF,$02,$23,$FC,$01
	.dc.b	$AF,$EF,$9A,$F0,$34,$43,$45,$3F,$80,$FC,$1F,$F0,$E2,$3F,$C3,$1A
	.dc.b	$FF,$0A,$6B,$C0,$D1,$0D,$14,$FE,$1B,$F0,$7F,$C6,$88,$FF,$18,$6B
	.dc.b	$FC,$59,$AF,$03,$44,$34,$53,$F8,$CF,$C1,$FF,$26,$23,$FC,$91,$AF
	.dc.b	$F2,$26,$BC,$0D,$10,$D1,$4F,$E4,$BF,$07,$FC,$C8,$8F,$F3,$06,$BF
	.dc.b	$CB,$9A,$F0,$34,$43,$45,$3F,$98,$FC,$1F,$F3,$E2,$3F,$CF,$1A,$FF
	.dc.b	$3A,$6B,$C0,$D1,$0D,$14,$FE,$7B,$F0,$7F,$D2,$88,$FF,$48,$6B,$FD
	.dc.b	$19,$AF,$03,$44,$34,$53,$FA,$4F,$C1,$FF,$56,$23,$FD,$51,$AF,$F5
	.dc.b	$26,$BC,$0D,$10,$D1,$4F,$EA,$BF,$07,$FD,$88,$8F,$F6,$06,$BF,$D7
	.dc.b	$9A,$F0,$34,$43,$45,$3F,$B0,$FC,$1F,$F6,$E2,$3F,$DB,$1A,$FF,$6A
	.dc.b	$6B,$C0,$D1,$0D,$14,$FE,$DB,$F0,$7F,$DE,$88,$FF,$78,$6B,$FD,$D9
	.dc.b	$AF,$03,$44,$34,$53,$FB,$50,$FB,$A7,$C2,$62,$3F,$B7,$46,$1C,$1B
	.dc.b	$8A,$D3,$31,$B7,$93,$5E,$ED,$7A,$C7,$28,$C4,$7F,$82,$1A,$F5,$39
	.dc.b	$C6,$FE,$12,$47,$81,$AF,$81,$AF,$81,$AF,$B1,$AF,$76,$BF,$0F,$E0
	.dc.b	$FC,$84,$7D,$AD,$7D,$AD,$7D,$AD,$7E,$4B,$5F,$63,$EF,$F8,$39,$1F
	.dc.b	$21,$AF,$C8,$6B,$F2,$1A,$FD,$46,$BF,$1D,$AF,$DB,$F8,$3F,$98,$47
	.dc.b	$EA,$B5,$FA,$AD,$7E,$AB,$5F,$C0,$B5,$FA,$8F,$D7,$EC,$91,$F8,$06
	.dc.b	$BF,$80,$6B,$F8,$06,$BF,$8C,$6B,$F7,$DA,$FE,$4F,$E0,$FF,$08,$47
	.dc.b	$F1,$AD,$7F,$1A,$D7,$F1,$AD,$7F,$32,$D7,$F1,$8F,$E3,$FE,$0E,$47
	.dc.b	$E6,$1A,$FE,$61,$AF,$E6,$1A,$FE,$91,$AF,$E5,$DA,$FE,$AF,$E0,$FF
	.dc.b	$38,$47,$F4,$AD,$7F,$4A,$D7,$F4,$AD,$7F,$62,$D7,$F4,$8F,$E9,$FE
	.dc.b	$0E,$47,$EC,$1A,$FE,$C1,$AF,$EC,$1A,$FE,$F1,$AF,$EB,$DA,$FF,$03
	.dc.b	$F8,$3F,$DA,$11,$FD,$EB,$5F,$DE,$B5,$FD,$EB,$5F,$E1,$2D,$7F,$78
	.dc.b	$FE,$FF,$E0,$E4,$7F,$08,$6B,$FC,$21,$AF,$F0,$86,$BF,$C5,$1A,$FF
	.dc.b	$07,$6B,$FC,$6F,$E0,$FF,$86,$11,$FE,$2A,$D7,$F8,$AB,$5F,$E2,$AD
	.dc.b	$7F,$90,$B5,$FE,$28,$FF,$17,$F8,$39,$1F,$C8,$1A,$FF,$20,$6B,$FC
	.dc.b	$81,$AF,$F2,$C6,$BF,$C7,$DA,$FF,$33,$F8,$3F,$E4,$84,$7F,$96,$B5
	.dc.b	$FE,$5A,$D7,$F9,$6B,$5F,$E7,$2D,$7F,$96,$3F,$CB,$FE,$0E,$47,$F3
	.dc.b	$86,$BF,$CE,$1A,$FF,$38,$6B,$FD,$11,$AF,$F3,$76,$BF,$D2,$FE,$0F
	.dc.b	$F9,$E1,$1F,$E8,$AD,$7F,$A2,$B5,$FE,$8A,$D7,$FA,$8B,$5F,$E8,$8F
	.dc.b	$F4,$7F,$83,$91,$FD,$41,$AF,$F5,$06,$BF,$D4,$1A,$FF,$5C,$6B,$FD
	.dc.b	$3D,$AF,$F6,$3F,$83,$FE,$A8,$47,$FA,$EB,$5F,$EB,$AD,$7F,$AE,$B5
	.dc.b	$FE,$D2,$D7,$FA,$E3,$FD,$7F,$E0,$E4,$7F,$68,$6B,$FD,$A1,$AF,$F6
	.dc.b	$86,$BF,$DD,$1A,$FF,$67,$6B,$FD,$EF,$E0,$FF,$B6,$11,$FE,$EA,$D7
	.dc.b	$FB,$AB,$5F,$EE,$AD,$7F,$C0,$2D,$7F,$BA,$3F,$E0,$0F,$BA,$FF,$00
	.dc.b	$C1,$A4,$74,$43,$FC,$12,$E5,$A6,$4F,$F0,$83,$96,$FC,$1F,$72,$3C
	.dc.b	$9A,$E8,$C1,$0C,$10,$FF,$0F,$C2,$38,$C6,$3E,$0F,$D8,$47,$CE,$D7
	.dc.b	$CA,$D7,$81,$A2,$1A,$29,$E0,$F8,$3F,$90,$47,$E1,$B5,$F8,$2D,$78
	.dc.b	$1A,$21,$A2,$9E,$CF,$83,$FA,$84,$7E,$7B,$5F,$9A,$D7,$81,$A2,$1A
	.dc.b	$29,$F2,$3E,$0F,$F0,$04,$7E,$DB,$5F,$B2,$D7,$81,$A2,$1A,$29,$F5
	.dc.b	$3E,$0F,$F1,$84,$7F,$0E,$D7,$F0,$AD,$78,$1A,$21,$A2,$9F,$80,$F8
	.dc.b	$3F,$CC,$11,$FC,$9B,$5F,$C8,$B5,$E0,$68,$86,$8A,$7E,$33,$E0,$FF
	.dc.b	$48,$47,$F3,$ED,$7F,$3A,$D7,$81,$A2,$1A,$29,$F9,$8F,$83,$FD,$81
	.dc.b	$1F,$D5,$B5,$FD,$4B,$5E,$06,$88,$68,$A7,$E9,$3E,$0F,$F7,$84,$7F
	.dc.b	$6E,$D7,$F6,$AD,$78,$1A,$21,$A2,$9F,$B0,$F8,$3F,$E1,$04,$7F,$81
	.dc.b	$B5,$FE,$02,$D7,$81,$A2,$1A,$29,$FB,$CF,$83,$FE,$28,$47,$F8,$7B
	.dc.b	$5F,$E1,$AD,$78,$1A,$21,$A2,$9F,$C2,$3E,$0F,$F9,$01,$1F,$E3,$6D
	.dc.b	$7F,$8C,$B5,$E0,$68,$86,$8A,$7F,$14,$F8,$3F,$E5,$84,$7F,$93,$B5
	.dc.b	$FE,$4A,$D7,$81,$A2,$1A,$29,$FC,$83,$E0,$FF,$9C,$11,$FE,$66,$D7
	.dc.b	$F9,$8B,$5E,$06,$88,$68,$A7,$F2,$CF,$83,$FE,$88,$47,$F9,$FB,$5F
	.dc.b	$E7,$AD,$78,$1A,$21,$A2,$9F,$CE,$3E,$0F,$FA,$81,$1F,$E9,$6D,$7F
	.dc.b	$A4,$B5,$E0,$68,$86,$8A,$7F,$44,$F8,$3F,$EB,$84,$7F,$AB,$B5,$FE
	.dc.b	$AA,$D7,$81,$A2,$1A,$29,$FD,$43,$E0,$FF,$B4,$11,$FE,$C6,$D7,$FB
	.dc.b	$0B,$5E,$06,$88,$68,$A7,$F5,$CF,$83,$FE,$E8,$47,$FB,$7B,$5F,$ED
	.dc.b	$AD,$78,$1A,$21,$A2,$9F,$DA,$3E,$0F,$FC,$00,$47,$FB,$DB,$5F,$EF
	.dc.b	$2D,$78,$1A,$21,$A2,$82,$0F,$BA,$7C,$26,$22,$46,$1C,$1C,$5C,$38
	.dc.b	$6B,$C2,$36,$ED,$AF,$76,$C7,$08,$F1,$B5,$E0,$72,$70,$F1,$31,$F0
	.dc.b	$9C,$88,$E0,$87,$95,$CB,$4C,$9E,$E7,$2D,$3C,$1F,$09,$E6,$23,$82
	.dc.b	$1F,$05,$CB,$4C,$9F,$29,$CB,$4F,$67,$C2,$7C,$22,$38,$21,$F3,$5C
	.dc.b	$B4,$C9,$F5,$9C,$B4,$F9,$1F,$09,$F3,$88,$E0,$87,$D9,$72,$D3,$27
	.dc.b	$E0,$9C,$B4,$FA,$9F,$09,$F6,$88,$E0,$87,$E1,$5C,$B4,$C9,$F8,$E7
	.dc.b	$2D,$3F,$01,$F0,$9F,$86,$23,$82,$1F,$91,$72,$D3,$27,$E6,$9C,$B4
	.dc.b	$FC,$67,$C2,$7E,$48,$8E,$08,$7E,$75,$CB,$4C,$9F,$A6,$72,$D3,$F3
	.dc.b	$1F,$09,$F9,$E2,$38,$21,$FA,$97,$2D,$32,$7E,$C9,$CB,$4F,$D2,$7C
	.dc.b	$27,$EA,$88,$E0,$87,$ED,$5C,$B4,$C9,$FB,$E7,$2D,$3F,$61,$F0,$9F
	.dc.b	$B6,$23,$82,$1F,$C0,$5C,$B4,$C9,$FC,$29,$CB,$4F,$DE,$7C,$27,$F0
	.dc.b	$22,$38,$21,$FC,$35,$CB,$4C,$9F,$C5,$9C,$B4,$FE,$11,$F0,$9F,$C3
	.dc.b	$88,$E0,$87,$F1,$97,$2D,$32,$7F,$22,$72,$D3,$F8,$A7,$C2,$7F,$1A
	.dc.b	$23,$82,$1F,$C9,$5C,$B4,$C9,$FC,$B9,$CB,$4F,$E4,$1F,$09,$FC,$98
	.dc.b	$8E,$08,$7F,$31,$72,$D3,$27,$F3,$A7,$2D,$3F,$96,$7C,$27,$F3,$22
	.dc.b	$38,$21,$FC,$F5,$CB,$4C,$9F,$D1,$9C,$B4,$FE,$71,$F0,$9F,$CF,$88
	.dc.b	$E0,$87,$F4,$97,$2D,$32,$7F,$52,$72,$D3,$FA,$27,$C2,$7F,$4A,$23
	.dc.b	$82,$1F,$D5,$5C,$B4,$C9,$FD,$79,$CB,$4F,$EA,$1F,$09,$FD,$58,$8E
	.dc.b	$08,$7F,$61,$72,$D3,$27,$F6,$A7,$2D,$3F,$AE,$7C,$27,$F6,$22,$38
	.dc.b	$21,$FD,$B5,$CB,$4C,$9F,$DD,$9C,$B4,$FE,$D1,$F0,$9F,$DB,$88,$E0
	.dc.b	$87,$F7,$97,$2D,$32,$7F,$80,$9C,$B4,$FE,$FB,$C4,$F8,$3F,$EF,$AD
	.dc.b	$8C,$C6,$08,$E0,$87,$F8,$32,$38,$36,$36,$D3,$67,$F8,$56,$1F,$07
	.dc.b	$E0,$24,$79,$1A,$F5,$6B,$C5,$AF,$F8,$8A,$39,$BC,$48,$79,$7E,$0F
	.dc.b	$DA,$47,$E0,$B5,$D1,$82,$18,$21,$E9,$78,$3E,$BF,$83,$F9,$24,$7E
	.dc.b	$6B,$5F,$8C,$D7,$81,$A2,$1A,$29,$F1,$FE,$0F,$EA,$91,$FB,$2D,$7E
	.dc.b	$93,$5E,$06,$88,$68,$A7,$D3,$F8,$3F,$C0,$91,$FC,$2B,$5F,$BC,$D7
	.dc.b	$81,$A2,$1A,$29,$F7,$FE,$0F,$F1,$A4,$7F,$22,$D7,$F1,$4D,$78,$1A
	.dc.b	$21,$A2,$9F,$8B,$F8,$3F,$CC,$91,$FC,$EB,$5F,$CB,$35,$E0,$68,$86
	.dc.b	$8A,$7E,$5F,$E0,$FF,$4A,$47,$F5,$2D,$7F,$44,$D7,$81,$A2,$1A,$29
	.dc.b	$FA,$3F,$83,$FD,$89,$1F,$DA,$B5,$FD,$73,$5E,$06,$88,$68,$A7,$EB
	.dc.b	$FE,$0F,$F7,$A4,$7F,$80,$B5,$FD,$D3,$5E,$06,$88,$68,$A7,$EE,$FE
	.dc.b	$0F,$F8,$49,$1F,$E1,$AD,$7F,$83,$35,$E0,$68,$86,$8A,$7F,$07,$F8
	.dc.b	$3F,$E2,$A4,$7F,$8C,$B5,$FE,$24,$D7,$81,$A2,$1A,$29,$FC,$4F,$E0
	.dc.b	$FF,$90,$91,$FE,$4A,$D7,$F8,$F3,$5E,$06,$88,$68,$A7,$F1,$FF,$83
	.dc.b	$FE,$5A,$47,$F9,$8B,$5F,$E5,$4D,$78,$1A,$21,$A2,$9F,$CA,$FE,$0F
	.dc.b	$F9,$C9,$1F,$E7,$AD,$7F,$9B,$35,$E0,$68,$86,$8A,$7F,$37,$F8,$3F
	.dc.b	$E8,$A4,$7F,$A4,$B5,$FE,$84,$D7,$81,$A2,$1A,$29,$FD,$0F,$E0,$FF
	.dc.b	$A8,$91,$FE,$AA,$D7,$FA,$73,$5E,$06,$88,$68,$A7,$F4,$FF,$83,$FE
	.dc.b	$BA,$47,$FB,$0B,$5F,$EB,$4D,$78,$1A,$21,$A2,$9F,$D6,$FE,$0F,$FB
	.dc.b	$49,$1F,$ED,$AD,$7F,$B3,$35,$E0,$68,$86,$8A,$7F,$67,$F8,$3F,$EE
	.dc.b	$A4,$7F,$BC,$B5,$FE,$E4,$D7,$81,$A2,$1A,$28,$20,$FB,$AF,$F7,$FB
	.dc.b	$4F,$F8,$02,$E0,$8E,$08,$72,$39,$BA,$70,$77,$70,$E1,$E9,$F0,$7D
	.dc.b	$89,$1C,$0F,$23,$87,$83,$5F,$04,$23,$11,$E2,$6B,$E1,$7D,$7F,$07
	.dc.b	$23,$B4,$A3,$11,$E5,$6B,$D9,$AF,$B9,$AF,$B5,$AF,$CB,$F8,$3F,$92
	.dc.b	$47,$BB,$5F,$84,$D7,$D6,$D7,$E5,$B5,$F9,$4F,$D4,$F8,$39,$1E,$A6
	.dc.b	$BF,$35,$AF,$C6,$6B,$F5,$9A,$FD,$56,$BF,$83,$F8,$3F,$C0,$91,$F8
	.dc.b	$ED,$7E,$D3,$5F,$A6,$D7,$F0,$6D,$7F,$04,$FE,$33,$E0,$E4,$7D,$26
	.dc.b	$BF,$85,$6B,$F7,$9A,$FE,$39,$AF,$E3,$5A,$FE,$6F,$E0,$FF,$32,$47
	.dc.b	$EF,$B5,$FC,$93,$5F,$C5,$B5,$FC,$DB,$5F,$CD,$3F,$A4,$F8,$39,$1F
	.dc.b	$8A,$6B,$F9,$D6,$BF,$96,$6B,$FA,$66,$BF,$A5,$6B,$FB,$3F,$83,$FD
	.dc.b	$89,$1F,$CB,$B5,$FD,$53,$5F,$D1,$B5,$FD,$9B,$5F,$D9,$3F,$BC,$F8
	.dc.b	$39,$1F,$A2,$6B,$FB,$56,$BF,$AE,$6B,$FB,$E6,$BF,$BD,$6B,$FC,$2F
	.dc.b	$E0,$FF,$84,$91,$FD,$7B,$5F,$E0,$4D,$7F,$76,$D7,$F8,$5B,$5F,$E1
	.dc.b	$4F,$F1,$4F,$83,$91,$FB,$A6,$BF,$C3,$5A,$FF,$06,$6B,$FC,$59,$AF
	.dc.b	$F1,$56,$BF,$C8,$FE,$0F,$F9,$09,$1F,$E0,$ED,$7F,$8D,$35,$FE,$26
	.dc.b	$D7,$F9,$1B,$5F,$E4,$4F,$F2,$CF,$83,$91,$FC,$49,$AF,$F2,$56,$BF
	.dc.b	$C7,$9A,$FF,$2E,$6B,$FC,$B5,$AF,$F3,$BF,$83,$FE,$72,$47,$F8,$FB
	.dc.b	$5F,$E6,$4D,$7F,$95,$B5,$FE,$76,$D7,$F9,$D3,$FD,$13,$E0,$E4,$7F
	.dc.b	$2A,$6B,$FC,$F5,$AF,$F3,$66,$BF,$D1,$9A,$FF,$45,$6B,$FD,$4F,$E0
	.dc.b	$FF,$A8,$91,$FE,$6E,$D7,$FA,$53,$5F,$E8,$6D,$7F,$A9,$B5,$FE,$A4
	.dc.b	$FF,$5C,$F8,$39,$1F,$D0,$9A,$FF,$55,$6B,$FD,$39,$AF,$F5,$E6,$BF
	.dc.b	$D7,$5A,$FF,$6B,$F8,$3F,$ED,$24,$7F,$A7,$B5,$FE,$C4,$D7,$FA,$DB
	.dc.b	$5F,$ED,$6D,$7F,$B5,$3F,$DD,$3E,$0E,$47,$F5,$A6,$BF,$DB,$5A,$FF
	.dc.b	$66,$6B,$FD,$D9,$AF,$F7,$56,$BF,$E0,$3F,$83,$FF,$00,$91,$FE,$CE
	.dc.b	$D7,$FB,$D3,$5F,$EE,$6D,$7F,$C0,$6D,$7F,$C0,$4F,$F7,$D5,$F7,$4F
	.dc.b	$84,$C3,$63,$FC,$04,$D7,$FC,$16,$D7,$FC,$20,$D7,$FC,$36,$D7,$FC
	.dc.b	$3C,$D7,$BC,$76,$23,$D9,$AF,$6F,$83,$F1,$91,$F1,$B5,$D1,$82,$18
	.dc.b	$21,$E6,$78,$3E,$8F,$83,$F7,$91,$F7,$B5,$F7,$B5,$E0,$68,$86,$8A
	.dc.b	$7C,$4F,$83,$F9,$64,$7E,$5B,$5F,$96,$D7,$81,$A2,$1A,$29,$F4,$3E
	.dc.b	$0F,$EB,$91,$FA,$ED,$7E,$BB,$5E,$06,$88,$68,$A7,$DC,$F4,$BF,$06
	.dc.b	$47,$F0,$6D,$7F,$06,$D7,$81,$A2,$1A,$29,$F8,$8F,$83,$FC,$79,$1F
	.dc.b	$C7,$B5,$FC,$7B,$5E,$06,$88,$68,$A7,$E5,$3E,$0F,$F3,$64,$7F,$36
	.dc.b	$D7,$F3,$6D,$78,$1A,$21,$A2,$9F,$A0,$F8,$3F,$D3,$91,$FD,$3B,$5F
	.dc.b	$D3,$B5,$E0,$68,$86,$8A,$7E,$B3,$E0,$FF,$66,$47,$F6,$6D,$7F,$66
	.dc.b	$D7,$81,$A2,$1A,$29,$FB,$8F,$83,$FD,$F9,$1F,$DF,$B5,$FD,$FB,$5E
	.dc.b	$06,$88,$68,$A7,$F0,$4F,$83,$FE,$16,$47,$F8,$5B,$5F,$E1,$6D,$78
	.dc.b	$1A,$21,$A2,$9F,$C4,$3E,$0F,$F8,$B9,$1F,$E2,$ED,$7F,$8B,$B5,$E0
	.dc.b	$68,$86,$8A,$7F,$1C,$F8,$3F,$E4,$64,$7F,$91,$B5,$FE,$46,$D7,$81
	.dc.b	$A2,$1A,$29,$FC,$A3,$E0,$FF,$97,$91,$FE,$5E,$D7,$F9,$7B,$5E,$06
	.dc.b	$88,$68,$A7,$F3,$4F,$83,$FE,$76,$47,$F9,$DB,$5F,$E7,$6D,$78,$1A
	.dc.b	$21,$A2,$9F,$D0,$3E,$0F,$FA,$39,$1F,$E8,$ED,$7F,$A3,$B5,$E0,$68
	.dc.b	$86,$8A,$7F,$4C,$F8,$3F,$EA,$64,$7F,$A9,$B5,$FE,$A6,$D7,$81,$A2
	.dc.b	$1A,$29,$FD,$63,$E0,$FF,$AF,$91,$FE,$BE,$D7,$FA,$FB,$5E,$06,$88
	.dc.b	$68,$A7,$F6,$4F,$83,$FE,$D6,$47,$FB,$5B,$5F,$ED,$6D,$78,$1A,$21
	.dc.b	$A2,$9F,$DC,$3E,$0F,$FB,$B9,$1F,$EE,$ED,$7F,$BB,$B5,$E0,$68,$86
	.dc.b	$8A,$08,$3E,$E9,$F0,$9F,$E0,$96,$C6,$8D,$B8,$70,$71,$8E,$04,$72
	.dc.b	$64,$B1,$C5,$E8,$D7,$AC,$71,$7A,$C6,$5F,$42,$F8,$4F,$31,$23,$AB
	.dc.b	$5E,$2D,$7B,$B5,$F5,$B5,$EA,$FC,$1F,$84,$91,$E8,$6B,$E4,$6B,$EA
	.dc.b	$6B,$F1,$9A,$FA,$1A,$FC,$FF,$84,$F9,$C4,$7E,$2B,$5F,$82,$D7,$E3
	.dc.b	$B5,$FA,$6D,$7E,$2B,$F6,$7E,$12,$47,$D0,$6B,$F3,$1A,$FD,$26,$BF
	.dc.b	$79,$AF,$D0,$6B,$F8,$7F,$84,$FC,$31,$1F,$BA,$D7,$EC,$B5,$FB,$ED
	.dc.b	$7F,$16,$D7,$EE,$BF,$91,$F8,$49,$1F,$88,$6B,$F8,$46,$BF,$8A,$6B
	.dc.b	$F9,$66,$BF,$88,$6B,$F9,$FF,$84,$FC,$F1,$1F,$CA,$B5,$FC,$8B,$5F
	.dc.b	$CB,$B5,$FD,$1B,$5F,$CA,$BF,$A9,$F8,$49,$1F,$A0,$6B,$F9,$C6,$BF
	.dc.b	$A2,$6B,$FA,$E6,$BF,$A0,$6B,$FB,$7F,$84,$FD,$B1,$1F,$D6,$B5,$FD
	.dc.b	$4B,$5F,$D7,$B5,$FD,$DB,$5F,$D6,$BF,$C0,$7E,$12,$47,$EE,$1A,$FE
	.dc.b	$D1,$AF,$EE,$9A,$FF,$06,$6B,$FB,$86,$BF,$C3,$FE,$13,$F8,$71,$1F
	.dc.b	$E0,$AD,$7F,$80,$B5,$FE,$0E,$D7,$F8,$9B,$5F,$E0,$AF,$F1,$9F,$84
	.dc.b	$91,$FC,$41,$AF,$F0,$C6,$BF,$C4,$9A,$FF,$1E,$6B,$FC,$41,$AF,$F2
	.dc.b	$7F,$84,$FE,$4C,$47,$F8,$EB,$5F,$E3,$2D,$7F,$8F,$B5,$FE,$56,$D7
	.dc.b	$F8,$EB,$FC,$C7,$E1,$24,$7F,$28,$6B,$FC,$91,$AF,$F2,$A6,$BF,$CD
	.dc.b	$9A,$FF,$28,$6B,$FC,$FF,$E1,$3F,$9F,$11,$FE,$6A,$D7,$F9,$8B,$5F
	.dc.b	$E6,$ED,$7F,$A1,$B5,$FE,$6A,$FF,$49,$F8,$49,$1F,$D0,$1A,$FF,$3C
	.dc.b	$6B,$FD,$09,$AF,$F4,$E6,$BF,$D0,$1A,$FF,$57,$F8,$4F,$EA,$C4,$7F
	.dc.b	$A6,$B5,$FE,$92,$D7,$FA,$7B,$5F,$EB,$6D,$7F,$A6,$BF,$D8,$7E,$12
	.dc.b	$47,$F5,$86,$BF,$D5,$1A,$FF,$5A,$6B,$FD,$99,$AF,$F5,$86,$BF,$DB
	.dc.b	$FE,$13,$FB,$71,$1F,$EC,$AD,$7F,$B0,$B5,$FE,$CE,$D7,$FB,$9B,$5F
	.dc.b	$EC,$AF,$F7,$9F,$84,$91,$FD,$C1,$AF,$F6,$C6,$BF,$DC,$9A,$FF,$7E
	.dc.b	$6B,$FD,$C1,$AB,$DE,$FE,$FA,$FF,$82,$DA,$E4,$CB,$82,$38,$21,$FE
	.dc.b	$06,$72,$D3,$27,$F8,$39,$7C,$27,$42,$47,$16,$BB,$6B,$FE,$19,$8E
	.dc.b	$2E,$C8,$6C,$A6,$FE,$0F,$D2,$47,$CC,$D7,$C8,$D7,$81,$A2,$1A,$29
	.dc.b	$E3,$F8,$3F,$8A,$47,$E1,$35,$F8,$0D,$78,$1A,$21,$A2,$9E,$FF,$83
	.dc.b	$FA,$24,$7E,$73,$5F,$98,$D7,$81,$A2,$1A,$29,$F2,$FE,$0F,$EE,$91
	.dc.b	$FB,$4D,$7E,$C3,$5E,$06,$88,$68,$A7,$D7,$BF,$C4,$91,$FC,$33,$5F
	.dc.b	$C2,$35,$E0,$68,$86,$8A,$7E,$0F,$E0,$FF,$2A,$47,$F2,$4D,$7F,$20
	.dc.b	$D7,$81,$A2,$1A,$29,$F8,$FF,$83,$FD,$09,$1F,$CF,$35,$FC,$E3,$5E
	.dc.b	$06,$88,$68,$A7,$E6,$FE,$0F,$F5,$A4,$7F,$54,$D7,$F5,$0D,$78,$1A
	.dc.b	$21,$A2,$9F,$A7,$F8,$3F,$DC,$91,$FD,$B3,$5F,$DA,$35,$E0,$68,$86
	.dc.b	$8A,$7E,$CF,$E0,$FF,$82,$91,$FE,$04,$D7,$F8,$03,$5E,$06,$88,$68
	.dc.b	$A7,$EF,$FE,$0F,$F8,$89,$1F,$E1,$CD,$7F,$86,$35,$E0,$68,$86,$8A
	.dc.b	$7F,$0B,$F8,$3F,$E3,$A4,$7F,$8D,$35,$FE,$30,$D7,$81,$A2,$1A,$29
	.dc.b	$FC,$5F,$E0,$FF,$94,$91,$FE,$4C,$D7,$F9,$23,$5E,$06,$88,$68,$A7
	.dc.b	$F2,$3F,$83,$FE,$6A,$47,$F9,$93,$5F,$E6,$0D,$78,$1A,$21,$A2,$9F
	.dc.b	$CB,$FE,$0F,$FA,$09,$1F,$E7,$CD,$7F,$9E,$35,$E0,$68,$86,$8A,$7F
	.dc.b	$3B,$F8,$3F,$E9,$A4,$7F,$A5,$35,$FE,$90,$D7,$81,$A2,$1A,$29,$FD
	.dc.b	$1F,$E0,$FF,$AC,$91,$FE,$AC,$D7,$FA,$A3,$5E,$06,$88,$68,$A7,$F5
	.dc.b	$3F,$83,$FE,$CA,$47,$FB,$13,$5F,$EC,$0D,$78,$1A,$21,$A2,$9F,$D7
	.dc.b	$FE,$0F,$FB,$89,$1F,$ED,$CD,$7F,$B6,$35,$E0,$68,$86,$8A,$7F,$6B
	.dc.b	$F8,$3F,$EF,$A4,$7F,$BD,$35,$FE,$F0,$D7,$81,$A2,$1A,$28,$20,$FB
	.dc.b	$A7,$C1,$FF,$84,$1B,$1A,$36,$E1,$C1,$C6,$38,$11,$C9,$92,$C7,$17
	.dc.b	$A3,$5E,$B1,$C5,$EB,$19,$7D,$0B,$E0,$FC,$C4,$8E,$AD,$78,$B5,$EE
	.dc.b	$D7,$D6,$D7,$AB,$F0,$7E,$0E,$47,$A1,$AF,$91,$AF,$A9,$AF,$C6,$6B
	.dc.b	$E8,$6B,$F3,$FE,$0F,$E7,$11,$F8,$AD,$7E,$0B,$5F,$8E,$D7,$E9,$B5
	.dc.b	$F8,$AF,$D9,$F8,$39,$1F,$41,$AF,$CC,$6B,$F4,$9A,$FD,$E6,$BF,$41
	.dc.b	$AF,$E1,$FE,$0F,$F0,$C4,$7E,$EB,$5F,$B2,$D7,$EF,$B5,$FC,$5B,$5F
	.dc.b	$BA,$FE,$47,$E0,$E4,$7E,$21,$AF,$E1,$1A,$FE,$29,$AF,$E5,$9A,$FE
	.dc.b	$21,$AF,$E7,$FE,$0F,$F3,$C4,$7F,$2A,$D7,$F2,$2D,$7F,$2E,$D7,$F4
	.dc.b	$6D,$7F,$2A,$FE,$A7,$E0,$E4,$7E,$81,$AF,$E7,$1A,$FE,$89,$AF,$EB
	.dc.b	$9A,$FE,$81,$AF,$ED,$FE,$0F,$F6,$C4,$7F,$5A,$D7,$F5,$2D,$7F,$5E
	.dc.b	$D7,$F7,$6D,$7F,$5A,$FF,$01,$F8,$39,$1F,$B8,$6B,$FB,$46,$BF,$BA
	.dc.b	$6B,$FC,$19,$AF,$EE,$1A,$FF,$0F,$F8,$3F,$E1,$C4,$7F,$82,$B5,$FE
	.dc.b	$02,$D7,$F8,$3B,$5F,$E2,$6D,$7F,$82,$BF,$C6,$7E,$0E,$47,$F1,$06
	.dc.b	$BF,$C3,$1A,$FF,$12,$6B,$FC,$79,$AF,$F1,$06,$BF,$C9,$FE,$0F,$F9
	.dc.b	$31,$1F,$E3,$AD,$7F,$8C,$B5,$FE,$3E,$D7,$F9,$5B,$5F,$E3,$AF,$F3
	.dc.b	$1F,$83,$91,$FC,$A1,$AF,$F2,$46,$BF,$CA,$9A,$FF,$36,$6B,$FC,$A1
	.dc.b	$AF,$F3,$FF,$83,$FE,$7C,$47,$F9,$AB,$5F,$E6,$2D,$7F,$9B,$B5,$FE
	.dc.b	$86,$D7,$F9,$AB,$FD,$27,$E0,$E4,$7F,$40,$6B,$FC,$F1,$AF,$F4,$26
	.dc.b	$BF,$D3,$9A,$FF,$40,$6B,$FD,$5F,$E0,$FF,$AB,$11,$FE,$9A,$D7,$FA
	.dc.b	$4B,$5F,$E9,$ED,$7F,$AD,$B5,$FE,$9A,$FF,$61,$F8,$39,$1F,$D6,$1A
	.dc.b	$FF,$54,$6B,$FD,$69,$AF,$F6,$66,$BF,$D6,$1A,$FF,$6F,$F8,$3F,$ED
	.dc.b	$C4,$7F,$B2,$B5,$FE,$C2,$D7,$FB,$3B,$5F,$EE,$6D,$7F,$B2,$BF,$DE
	.dc.b	$7E,$0E,$47,$F7,$06,$BF,$DB,$1A,$FF,$72,$6B,$FD,$F9,$AF,$F7,$06
	.dc.b	$AF,$BA,$7C,$27,$F7,$E8,$C3,$63,$31,$FE,$02,$6B,$FE,$08,$6B,$FE
	.dc.b	$0D,$6B,$FE,$14,$6B,$FE,$17,$6B,$D8,$E9,$1D,$C8,$F7,$F8,$3F,$11
	.dc.b	$1C,$10,$F1,$39,$69,$93,$C8,$E5,$A7,$AF,$E0,$FD,$C4,$70,$43,$DC
	.dc.b	$E5,$A6,$4F,$80,$E5,$A7,$C7,$F8,$3F,$94,$47,$04,$3E,$53,$96,$99
	.dc.b	$3E,$63,$96,$9F,$4F,$E0,$FE,$B1,$1C,$10,$FA,$CE,$5A,$64,$FB,$0E
	.dc.b	$5A,$7D,$FF,$83,$FC,$11,$1C,$10,$FC,$13,$96,$99,$3F,$08,$E5,$A7
	.dc.b	$E2,$FE,$0F,$F1,$C4,$70,$43,$F1,$CE,$5A,$64,$FC,$83,$96,$9F,$97
	.dc.b	$F8,$3F,$CD,$11,$C1,$0F,$CD,$39,$69,$93,$F3,$8E,$5A,$7E,$8F,$E0
	.dc.b	$FF,$4C,$47,$04,$3F,$4C,$E5,$A6,$4F,$D4,$39,$69,$FA,$FF,$83,$FD
	.dc.b	$91,$1C,$10,$FD,$93,$96,$99,$3F,$68,$E5,$A7,$FA,$0F,$EE,$84,$00

;--------------------------------------------------------------------------------
;	FMOVE.X <mem>,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovextoreg_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040+MC68060+FPSP060,-(sp)
	peamsg	'FMOVE.X <EA>,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVE.X <ea>,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fmovextoreg_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a0,a2			;source handle,...
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;<mem>,FPn
	fmove.l	#0,fpcr
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fmove.x	([a2]),fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)		;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVE.X <mem>='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmovextoreg_data_compressed::
	.dc.b	$00,$02,$55,$4C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FE
	.dc.b	$3D,$CB,$F2,$8E,$7F,$99,$74,$FC,$F3,$AF,$E8,$DD,$BF,$50,$EF,$FA
	.dc.b	$D7,$83,$F6,$4F,$0F,$ED,$DE,$2F,$DE,$3F,$01,$F0,$0F,$7E,$FB,$F8
	.dc.b	$F6,$F3,$ED,$EB,$DB,$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D
	.dc.b	$3D,$BE,$BE,$DF,$6F,$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$FC
	.dc.b	$97,$89,$F6,$FE,$03,$F2,$AB,$DA,$CF,$B5,$AF,$6B,$7E,$D7,$0F,$6B
	.dc.b	$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83
	.dc.b	$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BC
	.dc.b	$FE,$D7,$EA,$7E,$07,$EE,$74,$7E,$B1,$7B,$59,$F6,$B5,$ED,$6F,$DA
	.dc.b	$E1,$ED,$71,$F6,$B9,$7E,$01,$CF,$F0,$AE,$9F,$8A,$75,$FC,$7B,$B7
	.dc.b	$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$E5
	.dc.b	$F7,$BC,$DE,$F7,$9F,$DE,$FE,$01,$F0,$DF,$C0,$0F,$C1,$74,$7F,$04
	.dc.b	$5E,$EE,$BE,$1B,$7F,$0D,$C3,$E1,$B8,$FC,$37,$2F,$81,$E7,$F0,$3D
	.dc.b	$3E,$07,$AF,$C0,$F6,$F8,$1E,$FF,$03,$E0,$F8,$1F,$0F,$C0,$F8,$BE
	.dc.b	$07,$C7,$F0,$3E,$4F,$81,$F2,$FC,$0F,$9B,$E0,$7C,$FF,$03,$E8,$F8
	.dc.b	$1F,$E3,$1F,$0F,$F8,$7E,$83,$F8,$77,$F8,$EA,$F8,$0D,$FC,$07,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$6F,$80,$F3,$FC,$07,$A3,$E0,$3F,$98,$7C,$37,$E2,$39,$FF
	.dc.b	$8D,$7C,$06,$3E,$03,$3F,$01,$AF,$80,$DF,$C0,$70,$F8,$0E,$3F,$01
	.dc.b	$CB,$E0,$39,$FC,$07,$4F,$80,$EB,$F0,$1D,$BE,$03,$BF,$C0,$78,$3E
	.dc.b	$03,$C3,$F0,$1E,$2F,$80,$F1,$FC,$07,$93,$E0,$3C,$BF,$01,$E6,$F8
	.dc.b	$0F,$E8,$DF,$11,$FD,$18,$FC,$3F,$39,$FC,$3B,$FD,$31,$7B,$FC,$3D
	.dc.b	$FE,$3E,$FF,$2F,$7F,$9F,$BF,$D3,$DF,$EB,$EF,$F6,$F7,$FB,$FB,$FE
	.dc.b	$0F,$7F,$C3,$EF,$F8,$BD,$FF,$1F,$BF,$E4,$F7,$FC,$BE,$FF,$9B,$DF
	.dc.b	$F3,$FB,$FE,$8F,$7F,$D3,$EF,$FF,$5F,$F8,$9F,$C4,$1F,$87,$E7,$3F
	.dc.b	$87,$7F,$B2,$AF,$7B,$87,$BD,$C7,$DE,$E5,$EF,$73,$F7,$BA,$7B,$DD
	.dc.b	$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3,$F7,$BC
	.dc.b	$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$D1,$EF,$7F,$77,$F8,$AF,$C4
	.dc.b	$1F,$C3,$B9,$BF,$BD,$AF,$77,$5E,$EE,$FD,$DE,$1E,$EF,$1F,$77,$97
	.dc.b	$BB,$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7B,$BD,$FD,$DF,$07,$BB,$E1,$F7
	.dc.b	$7C,$5E,$EF,$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE,$F9,$FD,$DF,$F0
	.dc.b	$77,$C3,$FF,$83,$8F,$C4,$9F,$88,$E7,$FE,$A5,$EE,$EB,$DD,$DF,$BB
	.dc.b	$C3,$DD,$E3,$EE,$F2,$F7,$79,$FB,$BD,$3D,$DE,$BE,$EF,$6F,$77,$BF
	.dc.b	$BB,$E0,$F7,$7C,$3E,$EF,$8B,$DD,$F1,$FB,$BE,$4F,$77,$CB,$EE,$F9
	.dc.b	$BD,$DF,$3F,$BB,$FE,$25,$F8,$BF,$E1,$C7,$E1,$CF,$C3,$73,$1F,$C3
	.dc.b	$3F,$E2,$C5,$EE,$71,$F7,$39,$7B,$9C,$FD,$CE,$9E,$E7,$5F,$73,$B7
	.dc.b	$B9,$DF,$DC,$F0,$7B,$9E,$1F,$73,$C5,$EE,$78,$FD,$CF,$27,$B9,$E5
	.dc.b	$F7,$3C,$DE,$E7,$9F,$DC,$F4,$7B,$9F,$E3,$DF,$8C,$FE,$1C,$7E,$1F
	.dc.b	$94,$FE,$1D,$FF,$21,$AF,$6F,$87,$B7,$C7,$DB,$E5,$ED,$F3,$F6,$FA
	.dc.b	$7B,$7D,$7D,$BE,$DE,$DF,$7F,$6F,$C1,$ED,$F8,$7D,$BF,$17,$B7,$E3
	.dc.b	$F6,$FC,$9E,$DF,$97,$DB,$F3,$7B,$7E,$7F,$6F,$D1,$ED,$FF,$95,$7E
	.dc.b	$37,$FC,$AA,$7F,$10,$3F,$10,$7F,$0E,$E6,$FF,$2D,$AF,$6B,$87,$B5
	.dc.b	$C7,$DA,$E5,$ED,$73,$F6,$BA,$7B,$5D,$7D,$AE,$DE,$D7,$7F,$6B,$C1
	.dc.b	$ED,$78,$7D,$AF,$17,$B5,$E3,$F6,$BC,$9E,$D7,$97,$DA,$F3,$7B,$5E
	.dc.b	$7F,$6B,$D1,$ED,$7F,$9B,$7E,$0B,$FA,$B1,$FC,$3C,$FC,$49,$F8,$8E
	.dc.b	$6F,$F0,$3F,$B5,$BF,$6B,$87,$B5,$C7,$DA,$E5,$ED,$73,$F6,$BA,$7B
	.dc.b	$5D,$7D,$AE,$DE,$D7,$7F,$6B,$C1,$ED,$78,$7D,$AF,$17,$B5,$E3,$F6
	.dc.b	$BC,$9E,$D7,$97,$DA,$F3,$7B,$5F,$E8,$3F,$8E,$FE,$A8,$7E,$1C,$FC
	.dc.b	$37,$21,$FC,$33,$FE,$8A,$5E,$CF,$1F,$67,$97,$B3,$CF,$D9,$E9,$EC
	.dc.b	$F5,$F6,$7B,$7B,$3D,$FD,$9F,$07,$B3,$E1,$F6,$7C,$5E,$CF,$8F,$D9
	.dc.b	$F2,$7B,$3E,$5F,$67,$CD,$EC,$F9,$FD,$9F,$47,$B3,$FE,$9B,$F8,$FF
	.dc.b	$F4,$D9,$FD,$58,$FE,$1C,$7E,$1F,$94,$FE,$1D,$FF,$51,$AF,$63,$97
	.dc.b	$B1,$CF,$D8,$E9,$EC,$75,$F6,$3B,$7B,$1D,$FD,$8F,$07,$B1,$E1,$F6
	.dc.b	$3C,$5E,$C7,$8F,$D8,$F2,$7B,$1E,$5F,$63,$CD,$EC,$79,$FD,$8F,$47
	.dc.b	$B1,$E9,$F6,$3F,$D6,$9F,$21,$FE,$1D,$3F,$AB,$1F,$C4,$0F,$C4,$1F
	.dc.b	$C3,$B9,$7F,$D7,$6B,$D7,$E3,$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE
	.dc.b	$BF,$6F,$5F,$BF,$AF,$E0,$F5,$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F
	.dc.b	$5F,$CB,$EB,$F9,$BD,$7F,$3F,$AF,$FE,$CB,$F8,$CF,$F0,$E1,$FC,$3C
	.dc.b	$FC,$49,$F8,$8E,$4F,$F3,$E7,$AF,$BF,$5F,$87,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FF,$EE,$1F,$91,$FF
	.dc.b	$70,$9F,$E1,$C3,$FA,$A1,$F8,$73,$F0,$DC,$87,$F0,$CF,$FB,$A9,$7A
	.dc.b	$DC,$FD,$6E,$9E,$B7,$5F,$5B,$B7,$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B
	.dc.b	$C5,$EB,$78,$FD,$6F,$27,$AD,$E5,$F5,$BC,$DE,$B7,$9F,$D6,$F4,$7A
	.dc.b	$DE,$9F,$5B,$FD,$F5,$F2,$5F,$E6,$33,$FC,$38,$7F,$56,$3F,$87,$1F
	.dc.b	$87,$E4,$3F,$87,$46,$3D,$5E,$7E,$AF,$4F,$57,$AF,$AB,$DB,$D5,$EF
	.dc.b	$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93,$D5,$F2,$FA,$BE
	.dc.b	$6F,$57,$CF,$EA,$FA,$3D,$5F,$0F,$0B,$E4,$FF,$CC,$87,$F5,$73,$F8
	.dc.b	$89,$F8,$93,$F8,$7B,$E8,$5E,$A7,$25,$EA,$5F,$53,$1E,$A6,$7D,$4D
	.dc.b	$7A,$9B,$F5,$38,$7A,$9C,$7D,$4E,$5E,$A7,$3F,$53,$A7,$A9,$D7,$D4
	.dc.b	$ED,$EA,$77,$F5,$3C,$1E,$A7,$AB,$E3,$7D,$7C,$9F,$EA,$AF,$85,$C4
	.dc.b	$67,$E1,$75,$F0,$3B,$F7,$78,$7B,$1C,$7D,$8E,$5E,$C7,$3F,$63,$A7
	.dc.b	$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1,$7B,$1E,$3F
	.dc.b	$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3E,$8F,$CA,$7F,$AA,$CF
	.dc.b	$F3,$21,$FE,$1E,$3F,$AB,$1F,$89,$3F,$11,$D0,$5E,$C7,$2F,$63,$9F
	.dc.b	$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD
	.dc.b	$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$F3,$FB,$1F,$88,$F9,$5F
	.dc.b	$F5,$51,$FE,$1D,$3F,$AB,$9F,$C3,$CF,$C3,$F1,$9F,$C3,$BF,$A0,$7F
	.dc.b	$0E,$DF,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7,$87,$D6
	.dc.b	$F1,$7A,$DE,$3F,$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5,$BD,$1E
	.dc.b	$B7,$E8,$3E,$5B,$F0,$E7,$FA,$B0,$FF,$32,$1F,$D5,$CF,$E2,$27,$F1
	.dc.b	$0E,$13,$FC,$41,$FD,$CA,$F4,$7A,$7A,$3D,$7D,$1E,$DE,$8F,$7F,$47
	.dc.b	$C1,$E8,$F8,$7D,$1F,$17,$A3,$E3,$F4,$7C,$9E,$8F,$97,$D1,$F3,$7A
	.dc.b	$3E,$7F,$47,$D1,$E8,$FA,$7D,$1F,$DC,$FC,$4F,$FB,$CC,$FF,$56,$1F
	.dc.b	$E6,$63,$FC,$3C,$7F,$11,$3F,$12,$7E,$23,$8F,$F8,$8A,$F4,$39,$FA
	.dc.b	$1D,$3D,$0E,$BE,$87,$6F,$43,$BF,$A1,$E0,$F4,$3C,$3E,$87,$8B,$D0
	.dc.b	$F1,$FA,$1E,$4F,$43,$CB,$E8,$79,$BD,$0F,$3F,$A1,$FC,$43,$E5,$FF
	.dc.b	$DE,$47,$F9,$94,$FF,$0F,$9F,$D5,$CF,$C4,$9F,$88,$E2,$3F,$88,$6F
	.dc.b	$A1,$CF,$D0,$E9,$E8,$75,$F4,$3B,$7A,$1D,$FD,$0F,$07,$A1,$E1,$F4
	.dc.b	$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F,$43,$CD,$E8,$79,$FD,$0F,$E4
	.dc.b	$FF,$31,$FC,$38,$FF,$79,$9F,$EA,$B3,$FC,$3C,$7F,$58,$3F,$10,$7E
	.dc.b	$1F,$84,$FE,$1D,$FE,$82,$BC,$FE,$BE,$7F,$6F,$3F,$BF,$9F,$E0,$F3
	.dc.b	$FC,$3E,$7F,$8B,$CF,$F1,$F9,$FE,$4F,$3F,$CB,$E7,$F9,$BC,$FF,$3F
	.dc.b	$9F,$E8,$F3,$FD,$3E,$7F,$F4,$0F,$99,$7C,$8D,$F8,$3C,$7C,$96,$7E
	.dc.b	$43,$5F,$15,$BF,$87,$E1,$ED,$71,$F5,$B9,$7A,$DC,$FD,$6E,$9E,$B7
	.dc.b	$5F,$5B,$B7,$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$78,$FD,$6F
	.dc.b	$27,$AD,$E5,$F5,$BC,$DE,$B7,$F5,$6E,$47,$CB,$5F,$8C,$C7,$B3,$9F
	.dc.b	$8E,$D7,$C5,$6F,$E2,$38,$7C,$3F,$1F,$73,$97,$B9,$CF,$DC,$E9,$EE
	.dc.b	$75,$F7,$3B,$7B,$9D,$FD,$CF,$07,$B9,$E1,$F7,$3C,$5E,$E7,$8F,$DC
	.dc.b	$F2,$7B,$9E,$5F,$73,$CD,$EE,$7F,$6C,$F9,$AF,$EA,$63,$E2,$31,$F1
	.dc.b	$19,$F8,$8D,$7C,$46,$FE,$23,$87,$C4,$71,$F8,$5E,$5F,$05,$CF,$E0
	.dc.b	$BA,$7C,$17,$5F,$82,$ED,$F0,$5D,$FE,$0B,$C1,$F0,$5E,$1F,$82,$F1
	.dc.b	$7C,$17,$8F,$E0,$BC,$9F,$05,$E5,$F8,$2F,$37,$C1,$79,$FE,$0B,$FC
	.dc.b	$07,$F3,$7F,$84,$F8,$5F,$F0,$AA,$F8,$2C,$FA,$1A,$F8,$2D,$FC,$17
	.dc.b	$0F,$82,$E3,$F0,$5C,$BE,$0B,$9F,$C1,$74,$F8,$2E,$BF,$05,$DB,$E0
	.dc.b	$BB,$FC,$17,$83,$E0,$BC,$3F,$05,$E2,$F8,$2F,$1F,$C1,$79,$3E,$0B
	.dc.b	$CB,$F0,$5E,$6F,$82,$F3,$FC,$17,$F8,$6B,$E7,$3F,$0B,$D1,$FE,$2C
	.dc.b	$7F,$84,$1F,$C1,$C7,$B5,$BF,$6B,$87,$AB,$C7,$D5,$FF,$11,$F9,$EF
	.dc.b	$42,$B8,$F4,$33,$E8,$6B,$CB,$DF,$97,$C3,$CB,$E3,$E5,$F2,$F2,$F9
	.dc.b	$F9,$7D,$3C,$BE,$BE,$5F,$6F,$2F,$BF,$97,$E0,$F2,$FF,$C6,$7F,$3B
	.dc.b	$F4,$79,$CF,$D1,$7F,$C8,$EB,$CA,$DF,$95,$C3,$CA,$E3,$E5,$72,$F2
	.dc.b	$B9,$F9,$5D,$3C,$AE,$BE,$57,$6F,$2B,$BF,$95,$E0,$F2,$BC,$3E,$57
	.dc.b	$8B,$CA,$F1,$F9,$5E,$4F,$2B,$CB,$E5,$79,$BC,$AF,$F2,$4F,$CF,$7F
	.dc.b	$A3,$4F,$EF,$E7,$F3,$73,$F9,$B0,$F6,$F7,$ED,$F0,$F6,$38,$FB,$1F
	.dc.b	$E5,$2F,$9F,$FA,$1D,$5F,$E6,$47,$FD,$90,$7F,$25,$71,$EA,$6F,$D4
	.dc.b	$E1,$E3,$F1,$F1,$F9,$78,$FC,$FC,$7E,$9E,$3F,$5F,$1F,$B7,$8F,$DF
	.dc.b	$C7,$F0,$78,$FE,$1F,$1F,$C5,$E3,$FF,$99,$3E,$0F,$E9,$1F,$47,$80
	.dc.b	$FD,$17,$FC,$F2,$BC,$6E,$1E,$37,$1F,$1B,$97,$8D,$CF,$C6,$E9,$E3
	.dc.b	$75,$F1,$BB,$78,$DD,$FC,$6F,$07,$8D,$E1,$F1,$BC,$5E,$37,$8F,$C6
	.dc.b	$F2,$78,$DE,$5F,$1B,$CD,$E3,$79,$FD,$3F,$47,$A7,$FE,$7C,$FA,$0F
	.dc.b	$D1,$F3,$17,$C3,$67,$E1,$B5,$D1,$BF,$86,$E1,$F0,$DC,$7E,$1B,$97
	.dc.b	$C3,$73,$F8,$6E,$9E,$D7,$5F,$6B,$B7,$C3,$77,$F4,$BC,$1E,$97,$87
	.dc.b	$E1,$BC,$5F,$0D,$E3,$F7,$7C,$9E,$EF,$97,$E1,$BC,$DE,$BF,$9F,$D7
	.dc.b	$FF,$49,$FD,$0F,$F9,$E8,$FE,$51,$D1,$FE,$A6,$7F,$0A,$97,$BD,$C3
	.dc.b	$DE,$E3,$EF,$72,$F8,$5E,$7F,$0B,$D3,$E1,$7A,$FB,$BD,$BD,$DE,$FE
	.dc.b	$EF,$83,$DD,$F0,$FB,$BE,$2F,$77,$C7,$EE,$F9,$3D,$DF,$2F,$BB,$E6
	.dc.b	$F7,$7C,$FE,$EF,$A3,$DD,$F4,$FB,$BF,$EA,$CF,$A2,$FD,$D1,$F8,$93
	.dc.b	$F1,$1C,$E7,$F1,$0F,$FB,$09,$FF,$63,$1F,$E7,$E3,$FC,$6C,$EF,$DA
	.dc.b	$E9,$ED,$75,$F6,$BB,$7B,$5D,$FD,$AF,$07,$B5,$E1,$F6,$BC,$5E,$D7
	.dc.b	$8F,$DA,$F2,$7B,$5E,$5F,$6B,$CD,$ED,$79,$FD,$AF,$47,$B5,$FE,$C4
	.dc.b	$F7,$BF,$12,$7E,$23,$87,$FD,$AE,$BD,$9D,$7B,$3B,$F6,$78,$7B,$3C
	.dc.b	$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7,$D9,$ED,$EC,$F7,$F6,$7C,$1E
	.dc.b	$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9,$EC,$F9,$7D,$9F,$37,$B3,$E7
	.dc.b	$F6,$7F,$DB,$7F,$47,$FE,$C2,$3F,$13,$CE,$7F,$12,$7E,$21,$FF,$78
	.dc.b	$AF,$63,$8F,$B1,$CB,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8E,$FE
	.dc.b	$C7,$83,$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79,$3D,$8F,$2F,$B1,$E6
	.dc.b	$F6,$3C,$FE,$C7,$A3,$D8,$F4,$FB,$1F,$EF,$4F,$A4,$FE,$84,$7F,$11
	.dc.b	$3F,$12,$7E,$23,$98,$BE,$03,$87,$C0,$71,$F8,$0E,$5E,$DF,$3F,$6F
	.dc.b	$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6,$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E
	.dc.b	$3F,$6F,$C9,$ED,$F9,$7D,$BF,$37,$B7,$E7,$F6,$DC,$93,$E9,$7F,$88
	.dc.b	$9F,$89,$4F,$E2,$5D,$7B,$5C,$1C,$FB,$56,$7F,$D0,$07,$F9,$FC,$7B
	.dc.b	$19,$F6,$35,$EC,$6F,$D8,$E1,$EC,$71,$F6,$39,$7B,$1C,$FD,$8E,$9E
	.dc.b	$C7,$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C9,$EB,$F9,$78
	.dc.b	$FF,$0D,$F1,$58,$F8,$7C,$FC,$06,$AB,$7E,$D7,$0F,$6B,$8F,$B5,$CB
	.dc.b	$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB
	.dc.b	$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$CB
	.dc.b	$FA,$6F,$F0,$51,$FD,$64,$FE,$22,$7E,$24,$FC,$47,$47,$D8,$5E,$CF
	.dc.b	$1F,$67,$97,$B3,$CF,$D9,$E9,$EC,$F5,$F6,$7B,$7B,$3D,$FD,$9F,$07
	.dc.b	$B3,$E1,$F6,$7C,$5E,$CF,$8F,$D9,$F2,$7B,$3E,$5F,$67,$CD,$EC,$F9
	.dc.b	$FD,$9F,$C1,$7D,$3F,$F5,$83,$F1,$07,$E1,$F9,$0F,$F7,$27,$F1,$EB
	.dc.b	$D9,$E3,$EF,$F2,$F7,$F9,$FB,$5D,$3D,$AE,$BE,$D7,$6F,$6B,$BF,$B5
	.dc.b	$E0,$F6,$BC,$3E,$D7,$8B,$DA,$F1,$FB,$5E,$4F,$6B,$CB,$ED,$79,$BD
	.dc.b	$AF,$3F,$B5,$E8,$F6,$BF,$35,$F5,$1F,$87,$3F,$88,$9F,$D6,$0F,$E2
	.dc.b	$1C,$A7,$F8,$83,$FA,$85,$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF
	.dc.b	$6F,$5F,$BF,$AF,$E0,$F5,$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F
	.dc.b	$CB,$EB,$F9,$BD,$7F,$3F,$AF,$E8,$F5,$FD,$3E,$BF,$EC,$FC,$5F,$F2
	.dc.b	$91,$FE,$21,$3F,$AC,$9F,$C4,$4F,$C4,$F2,$FD,$1F,$5B,$87,$AD,$C7
	.dc.b	$D6,$E5,$EB,$73,$F5,$BA,$7A,$DD,$7D,$6E,$DE,$B7,$7F,$5B,$C1,$EB
	.dc.b	$78,$7D,$6F,$17,$AD,$E3,$F5,$BC,$9E,$B7,$97,$D6,$F3,$7A,$DF,$C2
	.dc.b	$7E,$A7,$FC,$3E,$7F,$0F,$3F,$10,$7E,$1F,$8C,$FE,$1D,$FE,$2E,$BD
	.dc.b	$6E,$5E,$B7,$3F,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7
	.dc.b	$87,$D6,$F1,$7A,$DE,$3F,$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5
	.dc.b	$BD,$1E,$B7,$F2,$1F,$AA,$FE,$1C,$7F,$58,$3F,$C3,$E7,$E2,$0F,$C3
	.dc.b	$F2,$1F,$C3,$BF,$CC,$17,$AB,$CF,$D5,$E9,$EA,$F5,$F5,$7B,$7A,$BD
	.dc.b	$FD,$5F,$07,$AB,$E1,$F5,$7C,$5E,$AF,$8F,$D5,$F2,$7A,$BE,$5F,$57
	.dc.b	$CD,$EA,$F9,$FD,$5F,$47,$AB,$E9,$F5,$7F,$9D,$7D,$5F,$FA,$64,$FF
	.dc.b	$34,$1F,$E2,$03,$FA,$C1,$F8,$83,$F8,$7F,$27,$F4,$87,$FC,$3E,$3D
	.dc.b	$3E,$7E,$9F,$4F,$4F,$AF,$A7,$DB,$D3,$EF,$E9,$F8,$3D,$3F,$0F,$A7
	.dc.b	$E2,$F4,$FC,$7E,$9F,$93,$D3,$F2,$FA,$7E,$6F,$4F,$CF,$E9,$FF,$51
	.dc.b	$F8,$1F,$F3,$39,$FD,$5C,$FE,$20,$7E,$24,$FC,$47,$17,$F6,$FF,$4F
	.dc.b	$87,$A7,$C7,$D3,$E5,$E9,$F3,$F4,$FA,$7A,$7D,$7D,$3E,$DE,$9F,$7F
	.dc.b	$4F,$C1,$E9,$F8,$7D,$3F,$17,$A7,$E3,$F4,$FC,$9E,$9F,$97,$D3,$F3
	.dc.b	$7A,$7F,$D9,$FE,$B3,$FA,$A1,$FE,$1E,$3F,$CC,$A7,$F0,$E3,$F0,$E7
	.dc.b	$E1,$B8,$CF,$E1,$9F,$EE,$EB,$D2,$E9,$E9,$75,$F4,$BB,$7A,$5D,$FD
	.dc.b	$2F,$07,$A5,$E1,$F4,$BC,$5E,$97,$8F,$D2,$F2,$7A,$5E,$5F,$4B,$CD
	.dc.b	$E9,$79,$FD,$2F,$47,$A5,$E9,$F4,$BF,$C0,$1F,$5B,$FE,$DF,$3F,$D5
	.dc.b	$C7,$F9,$98,$FF,$0F,$1F,$C3,$8F,$C3,$F1,$9F,$F3,$29,$F8,$77,$FC
	.dc.b	$22,$BD,$1E,$BE,$8F,$6F,$47,$BF,$A3,$E0,$F4,$7C,$3E,$8F,$8B,$D1
	.dc.b	$F1,$FA,$3E,$4F,$47,$CB,$E8,$F9,$BD,$1F,$3F,$A3,$E8,$F4,$7F,$C3
	.dc.b	$1F,$5D,$FE,$AE,$3F,$C3,$C7,$F5,$73,$F8,$81,$F8,$83,$F8,$77,$0F
	.dc.b	$F8,$99,$7A,$1C,$BD,$0E,$7E,$87,$4F,$43,$AF,$A1,$DB,$D0,$EF,$E8
	.dc.b	$78,$3D,$0F,$0F,$A1,$E2,$F4,$3C,$7E,$87,$93,$D0,$F2,$FA,$1E,$6F
	.dc.b	$43,$CF,$E8,$7F,$8B,$FD,$BF,$F0,$E1,$FE,$66,$3F,$D5,$A7,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$7F,$D5,$BD,$0E,$5E,$87,$3F,$43,$A7,$A1
	.dc.b	$D7,$D0,$ED,$E8,$77,$F4,$3C,$1E,$87,$87,$D0,$F1,$7A,$1E,$3F,$43
	.dc.b	$C9,$E8,$79,$7D,$0F,$37,$A1,$E7,$F4,$3F,$C8,$DF,$5E,$F9,$7B,$F2
	.dc.b	$98,$F9,$2C,$FC,$66,$BD,$8D,$FB,$FC,$3D,$FE,$3E,$DF,$2F,$63,$9F
	.dc.b	$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD
	.dc.b	$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$2E,$7D,$87,$FB,$C8
	.dc.b	$FF,$31,$1F,$E1,$B3,$FA,$A1,$FC,$30,$FC,$2F,$90,$7F,$0A,$FF,$9B
	.dc.b	$1F,$F3,$0B,$7D,$6E,$BE,$B7,$6F,$5B,$BF,$AD,$E0,$F5,$BC,$3E,$B7
	.dc.b	$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79,$BD,$6F,$3F,$AD,$E8,$F5
	.dc.b	$BF,$CE,$9F,$63,$FE,$61,$3F,$D5,$67,$FB,$C8,$FF,$0D,$9F,$D5,$0F
	.dc.b	$E1,$C7,$E1,$8F,$E1,$9C,$1F,$E8,$45,$EA,$F4,$F5,$7A,$FA,$BD,$BD
	.dc.b	$5E,$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F,$57,$C7,$EA,$F9,$3D,$5F,$2F
	.dc.b	$AB,$E6,$F5,$7C,$FE,$AF,$A3,$D5,$FF,$46,$7B,$BF,$87,$E2,$FF,$4A
	.dc.b	$AF,$53,$3E,$A6,$BD,$4D,$FA,$9C,$3D,$4E,$3E,$A7,$2F,$53,$9F,$A9
	.dc.b	$D3,$D4,$EB,$EA,$76,$F5,$3B,$FA,$9E,$0F,$53,$C3,$EA,$78,$BD,$4F
	.dc.b	$1F,$A9,$E4,$F5,$3C,$BE,$A7,$9B,$D4,$F3,$FA,$9F,$EA,$3F,$B2,$7C
	.dc.b	$D5,$F9,$3C,$7C,$8E,$7E,$3F,$5F,$0F,$BE,$CE,$1F,$01,$C7,$E0,$39
	.dc.b	$7B,$7C,$FD,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B,$7E,$1F
	.dc.b	$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE,$DF,$FA,$E7,$EC
	.dc.b	$FF,$D4,$C7,$FB,$C0,$7C,$76,$7E,$3B,$5F,$1D,$BF,$84,$E1,$F1,$3C
	.dc.b	$7E,$17,$97,$C0,$F3,$F8,$1E,$9F,$03,$D7,$E0,$7B,$7C,$0F,$7F,$81
	.dc.b	$F0,$7C,$0F,$87,$E0,$7C,$5F,$03,$E3,$F8,$1F,$27,$C0,$F9,$7E,$07
	.dc.b	$CD,$F0,$3E,$7F,$81,$FF,$68,$7D,$A7,$E1,$7E,$08,$FE,$15,$FF,$6D
	.dc.b	$AF,$81,$DF,$C0,$F0,$F8,$1E,$3F,$03,$CB,$E0,$79,$FC,$0F,$4F,$81
	.dc.b	$EB,$F0,$3D,$BE,$07,$BF,$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1
	.dc.b	$FC,$0F,$93,$E0,$7C,$BF,$03,$E6,$F8,$1F,$3F,$C0,$FA,$3E,$07,$FD
	.dc.b	$D1,$C3,$F8,$6E,$6F,$F7,$8B,$FE,$EF,$3F,$D5,$07,$F9,$84,$FF,$0C
	.dc.b	$9F,$C3,$8F,$E1,$AF,$FB,$03,$D9,$E7,$EC,$F4,$F6,$7A,$FB,$3D,$BD
	.dc.b	$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F,$67,$C7,$EC,$F9,$3D,$9F,$2F
	.dc.b	$B3,$E6,$F6,$7C,$FE,$CF,$FB,$FF,$ED,$7F,$DD,$83,$E4,$71,$ED,$67
	.dc.b	$DA,$D7,$B5,$BF,$6B,$87,$B5,$C7,$DA,$E5,$ED,$73,$F6,$BA,$7B,$5D
	.dc.b	$7D,$AE,$DE,$D7,$7F,$6B,$C1,$ED,$78,$7D,$AF,$17,$B5,$E3,$F6,$BC
	.dc.b	$9E,$D7,$97,$DA,$F3,$7B,$5E,$7F,$69,$E3,$4F,$B6,$FC,$41,$F8,$74
	.dc.b	$FE,$1D,$D7,$B3,$C1,$CF,$B3,$67,$FC,$36,$3D,$9C,$7B,$39,$F6,$75
	.dc.b	$EC,$EF,$D9,$E1,$EC,$F1,$F6,$79,$7B,$3C,$FD,$9E,$9E,$CF,$5F,$67
	.dc.b	$B7,$B3,$DF,$D9,$F0,$7B,$3E,$1F,$67,$C9,$F6,$FF,$89,$E4,$2F,$83
	.dc.b	$CF,$BF,$AF,$6F,$7E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6
	.dc.b	$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F8,$9E,$97,$F0,$6E,$1F,$9A
	.dc.b	$FF,$B9,$8F,$F4,$F9,$FE,$59,$3F,$C2,$27,$F4,$C7,$FD,$3F,$E8,$F2
	.dc.b	$F4,$79,$FA,$3D,$3D,$1E,$BE,$8F,$6F,$47,$BF,$A3,$E0,$F4,$7C,$3E
	.dc.b	$8F,$8B,$D1,$F1,$FA,$3E,$4F,$47,$CB,$E8,$F9,$BD,$1F,$B5,$F7,$1F
	.dc.b	$C1,$CF,$C0,$1F,$7F,$84,$BD,$2D,$FA,$5C,$3D,$DE,$3E,$97,$2F,$4B
	.dc.b	$9F,$A5,$D3,$D2,$EB,$E9,$76,$F4,$BB,$FA,$5E,$0F,$4B,$C3,$E9,$78
	.dc.b	$BD,$2F,$1F,$A5,$E4,$F4,$BC,$BE,$97,$9B,$D2,$F3,$FA,$5F,$91,$FB
	.dc.b	$9F,$BC,$7D,$DE,$6F,$CD,$AF,$85,$D7,$C2,$EF,$E1,$78,$7C,$2F,$1F
	.dc.b	$57,$97,$AB,$CF,$E0,$BA,$7A,$CF,$E0,$18,$5D,$3F,$40,$3E,$CE,$B8
	.dc.b	$6B,$FA,$56,$BF,$AE,$6B,$C4,$CB,$82,$38,$21,$D1,$CB,$E0,$F8,$1F
	.dc.b	$E7,$CF,$91,$AF,$ED,$5A,$FE,$F9,$AF,$8C,$7D,$77,$F9,$8B,$EE,$3F
	.dc.b	$AE,$FC,$7F,$F0,$48,$FF,$32,$CE,$DD,$7C,$27,$F8,$45,$CC,$7A,$DF
	.dc.b	$F3,$5F,$DC,$FF,$A0,$AB,$97,$AB,$F1,$B2,$B7,$FC,$42,$B4,$F8,$A7
	.dc.b	$E6,$E1,$E6,$25,$FC,$A7,$D3,$3D,$E7,$F8,$96,$3E,$88,$E4,$7D,$32
	.dc.b	$87,$58,$FA,$31,$F1,$CB,$1F,$1C,$E6,$38,$C6,$2F,$F0,$3F,$0B,$F2
	.dc.b	$89,$97,$DE,$7C,$B8,$7F,$2C,$F8,$0F,$8F,$36,$BE,$16,$1F,$C4,$DF
	.dc.b	$4E,$7F,$E3,$BA,$F0,$1F,$2B,$87,$13,$C2,$31,$7C,$EF,$F9,$97,$C2
	.dc.b	$FD,$C2,$47,$E4,$1F,$1C,$7C,$2D,$7E,$0D,$F8,$B5,$F2,$CF,$C6,$3E
	.dc.b	$F4,$7F,$3D,$D7,$E0,$47,$CB,$9F,$0B,$BF,$85,$FC,$59,$31,$FC,$FB
	.dc.b	$E0,$9E,$A7,$CB,$2D,$F8,$4E,$D9,$FE,$2D,$C5,$FE,$16,$30,$FF,$2C
	.dc.b	$BF,$D5,$CF,$E6,$88,$FE,$A2,$BA,$CF,$9E,$7E,$6C,$FD,$E8,$F6,$9D
	.dc.b	$5C,$3F,$8B,$30,$7C,$3F,$85,$FD,$39,$23,$F7,$6F,$B6,$7A,$CF,$C9
	.dc.b	$39,$1F,$8F,$D7,$D2,$FF,$03,$3F,$17,$86,$1F,$19,$F9,$17,$1E,$D7
	.dc.b	$ED,$89,$87,$F0,$A7,$C1,$3F,$DE,$34,$FB,$6F,$BC,$C5,$F5,$7C,$7A
	.dc.b	$7D,$A7,$C5,$9F,$DA,$3F,$7B,$ED,$7F,$04,$26,$3F,$C6,$1E,$F3,$EB
	.dc.b	$8F,$8F,$C1,$FA,$8C,$7C,$33,$D3,$4F,$B7,$5A,$8F,$7B,$F4,$7D,$AF
	.dc.b	$E2,$24,$7F,$92,$BD,$AB,$9C,$FC,$46,$BE,$AD,$7D,$98,$7E,$E3,$3F
	.dc.b	$64,$BD,$AF,$E3,$A4,$C7,$F9,$8B,$D7,$5E,$33,$FA,$36,$EB,$F0,$F4
	.dc.b	$FC,$8C,$FD,$3B,$FC,$47,$B5,$FC,$A0,$8F,$F3,$C7,$A3,$EF,$AF,$C8
	.dc.b	$CF,$C3,$3F,$16,$FE,$86,$BE,$2C,$7E,$BF,$DA,$FE,$66,$4A,$FE,$8E
	.dc.b	$F4,$8F,$C6,$67,$F3,$F1,$F3,$DF,$2C,$3A,$57,$68,$FE,$AB,$ED,$7F
	.dc.b	$3E,$26,$3F,$D4,$9E,$95,$C2,$BE,$E1,$D8,$FA,$E5,$7E,$23,$FE,$3F
	.dc.b	$F6,$BF,$A4,$92,$BF,$AE,$BD,$2F,$E4,$EF,$F6,$32,$FA,$DB,$FA,$30
	.dc.b	$F3,$1F,$E4,$9F,$6B,$FA,$99,$31,$FE,$CE,$F3,$8F,$CE,$1E,$6C,$BD
	.dc.b	$1C,$1F,$80,$8F,$C2,$57,$EF,$7D,$AF,$EB,$A4,$7F,$B8,$BC,$E3,$EA
	.dc.b	$1C,$F3,$FB,$F5,$CF,$2F,$3D,$FF,$07,$F6,$BF,$B3,$91,$FE,$F6,$F3
	.dc.b	$BF,$CB,$6F,$ED,$7F,$10,$7C,$9C,$3E,$B1,$FE,$08,$F6,$BF,$B8,$92
	.dc.b	$FF,$C0,$AF,$39,$FE,$13,$B7,$C9,$7E,$9E,$9F,$03,$6F,$E1,$1E,$D7
	.dc.b	$F7,$A2,$61,$FE,$0E,$79,$69,$F8,$78,$FC,$A6,$5E,$2C,$3E,$5E,$BF
	.dc.b	$A8,$F6,$BF,$C0,$64,$AF,$F0,$BB,$CA,$7F,$4D,$1C,$9F,$B2,$7C,$A9
	.dc.b	$FE,$19,$F7,$49,$FF,$07,$B5,$FF,$11,$B6,$7F,$C4,$ED,$7F,$C3,$87
	.dc.b	$63,$E4,$1F,$C6,$A3,$C0,$FC,$CC,$FE,$14,$7F,$78,$97,$C7,$F8,$09
	.dc.b	$1E,$07,$D5,$3F,$B4,$AF,$E5,$F7,$F7,$DF,$9B,$8F,$FA,$6F,$0F,$F7
	.dc.b	$26,$1E,$8F,$8B,$F5,$91,$FE,$9D,$3F,$D0,$30,$FD,$56,$DF,$5D,$FD
	.dc.b	$3E,$1F,$17,$6F,$E8,$38,$FF,$9A,$70,$F8,$7E,$E7,$C6,$23,$D2,$BE
	.dc.b	$01,$F0,$8F,$CB,$1F,$62,$BF,$1D,$9F,$D0,$F7,$3E,$79,$2B,$E7,$BF
	.dc.b	$03,$3F,$31,$F9,$EA,$E2,$79,$3D,$CF,$AE,$4A,$FA,$EF,$1C,$3C,$79
	.dc.b	$FD,$07,$C9,$8F,$E0,$3D,$CF,$BE,$4B,$FB,$EF,$19,$F4,$9F,$0E,$3E
	.dc.b	$DC,$FE,$23,$DC,$FC,$39,$23,$F0,$EF,$1D,$BE,$66,$7F,$99,$BE,$47
	.dc.b	$F8,$F3,$67,$B9,$F9,$02,$61,$F9,$07,$97,$37,$C0,$8F,$0E,$DF,$53
	.dc.b	$DC,$FC,$B9,$23,$F2,$EF,$07,$64,$C3,$C1,$EE,$7E,$78,$91,$F9,$E7
	.dc.b	$84,$7C,$6A,$F0,$3E,$7F,$73,$F4,$A4,$AF,$D2,$BC,$38,$3E,$71,$34
	.dc.b	$FC,$A7,$B9,$FA,$C2,$57,$EB,$1E,$11,$F7,$63,$CD,$28,$FC,$E7,$B9
	.dc.b	$FB,$42,$4F,$ED,$1E,$38,$7E,$36,$38,$CF,$E8,$FD,$CF,$DD,$92,$3F
	.dc.b	$76,$F0,$9F,$A7,$7E,$74,$37,$7B,$9F,$C1,$09,$7F,$C1,$1E,$53,$FB
	.dc.b	$4D,$3E,$3D,$FA,$63,$9D,$BD,$CF,$E1,$A4,$C3,$F8,$6B,$CA,$7F,$A1
	.dc.b	$9F,$46,$3F,$83,$7B,$9F,$C5,$08,$FF,$14,$78,$76,$FF,$40,$9E,$07
	.dc.b	$E3,$7D,$CF,$E3,$64,$AF,$E3,$6F,$08,$FE,$12,$7C,$93,$F2,$9E,$E7
	.dc.b	$F2,$22,$4F,$F2,$27,$86,$5F,$C1,$A7,$E0,$4F,$F1,$BF,$73,$F9,$49
	.dc.b	$1F,$E5,$2F,$0D,$FF,$45,$3C,$0F,$F4,$AF,$C7,$7F,$AF,$BA,$8D,$25
	.dc.b	$FF,$0D,$9F,$D9,$9F,$D6,$1F,$BE,$FA,$70,$FF,$5C,$FA,$5C,$C9,$3F
	.dc.b	$ED,$CB,$47,$8E,$BF,$A1,$9E,$3A,$FF,$60,$FB,$9E,$32,$5F,$17,$84
	.dc.b	$FF,$3B,$C7,$F3,$0F,$FA,$BB,$DC,$F5,$12,$BC,$0F,$0C,$7F,$9C,$61
	.dc.b	$FC,$B9,$7B,$9F,$04,$95,$E7,$78,$71,$FE,$33,$1F,$F4,$E7,$B9,$F1
	.dc.b	$C9,$5E,$D7,$F3,$97,$FE,$D4,$3F,$AE,$65,$FE,$84,$7F,$DD,$BC,$3F
	.dc.b	$87,$73,$1F,$E0,$03,$F4,$27,$FA,$28,$A6,$4F,$F2,$09,$FD,$83,$2F
	.dc.b	$EA,$8F,$F8,$03,$B3,$FD,$0F,$3F,$EE,$48,$E2,$D7,$89,$0F,$F7,$57
	.dc.b	$C3,$7E,$04,$98,$FD,$77,$F7,$BF,$74,$F9,$BF,$EA,$55,$FE,$DF,$AF
	.dc.b	$DB,$CE,$7C,$5F,$7E,$1F,$02,$38,$78,$D8,$AD,$E0,$BC,$61,$FE,$FB
	.dc.b	$F8,$8F,$C9,$12,$BF,$10,$F8,$AB,$C3,$8F,$57,$94,$FF,$08,$8C,$61
	.dc.b	$D3,$1F,$24,$E4,$FF,$6A,$7C,$47,$E7,$C8,$FE,$55,$ED,$E4,$71,$C0
	.dc.b	$DF,$23,$0F,$A3,$3F,$F8,$01,$F1,$1F,$A9,$26,$1F,$9F,$79,$BE,$18
	.dc.b	$63,$F2,$EB,$A4,$E6,$30,$EE,$59,$BE,$23,$F6,$C4,$9F,$D6,$3F,$17
	.dc.b	$CA,$F4,$79,$3E,$96,$CF,$3A,$7C,$47,$EF,$C9,$1F,$B5,$79,$3E,$3D
	.dc.b	$7F,$B2,$B6,$B9,$E3,$F8,$13,$E8,$7C,$47,$F0,$92,$4F,$EF,$DF,$B1
	.dc.b	$E5,$FC,$BF,$2F,$E6,$36,$7C,$9F,$11,$FC,$44,$8F,$F0,$77,$8B,$E3
	.dc.b	$9E,$2D,$3D,$9F,$1D,$FC,$64,$8F,$F0,$F7,$F1,$1E,$6F,$DB,$3E,$E7
	.dc.b	$9B,$67,$E2,$7C,$77,$F2,$12,$3F,$C5,$DE,$2F,$C3,$9E,$2D,$3E,$B7
	.dc.b	$9E,$F8,$4F,$DD,$EB,$FB,$A8,$C4,$FE,$EA,$3F,$DE,$13,$A6,$C6,$5F
	.dc.b	$F5,$5F,$0F,$F0,$4A,$76,$FE,$3E,$7F,$98,$1F,$44,$DD,$77,$8F,$CA
	.dc.b	$4F,$F6,$A8,$F2,$C6,$0F,$04,$FF,$AE,$0F,$E1,$B3,$DA,$7C,$4D,$8F
	.dc.b	$43,$5D,$CF,$28,$C4,$78,$67,$8C,$E4,$8E,$0E,$BF,$03,$C4,$8E,$0F
	.dc.b	$8C,$3F,$7D,$67,$FC,$1A,$1F,$BF,$7B,$70,$FC,$94,$6E,$75,$39,$CB
	.dc.b	$F8,$37,$A7,$FB,$BF,$4F,$F7,$8D,$BF,$B3,$F0,$7E,$B2,$3F,$09,$FA
	.dc.b	$CF,$53,$F7,$47,$8B,$1F,$26,$5E,$68,$CE,$3E,$17,$F8,$FD,$F7,$AF
	.dc.b	$D4,$75,$FF,$2B,$8C,$65,$F7,$FE,$0F,$E6,$12,$FE,$9B,$F3,$FE,$0F
	.dc.b	$EE,$9F,$CA,$38,$5F,$99,$3A,$73,$7F,$94,$AB,$9F,$ED,$F1,$7E,$A6
	.dc.b	$F7,$B3,$1F,$8F,$AE,$D7,$F8,$26,$DF,$E0,$9A,$FE,$21,$AF,$E7,$DA
	.dc.b	$FF,$8E,$B9,$51,$88,$43,$E1,$7F,$14,$4A,$FC,$8B,$E0,$BF,$E9,$7F
	.dc.b	$BA,$BE,$0F,$CF,$5F,$CF,$E3,$0F,$BC,$FE,$5D,$3E,$EF,$81,$DE,$3F
	.dc.b	$CA,$B4,$47,$E9,$BE,$1B,$F3,$44,$BF,$D0,$BE,$17,$B6,$3C,$B0,$63
	.dc.b	$E0,$C7,$F6,$59,$FB,$AF,$C3,$97,$67,$C9,$3F,$46,$30,$F0,$F8,$6F
	.dc.b	$D5,$12,$7F,$62,$FE,$B7,$1F,$DD,$9F,$7F,$C4,$76,$E2,$BF,$88,$CD
	.dc.b	$BF,$BD,$FA,$B3,$5C,$6B,$FC,$2F,$E1,$FF,$70,$4C,$3F,$7E,$F7,$79
	.dc.b	$9F,$D5,$F9,$7F,$A3,$E5,$CA,$3F,$8E,$65,$72,$FE,$AB,$97,$79,$E3
	.dc.b	$1F,$12,$7A,$3F,$C2,$9A,$FF,$10,$6B,$D2,$3F,$88,$C6,$23,$F1,$4D
	.dc.b	$7F,$12,$FF,$11,$F8,$79,$3F,$C6,$1B,$3D,$E3,$D5,$CC,$F0,$BD,$19
	.dc.b	$7E,$A1,$F5,$68,$C9,$BE,$1B,$87,$F3,$EF,$87,$FE,$40,$47,$F9,$3B
	.dc.b	$EE,$AF,$F1,$FE,$1D,$67,$E1,$C6,$31,$F1,$46,$0F,$F0,$C5,$F6,$CB
	.dc.b	$E4,$47,$E6,$A3,$15,$FD,$0F,$9F,$F9,$81,$27,$F9,$BB,$E1,$7F,$D5
	.dc.b	$38,$78,$6F,$99,$FD,$DD,$C5,$7F,$4D,$D7,$FC,$62,$FC,$1C,$FE,$01
	.dc.b	$FD,$2B,$97,$F9,$F1,$30,$FE,$8A,$DC,$3F,$A2,$B5,$FE,$96,$D7,$FB
	.dc.b	$13,$5F,$E0,$8F,$F4,$9F,$87,$93,$FD,$41,$B3,$FD,$41,$AF,$F5,$66
	.dc.b	$BF,$DB,$5A,$FF,$0F,$6B,$FC,$3B,$E1,$FF,$AB,$12,$7F,$AE,$B6,$3F
	.dc.b	$AE,$B5,$FE,$C6,$D7,$FB,$D3,$5F,$E3,$8F,$F6,$AF,$79,$34,$87,$F7
	.dc.b	$B7,$2D,$33,$3F,$E0,$38,$CB,$63,$2F,$FA,$02,$32,$D8,$CB,$FE,$86
	.dc.b	$87,$C3,$F8,$C9,$3E,$16,$CF,$76,$BD,$5A,$F3,$6B,$FE,$F2,$9E,$B3
	.dc.b	$EB,$F8,$7F,$80,$47,$04,$9F,$64,$65,$B1,$97,$D9,$19,$6C,$7C,$7F
	.dc.b	$53,$E6,$11,$C1,$23,$E4,$46,$5B,$19,$7E,$44,$65,$B1,$F4,$FD,$6F
	.dc.b	$B0,$47,$04,$8F,$A9,$19,$6C,$65,$FA,$91,$96,$C7,$DF,$F4,$BF,$08
	.dc.b	$47,$04,$8F,$C0,$46,$5B,$19,$7F,$01,$19,$6C,$7E,$2F,$D4,$FC,$81
	.dc.b	$1C,$1F,$8F,$F8,$2F,$F4,$A6,$47,$FC,$47,$C5,$87,$FD,$31,$F9,$EB
	.dc.b	$CE,$C5,$FF,$A3,$E3,$0F,$FA,$46,$F9,$37,$D2,$FD,$01,$1F,$CF,$3F
	.dc.b	$A3,$BE,$79,$7F,$8D,$3C,$9F,$F3,$0B,$FE,$57,$BE,$6D,$9D,$F2,$FF
	.dc.b	$B8,$B1,$F0,$CE,$71,$F1,$7A,$9F,$AF,$23,$FA,$D7,$C0,$F9,$EF,$DE
	.dc.b	$D2,$4F,$33,$FC,$D1,$5D,$AF,$FD,$E3,$87,$91,$BE,$9F,$EF,$49,$7F
	.dc.b	$BB,$7B,$DF,$39,$FF,$7E,$CF,$F9,$53,$5F,$5B,$FB,$5C,$6D,$06,$3E
	.dc.b	$19,$D1,$F3,$FD,$2F,$E1,$84,$7F,$85,$3E,$0C,$3C,$F7,$EE,$BF,$15
	.dc.b	$72,$69,$DE,$FB,$3F,$BB,$9F,$C2,$F8,$1F,$C6,$49,$5F,$C5,$DE,$CF
	.dc.b	$3B,$F6,$EB,$F6,$9F,$0C,$7C,$C6,$DF,$CC,$DF,$DE,$B5,$FB,$6F,$F1
	.dc.b	$B8,$FE,$50,$4A,$FC,$C1,$CC,$FF,$4C,$5F,$77,$C4,$49,$F5,$3F,$97
	.dc.b	$9E,$F4,$F2,$FF,$9A,$12,$7F,$99,$3E,$CA,$E2,$7E,$57,$5E,$39,$FA
	.dc.b	$34,$F8,$F3,$E9,$8F,$D5,$F9,$FF,$CF,$C9,$87,$F3,$D7,$A4,$EE,$3F
	.dc.b	$D0,$37,$D1,$F6,$47,$D3,$CB,$CF,$F1,$1F,$D2,$49,$3F,$D1,$DE,$79
	.dc.b	$FE,$4E,$3E,$1D,$F3,$BF,$BA,$7E,$2A,$FE,$0F,$C3,$FF,$53,$25,$7F
	.dc.b	$51,$79,$E5,$F8,$7D,$BE,$B9,$27,$E6,$5F,$37,$CE,$FD,$9F,$87,$FE
	.dc.b	$BC,$48,$FE,$B8,$F4,$7D,$E7,$0C,$BD,$13,$F3,$AB,$F7,$98,$FD,$6F
	.dc.b	$87,$FE,$D0,$4B,$FE,$CC,$F3,$AF,$A9,$1C,$F1,$FA,$95,$F0,$67,$FB
	.dc.b	$24,$7E,$C9,$B3,$F8,$A7,$FB,$64,$65,$FD,$21,$A3,$31,$B7,$07,$F6
	.dc.b	$FF,$87,$FE,$F8,$48,$FE,$F4,$FE,$25,$FF,$68,$FA,$FF,$BC,$69,$F2
	.dc.b	$3D,$37,$C9,$7E,$89,$B1,$FC,$07,$3F,$F0,$3E,$1F,$C0,$E7,$F4,$D3
	.dc.b	$FF,$00,$5B,$1F,$C1,$C8,$E1,$C0,$70,$FE,$96,$3C,$6F,$F0,$26,$CF
	.dc.b	$DF,$8F,$05,$79,$4F,$1C,$79,$23,$A7,$C4,$57,$C2,$D7,$D0,$D7,$75
	.dc.b	$F4,$67,$F8,$3E,$1F,$17,$94,$F7,$9F,$8D,$87,$81,$F6,$FC,$47,$3C
	.dc.b	$8D,$24,$75,$3F,$82,$D7,$53,$A6,$3F,$A4,$AF,$9E,$B1,$F6,$43,$E2
	.dc.b	$3B,$12,$FE,$83,$F7,$72,$F3,$C7,$EE,$0F,$92,$7E,$9A,$7A,$D7,$CB
	.dc.b	$BF,$A5,$97,$AD,$FB,$9F,$11,$EB,$25,$FD,$D7,$B6,$5E,$98,$F8,$64
	.dc.b	$9F,$43,$F8,$0F,$86,$F8,$84,$7F,$0C,$F2,$7C,$D9,$7A,$27,$89,$F7
	.dc.b	$A3,$F2,$07,$E2,$B0,$F8,$F8,$FC,$4F,$88,$FA,$04,$9F,$C9,$3D,$6B
	.dc.b	$F1,$57,$D6,$BE,$5D,$FB,$E7,$E0,$DF,$A6,$7E,$7F,$C8,$7D,$B2,$47
	.dc.b	$E7,$5E,$91,$F9,$38,$F9,$B3,$E7,$3D,$27,$43,$EE,$57,$7D,$1F,$21
	.dc.b	$F8,$72,$57,$EA,$5E,$AF,$E8,$F1,$F4,$3F,$04,$F8,$77,$DF,$0F,$B5
	.dc.b	$7F,$72,$7F,$5F,$F2,$1F,$93,$23,$FB,$57,$AD,$7F,$1D,$8F,$C9,$9D
	.dc.b	$1F,$45,$7C,$68,$F8,$87,$86,$3F,$72,$7B,$4F,$DD,$D7,$E6,$1B,$8F
	.dc.b	$67,$FA,$61,$FE,$2B,$D7,$E1,$47,$E9,$C9,$5F,$C2,$5F,$D5,$7C,$44
	.dc.b	$8F,$E1,$8D,$AF,$E4,$27,$38,$FC,$4D,$74,$7F,$C8,$5D,$D6,$7F,$C4
	.dc.b	$EF,$D1,$F8,$8F,$DB,$12,$7F,$8C,$3F,$C0,$FF,$01,$7C,$7F,$E1,$5B
	.dc.b	$3F,$40,$6C,$F9,$0D,$57,$DD,$F8,$B9,$5F,$2F,$F4,$67,$85,$8D,$3F
	.dc.b	$0B,$C0,$BF,$E1,$04,$BF,$E5,$4F,$8F,$0E,$B1,$E1,$CB,$C0,$FD,$7C
	.dc.b	$FC,$B8,$74,$BE,$4F,$D6,$C3,$B5,$74,$3F,$9B,$7C,$67,$F1,$72,$38
	.dc.b	$3E,$EC,$AF,$AF,$17,$F2,$D6,$CF,$F8,$7B,$5F,$2D,$78,$63,$C9,$9F
	.dc.b	$BE,$F8,$3F,$A1,$F8,$CF,$E4,$C4,$9F,$E9,$0F,$BD,$E2,$7B,$EF,$AD
	.dc.b	$7F,$22,$CE,$CF,$F4,$C6,$E1,$FD,$4F,$DF,$FE,$64,$49,$FE,$A8,$F5
	.dc.b	$8F,$DA,$E5,$EA,$7F,$DC,$3A,$FC,$C7,$39,$FA,$6F,$FA,$D5,$6E,$7F
	.dc.b	$AC,$61,$E0,$7F,$AD,$7C,$2F,$F4,$12,$57,$F6,$37,$C1,$97,$C2,$8F
	.dc.b	$61,$EE,$9F,$90,$78,$1F,$D9,$6D,$F6,$23,$07,$E2,$3E,$2F,$FA,$71
	.dc.b	$27,$FB,$83,$D9,$F3,$E5,$FC,$ED,$FC,$BE,$BE,$7B,$E9,$87,$8F,$0F
	.dc.b	$0B,$4F,$E9,$FE,$F7,$F5,$B2,$5F,$F7,$D7,$B4,$FF,$A0,$F4,$F9,$53
	.dc.b	$D8,$F6,$57,$37,$C6,$7B,$34,$F1,$7C,$57,$F6,$72,$47,$F8,$25,$EB
	.dc.b	$7F,$91,$CF,$D5,$B7,$D5,$C3,$A4,$F8,$67,$D9,$87,$D3,$C7,$FD,$2D
	.dc.b	$8F,$E6,$9B,$5F,$C6,$1A,$FF,$84,$9A,$F9,$8F,$9F,$7F,$E8,$8D,$3A
	.dc.b	$61,$DA,$FD,$A7,$E1,$3C,$9F,$E1,$04,$7C,$0F,$AB,$0F,$8B,$97,$A6
	.dc.b	$7B,$61,$EA,$C7,$DD,$1F,$14,$EA,$FC,$DE,$7F,$E2,$84,$8F,$63,$D7
	.dc.b	$1F,$8B,$97,$9A,$78,$D7,$E5,$9B,$8F,$E7,$63,$E0,$4F,$F1,$F2,$4F
	.dc.b	$C7,$AF,$D5,$BF,$6E,$7E,$1B,$E8,$23,$FE,$41,$CF,$C0,$FC,$D9,$FD
	.dc.b	$73,$F3,$5A,$FC,$D3,$13,$DC,$8F,$16,$BA,$3E,$C7,$C8,$FF,$35,$24
	.dc.b	$7D,$D7,$F2,$6F,$F3,$49,$F4,$E7,$FB,$7C,$67,$C4,$7E,$14,$FE,$79
	.dc.b	$F2,$1F,$CF,$C9,$7F,$87,$79,$D7,$1B,$FC,$89,$C6,$7E,$5D,$7F,$45
	.dc.b	$F8,$DF,$E9,$04,$8F,$C8,$3C,$A7,$C6,$71,$CB,$E5,$9F,$A9,$AF,$E9
	.dc.b	$DF,$1B,$FD,$44,$97,$F9,$97,$93,$F8,$43,$DC,$7F,$41,$8F,$EA,$C7
	.dc.b	$5D,$FF,$9B,$7C,$67,$F5,$B2,$4F,$E8,$DE,$77,$E2,$8E,$6F,$F2,$4A
	.dc.b	$E7,$1E,$E3,$B1,$FA,$B7,$FB,$31,$1F,$B3,$5F,$EE,$EC,$3D,$F1,$C2
	.dc.b	$BF,$61,$5B,$6C,$7E,$D7,$E3,$3F,$B8,$11,$C1,$D0,$FF,$75,$D7,$49
	.dc.b	$F3,$47,$F6,$1F,$87,$F3,$C7,$F3,$C7,$FB,$C7,$C4,$FF,$7B,$24,$FF
	.dc.b	$03,$7E,$0E,$3F,$C2,$A3,$E0,$BF,$D7,$71,$FF,$00,$57,$48,$FE,$C0
	.dc.b	$FF,$80,$FE,$0F,$FC,$0C,$4A,$FE,$1C,$F4,$7F,$4A,$74,$3F,$A4,$C7
	.dc.b	$47,$E2,$4F,$47,$FB,$D7,$C4,$FF,$84,$48,$FF,$19,$7A,$63,$FC,$7E
	.dc.b	$3F,$C0,$59,$73,$8F,$81,$1D,$0E,$E7,$F8,$17,$E1,$7F,$C3,$84,$7F
	.dc.b	$92,$3D,$23,$F8,$89,$FE,$70,$8F,$43,$FC,$F7,$0E,$D8,$7C,$59,$FA
	.dc.b	$04,$9F,$E5,$93,$FE,$29,$24,$7D,$98,$FE,$DB,$7F,$66,$BE,$04,$70
	.dc.b	$9F,$E6,$D5,$C2,$3F,$9B,$7C,$4F,$F8,$CC,$93,$FC,$ED,$F4,$71,$FF
	.dc.b	$2C,$C3,$CB,$3F,$EC,$FF,$5D,$E6,$E3,$00,$8C,$CF,$F8,$A7,$E1,$FF
	.dc.b	$C8,$04,$7F,$A4,$3F,$EC,$77,$FC,$7F,$CF,$FD,$48,$FE,$A7,$CF,$63
	.dc.b	$FC,$E7,$DC,$F8,$EF,$F1,$FC,$61,$B8,$FF,$A7,$27,$5E,$CF,$FA,$BD
	.dc.b	$E1,$E3,$7F,$93,$FE,$23,$FC,$A4,$4C,$3F,$AD,$3E,$3E,$C6,$7E,$1C
	.dc.b	$7E,$C4,$33,$C7,$1F,$FB,$DB,$2E,$B8,$F8,$F2,$F0,$68,$D2,$C7,$F9
	.dc.b	$7F,$E2,$3F,$CC,$44,$AF,$ED,$4D,$CF,$FB,$53,$5F,$ED,$4D,$7F,$C4
	.dc.b	$ED,$7F,$C0,$CF,$F9,$B7,$E2,$24,$FF,$75,$6C,$FF,$75,$6B,$FD,$D5
	.dc.b	$AF,$F8,$C9,$AF,$F8,$45,$AB,$F0,$24,$FF,$BF,$5A,$FF,$97,$1B,$1F
	.dc.b	$E9,$A6,$BF,$86,$3F,$0E,$E1,$C3,$5C,$C7,$52,$1D,$4A,$F5,$7C,$5F
	.dc.b	$13,$1F,$13,$63,$C4,$D7,$C4,$7A,$DC,$3E,$26,$BE,$66,$BF,$07,$E2
	.dc.b	$7E,$01,$27,$E0,$36,$3E,$03,$5E,$F4,$8C,$47,$C0,$6B,$F1,$1F,$99
	.dc.b	$F1,$32,$BE,$63,$67,$E6,$35,$EA,$D7,$E5,$B5,$F9,$ED,$7E,$D7,$C4
	.dc.b	$FD,$92,$4F,$D9,$6C,$7D,$96,$BE,$A6,$BF,$61,$AF,$DC,$7F,$09,$F1
	.dc.b	$32,$7F,$08,$D9,$FC,$23,$5F,$8A,$D7,$F0,$6D,$7F,$0E,$D7,$F2,$5F
	.dc.b	$13,$F9,$12,$4F,$E4,$5B,$1F,$91,$6B,$F4,$9A,$FE,$41,$AF,$E5,$1F
	.dc.b	$CE,$7C,$4C,$9F,$CE,$36,$7F,$38,$D7,$EE,$B5,$FC,$DB,$5F,$CF,$B5
	.dc.b	$FD,$57,$C4,$FE,$A4,$93,$FA,$96,$C7,$EA,$5A,$FE,$29,$AF,$EA,$1A
	.dc.b	$FE,$B1,$FD,$A7,$C4,$C9,$FD,$A3,$67,$F6,$8D,$7F,$2A,$D7,$F6,$6D
	.dc.b	$7F,$6E,$D7,$F8,$17,$C4,$FF,$01,$24,$FF,$01,$6C,$7F,$01,$6B,$FA
	.dc.b	$26,$BF,$C0,$1A,$FF,$04,$7F,$86,$7C,$4C,$9F,$E1,$8D,$9F,$E1,$8D
	.dc.b	$7F,$5A,$D7,$F8,$5B,$5F,$E1,$ED,$7F,$8D,$7C,$2F,$F1,$92,$4F,$F1
	.dc.b	$96,$C7,$F1,$96,$BF,$BA,$6B,$FC,$61,$AF,$F1,$C7,$F9,$27,$B3,$27
	.dc.b	$F9,$23,$67,$F9,$23,$5F,$E0,$AD,$7F,$91,$B5,$FE,$4E,$D7,$F9,$97
	.dc.b	$77,$F3,$12,$4F,$F3,$16,$C7,$F3,$16,$BF,$C4,$9A,$FF,$30,$6B,$FC
	.dc.b	$D1,$FE,$79,$CB,$27,$F9,$E3,$67,$F9,$E3,$5F,$E3,$AD,$7F,$9D,$B5
	.dc.b	$FE,$7E,$D7,$FA,$57,$37,$F4,$92,$4F,$F4,$96,$C7,$F4,$96,$BF,$CA
	.dc.b	$9A,$FF,$48,$6B,$FD,$31,$FE,$A9,$F1,$12,$7F,$AA,$36,$7F,$A0,$9F
	.dc.b	$D6,$BD,$DF,$EB,$6B,$AE,$B9,$AF,$F0,$EC,$69,$76,$BC,$EE,$33,$C2
	.dc.b	$57,$87,$FA,$EF,$C3,$DC,$3F,$B0,$B6,$9E,$C3,$CE,$D5,$7F,$6F,$AB
	.dc.b	$D6,$7F,$B8,$13,$0F,$EE,$0F,$A2,$7F,$B7,$47,$88,$F4,$98,$C7,$FD
	.dc.b	$FF,$97,$68,$D4,$FC,$17,$CE,$74,$23,$FD,$EE,$BF,$C0,$1F,$0F,$27
	.dc.b	$FC,$00,$D9,$F9,$B1,$D3,$3F,$89,$87,$82,$3C,$73,$FE,$48,$74,$78
	.dc.b	$1F,$11,$F4,$F3,$79,$FF,$0B,$1F,$91,$18,$AE,$75,$FE,$15,$F8,$7F
	.dc.b	$F0,$A1,$23,$FC,$28,$FC,$58,$F0,$61,$F0,$4F,$23,$E0,$3F,$00,$59
	.dc.b	$7C,$47,$C7,$3F,$C7,$BC,$16,$3C,$26,$8F,$C6,$CF,$F8,$9A,$7C,$0D
	.dc.b	$8C,$9B,$7E,$3D,$78,$E3,$E2,$9F,$8E,$D7,$13,$E3,$AE,$73,$C6,$3F
	.dc.b	$C3,$6D,$7F,$03,$1A,$F8,$89,$3E,$06,$CF,$82,$3D,$37,$F0,$B5,$FA
	.dc.b	$F1,$EC,$BF,$F2,$DE,$9E,$FB,$FF,$37,$78,$9F,$93,$AF,$83,$F1,$1E
	.dc.b	$42,$3E,$E7,$CF,$8F,$AB,$1F,$D5,$71,$FE,$A6,$BC,$27,$CA,$F8,$8F
	.dc.b	$CA,$1F,$7B,$CF,$B1,$93,$E4,$C7,$C2,$23,$F3,$2F,$B3,$F8,$DA,$ED
	.dc.b	$97,$DB,$9E,$6F,$E5,$AB,$8B,$6B,$EA,$F9,$BF,$38,$8E,$0E,$C7,$A2
	.dc.b	$3F,$D0,$BB,$F8,$1F,$CE,$57,$E5,$BC,$DE,$F9,$FF,$06,$E3,$47,$EC
	.dc.b	$7E,$23,$ED,$93,$1F,$C2,$BF,$21,$ED,$5F,$23,$1E,$D8,$FF,$9C,$BE
	.dc.b	$23,$F7,$47,$FA,$43,$9E,$DF,$E5,$23,$F0,$E4,$7F,$22,$F5,$9F,$89
	.dc.b	$1F,$90,$DB,$E3,$E1,$F8,$29,$FD,$15,$71,$7F,$7A,$D7,$F2,$3F,$11
	.dc.b	$F9,$42,$5F,$E7,$9E,$D5,$CC,$F5,$47,$B0,$FE,$2F,$5F,$EA,$8E,$7F
	.dc.b	$4D,$FE,$39,$FE,$2D,$F1,$1F,$A0,$25,$7E,$A9,$EA,$FF,$1D,$3D,$59
	.dc.b	$FC,$67,$F8,$6D,$7F,$00,$7F,$88,$E0,$FD,$65,$E6,$6F,$8F,$E2,$5F
	.dc.b	$EB,$1A,$F5,$3F,$50,$43,$57,$F8,$47,$06,$CC,$9F,$C9,$3E,$27,$F7
	.dc.b	$24,$8F,$E0,$6F,$AA,$3F,$D5,$BE,$E3,$E0,$7F,$19,$48,$C3,$F2,$67
	.dc.b	$FA,$C6,$3F,$E8,$D8,$C6,$5E,$DF,$3F,$F9,$F7,$C4,$FF,$08,$24,$FF
	.dc.b	$12,$7C,$0F,$5C,$7B,$E3,$DB,$1F,$E4,$5F,$13,$FD,$1C,$FF,$76,$7D
	.dc.b	$37,$F1,$9C,$BF,$B0,$46,$70,$FE,$3A,$FF,$16,$23,$F1,$63,$FB,$55
	.dc.b	$F5,$7F,$09,$1F,$D0,$E3,$0F,$F5,$A9,$E3,$7F,$D3,$1A,$FF,$20,$F8
	.dc.b	$1F,$E4,$64,$9F,$E5,$AF,$B2,$B9,$CF,$E0,$4F,$E9,$B3,$EF,$3F,$9C
	.dc.b	$D7,$B3,$C1,$F5,$4F,$F5,$2F,$59,$92,$7F,$B4,$7C,$4F,$F3,$22,$61
	.dc.b	$FC,$F1,$F0,$69,$F2,$8F,$ED,$DF,$13,$27,$F2,$3C,$0F,$4B,$19,$FF
	.dc.b	$4F,$AF,$E7,$A4,$AF,$E8,$EF,$47,$F0,$07,$A6,$38,$3A,$23,$FD,$CA
	.dc.b	$3F,$A3,$F7,$49,$FE,$A2,$D9,$FC,$DF,$07,$D4,$C7,$D1,$5F,$D7,$2F
	.dc.b	$FC,$2D,$8B,$D2,$F3,$DF,$F8,$82,$31,$5F,$D8,$BE,$27,$FA,$A9,$1F
	.dc.b	$EB,$EF,$A7,$1F,$E1,$E7,$D2,$E1,$FA,$AA,$5F,$13,$FE,$15,$3F,$1F
	.dc.b	$E2,$D8,$FE,$DF,$1F,$D8,$88,$FF,$6C,$7B,$63,$F8,$CA,$FD,$5E,$9C
	.dc.b	$5F,$F0,$AC,$F0,$6D,$7F,$74,$F8,$9F,$ED,$A4,$70,$7E,$64,$FC,$B5
	.dc.b	$7C,$69,$FF,$03,$57,$E9,$3C,$4E,$D8,$7F,$93,$39,$3F,$13,$1F,$E2
	.dc.b	$FB,$FE,$F2,$4C,$3F,$C0,$6F,$A1,$FE,$1D,$5F,$4D,$FF,$0E,$FE,$07
	.dc.b	$FE,$27,$C7,$FC,$2C,$D9,$FA,$5E,$97,$F9,$22,$31,$4A,$91,$C5,$C1
	.dc.b	$83,$FC,$2A,$E5,$CB,$5F,$D4,$57,$8D,$FB,$2D,$3F,$CC,$F1,$B8,$34
	.dc.b	$8F,$6F,$8A,$95,$FE,$26,$94,$62,$7F,$27,$3F,$38,$FC,$FE,$DF,$86
	.dc.b	$AF,$6C,$FF,$20,$F4,$FE,$03,$F9,$67,$E1,$7C,$57,$B0,$97,$DD,$F7
	.dc.b	$FC,$2F,$C7,$3F,$CE,$91,$FE,$7C,$C3,$0F,$D0,$F8,$AF,$98,$7F,$9D
	.dc.b	$FD,$57,$BD,$89,$D1,$EB,$9F,$96,$47,$DD,$7D,$32,$FE,$23,$E5,$7F
	.dc.b	$3B,$9F,$BB,$3C,$E7,$FD,$2B,$A2,$F9,$46,$1F,$AF,$F0,$92,$FE,$53
	.dc.b	$67,$8D,$F7,$DF,$FC,$5F,$A5,$BF,$BB,$87,$AF,$DF,$F0,$57,$FA,$6B
	.dc.b	$54,$7E,$2F,$E0,$FF,$0A,$4B,$FB,$0F,$CA,$7D,$27,$F8,$27,$0E,$D9
	.dc.b	$7E,$43,$E0,$FF,$28,$7F,$48,$F4,$AD,$7E,$72,$3F,$22,$47,$F0,$8F
	.dc.b	$59,$FC,$FE,$9F,$5B,$7F,$91,$18,$AE,$47,$02,$3F,$2A,$BF,$3B,$E5
	.dc.b	$C9,$FC,$83,$67,$E3,$C7,$A9,$FA,$B8,$FA,$9F,$CD,$D7,$F6,$0F,$47
	.dc.b	$D3,$1F,$D1,$7C,$54,$7E,$C6,$BF,$56,$4B,$FC,$F3,$EB,$DF,$F4,$F9
	.dc.b	$FA,$E3,$B7,$13,$E2,$BF,$74,$7E,$3F,$E0,$2E,$1F,$C0,$A3,$F6,$E4
	.dc.b	$7F,$54,$F5,$7F,$73,$7D,$5F,$66,$1F,$B6,$9E,$19,$FE,$39,$AF,$F0
	.dc.b	$43,$F7,$3F,$15,$FC,$10,$95,$FB,$77,$B4,$7F,$0C,$BE,$CF,$D3,$3F
	.dc.b	$8A,$D7,$E9,$FE,$07,$D3,$1F,$EB,$CC,$1C,$89,$FE,$31,$EF,$7F,$12
	.dc.b	$26,$5F,$C1,$5E,$F7,$FC,$33,$1E,$F8,$7C,$AA,$FE,$4B,$EE,$FF,$21
	.dc.b	$3F,$C2,$BE,$AD,$8C,$9F,$C3,$E3,$F8,$F9,$1F,$DA,$9F,$18,$FF,$44
	.dc.b	$6F,$F5,$23,$58,$7E,$D1,$AF,$C7,$DD,$F1,$5F,$CA,$89,$7F,$C7,$5F
	.dc.b	$4C,$32,$F4,$CF,$D6,$3C,$B5,$F8,$BF,$5F,$D0,$D8,$FE,$83,$F1,$5F
	.dc.b	$CD,$88,$FF,$29,$7A,$D7,$49,$FA,$93,$EB,$69,$F4,$77,$7C,$0F,$E4
	.dc.b	$2B,$F5,$D3,$EB,$6B,$E5,$3F,$4A,$43,$55,$F2,$1C,$1B,$32,$7F,$1A
	.dc.b	$F8,$AF,$E9,$A4,$8F,$E8,$0F,$E2,$97,$4F,$E8,$5F,$41,$8E,$3F,$E1
	.dc.b	$7C,$3F,$88,$CA,$9F,$8A,$9F,$EC,$BE,$EF,$F5,$A2,$3F,$D3,$5E,$D7
	.dc.b	$D7,$1F,$CB,$5F,$FB,$4B,$0F,$ED,$1E,$EC,$C3,$FB,$44,$FF,$79,$E9
	.dc.b	$7B,$79,$C3,$FB,$8D,$FF,$68,$25,$FF,$5B,$7B,$E1,$FC,$8A,$3D,$47
	.dc.b	$F7,$A8,$DC,$FF,$3E,$6B,$F8,$BF,$89,$FE,$E6,$49,$FE,$CC,$F4,$7E
	.dc.b	$2E,$5E,$BC,$BF,$B8,$DF,$F9,$D3,$DB,$F3,$47,$A5,$FF,$02,$7C,$4F
	.dc.b	$F7,$E2,$5F,$F7,$17,$A6,$3F,$12,$3F,$DD,$19,$7F,$82,$FE,$26,$47
	.dc.b	$FA,$A3,$F5,$CF,$F4,$03,$FE,$06,$8C,$D7,$F5,$87,$FC,$16,$48,$FC
	.dc.b	$36,$3F,$3D,$F4,$DF,$0A,$FD,$55,$6D,$E5,$3F,$C4,$DF,$F0,$D7,$CA
	.dc.b	$FF,$91,$63,$10,$8C,$3B,$6B,$FC,$6D,$AF,$E6,$EF,$C4,$FB,$CF,$F2
	.dc.b	$86,$9F,$2E,$BD,$87,$BF,$DA,$F6,$D7,$F9,$56,$8B,$2F,$F4,$25,$7E
	.dc.b	$20,$FF,$16,$E1,$FE,$CC,$7C,$ED,$AF,$37,$C0,$73,$7C,$CF,$D2,$BF
	.dc.b	$E3,$91,$FE,$06,$CF,$C1,$1B,$AF,$F5,$F4,$F8,$AF,$BF,$BF,$FE,$B4
	.dc.b	$F8,$0D,$C6,$0F,$24,$FC,$A9,$F0,$39,$3E,$43,$87,$51,$F6,$0F,$C1
	.dc.b	$9F,$60,$E6,$FD,$18,$E0,$D8,$F7,$C6,$2B,$CF,$3F,$4F,$CC,$C1,$2B
	.dc.b	$E2,$BF,$05,$FA,$C7,$82,$3D,$C7,$90,$FE,$E1,$F0,$1E,$E7,$18,$FE
	.dc.b	$DE,$31,$1F,$85,$F2,$FC,$44,$AF,$A6,$F7,$8F,$E4,$B5,$F8,$43,$E7
	.dc.b	$7C,$5F,$2C,$4F,$78,$FC,$07,$87,$98,$98,$7D,$E7,$E0,$11,$F9,$57
	.dc.b	$F8,$ED,$7F,$82,$70,$FE,$3E,$3F,$3B,$5B,$7F,$34,$D7,$F3,$5F,$17
	.dc.b	$F2,$C9,$7F,$8C,$7D,$51,$FD,$96,$3E,$F9,$FC,$41,$F9,$27,$E9,$E7
	.dc.b	$D5,$C2,$F0,$3D,$33,$FD,$D6,$31,$5F,$7B,$4F,$F6,$B3,$F9,$AF,$7F
	.dc.b	$EE,$11,$C1,$2B,$F6,$51,$96,$C6,$5F,$DA,$46,$5F,$DB,$7B,$BF,$87
	.dc.b	$24,$FE,$A1,$B1,$FB,$D6,$BC,$23,$44,$74,$D7,$F8,$17,$95,$F9,$32
	.dc.b	$47,$ED,$1B,$1F,$C2,$5A,$F0,$8D,$11,$D3,$5F,$E1,$DE,$4F,$E7,$C9
	.dc.b	$1F,$C0,$1B,$1F,$C5,$5A,$F0,$8D,$11,$D3,$5F,$E3,$5F,$17,$FA,$B2
	.dc.b	$47,$F0,$C6,$C7,$F2,$16,$BC,$23,$44,$74,$D7,$F9,$37,$C5,$FE,$DC
	.dc.b	$91,$FC,$61,$B1,$FC,$B5,$AF,$08,$D1,$1D,$35,$FE,$65,$EB,$FF,$03
	.dc.b	$24,$7F,$24,$7F,$84,$FC,$73,$E1,$BF,$73,$F8,$27,$F9,$B2,$BF,$87
	.dc.b	$CE,$5F,$E2,$31,$9B,$FC,$EE,$CF,$67,$F7,$07,$ED,$FF,$5E,$7C,$6E
	.dc.b	$30,$FC,$E4,$FF,$15,$23,$FC,$D9,$FE,$37,$87,$5E,$9F,$E6,$CD,$9F
	.dc.b	$F7,$4E,$5E,$3C,$BF,$5B,$39,$7F,$AB,$DF,$47,$FA,$A3,$5F,$EA,$9E
	.dc.b	$67,$F2,$52,$5F,$F4,$67,$C3,$EA,$7C,$23,$F8,$DE,$1D,$F1,$F1,$C7
	.dc.b	$AE,$FF,$CD,$99,$BE,$33,$15,$FD,$22,$31,$8F,$EE,$FC,$CF,$E6,$84
	.dc.b	$7F,$AA,$BE,$07,$E6,$E1,$E0,$DF,$C5,$1F,$C0,$F6,$FE,$48,$7F,$31
	.dc.b	$F8,$07,$98,$F8,$9C,$D6,$CF,$EC,$35,$FD,$0C,$8F,$F7,$AE,$1F,$91
	.dc.b	$B1,$BB,$FB,$E5,$32,$77,$C5,$F7,$27,$A9,$27,$FC,$05,$7F,$D4,$78
	.dc.b	$24,$FF,$72,$6C,$FF,$83,$6F,$C2,$78,$34,$F7,$C7,$E3,$2F,$DD,$9B
	.dc.b	$E2,$BD,$B3,$FE,$C1,$7F,$A3,$7C,$5F,$F5,$F2,$65,$FE,$00,$7E,$64
	.dc.b	$FE,$1B,$7F,$99,$3F,$38,$FE,$5D,$8F,$F4,$F3,$FA,$77,$A0,$F8,$6C
	.dc.b	$4E,$8F,$F0,$5C,$FF,$6D,$23,$F2,$EB,$D3,$C9,$FE,$0D,$E0,$F9,$58
	.dc.b	$FC,$99,$E6,$FF,$89,$EF,$8B,$6F,$EA,$7C,$6F,$E8,$88,$E0,$EC,$7A
	.dc.b	$63,$FA,$6D,$F6,$7E,$D5,$F6,$C3,$FC,$33,$93,$E1,$71,$87,$F9,$BA
	.dc.b	$30,$FE,$F3,$CD,$FD,$89,$1F,$C3,$3F,$2A,$BF,$43,$1F,$C1,$33,$F7
	.dc.b	$BF,$86,$C7,$F1,$5A,$7F,$0A,$3E,$CF,$C6,$59,$FE,$13,$1F,$BE,$23
	.dc.b	$F9,$37,$B3,$CF,$8B,$F0,$8F,$BF,$0F,$E0,$73,$C2,$FF,$14,$D7,$FD
	.dc.b	$D5,$5E,$C7,$F1,$D9,$FF,$0C,$25,$FC,$A8,$FC,$23,$EB,$3B,$1E,$1A
	.dc.b	$FC,$8F,$C5,$FA,$67,$FC,$47,$CF,$FB,$08,$FE,$4F,$F0,$3F,$C6,$09
	.dc.b	$7F,$AD,$7D,$75,$F8,$18,$EC,$F8,$09,$97,$CC,$7F,$DE,$DB,$7F,$A8
	.dc.b	$27,$C7,$1C,$4F,$D4,$6F,$F9,$C7,$DE,$D7,$E4,$35,$E0,$6A,$3E,$03
	.dc.b	$4E,$6F,$F0,$3F,$2B,$F9,$89,$1C,$1F,$1E,$FF,$99,$FC,$57,$E0,$2B
	.dc.b	$F8,$23,$61,$83,$D4,$C3,$6F,$FA,$4F,$91,$FC,$F8,$8F,$F1,$37,$E3
	.dc.b	$E1,$FC,$4F,$0E,$CD,$9F,$EA,$3E,$BF,$F5,$63,$F8,$2F,$C6,$58,$C9
	.dc.b	$FC,$76,$3F,$A5,$91,$F5,$1F,$8D,$AE,$BC,$47,$95,$FF,$15,$AE,$0D
	.dc.b	$AF,$E5,$19,$7F,$56,$23,$83,$AC,$7C,$47,$EC,$BE,$93,$FA,$BC,$FF
	.dc.b	$1C,$AE,$F9,$FF,$B7,$DF,$EE,$3E,$9F,$F6,$22,$5F,$F3,$77,$E2,$E3
	.dc.b	$E7,$8E,$AD,$C3,$FB,$1F,$A3,$FD,$E8,$FE,$37,$CB,$6B,$FB,$CC,$7F
	.dc.b	$6E,$23,$FD,$0D,$EB,$87,$C4,$38,$57,$F3,$78,$DB,$67,$FA,$57,$81
	.dc.b	$FD,$E0,$8E,$0F,$E9,$D5,$FE,$0C,$3A,$3E,$23,$FD,$4B,$97,$E2,$38
	.dc.b	$FB,$BF,$C6,$1F,$F0,$9F,$67,$F8,$10,$93,$FD,$65,$FF,$03,$D7,$58
	.dc.b	$F8,$6F,$F8,$6B,$29,$1F,$EB,$2E,$5F,$EC,$FF,$15,$8C,$FF,$C1,$35
	.dc.b	$FE,$10,$25,$7F,$65,$7A,$BF,$E2,$0B,$F4,$CF,$08,$D1,$1F,$EB,$93
	.dc.b	$FE,$1A,$C6,$4F,$F6,$F6,$CF,$F6,$1A,$F4,$3F,$C0,$A7,$F2,$87,$F8
	.dc.b	$BA,$BF,$B0,$FC,$5F,$9F,$3F,$F7,$16,$5F,$E0,$08,$FF,$1B,$F1,$7F
	.dc.b	$8A,$09,$7F,$DF,$5F,$4E,$1F,$E2,$6C,$7D,$38,$7E,$EA,$BD,$87,$D2
	.dc.b	$E4,$7C,$5F,$92,$75,$7E,$7A,$FB,$64,$70,$43,$F9,$8D,$F8,$1B,$1B
	.dc.b	$69,$B3,$F9,$F6,$BF,$E4,$02,$47,$F8,$55,$F4,$D7,$FB,$DF,$F4,$27
	.dc.b	$7F,$96,$BF,$A8,$46,$1F,$5D,$FF,$B6,$29,$FA,$FF,$F8,$6A,$73,$1F
	.dc.b	$EF,$C8,$C4,$FF,$92,$B5,$FF,$28,$92,$BF,$C4,$CF,$F9,$53,$3F,$F3
	.dc.b	$29,$FE,$DE,$F9,$5F,$EE,$1C,$BF,$22,$BF,$CC,$B3,$A7,$F3,$BF,$B0
	.dc.b	$B9,$F1,$8F,$F9,$62,$BA,$57,$F8,$CD,$B1,$FE,$25,$6B,$FD,$71,$AF
	.dc.b	$F6,$56,$BF,$E3,$56,$AF,$8F,$F6,$BF,$DD,$4C,$09,$C3,$FE,$3E,$8D
	.dc.b	$BF,$DE,$DA,$F5,$8F,$F2,$1E,$9C,$63,$87,$C4,$7B,$23,$11,$E1,$8C
	.dc.b	$4F,$0C,$5D,$38,$A5,$A7,$35,$D8,$F1,$61,$D2,$AC,$7A,$1F,$74,$65
	.dc.b	$F5,$3F,$2A,$BC,$6B,$0F,$04,$F9,$6B,$B3,$F0,$7E,$3F,$D9,$7C,$1F
	.dc.b	$F2,$8E,$58,$AF,$89,$16,$B8,$47,$8F,$04,$F8,$E7,$E9,$36,$33,$3F
	.dc.b	$7C,$F0,$CF,$8C,$FC,$23,$88,$F4,$36,$79,$11,$F0,$3D,$7E,$3A,$47
	.dc.b	$8D,$B5,$F8,$B8,$E9,$8F,$C2,$BE,$B3,$F4,$A7,$DA,$7F,$27,$F0,$7F
	.dc.b	$4B,$97,$C3,$8C,$4F,$C1,$3F,$1D,$F1,$FE,$D2,$47,$C3,$7E,$93,$F8
	.dc.b	$BC,$3C,$07,$1C,$1F,$1F,$4F,$99,$87,$E0,$FA,$71,$7F,$0E,$7E,$51
	.dc.b	$1F,$A0,$FC,$DB,$E8,$76,$8E,$06,$89,$8F,$E8,$30,$F0,$C7,$D9,$F8
	.dc.b	$EF,$AC,$47,$07,$DD,$AE,$AF,$EF,$8F,$96,$76,$AF,$E2,$1B,$FF,$73
	.dc.b	$CB,$F2,$38,$FB,$23,$15,$FA,$DF,$8F,$FC,$21,$1F,$C5,$3E,$B7,$AD
	.dc.b	$FE,$10,$FA,$C7,$EC,$1F,$C3,$7C,$77,$CB,$3F,$D8,$9B,$7F,$01,$8C
	.dc.b	$92,$7E,$04,$FE,$4C,$91,$ED,$3B,$E3,$FE,$F0,$E1,$F7,$D7,$80,$FE
	.dc.b	$19,$5C,$1E,$75,$FA,$18,$FC,$F7,$C0,$48,$FD,$23,$67,$E8,$69,$D3
	.dc.b	$0E,$D5,$E0,$3F,$88,$5F,$AB,$8B,$D9,$1F,$93,$7F,$11,$F1,$FF,$AB
	.dc.b	$24,$7E,$C5,$F9,$18,$79,$F8,$3F,$3B,$87,$68,$C6,$3E,$1B,$FC,$A1
	.dc.b	$FE,$12,$ED,$F9,$13,$A8,$FA,$2F,$F2,$9B,$FD,$3B,$6B,$BB,$67,$E5
	.dc.b	$3E,$E6,$B9,$30,$43,$07,$5F,$8F,$FE,$0C,$48,$FE,$18,$FB,$30,$FE
	.dc.b	$37,$DF,$E9,$AF,$E7,$0D,$8F,$F2,$D6,$77,$2F,$E1,$F5,$F0,$FE,$3F
	.dc.b	$F8,$91,$31,$FE,$30,$F5,$7F,$A0,$61,$D6,$7F,$A5,$3F,$B7,$F8,$E9
	.dc.b	$3F,$C1,$7A,$2D,$7C,$F9,$FE,$3C,$49,$FE,$48,$F5,$8F,$3C,$7D,$F6
	.dc.b	$9C,$5D,$91,$FD,$B4,$7F,$29,$F8,$E9,$3F,$CB,$DB,$3F,$BB,$BF,$E1
	.dc.b	$87,$A7,$6F,$F4,$9D,$7F,$2B,$BF,$D2,$F4,$FA,$35,$FF,$15,$F2,$23
	.dc.b	$F8,$A7,$C7,$7F,$37,$25,$FF,$3D,$7F,$9D,$63,$DA,$3D,$59,$7C,$67
	.dc.b	$F9,$C4,$66,$3F,$C2,$5B,$39,$B1,$97,$DA,$BF,$E9,$17,$D1,$A4,$BF
	.dc.b	$63,$5F,$C9,$46,$DC,$90,$CC,$7F,$53,$F8,$DF,$EA,$04,$70,$75,$AF
	.dc.b	$EE,$FF,$84,$3E,$33,$CB,$3F,$EF,$B8,$C3,$EC,$3F,$4D,$CF,$AC,$3F
	.dc.b	$BC,$E6,$8F,$E7,$9F,$1F,$FD,$7C,$95,$FD,$95,$F4,$63,$E0,$C7,$96
	.dc.b	$0C,$7B,$BF,$B9,$F8,$FF,$E8,$C7,$F1,$CE,$DB,$8F,$F6,$58,$FE,$D8
	.dc.b	$47,$FB,$93,$FD,$C6,$B9,$E1,$DE,$7D,$57,$FC,$6E,$73,$1F,$E1,$FA
	.dc.b	$E5,$3F,$D1,$5A,$FF,$01,$F8,$FF,$EF,$44,$9F,$F0,$03,$DF,$1F,$EC
	.dc.b	$97,$D2,$BF,$3B,$87,$85,$F6,$5F,$F6,$DE,$FF,$5C,$7F,$3C,$F2,$9C
	.dc.b	$89,$E7,$5E,$B2,$1F,$E2,$1C,$3F,$C7,$4D,$C7,$C2,$7C,$C8,$F1,$61
	.dc.b	$F3,$63,$33,$FC,$7F,$A6,$CF,$C3,$9F,$54,$F3,$24,$7F,$91,$AB,$EA
	.dc.b	$BF,$E4,$B6,$BF,$87,$9E,$2E,$DA,$ED,$AF,$E3,$7E,$3F,$EB,$12,$BD
	.dc.b	$8F,$2C,$FF,$C9,$7E,$13,$DD,$F3,$C6,$1B,$5F,$E5,$4E,$37,$4E,$31
	.dc.b	$FC,$AB,$F5,$3E,$3F,$F0,$84,$AF,$94,$D9,$F7,$35,$F7,$35,$FD,$23
	.dc.b	$51,$F9,$97,$F2,$1F,$1F,$27,$EB,$36,$7E,$53,$5F,$94,$D7,$F6,$0D
	.dc.b	$7A,$EA,$8C,$47,$E9,$FE,$37,$F3,$C4,$AF,$C2,$36,$3E,$C3,$5F,$B0
	.dc.b	$D7,$F7,$CD,$7B,$BF,$A9,$F8,$E9,$3F,$8F,$6C,$FE,$0D,$AF,$E0,$DA
	.dc.b	$FF,$09,$6B,$EA,$6B,$FB,$FF,$8D,$FD,$B1,$27,$F3,$8F,$F9,$ED,$7C
	.dc.b	$7B,$C6,$FF,$38,$46,$3C,$6B,$39,$8F,$F3,$D9,$93,$FC,$B5,$DF,$FE
	.dc.b	$24,$3F,$C4,$BD,$B6,$3F,$CE,$3A,$BE,$3B,$F8,$41,$23,$F5,$CF,$8E
	.dc.b	$FB,$67,$FE,$32,$9E,$CD,$8C,$F1,$FF,$A9,$67,$2F,$F4,$EC,$7C,$15
	.dc.b	$FB,$FE,$9F,$E5,$15,$E6,$3C,$7F,$1D,$FC,$64,$8F,$EF,$DB,$3F,$B3
	.dc.b	$6B,$FB,$36,$BF,$CE,$5A,$FE,$29,$FE,$49,$F1,$D2,$7F,$85,$36,$7F
	.dc.b	$7C,$D7,$F7,$CD,$7F,$A2,$35,$FC,$AB,$5F,$E7,$5F,$1D,$FC,$C4,$93
	.dc.b	$FC,$5D,$FE,$69,$F1,$DF,$C9,$8F,$F5,$57,$8E,$EA,$46,$6B,$EA,$B6
	.dc.b	$FE,$47,$99,$75,$FE,$B3,$3D,$2B,$F9,$23,$6B,$F8,$C3,$5F,$E3,$0D
	.dc.b	$7F,$AF,$35,$FD,$7B,$5F,$EA,$9F,$1D,$FD,$2C,$93,$FC,$C5,$ED,$5F
	.dc.b	$EE,$3C,$7D,$51,$E5,$9F,$F7,$66,$28,$C5,$BC,$1F,$E7,$FC,$B8,$6D
	.dc.b	$FE,$53,$E3,$BF,$AD,$11,$FE,$80,$FF,$5D,$F8,$EF,$E9,$E7,$CA,$C5
	.dc.b	$18,$9F,$CE,$CD,$C3,$F2,$1C,$F7,$4F,$F0,$0C,$F3,$AF,$E9,$8D,$AF
	.dc.b	$E8,$0D,$7F,$A0,$35,$FF,$03,$35,$FE,$2E,$D7,$FB,$E7,$C6,$FF,$73
	.dc.b	$24,$FF,$59,$7B,$57,$C9,$C5,$7E,$5A,$F9,$2D,$BF,$91,$C6,$F4,$31
	.dc.b	$A7,$F6,$87,$FB,$8F,$C6,$7F,$80,$49,$87,$F6,$77,$FC,$09,$E1,$7F
	.dc.b	$76,$3E,$46,$96,$7F,$81,$C5,$C7,$F6,$DE,$2D,$D7,$FC,$4F,$3C,$AB
	.dc.b	$FB,$93,$6B,$FB,$33,$5F,$EC,$CD,$7F,$C5,$8D,$7F,$9E,$B5,$FF,$11
	.dc.b	$7C,$5F,$F8,$64,$93,$FD,$FD,$ED,$5F,$1F,$1F,$24,$FE,$39,$B5,$F1
	.dc.b	$BC,$3B,$A7,$F8,$31,$FF,$0A,$FC,$4F,$F8,$94,$97,$FE,$0B,$4F,$F5
	.dc.b	$CA,$83,$F0,$87,$BD,$FF,$00,$5B,$3F,$D8,$B2,$FB,$5A,$66,$A4,$E4
	.dc.b	$E9,$5C,$2F,$F9,$17,$A7,$A7,$FD,$9D,$19,$7C,$2D,$78,$D7,$99,$F2
	.dc.b	$9E,$57,$0F,$06,$BC,$2B,$6E,$DB,$3D,$63,$FB,$DB,$5D,$1D,$9C,$13
	.dc.b	$E2,$FA,$C7,$06,$BF,$1D,$B0,$6D,$7E,$DF,$6F,$81,$97,$E8,$5B,$87
	.dc.b	$F8,$33,$5F,$F2,$1F,$A7,$8C,$BF,$42,$FD,$1F,$8A,$EA,$4A,$F5,$BE
	.dc.b	$0F,$8B,$FA,$C7,$F0,$5D,$EC,$FD,$3B,$F7,$CE,$1F,$EA,$9E,$AD,$CF
	.dc.b	$E5,$EC,$AE,$97,$F2,$1B,$7F,$15,$AF,$D8,$6B,$F8,$06,$BE,$08,$E4
	.dc.b	$7D,$2E,$6F,$8E,$47,$06,$A1,$7F,$CB,$FE,$4B,$F1,$4F,$9A,$FF,$A5
	.dc.b	$C6,$2B,$C0,$7B,$F5,$FF,$36,$FA,$DE,$D8,$F9,$D8,$A3,$F3,$9F,$19
	.dc.b	$F6,$09,$3F,$86,$7E,$DF,$C6,$7E,$84,$EF,$C7,$67,$E6,$69,$FA,$79
	.dc.b	$C3,$FD,$FF,$D8,$BA,$FC,$C9,$E7,$5F,$92,$6D,$7E,$3D,$AF,$E8,$9A
	.dc.b	$FE,$A9,$AF,$E6,$1A,$FE,$DF,$E3,$3F,$22,$49,$FC,$FB,$DA,$BE,$57
	.dc.b	$17,$92,$7F,$CB,$7E,$1B,$74,$E7,$0F,$E6,$3E,$C5,$E0,$FF,$4B,$BF
	.dc.b	$C0,$FE,$2F,$F3,$E4,$C7,$F5,$AF,$E8,$BE,$33,$F8,$41,$F5,$B8,$EC
	.dc.b	$FC,$8E,$44,$E3,$1F,$F1,$37,$B5,$76,$FC,$9C,$F3,$BF,$DC,$B6,$BF
	.dc.b	$6C,$D7,$F8,$4B,$5F,$E2,$2D,$7F,$7E,$D7,$F8,$F7,$C1,$7E,$E0,$93
	.dc.b	$FC,$19,$ED,$5F,$91,$E6,$F2,$4F,$C8,$DF,$FD,$72,$FA,$F0,$FE,$2D
	.dc.b	$5F,$C5,$F6,$7D,$7F,$FA,$E3,$2F,$E3,$F5,$AB,$FE,$01,$39,$BF,$D8
	.dc.b	$FC,$1F,$F0,$A2,$4F,$F1,$87,$F8,$6F,$C6,$FF,$33,$3F,$43,$D1,$67
	.dc.b	$E6,$45,$C7,$FD,$9D,$87,$FB,$13,$74,$E6,$7F,$97,$BD,$2B,$F9,$2B
	.dc.b	$63,$F9,$13,$5D,$C6,$88,$E9,$AF,$F3,$AF,$8D,$FE,$46,$48,$FE,$62
	.dc.b	$F5,$CB,$F7,$9C,$FE,$76,$CF,$A8,$FD,$D6,$2F,$77,$51,$F2,$1F,$EA
	.dc.b	$9F,$19,$FC,$C0,$91,$FC,$F9,$FE,$67,$F1,$BF,$D6,$4F,$E0,$FD,$56
	.dc.b	$7F,$25,$17,$3F,$8F,$EF,$5C,$BF,$91,$CF,$2A,$FE,$92,$DA,$FE,$8C
	.dc.b	$D7,$FA,$CB,$5F,$EC,$2D,$7F,$A7,$B5,$FE,$DD,$E6,$7F,$48,$24,$FF
	.dc.b	$56,$7B,$57,$F2,$1E,$9F,$24,$FE,$45,$B5,$FE,$4D,$D3,$FD,$91,$3E
	.dc.b	$C3,$E4,$FB,$DF,$DE,$7D,$FF,$EA,$A4,$8F,$EC,$6F,$F5,$8F,$8D,$FE
	.dc.b	$FC,$7F,$2C,$E6,$B3,$FB,$78,$B8,$7E,$33,$E0,$2E,$7F,$D2,$27,$95
	.dc.b	$7F,$6E,$6D,$7F,$6B,$6B,$FD,$F9,$AF,$F8,$19,$AF,$F7,$46,$BF,$E1
	.dc.b	$3F,$8B,$FE,$DA,$49,$FE,$F6,$F6,$AF,$E8,$DD,$9E,$49,$FD,$B3,$6B
	.dc.b	$E3,$7C,$0D,$EB,$FD,$AB,$FE,$1D,$F8,$8F,$EF,$24,$BF,$F0,$3B,$FD
	.dc.b	$FB,$F0,$8F,$F8,$03,$1F,$D0,$1B,$7F,$D5,$A3,$F8,$27,$6F,$EF,$60
	.dc.b	$27,$79,$7F,$86,$1B,$5F,$E3,$16,$CF,$DB,$F8,$3E,$4F,$12,$C7,$81
	.dc.b	$FF,$1B,$B5,$E4,$79,$C9,$3F,$E2,$28,$D4,$79,$1A,$F5,$8D,$3C,$0F
	.dc.b	$F1,$7C,$68,$9E,$2F,$C3,$24,$F5,$78,$5F,$F7,$7E,$DF,$3D,$FF,$76
	.dc.b	$6D,$7E,$F3,$E0,$6C,$FF,$16,$7D,$BF,$13,$F3,$C9,$3E,$97,$E9,$7C
	.dc.b	$4F,$C2,$3F,$C0,$3E,$05,$9F,$A9,$16,$FF,$90,$FC,$0D,$CB,$E8,$4F
	.dc.b	$2A,$F8,$8D,$AF,$7B,$5F,$9A,$D7,$EB,$35,$F9,$4D,$7E,$F7,$C4,$FE
	.dc.b	$1C,$93,$F4,$5E,$D5,$FE,$1B,$F0,$3C,$93,$FD,$2D,$B5,$FB,$BF,$82
	.dc.b	$BC,$9F,$D2,$9F,$C0,$FC,$4F,$E4,$C9,$7F,$75,$FC,$B7,$C4,$FE,$18
	.dc.b	$FF,$12,$F8,$36,$7F,$31,$17,$0F,$E8,$BF,$09,$73,$FC,$A4,$F2,$AF
	.dc.b	$C4,$36,$BF,$06,$D7,$F2,$2D,$7F,$34,$D7,$F1,$CD,$7F,$45,$F0,$BF
	.dc.b	$AB,$24,$FE,$55,$ED,$5F,$E3,$FF,$0F,$C9,$3F,$DB,$DB,$5F,$C7,$FE
	.dc.b	$12,$EB,$FD,$B9,$FD,$2F,$C4,$7E,$DC,$97,$FA,$17,$F7,$5F,$0D,$FA
	.dc.b	$A3,$FC,$93,$E1,$D9,$FD,$E4,$5C,$3F,$B4,$FC,$35,$CF,$F7,$13,$CA
	.dc.b	$BF,$58,$DA,$FD,$3B,$5F,$DA,$B5,$FD,$F3,$5F,$D9,$35,$FE,$0D,$F0
	.dc.b	$DF,$C3,$C9,$3F,$B9,$7B,$57,$F9,$8F,$C3,$F2,$4F,$F8,$4D,$B5,$FD
	.dc.b	$0F,$E1,$AE,$BF,$E1,$27,$F8,$4F,$C3,$7F,$1B,$25,$FF,$05,$7F,$8F
	.dc.b	$7C,$2F,$F0,$E3,$FC,$DB,$E3,$59,$FE,$41,$17,$0F,$F0,$6F,$C3,$5C
	.dc.b	$FF,$8E,$4F,$2A,$FE,$20,$DA,$FE,$16,$D7,$F8,$CB,$5F,$E4,$4D,$7F
	.dc.b	$8B,$35,$FE,$55,$F0,$BF,$CC,$C9,$3F,$C7,$5E,$D5,$FE,$87,$F1,$BC
	.dc.b	$93,$FE,$37,$6D,$7F,$67,$F8,$6B,$AF,$F8,$D9,$FE,$5B,$F0,$BF,$CF
	.dc.b	$C9,$7F,$CA,$5F,$E8,$5F,$09,$FC,$C8,$FE,$7D,$E3,$3F,$5E,$FE,$73
	.dc.b	$8F,$F3,$F9,$D5,$FF,$47,$7F,$A2,$63,$FE,$44,$F8,$7C,$D7,$F4,$93
	.dc.b	$F9,$A4,$F6,$AF,$E7,$0D,$8F,$E6,$6D,$7F,$A0,$B5,$FE,$94,$D7,$9C
	.dc.b	$72,$3F,$A1,$7C,$27,$F5,$B2,$38,$35,$1E,$3C,$7D,$87,$9B,$5F,$F7
	.dc.b	$93,$89,$F6,$1F,$E3,$6F,$87,$D3,$FE,$5C,$7F,$AB,$FC,$27,$F6,$92
	.dc.b	$47,$F5,$37,$FB,$67,$C2,$7F,$5C,$3E,$5E,$AF,$2B,$15,$FD,$BE,$6E
	.dc.b	$5F,$E5,$CF,$86,$B9,$FF,$6B,$9E,$75,$FD,$7D,$B5,$FD,$69,$AF,$F6
	.dc.b	$86,$BF,$DC,$5A,$FF,$65,$6B,$FD,$E7,$E1,$3F,$C0,$44,$9F,$ED,$8F
	.dc.b	$6A,$F8,$B8,$BC,$AF,$2D,$7F,$AB,$A3,$0F,$AC,$FF,$2E,$FC,$36,$B0
	.dc.b	$FF,$1D,$3F,$DF,$BE,$13,$FC,$20,$49,$FE,$F4,$FF,$84,$FE,$0F,$FC
	.dc.b	$06,$7C,$8F,$36,$CF,$F8,$56,$2D,$FE,$37,$95,$F6,$3F,$EE,$78,$FF
	.dc.b	$01,$4E,$9F,$F0,$5B,$D6,$BF,$C1,$0D,$8F,$F0,$1B,$5D,$C6,$88,$E9
	.dc.b	$AF,$F8,$88,$FF,$11,$FC,$3F,$FA,$8D,$C0,$70,$ED,$AF,$08,$F1,$6B
	.dc.b	$F3,$BC,$DF,$6B,$63,$E0,$1F,$F0,$07,$0E,$8E,$2B,$E1,$46,$1F,$D8
	.dc.b	$1E,$1F,$8B,$4F,$92,$3C,$F1,$87,$18,$23,$11,$EA,$8C,$4F,$18,$E8
	.dc.b	$D8,$CC,$72,$9E,$2F,$BD,$AF,$82,$78,$46,$23,$99,$C6,$3A,$7C,$0E
	.dc.b	$6B,$FD,$4D,$18,$24,$FA,$27,$44,$8C,$C7,$A6,$30,$EE,$32,$FD,$3F
	.dc.b	$80,$EC,$48,$F6,$36,$3E,$2B,$5F,$88,$D7,$94,$70,$23,$F3,$3D,$EF
	.dc.b	$51,$23,$E4,$36,$35,$18,$23,$82,$1F,$4E,$38,$3F,$57,$DD,$F8,$C4
	.dc.b	$9F,$A8,$D8,$FB,$AD,$78,$46,$88,$E9,$AF,$E0,$7D,$CF,$A4,$48,$FC
	.dc.b	$03,$63,$F1,$2D,$78,$46,$88,$E9,$AF,$E3,$7D,$BF,$BC,$48,$FC,$63
	.dc.b	$F8,$1F,$67,$FD,$0F,$99,$FE,$57,$E9,$E1,$95,$FE,$94,$DF,$F3,$35
	.dc.b	$B8,$FD,$47,$C4,$6B,$1F,$D2,$CE,$5F,$D0,$DF,$7A,$FC,$E3,$63,$31
	.dc.b	$82,$38,$21,$FA,$78,$E0,$FE,$A7,$D9,$FC,$D1,$27,$F5,$0F,$5C,$3F
	.dc.b	$96,$FA,$BE,$56,$CF,$AC,$FE,$11,$7F,$DB,$B8,$9F,$11,$B9,$FF,$59
	.dc.b	$F3,$7F,$22,$8F,$8F,$EC,$FE,$A4,$93,$FB,$77,$F5,$9E,$C7,$CC,$3F
	.dc.b	$AA,$FA,$D6,$7C,$1A,$FF,$1C,$9C,$3F,$C3,$BE,$22,$E9,$F3,$A7,$9D
	.dc.b	$7F,$03,$6D,$7F,$0C,$6B,$FC,$2D,$AF,$F2,$36,$BD,$E3,$91,$FC,$37
	.dc.b	$D8,$FE,$04,$47,$06,$A3,$FB,$E7,$7B,$D6,$F2,$CF,$F5,$56,$CB,$5F
	.dc.b	$E8,$FF,$0F,$86,$D7,$C0,$9F,$E7,$1E,$C7,$F1,$12,$47,$F1,$F7,$F8
	.dc.b	$A7,$AD,$F9,$43,$FB,$EF,$B1,$67,$97,$32,$71,$7F,$CD,$BE,$1A,$EF
	.dc.b	$F9,$69,$E7,$7F,$CA,$DB,$5F,$CD,$1A,$FF,$33,$6B,$FD,$2D,$AF,$F5
	.dc.b	$26,$BF,$D4,$7D,$6F,$E5,$44,$9F,$E7,$0F,$6A,$FE,$FB,$EC,$79,$27
	.dc.b	$FB,$E3,$6B,$E4,$73,$3E,$1B,$1B,$7F,$70,$7F,$6B,$EA,$FF,$37,$26
	.dc.b	$1F,$D1,$5F,$E7,$5E,$A7,$EE,$8F,$F1,$1F,$B1,$67,$E3,$C5,$C7,$FA
	.dc.b	$DF,$C3,$5D,$3F,$77,$3C,$AB,$FA,$83,$6B,$FA,$CB,$5F,$EB,$0D,$7F
	.dc.b	$B8,$35,$FE,$EE,$D7,$FB,$B7,$A7,$FD,$3C,$93,$FD,$75,$ED,$5F,$E2
	.dc.b	$2F,$67,$C9,$3F,$E2,$16,$D7,$C6,$F8,$5B,$AF,$F8,$51,$FE,$33,$E9
	.dc.b	$FF,$5B,$25,$FF,$69,$7F,$AF,$7A,$1F,$C7,$8F,$F2,$3F,$B3,$67,$F1
	.dc.b	$71,$70,$FE,$FF,$F0,$B7,$3F,$E3,$F3,$CA,$BF,$BA,$36,$BF,$BE,$B5
	.dc.b	$FE,$F8,$D7,$FC,$28,$D7,$FC,$36,$D5,$F8,$4B,$E2,$BF,$BC,$36,$BF
	.dc.b	$C0,$8D,$9F,$F0,$73,$5E,$F1,$FC,$33,$DA,$FF,$46,$7B,$9E,$98,$FF
	.dc.b	$25,$36,$7F,$1F,$F0,$F7,$7F,$F1,$EB,$FE,$17,$CF,$FC,$2B,$8F,$F8
	.dc.b	$7E,$BF,$C5,$ED,$7F,$C6,$8D,$47,$AB,$B3,$D3,$E5,$F3,$CB,$6F,$F9
	.dc.b	$DB,$DD,$C5,$70,$B4,$E2,$3E,$1B,$63,$35,$FE,$35,$F8,$6C,$4C,$8F
	.dc.b	$8A,$F8,$1A,$4A,$F8,$F1,$96,$C6,$5F,$6C,$65,$B1,$F4,$3C,$AF,$A6
	.dc.b	$47,$07,$5C,$BF,$CF,$3E,$F7,$8E,$BF,$93,$B6,$3F,$C8,$5A,$FF,$9A
	.dc.b	$BE,$1B,$17,$FD,$33,$B9,$1E,$FF,$2B,$F0,$24,$9F,$C0,$BF,$84,$F1
	.dc.b	$BE,$21,$FE,$A0,$F7,$AC,$FF,$2D,$DD,$38,$D3,$FC,$B3,$F0,$D7,$4F
	.dc.b	$BB,$3C,$EF,$F1,$AD,$AF,$C6,$B5,$FC,$0B,$5F,$C0,$B5,$FB,$4D,$7F
	.dc.b	$37,$E3,$7E,$68,$93,$F9,$A7,$B5,$7F,$A8,$3D,$FF,$24,$FF,$A7,$DB
	.dc.b	$5F,$C9,$3E,$1A,$EF,$FE,$92,$7F,$09,$E2,$FE,$98,$97,$FA,$67,$F5
	.dc.b	$3E,$1F,$E2,$4F,$F6,$2F,$BF,$67,$E3,$45,$C3,$FC,$FB,$F0,$D7,$3F
	.dc.b	$D0,$CF,$2A,$FD,$8B,$6B,$F6,$2D,$7F,$4A,$D7,$F4,$AD,$7F,$3C,$D7
	.dc.b	$F7,$FE,$17,$EF,$89,$3F,$BE,$7B,$57,$FB,$13,$DF,$F2,$4F,$FB,$0D
	.dc.b	$B5,$FD,$23,$E1,$AE,$BF,$EB,$67,$F5,$1E,$17,$F0,$A2,$5F,$F0,$A7
	.dc.b	$F8,$6F,$77,$EB,$4F,$D9,$F8,$0F,$D7,$FF,$D3,$75,$F8,$B9,$D3,$4F
	.dc.b	$F5,$37,$C3,$E3,$2F,$F5,$1F,$3A,$BA,$D7,$F1,$86,$CF,$F1,$86,$BF
	.dc.b	$C3,$1A,$FF,$0C,$6B,$FC,$15,$AF,$F2,$6E,$FF,$E4,$A4,$9F,$E4,$AF
	.dc.b	$6D,$BF,$D6,$19,$7A,$FE,$07,$CD,$18,$CF,$FD,$56,$FF,$1E,$C7,$FC
	.dc.b	$81,$F0,$FA,$C7,$F6,$8F,$F1,$1E,$DF,$E6,$64,$8F,$E6,$6F,$F3,$6E
	.dc.b	$BF,$E2,$C7,$CA,$C3,$FD,$57,$A7,$C4,$8C,$D7,$ED,$5F,$F2,$77,$C3
	.dc.b	$E2,$75,$3F,$CF,$A7,$A3,$49,$1F,$D0,$23,$2D,$8C,$BF,$CD,$A3,$2D
	.dc.b	$8F,$E9,$3D,$3F,$D3,$08,$E0,$EB,$9F,$C5,$F8,$2F,$23,$63,$D0,$7F
	.dc.b	$96,$39,$1F,$67,$FD,$02,$33,$82,$BF,$8F,$CE,$63,$F0,$BD,$3F,$D7
	.dc.b	$09,$5F,$D7,$1F,$EB,$FC,$DF,$CF,$CF,$95,$F0,$76,$7F,$8E,$45,$AE
	.dc.b	$D1,$FE,$74,$B7,$1E,$71,$FE,$B3,$7D,$2B,$FB,$4B,$6F,$FB,$4B,$5F
	.dc.b	$EB,$AD,$7F,$AE,$B5,$FE,$AC,$D7,$FB,$BF,$37,$F7,$62,$4F,$F7,$67
	.dc.b	$B5,$7E,$3F,$E0,$FC,$93,$FD,$9B,$0F,$D7,$46,$2F,$FC,$D1,$8B,$E2
	.dc.b	$31,$AF,$CE,$7F,$B1,$72,$7F,$81,$09,$8F,$F8,$11,$FF,$03,$F1,$7F
	.dc.b	$67,$3F,$25,$F0,$96,$7F,$A2,$45,$C7,$E4,$9E,$0F,$88,$C6,$BF,$E0
	.dc.b	$09,$E7,$5F,$E1,$06,$D7,$F8,$41,$AF,$F8,$11,$AF,$F8,$11,$AF,$F7
	.dc.b	$D6,$AF,$A3,$FC,$23,$FE,$08,$3F,$C1,$2E,$1C,$35,$C9,$E2,$AF,$DD
	.dc.b	$7C,$37,$A2,$FE,$67,$3A,$31,$5F,$33,$E2,$6E,$DF,$98,$7F,$C1,$D8
	.dc.b	$FE,$23,$0F,$15,$78,$38,$51,$88,$F2,$C6,$1C,$4F,$16,$EF,$D0,$FE
	.dc.b	$05,$F0,$96,$FF,$D4,$15,$F7,$5B,$3F,$E2,$2C,$7F,$57,$F1,$3B,$C3
	.dc.b	$EA,$9F,$1E,$7C,$0F,$9A,$F9,$3E,$0B,$F3,$3B,$6C,$6A,$38,$4F,$27
	.dc.b	$E6,$B5,$FA,$2D,$79,$1D,$48,$72,$9E,$BB,$EE,$8D,$24,$78,$AB,$F9
	.dc.b	$27,$C3,$FB,$77,$FE,$64,$D9,$FD,$D7,$C4,$5C,$7F,$46,$FD,$6D,$BB
	.dc.b	$12,$7D,$AF,$87,$3F,$B6,$7F,$31,$F8,$7B,$3F,$42,$2D,$FE,$83,$5F
	.dc.b	$E3,$3F,$11,$8C,$FE,$6E,$CB,$E9,$87,$CA,$6E,$1F,$80,$6B,$F4,$DA
	.dc.b	$FE,$61,$AF,$E7,$1A,$FE,$87,$3F,$90,$49,$FA,$EF,$6B,$FE,$AD,$F0
	.dc.b	$FE,$59,$FC,$EB,$6B,$F8,$3F,$C4,$5C,$FF,$99,$BF,$9C,$CF,$EA,$12
	.dc.b	$FF,$06,$FD,$78,$FD,$09,$FD,$5F,$E2,$2C,$FE,$5A,$2D,$FC,$7E,$9F
	.dc.b	$F8,$01,$39,$24,$7E,$8A,$7A,$47,$E4,$1B,$19,$8C,$11,$C1,$0F,$CE
	.dc.b	$47,$07,$F3,$DF,$0F,$F8,$D2,$4F,$E7,$1E,$B9,$FF,$7B,$F8,$7F,$3B
	.dc.b	$67,$CA,$FF,$52,$F8,$7C,$C7,$F4,$77,$F8,$09,$F9,$92,$3F,$A8,$7F
	.dc.b	$39,$F0,$BF,$C1,$8F,$EF,$9F,$0F,$67,$F7,$71,$71,$F8,$CD,$BF,$C5
	.dc.b	$BC,$A7,$F6,$6D,$9F,$E1,$2D,$7F,$78,$D7,$F9,$0B,$5F,$E4,$AD,$7F
	.dc.b	$95,$7C,$2F,$EC,$49,$3F,$C0,$1E,$D5,$FE,$1F,$F8,$7F,$24,$FF,$26
	.dc.b	$6D,$7F,$50,$AB,$C1,$FD,$F1,$FE,$4B,$F0,$9F,$BD,$25,$FF,$0C,$7F
	.dc.b	$80,$7C,$1F,$F2,$A3,$FC,$41,$F0,$F6,$7F,$8F,$C5,$C3,$F1,$55,$72
	.dc.b	$FD,$7C,$F2,$AF,$E2,$ED,$AF,$E5,$AD,$7F,$90,$35,$FE,$8A,$D7,$FA
	.dc.b	$4B,$5F,$E9,$DF,$05,$FC,$54,$93,$FC,$91,$ED,$5F,$E4,$3F,$87,$F2
	.dc.b	$4F,$F4,$A6,$D7,$F7,$4B,$B9,$FF,$88,$1F,$E9,$3F,$05,$FC,$84,$97
	.dc.b	$FC,$C1,$FE,$49,$F0,$1F,$D3,$8F,$F6,$FF,$CE,$3F,$0C,$46,$6B,$FA
	.dc.b	$2B,$71,$FD,$86,$97,$4F,$E2,$B3,$D2,$BF,$9E,$B6,$BF,$A9,$B5,$FE
	.dc.b	$8C,$D7,$FB,$5B,$5F,$ED,$ED,$7F,$BA,$7C,$07,$F3,$B2,$4F,$F4,$A7
	.dc.b	$B5,$7F,$61,$C5,$F8,$6F,$FA,$1A,$FF,$AB,$F7,$FB,$30,$47,$F8,$3A
	.dc.b	$B7,$1E,$B3,$FC,$71,$A5,$9F,$F1,$0B,$FD,$D3,$DF,$FE,$98,$49,$FE
	.dc.b	$B6,$D9,$FE,$E2,$D7,$FB,$23,$5F,$F0,$4B,$5F,$F0,$6B,$FD,$5F,$DE
	.dc.b	$93,$FD,$99,$B3,$FD,$F1,$AF,$F6,$F6,$BF,$E1,$46,$BF,$E1,$86,$BF
	.dc.b	$E1,$DF,$73,$FB,$21,$27,$FB,$9B,$63,$FC,$12,$D7,$FB,$E3,$5F,$F1
	.dc.b	$0B,$5F,$F1,$2B,$FD,$D7,$F0,$CF,$A1,$1F,$E1,$86,$BF,$E2,$B6,$BF
	.dc.b	$49,$A9,$FF,$50,$CE,$FD,$AF,$F4,$C7,$C5,$FF,$87,$76,$FF,$07,$D6
	.dc.b	$AF,$95,$66,$F8,$1F,$7A,$BF,$C4,$9F,$48,$FC,$59,$D5,$BC,$38,$DE
	.dc.b	$CE,$2F,$F8,$B7,$16,$3E,$D2,$4F,$F8,$72,$32,$D8,$CC,$F1,$6B,$F0
	.dc.b	$DA,$F6,$3E,$37,$C7,$FC,$92,$4F,$57,$9E,$5F,$EC,$8F,$D0,$BF,$C5
	.dc.b	$AF,$C2,$BF,$F7,$2C,$62,$BD,$33,$F0,$3F,$19,$C6,$1E,$E3,$FC,$DB
	.dc.b	$A2,$33,$1F,$47,$DC,$FB,$44,$BF,$73,$FE,$33,$5F,$2D,$F2,$9F,$91
	.dc.b	$3F,$E5,$48,$77,$7C,$D9,$DE,$D6,$73,$1F,$EB,$7F,$1F,$F6,$97,$FD
	.dc.b	$EF,$AB,$FC,$A1,$0E,$AB,$1F,$5E,$7F,$0D,$EE,$7E,$3C,$97,$F4,$DF
	.dc.b	$36,$1E,$59,$FF,$45,$CF,$A3,$2F,$16,$5E,$37,$AC,$78,$F0,$F0,$E3
	.dc.b	$8B,$FF,$2F,$F1,$3D,$AF,$CF,$12,$FF,$04,$F8,$3B,$FD,$C7,$BF,$BA
	.dc.b	$CF,$BF,$AB,$DB,$D5,$EE,$6C,$7C,$AC,$7F,$45,$EC,$FE,$AC,$91,$F8
	.dc.b	$F7,$B3,$F4,$0F,$A7,$87,$77,$1B,$A3,$B4,$7C,$FC,$2D,$FF,$8A,$5F
	.dc.b	$99,$9F,$FA,$56,$7F,$39,$EB,$FE,$EC,$8F,$E7,$FD,$6D,$9F,$A9,$F5
	.dc.b	$DE,$C3,$E4,$73,$5E,$BC,$17,$8C,$BF,$C7,$1E,$05,$EB,$FD,$DC,$FF
	.dc.b	$09,$25,$F7,$25,$7E,$CE,$32,$D8,$CB,$FC,$3E,$32,$D8,$FD,$B7,$AF
	.dc.b	$FC,$54,$8E,$0E,$B5,$FE,$A9,$CF,$D7,$F2,$3E,$59,$FF,$5C,$B6,$DE
	.dc.b	$43,$AB,$FA,$76,$1E,$AB,$FD,$1C,$E5,$F9,$BE,$BF,$F2,$52,$3F,$C2
	.dc.b	$5F,$CB,$F8,$BF,$A5,$EA,$FD,$33,$67,$9E,$1F,$C7,$EF,$C1,$E1,$7F
	.dc.b	$AD,$61,$E1,$D8,$FB,$7E,$AF,$F3,$32,$5F,$F1,$77,$BB,$F8,$6C,$7B
	.dc.b	$B4,$EF,$3C,$DF,$0D,$5D,$3F,$47,$EA,$7F,$3F,$24,$FF,$23,$7F,$63
	.dc.b	$DD,$FB,$F3,$F8,$07,$65,$9F,$5F,$85,$E9,$D3,$D4,$D8,$FC,$47,$A5
	.dc.b	$FD,$2C,$8F,$F2,$F7,$AB,$F9,$EB,$EA,$D3,$A9,$F2,$2B,$AB,$4F,$CA
	.dc.b	$7A,$3F,$D5,$C8,$FF,$3B,$7F,$AD,$7A,$1F,$A4,$3F,$3D,$E8,$3E,$47
	.dc.b	$15,$F9,$D9,$BA,$FF,$AE,$70,$FE,$0D,$39,$24,$7F,$4B,$9E,$B1,$FD
	.dc.b	$25,$B1,$98,$C1,$1C,$10,$FE,$D9,$1C,$1F,$ED,$D7,$FD,$C4,$93,$FD
	.dc.b	$55,$EB,$CB,$F9,$D5,$F2,$3E,$88,$C3,$67,$FC,$4B,$C9,$F9,$DF,$0F
	.dc.b	$18,$7E,$83,$24,$7F,$08,$F4,$BF,$C0,$04,$BF,$EC,$8F,$F3,$2F,$0B
	.dc.b	$FA,$07,$77,$F4,$16,$CF,$E7,$3C,$2F,$CD,$1F,$9B,$EF,$B1,$FC,$27
	.dc.b	$CE,$FF,$06,$11,$FE,$E0,$F5,$7F,$8F,$DF,$56,$9D,$49,$9F,$D5,$AE
	.dc.b	$AF,$F3,$CF,$0B,$FC,$30,$49,$FE,$F8,$FF,$5F,$FD,$FF,$F7,$B1,$FE
	.dc.b	$48,$0E,$25,$18,$EB,$FE,$C1,$39,$DF,$F8,$37,$85,$FB,$7F,$03,$F7
	.dc.b	$0D,$BF,$E4,$DE,$0F,$F8,$A4,$8F,$F8,$35,$F1,$3D,$0E,$B8,$57,$F8
	.dc.b	$2C,$F3,$C7,$F2,$4D,$E4,$F2,$C7,$F9,$82,$BF,$35,$5F,$E6,$28,$87
	.dc.b	$F8,$7B,$9F,$6E,$27,$6E,$8D,$4D,$7F,$C4,$BC,$BF,$DD,$67,$F1,$B3
	.dc.b	$FE,$04,$F0,$9C,$EF,$92,$D5,$CA,$FE,$C1,$7C,$30,$FD,$57,$85,$95
	.dc.b	$9F,$92,$75,$1E,$5B,$F3,$BE,$B8,$CB,$63,$2F,$C1,$6B,$E8,$BF,$54
	.dc.b	$F0,$8D,$46,$27,$73,$99,$E0,$7C,$37,$07,$B6,$79,$E1,$C0,$8F,$33
	.dc.b	$E2,$62,$F9,$2F,$76,$3F,$B5,$6C,$FE,$CB,$A9,$E0,$E3,$2F,$D9,$3F
	.dc.b	$3E,$BC,$A4,$AF,$5B,$F9,$9F,$0B,$F2,$7E,$0F,$E5,$1B,$3F,$D3,$3C
	.dc.b	$1F,$E9,$47,$F4,$BF,$02,$C7,$D4,$F9,$1F,$82,$47,$E3,$BD,$5F,$F0
	.dc.b	$97,$43,$0E,$D1,$87,$9E,$7F,$C9,$70,$ED,$3E,$13,$F0,$3F,$23,$F3
	.dc.b	$88,$FD,$47,$F6,$5A,$FE,$B0,$FD,$6F,$99,$67,$FB,$2F,$87,$FD,$8F
	.dc.b	$87,$FB,$23,$63,$F1,$BF,$23,$F6,$88,$FE,$01,$EA,$FA,$AF,$AB,$4E
	.dc.b	$A7,$F7,$2A,$EA,$D3,$F3,$3F,$23,$F8,$62,$3F,$8C,$7F,$13,$F2,$3F
	.dc.b	$A5,$3F,$A5,$EF,$67,$F4,$31,$75,$FF,$3A,$FB,$CF,$73,$1C,$DF,$A1
	.dc.b	$AE,$77,$F9,$86,$DF,$E4,$DA,$FE,$E5,$AF,$E9,$5A,$FF,$03,$6B,$FE
	.dc.b	$9E,$F9,$1F,$CF,$92,$7F,$4A,$F6,$AF,$E9,$5C,$1E,$49,$FF,$75,$F6
	.dc.b	$7E,$FA,$31,$6F,$5D,$F8,$FB,$E0,$73,$18,$7F,$84,$7C,$97,$EB,$89
	.dc.b	$8F,$ED,$1F,$E6,$5E,$6F,$F2,$CF,$2F,$F9,$6B,$67,$FC,$4B,$E6,$FF
	.dc.b	$88,$CF,$F1,$27,$99,$63,$F8,$6F,$C8,$7E,$E8,$8F,$F0,$07,$AB,$F9
	.dc.b	$AB,$EA,$D3,$A9,$34,$FE,$FD,$5D,$5F,$E2,$B1,$FC,$47,$E4,$7F,$83
	.dc.b	$91,$FE,$1A,$FF,$4B,$F3,$7F,$A2,$9F,$D1,$BC,$9B,$3F,$E3,$EF,$37
	.dc.b	$FC,$79,$E5,$FF,$8F,$5B,$1F,$C9,$BE,$47,$F8,$99,$1F,$E3,$2F,$57
	.dc.b	$F6,$77,$D5,$A7,$53,$FC,$97,$5D,$5A,$7F,$32,$F9,$1F,$E3,$E4,$7F
	.dc.b	$92,$BF,$C8,$BE,$47,$F9,$F1,$FE,$24,$F0,$2C,$FF,$3A,$8B,$A7,$C4
	.dc.b	$3F,$CC,$FF,$77,$FD,$2B,$A7,$F9,$D4,$EA,$7F,$A7,$3D,$6B,$F9,$9B
	.dc.b	$63,$F9,$73,$5D,$C6,$88,$E9,$AF,$F5,$9F,$92,$FE,$72,$48,$FE,$7E
	.dc.b	$F5,$AF,$F1,$27,$81,$E8,$6C,$FA,$CF,$F3,$DE,$1F,$D5,$63,$D3,$18
	.dc.b	$F3,$BF,$B0,$7C,$97,$F4,$62,$47,$F4,$C7,$FC,$15,$E9,$7F,$81,$3D
	.dc.b	$1F,$F0,$2B,$67,$AF,$A1,$FD,$DB,$0F,$EE,$FE,$9B,$2F,$E9,$9A,$FC
	.dc.b	$B8,$C7,$A1,$BC,$3F,$C6,$F5,$AB,$FE,$CD,$F2,$5F,$D6,$09,$3F,$D7
	.dc.b	$9F,$1B,$FC,$E3,$3F,$1B,$4F,$19,$31,$F0,$5F,$F5,$78,$C5,$FF,$73
	.dc.b	$F9,$2F,$EC,$A4,$8F,$ED,$6F,$F8,$A7,$D3,$FF,$11,$9F,$E2,$4F,$3E
	.dc.b	$CF,$CC,$F2,$FD,$58,$FE,$1B,$C6,$BB,$7C,$19,$B9,$A7,$FC,$1D,$F2
	.dc.b	$5F,$DD,$09,$87,$F7,$A7,$C0,$FF,$54,$CB,$C0,$D3,$C0,$7D,$C9,$EB
	.dc.b	$7D,$9F,$1E,$2F,$C4,$7F,$C8,$0F,$F8,$85,$AF,$F8,$A1,$AF,$F8,$AD
	.dc.b	$A8,$E5,$E4,$F4,$83,$FE,$8C,$F2,$F0,$CC,$38,$6F,$FE,$32,$AD,$C7
	.dc.b	$E4,$BA,$B5,$87,$EB,$4E,$B1,$FE,$33,$9D,$C7,$9E,$FF,$C3,$F1,$88
	.dc.b	$F4,$46,$09,$3C,$5B,$1C,$63,$04,$7E,$17,$C9,$74,$24,$78,$5E,$38
	.dc.b	$7F,$A3,$FC,$CF,$3B,$67,$DA,$7C,$7B,$FB,$BD,$5A,$8F,$CD,$BF,$1F
	.dc.b	$E4,$FD,$24,$8F,$73,$FE,$75,$F5,$BF,$CD,$FE,$9F,$F9,$C1,$B3,$FB
	.dc.b	$8D,$BF,$12,$7E,$2B,$2B,$1E,$2D,$3E,$B6,$AF,$93,$F8,$E4,$9F,$98
	.dc.b	$F7,$7F,$C3,$F8,$F7,$69,$DC,$95,$FA,$8B,$EE,$FD,$7B,$7C,$9F,$D4
	.dc.b	$25,$7D,$97,$FD,$21,$EA,$7F,$A2,$8F,$F4,$5F,$9B,$67,$F8,$9F,$9F
	.dc.b	$FA,$4F,$3B,$F4,$AD,$8F,$66,$3F,$99,$F9,$3F,$C0,$92,$3F,$0C,$F6
	.dc.b	$7F,$C7,$38,$76,$69,$D8,$FE,$43,$7D,$9A,$76,$CB,$FB,$E3,$F9,$1F
	.dc.b	$47,$F1,$E4,$7F,$29,$EA,$FE,$84,$FF,$75,$E5,$63,$94,$5C,$BF,$87
	.dc.b	$F9,$F7,$9B,$F4,$53,$F9,$A2,$57,$42,$57,$ED,$63,$2D,$8C,$BF,$B7
	.dc.b	$8C,$B6,$3F,$51,$F2,$7F,$A6,$23,$83,$AD,$7F,$BA,$73,$F2,$47,$F8
	.dc.b	$83,$44,$62,$FF,$AF,$F9,$D7,$C0,$FE,$74,$FF,$01,$F9,$3F,$D9,$93
	.dc.b	$0F,$DC,$3F,$EC,$FF,$4B,$FD,$93,$E8,$FF,$B2,$9B,$3F,$D4,$BC,$FF
	.dc.b	$E4,$67,$F2,$4F,$3E,$C7,$D3,$F5,$FF,$7E,$47,$F8,$23,$D5,$FF,$3B
	.dc.b	$DF,$56,$9D,$49,$9F,$F4,$0A,$EA,$FE,$67,$E2,$FF,$85,$92,$7F,$88
	.dc.b	$3F,$EE,$7F,$4B,$FD,$C2,$7F,$3C,$87,$C0,$7F,$80,$5F,$F7,$57,$BA
	.dc.b	$9F,$E7,$BE,$A7,$F3,$EF,$4E,$D7,$F7,$07,$F9,$47,$C5,$FF,$1A,$24
	.dc.b	$7F,$1F,$7C,$0F,$FA,$4A,$BC,$18,$77,$CB,$FD,$EB,$8F,$7A,$FE,$F7
	.dc.b	$8F,$76,$9D,$CF,$CE,$7C,$4F,$F2,$82,$3F,$CB,$5F,$E5,$7F,$17,$FC
	.dc.b	$EC,$FE,$B3,$E4,$7F,$B7,$BA,$F1,$5F,$A5,$9B,$8F,$F8,$8F,$DC,$7B
	.dc.b	$3E,$58,$FE,$9D,$3A,$7F,$A6,$3D,$AB,$F9,$E3,$63,$FA,$93,$5D,$C6
	.dc.b	$88,$E9,$AF,$F5,$7F,$88,$FE,$86,$48,$FE,$90,$F5,$EC,$47,$E7,$7A
	.dc.b	$7D,$31,$8A,$FE,$B3,$D8,$9C,$D7,$F5,$FC,$9F,$92,$7F,$B6,$23,$E1
	.dc.b	$1F,$CB,$E3,$3E,$8F,$F6,$3F,$89,$FE,$AA,$49,$FE,$B4,$FF,$BE,$1F
	.dc.b	$C4,$FB,$9F,$70,$FB,$BE,$DD,$8F,$F1,$AF,$B5,$FE,$06,$F6,$7F,$C0
	.dc.b	$ED,$8F,$E2,$DF,$11,$FD,$88,$8F,$F6,$77,$B7,$47,$F8,$F6,$3F,$BC
	.dc.b	$1E,$03,$FA,$31,$F9,$AB,$F0,$6F,$FE,$24,$BE,$CD,$9F,$F2,$3E,$1F
	.dc.b	$DF,$FE,$27,$FB,$89,$23,$FB,$B3,$E0,$F7,$7F,$2F,$EC,$FE,$61,$B3
	.dc.b	$FE,$50,$F6,$BF,$C4,$27,$F8,$93,$EF,$7F,$C2,$7C,$BF,$E4,$6F,$71
	.dc.b	$3F,$E0,$DF,$8A,$FE,$FE,$4B,$FF,$02,$3D,$DF,$5E,$EC,$BC,$11,$87
	.dc.b	$BB,$E1,$3F,$B3,$79,$4C,$BC,$31,$FE,$50,$C3,$FC,$39,$7F,$E0,$E2
	.dc.b	$47,$F8,$4D,$FC,$46,$DF,$E2,$33,$FC,$49,$F8,$97,$E3,$A7,$FA,$BB
	.dc.b	$63,$FD,$11,$DC,$EA,$79,$3F,$E3,$4A,$D6,$FD,$1B,$87,$56,$E0,$D3
	.dc.b	$D1,$8F,$1A,$87,$8E,$78,$3E,$66,$B9,$7C,$EE,$5A,$E5,$57,$A5,$E8
	.dc.b	$D7,$99,$CD,$C3,$C8,$C4,$75,$79,$B6,$1F,$15,$49,$1E,$18,$CB,$4F
	.dc.b	$E2,$79,$3C,$8F,$7D,$FC,$C8,$C4,$FD,$03,$F1,$9E,$EE,$B3,$FF,$48
	.dc.b	$EE,$8F,$A1,$F1,$5E,$32,$57,$85,$FD,$4F,$C0,$7F,$30,$3F,$98,$FC
	.dc.b	$05,$9F,$C4,$DB,$DF,$FC,$D7,$BB,$F9,$B6,$CF,$E0,$BE,$2B,$E0,$11
	.dc.b	$F7,$3D,$9F,$CC,$E1,$D9,$A7,$69,$E7,$AF,$FA,$9F,$0F,$04,$F8,$B0
	.dc.b	$FC,$57,$C4,$FC,$D2,$47,$CB,$7F,$81,$FC,$17,$F4,$AF,$81,$FE,$96
	.dc.b	$D9,$F5,$FB,$FF,$B3,$3F,$69,$EF,$58,$FC,$DF,$C5,$7D,$92,$3F,$5D
	.dc.b	$EA,$FE,$CA,$FA,$B4,$EA,$4C,$FD,$55,$D5,$FF,$40,$72,$BE,$27,$F0
	.dc.b	$C4,$AF,$C2,$3F,$89,$F8,$9F,$D4,$1F,$84,$F4,$AC,$FE,$C2,$2D,$FC
	.dc.b	$7F,$80,$BC,$FF,$AD,$9E,$55,$F8,$F6,$D7,$E6,$DA,$FE,$81,$AF,$EC
	.dc.b	$DA,$FF,$00,$6B,$FA,$8F,$47,$F3,$C4,$9F,$CE,$3D,$AB,$FA,$87,$67
	.dc.b	$F0,$3F,$5F,$CB,$5F,$E3,$16,$DF,$EF,$F9,$1E,$FE,$3B,$FF,$8C,$3F
	.dc.b	$C2,$7E,$27,$F5,$84,$C3,$F5,$4F,$F4,$2F,$84,$FF,$04,$9F,$E0,$AF
	.dc.b	$82,$B3,$F9,$AF,$82,$FE,$65,$F0,$3F,$CC,$DB,$1F,$C6,$BE,$27,$F7
	.dc.b	$04,$7F,$6C,$F5,$7F,$98,$DF,$56,$9D,$4F,$D3,$D7,$56,$9F,$C9,$BE
	.dc.b	$27,$F8,$21,$1F,$E0,$4F,$F5,$AF,$84,$FF,$10,$FB,$9F,$E2,$26,$CF
	.dc.b	$EC,$BE,$0F,$FA,$51,$FD,$2F,$E0,$AC,$7F,$32,$F6,$7F,$88,$11,$FE
	.dc.b	$1C,$F5,$7F,$A4,$DF,$56,$9D,$49,$C5,$FB,$1A,$EA,$FE,$63,$E2,$7F
	.dc.b	$8E,$12,$7F,$8D,$3F,$C7,$FE,$27,$F9,$F1,$FC,$6B,$D7,$B3,$FD,$1A
	.dc.b	$2D,$7E,$F7,$E1,$2F,$4F,$F4,$39,$E5,$5F,$C9,$5B,$5F,$CC,$5A,$FF
	.dc.b	$36,$6B,$FD,$25,$AF,$F5,$66,$BF,$CF,$BE,$27,$F9,$A1,$27,$F9,$93
	.dc.b	$DA,$BF,$2F,$E7,$3D,$9F,$2D,$7F,$A1,$DB,$7F,$97,$F8,$3B,$D5,$FD
	.dc.b	$B5,$FE,$BD,$7F,$D0,$49,$7F,$CF,$DF,$F0,$BF,$C0,$FF,$97,$8F,$F2
	.dc.b	$FF,$A9,$67,$F9,$7F,$C1,$FF,$81,$FE,$0F,$FC,$10,$D8,$FE,$DF,$F1
	.dc.b	$3F,$D3,$48,$FF,$4B,$7A,$BF,$E0,$6B,$EA,$D3,$A9,$FC,$F2,$BA,$B4
	.dc.b	$FE,$F7,$F1,$1F,$D6,$48,$FF,$57,$7F,$C5,$FE,$BF,$F9,$DB,$C8,$FF
	.dc.b	$3B,$B6,$7F,$A3,$FC,$1F,$F8,$7C,$FF,$10,$7C,$15,$8F,$F0,$3F,$C4
	.dc.b	$7F,$65,$23,$FD,$8D,$EA,$FF,$87,$AF,$AB,$4E,$A4,$E1,$FE,$8D,$5D
	.dc.b	$5F,$E5,$BF,$11,$FD,$C4,$93,$FD,$BD,$FE,$ED,$F8,$97,$F9,$E4,$FF
	.dc.b	$47,$79,$5F,$E1,$7F,$7F,$FC,$CD,$CF,$FE,$20,$9D,$7C,$66,$30,$FF
	.dc.b	$11,$47,$FA,$57,$E2,$1F,$93,$36,$D6,$81,$5C,$AF,$FC,$20,$B0,$F1
	.dc.b	$92,$32,$D9,$FF,$0D,$4E,$E7,$24,$8C,$CE,$DF,$F1,$24,$ED,$E3,$F1
	.dc.b	$3F,$00,$95,$E2,$7B,$65,$FC,$DB,$EA,$BF,$C8,$B5,$E9,$BF,$8B,$18
	.dc.b	$8F,$5E,$1E,$EC,$DE,$4F,$FA,$1E,$58,$7C,$9A,$D4,$E6,$3E,$67,$C5
	.dc.b	$7D,$22,$67,$F1,$1F,$F3,$DF,$A3,$FE,$C2,$3F,$D8,$7E,$1D,$9F,$F0
	.dc.b	$9F,$C5,$7F,$9A,$FE,$23,$FC,$D8,$D8,$FA,$18,$7D,$9F,$8A,$FB,$E4
	.dc.b	$8F,$A2,$F6,$7F,$CD,$78,$76,$69,$D8,$FF,$10,$DF,$66,$9D,$8F,$C1
	.dc.b	$FC,$4F,$E3,$08,$FD,$E7,$FD,$2B,$E2,$7F,$B7,$39,$7F,$DB,$AD,$9F
	.dc.b	$F1,$6F,$C4,$FF,$A1,$4F,$C4,$F8,$AF,$C0,$3F,$CC,$6F,$F9,$E6,$3F
	.dc.b	$1B,$C5,$F8,$CF,$85,$FC,$D9,$2F,$F1,$AF,$81,$FF,$43,$E3,$E0,$69
	.dc.b	$E0,$24,$FD,$87,$F3,$17,$E1,$7C,$87,$BE,$3F,$09,$F0,$FF,$AA,$24
	.dc.b	$7E,$71,$FD,$6F,$C1,$F9,$BA,$0F,$FB,$BB,$A1,$D9,$98,$97,$D7,$0F
	.dc.b	$D3,$C7,$C2,$D7,$FC,$CB,$1E,$53,$F2,$91,$C1,$FD,$2B,$E0,$9F,$D6
	.dc.b	$B6,$33,$18,$23,$82,$1F,$A8,$8E,$0F,$EA,$BE,$07,$F8,$39,$27,$F7
	.dc.b	$2F,$5E,$4F,$D4,$7C,$4F,$8D,$B3,$EB,$CF,$E0,$BF,$EB,$2C,$5F,$13
	.dc.b	$B8,$F9,$F1,$87,$87,$57,$F1,$62,$3F,$C2,$1F,$F7,$67,$F0,$07,$ED
	.dc.b	$A0,$D3,$E6,$BF,$EE,$07,$F7,$7C,$1F,$8B,$F8,$AF,$99,$F1,$1F,$35
	.dc.b	$B5,$E7,$F7,$7F,$92,$91,$FE,$2E,$F8,$37,$FE,$3B,$1E,$06,$E1,$E0
	.dc.b	$7E,$B4,$64,$F8,$15,$DA,$3E,$D6,$1E,$39,$F8,$98,$7E,$1F,$D9,$FE
	.dc.b	$6A,$48,$FE,$4E,$FC,$0D,$FB,$FC,$57,$81,$B3,$F3,$7E,$23,$F1,$07
	.dc.b	$E2,$7E,$1A,$C7,$D3,$F6,$BF,$A0,$91,$FE,$66,$F5,$7E,$45,$F5,$69
	.dc.b	$D4,$99,$FA,$B8,$3F,$CF,$51,$8E,$3F,$E9,$1E,$87,$F4,$E2,$47,$F4
	.dc.b	$07,$FA,$8F,$A7,$FA,$53,$F4,$98,$59,$FC,$44,$5C,$3B,$7C,$2D,$C3
	.dc.b	$FA,$A4,$F2,$AF,$E9,$6D,$AF,$E8,$ED,$7F,$A1,$B5,$FE,$B2,$D7,$F9
	.dc.b	$F3,$5F,$EA,$FE,$87,$F6,$62,$4F,$F5,$87,$B5,$7F,$47,$F8,$AF,$24
	.dc.b	$FF,$5F,$EB,$76,$27,$38,$7F,$BE,$F5,$7C,$36,$39,$7F,$8F,$3F,$C1
	.dc.b	$FD,$1F,$EE,$A4,$CB,$FB,$3B,$FA,$0C,$3F,$1E,$7E,$43,$E2,$AC,$FE
	.dc.b	$B7,$E1,$7F,$84,$7C,$27,$F0,$96,$C7,$EC,$FC,$9F,$F0,$09,$1F,$EE
	.dc.b	$6F,$57,$F5,$17,$D5,$A7,$53,$F7,$55,$D5,$A7,$EF,$FC,$CF,$F0,$69
	.dc.b	$1F,$EF,$EF,$EE,$3E,$2B,$F4,$FF,$13,$FA,$86,$CF,$F0,$5F,$85,$FE
	.dc.b	$40,$7F,$21,$F8,$4B,$1F,$C2,$FC,$CF,$F0,$C9,$1F,$F0,$5B,$D5,$FE
	.dc.b	$01,$7D,$5A,$75,$27,$17,$E6,$6B,$AB,$FC,$B4,$FF,$0B,$F9,$5F,$E2
	.dc.b	$62,$4F,$F8,$61,$3F,$E1,$D8,$83,$BF,$FC,$0C,$7F,$82,$B9,$BF,$07
	.dc.b	$C3,$FD,$62,$33,$34,$FE,$7D,$F0,$76,$FF,$A1,$4F,$F8,$A5,$EB,$3F
	.dc.b	$E2,$B8,$C3,$86,$BE,$26,$BE,$56,$BC,$4F,$49,$C4,$A7,$12,$9C,$4A
	.dc.b	$71,$8F,$1B,$EF,$32,$57,$31,$FC,$8E,$3C,$97,$A8,$C4,$79,$63,$0F
	.dc.b	$F6,$5D,$7E,$13,$FD,$7E,$3F,$57,$18,$E8,$F4,$F3,$7F,$81,$7E,$0B
	.dc.b	$18,$FF,$4A,$7E,$76,$3D,$C9,$1E,$47,$FA,$4F,$C4,$7F,$38,$3F,$9C
	.dc.b	$FC,$45,$9F,$EA,$DF,$09,$FD,$C7,$E0,$FF,$B9,$36,$3E,$AF,$83,$EB
	.dc.b	$23,$F0,$1E,$AF,$F5,$1B,$EA,$D3,$A9,$FD,$8E,$BA,$B4,$FC,$0F,$47
	.dc.b	$C7,$23,$F3,$1F,$EC,$3F,$11,$FD,$43,$E2,$3F,$A8,$B6,$7F,$B7,$7C
	.dc.b	$27,$F8,$24,$FF,$05,$7C,$1D,$8F,$C6,$F2,$FD,$32,$3F,$61,$EA,$FF
	.dc.b	$69,$BE,$AD,$3A,$93,$93,$FA,$15,$75,$7F,$55,$BB,$8F,$F0,$04,$AF
	.dc.b	$C2,$BF,$83,$E8,$FD,$21,$FE,$25,$2C,$7E,$89,$B5,$FE,$21,$F7,$5F
	.dc.b	$34,$CF,$3C,$DF,$9A,$C9,$87,$5C,$7F,$26,$DC,$7F,$50,$D7,$F5,$0D
	.dc.b	$7F,$50,$D7,$F6,$0D,$7F,$7D,$C3,$F9,$B2,$4F,$E8,$1E,$D7,$FE,$8B
	.dc.b	$7C,$B1,$FD,$17,$C4,$46,$29,$9B,$F1,$67,$1E,$63,$0F,$F0,$9E,$6F
	.dc.b	$D5,$13,$1F,$D7,$BF,$E2,$4F,$8A,$FF,$09,$9F,$E1,$4F,$8A,$B3,$FE
	.dc.b	$31,$F8,$3F,$F2,$67,$C2,$FF,$93,$5B,$1F,$C4,$39,$FF,$6C,$47,$F7
	.dc.b	$6F,$57,$FC,$59,$7D,$5A,$75,$3F,$C7,$D5,$D5,$A7,$F1,$CE,$6F,$E0
	.dc.b	$44,$7F,$83,$BF,$E3,$CF,$8A,$FF,$15,$FC,$57,$F8,$B1,$B3,$FE,$71
	.dc.b	$EE,$FC,$07,$C2,$FF,$A1,$B9,$FF,$D7,$DC,$BF,$C1,$17,$C3,$66,$7F
	.dc.b	$98,$69,$FC,$44,$98,$7F,$16,$7C,$0F,$F9,$1F,$2F,$03,$4F,$05,$F6
	.dc.b	$7F,$D0,$AD,$E4,$65,$F9,$3C,$7C,$15,$FC,$EB,$83,$F9,$01,$1F,$E4
	.dc.b	$AF,$F2,$3D,$3F,$A2,$1F,$E8,$5E,$1B,$3F,$D0,$A2,$E1,$FE,$48,$C5
	.dc.b	$F0,$D8,$DB,$F4,$15,$CE,$FF,$98,$B6,$FF,$A3,$35,$FE,$8C,$D7,$FA
	.dc.b	$33,$5F,$E9,$CD,$7F,$AE,$69,$FC,$E4,$93,$FC,$F9,$ED,$5F,$EE,$9E
	.dc.b	$1F,$24,$FE,$7D,$B5,$F1,$B1,$7C,$2E,$36,$FD,$D3,$FD,$97,$E2,$BF
	.dc.b	$A3,$13,$0F,$E9,$6F,$FA,$00,$FF,$37,$9F,$E7,$0F,$8A,$B1,$F5,$BE
	.dc.b	$1B,$F0,$9F,$0B,$F8,$56,$C7,$F6,$C9,$FE,$A4,$47,$FA,$BB,$D5,$FF
	.dc.b	$44,$57,$56,$9D,$4F,$C1,$D7,$56,$9F,$DE,$3E,$2B,$FA,$F1,$1F,$EC
	.dc.b	$6F,$FA,$60,$FF,$43,$FC,$4F,$FA,$21,$B1,$F8,$EF,$87,$FC,$E1,$F9
	.dc.b	$DF,$87,$B1,$FE,$04,$AF,$ED,$44,$7F,$B7,$BD,$5F,$F5,$03,$D5,$A7
	.dc.b	$52,$6D,$F8,$B9,$EA,$FF,$38,$F8,$9F,$EE,$C4,$9F,$EF,$6F,$F7,$9F
	.dc.b	$89,$FF,$0D,$1F,$ED,$BE,$BB,$3F,$E1,$88,$B5,$FC,$DF,$C3,$FE,$3F
	.dc.b	$39,$24,$7F,$86,$67,$A4,$7F,$82,$1B,$19,$8C,$11,$C1,$0F,$F1,$04
	.dc.b	$70,$7F,$C6,$7F,$8A,$FF,$8D,$A7,$FC,$2E,$D8,$FF,$15,$B5,$DC,$78
	.dc.b	$39,$7F,$C4,$BE,$77,$F9,$EB,$B7,$D7,$18,$F1,$51,$8A,$5F,$F3,$AC
	.dc.b	$9D,$99,$8C,$3F,$E3,$0C,$3F,$C5,$F9,$7F,$BF,$27,$CF,$F1,$DF,$E2
	.dc.b	$A8,$C4,$7F,$93,$5A,$FF,$94,$3E,$19,$BF,$FB,$40,$FF,$68,$ED,$73
	.dc.b	$FE,$2F,$F1,$7F,$C9,$7E,$27,$F9,$33,$63,$CF,$9F,$8B,$E2,$BC,$04
	.dc.b	$8F,$8E,$FF,$BB,$B0,$E9,$97,$FB,$73,$26,$58,$C3,$F9,$86,$3D,$6B
	.dc.b	$F8,$B3,$63,$DF,$F1,$3E,$D2,$65,$F5,$1F,$F2,$6A,$FE,$01,$FA,$D3
	.dc.b	$FE,$56,$8C,$6D,$FE,$F0,$9C,$C7,$F9,$95,$47,$F4,$3F,$8B,$93,$FC
	.dc.b	$E3,$E2,$AE,$5F,$E7,$DE,$07,$C5,$FC,$F2,$57,$E1,$9F,$1E,$3D,$AB
	.dc.b	$FC,$FE,$F6,$9F,$92,$78,$4F,$26,$3D,$B2,$FE,$A5,$97,$F5,$5C,$B1
	.dc.b	$5F,$6F,$E2,$FE,$F1,$1F,$CA,$BF,$81,$F8,$BF,$C3,$47,$E0,$5B,$93
	.dc.b	$A3,$FD,$9B,$E4,$E2,$1B,$7F,$BC,$9F,$75,$66,$31,$9F,$F6,$5F,$8B
	.dc.b	$C4,$CC,$7F,$4A,$F8,$5A,$4B,$FD,$3C,$65,$B1,$97,$F5,$F1,$96,$C7
	.dc.b	$EC,$3E,$2B,$F3,$84,$70,$75,$BF,$D6,$79,$5E,$3A,$FF,$40,$76,$23
	.dc.b	$13,$FD,$F7,$E2,$ED,$FF,$72,$7D,$DF,$11,$FA,$92,$47,$F0,$07,$ED
	.dc.b	$F5,$7E,$24,$FC,$57,$35,$9F,$EF,$BF,$17,$FE,$0B,$F8,$BF,$F0,$63
	.dc.b	$63,$EF,$E9,$FB,$EE,$07,$C4,$FE,$DC,$93,$FC,$39,$EE,$F0,$3F,$39
	.dc.b	$8F,$81,$C7,$42,$3B,$C7,$F8,$3B,$1B,$87,$F7,$F7,$ED,$7C,$47,$F0
	.dc.b	$52,$4F,$F1,$C7,$F2,$FD,$5F,$A5,$E7,$FD,$33,$67,$FC,$33,$F1,$5F
	.dc.b	$E2,$73,$FC,$51,$F1,$56,$3D,$B8,$FE,$2B,$E2,$7F,$89,$12,$3F,$94
	.dc.b	$BD,$9F,$4E,$1D,$9A,$76,$24,$FF,$85,$EB,$B3,$FB,$1F,$88,$FE,$3C
	.dc.b	$49,$FE,$6A,$FF,$21,$F8,$7F,$CC,$1F,$A0,$EC,$74,$3D,$5C,$DF,$A3
	.dc.b	$8D,$78,$C7,$FD,$0D,$F3,$6E,$EC,$75,$7F,$30,$AE,$D8,$7F,$44,$6D
	.dc.b	$FF,$44,$6B,$FD,$29,$AF,$F3,$36,$BF,$83,$6B,$FB,$2F,$88,$FE,$74
	.dc.b	$49,$FE,$A2,$F6,$AF,$E9,$5E,$8F,$92,$7F,$D0,$8D,$AF,$F4,$1F,$CD
	.dc.b	$3E,$23,$1F,$00,$8C,$5F,$C1,$F8,$9F,$E9,$04,$7F,$AF,$BF,$C6,$BC
	.dc.b	$3F,$E5,$07,$F2,$9F,$0E,$CF,$F9,$73,$E3,$3F,$CD,$5F,$19,$FE,$6B
	.dc.b	$6C,$7C,$DF,$83,$FE,$A8,$47,$FB,$5B,$D5,$FC,$DD,$F5,$69,$D4,$FF
	.dc.b	$34,$D7,$56,$9F,$67,$E2,$BF,$B0,$11,$FE,$EE,$FF,$32,$F1,$7F,$A0
	.dc.b	$78,$7F,$D0,$5B,$3F,$E7,$4F,$8C,$FF,$41,$1F,$C9,$BD,$E7,$E0,$9F
	.dc.b	$DD,$27,$47,$E1,$FE,$33,$FB,$71,$1F,$F0,$2B,$DD,$FD,$AE,$1D,$DA
	.dc.b	$72,$E3,$FF,$39,$61,$DE,$FB,$46,$27,$F2,$9F,$0F,$FD,$EC,$8F,$F8
	.dc.b	$49,$FE,$FD,$F0,$FF,$CD,$4F,$CF,$77,$3E,$03,$15,$F9,$D9,$B6,$E9
	.dc.b	$4F,$F8,$43,$5F,$7F,$81,$A9,$FF,$06,$DF,$6A,$FF,$10,$B6,$FF,$C4
	.dc.b	$2D,$7F,$C4,$ED,$7F,$C2,$0D,$7F,$90,$35,$7E,$2C,$F8,$8F,$1C,$7F
	.dc.b	$88,$89,$19,$AC,$46,$27,$6E,$27,$4F,$92,$7F,$97,$E1,$FC,$03,$E0
	.dc.b	$7D,$91,$FE,$E9,$6C,$FF,$AC,$36,$7C,$5E,$34,$FC,$0B,$FD,$97,$1F
	.dc.b	$F6,$1E,$1E,$2B,$FF,$1B,$C6,$1C,$35,$F2,$7C,$4F,$F8,$0B,$CC,$FF
	.dc.b	$08,$1F,$E1,$1F,$2A,$DF,$FB,$4F,$2F,$D9,$DF,$D5,$8F,$A9,$0F,$8E
	.dc.b	$FC,$37,$83,$F8,$77,$C9,$F1,$5F,$0A,$FE,$3B,$5F,$2B,$5F,$9E,$F8
	.dc.b	$EF,$FA,$4E,$3E,$36,$9E,$3B,$EC,$D3,$CD,$7F,$0F,$E2,$3A,$69,$FE
	.dc.b	$B2,$9F,$89,$8F,$DC,$9C,$9F,$E2,$2F,$2F,$FC,$61,$E3,$FF,$8C,$5B
	.dc.b	$5F,$07,$E2,$7F,$44,$7E,$8F,$E3,$AC,$7F,$4C,$C5,$F1,$5E,$02,$47
	.dc.b	$82,$32,$FD,$53,$FC,$85,$7D,$B2,$F9,$87,$F8,$C9,$C7,$47,$E4,$39
	.dc.b	$3F,$7F,$97,$82,$B7,$1F,$C0,$B2,$C5,$7E,$43,$17,$C5,$7C,$22,$65
	.dc.b	$E3,$7E,$2F,$C5,$7E,$24,$FD,$AF,$C5,$59,$F6,$45,$CB,$CB,$E5,$BE
	.dc.b	$37,$FD,$0B,$5A,$24,$FE,$3A,$BA,$C7,$C1,$6C,$66,$30,$47,$04,$3F
	.dc.b	$51,$1C,$1F,$D5,$7C,$5F,$DC,$24,$FC,$D7,$AE,$7F,$E9,$9F,$86,$F4
	.dc.b	$36,$7C,$AF,$EE,$B1,$7C,$C7,$FA,$E3,$0F,$A2,$7F,$00,$C7,$FC,$A7
	.dc.b	$3A,$CB,$F5,$DF,$0B,$F8,$C2,$63,$F7,$5F,$F3,$5F,$A5,$FE,$79,$3F
	.dc.b	$CF,$5E,$85,$9F,$D5,$7C,$57,$F2,$BB,$FE,$58,$D8,$FC,$57,$C4,$FE
	.dc.b	$60,$8F,$E2,$5E,$AF,$F8,$B2,$FA,$B4,$EA,$7E,$DE,$BA,$B4,$FC,$B7
	.dc.b	$C3,$7E,$90,$8F,$E5,$5F,$F4,$1F,$A3,$FE,$91,$F3,$FF,$D2,$4D,$9F
	.dc.b	$E0,$5F,$15,$FD,$0C,$FE,$89,$F1,$B6,$3F,$45,$F1,$1F,$B0,$23,$FA
	.dc.b	$17,$AB,$FE,$44,$BE,$AD,$3A,$93,$87,$F3,$BE,$22,$FB,$47,$D6,$E9
	.dc.b	$7C,$47,$EF,$89,$7F,$AF,$7F,$80,$FC,$3F,$F2,$53,$FA,$8F,$C5,$59
	.dc.b	$FD,$C4,$5B,$EF,$F1,$B7,$A7,$F5,$53,$CA,$BF,$74,$DA,$FE,$56,$D7
	.dc.b	$F9,$F3,$5F,$E8,$AD,$7F,$99,$B5,$FE,$9B,$F1,$1F,$C5,$89,$3F,$C1
	.dc.b	$DE,$D5,$FE,$F6,$FD,$8F,$FD,$6D,$5F,$D5,$2F,$FC,$91,$18,$CD,$1F
	.dc.b	$DD,$FE,$3B,$D5,$8F,$F3,$67,$FA,$37,$C4,$7F,$24,$24,$7F,$15,$7F
	.dc.b	$D8,$7E,$CF,$FB,$44,$FF,$69,$7A,$F6,$7F,$A1,$7C,$0F,$F8,$0F,$E3
	.dc.b	$3F,$C0,$8D,$8F,$E4,$5F,$11,$FC,$C0,$8F,$F2,$17,$AB,$FE,$84,$BE
	.dc.b	$AD,$3A,$9F,$D3,$EB,$AB,$4F,$E5,$DF,$0F,$FC,$F0,$8F,$F2,$D7,$FD
	.dc.b	$B7,$EC,$FF,$B2,$A1,$F7,$9F,$DC,$38,$7F,$DD,$DF,$18,$AD,$4E,$6B
	.dc.b	$FA,$DF,$C4,$FF,$87,$CF,$F1,$07,$C3,$D8,$FE,$7D,$F0,$BF,$D3,$08
	.dc.b	$FF,$3D,$7C,$2F,$FA,$8E,$BA,$72,$FF,$BB,$30,$EF,$EF,$73,$8F,$CE
	.dc.b	$E3,$DA,$FF,$B7,$3F,$DE,$FE,$13,$FA,$C9,$23,$FA,$53,$FD,$73,$E0
	.dc.b	$BF,$C0,$07,$E0,$BE,$3A,$CF,$F5,$38,$B7,$FD,$33,$67,$D6,$BB,$73
	.dc.b	$1F,$D1,$EB,$A5,$FF,$57,$6D,$FF,$82,$1A,$FF,$86,$DA,$FF,$88,$9A
	.dc.b	$FF,$84,$1A,$BF,$17,$FF,$24,$BF,$E2,$66,$BF,$E2,$B6,$A3,$81,$E2
	.dc.b	$AF,$EC,$5F,$81,$33,$F4,$E1,$FE,$45,$8C,$3F,$04,$FF,$6D,$CB,$F6
	.dc.b	$1F,$D3,$63,$FE,$3A,$89,$1F,$AC,$C5,$F0,$BF,$E1,$36,$DF,$F8,$21
	.dc.b	$B9,$7F,$8B,$5A,$FF,$90,$E3,$F1,$FF,$07,$F8,$CF,$82,$FC,$6B,$63
	.dc.b	$FC,$55,$F0,$DF,$E6,$B3,$FC,$D9,$F0,$D6,$3F,$DA,$B8,$7F,$8F,$9F
	.dc.b	$07,$C3,$78,$C8,$FF,$82,$79,$BF,$C7,$9A,$FF,$95,$23,$F3,$F9,$78
	.dc.b	$1F,$C9,$3D,$73,$FE,$57,$97,$82,$7C,$67,$83,$E1,$3E,$01,$1F,$80
	.dc.b	$FF,$8D,$DF,$D5,$FC,$1F,$EA,$0F,$D4,$FC,$15,$8F,$9F,$FB,$37,$E8
	.dc.b	$32,$3E,$68,$FB,$91,$8E,$9F,$A5,$18,$9D,$65,$F8,$1D,$BF,$C9,$93
	.dc.b	$F4,$7E,$17,$EA,$11,$F8,$59,$78,$6B,$FC,$53,$1F,$C0,$32,$C6,$5F
	.dc.b	$7F,$1F,$21,$D9,$F1,$4F,$A3,$87,$C7,$7E,$1F,$83,$FC,$11,$27,$F0
	.dc.b	$4F,$E1,$7E,$0B,$73,$F9,$D7,$F9,$9F,$C9,$63,$54,$52,$5F,$F9,$AF
	.dc.b	$E0,$AD,$FF,$84,$2B,$73,$F8,$FA,$EC,$D2,$4F,$E4,$63,$2D,$8C,$BF
	.dc.b	$AD,$8C,$B6,$3F,$35,$ED,$FE,$74,$8E,$0E,$B9,$FF,$4F,$F9,$3F,$2B
	.dc.b	$63,$F2,$B7,$FE,$C3,$F8,$2C,$C7,$FA,$09,$F8,$1E,$DF,$EA,$49,$1F
	.dc.b	$A9,$7F,$15,$CF,$FE,$54,$8F,$E4,$DF,$03,$FC,$CB,$E1,$3F,$99,$BF
	.dc.b	$4B,$1F,$DC,$78,$77,$B8,$DE,$1C,$3F,$3F,$DB,$FD,$C1,$31,$FD,$C3
	.dc.b	$E0,$C7,$AC,$7F,$35,$6F,$07,$F2,$5C,$3A,$CF,$F0,$86,$C6,$70,$EF
	.dc.b	$3F,$B6,$F6,$3F,$83,$11,$FE,$0C,$FB,$BE,$13,$FA,$69,$FD,$3B,$E0
	.dc.b	$AC,$FF,$AA,$FC,$6F,$F7,$D7,$E1,$3F,$ED,$E8,$C1,$FA,$63,$FD,$C0
	.dc.b	$FF,$15,$F6,$BF,$8A,$12,$3F,$8A,$3D,$DF,$EA,$17,$DD,$A7,$4C,$BF
	.dc.b	$D0,$F5,$DD,$FF,$16,$DF,$8A,$7F,$31,$EB,$7F,$21,$23,$FC,$85,$FE
	.dc.b	$49,$EA,$7E,$74,$FF,$08,$FC,$A5,$9F,$E3,$31,$6F,$FA,$87,$73,$C5
	.dc.b	$C7,$27,$FB,$93,$B1,$7D,$30,$FE,$5C,$DC,$3F,$97,$35,$FE,$90,$D7
	.dc.b	$FA,$43,$5F,$E1,$0D,$7F,$A8,$7A,$7F,$CE,$C9,$3F,$CE,$DE,$D7,$FE
	.dc.b	$2D,$F9,$4F,$2C,$FF,$51,$F2,$9E,$DA,$73,$82,$DF,$B3,$BA,$73,$18
	.dc.b	$7F,$81,$7A,$3F,$D2,$C9,$97,$F4,$B7,$F4,$7F,$0B,$FE,$12,$F8,$4F
	.dc.b	$F0,$9B,$67,$F0,$58,$BE,$25,$CD,$F9,$4C,$BF,$D8,$8D,$AF,$C3,$F5
	.dc.b	$FE,$33,$BB,$F8,$97,$AF,$FD,$6C,$93,$FD,$6D,$F0,$BF,$E1,$7C,$FC
	.dc.b	$2D,$3C,$38,$F8,$2B,$A4,$63,$06,$18,$CB,$F8,$D7,$9F,$FD,$A4,$95
	.dc.b	$FD,$A5,$FE,$03,$F0,$9F,$E3,$03,$FC,$63,$F0,$56,$7F,$61,$E4,$7E
	.dc.b	$FF,$AF,$F8,$03,$63,$FB,$2F,$A7,$FD,$D4,$8F,$F7,$57,$AB,$FE,$31
	.dc.b	$BE,$AD,$3A,$9F,$BD,$AE,$AD,$3F,$B8,$F9,$7F,$E0,$12,$3F,$E0,$17
	.dc.b	$FC,$09,$E4,$FF,$3D,$3F,$CE,$3F,$23,$67,$FB,$CC,$5D,$3F,$5F,$95
	.dc.b	$E3,$FD,$04,$F2,$AF,$F0,$63,$6B,$FC,$18,$D7,$FC,$4C,$D7,$FC,$4C
	.dc.b	$D7,$FB,$33,$5F,$F1,$77,$83,$FE,$19,$24,$FF,$86,$5E,$D5,$FE,$8B
	.dc.b	$F9,$2F,$24,$FE,$7F,$AD,$18,$BF,$C8,$71,$35,$C7,$27,$E4,$9F,$EC
	.dc.b	$3E,$07,$F0,$BA,$FF,$1C,$1F,$E3,$97,$0E,$1A,$E7,$1F,$E8,$7F,$05
	.dc.b	$FE,$72,$F8,$4F,$F3,$9B,$63,$FC,$17,$E9,$FF,$43,$F1,$0D,$F4,$F0
	.dc.b	$F8,$BA,$95,$E2,$C7,$C1,$18,$70,$7F,$13,$70,$EF,$2F,$E7,$EF,$F8
	.dc.b	$EF,$CE,$FF,$55,$D7,$F9,$EF,$3F,$2B,$4F,$29,$27,$F8,$FE,$3E,$48
	.dc.b	$FD,$CC,$7F,$21,$EB,$EE,$FC,$36,$BF,$1D,$AE,$E3,$FC,$7D,$F0,$7F
	.dc.b	$E9,$63,$FD,$2F,$F0,$56,$3F,$65,$E2,$7F,$4A,$E9,$FE,$96,$D8,$F0
	.dc.b	$E8,$E9,$EA,$D8,$F8,$78,$FF,$35,$9C,$9F,$DD,$B1,$ED,$1F,$E2,$CD
	.dc.b	$59,$63,$9F,$F8,$8E,$5D,$EB,$FA,$6B,$63,$E6,$E5,$F8,$9C,$BB,$B6
	.dc.b	$BE,$9B,$5F,$A5,$C7,$E6,$BF,$C6,$9F,$ED,$EF,$92,$B7,$EC,$8B,$5F
	.dc.b	$DE,$5B,$87,$81,$E5,$3E,$06,$CF,$81,$AF,$E0,$5A,$FE,$5D,$AF,$E6
	.dc.b	$DA,$FF,$5B,$D7,$E5,$92,$7D,$AF,$6A,$FF,$7C,$7C,$97,$92,$7F,$5C
	.dc.b	$DA,$FE,$95,$E8,$2F,$17,$FE,$97,$7F,$5B,$D3,$F6,$09,$87,$CA,$7F
	.dc.b	$C5,$7F,$07,$FE,$DA,$F8,$2F,$F6,$DB,$67,$F9,$B4,$FF,$83,$8F,$F0
	.dc.b	$7E,$96,$3F,$0A,$FE,$10,$8F,$D6,$7A,$BF,$ED,$BA,$EA,$D3,$A9,$32
	.dc.b	$FF,$06,$CF,$57,$FB,$34,$7F,$01,$CF,$F2,$24,$7F,$06,$FF,$91,$3E
	.dc.b	$03,$FD,$E6,$7F,$BD,$3D,$FB,$3F,$E8,$CF,$57,$F8,$17,$AD,$FC,$0D
	.dc.b	$B1,$F9,$EF,$90,$FC,$E9,$1F,$C7,$BD,$5F,$F7,$A5,$F5,$69,$D4,$FF
	.dc.b	$4C,$D7,$56,$9F,$AA,$F9,$1F,$D4,$91,$FC,$DB,$FA,$CF,$90,$FE,$46
	.dc.b	$7F,$A2,$BD,$1F,$F0,$4E,$18,$AF,$D9,$CD,$D3,$E2,$8F,$8F,$7B,$AF
	.dc.b	$6F,$5C,$BF,$AA,$BE,$D8,$7E,$A9,$B8,$7E,$A9,$AF,$F1,$B6,$BF,$CC
	.dc.b	$5A,$FF,$35,$6B,$FE,$35,$F9,$1F,$E0,$A4,$9F,$DB,$BD,$AB,$F9,$56
	.dc.b	$7F,$E2,$4D,$3C,$B5,$FD,$25,$B7,$EB,$C9,$ED,$E3,$4F,$6B,$FD,$23
	.dc.b	$E3,$FF,$89,$93,$0F,$E0,$AF,$F2,$4F,$BA,$FF,$76,$61,$F1,$9F,$F1
	.dc.b	$A4,$FF,$89,$1C,$E7,$FC,$F5,$CC,$FF,$9E,$F9,$FF,$D1,$07,$FA,$27
	.dc.b	$9A,$DF,$F2,$4F,$91,$FE,$46,$47,$F8,$AB,$E1,$A1,$FE,$C4,$BE,$B8
	.dc.b	$77,$8C,$D7,$8D,$6D,$D9,$F1,$47,$FA,$2F,$0C,$57,$F2,$4F,$91,$FE
	.dc.b	$66,$49,$FE,$4A,$F8,$7E,$23,$C1,$66,$FC,$46,$1F,$F4,$4F,$57,$FA
	.dc.b	$2B,$9A,$E1,$FE,$35,$7F,$BA,$FC,$8F,$F4,$02,$47,$F3,$27,$B3,$E9
	.dc.b	$AE,$AD,$3F,$C7,$D5,$D5,$A7,$53,$FA,$6F,$C7,$7F,$4C,$23,$FC,$F9
	.dc.b	$FE,$9F,$F1,$FF,$E0,$63,$F4,$9E,$93,$A7,$15,$FD,$5A,$6E,$9F,$E8
	.dc.b	$8F,$84,$BB,$FF,$4A,$9E,$75,$FD,$29,$B5,$FD,$29,$AF,$F7,$B6,$BF
	.dc.b	$E1,$16,$BF,$E1,$56,$BF,$E9,$5F,$8E,$FE,$CA,$49,$FE,$AE,$F6,$AF
	.dc.b	$D0,$E2,$EB,$F2,$D7,$F8,$A5,$B7,$FE,$85,$F8,$1F,$DD,$4F,$B4,$FF
	.dc.b	$3D,$DF,$FA,$23,$B7,$FC,$28,$FF,$84,$BF,$1A,$7C,$6F,$F8,$09,$B5
	.dc.b	$FD,$D9,$B8,$7F,$86,$9A,$FF,$8C,$3A,$3F,$2F,$F1,$BF,$88,$F8,$CF
	.dc.b	$C4,$B6,$FF,$CF,$1E,$0F,$FB,$40,$FF,$68,$F1,$58,$FF,$0B,$E2,$F8
	.dc.b	$DF,$11,$23,$C4,$FE,$96,$FA,$65,$FB,$50,$E3,$4F,$F4,$AE,$3D,$67
	.dc.b	$FD,$A3,$8D,$9F,$81,$E6,$FB,$C9,$8F,$BD,$FD,$97,$C6,$FE,$88,$FD
	.dc.b	$1F,$89,$67,$FD,$ED,$D1,$F2,$9F,$DB,$DB,$CC,$76,$67,$3F,$ED,$F7
	.dc.b	$6B,$53,$99,$6F,$FE,$4E,$C3,$E8,$7A,$FF,$4C,$8F,$F9,$2B,$8B,$F8
	.dc.b	$2E,$9E,$3B,$F8,$5A,$5C,$BE,$EE,$8D,$3C,$0F,$93,$1F,$49,$F8,$47
	.dc.b	$17,$FD,$10,$F1,$FA,$BF,$83,$24,$FE,$0D,$FC,$37,$95,$E9,$D4,$FE
	.dc.b	$BB,$C6,$C3,$2B,$EE,$BF,$0F,$C8,$7C,$6F,$FB,$CA,$72,$4C,$7F,$1D
	.dc.b	$5D,$A3,$F2,$4D,$8C,$C6,$08,$E0,$87,$F9,$CF,$0F,$2C,$6A,$31,$E5
	.dc.b	$7E,$7C,$91,$F9,$F7,$B6,$BF,$B0,$F2,$3D,$55,$FD,$A3,$C8,$FF,$7A
	.dc.b	$C6,$27,$FC,$FD,$9B,$E4,$FF,$DE,$98,$FA,$E7,$31,$8C,$3E,$76,$4F
	.dc.b	$0F,$F6,$04,$BF,$D8,$3F,$D0,$F5,$FE,$67,$F1,$3F,$CD,$1B,$3F,$BC
	.dc.b	$F6,$DF,$1D,$FE,$97,$3F,$D3,$1F,$13,$67,$F5,$B1,$FB,$D2,$3F,$BD
	.dc.b	$7B,$3F,$D2,$EF,$B3,$4E,$D1,$CF,$2F,$F0,$4E,$8C,$18,$66,$3F,$86
	.dc.b	$3F,$C2,$C9,$7F,$C2,$DF,$EB,$DE,$17,$F4,$E3,$FA,$7F,$9F,$67,$FD
	.dc.b	$FD,$E2,$FE,$A7,$DF,$7C,$03,$9F,$0F,$F2,$3A,$FD,$89,$FD,$A2,$77
	.dc.b	$3F,$95,$F8,$8F,$E3,$84,$9F,$E3,$8F,$89,$FE,$D5,$8F,$89,$A7,$88
	.dc.b	$EB,$AF,$FB,$52,$BC,$38,$F8,$9F,$D1,$7C,$2F,$F2,$92,$57,$F2,$97
	.dc.b	$F9,$67,$C3,$FE,$9C,$FE,$09,$F5,$2F,$C7,$3F,$DC,$35,$FC,$8A,$74
	.dc.b	$D3,$F6,$DE,$3E,$32,$FC,$4C,$F4,$9F,$E6,$CD,$8F,$E6,$CD,$7F,$44
	.dc.b	$D7,$FD,$CF,$CE,$8C,$47,$F8,$42,$31,$3D,$0F,$E8,$5F,$0B,$FD,$14
	.dc.b	$8E,$0D,$4F,$F9,$33,$A9,$F3,$AF,$03,$D1,$CD,$FA,$68,$C5,$FB,$6F
	.dc.b	$F1,$7E,$0F,$F7,$58,$C5,$7F,$1A,$F2,$7F,$87,$7C,$2F,$F5,$52,$4F
	.dc.b	$F5,$57,$FC,$51,$F2,$3F,$E1,$83,$FC,$33,$F2,$16,$7E,$CF,$47,$E2
	.dc.b	$BC,$2F,$C4,$FB,$3F,$8A,$6C,$FF,$52,$F8,$5F,$EC,$44,$7F,$B1,$3D
	.dc.b	$9F,$F1,$86,$1D,$9A,$76,$EC,$57,$4D,$BF,$CD,$F8,$F8,$6B,$58,$FF
	.dc.b	$29,$F8,$4F,$EE,$04,$8F,$EE,$0F,$F9,$0F,$E4,$3F,$C6,$7F,$21,$FE
	.dc.b	$34,$6C,$FF,$01,$EF,$FE,$99,$E1,$1B,$DB,$87,$F9,$B7,$C1,$FF,$7D
	.dc.b	$25,$7F,$7D,$7B,$3F,$E4,$BC,$3B,$34,$EC,$4D,$3F,$80,$E0,$C3,$B4
	.dc.b	$FF,$7A,$F8,$3F,$F0,$51,$1F,$F0,$53,$FE,$0D,$F8,$2F,$E9,$07,$F8
	.dc.b	$D3,$5B,$3F,$E0,$28,$B8,$7E,$2F,$E7,$DE,$E6,$3A,$3F,$C3,$15,$CE
	.dc.b	$FF,$C2,$CD,$BF,$F0,$B3,$5F,$E8,$4D,$7F,$42,$D7,$F4,$0D,$7F,$B9
	.dc.b	$FC,$27,$F8,$8C,$93,$FE,$23,$7B,$57,$F8,$D3,$E0,$7C,$93,$FE,$78
	.dc.b	$F3,$DE,$F2,$73,$87,$FA,$BF,$CB,$FE,$30,$FF,$42,$CB,$FB,$24,$63
	.dc.b	$16,$3F,$E8,$2A,$7E,$37,$FE,$80,$70,$7F,$A1,$9C,$38,$6B,$C5,$AB
	.dc.b	$D7,$FF,$46,$FC,$97,$F9,$F8,$FF,$3F,$FC,$95,$BF,$F0,$27,$9B,$FE
	.dc.b	$08,$F3,$7F,$C1,$2D,$8F,$E4,$78,$79,$37,$F3,$BE,$43,$11,$5A,$67
	.dc.b	$2F,$F4,$BC,$7F,$86,$E1,$8F,$85,$FF,$1E,$63,$73,$FF,$0D,$E3,$E1
	.dc.b	$9F,$E0,$ED,$8F,$0C,$7C,$6F,$84,$AF,$C1,$6B,$F1,$DA,$EF,$0F,$17
	.dc.b	$C9,$7F,$A8,$3E,$4B,$FD,$42,$D8,$FD,$E7,$0F,$F5,$03,$FA,$8F,$C3
	.dc.b	$58,$F8,$7A,$BE,$1F,$AB,$63,$E1,$37,$1F,$85,$3F,$EB,$FC,$3B,$47
	.dc.b	$F9,$6D,$B9,$FC,$CF,$07,$FB,$0E,$5D,$EB,$F8,$5F,$86,$CF,$17,$F0
	.dc.b	$B2,$FF,$31,$E7,$E0,$F8,$2F,$19,$1F,$1F,$C2,$7F,$94,$64,$FF,$A9
	.dc.b	$AF,$0C,$9F,$6B,$FC,$3B,$D1,$C6,$5F,$E4,$7B,$E3,$5D,$09,$1F,$7E
	.dc.b	$BE,$19,$23,$31,$F9,$78,$C3,$B8,$CB,$F9,$98,$F9,$7F,$01,$27,$E5
	.dc.b	$B6,$78,$1F,$E3,$5E,$A7,$C3,$F9,$B4,$FF,$54,$B6,$FF,$CE,$FF,$0E
	.dc.b	$F6,$71,$87,$FA,$A5,$F8,$9F,$05,$F6,$89,$5F,$69,$FF,$7B,$7C,$97
	.dc.b	$FB,$98,$FF,$73,$FC,$95,$9F,$E8,$58,$FF,$85,$BC,$2F,$F0,$BB,$63
	.dc.b	$F0,$FE,$FF,$E1,$88,$FE,$19,$EA,$FF,$BF,$6F,$AB,$4E,$A7,$F4,$FA
	.dc.b	$EA,$D3,$F2,$7E,$FF,$E4,$88,$FE,$49,$FD,$1F,$82,$FC,$17,$F8,$7C
	.dc.b	$FF,$87,$A3,$1F,$27,$FE,$99,$75,$16,$53,$FE,$8C,$F4,$64,$FF,$A2
	.dc.b	$FC,$9B,$C1,$F8,$5F,$80,$FD,$19,$27,$F4,$6F,$8B,$0F,$0C,$B8,$FF
	.dc.b	$85,$E5,$E2,$BF,$25,$6A,$72,$F8,$A3,$FD,$35,$8F,$82,$7F,$D4,$38
	.dc.b	$D9,$FE,$45,$C5,$FB,$42,$3F,$B4,$7F,$6F,$E9,$FE,$8C,$FF,$5C,$78
	.dc.b	$DF,$CE,$B7,$C5,$7E,$FA,$6E,$9F,$E1,$FF,$9C,$FE,$33,$39,$24,$7F
	.dc.b	$1F,$9E,$B1,$FC,$09,$B1,$98,$C1,$1C,$10,$FE,$79,$1C,$1F,$E7,$DE
	.dc.b	$B7,$F0,$E2,$4F,$F0,$E7,$AE,$7F,$CD,$31,$7D,$F7,$FB,$EA,$3F,$83
	.dc.b	$61,$EE,$8C,$5F,$F4,$78,$FF,$5B,$7C,$EF,$F0,$AC,$7F,$7D,$18,$F0
	.dc.b	$3F,$84,$7A,$DF,$C7,$89,$7F,$C7,$9F,$B9,$F2,$DF,$87,$3F,$11,$F2
	.dc.b	$D6,$7F,$D3,$1F,$03,$FE,$18,$F7,$3F,$C3,$2D,$8F,$E5,$7E,$C7,$F2
	.dc.b	$A2,$3F,$CA,$9E,$AF,$E1,$2F,$AB,$4E,$A7,$FA,$96,$BA,$B4,$FE,$6F
	.dc.b	$E8,$7F,$36,$23,$FC,$D9,$FC,$A7,$CA,$FE,$7F,$E5,$7F,$40,$D9,$FF
	.dc.b	$25,$FE,$1B,$F2,$37,$79,$A8,$C3,$7E,$4B,$FD,$35,$1B,$9F,$EF,$7E
	.dc.b	$A7,$F4,$62,$5F,$F4,$67,$C2,$FE,$83,$1F,$0B,$4F,$0F,$C9,$F4,$3A
	.dc.b	$E1,$63,$FA,$9F,$A3,$FD,$48,$8F,$F5,$27,$FA,$AF,$A7,$FD,$04,$FF
	.dc.b	$7A,$7E,$1A,$F9,$4C,$57,$F5,$C9,$B9,$FF,$9D,$BD,$4B,$D7,$FD,$3E
	.dc.b	$79,$D7,$F5,$E6,$D7,$F7,$F6,$BF,$E1,$56,$BF,$E1,$56,$BF,$E1,$B6
	.dc.b	$BF,$E2,$0F,$23,$FB,$59,$27,$FB,$5B,$DA,$BF,$C2,$23,$F2,$8F,$EC
	.dc.b	$93,$E9,$3F,$D7,$31,$9F,$53,$D3,$0B,$FE,$21,$F9,$4B,$9B,$31,$87
	.dc.b	$FB,$37,$9D,$FD,$F0,$95,$FD,$F1,$FE,$35,$F3,$3F,$C9,$7E,$67,$F9
	.dc.b	$33,$67,$FD,$93,$F1,$8F,$C8,$7F,$8C,$63,$FE,$A7,$6E,$9F,$E2,$A9
	.dc.b	$D7,$67,$B2,$FC,$D1,$C7,$CF,$C6,$1E,$9F,$C7,$BF,$94,$BF,$E4,$46
	.dc.b	$BF,$E4,$96,$BB,$AF,$39,$FC,$B3,$3F,$3B,$4F,$3E,$5E,$47,$C2,$D9
	.dc.b	$61,$E6,$9D,$E1,$E4,$96,$1E,$E7,$C9,$1F,$E3,$08,$C4,$79,$63,$FA
	.dc.b	$17,$CD,$FF,$4D,$3F,$A7,$7C,$CD,$BF,$B3,$F1,$0F,$8C,$FF,$4F,$FC
	.dc.b	$67,$FA,$81,$B3,$FE,$54,$C9,$B7,$72,$47,$F9,$4E,$32,$FF,$95,$FD
	.dc.b	$4F,$EC,$59,$77,$D3,$FA,$94,$FE,$16,$7F,$B3,$67,$9E,$0F,$C7,$E9
	.dc.b	$E3,$AF,$2E,$9E,$36,$9E,$3F,$1F,$E0,$11,$F1,$3F,$0F,$C4,$F9,$1A
	.dc.b	$1F,$E6,$9D,$F8,$65,$7B,$5F,$C0,$F0,$BF,$28,$F5,$ED,$E8,$3F,$09
	.dc.b	$3C,$1F,$F3,$A6,$1D,$6F,$E1,$4F,$85,$A4,$9D,$C6,$5B,$19,$7F,$1D
	.dc.b	$19,$7F,$21,$E1,$7D,$D2,$4F,$CF,$79,$6B,$FD,$4F,$CD,$FE,$1D,$3E
	.dc.b	$B3,$F6,$51,$9F,$03,$D0,$FE,$8F,$26,$B8,$AF,$D1,$C6,$1F,$C0,$74
	.dc.b	$7E,$34,$8F,$DF,$7F,$C1,$DF,$39,$FE,$19,$F9,$4F,$F0,$D3,$67,$F8
	.dc.b	$2F,$C8,$FF,$08,$D7,$E5,$64,$8C,$3F,$C3,$9F,$F2,$6C,$F0,$EF,$FB
	.dc.b	$3D,$9F,$9D,$24,$7E,$3D,$F0,$BF,$E1,$AC,$7C,$2D,$39,$13,$0F,$0D
	.dc.b	$F6,$6D,$FE,$1B,$A3,$F5,$44,$AF,$CE,$3F,$E2,$BF,$91,$FF,$1B,$1F
	.dc.b	$E3,$7F,$88,$B3,$FC,$4B,$F2,$D7,$C9,$7E,$29,$FF,$7D,$46,$63,$1F
	.dc.b	$23,$F9,$4E,$CF,$DC,$12,$7F,$54,$F7,$7F,$C6,$B8,$77,$69,$CC,$F0
	.dc.b	$4F,$7A,$EB,$1F,$B7,$E0,$FE,$08,$47,$F6,$CF,$F0,$7D,$BF,$4A,$7F
	.dc.b	$56,$F9,$CB,$3F,$BF,$8B,$B7,$D8,$ED,$7C,$9E,$38,$FF,$D8,$BD,$6B
	.dc.b	$E9,$87,$F0,$36,$E1,$FA,$66,$BF,$CA,$DA,$FF,$2B,$6B,$FC,$C5,$AF
	.dc.b	$F2,$0D,$7F,$8F,$12,$7F,$88,$3D,$AF,$F4,$5F,$3B,$E5,$9F,$EA,$5E
	.dc.b	$CF,$F8,$DA,$31,$7F,$E6,$3F,$86,$7B,$38,$C1,$18,$C3,$F9,$07,$27
	.dc.b	$F2,$D2,$3F,$C7,$DF,$F3,$77,$C7,$FF,$9E,$4F,$F3,$D7,$C0,$D9,$FE
	.dc.b	$AB,$F0,$7F,$D8,$DB,$A7,$F0,$E8,$C6,$7F,$C3,$7C,$34,$7F,$BE,$A3
	.dc.b	$F8,$81,$FD,$03,$0F,$E7,$C4,$BF,$E6,$0F,$89,$FF,$3D,$BE,$26,$9E
	.dc.b	$28,$F0,$5F,$85,$B1,$FC,$F7,$0F,$E9,$44,$7F,$9E,$3F,$E8,$BF,$82
	.dc.b	$FF,$4B,$7A,$5F,$E9,$76,$CF,$EF,$BE,$07,$FB,$4E,$DF,$BD,$6F,$B5
	.dc.b	$FE,$3B,$8C,$4F,$F4,$BA,$FE,$B0,$47,$FA,$53,$DD,$FF,$4A,$DF,$76
	.dc.b	$9D,$C9,$8F,$F0,$CB,$ED,$DE,$C7,$FA,$0D,$FF,$65,$25,$FF,$57,$7F
	.dc.b	$B4,$77,$FF,$46,$3F,$CF,$1F,$2D,$67,$FB,$04,$5C,$7E,$B7,$E7,$7F
	.dc.b	$E7,$9D,$3F,$43,$39,$9F,$F0,$33,$D2,$BF,$B2,$36,$3F,$A3,$B5,$DC
	.dc.b	$68,$8E,$9A,$FF,$84,$7E,$23,$FB,$F1,$23,$FB,$83,$D6,$91,$F8,$FF
	.dc.b	$17,$FD,$E4,$FA,$8F,$E8,$66,$7E,$17,$D9,$3F,$E2,$5F,$CF,$1E,$2E
	.dc.b	$2B,$FC,$4B,$18,$7F,$C0,$5B,$7F,$83,$C8,$FF,$80,$5F,$F6,$86,$7F
	.dc.b	$ED,$BC,$BF,$DB,$8D,$9F,$F0,$7F,$CB,$FE,$1B,$C0,$37,$E5,$70,$FF
	.dc.b	$87,$FD,$27,$BD,$EC,$E0,$FF,$26,$12,$33,$86,$23,$13,$B7,$13,$A7
	.dc.b	$5E,$9F,$FB,$D7,$2F,$24,$7F,$7D,$8C,$5F,$FB,$93,$3F,$1B,$E8,$3F
	.dc.b	$C4,$F9,$62,$BF,$6B,$83,$D2,$F0,$47,$95,$AF,$F9,$8E,$35,$5E,$78
	.dc.b	$C4,$E9,$FF,$16,$B6,$77,$1B,$6B,$EC,$8D,$12,$7A,$47,$26,$CF,$08
	.dc.b	$F9,$1E,$7F,$52,$4E,$9F,$3E,$2F,$89,$FD,$54,$FE,$22,$92,$E2,$FF
	.dc.b	$5B,$3B,$E0,$FF,$34,$B9,$DB,$FC,$57,$FA,$1B,$CF,$A6,$39,$2E,$5D
	.dc.b	$72,$FF,$26,$F8,$BF,$04,$91,$EA,$7C,$BA,$79,$27,$EC,$69,$E2,$8C
	.dc.b	$3F,$87,$CF,$C5,$7F,$E8,$6C,$BC,$4D,$CB,$C4,$D3,$C5,$B3,$A7,$E9
	.dc.b	$12,$3E,$5B,$72,$FA,$8D,$7F,$1E,$D7,$F2,$CD,$7F,$04,$FD,$DF,$2A
	.dc.b	$4F,$D6,$6C,$FD,$F6,$BF,$9A,$6B,$FA,$16,$BF,$8D,$6B,$FE,$76,$FA
	.dc.b	$0F,$C5,$12,$7F,$06,$D8,$FC,$63,$5F,$D3,$B5,$FD,$73,$5F,$CD,$3F
	.dc.b	$95,$F9,$F9,$3F,$8E,$6C,$FE,$5D,$AF,$EC,$9A,$FE,$E5,$AF,$E9,$5A
	.dc.b	$FF,$A3,$BE,$7F,$F4,$44,$9F,$CD,$B6,$3F,$48,$D7,$F7,$ED,$7F,$83
	.dc.b	$35,$FD,$93,$FA,$DF,$99,$93,$FA,$66,$CF,$EB,$DA,$FF,$0A,$6B,$FC
	.dc.b	$45,$AF,$EF,$5A,$FF,$A9,$BD,$0F,$DD,$12,$7F,$66,$D8,$FD,$E3,$5F
	.dc.b	$E2,$ED,$7F,$8F,$35,$FE,$14,$FF,$05,$F4,$64,$FE,$F9,$B3,$FC,$1D
	.dc.b	$AF,$F2,$26,$BF,$CA,$5A,$FF,$15,$6B,$FE,$BE,$F9,$6F,$E2,$44,$9F
	.dc.b	$E1,$6D,$8F,$E2,$8D,$7F,$97,$B5,$FE,$6C,$D7,$F9,$13,$FC,$77,$E2
	.dc.b	$A4,$FF,$16,$6C,$FF,$1F,$6B,$FC,$E9,$AF,$F4,$16,$BF,$CB,$5A,$FF
	.dc.b	$B5,$BE,$03,$F9,$51,$27,$F9,$1B,$63,$F9,$63,$5F,$E8,$ED,$7F,$A7
	.dc.b	$35,$FE,$74,$FF,$35,$F7,$64,$FF,$2E,$6C,$FF,$37,$6B,$FD,$49,$AF
	.dc.b	$F5,$96,$BF,$D1,$5A,$FF,$BB,$BD,$DF,$E8,$44,$9F,$E7,$6D,$8F,$E8
	.dc.b	$8D,$7F,$AF,$B5,$FE,$CC,$D7,$FA,$93,$FD,$37,$D5,$93,$FD,$19,$B3
	.dc.b	$FD,$3D,$AF,$F6,$A6,$BF,$DC,$5A,$FF,$5D,$6A,$E2,$FE,$FB,$3F,$D4
	.dc.b	$BC,$1F,$EB,$84,$70,$49,$FE,$BD,$19,$6C,$65,$FE,$F7,$19,$7F,$BE
	.dc.b	$F7,$FF,$67,$24,$FF,$60,$6C,$7F,$69,$6B,$C2,$34,$47,$4D,$7F,$C1
	.dc.b	$3F,$35,$FD,$CC,$91,$FD,$B1,$B1,$FD,$D5,$AF,$08,$D1,$1D,$35,$FF
	.dc.b	$0C,$7E,$47,$FE,$82,$69,$FE,$03,$70,$FF,$88,$1A,$83,$86,$9C,$49
	.dc.b	$1F,$E0,$48,$D1,$23,$27,$F8,$4D,$FF,$48,$3E,$4F,$3B,$C4,$49,$E4
	.dc.b	$D8,$E0,$D7,$84,$68,$8E,$9F,$17,$CD,$FB,$48,$E0,$91,$E3,$8C,$B6
	.dc.b	$32,$FC,$88,$CB,$63,$DD,$F3,$7F,$24,$8E,$09,$1E,$F8,$CB,$63,$2F
	.dc.b	$D4,$8C,$B6,$3E,$57,$C8,$7D,$52,$38,$24,$7C,$B8,$CB,$63,$2F,$E0
	.dc.b	$23,$2D,$8F,$AD,$F3,$3F,$81,$23,$82,$47,$D7,$8C,$B6,$32,$FE,$32
	.dc.b	$32,$D8,$FC,$17,$CC,$FE,$34,$8E,$09,$1F,$83,$8C,$B6,$32,$FE,$62
	.dc.b	$32,$D8,$FC,$77,$CC,$7E,$64,$8E,$09,$1F,$8F,$8C,$B6,$32,$FE,$92
	.dc.b	$32,$D8,$FC,$D7,$CC,$7E,$94,$8E,$09,$1F,$9B,$8C,$B6,$32,$FE,$C2
	.dc.b	$32,$D8,$FD,$37,$CC,$7E,$C4,$8E,$09,$1F,$A7,$8C,$B6,$32,$FE,$F2
	.dc.b	$32,$D8,$FD,$97,$B1,$FB,$D2,$38,$24,$7E,$CE,$32,$D8,$CB,$FC,$22
	.dc.b	$32,$D8,$FD,$F7,$A1,$FC,$24,$8E,$09,$1F,$BF,$8C,$B6,$32,$FF,$14
	.dc.b	$8C,$B6,$3F,$85,$7B,$9F,$C5,$48,$E0,$91,$FC,$2E,$32,$D8,$CB,$FC
	.dc.b	$82,$32,$D8,$FE,$2D,$F2,$FF,$C8,$48,$E0,$91,$FC,$5E,$32,$D8,$CB
	.dc.b	$FC,$B2,$32,$D8,$FE,$45,$F2,$DF,$CB,$48,$E0,$91,$FC,$8E,$32,$D8
	.dc.b	$CB,$FC,$E2,$32,$D8,$FE,$5D,$F2,$DF,$CE,$48,$E0,$91,$FC,$BE,$32
	.dc.b	$D8,$CB,$FD,$12,$32,$D8,$FE,$75,$EC,$FF,$45,$23,$82,$47,$F3,$B8
	.dc.b	$CB,$63,$2F,$F5,$08,$CB,$63,$FA,$37,$CA,$7F,$51,$23,$82,$47,$F4
	.dc.b	$78,$CB,$63,$2F,$F5,$C8,$CB,$63,$FA,$97,$CA,$7F,$5D,$23,$82,$47
	.dc.b	$F5,$38,$CB,$63,$2F,$F6,$88,$CB,$63,$FA,$F7,$C9,$FF,$69,$23,$82
	.dc.b	$47,$F5,$F8,$CB,$63,$2F,$F7,$48,$CB,$63,$FB,$57,$C9,$FF,$75,$23
	.dc.b	$82,$47,$F6,$B8,$CB,$63,$2F,$F8,$02,$32,$D8,$FE,$ED,$F2,$3F,$E0
	.dc.b	$12,$38,$24,$7F,$77,$8C,$B6,$32,$FF,$83,$23,$2D,$8F,$F0,$43,$07
	.dc.b	$E4,$68,$91,$93,$2E,$1E,$11,$B8,$D4,$69,$B1,$98,$E4,$D7,$8C,$73
	.dc.b	$7C,$71,$96,$C6,$5F,$F0,$F4,$65,$B1,$97,$C0,$76,$F9,$1C,$49,$A4
	.dc.b	$9F,$2B,$67,$84,$60,$8E,$08,$7C,$08,$EB,$DF,$27,$DE,$D9,$F7,$B5
	.dc.b	$F8,$ED,$7D,$ED,$7D,$AD,$7E,$9F,$C7,$7A,$89,$3F,$31,$B1,$F3,$1A
	.dc.b	$FD,$46,$BF,$31,$AF,$CA,$7E,$2F,$C8,$C9,$FA,$ED,$9F,$AE,$D7,$EF
	.dc.b	$B5,$FA,$ED,$7E,$AB,$5F,$C5,$FC,$8F,$D2,$24,$FE,$11,$B1,$F8,$46
	.dc.b	$BF,$8C,$6B,$F8,$46,$BF,$82,$7E,$EF,$C8,$49,$FC,$7B,$67,$F1,$ED
	.dc.b	$7F,$2E,$D7,$F1,$ED,$7F,$1A,$D7,$F4,$7E,$97,$E2,$89,$3F,$9C,$6C
	.dc.b	$7E,$71,$AF,$E9,$1A,$FE,$71,$AF,$E6,$9F,$CA,$FB,$12,$7F,$4E,$D9
	.dc.b	$FD,$3B,$5F,$D7,$B5,$FD,$3B,$5F,$D2,$B5,$FD,$DF,$C7,$FE,$88,$93
	.dc.b	$FB,$46,$C7,$ED,$1A,$FE,$F1,$AF,$ED,$1A,$FE,$C9,$FD,$6F,$C7,$C9
	.dc.b	$FD,$FB,$67,$F7,$ED,$7F,$83,$B5,$FD,$FB,$5F,$DE,$B5,$FE,$27,$F1
	.dc.b	$DF,$BA,$24,$FF,$0C,$6C,$7F,$0C,$6B,$FC,$51,$AF,$F0,$C6,$BF,$C2
	.dc.b	$9F,$E0,$BF,$1D,$27,$F8,$BB,$67,$F8,$BB,$5F,$E3,$ED,$7F,$8B,$B5
	.dc.b	$FE,$2A,$D7,$F9,$5F,$C6,$FF,$12,$24,$FF,$24,$6C,$7F,$24,$6B,$FC
	.dc.b	$B1,$AF,$F2,$46,$BF,$C8,$9F,$E3,$BF,$19,$27,$F9,$7B,$67,$F9,$7B
	.dc.b	$5F,$E6,$ED,$7F,$97,$B5,$FE,$5A,$D7,$FA,$1F,$C6,$7F,$2A,$24,$FF
	.dc.b	$3C,$6C,$7F,$3C,$6B,$FD,$11,$AF,$F3,$C6,$BF,$CE,$9F,$E6,$BF,$19
	.dc.b	$27,$FA,$3B,$67,$FA,$3B,$5F,$E9,$ED,$7F,$A3,$B5,$FE,$8A,$D7,$FA
	.dc.b	$DF,$C5,$FF,$42,$24,$FF,$54,$6C,$7F,$54,$6B,$FD,$71,$AF,$F5,$46
	.dc.b	$BF,$D4,$9F,$E9,$BF,$0F,$27,$FA,$FB,$67,$FA,$FB,$5F,$EC,$ED,$7F
	.dc.b	$AF,$B5,$FE,$BA,$D7,$FB,$9F,$C3,$FF,$5A,$24,$FF,$6C,$6C,$7F,$6C
	.dc.b	$6B,$FD,$D1,$AF,$F6,$C6,$BF,$DA,$9F,$EC,$BF,$0B,$27,$FB,$BB,$67
	.dc.b	$FB,$BB,$5F,$EF,$ED,$7F,$BB,$B5,$FE,$EA,$D5,$F9,$1B,$E1,$BF,$BB
	.dc.b	$36,$BF,$C0,$ED,$9F,$F0,$73,$5F,$F0,$A3,$5F,$F0,$BB,$51,$DA,$7A
	.dc.b	$92,$7A,$B6,$3A,$BE,$2F,$80,$93,$E2,$6C,$F8,$9A,$F8,$9A,$F8,$9A
	.dc.b	$F8,$A9,$18,$8F,$83,$F0,$1F,$00,$95,$F0,$1E,$FF,$86,$3E,$1F,$DB
	.dc.b	$EF,$FF,$8C,$23,$17,$D7,$0C,$56,$B0,$F8,$35,$AA,$F9,$F5,$FE,$38
	.dc.b	$F8,$6F,$F1,$B6,$4F,$83,$C6,$1E,$93,$E4,$CE,$A3,$37,$EB,$F8,$2F
	.dc.b	$AE,$4A,$FA,$EF,$DA,$CF,$C7,$3F,$05,$AF,$8A,$7C,$B9,$F9,$E1,$9F
	.dc.b	$9F,$25,$F8,$27,$F0,$ED,$87,$C1,$D8,$98,$7E,$21,$F1,$34,$95,$FE
	.dc.b	$4B,$AE,$B1,$F8,$63,$F1,$73,$FE,$5C,$69,$99,$F3,$18,$3F,$27,$1D
	.dc.b	$9F,$CA,$FB,$DF,$9A,$24,$FE,$69,$B1,$E1,$7B,$3F,$98,$6B,$F8,$76
	.dc.b	$BF,$A6,$78,$BF,$A0,$6B,$FA,$0F,$7B,$F5,$24,$9F,$D4,$BE,$0C,$7F
	.dc.b	$05,$F9,$1B,$C3,$7B,$7F,$4B,$0F,$06,$15,$56,$E7,$19,$B2,$FA,$5E
	.dc.b	$FF,$D3,$6E,$1F,$88,$3F,$D1,$3E,$DB,$DC,$FF,$39,$36,$3F,$D0,$6D
	.dc.b	$CB,$FD,$14,$D7,$FD,$21,$97,$85,$FC,$2F,$07,$FA,$52,$31,$AF,$E1
	.dc.b	$F1,$FC,$6C,$62,$7C,$38,$F8,$09,$8F,$81,$F1,$E3,$23,$C6,$DC,$7F
	.dc.b	$D3,$97,$CD,$FE,$16,$D7,$EA,$4F,$C7,$8C,$4F,$AF,$1F,$D8,$2F,$6F
	.dc.b	$F4,$2E,$71,$F8,$B1,$8F,$73,$E2,$93,$0F,$8A,$FF,$A6,$70,$F0,$1E
	.dc.b	$1D,$7E,$94,$F8,$77,$FB,$79,$FC,$63,$F9,$0E,$5E,$19,$FB,$8D,$8D
	.dc.b	$07,$C8,$F0,$C7,$CE,$7D,$82,$4F,$D8,$7C,$95,$E0,$CF,$F1,$AD,$CF
	.dc.b	$F3,$95,$E5,$DF,$C3,$7F,$9C,$C7,$C5,$3F,$EC,$D6,$E3,$F1,$23,$F6
	.dc.b	$0F,$AB,$D7,$FC,$49,$23,$F1,$2F,$8A,$FB,$4F,$B3,$3F,$8A,$FB,$B4
	.dc.b	$FA,$B5,$E1,$25,$F6,$9F,$85,$1E,$5F,$5F,$F2,$E4,$9F,$CB,$BD,$E7
	.dc.b	$A6,$9E,$F7,$F5,$26,$31,$F9,$F7,$F9,$49,$D4,$F6,$BF,$3E,$BF,$86
	.dc.b	$F5,$FF,$4A,$48,$FD,$2B,$DE,$3E,$6E,$1F,$B2,$E1,$E8,$F2,$DF,$F4
	.dc.b	$98,$F6,$D9,$9E,$B3,$FB,$F0,$F5,$FF,$66,$4D,$3F,$66,$F7,$C3,$9E
	.dc.b	$3D,$E7,$B5,$FE,$8B,$1F,$5E,$DF,$32,$73,$3F,$8B,$F5,$FF,$7E,$48
	.dc.b	$FD,$FB,$D6,$39,$5F,$57,$E4,$13,$5F,$7B,$E8,$D3,$D5,$1F,$AC,$F5
	.dc.b	$FF,$85,$93,$0F,$E1,$6F,$5A,$FE,$33,$97,$D0,$E1,$FC,$1D,$F4,$26
	.dc.b	$9E,$EA,$ED,$87,$AB,$67,$AF,$FC,$60,$91,$FC,$61,$ED,$B7,$EC,$F8
	.dc.b	$3C,$B8,$73,$C7,$AE,$3F,$99,$C3,$A7,$AF,$FA,$12,$3F,$91,$C9,$1F
	.dc.b	$D0,$31,$FD,$FE,$9E,$9C,$BA,$E1,$F2,$0F,$8D,$97,$F2,$ED,$B3,$1F
	.dc.b	$BB,$F5,$FF,$5C,$4C,$FB,$3D,$70,$FB,$27,$58,$FB,$38,$F8,$38,$3F
	.dc.b	$33,$1F,$0F,$D7,$FD,$C1,$1F,$43,$CE,$3F,$03,$C1,$F3,$B3,$F8,$44
	.dc.b	$CB,$D1,$BF,$F2,$4D,$3C,$FB,$BD,$7F,$E0,$84,$CF,$E2,$3D,$6F,$DE
	.dc.b	$7E,$BA,$FE,$6B,$F7,$B2,$FB,$18,$7C,$2D,$18,$7C,$B2,$1F,$BF,$72
	.dc.b	$DD,$7E,$1D,$7E,$4F,$0E,$B5,$F0,$F0,$FC,$29,$E2,$3F,$B2,$FA,$DF
	.dc.b	$5C,$98,$75,$7D,$15,$E6,$D7,$E2,$57,$C2,$D7,$E2,$3E,$93,$AC,$F8
	.dc.b	$3D,$6F,$C1,$13,$0F,$43,$D2,$B8,$DF,$4B,$E5,$7E,$8C,$FF,$81,$69
	.dc.b	$F0,$DF,$E8,$B3,$FD,$B1,$AF,$E8,$4C,$69,$FE,$68,$69,$FD,$8A,$36
	.dc.b	$FE,$44,$C9,$0E,$44,$34,$7F,$77,$F5,$BF,$CF,$CF,$F6,$C6,$BF,$DF
	.dc.b	$9A,$EF,$C6,$F1,$34,$E0,$E1,$FE,$E7,$1C,$1B,$19,$7F,$C3,$31,$96
	.dc.b	$C6,$76,$4E,$5B,$19,$F5,$BE,$16,$DF,$E1,$26,$DF,$F8,$45,$AF,$F8
	.dc.b	$11,$AF,$63,$FA,$85,$FC,$0C,$FF,$C6,$2D,$D7,$E7,$7E,$46,$F8,$3F
	.dc.b	$A1,$5A,$CB,$FC,$19,$7B,$C3,$DA,$7F,$5B,$F3,$37,$8F,$F8,$8F,$D8
	.dc.b	$F0,$EA,$69,$2B,$F4,$17,$CD,$F3,$63,$DD,$FF,$01,$E7,$E3,$7C,$39
	.dc.b	$47,$D6,$FF,$87,$F2,$F0,$C7,$F8,$93,$D8,$F0,$91,$FD,$73,$E4,$C3
	.dc.b	$C1,$97,$CE,$9E,$67,$83,$4E,$6F,$93,$4F,$C0,$67,$E5,$3F,$CB,$38
	.dc.b	$EA,$DE,$C7,$C0,$25,$FE,$F5,$F0,$67,$E9,$BE,$58,$7D,$0C,$79,$65
	.dc.b	$E4,$25,$7B,$70,$FF,$1E,$7B,$1F,$30,$8F,$F0,$97,$AB,$EA,$C3,$AB
	.dc.b	$FE,$D0,$D7,$E5,$39,$C7,$D6,$47,$FC,$97,$A7,$AF,$D8,$FB,$24,$8F
	.dc.b	$E2,$CF,$6B,$EB,$1F,$14,$ED,$9F,$AB,$2F,$76,$1D,$67,$FC,$F3,$87
	.dc.b	$C7,$F6,$3F,$0A,$48,$FE,$44,$F5,$9E,$1B,$7D,$5C,$7F,$D9,$19,$FA
	.dc.b	$9F,$0C,$66,$77,$87,$D3,$F6,$3F,$61,$8F,$FA,$21,$B4,$69,$8D,$BE
	.dc.b	$7C,$FC,$53,$D6,$7A,$8F,$CF,$E1,$E0,$BF,$C0,$7B,$1D,$09,$3F,$BB
	.dc.b	$7A,$69,$C6,$3E,$30,$CF,$D5,$9F,$BB,$1F,$0E,$3F,$8C,$AF,$F3,$D7
	.dc.b	$B1,$E8,$23,$FC,$1D,$EA,$F9,$CE,$B7,$D1,$F3,$DF,$5C,$3F,$15,$1F
	.dc.b	$E5,$1F,$63,$E1,$91,$FE,$24,$F4,$C3,$E6,$E3,$FB,$5C,$3A,$57,$AB
	.dc.b	$5F,$A0,$F6,$3F,$12,$7F,$A4,$7D,$8F,$9E,$47,$F8,$F3,$D7,$2F,$24
	.dc.b	$7E,$BC,$F8,$D3,$D1,$ED,$AF,$EB,$7D,$8F,$B2,$48,$FE,$50,$F3,$7C
	.dc.b	$47,$C2,$E1,$FD,$A4,$FE,$4E,$3E,$35,$23,$31,$FC,$C5,$FC,$29,$1F
	.dc.b	$E6,$99,$7D,$4C,$3F,$35,$97,$C2,$BE,$55,$FD,$E6,$FF,$8F,$91,$FF
	.dc.b	$3D,$4A,$BF,$26,$D8,$FE,$88,$D7,$FA,$53,$5F,$EA,$5A,$FC,$4C,$3E
	.dc.b	$96,$3F,$C5,$2B,$87,$47,$F5,$46,$DF,$F3,$AC,$3F,$A0,$C6,$23,$F8
	.dc.b	$17,$C3,$78,$88,$F8,$9F,$94,$FD,$DC,$3B,$BF,$EF,$19,$F3,$6D,$FB
	.dc.b	$2B,$F1,$1E,$3F,$5F,$C9,$EB,$FB,$89,$7E,$E7,$AE,$DF,$74,$FB,$D8
	.dc.b	$7E,$30,$57,$4A,$F0,$5F,$57,$C5,$1F,$28,$8F,$CA,$7A,$E3,$F5,$E3
	.dc.b	$F7,$27,$F4,$A8,$DB,$F3,$9A,$FC,$9F,$85,$FA,$A4,$7E,$AB,$F6,$35
	.dc.b	$FC,$56,$7F,$C8,$31,$F4,$47,$F8,$57,$D8,$FF,$1E,$DF,$F2,$FF,$86
	.dc.b	$C6,$FF,$DB,$1B,$1E,$49,$DB,$FC,$0E,$BB,$3F,$84,$6B,$F8,$47,$ED
	.dc.b	$91,$FC,$6B,$5D,$1C,$48,$7E,$33,$E1,$3F,$22,$47,$F2,$2D,$7C,$67
	.dc.b	$F1,$57,$F1,$24,$7F,$36,$ED,$FC,$83,$5F,$F0,$4F,$C2,$7E,$78,$8F
	.dc.b	$E7,$9E,$D9,$7F,$3A,$9F,$CD,$1F,$DB,$BF,$23,$FF,$80,$15,$F8,$F8
	.dc.b	$F8,$EF,$8C,$0C,$3E,$4E,$69,$DD,$7F,$8D,$A7,$E6,$AF,$5B,$FA,$1B
	.dc.b	$6F,$FC,$68,$D7,$FC,$74,$D7,$FC,$7E,$F9,$1F,$F0,$9D,$7A,$1F,$1E
	.dc.b	$1F,$B9,$9D,$37,$CA,$F1,$F8,$B2,$BD,$38,$F9,$27,$CE,$79,$09,$1E
	.dc.b	$46,$BE,$46,$BE,$46,$BF,$E6,$5B,$F3,$39,$39,$1E,$EF,$82,$F8,$24
	.dc.b	$70,$43,$E0,$B9,$69,$93,$E1,$B9,$69,$F2,$BE,$0B,$E6,$91,$C1,$0F
	.dc.b	$9A,$E5,$A6,$4F,$9E,$E5,$A7,$D6,$F8,$2F,$B2,$47,$04,$3E,$CB,$96
	.dc.b	$99,$3E,$DB,$96,$9F,$82,$F2,$7F,$0A,$47,$04,$3F,$0A,$E5,$A6,$4F
	.dc.b	$C3,$B9,$69,$F8,$EF,$5F,$F2,$24,$70,$43,$F2,$2E,$5A,$64,$FC,$9B
	.dc.b	$96,$9F,$9A,$F5,$FF,$3A,$47,$04,$3F,$3A,$E5,$A6,$4F,$CF,$B9,$69
	.dc.b	$FA,$6F,$5F,$F5,$24,$70,$43,$F5,$2E,$5A,$64,$FD,$5B,$96,$9F,$B2
	.dc.b	$F5,$FF,$6A,$47,$04,$3F,$6A,$E5,$A6,$4F,$DB,$B9,$69,$FB,$EF,$5F
	.dc.b	$F8,$09,$1C,$10,$FE,$02,$E5,$A6,$4F,$E0,$6E,$5A,$7F,$0A,$F5,$FF
	.dc.b	$86,$91,$C1,$0F,$E1,$AE,$5A,$64,$FE,$1E,$E5,$A7,$F1,$6F,$5F,$F8
	.dc.b	$C9,$1C,$10,$FE,$32,$E5,$A6,$4F,$E3,$6E,$5A,$7F,$22,$F5,$BF,$92
	.dc.b	$91,$C1,$0F,$E4,$AE,$5A,$64,$FE,$4E,$E5,$A7,$F2,$EF,$5B,$F9,$89
	.dc.b	$1C,$10,$FE,$62,$E5,$A6,$4F,$E6,$6E,$5A,$7F,$3A,$F5,$BF,$9E,$91
	.dc.b	$C1,$0F,$E7,$AE,$5A,$64,$FE,$7E,$E5,$A7,$FB,$E7,$C0,$FE,$CA,$FF
	.dc.b	$52,$3F,$C7,$9D,$BF,$E8,$E8,$CC,$7F,$56,$6B,$A3,$94,$6C,$8E,$08
	.dc.b	$75,$39,$3A,$70,$7F,$92,$FD,$6F,$01,$1F,$33,$5D,$CA,$31,$1D,$63
	.dc.b	$0F,$02,$1C,$23,$6F,$BE,$1E,$B7,$BC,$91,$F1,$5A,$F8,$1A,$F7,$6B
	.dc.b	$CD,$AF,$C7,$7E,$57,$AB,$23,$E8,$35,$F5,$B5,$F5,$35,$F3,$35,$FA
	.dc.b	$4D,$7F,$CE,$5E,$1F,$D7,$23,$F7,$5A,$FC,$86,$BF,$1D,$AF,$C3,$6B
	.dc.b	$F7,$DF,$C1,$7A,$B2,$3F,$10,$D7,$E9,$B5,$FA,$4D,$7E,$73,$5F,$C5
	.dc.b	$35,$FF,$45,$7A,$BF,$8F,$23,$F9,$56,$BF,$80,$6B,$F7,$DA,$FD,$B6
	.dc.b	$BF,$97,$7F,$35,$EA,$C8,$FD,$03,$5F,$C5,$B5,$FC,$53,$5F,$C3,$35
	.dc.b	$FD,$13,$5F,$F5,$17,$AB,$FA,$72,$3F,$AD,$6B,$F9,$86,$BF,$97,$6B
	.dc.b	$F9,$36,$BF,$AF,$7F,$65,$EA,$C8,$FD,$C3,$5F,$D1,$B5,$FD,$13,$5F
	.dc.b	$CF,$35,$FD,$D3,$5F,$F5,$D7,$AB,$FB,$F2,$3F,$C1,$5A,$FE,$C1,$AF
	.dc.b	$EB,$DA,$FE,$AD,$AF,$F0,$77,$F8,$57,$AB,$23,$F8,$83,$5F,$DD,$B5
	.dc.b	$FD,$D3,$5F,$DB,$35,$FE,$24,$D7,$FD,$A5,$EA,$FF,$17,$23,$FC,$75
	.dc.b	$AF,$F0,$86,$BF,$C1,$DA,$FF,$03,$6B,$FC,$7D,$FE,$45,$EA,$C8,$FE
	.dc.b	$50,$D7,$F8,$9B,$5F,$E2,$4D,$7F,$87,$35,$FE,$54,$D7,$FD,$D5,$EA
	.dc.b	$FF,$2F,$23,$FC,$D5,$AF,$F2,$06,$BF,$C7,$DA,$FF,$1B,$6B,$FC,$DD
	.dc.b	$FE,$75,$EA,$C8,$FE,$80,$D7,$F9,$5B,$5F,$E5,$4D,$7F,$93,$35,$FE
	.dc.b	$84,$D5,$C2,$F5,$7F,$A4,$12,$3F,$A7,$35,$FE,$72,$D7,$F9,$C3,$5F
	.dc.b	$E6,$8D,$7F,$A8,$3F,$D4,$FD,$59,$1F,$D6,$5A,$FF,$44,$6B,$FD,$0D
	.dc.b	$AF,$F3,$F6,$BF,$D6,$DA,$FA,$BD,$5F,$EC,$04,$7F,$B3,$35,$FE,$A2
	.dc.b	$D7,$FA,$83,$5F,$E9,$8D,$7F,$B4,$3F,$DA,$FE,$06,$47,$F7,$16,$BF
	.dc.b	$D7,$1A,$FF,$5B,$6B,$FD,$5D,$AF,$F7,$36,$BF,$4B,$E0,$7F,$BC,$11
	.dc.b	$FE,$FC,$D7,$FB,$4B,$5F,$ED,$0D,$7F,$B2,$35,$FF,$00,$3F,$DD,$9F
	.dc.b	$F6,$BF,$E4,$6F,$63,$10,$8C,$3B,$6B,$FD,$F9,$AF,$F8,$0D,$AF,$F8
	.dc.b	$45,$AF,$68,$EB,$18,$8E,$D1,$82,$3D,$DD,$FC,$6C,$8F,$13,$5F,$13
	.dc.b	$5F,$13,$5F,$13,$5F,$7B,$5F,$77,$C6,$F9,$08,$FB,$DA,$FB,$DA,$FB
	.dc.b	$DA,$FB,$DA,$FC,$C7,$DF,$A4,$8F,$94,$D7,$E5,$35,$F9,$4D,$7E,$53
	.dc.b	$5F,$AE,$D7,$EB,$7B,$5F,$30,$8F,$D7,$6B,$F5,$DA,$FD,$76,$BF,$5D
	.dc.b	$AF,$E1,$1F,$AF,$ED,$48,$FC,$13,$5F,$C1,$35,$FC,$13,$5F,$C1,$35
	.dc.b	$FC,$7B,$5F,$C7,$7B,$3F,$84,$23,$F8,$F6,$BF,$8F,$6B,$F8,$F6,$BF
	.dc.b	$8F,$6B,$F9,$C7,$F1,$FF,$09,$23,$F3,$4D,$7F,$34,$D7,$F3,$4D,$7F
	.dc.b	$34,$D7,$F4,$ED,$7F,$4D,$F0,$7F,$9C,$23,$FA,$76,$BF,$A7,$6B,$FA
	.dc.b	$76,$BF,$A7,$6B,$FB,$47,$F4,$FF,$09,$23,$F6,$4D,$7F,$64,$D7,$F6
	.dc.b	$4D,$7F,$64,$D7,$F7,$ED,$7F,$7D,$F0,$9F,$B4,$23,$FB,$F6,$BF,$BF
	.dc.b	$6B,$FB,$F6,$BF,$BF,$6B,$FC,$31,$FD,$FF,$B9,$23,$F8,$53,$5F,$E1
	.dc.b	$4D,$7F,$85,$35,$FE,$14,$D7,$F8,$BB,$5F,$E2,$DE,$EF,$F0,$C2,$3F
	.dc.b	$C5,$DA,$FF,$17,$6B,$FC,$5D,$AF,$F1,$76,$BF,$C9,$1F,$E2,$FE,$EC
	.dc.b	$8F,$E4,$4D,$7F,$91,$35,$FE,$44,$D7,$F9,$13,$5F,$E5,$ED,$7F,$97
	.dc.b	$7C,$57,$F2,$42,$3F,$CB,$DA,$FF,$2F,$6B,$FC,$BD,$AF,$F2,$F6,$BF
	.dc.b	$CF,$1F,$E5,$FF,$15,$23,$F9,$D3,$5F,$E7,$4D,$7F,$9D,$35,$FE,$74
	.dc.b	$D7,$FA,$3B,$5F,$E8,$DF,$05,$FC,$F0,$8F,$F4,$76,$BF,$D1,$DA,$FF
	.dc.b	$47,$6B,$FD,$1D,$AF,$F5,$47,$FA,$3F,$C3,$48,$FE,$A4,$D7,$FA,$93
	.dc.b	$5F,$EA,$4D,$7F,$A9,$35,$FE,$BE,$D7,$FA,$F7,$C3,$7F,$54,$23,$FD
	.dc.b	$7D,$AF,$F5,$F6,$BF,$D7,$DA,$FF,$5F,$6B,$FD,$B1,$FE,$BF,$EF,$C8
	.dc.b	$FE,$D4,$D7,$FB,$53,$5F,$ED,$4D,$7F,$B5,$35,$FE,$EE,$D7,$FB,$B7
	.dc.b	$C1,$7F,$6C,$23,$FD,$DD,$AF,$F7,$76,$BF,$DD,$DA,$FF,$77,$6B,$FE
	.dc.b	$04,$7F,$C0,$7F,$91,$FF,$81,$20,$D2,$3A,$21,$FE,$0D,$72,$D3,$20
	.dc.b	$85,$76,$E0,$C7,$C0,$78,$88,$F5,$6B,$B3,$CA,$E1,$E5,$08,$C4,$78
	.dc.b	$23,$0F,$32,$1C,$D7,$C1,$7C,$22,$47,$B5,$AE,$DA,$F8,$1A,$F7,$6B
	.dc.b	$EF,$7E,$6F,$C0,$48,$F9,$0D,$7C,$4D,$7D,$6D,$7D,$4D,$7E,$53,$5F
	.dc.b	$AF,$F0,$FF,$68,$8F,$D5,$6B,$EF,$6B,$F2,$1A,$FC,$76,$BF,$5D,$FC
	.dc.b	$2F,$C3,$C8,$FC,$03,$5F,$A9,$F4,$8F,$4F,$EF,$FC,$23,$D5,$FC,$0C
	.dc.b	$BD,$5C,$C6,$1B,$7F,$87,$9E,$4F,$E2,$6B,$93,$9F,$57,$6A,$51,$86
	.dc.b	$C7,$1B,$F0,$FA,$BC,$36,$7C,$5F,$E8,$09,$5F,$9B,$7F,$CC,$B7,$E2
	.dc.b	$86,$3E,$27,$C1,$97,$73,$C1,$9F,$87,$39,$3E,$13,$D0,$53,$D1,$97
	.dc.b	$8F,$E2,$FF,$62,$48,$FD,$63,$5F,$CB,$35,$FC,$FB,$5F,$CF,$35,$FD
	.dc.b	$73,$FB,$CF,$85,$91,$FB,$76,$BF,$AB,$F8,$57,$B9,$F3,$17,$B9,$4C
	.dc.b	$46,$7D,$C9,$5E,$BA,$47,$D1,$E1,$F9,$D1,$FC,$30,$93,$FC,$1D,$FF
	.dc.b	$48,$9E,$07,$95,$F7,$7A,$5F,$67,$E8,$D7,$F0,$AA,$EF,$1E,$1B,$F0
	.dc.b	$3F,$37,$85,$F0,$BF,$C7,$09,$1F,$C5,$DF,$E3,$FF,$0B,$EF,$F8,$09
	.dc.b	$3C,$63,$F0,$C7,$C3,$F8,$09,$19,$21,$FC,$26,$39,$BF,$C8,$DA,$E4
	.dc.b	$C1,$0C,$10,$FE,$32,$F0,$7F,$8D,$FC,$2F,$F3,$42,$3F,$CB,$DE,$BC
	.dc.b	$9F,$93,$C5,$5E,$8A,$FA,$9F,$07,$5F,$92,$7F,$31,$8C,$57,$E6,$F4
	.dc.b	$7C,$1F,$E9,$7E,$17,$FA,$19,$2F,$F9,$F3,$F6,$2F,$F0,$D9,$49,$FC
	.dc.b	$14,$7E,$26,$64,$7E,$15,$FC,$67,$C2,$FE,$C4,$FE,$9E,$48,$FE,$94
	.dc.b	$F5,$7F,$19,$5D,$5F,$48,$C3,$DE,$F5,$21,$A5,$5E,$08,$FC,$26,$5F
	.dc.b	$D7,$88,$FF,$59,$7F,$27,$C9,$FA,$12,$47,$E7,$79,$3F,$49,$7F,$A5
	.dc.b	$E0,$93,$FA,$18,$FE,$19,$A7,$F6,$A2,$47,$F6,$57,$AB,$FA,$7B,$EA
	.dc.b	$F9,$EF,$AB,$FA,$BB,$EA,$FA,$EF,$F8,$C7,$07,$F7,$62,$47,$F7,$17
	.dc.b	$FB,$CF,$07,$EE,$7A,$E4,$FE,$8E,$3F,$90,$1F,$BD,$EB,$91,$92,$1F
	.dc.b	$D8,$E3,$9B,$FD,$F5,$AE,$4C,$10,$C1,$0F,$EE,$6F,$07,$FB,$AF,$2F
	.dc.b	$F8,$38,$8F,$F8,$25,$EB,$C5,$F9,$B9,$F3,$CF,$F0,$DE,$DA,$FE,$0B
	.dc.b	$91,$38,$AF,$CD,$76,$FF,$3D,$E8,$FF,$0D,$92,$FF,$C2,$CF,$F8,$25
	.dc.b	$FF,$0A,$7E,$60,$F0,$7F,$A5,$B6,$3F,$9B,$35,$FC,$61,$FC,$97,$C1
	.dc.b	$FE,$4D,$E0,$C8,$FE,$45,$1F,$D3,$7A,$FF,$C5,$64,$8F,$F1,$33,$E1
	.dc.b	$55,$D7,$2F,$E9,$8D,$EC,$FF,$13,$B6,$7F,$C5,$5C,$7F,$E2,$7A,$FD
	.dc.b	$C6,$1E,$19,$FD,$06,$3E,$57,$FA,$77,$E4,$9F,$D5,$DF,$F1,$73,$5F
	.dc.b	$F1,$9B,$5F,$F1,$B3,$5F,$2F,$8D,$F1,$3C,$69,$1F,$12,$3E,$3F,$8D
	.dc.b	$23,$E3,$BF,$2B,$C6,$FE,$EB,$7F,$DA,$67,$C3,$E2,$7C,$18,$C4,$78
	.dc.b	$A3,$E2,$6A,$C3,$BC,$FF,$00,$86,$3F,$5A,$7C,$13,$9C,$6E,$9F,$2B
	.dc.b	$2F,$0F,$8B,$DC,$95,$F8,$1C,$FF,$C2,$4D,$F1,$7F,$03,$E4,$48,$F9
	.dc.b	$31,$EA,$3F,$0B,$E4,$48,$C9,$0F,$F2,$A5,$7F,$84,$1B,$3E,$F2,$3D
	.dc.b	$C8,$68,$A7,$D6,$76,$E4,$EB,$E3,$FC,$92,$3E,$B7,$79,$7F,$85,$6B
	.dc.b	$C4,$DA,$FC,$EF,$95,$EB,$BF,$44,$E5,$FC,$FF,$95,$F6,$3E,$1F,$EB
	.dc.b	$12,$3E,$43,$F5,$BC,$FF,$D3,$92,$3F,$4F,$E7,$FE,$AB,$CF,$FD,$5F
	.dc.b	$9F,$27,$F5,$71,$F8,$4F,$87,$FC,$11,$23,$EA,$3D,$5F,$D8,$DF,$57
	.dc.b	$F0,$57,$D5,$FD,$AD,$F5,$7F,$0D,$7F,$90,$F8,$7F,$C7,$12,$3F,$00
	.dc.b	$FE,$3B,$D5,$FD,$FF,$AB,$27,$F7,$F1,$FC,$0B,$D5,$91,$FC,$09,$FE
	.dc.b	$0B,$EA,$FE,$73,$E1,$FF,$34,$49,$FC,$63,$D5,$FE,$13,$7D,$5F,$CD
	.dc.b	$5F,$47,$F3,$97,$78,$3F,$80,$E1,$FA,$8F,$87,$FD,$31,$2B,$F3,$0F
	.dc.b	$EA,$7E,$1F,$F8,$BF,$B5,$27,$F7,$F1,$FB,$03,$F8,$DF,$B5,$23,$24
	.dc.b	$3F,$2F,$1C,$DF,$D2,$35,$C9,$82,$18,$21,$FC,$31,$E0,$FF,$0E,$F8
	.dc.b	$7F,$DF,$11,$FD,$83,$D7,$1F,$CB,$4F,$9E,$7F,$97,$FB,$B5,$FE,$FF
	.dc.b	$D2,$9C,$57,$F2,$6F,$77,$F8,$6F,$C3,$FF,$0B,$25,$FE,$F5,$FE,$5D
	.dc.b	$F7,$8F,$84,$FF,$43,$AF,$84,$A6,$23,$3F,$09,$2F,$FC,$ED,$48,$E3
	.dc.b	$F0,$9F,$C7,$3E,$23,$F8,$D1,$2B,$F8,$63,$E0,$BE,$B1,$CB,$0E,$CF
	.dc.b	$4B,$EC,$FF,$87,$30,$EC,$F8,$30,$FE,$53,$F1,$1F,$C9,$C9,$1F,$C6
	.dc.b	$5F,$6E,$3E,$C2,$47,$A2,$3D,$FF,$11,$F0,$3E,$22,$4F,$B2,$3F,$9A
	.dc.b	$9F,$CC,$C8,$FF,$25,$7A,$BF,$0A,$BA,$BE,$7A,$EA,$FC,$6A,$EA,$FA
	.dc.b	$EB,$FA,$0C,$7F,$3F,$24,$7F,$31,$7F,$A1,$67,$F5,$38,$A4,$FF,$58
	.dc.b	$8F,$E9,$27,$D8,$D6,$46,$48,$7F,$30,$8E,$6F,$F3,$D6,$B9,$30,$43
	.dc.b	$04,$3F,$B2,$BC,$1F,$EC,$FA,$FF,$57,$23,$FD,$25,$EB,$BF,$F2,$F9
	.dc.b	$F3,$CF,$E2,$39,$6B,$F9,$A8,$B3,$FE,$C9,$DB,$F3,$93,$FD,$DB,$7F
	.dc.b	$EC,$84,$7F,$AB,$3F,$9A,$F0,$7F,$33,$E0,$C9,$FC,$AC,$7E,$73,$B6
	.dc.b	$47,$E6,$1F,$CF,$F6,$FF,$72,$E3,$FE,$E0,$49,$FE,$C4,$F5,$7F,$3F
	.dc.b	$7D,$5F,$CA,$5F,$47,$F2,$D7,$72,$F5,$61,$FE,$F2,$F1,$1D,$1F,$E1
	.dc.b	$D6,$DF,$F8,$BD,$A8,$85,$72,$E0,$69,$FE,$F1,$BF,$91,$E7,$7E,$E8
	.dc.b	$FD,$DF,$93,$FB,$CB,$FD,$EF,$8F,$30,$FD,$CC,$79,$7A,$7C,$84,$8F
	.dc.b	$23,$FC,$1A,$FA,$47,$F0,$07,$AB,$C8,$FE,$07,$5D,$5F,$DC,$57,$C1
	.dc.b	$ED,$F8,$04,$8F,$80,$FC,$3E,$DF,$7E,$4F,$4B,$E0,$FE,$61,$FE,$DA
	.dc.b	$BF,$F4,$EB,$67,$51,$F1,$6B,$E7,$1F,$C6,$AA,$47,$CB,$9D,$AC,$15
	.dc.b	$C2,$3E,$93,$E3,$69,$0F,$A0,$71,$74,$E2,$39,$BB,$3E,$B3,$87,$51
	.dc.b	$C1,$CF,$85,$F8,$32,$3F,$83,$7A,$D7,$A2,$B9,$36,$BE,$3D,$FC,$6F
	.dc.b	$6F,$D5,$87,$FB,$0A,$F7,$3F,$8A,$6C,$7E,$37,$17,$AF,$F9,$5F,$17
	.dc.b	$F2,$84,$9F,$CA,$3F,$E7,$39,$7C,$1F,$82,$79,$7C,$1D,$53,$FE,$E6
	.dc.b	$7E,$86,$3F,$6A,$B5,$2F,$80,$A4,$AD,$9E,$5F,$73,$F5,$1E,$4F,$E9
	.dc.b	$C9,$1F,$A7,$7C,$97,$D3,$19,$3C,$8F,$1C,$75,$7F,$13,$77,$7F,$C1
	.dc.b	$E3,$FE,$A6,$9E,$1A,$BC,$9F,$DB,$12,$3F,$6C,$FC,$3F,$87,$F5,$92
	.dc.b	$3C,$14,$F8,$6F,$7F,$C3,$7C,$0F,$87,$95,$A8,$FE,$09,$E8,$FF,$03
	.dc.b	$24,$7F,$03,$7B,$3D,$30,$EC,$FE,$8A,$FA,$BE,$0B,$EA,$FE,$9A,$FF
	.dc.b	$88,$7A,$3F,$C3,$C9,$1F,$C3,$DF,$E2,$5E,$87,$D4,$F7,$64,$FC,$28
	.dc.b	$FE,$30,$7D,$8F,$85,$91,$92,$1F,$A1,$8E,$6F,$F1,$B6,$B9,$30,$43
	.dc.b	$04,$3F,$4E,$F0,$7F,$53,$E9,$FF,$27,$23,$FC,$9D,$EB,$C1,$FA,$59
	.dc.b	$F3,$CF,$D2,$F8,$7A,$FE,$8E,$2C,$FE,$03,$E1,$FF,$9A,$7A,$FF,$CC
	.dc.b	$C9,$5F,$CC,$DF,$D0,$7C,$2F,$E6,$3E,$12,$4F,$DE,$8F,$CD,$FC,$3C
	.dc.b	$8F,$C0,$BF,$9E,$F8,$7F,$E8,$1E,$DF,$F3,$F2,$4F,$F3,$F7,$AB,$F8
	.dc.b	$8B,$EA,$FF,$1E,$BE,$8F,$F2,$0B,$B9,$7E,$63,$0F,$E9,$9E,$EF,$F4
	.dc.b	$B2,$57,$F4,$B7,$FD,$D5,$E6,$7F,$B9,$60,$D3,$1D,$1F,$EE,$7F,$03
	.dc.b	$F9,$7F,$C2,$7E,$DB,$E1,$25,$FE,$72,$3F,$AD,$7B,$FF,$D6,$49,$1F
	.dc.b	$D6,$5E,$FF,$0D,$FC,$0F,$1E,$CF,$F0,$09,$FE,$03,$5D,$1F,$E7,$77
	.dc.b	$87,$F6,$3F,$07,$FD,$94,$8F,$F6,$57,$FB,$47,$C2,$75,$F8,$49,$3F
	.dc.b	$B8,$8F,$ED,$C7,$87,$E1,$24,$64,$87,$F3,$78,$E6,$FF,$71,$6B,$93
	.dc.b	$04,$30,$43,$F9,$EB,$C1,$FE,$7F,$3F,$DF,$48,$FF,$7D,$7A,$F0,$7F
	.dc.b	$3B,$9F,$3C,$FF,$17,$CF,$FD,$F8,$F9,$A7,$4D,$D7,$FC,$07,$3F,$EF
	.dc.b	$E8,$FE,$2F,$87,$F8,$2C,$8F,$F8,$2D,$FC,$0F,$2F,$D3,$24,$77,$E5
	.dc.b	$FA,$DC,$3F,$5F,$82,$4F,$8E,$3F,$C3,$1C,$1F,$E1,$72,$47,$F8,$5D
	.dc.b	$EA,$FA,$6F,$AA,$CF,$FB,$7E,$4C,$7B,$CF,$BF,$1E,$EE,$F3,$F6,$E5
	.dc.b	$FD,$2F,$F2,$5F,$ED,$F3,$FE,$22,$6B,$FE,$25,$6A,$6E,$0F,$0B,$F2
	.dc.b	$5D,$D2,$BE,$74,$7E,$57,$AA,$47,$D1,$7F,$31,$D9,$E1,$E5,$EF,$89
	.dc.b	$A4,$AF,$2E,$3D,$DF,$B5,$8F,$77,$D8,$7F,$8E,$E3,$4E,$0F,$D9,$E1
	.dc.b	$E2,$71,$08,$F1,$C7,$F5,$BB,$72,$F9,$09,$1F,$1A,$32,$DE,$9F,$D9
	.dc.b	$79,$12,$39,$47,$CD,$3F,$6D,$E2,$48,$C9,$0F,$F0,$FD,$27,$E1,$92
	.dc.b	$3B,$10,$DB,$96,$99,$3F,$CB,$11,$CD,$C1,$F7,$7B,$3E,$79,$1F,$63
	.dc.b	$CF,$8F,$FC,$73,$3C,$67,$E5,$36,$3F,$1F,$E7,$7B,$67,$D5,$1C,$2F
	.dc.b	$F8,$A7,$97,$F8,$DE,$EF,$BA,$4A,$F9,$4F,$F2,$8F,$57,$F8,$D1,$23
	.dc.b	$F6,$DE,$9F,$F1,$DF,$47,$F8,$F7,$A3,$27,$F7,$51,$F9,$9F,$0B,$F1
	.dc.b	$24,$8F,$AC,$F5,$7F,$80,$5F,$57,$F0,$37,$D5,$FE,$09,$7D,$5F,$C2
	.dc.b	$DF,$E9,$7C,$4F,$CA,$92,$3F,$04,$FE,$9B,$F3,$0F,$F7,$E4,$7F,$88
	.dc.b	$DF,$F7,$E3,$7D,$EF,$F1,$C4,$6A,$3F,$28,$FF,$BB,$9B,$EE,$68,$DF
	.dc.b	$B9,$FB,$5F,$23,$F4,$A4,$9F,$C9,$3E,$1F,$77,$AE,$12,$7C,$51,$CD
	.dc.b	$F2,$5D,$7C,$37,$FC,$07,$CB,$FD,$89,$23,$F3,$CF,$ED,$3C,$BF,$4F
	.dc.b	$C1,$49,$FD,$34,$7F,$09,$3D,$BF,$05,$23,$24,$3F,$3B,$1C,$DF,$D5
	.dc.b	$35,$C9,$82,$18,$21,$F9,$C7,$83,$F9,$EF,$47,$F8,$49,$1F,$DB,$3D
	.dc.b	$79,$FF,$37,$3E,$79,$FA,$DF,$09,$5F,$CE,$C5,$9F,$F4,$CE,$EF,$84
	.dc.b	$FE,$4D,$E9,$FF,$16,$25,$FF,$03,$7F,$05,$F0,$5F,$78,$91,$F8,$3F
	.dc.b	$85,$FC,$0F,$C2,$FE,$0B,$E1,$64,$FE,$1E,$3F,$99,$7A,$DF,$C8,$89
	.dc.b	$1F,$C3,$DE,$AF,$E3,$6F,$AB,$F8,$AB,$EA,$FE,$46,$FA,$BF,$8E,$BF
	.dc.b	$E7,$DE,$BF,$F2,$E2,$47,$F1,$B7,$F3,$5F,$09,$F9,$6F,$85,$93,$F9
	.dc.b	$B8,$FC,$CF,$C2,$C8,$FC,$F3,$F9,$CF,$2F,$FA,$57,$BB,$FC,$E8,$93
	.dc.b	$FC,$9D,$EA,$FE,$96,$FA,$BF,$A2,$BE,$8F,$E9,$2E,$F0,$FD,$2C,$3F
	.dc.b	$AB,$7B,$DF,$D1,$89,$5F,$CC,$DF,$E9,$3E,$F7,$EB,$BD,$E9,$3F,$CE
	.dc.b	$E3,$FA,$F1,$FB,$2F,$82,$91,$92,$1F,$CC,$A3,$9B,$FC,$FD,$AE,$4C
	.dc.b	$10,$C1,$0F,$E6,$2F,$07,$F9,$9F,$C0,$7F,$5E,$23,$FD,$2D,$EB,$8F
	.dc.b	$F3,$09,$F3,$CF,$F0,$BF,$83,$AF,$F3,$28,$B3,$F9,$6F,$84,$FE,$F5
	.dc.b	$F0,$9F,$DA,$89,$5F,$D5,$DF,$EF,$7F,$42,$7F,$D9,$4D,$8F,$F5,$D3
	.dc.b	$5F,$ED,$0F,$FB,$61,$CF,$C4,$FF,$1A,$F7,$24,$60,$FE,$D1,$F0,$DF
	.dc.b	$DE,$88,$FF,$67,$7C,$37,$E0,$24,$78,$23,$C7,$5D,$23,$F9,$2D,$61
	.dc.b	$FE,$E9,$1F,$E0,$62,$3F,$DC,$DF,$6F,$27,$B3,$82,$4F,$B2,$3D,$FA
	.dc.b	$48,$F7,$BF,$0B,$6F,$F7,$67,$E4,$CC,$7F,$C4,$2D,$AF,$F1,$2B,$51
	.dc.b	$0A,$E5,$C0,$D3,$FD,$F6,$FC,$EC,$7C,$AF,$CE,$C7,$C8,$FD,$0C,$6E
	.dc.b	$DD,$EF,$FD,$D1,$3F,$E1,$BC,$1B,$F9,$89,$1E,$67,$D3,$BF,$DE,$EB
	.dc.b	$93,$FE,$F7,$A4,$FC,$03,$F0,$7C,$F2,$75,$19,$7E,$0B,$D1,$A4,$3D
	.dc.b	$E2,$7A,$B9,$3E,$1B,$87,$6D,$73,$1D,$5C,$BF,$4B,$8E,$47,$D2,$6B
	.dc.b	$A3,$CE,$43,$F3,$3E,$0F,$A5,$F9,$13,$73,$F9,$9D,$BF,$80,$E5,$FB
	.dc.b	$E4,$AF,$BE,$FE,$A3,$CA,$FD,$39,$23,$F4,$FE,$57,$EA,$BC,$6F,$D5
	.dc.b	$F8,$72,$7F,$57,$1F,$8C,$EC,$FC,$59,$23,$F1,$6F,$57,$F6,$37,$D5
	.dc.b	$FD,$8D,$F5,$7F,$6B,$7D,$5F,$DA,$DF,$E6,$3B,$7F,$2E,$48,$FC,$BB
	.dc.b	$FC,$03,$D2,$FD,$FF,$AB,$27,$F7,$F1,$FC,$0B,$D2,$91,$FC,$09,$FE
	.dc.b	$0B,$E9,$7E,$93,$C1,$FD,$19,$27,$F4,$6F,$57,$F8,$4D,$F5,$7F,$84
	.dc.b	$DF,$47,$F8,$5D,$DE,$0F,$CE,$61,$FB,$0F,$07,$F5,$E4,$AF,$D7,$BF
	.dc.b	$B2,$F1,$7F,$8B,$FB,$72,$7F,$35,$1F,$B8,$3F,$8D,$FB,$32,$32,$43
	.dc.b	$F1,$F1,$CD,$FD,$DB,$5C,$98,$21,$82,$1F,$9A,$78,$3F,$9C,$F1,$FF
	.dc.b	$83,$91,$FE,$0E,$F5,$C7,$F8,$47,$4A,$BD,$15,$FC,$2A,$7F,$86,$7D
	.dc.b	$1B,$E0,$F3,$3F,$BF,$AB,$5F,$C0,$E3,$F8,$BF,$91,$FC,$58,$8F,$F1
	.dc.b	$67,$FD,$73,$7F,$ED,$E8,$FF,$51,$36,$7B,$FC,$3F,$F1,$97,$C3,$F0
	.dc.b	$D7,$C6,$FE,$39,$4F,$87,$C6,$1F,$C8,$21,$F0,$F8,$84,$61,$F6,$57
	.dc.b	$2A,$FE,$5B,$E6,$FF,$2C,$24,$FF,$2C,$7C,$F9,$73,$BF,$35,$F8,$A3
	.dc.b	$18,$E7,$0D,$63,$31,$EE,$79,$63,$F9,$DF,$A1,$FC,$E8,$91,$FC,$E9
	.dc.b	$F8,$BF,$0F,$DF,$E1,$E4,$F2,$3D,$71,$E3,$F8,$69,$19,$D1,$1E,$7F
	.dc.b	$87,$FE,$93,$E9,$FF,$48,$25,$7F,$48,$7B,$BC,$70,$EE,$FD,$4C,$3A
	.dc.b	$BF,$5A,$EF,$07,$E1,$30,$FE,$AB,$EB,$7F,$54,$25,$7F,$54,$7F,$AB
	.dc.b	$FA,$DF,$27,$E1,$24,$FF,$3F,$8F,$EB,$A7,$CD,$F8,$69,$19,$21,$FC
	.dc.b	$D2,$39,$BF,$D8,$1A,$E4,$C1,$0C,$10,$FE,$7E,$F0,$7F,$A0,$FB,$7F
	.dc.b	$DB,$08,$FF,$6C,$7A,$E3,$FA,$69,$F3,$CF,$CF,$F8,$7A,$FF,$61,$8B
	.dc.b	$3F,$A8,$F8,$7F,$EF,$3E,$F7,$F7,$82,$57,$F7,$87,$F4,$7F,$09,$F8
	.dc.b	$D2,$47,$DB,$F8,$4F,$C8,$7C,$3F,$E4,$7E,$1E,$4F,$DF,$8F,$F0,$2F
	.dc.b	$C1,$7F,$81,$09,$1F,$E0,$47,$AB,$F8,$5B,$EA,$FE,$C2,$FA,$BF,$89
	.dc.b	$BE,$AF,$ED,$2F,$FC,$23,$F0,$5F,$E1,$02,$47,$F8,$41,$FD,$DF,$C3
	.dc.b	$7E,$97,$E1,$A4,$FE,$4E,$3F,$51,$F0,$B2,$3F,$2C,$FE,$AF,$E1,$BF
	.dc.b	$6D,$F9,$32,$BF,$BB,$D7,$F8,$6A,$30,$E1,$AF,$F7,$C6,$BF,$E1,$CE
	.dc.b	$CC,$F0,$FF,$18,$C7,$C7,$3F,$C3,$AF,$C4,$FF,$10,$C6,$D7,$F1,$EC
	.dc.b	$7C,$77,$E2,$A6,$28,$C3,$63,$C9,$18,$BF,$E2,$1C,$7F,$EE,$9C,$3F
	.dc.b	$C6,$8D,$F0,$BF,$C6,$98,$2B,$E0,$9C,$34,$98,$FE,$FE,$74,$D9,$EC
	.dc.b	$FC,$36,$BF,$0D,$AF,$F8,$FA,$38,$B9,$38,$10,$C1,$0E,$19,$7B,$88
	.dc.b	$FD,$17,$BE,$3F,$E4,$F9,$EA,$DA,$F8,$3C,$BE,$CB,$F4,$10,$4F,$C4
	.dc.b	$E3,$FC,$0F,$17,$CC,$24,$FD,$F7,$E9,$68,$EE,$F9,$A4,$9F,$E4,$91
	.dc.b	$FE,$E3,$6F,$4F,$D2,$E7,$91,$83,$E6,$72,$FD,$A2,$3F,$8D,$7B,$DF
	.dc.b	$58,$F3,$F0,$7F,$9A,$F0,$F1,$47,$83,$BB,$EF,$E3,$DA,$36,$7F,$9C
	.dc.b	$A5,$8E,$A3,$EE,$72,$FE,$28,$8F,$E7,$5F,$7F,$8B,$F8,$AF,$22,$4F
	.dc.b	$BA,$3F,$1B,$E3,$48,$F8,$2F,$E4,$3C,$6F,$D5,$F5,$7E,$58,$93,$FA
	.dc.b	$97,$AB,$F2,$2F,$AB,$E8,$19,$7F,$80,$DE,$A4,$34,$AF,$C1,$1F,$97
	.dc.b	$8F,$D2,$91,$FD,$C3,$FA,$8E,$CF,$D2,$FA,$12,$7F,$3D,$1F,$C0,$8F
	.dc.b	$D4,$F9,$92,$32,$43,$F8,$04,$73,$7F,$82,$35,$C9,$82,$18,$21,$FA
	.dc.b	$57,$83,$FA,$7F,$03,$F7,$A4,$7F,$88,$3D,$79,$BF,$41,$3E,$79,$FC
	.dc.b	$8F,$AD,$5F,$95,$D4,$9C,$57,$F0,$2F,$4B,$F8,$FF,$85,$FC,$28,$97
	.dc.b	$FC,$75,$FC,$FF,$BD,$FC,$2B,$DA,$93,$F9,$E8,$FE,$1B,$EE,$48,$FD
	.dc.b	$0B,$FC,$43,$DA,$FE,$57,$E4,$FF,$16,$24,$FF,$29,$7A,$BF,$A8,$BE
	.dc.b	$AF,$DC,$BE,$8F,$DE,$BB,$97,$F1,$AC,$3F,$9B,$F9,$5F,$C8,$89,$5F
	.dc.b	$CD,$5F,$E5,$1F,$63,$FE,$F7,$AF,$F6,$A3,$7B,$7F,$DA,$8D,$AF,$F6
	.dc.b	$BB,$FE,$FB,$7F,$24,$BE,$17,$F9,$5C,$C9,$FF,$11,$4E,$5F,$E4,$DE
	.dc.b	$67,$F3,$42,$3F,$D1,$DF,$1F,$C3,$FF,$90,$E7,$C1,$26,$9E,$1C,$3C
	.dc.b	$F8,$77,$9F,$CF,$E1,$C6,$3F,$AB,$F9,$BF,$D0,$89,$3F,$D5,$5F,$E8
	.dc.b	$BE,$8F,$7F,$83,$93,$FC,$EE,$3F,$AF,$9E,$3F,$86,$91,$92,$1F,$D7
	.dc.b	$63,$9B,$FD,$85,$AE,$4C,$10,$C1,$0F,$E8,$4F,$07,$FA,$27,$A7,$FD
	.dc.b	$68,$8F,$F6,$D7,$AE,$5F,$CF,$67,$CF,$3F,$43,$E1,$6B,$F9,$E8,$B3
	.dc.b	$F7,$BE,$1B,$FB,$DF,$A9,$FD,$98,$95,$FD,$E5,$FC,$2F,$C1,$FD,$3F
	.dc.b	$83,$93,$F7,$23,$EB,$7C,$2C,$8F,$BE,$FD,$9F,$85,$FF,$03,$FB,$3F
	.dc.b	$DC,$89,$3F,$E0,$57,$AB,$F8,$6B,$EA,$FE,$32,$FA,$3F,$8E,$BB,$97
	.dc.b	$E5,$B0,$FF,$09,$FB,$5F,$DF,$89,$5F,$E1,$17,$F3,$BF,$0B,$F8,$B2
	.dc.b	$47,$E5,$3E,$17,$F1,$DF,$09,$F8,$FF,$85,$93,$F9,$88,$FF,$4A,$7C
	.dc.b	$2F,$F8,$36,$0D,$23,$D8,$FC,$FD,$F6,$7F,$4B,$7D,$9F,$D1,$DF,$67
	.dc.b	$F5,$37,$DB,$E0,$BF,$C3,$58,$1A,$49,$FF,$17,$C3,$E0,$BF,$4D,$F0
	.dc.b	$92,$7F,$D2,$13,$96,$CF,$EA,$FE,$12,$46,$08,$08,$59,$E4,$E0,$C4
	.dc.b	$F9,$88,$F9,$1A,$EC,$F8,$EE,$1E,$51,$D1,$B1,$97,$C3,$0F,$85,$F8
	.dc.b	$64,$8F,$82,$F6,$DD,$DB,$C2,$FE,$B9,$FC,$26,$6D,$7B,$AB,$6F,$A8
	.dc.b	$FE,$2B,$7F,$78,$FA,$24,$7E,$7B,$FE,$4D,$FC,$C3,$FD,$D9,$1F,$C7
	.dc.b	$0F,$E2,$C7,$FB,$65,$CB,$FC,$93,$3F,$E3,$44,$7F,$91,$E1,$F8,$CC
	.dc.b	$FE,$F9,$23,$EE,$BD,$F8,$3A,$C7,$73,$FD,$99,$7D,$B1,$FE,$6F,$87
	.dc.b	$59,$F0,$1D,$8F,$C1,$E9,$F8,$C2,$3F,$8A,$7D,$9D,$1E,$92,$47,$6E
	.dc.b	$AF,$67,$47,$B7,$A2,$4F,$8A,$3F,$4B,$C9,$F9,$82,$47,$E5,$9E,$AF
	.dc.b	$A2,$FA,$BF,$E1,$9C,$D8,$76,$8F,$6E,$1D,$9D,$61,$FB,$2E,$4F,$D2
	.dc.b	$92,$3F,$46,$FE,$A3,$8F,$E9,$78,$32,$7F,$47,$1F,$BA,$3E,$B7,$74
	.dc.b	$8C,$90,$FC,$C7,$1A,$7A,$47,$EC,$1A,$E9,$AF,$E7,$1A,$FE,$69,$AF
	.dc.b	$EA,$DF,$DE,$F3,$48,$FD,$DB,$5E,$87,$8F,$87,$F2,$EF,$F1,$09,$FC
	.dc.b	$77,$8F,$EB,$8D,$73,$F7,$7F,$0B,$E1,$FF,$18,$EB,$FE,$16,$48,$FE
	.dc.b	$12,$FE,$6F,$D0,$FC,$C7,$97,$27,$F1,$B1,$F9,$BF,$3A,$47,$E4,$1F
	.dc.b	$CF,$79,$7F,$C9,$3C,$0F,$E2,$E4,$9F,$E2,$AF,$57,$F2,$F7,$D5,$FC
	.dc.b	$05,$F4,$7F,$05,$77,$5F,$7F,$33,$1F,$E6,$3E,$07,$F2,$42,$5F,$F2
	.dc.b	$27,$F6,$9E,$E7,$EC,$49,$1F,$A6,$F7,$3F,$69,$EC,$7E,$D7,$D6,$93
	.dc.b	$FA,$A8,$FE,$7B,$E2,$7F,$30,$24,$7F,$2E,$7A,$BF,$B0,$BE,$AF,$E6
	.dc.b	$6F,$AB,$FB,$4B,$EA,$FE,$76,$FF,$A4,$F8,$DF,$CF,$09,$1F,$CE,$9F
	.dc.b	$E7,$FE,$37,$F0,$9F,$7E,$4F,$F3,$A8,$FE,$A2,$7F,$0D,$F8,$09,$19
	.dc.b	$21,$F9,$88,$E6,$FF,$46,$6B,$93,$04,$30,$43,$F9,$AB,$C1,$FE,$6F
	.dc.b	$E5,$7F,$54,$23,$FD,$49,$EB,$C1,$F9,$6E,$B5,$7A,$2B,$F8,$BF,$C1
	.dc.b	$D7,$FD,$31,$3A,$7D,$35,$F8,$0F,$83,$FE,$DF,$E6,$7F,$62,$24,$FF
	.dc.b	$60,$7F,$DF,$DF,$3C,$F8,$3F,$EA,$F7,$FA,$68,$7C,$1E,$3B,$BF,$AD
	.dc.b	$C6,$30,$FE,$BD,$5B,$BD,$47,$F6,$19,$FF,$80,$0D,$F8,$6C,$A8,$CB
	.dc.b	$7E,$17,$FC,$0B,$EA,$7F,$78,$24,$7F,$76,$7C,$F7,$CA,$BC,$53,$FE
	.dc.b	$B9,$6E,$1E,$38,$F2,$E3,$E8,$C7,$81,$23,$27,$9F,$1F,$F0,$AF,$C4
	.dc.b	$FF,$82,$09,$1F,$E0,$67,$E3,$FC,$37,$84,$91,$C8,$F6,$7C,$37,$97
	.dc.b	$E1,$BC,$DF,$0D,$27,$5A,$27,$FC,$47,$EC,$7F,$85,$89,$3F,$E1,$47
	.dc.b	$BB,$C3,$0E,$EF,$BF,$0E,$CF,$4C,$3B,$3F,$0E,$FE,$7F,$E4,$FF,$C8
	.dc.b	$23,$FC,$20,$D7,$FC,$2A,$D7,$FC,$30,$D4,$BD,$DF,$A1,$F0,$92,$7E
	.dc.b	$04,$E5,$B3,$F5,$7E,$1A,$46,$08,$7F,$86,$E7,$83,$67,$FC,$3C,$F8
	.dc.b	$9A,$F8,$9B,$1C,$23,$D7,$18,$23,$CA,$34,$FB,$A3,$2F,$87,$E0,$39
	.dc.b	$91,$E0,$EB,$8B,$FA,$3D,$79,$9B,$5F,$91,$F8,$7F,$9A,$CB,$F6,$CF
	.dc.b	$EA,$7E,$1F,$E8,$FC,$17,$A0,$93,$E5,$7F,$63,$F0,$BF,$9B,$F8,$49
	.dc.b	$3F,$86,$8F,$CF,$7C,$3C,$8F,$C4,$BF,$A1,$F8,$7F,$BB,$F0,$DF,$10
	.dc.b	$93,$F0,$5E,$AF,$E4,$2F,$AB,$FA,$7B,$E8,$FE,$A6,$EE,$5F,$C0,$70
	.dc.b	$FC,$4D,$7D,$02,$57,$CD,$7F,$84,$E7,$FB,$32,$47,$E7,$B6,$FD,$B6
	.dc.b	$7F,$B7,$CE,$4F,$E8,$A3,$F2,$B9,$FD,$C2,$47,$D9,$7A,$BF,$A8,$BE
	.dc.b	$AF,$EF,$EF,$AB,$FA,$CB,$EA,$FF,$03,$BF,$D0,$E7,$F8,$82,$47,$E1
	.dc.b	$5F,$C5,$E5,$FC,$53,$DA,$FF,$76,$4B,$4F,$F0,$3C,$61,$FC,$1C,$FE
	.dc.b	$B0,$FE,$25,$D7,$2B,$F4,$71,$97,$F2,$4F,$66,$90,$FD,$13,$96,$99
	.dc.b	$3F,$6E,$E5,$A7,$E7,$77,$FD,$19,$1C,$1D,$6B,$CE,$F9,$0F,$E2,$D6
	.dc.b	$ED,$F3,$47,$BF,$0F,$E0,$78,$62,$FE,$0F,$57,$F0,$AE,$1F,$D8,$92
	.dc.b	$FF,$5C,$FE,$DB,$17,$8F,$F2,$FC,$49,$5E,$08,$F9,$DE,$2C,$8F,$1B
	.dc.b	$F4,$7C,$2F,$E2,$FD,$1F,$BE,$24,$FE,$ED,$ED,$5D,$23,$D0,$7F,$AB
	.dc.b	$F0,$FE,$C4,$E0,$3E,$47,$B9,$0D,$2B,$F1,$47,$F1,$CE,$8F,$E1,$E4
	.dc.b	$7F,$86,$3E,$FF,$53,$F1,$04,$8F,$9B,$EA,$7E,$2F,$CF,$FC,$67,$9F
	.dc.b	$27,$E8,$C7,$F3,$4F,$03,$F8,$D9,$23,$F8,$C3,$D5,$FA,$F7,$D5,$F3
	.dc.b	$DF,$57,$ED,$DF,$57,$D7,$7F,$D0,$3C,$0F,$E4,$E4,$8F,$E4,$8F,$F2
	.dc.b	$AE,$FF,$D0,$7A,$F2,$7F,$9D,$C7,$F4,$83,$F4,$9E,$C4,$8C,$90,$FE
	.dc.b	$75,$1C,$DF,$E6,$0D,$72,$60,$86,$08,$7F,$54,$78,$3F,$D5,$BC,$2F
	.dc.b	$E7,$E4,$7F,$9E,$3D,$79,$BF,$9A,$CF,$9E,$7F,$3B,$EF,$D7,$F0,$5E
	.dc.b	$1F,$F6,$F7,$D1,$5F,$BC,$E2,$FE,$CB,$E4,$7F,$4C,$24,$FF,$49,$7F
	.dc.b	$39,$F0,$BF,$C1,$7C,$C9,$3F,$AA,$8F,$E1,$1F,$09,$23,$F5,$AF,$F0
	.dc.b	$BF,$82,$FE,$E3,$E6,$FF,$58,$24,$FF,$55,$7A,$BF,$B4,$BE,$AF,$E5
	.dc.b	$2F,$A3,$F9,$6B,$B9,$7F,$14,$C3,$FB,$EF,$9B,$FD,$90,$95,$FD,$85
	.dc.b	$FE,$D3,$F9,$87,$FB,$77,$C2,$4C,$8C,$37,$E1,$7F,$91,$3F,$C8,$FE
	.dc.b	$17,$F9,$27,$C3,$48,$FE,$13,$1F,$E0,$CF,$47,$FB,$99,$23,$FB,$83
	.dc.b	$E0,$AF,$F1,$FD,$F6,$93,$78,$7C,$71,$D0,$FE,$2B,$87,$77,$F6,$B8
	.dc.b	$7F,$92,$3F,$1C,$7A,$9F,$E0,$46,$CF,$F8,$21,$AF,$43,$DE,$E1,$C2
	.dc.b	$F5,$3C,$BF,$09,$27,$FC,$75,$39,$6C,$FA,$BE,$16,$46,$08,$77,$13
	.dc.b	$C5,$C3,$5F,$0C,$FA,$48,$FA,$5A,$F0,$6B,$C0,$D1,$EF,$F6,$3E,$11
	.dc.b	$1C,$1A,$3F,$91,$4F,$8F,$8B,$FD,$C2,$DE,$0F,$A3,$B3,$E1,$FE,$76
	.dc.b	$33,$0F,$6E,$1F,$EE,$8A,$7C,$3F,$D3,$F6,$BE,$91,$2B,$E9,$3F,$90
	.dc.b	$F8,$5F,$C0,$FC,$2C,$9F,$F3,$35,$A7,$F0,$BF,$0F,$27,$2F,$E2,$3E
	.dc.b	$1F,$F0,$1F,$05,$F7,$C9,$3F,$7D,$EC,$F4,$C3,$B3,$FD,$07,$0E,$CF
	.dc.b	$42,$19,$BE,$CF,$E0,$3E,$0B,$F1,$84,$7F,$18,$FE,$AB,$E1,$BF,$36
	.dc.b	$48,$F5,$FC,$3F,$E7,$BE,$1F,$F3,$FF,$0D,$27,$DF,$1F,$99,$F8,$5F
	.dc.b	$CC,$12,$3F,$30,$F5,$7E,$2D,$F5,$7D,$57,$D5,$F9,$37,$D5,$F7,$5F
	.dc.b	$E9,$7E,$1B,$F4,$84,$8F,$D2,$3F,$A7,$BF,$D9,$E1,$27,$F3,$F1,$FA
	.dc.b	$D3,$F6,$F3,$23,$24,$3F,$37,$1C,$DF,$D8,$35,$C9,$82,$18,$21,$FB
	.dc.b	$27,$83,$FB,$4C,$3F,$78,$47,$F7,$8F,$5E,$2F,$CC,$CF,$9E,$7F,$17
	.dc.b	$B5,$7F,$3F,$16,$7F,$31,$9F,$F0,$9C,$BF,$84,$12,$BF,$84,$3F,$C9
	.dc.b	$BA,$BF,$8B,$F2,$49,$FC,$BC,$7F,$1A,$E8,$91,$F9,$A7,$F8,$EF,$27
	.dc.b	$F1,$5D,$FF,$8A,$12,$7F,$8A,$3D,$5F,$D0,$DF,$57,$F2,$D7,$D1,$FC
	.dc.b	$C5,$DC,$BF,$61,$87,$F2,$1E,$4F,$E4,$04,$AF,$E4,$0F,$FA,$9B,$F3
	.dc.b	$0F,$F7,$D4,$7F,$B4,$DA,$FF,$88,$DB,$4E,$8F,$E6,$31,$FE,$E8,$6F
	.dc.b	$83,$FE,$F6,$87,$83,$4C,$47,$F3,$3E,$7F,$E6,$44,$C3,$F9,$93,$E3
	.dc.b	$F1,$F9,$12,$3C,$9E,$3F,$27,$CD,$7D,$5F,$45,$FF,$3F,$E8,$FE,$7C
	.dc.b	$48,$FE,$7C,$FF,$41,$EA,$F3,$79,$52,$7F,$9C,$C7,$F4,$73,$D5,$E5
	.dc.b	$C8,$C9,$0F,$E6,$B1,$CD,$FE,$94,$D7,$26,$08,$60,$87,$F4,$C7,$83
	.dc.b	$FD,$3B,$BB,$FA,$B1,$1F,$EA,$CF,$5D,$FF,$99,$CF,$9E,$7E,$C7,$A7
	.dc.b	$5F,$E7,$31,$67,$EC,$47,$F7,$AD,$DE,$7F,$F6,$5F,$03,$FB,$21,$27
	.dc.b	$FB,$23,$F8,$5F,$7B,$EE,$7B,$72,$7F,$0D,$1F,$7F,$D9,$91,$F8,$97
	.dc.b	$F0,$5E,$BF,$F7,$1F,$1F,$FB,$81,$27,$FB,$83,$D5,$FC,$85,$F5,$7F
	.dc.b	$21,$7D,$1F,$C9,$5D,$CF,$D7,$0C,$D8,$FF,$7F,$F1,$FF,$BF,$12,$7F
	.dc.b	$BF,$3F,$9F,$F8,$5F,$CB,$12,$3F,$41,$F0,$BF,$99,$F8,$1F,$CD,$7C
	.dc.b	$04,$9F,$D2,$47,$F8,$2F,$CF,$FF,$05,$12,$3F,$C1,$4F,$57,$F5,$57
	.dc.b	$D5,$FD,$55,$F5,$7F,$5D,$7D,$5F,$D7,$5F,$F8,$73,$3F,$6F,$E3,$BF
	.dc.b	$E0,$59,$FF,$0F,$B5,$FF,$19,$B5,$34,$79,$DF,$B3,$F8,$39,$3F,$E3
	.dc.b	$AA,$EB,$3D,$FC,$0E,$AF,$EE,$7E,$12,$56,$E7,$2D,$3C,$33,$E3,$C3
	.dc.b	$BC,$79,$DA,$FF,$87,$5A,$FF,$8B,$5A,$F4,$35,$1C,$C8,$F3,$6B,$A3
	.dc.b	$B7,$A3,$E1,$23,$83,$81,$E7,$3E,$2C,$7B,$DF,$49,$FC,$7F,$E1,$B0
	.dc.b	$E8,$E1,$5E,$AC,$7E,$3C,$7F,$25,$F8,$6D,$CB,$D7,$F8,$44,$9F,$84
	.dc.b	$FD,$7F,$AF,$FF,$74,$47,$8D,$7C,$37,$81,$B1,$9F,$87,$C7,$C4,$48
	.dc.b	$E0,$FF,$2D,$DB,$C3,$EE,$7D,$12,$4F,$D1,$7C,$15,$D7,$09,$3D,$4E
	.dc.b	$64,$3F,$DC,$18,$30,$ED,$1F,$0B,$DC,$FB,$C4,$7E,$F3,$ED,$F8,$6F
	.dc.b	$51,$23,$DD,$F1,$1E,$DF,$87,$F7,$7C,$44,$9F,$85,$1F,$2B,$E0,$3F
	.dc.b	$14,$48,$FC,$53,$D5,$F9,$17,$D5,$F3,$DF,$57,$E6,$5F,$57,$D7,$7F
	.dc.b	$5B,$E0,$BF,$2C,$48,$FC,$B3,$F9,$9F,$81,$FA,$5F,$0B,$27,$F0,$31
	.dc.b	$F7,$8F,$AD,$F0,$D2,$32,$43,$F3,$D1,$CD,$FD,$13,$5C,$98,$21,$82
	.dc.b	$1F,$A5,$78,$3F,$A7,$F8,$6F,$D7,$11,$FD,$73,$D7,$9F,$F8,$3C,$F9
	.dc.b	$E7,$F1,$DF,$11,$54,$E9,$F4,$3F,$CA,$B8,$5F,$11,$F9,$CF,$88,$FD
	.dc.b	$E1,$2B,$F7,$8F,$E6,$F4,$FC,$C6,$32,$7F,$39,$1F,$9B,$CA,$47,$E7
	.dc.b	$DF,$CF,$57,$EA,$32,$FE,$10,$49,$FE,$10,$F5,$7F,$4D,$7D,$5F,$CA
	.dc.b	$DF,$47,$F2,$F7,$72,$F5,$61,$FB,$4C,$7F,$8A,$12,$BF,$8A,$3F,$B3
	.dc.b	$E8,$FD,$81,$23,$F6,$9C,$FF,$B3,$E4,$FD,$A7,$24,$9F,$DC,$47,$F0
	.dc.b	$0E,$0F,$E4,$04,$8F,$E4,$0F,$57,$F7,$D7,$D5,$FD,$6D,$F5,$7F,$81
	.dc.b	$5F,$57,$F6,$37,$FC,$33,$83,$F9,$61,$23,$F9,$63,$FC,$BF,$93,$F8
	.dc.b	$47,$64,$9F,$E2,$F1,$FC,$50,$FE,$19,$DB,$23,$24,$3F,$9A,$47,$37
	.dc.b	$F9,$C3,$5C,$98,$21,$82,$1F,$CE,$DE,$0F,$F3,$DE,$6F,$E8,$84,$7F
	.dc.b	$A2,$3D,$77,$FE,$BB,$3E,$79,$FE,$6D,$E0,$7F,$BF,$FB,$13,$99,$FC
	.dc.b	$CF,$67,$F6,$28,$FE,$51,$1F,$CC,$BA,$3F,$A9,$11,$FE,$A4,$FF,$67
	.dc.b	$BF,$F6,$B4,$76,$F3,$3F,$9F,$C7,$FB,$52,$3C,$3E,$86,$3C,$D9,$1E
	.dc.b	$47,$7E,$47,$F4,$0E,$DF,$EC,$04,$9F,$EC,$0F,$7A,$EB,$84,$9E,$A7
	.dc.b	$32,$1E,$C0,$30,$61,$E0,$8F,$EA,$7E,$07,$F6,$F2,$3F,$DB,$DF,$77
	.dc.b	$B3,$EC,$24,$7C,$CF,$67,$DF,$EA,$FC,$0F,$4A,$4F,$D0,$8F,$EB,$7E
	.dc.b	$27,$F7,$B2,$47,$F7,$B7,$AB,$F5,$AF,$AB,$E9,$3D,$17,$D9,$FB,$77
	.dc.b	$D9,$D6,$4C,$3F,$B4,$F8,$9F,$E0,$92,$4F,$F8,$25,$FF,$0F,$7C,$17
	.dc.b	$F8,$FB,$97,$FC,$7F,$6F,$80,$FB,$27,$F7,$89,$D7,$2F,$CA,$C1,$18
	.dc.b	$7E,$6D,$3E,$03,$F0,$27,$1A,$93,$FE,$2F,$AF,$F2,$6B,$CC,$E7,$1E
	.dc.b	$42,$3B,$6B,$B3,$66,$48,$66,$3A,$BE,$76,$BF,$82,$5E,$4F,$C2,$24
	.dc.b	$79,$5E,$F8,$FF,$98,$A7,$DD,$3E,$03,$E7,$FC,$17,$B1,$FA,$D1,$A3
	.dc.b	$78,$BE,$0F,$E5,$79,$7F,$40,$95,$F0,$DF,$D5,$57,$EA,$3E,$0E,$4F
	.dc.b	$EE,$23,$F5,$7F,$0B,$23,$F7,$6F,$EB,$BE,$17,$EB,$7A,$1F,$70,$93
	.dc.b	$F3,$DE,$AF,$F0,$2B,$EA,$FE,$22,$FA,$3F,$8A,$BB,$A7,$E1,$30,$FC
	.dc.b	$17,$A1,$F8,$82,$57,$DB,$7F,$81,$7C,$27,$F0,$02,$47,$F1,$0F,$85
	.dc.b	$FE,$07,$F0,$BF,$C1,$3E,$16,$4F,$F1,$48,$FC,$77,$AF,$F9,$42,$47
	.dc.b	$E1,$DE,$AF,$F1,$AB,$EA,$FE,$82,$FA,$BF,$C7,$AF,$AB,$FA,$4B,$FC
	.dc.b	$D7,$AF,$FA,$02,$47,$E4,$DF,$D1,$F1,$7F,$95,$BF,$30,$7C,$1D,$3F
	.dc.b	$45,$12,$BF,$CC,$13,$FE,$F2,$6F,$C2,$65,$FC,$1C,$F4,$7F,$40,$D7
	.dc.b	$F2,$CD,$7F,$45,$C3,$FC,$55,$D9,$0D,$94,$DC,$7E,$EB,$DA,$91,$FA
	.dc.b	$D6,$BC,$0F,$19,$FE,$77,$9E,$CF,$F0,$DB,$7C,$37,$A7,$8B,$F4,$18
	.dc.b	$5E,$AF,$E0,$14,$F8,$6F,$E0,$9F,$03,$FC,$28,$98,$FE,$F5,$FF,$55
	.dc.b	$4B,$E1,$3E,$67,$C2,$4A,$F9,$31,$F3,$FE,$1A,$47,$CC,$7E,$97,$C3
	.dc.b	$7F,$11,$F8,$6F,$E2,$E4,$9F,$E1,$4F,$6A,$E9,$1E,$BB,$FB,$51,$D2
	.dc.b	$3D,$D7,$75,$FF,$6E,$E2,$C7,$F8,$F7,$C3,$7F,$24,$25,$FF,$17,$7D
	.dc.b	$79,$FE,$14,$91,$F8,$2C,$FF,$11,$3F,$89,$F8,$59,$3F,$86,$8F,$E5
	.dc.b	$51,$FC,$C0,$91,$FC,$8D,$EA,$FE,$32,$BA,$BF,$56,$BA,$BF,$90,$9E
	.dc.b	$AF,$D9,$AF,$E6,$D9,$FF,$3C,$24,$7F,$2F,$7F,$9F,$E9,$F9,$DD,$E4
	.dc.b	$FE,$0E,$3F,$9F,$1F,$A1,$E2,$91,$92,$1F,$CB,$63,$9B,$FC,$ED,$AE
	.dc.b	$4C,$10,$C1,$0F,$CF,$BC,$1F,$D0,$ED,$FD,$50,$8F,$F4,$77,$AE,$DF
	.dc.b	$CC,$E7,$CF,$3F,$B2,$EA,$AF,$F3,$28,$B3,$F7,$B9,$BF,$AD,$6F,$FD
	.dc.b	$80,$95,$FD,$4D,$FC,$BF,$87,$FC,$04,$91,$FB,$EF,$07,$F8,$27,$83
	.dc.b	$FC,$17,$BA,$4F,$F0,$28,$FE,$CD,$CF,$FD,$B0,$91,$FD,$7D,$EA,$FF
	.dc.b	$08,$BE,$AF,$EC,$6F,$AB,$FC,$32,$FA,$BF,$B5,$BF,$EE,$5C,$DF,$DE
	.dc.b	$09,$1F,$DA,$DF,$EC,$8F,$F8,$7B,$EC,$7F,$DC,$6F,$F8,$25,$FF,$70
	.dc.b	$B7,$CC,$FF,$79,$C6,$B5,$74,$FF,$7D,$8F,$F6,$B5,$7F,$BF,$5A,$F3
	.dc.b	$F2,$3F,$93,$4F,$F8,$17,$B7,$FC,$16,$49,$FE,$FC,$F9,$15,$77,$9E
	.dc.b	$D8,$78,$8F,$1F,$9B,$E4,$C2,$5F,$85,$EE,$FF,$19,$EE,$FF,$0C,$91
	.dc.b	$FF,$06,$3D,$FF,$16,$78,$7E,$EF,$52,$57,$F9,$0B,$5F,$7B,$FE,$44
	.dc.b	$AD,$B9,$CF,$FC,$3D,$18,$8F,$95,$E9,$FC,$83,$FC,$45,$5C,$A1,$3B
	.dc.b	$8C,$3C,$63,$FC,$3C,$D7,$AE,$1E,$72,$3F,$6A,$34,$F9,$63,$2F,$F8
	.dc.b	$86,$32,$F8,$23,$2F,$67,$E1,$78,$B2,$3A,$35,$E5,$1E,$AC,$3C,$71
	.dc.b	$F0,$4F,$01,$F8,$1F,$57,$4E,$0F,$F2,$A5,$FB,$E7,$8C,$7E,$3F,$D7
	.dc.b	$DB,$F1,$3C,$6F,$A8,$47,$DC,$FE,$17,$85,$F0,$BF,$94,$24,$FE,$47
	.dc.b	$E0,$7F,$2F,$F0,$3F,$98,$F7,$64,$FE,$56,$3B,$E6,$F2,$FF,$04,$49
	.dc.b	$F9,$8F,$7C,$3A,$C7,$E9,$5F,$D1,$E1,$DA,$3F,$41,$87,$67,$F5,$58
	.dc.b	$76,$F2,$FF,$1C,$48,$FB,$0F,$AF,$E1,$3F,$5B,$F0,$B2,$7F,$55,$1F
	.dc.b	$B0,$F8,$59,$1F,$AD,$7F,$67,$F0,$BE,$AF,$4B,$F3,$44,$9F,$C2,$3D
	.dc.b	$5F,$DA,$5F,$57,$F7,$D7,$D1,$FE,$01,$77,$9B,$F2,$DC,$0C,$7E,$3F
	.dc.b	$A5,$FA,$72,$5F,$E4,$5F,$D5,$7A,$5F,$C1,$BE,$0A,$4F,$E1,$4F,$CF
	.dc.b	$C7,$D0,$3F,$85,$FC,$1C,$8D,$72,$27,$51,$F9,$D7,$AB,$48,$7E,$75
	.dc.b	$CB,$4C,$9F,$AA,$72,$D3,$F5,$1E,$B7,$F0,$12,$38,$3A,$E7,$FB,$49
	.dc.b	$F1,$C7,$F2,$5F,$85,$B3,$98,$F4,$CF,$F2,$6F,$85,$FC,$6F,$B1,$FC
	.dc.b	$34,$93,$FB,$77,$F8,$F7,$E6,$0F,$88,$FD,$0B,$67,$E4,$35,$FE,$68
	.dc.b	$7F,$32,$F8,$7F,$E6,$7F,$0F,$23,$F9,$84,$7E,$73,$DE,$FE,$38,$48
	.dc.b	$FE,$0C,$F8,$2F,$B9,$23,$F9,$C1,$D4,$FE,$75,$5D,$5F,$E8,$95,$FA
	.dc.b	$8F,$7B,$F9,$41,$23,$F8,$93,$EC,$F8,$5F,$5F,$C2,$C9,$F5,$C7,$BB
	.dc.b	$E1,$64,$7B,$9F,$83,$F0,$DF,$B4,$F8,$6F,$E6,$84,$9F,$E3,$CF,$57
	.dc.b	$E2,$DF,$57,$FD,$C1,$AB,$0E,$B1,$8C,$2F,$0F,$F3,$1A,$63,$FC,$0B
	.dc.b	$E1,$BF,$A1,$12,$FF,$96,$3F,$D1,$7E,$17,$EA,$92,$3F,$2E,$FF,$0A
	.dc.b	$3E,$CF,$C3,$48,$C9,$0F,$E6,$51,$CD,$FE,$70,$D7,$26,$08,$60,$87
	.dc.b	$F4,$47,$83,$FD,$1A,$3F,$AD,$11,$FE,$88,$F5,$C3,$FA,$6C,$F9,$E7
	.dc.b	$F1,$BA,$57,$F9,$AC,$59,$F8,$D7,$FC,$9B,$3F,$EC,$C4,$AF,$EA,$0F
	.dc.b	$E6,$B9,$BF,$32,$48,$FC,$CF,$1F,$E7,$38,$FF,$3B,$C1,$27,$F3,$B1
	.dc.b	$FC,$CB,$83,$FB,$91,$23,$FA,$E3,$D5,$FD,$1D,$F5,$7E,$FD,$F5,$7F
	.dc.b	$4F,$7D,$5F,$C1,$DF,$F3,$EE,$0F,$EF,$C4,$8F,$ED,$0F,$EC,$BC,$1F
	.dc.b	$D8,$F8,$32,$7F,$63,$1F,$B4,$EC,$91,$FB,$47,$F6,$FD,$1F,$D2,$B8
	.dc.b	$FF,$C1,$44,$9F,$EE,$8F,$57,$F7,$77,$D5,$FC,$BD,$F4,$7F,$33,$77
	.dc.b	$83,$F2,$D8,$7F,$56,$E7,$FF,$0B,$12,$BF,$C0,$0F,$F8,$9F,$F1,$54
	.dc.b	$7F,$0A,$F2,$23,$FC,$C6,$32,$D3,$F8,$7F,$91,$1C,$10,$D9,$D2,$3F
	.dc.b	$C6,$89,$A4,$7C,$44,$3F,$C9,$AE,$5A,$64,$EA,$73,$3D,$5D,$31,$E2
	.dc.b	$D7,$71,$E4,$C7,$FC,$54,$DF,$2B,$DA,$D9,$F8,$7D,$5F,$E0,$F7,$D4
	.dc.b	$7F,$83,$A3,$4F,$84,$FE,$5F,$E6,$E7,$B3,$E6,$12,$3D,$6F,$F9,$57
	.dc.b	$5F,$F7,$0C,$7F,$5D,$6A,$C7,$FC,$7D,$18,$6F,$AB,$FE,$1E,$87,$AB
	.dc.b	$7D,$29,$7C,$E3,$CD,$E0,$FD,$D2,$47,$CC,$7C,$5E,$D7,$32,$47,$8F
	.dc.b	$0E,$D5,$D2,$3C,$9B,$7E,$0B,$16,$3A,$9F,$C3,$78,$5F,$8B,$23,$F6
	.dc.b	$9F,$81,$F0,$5E,$9F,$7A,$4F,$C1,$8F,$67,$C0,$48,$F8,$8F,$BF,$DE
	.dc.b	$FA,$1E,$4F,$E5,$C9,$3F,$86,$7A,$BF,$2E,$FA,$BE,$83,$F3,$B9,$30
	.dc.b	$ED,$19,$3F,$C1,$18,$63,$83,$E8,$E3,$F7,$FC,$9F,$D3,$12,$BF,$2A
	.dc.b	$FE,$A7,$CD,$FA,$9F,$07,$27,$F5,$11,$F8,$73,$EC,$7C,$24,$8C,$90
	.dc.b	$FC,$AC,$73,$7F,$42,$D7,$26,$08,$60,$87,$EC,$39,$D1,$C6,$31,$E8
	.dc.b	$7E,$FC,$8F,$EB,$9E,$D9,$7E,$92,$7D,$13,$F9,$5F,$86,$AF,$C7,$BE
	.dc.b	$AF,$A6,$BF,$2F,$F0,$DF,$A5,$F4,$BF,$86,$12,$7F,$76,$FE,$87,$E1
	.dc.b	$3F,$34,$48,$FD,$17,$C2,$7E,$77,$E1,$BF,$3D,$F0,$D2,$7F,$4D,$1F
	.dc.b	$B1,$F5,$7F,$8C,$12,$3F,$83,$BD,$5F,$D6,$5F,$57,$EF,$72,$B0,$ED
	.dc.b	$1F,$B1,$C3,$B3,$AC,$3F,$7D,$ED,$7F,$25,$24,$7F,$14,$7F,$75,$F0
	.dc.b	$DF,$B3,$F8,$69,$3F,$BB,$8F,$DB,$7C,$2C,$8F,$DF,$3F,$B9,$F8,$6F
	.dc.b	$E1,$5E,$F7,$F3,$12,$4F,$F2,$07,$AB,$FC,$16,$FA,$BE,$AB,$E8,$FB
	.dc.b	$2E,$F0,$FF,$08,$C3,$F8,$B7,$BB,$FC,$F4,$95,$FC,$B1,$FE,$81,$EF
	.dc.b	$7F,$AA,$3F,$30,$FF,$77,$DF,$FB,$35,$BE,$1F,$F2,$49,$FE,$40,$7F
	.dc.b	$11,$65,$FF,$45,$8C,$BF,$CE,$1E,$AD,$21,$FC,$C9,$CB,$4C,$9F,$99
	.dc.b	$72,$D3,$FA,$1F,$C0,$FF,$57,$23,$83,$AD,$FF,$3C,$9F,$1C,$7E,$F6
	.dc.b	$DF,$0D,$8A,$D4,$FA,$4F,$F4,$74,$BE,$1F,$F9,$F7,$C1,$FF,$65,$24
	.dc.b	$FF,$54,$7F,$A8,$BF,$DA,$25,$8F,$CB,$24,$74,$9F,$9D,$F0,$DF,$3F
	.dc.b	$E1,$A4,$F8,$23,$FA,$61,$FD,$CC,$8F,$F6,$27,$BA,$BE,$EF,$A2,$FB
	.dc.b	$BE,$FB,$EE,$FB,$2F,$BB,$F0,$EF,$FA,$D4,$7F,$80,$49,$1F,$DC,$1E
	.dc.b	$F7,$FE,$E6,$8F,$8B,$C5,$F8,$73,$F1,$1C,$7F,$89,$DE,$4F,$CA,$7F
	.dc.b	$19,$B7,$F6,$7D,$BF,$C1,$C4,$9F,$EF,$AF,$6A,$E9,$1D,$8F,$A6,$F4
	.dc.b	$7E,$D5,$5D,$BF,$2D,$7F,$DC,$F6,$FF,$0D,$12,$BF,$C1,$2F,$F8,$77
	.dc.b	$E1,$FF,$3D,$D7,$27,$FC,$39,$1F,$DE,$CF,$D1,$75,$C8,$C9,$0F,$F0
	.dc.b	$4C,$73,$7F,$C2,$AD,$72,$60,$86,$08,$7F,$35,$78,$3F,$E2,$6F,$C4
	.dc.b	$9C,$7F,$79,$B1,$F8,$E6,$BF,$E2,$06,$BC,$CF,$36,$5F,$E1,$CA,$FD
	.dc.b	$E7,$81,$ED,$9C,$DF,$F8,$7A,$7F,$80,$78,$3F,$E3,$CB,$EB,$83,$A3
	.dc.b	$C4,$4C,$3C,$4F,$E1,$3C,$FF,$E1,$04,$8F,$D4,$F9,$7F,$C2,$FC,$BF
	.dc.b	$E1,$9E,$3C,$9F,$D6,$C7,$7D,$3F,$C9,$78,$BB,$7E,$01,$2B,$E0,$3D
	.dc.b	$DF,$DA,$E3,$DD,$FD,$E6,$3D,$DF,$DC,$E3,$DD,$FE,$01,$8F,$77,$FC
	.dc.b	$BB,$83,$B7,$E7,$12,$3E,$73,$FA,$EF,$CC,$3F,$DD,$1A,$FF,$23,$9F
	.dc.b	$E4,$C7,$31,$1A,$D3,$FC,$CB,$D2,$F5,$3F,$96,$12,$BF,$86,$FA,$5C
	.dc.b	$72,$78,$1F,$7C,$93,$F7,$DF,$17,$AF,$E0,$8F,$E2,$A7,$87,$2F,$03
	.dc.b	$FC,$47,$0E,$CF,$F1,$F7,$F9,$E6,$1D,$FC,$6F,$C6,$12,$7F,$18,$FE
	.dc.b	$3F,$C5,$E9,$EE,$49,$FF,$41,$ED,$F8,$49,$F4,$1E,$1F,$72,$4E,$B7
	.dc.b	$FB,$73,$A8,$FC,$C3,$D5,$A4,$3F,$34,$E5,$A6,$4F,$D2,$39,$69,$FA
	.dc.b	$3F,$1B,$F4,$C4,$70,$75,$BF,$43,$E5,$3F,$03,$3F,$27,$E0,$BC,$D1
	.dc.b	$FE,$41,$E0,$FF,$02,$3B,$0E,$1F,$F2,$36,$CF,$82,$FB,$5E,$4F,$ED
	.dc.b	$C9,$3F,$B7,$6B,$FB,$76,$BF,$BF,$6B,$FA,$26,$BF,$96,$7F,$81,$79
	.dc.b	$B2,$3F,$81,$35,$FE,$04,$D7,$F8,$53,$5F,$D6,$B5,$FD,$0B,$5F,$6F
	.dc.b	$99,$FC,$3C,$8F,$F0,$F6,$BF,$C3,$DA,$FF,$17,$6B,$FB,$A6,$BF,$AE
	.dc.b	$7F,$8D,$79,$F2,$3F,$8D,$35,$FE,$35,$E7,$FE,$6F,$E1,$24,$7D,$1C
	.dc.b	$D3,$FA,$73,$F4,$1F,$0D,$27,$51,$97,$F9,$33,$FC,$A4,$8E,$08,$78
	.dc.b	$0A,$7F,$31,$76,$E4,$87,$F3,$5F,$53,$F9,$A1,$1F,$E6,$8F,$0E,$0F
	.dc.b	$E5,$33,$E3,$9F,$C3,$FC,$35,$9C,$C7,$A0,$FC,$76,$EF,$86,$FE,$09
	.dc.b	$EB,$FF,$41,$24,$FF,$41,$7F,$A2,$7A,$FF,$C0,$7E,$0E,$4F,$C7,$8F
	.dc.b	$E1,$87,$F0,$5F,$84,$91,$92,$1F,$D2,$63,$9B,$FD,$35,$AE,$4C,$10
	.dc.b	$C1,$0F,$EB,$0F,$07,$FA,$D7,$B3,$FD,$64,$8F,$F5,$97,$AE,$5F,$D3
	.dc.b	$A7,$CF,$3F,$0B,$E1,$AB,$FB,$38,$B3,$FD,$83,$37,$C3,$7F,$29,$F7
	.dc.b	$FF,$B3,$12,$FF,$B3,$3F,$DA,$7D,$FF,$E4,$DF,$09,$27,$F1,$91,$FC
	.dc.b	$C4,$FE,$55,$EE,$C8,$C9,$0F,$ED,$D1,$CD,$FE,$E4,$D7,$26,$08,$60
	.dc.b	$87,$F7,$D7,$83,$FD,$FF,$E0,$3F,$BF,$11,$FE,$FC,$F5,$CB,$FB,$9C
	.dc.b	$F9,$E7,$F0,$FF,$0B,$5F,$E3,$11,$67,$E3,$7C,$37,$F4,$DF,$82,$FF
	.dc.b	$05,$12,$BF,$C1,$4F,$F9,$2B,$CA,$7C,$4F,$F4,$BF,$87,$95,$FC,$86
	.dc.b	$72,$D9,$FE,$A1,$F0,$F2,$30,$40,$42,$CF,$27,$06,$27,$C4,$47,$FC
	.dc.b	$D7,$1B,$76,$D7,$A3,$5E,$47,$08,$E4,$F6,$3F,$B6,$BE,$FF,$86,$91
	.dc.b	$D9,$AE,$E7,$CD,$97,$F8,$B9,$BE,$47,$82,$FF,$B8,$1F,$E0,$18,$CD
	.dc.b	$78,$85,$71,$9F,$EE,$D7,$BC,$7E,$81,$2B,$DE,$FD,$3C,$7F,$C1,$9F
	.dc.b	$68,$DE,$9D,$E2,$57,$F5,$59,$69,$C0,$CC,$78,$B4,$F8,$F5,$D6,$3E
	.dc.b	$6B,$5F,$80,$D7,$E7,$B5,$E2,$70,$21,$C0,$FA,$FB,$FE,$2C,$8E,$0D
	.dc.b	$E9,$F2,$E7,$CA,$FF,$96,$B0,$71,$63,$0E,$F7,$E9,$3F,$13,$7F,$13
	.dc.b	$83,$E8,$70,$7E,$64,$93,$F8,$87,$F3,$9C,$5F,$2B,$96,$4F,$E6,$E3
	.dc.b	$EC,$1F,$3B,$9E,$46,$48,$7E,$0E,$39,$BF,$94,$6B,$93,$04,$30,$43
	.dc.b	$F3,$2F,$07,$F3,$7C,$9F,$B1,$23,$FA,$07,$AE,$BF,$91,$9F,$3C,$FD
	.dc.b	$FE,$EA,$FC,$53,$FC,$93,$18,$9F,$99,$D7,$F9,$5E,$9F,$DF,$12,$BF
	.dc.b	$5A,$FF,$01,$E8,$FC,$7F,$85,$27,$F8,$04,$7E,$74,$FC,$9F,$7C,$8C
	.dc.b	$90,$FD,$44,$73,$7F,$72,$D7,$26,$08,$60,$87,$EF,$9E,$0F,$F0,$0E
	.dc.b	$DF,$E2,$C4,$7F,$82,$BD,$71,$FD,$B4,$F9,$E7,$F4,$9E,$25,$7E,$37
	.dc.b	$67,$F4,$87,$D1,$5F,$91,$F1,$3F,$75,$DF,$FC,$8C,$93,$FC,$49,$FE
	.dc.b	$4D,$E0,$FE,$D3,$C9,$93,$FC,$96,$3F,$81,$1F,$B8,$F1,$A4,$64,$87
	.dc.b	$F0,$D8,$E6,$FF,$1E,$6B,$93,$04,$30,$43,$F9,$1B,$C1,$FE,$4B,$E1
	.dc.b	$FF,$3B,$23,$FC,$A9,$EB,$8F,$F1,$B9,$F3,$CF,$F0,$9F,$32,$BF,$1A
	.dc.b	$2C,$FE,$D7,$CA,$FE,$3D,$E4,$7F,$47,$25,$7F,$36,$7F,$A5,$78,$FF
	.dc.b	$C6,$3D,$19,$3F,$D2,$63,$F9,$31,$FC,$73,$79,$19,$21,$FC,$C6,$39
	.dc.b	$BF,$D0,$9A,$E4,$C1,$0C,$10,$FE,$8E,$F0,$7F,$A4,$F9,$7F,$D7,$C8
	.dc.b	$FF,$4E,$7A,$E3,$FC,$FE,$7C,$F3,$FC,$B7,$D7,$AF,$E2,$A2,$CF,$F1
	.dc.b	$9F,$5F,$FA,$17,$9B,$FD,$AC,$95,$FD,$69,$FE,$DD,$E7,$FF,$B1,$FD
	.dc.b	$A7,$B5,$4F,$E9,$51,$2B,$F9,$E7,$8D,$FD,$08,$C3,$F9,$A8,$E6,$FF
	.dc.b	$66,$6B,$FD,$65,$AF,$F6,$86,$BF,$DA,$1A,$FF,$74,$6B,$FE,$D2,$F4
	.dc.b	$7F,$C0,$84,$7F,$B9,$BD,$B2,$FE,$C9,$3E,$45,$5F,$DF,$F0,$7C,$0E
	.dc.b	$63,$C1,$87,$F8,$2E,$FE,$17,$BD,$FD,$A7,$D6,$FF,$09,$12,$7F,$C0
	.dc.b	$2F,$F7,$97,$FC,$17,$F8,$77,$F8,$23,$2E,$F1,$FE,$0E,$C9,$F0,$BF
	.dc.b	$34,$FF,$7A,$56,$F2,$FF,$09,$36,$B8,$7C,$2F,$D3,$36,$C3,$C1,$1E
	.dc.b	$17,$C2,$57,$09,$A4,$74,$43,$99,$4F,$F0,$A3,$C3,$D9,$E8,$48,$F0
	.dc.b	$BB,$C7,$DF,$1D,$CF,$31,$FD,$07,$17,$C3,$E2,$77,$7E,$FA,$E0,$7F
	.dc.b	$B6,$E9,$F0,$FF,$1B,$DB,$F6,$12,$7E,$0B,$EF,$F8,$0F,$CC,$7C,$2C
	.dc.b	$9F,$F7,$26,$09,$F9,$E7,$E7,$7E,$16,$4E,$A3,$2F,$CC,$7A,$34,$87
	.dc.b	$C4,$72,$D3,$27,$F9,$22,$7C,$0E,$0F,$BB,$1F,$57,$E0,$A4,$7D,$A6
	.dc.b	$BC,$CF,$ED,$13,$C9,$B3,$FB,$5F,$86,$F3,$C6,$0F,$51,$F2,$7E,$1F
	.dc.b	$F1,$BF,$07,$F8,$32,$47,$E2,$1F,$C3,$7C,$27,$EF,$FE,$1E,$4F,$C8
	.dc.b	$8F,$CA,$9F,$C0,$FE,$1E,$46,$48,$7E,$1A,$39,$BF,$94,$6B,$93,$04
	.dc.b	$30,$43,$E4,$3C,$1F,$95,$C1,$F9,$B2,$3F,$A0,$7A,$F0,$FE,$2A,$7C
	.dc.b	$F3,$F8,$99,$AF,$CA,$8B,$3F,$96,$97,$C3,$FE,$C8,$FD,$41,$2B,$F5
	.dc.b	$AF,$EA,$F0,$FE,$49,$8C,$9F,$C6,$C7,$EE,$8F,$E5,$18,$48,$C9,$0F
	.dc.b	$D5,$C7,$37,$F7,$2D,$72,$60,$86,$08,$7E,$35,$E0,$FE,$3F,$1F,$E0
	.dc.b	$04,$7F,$82,$BD,$71,$FD,$7C,$F9,$E7,$F5,$DA,$57,$F1,$F1,$67,$E3
	.dc.b	$6D,$FC,$5B,$5F,$E1,$84,$AF,$E2,$2F,$F0,$FD,$7F,$A4,$72,$49,$FD
	.dc.b	$8C,$7F,$1E,$3F,$A6,$71,$48,$C9,$0F,$E1,$F1,$CD,$FE,$3A,$D7,$26
	.dc.b	$08,$60,$87,$EC,$5E,$0F,$EC,$F8,$BF,$92,$11,$FE,$52,$F5,$C7,$F8
	.dc.b	$9C,$F9,$E7,$FD,$8B,$1F,$C7,$3F,$30,$74,$E6,$7F,$6B,$56,$BF,$19
	.dc.b	$1F,$CF,$39,$3F,$99,$11,$FE,$6E,$FF,$35,$E6,$FF,$68,$4F,$5E,$BB
	.dc.b	$D3,$FE,$D3,$C9,$1F,$D1,$AA,$61,$E4,$AD,$76,$F0,$8F,$F6,$DD,$A3
	.dc.b	$71,$FC,$2A,$BC,$0F,$F4,$86,$BF,$CF,$9A,$FF,$19,$6B,$FD,$51,$AF
	.dc.b	$F5,$E6,$BF,$D3,$7A,$7F,$A8,$11,$FE,$AA,$F6,$C7,$F9,$FC,$F9,$1F
	.dc.b	$99,$E1,$DC,$7B,$5F,$F3,$8A,$F6,$2C,$5E,$07,$F6,$8E,$DF,$EB,$C4
	.dc.b	$9F,$EC,$6F,$F6,$1E,$DE,$34,$F1,$A9,$FD,$C2,$25,$7C,$B3,$E4,$F8
	.dc.b	$7C,$0C,$BF,$CC,$A3,$A3,$FD,$C1,$AF,$F6,$66,$BF,$D0,$5A,$FF,$7C
	.dc.b	$6B,$FE,$06,$6B,$FD,$D7,$C1,$FE,$F0,$47,$FB,$EB,$DB,$5F,$EC,$F3
	.dc.b	$E4,$7F,$29,$E4,$5D,$7B,$5F,$F6,$0A,$F8,$3E,$57,$F8,$3B,$C3,$FF
	.dc.b	$02,$12,$7F,$C1,$2F,$F8,$1F,$C4,$F8,$FE,$8C,$9F,$ED,$31,$FE,$16
	.dc.b	$3E,$5F,$9D,$23,$24,$3F,$C0,$F1,$CD,$FF,$0A,$B5,$C9,$82,$18,$21
	.dc.b	$FD,$A5,$E0,$88,$3F,$0E,$FF,$07,$B8,$31,$97,$F8,$8E,$33,$1C,$1A
	.dc.b	$FF,$8D,$A7,$D8,$FA,$74,$FD,$F7,$AB,$88,$FE,$E3,$36,$7F,$81,$7A
	.dc.b	$7F,$94,$B7,$97,$DC,$97,$FE,$43,$8C,$B6,$32,$F9,$CF,$53,$87,$D5
	.dc.b	$1C,$1D,$35,$CC,$6D,$F6,$79,$7E,$F2,$3E,$16,$BC,$9A,$F6,$6B,$C8
	.dc.b	$E0,$43,$C3,$E6,$FC,$B2,$3E,$D6,$BA,$30,$43,$04,$3D,$AF,$07,$DF
	.dc.b	$E6,$FD,$72,$3F,$25,$AF,$C2,$6B,$C0,$D1,$0D,$14,$F9,$7E,$7F,$E0
	.dc.b	$C8,$FD,$56,$BF,$39,$AF,$03,$44,$34,$53,$EB,$F9,$FF,$8F,$23,$F8
	.dc.b	$16,$BF,$69,$AF,$03,$44,$34,$53,$F0,$7E,$8F,$E6,$C8,$FE,$35,$FC
	.dc.b	$F7,$A3,$FC,$7B,$D7,$93,$FC,$6C,$FD,$34,$7E,$5C,$FE,$47,$EC,$C8
	.dc.b	$D6,$A9,$D4,$7E,$65,$EA,$D2,$1F,$94,$72,$D3,$27,$E6,$9C,$B4,$FD
	.dc.b	$27,$AB,$FB,$52,$38,$3A,$F0,$7F,$A2,$AB,$C9,$1E,$6A,$FE,$77,$F0
	.dc.b	$3E,$68,$56,$E7,$0F,$F7,$7F,$77,$F6,$DE,$BF,$F0,$32,$57,$ED,$5F
	.dc.b	$E0,$DE,$CF,$F4,$6F,$84,$93,$F1,$23,$F7,$C7,$F4,$AF,$80,$91,$92
	.dc.b	$1F,$B6,$8E,$6F,$F0,$16,$B9,$30,$43,$04,$3F,$80,$BC,$1F,$E0,$7E
	.dc.b	$DF,$F1,$B2,$3F,$C3,$5E,$B8,$FC,$79,$F3,$CF,$E0,$7C,$47,$C3,$62
	.dc.b	$3E,$3F,$67,$F2,$3A,$C5,$FF,$87,$7E,$1B,$F8,$E7,$BF,$FC,$A4,$97
	.dc.b	$FC,$6D,$FE,$59,$EF,$FF,$74,$FC,$C1,$F0,$F4,$FE,$49,$12,$BF,$25
	.dc.b	$2F,$87,$E0,$66,$3C,$67,$F3,$DE,$55,$76,$8F,$E5,$4D,$7F,$91,$B5
	.dc.b	$FE,$4E,$D7,$F9,$F3,$5E,$8D,$7F,$A4,$7C,$27,$F4,$52,$3F,$CD,$DE
	.dc.b	$D9,$FE,$46,$7C,$CF,$FB,$73,$07,$C3,$63,$6E,$F8,$7E,$AA,$FE,$27
	.dc.b	$C3,$7F,$45,$F8,$6F,$EA,$44,$9F,$E8,$8F,$F5,$5B,$F9,$59,$49,$FD
	.dc.b	$9C,$7F,$4D,$3E,$74,$C8,$C9,$0F,$E8,$B1,$CD,$FE,$A0,$D7,$26,$08
	.dc.b	$60,$87,$F5,$07,$83,$FD,$4A,$BF,$B5,$11,$FE,$B8,$F5,$CB,$F6,$D3
	.dc.b	$E7,$9F,$BF,$B5,$7E,$E9,$FD,$DE,$31,$3F,$2F,$1F,$ED,$5B,$7F,$77
	.dc.b	$25,$7F,$69,$7F,$BD,$6D,$F8,$FE,$09,$3F,$C6,$23,$FB,$91,$F9,$3E
	.dc.b	$39,$19,$21,$FD,$AA,$39,$BF,$DD,$5A,$E4,$C1,$0C,$10,$FE,$EA,$F0
	.dc.b	$7F,$C2,$1F,$7F,$FA,$E7,$FB,$E0,$70,$FF,$7F,$6B,$C3,$8F,$C3,$97
	.dc.b	$F1,$FA,$F5,$CF,$EA,$3A,$EB,$F2,$B8,$FF,$54,$FB,$2B,$F2,$9D,$1F
	.dc.b	$E6,$6C,$5C,$BE,$2A,$F1,$36,$33,$1E,$66,$BE,$5E,$8F,$DE,$77,$48
	.dc.b	$FE,$95,$29,$EC,$7F,$01,$EB,$93,$A8,$CB,$D5,$F0,$11,$C1,$0F,$01
	.dc.b	$4F,$96,$ED,$C9,$0F,$9F,$D5,$EB,$23,$EB,$78,$64,$E8,$FF,$23,$E1
	.dc.b	$E9,$AF,$31,$FC,$53,$C0,$F3,$BF,$96,$8D,$37,$0F,$E3,$DE,$07,$CD
	.dc.b	$EE,$F9,$44,$9F,$94,$FC,$DE,$EF,$E4,$9E,$14,$9F,$93,$1F,$54,$FE
	.dc.b	$51,$E1,$48,$C9,$0F,$C3,$47,$37,$EB,$35,$C9,$82,$18,$21,$F9,$47
	.dc.b	$83,$F9,$6F,$0B,$F0,$44,$7F,$04,$F5,$D7,$E4,$47,$F9,$BF,$35,$7A
	.dc.b	$63,$F9,$DF,$89,$5F,$91,$16,$BF,$93,$78,$9F,$93,$F1,$3F,$20,$4A
	.dc.b	$FC,$83,$F9,$3F,$1F,$FD,$7B,$F8,$0B,$C8,$A7,$E6,$E2,$57,$F4,$AE
	.dc.b	$3F,$EA,$06,$1F,$B1,$1C,$DF,$CE,$35,$FD,$83,$5F,$DB,$B5,$FC,$2B
	.dc.b	$5F,$80,$D7,$F7,$5E,$47,$EA,$48,$FE,$A5,$ED,$97,$C8,$8F,$E0,$13
	.dc.b	$E5,$8C,$60,$F3,$F1,$87,$82,$FF,$5B,$7F,$0B,$CC,$FD,$D7,$99,$FB
	.dc.b	$72,$4F,$ED,$DF,$DD,$79,$9F,$23,$D2,$93,$FB,$68,$FE,$04,$7C,$CF
	.dc.b	$3E,$46,$48,$7F,$13,$8E,$6F,$F0,$36,$B9,$30,$43,$04,$3F,$90,$BC
	.dc.b	$1F,$E4,$7E,$87,$F0,$F2,$3F,$C3,$DE,$BA,$FC,$7E,$45,$7A,$2B,$FC
	.dc.b	$B1,$B3,$D4,$C4,$7E,$D6,$6D,$FC,$CF,$5B,$F9,$07,$A5,$FC,$74,$95
	.dc.b	$FC,$75,$FE,$41,$EA,$7E,$3B,$DC,$93,$FC,$72,$3F,$94,$1F,$92,$F6
	.dc.b	$A4,$64,$87,$F3,$98,$E6,$FF,$29,$6B,$93,$04,$30,$43,$FA,$3B,$C1
	.dc.b	$FE,$93,$EB,$7F,$35,$23,$FC,$D5,$EB,$8F,$C7,$9F,$3C,$FC,$6F,$7E
	.dc.b	$BF,$C6,$62,$CF,$E4,$3D,$CF,$E8,$9E,$D7,$F4,$12,$57,$F4,$17,$FA
	.dc.b	$27,$B5,$FB,$2F,$81,$93,$FD,$02,$3F,$A6,$1F,$B6,$F8,$39,$19,$21
	.dc.b	$FD,$76,$39,$BF,$D3,$5A,$E4,$C1,$0C,$10,$FE,$D6,$F0,$7F,$B6,$FB
	.dc.b	$DF,$D6,$48,$FF,$59,$7A,$E3,$F8,$B9,$F3,$CF,$E2,$BE,$1E,$BF,$CF
	.dc.b	$62,$CF,$ED,$3E,$1F,$FB,$47,$C0,$7F,$65,$25,$7F,$65,$7F,$B4,$7C
	.dc.b	$0F,$F1,$6F,$87,$93,$FD,$92,$3F,$B8,$1F,$C6,$BE,$1E,$46,$48,$7F
	.dc.b	$80,$63,$9B,$FD,$C5,$AE,$4C,$10,$C1,$0F,$F0,$7B,$C1,$10,$4F,$F8
	.dc.b	$05,$A7,$F8,$0D,$C1,$9E,$A4,$FA,$A7,$CF,$8F,$F1,$8F,$87,$C4,$7F
	.dc.b	$65,$9B,$3F,$C7,$3E,$1F,$C0,$FF,$21,$C1,$F0,$FD,$C9,$7F,$E2,$48
	.dc.b	$CB,$7E,$1F,$FA,$0C,$C8,$FF,$6C,$4A,$7A,$F8,$BE,$D8,$C4,$FF,$06
	.dc.b	$9F,$E9,$26,$9F,$87,$1A,$7F,$CA,$2D,$7E,$31,$F1,$A3,$E2,$11,$F2
	.dc.b	$10,$D9,$B3,$07,$28,$E8,$F9,$63,$93,$C2,$32,$F9,$32,$FA,$44,$7E
	.dc.b	$23,$C3,$0F,$95,$5E,$B9,$F0,$1F,$E6,$AF,$47,$EB,$BF,$1A,$35,$7E
	.dc.b	$FA,$E3,$3F,$CA,$8C,$E5,$F8,$12,$3F,$4D,$FC,$24,$F7,$B7,$15,$3A
	.dc.b	$C4,$AF,$C7,$DF,$F5,$FA,$6F,$C4,$D4,$FD,$0A,$EA,$FE,$05,$AF,$C5
	.dc.b	$6B,$F2,$DA,$FD,$F6,$BF,$71,$AF,$E7,$B7,$FC,$E1,$1F,$C7,$3D,$AF
	.dc.b	$F2,$93,$E4,$7D,$1C,$B7,$5E,$D7,$F9,$0A,$F8,$5C,$5F,$33,$8F,$F5
	.dc.b	$04,$9F,$CD,$3F,$AB,$E3,$F9,$1C,$F2,$7F,$33,$1F,$50,$F9,$9D,$32
	.dc.b	$32,$43,$F0,$D1,$CD,$FD,$33,$5C,$98,$21,$82,$1F,$93,$78,$3F,$95
	.dc.b	$E7,$FE,$00,$47,$F6,$4F,$5D,$3F,$71,$3E,$79,$FB,$B9,$FF,$0B,$DD
	.dc.b	$E0,$E6,$7F,$57,$56,$FF,$05,$DB,$F9,$2E,$9F,$E1,$C4,$AF,$E0,$0F
	.dc.b	$F1,$1E,$BF,$C7,$78,$92,$7F,$7F,$1F,$9A,$3F,$25,$E0,$48,$C9,$0F
	.dc.b	$D6,$47,$37,$F8,$63,$5C,$98,$21,$82,$1F,$B9,$78,$3F,$BB,$ED,$FE
	.dc.b	$4C,$47,$F8,$C3,$D7,$1F,$E3,$D3,$E7,$9F,$8E,$7F,$32,$D1,$E3,$66
	.dc.b	$3F,$88,$CD,$AF,$D4,$79,$1F,$B8,$F0,$7F,$9A,$12,$BF,$93,$3F,$CD
	.dc.b	$FC,$1F,$DA,$79,$D2,$7F,$92,$C7,$F0,$03,$F7,$1E,$5C,$8C,$90,$FE
	.dc.b	$25,$1C,$DF,$E6,$4D,$72,$60,$86,$08,$7F,$1F,$78,$3F,$C8,$7C,$5F
	.dc.b	$E9,$84,$7F,$9F,$3D,$71,$FE,$89,$3E,$79,$F8,$E3,$47,$A3,$98,$FE
	.dc.b	$6F,$36,$BF,$87,$79,$DF,$C7,$BC,$7F,$EB,$44,$AF,$E9,$8F,$F5,$DF
	.dc.b	$2B,$F8,$D7,$A5,$27,$FA,$5C,$7F,$26,$3F,$8F,$7A,$92,$32,$43,$F9
	.dc.b	$C4,$73,$7F,$AC,$35,$C9,$82,$18,$21,$FD,$15,$E0,$FF,$47,$F3,$3F
	.dc.b	$B9,$11,$FE,$C8,$F5,$C7,$FB,$54,$F9,$E7,$E3,$E8,$F7,$31,$1F,$D7
	.dc.b	$26,$D7,$F3,$3F,$63,$FA,$1F,$A1,$FD,$FC,$95,$FD,$C5,$AF,$F5,$56
	.dc.b	$BF,$D6,$DA,$FF,$6F,$6B,$FD,$B1,$FF,$05,$7A,$12,$3F,$BE,$35,$FE
	.dc.b	$C0,$D7,$FB,$33,$5F,$EF,$4D,$7F,$BB,$B5,$FF,$0C,$FA,$3F,$E1,$72
	.dc.b	$3F,$E0,$97,$FB,$9B,$FE,$20,$FC,$35,$F0,$5F,$D9,$D7,$C2,$53,$11
	.dc.b	$9F,$82,$95,$F9,$2A,$47,$1F,$80,$FE,$E1,$F8,$57,$FA,$EE,$BF,$C0
	.dc.b	$ED,$7F,$C5,$CD,$7F,$C6,$2D,$55,$E4,$7B,$61,$E3,$7C,$18,$78,$96
	.dc.b	$BF,$EC,$9A,$F1,$C7,$97,$1F,$23,$FA,$8C,$1E,$CF,$C2,$84,$61,$F8
	.dc.b	$AD,$7F,$C3,$2D,$7D,$71,$C6,$31,$1E,$C8,$C1,$1E,$6D,$79,$BC,$7D
	.dc.b	$99,$1E,$06,$BE,$06,$BF,$29,$AF,$CA,$6B,$F2,$9A,$FB,$3D,$BF,$31
	.dc.b	$1F,$6B,$5F,$6B,$5F,$AE,$D7,$EB,$B5,$FA,$EF,$C1,$F6,$E4,$7C,$86
	.dc.b	$BF,$21,$AF,$E0,$9A,$FE,$09,$AF,$E0,$9A,$FD,$4F,$73,$E7,$11,$FA
	.dc.b	$AF,$D1,$E5,$FC,$C7,$C3,$C9,$FB,$D1,$F9,$23,$F3,$9F,$0F,$23,$24
	.dc.b	$3E,$EC,$73,$7F,$02,$D7,$26,$08,$60,$87,$E7,$9E,$0F,$E8,$3E,$03
	.dc.b	$F0,$C4,$7F,$1A,$F5,$DD,$E7,$7E,$92,$BD,$15,$F9,$7F,$84,$AF,$E7
	.dc.b	$A7,$4F,$A6,$BF,$41,$F0,$DF,$B0,$F8,$2F,$CA,$12,$7F,$36,$F7,$AE
	.dc.b	$31,$CC,$FD,$8E,$1F,$B8,$C3,$F7,$3F,$0B,$FC,$0F,$E1,$E5,$7E,$A6
	.dc.b	$32,$FE,$99,$FD,$11,$1F,$8B,$97,$B1,$EA,$79,$F0,$FE,$03,$7C,$A3
	.dc.b	$F5,$B5,$C5,$A7,$EC,$BE,$17,$F5,$C4,$70,$7E,$DA,$7E,$2D,$F7,$8F
	.dc.b	$DD,$E3,$EB,$7D,$04,$3E,$01,$FC,$02,$FF,$90,$57,$F1,$8F,$87,$FD
	.dc.b	$E1,$27,$F8,$1B,$FB,$F8,$EF,$8F,$CA,$C3,$C1,$CA,$9F,$05,$F6,$E7
	.dc.b	$61,$FC,$B4,$CD,$F0,$AE,$95,$EB,$21,$FC,$41,$CB,$4C,$9F,$CB,$1C
	.dc.b	$B4,$C9,$FC,$BB,$4F,$E2,$C4,$7F,$8E,$BD,$27,$C5,$1F,$8F,$3F,$93
	.dc.b	$74,$F9,$B2,$FA,$F1,$93,$D9,$AF,$F3,$ED,$FF,$91,$12,$3F,$94,$BD
	.dc.b	$67,$8C,$7D,$0D,$3E,$33,$D8,$FE,$69,$C5,$FD,$47,$69,$18,$21,$FC
	.dc.b	$CA,$3F,$98,$11,$FE,$6F,$8F,$9D,$FE,$97,$A7,$F5,$28,$FE,$A7,$3B
	.dc.b	$9E,$47,$F3,$E7,$F9,$D7,$14,$8F,$E8,$2D,$7F,$A0,$CF,$57,$FA,$3D
	.dc.b	$75,$7C,$BA,$7E,$0A,$FB,$C7,$F6,$5E,$1F,$EC,$7D,$3F,$D1,$C9,$1F
	.dc.b	$D3,$9F,$E9,$58,$75,$CF,$E2,$CF,$C7,$8E,$CF,$E2,$F8,$FF,$BB,$18
	.dc.b	$C7,$F2,$33,$CA,$7E,$19,$0F,$49,$4F,$EE,$6E,$DC,$90,$C9,$B3,$FB
	.dc.b	$BF,$4F,$F5,$F2,$3F,$D9,$9E,$53,$E2,$8F,$B8,$7E,$BB,$F0,$27,$9B
	.dc.b	$E7,$E0,$FF,$00,$E1,$6B,$FC,$25,$CD,$FE,$08,$EC,$FE,$D8,$48,$FE
	.dc.b	$E6,$F6,$AE,$31,$DB,$1F,$E0,$18,$F6,$8F,$F3,$2B,$FE,$3F,$7F,$C2
	.dc.b	$DD,$1F,$E1,$D0,$EB,$C2,$AF,$9D,$FF,$02,$4F,$18,$C9,$C1,$C3,$86
	.dc.b	$BE,$43,$EC,$1E,$5B,$EB,$1E,$58,$EC,$7F,$8C,$9E,$2F,$59,$F0,$9E
	.dc.b	$2F,$06,$46,$DA,$FD,$8C,$3E,$83,$E0,$3E,$86,$9E,$C8,$F8,$B5,$70
	.dc.b	$F1,$B6,$BC,$07,$F8,$EF,$AF,$0B,$17,$8B,$F1,$09,$1E,$97,$EA,$3F
	.dc.b	$E3,$AD,$BE,$E3,$C7,$2E,$74,$F1,$7F,$C9,$DD,$D2,$BE,$BC,$68,$FA
	.dc.b	$24,$7E,$33,$D9,$FF,$1B,$DF,$6A,$F9,$9A,$7E,$0A,$B8,$57,$D8,$3D
	.dc.b	$4E,$1F,$B1,$E4,$FD,$E2,$3F,$4D,$FC,$0E,$3E,$AE,$4F,$F3,$1C,$FE
	.dc.b	$32,$3C,$CF,$A0,$87,$F3,$FF,$03,$DF,$E4,$7E,$28,$91,$F7,$DE,$AF
	.dc.b	$E1,$4F,$90,$75,$3F,$33,$E6,$7F,$9C,$FC,$39,$18,$21,$ED,$D1,$3F
	.dc.b	$98,$24,$7E,$37,$0F,$56,$BF,$49,$FA,$73,$A9,$E4,$D7,$F3,$9E,$67
	.dc.b	$E8,$88,$FE,$5D,$F4,$BF,$8F,$DB,$F2,$27,$EC,$AB,$AC,$79,$A3,$F6
	.dc.b	$B5,$FE,$8D,$E3,$FC,$1F,$9F,$FA,$E2,$57,$E8,$DE,$B5,$D2,$7F,$58
	.dc.b	$F5,$9F,$8C,$7F,$7A,$FC,$39,$E3,$FE,$61,$B1,$A9,$E0,$FC,$7C,$7D
	.dc.b	$2F,$EE,$5A,$FE,$D9,$AF,$EF,$DA,$FE,$6C,$F9,$84,$3E,$97,$A7,$FC
	.dc.b	$20,$8F,$EF,$5F,$5E,$1D,$63,$C8,$FF,$10,$B7,$95,$F5,$67,$27,$84
	.dc.b	$FF,$7F,$C3,$CA,$FD,$5F,$A7,$FC,$58,$93,$FC,$2D,$FE,$33,$EB,$FC
	.dc.b	$3F,$36,$4F,$F2,$48,$FD,$99,$F1,$FC,$E9,$19,$21,$F8,$D8,$E6,$FF
	.dc.b	$17,$6B,$93,$04,$30,$43,$F9,$33,$C1,$FE,$51,$EB,$7F,$2E,$23,$FC
	.dc.b	$8D,$EB,$B7,$F3,$C9,$F3,$CF,$F3,$88,$FF,$1F,$7E,$00,$F8,$5C,$CF
	.dc.b	$F4,$FA,$B5,$FE,$43,$D1,$E9,$7F,$12,$F7,$3F,$9E,$92,$FF,$99,$3E
	.dc.b	$0E,$0F,$E6,$F8,$72,$7F,$A2,$E5,$DE,$BF,$A6,$FB,$9F,$8F,$F5,$A4
	.dc.b	$60,$87,$ED,$63,$FA,$59,$1F,$8D,$9F,$B5,$F9,$EF,$B3,$4F,$E4,$F3
	.dc.b	$B9,$E4,$7F,$47,$7F,$AA,$FC,$04,$8F,$E9,$4D,$7E,$8D,$75,$7E,$21
	.dc.b	$F2,$71,$F5,$DF,$68,$F3,$CF,$F6,$AB,$FD,$57,$AF,$FC,$E3,$DF,$FE
	.dc.b	$C6,$4A,$FE,$B0,$FC,$77,$FB,$26,$9F,$1D,$FE,$DB,$A7,$C7,$7F,$C0
	.dc.b	$5F,$07,$FB,$5F,$6A,$47,$F7,$68,$FE,$8A,$7F,$70,$23,$FD,$95,$FF
	.dc.b	$7B,$4F,$3C,$BD,$84,$D3,$D7,$5C,$CE,$04,$3F,$93,$57,$F7,$CF,$83
	.dc.b	$91,$FD,$C5,$AF,$F4,$7B,$ED,$5F,$82,$8F,$AA,$F6,$7D,$67,$72,$9F
	.dc.b	$C4,$FD,$DF,$EB,$FF,$0F,$FE,$0A,$24,$7F,$7F,$7D,$7B,$FC,$78,$FE
	.dc.b	$DD,$A7,$F7,$0C,$3C,$31,$F2,$8F,$E3,$9E,$F4,$7E,$D1,$92,$88,$FE
	.dc.b	$C0,$FE,$60,$F9,$24,$32,$F6,$7C,$27,$B8,$FB,$35,$E3,$78,$1F,$14
	.dc.b	$A7,$F8,$92,$1F,$0F,$F5,$89,$1D,$9F,$7B,$D4,$EE,$7F,$AC,$A3,$C4
	.dc.b	$7F,$8B,$B5,$FD,$EC,$78,$4F,$E4,$3F,$01,$F1,$7E,$1F,$F0,$64,$8F
	.dc.b	$5B,$D9,$FC,$F9,$27,$F5,$8F,$C6,$AF,$81,$D0,$F8,$3F,$E9,$07,$D1
	.dc.b	$CA,$57,$C6,$7A,$3E,$F6,$BF,$83,$30,$6C,$87,$E2,$DD,$BF,$30,$FA
	.dc.b	$3F,$0F,$F9,$C2,$38,$38,$E1,$E8,$7E,$34,$FF,$8D,$3F,$18,$6F,$E6
	.dc.b	$9C,$9E,$D9,$FE,$C5,$F0,$BF,$8B,$F8,$7F,$D4,$92,$7F,$02,$FC,$7C
	.dc.b	$3B,$13,$8F,$FC,$C7,$8F,$6A,$F9,$1F,$11,$FD,$B7,$E2,$24,$60,$87
	.dc.b	$EA,$63,$F6,$E4,$7D,$75,$FA,$9B,$ED,$1F,$E7,$2B,$EC,$FE,$0E,$B8
	.dc.b	$3C,$CF,$CB,$BF,$C0,$AE,$47,$E6,$9A,$FC,$F3,$E1,$10,$FF,$7F,$CF
	.dc.b	$6A,$FE,$93,$5F,$AF,$24,$FF,$6E,$7F,$65,$8F,$F1,$02,$47,$EA,$1F
	.dc.b	$91,$87,$AC,$91,$EB,$CF,$F7,$5A,$FF,$84,$F3,$93,$F9,$18,$CB,$FB
	.dc.b	$27,$F8,$E9,$1F,$8E,$7E,$1C,$FB,$A7,$AA,$78,$BF,$C5,$E3,$83,$4F
	.dc.b	$DE,$F0,$FF,$28,$23,$83,$A9,$FC,$1F,$5F,$C5,$66,$C7,$F0,$38,$76
	.dc.b	$8F,$ED,$F8,$59,$FF,$0C,$6D,$FC,$5F,$8F,$F9,$A9,$31,$FE,$1A,$FC
	.dc.b	$67,$D5,$7F,$80,$5B,$7D,$3B,$FE,$DB,$CB,$FE,$39,$CE,$57,$ED,$E3
	.dc.b	$F9,$21,$FD,$08,$8F,$F1,$A7,$B3,$EB,$3F,$01,$7E,$B7,$EA,$1C,$CE
	.dc.b	$04,$3F,$A0,$47,$F4,$EE,$89,$1F,$C9,$9A,$FE,$34,$FA,$73,$EB,$C3
	.dc.b	$D5,$1F,$D8,$2B,$CE,$F7,$3F,$CA,$DC,$9F,$CF,$7A,$BF,$AD,$92,$3F
	.dc.b	$99,$BE,$AE,$5F,$E4,$79,$7E,$FF,$7F,$E8,$9D,$BF,$E5,$DE,$79,$87
	.dc.b	$F1,$B8,$CB,$FC,$F1,$FE,$CE,$47,$EF,$DF,$47,$FA,$3E,$7E,$87,$FA
	.dc.b	$44,$ED,$FE,$C4,$D7,$F9,$F7,$7F,$F7,$22,$3F,$D2,$9F,$E9,$EF,$DC
	.dc.b	$C3,$F1,$7B,$FE,$92,$30,$79,$C8,$7F,$41,$AF,$F3,$C3,$FD,$87,$C2
	.dc.b	$FE,$FE,$47,$FA,$BB,$5F,$EE,$CD,$7F,$A9,$35,$FF,$1B,$B5,$FF,$09
	.dc.b	$BF,$E0,$AF,$0A,$47,$F6,$26,$BF,$E0,$16,$BF,$D7,$5A,$FF,$90,$9A
	.dc.b	$FF,$87,$9A,$FF,$71,$F1,$7F,$C2,$E4,$7F,$B7,$B5,$FF,$07,$35,$FE
	.dc.b	$D4,$D7,$FC,$9E,$D7,$FC,$56,$FF,$88,$BC,$59,$1F,$DE,$9A,$FF,$86
	.dc.b	$5A,$FF,$75,$6B,$FE,$5A,$6B,$FE,$36,$6B,$FE,$09,$F2,$3F,$C5,$E4
	.dc.b	$7F,$C0,$ED,$7F,$C4,$CD,$7F,$C0,$4D,$7F,$CC,$ED,$7F,$C8,$6F,$E1
	.dc.b	$BC,$FF,$DA,$47,$FB,$D5,$A8,$81,$CB,$86,$BB,$6B,$BC,$BB,$11,$FF
	.dc.b	$11,$C6,$89,$19,$8F,$17,$9F,$E5,$23,$C1,$F4,$7C,$27,$F7,$1F,$2A
	.dc.b	$4F,$F8,$C2,$D3,$E8,$E7,$F6,$C3,$D1,$FE,$F1,$E2,$4B,$E1,$1B,$CB
	.dc.b	$D8,$FC,$52,$47,$A1,$F0,$BF,$B8,$C7,$C2,$B0,$EF,$8F,$38,$F8,$95
	.dc.b	$C9,$F9,$CD,$7E,$69,$F3,$7D,$5F,$A8,$47,$E3,$BE,$07,$FB,$4D,$F8
	.dc.b	$2F,$BC,$7A,$88,$78,$F0,$F5,$BF,$C3,$BC,$6F,$23,$F6,$FD,$7F,$C1
	.dc.b	$11,$FA,$AF,$7A,$FC,$2F,$33,$0F,$06,$1F,$81,$CF,$9E,$AF,$6B,$FC
	.dc.b	$49,$E4,$CC,$FE,$A4,$68,$FC,$81,$1F,$C1,$BD,$EF,$FC,$3F,$87,$69
	.dc.b	$FC,$66,$1C,$5D,$90,$FC,$C5,$7E,$6F,$DC,$91,$F8,$E6,$BF,$9D,$9E
	.dc.b	$8F,$F6,$0A,$F4,$C7,$58,$F9,$07,$E6,$EB,$19,$FC,$2F,$33,$DF,$EF
	.dc.b	$7E,$A4,$95,$F9,$C7,$D4,$FF,$8A,$F0,$EF,$A7,$E5,$23,$F2,$DF,$01
	.dc.b	$FE,$49,$F3,$A4,$60,$87,$EC,$E3,$F6,$C4,$7F,$53,$9F,$AB,$7F,$A9
	.dc.b	$1F,$A1,$8F,$A7,$3B,$8E,$47,$EC,$1F,$E0,$1F,$09,$23,$F6,$6D,$7E
	.dc.b	$31,$F9,$D2,$1F,$10,$FE,$05,$3D,$9F,$3E,$9F,$AE,$AF,$F3,$27,$A1
	.dc.b	$F8,$4F,$85,$FE,$1C,$4A,$FE,$02,$FC,$67,$E9,$1D,$AB,$F3,$6F,$7C
	.dc.b	$DF,$0F,$FE,$6F,$F4,$E4,$FE,$FE,$3F,$22,$7F,$1B,$23,$FC,$39,$ED
	.dc.b	$A7,$C6,$8F,$D4,$9F,$1C,$E4,$6C,$87,$F2,$38,$FE,$4D,$F0,$B2,$3F
	.dc.b	$8C,$B5,$F8,$E7,$42,$61,$EA,$3F,$88,$CF,$9A,$3B,$9F,$E6,$2F,$53
	.dc.b	$F5,$3F,$0F,$FC,$CC,$91,$FC,$99,$F4,$F0,$FE,$2B,$7F,$CB,$CF,$F8
	.dc.b	$23,$D6,$FF,$49,$18,$7E,$44,$7F,$07,$7F,$86,$EB,$F9,$17,$CE,$62
	.dc.b	$34,$FE,$EC,$FC,$69,$0D,$99,$29,$FC,$FF,$E1,$BF,$A5,$11,$FE,$7A
	.dc.b	$FD,$D9,$F0,$67,$F0,$FA,$3F,$D3,$7F,$05,$E7,$8C,$5F,$EF,$8F,$F1
	.dc.b	$1F,$B3,$FC,$33,$E1,$FF,$AB,$12,$3F,$A4,$BD,$72,$E5,$87,$49,$FE
	.dc.b	$85,$F0,$FF,$EB,$5F,$72,$46,$08,$7F,$5F,$8F,$EC,$44,$7E,$06,$3E
	.dc.b	$47,$F3,$E7,$F8,$DB,$2F,$E3,$F3,$B8,$E4,$7F,$5C,$7F,$B6,$7B,$F2
	.dc.b	$3F,$AF,$B5,$FC,$3D,$75,$D3,$E2,$E3,$E9,$9F,$D0,$5F,$67,$CF,$5F
	.dc.b	$CC,$BD,$FF,$E5,$FF,$0F,$FD,$E4,$93,$FD,$B1,$F8,$8F,$E7,$4F,$F5
	.dc.b	$7E,$5F,$CD,$BE,$1F,$FD,$B1,$F0,$32,$7F,$B4,$46,$5F,$EE,$AF,$F8
	.dc.b	$14,$8F,$C4,$CB,$F0,$58,$FC,$5E,$86,$1C,$6B,$FC,$15,$7C,$1A,$7F
	.dc.b	$5E,$3F,$C1,$9F,$80,$F0,$70,$1C,$3B,$6B,$FD,$99,$AF,$8B,$08,$FE
	.dc.b	$24,$F7,$9E,$C2,$1F,$C2,$E7,$FD,$F1,$AF,$F6,$2A,$7C,$47,$F9,$A1
	.dc.b	$B5,$E4,$6B,$E3,$8F,$9A,$FE,$B8,$FC,$21,$FE,$27,$CB,$FC,$57,$2F
	.dc.b	$88,$FE,$DF,$9B,$E1,$E9,$E1,$89,$7D,$DF,$11,$1F,$A0,$7B,$6F,$E8
	.dc.b	$69,$C2,$3E,$34,$ED,$A7,$A7,$E2,$3D,$84,$70,$74,$3F,$3A,$7A,$CF
	.dc.b	$92,$7F,$3F,$C7,$D4,$FB,$75,$FF,$21,$C6,$23,$55,$EE,$AF,$99,$8F
	.dc.b	$CB,$24,$FC,$B7,$E4,$E3,$EA,$8E,$87,$E1,$34,$F8,$F7,$23,$27,$E1
	.dc.b	$63,$EC,$1F,$5C,$8F,$D7,$7A,$E3,$F8,$48,$FF,$6E,$E5,$C1,$D1,$0F
	.dc.b	$C7,$4F,$E0,$76,$91,$F8,$16,$BF,$B8,$AE,$93,$FB,$88,$FF,$33,$69
	.dc.b	$F9,$8C,$7F,$03,$3D,$E7,$E5,$3F,$82,$D7,$F2,$1C,$1F,$8F,$24,$FE
	.dc.b	$3D,$FC,$97,$AB,$FE,$36,$CB,$FB,$0C,$FC,$26,$BF,$56,$3F,$35,$C9
	.dc.b	$BE,$19,$3F,$23,$85,$3A,$8F,$CE,$3F,$9F,$23,$F9,$FA,$F0,$76,$C8
	.dc.b	$E5,$1C,$5F,$DA,$4F,$06,$9F,$A7,$E8,$FD,$51,$1C,$1D,$4F,$55,$75
	.dc.b	$7D,$47,$9E,$7B,$6B,$FA,$A7,$E1,$72,$7E,$DF,$AF,$F6,$C4,$AF,$DB
	.dc.b	$3F,$21,$EA,$43,$A9,$F9,$BE,$BF,$99,$D1,$23,$E3,$47,$F0,$33,$F8
	.dc.b	$11,$1F,$E0,$4F,$57,$FC,$45,$B7,$D2,$C7,$FD,$52,$DE,$3F,$D2,$EB
	.dc.b	$CE,$B8,$90,$FE,$39,$7F,$C3,$FB,$64,$7F,$0F,$6B,$FC,$86,$7E,$36
	.dc.b	$1D,$B5,$FE,$31,$1D,$67,$D8,$7A,$8F,$C3,$74,$FF,$1D,$EF,$FE,$38
	.dc.b	$48,$FE,$38,$FA,$EF,$A4,$7C,$77,$F8,$CD,$7F,$25,$F0,$7F,$25,$DB
	.dc.b	$2B,$F1,$91,$97,$F9,$33,$FC,$A4,$8F,$DF,$3E,$3E,$3E,$AD,$3E,$2C
	.dc.b	$72,$38,$90,$FE,$81,$3F,$CC,$FC,$49,$1F,$CC,$DA,$FC,$53,$D1,$3F
	.dc.b	$ED,$6D,$BE,$94,$7F,$4E,$C7,$CC,$FF,$1B,$BB,$97,$D8,$EF,$FE,$85
	.dc.b	$E3,$7F,$41,$25,$7F,$41,$7E,$36,$3D,$73,$F5,$4F,$53,$F5,$DE,$4F
	.dc.b	$ED,$FC,$19,$18,$21,$FD,$6E,$3F,$A7,$11,$F8,$D8,$FD,$27,$E3,$9F
	.dc.b	$B0,$38,$9F,$D8,$63,$83,$CC,$FE,$AD,$E5,$FF,$59,$23,$83,$AE,$1E
	.dc.b	$97,$E3,$9E,$93,$F9,$F4,$FF,$63,$23,$FC,$3F,$C3,$FE,$CD,$E7,$7F
	.dc.b	$65,$24,$7F,$65,$7E,$2A,$1B,$FD,$26,$E7,$F9,$87,$D8,$7E,$AB,$C5
	.dc.b	$FE,$3C,$65,$FD,$B9,$FD,$FF,$83,$F6,$0D,$3F,$18,$53,$F0,$44,$30
	.dc.b	$43,$F1,$E6,$4F,$30,$FC,$BF,$FC,$22,$FA,$5A,$E4,$CB,$8E,$D4,$F9
	.dc.b	$F6,$FD,$3D,$7F,$35,$F4,$31,$B7,$82,$73,$3F,$C2,$3C,$8F,$F7,$8E
	.dc.b	$2F,$67,$A9,$27,$FC,$47,$19,$7E,$C9,$F5,$EB,$99,$FC,$43,$0E,$7E
	.dc.b	$E7,$F4,$7F,$32,$57,$F6,$8B,$4E,$A3,$AB,$EC,$23,$D8,$FF,$19,$67
	.dc.b	$EA,$7E,$E6,$EC,$39,$47,$F8,$EE,$FA,$BD,$08,$67,$DB,$F9,$24,$7D
	.dc.b	$AF,$E2,$1E,$E7,$E1,$32,$F6,$3E,$9C,$7B,$C3,$5F,$B3,$7E,$0A,$FE
	.dc.b	$67,$E6,$FC,$CF,$80,$FA,$E4,$8F,$96,$F7,$E0,$FB,$31,$DE,$FF,$84
	.dc.b	$67,$F5,$3E,$0F,$FB,$87,$A1,$2B,$E5,$46,$5F,$AA,$FE,$10,$8F,$C9
	.dc.b	$BE,$AF,$F5,$0C,$FD,$6F,$FB,$A7,$35,$70,$8F,$94,$D7,$F2,$A7,$E5
	.dc.b	$7E,$17,$F2,$24,$7F,$0A,$FA,$F8,$3E,$B4,$7E,$C3,$2F,$33,$DE,$AE
	.dc.b	$1F,$D5,$BD,$3F,$C8,$FC,$37,$E7,$09,$5F,$90,$7A,$3F,$14,$FC,$0D
	.dc.b	$FE,$63,$E1,$BF,$C2,$9E,$AC,$9F,$C7,$46,$5F,$CC,$BF,$A8,$23,$F1
	.dc.b	$72,$F3,$F2,$7E,$2E,$38,$9C,$08,$7E,$36,$7F,$63,$F0,$B2,$3F,$4A
	.dc.b	$D7,$F8,$0E,$7F,$A4,$E1,$FD,$81,$F7,$23,$A9,$F4,$6B,$CC,$FF,$7A
	.dc.b	$F5,$FF,$67,$F0,$DF,$BE,$24,$7E,$C9,$F8,$6F,$DB,$AE,$8F,$DB,$F8
	.dc.b	$6F,$F1,$B7,$B5,$23,$F6,$11,$FB,$E3,$F8,$49,$1F,$DE,$BD,$31,$F4
	.dc.b	$47,$CE,$38,$9A,$21,$FA,$F8,$FE,$27,$F0,$B2,$3F,$83,$B5,$F8,$77
	.dc.b	$F0,$9F,$E1,$0F,$A2,$BF,$55,$1F,$C7,$EB,$CC,$F8,$0F,$F0,$EF,$B9
	.dc.b	$FC,$5B,$E1,$BF,$90,$92,$3F,$8A,$B5,$FE,$04,$D7,$F3,$3C,$C8,$C4
	.dc.b	$7E,$E1,$AF,$F2,$F7,$F9,$6F,$C3,$48,$FE,$42,$D7,$F8,$73,$5E,$AD
	.dc.b	$7F,$81,$B5,$FE,$74,$D7,$F9,$F7,$C2,$7F,$3A,$23,$FC,$B9,$AF,$F1
	.dc.b	$B6,$BE,$A6,$BF,$C4,$1A,$FF,$47,$7F,$A2,$FC,$2C,$8F,$E7,$2D,$7F
	.dc.b	$93,$35,$F8,$AD,$7F,$8D,$B5,$FE,$A4,$D7,$FA,$B7,$C2,$FF,$52,$23
	.dc.b	$FD,$19,$AF,$F3,$36,$BF,$49,$AF,$F2,$86,$BF,$D7,$DF,$EB,$BF,$0B
	.dc.b	$23,$FA,$8B,$5F,$E7,$CD,$7E,$EB,$5F,$E6,$6D,$7F,$B5,$35,$FE,$DD
	.dc.b	$F0,$BF,$DA,$88,$FF,$5E,$6B,$FD,$2D,$AF,$E2,$9A,$FF,$40,$6B,$FD
	.dc.b	$DD,$FE,$EB,$F0,$B2,$3F,$B4,$B5,$FE,$AC,$D7,$F2,$AD,$7F,$A5,$B5
	.dc.b	$FF,$01,$35,$FF,$03,$7C,$2F,$F8,$08,$8F,$F7,$66,$BF,$D8,$DA,$FE
	.dc.b	$89,$AF,$F5,$86,$BF,$E0,$F7,$FC,$23,$F9,$7F,$ED,$23,$FA,$08,$70
	.dc.b	$FF,$44,$6B,$FE,$1A,$6B,$FE,$31,$6B,$C7,$87,$A9,$1E,$A6,$4B,$1C
	.dc.b	$09,$1E,$4E,$B9,$3A,$6B,$E6,$9F,$E7,$4D,$8C,$BE,$76,$BE,$76,$BE
	.dc.b	$77,$E0,$F0,$48,$F1,$B5,$EC,$D7,$E1,$35,$F8,$4D,$7E,$13,$5F,$95
	.dc.b	$C9,$F3,$88,$FC,$06,$BE,$B6,$BF,$3D,$AF,$CF,$6B,$F3,$DF,$B3,$CD
	.dc.b	$23,$E5,$B5,$F8,$CD,$7E,$D3,$5F,$B4,$D7,$ED,$35,$FC,$17,$5F,$E1
	.dc.b	$88,$FD,$86,$BF,$4D,$AF,$E1,$DA,$FE,$1D,$AF,$E1,$DF,$C8,$F6,$C8
	.dc.b	$FC,$1B,$5F,$BC,$D7,$F2,$4D,$7F,$24,$D7,$F2,$4D,$7F,$35,$E0,$FE
	.dc.b	$78,$8F,$E4,$1A,$FE,$2D,$AF,$E7,$DA,$FE,$7D,$AF,$E7,$DF,$D4,$F8
	.dc.b	$72,$3F,$36,$D7,$F2,$CD,$7F,$54,$D7,$F5,$4D,$7F,$54,$D7,$F6,$5E
	.dc.b	$37,$ED,$88,$FE,$A1,$AF,$E8,$DA,$FE,$DD,$AF,$ED,$DA,$FE,$DD,$FE
	.dc.b	$03,$E4,$48,$FD,$9B,$5F,$E0,$7F,$98,$38,$BF,$D3,$3D,$7F,$C1,$E3
	.dc.b	$13,$FE,$B1,$8C,$4E,$B8,$7F,$D3,$F4,$E3,$C5,$FF,$AD,$E1,$C7,$88
	.dc.b	$46,$1E,$95,$CA,$75,$E0,$7F,$17,$C7,$F8,$C7,$23,$CB,$FE,$36,$4C
	.dc.b	$3F,$86,$BF,$EA,$EC,$7C,$F0,$AF,$3B,$D8,$F2,$67,$E3,$9F,$44,$64
	.dc.b	$F2,$E3,$92,$39,$C7,$C9,$97,$F2,$5F,$2F,$F9,$69,$27,$F8,$FB,$FE
	.dc.b	$C0,$E2,$7C,$37,$7E,$99,$5D,$E7,$2F,$87,$67,$9D,$E5,$EA,$93,$98
	.dc.b	$F9,$53,$A8,$FE,$5B,$1F,$CF,$88,$FF,$32,$AE,$AF,$C3,$CB,$EB,$E1
	.dc.b	$E2,$8C,$1F,$EE,$D9,$DE,$FF,$42,$F9,$CF,$F3,$36,$BF,$D1,$FD,$1F
	.dc.b	$E9,$A4,$7F,$9F,$BF,$07,$3E,$D5,$F8,$BC,$3F,$05,$87,$6C,$7F,$A7
	.dc.b	$63,$EC,$26,$3F,$0F,$DD,$F7,$FA,$9F,$D6,$89,$1F,$D3,$1E,$CF,$D9
	.dc.b	$E1,$63,$E0,$C3,$FD,$E7,$8F,$7A,$FE,$B5,$8F,$67,$E0,$9F,$D8,$DF
	.dc.b	$9B,$8F,$CE,$F5,$FF,$B4,$92,$7F,$AD,$BE,$07,$8E,$3E,$07,$C5,$8F
	.dc.b	$BB,$4E,$DA,$7C,$B9,$FB,$47,$B4,$FC,$07,$B3,$FD,$D8,$8F,$F6,$87
	.dc.b	$B4,$7C,$98,$EA,$7E,$26,$7C,$EF,$43,$D7,$BF,$E1,$5B,$97,$E2,$3D
	.dc.b	$CF,$F0,$19,$23,$FB,$AB,$DB,$6F,$C2,$C7,$63,$E2,$ED,$F2,$5F,$F0
	.dc.b	$45,$7B,$27,$EC,$C7,$AC,$EE,$7F,$80,$23,$FC,$22,$47,$F2,$98,$7D
	.dc.b	$57,$F8,$36,$7F,$1D,$E4,$7F,$84,$E7,$83,$FE,$03,$7F,$C3,$7F,$01
	.dc.b	$23,$FC,$1A,$D7,$FC,$3B,$87,$9F,$5F,$A9,$1F,$5A,$FB,$3E,$62,$69
	.dc.b	$FC,$3B,$0C,$BF,$E2,$7B,$7C,$2F,$EE,$C9,$5F,$AD,$7E,$43,$EA,$3E
	.dc.b	$44,$F1,$CB,$EA,$C7,$6C,$FF,$C5,$98,$76,$7F,$2B,$F0,$7F,$C1,$C9
	.dc.b	$3F,$B9,$7A,$BF,$CA,$32,$F9,$0B,$2F,$A8,$7D,$5D,$98,$78,$32,$F1
	.dc.b	$57,$F0,$FF,$86,$FE,$28,$47,$F8,$33,$DB,$5F,$8F,$3F,$8F,$9F,$C0
	.dc.b	$E9,$E7,$7F,$C8,$57,$D7,$4F,$D6,$CF,$EF,$3E,$1F,$F9,$01,$2B,$F8
	.dc.b	$93,$D7,$4F,$4E,$9E,$8A,$E5,$5D,$0F,$CE,$10,$FB,$6F,$EB,$D1,$E0
	.dc.b	$3F,$16,$43,$65,$36,$7F,$1D,$79,$3C,$4D,$9F,$88,$F8,$6F,$E7,$04
	.dc.b	$7F,$95,$3F,$4A,$7D,$11,$FC,$3F,$C8,$F8,$8F,$D3,$AB,$2F,$77,$F8
	.dc.b	$FF,$5F,$F4,$32,$57,$F3,$57,$A5,$72,$D9,$87,$28,$E7,$7F,$CA,$3E
	.dc.b	$1B,$FA,$61,$30,$FE,$7C,$F2,$7E,$46,$FF,$DE,$6F,$94,$63,$6F,$E4
	.dc.b	$77,$FC,$CB,$E1,$7F,$AB,$12,$7F,$A4,$3C,$DF,$C0,$9E,$F3,$91,$F2
	.dc.b	$CA,$7E,$4E,$3F,$9E,$FC,$37,$F6,$12,$47,$F5,$37,$A6,$1F,$94,$8F
	.dc.b	$E5,$38,$79,$9E,$E7,$F1,$0A,$FE,$8F,$F0,$DF,$DA,$C9,$1F,$D7,$5E
	.dc.b	$78,$FD,$BB,$F8,$13,$E4,$DB,$EB,$BC,$CF,$EE,$1F,$0D,$FD,$D4,$8F
	.dc.b	$F6,$77,$9B,$ED,$3C,$A7,$F7,$7C,$3C,$D5,$FC,$15,$FE,$B7,$F0,$DF
	.dc.b	$DF,$C9,$3F,$DC,$5E,$6F,$E1,$4F,$94,$78,$CF,$7C,$FC,$33,$A3,$FE
	.dc.b	$05,$F8,$6F,$F0,$51,$1F,$EF,$8F,$4C,$3F,$8F,$67,$FD,$B3,$0F,$2E
	.dc.b	$5F,$7A,$FF,$C1,$DF,$0B,$FE,$13,$23,$FE,$05,$79,$70,$7F,$88,$EF
	.dc.b	$94,$FC,$98,$AF,$F4,$2F,$86,$FF,$0E,$11,$FF,$07,$3C,$9F,$6E,$7E
	.dc.b	$A7,$DB,$87,$D5,$3C,$8F,$C1,$3F,$6A,$7E,$24,$FF,$0C,$63,$FE,$29
	.dc.b	$3A,$3F,$48,$E0,$72,$38,$E5,$FA,$F7,$89,$FE,$21,$F8,$5F,$F1,$91
	.dc.b	$1C,$1F,$C1,$EF,$E8,$5F,$96,$3F,$CB,$19,$7C,$09,$F5,$12,$7F,$CB
	.dc.b	$FE,$9F,$F8,$6B,$E1,$BF,$C7,$C4,$BF,$F1,$83,$FE,$E2,$5F,$86,$3D
	.dc.b	$BF,$F2,$87,$B1,$27,$FC,$A1,$39,$7F,$C8,$9F,$0F,$FE,$55,$F6,$64
	.dc.b	$60,$87,$F5,$E8,$FF,$27,$11,$FF,$78,$CF,$37,$FC,$2B,$5D,$DF,$F0
	.dc.b	$FC,$67,$3F,$EB,$D5,$C6,$7A,$1F,$E4,$97,$FC,$B5,$EF,$C8,$FF,$26
	.dc.b	$35,$FB,$27,$BB,$1E,$D5,$FE,$89,$C3,$FD,$01,$87,$7D,$3F,$C6,$78
	.dc.b	$FF,$90,$89,$87,$C2,$C3,$6F,$C9,$AF,$94,$FC,$A3,$06,$63,$BD,$7F
	.dc.b	$A4,$AF,$BC,$74,$9F,$F0,$CD,$DC,$17,$FD,$13,$0F,$25,$7F,$4C,$5F
	.dc.b	$0F,$E1,$24,$7C,$C7,$C0,$FF,$93,$F1,$F0,$BC,$71,$ED,$7E,$E8,$FA
	.dc.b	$F8,$78,$1D,$1F,$80,$F8,$6F,$71,$1F,$B2,$F6,$7E,$C9,$DC,$F8,$C7
	.dc.b	$49,$EC,$FB,$8C,$9F,$7B,$E1,$FE,$51,$1F,$C2,$BD,$6B,$EB,$47,$53
	.dc.b	$F0,$B7,$EA,$7F,$D0,$99,$D7,$EF,$1F,$98,$F8,$7F,$AC,$47,$F2,$2F
	.dc.b	$5C,$3E,$A4,$75,$BF,$84,$F2,$3B,$9E,$B3,$B9,$EA,$77,$A3,$E1,$7F
	.dc.b	$08,$49,$FC,$FB,$DD,$FF,$68,$69,$F5,$1E,$D5,$CB,$2F,$57,$23,$3F
	.dc.b	$C3,$D7,$EC,$BE,$1B,$F2,$04,$7F,$56,$F5,$7F,$DB,$38,$75,$CB,$F1
	.dc.b	$31,$DB,$3F,$A7,$97,$EB,$23,$D5,$1F,$8F,$F8,$6F,$CE,$12,$3F,$6E
	.dc.b	$F5,$F0,$98,$75,$D3,$CB,$5F,$74,$98,$7A,$DF,$E1,$3F,$0B,$FA,$72
	.dc.b	$3F,$C0,$9E,$95,$CE,$3C,$67,$CE,$39,$E3,$FA,$3B,$FD,$BF,$C3,$7E
	.dc.b	$C4,$93,$FC,$31,$E6,$FD,$B9,$E0,$F4,$3F,$CA,$9A,$FE,$1F,$4F,$8A
	.dc.b	$7F,$13,$8F,$E7,$ED,$78,$9F,$C2,$9C,$3F,$A6,$31,$7F,$A3,$78,$B5
	.dc.b	$E2,$60,$FE,$42,$FF,$09,$F8,$59,$1F,$C9,$1A,$F4,$3D,$13,$F6,$B4
	.dc.b	$4F,$D0,$D3,$E7,$57,$E5,$8F,$F5,$FF,$C3,$7F,$1C,$3F,$8B,$11,$FE
	.dc.b	$62,$FE,$1E,$FE,$0E,$1F,$E5,$2A,$FD,$0E,$1D,$AF,$C3,$3D,$F2,$FF
	.dc.b	$2B,$CB,$0F,$E5,$3A,$7F,$23,$25,$FF,$3E,$7B,$5F,$18,$FF,$32,$D7
	.dc.b	$57,$D1,$87,$A8,$98,$76,$C3,$F9,$AF,$3F,$F2,$F2,$57,$F4,$A7,$AB
	.dc.b	$F7,$6F,$A5,$F2,$8F,$AC,$D5,$1E,$97,$FA,$EF,$57,$F3,$B2,$3F,$D5
	.dc.b	$9E,$AF,$9B,$0F,$83,$A7,$AE,$3F,$58,$76,$8F,$43,$FD,$A3,$AB,$FA
	.dc.b	$31,$1F,$EC,$2F,$4C,$3C,$98,$7A,$2B,$F1,$F1,$E7,$D7,$F0,$19,$FF
	.dc.b	$56,$F2,$7F,$A8,$13,$1F,$ED,$6F,$3B,$E2,$79,$A3,$EC,$9F,$16,$7F
	.dc.b	$AF,$FA,$FF,$D6,$49,$1F,$DD,$1E,$4F,$AC,$FA,$F9,$FB,$F0,$FC,$3C
	.dc.b	$7F,$A7,$B1,$FC,$3D,$FF,$6A,$F6,$BF,$B2,$12,$7F,$BF,$BD,$1F,$8E
	.dc.b	$7D,$2B,$E9,$BF,$CB,$D3,$EA,$4F,$F8,$0B,$DC,$FE,$DC,$47,$FC,$12
	.dc.b	$F3,$CB,$F4,$97,$CA,$7D,$C7,$F8,$63,$DB,$FE,$EC,$47,$FC,$22,$F1
	.dc.b	$C3,$F4,$1A,$FC,$79,$FB,$C7,$C1,$3C,$F8,$F9,$34,$FE,$8B,$F9,$87
	.dc.b	$F4,$69,$FF,$22,$B5,$FF,$5D,$B5,$FF,$61,$B5,$F8,$39,$7C,$A5,$3F
	.dc.b	$B0,$CB,$D5,$1E,$23,$F2,$FA,$F9,$9E,$7F,$0D,$FE,$39,$C0,$D2,$47
	.dc.b	$F0,$E3,$F0,$27,$6C,$BD,$D8,$7D,$88,$CB,$5D,$1E,$8F,$86,$EC,$47
	.dc.b	$FC,$56,$2D,$19,$8F,$F1,$7C,$61,$DB,$5C,$C6,$DF,$92,$D7,$E4,$BE
	.dc.b	$DF,$86,$91,$C9,$AF,$66,$BF,$49,$AF,$D2,$6B,$F4,$9A,$FA,$7E,$1B
	.dc.b	$E5,$11,$F3,$B5,$F5,$B5,$FB,$ED,$7E,$FB,$5F,$BE,$FD,$5E,$D9,$1F
	.dc.b	$09,$AF,$C6,$6B,$F8,$A6,$BF,$8A,$6B,$F8,$A6,$BF,$47,$E1,$BF,$04
	.dc.b	$47,$E7,$B5,$FA,$6D,$7F,$2E,$D7,$F2,$ED,$7F,$2E,$FE,$37,$E1,$A4
	.dc.b	$7D,$A6,$BF,$79,$AF,$E8,$9A,$FE,$89,$AF,$E8,$9A,$FE,$27,$E1,$BF
	.dc.b	$34,$47,$F0,$ED,$7F,$16,$D7,$F5,$ED,$7F,$5E,$D7,$F5,$EF,$E9,$7E
	.dc.b	$1A,$47,$E4,$9A,$FE,$59,$AF,$EE,$9A,$FE,$E9,$AF,$EE,$9A,$FE,$87
	.dc.b	$E1,$BF,$64,$47,$F3,$ED,$7F,$46,$D7,$F8,$3B,$5F,$E0,$ED,$7F,$83
	.dc.b	$BF,$BD,$F8,$69,$1F,$AA,$6B,$FA,$EF,$85,$FE,$08,$FF,$5C,$F0,$6F
	.dc.b	$31,$CF,$02,$A9,$AE,$79,$7F,$CE,$F8,$FF,$8C,$CF,$0F,$86,$FE,$1C
	.dc.b	$49,$FD,$DB,$FC,$46,$3B,$E1,$27,$91,$FC,$82,$3B,$B8,$3F,$B8,$DF
	.dc.b	$81,$FE,$1E,$7F,$0B,$8F,$3F,$C2,$FF,$1D,$24,$7F,$0A,$7C,$0F,$3C
	.dc.b	$BD,$CD,$BE,$EF,$4C,$BE,$03,$BC,$3C,$1C,$BF,$CE,$B1,$DC,$FF,$3A
	.dc.b	$7A,$BF,$34,$FE,$32,$E1,$D1,$82,$1B,$21,$A3,$F9,$1F,$C3,$7F,$37
	.dc.b	$23,$FC,$91,$EF,$3E,$38,$F8,$D7,$E8,$7E,$2E,$2F,$1B,$DB,$5B,$9F
	.dc.b	$76,$1F,$80,$F1,$BE,$FF,$C3,$7F,$45,$24,$FF,$32,$7B,$EF,$F7,$A3
	.dc.b	$BD,$7E,$23,$0E,$B3,$99,$EA,$7C,$FB,$FC,$5F,$C3,$7F,$51,$24,$7F
	.dc.b	$3E,$7F,$AA,$7C,$07,$E4,$BC,$B9,$3E,$C8,$FC,$99,$F9,$6F,$2E,$46
	.dc.b	$4F,$83,$F0,$DF,$D7,$48,$FF,$4A,$7A,$BF,$81,$AE,$AF,$DF,$AE,$AF
	.dc.b	$E1,$6B,$AB,$93,$B1,$FD,$5A,$39,$B4,$FE,$AE,$E1,$FE,$FA,$D7,$FB
	.dc.b	$EB,$5F,$EF,$AD,$7F,$B2,$7C,$37,$F7,$52,$3F,$D8,$9E,$FB,$7F,$82
	.dc.b	$A7,$CF,$3F,$7B,$CF,$B3,$98,$F3,$BF,$E7,$4F,$3F,$F7,$FF,$98,$3E
	.dc.b	$1F,$FB,$9D,$23,$0F,$F7,$66,$A2,$15,$CB,$81,$F9,$7B,$D5,$FE,$17
	.dc.b	$10,$FD,$D4,$65,$FD,$97,$AB,$FC,$43,$D5,$91,$83,$F8,$05,$BE,$23
	.dc.b	$CE,$D8,$F3,$46,$0F,$5D,$F5,$8F,$D0,$9E,$13,$F8,$3B,$9C,$FF,$D2
	.dc.b	$F8,$78,$27,$F4,$F8,$7F,$22,$8F,$8F,$F0,$9E,$02,$4F,$C5,$6B,$F0
	.dc.b	$4F,$94,$E1,$F8,$98,$A3,$11,$F2,$A3,$0F,$32,$1C,$D7,$C2,$7C,$02
	.dc.b	$47,$D5,$7C,$38,$39,$BA,$5F,$A2,$BE,$7F,$C2,$7F,$91,$97,$E3,$CE
	.dc.b	$4D,$C7,$CE,$AE,$0F,$F2,$CC,$97,$F8,$8F,$85,$FA,$24,$9F,$C3,$3F
	.dc.b	$42,$3B,$61,$D6,$7F,$0D,$7F,$88,$8C,$47,$C6,$3F,$18,$E3,$6E,$F9
	.dc.b	$E3,$2F,$C7,$C3,$DC,$FE,$AE,$DB,$FC,$5C,$63,$DC,$FC,$EF,$C2,$FE
	.dc.b	$18,$93,$F9,$B7,$CD,$87,$58,$F2,$94,$F2,$1F,$1B,$0F,$0E,$7E,$C7
	.dc.b	$C7,$87,$88,$93,$F9,$CC,$3F,$5B,$F0,$BF,$96,$24,$7E,$AD,$F0,$BD
	.dc.b	$4F,$09,$4F,$0E,$7F,$0F,$0F,$04,$7E,$C7,$37,$C0,$7C,$2F,$80,$93
	.dc.b	$FA,$C8,$FD,$EF,$C2,$FE,$98,$91,$FB,$B7,$C3,$87,$D2,$E1,$FC,$8C
	.dc.b	$F8,$5B,$1E,$1D,$BE,$26,$3E,$0A,$FD,$CE,$3E,$07,$F5,$78,$FF,$0C
	.dc.b	$F8,$5F,$DA,$92,$3F,$85,$3E,$0D,$FF,$1F,$1D,$F8,$7F,$09,$1F,$3B
	.dc.b	$1E,$FF,$0B,$F7,$BE,$12,$5F,$F0,$88,$FE,$32,$7F,$02,$23,$FC,$5D
	.dc.b	$EC,$FC,$CD,$3D,$D8,$76,$8F,$95,$BF,$CC,$7F,$8B,$D7,$67,$F8,$2D
	.dc.b	$FE,$3A,$3F,$93,$BF,$C4,$09,$1F,$C9,$5E,$EF,$D8,$C7,$DB,$87,$C9
	.dc.b	$8F,$B9,$AB,$7F,$CC,$92,$3F,$91,$E9,$FC,$D3,$5F,$E3,$A4,$8F,$E6
	.dc.b	$4F,$6D,$FE,$AE,$3D,$8F,$91,$3D,$AF,$A3,$FC,$79,$FD,$3D,$FF,$40
	.dc.b	$DF,$F9,$49,$27,$F9,$F3,$D7,$6F,$8D,$1E,$B3,$E3,$EF,$F8,$2E,$3F
	.dc.b	$D8,$F0,$C9,$FE,$77,$1F,$D3,$38,$FF,$9A,$92,$3F,$A5,$3D,$5F,$D7
	.dc.b	$69,$EA,$7A,$9F,$1A,$7A,$3F,$CD,$AB,$F7,$F1,$FE,$56,$D1,$D1,$FD
	.dc.b	$61,$B5,$FE,$CC,$6B,$FC,$CD,$AF,$F5,$5D,$FF,$86,$BF,$91,$3C,$79
	.dc.b	$FC,$08,$CF,$07,$F1,$6A,$E3,$5F,$D8,$1A,$FF,$63,$6B,$FC,$AE,$FC
	.dc.b	$9D,$B2,$3C,$8D,$7F,$A3,$5F,$6A,$FB,$51,$F9,$77,$ED,$DF,$86,$FE
	.dc.b	$01,$27,$F9,$DE,$BF,$C5,$E9,$DB,$F0,$C9,$3F,$0D,$FA,$0F,$BC,$FC
	.dc.b	$9F,$2F,$E8,$71,$F1,$47,$AF,$0F,$C0,$DF,$CA,$C7,$F3,$47,$C3,$F0
	.dc.b	$3E,$81,$2B,$E8,$3D,$A3,$F5,$F1,$FC,$23,$0E,$C7,$EB,$EB,$A1,$F2
	.dc.b	$EF,$E8,$46,$63,$EE,$12,$7E,$E3,$D4,$F8,$EF,$F8,$3A,$3A,$E5,$F0
	.dc.b	$EB,$D2,$74,$7E,$37,$89,$F8,$72,$57,$E1,$DE,$8F,$D1,$3F,$27,$87
	.dc.b	$C0,$C3,$CD,$1D,$4F,$65,$7D,$1F,$13,$F2,$44,$8F,$C9,$3D,$1F,$D5
	.dc.b	$1E,$78,$E2,$FF,$3F,$9F,$86,$FD,$13,$35,$F9,$1F,$2F,$F3,$C4,$8F
	.dc.b	$CF,$3D,$78,$BF,$55,$1D,$4F,$BB,$7E,$77,$B6,$5F,$0D,$F0,$79,$7F
	.dc.b	$A9,$23,$FA,$97,$A7,$0E,$F1,$F8,$53,$E5,$39,$9E,$6C,$BF,$49,$ED
	.dc.b	$7E,$CC,$91,$FB,$37,$9B,$B3,$E6,$C7,$CC,$79,$9E,$83,$E2,$7B,$5F
	.dc.b	$BD,$23,$FB,$D7,$9B,$BA,$F2,$BF,$B1,$3E,$03,$89,$7C,$27,$F0,$72
	.dc.b	$4F,$F0,$77,$9C,$F2,$E2,$FD,$1C,$F3,$3C,$B1,$CF,$E1,$3F,$88,$12
	.dc.b	$3F,$88,$3C,$AF,$F7,$24,$8F,$91,$3F,$BB,$7E,$11,$FE,$31,$F8,$4F
	.dc.b	$E3,$44,$7F,$8D,$3C,$DF,$8E,$43,$C9,$1C,$DF,$85,$3F,$C5,$7E,$13
	.dc.b	$F9,$21,$23,$F9,$23,$CD,$FC,$34,$FC,$5D,$BF,$4F,$1E,$4A,$7C,$27
	.dc.b	$F2,$D2,$47,$F2,$D7,$96,$5F,$18,$F8,$F3,$F9,$A9,$F2,$BF,$05,$FF
	.dc.b	$27,$7C,$27,$F3,$72,$3F,$CD,$DE,$7C,$DF,$BA,$8F,$9F,$3F,$85,$C7
	.dc.b	$ED,$7C,$27,$F3,$F2,$4F,$F3,$F7,$8F,$1F,$A7,$3F,$5C,$FC,$43,$F9
	.dc.b	$B7,$C2,$7F,$47,$23,$FD,$1D,$E2,$EC,$F4,$9F,$63,$0F,$50,$F8,$4F
	.dc.b	$EA,$04,$8F,$EA,$0F,$2C,$7E,$64,$7E,$4A,$7E,$17,$E6,$1F,$E7,$58
	.dc.b	$FF,$43,$07,$0F,$FA,$25,$AF,$FA,$AD,$AF,$CB,$E7,$FD,$F9,$FC,$03
	.dc.b	$6F,$A5,$AF,$F2,$DC,$FE,$0E,$3F,$A4,$C1,$F0,$9F,$EB,$E6,$E3,$E2
	.dc.b	$6B,$E1,$8F,$56,$5F,$02,$F9,$BE,$AF,$84,$E4,$4A,$E4,$F0,$7F,$2D
	.dc.b	$1F,$83,$3F,$09,$97,$EB,$30,$7C,$27,$8C,$95,$E3,$79,$3F,$96,$AF
	.dc.b	$C1,$11,$FD,$14,$BE,$13,$D8,$48,$F6,$3C,$AF,$F2,$B8,$FA,$A7,$8F
	.dc.b	$C2,$7C,$32,$3F,$0D,$E1,$AF,$CB,$8F,$B4,$78,$3E,$13,$E5,$11,$F9
	.dc.b	$4F,$0E,$2D,$C7,$E2,$8F,$51,$FE,$E9,$C9,$F0,$9F,$4C,$91,$F4,$DE
	.dc.b	$4F,$E8,$8F,$C9,$1F,$03,$E1,$3E,$D1,$1F,$B4,$F0,$E1,$FE,$08,$4D
	.dc.b	$BF,$23,$F0,$9F,$81,$25,$7E,$05,$E1,$1F,$36,$3D,$1B,$7D,$7F,$84
	.dc.b	$FC,$41,$23,$F1,$0F,$07,$CA,$70,$D7,$FC,$D9,$9B,$E1,$3F,$20,$49
	.dc.b	$FC,$83,$C5,$F8,$D8,$7E,$1A,$B8,$FC,$27,$E5,$C9,$3F,$97,$78,$4F
	.dc.b	$E3,$78,$7F,$AB,$29,$F9,$DF,$09,$F9,$F2,$3F,$9F,$78,$CF,$E0,$E3
	.dc.b	$EA,$9F,$B3,$F8,$0F,$D3,$11,$FD,$33,$C1,$FB,$BA,$7E,$6F,$E1,$3F
	.dc.b	$58,$48,$FD,$63,$B7,$ED,$1D,$B4,$7C,$17,$EC,$C9,$1F,$B3,$78,$70
	.dc.b	$7D,$98,$FC,$FF,$C1,$7E,$E4,$8F,$EE,$5D,$E9,$F6,$71,$FD,$77,$C1
	.dc.b	$7E,$FC,$95,$FB,$F7,$6F,$D8,$8F,$DA,$7C,$0F,$F0,$52,$47,$F0,$57
	.dc.b	$6F,$DD,$E6,$FF,$73,$66,$F8,$2F,$E1,$84,$9F,$E1,$8F,$0C,$7E,$E5
	.dc.b	$7F,$B2,$BE,$0B,$F9,$0D,$31,$46,$1B,$1F,$C9,$9A,$FF,$9A,$38,$FF
	.dc.b	$D7,$1D,$FF,$C8,$BC,$BF,$C0,$CF,$F1,$0E,$3F,$B9,$7F,$D5,$7E,$07
	.dc.b	$A1,$27,$A3,$E5,$9E,$31,$E3,$F8,$9F,$11,$1F,$13,$B7,$FB,$CD,$3E
	.dc.b	$27,$D0,$48,$F4,$3B,$7F,$A6,$43,$E2,$3D,$C4,$8F,$73,$BE,$4F,$F0
	.dc.b	$97,$C4,$FC,$32,$4F,$C3,$7E,$36,$AD,$FF,$41,$A7,$F9,$C2,$8D,$3F
	.dc.b	$DA,$44,$E0,$FF,$18,$3F,$E0,$93,$11,$FA,$9C,$9B,$7D,$32,$47,$D3
	.dc.b	$7C,$D5,$FD,$F6,$5F,$09,$F6,$49,$5F,$65,$DB,$FE,$19,$8F,$F5,$8B
	.dc.b	$FE,$B3,$C7,$D4,$FD,$DC,$7F,$01,$18,$8D,$46,$08,$EC,$86,$8A,$7F
	.dc.b	$B9,$0F,$F6,$07,$C1,$BF,$01,$FC,$63,$63,$FD,$66,$D7,$71,$A2,$3A
	.dc.b	$6B,$A3,$04,$30,$7F,$13,$FC,$40,$40

;--------------------------------------------------------------------------------
;	FMOVECR.X #ccc,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovecr881_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+FPSP040+FPSP060,-(sp)
	peamsg	'FMOVECR.X #CCC,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVECR.X #ccc,FPn',13,10
;------------------------------------------------
;	d1-d3	actual result
;	d4	actual status
;	d5	0=failed,1=successful
;	d6	fpcr=(XRN..XRP)<<4
;	d7	ccc=0..63
;	a0	expected result handle,expected status,...
;	a1-a3	expected result
;	a4	expected status
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a3
;decompress data
	move.l	a3,-(sp)
	pea.l	fmovecr881_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a3,a0			;expected result handle,expected status,...
@@:
	add.l	d0,(a3)+		;expected result handle
	addq.l	#4,a3			;expected status
	tst.l	(a3)
	bpl	@b
;	addq.l	#4,a3			;-1
;
	moveq.l	#0,d7			;ccc=0..63
77:
	move.l	#0,d6		;fpcr=(XRN..DRP)<<4
66:
;#ccc,FPn
	fmove.l	#0,fpcr
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
					;ccc
	jsr	([fmovecr881_execute,za0,d7.l*4])	;EXECUTE
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
	fmove.x	fp7,-(sp)
	movem.l	(sp)+,d1-d3		;actual result
;
	movem.l	([a0]),a1-a3		;expected result
	movea.l	(4,a0),a4		;expected status
;
	move.l	#-1,-(sp)		;fpcr(rp<<6,-1=strict)
	movem.l	d1-d4/a1-a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d5			;0=failed,1=successful
;
	move.l	d5,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVECR.X #'
	puthex2	d7			;ccc
	putmsg	',FPn @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d5,-(sp)		;0=failed,1=successful
	movem.l	d1-d4/a1-a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a0			;expected result handle,expected status,...
;
	add.w	#1<<4,d6		;rprm++
	cmp.w	#176,d6		;fpcr=(XRN..DRP)<<4
	bls	66b
;
	addq.w	#1,d7			;ccc++
	cmp.w	#63,d7			;ccc=0..63
	bls	77b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
fmovecr881_execute::
	.dc.l	fmovecr881_execute_0
	.dc.l	fmovecr881_execute_1
	.dc.l	fmovecr881_execute_2
	.dc.l	fmovecr881_execute_3
	.dc.l	fmovecr881_execute_4
	.dc.l	fmovecr881_execute_5
	.dc.l	fmovecr881_execute_6
	.dc.l	fmovecr881_execute_7
	.dc.l	fmovecr881_execute_8
	.dc.l	fmovecr881_execute_9
	.dc.l	fmovecr881_execute_10
	.dc.l	fmovecr881_execute_11
	.dc.l	fmovecr881_execute_12
	.dc.l	fmovecr881_execute_13
	.dc.l	fmovecr881_execute_14
	.dc.l	fmovecr881_execute_15
	.dc.l	fmovecr881_execute_16
	.dc.l	fmovecr881_execute_17
	.dc.l	fmovecr881_execute_18
	.dc.l	fmovecr881_execute_19
	.dc.l	fmovecr881_execute_20
	.dc.l	fmovecr881_execute_21
	.dc.l	fmovecr881_execute_22
	.dc.l	fmovecr881_execute_23
	.dc.l	fmovecr881_execute_24
	.dc.l	fmovecr881_execute_25
	.dc.l	fmovecr881_execute_26
	.dc.l	fmovecr881_execute_27
	.dc.l	fmovecr881_execute_28
	.dc.l	fmovecr881_execute_29
	.dc.l	fmovecr881_execute_30
	.dc.l	fmovecr881_execute_31
	.dc.l	fmovecr881_execute_32
	.dc.l	fmovecr881_execute_33
	.dc.l	fmovecr881_execute_34
	.dc.l	fmovecr881_execute_35
	.dc.l	fmovecr881_execute_36
	.dc.l	fmovecr881_execute_37
	.dc.l	fmovecr881_execute_38
	.dc.l	fmovecr881_execute_39
	.dc.l	fmovecr881_execute_40
	.dc.l	fmovecr881_execute_41
	.dc.l	fmovecr881_execute_42
	.dc.l	fmovecr881_execute_43
	.dc.l	fmovecr881_execute_44
	.dc.l	fmovecr881_execute_45
	.dc.l	fmovecr881_execute_46
	.dc.l	fmovecr881_execute_47
	.dc.l	fmovecr881_execute_48
	.dc.l	fmovecr881_execute_49
	.dc.l	fmovecr881_execute_50
	.dc.l	fmovecr881_execute_51
	.dc.l	fmovecr881_execute_52
	.dc.l	fmovecr881_execute_53
	.dc.l	fmovecr881_execute_54
	.dc.l	fmovecr881_execute_55
	.dc.l	fmovecr881_execute_56
	.dc.l	fmovecr881_execute_57
	.dc.l	fmovecr881_execute_58
	.dc.l	fmovecr881_execute_59
	.dc.l	fmovecr881_execute_60
	.dc.l	fmovecr881_execute_61
	.dc.l	fmovecr881_execute_62
	.dc.l	fmovecr881_execute_63
fmovecr881_execute_0::
	fmovecr.x	#0,fp7
	rts
fmovecr881_execute_1::
	fmovecr.x	#1,fp7
	rts
fmovecr881_execute_2::
	fmovecr.x	#2,fp7
	rts
fmovecr881_execute_3::
	fmovecr.x	#3,fp7
	rts
fmovecr881_execute_4::
	fmovecr.x	#4,fp7
	rts
fmovecr881_execute_5::
	fmovecr.x	#5,fp7
	rts
fmovecr881_execute_6::
	fmovecr.x	#6,fp7
	rts
fmovecr881_execute_7::
	fmovecr.x	#7,fp7
	rts
fmovecr881_execute_8::
	fmovecr.x	#8,fp7
	rts
fmovecr881_execute_9::
	fmovecr.x	#9,fp7
	rts
fmovecr881_execute_10::
	fmovecr.x	#10,fp7
	rts
fmovecr881_execute_11::
	fmovecr.x	#11,fp7
	rts
fmovecr881_execute_12::
	fmovecr.x	#12,fp7
	rts
fmovecr881_execute_13::
	fmovecr.x	#13,fp7
	rts
fmovecr881_execute_14::
	fmovecr.x	#14,fp7
	rts
fmovecr881_execute_15::
	fmovecr.x	#15,fp7
	rts
fmovecr881_execute_16::
	fmovecr.x	#16,fp7
	rts
fmovecr881_execute_17::
	fmovecr.x	#17,fp7
	rts
fmovecr881_execute_18::
	fmovecr.x	#18,fp7
	rts
fmovecr881_execute_19::
	fmovecr.x	#19,fp7
	rts
fmovecr881_execute_20::
	fmovecr.x	#20,fp7
	rts
fmovecr881_execute_21::
	fmovecr.x	#21,fp7
	rts
fmovecr881_execute_22::
	fmovecr.x	#22,fp7
	rts
fmovecr881_execute_23::
	fmovecr.x	#23,fp7
	rts
fmovecr881_execute_24::
	fmovecr.x	#24,fp7
	rts
fmovecr881_execute_25::
	fmovecr.x	#25,fp7
	rts
fmovecr881_execute_26::
	fmovecr.x	#26,fp7
	rts
fmovecr881_execute_27::
	fmovecr.x	#27,fp7
	rts
fmovecr881_execute_28::
	fmovecr.x	#28,fp7
	rts
fmovecr881_execute_29::
	fmovecr.x	#29,fp7
	rts
fmovecr881_execute_30::
	fmovecr.x	#30,fp7
	rts
fmovecr881_execute_31::
	fmovecr.x	#31,fp7
	rts
fmovecr881_execute_32::
	fmovecr.x	#32,fp7
	rts
fmovecr881_execute_33::
	fmovecr.x	#33,fp7
	rts
fmovecr881_execute_34::
	fmovecr.x	#34,fp7
	rts
fmovecr881_execute_35::
	fmovecr.x	#35,fp7
	rts
fmovecr881_execute_36::
	fmovecr.x	#36,fp7
	rts
fmovecr881_execute_37::
	fmovecr.x	#37,fp7
	rts
fmovecr881_execute_38::
	fmovecr.x	#38,fp7
	rts
fmovecr881_execute_39::
	fmovecr.x	#39,fp7
	rts
fmovecr881_execute_40::
	fmovecr.x	#40,fp7
	rts
fmovecr881_execute_41::
	fmovecr.x	#41,fp7
	rts
fmovecr881_execute_42::
	fmovecr.x	#42,fp7
	rts
fmovecr881_execute_43::
	fmovecr.x	#43,fp7
	rts
fmovecr881_execute_44::
	fmovecr.x	#44,fp7
	rts
fmovecr881_execute_45::
	fmovecr.x	#45,fp7
	rts
fmovecr881_execute_46::
	fmovecr.x	#46,fp7
	rts
fmovecr881_execute_47::
	fmovecr.x	#47,fp7
	rts
fmovecr881_execute_48::
	fmovecr.x	#48,fp7
	rts
fmovecr881_execute_49::
	fmovecr.x	#49,fp7
	rts
fmovecr881_execute_50::
	fmovecr.x	#50,fp7
	rts
fmovecr881_execute_51::
	fmovecr.x	#51,fp7
	rts
fmovecr881_execute_52::
	fmovecr.x	#52,fp7
	rts
fmovecr881_execute_53::
	fmovecr.x	#53,fp7
	rts
fmovecr881_execute_54::
	fmovecr.x	#54,fp7
	rts
fmovecr881_execute_55::
	fmovecr.x	#55,fp7
	rts
fmovecr881_execute_56::
	fmovecr.x	#56,fp7
	rts
fmovecr881_execute_57::
	fmovecr.x	#57,fp7
	rts
fmovecr881_execute_58::
	fmovecr.x	#58,fp7
	rts
fmovecr881_execute_59::
	fmovecr.x	#59,fp7
	rts
fmovecr881_execute_60::
	fmovecr.x	#60,fp7
	rts
fmovecr881_execute_61::
	fmovecr.x	#61,fp7
	rts
fmovecr881_execute_62::
	fmovecr.x	#62,fp7
	rts
fmovecr881_execute_63::
	fmovecr.x	#63,fp7
	rts
	.cpu	68000

	.align	4
fmovecr881_expected_compressed::
	.dc.b	$00,$00,$18,$04,$13,$EC,$DF,$AF,$62,$14,$D3,$14,$FD,$AB,$53,$9A
	.dc.b	$E5,$18,$AD,$DD,$9E,$DF,$BB,$55,$F6,$FD,$DE,$F5,$E1,$8C,$61,$D6
	.dc.b	$58,$AF,$95,$38,$E9,$2B,$DB,$7C,$65,$DF,$E1,$F3,$1E,$FB,$83,$C5
	.dc.b	$87,$C4,$9C,$C6,$21,$18,$76,$D7,$D4,$D7,$E5,$35,$E9,$3D,$DE,$2D
	.dc.b	$78,$B5,$E2,$D7,$8C,$F8,$E7,$A4,$79,$27,$4D,$9D,$46,$27,$64,$70
	.dc.b	$48,$D9,$27,$CA,$D7,$CA,$D7,$CA,$D7,$CA,$70,$21,$C0,$A7,$02,$9C
	.dc.b	$0D,$10,$D1,$4D,$18,$21,$82,$1F,$04,$EA,$53,$91,$4E,$25,$38,$CF
	.dc.b	$C5,$8F,$64,$64,$91,$97,$AB,$5E,$AD,$7A,$B5,$EB,$3F,$3D,$E2,$D7
	.dc.b	$8B,$5E,$2D,$78,$CF,$92,$7A,$47,$92,$74,$D9,$D4,$62,$76,$47,$04
	.dc.b	$8D,$92,$7C,$AD,$7C,$AD,$7C,$AD,$7E,$97,$13,$F7,$97,$EB,$5F,$C0
	.dc.b	$63,$0D,$8C,$CF,$ED,$B7,$FE,$27,$4F,$DE,$9E,$46,$FF,$71,$E1,$12
	.dc.b	$F8,$0A,$F0,$4F,$79,$FD,$66,$FF,$C5,$67,$A7,$7E,$4D,$69,$FC,$57
	.dc.b	$A1,$FB,$FF,$F0,$BC,$BA,$C6,$38,$F5,$39,$9F,$E2,$E2,$78,$F1,$6B
	.dc.b	$D2,$93,$CA,$91,$CB,$DD,$E7,$F1,$12,$B4,$F6,$F9,$4F,$E5,$19,$3F
	.dc.b	$55,$7C,$76,$17,$CC,$D3,$11,$9F,$A5,$98,$FC,$6B,$E5,$1C,$7E,$C7
	.dc.b	$97,$DE,$CA,$D6,$28,$ED,$F8,$3F,$7F,$C4,$FF,$91,$63,$E3,$6D,$1A
	.dc.b	$63,$F1,$FF,$37,$E5,$73,$1E,$B1,$E9,$FC,$DE,$47,$81,$F6,$7E,$87
	.dc.b	$FC,$D7,$C1,$FC,$A4,$FE,$48,$87,$E4,$8A,$7E,$48,$A7,$E9,$8A,$7E
	.dc.b	$8C,$BE,$17,$E8,$7A,$7F,$31,$19,$3A,$10,$E8,$53,$A1,$4E,$06,$88
	.dc.b	$68,$A6,$8C,$10,$C1,$23,$C8,$F9,$A7,$E0,$BD,$E3,$2D,$8C,$BD,$A3
	.dc.b	$2D,$8C,$BD,$E3,$2D,$8C,$BE,$16,$C6,$A3,$13,$F1,$A3,$04,$9D,$92
	.dc.b	$34,$D9,$F5,$46,$08,$E0,$87,$AD,$CB,$4C,$9E,$B7,$2D,$32,$7A,$DC
	.dc.b	$B4,$C9,$EA,$2C,$FA,$63,$71,$93,$C0,$43,$B1,$4E,$85,$38,$1A,$21
	.dc.b	$A2,$9A,$30,$43,$04,$8F,$23,$E6,$9F,$94,$F7,$8C,$B6,$32,$F6,$8C
	.dc.b	$B6,$32,$F7,$8C,$B6,$32,$F8,$5B,$1A,$8C,$4F,$C6,$8C,$12,$76,$48
	.dc.b	$D3,$67,$D5,$18,$23,$82,$1E,$B7,$2D,$32,$7A,$DC,$B4,$C9,$FE,$01
	.dc.b	$F2,$9F,$AC,$FF,$49,$8F,$E8,$B1,$8F,$CF,$7F,$CA,$4D,$8C,$CE,$DF
	.dc.b	$E7,$71,$CB,$F5,$4D,$CF,$F5,$69,$CB,$7C,$0F,$F3,$05,$6E,$79,$FE
	.dc.b	$D1,$C0,$8F,$6F,$0F,$C3,$FB,$A7,$F9,$BA,$3C,$AF,$F9,$AD,$BE,$6E
	.dc.b	$BC,$DD,$F9,$9E,$8F,$DD,$7F,$CC,$7E,$57,$F9,$E3,$34,$63,$04,$31
	.dc.b	$FF,$33,$47,$C2,$9F,$EB,$1A,$7A,$1F,$F3,$1C,$67,$C5,$ED,$D5,$FE
	.dc.b	$75,$F8,$D7,$E2,$DF,$E6,$1A,$46,$1F,$F3,$23,$5F,$F3,$4B,$5F,$F3
	.dc.b	$E3,$5F,$F4,$63,$5F,$F4,$8B,$5E,$B1,$DE,$30,$47,$07,$8F,$E6,$5F
	.dc.b	$5B,$49,$5F,$E7,$78,$CB,$63,$2F,$F9,$FA,$32,$D8,$CB,$FE,$93,$8C
	.dc.b	$B6,$32,$FF,$AA,$35,$FE,$E9,$F9,$67,$FA,$15,$B1,$FE,$8A,$6B,$FE
	.dc.b	$9D,$6B,$FE,$AD,$6B,$CF,$BD,$3C,$DB,$1C,$DA,$F3,$6B,$CD,$AF,$38
	.dc.b	$E4,$7F,$78,$FD,$1B,$D7,$1F,$EC,$0C,$D1,$88,$FF,$55,$36,$33,$87
	.dc.b	$12,$4F,$FA,$9A,$32,$D9,$E0,$49,$DC,$7F,$AF,$67,$6D,$B7,$EA,$9F
	.dc.b	$E0,$A2,$47,$58,$CB,$63,$2F,$16,$BF,$ED,$79,$FF,$13,$46,$23,$DD
	.dc.b	$3C,$9E,$07,$FA,$FB,$CD,$75,$3E,$EF,$FD,$7D,$18,$6C,$66,$78,$35
	.dc.b	$F5,$7E,$BB,$FE,$33,$E7,$91,$FE,$2A,$8F,$F2,$64,$EE,$3F,$DD,$97
	.dc.b	$D5,$E7,$7D,$8F,$D7,$7D,$63,$D1,$FF,$73,$3F,$E5,$AE,$0F,$F7,$A3
	.dc.b	$6B,$FD,$F2,$ED,$FF,$31,$8F,$AE,$FE,$02,$FF,$95,$87,$D7,$FF,$07
	.dc.b	$69,$FE,$F3,$8D,$F4,$7F,$BB,$E7,$F8,$3D,$AA,$47,$F0,$87,$A3,$67
	.dc.b	$87,$07,$F9,$57,$B2,$57,$F9,$31,$FF,$2C,$76,$7F,$97,$3F,$5B,$FE
	.dc.b	$31,$46,$91,$D7,$56,$FD,$39,$87,$F9,$6E,$BC,$19,$FE,$6D,$7A,$7D
	.dc.b	$7D,$39,$5B,$8F,$07,$A7,$E1,$FB,$79,$3C,$E3,$C9,$F8,$0F,$97,$F0
	.dc.b	$C9,$3D,$E3,$D1,$F8,$AF,$A7,$F1,$F9,$3E,$38,$F6,$7E,$55,$ED,$FC
	.dc.b	$D2,$4F,$9E,$3E,$07,$E7,$FF,$07,$F4,$99,$3E,$B8,$F8,$9F,$A9,$FC
	.dc.b	$5E,$09,$3E,$F8,$F9,$1F,$B2,$7C,$9F,$DB,$A4,$FC,$38,$F9,$9F,$BB
	.dc.b	$48,$F8,$CF,$CF,$FD,$FF,$ED,$47,$E0,$F8,$58,$63,$AD,$CD,$A9,$CD
	.dc.b	$7E,$31,$AF,$1C,$FF,$D2,$BF,$B0,$70,$8F,$F4,$72,$F9,$6C,$9A,$BD
	.dc.b	$FD,$07,$0F,$A9,$97,$C2,$39,$FB,$3D,$3E,$F2,$4F,$48,$EF,$E6,$C8
	.dc.b	$EE,$F8,$BC,$7F,$1F,$E2,$BE,$4F,$C7,$65,$79,$23,$CF,$E9,$FA,$3F
	.dc.b	$30,$93,$E8,$8F,$5F,$A7,$23,$D6,$FB,$BF,$43,$F7,$F9,$FF,$03,$F5
	.dc.b	$59,$5F,$02,$3E,$57,$0B,$F8,$06,$00

;--------------------------------------------------------------------------------
;	FMOVECR.X #ccc,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmovecr882_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68882,-(sp)
	peamsg	'FMOVECR.X #CCC,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMOVECR.X #ccc,FPn',13,10
;------------------------------------------------
;	d1-d3	actual result
;	d4	actual status
;	d5	0=failed,1=successful
;	d6	fpcr=(XRN..XRP)<<4
;	d7	ccc=0..63
;	a0	expected result handle,expected status,...
;	a1-a3	expected result
;	a4	expected status
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a3
;decompress data
	move.l	a3,-(sp)
	pea.l	fmovecr882_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a3,a0			;expected result handle,expected status,...
@@:
	add.l	d0,(a3)+		;expected result handle
	addq.l	#4,a3			;expected status
	tst.l	(a3)
	bpl	@b
;	addq.l	#4,a3			;-1
;
	moveq.l	#0,d7			;ccc=0..63
77:
	move.l	#0,d6		;fpcr=(XRN..DRP)<<4
66:
;#ccc,FPn
	fmove.l	#0,fpcr
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d6,fpcr			;fpcr
	fmove.l	#0,fpsr
					;ccc
	jsr	([fmovecr882_execute,za0,d7.l*4])	;EXECUTE
	fmove.l	fpsr,d4			;actual status
	fmove.l	#0,fpcr
	fmove.x	fp7,-(sp)
	movem.l	(sp)+,d1-d3		;actual result
;
	movem.l	([a0]),a1-a3		;expected result
	movea.l	(4,a0),a4		;expected status
;
	move.l	#-1,-(sp)		;fpcr(rp<<6,-1=strict)
	movem.l	d1-d4/a1-a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d5			;0=failed,1=successful
;
	move.l	d5,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMOVECR.X #'
	puthex2	d7			;ccc
	putmsg	',FPn @'
	move.l	d6,-(sp)		;fpcr
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d5,-(sp)		;0=failed,1=successful
	movem.l	d1-d4/a1-a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a0			;expected result handle,expected status,...
;
	add.w	#1<<4,d6		;rprm++
	cmp.w	#176,d6		;fpcr=(XRN..DRP)<<4
	bls	66b
;
	addq.w	#1,d7			;ccc++
	cmp.w	#63,d7			;ccc=0..63
	bls	77b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
fmovecr882_execute::
	.dc.l	fmovecr882_execute_0
	.dc.l	fmovecr882_execute_1
	.dc.l	fmovecr882_execute_2
	.dc.l	fmovecr882_execute_3
	.dc.l	fmovecr882_execute_4
	.dc.l	fmovecr882_execute_5
	.dc.l	fmovecr882_execute_6
	.dc.l	fmovecr882_execute_7
	.dc.l	fmovecr882_execute_8
	.dc.l	fmovecr882_execute_9
	.dc.l	fmovecr882_execute_10
	.dc.l	fmovecr882_execute_11
	.dc.l	fmovecr882_execute_12
	.dc.l	fmovecr882_execute_13
	.dc.l	fmovecr882_execute_14
	.dc.l	fmovecr882_execute_15
	.dc.l	fmovecr882_execute_16
	.dc.l	fmovecr882_execute_17
	.dc.l	fmovecr882_execute_18
	.dc.l	fmovecr882_execute_19
	.dc.l	fmovecr882_execute_20
	.dc.l	fmovecr882_execute_21
	.dc.l	fmovecr882_execute_22
	.dc.l	fmovecr882_execute_23
	.dc.l	fmovecr882_execute_24
	.dc.l	fmovecr882_execute_25
	.dc.l	fmovecr882_execute_26
	.dc.l	fmovecr882_execute_27
	.dc.l	fmovecr882_execute_28
	.dc.l	fmovecr882_execute_29
	.dc.l	fmovecr882_execute_30
	.dc.l	fmovecr882_execute_31
	.dc.l	fmovecr882_execute_32
	.dc.l	fmovecr882_execute_33
	.dc.l	fmovecr882_execute_34
	.dc.l	fmovecr882_execute_35
	.dc.l	fmovecr882_execute_36
	.dc.l	fmovecr882_execute_37
	.dc.l	fmovecr882_execute_38
	.dc.l	fmovecr882_execute_39
	.dc.l	fmovecr882_execute_40
	.dc.l	fmovecr882_execute_41
	.dc.l	fmovecr882_execute_42
	.dc.l	fmovecr882_execute_43
	.dc.l	fmovecr882_execute_44
	.dc.l	fmovecr882_execute_45
	.dc.l	fmovecr882_execute_46
	.dc.l	fmovecr882_execute_47
	.dc.l	fmovecr882_execute_48
	.dc.l	fmovecr882_execute_49
	.dc.l	fmovecr882_execute_50
	.dc.l	fmovecr882_execute_51
	.dc.l	fmovecr882_execute_52
	.dc.l	fmovecr882_execute_53
	.dc.l	fmovecr882_execute_54
	.dc.l	fmovecr882_execute_55
	.dc.l	fmovecr882_execute_56
	.dc.l	fmovecr882_execute_57
	.dc.l	fmovecr882_execute_58
	.dc.l	fmovecr882_execute_59
	.dc.l	fmovecr882_execute_60
	.dc.l	fmovecr882_execute_61
	.dc.l	fmovecr882_execute_62
	.dc.l	fmovecr882_execute_63
fmovecr882_execute_0::
	fmovecr.x	#0,fp7
	rts
fmovecr882_execute_1::
	fmovecr.x	#1,fp7
	rts
fmovecr882_execute_2::
	fmovecr.x	#2,fp7
	rts
fmovecr882_execute_3::
	fmovecr.x	#3,fp7
	rts
fmovecr882_execute_4::
	fmovecr.x	#4,fp7
	rts
fmovecr882_execute_5::
	fmovecr.x	#5,fp7
	rts
fmovecr882_execute_6::
	fmovecr.x	#6,fp7
	rts
fmovecr882_execute_7::
	fmovecr.x	#7,fp7
	rts
fmovecr882_execute_8::
	fmovecr.x	#8,fp7
	rts
fmovecr882_execute_9::
	fmovecr.x	#9,fp7
	rts
fmovecr882_execute_10::
	fmovecr.x	#10,fp7
	rts
fmovecr882_execute_11::
	fmovecr.x	#11,fp7
	rts
fmovecr882_execute_12::
	fmovecr.x	#12,fp7
	rts
fmovecr882_execute_13::
	fmovecr.x	#13,fp7
	rts
fmovecr882_execute_14::
	fmovecr.x	#14,fp7
	rts
fmovecr882_execute_15::
	fmovecr.x	#15,fp7
	rts
fmovecr882_execute_16::
	fmovecr.x	#16,fp7
	rts
fmovecr882_execute_17::
	fmovecr.x	#17,fp7
	rts
fmovecr882_execute_18::
	fmovecr.x	#18,fp7
	rts
fmovecr882_execute_19::
	fmovecr.x	#19,fp7
	rts
fmovecr882_execute_20::
	fmovecr.x	#20,fp7
	rts
fmovecr882_execute_21::
	fmovecr.x	#21,fp7
	rts
fmovecr882_execute_22::
	fmovecr.x	#22,fp7
	rts
fmovecr882_execute_23::
	fmovecr.x	#23,fp7
	rts
fmovecr882_execute_24::
	fmovecr.x	#24,fp7
	rts
fmovecr882_execute_25::
	fmovecr.x	#25,fp7
	rts
fmovecr882_execute_26::
	fmovecr.x	#26,fp7
	rts
fmovecr882_execute_27::
	fmovecr.x	#27,fp7
	rts
fmovecr882_execute_28::
	fmovecr.x	#28,fp7
	rts
fmovecr882_execute_29::
	fmovecr.x	#29,fp7
	rts
fmovecr882_execute_30::
	fmovecr.x	#30,fp7
	rts
fmovecr882_execute_31::
	fmovecr.x	#31,fp7
	rts
fmovecr882_execute_32::
	fmovecr.x	#32,fp7
	rts
fmovecr882_execute_33::
	fmovecr.x	#33,fp7
	rts
fmovecr882_execute_34::
	fmovecr.x	#34,fp7
	rts
fmovecr882_execute_35::
	fmovecr.x	#35,fp7
	rts
fmovecr882_execute_36::
	fmovecr.x	#36,fp7
	rts
fmovecr882_execute_37::
	fmovecr.x	#37,fp7
	rts
fmovecr882_execute_38::
	fmovecr.x	#38,fp7
	rts
fmovecr882_execute_39::
	fmovecr.x	#39,fp7
	rts
fmovecr882_execute_40::
	fmovecr.x	#40,fp7
	rts
fmovecr882_execute_41::
	fmovecr.x	#41,fp7
	rts
fmovecr882_execute_42::
	fmovecr.x	#42,fp7
	rts
fmovecr882_execute_43::
	fmovecr.x	#43,fp7
	rts
fmovecr882_execute_44::
	fmovecr.x	#44,fp7
	rts
fmovecr882_execute_45::
	fmovecr.x	#45,fp7
	rts
fmovecr882_execute_46::
	fmovecr.x	#46,fp7
	rts
fmovecr882_execute_47::
	fmovecr.x	#47,fp7
	rts
fmovecr882_execute_48::
	fmovecr.x	#48,fp7
	rts
fmovecr882_execute_49::
	fmovecr.x	#49,fp7
	rts
fmovecr882_execute_50::
	fmovecr.x	#50,fp7
	rts
fmovecr882_execute_51::
	fmovecr.x	#51,fp7
	rts
fmovecr882_execute_52::
	fmovecr.x	#52,fp7
	rts
fmovecr882_execute_53::
	fmovecr.x	#53,fp7
	rts
fmovecr882_execute_54::
	fmovecr.x	#54,fp7
	rts
fmovecr882_execute_55::
	fmovecr.x	#55,fp7
	rts
fmovecr882_execute_56::
	fmovecr.x	#56,fp7
	rts
fmovecr882_execute_57::
	fmovecr.x	#57,fp7
	rts
fmovecr882_execute_58::
	fmovecr.x	#58,fp7
	rts
fmovecr882_execute_59::
	fmovecr.x	#59,fp7
	rts
fmovecr882_execute_60::
	fmovecr.x	#60,fp7
	rts
fmovecr882_execute_61::
	fmovecr.x	#61,fp7
	rts
fmovecr882_execute_62::
	fmovecr.x	#62,fp7
	rts
fmovecr882_execute_63::
	fmovecr.x	#63,fp7
	rts
	.cpu	68000

	.align	4
fmovecr882_expected_compressed::
	.dc.b	$00,$00,$18,$04,$13,$EC,$DF,$AF,$62,$14,$D3,$14,$FD,$AB,$53,$9A
	.dc.b	$E5,$18,$AD,$DD,$9E,$DF,$BB,$55,$F6,$FD,$DE,$F5,$E1,$8C,$61,$D6
	.dc.b	$58,$AF,$95,$38,$E9,$2B,$DB,$7C,$65,$DF,$E3,$A7,$EA,$4F,$D3,$BC
	.dc.b	$58,$7C,$49,$CC,$62,$3B,$4E,$DD,$CE,$9B,$0F,$D6,$7B,$36,$3C,$7A
	.dc.b	$75,$8C,$47,$C0,$3A,$FE,$C9,$C3,$0F,$06,$1D,$67,$E3,$35,$FA,$CD
	.dc.b	$7A,$7E,$DF,$C2,$CD,$24,$79,$BF,$7A,$F2,$91,$F2,$9F,$2F,$F5,$77
	.dc.b	$2F,$AA,$B8,$CF,$8A,$FA,$BF,$02,$3C,$F4,$EF,$72,$FC,$68,$C3,$63
	.dc.b	$35,$EC,$9C,$B7,$CA,$FC,$8B,$F1,$73,$4B,$D6,$E3,$C7,$F9,$19,$CC
	.dc.b	$63,$83,$ED,$53,$0E,$77,$B8,$EF,$87,$CA,$6B,$C3,$1F,$2B,$96,$BF
	.dc.b	$A5,$F5,$1F,$9E,$7E,$16,$11,$87,$F1,$0D,$7F,$1A,$D7,$F4,$0B,$F4
	.dc.b	$8F,$CA,$5B,$E1,$29,$FA,$79,$D5,$7D,$63,$BD,$70,$F9,$09,$19,$9D
	.dc.b	$BF,$26,$DF,$39,$F5,$89,$1F,$55,$F5,$7B,$5A,$8F,$C1,$F0,$7D,$0F
	.dc.b	$A8,$F2,$9F,$B9,$9D,$4F,$5A,$E6,$FD,$B6,$BF,$37,$ED,$3F,$0A,$47
	.dc.b	$F0,$8D,$7F,$04,$D7,$81,$A2,$1A,$29,$F6,$7F,$00,$FA,$AF,$E6,$5B
	.dc.b	$87,$E2,$1B,$89,$BF,$85,$7E,$1E,$BF,$40,$DA,$FC,$7B,$69,$F8,$B5
	.dc.b	$BE,$64,$E6,$FC,$F3,$DB,$F2,$29,$1F,$A1,$6E,$3F,$9F,$6B,$F9,$16
	.dc.b	$BF,$98,$6B,$FB,$36,$E3,$FA,$2F,$CB,$BF,$5E,$49,$FD,$73,$5F,$D6
	.dc.b	$35,$FC,$F3,$5F,$D2,$B5,$FE,$00,$FE,$EB,$F3,$A9,$1F,$B9,$6B,$FB
	.dc.b	$76,$BF,$A9,$6B,$FB,$06,$BF,$C2,$DB,$E4,$BF,$7A,$FE,$43,$D1,$FD
	.dc.b	$26,$31,$E1,$7F,$47,$6D,$FF,$14,$5E,$07,$EE,$E3,$17,$FC,$49,$FE
	.dc.b	$55,$3D,$67,$94,$F0,$9D,$F7,$75,$9F,$E3,$B1,$D5,$FE,$30,$DE,$BD
	.dc.b	$78,$1F,$D6,$7C,$A7,$EF,$FF,$D3,$EB,$F9,$3C,$63,$7D,$4E,$67,$F9
	.dc.b	$E4,$F0,$F1,$73,$F0,$92,$78,$D2,$38,$FC,$6F,$2F,$98,$95,$A7,$AF
	.dc.b	$D1,$FF,$59,$FD,$61,$E0,$DE,$E9,$7F,$B8,$9C,$B7,$C0,$E0,$47,$5D
	.dc.b	$FC,$FB,$24,$71,$7B,$77,$F7,$EE,$FE,$A7,$87,$E8,$AB,$FB,$0B,$5F
	.dc.b	$EA,$AD,$7F,$AA,$AE,$DF,$47,$6C,$9F,$34,$7A,$FB,$64,$7A,$5F,$77
	.dc.b	$E9,$1F,$C8,$7E,$63,$EB,$CF,$F2,$86,$BF,$CC,$DA,$FF,$87,$3C,$34
	.dc.b	$62,$3F,$9A,$B6,$33,$D3,$FC,$DA,$30,$47,$04,$9F,$E7,$31,$96,$C6
	.dc.b	$5F,$23,$5F,$23,$5F,$24,$71,$76,$72,$21,$82,$1C,$8C,$94,$C9,$D2
	.dc.b	$3B,$3E,$E6,$CF,$BA,$38,$9B,$29,$B2,$BB,$7A,$99,$2B,$97,$C4,$43
	.dc.b	$A1,$A9,$F7,$39,$39,$16,$39,$3C,$0C,$47,$07,$04,$38,$96,$38,$B8
	.dc.b	$23,$82,$47,$38,$CB,$63,$33,$F1,$A3,$04,$70,$49,$F8,$B1,$96,$C6
	.dc.b	$5F,$80,$D7,$D6,$D7,$D1,$1C,$5D,$9C,$88,$60,$87,$23,$25,$32,$74
	.dc.b	$8E,$CF,$B9,$B3,$EE,$8E,$26,$CA,$6C,$AE,$DE,$A6,$4A,$E5,$F1,$10
	.dc.b	$E8,$6A,$7D,$CE,$4E,$45,$8E,$4F,$03,$11,$C1,$C1,$0E,$25,$8E,$2E
	.dc.b	$08,$E0,$91,$CE,$32,$D8,$CC,$FC,$68,$C1,$1C,$12,$7E,$2C,$65,$B1
	.dc.b	$97,$E0,$35,$F5,$B5,$F4,$47,$17,$67,$22,$18,$21,$C8,$C9,$4C,$9D
	.dc.b	$23,$B3,$EE,$6C,$FB,$A3,$89,$B2,$9B,$2B,$B7,$A9,$92,$B9,$7C,$44
	.dc.b	$3A,$1A,$9F,$73,$93,$91,$63,$93,$C0,$C4,$70,$70,$43,$89,$63,$8B
	.dc.b	$82,$38,$24,$73,$8F,$C4,$FA,$6F,$D6,$BF,$C0,$91,$FE,$B1,$8C,$79
	.dc.b	$1F,$ED,$76,$C7,$FA,$91,$E1,$DD,$9A,$FF,$61,$4F,$17,$FC,$37,$E5
	.dc.b	$7F,$BB,$2B,$B7,$87,$BF,$37,$87,$9D,$30,$E9,$82,$3A,$78,$7E,$1F
	.dc.b	$DD,$3F,$DE,$75,$E1,$7F,$DE,$0D,$F5,$35,$EA,$6F,$D3,$F4,$7E,$EB
	.dc.b	$FE,$E8,$DB,$C4,$DB,$6C,$C7,$FD,$D5,$1E,$D9,$FF,$0C,$79,$7F,$EF
	.dc.b	$BA,$47,$FB,$A2,$75,$E3,$76,$F2,$3F,$DF,$5F,$22,$FC,$67,$FD,$44
	.dc.b	$DA,$FF,$74,$35,$FF,$76,$B5,$FF,$80,$0D,$44,$70,$D8,$EA,$47,$AB
	.dc.b	$5E,$AF,$FA,$CF,$E7,$1F,$61,$49,$3F,$EF,$B8,$CB,$63,$2F,$88,$91
	.dc.b	$A7,$07,$99,$C3,$86,$BC,$23,$9B,$97,$8B,$EE,$D5,$DE,$FC,$BF,$32
	.dc.b	$9C,$C6,$1B,$1C,$9A,$F9,$1A,$FA,$5A,$F7,$9E,$CF,$16,$BC,$5A,$F1
	.dc.b	$6B,$C6,$7C,$51,$88,$64,$FD,$27,$E0,$C6,$1B,$19,$C3,$DA,$D9,$F8
	.dc.b	$6D,$7E,$6B,$5F,$A4,$D7,$F0,$AD,$7B,$47,$49,$EE,$47,$FC,$AF,$FA
	.dc.b	$B4,$9F,$F2,$BB,$5F,$F2,$BB,$5F,$F2,$BB,$5E,$F6,$8C,$47,$C0,$8C
	.dc.b	$3D,$0F,$BB,$E7,$BB,$1F,$7D,$F7,$23,$0D,$8C,$CF,$06,$BF,$E6,$AF
	.dc.b	$D7,$7F,$CF,$DC,$F2,$3F,$CF,$91,$FE,$93,$9D,$C7,$E5,$EF,$AB,$CE
	.dc.b	$FB,$1F,$8C,$5F,$5F,$FE,$9E,$5C,$DF,$8F,$7F,$63,$1F,$E6,$E8,$C4
	.dc.b	$70,$F0,$BF,$49,$18,$9E,$4E,$BE,$C7,$F8,$5D,$74,$E2,$FD,$57,$D9
	.dc.b	$7F,$11,$83,$67,$75,$DA,$3F,$88,$D7,$0E,$7E,$ED,$BF,$E0,$B5,$BE
	.dc.b	$3F,$F5,$17,$64,$9F,$F5,$03,$FE,$A5,$EC,$E3,$1F,$96,$FD,$77,$F9
	.dc.b	$15,$1A,$47,$66,$71,$B1,$FE,$B5,$9F,$0E,$3F,$92,$8C,$47,$BC,$ED
	.dc.b	$3C,$3D,$3E,$3E,$9C,$9E,$27,$8E,$3A,$FA,$7D,$BE,$EA,$4E,$B2,$F3
	.dc.b	$4F,$8B,$F0,$4F,$1F,$E1,$F2,$B5,$9A,$7C,$FF,$8C,$7A,$3F,$22,$95
	.dc.b	$A8,$F5,$FE,$5B,$EC,$FC,$DE,$4F,$28,$F7,$FE,$85,$F0,$3F,$4C,$93
	.dc.b	$DA,$3E,$1F,$EA,$BF,$13,$39,$3E,$28,$F8,$FF,$B2,$7C,$8F,$DB,$A4
	.dc.b	$F9,$A3,$E5,$FE,$ED,$23,$D2,$FC,$EF,$DF,$FE,$BC,$38,$58,$61,$B1
	.dc.b	$F2,$8D,$46,$71,$FB,$3F,$1D,$F9,$8E,$BF,$CE,$72,$21,$F3,$7A,$37
	.dc.b	$D5,$C3,$E9,$78,$FD,$64,$CF,$E9,$5A,$7A,$FB,$3D,$BE,$FA,$56,$A3
	.dc.b	$C3,$E6,$48,$E2,$F9,$3C,$6F,$2F,$E2,$FE,$6F,$C7,$E5,$76,$8F,$4F
	.dc.b	$A5,$EA,$FC,$CA,$4F,$8A,$3D,$BE,$94,$8F,$2B,$F0,$3F,$45,$F8,$3E
	.dc.b	$77,$C2,$F9,$39,$5E,$A8,$FD,$5F,$F0,$08,$80

;--------------------------------------------------------------------------------
;	FMUL.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fmul_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#63,-(sp)
	peamsg	'FMUL.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FMUL.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fmul_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fmul.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMUL.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fmul.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FMUL.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fmul_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3A,$F9,$AF,$EC,$ED,$2D,$FC,$E8,$CC,$76,$B4,$62
	.dc.b	$75,$2F,$A0,$F5,$BE,$9C,$FF,$BA,$46,$23,$CF,$18,$CF,$D7,$87,$49
	.dc.b	$E5,$18,$C3,$FB,$ED,$F1,$C3,$DD,$5A,$3B,$1A,$AE,$0E,$4F,$45,$7F
	.dc.b	$81,$BE,$7F,$96,$3F,$E0,$76,$CE,$6B,$8C,$77,$C7,$C9,$8E,$AB,$A6
	.dc.b	$3E,$28,$D6,$3E,$19,$F9,$51,$B2,$61,$E4,$7E,$2F,$CA,$BF,$5D,$7E
	.dc.b	$9B,$FE,$12,$F8,$5F,$F0,$C4,$63,$0F,$F0,$EE,$3F,$C2,$6F,$85,$EE
	.dc.b	$7F,$C2,$B4,$EB,$C5,$75,$F2,$AC,$FF,$7C,$AE,$78,$F6,$9C,$F9,$5A
	.dc.b	$ED,$DF,$87,$2F,$67,$68,$F0,$7C,$EF,$EE,$BC,$EF,$C2,$B6,$7F,$16
	.dc.b	$D7,$FC,$66,$D7,$FC,$43,$E4,$7E,$71,$CB,$5C,$F8,$5D,$23,$11,$FE
	.dc.b	$3B,$FD,$7B,$FC,$7B,$C3,$F0,$DB,$AE,$67,$E0,$C7,$F8,$31,$FC,$BC
	.dc.b	$7F,$91,$BE,$51,$ED,$BE,$0B,$E5,$46,$1B,$19,$AF,$83,$C4,$C3,$F7
	.dc.b	$18,$FF,$3B,$F9,$F7,$E3,$3F,$A6,$B4,$61,$FD,$4B,$5F,$D7,$B5,$FF
	.dc.b	$2B,$DF,$5F,$93,$F4,$92,$3C,$38,$7F,$89,$1F,$58,$DF,$F9,$5F,$C2
	.dc.b	$7D,$4D,$FF,$82,$67,$FD,$3E,$73,$5C,$BE,$4B,$5D,$BF,$34,$91,$F6
	.dc.b	$2F,$F9,$6B,$7C,$1E,$84,$7C,$7D,$BD,$FE,$66,$47,$E1,$B1,$FE,$7B
	.dc.b	$1F,$4B,$F0,$3F,$CB,$66,$FD,$BF,$FC,$B5,$7F,$C4,$23,$1E,$47,$F0
	.dc.b	$E6,$C6,$67,$6F,$E1,$BD,$9F,$71,$1F,$8B,$F0,$FF,$07,$C2,$91,$E3
	.dc.b	$DB,$F5,$4F,$C8,$E4,$FB,$5F,$0D,$FA,$98,$FE,$7B,$E5,$FD,$9F,$3C
	.dc.b	$D2,$47,$F2,$6A,$FC,$17,$A1,$FC,$91,$B1,$9E,$7F,$E9,$4D,$9D,$47
	.dc.b	$F5,$58,$FB,$ED,$7F,$CD,$AD,$7F,$3D,$F2,$BE,$3F,$81,$91,$FC,$CB
	.dc.b	$14,$62,$39,$FC,$AF,$62,$3E,$37,$F6,$BE,$7F,$F4,$F8,$FF,$48,$B5
	.dc.b	$FF,$54,$F3,$7F,$A8,$7E,$5F,$67,$E6,$FD,$8F,$EF,$F3,$FD,$09,$AF
	.dc.b	$F4,$86,$BF,$EB,$06,$BD,$2B,$FC,$00,$F8,$3B,$BC,$64,$7F,$83,$C7
	.dc.b	$E7,$5F,$2F,$CB,$FA,$08,$F8,$7E,$47,$C8,$47,$07,$2E,$FF,$79,$1C
	.dc.b	$1F,$D7,$BC,$5F,$EE,$5D,$FF,$C0,$3D,$2F,$F0,$C4,$FE,$B2,$72,$FE
	.dc.b	$6F,$4F,$F3,$E3,$FC,$88,$7D,$9F,$07,$FD,$9E,$D7,$FD,$A4,$D7,$FD
	.dc.b	$AA,$D9,$FF,$61,$FC,$4F,$F1,$0F,$92,$34,$FC,$A4,$FE,$CA,$BF,$31
	.dc.b	$CB,$FC,$61,$FE,$07,$E4,$7F,$1E,$3F,$8F,$FC,$6F,$F2,$0F,$81,$93
	.dc.b	$FC,$21,$FE,$47,$F3,$1F,$C9,$3E,$3A,$4F,$F0,$C7,$F9,$3F,$C7,$7F
	.dc.b	$28,$CF,$F9,$4F,$CC,$4A,$FE,$23,$1F,$C9,$BF,$20,$FE,$43,$E8,$FF
	.dc.b	$BB,$FE,$CF,$F8,$A7,$93,$FE,$27,$6C,$7F,$2F,$9D,$DF,$FB,$DA,$7F
	.dc.b	$06,$F1,$EA,$FE,$51,$18,$8D,$46,$39,$3F,$9A,$FE,$4D,$F4,$A7,$EB
	.dc.b	$37,$1F,$04,$FF,$40,$F9,$B9,$5F,$CA,$1F,$EC,$7E,$D7,$F8,$EE,$BF
	.dc.b	$B5,$3F,$E7,$16,$FC,$D6,$8D,$FC,$D7,$0F,$91,$98,$7F,$42,$B4,$74
	.dc.b	$EE,$EB,$F1,$32,$B4,$F8,$3E,$3F,$F8,$77,$B4,$F7,$A9,$2B,$F1,$9F
	.dc.b	$B1,$FF,$A3,$1F,$C7,$7C,$B7,$EF,$9F,$F1,$8B,$5F,$F1,$BB,$51,$FE
	.dc.b	$86,$3F,$C6,$FF,$8E,$FF,$7F,$7F,$A3,$87,$1B,$FF,$97,$5B,$CD,$FE
	.dc.b	$8F,$9F,$77,$83,$F2,$08,$FC,$1F,$9C,$F9,$7F,$31,$F3,$3C,$D9,$3F
	.dc.b	$16,$3F,$B5,$FE,$93,$FD,$A6,$3F,$58,$DC,$BF,$D4,$5E,$32,$31,$97
	.dc.b	$F7,$57,$F9,$FC,$EB,$F4,$D7,$47,$F7,$66,$E1,$FE,$14,$C1,$18,$8E
	.dc.b	$FF,$1D,$FD,$F3,$F5,$0F,$91,$87,$81,$B9,$7F,$80,$DB,$5F,$DA,$FE
	.dc.b	$E7,$F0,$93,$FE,$BF,$F5,$3F,$BE,$7C,$EF,$F3,$5A,$FF,$62,$46,$0F
	.dc.b	$F3,$2E,$6C,$BF,$A0,$CF,$F0,$B6,$E5,$F0,$1B,$87,$FB,$2F,$D6,$FF
	.dc.b	$08,$FE,$BD,$F8,$5B,$F6,$35,$FE,$28,$DB,$F9,$3F,$AF,$FF,$88,$25
	.dc.b	$8B,$B9,$EA,$7F,$87,$1B,$3F,$1D,$AF,$4F,$8C,$FD,$2F,$CF,$C8,$FC
	.dc.b	$F3,$FA,$8F,$96,$FD,$4F,$CC,$49,$FD,$13,$FA,$CF,$97,$FF,$21,$E2
	.dc.b	$F3,$BF,$A7,$92,$BF,$9E,$BF,$D9,$F2,$F8,$D1,$FA,$BE,$1F,$DA,$9F
	.dc.b	$B6,$FA,$2F,$DB,$F8,$5F,$B8,$F8,$79,$5F,$B1,$8E,$FE,$87,$F5,$C2
	.dc.b	$47,$F4,$A7,$FC,$A7,$F3,$7F,$EB,$98,$FD,$29,$FE,$27,$70,$FF,$60
	.dc.b	$6B,$9E,$6F,$37,$C1,$7F,$08,$24,$7F,$01,$FA,$0F,$F3,$6F,$EC,$3F
	.dc.b	$C4,$E3,$AF,$C2,$23,$11,$FB,$48,$C7,$37,$F8,$1E,$73,$29,$E1,$F3
	.dc.b	$3D,$09,$1F,$E5,$09,$FF,$7D,$37,$E8,$7C,$24,$75,$F2,$9E,$4E,$49
	.dc.b	$1C,$5F,$3F,$C9,$7A,$3E,$07,$C4,$49,$FF,$0E,$61,$FE,$AC,$7D,$6F
	.dc.b	$F5,$1E,$DF,$F4,$83,$77,$F6,$36,$7D,$4F,$F6,$A9,$FF,$74,$7D,$23
	.dc.b	$D2,$FE,$D3,$F5,$9F,$ED,$CB,$4E,$78,$3F,$8C,$CE,$63,$12,$8C,$7A
	.dc.b	$7F,$69,$D5,$F4,$DF,$EE,$DF,$09,$CE,$1D,$27,$37,$C5,$DD,$F1,$C3
	.dc.b	$F9,$4E,$1C,$67,$37,$C6,$73,$7C,$5D,$DF,$17,$D6,$DB,$E5,$1C,$0E
	.dc.b	$46,$AA,$C6,$A3,$07,$BE,$FC,$EE,$AB,$A1,$2B,$8D,$F8,$E7,$8E,$1F
	.dc.b	$01,$E4,$79,$A7,$6F,$36,$3A,$38,$DF,$81,$C6,$1E,$52,$61,$ED,$C3
	.dc.b	$C1,$18,$BE,$78,$7D,$6B,$B8,$7C,$A7,$9E,$1C,$5C,$9E,$53,$35,$E6
	.dc.b	$69,$FB,$83,$45,$BF,$C2,$B6,$7F,$57,$18,$8F,$01,$99,$E4,$F2,$3C
	.dc.b	$97,$F1,$C9,$7F,$05,$B1,$E5,$6B,$F9,$8C,$3E,$0E,$1F,$4E,$3C,$F3
	.dc.b	$F0,$5B,$5D,$88,$FE,$0C,$C9,$D5,$CB,$D4,$F8,$A7,$D4,$34,$71,$8F
	.dc.b	$1E,$1F,$46,$3E,$BD,$F9,$30,$E3,$3F,$66,$31,$3E,$03,$A1,$DC,$C4
	.dc.b	$73,$AF,$51,$DB,$0F,$73,$8C,$3C,$F3,$B7,$E4,$1F,$C1,$3E,$53,$F9
	.dc.b	$29,$2F,$F8,$D7,$C2,$FF,$11,$7F,$87,$ED,$FC,$45,$FF,$70,$EB,$FC
	.dc.b	$45,$FE,$1B,$C1,$FC,$69,$FF,$5C,$47,$F5,$8D,$BF,$89,$F0,$7F,$21
	.dc.b	$9C,$E5,$67,$F8,$BF,$0F,$F2,$A9,$E3,$C5,$67,$F8,$DF,$CB,$FF,$3A
	.dc.b	$E2,$FE,$68,$4A,$FE,$46,$FF,$42,$7F,$C8,$1C,$DF,$D0,$9F,$F6,$57
	.dc.b	$6B,$9F,$FA,$6E,$1F,$EE,$FE,$8F,$E9,$7F,$14,$69,$3A,$3F,$AB,$E1
	.dc.b	$F9,$D9,$FE,$EF,$D5,$88,$FF,$5A,$BF,$D9,$FA,$FF,$B0,$C7,$83,$B3
	.dc.b	$FB,$2B,$FD,$A4,$87,$93,$BB,$FB,$53,$FE,$F3,$E5,$FE,$E8,$FF,$72
	.dc.b	$21,$E1,$72,$FE,$BF,$B7,$FC,$15,$E0,$FF,$7F,$8F,$7B,$FD,$FD,$FF
	.dc.b	$02,$90,$F8,$7E,$1F,$F8,$1D,$FF,$07,$57,$F8,$44,$8F,$F8,$73,$C4
	.dc.b	$FF,$0F,$FA,$D2,$3F,$C3,$EF,$CA,$DB,$FC,$B7,$C8,$F1,$BF,$C3,$CF
	.dc.b	$F8,$B7,$C8,$D5,$FF,$88,$35,$FF,$15,$B6,$3F,$DF,$9E,$56,$20,$FF
	.dc.b	$8E,$5B,$E5,$7F,$8E,$5F,$E0,$A6,$0F,$F6,$37,$99,$FE,$44,$C0,$D3
	.dc.b	$FC,$8F,$C8,$F3,$3F,$C8,$F5,$FE,$32,$F3,$7F,$CA,$5D,$1F,$E5,$9F
	.dc.b	$3B,$FC,$B0,$7F,$99,$35,$FF,$32,$F3,$CB,$FF,$2D,$79,$FF,$E6,$27
	.dc.b	$FC,$76,$FF,$98,$B8,$7F,$C1,$5D,$8F,$43,$FC,$DF,$5F,$92,$F4,$7F
	.dc.b	$CE,$6F,$F9,$E0,$87,$E5,$BD,$3F,$F5,$0F,$A7,$F2,$23,$FD,$05,$77
	.dc.b	$D4,$FF,$40,$C7,$FA,$29,$AB,$D5,$F0,$3F,$E9,$4C,$3E,$3B,$96,$C7
	.dc.b	$FA,$4E,$3F,$D2,$47,$F3,$0D,$3F,$CB,$19,$FF,$8C,$9B,$1F,$E9,$86
	.dc.b	$AF,$5B,$FD,$44,$FE,$BB,$D1,$FF,$56,$34,$FF,$59,$7A,$FF,$EB,$47
	.dc.b	$F6,$7E,$C5,$E3,$FF,$5E,$FA,$7F,$EB,$B9,$FF,$5C,$7B,$55,$FD,$CF
	.dc.b	$B5,$FE,$C6,$7F,$DD,$BC,$D6,$FF,$D8,$BA,$FC,$7F,$7B,$FD,$B3,$D8
	.dc.b	$F7,$7F,$DA,$13,$F9,$66,$E1,$FE,$DD,$F7,$78,$7B,$DF,$ED,$C2,$47
	.dc.b	$E3,$B0,$FF,$6F,$60,$F7,$FF,$DC,$B2,$9F,$F7,$4B,$FE,$E7,$F8,$9F
	.dc.b	$F7,$86,$1F,$EF,$58,$FF,$5F,$E6,$CF,$FC,$9D,$F0,$EF,$3B,$FD,$F1
	.dc.b	$4A,$D3,$FE,$4F,$97,$C1,$7E,$AB,$CE,$43,$E0,$F0,$DA,$EF,$8F,$69
	.dc.b	$CF,$C0,$6B,$E1,$65,$68,$EC,$F2,$F8,$5E,$7F,$0F,$27,$59,$78,$23
	.dc.b	$BF,$C3,$FE,$F3,$8D,$E2,$79,$AF,$FD,$7C,$D7,$D9,$A7,$F0,$68,$C6
	.dc.b	$FE,$E9,$7C,$47,$F9,$55,$BF,$03,$F9,$88,$34,$91,$FE,$06,$C1,$F0
	.dc.b	$3F,$E5,$BC,$7F,$89,$B6,$7F,$D5,$ED,$F3,$5C,$BF,$82,$7F,$D6,$3F
	.dc.b	$13,$FC,$31,$B5,$FE,$12,$6B,$FC,$C6,$3F,$57,$F1,$5F,$E6,$78,$FE
	.dc.b	$26,$D7,$FC,$32,$D7,$FC,$48,$D7,$FE,$00,$CF,$FB,$03,$A3,$C1,$EB
	.dc.b	$9A,$49,$F1,$43,$B3,$64,$8C,$BE,$4F,$0E,$47,$91,$AF,$F4,$9C,$BB
	.dc.b	$91,$EA,$FF,$2C,$E4,$FC,$6F,$C5,$DE,$A3,$49,$5F,$CC,$BE,$33,$F8
	.dc.b	$52,$0E,$1B,$E5,$7F,$12,$8D,$DF,$E6,$BD,$4E,$C4,$AF,$F2,$0C,$65
	.dc.b	$B8,$7F,$BD,$9F,$F5,$CB,$5F,$D5,$FA,$DE,$73,$B7,$BF,$23,$CC,$D7
	.dc.b	$FB,$D7,$C7,$3E,$3F,$FA,$AF,$12,$31,$3F,$A5,$8C,$4F,$EB,$BE,$3F
	.dc.b	$74,$F9,$1A,$4B,$FD,$3E,$4F,$91,$FE,$C3,$DA,$F9,$27,$C1,$FF,$5A
	.dc.b	$8C,$36,$33,$7F,$E6,$29,$CB,$F9,$3C,$5F,$29,$E2,$24,$78,$3E,$53
	.dc.b	$CB,$E0,$48,$F5,$57,$E8,$F8,$BF,$84,$FC,$AF,$AB,$E3,$FF,$41,$26
	.dc.b	$92,$BC,$1F,$1D,$F0,$08,$FB,$BC,$0F,$E2,$3D,$B2,$3E,$0B,$FC,$53
	.dc.b	$BF,$FC,$03,$F2,$CF,$97,$FE,$A6,$DA,$FE,$F6,$D7,$FD,$02,$D7,$FD
	.dc.b	$19,$E2,$BC,$9F,$EE,$0D,$F6,$BF,$89,$47,$F7,$3D,$FF,$92,$F8,$5F
	.dc.b	$C9,$2E,$1F,$50,$FE,$4F,$1F,$6B,$E3,$FE,$09,$23,$E7,$3F,$13,$E3
	.dc.b	$A4,$7D,$16,$BF,$E3,$1F,$99,$7C,$CF,$F3,$57,$FB,$3D,$9A,$6A,$7F
	.dc.b	$95,$BF,$DE,$A3,$5F,$03,$C4,$E5,$F1,$9C,$FC,$19,$7C,$1E,$DF,$35
	.dc.b	$FE,$05,$F9,$CF,$B1,$DB,$FB,$1F,$9C,$75,$FE,$CF,$05,$7F,$26,$75
	.dc.b	$19,$8C,$42,$78,$BF,$BB,$8E,$AE,$67,$87,$57,$F5,$78,$C1,$BF,$3F
	.dc.b	$C1,$7C,$A3,$C3,$18,$AE,$38,$78,$A7,$B5,$F3,$F9,$4E,$18,$F9,$1B
	.dc.b	$87,$AD,$CC,$EF,$0E,$77,$E2,$3B,$9D,$4A,$66,$FD,$78,$F2,$C7,$51
	.dc.b	$F0,$A3,$BC,$E2,$FD,$D1,$B8,$FF,$6A,$CF,$C4,$7F,$DC,$0D,$7C,$66
	.dc.b	$3B,$9F,$61,$FE,$58,$9F,$EB,$6D,$8C,$CF,$46,$BD,$2F,$E7,$37,$0E
	.dc.b	$51,$88,$EE,$43,$E5,$C6,$88,$FA,$23,$83,$A3,$11,$A7,$B1,$82,$18
	.dc.b	$21,$DC,$FC,$6C,$74,$6C,$73,$75,$7F,$15,$F0,$1C,$8A,$6A,$79,$46
	.dc.b	$E3,$99,$E2,$35,$3D,$5E,$24,$36,$75,$29,$C0,$E4,$6E,$3B,$12,$7C
	.dc.b	$24,$9E,$D3,$C8,$E1,$19,$79,$99,$24,$F3,$9E,$2D,$8D,$B5,$F3,$35
	.dc.b	$F3,$18,$9E,$CD,$9E,$71,$88,$D9,$0F,$2C,$68,$8F,$82,$38,$3A,$31
	.dc.b	$1A,$7B,$18,$21,$82,$1D,$CF,$7C,$74,$6C,$73,$75,$3E,$D7,$C0,$72
	.dc.b	$29,$A9,$E5,$1B,$8E,$67,$88,$D4,$F5,$78,$90,$D9,$D4,$A7,$03,$91
	.dc.b	$B8,$EC,$49,$F0,$92,$7B,$4F,$23,$84,$65,$E6,$64,$93,$CE,$78,$B6
	.dc.b	$36,$D7,$CC,$D7,$CC,$62,$7B,$36,$79,$C6,$23,$64,$3C,$B1,$A2,$3E
	.dc.b	$08,$E0,$E8,$C4,$69,$EC,$60,$86,$08,$77,$3D,$F1,$D1,$B1,$CD,$D4
	.dc.b	$FB,$5F,$01,$C8,$A6,$A7,$94,$6E,$39,$9E,$23,$53,$D5,$E2,$43,$67
	.dc.b	$52,$9C,$0E,$46,$E3,$B1,$F8,$2A,$FE,$67,$F0,$0F,$3B,$FB,$34,$61
	.dc.b	$B8,$FF,$97,$E7,$57,$FD,$9E,$B4,$D9,$D5,$70,$85,$70,$7F,$BA,$D7
	.dc.b	$06,$D7,$07,$FB,$D0,$F9,$9F,$F7,$DB,$4F,$F7,$EB,$82,$38,$21,$C8
	.dc.b	$FF,$3B,$3A,$70,$73,$2B,$97,$83,$5F,$1B,$5F,$F0,$73,$5E,$A7,$80
	.dc.b	$E2,$53,$89,$4E,$25,$38,$9E,$23,$07,$94,$E6,$64,$86,$08,$76,$32
	.dc.b	$53,$27,$80,$C9,$4C,$9E,$22,$9E,$13,$91,$0E,$05,$38,$14,$E0,$68
	.dc.b	$86,$8A,$68,$C1,$0C,$11,$F9,$CD,$3E,$2B,$82,$38,$21,$EE,$3A,$3A
	.dc.b	$70,$73,$2B,$97,$E0,$B5,$F6,$35,$F4,$B5,$EA,$78,$0E,$25,$38,$94
	.dc.b	$E2,$53,$89,$E2,$30,$79,$4E,$66,$48,$60,$87,$63,$25,$32,$78,$0C
	.dc.b	$94,$C9,$E2,$29,$E1,$39,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$C1,$1F,$9C,$D3,$E2,$B8,$23,$82,$1E,$E3,$A3,$A7,$07,$32,$B9
	.dc.b	$7E,$0B,$5F,$63,$5F,$4B,$5E,$A7,$80,$E2,$53,$89,$4E,$25,$38,$9E
	.dc.b	$23,$07,$94,$E6,$64,$86,$08,$76,$32,$53,$27,$80,$C9,$4C,$9E,$22
	.dc.b	$9E,$13,$91,$0E,$07,$F0,$3C,$3F,$CF,$AF,$F9,$EE,$7F,$CF,$C4,$7F
	.dc.b	$A8,$9F,$E1,$CC,$FF,$D1,$EF,$FA,$57,$6F,$F5,$33,$FE,$75,$8F,$F4
	.dc.b	$81,$A2,$6B,$FE,$A0,$7F,$D2,$9B,$72,$24,$7F,$A9,$1E,$2F,$FA,$FE
	.dc.b	$BA,$D7,$FA,$92,$B2,$F7,$9D,$3F,$EA,$49,$DE,$1F,$EA,$89,$F0,$1A
	.dc.b	$9F,$41,$FE,$A7,$D6,$38,$25,$F6,$7F,$CF,$6F,$A9,$FF,$64,$1C,$1D
	.dc.b	$57,$FA,$CE,$BB,$46,$36,$FF,$60,$57,$01,$8F,$C0,$7F,$D3,$8F,$2C
	.dc.b	$7D,$27,$AA,$7F,$D7,$77,$EA,$AE,$57,$E3,$7C,$B1,$F2,$4F,$75,$F5
	.dc.b	$8F,$F4,$9B,$D5,$F8,$C7,$A6,$3A,$3F,$35,$F4,$98,$C3,$FD,$D7,$3C
	.dc.b	$DF,$F6,$F4,$77,$9F,$A4,$79,$1E,$2E,$0F,$29,$92,$18,$3C,$AF,$A0
	.dc.b	$8F,$C4,$7C,$58,$7C,$48,$F4,$1D,$67,$E8,$3D,$1F,$C0,$3C,$4F,$A1
	.dc.b	$5E,$9C,$3E,$9C,$78,$67,$8B,$F0,$4E,$91,$F4,$1E,$95,$F4,$F1,$FC
	.dc.b	$3D,$FC,$38,$FF,$7F,$61,$F0,$EB,$A4,$79,$DF,$F7,$D5,$F3,$72,$7D
	.dc.b	$23,$37,$D2,$BE,$0C,$30,$F9,$EF,$AC,$E2,$7E,$5A,$3E,$B5,$75,$9F
	.dc.b	$59,$1C,$1F,$01,$E7,$3D,$6F,$D2,$F5,$8F,$92,$FC,$32,$38,$24,$7B
	.dc.b	$71,$FC,$F6,$0A,$FC,$EE,$1E,$AB,$EE,$7E,$29,$FB,$71,$F0,$6B,$8B
	.dc.b	$EB,$3E,$D6,$1E,$8B,$D5,$7E,$32,$7E,$C4,$FB,$F0,$F1,$4F,$C2,$AD
	.dc.b	$D7,$27,$F0,$EF,$AA,$7E,$D3,$F4,$0F,$01,$EE,$AF,$03,$8C,$3E,$55
	.dc.b	$72,$7E,$6D,$72,$C3,$F5,$95,$BB,$E3,$1F,$B5,$AE,$33,$DE,$31,$5C
	.dc.b	$70,$F8,$73,$C7,$0F,$C2,$3F,$E6,$AB,$57,$F9,$82,$3E,$C1,$FE,$62
	.dc.b	$8C,$B7,$C1,$FA,$4A,$9E,$A6,$29,$5A,$9C,$D7,$D9,$9C,$BF,$6E,$32
	.dc.b	$F8,$0E,$95,$F7,$FB,$5F,$91,$BC,$EE,$93,$98,$C5,$73,$8C,$4E,$A1
	.dc.b	$5C,$BE,$5F,$0F,$4F,$2A,$E0,$69,$8B,$E5,$E1,$3F,$4C,$79,$38,$6E
	.dc.b	$5E,$11,$3B,$9D,$79,$9B,$E1,$95,$BA,$E9,$1C,$EB,$D1,$3F,$B5,$F5
	.dc.b	$AF,$01,$A6,$BF,$6A,$FF,$51,$F0,$FC,$A9,$78,$9F,$6D,$B7,$F6,$E3
	.dc.b	$06,$7E,$6B,$A9,$23,$D4,$7F,$05,$8F,$83,$FA,$93,$E3,$E9,$27,$D5
	.dc.b	$1F,$EB,$9C,$3E,$FE,$A8,$7C,$87,$22,$57,$0B,$FA,$7F,$A8,$BF,$50
	.dc.b	$A4,$9F,$99,$CC,$97,$87,$C3,$C4,$95,$A9,$47,$18,$FC,$59,$F8,$2F
	.dc.b	$DA,$1E,$D7,$FA,$FB,$04,$62,$7F,$D8,$91,$89,$D4,$E6,$7F,$DA,$3E
	.dc.b	$4B,$E7,$DF,$5D,$F8,$F8,$C3,$63,$33,$EA,$AE,$5C,$3F,$A4,$C5,$FB
	.dc.b	$7B,$CC,$FC,$AB,$6F,$E0,$B5,$F5,$CF,$E5,$FD,$5E,$DE,$C4,$9F,$C9
	.dc.b	$69,$E5,$8E,$78,$7E,$83,$F7,$67,$C5,$7C,$16,$DF,$C0,$6B,$FE,$EE
	.dc.b	$C5,$18,$8F,$E0,$38,$7C,$1F,$AC,$F7,$12,$3E,$BC,$F2,$D3,$DD,$FB
	.dc.b	$FB,$DA,$FA,$0D,$AF,$9E,$D7,$C0,$D7,$BF,$03,$DF,$F9,$84,$8F,$C2
	.dc.b	$CF,$1B,$FE,$1F,$F0,$2F,$B9,$70,$FF,$03,$8C,$36,$33,$7F,$B9,$97
	.dc.b	$A3,$CF,$CC,$FE,$A1,$BF,$DF,$EB,$FB,$84,$BE,$18,$72,$7F,$1F,$83
	.dc.b	$EF,$97,$FC,$A2,$7F,$94,$CE,$63,$11,$FD,$76,$30,$E1,$AF,$E4,$36
	.dc.b	$FC,$EB,$F7,$9A,$FE,$7A,$7F,$29,$D5,$EF,$BF,$0C,$FE,$F3,$E2,$5E
	.dc.b	$AF,$F0,$E6,$C6,$63,$1D,$79,$F4,$FF,$A2,$51,$A4,$AF,$D3,$FC,$53
	.dc.b	$E3,$FF,$96,$B6,$3F,$72,$D7,$F6,$62,$BA,$3F,$C4,$3E,$5F,$F3,$DC
	.dc.b	$9F,$D1,$1B,$C5,$FD,$41,$B3,$FE,$03,$5E,$DF,$F2,$87,$F8,$2C,$7C
	.dc.b	$C0,$E1,$F9,$ED,$7E,$A3,$5F,$F0,$77,$8E,$FD,$9F,$F9,$EC,$7F,$0A
	.dc.b	$8C,$79,$DF,$D1,$27,$38,$FF,$5A,$FD,$AF,$F9,$16,$5F,$6F,$ED,$9F
	.dc.b	$59,$F9,$16,$E5,$F8,$F6,$BF,$88,$FC,$21,$F1,$9F,$D0,$BE,$7D,$F5
	.dc.b	$34,$95,$FC,$DB,$E2,$5F,$53,$BF,$47,$FB,$D5,$FF,$30,$AF,$84,$D7
	.dc.b	$F1,$ED,$7F,$27,$F5,$8F,$96,$FD,$33,$63,$F4,$8D,$7E,$CE,$0F,$89
	.dc.b	$FE,$DD,$E2,$3F,$2C,$7C,$9F,$D1,$8C,$36,$33,$7F,$A0,$9C,$B6,$1E
	.dc.b	$F6,$C9,$19,$7F,$C2,$D8,$BE,$87,$FB,$EC,$FE,$AA,$31,$E7,$FE,$05
	.dc.b	$B1,$99,$DB,$FB,$38,$FC,$2F,$8B,$BE,$CF,$D3,$CE,$5A,$FF,$88,$27
	.dc.b	$6F,$EE,$F2,$7E,$51,$F6,$3A,$3D,$8D,$9D,$35,$FF,$14,$CF,$0E,$5F
	.dc.b	$F1,$0C,$7F,$97,$63,$9F,$C0,$62,$3F,$89,$D2,$31,$3A,$BE,$F1,$BF
	.dc.b	$25,$8B,$E7,$7F,$C6,$3D,$BF,$C9,$23,$17,$FD,$FB,$EC,$5F,$47,$A8
	.dc.b	$FE,$25,$3A,$D3,$F6,$FF,$77,$5F,$E3,$4D,$7F,$C8,$6D,$7F,$CE,$CD
	.dc.b	$4B,$E8,$3B,$46,$9A,$BF,$66,$FC,$FD,$FF,$84,$AB,$83,$FA,$7F,$B2
	.dc.b	$E3,$AB,$C1,$46,$3C,$1F,$EF,$B5,$B3,$FC,$5D,$F5,$1F,$E8,$96,$E5
	.dc.b	$E0,$FD,$9F,$FB,$1C,$60,$FF,$10,$FD,$F7,$F4,$F7,$F8,$53,$67,$F9
	.dc.b	$BB,$5F,$E7,$8D,$7F,$CC,$AB,$DB,$F6,$B6,$7A,$FB,$D8,$9F,$87,$D8
	.dc.b	$FC,$97,$F3,$95,$FE,$20,$6E,$1F,$CA,$3F,$0C,$F5,$3F,$E0,$28,$CF
	.dc.b	$37,$F9,$AA,$73,$3F,$D3,$DA,$FF,$9E,$1A,$FF,$AA,$DA,$BF,$0C,$FE
	.dc.b	$17,$7F,$E4,$26,$C7,$F3,$87,$F8,$04,$67,$C8,$FB,$7D,$FF,$E2,$3F
	.dc.b	$86,$BE,$CF,$F8,$DA,$BF,$B6,$7B,$1F,$BC,$7F,$AC,$4F,$DA,$8C,$63
	.dc.b	$FD,$97,$9B,$FB,$47,$89,$F9,$B9,$CC,$FF,$68,$6B,$FD,$D3,$0F,$F1
	.dc.b	$46,$1F,$D8,$FE,$0B,$F0,$BE,$DF,$E0,$AB,$FB,$25,$69,$FE,$EB,$19
	.dc.b	$DF,$FC,$0E,$DB,$FE,$E5,$E1,$3E,$77,$9F,$C3,$7B,$3D,$69,$5F,$E1
	.dc.b	$0F,$2E,$E5,$FE,$34,$F8,$0F,$E4,$33,$FE,$1C,$9F,$DD,$73,$A7,$35
	.dc.b	$FE,$1E,$9F,$DF,$71,$FF,$92,$FC,$DF,$92,$7E,$13,$C7,$7C,$CE,$32
	.dc.b	$FF,$31,$C6,$1F,$F2,$BB,$70,$FF,$1C,$35,$FF,$5B,$E5,$FC,$0E,$BB
	.dc.b	$F0,$FD,$CE,$1F,$F5,$AE,$5F,$E5,$BE,$7F,$C9,$FD,$0F,$F8,$A3,$3F
	.dc.b	$F0,$74,$63,$E0,$A4,$7F,$9D,$3E,$4D,$F4,$98,$D7,$FC,$25,$ED,$7E
	.dc.b	$5B,$3E,$AD,$BF,$F5,$E7,$C5,$7C,$D7,$FC,$AF,$19,$AF,$F3,$B7,$37
	.dc.b	$C8,$AD,$47,$F9,$4D,$AF,$B1,$AF,$FB,$1F,$DA,$FF,$27,$5F,$FA,$21
	.dc.b	$B1,$FE,$68,$7D,$8D,$F8,$7F,$D8,$C6,$31,$FF,$3A,$36,$BA,$7C,$FE
	.dc.b	$B5,$FE,$33,$EF,$BE,$C7,$FD,$03,$5F,$EB,$B6,$E5,$FE,$07,$7F,$C5
	.dc.b	$F1,$9F,$85,$E3,$18,$9D,$7C,$36,$3E,$1B,$2F,$F2,$48,$7D,$15,$31
	.dc.b	$B5,$C7,$FD,$6F,$3E,$E3,$FC,$97,$1F,$D2,$51,$93,$FC,$A4,$E5,$F7
	.dc.b	$1F,$EF,$CC,$FF,$99,$FC,$53,$E1,$FC,$51,$F0,$3E,$2B,$FC,$57,$87
	.dc.b	$FB,$65,$B7,$FE,$25,$7F,$25,$C3,$FE,$C2,$72,$DE,$BE,$47,$3F,$66
	.dc.b	$DF,$FA,$FB,$5F,$89,$5F,$EA,$56,$C7,$AB,$B7,$FC,$B3,$C3,$FC,$A2
	.dc.b	$73,$97,$FB,$F9,$FA,$37,$C6,$75,$C1,$B7,$E6,$71,$CC,$7E,$4F,$65
	.dc.b	$C7,$FC,$AA,$7F,$9E,$23,$FC,$94,$D3,$FC,$99,$1F,$24,$F9,$4B,$E7
	.dc.b	$BE,$7F,$CE,$FF,$77,$8F,$F5,$3D,$63,$DD,$FF,$3F,$D5,$9C,$8A,$E0
	.dc.b	$FD,$83,$A1,$99,$FF,$68,$D7,$18,$D5,$57,$CD,$5C,$9F,$C0,$C6,$6D
	.dc.b	$5B,$9E,$B5,$B9,$CC,$77,$3F,$D3,$15,$C6,$B8,$36,$34,$78,$8A,$78
	.dc.b	$AF,$DE,$E2,$BA,$DF,$96,$3C,$33,$EC,$78,$98,$3E,$2C,$6E,$BC,$0D
	.dc.b	$8E,$97,$E4,$69,$CA,$77,$5E,$2A,$F7,$56,$DB,$5B,$38,$BD,$4C,$96
	.dc.b	$7C,$CD,$8F,$2C,$75,$32,$57,$27,$CA,$3A,$B9,$7F,$D9,$B3,$E9,$F9
	.dc.b	$EF,$19,$ED,$2C,$EE,$7E,$61,$CE,$BC,$B1,$9C,$3E,$6B,$6F,$A3,$F3
	.dc.b	$6F,$A3,$4D,$5F,$C1,$BE,$6F,$58,$F5,$B4,$F5,$9F,$7E,$7C,$AF,$B5
	.dc.b	$AF,$E0,$E7,$9B,$67,$9B,$4E,$67,$E5,$8F,$8A,$73,$7C,$CD,$3F,$4F
	.dc.b	$3F,$45,$F9,$27,$A6,$3A,$BC,$E7,$C2,$D8,$CB,$FA,$C9,$F3,$BE,$16
	.dc.b	$BF,$91,$3A,$CF,$77,$E8,$9E,$93,$33,$F2,$A3,$E6,$C7,$9A,$3F,$43
	.dc.b	$3D,$30,$FC,$9C,$FC,$B3,$89,$4E,$A7,$9C,$F8,$A7,$9E,$7E,$1B,$F4
	.dc.b	$5A,$EC,$E6,$79,$4E,$67,$AC,$B1,$E2,$79,$1E,$73,$E7,$1F,$32,$7E
	.dc.b	$06,$1F,$8C,$8F,$33,$5E,$CD,$7C,$47,$61,$1F,$05,$B5,$E8,$7C,$93
	.dc.b	$C5,$B8,$7E,$D6,$7E,$BD,$F9,$E7,$AD,$6E,$31,$1C,$0F,$C1,$1F,$21
	.dc.b	$C9,$1F,$C6,$9E,$8A,$FD,$2B,$F0,$4F,$86,$7C,$1A,$FB,$84,$7E,$0C
	.dc.b	$79,$5E,$C7,$F4,$69,$FE,$8E,$FF,$38,$3F,$9D,$1F,$C9,$4B,$F2,$CD
	.dc.b	$9E,$4E,$A7,$FB,$BC,$E6,$31,$08,$C3,$B6,$BD,$36,$6F,$C2,$51,$89
	.dc.b	$E0,$D7,$84,$EE,$35,$AF,$F5,$2A,$7C,$F7,$F5,$38,$C3,$74,$FE,$AD
	.dc.b	$3D,$AF,$FA,$A3,$70,$FE,$8A,$F2,$73,$5C,$5F,$7E,$1E,$0B,$E6,$F9
	.dc.b	$F6,$EF,$3F,$E1,$A6,$BF,$8A,$38,$E1,$9A,$E2,$F6,$88,$78,$AB,$B3
	.dc.b	$C0,$F2,$DF,$8D,$F9,$CD,$C3,$D5,$8F,$C3,$AE,$35,$A3,$AC,$66,$FE
	.dc.b	$0C,$EA,$BE,$05,$76,$33,$31,$C1,$2B,$93,$E7,$8E,$B8,$7C,$EC,$3C
	.dc.b	$98,$75,$3E,$34,$66,$BE,$11,$F8,$A9,$FF,$05,$63,$D0,$EB,$87,$CA
	.dc.b	$25,$7C,$8A,$F0,$1E,$37,$A3,$88,$E5,$18,$9F,$A6,$62,$7E,$64,$F4
	.dc.b	$7C,$47,$BD,$F0,$CF,$47,$51,$F9,$DB,$F6,$61,$EA,$8F,$9C,$7D,$F2
	.dc.b	$DF,$77,$84,$7C,$67,$B9,$E7,$BF,$65,$F6,$C7,$E9,$BE,$29,$FC,$15
	.dc.b	$FA,$AB,$D8,$7C,$37,$51,$E2,$9E,$58,$79,$A3,$E3,$1E,$08,$DD,$FE
	.dc.b	$86,$BA,$4F,$91,$F4,$47,$E3,$2F,$E1,$12,$7E,$4B,$E8,$30,$F8,$5C
	.dc.b	$3E,$87,$CE,$F5,$7D,$06,$9E,$07,$95,$C4,$78,$63,$04,$3E,$59,$F8
	.dc.b	$29,$F4,$BF,$42,$7A,$BE,$63,$84,$FC,$C8,$EB,$7F,$A1,$C7,$F5,$47
	.dc.b	$E6,$31,$FD,$06,$1D,$2F,$F4,$AF,$C4,$AE,$18,$7D,$B3,$C7,$8F,$CC
	.dc.b	$C3,$9C,$7E,$2F,$1F,$1B,$E5,$3E,$11,$F1,$4F,$31,$F0,$89,$87,$C1
	.dc.b	$7A,$C7,$CE,$8F,$54,$70,$74,$61,$E3,$7F,$81,$BF,$C6,$5F,$B6,$7F
	.dc.b	$20,$7C,$47,$11,$E9,$8F,$65,$74,$AF,$7C,$7E,$28,$F1,$9E,$1A,$F7
	.dc.b	$BD,$0C,$9E,$A7,$AD,$FD,$C7,$77,$F9,$2A,$EA,$FE,$63,$1F,$91,$7E
	.dc.b	$52,$9D,$30,$FE,$2F,$1F,$E8,$28,$FD,$8B,$63,$E1,$FC,$F7,$FA,$91
	.dc.b	$34,$93,$FD,$63,$A1,$DB,$FE,$9B,$BE,$91,$89,$FF,$6A,$C6,$27,$5C
	.dc.b	$1F,$D3,$2E,$D3,$91,$F3,$AF,$0F,$31,$86,$C6,$67,$AD,$FF,$B2,$61
	.dc.b	$F3,$BC,$3F,$32,$F1,$57,$F6,$8A,$FF,$67,$46,$33,$FF,$51,$E2,$8C
	.dc.b	$4F,$3C,$7C,$BE,$0B,$E7,$9F,$23,$87,$FD,$77,$8F,$FA,$B2,$75,$C9
	.dc.b	$9A,$D3,$CB,$E6,$1D,$D4,$CE,$BE,$A3,$FC,$0B,$F3,$DF,$11,$FF,$02
	.dc.b	$7C,$E5,$E5,$36,$71,$F3,$BF,$5C,$95,$FE,$05,$7F,$DB,$BC,$CF,$94
	.dc.b	$E8,$D9,$F0,$07,$0F,$C7,$3F,$C2,$F3,$F0,$BE,$75,$F3,$B7,$E5,$A5
	.dc.b	$7A,$78,$15,$A8,$C2,$9F,$97,$F3,$59,$9E,$D4,$8C,$71,$7C,$BB,$E1
	.dc.b	$5F,$17,$9D,$C2,$F1,$FF,$0A,$F1,$9D,$35,$F5,$4F,$09,$DE,$1F,$8C
	.dc.b	$FA,$A7,$95,$4C,$DF,$D3,$F1,$9E,$A3,$C9,$FF,$1B,$C6,$1B,$19,$9F
	.dc.b	$F1,$F4,$E5,$FF,$20,$64,$F8,$8F,$5C,$79,$1B,$B7,$E4,$9B,$DA,$95
	.dc.b	$F3,$7E,$A3,$E4,$F8,$6F,$D4,$B6,$3D,$18,$A7,$F5,$3F,$3C,$F7,$FF
	.dc.b	$57,$E0,$3C,$DA,$4C,$38,$DF,$4C,$1F,$0F,$EE,$24,$FB,$A3,$2F,$BB
	.dc.b	$F7,$4F,$C4,$ED,$FA,$C6,$EB,$EB,$DD,$18,$AF,$6D,$F2,$7D,$AF,$E4
	.dc.b	$DE,$DD,$7F,$38,$8F,$CC,$FD,$ED,$F2,$BF,$C1,$9B,$1F,$87,$6B,$D2
	.dc.b	$7D,$9C,$F2,$3E,$3B,$FC,$4B,$E1,$9F,$08,$F5,$71,$E9,$5B,$FD,$B5
	.dc.b	$7F,$9F,$27,$27,$38,$F4,$F5,$FE,$10,$8F,$C7,$7F,$CF,$F3,$F7,$3E
	.dc.b	$03,$C1,$D0,$69,$F4,$67,$F8,$17,$CE,$FF,$A2,$EF,$F6,$EF,$FA,$01
	.dc.b	$BE,$96,$BD,$5F,$E4,$FD,$CF,$9E,$FF,$46,$F2,$FE,$79,$B1,$9B,$FC
	.dc.b	$AB,$5F,$F4,$DC,$BF,$29,$F2,$47,$EC,$5A,$FE,$B1,$BD,$1F,$CA,$AB
	.dc.b	$93,$FD,$37,$E6,$7F,$21,$5F,$C3,$97,$C5,$FF,$1D,$6C,$67,$BF,$F9
	.dc.b	$07,$CF,$BE,$93,$FD,$4B,$28,$C3,$FE,$A8,$6B,$FE,$AD,$6B,$FA,$91
	.dc.b	$C6,$F9,$3F,$C1,$E4,$EE,$FD,$74,$E7,$B3,$F0,$55,$A9,$FD,$97,$92
	.dc.b	$F8,$4F,$E8,$38,$FE,$0B,$E5,$5F,$75,$F8,$96,$C6,$63,$1E,$E7,$F3
	.dc.b	$BF,$A1,$FD,$04,$9F,$4E,$BF,$DC,$DB,$97,$A5,$DF,$DA,$7C,$AA,$FF
	.dc.b	$65,$35,$FF,$68,$B5,$F8,$0D,$7E,$0F,$CE,$3E,$27,$F3,$CD,$8F,$CD
	.dc.b	$B5,$FE,$81,$C0,$87,$D1,$7F,$0E,$DB,$F8,$9D,$7F,$1E,$6B,$FE,$E2
	.dc.b	$F0,$3E,$31,$FC,$9A,$33,$1F,$4F,$AB,$F9,$7F,$C4,$7E,$48,$91,$F9
	.dc.b	$27,$EF,$E0,$FA,$0F,$F0,$24,$FF,$BA,$63,$15,$FE,$E8,$6C,$66,$76
	.dc.b	$FF,$BD,$21,$D3,$E2,$8F,$E5,$91,$8C,$79,$4E,$67,$E0,$3B,$B7,$CC
	.dc.b	$7F,$84,$F8,$3F,$9D,$B6,$BF,$4E,$D7,$DC,$D7,$F9,$D5,$FE,$27,$CE
	.dc.b	$FF,$18,$46,$34,$FE,$65,$7F,$2E,$73,$3F,$D0,$27,$2F,$E6,$CF,$B7
	.dc.b	$DF,$F9,$97,$EB,$E7,$FB,$FB,$FB,$7F,$21,$9E,$4D,$70,$FD,$BC,$31
	.dc.b	$5F,$A5,$FC,$C9,$F6,$BE,$0C,$91,$89,$F0,$47,$F5,$38,$C3,$86,$BE
	.dc.b	$5F,$9E,$7D,$07,$68,$FE,$B5,$1E,$57,$FC,$18,$1D,$D7,$F8,$49,$D6
	.dc.b	$5E,$8E,$B7,$DB,$F3,$D1,$18,$BF,$D9,$7D,$06,$63,$18,$FE,$6B,$BD
	.dc.b	$F3,$5C,$5B,$7F,$E1,$7F,$67,$FB,$73,$FC,$F2,$33,$5E,$6F,$A0,$7D
	.dc.b	$3F,$C5,$6D,$7C,$49,$46,$1F,$8E,$D7,$EC,$7D,$0F,$F6,$A9,$FF,$17
	.dc.b	$BF,$E2,$F6,$FB,$3F,$1E,$53,$97,$FC,$33,$F9,$5B,$E7,$3E,$67,$6A
	.dc.b	$31,$87,$F8,$97,$B7,$F5,$0F,$F8,$9E,$33,$7F,$E2,$29,$CB,$F9,$8E
	.dc.b	$FF,$C7,$DF,$E7,$9B,$3A,$9F,$C8,$7D,$0B,$E7,$3C,$DF,$45,$F2,$27
	.dc.b	$ED,$CE,$72,$FF,$25,$CE,$79,$FF,$23,$F8,$47,$CC,$6F,$27,$E1,$7C
	.dc.b	$2F,$F1,$6F,$A1,$F7,$6B,$58,$7E,$A3,$DE,$FE,$8E,$FF,$83,$E7,$F4
	.dc.b	$B5,$C2,$30,$7E,$5B,$A9,$ED,$3E,$13,$F1,$0D,$9F,$F1,$2F,$A7,$89
	.dc.b	$F1,$63,$FE,$25,$9C,$9F,$DF,$BD,$CF,$1F,$07,$F8,$5A,$FE,$8F,$BE
	.dc.b	$FB,$8A,$72,$FA,$2B,$87,$F8,$C9,$F3,$E2,$69,$23,$FC,$8B,$FB,$B3
	.dc.b	$EB,$FE,$EB,$63,$FC,$77,$F0,$18,$9F,$3B,$D0,$91,$FE,$33,$7E,$DF
	.dc.b	$C3,$7F,$38,$8F,$F3,$43,$74,$FF,$16,$F4,$7E,$2A,$31,$CD,$FC,$2F
	.dc.b	$8F,$FC,$BF,$3F,$10,$FC,$7F,$D1,$FF,$88,$F0,$FF,$26,$C7,$E4,$BC
	.dc.b	$47,$D5,$D3,$EE,$7A,$1F,$81,$AF,$F2,$04,$E9,$FF,$36,$47,$E2,$73
	.dc.b	$FE,$ED,$8F,$F6,$D7,$EE,$C6,$5F,$E4,$C7,$E5,$BD,$87,$D0,$FE,$8B
	.dc.b	$9F,$FC,$0D,$87,$FA,$16,$72,$FF,$A1,$E3,$35,$FE,$82,$9F,$53,$6B
	.dc.b	$F5,$71,$8D,$7F,$D8,$6F,$29,$FF,$4A,$FD,$17,$FA,$07,$0F,$F4,$F7
	.dc.b	$C2,$FF,$A6,$DF,$F5,$04,$66,$7F,$D3,$73,$97,$8F,$C1,$FF,$23,$F7
	.dc.b	$95,$BC,$7F,$D9,$0A,$74,$E7,$0F,$F0,$D7,$C4,$7E,$E2,$BF,$DD,$BE
	.dc.b	$2F,$F0,$7C,$5E,$D7,$F0,$9A,$D5,$FF,$47,$8F,$E7,$93,$A8,$FF,$15
	.dc.b	$C6,$0C,$FC,$3F,$3D,$BA,$36,$7F,$D7,$9F,$11,$FD,$A9,$FF,$5F,$7C
	.dc.b	$6F,$F4,$2C,$FF,$DC,$13,$9C,$7F,$D7,$73,$9A,$FF,$14,$CF,$FB,$81
	.dc.b	$E7,$EE,$FC,$38,$C5,$7F,$B0,$9E,$57,$FE,$F4,$F9,$AF,$E7,$6D,$F5
	.dc.b	$BF,$90,$F0,$7C,$7C,$7F,$D8,$95,$A7,$FD,$1D,$19,$C3,$FD,$82,$DB
	.dc.b	$FF,$7F,$BF,$36,$1E,$6F,$D1,$8F,$95,$94,$BF,$F7,$2D,$7F,$72,$BF
	.dc.b	$EE,$0F,$FB,$9E,$7E,$97,$BD,$F3,$89,$5F,$E5,$87,$E8,$CB,$E8,$7F
	.dc.b	$D7,$9F,$3E,$F9,$AF,$F1,$45,$F6,$AD,$09,$D3,$FA,$68,$D5,$72,$AE
	.dc.b	$33,$B7,$FB,$BC,$66,$38,$19,$3F,$C1,$4E,$5F,$F0,$AC,$D7,$FD,$A1
	.dc.b	$17,$AF,$FC,$1D,$7E,$29,$FF,$36,$4E,$6B,$84,$7A,$27,$4F,$FB,$62
	.dc.b	$33,$3B,$8F,$11,$C8,$A6,$CA,$73,$BF,$6D,$74,$8C,$1E,$C7,$13,$D5
	.dc.b	$B1,$D2,$FD,$35,$F1,$23,$17,$E4,$69,$E0,$BE,$53,$BB,$F5,$DF,$2A
	.dc.b	$F9,$27,$08,$EE,$6A,$FA,$36,$3A,$1B,$AE,$87,$B0,$AF,$AD,$AF,$D4
	.dc.b	$38,$C7,$33,$DB,$F4,$3D,$4B,$3F,$82,$7C,$6F,$59,$F3,$CF,$18,$EB
	.dc.b	$1F,$3E,$36,$F8,$4F,$6C,$F5,$9F,$90,$D8,$E7,$1D,$5F,$B5,$3D,$5C
	.dc.b	$9C,$CC,$C6,$DF,$C6,$89,$E8,$FE,$0D,$AE,$A7,$A3,$C8,$E8,$66,$7A
	.dc.b	$3F,$84,$3E,$44,$7C,$47,$F4,$33,$5F,$98,$FC,$07,$73,$F1,$2F,$EF
	.dc.b	$47,$73,$27,$B0,$A7,$9E,$3E,$81,$F1,$8A,$EC,$F9,$CF,$5B,$FC,$2B
	.dc.b	$E8,$C3,$F1,$F3,$F4,$23,$8B,$D4,$F4,$BC,$5F,$C1,$BC,$5F,$C6,$1D
	.dc.b	$EB,$D2,$FC,$EB,$F9,$6F,$F0,$C6,$D7,$82,$3D,$C7,$98,$E5,$7E,$67
	.dc.b	$E7,$9C,$6F,$CA,$F5,$6B,$F4,$4E,$47,$4B,$F1,$B9,$6D,$78,$63,$91
	.dc.b	$EE,$3E,$11,$4E,$55,$E3,$7E,$89,$F8,$13,$CB,$1C,$DE,$D1,$B7,$C6
	.dc.b	$6C,$E2,$68,$F9,$67,$23,$E5,$9E,$03,$27,$08,$3F,$99,$AF,$83,$18
	.dc.b	$BF,$9B,$3C,$8F,$01,$E2,$3C,$18,$7E,$45,$C4,$FF,$A2,$DF,$D2,$9F
	.dc.b	$E8,$D7,$0F,$F4,$43,$FD,$85,$F3,$5F,$E0,$7F,$33,$FD,$24,$D8,$FF
	.dc.b	$5B,$B5,$DE,$DF,$E6,$47,$3F,$1C,$DF,$8B,$63,$31,$82,$3E,$F3,$E0
	.dc.b	$D3,$E8,$7E,$1C,$64,$46,$5F,$89,$A6,$A1,$7D,$F4,$FE,$99,$18,$C3
	.dc.b	$77,$DB,$0E,$B5,$A6,$E1,$CF,$0E,$4D,$BE,$37,$E2,$AD,$BC,$CC,$E1
	.dc.b	$DD,$F4,$37,$0E,$D1,$A9,$E2,$FB,$30,$EE,$FC,$26,$BD,$CE,$07,$67
	.dc.b	$35,$D7,$0F,$85,$5D,$23,$E3,$57,$8A,$B6,$70,$9D,$1B,$73,$87,$69
	.dc.b	$F4,$CE,$63,$C0,$67,$0F,$5C,$F5,$8C,$BC,$0D,$61,$F2,$5D,$19,$24
	.dc.b	$FD,$6B,$F0,$C6,$27,$D6,$F4,$9D,$57,$48,$FA,$07,$48,$CB,$C4,$CB
	.dc.b	$82,$4F,$48,$D1,$F4,$67,$B4,$66,$7D,$A7,$09,$FA,$93,$E1,$7E,$A1
	.dc.b	$E1,$9F,$74,$F0,$79,$18,$24,$79,$1C,$BD,$0E,$46,$48,$79,$4C,$9E
	.dc.b	$08,$EC,$F1,$7C,$CE,$63,$AB,$C6,$33,$87,$E1,$1E,$31,$99,$F2,$BC
	.dc.b	$4F,$0C,$FE,$35,$ED,$1D,$23,$93,$C1,$D4,$71,$23,$82,$61,$F9,$08
	.dc.b	$F9,$E7,$43,$BB,$F8,$F7,$DF,$1D,$9F,$1E,$1F,$86,$BF,$4C,$E6,$7D
	.dc.b	$47,$33,$C3,$5C,$E7,$E0,$BE,$18,$EB,$5F,$A3,$7E,$E5,$79,$E3,$11
	.dc.b	$F8,$68,$F4,$BF,$C9,$27,$F9,$64,$7E,$BA,$3F,$94,$C7,$F1,$87,$F8
	.dc.b	$CF,$0F,$F1,$4F,$94,$C4,$7F,$3E,$8C,$38,$6B,$C5,$AF,$2B,$FF,$5D
	.dc.b	$3B,$6B,$B6,$BB,$CF,$BB,$FE,$92,$FA,$07,$D7,$62,$D1,$89,$F1,$35
	.dc.b	$F1,$35,$F5,$AF,$A1,$E4,$D8,$E3,$18,$3C,$3E,$03,$E6,$35,$3F,$16
	.dc.b	$31,$3A,$9C,$CF,$F4,$09,$CB,$F2,$17,$D0,$BD,$DE,$AD,$9D,$35,$EC
	.dc.b	$D7,$A2,$F9,$1E,$24,$8E,$2F,$4F,$A8,$91,$D1,$AF,$4F,$A2,$F0,$91
	.dc.b	$F0,$BF,$37,$F4,$97,$C6,$D2,$4F,$DF,$8C,$B7,$E1,$B8,$11,$C1,$F6
	.dc.b	$BE,$BF,$FB,$B9,$1F,$EF,$8F,$F7,$AE,$67,$D1,$7E,$50,$93,$FD,$E9
	.dc.b	$FC,$BF,$D1,$C8,$FE,$FA,$D7,$EF,$FC,$67,$DC,$23,$F8,$47,$EF,$F8
	.dc.b	$B2,$3F,$0E,$D7,$EE,$71,$3E,$67,$EE,$12,$3E,$D4,$65,$BF,$43,$F8
	.dc.b	$12,$38,$37,$F4,$9F,$85,$23,$83,$9F,$D3,$7E,$24,$8E,$0E,$FE,$2F
	.dc.b	$E3,$48,$E0,$FD,$87,$96,$F8,$5F,$D2,$36,$33,$18,$33,$E6,$48,$E2
	.dc.b	$D7,$F7,$1F,$B3,$BE,$5B,$F5,$CD,$8F,$DF,$35,$DC,$6B,$E5,$24,$69
	.dc.b	$AF,$F0,$3F,$69,$EA,$FE,$E5,$B1,$FC,$25,$AE,$E3,$5E,$84,$8D,$35
	.dc.b	$FE,$1D,$FB,$8B,$E1,$FF,$82,$36,$3F,$8A,$35,$DC,$6B,$49,$1A,$6B
	.dc.b	$FC,$67,$E8,$9F,$6F,$FC,$3D,$B1,$FC,$7D,$AE,$E3,$5F,$33,$23,$4D
	.dc.b	$7F,$92,$7C,$53,$D9,$FE,$34,$D8,$FE,$54,$D7,$71,$AF,$66,$46,$9A
	.dc.b	$FF,$37,$F8,$CF,$F3,$8F,$C4,$FF,$28,$6C,$7F,$38,$6B,$B1,$1B,$7E
	.dc.b	$31,$FE,$58,$8F,$F3,$DE,$99,$24,$73,$F6,$5C,$BD,$DB,$19,$8C,$19
	.dc.b	$F7,$64,$71,$6B,$FD,$33,$87,$C6,$FF,$40,$6B,$9C,$7F,$D0,$51,$88
	.dc.b	$FE,$AC,$FA,$1F,$F3,$7B,$5F,$F4,$9B,$53,$4F,$D3,$3E,$1E,$0E,$A4
	.dc.b	$7F,$A8,$FC,$F3,$EA,$3F,$D2,$4D,$8C,$C6,$0C,$F7,$C8,$E2,$D7,$FB
	.dc.b	$4F,$CF,$F8,$C9,$AF,$FA,$75,$F3,$7C,$33,$E2,$BF,$D4,$0D,$AF,$EC
	.dc.b	$6D,$7F,$B9,$B5,$FE,$F1,$F6,$CE,$0F,$F5,$43,$63,$FB,$5B,$5F,$EF
	.dc.b	$6D,$7F,$C0,$1F,$80,$3B,$3F,$D6,$0D,$8F,$EE,$6D,$7F,$C0,$6D,$7C
	.dc.b	$27,$80,$87,$80,$FF,$5D,$7E,$14,$F9,$4A,$48,$FE,$FF,$19,$6F,$A5
	.dc.b	$C0,$8E,$0F,$F0,$2F,$D0,$7F,$0C,$F6,$6F,$80,$69,$27,$FC,$33,$3F
	.dc.b	$82,$D3,$8F,$2F,$C1,$9F,$F6,$CC,$E5,$DC,$65,$7C,$EF,$F8,$4B,$C6
	.dc.b	$FF,$0A,$B6,$3F,$C4,$AD,$7F,$C5,$4D,$7F,$D8,$6F,$F8,$46,$38,$FB
	.dc.b	$26,$92,$7F,$C5,$FF,$3A,$F5,$F5,$18,$6B,$FE,$37,$8D,$4F,$F8,$A1
	.dc.b	$B1,$FE,$38,$FC,$33,$F3,$53,$FE,$2C,$6C,$67,$A7,$9B,$FE,$D8,$8C
	.dc.b	$BF,$E4,$3F,$B2,$F4,$3F,$E4,$26,$B9,$C7,$8A,$11,$A7,$FD,$C7,$F6
	.dc.b	$6F,$37,$70,$8C,$63,$FE,$40,$6C,$7F,$94,$1A,$FF,$95,$7E,$8D,$E3
	.dc.b	$F2,$6D,$7F,$92,$1B,$3F,$E5,$86,$BF,$E5,$DF,$A3,$FF,$0E,$CF,$F9
	.dc.b	$39,$B1,$FE,$5E,$6B,$BF,$6F,$CF,$1F,$A5,$9E,$1D,$58,$23,$FE,$58
	.dc.b	$FA,$17,$8B,$EE,$6C,$66,$30,$67,$E8,$A4,$71,$6E,$3F,$E7,$0A,$7D
	.dc.b	$5F,$32,$57,$77,$E8,$F9,$2F,$9F,$7D,$E6,$1B,$5F,$D0,$DB,$87,$F4
	.dc.b	$96,$BF,$D0,$57,$D3,$71,$24,$F1,$7B,$FC,$77,$E6,$63,$FA,$73,$6B
	.dc.b	$FA,$A3,$5F,$E9,$8D,$7F,$0D,$83,$E3,$FF,$C2,$15,$FD,$5D,$B3,$FD
	.dc.b	$75,$AF,$F5,$56,$BF,$E9,$3F,$DF,$3E,$EB,$78,$FF,$27,$E0,$27,$33
	.dc.b	$FC,$15,$B1,$9E,$ED,$C7,$F4,$6C,$3F,$D3,$8D,$AF,$F4,$E3,$5F,$F5
	.dc.b	$2F,$CF,$BE,$37,$BB,$6B,$FD,$48,$D9,$FF,$52,$35,$E3,$F6,$3C,$09
	.dc.b	$3C,$1F,$4F,$CF,$BE,$8A,$92,$7F,$D6,$11,$96,$F4,$F0,$23,$83,$FD
	.dc.b	$6D,$FB,$7B,$E8,$BE,$2B,$63,$31,$83,$3B,$48,$E2,$DC,$BF,$D7,$DF
	.dc.b	$B5,$FF,$89,$A7,$FD,$84,$D8,$CF,$B7,$F4,$A3,$11,$FE,$CC,$7E,$89
	.dc.b	$38,$74,$7D,$8F,$9F,$7B,$54,$98,$7F,$B3,$E3,$2D,$F9,$BE,$04,$70
	.dc.b	$7F,$AC,$7E,$7B,$FB,$63,$7D,$5F,$E2,$FF,$26,$9C,$E7,$FE,$BD,$6B
	.dc.b	$FE,$BB,$7F,$C9,$D5,$CF,$D8,$E3,$18,$9E,$11,$8F,$6E,$57,$85,$B3
	.dc.b	$D7,$E5,$B8,$92,$7C,$CF,$4E,$89,$1E,$96,$CF,$FB,$B6,$7F,$C6,$33
	.dc.b	$FE,$F1,$6C,$67,$DE,$FD,$0C,$62,$3F,$DF,$6B,$C4,$C9,$2B,$93,$FC
	.dc.b	$D2,$7F,$CA,$D1,$FE,$F2,$69,$FE,$F4,$F7,$F4,$6F,$82,$4E,$DB,$3F
	.dc.b	$D0,$C9,$E3,$3E,$BB,$FA,$0B,$6B,$FD,$FE,$D9,$E0,$7F,$4B,$70,$F3
	.dc.b	$E6,$7E,$9F,$FB,$A2,$72,$FF,$00,$71,$EB,$D2,$61,$D2,$FF,$C0,$D3
	.dc.b	$FE,$94,$8F,$F7,$B6,$CF,$89,$FC,$2B,$63,$31,$83,$3E,$7C,$8E,$2D
	.dc.b	$BF,$F7,$CF,$7F,$E3,$BE,$57,$F6,$D1,$8A,$FF,$7F,$CE,$7E,$22,$47
	.dc.b	$97,$A1,$7F,$C7,$AD,$18,$AF,$CC,$FC,$67,$F8,$36,$72,$D9,$34,$D6
	.dc.b	$EE,$57,$CC,$FE,$18,$99,$7E,$2D,$FC,$4F,$3C,$8F,$C7,$37,$1F,$AB
	.dc.b	$6F,$B9,$FF,$00,$57,$CD,$8C,$4F,$F3,$29,$CB,$C1,$AF,$DB,$AF,$D7
	.dc.b	$BF,$E0,$99,$FE,$59,$D9,$FC,$EA,$73,$3F,$E3,$D6,$BF,$E0,$FF,$77
	.dc.b	$F3,$04,$C3,$F3,$0F,$E6,$E1,$F3,$DF,$E1,$5F,$9D,$7C,$F7,$E3,$AB
	.dc.b	$FC,$5B,$18,$AF,$C8,$D6,$EB,$57,$67,$22,$B8,$3E,$38,$D3,$FA,$1A
	.dc.b	$AF,$F9,$D6,$2B,$E4,$39,$57,$66,$9D,$AB,$AB,$C4,$C9,$D4,$A6,$5D
	.dc.b	$BD,$5E,$37,$EE,$9F,$D1,$5F,$92,$31,$5D,$E7,$C5,$5C,$AB,$57,$E8
	.dc.b	$6D,$78,$0E,$D5,$C0,$EA,$F9,$2F,$AB,$E7,$BF,$1D,$F3,$8D,$36,$33
	.dc.b	$1D,$6F,$BC,$78,$E3,$66,$4A,$E4,$E2,$5B,$F2,$47,$B4,$FE,$29,$7E
	.dc.b	$47,$FC,$B1,$5D,$A7,$D3,$5C,$9A,$72,$9F,$53,$5E,$27,$89,$DB,$DA
	.dc.b	$FD,$91,$B8,$F8,$E6,$4F,$54,$F7,$7C,$C7,$E2,$CC,$C7,$1A,$F3,$46
	.dc.b	$8F,$C0,$47,$81,$E1,$3F,$05,$EC,$D7,$31,$B7,$89,$CE,$FE,$23,$E6
	.dc.b	$36,$66,$FD,$6F,$49,$E5,$3C,$2F,$CB,$3F,$18,$D1,$C4,$F5,$98,$34
	.dc.b	$76,$29,$E4,$2C,$75,$7E,$71,$CE,$FD,$8E,$CF,$A6,$FC,$B7,$8D,$FA
	.dc.b	$1E,$C6,$AF,$E6,$3D,$0D,$9E,$63,$67,$C3,$3B,$16,$BC,$95,$DE,$3A
	.dc.b	$1E,$D2,$9C,$AF,$E2,$BF,$88,$3F,$AA,$1F,$C0,$CA,$FF,$71,$70,$7F
	.dc.b	$6B,$F9,$6F,$E7,$5E,$77,$06,$C7,$F5,$D6,$BF,$DA,$9A,$F1,$C9,$E8
	.dc.b	$71,$6C,$71,$6B,$C5,$AF,$18,$FF,$5F,$7C,$FF,$F2,$0A,$7C,$FF,$F7
	.dc.b	$F8,$C3,$74,$E1,$69,$E1,$87,$81,$DC,$74,$9F,$75,$71,$8F,$F6,$65
	.dc.b	$FF,$9B,$F0,$D4,$63,$0F,$4E,$7C,$A3,$F9,$2D,$79,$1F,$E6,$6F,$17
	.dc.b	$33,$E1,$C3,$BC,$74,$13,$C5,$F4,$47,$AB,$1F,$4E,$1B,$AD,$47,$84
	.dc.b	$CD,$70,$3C,$D3,$BC,$7D,$91,$D9,$B5,$E5,$AE,$27,$3B,$F2,$D7,$33
	.dc.b	$CB,$87,$A5,$EA,$66,$BC,$73,$C2,$B9,$BE,$93,$04,$9E,$31,$E8,$C7
	.dc.b	$E5,$12,$7E,$CB,$CA,$7E,$9E,$1E,$83,$D3,$87,$8E,$3C,$B8,$FB,$D8
	.dc.b	$79,$63,$58,$FD,$3B,$ED,$3D,$0F,$AF,$3F,$73,$0F,$44,$64,$F7,$61
	.dc.b	$E0,$C3,$E9,$57,$48,$F2,$4E,$26,$3F,$84,$8F,$F2,$A5,$F8,$23,$CE
	.dc.b	$7E,$12,$7C,$F8,$FD,$18,$F8,$B8,$78,$9F,$B0,$69,$C6,$3E,$B7,$EB
	.dc.b	$DF,$58,$C4,$F2,$3D,$55,$D5,$F0,$BF,$35,$EA,$7B,$8F,$4B,$F5,$6B
	.dc.b	$A9,$23,$F0,$35,$C6,$BF,$03,$1D,$A7,$CB,$8F,$C5,$8F,$41,$E3,$3E
	.dc.b	$0D,$79,$F1,$F8,$B5,$F8,$03,$DA,$FC,$C9,$FB,$58,$F9,$27,$F1,$51
	.dc.b	$E8,$C7,$C5,$87,$D9,$8F,$B1,$87,$C3,$8C,$47,$C0,$AE,$38,$FA,$23
	.dc.b	$07,$95,$C1,$27,$F1,$27,$27,$F6,$D5,$D7,$0F,$C4,$9E,$EA,$F2,$5F
	.dc.b	$C2,$8C,$57,$37,$F2,$B3,$FC,$9F,$1F,$F5,$04,$FF,$01,$8C,$1F,$CB
	.dc.b	$FC,$EF,$F4,$E5,$36,$FE,$69,$18,$6C,$8F,$E5,$EF,$EE,$A7,$84,$7F
	.dc.b	$96,$F8,$F8,$34,$E1,$1A,$E1,$69,$FE,$79,$9C,$DF,$F9,$9E,$3F,$CF
	.dc.b	$71,$83,$FC,$CC,$F4,$9E,$2D,$39,$C6,$E3,$AD,$EF,$E7,$9E,$F5,$25
	.dc.b	$6E,$1F,$3F,$BF,$46,$4E,$5D,$BF,$00,$FF,$4C,$F5,$BE,$7F,$DF,$D7
	.dc.b	$FE,$9E,$6C,$66,$7D,$53,$FE,$9A,$C7,$3F,$0F,$27,$34,$8E,$37,$CB
	.dc.b	$E2,$65,$69,$E0,$F8,$F0,$FF,$56,$FC,$FF,$80,$91,$D3,$E1,$FC,$7C
	.dc.b	$D2,$3B,$BA,$7D,$9C,$0F,$9F,$7C,$FF,$CF,$6C,$FC,$D6,$BC,$17,$B1
	.dc.b	$BF,$82,$93,$B7,$95,$7E,$12,$3F,$A9,$FC,$F5,$EA,$34,$D7,$CF,$FF
	.dc.b	$BA,$3A,$BF,$1B,$5F,$D6,$BE,$7F,$F3,$64,$7F,$C1,$0F,$E7,$BE,$D7
	.dc.b	$A1,$23,$F1,$47,$13,$F0,$1F,$3F,$F8,$52,$3F,$83,$FA,$0F,$C4,$7B
	.dc.b	$72,$3F,$0C,$E9,$FA,$1E,$53,$E8,$3F,$02,$D8,$FB,$ED,$77,$F3,$1A
	.dc.b	$FB,$59,$1A,$78,$D7,$2F,$42,$4F,$27,$AF,$A3,$DB,$C5,$FF,$16,$F5
	.dc.b	$FF,$B9,$1B,$97,$C6,$BF,$DB,$1F,$EE,$A8,$DF,$EC,$4F,$4B,$F2,$ED
	.dc.b	$9F,$CA,$B5,$F6,$CF,$EB,$FE,$89,$EA,$52,$4F,$EB,$E1,$EE,$EF,$E9
	.dc.b	$64,$E5,$DB,$F8,$EF,$A3,$76,$7E,$ED,$B1,$FB,$96,$BB,$F5,$F5,$DD
	.dc.b	$23,$4F,$1A,$FE,$09,$F4,$6F,$8A,$FF,$28,$F8,$48,$C5,$FF,$BE,$DA
	.dc.b	$FF,$C0,$06,$BF,$22,$38,$71,$F4,$24,$7F,$0D,$9E,$3D,$BF,$C6,$FE
	.dc.b	$8F,$E0,$CF,$F1,$16,$BF,$AC,$6C,$6F,$E0,$FF,$8D,$43,$6A,$4A,$E5
	.dc.b	$BF,$F1,$FF,$A3,$7B,$FE,$C6,$CF,$DE,$3F,$9A,$68,$8C,$C6,$63,$13
	.dc.b	$C7,$DF,$E8,$49,$FE,$51,$5C,$33,$FE,$6D,$F4,$6F,$83,$F9,$6D,$AF
	.dc.b	$00,$C5,$19,$8F,$03,$5F,$98,$BE,$1F,$F9,$D3,$63,$F9,$C3,$5F,$E2
	.dc.b	$6F,$3B,$FE,$9B,$F1,$DF,$D4,$77,$FE,$9A,$DA,$FE,$8C,$D7,$F9,$B3
	.dc.b	$FD,$5F,$E8,$F7,$1F,$4B,$AB,$FA,$D4,$FF,$5B,$DF,$26,$AF,$AF,$C0
	.dc.b	$3E,$17,$FD,$37,$48,$C6,$1F,$0A,$31,$3A,$9C,$CF,$D5,$9E,$1F,$11
	.dc.b	$D0,$91,$E8,$9D,$E7,$FC,$7B,$9D,$8F,$F2,$76,$D7,$F1,$A6,$BF,$D4
	.dc.b	$CF,$D0,$69,$BC,$FF,$D6,$31,$FA,$36,$CF,$E5,$23,$F1,$0D,$7F,$83
	.dc.b	$FD,$9B,$C6,$FE,$6C,$D8,$FE,$52,$D7,$C1,$3F,$2D,$FF,$02,$7D,$26
	.dc.b	$72,$FE,$6E,$DD,$FF,$89,$B6,$77,$F3,$9F,$D0,$E1,$E4,$7A,$1B,$3F
	.dc.b	$EC,$A6,$BF,$90,$6B,$FA,$1F,$DC,$1E,$77,$F4,$D6,$C7,$F4,$06,$BF
	.dc.b	$1F,$E9,$7E,$4E,$5F,$8C,$FA,$17,$CB,$7E,$31,$B5,$F8,$C6,$BF,$3F
	.dc.b	$E9,$5F,$17,$FD,$6F,$E9,$3F,$B0,$CF,$F4,$E6,$BF,$55,$B8,$7F,$5A
	.dc.b	$FA,$5F,$C5,$4E,$FD,$9F,$AF,$5F,$DB,$23,$F7,$8D,$7F,$9D,$B5,$75
	.dc.b	$32,$7D,$CF,$A4,$93,$FD,$83,$0F,$ED,$5C,$7F,$E3,$9F,$8D,$7D,$E3
	.dc.b	$DF,$FF,$7B,$C6,$1B,$19,$BF,$DF,$CE,$5B,$0F,$A9,$D9,$23,$2F,$E4
	.dc.b	$B2,$78,$DF,$C1,$E7,$F8,$4C,$63,$C4,$FE,$17,$1E,$B9,$D3,$FC,$76
	.dc.b	$74,$F0,$EA,$F9,$2E,$5A,$FF,$60,$8F,$E8,$5E,$07,$46,$C7,$F2,$7C
	.dc.b	$9F,$23,$F2,$35,$F6,$B6,$74,$D7,$FB,$5C,$F0,$DF,$FB,$74,$23,$9F
	.dc.b	$C2,$E1,$B7,$D2,$31,$7F,$E3,$D8,$FE,$11,$F8,$97,$F9,$BD,$FE,$32
	.dc.b	$DB,$FD,$E7,$D3,$FD,$59,$C9,$FE,$6E,$E6,$FF,$11,$FD,$3B,$F0,$CF
	.dc.b	$1D,$A3,$15,$F3,$1B,$1F,$CE,$1A,$FF,$49,$7F,$80,$47,$F8,$99,$B1
	.dc.b	$FC,$37,$EA,$3F,$CD,$33,$9F,$13,$C4,$FF,$2A,$DD,$EA,$E2,$FF,$2B
	.dc.b	$18,$9F,$F1,$8F,$D4,$3D,$ED,$75,$FE,$67,$0E,$18,$7F,$20,$E5,$BE
	.dc.b	$57,$F7,$BC,$7F,$C7,$3F,$51,$FD,$89,$FF,$1D,$C6,$6B,$F9,$17,$D4
	.dc.b	$77,$7F,$AA,$B5,$FE,$CC,$D7,$FC,$4A,$D7,$FD,$38,$FF,$96,$A7,$FC
	.dc.b	$8E,$D8,$EB,$EC,$79,$A3,$1C,$5F,$E5,$4E,$3F,$F3,$3D,$FF,$93,$9B
	.dc.b	$7F,$CD,$7E,$A3,$FA,$54,$E7,$C2,$F7,$C6,$27,$FB,$5B,$5F,$EF,$8D
	.dc.b	$7F,$C7,$6D,$7F,$D6,$4B,$CF,$E6,$74,$FA,$8B,$7F,$D9,$3E,$9D,$F5
	.dc.b	$9F,$41,$B1,$9F,$A9,$C4,$E5,$FC,$EF,$1A,$91,$86,$D7,$F6,$16,$E5
	.dc.b	$FD,$7F,$BB,$FC,$9D,$3F,$DD,$1B,$1F,$DD,$36,$FC,$F1,$FD,$F2,$32
	.dc.b	$FF,$7E,$33,$3F,$DD,$9B,$3F,$C8,$27,$EF,$46,$E7,$F2,$1E,$13,$EA
	.dc.b	$7F,$87,$C7,$FB,$0D,$B7,$FE,$9E,$7F,$14,$DF,$0F,$8C,$62,$3F,$C8
	.dc.b	$4D,$7F,$C5,$AD,$7F,$C4,$0D,$A7,$D4,$BE,$77,$C6,$DC,$3F,$DA,$8F
	.dc.b	$FB,$81,$B1,$AF,$9C,$FE,$17,$5C,$9B,$5B,$C3,$FC,$74,$D7,$FC,$AD
	.dc.b	$F3,$3F,$0B,$0F,$F2,$0B,$63,$FC,$68,$FC,$C8,$CF,$5F,$C6,$8C,$4E
	.dc.b	$BE,$23,$1F,$09,$97,$E5,$FD,$4B,$E7,$E9,$8F,$A9,$B9,$7F,$BB,$E5
	.dc.b	$1E,$A6,$C7,$F9,$1B,$54,$63,$1F,$F2,$A3,$5F,$F3,$87,$D4,$3E,$9B
	.dc.b	$ED,$B6,$FF,$DF,$8B,$EA,$73,$3A,$64,$EB,$E1,$AD,$FF,$B7,$3E,$49
	.dc.b	$F5,$1F,$86,$6D,$27,$5F,$0F,$9A,$D3,$27,$5D,$07,$FC,$DF,$7F,$E5
	.dc.b	$B7,$26,$27,$F9,$17,$C9,$3E,$A5,$F0,$D8,$6D,$70,$C9,$5B,$F7,$35
	.dc.b	$7B,$7F,$A7,$C6,$6B,$FC,$E2,$DA,$FF,$41,$35,$FF,$60,$BE,$1F,$27
	.dc.b	$C0,$78,$7E,$1A,$D7,$86,$7C,$DF,$51,$E5,$3C,$DF,$4B,$67,$59,$A3
	.dc.b	$31,$8E,$DB,$93,$7F,$8C,$FF,$A1,$6B,$F8,$F4,$63,$9B,$97,$C4,$FE
	.dc.b	$E2,$7F,$D9,$B8,$7C,$39,$CB,$63,$37,$D6,$3F,$D3,$F1,$8B,$E8,$FF
	.dc.b	$23,$8E,$1F,$27,$FD,$32,$BF,$C2,$D1,$F3,$DF,$C0,$7B,$B9,$E4,$D3
	.dc.b	$67,$FC,$D3,$1F,$DB,$27,$F0,$E7,$F8,$59,$C7,$1F,$53,$FA,$CE,$7F
	.dc.b	$97,$9F,$F3,$A1,$FE,$A9,$AF,$67,$D4,$FE,$65,$BC,$FF,$C6,$23,$F8
	.dc.b	$9E,$9F,$15,$D5,$7B,$CF,$F5,$73,$AA,$FF,$5F,$B6,$33,$E2,$73,$E8
	.dc.b	$FC,$8E,$BF,$D9,$70,$FA,$7C,$7F,$EF,$AF,$4B,$3F,$25,$A6,$E7,$F8
	.dc.b	$3B,$4F,$2F,$53,$39,$F8,$F0,$FA,$F7,$B3,$F3,$FF,$47,$FC,$32,$33
	.dc.b	$3B,$8C,$13,$C5,$DF,$87,$C1,$97,$F9,$3F,$97,$7D,$2E,$30,$F8,$5D
	.dc.b	$1F,$91,$E8,$FC,$41,$F8,$9E,$BB,$87,$E1,$DC,$3F,$89,$EA,$B7,$C9
	.dc.b	$7D,$77,$F4,$86,$CF,$E4,$7E,$03,$F8,$23,$F9,$38,$CD,$78,$A7,$2D
	.dc.b	$39,$78,$1F,$D4,$A3,$3C,$BF,$A2,$9F,$E1,$DF,$25,$F9,$33,$F2,$9F
	.dc.b	$09,$71,$FC,$A4,$FF,$B3,$5F,$F5,$CC,$7E,$44,$8F,$FA,$F2,$1F,$55
	.dc.b	$FC,$67,$EA,$9F,$49,$FC,$6E,$7F,$8E,$5F,$EC,$A7,$31,$93,$88,$72
	.dc.b	$F0,$BE,$95,$FE,$FE,$BB,$3D,$6F,$A4,$E6,$39,$3D,$63,$31,$DE,$7F
	.dc.b	$97,$79,$9D,$A3,$DC,$7F,$4D,$AE,$51,$9A,$FE,$12,$EE,$7C,$F8,$67
	.dc.b	$0F,$1C,$7B,$F0,$AF,$49,$E1,$18,$BF,$1C,$EC,$F3,$9E,$BA,$E2,$FC
	.dc.b	$4B,$EC,$F9,$4F,$51,$4F,$8C,$53,$FA,$55,$77,$9E,$6D,$8E,$E7,$86
	.dc.b	$FA,$9E,$67,$27,$A7,$0F,$86,$7C,$BA,$F9,$2E,$23,$D0,$78,$4F,$8A
	.dc.b	$ED,$E6,$60,$DC,$F7,$3C,$31,$F6,$5B,$1E,$EB,$F8,$27,$94,$B3,$B3
	.dc.b	$B4,$F3,$69,$99,$EE,$D7,$E3,$C7,$38,$F3,$4D,$9F,$5B,$CD,$E2,$7D
	.dc.b	$09,$F2,$BF,$26,$7B,$B9,$6B,$F7,$67,$CE,$FA,$E3,$B9,$E3,$7E,$71
	.dc.b	$C4,$EE,$75,$29,$F7,$0F,$AB,$87,$E2,$9C,$4F,$67,$C1,$3D,$9E,$87
	.dc.b	$C0,$7C,$13,$C6,$3E,$A1,$93,$67,$90,$E6,$7C,$48,$FC,$03,$F9,$09
	.dc.b	$F8,$CF,$13,$D0,$53,$D1,$5F,$21,$B3,$F1,$9A,$7B,$8F,$19,$E1,$32
	.dc.b	$73,$33,$5E,$9C,$3F,$39,$5E,$29,$E0,$6B,$0F,$D0,$BF,$21,$B8,$7C
	.dc.b	$78,$F9,$38,$7C,$77,$D2,$72,$BE,$B1,$F2,$2B,$9D,$FA,$EB,$F8,$C9
	.dc.b	$FC,$6C,$FA,$C5,$32,$7F,$15,$3F,$90,$7D,$1F,$F8,$67,$D3,$D3,$63
	.dc.b	$93,$5D,$F1,$7F,$BF,$5C,$F0,$7F,$87,$5B,$B7,$28,$C4,$76,$6B,$FC
	.dc.b	$CE,$91,$C7,$0F,$E5,$F1,$FE,$87,$9F,$E3,$6D,$7F,$83,$BC,$2B,$35
	.dc.b	$C7,$2E,$9E,$97,$F3,$FC,$39,$B6,$3C,$13,$BC,$3F,$CB,$D8,$73,$7F
	.dc.b	$8D,$3D,$F5,$E3,$1F,$C3,$8D,$CF,$BE,$3C,$EF,$38,$4F,$9F,$0E,$31
	.dc.b	$E8,$9D,$4F,$79,$C9,$2F,$CB,$8F,$B0,$F8,$38,$FA,$70,$F3,$B5,$EF
	.dc.b	$87,$43,$AC,$67,$0F,$15,$F4,$AF,$15,$F4,$61,$82,$61,$E3,$7B,$9E
	.dc.b	$33,$58,$FB,$09,$5D,$5F,$87,$5D,$2B,$95,$79,$A3,$94,$62,$FC,$B3
	.dc.b	$F3,$27,$B5,$74,$8F,$09,$F5,$31,$FA,$18,$7A,$F0,$E5,$8F,$BE,$7E
	.dc.b	$16,$1E,$1A,$FC,$1D,$F8,$1F,$31,$87,$EF,$D7,$F8,$17,$0F,$4C,$77
	.dc.b	$AE,$4F,$8B,$0C,$92,$F9,$3E,$BC,$3C,$33,$D0,$FC,$41,$F8,$4C,$3B
	.dc.b	$57,$8A,$3E,$01,$F9,$C8,$FB,$07,$AE,$FD,$11,$E1,$9E,$E7,$C1,$39
	.dc.b	$3D,$4E,$F8,$FE,$66,$32,$FE,$3D,$E3,$1C,$CF,$0E,$1E,$B7,$07,$33
	.dc.b	$F0,$71,$CA,$FD,$31,$F5,$31,$F8,$E7,$17,$85,$7C,$3C,$3D,$31,$8B
	.dc.b	$F0,$BF,$83,$3F,$16,$FD,$3A,$FC,$04,$65,$E9,$87,$E1,$A3,$91,$E8
	.dc.b	$9F,$84,$FA,$23,$47,$BD,$F9,$B3,$D2,$3E,$24,$62,$78,$BD,$09,$1C
	.dc.b	$5F,$35,$FE,$EA,$FF,$CB,$13,$FC,$4D,$B8,$FE,$F3,$EA,$BF,$97,$5F
	.dc.b	$F0,$26,$D1,$A6,$BC,$47,$99,$FE,$55,$C3,$FC,$C9,$18,$BF,$F5,$A3
	.dc.b	$5F,$F5,$01,$D4,$FF,$39,$61,$C7,$7F,$F7,$C7,$3F,$F2,$C8,$C4,$A3
	.dc.b	$0F,$F9,$E7,$7C,$C3,$F3,$C6,$18,$6D,$7F,$B4,$5A,$FF,$A1,$EF,$8F
	.dc.b	$D5,$BD,$4A,$49,$FF,$6C,$43,$EA,$FB,$FD,$77,$AF,$5F,$F7,$03,$75
	.dc.b	$F4,$B6,$FD,$17,$D5,$FF,$7A,$78,$2F,$8C,$F6,$F8,$68,$C4,$EA,$73
	.dc.b	$0F,$AC,$BF,$59,$2B,$9B,$AA,$C4,$7F,$AB,$3F,$08,$7C,$0F,$C6,$6C
	.dc.b	$70,$3C,$7F,$5B,$DB,$AA,$46,$0F,$3B,$E6,$FA,$5F,$F6,$4C,$0F,$F6
	.dc.b	$43,$87,$FD,$89,$1F,$53,$EB,$5F,$0F,$7E,$2A,$57,$1E,$37,$1F,$E0
	.dc.b	$2F,$73,$98,$C5,$7D,$EF,$91,$79,$94,$E7,$5F,$8C,$EF,$7D,$6F,$E2
	.dc.b	$23,$FA,$DF,$CE,$D9,$36,$71,$F5,$9F,$9E,$25,$7F,$82,$9F,$CF,$7D
	.dc.b	$0F,$3E,$F3,$4E,$33,$F8,$6F,$0B,$F1,$64,$7F,$13,$F5,$9F,$88,$FB
	.dc.b	$A9,$1E,$D3,$F3,$9D,$C8,$E1,$1F,$EF,$3F,$49,$F5,$9F,$86,$5F,$49
	.dc.b	$4C,$45,$9E,$9E,$03,$EA,$38,$C6,$1A,$F5,$AE,$9E,$BF,$5E,$69,$3F
	.dc.b	$EF,$78,$FC,$AF,$D6,$3D,$8F,$F1,$96,$28,$C5,$7F,$8D,$E3,$13,$A9
	.dc.b	$CC,$FF,$B5,$65,$FB,$2B,$D8,$FD,$0B,$67,$D2,$D5,$D9,$F8,$65,$5D
	.dc.b	$9F,$95,$E6,$3C,$07,$B7,$FC,$11,$F8,$53,$A6,$BD,$35,$FE,$1F,$1C
	.dc.b	$23,$5B,$FF,$0C,$F9,$27,$C4,$D3,$A4,$7F,$10,$FA,$D7,$83,$7E,$3A
	.dc.b	$61,$E4,$9C,$B6,$3F,$8B,$FC,$E7,$B7,$7F,$66,$C8,$C6,$BF,$C8,$B0
	.dc.b	$5F,$F1,$EF,$AD,$FD,$76,$3F,$B1,$3F,$92,$38,$7F,$94,$B5,$CF,$C0
	.dc.b	$FF,$9A,$71,$7D,$FB,$E8,$FF,$94,$B6,$BE,$5B,$5F,$8C,$72,$3E,$7F
	.dc.b	$D7,$7C,$AD,$7D,$2D,$D3,$F9,$CB,$67,$F9,$C2,$F8,$EF,$E6,$FF,$7F
	.dc.b	$9C,$8D,$31,$39,$FA,$F7,$C5,$7A,$5B,$87,$FA,$16,$D1,$88,$F8,$87
	.dc.b	$F4,$77,$0F,$C7,$87,$E0,$4E,$6F,$E9,$2D,$9F,$C5,$B5,$F8,$33,$DA
	.dc.b	$FF,$A7,$FD,$83,$E1,$BE,$63,$6B,$C8,$D7,$C7,$48,$C4,$78,$FF,$01
	.dc.b	$D6,$FF,$82,$6E,$FF,$D8,$9B,$3F,$D8,$6A,$FC,$EF,$F3,$1E,$0F,$EC
	.dc.b	$AD,$9F,$E1,$8D,$7E,$CB,$FC,$DF,$E5,$77,$F7,$2F,$8A,$A4,$AE,$27
	.dc.b	$D7,$39,$AF,$9F,$FF,$59,$5F,$16,$C6,$67,$FD,$71,$39,$6F,$C3,$BE
	.dc.b	$3B,$CE,$D8,$F3,$B5,$F3,$D7,$37,$FC,$0F,$F2,$0C,$BF,$9E,$3E,$58
	.dc.b	$CE,$78,$9F,$E8,$1F,$64,$FA,$7F,$F0,$DB,$6B,$FD,$94,$D7,$FD,$9C
	.dc.b	$D7,$F5,$9E,$33,$ED,$DE,$FF,$F7,$66,$CF,$F2,$E6,$BF,$E1,$CA,$E9
	.dc.b	$C9,$FC,$5E,$1F,$67,$8E,$5F,$DC,$3F,$CE,$A3,$FA,$11,$AF,$3F,$FA
	.dc.b	$C4,$6E,$3F,$B7,$7D,$97,$C1,$7F,$DC,$4D,$7F,$DC,$CD,$7F,$42,$D7
	.dc.b	$F8,$37,$D8,$BE,$57,$FB,$23,$63,$C2,$D7,$FC,$6B,$F6,$7F,$D9,$9F
	.dc.b	$C5,$FD,$63,$EF,$FF,$1E,$D8,$CC,$63,$ED,$3F,$B8,$7D,$A7,$F1,$89
	.dc.b	$F6,$BF,$C8,$9F,$F2,$0B,$76,$F8,$CE,$E3,$FC,$8B,$E5,$3E,$87,$FC
	.dc.b	$A6,$D8,$FD,$E9,$FE,$11,$70,$E1,$AF,$F2,$5E,$C6,$5F,$E2,$98,$F5
	.dc.b	$FC,$77,$A8,$F5,$FD,$97,$F9,$2B,$D3,$E3,$3F,$E5,$26,$BF,$E5,$16
	.dc.b	$F8,$9F,$D1,$32,$FE,$59,$F0,$5F,$E5,$5A,$FE,$28,$D7,$F8,$0B,$5F
	.dc.b	$E2,$4D,$7F,$9D,$63,$FC,$7A,$BD,$B1,$E8,$F1,$F4,$D8,$CC,$EE,$3E
	.dc.b	$1A,$EF,$FD,$E4,$61,$AF,$F2,$48,$D7,$81,$CA,$73,$3F,$E7,$0F,$D7
	.dc.b	$3F,$A2,$75,$7F,$2D,$6C,$7F,$1A,$6B,$FC,$AD,$AF,$F9,$E3,$EF,$BF
	.dc.b	$CD,$53,$FC,$D5,$B1,$FC,$89,$AF,$F3,$36,$BF,$E4,$BE,$0F,$AF,$EE
	.dc.b	$7F,$7F,$E4,$FF,$00,$46,$2F,$F8,$EF,$DA,$7F,$92,$1F,$CA,$D6,$DF
	.dc.b	$EC,$9E,$1B,$E8,$FF,$65,$8E,$1A,$FF,$23,$6B,$FC,$35,$BF,$69,$FC
	.dc.b	$26,$BF,$6A,$FF,$90,$1B,$E2,$F7,$AF,$F3,$14,$E5,$FE,$5F,$C1,$FD
	.dc.b	$E2,$BF,$53,$EB,$6F,$4F,$C9,$FC,$65,$AD,$5F,$FA,$93,$ED,$1F,$05
	.dc.b	$D6,$11,$8C,$7F,$83,$EB,$F7,$DF,$E1,$31,$99,$FF,$56,$FD,$A3,$E9
	.dc.b	$7D,$AD,$AF,$C6,$B6,$7F,$56,$D7,$FB,$4F,$DA,$79,$38,$FF,$8D,$3F
	.dc.b	$E7,$0A,$FE,$2E,$E5,$FF,$09,$39,$9F,$F0,$3F,$98,$F0,$3F,$D0,$D5
	.dc.b	$FE,$54,$6C,$FF,$B0,$BE,$D3,$FD,$0B,$39,$8F,$81,$18,$7F,$7E,$D7
	.dc.b	$FB,$EB,$5F,$F1,$13,$5F,$F4,$A7,$DA,$7F,$57,$9F,$E5,$AF,$FB,$11
	.dc.b	$BE,$97,$CD,$8C,$4F,$F6,$6F,$C0,$DF,$37,$F5,$DB,$7F,$DB,$3E,$D3
	.dc.b	$FC,$10,$FF,$6E,$8C,$D7,$F6,$E9,$CB,$FE,$12,$E5,$46,$23,$F8,$A3
	.dc.b	$5F,$D0,$35,$43,$D6,$78,$DF,$89,$6E,$1F,$DF,$BE,$9F,$FC,$E8,$FF
	.dc.b	$80,$23,$37,$FE,$00,$9C,$B7,$B7,$F2,$5C,$5F,$A9,$8C,$5F,$F9,$AB
	.dc.b	$D9,$FE,$7C,$DF,$33,$F3,$DE,$2B,$ED,$3F,$C6,$73,$93,$FC,$E2,$E2
	.dc.b	$FF,$C3,$ED,$7F,$C6,$0D,$7F,$DF,$AD,$7F,$DF,$9D,$BC,$71,$FF,$31
	.dc.b	$B6,$BF,$D9,$D6,$FB,$47,$D4,$F9,$2B,$58,$2B,$53,$FB,$E6,$DF,$F9
	.dc.b	$1E,$F8,$B6,$75,$F7,$9F,$D3,$E3,$FD,$16,$DB,$FF,$20,$FD,$2F,$E7
	.dc.b	$72,$FF,$3B,$CE,$5F,$F3,$CC,$66,$BC,$ED,$AF,$23,$F4,$A7,$79,$7C
	.dc.b	$CF,$85,$FC,$3C,$7F,$A0,$9B,$7F,$03,$ED,$1F,$57,$BF,$84,$FA,$71
	.dc.b	$89,$FF,$31,$37,$2F,$F2,$7B,$5F,$F3,$03,$5F,$97,$F6,$9F,$E9,$C9
	.dc.b	$FF,$54,$FC,$67,$E8,$B2,$FF,$49,$CE,$6B,$FC,$C3,$39,$AE,$3E,$5F
	.dc.b	$E1,$2B,$57,$C2,$31,$EA,$7E,$23,$E6,$5F,$65,$F8,$AD,$3F,$96,$67
	.dc.b	$FE,$6C,$9C,$E5,$FE,$BA,$9C,$E3,$FE,$A4,$97,$D9,$73,$C3,$A3,$6F
	.dc.b	$FD,$8F,$EE,$FF,$33,$F1,$BF,$D9,$AD,$FB,$37,$DA,$67,$E5,$BF,$C4
	.dc.b	$57,$FE,$E6,$9C,$BF,$ED,$96,$EB,$FE,$D7,$6B,$FE,$44,$6A,$3F,$DA
	.dc.b	$6F,$F9,$56,$3A,$5F,$87,$CE,$F1,$3E,$3F,$97,$F3,$7C,$AF,$FB,$EA
	.dc.b	$BC,$51,$8F,$85,$F4,$7D,$7E,$E7,$CC,$DC,$BF,$D2,$7F,$3A,$FB,$9F
	.dc.b	$DA,$B6,$BF,$DC,$75,$F5,$EB,$F4,$93,$97,$FD,$37,$19,$F0,$4D,$3F
	.dc.b	$DD,$35,$CB,$E0,$3F,$A9,$1F,$AE,$A3,$5F,$D1,$47,$FB,$4E,$31,$F0
	.dc.b	$3F,$DB,$27,$F5,$47,$F8,$E1,$C1,$9E,$3F,$EB,$D1,$F8,$23,$FD,$9E
	.dc.b	$7C,$D7,$3C,$9F,$C4,$BE,$D7,$FB,$E6,$1F,$F0,$03,$E9,$BF,$58,$FF
	.dc.b	$06,$8C,$DF,$FA,$E2,$72,$FF,$05,$8F,$EC,$1D,$1F,$EF,$07,$6D,$F5
	.dc.b	$FB,$76,$7E,$2F,$E7,$3F,$B7,$5F,$F5,$CF,$9C,$E3,$97,$E0,$7A,$D5
	.dc.b	$AC,$53,$9A,$DD,$AF,$7F,$35,$D7,$E1,$3B,$36,$FF,$19,$87,$2B,$E1
	.dc.b	$8F,$F3,$78,$C7,$1B,$EA,$BF,$AF,$57,$F7,$0A,$D5,$FF,$28,$9C,$DF
	.dc.b	$3C,$BE,$EC,$F6,$3D,$55,$C2,$FC,$DF,$5B,$F2,$2B,$71,$FD,$F2,$30
	.dc.b	$4F,$23,$F6,$0F,$CE,$C9,$CE,$FA,$3F,$CA,$9F,$D2,$BA,$3F,$1F,$9F
	.dc.b	$87,$1F,$CA,$CE,$66,$4F,$E6,$BA,$ED,$7F,$BC,$BE,$C7,$F2,$AD,$D7
	.dc.b	$FA,$AF,$D8,$FF,$B3,$6F,$A6,$6B,$DE,$FF,$22,$B5,$87,$FB,$F1,$B1
	.dc.b	$F3,$3E,$D3,$FC,$2A,$EF,$E6,$7F,$43,$C9,$32,$FF,$5B,$E1,$FF,$00
	.dc.b	$5F,$F5,$CF,$6F,$E0,$5F,$E0,$91,$FC,$17,$E6,$FF,$44,$48,$FF,$3E
	.dc.b	$C7,$12,$FD,$A7,$F0,$BF,$9B,$7D,$7F,$29,$FF,$7B,$57,$81,$56,$E7
	.dc.b	$5F,$21,$FA,$A9,$C8,$BE,$6F,$92,$3F,$19,$18,$8F,$3D,$73,$75,$75
	.dc.b	$C9,$FD,$4A,$FB,$C7,$F2,$1F,$47,$94,$FF,$23,$AC,$56,$A3,$81,$EB
	.dc.b	$8E,$8F,$B0,$F3,$47,$FA,$F2,$33,$87,$9A,$31,$3A,$3C,$A5,$3C,$B8
	.dc.b	$78,$A7,$84,$77,$3C,$27,$F3,$77,$F5,$97,$E8,$8F,$8E,$F1,$C3,$D3
	.dc.b	$1C,$6B,$0F,$73,$DC,$5A,$EE,$F8,$AF,$E0,$4F,$5C,$3D,$77,$F0,$2B
	.dc.b	$BD,$7C,$5C,$35,$1E,$63,$E2,$38,$BF,$09,$67,$A9,$F1,$1C,$B4,$CC
	.dc.b	$FC,$19,$F4,$9D,$8B,$1B,$8F,$97,$3D,$5B,$19,$7E,$45,$7A,$6F,$D7
	.dc.b	$1F,$0C,$F5,$9F,$52,$FB,$57,$B4,$E5,$1F,$7E,$FB,$3D,$0E,$55,$D2
	.dc.b	$33,$1B,$9F,$C0,$D7,$A2,$3D,$91,$D5,$FA,$CF,$C0,$7E,$91,$EA,$32
	.dc.b	$53,$27,$86,$7C,$EF,$A6,$3B,$9D,$5D,$9E,$33,$99,$C5,$F6,$E1,$F8
	.dc.b	$28,$FC,$2C,$FD,$47,$CE,$75,$29,$F0,$6B,$DF,$3C,$63,$F7,$33,$C5
	.dc.b	$F0,$C6,$6F,$E3,$5F,$83,$0F,$A9,$5F,$93,$6D,$F1,$8F,$1C,$FC,$56
	.dc.b	$9F,$0C,$A7,$AC,$F6,$9E,$F3,$C3,$3D,$B0,$FA,$51,$FC,$11,$AF,$23
	.dc.b	$81,$70,$F4,$3E,$53,$EF,$96,$FF,$42,$FF,$1F,$3F,$5A,$53,$F8,$D9
	.dc.b	$70,$C7,$D6,$FF,$86,$3D,$CB,$CC,$69,$2B,$A4,$B8,$9C,$3C,$9B,$3A
	.dc.b	$6B,$FC,$C6,$78,$4E,$CF,$E4,$8F,$FA,$4A,$2F,$AF,$FD,$03,$2F,$E5
	.dc.b	$93,$C2,$7F,$8B,$36,$33,$3B,$DB,$FA,$3C,$F3,$AD,$E1,$E4,$BE,$6F
	.dc.b	$9B,$5E,$33,$FC,$58,$57,$18,$DE,$39,$8F,$06,$1D,$B1,$F0,$5F,$F9
	.dc.b	$66,$7C,$18,$F0,$8F,$0D,$FB,$18,$60,$FE,$65,$AB,$F0,$47,$95,$FD
	.dc.b	$82,$BF,$C3,$13,$98,$C7,$07,$F3,$1A,$DD,$FF,$80,$6B,$97,$CF,$6A
	.dc.b	$11,$D7,$D0,$C3,$63,$85,$7F,$9F,$1B,$86,$BE,$77,$FD,$0F,$1D,$5A
	.dc.b	$FE,$03,$1F,$E8,$51,$8C,$BC,$EF,$FA,$5A,$78,$79,$7C,$C8,$F3,$F2
	.dc.b	$FB,$70,$7E,$57,$35,$FF,$4B,$AF,$4B,$63,$33,$B8,$7C,$F7,$9C,$95
	.dc.b	$E7,$9F,$57,$D2,$7A,$FE,$CA,$4F,$AE,$3D,$DF,$11,$CC,$91,$FA,$9C
	.dc.b	$7C,$71,$DF,$E6,$7E,$21,$27,$E2,$7C,$37,$C7,$F8,$39,$1F,$1D,$F9
	.dc.b	$5F,$37,$F2,$FD,$BF,$41,$27,$B6,$C8,$F6,$7D,$C7,$D1,$24,$FD,$1F
	.dc.b	$B2,$FA,$9F,$53,$23,$EA,$3F,$5F,$E5,$FE,$C7,$9F,$F1,$49,$3D,$9F
	.dc.b	$91,$F1,$5F,$74,$91,$F8,$9F,$DC,$3F,$CC,$CF,$DD,$3F,$12,$BE,$5F
	.dc.b	$F1,$2D,$8E,$11,$8F,$57,$3E,$57,$DB,$A3,$49,$5F,$36,$1F,$3D,$D0
	.dc.b	$91,$E3,$F6,$7B,$FB,$32,$3C,$CF,$8B,$E1,$3C,$7F,$59,$DC,$93,$D5
	.dc.b	$F1,$7C,$07,$A0,$91,$EF,$F9,$CF,$5F,$D7,$C8,$F8,$4F,$BB,$E0,$3D
	.dc.b	$FF,$5F,$EB,$24,$FA,$5F,$77,$C4,$7C,$42,$47,$CB,$F7,$3E,$3F,$C2
	.dc.b	$48,$F9,$CF,$CA,$FA,$4F,$97,$F1,$DF,$1C,$93,$F1,$5F,$95,$F6,$DF
	.dc.b	$40,$91,$F5,$FE,$E3,$F7,$74,$F7,$FF,$80,$7C,$27,$FA,$1F,$D6,$B9
	.dc.b	$FF,$23,$F3,$FF,$85,$DE,$F1,$FF,$37,$F5,$FF,$9A,$BD,$DF,$B9,$1F
	.dc.b	$D2,$5A,$FF,$9A,$5A,$FF,$B5,$7D,$1F,$E1,$CF,$FB,$69,$B1,$F8,$AB
	.dc.b	$FC,$5B,$D1,$F0,$FB,$1E,$8F,$B9,$97,$FC,$67,$14,$E2,$3F,$25,$E6
	.dc.b	$A5,$F6,$7F,$00,$D6,$1E,$CF,$B5,$F8,$64,$BE,$9F,$6B,$F1,$BE,$E6
	.dc.b	$47,$77,$4F,$E7,$F8,$7F,$D5,$D3,$F2,$2A,$4F,$FA,$B6,$72,$DE,$04
	.dc.b	$FC,$EC,$64,$E6,$2B,$D7,$F6,$D7,$37,$D6,$6C,$73,$AF,$B7,$C7,$23
	.dc.b	$E2,$3F,$7B,$DF,$FB,$FF,$03,$27,$E4,$3B,$7E,$3E,$EF,$88,$FC,$2F
	.dc.b	$DC,$C9,$F9,$D1,$5D,$FD,$0F,$E2,$8C,$71,$7E,$1B,$9B,$EE,$FE,$E4
	.dc.b	$F8,$8A,$7E,$42,$3F,$03,$EE,$7F,$44,$C3,$FD,$5C,$D7,$FD,$72,$D4
	.dc.b	$7F,$AD,$3D,$EF,$F5,$CC,$3E,$23,$99,$27,$FD,$B8,$6C,$FC,$37,$C5
	.dc.b	$FE,$80,$8F,$E2,$7E,$0B,$F4,$7F,$0B,$23,$F1,$8E,$9F,$BD,$FB,$5B
	.dc.b	$C7,$FD,$3B,$63,$EF,$35,$DF,$8F,$AF,$22,$46,$9E,$35,$FB,$2F,$94
	.dc.b	$E7,$E8,$4A,$FC,$E4,$76,$F4,$7F,$C6,$5E,$03,$2F,$E4,$35,$FE,$37
	.dc.b	$8C,$4E,$A7,$33,$FE,$3B,$FC,$57,$E9,$3F,$11,$AF,$F3,$70,$E3,$95
	.dc.b	$E3,$FD,$0C,$3A,$D7,$EB,$FE,$23,$F8,$49,$1F,$D9,$7C,$AF,$F0,$CE
	.dc.b	$79,$1F,$B5,$74,$FE,$D3,$E9,$BF,$DD,$98,$FF,$85,$9B,$8F,$FB,$15
	.dc.b	$B3,$FE,$12,$7F,$C2,$6F,$F8,$5D,$FC,$03,$79,$7F,$C2,$B1,$88,$EC
	.dc.b	$63,$9F,$FC,$0D,$1F,$8B,$F2,$0D,$3F,$1D,$F4,$4F,$C5,$1E,$BF,$F8
	.dc.b	$8E,$30,$D8,$CE,$1E,$3C,$3C,$3F,$31,$23,$CC,$F9,$3E,$33,$CB,$DB
	.dc.b	$27,$D4,$ED,$F8,$9F,$19,$E6,$23,$FE,$0B,$CF,$FC,$2E,$FF,$9D,$73
	.dc.b	$F3,$7A,$FE,$32,$47,$C5,$AF,$87,$F1,$7F,$00,$91,$CD,$FF,$3E,$3F
	.dc.b	$E7,$3F,$6B,$F9,$9F,$92,$69,$CA,$3E,$87,$C6,$7C,$E2,$3F,$4F,$E3
	.dc.b	$3E,$8F,$C6,$48,$FA,$CE,$9F,$C0,$7C,$67,$D2,$23,$ED,$7F,$D1,$DF
	.dc.b	$20,$FA,$9F,$E4,$B3,$FC,$03,$A7,$F2,$6F,$FA,$36,$32,$FF,$83,$E7
	.dc.b	$87,$A9,$F8,$2B,$FC,$1F,$C7,$7E,$13,$E3,$A5,$FE,$26,$3F,$0F,$F1
	.dc.b	$BF,$84,$24,$7C,$F7,$F0,$FF,$1B,$F8,$C2,$47,$E4,$63,$FB,$3F,$ED
	.dc.b	$5F,$C9,$E3,$F9,$F2,$F5,$72,$6B,$D3,$DF,$1F,$EE,$A7,$FD,$45,$5F
	.dc.b	$B9,$6B,$FC,$78,$FD,$83,$87,$1D,$FF,$98,$F8,$EF,$CF,$92,$3F,$A7
	.dc.b	$7B,$2F,$33,$F4,$7F,$1D,$27,$F5,$0F,$E9,$BE,$3B,$F4,$FF,$1D,$FA
	.dc.b	$8F,$8E,$95,$FA,$D8,$F6,$FC,$0F,$9C,$91,$E7,$7F,$CE,$1F,$3F,$FD
	.dc.b	$DF,$CA,$FE,$6B,$AA,$31,$5F,$E8,$A6,$C6,$7E,$17,$FD,$17,$0F,$8F
	.dc.b	$E6,$4A,$FF,$35,$46,$5B,$F1,$FE,$02,$38,$3F,$DA,$DF,$59,$FE,$44
	.dc.b	$F8,$9B,$EB,$7F,$85,$E7,$E6,$09,$D3,$FE,$DF,$6B,$A7,$FC,$A3,$97
	.dc.b	$4B,$FE,$23,$39,$6B,$F5,$E7,$6F,$69,$F8,$55,$FE,$64,$8F,$8B,$87
	.dc.b	$F9,$63,$3F,$6C,$E6,$7E,$33,$6B,$E5,$70,$DB,$FB,$97,$F1,$23,$06
	.dc.b	$7C,$FF,$D2,$F8,$3F,$EC,$0A,$FE,$33,$5F,$EB,$D5,$F1,$FA,$8F,$1D
	.dc.b	$5C,$9E,$AB,$9B,$FC,$E3,$5F,$D3,$5B,$3F,$2A,$75,$B7,$A5,$B5,$F6
	.dc.b	$3C,$7E,$9A,$FE,$0F,$E3,$EE,$9F,$C3,$6B,$C9,$18,$F2,$E4,$79,$FE
	.dc.b	$3E,$E7,$FE,$3A,$B7,$B9,$C8,$F4,$66,$F9,$F7,$9B,$9C,$7F,$46,$79
	.dc.b	$77,$FE,$B6,$FA,$CC,$1E,$9F,$11,$D7,$D1,$76,$58,$C3,$EA,$F7,$65
	.dc.b	$7C,$DF,$7A,$ED,$E7,$7E,$C1,$23,$F3,$6F,$DB,$F8,$AF,$87,$F2,$1F
	.dc.b	$11,$B3,$FC,$7C,$FF,$0F,$F9,$AF,$D0,$15,$F5,$E3,$0F,$CC,$F8,$9C
	.dc.b	$4E,$AB,$F0,$11,$F8,$92,$3F,$AC,$7F,$19,$F1,$9F,$50,$FA,$BF,$23
	.dc.b	$67,$F1,$0E,$1F,$C5,$7C,$8D,$D3,$FD,$7D,$DE,$FD,$BB,$E6,$C6,$1B
	.dc.b	$1F,$CE,$3B,$1F,$69,$FE,$8F,$C3,$DD,$6A,$E5,$1F,$88,$9D,$4F,$EA
	.dc.b	$DA,$FF,$22,$6B,$FC,$41,$AB,$DA,$FE,$F5,$B7,$E7,$5B,$5E,$0C,$3C
	.dc.b	$7F,$25,$9F,$81,$F9,$18,$7F,$A2,$31,$FF,$45,$DF,$FB,$ED,$B1,$E3
	.dc.b	$AF,$DB,$46,$23,$C9,$F2,$7E,$93,$D5,$F2,$56,$FC,$15,$FE,$92,$6C
	.dc.b	$7B,$FD,$9D,$C7,$C4,$CF,$FC,$25,$BF,$94,$F8,$9F,$27,$67,$9B,$E2
	.dc.b	$23,$E7,$7C,$BE,$7F,$CB,$3E,$67,$C9,$D9,$F1,$B8,$7C,$BE,$AD,$D7
	.dc.b	$E1,$7A,$FF,$4C,$FA,$9E,$BD,$AF,$5B,$87,$DB,$F2,$76,$BF,$C9,$9C
	.dc.b	$BF,$D8,$E7,$FC,$E6,$D8,$FF,$54,$73,$7F,$A8,$A7,$38,$7E,$F2,$31
	.dc.b	$3F,$C3,$DA,$FF,$65,$F1,$7F,$21,$C1,$FC,$46,$73,$3F,$83,$F9,$3E
	.dc.b	$5C,$5F,$8B,$6E,$3F,$EB,$3F,$DB,$BF,$D8,$8D,$EC,$FE,$17,$1F,$82
	.dc.b	$D3,$FC,$1F,$9A,$37,$8F,$FA,$FD,$D5,$FF,$99,$5B,$1F,$CA,$97,$7F
	.dc.b	$4E,$BF,$CE,$7C,$9C,$BF,$C1,$FC,$9D,$C3,$F2,$0F,$E8,$4F,$D1,$7C
	.dc.b	$9D,$9F,$93,$CD,$FC,$DE,$BF,$50,$DD,$1C,$F6,$30,$FF,$4D,$9F,$D7
	.dc.b	$36,$77,$5F,$D9,$27,$FC,$5D,$3F,$9C,$F9,$3E,$D9,$7F,$30,$AF,$CC
	.dc.b	$C6,$3C,$CF,$DB,$7C,$9F,$ED,$FE,$46,$D7,$E7,$9F,$DD,$7C,$8E,$32
	.dc.b	$FD,$63,$73,$FE,$09,$DF,$FE,$C5,$BF,$D2,$FC,$8F,$F4,$8F,$80,$FF
	.dc.b	$5F,$C6,$0F,$DC,$38,$C3,$FD,$FA,$DB,$FE,$1C,$BC,$5E,$53,$F7,$3D
	.dc.b	$AF,$E8,$FF,$7C,$DF,$18,$FF,$BD,$63,$0F,$E1,$7D,$9C,$DF,$DF,$6E
	.dc.b	$5F,$84,$E8,$FE,$F5,$39,$AF,$C6,$7B,$9F,$8A,$BF,$57,$C9,$49,$F4
	.dc.b	$FC,$95,$BF,$4B,$EF,$3E,$07,$C9,$59,$F7,$FC,$97,$C3,$F9,$2F,$88
	.dc.b	$D8,$F8,$7F,$23,$86,$9C,$0F,$F7,$0F,$EC,$3F,$E0,$8C,$7F,$D7,$CD
	.dc.b	$DB,$FC,$D7,$08,$F5,$C6,$1D,$4F,$F9,$4D,$F7,$CF,$F8,$A9,$AE,$75
	.dc.b	$FC,$74,$62,$3E,$97,$1F,$F7,$4A,$FD,$87,$C3,$FD,$DF,$73,$F5,$EF
	.dc.b	$DF,$F9,$29,$3F,$7B,$E4,$AE,$1F,$7B,$E4,$BF,$C4,$98,$9A,$4A,$FC
	.dc.b	$D7,$42,$5F,$23,$F8,$B3,$F1,$9F,$21,$6B,$F1,$7F,$23,$F8,$F3,$F2
	.dc.b	$1F,$23,$63,$F1,$FF,$05,$E0,$23,$DB,$04,$79,$3E,$43,$F3,$07,$E6
	.dc.b	$62,$CF,$E6,$3E,$43,$F3,$87,$E7,$7E,$12,$C7,$E7,$3E,$13,$DA,$47
	.dc.b	$B3,$FE,$43,$FB,$AF,$ED,$71,$FE,$6E,$6F,$07,$E7,$D7,$1F,$F8,$C1
	.dc.b	$CF,$01,$FF,$37,$57,$F6,$16,$D7,$F7,$18,$E9,$DD,$C7,$E4,$79,$74
	.dc.b	$78,$49,$3F,$3A,$FE,$F6,$1D,$7C,$3F,$01,$E1,$F8,$EB,$5D,$FE,$3F
	.dc.b	$C8,$79,$7E,$3E,$C7,$8F,$E3,$FD,$C4,$7B,$3F,$07,$E3,$FD,$87,$B7
	.dc.b	$E3,$EC,$FA,$FC,$FF,$80,$7C,$1F,$8C,$B1,$EF,$F8,$CF,$94,$47,$D4
	.dc.b	$FC,$DF,$63,$E4,$1F,$27,$D8,$B3,$F1,$FD,$AF,$98,$7C,$DF,$8A,$B1
	.dc.b	$F2,$FE,$2B,$EB,$11,$F8,$CF,$D9,$F8,$6F,$F3,$1F,$D4,$7F,$87,$3C
	.dc.b	$AF,$F1,$27,$32,$73,$F2,$5A,$F8,$B9,$8F,$F8,$5F,$07,$C5,$E3,$B7
	.dc.b	$EC,$1F,$C7,$E3,$33,$FA,$26,$C7,$FA,$4D,$AF,$FB,$29,$AF,$FA,$E1
	.dc.b	$FF,$0A,$3D,$3A,$91,$8C,$FF,$C7,$4F,$F7,$46,$FC,$1F,$B3,$E1,$66
	.dc.b	$1E,$7F,$7E,$E3,$CF,$07,$C0,$7C,$23,$E1,$FC,$55,$BF,$73,$87,$E0
	.dc.b	$7C,$55,$BF,$79,$FE,$E5,$F9,$E2,$F9,$D7,$B1,$FE,$B1,$BC,$C6,$27
	.dc.b	$FD,$83,$58,$AD,$47,$17,$FB,$14,$66,$3F,$DD,$F5,$5F,$01,$B3,$25
	.dc.b	$33,$C3,$DD,$FF,$71,$C7,$9A,$31,$1E,$89,$C5,$78,$A7,$32,$9F,$03
	.dc.b	$FD,$6E,$AB,$93,$C0,$64,$A6,$6F,$BF,$02,$78,$D7,$48,$C5,$70,$8F
	.dc.b	$F0,$13,$EE,$2B,$87,$D0,$7F,$81,$FE,$79,$F1,$6F,$91,$E3,$18,$6C
	.dc.b	$66,$78,$35,$FF,$79,$CB,$4F,$F4,$14,$75,$6B,$A6,$BF,$80,$11,$C7
	.dc.b	$8F,$C3,$3C,$B3,$FA,$52,$DB,$84,$61,$BB,$73,$9E,$58,$7D,$B6,$D6
	.dc.b	$E3,$F0,$51,$CD,$B7,$DE,$B9,$3E,$AD,$B8,$49,$A7,$92,$35,$3C,$31
	.dc.b	$EB,$8F,$92,$FF,$3B,$87,$77,$C9,$7D,$DC,$B5,$F8,$73,$C5,$A7,$9A
	.dc.b	$BF,$39,$AB,$E7,$9F,$0D,$F2,$E7,$FD,$6D,$18,$AF,$F5,$33,$5F,$E0
	.dc.b	$8B,$E2,$78,$70,$FE,$CB,$0E,$9F,$11,$8B,$FF,$19,$5A,$31,$3A,$9E
	.dc.b	$B8,$E3,$E7,$9F,$01,$49,$3F,$EB,$78,$7C,$17,$2E,$7F,$C8,$CF,$F2
	.dc.b	$5E,$4C,$1C,$1F,$E3,$54,$F8,$1F,$CE,$E4,$8C,$4F,$F8,$21,$AF,$E3
	.dc.b	$4F,$F2,$53,$87,$F9,$47,$C8,$78,$48,$F9,$6B,$D5,$F0,$1E,$BF,$80
	.dc.b	$93,$E9,$8F,$77,$BF,$E2,$C3,$FC,$2B,$3E,$18,$C1,$E6,$CD,$F2,$1E
	.dc.b	$F2,$4F,$C3,$F8,$1F,$91,$F0,$3F,$0F,$E3,$E4,$FC,$78,$F8,$DF,$1B
	.dc.b	$CC,$91,$FD,$15,$ED,$F1,$DF,$2C,$91,$F3,$FE,$0B,$EA,$7C,$0F,$CF
	.dc.b	$EA,$93,$F4,$E3,$E9,$7C,$5F,$9C,$91,$FD,$45,$F5,$74,$FD,$72,$47
	.dc.b	$DB,$F8,$0F,$C1,$FC,$F7,$F8,$62,$3F,$AF,$75,$FD,$D3,$3F,$03,$F7
	.dc.b	$FE,$3F,$E2,$12,$7F,$AF,$3F,$1F,$E3,$BF,$08,$48,$FF,$7F,$5B,$E0
	.dc.b	$7B,$FC,$07,$E4,$BF,$72,$FB,$DE,$27,$F2,$2E,$F4,$63,$C4,$FC,$E6
	.dc.b	$0B,$E1,$5F,$8A,$AF,$F4,$76,$3F,$E5,$06,$BF,$89,$8E,$66,$FE,$37
	.dc.b	$C0,$48,$F2,$FC,$07,$C2,$F7,$FC,$9E,$2C,$9F,$4C,$79,$FE,$37,$C0
	.dc.b	$48,$F0,$3E,$3F,$1F,$D8,$48,$F8,$3E,$F7,$CE,$F7,$7E,$07,$9D,$27
	.dc.b	$E2,$C7,$C3,$F2,$3D,$84,$8F,$63,$EF,$F8,$CF,$90,$48,$F9,$BE,$E7
	.dc.b	$EF,$FE,$47,$FD,$BF,$E7,$FF,$01,$97,$87,$F7,$5F,$EF,$2D,$E7,$FF
	.dc.b	$76,$FC,$07,$F1,$D7,$FD,$27,$1F,$ED,$79,$E3,$D1,$FD,$46,$30,$7F
	.dc.b	$B6,$F0,$FF,$76,$FA,$FE,$1C,$3F,$C1,$B7,$FE,$04,$C3,$AC,$F8,$EF
	.dc.b	$FD,$3F,$1F,$7F,$E0,$3F,$C5,$75,$BC,$7F,$8D,$43,$DB,$FE,$2E,$D9
	.dc.b	$D1,$FE,$CD,$78,$BF,$E2,$77,$4F,$F8,$A7,$A5,$E7,$FE,$3C,$E7,$5D
	.dc.b	$32,$95,$D3,$A9,$87,$AA,$3A,$7B,$7F,$97,$F4,$64,$6F,$24,$72,$8F
	.dc.b	$F2,$1E,$ED,$D7,$F1,$73,$95,$EA,$AD,$4E,$23,$F9,$04,$71,$9F,$F0
	.dc.b	$1B,$63,$F9,$20,$70,$F0,$F9,$67,$C0,$FE,$0D,$B1,$DD,$AF,$9B,$0F
	.dc.b	$8F,$C1,$F4,$F8,$BE,$7F,$AB,$2B,$E2,$6D,$F8,$5E,$2F,$AB,$3F,$B9
	.dc.b	$F4,$25,$6A,$3E,$BF,$A3,$23,$8B,$F6,$BD,$5F,$B7,$E9,$49,$EA,$ED
	.dc.b	$FF,$26,$FC,$FB,$C8,$FC,$D7,$77,$F5,$09,$FC,$09,$FE,$4E,$F0,$FF
	.dc.b	$3D,$5B,$8C,$CB,$C5,$FF,$14,$36,$3F,$C5,$AD,$7E,$5B,$5F,$EA,$1D
	.dc.b	$9F,$1B,$87,$F4,$7C,$9F,$99,$3C,$77,$F9,$4F,$3A,$5F,$C6,$7F,$2F
	.dc.b	$E9,$7F,$1D,$F5,$64,$FE,$1D,$DB,$F8,$6F,$9F,$7C,$27,$EE,$57,$C2
	.dc.b	$D3,$11,$67,$ED,$78,$FF,$74,$C3,$CF,$E2,$79,$79,$F2,$3F,$1D,$82
	.dc.b	$7F,$82,$7C,$FF,$C7,$9F,$C5,$B5,$FE,$E0,$D7,$FC,$7C,$D7,$F2,$12
	.dc.b	$F8,$CF,$41,$23,$EA,$CF,$5E,$0F,$DF,$7A,$9F,$BA,$E6,$95,$F8,$C8
	.dc.b	$AF,$ED,$FD,$AF,$E8,$86,$2B,$FA,$27,$CF,$FF,$43,$97,$90,$8F,$B5
	.dc.b	$3A,$E4,$FF,$0E,$C6,$E7,$51,$FD,$3F,$E4,$A5,$FE,$FE,$B8,$F8,$1F
	.dc.b	$E9,$AF,$98,$7C,$CF,$F4,$FA,$FD,$7C,$63,$D6,$FF,$3A,$B6,$33,$3B
	.dc.b	$7F,$A8,$79,$8F,$94,$FE,$BC,$48,$FE,$33,$5C,$5F,$F5,$F6,$3F,$E9
	.dc.b	$7D,$7D,$FD,$DE,$A3,$D7,$EA,$FF,$8E,$EF,$CB,$F1,$F7,$33,$4E,$18
	.dc.b	$F6,$87,$C7,$FA,$1B,$1E,$88,$C1,$FD,$AF,$E0,$7F,$BA,$12,$39,$47
	.dc.b	$F4,$BF,$85,$EF,$1F,$E9,$66,$BF,$C5,$1A,$FF,$8E,$5A,$FF,$95,$76
	.dc.b	$7C,$A7,$F8,$04,$91,$FB,$C9,$E7,$AF,$E0,$3C,$BF,$B6,$49,$FB,$1E
	.dc.b	$AF,$C1,$F0,$64,$7B,$DD,$3F,$EF,$4F,$8A,$7C,$FF,$F3,$28,$FE,$53
	.dc.b	$18,$F5,$7F,$D5,$ED,$8C,$CE,$DF,$B7,$F0,$9F,$E1,$72,$3F,$D4,$A7
	.dc.b	$83,$F6,$BA,$7F,$23,$CF,$23,$F1,$D1,$F9,$4E,$1F,$F3,$0F,$39,$A4
	.dc.b	$9F,$F0,$D7,$CA,$FF,$89,$48,$FE,$5E,$3F,$BF,$7C,$FF,$F9,$C2,$3F
	.dc.b	$A7,$BF,$EB,$97,$FA,$63,$5C,$FA,$5F,$D0,$FC,$5F,$E9,$7E,$B7,$F0
	.dc.b	$39,$FE,$97,$EB,$A3,$11,$FD,$2E,$7C,$F8,$FE,$B0,$91,$FA,$9E,$6F
	.dc.b	$F1,$D7,$CB,$48,$FE,$C8,$FE,$CB,$83,$F6,$7F,$0D,$E8,$F6,$25,$7E
	.dc.b	$C6,$3F,$AB,$79,$7E,$42,$47,$93,$3F,$F2,$43,$FE,$3D,$9F,$F2,$41
	.dc.b	$1F,$EF,$1F,$3D,$FE,$0C,$F1,$3F,$D4,$5E,$47,$FB,$B2,$7F,$C8,$4D
	.dc.b	$7F,$DB,$1E,$3F,$FA,$C2,$7F,$D2,$73,$99,$E3,$F0,$F8,$23,$FD,$C2
	.dc.b	$BF,$D6,$F1,$FE,$F3,$F9,$FF,$F2,$E7,$CA,$7F,$BE,$1B,$5F,$DD,$9A
	.dc.b	$FF,$A6,$DA,$FF,$BA,$5F,$E5,$55,$C7,$CE,$FF,$2C,$CE,$5A,$84,$F0
	.dc.b	$7D,$BF,$31,$E9,$AF,$F6,$5B,$71,$34,$C5,$CC,$BD,$93,$FC,$ED,$B1
	.dc.b	$C7,$64,$7F,$89,$67,$FC,$F4,$FD,$0F,$2F,$38,$BC,$3F,$F3,$D5,$6A
	.dc.b	$3F,$C4,$7F,$3F,$FE,$32,$BF,$C7,$B6,$3F,$C3,$ED,$7D,$C7,$F8,$75
	.dc.b	$C3,$FE,$DD,$DD,$F1,$3E,$0C,$3C,$3D,$3F,$99,$E1,$C6,$3B,$AB,$87
	.dc.b	$F8,$9F,$8D,$58,$E0,$FF,$78,$F4,$DD,$7F,$B3,$63,$F8,$CF,$23,$13
	.dc.b	$BD,$5F,$13,$4C,$78,$B6,$FF,$C7,$9C,$0F,$86,$FC,$67,$02,$31,$3F
	.dc.b	$0A,$73,$18,$DF,$FC,$83,$E0,$3A,$7D,$38,$7E,$D7,$0B,$39,$7F,$C7
	.dc.b	$87,$A7,$92,$3F,$A4,$F5,$AE,$1F,$56,$3E,$54,$7E,$1F,$5F,$F0,$8C
	.dc.b	$7E,$75,$B3,$FE,$A9,$6B,$FA,$85,$F1,$1F,$0F,$D7,$F8,$8D,$AF,$C3
	.dc.b	$7C,$CF,$CB,$3E,$8F,$90,$F9,$FE,$33,$A3,$F9,$53,$8C,$7F,$E0,$04
	.dc.b	$EE,$FF,$68,$DA,$FE,$73,$5B,$AF,$C3,$FA,$5F,$DE,$31,$FB,$ED,$AF
	.dc.b	$E0,$BE,$EE,$3D,$25,$7D,$B8,$93,$FC,$2B,$E2,$6E,$1F,$1C,$FF,$07
	.dc.b	$F4,$FF,$6E,$9F,$D1,$7A,$9F,$8C,$EF,$7C,$FF,$E9,$A7,$27,$FA,$5A
	.dc.b	$32,$FF,$0D,$6B,$FE,$7A,$6D,$FF,$B8,$A7,$93,$FC,$63,$C0,$FC,$A1
	.dc.b	$E9,$E0,$B7,$FD,$F5,$FE,$31,$B7,$F8,$61,$BB,$7C,$5B,$FF,$0E,$35
	.dc.b	$4F,$F8,$7B,$5F,$88,$7E,$8F,$E7,$DF,$35,$8C,$7F,$D4,$91,$86,$C6
	.dc.b	$FE,$2B,$83,$FA,$9F,$02,$4F,$F8,$47,$E6,$EE,$5F,$2F,$2F,$F1,$6C
	.dc.b	$79,$89,$3F,$E2,$A7,$F8,$D6,$2F,$8D,$FD,$A7,$83,$E5,$6D,$79,$3C
	.dc.b	$BC,$34,$F3,$C7,$AF,$D0,$FD,$E7,$87,$67,$F8,$D3,$EF,$F8,$9C,$69
	.dc.b	$F7,$FC,$66,$69,$CF,$37,$F0,$EE,$67,$CD,$FE,$D1,$B9,$7F,$A5,$25
	.dc.b	$9E,$6B,$F8,$8E,$5F,$C9,$63,$FA,$73,$6B,$FD,$76,$D3,$F0,$91,$FC
	.dc.b	$22,$FF,$DC,$53,$FE,$C7,$7F,$D9,$6D,$E3,$F1,$7B,$1F,$C1,$BE,$7B
	.dc.b	$FD,$B3,$3F,$EC,$DF,$9F,$FF,$55,$3F,$EC,$F8,$CE,$3F,$ED,$69,$CB
	.dc.b	$FE,$D8,$8F,$D9,$67,$FC,$B1,$DB,$7A,$7B,$7C,$77,$F2,$7F,$12,$63
	.dc.b	$F8,$6F,$8A,$B8,$FC,$BE,$14,$7E,$34,$FC,$77,$C4,$DA,$FF,$5D,$C7
	.dc.b	$FB,$9D,$BF,$1F,$F2,$33,$FC,$14,$60,$F9,$DC,$DF,$2A,$33,$8F,$EA
	.dc.b	$E3,$07,$CE,$BF,$F5,$5D,$7D,$5E,$3F,$A7,$18,$AF,$DF,$1F,$D6,$63
	.dc.b	$F4,$5E,$0C,$AF,$D0,$FC,$AD,$AF,$63,$FA,$73,$F5,$1F,$2B,$67,$F4
	.dc.b	$FF,$13,$FD,$2F,$C6,$FD,$63,$63,$FB,$D6,$9F,$DA,$3A,$9E,$CF,$F7
	.dc.b	$D9,$E2,$68,$FF,$00,$46,$B2,$E7,$1F,$77,$E3,$7F,$A2,$F3,$9F,$C1
	.dc.b	$7A,$78,$6C,$FF,$8F,$BA,$3F,$AE,$DF,$F6,$E1,$E8,$23,$31,$FD,$CA
	.dc.b	$31,$7F,$D6,$A7,$FC,$0D,$3F,$87,$8F,$C8,$7A,$FD,$A3,$F5,$5C,$3F
	.dc.b	$6D,$BC,$BF,$DB,$F5,$FF,$7A,$65,$FE,$0B,$8C,$7C,$4C,$8F,$81,$5F
	.dc.b	$EB,$EA,$7C,$7E,$63,$F6,$D0,$F8,$AF,$EF,$1D,$BD,$5B,$7C,$B8,$7E
	.dc.b	$6F,$CA,$66,$31,$68,$C6,$1F,$A1,$F7,$3F,$54,$7E,$E3,$E7,$FE,$3F
	.dc.b	$2F,$E8,$E3,$1C,$3F,$BA,$9F,$8F,$5F,$E1,$86,$BF,$E3,$76,$F1,$FC
	.dc.b	$DF,$6F,$A7,$BB,$FB,$76,$E9,$FE,$2D,$F0,$BF,$C2,$47,$C7,$F8,$8B
	.dc.b	$1F,$E3,$1F,$43,$FC,$78,$47,$FC,$72,$FF,$90,$77,$7C,$87,$F8,$6F
	.dc.b	$D5,$FA,$0D,$AF,$F1,$D7,$9F,$FE,$1F,$3E,$A7,$C3,$D8,$F9,$1F,$22
	.dc.b	$F8,$3F,$F2,$63,$63,$FC,$C4,$DE,$1F,$C5,$73,$23,$11,$FE,$6E,$CF
	.dc.b	$FD,$7F,$5D,$3E,$73,$F0,$5F,$0D,$FE,$31,$B4,$5F,$5E,$57,$D6,$F8
	.dc.b	$7B,$8F,$F3,$5B,$FE,$46,$F8,$1A,$FF,$9F,$1A,$FE,$A1,$AB,$27,$CF
	.dc.b	$FF,$6E,$EA,$FD,$95,$7F,$2C,$AF,$E8,$53,$9C,$39,$7A,$1E,$7F,$2E
	.dc.b	$47,$E3,$67,$FC,$DA,$FC,$88,$F6,$79,$BF,$20,$91,$F2,$1F,$F3,$B4
	.dc.b	$FE,$32,$3C,$DE,$B7,$F9,$4E,$31,$87,$9F,$14,$FC,$7F,$8E,$97,$E1
	.dc.b	$F8,$0B,$7F,$17,$D6,$FA,$E4,$9F,$AE,$F8,$3D,$BF,$A1,$EC,$77,$6C
	.dc.b	$FE,$03,$E5,$7E,$A1,$E3,$F6,$AC,$7E,$13,$E5,$7F,$06,$47,$F0,$6F
	.dc.b	$B3,$CC,$FB,$9E,$B7,$AD,$B3,$F8,$CF,$5F,$F0,$07,$BF,$C3,$B1,$F2
	.dc.b	$FC,$2F,$8A,$47,$FD,$61,$C6,$8F,$F4,$F7,$CD,$7E,$47,$A3,$F9,$54
	.dc.b	$BC,$3F,$F3,$E4,$E6,$FF,$94,$CE,$5F,$CE,$FC,$D7,$E4,$4D,$F7,$58
	.dc.b	$FE,$11,$5F,$94,$7F,$D8,$0D,$E4,$FC,$FB,$5F,$F6,$5B,$5F,$C8,$7A
	.dc.b	$9F,$9D,$3C,$7E,$E5,$AE,$AE,$1E,$FD,$B6,$BF,$3D,$ED,$7E,$94,$F5
	.dc.b	$F6,$DA,$F2,$B8,$7C,$FD,$96,$BF,$AD,$9F,$EC,$7F,$9F,$FF,$64,$7C
	.dc.b	$FF,$F2,$78,$FF,$70,$D6,$3D,$2F,$E8,$50,$AC,$3A,$15,$C5,$FF,$5A
	.dc.b	$C6,$9F,$EA,$B5,$5F,$01,$C2,$BF,$82,$C6,$63,$B1,$FC,$02,$32,$D3
	.dc.b	$FA,$A4,$FF,$B7,$F6,$F0,$47,$EE,$6B,$15,$E8,$AF,$07,$85,$FE,$D2
	.dc.b	$C3,$77,$D9,$E3,$19,$AE,$13,$E5,$3D,$D3,$C2,$B9,$DF,$9D,$FF,$1A
	.dc.b	$CF,$B0,$FF,$1B,$B8,$7D,$67,$F6,$BF,$9F,$FF,$20,$FC,$3F,$F8,$F1
	.dc.b	$B1,$F3,$9A,$EF,$C0,$4E,$FC,$2F,$F7,$54,$ED,$AF,$56,$BF,$EE,$6A
	.dc.b	$46,$20,$23,$3C,$FF,$5D,$FE,$17,$17,$6F,$F7,$7E,$7F,$76,$77,$3F
	.dc.b	$14,$E5,$7E,$49,$D5,$70,$D3,$C4,$DA,$F0,$DF,$2A,$E9,$7E,$CD,$B8
	.dc.b	$D7,$DA,$37,$7F,$F0,$02,$37,$8E,$67,$C2,$F9,$D8,$79,$70,$ED,$8F
	.dc.b	$C3,$C3,$CC,$F9,$27,$78,$F4,$7F,$01,$B3,$E7,$BF,$C1,$B1,$FE,$07
	.dc.b	$6B,$FE,$83,$6E,$3F,$E3,$75,$F3,$1F,$3A,$75,$EB,$5A,$D3,$78,$7F
	.dc.b	$C7,$37,$FE,$0F,$AE,$DE,$C6,$5F,$EA,$9F,$3E,$F8,$1C,$66,$87,$C0
	.dc.b	$61,$B5,$FE,$1D,$78,$36,$3A,$7C,$33,$F1,$2F,$F5,$DC,$FF,$8B,$23
	.dc.b	$1E,$7F,$3A,$EF,$E6,$BD,$FF,$F4,$9E,$5F,$DE,$DB,$19,$C3,$FC,$03
	.dc.b	$1F,$E9,$8C,$1F,$8E,$BC,$CF,$0B,$6B,$C0,$D7,$FB,$D5,$78,$35,$F0
	.dc.b	$E5,$27,$FC,$23,$1F,$0B,$F0,$9F,$D8,$E6,$69,$27,$CE,$FF,$7D,$7F
	.dc.b	$C8,$0D,$7F,$D7,$8D,$E3,$71,$7E,$9B,$4F,$4F,$97,$DF,$E0,$7D,$3E
	.dc.b	$24,$C7,$9C,$78,$F9,$3B,$92,$3F,$94,$E7,$E5,$8F,$1F,$3F,$A4,$93
	.dc.b	$E4,$E7,$F6,$74,$FC,$6E,$09,$3E,$88,$F8,$1F,$43,$EC,$24,$76,$E0
	.dc.b	$47,$C0,$F9,$4F,$8C,$49,$F8,$3E,$E7,$C9,$FA,$19,$1F,$11,$F9,$9F
	.dc.b	$3B,$F3,$7D,$EF,$92,$49,$EC,$FC,$CF,$9F,$FA,$5E,$F4,$9F,$9B,$1F
	.dc.b	$57,$E7,$64,$7D,$07,$EC,$7C,$B7,$D9,$F7,$BE,$A9,$27,$D4,$FD,$8F
	.dc.b	$7F,$EF,$7B,$F2,$7E,$CC,$7E,$07,$E6,$E4,$7D,$C7,$F0,$9F,$31,$FA
	.dc.b	$6F,$C0,$BF,$DD,$7D,$26,$93,$BB,$F3,$6B,$E1,$7F,$4B,$F8,$AA,$E9
	.dc.b	$CE,$3A,$7C,$5F,$EA,$BE,$2F,$86,$7F,$A9,$6A,$F0,$FF,$D7,$58,$AB
	.dc.b	$C9,$F1,$BE,$02,$57,$E1,$A5,$1E,$4F,$8A,$ED,$84,$AF,$F4,$C5,$27
	.dc.b	$DD,$3F,$C1,$FE,$2D,$FB,$8F,$F1,$4D,$BF,$9F,$C6,$3C,$3D,$E3,$FE
	.dc.b	$AA,$1F,$B2,$FB,$DF,$EF,$E6,$37,$FE,$BB,$19,$8F,$F0,$1B,$5F,$F6
	.dc.b	$32,$E2,$FE,$31,$1D,$1B,$19,$8F,$03,$CF,$7F,$8B,$F3,$9F,$5C,$93
	.dc.b	$EE,$F8,$BF,$9C,$47,$E7,$3A,$7E,$76,$7F,$EE,$2D,$DF,$35,$F8,$12
	.dc.b	$4F,$C8,$F9,$AA,$F0,$1E,$63,$85,$3F,$74,$DE,$7F,$C1,$2B,$FD,$C7
	.dc.b	$C3,$FC,$1F,$2F,$9A,$FF,$B9,$E7,$5C,$3F,$4A,$BE,$A4,$ED,$E2,$7F
	.dc.b	$3D,$9F,$E6,$AF,$FB,$71,$AF,$96,$79,$39,$F3,$7B,$5F,$DC,$FC,$75
	.dc.b	$F1,$FF,$81,$6D,$6C,$FD,$17,$C7,$7E,$5F,$E6,$A4,$7E,$2A,$2B,$F0
	.dc.b	$F6,$F8,$53,$8C,$FF,$33,$5F,$AB,$89,$5F,$92,$7F,$5D,$F1,$FF,$A3
	.dc.b	$F9,$99,$3F,$A3,$76,$FC,$FE,$EF,$D9,$47,$E1,$BD,$7F,$83,$8F,$CD
	.dc.b	$AF,$F6,$36,$DF,$B9,$AF,$C5,$FD,$2F,$F8,$52,$7F,$CA,$6D,$7F,$B7
	.dc.b	$07,$0F,$F7,$16,$BF,$E2,$05,$F0,$BD,$09,$1F,$A3,$8E,$D7,$FB,$8F
	.dc.b	$99,$FE,$14,$49,$FD,$8F,$C8,$D7,$E4,$73,$23,$F8,$21,$9A,$FC,$1F
	.dc.b	$EE,$2F,$67,$F0,$1F,$BB,$7F,$5F,$9F,$C6,$07,$1C,$DF,$B8,$AE,$18
	.dc.b	$FE,$16,$72,$FF,$0D,$6D,$71,$F7,$7F,$81,$57,$0A,$FA,$1D,$8F,$99
	.dc.b	$CF,$C3,$C9,$FE,$17,$1C,$38,$3F,$80,$F5,$B0,$F8,$F5,$FE,$82,$8C
	.dc.b	$4E,$A7,$33,$F5,$3E,$37,$EB,$91,$FF,$27,$E7,$F0,$AB,$F9,$A7,$C9
	.dc.b	$FF,$29,$24,$FF,$1C,$6B,$F3,$1B,$3F,$CA,$9F,$E7,$9F,$27,$FC,$C7
	.dc.b	$E4,$E5,$7F,$24,$8A,$FD,$48,$B5,$FC,$CA,$7F,$A2,$47,$F4,$02,$4F
	.dc.b	$F3,$17,$FC,$F7,$F1,$2F,$13,$F8,$DB,$67,$F9,$43,$5F,$9B,$3F,$D3
	.dc.b	$FE,$63,$FC,$6B,$6F,$98,$A6,$6B,$E1,$F2,$7F,$9C,$BB,$7F,$DB,$13
	.dc.b	$FE,$34,$FC,$85,$F1,$D4,$E3,$7C,$05,$7F,$B7,$E3,$FC,$67,$87,$09
	.dc.b	$FF,$07,$B5,$FD,$2C,$ED,$FF,$5D,$FC,$7B,$F0,$5F,$E6,$91,$FC,$8E
	.dc.b	$31,$E6,$FF,$94,$5B,$1E,$1A,$E4,$FE,$EE,$FF,$C9,$2F,$66,$BF,$E6
	.dc.b	$18,$FF,$7A,$7C,$5F,$98,$FF,$22,$64,$F9,$4F,$80,$4A,$ED,$7D,$38
	.dc.b	$BF,$98,$61,$FE,$FD,$E5,$F9,$A4,$AD,$DF,$D0,$F2,$A4,$72,$74,$F9
	.dc.b	$F9,$BE,$39,$1F,$8E,$F9,$39,$3E,$91,$23,$C7,$1F,$73,$AB,$EE,$CC
	.dc.b	$9F,$3C,$7E,$03,$E4,$BE,$99,$23,$E9,$BF,$5B,$E6,$BF,$0A,$48,$F7
	.dc.b	$C7,$F7,$9F,$A3,$7B,$D7,$E0,$25,$7E,$07,$B9,$5A,$8C,$2F,$80,$FD
	.dc.b	$CB,$63,$F7,$0D,$7F,$93,$CF,$29,$FC,$4F,$85,$F9,$82,$4F,$7F,$C0
	.dc.b	$3F,$B1,$BF,$53,$E7,$3F,$3A,$47,$EB,$F8,$DF,$A0,$F9,$CF,$D0,$FC
	.dc.b	$DC,$9F,$B7,$1E,$DF,$86,$F2,$92,$3C,$AF,$E6,$BE,$47,$F5,$24,$8F
	.dc.b	$C1,$FC,$D7,$EB,$3E,$4F,$F3,$E4,$8F,$CF,$BF,$A2,$F4,$A4,$7E,$89
	.dc.b	$AF,$F4,$61,$F4,$DF,$E2,$9F,$17,$FC,$3E,$D8,$D3,$5F,$E9,$AD,$7F
	.dc.b	$D7,$8F,$F8,$DB,$B7,$8F,$C1,$7F,$A2,$E7,$2D,$7F,$DD,$F3,$B7,$F4
	.dc.b	$DF,$23,$FD,$7D,$FE,$05,$E0,$FF,$AE,$BC,$8F,$DC,$CE,$67,$F7,$F3
	.dc.b	$97,$F7,$7F,$BA,$7F,$BE,$AB,$FD,$1A,$D8,$C9,$C7,$14,$66,$3F,$68
	.dc.b	$FF,$0A,$9F,$DE,$36,$3F,$BB,$2F,$95,$FE,$AB,$3F,$C5,$E3,$5E,$27
	.dc.b	$F8,$7B,$9B,$F9,$34,$FF,$45,$6D,$FE,$39,$AF,$FB,$8F,$A5,$18,$8E
	.dc.b	$9F,$09,$D9,$FE,$3B,$87,$FB,$AA,$FC,$B3,$9C,$BD,$7F,$2B,$69,$F2
	.dc.b	$B4,$C7,$CA,$DC,$3D,$BF,$29,$B3,$B6,$8F,$8F,$7C,$86,$6B,$F4,$1F
	.dc.b	$17,$F4,$6B,$CE,$DD,$FF,$CE,$13,$EF,$8C,$7C,$67,$49,$FF,$3C,$CF
	.dc.b	$EA,$63,$1C,$DE,$7F,$94,$93,$F4,$3E,$52,$F1,$FD,$07,$EC,$70,$1A
	.dc.b	$49,$FB,$FF,$29,$F0,$7B,$FE,$13,$63,$FD,$F5,$E8,$BC,$BF,$99,$F1
	.dc.b	$08,$C4,$FE,$03,$E5,$33,$18,$CF,$F0,$31,$E2,$24,$FE,$39,$F2,$F9
	.dc.b	$3F,$44,$FA,$5E,$5D,$9F,$C5,$B8,$7F,$1B,$E6,$5B,$FF,$27,$60,$F1
	.dc.b	$FC,$27,$DB,$E6,$B7,$F9,$47,$E1,$91,$FC,$FB,$FA,$5F,$55,$F8,$CF
	.dc.b	$E9,$7C,$0F,$F3,$B4,$63,$C3,$FC,$37,$5F,$E3,$63,$18,$7E,$82,$3E
	.dc.b	$79,$1F,$D5,$BF,$4B,$E5,$3E,$2F,$93,$F8,$F6,$CF,$EA,$3E,$57,$0D
	.dc.b	$3F,$57,$1F,$33,$C9,$91,$FA,$EF,$66,$F0,$7E,$B9,$FB,$E4,$8F,$DD
	.dc.b	$BF,$C0,$79,$5F,$8E,$7E,$F7,$6F,$F4,$94,$62,$FF,$45,$F2,$9F,$7A
	.dc.b	$B5,$87,$ED,$E3,$13,$A8,$FF,$05,$57,$16,$D7,$EF,$DB,$E4,$FE,$BF
	.dc.b	$F7,$54,$75,$8C,$34,$FF,$02,$46,$9E,$27,$FB,$37,$C4,$FD,$AB,$7C
	.dc.b	$7F,$D8,$FC,$23,$D4,$FE,$01,$9A,$F6,$FF,$A8,$67,$57,$FE,$51,$6C
	.dc.b	$7F,$5A,$6B,$FC,$35,$AF,$F8,$B7,$0F,$F7,$E4,$FF,$69,$9F,$CF,$6B
	.dc.b	$FE,$0C,$9C,$D7,$7E,$EF,$E2,$52,$FD,$DD,$F1,$F8,$6E,$3C,$FD,$DC
	.dc.b	$4F,$B1,$B1,$93,$FD,$79,$3C,$30,$FF,$51,$62,$F2,$3E,$C7,$9B,$F3
	.dc.b	$DB,$7F,$31,$C3,$F3,$BC,$9B,$7F,$ED,$93,$E4,$70,$FC,$AD,$FE,$D4
	.dc.b	$65,$FE,$6F,$5F,$CC,$67,$5E,$77,$F4,$DA,$FA,$0D,$AF,$E8,$51,$FE
	.dc.b	$06,$8F,$E9,$59,$7F,$20,$8F,$07,$AC,$F8,$5F,$C3,$79,$DF,$6A,$31
	.dc.b	$97,$E3,$3D,$0F,$CA,$1F,$8B,$F8,$AB,$1F,$88,$70,$FE,$2B,$E5,$6E
	.dc.b	$3F,$ED,$3E,$97,$A1,$F9,$D3,$F2,$9F,$0F,$6F,$F2,$2D,$AF,$9A,$DE
	.dc.b	$7F,$F5,$6F,$2F,$CF,$AF,$F1,$57,$C9,$7F,$07,$7F,$D2,$B1,$9B,$FF
	.dc.b	$78,$4E,$5F,$CA,$7D,$DF,$FA,$A6,$BE,$DF,$37,$F4,$A8,$C4,$B0,$EF
	.dc.b	$39,$E1,$D7,$5D,$C3,$FC,$75,$87,$0B,$FC,$97,$CB,$7E,$E7,$D1,$FD
	.dc.b	$83,$6B,$F5,$AE,$1F,$D7,$FC,$B5,$CF,$E3,$7A,$5F,$C0,$4F,$DC,$7C
	.dc.b	$AD,$AF,$DA,$B8,$7F,$6F,$DF,$6B,$F1,$5F,$49,$FD,$E3,$3F,$F7,$93
	.dc.b	$69,$32,$1C,$BF,$B8,$75,$19,$8F,$C6,$46,$A7,$FC,$61,$28,$C3,$FE
	.dc.b	$33,$6B,$FE,$87,$EA,$7C,$B7,$CD,$F9,$3F,$E2,$4D,$BF,$E1,$FF,$2B
	.dc.b	$FC,$7F,$D6,$FE,$2C,$D8,$FE,$27,$5F,$98,$23,$FE,$41,$EF,$60,$9F
	.dc.b	$CE,$FC,$A7,$F2,$9F,$37,$F9,$03,$6F,$F8,$EF,$37,$F2,$DF,$33,$F9
	.dc.b	$23,$63,$F9,$0F,$47,$E9,$88,$F7,$C7,$F4,$7B,$7F,$AC,$97,$BF,$FC
	.dc.b	$B4,$FE,$5D,$F2,$36,$BF,$95,$FC,$8F,$F8,$7F,$F7,$7F,$F6,$54,$60
	.dc.b	$FF,$59,$38,$C3,$F2,$11,$FE,$CB,$8F,$F6,$4B,$EF,$6C,$7F,$B0,$9A
	.dc.b	$FF,$AD,$5A,$F1,$F9,$0F,$01,$E1,$F9,$0B,$5E,$0F,$37,$C9,$F2,$1E
	.dc.b	$56,$C7,$93,$E4,$3F,$CA,$18,$9A,$48,$F2,$7C,$7F,$B4,$F7,$7C,$85
	.dc.b	$8F,$6F,$C8,$7C,$13,$E1,$7A,$76,$3E,$0F,$A9,$D8,$8F,$BD,$F0,$FC
	.dc.b	$7F,$F8,$83,$E5,$9C,$3F,$2E,$30,$FC,$DF,$46,$5F,$CC,$F8,$2B,$C7
	.dc.b	$F3,$0F,$F7,$7E,$EF,$90,$F6,$92,$BE,$73,$F0,$3E,$43,$FC,$5F,$EA
	.dc.b	$39,$7E,$D4,$61,$FB,$9E,$FC,$BF,$EE,$74,$E6,$C6,$5E,$1F,$05,$9B
	.dc.b	$E4,$3E,$69,$30,$FB,$CF,$D0,$F7,$FF,$11,$F0,$7F,$89,$6C,$F6,$E7
	.dc.b	$F2,$9F,$E6,$BF,$81,$FF,$03,$E4,$F8,$06,$AA,$D4,$FF,$A2,$EB,$FD
	.dc.b	$14,$FF,$04,$D1,$18,$84,$61,$E0,$FC,$4E,$8F,$03,$71,$FE,$1B,$E6
	.dc.b	$7F,$BF,$AF,$FA,$CB,$67,$CD,$EE,$F0,$AD,$9E,$1F,$A4,$F9,$CE,$31
	.dc.b	$F0,$C6,$1F,$94,$BB,$FF,$D1,$4E,$70,$FF,$0E,$CF,$09,$C9,$AE,$2B
	.dc.b	$7D,$67,$F9,$18,$FA,$4F,$E4,$BF,$47,$F2,$23,$39,$F4,$8C,$47,$09
	.dc.b	$DE,$FC,$A3,$FD,$61,$3A,$7F,$97,$CE,$9E,$77,$5E,$42,$78,$5F,$7B
	.dc.b	$FF,$6F,$5F,$F3,$8F,$96,$CE,$16,$73,$82,$72,$D8,$CE,$18,$3C,$91
	.dc.b	$EF,$3C,$11,$E9,$C3,$C1,$1E,$BC,$3C,$CF,$B2,$79,$9B,$C3,$31,$CC
	.dc.b	$FE,$8B,$3F,$D1,$9F,$82,$57,$FA,$4F,$D2,$3C,$17,$C4,$61,$B5,$FE
	.dc.b	$FD,$9D,$4E,$4E,$3C,$7F,$DD,$67,$4D,$7E,$94,$FE,$7F,$7D,$92,$3D
	.dc.b	$E7,$F6,$28,$78,$7F,$82,$D7,$EE,$C6,$27,$EB,$3C,$DE,$21,$C4,$70
	.dc.b	$9E,$5A,$7E,$32,$79,$57,$07,$71,$9C,$3D,$5B,$71,$AF,$B8,$79,$6F
	.dc.b	$F2,$B1,$E4,$9D,$D7,$1C,$7C,$B5,$E0,$BE,$07,$A2,$BE,$36,$3C,$6B
	.dc.b	$73,$0F,$24,$EA,$FF,$81,$7D,$27,$FB,$CA,$7F,$65,$BA,$31,$1F,$C6
	.dc.b	$DB,$7F,$B0,$9D,$7C,$A7,$EB,$23,$1C,$7F,$7E,$F9,$B5,$FD,$DE,$FB
	.dc.b	$E1,$F0,$FC,$45,$9F,$E2,$3F,$48,$EB,$C7,$53,$E9,$1E,$56,$1B,$7E
	.dc.b	$8B,$E0,$FF,$A9,$A3,$2F,$ED,$EB,$95,$7F,$87,$1B,$1F,$BB,$3F,$95
	.dc.b	$4F,$07,$F8,$3C,$3E,$93,$FD,$41,$1B,$8C,$79,$5C,$E7,$33,$DD,$AF
	.dc.b	$F0,$CE,$D7,$07,$BF,$87,$F9,$8B,$63,$33,$FE,$BE,$AE,$7C,$3F,$3F
	.dc.b	$E7,$FD,$B3,$FD,$37,$E9,$3F,$89,$4F,$D3,$6F,$0F,$F3,$DA,$DC,$FF
	.dc.b	$B4,$2D,$E4,$F9,$C9,$87,$AE,$32,$FF,$8B,$F1,$7C,$6F,$94,$91,$F8
	.dc.b	$6C,$BF,$A3,$BE,$7F,$23,$D5,$C9,$23,$07,$AF,$C5,$F8,$C4,$7C,$0F
	.dc.b	$C9,$F1,$7E,$FE,$46,$9F,$1F,$0F,$1D,$7C,$5E,$6F,$91,$D5,$23,$07
	.dc.b	$FB,$9B,$E9,$3C,$AF,$E0,$5A,$FD,$C6,$BF,$E6,$6E,$54,$62,$3F,$B6
	.dc.b	$3F,$51,$FF,$06,$46,$B8,$FF,$2B,$39,$E1,$FA,$5F,$13,$F6,$FE,$6B
	.dc.b	$FC,$D3,$CD,$FD,$B9,$BC,$66,$9F,$93,$BC,$65,$F8,$D6,$BE,$98,$D1
	.dc.b	$D3,$57,$B9,$49,$3F,$E6,$78,$CB,$F6,$F1,$F0,$91,$C1,$DF,$E5,$BC
	.dc.b	$44,$70,$73,$C7,$C0,$47,$07,$9F,$E4,$BD,$44,$70,$7A,$E3,$DC,$47
	.dc.b	$07,$BF,$E4,$BE,$11,$1C,$1F,$AF,$F8,$F7,$D2,$7E,$E5,$B1,$98,$C1
	.dc.b	$9F,$93,$91,$C5,$AF,$F0,$39,$1C,$C8,$F7,$7F,$D4,$3F,$26,$FA,$3F
	.dc.b	$BD,$DF,$FE,$0A,$FA,$46,$5F,$C6,$70,$46,$1F,$F0,$E5,$F1,$E5,$D4
	.dc.b	$BE,$07,$FB,$43,$6B,$FD,$5E,$D7,$F8,$A3,$5F,$F0,$CD,$78,$FD,$2F
	.dc.b	$E5,$0D,$8E,$F9,$27,$BD,$FF,$2A,$CD,$F2,$74,$97,$FD,$03,$E8,$FF
	.dc.b	$D8,$AF,$F4,$2E,$07,$CE,$7F,$15,$24,$7F,$8D,$FA,$FE,$4D,$FF,$1B
	.dc.b	$B7,$C8,$77,$F9,$09,$7D,$62,$BF,$E6,$5C,$5F,$23,$FC,$84,$CF,$77
	.dc.b	$F8,$F3,$83,$E8,$57,$F4,$58,$FE,$08,$DF,$07,$F8,$FA,$9D,$BF,$E4
	.dc.b	$B9,$DB,$FC,$F5,$AF,$F2,$89,$FF,$0C,$46,$23,$E6,$FC,$A7,$F9,$ED
	.dc.b	$FB,$2D,$7F,$D1,$94,$8C,$4F,$F9,$48,$EA,$7F,$95,$35,$FC,$6F,$C8
	.dc.b	$7D,$12,$5F,$CF,$F9,$1F,$E8,$1F,$39,$23,$E8,$3A,$7F,$D1,$5B,$AB
	.dc.b	$ED,$7C,$7C,$9F,$B1,$15,$F9,$9F,$27,$FD,$2C,$C6,$5F,$E6,$68,$79
	.dc.b	$DF,$12,$31,$F0,$FF,$E1,$86,$FB,$3F,$9B,$9D,$77,$7F,$9A,$2B,$FD
	.dc.b	$89,$7F,$D9,$C3,$87,$FB,$53,$5F,$F3,$33,$5F,$F3,$8F,$ED,$9F,$E8
	.dc.b	$D7,$FD,$6C,$1C,$61,$E7,$6D,$78,$7E,$27,$F1,$DF,$23,$F9,$BF,$90
	.dc.b	$98,$7E,$66,$2B,$F8,$DF,$1B,$FB,$61,$8C,$7E,$B7,$D2,$FF,$6C,$C7
	.dc.b	$FD,$32,$DA,$4A,$31,$CC,$FA,$5F,$EC,$B5,$F8,$25,$F1,$BF,$B2,$6C
	.dc.b	$7F,$00,$6B,$FE,$D3,$BE,$9E,$C6,$F8,$78,$6D,$2B,$84,$7F,$0B,$E0
	.dc.b	$FF,$69,$47,$CF,$6B,$F3,$BE,$2B,$FC,$E1,$18,$8F,$E0,$B8,$FE,$53
	.dc.b	$54,$7E,$96,$30,$DF,$86,$ED,$7F,$C1,$7E,$46,$4F,$EF,$7C,$44,$7E
	.dc.b	$F4,$FF,$0B,$7A,$12,$7F,$83,$3C,$1F,$EC,$B2,$F8,$FF,$E2,$04,$8F
	.dc.b	$E1,$DF,$21,$FE,$1F,$F9,$F9,$1F,$C3,$DD,$3F,$84,$F8,$5F,$E2,$EF
	.dc.b	$E3,$DA,$FF,$8D,$DA,$FF,$AC,$DA,$FF,$B7,$7C,$57,$CF,$7F,$8B,$49
	.dc.b	$1F,$EC,$39,$E7,$C9,$FE,$5B,$C1,$F4,$7F,$E5,$0F,$A0,$FF,$57,$5F
	.dc.b	$F1,$F6,$BF,$EF,$D9,$DB,$9F,$8A,$FF,$26,$4E,$A3,$FC,$92,$D7,$FC
	.dc.b	$1C,$D7,$FD,$47,$F4,$BF,$CF,$5F,$AF,$F1,$A8,$C4,$6B,$D0,$FD,$84
	.dc.b	$E5,$FF,$05,$FB,$1F,$E4,$92,$4F,$FB,$AE,$74,$FA,$7E,$3F,$CA,$47
	.dc.b	$FC,$B5,$87,$FB,$B5,$FF,$20,$7C,$C7,$F9,$4C,$8F,$13,$FD,$0B,$9F
	.dc.b	$E2,$BD,$F3,$49,$CF,$F2,$7B,$65,$7C,$97,$FC,$C1,$3F,$92,$F3,$58
	.dc.b	$B8,$BE,$49,$2F,$CA,$FF,$8A,$39,$3F,$CC,$C4,$8F,$A9,$1D,$0E,$BF
	.dc.b	$17,$23,$CF,$1E,$0F,$91,$FF,$19,$E2,$69,$27,$FC,$C1,$C3,$FE,$73
	.dc.b	$23,$F0,$23,$FB,$7F,$D2,$FF,$8B,$23,$FA,$1A,$F1,$B2,$6B,$B3,$FB
	.dc.b	$AF,$D2,$3D,$6F,$F0,$5B,$6F,$FD,$21,$E3,$A3,$11,$FB,$83,$F7,$31
	.dc.b	$99,$F4,$FC,$5F,$CA,$24,$FD,$4F,$1F,$FD,$15,$EF,$48,$FE,$F4,$FD
	.dc.b	$0F,$27,$E8,$FC,$B7,$AB,$D0,$95,$F7,$23,$CF,$EC,$F9,$89,$1E,$6D
	.dc.b	$3F,$BE,$BF,$EA,$7E,$A7,$E2,$3F,$D3,$7C,$3F,$F6,$23,$7C,$BF,$E2
	.dc.b	$33,$A7,$F0,$3F,$19,$FE,$A1,$9F,$F1,$CD,$FB,$C5,$74,$F8,$3F,$F3
	.dc.b	$4C,$EA,$7E,$24,$E5,$F1,$43,$E8,$36,$48,$CB,$FD,$5D,$7D,$33,$E9
	.dc.b	$3F,$73,$1E,$18,$C7,$93,$FD,$75,$B1,$FE,$C5,$6B,$BF,$86,$FF,$43
	.dc.b	$41,$A4,$C7,$CA,$FF,$71,$9F,$87,$8E,$F4,$C9,$23,$E1,$57,$CC,$C7
	.dc.b	$84,$7C,$4A,$F4,$57,$C0,$9C,$D6,$E3,$E1,$F9,$DF,$E1,$37,$BB,$6F
	.dc.b	$E6,$FD,$37,$CE,$9F,$F7,$FE,$1F,$E6,$27,$FC,$AF,$E5,$0F,$0F,$55
	.dc.b	$69,$98,$FC,$CB,$6F,$D3,$7F,$EC,$89,$FF,$30,$F1,$75,$9F,$ED,$B1
	.dc.b	$F8,$49,$D4,$7F,$8B,$27,$5C,$5F,$8A,$FA,$2F,$46,$9F,$8E,$6D,$BE
	.dc.b	$47,$3A,$27,$3F,$1D,$33,$FE,$C3,$89,$F0,$7C,$6E,$74,$FF,$1D,$EB
	.dc.b	$FD,$1A,$7D,$7E,$8F,$B8,$8F,$E8,$78,$BE,$57,$07,$ED,$F8,$AF,$0F
	.dc.b	$DB,$C7,$D5,$1E,$4F,$86,$98,$78,$8F,$81,$F1,$98,$AF,$86,$7D,$47
	.dc.b	$ED,$F1,$1A,$49,$FD,$84,$3E,$0F,$E1,$FC,$47,$C4,$6C,$FD,$CE,$CF
	.dc.b	$8A,$7E,$07,$CE,$FE,$5F,$E4,$62,$7F,$AE,$3F,$8E,$8C,$C7,$F4,$48
	.dc.b	$C7,$9D,$FC,$9E,$3F,$A3,$F8,$5C,$E3,$FC,$43,$1E,$DD,$3F,$1B,$E4
	.dc.b	$66,$BE,$BB,$27,$5D,$6D,$7F,$DB,$31,$87,$F9,$53,$F9,$4F,$1F,$EE
	.dc.b	$9D,$FE,$76,$C7,$F7,$37,$CA,$47,$FA,$DB,$FA,$4E,$C7,$C3,$7E,$18
	.dc.b	$FD,$17,$AA,$FA,$67,$C8,$E7,$0F,$ED,$11,$87,$F8,$A6,$CA,$E1,$5F
	.dc.b	$B4,$6C,$E9,$B3,$FC,$6F,$87,$FA,$EF,$CF,$7C,$A2,$57,$F7,$07,$EF
	.dc.b	$FA,$FF,$2D,$FC,$EB,$63,$F6,$BF,$0F,$F9,$E3,$E8,$7C,$5D,$8F,$F0
	.dc.b	$A3,$F5,$88,$FF,$7C,$7E,$17,$D1,$7E,$98,$FC,$17,$D3,$7F,$91,$27
	.dc.b	$F8,$0C,$63,$29,$1F,$4F,$D0,$B1,$F9,$6E,$4F,$E7,$FF,$17,$FB,$03
	.dc.b	$9F,$C7,$DB,$FF,$62,$63,$FE,$D1,$6C,$7E,$EB,$E8,$BF,$65,$18,$3F
	.dc.b	$DC,$BF,$47,$F8,$AC,$7F,$D5,$AD,$F2,$7F,$D3,$1E,$2F,$EF,$EB,$7D
	.dc.b	$9F,$C6,$EF,$FD,$81,$E6,$7F,$26,$6C,$FF,$BE,$23,$15,$C6,$7F,$A7
	.dc.b	$2C,$3F,$39,$1E,$0F,$57,$F7,$B1,$8C,$7F,$DA,$4A,$33,$F0,$D3,$1F
	.dc.b	$F6,$77,$D3,$7F,$AD,$5F,$5C,$7F,$2F,$B4,$E7,$0F,$F4,$13,$63,$AF
	.dc.b	$C1,$6E,$31,$7F,$E7,$8E,$87,$07,$F2,$6B,$FE,$DB,$F1,$9F,$6B,$E6
	.dc.b	$BF,$90,$BF,$37,$D6,$C6,$1F,$AE,$E2,$FE,$C8,$FD,$4A,$F9,$1E,$DF
	.dc.b	$F9,$D6,$3F,$71,$AA,$30,$FF,$AD,$38,$7E,$D3,$FE,$B6,$8C,$E5,$F5
	.dc.b	$39,$3E,$69,$2F,$FD,$89,$7C,$2F,$FB,$E5,$FF,$A2,$A7,$BC,$FF,$51
	.dc.b	$9E,$9F,$13,$F6,$E3,$18,$7F,$BA,$75,$7A,$9F,$CF,$BE,$3B,$F3,$2D
	.dc.b	$C7,$DD,$EB,$E1,$A7,$E4,$A3,$FA,$2F,$34,$8F,$CA,$F9,$F7,$4F,$E6
	.dc.b	$3C,$7F,$3A,$3F,$4A,$7E,$5F,$D0,$B5,$FC,$77,$5F,$F8,$01,$E2,$7F
	.dc.b	$42,$F9,$2F,$E0,$D8,$FD,$CE,$5C,$4F,$DB,$7F,$BC,$F0,$7E,$09,$BC
	.dc.b	$FF,$52,$BF,$C4,$13,$9C,$7F,$C8,$7F,$4D,$F9,$57,$F1,$B7,$FD,$42
	.dc.b	$31,$5D,$E7,$F5,$FF,$29,$FD,$8B,$93,$F7,$AD,$8F,$C1,$7A,$38,$69
	.dc.b	$C0,$FF,$00,$FC,$27,$F6,$A3,$F8,$2F,$C0,$D8,$FD,$DB,$87,$F1,$9E
	.dc.b	$65,$C7,$FA,$CF,$03,$37,$3F,$9C,$FD,$B7,$D3,$38,$F1,$8F,$E4,$23
	.dc.b	$0F,$F2,$EC,$11,$87,$F9,$8B,$5F,$F2,$33,$5F,$F3,$2D,$BB,$BB,$FB
	.dc.b	$BF,$1D,$B7,$FC,$53,$E0,$FF,$AF,$76,$7F,$50,$C3,$F9,$B3,$5F,$F3
	.dc.b	$87,$4F,$EC,$27,$7E,$E4,$BF,$E8,$73,$FB,$48,$C5,$72,$F3,$7A,$92
	.dc.b	$B7,$82,$3C,$3F,$0B,$27,$2D,$AF,$E9,$1F,$21,$FC,$99,$B3,$FD,$D6
	.dc.b	$31,$1E,$66,$BF,$C1,$DF,$C6,$39,$E8,$E0,$49,$FF,$3C,$7D,$2F,$FB
	.dc.b	$79,$BF,$07,$FE,$4A,$8C,$57,$F9,$49,$AF,$7D,$FC,$44,$AF,$87,$87
	.dc.b	$F8,$C5,$FF,$39,$FC,$87,$F7,$12,$38,$3C,$3E,$4F,$72,$38,$39,$FA
	.dc.b	$1E,$F2,$38,$3F,$CF,$FF,$4D,$FD,$F4,$8E,$0F,$1F,$89,$FE,$1F,$C3
	.dc.b	$F4,$FE,$07,$F9,$07,$3F,$F4,$6E,$8F,$8B,$D5,$EF,$BB,$E7,$13,$1F
	.dc.b	$6B,$F4,$72,$FF,$29,$FD,$FF,$C6,$CB,$A7,$33,$45,$6B,$87,$FC,$61
	.dc.b	$E3,$BE,$46,$9F,$E5,$CD,$BF,$98,$69,$FE,$29,$C7,$F8,$FC,$F0,$FA
	.dc.b	$1F,$F1,$A5,$9A,$4A,$FC,$9C,$3E,$47,$CB,$5E,$06,$CF,$F8,$C3,$5E
	.dc.b	$4D,$BF,$17,$B3,$EC,$7C,$FC,$9E,$02,$4F,$67,$F0,$DE,$4F,$A7,$E8
	.dc.b	$3D,$4D,$9E,$D8,$FC,$63,$E0,$FC,$85,$8F,$76,$9F,$EF,$F7,$D3,$1E
	.dc.b	$E2,$3E,$CF,$84,$F8,$24,$70,$7F,$B4,$7E,$47,$FA,$3B,$FD,$9E,$BF
	.dc.b	$BC,$3F,$ED,$08,$CF,$43,$83,$FA,$5D,$FF,$A3,$3E,$65,$F2,$14,$FF
	.dc.b	$47,$FD,$1F,$CB,$C7,$FC,$AF,$3A,$7F,$CB,$31,$9C,$3F,$C4,$8D,$E8
	.dc.b	$FC,$77,$BA,$87,$B3,$D3,$7F,$F0,$C5,$87,$D3,$DF,$A6,$FE,$BF,$1F
	.dc.b	$ED,$EC,$B1,$C1,$FE,$52,$AB,$39,$15,$C1,$E8,$74,$33,$3F,$EC,$CA
	.dc.b	$E3,$1A,$AA,$FF,$73,$AE,$4F,$96,$3C,$DD,$18,$CF,$F5,$4E,$A7,$B3
	.dc.b	$9B,$DE,$1E,$79,$F0,$9E,$7C,$38,$61,$E0,$C3,$F6,$91,$89,$D9,$F1
	.dc.b	$63,$35,$CE,$7B,$4E,$8F,$31,$FE,$FD,$AF,$F7,$F3,$F0,$4F,$E9,$8F
	.dc.b	$F4,$DF,$A7,$FE,$01,$F0,$97,$C1,$FF,$3E,$57,$F9,$FA,$72,$D8,$CB
	.dc.b	$F3,$04,$39,$5C,$1D,$5B,$3C,$DA,$F1,$9E,$13,$B3,$E3,$C7,$E0,$E1
	.dc.b	$F3,$7F,$6B,$6F,$B9,$18,$9D,$9E,$2B,$EF,$3A,$6E,$9F,$82,$AE,$15
	.dc.b	$FC,$46,$31,$7D,$5D,$57,$09,$F1,$35,$F7,$CE,$DF,$33,$5F,$16,$1E
	.dc.b	$4B,$F6,$C7,$73,$66,$67,$84,$6B,$0F,$45,$66,$A1,$C4,$EA,$73,$D5
	.dc.b	$F4,$FF,$91,$84,$63,$2F,$F0,$53,$63,$FD,$72,$D7,$ED,$3C,$BE,$53
	.dc.b	$74,$69,$27,$8F,$23,$AB,$AB,$63,$31,$83,$FC,$35,$D0,$FA,$5F,$F4
	.dc.b	$44,$FF,$82,$1A,$FF,$B1,$5A,$FF,$84,$1A,$FF,$A0,$7E,$9B,$F8,$4B
	.dc.b	$FE,$10,$6B,$FE,$CD,$6B,$FE,$18,$6B,$FA,$9F,$C5,$FE,$2B,$F0,$9A
	.dc.b	$FC,$26,$BF,$C2,$33,$46,$23,$F8,$B7,$87,$F9,$4F,$9A,$FF,$1B,$36
	.dc.b	$3F,$DC,$AD,$7F,$4C,$D7,$F5,$1E,$37,$91,$FE,$40,$D7,$FB,$47,$5F
	.dc.b	$F9,$05,$A7,$D8,$F1,$38,$62,$69,$27,$EF,$AF,$0A,$92,$B9,$3F,$53
	.dc.b	$F3,$E7,$C9,$52,$4F,$F8,$F2,$32,$DF,$9F,$E0,$47,$06,$FD,$3E,$84
	.dc.b	$70,$73,$F9,$AF,$01,$1C,$1D,$FD,$3F,$21,$1C,$1E,$3F,$1F,$D0,$47
	.dc.b	$07,$9F,$D3,$F6,$11,$C1,$EB,$F8,$DF,$80,$47,$07,$F9,$6B,$BB,$FD
	.dc.b	$D1,$CF,$FC,$73,$C5,$FC,$CC,$60,$9F,$39,$82,$57,$17,$7E,$AC,$8E
	.dc.b	$8D,$7F,$94,$5F,$F2,$8F,$9B,$EA,$48,$F1,$3E,$0F,$9A,$91,$E5,$6B
	.dc.b	$CB,$07,$C4,$79,$89,$1E,$B7,$D3,$EC,$48,$F7,$35,$E5,$EF,$7B,$88
	.dc.b	$FC,$37,$E0,$FC,$1C,$8F,$8C,$D7,$FA,$17,$C0,$BE,$A5,$EB,$7F,$B8
	.dc.b	$1B,$3F,$D1,$C4,$EE,$74,$7E,$7F,$E9,$9C,$94,$95,$FE,$A3,$8C,$B7
	.dc.b	$97,$81,$1C,$1F,$A7,$FA,$0F,$D1,$91,$C1,$FA,$BF,$37,$F4,$E4,$70
	.dc.b	$7E,$BF,$CE,$FD,$59,$1C,$1F,$B3,$F3,$FF,$5E,$47,$07,$FA,$E7,$F7
	.dc.b	$87,$CF,$7E,$ED,$B1,$98,$C1,$9F,$8D,$91,$C5,$AF,$C9,$9F,$E6,$8F
	.dc.b	$5F,$3B,$A1,$1F,$E0,$DE,$CF,$F0,$22,$38,$3F,$85,$7C,$97,$F0,$62
	.dc.b	$38,$3F,$87,$7C,$A7,$F0,$A2,$38,$3F,$89,$7C,$87,$F0,$E2,$38,$3F
	.dc.b	$DB,$1F,$0B,$FE,$80,$6F,$C0,$FF,$1B,$8C,$47,$E0,$DA,$FF,$82,$D7
	.dc.b	$C2,$EC,$91,$D1,$E5,$F0,$72,$3B,$B5,$FE,$B7,$E6,$3E,$6B,$F9,$19
	.dc.b	$23,$F9,$53,$FC,$9B,$D2,$91,$FC,$B5,$AF,$E7,$3A,$DC,$DF,$9B,$24
	.dc.b	$7E,$6F,$AF,$F4,$11,$F9,$FF,$3E,$4E,$6B,$F9,$F5,$7E,$96,$7E,$17
	.dc.b	$C5,$CA,$F8,$ED,$7E,$27,$C4,$FC,$72,$3F,$31,$F9,$5F,$2D,$23,$E7
	.dc.b	$B5,$FD,$9F,$D7,$7F,$A8,$5F,$33,$5C,$FC,$1F,$EC,$DF,$D9,$C6,$5F
	.dc.b	$F3,$3E,$3F,$DF,$70,$D3,$63,$E0,$7B,$69,$CB,$63,$FA,$B7,$C2,$FF
	.dc.b	$52,$23,$83,$FA,$D7,$87,$FD,$58,$8E,$0F,$27,$E3,$6F,$91,$FE,$C4
	.dc.b	$D8,$CC,$60,$F5,$7C,$23,$F1,$A7,$81,$86,$D7,$F1,$19,$D4,$E4,$FE
	.dc.b	$CF,$94,$9F,$ED,$CD,$7F,$9B,$FE,$80,$F7,$BF,$C3,$6D,$8F,$E3,$9E
	.dc.b	$8A,$31,$1C,$27,$7F,$17,$23,$73,$FE,$D9,$6C,$7F,$78,$F9,$B9,$1F
	.dc.b	$DF,$DA,$F8,$B9,$BC,$24,$79,$BF,$E0,$0F,$8C,$91,$FE,$07,$6B,$FE
	.dc.b	$3E,$FA,$8F,$F0,$67,$C0,$9A,$48,$F6,$F5,$C7,$CE,$D7,$E1,$F9,$9F
	.dc.b	$08,$8F,$B1,$FF,$27,$CF,$F8,$9E,$3F,$CA,$6D,$AF,$01,$A3,$7F,$33
	.dc.b	$27,$FC,$AB,$28,$C4,$7A,$7E,$B3,$A9,$1D,$BE,$0F,$9B,$91,$C9,$AF
	.dc.b	$F5,$6F,$DE,$1E,$EF,$E7,$1B,$1F,$BB,$6B,$B8,$FD,$0F,$C0,$3D,$FA
	.dc.b	$49,$FE,$05,$19,$7D,$3F,$3D,$F0,$88,$E0,$F6,$FC,$6F,$C6,$23,$83
	.dc.b	$9F,$C1,$78,$08,$E0,$EF,$EC,$F9,$08,$E0,$F9,$3E,$9F,$D2,$23,$83
	.dc.b	$FB,$51,$FD,$BC,$86,$0F,$E0,$9F,$4F,$FB,$B7,$FB,$C3,$5F,$F6,$03
	.dc.b	$5F,$F7,$6B,$5F,$F7,$D2,$F8,$0F,$D1,$C6,$9A,$E9,$AA,$38,$75,$78
	.dc.b	$BE,$07,$F1,$50,$69,$39,$3F,$8D,$71,$3E,$5F,$EC,$E1,$FF,$00,$23
	.dc.b	$1C,$3A,$9C,$C8,$4E,$9F,$F0,$B7,$ED,$CF,$63,$70,$8C,$5F,$F2,$96
	.dc.b	$C7,$09,$EB,$87,$EB,$7C,$1B,$D1,$FC,$53,$0F,$CE,$C6,$0F,$1F,$93
	.dc.b	$FE,$BA,$E4,$BF,$20,$69,$2B,$FA,$84,$3A,$76,$48,$CB,$E0,$F6,$E4
	.dc.b	$7E,$B1,$BA,$7F,$03,$F1,$FC,$A4,$9F,$D8,$BE,$2E,$C9,$1E,$06,$CF
	.dc.b	$F9,$03,$0F,$95,$3E,$48,$F4,$91,$FD,$DF,$D3,$FC,$22,$38,$3F,$7F
	.dc.b	$F3,$BF,$18,$8E,$0F,$E0,$7F,$19,$F2,$88,$E0,$FE,$0F,$C9,$F3,$88
	.dc.b	$E0,$FE,$5F,$EF,$FE,$03,$C8,$FE,$6C,$DF,$20,$D3,$1E,$F4,$AF,$E6
	.dc.b	$ED,$CF,$F9,$57,$C5,$FD,$C2,$4F,$F1,$87,$EF,$FC,$5C,$8F,$E3,$6D
	.dc.b	$9F,$E9,$7E,$FF,$84,$93,$FD,$11,$FC,$3F,$89,$23,$F9,$1B,$67,$FA
	.dc.b	$9F,$B5,$E9,$24,$FF,$4C,$7D,$9E,$9C,$8F,$E9,$ED,$9F,$E8,$5F,$07
	.dc.b	$F9,$42,$4F,$F3,$07,$F4,$BE,$8B,$EF,$FF,$CB,$33,$FD,$A5,$B5,$FA
	.dc.b	$93,$F5,$53,$BE,$8E,$33,$A9,$D0,$FC,$7B,$FC,$7B,$7F,$E1,$16,$C7
	.dc.b	$FA,$15,$AF,$06,$BF,$B7,$EF,$FE,$B1,$E4,$7E,$B0,$95,$FD,$21,$FD
	.dc.b	$7F,$93,$23,$FA,$5B,$6B,$FD,$2F,$F2,$9F,$B8,$7F,$C3,$ED,$7F,$82
	.dc.b	$78,$08,$C4,$7F,$A6,$1B,$0F,$B3,$F2,$CF,$FA,$71,$B1,$9F,$82,$FD
	.dc.b	$EC,$6A,$7F,$D4,$3C,$9F,$C3,$28,$D2,$63,$FE,$30,$E5,$D1,$1C,$1F
	.dc.b	$E3,$4E,$AE,$44,$70,$7F,$95,$7D,$97,$D0,$FF,$0E,$6C,$66,$30,$67
	.dc.b	$E7,$64,$71,$6E,$5F,$E0,$9F,$47,$F8,$D9,$27,$FB,$BB,$FC,$83,$D7
	.dc.b	$FE,$DB,$1F,$EB,$FE,$24,$62,$BF,$07,$18,$9E,$07,$C3,$F0,$A3,$FE
	.dc.b	$50,$6C,$FF,$9B,$FC,$5F,$90,$49,$FF,$2A,$BF,$2F,$9E,$47,$F9,$61
	.dc.b	$B3,$E0,$C1,$F2,$3E,$72,$57,$9E,$32,$DF,$91,$F6,$11,$C1,$BF,$3F
	.dc.b	$E0,$11,$C1,$CF,$D2,$F8,$84,$70,$7F,$9C,$7F,$15,$7C,$D7,$F4,$E6
	.dc.b	$C6,$63,$07,$F5,$2F,$0D,$F1,$54,$93,$FE,$7B,$8C,$BD,$7D,$5E,$84
	.dc.b	$70,$7B,$FE,$4B,$EC,$11,$C1,$F0,$FE,$4B,$EE,$11,$C1,$F1,$FE,$4B
	.dc.b	$F0,$04,$70,$7C,$BF,$92,$FC,$21,$1C,$1F,$E9,$1F,$A1,$7D,$4F,$F6
	.dc.b	$E6,$C6,$63,$06,$66,$47,$16,$F1,$FF,$0E,$EF,$76,$33,$FD,$01,$2F
	.dc.b	$F4,$8F,$F4,$3F,$80,$FE,$2D,$7F,$E0,$3F,$43,$FD,$7F,$D1,$92,$5F
	.dc.b	$F9,$FA,$BF,$D2,$8D,$E9,$E4,$47,$5D,$5D,$48,$E0,$E5,$E9,$78,$48
	.dc.b	$E0,$ED,$E9,$79,$48,$E0,$F1,$78,$3E,$92,$38,$3C,$DE,$17,$B4,$8E
	.dc.b	$0F,$57,$87,$F0,$48,$E0,$F7,$79,$FF,$14,$8E,$0F,$85,$E6,$FC,$92
	.dc.b	$38,$3E,$37,$95,$F3,$48,$E0,$F9,$5E,$57,$D1,$23,$83,$E7,$78,$BF
	.dc.b	$54,$8E,$0F,$F5,$7F,$E8,$7F,$C2,$F8,$FF,$A9,$71,$A3,$11,$F6,$DA
	.dc.b	$A3,$53,$FE,$03,$9F,$DF,$CF,$F5,$69,$E5,$E2,$E4,$95,$E0,$F1,$38
	.dc.b	$11,$C1,$E4,$F1,$3A,$11,$C1,$FC,$37,$AF,$F9,$31,$1C,$1F,$C4,$7E
	.dc.b	$93,$F9,$51,$1C,$1F,$EF,$2F,$5F,$FA,$41,$1C,$1F,$DB,$B2,$FF,$6B
	.dc.b	$7E,$98,$E3,$FF,$2F,$F8,$48,$C5,$6E,$31,$3A,$9C,$D7,$F2,$7F,$A9
	.dc.b	$FE,$78,$4B,$FE,$4C,$FC,$9E,$D9,$1F,$51,$B3,$F2,$FE,$09,$F2,$FF
	.dc.b	$E3,$16,$D7,$FB,$A1,$B3,$B9,$FF,$34,$F2,$FF,$93,$23,$FA,$3B,$5F
	.dc.b	$F0,$87,$98,$8C,$47,$F8,$15,$AA,$1F,$47,$FC,$B2,$3F,$A8,$B5,$E6
	.dc.b	$D7,$FC,$1C,$D7,$92,$F8,$9F,$F3,$EB,$71,$FE,$BC,$D9,$EB,$3F,$E8
	.dc.b	$87,$FD,$1D,$F4,$5F,$E8,$FF,$31,$9F,$F6,$2C,$30,$FD,$99,$B5,$EE
	.dc.b	$9A,$FF,$8B,$E2,$BE,$D8,$AA,$33,$8F,$FB,$72,$FB,$CF,$F8,$CA,$73
	.dc.b	$5C,$1B,$1F,$D4,$AF,$C4,$D3,$C4,$53,$C4,$53,$C4,$53,$C4,$66,$FC
	.dc.b	$51,$D2,$FC,$73,$FE,$3B,$BF,$13,$F9,$93,$E3,$94,$E6,$66,$F5,$F3
	.dc.b	$7F,$E7,$8F,$62,$FB,$86,$92,$BF,$49,$3F,$92,$FA,$2E,$3B,$9A,$49
	.dc.b	$DC,$65,$7D,$1F,$A2,$9F,$17,$FE,$61,$8C,$37,$5F,$F3,$24,$F7,$AE
	.dc.b	$31,$8B,$E9,$7C,$27,$FD,$A4,$D7,$67,$FB,$56,$B9,$47,$86,$7C,$9A
	.dc.b	$EC,$97,$D7,$1F,$C3,$DF,$08,$F1,$E3,$E5,$C7,$A3,$54,$ED,$F0,$61
	.dc.b	$E6,$AE,$31,$E6,$6B,$E6,$38,$C6,$DF,$D8,$7C,$0B,$E2,$BF,$D0,$D8
	.dc.b	$B7,$FF,$3D,$4E,$6B,$FD,$E5,$39,$78,$D7,$0A,$DE,$16,$7F,$67,$F0
	.dc.b	$F7,$87,$F7,$AD,$E3,$FF,$7C,$4F,$4B,$EF,$87,$1F,$01,$F4,$6F,$83
	.dc.b	$C3,$6F,$81,$FD,$1D,$C3,$F8,$7B,$FF,$49,$7B,$EF,$8F,$F1,$74,$BE
	.dc.b	$2E,$92,$F9,$E5,$E0,$CB,$D1,$F4,$5F,$C5,$67,$FC,$CC,$D7,$FC,$CA
	.dc.b	$D7,$F8,$FB,$5F,$F5,$13,$FE,$AC,$E1,$F9,$1C,$08,$C6,$9A,$AE,$F3
	.dc.b	$F2,$57,$D1,$3E,$47,$0D,$8F,$82,$6A,$3E,$87,$85,$FA,$0F,$9A,$BF
	.dc.b	$92,$78,$E9,$28,$C3,$70,$FD,$E3,$67,$FC,$DD,$F4,$DF,$67,$E0,$CD
	.dc.b	$3E,$E7,$CE,$F8,$E7,$E8,$1F,$CE,$E5,$F3,$7C,$49,$1F,$63,$05,$3E
	.dc.b	$93,$B7,$CE,$FF,$0D,$9F,$F4,$9B,$70,$34,$C7,$CD,$F9,$49,$87,$DA
	.dc.b	$FA,$4F,$ED,$8F,$7A,$7C,$57,$AC,$91,$BF,$67,$DD,$F1,$7D,$C9,$1F
	.dc.b	$E8,$F7,$C5,$F4,$1F,$0C,$91,$DF,$DA,$F8,$DF,$1D,$23,$C0,$FE,$6F
	.dc.b	$2F,$EC,$51,$E3,$F8,$BF,$61,$1F,$F4,$E3,$F0,$7E,$A3,$EB,$C7,$C7
	.dc.b	$6B,$F3,$B7,$46,$3A,$FF,$69,$DA,$C1,$EE,$7E,$C6,$BF,$D3,$38,$73
	.dc.b	$F3,$30,$7E,$BF,$CB,$FD,$E4,$75,$9F,$ED,$4F,$FB,$15,$AF,$FB,$11
	.dc.b	$AF,$F7,$E6,$BF,$F0,$05,$7D,$07,$84,$91,$FA,$E9,$7B,$7E,$5E,$E9
	.dc.b	$3E,$F7,$C9,$ED,$FA,$7A,$7E,$21,$27,$C8,$FC,$7F,$85,$F7,$12,$3B
	.dc.b	$7A,$DF,$07,$D1,$91,$F2,$DF,$81,$F1,$9F,$17,$D2,$FA,$04,$9F,$80
	.dc.b	$FD,$3F,$57,$E5,$12,$3D,$5F,$15,$FB,$7F,$A5,$FB,$B1,$FA,$20,$E3
	.dc.b	$9B,$E6,$B7,$C9,$4B,$E2,$DF,$45,$FE,$A4,$8F,$DE,$D3,$B7,$FB,$73
	.dc.b	$63,$FB,$93,$5F,$F2,$63,$5E,$A7,$59,$7C,$AF,$F3,$5B,$EE,$D8,$CC
	.dc.b	$F8,$2F,$F7,$BB,$48,$FD,$C7,$8D,$FD,$B1,$BC,$3F,$8D,$F5,$A4,$7E
	.dc.b	$5D,$E0,$FE,$5E,$5B,$FF,$08,$24,$74,$E0,$FC,$B7,$9F,$23,$F4,$0E
	.dc.b	$9F,$99,$F4,$AF,$9B,$FE,$4C,$D8,$FE,$46,$D7,$7F,$1D,$AF,$62,$46
	.dc.b	$9E,$35,$F0,$FF,$04,$73,$D2,$56,$CF,$83,$8F,$F2,$02,$3F,$13,$5F
	.dc.b	$D7,$7D,$34,$8F,$DC,$3A,$7E,$17,$CA,$7F,$28,$23,$F3,$23,$F6,$D1
	.dc.b	$23,$F8,$03,$A7,$F1,$9F,$4C,$F2,$7F,$A3,$36,$3F,$A1,$B5,$DF,$93
	.dc.b	$AF,$2B,$FC,$6D,$26,$92,$7F,$C7,$90,$F2,$FA,$92,$3A,$CF,$F4,$0F
	.dc.b	$56,$47,$E1,$1F,$E8,$7F,$37,$FC,$3C,$91,$FC,$65,$DB,$FD,$1B,$B3
	.dc.b	$F3,$FE,$27,$E4,$49,$1F,$DF,$23,$F4,$7D,$3F,$95,$F0,$E4,$E7,$B7
	.dc.b	$FA,$43,$6F,$E1,$FA,$52,$3E,$1B,$F1,$BD,$2F,$8F,$E9,$FC,$22,$4F
	.dc.b	$C0,$7E,$2F,$A9,$F3,$09,$1F,$32,$3F,$51,$ED,$AA,$92,$7D,$1F,$4D
	.dc.b	$FC,$81,$F5,$4F,$F5,$06,$FA,$BC,$8E,$0F,$E9,$7E,$57,$FB,$59,$1F
	.dc.b	$D7,$F8,$DF,$CE,$78,$A4,$7F,$43,$74,$FF,$70,$F8,$47,$D2,$FE,$86
	.dc.b	$3F,$80,$C6,$3C,$0F,$F3,$C3,$63,$FD,$A4,$D7,$F8,$39,$FE,$06,$F9
	.dc.b	$87,$D0,$7F,$83,$1B,$1F,$E3,$0F,$71,$18,$8F,$16,$1F,$B1,$F3,$BF
	.dc.b	$A9,$FC,$FF,$F5,$5F,$3F,$F8,$FF,$B4,$8C,$5F,$FB,$FD,$BE,$DF,$F8
	.dc.b	$13,$8B,$FA,$C7,$04,$AF,$EC,$4F,$F9,$16,$B7,$E7,$3D,$DF,$F0,$43
	.dc.b	$6B,$FD,$DD,$87,$F8,$3A,$72,$7F,$6D,$E3,$FE,$C8,$49,$FE,$3B,$1F
	.dc.b	$E1,$CF,$A5,$E6,$48,$E7,$68,$FE,$ED,$97,$F6,$D2,$4F,$F2,$68,$F3
	.dc.b	$77,$48,$FF,$26,$E4,$8F,$F3,$27,$D1,$7F,$90,$A7,$FD,$C5,$ED,$FA
	.dc.b	$5F,$EE,$59,$7F,$79,$39,$3F,$E0,$19,$FF,$1E,$68,$69,$1F,$F0,$56
	.dc.b	$3F,$E0,$12,$4F,$F8,$C6,$7F,$CB,$5C,$1F,$E5,$47,$FC,$C9,$19,$F8
	.dc.b	$3F,$29,$E6,$FA,$2F,$3F,$C8,$FF,$8D,$68,$D2,$57,$D0,$FC,$77,$D9
	.dc.b	$91,$A6,$2B,$8C,$FE,$FF,$E6,$3F,$C2,$ED,$F8,$2F,$EA,$D1,$88,$FE
	.dc.b	$11,$F4,$DF,$E1,$6D,$3F,$D2,$5E,$43,$E3,$BF,$DA,$F5,$FD,$82,$31
	.dc.b	$F1,$DF,$ED,$76,$C7,$E7,$8F,$EE,$13,$C1,$FE,$E7,$A7,$9C,$FF,$4C
	.dc.b	$69,$F1,$BA,$B2,$D9,$4F,$D1,$EE,$DF,$04,$9F,$F2,$F5,$23,$EF,$E5
	.dc.b	$EC,$F7,$65,$7C,$17,$DF,$F3,$3F,$DD,$7E,$E1,$ED,$FA,$5B,$5E,$9D
	.dc.b	$91,$88,$FD,$CC,$62,$7F,$CD,$59,$7D,$C9,$FF,$24,$37,$7F,$F4,$9B
	.dc.b	$76,$F8,$3C,$7F,$EB,$FA,$FF,$0D,$7D,$37,$E7,$DF,$F8,$00,$D7,$E5
	.dc.b	$35,$FE,$12,$D7,$FC,$23,$F4,$FF,$E7,$87,$E8,$07,$0F,$D3,$6B,$FC
	.dc.b	$49,$AF,$2F,$A4,$FF,$30,$B7,$C5,$E7,$29,$CC,$FF,$8B,$23,$13,$AF
	.dc.b	$86,$FF,$51,$41,$A4,$E4,$F2,$F5,$3E,$67,$FC,$C7,$87,$F8,$AA,$31
	.dc.b	$A6,$A7,$33,$FE,$3A,$9C,$BF,$13,$DD,$CA,$D3,$FD,$B5,$7F,$2A,$73
	.dc.b	$18,$9F,$E6,$38,$7E,$9F,$B3,$FD,$DD,$C1,$F5,$9B,$87,$D6,$F2,$AD
	.dc.b	$FE,$4B,$87,$F6,$73,$FD,$A7,$EB,$7E,$23,$F8,$76,$CF,$E1,$9A,$FE
	.dc.b	$9D,$AF,$F9,$53,$93,$F5,$9A,$F6,$6D,$7F,$75,$EB,$EF,$5F,$DE,$DF
	.dc.b	$F7,$64,$7E,$83,$AF,$CB,$AF,$F0,$16,$F2,$FE,$72,$74,$7E,$9B,$AB
	.dc.b	$7A,$1D,$FD,$1E,$1F,$17,$E0,$5F,$B0,$AF,$D1,$37,$3F,$D0,$B5,$FE
	.dc.b	$04,$D7,$D9,$9F,$EC,$E7,$D5,$DB,$ED,$8F,$F3,$9C,$60,$FD,$CF,$C1
	.dc.b	$F9,$0F,$4F,$D2,$58,$FE,$54,$FE,$0B,$6F,$E6,$4D,$D4,$D3,$1F,$49
	.dc.b	$F0,$8F,$87,$F4,$96,$FF,$83,$F2,$7C,$23,$E3,$FD,$25,$8F,$E7,$0F
	.dc.b	$52,$3F,$D7,$B0,$47,$8F,$87,$E7,$6D,$FC,$03,$BF,$F0,$FE,$7F,$F7
	.dc.b	$1E,$BF,$9F,$18,$F2,$FF,$25,$D7,$FD,$E2,$7E,$BF,$6E,$5F,$AB,$CF
	.dc.b	$35,$FE,$95,$F2,$97,$3F,$EA,$8F,$C1,$24,$79,$5F,$27,$17,$DE,$3F
	.dc.b	$07,$E3,$D9,$FE,$A8,$FC,$92,$3E,$D7,$D9,$C7,$F8,$63,$D1,$E8,$59
	.dc.b	$FE,$DA,$E1,$FE,$BD,$E3,$DC,$5A,$A7,$FD,$63,$E2,$3F,$19,$FF,$32
	.dc.b	$47,$F8,$75,$B5,$FE,$2D,$FA,$AF,$97,$7C,$61,$7C,$7E,$07,$F6,$7E
	.dc.b	$B2,$FE,$8D,$7F,$8C,$9A,$FF,$B0,$23,$5C,$9F,$C2,$9B,$97,$F6,$9E
	.dc.b	$AF,$D1,$9F,$A8,$F9,$AB,$1F,$DE,$5C,$3F,$DE,$FC,$CB,$87,$F9,$B7
	.dc.b	$E6,$FE,$DC,$FE,$ED,$B1,$FE,$CF,$5F,$15,$FB,$D7,$3E,$31,$CD,$7F
	.dc.b	$9C,$DB,$5F,$E7,$9F,$A9,$FF,$49,$4E,$6F,$F8,$9C,$FF,$B6,$F9,$6E
	.dc.b	$06,$F3,$E1,$B7,$FD,$D6,$3F,$93,$C9,$A6,$3A,$B9,$9D,$3B,$AD,$7F
	.dc.b	$A1,$FD,$8F,$AF,$CB,$FE,$04,$6C,$7F,$1B,$7F,$D3,$8E,$F9,$3F,$8E
	.dc.b	$46,$33,$FF,$4A,$36,$B9,$7B,$9A,$F8,$6F,$E2,$B8,$CC,$7F,$C5,$5E
	.dc.b	$A5,$C7,$FC,$68,$FF,$1E,$3F,$90,$FA,$DF,$8F,$9D,$9F,$F0,$03,$1F
	.dc.b	$C5,$57,$FC,$01,$ED,$F8,$E7,$C8,$EE,$B5,$B3,$FC,$38,$FC,$62,$3F
	.dc.b	$E1,$D7,$E4,$F7,$FD,$03,$E8,$F5,$59,$FF,$09,$F0,$7F,$2E,$3C,$BF
	.dc.b	$4D,$63,$B6,$3F,$E1,$CC,$D3,$A9,$CF,$8F,$FE,$6F,$F9,$6C,$5F,$F8
	.dc.b	$33,$93,$F9,$C9,$F0,$3D,$3B,$7C,$31,$FE,$E0,$D8,$F8,$BE,$AE,$E3
	.dc.b	$FD,$81,$C7,$FA,$C6,$F8,$3F,$8C,$FA,$97,$CD,$7F,$53,$AD,$67,$FB
	.dc.b	$CB,$32,$B7,$1E,$C8,$46,$1E,$07,$0F,$8E,$7C,$6F,$7C,$3F,$37,$18
	.dc.b	$BE,$4D,$AE,$27,$4C,$77,$7F,$D6,$BB,$65,$FC,$1F,$98,$B7,$E0,$E1
	.dc.b	$FF,$4C,$47,$E1,$8F,$C3,$FC,$8D,$AF,$F2,$4F,$93,$FD,$90,$FC,$5F
	.dc.b	$C2,$58,$F9,$8E,$1F,$9D,$F3,$17,$1F,$CD,$DF,$E7,$8F,$DA,$FD,$4B
	.dc.b	$EA,$31,$7F,$16,$30,$FF,$8D,$9B,$5F,$E6,$76,$BF,$EC,$66,$BF,$40
	.dc.b	$2B,$EE,$7E,$BB,$2B,$5F,$E6,$4F,$27,$FB,$F9,$FD,$53,$E1,$3F,$96
	.dc.b	$C7,$F9,$BA,$3F,$C0,$7E,$57,$F5,$B2,$47,$FA,$35,$FE,$BD,$E1,$7F
	.dc.b	$82,$0E,$5E,$05,$9F,$CB,$FE,$1F,$FB,$1D,$3E,$83,$7C,$3F,$E7,$5E
	.dc.b	$9F,$FA,$CF,$C4,$FF,$0B,$CE,$6F,$FD,$06,$FE,$89,$FF,$65,$B6,$FF
	.dc.b	$C6,$ED,$7E,$17,$22,$31,$1C,$BC,$0E,$A7,$BF,$C5,$B7,$E6,$EF,$F0
	.dc.b	$9F,$0F,$C3,$B1,$EA,$F0,$3F,$C1,$D8,$FF,$8D,$B1,$FF,$48,$B7,$2F
	.dc.b	$04,$EB,$8F,$D6,$7C,$DE,$AB,$87,$C4,$E8,$F7,$9F,$47,$AA,$C7,$C8
	.dc.b	$EA,$EE,$47,$E0,$BE,$2E,$8F,$8E,$7D,$9E,$6B,$3F,$43,$AB,$E5,$9F
	.dc.b	$77,$9E,$C7,$D4,$FA,$9F,$59,$1F,$9A,$FB,$BE,$A3,$E9,$9F,$E1,$8F
	.dc.b	$AA,$FA,$1D,$3F,$DC,$78,$7E,$74,$FD,$8F,$90,$98,$7F,$1A,$EA,$60
	.dc.b	$E9,$CF,$37,$E0,$BE,$43,$E5,$92,$FE,$EB,$F3,$BE,$47,$F0,$67,$F3
	.dc.b	$AD,$EC,$FE,$33,$E4,$7F,$0E,$7F,$3E,$DE,$C7,$F2,$AF,$0D,$F4,$4F
	.dc.b	$7B,$FD,$0F,$18,$6C,$66,$7F,$DA,$93,$97,$FD,$B1,$F0,$FF,$DA,$8F
	.dc.b	$F3,$FC,$3E,$AB,$FD,$1F,$C9,$FD,$2B,$D4,$FF,$39,$BF,$17,$BD,$E9
	.dc.b	$7E,$70,$FE,$A7,$CD,$72,$FE,$68,$FF,$96,$7C,$E3,$49,$1F,$EE,$7E
	.dc.b	$87,$C8,$3E,$8B,$FC,$FA,$D9,$FA,$1F,$0E,$F9,$EC,$D7,$F9,$42,$31
	.dc.b	$3A,$9F,$F4,$2F,$97,$FE,$70,$E9,$FE,$67,$D9,$FE,$81,$9F,$F4,$4F
	.dc.b	$B6,$F9,$4A,$7F,$A3,$3E,$9F,$F9,$D7,$17,$E0,$E7,$59,$7F,$56,$9C
	.dc.b	$BF,$EF,$C6,$E3,$FE,$F8,$5F,$55,$CF,$D4,$FF,$0C,$CF,$F4,$AF,$AA
	.dc.b	$FE,$97,$F5,$5F,$E9,$38,$FD,$EE,$18,$DF,$F9,$04,$6C,$4E,$DF,$EA
	.dc.b	$7C,$5F,$E4,$78,$C4,$6D,$B1,$99,$E7,$2B,$EC,$F8,$0E,$B1,$A8,$FE
	.dc.b	$BF,$87,$2C,$2C,$E6,$BF,$B3,$BF,$EC,$C8,$CC,$7A,$2B,$AB,$E6,$8A
	.dc.b	$F9,$67,$83,$E0,$38,$14,$E0,$53,$81,$E7,$C3,$DE,$FA,$CF,$F8,$02
	.dc.b	$53,$DE,$53,$D0,$7C,$CF,$A6,$7D,$23,$E1,$37,$18,$6C,$66,$7E,$4C
	.dc.b	$E5,$BC,$3F,$E8,$06,$F0,$72,$8C,$47,$76,$BF,$D6,$E7,$F3,$53,$F1
	.dc.b	$97,$D2,$FF,$5F,$C7,$EF,$46,$2B,$57,$DE,$72,$DC,$BA,$5F,$16,$D7
	.dc.b	$0C,$23,$F6,$0F,$F0,$DE,$1D,$5C,$07,$1A,$F6,$9F,$BB,$39,$C3,$C0
	.dc.b	$F2,$C3,$C0,$D3,$BE,$3C,$5F,$B9,$87,$78,$F6,$CF,$09,$E9,$7F,$94
	.dc.b	$D9,$F1,$DF,$C7,$67,$FC,$8A,$BE,$03,$FC,$09,$5F,$CD,$E3,$13,$AF
	.dc.b	$0B,$3C,$1F,$A6,$9C,$CF,$F8,$95,$7C,$0E,$A1,$8F,$82,$30,$F8,$B1
	.dc.b	$BC,$AE,$CB,$F3,$13,$0F,$F2,$CC,$E5,$B1,$CB,$E5,$FF,$77,$AF,$F4
	.dc.b	$D6,$E9,$FB,$AB,$57,$8F,$C3,$FC,$1F,$57,$E6,$9B,$8F,$F3,$AE,$0F
	.dc.b	$E3,$0D,$9F,$D0,$FD,$43,$E7,$BD,$92,$8C,$3F,$E5,$76,$BF,$E5,$6B
	.dc.b	$EA,$DF,$8A,$7C,$27,$8D,$B1,$FD,$45,$AF,$F9,$3F,$EA,$5F,$29,$BF
	.dc.b	$1B,$D1,$5F,$D1,$AB,$FA,$8B,$5F,$E3,$B8,$23,$11,$FE,$3F,$F2,$1F
	.dc.b	$93,$FC,$68,$FF,$71,$FD,$D3,$DD,$F8,$CD,$8F,$F4,$6D,$70,$7E,$CF
	.dc.b	$97,$E6,$E6,$FE,$B1,$B6,$0C,$FC,$66,$BE,$9B,$F8,$3F,$5F,$F8,$B5
	.dc.b	$B6,$69,$8F,$88,$FC,$2B,$FE,$F6,$C7,$FD,$32,$D7,$FC,$87,$1A,$F0
	.dc.b	$E4,$69,$E3,$E4,$72,$F3,$3C,$37,$FD,$86,$BF,$B2,$35,$FE,$87,$8F
	.dc.b	$8E,$77,$CF,$E2,$FA,$49,$5E,$28,$F2,$FC,$44,$8F,$2B,$E8,$EB,$F4
	.dc.b	$F8,$FE,$02,$4F,$81,$F1,$F9,$3E,$EF,$8F,$93,$EE,$8F,$83,$F2,$92
	.dc.b	$3E,$0B,$F1,$3E,$9F,$E2,$FC,$6F,$B0,$93,$EC,$7D,$FF,$4D,$F2,$BC
	.dc.b	$29,$3F,$2A,$3E,$6F,$CB,$48,$F9,$AF,$D0,$F9,$3F,$A3,$F4,$7F,$20
	.dc.b	$93,$F2,$1F,$97,$E5,$7D,$6F,$1E,$4F,$D6,$8F,$DE,$FD,$4F,$E3,$79
	.dc.b	$8D,$23,$AF,$1B,$F0,$B1,$F9,$EA,$FF,$0C,$35,$FA,$AD,$7F,$CE,$EB
	.dc.b	$E7,$FB,$FB,$92,$7F,$DB,$F3,$97,$FD,$95,$F3,$8F,$9F,$FD,$CB,$63
	.dc.b	$31,$E8,$3F,$7B,$F5,$1F,$EC,$C8,$FE,$BE,$D7,$FB,$FB,$5F,$F0,$5B
	.dc.b	$5F,$F5,$63,$F9,$D7,$FD,$B4,$D7,$FD,$A2,$DF,$11,$E7,$F0,$F0,$BE
	.dc.b	$4F,$AF,$F2,$BE,$7E,$61,$EC,$8A,$FF,$B4,$F3,$FE,$74,$DF,$5B,$38
	.dc.b	$FF,$9B,$E7,$F9,$AB,$7D,$5F,$AA,$7D,$6F,$56,$DF,$FB,$CA,$7F,$CE
	.dc.b	$DE,$F3,$A7,$F9,$BB,$67,$E6,$DF,$F3,$DF,$31,$F4,$FF,$CA,$27,$FD
	.dc.b	$EB,$18,$BD,$4F,$19,$CB,$BC,$3E,$0C,$2B,$83,$FC,$43,$EA,$3F,$C1
	.dc.b	$6D,$F0,$79,$E1,$FD,$36,$DF,$BC,$FD,$FA,$47,$43,$3C,$FB,$9D,$47
	.dc.b	$83,$16,$3D,$AF,$D2,$FF,$79,$9D,$37,$8D,$3E,$78,$C3,$FD,$7B,$77
	.dc.b	$CC,$FC,$59,$78,$F4,$95,$ED,$F1,$FF,$4F,$F3,$32,$3F,$32,$FE,$AB
	.dc.b	$D5,$FD,$5F,$CC,$C9,$FC,$EB,$B7,$F1,$7E,$C7,$E6,$08,$FD,$BD,$3F
	.dc.b	$DF,$DF,$48,$F5,$3F,$3B,$F1,$5F,$E3,$CE,$E3,$49,$7F,$A0,$F8,$CF
	.dc.b	$D1,$91,$FC,$DC,$7E,$FB,$E6,$7F,$7F,$F3,$B2,$7F,$63,$15,$FC,$B7
	.dc.b	$CC,$FF,$04,$31,$CB,$F4,$7A,$FF,$B3,$B7,$D3,$FC,$39,$F8,$8F,$53
	.dc.b	$F1,$3E,$AC,$C3,$FA,$9D,$FF,$39,$BE,$11,$FD,$5B,$F0,$4F,$EA,$9C
	.dc.b	$3F,$5B,$AD,$18,$9F,$C7,$37,$CB,$FE,$2B,$F3,$3F,$9B,$8F,$EC,$F3
	.dc.b	$C5,$AF,$0E,$5F,$F3,$0D,$F4,$7E,$57,$C8,$7F,$08,$24,$F8,$E3,$F9
	.dc.b	$1F,$C6,$C8,$FE,$2A,$FF,$26,$F9,$7F,$D6,$FA,$92,$7E,$C1,$FE,$28
	.dc.b	$F1,$D1,$C6,$3F,$80,$7C,$BF,$F2,$E2,$3C,$35,$FD,$1E,$DF,$E2,$EF
	.dc.b	$11,$F3,$F4,$95,$FC,$9A,$FF,$C6,$DF,$1B,$FE,$99,$F6,$BF,$CF,$AD
	.dc.b	$9F,$F4,$EB,$5F,$F5,$4B,$5F,$F5,$5F,$D8,$7F,$0E,$C7,$F4,$B0,$8C
	.dc.b	$3F,$A8,$6C,$F3,$E4,$FE,$03,$9F,$EE,$BE,$5F,$D1,$F3,$D3,$0F,$66
	.dc.b	$A9,$C4,$7E,$FB,$2F,$F5,$BC,$BE,$7B,$E0,$1A,$C7,$AC,$7F,$86,$EB
	.dc.b	$FB,$FB,$5F,$3E,$3F,$C2,$BE,$23,$FC,$2F,$39,$9F,$DF,$FC,$A3,$E7
	.dc.b	$9E,$1F,$F9,$76,$30,$D8,$CC,$FA,$2F,$E9,$47,$42,$47,$FB,$D6,$FA
	.dc.b	$3F,$EC,$AB,$ED,$CF,$CC,$91,$F0,$AB,$EE,$7C,$F7,$DD,$F9,$E9,$3F
	.dc.b	$1A,$2B,$F1,$4F,$F3,$8E,$0F,$9B,$FC,$21,$AD,$7E,$2F,$5F,$C0,$25
	.dc.b	$79,$5F,$F5,$FE,$7F,$07,$DA,$F7,$12,$7E,$AC,$FF,$43,$FA,$9F,$C7
	.dc.b	$C7,$F9,$21,$AF,$F4,$46,$DF,$D4,$F9,$AE,$1C,$7F,$AC,$9F,$E8,$75
	.dc.b	$FE,$F6,$EF,$46,$23,$FC,$B4,$D7,$F7,$DE,$33,$AF,$FA,$A3,$63,$FC
	.dc.b	$B2,$D7,$C5,$7A,$EF,$91,$A7,$FA,$E7,$D3,$FF,$81,$63,$FA,$DD,$3C
	.dc.b	$3F,$E0,$2D,$8F,$F8,$03,$92,$31,$1E,$96,$BF,$E7,$15,$E1,$7F,$69
	.dc.b	$F9,$6F,$D2,$D1,$B3,$AA,$FF,$38,$C7,$F7,$ED,$BF,$37,$F4,$1F,$BD
	.dc.b	$39,$7B,$BE,$8F,$82,$97,$FE,$78,$D1,$1F,$C0,$FE,$7F,$DB,$ED,$FF
	.dc.b	$1A,$9F,$ED,$15,$FE,$85,$6B,$FC,$A5,$AF,$F7,$5C,$9F,$03,$FC,$37
	.dc.b	$BE,$4F,$9A,$3F,$D4,$BF,$53,$FC,$A2,$0D,$27,$7F,$F0,$68,$90,$CB
	.dc.b	$11,$FC,$2A,$30,$E0,$E7,$15,$A6,$63,$F8,$17,$CB,$5F,$0F,$FD,$11
	.dc.b	$3C,$DA,$FF,$7A,$9D,$BE,$05,$E2,$FF,$B6,$23,$0D,$7C,$ED,$7C,$CD
	.dc.b	$7F,$C0,$BA,$BD,$2F,$CB,$47,$F9,$5D,$B1,$9E,$8F,$63,$67,$E0,$BF
	.dc.b	$EB,$98,$CF,$B1,$C0,$8F,$E4,$3D,$77,$87,$FE,$1F,$8F,$E4,$7E,$AD
	.dc.b	$C3,$FC,$59,$5A,$9F,$E5,$73,$9E,$97,$E1,$1F,$A9,$9F,$CD,$B6,$FF
	.dc.b	$64,$D7,$FD,$94,$D7,$FB,$72,$ED,$E8,$D9,$FE,$73,$C5,$89,$F0,$63
	.dc.b	$FE,$CF,$9F,$E7,$BF,$2B,$8F,$27,$D1,$E5,$7F,$8F,$23,$CD,$F3,$7F
	.dc.b	$E7,$9A,$FF,$69,$DF,$F3,$F7,$FD,$AB,$19,$CF,$FC,$A3,$39,$7F,$CA
	.dc.b	$71,$FE,$87,$9F,$DB,$B6,$3F,$87,$35,$FF,$73,$35,$FF,$05,$3F,$AA
	.dc.b	$8E,$FD,$7E,$EF,$07,$CA,$F9,$BE,$26,$4F,$92,$E9,$E2,$3F,$8C,$6E
	.dc.b	$C3,$FC,$97,$3E,$49,$CC,$61,$F2,$E7,$F0,$8F,$8B,$E5,$7F,$8E,$EB
	.dc.b	$E0,$C7,$DF,$F2,$5F,$A4,$3E,$5B,$0D,$AF,$ED,$7F,$2B,$89,$E6,$FF
	.dc.b	$2B,$BF,$F3,$5B,$70,$FF,$7B,$57,$F5,$78,$C7,$99,$F5,$BF,$20,$77
	.dc.b	$F9,$23,$0E,$71,$C5,$FF,$85,$6B,$C5,$A7,$F5,$5C,$BF,$CF,$38,$79
	.dc.b	$BD,$B9,$7B,$F9,$7B,$7B,$C1,$1E,$76,$CF,$F0,$8C,$11,$8A,$FF,$7E
	.dc.b	$B5,$FF,$0A,$ED,$F1,$BE,$0F,$E2,$9E,$BF,$76,$E1,$E7,$7B,$11,$ED
	.dc.b	$68,$F1,$7B,$DF,$0F,$3F,$88,$DA,$F8,$1F,$09,$86,$9F,$0E,$3F,$6D
	.dc.b	$93,$EC,$DB,$61,$B5,$FE,$44,$F6,$7F,$D3,$BF,$09,$9A,$FD,$F4,$62
	.dc.b	$7F,$C6,$FE,$82,$BF,$B3,$4E,$70,$FF,$1C,$CE,$6B,$D7,$B7,$DB,$3E
	.dc.b	$E7,$CC,$58,$FB,$0E,$1F,$B5,$F1,$76,$FE,$D7,$C8,$7E,$FD,$BE,$17
	.dc.b	$E1,$7E,$63,$F7,$73,$9C,$BF,$C7,$FC,$69,$D6,$3F,$85,$F9,$8B,$8F
	.dc.b	$F8,$64,$FF,$62,$7A,$FF,$1B,$BF,$FB,$4D,$7F,$99,$97,$D1,$7F,$AB
	.dc.b	$A3,$FC,$D3,$5C,$1D,$46,$63,$FC,$18,$D7,$FC,$AF,$3C,$B9,$7F,$6B
	.dc.b	$39,$9F,$F4,$52,$E6,$FE,$50,$DB,$FC,$E7,$C1,$62,$7F,$96,$7D,$47
	.dc.b	$EA,$2F,$FD,$C6,$DD,$BF,$DA,$0F,$EE,$63,$3D,$FC,$3E,$2E,$63,$ED
	.dc.b	$A7,$91,$8B,$FD,$17,$CD,$7E,$C7,$E4,$7F,$64,$DB,$FD,$77,$CD,$61
	.dc.b	$A7,$EC,$63,$E5,$7C,$6C,$8F,$8F,$C3,$F8,$3C,$3E,$82,$9F,$A1,$D9
	.dc.b	$FD,$6A,$35,$3F,$E3,$58,$C1,$9F,$4B,$EB,$1F,$9C,$F6,$3F,$4D,$87
	.dc.b	$68,$FE,$11,$F1,$52,$3F,$82,$7C,$D5,$C3,$F7,$CF,$F0,$D3,$F8,$77
	.dc.b	$CD,$59,$FC,$04,$3D,$3F,$EB,$4D,$8C,$FA,$B8,$9C,$B2,$3F,$89,$7C
	.dc.b	$CD,$C3,$FC,$CD,$CD,$FC,$23,$C5,$FE,$D0,$DA,$EF,$AE,$3B,$FF,$67
	.dc.b	$7F,$73,$D7,$F0,$CF,$E0,$1B,$5F,$67,$E2,$70,$D3,$EE,$47,$F2,$9F
	.dc.b	$9A,$FE,$BF,$9A,$70,$7F,$AE,$DC,$3F,$EB,$DF,$13,$FA,$26,$7F,$AA
	.dc.b	$AF,$EC,$F1,$89,$64,$F3,$7F,$91,$D6,$9D,$FA,$96,$BA,$3C,$BD,$AC
	.dc.b	$5F,$F3,$5C,$3F,$B3,$63,$FE,$B6,$7A,$5F,$E2,$3E,$5F,$F6,$87,$ED
	.dc.b	$7E,$62,$C7,$E3,$1F,$C4,$11,$FF,$74,$BF,$E2,$4F,$A8,$69,$FC,$89
	.dc.b	$B3,$EF,$F6,$71,$3E,$EF,$9B,$F7,$FC,$DD,$AF,$88,$FC,$2F,$84,$FE
	.dc.b	$B5,$9F,$F9,$5A,$7F,$A2,$37,$5F,$F2,$87,$07,$F8,$5E,$B7,$F1,$1F
	.dc.b	$28,$F9,$7A,$DC,$3E,$86,$DF,$3B,$E6,$BE,$7B,$63,$EA,$7C,$CF,$80
	.dc.b	$8F,$81,$F1,$F7,$FD,$63,$EB,$F8,$16,$7E,$E7,$85,$F6,$BE,$13,$ED
	.dc.b	$B6,$3F,$01,$F1,$9E,$C2,$3E,$C7,$DF,$F0,$5F,$82,$3F,$07,$E8,$D9
	.dc.b	$FC,$47,$17,$E1,$BE,$37,$F0,$ED,$8F,$C3,$F4,$FF,$BC,$BF,$7A,$FC
	.dc.b	$7F,$2F,$C6,$6B,$FE,$E6,$6B,$F8,$EE,$54,$62,$3F,$20,$BE,$5B,$FC
	.dc.b	$76,$DB,$EE,$71,$7F,$DF,$8E,$27,$8C,$F0,$F8,$9E,$3F,$17,$6B,$58
	.dc.b	$78,$23,$AF,$C4,$FF,$86,$71,$FF,$26,$63,$E9,$8C,$19,$F9,$4F,$19
	.dc.b	$E4,$F0,$EE,$3D,$2D,$E2,$7A,$3E,$53,$D2,$D9,$D7,$CA,$77,$23,$FE
	.dc.b	$11,$A4,$78,$FE,$53,$E0,$9F,$0B,$E4,$ED,$7F,$59,$F9,$F7,$67,$F9
	.dc.b	$DF,$E2,$DF,$3F,$CA,$B5,$39,$8C,$70,$7F,$BF,$69,$F5,$3F,$A8,$8F
	.dc.b	$F5,$64,$63,$C7,$FF,$73,$37,$0F,$F5,$6B,$67,$74,$F3,$3C,$4D,$AF
	.dc.b	$89,$E6,$E2,$7F,$B9,$79,$52,$7E,$C6,$0F,$A7,$FE,$91,$1F,$D3,$27
	.dc.b	$D9,$18,$C7,$CA,$D7,$F8,$B9,$FE,$4A,$70,$FF,$1A,$AE,$9E,$0F,$F9
	.dc.b	$57,$ED,$BD,$DD,$BF,$C9,$EB,$F8,$F4,$3D,$5F,$F4,$4B,$AC,$BF,$2C
	.dc.b	$E5,$FC,$4B,$ED,$9F,$F2,$93,$B7,$9E,$7C,$B6,$FF,$49,$FC,$0F,$F9
	.dc.b	$7A,$FF,$23,$EE,$FF,$7B,$7F,$45,$19,$BF,$F2,$E4,$E5,$FE,$64,$FF
	.dc.b	$48,$CB,$D1,$DD,$FE,$18,$D7,$FD,$3F,$39,$C3,$9C,$7F,$91,$2F,$F9
	.dc.b	$DF,$D4,$7F,$3C,$FA,$77,$A1,$FC,$6A,$F0,$FF,$BA,$26,$D7,$CC,$C7
	.dc.b	$E3,$5F,$08,$FF,$78,$4E,$9F,$DD,$5D,$7B,$89,$E0,$E8,$C5,$F7,$7F
	.dc.b	$D1,$CF,$8D,$A7,$62,$9C,$7E,$47,$9B,$DA,$6E,$1E,$6A,$E9,$18,$BF
	.dc.b	$25,$30,$F3,$5E,$E3,$B5,$6D,$F2,$9E,$13,$F1,$A7,$E3,$A6,$B9,$3D
	.dc.b	$46,$BE,$A1,$F1,$4F,$8D,$DC,$A7,$31,$86,$C7,$46,$BC,$FE,$31,$BE
	.dc.b	$DB,$1B,$6B,$B8,$D1,$FD,$EB,$5F,$A3,$4F,$8D,$FD,$3C,$61,$BB,$7E
	.dc.b	$AA,$58,$76,$AE,$2D,$9D,$57,$0C,$3F,$C0,$F7,$D5,$F0,$4F,$3A,$E1
	.dc.b	$1A,$6B,$F0,$EB,$93,$4E,$26,$F0,$AF,$C9,$9E,$37,$DE,$3F,$0A,$DC
	.dc.b	$3C,$F5,$C2,$3D,$6D,$BE,$F1,$BC,$3C,$D3,$E0,$D7,$FD,$D9,$E9,$7D
	.dc.b	$86,$E3,$FE,$C5,$F9,$54,$E5,$B1,$9E,$EC,$11,$C1,$FE,$CD,$F9,$AF
	.dc.b	$C1,$BF,$E2,$D6,$D9,$A6,$BE,$4D,$F6,$AF,$86,$FC,$5C,$61,$B1,$9B
	.dc.b	$FC,$AC,$E5,$FC,$8F,$D4,$3E,$53,$C2,$D8,$CC,$60,$FC,$BF,$43,$EA
	.dc.b	$BF,$C8,$53,$FD,$01,$AF,$F4,$66,$BF,$CC,$A3,$86,$7C,$BD,$AF,$F3
	.dc.b	$74,$FF,$48,$6B,$FD,$39,$AF,$16,$BF,$E5,$2F,$89,$79,$1F,$21,$B1
	.dc.b	$FA,$C6,$BF,$B4,$6B,$FE,$11,$F9,$DF,$B5,$E6,$9A,$48,$FE,$E1,$F0
	.dc.b	$FA,$23,$83,$E5,$FD,$AB,$DC,$FA,$2D,$8C,$C6,$0F,$AB,$F5,$9F,$E0
	.dc.b	$28,$FA,$6D,$7F,$85,$F0,$A3,$11,$FD,$6A,$31,$E9,$7D,$EF,$D2,$3F
	.dc.b	$CB,$B3,$F7,$1A,$F2,$6B,$C5,$AF,$F9,$EA,$5F,$A4,$FD,$68,$C3,$54
	.dc.b	$6B,$83,$F1,$B3,$98,$E3,$F1,$38,$25,$7E,$3A,$51,$C3,$CF,$93,$96
	.dc.b	$BF,$03,$BF,$FC,$78,$FE,$59,$AE,$7D,$CF,$CE,$C6,$23,$E2,$5B,$E1
	.dc.b	$BC,$A4,$8F,$13,$E8,$F8,$59,$1E,$56,$BF,$42,$5F,$0B,$EE,$24,$7A
	.dc.b	$DF,$83,$CF,$23,$DC,$D7,$97,$A5,$F1,$88,$FC,$37,$D3,$ED,$C8,$F7
	.dc.b	$35,$F1,$7C,$1F,$CE,$23,$F2,$DF,$A3,$F0,$32,$3E,$73,$5F,$C4,$FB
	.dc.b	$DF,$E1,$96,$FA,$1F,$C1,$23,$11,$F8,$B6,$BF,$83,$5F,$49,$B2,$47
	.dc.b	$F0,$88,$CB,$FD,$63,$EA,$5F,$1B,$FC,$61,$B1,$98,$C1,$9F,$95,$91
	.dc.b	$C5,$AF,$F6,$89,$7C,$AF,$32,$47,$77,$F9,$87,$D5,$BE,$32,$92,$7F
	.dc.b	$B8,$F5,$27,$51,$FE,$69,$F9,$1F,$DF,$91,$C1,$C3,$EA,$1F,$2D,$FC
	.dc.b	$F5,$B1,$98,$C1,$9F,$96,$91,$C5,$AF,$F8,$13,$C7,$7D,$5F,$F8,$02
	.dc.b	$3F,$89,$35,$D3,$5F,$F2,$BC,$70,$F6,$BF,$C0,$BE,$AB,$E8,$9F,$2B
	.dc.b	$86,$DF,$F8,$42,$75,$39,$3F,$91,$61,$FD,$E2,$7C,$3E,$DE,$08,$F9
	.dc.b	$3E,$73,$64,$70,$7A,$3D,$2E,$64,$70,$7D,$0E,$07,$CA,$7D,$62,$46
	.dc.b	$5F,$E5,$FF,$35,$23,$FD,$38,$D7,$F0,$5E,$7F,$E2,$C8,$FE,$3D,$FC
	.dc.b	$77,$A1,$23,$F2,$4D,$7F,$15,$F1,$3F,$93,$23,$F9,$77,$F1,$3E,$84
	.dc.b	$8F,$43,$5F,$F6,$A7,$4F,$F4,$42,$3F,$EB,$17,$FA,$67,$D5,$BE,$8A
	.dc.b	$92,$7F,$A7,$46,$5F,$EA,$1F,$14,$CB,$FA,$A3,$63,$31,$83,$FA,$C7
	.dc.b	$D5,$FF,$5E,$87,$3A,$30,$DE,$BD,$9F,$B1,$F9,$19,$5F,$B5,$6B,$F9
	.dc.b	$3C,$1F,$21,$E2,$24,$79,$9F,$D0,$7C,$EC,$8F,$D2,$35,$FF,$2F,$F7
	.dc.b	$BB,$DF,$3B,$FA,$72,$4F,$EA,$DF,$F1,$37,$6F,$F3,$78,$FF,$7E,$35
	.dc.b	$57,$E7,$7D,$2D,$9F,$DD,$3C,$49,$3F,$EE,$16,$BF,$E6,$EF,$AB,$7A
	.dc.b	$3F,$E2,$86,$C7,$F9,$B1,$AE,$E3,$FC,$DF,$F1,$2F,$B5,$7C,$EE,$1B
	.dc.b	$5F,$E7,$A9,$D4,$E4,$F5,$FC,$CC,$9F,$E1,$FA,$23,$11,$FC,$CB,$83
	.dc.b	$F9,$81,$1F,$E5,$4F,$F3,$3E,$C9,$1F,$CB,$5A,$FC,$EF,$7F,$F9,$5B
	.dc.b	$7C,$CF,$F0,$B9,$1F,$E6,$9C,$DF,$CF,$48,$E0,$FE,$47,$FB,$DB,$E0
	.dc.b	$BF,$CB,$0D,$8C,$C6,$0F,$AF,$F3,$32,$3D,$AD,$7F,$A6,$F7,$FF,$4B
	.dc.b	$23,$FD,$15,$FE,$9D,$84,$8F,$E9,$0D,$7F,$13,$6F,$8E,$F1,$92,$3C
	.dc.b	$EF,$F8,$E7,$EB,$7F,$AA,$C7,$E8,$9B,$DA,$69,$8F,$09,$F3,$BF,$E4
	.dc.b	$16,$D7,$F9,$05,$AF,$F3,$56,$BF,$CC,$3E,$37,$FD,$92,$FF,$93,$5A
	.dc.b	$FE,$C9,$B6,$6C,$BD,$8F,$01,$2B,$C0,$F8,$FE,$7A,$47,$8D,$AF,$F4
	.dc.b	$7F,$8D,$FE,$AE,$DF,$57,$FC,$B5,$18,$8F,$E8,$AD,$47,$EA,$BE,$9B
	.dc.b	$FC,$44,$FF,$62,$6B,$FB,$86,$BF,$EF,$96,$BF,$E6,$BF,$AD,$FF,$1B
	.dc.b	$BF,$E6,$76,$BF,$D6,$DA,$FF,$56,$6B,$FE,$73,$F1,$7E,$A9,$3D,$BF
	.dc.b	$E3,$E7,$E4,$3E,$A6,$4F,$E5,$1A,$FF,$BA,$3E,$B7,$F9,$23,$7C,$BF
	.dc.b	$F5,$1C,$62,$3F,$DD,$ED,$7E,$08,$7F,$8E,$BF,$EE,$C6,$BD,$7B,$39
	.dc.b	$9D,$3B,$7F,$45,$1D,$1A,$F4,$9F,$DE,$75,$73,$9C,$CF,$D0,$97,$D1
	.dc.b	$7F,$99,$DB,$E3,$F8,$63,$07,$F9,$AF,$D7,$FF,$43,$69,$FF,$00,$1B
	.dc.b	$7E,$A3,$ED,$B8,$70,$D7,$DE,$FF,$6E,$8E,$5E,$97,$E9,$A3,$13,$CC
	.dc.b	$FF,$03,$C7,$F8,$FA,$DD,$7F,$D6,$6F,$F8,$2F,$A1,$98,$C5,$FE,$DA
	.dc.b	$3F,$C8,$BD,$3F,$ED,$1F,$12,$F2,$FF,$28,$C3,$FD,$58,$DB,$FC,$D4
	.dc.b	$F0,$FA,$D7,$45,$25,$FF,$AC,$E3,$2F,$E5,$FA,$3C,$DD,$1F,$E3,$D8
	.dc.b	$C1,$30,$FF,$16,$7D,$6B,$E9,$3C,$AD,$9F,$F5,$F3,$67,$73,$EA,$ED
	.dc.b	$F3,$47,$EF,$F3,$C1,$2F,$FD,$8D,$3F,$E2,$E6,$F1,$71,$23,$FA,$5E
	.dc.b	$1F,$8B,$D7,$FD,$6A,$30,$4C,$3F,$C8,$57,$FE,$A2,$8E,$BF,$17,$E3
	.dc.b	$23,$E0,$9F,$31,$1C,$1F,$CE,$BD,$8F,$AC,$47,$07,$FB,$F3,$EB,$7F
	.dc.b	$96,$37,$C3,$FF,$3E,$FC,$6A,$73,$7F,$EC,$E6,$BF,$EC,$D7,$F0,$2D
	.dc.b	$E6,$E3,$18,$8F,$F6,$43,$5F,$EF,$3F,$51,$FD,$37,$0F,$F3,$8B,$63
	.dc.b	$FC,$0E,$D7,$F6,$FA,$A3,$11,$FC,$6A,$3F,$2B,$F5,$3F,$93,$25,$7F
	.dc.b	$5A,$7C,$64,$7E,$03,$67,$FA,$6F,$BB,$F9,$B2,$4F,$F6,$27,$D7,$F3
	.dc.b	$92,$3E,$43,$67,$E2,$7C,$17,$C0,$24,$FC,$D7,$F5,$58,$B0,$A4,$9F
	.dc.b	$82,$FF,$AB,$3D,$17,$D8,$BE,$6F,$0D,$AF,$F0,$73,$71,$FE,$05,$5C
	.dc.b	$2B,$FB,$0F,$D7,$7F,$AB,$AF,$F7,$94,$8C,$3F,$BF,$6B,$FC,$2D,$AF
	.dc.b	$CD,$F0,$1A,$7E,$FC,$95,$FE,$06,$7D,$A4,$7E,$DB,$67,$FA,$57,$77
	.dc.b	$ED,$89,$3F,$D3,$DF,$DC,$FC,$5C,$8F,$EA,$4D,$9F,$F5,$27,$D8,$BE
	.dc.b	$23,$F8,$6B,$6B,$F8,$C7,$86,$8C,$57,$0A,$FF,$00,$E1,$2F,$E1,$DF
	.dc.b	$F0,$AA,$E3,$3F,$C6,$3E,$8D,$C7,$49,$7D,$A3,$2F,$83,$17,$D7,$F9
	.dc.b	$09,$19,$7F,$91,$FD,$23,$DE,$A4,$9F,$4C,$65,$FF,$18,$FC,$4F,$F2
	.dc.b	$D2,$38,$3F,$4F,$F1,$9F,$8C,$23,$83,$DB,$F5,$9F,$E2,$02,$38,$3E
	.dc.b	$0F,$C9,$7F,$89,$08,$E0,$F8,$BE,$67,$F8,$A0,$8E,$0F,$F2,$2F,$5F
	.dc.b	$F4,$12,$38,$3F,$77,$F0,$DF,$A0,$23,$83,$EB,$FC,$2B,$E2,$FF,$A3
	.dc.b	$B6,$33,$18,$33,$EF,$C8,$E2,$DE,$4F,$F7,$45,$3E,$07,$99,$2B,$BA
	.dc.b	$7F,$DC,$1E,$07,$E7,$A9,$F1,$DF,$C3,$7D,$FD,$47,$F6,$D8,$C5,$70
	.dc.b	$9F,$3F,$C6,$48,$F5,$B6,$BF,$C5,$1E,$FF,$F5,$42,$4F,$F8,$B5,$FE
	.dc.b	$D9,$F4,$FE,$38,$FE,$E1,$E3,$23,$15,$FE,$A8,$6B,$FC,$89,$AF,$ED
	.dc.b	$7A,$3F,$B9,$12,$BF,$CF,$2F,$F7,$9E,$FF,$E9,$53,$FE,$B2,$6B,$FC
	.dc.b	$B1,$AF,$F2,$B6,$BF,$E4,$6E,$E7,$CA,$FF,$9A,$89,$7F,$AA,$7F,$CD
	.dc.b	$BF,$47,$23,$F5,$AD,$AF,$DA,$FC,$4F,$F9,$C8,$93,$FB,$27,$FA,$3E
	.dc.b	$52,$3F,$AA,$36,$7F,$C3,$DE,$63,$F4,$BF,$F2,$CD,$7F,$85,$E3,$13
	.dc.b	$FD,$62,$72,$F0,$6B,$FE,$40,$8E,$5E,$86,$C9,$5C,$EB,$E0,$CF,$7F
	.dc.b	$9B,$E8,$49,$D3,$FD,$9B,$C4,$67,$49,$3E,$58,$CB,$D7,$E6,$BC,$C4
	.dc.b	$70,$78,$74,$F5,$11,$C1,$E5,$D7,$DC,$47,$07,$A7,$E6,$7E,$11,$1C
	.dc.b	$1E,$DE,$FF,$8C,$47,$07,$9F,$C0,$F6,$11,$C1,$F1,$7E,$1B,$E7,$11
	.dc.b	$C1,$F2,$78,$FE,$91,$1C,$1F,$5F,$F4,$CF,$CC,$F1,$38,$11,$86,$CF
	.dc.b	$F8,$4A,$30,$3E,$43,$04,$8E,$8F,$F8,$63,$E7,$5F,$13,$49,$3F,$87
	.dc.b	$8C,$BF,$89,$F9,$FF,$01,$1F,$F1,$0B,$E6,$EA,$FF,$30,$FD,$8F,$F9
	.dc.b	$7C,$93,$FE,$E6,$7F,$CC,$5F,$63,$23,$FD,$D2,$DD,$3F,$CE,$DD,$6F
	.dc.b	$A3,$FE,$F1,$1F,$E9,$28,$C7,$AB,$FE,$68,$9C,$CF,$EC,$5A,$FF,$BC
	.dc.b	$FE,$B9,$F5,$9F,$C5,$A3,$FD,$37,$F1,$89,$CE,$5F,$E7,$B8,$C5,$6E
	.dc.b	$BF,$D8,$31,$89,$D7,$D3,$7F,$9A,$67,$FD,$19,$28,$C4,$76,$6B,$C9
	.dc.b	$AF,$F0,$7E,$EF,$F7,$79,$2B,$FB,$7B,$FE,$4C,$FA,$AF,$F1,$14,$7E
	.dc.b	$E5,$B3,$FB,$E6,$BF,$B7,$6B,$FE,$3F,$FA,$DF,$B2,$FF,$A8,$78,$11
	.dc.b	$88,$FF,$23,$B5,$E1,$C7,$FE,$9B,$F4,$DF,$2B,$49,$8F,$F3,$28,$CB
	.dc.b	$7D,$5E,$04,$70,$7F,$8B,$2F,$FD,$D7,$F5,$9F,$EE,$CF,$83,$69,$E5
	.dc.b	$AC,$3F,$EA,$D9,$B5,$FE,$5C,$9A,$FF,$B5,$22,$BF,$E8,$51,$5D,$1F
	.dc.b	$F6,$04,$6B,$4F,$EA,$F3,$99,$E4,$70,$BF,$13,$4E,$25,$3F,$CD,$57
	.dc.b	$E0,$57,$D6,$30,$D8,$E0,$7C,$02,$DF,$8A,$BD,$D1,$8A,$E2,$FC,$7A
	.dc.b	$F9,$0F,$63,$07,$FB,$47,$E9,$DE,$AB,$E0,$3F,$CB,$31,$86,$C6,$67
	.dc.b	$83,$5F,$F7,$24,$BD,$9F,$EF,$B1,$D5,$AE,$9A,$89,$1C,$7D,$1F,$0C
	.dc.b	$F2,$FA,$DF,$9D,$3F,$E6,$4A,$E1,$18,$F7,$F9,$42,$37,$5C,$74,$CD
	.dc.b	$F5,$AD,$46,$2F,$8D,$F3,$9E,$B3,$C2,$F9,$47,$99,$F6,$ED,$B8,$FB
	.dc.b	$98,$79,$5B,$3A,$AE,$58,$F4,$9F,$7E,$3E,$09,$F1,$46,$45,$F9,$F1
	.dc.b	$F5,$DE,$63,$B4,$78,$4C,$CF,$B0,$6C,$F8,$7F,$F0,$5C,$7C,$FB,$FF
	.dc.b	$17,$CE,$7C,$9F,$93,$18,$AD,$D7,$F8,$DB,$83,$FD,$81,$1F,$EA,$58
	.dc.b	$E5,$1F,$EC,$3F,$79,$E0,$72,$8C,$36,$7C,$73,$A7,$89,$E6,$AE,$F7
	.dc.b	$E4,$F9,$0F,$4C,$EB,$24,$62,$3C,$87,$B9,$C4,$FF,$92,$BC,$67,$D6
	.dc.b	$3E,$B7,$C5,$83,$E0,$9D,$B8,$6E,$5D,$DA,$F6,$9F,$09,$CB,$E8,$1F
	.dc.b	$17,$FE,$BE,$84,$62,$7F,$C9,$AD,$7F,$94,$35,$FF,$26,$7C,$BF,$93
	.dc.b	$D8,$34,$ED,$7F,$27,$C8,$FE,$B5,$5F,$2F,$C4,$7D,$7B,$E4,$70,$D9
	.dc.b	$F9,$B5,$C1,$F9,$5F,$3D,$FE,$77,$7F,$CC,$2D,$7F,$9E,$35,$FF,$80
	.dc.b	$2D,$7F,$E0,$07,$D9,$BE,$4B,$E5,$B6,$3E,$B3,$5F,$90,$FC,$AE,$4F
	.dc.b	$C7,$FC,$67,$D3,$9F,$B8,$BE,$2B,$EA,$B6,$33,$C9,$8F,$8C,$F9,$D3
	.dc.b	$FE,$D3,$9F,$F3,$CB,$5F,$EA,$8D,$7F,$81,$F1,$A1,$F1,$5D,$89,$3E
	.dc.b	$2F,$88,$F0,$F9,$9E,$2E,$59,$3E,$68,$F2,$F9,$FE,$22,$47,$89,$FC
	.dc.b	$17,$EF,$9F,$E9,$9F,$07,$F8,$E6,$68,$C5,$7E,$D1,$B1,$93,$F2,$AB
	.dc.b	$E0,$EF,$14,$AF,$D1,$4E,$5B,$D9,$E0,$23,$FE,$9C,$8F,$CC,$BF,$09
	.dc.b	$7C,$A7,$62,$47,$4F,$1F,$C3,$F0,$3E,$2F,$97,$93,$E0,$8F,$2F,$C1
	.dc.b	$7B,$89,$1D,$DF,$83,$F0,$9E,$A2,$47,$A3,$C9,$F6,$FB,$3E,$EF,$83
	.dc.b	$93,$EC,$8F,$83,$F1,$DF,$28,$91,$EB,$7E,$6F,$C2,$7C,$62,$47,$C4
	.dc.b	$F8,$4F,$93,$F3,$DF,$2B,$E1,$64,$FC,$88,$F9,$BE,$57,$D6,$24,$7C
	.dc.b	$77,$EC,$FC,$2F,$D2,$24,$7D,$0F,$89,$FE,$34,$BE,$A5,$EE,$61,$FF
	.dc.b	$0E,$C7,$F1,$C3,$5E,$F6,$67,$5F,$21,$FE,$19,$AF,$F7,$95,$7F,$58
	.dc.b	$3F,$AC,$B8,$70,$77,$8F,$F1,$07,$E4,$3F,$AB,$EE,$FE,$BE,$D9,$FE
	.dc.b	$D4,$D7,$81,$FC,$A7,$BF,$FC,$B1,$E4,$7F,$40,$6D,$7E,$59,$B3,$FD
	.dc.b	$23,$CA,$7C,$8F,$F2,$89,$FF,$41,$4F,$3F,$89,$C1,$9B,$FE,$5B,$F7
	.dc.b	$5F,$B9,$9F,$E5,$0D,$C7,$FA,$7B,$67,$F4,$BE,$A7,$EE,$67,$C9,$F5
	.dc.b	$BD,$C9,$7F,$D0,$6F,$82,$F4,$1F,$59,$FE,$0F,$8F,$E3,$ED,$7F,$A2
	.dc.b	$C6,$BE,$03,$FA,$96,$0F,$A6,$FE,$2C,$B3,$FF,$78,$7D,$DB,$E8,$7F
	.dc.b	$98,$37,$1F,$E5,$6D,$7A,$E5,$DE,$F9,$47,$96,$75,$1F,$3E,$BE,$51
	.dc.b	$D0,$FE,$79,$5F,$D1,$63,$8E,$5E,$CF,$9B,$C9,$AF,$6A,$E3,$F6,$FE
	.dc.b	$8F,$EA,$18,$7F,$AB,$72,$7E,$27,$0F,$BD,$ED,$7D,$93,$FB,$0C,$67
	.dc.b	$5F,$C0,$47,$DD,$FA,$E9,$7F,$87,$8F,$CE,$7C,$D7,$E7,$7E,$62,$4F
	.dc.b	$E7,$7A,$FF,$A4,$65,$FE,$A3,$7F,$AC,$C7,$E6,$0F,$D3,$7B,$FF,$9F
	.dc.b	$23,$98,$7D,$AF,$EA,$49,$1D,$67,$F2,$1F,$61,$23,$F3,$4F,$AB,$C0
	.dc.b	$FC,$A7,$D9,$49,$FC,$F3,$B7,$F0,$3E,$C3,$EA,$3F,$B5,$3F,$9C,$8F
	.dc.b	$EC,$AD,$77,$1A,$F0,$3F,$B8,$7E,$11,$EE,$E2,$FF,$2B,$73,$A3,$17
	.dc.b	$F9,$3C,$39,$6F,$F0,$FE,$BF,$BE,$D2,$BE,$44,$78,$B7,$F8,$D8,$FF
	.dc.b	$BA,$1B,$87,$C4,$AF,$E0,$33,$96,$E2,$BF,$29,$23,$D7,$7F,$E1,$5F
	.dc.b	$77,$F6,$7F,$01,$27,$FC,$55,$E1,$27,$11,$FE,$1D,$F1,$1F,$5B,$FB
	.dc.b	$F3,$3B,$7F,$91,$BA,$BF,$8F,$D7,$F0,$BC,$FF,$82,$1F,$8C,$F2,$2C
	.dc.b	$7F,$0F,$D3,$F8,$41,$87,$FC,$6F,$F0,$DF,$E3,$33,$F0,$C5,$7F,$DD
	.dc.b	$1F,$06,$8C,$B6,$3F,$DD,$9F,$6F,$FE,$46,$E4,$FF,$75,$B6,$38,$36
	.dc.b	$BA,$47,$B7,$93,$F8,$D1,$1F,$F2,$8F,$C4,$BB,$FF,$33,$1F,$0F,$B7
	.dc.b	$F9,$01,$1F,$E5,$55,$FC,$8F,$BB,$FC,$81,$C5,$FE,$25,$9F,$F3,$7B
	.dc.b	$7E,$08,$D3,$F2,$51,$FC,$CF,$AF,$F9,$61,$CD,$FE,$6D,$D7,$D0,$8F
	.dc.b	$F4,$2A,$E7,$1E,$0F,$37,$F9,$A1,$27,$89,$FC,$FF,$BB,$F9,$C1,$1F
	.dc.b	$7E,$9F,$D5,$9F,$C0,$71,$FF,$3D,$23,$C4,$FE,$91,$F4,$3E,$E2,$3E
	.dc.b	$3C,$D1,$FE,$61,$FC,$9F,$EB,$4F,$F9,$9D,$BD,$1F,$E8,$E6,$F2,$7F
	.dc.b	$4F,$AF,$F2,$E7,$A6,$F4,$7F,$CB,$DF,$1B,$FE,$EB,$C3,$F9,$1B,$6F
	.dc.b	$F6,$35,$BC,$7F,$D0,$9F,$57,$FE,$1C,$7E,$33,$5F,$ED,$9C,$28,$C4
	.dc.b	$7F,$91,$1A,$FF,$8D,$BE,$F3,$FD,$13,$76,$BF,$D0,$F1,$89,$FF,$55
	.dc.b	$D7,$5C,$7F,$9B,$E1,$FE,$8B,$9F,$F3,$7F,$C7,$BE,$03,$FC,$F0,$D8
	.dc.b	$FF,$39,$CE,$DF,$EA,$B0,$F8,$2F,$E0,$6D,$8F,$E0,$2D,$7F,$C0,$DE
	.dc.b	$37,$FA,$8A,$7F,$D2,$18,$BE,$0B,$FD,$3A,$DA,$F8,$B1,$FE,$A0,$F4
	.dc.b	$FF,$3B,$9F,$D3,$F9,$5F,$F5,$1F,$19,$A7,$C0,$F8,$8F,$F4,$87,$D5
	.dc.b	$FF,$66,$E2,$FE,$43,$E6,$7F,$81,$B6,$36,$75,$A7,$F7,$5D,$FC,$4F
	.dc.b	$A1,$AF,$FB,$DA,$BF,$D7,$7A,$FC,$03,$F1,$5F,$5F,$E4,$24,$FF,$B0
	.dc.b	$74,$7D,$67,$F8,$CB,$EF,$5F,$1D,$FD,$3E,$91,$87,$FA,$93,$5F,$F1
	.dc.b	$5B,$5F,$F1,$87,$D8,$7F,$2F,$7F,$BC,$80,$E7,$6D,$F9,$59,$9D,$9A
	.dc.b	$71,$1D,$3D,$6F,$D1,$CE,$E9,$18,$8E,$D1,$8F,$27,$FB,$43,$63,$FC
	.dc.b	$82,$FF,$84,$FA,$F3,$ED,$7F,$B8,$67,$A3,$5E,$6D,$7F,$D6,$DF,$07
	.dc.b	$FE,$14,$C7,$FC,$9D,$CA,$8C,$4F,$AE,$31,$3F,$04,$DC,$EA,$7E,$26
	.dc.b	$8F,$A3,$F8,$31,$86,$BF,$49,$B1,$BA,$FE,$35,$C1,$FE,$E8,$7E,$E7
	.dc.b	$B5,$FE,$28,$E9,$FE,$0D,$ED,$D9,$FF,$54,$D6,$F1,$FF,$1D,$FD,$7B
	.dc.b	$DA,$F0,$F4,$23,$17,$FD,$A1,$B5,$FA,$86,$BF,$91,$FB,$0F,$E2,$13
	.dc.b	$FB,$8F,$8F,$F0,$4F,$F9,$36,$72,$FE,$DA,$33,$C4,$C9,$F5,$38,$6E
	.dc.b	$1F,$89,$D7,$FC,$87,$4F,$15,$F2,$5B,$C3,$9F,$87,$FC,$07,$0F,$E3
	.dc.b	$7E,$27,$6F,$11,$E7,$6A,$3C,$BC,$3E,$FF,$80,$F8,$3C,$5F,$F0,$07
	.dc.b	$0F,$81,$18,$3E,$2F,$93,$FE,$C0,$8F,$21,$FE,$1A,$9F,$C9,$B8,$E2
	.dc.b	$FF,$36,$4E,$DF,$9F,$DD,$F0,$37,$7C,$8D,$3D,$3F,$23,$70,$FE,$66
	.dc.b	$FE,$13,$8B,$F9,$83,$78,$0D,$3C,$BF,$27,$EC,$F2,$15,$EE,$D3,$FD
	.dc.b	$C5,$87,$97,$75,$6A,$3F,$82,$7C,$F7,$D1,$3E,$97,$C3,$D8,$FE,$10
	.dc.b	$F8,$FA,$0D,$24,$7E,$53,$E7,$FE,$D1,$F6,$FE,$AE,$C7,$F1,$17,$0F
	.dc.b	$F8,$23,$EC,$3F,$89,$B9,$3F,$8B,$7C,$FF,$E0,$CF,$C2,$78,$76,$3F
	.dc.b	$8D,$3E,$32,$3F,$A4,$7C,$DF,$01,$F8,$B3,$F1,$9F,$37,$67,$F9,$13
	.dc.b	$EB,$23,$FA,$C7,$DD,$F0,$FF,$93,$3F,$29,$F3,$56,$7F,$95,$38,$7D
	.dc.b	$59,$DE,$9F,$F1,$AF,$EE,$6F,$1B,$F4,$7E,$6A,$31,$7F,$87,$87,$C2
	.dc.b	$7E,$66,$F8,$DF,$08,$C4,$7F,$27,$9D,$F8,$1F,$C2,$1B,$7F,$E3,$C8
	.dc.b	$FA,$DF,$11,$FA,$BF,$5B,$F5,$8D,$9F,$CB,$75,$3E,$71,$D1,$A9,$CC
	.dc.b	$61,$FE,$91,$5F,$99,$F9,$AF,$DB,$7A,$96,$7F,$A6,$3F,$B2,$F9,$6C
	.dc.b	$6A,$F2,$FF,$73,$E1,$27,$31,$FE,$B6,$FC,$17,$F7,$78,$7E,$55,$B7
	.dc.b	$FE,$D0,$FB,$07,$47,$EE,$27,$37,$FE,$BE,$9C,$DF,$F9,$F9,$B7,$FD
	.dc.b	$F9,$AF,$F6,$76,$BF,$DA,$70,$FE,$07,$0E,$8F,$E3,$33,$FE,$E5,$9C
	.dc.b	$C6,$BA,$FF,$84,$FC,$27,$F9,$53,$0F,$D6,$60,$9C,$FC,$17,$4E,$FE
	.dc.b	$15,$A6,$E5,$FE,$3D,$9F,$E4,$07,$F2,$4F,$C7,$BF,$A4,$46,$6B,$EB
	.dc.b	$C6,$34,$FE,$46,$D8,$FE,$DD,$CB,$F6,$FC,$1E,$57,$EA,$F1,$3F,$A9
	.dc.b	$BF,$4E,$BF,$C1,$1E,$B6,$7E,$03,$4D,$8F,$F5,$5C,$7F,$52,$A3,$4C
	.dc.b	$7C,$DF,$5F,$9B,$EC,$DA,$F3,$E5,$EB,$F2,$BD,$8D,$8F,$C6,$F8,$78
	.dc.b	$69,$C0,$FE,$2D,$D6,$E4,$F8,$9E,$D7,$F3,$88,$C4,$FC,$48,$C7,$A5
	.dc.b	$9A,$B3,$C9,$FE,$7F,$1F,$C3,$B7,$F8,$47,$F2,$6E,$1F,$C5,$D7,$FB
	.dc.b	$52,$75,$C5,$F3,$BD,$5E,$3E,$F4,$AF,$2F,$C5,$5C,$79,$F1,$FF,$B0
	.dc.b	$E3,$CC,$7F,$62,$C1,$F6,$3F,$E7,$EB,$E2,$7D,$CF,$67,$F0,$07,$F4
	.dc.b	$5F,$9D,$B1,$FB,$6F,$55,$19,$8F,$F3,$BF,$C3,$DC,$FD,$7C,$8E,$0F
	.dc.b	$E9,$B5,$F8,$8F,$5B,$F1,$9F,$09,$F8,$D6,$C7,$E3,$23,$FB,$9F,$90
	.dc.b	$EF,$CE,$3F,$18,$7D,$17,$E3,$DF,$F1,$E3,$57,$13,$E7,$FF,$53,$39
	.dc.b	$8C,$63,$FE,$F6,$9F,$C6,$11,$FF,$7C,$FB,$3F,$97,$FA,$0F,$CC,$36
	.dc.b	$3F,$43,$E7,$FE,$94,$FD,$37,$C5,$58,$FD,$2C,$FF,$67,$DF,$FB,$2C
	.dc.b	$7F,$AF,$DB,$9F,$FA,$69,$AF,$F9,$47,$83,$E4,$63,$FB,$1C,$BF,$64
	.dc.b	$DB,$FD,$8F,$9D,$FB,$53,$F6,$DC,$76,$3F,$6B,$E0,$F8,$48,$F8,$5F
	.dc.b	$27,$85,$FB,$DF,$0F,$F7,$CD,$9F,$DE,$F8,$9F,$C0,$4F,$E0,$5E,$55
	.dc.b	$8F,$E0,$3F,$41,$ED,$23,$ED,$7E,$07,$4F,$F8,$B3,$F7,$D4,$FF,$0B
	.dc.b	$8C,$39,$FA,$08,$FF,$42,$97,$D0,$63,$8B,$F1,$70,$F8,$9F,$F0,$93
	.dc.b	$6B,$FD,$D4,$D9,$FF,$62,$B5,$F9,$8F,$F8,$91,$E7,$F0,$7E,$5F,$82
	.dc.b	$F3,$3E,$7F,$91,$97,$E3,$F8,$8B,$7E,$2F,$9C,$F0,$92,$7C,$2F,$93
	.dc.b	$E6,$BE,$07,$CC,$FC,$16,$CF,$B7,$E3,$FE,$21,$F1,$7E,$5E,$C7,$C1
	.dc.b	$F9,$8F,$69,$1F,$6B,$F0,$3C,$CF,$99,$F3,$1F,$35,$B3,$F2,$7E,$63
	.dc.b	$E8,$1F,$47,$D1,$B1,$FD,$AB,$E2,$3F,$C7,$5E,$57,$F9,$E7,$7F,$F7
	.dc.b	$5B,$5F,$C9,$35,$FF,$54,$79,$0F,$AE,$FD,$4C,$7F,$18,$F8,$4F,$D9
	.dc.b	$F0,$FF,$96,$A7,$3E,$97,$E0,$29,$5A,$CB,$FB,$FF,$DF,$B7,$FF,$38
	.dc.b	$B6,$3F,$C0,$7F,$06,$CB,$35,$FE,$76,$8C,$4F,$FA,$DF,$0F,$E9,$F1
	.dc.b	$87,$1E,$4F,$FB,$2A,$FF,$CB,$13,$99,$FD,$77,$CE,$BE,$5F,$FD,$02
	.dc.b	$DC,$3F,$C2,$7F,$5A,$F7,$33,$5F,$E8,$68,$C4,$EA,$7F,$D1,$BF,$3B
	.dc.b	$FE,$6A,$F8,$BD,$47,$FA,$33,$E1,$3F,$C0,$3C,$3F,$E8,$86,$DF,$F6
	.dc.b	$6F,$AD,$FF,$41,$BF,$E2,$28,$CE,$9F,$E6,$89,$CB,$FC,$B1,$B5,$FE
	.dc.b	$92,$7F,$A8,$C6,$79,$FF,$C0,$EF,$FB,$7F,$EB,$7F,$DC,$1F,$57,$FE
	.dc.b	$7E,$8F,$E1,$D8,$E3,$DB,$F9,$50,$AC,$3A,$15,$C5,$FF,$73,$46,$9F
	.dc.b	$EF,$F5,$5F,$01,$C2,$BE,$9C,$66,$3B,$0F,$9E,$E8,$FF,$A0,$4C,$4F
	.dc.b	$92,$BF,$DF,$13,$AE,$4F,$C6,$3C,$2B,$9E,$19,$C3,$B5,$F8,$EF,$F2
	.dc.b	$53,$B8,$C4,$EA,$BC,$C7,$BE,$74,$FC,$43,$D6,$59,$C3,$80,$4F,$AC
	.dc.b	$7C,$2B,$EB,$79,$46,$1B,$19,$9F,$A3,$39,$6F,$C4,$7F,$95,$DB,$BF
	.dc.b	$28,$C4,$7E,$86,$7F,$D4,$33,$93,$EA,$3F,$E6,$A8,$BF,$1F,$E6,$7D
	.dc.b	$E2,$38,$B8,$6E,$5F,$7D,$E5,$19,$9E,$53,$9B,$F3,$36,$FE,$BE,$DD
	.dc.b	$F0,$E3,$B7,$3C,$3C,$4E,$F1,$FB,$E7,$C9,$C7,$9D,$F1,$AE,$ED,$9F
	.dc.b	$6C,$E9,$E1,$87,$B9,$87,$AE,$35,$B3,$E6,$1E,$CF,$C3,$BF,$F7,$DC
	.dc.b	$62,$FF,$D2,$6D,$7F,$C3,$6F,$F8,$81,$CC,$F1,$E0,$FF,$2F,$FD,$6B
	.dc.b	$C3,$C5,$7F,$B8,$23,$0F,$03,$FD,$3F,$F4,$2F,$8C,$DF,$53,$EB,$5F
	.dc.b	$2F,$86,$DF,$0A,$C4,$F6,$DB,$FD,$35,$F1,$37,$DA,$3A,$BF,$F5,$6F
	.dc.b	$83,$FE,$C1,$7F,$7F,$48,$C4,$7F,$28,$6B,$FC,$9B,$BF,$F4,$52,$F0
	.dc.b	$9E,$67,$85,$B5,$FC,$E2,$3C,$97,$D3,$93,$FA,$15,$BC,$67,$7F,$A5
	.dc.b	$B8,$7F,$44,$8C,$7C,$6F,$B7,$E9,$7F,$A3,$4F,$A1,$AF,$F3,$E6,$BF
	.dc.b	$E2,$96,$BF,$E9,$2F,$31,$F3,$3F,$1D,$FF,$5C,$C7,$FB,$33,$E5,$71
	.dc.b	$3F,$4B,$F0,$37,$D0,$52,$57,$FB,$5B,$EB,$BE,$6B,$FE,$17,$CB,$E2
	.dc.b	$FB,$9F,$E8,$58,$FF,$09,$C6,$0F,$E1,$5D,$2C,$1E,$07,$D9,$B7,$D3
	.dc.b	$D3,$C5,$19,$FA,$79,$8F,$F8,$95,$FC,$01,$FD,$B1,$FF,$74,$BF,$CA
	.dc.b	$DF,$BA,$D7,$FB,$1B,$5F,$F2,$73,$5F,$E3,$38,$3E,$47,$FD,$A1,$1D
	.dc.b	$3D,$0E,$E7,$E5,$A3,$31,$FE,$7E,$A7,$CA,$79,$09,$1E,$4F,$A5,$F3
	.dc.b	$F5,$FE,$2A,$BF,$1B,$3F,$EC,$66,$AE,$74,$63,$27,$EA,$1F,$DF,$FA
	.dc.b	$7F,$17,$3F,$5A,$FE,$0F,$1F,$EA,$AB,$5E,$2F,$5F,$27,$E3,$92,$BE
	.dc.b	$3F,$07,$CA,$F3,$7C,$84,$8F,$92,$FD,$8B,$FE,$C1,$97,$83,$E5,$3E
	.dc.b	$89,$DF,$CB,$FA,$7F,$35,$30,$FA,$6F,$D6,$F2,$BE,$BF,$5F,$C1,$24
	.dc.b	$FD,$67,$B7,$C8,$7D,$CE,$D9,$3F,$D3,$32,$FF,$15,$71,$7C,$2A,$FC
	.dc.b	$0F,$CD,$CA,$FC,$0B,$F8,$4F,$23,$F0,$BE,$B7,$CF,$24,$FE,$11,$F6
	.dc.b	$7C,$67,$E2,$BE,$7A,$4F,$7E,$5F,$C1,$F1,$2B,$F8,$DF,$BC,$FC,$21
	.dc.b	$F0,$B8,$6E,$1C,$4F,$F5,$8B,$87,$F0,$DE,$67,$F2,$45,$F8,$2F,$FB
	.dc.b	$E9,$FE,$DC,$D7,$FC,$26,$D7,$FC,$7C,$D7,$FC,$A7,$2F,$88,$F1,$12
	.dc.b	$3C,$47,$6F,$91,$FD,$3F,$CF,$48,$FD,$3B,$FC,$A7,$F7,$EF,$C3,$C7
	.dc.b	$E4,$5F,$C9,$3E,$BF,$05,$18,$8F,$F6,$EB,$79,$BF,$27,$5C,$FD,$4E
	.dc.b	$9F,$3F,$2F,$95,$FF,$3F,$6C,$78,$7E,$7E,$46,$0F,$95,$C1,$F3,$FE
	.dc.b	$3B,$E7,$12,$39,$FA,$DF,$47,$E4,$3F,$C2,$D5,$F5,$5B,$C1,$FE,$A7
	.dc.b	$AF,$F1,$EC,$E5,$B0,$F9,$0F,$85,$E1,$49,$F0,$CE,$5F,$B9,$DC,$FD
	.dc.b	$4B,$FD,$6B,$18,$6A,$E1,$FC,$BC,$61,$FE,$83,$BF,$F0,$D9,$F8,$7F
	.dc.b	$3B,$F4,$48,$F3,$F9,$3F,$A9,$F3,$BF,$57,$E7,$64,$FC,$A8,$AF,$D2
	.dc.b	$F9,$CF,$B4,$63,$1F,$F4,$84,$3E,$B9,$F3,$1F,$91,$6C,$7E,$19,$AF
	.dc.b	$DC,$F1,$B5,$E1,$7F,$8F,$A7,$9C,$FF,$A3,$DA,$FF,$8A,$DA,$FF,$90
	.dc.b	$BF,$0B,$7C,$D7,$E7,$DB,$1F,$94,$6B,$F1,$EB,$91,$F9,$3F,$87,$FC
	.dc.b	$A7,$CF,$49,$FC,$4C,$57,$F2,$3F,$3D,$F9,$93,$1A,$7F,$AB,$CF,$E5
	.dc.b	$3C,$FF,$E1,$26,$CF,$ED,$5A,$F8,$47,$43,$E1,$BF,$C3,$D1,$98,$C6
	.dc.b	$1F,$C0,$A7,$FC,$49,$5F,$EB,$D9,$CD,$75,$F5,$BF,$10,$F5,$1D,$AF
	.dc.b	$47,$F8,$13,$63,$F6,$ED,$7F,$5B,$F3,$DA,$F4,$FF,$CD,$F2,$69,$27
	.dc.b	$FC,$ED,$1F,$E7,$07,$9F,$A9,$D8,$E4,$FE,$B0,$F4,$F0,$7F,$B7,$1A
	.dc.b	$7C,$22,$F8,$DF,$C0,$3C,$7F,$E0,$DE,$64,$9F,$DD,$BF,$C9,$1E,$3E
	.dc.b	$07,$E9,$7E,$73,$FC,$77,$E0,$9A,$7F,$0F,$E2,$FE,$21,$F3,$D2,$FF
	.dc.b	$82,$BA,$7F,$83,$F4,$FF,$8C,$3C,$2F,$C3,$C7,$F9,$E7,$4F,$E2,$73
	.dc.b	$93,$FC,$6B,$FB,$FF,$E8,$39,$BF,$C5,$ED,$DF,$FD,$2B,$48,$C5,$FF
	.dc.b	$1F,$BF,$F2,$54,$7E,$8A,$FF,$73,$F3,$9E,$A3,$3D,$F6,$FF,$92,$57
	.dc.b	$EF,$71,$EA,$D8,$6F,$FC,$1E,$FA,$F8,$FE,$62,$4F,$F9,$4A,$BF,$8B
	.dc.b	$3F,$EF,$87,$F1,$38,$FA,$BA,$FC,$A7,$29,$F0,$F2,$FC,$02,$3C,$DF
	.dc.b	$87,$E6,$FB,$09,$1C,$0F,$4F,$CF,$FF,$9C,$3F,$54,$FF,$15,$47,$E8
	.dc.b	$F4,$46,$31,$F9,$75,$BF,$9C,$F8,$FE,$9C,$C3,$F9,$2D,$EF,$7F,$87
	.dc.b	$CD,$F5,$09,$3E,$F7,$EB,$FC,$DF,$D0,$24,$7C,$C8,$FA,$7F,$41,$23
	.dc.b	$E7,$BF,$5B,$E8,$3E,$BF,$C8,$FE,$00,$93,$F2,$DF,$C1,$F2,$7D,$C2
	.dc.b	$47,$D8,$8F,$EF,$7F,$58,$AA,$49,$FE,$A3,$1F,$E7,$C7,$F8,$12,$BF
	.dc.b	$E9,$ED,$8F,$E9,$AD,$7F,$D9,$C7,$F0,$C7,$0F,$E6,$79,$3F,$CF,$3C
	.dc.b	$1F,$96,$FA,$1F,$C8,$9F,$92,$FA,$0B,$5F,$8E,$F9,$2F,$CA,$98,$7C
	.dc.b	$7B,$7A,$88,$FE,$53,$E3,$FD,$BB,$F9,$89,$1E,$67,$F4,$FF,$1B,$FA
	.dc.b	$0F,$22,$4F,$E7,$23,$FD,$59,$F9,$2F,$E1,$38,$FF,$60,$D9,$FD,$5B
	.dc.b	$5F,$F2,$C3,$7B,$CF,$F9,$C6,$78,$FC,$D7,$FB,$C5,$B4,$3F,$AB,$D7
	.dc.b	$09,$D4,$7F,$BD,$FE,$C7,$FB,$B7,$4F,$CC,$8E,$11,$87,$0D,$7E,$8B
	.dc.b	$5E,$5D,$FF,$E1,$96,$DF,$35,$29,$D4,$76,$6C,$FB,$3E,$41,$8F,$A4
	.dc.b	$F8,$AE,$2B,$D7,$92,$31,$3A,$38,$4F,$CA,$F8,$FC,$B6,$32,$FC,$C8
	.dc.b	$CB,$63,$37,$F3,$BE,$3F,$FB,$04,$7D,$4F,$8F,$B5,$F5,$26,$D7,$D4
	.dc.b	$9C,$FA,$0F,$A6,$7B,$9F,$DF,$23,$0D,$8C,$E3,$FE,$2A,$6E,$3F,$9E
	.dc.b	$AE,$5F,$23,$D2,$7F,$D5,$55,$FE,$49,$D7,$FD,$96,$FF,$8F,$23,$37
	.dc.b	$FD,$C7,$2F,$C3,$FC,$87,$9A,$FF,$D8,$B7,$C6,$31,$F2,$19,$F4,$B5
	.dc.b	$16,$74,$75,$F8,$2F,$41,$1F,$43,$EB,$F8,$4F,$31,$E7,$F8,$4B,$3E
	.dc.b	$7F,$5B,$C4,$78,$FD,$8B,$1E,$B7,$FC,$43,$A7,$E7,$9B,$A7,$F7,$D9
	.dc.b	$FC,$B4,$E5,$B0,$F9,$0F,$71,$EF,$F9,$0B,$3F,$1F,$E4,$3E,$11,$F0
	.dc.b	$FE,$42,$C7,$CB,$7B,$91,$EA,$FF,$8D,$3D,$0F,$E7,$3D,$5F,$2F,$E4
	.dc.b	$3F,$07,$18,$AF,$A9,$1F,$3F,$E4,$33,$E8,$7E,$5E,$7E,$C4,$62,$FE
	.dc.b	$A7,$C8,$49,$FB,$9F,$1F,$7A,$3E,$D6,$E8,$F8,$64,$9F,$82,$FC,$6F
	.dc.b	$8E,$FB,$C7,$E5,$3D,$AF,$D1,$F9,$DF,$A9,$E3,$FE,$13,$39,$C7,$E7
	.dc.b	$D7,$E9,$1F,$EF,$6D,$7F,$DD,$0D,$DF,$FA,$CF,$5F,$5B,$F6,$B7,$2F
	.dc.b	$F4,$E5,$3E,$3B,$31,$86,$E1,$F9,$59,$74,$F9,$CF,$47,$4D,$BE,$6F
	.dc.b	$98,$8F,$E0,$A5,$1E,$AE,$AF,$80,$7D,$0E,$7B,$5F,$DA,$9C,$3E,$6F
	.dc.b	$8E,$B8,$F8,$FB,$FE,$41,$FB,$AE,$0F,$B1,$19,$AF,$F5,$4C,$60,$FF
	.dc.b	$2B,$15,$F9,$3E,$0F,$1E,$5F,$E0,$F5,$FE,$18,$E2,$BC,$46,$F1,$E1
	.dc.b	$B8,$7F,$29,$9F,$C0,$36,$7F,$D0,$71,$8A,$FE,$70,$B9,$7A,$9F,$81
	.dc.b	$F8,$BB,$7F,$EC,$0B,$7C,$6F,$F1,$66,$C7,$F9,$E7,$DF,$E4,$EE,$7F
	.dc.b	$8F,$EB,$FE,$D6,$C2,$D6,$B2,$FF,$3D,$61,$FE,$A2,$C3,$95,$7E,$47
	.dc.b	$E3,$7F,$25,$F1,$B6,$7F,$0A,$FE,$57,$E3,$71,$87,$E0,$DB,$97,$FA
	.dc.b	$BB,$CD,$7B,$BF,$29,$B8,$7F,$BC,$97,$BD,$99,$D3,$FE,$A8,$8C,$CF
	.dc.b	$F5,$36,$E1,$FD,$7D,$AF,$F8,$31,$FF,$80,$11,$D7,$DF,$FD,$57,$C4
	.dc.b	$F8,$5F,$D6,$7C,$5F,$EB,$7E,$2E,$D7,$E8,$1F,$D8,$7C,$5E,$31,$FC
	.dc.b	$F3,$74,$FD,$97,$73,$E2,$FF,$6C,$7F,$06,$E8,$B8,$FF,$90,$1C,$3F
	.dc.b	$AC,$F8,$BB,$8F,$FA,$A7,$EA,$5C,$5F,$BF,$F8,$BF,$EA,$31,$8B,$FD
	.dc.b	$9C,$63,$93,$FA,$94,$9A,$49,$FF,$7D,$C3,$97,$99,$FC,$2F,$E2,$EC
	.dc.b	$FE,$FB,$5F,$EC,$3F,$C0,$1F,$DE,$23,$07,$F8,$47,$7F,$E2,$11,$9D
	.dc.b	$3F,$AA,$3F,$87,$7F,$81,$57,$2F,$8C,$FE,$30,$F4,$EC,$98,$73,$3F
	.dc.b	$D1,$F7,$FE,$0B,$71,$6F,$25,$3F,$9C,$8E,$FE,$D7,$98,$F3,$FC,$65
	.dc.b	$BE,$3E,$C3,$7F,$EE,$3A,$FF,$B4,$5B,$3F,$E2,$BC,$1F,$74,$F8,$3F
	.dc.b	$9F,$E3,$7E,$FA,$31,$5F,$E2,$16,$D7,$F8,$B5,$AF,$F9,$69,$AF,$F8
	.dc.b	$DF,$1F,$17,$8F,$F6,$8F,$9B,$F1,$B6,$FD,$FE,$3B,$D9,$79,$5F,$4B
	.dc.b	$C1,$FE,$A7,$18,$AF,$8B,$C7,$F1,$27,$EB,$F8,$9F,$DB,$B9,$3F,$43
	.dc.b	$AF,$CC,$8C,$19,$F1,$7E,$E9,$F7,$BE,$0E,$E3,$E2,$CB,$F1,$D7,$F7
	.dc.b	$38,$BC,$3D,$DF,$84,$7F,$0B,$F1,$5D,$C9,$7F,$DE,$B3,$FF,$40,$37
	.dc.b	$E2,$FF,$16,$7E,$33,$E3,$2C,$77,$F0,$BF,$1F,$F1,$BE,$E6,$C7,$BA
	.dc.b	$DF,$15,$ED,$24,$77,$7E,$07,$C5,$FE,$60,$FC,$CF,$C5,$D9,$F9,$BF
	.dc.b	$85,$7E,$27,$BB,$F3,$3F,$19,$FD,$FA,$31,$3F,$92,$8F,$D0,$FC,$5F
	.dc.b	$CB,$24,$7B,$5F,$9D,$F1,$9F,$A6,$3F,$C8,$FF,$C0,$2F,$4B,$15,$E0
	.dc.b	$E8,$CF,$C6,$F7,$6C,$7D,$AA,$FE,$A7,$CF,$FE,$F7,$6B,$FE,$F2,$6B
	.dc.b	$FA,$36,$F6,$17,$C6,$78,$8F,$1F,$C6,$5B,$F8,$3C,$DE,$6F,$8C,$F8
	.dc.b	$0D,$8F,$17,$CB,$7F,$8D,$B0,$FF,$28,$E5,$FE,$B4,$CB,$B4,$E4,$FF
	.dc.b	$14,$70,$FD,$4B,$F9,$BC,$BF,$0B,$B7,$E6,$3F,$E6,$AF,$97,$75,$58
	.dc.b	$C3,$6B,$F3,$7F,$35,$FE,$43,$CF,$C0,$D7,$F6,$D9,$A3,$11,$FE,$A3
	.dc.b	$8C,$7C,$67,$F9,$3B,$E6,$FF,$23,$3F,$E6,$76,$F4,$FF,$89,$DF,$F6
	.dc.b	$D6,$1F,$AA,$72,$F3,$F9,$CF,$3B,$FA,$E6,$D7,$F9,$C7,$DE,$78,$9F
	.dc.b	$E8,$69,$CE,$1F,$E3,$39,$CD,$FF,$97,$1B,$7F,$2D,$AF,$B1,$AF,$FA
	.dc.b	$F7,$C8,$FE,$03,$CF,$FE,$92,$F9,$D7,$57,$F9,$53,$C9,$FB,$F8,$FF
	.dc.b	$02,$9C,$BF,$C1,$3E,$A1,$F1,$BB,$C3,$FD,$37,$18,$9F,$E1,$F7,$CB
	.dc.b	$2F,$F4,$67,$D3,$FD,$B7,$FD,$1D,$F5,$0F,$87,$FF,$51,$CE,$77,$FF
	.dc.b	$28,$4E,$70,$FE,$27,$39,$7E,$A3,$6F,$F9,$D2,$F1,$B9,$F4,$7F,$8A
	.dc.b	$A7,$FD,$B9,$F5,$1F,$ED,$DF,$A7,$FE,$43,$1F,$80,$C3,$1A,$7F,$B6
	.dc.b	$EA,$CE,$45,$70,$7F,$9E,$46,$9F,$C3,$55,$7B,$9C,$2B,$F3,$D1,$98
	.dc.b	$4F,$FB,$BF,$87,$A4,$7F,$53,$9C,$57,$8E,$3E,$AB,$C9,$A7,$F4,$D3
	.dc.b	$99,$DC,$F4,$5F,$47,$2F,$6E,$8F,$F7,$23,$9C,$3D,$11,$8A,$F0,$4E
	.dc.b	$EF,$D1,$5F,$EF,$5B,$F6,$9F,$EF,$77,$0F,$88,$F2,$7D,$3B,$E5,$DF
	.dc.b	$51,$FE,$FD,$8C,$36,$33,$3E,$B1,$3A,$70,$BE,$45,$C3,$CD,$B3,$A6
	.dc.b	$BE,$26,$BF,$DC,$E7,$94,$FC,$E8,$1E,$0F,$03,$FB,$C6,$DF,$7E,$BF
	.dc.b	$B4,$4E,$CF,$2D,$F8,$27,$4D,$D7,$B5,$F2,$6D,$71,$BF,$0B,$F6,$8E
	.dc.b	$78,$66,$B8,$47,$92,$7C,$2E,$DE,$38,$78,$2B,$84,$C3,$58,$78,$DE
	.dc.b	$4D,$7D,$C6,$E7,$93,$F8,$CD,$9F,$50,$F8,$8F,$A9,$3F,$E2,$78,$C3
	.dc.b	$FE,$EC,$6B,$FC,$ED,$AF,$F2,$CF,$A8,$FD,$F3,$FE,$F2,$6B,$FD,$05
	.dc.b	$AE,$F9,$BB,$DB,$EA,$3F,$D7,$F3,$FE,$FA,$6B,$FD,$25,$AF,$F3,$A6
	.dc.b	$BF,$C7,$FE,$A3,$FD,$3C,$FF,$A8,$1A,$FE,$A7,$B5,$18,$8F,$F1,$C4
	.dc.b	$63,$E3,$BF,$D5,$9F,$85,$3E,$3E,$92,$BF,$D5,$71,$97,$F9,$A7,$8F
	.dc.b	$EA,$F0,$BF,$AA,$C6,$08,$7F,$61,$F0,$DE,$97,$C0,$6C,$7E,$8F,$54
	.dc.b	$62,$38,$4F,$93,$E3,$64,$78,$9A,$FF,$5E,$F9,$AF,$97,$E5,$1A,$48
	.dc.b	$ED,$E5,$BD,$5F,$A2,$D8,$CC,$60,$FF,$67,$7D,$47,$EC,$E3,$FD,$96
	.dc.b	$D7,$E6,$35,$F9,$6D,$7F,$A9,$2F,$9E,$E6,$48,$E6,$FF,$B6,$FE,$0B
	.dc.b	$F0,$B1,$FD,$65,$AB,$8D,$18,$6F,$37,$1F,$9D,$95,$FE,$E0,$A4,$62
	.dc.b	$3F,$05,$1F,$D7,$1D,$79,$9E,$12,$3C,$79,$BC,$84,$70,$75,$E3,$F4
	.dc.b	$11,$C1,$FE,$F0,$FA,$8F,$F4,$83,$79,$3F,$DF,$11,$88,$FC,$A3,$5F
	.dc.b	$C9,$BF,$9C,$7E,$03,$73,$34,$9F,$33,$D0,$93,$E1,$7B,$E5,$23,$C8
	.dc.b	$D7,$C3,$F2,$FE,$42,$3E,$97,$CF,$F2,$32,$3D,$8D,$7D,$3D,$5E,$C2
	.dc.b	$3F,$05,$F7,$F7,$C8,$F8,$8D,$7E,$0F,$CA,$FC,$42,$3F,$25,$F8,$FF
	.dc.b	$29,$23,$E6,$35,$FE,$C9,$F1,$5F,$AC,$F9,$DF,$E3,$2D,$8F,$E9,$CD
	.dc.b	$77,$D8,$9F,$EB,$DF,$4C,$F9,$2A,$4A,$FE,$47,$19,$79,$D7,$E2,$5E
	.dc.b	$7C,$BE,$08,$C4,$7F,$B0,$F0,$FD,$FC,$FF,$1E,$F7,$7F,$49,$3F,$E7
	.dc.b	$1F,$23,$04,$7F,$81,$47,$F6,$B6,$F9,$BC,$48,$EB,$AF,$A1,$1C,$1C
	.dc.b	$BD,$BF,$01,$1C,$1E,$AE,$BF,$41,$1C,$1F,$D0,$B3,$FE,$A6,$FF,$84
	.dc.b	$5A,$FF,$B7,$DA,$FF,$6C,$6E,$7F,$C7,$BB,$3E,$21,$27,$E4,$BF,$0F
	.dc.b	$E8,$24,$7C,$06,$BF,$E8,$7C,$FF,$80,$F5,$FF,$A5,$6B,$E5,$12,$7E
	.dc.b	$3E,$3F,$34,$8E,$0F,$97,$E0,$FD,$12,$38,$3F,$AF,$FD,$33,$E6,$3F
	.dc.b	$C4,$0D,$8C,$C6,$0C,$F4,$C8,$E2,$D7,$FC,$39,$2E,$AE,$64,$8E,$EF
	.dc.b	$E0,$F8,$E4,$7E,$01,$AF,$C1,$E6,$FC,$41,$1F,$C2,$BF,$8F,$F9,$C9
	.dc.b	$1F,$94,$6B,$F9,$EF,$86,$FC,$A1,$1F,$CC,$BF,$E3,$DF,$91,$7C,$2D
	.dc.b	$24,$FF,$80,$A3,$2D,$F8,$6E,$04,$70,$6F,$E0,$3A,$11,$C1,$FE,$3F
	.dc.b	$BF,$F1,$8F,$C6,$7F,$83,$9B,$19,$11,$96,$99,$F8,$EC,$11,$C1,$FA
	.dc.b	$8F,$23,$F5,$C4,$70,$7E,$B3,$C8,$FD,$91,$1C,$1F,$E5,$3F,$21,$F4
	.dc.b	$DE,$08,$FF,$04,$35,$FF,$03,$B5,$FF,$57,$35,$FF,$76,$7C,$DF,$EF
	.dc.b	$C8,$FE,$F1,$FF,$0F,$FE,$7F,$FE,$6D,$EA,$34,$93,$FE,$32,$BF,$1F
	.dc.b	$C1,$5C,$3F,$D4,$13,$E7,$6B,$E6,$8D,$7C,$74,$8D,$35,$FF,$79,$5F
	.dc.b	$F4,$F7,$FD,$C0,$D5,$87,$90,$F3,$9E,$1F,$02,$47,$85,$AF,$4F,$A1
	.dc.b	$F6,$91,$FF,$0E,$52,$3E,$0F,$D0,$C9,$CB,$5F,$F1,$B7,$D1,$FF,$B5
	.dc.b	$9F,$A8,$D7,$3E,$EF,$C5,$8C,$47,$FA,$97,$F2,$FF,$E5,$0F,$FA,$35
	.dc.b	$AE,$7C,$8F,$F4,$E4,$62,$3F,$D0,$63,$EA,$9F,$49,$FE,$A1,$6C,$7F
	.dc.b	$99,$1A,$FF,$A6,$5A,$F2,$D7,$ED,$FA,$7F,$78,$91,$F0,$9F,$C0,$D4
	.dc.b	$8F,$8A,$D7,$EA,$FC,$27,$E1,$88,$FC,$A7,$FA,$EF,$D3,$7F,$94,$63
	.dc.b	$FB,$0B,$5F,$EA,$0D,$7F,$D3,$FE,$5A,$31,$08,$C7,$CC,$F4,$24,$7F
	.dc.b	$68,$94,$7E,$63,$D6,$93,$F7,$1A,$FE,$33,$E7,$7F,$3A,$47,$F0,$2F
	.dc.b	$A3,$C9,$91,$D5,$AF,$AB,$C9,$F6,$91,$F1,$3F,$A8,$F2,$24,$7E,$31
	.dc.b	$AF,$E7,$23,$ED,$FC,$57,$D3,$E7,$46,$27,$F3,$ED,$7F,$3F,$1A,$7F
	.dc.b	$D8,$B3,$CB,$F7,$1F,$F3,$24,$23,$1E,$E7,$15,$3A,$73,$3C,$8F,$C6
	.dc.b	$FD,$43,$E8,$5E,$37,$FA,$BA,$30,$D8,$CC,$FF,$B9,$67,$2D,$FA,$2F
	.dc.b	$DB,$41,$A4,$E6,$FF,$76,$60,$F9,$EF,$F4,$7B,$E3,$BF,$54,$61,$F6
	.dc.b	$1F,$01,$C4,$7C,$A6,$C7,$F0,$6D,$F7,$3F,$35,$B1,$98,$ED,$5F,$C7
	.dc.b	$27,$33,$F3,$B8,$BB,$BF,$55,$AF,$F1,$26,$BB,$AF,$F1,$CF,$D2,$BD
	.dc.b	$DC,$F3,$23,$13,$A9,$CC,$62,$BF,$C8,$AF,$FB,$76,$79,$FB,$DA,$23
	.dc.b	$FC,$5B,$DF,$F0,$E7,$FC,$D1,$BC,$7F,$EC,$39,$D1,$FC,$73,$E9,$B2
	.dc.b	$47,$07,$A2,$7C,$64,$70,$7B,$35,$F3,$91,$C1,$FC,$A3,$0F,$09,$1C
	.dc.b	$1F,$CB,$33,$F2,$91,$C1,$FC,$C3,$E1,$7D,$24,$70,$7F,$BF,$BE,$E7
	.dc.b	$FC,$02,$DF,$4F,$E8,$EC,$9C,$BF,$C9,$81,$98,$DF,$E7,$4F,$36,$92
	.dc.b	$F8,$57,$09,$D4,$7F,$3C,$F4,$BE,$FF,$9C,$F1,$D1,$86,$CE,$CF,$E9
	.dc.b	$7F,$45,$F5,$08,$E0,$FE,$9F,$F0,$9F,$60,$8E,$0F,$E9,$7E,$AF,$52
	.dc.b	$38,$3F,$AB,$F3,$7E,$00,$8E,$0F,$EA,$7F,$3D,$E5,$23,$83,$FA,$BF
	.dc.b	$97,$E9,$23,$83,$FB,$1F,$D3,$FE,$30,$8E,$0F,$8F,$F9,$6F,$F1,$26
	.dc.b	$EF,$FC,$45,$A7,$E9,$A3,$51,$FD,$A3,$E0,$FE,$41,$27,$FB,$4B,$B8
	.dc.b	$7D,$4D,$24,$F6,$75,$F1,$12,$3C,$2D,$9F,$C1,$78,$8F,$94,$79,$FF
	.dc.b	$B8,$8C,$36,$33,$7F,$E4,$16,$D7,$25,$E9,$EC,$93,$B7,$F2,$7F,$A2
	.dc.b	$FF,$74,$8F,$C5,$D2,$30,$FE,$39,$AF,$E5,$DA,$FF,$7E,$7C,$2E,$6F
	.dc.b	$A9,$27,$AC,$F9,$1B,$5E,$22,$3F,$EC,$CF,$B4,$FE,$76,$DC,$7F,$8A
	.dc.b	$C6,$30,$FF,$6A,$B5,$FF,$62,$7D,$DB,$E9,$FF,$8D,$B6,$FF,$DA,$ED
	.dc.b	$9F,$F6,$EB,$5F,$F0,$BE,$FF,$B2,$F2,$7E,$E1,$2B,$F0,$EF,$DF,$F2
	.dc.b	$E4,$7E,$29,$B5,$FE,$73,$C3,$F7,$F9,$75,$8F,$F1,$1B,$5F,$DF,$B5
	.dc.b	$FE,$B3,$EC,$F0,$25,$FF,$9E,$DE,$7E,$74,$8F,$F3,$F3,$67,$F8,$56
	.dc.b	$5F,$C2,$BE,$07,$F2,$84,$AF,$CF,$BF,$97,$FA,$A9,$1F,$A2,$6D,$72
	.dc.b	$1F,$55,$F9,$D2,$4F,$EA,$1F,$D0,$7D,$4C,$8F,$D5,$B6,$7E,$6F,$D3
	.dc.b	$7D,$22,$4F,$D2,$7F,$91,$FE,$F5,$FD,$AE,$3F,$22,$D9,$FE,$3A,$D7
	.dc.b	$F9,$33,$5F,$F1,$8F,$0A,$FE,$F9,$2B,$EF,$BF,$B5,$B9,$1F,$C0,$1B
	.dc.b	$5E,$FF,$85,$FD,$D1,$27,$F8,$2B,$FB,$DF,$4E,$47,$F0,$86,$CF,$C7
	.dc.b	$F8,$6F,$E0,$44,$9F,$E1,$AF,$E4,$B3,$91,$F9,$26,$CF,$FA,$E7,$5F
	.dc.b	$C3,$92,$7F,$D7,$4F,$E6,$88,$FE,$69,$B1,$FD,$27,$C6,$7C,$FB,$DB
	.dc.b	$FE,$82,$DB,$FF,$59,$36,$B8,$35,$FF,$3B,$D7,$2F,$6F,$89,$2B,$8B
	.dc.b	$D3,$D1,$91,$D1,$B5,$D3,$D2,$F0,$92,$7C,$2F,$93,$D3,$91,$E4,$6C
	.dc.b	$F9,$3D,$1F,$49,$27,$D2,$FB,$3C,$A9,$1E,$C6,$CF,$B3,$C9,$F8,$24
	.dc.b	$9F,$82,$FF,$64,$FA,$9F,$E6,$F1,$FE,$D5,$6C,$FC,$B6,$BF,$11,$AF
	.dc.b	$F7,$6D,$11,$FE,$47,$9E,$5D,$FD,$48,$F5,$EF,$F0,$11,$C1,$E1,$E2
	.dc.b	$F2,$11,$C1,$E5,$E6,$F4,$11,$C1,$E9,$E6,$F6,$11,$C1,$ED,$E6,$F8
	.dc.b	$04,$70,$7C,$1F,$9A,$F8,$84,$70,$7C,$5F,$AA,$F9,$04,$70,$7F,$84
	.dc.b	$BD,$87,$C9,$3C,$EF,$F1,$FB,$67,$4D,$E0,$FF,$10,$57,$0A,$FF,$22
	.dc.b	$FC,$CB,$0A,$4C,$3B,$46,$5F,$F1,$0F,$0B,$F0,$5F,$EF,$58,$FE,$8D
	.dc.b	$AF,$E7,$1A,$FD,$96,$BF,$EF,$4F,$AB,$6D,$FE,$56,$6D,$FB,$5B,$3E
	.dc.b	$86,$BF,$ED,$5F,$9D,$F2,$3E,$C6,$B9,$CB,$FC,$C3,$18,$8F,$17,$1F
	.dc.b	$F8,$DB,$1F,$C5,$47,$A6,$30,$7F,$8E,$BE,$15,$F2,$14,$99,$7C,$28
	.dc.b	$CB,$F4,$FE,$A9,$F2,$BF,$E7,$06,$C6,$63,$07,$F9,$D3,$EA,$9F,$33
	.dc.b	$49,$3F,$76,$32,$FD,$FF,$AA,$73,$7F,$9F,$1B,$19,$8C,$1F,$E6,$0D
	.dc.b	$3F,$CB,$BF,$51,$FE,$5E,$F9,$06,$9F,$82,$BC,$3F,$E5,$C9,$B5,$FF
	.dc.b	$00,$66,$BF,$E6,$61,$18,$5C,$FF,$EB,$0B,$EB,$1F,$E6,$D9,$D4,$E6
	.dc.b	$FB,$4A,$B6,$FF,$9C,$A7,$4F,$13,$53,$E7,$69,$99,$E2,$F8,$AF,$C8
	.dc.b	$F9,$8D,$9D,$5C,$DF,$91,$B1,$A3,$91,$4E,$47,$F8,$DA,$3F,$C6,$EF
	.dc.b	$CC,$2B,$FE,$3A,$FA,$77,$B8,$FA,$8C,$36,$BF,$D1,$33,$A9,$C9,$C7
	.dc.b	$E7,$FF,$CA,$53,$A6,$BF,$E9,$79,$46,$23,$EA,$46,$3D,$AE,$47,$FA
	.dc.b	$73,$EA,$7E,$6C,$7F,$56,$9C,$69,$C3,$CB,$FB,$95,$AC,$3C,$2D,$C3
	.dc.b	$77,$FE,$A6,$C3,$92,$9E,$B7,$CB,$0F,$1E,$1C,$9B,$87,$E2,$B0,$E3
	.dc.b	$AF,$0C,$71,$7F,$02,$3C,$11,$D6,$FC,$CF,$88,$57,$8B,$1D,$92,$77
	.dc.b	$3C,$4F,$46,$3E,$57,$D5,$87,$95,$FC,$CF,$D4,$FF,$BC,$DF,$F6,$1E
	.dc.b	$4D,$BF,$29,$39,$9F,$F4,$34,$63,$E2,$7F,$49,$5B,$9E,$7E,$4E,$2B
	.dc.b	$F3,$9F,$52,$FA,$0C,$57,$EA,$23,$0F,$82,$77,$3A,$E3,$65,$E1,$C7
	.dc.b	$B4,$3E,$A7,$FD,$2F,$87,$13,$F6,$6E,$1F,$39,$8F,$80,$E9,$0F,$96
	.dc.b	$F1,$F8,$8F,$A9,$7C,$F6,$1B,$87,$59,$D4,$F2,$C3,$8F,$D4,$BE,$47
	.dc.b	$FD,$37,$48,$C5,$FF,$48,$6B,$FE,$29,$6B,$FC,$37,$3F,$0F,$D5,$7C
	.dc.b	$1C,$4D,$3B,$E1,$F1,$7C,$6F,$F4,$85,$FC,$1B,$7C,$0B,$E0,$BE,$53
	.dc.b	$67,$55,$C1,$F8,$B9,$FF,$5B,$F7,$7D,$3E,$4F,$F9,$76,$7F,$AF,$B5
	.dc.b	$FE,$BD,$1F,$EB,$8C,$5F,$54,$FA,$4F,$B0,$DA,$F2,$35,$FA,$73,$C3
	.dc.b	$E9,$3E,$8F,$D5,$7E,$F2,$7D,$5E,$8F,$A9,$BA,$EB,$BB,$7F,$51,$F3
	.dc.b	$AC,$D2,$57,$F7,$68,$FE,$26,$FD,$96,$BF,$01,$BD,$3F,$8C,$FB,$4F
	.dc.b	$E6,$6F,$E3,$BE,$27,$F9,$B5,$7E,$37,$E2,$0F,$F8,$1C,$C1,$4F,$6C
	.dc.b	$70,$7E,$6F,$C1,$FF,$64,$7F,$C1,$2D,$7F,$CD,$4D,$7F,$CC,$6D,$7F
	.dc.b	$C8,$1F,$AC,$FC,$57,$EB,$B6,$FF,$61,$C8,$8C,$78,$3E,$6C,$7C,$FE
	.dc.b	$6F,$EC,$1F,$4F,$55,$C5,$F4,$D4,$C4,$7E,$93,$0F,$9B,$3F,$E2,$06
	.dc.b	$BF,$E8,$16,$BF,$E7,$A6,$BE,$8F,$91,$E4,$47,$E3,$FD,$37,$87,$E3
	.dc.b	$7C,$5C,$92,$7E,$5C,$78,$7E,$AB,$F5,$78,$1A,$49,$F2,$78,$FE,$72
	.dc.b	$3F,$53,$8B,$E0,$7A,$1F,$A1,$8F,$E0,$7B,$7F,$55,$F4,$32,$6B,$C1
	.dc.b	$F8,$1D,$DE,$02,$57,$C1,$7C,$7D,$9F,$14,$91,$F8,$5C,$3F,$C4,$ED
	.dc.b	$F9,$8F,$0F,$CB,$48,$C1,$FC,$2F,$6F,$E1,$AF,$CB,$F1,$7D,$E4,$7E
	.dc.b	$87,$CD,$7D,$32,$3D,$FE,$6F,$DB,$F0,$3F,$CB,$3E,$79,$F8,$2B,$E8
	.dc.b	$30,$DB,$F3,$57,$F3,$25,$F4,$DF,$CF,$5F,$F0,$F1,$8F,$57,$FB,$FC
	.dc.b	$66,$3F,$C4,$4D,$58,$78,$3E,$43,$B9,$23,$F9,$A1,$CC,$F8,$DE,$BF
	.dc.b	$D4,$F2,$24,$60,$F9,$4F,$D8,$F6,$E4,$60,$F9,$DE,$4F,$DC,$23,$F4
	.dc.b	$7C,$0F,$BF,$EF,$C8,$FA,$8E,$9F,$E8,$79,$7C,$6F,$A5,$F8,$BE,$4C
	.dc.b	$9F,$E9,$53,$96,$F0,$7F,$B7,$5F,$8A,$7C,$07,$0F,$C8,$8F,$A3,$3A
	.dc.b	$8D,$36,$3E,$AF,$C9,$7D,$42,$4F,$69,$FC,$BF,$83,$F9,$8F,$06,$4F
	.dc.b	$E3,$62,$BF,$9A,$EB,$FE,$15,$E5,$7F,$71,$F0,$7F,$C6,$1F,$05,$A8
	.dc.b	$C4,$FF,$31,$6F,$81,$FA,$43,$F4,$BE,$07,$E9,$BE,$AA,$63,$F9,$B7
	.dc.b	$4F,$E7,$7B,$BF,$11,$F2,$8E,$FA,$7E,$BA,$32,$7F,$91,$A3,$F3,$B3
	.dc.b	$F8,$C6,$BF,$E6,$B6,$BF,$CB,$15,$76,$24,$7E,$2A,$3F,$4F,$8F,$B2
	.dc.b	$FF,$7A,$49,$FE,$6D,$1F,$9A,$3E,$21,$FE,$76,$73,$E8,$FF,$23,$8C
	.dc.b	$47,$CB,$9F,$D5,$7A,$5F,$1F,$CE,$FE,$7F,$C9,$FD,$7B,$7C,$9A,$DA
	.dc.b	$E1,$F8,$8B,$FF,$0D,$BF,$8E,$6E,$3F,$E2,$5B,$E5,$3E,$6F,$A9,$FD
	.dc.b	$97,$E3,$DF,$DC,$EB,$FC,$55,$28,$C6,$5F,$CE,$E5,$5C,$3C,$CF,$F3
	.dc.b	$9D,$7F,$67,$BF,$CA,$B5,$FD,$DB,$5F,$F4,$D2,$F3,$3C,$24,$8F,$DB
	.dc.b	$C7,$3B,$F9,$3F,$39,$F1,$FD,$B9,$5F,$3A,$2B,$FC,$B2,$7F,$C5,$8F
	.dc.b	$D2,$F7,$73,$1F,$4B,$E0,$BE,$87,$5C,$AF,$AF,$15,$E5,$D9,$F5,$4C
	.dc.b	$57,$F5,$EF,$2F,$F9,$D4,$FF,$4F,$F3,$BF,$A2,$91,$FD,$AF,$62,$9E
	.dc.b	$07,$DF,$24,$FF,$31,$AF,$C1,$7D,$44,$8F,$C4,$BF,$85,$F2,$3F,$C8
	.dc.b	$BF,$AD,$7F,$81,$39,$7F,$C0,$3E,$BA,$31,$68,$C6,$DC,$E3,$FB,$BF
	.dc.b	$A5,$FD,$68,$8F,$F6,$B7,$F8,$36,$BE,$7F,$3F,$F2,$84,$9F,$E8,$D3
	.dc.b	$FE,$56,$FA,$4F,$83,$1E,$26,$BE,$16,$E0,$73,$DE,$FB,$6F,$F1,$D7
	.dc.b	$0F,$F9,$A9,$BA,$FF,$B4,$5B,$87,$3E,$0F,$D3,$7C,$DC,$8F,$D6,$BF
	.dc.b	$A9,$F9,$BF,$8B,$CB,$27,$FC,$CD,$C2,$8E,$11,$FE,$72,$1F,$59,$FE
	.dc.b	$1B,$7F,$A7,$B5,$FF,$77,$35,$FF,$17,$B5,$FF,$40,$5F,$F5,$9F,$86
	.dc.b	$34,$FE,$B9,$3F,$E2,$29,$71,$7E,$0F,$E1,$3F,$80,$7A,$B2,$7F,$83
	.dc.b	$C7,$F0,$3F,$89,$FA,$7C,$06,$92,$7F,$1B,$0F,$A6,$FE,$14,$48,$FE
	.dc.b	$23,$3F,$DC,$FE,$53,$F7,$75,$FC,$07,$E7,$1F,$4D,$FD,$DD,$B7,$FD
	.dc.b	$42,$B7,$C9,$FC,$42,$47,$F4,$92,$CF,$F5,$2F,$1F,$31,$93,$D3,$F3
	.dc.b	$DF,$C7,$49,$3F,$C9,$23,$F9,$06,$12,$3F,$93,$BF,$E4,$5A,$F0,$C7
	.dc.b	$C8,$5F,$4D,$9F,$A6,$93,$FE,$63,$9C,$BF,$DA,$2F,$E2,$3E,$56,$BF
	.dc.b	$D9,$5A,$FF,$9A,$8F,$F0,$4B,$87,$FC,$B1,$0F,$98,$F2,$12,$3C,$1F
	.dc.b	$3F,$E7,$F4,$79,$69,$CF,$8F,$FC,$27,$18,$AE,$6D,$7D,$8D,$8F,$F0
	.dc.b	$CF,$D2,$5C,$4D,$25,$7F,$A0,$2F,$FC,$4B,$F5,$4C,$7F,$62,$DE,$CF
	.dc.b	$F0,$03,$6B,$FC,$0E,$D7,$8E,$7F,$D4,$1F,$85,$1E,$FE,$EE,$33,$99
	.dc.b	$FF,$47,$CE,$5F,$F1,$AE,$7F,$E3,$0F,$A1,$FF,$78,$65,$F0,$74,$46
	.dc.b	$2B,$E5,$B5,$FB,$8D,$AF,$F1,$CF,$C7,$DA,$34,$97,$F0,$6F,$EA,$FD
	.dc.b	$0F,$FB,$FF,$9B,$FC,$21,$68,$C5,$7D,$AF,$9F,$B7,$F8,$B9,$B4,$F9
	.dc.b	$1F,$EA,$57,$F9,$B6,$D7,$F8,$AD,$AF,$E1,$E7,$87,$A5,$FD,$9A,$3F
	.dc.b	$0F,$F4,$3D,$B4,$FF,$29,$FD,$1D,$B4,$D8,$C7,$D0,$CC,$BF,$35,$F2
	.dc.b	$57,$1F,$F4,$2E,$1F,$EF,$09,$FC,$F7,$C8,$F3,$F9,$2E,$8F,$AB,$5D
	.dc.b	$5F,$9E,$3C,$8F,$F9,$92,$32,$FA,$5A,$FF,$83,$16,$3C,$F1,$E8,$DC
	.dc.b	$3F,$5D,$F2,$5E,$B3,$D9,$B5,$8F,$D9,$3E,$0D,$BF,$CE,$6D,$D3,$E4
	.dc.b	$E5,$FA,$C8,$C7,$B5,$F1,$7E,$4B,$E3,$37,$0F,$F3,$BF,$CD,$3E,$1F
	.dc.b	$EC,$E0,$8C,$4F,$F9,$F3,$DB,$FC,$27,$37,$F8,$62,$73,$5F,$CF,$E7
	.dc.b	$F1,$0F,$FA,$31,$AF,$F0,$16,$BF,$EE,$C6,$FA,$5F,$DB,$7D,$8F,$C4
	.dc.b	$B7,$0F,$F7,$97,$70,$CF,$A1,$FD,$FA,$33,$1F,$83,$34,$F0,$F5,$31
	.dc.b	$1F,$94,$F0,$1F,$6C,$F8,$1C,$36,$FF,$DF,$FB,$7F,$BB,$BD,$67,$D1
	.dc.b	$6A,$FC,$B1,$8A,$FF,$42,$E9,$FD,$C6,$BF,$06,$D8,$CC,$EF,$7F,$DD
	.dc.b	$E7,$71,$FF,$3D,$36,$BF,$E0,$0E,$1E,$B3,$E1,$FB,$9C,$6F,$FD,$E1
	.dc.b	$1D,$A3,$11,$C2,$7C,$7B,$3D,$BF,$8E,$7C,$5F,$95,$B3,$F3,$1C,$3F
	.dc.b	$3B,$E5,$6D,$FF,$BA,$32,$7C,$AF,$CD,$3C,$77,$6F,$FD,$B8,$F8,$C8
	.dc.b	$FC,$4C,$91,$F9,$7A,$7C,$AF,$D7,$F9,$4F,$B0,$DB,$FB,$FF,$29,$86
	.dc.b	$9F,$82,$8F,$BD,$F2,$9F,$7F,$E5,$2C,$FE,$19,$FC,$17,$CA,$63,$5F
	.dc.b	$53,$78,$BF,$CD,$FF,$51,$F1,$6F,$FD,$AE,$DA,$FF,$0C,$2F,$86,$FD
	.dc.b	$1B,$9A,$E3,$1F,$C0,$25,$3B,$AF,$F6,$13,$63,$FD,$92,$D7,$FA,$C6
	.dc.b	$AF,$95,$EB,$D9,$F9,$96,$DF,$FB,$53,$F0,$0F,$F5,$83,$FE,$D6,$FB
	.dc.b	$BF,$D8,$BF,$E9,$68,$CE,$1F,$ED,$E9,$CB,$FE,$F6,$8F,$F5,$6C,$F3
	.dc.b	$F9,$1F,$E6,$71,$89,$FF,$72,$BC,$B4,$FC,$9F,$CB,$7E,$AC,$FD,$67
	.dc.b	$87,$70,$FD,$9B,$87,$F6,$BF,$25,$70,$FF,$20,$77,$38,$7F,$85,$B6
	.dc.b	$BF,$5A,$7F,$80,$D7,$16,$A7,$6F,$CD,$8C,$CF,$FB,$B5,$B5,$FB,$09
	.dc.b	$FC,$67,$7F,$6E,$4F,$E0,$4D,$BF,$E1,$3F,$2B,$FC,$14,$FE,$0D,$F2
	.dc.b	$96,$3F,$86,$B8,$7F,$87,$FC,$AD,$C7,$F4,$5C,$3F,$D3,$F3,$F9,$BE
	.dc.b	$67,$F1,$38,$C4,$E7,$C5,$FE,$25,$17,$0F,$E3,$2F,$D0,$EC,$91,$F7
	.dc.b	$7A,$ED,$FD,$27,$FA,$5F,$98,$E6,$FD,$3D,$7F,$90,$5B,$5F,$E2,$B3
	.dc.b	$EF,$57,$08,$FD,$BC,$F5,$3F,$15,$E0,$7E,$02,$7D,$5F,$29,$27,$E9
	.dc.b	$DF,$F9,$A3,$57,$93,$98,$FF,$56,$49,$D4,$7C,$33,$E2,$7C,$9D,$9E
	.dc.b	$7B,$FF,$A9,$7E,$4F,$E4,$7C,$97,$C9,$6C,$EB,$17,$C9,$E6,$30,$DC
	.dc.b	$BF,$2B,$E3,$FE,$60,$FE,$AB,$F6,$DF,$DF,$AB,$F4,$D1,$8F,$6B,$F6
	.dc.b	$93,$9A,$F8,$C1,$C3,$F2,$5A,$FB,$73,$FF,$5E,$7C,$8F,$E9,$BC,$DF
	.dc.b	$C0,$37,$1F,$F4,$FF,$C6,$7F,$52,$3F,$09,$E3,$58,$FF,$51,$FC,$57
	.dc.b	$F4,$2F,$3C,$D2,$47,$FB,$63,$7F,$E0,$DE,$32,$3C,$51,$FC,$3D,$BE
	.dc.b	$5F,$E4,$7C,$FF,$EC,$3F,$15,$27,$E5,$7C,$A5,$E0,$FA,$E7,$FB,$57
	.dc.b	$E0,$FA,$4F,$F5,$66,$C7,$E2,$71,$FF,$24,$CF,$F5,$89,$CC,$FF,$9B
	.dc.b	$67,$FD,$5F,$3F,$5F,$E4,$FF,$45,$F2,$5F,$A3,$7F,$BB,$DB,$E3,$29
	.dc.b	$FA,$7F,$90,$B7,$F8,$EC,$3F,$CA,$FF,$13,$DC,$91,$E5,$7F,$C3,$90
	.dc.b	$F9,$1F,$D8,$7C,$8F,$EC,$5B,$5C,$F1,$7C,$77,$88,$FD,$B7,$35,$9D
	.dc.b	$7C,$77,$B4,$8F,$C3,$7B,$F2,$FE,$F7,$E3,$FF,$7C,$D9,$ED,$F1,$BE
	.dc.b	$E3,$F8,$17,$25,$8F,$67,$D4,$7F,$9A,$9F,$F0,$1B,$79,$3E,$73,$5F
	.dc.b	$E4,$0D,$7F,$59,$E6,$BE,$A3,$FC,$05,$1F,$ED,$AF,$73,$EB,$F5,$FE
	.dc.b	$16,$15,$BE,$FF,$F3,$15,$6B,$1F,$DB,$72,$7F,$A1,$3E,$3B,$EA,$9A
	.dc.b	$F9,$2B,$3E,$8F,$8B,$FF,$0B,$EA,$69,$23,$D7,$F1,$7F,$E2,$9D,$BF
	.dc.b	$86,$EB,$F8,$EC,$98,$7F,$97,$2F,$F8,$5E,$BE,$BF,$3F,$FC,$61,$F1
	.dc.b	$9F,$B3,$E0,$F5,$4F,$FA,$0F,$C8,$CF,$C7,$E9,$FF,$20,$FE,$13,$FD
	.dc.b	$57,$4F,$E8,$CD,$7D,$AD,$5A,$23,$16,$F9,$4F,$A5,$8F,$81,$FF,$4C
	.dc.b	$69,$E6,$72,$FD,$27,$38,$7E,$AA,$9F,$40,$CF,$FC,$94,$DB,$FC,$E7
	.dc.b	$D0,$7E,$85,$F8,$31,$9B,$F8,$D3,$97,$F6,$0D,$7F,$B1,$35,$FB,$2D
	.dc.b	$7D,$BE,$8B,$E3,$3F,$CC,$AD,$BF,$9F,$F5,$2F,$85,$CD,$7F,$97,$E3
	.dc.b	$13,$A9,$C7,$C2,$E5,$FF,$32,$7D,$4F,$F7,$5C,$7E,$03,$77,$FA,$FF
	.dc.b	$53,$EB,$7E,$CC,$67,$0F,$CC,$CE,$5B,$B7,$F9,$BE,$3F,$09,$7F,$2D
	.dc.b	$D3,$FD,$67,$EA,$7F,$A6,$7D,$43,$6F,$EB,$9C,$3F,$9F,$8C,$56,$C3
	.dc.b	$87,$F5,$38,$F2,$AF,$E7,$F7,$67,$FA,$C1,$FA,$89,$C3,$88,$AE,$4E
	.dc.b	$B3,$C9,$FA,$0F,$92,$7F,$B1,$6F,$E1,$9F,$EC,$95,$8A,$D4,$70,$3D
	.dc.b	$17,$BC,$3D,$93,$D6,$31,$87,$8A,$97,$E4,$BD,$C7,$2B,$EB,$7E,$18
	.dc.b	$D9,$FD,$78,$F1,$94,$C9,$9A,$D7,$D4,$BE,$1D,$F4,$58,$6D,$7F,$68
	.dc.b	$9D,$4E,$4E,$3F,$1E,$E0,$A4,$AE,$11,$96,$C7,$D0,$A7,$C2,$7D,$4C
	.dc.b	$BE,$D4,$62,$B5,$7C,$67,$2D,$CF,$ED,$D6,$EB,$EB,$9F,$E1,$57,$15
	.dc.b	$C3,$0E,$67,$D5,$9D,$61,$F8,$4D,$BC,$2D,$C3,$AE,$3E,$99,$E5,$5D
	.dc.b	$86,$3D,$AB,$C7,$7E,$B7,$8E,$1E,$0A,$E1,$5B,$7C,$26,$71,$FC,$EE
	.dc.b	$AF,$A9,$79,$7F,$98,$C3,$F9,$3C,$62,$FF,$96,$B5,$FE,$2C,$FF,$83
	.dc.b	$9C,$CF,$1E,$0F,$E9,$BF,$2A,$F8,$EC,$57,$F4,$A8,$C3,$C0,$FE,$73
	.dc.b	$F4,$1F,$AD,$C7,$FA,$87,$73,$EA,$5F,$15,$CA,$30,$F4,$D5,$7C,$73
	.dc.b	$FE,$A7,$7C,$31,$FE,$8B,$E7,$65,$FE,$B6,$DC,$79,$C7,$F8,$1F,$EA
	.dc.b	$7F,$C0,$4F,$FB,$B9,$AF,$F8,$65,$AF,$FA,$2D,$AF,$F6,$CF,$8C,$7C
	.dc.b	$F7,$95,$B1,$FD,$59,$AF,$F5,$B9,$EB,$AF,$F5,$6F,$9A,$FE,$61,$3F
	.dc.b	$D7,$1A,$FF,$45,$6C,$6F,$D0,$F8,$9E,$37,$BE,$BF,$C5,$CD,$7F,$AD
	.dc.b	$F0,$7F,$A6,$E3,$11,$FE,$02,$FA,$57,$C8,$7C,$B5,$9D,$31,$1F,$3F
	.dc.b	$CA,$7C,$E5,$25,$FF,$72,$5F,$31,$F5,$FE,$2F,$FA,$34,$FF,$99,$38
	.dc.b	$7F,$CE,$3E,$3F,$FA,$B6,$73,$3F,$E4,$56,$BF,$CA,$B9,$FB,$F8,$1F
	.dc.b	$8A,$F8,$9F,$F4,$27,$A5,$68,$FE,$0F,$3F,$8B,$D3,$FE,$14,$FA,$9F
	.dc.b	$F6,$25,$7F,$90,$DA,$FF,$82,$DB,$CB,$FE,$06,$FA,$9F,$F4,$2C,$FF
	.dc.b	$A5,$A1,$F6,$CE,$8F,$E4,$11,$86,$C6,$67,$FD,$03,$39,$6C,$3E,$83
	.dc.b	$D5,$ED,$7F,$96,$67,$FD,$32,$D7,$FC,$D7,$5E,$09,$FD,$07,$C0,$3D
	.dc.b	$5A,$4A,$E3,$F0,$7F,$A5,$FB,$67,$C8,$FF,$B6,$5B,$3F,$45,$AF,$F9
	.dc.b	$CD,$AF,$FB,$96,$FE,$5B,$E4,$FA,$6F,$39,$E2,$E2,$F1,$FD,$24,$9F
	.dc.b	$F3,$23,$EC,$F8,$5F,$3F,$D2,$7F,$BA,$70,$34,$95,$FE,$85,$D7,$E1
	.dc.b	$FC,$34,$8F,$F3,$74,$7C,$0E,$69,$1F,$E7,$17,$E4,$DF,$BF,$E2,$3D
	.dc.b	$C4,$9F,$73,$F5,$39,$FF,$88,$FC,$D3,$F1,$BF,$6D,$7F,$A6,$23,$1B
	.dc.b	$7F,$16,$E4,$FB,$F7,$FC,$3F,$F1,$7F,$C5,$DF,$F9,$01,$BE,$01,$CF
	.dc.b	$63,$A1,$F1,$3C,$7E,$8F,$FB,$35,$ED,$B6,$69,$8F,$98,$FC,$2F,$C5
	.dc.b	$4C,$BF,$D3,$54,$9F,$E5,$70,$FB,$CF,$9A,$FF,$BB,$0C,$73,$6A,$33
	.dc.b	$1F,$EB,$6D,$3C,$DF,$51,$CB,$F1,$87,$CA,$FF,$79,$6D,$7F,$7D,$6B
	.dc.b	$FE,$DC,$6B,$FE,$1E,$3C,$53,$DC,$8F,$8E,$38,$F0,$FF,$A2,$EF,$E3
	.dc.b	$78,$1F,$A2,$F8,$39,$7F,$A2,$8A,$F9,$F8,$BE,$BD,$F1,$6F,$1F,$FB
	.dc.b	$A5,$FF,$33,$4F,$EA,$BE,$37,$F5,$A4,$7F,$5B,$DD,$FB,$0F,$52,$47
	.dc.b	$EC,$1D,$3F,$83,$FA,$CF,$D7,$23,$19,$7E,$2A,$33,$1E,$5E,$1C,$72
	.dc.b	$7F,$88,$FA,$54,$8C,$36,$BE,$40,$9D,$CE,$B8,$7F,$82,$F7,$4A,$F1
	.dc.b	$6F,$FE,$B3,$BE,$B5,$FC,$2B,$C6,$FE,$17,$DF,$2F,$F8,$5C,$57,$F0
	.dc.b	$1F,$98,$FF,$34,$79,$F1,$7B,$79,$13,$99,$FA,$17,$FC,$BE,$3F,$3A
	.dc.b	$4F,$2B,$1E,$AE,$7B,$25,$F4,$18,$74,$8E,$55,$F9,$8F,$11,$F4,$74
	.dc.b	$97,$C6,$FC,$7D,$BB,$FA,$4F,$F2,$65,$9A,$4A,$FF,$29,$47,$F8,$31
	.dc.b	$FF,$23,$6B,$FE,$72,$8C,$70,$FF,$B6,$E7,$FD,$F3,$EE,$F5,$7C,$1B
	.dc.b	$3E,$3F,$E0,$12,$33,$F4,$7F,$0F,$E1,$64,$70,$74,$FF,$0B,$FB,$EF
	.dc.b	$F1,$36,$3F,$5D,$79,$3F,$C4,$9B,$3F,$1F,$9D,$1C,$35,$FC,$84,$62
	.dc.b	$B8,$7C,$BC,$BE,$17,$C9,$F0,$7C,$FF,$52,$3E,$CF,$89,$FE,$42,$FF
	.dc.b	$9E,$4F,$C8,$17,$0F,$CF,$7C,$4F,$E6,$63,$17,$FD,$89,$F6,$7B,$1E
	.dc.b	$93,$9C,$FE,$17,$E2,$FF,$0D,$F4,$D2,$BE,$5C,$57,$F2,$3F,$4F,$F8
	.dc.b	$B3,$1B,$FF,$03,$E3,$7D,$4F,$E3,$FD,$D9,$7F,$52,$2B,$F9,$AF,$92
	.dc.b	$FC,$A1,$8B,$FF,$23,$FA,$1F,$81,$9F,$F5,$8F,$A3,$FC,$D7,$A7,$F9
	.dc.b	$79,$23,$F6,$3D,$BF,$A0,$8F,$CF,$12,$3F,$97,$BF,$A1,$EC,$FE,$85
	.dc.b	$DD,$27,$F0,$F1,$5F,$D6,$C7,$E9,$CC,$63,$E4,$F8,$5F,$E7,$E4,$9F
	.dc.b	$F1,$16,$5F,$E3,$27,$F4,$FB,$7F,$8C,$9F,$C9,$FC,$17,$EC,$4F,$EA
	.dc.b	$55,$FE,$61,$FC,$71,$F0,$3F,$D3,$1B,$5F,$E2,$D6,$BF,$CD,$76,$46
	.dc.b	$23,$FD,$AB,$B7,$F0,$0F,$84,$FD,$F1,$23,$F7,$55,$C2,$FC,$1F,$51
	.dc.b	$27,$38,$23,$F8,$36,$39,$9F,$E9,$DD,$FF,$D1,$76,$FE,$80,$D3,$F9
	.dc.b	$66,$3E,$4B,$FE,$B7,$F5,$FF,$EA,$DB,$FE,$98,$D7,$FD,$D6,$D7,$F9
	.dc.b	$A3,$5F,$F4,$0F,$AF,$FD,$93,$E3,$3F,$A7,$E1,$FE,$F2,$6C,$F0,$F0
	.dc.b	$FF,$C9,$53,$F1,$B2,$F2,$EB,$83,$39,$6B,$EA,$7F,$DB,$DC,$0B,$46
	.dc.b	$1B,$8F,$E3,$1A,$FF,$65,$9E,$1F,$53,$E0,$24,$7F,$1F,$9D,$E3,$FC
	.dc.b	$F7,$CD,$79,$BF,$E1,$86,$D7,$F0,$76,$BF,$E1,$97,$FC,$14,$E7,$CC
	.dc.b	$FE,$A5,$1F,$E8,$88,$FE,$B4,$7F,$4F,$70,$FF,$0E,$6B,$FC,$E2,$FF
	.dc.b	$C8,$B7,$F6,$3C,$1F,$8E,$7C,$8F,$83,$B5,$E9,$F0,$FE,$59,$87,$C9
	.dc.b	$F1,$1E,$92,$3E,$FF,$A8,$FF,$19,$FC,$57,$FA,$2A,$3F,$B1,$F9,$08
	.dc.b	$C4,$7F,$83,$63,$1A,$7F,$6D,$9C,$CF,$E6,$B4,$79,$5F,$73,$E9,$E4
	.dc.b	$FC,$D8,$FB,$FF,$4F,$F0,$AB,$83,$67,$FB,$14,$23,$11,$F0,$DB,$C7
	.dc.b	$F8,$1E,$9B,$5F,$E2,$43,$C4,$E0,$E3,$4F,$AF,$7C,$47,$F6,$D8,$FA
	.dc.b	$71,$8D,$FF,$C0,$71,$B9,$D3,$FE,$11,$9D,$3F,$E0,$85,$F1,$5F,$BB
	.dc.b	$8D,$35,$F0,$36,$37,$D5,$F9,$5F,$A8,$FC,$D5,$7F,$63,$CF,$F3,$91
	.dc.b	$8C,$FE,$63,$67,$E2,$B5,$FC,$37,$D4,$5A,$FD,$23,$68,$D3,$1F,$4F
	.dc.b	$E0,$E8,$FE,$07,$E9,$DA,$FA,$33,$70,$FC,$94,$DE,$57,$D2,$3C,$4F
	.dc.b	$EF,$55,$FD,$9A,$31,$87,$E2,$9B,$1F,$89,$6B,$FA,$2A,$E7,$E3,$75
	.dc.b	$9F,$F6,$FD,$7E,$CF,$D5,$C4,$F8,$F5,$FE,$01,$F3,$16,$9F,$2D,$63
	.dc.b	$0E,$7E,$9F,$13,$FB,$D6,$E7,$FE,$8F,$CB,$FC,$43,$3E,$7F,$3B,$D7
	.dc.b	$3F,$E4,$39,$CB,$E9,$8C,$E9,$F0,$BD,$0F,$41,$E9,$F4,$2E,$1F,$C2
	.dc.b	$FE,$9F,$CC,$7A,$7F,$48,$FE,$31,$19,$3F,$C5,$CE,$31,$FE,$2C,$D7
	.dc.b	$FA,$2B,$5F,$F2,$33,$5F,$F1,$EA,$FA,$8E,$C7,$7F,$A9,$B5,$FE,$93
	.dc.b	$F9,$D7,$BB,$F7,$70,$46,$27,$FD,$31,$E6,$FF,$5C,$78,$56,$E3,$15
	.dc.b	$FE,$68,$6B,$FE,$91,$6B,$FC,$79,$AB,$D2,$FE,$C3,$E7,$7C,$3F,$03
	.dc.b	$F1,$D1,$8C,$3E,$1C,$63,$D3,$55,$3E,$57,$85,$6B,$E4,$FC,$07,$CD
	.dc.b	$38,$79,$D6,$3F,$B2,$BF,$73,$1F,$EE,$2D,$C0,$D3,$F0,$3D,$4E,$17
	.dc.b	$CA,$FD,$93,$C9,$E7,$5C,$7F,$B8,$BC,$C8,$FF,$80,$5E,$DE,$57,$E0
	.dc.b	$7E,$57,$D8,$D9,$FC,$0F,$C9,$E1,$A7,$F7,$C8,$FC,$47,$C4,$C8,$FE
	.dc.b	$FF,$E6,$5E,$3F,$F3,$69,$F3,$76,$FF,$64,$CF,$F8,$46,$7D,$C4,$7F
	.dc.b	$C5,$FA,$3E,$23,$F2,$9F,$19,$F9,$56,$CF,$E5,$3E,$47,$0D,$3F,$2F
	.dc.b	$1F,$9B,$F9,$4F,$CE,$7C,$8D,$9F,$CD,$BF,$9F,$F9,$1C,$69,$FE,$DF
	.dc.b	$C9,$18,$F3,$7F,$D5,$5F,$7F,$FE,$B5,$C3,$FC,$E4,$DB,$F6,$F8,$8A
	.dc.b	$7B,$BF,$C4,$EB,$51,$FE,$B3,$33,$7F,$D6,$9B,$7F,$D6,$5A,$FF,$11
	.dc.b	$6B,$FE,$7E,$97,$1F,$53,$F7,$1F,$25,$6F,$CA,$78,$7C,$3F,$13,$4F
	.dc.b	$F4,$57,$88,$F2,$73,$18,$6E,$3F,$0F,$E4,$BF,$82,$1F,$C1,$7E,$46
	.dc.b	$D7,$F0,$47,$0F,$F0,$6F,$91,$B5,$F8,$03,$F5,$9F,$0F,$FB,$BF,$04
	.dc.b	$DF,$89,$C3,$6F,$FA,$EC,$A7,$F8,$FD,$7E,$D5,$B1,$99,$DF,$C4,$F5
	.dc.b	$3F,$8D,$7B,$F6,$3F,$D8,$DF,$46,$F7,$7F,$9B,$36,$33,$F0,$78,$9C
	.dc.b	$B2,$38,$FC,$25,$CF,$F3,$7E,$3B,$07,$C9,$74,$F9,$2E,$AD,$C3,$C3
	.dc.b	$F2,$58,$69,$E4,$8F,$EB,$1C,$AF,$86,$F2,$FC,$96,$1B,$5F,$63,$CB
	.dc.b	$FE,$7B,$46,$92,$3F,$C4,$B1,$FC,$21,$FB,$FE,$6F,$C0,$EF,$EE,$FC
	.dc.b	$2F,$4E,$4F,$C6,$F9,$0B,$C1,$EE,$DD,$1F,$20,$F9,$3F,$23,$6B,$E6
	.dc.b	$FC,$8F,$CC,$E7,$F9,$AD,$8F,$A3,$D3,$86,$9C,$06,$3F,$87,$FD,$ED
	.dc.b	$F2,$1F,$D0,$5B,$3A,$F1,$7F,$DE,$9F,$5A,$F3,$75,$7F,$D1,$E3,$15
	.dc.b	$FE,$26,$6E,$1F,$E5,$C3,$BB,$87,$0D,$7F,$CC,$38,$7F,$6C,$C3,$FC
	.dc.b	$43,$08,$C5,$7A,$3A,$7F,$05,$DF,$E7,$F7,$BF,$A3,$FE,$21,$FD,$7E
	.dc.b	$BF,$23,$EF,$E3,$C7,$FC,$8B,$AF,$6F,$FA,$05,$7F,$8D,$AB,$CD,$9F
	.dc.b	$F5,$38,$C3,$FE,$B5,$6B,$FE,$D8,$7F,$C8,$F1,$FE,$B2,$6C,$7D,$9F
	.dc.b	$7F,$C7,$ED,$FF,$89,$F6,$7D,$5F,$FA,$CF,$0F,$98,$7D,$2F,$3D,$F0
	.dc.b	$BF,$CD,$B1,$FE,$73,$7C,$3E,$47,$FC,$5D,$39,$8F,$F2,$2F,$95,$FC
	.dc.b	$83,$DB,$FE,$B5,$E0,$7F,$9B,$B2,$F5,$CE,$5B,$19,$C7,$D9,$EE,$FC
	.dc.b	$83,$F6,$BC,$76,$BD,$9E,$97,$EE,$3A,$7E,$6B,$63,$E0,$7B,$1F,$28
	.dc.b	$8F,$83,$8F,$FA,$53,$7D,$9F,$AA,$7F,$02,$F8,$6B,$1F,$27,$DA,$FE
	.dc.b	$0B,$F0,$FF,$69,$B1,$FE,$DE,$D9,$E4,$7B,$89,$1E,$C7,$ED,$FB,$3F
	.dc.b	$82,$3F,$C8,$9F,$A4,$FE,$7A,$7E,$B4,$7D,$3F,$BF,$7A,$1A,$F8,$EC
	.dc.b	$36,$BE,$AF,$27,$F4,$DA,$7A,$1F,$E4,$2E,$84,$62,$BF,$13,$18,$9D
	.dc.b	$4E,$6B,$F1,$5E,$C7,$E5,$8F,$2F,$5D,$C3,$BE,$5F,$E6,$AF,$53,$FC
	.dc.b	$B9,$F3,$5F,$E4,$1C,$FF,$D0,$73,$9A,$F2,$F8,$CA,$D4,$7F,$BE,$3E
	.dc.b	$BD,$CD,$FE,$1D,$6C,$7E,$11,$BA,$7E,$D7,$B5,$18,$8F,$F6,$9E,$3F
	.dc.b	$E7,$CF,$57,$FD,$00,$DB,$F3,$FC,$53,$D7,$CB,$FC,$D2,$3D,$7F,$29
	.dc.b	$FD,$C9,$FE,$1F,$5F,$EC,$9E,$DE,$73,$9C,$53,$9A,$FA,$39,$BF,$12
	.dc.b	$FD,$E5,$79,$DA,$FF,$BA,$23,$5E,$5F,$F8,$EE,$73,$3E,$0F,$B0,$FF
	.dc.b	$1B,$DF,$CE,$6C,$7F,$89,$B0,$43,$E2,$3D,$EE,$B8,$FE,$8F,$93,$F2
	.dc.b	$1B,$C5,$FE,$56,$FB,$0F,$2C,$63,$2F,$4C,$78,$B2,$FF,$2F,$37,$53
	.dc.b	$4C,$7E,$3A,$F8,$0F,$9F,$F4,$9F,$D9,$EF,$FC,$77,$6A,$D6,$1F,$88
	.dc.b	$AD,$63,$F8,$6F,$B1,$7B,$3F,$E6,$AB,$DE,$3E,$F9,$CE,$1F,$3A,$73
	.dc.b	$7F,$8E,$FB,$0F,$DC,$DF,$C5,$FB,$17,$B9,$F8,$29,$CD,$7E,$52,$73
	.dc.b	$87,$E6,$A7,$2F,$E0,$1B,$7F,$79,$F9,$51,$9E,$7F,$F4,$84,$FF,$54
	.dc.b	$F7,$BF,$AA,$FD,$7F,$F9,$CA,$3F,$27,$58,$F8,$1F,$E6,$F5,$67,$22
	.dc.b	$B8,$3F,$D7,$8F,$F0,$41,$99,$FE,$A9,$CF,$FE,$EE,$9D,$CF,$84,$D1
	.dc.b	$D1,$CB,$CB,$E1,$37,$3F,$EF,$67,$37,$FD,$B1,$CE,$17,$0F,$3D,$F4
	.dc.b	$3F,$47,$1E,$97,$CD,$55,$FF,$55,$E1,$D1,$E0,$62,$B9,$C7,$68,$C9
	.dc.b	$EE,$3F,$AD,$94,$F0,$98,$9C,$FD,$87,$F8,$3B,$EC,$6F,$C1,$9A,$4A
	.dc.b	$FE,$CB,$3F,$EC,$6D,$F8,$FC,$27,$F9,$C6,$72,$D7,$FC,$01,$3B,$73
	.dc.b	$3F,$0D,$FF,$08,$C7,$F9,$E1,$C0,$70,$B4,$FB,$71,$CD,$B5,$E1,$73
	.dc.b	$E7,$FD,$F6,$D7,$4A,$E1,$A7,$E0,$DB,$1E,$2A,$E1,$39,$6C,$79,$AF
	.dc.b	$B4,$F1,$6C,$71,$3B,$4F,$07,$D5,$1A,$6E,$1A,$61,$ED,$AE,$91,$F1
	.dc.b	$76,$7D,$8F,$FA,$1E,$BF,$C2,$B1,$8F,$8B,$FB,$AD,$8C,$CE,$DF,$CE
	.dc.b	$FC,$B3,$DF,$DB,$63,$9B,$5D,$C7,$4F,$A2,$FF,$2F,$47,$81,$AF,$EA
	.dc.b	$1A,$FF,$46,$6B,$FE,$4C,$EC,$7D,$C7,$F3,$C8,$FF,$09,$B5,$FE,$A0
	.dc.b	$D7,$F9,$D3,$5E,$5F,$63,$FE,$24,$7F,$C3,$6D,$7F,$AC,$35,$FE,$84
	.dc.b	$D7,$FA,$7F,$B2,$F8,$FF,$8E,$D8,$F9,$6D,$7F,$81,$B5,$FE,$C3,$F1
	.dc.b	$FF,$E8,$E7,$E8,$B5,$FE,$12,$D7,$7E,$6F,$FA,$AB,$F1,$DF,$54,$32
	.dc.b	$46,$1B,$BE,$CF,$B7,$F6,$2F,$76,$92,$FF,$03,$19,$7F,$07,$E3,$7E
	.dc.b	$95,$BD,$3F,$84,$8C,$47,$F1,$E6,$BF,$D9,$FE,$3F,$FD,$86,$FF,$91
	.dc.b	$9A,$FF,$80,$DA,$EF,$E2,$BF,$D8,$B0,$F8,$AA,$49,$E4,$FF,$10,$E8
	.dc.b	$79,$2E,$0C,$36,$BF,$C4,$ED,$7F,$D0,$D3,$C2,$77,$EC,$C9,$DB,$5F
	.dc.b	$F3,$A6,$5F,$ED,$8F,$81,$FE,$D1,$1F,$E5,$18,$C1,$FE,$DC,$F9,$0F
	.dc.b	$F6,$94,$7F,$99,$1A,$FF,$88,$5A,$FF,$82,$DA,$FE,$6F,$EB,$DF,$37
	.dc.b	$FE,$E6,$6C,$7F,$9B,$9A,$FF,$8A,$DA,$F3,$DF,$C4,$47,$C4,$F9,$7D
	.dc.b	$C9,$1E,$56,$BE,$5E,$6F,$51,$1F,$53,$ED,$F6,$64,$7B,$5A,$FB,$7E
	.dc.b	$03,$E1,$11,$F8,$4F,$AF,$E8,$24,$7A,$DA,$FD,$4F,$C1,$FF,$86,$3F
	.dc.b	$DA,$F8,$51,$88,$FF,$63,$35,$46,$BC,$3F,$A5,$F0,$52,$7E,$93,$5F
	.dc.b	$A5,$E3,$FD,$72,$3F,$5D,$FB,$5E,$3C,$8F,$B4,$D7,$FD,$73,$F8,$2B
	.dc.b	$E9,$FF,$B0,$C2,$30,$FF,$64,$6B,$C2,$37,$0F,$A8,$A4,$9E,$0E,$BE
	.dc.b	$3A,$47,$36,$BF,$ED,$FF,$84,$7C,$43,$E8,$7F,$B9,$C6,$1B,$19,$9F
	.dc.b	$F6,$8C,$E5,$B1,$F8,$A7,$FD,$C0,$D5,$1D,$B9,$F9,$9F,$D1,$3E,$13
	.dc.b	$F9,$E4,$FF,$9E,$9A,$FF,$9C,$9A,$FF,$9C,$1B,$E2,$17,$C7,$73,$24
	.dc.b	$FF,$52,$8C,$BF,$EE,$8F,$91,$FD,$F7,$37,$F8,$41,$B1,$9F,$03,$FD
	.dc.b	$5E,$D9,$FE,$0D,$DC,$F8,$EA,$49,$FF,$55,$C6,$5E,$5D,$5D,$6F,$FD
	.dc.b	$21,$18,$21,$FE,$3F,$18,$7F,$11,$EC,$E0,$48,$F2,$3C,$FE,$66,$47
	.dc.b	$9D,$AF,$DB,$E0,$7D,$4F,$D8,$24,$7D,$67,$ED,$FA,$92,$3E,$CB,$5F
	.dc.b	$20,$FA,$2F,$86,$E1,$AF,$A5,$AF,$F9,$65,$A8,$E5,$9F,$E7,$A3,$E1
	.dc.b	$C7,$38,$C1,$9F,$02,$47,$F9,$12,$7F,$AC,$3D,$63,$F2,$7E,$44,$8F
	.dc.b	$C8,$B5,$F8,$5E,$27,$E6,$08,$FE,$59,$FF,$27,$E7,$E5,$F9,$9B,$26
	.dc.b	$91,$F8,$71,$EB,$F7,$76,$48,$C8,$79,$FC,$B4,$8C,$B5,$F9,$BE,$A3
	.dc.b	$E8,$BF,$80,$B6,$3F,$CC,$6D,$77,$1E,$6F,$98,$91,$E0,$6B,$FE,$86
	.dc.b	$E8,$7C,$B7,$EB,$49,$1F,$A3,$7F,$61,$EA,$C8,$FD,$33,$5F,$C5,$78
	.dc.b	$1F,$C0,$08,$FE,$F9,$FE,$07,$E0,$C8,$FE,$02,$D7,$FD,$ED,$F5,$9F
	.dc.b	$B8,$F5,$4D,$27,$B5,$FC,$33,$C6,$95,$FC,$29,$AF,$E6,$FE,$8B,$F8
	.dc.b	$89,$1F,$E1,$EF,$F6,$79,$91,$FD,$9D,$AF,$EB,$6D,$F4,$7F,$CB,$9B
	.dc.b	$1F,$E9,$06,$BF,$A0,$6B,$FA,$BF,$A9,$FE,$42,$47,$F8,$FB,$C6,$5F
	.dc.b	$2D,$49,$3C,$E3,$2F,$F2,$8F,$86,$FE,$7F,$E4,$9A,$48,$FE,$5F,$F5
	.dc.b	$31,$FE,$5A,$D7,$FC,$05,$C3,$FE,$07,$23,$FE,$07,$7F,$A5,$FD,$5B
	.dc.b	$D0,$A4,$9F,$F5,$7C,$65,$BE,$4F,$02,$38,$3F,$94,$7C,$CB,$5F,$E5
	.dc.b	$0D,$8C,$C6,$0C,$F0,$48,$E2,$D7,$F8,$B7,$D5,$7F,$4C,$23,$FD,$29
	.dc.b	$EB,$CB,$23,$C2,$D7,$F9,$47,$90,$C3,$FB,$14,$79,$B9,$BC,$A4,$7F
	.dc.b	$AA,$FA,$DF,$D6,$88,$E0,$FF,$4B,$FE,$06,$C3,$FD,$2E,$D8,$CC,$60
	.dc.b	$FE,$91,$1F,$72,$3C,$6D,$7F,$DF,$58,$23,$11,$FC,$8E,$31,$C9,$CB
	.dc.b	$CF,$93,$FD,$56,$BF,$CE,$75,$C6,$7F,$B8,$FD,$04,$AF,$ED,$ED,$7F
	.dc.b	$A5,$72,$3F,$04,$7B,$BF,$CC,$67,$F1,$C7,$F3,$39,$D4,$65,$93,$C2
	.dc.b	$74,$6D,$C3,$C8,$FC,$C2,$F9,$8E,$51,$B6,$BA,$6C,$F9,$5C,$C3,$EB
	.dc.b	$5E,$27,$F9,$6A,$30,$D8,$CF,$8D,$FD,$52,$72,$D8,$78,$7F,$E9,$A6
	.dc.b	$E9,$98,$C1,$FE,$B7,$C5,$FB,$93,$A3,$E2,$C7,$C3,$8C,$3F,$0B,$1F
	.dc.b	$DB,$46,$1F,$85,$1C,$A7,$8C,$7C,$DF,$98,$FF,$45,$4F,$10,$E2,$38
	.dc.b	$B6,$3E,$93,$63,$FC,$53,$F5,$DF,$3D,$B3,$FE,$2E,$8C,$47,$F8,$B3
	.dc.b	$D2,$46,$34,$FF,$05,$FD,$0D,$E3,$34,$98,$7F,$5C,$87,$D6,$6C,$91
	.dc.b	$97,$BE,$DF,$9B,$EC,$BC,$C6,$92,$BF,$C8,$7F,$5D,$E6,$F0,$FF,$C9
	.dc.b	$AD,$D4,$D3,$D7,$93,$E0,$1D,$98,$6E,$1F,$E3,$96,$F0,$7F,$02,$6B
	.dc.b	$FB,$05,$E2,$71,$24,$F1,$7C,$51,$23,$FC,$B2,$D9,$E9,$E0,$F8,$49
	.dc.b	$3E,$17,$C9,$E4,$48,$F2,$36,$7F,$C5,$3E,$EF,$F9,$AE,$7F,$30,$D8
	.dc.b	$CB,$F6,$9F,$F3,$83,$4F,$E9,$71,$C3,$BF,$04,$7F,$D8,$9E,$5F,$E0
	.dc.b	$BC,$97,$A2,$8C,$36,$76,$7F,$9D,$BE,$93,$E9,$11,$C1,$FE,$7C,$FA
	.dc.b	$DF,$D5,$B7,$BB,$ED,$46,$33,$FF,$44,$35,$FE,$1D,$8F,$F6,$EF,$7F
	.dc.b	$C9,$E6,$FF,$B4,$B0,$FE,$C5,$18,$3F,$09,$D3,$FD,$86,$3F,$D1,$6D
	.dc.b	$BF,$F4,$A3,$5F,$E3,$5C,$E8,$C4,$7F,$93,$70,$FE,$FB,$A7,$80,$97
	.dc.b	$FD,$C5,$F8,$DF,$AE,$FE,$EE,$B8,$35,$52,$30,$D8,$F3,$FD,$1C,$BF
	.dc.b	$3B,$63,$CF,$F5,$FE,$C2,$4F,$B1,$F8,$7C,$92,$3F,$C0,$8D,$9F,$7E
	.dc.b	$FF,$10,$93,$F1,$1F,$8F,$E2,$C8,$F8,$ED,$9F,$F5,$D0,$C3,$F6,$0D
	.dc.b	$9F,$F5,$BB,$4D,$C6,$B0,$93,$A6,$CF,$FA,$57,$C0,$FF,$72,$E3,$FC
	.dc.b	$37,$E1,$BF,$C0,$57,$FE,$64,$6D,$6D,$AF,$FA,$77,$E9,$FF,$79,$3F
	.dc.b	$CE,$5B,$1F,$E8,$E6,$BB,$E9,$ED,$D3,$2B,$B3,$67,$B7,$1F,$8C,$93
	.dc.b	$E3,$7C,$DD,$32,$3C,$CD,$9F,$C6,$FB,$9F,$92,$24,$FE,$49,$FC,$AF
	.dc.b	$D4,$C8,$FC,$AB,$67,$D1,$F0,$1F,$B9,$CF,$FD,$5F,$5E,$D8,$C1,$9C
	.dc.b	$64,$7C,$16,$D7,$C1,$C7,$91,$27,$E3,$3F,$A6,$F2,$E4,$7E,$99,$B3
	.dc.b	$FA,$6F,$84,$FD,$59,$27,$F5,$6F,$F2,$CF,$C1,$1C,$34,$93,$FE,$E3
	.dc.b	$8C,$BF,$EE,$5F,$C0,$DF,$0D,$FC,$D1,$B1,$98,$C1,$9F,$46,$47,$16
	.dc.b	$E3,$FE,$DD,$E3,$7D,$9F,$FB,$DA,$BF,$D9,$B1,$8F,$88,$FE,$B6,$D8
	.dc.b	$CC,$EE,$BF,$C7,$F8,$BE,$5B,$C8,$4A,$F4,$3F,$A0,$EB,$91,$FA,$06
	.dc.b	$D7,$E9,$BA,$FF,$4A,$49,$FD,$2B,$FC,$53,$E0,$24,$7F,$14,$6C,$FF
	.dc.b	$14,$F8,$2F,$E3,$24,$9F,$E3,$2F,$F3,$BE,$0F,$8F,$3F,$EC,$8D,$F0
	.dc.b	$4A,$FF,$68,$FC,$4E,$C8,$E0,$FF,$6A,$F0,$73,$23,$83,$FD,$2D,$E0
	.dc.b	$7F,$40,$6F,$83,$FC,$78,$8E,$BD,$9F,$E4,$24,$70,$72,$F6,$BF,$92
	.dc.b	$91,$C1,$F5,$BE,$9F,$F9,$D6,$7F,$6D,$AF,$F9,$37,$91,$18,$8F,$F8
	.dc.b	$01,$18,$EA,$FE,$AD,$F1,$BE,$79,$FF,$5A,$35,$FE,$92,$D7,$F9,$E3
	.dc.b	$5F,$E6,$B9,$BC,$0F,$E7,$64,$C7,$E1,$BF,$CF,$BD,$29,$1F,$19,$B5
	.dc.b	$FE,$C3,$AF,$E2,$AF,$F4,$2F,$0F,$FA,$29,$2B,$E6,$F9,$DF,$D2,$08
	.dc.b	$E0,$FA,$3E,$F7,$F4,$C2,$38,$3F,$D9,$59,$7F,$1C,$BF,$5F,$8B,$F0
	.dc.b	$89,$3F,$6B,$D0,$FE,$AA,$47,$07,$DE,$F7,$7F,$AC,$91,$C1,$FA,$AF
	.dc.b	$7B,$FC,$A9,$7F,$4D,$AF,$D2,$6A,$E0,$FF,$54,$7B,$7F,$B0,$9F,$F6
	.dc.b	$35,$73,$F8,$0C,$11,$FF,$72,$7D,$5B,$C5,$FF,$08,$36,$33,$18,$33
	.dc.b	$DB,$23,$8B,$72,$FF,$4A,$1F,$DE,$FE,$15,$F1,$9F,$E4,$E6,$D6,$A7
	.dc.b	$53,$93,$FC,$A7,$F5,$7E,$18,$E7,$E9,$23,$11,$FE,$A1,$8C,$63,$C0
	.dc.b	$FF,$14,$7D,$5F,$F8,$2A,$7F,$C0,$2D,$F4,$CD,$35,$4F,$AB,$7A,$5F
	.dc.b	$E6,$46,$E7,$E7,$9D,$4E,$4F,$F3,$4F,$73,$6A,$4A,$E1,$19,$7E,$06
	.dc.b	$2F,$AB,$F2,$12,$3F,$C8,$0F,$9F,$EA,$7F,$D7,$F1,$C1,$BA,$7C,$EA
	.dc.b	$46,$23,$FC,$6E,$D4,$7F,$8B,$7E,$A9,$F6,$1F,$E8,$06,$DF,$B5,$B3
	.dc.b	$F4,$1A,$FC,$FF,$AA,$7D,$3F,$FA,$21,$B5,$F1,$5B,$3F,$61,$AF,$D7
	.dc.b	$FA,$A7,$9B,$FE,$90,$6D,$7C,$D6,$CF,$E0,$1A,$FF,$B1,$7E,$A7,$D7
	.dc.b	$F3,$6F,$98,$FC,$2D,$E1,$FC,$64,$DA,$FC,$34,$D7,$FD,$D1,$15,$FF
	.dc.b	$28,$45,$51,$9E,$3F,$F7,$6D,$F7,$9F,$CE,$CE,$6B,$83,$61,$3B,$7B
	.dc.b	$98,$BF,$3B,$4E,$65,$3E,$BD,$F8,$23,$D7,$19,$3C,$E5,$3C,$E6,$63
	.dc.b	$83,$D0,$F9,$07,$FB,$B4,$D9,$83,$F6,$FF,$51,$F7,$3E,$A7,$FD,$E4
	.dc.b	$D8,$FD,$B3,$5D,$EA,$9D,$FC,$E7,$F1,$19,$DB,$5E,$AD,$7F,$26,$29
	.dc.b	$19,$8F,$CB,$C6,$3C,$1F,$CC,$FD,$57,$CB,$87,$8D,$CA,$30,$DC,$FA
	.dc.b	$CF,$1B,$57,$1C,$3C,$CF,$08,$EB,$5C,$2F,$94,$78,$AB,$73,$D1,$F7
	.dc.b	$69,$C6,$7E,$C6,$1E,$47,$F1,$CF,$86,$3C,$18,$F6,$9F,$6E,$1D,$9F
	.dc.b	$84,$DA,$F4,$63,$C2,$7D,$38,$F9,$5E,$47,$02,$CF,$08,$FC,$4F,$D5
	.dc.b	$3E,$1F,$FB,$D7,$06,$1B,$7F,$EA,$F9,$D3,$FD,$F2,$7F,$23,$F8,$A1
	.dc.b	$DF,$03,$EA,$DE,$F6,$1B,$7C,$4F,$F5,$D3,$87,$FA,$A5,$78,$5D,$53
	.dc.b	$EA,$FF,$DB,$B1,$86,$BC,$0F,$F6,$23,$8D,$FC,$91,$8F,$57,$D7,$D8
	.dc.b	$F9,$C7,$C1,$E1,$B5,$E5,$9D,$4F,$B3,$0E,$99,$7F,$B3,$FE,$AD,$F2
	.dc.b	$BF,$D5,$70,$FF,$64,$C6,$1F,$E7,$2D,$7F,$99,$35,$FE,$A9,$F1,$CF
	.dc.b	$73,$E4,$B6,$3E,$0B,$5F,$76,$4A,$E9,$AF,$BB,$67,$BA,$FA,$1F,$A6
	.dc.b	$BD,$2A,$62,$3E,$1F,$D5,$BE,$47,$D5,$92,$31,$9F,$F5,$26,$BF,$D2
	.dc.b	$DA,$FF,$6D,$FC,$0F,$FC,$FD,$C1,$FC,$51,$B1,$FE,$F7,$6E,$67,$FD
	.dc.b	$F5,$1F,$3F,$E8,$9F,$53,$F8,$AF,$84,$FF,$01,$CF,$E0,$9A,$FC,$29
	.dc.b	$DB,$C7,$E8,$7D,$30,$69,$23,$FB,$5F,$E0,$3F,$C2,$9F,$BC,$D7,$E6
	.dc.b	$B7,$87,$F1,$3F,$29,$FD,$93,$E6,$FF,$47,$0F,$AA,$FA,$8E,$BE,$4D
	.dc.b	$F2,$BF,$A8,$B4,$63,$1F,$D0,$57,$1E,$9F,$C9,$FD,$57,$FA,$BA,$7F
	.dc.b	$BF,$B5,$FE,$F0,$D7,$F9,$1F,$22,$31,$1F,$DE,$21,$F7,$8F,$B4,$FC
	.dc.b	$FB,$67,$F3,$8D,$7F,$51,$C8,$FB,$CF,$CE,$C7,$F9,$BB,$BB,$D5,$2F
	.dc.b	$A2,$A7,$EA,$BE,$2F,$F2,$3D,$3E,$15,$F1,$1F,$8A,$6E,$1F,$E9,$68
	.dc.b	$C1,$9F,$2B,$91,$23,$F8,$27,$13,$17,$D4,$FE,$E3,$F5,$E7,$CE,$F9
	.dc.b	$3E,$D2,$61,$FB,$79,$CB,$7E,$5F,$F3,$70,$69,$23,$FC,$5B,$0F,$97
	.dc.b	$EC,$48,$F1,$E2,$EB,$F1,$7C,$F7,$C4,$FA,$59,$5E,$18,$F3,$7C,$47
	.dc.b	$80,$91,$D9,$FE,$0D,$D6,$FA,$EF,$93,$3F,$E7,$36,$C6,$79,$BF,$98
	.dc.b	$FD,$1F,$E6,$F0,$F8,$5F,$01,$F5,$3E,$DE,$63,$F0,$63,$E3,$74,$FC
	.dc.b	$02,$47,$B9,$F8,$7C,$9F,$2C,$91,$EE,$F9,$DF,$BF,$E0,$7F,$40,$87
	.dc.b	$DE,$FF,$0C,$7F,$9C,$18,$F3,$F6,$E3,$EF,$1F,$31,$FE,$09,$6C,$7F
	.dc.b	$25,$6B,$FE,$7B,$6B,$FE,$82,$FC,$5F,$F8,$3A,$69,$88,$E3,$C3,$FB
	.dc.b	$4E,$6F,$25,$7F,$41,$EC,$FC,$46,$3F,$98,$3F,$33,$E6,$7E,$CE,$1F
	.dc.b	$41,$49,$5F,$7A,$B5,$F0,$5F,$D1,$BF,$01,$FE,$87,$3F,$CC,$1B,$97
	.dc.b	$F0,$4C,$3F,$C7,$33,$FD,$1E,$73,$E2,$7F,$9A,$A4,$7F,$18,$AE,$1F
	.dc.b	$29,$D7,$DA,$FE,$DB,$9F,$F0,$6F,$0E,$F9,$5F,$DB,$B3,$FE,$B9,$28
	.dc.b	$C6,$1F,$D7,$23,$84,$FF,$6B,$75,$E2,$FF,$53,$6C,$7F,$AA,$5A,$FF
	.dc.b	$AC,$1A,$FF,$8B,$9F,$E2,$F9,$F9,$3C,$CF,$ED,$33,$E2,$6D,$F8,$70
	.dc.b	$F1,$D7,$FB,$FB,$EF,$1F,$3D,$FC,$99,$B8,$7F,$20,$6B,$F5,$F6,$7D
	.dc.b	$9F,$D5,$F6,$64,$FF,$02,$9F,$EB,$4F,$1B,$FB,$5E,$D7,$DB,$F9,$A9
	.dc.b	$3B,$8A,$F5,$F8,$8F,$C0,$18,$D7,$FC,$07,$AF,$FA,$7B,$E4,$3F,$3D
	.dc.b	$F0,$3F,$BD,$BF,$CE,$B5,$FE,$DE,$7F,$A8,$8C,$F9,$FF,$E6,$58,$34
	.dc.b	$91,$FE,$03,$8F,$E3,$2F,$4F,$0B,$B9,$E0,$9F,$CA,$7C,$57,$F1,$DE
	.dc.b	$B9,$5F,$06,$2B,$F1,$FE,$33,$F3,$46,$33,$FF,$0E,$F9,$2F,$BD,$FF
	.dc.b	$66,$57,$F8,$B6,$31,$F2,$BF,$C2,$A7,$33,$F8,$CD,$9F,$47,$F9,$0F
	.dc.b	$87,$93,$F8,$B9,$47,$1B,$FD,$57,$DD,$7E,$5B,$E9,$25,$FD,$78,$ED
	.dc.b	$F3,$3F,$E2,$3C,$FC,$8D,$CF,$FD,$CF,$5F,$EF,$B9,$CB,$F9,$5C,$BF
	.dc.b	$52,$FE,$D9,$BD,$DF,$A1,$F0,$7A,$CF,$EB,$F2,$FD,$B1,$1F,$C4,$72
	.dc.b	$FE,$FF,$EE,$E4,$7E,$2D,$D3,$FE,$2A,$FC,$AF,$F7,$2D,$F1,$BF,$C8
	.dc.b	$B1,$8E,$2F,$DF,$7B,$FF,$C2,$7E,$EA,$5F,$E5,$5D,$DF,$0F,$07,$FC
	.dc.b	$53,$87,$E7,$2B,$FD,$F2,$D7,$FD,$E4,$D7,$FD,$3B,$E0,$7F,$06,$F8
	.dc.b	$7F,$E2,$C4,$8F,$F2,$94,$F1,$CB,$F8,$87,$C3,$7F,$1B,$24,$FE,$9F
	.dc.b	$E3,$FD,$3E,$0C,$8E,$2E,$9F,$81,$E0,$F9,$88,$F9,$9F,$F2,$D6,$5F
	.dc.b	$C6,$3E,$DF,$F9,$41,$27,$C7,$3F,$E6,$AF,$A9,$4F,$F9,$21,$B3,$FE
	.dc.b	$B8,$3F,$C2,$8E,$1C,$73,$23,$11,$F0,$AD,$F1,$FF,$CD,$C9,$1F,$E7
	.dc.b	$39,$E5,$A7,$F3,$1F,$BC,$FE,$78,$49,$FF,$04,$FE,$9B,$FE,$50,$C5
	.dc.b	$BB,$B7,$EE,$92,$BF,$87,$5F,$F5,$FF,$B9,$FA,$91,$FE,$D0,$6A,$0E
	.dc.b	$1B,$C7,$FE,$E7,$FC,$5B,$FD,$F5,$9F,$F2,$86,$F9,$BF,$ED,$A6,$D7
	.dc.b	$3B,$FB,$93,$FD,$1F,$4F,$F3,$1C,$FF,$5B,$7F,$CD,$F7,$FE,$52,$6C
	.dc.b	$EF,$D4,$FF,$27,$47,$F5,$6A,$FC,$5B,$5E,$9D,$BF,$72,$31,$1F,$D5
	.dc.b	$17,$6F,$22,$47,$CA,$9F,$EA,$BD,$DF,$DA,$23,$FA,$77,$7F,$F9,$57
	.dc.b	$2F,$EA,$4D,$CB,$E3,$57,$F6,$E3,$FA,$EC,$7F,$76,$AF,$E9,$CD,$76
	.dc.b	$D7,$FB,$AB,$79,$19,$3E,$CB,$DE,$4A,$FB,$FF,$33,$F0,$BE,$46,$47
	.dc.b	$E0,$9F,$8B,$CD,$F1,$BC,$6F,$8F,$E3,$CA,$FC,$3C,$79,$BD,$DF,$ED
	.dc.b	$1A,$1A,$49,$FC,$E1,$F2,$7E,$BB,$E8,$91,$FC,$8F,$A3,$F5,$3E,$BB
	.dc.b	$F1,$11,$FD,$AB,$7F,$5C,$63,$87,$EB,$B6,$7F,$04,$D7,$FD,$D9,$FB
	.dc.b	$9A,$31,$28,$C4,$FE,$56,$3A,$46,$23,$FB,$CC,$6E,$7A,$43,$EE,$BF
	.dc.b	$CC,$1C,$7F,$ED,$96,$C7,$B9,$AF,$EE,$DA,$FB,$1F,$F3,$67,$3F,$1F
	.dc.b	$8D,$FC,$CC,$E5,$AF,$EA,$67,$6F,$F0,$5C,$1F,$75,$FC,$29,$F5,$47
	.dc.b	$AE,$30,$FA,$9B,$1F,$25,$B1,$FA,$9F,$81,$FF,$7F,$CF,$E8,$9B,$19
	.dc.b	$8E,$AF,$E8,$63,$2F,$E8,$FE,$C5,$3F,$C2,$9B,$1F,$B3,$FB,$1B,$D5
	.dc.b	$FB,$39,$B5,$FA,$0F,$A4,$79,$9F,$E2,$8C,$30,$D7,$FC,$A3,$1A,$9F
	.dc.b	$F1,$A3,$63,$F9,$47,$0F,$E8,$3E,$C7,$AC,$7E,$DF,$E3,$2C,$E6,$FF
	.dc.b	$8F,$C6,$3B,$9F,$71,$4C,$7C,$A5,$AF,$E4,$BF,$05,$B3,$9F,$E9,$FF
	.dc.b	$02,$32,$7F,$92,$FE,$E5,$EE,$FA,$7A,$11,$89,$FF,$26,$37,$5F,$F2
	.dc.b	$63,$5F,$F2,$63,$FA,$B8,$E9,$F1,$9E,$CA,$E7,$DD,$69,$3E,$4F,$86
	.dc.b	$97,$FE,$C4,$CB,$FD,$DA,$D8,$F4,$FC,$DF,$CD,$E1,$56,$B3,$F9,$6D
	.dc.b	$7F,$B7,$B5,$FF,$3B,$3F,$EC,$5B,$ED,$F2,$DF,$13,$B3,$E2,$BF,$4A
	.dc.b	$9F,$46,$F6,$30,$DB,$FF,$52,$F9,$7F,$C5,$DF,$F5,$3C,$66,$9D,$7F
	.dc.b	$E6,$CA,$D4,$7F,$0C,$6B,$FD,$A5,$AF,$F0,$A6,$BF,$E8,$3F,$49,$EE
	.dc.b	$FE,$1B,$D4,$FE,$BE,$FD,$CF,$7F,$13,$AC,$BE,$54,$65,$F9,$9F,$27
	.dc.b	$70,$FF,$55,$6E,$FB,$8F,$C2,$9D,$7C,$EF,$EE,$B7,$F4,$E3,$1E,$C7
	.dc.b	$E7,$74,$FE,$F8,$DA,$DE,$1F,$E0,$06,$BF,$4B,$E5,$BF,$22,$79,$BE
	.dc.b	$02,$E3,$FE,$20,$7A,$6A,$69,$23,$FC,$5B,$EF,$FB,$8F,$CD,$FC,$A5
	.dc.b	$8F,$C4,$38,$7F,$15,$F0,$97,$1F,$7F,$CA,$7E,$88,$F8,$FF,$05,$6B
	.dc.b	$FC,$62,$F8,$88,$FF,$8F,$1F,$C8,$70,$3E,$27,$F5,$67,$D0,$DE,$D7
	.dc.b	$F8,$E9,$C3,$F9,$BF,$89,$B8,$FD,$0F,$98,$FD,$99,$F6,$FE,$47,$FA
	.dc.b	$05,$7E,$86,$31,$EA,$48,$FD,$27,$CC,$5B,$FF,$63,$C3,$D0,$FD,$73
	.dc.b	$4F,$F4,$20,$F8,$3C,$C7,$F0,$EF,$6B,$F9,$0C,$7F,$17,$6C,$FF,$9D
	.dc.b	$CE,$05,$8F,$83,$F2,$FD,$B1,$FE,$16,$D9,$FF,$6A,$1F,$E3,$F2,$F4
	.dc.b	$77,$5C,$BE,$68,$C3,$23,$8F,$BD,$73,$FD,$4C,$7F,$8C,$FA,$9F,$1A
	.dc.b	$F6,$BF,$8D,$F9,$9F,$A3,$8C,$57,$F0,$98,$FE,$3F,$E5,$E6,$AC,$FF
	.dc.b	$8F,$7C,$4F,$F6,$2D,$F0,$C7,$B9,$FE,$41,$F9,$27,$AB,$FC,$0D,$B7
	.dc.b	$C4,$FE,$FE,$BD,$8D,$4E,$DF,$E8,$54,$9D,$57,$F8,$32,$31,$3A,$9C
	.dc.b	$C7,$F9,$A3,$07,$8B,$E1,$E5,$FE,$01,$18,$C3,$C5,$5E,$79,$FE,$7B
	.dc.b	$E8,$FF,$3E,$F9,$4B,$5F,$CA,$DF,$E9,$7E,$CB,$D3,$CE,$1F,$03,$D0
	.dc.b	$FE,$B7,$18,$CB,$FB,$1C,$7F,$5E,$AE,$27,$F5,$68,$7D,$9B,$D3,$CE
	.dc.b	$5E,$49,$CC,$6E,$75,$3E,$8B,$FE,$D3,$F2,$B9,$AF,$E8,$DF,$20,$F1
	.dc.b	$A9,$8F,$4A,$E1,$F0,$34,$FF,$71,$47,$F5,$7A,$FE,$61,$AF,$F8,$CA
	.dc.b	$73,$5F,$E3,$A6,$C7,$FA,$5D,$FF,$75,$75,$F5,$9E,$95,$F9,$66,$DF
	.dc.b	$ED,$7D,$DF,$93,$F1,$56,$7D,$5C,$08,$F9,$BF,$29,$9C,$FE,$64,$A3
	.dc.b	$1C,$3F,$32,$1F,$1F,$F5,$0F,$AB,$F2,$97,$1E,$AE,$1E,$FF,$21,$71
	.dc.b	$FC,$B8,$FD,$F7,$FC,$53,$8F,$F8,$6A,$31,$F2,$3F,$D6,$FF,$03,$79
	.dc.b	$BF,$D7,$6F,$78,$FF,$96,$27,$2F,$E5,$5B,$7F,$E6,$26,$BF,$EB,$F6
	.dc.b	$BF,$EA,$7E,$7F,$F1,$1F,$81,$D7,$9B,$F1,$ED,$BF,$CB,$F8,$9F,$BA
	.dc.b	$F1,$BF,$70,$7E,$53,$E5,$6C,$EA,$9F,$1B,$FD,$3D,$B1,$FE,$74,$BF
	.dc.b	$E0,$73,$93,$5F,$19,$F9,$EF,$93,$FE,$08,$DA,$FC,$A7,$D9,$3E,$43
	.dc.b	$F3,$4D,$8F,$D5,$FA,$0F,$9A,$EF,$C1,$FC,$13,$95,$7B,$8F,$E0,$FA
	.dc.b	$7F,$14,$F4,$3C,$C4,$C7,$CD,$B7,$F3,$CA,$FF,$71,$AE,$DF,$D9,$FC
	.dc.b	$07,$F1,$96,$D7,$C1,$F2,$FF,$8E,$1F,$B8,$F9,$7B,$1F,$17,$E3,$3E
	.dc.b	$21,$1E,$EF,$C7,$F8,$BF,$DF,$FB,$FF,$C9,$5B,$3F,$37,$E2,$BF,$94
	.dc.b	$1F,$C1,$3E,$5E,$C7,$D6,$FC,$7F,$F6,$FB,$7F,$BE,$9A,$FF,$A7,$B8
	.dc.b	$11,$88,$FE,$6C,$D5,$1A,$F8,$BF,$E2,$3E,$57,$F8,$66,$BF,$C7,$1C
	.dc.b	$9F,$A3,$8C,$7C,$54,$C7,$F5,$BE,$55,$CB,$E7,$FC,$FF,$F3,$27,$FD
	.dc.b	$5E,$76,$70,$F9,$5A,$FF,$49,$6B,$FE,$34,$FB,$8F,$D8,$D7,$F1,$3F
	.dc.b	$BD,$FE,$85,$97,$FA,$02,$73,$5F,$C9,$67,$35,$FD,$4F,$E2,$9F,$15
	.dc.b	$FE,$33,$6C,$7F,$1D,$DD,$C3,$9A,$FF,$1C,$7D,$CB,$E2,$BF,$C5,$4D
	.dc.b	$BF,$CE,$37,$2F,$6E,$08,$C4,$7D,$17,$BB,$FD,$63,$DD,$FE,$67,$E8
	.dc.b	$76,$C7,$FD,$17,$18,$C3,$FA,$3A,$FB,$97,$C6,$E9,$B3,$A7,$FC,$88
	.dc.b	$D8,$FF,$28,$7E,$3C,$F9,$3F,$F2,$13,$67,$F6,$8D,$CB,$D2,$D7,$FD
	.dc.b	$CB,$83,$EE,$7F,$CE,$51,$E9,$5C,$DE,$87,$3C,$9F,$EB,$4A,$E1,$3E
	.dc.b	$CF,$13,$E3,$4E,$5B,$8F,$F8,$1F,$D4,$FA,$2F,$C0,$8C,$E3,$F9,$49
	.dc.b	$CB,$F3,$9A,$FE,$75,$AF,$3E,$3F,$F3,$07,$DC,$3E,$1F,$FC,$E4,$DC
	.dc.b	$3E,$6F,$DC,$3C,$FC,$D7,$D2,$8C,$4E,$A7,$1E,$76,$5F,$AF,$F7,$1F
	.dc.b	$E1,$E9,$FA,$ED,$CB,$EF,$7D,$C7,$D4,$7F,$01,$19,$DB,$E3,$4E,$5B
	.dc.b	$D7,$F8,$28,$FC,$3C,$62,$FF,$C1,$36,$27,$CD,$5F,$B7,$FE,$4D,$1F
	.dc.b	$62,$71,$E8,$FE,$C2,$15,$87,$42,$B8,$BD,$4D,$99,$9F,$DC,$F2,$7F
	.dc.b	$BC,$27,$73,$E2,$34,$74,$72,$F2,$EB,$DC,$FF,$BE,$1C,$DF,$AD,$CE
	.dc.b	$17,$0F,$44,$F9,$EF,$E0,$5F,$86,$73,$1E,$03,$35,$B7,$C9,$87,$47
	.dc.b	$46,$27,$D7,$3B,$8F,$79,$FE,$3F,$BF,$F2,$03,$ED,$30,$7C,$BF,$B7
	.dc.b	$FC,$47,$DC,$7F,$91,$5B,$19,$DD,$39,$6C,$71,$F4,$3F,$D9,$ED,$D7
	.dc.b	$94,$62,$3F,$93,$DF,$F9,$AA,$72,$7C,$35,$EE,$FD,$7D,$3E,$CC,$62
	.dc.b	$B5,$7D,$A7,$2D,$D3,$EF,$56,$EB,$E5,$36,$B7,$87,$23,$EA,$4F,$E4
	.dc.b	$67,$86,$DE,$0C,$7F,$10,$F0,$93,$6F,$51,$E9,$9F,$46,$1E,$2C,$7C
	.dc.b	$93,$D5,$AB,$0F,$24,$FA,$2B,$8C,$EF,$1F,$CE,$6C,$FB,$7F,$F2,$F4
	.dc.b	$FF,$2A,$6B,$FC,$C5,$B3,$FC,$5D,$79,$5F,$A1,$8C,$7C,$25,$E1,$FC
	.dc.b	$8F,$CD,$BE,$03,$18,$FF,$A7,$63,$0F,$7A,$CE,$4F,$B7,$FF,$12,$CF
	.dc.b	$F3,$C6,$BF,$D0,$DB,$8F,$F2,$95,$ED,$72,$F4,$5F,$23,$FE,$93,$C1
	.dc.b	$18,$C3,$F7,$D1,$89,$D4,$E6,$7F,$AA,$FD,$BF,$07,$FC,$68,$D7,$31
	.dc.b	$FE,$9F,$F2,$D3,$99,$FE,$87,$F2,$EF,$8D,$F6,$B6,$3F,$AB,$B5,$FF
	.dc.b	$17,$CF,$5D,$3F,$A2,$53,$C4,$F8,$AD,$AF,$EB,$ED,$7F,$B3,$3F,$95
	.dc.b	$97,$0F,$FB,$37,$ED,$9F,$51,$E7,$6D,$FF,$93,$DA,$FF,$74,$6B,$FD
	.dc.b	$73,$ED,$9F,$6F,$F5,$97,$D0,$D3,$11,$F6,$BE,$4B,$FD,$11,$5F,$DD
	.dc.b	$1A,$FF,$7B,$6E,$5F,$D8,$5F,$F4,$54,$7F,$9A,$B8,$7F,$0A,$FF,$81
	.dc.b	$9A,$FF,$8D,$5A,$FF,$BA,$5A,$BE,$23,$FA,$07,$CD,$7E,$41,$FE,$85
	.dc.b	$E5,$59,$FE,$BC,$7F,$AD,$C4,$FE,$5B,$8D,$E3,$BE,$7F,$0D,$AF,$F3
	.dc.b	$FB,$5F,$F6,$77,$CB,$BE,$8B,$97,$C5,$FF,$B4,$E7,$FC,$F8,$D7,$FC
	.dc.b	$42,$D7,$FC,$A0,$D7,$FC,$79,$F0,$7E,$9E,$BE,$DC,$1F,$ED,$86,$F3
	.dc.b	$6A,$BF,$D2,$3F,$68,$F8,$1F,$53,$63,$31,$FA,$EF,$B4,$7D,$5F,$ED
	.dc.b	$27,$FD,$03,$F3,$DF,$CE,$BD,$73,$49,$1F,$EF,$DF,$C4,$DF,$3F,$FD
	.dc.b	$01,$B1,$98,$C1,$FD,$AD,$FA,$9E,$87,$FB,$85,$FE,$DD,$D1,$FC,$83
	.dc.b	$E4,$B4,$D8,$CC,$ED,$FE,$E9,$F0,$AF,$8C,$FE,$9E,$D8,$EC,$D7,$71
	.dc.b	$FD,$53,$F8,$03,$F8,$2C,$78,$DA,$FF,$7F,$6B,$FE,$70,$6B,$FE,$C1
	.dc.b	$AF,$C6,$DF,$F3,$6F,$37,$CF,$18,$3F,$23,$D7,$FE,$88,$FA,$6F,$E0
	.dc.b	$9F,$16,$69,$30,$F6,$F9,$5C,$88,$E0,$F8,$3F,$45,$D8,$8E,$0F,$F6
	.dc.b	$35,$7E,$61,$BE,$07,$80,$8F,$CE,$FD,$4B,$FB,$AE,$86,$92,$3E,$B7
	.dc.b	$CF,$3A,$FF,$8A,$76,$7F,$0A,$94,$E7,$CB,$FE,$76,$D8,$FF,$17,$4E
	.dc.b	$F0,$ED,$7F,$8D,$FC,$45,$F4,$5F,$E0,$66,$DF,$E1,$5A,$FF,$8C,$5A
	.dc.b	$FF,$8B,$70,$F2,$74,$F8,$2F,$F2,$B3,$E0,$EF,$FF,$1F,$56,$A3,$FC
	.dc.b	$83,$7D,$32,$F2,$69,$FC,$6A,$78,$1F,$C1,$DF,$E3,$6E,$33,$F8,$DE
	.dc.b	$57,$F0,$99,$E1,$2A,$E7,$39,$F3,$64,$F1,$AE,$6E,$F2,$FF,$29,$61
	.dc.b	$F4,$7E,$C3,$A1,$27,$87,$B3,$DF,$CB,$91,$CD,$D3,$CC,$FA,$7B,$7F
	.dc.b	$1A,$71,$B7,$D9,$F2,$FE,$31,$E9,$F1,$FF,$B8,$FC,$CF,$DD,$F9,$AF
	.dc.b	$DA,$36,$FF,$CC,$0D,$7F,$DA,$2D,$7E,$96,$1F,$8F,$7C,$5A,$7E,$2F
	.dc.b	$C5,$F8,$C7,$D5,$E7,$F9,$18,$7C,$9F,$56,$5F,$C5,$8A,$F2,$E1,$7D
	.dc.b	$87,$CF,$33,$CB,$F6,$3E,$6B,$E9,$12,$BE,$77,$CC,$FD,$5E,$A9,$1F
	.dc.b	$45,$D3,$F1,$C7,$E1,$BD,$9F,$F4,$2B,$5D,$36,$FF,$0F,$E2,$FF,$25
	.dc.b	$8F,$E0,$73,$FE,$8B,$6B,$D3,$4F,$F3,$0C,$60,$FF,$0A,$F4,$7F,$A1
	.dc.b	$5B,$3C,$75,$E9,$5F,$8D,$FA,$79,$3F,$89,$7F,$21,$E7,$FE,$47,$DA
	.dc.b	$93,$F8,$D7,$6F,$E0,$39,$3F,$D9,$AF,$E7,$3A,$3F,$75,$E8,$FE,$9C
	.dc.b	$FD,$47,$C6,$D9,$D5,$7F,$8E,$9B,$EF,$E5,$F9,$1F,$98,$FF,$57,$E1
	.dc.b	$FE,$05,$08,$C7,$07,$FA,$85,$B0,$EA,$7E,$4D,$FE,$48,$AF,$F5,$37
	.dc.b	$D1,$BD,$AF,$F4,$3B,$6F,$FA,$33,$5F,$F5,$93,$5F,$F5,$70,$7F,$8A
	.dc.b	$CF,$4F,$1F,$FA,$D4,$62,$7F,$D8,$0F,$3C,$BF,$80,$FC,$9F,$EE,$FE
	.dc.b	$46,$57,$ED,$E2,$BF,$9C,$F8,$9F,$7D,$7F,$1B,$6E,$3F,$BA,$EE,$61
	.dc.b	$FE,$0D,$9F,$75,$FF,$84,$5F,$98,$F3,$DF,$5F,$09,$B7,$87,$1C,$9F
	.dc.b	$E2,$15,$C3,$2F,$E6,$1E,$83,$F6,$27,$D0,$7F,$93,$DB,$7F,$ED,$79
	.dc.b	$D4,$E4,$FE,$D7,$7F,$84,$9F,$67,$37,$9C,$FF,$0F,$5F,$28,$F8,$1F
	.dc.b	$55,$EB,$23,$E4,$F9,$7F,$77,$C0,$F4,$24,$7F,$7A,$C3,$FC,$4E,$F8
	.dc.b	$EF,$FC,$20,$FD,$8C,$BD,$5E,$FF,$A4,$91,$C7,$E6,$7E,$51,$1F,$E6
	.dc.b	$B3,$C1,$B5,$FE,$03,$7F,$DE,$11,$F1,$7D,$CF,$88,$48,$F2,$7C,$DF
	.dc.b	$D5,$23,$CA,$3E,$0F,$B7,$23,$AB,$F6,$FD,$AC,$47,$E0,$FE,$13,$EF
	.dc.b	$18,$E2,$FA,$B9,$3D,$AF,$A8,$4A,$F8,$4F,$D2,$FA,$DF,$C3,$12,$3D
	.dc.b	$71,$F8,$9F,$A6,$91,$F8,$27,$F1,$9F,$59,$F8,$DF,$47,$ED,$92,$7E
	.dc.b	$83,$F7,$BD,$5F,$C9,$12,$3F,$17,$1F,$DD,$BE,$57,$FD,$0D,$9F,$F7
	.dc.b	$BF,$21,$18,$F0,$3F,$B4,$B6,$FF,$BF,$BF,$E8,$B8,$CF,$85,$FE,$06
	.dc.b	$8F,$D9,$4F,$ED,$0D,$C6,$4F,$C8,$FB,$BF,$A5,$F8,$39,$3F,$9E,$8A
	.dc.b	$FE,$BF,$C5,$FD,$51,$8C,$BC,$D8,$7F,$A1,$27,$FA,$EB,$63,$3F,$51
	.dc.b	$FB,$1F,$1F,$DD,$F5,$3E,$52,$5F,$97,$5F,$7C,$7F,$93,$BF,$3F,$FE
	.dc.b	$2F,$ED,$1F,$F7,$8D,$7F,$80,$5A,$FF,$5E,$76,$3E,$B1,$13,$82,$BF
	.dc.b	$E9,$19,$D3,$93,$FC,$02,$E1,$FE,$57,$5E,$1A,$E3,$5F,$9C,$FD,$51
	.dc.b	$F3,$18,$BE,$FD,$E9,$CC,$75,$11,$99,$E1,$18,$9D,$BF,$CE,$87,$DA
	.dc.b	$BC,$7F,$F4,$84,$61,$B1,$9D,$7F,$CF,$4D,$9F,$C0,$3F,$E9,$FF,$4B
	.dc.b	$3E,$57,$F9,$26,$72,$D7,$F9,$CC,$ED,$FE,$09,$FA,$67,$FA,$C3,$4F
	.dc.b	$56,$88,$C4,$FF,$50,$8C,$74,$FF,$94,$EB,$F8,$0D,$69,$FA,$3B,$FD
	.dc.b	$BE,$1D,$61,$F4,$1A,$FD,$56,$C6,$E7,$E5,$3F,$83,$E3,$FF,$30,$72
	.dc.b	$7F,$96,$F8,$FF,$A8,$57,$E1,$67,$2F,$DC,$E9,$B8,$FF,$99,$BC,$F7
	.dc.b	$C0,$FC,$F8,$C3,$78,$BF,$CE,$4D,$9D,$DF,$E5,$D7,$C1,$74,$AF,$F3
	.dc.b	$C3,$DB,$4F,$CC,$7C,$A6,$6B,$77,$F9,$9F,$99,$D6,$BF,$8A,$8C,$1F
	.dc.b	$9D,$E8,$CF,$CD,$EA,$AC,$EB,$5F,$F2,$5E,$0F,$87,$F3,$12,$7C,$CF
	.dc.b	$A7,$E5,$3C,$47,$8F,$E2,$6C,$FE,$AF,$E7,$FC,$67,$93,$D5,$B1,$FA
	.dc.b	$F7,$E3,$6F,$FB,$36,$ED,$FD,$82,$7F,$D8,$ED,$CF,$FC,$07,$5F,$B9
	.dc.b	$FA,$BF,$79,$F0,$3E,$AE,$DF,$EE,$FE,$AF,$E1,$9F,$5B,$B1,$FB,$1F
	.dc.b	$FA,$7E,$74,$7F,$50,$71,$8F,$F5,$96,$BF,$D4,$1A,$FE,$D1,$AF,$EC
	.dc.b	$96,$BF,$4B,$D5,$F0,$46,$2B,$F8,$74,$62,$64,$7F,$11,$FA,$CB,$7F
	.dc.b	$C4,$7D,$4F,$B6,$72,$F8,$FB,$5E,$77,$F1,$D9,$7B,$1B,$91,$A7,$CC
	.dc.b	$F5,$7F,$0A,$79,$3A,$AD,$FC,$17,$89,$1F,$F0,$75,$7E,$3B,$1F,$A3
	.dc.b	$C2,$F9,$2F,$AE,$4A,$FA,$EF,$DA,$F4,$BF,$26,$7E,$67,$DE,$FE,$49
	.dc.b	$E1,$7E,$B6,$31,$E4,$4B,$FA,$5F,$11,$72,$FF,$50,$6F,$F6,$23,$DA
	.dc.b	$49,$F2,$BF,$8A,$C5,$D5,$DF,$8F,$C0,$DA,$FF,$5F,$4F,$E7,$5B,$1E
	.dc.b	$5E,$BD,$C6,$B8,$BD,$1C,$3E,$96,$C7,$2F,$0B,$0D,$38,$1F,$EC,$ED
	.dc.b	$1F,$6A,$D3,$F8,$3C,$61,$B0,$F9,$0B,$18,$7F,$85,$4F,$26,$CE,$EB
	.dc.b	$FB,$94,$E5,$B0,$F9,$1E,$A7,$F1,$6F,$86,$7D,$D7,$D2,$AF,$EB,$35
	.dc.b	$B7,$FD,$15,$F6,$BF,$D1,$5B,$F0,$9F,$C4,$E3,$18,$FF,$B4,$14,$67
	.dc.b	$E1,$75,$E9,$5B,$E6,$FE,$07,$39,$1F,$4B,$4B,$8F,$FB,$33,$8F,$D8
	.dc.b	$D8,$FC,$46,$DF,$89,$7E,$4F,$D9,$BD,$8A,$63,$DA,$B8,$FD,$19,$47
	.dc.b	$CA,$6C,$7C,$86,$E1,$FE,$1B,$6B,$FE,$90,$5E,$D7,$80,$FC,$D7,$C7
	.dc.b	$5A,$FC,$5F,$CA,$7E,$74,$FC,$F7,$CA,$58,$FC,$7B,$87,$F2,$3F,$29
	.dc.b	$71,$F8,$1F,$29,$FA,$5F,$94,$FD,$33,$6B,$F2,$FF,$27,$86,$9F,$9A
	.dc.b	$8F,$E8,$DF,$6E,$C3,$F5,$DF,$29,$86,$D7,$F8,$63,$C5,$3F,$E1,$AE
	.dc.b	$C7,$A1,$F2,$E3,$FC,$79,$18,$E3,$FE,$B4,$D8,$CC,$EF,$5F,$F5,$AF
	.dc.b	$B0,$CD,$BF,$90,$FE,$01,$F2,$97,$0F,$D8,$FC,$97,$F0,$33,$F8,$27
	.dc.b	$C9,$58,$FD,$AB,$87,$D0,$7F,$8C,$32,$FF,$11,$39,$C9,$AF,$B1,$F5
	.dc.b	$FC,$1B,$AF,$FB,$42,$EE,$B1,$87,$3F,$07,$8C,$3B,$DF,$28,$C3,$DA
	.dc.b	$3F,$C8,$3B,$FB,$FA,$3E,$89,$FB,$CF,$B3,$7C,$4E,$2B,$F8,$14,$61
	.dc.b	$FD,$BB,$6F,$FC,$D4,$D7,$FD,$BE,$D7,$BF,$17,$F0,$99,$FF,$7F,$36
	.dc.b	$3C,$87,$C0,$F0,$7F,$07,$E0,$7C,$0F,$0B,$FB,$6E,$FF,$B6,$C3,$FC
	.dc.b	$F1,$C3,$FE,$4B,$9C,$9F,$D2,$BE,$41,$F2,$74,$95,$FE,$CF,$8C,$BC
	.dc.b	$AE,$E5,$FD,$D1,$AF,$F8,$D1,$AF,$D9,$6B,$F5,$BE,$4F,$CA,$49,$E8
	.dc.b	$F8,$FE,$42,$47,$85,$B3,$DE,$FD,$A4,$9F,$23,$FC,$A3,$F7,$D7,$85
	.dc.b	$49,$3F,$CA,$23,$2F,$F2,$AF,$C8,$FF,$7E,$DF,$13,$F9,$8C,$63,$0F
	.dc.b	$F5,$03,$5F,$77,$CA,$7C,$E2,$57,$C5,$7F,$9C,$F7,$B9,$A9,$27,$F9
	.dc.b	$CC,$65,$FE,$77,$F9,$1F,$F8,$E6,$FF,$D5,$CD,$7F,$D1,$E1,$C3,$FE
	.dc.b	$92,$6B,$FE,$84,$D1,$F2,$3F,$E2,$16,$D7,$E7,$B8,$FF,$D1,$9F,$2B
	.dc.b	$9A,$FF,$13,$7C,$DF,$E9,$EB,$5F,$38,$F8,$1C,$F7,$FF,$94,$E3,$13
	.dc.b	$FE,$F8,$FB,$57,$99,$FE,$14,$6C,$7F,$B1,$FD,$A4,$63,$6F,$5C,$62
	.dc.b	$BF,$C3,$7F,$82,$7F,$A0,$BC,$5F,$DE,$4A,$31,$87,$F0,$3F,$B6,$FE
	.dc.b	$B5,$87,$FA,$8A,$72,$FF,$1C,$A7,$93,$FE,$69,$5E,$4D,$8C,$3F,$E4
	.dc.b	$5F,$B6,$77,$7F,$8B,$1B,$97,$A5,$B8,$7A,$1A,$FF,$B0,$BD,$E7,$E0
	.dc.b	$0F,$98,$FF,$42,$C6,$1B,$F7,$4F,$2F,$35,$BC,$7F,$88,$DB,$EE,$3F
	.dc.b	$B7,$F0,$FF,$1D,$8C,$78,$FD,$27,$35,$FE,$D4,$9C,$BF,$C6,$B8,$1F
	.dc.b	$70,$F5,$3F,$C8,$ED,$9F,$A8,$DD,$BE,$9B,$5F,$67,$DC,$7F,$2B,$9F
	.dc.b	$2B,$63,$F8,$8E,$08,$FC,$44,$61,$D6,$7E,$9F,$A9,$78,$1F,$11,$B8
	.dc.b	$79,$FE,$5B,$EA,$BF,$0A,$33,$87,$F9,$0A,$72,$FE,$21,$AF,$FB,$CB
	.dc.b	$44,$62,$3F,$DD,$2D,$59,$3F,$03,$FA,$11,$FA,$C8,$C7,$1F,$FA,$E5
	.dc.b	$B8,$FE,$C2,$78,$D7,$07,$F0,$1F,$70,$F9,$6F,$BE,$DA,$FB,$6F,$EE
	.dc.b	$F5,$FB,$4E,$6B,$E4,$FD,$BF,$F3,$57,$E5,$FD,$C6,$6B,$F0,$13,$97
	.dc.b	$F0,$51,$9B,$FC,$04,$E5,$BC,$9F,$90,$8F,$E6,$B5,$CA,$3F,$CA,$36
	.dc.b	$1F,$45,$7E,$9B,$FD,$A3,$1F,$9B,$C3,$1F,$15,$F0,$6A,$CE,$45,$70
	.dc.b	$7A,$1D,$0C,$CF,$CD,$AE,$31,$AA,$AF,$FA,$C2,$B9,$3E,$58,$F3,$78
	.dc.b	$78,$DF,$FB,$F3,$A9,$EC,$E6,$F7,$87,$9E,$7C,$27,$9F,$0E,$18,$78
	.dc.b	$30,$FF,$04,$46,$27,$67,$EA,$EF,$CA,$FA,$E3,$B3,$A3,$CC,$7F,$9C
	.dc.b	$8F,$80,$7F,$9D,$2B,$0E,$0F,$8B,$F6,$EF,$9C,$7D,$C7,$F9,$EA,$30
	.dc.b	$D8,$CC,$FE,$46,$72,$DF,$9E,$FF,$10,$B7,$6E,$51,$88,$FE,$8F,$28
	.dc.b	$C4,$78,$9E,$3A,$FD,$99,$FC,$D4,$FC,$A3,$F0,$CE,$14,$6E,$78,$EB
	.dc.b	$CE,$F8,$52,$F8,$5F,$5E,$2F,$C2,$BC,$6B,$73,$E5,$6C,$EA,$FE,$3B
	.dc.b	$6F,$C7,$87,$81,$B5,$C1,$87,$9A,$BC,$8D,$36,$53,$CD,$5D,$5A,$79
	.dc.b	$2B,$B6,$CF,$B8,$7B,$FF,$EB,$08,$C3,$63,$35,$FC,$1E,$72,$DF,$B8
	.dc.b	$FF,$74,$37,$E5,$79,$46,$23,$F8,$93,$5F,$E9,$9F,$71,$FE,$1A,$7A
	.dc.b	$35,$FE,$30,$D7,$7E,$D7,$FB,$13,$B1,$F8,$0B,$E4,$B0,$DA,$FE,$E8
	.dc.b	$D7,$FB,$9B,$5F,$E5,$9F,$73,$FD,$A1,$FE,$F4,$D7,$FB,$CB,$5D,$F9
	.dc.b	$7F,$17,$3F,$EC,$51,$F0,$DA,$FF,$2F,$6B,$FD,$A5,$AF,$F5,$BF,$89
	.dc.b	$FC,$C3,$F2,$DA,$FF,$37,$6B,$FD,$C5,$AF,$F1,$DC,$7F,$25,$F0,$FF
	.dc.b	$0F,$97,$FD,$09,$3F,$E1,$36,$BF,$E1,$26,$BF,$DE,$FE,$E1,$F2,$DF
	.dc.b	$7D,$B1,$F7,$5A,$FF,$49,$6B,$FE,$03,$F0,$DF,$4B,$F8,$76,$C7,$E1
	.dc.b	$5A,$FF,$51,$6B,$FD,$6F,$7F,$C1,$3E,$B6,$AC,$BE,$E6,$9F,$7E,$73
	.dc.b	$0F,$C1,$9F,$4B,$F8,$36,$DF,$F8,$CD,$AF,$F8,$C9,$AF,$F6,$BF,$C4
	.dc.b	$9F,$35,$F8,$B6,$C7,$F8,$ED,$AF,$F8,$E9,$AF,$1D,$78,$11,$E0,$FE
	.dc.b	$57,$D9,$FC,$44,$7D,$76,$BE,$D6,$E3,$FE,$48,$E2,$FE,$5D,$E5,$FF
	.dc.b	$BD,$2B,$FB,$A0,$70,$FF,$77,$6B,$FE,$68,$6B,$FD,$5B,$ED,$DC,$9F
	.dc.b	$CE,$5B,$1D,$1A,$FF,$80,$5A,$FF,$9C,$BF,$43,$FC,$7B,$FD,$AD,$A9
	.dc.b	$A6,$29,$F5,$BE,$A2,$4F,$AA,$32,$F8,$7E,$47,$C8,$47,$07,$C0,$FE
	.dc.b	$00,$7C,$87,$F5,$16,$C6,$63,$07,$F0,$3F,$49,$F3,$34,$93,$FE,$BD
	.dc.b	$7E,$14,$C8,$F8,$4D,$7F,$CE,$F3,$F0,$9F,$D7,$35,$FF,$48,$69,$DC
	.dc.b	$F0,$73,$C8,$FF,$65,$35,$F9,$9E,$1F,$D1,$23,$F4,$5F,$E3,$5F,$B3
	.dc.b	$7F,$99,$E3,$FD,$14,$D7,$FD,$12,$D7,$FC,$84,$D7,$FD,$EF,$F4,$6F
	.dc.b	$AA,$FC,$3F,$B8,$8C,$4F,$F8,$F6,$31,$3A,$9C,$CF,$F9,$07,$ED,$DF
	.dc.b	$33,$FD,$FD,$B1,$F9,$26,$BB,$8F,$F5,$EF,$E1,$6F,$9C,$A4,$9F,$F2
	.dc.b	$67,$C7,$A7,$51,$FE,$ED,$F9,$5F,$CE,$FC,$7B,$E5,$91,$86,$CE,$CF
	.dc.b	$F7,$A7,$DA,$E4,$8E,$0F,$F7,$C7,$DA,$F1,$23,$83,$FD,$F9,$D1,$D4
	.dc.b	$8E,$0F,$8D,$B7,$F0,$67,$E6,$F8,$9F,$1C,$8F,$ED,$3E,$EF,$FA,$53
	.dc.b	$FE,$44,$6B,$FE,$51,$6A,$73,$C5,$F5,$7D,$8F,$A4,$49,$F0,$BF,$07
	.dc.b	$B6,$47,$F9,$D7,$6F,$8D,$18,$8F,$F7,$3C,$7C,$9F,$92,$F9,$04,$8E
	.dc.b	$1B,$7F,$B9,$9B,$F3,$1F,$34,$8E,$0D,$F6,$FD,$12,$38,$3F,$04,$7E
	.dc.b	$14,$86,$0F,$E2,$FF,$72,$FA,$AF,$F2,$43,$63,$31,$83,$3E,$64,$8E
	.dc.b	$2D,$7D,$B8,$7E,$C7,$2F,$C3,$77,$7E,$14,$93,$EE,$EE,$FC,$41,$1C
	.dc.b	$1F,$A1,$F7,$FF,$B8,$BF,$ED,$56,$F0,$7F,$9F,$A7,$F5,$31,$83,$F8
	.dc.b	$5C,$F1,$D3,$24,$7F,$9A,$F9,$DF,$C7,$1F,$F5,$53,$5F,$F4,$39,$FD
	.dc.b	$9B,$83,$FC,$D5,$19,$8F,$F4,$5A,$E5,$C9,$23,$AB,$C3,$9A,$47,$81
	.dc.b	$AF,$6C,$9F,$83,$BE,$67,$F8,$1B,$67,$F8,$13,$5E,$0D,$7F,$D3,$B3
	.dc.b	$CB,$E9,$78,$92,$38,$BF,$B8,$FD,$67,$F8,$94,$38,$11,$86,$D6,$CE
	.dc.b	$1F,$03,$2B,$F3,$95,$FE,$70,$6C,$7F,$24,$FB,$29,$1F,$C9,$1A,$FE
	.dc.b	$82,$5F,$73,$E3,$24,$74,$7C,$DF,$6F,$23,$BB,$5F,$E5,$BF,$37,$FC
	.dc.b	$C8,$8F,$F3,$27,$F9,$AF,$D9,$48,$FE,$6A,$D7,$C5,$F6,$FF,$0C,$8F
	.dc.b	$B1,$FE,$7B,$D3,$23,$F9,$EB,$5F,$57,$1F,$CB,$23,$F1,$1F,$9D,$F2
	.dc.b	$12,$3E,$3B,$5F,$C6,$FA,$8F,$D0,$3F,$D6,$CD,$8F,$F3,$34,$60,$CF
	.dc.b	$D9,$C8,$FF,$35,$35,$FA,$7F,$11,$F7,$48,$E0,$FA,$FF,$11,$F8,$12
	.dc.b	$38,$3E,$DF,$89,$F8,$52,$38,$3E,$FF,$8D,$F8,$92,$38,$3F,$AB,$7D
	.dc.b	$C7,$F9,$FD,$FF,$7D,$73,$23,$10,$8C,$7C,$2F,$F6,$09,$CC,$6F,$AF
	.dc.b	$04,$AE,$AF,$E6,$7E,$E2,$47,$E4,$9A,$FE,$37,$D2,$FC,$F1,$1F,$CB
	.dc.b	$BD,$FE,$A6,$47,$A1,$AF,$EC,$F0,$76,$F9,$49,$1E,$E7,$F5,$5D,$72
	.dc.b	$3F,$46,$D7,$EE,$F8,$DF,$DD,$BE,$41,$1F,$E8,$18,$C3,$71,$CA,$9B
	.dc.b	$C9,$FD,$B6,$73,$18,$AF,$F7,$03,$78,$B3,$7D,$5A,$FF,$9C,$E7,$6F
	.dc.b	$6F,$BA,$FF,$1A,$EF,$FE,$56,$6C,$7F,$20,$0E,$1F,$E4,$6D,$79,$7A
	.dc.b	$EF,$8C,$E4,$D8,$E5,$3F,$ED,$49,$C9,$D7,$37,$DD,$3C,$EF,$81,$70
	.dc.b	$FF,$45,$46,$74,$F9,$27,$C4,$71,$1B,$AE,$75,$CA,$B5,$E9,$DA,$FF
	.dc.b	$58,$5F,$F8,$5A,$79,$C6,$27,$4F,$81,$BE,$2F,$F2,$4F,$BA,$F2,$4F
	.dc.b	$DA,$8C,$57,$F2,$29,$CC,$FF,$2B,$F1,$51,$8D,$9F,$AD,$7F,$A6,$2F
	.dc.b	$FC,$78,$DB,$FE,$84,$D7,$F9,$DB,$5F,$D5,$BF,$EC,$08,$E3,$F0,$9D
	.dc.b	$63,$13,$CA,$3C,$78,$FF,$3C,$F1,$6E,$46,$93,$0F,$E7,$7F,$77,$A2
	.dc.b	$38,$3C,$FF,$73,$DC,$8E,$0F,$5E,$3E,$32,$38,$3F,$A3,$F8,$FE,$22
	.dc.b	$38,$3F,$A5,$FD,$DF,$98,$8E,$0F,$E9,$FE,$3F,$A8,$8E,$0F,$EA,$7E
	.dc.b	$A7,$B8,$8E,$0F,$C8,$78,$AC,$7E,$BF,$2A,$31,$3A,$9C,$C6,$32,$FE
	.dc.b	$6F,$E3,$3E,$EF,$8B,$6F,$F3,$8D,$9D,$CF,$D2,$E2,$FC,$7C,$7E,$FB
	.dc.b	$87,$F0,$6E,$49,$7F,$E4,$79,$FD,$63,$7A,$39,$11,$EF,$F1,$5E,$22
	.dc.b	$38,$39,$67,$E1,$23,$83,$CF,$F2,$BE,$A2,$38,$3D,$7F,$71,$EE,$23
	.dc.b	$83,$CD,$9F,$B4,$8E,$0F,$57,$37,$C1,$23,$83,$E3,$F8,$9F,$28,$8E
	.dc.b	$0F,$85,$8F,$C9,$23,$83,$E3,$61,$F3,$48,$E0,$FF,$54,$79,$CF,$33
	.dc.b	$F6,$4D,$8C,$C6,$0C,$F9,$B2,$38,$B7,$7F,$E4,$9F,$74,$FA,$BF,$DF
	.dc.b	$C2,$31,$5F,$F0,$01,$B1,$C2,$77,$1F,$EA,$F9,$D7,$C1,$60,$8F,$1F
	.dc.b	$B7,$D9,$1C,$1F,$EC,$3F,$1B,$FC,$A0,$DF,$5F,$F8,$74,$63,$1F,$F2
	.dc.b	$63,$5F,$F2,$2A,$F5,$F6,$4A,$E8,$FE,$5F,$E4,$64,$7E,$4D,$B5,$FE
	.dc.b	$98,$E7,$72,$FE,$74,$95,$F9,$97,$E2,$AF,$4A,$91,$FE,$9D,$5F,$D7
	.dc.b	$E3,$7E,$9C,$9C,$87,$13,$FD,$56,$1E,$A7,$62,$3C,$1F,$D8,$7A,$12
	.dc.b	$3F,$58,$D9,$F5,$73,$FE,$D4,$93,$FB,$17,$F7,$1E,$9C,$8F,$DA,$36
	.dc.b	$7E,$17,$A9,$FB,$D2,$4F,$EE,$5F,$A7,$E7,$C8,$FB,$0D,$9F,$ED,$DF
	.dc.b	$A7,$23,$F1,$6D,$9F,$8A,$D9,$DC,$FD,$EF,$AE,$95,$F8,$36,$CF,$D5
	.dc.b	$F8,$7F,$E2,$04,$9F,$E1,$8F,$F1,$3E,$59,$1F,$C3,$DB,$3F,$77,$9B
	.dc.b	$F8,$C1,$27,$F8,$A3,$FC,$6F,$A2,$47,$F1,$76,$CF,$FA,$A3,$E1,$FF
	.dc.b	$BE,$7A,$DF,$D9,$1B,$5F,$E0,$D6,$CE,$FD,$35,$7E,$B3,$EE,$9F,$09
	.dc.b	$49,$87,$E2,$23,$2F,$E8,$CF,$D0,$90,$C1,$FE,$0F,$FC,$C9,$B7,$F4
	.dc.b	$06,$C6,$63,$07,$EF,$3D,$37,$C6,$D2,$4F,$E6,$23,$2F,$5E,$1E,$84
	.dc.b	$70,$78,$67,$C0,$47,$07,$38,$F0,$11,$C1,$E9,$FA,$1F,$41,$1C,$1F
	.dc.b	$AA,$EE,$7A,$7F,$C5,$1B,$19,$8C,$1F,$C6,$38,$3F,$89,$70,$7D,$3E
	.dc.b	$2C,$12,$BF,$6B,$F6,$BF,$C2,$67,$FC,$AA,$D7,$FC,$B6,$D4,$DF,$CA
	.dc.b	$BF,$DB,$53,$FB,$E6,$CF,$F9,$35,$AF,$FB,$B9,$AF,$F9,$8A,$DD,$5E
	.dc.b	$72,$57,$F0,$88,$CB,$FC,$2B,$ED,$7F,$81,$B7,$B7,$F9,$AC,$62,$FF
	.dc.b	$CB,$8D,$7F,$CD,$BF,$68,$F9,$DF,$F0,$1B,$6F,$FC,$AC,$D9,$FF,$5B
	.dc.b	$B5,$FF,$3A,$FB,$EF,$98,$FF,$05,$B6,$BF,$CB,$CD,$9F,$F5,$FB,$5E
	.dc.b	$3F,$77,$C0,$93,$C1,$FF,$0A,$FD,$93,$EE,$69,$27,$FC,$D5,$19,$7F
	.dc.b	$CD,$9F,$67,$F9,$4B,$FF,$66,$B5,$FF,$45,$B5,$FF,$49,$FB,$48,$C4
	.dc.b	$7F,$9E,$FE,$C9,$F2,$3F,$E5,$CF,$65,$18,$C3,$FD,$0D,$18,$9D,$4E
	.dc.b	$6B,$FD,$36,$7F,$A1,$7E,$CB,$F8,$D3,$FE,$9E,$A4,$62,$3F,$D4,$B1
	.dc.b	$8E,$BF,$04,$E6,$59,$7F,$7B,$F1,$3F,$C7,$3E,$A7,$F6,$E9,$FF,$53
	.dc.b	$B5,$F6,$B5,$FF,$4B,$7D,$87,$FA,$B7,$4F,$4B,$63,$CF,$92,$31,$1E
	.dc.b	$76,$BE,$6F,$9D,$6F,$F1,$5B,$7F,$11,$B3,$D1,$AF,$9B,$ED,$3D,$24
	.dc.b	$9F,$F5,$26,$28,$FF,$2A,$FE,$0E,$F9,$AA,$4A,$FF,$55,$46,$5F,$A7
	.dc.b	$F6,$3F,$E1,$DC,$3D,$6D,$7E,$83,$5F,$9E,$D7,$FD,$75,$F6,$37,$53
	.dc.b	$49,$97,$E0,$FE,$C7,$F0,$F2,$69,$3C,$6F,$F1,$2D,$FF,$B8,$3E,$C7
	.dc.b	$FD,$C3,$F2,$CF,$80,$F9,$FA,$61,$FF,$68,$4D,$AF,$C3,$4D,$7F,$DC
	.dc.b	$B1,$5F,$F3,$28,$AE,$8F,$E4,$23,$58,$FF,$B3,$27,$33,$C8,$E1,$7E
	.dc.b	$26,$9C,$4A,$7E,$52,$FC,$0A,$FA,$C6,$1B,$1C,$0F,$80,$5B,$F1,$57
	.dc.b	$BA,$31,$5C,$5F,$F1,$19,$D4,$B5,$87,$07,$EB,$3E,$AB,$FC,$49,$F6
	.dc.b	$3F,$E2,$A6,$C7,$FB,$E5,$AE,$FD,$94,$EF,$C4,$7D,$8D,$25,$71,$8C
	.dc.b	$B6,$0B,$E7,$7E,$5E,$0F,$0B,$33,$9D,$2C,$E6,$38,$D7,$0C,$3A,$B7
	.dc.b	$0C,$DF,$0C,$39,$CF,$66,$C7,$4A,$E2,$DA,$FC,$36,$1D,$F0,$FB,$2D
	.dc.b	$DB,$B4,$62,$F8,$BE,$C9,$E1,$8F,$78,$F1,$99,$9D,$A9,$FF,$28,$DE
	.dc.b	$A3,$18,$F9,$9E,$8D,$78,$57,$83,$67,$D8,$FF,$83,$A7,$E6,$DB,$7F
	.dc.b	$E4,$73,$99,$FA,$F1,$8E,$0F,$9F,$5B,$95,$F4,$F3,$F3,$7F,$CB,$3E
	.dc.b	$C5,$F1,$D8,$BD,$C6,$1F,$0C,$EE,$75,$7E,$6C,$6F,$D8,$BD,$BB,$ED
	.dc.b	$CC,$78,$9E,$29,$D3,$8B,$42,$7F,$D0,$31,$86,$C6,$67,$FA,$6B,$5F
	.dc.b	$E9,$7E,$13,$EC,$1E,$57,$FA,$2A,$30,$D8,$CC,$FF,$52,$3F,$A4,$B8
	.dc.b	$7F,$9F,$FE,$60,$F5,$FD,$ED,$8F,$43,$5F,$A2,$72,$AF,$8B,$E5,$7F
	.dc.b	$A5,$63,$DC,$FF,$A7,$DB,$F0,$5A,$EC,$FA,$19,$3F,$08,$7C,$37,$A5
	.dc.b	$B7,$E7,$9D,$4E,$4F,$8F,$F6,$4F,$8D,$FA,$AD,$AF,$9C,$D7,$CF,$C2
	.dc.b	$8F,$E4,$3D,$1F,$87,$C1,$F6,$3F,$DD,$A3,$F1,$54,$FA,$6F,$F7,$46
	.dc.b	$1F,$65,$AF,$C2,$BE,$2F,$B2,$FF,$22,$FA,$5A,$FD,$86,$BF,$C1,$A3
	.dc.b	$C3,$5F,$67,$EC,$9F,$45,$F1,$9B,$5F,$8C,$A4,$62,$3E,$9F,$CA,$FF
	.dc.b	$14,$F8,$7F,$9B,$E8,$7F,$B7,$67,$F2,$72,$FC,$23,$F5,$5A,$FE,$9E
	.dc.b	$31,$D9,$C3,$E0,$FE,$AF,$C6,$FF,$4B,$C3,$E9,$35,$FC,$8B,$5F,$53
	.dc.b	$5F,$F1,$5F,$D9,$FE,$37,$0F,$9C,$DF,$07,$3D,$5F,$B1,$CB,$F6,$FE
	.dc.b	$97,$EC,$F0,$7D,$9B,$EA,$BF,$A1,$B6,$3F,$DC,$ED,$7F,$38,$D7,$FC
	.dc.b	$5F,$F6,$7F,$EE,$47,$FD,$DA,$D7,$F4,$4D,$77,$E4,$FE,$CB,$F0,$7F
	.dc.b	$F3,$7C,$7F,$8F,$1A,$FF,$89,$1A,$FF,$8E,$5A,$FF,$40,$FA,$E7,$C4
	.dc.b	$FF,$5A,$6C,$7F,$BF,$9A,$FE,$CD,$AF,$F9,$1B,$F0,$77,$C9,$7F,$62
	.dc.b	$6C,$7E,$E4,$38,$7F,$78,$D7,$FA,$8E,$AF,$95,$7E,$05,$C6,$30,$D8
	.dc.b	$CC,$FF,$05,$9C,$B6,$1F,$75,$B2,$46,$5F,$E2,$1F,$22,$FB,$EA,$49
	.dc.b	$FF,$68,$3D,$FE,$02,$47,$56,$BF,$E6,$6C,$3D,$1F,$13,$E5,$24,$78
	.dc.b	$DE,$FF,$63,$23,$FD,$B8,$D7,$96,$5F,$C7,$5F,$86,$D7,$FB,$CD,$7C
	.dc.b	$5E,$6F,$84,$7C,$5F,$A1,$FE,$95,$3F,$C7,$BC,$F7,$EB,$6F,$A8,$FE
	.dc.b	$91,$18,$7F,$35,$5C,$3B,$78,$FE,$11,$FC,$DF,$0F,$F4,$0B,$5F,$F0
	.dc.b	$2B,$5F,$F0,$5B,$6F,$BF,$D5,$F6,$C4,$D3,$F4,$D3,$FD,$0E,$FF,$8D
	.dc.b	$70,$FE,$8D,$BD,$9F,$CB,$A3,$1D,$DF,$03,$3F,$F5,$F6,$7F,$C1,$17
	.dc.b	$CC,$7A,$89,$1F,$C3,$EB,$8A,$DD,$4F,$8D,$FE,$77,$F3,$DF,$D1,$73
	.dc.b	$FE,$58,$DC,$BC,$13,$AD,$FE,$3F,$8B,$32,$FE,$39,$7B,$7F,$A9,$70
	.dc.b	$3E,$BF,$FB,$B7,$CB,$7E,$6F,$E7,$6F,$31,$F8,$D3,$87,$FD,$C9,$19
	.dc.b	$BF,$EC,$15,$C5,$B1,$97,$FD,$6B,$97,$DE,$F3,$BE,$81,$23,$FB,$DD
	.dc.b	$71,$7F,$A0,$64,$FA,$4F,$82,$48,$F9,$9F,$03,$F8,$4F,$95,$91,$F3
	.dc.b	$DD,$3F,$CE,$38,$BF,$D5,$2F,$C2,$F8,$CF,$96,$7E,$33,$CF,$B1,$F5
	.dc.b	$FE,$E7,$F1,$E6,$1E,$5F,$80,$FF,$29,$FB,$6F,$EE,$51,$FD,$D8,$38
	.dc.b	$E7,$FE,$81,$8A,$B8,$79,$7F,$E7,$0A,$FE,$49,$7F,$E5,$A6,$BF,$E6
	.dc.b	$06,$BD,$3E,$23,$C0,$47,$FC,$42,$78,$27,$F4,$BE,$2F,$E1,$BD,$E9
	.dc.b	$3F,$91,$8A,$FC,$3F,$B0,$FD,$59,$8C,$FF,$C7,$9C,$7F,$C6,$FE,$D7
	.dc.b	$F0,$24,$AF,$F1,$7F,$37,$ED,$A3,$F5,$1F,$1B,$27,$3D,$7F,$D6,$70
	.dc.b	$E7,$D2,$BF,$C6,$FE,$11,$FB,$9E,$FF,$F3,$CF,$D5,$7C,$5C,$7D,$6D
	.dc.b	$7F,$C4,$3A,$2B,$84,$7F,$A1,$25,$CB,$FE,$9A,$6C,$7F,$39,$6B,$FE
	.dc.b	$33,$6B,$FE,$88,$0F,$E1,$1F,$B4,$DC,$BF,$8A,$B7,$0F,$3D,$7E,$CB
	.dc.b	$BA,$5F,$EB,$1F,$E2,$BE,$AF,$F1,$6F,$62,$4F,$F0,$17,$6F,$E7,$3C
	.dc.b	$9F,$DD,$91,$FE,$0D,$F5,$3F,$C7,$BC,$C9,$1F,$C2,$5D,$3F,$A4,$FB
	.dc.b	$0F,$F2,$97,$84,$69,$23,$FA,$77,$6B,$E9,$FF,$94,$92,$3F,$81,$CF
	.dc.b	$F2,$CF,$77,$F9,$6F,$99,$27,$F8,$C4,$57,$FA,$45,$FF,$80,$5F,$E3
	.dc.b	$D3,$FC,$EC,$FF,$32,$7E,$85,$FE,$30,$8F,$F5,$CF,$C0,$23,$10,$8C
	.dc.b	$7C,$0F,$03,$FA,$27,$C4,$FF,$3E,$24,$FF,$8C,$EB,$FC,$7B,$CD,$FE
	.dc.b	$4B,$5F,$5D,$E0,$24,$FF,$96,$E3,$2F,$C2,$F8,$5F,$E9,$24,$7A,$4F
	.dc.b	$F8,$97,$B7,$F9,$27,$A1,$FD,$4B,$C0,$C1,$9F,$03,$FC,$D7,$8F,$F9
	.dc.b	$7F,$2F,$17,$C3,$E0,$DE,$5F,$E4,$9F,$AF,$F7,$EE,$69,$37,$FF,$03
	.dc.b	$69,$FA,$5A,$FF,$24,$B5,$FF,$27,$B7,$0F,$F3,$55,$7C,$2E,$DF,$ED
	.dc.b	$B3,$FC,$96,$B8,$FD,$BF,$AC,$F6,$79,$FE,$DF,$03,$B9,$27,$FC,$25
	.dc.b	$5F,$A2,$6F,$9B,$F0,$C8,$F9,$27,$FB,$9F,$8A,$FA,$DB,$F5,$12,$BF
	.dc.b	$82,$CE,$5B,$DB,$FE,$2B,$E4,$FF,$13,$36,$4D,$3E,$1F,$99,$F4,$C9
	.dc.b	$5F,$13,$CC,$FA,$DE,$64,$8F,$8E,$FD,$9F,$8E,$FB,$5E,$9F,$DB,$F2
	.dc.b	$25,$7C,$C8,$F4,$FC,$6F,$84,$91,$FE,$B7,$D3,$FC,$75,$1F,$53,$D5
	.dc.b	$FC,$21,$27,$EA,$FB,$7F,$87,$F2,$51,$39,$FF,$C8,$7C,$BF,$E1,$98
	.dc.b	$C1,$F5,$A7,$6F,$F8,$83,$BF,$FC,$38,$E8,$FD,$27,$D7,$7F,$A8,$2F
	.dc.b	$FD,$30,$D8,$FD,$5B,$5D,$F3,$F4,$9F,$E3,$9F,$6C,$F6,$7F,$D9,$55
	.dc.b	$E0,$8C,$6D,$FE,$5D,$6C,$7F,$B5,$DA,$F2,$F6,$BF,$6F,$1F,$E1,$1E
	.dc.b	$FE,$11,$83,$F7,$DE,$03,$ED,$BF,$D1,$59,$7A,$E3,$1E,$42,$31,$3A
	.dc.b	$C7,$FC,$83,$5A,$9F,$85,$B7,$F9,$0B,$EB,$B8,$57,$C7,$E1,$46,$2B
	.dc.b	$E7,$B6,$7F,$46,$D7,$FC,$55,$2F,$B5,$E7,$E2,$7F,$AC,$E3,$17,$FB
	.dc.b	$0F,$37,$2F,$EB,$2A,$DF,$FB,$DF,$F4,$A7,$D4,$7F,$96,$23,$F0,$3C
	.dc.b	$49,$CD,$FE,$2D,$B3,$FE,$20,$6B,$F8,$6F,$B6,$EB,$AF,$F9,$83,$E7
	.dc.b	$2E,$89,$B1,$AE,$0F,$C5,$ED,$FE,$CB,$6E,$5F,$E6,$E5,$C7,$BC,$B8
	.dc.b	$7A,$12,$BF,$DA,$B5,$FC,$A1,$B1,$EE,$D7,$D5,$F3,$9E,$DB,$F8,$0D
	.dc.b	$8C,$D7,$B6,$31,$B7,$F0,$38,$FF,$02,$BF,$EA,$C6,$F8,$FD,$63,$17
	.dc.b	$FE,$EA,$C5,$F6,$BE,$D3,$DD,$E6,$DC,$3C,$EF,$8F,$87,$F1,$CD,$A3
	.dc.b	$4C,$7D,$17,$C7,$F0,$7E,$43,$6F,$E0,$FC,$AF,$CB,$3E,$EF,$4F,$F3
	.dc.b	$5E,$0C,$4F,$F3,$97,$E1,$46,$63,$FC,$11,$A7,$F5,$9F,$3F,$FC,$EB
	.dc.b	$18,$9D,$4F,$F8,$35,$FE,$C1,$7D,$BD,$03,$F6,$FC,$BC,$36,$FE,$AC
	.dc.b	$7D,$FF,$37,$18,$7C,$7E,$2F,$C1,$1C,$E2,$D7,$F6,$57,$CF,$81,$A4
	.dc.b	$8F,$F6,$86,$FF,$4B,$9B,$F1,$5F,$1F,$E5,$C3,$F1,$CD,$AF,$ED,$51
	.dc.b	$D0,$8F,$2E,$5F,$A4,$47,$07,$6F,$9E,$FC,$C4,$7E,$65,$B1,$F9,$2F
	.dc.b	$92,$FC,$E1,$F1,$4A,$FF,$7F,$7D,$64,$7F,$BE,$3F,$81,$ED,$FD,$21
	.dc.b	$F3,$7E,$42,$CF,$F8,$1D,$C3,$FA,$06,$ED,$F9,$1F,$90,$FD,$61,$F5
	.dc.b	$7E,$3E,$D7,$F8,$3D,$C3,$FA,$8F,$90,$B5,$FE,$7F,$FA,$4F,$E3,$93
	.dc.b	$FE,$72,$6C,$7F,$A0,$FE,$D5,$F2,$5F,$C2,$A1,$3A,$E1,$FF,$70,$D6
	.dc.b	$AF,$F1,$B4,$8C,$3F,$90,$6B,$FE,$59,$6B,$FD,$4B,$E1,$DF,$84,$FE
	.dc.b	$52,$3C,$59,$FF,$1A,$9C,$F5,$F8,$ED,$3F,$C0,$7F,$53,$7C,$0D,$38
	.dc.b	$79,$0F,$87,$C6,$9F,$EB,$68,$C3,$63,$97,$9B,$C1,$FD,$6F,$EC,$EF
	.dc.b	$B3,$FE,$35,$96,$1B,$7F,$C5,$38,$B0,$D3,$FB,$6F,$0F,$F4,$2F,$05
	.dc.b	$87,$39,$FF,$50,$ED,$FC,$81,$BE,$57,$B7,$8B,$DC,$D8,$F9,$B3,$FD
	.dc.b	$BE,$7F,$B9,$B5,$FE,$F6,$DF,$1D,$C3,$F5,$EB,$FB,$77,$97,$FC,$E2
	.dc.b	$31,$7F,$E0,$18,$F7,$FA,$59,$DA,$DF,$A8,$FF,$03,$E1,$D2,$39,$7C
	.dc.b	$AE,$B0,$6D,$FD,$82,$BC,$33,$98,$C6,$4F,$4B,$C4,$7C,$EF,$92,$B9
	.dc.b	$77,$C7,$FC,$DD,$BA,$75,$39,$CB,$DB,$A6,$2F,$C3,$3F,$D7,$31,$FF
	.dc.b	$6C,$3F,$EC,$06,$C7,$FA,$D3,$97,$E9,$3C,$07,$9F,$F2,$3E,$4B,$F1
	.dc.b	$3B,$FF,$4B,$EA,$5F,$0B,$FF,$11,$C6,$27,$53,$99,$F1,$7C,$6F,$2C
	.dc.b	$BC,$4D,$8F,$07,$2F,$E1,$3E,$47,$EA,$7C,$A7,$E4,$1B,$7E,$FD,$5F
	.dc.b	$29,$98,$C3,$70,$F5,$70,$7E,$03,$87,$EF,$1F,$9A,$F9,$0B,$7F,$37
	.dc.b	$3F,$C9,$C6,$63,$B6,$97,$0F,$89,$C3,$CE,$7E,$56,$1F,$C0,$FF,$0A
	.dc.b	$7C,$87,$EC,$63,$0F,$F7,$1B,$4E,$AB,$F8,$2B,$63,$FA,$9B,$5F,$F2
	.dc.b	$6E,$7F,$46,$7E,$C6,$FF,$97,$E3,$FD,$F3,$F9,$9E,$0F,$ED,$9C,$06
	.dc.b	$92,$BF,$AE,$72,$FF,$8D,$F1,$FF,$2C,$F0,$7E,$CB,$BF,$FB,$2C,$E6
	.dc.b	$7F,$D7,$ED,$59,$FE,$AB,$E5,$3B,$93,$0F,$ED,$71,$96,$E5,$FC,$32
	.dc.b	$BF,$DE,$D3,$FB,$2F,$03,$CE,$4A,$FE,$E1,$FA,$9B,$E5,$7F,$BB,$B6
	.dc.b	$33,$18,$3D,$FC,$B2,$3B,$B7,$0E,$F8,$BE,$53,$E2,$92,$BC,$AF,$D5
	.dc.b	$A7,$C2,$D2,$4F,$F3,$A8,$CB,$FC,$F3,$F0,$A7,$B3,$F7,$5B,$19,$8C
	.dc.b	$1F,$53,$9A,$47,$C4,$6E,$5F,$E1,$2F,$B8,$FF,$18,$37,$D3,$FC,$44
	.dc.b	$62,$3F,$D1,$E1,$C3,$FB,$08,$FC,$77,$7E,$23,$FA,$54,$60,$FF,$10
	.dc.b	$FC,$67,$F5,$2F,$0F,$F9,$93,$71,$FE,$37,$F2,$4F,$89,$FF,$15,$CE
	.dc.b	$70,$FF,$7F,$7D,$03,$D5,$FF,$15,$B7,$0F,$F1,$43,$67,$FC,$50,$D7
	.dc.b	$F1,$FF,$84,$B8,$7F,$CC,$FE,$93,$CB,$FF,$1C,$CE,$70,$FE,$39,$39
	.dc.b	$7F,$6D,$E1,$BC,$1F,$F1,$FB,$63,$F9,$17,$80,$F5,$33,$5F,$E4,$4F
	.dc.b	$B7,$FE,$5D,$5F,$E3,$D6,$E9,$FE,$3D,$6B,$FA,$46,$BF,$EC,$4B,$43
	.dc.b	$CD,$FF,$45,$F9,$EF,$B7,$FF,$51,$72,$FF,$5D,$AD,$63,$FC,$E6,$B5
	.dc.b	$E2,$F0,$FB,$7F,$E5,$F7,$FE,$A4,$94,$62,$BF,$DA,$71,$8C,$78,$79
	.dc.b	$DF,$E5,$EF,$C2,$3E,$66,$BE,$56,$AD,$51,$86,$F3,$7F,$A0,$F0,$FF
	.dc.b	$03,$FA,$9E,$A3,$EF,$38,$CB,$D1,$18,$FB,$7F,$E5,$F1,$EB,$6E,$3F
	.dc.b	$CE,$3E,$3B,$E8,$4E,$7D,$1F,$F3,$9C,$62,$7F,$CD,$4D,$7F,$88,$B5
	.dc.b	$F9,$CD,$7F,$D4,$3F,$6C,$F9,$8F,$43,$6F,$FC,$F3,$92,$31,$5F,$A6
	.dc.b	$8C,$57,$0F,$9A,$FA,$F5,$AE,$2F,$A0,$A3,$33,$F7,$3E,$DB,$58,$FD
	.dc.b	$26,$E3,$FE,$B4,$FB,$6F,$BA,$FF,$67,$C9,$3A,$53,$AC,$BE,$F3,$6F
	.dc.b	$EE,$BF,$92,$8C,$F1,$FF,$93,$70,$1F,$E4,$E1,$F6,$BF,$C4,$1C,$9F
	.dc.b	$E2,$73,$3E,$47,$FA,$73,$B7,$8C,$EE,$7F,$17,$8E,$2B,$51,$C0,$E7
	.dc.b	$97,$FB,$DA,$B8,$47,$8A,$F8,$3A,$BA,$F8,$A3,$CF,$E2,$70,$BF,$F7
	.dc.b	$FB,$CA,$7D,$4E,$6B,$78,$78,$B0,$C5,$78,$CE,$85,$3D,$CF,$3C,$3C
	.dc.b	$31,$89,$E4,$7E,$06,$B3,$1F,$24,$FF,$39,$CF,$F9,$D1,$F6,$15,$FF
	.dc.b	$3B,$FD,$AB,$E6,$DF,$6D,$86,$D7,$FB,$22,$75,$39,$38,$F8,$DF,$E4
	.dc.b	$09,$D3,$5F,$E7,$32,$8C,$47,$DA,$8C,$6D,$C8,$F9,$AB,$E7,$7E,$FE
	.dc.b	$9F,$6E,$31,$5A,$BF,$04,$E5,$BA,$75,$BE,$2D,$AE,$18,$45,$AF,$5A
	.dc.b	$FC,$6B,$63,$EB,$9F,$94,$C7,$CE,$F0,$BF,$56,$1E,$08,$D5,$F8,$1A
	.dc.b	$77,$C7,$CD,$8F,$BA,$75,$2C,$3D,$AF,$E0,$DE,$F5,$FA,$1D,$9F,$6D
	.dc.b	$FE,$42,$AF,$E5,$CD,$7F,$68,$DC,$3F,$97,$3F,$E6,$98,$DF,$3F,$EA
	.dc.b	$A7,$F9,$97,$C2,$7F,$6E,$6C,$E7,$87,$F9,$9D,$F3,$7F,$9A,$FD,$AF
	.dc.b	$F6,$39,$FE,$7C,$D7,$F8,$23,$6B,$F9,$F2,$F8,$8E,$92,$C9,$F0,$EF
	.dc.b	$90,$FF,$64,$C6,$1B,$19,$9F,$E4,$2D,$7F,$C0,$BE,$07,$C9,$CB,$FD
	.dc.b	$A2,$D8,$FF,$04,$9F,$D1,$77,$FF,$2A,$46,$63,$F9,$47,$DA,$7E,$09
	.dc.b	$FE,$B0,$D7,$F8,$F3,$74,$FF,$5D,$52,$FD,$79,$7F,$B4,$3E,$7A,$FC
	.dc.b	$49,$A6,$3E,$7F,$E0,$E6,$FB,$47,$CA,$7A,$1B,$AF,$A2,$75,$39,$3E
	.dc.b	$2F,$DA,$3E,$5F,$C2,$FC,$3D,$BF,$B6,$37,$5F,$ED,$6E,$F0,$FE,$E9
	.dc.b	$F4,$DF,$8E,$AF,$EE,$4D,$AF,$EE,$2D,$9F,$AC,$FE,$E5,$FF,$44,$7D
	.dc.b	$A3,$E9,$FE,$7B,$63,$D0,$D7,$EA,$B5,$FF,$3E,$8F,$3F,$F0,$7F,$0B
	.dc.b	$FD,$03,$86,$CF,$FB,$E8,$FF,$5F,$1B,$3F,$C4,$7E,$F7,$CA,$F5,$BF
	.dc.b	$D2,$31,$83,$FD,$89,$F6,$7F,$E2,$77,$FD,$3D,$D3,$FE,$6D,$7E,$63
	.dc.b	$5F,$C4,$B5,$FF,$42,$B5,$FF,$11,$7E,$2F,$FF,$00,$5F,$F6,$87,$3F
	.dc.b	$F8,$ED,$BE,$DF,$DC,$9D,$47,$FA,$CB,$C5,$7D,$0F,$E4,$A3,$FD,$17
	.dc.b	$18,$7F,$D1,$2D,$7F,$D0,$ED,$7F,$CC,$3E,$8F,$B9,$FC,$DB,$5C,$E9
	.dc.b	$FC,$F6,$31,$1F,$D9,$BE,$C7,$F6,$8F,$F9,$A9,$AE,$7D,$4F,$E8,$91
	.dc.b	$88,$FD,$A7,$D9,$3E,$07,$F6,$6D,$8F,$F0,$8B,$5F,$F0,$83,$5F,$F5
	.dc.b	$6F,$C4,$FF,$4B,$7F,$D3,$BE,$77,$ED,$23,$11,$FB,$86,$BF,$EE,$AF
	.dc.b	$CD,$5F,$1F,$F1,$FE,$15,$18,$9F,$DD,$C6,$27,$53,$99,$FF,$50,$FD
	.dc.b	$77,$CF,$E8,$7C,$5A,$30,$D9,$FF,$5B,$C7,$97,$C5,$91,$FE,$B4,$6B
	.dc.b	$FE,$AC,$FA,$EE,$24,$7F,$D7,$EF,$4E,$C9,$1F,$EC,$26,$BE,$FE,$5F
	.dc.b	$84,$47,$FD,$82,$F9,$3B,$24,$7F,$B2,$9A,$FF,$1A,$FC,$9F,$F0,$5E
	.dc.b	$17,$F1,$18,$FF,$13,$C6,$3C,$2F,$F1,$2B,$61,$C0,$FA,$87,$CB,$7F
	.dc.b	$32,$97,$D9,$BE,$83,$FC,$0C,$D8,$FC,$C3,$5F,$E5,$6D,$7F,$D6,$B8
	.dc.b	$74,$7F,$C7,$39,$79,$54,$65,$46,$63,$0E,$E7,$B3,$ED,$F2,$9C,$DF
	.dc.b	$EA,$7B,$EB,$A7,$FA,$56,$31,$1D,$23,$1F,$17,$FC,$33,$91,$E4,$7F
	.dc.b	$0E,$F9,$27,$CC,$52,$61,$D0,$FF,$45,$63,$D3,$3F,$9D,$D1,$FE,$F5
	.dc.b	$F0,$BF,$9C,$F6,$3E,$7A,$9D,$62,$E1,$CE,$BF,$DF,$99,$78,$4C,$C7
	.dc.b	$9C,$F1,$FB,$FF,$AB,$C7,$EE,$67,$FE,$AF,$F0,$3F,$B3,$B6,$BA,$5B
	.dc.b	$EC,$3F,$B5,$F4,$3E,$A7,$EF,$57,$E9,$9A,$FF,$BA,$A3,$5D,$3F,$AD
	.dc.b	$9C,$CF,$F9,$B3,$0F,$F5,$F7,$D2,$FC,$7F,$86,$93,$F1,$67,$2F,$C7
	.dc.b	$E7,$FF,$24,$7D,$47,$C4,$24,$79,$BE,$1F,$E8,$7C,$4C,$8F,$4B,$A7
	.dc.b	$E1,$71,$37,$FE,$A1,$87,$F9,$4F,$CF,$F9,$E7,$D9,$F8,$2B,$7F,$0B
	.dc.b	$E0,$FE,$E1,$87,$E6,$0F,$B2,$65,$F9,$F6,$C7,$DF,$35,$1F,$85,$E9
	.dc.b	$D6,$9F,$E6,$7A,$FF,$59,$57,$F3,$96,$BF,$E0,$C6,$BF,$E9,$43,$C1
	.dc.b	$BF,$72,$3D,$E3,$8F,$37,$E3,$35,$7D,$87,$E3,$39,$E5,$FD,$E8,$AF
	.dc.b	$C6,$9F,$F0,$63,$F8,$2F,$73,$37,$FC,$0F,$9F,$F0,$5B,$7E,$13,$9B
	.dc.b	$FB,$C5,$FE,$3E,$72,$FD,$BF,$D3,$FF,$3D,$E0,$FF,$AC,$39,$D0,$FA
	.dc.b	$DF,$C8,$B6,$B8,$47,$5C,$73,$E5,$C9,$E8,$FE,$BC,$F1,$77,$7E,$3C
	.dc.b	$91,$F9,$57,$FB,$26,$1E,$0D,$FB,$92,$7F,$69,$9F,$E7,$6B,$F6,$3F
	.dc.b	$51,$27,$F2,$5C,$A8,$FD,$F7,$CE,$7E,$FF,$E6,$A5,$7E,$A1,$E0,$FE
	.dc.b	$C7,$AF,$FD,$41,$F5,$1F,$BC,$24,$7E,$BB,$EC,$7F,$84,$7D,$7C,$8F
	.dc.b	$D8,$BA,$79,$64,$F9,$4F,$F2,$8B,$63,$FC,$74,$D7,$FC,$AE,$D7,$FD
	.dc.b	$B7,$E0,$7E,$87,$D0,$FE,$2C,$48,$F6,$D7,$F1,$9F,$12,$47,$F0,$27
	.dc.b	$F8,$E7,$2F,$F1,$DF,$2F,$FB,$07,$67,$F2,$4A,$EA,$D7,$FD,$DB,$8F
	.dc.b	$EB,$E7,$FC,$CF,$FA,$03,$C8,$A4,$AF,$F2,$94,$65,$FF,$35,$D7,$F0
	.dc.b	$F7,$FC,$73,$D9,$FE,$89,$EC,$EA,$D8,$71,$2F,$94,$7D,$15,$F5,$1E
	.dc.b	$22,$47,$81,$FE,$F5,$5F,$E8,$78,$FE,$58,$D7,$F8,$4B,$7B,$BF,$A4
	.dc.b	$57,$49,$FE,$2C,$FF,$90,$DF,$07,$CE,$7F,$9C,$A3,$13,$FC,$FA,$BF
	.dc.b	$9E,$57,$0A,$F6,$FC,$24,$8C,$1D,$3E,$0F,$E4,$91,$F8,$EF,$A3,$EC
	.dc.b	$3E,$29,$23,$81,$EA,$F2,$FE,$49,$1F,$6F,$C9,$FC,$CF,$88,$F8,$24
	.dc.b	$8F,$A3,$97,$F9,$ED,$FE,$21,$C1,$FD,$07,$CD,$FA,$64,$8F,$8F,$5F
	.dc.b	$D2,$A9,$F5,$94,$93,$FC,$F7,$F4,$FF,$EF,$1B,$FF,$52,$AF,$CB,$4F
	.dc.b	$F0,$56,$BF,$DA,$9B,$E2,$FF,$51,$EA,$FA,$FF,$17,$F8,$4F,$46,$5F
	.dc.b	$D8,$8A,$FD,$8F,$33,$F1,$26,$31,$F0,$7C,$7F,$9C,$93,$F8,$0F,$AE
	.dc.b	$F5,$7D,$0F,$F0,$8D,$0D,$24,$FC,$6F,$73,$F2,$BE,$4C,$8F,$C4,$C7
	.dc.b	$FA,$CB,$EC,$FF,$CA,$7D,$C9,$9D,$C9,$CC,$61,$AF,$E7,$1B,$0E,$2F
	.dc.b	$E6,$B3,$FE,$21,$8C,$7A,$3D,$63,$13,$AB,$C5,$6A,$7C,$51,$FD,$2F
	.dc.b	$EC,$FF,$4B,$C9,$FE,$89,$6C,$FF,$B7,$1A,$EF,$C0,$FF,$0E,$B9,$FA
	.dc.b	$FF,$D3,$3E,$19,$46,$23,$E0,$46,$3E,$3B,$B4,$E7,$17,$EB,$AF,$33
	.dc.b	$FC,$A1,$9F,$A2,$30,$FA,$31,$F9,$31,$87,$E1,$36,$3D,$FF,$65,$FE
	.dc.b	$73,$78,$4F,$F0,$68,$C3,$F2,$5B,$1C,$3E,$1F,$C3,$3F,$C3,$BE,$33
	.dc.b	$FC,$C5,$D1,$FB,$9F,$97,$B7,$93,$F8,$3B,$8A,$FF,$20,$F8,$DF,$ED
	.dc.b	$19,$FE,$F1,$18,$F8,$8F,$F2,$C3,$FD,$8E,$74,$FF,$7D,$E2,$FF,$70
	.dc.b	$4E,$A3,$9F,$C4,$F5,$9F,$F7,$23,$DF,$C4,$FF,$1A,$7A,$79,$C1,$39
	.dc.b	$EC,$FC,$CE,$5F,$C7,$3B,$B1,$DB,$97,$5E,$16,$27,$F9,$0B,$74,$FF
	.dc.b	$25,$69,$FE,$F3,$F8,$FF,$23,$6B,$DA,$D9,$F5,$66,$AD,$AF,$91,$F3
	.dc.b	$1E,$7F,$AB,$B8,$7F,$30,$F1,$BD,$07,$B7,$B9,$18,$3F,$CC,$0E,$1F
	.dc.b	$E8,$ED,$C7,$F8,$BB,$5F,$EC,$2D,$7F,$AA,$AF,$A4,$EC,$77,$F4,$2D
	.dc.b	$76,$ED,$F1,$1F,$77,$B7,$FA,$77,$C7,$E2,$7F,$A0,$BF,$5E,$33,$1F
	.dc.b	$D7,$23,$1E,$8F,$FB,$42,$3F,$AF,$FD,$6F,$B7,$C3,$E9,$28,$C4,$FF
	.dc.b	$4C,$72,$7C,$3F,$4A,$C7,$C2,$F9,$EF,$8C,$72,$E1,$B1,$FD,$51,$F0
	.dc.b	$91,$FE,$DE,$FE,$37,$AF,$F8,$67,$A7,$F3,$CF,$1F,$1D,$AF,$EB,$AF
	.dc.b	$A8,$8F,$F7,$87,$97,$0F,$FB,$02,$3E,$9F,$C7,$E0,$8F,$D6,$DB,$EE
	.dc.b	$1F,$77,$EB,$EC,$7D,$CD,$7F,$00,$7C,$4E,$FB,$1F,$DB,$9F,$96,$47
	.dc.b	$FC,$12,$F8,$BE,$BB,$F1,$07,$CC,$F8,$EB,$3F,$DD,$9C,$3F,$8A,$DA
	.dc.b	$F3,$FF,$A1,$B4,$6B,$F9,$13,$F0,$1F,$9B,$7F,$A2,$6F,$F2,$31,$8F
	.dc.b	$6A,$47,$E5,$3B,$2E,$1F,$A8,$FC,$CD,$DB,$FB,$96,$D7,$F9,$43,$E2
	.dc.b	$1F,$29,$9A,$FE,$15,$F0,$DF,$E6,$68,$FE,$04,$DA,$FE,$16,$DC,$3F
	.dc.b	$D6,$17,$CA,$7F,$67,$87,$FB,$25,$FF,$56,$36,$7D,$DE,$5F,$8F,$E4
	.dc.b	$3F,$86,$7E,$76,$F8,$5A,$78,$B1,$72,$63,$2F,$43,$93,$8F,$A3,$63
	.dc.b	$FD,$71,$C4,$F8,$EF,$DD,$79,$9F,$29,$B3,$FB,$AE,$8C,$34,$E3,$1F
	.dc.b	$CA,$F6,$7D,$AB,$E0,$B0,$DA,$FF,$57,$3F,$32,$7F,$D6,$AE,$5B,$C9
	.dc.b	$FC,$AD,$E6,$D8,$E1,$1F,$EA,$D6,$BF,$DA,$DF,$91,$E8,$FF,$12,$8F
	.dc.b	$EF,$51,$8A,$FE,$17,$CB,$E4,$E2,$FC,$3F,$C6,$FF,$19,$EC,$B5,$FC
	.dc.b	$61,$FE,$71,$C4,$F3,$33,$97,$FB,$5A,$31,$5F,$2A,$F7,$F1,$72,$3E
	.dc.b	$6F,$B3,$6F,$FD,$1D,$C6,$8F,$CB,$9D,$BB,$2D,$7F,$BB,$7A,$1F,$1F
	.dc.b	$E2,$F9,$2F,$C0,$B6,$7F,$29,$E7,$7E,$43,$5F,$F3,$C4,$7E,$41,$B9
	.dc.b	$FE,$42,$7F,$A1,$E0,$C3,$F1,$67,$BB,$0B,$87,$E1,$3E,$33,$E0,$F3
	.dc.b	$FE,$B1,$B1,$FE,$25,$E5,$F9,$DE,$4F,$FB,$1B,$E4,$75,$39,$3F,$D1
	.dc.b	$BA,$FF,$BC,$1E,$86,$75,$57,$FB,$3B,$E4,$BE,$79,$FB,$AF,$8D,$B5
	.dc.b	$F9,$67,$0F,$19,$FF,$6A,$F2,$62,$FF,$DF,$FF,$A8,$7C,$8A,$FF,$7E
	.dc.b	$36,$19,$FE,$15,$7D,$AF,$F8,$3A,$31,$3B,$75,$1F,$E1,$29,$FF,$7E
	.dc.b	$35,$FF,$1D,$35,$FF,$24,$35,$4F,$CF,$AF,$1F,$07,$F0,$8F,$AB,$7C
	.dc.b	$87,$EA,$6B,$57,$F8,$DF,$91,$FE,$81,$B9,$A4,$C7,$FC,$2F,$EB,$3F
	.dc.b	$68,$7C,$27,$F5,$D6,$CE,$9B,$87,$F9,$44,$FF,$59,$B8,$7F,$DA,$38
	.dc.b	$3E,$53,$C2,$4A,$FF,$5F,$C6,$5B,$D7,$E6,$23,$83,$C1,$F2,$1E,$32
	.dc.b	$38,$39,$FC,$AF,$B8,$8E,$0F,$E5,$3F,$9D,$7E,$FB,$17,$9E,$8C,$37
	.dc.b	$C7,$FE,$63,$1B,$DB,$E2,$7C,$7F,$C3,$26,$1F,$0D,$F9,$BF,$2D,$23
	.dc.b	$DA,$DC,$7F,$C8,$7F,$6C,$D3,$F8,$A3,$6B,$FC,$8F,$28,$C3,$C2,$7F
	.dc.b	$8D,$7E,$11,$FD,$5A,$B6,$D9,$FF,$27,$B5,$FA,$8D,$7F,$DE,$BF,$1D
	.dc.b	$FE,$13,$7F,$D7,$AD,$7F,$D1,$5D,$9F,$E9,$5F,$6F,$FC,$0D,$39,$9F
	.dc.b	$F4,$CF,$6B,$E1,$BF,$CA,$5E,$4B,$5F,$F1,$34,$E7,$3F,$D6,$4E,$5F
	.dc.b	$DB,$6E,$FB,$67,$C8,$FF,$8A,$5B,$3F,$92,$6E,$5F,$8D,$AE,$15,$FE
	.dc.b	$2F,$FB,$6F,$F1,$45,$F4,$6D,$7E,$D2,$3F,$D6,$1B,$3E,$C1,$F3,$7F
	.dc.b	$E3,$BB,$DF,$89,$FC,$43,$27,$C7,$FF,$9B,$57,$C7,$D8,$C3,$FE,$43
	.dc.b	$FB,$5F,$F2,$56,$7F,$A8,$6E,$9E,$76,$BF,$EB,$26,$BF,$EC,$3F,$05
	.dc.b	$F6,$DF,$C1,$EB,$0D,$9F,$D8,$BF,$E7,$C8,$FE,$3D,$3C,$3D,$EF,$F4
	.dc.b	$82,$FC,$D3,$FD,$E3,$3A,$FB,$7F,$F3,$26,$BF,$D6,$6B,$5F,$21,$C2
	.dc.b	$73,$5F,$ED,$4F,$B7,$7C,$CF,$F9,$6D,$B1,$FC,$11,$BA,$7D,$36,$BF
	.dc.b	$ED,$1C,$1F,$27,$FE,$90,$8F,$A4,$7F,$0A,$71,$97,$8D,$FC,$34,$6A
	.dc.b	$78,$FC,$A7,$B3,$EA,$5F,$23,$4F,$F3,$97,$AB,$F5,$33,$F8,$73,$A7
	.dc.b	$F1,$B1,$9C,$3F,$10,$D7,$FD,$C6,$D7,$FD,$E6,$D7,$F8,$FF,$E2,$DF
	.dc.b	$E2,$B7,$F8,$E6,$28,$C4,$7B,$DA,$FD,$B6,$BF,$CA,$3E,$E3,$FC,$90
	.dc.b	$FB,$5B,$A7,$D3,$7F,$9A,$B7,$DD,$FB,$7F,$71,$F5,$2B,$F9,$8F,$AB
	.dc.b	$FD,$27,$3F,$C4,$4E,$72,$4E,$6B,$F9,$BC,$E4,$9F,$31,$D2,$FF,$C5
	.dc.b	$92,$3F,$48,$3E,$DF,$F8,$A3,$F6,$E3,$51,$88,$E0,$70,$70,$FF,$42
	.dc.b	$6B,$9F,$2B,$E4,$CF,$27,$FD,$C7,$E0,$59,$EC,$78,$63,$2B,$C4,$EB
	.dc.b	$9F,$F8,$C5,$EB,$5E,$97,$37,$6F,$8E,$18,$C3,$A1,$FD,$56,$FC,$0F
	.dc.b	$EC,$6A,$BF,$0E,$2E,$1F,$E3,$FA,$EF,$3C,$8E,$05,$38,$1F,$CE,$0F
	.dc.b	$96,$7F,$3A,$9C,$38,$3E,$4F,$D9,$FF,$71,$FB,$8F,$E7,$ED,$8F,$ED
	.dc.b	$41,$C3,$FD,$B5,$AF,$2F,$1F,$FC,$22,$F0,$84,$62,$38,$35,$E1,$C1
	.dc.b	$E0,$8E,$3A,$7D,$DA,$7B,$BF,$7A,$30,$DD,$78,$E1,$F5,$5B,$3A,$8F
	.dc.b	$A4,$7E,$36,$B9,$B8,$3D,$97,$E0,$AE,$27,$C4,$BE,$D3,$C5,$AF,$0C
	.dc.b	$39,$5F,$B9,$B3,$A8,$EF,$3E,$53,$C5,$7D,$DB,$5C,$18,$6C,$EC,$71
	.dc.b	$D1,$F6,$EF,$97,$FE,$EB,$18,$6C,$66,$7E,$CC,$E5,$BF,$71,$FE,$2A
	.dc.b	$6C,$72,$8C,$47,$F0,$E6,$BF,$CA,$FE,$E3,$FA,$8B,$D1,$AF,$F1,$46
	.dc.b	$BB,$F1,$7F,$C1,$3E,$3B,$F1,$6F,$E7,$73,$FE,$3C,$6B,$FD,$AD,$AF
	.dc.b	$EF,$23,$87,$83,$CB,$67,$35,$25,$7F,$91,$A3,$2F,$C0,$FB,$97,$57
	.dc.b	$C6,$6C,$66,$30,$7C,$AF,$8D,$FF,$04,$47,$C8,$6B,$FC,$C1,$AF,$F4
	.dc.b	$D6,$BF,$C6,$71,$7D,$CB,$EA,$3E,$0B,$67,$FC,$B6,$D7,$FC,$14,$D7
	.dc.b	$FA,$C7,$DC,$BD,$9F,$BA,$D8,$FB,$6D,$7F,$A1,$B5,$FE,$C1,$F1,$FF
	.dc.b	$E6,$B7,$F0,$2D,$7F,$A4,$B5,$DF,$C1,$FD,$6F,$B9,$FF,$6E,$47,$F9
	.dc.b	$C1,$AF,$F8,$7D,$AF,$D5,$6B,$F0,$7F,$05,$F8,$6B,$34,$61,$BE,$C7
	.dc.b	$E1,$23,$1F,$42,$F9,$7F,$C3,$B6,$FF,$D0,$2D,$7F,$C6,$0D,$7F,$0D
	.dc.b	$F5,$DF,$EB,$F7,$FD,$0E,$D7,$FC,$6C,$D7,$7F,$17,$F9,$4F,$B9,$7D
	.dc.b	$3D,$24,$FF,$A4,$23,$2F,$F9,$8B,$EE,$7F,$C6,$6D,$F6,$BF,$97,$46
	.dc.b	$23,$F9,$3B,$5F,$F4,$97,$E0,$AF,$A2,$FE,$68,$D8,$FF,$36,$B5,$FE
	.dc.b	$5E,$D7,$FD,$35,$F8,$23,$EC,$3F,$9E,$36,$3F,$CE,$AD,$7F,$9B,$B5
	.dc.b	$FD,$57,$DB,$7D,$F7,$F2,$CD,$7E,$B3,$78,$8E,$61,$6F,$A9,$F4,$92
	.dc.b	$3D,$2F,$EF,$BE,$4B,$C1,$1F,$A3,$6B,$FE,$B2,$C7,$46,$FC,$19,$1F
	.dc.b	$EB,$96,$BF,$0B,$7F,$8E,$47,$E3,$BD,$FE,$16,$47,$FB,$05,$AF,$F9
	.dc.b	$D6,$DF,$5D,$E5,$24,$7F,$B1,$DF,$A7,$F3,$92,$3E,$9B,$5E,$5C,$3E
	.dc.b	$D2,$3F,$EC,$F7,$FB,$77,$E2,$9F,$DB,$09,$18,$3F,$B9,$7D,$BB,$ED
	.dc.b	$E9,$23,$FD,$5B,$19,$6F,$C4,$F0,$23,$83,$FD,$AB,$F3,$9F,$EB,$4F
	.dc.b	$8B,$7C,$BA,$30,$D9,$D9,$FE,$DC,$FA,$6C,$91,$C1,$FD,$FA,$9F,$20
	.dc.b	$F2,$7B,$36,$74,$D7,$FD,$6F,$1A,$F1,$64,$69,$AF,$FB,$CF,$E8,$9F
	.dc.b	$89,$FF,$32,$8C,$35,$FF,$1D,$C6,$BC,$CF,$F6,$8B,$63,$FC,$27,$F3
	.dc.b	$0F,$C4,$9F,$41,$FE,$D6,$8C,$36,$33,$3C,$1A,$F8,$3D,$EE,$B8,$FF
	.dc.b	$8F,$63,$FC,$3F,$87,$9A,$72,$78,$BE,$05,$9D,$25,$70,$8C,$B7,$EA
	.dc.b	$38,$11,$C1,$F3,$F8,$5F,$55,$F6,$09,$19,$7E,$DE,$92,$38,$35,$FF
	.dc.b	$1A,$FD,$1B,$EB,$3F,$C8,$0D,$8F,$F7,$CB,$5D,$C6,$BC,$79,$1A,$6B
	.dc.b	$D3,$EB,$B9,$91,$E6,$FE,$3B,$E9,$24,$7A,$DA,$FF,$9D,$07,$37,$E5
	.dc.b	$BA,$7F,$2A,$48,$F8,$4F,$DE,$FC,$4F,$FB,$14,$7F,$A0,$F8,$91,$88
	.dc.b	$46,$3D,$8E,$07,$1F,$5E,$4F,$DA,$AF,$EC,$8F,$36,$BC,$97,$D0,$F8
	.dc.b	$49,$1C,$5F,$97,$F1,$52,$3E,$5B,$5F,$CE,$4B,$EA,$BD,$44,$8F,$13
	.dc.b	$FB,$7F,$C4,$FD,$31,$FA,$F6,$BF,$1F,$44,$62,$3F,$D3,$B1,$8C,$F9
	.dc.b	$74,$49,$E4,$D7,$97,$2F,$F0,$B6,$F8,$5F,$C1,$08,$FE,$FF,$C5,$FF
	.dc.b	$43,$60,$69,$23,$FC,$73,$F4,$1F,$D8,$9F,$B9,$E8,$23,$11,$FE,$37
	.dc.b	$6B,$FE,$23,$6B,$F8,$29,$FF,$25,$BF,$EB,$FA,$F0,$7C,$2F,$F1,$B9
	.dc.b	$CC,$EE,$31,$AC,$8F,$43,$5F,$1D,$FF,$BD,$1F,$F4,$BE,$1E,$5F,$94
	.dc.b	$F4,$92,$33,$2F,$63,$DA,$48,$CB,$E3,$E2,$91,$F1,$DA,$F8,$27,$E2
	.dc.b	$91,$EA,$FA,$FE,$B6,$47,$CF,$6B,$F7,$F3,$7D,$DF,$C1,$24,$7D,$57
	.dc.b	$E8,$FD,$94,$8F,$4B,$5F,$A7,$F6,$9F,$E7,$F6,$FC,$67,$F9,$C2,$31
	.dc.b	$1F,$E6,$B6,$BF,$46,$FF,$DC,$AF,$D5,$F9,$0F,$A4,$47,$F9,$1F,$8F
	.dc.b	$FD,$EB,$9B,$F9,$87,$AD,$FC,$D2,$30,$4E,$CC,$12,$B8,$BF,$77,$EE
	.dc.b	$E4,$7E,$45,$AF,$F4,$59,$7B,$FD,$49,$1E,$27,$C1,$DD,$23,$CA,$D7
	.dc.b	$E6,$47,$E2,$C8,$FE,$7D,$FD,$97,$C1,$48,$FD,$93,$5F,$EA,$DF,$75
	.dc.b	$FB,$17,$F8,$93,$5F,$E2,$2D,$CC,$DE,$4F,$E8,$33,$FD,$7D,$AF,$F5
	.dc.b	$C6,$BF,$C7,$5A,$FE,$97,$0F,$D1,$11,$FD,$B3,$FE,$AA,$F4,$3F,$D9
	.dc.b	$7E,$0F,$FA,$95,$B3,$FE,$91,$6B,$FE,$71,$6F,$4B,$ED,$7F,$D0,$55
	.dc.b	$FE,$AD,$6D,$7F,$A6,$5A,$FF,$9E,$5A,$FB,$BE,$DF,$E1,$64,$69,$33
	.dc.b	$FE,$81,$23,$E2,$FB,$7F,$BC,$AF,$F3,$7F,$76,$A7,$2D,$8C,$C7,$F0
	.dc.b	$EF,$BA,$FE,$CB,$97,$33,$FD,$88,$E2,$11,$8F,$36,$C4,$3F,$D2,$53
	.dc.b	$B9,$FF,$66,$7D,$D7,$FA,$3D,$E9,$3F,$EB,$98,$C3,$C7,$54,$62,$78
	.dc.b	$FC,$37,$67,$F9,$AF,$4B,$EE,$FC,$18,$7E,$B5,$AF,$F2,$16,$F8,$9F
	.dc.b	$CB,$DC,$EE,$FD,$6D,$F4,$38,$6E,$3F,$E1,$46,$E1,$FE,$2B,$6B,$FE
	.dc.b	$28,$5F,$27,$C4,$93,$C5,$F0,$F9,$7F,$E4,$78,$FE,$A5,$EE,$60,$8F
	.dc.b	$F4,$5E,$EF,$3E,$BF,$ED,$B8,$C1,$31,$FF,$1E,$7E,$11,$FC,$6A,$3F
	.dc.b	$C5,$2D,$8C,$F8,$BE,$E8,$C4,$7F,$8F,$9F,$EE,$33,$C3,$B3,$44,$7F
	.dc.b	$AB,$78,$3E,$22,$38,$3B,$F9,$5D,$48,$E0,$F1,$DF,$84,$8E,$0F,$EC
	.dc.b	$5F,$09,$EE,$23,$83,$FA,$5F,$8B,$FE,$A1,$6F,$C2,$FD,$CE,$84,$E6
	.dc.b	$FF,$99,$CE,$7C,$AF,$F6,$05,$FF,$50,$F5,$30,$47,$FB,$97,$C2,$FC
	.dc.b	$E2,$38,$3F,$BB,$7C,$27,$D2,$23,$83,$FA,$DF,$D4,$77,$23,$83,$FB
	.dc.b	$F7,$C2,$FD,$A2,$38,$3F,$B1,$F6,$F9,$C8,$E0,$FE,$CF,$F0,$5E,$B2
	.dc.b	$38,$3F,$C1,$5D,$3F,$86,$23,$83,$FC,$1D,$D7,$F8,$A2,$38,$3F,$BB
	.dc.b	$FE,$9D,$FE,$3C,$6F,$6F,$CD,$8C,$70,$7F,$84,$1A,$FE,$C3,$EB,$BF
	.dc.b	$2A,$4A,$FF,$0E,$3F,$98,$FA,$69,$1F,$E1,$F6,$CF,$F8,$BF,$A7,$F2
	.dc.b	$6F,$F8,$4D,$AF,$F9,$61,$BE,$09,$A3,$D0,$FD,$DB,$6F,$FD,$54,$DA
	.dc.b	$ED,$3F,$BE,$F8,$F7,$C3,$D2,$5F,$FA,$42,$32,$FF,$25,$FA,$0F,$F0
	.dc.b	$B3,$7D,$7F,$E1,$78,$27,$2F,$F3,$06,$BF,$EA,$AA,$7A,$3E,$12,$5F
	.dc.b	$77,$C9,$F4,$B2,$3C,$4D,$AF,$F5,$C5,$3E,$0B,$D4,$4A,$F4,$3F,$C1
	.dc.b	$BE,$A6,$47,$F9,$35,$B5,$FE,$5D,$F3,$3F,$9B,$CF,$FB,$3D,$B1,$98
	.dc.b	$FE,$4B,$18,$8F,$F4,$57,$DD,$3A,$7E,$0B,$6B,$FA,$53,$67,$FA,$8B
	.dc.b	$5E,$3D,$7C,$09,$3C,$1E,$7D,$92,$39,$B6,$7D,$1F,$37,$E4,$24,$FA
	.dc.b	$5F,$3F,$91,$23,$D8,$D9,$FF,$70,$7A,$BF,$C0,$9B,$E0,$7F,$40,$8C
	.dc.b	$47,$FB,$81,$AF,$FA,$85,$7C,$26,$C9,$7F,$ED,$68,$CB,$7D,$CE,$84
	.dc.b	$70,$7F,$BA,$3E,$2B,$FA,$ED,$FF,$75,$1F,$17,$FE,$F4,$F0,$FF,$AB
	.dc.b	$B6,$B8,$47,$9F,$C6,$91,$E1,$6D,$7E,$AF,$DD,$FD,$19,$27,$F4,$6F
	.dc.b	$E9,$BD,$C9,$1F,$A6,$6C,$FE,$17,$D9,$FC,$09,$27,$F0,$CF,$E1,$3D
	.dc.b	$A9,$1F,$89,$6C,$FF,$84,$21,$F6,$BF,$A6,$69,$FE,$0F,$2C,$6E,$35
	.dc.b	$F3,$12,$74,$D9,$E9,$F4,$5C,$C9,$3C,$DF,$E2,$D5,$23,$F8,$DB,$67
	.dc.b	$FC,$47,$E4,$3E,$0F,$F7,$2D,$AF,$F1,$13,$67,$73,$FC,$87,$ED,$65
	.dc.b	$7F,$26,$6C,$FF,$23,$FB,$5F,$E4,$C4,$9F,$E5,$6F,$6C,$E4,$76,$6C
	.dc.b	$FF,$2B,$F0,$BF,$97,$12,$7F,$99,$BF,$CC,$7D,$99,$1F,$CD,$9B,$3F
	.dc.b	$E3,$EE,$37,$97,$FC,$51,$B5,$FE,$3D,$6C,$EE,$7F,$B4,$7C,$A3,$E0
	.dc.b	$E9,$2F,$FD,$73,$19,$7F,$C8,$DD,$5F,$C6,$9B,$97,$F2,$28,$C6,$1F
	.dc.b	$E5,$06,$BF,$9D,$3F,$3B,$DF,$E5,$25,$79,$5F,$E5,$7F,$67,$FA,$49
	.dc.b	$FD,$23,$5F,$F2,$D3,$53,$9F,$59,$C9,$FE,$32,$9F,$F3,$04,$63,$D1
	.dc.b	$FF,$03,$4E,$65,$39,$C3,$FD,$AF,$F6,$6F,$63,$FC,$18,$D8,$CC,$60
	.dc.b	$FF,$08,$7D,$9F,$F8,$3E,$FF,$75,$68,$C4,$78,$9A,$F8,$1B,$1F,$EE
	.dc.b	$1E,$AF,$F0,$D7,$9C,$69,$30,$FF,$75,$FD,$43,$D4,$FF,$10,$36,$33
	.dc.b	$18,$3F,$CB,$FF,$0C,$FB,$27,$C9,$61,$B5,$FE,$86,$9D,$4E,$4E,$3F
	.dc.b	$65,$D7,$5F,$B5,$28,$C4,$7F,$19,$8C,$61,$C0,$FF,$1A,$7D,$8F,$F9
	.dc.b	$5E,$7F,$05,$AA,$31,$1F,$21,$AF,$D8,$6A,$AF,$81,$ED,$FB,$2F,$FC
	.dc.b	$DB,$97,$A5,$B5,$E6,$6B,$E5,$FB,$1F,$F5,$34,$FB,$DB,$1E,$C6,$BB
	.dc.b	$EE,$FF,$26,$7A,$DF,$D6,$63,$F1,$6D,$7F,$12,$DC,$CD,$FC,$3D,$F7
	.dc.b	$DF,$E5,$56,$E3,$FE,$B4,$A4,$62,$FE,$2B,$5F,$99,$F6,$4F,$95,$FB
	.dc.b	$6D,$BF,$B0,$D9,$FA,$AD,$7E,$A7,$D9,$3D,$BF,$C1,$B6,$BF,$00,$D9
	.dc.b	$FB,$AD,$7E,$E7,$63,$DA,$FC,$5B,$6B,$F1,$0D,$9F,$C2,$B5,$FF,$14
	.dc.b	$1E,$9F,$7B,$F1,$33,$FE,$D3,$CB,$FC,$71,$3F,$C1,$7D,$DC,$56,$A3
	.dc.b	$81,$92,$9F,$ED,$BE,$4F,$F6,$6D,$72,$8F,$ED,$55,$C9,$FC,$A5,$D7
	.dc.b	$27,$23,$F1,$8E,$5F,$F7,$34,$5A,$EB,$7F,$E3,$B9,$E1,$18,$7C,$97
	.dc.b	$E1,$79,$9D,$4F,$F2,$21,$C2,$FB,$38,$8E,$13,$CE,$3C,$A7,$C8,$21
	.dc.b	$E1,$3F,$95,$7D,$5B,$CF,$7D,$8F,$FB,$E6,$30,$D8,$CC,$FF,$31,$13
	.dc.b	$A7,$0F,$F2,$8F,$B1,$E5,$E1,$9A,$49,$DC,$65,$7C,$FF,$D1,$A7,$7F
	.dc.b	$F8,$6E,$30,$DD,$7F,$C3,$D3,$DE,$B8,$C6,$2F,$A5,$F0,$91,$A2,$D7
	.dc.b	$67,$31,$E1,$9E,$35,$CF,$0F,$C0,$5E,$EF,$F1,$2D,$D3,$84,$EA,$F8
	.dc.b	$D7,$6C,$7D,$0D,$3B,$E1,$D5,$F2,$1F,$32,$78,$BF,$3A,$78,$B4,$E2
	.dc.b	$7A,$A3,$BE,$8F,$B1,$7C,$4F,$F9,$0F,$16,$FF,$9F,$9C,$E1,$F1,$27
	.dc.b	$2F,$F2,$58,$CB,$F6,$67,$87,$AF,$CD,$FD,$47,$D8,$BE,$33,$15,$FC
	.dc.b	$B6,$30,$F5,$9D,$CE,$B2,$F3,$65,$7E,$C5,$CB,$7D,$19,$8F,$13,$F9
	.dc.b	$C4,$E9,$C5,$A3,$FC,$2F,$1F,$D4,$5A,$FF,$4E,$6B,$FD,$35,$AF,$F7
	.dc.b	$CE,$27,$C0,$7E,$76,$3F,$BD,$C6,$1F,$E5,$CD,$7F,$A8,$B5,$FE,$B1
	.dc.b	$F6,$1F,$C9,$5F,$2B,$5F,$94,$7F,$51,$71,$C1,$F4,$3B,$BF,$57,$7F
	.dc.b	$17,$E3,$1E,$B5,$F8,$99,$8F,$C0,$9C,$BF,$16,$DF,$61,$FE,$0F,$8F
	.dc.b	$E8,$8D,$7F,$B3,$B5,$FF,$06,$B5,$FE,$45,$F6,$0F,$97,$FA,$CD,$8F
	.dc.b	$9E,$D7,$D7,$BA,$1F,$2B,$F8,$2C,$1F,$61,$FC,$AF,$6F,$C6,$DB,$E9
	.dc.b	$FF,$CF,$78,$FD,$A6,$BF,$1A,$F8,$F6,$F3,$C1,$F0,$9F,$56,$7A,$B6
	.dc.b	$33,$5F,$D8,$DA,$FF,$81,$BE,$C3,$E3,$BD,$5B,$97,$D1,$6F,$57,$F0
	.dc.b	$6F,$92,$FE,$0F,$9F,$AF,$D6,$7D,$47,$E7,$A5,$F8,$2F,$F7,$4C,$BF
	.dc.b	$2D,$18,$F3,$B8,$57,$C3,$F8,$87,$DD,$FA,$5B,$7F,$DE,$5A,$FF,$87
	.dc.b	$1A,$FF,$90,$BE,$C9,$E9,$FC,$D6,$C7,$E0,$5A,$FE,$3B,$1F,$F0,$F7
	.dc.b	$C7,$FE,$D2,$1F,$63,$F5,$9F,$CF,$35,$FE,$78,$D7,$FC,$DC,$D7,$FD
	.dc.b	$99,$F8,$C7,$67,$FC,$E4,$D7,$38,$7E,$A5,$46,$55,$BE,$61,$ED,$7E
	.dc.b	$BE,$30,$D8,$CC,$FF,$4F,$9C,$BF,$E9,$5F,$C1,$7F,$8A,$BF,$ED,$FF
	.dc.b	$8A,$46,$23,$F9,$FC,$63,$C9,$FF,$4C,$4E,$67,$FC,$EF,$F1,$2F,$9D
	.dc.b	$FC,$63,$63,$FD,$2F,$18,$3F,$82,$77,$3F,$06,$D4,$FF,$96,$9A,$FF
	.dc.b	$96,$5A,$FF,$83,$5A,$FC,$2F,$9A,$FF,$64,$3E,$D6,$BF,$E9,$DC,$D1
	.dc.b	$88,$FF,$7E,$46,$3E,$43,$F4,$5F,$38,$FA,$FA,$4A,$FF,$22,$46,$5E
	.dc.b	$DD,$3F,$3F,$0F,$CC,$47,$76,$BD,$9A,$FE,$8D,$75,$F1,$24,$7A,$5F
	.dc.b	$DA,$7E,$D3,$FE,$6D,$8F,$F2,$A3,$5F,$F3,$BB,$5F,$F4,$EB,$5E,$DF
	.dc.b	$55,$E5,$23,$F1,$5F,$F6,$87,$A1,$FE,$49,$8F,$E8,$83,$F2,$7F,$F4
	.dc.b	$9B,$86,$BF,$CC,$39,$3F,$A3,$7B,$FF,$A3,$FC,$1F,$E9,$C9,$A4,$E3
	.dc.b	$FC,$D7,$D1,$3D,$0F,$D5,$36,$BC,$2D,$7F,$9E,$9C,$E7,$F8,$36,$DF
	.dc.b	$DA,$BC,$7F,$ED,$BC,$9F,$73,$D9,$C9,$AE,$AB,$87,$F0,$AB,$7C,$E7
	.dc.b	$23,$31,$FB,$D8,$7E,$09,$D8,$91,$C7,$E5,$7C,$3E,$1C,$8E,$8E,$9F
	.dc.b	$EA,$7F,$35,$FE,$C5,$D7,$FB,$CF,$13,$F4,$2F,$E0,$9B,$FF,$81,$23
	.dc.b	$1E,$97,$F7,$EB,$E1,$87,$F7,$59,$CB,$FE,$45,$6B,$FE,$F7,$6B,$FB
	.dc.b	$7B,$FE,$18,$A7,$FD,$AD,$5D,$7E,$4B,$C3,$E7,$7F,$8E,$27,$35,$DE
	.dc.b	$F2,$FE,$AF,$E4,$7E,$7F,$24,$8F,$95,$15,$F9,$1F,$3B,$F5,$0C,$69
	.dc.b	$F5,$38,$7F,$9E,$CF,$D1,$F3,$BE,$C9,$F6,$BD,$9F,$B7,$E8,$C9,$FA
	.dc.b	$CE,$9F,$E8,$3F,$82,$7A,$DF,$CA,$FD,$2F,$FB,$C5,$BC,$1F,$D2,$67
	.dc.b	$51,$F8,$89,$FF,$05,$3F,$E5,$60,$E1,$FF,$2C,$B5,$CF,$BB,$FC,$BE
	.dc.b	$1F,$45,$FC,$FD,$B3,$E2,$8C,$4F,$87,$E2,$24,$FA,$1F,$27,$DB,$79
	.dc.b	$7E,$3E,$4F,$B1,$DB,$F8,$8F,$D5,$5F,$67,$F8,$35,$F6,$D4,$C4,$59
	.dc.b	$FF,$1D,$73,$A9,$F6,$FC,$4D,$4F,$F5,$9D,$BF,$B1,$F2,$7F,$95,$2F
	.dc.b	$87,$D1,$F7,$3E,$57,$B9,$F9,$8F,$B7,$7D,$97,$F8,$E7,$BD,$18,$CB
	.dc.b	$F2,$8D,$7F,$28,$D7,$E8,$61,$E5,$FA,$6F,$61,$23,$F3,$D3,$C7,$2F
	.dc.b	$DD,$E7,$F5,$A7,$F0,$3F,$35,$F7,$4F,$BD,$CD,$63,$F0,$BF,$27,$F8
	.dc.b	$13,$0F,$BF,$E4,$FF,$09,$F2,$B2,$3F,$17,$15,$F8,$9F,$4F,$F8,$93
	.dc.b	$19,$7F,$92,$7D,$4F,$F3,$24,$FF,$2E,$F9,$2F,$E6,$64,$7F,$B5,$68
	.dc.b	$DB,$F2,$44,$8F,$F2,$55,$72,$BF,$6F,$1F,$E5,$FE,$9A,$57,$E7,$A2
	.dc.b	$BF,$D5,$F0,$4F,$F7,$B6,$C7,$82,$3F,$C7,$1F,$B5,$7F,$A3,$A3,$F3
	.dc.b	$C3,$E3,$11,$98,$FD,$14,$63,$C9,$E2,$75,$F9,$1F,$D5,$12,$7F,$CC
	.dc.b	$55,$D7,$2F,$F3,$63,$FA,$9B,$EB,$E0,$78,$89,$5F,$13,$D4,$FD,$A1
	.dc.b	$1E,$71,$FB,$7E,$99,$1F,$BE,$7F,$75,$EE,$FF,$97,$F0,$FF,$18,$E3
	.dc.b	$E0,$F8,$6C,$1C,$2F,$FC,$B9,$EA,$7F,$38,$7F,$DD,$59,$FF,$AD,$BD
	.dc.b	$3F,$F2,$E4,$E6,$7E,$2C,$E5,$FB,$3D,$FF,$C2,$48,$FF,$A0,$67,$86
	.dc.b	$1F,$94,$F0,$BF,$87,$78,$52,$BF,$8A,$C7,$F9,$67,$F0,$8F,$F1,$B7
	.dc.b	$AE,$69,$27,$FC,$68,$FF,$79,$BF,$EB,$2D,$E3,$FF,$33,$36,$7F,$C7
	.dc.b	$AF,$DC,$9F,$DC,$7E,$AD,$FE,$A7,$6F,$93,$FD,$26,$3F,$8A,$D7,$F5
	.dc.b	$27,$57,$FC,$FD,$FC,$14,$58,$FF,$5A,$78,$69,$CC,$FF,$49,$6E,$3E
	.dc.b	$BF,$97,$F7,$12,$7F,$CE,$B9,$3E,$C7,$E1,$7D,$CC,$9C,$BF,$17,$EC
	.dc.b	$FE,$37,$CA,$7C,$7F,$94,$95,$CA,$3F,$B1,$7D,$AB,$E9,$7C,$AD,$9F
	.dc.b	$CE,$9F,$D7,$5C,$3F,$A0,$D9,$18,$8F,$F1,$EE,$6F,$93,$FA,$C4,$8F
	.dc.b	$4F,$D7,$7F,$AB,$7E,$E1,$E2,$7F,$47,$E3,$FD,$3C,$62,$87,$EB,$FA
	.dc.b	$91,$98,$C3,$67,$8F,$91,$CE,$78,$D7,$08,$C4,$7E,$DE,$BF,$D6,$F5
	.dc.b	$C2,$3B,$BF,$BC,$FB,$8F,$F4,$BE,$BF,$EE,$B6,$CF,$C4,$6B,$BE,$EF
	.dc.b	$EC,$6F,$F6,$18,$7A,$DF,$E4,$58,$46,$3E,$3F,$A4,$E6,$31,$3F,$A8
	.dc.b	$F6,$FF,$CC,$98,$F9,$23,$D7,$18,$7E,$23,$63,$F8,$7B,$63,$7E,$3A
	.dc.b	$79,$47,$F1,$79,$FF,$7A,$C6,$35,$FF,$3B,$D6,$EB,$AB,$67,$EE,$FC
	.dc.b	$DE,$7B,$BF,$56,$D8,$FD,$5F,$C0,$DB,$FD,$5C,$DD,$9D,$5F,$ED,$2B
	.dc.b	$FC,$6C,$63,$DD,$FD,$3B,$63,$33,$BB,$FF,$46,$73,$7F,$A7,$E7,$F4
	.dc.b	$7F,$5B,$DF,$BF,$F8,$17,$C1,$DA,$4D,$FB,$39,$7F,$AA,$FB,$1B,$8F
	.dc.b	$FB,$3E,$DF,$6B,$C4,$E5,$F3,$16,$FF,$C0,$3A,$3E,$17,$CE,$D8,$F3
	.dc.b	$C6,$0C,$FC,$4F,$C1,$97,$C4,$D8,$C3,$E7,$F6,$25,$FE,$FB,$E8,$EE
	.dc.b	$9E,$57,$F8,$5F,$0F,$F1,$F6,$EA,$69,$8F,$13,$E1,$7C,$77,$C3,$6E
	.dc.b	$1F,$EF,$EF,$27,$F0,$8D,$3E,$C7,$5B,$CF,$6D,$F3,$6B,$7A,$7F,$42
	.dc.b	$8F,$E3,$B5,$B6,$D7,$0B,$FE,$8A,$DA,$FE,$87,$E9,$7F,$AC,$A7,$FA
	.dc.b	$FF,$C9,$75,$F3,$FC,$33,$9A,$77,$7F,$A4,$33,$FF,$22,$E1,$E9,$BF
	.dc.b	$27,$AD,$C6,$3D,$9E,$1F,$9C,$E7,$9D,$8F,$ED,$AF,$A3,$2F,$74,$7F
	.dc.b	$5A,$6E,$5E,$69,$F9,$F8,$BC,$5F,$84,$79,$BE,$3A,$E1,$FD,$E1,$E8
	.dc.b	$47,$A3,$CB,$E5,$64,$7F,$2C,$E5,$FE,$75,$18,$C7,$FD,$D1,$83,$E5
	.dc.b	$3E,$84,$FD,$16,$DF,$D9,$E0,$FA,$87,$C8,$E8,$B1,$FE,$0A,$7D,$C4
	.dc.b	$7D,$CF,$B3,$E5,$3E,$E6,$BF,$75,$B3,$F8,$5D,$30,$D3,$F1,$11,$F8
	.dc.b	$3F,$91,$FC,$27,$B3,$67,$F1,$8F,$E1,$F9,$71,$B7,$DA,$9F,$E7,$7B
	.dc.b	$FF,$59,$9C,$F2,$FF,$99,$7F,$5A,$7A,$7F,$C1,$DB,$7F,$EB,$5F,$4F
	.dc.b	$FC,$80,$FF,$9C,$A3,$38,$FF,$B9,$67,$2F,$FA,$DD,$AF,$FA,$4F,$E2
	.dc.b	$3F,$AB,$FC,$17,$F7,$46,$DF,$D3,$E5,$FD,$AB,$D7,$C1,$FD,$07,$1F
	.dc.b	$8D,$FE,$35,$FB,$0F,$FA,$02,$7F,$D0,$6F,$FA,$21,$BE,$EE,$BD,$CD
	.dc.b	$FB,$FC,$22,$CE,$8F,$D9,$78,$DF,$AD,$3F,$5D,$F2,$56,$3F,$6C,$E1
	.dc.b	$FD,$C7,$C9,$5D,$FF,$DE,$FF,$55,$F8,$79,$FE,$CF,$C9,$FE,$55,$8C
	.dc.b	$4E,$BB,$BF,$9A,$28,$CC,$EE,$7F,$C2,$70,$69,$2B,$E3,$7A,$3C,$A7
	.dc.b	$F4,$91,$FC,$2E,$31,$E0,$7F,$07,$F9,$39,$3F,$C3,$FE,$4E,$E5,$F8
	.dc.b	$31,$D7,$FE,$1B,$8F,$E1,$C7,$F3,$BF,$63,$FD,$D9,$5C,$4F,$69,$F8
	.dc.b	$A5,$F2,$7B,$F9,$3E,$0D,$9D,$E1,$DA,$9F,$27,$A9,$CE,$3F,$DD,$23
	.dc.b	$15,$E5,$F5,$DC,$FF,$ED,$AB,$FB,$FF,$1F,$FD,$77,$B3,$9F,$2F,$E1
	.dc.b	$23,$18,$A3,$F1,$CD,$AF,$F1,$5D,$70,$6C,$67,$1F,$F1,$66,$0F,$92
	.dc.b	$E7,$87,$E4,$1B,$7E,$8D,$34,$D8,$F9,$19,$7C,$96,$C6,$E3,$E6,$7C
	.dc.b	$8E,$31,$FA,$5A,$FF,$85,$3E,$47,$E8,$7C,$87,$D1,$6E,$1D,$FE,$47
	.dc.b	$0D,$3C,$51,$F8,$CF,$C5,$9C,$74,$FC,$7F,$E1,$6E,$CC,$5F,$D4,$CD
	.dc.b	$39,$8F,$F2,$7C,$62,$7F,$CC,$B3,$98,$FF,$29,$B5,$FD,$E7,$E1,$4F
	.dc.b	$7B,$C5,$CF,$F5,$63,$18,$77,$CB,$F0,$EF,$C9,$F8,$7F,$ED,$7C,$66
	.dc.b	$92,$3F,$D3,$BF,$05,$FC,$2B,$2F,$F2,$5B,$5F,$F6,$03,$7D,$83,$9E
	.dc.b	$2F,$F1,$FF,$93,$D8,$97,$FD,$07,$1D,$1F,$E6,$27,$1F,$11,$E4,$23
	.dc.b	$E2,$F1,$BF,$B7,$E0,$69,$23,$A7,$C9,$7B,$08,$E0,$F0,$7C,$37,$C0
	.dc.b	$23,$83,$FD,$77,$F3,$7F,$E4,$4D,$7F,$D7,$2D,$7F,$DB,$6D,$7F,$CD
	.dc.b	$AD,$5D,$EE,$1F,$F4,$B5,$7E,$81,$B3,$FE,$E8,$9D,$79,$BF,$CC,$E7
	.dc.b	$33,$FE,$80,$3F,$82,$FE,$1D,$FC,$D1,$FF,$77,$F0,$23,$11,$B8,$C7
	.dc.b	$9D,$FE,$26,$9C,$CF,$3F,$4B,$E4,$65,$F8,$E6,$C7,$FA,$79,$AF,$FB
	.dc.b	$F1,$AF,$8F,$F0,$BF,$F2,$83,$FE,$A4,$6B,$E7,$0E,$1E,$1F,$11,$FE
	.dc.b	$26,$EF,$FF,$26,$DF,$F1,$4E,$64,$62,$BF,$67,$D3,$F8,$37,$F9,$34
	.dc.b	$EB,$E1,$BF,$C6,$BF,$71,$D6,$52,$8C,$37,$1F,$F1,$AC,$63,$EE,$1E
	.dc.b	$D7,$F9,$9F,$CF,$FE,$B5,$87,$69,$CE,$9F,$C9,$27,$3E,$8F,$FA,$0F
	.dc.b	$27,$BA,$F8,$1F,$F2,$3B,$6F,$F9,$5C,$BE,$27,$35,$FE,$4C,$F1,$FF
	.dc.b	$CA,$D5,$FE,$CC,$A4,$63,$3F,$95,$18,$F8,$6F,$F2,$6C,$E6,$7F,$09
	.dc.b	$3C,$3E,$23,$FD,$1D,$F1,$AF,$B8,$FC,$36,$BE,$7C,$15,$A9,$CE,$5F
	.dc.b	$EC,$CB,$7D,$C6,$A3,$FA,$3C,$62,$BA,$CF,$97,$0F,$F5,$24,$E6,$BE
	.dc.b	$D7,$DC,$3E,$4B,$FC,$C6,$D8,$F7,$37,$4F,$6B,$5F,$CF,$57,$FB,$C7
	.dc.b	$7F,$5F,$C5,$FE,$0F,$1E,$EF,$CB,$C3,$CA,$E5,$F5,$7D,$C7,$FB,$39
	.dc.b	$F4,$B7,$0F,$C4,$FA,$5C,$5F,$EE,$18,$A7,$58,$7D,$DA,$D5,$7C,$E6
	.dc.b	$BF,$EE,$A6,$BF,$90,$6D,$3E,$DF,$FB,$BC,$FE,$8E,$31,$EC,$FF,$B2
	.dc.b	$E3,$DF,$3A,$6C,$65,$FA,$9F,$87,$BD,$BF,$BB,$5D,$31,$FA,$33,$9C
	.dc.b	$FE,$BC,$E5,$FB,$9F,$71,$FD,$49,$FB,$6D,$C3,$E0,$F1,$7D,$B9,$DF
	.dc.b	$AD,$F8,$A8,$C4,$B4,$FC,$4F,$3E,$78,$34,$FF,$8A,$A4,$7F,$8A,$C7
	.dc.b	$DB,$FF,$96,$9F,$BB,$8E,$A3,$11,$FC,$77,$AF,$1F,$11,$F1,$EA,$CF
	.dc.b	$F0,$B9,$AF,$E5,$E2,$BE,$08,$B8,$FF,$BD,$EF,$AC,$E4,$FF,$12,$DF
	.dc.b	$56,$9D,$4A,$75,$29,$C6,$39,$CF,$59,$FE,$3D,$3E,$F9,$CD,$F9,$23
	.dc.b	$13,$DE,$7C,$71,$83,$E1,$9F,$1A,$39,$DF,$F8,$02,$3D,$07,$F8,$09
	.dc.b	$C3,$F3,$4F,$E5,$DF,$6A,$F6,$9F,$71,$FE,$08,$8C,$36,$33,$3E,$89
	.dc.b	$CB,$7C,$6F,$F6,$6B,$78,$F9,$46,$23,$BB,$5F,$EA,$93,$FC,$E2,$7E
	.dc.b	$AB,$FC,$6A,$2F,$6F,$F5,$9C,$FE,$FC,$EE,$7E,$01,$F8,$87,$11,$A9
	.dc.b	$E1,$9F,$66,$D6,$F0,$F3,$D7,$47,$D3,$E3,$71,$9F,$C9,$B5,$FB,$B8
	.dc.b	$78,$63,$86,$3E,$69,$87,$90,$4F,$93,$1F,$7D,$F8,$1F,$24,$F0,$C3
	.dc.b	$9B,$EC,$D9,$F7,$0F,$8D,$FE,$F5,$97,$F2,$88,$C5,$7F,$2E,$6B,$FC
	.dc.b	$51,$7C,$6F,$0E,$1F,$E6,$D1,$AA,$FE,$63,$F7,$0F,$47,$58,$7E,$DE
	.dc.b	$30,$F5,$C7,$39,$3E,$AD,$CD,$86,$D7,$F4,$06,$BF,$C9,$97,$CB,$76
	.dc.b	$F1,$9F,$84,$3E,$67,$FC,$AF,$18,$6C,$66,$FF,$AF,$35,$FE,$BB,$C8
	.dc.b	$F9,$7E,$2D,$8F,$EC,$8D,$77,$1F,$E6,$4C,$DF,$73,$FE,$85,$9F,$EA
	.dc.b	$ED,$7F,$9E,$B7,$83,$FB,$C5,$F3,$7F,$AC,$FC,$5F,$F9,$26,$BF,$AF
	.dc.b	$B5,$FE,$8A,$D8,$DF,$D2,$FC,$1F,$B9,$FE,$F9,$5F,$DE,$5A,$FF,$78
	.dc.b	$6B,$FD,$D9,$AF,$F6,$FF,$B9,$7A,$9F,$51,$7C,$F5,$31,$1F,$63,$C8
	.dc.b	$7C,$2D,$25,$FF,$76,$87,$C2,$7D,$5F,$B9,$FE,$E5,$3F,$E0,$D6,$BF
	.dc.b	$E0,$C6,$BF,$E0,$A6,$BF,$CE,$BD,$EF,$E1,$7E,$6F,$E3,$1F,$E1,$BE
	.dc.b	$1D,$F1,$BF,$D9,$F5,$FE,$C8,$55,$F8,$5F,$5B,$FA,$C4,$FF,$8F,$5A
	.dc.b	$FF,$8E,$9B,$C2,$F0,$7C,$1F,$71,$FE,$2C,$AF,$F2,$27,$D2,$3E,$3B
	.dc.b	$F1,$AD,$8F,$F1,$43,$5F,$F1,$3B,$5F,$F1,$2F,$DC,$3E,$7F,$D5,$D6
	.dc.b	$F9,$8A,$7E,$6E,$75,$5F,$EE,$2D,$DF,$33,$FE,$B2,$E1,$FF,$76,$B6
	.dc.b	$33,$3F,$D0,$A7,$2F,$F8,$EF,$DF,$7C,$C7,$E9,$1B,$19,$8C,$1D,$7E
	.dc.b	$DB,$FC,$6B,$1C,$DA,$FF,$9F,$9A,$FF,$9F,$5A,$FD,$7F,$86,$FF,$77
	.dc.b	$3E,$36,$BF,$E8,$66,$BF,$E8,$56,$BF,$DE,$3C,$57,$DA,$FE,$4A,$3F
	.dc.b	$CA,$8D,$7F,$CA,$6D,$7F,$CA,$2D,$7F,$11,$F4,$CF,$9C,$F9,$0D,$8F
	.dc.b	$8C,$D7,$FD,$32,$D7,$FC,$B7,$C6,$FA,$AF,$E1,$6D,$8F,$F3,$33,$5F
	.dc.b	$F3,$2B,$5F,$F3,$1F,$ED,$7F,$DA,$71,$34,$91,$FC,$59,$FF,$0E,$3F
	.dc.b	$E7,$16,$BF,$E7,$06,$BD,$38,$F8,$5C,$8E,$0D,$78,$7A,$1D,$48,$F5
	.dc.b	$7C,$1F,$41,$23,$C0,$D7,$FD,$51,$D8,$FD,$97,$F9,$F7,$5F,$F8,$B2
	.dc.b	$31,$E8,$FF,$3E,$6D,$7F,$8C,$3B,$7F,$54,$B9,$9F,$2C,$FA,$FF,$E8
	.dc.b	$B2,$FB,$57,$D5,$FF,$83,$DB,$1F,$E1,$C3,$51,$93,$18,$F9,$7E,$A3
	.dc.b	$C5,$A9,$A7,$6B,$E7,$6C,$FC,$B9,$F9,$3E,$67,$F8,$66,$8F,$3A,$98
	.dc.b	$8B,$7E,$DD,$FF,$8A,$57,$EC,$23,$A4,$63,$BF,$71,$83,$E1,$4B,$E9
	.dc.b	$7B,$92,$3A,$1A,$FB,$0F,$E3,$DF,$67,$FE,$D9,$D3,$CE,$DC,$7F,$92
	.dc.b	$36,$BA,$DF,$C2,$F0,$64,$74,$E1,$FB,$51,$D3,$A3,$FB,$AF,$CC,$BF
	.dc.b	$5E,$F8,$B7,$FE,$FD,$8C,$72,$FD,$16,$C7,$F9,$39,$AF,$E1,$B9,$7F
	.dc.b	$93,$3E,$FE,$36,$9F,$A5,$DB,$D3,$D5,$F5,$BC,$AF,$6F,$5F,$BB,$AE
	.dc.b	$63,$F0,$B1,$E3,$38,$8E,$98,$BE,$C7,$E2,$99,$CB,$E2,$FC,$17,$C8
	.dc.b	$25,$73,$A7,$D8,$FC,$CF,$32,$4E,$5D,$BF,$5C,$7E,$52,$F9,$9F,$CB
	.dc.b	$B6,$3E,$B9,$A8,$FA,$FF,$29,$AF,$93,$FF,$34,$57,$F6,$EA,$FF,$49
	.dc.b	$B5,$EB,$1A,$FA,$4E,$A4,$8E,$B3,$FD,$EB,$4F,$C0,$FD,$17,$E1,$09
	.dc.b	$3E,$DE,$6F,$C3,$FB,$32,$3E,$03,$A7,$F2,$5D,$6F,$D9,$7F,$9D,$E1
	.dc.b	$FB,$68,$C7,$C1,$FE,$EE,$7F,$C4,$93,$97,$7D,$FF,$C0,$5C,$B7,$7E
	.dc.b	$07,$1A,$FD,$EF,$EC,$7F,$E1,$F9,$FF,$59,$E2,$8C,$3F,$EB,$66,$F3
	.dc.b	$FE,$23,$DD,$E3,$F5,$9F,$E9,$19,$34,$95,$FC,$16,$1F,$27,$E0,$24
	.dc.b	$7D,$68,$E2,$FE,$0F,$8B,$F5,$BC,$72,$30,$7E,$1D,$FD,$8E,$92,$30
	.dc.b	$7E,$2F,$4F,$DA,$FC,$9C,$8C,$1F,$8F,$7F,$73,$F2,$72,$30,$7E,$FB
	.dc.b	$D0,$FE,$87,$1F,$D0,$49,$1F,$D0,$5F,$EF,$DD,$5F,$E6,$0E,$2F,$E0
	.dc.b	$3F,$17,$FC,$10,$FF,$30,$7A,$2A,$FE,$0F,$F1,$53,$0F,$D0,$3F,$E2
	.dc.b	$7F,$B9,$FF,$7F,$61,$FC,$D3,$2F,$EA,$73,$FD,$44,$95,$FD,$45,$FF
	.dc.b	$34,$FE,$C2,$FB,$4A,$49,$FF,$67,$D2,$75,$1F,$E1,$1C,$2B,$FC,$0D
	.dc.b	$B8,$7F,$30,$6F,$3F,$F7,$FA,$B4,$BE,$9F,$19,$E2,$24,$78,$27,$FC
	.dc.b	$E4,$FE,$33,$A1,$83,$E7,$FE,$01,$27,$E8,$57,$F9,$BF,$1F,$07,$D8
	.dc.b	$7F,$A2,$60,$D3,$E4,$63,$6B,$F8,$A6,$1A,$6C,$E5,$FE,$7D,$97,$C3
	.dc.b	$F8,$3F,$92,$48,$F4,$3F,$E3,$9A,$FF,$08,$69,$F0,$FD,$CF,$AC,$78
	.dc.b	$E7,$EC,$12,$BC,$2F,$DB,$F2,$BE,$E7,$A1,$CC,$93,$F4,$5E,$DE,$C7
	.dc.b	$E0,$49,$1E,$A8,$FE,$F1,$FB,$2F,$EA,$23,$FA,$33,$5F,$E0,$ED,$C7
	.dc.b	$F8,$33,$FA,$A8,$CF,$C1,$7F,$72,$8F,$F3,$8C,$7F,$82,$1A,$FF,$00
	.dc.b	$F0,$D1,$88,$FF,$3E,$68,$FB,$4F,$CB,$7B,$B2,$7E,$7C,$57,$F1,$5F
	.dc.b	$2B,$F9,$B3,$19,$79,$31,$F5,$92,$7E,$BF,$D3,$7F,$85,$7D,$2F,$F3
	.dc.b	$FC,$78,$DA,$FF,$86,$DA,$FF,$A2,$1B,$D1,$F8,$CF,$A7,$FD,$57,$C4
	.dc.b	$4A,$FC,$1C,$7E,$B7,$EC,$7F,$8E,$F5,$7E,$4F,$7F,$F2,$FB,$63,$35
	.dc.b	$F0,$1A,$FF,$89,$65,$F6,$18,$C7,$F3,$6D,$8C,$C8,$FF,$16,$9C,$63
	.dc.b	$9B,$FE,$2E,$E9,$FF,$50,$BF,$CC,$0E,$1E,$F7,$E7,$B6,$FF,$16,$36
	.dc.b	$34,$71,$8C,$F5,$FF,$94,$23,$8B,$5F,$F1,$DB,$67,$FD,$D1,$EE,$7F
	.dc.b	$01,$F8,$9F,$F2,$0E,$BF,$0D,$B7,$F8,$06,$BF,$05,$BA,$7E,$EB,$E4
	.dc.b	$EE,$A6,$92,$FE,$11,$F3,$FE,$43,$FC,$9F,$C9,$FE,$3C,$F3,$FF,$D8
	.dc.b	$33,$FB,$89,$CB,$FE,$CB,$8C,$E5,$FD,$93,$F6,$87,$D7,$7C,$C8,$FF
	.dc.b	$2F,$46,$2B,$E2,$F0,$23,$13,$FE,$66,$6B,$F9,$F7,$FD,$C3,$3F,$E3
	.dc.b	$16,$C7,$F8,$96,$DD,$9E,$48,$FF,$6B,$CE,$B2,$F5,$7A,$39,$DB,$FB
	.dc.b	$0E,$14,$FE,$BF,$8F,$F8,$9D,$FE,$CA,$DF,$B2,$D5,$6D,$B1,$E3,$FB
	.dc.b	$0F,$76,$5E,$06,$E9,$ED,$6E,$66,$FC,$FF,$9E,$7E,$13,$65,$F4,$36
	.dc.b	$3C,$DF,$1B,$31,$F3,$7D,$35,$CB,$E1,$3E,$2D,$BF,$BE,$B7,$63,$4C
	.dc.b	$7D,$17,$C1,$3E,$16,$37,$0F,$85,$97,$C5,$3E,$67,$95,$FD,$57,$C6
	.dc.b	$C4,$FF,$42,$7E,$0C,$66,$3F,$A6,$46,$3C,$B3,$7C,$5F,$A1,$C3,$D2
	.dc.b	$7F,$8C,$57,$D1,$8C,$4C,$8F,$A9,$F3,$37,$7F,$C0,$7C,$CF,$D9,$39
	.dc.b	$78,$F6,$BF,$AA,$BE,$12,$3F,$D7,$DF,$89,$D3,$F8,$13,$C5,$E3,$D9
	.dc.b	$FE,$BA,$FA,$48,$FF,$67,$7D,$9E,$34,$8F,$ED,$4D,$C3,$F8,$37,$27
	.dc.b	$F8,$BB,$EC,$3F,$1C,$7E,$3F,$E4,$AD,$FE,$3F,$E4,$3F,$24,$7C,$3F
	.dc.b	$1E,$C7,$F6,$E7,$E5,$11,$FE,$F0,$FD,$DF,$91,$FC,$D7,$C8,$FE,$6D
	.dc.b	$B3,$F9,$BF,$90,$C3,$4F,$CF,$47,$E8,$7E,$47,$F4,$9F,$69,$FC,$0B
	.dc.b	$5F,$D6,$46,$1C,$F9,$3B,$BF,$F4,$65,$23,$13,$FE,$7F,$5F,$19,$FB
	.dc.b	$38,$D3,$63,$FD,$21,$5F,$E8,$07,$6D,$F8,$DF,$D5,$D7,$56,$CF,$08
	.dc.b	$FE,$5A,$D7,$F9,$A2,$F4,$7B,$1E,$CF,$8F,$B7,$FE,$B9,$FD,$9F,$FD
	.dc.b	$24,$DF,$82,$FD,$94,$63,$0F,$F3,$EA,$8C,$FC,$2E,$BE,$12,$DF,$37
	.dc.b	$E4,$FC,$84,$8F,$8B,$95,$C7,$FD,$75,$B7,$96,$72,$7C,$AD,$7E,$8D
	.dc.b	$7C,$7F,$A6,$74,$D3,$1E,$C5,$BF,$F5,$3C,$A3,$E3,$B6,$3F,$DF,$CD
	.dc.b	$C7,$FA,$9B,$5F,$EB,$4B,$AF,$C0,$7E,$0B,$8A,$D7,$DE,$F9,$1F,$C2
	.dc.b	$9F,$86,$F9,$1B,$1F,$82,$70,$FE,$13,$E4,$6E,$3F,$03,$E4,$3F,$1B
	.dc.b	$F2,$1F,$8E,$6D,$7E,$2B,$E4,$30,$D3,$F1,$B1,$F5,$37,$FC,$AF,$C7
	.dc.b	$D9,$FC,$8B,$FC,$FF,$E4,$DF,$6D,$FD,$0A,$7F,$DD,$8D,$CB,$FB,$C4
	.dc.b	$EB,$D9,$FE,$6E,$DA,$FE,$A7,$D4,$E7,$FC,$19,$FA,$4E,$9B,$3F,$A0
	.dc.b	$F9,$0F,$D3,$9F,$A8,$EA,$B1,$FA,$47,$0F,$E9,$BA,$EE,$7F,$82,$F9
	.dc.b	$0F,$D7,$FC,$6F,$EC,$1B,$5F,$AC,$F8,$FC,$34,$FD,$7C,$7E,$5B,$C0
	.dc.b	$91,$FE,$E8,$FB,$57,$C1,$63,$1F,$F7,$84,$62,$7F,$C2,$70,$8C,$3F
	.dc.b	$E1,$66,$BF,$E1,$D6,$BF,$E2,$5F,$71,$C1,$FC,$17,$DF,$FF,$6F,$36
	.dc.b	$FF,$C5,$5E,$6B,$C4,$7C,$7F,$EB,$31,$FE,$93,$FB,$67,$F7,$87,$FD
	.dc.b	$B3,$EC,$23,$11,$FE,$33,$8C,$7A,$7F,$D3,$27,$33,$FE,$9B,$7F,$C0
	.dc.b	$AE,$DA,$E7,$D2,$E0,$71,$F4,$A6,$7F,$EA,$EA,$46,$2B,$FC,$79,$87
	.dc.b	$8F,$D2,$F0,$12,$F8,$3E,$3F,$4A,$47,$36,$D7,$2C,$1E,$7F,$A4,$95
	.dc.b	$E1,$7F,$BB,$7C,$2F,$A3,$73,$49,$A7,$F6,$CF,$B5,$F6,$E8,$69,$23
	.dc.b	$FC,$01,$F8,$77,$C6,$B3,$49,$BF,$F8,$17,$E7,$7F,$80,$67,$FE,$56
	.dc.b	$A4,$63,$5F,$EB,$2D,$7F,$D8,$8D,$7E,$B7,$E4,$2F,$92,$FF,$00,$B6
	.dc.b	$FF,$DB,$6D,$9F,$9A,$D7,$FD,$9D,$F8,$57,$F8,$12,$7C,$4D,$8F,$ED
	.dc.b	$2D,$77,$EF,$FF,$83,$BE,$05,$E2,$7F,$8E,$FC,$77,$BB,$FE,$16,$9C
	.dc.b	$E3,$FE,$7D,$EC,$3F,$96,$AD,$B9,$3F,$21,$F6,$CF,$3F,$FC,$36,$D8
	.dc.b	$FF,$7A,$B7,$3F,$C3,$B5,$FC,$DF,$DB,$3E,$57,$FC,$4E,$DA,$ED,$2E
	.dc.b	$27,$AF,$F8,$69,$CF,$42,$F8,$5B,$4F,$4B,$C6,$5E,$7F,$F9,$6D,$F7
	.dc.b	$C6,$EF,$F6,$71,$A7,$C3,$8A,$37,$7D,$17,$C2,$FF,$5F,$AF,$F4,$CC
	.dc.b	$63,$CB,$FF,$1E,$CE,$70,$FD,$3B,$5F,$F5,$51,$FE,$03,$FB,$67,$C1
	.dc.b	$7F,$9C,$FC,$A4,$63,$1F,$87,$18,$F6,$B5,$39,$BF,$51,$FE,$0D,$FB
	.dc.b	$6F,$EE,$9A,$7C,$2B,$47,$8E,$30,$EB,$0F,$F5,$E4,$EB,$3F,$F2,$CF
	.dc.b	$DB,$3D,$BF,$F2,$BB,$6B,$DE,$DC,$7F,$81,$B5,$F6,$DB,$D0,$F9,$91
	.dc.b	$E6,$8C,$70,$FB,$E7,$E6,$DF,$89,$DE,$7E,$9F,$B6,$7C,$D7,$C5,$6D
	.dc.b	$7E,$0B,$E2,$F8,$E5,$FE,$2D,$9C,$BF,$13,$8F,$FC,$F7,$DB,$FE,$70
	.dc.b	$AF,$C0,$CF,$F2,$C6,$BF,$CA,$DA,$FE,$07,$F0,$8E,$6F,$F3,$26,$BF
	.dc.b	$CC,$5A,$EF,$1F,$F4,$EE,$AF,$B7,$FE,$93,$A7,$D3,$6E,$9F,$41,$FF
	.dc.b	$67,$3F,$79,$CD,$FF,$91,$FD,$5F,$B1,$3F,$5B,$ED,$BF,$BB,$5F,$DD
	.dc.b	$9C,$D7,$F9,$36,$73,$5F,$72,$73,$81,$A5,$8F,$C6,$BC,$AB,$F2,$BD
	.dc.b	$7F,$EF,$FF,$B6,$FF,$2A,$C7,$DC,$AC,$7C,$1F,$F8,$9A,$13,$A7,$F1
	.dc.b	$73,$A7,$FC,$52,$6A,$7F,$CA,$31,$98,$EE,$64,$EC,$53,$3E,$0F,$57
	.dc.b	$FC,$D5,$07,$FD,$B7,$3E,$39,$D6,$3F,$C0,$9E,$13,$FC,$28,$E8,$72
	.dc.b	$29,$9B,$E3,$7E,$58,$ED,$85,$BF,$E1,$B5,$D2,$31,$5B,$AE,$C7,$F9
	.dc.b	$30,$FF,$05,$98,$3E,$41,$FC,$08,$7D,$8F,$FB,$6B,$EB,$BF,$C2,$2D
	.dc.b	$8D,$1A,$F9,$14,$E9,$B1,$A9,$78,$BF,$C8,$E3,$0D,$7C,$11,$AD,$FB
	.dc.b	$CE,$67,$F8,$AC,$E5,$F9,$A7,$CF,$7F,$90,$BF,$47,$4F,$C0,$46,$23
	.dc.b	$67,$B2,$7C,$53,$A6,$EB,$DB,$C0,$FE,$47,$3C,$9B,$1E,$29,$DD,$78
	.dc.b	$5A,$FC,$89,$DB,$E5,$6B,$E1,$C3,$CD,$5C,$DB,$1B,$29,$CC,$B8,$7C
	.dc.b	$1C,$3D,$13,$8B,$EB,$3E,$F9,$CD,$7C,$6D,$9F,$6D,$FE,$87,$8F,$F5
	.dc.b	$8B,$5F,$F5,$8B,$5F,$E4,$ED,$7F,$99,$FD,$B7,$FA,$D5,$FF,$5C,$B5
	.dc.b	$FF,$5C,$B5,$FE,$5E,$D7,$F9,$DF,$E0,$BF,$EC,$22,$6F,$FE,$5C,$B4
	.dc.b	$7F,$B2,$3A,$5F,$6C,$F8,$0F,$F3,$0C,$61,$B1,$9C,$3F,$AF,$B5,$FF
	.dc.b	$21,$E8,$F9,$FE,$2D,$8F,$EC,$AD,$77,$1E,$3F,$B6,$7C,$CD,$24,$FF
	.dc.b	$B6,$23,$2F,$FB,$6B,$E7,$9F,$2B,$EB,$6C,$66,$30,$78,$BE,$96,$47
	.dc.b	$89,$AF,$FB,$8B,$ED,$7F,$C4,$AF,$FB,$99,$AE,$7C,$2F,$91,$18,$8F
	.dc.b	$EB,$BF,$40,$F6,$FE,$63,$63,$FD,$DE,$D7,$FD,$DE,$D7,$FC,$F7,$F6
	.dc.b	$AF,$90,$F9,$ED,$8F,$F0,$5B,$5F,$F3,$7B,$5F,$F4,$1F,$E1,$0F,$83
	.dc.b	$FA,$ED,$8F,$F0,$9B,$5F,$F3,$BB,$5F,$0F,$9D,$E0,$23,$E0,$7C,$7F
	.dc.b	$19,$23,$C6,$D7,$FD,$03,$F6,$AF,$6F,$F0,$ED,$8F,$F1,$1B,$5D,$C7
	.dc.b	$E3,$3F,$25,$7C,$F5,$24,$FF,$8A,$A3,$2F,$FA,$CB,$F1,$3F,$C4,$3F
	.dc.b	$C9,$DA,$FF,$8A,$9A,$FF,$43,$6F,$5B,$E0,$39,$CF,$F9,$B2,$FF,$3B
	.dc.b	$18,$8F,$F5,$6B,$5F,$F5,$67,$E2,$8F,$84,$FD,$1B,$63,$FC,$7E,$D7
	.dc.b	$FD,$46,$D7,$CD,$F5,$1F,$D0,$31,$34,$91,$FE,$48,$F8,$7F,$F5,$0B
	.dc.b	$7E,$27,$F6,$31,$88,$FF,$57,$B5,$FF,$61,$FD,$77,$F4,$CD,$0D,$24
	.dc.b	$7F,$50,$F0,$7F,$CD,$CF,$FB,$71,$AF,$F4,$C6,$BF,$D4,$DA,$FF,$B1
	.dc.b	$FB,$3F,$81,$FC,$5F,$EB,$A3,$FD,$71,$18,$33,$F4,$32,$3F,$D9,$0D
	.dc.b	$7B,$7C,$4F,$8C,$8F,$8D,$EB,$F4,$32,$3F,$DA,$0D,$7F,$B3,$7C,$9B
	.dc.b	$F6,$B7,$9D,$FE,$F6,$8C,$36,$33,$3F,$DA,$27,$2F,$FB,$F3,$EC,$DD
	.dc.b	$7F,$DD,$1B,$19,$8C,$1C,$E6,$47,$3A,$46,$23,$FB,$D7,$C3,$BF,$05
	.dc.b	$FF,$48,$C6,$1A,$FF,$A0,$23,$5E,$7F,$F7,$C8,$FF,$4F,$4E,$9F,$F0
	.dc.b	$E4,$F3,$FB,$1D,$36,$3F,$DA,$52,$9C,$B6,$33,$EF,$60,$8E,$0E,$3C
	.dc.b	$FC,$08,$E0,$F9,$1F,$B5,$BC,$BF,$9C,$D8,$CC,$60,$FF,$0E,$FE,$9E
	.dc.b	$FA,$3A,$49,$F7,$46,$5E,$BF,$39,$D0,$8E,$0F,$6F,$A3,$EC,$23,$83
	.dc.b	$E0,$F2,$FC,$02,$38,$3E,$2F,$C7,$FC,$42,$38,$3E,$4F,$A1,$F2,$08
	.dc.b	$E0,$FC,$37,$D0,$FE,$14,$8E,$0F,$C5,$7C,$FF,$E2,$48,$E0,$FC,$77
	.dc.b	$D3,$FE,$34,$8E,$0F,$B3,$E4,$FD,$82,$38,$3E,$EF,$D9,$FD,$C2,$38
	.dc.b	$3E,$AE,$1F,$50,$8E,$0F,$B3,$97,$D8,$23,$83,$F2,$1F,$B5,$7D,$B7
	.dc.b	$FC,$9A,$D7,$FC,$A8,$D7,$FD,$AE,$D7,$FC,$E0,$B5,$FF,$2E,$B6,$3A
	.dc.b	$C6,$0C,$FD,$54,$8F,$0B,$5F,$DE,$FC,$5F,$F9,$25,$FF,$72,$F5,$23
	.dc.b	$11,$FE,$7F,$8C,$7C,$07,$F0,$19,$CC,$FE,$D1,$FF,$03,$BD,$3D,$AD
	.dc.b	$11,$FD,$4F,$9D,$FA,$82,$38,$3F,$5B,$F6,$7F,$AC,$23,$83,$FD,$FF
	.dc.b	$F5,$8F,$A8,$FF,$13,$B6,$33,$18,$3F,$6B,$F2,$F2,$3F,$72,$D7,$F8
	.dc.b	$37,$CB,$FE,$E8,$8F,$EF,$9E,$BE,$F4,$8F,$0B,$5F,$F4,$2F,$D9,$7F
	.dc.b	$A7,$9F,$F5,$23,$53,$4C,$7D,$D3,$EA,$FF,$D2,$AD,$9F,$E2,$8D,$7F
	.dc.b	$D2,$2D,$7F,$D6,$03,$D1,$66,$F6,$BE,$19,$27,$E3,$BF,$C5,$BE,$B2
	.dc.b	$47,$F1,$A6,$BF,$C9,$FE,$B7,$F8,$D9,$1F,$E3,$EF,$CE,$F9,$29,$1F
	.dc.b	$49,$AF,$FA,$AF,$07,$D7,$BA,$FF,$CB,$6D,$9F,$F5,$5C,$EA,$72,$7F
	.dc.b	$2A,$FB,$49,$3F,$CB,$9A,$FF,$3B,$E9,$FE,$5E,$47,$F9,$9B,$FC,$CB
	.dc.b	$AA,$47,$F3,$66,$BF,$D0,$FC,$6F,$E6,$E4,$7F,$9D,$BF,$D6,$FE,$CB
	.dc.b	$FB,$74,$7E,$89,$AF,$E8,$5A,$FF,$52,$6F,$61,$FF,$03,$46,$FD,$AE
	.dc.b	$84,$7F,$A4,$FC,$2F,$F4,$82,$38,$3F,$D9,$FE,$7B,$F0,$BF,$EB,$F1
	.dc.b	$86,$BF,$D8,$1A,$FF,$5D,$6B,$E7,$F8,$4F,$61,$1F,$EE,$98,$7F,$94
	.dc.b	$DF,$67,$D4,$FC,$12,$3A,$7F,$B0,$D7,$F7,$C8,$F0,$7A,$F8,$23,$E5
	.dc.b	$FB,$5F,$CA,$FA,$BE,$36,$C7,$F6,$CA,$46,$21,$97,$BE,$30,$FC,$1F
	.dc.b	$B4,$B4,$69,$28,$7F,$7E,$FA,$BF,$F5,$16,$FF,$DF,$5B,$1F,$EF,$A6
	.dc.b	$BF,$F0,$05,$AF,$EB,$1F,$F2,$3F,$C8,$F1,$F5,$4D,$24,$FD,$1F,$B5
	.dc.b	$69,$E2,$3E,$93,$88,$FF,$09,$E7,$F6,$E3,$13,$A3,$84,$FC,$5F,$23
	.dc.b	$2D,$8C,$BF,$1B,$4F,$F5,$BC,$62,$FF,$0D,$5C,$2B,$5E,$4F,$F8,$6E
	.dc.b	$7F,$14,$DB,$FE,$79,$7F,$8B,$8F,$F1,$06,$8F,$2D,$EF,$61,$B5,$FE
	.dc.b	$3B,$6E,$3F,$E4,$46,$BF,$D2,$1F,$F6,$2C,$6F,$DF,$E4,$47,$F3,$5E
	.dc.b	$6F,$8C,$8E,$0F,$CF,$7C,$BF,$9C,$8E,$0F,$17,$3F,$94,$8E,$0F,$37
	.dc.b	$D6,$FA,$48,$E0,$FD,$57,$CC,$FC,$32,$38,$3F,$5D,$E3,$FC,$72,$38
	.dc.b	$3F,$65,$F5,$5F,$2C,$8E,$0F,$DB,$7C,$57,$CF,$23,$83,$F9,$0F,$8A
	.dc.b	$F5,$3F,$CA,$1E,$37,$F2,$86,$C6,$67,$71,$AC,$1F,$5D,$49,$3F,$D9
	.dc.b	$1F,$BD,$F2,$52,$3F,$84,$37,$8B,$F8,$46,$BF,$83,$24,$FF,$0D,$7C
	.dc.b	$7F,$B6,$FF,$97,$63,$F9,$BB,$67,$F9,$F3,$5F,$EA,$79,$A3,$11,$FD
	.dc.b	$5B,$F5,$0F,$F4,$3D,$7F,$9B,$DB,$1F,$D6,$8F,$DC,$B8,$70,$D7,$FA
	.dc.b	$FF,$E1,$5F,$C2,$27,$FC,$F0,$D8,$E4,$1C,$3D,$5A,$FF,$06,$BF,$37
	.dc.b	$C1,$FB,$89,$5E,$E7,$F4,$DE,$24,$8F,$E6,$0D,$AF,$E6,$1C,$7F,$AB
	.dc.b	$24,$FF,$35,$7E,$4F,$C0,$48,$F9,$2D,$9F,$E7,$1E,$47,$EC,$C9,$3F
	.dc.b	$CF,$5F,$DB,$74,$48,$FE,$80,$D9,$FF,$59,$FE,$98,$CB,$F7,$2D,$AF
	.dc.b	$F5,$83,$67,$73,$AF,$86,$95,$A6,$CF,$CA,$F0,$FF,$C5,$93,$FE,$A5
	.dc.b	$6C,$67,$DA,$FF,$53,$3F,$C9,$1B,$0F,$AF,$A4,$AE,$4E,$BE,$5E,$47
	.dc.b	$56,$DF,$FB,$3B,$EE,$5F,$01,$FC,$51,$B5,$FE,$CB,$6C,$EE,$75,$BC
	.dc.b	$AD,$36,$7F,$B3,$78,$5F,$C8,$49,$3F,$DA,$DF,$E4,$9F,$1B,$23,$FB
	.dc.b	$73,$67,$F1,$BF,$9A,$7F,$28,$6F,$8D,$F0,$A3,$11,$F9,$B6,$BF,$E3
	.dc.b	$67,$F9,$BB,$CB,$C9,$E0,$4B,$FF,$69,$4B,$CA,$E8,$48,$CB,$FC,$EF
	.dc.b	$CB,$FE,$87,$7F,$ED,$66,$BF,$DA,$9B,$E5,$9A,$7F,$5A,$AF,$F0,$6B
	.dc.b	$67,$FD,$D2,$D7,$FD,$CC,$D7,$B7,$C2,$78,$49,$3C,$9F,$27,$C2,$C8
	.dc.b	$EA,$D9,$FD,$F7,$98,$FD,$B5,$F7,$3F,$D4,$1B,$7F,$BA,$9D,$4E,$4F
	.dc.b	$C7,$FB,$9F,$D3,$AF,$FC,$C2,$D7,$FC,$02,$D4,$E6,$7C,$1F,$05,$DC
	.dc.b	$97,$DF,$04,$7C,$DF,$88,$93,$F1,$9B,$87,$C6,$E6,$FA,$44,$9F,$96
	.dc.b	$FD,$5F,$A4,$91,$F3,$9B,$3F,$3B,$EB,$FE,$D1,$27,$E9,$BF,$0F,$EB
	.dc.b	$A4,$79,$5B,$3F,$C7,$BE,$A1,$F1,$3F,$B2,$6D,$7F,$22,$94,$61,$E1
	.dc.b	$3F,$3F,$B6,$57,$C1,$6C,$FF,$3C,$F7,$FF,$A0,$92,$7F,$A0,$BF,$C0
	.dc.b	$3F,$30,$FE,$19,$3F,$CE,$5A,$FF,$A5,$1A,$FF,$8D,$1B,$E4,$FF,$45
	.dc.b	$F8,$0F,$E9,$A4,$BF,$E9,$AF,$F5,$0F,$80,$91,$FD,$41,$B7,$FC,$FA
	.dc.b	$FF,$A6,$BD,$7C,$5F,$21,$2B,$FD,$1B,$F2,$CF,$2F,$FB,$6B,$63,$31
	.dc.b	$83,$FB,$8F,$8B,$9B,$FF,$4D,$35,$FE,$BA,$D7,$FA,$53,$5F,$F7,$27
	.dc.b	$89,$FC,$93,$A0,$D2,$5F,$FA,$83,$F0,$47,$2F,$F8,$01,$B1,$98,$C1
	.dc.b	$FE,$04,$F9,$CF,$EB,$38,$7F,$AF,$DB,$C0,$69,$A8,$63,$FC,$06,$FF
	.dc.b	$08,$D8,$FD,$77,$7F,$9C,$FF,$08,$FC,$83,$E3,$E9,$2F,$FD,$6F,$19
	.dc.b	$7F,$D7,$3E,$EB,$A7,$FC,$32,$D8,$CC,$60,$FF,$0E,$FB,$CF,$93,$A4
	.dc.b	$9F,$F6,$1C,$65,$BF,$1D,$C0,$8E,$0F,$F6,$4F,$C5,$7F,$9C,$1A,$7F
	.dc.b	$8A,$9C,$69,$FE,$D3,$6B,$FE,$16,$AF,$DF,$CF,$83,$C1,$F2,$91,$FE
	.dc.b	$C5,$F6,$7E,$FA,$FF,$0B,$7A,$08,$C4,$7F,$BC,$A3,$1B,$7F,$9F,$E3
	.dc.b	$13,$A7,$FD,$2B,$2F,$B3,$FE,$3F,$1D,$1A,$F3,$6B,$C1,$AF,$FB,$FF
	.dc.b	$CE,$74,$7F,$92,$1B,$8F,$FB,$B1,$B3,$FE,$F3,$6B,$E7,$FB,$2F,$EF
	.dc.b	$6D,$F4,$FD,$0F,$9A,$32,$D8,$CB,$FE,$A2,$FB,$2F,$D8,$E4,$69,$3B
	.dc.b	$BC,$9C,$73,$2F,$23,$6F,$FC,$81,$D2,$FB,$2F,$AB,$1F,$E3,$56,$A8
	.dc.b	$D5,$7F,$99,$A7,$38,$BE,$C9,$F1,$FE,$68,$C3,$63,$35,$FE,$1D,$6E
	.dc.b	$1F,$73,$EC,$BF,$4F,$3F,$E2,$16,$C6,$7B,$3A,$46,$23,$F0,$1F,$24
	.dc.b	$F9,$4F,$03,$6F,$FC,$54,$D9,$FC,$4B,$5F,$E6,$DE,$5F,$FB,$63,$E4
	.dc.b	$7F,$81,$C7,$E2,$EB,$1E,$17,$F3,$F8,$4E,$9F,$C5,$CE,$9F,$F3,$55
	.dc.b	$57,$27,$0A,$E5,$5C,$63,$FC,$00,$D7,$FA,$39,$B3,$C4,$53,$FD,$DA
	.dc.b	$69,$E5,$AF,$FB,$75,$E5,$5E,$19,$EB,$5E,$03,$F5,$37,$98,$E5,$18
	.dc.b	$7B,$1F,$EF,$A9,$E5,$7D,$67,$E1,$4F,$C8,$3F,$AF,$1E,$83,$07,$FA
	.dc.b	$2F,$EA,$DF,$1A,$FB,$1F,$EC,$91,$86,$C6,$67,$FD,$13,$39,$6F,$84
	.dc.b	$FB,$1E,$2D,$8C,$C6,$08,$BE,$CB,$EB,$53,$C1,$F9,$31,$86,$E5,$F3
	.dc.b	$25,$87,$5A,$E2,$D9,$D5,$70,$B0,$BE,$CE,$2F,$8E,$16,$FB,$61,$E9
	.dc.b	$DB,$C5,$7F,$87,$C3,$8B,$F7,$DF,$1D,$E6,$BC,$53,$E0,$9D,$E1,$CA
	.dc.b	$3E,$06,$1D,$AF,$C1,$1B,$BE,$2F,$9F,$1F,$53,$4D,$9E,$B8,$EF,$A3
	.dc.b	$EC,$9F,$33,$F9,$2C,$1C,$1F,$CB,$27,$38,$7C,$49,$CA,$BE,$57,$C6
	.dc.b	$F8,$7B,$16,$B5,$F6,$4F,$5F,$18,$23,$13,$D6,$FF,$9D,$C6,$23,$F0
	.dc.b	$D1,$E3,$C3,$8F,$D9,$3B,$BB,$76,$BE,$9F,$FD,$F3,$87,$33,$F8,$F3
	.dc.b	$87,$99,$E9,$AE,$4E,$E7,$97,$C8,$7F,$8F,$DB,$1F,$EE,$26,$BF,$CB
	.dc.b	$AB,$A3,$FE,$F6,$E4,$FF,$56,$BB,$6B,$FC,$D1,$AE,$FE,$1B,$FD,$85
	.dc.b	$F6,$2F,$1B,$E2,$B6,$7D,$5B,$A3,$11,$EA,$C7,$F2,$D7,$F3,$3D,$87
	.dc.b	$C3,$DF,$87,$97,$CA,$72,$FB,$FF,$07,$78,$1E,$F6,$C7,$BD,$AF,$F4
	.dc.b	$C6,$BF,$DD,$7E,$CB,$FD,$1E,$F9,$9A,$FC,$ED,$11,$8E,$8F,$C4,$C6
	.dc.b	$3C,$2F,$B3,$F6,$4F,$2F,$F0,$BE,$EB,$E4,$E9,$A8,$DF,$99,$F5,$6B
	.dc.b	$FC,$4D,$9F,$D5,$6B,$FD,$89,$AF,$F2,$EE,$27,$D9,$3E,$43,$F1,$2D
	.dc.b	$9F,$C1,$35,$FC,$3F,$C5,$7F,$1F,$F9,$FF,$CE,$6E,$FD,$03,$FC,$28
	.dc.b	$FE,$93,$EF,$9F,$89,$BE,$9F,$F4,$51,$87,$8D,$70,$F0,$FF,$59,$F9
	.dc.b	$13,$D7,$FF,$47,$62,$8C,$69,$FD,$18,$DC,$EA,$33,$19,$9E,$1D,$3E
	.dc.b	$69,$FB,$31,$FB,$83,$33,$FA,$5C,$BF,$CF,$5F,$55,$FB,$7D,$BF,$B1
	.dc.b	$F6,$FF,$A7,$DB,$5E,$10,$E1,$F2,$35,$F9,$DF,$02,$ED,$FD,$53,$63
	.dc.b	$FC,$BF,$18,$3B,$7E,$12,$FB,$8A,$49,$F0,$5A,$75,$18,$FB,$DF,$F0
	.dc.b	$AB,$7D,$AF,$44,$62,$3C,$AD,$7F,$05,$F0,$FF,$ED,$47,$FD,$11,$F0
	.dc.b	$5F,$BF,$8C,$47,$F8,$A9,$AF,$F7,$6C,$1F,$84,$B6,$F8,$CD,$9F,$3B
	.dc.b	$5F,$8A,$D7,$F2,$5F,$6B,$FC,$A5,$FE,$18,$D7,$FC,$72,$D7,$7E,$1F
	.dc.b	$F1,$3F,$8C,$7C,$E5,$24,$FF,$AD,$1F,$E3,$1F,$34,$F9,$7A,$49,$FE
	.dc.b	$35,$19,$7F,$8E,$7E,$DD,$FE,$9C,$6F,$C1,$7F,$21,$8C,$47,$F9,$31
	.dc.b	$AF,$EA,$DF,$F1,$4B,$CB,$DE,$E0,$47,$AF,$83,$CC,$8E,$0F,$BB,$F1
	.dc.b	$BF,$E9,$97,$F9,$C5,$B0,$91,$86,$BF,$B4,$9E,$0F,$F3,$C1,$F5,$DF
	.dc.b	$E7,$A7,$FD,$F0,$D7,$F3,$4D,$7F,$58,$D7,$FC,$03,$5E,$1F,$9B,$FE
	.dc.b	$B4,$D8,$F1,$47,$F5,$D7,$BF,$C4,$FF,$52,$CD,$F5,$94,$93,$FD,$4F
	.dc.b	$91,$93,$E3,$FF,$60,$F9,$7C,$0F,$ED,$B3,$AC,$72,$72,$BF,$81,$F2
	.dc.b	$DD,$FD,$79,$1D,$70,$4E,$23,$E3,$79,$FD,$8C,$63,$FD,$EB,$6F,$9D
	.dc.b	$C3,$FC,$5B,$E4,$DF,$49,$4F,$83,$C6,$9C,$63,$FC,$0F,$C5,$FF,$79
	.dc.b	$D7,$F0,$58,$46,$1D,$5F,$47,$F8,$56,$1F,$77,$97,$A9,$23,$F8,$A4
	.dc.b	$ED,$FE,$BD,$C7,$FE,$50,$7F,$BB,$35,$CF,$C4,$7C,$AF,$4B,$E5,$F8
	.dc.b	$12,$7E,$23,$F5,$67,$E7,$FC,$E7,$D7,$FA,$89,$5F,$26,$2B,$F8,$DF
	.dc.b	$5F,$EA,$98,$E1,$FA,$BC,$8F,$B5,$FC,$17,$E0,$FE,$79,$FF,$6B,$37
	.dc.b	$D0,$FF,$25,$F5,$7E,$02,$B7,$9F,$F9,$BA,$8D,$26,$5F,$2E,$1F,$57
	.dc.b	$D8,$91,$FE,$1D,$11,$C9,$FC,$17,$A1,$F9,$3F,$A5,$91,$83,$E4,$BF
	.dc.b	$90,$F3,$64,$60,$FA,$3C,$BF,$83,$23,$FB,$4F,$B7,$FF,$21,$BF,$E2
	.dc.b	$AE,$74,$7F,$5F,$8D,$7A,$BF,$A8,$8C,$47,$F8,$C3,$A5,$D3,$FC,$22
	.dc.b	$3F,$75,$EA,$7E,$AF,$EB,$77,$C9,$FE,$FF,$6F,$27,$CA,$E5,$FE,$BF
	.dc.b	$39,$FA,$0F,$F4,$BD,$1A,$4C,$3F,$29,$0F,$9A,$EC,$48,$F9,$51,$C5
	.dc.b	$FC,$6F,$95,$FB,$B2,$3F,$AD,$F8,$7F,$EC,$DF,$85,$7F,$79,$8F,$EC
	.dc.b	$6D,$EA,$39,$F7,$9E,$0F,$F3,$8A,$FF,$27,$7E,$DF,$FD,$DF,$7F,$F4
	.dc.b	$7D,$23,$13,$FE,$F5,$6E,$1D,$3E,$2F,$C3,$F5,$12,$BF,$86,$64,$8F
	.dc.b	$F3,$1D,$F2,$C3,$A7,$A1,$FE,$5B,$8F,$E8,$17,$FE,$0B,$9C,$BF,$E5
	.dc.b	$D8,$CB,$FD,$7E,$FF,$8B,$BD,$B8,$BE,$09,$DA,$FE,$1F,$27,$83,$E3
	.dc.b	$E5,$79,$22,$BD,$F0,$FF,$3E,$3F,$E3,$36,$BF,$09,$AF,$FA,$CB,$ED
	.dc.b	$2A,$69,$2B,$FD,$AF,$F8,$6B,$C4,$FE,$68,$DA,$FD,$33,$5F,$E8,$D5
	.dc.b	$FE,$8D,$E7,$FF,$4D,$EB,$F4,$7E,$9B,$E3,$12,$BE,$23,$FE,$23,$DB
	.dc.b	$FC,$F7,$5F,$CD,$9B,$C5,$FB,$E6,$D7,$D9,$AF,$C3,$FB,$32,$3E,$BB
	.dc.b	$F8,$AF,$A1,$FA,$FE,$9C,$9F,$C2,$BB,$7F,$CF,$DF,$5F,$FE,$20,$7E
	.dc.b	$C3,$5F,$F2,$E3,$5F,$C5,$B5,$FC,$AF,$C0,$FE,$60,$8F,$FA,$7E,$3A
	.dc.b	$67,$EB,$FB,$0F,$CE,$FD,$B4,$AF,$C6,$C7,$E8,$3B,$BE,$11,$23,$FC
	.dc.b	$6B,$AF,$EB,$5B,$DF,$FA,$62,$3F,$96,$9F,$ED,$7E,$4B,$EA,$AF,$4C
	.dc.b	$AF,$F0,$0C,$E5,$BF,$59,$FD,$37,$C7,$FE,$3E,$D9,$34,$FA,$DE,$3F
	.dc.b	$ED,$89,$5F,$A6,$F9,$7F,$DC,$FC,$CC,$8F,$D4,$BF,$BC,$FB,$6F,$DE
	.dc.b	$FC,$37,$EF,$BE,$B2,$57,$EB,$A3,$D2,$F8,$48,$7F,$5C,$CB,$FB,$13
	.dc.b	$FC,$13,$E2,$7F,$83,$91,$FD,$BF,$C9,$7F,$9B,$FF,$12,$FF,$16,$F6
	.dc.b	$FF,$93,$5B,$3F,$D2,$1A,$FF,$9C,$1B,$E4,$1F,$E5,$53,$C7,$D0,$54
	.dc.b	$8C,$34,$E9,$1A,$7F,$DF,$67,$0F,$B6,$FE,$6B,$C5,$FD,$ED,$B1,$FE
	.dc.b	$81,$6B,$FE,$93,$6B,$CD,$FF,$0E,$FC,$3F,$1F,$27,$FC,$B5,$3D,$5A
	.dc.b	$A7,$6F,$FA,$2B,$1F,$F0,$76,$3E,$78,$F8,$1E,$F2,$31,$9F,$C1,$8C
	.dc.b	$4F,$AA,$31,$3A,$AF,$F1,$9B,$7E,$13,$6F,$CF,$6D,$7C,$F6,$BF,$3F
	.dc.b	$E1,$7F,$D5,$13,$F4,$DB,$1F,$4F,$E0,$6E,$1F,$4E,$73,$AB,$E6,$3D
	.dc.b	$57,$FE,$4C,$6A,$8D,$3F,$48,$D1,$4F,$C0,$7C,$2F,$6C,$DF,$41,$63
	.dc.b	$0E,$7E,$82,$69,$FE,$4F,$A7,$C1,$63,$0F,$F6,$DC,$3C,$6E,$87,$5F
	.dc.b	$7E,$DF,$F9,$CB,$E7,$7F,$D1,$EF,$F8,$E1,$B7,$FE,$D3,$D1,$18,$8F
	.dc.b	$C6,$46,$3D,$9F,$53,$AF,$AD,$F8,$B5,$FE,$F1,$BE,$D1,$8F,$97,$F7
	.dc.b	$7C,$1C,$9F,$57,$C4,$5C,$BD,$2F,$C5,$E4,$FD,$13,$76,$FF,$2C,$D7
	.dc.b	$FB,$21,$B5,$F8,$DF,$99,$F9,$5F,$09,$F2,$DB,$5F,$1F,$3F,$9C,$7D
	.dc.b	$EF,$57,$E3,$7C,$0E,$27,$F9,$43,$F5,$63,$31,$FE,$E5,$6E,$1F,$CD
	.dc.b	$A7,$5E,$DB,$E2,$BA,$36,$FE,$B6,$F8,$9F,$BC,$C8,$FB,$5C,$57,$0F
	.dc.b	$B3,$F2,$9F,$83,$39,$FC,$55,$AF,$F0,$6B,$E0,$23,$FD,$F5,$FC,$64
	.dc.b	$BF,$24,$FF,$12,$4F,$F9,$2D,$B1,$FE,$AB,$FB,$3F,$D4,$BF,$EA,$D8
	.dc.b	$CF,$AF,$FC,$22,$73,$3F,$C2,$C3,$87,$F8,$73,$5F,$E2,$AD,$DB,$FC
	.dc.b	$41,$F1,$9F,$18,$98,$7F,$84,$DF,$B9,$F3,$DF,$21,$FD,$13,$63,$FD
	.dc.b	$59,$F8,$57,$F1,$DC,$7F,$D5,$2D,$8F,$F0,$67,$C7,$7E,$DA,$31,$97
	.dc.b	$F8,$B6,$3E,$A9,$27,$FC,$44,$FE,$36,$5E,$4F,$13,$EC,$FC,$6D,$AF
	.dc.b	$F1,$DB,$87,$C7,$EA,$DC,$18,$F9,$3D,$6F,$11,$F8,$DF,$73,$F8,$E5
	.dc.b	$FA,$A3,$0C,$3D,$BE,$C5,$BF,$F6,$F6,$4F,$8E,$FE,$27,$1E,$66,$C7
	.dc.b	$FA,$56,$BF,$D0,$CE,$DB,$AF,$E0,$A3,$55,$F6,$4F,$7C,$64,$F8,$5F
	.dc.b	$0D,$D3,$E1,$BE,$53,$6F,$FD,$EF,$F9,$1B,$E1,$3F,$90,$B6,$3F,$A9
	.dc.b	$ED,$FC,$DB,$E3,$9F,$7C,$F7,$77,$87,$F2,$C9,$CC,$59,$D5,$F6,$F4
	.dc.b	$31,$3F,$27,$D6,$FB,$9F,$4F,$6B,$F0,$6F,$DF,$F8,$8C,$70,$7D,$29
	.dc.b	$FF,$59,$F4,$27,$3A,$FF,$62,$FB,$5F,$E3,$AD,$F0,$7E,$6C,$7F,$17
	.dc.b	$C3,$91,$FD,$B9,$DE,$1F,$DA,$DD,$47,$E8,$DA,$E7,$C2,$FD,$8C,$62
	.dc.b	$1E,$5B,$27,$D9,$78,$3C,$8F,$CC,$37,$1F,$96,$7F,$A2,$3B,$3E,$76
	.dc.b	$1E,$4C,$39,$7C,$4F,$E3,$79,$7F,$40,$FE,$87,$E5,$B1,$87,$EA,$1B
	.dc.b	$7C,$73,$7C,$B7,$E9,$CF,$CB,$7C,$55,$BF,$0D,$3C,$56,$8A,$DD,$7E
	.dc.b	$D3,$CA,$B7,$FE,$C7,$FC,$8D,$E2,$FF,$66,$6D,$74,$EC,$FC,$2F,$4E
	.dc.b	$6B,$F0,$4F,$F6,$38,$CC,$FE,$5D,$B7,$FC,$45,$AF,$F5,$2B,$FD,$03
	.dc.b	$FE,$D1,$EC,$5F,$87,$C2,$EB,$87,$B3,$5F,$5F,$C5,$7F,$07,$F9,$7F
	.dc.b	$DB,$36,$7F,$87,$F9,$98,$69,$C7,$83,$FD,$A4,$FF,$14,$F2,$BF,$88
	.dc.b	$1F,$BF,$F0,$2C,$6D,$C3,$FC,$6B,$C4,$BB,$7C,$7F,$17,$E3,$1F,$BE
	.dc.b	$FC,$31,$F0,$B8,$BF,$A7,$49,$CC,$7E,$EE,$31,$3F,$D0,$27,$31,$FB
	.dc.b	$F6,$BF,$E4,$3E,$CF,$C9,$5F,$D5,$CB,$F9,$43,$70,$F7,$7C,$47,$F6
	.dc.b	$1D,$0D,$24,$7F,$63,$8E,$B8,$7E,$5F,$2E,$04,$9D,$7C,$37,$32,$38
	.dc.b	$3F,$B6,$7E,$E1,$FE,$27,$6F,$B7,$FD,$C6,$31,$7F,$ED,$B6,$BF,$E5
	.dc.b	$DE,$37,$DD,$B0,$FE,$F0,$DC,$3F,$BC,$36,$BF,$DC,$8D,$79,$7C,$E3
	.dc.b	$D9,$FF,$00,$36,$BF,$C0,$0D,$9F,$F7,$63,$5F,$99,$F1,$0F,$77,$FC
	.dc.b	$30,$DA,$FE,$61,$C5,$FE,$63,$9C,$9F,$E0,$DF,$C0,$B5,$7F,$EF,$76
	.dc.b	$BE,$D6,$BF,$E8,$0D,$11,$88,$FE,$1B,$F8,$5B,$E0,$7F,$C5,$0D,$C7
	.dc.b	$C2,$D9,$FE,$F5,$8A,$31,$1F,$85,$FA,$2F,$CE,$BA,$6B,$9E,$BF,$F1
	.dc.b	$A4,$62,$3C,$9F,$82,$BE,$67,$FC,$50,$DC,$7F,$C5,$0D,$9F,$F0,$56
	.dc.b	$88,$C4,$7E,$57,$F0,$B7,$CC,$7F,$90,$5B,$7F,$29,$B3,$EA,$6B,$FE
	.dc.b	$9F,$F4,$DF,$70,$F5,$BF,$CF,$71,$87,$FA,$97,$E7,$AF,$3F,$35,$BC
	.dc.b	$7F,$90,$DF,$2F,$73,$FC,$EC,$D8,$FE,$D4,$DC,$7F,$C4,$8D,$7C,$55
	.dc.b	$FE,$D8,$9F,$E4,$1E,$07,$F9,$0E,$7F,$A6,$D7,$9B,$DE,$FE,$C9,$39
	.dc.b	$8F,$F4,$FE,$EF,$CF,$DF,$43,$FE,$57,$6E,$3F,$CD,$B9,$BF,$9B,$7A
	.dc.b	$19,$AF,$F2,$E7,$C7,$3C,$BF,$F2,$D3,$6F,$FC,$B4,$DD,$7E,$9B,$5F
	.dc.b	$D6,$3E,$38,$FF,$67,$FC,$03,$EF,$3F,$A1,$C7,$F2,$B9,$D5,$7F,$8A
	.dc.b	$A7,$2F,$F6,$AA,$7D,$E7,$7D,$BF,$D7,$71,$8C,$7F,$D5,$EE,$BC,$EB
	.dc.b	$5B,$CB,$F4,$1F,$85,$B9,$3F,$CE,$8D,$8F,$F3,$A3,$72,$FC,$6B,$5F
	.dc.b	$F7,$16,$0F,$8A,$FC,$05,$7F,$5E,$8C,$7B,$1E,$29,$CD,$7A,$5F,$EC
	.dc.b	$F5,$BE,$6F,$F1,$DC,$FB,$5B,$AF,$D8,$F9,$2F,$87,$8F,$E4,$AF,$FD
	.dc.b	$1D,$1F,$81,$BF,$E3,$0D,$7F,$78,$D7,$FB,$83,$79,$5F,$84,$FA,$6B
	.dc.b	$FD,$24,$DA,$FD,$5B,$5F,$E4,$8D,$7E,$CF,$DE,$BE,$73,$FD,$41,$3C
	.dc.b	$72,$FA,$73,$9B,$FE,$CD,$39,$7F,$03,$F0,$9F,$CA,$6B,$EC,$B6,$3F
	.dc.b	$0B,$EB,$B8,$FE,$DC,$E7,$2F,$F2,$C4,$7D,$E9,$D5,$7D,$15,$EA,$66
	.dc.b	$7F,$CB,$F4,$1F,$79,$7E,$EF,$F7,$91,$F7,$30,$C7,$85,$FE,$5D,$AB
	.dc.b	$39,$15,$C1,$E8,$74,$33,$3F,$E9,$DE,$15,$F2,$9D,$DD,$7F,$D0,$F1
	.dc.b	$5E,$5F,$39,$D5,$C8,$70,$EA,$FB,$4F,$B6,$73,$5D,$6F,$C3,$7E,$43
	.dc.b	$FD,$61,$19,$AF,$03,$63,$47,$98,$A7,$8B,$0E,$AF,$F8,$1C,$F8,$07
	.dc.b	$F8,$26,$30,$E0,$F8,$BF,$70,$F9,$87,$DE,$7F,$83,$E3,$0D,$8C,$CF
	.dc.b	$FA,$36,$72,$DF,$9A,$FF,$52,$B7,$87,$94,$62,$3F,$D9,$92,$8C,$47
	.dc.b	$89,$E3,$8F,$D9,$97,$7F,$DA,$8C,$37,$6E,$18,$7D,$9A,$E8,$D9,$D6
	.dc.b	$1F,$8D,$9E,$6E,$F0,$EF,$7C,$A3,$ED,$3F,$12,$BF,$13,$B7,$82,$7E
	.dc.b	$D4,$EE,$3C,$D5,$CA,$7B,$57,$81,$5F,$81,$FB,$2F,$99,$B3,$C5,$DC
	.dc.b	$F1,$9E,$18,$F6,$C7,$F3,$FB,$3E,$F3,$F7,$D8,$7F,$1D,$6B,$FC,$CD
	.dc.b	$B3,$FC,$C9,$7D,$0F,$F6,$9E,$0F,$E7,$F1,$A7,$FA,$0F,$DE,$7E,$A6
	.dc.b	$BF,$9C,$B6,$37,$D1,$C7,$0F,$F7,$5F,$DE,$3E,$AF,$1D,$6F,$92,$7D
	.dc.b	$66,$1B,$7F,$CC,$EF,$83,$FE,$77,$8C,$BF,$E0,$86,$BF,$E4,$16,$BF
	.dc.b	$E2,$56,$BF,$E9,$9E,$0F,$F6,$C3,$FE,$E1,$BF,$A3,$F1,$DF,$E6,$96
	.dc.b	$C6,$67,$6F,$F0,$3F,$BC,$FA,$DB,$7F,$58,$6E,$BF,$D9,$1B,$3D,$F0
	.dc.b	$F8,$5C,$2F,$D6,$FF,$C0,$B7,$FD,$A9,$B4,$9D,$3C,$23,$0F,$FA,$5F
	.dc.b	$EF,$7F,$AC,$3E,$96,$BF,$E2,$A6,$BF,$E6,$36,$BF,$DE,$3E,$F5,$E8
	.dc.b	$79,$17,$81,$4C,$47,$A3,$E0,$BF,$CC,$55,$FD,$95,$AF,$F7,$B6,$EB
	.dc.b	$FD,$E9,$7C,$A7,$DE,$FB,$DF,$E7,$75,$F3,$DA,$FF,$90,$1A,$FF,$9E
	.dc.b	$5A,$FF,$3F,$F4,$BF,$87,$FD,$47,$E2,$DF,$E2,$3F,$4F,$7C,$33,$67
	.dc.b	$F1,$BF,$A6,$FF,$2F,$BF,$F5,$F3,$5F,$F7,$DF,$32,$31,$A7,$F2,$EC
	.dc.b	$5F,$8B,$3E,$C7,$F2,$FF,$34,$F3,$BF,$0C,$DC,$BF,$14,$D7,$FC,$B6
	.dc.b	$D7,$FD,$29,$F8,$77,$FB,$9D,$FF,$1B,$73,$FE,$AD,$63,$F8,$1F,$03
	.dc.b	$F4,$B5,$B9,$FF,$7F,$78,$EF,$A2,$FF,$5F,$5F,$F4,$B6,$C6,$67,$FD
	.dc.b	$1A,$D7,$FC,$DB,$9F,$FA,$8D,$FD,$03,$5C,$F2,$FE,$86,$31,$1F,$EC
	.dc.b	$5F,$BC,$FC,$4B,$E6,$6E,$06,$98,$F8,$4F,$F6,$2C,$FF,$A7,$9A,$FF
	.dc.b	$A2,$7E,$4B,$D3,$18,$8F,$F5,$C7,$DE,$7E,$01,$72,$23,$0D,$F4,$7F
	.dc.b	$6D,$1B,$C5,$F7,$9F,$E8,$6B,$FF,$56,$B5,$FF,$4E,$B5,$FF,$70,$35
	.dc.b	$FF,$55,$7D,$E3,$E0,$BF,$81,$36,$3F,$83,$35,$FF,$4B,$B5,$FF,$63
	.dc.b	$FD,$4F,$F0,$9F,$90,$34,$91,$FC,$3F,$EE,$5F,$2B,$FC,$3D,$B1,$98
	.dc.b	$C1,$FC,$4F,$EC,$7F,$CB,$D1,$FC,$5D,$AF,$FA,$A1,$AF,$FB,$85,$AF
	.dc.b	$FB,$17,$E5,$BF,$40,$FF,$B9,$5A,$E7,$97,$F9,$04,$62,$3E,$8A,$FB
	.dc.b	$2C,$92,$3F,$DE,$AF,$F8,$27,$DE,$7D,$AF,$F3,$F9,$46,$1F,$F1,$C7
	.dc.b	$81,$FE,$1A,$73,$3F,$BB,$FA,$4F,$E2,$31,$AF,$37,$FC,$67,$FA,$37
	.dc.b	$E6,$63,$FC,$6A,$1C,$3A,$ED,$FF,$0D,$46,$23,$FC,$8D,$8F,$F3,$8C
	.dc.b	$7C,$0D,$BF,$F1,$FB,$67,$FC,$99,$5F,$A7,$AF,$F5,$7E,$6F,$17,$FA
	.dc.b	$BB,$6B,$CC,$7C,$68,$F5,$7D,$27,$AF,$E1,$24,$FF,$AC,$F4,$FE,$6B
	.dc.b	$87,$4B,$F8,$1F,$17,$D3,$C6,$97,$B8,$AA,$BF,$8A,$7C,$47,$C9,$6E
	.dc.b	$DA,$7F,$C8,$BC,$9F,$E8,$4F,$98,$FE,$E0,$DB,$F3,$B5,$E9,$5F,$CB
	.dc.b	$6F,$FB,$5D,$FF,$93,$BE,$91,$F6,$BF,$CF,$AF,$FC,$CB,$18,$F9,$CF
	.dc.b	$A7,$1F,$E7,$38,$C3,$86,$BF,$63,$1F,$F4,$BF,$CD,$7F,$39,$C7,$FB
	.dc.b	$DC,$F8,$8F,$E4,$2E,$31,$F5,$FC,$52,$BE,$8F,$C3,$CB,$FA,$3E,$2B
	.dc.b	$97,$FC,$A9,$1F,$EE,$1C,$5F,$47,$DF,$E6,$64,$E5,$E4,$F8,$7C,$5F
	.dc.b	$F6,$57,$BF,$F7,$B8,$E4,$F3,$8A,$EF,$0F,$C4,$57,$FB,$47,$DD,$CE
	.dc.b	$FF,$9D,$E3,$7D,$AF,$E8,$DB,$87,$E2,$0F,$9B,$1F,$0F,$E6,$75,$C1
	.dc.b	$FE,$73,$AF,$5D,$7C,$E6,$BF,$A2,$6B,$FE,$98,$7F,$CC,$CF,$56,$BF
	.dc.b	$34,$FB,$7F,$15,$D9,$F8,$DE,$87,$E7,$39,$E4,$60,$F9,$BC,$5F,$DF
	.dc.b	$3A,$3F,$0D,$F4,$FF,$64,$DC,$FD,$9E,$9F,$C3,$F8,$DF,$E3,$F6,$C7
	.dc.b	$E9,$9A,$FD,$FE,$AF,$F6,$CF,$8D,$F9,$6E,$BF,$E4,$33,$96,$F1,$7E
	.dc.b	$0A,$BA,$6E,$C7,$F3,$BF,$A7,$3E,$DE,$93,$1F,$C1,$7D,$27,$8F,$EF
	.dc.b	$1F,$4B,$F9,$89,$6C,$9C,$C7,$E6,$5A,$FF,$10,$8D,$4F,$1C,$D8,$BE
	.dc.b	$87,$C6,$4B,$F8,$18,$72,$D7,$F8,$5F,$CA,$FF,$0A,$24,$FE,$5F,$EE
	.dc.b	$7F,$49,$72,$3F,$34,$E9,$FF,$0B,$FE,$D1,$FE,$46,$F0,$CD,$3F,$C9
	.dc.b	$3D,$CE,$FF,$E6,$15,$F9,$7E,$1F,$D8,$11,$FD,$4D,$71,$FB,$3F,$F3
	.dc.b	$02,$69,$23,$F7,$F7,$D7,$EA,$FF,$74,$48,$E7,$3F,$BD,$F7,$E4,$7E
	.dc.b	$CD,$FE,$01,$F2,$9F,$E8,$0C,$1F,$67,$E0,$25,$7F,$9C,$E3,$2F,$D7
	.dc.b	$FB,$CF,$EB,$6F,$FC,$01,$6A,$69,$8F,$BD,$FF,$26,$F1,$7F,$4F,$6F
	.dc.b	$77,$FB,$F9,$B5,$CF,$7F,$27,$DA,$7C,$22,$47,$8D,$FF,$17,$C7,$F8
	.dc.b	$6A,$FF,$99,$FC,$0F,$F1,$B3,$8D,$7C,$CF,$86,$F7,$11,$FE,$29,$D5
	.dc.b	$5F,$E1,$5F,$09,$FC,$8C,$C5,$FF,$BF,$B9,$7E,$EF,$D3,$7D,$42,$57
	.dc.b	$C5,$7F,$C8,$3F,$8B,$7E,$9B,$6F,$F3,$33,$67,$FC,$96,$DA,$E7,$1F
	.dc.b	$D6,$39,$93,$FE,$60,$AE,$3E,$1F,$F9,$B6,$31,$5C,$BD,$CC,$7A,$BE
	.dc.b	$02,$47,$E1,$27,$FB,$23,$FD,$DB,$0E,$BF,$17,$CC,$91,$E1,$9D,$E9
	.dc.b	$FD,$63,$D0,$FB,$F3,$FC,$15,$B1,$9E,$7F,$EB,$71,$FE,$CD,$9F,$F1
	.dc.b	$4B,$F8,$4F,$0B,$FD,$95,$C9,$FB,$58,$C1,$F8,$8C,$BD,$FE,$2F,$B7
	.dc.b	$D6,$98,$78,$E2,$BF,$D0,$F3,$FF,$5C,$D7,$C6,$F9,$CC,$E9,$E7,$EE
	.dc.b	$F1,$92,$FD,$9E,$77,$9B,$E0,$FC,$44,$8F,$EC,$FB,$7F,$B7,$E3,$EC
	.dc.b	$FA,$1F,$3F,$D2,$95,$F1,$63,$E9,$7A,$E8,$9C,$F5,$FB,$B1,$63,$1A
	.dc.b	$FD,$C8,$C4,$FE,$11,$AF,$E1,$3D,$5B,$3F,$82,$6E,$E6,$9C,$25,$F5
	.dc.b	$AF,$5F,$F9,$BC,$78,$E3,$1A,$FF,$2D,$6C,$7F,$81,$9A,$FF,$8E,$9F
	.dc.b	$EE,$BF,$17,$BF,$67,$FD,$BB,$39,$6B,$D5,$AF,$E7,$3E,$F1,$AF,$C4
	.dc.b	$8F,$D0,$C6,$1F,$F2,$76,$9F,$9E,$8C,$4F,$C6,$D1,$18,$AF,$D4,$FD
	.dc.b	$DD,$7F,$52,$DB,$FD,$3B,$68,$DF,$63,$C1,$CB,$FE,$C9,$9F,$F7,$0E
	.dc.b	$3F,$9F,$9C,$BF,$B6,$F9,$2B,$8F,$FA,$0B,$E6,$5F,$45,$FE,$65,$8F
	.dc.b	$C0,$70,$27,$37,$F8,$A6,$CF,$F8,$B5,$AF,$E1,$67,$87,$D3,$F6,$6D
	.dc.b	$7E,$C7,$ED,$F1,$6A,$F1,$ED,$FC,$13,$D5,$BA,$FF,$A6,$B0,$FD,$C4
	.dc.b	$63,$E2,$B3,$F1,$1A,$8B,$3A,$3F,$C9,$F8,$BE,$9B,$D9,$AF,$BD,$BA
	.dc.b	$FB,$5B,$3B,$AF,$77,$DB,$79,$8F,$3F,$97,$63,$F8,$D7,$E3,$1F,$81
	.dc.b	$E4,$FF,$7C,$36,$3F,$51,$BA,$1D,$1E,$E7,$5A,$7F,$BC,$1C,$D7,$F5
	.dc.b	$E6,$C7,$F1,$B6,$BF,$AE,$6B,$D9,$FF,$64,$47,$93,$BF,$B7,$83,$DD
	.dc.b	$F0,$78,$D2,$7F,$B8,$FD,$65,$C3,$FB,$EB,$F1,$BC,$DF,$EF,$AD,$F5
	.dc.b	$4D,$31,$D3,$EB,$F2,$BF,$17,$5F,$E0,$48,$C6,$1F,$DF,$9C,$DF,$E2
	.dc.b	$37,$F8,$FF,$4F,$89,$F8,$9E,$A4,$BF,$F0,$2F,$DA,$DD,$3F,$C2,$2F
	.dc.b	$8C,$91,$FD,$DD,$FB,$B0,$FB,$B7,$C4,$FE,$7A,$30,$DF,$BB,$7A,$39
	.dc.b	$AD,$E1,$FE,$65,$9C,$BF,$E3,$76,$DF,$FA,$41,$AF,$FB,$11,$AF,$F2
	.dc.b	$7D,$BF,$C4,$BF,$3F,$F2,$49,$5F,$E0,$D7,$F2,$7C,$FF,$85,$F6,$3F
	.dc.b	$08,$7E,$17,$AA,$D7,$E3,$BA,$BF,$10,$7C,$AF,$8F,$B1,$FE,$2A,$7E
	.dc.b	$B1,$1F,$F0,$D3,$DB,$AB,$F3,$99,$FF,$A3,$DB,$C9,$FD,$FF,$95,$F1
	.dc.b	$BF,$95,$EE,$FA,$EC,$C7,$F3,$FD,$37,$1E,$5E,$27,$E6,$CF,$C0,$FE
	.dc.b	$E2,$FA,$8C,$5F,$FA,$9A,$30,$D8,$FD,$17,$57,$07,$F8,$37,$E7,$2F
	.dc.b	$6E,$9F,$C5,$BE,$27,$FC,$E1,$EF,$E3,$0F,$F3,$E7,$B2,$8F,$DE,$CE
	.dc.b	$A7,$FD,$85,$39,$6C,$FF,$00,$B7,$8F,$E0,$3F,$69,$F3,$76,$BF,$DA
	.dc.b	$FE,$17,$F9,$F5,$BF,$0D,$E3,$C5,$F1,$1F,$C7,$6B,$5C,$3F,$E9,$39
	.dc.b	$CD,$F9,$AB,$5F,$2D,$C5,$F8,$FE,$FF,$F0,$3F,$16,$D7,$F0,$C7,$F8
	.dc.b	$37,$81,$8C,$7E,$C3,$7A,$BF,$A7,$47,$F0,$C3,$F3,$DE,$5F,$F0,$C5
	.dc.b	$F7,$8F,$3F,$31,$FA,$38,$C6,$3F,$E9,$36,$C7,$FB,$B1,$AF,$F7,$F3
	.dc.b	$9C,$FE,$BA,$3F,$9A,$47,$FA,$F5,$B7,$FC,$8B,$E9,$3B,$7A,$3F,$84
	.dc.b	$F9,$3F,$E4,$1E,$15,$CB,$F9,$4B,$FC,$8F,$E5,$B1,$97,$E5,$5B,$8F
	.dc.b	$F2,$CF,$03,$F9,$41,$F8,$EE,$1B,$87,$E3,$5C,$3F,$E4,$DF,$C6,$FF
	.dc.b	$BA,$B9,$3E,$DF,$B7,$FD,$3A,$1F,$57,$F6,$63,$0F,$DD,$3F,$17,$CD
	.dc.b	$AA,$FC,$5C,$60,$FE,$93,$CF,$F9,$CD,$3F,$D9,$8D,$CF,$FC,$05,$E5
	.dc.b	$32,$C6,$1F,$DB,$2A,$DB,$CB,$78,$7F,$8A,$C7,$8E,$0D,$3F,$5B,$D7
	.dc.b	$E9,$7D,$5E,$F6,$33,$F8,$0D,$C9,$87,$2E,$BF,$80,$7E,$DB,$E3,$2D
	.dc.b	$FE,$D5,$C3,$F2,$7D,$1B,$7F,$EF,$9F,$E0,$1F,$E7,$13,$FE,$41,$6C
	.dc.b	$7F,$91,$7E,$F1,$F1,$7F,$CB,$E7,$32,$9C,$D7,$FB,$85,$B1,$F8,$D3
	.dc.b	$55,$FC,$EA,$33,$1F,$91,$AF,$F0,$35,$7F,$94,$F2,$FD,$0C,$62,$7D
	.dc.b	$F3,$FD,$C3,$67,$CC,$F8,$FC,$BF,$E2,$B9,$7F,$5D,$E1,$EB,$87,$F3
	.dc.b	$76,$BF,$4D,$BC,$5F,$E0,$9F,$17,$FB,$27,$9B,$FC,$23,$3F,$44,$E5
	.dc.b	$B1,$9B,$FD,$44,$FF,$84,$7C,$AE,$24,$AD,$BF,$DB,$FE,$57,$B9,$32
	.dc.b	$C1,$FD,$CF,$EF,$1F,$1F,$49,$7E,$28,$CB,$E4,$F9,$FF,$F2,$2B,$7E
	.dc.b	$17,$FB,$E4,$63,$3F,$2B,$5F,$F3,$CF,$C3,$BE,$27,$FA,$AB,$6F,$FD
	.dc.b	$E2,$D9,$FF,$57,$35,$FF,$42,$6A,$FB,$CC,$47,$F6,$36,$BF,$9F,$6C
	.dc.b	$6E,$FF,$B1,$FD,$E3,$E3,$69,$31,$FF,$80,$31,$97,$F1,$1E,$0B,$E9
	.dc.b	$7F,$97,$E5,$FE,$C1,$6B,$FE,$94,$6B,$FE,$34,$6B,$FB,$1F,$BC,$FD
	.dc.b	$75,$7D,$C6,$C7,$DA,$6B,$EE,$6B,$FE,$3E,$FB,$CF,$F0,$4B,$FE,$D1
	.dc.b	$6B,$FE,$A4,$6B,$BE,$EF,$F1,$77,$C8,$FF,$72,$BF,$F2,$3F,$37,$B6
	.dc.b	$BF,$36,$7F,$05,$9D,$2C,$BF,$B3,$69,$FE,$30,$FC,$65,$EF,$FF,$8E
	.dc.b	$5B,$8F,$E5,$9B,$7F,$94,$6B,$F6,$3F,$19,$FF,$56,$4F,$EB,$4F,$F0
	.dc.b	$D3,$88,$F1,$E0,$FB,$8F,$F4,$6C,$79,$B8,$95,$BF,$0B,$F9,$0C,$78
	.dc.b	$1B,$D9,$FE,$7E,$8C,$63,$FE,$66,$7D,$A3,$E5,$7F,$C2,$F8,$7E,$31
	.dc.b	$AF,$F8,$B5,$AF,$F2,$96,$FA,$CF,$BC,$7A,$7F,$E9,$E8,$C3,$61,$F4
	.dc.b	$76,$35,$F7,$9F,$3A,$8F,$F9,$83,$2F,$89,$8F,$F9,$EA,$34,$65,$F8
	.dc.b	$5F,$79,$FE,$72,$7F,$80,$B5,$FD,$2B,$5F,$F2,$33,$5F,$F4,$2F,$AA
	.dc.b	$F8,$8F,$43,$6F,$C3,$82,$1F,$11,$AA,$F7,$7E,$5F,$FE,$3A,$BD,$B7
	.dc.b	$0F,$AD,$F3,$7E,$D8,$C7,$17,$FB,$5F,$85,$3F,$31,$BC,$5F,$E8,$38
	.dc.b	$C1,$9F,$A6,$7B,$BF,$E8,$76,$E5,$FC,$91,$B9,$7F,$22,$6A,$26,$CF
	.dc.b	$AA,$7D,$1F,$FA,$2A,$B8,$E9,$FE,$EB,$AD,$4E,$71,$FB,$87,$D1,$9D
	.dc.b	$7C,$BF,$F2,$2C,$7F,$53,$7E,$6F,$8C,$FC,$14,$E6,$BA,$5F,$1C,$FC
	.dc.b	$6D,$C3,$FD,$6E,$FF,$4D,$9D,$DF,$FA,$2E,$7D,$5D,$5E,$BF,$BC,$7B
	.dc.b	$9F,$E2,$7C,$D9,$66,$3F,$AF,$55,$BF,$84,$7F,$AF,$AF,$8B,$FE,$2F
	.dc.b	$8D,$3F,$1A,$AB,$E0,$13,$C1,$D1,$8A,$EE,$FF,$9A,$DF,$24,$FD,$1F
	.dc.b	$0B,$84,$F3,$6E,$1E,$88,$E0,$78,$CE,$87,$A1,$D5,$F2,$8F,$E0,$86
	.dc.b	$8F,$1B,$9C,$3D,$33,$C3,$0F,$2C,$F8,$30,$F6,$DF,$23,$FC,$08,$53
	.dc.b	$D6,$62,$F3,$F3,$0F,$9F,$7D,$A6,$1B,$5F,$E0,$D9,$D4,$E4,$E3,$E2
	.dc.b	$FF,$89,$A7,$4D,$7F,$CC,$12,$8C,$47,$D8,$8C,$6D,$C8,$F9,$99,$E7
	.dc.b	$C0,$FB,$BB,$7D,$5A,$FE,$39,$7A,$3F,$44,$F2,$7E,$78,$71,$1C,$67
	.dc.b	$9D,$75,$AE,$46,$F0,$F5,$57,$76,$CE,$E7,$AD,$F8,$9B,$3A,$7C,$6D
	.dc.b	$7C,$57,$E0,$7C,$ED,$7C,$18,$78,$AB,$CF,$7D,$F0,$F4,$56,$70,$ED
	.dc.b	$B3,$EF,$3F,$D7,$13,$FE,$41,$8C,$7C,$77,$E0,$1B,$19,$9D,$BF,$AD
	.dc.b	$FB,$C7,$CC,$6D,$B1,$CD,$AE,$E3,$A7,$DE,$7F,$9A,$E3,$C0,$D7,$F7
	.dc.b	$0D,$7F,$3C,$D7,$FD,$F9,$F0,$35,$FF,$3C,$37,$B4,$D3,$1B,$7F,$94
	.dc.b	$A7,$FC,$8E,$D7,$FC,$8A,$D7,$FC,$52,$D7,$F9,$2F,$DE,$7E,$45,$F8
	.dc.b	$CD,$7F,$87,$35,$FD,$A3,$5F,$E5,$BF,$78,$F9,$EF,$98,$D8,$FA,$0D
	.dc.b	$7F,$8C,$35,$FF,$1D,$F9,$FF,$EC,$E6,$F4,$7E,$D3,$2F,$F0,$DC,$7F
	.dc.b	$3B,$6A,$DD,$F7,$6F,$8D,$FB,$8D,$9F,$C0,$35,$FE,$50,$D7,$F8,$77
	.dc.b	$BA,$FA,$2F,$C2,$36,$3F,$10,$D7,$F9,$83,$5F,$F2,$9F,$DD,$BE,$53
	.dc.b	$D0,$D8,$FF,$3C,$B5,$FF,$3B,$B5,$FF,$2D,$FE,$18,$F9,$2F,$80,$D8
	.dc.b	$FF,$40,$B5,$FF,$3F,$B5,$FF,$31,$FD,$E3,$EA,$BE,$43,$63,$FD,$12
	.dc.b	$D7,$FD,$0E,$D7,$FC,$D7,$E6,$FE,$DD,$BE,$07,$D1,$8C,$47,$D1,$6B
	.dc.b	$FE,$CB,$FB,$B7,$CA,$7D,$76,$C7,$FA,$61,$AF,$FA,$59,$AF,$F9,$EB
	.dc.b	$F1,$AF,$EE,$6F,$FA,$79,$AF,$FA,$71,$AE,$FC,$BF,$E6,$EB,$CC,$FD
	.dc.b	$33,$7E,$4B,$31,$83,$F5,$1E,$FF,$29,$FC,$53,$5F,$F3,$A5,$7E,$3E
	.dc.b	$31,$CE,$FC,$43,$EF,$57,$E4,$5A,$FF,$BA,$DB,$1B,$E5,$F0,$FC,$FC
	.dc.b	$AF,$E9,$B8,$A3,$11,$FE,$B8,$FC,$97,$F3,$4F,$FA,$F9,$AE,$7A,$BF
	.dc.b	$45,$18,$8F,$F5,$0F,$DD,$BE,$AB,$F4,$CD,$8F,$F6,$3B,$5F,$F6,$2B
	.dc.b	$5F,$F0,$18,$FC,$27,$F8,$5B,$FD,$89,$AF,$FB,$9D,$AF,$06,$BF,$B3
	.dc.b	$FB,$C7,$99,$FB,$76,$C7,$FB,$61,$AF,$FB,$59,$AF,$FA,$EB,$EE,$AB
	.dc.b	$FB,$D6,$AD,$91,$86,$D7,$E0,$AD,$F3,$54,$97,$EB,$7F,$85,$7B,$4F
	.dc.b	$B3,$A4,$9F,$F7,$5B,$F8,$CF,$C2,$DF,$5F,$49,$3F,$E0,$68,$CB,$76
	.dc.b	$E0,$47,$06,$FC,$9E,$84,$70,$7F,$B3,$3E,$6B,$FD,$12,$DF,$84,$FF
	.dc.b	$09,$FC,$42,$73,$3F,$62,$31,$3A,$9F,$F4,$6C,$F3,$F8,$6C,$11,$FF
	.dc.b	$6C,$FE,$37,$C9,$FF,$21,$DA,$31,$1F,$E6,$78,$C5,$F8,$A7,$33,$FE
	.dc.b	$C3,$7B,$BE,$36,$BD,$1A,$F3,$6A,$F9,$87,$E3,$6F,$99,$FF,$18,$C6
	.dc.b	$1B,$19,$9F,$CB,$9C,$5C,$3F,$9A,$FA,$A7,$D6,$BE,$17,$FC,$77,$18
	.dc.b	$6C,$66,$79,$CE,$5E,$B8,$79,$34,$F0,$92,$33,$87,$AA,$7D,$3F,$3D
	.dc.b	$E6,$24,$E9,$F4,$FB,$12,$38,$B5,$FB,$5B,$FA,$9F,$A6,$D7,$E9,$35
	.dc.b	$FF,$79,$37,$7F,$C2,$79,$1F,$81,$24,$FE,$2D,$FE,$4B,$F7,$6F,$9D
	.dc.b	$A4,$9F,$F2,$FC,$65,$BE,$FF,$02,$38,$37,$F4,$5D,$08,$E0,$FC,$D7
	.dc.b	$DD,$FE,$4C,$8E,$0F,$CF,$7D,$5F,$E5,$C8,$E0,$FA,$3F,$1B,$FD,$33
	.dc.b	$7F,$F6,$2F,$A3,$F8,$48,$C1,$3E,$9F,$04,$AE,$2F,$E2,$BF,$4A,$7B
	.dc.b	$54,$93,$F7,$A3,$2F,$C2,$F0,$BE,$39,$1C,$1B,$DF,$A1,$1C,$1F,$AC
	.dc.b	$EA,$FD,$31,$1C,$1F,$B0,$E1,$FD,$51,$1C,$1F,$C0,$7E,$9F,$F7,$44
	.dc.b	$70,$7F,$05,$F4,$7F,$7C,$47,$07,$E9,$3C,$77,$8A,$F9,$EF,$E3,$E4
	.dc.b	$9D,$46,$5F,$E2,$1F,$79,$FC,$24,$8E,$0F,$E2,$9C,$BF,$C3,$48,$E0
	.dc.b	$E7,$F7,$3F,$CB,$08,$E0,$FF,$55,$FD,$E7,$F4,$F7,$FD,$A2,$D7,$FD
	.dc.b	$E8,$D4,$D3,$FC,$88,$F0,$F1,$7F,$BF,$C6,$23,$8C,$7F,$3C,$FB,$A9
	.dc.b	$1E,$E6,$BF,$AE,$EA,$FE,$4C,$47,$F9,$83,$FC,$A7,$86,$47,$F3,$36
	.dc.b	$BF,$D9,$B5,$F9,$BE,$A7,$F4,$A2,$47,$CC,$7F,$CA,$DF,$69,$F4,$A3
	.dc.b	$EA,$35,$E4,$D7,$FD,$E7,$28,$C4,$7F,$00,$E6,$FE,$78,$47,$FA,$33
	.dc.b	$FC,$9B,$F1,$BF,$E3,$51,$FD,$EB,$AD,$18,$8F,$F7,$FC,$63,$C7,$E0
	.dc.b	$71,$C6,$4F,$F2,$0A,$FF,$0D,$D7,$29,$FE,$A3,$F7,$1F,$CD,$EB,$F9
	.dc.b	$8F,$82,$8C,$47,$F5,$36,$BF,$D3,$1A,$8F,$E8,$16,$FA,$9F,$51,$23
	.dc.b	$C4,$FB,$77,$91,$E5,$6B,$DF,$D1,$F2,$91,$F2,$DF,$F8,$23,$07,$E3
	.dc.b	$74,$95,$FD,$B1,$FF,$06,$E7,$FE,$E1,$FB,$4F,$EA,$3F,$1A,$69,$31
	.dc.b	$FE,$8B,$1F,$E7,$77,$D9,$3F,$EE,$9F,$87,$E2,$48,$F1,$FA,$9F,$28
	.dc.b	$8E,$0F,$97,$87,$F9,$CF,$E2,$BF,$B9,$B6,$3F,$B9,$35,$67,$EC,$F8
	.dc.b	$5F,$99,$0A,$FA,$11,$86,$C7,$D1,$6B,$B3,$FB,$F6,$BD,$57,$D5,$7F
	.dc.b	$8A,$B3,$FE,$DF,$18,$F1,$3E,$DB,$63,$33,$B8,$FE,$EA,$BE,$0B,$FC
	.dc.b	$AD,$1F,$81,$8C,$7B,$DC,$27,$33,$E0,$77,$9F,$E3,$71,$F3,$65,$F8
	.dc.b	$B6,$D7,$A1,$AF,$C4,$7F,$29,$1E,$F8,$F9,$31,$A9,$FC,$1B,$74,$DC
	.dc.b	$FC,$33,$E6,$B8,$AF,$CD,$79,$1F,$ED,$39,$EC,$DB,$FD,$13,$6C,$EF
	.dc.b	$37,$E0,$EE,$4C,$37,$1F,$F0,$7B,$71,$FD,$D3,$5F,$DC,$BF,$A7,$8D
	.dc.b	$F2,$F2,$23,$F8,$AF,$2B,$C9,$A1,$A4,$8F,$D2,$BE,$72,$38,$3F,$C4
	.dc.b	$3E,$77,$FB,$45,$BF,$13,$EB,$8C,$63,$FC,$3D,$AF,$F0,$E7,$FD,$D1
	.dc.b	$3C,$BA,$B8,$11,$FD,$AF,$9B,$F1,$C8,$E0,$F0,$FD,$EF,$72,$38,$3C
	.dc.b	$B9,$F8,$C8,$E0,$FE,$03,$F2,$BF,$4C,$8E,$0F,$E6,$7F,$61,$F6,$FC
	.dc.b	$2F,$E6,$0D,$F0,$8D,$35,$E5,$3E,$DA,$92,$FF,$B5,$BF,$81,$F9,$B9
	.dc.b	$1F,$C4,$9B,$BF,$F2,$5F,$AA,$FC,$31,$27,$F8,$BB,$E3,$FD,$CD,$D9
	.dc.b	$49,$3F,$DF,$63,$2F,$ED,$3E,$EF,$EF,$37,$97,$F2,$D8,$27,$2F,$F9
	.dc.b	$A1,$AF,$FB,$01,$FC,$BC,$F3,$EA,$E3,$18,$9E,$11,$E5,$F2,$E5,$78
	.dc.b	$DB,$3F,$D0,$7D,$0F,$D1,$12,$7F,$99,$BF,$A5,$FB,$F9,$1F,$CD,$9B
	.dc.b	$3F,$DB,$72,$FF,$57,$CF,$BF,$E1,$3E,$21,$1F,$E7,$BE,$87,$EB,$C8
	.dc.b	$E0,$FE,$83,$CD,$FB,$32,$38,$3F,$97,$F4,$FE,$98,$8E,$0F,$E6,$7E
	.dc.b	$67,$EA,$88,$E0,$FF,$42,$FE,$34,$F5,$BF,$7C,$D8,$CC,$60,$FA,$7E
	.dc.b	$27,$F2,$38,$FF,$5D,$F8,$B8,$23,$D3,$DF,$E6,$47,$07,$83,$E2,$FB
	.dc.b	$91,$C1,$E4,$FB,$8F,$19,$1C,$1F,$D9,$3C,$6F,$E2,$C4,$70,$7F,$68
	.dc.b	$F0,$7F,$8D,$11,$C1,$FD,$0B,$A5,$F4,$9F,$8E,$6C,$66,$30,$67,$BA
	.dc.b	$47,$16,$F2,$FF,$AE,$A9,$F5,$FC,$C9,$5D,$DF,$E5,$9C,$52,$3F,$BE
	.dc.b	$B6,$BF,$C2,$1C,$5F,$CC,$49,$3F,$E0,$27,$F9,$97,$DF,$BE,$0A,$92
	.dc.b	$7F,$D7,$F1,$96,$F9,$DC,$08,$E0,$DF,$07,$42,$38,$3F,$D9,$7E,$6F
	.dc.b	$FA,$41,$BE,$B7,$F4,$9F,$85,$FF,$14,$36,$75,$9F,$F8,$E7,$17,$1F
	.dc.b	$A0,$98,$79,$9F,$E7,$BE,$64,$8F,$F0,$4B,$6B,$FC,$06,$7F,$5A,$F1
	.dc.b	$BF,$A1,$92,$7F,$C1,$EF,$F5,$79,$91,$FE,$28,$6C,$FF,$8E,$E7,$FA
	.dc.b	$E1,$27,$FC,$5A,$FF,$0E,$FB,$BF,$E1,$73,$FE,$12,$6B,$FE,$3F,$6A
	.dc.b	$69,$FC,$35,$F3,$B5,$E5,$E1,$F5,$3B,$65,$2F,$FC,$27,$5F,$DE,$2B
	.dc.b	$8C,$FF,$29,$FB,$E9,$5F,$C9,$DB,$3F,$10,$7E,$33,$FB,$37,$0D,$7E
	.dc.b	$4B,$5F,$F1,$6B,$57,$99,$FE,$AE,$F4,$BF,$A2,$92,$FF,$D5,$AF,$F4
	.dc.b	$DF,$82,$FF,$78,$47,$FA,$D5,$B5,$FE,$8B,$6B,$FE,$82,$6B,$E0,$E2
	.dc.b	$FE,$77,$81,$A4,$AF,$F6,$07,$DC,$3C,$4F,$EB,$8D,$8C,$C6,$0F,$EC
	.dc.b	$1D,$3F,$EC,$58,$FF,$64,$37,$0F,$F4,$F3,$5F,$F4,$AB,$5F,$F4,$5F
	.dc.b	$3F,$F4,$EC,$FF,$51,$1F,$35,$AF,$FA,$4E,$3F,$B7,$7E,$11,$FD,$62
	.dc.b	$3F,$D5,$4D,$7F,$D4,$2D,$7F,$D0,$DA,$A3,$11,$F4,$3F,$38,$7C,$07
	.dc.b	$F7,$96,$E5,$FE,$E1,$6C,$FF,$AD,$DA,$FF,$A9,$7A,$BD,$04,$9F,$F2
	.dc.b	$EE,$48,$FF,$02,$7C,$73,$E8,$69,$2B,$FD,$DD,$19,$7F,$DE,$1E,$AF
	.dc.b	$F1,$A6,$ED,$FE,$0D,$8C,$63,$FE,$C9,$6B,$FE,$C0,$FB,$67,$D8,$FF
	.dc.b	$84,$DB,$7F,$EF,$B6,$CF,$FB,$45,$AF,$FB,$23,$ED,$BF,$29,$C0,$69
	.dc.b	$3C,$9F,$F0,$F7,$E1,$AF,$A5,$A4,$C7,$F4,$96,$9D,$46,$3D,$B7,$D0
	.dc.b	$FF,$89,$DB,$19,$8C,$1F,$E6,$5F,$21,$F6,$EF,$93,$C3,$6B,$FA,$03
	.dc.b	$75,$FE,$2F,$5C,$2B,$97,$DB,$FC,$69,$FF,$64,$4A,$31,$1F,$E2,$66
	.dc.b	$C6,$78,$BF,$C7,$9F,$6F,$FE,$81,$AF,$F7,$6B,$5F,$C6,$B5,$FD,$DB
	.dc.b	$65,$F6,$EF,$AD,$F6,$37,$3F,$EB,$0D,$BF,$4B,$5F,$F5,$9D,$3E,$DF
	.dc.b	$9D,$7F,$5F,$6C,$FC,$06,$BC,$1A,$FE,$8B,$DF,$F7,$F0,$FF,$9E,$F3
	.dc.b	$FF,$2C,$46,$0C,$FD,$1B,$E2,$FF,$CB,$8D,$C7,$EE,$B7,$0F,$E8,$B8
	.dc.b	$23,$11,$E8,$FB,$77,$CB,$7D,$B6,$DF,$F7,$76,$CF,$D8,$6B,$F0,$3D
	.dc.b	$3F,$D7,$BF,$61,$AF,$F9,$71,$BA,$9B,$F6,$EF,$5B,$F1,$4D,$C3,$FC
	.dc.b	$0C,$DB,$FC,$3B,$5F,$F6,$AF,$DB,$17,$C9,$3E,$D3,$F1,$B7,$98,$C4
	.dc.b	$FF,$66,$AC,$56,$A3,$8B,$FD,$1E,$33,$1F,$72,$AB,$FE,$F0,$8A,$FF
	.dc.b	$1D,$AE,$AE,$4E,$67,$12,$99,$76,$F1,$AF,$03,$A3,$07,$9C,$C6,$5F
	.dc.b	$EF,$99,$E3,$0A,$F1,$CE,$6B,$9D,$76,$8C,$57,$33,$E4,$9F,$EF,$E3
	.dc.b	$91,$6B,$FE,$00,$FC,$93,$E2,$9F,$6D,$86,$D7,$F6,$D9,$D4,$E4,$E3
	.dc.b	$E2,$3E,$DA,$92,$B8,$46,$5B,$1F,$6F,$ED,$FE,$5D,$BC,$0F,$BD,$49
	.dc.b	$CC,$61,$DD,$F3,$6D,$F3,$D7,$8E,$1C,$F0,$C1,$F8,$97,$0F,$58,$DD
	.dc.b	$F3,$6D,$7E,$0A,$B7,$AF,$8A,$F1,$7E,$B8,$F0,$C7,$33,$B6,$3E,$7C
	.dc.b	$3C,$6D,$8F,$15,$77,$6E,$1E,$65,$87,$9B,$1F,$04,$75,$33,$5C,$75
	.dc.b	$7D,$BB,$E6,$7E,$4E,$2E,$0F,$E5,$B3,$9C,$3F,$55,$39,$7F,$DA,$71
	.dc.b	$97,$ED,$4F,$0F,$7B,$9B,$FC,$DB,$ED,$FF,$A3,$CF,$28,$C7,$89,$A8
	.dc.b	$FF,$6E,$CE,$9E,$67,$6C,$75,$93,$ED,$FF,$BA,$4F,$33,$FC,$63,$E9
	.dc.b	$FF,$9E,$DC,$76,$7F,$9F,$63,$8C,$F0,$9E,$B3,$97,$FA,$FC,$F1,$7F
	.dc.b	$91,$B5,$FC,$6C,$6D,$FE,$4F,$C0,$F3,$7F,$C9,$51,$FE,$42,$8C,$3F
	.dc.b	$DA,$9A,$F0,$8F,$F5,$47,$DB,$FE,$A6,$3D,$58,$A3,$11,$F0,$A3,$1C
	.dc.b	$FF,$1F,$0F,$BF,$7F,$13,$ED,$DF,$53,$7E,$3A,$63,$D2,$72,$FC,$4F
	.dc.b	$4B,$FC,$90,$FF,$7E,$6B,$FD,$09,$AF,$C0,$6B,$FD,$C3,$ED,$FF,$B1
	.dc.b	$BE,$86,$BE,$76,$F3,$7E,$1F,$44,$3F,$00,$F5,$7D,$C3,$C0,$FB,$BD
	.dc.b	$1F,$D0,$71,$F8,$6D,$7E,$BC,$ED,$FB,$AF,$F8,$65,$FF,$0B,$B5,$FE
	.dc.b	$B6,$D7,$EC,$B5,$FF,$25,$FD,$C3,$E1,$7E,$63,$63,$F1,$2D,$7F,$27
	.dc.b	$F1,$FF,$C7,$BE,$1B,$E9,$7D,$C3,$E1,$BF,$35,$2F,$C3,$1F,$23,$86
	.dc.b	$E1,$F9,$4B,$E0,$FE,$5B,$BD,$E1,$7E,$2B,$A7,$FC,$5A,$D8,$CC,$FF
	.dc.b	$8D,$1A,$FF,$78,$FB,$97,$D6,$7E,$15,$B1,$FA,$26,$BE,$AE,$57,$D4
	.dc.b	$7A,$3F,$32,$FE,$7D,$B1,$A4,$AF,$27,$E8,$6F,$84,$A4,$8F,$F2,$3C
	.dc.b	$65,$FF,$56,$53,$F0,$EF,$C0,$46,$1A,$FF,$B5,$23,$5D,$3F,$AA,$6C
	.dc.b	$72,$FB,$EF,$E1,$CF,$F3,$E8,$46,$23,$FB,$F4,$63,$C7,$E5,$39,$9F
	.dc.b	$F5,$AF,$C4,$FF,$88,$DF,$F2,$EB,$5F,$F4,$8F,$AF,$F4,$63,$11,$DF
	.dc.b	$F0,$EF,$DC,$B7,$C5,$F3,$46,$23,$C2,$D7,$FD,$87,$F7,$4F,$17,$EE
	.dc.b	$36,$3F,$CD,$AD,$7F,$C6,$2D,$7F,$87,$FC,$6F,$E4,$5B,$DF,$F8,$58
	.dc.b	$C4,$7F,$A9,$1A,$FF,$B4,$BE,$E9,$F5,$9F,$8B,$6C,$7F,$9F,$1A,$FF
	.dc.b	$90,$1A,$FF,$1E,$FB,$AA,$FF,$AC,$1A,$FF,$B7,$9B,$D8,$6F,$B3,$2F
	.dc.b	$F3,$4D,$AF,$F4,$5F,$84,$F9,$8F,$CF,$36,$33,$18,$3F,$05,$F9,$17
	.dc.b	$F7,$C8,$FE,$86,$3F,$1B,$7D,$07,$F4,$98,$C3,$FB,$29,$E1,$E6,$7F
	.dc.b	$57,$97,$D0,$BE,$C3,$FC,$38,$D8,$FE,$72,$D7,$C0,$70,$8F,$F1,$17
	.dc.b	$D9,$FF,$34,$AE,$8D,$F6,$7F,$C2,$2D,$AF,$F2,$85,$F3,$AF,$ED,$1D
	.dc.b	$8C,$5E,$1F,$9F,$EA,$26,$1F,$D6,$EB,$53,$88,$FE,$23,$5F,$CE,$9B
	.dc.b	$EA,$67,$0E,$B3,$F8,$B6,$F4,$FC,$33,$E2,$70,$7C,$5F,$7A,$61,$DE
	.dc.b	$FF,$CA,$B7,$C2,$3F,$B6,$78,$8F,$8B,$72,$7F,$77,$FA,$0F,$F3,$57
	.dc.b	$E1,$17,$23,$FD,$EE,$F8,$56,$63,$FB,$2D,$70,$C3,$E6,$B5,$FB,$CD
	.dc.b	$7F,$C7,$75,$E1,$F4,$7A,$63,$FC,$A2,$7F,$C0,$B3,$97,$FC,$05,$5C
	.dc.b	$F2,$F8,$7F,$1B,$F8,$3F,$3A,$57,$BE,$2B,$F3,$3E,$5B,$F1,$06,$2B
	.dc.b	$E5,$FC,$67,$E2,$C9,$3F,$1F,$B3,$F1,$DF,$47,$23,$E5,$3A,$7C,$BF
	.dc.b	$03,$E7,$BF,$C9,$7D,$77,$FB,$86,$7F,$C3,$01,$C3,$C3,$7F,$CE,$43
	.dc.b	$DD,$FF,$39,$36,$7F,$C1,$ED,$7F,$2E,$D7,$F5,$0F,$F8,$6E,$BC,$3F
	.dc.b	$05,$D0,$EB,$5F,$A7,$FA,$D9,$7F,$83,$7F,$55,$F4,$3F,$AB,$E4,$93
	.dc.b	$F8,$77,$6F,$D2,$FD,$6F,$FD,$8A,$FE,$21,$7D,$C7,$E0,$DB,$19,$F9
	.dc.b	$0C,$4E,$7D,$A9,$3C,$63,$F7,$5E,$97,$1D,$24,$FE,$5E,$3F,$25,$F8
	.dc.b	$F7,$F2,$0E,$83,$49,$E7,$FF,$8D,$3A,$9A,$7F,$90,$78,$7F,$CB,$51
	.dc.b	$8F,$9E,$F3,$1D,$2B,$F8,$57,$93,$FC,$2F,$E1,$26,$9F,$A7,$8A,$FE
	.dc.b	$D3,$E6,$FF,$88,$18,$C7,$F6,$7F,$2F,$FC,$4F,$DA,$95,$FA,$F8,$AF
	.dc.b	$F8,$33,$F7,$3F,$E2,$D5,$FC,$19,$BE,$77,$F5,$CF,$1B,$FB,$3C,$FF
	.dc.b	$9C,$7D,$3F,$F7,$27,$D1,$7F,$2E,$26,$1F,$C6,$23,$2F,$F7,$2C,$1F
	.dc.b	$51,$FE,$16,$6C,$7F,$0F,$6B,$FE,$64,$F0,$97,$BF,$A4,$FF,$28,$7E
	.dc.b	$E9,$FE,$5A,$CB,$C4,$DD,$3F,$B8,$E3,$FE,$4A,$9C,$FB,$9F,$E1,$BA
	.dc.b	$FE,$C9,$A1,$A7,$F6,$7C,$DF,$93,$3E,$AB,$FD,$A0,$D8,$FF,$36,$B5
	.dc.b	$FE,$6A,$D7,$FA,$37,$E8,$2F,$B9,$FE,$88,$D8,$FD,$7B,$5F,$E9,$73
	.dc.b	$CF,$2F,$5F,$D5,$F9,$C9,$3F,$2F,$14,$7F,$85,$74,$7D,$37,$CB,$25
	.dc.b	$7F,$A4,$2B,$6F,$CD,$EA,$FA,$04,$7E,$87,$81,$F3,$3C,$29,$1C,$5D
	.dc.b	$3F,$4F,$E7,$BF,$B5,$91,$FB,$DC,$BF,$E6,$57,$FC,$49,$97,$F0,$3D
	.dc.b	$3F,$15,$E4,$7D,$F3,$FD,$11,$5F,$E6,$B8,$FC,$77,$D2,$7E,$0F,$92
	.dc.b	$61,$F8,$38,$FC,$37,$C8,$7B,$C9,$1F,$8B,$7E,$17,$07,$E2,$C9,$1F
	.dc.b	$8B,$8F,$E8,$58,$FF,$72,$E6,$4F,$F9,$27,$EE,$3F,$CD,$99,$7E,$C9
	.dc.b	$B3,$AF,$0B,$F8,$D4,$E7,$9F,$F7,$55,$F7,$EF,$4D,$7A,$47,$D5,$F7
	.dc.b	$FF,$3C,$48,$FC,$F7,$0F,$E8,$7C,$89,$1F,$A1,$7F,$49,$E4,$FE,$97
	.dc.b	$EF,$3F,$4D,$F0,$12,$BF,$4D,$1F,$D7,$72,$7D,$7F,$90,$93,$FE,$08
	.dc.b	$D5,$3A,$8F,$D9,$7C,$7F,$EC,$48,$FE,$C7,$EB,$7F,$69,$F1,$BF,$1A
	.dc.b	$3F,$C3,$1F,$04,$E9,$FF,$77,$71,$FD,$EA,$D3,$6B,$E6,$BF,$82,$F9
	.dc.b	$7C,$C7,$ED,$1A,$AF,$A4,$61,$B1,$C8,$F1,$7D,$D3,$BB,$FC,$F9,$1F
	.dc.b	$E3,$38,$C6,$BF,$EE,$06,$C7,$F8,$A1,$AE,$FC,$0F,$F0,$FB,$96,$B9
	.dc.b	$F8,$4E,$51,$88,$FF,$3D,$FD,$DF,$FA,$3F,$3F,$24,$7A,$63,$0F,$C6
	.dc.b	$6C,$7F,$A1,$5B,$1C,$BE,$8D,$3C,$B3,$52,$73,$5C,$DB,$5C,$DB,$1E
	.dc.b	$4F,$8A,$DF,$5F,$F8,$F2,$FF,$CD,$F7,$FE,$39,$7F,$89,$36,$76,$FE
	.dc.b	$77,$F5,$D7,$C2,$FF,$95,$70,$C3,$5F,$F5,$74,$6A,$7F,$18,$D8,$FF
	.dc.b	$54,$5F,$FB,$46,$7C,$1F,$0B,$DA,$7E,$37,$BD,$74,$F7,$CD,$B7,$D4
	.dc.b	$D3,$1F,$4B,$71,$F5,$7C,$CE,$CE,$1F,$41,$63,$FC,$8B,$A3,$E3,$3D
	.dc.b	$1C,$A8,$C4,$FA,$27,$31,$8D,$BF,$CF,$95,$BE,$4F,$43,$C7,$E3,$6D
	.dc.b	$7F,$99,$B0,$EF,$1E,$9F,$BA,$91,$F3,$FE,$E6,$E3,$FE,$F6,$DD,$1F
	.dc.b	$2F,$87,$EF,$B6,$33,$5F,$ED,$66,$D7,$CC,$FB,$9F,$8D,$EB,$FC,$76
	.dc.b	$D7,$DD,$FB,$8F,$94,$7C,$FF,$5D,$E8,$FF,$59,$9D,$1F,$C8,$3A,$BF
	.dc.b	$BC,$46,$71,$FF,$4E,$57,$0F,$33,$F0,$0D,$BF,$E7,$B1,$8F,$51,$E4
	.dc.b	$77,$6C,$FE,$3F,$EE,$31,$3F,$75,$91,$F9,$3F,$B8,$B8,$FC,$5F,$2B
	.dc.b	$F0,$47,$3F,$26,$D7,$F5,$97,$C0,$47,$FB,$13,$F9,$59,$7E,$23,$FE
	.dc.b	$3E,$9F,$F2,$DB,$63,$FD,$6F,$F7,$6E,$6F,$CA,$47,$0F,$3E,$CE,$6F
	.dc.b	$FD,$5F,$28,$C3,$FE,$B2,$6B,$FE,$C3,$6B,$FD,$33,$7F,$F0,$67,$99
	.dc.b	$F1,$C9,$5F,$DE,$5F,$C0,$7A,$FF,$25,$FD,$13,$63,$F7,$1C,$3F,$A5
	.dc.b	$3E,$E7,$DD,$BA,$F1,$3F,$6C,$F0,$46,$5F,$F1,$83,$71,$FB,$AD,$7F
	.dc.b	$9E,$62,$8C,$47,$F1,$EF,$85,$FD,$B7,$99,$FC,$C7,$9F,$F8,$64,$62
	.dc.b	$73,$E0,$7E,$E6,$2E,$3F,$C3,$DE,$39,$72,$F0,$6C,$FF,$B2,$F1,$FD
	.dc.b	$86,$3F,$ED,$25,$F2,$DA,$C9,$39,$8C,$3F,$E6,$6F,$93,$70,$7E,$D9
	.dc.b	$B8,$7F,$AC,$3E,$AF,$FC,$8E,$FF,$B2,$23,$38,$74,$8F,$F0,$0F,$C5
	.dc.b	$AB,$73,$FE,$C0,$6B,$FD,$86,$35,$E0,$7F,$23,$6D,$7F,$BD,$3C,$3F
	.dc.b	$EA,$9F,$2F,$E3,$F5,$BE,$4B,$67,$C5,$EF,$79,$FD,$CF,$43,$63,$CD
	.dc.b	$EE,$61,$A7,$03,$E4,$7A,$3F,$50,$FA,$BB,$58,$F9,$8E,$1F,$9D,$F2
	.dc.b	$D7,$6F,$80,$7F,$88,$34,$7C,$27,$E8,$9B,$5F,$EC,$27,$FC,$EA,$D8
	.dc.b	$D7,$C2,$FE,$BE,$73,$5F,$06,$30,$7F,$30,$FC,$AF,$FC,$AB,$7C,$CF
	.dc.b	$E5,$6E,$27,$C6,$7F,$80,$63,$F1,$FF,$2F,$1F,$C3,$7C,$BD,$9F,$4E
	.dc.b	$1F,$E0,$18,$FC,$A9,$F9,$6F,$2E,$D7,$E3,$7E,$5F,$F3,$3F,$2B,$F9
	.dc.b	$A6,$C7,$E4,$7E,$57,$0D,$38,$1F,$AB,$C5,$EA,$3F,$50,$7C,$97,$F4
	.dc.b	$46,$D7,$FB,$FB,$F1,$3F,$C6,$F7,$7F,$AE,$A7,$38,$7D,$0E,$B5,$7E
	.dc.b	$4A,$3F,$18,$D7,$F2,$6D,$51,$C3,$BB,$FA,$DD,$3E,$4F,$FB,$64,$E7
	.dc.b	$1F,$46,$7E,$79,$CE,$3E,$49,$FC,$07,$82,$F9,$7F,$DE,$9F,$BE,$F9
	.dc.b	$7B,$3F,$B5,$70,$FE,$DE,$EE,$9F,$4A,$FD,$BA,$FE,$E7,$E5,$7F,$83
	.dc.b	$EF,$FC,$21,$FE,$13,$F2,$98,$C7,$F8,$03,$70,$FC,$0E,$4E,$5F,$E6
	.dc.b	$6D,$AF,$A5,$2F,$C9,$FF,$C5,$B9,$7F,$89,$21,$5B,$7F,$17,$3F,$E3
	.dc.b	$27,$FC,$B0,$71,$DF,$FA,$34,$66,$3F,$CB,$8D,$7F,$CA,$5A,$3E,$5B
	.dc.b	$C7,$D3,$FC,$89,$B9,$7E,$17,$CD,$7E,$EA,$F6,$FF,$B3,$B6,$7F,$DC
	.dc.b	$0D,$C7,$FB,$6B,$5F,$E6,$3E,$7B,$A7,$FB,$7B,$6B,$F9,$CC,$60,$FE
	.dc.b	$E7,$F8,$9B,$E2,$E9,$27,$FD,$DD,$19,$7F,$DE,$1F,$25,$F8,$BC,$FF
	.dc.b	$D8,$FD,$9F,$DF,$23,$06,$7F,$13,$91,$FE,$00,$6D,$7F,$BE,$7E,$8D
	.dc.b	$F0,$3F,$E0,$46,$C6,$63,$07,$F8,$77,$E3,$BF,$B1,$ED,$FE,$7D,$6F
	.dc.b	$88,$69,$AC,$DF,$89,$D2,$7B,$9F,$E1,$27,$8F,$E2,$6F,$88,$FF,$15
	.dc.b	$37,$3F,$E6,$4D,$CB,$F0,$39,$23,$11,$FE,$9B,$87,$D4,$FF,$13,$8C
	.dc.b	$35,$FE,$4C,$D7,$FD,$2C,$DC,$9F,$88,$FF,$28,$9F,$F1,$4B,$5F,$ED
	.dc.b	$AD,$7F,$C0,$8D,$7F,$D4,$1F,$89,$BE,$6F,$FC,$88,$DC,$BF,$A6,$36
	.dc.b	$7D,$6D,$7F,$BC,$FC,$6B,$D1,$FF,$20,$36,$BF,$8D,$67,$FE,$8B,$D7
	.dc.b	$35,$FE,$45,$FC,$0F,$58,$7F,$90,$1A,$FF,$81,$1A,$FF,$86,$DA,$B4
	.dc.b	$7E,$07,$FD,$0B,$C4,$FF,$4B,$E2,$FB,$D7,$C5,$FF,$94,$EB,$5A,$7F
	.dc.b	$51,$9C,$BF,$AB,$FC,$0F,$F8,$5B,$FE,$29,$6E,$5F,$EC,$5C,$9E,$1F
	.dc.b	$F9,$76,$73,$0F,$92,$7B,$1F,$E5,$86,$E1,$FC,$0B,$0F,$44,$61,$FF
	.dc.b	$2F,$B5,$FF,$11,$BF,$E9,$BD,$7F,$D6,$3F,$0A,$FC,$0D,$EE,$7F,$9B
	.dc.b	$6B,$58,$30,$5E,$EB,$FD,$40,$BF,$03,$F9,$F8,$F4,$6D,$F1,$9D,$6B
	.dc.b	$C2,$73,$87,$CF,$FA,$6F,$F3,$B3,$FE,$38,$6B,$FE,$4F,$6B,$BF,$89
	.dc.b	$FF,$5C,$5B,$D1,$7C,$17,$91,$F4,$D7,$A1,$CE,$5F,$D8,$6B,$87,$C0
	.dc.b	$7A,$BF,$03,$7A,$B4,$FF,$47,$7C,$BF,$D4,$BE,$13,$A7,$F9,$34,$67
	.dc.b	$1F,$F4,$6B,$5F,$F2,$8B,$5F,$F3,$43,$5F,$C7,$7E,$07,$FC,$8D,$FF
	.dc.b	$4A,$B5,$FF,$2C,$B5,$FF,$38,$35,$FA,$7F,$81,$FF,$A0,$78,$7F,$DD
	.dc.b	$7C,$CA,$E9,$1F,$7A,$75,$87,$2F,$3F,$EC,$FC,$73,$E4,$A9,$F7,$3D
	.dc.b	$BF,$66,$1F,$86,$9D,$5F,$C9,$9C,$BF,$75,$B9,$FE,$19,$FE,$B1,$5F
	.dc.b	$8E,$73,$1F,$8B,$3F,$54,$3F,$02,$FF,$7A,$B9,$E1,$5F,$E4,$67,$33
	.dc.b	$C7,$B7,$1C,$5F,$D8,$AF,$73,$CE,$3F,$49,$3A,$8E,$97,$5E,$46,$CC
	.dc.b	$94,$CF,$CF,$78,$1F,$F7,$20,$70,$FA,$9D,$5F,$89,$CD,$77,$9E,$38
	.dc.b	$61,$EC,$63,$0F,$3B,$8B,$F5,$CE,$CE,$B8,$73,$9F,$92,$7F,$AE,$AF
	.dc.b	$FD,$76,$FE,$DD,$AE,$8F,$A1,$F1,$BF,$D3,$7F,$03,$FF,$60,$B6,$3F
	.dc.b	$85,$35,$FF,$06,$B5,$E3,$E3,$FF,$A6,$1D,$B5,$DB,$5D,$EB,$DA,$38
	.dc.b	$63,$F6,$AB,$F8,$E5,$7D,$36,$BF,$45,$4F,$0C,$F9,$CE,$7B,$BF,$15
	.dc.b	$7C,$DB,$3A,$8E,$D5,$E7,$D7,$57,$FC,$6D,$AF,$DB,$AE,$B1,$C3,$1F
	.dc.b	$26,$1D,$2B,$F9,$16,$3E,$6C,$75,$1E,$7C,$7C,$CE,$AB,$F3,$D7,$FE
	.dc.b	$3B,$BF,$6B,$6B,$7A,$FE,$C7,$DD,$DE,$DF,$EB,$8F,$98,$DB,$71,$CE
	.dc.b	$5F,$CC,$B8,$BF,$CF,$2F,$F9,$EF,$E1,$17,$C2,$3F,$9C,$7C,$7B,$E2
	.dc.b	$F8,$C2,$30,$F1,$9E,$19,$BE,$BB,$FB,$C4,$F5,$FD,$9D,$F2,$7E,$2C
	.dc.b	$91,$89,$FE,$59,$5C,$1F,$F4,$24,$65,$FF,$51,$FE,$4B,$FE,$51,$7F
	.dc.b	$CC,$2D,$77,$7D,$7F,$61,$7C,$DF,$F5,$D9,$FF,$43,$C6,$1F,$EC,$18
	.dc.b	$A3,$11,$FE,$64,$8C,$4F,$E9,$2F,$F9,$EC,$F8,$3C,$2F,$87,$18,$8F
	.dc.b	$F4,$D5,$F2,$E4,$FF,$39,$7E,$59,$FD,$32,$7F,$CE,$ED,$7F,$D7,$AD
	.dc.b	$8F,$ED,$7F,$82,$5E,$4F,$E3,$DE,$0F,$F9,$6E,$FD,$4D,$7F,$BC,$35
	.dc.b	$FF,$3D,$B5,$E3,$F4,$FF,$62,$1F,$03,$4C,$47,$D3,$FC,$A1,$F2,$F4
	.dc.b	$97,$FD,$7F,$E1,$9E,$67,$D9,$BF,$8F,$3F,$F0,$05,$AF,$F7,$FE,$24
	.dc.b	$62,$3F,$D6,$ED,$F5,$9E,$1B,$CA,$FC,$27,$E0,$AF,$94,$A7,$E6,$2B
	.dc.b	$73,$F9,$2F,$2F,$FA,$86,$3F,$EA,$26,$BF,$EE,$86,$F1,$FF,$8C,$3F
	.dc.b	$03,$7C,$47,$E6,$BF,$1D,$7C,$C7,$F2,$06,$E1,$F8,$76,$BF,$E2,$56
	.dc.b	$BF,$E6,$5F,$03,$E5,$BF,$EB,$2F,$BA,$F7,$B7,$83,$D1,$3A,$86,$5F
	.dc.b	$D3,$BC,$7F,$F1,$1D,$7E,$D1,$B3,$FE,$D6,$AE,$6F,$FB,$F3,$F0,$2F
	.dc.b	$F1,$4B,$FE,$DA,$FC,$E3,$FD,$1E,$FF,$8E,$69,$18,$8F,$F5,$EC,$63
	.dc.b	$E2,$FF,$44,$D8,$FF,$20,$70,$BE,$87,$FD,$91,$9F,$97,$F4,$3F,$F2
	.dc.b	$AD,$7F,$BA,$1A,$FA,$BB,$BF,$C7,$97,$C5,$FA,$9E,$44,$9F,$A8,$FE
	.dc.b	$FF,$D4,$FC,$6C,$7F,$B6,$3F,$27,$FE,$9A,$47,$FA,$8B,$FD,$4B,$AD
	.dc.b	$F9,$07,$EF,$73,$FF,$6E,$71,$FF,$16,$BF,$A1,$8F,$F0,$BA,$D6,$5F
	.dc.b	$E4,$FD,$9F,$8A,$BE,$DF,$E7,$37,$0F,$F2,$A6,$5F,$4E,$31,$1E,$D3
	.dc.b	$FD,$69,$F0,$DF,$57,$F2,$6F,$EF,$11,$F3,$DA,$FC,$E6,$BF,$E1,$5C
	.dc.b	$11,$88,$F8,$FF,$69,$FE,$93,$7C,$39,$79,$FE,$C8,$7E,$0C,$B1,$FE
	.dc.b	$96,$9E,$0F,$E7,$7E,$B3,$F4,$44,$7F,$D3,$7F,$91,$BD,$EF,$E6,$FF
	.dc.b	$8D,$BE,$B3,$FB,$EB,$6B,$F9,$23,$5F,$F4,$13,$5F,$F6,$17,$DF,$FF
	.dc.b	$8B,$DF,$88,$39,$3E,$3F,$0F,$D4,$E2,$FD,$B5,$6E,$3F,$3D,$ED,$FF
	.dc.b	$B4,$EF,$DE,$38,$7F,$D2,$9F,$23,$FA,$19,$D4,$FD,$1F,$84,$FC,$A5
	.dc.b	$7F,$50,$D9,$F3,$1F,$E1,$66,$D7,$F3,$D6,$BF,$E9,$B6,$BE,$97,$AE
	.dc.b	$3F,$E3,$E8,$46,$38,$FF,$C9,$8D,$8F,$F4,$C3,$D2,$7F,$D7,$3F,$8C
	.dc.b	$FE,$9A,$BF,$CA,$5E,$8F,$F6,$2B,$F8,$AD,$AD,$CF,$F6,$88,$C3,$DB
	.dc.b	$1F,$61,$9F,$7B,$F8,$F6,$1A,$C3,$FD,$51,$EC,$FF,$61,$C7,$D0,$DC
	.dc.b	$3F,$63,$5B,$AF,$F4,$8F,$B7,$FD,$DB,$F2,$97,$DE,$D3,$FB,$CE,$1C
	.dc.b	$6B,$E9,$70,$7F,$89,$9F,$E5,$5C,$7F,$CB,$4F,$A9,$F0,$56,$3E,$C8
	.dc.b	$C1,$D5,$DC,$D4,$7E,$1A,$1F,$5D,$E3,$24,$72,$8E,$2F,$4F,$AE,$FC
	.dc.b	$46,$DF,$E7,$D6,$C7,$FB,$25,$BE,$33,$7E,$B3,$E4,$FC,$A9,$F5,$9F
	.dc.b	$E1,$66,$C6,$63,$FC,$33,$7D,$23,$FC,$2D,$E9,$7F,$87,$3E,$13,$FA
	.dc.b	$2C,$7B,$9A,$FF,$03,$C7,$C1,$AF,$5F,$8A,$93,$E1,$79,$5F,$F8,$9B
	.dc.b	$F2,$C7,$99,$49,$5F,$A2,$ED,$7E,$1B,$FC,$FB,$A7,$F2,$99,$70,$AF
	.dc.b	$F0,$C3,$5F,$F7,$B3,$5F,$DA,$35,$FF,$52,$3F,$E1,$1E,$1B,$F0,$DF
	.dc.b	$E3,$88,$C4,$FF,$BC,$DE,$95,$FA,$3F,$D8,$DF,$71,$FE,$3E,$7F,$9F
	.dc.b	$C6,$7C,$BC,$4E,$5E,$9D,$13,$0F,$C2,$DF,$0C,$BF,$D6,$DF,$A3,$BE
	.dc.b	$D7,$FC,$92,$DA,$FC,$5B,$5F,$E1,$AF,$D0,$CB,$F6,$DF,$8E,$7F,$87
	.dc.b	$DB,$F0,$9F,$E5,$28,$C6,$1D,$B0,$FF,$2A,$FC,$73,$EE,$29,$32,$FC
	.dc.b	$CF,$EC,$4F,$9B,$E5,$E0,$FF,$A8,$32,$FE,$83,$5F,$C7,$DA,$FF,$B3
	.dc.b	$DA,$83,$E7,$CB,$5D,$1E,$53,$AC,$FE,$A3,$F4,$D7,$DB,$7F,$9A,$9B
	.dc.b	$19,$8F,$F3,$6E,$4F,$A0,$F7,$79,$32,$BF,$63,$97,$EF,$F0,$E3,$1F
	.dc.b	$CD,$3E,$47,$EE,$BF,$E5,$83,$F6,$EE,$39,$1F,$7F,$FE,$7A,$FC,$89
	.dc.b	$F3,$74,$97,$C3,$B3,$FD,$0F,$F1,$CF,$CF,$BF,$C9,$93,$FE,$38,$6C
	.dc.b	$67,$A3,$FC,$DC,$D9,$FF,$38,$7E,$43,$FE,$F0,$EA,$FE,$60,$DE,$5F
	.dc.b	$F7,$D3,$67,$A5,$FF,$22,$FB,$7F,$F1,$27,$C0,$7F,$8A,$70,$FF,$14
	.dc.b	$37,$0F,$E4,$33,$AF,$94,$FF,$39,$FD,$4C,$C7,$FA,$7F,$AB,$FE,$0A
	.dc.b	$9F,$5E,$BF,$CD,$E7,$ED,$1E,$BE,$6F,$4C,$66,$3E,$F6,$DF,$06,$7F
	.dc.b	$DC,$BF,$85,$3E,$DF,$FD,$46,$DA,$FE,$4A,$D7,$FC,$15,$5C,$5F,$EC
	.dc.b	$2B,$4F,$F4,$E4,$78,$A3,$1E,$BF,$CA,$9C,$CF,$89,$AF,$E4,$6B,$FA
	.dc.b	$B2,$69,$FD,$66,$B9,$C7,$F7,$0F,$37,$F9,$47,$D6,$FF,$93,$FF,$5A
	.dc.b	$FE,$21,$87,$F8,$9D,$B4,$AD,$BF,$2A,$32,$BE,$77,$FC,$11,$C7,$FD
	.dc.b	$AD,$B7,$CF,$E4,$7F,$13,$F9,$07,$FB,$1E,$FF,$C1,$47,$FA,$01,$C3
	.dc.b	$FE,$0D,$CD,$18,$8F,$88,$D7,$EB,$FD,$A7,$88,$95,$DB,$E8,$7F,$DB
	.dc.b	$BE,$8B,$F4,$17,$D4,$61,$B8,$7F,$31,$6B,$FC,$F7,$DB,$7D,$67,$FA
	.dc.b	$F3,$E5,$FF,$D8,$04,$9F,$F6,$56,$FF,$E7,$27,$D6,$FF,$4B,$23,$C4
	.dc.b	$FE,$CF,$EC,$7E,$C1,$BE,$27,$F4,$53,$FD,$01,$F8,$3F,$EF,$E3,$FB
	.dc.b	$6E,$EF,$C0,$FF,$DB,$10,$8C,$3F,$ED,$B6,$BF,$31,$AF,$F8,$DD,$FF
	.dc.b	$33,$3F,$E3,$85,$F8,$67,$FB,$E9,$B1,$E0,$F9,$3F,$F7,$EF,$DF,$FF
	.dc.b	$AF,$A3,$D3,$F1,$5F,$D6,$CC,$FD,$1F,$EA,$3A,$7F,$C5,$5E,$FF,$2A
	.dc.b	$FF,$22,$B6,$33,$96,$FA,$BF,$5F,$3F,$B9,$AF,$E5,$59,$BE,$4B,$DE
	.dc.b	$D8,$F7,$46,$0C,$FA,$DF,$B8,$FA,$C7,$E1,$FF,$84,$F6,$30,$DB,$FE
	.dc.b	$6D,$3A,$7B,$D6,$9F,$E7,$5B,$FF,$60,$77,$5F,$C1,$23,$10,$AE,$4A
	.dc.b	$33,$1D,$7A,$FF,$C4,$9F,$17,$D6,$BE,$C7,$4D,$D3,$F1,$B1,$D8,$F8
	.dc.b	$31,$BD,$7F,$C1,$F8,$76,$A7,$99,$FE,$67,$BF,$BC,$D7,$EF,$35,$FD
	.dc.b	$2B,$5F,$9B,$F0,$5F,$97,$7F,$52,$D7,$3D,$7F,$8B,$8C,$47,$E3,$FC
	.dc.b	$9F,$E2,$CF,$C1,$DB,$F6,$31,$8C,$7E,$1E,$5F,$C6,$F8,$BD,$97,$C5
	.dc.b	$BB,$BE,$07,$86,$3F,$35,$B4,$E0,$FF,$4B,$73,$62,$3E,$74,$DF,$27
	.dc.b	$FD,$39,$8F,$65,$C5,$9F,$94,$D3,$BF,$C0,$7F,$CF,$D8,$7F,$3E,$A4
	.dc.b	$62,$3F,$BD,$46,$2F,$F8,$AC,$23,$0F,$F1,$8F,$C7,$DF,$81,$7C,$4E
	.dc.b	$BF,$F4,$54,$62,$FF,$CC,$AF,$F1,$06,$CE,$DF,$C6,$7C,$FD,$BF,$F0
	.dc.b	$57,$0B,$EA,$71,$1F,$8D,$9B,$39,$AF,$F3,$8E,$5F,$DE,$1B,$F1,$BC
	.dc.b	$23,$17,$FE,$B6,$7F,$CC,$73,$C3,$F0,$6F,$81,$3E,$D6,$C9,$A6,$B3
	.dc.b	$F8,$3A,$7B,$B5,$F4,$7A,$AF,$C3,$FF,$E0,$06,$3F,$EC,$4B,$DE,$5F
	.dc.b	$A9,$F0,$BD,$7C,$FF,$EB,$88,$FF,$44,$AF,$97,$CC,$E9,$F2,$C6,$67
	.dc.b	$E9,$B6,$FF,$A4,$B5,$FE,$F3,$CE,$FA,$17,$DA,$7C,$D8,$C3,$F3,$FD
	.dc.b	$4F,$F4,$46,$99,$AD,$E3,$F3,$3F,$05,$F9,$47,$D5,$F5,$3F,$29,$F0
	.dc.b	$18,$9F,$EF,$39,$7C,$EC,$FF,$BC,$35,$FF,$03,$B5,$57,$29,$FE,$C3
	.dc.b	$2E,$27,$B7,$F1,$BE,$AF,$D1,$18,$AF,$8F,$1E,$9F,$4B,$F0,$58,$FF
	.dc.b	$B9,$6B,$4F,$F5,$9E,$BE,$38,$7E,$2B,$F0,$8F,$F3,$16,$3F,$EB,$86
	.dc.b	$D7,$F8,$53,$27,$D0,$6F,$E3,$7E,$04,$62,$FF,$BB,$EB,$FB,$5A,$FD
	.dc.b	$43,$63,$33,$BB,$FE,$FB,$87,$EF,$E7,$C1,$E1,$7C,$BE,$9F,$E6,$F1
	.dc.b	$83,$FC,$46,$E2,$BF,$C0,$4D,$7F,$C4,$8D,$7F,$C6,$6D,$4B,$E5,$9F
	.dc.b	$1F,$F9,$0F,$C3,$7F,$2F,$18,$BF,$C8,$C7,$CB,$F9,$1C,$D5,$9F,$F2
	.dc.b	$3D,$6F,$0F,$AD,$1F,$E4,$6C,$98,$7C,$D3,$F0,$DF,$4B,$FE,$08,$9F
	.dc.b	$9D,$18,$F4,$3E,$3C,$E6,$FF,$C3,$CD,$7F,$C8,$0D,$7F,$CA,$7F,$80
	.dc.b	$BD,$6F,$E0,$6D,$C3,$FD,$E3,$F9,$87,$FA,$65,$FF,$79,$C6,$6B,$CE
	.dc.b	$7F,$A5,$67,$4E,$03,$8E,$6F,$E7,$B0,$F9,$27,$5F,$F1,$2A,$D6,$1E
	.dc.b	$4C,$FC,$66,$7E,$4A,$C7,$91,$C3,$FE,$A5,$C7,$AF,$C9,$FF,$23,$CB
	.dc.b	$C5,$87,$FA,$76,$B4,$7F,$9C,$BF,$3B,$FE,$76,$E2,$7F,$D5,$D1,$87
	.dc.b	$D4,$FF,$42,$6F,$9D,$FC,$A2,$31,$1F,$E7,$C0,$E1,$ED,$E7,$FF,$22
	.dc.b	$8C,$47,$FA,$EF,$F6,$47,$B3,$FC,$C9,$B8,$78,$E5,$0F,$A1,$D5,$78
	.dc.b	$7E,$5E,$5E,$EB,$F6,$FA,$FC,$EC,$7E,$5E,$3D,$FC,$1E,$C7,$7F,$1A
	.dc.b	$E1,$C7,$DA,$F6,$57,$8D,$B1,$D7,$D9,$C3,$4E,$07,$9F,$F4,$87,$AB
	.dc.b	$FD,$31,$B1,$9F,$82,$C4,$E5,$FE,$8B,$3F,$EC,$69,$FE,$C6,$DC,$3F
	.dc.b	$CF,$2D,$7E,$76,$BF,$E0,$17,$F1,$CD,$D7,$FC,$13,$E1,$BE,$B1,$F8
	.dc.b	$8F,$CD,$C7,$C9,$DF,$1F,$9D,$F7,$3F,$87,$C3,$F2,$B9,$FE,$C2,$73
	.dc.b	$F1,$3F,$D9,$AB,$51,$FE,$F3,$F4,$1E,$6B,$E0,$B5,$7F,$DB,$32,$FF
	.dc.b	$30,$DF,$DB,$6E,$FF,$A5,$C7,$F1,$99,$FE,$92,$72,$FE,$9A,$33,$5F
	.dc.b	$EC,$86,$D7,$FA,$9F,$1F,$C3,$BC,$FA,$BC,$1F,$1D,$2F,$C1,$CB,$6B
	.dc.b	$C1,$EC,$79,$63,$D8,$D9,$FF,$6C,$FE,$43,$F2,$DF,$04,$7F,$72,$8C
	.dc.b	$50,$65,$FD,$C9,$C3,$FE,$04,$CF,$F7,$FF,$83,$BC,$1F,$95,$EA,$7F
	.dc.b	$84,$23,$19,$7F,$B1,$F6,$FC,$F3,$F5,$27,$FA,$97,$81,$C2,$3E,$BF
	.dc.b	$DB,$F1,$F9,$5F,$A8,$D8,$E5,$F1,$18,$69,$C0,$FD,$67,$E8,$9F,$D5
	.dc.b	$34,$FE,$26,$D8,$FF,$7A,$AF,$84,$FF,$12,$B9,$DF,$E0,$39,$9F,$F1
	.dc.b	$CB,$63,$FD,$9E,$D7,$F8,$23,$5F,$91,$E9,$BF,$46,$FF,$1D,$47,$F2
	.dc.b	$A8,$C7,$C2,$FF,$82,$FB,$BF,$25,$39,$AD,$E5,$FE,$66,$EC,$FF,$3B
	.dc.b	$DF,$F8,$EB,$9B,$F7,$F1,$8C,$FF,$85,$47,$F0,$2C,$1F,$94,$7E,$92
	.dc.b	$7F,$CA,$2D,$8F,$F7,$3B,$5F,$E3,$8D,$7F,$00,$3F,$30,$FF,$1D,$4F
	.dc.b	$F1,$76,$C7,$F4,$4B,$7D,$A7,$F8,$E6,$B5,$F0,$BF,$E2,$5F,$AF,$FC
	.dc.b	$0D,$FC,$1F,$CD,$BF,$C9,$2F,$18,$CE,$BF,$E8,$4E,$05,$6A,$31,$F9
	.dc.b	$CB,$EE,$FF,$CD,$D4,$8C,$65,$FC,$D7,$1F,$EF,$31,$87,$F9,$C4,$EE
	.dc.b	$B8,$FD,$9F,$E7,$6F,$FA,$24,$A3,$15,$ED,$FC,$EF,$E9,$D7,$DC,$9C
	.dc.b	$E3,$FE,$9A,$F6,$FF,$C0,$DD,$9F,$E6,$EF,$CF,$9F,$83,$7F,$9C,$E7
	.dc.b	$38,$7F,$3F,$9C,$BF,$D0,$47,$D6,$FF,$9A,$9A,$7F,$A2,$0B,$9F,$B5
	.dc.b	$AF,$F5,$3F,$D1,$FF,$C3,$D9,$7F,$70,$6C,$FE,$E7,$07,$CD,$BE,$4F
	.dc.b	$FD,$07,$5A,$C1,$CC,$F1,$1F,$0B,$FE,$74,$5E,$57,$F9,$7D,$CE,$7F
	.dc.b	$21,$CB,$F2,$BF,$0F,$FE,$71,$87,$E2,$B9,$11,$88,$FD,$2B,$5F,$F1
	.dc.b	$7B,$61,$FA,$8F,$F8,$22,$BE,$AB,$6B,$F2,$15,$FE,$DA,$9C,$FC,$0F
	.dc.b	$F9,$CF,$8F,$E3,$F0,$FF,$84,$BF,$54,$7D,$5E,$79,$3F,$D4,$71,$8C
	.dc.b	$3F,$B9,$69,$FE,$60,$CF,$F0,$FF,$1D,$FE,$9C,$8C,$47,$AE,$3F,$D3
	.dc.b	$FF,$AF,$71,$8F,$F1,$13,$72,$FD,$C7,$CD,$3E,$A3,$FD,$63,$39,$C7
	.dc.b	$FA,$67,$B4,$F3,$BF,$0D,$FB,$07,$F0,$2C,$7F,$80,$4E,$6B,$FC,$31
	.dc.b	$D6,$AD,$47,$DD,$FD,$6B,$E8,$E9,$FE,$15,$A4,$62,$3F,$77,$18,$E7
	.dc.b	$FF,$6B,$36,$BF,$50,$BF,$0B,$FF,$15,$57,$F9,$02,$31,$F1,$3F,$EC
	.dc.b	$DF,$A0,$7E,$23,$E0,$D7,$F4,$58,$70,$7F,$C8,$37,$F5,$DB,$E2,$FF
	.dc.b	$B4,$A3,$19,$79,$23,$1F,$11,$FE,$CA,$9F,$F2,$6E,$7E,$B8,$C1,$FE
	.dc.b	$CE,$FD,$4B,$F9,$B4,$1F,$E7,$19,$BD,$FF,$F4,$46,$5F,$E5,$29,$DF
	.dc.b	$E7,$9F,$E2,$8C,$FF,$8E,$B7,$06,$5F,$86,$DF,$FD,$CF,$39,$8C,$57
	.dc.b	$C6,$EB,$F9,$97,$FA,$D6,$BC,$9B,$3F,$E3,$CF,$A5,$E7,$F1,$FF,$C0
	.dc.b	$FE,$1B,$FC,$03,$87,$8A,$3E,$67,$E1,$7F,$D9,$31,$FA,$15,$A5,$BA
	.dc.b	$75,$B7,$F1,$1A,$D6,$4F,$DA,$1F,$4D,$EA,$F8,$4F,$E2,$51,$8D,$3F
	.dc.b	$CA,$1B,$FE,$A7,$4F,$F3,$EE,$8F,$85,$FC,$84,$7E,$06,$F8,$3A,$8C
	.dc.b	$E9,$FC,$86,$72,$FF,$A0,$27,$FC,$11,$DF,$FE,$09,$FC,$45,$F7,$DF
	.dc.b	$E6,$9C,$59,$FF,$98,$6F,$62,$B6,$FF,$52,$9D,$BF,$E8,$F3,$71,$A7
	.dc.b	$FC,$2D,$1A,$8C,$46,$AB,$C1,$5E,$37,$FA,$35,$FF,$87,$7C,$6D,$E1
	.dc.b	$D9,$E3,$69,$D5,$72,$8F,$E7,$AF,$27,$D0,$E7,$0E,$33,$DE,$7D,$58
	.dc.b	$57,$FB,$6C,$57,$EE,$C5,$7C,$07,$43,$89,$4E,$27,$F5,$F3,$E4,$96
	.dc.b	$30,$E0,$FE,$F7,$F8,$8B,$E9,$1F,$3D,$C6,$30,$D8,$CC,$FF,$6B,$9C
	.dc.b	$B7,$C9,$6B,$C5,$B1,$98,$C1,$1F,$8E,$7F,$92,$E5,$F8,$27,$F8,$3B
	.dc.b	$5F,$B5,$18,$9F,$9E,$6E,$FA,$CE,$9B,$AF,$B2,$F9,$36,$B8,$8B,$FE
	.dc.b	$DB,$86,$1B,$1E,$43,$55,$C9,$F3,$35,$E5,$7E,$07,$27,$C9,$8F,$9F
	.dc.b	$3E,$98,$F9,$11,$8F,$A2,$91,$F3,$DA,$FF,$85,$FE,$C3,$F6,$EF,$DB
	.dc.b	$6B,$9E,$5F,$B7,$18,$8F,$F6,$93,$FA,$F7,$FC,$6E,$D7,$FC,$F0,$D4
	.dc.b	$DF,$AF,$7C,$1F,$E1,$9B,$3F,$8B,$6B,$FE,$2B,$6B,$FE,$E1,$7F,$97
	.dc.b	$3B,$F0,$39,$11,$E5,$F3,$5D,$48,$E0,$ED,$CF,$E1,$23,$83,$F3,$9F
	.dc.b	$88,$3E,$E3,$F3,$6D,$8C,$C6,$0C,$FD,$9C,$8E,$2D,$7F,$C8,$92,$FA
	.dc.b	$4E,$64,$8E,$EF,$C3,$F9,$19,$1F,$0D,$AF,$2F,$C1,$7C,$64,$7C,$EF
	.dc.b	$CB,$F8,$A9,$1F,$2D,$AF,$F9,$57,$F3,$87,$C8,$FE,$D9,$B1,$FB,$B6
	.dc.b	$BB,$8F,$DD,$FE,$1E,$F6,$A9,$27,$F8,$14,$65,$E5,$F0,$9D,$48,$E0
	.dc.b	$FB,$DF,$63,$F8,$12,$38,$3F,$86,$7E,$1A,$F8,$CF,$E1,$6D,$8C,$C6
	.dc.b	$0F,$C5,$54,$8F,$C5,$35,$FC,$54,$7E,$3C,$8F,$E3,$DF,$E2,$FF,$16
	.dc.b	$FA,$CA,$49,$FE,$3D,$19,$7F,$2D,$9F,$E6,$48,$E0,$FC,$D7,$A9,$F9
	.dc.b	$D2,$38,$3F,$94,$7E,$17,$FE,$D7,$6F,$0F,$F2,$88,$C4,$7F,$A2,$DA
	.dc.b	$FE,$87,$6F,$D4,$11,$FD,$43,$FC,$C3,$F4,$0F,$34,$7F,$A4,$DA,$FF
	.dc.b	$AB,$03,$87,$FD,$5E,$D7,$FD,$E1,$2F,$B4,$F2,$12,$3D,$2F,$F3,$EF
	.dc.b	$CB,$3C,$B1,$FE,$B8,$6B,$FE,$F9,$6A,$6F,$C8,$BE,$D3,$FA,$3B,$67
	.dc.b	$FA,$63,$5F,$F5,$63,$5F,$53,$FD,$2D,$DF,$93,$C8,$8F,$2F,$1B,$A9
	.dc.b	$1C,$1D,$BE,$17,$C2,$47,$07,$F5,$DF,$C2,$FF,$80,$3F,$EB,$D6,$BF
	.dc.b	$2D,$AF,$FB,$61,$A8,$2F,$9C,$FE,$CA,$D8,$ED,$18,$33,$F5,$52,$3C
	.dc.b	$4D,$7C,$5F,$4F,$C8,$8F,$9D,$EB,$DB,$23,$D4,$D7,$D5,$E4,$78,$88
	.dc.b	$FB,$DF,$EF,$3F,$85,$BB,$29,$27,$FC,$01,$19,$6F,$5F,$02,$38,$37
	.dc.b	$E7,$74,$23,$83,$9F,$DA,$F8,$08,$E0,$FF,$07,$65,$FE,$C9,$6F,$C9
	.dc.b	$FF,$83,$A3,$11,$FE,$F1,$6B,$F9,$F7,$FD,$62,$F2,$F7,$F8,$11,$EB
	.dc.b	$F8,$4F,$32,$38,$3C,$3C,$5D,$C8,$E0,$F2,$FA,$1E,$32,$38,$3F,$C4
	.dc.b	$DF,$84,$7E,$11,$34,$9E,$57,$F8,$A1,$79,$54,$93,$C9,$D7,$1C,$8E
	.dc.b	$AD,$7F,$7C,$76,$FA,$7E,$84,$7C,$6F,$7F,$94,$91,$E6,$6B,$CA,$FF
	.dc.b	$C0,$0F,$E6,$CF,$4F,$27,$9C,$8F,$BB,$D9,$F5,$11,$C1,$FE,$47,$FC
	.dc.b	$15,$C9,$FE,$46,$6C,$66,$30,$67,$C8,$91,$C5,$AF,$BB,$F0,$5F,$F3
	.dc.b	$3B,$FA,$F6,$A6,$98,$F4,$DF,$11,$FE,$11,$24,$FF,$84,$5F,$F0,$A7
	.dc.b	$C7,$C8,$FF,$0A,$35,$FF,$0A,$7C,$67,$F8,$64,$8F,$F8,$65,$FF,$30
	.dc.b	$FE,$07,$FE,$F1,$8F,$F3,$33,$5F,$AA,$D7,$F9,$A3,$5F,$9D,$7E,$7E
	.dc.b	$BF,$F1,$31,$23,$9F,$85,$D8,$8E,$0F,$F1,$57,$67,$F8,$B4,$8E,$0F
	.dc.b	$F1,$77,$DC,$FF,$8C,$48,$E0,$FF,$19,$78,$FF,$E3,$52,$38,$3F,$D0
	.dc.b	$1F,$7D,$F9,$66,$F4,$7F,$A0,$23,$11,$F9,$26,$BF,$E3,$9F,$A5,$FF
	.dc.b	$20,$11,$FF,$20,$3F,$E4,$3F,$84,$91,$FE,$43,$6B,$FE,$43,$3F,$C9
	.dc.b	$04,$3F,$C9,$07,$FA,$43,$EF,$3F,$AD,$26,$93,$D3,$F3,$7D,$FC,$9F
	.dc.b	$5B,$5F,$47,$DC,$FA,$C8,$FC,$07,$B7,$D5,$48,$FF,$51,$5A,$31,$1F
	.dc.b	$B2,$8F,$E2,$6F,$EA,$5A,$FF,$45,$6B,$E7,$F1,$BD,$5E,$2C,$8E,$AD
	.dc.b	$7E,$67,$95,$F3,$C8,$FD,$47,$E9,$78,$3F,$B7,$EE,$FF,$00,$6E,$8C
	.dc.b	$57,$EE,$E3,$0F,$F5,$79,$DC,$EA,$3C,$DE,$35,$7E,$63,$72,$34,$D5
	.dc.b	$BE,$3E,$92,$BB,$1E,$17,$7F,$18,$F3,$BF,$1A,$D9,$FC,$53,$5F,$5B
	.dc.b	$5F,$57,$94,$E3,$FC,$AB,$63,$F2,$4D,$7E,$1B,$5F,$85,$E2,$7F,$BF
	.dc.b	$B5,$F8,$ED,$8C,$C7,$A1,$FE,$F9,$92,$75,$1F,$6F,$EF,$2B,$F9,$16
	.dc.b	$EA,$69,$AE,$87,$C0,$D2,$63,$DE,$FF,$C6,$51,$BF,$5F,$FD,$5D,$46
	.dc.b	$92,$BF,$27,$F3,$5F,$EB,$3C,$4D,$24,$7E,$63,$E1,$74,$47,$07,$E7
	.dc.b	$3E,$D3,$91,$1C,$1F,$F0,$07,$0F,$E7,$6D,$F2,$7C,$A4,$75,$E4,$FA
	.dc.b	$08,$E0,$FD,$4F,$DA,$79,$C8,$E0,$FD,$6F,$AF,$EB,$23,$83,$F6,$3E
	.dc.b	$27,$BC,$8E,$0F,$DA,$FA,$5F,$0C,$8E,$0F,$57,$B7,$F3,$08,$E0,$FD
	.dc.b	$EF,$A7,$F2,$C8,$E0,$FE,$03,$F1,$1F,$3C,$8E,$0F,$E0,$BF,$87,$7D
	.dc.b	$32,$38,$3F,$BB,$F2,$3E,$F3,$FD,$25,$18,$6C,$7F,$7C,$6B,$FE,$7B
	.dc.b	$6B,$FC,$37,$ED,$BF,$00,$49,$FE,$2A,$FF,$BD,$67,$F2,$51,$D7,$9F
	.dc.b	$04,$7F,$8D,$FE,$01,$F8,$92,$38,$3F,$8F,$FD,$B7,$E3,$48,$E0,$FE
	.dc.b	$47,$F2,$1F,$91,$23,$83,$D9,$1E,$32,$38,$3E,$1F,$DE,$BC,$A4,$61
	.dc.b	$B9,$FF,$87,$A7,$53,$93,$8F,$95,$A9,$FF,$50,$4A,$31,$1F,$E1,$B8
	.dc.b	$C4,$F0,$3C,$3F,$73,$A9,$FF,$3D,$4A,$31,$08,$C4,$F0,$3D,$3E,$2E
	.dc.b	$A3,$F9,$14,$60,$93,$FE,$36,$FB,$B7,$C8,$74,$6C,$FE,$11,$B3,$B9
	.dc.b	$D7,$83,$2B,$4D,$9E,$9F,$21,$CC,$93,$CD,$ED,$CD,$23,$B3,$67,$C9
	.dc.b	$EC,$F8,$C9,$3E,$37,$CD,$E3,$C8,$F3,$36,$7D,$87,$AC,$91,$EB,$7E
	.dc.b	$4F,$DD,$3C,$2A,$49,$FD,$1C,$65,$BE,$17,$02,$38,$37,$DF,$D0,$8E
	.dc.b	$0E,$7C,$DE,$02,$38,$3B,$F2,$F9,$08,$E0,$FD,$AB,$F8,$6C,$7F,$CD
	.dc.b	$AD,$7F,$DD,$8D,$7F,$CC,$CD,$7F,$D8,$4B,$E7,$7F,$12,$D9,$EB,$18
	.dc.b	$33,$C7,$23,$C2,$DA,$F0,$FC,$F7,$22,$4F,$99,$EB,$E5,$C8,$F4,$B6
	.dc.b	$7F,$D0,$BF,$54,$E2,$FC,$DB,$6B,$F8,$9B,$67,$73,$AF,$26,$56,$9B
	.dc.b	$3D,$38,$B9,$92,$79,$BD,$BD,$D9,$1D,$9B,$3E,$43,$C6,$48,$F1,$BF
	.dc.b	$B0,$FB,$6F,$E3,$EF,$FB,$B5,$A9,$A4,$F5,$78,$12,$7F,$97,$C7,$F8
	.dc.b	$DD,$FF,$1C,$FA,$1F,$D9,$C8,$E0,$FE,$6B,$F6,$AF,$13,$F8,$0B,$63
	.dc.b	$31,$83,$F8,$2F,$CF,$3E,$1A,$92,$7F,$9F,$C6,$5B,$ED,$F0,$23,$83
	.dc.b	$FA,$2F,$CE,$3D,$8F,$E2,$2D,$8C,$C6,$0F,$0F,$D4,$C8,$EE,$DD,$BF
	.dc.b	$D8,$B8,$3E,$63,$CE,$4A,$F2,$BF,$C7,$97,$9F,$48,$FF,$58,$32,$7B
	.dc.b	$3C,$8F,$82,$47,$07,$C0,$F2,$3E,$29,$1C,$1F,$D8,$BF,$03,$FE,$77
	.dc.b	$8F,$FB,$69,$AF,$FB,$44,$38,$7F,$DA,$8D,$43,$1F,$D7,$3F,$03,$B3
	.dc.b	$EA,$92,$BE,$9F,$DC,$7D,$82,$38,$3F,$BA,$F1,$FF,$BD,$1B,$C1,$FD
	.dc.b	$0A,$31,$7F,$EF,$06,$BE,$7C,$3E,$B3,$F4,$37,$FC,$29,$2F,$F0,$7F
	.dc.b	$17,$F8,$82,$38,$3C,$3D,$BD,$C8,$E0,$FC,$5F,$BB,$F9,$02,$38,$3F
	.dc.b	$C1,$BF,$44,$F9,$6F,$EB,$2D,$8C,$C6,$0C,$F9,$32,$38,$B7,$2F,$A4
	.dc.b	$2B,$FD,$D7,$3D,$BB,$7A,$91,$FC,$FF,$C6,$FE,$90,$8E,$0F,$D1,$FD
	.dc.b	$E7,$EA,$08,$E0,$FD,$3F,$A1,$FA,$C2,$38,$3F,$57,$E2,$FE,$C0,$8E
	.dc.b	$0F,$D7,$EB,$FB,$42,$38,$3F,$67,$E2,$FE,$E0,$8E,$0F,$F1,$AE,$6F
	.dc.b	$85,$FF,$00,$B6,$33,$18,$33,$E6,$C8,$E2,$D9,$E7,$DD,$F9,$6F,$49
	.dc.b	$F1,$1F,$9D,$AF,$F0,$EC,$62,$BF,$41,$5B,$AD,$5D,$9C,$8A,$E0,$FE
	.dc.b	$8A,$F9,$D7,$D8,$BE,$13,$9B,$AF,$EA,$22,$D7,$08,$EB,$7D,$1F,$F5
	.dc.b	$75,$70,$55,$F1,$9A,$7B,$88,$7E,$E7,$E7,$FF,$85,$FD,$0F,$F9,$81
	.dc.b	$B1,$F8,$C6,$BB,$F0,$13,$BF,$6D,$F4,$34,$95,$C6,$32,$D8,$FE,$07
	.dc.b	$F4,$5F,$13,$07,$BB,$FC,$1A,$93,$98,$C3,$BB,$E6,$DB,$E7,$B7,$1C
	.dc.b	$39,$E1,$83,$F8,$8B,$87,$AC,$6E,$F9,$BC,$70,$FC,$6D,$6E,$BE,$34
	.dc.b	$60,$CF,$BB,$E7,$C3,$FC,$B9,$48,$C6,$BE,$36,$DF,$8A,$EB,$FE,$2F
	.dc.b	$F0,$AF,$AD,$FE,$13,$8F,$E4,$F8,$FF,$99,$E3,$11,$F3,$B0,$E8,$E9
	.dc.b	$F9,$11,$FE,$6D,$9E,$11,$D0,$EB,$F0,$9E,$2E,$BF,$F3,$86,$57,$C6
	.dc.b	$F4,$63,$CC,$FE,$14,$E1,$F0,$1F,$CE,$B1,$ED,$87,$D7,$F8,$DF,$F3
	.dc.b	$3C,$F1,$0E,$1E,$26,$3D,$AF,$83,$0F,$A2,$F2,$7D,$DF,$EE,$EB,$C5
	.dc.b	$9A,$31,$1E,$28,$C7,$4F,$2C,$39,$FA,$BF,$C1,$31,$FF,$15,$5F,$FA
	.dc.b	$E9,$AF,$F2,$66,$BF,$8F,$B7,$E1,$0F,$8B,$F8,$CD,$9F,$1B,$5F,$5D
	.dc.b	$74,$E1,$FF,$4E,$7D,$23,$CC,$FB,$AD,$AF,$88,$D7,$E1,$BF,$E1,$37
	.dc.b	$3E,$8F,$F1,$C8,$34,$91,$FE,$D4,$FB,$E7,$C8,$FE,$29,$B1,$F4,$5A
	.dc.b	$F6,$F4,$FF,$01,$F8,$6B,$C6,$A4,$9F,$B1,$A3,$DB,$FC,$36,$9E,$F9
	.dc.b	$F8,$CD,$7F,$BD,$F0,$7B,$F8,$3F,$AD,$CE,$67,$FA,$97,$E2,$6F,$A2
	.dc.b	$FC,$83,$63,$EE,$B5,$F9,$7F,$4F,$FA,$4F,$AA,$7A,$54,$93,$F8,$DF
	.dc.b	$A8,$FE,$D0,$FE,$3F,$EA,$1E,$97,$F4,$46,$C7,$FB,$E9,$AF,$F6,$66
	.dc.b	$BF,$E1,$9E,$3F,$F0,$D7,$3F,$FB,$55,$BD,$DF,$C4,$DB,$3F,$E1,$AC
	.dc.b	$1F,$21,$FB,$CF,$65,$E7,$FF,$8E,$6B,$F5,$51,$8E,$8D,$4E,$67,$F1
	.dc.b	$3F,$41,$FB,$3F,$C2,$CF,$F8,$11,$B1,$FD,$F3,$99,$18,$8F,$F1,$84
	.dc.b	$62,$7F,$C4,$7F,$82,$3E,$57,$C8,$DA,$F1,$B5,$FC,$CE,$7F,$3F,$D7
	.dc.b	$91,$FB,$9E,$44,$7E,$DB,$E5,$3F,$8B,$FD,$4B,$E1,$FF,$B8,$B6,$FF
	.dc.b	$C2,$21,$C3,$FE,$14,$6B,$FE,$47,$F8,$37,$C9,$7F,$22,$6C,$7F,$09
	.dc.b	$6B,$FA,$EF,$7F,$E7,$7D,$F3,$DF,$A4,$9F,$CC,$F7,$FD,$3F,$C4,$1C
	.dc.b	$FF,$E4,$66,$CF,$F9,$0D,$AF,$F9,$66,$11,$88,$FE,$A5,$1F,$E8,$DF
	.dc.b	$96,$FB,$DF,$12,$69,$FC,$6B,$E0,$FF,$31,$E4,$4B,$F9,$4F,$F3,$EF
	.dc.b	$AB,$7C,$FF,$F4,$0F,$AB,$FE,$37,$5F,$E6,$96,$BF,$D8,$5A,$FF,$9A
	.dc.b	$1A,$B8,$3F,$D0,$2F,$C7,$F8,$1F,$94,$7F,$4C,$FA,$BF,$DD,$47,$F3
	.dc.b	$3F,$AB,$DB,$FC,$C9,$BE,$2B,$C5,$FE,$E1,$18,$6F,$CE,$7F,$56,$87
	.dc.b	$C1,$7F,$8C,$1B,$7F,$80,$6B,$FE,$52,$6B,$F2,$BE,$25,$F3,$3F,$A7
	.dc.b	$6C,$7E,$99,$AF,$96,$BF,$B2,$FD,$5F,$F2,$58,$FE,$92,$DC,$BD,$11
	.dc.b	$FE,$8C,$8C,$65,$EA,$E5,$FD,$A7,$EB,$3F,$9A,$AF,$D9,$B5,$FE,$B7
	.dc.b	$08,$C5,$F4,$95,$F4,$7B,$7C,$6F,$90,$8F,$F5,$F8,$FF,$61,$3F,$D8
	.dc.b	$7F,$46,$7C,$D7,$F7,$E6,$C6,$63,$8B,$EC,$FA,$DF,$79,$1E,$9F,$49
	.dc.b	$F0,$BE,$1E,$47,$77,$4F,$AB,$F5,$E7,$D3,$FF,$13,$6C,$7F,$12,$6B
	.dc.b	$F2,$A7,$FC,$23,$F5,$DF,$C6,$A3,$FB,$C3,$5F,$EE,$ED,$D7,$E2,$3F
	.dc.b	$C4,$5F,$A3,$BB,$F2,$4F,$E9,$95,$FC,$81,$AF,$CE,$6C,$F2,$F3,$3C
	.dc.b	$3F,$55,$2B,$C2,$F9,$3E,$43,$FC,$4B,$F8,$9F,$EF,$E7,$FC,$18,$D7
	.dc.b	$FC,$16,$DB,$F2,$D7,$F8,$B3,$67,$D5,$D2,$5F,$F8,$56,$3E,$1D,$7D
	.dc.b	$F7,$FA,$C4,$1A,$4F,$80,$FF,$47,$FE,$4F,$FD,$52,$7F,$C3,$CD,$7F
	.dc.b	$C3,$AD,$AE,$93,$E3,$F9,$29,$5E,$17,$CD,$D5,$FE,$42,$FC,$4B,$F9
	.dc.b	$A5,$FE,$45,$BD,$7F,$E6,$36,$DF,$FB,$AB,$2F,$F2,$D5,$7E,$5F,$EC
	.dc.b	$BF,$C1,$8D,$F8,$4F,$F2,$6C,$63,$0F,$F3,$63,$F9,$BB,$7E,$27,$FA
	.dc.b	$09,$FF,$38,$36,$3F,$DE,$ED,$53,$CD,$FF,$7D,$BE,$F9,$EB,$F0,$7E
	.dc.b	$03,$C3,$DD,$E2,$F4,$3F,$D6,$F5,$F6,$6F,$FC,$FC,$74,$8C,$9F,$EB
	.dc.b	$CF,$D3,$FE,$CC,$7F,$92,$DA,$FF,$92,$9B,$9F,$C1,$9F,$7E,$32,$BF
	.dc.b	$CA,$57,$FE,$56,$7C,$FD,$5F,$E7,$0C,$DF,$39,$49,$3C,$2F,$FC,$EB
	.dc.b	$F6,$8F,$66,$92,$7F,$CB,$D1,$FA,$59,$FA,$DF,$2B,$F6,$C9,$3A,$AF
	.dc.b	$F2,$FF,$C3,$FF,$95,$49,$1F,$E5,$57,$FB,$4F,$73,$E1,$1E,$9F,$F8
	.dc.b	$35,$B5,$FE,$0C,$6B,$FE,$52,$AF,$F3,$5F,$D9,$C8,$FF,$1D,$BF,$E6
	.dc.b	$EF,$A7,$FF,$0B,$7E,$29,$FE,$11,$E3,$FD,$CB,$75,$5F,$F6,$88,$C3
	.dc.b	$8F,$8E,$61,$FE,$78,$F0,$BF,$CD,$C4,$9F,$F3,$73,$EA,$FC,$EB,$E4
	.dc.b	$47,$F8,$89,$AF,$F3,$06,$E3,$FC,$BE,$FF,$DC,$AD,$F0,$BF,$CF,$44
	.dc.b	$AF,$F3,$D3,$F1,$BD,$6F,$EA,$D1,$FD,$F9,$BC,$1F,$DB,$BD,$4F,$F4
	.dc.b	$6D,$7F,$A3,$BE,$CA,$5F,$F9,$79,$FF,$48,$FC,$97,$F8,$E7,$F1,$4F
	.dc.b	$F3,$7C,$FF,$8E,$1A,$FF,$46,$6E,$5E,$7A,$FF,$4C,$FC,$CF,$FA,$3C
	.dc.b	$95,$FE,$8F,$7F,$C4,$7F,$80,$3E,$AF,$FA,$53,$7E,$AF,$FD,$86,$DA
	.dc.b	$FF,$38,$D7,$FA,$91,$87,$F9,$D0,$FF,$53,$FC,$87,$FB,$23,$ED,$7F
	.dc.b	$C1,$71,$FE,$30,$6B,$FA,$C0,$E1,$FD,$7B,$5F,$BF,$4F,$AE,$7B,$BE
	.dc.b	$46,$CF,$FB,$0E,$75,$3E,$1E,$2F,$F6,$CF,$DB,$3E,$8A,$92,$FF,$D8
	.dc.b	$51,$FE,$17,$6F,$D2,$7F,$91,$94,$7F,$B7,$1F,$E0,$6D,$7F,$26,$6A
	.dc.b	$38,$7E,$0F,$E4,$8F,$F6,$5B,$78,$7F,$BC,$37,$B3,$8E,$BF,$E5,$29
	.dc.b	$FF,$03,$E8,$E2,$FE,$13,$5F,$09,$B8,$27,$5E,$9F,$CF,$F6,$7D,$FF
	.dc.b	$5D,$2B,$DA,$FD,$39,$F8,$7F,$5B,$E8,$FA,$B7,$D9,$7A,$B8,$1E,$86
	.dc.b	$63,$0D,$C7,$F6,$33,$FE,$0A,$55,$C9,$CF,$A3,$B5,$C3,$FE,$3C,$9D
	.dc.b	$46,$3B,$B5,$7F,$D5,$67,$3C,$5F,$ED,$97,$7B,$67,$3F,$9E,$DC,$3F
	.dc.b	$81,$D6,$E7,$D5,$F1,$5D,$3E,$BA,$57,$F0,$8C,$3C,$38,$FF,$80,$1B
	.dc.b	$E7,$7F,$9A,$E7,$DA,$D7,$F0,$2D,$9F,$E6,$7E,$17,$FA,$29,$FE,$E5
	.dc.b	$D9,$7D,$5F,$F5,$6D,$FE,$76,$72,$D8,$CB,$F8,$AF,$47,$FC,$8B,$8F
	.dc.b	$93,$E0,$2E,$BF,$4E,$6D,$7D,$09,$AF,$EA,$3E,$4A,$F0,$7F,$99,$7A
	.dc.b	$BD,$D8,$7E,$32,$73,$1C,$3A,$BF,$CB,$AE,$AF,$F9,$5D,$B9,$F5,$F0
	.dc.b	$FD,$1E,$1F,$43,$C7,$E9,$AE,$1C,$70,$FC,$B4,$75,$FA,$1F,$F2,$FF
	.dc.b	$7F,$F9,$57,$34,$62,$FF,$3F,$F3,$77,$0F,$E7,$75,$A9,$FD,$D7,$DC
	.dc.b	$7F,$A2,$5B,$F1,$FB,$AF,$E0,$FE,$C7,$1F,$0D,$7B,$96,$7F,$E6,$5E
	.dc.b	$0F,$F6,$2E,$3F,$D2,$23,$1F,$35,$CF,$B9,$F6,$9D,$6F,$78,$72,$C3
	.dc.b	$F7,$BA,$FF,$B8,$1F,$F6,$4B,$63,$F8,$D3,$5F,$F0,$3B,$5E,$3F,$0F
	.dc.b	$E9,$8F,$F7,$9C,$F8,$CE,$BE,$DE,$67,$D9,$F6,$8F,$9A,$A7,$9F,$E3
	.dc.b	$6D,$F1,$C7,$F4,$31,$FE,$83,$7F,$85,$B5,$CC,$7F,$31,$8C,$47,$F3
	.dc.b	$0F,$3B,$E1,$1F,$0F,$C6,$B5,$FC,$AF,$F0,$57,$C9,$7C,$F6,$C7,$86
	.dc.b	$FF,$9C,$FC,$46,$6B,$E6,$3D,$B8,$7F,$D6,$75,$FD,$06,$30,$79,$79
	.dc.b	$7E,$BF,$C3,$FD,$BE,$53,$F5,$7E,$27,$0D,$BF,$06,$09,$FB,$5F,$5D
	.dc.b	$9C,$FF,$D4,$FF,$8F,$72,$BF,$8D,$E5,$23,$13,$F7,$4F,$E1,$3F,$35
	.dc.b	$F9,$39,$D6,$5F,$85,$8F,$51,$2F,$FA,$D3,$F3,$67,$CF,$5C,$6B,$FC
	.dc.b	$2B,$5E,$6F,$9A,$C4,$FE,$3D,$F8,$84,$8F,$EC,$8F,$3F,$C4,$9F,$57
	.dc.b	$E3,$6C,$FC,$0F,$23,$13,$F9,$96,$47,$C4,$E8,$B8,$7F,$B2,$3E,$2D
	.dc.b	$F2,$1E,$D6,$D7,$CA,$FB,$4C,$4F,$E8,$DF,$B0,$48,$FE,$F0,$FE,$23
	.dc.b	$E4,$58,$7E,$CD,$B3,$F8,$3F,$BA,$FD,$97,$2F,$FA,$16,$BF,$71,$18
	.dc.b	$9F,$EC,$ED,$CB,$FC,$76,$FF,$75,$8E,$5E,$37,$EF,$FD,$A7,$AD,$4E
	.dc.b	$7D,$76,$BC,$0E,$1F,$17,$AF,$71,$FB,$C3,$F5,$3E,$73,$F9,$96,$C6
	.dc.b	$BE,$ED,$F6,$5F,$C3,$E7,$3B,$7F,$BB,$E7,$37,$FE,$19,$B4,$61,$FF
	.dc.b	$0E,$35,$FF,$44,$B5,$FF,$03,$BF,$CC,$9F,$07,$A5,$FC,$76,$1E,$85
	.dc.b	$8F,$E4,$1F,$5A,$FB,$4A,$6F,$E4,$EE,$3F,$B3,$C3,$FC,$E6,$D8,$E9
	.dc.b	$F2,$5B,$8F,$DC,$5F,$87,$34,$7F,$86,$A3,$0F,$9B,$DD,$F0,$E5,$E5
	.dc.b	$64,$77,$D2,$9F,$3A,$9E,$B7,$A0,$FE,$71,$F7,$6F,$57,$17,$FC,$26
	.dc.b	$30,$FC,$B6,$DF,$F9,$15,$AF,$FA,$C9,$AF,$F8,$C1,$FE,$73,$1C,$FD
	.dc.b	$3E,$FE,$77,$81,$F9,$3E,$BC,$9F,$07,$81,$70,$FC,$2E,$1F,$E2,$29
	.dc.b	$C9,$E5,$EC,$F4,$D7,$F0,$5C,$5E,$A5,$3F,$AA,$4F,$E6,$70,$FE,$19
	.dc.b	$E3,$2B,$78,$7F,$27,$AD,$70,$FD,$0B,$FF,$4D,$C6,$27,$FC,$96,$BE
	.dc.b	$13,$AF,$C2,$FF,$15,$FC,$87,$FA,$25,$7F,$1C,$6D,$FE,$7A,$DD,$9B
	.dc.b	$F5,$3C,$AC,$BF,$17,$7D,$BE,$0F,$A8,$F5,$FC,$15,$9F,$C3,$FA,$DF
	.dc.b	$92,$3F,$B9,$7C,$47,$E5,$A3,$27,$F4,$37,$17,$FB,$06,$ED,$EF,$8F
	.dc.b	$F3,$D7,$3F,$2E,$FF,$E6,$1F,$A1,$FE,$77,$0F,$E6,$ED,$AE,$5D,$BA
	.dc.b	$F0,$3C,$2C,$AF,$07,$CA,$DA,$FA,$3F,$9D,$FF,$BA,$27,$FA,$0B,$63
	.dc.b	$C8,$7F,$3F,$71,$9F,$97,$4F,$E7,$33,$FE,$95,$6B,$FE,$7E,$3F,$D2
	.dc.b	$47,$F6,$A7,$2F,$FA,$0B,$EF,$9F,$19,$FE,$17,$6E,$1F,$D9,$D7,$D6
	.dc.b	$E6,$7F,$C3,$9D,$DF,$B8,$BF,$E2,$ED,$C7,$E4,$7B,$0E,$CD,$F8,$BF
	.dc.b	$E2,$3F,$90,$7B,$1F,$DF,$1B,$AF,$F1,$56,$E5,$FE,$AF,$6B,$EB,$FC
	.dc.b	$4B,$FC,$26,$FF,$26,$6E,$5F,$5F,$EF,$9F,$25,$FE,$31,$9C,$E1,$FE
	.dc.b	$F2,$CD,$F5,$FF,$C0,$23,$FB,$7B,$71,$FE,$63,$87,$F7,$0C,$7F,$41
	.dc.b	$7C,$3A,$7F,$B8,$7E,$24,$F8,$4F,$ED,$8D,$CB,$FD,$8D,$48,$C6,$3F
	.dc.b	$4E,$31,$5F,$DC,$FF,$11,$79,$BF,$DE,$DB,$87,$F8,$5F,$3F,$ED,$78
	.dc.b	$7E,$DA,$72,$FF,$46,$FC,$47,$FA,$DB,$7C,$4F,$F2,$85,$3E,$7F,$FC
	.dc.b	$A5,$5A,$DB,$F3,$98,$FF,$B5,$0F,$E8,$85,$C3,$FD,$BC,$D7,$F1,$4D
	.dc.b	$7F,$D9,$3F,$88,$BD,$FF,$F2,$F3,$67,$D1,$EC,$62,$7F,$C2,$DD,$BF
	.dc.b	$E3,$2B,$FE,$1E,$DC,$3E,$C7,$95,$AF,$5F,$FC,$DD,$F6,$FE,$9F,$4F
	.dc.b	$FC,$88,$DC,$FF,$DE,$4D,$C7,$F3,$CD,$7F,$DB,$BC,$8F,$87,$7D,$9F
	.dc.b	$F8,$9D,$B7,$F3,$7D,$0C,$4F,$F9,$DF,$F0,$FF,$E4,$38,$7F,$72,$6E
	.dc.b	$BF,$8D,$F9,$AD,$7A,$9F,$E7,$DC,$FF,$1F,$97,$F6,$D6,$E1,$F8,$2E
	.dc.b	$FF,$F2,$CC,$62,$3F,$69,$F8,$7F,$4A,$FF,$79,$B6,$03,$A3,$FD,$3D
	.dc.b	$3F,$E5,$99,$D6,$EF,$C3,$FE,$85,$FF,$91,$63,$1C,$BF,$82,$F9,$CF
	.dc.b	$97,$5A,$DB,$F9,$D5,$3F,$0F,$FA,$98,$78,$04,$67,$97,$E8,$CE,$A7
	.dc.b	$CB,$39,$CF,$E9,$FE,$1C,$F3,$BE,$CB,$63,$F3,$5F,$3D,$89,$FF,$52
	.dc.b	$7E,$A7,$FC,$63,$3F,$E5,$CD,$AF,$16,$1F,$D1,$A7,$38,$FF,$AA,$E3
	.dc.b	$EC,$BF,$82,$E6,$FF,$42,$C6,$23,$F0,$71,$8F,$2B,$81,$C7,$D1,$B8
	.dc.b	$7F,$A4,$ED,$F0,$9F,$D4,$A3,$0D,$C3,$F8,$1F,$D5,$FE,$52,$31,$97
	.dc.b	$F0,$F8,$FF,$4A,$F2,$9A,$4B,$FC,$BF,$E1,$DF,$D9,$F0,$F9,$AD,$7E
	.dc.b	$F7,$17,$E8,$23,$17,$FA,$5A,$FA,$8D,$AF,$C2,$BE,$28,$CD,$7E,$A3
	.dc.b	$F0,$E7,$CE,$D3,$F8,$1F,$B9,$6B,$F8,$63,$87,$F8,$77,$B9,$71,$FF
	.dc.b	$28,$F8,$4F,$A0,$FF,$6D,$36,$BF,$D5,$52,$FA,$3C,$D7,$FB,$7E,$31
	.dc.b	$3A,$9C,$79,$B9,$6C,$7F,$94,$E7,$F8,$D9,$FD,$5C,$A7,$F5,$99,$FF
	.dc.b	$60,$BF,$A4,$FA,$1F,$D2,$FE,$12,$FA,$AD,$5E,$1E,$53,$6B,$F3,$97
	.dc.b	$FE,$19,$BE,$11,$FD,$96,$74,$FE,$C2,$75,$5C,$A1,$5C,$1D,$5F,$07
	.dc.b	$C8,$7A,$EF,$D8,$9A,$7F,$9C,$7F,$09,$7C,$E3,$F0,$AC,$36,$BE,$14
	.dc.b	$EA,$72,$71,$F9,$D7,$3D,$25,$70,$8C,$B6,$3D,$B5,$FE,$D1,$8B,$F3
	.dc.b	$9F,$BF,$CF,$D9,$2B,$E3,$18,$6E,$9F,$C1,$67,$95,$FC,$48,$C5,$F3
	.dc.b	$75,$5C,$27,$FC,$04,$BE,$7B,$F9,$33,$63,$FB,$3B,$5F,$EC,$CD,$7F
	.dc.b	$9D,$BF,$21,$FF,$72,$B5,$FF,$2C,$B7,$1F,$EE,$EF,$36,$BF,$C7,$5B
	.dc.b	$1C,$A3,$1A,$FF,$81,$5E,$4F,$E0,$FD,$47,$CB,$6A,$BF,$BB,$C7,$FB
	.dc.b	$31,$FE,$6F,$7F,$E6,$5E,$0F,$EE,$1E,$B6,$DF,$EE,$57,$6F,$85,$FF
	.dc.b	$BE,$E7,$FB,$B7,$C6,$E4,$D5,$F1,$FB,$3F,$F3,$2D,$76,$CB,$F2,$D1
	.dc.b	$88,$E4,$DC,$BC,$4B,$AF,$D1,$F3,$FF,$ED,$46,$F5,$66,$FA,$09,$D3
	.dc.b	$8C,$D7,$D7,$1F,$D3,$7E,$93,$FE,$3A,$AF,$DC,$52,$31,$1F,$B9,$6B
	.dc.b	$FB,$26,$BC,$B8,$BE,$34,$7E,$33,$3F,$94,$DB,$EF,$F0,$5F,$2F,$F2
	.dc.b	$2F,$D9,$57,$CF,$5F,$61,$93,$5C,$BB,$F2,$BF,$DC,$1A,$2B,$FA,$B6
	.dc.b	$9E,$86,$BF,$C3,$DA,$FF,$08,$F8,$7F,$F4,$16,$3F,$75,$B9,$FF,$07
	.dc.b	$6C,$FF,$8E,$BF,$55,$7B,$BF,$81,$F5,$BF,$CD,$B3,$F2,$DA,$FE,$05
	.dc.b	$B8,$7E,$27,$F5,$DF,$F3,$C3,$F8,$BF,$D6,$7F,$69,$5F,$39,$AF,$F2
	.dc.b	$56,$BF,$C7,$9A,$FF,$82,$BF,$17,$FE,$1D,$C7,$F9,$97,$F8,$87,$65
	.dc.b	$F0,$39,$E1,$FE,$BD,$9D,$DF,$E1,$17,$E4,$AF,$5F,$92,$FA,$CA,$62
	.dc.b	$3B,$7C,$5B,$EA,$7F,$9E,$60,$C1,$39,$8F,$E8,$11,$89,$DB,$63,$3C
	.dc.b	$AF,$A7,$F5,$36,$3C,$CD,$7F,$92,$F2,$FC,$18,$FF,$41,$C7,$AF,$F2
	.dc.b	$5F,$EB,$0D,$C3,$E2,$47,$8E,$38,$9E,$77,$6E,$30,$7E,$33,$F0,$E3
	.dc.b	$D2,$D7,$FD,$61,$1A,$CF,$D7,$39,$9F,$F1,$77,$7B,$F2,$A7,$D4,$FE
	.dc.b	$D9,$B3,$FB,$49,$D4,$FD,$9F,$35,$F6,$3F,$BC,$FD,$23,$F1,$3B,$FF
	.dc.b	$92,$2D,$18,$C3,$F7,$8D,$D1,$F1,$7F,$12,$7D,$9F,$50,$FB,$1F,$A2
	.dc.b	$D8,$FA,$AD,$7D,$AD,$7F,$C9,$14,$FC,$FB,$FC,$45,$1F,$89,$5F,$05
	.dc.b	$F9,$16,$C6,$7D,$CF,$CA,$7D,$A3,$EB,$6F,$BF,$2B,$F2,$53,$97,$F2
	.dc.b	$9F,$79,$FE,$F5,$7C,$2D,$7F,$40,$D7,$F2,$0D,$DD,$F9,$EB,$D3,$FE
	.dc.b	$4A,$DA,$FA,$0D,$7F,$01,$ED,$BE,$0F,$F9,$5F,$E7,$B9,$F0,$DC,$DF
	.dc.b	$46,$30,$DC,$7F,$61,$F8,$43,$EB,$7F,$65,$F8,$6B,$ED,$FF,$3E,$DD
	.dc.b	$3F,$46,$D7,$F1,$6D,$7E,$7F,$CD,$7F,$B2,$1F,$DB,$3F,$96,$CF,$F7
	.dc.b	$4D,$CB,$F8,$13,$B9,$FC,$D7,$E4,$3F,$C6,$17,$E2,$DF,$8C,$7F,$A1
	.dc.b	$98,$EE,$FE,$11,$3B,$D3,$FD,$F9,$3F,$BC,$6B,$FE,$39,$6B,$FE,$8B
	.dc.b	$0E,$1F,$F4,$77,$E1,$5F,$91,$CB,$E4,$C7,$FA,$82,$31,$96,$BD,$1F
	.dc.b	$EE,$13,$CB,$E5,$7F,$B6,$7E,$A3,$FE,$F1,$8F,$EA,$CD,$7F,$B6,$37
	.dc.b	$0F,$EE,$8B,$EB,$38,$FD,$34,$AE,$2F,$4F,$A1,$F7,$FC,$EC,$9F,$E4
	.dc.b	$5B,$23,$E1,$FC,$FF,$C4,$F6,$65,$69,$E0,$FF,$0A,$1F,$84,$BE,$B7
	.dc.b	$D6,$D8,$F9,$ED,$7E,$AF,$C4,$6B,$E9,$64,$69,$E3,$5C,$B4,$93,$C9
	.dc.b	$DB,$F7,$BC,$EF,$B2,$47,$D7,$EA,$FD,$CE,$39,$1E,$E7,$4F,$F1,$FF
	.dc.b	$C4,$3E,$EF,$1F,$F4,$56,$F2,$FF,$4D,$6C,$FF,$3D,$7F,$C4,$EF,$F4
	.dc.b	$97,$FC,$3F,$3C,$7E,$2B,$FA,$94,$62,$B9,$73,$7D,$ED,$1F,$61,$79
	.dc.b	$66,$5F,$94,$9C,$BF,$96,$F0,$DF,$25,$FE,$8B,$6C,$7F,$50,$6B,$FE
	.dc.b	$8D,$6B,$F9,$55,$F6,$9F,$8E,$6C,$75,$8C,$4F,$A3,$C3,$93,$ED,$7D
	.dc.b	$7F,$41,$EC,$FB,$C9,$3F,$05,$DB,$FD,$33,$47,$E8,$5F,$C0,$E3,$FC
	.dc.b	$63,$18,$F1,$BF,$C9,$CD,$8C,$CB,$EE,$B2,$FE,$1F,$8F,$F5,$35,$D3
	.dc.b	$6C,$9A,$C3,$E7,$FC,$6F,$0F,$02,$65,$F5,$63,$F5,$BF,$84,$7F,$0D
	.dc.b	$81,$1C,$37,$C9,$FF,$5C,$46,$20,$FF,$92,$DF,$D0,$37,$2F,$DE,$C6
	.dc.b	$BE,$4B,$BD,$7E,$0B,$E1,$E4,$FE,$1D,$FC,$2F,$C3,$FE,$1B,$E5,$E4
	.dc.b	$FE,$2D,$DB,$F9,$5F,$95,$FC,$59,$1F,$C8,$7D,$BF,$E3,$B9,$E4,$7E
	.dc.b	$4D,$D3,$F9,$DE,$7F,$F0,$7F,$67,$E1,$23,$F3,$91,$83,$3E,$EF,$E6
	.dc.b	$09,$1F,$9D,$8F,$CD,$FE,$03,$F9,$CF,$66,$4F,$E8,$62,$BF,$AB,$F6
	.dc.b	$7F,$42,$63,$87,$F5,$9F,$51,$E2,$24,$FF,$89,$32,$47,$9B,$EC,$BF
	.dc.b	$52,$49,$FD,$74,$7E,$B3,$E4,$7D,$64,$8E,$4F,$F8,$B3,$EB,$FF,$C9
	.dc.b	$BE,$27,$F9,$15,$B3,$FE,$35,$6F,$C0,$B1,$FF,$04,$47,$F4,$9C,$BF
	.dc.b	$C8,$CD,$BE,$5F,$17,$FE,$4D,$8F,$EA,$18,$73,$F8,$BC,$11,$FF,$26
	.dc.b	$47,$17,$C1,$F8,$27,$94,$8F,$F9,$12,$11,$E9,$FC,$17,$B1,$27,$9C
	.dc.b	$78,$7E,$5F,$C5,$F3,$12,$7B,$C7,$97,$E6,$3E,$11,$23,$B3,$F1,$7E
	.dc.b	$07,$D4,$48,$F3,$C7,$ED,$0F,$E8,$4F,$EE,$17,$D6,$E4,$D7,$B3,$BF
	.dc.b	$93,$FF,$30,$F9,$0A,$FF,$34,$5F,$F6,$51,$19,$9D,$CE,$A7,$FC,$97
	.dc.b	$AB,$E9,$FE,$71,$23,$F8,$4F,$86,$FB,$7F,$A5,$F5,$32,$7E,$7B,$F5
	.dc.b	$7D,$3F,$57,$CA,$7E,$FA,$5F,$47,$FE,$21,$D7,$CD,$39,$8C,$69,$FE
	.dc.b	$5D,$7F,$C4,$DD,$4D,$7F,$B8,$46,$BE,$37,$87,$A9,$FC,$2B,$1F,$81
	.dc.b	$F1,$F8,$CF,$DF,$97,$F9,$09,$F0,$D6,$E3,$FC,$93,$18,$3F,$29,$E7
	.dc.b	$F1,$CF,$C1,$F8,$A5,$D0,$D9,$ED,$4F,$B6,$FF,$24,$3C,$CF,$D5,$E5
	.dc.b	$AF,$0B,$C5,$B7,$F8,$FE,$B6,$FF,$90,$7E,$43,$FD,$DC,$DF,$A2,$EB
	.dc.b	$06,$FD,$D6,$3E,$9F,$FC,$3D,$92,$91,$86,$F0,$FF,$A9,$1B,$3F,$83
	.dc.b	$5E,$E7,$F8,$DE,$3E,$B4,$63,$C3,$E1,$39,$9F,$E6,$F3,$97,$97,$C6
	.dc.b	$7F,$8B,$33,$F5,$7D,$45,$E5,$E7,$36,$B8,$CD,$7E,$47,$99,$78,$3F
	.dc.b	$CB,$BE,$3F,$F4,$4C,$7E,$34,$E6,$39,$73,$7E,$0F,$E1,$2F,$3B,$E1
	.dc.b	$6C,$61,$F0,$70,$FF,$98,$7D,$CB,$E0,$7F,$5D,$C7,$B7,$07,$E2,$63
	.dc.b	$FC,$51,$18,$DB,$F4,$F1,$B7,$F0,$D7,$FE,$6B,$F7,$39,$9E,$2C,$2D
	.dc.b	$F1,$D3,$F3,$13,$FE,$50,$F3,$3F,$8D,$B6,$7F,$92,$7B,$D7,$1F,$E4
	.dc.b	$93,$6F,$F4,$1E,$1D,$DF,$F6,$18,$7F,$7A,$F2,$1D,$59,$AE,$5A,$FF
	.dc.b	$9F,$F8,$3F,$D2,$CD,$BF,$E0,$4D,$CF,$FC,$D0,$D7,$F6,$CF,$FA,$1F
	.dc.b	$BF,$9F,$03,$8A,$C6,$1F,$37,$99,$F6,$2F,$FD,$B0,$DE,$1F,$F7,$F3
	.dc.b	$FE,$CA,$C7,$FD,$FC,$E5,$FF,$00,$EE,$8D,$57,$FB,$71,$AF,$FB,$01
	.dc.b	$AF,$06,$BF,$EF,$F5,$CF,$E4,$EA,$F2,$B6,$FC,$3D,$9E,$83,$D3,$DB
	.dc.b	$63,$CA,$FF,$AF,$71,$F2,$37,$0F,$D9,$4F,$F2,$B9,$CB,$63,$E9,$BE
	.dc.b	$EE,$1F,$8B,$ED,$FC,$66,$C7,$C2,$8F,$93,$BE,$34,$F9,$3E,$1F,$CC
	.dc.b	$3E,$6F,$83,$6F,$E4,$BE,$32,$3E,$17,$CD,$E2,$7D,$43,$EA,$F9,$16
	.dc.b	$7E,$8B,$EB,$23,$E9,$7D,$DE,$57,$DC,$F2,$BE,$EB,$67,$EC,$FC,$4E
	.dc.b	$1A,$7D,$C8,$FC,$1F,$C4,$C8,$FB,$FF,$11,$74,$FC,$2B,$F2,$C9,$1F
	.dc.b	$25,$F9,$DF,$11,$F8,$C3,$F1,$BF,$11,$67,$F1,$2E,$1F,$C5,$FC,$45
	.dc.b	$B7,$4F,$FA,$FB,$95,$F9,$F7,$F6,$0B,$FF,$27,$CE,$63,$15,$C2,$7F
	.dc.b	$27,$0F,$A7,$FD,$93,$AA,$E1,$18,$33,$F1,$56,$BF,$DB,$7F,$31,$F9
	.dc.b	$DC,$3F,$CB,$AD,$5F,$7F,$FB,$A7,$FD,$71,$19,$8C,$47,$FB,$6D,$AF
	.dc.b	$F6,$96,$BF,$E6,$86,$E5,$EE,$F2,$7F,$80,$FE,$B6,$E3,$F4,$C6,$1C
	.dc.b	$F3,$4C,$BF,$D0,$74,$E7,$C6,$5F,$E8,$98,$C4,$EA,$7A,$F4,$76,$7E
	.dc.b	$2F,$C6,$49,$F9,$A7,$CA,$F8,$CC,$61,$F3,$46,$9F,$52,$3E,$69,$F3
	.dc.b	$BE,$2E,$D7,$2A,$FF,$42,$FD,$FB,$E0,$BF,$93,$42,$31,$3B,$8C,$7C
	.dc.b	$1E,$A7,$31,$FE,$E5,$51,$FE,$86,$8F,$F0,$A4,$62,$7F,$C4,$D1,$C9
	.dc.b	$B1,$98,$FF,$15,$7D,$7F,$F1,$6E,$6F,$F4,$7B,$FC,$42,$BC,$4E,$5F
	.dc.b	$F1,$03,$99,$F8,$59,$FE,$53,$E3,$3F,$1A,$7E,$3B,$E3,$2D,$FC,$19
	.dc.b	$FC,$CC,$66,$31,$D2,$F8,$CC,$61,$D2,$1F,$19,$F9,$83,$F3,$3F,$19
	.dc.b	$6F,$8B,$87,$9F,$C6,$5B,$FA,$C3,$B3,$FC,$9F,$3F,$35,$B1,$A5,$E2
	.dc.b	$FD,$47,$3B,$FF,$AC,$9C,$D7,$19,$D7,$4E,$DB,$1D,$63,$F5,$FF,$17
	.dc.b	$8C,$3D,$9F,$19,$FB,$23,$F6,$7E,$D5,$BF,$86,$E1,$F8,$BC,$B6,$BE
	.dc.b	$27,$37,$EE,$8F,$DD,$FB,$96,$BE,$5B,$87,$E6,$FC,$65,$AF,$87,$F3
	.dc.b	$AF,$67,$F2,$2D,$AF,$C8,$3F,$E6,$F6,$C6,$BE,$13,$F1,$D5,$FE,$DE
	.dc.b	$AD,$3F,$E7,$36,$BF,$E6,$50,$E1,$FE,$CB,$F2,$5F,$32,$FF,$CD,$5F
	.dc.b	$8E,$BE,$1F,$FC,$0D,$39,$AF,$F7,$5C,$E6,$BF,$15,$EB,$7F,$86,$79
	.dc.b	$7F,$92,$3F,$E1,$EA,$FC,$9B,$99,$FE,$E6,$E6,$7F,$95,$78,$4F,$8F
	.dc.b	$FE,$C8,$DA,$FE,$C4,$DB,$F5,$77,$23,$11,$FE,$81,$F9,$C7,$E6,$CF
	.dc.b	$47,$FC,$41,$18,$7F,$C4,$1F,$37,$FD,$AA,$BF,$84,$4E,$71,$FE,$08
	.dc.b	$FF,$17,$F2,$F8,$F9,$3F,$C7,$E7,$F5,$1F,$33,$FE,$17,$77,$19,$BF
	.dc.b	$D8,$FB,$AF,$90,$FE,$FC,$DA,$FE,$F8,$D9,$F5,$B5,$FF,$4D,$6C,$FC
	.dc.b	$97,$E4,$47,$F1,$C7,$E4,$57,$F8,$71,$CC,$EF,$E2,$3F,$C7,$9F,$17
	.dc.b	$FB,$9A,$FD,$F2,$F2,$F3,$E5,$69,$FF,$21,$7D,$07,$F9,$F2,$BF,$C2
	.dc.b	$0D,$CB,$E7,$B5,$F2,$B5,$0E,$07,$EA,$FF,$C9,$39,$D7,$A9,$CF,$83
	.dc.b	$FE,$51,$F5,$9E,$77,$D6,$C3,$85,$EF,$F4,$7F,$C8,$60,$7F,$25,$C1
	.dc.b	$F9,$AF,$7F,$EE,$C6,$3F,$2F,$FA,$9A,$FE,$19,$AF,$CC,$6B,$FE,$44
	.dc.b	$6B,$FD,$23,$E9,$FF,$D2,$F8,$FF,$25,$7F,$97,$46,$79,$7F,$CD,$11
	.dc.b	$8C,$7F,$BD,$7A,$BF,$E6,$BF,$6C,$DE,$3C,$3F,$E6,$FF,$91,$FE,$BF
	.dc.b	$8F,$E6,$9A,$FE,$01,$AF,$F9,$61,$AF,$F5,$CF,$90,$79,$FF,$E7,$95
	.dc.b	$EA,$58,$C3,$FE,$7D,$E1,$FC,$77,$37,$F8,$5D,$FF,$31,$6F,$FD,$05
	.dc.b	$CB,$FE,$66,$73,$87,$F4,$A3,$F9,$69,$FB,$02,$9F,$96,$29,$FE,$6E
	.dc.b	$28,$C5,$F9,$47,$CC,$8F,$EE,$D1,$8E,$0F,$F4,$A3,$AF,$1A,$D6,$F1
	.dc.b	$FC,$4D,$F3,$FC,$33,$CE,$DE,$FE,$3F,$B0,$7F,$54,$D3,$FB,$E6,$4B
	.dc.b	$DA,$F9,$37,$D4,$FF,$28,$8C,$36,$33,$5F,$C9,$F1,$F5,$46,$1F,$E5
	.dc.b	$55,$CF,$E7,$3D,$33,$EF,$AF,$0A,$8C,$CF,$BB,$D2,$7C,$55,$31,$D9
	.dc.b	$70,$F4,$CA,$3C,$CD,$8F,$25,$23,$18,$F9,$1A,$FF,$23,$FC,$8F,$F9
	.dc.b	$B3,$FC,$79,$BB,$FC,$CA,$7E,$3D,$F9,$48,$F9,$FC,$8A,$DE,$3F,$BE
	.dc.b	$BD,$F9,$1F,$63,$0E,$13,$95,$ED,$66,$7E,$EF,$91,$F8,$CC,$BC,$0B
	.dc.b	$E7,$73,$F1,$FA,$77,$CD,$8C,$BE,$DD,$7D,$A3,$93,$FC,$43,$E5,$3F
	.dc.b	$84,$FA,$7F,$ED,$4F,$1F,$FA,$36,$5F,$88,$97,$E1,$3F,$C5,$AF,$FC
	.dc.b	$81,$1C,$1B,$19,$E2,$FF,$25,$4F,$07,$A7,$1F,$FB,$82,$75,$7D,$70
	.dc.b	$FA,$58,$7F,$B8,$1B,$7F,$EE,$06,$BF,$EE,$06,$BF,$8D,$F0,$3F,$94
	.dc.b	$FD,$CF,$C3,$8F,$E5,$67,$E2,$C3,$97,$31,$C3,$6F,$C9,$DD,$9F,$EF
	.dc.b	$D3,$5F,$B1,$15,$FE,$6F,$17,$8F,$F5,$B7,$D6,$72,$7F,$95,$84,$D6
	.dc.b	$99,$7F,$CB,$7F,$73,$FE,$47,$FB,$3B,$E7,$9A,$4A,$FD,$64,$FF,$40
	.dc.b	$D3,$8E,$9F,$EA,$C9,$CB,$5F,$F0,$C4,$ED,$CC,$FB,$63,$FA,$E4,$5F
	.dc.b	$1B,$E4,$E5,$F1,$E7,$83,$6F,$B3,$99,$DE,$BF,$42,$B9,$DF,$D0,$8C
	.dc.b	$5F,$07,$F2,$0F,$F9,$96,$3F,$01,$97,$F9,$F3,$83,$FA,$DB,$67,$53
	.dc.b	$F8,$37,$EE,$E5,$F8,$96,$ED,$D1,$B2,$BE,$6F,$E6,$3F,$59,$FE,$82
	.dc.b	$DF,$1B,$72,$9C,$CF,$E6,$AB,$FB,$6B,$FC,$93,$D6,$FF,$38,$37,$E3
	.dc.b	$7B,$F1,$7F,$89,$DB,$3A,$9F,$E6,$EA,$37,$F9,$77,$F9,$5E,$7F,$97
	.dc.b	$B5,$DC,$7A,$BB,$9F,$2F,$FC,$82,$7F,$C5,$AD,$7F,$C6,$2D,$DB,$F8
	.dc.b	$A7,$DD,$BE,$63,$CB,$F3,$FD,$67,$F8,$C3,$78,$7F,$8A,$37,$0F,$F0
	.dc.b	$96,$1C,$E7,$F8,$C7,$DD,$FD,$C9,$FD,$6E,$C8,$C4,$6B,$9B,$FD,$6B
	.dc.b	$39,$6F,$E2,$DF,$EA,$09,$FE,$9A,$D7,$FA,$DB,$72,$FE,$59,$7F,$8A
	.dc.b	$7F,$97,$7D,$E7,$F9,$0E,$BF,$AA,$AF,$86,$FA,$2D,$8C,$F8,$3F,$53
	.dc.b	$EF,$1F,$37,$E4,$6C,$FF,$42,$6B,$FD,$51,$AF,$BB,$C2,$7D,$27,$DE
	.dc.b	$6C,$7F,$66,$6B,$FD,$D1,$FE,$DE,$FF,$76,$FB,$CF,$EF,$4F,$F6,$C6
	.dc.b	$B9,$DF,$F1,$3F,$37,$FE,$79,$8F,$EE,$3F,$79,$FA,$07,$E7,$B5,$FE
	.dc.b	$94,$D7,$FC,$0C,$DF,$6B,$F8,$47,$67,$E8,$9F,$E1,$5E,$05,$F1,$3F
	.dc.b	$57,$C5,$FE,$04,$9D,$E1,$FE,$DB,$FB,$CF,$F6,$3B,$7E,$83,$94,$1B
	.dc.b	$F2,$9D,$7E,$F1,$F5,$BF,$EF,$4C,$11,$8C,$3F,$C0,$51,$89,$D4,$E6
	.dc.b	$7F,$D2,$73,$FB,$77,$FD,$D8,$D7,$FD,$E6,$DC,$7D,$1B,$BE,$87,$DD
	.dc.b	$F7,$9F,$E0,$7B,$FF,$7A,$3D,$6B,$DC,$DE,$DF,$F3,$7E,$08,$E1,$5F
	.dc.b	$F0,$03,$E0,$BE,$0B,$FE,$A6,$6B,$FE,$8C,$6B,$C9,$AA,$7F,$C4,$BF
	.dc.b	$4B,$F4,$7D,$B3,$4F,$DE,$7D,$E3,$E9,$7E,$B7,$CE,$7B,$73,$F2,$B7
	.dc.b	$1E,$8D,$F2,$1F,$79,$F8,$79,$6D,$F6,$FE,$2D,$F3,$9F,$4D,$B3,$F4
	.dc.b	$9A,$FF,$B0,$9A,$FF,$A5,$3D,$AF,$C4,$7C,$77,$F8,$4A,$3F,$D2,$B7
	.dc.b	$F8,$78,$C4,$78,$33,$F9,$70,$AF,$C8,$7C,$7B,$E3,$EF,$CD,$CC,$3F
	.dc.b	$D5,$71,$F9,$5F,$BC,$FD,$4C,$7F,$AE,$1A,$FD,$66,$BF,$35,$BC,$8F
	.dc.b	$8B,$7D,$8F,$E7,$DB,$5E,$8D,$3E,$FC,$62,$3F,$DA,$FF,$79,$FC,$CD
	.dc.b	$FF,$69,$FC,$4F,$E9,$FE,$57,$FC,$BD,$18,$3F,$DB,$7F,$79,$B7,$E0
	.dc.b	$1F,$7F,$E6,$1F,$51,$FA,$16,$C7,$E8,$1A,$FF,$BC,$1A,$FF,$B5,$BE
	.dc.b	$F3,$FC,$D6,$F9,$CF,$C5,$3F,$94,$E4,$FD,$B3,$74,$FC,$CB,$C2,$7C
	.dc.b	$63,$EF,$3F,$89,$D7,$F1,$2C,$BF,$15,$26,$CE,$AF,$FC,$4F,$F7,$8F
	.dc.b	$37,$FC,$30,$DB,$FF,$3D,$07,$0F,$F9,$F5,$AF,$F8,$C3,$E0,$5F,$8E
	.dc.b	$FF,$9C,$63,$FC,$8E,$D7,$FC,$8A,$DC,$BF,$93,$D7,$47,$FC,$D9,$F7
	.dc.b	$AF,$B1,$FE,$DA,$DB,$FF,$4A,$B5,$FF,$51,$AF,$BC,$DF,$D9,$C9,$DB
	.dc.b	$CA,$BF,$92,$7D,$2F,$F4,$0C,$3F,$91,$37,$5F,$E0,$AD,$BF,$DE,$2F
	.dc.b	$B3,$FE,$59,$F5,$DF,$B1,$BF,$E5,$4D,$BF,$E1,$AD,$BE,$3E,$DF,$F8
	.dc.b	$51,$7D,$EB,$ED,$70,$D8,$FC,$11,$A8,$CF,$C9,$E9,$DE,$92,$77,$5C
	.dc.b	$5F,$C1,$BF,$17,$2F,$E7,$DF,$A7,$FF,$99,$27,$F3,$3D,$E8,$C4,$FF
	.dc.b	$58,$AE,$1A,$7C,$2F,$2F,$E3,$F3,$4C,$7F,$2D,$8B,$83,$F8,$D4,$61
	.dc.b	$FB,$57,$FC,$BB,$2F,$EB,$F9,$78,$BD,$2E,$57,$E2,$FB,$CF,$E3,$D8
	.dc.b	$9F,$F1,$45,$7F,$55,$6E,$1F,$CE,$1D,$98,$AF,$E6,$18,$7E,$8E,$BD
	.dc.b	$FE,$87,$F6,$1A,$D5,$38,$3F,$0C,$FC,$EF,$0B,$FC,$7F,$7F,$DE,$1B
	.dc.b	$8F,$FA,$2A,$B7,$5F,$AE,$97,$D8,$FF,$9F,$5B,$1F,$A6,$6B,$FE,$E1
	.dc.b	$6B,$F4,$65,$F7,$9F,$83,$6C,$7E,$01,$AF,$7A,$F5,$7D,$B4,$8F,$A3
	.dc.b	$92,$3D,$DF,$1D,$FD,$FF,$5F,$E0,$D5,$F1,$3E,$7F,$07,$17,$FD,$6B
	.dc.b	$F7,$CF,$8E,$FF,$30,$BF,$16,$3F,$D7,$AD,$77,$1A,$CF,$FC,$CF,$F2
	.dc.b	$7F,$47,$C2,$FF,$34,$7A,$D9,$BE,$3F,$7B,$F9,$FF,$BE,$FD,$F7,$1F
	.dc.b	$FA,$D1,$B3,$FE,$75,$CD,$5B,$FA,$2F,$F5,$45,$7F,$97,$6B,$F3,$0D
	.dc.b	$79,$47,$FA,$1E,$BE,$AF,$2F,$FB,$5D,$B8,$7F,$B7,$5B,$41,$7F,$E5
	.dc.b	$D7,$FB,$C7,$17,$F9,$3E,$BF,$1D,$F0,$BF,$89,$3F,$15,$F2,$36,$BF
	.dc.b	$43,$C5,$FE,$0B,$7F,$23,$F0,$59,$7F,$39,$F6,$FF,$91,$E7,$91,$F6
	.dc.b	$22,$BF,$B2,$A7,$CF,$FD,$93,$3B,$7E,$87,$ED,$FF,$C5,$9A,$9A,$4B
	.dc.b	$FF,$16,$DF,$FB,$76,$BF,$7F,$F3,$7F,$83,$3F,$45,$5F,$A3,$FB,$7F
	.dc.b	$D2,$7B,$33,$0F,$C5,$C5,$7C,$9F,$25,$F8,$B3,$18,$7E,$CF,$E8,$7C
	.dc.b	$64,$9F,$F1,$EE,$68,$FF,$21,$67,$FE,$73,$9F,$77,$CA,$FE,$54,$FD
	.dc.b	$A4,$FF,$95,$7E,$F7,$FD,$CD,$E2,$1A,$4A,$E7,$F7,$1F,$9D,$23,$FC
	.dc.b	$4A,$3F,$8C,$F3,$7F,$8F,$1F,$D0,$D7,$3F,$03,$B1,$DF,$C6,$FD,$37
	.dc.b	$BB,$30,$FB,$6E,$9E,$9C,$9F,$E5,$BC,$3E,$AC,$7F,$96,$9A,$FF,$2B
	.dc.b	$C3,$D1,$3F,$03,$C5,$FD,$91,$27,$99,$E7,$E1,$F8,$1B,$48,$C1,$EB
	.dc.b	$D7,$C0,$47,$C0,$FC,$CF,$B7,$FD,$F1,$23,$81,$FE,$09,$F0,$7F,$D4
	.dc.b	$9F,$1F,$FE,$DC,$BF,$F2,$DF,$23,$87,$F4,$55,$FE,$A9,$F8,$FD,$5F
	.dc.b	$FB,$8B,$EA,$1F,$7B,$FE,$F3,$6E,$5F,$D2,$4F,$F7,$63,$87,$FA,$65
	.dc.b	$A3,$11,$FA,$F3,$FC,$C9,$C2,$D3,$F0,$04,$8F,$B5,$C5,$FC,$67,$D7
	.dc.b	$91,$FA,$E7,$F0,$DC,$3E,$BF,$A7,$E9,$F4,$DF,$E3,$9C,$FF,$D8,$5F
	.dc.b	$03,$8F,$37,$DB,$C7,$FE,$30,$AD,$3F,$E5,$86,$FC,$47,$97,$8F,$FC
	.dc.b	$31,$3F,$E5,$9D,$3F,$83,$46,$27,$51,$D1,$AF,$B3,$DB,$F0,$34,$FE
	.dc.b	$1D,$7F,$E4,$DE,$9F,$CE,$4E,$9F,$CF,$79,$7C,$29,$F7,$CF,$07,$1F
	.dc.b	$11,$E3,$96,$3F,$E5,$18,$C3,$FE,$53,$AE,$13,$E0,$D7,$E8,$7C,$4F
	.dc.b	$FC,$01,$8C,$1F,$C8,$57,$D0,$78,$5B,$1A,$0F,$17,$C3,$3D,$3E,$BB
	.dc.b	$FD,$69,$17,$8B,$FC,$8B,$49,$CC,$F7,$6B,$D3,$D6,$FF,$70,$BE,$66
	.dc.b	$BE,$56,$BB,$F0,$BA,$FC,$EF,$E6,$63,$C6,$7A,$FE,$4F,$1A,$FF,$9D
	.dc.b	$71,$FF,$7B,$F2,$7A,$DF,$F5,$8B,$ED,$DB,$FD,$4A,$FB,$F7,$FD,$23
	.dc.b	$9F,$90,$E7,$BF,$E6,$67,$39,$78,$38,$FF,$8B,$7C,$06,$A3,$F3,$AD
	.dc.b	$8F,$E3,$5F,$01,$7B,$FF,$C7,$D9,$7D,$2E,$3F,$D3,$7C,$67,$AE,$3F
	.dc.b	$CD,$1E,$0A,$31,$97,$ED,$A3,$13,$F9,$48,$C4,$E9,$FD,$9F,$B7,$E3
	.dc.b	$BF,$D9,$7B,$77,$0F,$EF,$D8,$78,$5F,$F1,$13,$F8,$4C,$FF,$C8,$5D
	.dc.b	$5F,$37,$C7,$FF,$23,$B6,$7F,$96,$63,$72,$FF,$17,$CD,$BF,$E7,$19
	.dc.b	$DD,$3C,$F5,$FE,$4E,$7F,$85,$E1,$FC,$DE,$79,$68,$F2,$1F,$45,$F7
	.dc.b	$70,$FF,$45,$46,$2F,$F8,$85,$6E,$B5,$5F,$C6,$AB,$4F,$F1,$2D,$7F
	.dc.b	$D1,$17,$FC,$8D,$FA,$31,$FE,$B8,$BF,$E2,$DB,$F8,$A7,$3B,$7F,$99
	.dc.b	$BB,$BD,$9E,$71,$F7,$79,$9F,$82,$9D,$65,$FE,$62,$8F,$C2,$E5,$FC
	.dc.b	$7F,$7F,$EF,$73,$E6,$84,$62,$3F,$95,$36,$3F,$A5,$4F,$FB,$6E,$7A
	.dc.b	$75,$F8,$3B,$7C,$2F,$93,$49,$3E,$1E,$EB,$87,$F9,$F3,$0F,$D9,$4F
	.dc.b	$66,$C7,$69,$F2,$63,$FE,$10,$6E,$9F,$0A,$FF,$CF,$1C,$DF,$03,$87
	.dc.b	$E1,$B6,$3E,$04,$7C,$4F,$6B,$18,$FF,$86,$F9,$FE,$49,$F2,$3C,$4B
	.dc.b	$7F,$1D,$F0,$E9,$FE,$E8,$8F,$34,$60,$F1,$F4,$7D,$23,$E8,$F8,$B6
	.dc.b	$7E,$83,$C8,$8F,$BD,$EB,$F0,$BF,$6B,$A3,$EC,$B6,$7E,$C7,$99,$86
	.dc.b	$9F,$6E,$3F,$03,$E6,$48,$FB,$DF,$09,$75,$F9,$CF,$A8,$91,$F3,$DF
	.dc.b	$6F,$C1,$FE,$97,$F2,$0F,$C4,$4F,$F9,$47,$97,$FC,$7F,$39,$8F,$C6
	.dc.b	$FC,$C7,$E3,$A2,$C7,$E3,$5F,$D7,$7C,$3B,$BB,$A7,$56,$1B,$5F,$EB
	.dc.b	$49,$7D,$96,$63,$0D,$D3,$F4,$5F,$7C,$D7,$F6,$6D,$AF,$F6,$8F,$CA
	.dc.b	$FA,$2B,$C1,$39,$5D,$FF,$B7,$9F,$D0,$C7,$F7,$F6,$BF,$E6,$F6,$A7
	.dc.b	$39,$29,$F7,$1F,$C3,$69,$F7,$FF,$C2,$1F,$3C,$66,$31,$87,$F4,$3B
	.dc.b	$E5,$CD,$FC,$33,$E6,$9F,$05,$EC,$8C,$3E,$EF,$4F,$1A,$FF,$89,$F4
	.dc.b	$46,$34,$FF,$72,$F2,$FF,$48,$6F,$A1,$F4,$BE,$EF,$EE,$37,$0F,$A1
	.dc.b	$1F,$5B,$E2,$B1,$5F,$1A,$7F,$C4,$5F,$88,$FF,$30,$8C,$1F,$E9,$37
	.dc.b	$1B,$FF,$21,$FC,$01,$F0,$5F,$0E,$73,$7F,$E2,$B9,$CB,$FE,$33,$B4
	.dc.b	$61,$FF,$1B,$35,$FF,$80,$0D,$7F,$C6,$9F,$80,$7B,$70,$F0,$7B,$3F
	.dc.b	$46,$31,$7D,$E3,$F2,$3F,$11,$F9,$BF,$36,$CF,$FA,$AE,$FF,$BA,$77
	.dc.b	$B8,$3F,$8D,$65,$EE,$87,$63,$E9,$38,$CE,$63,$18,$7E,$67,$2F,$CF
	.dc.b	$FC,$47,$EA,$7C,$EB,$3D,$CF,$D5,$65,$FC,$73,$3F,$0E,$0F,$B8,$DE
	.dc.b	$5D,$5B,$3F,$D3,$7E,$39,$E3,$FE,$D7,$DE,$FC,$84,$63,$2E,$D8,$7F
	.dc.b	$38,$CB,$FC,$4F,$18,$84,$BE,$13,$8E,$1C,$9B,$5B,$9D,$36,$3B,$F6
	.dc.b	$EE,$3B,$F0,$FC,$9F,$43,$F8,$41,$FC,$03,$E2,$6C,$FE,$F5,$C3,$FB
	.dc.b	$FF,$72,$E7,$F9,$8F,$4B,$F8,$81,$FC,$23,$AE,$D7,$F0,$57,$0F,$F0
	.dc.b	$7F,$6E,$D7,$F6,$5F,$B3,$FD,$6C,$FF,$A1,$9B,$1F,$EE,$9F,$C0,$3F
	.dc.b	$BC,$4F,$FA,$3E,$73,$29,$CB,$FE,$87,$6B,$FE,$83,$35,$ED,$FF,$65
	.dc.b	$6D,$7E,$6F,$F0,$0F,$F4,$33,$F3,$E9,$F0,$9F,$C1,$67,$35,$FA,$A3
	.dc.b	$C1,$1F,$E1,$3E,$9F,$D3,$C7,$F7,$07,$FC,$42,$FE,$4D,$CF,$07,$F9
	.dc.b	$5F,$D5,$FF,$0D,$78,$9F,$ED,$9C,$BF,$BA,$B6,$FF,$D7,$CD,$7F,$BF
	.dc.b	$35,$FE,$79,$F3,$BF,$86,$7F,$40,$DE,$AF,$F2,$C7,$D1,$BF,$00,$7D
	.dc.b	$9F,$F8,$A6,$B5,$8F,$F4,$5F,$9F,$FF,$08,$3F,$84,$6E,$3F,$C2,$30
	.dc.b	$FE,$2F,$E8,$FF,$8C,$67,$39,$7E,$CB,$C5,$7C,$97,$F7,$E6,$E3,$FE
	.dc.b	$10,$6E,$3F,$ED,$96,$BF,$E1,$9F,$C0,$3E,$29,$F2,$2F,$F7,$AE,$37
	.dc.b	$FF,$20,$7C,$4B,$C7,$BE,$46,$1B,$97,$E7,$3E,$97,$F8,$1E,$5F,$E2
	.dc.b	$56,$BF,$D8,$DA,$FF,$89,$9B,$C9,$FD,$1E,$BF,$C9,$DE,$B7,$7B,$FD
	.dc.b	$AF,$0B,$F0,$0F,$E4,$1B,$7F,$12,$AD,$65,$FC,$47,$A9,$DF,$FE,$57
	.dc.b	$6E,$3C,$BF,$00,$FE,$01,$5B,$AF,$F2,$E7,$E0,$1F,$42,$3F,$C7,$8D
	.dc.b	$7F,$BD,$35,$FF,$1E,$B7,$57,$D5,$7F,$A4,$A7,$F9,$13,$FC,$B1,$FF
	.dc.b	$21,$B9,$CF,$F9,$4B,$99,$FE,$F3,$E9,$FF,$9B,$7D,$C3,$7F,$00,$C3
	.dc.b	$FE,$71,$F9,$1F,$F1,$56,$3F,$AD,$B4,$62,$3F,$CA,$8D,$7F,$C2,$4D
	.dc.b	$7F,$CA,$BF,$23,$FC,$A7,$6F,$E2,$6B,$E6,$BF,$CF,$CE,$67,$F9,$E3
	.dc.b	$99,$FE,$33,$C7,$FE,$60,$9F,$E8,$0F,$E3,$B5,$FE,$8A,$E5,$BD,$7F
	.dc.b	$E8,$99,$79,$9F,$E5,$46,$DF,$F9,$99,$A7,$B4,$A0,$99,$3F,$00,$F9
	.dc.b	$CD,$F6,$BF,$D3,$0E,$7B,$AC,$EB,$EF,$FC,$D8,$72,$6E,$1F,$29,$7C
	.dc.b	$CF,$27,$38,$3F,$00,$7B,$3F,$E7,$F8,$C3,$63,$38,$7F,$21,$C3,$EA
	.dc.b	$C6,$1F,$E4,$95,$FE,$C1,$3C,$FF,$14,$FA,$2B,$18,$7D,$7F,$9A,$FF
	.dc.b	$1E,$83,$7E,$7B,$F3,$B7,$FE,$0E,$9C,$FC,$87,$C2,$9C,$CF,$C9,$3F
	.dc.b	$0E,$E1,$F9,$6D,$73,$DB,$ED,$FA,$A7,$C6,$7C,$C6,$EB,$FB,$4F,$C0
	.dc.b	$7F,$9A,$F3,$FF,$BD,$F3,$56,$B1,$FE,$35,$5A,$BF,$B5,$D0,$8C,$4F
	.dc.b	$E6,$BF,$01,$FF,$1A,$D7,$5F,$A0,$FB,$7B,$7D,$2F,$63,$FC,$A9,$A7
	.dc.b	$78,$C7,$C7,$6B,$E5,$37,$DB,$71,$FE,$FB,$BA,$FE,$FF,$C7,$3E,$99
	.dc.b	$EF,$FF,$B5,$9B,$7F,$E0,$49,$FF,$0D,$7D,$C6,$6B,$FD,$B7,$E8,$22
	.dc.b	$9F,$ED,$EF,$7B,$FD,$F1,$B7,$F9,$5E,$75,$87,$F6,$9E,$2F,$F6,$36
	.dc.b	$2B,$FF,$7F,$47,$D5,$8C,$3A,$8C,$BF,$8B,$F1,$3F,$45,$F8,$0F,$F4
	.dc.b	$08,$FB,$9B,$30,$CE,$FB,$95,$6D,$FD,$4C,$E9,$FE,$ED,$75,$C9,$C2
	.dc.b	$B9,$57,$18,$FE,$1F,$5B,$7F,$69,$5F,$E5,$56,$95,$C2,$7F,$CB,$5F
	.dc.b	$80,$BE,$9D,$F7,$18,$6D,$7E,$A6,$75,$39,$38,$EB,$FE,$94,$9D,$35
	.dc.b	$FF,$08,$4F,$F9,$47,$6D,$92,$3D,$87,$F0,$B8,$FE,$B8,$FC,$16,$BA
	.dc.b	$0F,$37,$0D,$CB,$E6,$3C,$1B,$EA,$7E,$BD,$CD,$F9,$DF,$B1,$BB,$ED
	.dc.b	$5F,$5B,$FE,$7D,$8C,$36,$33,$8F,$F0,$B9,$CB,$61,$F3,$3F,$BD,$CF
	.dc.b	$FD,$4D,$18,$9F,$CD,$35,$FF,$49,$D7,$26,$C7,$EF,$9D,$F7,$F3,$E0
	.dc.b	$FE,$61,$5E,$78,$C4,$FF,$B5,$6B,$A5,$FF,$01,$F7,$DF,$03,$AE,$4F
	.dc.b	$F1,$0B,$63,$37,$FB,$47,$FA,$FB,$7D,$2D,$11,$D4,$F3,$FB,$F7,$D1
	.dc.b	$52,$4F,$EF,$A1,$F3,$D8,$8F,$F1,$57,$C8,$61,$FF,$05,$B5,$FE,$51
	.dc.b	$3D,$1F,$E0,$FF,$7D,$FE,$99,$7F,$9D,$66,$8C,$47,$F3,$A6,$BF,$B2
	.dc.b	$6B,$FD,$59,$FC,$F3,$D3,$A3,$E2,$46,$23,$F9,$C3,$D3,$5F,$E5,$5F
	.dc.b	$7C,$F8,$BF,$94,$BE,$AE,$98,$8E,$7E,$07,$FB,$17,$2F,$33,$70,$FE
	.dc.b	$3B,$8F,$A9,$AF,$F5,$2F,$BF,$B9,$7F,$39,$6F,$37,$F6,$E6,$CF,$F8
	.dc.b	$3A,$9F,$1F,$F7,$7E,$FB,$FA,$C4,$FF,$2E,$5F,$5D,$93,$5E,$97,$E0
	.dc.b	$FE,$FB,$FB,$4D,$7C,$C6,$BF,$D9,$1A,$FF,$1C,$6B,$FE,$01,$F7,$BF
	.dc.b	$83,$FA,$FF,$92,$6C,$7F,$AB,$3C,$4E,$7B,$7F,$B5,$9F,$F6,$34,$7F
	.dc.b	$AC,$3E,$F9,$F4,$DC,$5B,$19,$8C,$7C,$77,$9F,$E7,$3E,$66,$66,$92
	.dc.b	$BF,$BF,$64,$1F,$09,$FA,$6F,$59,$F9,$BF,$C4,$AF,$F2,$DB,$5F,$F7
	.dc.b	$45,$70,$7F,$D8,$BF,$7C,$F4,$FD,$CD,$8C,$C6,$3C,$FF,$85,$F0,$FF
	.dc.b	$ED,$E9,$F1,$B5,$FF,$23,$35,$FF,$01,$5F,$F9,$63,$E6,$FF,$C2,$D1
	.dc.b	$E9,$E4,$7C,$27,$A7,$B9,$E5,$D3,$E8,$4E,$AB,$FD,$B7,$F3,$7E,$97
	.dc.b	$FD,$31,$E1,$FE,$11,$B3,$F5,$E7,$5D,$1F,$E6,$2F,$37,$C3,$A7,$F9
	.dc.b	$A9,$BA,$FF,$17,$C3,$CE,$D7,$FD,$15,$F3,$7F,$81,$F9,$0F,$F7,$A4
	.dc.b	$7F,$A4,$63,$1B,$FE,$17,$E4,$5F,$05,$49,$5F,$EF,$28,$FE,$F4,$FF
	.dc.b	$BD,$3E,$F9,$F1,$5E,$46,$C7,$DE,$6B,$FE,$71,$6B,$F3,$38,$DF,$65
	.dc.b	$F8,$1F,$BE,$7A,$54,$FC,$C4,$EA,$BF,$D8,$1F,$57,$F8,$7F,$BE,$DC
	.dc.b	$7F,$A8,$1A,$B8,$FF,$BF,$3F,$8F,$F9,$95,$7E,$8F,$3F,$E3,$53,$E5
	.dc.b	$F5,$3D,$44,$7F,$2F,$F7,$CF,$AA,$FD,$53,$63,$FC,$DD,$8F,$6E,$74
	.dc.b	$E6,$7C,$5E,$7F,$EC,$E3,$F2,$91,$83,$F4,$5F,$74,$FA,$EF,$E8,$F0
	.dc.b	$F9,$CF,$D0,$63,$EA,$8C,$79,$DF,$BB,$BE,$11,$E2,$C7,$FC,$B1,$F7
	.dc.b	$BF,$7A,$3F,$CD,$8D,$7F,$CB,$81,$C3,$C3,$AB,$FC,$29,$1F,$E9,$08
	.dc.b	$E8,$7F,$A2,$DC,$3E,$93,$97,$CB,$7A,$23,$C1,$8F,$C3,$C3,$F8,$3D
	.dc.b	$7B,$F5,$46,$27,$5E,$6F,$AF,$CD,$F6,$7C,$54,$C3,$C3,$87,$08,$F8
	.dc.b	$15,$FD,$9F,$EF,$5F,$63,$49,$7F,$E5,$A8,$FF,$2D,$37,$83,$81,$1D
	.dc.b	$4F,$D0,$FB,$DF,$C7,$47,$B5,$B9,$FD,$16,$F0,$7F,$A1,$7E,$0F,$F9
	.dc.b	$FD,$27,$FA,$83,$F2,$0C,$7B,$9D,$4E,$DD,$F6,$35,$3F,$D6,$5B,$E5
	.dc.b	$E5,$F9,$1F,$53,$F9,$6C,$BF,$1C,$FF,$73,$6F,$C0,$FF,$3D,$8D,$67
	.dc.b	$FD,$05,$D3,$7E,$BF,$46,$E7,$87,$34,$C3,$E7,$DF,$DE,$C3,$BF,$3F
	.dc.b	$5F,$1F,$B7,$1C,$BD,$C7,$E6,$3F,$15,$FE,$81,$1F,$05,$FE,$C2,$DF
	.dc.b	$73,$5E,$BF,$E7,$BC,$5F,$F0,$0C,$FF,$B5,$DA,$FF,$2C,$F8,$9F,$F3
	.dc.b	$34,$60,$FF,$27,$E7,$FA,$36,$F0,$7F,$AD,$DB,$5C,$EB,$E2,$75,$C8
	.dc.b	$F6,$BF,$1F,$AB,$E4,$7C,$34,$9F,$82,$ED,$FF,$20,$FD,$03,$EE,$3F
	.dc.b	$C7,$7E,$AA,$1F,$29,$86,$D7,$FA,$8D,$E0,$D8,$D7,$DD,$C8,$C1,$F8
	.dc.b	$7F,$D2,$BF,$7D,$BF,$EF,$FF,$32,$7D,$67,$F0,$46,$D6,$F8,$7F,$83
	.dc.b	$FD,$FB,$EC,$3F,$D1,$8D,$9F,$EA,$ED,$7F,$DD,$CD,$7F,$81,$AF,$8E
	.dc.b	$E6,$48,$FC,$D4,$71,$C7,$F9,$5E,$CF,$D4,$F8,$57,$FB,$05,$FE,$BC
	.dc.b	$FF,$9A,$1B,$3C,$1D,$C6,$79,$7E,$5C,$3E,$BE,$98,$8F,$CE,$78,$BF
	.dc.b	$9D,$F7,$25,$FE,$5A,$2B,$FA,$6F,$93,$FD,$11,$8C,$7F,$9A,$7D,$47
	.dc.b	$F7,$49,$FF,$08,$B5,$FE,$47,$E6,$23,$11,$FE,$00,$6B,$FE,$2B,$7F
	.dc.b	$B9,$F0,$5F,$87,$FF,$16,$46,$27,$FC,$38,$F5,$BF,$0F,$C0,$68,$93
	.dc.b	$FC,$72,$5F,$7D,$CB,$E5,$24,$E5,$DB,$FC,$A6,$7F,$D0,$2F,$A1,$AF
	.dc.b	$9D,$FF,$02,$98,$7D,$1E,$E7,$98,$8F,$86,$3D,$3F,$2F,$C0,$91,$F1
	.dc.b	$9E,$7E,$0F,$B8,$91,$E9,$8F,$83,$F0,$B2,$3D,$8F,$C4,$F8,$EC,$47
	.dc.b	$F5,$FF,$8B,$7E,$EE,$FC,$1F,$FC,$80,$D9,$E2,$7C,$5F,$B5,$F5,$91
	.dc.b	$FB,$2F,$BB,$EB,$3E,$89,$23,$81,$F3,$6B,$EA,$FC,$B4,$8C,$1F,$47
	.dc.b	$CB,$F8,$E4,$7F,$0A,$FC,$AF,$2F,$EE,$92,$38,$1F,$D9,$3B,$BF,$D2
	.dc.b	$1E,$8F,$F0,$4F,$51,$EA,$53,$11,$9F,$53,$FD,$B1,$3F,$EA,$BB,$1F
	.dc.b	$EC,$87,$0D,$8F,$B7,$F6,$BF,$8E,$24,$78,$63,$FC,$91,$D5,$F9,$EF
	.dc.b	$A3,$FC,$9F,$C5,$CA,$FC,$74,$7E,$5B,$E8,$7B,$C3,$E8,$3F,$22,$FF
	.dc.b	$96,$32,$FF,$11,$C6,$35,$FF,$56,$7C,$37,$08,$DF,$85,$FA,$7C,$31
	.dc.b	$DF,$E9,$FA,$7E,$BE,$1F,$6C,$7F,$D6,$91,$8C,$F4,$7C,$09,$D6,$69
	.dc.b	$D4,$FF,$A1,$7F,$14,$FF,$0B,$E7,$E0,$6E,$3F,$E9,$D6,$CF,$F9,$C3
	.dc.b	$81,$F3,$BE,$48,$F9,$BB,$FD,$47,$E7,$E1,$F0,$14,$65,$F8,$98,$7C
	.dc.b	$68,$C4,$EF,$F0,$2F,$EB,$B5,$A3,$EC,$FC,$8F,$F2,$D9,$E8,$FF,$1E
	.dc.b	$6F,$9F,$E3,$94,$EF,$0F,$A6,$FF,$21,$DF,$FD,$1A,$D9,$FF,$1B,$9C
	.dc.b	$3C,$DF,$E2,$6D,$9D,$F9,$BF,$02,$37,$5F,$E9,$4D,$32,$48,$CF,$C4
	.dc.b	$FF,$99,$33,$F3,$FC,$8D,$CF,$C7,$36,$BD,$B3,$5F,$8F,$C7,$77,$39
	.dc.b	$DF,$FD,$BE,$FC,$EC,$3E,$53,$CF,$95,$C5,$F5,$A3,$58,$7D,$FD,$3C
	.dc.b	$1B,$BF,$55,$FE,$C3,$5A,$C3,$8D,$6B,$8F,$F1,$58,$7F,$99,$70,$46
	.dc.b	$2F,$F2,$4D,$9F,$D7,$CE,$9F,$F4,$2C,$FF,$AF,$B1,$7C,$AF,$8B,$F4
	.dc.b	$B7,$47,$A6,$73,$5E,$8C,$15,$A8,$F3,$4F,$FB,$F1,$FF,$2C,$BF,$E8
	.dc.b	$96,$FC,$6F,$FA,$53,$B3,$FD,$11,$39,$CB,$FC,$CC,$FF,$B6,$EB,$8F
	.dc.b	$CC,$6D,$F4,$EE,$FB,$0C,$5F,$E9,$EA,$E7,$E4,$EE,$FF,$44,$4F,$C0
	.dc.b	$D3,$FD,$7B,$8F,$E3,$2B,$4D,$8C,$CF,$FA,$1A,$7D,$BD,$FD,$27,$F0
	.dc.b	$33,$FE,$99,$9C,$D7,$F2,$39,$CE,$9E,$CF,$99,$F2,$F1,$FE,$01,$BB
	.dc.b	$7A,$6F,$E2,$C6,$35,$F3,$E5,$FC,$05,$B7,$F8,$86,$DF,$19,$FE,$F2
	.dc.b	$FE,$19,$F9,$3D,$FD,$BD,$BE,$EF,$83,$E2,$E6,$5F,$53,$E2,$ED,$FD
	.dc.b	$07,$A9,$23,$F6,$39,$7E,$62,$3C,$3F,$01,$F4,$3C,$EF,$EB,$7E,$57
	.dc.b	$B2,$31,$E7,$23,$18,$7E,$0A,$7E,$0F,$C0,$63,$2F,$BF,$F0,$5F,$10
	.dc.b	$F8,$BF,$81,$DB,$FC,$4B,$F0,$08,$F9,$38,$11,$F1,$3F,$03,$F9,$A7
	.dc.b	$CE,$F8,$EB,$5F,$92,$7E,$49,$1E,$4F,$CC,$FA,$BF,$AA,$7D,$6F,$90
	.dc.b	$B3,$F9,$A7,$0F,$E7,$3E,$C2,$E7,$F9,$8F,$98,$FB,$A7,$E0,$3F,$33
	.dc.b	$FE,$0B,$5F,$A2,$8C,$3F,$72,$33,$3F,$E3,$26,$BF,$D5,$B9,$11,$88
	.dc.b	$FE,$CF,$18,$F2,$BB,$79,$5D,$DB,$7F,$EC,$8B,$79,$B9,$8C,$37,$0F
	.dc.b	$CD,$7E,$26,$CB,$F5,$3C,$5E,$4E,$EF,$F2,$FC,$E7,$0F,$F0,$04,$E5
	.dc.b	$FE,$17,$1F,$EF,$77,$FC,$08,$EE,$FA,$ED,$FB,$48,$C5,$FF,$BE,$FE
	.dc.b	$E1,$AF,$EE,$AD,$18,$C3,$FD,$94,$B5,$D3,$99,$FF,$79,$39,$9F,$E8
	.dc.b	$F1,$CA,$7F,$56,$D8,$CC,$EE,$3F,$A6,$FD,$FB,$E4,$FC,$5F,$9D,$3E
	.dc.b	$1F,$C9,$39,$C3,$FD,$B9,$39,$7F,$C4,$5F,$92,$B9,$FF,$88,$BF,$EF
	.dc.b	$48,$CC,$63,$9F,$3C,$3A,$8B,$3A,$3E,$C7,$AB,$F8,$93,$F1,$5E,$B5
	.dc.b	$8F,$B8,$E1,$FB,$DF,$1B,$76,$71,$7F,$54,$7F,$C5,$BF,$83,$7F,$4F
	.dc.b	$73,$5F,$E2,$28,$C4,$EE,$51,$FD,$32,$30,$EA,$78,$4F,$CC,$8F,$EA
	.dc.b	$2F,$F8,$B0,$E2,$64,$FA,$3B,$7E,$4A,$7F,$C7,$F5,$DE,$3F,$D4,$11
	.dc.b	$83,$FC,$6A,$EE,$7C,$E7,$F9,$19,$DB,$8E,$9F,$D2,$FB,$3F,$AF,$F6
	.dc.b	$BF,$60,$DB,$FD,$17,$A9,$FB,$33,$F6,$9E,$E5,$8F,$D3,$7B,$9F,$B7
	.dc.b	$F9,$7F,$DC,$36,$3F,$55,$F2,$F8,$69,$C0,$F4,$67,$FE,$5A,$F8,$B7
	.dc.b	$C7,$7E,$6D,$FC,$34,$7A,$83,$A9,$E3,$18,$E9,$E1,$52,$73,$4F,$8E
	.dc.b	$C5,$F3,$8C,$4E,$A7,$F5,$5F,$80,$FE,$CE,$B5,$1F,$C4,$FB,$25,$7F
	.dc.b	$05,$F9,$7B,$5F,$2F,$C5,$47,$F1,$93,$F8,$D7,$CB,$DA,$F7,$F2,$3F
	.dc.b	$02,$FE,$7D,$18,$6E,$1F,$55,$FF,$11,$B7,$CE,$FE,$C7,$18,$E8,$FC
	.dc.b	$B5,$F3,$F8,$7C,$FC,$3E,$9F,$E0,$9F,$7C,$F3,$69,$8F,$0E,$E9,$F2
	.dc.b	$A7,$F6,$1A,$7F,$92,$1A,$FE,$55,$AF,$FA,$09,$AB,$17,$CF,$3A,$7E
	.dc.b	$5F,$89,$FD,$EA,$31,$87,$D8,$E5,$4F,$F7,$EF,$0F,$F8,$14,$FC,$66
	.dc.b	$E3,$F9,$77,$CB,$BF,$C0,$72,$FC,$2F,$96,$FF,$15,$4F,$FB,$D1,$AF
	.dc.b	$F4,$36,$BF,$D0,$9A,$FA,$7F,$01,$FD,$2C,$FF,$01,$F5,$9F,$81,$7F
	.dc.b	$19,$AF,$D4,$56,$B1,$FF,$58,$D6,$BA,$7F,$C2,$FF,$4B,$EC,$9F,$CE
	.dc.b	$CF,$FA,$F3,$7C,$EB,$FE,$01,$F0,$7E,$1E,$5F,$B3,$F4,$BF,$C7,$0F
	.dc.b	$F0,$3E,$2F,$EA,$3F,$07,$FE,$1A,$9C,$CF,$F5,$D6,$BF,$D7,$3F,$02
	.dc.b	$F8,$79,$DF,$33,$FC,$53,$1F,$22,$30,$7E,$4B,$E3,$9F,$2D,$FE,$2E
	.dc.b	$6C,$67,$E5,$F1,$3F,$E3,$2F,$C0,$7F,$C4,$D8,$FF,$6D,$6B,$FD,$B1
	.dc.b	$AF,$E0,$DB,$A7,$F5,$4A,$FF,$10,$7E,$05,$F9,$EE,$6F,$CC,$FA,$89
	.dc.b	$CF,$85,$FE,$40,$ED,$7A,$3F,$E2,$7F,$9F,$FF,$72,$ED,$FA,$A9,$CB
	.dc.b	$F3,$A3,$3A,$FF,$1D,$D7,$E9,$3F,$2D,$AF,$A7,$97,$FC,$8B,$18,$8F
	.dc.b	$F0,$2F,$AD,$FD,$BA,$EF,$B5,$FE,$52,$FA,$CF,$EF,$15,$F9,$FA,$D3
	.dc.b	$FA,$1F,$9D,$7C,$EF,$F8,$F7,$EB,$BC,$3C,$5F,$B6,$9C,$E5,$F8,$49
	.dc.b	$CB,$FE,$14,$BF,$E6,$A7,$F8,$64,$A6,$7C,$3F,$F2,$DB,$83,$F6,$1F
	.dc.b	$2E,$F9,$CF,$F3,$3B,$6B,$F8,$7A,$F9,$CC,$CF,$F9,$B3,$C1,$FF,$47
	.dc.b	$5F,$F4,$36,$E7,$FC,$31,$FF,$46,$37,$E4,$FF,$CE,$73,$F3,$8F,$37
	.dc.b	$83,$E4,$26,$5F,$ED,$0F,$96,$7E,$05,$FE,$07,$C3,$FB,$64,$7F,$89
	.dc.b	$F8,$FF,$D2,$2E,$B8,$AC,$7C,$C8,$7E,$01,$F1,$B2,$FF,$7A,$B6,$1F
	.dc.b	$89,$EE,$31,$F2,$DF,$E3,$FA,$FF,$7D,$B6,$BF,$D3,$39,$F9,$63,$1F
	.dc.b	$11,$FD,$DE,$14,$A7,$EC,$1F,$C3,$34,$F4,$DE,$EB,$53,$9B,$E3,$6C
	.dc.b	$1F,$09,$EE,$6C,$FF,$60,$9D,$77,$66,$B4,$FF,$A6,$A3,$33,$E7,$8C
	.dc.b	$57,$E8,$30,$FF,$04,$53,$F0,$07,$C1,$FF,$A3,$7F,$01,$7C,$97,$CE
	.dc.b	$9C,$E5,$FC,$52,$72,$F8,$E3,$38,$7F,$82,$30,$E9,$87,$1F,$84,$E4
	.dc.b	$FD,$8F,$63,$FD,$B7,$5B,$6F,$07,$DE,$7F,$17,$8F,$C4,$72,$DE,$7E
	.dc.b	$47,$3F,$AF,$B1,$FE,$04,$F9,$9F,$CE,$75,$FF,$B5,$0F,$CE,$63,$FE
	.dc.b	$C9,$7F,$39,$3B,$73,$8F,$F8,$E6,$BF,$49,$D9,$C4,$E5,$F4,$B6,$F8
	.dc.b	$3F,$EC,$89,$FD,$53,$6C,$D3,$FC,$C7,$87,$E1,$BB,$BF,$0F,$F7,$CF
	.dc.b	$9A,$FA,$78,$E1,$FF,$1E,$52,$B0,$FE,$38,$FF,$26,$57,$17,$FC,$85
	.dc.b	$1A,$7F,$17,$55,$F0,$09,$E0,$E8,$C5,$78,$2B,$FC,$F7,$1C,$0F,$F3
	.dc.b	$EB,$87,$C8,$79,$BE,$FF,$F8,$57,$DC,$FF,$A0,$9B,$19,$C9,$39,$6C
	.dc.b	$71,$CB,$FC,$D2,$DD,$79,$46,$23,$FC,$BD,$7F,$32,$72,$7B,$D7,$D4
	.dc.b	$7C,$5C,$FD,$B1,$8A,$D5,$F6,$9C,$B7,$3F,$99,$5B,$AF,$86,$7E,$F1
	.dc.b	$C5,$70,$C3,$E9,$EA,$FB,$FF,$F2,$8D,$FD,$88,$C7,$C5,$7F,$A9,$1B
	.dc.b	$19,$9D,$BF,$E7,$98,$E5,$F1,$3F,$76,$FF,$25,$5F,$E5,$1A,$FC,$E3
	.dc.b	$5F,$F3,$AD,$74,$6F,$E7,$FF,$EA,$B7,$F4,$4D,$73,$EF,$75,$E2,$FE
	.dc.b	$2F,$5E,$09,$FE,$19,$F5,$DF,$DA,$DB,$EA,$6D,$FE,$39,$19,$7F,$7C
	.dc.b	$FF,$58,$7F,$8A,$7E,$23,$FC,$21,$BD,$FC,$7B,$53,$99,$E3,$79,$FC
	.dc.b	$07,$F8,$8C,$7F,$88,$3C,$37,$7F,$87,$CC,$FE,$46,$EF,$C0,$C4,$EA
	.dc.b	$BF,$92,$7C,$CF,$5E,$3F,$E8,$ED,$E0,$FF,$17,$D6,$EB,$F8,$B7,$DF
	.dc.b	$FF,$61,$7F,$A8,$35,$FE,$4A,$D7,$F6,$1B,$23,$11,$FC,$E3,$F0,$07
	.dc.b	$D7,$FB,$DB,$1F,$CD,$DA,$F8,$3B,$BE,$27,$E0,$3E,$58,$FE,$91,$6F
	.dc.b	$6F,$E4,$36,$3F,$A0,$4E,$F1,$F1,$7C,$6F,$FB,$C1,$B3,$EA,$8C,$51
	.dc.b	$D7,$D5,$7F,$C0,$09,$FE,$94,$D7,$F9,$C3,$77,$FE,$DF,$9F,$86,$7F
	.dc.b	$9E,$7E,$02,$F2,$3E,$E2,$F2,$29,$88,$FC,$07,$E0,$1F,$02,$BE,$4B
	.dc.b	$5F,$EB,$ED,$7F,$B4,$37,$E0,$BF,$83,$FD,$97,$E4,$5B,$3F,$EA,$2F
	.dc.b	$25,$39,$7F,$C2,$D1,$FE,$A6,$FC,$05,$F2,$BC,$5B,$19,$8C,$78,$5E
	.dc.b	$6F,$AD,$F9,$59,$1A,$4A,$FE,$67,$DA,$F7,$7E,$14,$7F,$AD,$34,$FF
	.dc.b	$66,$B7,$E6,$7F,$31,$3A,$8F,$F7,$47,$E0,$2F,$AD,$F6,$36,$3F,$D7
	.dc.b	$AD,$77,$F1,$FF,$07,$E1,$3F,$75,$1F,$D0,$71,$46,$23,$FC,$23,$18
	.dc.b	$D7,$FA,$4C,$7F,$AB,$67,$4B,$37,$D6,$3C,$AF,$9D,$F6,$0F,$84,$A7
	.dc.b	$EC,$67,$57,$FE,$94,$F0,$55,$F5,$78,$3E,$0C,$BA,$FF,$D9,$B1,$87
	.dc.b	$FD,$49,$5C,$27,$FD,$B7,$9B,$E7,$BC,$FB,$7B,$1B,$19,$9F,$EC,$73
	.dc.b	$96,$FE,$03,$FC,$77,$0F,$F6,$DB,$7C,$6F,$86,$D9,$FF,$67,$3F,$EC
	.dc.b	$D7,$FD,$D7,$F9,$1F,$F6,$B7,$D1,$3F,$2A,$31,$5F,$EE,$66,$C7,$CE
	.dc.b	$F9,$1F,$F1,$B5,$7F,$BD,$3F,$00,$7C,$AF,$A9,$B1,$E8,$6B,$FE,$B5
	.dc.b	$6B,$FD,$EF,$8D,$F8,$17,$F2,$DC,$BF,$DE,$B6,$FA,$FF,$D1,$36,$FF
	.dc.b	$D8,$B5,$BC,$BF,$D9,$7F,$81,$59,$FF,$76,$78,$69,$FD,$53,$63,$FD
	.dc.b	$9B,$3B,$DD,$A7,$F2,$59,$FC,$73,$5F,$C6,$B5,$FF,$7C,$36,$1F,$81
	.dc.b	$7F,$83,$34,$ED,$18,$F7,$7F,$75,$39,$CB,$F2,$EF,$F0,$9C,$3F,$CD
	.dc.b	$37,$E5,$5E,$87,$EA,$63,$1F,$81,$7D,$99,$FA,$E7,$F9,$C1,$C4,$71
	.dc.b	$9E,$17,$FE,$27,$97,$D4,$7F,$87,$1B,$7F,$A6,$6B,$FE,$30,$6B,$F8
	.dc.b	$5F,$CA,$FF,$9C,$E1,$E5,$6D,$FF,$8C,$DB,$3E,$3B,$FE,$31,$F4,$7F
	.dc.b	$C4,$23,$CF,$A2,$31,$87,$C6,$6E,$1F,$BA,$7F,$31,$2F,$A6,$79,$DF
	.dc.b	$C8,$DB,$87,$C8,$BF,$E0,$31,$88,$EF,$7D,$1F,$F4,$A7,$E0,$7F,$31
	.dc.b	$FF,$45,$35,$FF,$3C,$37,$1F,$F5,$7A,$FA,$6F,$EF,$4E,$70,$FF,$4B
	.dc.b	$AF,$8F,$C9,$AD,$FF,$C1,$18,$3F,$03,$7D,$BE,$1F,$EC,$11,$9F,$80
	.dc.b	$C4,$E5,$D7,$89,$27,$DC,$7F,$AB,$AB,$89,$FC,$69,$CF,$03,$EA,$DF
	.dc.b	$05,$FD,$25,$B3,$C8,$F2,$5F,$F4,$DF,$C0,$FF,$C0,$9C,$1F,$98,$5F
	.dc.b	$33,$93,$5F,$03,$6F,$BC,$E9,$B3,$A7,$3F,$5B,$D4,$8F,$0E,$7F,$07
	.dc.b	$9B,$E5,$F8,$89,$3D,$23,$E4,$FE,$07,$FD,$E6,$3D,$1E,$1F,$FA,$31
	.dc.b	$BF,$1B,$AF,$7B,$E7,$CF,$F9,$66,$7F,$8E,$35,$FF,$44,$35,$FF,$6B
	.dc.b	$B5,$57,$F7,$FD,$BF,$64,$DE,$9F,$F4,$83,$67,$8F,$C7,$7C,$9E,$39
	.dc.b	$1F,$11,$F9,$9F,$5D,$88,$FD,$0F,$27,$F9,$83,$1F,$E0,$AD,$C7,$FA
	.dc.b	$CF,$17,$D6,$AF,$F2,$EF,$D0,$FF,$1F,$6F,$DC,$7F,$94,$FD,$54,$A3
	.dc.b	$0E,$F1,$FF,$5D,$39,$AF,$F6,$23,$5E,$66,$3C,$BF,$CB,$FE,$04,$FB
	.dc.b	$BF,$CC,$FD,$7B,$E6,$E9,$A8,$CF,$C6,$7F,$A9,$A7,$FD,$EF,$87,$F7
	.dc.b	$06,$BF,$EB,$A6,$BF,$D3,$FF,$2F,$7D,$07,$E9,$9B,$1F,$95,$6B,$FC
	.dc.b	$78,$4F,$4D,$3F,$DA,$DF,$81,$FF,$4C,$6F,$BB,$F1,$63,$FB,$D5,$7F
	.dc.b	$B8,$DD,$71,$7F,$A0,$1C,$C7,$FA,$29,$B5,$CA,$38,$74,$7E,$97,$E5
	.dc.b	$BF,$4D,$ED,$C9,$FD,$0C,$57,$F8,$EF,$2F,$F7,$76,$CF,$17,$F6,$BF
	.dc.b	$56,$FA,$5F,$EF,$ED,$8F,$EF,$AD,$7F,$BA,$78,$C8,$C4,$7F,$88,$7A
	.dc.b	$7B,$67,$F8,$AC,$3F,$7C,$7F,$91,$EB,$94,$7F,$52,$7F,$D2,$D5,$FC
	.dc.b	$C9,$B1,$9F,$A6,$E1,$F0,$1F,$E4,$FF,$D4,$7F,$C3,$F7,$FC,$E1,$B5
	.dc.b	$FD,$71,$BC,$1F,$E3,$0C,$39,$B7,$E5,$FD,$A4,$BF,$6B,$FE,$1E,$AF
	.dc.b	$2C,$FF,$8E,$1B,$2D,$11,$87,$E1,$57,$B7,$C3,$F9,$24,$AF,$92,$F5
	.dc.b	$CB,$D4,$48,$FA,$31,$F1,$7E,$0B,$FD,$59,$F7,$CF,$C1,$29,$9A,$FE
	.dc.b	$5F,$F0,$09,$D4,$7F,$2E,$CD,$F8,$17,$C7,$38,$C7,$DE,$F9,$0F,$BA
	.dc.b	$47,$EE,$BF,$07,$87,$E7,$12,$38,$1F,$86,$F9,$7F,$B5,$E5,$48,$C1
	.dc.b	$F8,$AF,$B7,$FC,$49,$1F,$C4,$BF,$37,$8F,$ED,$12,$38,$1F,$D2,$BC
	.dc.b	$7F,$F3,$0F,$A1,$FC,$2F,$D2,$FE,$83,$CD,$FE,$E4,$6D,$7F,$99,$5F
	.dc.b	$E7,$CF,$FB,$C3,$D2,$FF,$2D,$3F,$EF,$96,$A3,$FD,$DA,$E1,$BF,$13
	.dc.b	$FA,$0F,$C1,$3F,$2C,$48,$FD,$2F,$C7,$FE,$67,$E9,$FF,$25,$D5,$27
	.dc.b	$F5,$31,$E7,$F9,$CE,$D0,$FA,$3F,$A6,$F9,$E3,$31,$8D,$BF,$CF,$F1
	.dc.b	$8E,$5F,$CC,$7C,$15,$BF,$F4,$16,$AF,$9B,$C6,$1F,$EF,$26,$C6,$67
	.dc.b	$DD,$8F,$5D,$DF,$73,$CF,$C0,$7E,$03,$FE,$6D,$AF,$2C,$7E,$73,$D7
	.dc.b	$D7,$61,$E1,$5B,$DB,$DB,$C7,$F4,$30,$F6,$E6,$8C,$4E,$B9,$BD,$52
	.dc.b	$FC,$11,$D3,$8F,$8F,$FE,$75,$3E,$8B,$F9,$31,$FC,$DB,$1E,$AD,$8C
	.dc.b	$FA,$7C,$7C,$EF,$E0,$77,$C1,$AF,$DF,$AF,$F5,$47,$27,$FA,$C1,$B1
	.dc.b	$FE,$8D,$0F,$FA,$1B,$93,$5E,$6F,$CB,$9F,$F4,$94,$E5,$E0,$D7,$3F
	.dc.b	$5D,$FE,$79,$CF,$CF,$F8,$05,$D3,$D7,$36,$BC,$93,$5F,$91,$DD,$78
	.dc.b	$7F,$D0,$8F,$C3,$8F,$CC,$70,$3C,$0D,$4F,$3E,$6F,$B7,$B5,$E6,$79
	.dc.b	$19,$E6,$FC,$45,$F3,$BF,$E3,$33,$E3,$E2,$7E,$09,$FD,$B6,$BC,$11
	.dc.b	$8F,$5B,$5D,$9F,$A9,$9C,$D6,$F6,$FC,$F7,$17,$EB,$70,$7E,$09,$FE
	.dc.b	$64,$BE,$6D,$AD,$9F,$CA,$F4,$FF,$6B,$CE,$BB,$BD,$1F,$25,$FE,$A3
	.dc.b	$F3,$BF,$9E,$79,$B6,$BF,$DD,$76,$AC,$61,$FC,$D7,$85,$78,$E1,$FC
	.dc.b	$FF,$CF,$DC,$0D,$F9,$8C,$3C,$B6,$F8,$7F,$35,$FE,$9F,$CB,$D9,$C2
	.dc.b	$8C,$47,$FB,$D6,$31,$AF,$E2,$EB,$8B,$78,$3E,$ED,$7F,$A9,$27,$E2
	.dc.b	$F8,$DD,$E7,$3C,$1F,$EC,$19,$CD,$FF,$3B,$FA,$BF,$57,$01,$F5,$7C
	.dc.b	$A6,$1B,$3F,$EF,$2D,$D3,$E6,$E0,$FD,$FB,$63,$33,$BF,$1F,$E2,$6F
	.dc.b	$F1,$5B,$1F,$D2,$FE,$97,$C2,$7C,$9F,$8B,$B1,$FE,$7C,$B4,$75,$24
	.dc.b	$78,$DF,$8F,$E9,$FF,$57,$F4,$3D,$7E,$97,$D8,$8C,$57,$F5,$98,$FA
	.dc.b	$DF,$03,$9F,$51,$78,$AF,$EC,$31,$87,$FB,$07,$AB,$F1,$8F,$BD,$F6
	.dc.b	$96,$3C,$CF,$C1,$23,$F1,$DF,$8D,$C5,$CF,$9F,$FC,$19,$87,$E1,$FE
	.dc.b	$E7,$E8,$B6,$FE,$74,$7C,$B2,$3F,$C2,$F8,$FF,$C0,$BE,$27,$D8,$3F
	.dc.b	$21,$F5,$D6,$7E,$43,$87,$E5,$7C,$85,$C7,$FC,$39,$E8,$7E,$00,$FC
	.dc.b	$C7,$C1,$DA,$FA,$0F,$DB,$23,$F8,$17,$F0,$1E,$EB,$CC,$FD,$13,$67
	.dc.b	$FD,$66,$BC,$DC,$CE,$9B,$3A,$6E,$1F,$99,$E0,$FE,$2A,$D9,$FD,$47
	.dc.b	$E0,$7F,$E6,$87,$FB,$D4,$EB,$BF,$F3,$93,$9A,$E1,$5F,$DF,$E3,$13
	.dc.b	$FA,$1A,$E1,$3F,$DA,$BF,$0C,$FC,$BD,$FF,$9B,$DB,$49,$FE,$DF,$F7
	.dc.b	$3E,$DA,$D7,$7F,$F0,$45,$19,$7F,$BA,$98,$CF,$F7,$6D,$AF,$E8,$8D
	.dc.b	$3F,$A3,$7E,$08,$F8,$1E,$97,$F2,$6B,$FD,$B1,$39,$7F,$DB,$51,$9A
	.dc.b	$FE,$01,$F8,$27,$FB,$DE,$7F,$C3,$CD,$8C,$F3,$7F,$11,$8C,$67,$FC
	.dc.b	$0D,$46,$7E,$63,$4D,$BF,$B7,$EC,$7F,$06,$F7,$FE,$FB,$63,$24,$C7
	.dc.b	$FD,$7F,$F1,$5A,$8F,$F7,$87,$03,$9F,$E4,$36,$3F,$DE,$9F,$82,$7F
	.dc.b	$4A,$BF,$A5,$5F,$36,$31,$1F,$31,$AF,$F8,$89,$AF,$F8,$C5,$AB,$F0
	.dc.b	$46,$3C,$E3,$EC,$46,$2F,$C3,$1F,$9A,$F8,$1F,$E4,$7E,$D5,$9F,$F3
	.dc.b	$03,$F9,$EF,$81,$91,$F9,$EF,$5E,$DF,$FB,$7B,$F0,$0F,$D1,$9F,$A4
	.dc.b	$F9,$6B,$5F,$A4,$6C,$FF,$B9,$1B,$C9,$F8,$F8,$FF,$25,$7C,$47,$DF
	.dc.b	$7F,$D2,$5E,$9F,$E3,$E3,$13,$FE,$93,$8C,$7C,$5E,$7A,$AE,$3F,$E5
	.dc.b	$3A,$E9,$83,$0E,$33,$FE,$4A,$5F,$2D,$C2,$7F,$DA,$11,$89,$E4,$6F
	.dc.b	$0F,$DF,$FC,$B7,$F4,$8F,$72,$CF,$FA,$15,$B7,$FF,$00,$27,$F8,$2F
	.dc.b	$CA,$63,$2F,$77,$92,$F8,$EF,$F7,$76,$3D,$B0,$FF,$51,$46,$15,$70
	.dc.b	$F9,$DF,$E5,$4F,$29,$D3,$9F,$33,$FB,$A4,$EA,$FF,$CE,$98,$23,$FC
	.dc.b	$3D,$1D,$E3,$13,$F9,$4F,$16,$4F,$8F,$17,$DF,$3C,$CC,$C7,$FB,$D6
	.dc.b	$31,$3F,$BC,$6C,$7F,$90,$5A,$FF,$04,$6B,$FE,$7A,$FC,$11,$7F,$C2
	.dc.b	$E7,$F6,$B5,$98,$C7,$87,$FE,$0F,$ED,$7E,$0B,$4C,$78,$96,$FD,$9F
	.dc.b	$28,$E9,$FE,$EC,$D8,$FE,$94,$DD,$3F,$DD,$ED,$7F,$D1,$BE,$B3,$F0
	.dc.b	$57,$37,$F8,$7F,$F0,$57,$C6,$FF,$6D,$9C,$E3,$F2,$E7,$2F,$FB,$2B
	.dc.b	$E3,$DE,$07,$F8,$9F,$F0,$57,$CD,$FF,$75,$9C,$D7,$D3,$9C,$BF,$57
	.dc.b	$1F,$F1,$93,$FD,$85,$BA,$7F,$2E,$E8,$F1,$FC,$4F,$F8,$4E,$73,$3F
	.dc.b	$EA,$69,$7E,$0B,$F2,$2F,$F0,$0D,$9F,$C6,$C7,$E1,$A3,$F9,$1B,$AE
	.dc.b	$3F,$F1,$F7,$C8,$BE,$5E,$98,$F9,$7B,$5F,$8B,$FC,$13,$FC,$6F,$8F
	.dc.b	$95,$AF,$8D,$AF,$E5,$1A,$B9,$1F,$82,$FE,$66,$1A,$7F,$4E,$C3,$FC
	.dc.b	$39,$39,$8C,$61,$FE,$F8,$F9,$E7,$0F,$F9,$57,$B5,$E9,$FF,$88,$E7
	.dc.b	$35,$FA,$19,$CB,$E9,$E4,$FF,$1E,$6B,$F4,$B4,$F5,$46,$0F,$F2,$07
	.dc.b	$B6,$FA,$4F,$F3,$13,$6F,$F5,$8F,$F7,$07,$F6,$2E,$6B,$E7,$9E,$FF
	.dc.b	$B1,$FD,$0C,$FE,$CF,$E7,$5F,$29,$FE,$39,$9C,$D7,$0D,$D5,$AB,$FD
	.dc.b	$D4,$FE,$40,$FC,$21,$4F,$C1,$14,$FE,$06,$53,$E6,$FC,$CF,$FA,$72
	.dc.b	$3F,$82,$36,$3C,$0F,$FA,$02,$33,$F2,$FF,$E8,$1E,$4F,$F4,$6D,$79
	.dc.b	$5B,$E5,$7F,$1E,$7F,$D0,$75,$FC,$3D,$CB,$E2,$F4,$FF,$CD,$54,$69
	.dc.b	$23,$FD,$95,$F3,$0F,$84,$F9,$95,$FE,$F7,$FC,$17,$FC,$36,$FF,$BE
	.dc.b	$63,$32,$9C,$AF,$C2,$DF,$31,$FE,$9C,$8C,$37,$F0,$5F,$9F,$A7,$F5
	.dc.b	$39,$CE,$5F,$CD,$57,$E0,$DF,$E2,$FC,$C4,$8C,$E7,$FE,$9A,$6D,$7A
	.dc.b	$AB,$85,$6F,$F0,$6F,$F5,$C6,$1D,$FC,$BF,$3C,$62,$FA,$47,$AB,$2F
	.dc.b	$E4,$B6,$C5,$18,$7F,$90,$76,$7A,$FC,$EF,$75,$FE,$7A,$FF,$C6,$B5
	.dc.b	$AC,$3E,$6C,$E6,$7D,$67,$57,$0F,$BB,$85,$F8,$37,$FA,$52,$BE,$27
	.dc.b	$E1,$4F,$A8,$FF,$48,$4E,$6F,$F3,$F8,$AB,$57,$F2,$FF,$07,$7C,$17
	.dc.b	$0F,$96,$FA,$91,$89,$FE,$C9,$9F,$82,$7E,$C7,$C1,$BD,$CA,$7D,$BF
	.dc.b	$67,$F1,$BA,$7C,$59,$D6,$1F,$66,$73,$A7,$F5,$BF,$A6,$E1,$F4,$1C
	.dc.b	$5F,$F6,$B7,$E0,$9C,$31,$FF,$66,$37,$9F,$FD,$95,$F4,$5F,$A3,$9C
	.dc.b	$E9,$FE,$DD,$8C,$4E,$9C,$7C,$FE,$5B,$A7,$FB,$6D,$FF,$68,$36,$3F
	.dc.b	$DC,$3F,$41,$FE,$EE,$7F,$CD,$38,$7E,$3B,$C2,$FC,$7F,$E0,$DF,$ED
	.dc.b	$28,$FA,$FB,$7F,$8A,$63,$11,$F6,$A7,$7F,$29,$F8,$0C,$6C,$FF,$9D
	.dc.b	$A6,$BF,$99,$8A,$F7,$15,$D1,$FF,$2B,$35,$4F,$F9,$F1,$E0,$7F,$9F
	.dc.b	$9C,$3A,$3F,$D0,$A3,$F0,$76,$6F,$C0,$3F,$D0,$B1,$86,$C6,$67,$8C
	.dc.b	$E5,$BE,$37,$F8,$99,$BB,$72,$8C,$47,$FA,$52,$51,$88,$FC,$FC,$62
	.dc.b	$7E,$1A,$FA,$9F,$D1,$E5,$F3,$23,$15,$AB,$F0,$4E,$5B,$A7,$4B,$E2
	.dc.b	$DA,$E1,$87,$D2,$FC,$61,$FA,$E3,$E2,$FF,$D4,$B1,$86,$C6,$6B,$E4
	.dc.b	$CE,$5B,$FB,$7B,$ED,$3F,$73,$87,$F9,$AE,$93,$9A,$FF,$27,$46,$2B
	.dc.b	$7C,$3F,$C6,$A3,$A4,$FF,$56,$78,$FD,$3E,$A1,$18,$79,$DA,$75,$3D
	.dc.b	$67,$BF,$33,$DA,$CB,$FE,$0B,$8C,$FB,$78,$9C,$BA,$F4,$E4,$E8,$F4
	.dc.b	$FC,$57,$F2,$16,$9E,$07,$FD,$A3,$AF,$C1,$75,$1E,$B7,$3A,$7F,$A1
	.dc.b	$7C,$7B,$3F,$E6,$E6,$E6,$AC,$CB,$F0,$7F,$73,$FE,$DA,$6B,$F9,$C6
	.dc.b	$BF,$EE,$26,$BF,$CB,$17,$D6,$7C,$B8,$E8,$D8,$CC,$7F,$9C,$5E,$B8
	.dc.b	$FD,$BF,$B3,$7D,$1F,$CE,$5F,$1F,$4C,$47,$D2,$FC,$DB,$F5,$B8,$7E
	.dc.b	$21,$AF,$FB,$DB,$34,$62,$3F,$A0,$C6,$3B,$3F,$27,$3F,$A8,$9F,$C4
	.dc.b	$B5,$FC,$33,$71,$FF,$5D,$0F,$C7,$5E,$2F,$E1,$3C,$CF,$F1,$9D,$FE
	.dc.b	$49,$AF,$E3,$DB,$5F,$DC,$7F,$07,$FE,$EF,$3F,$98,$EB,$F7,$BF,$A2
	.dc.b	$6B,$EB,$6B,$EA,$6B,$F9,$DF,$CD,$BE,$5F,$25,$FA,$FF,$CB,$C1,$BD
	.dc.b	$5F,$9A,$FD,$8D,$F3,$94,$98,$7F,$AA,$7F,$09,$FE,$C0,$F9,$7B,$3F
	.dc.b	$73,$9F,$E7,$E3,$CD,$18,$3F,$90,$AF,$29,$4F,$82,$FD,$74,$7F,$98
	.dc.b	$FD,$BB,$E9,$7F,$88,$6F,$FD,$8D,$3C,$2B,$7E,$63,$F0,$9A,$4A,$FF
	.dc.b	$5F,$7E,$11,$F8,$C7,$FC,$A3,$6F,$D0,$5E,$87,$92,$30,$D8,$CC,$FF
	.dc.b	$38,$9C,$B7,$F3,$67,$D9,$FF,$00,$F8,$A7,$D8,$D3,$D7,$3A,$AF,$4F
	.dc.b	$D6,$FF,$13,$7C,$6D,$EB,$75,$7F,$8B,$32,$4E,$63,$FD,$B7,$F4,$5F
	.dc.b	$E6,$6B,$F8,$7A,$7F,$BF,$5F,$33,$5F,$F0,$D9,$F3,$DC,$3F,$E1,$EB
	.dc.b	$F8,$53,$F0,$7E,$55,$F6,$7F,$C6,$BF,$3B,$F0,$5F,$F2,$06,$DF,$FB
	.dc.b	$0A,$7E,$BF,$CD,$FF,$86,$AB,$EA,$B5,$F9,$ED,$D3,$FE,$00,$FE,$13
	.dc.b	$FC,$51,$F7,$CF,$E0,$FF,$03,$FE,$7C,$FA,$30,$46,$23,$D0,$D7,$E3
	.dc.b	$35,$FC,$67,$5F,$C1,$AF,$E8,$5F,$45,$F5,$27,$FA,$0B,$5F,$C1,$CE
	.dc.b	$F7,$FA,$1F,$0B,$F5,$E7,$F0,$ED,$BF,$3E,$3F,$E4,$2E,$5F,$E8,$9F
	.dc.b	$22,$F9,$CF,$CD,$61,$F9,$6D,$3F,$D6,$6F,$E2,$5A,$FB,$5A,$FB,$3E
	.dc.b	$63,$F2,$5F,$A7,$FE,$A6,$3F,$A9,$7B,$9F,$F0,$01,$FE,$B2,$D7,$F3
	.dc.b	$71,$BC,$FC,$BE,$EF,$F1,$3D,$BF,$CB,$7F,$11,$FB,$AB,$FF,$3B,$D7
	.dc.b	$D0,$51,$97,$F4,$B5,$C5,$FF,$3E,$FE,$0D,$FF,$00,$1F,$DC,$B5,$FF
	.dc.b	$0D,$B5,$FC,$7B,$5F,$F3,$5B,$F9,$A7,$C3,$78,$EA,$FE,$85,$39,$9F
	.dc.b	$F6,$28,$7B,$61,$E7,$FC,$11,$F1,$1F,$D2,$5B,$5D,$0F,$F4,$9B,$87
	.dc.b	$FD,$91,$B7,$F1,$67,$AC,$3E,$37,$91,$23,$90,$8E,$57,$E1,$F8,$1F
	.dc.b	$01,$27,$7F,$55,$FC,$CA,$75,$1C,$A9,$F1,$9E,$73,$84,$7F,$12,$FD
	.dc.b	$65,$F1,$9E,$BE,$59,$3F,$C7,$A7,$2F,$38,$F8,$1E,$B4,$8C,$1D,$DF
	.dc.b	$89,$52,$30,$7B,$3C,$17,$C6,$FF,$71,$5F,$65,$4C,$45,$9E,$11,$5F
	.dc.b	$AB,$3C,$9F,$E6,$7F,$BC,$D7,$F9,$6B,$75,$FD,$3B,$79,$7F,$D8,$FD
	.dc.b	$7F,$D8,$2B,$9F,$DE,$F8,$FC,$B9,$1E,$78,$FF,$03,$7C,$0F,$EC,$6F
	.dc.b	$FD,$6A,$BF,$02,$C9,$AF,$03,$D9,$F1,$1F,$E7,$3D,$7F,$C5,$D8,$FF
	.dc.b	$9C,$1A,$FF,$24,$6B,$FE,$C0,$7F,$96,$3E,$EF,$88,$F8,$C7,$47,$F0
	.dc.b	$BF,$54,$E8,$FF,$0D,$36,$33,$1F,$D6,$BE,$89,$F5,$7A,$8D,$FC,$B4
	.dc.b	$AD,$9D,$3D,$CE,$7E,$3C,$8C,$1E,$07,$0F,$E9,$B1,$FE,$17,$CB,$FB
	.dc.b	$B7,$F7,$4F,$F1,$6F,$A2,$7D,$4F,$F6,$CF,$D4,$BF,$79,$E5,$9B,$3A
	.dc.b	$C3,$DB,$E7,$7F,$A6,$F1,$34,$91,$FD,$EA,$3F,$6B,$CD,$FE,$E8,$6C
	.dc.b	$03,$86,$CF,$F9,$07,$F0,$B7,$DB,$5F,$22,$57,$EC,$A7,$2D,$8C,$BF
	.dc.b	$C1,$A3,$F6,$B8,$7F,$BE,$1D,$C6,$7D,$3E,$DE,$3C,$8E,$91,$5E,$51
	.dc.b	$77,$FE,$49,$5F,$D8,$A7,$FC,$1F,$C5,$FE,$5A,$7F,$C3,$0D,$DF,$FE
	.dc.b	$00,$7C,$1F,$F8,$26,$35,$F1,$6F,$D9,$3F,$C1,$37,$FE,$8B,$6E,$3F
	.dc.b	$E9,$56,$E9,$FD,$FE,$FF,$C6,$B5,$FE,$25,$FC,$DF,$FC,$38,$FF,$88
	.dc.b	$1F,$F8,$00,$DF,$94,$FF,$13,$C6,$A3,$FC,$3A,$E9,$FF,$4D,$7E,$F6
	.dc.b	$E4,$FF,$4A,$B6,$39,$1F,$E2,$CF,$C5,$3E,$03,$FD,$C5,$AF,$F8,$1D
	.dc.b	$AF,$F8,$55,$AF,$FA,$93,$F2,$87,$D4,$7F,$A8,$9B,$1F,$EB,$06,$BF
	.dc.b	$E3,$29,$EF,$AF,$F6,$CF,$B7,$FF,$36,$4F,$F8,$55,$AF,$F8,$8D,$AF
	.dc.b	$F9,$25,$BB,$BF,$60,$7D,$D7,$FA,$C9,$B5,$FE,$C0,$6B,$FE,$42,$AF
	.dc.b	$D1,$FE,$71,$FD,$17,$0F,$EF,$AF,$CB,$6F,$C4,$7F,$95,$23,$5C,$5F
	.dc.b	$E3,$57,$4D,$E1,$D1,$B9,$FF,$64,$7C,$5B,$F2,$8F,$F0,$B3,$7E,$3F
	.dc.b	$FC,$C6,$DC,$FF,$C6,$E1,$C3,$FE,$3A,$79,$72,$F4,$AF,$1C,$ED,$B3
	.dc.b	$C3,$3F,$F1,$6A,$FD,$1F,$F4,$B1,$FE,$01,$E6,$FF,$1E,$D9,$FF,$1F
	.dc.b	$D6,$FB,$BF,$D6,$D1,$FD,$6A,$7F,$CF,$2D,$7C,$ED,$7F,$C9,$ED,$7F
	.dc.b	$81,$1D,$7A,$79,$91,$FE,$DF,$1C,$EF,$FC,$29,$F9,$CB,$DF,$FF,$63
	.dc.b	$3F,$E1,$DE,$0E,$6D,$BF,$E1,$CE,$DF,$EF,$2F,$F3,$A7,$FC,$C2,$D7
	.dc.b	$8F,$95,$FE,$92,$51,$97,$9F,$C4,$78,$08,$F4,$F3,$7C,$7F,$61,$F1
	.dc.b	$3D,$89,$3E,$08,$FF,$4C,$FE,$19,$FE,$7D,$FA,$8F,$F3,$B7,$86,$8C
	.dc.b	$67,$FE,$7D,$6D,$BD,$77,$D5,$FF,$BE,$30,$E3,$5F,$3D,$BE,$FF,$CC
	.dc.b	$6D,$FF,$2F,$77,$C7,$9C,$7F,$D1,$5E,$7F,$EF,$E9,$5F,$E8,$9E,$6F
	.dc.b	$F6,$AF,$AB,$B9,$FC,$87,$BB,$FE,$61,$AF,$4D,$F1,$7F,$B2,$FB,$BF
	.dc.b	$E5,$A6,$C7,$1A,$FF,$06,$B7,$F0,$D7,$E1,$39,$FB,$87,$E1,$D7,$86
	.dc.b	$61,$E9,$DB,$F0,$F5,$A8,$C1,$FD,$2D,$FE,$A7,$C7,$FE,$CA,$6C,$7E
	.dc.b	$A1,$AF,$FA,$81,$AF,$4F,$6D,$F2,$DC,$5B,$1C,$5A,$F1,$6B,$E4,$F9
	.dc.b	$7F,$F4,$BE,$DF,$E0,$6B,$FF,$5A,$F6,$79,$A9,$1F,$ED,$B8,$C3,$A8
	.dc.b	$FA,$39,$7E,$93,$83,$FD,$7F,$5C,$5B,$3A,$3F,$B5,$6E,$FA,$2F,$0D
	.dc.b	$F1,$BB,$5F,$EE,$BF,$51,$C5,$8E,$57,$C8,$BF,$02,$DF,$12,$F7,$8F
	.dc.b	$CE,$BD,$D6,$BF,$17,$7E,$01,$4C,$7B,$D6,$FE,$AE,$2A,$7E,$27,$C4
	.dc.b	$F2,$F7,$FE,$73,$C1,$C5,$FB,$FD,$6F,$F6,$0F,$8F,$FE,$93,$E5,$79
	.dc.b	$F8,$AF,$8D,$57,$8F,$FC,$19,$5A,$9F,$F3,$BB,$F8,$C6,$F6,$EE,$7F
	.dc.b	$CB,$FD,$B6,$B5,$1F,$E8,$EF,$C7,$7F,$C2,$37,$FC,$C2,$11,$87,$F9
	.dc.b	$9B,$5F,$F7,$5B,$5F,$E3,$AF,$FA,$2E,$3F,$D1,$2F,$FA,$59,$BF,$1D
	.dc.b	$E3,$8C,$4F,$87,$DD,$C1,$9E,$6B,$7C,$9F,$F5,$B4,$F6,$6B,$FE,$2D
	.dc.b	$AF,$B5,$3E,$96,$C7,$F4,$7F,$9A,$7C,$AF,$87,$C7,$F9,$5D,$29,$CE
	.dc.b	$1E,$89,$CF,$E6,$1F,$DE,$A3,$FD,$59,$48,$C6,$0E,$2F,$E9,$B0,$FB
	.dc.b	$5E,$3C,$5C,$9F,$3D,$F1,$6C,$EF,$0F,$EB,$F3,$96,$DF,$E5,$DF,$0F
	.dc.b	$B6,$F3,$3D,$33,$9C,$BC,$55,$FE,$4F,$B5,$EE,$38,$F5,$70,$ED,$F8
	.dc.b	$32,$8C,$4F,$FB,$26,$3F,$33,$E0,$FE,$3D,$B1,$FE,$CD,$97,$DD,$78
	.dc.b	$72,$4F,$F9,$7E,$B7,$97,$90,$ED,$5E,$77,$FB,$1D,$F2,$C7,$ED,$B6
	.dc.b	$FF,$C0,$2D,$9F,$ED,$27,$B3,$C3,$7C,$6F,$95,$B5,$E5,$F6,$3F,$8E
	.dc.b	$BE,$B8,$CD,$78,$ED,$5A,$8E,$B1,$8F,$2B,$C7,$39,$9F,$3B,$5F,$F3
	.dc.b	$E4,$BF,$39,$E7,$87,$C9,$FD,$F1,$ED,$7E,$DE,$B5,$87,$CB,$9C,$D7
	.dc.b	$CB,$BF,$94,$F4,$79,$E1,$E7,$72,$FD,$77,$3B,$FD,$08,$FB,$FD,$7F
	.dc.b	$EA,$EA,$FB,$4D,$8F,$81,$F9,$D7,$FB,$7A,$73,$F1,$BE,$58,$C4,$FC
	.dc.b	$26,$BF,$59,$AF,$C0,$6B,$FE,$9D,$FC,$3B,$FD,$85,$5F,$E9,$FC,$3C
	.dc.b	$92,$9C,$BF,$83,$31,$AF,$F1,$58,$CC,$F0,$7E,$2C,$ED,$CD,$7F,$8A
	.dc.b	$9C,$CE,$FE,$3B,$97,$C7,$4A,$F1,$1B,$F3,$71,$87,$36,$DF,$76,$CF
	.dc.b	$5F,$8C,$F2,$1E,$5F,$4E,$C7,$4B,$F8,$11,$FC,$49,$F7,$37,$23,$4F
	.dc.b	$CE,$F2,$3D,$5F,$C2,$7C,$97,$F5,$BC,$3F,$59,$39,$7F,$5D,$19,$C3
	.dc.b	$E4,$7E,$19,$FC,$CF,$4F,$9A,$DA,$FF,$26,$BF,$CB,$23,$3F,$01,$FD
	.dc.b	$06,$31,$3A,$8C,$7C,$76,$5D,$79,$1B,$E5,$B5,$EF,$C5,$82,$73,$1E
	.dc.b	$4A,$FC,$F7,$E1,$DF,$EA,$DB,$E1,$7F,$CC,$B2,$F3,$FE,$01,$E3,$3C
	.dc.b	$9F,$31,$6B,$6F,$CB,$A3,$49,$1F,$D5,$7F,$0D,$FE,$B3,$7F,$8E,$7F
	.dc.b	$C8,$F1,$E7,$72,$FF,$9C,$1C,$E5,$F9,$1F,$C9,$DE,$7F,$F6,$55,$F7
	.dc.b	$16,$30,$FE,$E7,$E0,$1F,$49,$4F,$EE,$5E,$BF,$F9,$5B,$9B,$19,$FC
	.dc.b	$7C,$57,$D4,$ED,$E2,$D9,$D3,$73,$F2,$37,$5F,$D4,$DF,$82,$BA,$7C
	.dc.b	$CF,$9F,$3F,$43,$E9,$F8,$5C,$4F,$85,$B8,$FF,$1F,$E8,$FE,$7D,$39
	.dc.b	$3F,$1B,$E3,$71,$7E,$F7,$C3,$7E,$BA,$81,$FE,$1F,$1F,$E9,$A7,$51
	.dc.b	$9F,$1F,$FC,$1D,$18,$33,$F7,$1F,$95,$9E,$71,$9F,$7B,$67,$0F,$0E
	.dc.b	$E1,$E2,$70,$FE,$BF,$8D,$F8,$6B,$EA,$B3,$8F,$F0,$BC,$7E,$EF,$17
	.dc.b	$F1,$06,$E1,$E7,$F1,$3C,$BE,$DF,$99,$FD,$C4,$9A,$49,$FE,$A3,$FA
	.dc.b	$3F,$F9,$4A,$FF,$8D,$ED,$FC,$AD,$FE,$61,$C0,$9D,$57,$FA,$FA,$B4
	.dc.b	$FF,$46,$FD,$43,$FC,$49,$9F,$F6,$3F,$49,$FA,$87,$CF,$C5,$5A,$CB
	.dc.b	$F9,$E5,$6B,$CB,$FE,$95,$E1,$7F,$9A,$F1,$FE,$AE,$D7,$FB,$5B,$5F
	.dc.b	$EA,$AD,$7F,$BF,$7B,$9F,$E7,$1B,$F4,$B7,$3F,$BF,$C6,$FA,$BC,$5F
	.dc.b	$0F,$A8,$FE,$B5,$F8,$C7,$F0,$A8,$FF,$06,$FE,$5F,$FC,$7B,$0F,$F0
	.dc.b	$24,$E7,$54,$E6,$BF,$B8,$7E,$5E,$FC,$4F,$FB,$FB,$70,$FE,$FE,$D9
	.dc.b	$FE,$E6,$D7,$FC,$0F,$F9,$0F,$F2,$57,$FB,$D3,$76,$FE,$C5,$AB,$F1
	.dc.b	$2E,$51,$F5,$A7,$58,$7F,$78,$FA,$E7,$83,$FE,$1D,$F8,$F7,$CD,$FF
	.dc.b	$77,$9C,$D7,$F8,$8E,$72,$FF,$88,$7F,$24,$FF,$43,$E9,$FE,$08,$6B
	.dc.b	$FE,$1C,$6B,$FE,$06,$6B,$FE,$2F,$FD,$7D,$5F,$E0,$56,$D7,$E7,$FD
	.dc.b	$8F,$0D,$7F,$85,$E7,$2F,$F8,$67,$F1,$2F,$F2,$43,$F8,$BF,$8A,$FC
	.dc.b	$26,$DF,$94,$9C,$BF,$8A,$8C,$E1,$E3,$E6,$FF,$17,$E7,$FE,$B3,$C3
	.dc.b	$D1,$18,$3F,$C6,$7F,$0F,$FC,$EE,$3F,$C5,$0D,$C3,$E7,$FC,$B7,$EF
	.dc.b	$7B,$BF,$CC,$9F,$4D,$FE,$1F,$FD,$1F,$FA,$C5,$7E,$99,$FF,$70,$E7
	.dc.b	$FE,$49,$72,$FF,$8A,$9C,$E3,$FA,$5F,$C0,$7F,$C0,$2F,$F1,$23,$FC
	.dc.b	$B6,$E1,$FF,$27,$35,$CF,$C0,$FF,$A1,$3B,$1F,$77,$FE,$7C,$9F,$D8
	.dc.b	$37,$1F,$C3,$FC,$D3,$E7,$FF,$CB,$D3,$9C,$FF,$CB,$5F,$9B,$FF,$31
	.dc.b	$C3,$F7,$0B,$D3,$FF,$32,$39,$C7,$F6,$EE,$6B,$FC,$D5,$A7,$D6,$BF
	.dc.b	$56,$68,$C4,$7A,$BE,$43,$D1,$18,$8F,$F3,$CF,$E2,$2F,$93,$FF,$39
	.dc.b	$37,$0F,$F0,$97,$C6,$FB,$BF,$14,$CD,$7F,$9D,$FE,$D3,$FD,$AF,$5F
	.dc.b	$E7,$B6,$E1,$FE,$73,$D5,$F8,$97,$F9,$33,$55,$6A,$7F,$88,$BF,$ED
	.dc.b	$AB,$FF,$38,$DF,$FA,$32,$30,$FF,$9D,$2B,$8F,$CE,$7D,$19,$CC,$FC
	.dc.b	$0F,$D0,$7F,$27,$7F,$E8,$55,$F5,$9F,$E8,$D7,$33,$FE,$98,$73,$3F
	.dc.b	$E1,$1F,$63,$FC,$C3,$3F,$D0,$7E,$53,$FD,$B9,$FA,$4F,$F5,$C8,$FF
	.dc.b	$26,$4E,$A3,$11,$FE,$88,$6B,$F8,$B6,$BF,$D7,$9B,$9B,$E1,$FF,$7F
	.dc.b	$3F,$1F,$EB,$BF,$91,$CE,$7C,$9F,$F7,$96,$E9,$CD,$FF,$95,$33,$FF
	.dc.b	$51,$3C,$A7,$F9,$DB,$9C,$3F,$D6,$AE,$67,$FD,$51,$F9,$C7,$F2,$7A
	.dc.b	$FE,$82,$FF,$5A,$AF,$F5,$DB,$99,$DF,$BB,$FE,$B5,$8F,$EB,$51,$FE
	.dc.b	$B9,$75,$3C,$3D,$CD,$7C,$97,$FA,$EF,$F0,$EF,$D2,$DF,$FA,$FD,$B5
	.dc.b	$FD,$55,$73,$EF,$BB,$83,$2B,$FD,$8D,$F8,$7F,$E2,$6B,$71,$8E,$3F
	.dc.b	$D9,$4F,$FA,$FF,$0F,$F6,$23,$5F,$F6,$AB,$5F,$DF,$78,$3F,$E1,$0F
	.dc.b	$C4,$7F,$DE,$F1,$FC,$07,$B3,$FD,$3D,$18,$8D,$87,$0F,$FA,$ED,$AA
	.dc.b	$7A,$7B,$5F,$EC,$AA,$E6,$FE,$72,$33,$3D,$F1,$AF,$63,$F4,$27,$32
	.dc.b	$94,$CC,$6B,$F1,$27,$E1,$CF,$C4,$F0,$DA,$FD,$5C,$EA,$72,$71,$FC
	.dc.b	$4D,$CB,$49,$5C,$23,$2D,$8F,$3D,$7F,$1B,$8B,$E3,$FF,$29,$D7,$E4
	.dc.b	$4A,$F8,$C6,$1B,$A7,$F2,$F9,$E5,$7F,$36,$31,$7C,$DD,$57,$09,$75
	.dc.b	$38,$5F,$2B,$F2,$23,$0D,$8C,$CF,$F0,$C9,$CB,$61,$F5,$DF,$D0,$74
	.dc.b	$FF,$11,$C6,$27,$F3,$13,$9E,$0F,$F2,$BD,$F2,$9F,$CD,$78,$99,$7C
	.dc.b	$B7,$FC,$AD,$FD,$04,$66,$7D,$17,$7E,$A3,$FC,$0B,$5F,$EB,$F9,$47
	.dc.b	$F7,$88,$C3,$5F,$F1,$E3,$C1,$BE,$BF,$C1,$8D,$47,$F4,$26,$BE,$86
	.dc.b	$BF,$2B,$E1,$BF,$B0,$3F,$CD,$FE,$4E,$E9,$FE,$FF,$3F,$D2,$F1,$E9
	.dc.b	$9C,$E9,$FE,$F1,$AF,$CA,$D6,$AB,$F9,$CB,$5E,$71,$F9,$A5,$F8,$93
	.dc.b	$F1,$3F,$CE,$B6,$3F,$42,$D7,$FA,$1B,$5F,$3A,$FC,$3B,$E8,$47,$46
	.dc.b	$C6,$63,$FD,$98,$F3,$BF,$F2,$FF,$E0,$2F,$C4,$FE,$A7,$E2,$7F,$AA
	.dc.b	$BF,$A4,$D7,$FC,$D5,$3B,$7F,$DE,$95,$FD,$B7,$E6,$7F,$6E,$D8,$F8
	.dc.b	$3E,$8F,$FB,$C5,$B3,$F9,$5F,$CB,$FF,$DE,$F1,$F7,$DA,$FF,$9D,$DB
	.dc.b	$5F,$E7,$4C,$3F,$55,$3F,$85,$FC,$53,$6F,$F7,$E6,$BF,$94,$6B,$FA
	.dc.b	$66,$AF,$96,$FD,$7B,$F8,$06,$BF,$E2,$16,$BF,$EB,$46,$F2,$3F,$13
	.dc.b	$F9,$93,$FD,$CD,$AF,$F8,$41,$AF,$E9,$9A,$FE,$E3,$F1,$37,$D7,$FE
	.dc.b	$DD,$B1,$FB,$56,$BF,$DF,$DA,$FE,$D7,$C6,$7B,$1F,$41,$B1,$F9,$06
	.dc.b	$BF,$9A,$6B,$FE,$3D,$F7,$9F,$5D,$FE,$38,$8F,$F0,$7C,$63,$CB,$FE
	.dc.b	$62,$D8,$CC,$ED,$FF,$09,$79,$3F,$BD,$6F,$7F,$F1,$18,$C4,$7F,$8C
	.dc.b	$9A,$FF,$93,$7F,$16,$F8,$69,$A4,$E9,$FE,$35,$F8,$97,$FB,$46,$7F
	.dc.b	$8C,$35,$FF,$11,$35,$FF,$1F,$B5,$FE,$2B,$F3,$EF,$A2,$FE,$48,$D8
	.dc.b	$FE,$44,$D7,$FC,$58,$D7,$FC,$93,$F8,$9F,$F8,$D9,$FE,$50,$D7,$FC
	.dc.b	$64,$D7,$7E,$E7,$F3,$1F,$C4,$BF,$97,$C7,$F8,$DD,$AF,$F9,$51,$AF
	.dc.b	$F2,$A6,$BF,$CF,$3F,$0E,$BC,$C6,$93,$C9,$FE,$C7,$F0,$3F,$C0,$EB
	.dc.b	$FC,$F0,$D4,$1C,$37,$C3,$FE,$8D,$EB,$7C,$78,$FE,$96,$DC,$3F,$DE
	.dc.b	$DE,$EE,$8F,$E9,$9F,$9A,$BE,$D2,$92,$FF,$A6,$46,$5F,$E9,$DF,$AD
	.dc.b	$FC,$B7,$FA,$3B,$5F,$EB,$2D,$B3,$7D,$17,$E3,$9F,$D6,$DB,$5F,$D6
	.dc.b	$5A,$FF,$98,$DA,$FF,$3A,$FC,$4D,$F4,$BF,$C9,$3E,$57,$F8,$5F,$9F
	.dc.b	$FE,$9D,$6D,$7C,$6A,$DB,$F2,$3C,$85,$6F,$E4,$3F,$C2,$36,$69,$2B
	.dc.b	$FC,$2F,$0F,$8F,$F0,$12,$3A,$57,$36,$9F,$5B,$95,$F7,$5F,$CE,$9B
	.dc.b	$1A,$3F,$05,$19,$FB,$1D,$3B,$FB,$59,$3B,$AE,$2F,$F8,$72,$FF,$D7
	.dc.b	$91,$FB,$7C,$FD,$B3,$E1,$F0,$BC,$07,$87,$A6,$C7,$3F,$63,$C8,$61
	.dc.b	$E8,$7A,$7E,$BB,$D1,$E8,$C8,$F1,$45,$7F,$D3,$78,$FF,$43,$6C,$FF
	.dc.b	$8F,$23,$11,$FD,$C9,$7D,$E7,$C3,$24,$7B,$67,$FA,$FF,$C2,$FE,$CF
	.dc.b	$3F,$5C,$A3,$1A,$7F,$B6,$DB,$1F,$A4,$7F,$80,$4F,$05,$F8,$0F,$F9
	.dc.b	$39,$B3,$FE,$52,$6B,$FE,$B4,$6B,$FE,$08,$5F,$5F,$C4,$91,$C5,$E9
	.dc.b	$F5,$DE,$8F,$85,$93,$F7,$69,$3F,$DE,$BD,$9F,$F7,$04,$E7,$E7,$BD
	.dc.b	$E6,$7E,$57,$E0,$FC,$0F,$9C,$93,$E7,$7D,$5F,$59,$F1,$89,$1D,$E3
	.dc.b	$FC,$0F,$F8,$73,$E6,$29,$27,$F2,$10,$FA,$AC,$47,$E0,$04,$3E,$4F
	.dc.b	$99,$AE,$1F,$A7,$0F,$B7,$EE,$4A,$EB,$E1,$78,$BD,$CF,$F3,$ED,$7F
	.dc.b	$5E,$9F,$E3,$27,$68,$C9,$FD,$53,$F0,$FF,$F0,$1C,$78,$E5,$E6,$7B
	.dc.b	$0F,$F2,$97,$B2,$AD,$E5,$E4,$E4,$EE,$4C,$3F,$D1,$F8,$7F,$81,$8F
	.dc.b	$DF,$3F,$0B,$C5,$F8,$E7,$C8,$9F,$F1,$7F,$E1,$EF,$12,$92,$BF,$71
	.dc.b	$0F,$B8,$DF,$8B,$FE,$9A,$8F,$DA,$CF,$F9,$41,$AF,$C0,$6B,$FC,$E9
	.dc.b	$FE,$97,$97,$E7,$9B,$DB,$BF,$0F,$C7,$F7,$1E,$4F,$17,$B1,$30,$FF
	.dc.b	$55,$DF,$F7,$27,$FD,$13,$F8,$7F,$E4,$1F,$F4,$23,$5F,$F3,$1F,$9C
	.dc.b	$8C,$47,$D4,$8C,$77,$FF,$81,$BA,$BF,$C9,$99,$FF,$50,$E4,$FC,$67
	.dc.b	$A5,$F4,$A7,$38,$78,$2F,$EA,$FE,$11,$FE,$7E,$8F,$F4,$07,$E1,$EF
	.dc.b	$A3,$FF,$43,$36,$3F,$93,$E2,$AE,$11,$FC,$4B,$35,$6E,$3D,$2D,$7D
	.dc.b	$0D,$7F,$D3,$2D,$7F,$D5,$7F,$87,$FE,$45,$BD,$7F,$EA,$B8,$C5,$1D
	.dc.b	$60,$F9,$DF,$A7,$68,$C7,$1F,$E4,$A7,$37,$C7,$0F,$F5,$17,$E1,$EF
	.dc.b	$A1,$FF,$5C,$37,$1F,$F5,$B3,$5F,$A6,$D7,$FC,$75,$F8,$7F,$FB,$23
	.dc.b	$9F,$F9,$DE,$68,$C7,$1F,$D3,$6D,$7F,$B2,$2F,$A5,$F0,$FC,$3F,$F7
	.dc.b	$53,$F3,$9F,$9B,$19,$F9,$5F,$D5,$C6,$A7,$FA,$0E,$28,$FA,$71,$FE
	.dc.b	$4D,$6B,$F8,$36,$BF,$15,$BF,$0E,$FC,$3F,$F8,$94,$FF,$AC,$8F,$F2
	.dc.b	$AB,$87,$FD,$6B,$1B,$E6,$D4,$69,$B1,$CF,$C7,$FF,$69,$36,$7F,$B7
	.dc.b	$B5,$FE,$B2,$F3,$7F,$B7,$72,$FD,$A7,$F1,$DD,$7F,$EA,$C8,$FE,$D1
	.dc.b	$BF,$F1,$47,$47,$FB,$96,$7D,$0D,$7F,$D9,$6D,$78,$7A,$9E,$87,$F3
	.dc.b	$5D,$5D,$48,$F8,$A3,$FD,$77,$F8,$73,$EF,$AF,$97,$2B,$FC,$09,$39
	.dc.b	$6F,$9D,$D6,$3F,$DC,$91,$FE,$DC,$6B,$FC,$B1,$AF,$F9,$19,$7E,$0F
	.dc.b	$E0,$24,$76,$F3,$BE,$17,$AB,$E9,$D3,$FB,$AC,$FD,$26,$BF,$0A,$3F
	.dc.b	$D1,$3F,$87,$7F,$98,$BE,$87,$FD,$F5,$CC,$8F,$91,$18,$6E,$9F,$C2
	.dc.b	$5E,$0D,$E1,$FE,$87,$C3,$FE,$99,$9F,$F4,$83,$63,$38,$EC,$FF,$30
	.dc.b	$C7,$F4,$AE,$27,$AE,$FC,$07,$FA,$6E,$4C,$BF,$CC,$F5,$A9,$CC,$62
	.dc.b	$B8,$61,$FE,$6C,$BF,$5F,$A5,$FC,$3D,$E5,$1F,$E7,$0F,$8A,$FE,$C0
	.dc.b	$F8,$8F,$EB,$6E,$1F,$F0,$B9,$FE,$95,$9F,$4B,$C5,$AB,$F0,$E7,$E1
	.dc.b	$7E,$48,$FF,$3E,$C3,$EA,$31,$1B,$6C,$70,$8F,$46,$5F,$DB,$23,$FB
	.dc.b	$3F,$07,$EA,$9B,$3F,$B1,$6B,$FD,$C8,$F1,$7A,$BF,$C7,$5D,$7A,$4F
	.dc.b	$91,$E4,$D8,$CC,$60,$ED,$F0,$3F,$E5,$0D,$BE,$3C,$FB,$7D,$5C,$67
	.dc.b	$E5,$AB,$3E,$E9,$AF,$FA,$EF,$8F,$FC,$B2,$79,$78,$3F,$CE,$EF,$CF
	.dc.b	$8F,$A4,$E9,$FA,$2D,$7F,$DC,$3F,$87,$3E,$53,$BF,$2A,$31,$5F,$22
	.dc.b	$BF,$9E,$79,$BF,$E1,$B8,$C5,$6E,$FC,$FC,$2F,$C3,$5F,$87,$77,$8C
	.dc.b	$36,$33,$7E,$79,$CB,$61,$F5,$FD,$17,$E0,$D6,$30,$FC,$4D,$5E,$57
	.dc.b	$E3,$35,$FC,$93,$5F,$F7,$13,$63,$78,$7F,$2A,$5E,$7F,$FA,$13,$CC
	.dc.b	$FF,$35,$C6,$27,$FD,$24,$FE,$49,$B3,$B7,$F3,$BD,$3D,$48,$F5,$7C
	.dc.b	$5F,$87,$7F,$89,$E3,$FD,$38,$FF,$89,$5B,$EE,$73,$CB,$FD,$B3,$39
	.dc.b	$7F,$CA,$18,$7F,$9F,$6B,$C7,$D3,$B9,$53,$DA,$C5,$EE,$7F,$CB,$13
	.dc.b	$9B,$FE,$92,$D8,$FE,$A0,$D7,$F4,$0D,$7F,$CA,$6B,$C1,$F0,$2E,$5B
	.dc.b	$18,$7D,$5C,$AF,$B5,$FC,$45,$FF,$50,$6E,$FF,$EB,$29,$D6,$FC,$27
	.dc.b	$39,$7F,$14,$7E,$95,$FF,$62,$6B,$FC,$75,$AE,$E7,$F0,$54,$F5,$3F
	.dc.b	$9E,$DF,$F6,$D8,$C7,$91,$A9,$CD,$FF,$6E,$6B,$FC,$A7,$1F,$D1,$BF
	.dc.b	$DA,$23,$71,$E9,$32,$7A,$DC,$DF,$F9,$0F,$F6,$2F,$F1,$23,$FE,$3A
	.dc.b	$6D,$7F,$8E,$79,$5F,$23,$F8,$1E,$85,$6B,$1F,$F6,$BE,$3F,$A4,$AF
	.dc.b	$CF,$37,$1F,$F0,$43,$71,$FF,$2C,$E3,$D2,$7F,$C1,$98,$BE,$7F,$FD
	.dc.b	$9B,$1F,$E4,$C6,$D7,$F9,$2F,$F0,$BF,$F5,$DC,$E7,$CA,$FC,$A6,$29
	.dc.b	$CD,$7F,$99,$9A,$FF,$99,$5A,$8F,$A4,$E1,$FA,$9F,$87,$7F,$AC,$6F
	.dc.b	$D2,$D8,$FF,$38,$B5,$FF,$38,$35,$FF,$37,$7E,$1D,$EB,$9E,$DA,$FC
	.dc.b	$6A,$E4,$FF,$91,$27,$55,$FE,$67,$9C,$D7,$F9,$DF,$F0,$EF,$E5,$93
	.dc.b	$FE,$3C,$6D,$FD,$0E,$87,$C8,$7D,$59,$CD,$FB,$73,$FD,$84,$FD,$4A
	.dc.b	$DC,$FD,$46,$E1,$FE,$8C,$6B,$FD,$37,$EF,$FC,$AD,$DB,$F8,$C4,$62
	.dc.b	$3F,$D8,$6D,$7E,$63,$FE,$D5,$AF,$82,$6B,$D7,$FE,$09,$1E,$E9,$DC
	.dc.b	$63,$C0,$EA,$4A,$FF,$25,$D7,$F2,$E6,$FC,$4F,$23,$9F,$9D,$63,$BE
	.dc.b	$1E,$2F,$07,$B9,$E0,$F3,$2C,$EA,$58,$7F,$B5,$A3,$BC,$FB,$88,$FE
	.dc.b	$5B,$E0,$FF,$31,$5F,$EE,$FA,$FE,$35,$C3,$FE,$7E,$9C,$AA,$F8,$BE
	.dc.b	$17,$F2,$3F,$C3,$5F,$6B,$FD,$05,$B3,$FE,$68,$6E,$3F,$E6,$F6,$BF
	.dc.b	$DF,$DF,$E6,$17,$D6,$3F,$35,$18,$9F,$F7,$AB,$FC,$EE,$13,$B5,$1B
	.dc.b	$78,$F0,$F9,$7E,$17,$CC,$DB,$F8,$D5,$FC,$CE,$F9,$36,$3B,$F8,$5C
	.dc.b	$1D,$FC,$B7,$BC,$F8,$1F,$84,$D8,$E6,$FC,$12,$3F,$05,$FE,$C9,$F7
	.dc.b	$1F,$E2,$98,$F0,$FC,$B7,$F4,$B8,$C6,$9E,$C9,$F7,$CF,$F4,$FF,$A1
	.dc.b	$7D,$7D,$F8,$7C,$36,$BF,$01,$3F,$AC,$BF,$F1,$53,$5F,$E0,$81,$C3
	.dc.b	$FC,$1D,$AE,$7E,$7B,$FB,$D4,$1A,$49,$FF,$51,$C7,$F9,$75,$F0,$FB
	.dc.b	$9E,$7F,$97,$F4,$3E,$9F,$95,$93,$E6,$F8,$4B,$8F,$BC,$F1,$3E,$02
	.dc.b	$47,$67,$C7,$F0,$9F,$0D,$F8,$8D,$8F,$D0,$7C,$1F,$F1,$0E,$4F,$C2
	.dc.b	$3F,$D1,$A3,$E0,$B9,$6F,$B1,$FD,$EE,$31,$87,$F2,$96,$A7,$F8,$03
	.dc.b	$E7,$31,$E1,$71,$F0,$79,$79,$77,$0F,$1B,$6B,$85,$FE,$FD,$BE,$16
	.dc.b	$72,$F3,$FC,$07,$8C,$F2,$78,$57,$1F,$6B,$FE,$2D,$C3,$F0,$F8,$7F
	.dc.b	$99,$DA,$FB,$A3,$FC,$65,$F6,$1F,$C8,$B0,$F8,$7E,$1F,$F8,$66,$31
	.dc.b	$87,$C4,$8F,$E9,$FF,$85,$FB,$B0,$FC,$3F,$90,$8C,$47,$F3,$16,$BF
	.dc.b	$DF,$DB,$0E,$1F,$EF,$33,$C9,$AF,$F3,$96,$BF,$E0,$76,$C7,$FB,$47
	.dc.b	$DE,$FE,$DF,$C6,$69,$31,$FE,$E3,$F8,$47,$F8,$BF,$0F,$3B,$5F,$E8
	.dc.b	$CD,$7F,$C2,$8D,$7F,$CB,$3D,$5F,$E5,$7F,$6F,$F5,$B5,$FE,$FE,$6C
	.dc.b	$FF,$A4,$CD,$C7,$F8,$03,$F1,$EF,$F0,$BC,$7F,$53,$6B,$FE,$21,$6E
	.dc.b	$46,$FC,$0B,$E6,$7F,$C1,$6D,$BF,$F0,$5B,$6F,$FD,$45,$7F,$EC,$7A
	.dc.b	$FF,$08,$FE,$31,$FD,$CE,$7F,$C5,$CD,$93,$4C,$79,$8F,$83,$FF,$0C
	.dc.b	$B7,$1F,$F0,$CB,$6F,$BB,$5F,$C0,$FE,$07,$F4,$67,$FC,$3E,$D8,$F2
	.dc.b	$B5,$DE,$FF,$E8,$1F,$BC,$F9,$C4,$F3,$B0,$7F,$8A,$BF,$08,$FE,$D9
	.dc.b	$A1,$A4,$EF,$FF,$46,$7B,$2F,$C0,$DF,$43,$86,$EB,$FE,$2E,$6E,$9F
	.dc.b	$EC,$9A,$E1,$5F,$E9,$5E,$C7,$E3,$4F,$AB,$C3,$71,$FB,$D3,$A9,$C9
	.dc.b	$FE,$40,$FB,$07,$D7,$D2,$57,$F9,$0A,$32,$FF,$91,$3F,$04,$79,$3F
	.dc.b	$E4,$86,$C6,$63,$07,$FA,$9F,$F6,$97,$8F,$49,$3F,$91,$D9,$3A,$8F
	.dc.b	$F2,$8F,$D7,$7F,$A0,$DB,$EA,$FF,$95,$A3,$15,$F8,$A6,$BF,$BC,$F9
	.dc.b	$97,$77,$F9,$39,$B7,$FE,$0E,$E6,$F8,$DF,$1F,$9A,$FF,$53,$70,$BF
	.dc.b	$05,$FD,$CB,$7C,$FF,$9D,$18,$3F,$77,$F8,$2B,$E7,$FF,$CB,$7C,$BF
	.dc.b	$EC,$EC,$FB,$4E,$5F,$04,$67,$5F,$E5,$DF,$82,$FF,$9F,$1F,$F2,$E3
	.dc.b	$5F,$EA,$98,$FF,$AC,$E3,$1E,$B7,$F0,$C9,$CD,$FC,$98,$C1,$E8,$FA
	.dc.b	$1F,$CF,$4F,$F9,$69,$AF,$F8,$31,$BA,$1C,$EE,$F6,$BF,$6D,$3F,$ED
	.dc.b	$99,$FD,$53,$FE,$88,$8C,$E1,$D7,$C5,$FF,$3D,$FE,$0B,$FC,$DB,$1F
	.dc.b	$F5,$33,$71,$34,$9E,$17,$0F,$4B,$8B,$70,$FA,$DF,$7D,$FE,$33,$69
	.dc.b	$FE,$88,$FC,$01,$DD,$FE,$8A,$9D,$65,$FD,$BF,$25,$6A,$FF,$00,$D9
	.dc.b	$D3,$FE,$38,$D7,$FD,$46,$E5,$BF,$07,$FE,$A4,$8F,$F1,$DC,$FF,$9D
	.dc.b	$5A,$FF,$B0,$DA,$FF,$84,$BE,$67,$FD,$2D,$F8,$23,$E0,$BF,$C2,$B2
	.dc.b	$F9,$BF,$F2,$14,$E6,$33,$5C,$2B,$75,$A7,$F9,$B4,$66,$7F,$C3,$F3
	.dc.b	$97,$88,$AE,$8F,$FA,$9E,$34,$F9,$4F,$D8,$09,$AD,$32,$FE,$D7,$F0
	.dc.b	$47,$D8,$3F,$05,$C3,$6B,$FA,$54,$EA,$72,$71,$FC,$13,$FC,$B7,$3A
	.dc.b	$6B,$F1,$25,$18,$8F,$F1,$DC,$63,$5E,$47,$C1,$84,$3F,$05,$FF,$20
	.dc.b	$E7,$F0,$23,$13,$B7,$93,$F1,$74,$EB,$87,$92,$B8,$5F,$29,$F0,$B6
	.dc.b	$BF,$DA,$17,$CA,$BE,$C7,$07,$E3,$BE,$9B,$F8,$D3,$6B,$FD,$D3,$BA
	.dc.b	$72,$D8,$CF,$CE,$7F,$96,$34,$FC,$8B,$78,$BF,$1C,$D9,$FC,$75,$DC
	.dc.b	$3F,$3B,$39,$F3,$9F,$AE,$BE,$1F,$0D,$C3,$F4,$35,$C7,$DB,$FE,$4B
	.dc.b	$1F,$C8,$FF,$70,$7C,$97,$56,$C7,$F2,$79,$DB,$FA,$D8,$CB,$67,$81
	.dc.b	$D7,$C8,$F4,$6D,$FE,$7C,$9F,$47,$ED,$8F,$81,$A6,$BE,$03,$F4,$B5
	.dc.b	$FE,$A5,$DD,$5B,$BF,$E7,$0D,$7F,$7C,$D9,$FF,$56,$1D,$E3,$FD,$8D
	.dc.b	$F8,$47,$F5,$C7,$FA,$73,$5F,$F1,$03,$5F,$DA,$B5,$FF,$79,$7E,$0E
	.dc.b	$F8,$9F,$A0,$D8,$FE,$94,$D7,$FC,$D1,$3D,$6B,$C7,$F8,$43,$EA,$7E
	.dc.b	$1B,$6B,$FD,$BF,$28,$C4,$7F,$19,$FC,$23,$F5,$EF,$F5,$77,$F4,$AF
	.dc.b	$F8,$DF,$34,$62,$3F,$C5,$6D,$7F,$CB,$AD,$7D,$7E,$2F,$E2,$FC,$27
	.dc.b	$C8,$A3,$0D,$9F,$6C,$7E,$3F,$F0,$87,$C6,$71,$F7,$BF,$86,$E9,$FE
	.dc.b	$DB,$C7,$FC,$23,$E1,$6E,$BF,$2D,$F0,$3F,$EA,$DC,$BD,$0D,$7F,$C8
	.dc.b	$2D,$7F,$CE,$6D,$43,$57,$E1,$0F,$B0,$FD,$13,$67,$FB,$E3,$5F,$E8
	.dc.b	$4F,$D3,$F0,$79,$D3,$F0,$87,$CE,$7E,$B7,$E1,$9F,$8F,$7E,$73,$7F
	.dc.b	$F7,$B4,$63,$BB,$F6,$17,$C3,$3F,$F0,$56,$BF,$E8,$37,$DB,$F2,$1F
	.dc.b	$C0,$A3,$11,$EE,$8F,$F0,$F7,$8B,$DB,$5F,$DC,$67,$FE,$FD,$CD,$1F
	.dc.b	$02,$75,$7F,$0A,$72,$A1,$F4,$FF,$D8,$A3,$55,$C7,$8D,$E1,$FB,$7F
	.dc.b	$06,$FE,$21,$5F,$ED,$3D,$D1,$8C,$BA,$E5,$FD,$56,$FF,$86,$7C,$7F
	.dc.b	$E4,$B0,$FF,$28,$35,$FF,$77,$CF,$F9,$D3,$D0,$F6,$CE,$67,$D5,$F1
	.dc.b	$DF,$C4,$5F,$F7,$15,$23,$18,$78,$1B,$A7,$D6,$FC,$1F,$F0,$D3,$E4
	.dc.b	$F6,$5E,$57,$D7,$F9,$6F,$9B,$3F,$59,$B7,$F2,$6B,$79,$7C,$43,$E4
	.dc.b	$66,$FA,$9F,$39,$23,$FD,$45,$A7,$FA,$3A,$3E,$BF,$80,$F8,$8F,$E5
	.dc.b	$CD,$AE,$95,$FA,$99,$FE,$69,$F8,$3E,$67,$DF,$F8,$97,$F9,$15,$F1
	.dc.b	$72,$3E,$C9,$F5,$5F,$91,$6C,$F4,$AE,$0F,$C8,$FC,$73,$F9,$7B,$F4
	.dc.b	$1A,$FE,$1D,$AF,$E1,$5B,$97,$F2,$FD,$3F,$A3,$7A,$1F,$06,$BD,$2D
	.dc.b	$7F,$C5,$1E,$27,$D0,$E1,$FC,$97,$E3,$DF,$52,$7F,$D3,$5F,$45,$FE
	.dc.b	$AF,$7D,$6D,$7E,$EC,$6F,$16,$4F,$AF,$FD,$0B,$6B,$E7,$B5,$FF,$4E
	.dc.b	$35,$FC,$6F,$E3,$DF,$71,$FD,$03,$68,$E7,$C6,$FF,$17,$7E,$13,$FD
	.dc.b	$9E,$9F,$84,$79,$7A,$BF,$53,$18,$E9,$D4,$E7,$6F,$E1,$F3,$C3,$A3
	.dc.b	$3F,$39,$FE,$26,$AF,$F7,$B4,$FF,$A5,$DA,$FF,$BA,$5A,$FE,$93,$F4
	.dc.b	$A7,$DB,$7F,$75,$6C,$7E,$21,$AB,$2F,$2D,$74,$7C,$FF,$84,$BE,$BF
	.dc.b	$D8,$DA,$FC,$E0,$70,$F3,$CF,$D7,$0F,$C1,$78,$D7,$F1,$FF,$C3,$3F
	.dc.b	$B7,$D7,$F2,$EE,$D4,$63,$0F,$E6,$38,$74,$C7,$C7,$E1,$78,$BE,$0A
	.dc.b	$57,$8A,$BF,$5E,$F8,$4E,$31,$FC,$B7,$F2,$97,$CD,$53,$8C,$EC,$F5
	.dc.b	$7B,$7E,$EF,$0A,$47,$58,$AE,$F0,$FD,$3B,$67,$E2,$CE,$B2,$FE,$37
	.dc.b	$F0,$7F,$C9,$EB,$F6,$ED,$78,$1F,$2B,$97,$FD,$E2,$DF,$A1,$FC,$54
	.dc.b	$62,$3C,$CF,$FB,$B5,$FA,$3F,$31,$FD,$0D,$FF,$48,$7E,$5D,$FD,$3D
	.dc.b	$FE,$B0,$D5,$1B,$C3,$F2,$74,$FB,$3E,$FE,$74,$BF,$24,$7F,$55,$F9
	.dc.b	$37,$89,$49,$3E,$9A,$F3,$FC,$3C,$8F,$63,$EA,$F5,$BD,$7E,$7F,$FA
	.dc.b	$03,$23,$49,$5F,$EF,$3F,$C3,$3F,$C5,$9D,$5F,$A4,$6C,$7F,$74,$6B
	.dc.b	$FE,$09,$9F,$D8,$7E,$17,$F7,$A3,$F2,$AB,$EA,$72,$6B,$9E,$CF,$F2
	.dc.b	$4D,$9F,$65,$EF,$33,$1F,$85,$BF,$E0,$DC,$3F,$BA,$6E,$7F,$B8,$6C
	.dc.b	$FF,$82,$1F,$9D,$1F,$BC,$FC,$73,$EC,$4F,$E3,$F4,$FE,$18,$9B,$1A
	.dc.b	$BE,$DA,$FB,$09,$5E,$CF,$37,$DF,$E3,$7F,$AA,$B2,$FE,$45,$3F,$EB
	.dc.b	$E6,$BE,$58,$FF,$00,$FE,$C7,$F2,$BC,$B3,$4F,$97,$5F,$C4,$EB,$FC
	.dc.b	$43,$F8,$67,$EE,$5F,$DF,$B5,$FD,$43,$72,$FE,$3A,$BF,$01,$FE,$45
	.dc.b	$F9,$87,$CC,$AF,$EA,$FF,$A0,$FF,$0D,$E0,$FF,$07,$36,$3F,$C5,$F3
	.dc.b	$B7,$AF,$89,$FC,$AF,$0F,$DE,$E3,$DA,$31,$EC,$7F,$A4,$7D,$CC,$E5
	.dc.b	$FE,$19,$F9,$EF,$E8,$47,$F4,$38,$CE,$3F,$E1,$9E,$7F,$F1,$B3,$FE
	.dc.b	$0A,$6B,$FE,$21,$6B,$BF,$87,$FF,$14,$FE,$BB,$FE,$E9,$8F,$EE,$2D
	.dc.b	$7F,$C1,$AD,$CF,$FC,$37,$5F,$E4,$37,$FD,$19,$F8,$73,$EF,$7F,$D5
	.dc.b	$1F,$1B,$FE,$1D,$9F,$F1,$0B,$74,$FF,$23,$D6,$F1,$FE,$FF,$39,$7F
	.dc.b	$C6,$4D,$E5,$39,$F5,$95,$C6,$BF,$C4,$3F,$47,$FE,$05,$7F,$C5,$2D
	.dc.b	$7F,$C8,$0D,$7F,$C7,$CD,$B7,$D4,$FA,$F1,$FF,$23,$36,$BF,$C8,$8D
	.dc.b	$95,$7F,$84,$F2,$FF,$1E,$FC,$87,$FB,$92,$FF,$AF,$9F,$E4,$B7,$0F
	.dc.b	$F8,$F1,$AF,$F9,$4D,$AF,$F9,$47,$E2,$FF,$92,$4F,$F9,$21,$B8,$FF
	.dc.b	$96,$5B,$3D,$7E,$77,$FD,$AD,$F8,$67,$F9,$52,$3F,$C6,$9F,$17,$FE
	.dc.b	$70,$F3,$BF,$DB,$CD,$8F,$F1,$6C,$ED,$FE,$99,$19,$6F,$C6,$76,$8C
	.dc.b	$65,$FE,$5A,$AE,$0F,$F9,$A2,$1F,$86,$7F,$89,$9B,$F1,$FF,$0B,$0F
	.dc.b	$F4,$7C,$61,$C3,$5F,$8E,$BE,$7B,$FC,$4D,$1F,$E7,$26,$C6,$7E,$9F
	.dc.b	$D7,$7F,$E8,$BD,$BF,$C1,$BF,$85,$FF,$9B,$DB,$ED,$7F,$85,$E3,$18
	.dc.b	$FF,$A8,$1F,$D3,$A7,$F4,$6D,$EB,$F4,$3B,$6B,$93,$31,$FD,$44,$9F
	.dc.b	$67,$DE,$7F,$C4,$DF,$86,$7F,$22,$27,$A3,$8F,$4B,$3E,$97,$49,$FE
	.dc.b	$8F,$8F,$81,$AF,$C9,$6B,$F8,$C7,$FC,$99,$C3,$FE,$30,$CB,$FB,$BB
	.dc.b	$7D,$AF,$0F,$C5,$FE,$03,$DB,$7E,$0D,$49,$8F,$F8,$7B,$0E,$3C,$1F
	.dc.b	$EA,$5F,$C3,$3F,$D3,$7F,$50,$DB,$1E,$57,$F8,$EA,$31,$87,$FA,$9E
	.dc.b	$BF,$C2,$B1,$8F,$01,$1C,$AF,$8D,$7F,$B4,$2B,$5F,$31,$FB,$96,$F8
	.dc.b	$DD,$6B,$74,$9E,$FD,$3F,$EF,$FF,$C0,$35,$3F,$EB,$6D,$D1,$89,$D4
	.dc.b	$F7,$BF,$F6,$06,$3D,$DF,$F2,$9F,$C4,$FF,$83,$71,$F5,$D3,$F0,$C7
	.dc.b	$D1,$76,$8C,$36,$B8,$5F,$F9,$72,$9E,$E6,$DB,$7E,$13,$FD,$DE,$E1
	.dc.b	$F6,$4F,$1B,$E1,$EC,$7F,$8A,$67,$A3,$5E,$2D,$7C,$CD,$A7,$B3,$FC
	.dc.b	$DE,$7C,$8D,$7C,$2D,$7D,$CD,$87,$B3,$FE,$E2,$CF,$E1,$C6,$39,$3D
	.dc.b	$57,$F1,$7E,$33,$59,$FC,$1C,$2C,$8F,$F3,$AD,$61,$E0,$F2,$D7,$FD
	.dc.b	$E6,$F4,$C1,$87,$2D,$FF,$D7,$5F,$61,$7E,$54,$F6,$E6,$FF,$64,$71
	.dc.b	$66,$31,$F8,$63,$ED,$35,$EC,$2B,$31,$87,$7F,$81,$63,$5F,$AD,$8B
	.dc.b	$04,$E7,$7F,$AE,$34,$FF,$67,$7A,$CC,$3F,$97,$EB,$FE,$72,$7F,$D1
	.dc.b	$3C,$0F,$63,$58,$7F,$2E,$AD,$3F,$E7,$59,$FF,$5D,$57,$0A,$DF,$85
	.dc.b	$FE,$AC,$9C,$D6,$E7,$35,$F9,$AE,$8F,$F5,$8C,$D8,$FE,$33,$5F,$9A
	.dc.b	$6C,$EA,$7E,$B4,$FE,$62,$7C,$FD,$3D,$27,$ED,$7E,$18,$FC,$43,$11
	.dc.b	$FE,$B0,$9B,$6F,$B7,$A6,$3A,$EE,$3F,$09,$F0,$5F,$FA,$E1,$B5,$F9
	.dc.b	$89,$F3,$C6,$B1,$7E,$0F,$E0,$6C,$FF,$AD,$17,$36,$67,$D1,$C2,$F9
	.dc.b	$EF,$F5,$A5,$FF,$62,$6E,$5F,$D7,$BF,$0D,$FC,$24,$E7,$E5,$78,$C6
	.dc.b	$27,$FA,$9B,$5F,$F6,$CD,$A3,$11,$FD,$71,$AF,$E1,$7E,$FF,$FC,$8D
	.dc.b	$7F,$DC,$9B,$1F,$DB,$BD,$E7,$95,$E8,$9C,$F2,$FF,$6B,$9C,$DF,$1F
	.dc.b	$85,$F6,$C6,$30,$FE,$EF,$7C,$2B,$F6,$1D,$6F,$DA,$DF,$6D,$F6,$E3
	.dc.b	$0D,$FC,$39,$19,$9D,$D7,$EE,$73,$FC,$BC,$AB,$73,$FD,$EE,$30,$7E
	.dc.b	$2C,$AE,$63,$78,$FF,$88,$A7,$C7,$F7,$FE,$3C,$4F,$97,$EA,$5E,$8E
	.dc.b	$63,$0F,$58,$CE,$3E,$16,$E1,$D5,$AF,$F1,$13,$DB,$F9,$CF,$E1,$A7
	.dc.b	$D8,$7C,$67,$11,$FE,$EB,$D9,$F8,$6F,$ED,$35,$F8,$35,$A7,$F7,$33
	.dc.b	$AF,$07,$E6,$4E,$70,$FE,$41,$39,$7B,$E1,$E2,$E8,$FA,$5D,$3F,$C8
	.dc.b	$23,$FD,$F5,$BA,$31,$97,$D0,$FC,$37,$F9,$4D,$6A,$73,$87,$DE,$9C
	.dc.b	$CF,$BE,$D1,$88,$D7,$67,$D5,$8C,$47,$E0,$3F,$47,$FE,$17,$8F,$DE
	.dc.b	$FD,$C3,$D5,$39,$FA,$6F,$C4,$46,$21,$F8,$43,$8B,$F8,$1B,$74,$F7
	.dc.b	$CE,$BE,$0B,$35,$FC,$23,$F0,$3F,$E2,$53,$F0,$3E,$4B,$F9,$B5,$F0
	.dc.b	$9C,$E7,$F9,$09,$CD,$23,$F9,$4B,$E2,$F2,$BC,$27,$8B,$A2,$E1,$F4
	.dc.b	$38,$7F,$A0,$7D,$87,$E8,$7C,$D3,$E8,$F2,$30,$DA,$FF,$1F,$EF,$FC
	.dc.b	$8F,$2F,$F2,$C3,$5F,$F3,$DB,$5F,$C7,$2F,$57,$E1,$9F,$13,$C8,$B5
	.dc.b	$F1,$3F,$4C,$E9,$B7,$AB,$83,$F9,$06,$3E,$DA,$FC,$C5,$69,$FC,$D4
	.dc.b	$EA,$BE,$0F,$E2,$1F,$B6,$BE,$07,$F6,$F7,$11,$AF,$33,$F9,$BC,$62
	.dc.b	$FF,$48,$FF,$AD,$5F,$CA,$FD,$2F,$F8,$63,$ED,$37,$EE,$7F,$41,$8C
	.dc.b	$4F,$DC,$AF,$E6,$38,$F8,$03,$CD,$C3,$AC,$7C,$F7,$FD,$23,$F3,$1F
	.dc.b	$15,$7F,$81,$5A,$F2,$C7,$5E,$2F,$A1,$FE,$8F,$81,$A4,$C7,$FC,$EB
	.dc.b	$0F,$9D,$F7,$9F,$03,$E8,$2C,$FB,$B9,$1F,$11,$ED,$6C,$7E,$63,$F1
	.dc.b	$17,$C9,$E6,$BF,$AF,$C3,$EF,$6C,$E1,$B5,$FA,$0D,$BF,$A6,$F0,$7F
	.dc.b	$06,$F9,$CF,$F1,$14,$E7,$D4,$FE,$D9,$E4,$A7,$3C,$5F,$2F,$37,$C7
	.dc.b	$F8,$0F,$0F,$C2,$5C,$7C,$DF,$85,$7F,$3B,$FC,$C3,$F9,$0E,$9F,$A4
	.dc.b	$C1,$39,$F4,$7D,$2C,$BF,$77,$D9,$DC,$FC,$0D,$C3,$93,$67,$F9,$57
	.dc.b	$C7,$7F,$8C,$1F,$C3,$FE,$69,$F0,$DF,$DF,$C6,$6F,$F2,$7D,$8A,$D4
	.dc.b	$7E,$1E,$7F,$83,$37,$E1,$3F,$C1,$71,$8C,$7F,$7A,$F3,$6F,$E2,$7F
	.dc.b	$93,$FC,$CB,$F4,$97,$FE,$4D,$A4,$E7,$A7,$9B,$33,$F2,$F9,$56,$FF
	.dc.b	$97,$FE,$99,$F9,$59,$FE,$CE,$D8,$E9,$F0,$7F,$D9,$E3,$10,$C7,$F8
	.dc.b	$CB,$FE,$FC,$6B,$F1,$A7,$FC,$3B,$18,$8F,$36,$3E,$E3,$DF,$E3,$DC
	.dc.b	$3F,$8B,$7E,$51,$FD,$E6,$BF,$B9,$B7,$A5,$F8,$AB,$D1,$FE,$F5,$3B
	.dc.b	$F3,$BF,$41,$F1,$9F,$DC,$7C,$4F,$F3,$B6,$3F,$DB,$DA,$FF,$45,$6B
	.dc.b	$FE,$06,$F1,$BF,$86,$67,$FE,$77,$E7,$7D,$D7,$F9,$CA,$3F,$95,$4E
	.dc.b	$A9,$5A,$F8,$CF,$F0,$67,$E5,$1F,$E4,$5B,$FE,$60,$DA,$FE,$D9,$F8
	.dc.b	$B7,$1D,$13,$9E,$9F,$F0,$6F,$E4,$EF,$BA,$FF,$05,$37,$1F,$F0,$73
	.dc.b	$71,$FF,$05,$35,$FE,$C1,$F9,$33,$CC,$FE,$31,$F9,$3F,$F1,$3C,$7F
	.dc.b	$C3,$33,$9C,$3F,$8F,$CE,$5F,$F1,$3F,$DD,$BC,$2F,$F1,$33,$63,$FB
	.dc.b	$CF,$47,$FB,$87,$EC,$B3,$5F,$E2,$BF,$C9,$7F,$C3,$35,$FE,$23,$6E
	.dc.b	$9F,$E1,$F6,$BF,$DD,$5A,$FF,$8D,$BF,$27,$7D,$07,$F2,$FF,$6B,$FD
	.dc.b	$DD,$7F,$D6,$67,$2F,$F3,$68,$CD,$7F,$82,$3F,$27,$FE,$F3,$3F,$E4
	.dc.b	$06,$C3,$37,$E0,$BF,$6E,$38,$7C,$1F,$A7,$DD,$FF,$1C,$69,$FD,$4A
	.dc.b	$7C,$D1,$83,$FC,$79,$F2,$2F,$CA,$3F,$A2,$75,$FE,$46,$31,$A7,$F4
	.dc.b	$A8,$FC,$A7,$C8,$3D,$FF,$F2,$83,$67,$FB,$A2,$FA,$AC,$CF,$F2,$4F
	.dc.b	$C2,$FF,$3D,$5F,$E3,$9C,$FF,$C5,$71,$87,$1B,$A3,$11,$FE,$3F,$8C
	.dc.b	$78,$DF,$E9,$D5,$F8,$B3,$E9,$FF,$AE,$7D,$0F,$F2,$6C,$7E,$EC,$E5
	.dc.b	$FE,$C7,$19,$C3,$FC,$89,$F8,$DB,$EB,$BF,$9C,$B6,$BF,$0A,$EF,$BF
	.dc.b	$FC,$8A,$E6,$BF,$C9,$9A,$7F,$AE,$1F,$33,$5F,$6F,$2F,$9E,$31,$1F
	.dc.b	$67,$F1,$AF,$E2,$9F,$13,$EA,$F8,$8F,$E9,$31,$8D,$7F,$55,$9F,$F9
	.dc.b	$43,$E5,$7F,$0B,$A7,$FB,$3B,$07,$E5,$7F,$A1,$CB,$FC,$BD,$5A,$BF
	.dc.b	$F7,$B5,$6B,$C5,$FE,$AB,$C9,$FE,$30,$BF,$98,$D7,$FC,$BA,$D7,$F1
	.dc.b	$CD,$C5,$F6,$7F,$CE,$70,$FF,$45,$36,$B7,$EC,$FF,$B3,$6B,$7E,$DF
	.dc.b	$F6,$4C,$FF,$D0,$F1,$FE,$6E,$6E,$1F,$E7,$BF,$C6,$FF,$D4,$13,$9F
	.dc.b	$37,$FD,$A7,$1F,$EA,$3C,$7F,$10,$D7,$87,$CE,$FF,$B4,$23,$11,$FA
	.dc.b	$5F,$C4,$FF,$BF,$E1,$FB,$E6,$C7,$F8,$B3,$07,$DA,$FF,$2C,$7F,$D0
	.dc.b	$33,$A7,$FD,$EB,$19,$AF,$F4,$0C,$E5,$FF,$55,$46,$6B,$C1,$F1,$7F
	.dc.b	$E0,$2F,$DC,$3F,$45,$5F,$C3,$9B,$3F,$E9,$3F,$BA,$FE,$BF,$DC,$9C
	.dc.b	$E5,$FE,$92,$9C,$C6,$2F,$FD,$23,$18,$F0,$3F,$C1,$FF,$8D,$FF,$A2
	.dc.b	$31,$FF,$5F,$FE,$0A,$F8,$AC,$E1,$FE,$C5,$8C,$3F,$C8,$2B,$86,$66
	.dc.b	$FE,$31,$FA,$C9,$D1,$FE,$1C,$71,$87,$ED,$9A,$FE,$D5,$AF,$FA,$99
	.dc.b	$AF,$EF,$FC,$22,$FB,$7F,$F2,$03,$FE,$BB,$EB,$5E,$A7,$33,$FE,$BF
	.dc.b	$AC,$7B,$7F,$C2,$6F,$73,$CE,$3F,$86,$29,$DB,$D6,$EB,$CC,$E0,$64
	.dc.b	$A6,$4E,$67,$32,$94,$CC,$6B,$F1,$97,$E0,$8F,$C6,$F0,$DA,$F2,$CE
	.dc.b	$A7,$27,$1F,$C6,$FF,$D4,$53,$A6,$BF,$C9,$E5,$18,$8F,$F7,$14,$63
	.dc.b	$93,$91,$EE,$7F,$8E,$45,$F2,$3F,$DD,$1A,$FC,$C9,$E1,$3F,$01,$B1
	.dc.b	$99,$DE,$9D,$9B,$5F,$6A,$B7,$3F,$7F,$C9,$F8,$FC,$2E,$0F,$E6,$97
	.dc.b	$F5,$23,$13,$A9,$CC,$FE,$2F,$E5,$BF,$DF,$99,$FF,$BC,$A7,$FC,$F7
	.dc.b	$3F,$DD,$9A,$FF,$BE,$DE,$31,$B9,$E3,$C1,$F9,$1C,$6D,$FF,$67,$BF
	.dc.b	$C5,$DE,$E3,$FD,$F3,$C2,$EE,$F8,$F1,$FE,$B0,$8C,$6B,$C9,$B1,$C2
	.dc.b	$7E,$9B,$67,$51,$F8,$28,$D7,$0E,$27,$97,$E7,$EF,$9C,$F2,$FD,$5B
	.dc.b	$E2,$9F,$3D,$86,$E3,$E6,$BE,$0F,$EA,$B4,$FD,$0D,$F1,$BE,$04,$9E
	.dc.b	$18,$7A,$7E,$9D,$F8,$47,$E7,$36,$46,$2B,$F5,$31,$89,$D4,$E6,$7F
	.dc.b	$60,$BF,$1C,$E8,$48,$F9,$71,$BC,$BF,$6B,$CA,$FC,$67,$E6,$C6,$1B
	.dc.b	$C9,$FC,$12,$75,$C1,$F3,$FF,$1A,$7D,$5F,$DA,$9E,$5F,$31,$F8,$EC
	.dc.b	$3F,$D4,$79,$FE,$35,$AF,$F8,$29,$AF,$F8,$25,$FF,$19,$BF,$E4,$07
	.dc.b	$FC,$88,$DF,$83,$ED,$1A,$8F,$F7,$8F,$13,$F5,$67,$C8,$7D,$86,$CF
	.dc.b	$D3,$D7,$F8,$ED,$6E,$BF,$11,$F8,$DF,$E6,$37,$FE,$F1,$FA,$07,$F3
	.dc.b	$5C,$3E,$7B,$5F,$E3,$8D,$7F,$92,$35,$FF,$3D,$78,$5F,$EC,$A7,$FD
	.dc.b	$FA,$FF,$39,$6F,$89,$FA,$0F,$90,$FE,$25,$F3,$F7,$F0,$57,$D5,$5F
	.dc.b	$52,$61,$E9,$3F,$D0,$33,$A7,$FC,$CB,$3F,$EA,$57,$F7,$4D,$7F,$C6
	.dc.b	$ED,$7F,$C6,$CD,$7F,$2D,$F9,$AB,$ED,$BF,$67,$AA,$31,$3E,$18,$C4
	.dc.b	$FA,$72,$7D,$EF,$AF,$F1,$C7,$E0,$37,$AA,$63,$E9,$B5,$6A,$31,$F1
	.dc.b	$6F,$BB,$FE,$77,$48,$C4,$FF,$48,$8C,$4E,$A7,$33,$FE,$57,$E6,$7E
	.dc.b	$0B,$E4,$8F,$91,$18,$CF,$F7,$73,$99,$FF,$73,$FD,$37,$EF,$5D,$F9
	.dc.b	$2F,$9D,$FE,$02,$D8,$CC,$7D,$9F,$A1,$FC,$AC,$38,$7F,$AF,$FB,$FF
	.dc.b	$DA,$79,$BF,$AA,$B6,$7F,$B0,$B5,$FF,$37,$35,$FE,$E2,$7F,$AF,$BF
	.dc.b	$40,$F6,$D7,$F6,$AE,$04,$3E,$5F,$F1,$4D,$AF,$9C,$F0,$EF,$FF,$66
	.dc.b	$7D,$F3,$F0,$2F,$C8,$3F,$EB,$7A,$FA,$8D,$7D,$53,$C2,$7F,$2D,$F8
	.dc.b	$6B,$E9,$7E,$AB,$6B,$EA,$35,$FF,$44,$64,$8C,$47,$B3,$91,$F8,$0F
	.dc.b	$F3,$18,$FD,$CB,$75,$FE,$4E,$DC,$BF,$DC,$9F,$2E,$FB,$4F,$E6,$7F
	.dc.b	$A0,$67,$06,$28,$C3,$79,$7F,$5D,$F1,$AC,$3F,$A7,$F1,$BF,$1B,$7E
	.dc.b	$07,$F5,$9B,$97,$D6,$6B,$F5,$9A,$FD,$6E,$07,$D1,$FF,$09,$DB,$FD
	.dc.b	$F4,$FF,$60,$6F,$BF,$FB,$F8,$D5,$7F,$8B,$BF,$27,$FF,$57,$71,$7F
	.dc.b	$89,$A1,$18,$BF,$C1,$B6,$3F,$E0,$0E,$3F,$C7,$1F,$AE,$7F,$8E,$3F
	.dc.b	$17,$FF,$7D,$3F,$E2,$46,$BF,$E3,$06,$BF,$EB,$D6,$BF,$90,$7F,$D9
	.dc.b	$CF,$F0,$F0,$E1,$EE,$63,$E7,$BF,$BC,$BA,$FA,$7F,$E3,$94,$F9,$CA
	.dc.b	$49,$FE,$05,$1F,$C6,$5F,$4E,$BF,$4E,$78,$F6,$C9,$DC,$7F,$27,$FB
	.dc.b	$07,$E6,$17,$E7,$65,$7F,$0D,$9C,$B6,$38,$7C,$04,$8C,$1E,$DE,$57
	.dc.b	$E0,$1E,$A7,$FD,$D1,$19,$F9,$9C,$4E,$59,$19,$39,$47,$1F,$90,$91
	.dc.b	$D5,$DB,$FE,$17,$FC,$FF,$FD,$C6,$FF,$87,$1A,$E7,$E7,$3F,$C3,$57
	.dc.b	$F8,$3E,$6F,$E0,$1F,$11,$F8,$BA,$7C,$37,$FB,$F7,$DC,$D9,$FD,$2E
	.dc.b	$75,$C7,$FE,$17,$F3,$71,$E9,$CC,$3F,$99,$57,$F2,$66,$FA,$7C,$48
	.dc.b	$EA,$7E,$EF,$E0,$2E,$3A,$49,$FC,$34,$7F,$88,$1F,$C2,$FE,$28,$E8
	.dc.b	$FF,$39,$B6,$3F,$CE,$ED,$7F,$D0,$6D,$7F,$90,$87,$F0,$DB,$7E,$95
	.dc.b	$BD,$DF,$C9,$9B,$3F,$E9,$87,$A4,$FF,$32,$FC,$D7,$F3,$0D,$F5,$FF
	.dc.b	$80,$C7,$F5,$CA,$FD,$4F,$88,$8D,$E1,$FE,$A2,$8C,$57,$0F,$26,$5F
	.dc.b	$F7,$39,$CB,$7F,$5D,$7E,$01,$FC,$39,$B1,$98,$C7,$3E,$7E,$8F,$5F
	.dc.b	$84,$CA,$E5,$1F,$AC,$FC,$6D,$F7,$F4,$93,$FB,$5F,$BF,$FE,$16,$FC
	.dc.b	$7F,$AB,$F9,$E4,$7E,$7B,$F4,$B7,$FD,$BF,$17,$E3,$5F,$94,$62,$3F
	.dc.b	$92,$FE,$F1,$FE,$6F,$AF,$73,$5F,$8E,$DE,$3F,$EB,$EF,$EF,$E3,$7E
	.dc.b	$CF,$2F,$7E,$46,$0F,$91,$F4,$1F,$C6,$F8,$FF,$DD,$91,$FC,$D2,$30
	.dc.b	$7E,$79,$F9,$DF,$93,$BE,$8B,$F9,$CA,$FC,$0E,$98,$8F,$D1,$74,$7F
	.dc.b	$82,$BB,$FE,$F5,$FF,$37,$6B,$FE,$DE,$F0,$7F,$CF,$93,$BE,$AF,$D5
	.dc.b	$12,$7F,$4F,$1F,$AD,$F3,$E4,$7E,$A9,$FD,$87,$9F,$88,$FE,$37,$83
	.dc.b	$F2,$0F,$DA,$19,$E2,$FD,$DF,$3F,$94,$95,$FE,$1D,$AF,$BE,$DE,$AF
	.dc.b	$DE,$11,$FD,$CC,$FE,$FF,$AB,$DC,$48,$E4,$FC,$1E,$CF,$E0,$84,$8F
	.dc.b	$E0,$31,$FC,$1F,$B3,$FD,$3B,$F3,$1F,$D2,$F2,$FF,$4B,$B7,$AB,$E3
	.dc.b	$67,$FC,$5A,$73,$E3,$70,$26,$3F,$C3,$30,$FF,$2E,$7E,$AD,$FD,$6E
	.dc.b	$3F,$B3,$35,$FA,$0D,$5C,$68,$C3,$7B,$39,$92,$7F,$C9,$53,$F9,$97
	.dc.b	$8B,$E7,$6D,$F8,$8F,$1F,$9B,$E4,$F7,$BF,$C9,$B5,$FE,$14,$AF,$1C
	.dc.b	$60,$CE,$7E,$A2,$47,$82,$3F,$C1,$3E,$47,$F8,$B6,$1D,$9F,$EA,$2E
	.dc.b	$DC,$9A,$EE,$DF,$E2,$1F,$E4,$2B,$34,$97,$FE,$44,$FD,$EF,$FD,$0C
	.dc.b	$FF,$A7,$3E,$D7,$FD,$4F,$C5,$FE,$CA,$6F,$1F,$F1,$5A,$DC,$A1,$57
	.dc.b	$F7,$CF,$E6,$93,$98,$C7,$A7,$FE,$F0,$A5,$69,$FF,$52,$E1,$DA,$BD
	.dc.b	$B9,$3E,$BB,$FC,$AB,$F4,$FF,$E6,$DD,$FF,$D8,$3D,$18,$F8,$BF,$E5
	.dc.b	$17,$FE,$E4,$C1,$7B,$8E,$51,$E6,$8C,$57,$18,$EF,$F8,$25,$C3,$FD
	.dc.b	$9A,$DB,$E9,$7F,$E7,$DF,$83,$7D,$CF,$F0,$7C,$3B,$E5,$FE,$BE,$8C
	.dc.b	$56,$DB,$1E,$6C,$3C,$15,$FE,$02,$DE,$FE,$26,$FC,$12,$9A,$F6,$FF
	.dc.b	$AF,$65,$FE,$89,$9F,$F4,$45,$F8,$0F,$A2,$E2,$35,$D3,$CA,$7E,$06
	.dc.b	$5F,$E7,$EA,$FF,$36,$6D,$FD,$82,$78,$F4,$FE,$31,$C4,$E9,$B1,$97
	.dc.b	$FB,$3C,$7F,$97,$7D,$8E,$50,$8C,$3C,$BB,$3E,$33,$E1,$6B,$E8,$ED
	.dc.b	$FC,$36,$7E,$D6,$E5,$E7,$EE,$B5,$BE,$7F,$5B,$F9,$0D,$BF,$25,$18
	.dc.b	$8F,$14,$F0,$7F,$C5,$F1,$72,$4F,$4E,$3F,$B7,$E2,$5E,$43,$E0,$E0
	.dc.b	$75,$BE,$57,$53,$98,$7D,$15,$9C,$37,$3F,$3D,$F2,$FA,$DC,$C6,$2D
	.dc.b	$18,$D7,$EE,$4B,$D5,$70,$FE,$AF,$BF,$F5,$93,$98,$C3,$FB,$2D,$9F
	.dc.b	$21,$FD,$0E,$3F,$DF,$46,$CF,$F0,$57,$E6,$7F,$B8,$7F,$DA,$F1,$BC
	.dc.b	$7F,$C2,$B3,$97,$FC,$FF,$5D,$3C,$BE,$F1,$FC,$4B,$47,$D2,$FF,$2F
	.dc.b	$DB,$E3,$B6,$33,$97,$F0,$F8,$C7,$D6,$F2,$AF,$9A,$DA,$7C,$6D,$8E
	.dc.b	$FE,$57,$F3,$7E,$FB,$E5,$7E,$9F,$2F,$E6,$BF,$19,$89,$CB,$FD,$06
	.dc.b	$BF,$9F,$36,$BF,$D6,$35,$FE,$FD,$6E,$9F,$EC,$A7,$FC,$FF,$1F,$D0
	.dc.b	$3E,$1B,$BF,$C0,$78,$1F,$0F,$89,$27,$FA,$2F,$8F,$70,$FF,$46,$E0
	.dc.b	$8F,$01,$27,$C0,$FF,$55,$E2,$D7,$85,$F3,$1B,$3F,$D4,$3C,$6C,$57
	.dc.b	$FB,$2E,$30,$7F,$30,$E8,$FE,$D3,$7F,$12,$B6,$FC,$6F,$22,$57,$F5
	.dc.b	$AF,$2E,$E7,$E6,$7E,$21,$23,$E2,$3F,$1F,$CD,$F9,$E7,$D0,$F3,$6C
	.dc.b	$FF,$64,$7E,$61,$1F,$98,$FC,$FF,$A8,$FA,$FF,$11,$F6,$1B,$3F,$DB
	.dc.b	$1C,$3F,$DB,$BE,$6A,$E3,$F2,$BE,$0B,$EF,$9F,$82,$FC,$9F,$FB,$0D
	.dc.b	$7F,$74,$8C,$76,$7D,$F9,$CD,$7F,$9F,$3A,$11,$88,$FF,$75,$35,$FF
	.dc.b	$38,$7C,$4B,$E8,$FF,$28,$BE,$92,$C6,$1C,$FD,$F4,$CF,$FC,$E9,$4F
	.dc.b	$3F,$18,$7F,$9F,$BF,$78,$FE,$1F,$5F,$E7,$97,$F7,$6D,$F3,$BF,$45
	.dc.b	$18,$9E,$13,$83,$F4,$DC,$8F,$63,$11,$FE,$E7,$FB,$FF,$CC,$F5,$7F
	.dc.b	$BE,$DB,$1F,$DD,$2D,$1F,$A7,$7F,$DD,$6E,$9F,$EF,$2E,$5F,$F5,$1B
	.dc.b	$5E,$4D,$77,$E1,$7F,$8A,$FF,$0C,$7D,$87,$F0,$49,$46,$2F,$FD,$F9
	.dc.b	$EF,$7F,$A5,$9F,$F7,$F4,$66,$7F,$D4,$7F,$BE,$3E,$C7,$F8,$23,$9C
	.dc.b	$FF,$C1,$5C,$E8,$D6,$1F,$EA,$37,$4F,$FA,$A5,$7D,$07,$23,$9F,$CD
	.dc.b	$59,$FF,$55,$7E,$47,$FC,$7F,$0E,$AF,$FA,$82,$33,$E7,$FF,$15,$8C
	.dc.b	$67,$FE,$B8,$5F,$0B,$99,$D3,$6B,$F6,$87,$D1,$FC,$3D,$F2,$7F,$15
	.dc.b	$B1,$AC,$3F,$C5,$30,$F9,$4D,$56,$DF,$F2,$C4,$66,$BF,$B7,$B6,$BF
	.dc.b	$BD,$35,$FF,$6D,$FE,$5D,$FC,$69,$67,$FE,$33,$51,$98,$C3,$F8,$38
	.dc.b	$CF,$9D,$FA,$FF,$BC,$98,$79,$FC,$2F,$D5,$FA,$F8,$BE,$B3,$FE,$9D
	.dc.b	$F0,$FF,$44,$ED,$FE,$A6,$DE,$FF,$DD,$57,$EE,$FC,$1F,$DE,$78,$76
	.dc.b	$B8,$3F,$BF,$F0,$E4,$73,$FB,$2B,$97,$80,$FF,$28,$FE,$1F,$FD,$D6
	.dc.b	$7E,$64,$23,$0E,$CC,$7B,$9F,$97,$8C,$F0,$FC,$57,$FB,$9D,$72,$6D
	.dc.b	$71,$7F,$BF,$C6,$6B,$97,$BB,$CD,$B1,$93,$E4,$7E,$4F,$FD,$0D,$BE
	.dc.b	$8F,$F5,$C8,$C6,$1E,$BE,$14,$27,$5E,$86,$FD,$2B,$87,$47,$FB,$2F
	.dc.b	$E7,$7E,$4C,$79,$3F,$DC,$1B,$9E,$BD,$CD,$B2,$FE,$E5,$7B,$B6,$FF
	.dc.b	$3C,$68,$8E,$91,$F8,$9E,$37,$A7,$FB,$F6,$DF,$D0,$7F,$D1,$ED,$8D
	.dc.b	$79,$7F,$EB,$19,$CE,$7F,$E5,$16,$BF,$C5,$CF,$F0,$CC,$7F,$AD,$FF
	.dc.b	$33,$FF,$19,$DF,$DC,$6C,$FC,$7F,$9F,$F9,$38,$FE,$76,$72,$FA,$BE
	.dc.b	$27,$F6,$F7,$F9,$95,$F2,$3F,$E0,$C7,$38,$7D,$17,$33,$FB,$5F,$23
	.dc.b	$F8,$5C,$7F,$87,$7C,$73,$49,$3F,$E6,$BF,$65,$F8,$9F,$92,$3F,$C6
	.dc.b	$8D,$C3,$FD,$A3,$F8,$7F,$F9,$61,$FF,$16,$C6,$69,$EE,$BD,$EF,$F0
	.dc.b	$23,$70,$FD,$97,$E6,$7E,$6C,$3F,$C6,$B3,$97,$F1,$5E,$F3,$EA,$BF
	.dc.b	$C4,$4D,$BF,$F4,$07,$3A,$31,$5F,$D9,$DA,$FF,$49,$87,$E1,$BD,$F8
	.dc.b	$BF,$0A,$FF,$1A,$9F,$CC,$39,$BF,$F6,$1D,$70,$F3,$3F,$C7,$3F,$19
	.dc.b	$FC,$12,$FF,$3C,$BC,$9C,$F9,$1A,$7F,$C8,$9F,$5F,$FE,$C1,$C7,$FD
	.dc.b	$2D,$CC,$8C,$61,$F8,$78,$C7,$83,$FE,$51,$9C,$CF,$F6,$1E,$37,$E8
	.dc.b	$3F,$9A,$C3,$FC,$CC,$D9,$6B,$FD,$9B,$B7,$FC,$65,$39,$B7,$D1,$7F
	.dc.b	$67,$6F,$C3,$7F,$95,$3E,$57,$F0,$79,$7F,$3B,$9C,$BF,$C9,$FF,$30
	.dc.b	$FA,$D9,$FC,$B6,$BF,$19,$AF,$89,$B9,$FF,$A5,$BD,$BF,$F3,$0F,$C0
	.dc.b	$BE,$6A,$C6,$1F,$F3,$9F,$3B,$ED,$FF,$C1,$F8,$FF,$25,$8C,$5F,$F9
	.dc.b	$AB,$93,$F9,$9C,$E6,$B7,$5F,$EF,$7F,$BB,$7E,$14,$F8,$7F,$E9,$17
	.dc.b	$FE,$FE,$9C,$C6,$21,$18,$DB,$FD,$98,$D8,$FF,$19,$7C,$BF,$F8,$B6
	.dc.b	$7F,$A5,$3F,$83,$C3,$F8,$7B,$99,$FE,$44,$E6,$7F,$17,$C0,$F9,$8F
	.dc.b	$F4,$1F,$4B,$E7,$FF,$D0,$B3,$9A,$FE,$AD,$39,$7F,$AC,$61,$FD,$8C
	.dc.b	$FC,$E1,$4F,$CB,$94,$FC,$19,$65,$FB,$97,$38,$FF,$51,$B7,$25,$27
	.dc.b	$CD,$F1,$3F,$C4,$A3,$FD,$69,$5B,$7E,$57,$D4,$BE,$3B,$FD,$57,$B2
	.dc.b	$1F,$1F,$87,$55,$FE,$07,$9D,$4E,$7E,$43,$E9,$47,$16,$D1,$A6,$3D
	.dc.b	$2E,$87,$FA,$EB,$81,$F8,$4B,$E8,$F3,$97,$F8,$56,$71,$E4,$6E,$7F
	.dc.b	$D2,$5B,$FF,$71,$9E,$1E,$7E,$6B,$4F,$E1,$63,$37,$FC,$F6,$BE,$D5
	.dc.b	$FE,$A1,$FE,$81,$3A,$7F,$A1,$71,$7F,$9E,$3E,$C3,$FD,$49,$F0,$47
	.dc.b	$D1,$F9,$D7,$EC,$E3,$18,$F9,$E3,$13,$AF,$0F,$8B,$EF,$F9,$DF,$F6
	.dc.b	$E7,$3F,$D8,$CB,$F1,$9F,$96,$FD,$57,$F1,$D3,$AC,$7E,$E6,$8A,$D2
	.dc.b	$9D,$57,$03,$B4,$77,$7E,$EF,$DE,$7F,$9F,$70,$FD,$73,$F2,$9B,$F3
	.dc.b	$BA,$E1,$DB,$32,$FC,$05,$7E,$D9,$FD,$93,$CD,$BE,$17,$E0,$A3,$F8
	.dc.b	$0E,$1F,$E8,$1E,$74,$62,$37,$5F,$4D,$FF,$43,$77,$7E,$67,$F0,$6F
	.dc.b	$F3,$94,$7F,$A2,$74,$F8,$D1,$88,$D8,$70,$FE,$C4,$C7,$C9,$7D,$FA
	.dc.b	$E8,$FD,$EB,$B3,$FD,$FE,$6B,$F3,$A2,$BF,$AB,$8F,$33,$4A,$E1,$3E
	.dc.b	$BF,$C1,$9F,$3A,$FB,$FC,$36,$BF,$25,$3A,$9C,$9C,$7E,$7B,$FC,$CF
	.dc.b	$3A,$6B,$FE,$15,$94,$62,$3F,$D5,$D1,$8E,$1E,$47,$B1,$7C,$57,$FA
	.dc.b	$CF,$4F,$95,$18,$AD,$5F,$82,$72,$DD,$3A,$DF,$16,$D7,$0C,$3E,$5E
	.dc.b	$EF,$C1,$BF,$D6,$73,$FC,$C9,$AF,$F8,$B6,$35,$F3,$3F,$81,$6C,$6E
	.dc.b	$FF,$D9,$17,$FD,$0E,$FF,$00,$D8,$CE,$FF,$D0,$E7,$AC,$E6,$BF,$8B
	.dc.b	$7E,$F0,$E4,$DC,$7F,$46,$8C,$3F,$C6,$EB,$AD,$FE,$BB,$C4,$FB,$33
	.dc.b	$FC,$45,$AF,$F0,$F6,$E1,$FB,$9E,$3F,$E5,$5A,$F6,$7F,$96,$71,$5F
	.dc.b	$C7,$DF,$53,$4D,$61,$CB,$2F,$E7,$1D,$EF,$A9,$7C,$67,$A2,$30,$D8
	.dc.b	$CE,$BF,$C9,$7F,$08,$FF,$1B,$BF,$C9,$FE,$B3,$BB,$72,$F2,$46,$23
	.dc.b	$FD,$31,$87,$47,$F9,$77,$E1,$0F,$8E,$FD,$A6,$E8,$C4,$FF,$29,$8C
	.dc.b	$4E,$A7,$33,$FC,$AB,$F0,$B7,$C9,$FC,$36,$C7,$F3,$B6,$BF,$CE,$67
	.dc.b	$8F,$1F,$F8,$07,$F0,$87,$CB,$7D,$8E,$A4,$62,$FF,$A9,$35,$FF,$59
	.dc.b	$46,$B9,$B8,$4B,$E6,$3F,$1C,$DA,$FF,$02,$B5,$FC,$A3,$5F,$E3,$0B
	.dc.b	$E6,$3E,$E7,$AD,$FE,$70,$9F,$EA,$8D,$7C,$33,$B7,$FA,$C7,$E0,$FF
	.dc.b	$D1,$DB,$E5,$7E,$16,$31,$1F,$4B,$F0,$8F,$C9,$BF,$D6,$FD,$B7,$D7
	.dc.b	$7D,$26,$C7,$D1,$6B,$FD,$61,$AF,$F9,$9F,$C7,$FE,$5E,$FF,$B9,$D7
	.dc.b	$E4,$DF,$A2,$F8,$1F,$E2,$8D,$8D,$CF,$FB,$07,$F0,$9F,$F1,$83,$FE
	.dc.b	$EF,$FC,$49,$EF,$71,$6C,$7F,$BC,$E7,$6F,$D2,$FA,$DC,$BF,$BF,$6B
	.dc.b	$FD,$0D,$AF,$F8,$ED,$AF,$F9,$8B,$1F,$A9,$CD,$FE,$C2,$6F,$85,$FE
	.dc.b	$7C,$6C,$A5,$F8,$57,$FA,$B2,$3F,$C4,$51,$FE,$CA,$FC,$2B,$F4,$0D
	.dc.b	$F9,$9F,$85,$1B,$9F,$F7,$03,$AC,$3F,$D1,$3F,$1B,$FE,$19,$9F,$F4
	.dc.b	$BB,$5F,$F6,$36,$3F,$E8,$3F,$7F,$FA,$D3,$63,$FD,$3B,$F8,$53,$EB
	.dc.b	$7E,$83,$63,$CB,$87,$FA,$5A,$31,$1F,$B4,$FA,$77,$C9,$7D,$7F,$91
	.dc.b	$FC,$F6,$DF,$EE,$46,$E1,$FE,$9E,$6D,$BC,$7F,$EA,$3F,$39,$FC,$83
	.dc.b	$97,$FB,$37,$C3,$FF,$6F,$6D,$7F,$A7,$3D,$8F,$F1,$FB,$67,$D7,$F8
	.dc.b	$57,$BF,$3F,$F7,$9B,$FE,$47,$9F,$49,$FE,$98,$70,$FD,$3C,$FF,$D6
	.dc.b	$31,$C5,$FF,$59,$FE,$14,$FA,$DF,$C9,$2C,$E9,$88,$FC,$B7,$E1,$5F
	.dc.b	$C8,$F0,$FF,$63,$B5,$FA,$2D,$7F,$D8,$CD,$7F,$D8,$3F,$02,$FA,$CF
	.dc.b	$D0,$36,$3D,$3C,$BF,$ED,$19,$FE,$5D,$F8,$57,$F3,$39,$FF,$59,$E1
	.dc.b	$FB,$DD,$7F,$C6,$D7,$E9,$F8,$AF,$D5,$4E,$75,$FE,$73,$F8,$57,$1E
	.dc.b	$27,$D2,$F8,$F2,$FD,$4B,$63,$3E,$C7,$D3,$8D,$64,$FC,$2B,$F8,$2E
	.dc.b	$9E,$05,$D3,$FB,$96,$E3,$EA,$77,$9F,$FA,$06,$1F,$84,$FF,$AD,$DF
	.dc.b	$AE,$63,$E5,$FF,$B1,$B4,$CF,$E1,$5C,$5F,$B4,$7F,$A2,$E7,$E4,$3F
	.dc.b	$E6,$96,$BB,$6B,$E1,$6B,$FC,$23,$F5,$9F,$CE,$E5,$FE,$5F,$D9,$18
	.dc.b	$C7,$FD,$31,$18,$EF,$F2,$E1,$E0,$BF,$A3,$B7,$F0,$E6,$FC,$BF,$F1
	.dc.b	$B8,$C6,$1F,$F0,$01,$FC,$DE,$5E,$AD,$FE,$9D,$7F,$21,$FC,$C7,$FD
	.dc.b	$75,$3F,$82,$6B,$BA,$E6,$FF,$BD,$3F,$0C,$7D,$47,$F2,$97,$FB,$7C
	.dc.b	$75,$6E,$5F,$86,$77,$1F,$8E,$73,$97,$FA,$D7,$24,$62,$7B,$7D,$3F
	.dc.b	$5C,$3E,$AF,$57,$FA,$CE,$7E,$CB,$5F,$AA,$1C,$63,$FE,$8B,$8C,$7E
	.dc.b	$03,$CB,$EF,$3F,$CB,$34,$69,$27,$FC,$E5,$F9,$8F,$8F,$DC,$7E,$17
	.dc.b	$FE,$2C,$DB,$FD,$E0,$DA,$FF,$67,$57,$A7,$6F,$F0,$C7,$B1,$FE,$21
	.dc.b	$F9,$DF,$EB,$7A,$FF,$89,$5B,$8F,$C1,$EC,$FC,$BE,$3F,$EE,$B9,$7C
	.dc.b	$AF,$12,$47,$FB,$06,$38,$AC,$7F,$C4,$DF,$53,$FD,$6E,$FF,$9C,$4E
	.dc.b	$5B,$3A,$8F,$DF,$FC,$DF,$F9,$A1,$BF,$29,$FD,$8E,$31,$1F,$87,$AE
	.dc.b	$6B,$F0,$AF,$C1,$B7,$A7,$F5,$31,$8F,$C3,$3E,$FC,$FD,$48,$7A,$FF
	.dc.b	$E8,$86,$C7,$FA,$7D,$AF,$FB,$4D,$AF,$EC,$65,$F7,$1F,$B6,$6C,$7E
	.dc.b	$E9,$AF,$7A,$FE,$7D,$F5,$3F,$D6,$F8,$7F,$5C,$BF,$08,$C9,$AF,$66
	.dc.b	$DF,$F3,$AF,$09,$5B,$8C,$35,$FE,$07,$1A,$9F,$E9,$9F,$82,$BE,$3E
	.dc.b	$FC,$F4,$BE,$93,$97,$B7,$C9,$BE,$A7,$FA,$A3,$63,$31,$C5,$F9,$5D
	.dc.b	$1F,$38,$8F,$F1,$AE,$CE,$0F,$CF,$EA,$F9,$C4,$7E,$73,$FE,$B8,$E1
	.dc.b	$FE,$6D,$00,$E1,$BE,$7F,$DC,$9F,$ED,$1F,$85,$FF,$59,$9F,$F1,$0A
	.dc.b	$FA,$6C,$9A,$F7,$BF,$B7,$7E,$AA,$FA,$9A,$4B,$F9,$D7,$C5,$FA,$1F
	.dc.b	$85,$BE,$3B,$FB,$B2,$FA,$8A,$62,$3F,$BD,$7D,$F3,$E0,$A9,$2F,$9D
	.dc.b	$7E,$8F,$AB,$FC,$27,$E0,$7D,$99,$FE,$89,$E1,$FF,$A3,$67,$27,$F8
	.dc.b	$63,$C2,$FE,$C1,$06,$98,$AF,$C1,$77,$7F,$62,$5F,$80,$7E,$C4,$98
	.dc.b	$7F,$6C,$8F,$E7,$1C,$7F,$BF,$9F,$D7,$7D,$E4,$AD,$47,$EE,$7B,$3C
	.dc.b	$E4,$8F,$EA,$47,$EE,$DF,$37,$CD,$7E,$DC,$91,$DE,$3F,$81,$7E,$17
	.dc.b	$F0,$09,$1C,$B4,$FF,$56,$47,$BF,$E6,$FF,$80,$12,$7D,$11,$FE,$AF
	.dc.b	$EE,$FE,$1F,$F0,$FF,$EB,$2F,$A4,$C1,$FC,$3F,$A3,$F8,$47,$C9,$CC
	.dc.b	$3F,$86,$BA,$7F,$C1,$BF,$83,$FF,$1C,$7C,$4D,$7F,$C9,$2D,$7E,$AB
	.dc.b	$5F,$E8,$DC,$7F,$E7,$4E,$3F,$F0,$8B,$78,$B3,$F5,$5F,$E7,$BB,$E9
	.dc.b	$5F,$E1,$67,$F3,$FB,$9A,$4C,$F8,$72,$7F,$96,$23,$FC,$22,$47,$F0
	.dc.b	$9F,$29,$FC,$94,$99,$FB,$23,$FC,$99,$F4,$0F,$BD,$A4,$9F,$E1,$91
	.dc.b	$F9,$F7,$F8,$4F,$E1,$1F,$E8,$E7,$FC,$2D,$E5,$A3,$11,$FE,$02,$6B
	.dc.b	$FE,$90,$6F,$13,$ED,$7F,$CB,$FC,$0F,$C2,$FF,$B5,$37,$DD,$FF,$74
	.dc.b	$4F,$F9,$93,$83,$F3,$9E,$D7,$F3,$F2,$65,$FE,$5A,$BE,$0F,$A3,$9F
	.dc.b	$FC,$55,$F5,$DF,$E7,$28,$FF,$50,$ED,$FE,$F0,$51,$9D,$BF,$DF,$B5
	.dc.b	$C3,$E4,$FF,$C3,$76,$FB,$8C,$61,$CB,$F0,$1F,$F1,$B6,$5F,$0A,$91
	.dc.b	$89,$E2,$D7,$81,$FC,$66,$7F,$C7,$DC,$0F,$C2,$FF,$C3,$3A,$79,$E1
	.dc.b	$F7,$BF,$EF,$48,$C3,$73,$FE,$45,$3C,$7E,$A7,$CB,$8B,$E0,$FF,$C3
	.dc.b	$B8,$7C,$18,$7D,$E7,$FA,$5A,$30,$DC,$3F,$CA,$15,$C1,$FF,$29,$47
	.dc.b	$F9,$DE,$77,$D1,$DA,$31,$3A,$F8,$AC,$7A,$5F,$EC,$29,$34,$9C,$7F
	.dc.b	$C1,$32,$F1,$FC,$2F,$F8,$E1,$EE,$D7,$F9,$F5,$7F,$89,$74,$E0,$D8
	.dc.b	$FE,$1A,$BB,$9C,$3F,$D1,$75,$F6,$C6,$2B,$EF,$E9,$EC,$F5,$B1,$5A
	.dc.b	$BE,$13,$F0,$9C,$BB,$73,$5F,$D4,$77,$EF,$18,$8F,$9B,$7C,$1F,$F5
	.dc.b	$D7,$13,$C7,$C7,$13,$F3,$5F,$F7,$9E,$1D,$23,$F0,$53,$A9,$CD,$FA
	.dc.b	$0F,$D5,$78,$BE,$BB,$FD,$8F,$7A,$31,$3F,$B1,$6B,$C2,$7F,$20,$FE
	.dc.b	$D7,$C9,$F3,$71,$FE,$E5,$B6,$F2,$6C,$7A,$FB,$BF,$A6,$7C,$1F,$CB
	.dc.b	$F1,$BF,$8D,$B7,$2F,$CE,$7B,$5F,$4D,$FF,$1B,$C6,$73,$FF,$6E,$7A
	.dc.b	$1F,$D4,$27,$F7,$15,$B9,$FF,$4B,$F0,$7F,$65,$F3,$78,$52,$B5,$8F
	.dc.b	$FA,$47,$F0,$CF,$5E,$1F,$C4,$9A,$FF,$A4,$9A,$B7,$46,$1F,$E3,$1E
	.dc.b	$7F,$95,$B8,$7F,$8A,$F8,$1F,$07,$95,$5E,$BD,$FF,$B7,$73,$DF,$3B
	.dc.b	$F6,$18,$7F,$48,$F0,$31,$3B,$F3,$51,$FE,$7D,$D7,$FD,$15,$18,$AF
	.dc.b	$E9,$B3,$99,$FF,$33,$4E,$5F,$E9,$FF,$6F,$CC,$E9,$F0,$56,$FF,$CC
	.dc.b	$BE,$A7,$8C,$F0,$79,$B6,$3F,$AD,$BD,$74,$FE,$98,$DC,$4D,$35,$A7
	.dc.b	$F8,$76,$BF,$CE,$3E,$B7,$DB,$E6,$79,$9F,$E9,$99,$C9,$FE,$35,$E5
	.dc.b	$FD,$84,$EB,$C6,$E1,$5B,$DF,$FD,$79,$9F,$F2,$8C,$38,$1F,$C7,$27
	.dc.b	$58,$7F,$BE,$3F,$68,$FC,$CB,$6F,$E6,$79,$BF,$4E,$73,$7F,$32,$3D
	.dc.b	$84,$7F,$BC,$BF,$06,$1E,$47,$13,$EB,$7B,$96,$BF,$C0,$2F,$C5,$23
	.dc.b	$FE,$02,$7E,$47,$CE,$78,$7E,$7B,$EF,$36,$7F,$C1,$2E,$1F,$F0,$5F
	.dc.b	$D8,$5C,$BF,$DB,$98,$FF,$3E,$9F,$83,$1F,$EA,$58,$C3,$F1,$BE,$23
	.dc.b	$DD,$E8,$7E,$2D,$FB,$04,$AF,$F0,$AB,$F8,$9E,$E7,$C0,$7C,$4B,$FD
	.dc.b	$13,$63,$FD,$95,$F5,$BF,$A1,$8C,$7A,$AF,$AB,$FD,$22,$9B,$18,$7F
	.dc.b	$51,$F2,$CD,$A9,$FA,$BF,$C3,$7F,$6F,$A7,$F9,$DF,$25,$6E,$3F,$D3
	.dc.b	$13,$FE,$C8,$6F,$A7,$C7,$D1,$E4,$FF,$09,$D5,$E5,$BD,$6C,$36,$FE
	.dc.b	$4C,$BF,$0D,$78,$9A,$BE,$51,$8A,$FE,$A4,$DA,$FF,$2B,$B5,$FF,$47
	.dc.b	$1F,$E8,$F7,$0E,$1F,$DA,$75,$7F,$BD,$DB,$3C,$3F,$0D,$7D,$CF,$83
	.dc.b	$C7,$FE,$1D,$FB,$97,$C1,$C7,$FD,$66,$DC,$7F,$AB,$70,$BF,$0E,$E3
	.dc.b	$97,$FA,$F6,$B4,$FC,$B8,$7D,$DF,$F1,$97,$78,$7F,$AF,$9C,$CF,$FB
	.dc.b	$15,$CE,$1D,$BF,$0E,$FF,$5D,$4F,$FB,$25,$B5,$FE,$C6,$7F,$9B,$37
	.dc.b	$E3,$BF,$93,$46,$2B,$94,$63,$D3,$CB,$F5,$F6,$FF,$55,$61,$F5,$1B
	.dc.b	$82,$FF,$BA,$2F,$C3,$9F,$2B,$C7,$E2,$3F,$01,$18,$C3,$FC,$10,$DA
	.dc.b	$FF,$46,$B5,$FA,$E7,$3A,$FE,$54,$F5,$FA,$1E,$67,$E2,$FE,$33,$FD
	.dc.b	$EB,$7F,$88,$B4,$FF,$0B,$FA,$39,$3F,$B4,$F4,$7F,$D5,$9F,$82,$63
	.dc.b	$1F,$54,$FF,$9F,$BD,$BF,$E8,$53,$FD,$0D,$B1,$C7,$C7,$7D,$47,$F1
	.dc.b	$73,$FA,$47,$89,$67,$B7,$A6,$FB,$6C,$C6,$1B,$97,$80,$FF,$71,$71
	.dc.b	$FD,$8E,$3F,$F0,$F7,$CE,$BF,$10,$FE,$01,$4A,$D6,$1F,$EE,$9A,$D4
	.dc.b	$FE,$B7,$F0,$FF,$F0,$54,$FF,$8D,$5B,$1F,$EB,$A6,$BF,$39,$AF,$E9
	.dc.b	$9F,$F0,$44,$7E,$C9,$B1,$E7,$0B,$E0,$7C,$1F,$05,$FD,$AF,$F0,$FF
	.dc.b	$E8,$55,$ED,$E3,$46,$34,$FF,$1D,$46,$3D,$6E,$1F,$19,$C5,$B9,$FD
	.dc.b	$7B,$FE,$AA,$D9,$FE,$F1,$F8,$7F,$F7,$A8,$FB,$9F,$47,$FE,$56,$7E
	.dc.b	$F4,$67,$2F,$14,$E5,$F2,$46,$72,$F1,$B6,$92,$C9,$3E,$F9,$EB,$F0
	.dc.b	$5F,$E0,$FD,$7F,$C0,$F7,$F6,$FC,$5F,$89,$F0,$5F,$C7,$FA,$5E,$CE
	.dc.b	$E3,$0F,$E7,$23,$38,$FF,$9A,$DB,$7F,$EF,$06,$BF,$A4,$7F,$D7,$D1
	.dc.b	$FC,$99,$B1,$F5,$7F,$0F,$FF,$2B,$4E,$7E,$4F,$FC,$3D,$F2,$0F,$9E
	.dc.b	$BF,$31,$86,$D7,$F5,$8D,$3F,$3B,$E8,$7F,$51,$EF,$FE,$69,$9F,$ED
	.dc.b	$27,$35,$FB,$68,$FF,$1A,$4E,$9F,$E5,$FC,$4F,$D0,$3F,$83,$F2,$7E
	.dc.b	$01,$B7,$FE,$A0,$FC,$3F,$EE,$DF,$E1,$E7,$2F,$D1,$FC,$3D,$F2,$DF
	.dc.b	$E3,$76,$DF,$F6,$EF,$87,$FF,$0F,$FD,$3E,$6B,$FC,$79,$F1,$0F,$85
	.dc.b	$FF,$13,$36,$FD,$1A,$A3,$19,$7F,$10,$8C,$57,$F8,$B7,$F0,$FF,$E2
	.dc.b	$DB,$7E,$61,$F0,$D7,$E7,$5C,$E1,$F8,$8B,$46,$AB,$F3,$FF,$30,$F2
	.dc.b	$FF,$CA,$0B,$A6,$C6,$1F,$F2,$D7,$DB,$7D,$CC,$BF,$DE,$AD,$7F,$B9
	.dc.b	$35,$FF,$32,$37,$1F,$ED,$15,$FE,$5D,$F6,$FC,$19,$FC,$7F,$C3,$FF
	.dc.b	$35,$87,$ED,$67,$2F,$ED,$FE,$6F,$FC,$6F,$5E,$B6,$C3,$89,$F4,$DF
	.dc.b	$DB,$F6,$FD,$DD,$6B,$37,$E1,$FF,$1B,$0F,$C1,$B5,$F8,$CD,$7F,$5C
	.dc.b	$D8,$7C,$1F,$F8,$3B,$2F,$E7,$0F,$F5,$77,$E7,$39,$AF,$E1,$CE,$5F
	.dc.b	$DC,$6B,$EC,$C7,$F8,$7B,$6B,$FC,$B9,$F7,$DF,$E1,$7A,$FE,$2D,$39
	.dc.b	$5E,$6B,$F0,$DF,$F5,$CC,$7F,$82,$A3,$1D,$BF,$E7,$D6,$E1,$FA,$36
	.dc.b	$CF,$E2,$FE,$7D,$F3,$FF,$E8,$F7,$FA,$25,$7F,$53,$72,$DF,$4F,$FD
	.dc.b	$29,$87,$F9,$DE,$FF,$95,$71,$7F,$90,$77,$FF,$51,$7C,$A7,$FA,$66
	.dc.b	$31,$3F,$E1,$C3,$F8,$C1,$67,$F8,$DF,$C3,$7F,$A1,$E3,$5F,$24,$F7
	.dc.b	$FF,$A5,$C7,$F9,$A9,$B6,$FC,$3B,$C5,$C7,$FE,$F1,$BE,$1F,$03,$FE
	.dc.b	$C3,$E1,$FF,$1D,$FC,$6F,$FB,$16,$5F,$8C,$7F,$98,$2F,$FA,$DF,$C5
	.dc.b	$FF,$88,$34,$FF,$3B,$CE,$78,$BF,$AD,$4B,$E5,$3E,$63,$63,$E6,$36
	.dc.b	$BF,$C0,$B9,$7F,$92,$AB,$7F,$25,$E3,$AF,$F3,$3D,$FF,$9C,$AB,$FC
	.dc.b	$9F,$E9,$66,$BC,$DF,$A0,$FE,$1F,$43,$7E,$67,$F3,$F1,$8F,$89,$F6
	.dc.b	$47,$FA,$7F,$4F,$F2,$5C,$A3,$11,$FD,$22,$31,$E4,$FF,$B0,$5B,$9F
	.dc.b	$F2,$75,$F9,$AF,$9A,$BF,$C0,$51,$8C,$BF,$D6,$9F,$06,$FB,$0F,$99
	.dc.b	$7B,$E2,$FF,$59,$CE,$71,$FE,$44,$DC,$79,$FE,$23,$FB,$59,$FE,$B7
	.dc.b	$CE,$AD,$5F,$DB,$FC,$47,$EE,$D7,$A7,$DB,$FE,$97,$39,$ED,$FB,$91
	.dc.b	$89,$F7,$57,$0C,$BF,$D6,$2D,$D9,$F4,$BF,$E9,$75,$1B,$E5,$7D,$4F
	.dc.b	$F0,$C5,$F8,$87,$F4,$ED,$3F,$C4,$33,$98,$C6,$5F,$E1,$F8,$E0,$4F
	.dc.b	$97,$FC,$93,$63,$F6,$4A,$3F,$6C,$D8,$CC,$7F,$93,$33,$FE,$FF,$E2
	.dc.b	$7F,$80,$3F,$01,$FF,$39,$C7,$DE,$AC,$7C,$AF,$F5,$FA,$B3,$91,$5C
	.dc.b	$1F,$F4,$7C,$69,$FC,$25,$57,$B9,$C2,$BF,$CB,$FA,$A7,$52,$AF,$ED
	.dc.b	$CF,$42,$99,$2B,$AF,$C4,$5F,$88,$3F,$08,$C3,$6B,$FB,$C4,$EA,$72
	.dc.b	$71,$FA,$7F,$F3,$E4,$E9,$AF,$F9,$D2,$51,$88,$FF,$56,$C6,$37,$E4
	.dc.b	$7C,$15,$F8,$67,$FA,$CB,$5F,$8D,$18,$AD,$5F,$82,$72,$DD,$7A,$DF
	.dc.b	$16,$D7,$0C,$3E,$9F,$94,$FD,$E9,$F6,$FF,$22,$30,$D8,$CD,$7F,$86
	.dc.b	$67,$2D,$87,$91,$FE,$C8,$BF,$F5,$DC,$62,$7F,$C4,$CD,$7F,$C8,$0F
	.dc.b	$46,$CF,$1E,$0F,$F7,$44,$FE,$92,$78,$F3,$E5,$61,$E9,$BC,$71,$7F
	.dc.b	$23,$F8,$AB,$E3,$FF,$9B,$EF,$FC,$AB,$9F,$F3,$38,$C4,$7F,$0A,$53
	.dc.b	$C3,$E2,$A4,$F0,$D3,$F8,$B5,$70,$BE,$57,$E0,$FC,$41,$F2,$14,$98
	.dc.b	$71,$19,$25,$F1,$FA,$9F,$F4,$E7,$83,$FE,$FD,$FA,$7F,$73,$63,$95
	.dc.b	$F1,$C3,$DF,$F8,$7B,$E4,$FF,$3F,$28,$C5,$7F,$29,$6B,$FA,$86,$BF
	.dc.b	$CD,$17,$D7,$F4,$24,$78,$23,$8E,$BF,$D6,$FF,$17,$FF,$32,$4F,$91
	.dc.b	$AF,$F3,$46,$C6,$FE,$3F,$EA,$C3,$F0,$6F,$3B,$6B,$F9,$F3,$5F,$DE
	.dc.b	$B5,$FE,$93,$FA,$BF,$F5,$F7,$E1,$F9,$0F,$90,$FC,$0B,$63,$E3,$CE
	.dc.b	$DF,$E8,$FF,$88,$7F,$B7,$9F,$96,$FF,$BC,$1B,$F4,$5F,$8A,$8D,$47
	.dc.b	$F4,$EF,$81,$FE,$62,$FF,$53,$6B,$FC,$49,$AF,$F5,$E6,$BF,$E5,$DF
	.dc.b	$11,$F5,$7F,$99,$F7,$BF,$87,$36,$33,$3F,$EE,$F5,$EC,$EF,$F0,$FF
	.dc.b	$F3,$14,$FF,$BC,$DA,$FF,$B1,$DB,$C3,$FE,$C6,$7F,$CC,$D1,$FE,$F9
	.dc.b	$FC,$D3,$FD,$E2,$FF,$9E,$9A,$FF,$9E,$5A,$FF,$A4,$9A,$FF,$9D,$71
	.dc.b	$7D,$07,$A5,$7D,$6D,$31,$1E,$DF,$C3,$9C,$B4,$97,$FA,$3B,$4F,$DF
	.dc.b	$CB,$FD,$BB,$F1,$DF,$5E,$BF,$D4,$6D,$7F,$D0,$CD,$7F,$D9,$CD,$AF
	.dc.b	$F4,$C7,$C2,$7E,$A3,$E9,$7C,$B5,$FE,$9F,$6B,$FE,$ED,$F9,$34,$FD
	.dc.b	$3F,$91,$FE,$59,$3F,$B2,$F4,$BF,$C1,$4F,$FA,$9D,$AF,$FB,$D6,$37
	.dc.b	$97,$FB,$D3,$E4,$3E,$CC,$FF,$A8,$DA,$FF,$AF,$1A,$FF,$A7,$DA,$FF
	.dc.b	$B9,$FF,$0E,$7C,$3F,$E1,$5B,$1F,$1D,$AA,$BF,$B0,$3C,$7F,$02,$FC
	.dc.b	$67,$E1,$DF,$DA,$E7,$3B,$23,$1E,$67,$FA,$BD,$B7,$FE,$F7,$5F,$5F
	.dc.b	$F9,$5F,$C3,$9F,$61,$FD,$AB,$04,$62,$FF,$D8,$6D,$7F,$D8,$4D,$7F
	.dc.b	$DB,$7F,$0E,$FB,$2F,$E3,$7F,$87,$7F,$0A,$CF,$FC,$5A,$DA,$61,$FE
	.dc.b	$CD,$AF,$D3,$7D,$6F,$EE,$E7,$F9,$2F,$E1,$DB,$F3,$CF,$1B,$56,$F7
	.dc.b	$F8,$3F,$8B,$BE,$E7,$D4,$DC,$7F,$DB,$6D,$7F,$DB,$4D,$7F,$DE,$7F
	.dc.b	$87,$FF,$05,$7E,$86,$BF,$ED,$78,$C5,$FC,$E6,$C6,$7E,$87,$FA,$B5
	.dc.b	$7F,$3F,$C2,$FA,$DD,$F0,$FB,$33,$93,$F4,$9F,$85,$FF,$BB,$5F,$E9
	.dc.b	$03,$F0,$DF,$F1,$32,$21,$9F,$81,$FF,$0E,$C6,$A3,$2D,$78,$3F,$C0
	.dc.b	$2B,$C2,$DC,$B7,$F5,$1E,$88,$FC,$2E,$EF,$C9,$7F,$B7,$E1,$FB,$96
	.dc.b	$BF,$10,$FB,$2E,$30,$FC,$8B,$FC,$53,$0F,$F7,$F3,$96,$F7,$7F,$23
	.dc.b	$8F,$E2,$F5,$CE,$FC,$1F,$01,$FC,$9F,$EB,$FF,$B6,$CF,$F0,$96,$BB
	.dc.b	$6E,$3F,$A0,$7F,$B6,$C6,$FD,$CE,$44,$79,$7B,$9F,$E0,$BB,$7E,$8C
	.dc.b	$F9,$0C,$36,$BF,$BD,$4E,$A7,$F9,$D7,$B1,$F4,$A7,$E1,$0E,$DF,$F0
	.dc.b	$6B,$96,$BF,$61,$AF,$DB,$6B,$C7,$E0,$3F,$13,$1F,$E5,$48,$FC,$1B
	.dc.b	$5E,$EF,$F8,$61,$FE,$59,$3F,$D6,$34,$FE,$02,$FE,$52,$3F,$5A,$DD
	.dc.b	$7F,$B4,$79,$5E,$AF,$5F,$D7,$D3,$2F,$9C,$7F,$8B,$3D,$3F,$CA,$47
	.dc.b	$FA,$C1,$B8,$FF,$B7,$5B,$C9,$FE,$BB,$5F,$5B,$C7,$C4,$97,$FE,$1A
	.dc.b	$9F,$E1,$17,$C7,$5F,$D3,$FD,$A7,$F4,$8A,$FD,$0B,$5F,$9A,$D9,$FD
	.dc.b	$4F,$E1,$7F,$BF,$CB,$F4,$7F,$86,$BE,$77,$FC,$FE,$DC,$3F,$D0,$6D
	.dc.b	$7E,$C0,$70,$FD,$B5,$F6,$3F,$BA,$6C,$7E,$D1,$AF,$E0,$1E,$98,$7F
	.dc.b	$30,$FB,$7F,$D4,$93,$BF,$FC,$00,$FF,$2E,$AF,$EC,$4F,$F4,$C9,$E1
	.dc.b	$F3,$7A,$37,$5F,$92,$F8,$AF,$C9,$F3,$CC,$3F,$27,$D5,$FE,$53,$D7
	.dc.b	$F8,$27,$57,$E6,$AB,$F3,$3E,$9C,$BD,$C7,$E7,$3C,$09,$1C,$9F,$2F
	.dc.b	$4F,$F1,$8F,$35,$F4,$5F,$EA,$86,$D7,$FA,$B1,$AF,$CA,$6B,$FA,$35
	.dc.b	$F8,$4F,$32,$47,$F1,$C8,$E3,$C1,$F1,$FA,$FE,$2F,$64,$AF,$8B,$15
	.dc.b	$FE,$09,$C7,$FE,$4C,$DF,$F9,$17,$6A,$31,$3F,$D2,$95,$FD,$0F,$6B
	.dc.b	$E8,$FC,$34,$AF,$A3,$15,$EB,$E8,$7D,$63,$19,$FD,$BF,$A0,$FE,$C5
	.dc.b	$89,$A4,$AF,$C5,$E5,$E3,$F8,$EF,$BE,$48,$FB,$F3,$F8,$2F,$8F,$FC
	.dc.b	$1F,$C8,$C9,$FC,$1C,$57,$DF,$17,$1F,$C6,$4F,$E2,$B8,$7C,$64,$AF
	.dc.b	$0E,$5F,$41,$BC,$5F,$90,$23,$F9,$09,$FF,$29,$75,$7F,$52,$8F,$E9
	.dc.b	$7A,$23,$11,$F3,$9A,$FF,$76,$6C,$6F,$EA,$39,$92,$7F,$C7,$D3,$FD
	.dc.b	$7B,$2F,$F0,$9A,$F9,$8F,$09,$27,$7E,$1D,$7F,$A1,$4F,$F0,$A7,$9F
	.dc.b	$8D,$9B,$E7,$E3,$FF,$85,$6B,$FC,$0D,$5E,$A6,$BE,$98,$D7,$8F,$F0
	.dc.b	$09,$1E,$68,$F8,$7E,$F4,$8F,$4B,$F1,$BF,$02,$F8,$FF,$61,$DC,$93
	.dc.b	$F8,$0C,$11,$E3,$F5,$3E,$69,$27,$E1,$C7,$F4,$FF,$C1,$3F,$C2,$F1
	.dc.b	$FD,$31,$AF,$F6,$BF,$0D,$18,$D3,$FC,$7F,$18,$F4,$F9,$7E,$05,$FE
	.dc.b	$F7,$A3,$49,$7F,$EF,$7F,$CB,$DF,$57,$F8,$0F,$4A,$4F,$F5,$C9,$CB
	.dc.b	$F6,$23,$F0,$BF,$2F,$23,$07,$9F,$E1,$BF,$C2,$3F,$3D,$C3,$4F,$3C
	.dc.b	$63,$1F,$E2,$79,$A3,$17,$F8,$9F,$5A,$DF,$BB,$E5,$FF,$C3,$7E,$3F
	.dc.b	$F8,$72,$31,$5F,$E1,$F8,$C6,$1F,$EE,$A8,$C1,$DA,$B8,$BF,$C8,$B5
	.dc.b	$FF,$05,$E7,$EC,$EC,$FC,$E7,$A5,$E2,$6D,$7F,$8A,$E7,$D5,$5C,$35
	.dc.b	$FF,$16,$FD,$97,$F9,$76,$7E,$2D,$23,$13,$AE,$AF,$6C,$FF,$82,$79
	.dc.b	$AD,$7F,$96,$6F,$8B,$6F,$AF,$56,$B0,$DF,$5F,$F8,$B3,$1F,$0B,$5F
	.dc.b	$F2,$05,$FF,$6C,$E0,$FE,$B3,$1E,$39,$D3,$E5,$EC,$F8,$CF,$F8,$56
	.dc.b	$BF,$C2,$9B,$F0,$6C,$7F,$62,$6B,$BE,$DF,$F4,$6E,$9E,$86,$E9,$F1
	.dc.b	$3B,$AD,$79,$26,$BF,$16,$2D,$9C,$FB,$0E,$4C,$C7,$2D,$5F,$BF,$7F
	.dc.b	$B8,$27,$EA,$65,$FC,$47,$D6,$D4,$E7,$26,$9F,$77,$BF,$FC,$01,$87
	.dc.b	$09,$62,$E4,$7D,$7F,$FA,$A2,$BF,$1D,$18,$7F,$17,$5C,$23,$0F,$F9
	.dc.b	$DA,$B9,$57,$18,$FC,$74,$7A,$67,$BD,$78,$79,$BF,$95,$7D,$4E,$6B
	.dc.b	$E3,$C7,$F2,$DF,$9D,$FF,$5C,$F8,$3F,$C9,$3E,$32,$E1,$FC,$BE,$6E
	.dc.b	$5F,$CB,$E6,$E1,$F1,$BE,$57,$38,$F3,$C3,$FC,$AB,$BA,$73,$B3,$2E
	.dc.b	$3F,$9C,$BF,$05,$FE,$7B,$92,$31,$8F,$F8,$92,$31,$3A,$9C,$E9,$FE
	.dc.b	$5F,$9E,$1A,$74,$3A,$FC,$75,$C3,$FA,$0B,$E2,$3C,$78,$58,$FE,$8A
	.dc.b	$FF,$4E,$D7,$FA,$83,$73,$EF,$8F,$F3,$B8,$D7,$E5,$7F,$3A,$3F,$D0
	.dc.b	$CD,$DF,$FD,$BE,$FD,$86,$E7,$F1,$A1,$EC,$D3,$1C,$D7,$3F,$F5,$65
	.dc.b	$F2,$8E,$E4,$8F,$EF,$B5,$FC,$A9,$BF,$63,$D7,$E5,$7B,$36,$39,$9F
	.dc.b	$E1,$69,$46,$A3,$39,$7F,$B8,$25,$F2,$D9,$8F,$6F,$D1,$FA,$4F,$57
	.dc.b	$D2,$5B,$E3,$C0,$C1,$3F,$18,$93,$E8,$7E,$4F,$C0,$7C,$33,$E2,$7C
	.dc.b	$05,$9E,$74,$8F,$9E,$48,$F8,$2F,$D2,$F8,$2F,$98,$7C,$DF,$A3,$B3
	.dc.b	$C9,$B9,$FD,$66,$F0,$7D,$6F,$A3,$FA,$87,$D5,$FA,$7B,$7E,$27,$EF
	.dc.b	$91,$FA,$2F,$E0,$BF,$0D,$FB,$9D,$9F,$75,$B3,$EA,$F8,$BC,$34,$F6
	.dc.b	$C7,$E0,$FC,$5F,$C2,$7D,$7D,$9F,$82,$FE,$1F,$E9,$31,$AF,$E3,$9B
	.dc.b	$A7,$FB,$E3,$E7,$3F,$BF,$DF,$FB,$D9,$B5,$FE,$C2,$5E,$D7,$EE,$A1
	.dc.b	$1A,$AE,$53,$FE,$2C,$8D,$D7,$F3,$C6,$D7,$FA,$59,$AF,$F6,$5F,$C3
	.dc.b	$7F,$B0,$DF,$F8,$11,$B5,$FE,$64,$FA,$AB,$5F,$C8,$A6,$F7,$38,$9C
	.dc.b	$5F,$C3,$17,$15,$8C,$39,$E3,$C6,$5F,$EE,$2A,$46,$31,$FF,$7C,$64
	.dc.b	$F3,$BF,$8A,$B6,$FF,$2E,$FF,$80,$DB,$1A,$F5,$E4,$6B,$DA,$B7,$FE
	.dc.b	$EC,$FC,$0D,$F0,$DF,$2D,$B5,$FA,$A9,$43,$E2,$35,$5B,$7F,$CB,$31
	.dc.b	$9A,$FE,$D2,$DA,$FF,$61,$B5,$FF,$7E,$7E,$08,$F6,$7E,$D3,$6B,$C0
	.dc.b	$FD,$B5,$32,$34,$FC,$4F,$9A,$F8,$67,$C4,$F8,$0B,$5F,$11,$C3,$F1
	.dc.b	$BC,$AB,$4E,$6F,$ED,$0B,$C3,$7D,$87,$F1,$CB,$FE,$06,$FF,$21,$7F
	.dc.b	$D4,$AE,$67,$94,$63,$D3,$FE,$93,$3F,$C6,$E7,$27,$57,$85,$F8,$2F
	.dc.b	$BF,$95,$C3,$CB,$FC,$E7,$E0,$6F,$02,$F8,$18,$6E,$9F,$56,$5E,$0F
	.dc.b	$FA,$29,$B1,$FE,$1A,$6E,$9F,$EF,$46,$BF,$E4,$3E,$B7,$99,$F9,$73
	.dc.b	$FB,$0F,$BB,$FE,$60,$BF,$CA,$47,$F6,$4F,$C0,$DF,$0B,$4C,$7C,$2D
	.dc.b	$AF,$E5,$33,$F6,$71,$F9,$8D,$E0,$3F,$6B,$87,$E1,$39,$FC,$0F,$F9
	.dc.b	$7D,$FF,$32,$5E,$7F,$F7,$47,$38,$7C,$CC,$11,$AA,$E1,$E7,$EC,$E1
	.dc.b	$E7,$D8,$F9,$1C,$BF,$38,$F9,$FC,$B6,$3E,$63,$F2,$C8,$FC,$B7,$FD
	.dc.b	$6F,$F8,$1F,$FB,$CA,$3F,$4C,$DC,$BC,$9F,$39,$ED,$9F,$CC,$C2,$75
	.dc.b	$B7,$E0,$7E,$09,$F2,$5F,$E1,$06,$DF,$F5,$27,$FD,$DE,$D8,$FF,$0A
	.dc.b	$FA,$BF,$6B,$0F,$F7,$BB,$5F,$CD,$37,$7F,$EC,$BD,$1F,$E0,$3F,$C0
	.dc.b	$7C,$38,$7E,$B3,$E8,$9F,$81,$7F,$24,$D3,$F2,$95,$AC,$95,$AF,$4B
	.dc.b	$FC,$4B,$F8,$17,$9A,$7F,$3B,$3F,$BB,$F8,$8C,$ED,$FE,$2E,$8C,$67
	.dc.b	$FD,$B3,$E2,$3F,$8C,$5F,$C4,$3F,$60,$E1,$FF,$73,$B5,$FE,$03,$B7
	.dc.b	$F9,$BA,$78,$47,$ED,$FF,$02,$7C,$DF,$F8,$E1,$B3,$FD,$E9,$7C,$D6
	.dc.b	$67,$FC,$7B,$E7,$FF,$8C,$38,$3F,$64,$DB,$EF,$F1,$5F,$E2,$17,$C3
	.dc.b	$19,$C7,$F8,$37,$D4,$FF,$BB,$E7,$C7,$9A,$31,$3E,$D8,$C7,$C6,$7F
	.dc.b	$8C,$E7,$33,$F8,$49,$FE,$B3,$EC,$FC,$7F,$C0,$BF,$3B,$9F,$F0,$5F
	.dc.b	$73,$C5,$7F,$4A,$73,$97,$C4,$FC,$07,$BD,$FF,$1A,$E2,$7E,$04,$EC
	.dc.b	$FF,$2D,$D6,$AF,$F0,$13,$97,$FC,$3F,$F8,$17,$DB,$CB,$C8,$D7,$E8
	.dc.b	$B5,$FE,$4C,$DD,$7F,$D2,$5F,$3B,$FE,$68,$F9,$13,$7D,$9C,$3F,$E6
	.dc.b	$CF,$4F,$D7,$4F,$DD,$1F,$15,$9E,$3F,$F3,$8C,$63,$4F,$F1,$9F,$BD
	.dc.b	$FD,$DB,$5F,$C1,$B5,$FE,$70,$DC,$3F,$B3,$78,$1F,$E6,$1F,$91,$FE
	.dc.b	$71,$5F,$CF,$5B,$3F,$E9,$EF,$23,$FC,$FF,$18,$CB,$F9,$3F,$2F,$F3
	.dc.b	$56,$E5,$F9,$27,$F9,$A3,$FE,$46,$73,$87,$F4,$91,$E0,$FF,$9C,$5B
	.dc.b	$3F,$86,$69,$F8,$52,$9F,$C1,$B3,$7C,$CB,$E7,$7F,$D2,$F1,$8E,$2B
	.dc.b	$39,$BF,$F2,$FC,$E5,$BF,$AA,$BC,$DE,$3F,$81,$7E,$C3,$5F,$EB,$F6
	.dc.b	$AD,$5F,$2A,$D3,$FD,$C2,$5F,$82,$7F,$A8,$63,$F6,$31,$8F,$8C,$FE
	.dc.b	$D3,$AF,$A6,$B4,$FE,$F6,$75,$5C,$27,$FD,$9D,$8F,$89,$B1,$DD,$F9
	.dc.b	$73,$E0,$F8,$CF,$83,$EE,$7F,$B7,$EB,$DE,$DC,$3F,$B6,$D7,$F6,$7F
	.dc.b	$0F,$7D,$7C,$1F,$2C,$FA,$EB,$F6,$2D,$8C,$BF,$E2,$D8,$CB,$FE,$2F
	.dc.b	$FC,$05,$ED,$FF,$1B,$F7,$DF,$01,$F5,$27,$39,$FF,$A4,$3C,$4F,$F0
	.dc.b	$4C,$E6,$E4,$FD,$6A,$FC,$0E,$FF,$49,$E3,$8F,$AB,$F0,$1F,$86,$FF
	.dc.b	$8F,$DB,$B1,$FC,$8C,$6F,$7F,$C2,$C6,$2F,$EE,$BC,$0C,$61,$F6,$DE
	.dc.b	$BD,$2F,$DE,$FC,$35,$FE,$22,$8C,$75,$7F,$B9,$27,$37,$FE,$32,$9C
	.dc.b	$BF,$EA,$E7,$96,$9B,$38,$78,$D6,$75,$5F,$C3,$F3,$FF,$6F,$35,$FE
	.dc.b	$CB,$7E,$3A,$FF,$74,$CE,$67,$F9,$FF,$67,$F9,$0B,$F0,$0F,$A3,$1F
	.dc.b	$A7,$C3,$1C,$1A,$84,$E9,$FC,$D4,$E9,$FE,$A1,$55,$C9,$C2,$B9,$57
	.dc.b	$18,$FE,$AE,$D5,$3F,$D4,$9F,$EF,$AD,$7F,$CA,$ED,$76,$7F,$58,$F8
	.dc.b	$BF,$DA,$7D,$D7,$16,$C6,$7B,$13,$96,$C7,$1F,$73,$FC,$B0,$DE,$0E
	.dc.b	$51,$88,$FE,$D9,$7F,$ED,$89,$C9,$EF,$7F,$C5,$71,$7D,$EF,$F0,$AE
	.dc.b	$5F,$1E,$78,$4F,$C3,$3F,$CE,$2E,$23,$53,$C3,$3F,$F0,$EC,$F4,$AE
	.dc.b	$01,$E1,$1F,$89,$FC,$01,$F4,$FF,$27,$85,$18,$BF,$EF,$B1,$89,$D4
	.dc.b	$E6,$7F,$C1,$13,$C3,$E9,$BF,$C5,$B9,$FE,$66,$91,$8E,$0F,$03,$6D
	.dc.b	$7D,$5F,$CD,$57,$07,$F3,$9F,$46,$FA,$5E,$14,$8C,$4F,$0B,$E9,$D2
	.dc.b	$F9,$DF,$F5,$CC,$FF,$34,$7F,$D8,$8D,$F8,$9D,$C6,$A3,$F8,$CB,$A6
	.dc.b	$C6,$8D,$FC,$9F,$A7,$BB,$F8,$9F,$D2,$5E,$B3,$4F,$1F,$D8,$BE,$9B
	.dc.b	$18,$FF,$46,$FA,$87,$E0,$DE,$86,$C7,$F1,$9C,$39,$BF,$D3,$3F,$01
	.dc.b	$FF,$6D,$BF,$D0,$1A,$FF,$3D,$6B,$FD,$65,$AF,$F4,$9F,$C7,$77,$CD
	.dc.b	$FD,$41,$BC,$D9,$F9,$CF,$93,$78,$BF,$F6,$F7,$E0,$5F,$EC,$F7,$FA
	.dc.b	$EB,$5F,$E7,$ED,$C3,$FA,$DB,$EA,$8F,$E8,$7F,$81,$7D,$37,$FA,$BB
	.dc.b	$5F,$EA,$8D,$7F,$B8,$35,$FE,$C1,$F2,$DF,$D0,$1F,$ED,$AF,$F7,$C6
	.dc.b	$F8,$DF,$80,$8D,$47,$F5,$2F,$C0,$BF,$CF,$6F,$F7,$45,$F5,$FF,$86
	.dc.b	$6C,$67,$E2,$BF,$13,$F8,$13,$E8,$FF,$86,$75,$A3,$15,$FD,$C1,$AF
	.dc.b	$F6,$D6,$BF,$E4,$BF,$23,$FD,$86,$FF,$BA,$36,$FF,$67,$46,$23,$FD
	.dc.b	$50,$D8,$CF,$A9,$FA,$5F,$6D,$F8,$13,$E5,$70,$DC,$3A,$72,$FF,$9D
	.dc.b	$23,$11,$FE,$B3,$5F,$03,$CF,$F0,$2F,$E5,$13,$FE,$ED,$6E,$06,$98
	.dc.b	$5F,$51,$E3,$7F,$86,$67,$FE,$E7,$6E,$3F,$E8,$47,$77,$E4,$3F,$DF
	.dc.b	$5F,$81,$3E,$93,$D8,$D8,$D4,$A3,$11,$FE,$CB,$7F,$CB,$2E,$BF,$02
	.dc.b	$7D,$67,$F3,$FA,$46,$27,$FD,$02,$D7,$FD,$00,$D7,$FD,$83,$E2,$7F
	.dc.b	$A4,$B3,$F9,$DF,$4E,$F8,$6A,$7E,$F7,$E4,$FD,$77,$FE,$AC,$F5,$BF
	.dc.b	$D3,$AF,$A3,$E1,$DE,$BF,$D7,$6C,$7F,$84,$67,$6F,$FB,$97,$E3,$DE
	.dc.b	$EF,$A5,$B1,$FE,$98,$6B,$FE,$97,$6B,$FE,$95,$F9,$07,$D6,$7E,$15
	.dc.b	$B1,$F2,$9A,$FF,$BC,$DF,$85,$CA,$E5,$FF,$5C,$FE,$05,$E4,$8C,$53
	.dc.b	$8D,$39,$AD,$FC,$AF,$E4,$E3,$FB,$D5,$F0,$A7,$DB,$3E,$8F,$EA,$37
	.dc.b	$1F,$F5,$9B,$5F,$F5,$93,$5F,$A7,$F8,$0F,$F2,$FB,$E9,$F3,$5F,$E6
	.dc.b	$F9,$FF,$6B,$7C,$87,$F2,$FA,$D6,$9F,$EE,$7F,$C0,$6D,$7A,$7E,$A7
	.dc.b	$F8,$22,$C7,$CD,$18,$7F,$D6,$93,$C3,$47,$E0,$3F,$CD,$6F,$FD,$92
	.dc.b	$D7,$FD,$90,$D7,$FD,$8C,$D7,$FD,$C7,$F8,$0F,$F1,$FB,$F8,$8D,$C7
	.dc.b	$3E,$B7,$EE,$E3,$F8,$9D,$79,$64,$CF,$C0,$7E,$6C,$62,$11,$87,$5F
	.dc.b	$09,$FD,$A5,$7B,$4F,$AE,$C5,$3E,$B7,$FC,$44,$DA,$FF,$13,$35,$FF
	.dc.b	$7C,$B5,$F0,$1D,$3E,$AF,$B5,$9A,$76,$AF,$E2,$DF,$80,$BB,$A9,$2F
	.dc.b	$F7,$B1,$FD,$F9,$B1,$FC,$77,$F1,$77,$87,$4E,$13,$F1,$4F,$CA,$7E
	.dc.b	$05,$FD,$89,$C7,$1A,$1F,$79,$FC,$9D,$B3,$FC,$31,$E1,$C7,$F9,$77
	.dc.b	$31,$CB,$EB,$FF,$4B,$C5,$F4,$B0,$46,$27,$C3,$87,$CF,$EA,$FF,$03
	.dc.b	$63,$FE,$AE,$5F,$6D,$93,$5F,$0B,$6F,$F9,$10,$9F,$F6,$13,$86,$BE
	.dc.b	$F3,$53,$FD,$07,$E0,$9D,$74,$93,$C7,$C6,$D7,$B7,$23,$F9,$5D,$7F
	.dc.b	$5C,$7E,$3B,$F9,$0C,$BF,$71,$F2,$DF,$C5,$A7,$E5,$B6,$33,$E4,$F3
	.dc.b	$9F,$F5,$DF,$93,$FA,$9E,$1F,$F6,$93,$61,$EC,$FF,$1C,$F7,$BF,$C6
	.dc.b	$33,$EF,$9F,$F7,$1B,$FD,$D3,$0F,$F6,$FB,$6F,$9E,$5C,$2F,$F3,$9F
	.dc.b	$80,$3E,$F6,$93,$1F,$97,$F8,$17,$F4,$47,$E4,$4F,$DC,$7F,$32,$D7
	.dc.b	$F5,$4D,$7F,$5E,$D5,$1F,$B7,$7F,$66,$D7,$3E,$1F,$93,$97,$FB,$A7
	.dc.b	$E0,$1F,$C9,$67,$F9,$47,$C9,$BF,$03,$FE,$F0,$D8,$FE,$BB,$3B,$EA
	.dc.b	$FF,$39,$56,$BC,$E9,$5F,$D8,$AF,$83,$FA,$1F,$91,$7E,$05,$F8,$6E
	.dc.b	$3F,$F6,$23,$78,$BF,$D8,$4D,$AF,$F3,$5B,$F9,$47,$FC,$D3,$8F,$F9
	.dc.b	$A3,$E4,$BA,$DF,$FB,$35,$B7,$CB,$A7,$F8,$67,$9E,$FA,$3A,$4C,$7F
	.dc.b	$D9,$D9,$7E,$FD,$FC,$EC,$FF,$9C,$9F,$E1,$27,$F4,$4F,$94,$FF,$59
	.dc.b	$46,$63,$F8,$6B,$5F,$4F,$C0,$7F,$9F,$2B,$FC,$2F,$1D,$A3,$C7,$5F
	.dc.b	$43,$8E,$4F,$CC,$7E,$9F,$6F,$D0,$F6,$24,$FC,$B7,$6F,$F0,$8E,$4F
	.dc.b	$F0,$6E,$FF,$D6,$5B,$B7,$F0,$16,$CF,$F6,$37,$FC,$B9,$1F,$C1,$78
	.dc.b	$BF,$54,$DF,$87,$FB,$E7,$E1,$3E,$03,$F0,$BF,$2D,$FE,$15,$BF,$F5
	.dc.b	$36,$1F,$CE,$F9,$D1,$88,$F0,$4F,$F8,$6F,$1F,$DC,$65,$FD,$B5,$B9
	.dc.b	$78,$74,$EB,$7F,$8C,$F9,$0F,$C7,$ED,$23,$07,$E2,$DC,$3F,$CA,$2F
	.dc.b	$F3,$D9,$F8,$AB,$F3,$7C,$7E,$72,$57,$DA,$9F,$3C,$79,$7E,$9F,$F3
	.dc.b	$A4,$9F,$CE,$C7,$F9,$5F,$D3,$7D,$0F,$F8,$D5,$B3,$CD,$AF,$F4,$EB
	.dc.b	$46,$23,$E3,$63,$F4,$1F,$D6,$FD,$2F,$EB,$0E,$D7,$DF,$E7,$A4,$FF
	.dc.b	$92,$AD,$1E,$3F,$A2,$CC,$FC,$1E,$3F,$DA,$98,$E1,$FF,$0D,$5F,$F8
	.dc.b	$22,$7E,$B3,$5F,$E8,$8D,$7F,$C4,$CD,$CF,$FD,$D9,$EE,$7F,$00,$24
	.dc.b	$F8,$FC,$4F,$89,$EA,$C8,$F3,$3F,$1F,$DD,$F9,$1E,$CF,$C7,$24,$FC
	.dc.b	$77,$F0,$9C,$FF,$C3,$09,$1E,$F8,$FF,$71,$FE,$04,$FA,$DA,$49,$FF
	.dc.b	$39,$78,$89,$FE,$97,$18,$FC,$0B,$FD,$0E,$FF,$83,$E5,$18,$8F,$F0
	.dc.b	$AC,$63,$E3,$7F,$7B,$1F,$EF,$89,$D3,$F1,$3E,$67,$F0,$5E,$34,$8F
	.dc.b	$A9,$15,$FC,$97,$CE,$FF,$26,$31,$C7,$FE,$52,$C7,$CD,$EC,$FF,$88
	.dc.b	$7E,$87,$FD,$3D,$1E,$9F,$4B,$1A,$79,$35,$FE,$A3,$97,$F8,$AA,$F7
	.dc.b	$1F,$E0,$C8,$C4,$FA,$B5,$FE,$31,$0A,$EB,$18,$6F,$B9,$DA,$58,$F8
	.dc.b	$5E,$67,$E0,$BC,$ED,$73,$FE,$69,$BD,$1F,$E0,$EC,$39,$D7,$01,$F5
	.dc.b	$FF,$E5,$68,$FF,$5B,$B7,$3F,$F2,$03,$67,$FC,$8B,$F6,$9F,$E3,$D7
	.dc.b	$94,$7F,$B7,$AB,$8F,$A9,$E7,$9E,$AD,$9E,$51,$E0,$EE,$FF,$19,$3F
	.dc.b	$E3,$E8,$FD,$3E,$9F,$71,$B1,$FC,$E9,$AE,$FD,$AF,$9E,$DE,$1E,$11
	.dc.b	$88,$F4,$B5,$FF,$2E,$CB,$BF,$FD,$1F,$97,$C0,$6E,$9F,$03,$C0,$B7
	.dc.b	$C2,$6B,$F2,$22,$DF,$F0,$EF,$23,$E6,$6D,$FE,$9D,$7E,$2C,$F2,$E1
	.dc.b	$FE,$1B,$EA,$DE,$1F,$F0,$65,$7D,$AD,$DF,$80,$FF,$AD,$23,$EE,$DF
	.dc.b	$07,$51,$9C,$FF,$11,$AF,$E9,$23,$1B,$7E,$2B,$CF,$F4,$E3,$D3,$D7
	.dc.b	$BC,$FF,$8A,$CB,$C0,$FF,$BD,$1B,$3A,$AF,$37,$C8,$7F,$AD,$3A,$FF
	.dc.b	$A1,$79,$76,$FF,$89,$CD,$D7,$FC,$23,$37,$6F,$E5,$7F,$43,$78,$3F
	.dc.b	$DC,$F8,$7F,$25,$8C,$61,$C7,$3F,$83,$F2,$BF,$EC,$2C,$BF,$81,$70
	.dc.b	$23,$11,$EF,$6B,$FA,$48,$D6,$1F,$E0,$0A,$7E,$03,$FE,$D6,$8F,$E2
	.dc.b	$2D,$AF,$F0,$E3,$E4,$6F,$C7,$7A,$7C,$D7,$AB,$4F,$57,$D7,$5C,$3F
	.dc.b	$9F,$3E,$2D,$FF,$A4,$B7,$73,$4C,$61,$F0,$CF,$8D,$F8,$0F,$D3,$CB
	.dc.b	$FD,$17,$5E,$5E,$2F,$EE,$1E,$8F,$8C,$DF,$87,$73,$FF,$5B,$CF,$FB
	.dc.b	$87,$0F,$F5,$04,$7F,$67,$6E,$1F,$D7,$5A,$FF,$80,$5A,$FF,$A3,$D7
	.dc.b	$DE,$7D,$9F,$D8,$BF,$35,$3F,$E9,$E9,$CD,$78,$0F,$F4,$B4,$EA,$FC
	.dc.b	$57,$EC,$3D,$BF,$75,$6B,$86,$1F,$EF,$DB,$FD,$43,$FE,$AA,$6D,$FA
	.dc.b	$FE,$A7,$C5,$DF,$F1,$9F,$9D,$91,$A4,$C3,$FC,$0D,$7F,$BD,$7B,$F0
	.dc.b	$23,$18,$FA,$FD,$EF,$77,$87,$F4,$9E,$A4,$BF,$94,$FD,$DA,$FF,$54
	.dc.b	$47,$46,$D7,$C5,$F8,$7F,$91,$E2,$FD,$C6,$D7,$8B,$34,$62,$BE,$07
	.dc.b	$8F,$F3,$CF,$C1,$FD,$2D,$BE,$0F,$C0,$23,$F7,$5F,$27,$93,$F5,$CF
	.dc.b	$C5,$F6,$D9,$F0,$38,$7C,$5F,$41,$70,$FA,$1E,$B7,$DF,$3F,$27,$F5
	.dc.b	$D6,$BD,$0E,$1F,$57,$95,$6B,$FD,$09,$F8,$17,$FA,$C2,$7F,$BE,$B6
	.dc.b	$3F,$D9,$5F,$33,$FE,$F7,$9C,$F9,$DF,$C1,$E5,$39,$7F,$CE,$ED,$7F
	.dc.b	$E0,$0B,$51,$1C,$37,$E6,$BA,$41,$FE,$19,$7C,$2F,$FD,$C3,$39,$7F
	.dc.b	$C2,$7F,$BF,$3D,$5F,$E2,$4D,$8F,$F0,$BF,$D5,$BF,$03,$CD,$7F,$17
	.dc.b	$8C,$4E,$A7,$1C,$59,$7F,$55,$EE,$7E,$AF,$3B,$5F,$89,$7F,$87,$6B
	.dc.b	$8D,$7F,$CC,$8D,$EE,$FE,$45,$FA,$87,$EB,$6F,$E0,$D2,$30,$FC,$4F
	.dc.b	$C1,$1F,$03,$F3,$67,$32,$9C,$D7,$F8,$A1,$B1,$FE,$A6,$6B,$D8,$E1
	.dc.b	$5F,$CB,$9E,$5F,$05,$F6,$64,$FD,$BF,$C1,$1F,$01,$98,$FE,$15,$F0
	.dc.b	$53,$1F,$F6,$67,$C9,$5C,$7F,$D0,$4F,$F2,$F3,$F9,$87,$CC,$59,$FF
	.dc.b	$6B,$C7,$F4,$BF,$89,$3F,$D3,$7E,$4F,$FA,$C5,$FF,$AA,$E3,$1E,$CC
	.dc.b	$BF,$E0,$35,$FC,$8B,$81,$1F,$8A,$F2,$13,$CA,$73,$82,$5E,$0F,$FB
	.dc.b	$AE,$FF,$85,$B6,$55,$FD,$5F,$F0,$47,$D5,$FF,$61,$AD,$4E,$7C,$0F
	.dc.b	$D0,$56,$B0,$FE,$47,$82,$13,$AC,$37,$E0,$70,$8B,$7C,$2B,$6F,$F1
	.dc.b	$EF,$97,$FC,$E6,$1E,$48,$F7,$7E,$66,$C3,$FB,$A5,$6B,$2F,$7E,$8A
	.dc.b	$D5,$F3,$EB,$DF,$5F,$06,$C7,$F2,$CF,$C1,$7F,$99,$3F,$CB,$BE,$6B
	.dc.b	$F9,$1E,$3D,$27,$3A,$78,$CF,$99,$3A,$AF,$A5,$3A,$AE,$1E,$7F,$F8
	.dc.b	$26,$1F,$33,$63,$F8,$47,$E1,$AF,$8F,$A7,$CC,$F8,$6B,$8F,$C5,$7F
	.dc.b	$31,$C4,$69,$23,$FD,$13,$1F,$85,$7F,$5A,$DD,$3F,$87,$FE,$0B,$FE
	.dc.b	$02,$9F,$A1,$EC,$7F,$87,$7C,$F7,$CA,$D3,$1F,$2B,$6B,$FA,$8F,$B3
	.dc.b	$FE,$37,$C7,$FD,$E0,$D7,$F2,$4D,$7F,$20,$DF,$A8,$7E,$0B,$F5,$EB
	.dc.b	$F3,$1F,$82,$BE,$27,$FC,$61,$5A,$E8,$FF,$18,$4E,$6B,$FB,$7C,$FE
	.dc.b	$D3,$D4,$FF,$1A,$7E,$0B,$E8,$9C,$F9,$29,$CE,$DF,$E3,$9F,$6B,$F9
	.dc.b	$46,$DF,$EE,$56,$BF,$C4,$BB,$91,$88,$FD,$63,$5F,$D5,$7E,$0B,$FA
	.dc.b	$2C,$7F,$60,$D8,$CF,$CF,$7F,$91,$E3,$17,$FE,$C9,$F8,$5F,$F2,$57
	.dc.b	$CB,$FF,$8A,$B0,$FE,$F6,$FF,$89,$5F,$DD,$39,$C3,$F1,$D4,$FA,$BF
	.dc.b	$F0,$6C,$79,$DA,$FF,$01,$6B,$FB,$F6,$EE,$FC,$17,$F2,$38,$7F,$82
	.dc.b	$BD,$2F,$05,$AB,$5E,$07,$F9,$7E,$73,$0F,$C1,$3F,$E0,$0D,$FF,$84
	.dc.b	$BF,$05,$FF,$0E,$BF,$D6,$A7,$5D,$7F,$E6,$79,$CE,$3F,$C4,$7F,$05
	.dc.b	$FA,$CF,$F3,$3D,$51,$88,$FA,$4D,$7F,$8E,$35,$FE,$35,$F0,$EF,$9F
	.dc.b	$FF,$3A,$37,$2F,$F1,$2A,$F9,$FC,$CF,$F9,$E7,$F0,$5F,$99,$E4,$FF
	.dc.b	$06,$7E,$0E,$47,$38,$7F,$2B,$73,$EE,$3C,$CF,$F3,$23,$67,$FA,$66
	.dc.b	$7F,$CB,$63,$0F,$B5,$AF,$E3,$7E,$63,$FC,$C9,$8F,$F1,$66,$D7,$F3
	.dc.b	$97,$F9,$C3,$7E,$4F,$FD,$20,$7F,$A9,$67,$F8,$EB,$63,$FA,$03,$FC
	.dc.b	$7A,$BF,$0E,$E5,$FE,$80,$7F,$40,$8F,$3E,$5F,$E1,$0F,$53,$FC,$FD
	.dc.b	$39,$3E,$A1,$4F,$EE,$B8,$BE,$3B,$FB,$2D,$FF,$98,$23,$1E,$F7,$FA
	.dc.b	$9D,$D7,$BB,$6B,$FB,$9E,$7F,$D4,$A5,$F8,$27,$F8,$AE,$30,$BE,$7B
	.dc.b	$93,$9C,$7F,$C4,$B5,$C2,$7B,$7E,$0A,$F8,$FF,$F5,$3C,$A3,$15,$F5
	.dc.b	$23,$13,$A9,$CE,$9F,$E0,$39,$7E,$0B,$FE,$AE,$8F,$E7,$91,$8F,$8D
	.dc.b	$F6,$4F,$CB,$8C,$61,$F0,$7F,$0C,$F6,$BE,$66,$DA,$7F,$B0,$F4,$7C
	.dc.b	$59,$CC,$62,$BC,$DE,$EF,$EE,$23,$17,$E5,$8F,$DC,$FB,$EF,$94,$F9
	.dc.b	$DF,$86,$7F,$BB,$AB,$FA,$97,$81,$FE,$00,$9C,$FC,$FF,$F2,$AB,$E1
	.dc.b	$97,$F2,$86,$DF,$BB,$F0,$7F,$C0,$72,$7F,$81,$B1,$E5,$7F,$77,$F0
	.dc.b	$7F,$C6,$4F,$D9,$6C,$7D,$47,$E2,$18,$BF,$80,$E3,$E0,$FF,$08,$D9
	.dc.b	$FF,$53,$F6,$3F,$07,$FE,$6D,$18,$FA,$7F,$C7,$3A,$8C,$E1,$FD,$BE
	.dc.b	$95,$A8,$C4,$EB,$1F,$F0,$FD,$75,$F4,$BF,$DE,$53,$EE,$C3,$FC,$DF
	.dc.b	$39,$7B,$77,$7F,$BC,$23,$C8,$D9,$FD,$16,$1C,$1A,$FF,$92,$FC,$0F
	.dc.b	$F3,$CF,$E0,$FF,$56,$3F,$4F,$86,$38,$BF,$82,$D5,$9C,$8A,$E0,$FF
	.dc.b	$98,$A3,$4F,$E8,$2A,$BF,$E8,$68,$BC,$BF,$E3,$CA,$E3,$39,$3F,$89
	.dc.b	$94,$B3,$87,$07,$F9,$C3,$E7,$BF,$55,$F7,$FC,$1B,$1F,$C7,$1A,$FF
	.dc.b	$9A,$5A,$F1,$F3,$7F,$C8,$EE,$DA,$ED,$AE,$F7,$ED,$1C,$30,$F8,$35
	.dc.b	$FE,$E1,$AF,$82,$D7,$E3,$67,$C5,$A7,$FA,$A2,$3F,$C5,$77,$C3,$E1
	.dc.b	$3F,$D1,$6F,$F9,$82,$53,$CE,$30,$F0,$8E,$63,$F3,$CF,$8C,$D3,$E4
	.dc.b	$38,$33,$0F,$9E,$EF,$B7,$CB,$6D,$79,$5A,$F3,$3F,$CC,$B3,$C5,$BE
	.dc.b	$FE,$A0,$D9,$E7,$BB,$E9,$1F,$2B,$D3,$F3,$DF,$F3,$AE,$3F,$E7,$06
	.dc.b	$BF,$C6,$A7,$6E,$7E,$3B,$C7,$BB,$D8,$A7,$38,$EF,$F6,$CF,$C2,$3B
	.dc.b	$78,$73,$0F,$E8,$13,$97,$C1,$1E,$EF,$93,$F9,$DC,$3F,$EB,$C6,$CF
	.dc.b	$FB,$7B,$4F,$F4,$1B,$FE,$86,$F0,$9F,$47,$F9,$56,$C7,$E6,$9A,$FF
	.dc.b	$6E,$6B,$FE,$BB,$5E,$EF,$CE,$6C,$75,$8C,$4E,$7F,$0B,$93,$E1,$7E
	.dc.b	$AF,$8A,$FA,$8E,$5E,$3F,$8C,$95,$E3,$7C,$DF,$59,$DF,$82,$4F,$AE
	.dc.b	$3C,$5E,$5C,$8F,$73,$F6,$BF,$05,$FF,$0C,$47,$FA,$9B,$F1,$47,$D9
	.dc.b	$7E,$D5,$B1,$FB,$A6,$BF,$E0,$E6,$BF,$E5,$7F,$0F,$E5,$BF,$E9,$33
	.dc.b	$FD,$7B,$CD,$F9,$0F,$1B,$F0,$7E,$0F,$FA,$92,$B6,$FE,$96,$1F,$5D
	.dc.b	$B2,$58,$FF,$28,$FE,$0B,$F9,$47,$FD,$7B,$3F,$99,$7F,$C4,$2D,$7F
	.dc.b	$DD,$CD,$7F,$DF,$AD,$7F,$C9,$C3,$EA,$3F,$4B,$F3,$4D,$E9,$FA,$AF
	.dc.b	$43,$C5,$3F,$AA,$F9,$BF,$93,$7E,$F6,$E7,$FB,$F6,$F6,$7F,$B8,$FF
	.dc.b	$04,$FF,$56,$57,$FB,$52,$7F,$7A,$47,$F9,$23,$FC,$97,$DE,$FF,$09
	.dc.b	$7D,$07,$EE,$3F,$33,$FF,$07,$57,$FB,$71,$AF,$F9,$D3,$DC,$4E,$FF
	.dc.b	$04,$93,$B7,$F8,$1D,$BE,$63,$8E,$9F,$C5,$7C,$B7,$67,$F3,$56,$E1
	.dc.b	$FC,$E9,$AF,$F9,$39,$AF,$FA,$5F,$F1,$97,$AB,$FC,$3A,$7F,$D3,$33
	.dc.b	$FE,$F2,$6B,$FE,$8A,$9D,$BF,$EF,$6F,$A0,$F8,$AF,$E2,$5B,$87,$F2
	.dc.b	$96,$F0,$FE,$0F,$37,$E0,$BF,$EC,$6B,$E5,$F4,$7F,$D3,$6C,$D2,$47
	.dc.b	$F9,$43,$E5,$1F,$71,$FC,$93,$EF,$FF,$8A,$CF,$E6,$5B,$9F,$FB,$EA
	.dc.b	$B7,$3F,$93,$8F,$E7,$92,$FB,$FE,$44,$9D,$69,$E7,$C7,$FC,$F1,$18
	.dc.b	$3F,$39,$F4,$BF,$C1,$9F,$F3,$CB,$5F,$D0,$1A,$F2,$FF,$B4,$46,$63
	.dc.b	$FC,$13,$E8,$7E,$8D,$FD,$84,$76,$6F,$C6,$7F,$42,$E3,$56,$F4,$FE
	.dc.b	$A9,$E7,$BE,$0B,$F9,$D1,$30,$FF,$6E,$FC,$87,$96,$BF,$AD,$70,$6E
	.dc.b	$7F,$BE,$B5,$FF,$48,$B5,$F4,$9F,$ED,$AA,$FF,$00,$47,$EF,$A3,$E1
	.dc.b	$1F,$ED,$EF,$17,$FA,$AC,$66,$3F,$85,$3D,$6B,$FB,$57,$DD,$7E,$B2
	.dc.b	$7F,$82,$3F,$EA,$D7,$9B,$5F,$E2,$4E,$D3,$98,$F5,$FB,$1E,$03,$FB
	.dc.b	$2C,$67,$87,$E8,$FC,$23,$E2,$3F,$B4,$AF,$B1,$A6,$22,$E3,$EC,$C9
	.dc.b	$F0,$5E,$C3,$31,$FC,$96,$DF,$5B,$D3,$C1,$93,$F1,$27,$2D,$8F,$8B
	.dc.b	$EF,$C8,$C1,$DF,$E9,$BE,$49,$1F,$14,$7F,$80,$3F,$0F,$FE,$0B,$1F
	.dc.b	$41,$AF,$F3,$76,$F0,$7F,$85,$D7,$CF,$FF,$82,$3F,$00,$FE,$C3,$3F
	.dc.b	$E1,$B6,$F0,$7E,$59,$B7,$FC,$DE,$F9,$CF,$E6,$BF,$00,$F9,$93,$FC
	.dc.b	$ED,$A9,$B2,$F4,$7C,$1E,$3C,$AF,$C3,$E1,$FC,$4B,$1F,$F1,$66,$1C
	.dc.b	$BE,$0F,$FA,$6E,$5F,$A8,$6D,$FF,$49,$6B,$DF,$E8,$7D,$9F,$63,$FE
	.dc.b	$53,$C4,$D2,$4F,$F9,$6A,$3E,$1B,$E2,$D7,$E2,$1C,$9F,$CD,$FE,$8B
	.dc.b	$FE,$3F,$CF,$CB,$F4,$7F,$E3,$0D,$0F,$F8,$CA,$75,$A7,$C5,$8F,$9D
	.dc.b	$F1,$F3,$0F,$D2,$DF,$07,$F4,$FF,$8F,$7F,$12,$7F,$B7,$37,$1F,$F2
	.dc.b	$73,$70,$FA,$1F,$81,$7D,$9F,$80,$97,$F2,$9F,$F2,$1D,$6E,$3F,$B7
	.dc.b	$7E,$55,$FE,$C7,$7F,$6A,$D7,$FB,$E3,$70,$FA,$3F,$17,$FE,$4C,$FC
	.dc.b	$0D,$F5,$7F,$E7,$D6,$DF,$FA,$09,$AF,$F0,$40,$E1,$FE,$0E,$FE,$D9
	.dc.b	$FE,$0A,$D7,$FC,$18,$DA,$FF,$36,$3D,$1F,$F3,$77,$E4,$DF,$C2,$5B
	.dc.b	$E2,$7F,$98,$23,$1A,$FE,$F3,$DC,$F4,$7C,$F4,$C7,$BB,$BB,$FB,$1E
	.dc.b	$D4,$9F,$A6,$FF,$6A,$FB,$CF,$C1,$C7,$93,$A7,$E1,$11,$D9,$FE,$63
	.dc.b	$AF,$F5,$07,$97,$FC,$76,$3F,$92,$B5,$FE,$7E,$D7,$FC,$41,$F8,$EF
	.dc.b	$F9,$95,$FE,$4E,$D7,$FC,$5C,$DD,$7F,$9D,$D7,$97,$AF,$FC,$E9,$E3
	.dc.b	$CB,$FE,$A1,$EC,$33,$FC,$FC,$7F,$9D,$E7,$FD,$0B,$EC,$7E,$98,$FD
	.dc.b	$3E,$7F,$C8,$67,$FC,$F9,$E5,$7A,$88,$FF,$AF,$A3,$F9,$23,$FE,$C0
	.dc.b	$9F,$ED,$3C,$3F,$D6,$9B,$1F,$E3,$B6,$A9,$E2,$FF,$92,$57,$DC,$FC
	.dc.b	$82,$47,$77,$FD,$95,$2F,$1F,$D2,$49,$FF,$49,$4F,$D7,$F9,$AF,$F4
	.dc.b	$B7,$9D,$27,$E1,$E9,$FE,$FA,$7F,$D3,$11,$F0,$BC,$89,$18,$3F,$D3
	.dc.b	$9E,$5F,$DA,$23,$F0,$5F,$F6,$E7,$E0,$7F,$E7,$98,$FE,$BA,$D7,$FC
	.dc.b	$C8,$DE,$3F,$EE,$33,$FE,$A9,$E8,$95,$FE,$A9,$7F,$BC,$7E,$4C,$F0
	.dc.b	$7F,$D7,$38,$3E,$43,$B9,$2F,$FD,$DF,$19,$6F,$95,$C0,$8F,$92,$7F
	.dc.b	$D8,$7F,$A4,$7E,$92,$3F,$CD,$CD,$4D,$F4,$FF,$CC,$91,$EC,$F8,$FF
	.dc.b	$79,$1F,$0B,$F0,$B8,$7F,$C3,$3F,$38,$F4,$7F,$C3,$79,$FE,$42,$D3
	.dc.b	$97,$FC,$FD,$19,$8C,$42,$3E,$B3,$C2,$79,$F6,$76,$7A,$F9,$B6,$79
	.dc.b	$B5,$E5,$37,$A7,$F1,$11,$FE,$8D,$C3,$BE,$1F,$D9,$1B,$F0,$9B,$E9
	.dc.b	$E5,$2F,$B2,$A4,$BE,$59,$FB,$AF,$7E,$DC,$9C,$9E,$E7,$83,$EB,$3E
	.dc.b	$0F,$9D,$D8,$8F,$03,$FC,$6D,$AF,$8B,$D3,$91,$9C,$BF,$B4,$36,$BC
	.dc.b	$FF,$57,$FE,$6E,$8F,$F1,$9C,$A3,$11,$F5,$1A,$FD,$26,$BF,$47,$D8
	.dc.b	$FE,$84,$FD,$A6,$BF,$61,$AE,$FE,$6B,$F2,$7F,$80,$BC,$3F,$33,$67
	.dc.b	$E0,$1F,$77,$D2,$C7,$17,$D2,$9B,$C1,$F8,$89,$B5,$F8,$66,$BF,$E5
	.dc.b	$9B,$B9,$A6,$C4,$3E,$57,$CD,$BE,$9F,$17,$E2,$8C,$4E,$A7,$AC,$3F
	.dc.b	$02,$B1,$E0,$EA,$FF,$03,$D7,$49,$C7,$47,$07,$8F,$BD,$CB,$D3,$B5
	.dc.b	$E2,$7A,$FC,$3E,$33,$F1,$37,$83,$F4,$3A,$3E,$E7,$E4,$57,$F9,$B2
	.dc.b	$31,$87,$F6,$B8,$46,$1F,$ED,$CD,$7F,$CE,$7D,$FF,$CB,$27,$FC,$E2
	.dc.b	$D8,$FF,$7C,$7E,$51,$F0,$9F,$F7,$D4,$67,$C7,$F0,$DB,$D4,$FB,$51
	.dc.b	$86,$E1,$FE,$98,$97,$D7,$63,$0F,$F7,$D4,$71,$9F,$F4,$0B,$A7,$D4
	.dc.b	$D8,$F4,$CF,$C0,$7A,$35,$0C,$7E,$06,$1F,$19,$B7,$AF,$85,$77,$3D
	.dc.b	$6F,$26,$1C,$75,$B9,$BE,$66,$9A,$CB,$E6,$E5,$E8,$8C,$4F,$8E,$FC
	.dc.b	$4D,$AD,$C7,$F8,$7D,$AF,$FA,$93,$F4,$27,$D0,$F8,$3E,$2F,$C7,$18
	.dc.b	$C7,$E2,$C7,$C9,$F5,$FF,$BC,$79,$5A,$BC,$7A,$68,$C5,$FC,$CD,$BD
	.dc.b	$67,$4F,$AB,$B1,$FE,$B0,$C9,$1E,$96,$C7,$E2,$DB,$9F,$F8,$DD,$AF
	.dc.b	$FA,$F3,$F6,$4F,$F5,$24,$FF,$AE,$9B,$0E,$3F,$CB,$D3,$8F,$E7,$D6
	.dc.b	$A7,$31,$E1,$25,$F8,$5F,$99,$F9,$37,$F5,$F8,$ED,$8F,$D8,$8C,$63
	.dc.b	$DE,$7F,$DC,$F8,$7A,$30,$FE,$1D,$B7,$F8,$7F,$E6,$78,$7A,$1F,$21
	.dc.b	$B7,$E9,$F9,$4F,$29,$D7,$9E,$C7,$FB,$5B,$3F,$C4,$BF,$AF,$6E,$7F
	.dc.b	$E5,$B6,$BF,$EE,$06,$BF,$EF,$AF,$C1,$1F,$47,$F9,$06,$D7,$FB,$8B
	.dc.b	$E2,$FE,$23,$B9,$FC,$AC,$62,$BF,$D3,$5E,$FF,$E0,$3D,$6C,$BC,$D9
	.dc.b	$7F,$EE,$CF,$BF,$7C,$1E,$2F,$E3,$1F,$A8,$8C,$CE,$E7,$F4,$7F,$12
	.dc.b	$FB,$DF,$D2,$D6,$AF,$88,$D3,$F9,$0B,$86,$F1,$FF,$16,$7F,$D9,$38
	.dc.b	$FF,$BF,$9C,$BF,$0F,$D5,$FD,$44,$6A,$BF,$D0,$8D,$7F,$7A,$7E,$F9
	.dc.b	$C3,$8F,$C0,$9F,$2B,$FB,$A6,$DF,$F0,$4C,$3F,$D0,$7D,$D9,$AF,$A1
	.dc.b	$EB,$4A,$F8,$DE,$ED,$C7,$DB,$A7,$F8,$BE,$72,$73,$3C,$7E,$57,$F4
	.dc.b	$37,$33,$F1,$3F,$03,$FB,$6F,$C3,$F4,$BF,$86,$C6,$34,$F8,$91,$E7
	.dc.b	$EA,$CF,$3D,$C3,$B6,$1F,$C7,$30,$E1,$1F,$0B,$EE,$7F,$D5,$AE,$A5
	.dc.b	$18,$7E,$AA,$F3,$FF,$8E,$B9,$CB,$FA,$4B,$9A,$FE,$90,$DA,$FF,$5A
	.dc.b	$35,$FA,$27,$6A,$FE,$3D,$C3,$F6,$A3,$F9,$8B,$52,$F7,$FE,$A7,$A5
	.dc.b	$E2,$F3,$A6,$3E,$9A,$FC,$06,$5E,$49,$FB,$5C,$D2,$B5,$EE,$5C,$BD
	.dc.b	$BD,$5E,$C3,$F0,$1E,$ED,$AE,$8E,$1E,$DE,$05,$AF,$8B,$F8,$1F,$C4
	.dc.b	$3E,$2F,$99,$6B,$C9,$EF,$61,$A7,$CF,$E2,$4F,$E3,$FA,$BF,$21,$56
	.dc.b	$B7,$87,$E9,$B8,$3F,$C5,$D1,$C6,$3F,$2D,$EF,$7D,$47,$F3,$1E,$F6
	.dc.b	$27,$E2,$36,$39,$7B,$BF,$D5,$BF,$02,$FE,$05,$1F,$3A,$31,$E2,$FE
	.dc.b	$87,$B3,$F4,$51,$70,$F9,$CF,$1F,$84,$91,$F8,$9C,$1F,$09,$8C,$7F
	.dc.b	$7B,$18,$9F,$F0,$63,$63,$FD,$F0,$D7,$F0,$AD,$7F,$69,$1F,$B3,$3F
	.dc.b	$C0,$1F,$45,$FD,$BF,$E1,$B1,$1F,$9E,$F4,$5F,$94,$7E,$0A,$3F,$7F
	.dc.b	$E8,$7F,$91,$DD,$46,$70,$FE,$FF,$39,$7F,$89,$FC,$27,$F1,$7A,$FF
	.dc.b	$11,$36,$3F,$32,$1C,$3F,$9C,$6B,$FC,$3B,$F1,$4F,$EE,$0F,$F1,$E6
	.dc.b	$F7,$FF,$4A,$FC,$0B,$F4,$CF,$F8,$CA,$33,$7E,$4F,$C0,$9F,$2D,$FE
	.dc.b	$0C,$9F,$E7,$19,$27,$35,$FB,$99,$FB,$63,$5F,$F0,$53,$63,$FC,$78
	.dc.b	$DC,$BD,$6D,$7F,$6B,$F0,$5F,$E0,$19,$FE,$08,$D8,$FE,$7F,$83,$F0
	.dc.b	$27,$CA,$FF,$88,$2B,$5A,$7F,$66,$F8,$07,$D7,$67,$AF,$FC,$21,$18
	.dc.b	$CF,$AC,$7F,$8A,$3F,$66,$FF,$1D,$69,$F3,$9A,$FF,$08,$6B,$FC,$E9
	.dc.b	$AF,$F8,$E1,$FF,$19,$4F,$F1,$B6,$C7,$A3,$C1,$FE,$76,$FF,$50,$8C
	.dc.b	$E3,$FD,$F3,$F0,$0F,$C1,$3F,$E6,$4B,$7D,$77,$F8,$9E,$73,$1F,$D2
	.dc.b	$8C,$9F,$CA,$47,$D0,$7F,$8E,$5A,$7F,$9A,$CB,$87,$E2,$9A,$FF,$24
	.dc.b	$FA,$D7,$B9,$FE,$4A,$6C,$FF,$3A,$E2,$FF,$49,$7C,$3E,$6B,$F4,$79
	.dc.b	$BC,$5F,$F2,$8F,$E5,$6F,$BF,$CC,$61,$CF,$17,$F9,$36,$7F,$C9,$3A
	.dc.b	$7F,$64,$6B,$FE,$62,$6A,$EF,$78,$5F,$E5,$E6,$EB,$E5,$E2,$FE,$9B
	.dc.b	$5F,$E6,$A9,$CB,$F8,$8F,$03,$F8,$7C,$7F,$8A,$EB,$F6,$31,$8F,$84
	.dc.b	$FF,$34,$BF,$85,$AF,$D7,$B5,$FE,$92,$D7,$FB,$DB,$5F,$F3,$A7,$93
	.dc.b	$C8,$FF,$43,$78,$3F,$EB,$5F,$8B,$C6,$1F,$D7,$FF,$1F,$F8,$CD,$F8
	.dc.b	$CF,$F3,$CC,$63,$0F,$F0,$AB,$FE,$80,$7F,$D0,$AF,$E4,$13,$49,$EA
	.dc.b	$7F,$9E,$5F,$F4,$87,$1F,$FA,$B9,$B7,$FE,$12,$F8,$A7,$C9,$F8,$32
	.dc.b	$F0,$FC,$1C,$AF,$37,$AB,$73,$F1,$BD,$49,$1F,$EB,$7C,$7F,$C5,$93
	.dc.b	$FD,$BB,$E5,$FF,$12,$FA,$B6,$FF,$4A,$C6,$34,$F2,$47,$C6,$FB,$4F
	.dc.b	$E9,$B5,$FE,$34,$AD,$61,$FE,$A4,$A5,$6A,$FF,$D9,$35,$AB,$FF,$54
	.dc.b	$BF,$DD,$EF,$FC,$19,$2F,$89,$F0,$FB,$9F,$EA,$68,$C4,$FF,$85,$B1
	.dc.b	$7D,$2F,$F5,$4C,$39,$63,$FE,$4E,$9C,$FA,$9C,$AF,$FC,$31,$3F,$EA
	.dc.b	$4C,$3F,$D7,$9E,$BF,$FA,$CA,$31,$3F,$E1,$08,$C7,$64,$8F,$F1,$06
	.dc.b	$3F,$E7,$0B,$FF,$62,$4F,$C6,$F6,$7F,$D6,$53,$FE,$70,$C3,$DB,$18
	.dc.b	$3F,$D6,$CF,$FB,$23,$CA,$FF,$64,$FD,$D7,$F2,$48,$FF,$27,$E9,$E3
	.dc.b	$53,$AF,$33,$FA,$46,$7E,$A9,$C8,$BE,$4F,$D1,$8E,$0F,$F3,$63,$53
	.dc.b	$C9,$A6,$4F,$F7,$D5,$7F,$BE,$DF,$31,$5F,$F7,$F7,$DC,$7F,$AE,$3E
	.dc.b	$D2,$F7,$1A,$4A,$FE,$C3,$3F,$C7,$B6,$E3,$E4,$FF,$AD,$67,$2D,$7F
	.dc.b	$C7,$F3,$B7,$8F,$DD,$78,$E0,$D2,$47,$93,$E6,$F4,$47,$07,$A3,$9B
	.dc.b	$91,$1C,$1E,$CF,$4F,$B1,$1C,$1F,$03,$B7,$C4,$47,$07,$F7,$E3,$FA
	.dc.b	$EF,$27,$C9,$DC,$D2,$46,$E1,$F2,$7B,$24,$65,$FA,$5F,$73,$F0,$23
	.dc.b	$B3,$5F,$F0,$6B,$5F,$F3,$3B,$5F,$F4,$CA,$CB,$99,$23,$9B,$F7,$FE
	.dc.b	$E7,$FD,$37,$1E,$B6,$BF,$E1,$A6,$BF,$E7,$46,$BF,$EA,$67,$FB,$CB
	.dc.b	$C7,$D4,$E8,$47,$A7,$C8,$F7,$23,$83,$C1,$F1,$5E,$32,$38,$3C,$9D
	.dc.b	$DE,$72,$38,$3D,$1C,$FE,$B2,$38,$3E,$B7,$DC,$7F,$B3,$5B,$E1,$FE
	.dc.b	$7E,$31,$1F,$E3,$86,$BF,$E9,$2F,$3D,$F2,$1F,$A5,$6C,$7E,$05,$AF
	.dc.b	$F9,$01,$AF,$1F,$07,$81,$1E,$0F,$3E,$49,$1C,$DA,$FF,$92,$BE,$E1
	.dc.b	$E4,$7E,$D5,$B1,$F9,$16,$BB,$8F,$DD,$7D,$BB,$D4,$A4,$9F,$CC,$46
	.dc.b	$5F,$CD,$7D,$BF,$F8,$69,$BE,$CF,$F0,$38,$C4,$7F,$97,$1A,$FF,$AD
	.dc.b	$9F,$F0,$DB,$CB,$BB,$81,$1E,$BD,$9C,$C8,$E0,$FD,$52,$E3,$FE,$24
	.dc.b	$D8,$CC,$60,$CF,$C2,$48,$E2,$D7,$FC,$E3,$F6,$BF,$E1,$D7,$F6,$CD
	.dc.b	$73,$EB,$FF,$1F,$8C,$47,$FB,$41,$FD,$CB,$C3,$BB,$44,$7F,$80,$7D
	.dc.b	$AF,$F8,$79,$BE,$BF,$F2,$98,$C4,$7F,$A0,$9A,$FF,$B7,$17,$8D,$B2
	.dc.b	$47,$47,$F9,$97,$DA,$7E,$B2,$3F,$D1,$6D,$7F,$DC,$AD,$44,$70,$DF
	.dc.b	$57,$99,$1F,$E3,$10,$8F,$E8,$1F,$12,$FA,$2A,$4A,$FE,$3D,$19,$6F
	.dc.b	$D5,$70,$23,$83,$7D,$3D,$08,$E0,$E7,$E8,$78,$08,$E0,$EF,$A7,$90
	.dc.b	$8E,$0F,$1F,$1F,$A0,$8E,$0F,$3F,$93,$EC,$23,$83,$D7,$D7,$F0,$08
	.dc.b	$E0,$F7,$F9,$3F,$10,$8E,$0F,$87,$F6,$FF,$20,$8E,$0F,$E8,$9F,$64
	.dc.b	$F9,$6F,$ED,$8D,$8C,$C6,$0C,$FC,$D4,$8E,$2D,$7F,$D9,$72,$F1,$F9
	.dc.b	$92,$3B,$BD,$BA,$A4,$78,$9A,$F2,$F9,$5F,$19,$1F,$3B,$E6,$F9,$59
	.dc.b	$1E,$A6,$BE,$2F,$95,$F5,$91,$F7,$BE,$EE,$69,$1F,$09,$AF,$AB,$C5
	.dc.b	$F8,$64,$7E,$3B,$F1,$BE,$4A,$47,$CA,$6B,$F0,$BC,$AF,$96,$47,$E7
	.dc.b	$BF,$E1,$9F,$AF,$7D,$2D,$24,$FF,$77,$8C,$B7,$E6,$38,$11,$C1,$BF
	.dc.b	$9A,$E8,$47,$07,$3F,$83,$F0,$11,$C1,$DF,$E0,$FC,$84,$70,$78,$FE
	.dc.b	$0F,$D0,$47,$07,$9F,$E4,$BD,$84,$70,$7F,$84,$7E,$BB,$F8,$BB,$FC
	.dc.b	$08,$8E,$1C,$07,$0E,$BC,$6F,$F2,$04,$7F,$85,$23,$A7,$B3,$82,$3F
	.dc.b	$E1,$FF,$AD,$FE,$82,$F7,$6B,$FC,$49,$A9,$A7,$F6,$47,$87,$A7,$FE
	.dc.b	$4F,$8C,$47,$18,$C7,$85,$23,$C0,$D7,$D6,$75,$F0,$B9,$11,$F3,$3D
	.dc.b	$7D,$99,$1E,$96,$BC,$AF,$FD,$20,$FB,$7C,$EF,$21,$1F,$81,$E1,$79
	.dc.b	$C8,$E0,$F8,$9E,$17,$AC,$8E,$0F,$F2,$37,$D5,$7F,$34,$4D,$27,$B3
	.dc.b	$FE,$71,$7F,$DF,$B1,$C3,$E3,$70,$47,$9F,$D1,$EC,$8E,$0E,$FE,$17
	.dc.b	$32,$38,$3C,$7E,$17,$72,$38,$3C,$FE,$BF,$8C,$8E,$0F,$5F,$B7,$E7
	.dc.b	$23,$83,$DF,$E2,$7E,$4F,$BF,$F2,$39,$A3,$13,$F9,$78,$C3,$FD,$6A
	.dc.b	$77,$3A,$8F,$EB,$B2,$F1,$31,$1F,$29,$B1,$98,$E4,$75,$8C,$F8,$92
	.dc.b	$32,$74,$70,$AF,$CF,$F3,$BD,$09,$1C,$0F,$B2,$F8,$3E,$1A,$46,$70
	.dc.b	$46,$23,$A1,$EF,$F1,$BC,$E4,$78,$3E,$AF,$1A,$47,$36,$BC,$8F,$87
	.dc.b	$F2,$8F,$8D,$FC,$6B,$63,$F4,$B8,$23,$11,$F4,$9A,$FD,$1F,$45,$F1
	.dc.b	$1F,$96,$6C,$72,$6B,$F6,$9A,$FD,$9F,$67,$DB,$AF,$DE,$6C,$66,$3E
	.dc.b	$24,$62,$3E,$DF,$97,$5F,$D1,$B7,$13,$4C,$54,$3E,$71,$67,$FD,$E7
	.dc.b	$F5,$5C,$FC,$43,$49,$3F,$66,$1F,$55,$B2,$46,$5E,$5E,$6C,$8E,$0D
	.dc.b	$BF,$F8,$03,$E6,$77,$24,$F5,$7C,$5E,$67,$D7,$9F,$E6,$98,$23,$11
	.dc.b	$FE,$38,$8C,$57,$FB,$2D,$B1,$FC,$79,$7C,$7F,$08,$FF,$66,$B6,$33
	.dc.b	$3C,$63,$1E,$E4,$8E,$8D,$AE,$9E,$77,$12,$4F,$85,$E9,$E7,$48,$F2
	.dc.b	$36,$7C,$9F,$81,$78,$49,$3E,$97,$C9,$F0,$32,$3D,$8D,$9F,$67,$D0
	.dc.b	$7A,$49,$3F,$05,$F6,$7C,$74,$8F,$88,$D9,$F8,$9A,$7C,$12,$4F,$C9
	.dc.b	$7E,$27,$9F,$23,$E6,$36,$7F,$A4,$FD,$3B,$DC,$F9,$8D,$AF,$F7,$E3
	.dc.b	$67,$73,$AF,$72,$56,$9B,$3D,$3E,$DF,$99,$27,$9B,$DB,$E0,$24,$76
	.dc.b	$6C,$F9,$3D,$1F,$19,$27,$C6,$F9,$BD,$09,$1E,$66,$CF,$F6,$4E,$CF
	.dc.b	$E0,$93,$FD,$9A,$11,$87,$5E,$07,$E2,$E3,$11,$FE,$06,$FA,$6F,$F2
	.dc.b	$C5,$71,$6C,$7F,$70,$6B,$BF,$73,$F2,$90,$F3,$A9,$2B,$93,$AF,$3A
	.dc.b	$47,$56,$DF,$5F,$33,$99,$27,$C4,$F6,$F3,$24,$79,$5B,$3E,$5F,$33
	.dc.b	$C6,$49,$F5,$3E,$6F,$2A,$47,$B5,$B3,$ED,$F2,$BD,$64,$9F,$84,$FB
	.dc.b	$B5,$91,$F1,$5B,$3F,$17,$C5,$F8,$64,$9F,$94,$FE,$E7,$E6,$FF,$88
	.dc.b	$C7,$D5,$6C,$FF,$87,$1A,$FF,$8E,$1A,$FF,$8F,$5F,$F2,$EC,$71,$F1
	.dc.b	$3A,$11,$E9,$E2,$77,$23,$83,$C1,$E1,$78,$C8,$E0,$F2,$78,$5E,$72
	.dc.b	$38,$3D,$1F,$15,$EB,$23,$83,$D9,$ED,$FB,$C8,$E0,$F8,$1F,$1B,$F0
	.dc.b	$C8,$E0,$F8,$9E,$17,$C7,$23,$83,$E4,$78,$5F,$2C,$8E,$0F,$99,$E6
	.dc.b	$7C,$F2,$38,$3E,$86,$DF,$4C,$8E,$0F,$D4,$7D,$0F,$F0,$18,$FF,$34
	.dc.b	$B5,$FF,$37,$35,$FF,$78,$B5,$2F,$03,$F9,$C3,$67,$AC,$60,$CF,$B5
	.dc.b	$23,$C2,$DA,$FF,$2F,$D3,$DA,$E6,$4A,$F3,$BD,$BE,$8A,$47,$A9,$B5
	.dc.b	$FE,$68,$F9,$BF,$ED,$B3,$FC,$15,$B1,$9E,$5F,$E9,$D1,$88,$FF,$44
	.dc.b	$E6,$F2,$7F,$A9,$36,$FF,$87,$36,$7F,$CE,$AD,$7F,$D2,$3F,$3E,$F8
	.dc.b	$CF,$EB,$4D,$AF,$E2,$CD,$9F,$F3,$EB,$5E,$3E,$47,$02,$4F,$07,$9F
	.dc.b	$8B,$23,$9B,$67,$FD,$0F,$F2,$AF,$6B,$FB,$53,$6B,$F9,$33,$67,$73
	.dc.b	$AF,$5E,$56,$9B,$3F,$E9,$2F,$9B,$75,$FF,$75,$6D,$7F,$31,$6C,$EE
	.dc.b	$75,$E7,$CA,$D3,$67,$A1,$CC,$91,$CD,$FF,$01,$2F,$32,$91,$FE,$80
	.dc.b	$64,$FE,$85,$F3,$AF,$7B,$FC,$14,$D8,$CC,$60,$FF,$07,$7C,$EF,$F8
	.dc.b	$32,$FF,$D5,$ED,$7F,$D9,$6D,$7F,$DA,$0D,$41,$75,$F4,$25,$7F,$55
	.dc.b	$8C,$BF,$D5,$FE,$75,$D7,$FE,$1F,$6C,$66,$30,$67,$D1,$91,$C5,$B8
	.dc.b	$7F,$B1,$BD,$07,$C4,$FF,$8A,$5B,$5F,$D9,$5B,$3B,$9D,$79,$52,$B4
	.dc.b	$D9,$FF,$68,$FC,$EF,$F8,$D9,$FF,$76,$B5,$FF,$79,$35,$F8,$6D,$4F
	.dc.b	$E1,$1E,$3C,$3F,$E3,$D8,$C4,$72,$8C,$78,$D2,$FF,$BE,$52,$31,$5E
	.dc.b	$4E,$CE,$44,$AD,$BD,$7C,$69,$1C,$9B,$3E,$CF,$57,$C4,$49,$EE,$F9
	.dc.b	$7B,$A4,$78,$9B,$3F,$EF,$4F,$99,$7A,$BF,$E5,$76,$D7,$F8,$4D,$B3
	.dc.b	$B9,$D7,$6C,$AD,$36,$7F,$C3,$5E,$3F,$F9,$85,$34,$93,$FE,$1F,$8F
	.dc.b	$C5,$9F,$E2,$1E,$6E,$04,$8D,$75,$73,$23,$83,$96,$7D,$C8,$E0,$ED
	.dc.b	$9F,$8C,$8E,$0F,$15,$BB,$A9,$1F,$33,$FE,$35,$F9,$9F,$DD,$7C,$D5
	.dc.b	$ED,$34,$95,$FA,$C2,$4B,$E4,$1C,$1C,$DB,$3B,$3F,$67,$3B,$9D,$1E
	.dc.b	$1E,$0E,$8F,$FB,$36,$3E,$CC,$FE,$EA,$31,$3A,$CB,$C3,$F2,$16,$BF
	.dc.b	$7F,$D7,$FC,$B5,$E1,$3E,$78,$C4,$79,$A3,$1E,$17,$18,$63,$E0,$74
	.dc.b	$7A,$3E,$27,$EE,$41,$A4,$8F,$95,$F1,$3C,$E4,$D3,$13,$DF,$07,$AD
	.dc.b	$49,$7E,$83,$E3,$D7,$97,$CF,$71,$D2,$4E,$C4,$39,$3B,$78,$72,$75
	.dc.b	$8F,$F3,$17,$85,$74,$79,$7C,$D7,$C1,$E1,$FF,$7F,$37,$7F,$C0,$36
	.dc.b	$7F,$96,$AE,$0F,$95,$F9,$E3,$D4,$A4,$AE,$BC,$7F,$3F,$E6,$FC,$51
	.dc.b	$F8,$50,$E1,$FC,$43,$5F,$C8,$35,$FE,$7A,$BE,$33,$A1,$23,$E2,$47
	.dc.b	$6C,$D9,$F2,$E5,$F1,$12,$B7,$EE,$79,$7E,$33,$F7,$58,$1A,$49,$FF
	.dc.b	$4E,$47,$DA,$7B,$F2,$FB,$4F,$77,$BB,$F8,$DF,$9B,$7C,$45,$25,$79
	.dc.b	$B3,$D7,$B5,$FC,$26,$8D,$24,$FF,$AA,$FD,$17,$CA,$FE,$65,$B1,$F8
	.dc.b	$36,$BF,$9A,$0B,$E4,$38,$7D,$8C,$9E,$0F,$3F,$77,$A7,$77,$F1,$BB
	.dc.b	$FC,$DD,$78,$63,$07,$F1,$FF,$9C,$6D,$FA,$B6,$CF,$E5,$5A,$F8,$DF
	.dc.b	$98,$FE,$67,$E7,$1E,$5F,$EC,$DB,$19,$8C,$79,$9F,$B7,$F9,$CF,$54
	.dc.b	$FF,$B3,$5A,$FE,$2D,$AF,$F1,$16,$BF,$C6,$5F,$CC,$3F,$A9,$6B,$F1
	.dc.b	$DB,$D1,$F8,$88,$FF,$05,$C7,$F1,$F7,$87,$89,$E3,$23,$CF,$CD,$FE
	.dc.b	$18,$7B,$9F,$CF,$35,$FE,$48,$D7,$F9,$63,$5F,$F0,$CA,$F7,$79,$92
	.dc.b	$3F,$80,$47,$FB,$B1,$FE,$05,$F3,$9F,$DC,$DB,$E8,$7F,$1D,$8C,$47
	.dc.b	$E0,$DF,$F2,$3B,$C3,$C8,$FE,$B1,$1F,$C0,$E3,$F6,$8D,$7F,$9D,$35
	.dc.b	$FE,$85,$F3,$8F,$80,$FE,$54,$D8,$FE,$28,$D7,$F2,$4F,$FA,$25,$CF
	.dc.b	$99,$C0,$8F,$0F,$3B,$F9,$A7,$CE,$7E,$F6,$3F,$83,$07,$0F,$F0,$96
	.dc.b	$BF,$D3,$DA,$FF,$57,$7F,$42,$FF,$27,$6B,$FA,$56,$EF,$F7,$E3,$B7
	.dc.b	$C3,$FF,$45,$F9,$37,$C0,$52,$4F,$F3,$18,$FD,$7D,$71,$7F,$D6,$11
	.dc.b	$E8,$6B,$FC,$7D,$AF,$F6,$56,$BF,$DB,$7E,$A9,$F1,$5F,$D5,$5B,$1F
	.dc.b	$CF,$DA,$FE,$F6,$7A,$D7,$87,$E1,$BD,$04,$9F,$37,$A3,$FD,$7F,$E2
	.dc.b	$3F,$D5,$51,$FD,$25,$AF,$F0,$76,$D7,$38,$F7,$7A,$1F,$0C,$8F,$C1
	.dc.b	$DF,$FB,$57,$CF,$BE,$C6,$92,$78,$9F,$1F,$C0,$F9,$84,$7E,$57,$57
	.dc.b	$CF,$F9,$F9,$1F,$35,$D3,$FC,$6E,$DD,$5E,$02,$46,$CF,$EC,$9F,$3F
	.dc.b	$FE,$70,$6F,$C1,$7F,$7F,$8C,$47,$F2,$55,$F2,$3F,$E0,$3F,$B6,$7C
	.dc.b	$DD,$25,$72,$3A,$FA,$9C,$C8,$F8,$3D,$4F,$F0,$67,$D6,$3E,$5A,$92
	.dc.b	$78,$9E,$6E,$AF,$21,$1F,$4F,$A9,$E7,$F9,$DF,$F2,$66,$A6,$92,$7F
	.dc.b	$4B,$F9,$17,$E5,$9F,$F0,$23,$5F,$E8,$4D,$E8,$FC,$F5,$73,$9F,$8F
	.dc.b	$E7,$7C,$52,$3F,$2B,$C0,$FF,$13,$7D,$B7,$F5,$A8,$FF,$08,$35,$FE
	.dc.b	$9C,$DB,$EB,$5F,$3F,$D6,$95,$F5,$1F,$A5,$E5,$7C,$FF,$9F,$93,$FE
	.dc.b	$1B,$E4,$47,$08,$FE,$B5,$F4,$6F,$03,$FC,$70,$D8,$FF,$11,$35,$DF
	.dc.b	$BD,$AF,$9E,$FF,$3A,$49,$A4,$9F,$E1,$BC,$BF,$E7,$C7,$FC,$58,$D7
	.dc.b	$FB,$53,$75,$FE,$0D,$5C,$E7,$C3,$F3,$BE,$42,$3E,$4F,$3F,$F4,$BE
	.dc.b	$2C,$8F,$F1,$C7,$7F,$FB,$39,$E2,$7A,$BE,$8F,$DA,$47,$DB,$ED,$FE
	.dc.b	$E7,$D4,$91,$FB,$97,$4F,$F7,$FF,$A3,$FE,$6C,$FF,$B8,$9A,$9A,$63
	.dc.b	$EB,$FC,$51,$FE,$07,$6E,$BF,$C8,$DB,$CB,$FE,$81,$AF,$E8,$D3,$FC
	.dc.b	$9F,$E9,$3F,$B8,$B7,$DD,$FF,$36,$46,$2B,$FD,$10,$BE,$6B,$FC,$E1
	.dc.b	$87,$F8,$33,$1F,$07,$95,$83,$8E,$1C,$BC,$EF,$F5,$66,$1F,$D2,$6B
	.dc.b	$FC,$90,$D7,$FD,$1C,$D7,$FD,$2D,$7B,$CB,$D5,$5E,$03,$A4,$FF,$9A
	.dc.b	$6B,$F8,$5B,$FD,$39,$AF,$6F,$27,$C2,$7F,$AF,$7D,$FF,$EE,$11,$FE
	.dc.b	$2D,$6D,$FF,$49,$6E,$9F,$02,$BE,$1E,$72,$FC,$0F,$C6,$F3,$3F,$D2
	.dc.b	$5F,$63,$FE,$7A,$9F,$F3,$E3,$5F,$F1,$D3,$71,$FA,$15,$FE,$99,$FA
	.dc.b	$5F,$EB,$F4,$0E,$1B,$F0,$DF,$ED,$58,$C6,$7F,$E0,$D9,$FF,$45,$35
	.dc.b	$F6,$78,$1D,$6B,$B7,$95,$27,$FD,$21,$68,$F1,$7D,$0F,$8F,$CC,$EC
	.dc.b	$4A,$FE,$F3,$5F,$EB,$27,$FD,$89,$F6,$5F,$85,$7D,$0D,$4D,$8F,$F5
	.dc.b	$CE,$9F,$DD,$5F,$F2,$5B,$7E,$0B,$E1,$1F,$0F,$E2,$2D,$FC,$3F,$91
	.dc.b	$FC,$75,$FF,$9D,$E3,$FD,$C9,$BF,$FB,$6E,$31,$3A,$BF,$F6,$F3,$C3
	.dc.b	$CE,$B5,$BF,$3E,$E1,$F3,$F0,$FE,$E7,$7B,$8F,$05,$6D,$A7,$D5,$FB
	.dc.b	$37,$BB,$CA,$11,$87,$67,$DA,$C3,$58,$7F,$37,$8F,$49,$E1,$F4,$F2
	.dc.b	$6B,$D2,$B1,$CA,$5F,$31,$C8,$CC,$72,$3F,$0E,$F8,$F0,$7C,$CD,$38
	.dc.b	$4F,$3F,$4B,$C7,$E8,$C8,$C1,$DD,$F3,$FA,$1C,$A3,$B6,$3F,$CF,$2F
	.dc.b	$51,$E5,$9D,$3F,$99,$3E,$63,$87,$EB,$35,$CC,$7E,$33,$DC,$7C,$AD
	.dc.b	$24,$FF,$B3,$6D,$3A,$8C,$7D,$1F,$F9,$F7,$8B,$F0,$B3,$97,$F0,$F1
	.dc.b	$9E,$2F,$C3,$4E,$5F,$C4,$76,$66,$FF,$05,$55,$FD,$8C,$5C,$13,$7C
	.dc.b	$1D,$77,$7F,$B3,$5F,$AC,$63,$A3,$97,$CA,$7F,$B2,$23,$27,$E1,$1C
	.dc.b	$63,$F3,$23,$7F,$4F,$C3,$E3,$EC,$6D,$E7,$EA,$F8,$3B,$2C,$FE,$3B
	.dc.b	$2F,$F5,$CB,$D7,$D6,$CE,$3F,$85,$CB,$FB,$84,$F8,$70,$F2,$FA,$DE
	.dc.b	$6B,$B3,$C4,$F3,$47,$FA,$A7,$BF,$FC,$D9,$7F,$DC,$67,$F8,$B4,$63
	.dc.b	$E9,$F5,$1F,$81,$9D,$63,$FE,$D9,$F0,$5E,$56,$A3,$F8,$3B,$54,$BE
	.dc.b	$73,$15,$C2,$FB,$9D,$63,$FC,$FD,$56,$3F,$D1,$B5,$C5,$B3,$A9,$FE
	.dc.b	$59,$2F,$57,$C0,$DA,$EB,$3A,$F5,$B3,$5E,$5F,$29,$F6,$54,$E3,$EF
	.dc.b	$5C,$38,$E1,$EA,$8F,$03,$63,$CB,$18,$C7,$FC,$99,$1F,$EA,$4F,$41
	.dc.b	$F2,$1F,$0D,$B1,$D2,$5F,$39,$9A,$F4,$B2,$75,$F3,$97,$1F,$F0,$17
	.dc.b	$C9,$F3,$3D,$F8,$5A,$E8,$F8,$C8,$FC,$A7,$D3,$8B,$D2,$F2,$1F,$5F
	.dc.b	$EA,$3F,$D5,$75,$F0,$27,$4F,$C2,$8C,$E1,$FC,$0C,$FE,$B8,$E1,$C3
	.dc.b	$5F,$E1,$0D,$7F,$B1,$7C,$57,$EB,$57,$37,$FA,$F6,$30,$FF,$AF,$E5
	.dc.b	$F3,$FF,$8B,$9C,$DF,$48,$F4,$1E,$9F,$42,$C7,$FA,$CF,$B7,$81,$ED
	.dc.b	$F9,$8B,$1E,$B7,$ED,$E0,$69,$23,$FD,$BB,$D8,$F7,$BF,$38,$D8,$FC
	.dc.b	$2F,$87,$FC,$D1,$FF,$5C,$4F,$E8,$63,$1A,$FB,$A7,$F9,$54,$63,$5F
	.dc.b	$F6,$E7,$D4,$7F,$A7,$A7,$9F,$97,$FA,$9F,$80,$7C,$55,$3F,$59,$4F
	.dc.b	$63,$17,$F9,$68,$C3,$63,$7E,$CF,$07,$F6,$9F,$1C,$F4,$E9,$D3,$8A
	.dc.b	$DF,$37,$F3,$F8,$7F,$12,$DB,$F0,$01,$C3,$F8,$36,$BF,$D1,$D7,$B3
	.dc.b	$E8,$3D,$3E,$CD,$BF,$D6,$E2,$F5,$7C,$A7,$BB,$D7,$B3,$ED,$70,$FB
	.dc.b	$F3,$B9,$7C,$D3,$FC,$5D,$F6,$4F,$9D,$FE,$24,$DA,$F8,$DC,$98,$9F
	.dc.b	$96,$C8,$F9,$5E,$9D,$BE,$83,$F3,$CF,$34,$FF,$04,$6C,$6B,$EA,$9E
	.dc.b	$9F,$F2,$19,$CF,$42,$FF,$B5,$56,$9F,$97,$68,$C3,$F3,$9A,$FE,$BD
	.dc.b	$AF,$2F,$47,$F9,$74,$1B,$D7,$87,$F9,$97,$D9,$F8,$6F,$F8,$DB,$6F
	.dc.b	$D1,$E7,$6A,$79,$B1,$FE,$43,$87,$F2,$B9,$DC,$7F,$24,$DF,$FA,$FE
	.dc.b	$5E,$D9,$FA,$3E,$A7,$93,$E1,$BE,$8B,$6F,$9C,$7A,$3E,$1F,$15,$F3
	.dc.b	$B0,$F8,$38,$73,$84,$62,$FE,$4F,$17,$DC,$E0,$F8,$2C,$8F,$3F,$97
	.dc.b	$6F,$EA,$3F,$18,$F8,$FE,$5D,$9F,$57,$07,$E1,$8F,$97,$EE,$58,$F8
	.dc.b	$2E,$1F,$87,$74,$F4,$71,$7E,$38,$FE,$C1,$F5,$7F,$C9,$B5,$F5,$6B
	.dc.b	$FD,$DA,$BF,$17,$39,$BF,$F1,$63,$63,$FD,$0C,$FF,$46,$77,$D5,$F9
	.dc.b	$AF,$5E,$57,$F5,$4A,$7C,$4E,$2F,$AB,$FD,$5A,$33,$3F,$8E,$6C,$FE
	.dc.b	$E9,$AF,$F3,$B7,$FB,$A3,$E6,$F5,$FD,$7F,$2F,$EC,$7F,$BE,$7D,$5F
	.dc.b	$F8,$A2,$FF,$B2,$B6,$BF,$4E,$BA,$F7,$D9,$FE,$03,$E0,$FE,$CB,$5F
	.dc.b	$C2,$9A,$FF,$4B,$6B,$FE,$4F,$6B,$FE,$A9,$7F,$D5,$57,$FD,$D1,$B1
	.dc.b	$FB,$6F,$AB,$FF,$6D,$4E,$63,$FC,$25,$0F,$72,$98,$DE,$D7,$F7,$E9
	.dc.b	$FE,$84,$FE,$D1,$B5,$FC,$81,$AF,$F6,$16,$BF,$E9,$4F,$1F,$FC,$41
	.dc.b	$F0,$3F,$EB,$1A,$FF,$04,$AF,$86,$CC,$EB,$C2,$95,$AF,$12,$E1,$FC
	.dc.b	$43,$EA,$DE,$FF,$F8,$41,$B5,$FC,$21,$B3,$FC,$D1,$AF,$F7,$55,$D7
	.dc.b	$FE,$33,$6C,$FF,$87,$1F,$F1,$AB,$63,$5B,$C8,$D5,$53,$F9,$1B,$FD
	.dc.b	$D0,$FE,$32,$57,$FA,$2B,$5F,$F0,$33,$5F,$F4,$57,$D5,$BE,$5B,$FC
	.dc.b	$8C,$D9,$FF,$17,$AF,$A6,$CC,$FF,$92,$FE,$AF,$FC,$A3,$DB,$FC,$15
	.dc.b	$B7,$FE,$95,$7E,$1C,$67,$B3,$87,$81,$FE,$38,$9F,$F1,$7D,$7F,$30
	.dc.b	$6E,$5F,$D7,$1A,$FF,$87,$7E,$AD,$F3,$5F,$E5,$B6,$D7,$F9,$11,$FD
	.dc.b	$5E,$3F,$C5,$DC,$D7,$F4,$85,$DB,$C4,$E5,$EC,$DB,$FF,$26,$FA,$BF
	.dc.b	$E0,$1B,$FE,$DA,$D7,$FC,$5C,$D7,$FD,$64,$D4,$7F,$B1,$3E,$37,$FC
	.dc.b	$E1,$06,$F9,$F8,$7F,$CE,$9F,$57,$FC,$4F,$3F,$E5,$CD,$C3,$FD,$7C
	.dc.b	$FC,$CC,$3A,$B9,$6F,$37,$F9,$8A,$3F,$1F,$1C,$23,$06,$7C,$EE,$E7
	.dc.b	$83,$E1,$6D,$FF,$9B,$2F,$F8,$1E,$1E,$2F,$83,$F2,$FD,$1F,$99,$B5
	.dc.b	$FD,$9A,$93,$E1,$24,$7F,$6A,$F2,$3F,$15,$9F,$F7,$86,$E1,$EE,$FA
	.dc.b	$2F,$F4,$A4,$62,$BD,$F1,$F1,$0F,$8B,$D5,$63,$FD,$03,$3F,$BB,$BE
	.dc.b	$9D,$FF,$E9,$DC,$3F,$82,$9F,$D5,$E7,$4F,$F5,$93,$35,$FE,$E9,$6D
	.dc.b	$7F,$AB,$BE,$4B,$97,$BD,$E4,$F8,$F9,$7F,$E9,$0A,$FC,$AD,$F5,$9F
	.dc.b	$F5,$64,$62,$7F,$CE,$5D,$58,$EF,$CB,$63,$FC,$1B,$3F,$C0,$0F,$F0
	.dc.b	$7C,$FF,$A6,$9C,$1F,$E7,$BC,$BF,$DE,$73,$F0,$3E,$3F,$BE,$DF,$21
	.dc.b	$B1,$E5,$8F,$1E,$D8,$BF,$F1,$39,$F9,$4F,$17,$C2,$9A,$47,$FC,$97
	.dc.b	$F5,$4F,$83,$7D,$5F,$F8,$92,$30,$D8,$CC,$F0,$6B,$FE,$93,$97,$D0
	.dc.b	$7F,$A3,$23,$AB,$5D,$35,$FD,$48,$8E,$3C,$1F,$B2,$FD,$63,$F8,$B4
	.dc.b	$FF,$B2,$BB,$51,$88,$FF,$28,$46,$3C,$4F,$BD,$B7,$ED,$23,$7F,$43
	.dc.b	$8A,$47,$2F,$1B,$0D,$8E,$D8,$7F,$21,$6E,$1A,$8F,$F4,$04,$79,$7E
	.dc.b	$0F,$FC,$93,$5E,$46,$CE,$A7,$F8,$4A,$E9,$E4,$48,$EF,$3D,$7C,$C9
	.dc.b	$1E,$27,$4F,$F9,$B7,$89,$F2,$FF,$C9,$23,$FD,$DB,$18,$F1,$FD,$CD
	.dc.b	$8C,$CB,$E3,$72,$FF,$B9,$23,$3F,$1D,$83,$37,$CB,$EE,$FF,$B4,$4F
	.dc.b	$66,$BE,$36,$F1,$7F,$94,$71,$47,$BE,$78,$65,$F4,$7C,$AF,$F4,$76
	.dc.b	$B7,$CF,$D5,$73,$AF,$4F,$E3,$FE,$D8,$FF,$0A,$35,$FF,$0A,$35,$FF
	.dc.b	$0A,$35,$62,$8F,$81,$1F,$19,$AF,$CB,$6E,$5E,$97,$C0,$FA,$FF,$1F
	.dc.b	$79,$1F,$31,$FF,$01,$D7,$D0,$6B,$F5,$5D,$B9,$F3,$FF,$D1,$B8,$25
	.dc.b	$18,$6C,$FF,$8C,$DA,$FF,$8C,$FE,$75,$E6,$FD,$C6,$C7,$E0,$1A,$FE
	.dc.b	$17,$F1,$FF,$CC,$3F,$87,$FC,$05,$F2,$FF,$84,$5E,$75,$31,$1F,$88
	.dc.b	$FC,$7F,$F0,$95,$FE,$43,$6B,$FE,$43,$6B,$FE,$43,$6A,$F7,$7F,$84
	.dc.b	$E3,$FA,$F6,$C7,$F6,$6D,$FF,$D2,$D1,$88,$FF,$64,$3F,$C0,$E3,$3E
	.dc.b	$B3,$CE,$E4,$DA,$E5,$E1,$7F,$70,$B4,$E6,$79,$AF,$3B,$C5,$EC,$BE
	.dc.b	$5F,$FD,$7F,$68,$C5,$FF,$BB,$23,$13,$A9,$CC,$FF,$16,$3D,$D3,$FC
	.dc.b	$79,$F1,$8E,$A7,$EB,$BF,$C5,$74,$FF,$5A,$46,$3C,$4F,$89,$7C,$6F
	.dc.b	$E0,$FE,$07,$F8,$0B,$F8,$AD,$D7,$DC,$D9,$F9,$CB,$E1,$3E,$77,$E4
	.dc.b	$1F,$C8,$E7,$F9,$5E,$08,$C4,$7C,$76,$BF,$5D,$B0,$F3,$DF,$1F,$FB
	.dc.b	$7F,$8A,$FE,$89,$7E,$ED,$51,$88,$FB,$35,$C1,$F8,$76,$F9,$37,$BB
	.dc.b	$FC,$07,$D1,$FF,$04,$D7,$36,$BF,$82,$9D,$BC,$BE,$4B,$FD,$10,$FD
	.dc.b	$56,$BF,$4E,$D1,$88,$F8,$2D,$7F,$0D,$F8,$23,$F1,$2F,$CA,$36,$3C
	.dc.b	$ED,$7F,$26,$FA,$5F,$5F,$D2,$3E,$43,$F3,$8D,$8C,$C6,$3D,$EF,$D0
	.dc.b	$7D,$3F,$F9,$06,$7F,$18,$D7,$D0,$D7,$EB,$B5,$FC,$E7,$B7,$F7,$5F
	.dc.b	$A1,$F9,$07,$F1,$6E,$7F,$F2,$7F,$1A,$31,$3F,$B1,$AE,$0F,$EB,$3D
	.dc.b	$EF,$07,$4F,$E4,$BF,$20,$DE,$07,$8F,$0F,$FA,$5B,$E1,$7F,$71,$F9
	.dc.b	$03,$E6,$BF,$4E,$DC,$BF,$4C,$D7,$E9,$35,$FC,$8F,$E4,$1F,$C4,$1F
	.dc.b	$C1,$B5,$FD,$FB,$72,$FE,$06,$FF,$14,$8F,$D5,$8F,$9C,$FE,$27,$2D
	.dc.b	$FF,$C0,$F1,$F0,$27,$5E,$D7,$F7,$D7,$87,$B3,$78,$3F,$5D,$1F,$B9
	.dc.b	$6B,$FE,$10,$6B,$FE,$7A,$6A,$5F,$63,$CC,$91,$E6,$8F,$DD,$63,$FC
	.dc.b	$73,$F1,$F7,$D4,$FF,$21,$6D,$7E,$91,$AF,$F2,$05,$F4,$9B,$F3,$24
	.dc.b	$ED,$E5,$5F,$CA,$FE,$D7,$FB,$3E,$1F,$CA,$9B,$B7,$F2,$36,$DF,$F3
	.dc.b	$37,$FC,$25,$3B,$F7,$F9,$12,$7F,$6F,$3C,$17,$D9,$7F,$4A,$AF,$A1
	.dc.b	$1F,$E9,$E8,$C5,$FF,$8E,$1B,$1F,$E2,$CF,$C4,$FF,$B1,$D7,$08,$E7
	.dc.b	$87,$D5,$AF,$F0,$CC,$7D,$E6,$A9,$E1,$1D,$23,$89,$BF,$82,$E4,$73
	.dc.b	$AF,$F1,$CD,$FE,$26,$7F,$C8,$6B,$E3,$F2,$6B,$D6,$B3,$FE,$4B,$9F
	.dc.b	$C9,$65,$FE,$47,$6C,$7A,$EB,$FA,$D7,$DE,$BE,$7A,$92,$7F,$8F,$C7
	.dc.b	$F4,$6D,$3F,$AE,$AF,$A5,$FE,$C5,$F1,$79,$CB,$FA,$C3,$6F,$FB,$45
	.dc.b	$6D,$FD,$0F,$E1,$5F,$E9,$9A,$FF,$32,$FD,$9E,$67,$E3,$46,$30,$FD
	.dc.b	$05,$6E,$7E,$B7,$CC,$FF,$A4,$38,$4D,$24,$7F,$5B,$3E,$E7,$C3,$FD
	.dc.b	$52,$3F,$56,$3F,$C9,$7F,$35,$FE,$0E,$D7,$FC,$7C,$DC,$BF,$DC,$3F
	.dc.b	$0A,$8C,$5F,$F8,$C9,$FE,$23,$3A,$5F,$63,$D0,$93,$FE,$5A,$8E,$3F
	.dc.b	$1C,$FD,$61,$9F,$F9,$D9,$FD,$A5,$7F,$A0,$DA,$FF,$A0,$9D,$B9,$ED
	.dc.b	$E1,$F2,$F2,$30,$7F,$56,$FC,$23,$F4,$B9,$FF,$A1,$FE,$C5,$F4,$FF
	.dc.b	$E8,$46,$D6,$FB,$3F,$87,$79,$FF,$E4,$08,$FF,$02,$35,$FF,$4D,$B5
	.dc.b	$07,$0D,$F8,$DE,$64,$7F,$90,$C7,$EA,$B0,$FD,$E7,$CA,$7F,$01,$F8
	.dc.b	$19,$5F,$B8,$8A,$FF,$8A,$BC,$17,$CA,$FF,$07,$33,$CD,$FE,$30,$DB
	.dc.b	$F5,$ED,$BF,$F1,$B4,$7E,$EB,$CF,$3A,$CB,$FC,$5C,$E4,$FF,$0E,$38
	.dc.b	$CB,$8D,$72,$C3,$FC,$5B,$5E,$2C,$BF,$C5,$8D,$7F,$BC,$35,$FF,$60
	.dc.b	$DB,$F1,$9F,$E5,$6F,$C8,$31,$F2,$FF,$EB,$96,$CF,$D9,$BF,$F2,$77
	.dc.b	$C4,$7F,$AF,$23,$FC,$22,$D7,$FC,$9A,$DD,$3F,$CA,$2F,$F7,$F6,$FB
	.dc.b	$BC,$89,$3C,$BA,$3A,$F9,$1F,$E4,$8B,$FE,$73,$3F,$E2,$E6,$BF,$E5
	.dc.b	$36,$BF,$E6,$37,$F9,$EB,$E4,$F2,$BC,$E7,$47,$CF,$E5,$75,$23,$FE
	.dc.b	$53,$9F,$F2,$E3,$F1,$B5,$7D,$27,$C3,$6C,$7F,$94,$B5,$F9,$53,$F1
	.dc.b	$2F,$E2,$DC,$AF,$8A,$FC,$8F,$37,$11,$F1,$B7,$F9,$66,$36,$F8,$FE
	.dc.b	$FF,$C1,$24,$F9,$5F,$89,$F0,$1F,$48,$91,$F4,$A3,$FD,$DB,$F0,$4F
	.dc.b	$AC,$BF,$09,$2B,$FD,$F9,$39,$6F,$89,$FE,$EC,$8F,$E0,$F1,$FC,$5F
	.dc.b	$8F,$FC,$A3,$39,$3F,$DE,$27,$F9,$66,$3F,$CE,$71,$FC,$3B,$E1,$3F
	.dc.b	$DF,$B3,$99,$FE,$52,$27,$9B,$E2,$F6,$BF,$6E,$BE,$6F,$FB,$CC,$FF
	.dc.b	$98,$3E,$4B,$31,$8C,$7F,$D6,$2D,$BC,$F3,$FF,$00,$BE,$18,$7E,$6C
	.dc.b	$C5,$FF,$AE,$39,$7B,$CE,$A7,$FB,$43,$63,$39,$7F,$B7,$27,$F8,$5E
	.dc.b	$3E,$48,$C4,$79,$F0,$F0,$46,$27,$CD,$97,$92,$BF,$D6,$DF,$79,$FE
	.dc.b	$9B,$9F,$F0,$AF,$72,$31,$87,$FA,$B5,$B5,$F1,$F0,$E9,$3E,$DF,$D3
	.dc.b	$BF,$D5,$D3,$E4,$8C,$73,$77,$D1,$5A,$AF,$31,$FA,$5F,$A2,$D4,$67
	.dc.b	$3F,$9E,$FF,$8C,$27,$A1,$F5,$DC,$3C,$AB,$AD,$F2,$E7,$F7,$69,$FE
	.dc.b	$C1,$E9,$B3,$FA,$6B,$E1,$87,$2C,$7F,$C6,$7F,$6B,$FD,$9A,$7B,$87
	.dc.b	$FA,$5B,$E8,$31,$E5,$FC,$B7,$5E,$75,$9D,$D7,$4F,$2B,$F8,$A3,$75
	.dc.b	$EB,$3A,$DF,$F1,$EA,$BF,$8C,$61,$86,$BF,$D2,$A3,$55,$FD,$E6,$3C
	.dc.b	$D3,$A7,$D1,$E5,$FE,$11,$EA,$D7,$93,$5D,$F9,$DC,$ED,$F4,$FF,$E4
	.dc.b	$0C,$BE,$2D,$FE,$45,$F8,$F1,$9C,$BA,$DF,$B3,$EA,$31,$8F,$9E,$AB
	.dc.b	$F2,$E2,$E3,$DB,$9F,$EE,$E8,$FE,$00,$7D,$C7,$2E,$4F,$F7,$D7,$D2
	.dc.b	$74,$AD,$4E,$79,$13,$9B,$7D,$55,$AC,$37,$1E,$F3,$A6,$C7,$3F,$A3
	.dc.b	$DC,$7E,$43,$2F,$D2,$6F,$E3,$EB,$3E,$3F,$D1,$FF,$C7,$D5,$B9,$CB
	.dc.b	$EC,$8C,$F0,$7E,$5D,$BC,$7F,$B9,$8E,$5E,$CF,$B7,$B7,$F8,$35,$FF
	.dc.b	$00,$9F,$F3,$8B,$4F,$E1,$4F,$EC,$CC,$9F,$C3,$FF,$59,$7C,$DF,$38
	.dc.b	$F9,$11,$89,$F9,$F5,$FC,$4F,$0F,$E3,$B8,$BE,$6B,$FC,$FF,$E4,$FF
	.dc.b	$A0,$39,$53,$9C,$3F,$96,$E3,$F5,$9F,$E5,$F1,$9C,$BF,$96,$D2,$B5
	.dc.b	$1F,$CB,$0F,$E6,$0E,$1F,$E4,$2D,$7D,$5C,$7F,$E8,$97,$D9,$ED,$79
	.dc.b	$30,$F7,$71,$D9,$E7,$37,$0F,$04,$E7,$89,$CC,$F8,$7C,$37,$1F,$F0
	.dc.b	$3F,$C3,$E2,$72,$F8,$72,$FE,$F4,$DB,$E6,$DB,$37,$2E,$B8,$F6,$6E
	.dc.b	$3F,$E1,$47,$0F,$F8,$5B,$3B,$8F,$F5,$48,$7C,$87,$F9,$AD,$FE,$F8
	.dc.b	$D8,$FF,$0D,$F1,$FC,$19,$DD,$FB,$23,$C8,$48,$FF,$05,$3E,$9F,$D9
	.dc.b	$BF,$A9,$70,$FF,$AF,$35,$46,$2B,$FC,$AF,$18,$E7,$F9,$53,$9C,$53
	.dc.b	$9A,$FF,$45,$36,$3E,$F3,$5F,$ED,$6D,$7A,$CB,$CA,$EA,$76,$FB,$8B
	.dc.b	$7F,$E6,$8C,$7C,$D4,$9D,$4E,$4E,$7F,$23,$8C,$3E,$56,$BF,$C1,$27
	.dc.b	$E4,$F6,$FA,$BE,$AB,$E4,$4E,$6B,$FC,$07,$A7,$C2,$9F,$8C,$4B,$F6
	.dc.b	$E4,$8F,$95,$F0,$FF,$13,$A7,$E2,$B6,$BC,$98,$7C,$6B,$FE,$03,$AF
	.dc.b	$AA,$BE,$5F,$67,$0B,$DE,$4F,$C0,$9E,$17,$E6,$5F,$E4,$D9,$79,$9C
	.dc.b	$D7,$07,$38,$FD,$B5,$E2,$F1,$BE,$4D,$AF,$2D,$BE,$65,$F4,$FF,$AB
	.dc.b	$8C,$36,$1F,$07,$63,$0F,$E9,$67,$93,$67,$75,$FD,$55,$B3,$F8,$4D
	.dc.b	$9E,$8F,$ED,$FE,$37,$FB,$05,$FC,$29,$CB,$FE,$3A,$8C,$D7,$F0,$CF
	.dc.b	$D6,$7F,$D6,$B3,$93,$F8,$03,$88,$FE,$23,$29,$CF,$D9,$EB,$E1,$EC
	.dc.b	$7F,$91,$AF,$87,$07,$E3,$BE,$67,$F3,$E7,$E8,$39,$ED,$79,$5C,$3E
	.dc.b	$7F,$8F,$B9,$FF,$93,$0F,$99,$E7,$7C,$93,$E5,$71,$DA,$FA,$3C,$1F
	.dc.b	$6A,$BF,$DE,$4D,$7F,$93,$35,$FF,$29,$B5,$FF,$65,$FE,$1F,$F6,$27
	.dc.b	$ED,$FD,$BF,$FB,$01,$FB,$B1,$99,$F2,$4E,$5F,$CF,$FC,$8F,$F5,$67
	.dc.b	$F4,$3F,$0C,$F6,$FF,$94,$4E,$77,$FF,$64,$46,$3C,$29,$97,$E6,$2B
	.dc.b	$F6,$8D,$8F,$E6,$5F,$81,$3E,$CA,$FD,$66,$1B,$5F,$E7,$69,$73,$E6
	.dc.b	$30,$DD,$3E,$67,$DC,$38,$FF,$2C,$BE,$AA,$C6,$1B,$7F,$9E,$9E,$0D
	.dc.b	$E5,$CD,$F4,$79,$9E,$8A,$B3,$FA,$DF,$6B,$F6,$47,$EC,$FD,$CB,$1F
	.dc.b	$B1,$70,$FE,$CF,$DC,$B8,$7F,$0C,$FC,$0D,$F1,$5F,$D4,$DB,$5F,$C0
	.dc.b	$17,$C4,$E6,$74,$C9,$D7,$A9,$6F,$E1,$F9,$FF,$08,$F8,$6D,$9F,$80
	.dc.b	$FE,$D7,$6F,$F0,$3B,$FE,$9C,$6A,$0E,$17,$57,$F0,$A9,$FC,$5E,$BF
	.dc.b	$C1,$63,$13,$F9,$F7,$F6,$BD,$1F,$D2,$E7,$FC,$04,$DF,$2F,$FB,$94
	.dc.b	$7F,$22,$BF,$E9,$AE,$9F,$E9,$FA,$7F,$6A,$7F,$A5,$B7,$0F,$E9,$6D
	.dc.b	$7F,$A5,$B5,$FE,$AF,$F8,$73,$9F,$FC,$09,$F4,$2F,$C7,$9C,$7F,$E0
	.dc.b	$3A,$D6,$3F,$CD,$AB,$53,$FE,$07,$FC,$3F,$C9,$8A,$5F,$13,$F0,$36
	.dc.b	$FE,$D1,$5A,$9C,$F9,$9F,$E0,$FF,$98,$FF,$17,$61,$FD,$89,$AF,$F6
	.dc.b	$26,$BF,$DB,$9A,$FF,$6C,$FC,$3D,$EA,$FF,$86,$5B,$A7,$F4,$67,$FA
	.dc.b	$93,$63,$FC,$3D,$EB,$3E,$62,$98,$F9,$EB,$5F,$D3,$BF,$1E,$FE,$48
	.dc.b	$FF,$78,$6E,$3F,$DE,$1A,$FF,$78,$6B,$FE,$04,$E8,$7C,$B7,$CE,$FC
	.dc.b	$7B,$F3,$37,$FD,$6A,$72,$FC,$D8,$CD,$7F,$6E,$FD,$8F,$F9,$7C,$FE
	.dc.b	$47,$81,$F4,$3F,$C8,$F8,$FF,$C2,$95,$AB,$FA,$53,$9B,$FF,$0A,$FE
	.dc.b	$2F,$FE,$6C,$9F,$F0,$6B,$5C,$FB,$7F,$E3,$78,$C4,$7F,$86,$3D,$0F
	.dc.b	$E9,$AF,$F8,$01,$B8,$FF,$88,$D7,$A5,$FE,$49,$73,$87,$F8,$93,$F1
	.dc.b	$FF,$6D,$AF,$FA,$73,$F8,$7F,$98,$FE,$51,$5B,$E0,$FC,$9C,$62,$7F
	.dc.b	$C5,$3E,$1B,$C7,$FF,$26,$36,$FF,$C5,$6D,$9F,$F1,$5B,$5F,$F1,$5F
	.dc.b	$DF,$F8,$B1,$FF,$1D,$36,$BF,$C7,$2B,$E7,$7F,$CC,$4E,$6B,$FC,$73
	.dc.b	$F1,$2F,$8F,$FF,$33,$36,$FF,$C2,$6F,$F9,$96,$7F,$C4,$AE,$5F,$F1
	.dc.b	$39,$FD,$D0,$FF,$1E,$94,$FF,$24,$94,$FF,$23,$97,$17,$B5,$FC,$D2
	.dc.b	$3E,$55,$7F,$13,$9C,$F0,$7F,$9F,$5D,$F9,$F6,$FE,$AD,$7F,$8B,$A5
	.dc.b	$F8,$FF,$F9,$55,$B8,$71,$FC,$7D,$C1,$CE,$B5,$F8,$FF,$AE,$BF,$91
	.dc.b	$1F,$EB,$C7,$0F,$8D,$AE,$75,$FF,$42,$60,$9F,$8D,$C1,$F8,$28,$C7
	.dc.b	$C3,$FA,$67,$39,$F3,$8F,$67,$E7,$FF,$9D,$A3,$7E,$2F,$F6,$78,$73
	.dc.b	$8C,$5F,$82,$7F,$67,$97,$FB,$49,$B7,$FE,$B3,$9F,$F5,$74,$FF,$B5
	.dc.b	$BF,$17,$FF,$22,$C7,$EB,$FF,$15,$FE,$17,$3F,$40,$D7,$C0,$FF,$28
	.dc.b	$8C,$E5,$EB,$BE,$35,$AB,$FA,$11,$FC,$8D,$FB,$0E,$B0,$7D,$67,$F9
	.dc.b	$B6,$FE,$23,$6B,$EF,$3F,$8B,$8C,$E9,$F7,$E3,$13,$AD,$B1,$F2,$39
	.dc.b	$7F,$D7,$DF,$41,$CF,$37,$3A,$30,$D9,$DE,$3C,$19,$87,$F9,$DA,$78
	.dc.b	$7C,$E6,$30,$FF,$6C,$E0,$8F,$F6,$14,$EA,$BE,$6B,$6B,$FD,$B8,$D7
	.dc.b	$F1,$6F,$8E,$BC,$E7,$A3,$E8,$6D,$78,$61,$19,$8C,$72,$DB,$FF,$26
	.dc.b	$1F,$DD,$FE,$5F,$EE,$BF,$98,$6B,$92,$BA,$F8,$3F,$89,$F8,$1D,$C4
	.dc.b	$D2,$57,$E2,$27,$F4,$FC,$1C,$70,$FE,$3D,$39,$6B,$FE,$FB,$9D,$BF
	.dc.b	$0B,$F0,$3F,$E8,$AF,$CD,$6B,$F8,$FD,$11,$8B,$FE,$53,$18,$EF,$FD
	.dc.b	$9C,$63,$F5,$9F,$F4,$AD,$FD,$76,$BC,$DB,$1C,$AF,$A6,$BF,$E1,$DC
	.dc.b	$3B,$FE,$07,$F2,$1B,$87,$DA,$A4,$E5,$B1,$9F,$91,$F1,$DF,$39,$EE
	.dc.b	$D8,$CC,$FE,$4A,$78,$BE,$5F,$91,$CE,$0C,$BF,$D5,$73,$EB,$9C,$D7
	.dc.b	$48,$73,$BD,$07,$CC,$F6,$8C,$36,$33,$3F,$C8,$DF,$E7,$8F,$F2,$5F
	.dc.b	$C3,$BF,$CD,$0D,$F1,$39,$46,$23,$F8,$15,$F4,$7F,$96,$7E,$0B,$E3
	.dc.b	$F4,$3F,$C5,$3C,$1E,$E6,$CE,$B7,$C7,$B9,$F0,$38,$BF,$91,$7C,$97
	.dc.b	$F9,$D6,$BF,$9B,$35,$FE,$21,$7C,$9F,$0F,$E0,$AE,$FF,$CE,$E8,$8C
	.dc.b	$4F,$F1,$C8,$C4,$EA,$73,$3F,$CE,$8F,$93,$F9,$5F,$C2,$7D,$AD,$C7
	.dc.b	$F9,$B3,$79,$BF,$A8,$57,$29,$FE,$75,$F8,$37,$FA,$66,$FF,$A7,$3F
	.dc.b	$E4,$D6,$FC,$1F,$D3,$8D,$47,$CA,$FC,$1B,$F3,$EF,$F4,$CC,$D1,$88
	.dc.b	$FE,$BC,$D7,$EF,$B6,$1F,$3A,$FA,$AF,$C0,$36,$BF,$B0,$35,$FB,$2F
	.dc.b	$FA,$E5,$FE,$C9,$F8,$37,$FA,$B1,$BF,$65,$F8,$98,$37,$E4,$BF,$19
	.dc.b	$F8,$33,$E8,$BE,$8B,$6F,$E8,$35,$F9,$D9,$23,$11,$FD,$AB,$E8,$3F
	.dc.b	$C4,$93,$FA,$46,$C7,$FA,$3F,$7F,$F1,$3C,$62,$3F,$A9,$3F,$C5,$33
	.dc.b	$E1,$F8,$33,$EB,$69,$2B,$97,$1F,$FA,$66,$D5,$FE,$AA,$9C,$88,$7E
	.dc.b	$0C,$F8,$0F,$E5,$31,$86,$C6,$67,$FB,$4C,$E5,$FE,$D7,$D0,$FC,$1B
	.dc.b	$FC,$59,$1D,$A7,$B7,$97,$EF,$7F,$A9,$47,$FA,$36,$76,$FB,$BF,$06
	.dc.b	$76,$7C,$56,$C7,$36,$BB,$E9,$F9,$5F,$82,$BE,$DB,$D2,$D9,$F3,$B5
	.dc.b	$FD,$16,$28,$C4,$7F,$87,$B3,$7D,$37,$9A,$3E,$8F,$D0,$3E,$D3,$EC
	.dc.b	$36,$3F,$D4,$D3,$BE,$5F,$F2,$FD,$FF,$6F,$F9,$FF,$BD,$F6,$CF,$B5
	.dc.b	$A7,$F0,$7E,$24,$FE,$0F,$E9,$DE,$FF,$D2,$6E,$7F,$41,$AF,$A9,$AF
	.dc.b	$E3,$7F,$04,$7D,$AF,$E3,$5B,$1F,$4D,$AF,$E3,$17,$E0,$BF,$93,$FC
	.dc.b	$17,$F9,$06,$DF,$92,$6F,$67,$E2,$DB,$7F,$89,$7F,$D0,$6F,$DF,$FB
	.dc.b	$B7,$DA,$FE,$29,$B1,$F8,$86,$BF,$49,$AF,$D1,$FC,$17,$F8,$75,$FE
	.dc.b	$67,$E7,$7E,$FE,$3F,$A4,$6C,$7F,$A9,$67,$75,$FE,$D3,$FC,$16,$F4
	.dc.b	$FF,$B1,$B2,$FD,$77,$CE,$78,$32,$FA,$B5,$AF,$C1,$3F,$A7,$E1,$F8
	.dc.b	$46,$BF,$B7,$6B,$FA,$B6,$BF,$A9,$FC,$13,$F8,$3B,$F9,$F6,$BF,$B8
	.dc.b	$6D,$7E,$F1,$FE,$7C,$FF,$A9,$23,$FC,$87,$E9,$7F,$66,$5A,$FF,$9D
	.dc.b	$1F,$DB,$98,$F8,$1D,$7C,$47,$C0,$8D,$F9,$1F,$E1,$EA,$FE,$09,$1F
	.dc.b	$E2,$26,$BF,$E6,$06,$BF,$D0,$E7,$F9,$F2,$3F,$CA,$BF,$0F,$D8,$F2
	.dc.b	$4F,$2C,$7F,$8A,$FE,$09,$FC,$2E,$BF,$8A,$B6,$33,$F5,$7F,$C7,$63
	.dc.b	$F1,$F1,$FC,$73,$F1,$57,$D0,$FF,$21,$6D,$FF,$07,$6B,$D2,$FF,$93
	.dc.b	$7E,$03,$FD,$FE,$3F,$86,$35,$FE,$4A,$DB,$FE,$52,$FF,$6F,$E6,$FE
	.dc.b	$44,$E5,$FE,$49,$F6,$FF,$C4,$27,$F9,$7B,$63,$7F,$05,$F4,$FE,$97
	.dc.b	$FC,$4F,$87,$C5,$79,$36,$35,$F3,$7B,$78,$7D,$24,$9F,$B3,$5C,$12
	.dc.b	$8F,$E8,$5F,$00,$FB,$9A,$71,$AF,$F2,$27,$D0,$FE,$C3,$E3,$0F,$E2
	.dc.b	$B3,$FF,$1D,$36,$7F,$CF,$EE,$CC,$4F,$0F,$B1,$FF,$17,$C6,$0F,$DD
	.dc.b	$F7,$BF,$05,$FE,$7D,$7F,$D4,$1B,$C1,$FD,$51,$B8,$FF,$4D,$5F,$75
	.dc.b	$FD,$6F,$E4,$5F,$61,$7D,$B9,$8F,$F4,$39,$CB,$F9,$CF,$C1,$79,$BF
	.dc.b	$37,$E8,$FE,$81,$FA,$38,$CC,$7C,$8E,$1F,$F3,$DF,$19,$A4,$8F,$F4
	.dc.b	$0F,$E2,$BF,$EE,$A7,$F5,$2D,$7F,$D7,$81,$C6,$DF,$B3,$AE,$93,$FE
	.dc.b	$E9,$F8,$3F,$E1,$1B,$FF,$BC,$5B,$1F,$E5,$26,$FC,$17,$F8,$17,$EE
	.dc.b	$3F,$CA,$51,$81,$F5,$F8,$33,$E5,$FF,$98,$3F,$05,$FD,$FE,$1F,$EA
	.dc.b	$F5,$F6,$99,$35,$E2,$59,$D4,$57,$FD,$11,$3C,$9F,$F4,$57,$CB,$FB
	.dc.b	$34,$FF,$7C,$FA,$1F,$E3,$2D,$FF,$88,$36,$3F,$D0,$53,$BA,$FE,$05
	.dc.b	$3F,$C6,$9F,$F2,$43,$5E,$7F,$15,$FE,$AC,$8C,$47,$FB,$3A,$5F,$86
	.dc.b	$FF,$9F,$1B,$1F,$EC,$B6,$BF,$4A,$7F,$92,$DF,$ED,$BE,$6B,$F7,$FF
	.dc.b	$07,$2B,$F7,$F1,$5F,$DF,$C5,$CB,$F7,$F3,$FE,$2B,$FB,$BF,$DB,$F7
	.dc.b	$1F,$DE,$5F,$F8,$7D,$B1,$9F,$17,$85,$4A,$FF,$0C,$56,$8F,$E0,$38
	.dc.b	$BF,$06,$FE,$69,$3F,$E6,$B8,$C7,$A1,$FE,$D0,$6C,$7F,$9B,$C3,$87
	.dc.b	$FC,$23,$F7,$FF,$B7,$C8,$D3,$F7,$53,$BC,$FF,$C8,$7F,$74,$F8,$2F
	.dc.b	$F2,$43,$70,$FF,$17,$35,$FF,$23,$BF,$F0,$01,$FF,$19,$5F,$F9,$7D
	.dc.b	$FF,$25,$B6,$BF,$C9,$ED,$BF,$F2,$45,$73,$9F,$0F,$DC,$FF,$91,$20
	.dc.b	$D2,$5F,$F9,$12,$FE,$DB,$FE,$4D,$6E,$3E,$EE,$3F,$4D,$7F,$B1,$31
	.dc.b	$F9,$9F,$47,$F2,$75,$34,$98,$F5,$FB,$DF,$F3,$46,$5F,$99,$8E,$D1
	.dc.b	$CE,$B8,$C7,$C4,$CF,$E3,$ED,$23,$07,$C8,$79,$FC,$5C,$8C,$1F,$33
	.dc.b	$E9,$3E,$01,$1F,$F3,$0D,$23,$FC,$CF,$8B,$F0,$9F,$29,$2B,$91,$FE
	.dc.b	$70,$F3,$7F,$C9,$BD,$DF,$EE,$0F,$8A,$FE,$17,$1F,$E5,$0E,$FC,$4F
	.dc.b	$FB,$91,$7C,$5F,$FB,$D5,$B7,$F8,$DA,$FF,$17,$CE,$4F,$F3,$CF,$AB
	.dc.b	$FE,$93,$E6,$FF,$45,$B7,$8F,$FD,$08,$DB,$5F,$27,$CD,$97,$FA,$9D
	.dc.b	$7D,$2F,$F9,$2A,$7F,$CC,$BE,$27,$F8,$1E,$73,$F0,$DF,$E0,$99,$CC
	.dc.b	$7F,$82,$EB,$85,$FF,$AC,$DE,$5A,$7F,$BE,$A7,$57,$C6,$31,$1D,$0F
	.dc.b	$2B,$89,$E7,$87,$85,$E7,$E3,$5B,$FF,$50,$7E,$09,$FE,$DF,$9F,$F3
	.dc.b	$0F,$5A,$31,$3D,$EB,$86,$3F,$EA,$39,$CB,$FE,$A5,$FC,$5B,$FC,$53
	.dc.b	$87,$26,$DF,$69,$E3,$A3,$F0,$87,$36,$1B,$3E,$E3,$E7,$B8,$7D,$D9
	.dc.b	$F0,$F8,$EE,$1F,$03,$27,$86,$1D,$63,$A6,$1E,$8B,$EF,$DB,$63,$F4
	.dc.b	$37,$D1,$BA,$7D,$9C,$BD,$33,$FE,$3F,$FC,$21,$E8,$FC,$28,$47,$F6
	.dc.b	$27,$E9,$18,$F4,$35,$1C,$23,$D8,$64,$AF,$F5,$AD,$2F,$99,$FC,$4F
	.dc.b	$1F,$F4,$8C,$F5,$6C,$7F,$17,$AF,$C6,$1E,$27,$0F,$E3,$E7,$A3,$FC
	.dc.b	$B5,$AF,$07,$FC,$D3,$C5,$FC,$C6,$3D,$7E,$6F,$E3,$1E,$4D,$7E,$13
	.dc.b	$5F,$83,$F5,$BF,$CE,$1F,$83,$5D,$1B,$19,$C3,$D7,$5F,$9B,$7D,$BF
	.dc.b	$6D,$9E,$3F,$45,$5B,$FA,$13,$5E,$FD,$5F,$85,$E2,$FC,$4F,$D6,$6D
	.dc.b	$B1,$F8,$2F,$C2,$3F,$D8,$AE,$F8,$D5,$D7,$81,$5A,$C3,$F5,$51,$8F
	.dc.b	$91,$D4,$BF,$06,$B1,$86,$63,$D7,$E1,$EE,$1F,$96,$EB,$71,$F9,$7A
	.dc.b	$DE,$3F,$9E,$73,$87,$AE,$72,$FE,$6B,$F0,$87,$B7,$BC,$3D,$11,$89
	.dc.b	$F8,$57,$CB,$1F,$4C,$E5,$B7,$EF,$F0,$3F,$D3,$17,$D6,$31,$32,$3F
	.dc.b	$53,$F8,$43,$E7,$F1,$3E,$BF,$83,$FB,$DE,$37,$DF,$E6,$FE,$0D,$18
	.dc.b	$C3,$F8,$E6,$7F,$E5,$C7,$FD,$4D,$5B,$D3,$FC,$F1,$1F,$CF,$67,$58
	.dc.b	$7E,$6B,$91,$18,$8F,$E5,$11,$8F,$7F,$FD,$85,$1D,$9A,$FF,$34,$D3
	.dc.b	$EB,$5F,$F4,$97,$EE,$63,$F9,$5B,$46,$A9,$39,$8D,$D7,$FA,$F7,$3F
	.dc.b	$F0,$7F,$CE,$7B,$6F,$FB,$B3,$63,$F8,$CF,$C6,$DF,$35,$39,$EA,$4F
	.dc.b	$9A,$B7,$87,$F6,$48,$C4,$EA,$3F,$B3,$F3,$F3,$F6,$F9,$36,$3F,$C3
	.dc.b	$3E,$EE,$1A,$7F,$87,$23,$C5,$F6,$F2,$3E,$89,$F8,$0F,$98,$56,$75
	.dc.b	$F8,$33,$98,$F9,$6A,$76,$FF,$8D,$4C,$46,$A3,$33,$FC,$4B,$37,$EC
	.dc.b	$DF,$E3,$AB,$F0,$CA,$73,$1F,$1E,$3E,$54,$ED,$FF,$17,$9F,$1F,$6F
	.dc.b	$CB,$BF,$E4,$23,$07,$CB,$87,$81,$C8,$FA,$5E,$85,$BF,$F1,$DB,$87
	.dc.b	$FC,$7B,$F1,$D6,$FE,$09,$E5,$B7,$87,$E4,$3F,$01,$5F,$0F,$C3,$C3
	.dc.b	$EC,$30,$7C,$5D,$3F,$86,$B6,$3E,$31,$EB,$E7,$FC,$31,$F8,$7F,$86
	.dc.b	$B3,$FE,$50,$6E,$3F,$E5,$06,$F3,$FF,$BC,$31,$FC,$87,$C6,$79,$AF
	.dc.b	$E6,$F0,$23,$E1,$C6,$1D,$61,$EE,$9D,$6F,$CA,$F9,$B6,$FF,$C6,$1F
	.dc.b	$47,$FE,$C7,$6F,$99,$FA,$1C,$1F,$34,$FA,$AF,$D2,$76,$B0,$E1,$5F
	.dc.b	$EC,$8A,$D6,$1F,$D3,$5B,$1F,$E7,$F6,$BF,$D6,$9A,$FF,$89,$67,$FD
	.dc.b	$7B,$3F,$8A,$8F,$0F,$C2,$F8,$27,$31,$DE,$31,$30,$FF,$25,$7E,$12
	.dc.b	$E9,$C5,$7E,$03,$14,$E6,$BE,$09,$EC,$71,$3B,$C3,$F2,$DC,$1F,$E8
	.dc.b	$2F,$BA,$FC,$CF,$9F,$FC,$05,$B8,$7F,$A2,$9C,$3F,$E8,$CE,$6B,$87
	.dc.b	$FB,$B3,$F0,$F7,$2F,$CD,$6D,$7B,$F7,$FB,$39,$7F,$6D,$B5,$69,$F9
	.dc.b	$DC,$7F,$C8,$9F,$AA,$DA,$FC,$0C,$7C,$D8,$D7,$E1,$4F,$88,$FB,$95
	.dc.b	$F8,$8A,$FF,$7F,$CE,$71,$F1,$72,$FF,$1F,$F8,$67,$B7,$7E,$AB,$0D
	.dc.b	$AF,$C9,$FC,$7F,$F5,$3D,$BF,$A3,$36,$3F,$96,$3F,$E7,$67,$7E,$7F
	.dc.b	$E2,$BF,$0A,$7C,$25,$31,$F0,$D6,$BF,$D0,$92,$8C,$6D,$F6,$E3,$13
	.dc.b	$AA,$FC,$C7,$E0,$6F,$94,$FE,$71,$DF,$F2,$74,$F9,$93,$9B,$FD,$5C
	.dc.b	$E5,$FF,$3C,$43,$ED,$39,$DF,$46,$CF,$E5,$67,$F9,$43,$63,$C3,$D1
	.dc.b	$B8,$FD,$A7,$BD,$FD,$27,$D7,$FE,$53,$97,$F6,$BF,$13,$F3,$6F,$F6
	.dc.b	$D8,$CD,$7F,$53,$9C,$B6,$BF,$81,$E7,$FC,$11,$FE,$A9,$F3,$1F,$88
	.dc.b	$9F,$EB,$4D,$DF,$F9,$97,$E1,$5F,$ED,$49,$FE,$13,$39,$F1,$93,$9A
	.dc.b	$FE,$10,$DC,$FF,$C0,$0B,$E1,$FF,$B1,$39,$C3,$FD,$3D,$DD,$FD,$89
	.dc.b	$FD,$C3,$6F,$F9,$78,$E0,$FC,$22,$3F,$3B,$E9,$FD,$8E,$6F,$CE,$4F
	.dc.b	$F8,$76,$7F,$CB,$ED,$7E,$AC,$7E,$87,$F0,$AF,$8B,$1C,$E7,$FA,$4E
	.dc.b	$7F,$A9,$9C,$CF,$84,$FE,$F5,$B7,$DF,$7F,$AB,$37,$53,$FE,$16,$AF
	.dc.b	$F1,$03,$9F,$8D,$FE,$F1,$5F,$D8,$5A,$FF,$AF,$EB,$FE,$00,$FA,$BF
	.dc.b	$E0,$19,$CC,$FB,$FE,$9D,$F8,$57,$F2,$8C,$7F,$C5,$9F,$85,$7E,$6A
	.dc.b	$BF,$AA,$D6,$B9,$3F,$C2,$13,$9A,$FF,$38,$4B,$E3,$5E,$4F,$F8,$57
	.dc.b	$F0,$AF,$64,$E6,$31,$E4,$FF,$86,$7E,$63,$FC,$C1,$9F,$9D,$AF,$F8
	.dc.b	$85,$AF,$F7,$96,$EE,$FC,$29,$EB,$7F,$88,$DB,$D1,$FD,$51,$7D,$36
	.dc.b	$67,$FC,$50,$3E,$E2,$98,$FA,$0B,$7F,$D7,$3C,$FF,$E6,$39,$FC,$B6
	.dc.b	$BF,$E3,$56,$BF,$E0,$D6,$EC,$FC,$27,$F2,$D5,$FD,$9B,$A9,$F4,$5F
	.dc.b	$E3,$BA,$D7,$81,$FE,$3B,$9C,$D7,$F8,$62,$5F,$45,$FC,$97,$9B,$FC
	.dc.b	$C1,$F8,$4B,$E0,$B9,$4E,$6F,$FC,$91,$39,$7F,$BA,$FE,$13,$FE,$5B
	.dc.b	$DB,$FC,$98,$D7,$FC,$48,$DE,$43,$7D,$2F,$F3,$D6,$5F,$E6,$F6,$DF
	.dc.b	$F9,$55,$7C,$C7,$F9,$4D,$CE,$0D,$3F,$D1,$FE,$1F,$F9,$5F,$F0,$9F
	.dc.b	$6C,$FE,$33,$E3,$BF,$63,$A2,$73,$9F,$F9,$F6,$FF,$BD,$DF,$E6,$DA
	.dc.b	$FF,$99,$5A,$FF,$8F,$5B,$4F,$C2,$79,$E9,$FE,$6B,$6D,$7F,$90,$D7
	.dc.b	$D0,$FF,$9A,$DC,$D3,$DE,$FF,$15,$D7,$F8,$65,$B8,$FF,$A4,$57,$C8
	.dc.b	$FF,$9C,$DC,$E3,$FE,$79,$AF,$D1,$BF,$EE,$67,$FC,$A4,$60,$FF,$2C
	.dc.b	$14,$FD,$C1,$4F,$E2,$99,$7F,$82,$7D,$DF,$F4,$13,$FE,$0B,$F4,$6F
	.dc.b	$C0,$26,$C6,$BF,$42,$FF,$55,$61,$FE,$62,$B4,$63,$0F,$F2,$8C,$3F
	.dc.b	$09,$FE,$2F,$5E,$0A,$D6,$9F,$EE,$DA,$7E,$15,$FE,$C6,$C3,$FE,$00
	.dc.b	$C6,$38,$BF,$D4,$6D,$C3,$E0,$36,$77,$F8,$57,$C6,$7F,$DA,$AF,$CA
	.dc.b	$9F,$33,$99,$39,$9F,$3F,$E4,$9F,$E3,$FA,$F6,$36,$3D,$7F,$85,$FF
	.dc.b	$8A,$32,$F4,$4E,$5F,$54,$7E,$E3,$03,$49,$BF,$FA,$F3,$99,$F8,$5F
	.dc.b	$F0,$AC,$BE,$57,$E4,$9F,$E4,$AC,$7F,$90,$CE,$73,$F9,$73,$9A,$FB
	.dc.b	$3F,$86,$7F,$61,$C3,$9F,$91,$F6,$A3,$18,$71,$C7,$E7,$7C,$93,$E2
	.dc.b	$FE,$FB,$71,$FB,$5F,$65,$F1,$63,$1A,$7D,$A8,$F9,$04,$E9,$FC,$3B
	.dc.b	$F3,$FA,$5F,$85,$BC,$AF,$F3,$4B,$71,$FF,$53,$CB,$EE,$F8,$4E,$71
	.dc.b	$FF,$2C,$47,$32,$63,$FE,$72,$6D,$FE,$7E,$F9,$4F,$FB,$99,$B1,$99
	.dc.b	$FB,$95,$FE,$EE,$BF,$E6,$93,$E2,$6C,$78,$27,$CF,$E6,$77,$65,$7A
	.dc.b	$17,$D4,$E2,$BF,$BF,$9F,$DF,$7E,$8B,$F6,$4F,$F7,$F0,$E1,$D1,$5D
	.dc.b	$FC,$47,$C3,$FC,$32,$D9,$A4,$AF,$DF,$4B,$B3,$FD,$FB,$18,$6B,$E1
	.dc.b	$11,$BE,$2E,$F3,$A9,$FF,$0A,$7E,$20,$FC,$F1,$F2,$BE,$9B,$6B,$A9
	.dc.b	$CD,$FF,$B8,$2F,$85,$EE,$B5,$F4,$36,$95,$A7,$85,$FF,$A1,$1B,$EC
	.dc.b	$6E,$75,$1F,$ED,$85,$FA,$AF,$F8,$06,$3F,$8F,$46,$3D,$CF,$9E,$D8
	.dc.b	$CC,$ED,$F8,$51,$CB,$E4,$3B,$49,$A6,$2A,$D7,$B6,$5F,$47,$C0,$CC
	.dc.b	$79,$7F,$7D,$FD,$53,$F0,$DA,$E7,$B3,$D9,$1F,$EC,$B8,$D4,$63,$F1
	.dc.b	$39,$18,$3E,$67,$EF,$8F,$9C,$F9,$5C,$5F,$C2,$BE,$BE,$F8,$E7,$E4
	.dc.b	$56,$EF,$FC,$9D,$5F,$61,$3C,$5C,$F1,$EA,$F9,$46,$0D,$B8,$7E,$95
	.dc.b	$75,$7E,$A7,$E4,$AF,$A3,$FA,$2D,$8F,$AA,$D7,$E8,$B5,$FD,$27,$12
	.dc.b	$1F,$13,$F8,$16,$CF,$E3,$5A,$F8,$2B,$F0,$DE,$1E,$33,$F7,$60,$8C
	.dc.b	$67,$F9,$C6,$DF,$E4,$9F,$F3,$E3,$F9,$95,$F5,$7F,$8A,$6C,$7E,$39
	.dc.b	$AF,$E2,$9A,$FA,$BC,$5F,$F1,$F3,$FA,$37,$F9,$03,$7C,$5F,$CE,$46
	.dc.b	$A3,$E9,$DF,$D6,$F9,$DF,$F3,$8C,$7D,$1C,$BF,$DB,$1F,$1B,$F8,$09
	.dc.b	$CC,$FE,$8F,$E3,$DF,$3D,$FA,$7F,$B5,$7D,$AF,$E8,$9B,$5F,$A6,$6B
	.dc.b	$FA,$26,$BF,$4B,$D5,$7E,$25,$FB,$15,$F2,$54,$C4,$7E,$D7,$F0,$C7
	.dc.b	$DA,$DF,$9D,$98,$7F,$03,$9C,$BF,$CF,$87,$6F,$F2,$4F,$43,$FD,$44
	.dc.b	$D8,$C9,$FC,$EC,$A7,$EE,$3F,$0B,$FD,$D7,$17,$DE,$6F,$1F,$F1,$86
	.dc.b	$CF,$F8,$AF,$D6,$FF,$0B,$3F,$D3,$BD,$9F,$93,$7F,$A3,$C5,$1F,$E8
	.dc.b	$A9,$D7,$C9,$7F,$17,$9C,$F3,$5F,$C8,$9C,$FE,$96,$DF,$F5,$16,$BF
	.dc.b	$CB,$E9,$18,$8F,$CF,$FC,$47,$D2,$CB,$F2,$ED,$D7,$F9,$8B,$67,$FC
	.dc.b	$89,$FA,$3D,$EE,$7E,$33,$FC,$FE,$7F,$98,$35,$FE,$4C,$DC,$3D,$DF
	.dc.b	$4E,$FB,$2F,$67,$E0,$EF,$86,$FA,$4D,$C3,$FB,$53,$5F,$53,$5F,$EB
	.dc.b	$5B,$3E,$63,$A4,$7F,$7A,$5F,$69,$F2,$5B,$19,$C3,$E6,$7C,$DF,$E1
	.dc.b	$23,$F4,$27,$F9,$85,$75,$68,$DF,$5F,$D3,$FC,$D1,$F5,$3F,$89,$6D
	.dc.b	$FF,$80,$5A,$FD,$16,$BF,$C3,$B9,$DF,$84,$FF,$4D,$6C,$7F,$59,$6B
	.dc.b	$E4,$D5,$F4,$9F,$CD,$E3,$FA,$97,$E0,$9F,$C2,$1B,$F3,$5F,$D5,$E3
	.dc.b	$1C,$9F,$8D,$F8,$7E,$6F,$E3,$FC,$E7,$DE,$FE,$85,$B8,$7F,$86,$5A
	.dc.b	$FE,$25,$AF,$D0,$FD,$AB,$C5,$5F,$99,$7F,$A6,$B7,$D7,$FC,$C4,$6B
	.dc.b	$BB,$FC,$21,$F9,$1F,$F5,$8B,$FE,$78,$1C,$3F,$48,$C7,$B5,$FD,$DD
	.dc.b	$D7,$B9,$7A,$7F,$35,$1F,$E1,$B6,$A3,$F0,$9E,$1F,$F8,$76,$33,$1F
	.dc.b	$87,$F3,$79,$11,$F1,$C7,$EA,$34,$FD,$A7,$E2,$0F,$47,$F6,$4D,$AF
	.dc.b	$F2,$3B,$5F,$AC,$FF,$B5,$5C,$C7,$0E,$A9,$18,$3F,$C9,$FF,$39,$FC
	.dc.b	$A1,$BF,$59,$FB,$F8,$37,$D3,$C3,$9F,$AD,$95,$C0,$F4,$FE,$1D,$F1
	.dc.b	$77,$91,$83,$DA,$E1,$F9,$DA,$FF,$16,$7F,$55,$3F,$37,$82,$47,$C4
	.dc.b	$78,$6D,$FC,$5F,$F3,$CF,$74,$FE,$55,$B1,$9F,$A1,$FE,$31,$0F,$95
	.dc.b	$BC,$18,$E0,$98,$7F,$9C,$2B,$E5,$61,$FC,$8F,$F3,$4F,$EF,$CD,$F9
	.dc.b	$0F,$E4,$51,$FE,$5E,$AF,$F3,$BB,$AE,$7F,$C0,$78,$08,$FF,$3D,$4E
	.dc.b	$BE,$3F,$89,$CA,$FF,$19,$F8,$47,$EF,$A7,$F2,$6D,$7F,$24,$DC,$7F
	.dc.b	$C3,$10,$FC,$27,$FC,$9E,$FC,$AF,$A2,$FC,$71,$1F,$F3,$7F,$1A,$3F
	.dc.b	$25,$A7,$E6,$49,$3F,$90,$8F,$E7,$DF,$84,$BE,$B6,$FA,$F2,$BF,$D2
	.dc.b	$B3,$96,$C7,$0F,$22,$46,$0F,$F4,$DF,$E7,$CF,$A9,$FE,$94,$D8,$CC
	.dc.b	$63,$E9,$73,$F0,$BA,$F9,$99,$5C,$A3,$F6,$5F,$12,$FB,$EA,$49,$FD
	.dc.b	$DF,$C4,$FD,$37,$F6,$5F,$85,$7E,$75,$FF,$40,$07,$0F,$FA,$0D,$AF
	.dc.b	$FA,$55,$AF,$2F,$0B,$F8,$14,$1A,$63,$C4,$FE,$CD,$FA,$1F,$FA,$37
	.dc.b	$9F,$FB,$63,$79,$7F,$0A,$DC,$1D,$88,$FE,$E3,$1F,$EC,$2F,$CF,$FF
	.dc.b	$A5,$BF,$CC,$5F,$E9,$2D,$F6,$BF,$BA,$46,$B0,$FE,$F5,$F7,$97,$1F
	.dc.b	$C9,$B6,$FB,$B7,$17,$E8,$DE,$35,$1D,$AF,$9F,$C7,$7F,$80,$60,$DC
	.dc.b	$BF,$C0,$9F,$90,$F9,$31,$FE,$C6,$D7,$F4,$6D,$7D,$ED,$C0,$FD,$B8
	.dc.b	$FF,$70,$B5,$FE,$A0,$DC,$3F,$C2,$CF,$97,$2F,$D7,$FE,$61,$F9,$3A
	.dc.b	$FE,$AC,$DA,$FF,$0E,$36,$77,$F8,$5F,$F8,$6E,$31,$5F,$49,$FF,$11
	.dc.b	$CF,$4E,$AD,$76,$FF,$88,$7F,$1E,$FA,$B7,$FE,$1B,$6B,$FE,$02,$6B
	.dc.b	$FC,$09,$AF,$DE,$D3,$E8,$13,$E0,$7A,$7C,$57,$F8,$B7,$F5,$77,$6F
	.dc.b	$F9,$0D,$B7,$FE,$FB,$6B,$FE,$2B,$6B,$FE,$12,$7F,$CD,$AF,$F7,$76
	.dc.b	$ED,$FE,$3D,$6E,$9F,$C5,$9E,$71,$FE,$41,$FD,$4F,$E1,$FC,$4B,$F2
	.dc.b	$EF,$E5,$18,$7F,$B6,$DA,$FF,$80,$E7,$71,$8A,$FF,$47,$B7,$0F,$F7
	.dc.b	$4B,$6B,$F3,$F5,$F4,$3E,$39,$F8,$EB,$8F,$F6,$B4,$9C,$C6,$1B,$1F
	.dc.b	$BA,$6B,$FE,$FB,$79,$F1,$7F,$BB,$5B,$1F,$EB,$96,$BF,$E2,$E7,$8E
	.dc.b	$5F,$E6,$1F,$8A,$FF,$49,$37,$C3,$FF,$30,$46,$34,$FE,$7A,$FC,$AB
	.dc.b	$FD,$CB,$FE,$23,$7F,$C8,$6D,$77,$F8,$3F,$F9,$F6,$31,$1F,$C0,$3F
	.dc.b	$42,$7D,$5F,$F9,$B5,$B1,$FE,$CF,$EA,$46,$23,$FC,$65,$5F,$E1,$B7
	.dc.b	$C3,$EC,$7F,$83,$D7,$D5,$7F,$AD,$AF,$FC,$B9,$39,$8C,$52,$31,$F3
	.dc.b	$7E,$CF,$83,$B8,$74,$3F,$C3,$57,$C3,$E5,$B3,$8F,$F9,$9D,$B8,$F3
	.dc.b	$AD,$E1,$F1,$A7,$9F,$03,$F1,$CF,$D4,$E3,$FE,$CA,$8C,$75,$72,$AD
	.dc.b	$D6,$AF,$FC,$55,$5A,$C3,$A9,$F8,$59,$D3,$F4,$AB,$CB,$B7,$F1,$68
	.dc.b	$DE,$7F,$3B,$C1,$C9,$AE,$CB,$3A,$AF,$E5,$BA,$7C,$96,$E7,$FD,$F1
	.dc.b	$E9,$3F,$3A,$31,$DB,$E0,$24,$73,$C3,$E5,$AF,$C4,$BF,$8B,$61,$FE
	.dc.b	$E8,$6D,$FD,$96,$CF,$F9,$25,$FD,$23,$F0,$0E,$5E,$07,$48,$74,$7F
	.dc.b	$4A,$9F,$21,$C5,$C5,$F7,$AE,$11,$B9,$FC,$41,$E5,$70,$F8,$DA,$F3
	.dc.b	$D3,$F6,$B6,$F4,$3F,$8E,$B7,$87,$31,$83,$3F,$7A,$E8,$F5,$56,$F0
	.dc.b	$F3,$9F,$16,$74,$E2,$35,$9F,$C7,$9C,$B6,$B9,$FA,$78,$D7,$E5,$55
	.dc.b	$C3,$FB,$6F,$E8,$7F,$E7,$3B,$FC,$1B,$63,$E3,$E8,$FA,$6F,$F5,$BD
	.dc.b	$6A,$79,$4E,$78,$6D,$F1,$CF,$F5,$D8,$70,$BE,$73,$9C,$BF,$C8,$73
	.dc.b	$9A,$DD,$F4,$8F,$F6,$96,$3F,$96,$75,$5F,$A0,$E5,$7B,$BE,$BE,$7F
	.dc.b	$D5,$3E,$4E,$B7,$D2,$E6,$B7,$8F,$E8,$27,$2F,$EA,$2F,$8B,$9A,$E1
	.dc.b	$7C,$FE,$97,$A4,$59,$D1,$CE,$31,$B7,$C8,$C1,$F8,$AE,$36,$FE,$21
	.dc.b	$F9,$E7,$CD,$F0,$BE,$DF,$12,$31,$97,$F9,$9B,$07,$BD,$F8,$0B,$DD
	.dc.b	$6B,$2F,$E4,$D3,$FE,$60,$78,$35,$CD,$FF,$8F,$30,$FE,$4F,$39,$7F
	.dc.b	$94,$FD,$77,$59,$F0,$36,$BF,$99,$7C,$05,$AD,$D7,$F4,$7F,$BC,$FF
	.dc.b	$72,$B7,$E5,$B8,$71,$27,$FC,$73,$39,$38,$57,$96,$74,$D9,$DE,$1F
	.dc.b	$D5,$DA,$FF,$39,$F3,$1E,$FB,$F1,$4F,$34,$61,$F0,$AE,$3C,$CE,$F2
	.dc.b	$FE,$BD,$39,$7F,$DA,$5C,$1F,$AC,$8E,$51,$AA,$E3,$7F,$07,$F5,$1E
	.dc.b	$18,$FF,$BF,$DB,$87,$DA,$7E,$BD,$FF,$55,$72,$FF,$A8,$1C,$CF,$F0
	.dc.b	$CC,$D1,$88,$FE,$AD,$18,$E0,$FA,$D3,$99,$FF,$54,$FC,$7F,$F9,$B6
	.dc.b	$FF,$B1,$36,$18,$7E,$32,$3F,$DA,$B3,$AF,$6F,$ED,$46,$2B,$C4,$D7
	.dc.b	$C2,$D7,$BF,$C6,$FE,$1F,$ED,$5C,$7F,$83,$7D,$78,$FE,$69,$CE,$3F
	.dc.b	$DD,$5C,$BF,$EC,$07,$37,$E6,$8C,$57,$3E,$4E,$8F,$E3,$FD,$F7,$D8
	.dc.b	$D3,$1F,$65,$6F,$FC,$05,$28,$FC,$E3,$63,$F4,$0D,$C3,$E8,$35,$F9
	.dc.b	$FF,$BD,$BE,$5F,$F3,$CB,$E6,$2C,$61,$CF,$CD,$CC,$3F,$C2,$54,$FA
	.dc.b	$DC,$61,$ED,$3F,$D3,$3F,$54,$BF,$D5,$3D,$EF,$FC,$34,$E5,$B8,$EA
	.dc.b	$E4,$6B,$C5,$B8,$75,$1F,$48,$F8,$BF,$DC,$46,$1B,$0F,$6A,$C6,$1F
	.dc.b	$E0,$D3,$C9,$B3,$BB,$FF,$27,$B6,$7F,$D8,$BF,$84,$BF,$01,$EE,$DA
	.dc.b	$FD,$59,$FE,$2F,$F9,$17,$DF,$EA,$BF,$85,$7E,$8B,$FE,$EF,$AF,$F1
	.dc.b	$D2,$F4,$73,$E9,$69,$DC,$C3,$FC,$81,$EF,$BE,$63,$1A,$7C,$FF,$0D
	.dc.b	$F8,$8E,$6F,$FC,$8D,$F1,$FF,$DB,$5F,$F2,$44,$66,$B8,$4E,$5E,$50
	.dc.b	$F9,$9F,$E4,$0E,$FC,$1F,$E6,$EE,$6F,$9E,$3F,$3B,$E9,$A4,$7C,$2C
	.dc.b	$07,$D6,$F8,$5C,$BF,$67,$93,$FC,$F9,$3F,$3A,$7E,$89,$1F,$F3,$FF
	.dc.b	$27,$E6,$72,$FF,$00,$DA,$31,$3F,$E5,$DF,$A7,$7C,$E7,$D4,$F0,$7F
	.dc.b	$9C,$7A,$BF,$E6,$AA,$FF,$34,$37,$0F,$90,$FF,$62,$6F,$C1,$70,$F3
	.dc.b	$B8,$FA,$F6,$F8,$BF,$D0,$FF,$44,$7D,$C5,$CB,$0D,$AF,$F3,$B4,$FF
	.dc.b	$31,$C7,$EB,$37,$6F,$EC,$2B,$EF,$7A,$7A,$5F,$9A,$FC,$1B,$D7,$87
	.dc.b	$CB,$C1,$18,$7F,$44,$FD,$36,$EF,$C7,$9E,$57,$1A,$FF,$00,$D7,$09
	.dc.b	$ED,$EE,$6A,$39,$B7,$3F,$EE,$BF,$88,$3F,$11,$FE,$C4,$B9,$2C,$61
	.dc.b	$CF,$5E,$33,$FF,$4B,$52,$31,$8F,$E5,$FF,$64,$7D,$2F,$F6,$D7,$FD
	.dc.b	$21,$7F,$E9,$D7,$2D,$F8,$FD,$4C,$8D,$7D,$2D,$BF,$87,$E3,$7C,$93
	.dc.b	$E5,$78,$D6,$BE,$13,$FC,$03,$87,$F7,$15,$FE,$5E,$3E,$6C,$64,$FE
	.dc.b	$0B,$F8,$7B,$EF,$BF,$86,$B6,$FE,$A2,$F0,$73,$3F,$E0,$7E,$AF,$CE
	.dc.b	$E1,$FB,$57,$83,$7C,$0D,$4F,$F8,$37,$F1,$4F,$E0,$93,$FD,$95,$AF
	.dc.b	$F6,$46,$BF,$AC,$F1,$51,$88,$7B,$DF,$CB,$F8,$7F,$C0,$BE,$97,$CC
	.dc.b	$9F,$E0,$DD,$EF,$93,$CD,$7F,$82,$BF,$1F,$FE,$3D,$D7,$FE,$03,$6F
	.dc.b	$0F,$7F,$C4,$DF,$23,$BF,$47,$FC,$49,$E1,$7F,$A3,$AB,$FC,$5C,$DD
	.dc.b	$3F,$BF,$35,$FE,$FA,$D7,$D5,$F9,$07,$E1,$67,$FD,$B8,$FE,$3D,$BD
	.dc.b	$BF,$E3,$18,$C4,$FF,$1F,$FD,$39,$EC,$7F,$8D,$57,$CA,$58,$C3,$FE
	.dc.b	$39,$F5,$FF,$A7,$57,$F8,$45,$AF,$F8,$41,$AF,$CD,$6B,$F3,$3C,$F7
	.dc.b	$CB,$7F,$8C,$1B,$9F,$FB,$CF,$E3,$9F,$7D,$9A,$FF,$1A,$FD,$D6,$B0
	.dc.b	$FE,$C9,$D2,$8C,$72,$FF,$98,$BF,$38,$7C,$FF,$0C,$3F,$CA,$1F,$93
	.dc.b	$7D,$7D,$3F,$CB,$4D,$CB,$FC,$54,$D7,$FC,$52,$D7,$F0,$DF,$1D,$FD
	.dc.b	$36,$7F,$9A,$77,$FF,$95,$7E,$6B,$FC,$B9,$39,$AF,$E9,$4F,$D9,$CF
	.dc.b	$FC,$89,$7E,$C3,$F9,$E6,$BF,$EB,$7D,$BF,$CC,$91,$FC,$CA,$FF,$22
	.dc.b	$ED,$FE,$A9,$4F,$C4,$FF,$CE,$0D,$9F,$F3,$A3,$67,$FC,$88,$D7,$FC
	.dc.b	$C1,$ED,$7F,$1F,$8F,$F2,$B7,$E9,$3F,$D5,$DF,$F2,$C4,$66,$BF,$A7
	.dc.b	$E2,$AD,$46,$3F,$0F,$FC,$2B,$EA,$7F,$D5,$B8,$7D,$47,$2D,$F9,$EF
	.dc.b	$F3,$6F,$E9,$4F,$B7,$FF,$46,$62,$8C,$61,$FC,$26,$31,$3A,$9C,$DF
	.dc.b	$FB,$AB,$37,$E0,$9D,$9F,$F4,$E3,$70,$F4,$FC,$EF,$F0,$29,$CF,$5F
	.dc.b	$F9,$FB,$F3,$8F,$F2,$5E,$1F,$BE,$7F,$CD,$AD,$F9,$7D,$7C,$07,$FA
	.dc.b	$13,$F3,$67,$CB,$7A,$1B,$9F,$9D,$B7,$FE,$FE,$DD,$18,$8F,$F4,$83
	.dc.b	$F9,$49,$FC,$0B,$63,$F3,$6B,$EC,$FC,$1D,$9F,$E9,$4F,$B2,$C5,$FE
	.dc.b	$21,$B7,$FA,$27,$FC,$71,$19,$AE,$1F,$1D,$F3,$A7,$35,$E5,$6B,$FE
	.dc.b	$A3,$6B,$FE,$9A,$FA,$BF,$E3,$DC,$1F,$EC,$76,$E1,$F8,$4B,$FF,$2B
	.dc.b	$3F,$E1,$59,$4E,$AB,$F9,$4D,$69,$FF,$2D,$4E,$BC,$EF,$F5,$7C,$E6
	.dc.b	$3B,$47,$FA,$CB,$F4,$CF,$3C,$7F,$AA,$DF,$7B,$67,$59,$6D,$98,$FE
	.dc.b	$B2,$BF,$97,$E1,$FC,$E9,$E7,$1F,$9C,$FA,$4F,$F4,$4E,$1F,$EB,$77
	.dc.b	$F0,$7B,$1C,$CE,$F8,$7F,$D8,$B1,$8C,$3F,$A2,$B6,$B8,$7C,$CE,$BE
	.dc.b	$67,$6F,$E7,$A4,$D2,$61,$F9,$FF,$8B,$FE,$11,$97,$FB,$25,$7D,$37
	.dc.b	$FB,$51,$CC,$FE,$D1,$CB,$FC,$29,$CE,$5F,$D6,$1B,$91,$FD,$14,$F1
	.dc.b	$F9,$1F,$F3,$ED,$7F,$93,$3E,$93,$FB,$C3,$FE,$4E,$0E,$1D,$15,$DF
	.dc.b	$E1,$DF,$EB,$AF,$C3,$EE,$C6,$92,$BF,$C0,$93,$F0,$B4,$E3,$AF,$FB
	.dc.b	$3A,$72,$D7,$C4,$27,$83,$FA,$DF,$C3,$FF,$C8,$EF,$F2,$A6,$BF,$D1
	.dc.b	$1B,$5F,$EF,$37,$FD,$79,$1B,$F6,$FD,$D3,$FE,$F7,$FD,$23,$FD,$11
	.dc.b	$3F,$CD,$1A,$F3,$9D,$EB,$FC,$93,$EA,$7F,$C3,$53,$FC,$5D,$AF,$F0
	.dc.b	$16,$BF,$2F,$81,$18,$8E,$5F,$11,$D1,$FF,$07,$C2,$31,$3C,$1B,$1D
	.dc.b	$B1,$5F,$2B,$FE,$97,$E5,$66,$3F,$9C,$B6,$3A,$DF,$4F,$C4,$3F,$B8
	.dc.b	$C7,$F5,$26,$BF,$DB,$1B,$7D,$A3,$FC,$47,$1F,$E1,$F7,$87,$A1,$E6
	.dc.b	$FB,$67,$CC,$D3,$B4,$7B,$6D,$F8,$83,$C4,$C3,$70,$FE,$14,$D7,$FB
	.dc.b	$EF,$E2,$0F,$98,$F8,$6B,$E6,$69,$27,$8D,$FC,$8F,$BB,$FF,$75,$47
	.dc.b	$F4,$B6,$BF,$CD,$5A,$FD,$56,$BF,$D8,$1F,$F7,$73,$FD,$D9,$AF,$F8
	.dc.b	$41,$B8,$FC,$5E,$24,$76,$8C,$7D,$23,$DD,$FB,$6D,$BF,$F0,$0B,$5F
	.dc.b	$F0,$DB,$C7,$8F,$FB,$8F,$99,$FE,$EF,$9F,$EC,$8D,$7F,$A7,$35,$FC
	.dc.b	$73,$5E,$3E,$27,$E2,$BF,$32,$FA,$B1,$FE,$10,$6B,$FE,$2A,$9F,$F8
	.dc.b	$00,$FF,$85,$3E,$65,$F5,$3F,$92,$6C,$66,$3E,$E7,$A6,$F6,$BF,$2F
	.dc.b	$E8,$3D,$BF,$E1,$7C,$3F,$DA,$1B,$19,$C3,$FB,$F3,$5F,$ED,$1F,$88
	.dc.b	$7F,$90,$1F,$F1,$2B,$FC,$65,$BE,$47,$E9,$23,$51,$FE,$42,$F5,$7F
	.dc.b	$D5,$4F,$F8,$B1,$FF,$57,$37,$E6,$BF,$57,$1A,$8F,$F7,$F1,$FE,$A8
	.dc.b	$3F,$D3,$18,$7F,$BB,$BE,$5F,$F9,$1B,$63,$26,$CF,$E2,$AB,$EB,$5F
	.dc.b	$2D,$FB,$B8,$C3,$63,$33,$FE,$40,$FC,$43,$97,$47,$EB,$6D,$F6,$7F
	.dc.b	$D0,$2B,$FC,$9A,$FF,$12,$9D,$37,$9F,$FD,$30,$EE,$53,$CF,$F1,$0F
	.dc.b	$F3,$0C,$7B,$1A,$FA,$1A,$FF,$3F,$B4,$62,$3F,$D4,$9F,$0C,$FB,$1F
	.dc.b	$63,$63,$FC,$C2,$D7,$FC,$E7,$F8,$83,$F1,$1F,$85,$F1,$6F,$81,$A4
	.dc.b	$AF,$F3,$0F,$E2,$3F,$D9,$1F,$F3,$5F,$E2,$3F,$19,$F9,$AD,$7C,$CD
	.dc.b	$7F,$D7,$0D,$7F,$21,$F8,$2F,$77,$FC,$EA,$FF,$A6,$9B,$DF,$F5,$63
	.dc.b	$51,$F2,$38,$7F,$D8,$5F,$88,$FE,$17,$87,$EA,$3F,$8A,$9D,$1F,$9D
	.dc.b	$71,$1A,$FA,$5F,$C0,$7E,$23,$FE,$BC,$AF,$C5,$B5,$FC,$3B,$5F,$A0
	.dc.b	$D7,$FD,$A5,$F8,$8F,$F2,$D7,$BF,$22,$31,$7F,$87,$6E,$3F,$E7,$6F
	.dc.b	$03,$E9,$4F,$E2,$FF,$10,$FD,$63,$6F,$F2,$91,$8A,$FF,$3E,$FE,$23
	.dc.b	$FC,$69,$FF,$71,$7D,$A7,$F1,$EC,$BF,$08,$D7,$3F,$41,$FA,$48,$C4
	.dc.b	$7E,$FF,$F1,$1F,$1B,$7D,$3F,$D1,$47,$F4,$AB,$FF,$77,$3A,$97,$0F
	.dc.b	$FA,$0D,$1F,$38,$CF,$8F,$FD,$95,$A6,$B5,$FE,$A8,$BF,$11,$FE,$D0
	.dc.b	$E7,$0F,$F0,$D5,$FF,$9E,$23,$FC,$3A,$D7,$FC,$1A,$78,$BC,$7F,$F0
	.dc.b	$F4,$66,$3F,$C2,$35,$FD,$A5,$ED,$F4,$9E,$38,$C4,$F3,$C3,$F7,$5F
	.dc.b	$88,$BE,$BB,$F8,$13,$70,$FF,$63,$B5,$FB,$AB,$CD,$DC,$FF,$09,$FD
	.dc.b	$93,$F5,$55,$FE,$CE,$6B,$C2,$3F,$87,$7E,$24,$FB,$EB,$E8,$4A,$FF
	.dc.b	$6A,$CE,$5B,$19,$7F,$1B,$1F,$CF,$1B,$3F,$3F,$3F,$E2,$BB,$FC,$AF
	.dc.b	$2F,$06,$7E,$3F,$5E,$FF,$C8,$26,$1F,$39,$FF,$24,$7D,$8B,$E4,$BE
	.dc.b	$D7,$AF,$FD,$76,$BE,$FB,$54,$ED,$FF,$76,$FE,$25,$F1,$5B,$BF,$F2
	.dc.b	$D8,$F9,$5C,$7F,$CA,$5D,$4F,$EA,$32,$FD,$D6,$9F,$8C,$EE,$46,$27
	.dc.b	$FA,$B3,$61,$5D,$2B,$87,$E2,$4F,$91,$FE,$78,$BE,$36,$98,$8F,$E8
	.dc.b	$1D,$BF,$E4,$EC,$BF,$72,$DC,$FF,$9F,$37,$0E,$B7,$F3,$FE,$E6,$5F
	.dc.b	$E6,$4F,$D1,$F1,$A3,$F4,$5F,$89,$7A,$67,$F5,$37,$F9,$82,$3F,$9E
	.dc.b	$7F,$CF,$1D,$7F,$DD,$63,$A1,$FA,$7F,$13,$F3,$47,$DF,$9F,$EB,$1F
	.dc.b	$81,$FF,$4B,$D3,$FA,$A3,$C6,$BF,$89,$B6,$33,$F2,$7C,$2E,$57,$A7
	.dc.b	$85,$7B,$C3,$E9,$3D,$5F,$A9,$F8,$97,$F8,$42,$BF,$8F,$BF,$E1,$DC
	.dc.b	$7F,$B2,$36,$77,$F0,$1F,$DB,$7E,$55,$F4,$14,$98,$78,$6F,$83,$FC
	.dc.b	$33,$F1,$27,$D5,$7F,$0B,$6C,$7C,$46,$BF,$C4,$7E,$1B,$F2,$4F,$F1
	.dc.b	$38,$FD,$33,$FE,$94,$6B,$FE,$A7,$6B,$DF,$C2,$FF,$54,$76,$FF,$96
	.dc.b	$63,$E8,$B5,$FE,$3C,$DD,$7F,$8D,$D7,$EC,$DF,$F0,$27,$C6,$BE,$83
	.dc.b	$FC,$18,$FF,$58,$8C,$FA,$F8,$9F,$F0,$87,$DA,$FF,$06,$9F,$F0,$63
	.dc.b	$78,$7F,$44,$DC,$BF,$C6,$D8,$73,$9F,$D3,$7E,$21,$67,$FC,$74,$FF
	.dc.b	$69,$7F,$A9,$B6,$FF,$C3,$0E,$F9,$3F,$7D,$E8,$7F,$88,$7C,$4F,$E9
	.dc.b	$97,$FE,$43,$6C,$AB,$96,$1F,$B5,$FC,$43,$E7,$DF,$F8,$4D,$AE,$6F
	.dc.b	$FC,$67,$18,$8F,$F2,$73,$C9,$BE,$C7,$F8,$C6,$31,$81,$FB,$79,$7F
	.dc.b	$8B,$FF,$10,$FC,$A5,$FF,$95,$DB,$46,$9F,$DE,$67,$87,$57,$F8,$F7
	.dc.b	$F1,$4F,$E3,$54,$6D,$72,$E1,$FB,$18,$7F,$93,$1A,$FF,$92,$1A,$FF
	.dc.b	$8C,$5A,$FF,$17,$FA,$1F,$F7,$0E,$3F,$E4,$56,$EB,$99,$FF,$27,$57
	.dc.b	$4A,$FF,$25,$7E,$23,$FA,$B9,$FF,$1E,$35,$FE,$44,$D7,$FC,$EC,$DC
	.dc.b	$8B,$E8,$BF,$CB,$4D,$9F,$DE,$35,$FF,$0A,$BF,$B9,$AF,$F5,$17,$E2
	.dc.b	$3F,$B4,$9F,$E3,$5F,$1E,$8C,$47,$F8,$AA,$31,$DB,$FE,$67,$FC,$47
	.dc.b	$E7,$69,$FC,$DD,$FA,$64,$E0,$FF,$71,$E1,$CB,$1F,$F6,$A7,$C5,$BF
	.dc.b	$11,$FE,$2D,$9F,$F8,$28,$FF,$3A,$38,$7F,$C1,$B1,$BB,$FF,$00,$C6
	.dc.b	$9B,$1C,$FE,$DB,$FD,$E8,$D9,$FE,$68,$D7,$FC,$86,$F3,$7F,$90,$7E
	.dc.b	$23,$FE,$AC,$C3,$FC,$F8,$DC,$FF,$A9,$36,$7F,$D4,$CF,$F1,$6E,$07
	.dc.b	$E2,$1F,$E9,$F9,$FF,$46,$B5,$FF,$3B,$35,$FE,$B8,$D7,$FD,$25,$F6
	.dc.b	$4F,$80,$F4,$B6,$3D,$2D,$7F,$C9,$B3,$DF,$1F,$3F,$A7,$F2,$7E,$9F
	.dc.b	$FD,$AD,$8B,$27,$A3,$9F,$98,$F5,$B7,$2F,$5E,$16,$F8,$4F,$D3,$D7
	.dc.b	$FD,$1D,$87,$2B,$FF,$1F,$C6,$27,$51,$D0,$C9,$F6,$3E,$CF,$FD,$4A
	.dc.b	$FF,$91,$B2,$FF,$75,$6E,$9C,$CA,$73,$18,$7D,$78,$78,$67,$59,$F5
	.dc.b	$F9,$2F,$56,$0F,$B1,$7C,$DE,$1B,$7E,$1B,$E0,$D3,$FD,$C1,$F3,$5F
	.dc.b	$C9,$F2,$F9,$2D,$9F,$F0,$93,$67,$7E,$47,$C7,$8F,$F4,$9D,$70,$EE
	.dc.b	$C1,$9F,$0E,$DF,$7C,$DE,$97,$F9,$A6,$3D,$67,$FB,$4D,$C5,$F3,$8F
	.dc.b	$30,$FC,$47,$FD,$29,$0C,$FF,$51,$18,$6F,$4F,$DD,$AE,$1F,$7D,$6B
	.dc.b	$51,$FE,$78,$5D,$7F,$EF,$DC,$3A,$4E,$63,$18,$74,$8F,$1D,$70,$E6
	.dc.b	$FD,$9B,$63,$FC,$EA,$D7,$F7,$87,$A6,$93,$C3,$2F,$C0,$CF,$A9,$AF
	.dc.b	$95,$AF,$C1,$6D,$3E,$B1,$CF,$E4,$9F,$C7,$E5,$EA,$BF,$91,$5A,$7E
	.dc.b	$79,$CF,$93,$D7,$19,$9E,$18,$75,$C3,$DF,$EC,$67,$4F,$9F,$77,$1F
	.dc.b	$5F,$A9,$FA,$0C,$7F,$3B,$F6,$EF,$CA,$B5,$B2,$B5,$8F,$F1,$CA,$D6
	.dc.b	$7F,$93,$AD,$63,$C2,$F6,$FC,$8F,$A8,$FF,$7A,$56,$E7,$34,$9C,$D7
	.dc.b	$E3,$23,$1C,$BE,$29,$FF,$0A,$5F,$77,$79,$FE,$C3,$E0,$7D,$9C,$AF
	.dc.b	$CF,$3E,$BD,$FE,$67,$B7,$FD,$4C,$FE,$6E,$33,$9F,$E6,$AB,$F7,$97
	.dc.b	$CA,$F7,$3E,$3F,$05,$F4,$B9,$AF,$8B,$18,$C7,$F8,$4D,$7E,$1A,$30
	.dc.b	$FF,$B2,$35,$FE,$5F,$7F,$6E,$BF,$96,$B6,$7F,$96,$F7,$74,$8E,$AD
	.dc.b	$9F,$D9,$7C,$CB,$E8,$7F,$09,$E2,$7F,$54,$8C,$4F,$F9,$8F,$F1,$3F
	.dc.b	$E3,$73,$9A,$DC,$FF,$A5,$67,$CF,$8D,$3F,$DB,$A5,$8F,$F6,$BE,$49
	.dc.b	$DC,$FB,$7E,$13,$9B,$6B,$FC,$E5,$A3,$EB,$F3,$87,$F4,$EB,$DF,$03
	.dc.b	$EE,$1A,$E1,$B8,$FF,$60,$5F,$41,$99,$D3,$E6,$CD,$7C,$E3,$0D,$BF
	.dc.b	$F7,$7B,$67,$F7,$6B,$D0,$7D,$3F,$91,$B5,$FE,$0A,$F8,$47,$DE,$66
	.dc.b	$BD,$F5,$FE,$8E,$8B,$3A,$AF,$EE,$F3,$FC,$E9,$E5,$3F,$0B,$F1,$3E
	.dc.b	$B8,$B8,$FF,$79,$BA,$73,$18,$79,$4E,$EF,$FD,$99,$1A,$6F,$27,$C3
	.dc.b	$BE,$4D,$9E,$35,$FE,$15,$9C,$D7,$DC,$F9,$6F,$F4,$AB,$FE,$A9,$6C
	.dc.b	$F7,$3F,$8F,$AF,$83,$FB,$6E,$B1,$F4,$B9,$AE,$93,$E3,$8C,$57,$F2
	.dc.b	$AB,$46,$23,$C9,$F8,$9F,$FB,$7E,$FF,$D7,$5F,$9A,$7E,$6A,$7C,$56
	.dc.b	$AD,$46,$27,$58,$FC,$A9,$CB,$79,$3F,$23,$1F,$EE,$9C,$3F,$D8,$CE
	.dc.b	$9F,$F7,$B7,$C2,$BE,$A3,$F2,$ED,$8F,$F6,$62,$FC,$53,$33,$A7,$E8
	.dc.b	$46,$67,$F4,$ED,$CB,$E0,$35,$FA,$5F,$BF,$3F,$07,$FD,$2B,$6B,$FD
	.dc.b	$B5,$F8,$A3,$F1,$5C,$D7,$EA,$A3,$FC,$5D,$3A,$9C,$7B,$79,$7F,$5F
	.dc.b	$F1,$FF,$C0,$25,$8F,$F5,$88,$C3,$FD,$6B,$F0,$3E,$6F,$CE,$8C,$F1
	.dc.b	$25,$F1,$5F,$BA,$77,$C5,$F9,$87,$38,$7F,$BC,$5C,$BF,$E6,$3D,$BF
	.dc.b	$77,$8F,$FB,$43,$BB,$F8,$3A,$FA,$A6,$78,$AD,$E3,$FD,$7E,$B5,$3F
	.dc.b	$BF,$6E,$1F,$E6,$AB,$E2,$D9,$D4,$3F,$05,$EE,$D9,$F8,$77,$D7,$F1
	.dc.b	$1F,$BD,$87,$E7,$CF,$25,$FF,$18,$F4,$5F,$7F,$4F,$E3,$7F,$89,$FF
	.dc.b	$6D,$AF,$CA,$CE,$B1,$EA,$7D,$0E,$05,$3D,$5E,$3E,$AF,$26,$D7,$0C
	.dc.b	$3F,$C1,$BF,$29,$FE,$04,$8C,$61,$FB,$66,$FE,$26,$F8,$9F,$E5,$B5
	.dc.b	$BB,$FD,$3C,$E6,$B8,$CE,$77,$FE,$1F,$39,$AF,$D6,$CE,$6B,$FD,$35
	.dc.b	$F2,$0F,$7F,$EB,$B6,$3C,$D5,$F6,$7F,$0D,$CD,$7C,$E7,$E8,$92,$7E
	.dc.b	$B3,$F5,$3E,$03,$E9,$9F,$82,$E3,$FD,$05,$FD,$1E,$44,$FF,$48,$FC
	.dc.b	$B5,$F1,$14,$FE,$97,$E6,$BE,$13,$19,$FC,$68,$FE,$DF,$C1,$FC,$B1
	.dc.b	$D4,$67,$D6,$E3,$5F,$D5,$3F,$4C,$FB,$1A,$7F,$0E,$6E,$1F,$CD,$3F
	.dc.b	$15,$F2,$BF,$CD,$A3,$38,$7D,$D9,$7E,$01,$FD,$79,$DC,$66,$31,$E2
	.dc.b	$FE,$83,$F3,$BF,$95,$8F,$E3,$F9,$3F,$CD,$77,$F3,$27,$39,$7E,$86
	.dc.b	$73,$5F,$95,$FC,$49,$F8,$AF,$E8,$DB,$87,$FB,$31,$B3,$FE,$58,$0E
	.dc.b	$1F,$F2,$DA,$FA,$BF,$D9,$B6,$7E,$EB,$FE,$20,$C3,$F6,$4E,$6B,$F9
	.dc.b	$9F,$D3,$3E,$D3,$FB,$EB,$6F,$F2,$EF,$F3,$47,$F4,$4E,$67,$8F,$BF
	.dc.b	$C2,$B8,$B6,$BF,$3F,$F8,$AF,$E0,$1F,$D0,$FE,$CC,$FC,$5F,$FC,$17
	.dc.b	$39,$C3,$FA,$2C,$E6,$B8,$F3,$E8,$DF,$D6,$58,$FC,$CF,$67,$E8,$76
	.dc.b	$FE,$05,$E0,$FC,$2F,$C5,$FF,$28,$DA,$F9,$91,$F3,$C8,$FC,$FF,$53
	.dc.b	$E4,$9F,$C7,$3E,$8D,$F6,$58,$9F,$DC,$F4,$61,$BF,$5D,$FE,$24,$8C
	.dc.b	$6F,$FD,$79,$47,$F8,$A3,$F1,$7F,$E1,$D3,$FD,$E5,$AF,$F0,$D6,$BF
	.dc.b	$E9,$46,$F9,$1F,$94,$FC,$1B,$FC,$63,$DA,$7F,$C3,$3E,$A7,$C9,$E0
	.dc.b	$FF,$08,$CE,$73,$FE,$1F,$F8,$AF,$EC,$9B,$E3,$7F,$88,$3F,$16,$FE
	.dc.b	$55,$A7,$F7,$4A,$D3,$FD,$DB,$E0,$FF,$3A,$FF,$83,$DA,$FF,$23,$6B
	.dc.b	$FE,$AA,$6E,$8F,$C5,$BF,$11,$3F,$EA,$E7,$F1,$19,$FF,$89,$9C,$CE
	.dc.b	$FB,$7F,$C9,$BF,$28,$F9,$AB,$F2,$98,$6E,$DF,$CA,$BE,$5B,$FA,$FE
	.dc.b	$7F,$E2,$36,$BF,$CD,$DA,$FF,$AF,$9B,$8F,$FB,$FF,$DC,$FF,$2D,$FC
	.dc.b	$97,$F6,$5B,$FE,$71,$F1,$5F,$D6,$FE,$7B,$35,$FE,$60,$F9,$FF,$F0
	.dc.b	$FD,$7F,$3A,$6E,$5F,$E2,$DB,$7E,$04,$C3,$87,$C5,$7F,$9A,$3F,$16
	.dc.b	$FA,$97,$FE,$29,$6B,$FE,$3B,$6B,$FD,$3D,$AF,$FB,$5B,$E4,$BF,$DC
	.dc.b	$B9,$7F,$4A,$7F,$86,$BF,$D5,$5C,$C7,$F8,$B9,$CB,$FE,$A8,$F8,$9F
	.dc.b	$F3,$CC,$FF,$86,$F9,$7F,$DA,$59,$FF,$B6,$3F,$0C,$CD,$7F,$9F,$A7
	.dc.b	$F6,$57,$FE,$3F,$6B,$FE,$51,$6B,$FD,$95,$AF,$F9,$CF,$F1,$6F,$EA
	.dc.b	$4D,$F7,$FF,$CD,$F1,$8C,$0F,$F6,$BC,$7E,$8F,$E2,$DF,$EC,$4B,$F9
	.dc.b	$9F,$8F,$FF,$87,$AD,$39,$C7,$D3,$39,$FC,$7F,$FD,$83,$7F,$E2,$63
	.dc.b	$FD,$08,$E1,$FF,$14,$DA,$31,$1A,$E5,$FF,$49,$4B,$A3,$FC,$09,$87
	.dc.b	$C6,$6E,$5F,$E1,$0F,$C2,$FF,$2C,$FE,$22,$33,$87,$FA,$4F,$F1,$7F
	.dc.b	$F3,$FD,$FC,$EF,$88,$F4,$5F,$8A,$72,$F9,$23,$35,$C7,$E4,$BC,$75
	.dc.b	$FE,$66,$8F,$66,$48,$C4,$73,$9F,$4F,$E2,$FF,$94,$C7,$EE,$B7,$1F
	.dc.b	$82,$FF,$60,$AF,$F5,$43,$97,$FD,$55,$ED,$E7,$13,$47,$D7,$7F,$AA
	.dc.b	$1C,$3F,$21,$CD,$FF,$57,$9E,$33,$93,$E6,$FC,$6F,$F9,$3A,$FF,$50
	.dc.b	$DC,$7F,$DC,$3F,$5F,$FC,$6F,$0F,$AF,$39,$7F,$C7,$F1,$9A,$FF,$58
	.dc.b	$37,$0F,$F2,$8F,$E3,$7F,$ED,$AC,$13,$9F,$C6,$3C,$B7,$FB,$7F,$C7
	.dc.b	$7E,$7E,$09,$CD,$FE,$92,$73,$18,$C7,$F1,$2D,$BF,$F6,$83,$FE,$58
	.dc.b	$7F,$81,$39,$B7,$D3,$FF,$A5,$67,$FD,$AD,$D7,$FD,$12,$FF,$87,$CE
	.dc.b	$69,$39,$AF,$E1,$93,$92,$7D,$F7,$FB,$61,$B7,$F3,$DF,$F5,$24,$FE
	.dc.b	$84,$FF,$72,$38,$7F,$12,$D7,$F4,$7F,$0F,$FE,$14,$CF,$F8,$1F,$E2
	.dc.b	$DF,$C2,$32,$FF,$1A,$7D,$67,$FB,$9A,$31,$5B,$C7,$8C,$FF,$BA,$A3
	.dc.b	$F4,$35,$C2,$7F,$C9,$B8,$7F,$A6,$3E,$A7,$FC,$34,$FE,$D5,$AE,$4A
	.dc.b	$EB,$F1,$86,$8F,$C6,$70,$DA,$FF,$12,$4E,$A7,$27,$1F,$A7,$FF,$7A
	.dc.b	$4E,$9A,$FC,$59,$46,$23,$F8,$44,$63,$5F,$17,$AD,$FC,$2D,$FF,$4D
	.dc.b	$47,$FA,$AA,$7F,$CD,$2D,$7F,$D2,$1F,$6F,$F4,$67,$75,$C7,$3E,$BC
	.dc.b	$B6,$FF,$D7,$78,$23,$10,$8E,$97,$FC,$DB,$EB,$5F,$31,$F2,$B7,$46
	.dc.b	$2F,$F6,$0D,$7F,$05,$1A,$9E,$13,$FE,$99,$9F,$EB,$ED,$7F,$AD,$36
	.dc.b	$BC,$13,$E7,$AE,$FE,$35,$AF,$F5,$E5,$79,$72,$C4,$FB,$7E,$BF,$F8
	.dc.b	$3F,$67,$1F,$C6,$3F,$D2,$59,$FF,$6E,$6A,$C1,$87,$FB,$7A,$F8,$CE
	.dc.b	$FE,$63,$97,$D8,$C8,$C1,$FE,$AF,$97,$E2,$DF,$ED,$28,$FE,$F7,$F8
	.dc.b	$BF,$FB,$C5,$BF,$47,$E5,$9D,$47,$F7,$99,$D3,$7E,$93,$E0,$41,$B5
	.dc.b	$F5,$BF,$0E,$FF,$6A,$57,$E7,$1A,$FF,$9D,$DA,$FE,$CD,$AF,$F5,$57
	.dc.b	$FC,$74,$FF,$AC,$9A,$FF,$B6,$5B,$A7,$FB,$A5,$EA,$FF,$78,$FC,$0F
	.dc.b	$F9,$9C,$FF,$86,$1A,$FF,$84,$DB,$1F,$32,$19,$3F,$38,$7D,$A7,$DD
	.dc.b	$BF,$87,$5F,$D3,$9A,$FF,$68,$6A,$E4,$46,$21,$8B,$E4,$3F,$11,$EF
	.dc.b	$7F,$88,$E7,$FC,$5A,$D7,$DD,$1F,$A0,$E1,$7D,$7D,$24,$FD,$DF,$41
	.dc.b	$F2,$DB,$ED,$FE,$13,$7F,$B1,$AF,$E4,$0D,$7F,$C0,$4D,$5F,$90,$FE
	.dc.b	$3F,$F2,$BF,$D7,$FE,$54,$FC,$42,$9F,$B5,$9D,$57,$F9,$13,$27,$C9
	.dc.b	$FE,$8D,$B1,$98,$E1,$F8,$D7,$F0,$C8,$FF,$7F,$8E,$BF,$E4,$3E,$C2
	.dc.b	$30,$7F,$AF,$5C,$46,$A3,$27,$F3,$BC,$3D,$3E,$AF,$F0,$9A,$7D,$B7
	.dc.b	$FA,$AE,$BA,$75,$7D,$F9,$CF,$CE,$7A,$F0,$61,$FC,$3B,$F2,$F5,$52
	.dc.b	$61,$F8,$BF,$13,$F8,$9F,$1F,$E6,$23,$D0,$D4,$D3,$5D,$5F,$84,$F0
	.dc.b	$B8,$FC,$A4,$AF,$CB,$47,$A7,$EA,$1F,$4F,$49,$3E,$A9,$F8,$DA,$BE
	.dc.b	$C3,$73,$93,$D3,$F8,$C3,$EF,$FE,$6B,$63,$E6,$35,$FE,$C9,$B2,$31
	.dc.b	$1F,$E4,$7B,$7C,$77,$F3,$38,$ED,$4F,$C3,$7C,$31,$FC,$5F,$2F,$B5
	.dc.b	$39,$CF,$F1,$17,$C6,$B4,$7D,$FF,$C5,$71,$CB,$FB,$38,$46,$30,$FF
	.dc.b	$2C,$36,$3F,$2F,$9B,$EE,$7F,$A2,$FE,$56,$F8,$EF,$ED,$2D,$AF,$ED
	.dc.b	$AD,$7F,$C0,$41,$C3,$EF,$F6,$DF,$55,$D3,$F2,$19,$3F,$37,$85,$5A
	.dc.b	$8F,$87,$FA,$07,$F1,$4D,$7E,$0B,$63,$F1,$6D,$AF,$F7,$F7,$91,$F9
	.dc.b	$7F,$CB,$7F,$5D,$5F,$9F,$6B,$F5,$1A,$FF,$9E,$78,$D1,$88,$FC,$1F
	.dc.b	$CD,$7F,$02,$7E,$AF,$E3,$3F,$D2,$B8,$FF,$40,$DC,$F7,$E5,$7F,$6C
	.dc.b	$F8,$97,$E0,$D4,$98,$7E,$2C,$7E,$49,$FB,$C7,$EC,$9F,$81,$FA,$BF
	.dc.b	$F0,$BF,$59,$A4,$8F,$F0,$EC,$7F,$21,$7A,$35,$E6,$DC,$FF,$04,$F3
	.dc.b	$8F,$A1,$F4,$3F,$D3,$DF,$C8,$FE,$34,$FB,$8F,$E0,$0D,$8D,$FB,$58
	.dc.b	$9F,$F0,$4D,$BE,$A6,$99,$AF,$77,$E1,$FF,$DE,$38,$3F,$D4,$DA,$BE
	.dc.b	$6B,$FC,$22,$DA,$DC,$E2,$7A,$FD,$34,$9F,$91,$5B,$7F,$3F,$CD,$F9
	.dc.b	$7F,$9E,$FE,$2D,$FA,$FF,$F2,$39,$CB,$78,$7F,$8B,$78,$3F,$D0,$DF
	.dc.b	$E3,$11,$9C,$FF,$D7,$5B,$7F,$22,$9F,$8F,$E8,$F4,$F8,$AF,$7F,$7C
	.dc.b	$CB,$E7,$BF,$0A,$7F,$C5,$7F,$8B,$3E,$9A,$92,$BD,$3F,$59,$F2,$3D
	.dc.b	$19,$1F,$61,$D3,$EC,$FC,$5F,$F8,$93,$FC,$11,$AE,$7C,$EF,$E9,$1F
	.dc.b	$9A,$7E,$27,$5F,$DF,$7E,$2C,$FB,$2F,$F1,$EB,$6F,$F2,$0D,$79,$1F
	.dc.b	$93,$FD,$13,$FB,$BB,$7E,$23,$FC,$8B,$18,$8F,$CA,$F4,$3E,$BF,$AF
	.dc.b	$DC,$FF,$9B,$30,$34,$95,$FE,$72,$8F,$F0,$0B,$E2,$F8,$9F,$29,$C9
	.dc.b	$FE,$05,$E5,$7F,$89,$B4,$FE,$99,$E7,$79,$DB,$AF,$F6,$79,$D7,$4F
	.dc.b	$F2,$4A,$D3,$FE,$F3,$C7,$F7,$18,$7E,$B1,$B8,$72,$8F,$D9,$7E,$31
	.dc.b	$FE,$D0,$6F,$C8,$7F,$99,$23,$11,$FB,$4F,$C1,$7F,$C0,$EF,$CA,$F1
	.dc.b	$7E,$69,$1E,$27,$F3,$AF,$DF,$FF,$61,$97,$F3,$86,$C7,$F6,$26,$FC
	.dc.b	$13,$F2,$FF,$F3,$15,$7A,$7F,$1D,$F9,$84,$BF,$94,$FC,$FF,$C5,$3E
	.dc.b	$E9,$23,$EE,$C7,$EB,$3E,$2E,$47,$EA,$1F,$91,$EA,$FE,$13,$E2,$24
	.dc.b	$FE,$11,$DB,$FA,$CF,$C6,$BF,$A8,$F0,$FF,$7D,$6E,$5F,$D4,$9B,$3F
	.dc.b	$E1,$EF,$B9,$FD,$44,$FD,$1E,$9F,$C8,$92,$7F,$AE,$D7,$0F,$9D,$7E
	.dc.b	$5A,$F3,$BF,$D6,$F1,$86,$C6,$6B,$FA,$6C,$E5,$FF,$5E,$FC,$03,$ED
	.dc.b	$7F,$D3,$6D,$8F,$E5,$CD,$7F,$B2,$CF,$17,$FB,$3F,$BA,$D3,$F9,$D4
	.dc.b	$7F,$85,$BE,$AF,$F3,$6F,$F7,$46,$BC,$23,$7E,$92,$73,$BF,$F8,$8D
	.dc.b	$FC,$FD,$7F,$16,$6C,$67,$C1,$FF,$16,$7D,$6B,$E9,$3F,$C6,$ED,$BF
	.dc.b	$F1,$B3,$5F,$F0,$5B,$5F,$F1,$E1,$FE,$65,$FD,$F7,$F9,$E3,$FE,$02
	.dc.b	$F8,$C4,$63,$0F,$F3,$A4,$63,$DD,$F0,$E1,$FE,$C5,$D9,$F2,$2C,$BE
	.dc.b	$9B,$72,$FA,$2D,$7F,$E0,$04,$F0,$9F,$F7,$6F,$07,$F8,$FA,$7E,$8E
	.dc.b	$DF,$50,$ED,$7C,$73,$FF,$43,$70,$7D,$1E,$3F,$F2,$77,$E7,$9F,$BA
	.dc.b	$C3,$FC,$8E,$D9,$D4,$7F,$89,$A3,$58,$1F,$F0,$44,$7F,$44,$6D,$FF
	.dc.b	$95,$5E,$9D,$CF,$A6,$DD,$7C,$A8,$C7,$C3,$FF,$86,$A3,$FC,$BF,$3A
	.dc.b	$7F,$CD,$53,$A7,$FC,$D5,$F9,$4F,$F3,$1E,$BF,$F1,$AB,$63,$FC,$BE
	.dc.b	$D9,$E1,$1F,$DF,$BF,$44,$FE,$1B,$97,$F8,$D5,$B7,$F7,$1B,$3F,$E7
	.dc.b	$9F,$9B,$FF,$10,$2F,$C6,$9E,$5F,$9E,$30,$D8,$CC,$7A,$DA,$FF,$8D
	.dc.b	$9F,$EC,$57,$F8,$A6,$C7,$FA,$1D,$AF,$43,$FC,$30,$BE,$4F,$F2,$3A
	.dc.b	$3E,$EF,$94,$61,$B5,$FE,$91,$9D,$3F,$E9,$3E,$CF,$F0,$83,$F4,$E7
	.dc.b	$C5,$F0,$9C,$63,$C4,$D7,$E0,$37,$0F,$14,$75,$9E,$2D,$3D,$5F,$86
	.dc.b	$6C,$E1,$56,$34,$7C,$B8,$C3,$E3,$3C,$27,$3E,$37,$EF,$7F,$A1,$CB
	.dc.b	$D7,$18,$DB,$D0,$DA,$E1,$5B,$C7,$FC,$F2,$AB,$6D,$9D,$C6,$3A,$72
	.dc.b	$FC,$3E,$2F,$F4,$27,$8B,$7E,$07,$FC,$91,$9F,$8B,$1E,$58,$FA,$4F
	.dc.b	$C3,$7C,$F7,$C7,$23,$F0,$37,$FE,$58,$ED,$47,$CD,$F8,$0F,$DD,$5F
	.dc.b	$D8,$AF,$C8,$46,$35,$FC,$45,$6F,$6F,$C7,$4F,$F0,$1A,$FE,$2A,$70
	.dc.b	$8C,$9F,$93,$F8,$1F,$F0,$64,$7E,$45,$AF,$E4,$5A,$FE,$7D,$B8,$BE
	.dc.b	$A1,$CB,$EB,$9F,$F0,$CE,$7E,$7A,$D4,$E7,$DB,$F8,$17,$C3,$0F,$80
	.dc.b	$DB,$E1,$EF,$DD,$7F,$49,$36,$BF,$DB,$3F,$4F,$FC,$97,$1F,$C7,$71
	.dc.b	$7F,$A6,$3E,$CF,$5B,$2B,$57,$F8,$BA,$D5,$FE,$2E,$73,$9F,$E2,$9B
	.dc.b	$97,$FA,$F9,$FC,$34,$EF,$97,$AC,$63,$0F,$CB,$63,$DE,$73,$CB,$A7
	.dc.b	$D1,$3F,$C1,$EE,$EF,$F2,$1F,$D0,$FC,$CF,$F9,$1D,$FC,$21,$C1,$C4
	.dc.b	$FE,$16,$31,$EE,$72,$69,$C9,$E1,$E0,$62,$7F,$25,$5F,$EC,$0A,$EF
	.dc.b	$3E,$4C,$7F,$D3,$91,$83,$C7,$F0,$9F,$57,$C0,$FF,$6C,$5F,$F9,$AE
	.dc.b	$7F,$D5,$0B,$E4,$75,$1F,$D8,$A3,$58,$FF,$AA,$9D,$47,$F3,$B6,$C7
	.dc.b	$FB,$59,$AF,$E5,$9A,$FF,$31,$F2,$7A,$A1,$E1,$7F,$77,$AF,$F5,$D7
	.dc.b	$E8,$1F,$1B,$1F,$E7,$57,$C3,$81,$F4,$3F,$07,$1F,$F4,$BB,$6B,$F2
	.dc.b	$5C,$AF,$BB,$7C,$57,$1A,$D6,$7F,$D4,$27,$2F,$FB,$BA,$B8,$57,$FB
	.dc.b	$F7,$0D,$35,$46,$A7,$87,$9E,$F8,$0F,$6C,$A3,$F1,$99,$79,$1D,$57
	.dc.b	$C1,$8C,$71,$7C,$4F,$2F,$FC,$51,$E2,$F2,$BD,$E5,$FE,$25,$8C,$65
	.dc.b	$F2,$A1,$C5,$FE,$C1,$CB,$ED,$FC,$77,$CE,$F4,$3C,$FE,$3E,$EB,$57
	.dc.b	$FE,$38,$D9,$3E,$DA,$DE,$1F,$E4,$49,$CB,$67,$E1,$66,$F8,$97,$81
	.dc.b	$F7,$23,$0F,$D7,$F9,$9F,$9F,$97,$C3,$AD,$E1,$F0,$4F,$C1,$72,$AB
	.dc.b	$EE,$57,$F8,$CA,$31,$D7,$F4,$5B,$5F,$C8,$A5,$18,$7F,$92,$AF,$C5
	.dc.b	$7E,$4F,$27,$F1,$58,$C7,$27,$E2,$E7,$37,$E2,$7E,$BB,$F9,$5E,$4F
	.dc.b	$F2,$D7,$97,$BE,$4F,$AD,$EB,$FF,$BD,$6B,$D5,$9F,$8B,$0E,$6F,$E7
	.dc.b	$7F,$44,$7E,$1F,$F5,$DB,$5C,$39,$7F,$0D,$D7,$9A,$FB,$1C,$9F,$9E
	.dc.b	$AF,$F3,$AB,$70,$FF,$22,$B5,$FF,$2B,$DF,$29,$FF,$17,$78,$EF,$8D
	.dc.b	$FC,$33,$63,$CF,$C2,$FC,$57,$C7,$83,$8F,$F9,$D3,$FD,$0F,$0F,$D8
	.dc.b	$4E,$A7,$E5,$36,$FE,$36,$FF,$E1,$97,$AC,$66,$FF,$7D,$F4,$3E,$9B
	.dc.b	$EA,$DA,$FF,$37,$FC,$D7,$FA,$3D,$F9,$74,$9D,$3F,$C1,$E7,$4F,$9E
	.dc.b	$78,$FB,$DF,$A9,$F8,$C7,$E2,$BF,$A5,$BD,$E3,$E0,$FB,$CF,$D4,$64
	.dc.b	$6C,$E1,$B8,$FF,$95,$E5,$3F,$C0,$72,$F2,$36,$33,$3F,$93,$D9,$FB
	.dc.b	$E7,$F8,$02,$3E,$7B,$6F,$D0,$7F,$0A,$F8,$A7,$89,$FC,$76,$75,$9F
	.dc.b	$C6,$D3,$CE,$7E,$FF,$F0,$BF,$D9,$E1,$C8,$F8,$D1,$9F,$0A,$65,$F2
	.dc.b	$BE,$49,$F8,$5E,$2F,$C3,$AB,$E3,$DE,$CF,$C4,$F7,$FF,$97,$46,$31
	.dc.b	$FA,$7C,$3F,$E6,$47,$F3,$FE,$13,$F3,$9F,$1F,$07,$EB,$6F,$86,$7F
	.dc.b	$EC,$29,$CD,$FE,$FA,$73,$5F,$82,$F8,$6F,$A6,$7D,$4F,$8C,$B5,$E0
	.dc.b	$7E,$71,$1F,$A8,$FD,$3F,$C3,$3F,$84,$47,$E2,$1B,$8F,$C5,$F9,$3F
	.dc.b	$F0,$3E,$1F,$9C,$9C,$BF,$92,$F1,$FF,$06,$7F,$4D,$FD,$2F,$FA,$A6
	.dc.b	$1F,$02,$31,$84,$8F,$83,$F7,$1A,$8F,$97,$56,$FF,$3B,$F1,$3F,$8E
	.dc.b	$3E,$C7,$CD,$5C,$3F,$3B,$D2,$8C,$C7,$E8,$3D,$AF,$E4,$F8,$AB,$53
	.dc.b	$F4,$B9,$7F,$32,$7E,$6B,$E0,$EC,$7E,$F3,$EB,$5F,$7D,$98,$C3,$70
	.dc.b	$FC,$0F,$B9,$FD,$22,$72,$7F,$04,$71,$1F,$86,$AE,$73,$FB,$4F,$92
	.dc.b	$FD,$9C,$FE,$D7,$E1,$7F,$A9,$46,$3A,$91,$FD,$57,$91,$F8,$6E,$BD
	.dc.b	$0C,$37,$1F,$E2,$7F,$87,$FE,$E1,$FE,$03,$F8,$8F,$F8,$05,$FE,$8B
	.dc.b	$19,$CF,$F4,$13,$95,$DC,$FA,$CF,$F5,$45,$F4,$F1,$7F,$B6,$46,$2B
	.dc.b	$F4,$38,$73,$E8,$FE,$3F,$8B,$E5,$7F,$B6,$37,$1F,$1F,$CF,$62,$72
	.dc.b	$FF,$03,$93,$49,$3F,$ED,$FD,$BF,$C6,$78,$FE,$DD,$B8,$7E,$4F,$E9
	.dc.b	$3E,$37,$83,$FE,$01,$E8,$7E,$1D,$FE,$1F,$EC,$FE,$CF,$8F,$F0,$8E
	.dc.b	$1F,$DB,$CE,$7D,$FF,$F1,$27,$A7,$FC,$C6,$BF,$B0,$B5,$FE,$D4,$D7
	.dc.b	$FB,$7B,$5F,$ED,$3F,$A6,$7F,$6A,$DB,$FA,$2B,$63,$C1,$F9,$73,$F1
	.dc.b	$AF,$F1,$94,$E7,$7F,$F4,$06,$6F,$CE,$BD,$38,$FF,$3A,$F9,$2F,$F2
	.dc.b	$2D,$6A,$73,$97,$F1,$99,$CD,$7F,$90,$E1,$1F,$C6,$23,$FB,$F4,$63
	.dc.b	$CE,$FE,$65,$39,$8F,$F0,$13,$5F,$F5,$2A,$FB,$7F,$F1,$33,$6B,$F8
	.dc.b	$CF,$A8,$FA,$0C,$D7,$F8,$C7,$E2,$3F,$2F,$5F,$D9,$B2,$E1,$8F,$D1
	.dc.b	$9C,$BF,$CA,$A3,$39,$FF,$83,$7E,$4F,$FB,$5C,$FF,$86,$5A,$FF,$87
	.dc.b	$9A,$FF,$86,$1B,$E9,$BE,$0B,$C9,$8F,$F7,$66,$D7,$E0,$7B,$3F,$07
	.dc.b	$87,$F8,$6E,$72,$FF,$AC,$7E,$05,$F1,$D8,$C3,$E6,$7C,$FF,$F8,$CA
	.dc.b	$73,$BF,$F9,$26,$3F,$87,$9F,$D0,$E7,$E7,$B7,$2F,$F1,$73,$5F,$F6
	.dc.b	$8B,$5E,$BF,$19,$CC,$FF,$21,$61,$FC,$8E,$B2,$7F,$48,$F8,$EF,$4B
	.dc.b	$FE,$40,$1F,$55,$FE,$61,$8C,$DF,$E4,$E3,$FD,$01,$9B,$E8,$FF,$C8
	.dc.b	$4D,$BF,$EE,$0D,$AF,$F2,$0B,$5F,$C6,$68,$FC,$3B,$FC,$71,$1F,$DC
	.dc.b	$A3,$1E,$BF,$F9,$C7,$5F,$0C,$63,$0F,$EB,$FC,$2F,$B6,$FF,$38,$7E
	.dc.b	$8F,$FD,$73,$1F,$E2,$33,$9A,$FF,$0C,$CE,$64,$FF,$2C,$7F,$DF,$2D
	.dc.b	$7F,$40,$D7,$FD,$48,$D7,$FD,$1F,$F1,$DF,$E7,$FF,$92,$FF,$2A,$C7
	.dc.b	$E9,$BE,$5B,$B4,$62,$3F,$D2,$1E,$5F,$F9,$CF,$E7,$BF,$9C,$C7,$F8
	.dc.b	$1F,$4F,$E3,$8F,$C1,$8C,$F0,$7E,$27,$F4,$87,$D8,$7F,$9F,$9B,$AF
	.dc.b	$F9,$BF,$04,$63,$0F,$DC,$1F,$E9,$F7,$0F,$D0,$5F,$73,$FE,$85,$6D
	.dc.b	$7F,$9A,$3C,$17,$E1,$DF,$26,$FE,$A5,$69,$F4,$78,$DF,$E9,$FC,$7F
	.dc.b	$89,$61,$FE,$46,$7F,$2D,$9F,$F9,$61,$CD,$7B,$FE,$13,$FC,$71,$1F
	.dc.b	$97,$8F,$F6,$83,$5F,$F5,$B4,$7F,$A7,$3C,$A7,$B5,$FE,$9B,$5F,$7D
	.dc.b	$63,$0F,$FB,$6B,$5F,$F4,$77,$B3,$9C,$FD,$9F,$2B,$FD,$89,$FF,$42
	.dc.b	$E7,$FE,$A6,$C3,$FD,$4D,$FA,$D7,$F5,$3D,$7F,$70,$BF,$07,$CF,$8F
	.dc.b	$A7,$B7,$D7,$FC,$5A,$DC,$E3,$B7,$79,$76,$FC,$2F,$FD,$23,$99,$FF
	.dc.b	$10,$C7,$C0,$72,$63,$8B,$FD,$AC,$E3,$E6,$BF,$D7,$4F,$F9,$A2,$78
	.dc.b	$46,$3E,$57,$FD,$83,$3F,$E7,$2A,$FD,$6B,$5E,$D1,$FE,$67,$7F,$C1
	.dc.b	$79,$7F,$BA,$FC,$EF,$F6,$C4,$63,$0F,$C4,$C6,$3C,$E9,$1F,$E8,$2F
	.dc.b	$9A,$FF,$32,$BF,$E8,$4A,$E0,$7E,$43,$E9,$3F,$A2,$A0,$E1,$A6,$9F
	.dc.b	$E9,$7F,$18,$FC,$29,$F3,$98,$6D,$7F,$B3,$67,$53,$93,$8F,$D0,$7F
	.dc.b	$A9,$E7,$4D,$7F,$47,$28,$C4,$7F,$5D,$8C,$69,$E2,$E4,$F4,$C7,$E0
	.dc.b	$67,$04,$3C,$5C,$7A,$23,$EA,$7F,$B2,$F1,$3F,$09,$FF,$18,$CF,$ED
	.dc.b	$DA,$FE,$DA,$35,$D3,$C2,$73,$3F,$C9,$9F,$D3,$93,$6D,$1D,$BF,$1E
	.dc.b	$78,$F4,$95,$FC,$16,$32,$FF,$07,$FA,$7F,$F4,$6B,$7E,$03,$CF,$18
	.dc.b	$8F,$23,$5F,$E1,$7F,$6C,$FB,$CF,$BA,$D8,$FF,$01,$B5,$FE,$22,$1C
	.dc.b	$3F,$D0,$3F,$09,$FE,$36,$FF,$19,$6B,$F0,$DA,$EF,$C9,$EF,$BC,$8E
	.dc.b	$ED,$7B,$FE,$0D,$FE,$5C,$7F,$C2,$AD,$73,$F1,$9F,$94,$8C,$47,$F2
	.dc.b	$65,$F3,$59,$24,$72,$7E,$DF,$E0,$EF,$22,$92,$7F,$99,$46,$5F,$0F
	.dc.b	$CD,$77,$23,$83,$F9,$C7,$A4,$F8,$9F,$C3,$B6,$33,$18,$3F,$57,$F5
	.dc.b	$BF,$C8,$A3,$FC,$62,$D7,$E7,$B5,$FE,$82,$D7,$FC,$74,$FE,$6D,$DF
	.dc.b	$57,$42,$3F,$E3,$CF,$C1,$DF,$53,$FB,$A6,$C6,$63,$07,$E7,$97,$DB
	.dc.b	$D2,$3F,$D5,$4C,$9D,$7E,$CB,$A1,$1C,$1F,$D6,$FE,$D5,$8F,$EA,$9B
	.dc.b	$19,$8C,$1F,$C3,$BD,$67,$CB,$52,$4F,$F9,$5A,$32,$DE,$3E,$04,$70
	.dc.b	$6F,$DE,$E8,$47,$07,$F6,$FF,$CC,$DF,$92,$7E,$F9,$B1,$98,$C1,$E5
	.dc.b	$F0,$24,$78,$DA,$FF,$01,$FB,$87,$9D,$FC,$21,$B1,$FD,$F5,$AE,$E3
	.dc.b	$F9,$6F,$E0,$CE,$BA,$49,$FF,$3A,$C6,$5F,$F3,$BF,$CC,$BE,$AF,$F9
	.dc.b	$AB,$63,$31,$83,$3F,$2F,$23,$8B,$5F,$E3,$13,$F8,$46,$F2,$FF,$20
	.dc.b	$8C,$47,$F8,$F9,$AF,$FA,$2D,$FE,$18,$DF,$13,$C6,$47,$FC,$39,$E1
	.dc.b	$7F,$08,$7F,$C0,$AD,$5E,$4F,$F0,$DF,$13,$F9,$7C,$E6,$5F,$6F,$FA
	.dc.b	$B7,$F8,$9B,$5F,$F1,$2B,$5F,$F4,$F3,$5F,$EB,$0B,$BF,$FA,$DB,$63
	.dc.b	$FD,$42,$D7,$AC,$6B,$AA,$46,$9A,$F8,$B9,$79,$91,$E6,$F6,$FA,$B9
	.dc.b	$1D,$9A,$FA,$BE,$97,$C6,$47,$C6,$F9,$BA,$E4,$79,$9A,$FC,$2D,$7D
	.dc.b	$64,$7D,$6F,$BB,$E8,$E4,$7B,$9A,$FC,$AF,$B0,$F8,$64,$7E,$1B,$F1
	.dc.b	$BC,$E9,$1F,$19,$AF,$D2,$E9,$F9,$64,$7E,$5B,$FE,$08,$FB,$DF,$EA
	.dc.b	$F1,$FD,$29,$AF,$F9,$71,$AF,$FB,$55,$AF,$F8,$3D,$7D,$97,$32,$47
	.dc.b	$FB,$62,$32,$DF,$88,$F0,$11,$C1,$BF,$A0,$F2,$11,$C1,$CF,$EA,$3D
	.dc.b	$04,$70,$77,$FB,$5F,$61,$1C,$1E,$3F,$88,$F8,$04,$70,$79,$FE,$23
	.dc.b	$E2,$11,$C1,$EB,$F9,$2F,$90,$47,$07,$BF,$EF,$3E,$61,$1C,$1F,$0F
	.dc.b	$E4,$FE,$81,$1C,$1F,$1F,$E7,$3E,$A1,$1C,$1F,$2F,$EB,$FE,$C1,$1C
	.dc.b	$1F,$3F,$E6,$3E,$E1,$1C,$1F,$4F,$CE,$FC,$01,$1C,$1F,$5F,$E8,$BF
	.dc.b	$08,$47,$07,$F8,$38,$79,$3F,$C9,$9F,$F5,$0B,$5F,$F2,$91,$AF,$27
	.dc.b	$FC,$AB,$19,$8F,$F2,$7F,$E6,$3F,$E1,$07,$FC,$B1,$F0,$48,$C4,$6B
	.dc.b	$DA,$F0,$4E,$5F,$17,$CB,$3E,$51,$CD,$86,$DF,$F8,$92,$75,$39,$3F
	.dc.b	$C8,$FF,$5A,$FB,$9A,$4A,$FF,$5F,$46,$5F,$F6,$07,$DE,$FF,$17,$F2
	.dc.b	$CD,$27,$B5,$E4,$FC,$4F,$FD,$4F,$5E,$16,$BF,$EC,$6C,$91,$88,$F8
	.dc.b	$F1,$8F,$96,$F8,$EC,$8F,$A1,$C0,$8C,$47,$FB,$43,$77,$EF,$1F,$CA
	.dc.b	$DB,$8F,$C8,$8C,$1F,$CB,$E9,$F7,$FF,$B4,$BF,$23,$5F,$1B,$5F,$F4
	.dc.b	$64,$70,$F0,$BE,$8C,$BC,$1A,$4A,$FA,$11,$97,$E9,$7D,$FB,$AB,$EC
	.dc.b	$B6,$33,$18,$3E,$EF,$E1,$9F,$E3,$D8,$F9,$0D,$7E,$3B,$5F,$2B,$5F
	.dc.b	$F5,$17,$DF,$FF,$93,$9F,$CC,$35,$FF,$33,$B5,$FF,$7C,$B5,$FC,$CA
	.dc.b	$F5,$BF,$D5,$AD,$8E,$71,$82,$3F,$EB,$1F,$90,$FF,$01,$29,$FE,$C9
	.dc.b	$18,$6F,$B5,$DA,$BF,$4F,$39,$F4,$6C,$F8,$9B,$7E,$C6,$BD,$7D,$2B
	.dc.b	$82,$7F,$63,$E5,$FB,$1B,$5C,$CF,$F6,$34,$65,$A6,$4F,$F6,$4D,$FC
	.dc.b	$2F,$4B,$A9,$DA,$AC,$F0,$3E,$7C,$76,$69,$93,$E1,$39,$E1,$7E,$00
	.dc.b	$F8,$4F,$6C,$61,$B1,$9A,$F9,$98,$7E,$4A,$3F,$69,$1E,$63,$F4,$8E
	.dc.b	$D6,$3F,$AC,$8C,$36,$BC,$51,$89,$E9,$1F,$3B,$E7,$FF,$C3,$7C,$17
	.dc.b	$D7,$FC,$B5,$77,$CB,$CC,$F9,$0F,$C7,$FA,$BF,$05,$FC,$03,$76,$FF
	.dc.b	$35,$C7,$90,$FC,$6C,$67,$D5,$FD,$E4,$7A,$CF,$C4,$38,$E1,$FC,$43
	.dc.b	$76,$E3,$3E,$13,$78,$7E,$7A,$79,$C7,$E7,$1A,$FE,$71,$AF,$E9,$BD
	.dc.b	$57,$CE,$FE,$99,$B1,$FA,$66,$BF,$A6,$6B,$FA,$67,$78,$7E,$75,$B8
	.dc.b	$7E,$AB,$D6,$F8,$13,$9B,$F4,$66,$9F,$89,$C3,$FB,$26,$E3,$C6,$3F
	.dc.b	$15,$3B,$7F,$21,$9B,$E3,$3F,$9D,$47,$D3,$E0,$69,$FE,$C9,$8F,$AD
	.dc.b	$5B,$F1,$7C,$7A,$B2,$61,$C3,$E2,$73,$1F,$CB,$F2,$EA,$FD,$93,$17
	.dc.b	$E6,$6E,$3E,$49,$DB,$FC,$F1,$E3,$B7,$FA,$17,$2F,$14,$F9,$27,$F4
	.dc.b	$4D,$BF,$0D,$FF,$09,$8E,$CD,$7E,$19,$8D,$3F,$3E,$FE,$BF,$E5,$5F
	.dc.b	$1F,$CA,$73,$5F,$8D,$9C,$D7,$D2,$3B,$B8,$8F,$70,$71,$1F,$DF,$23
	.dc.b	$15,$E4,$9C,$C7,$F8,$01,$E5,$97,$F2,$26,$EB,$9E,$FE,$F7,$E0,$F6
	.dc.b	$3C,$3E,$55,$CB,$F3,$98,$7F,$7C,$6C,$F9,$32,$FB,$12,$8C,$47,$9D
	.dc.b	$B1,$99,$FF,$0C,$2F,$63,$24,$8C,$BD,$77,$F9,$3F,$01,$72,$FF,$7A
	.dc.b	$D7,$F8,$2D,$B8,$FF,$7B,$6B,$FD,$EB,$E1,$FF,$CC,$73,$FE,$11,$6C
	.dc.b	$7F,$80,$5A,$EF,$4F,$67,$1F,$FA,$BB,$E0,$2F,$9A,$69,$30,$FA,$57
	.dc.b	$F9,$89,$FC,$4B,$F5,$6B,$FD,$6F,$DB,$C8,$95,$C3,$C9,$7E,$FA,$FC
	.dc.b	$3F,$E6,$46,$1B,$19,$9E,$0D,$C3,$C3,$C2,$F5,$FE,$79,$2B,$E7,$BF
	.dc.b	$2F,$E9,$9E,$6D,$24,$FF,$8D,$63,$2F,$B3,$57,$E0,$3F,$B9,$C7,$F2
	.dc.b	$11,$8E,$DF,$C1,$36,$B8,$56,$EB,$FC,$A3,$3F,$DE,$7E,$83,$7E,$51
	.dc.b	$A4,$BF,$F2,$37,$D0,$FC,$AB,$FF,$19,$B5,$FF,$10,$B5,$FF,$19,$37
	.dc.b	$81,$F6,$7E,$A9,$FF,$1A,$35,$FF,$31,$C6,$B5,$F9,$F3,$99,$FF,$33
	.dc.b	$FD,$53,$E7,$7F,$36,$DC,$7F,$CE,$0D,$9F,$C0,$B5,$FD,$A7,$8A,$F0
	.dc.b	$3F,$46,$DA,$FF,$3C,$36,$7F,$12,$D7,$8F,$E0,$3C,$09,$3C,$1E,$7D
	.dc.b	$D2,$39,$B6,$7F,$27,$F6,$CE,$DF,$D9,$B6,$BF,$D1,$0D,$9D,$CE,$BB
	.dc.b	$E5,$69,$B3,$D3,$E3,$79,$92,$79,$BD,$BE,$FE,$47,$66,$CF,$93,$9F
	.dc.b	$C6,$49,$F1,$BE,$6F,$B7,$91,$E6,$6C,$FB,$3E,$4F,$D6,$49,$F5,$BE
	.dc.b	$EF,$99,$91,$EE,$6C,$FC,$4E,$8F,$86,$49,$F8,$6F,$C6,$E5,$91,$F1
	.dc.b	$9B,$3F,$33,$E3,$FE,$59,$27,$E5,$BF,$3B,$D6,$91,$F3,$9B,$3F,$53
	.dc.b	$D6,$FA,$64,$9F,$A6,$FD,$6F,$5A,$47,$D6,$6C,$FD,$CF,$8F,$FB,$64
	.dc.b	$9F,$B6,$FD,$EF,$52,$47,$DE,$6C,$FE,$13,$D4,$FC,$19,$27,$F0,$6F
	.dc.b	$E1,$BD,$39,$1F,$86,$6C,$FE,$33,$A7,$F1,$64,$9F,$C5,$BF,$8E,$F8
	.dc.b	$F9,$1F,$8E,$6C,$FE,$52,$3F,$26,$43,$F2,$67,$F4,$8F,$AF,$FE,$C1
	.dc.b	$1F,$D7,$5A,$FF,$B6,$9A,$FF,$81,$DA,$9F,$E7,$6E,$FC,$DE,$84,$9F
	.dc.b	$F7,$44,$FF,$A9,$DB,$EE,$F8,$48,$EB,$D2,$F2,$11,$C1,$FC,$97,$E7
	.dc.b	$FF,$80,$61,$FE,$19,$F1,$D1,$88,$46,$3C,$3F,$E4,$F3,$99,$7D,$97
	.dc.b	$F8,$9A,$FF,$DF,$8D,$8F,$E9,$AD,$7F,$C0,$0D,$7C,$1F,$30,$EE,$FE
	.dc.b	$66,$DA,$FE,$74,$D9,$FE,$72,$D7,$FC,$15,$ED,$7F,$75,$7F,$AD,$B5
	.dc.b	$CF,$AD,$FD,$F2,$31,$1F,$E3,$73,$FC,$AF,$F1,$D9,$25,$FF,$64,$8C
	.dc.b	$B6,$5F,$19,$CC,$91,$97,$FA,$5F,$D9,$3E,$1A,$92,$7F,$AA,$46,$5E
	.dc.b	$DE,$DB,$AF,$FC,$22,$D8,$CC,$60,$FE,$AF,$E4,$7F,$81,$B2,$FB,$0D
	.dc.b	$7E,$BB,$78,$8E,$61,$F5,$DF,$90,$27,$B7,$FE,$20,$71,$F2,$94,$98
	.dc.b	$FC,$38,$46,$FE,$0A,$4E,$5B,$7F,$DB,$A3,$FC,$20,$7F,$73,$2B,$9F
	.dc.b	$77,$FB,$8C,$62,$3F,$C8,$9F,$5D,$FA,$27,$FC,$D8,$D4,$D3,$59,$BE
	.dc.b	$93,$FC,$19,$3F,$DE,$1A,$FE,$75,$AF,$E7,$1A,$F4,$F9,$3F,$F2,$0D
	.dc.b	$1A,$4B,$FC,$24,$3E,$9F,$64,$8C,$BF,$E4,$AF,$81,$FF,$3F,$DF,$F8
	.dc.b	$AF,$E1,$30,$4A,$FF,$0B,$7C,$83,$F0,$2F,$F0,$9B,$63,$31,$83,$FC
	.dc.b	$AB,$A7,$F0,$1B,$FF,$0F,$78,$38,$25,$7F,$88,$7E,$B9,$F2,$7F,$E1
	.dc.b	$E6,$C6,$63,$07,$4F,$87,$91,$FA,$1C,$3F,$CB,$CD,$8F,$0F,$1C,$8D
	.dc.b	$37,$1F,$F1,$87,$5F,$98,$93,$CD,$FF,$17,$FA,$BF,$A1,$9F,$F1,$9B
	.dc.b	$5F,$F1,$7B,$5F,$E2,$8D,$7F,$D5,$DC,$6F,$A9,$FF,$3B,$B6,$FF,$70
	.dc.b	$D9,$FF,$2E,$B5,$FF,$1D,$FD,$73,$E9,$3F,$C8,$2D,$AF,$F2,$43,$67
	.dc.b	$FC,$8E,$D7,$FC,$87,$F5,$7F,$E3,$9D,$4D,$27,$C5,$FF,$93,$1F,$C0
	.dc.b	$CF,$F2,$D6,$BF,$CB,$1B,$3C,$B9,$38,$7D,$64,$C7,$FC,$B5,$48,$C5
	.dc.b	$FF,$15,$DB,$FD,$2E,$FF,$14,$AF,$E3,$0D,$34,$E3,$9F,$64,$70,$70
	.dc.b	$87,$81,$48,$F4,$7F,$92,$79,$3F,$47,$E1,$2F,$09,$A4,$AF,$F4,$6C
	.dc.b	$BE,$B3,$F6,$D1,$86,$BF,$EC,$38,$D7,$17,$69,$CC,$FF,$37,$9F,$F2
	.dc.b	$A3,$F8,$2C,$71,$F0,$7F,$CE,$EF,$FD,$D9,$5A,$AF,$E0,$95,$C7,$8F
	.dc.b	$4F,$F4,$1F,$37,$FD,$17,$3C,$CF,$F7,$D3,$87,$89,$8F,$6B,$8B,$FE
	.dc.b	$FF,$BE,$D8,$2F,$FB,$57,$D6,$7E,$D8,$8F,$EA,$DF,$DC,$FD,$67,$FB
	.dc.b	$B2,$FC,$1C,$A8,$C4,$78,$23,$1E,$37,$81,$57,$79,$E3,$FB,$7F,$FA
	.dc.b	$4D,$BF,$15,$B8,$C4,$73,$C3,$AE,$5C,$61,$F4,$9F,$EC,$57,$E1,$18
	.dc.b	$F1,$FD,$ED,$3C,$FF,$26,$F6,$FC,$8B,$DB,$A4,$9E,$77,$D5,$F9,$67
	.dc.b	$83,$93,$AC,$7C,$AE,$0F,$EE,$8D,$9F,$F0,$8F,$D6,$FF,$1A,$7E,$3A
	.dc.b	$F5,$7E,$23,$63,$31,$D2,$FD,$3F,$5A,$F8,$4F,$D2,$B6,$BF,$BF,$35
	.dc.b	$FF,$71,$B5,$FF,$03,$05,$F0,$9D,$09,$1F,$22,$39,$6F,$F3,$57,$C2
	.dc.b	$78,$89,$3F,$80,$FA,$D7,$8F,$F7,$7C,$6E,$B3,$FE,$D9,$6B,$FE,$D8
	.dc.b	$6B,$FE,$D4,$6B,$FE,$39,$FC,$47,$BB,$F8,$C6,$BF,$8B,$6E,$3F,$89
	.dc.b	$E9,$57,$E7,$E3,$F0,$9F,$07,$FC,$B6,$FF,$06,$D7,$EC,$B6,$37,$D3
	.dc.b	$F8,$68,$FD,$84,$FC,$86,$BF,$E2,$F6,$A0,$E1,$7D,$77,$F9,$0D,$FC
	.dc.b	$9B,$5F,$C4,$B6,$BF,$34,$BE,$2F,$F4,$DF,$36,$FB,$9A,$4B,$FD,$78
	.dc.b	$FA,$6F,$DD,$FC,$87,$8F,$AB,$FC,$80,$D9,$FC,$08,$70,$FE,$11,$AF
	.dc.b	$F9,$2D,$7A,$EF,$98,$EA,$D9,$FE,$1D,$3A,$9F,$DE,$7D,$7F,$C2,$9F
	.dc.b	$D8,$F5,$7A,$DB,$A7,$EC,$A3,$11,$FA,$A7,$FD,$CC,$F0,$E9,$FE,$83
	.dc.b	$1F,$99,$8F,$F2,$F3,$5F,$A0,$D7,$F5,$5E,$FF,$52,$3E,$2F,$8D,$F8
	.dc.b	$BF,$01,$FC,$2B,$8B,$F8,$86,$9F,$E4,$CE,$4C,$9F,$C5,$7E,$BD,$F5
	.dc.b	$FF,$33,$E9,$1F,$6B,$F3,$1B,$9F,$CC,$C1,$18,$8F,$98,$D7,$E6,$7D
	.dc.b	$0F,$F9,$AD,$F9,$2D,$7F,$9A,$CA,$31,$97,$E7,$69,$F5,$5F,$C6,$BE
	.dc.b	$C1,$F3,$54,$98,$FF,$20,$8F,$E2,$0B,$EC,$1F,$35,$FD,$B1,$B1,$FE
	.dc.b	$85,$6B,$F8,$66,$BF,$C2,$1F,$D6,$79,$5F,$E8,$F6,$C7,$E8,$5A,$F7
	.dc.b	$F1,$FF,$27,$F8,$77,$F0,$0D,$3E,$7B,$6F,$F5,$1B,$23,$17,$FE,$B6
	.dc.b	$C3,$F0,$B7,$F5,$BF,$04,$FC,$3C,$FE,$55,$AF,$26,$BF,$9D,$6D,$BF
	.dc.b	$0D,$FE,$7D,$3F,$98,$6B,$F4,$1B,$5F,$C2,$FC,$DF,$E7,$CF,$F4,$DE
	.dc.b	$1F,$E1,$F9,$FF,$AB,$DB,$7F,$C0,$5A,$FD,$9F,$5B,$FA,$7C,$3E,$6F
	.dc.b	$FC,$36,$D9,$FF,$5A,$35,$FD,$AB,$5F,$E6,$FF,$8F,$FF,$89,$DF,$EB
	.dc.b	$6D,$7F,$A8,$B7,$0F,$35,$78,$3A,$E5,$7F,$62,$BF,$F5,$A3,$7A,$FF
	.dc.b	$83,$FE,$1A,$FB,$3A,$49,$E1,$7E,$AF,$26,$47,$37,$DB,$F3,$DF,$DB
	.dc.b	$3F,$0B,$FE,$E7,$BF,$FB,$4D,$B5,$FC,$95,$AF,$E7,$6B,$E2,$F8,$F2
	.dc.b	$3C,$CF,$C8,$F5,$BE,$4F,$CF,$C9,$F5,$3B,$7F,$4F,$FB,$17,$F6,$76
	.dc.b	$F8,$BF,$C8,$A3,$1C,$5F,$66,$BE,$9E,$72,$FE,$23,$F5,$BE,$83,$FC
	.dc.b	$0B,$EA,$7E,$0A,$3F,$DD,$8D,$9F,$E7,$CD,$7F,$6D,$1F,$E0,$CF,$B7
	.dc.b	$7D,$0D,$24,$FF,$84,$A3,$E1,$F5,$BE,$A3,$F8,$6C,$FF,$2A,$6B,$FC
	.dc.b	$09,$B9,$FF,$87,$6B,$AC,$F3,$FA,$1F,$01,$1E,$DF,$29,$E3,$F9,$FF
	.dc.b	$F3,$4D,$1A,$49,$FE,$86,$3F,$08,$7C,$77,$F8,$99,$B1,$FE,$2A,$6B
	.dc.b	$FE,$20,$AE,$8F,$B7,$E4,$FE,$11,$1F,$81,$F2,$9F,$D5,$3D,$2F,$4C
	.dc.b	$7F,$1E,$6E,$9F,$CC,$9B,$C1,$FD,$AE,$58,$7F,$BD,$9F,$95,$D9,$F3
	.dc.b	$C8,$FC,$DF,$B8,$FF,$1E,$FD,$D3,$E3,$29,$27,$89,$F4,$FE,$8B,$EC
	.dc.b	$11,$FA,$DF,$73,$F6,$FE,$7B,$E4,$92,$3E,$3E,$5F,$EE,$B7,$FA,$07
	.dc.b	$D4,$57,$F9,$CB,$76,$FE,$9A,$DD,$7D,$95,$6B,$FC,$10,$BE,$BB,$F1
	.dc.b	$44,$8F,$C4,$7D,$0F,$F9,$63,$9F,$84,$1A,$62,$B9,$47,$E3,$F9,$7F
	.dc.b	$28,$47,$F2,$5F,$01,$FE,$06,$FC,$0F,$FA,$6E,$5F,$CC,$9B,$3F,$E0
	.dc.b	$56,$BF,$4A,$BF,$D0,$FF,$2C,$FB,$AF,$F2,$D4,$FF,$93,$23,$1F,$3F
	.dc.b	$F8,$C6,$C7,$F8,$41,$AF,$F8,$3F,$F0,$EE,$FD,$E6,$9E,$29,$DF,$0F
	.dc.b	$F8,$F7,$EE,$DF,$41,$FE,$89,$6E,$1F,$E9,$06,$BF,$DA,$BE,$93,$E3
	.dc.b	$3F,$E1,$67,$F2,$EE,$7C,$0F,$F1,$3C,$62,$FB,$5F,$F8,$AB,$EE,$BF
	.dc.b	$BA,$5F,$F8,$8D,$AF,$F5,$86,$BF,$E2,$B6,$E7,$E6,$EA,$FF,$19,$D7
	.dc.b	$E4,$EB,$FC,$4E,$D7,$FB,$FD,$7B,$FE,$EB,$FD,$6D,$1F,$A1,$D1,$18
	.dc.b	$8F,$F2,$74,$63,$D6,$F9,$D3,$99,$FF,$1E,$7E,$61,$FD,$53,$23,$4F
	.dc.b	$EA,$F7,$C6,$7F,$C7,$3F,$76,$F1,$BF,$C9,$2D,$8C,$C6,$3C,$8C,$F5
	.dc.b	$49,$E2,$F0,$9F,$F2,$8F,$C7,$7F,$A0,$A7,$FC,$48,$D7,$FC,$9C,$D7
	.dc.b	$FC,$9C,$D7,$FC,$D5,$F1,$BF,$EF,$89,$FF,$6A,$B6,$3F,$DB,$8D,$7F
	.dc.b	$C4,$F3,$D7,$87,$FD,$85,$F7,$3F,$E0,$A9,$34,$9E,$3F,$FB,$09,$FF
	.dc.b	$49,$FD,$1F,$FB,$B5,$FF,$73,$37,$87,$F8,$93,$9F,$82,$BF,$31,$FC
	.dc.b	$56,$B6,$DA,$CE,$1F,$E8,$39,$CB,$6B,$2F,$93,$DB,$B2,$69,$8F,$76
	.dc.b	$CF,$FA,$37,$1F,$37,$B7,$C0,$CC,$23,$2E,$F2,$F3,$4E,$E3,$07,$96
	.dc.b	$F0,$F1,$9F,$E6,$1D,$5F,$E8,$58,$D8,$70,$EC,$F7,$57,$C0,$E6,$7D
	.dc.b	$D7,$FA,$CA,$7B,$C6,$3E,$53,$53,$99,$E5,$3C,$3A,$F3,$5F,$2B,$8F
	.dc.b	$FC,$C5,$E3,$DF,$7C,$EB,$0F,$5C,$E9,$B3,$A6,$CE,$DF,$41,$F8,$4F
	.dc.b	$73,$E8,$FB,$7E,$FE,$0F,$C1,$37,$7F,$BD,$3A,$D3,$E1,$F1,$2B,$58
	.dc.b	$7D,$F6,$BD,$4D,$BF,$21,$FC,$43,$5E,$13,$FD,$CA,$7A,$37,$DC,$7D
	.dc.b	$7D,$25,$7F,$AE,$A9,$3A,$8C,$37,$2F,$CE,$37,$2F,$CC,$7B,$6F,$2F
	.dc.b	$D7,$39,$BF,$CC,$4E,$5F,$8F,$87,$6C,$7E,$0C,$E7,$1E,$B1,$F4,$AF
	.dc.b	$FE,$00,$E1,$7E,$63,$59,$7F,$9F,$BE,$73,$F9,$CC,$E4,$FC,$2B,$8C
	.dc.b	$FF,$19,$C3,$EC,$BE,$31,$FE,$D7,$9D,$64,$C3,$F8,$3E,$3E,$0A,$FC
	.dc.b	$5D,$6A,$7C,$1C,$7C,$B6,$4B,$8F,$0E,$EF,$F7,$1A,$78,$1D,$75,$DE
	.dc.b	$0F,$91,$3F,$A4,$B4,$E6,$21,$FA,$8F,$95,$7D,$0E,$30,$FC,$8C,$62
	.dc.b	$7F,$16,$77,$71,$1F,$42,$31,$31,$E9,$E1,$59,$E7,$E1,$71,$39,$79
	.dc.b	$36,$BC,$4D,$9F,$0B,$78,$7E,$54,$0F,$F7,$F4,$7E,$07,$C1,$FC,$14
	.dc.b	$62,$B5,$18,$C9,$ED,$6E,$7F,$C7,$11,$FE,$4F,$7F,$9B,$FE,$00,$FA
	.dc.b	$6D,$60,$E7,$4E,$6B,$A9,$FE,$C4,$D9,$F2,$FC,$2B,$78,$72,$E4,$F9
	.dc.b	$B3,$E6,$C7,$83,$54,$6B,$0F,$F7,$B6,$1D,$1F,$F2,$74,$FF,$50,$F9
	.dc.b	$BF,$C9,$6B,$FA,$1C,$91,$8A,$D7,$D0,$78,$E3,$17,$EB,$51,$93,$5C
	.dc.b	$96,$BD,$B5,$E4,$FC,$A7,$F2,$51,$8C,$D1,$89,$65,$FB,$3F,$31,$F7
	.dc.b	$6F,$BC,$F8,$57,$BA,$D4,$E7,$1F,$F1,$83,$70,$FF,$1B,$35,$FF,$07
	.dc.b	$35,$FF,$79,$EF,$F9,$2A,$FE,$05,$97,$D2,$DF,$E8,$7D,$E3,$CC,$FA
	.dc.b	$57,$BC,$BF,$DF,$73,$97,$E6,$E3,$F4,$A3,$0F,$D0,$C1,$18,$8F,$F4
	.dc.b	$04,$62,$FE,$D2,$FB,$CF,$9E,$DE,$EF,$C5,$F1,$3C,$D7,$CE,$FE,$16
	.dc.b	$FE,$AC,$E6,$37,$3A,$CF,$E8,$57,$C9,$7F,$45,$5B,$7E,$A7,$C1,$BE
	.dc.b	$1B,$F2,$0D,$8F,$B7,$C8,$FB,$CF,$6F,$17,$C8,$AD,$6F,$F6,$AF,$FC
	.dc.b	$73,$5E,$F8,$BB,$FE,$5F,$0E,$4D,$9D,$4F,$E6,$BF,$0D,$7D,$9F,$E8
	.dc.b	$5B,$87,$E3,$25,$E5,$7F,$25,$8F,$C6,$4E,$A7,$C4,$7E,$FE,$74,$E0
	.dc.b	$E3,$E3,$FE,$D9,$E7,$F3,$76,$77,$AF,$F0,$4F,$13,$FD,$31,$8F,$85
	.dc.b	$B7,$CE,$BF,$63,$39,$F8,$4E,$5C,$D2,$3D,$1F,$42,$FB,$CC,$61,$E6
	.dc.b	$A7,$EB,$EF,$63,$F8,$03,$70,$E3,$D9,$89,$F4,$63,$FB,$DC,$3F,$22
	.dc.b	$D7,$F7,$ED,$7F,$D9,$CD,$7F,$91,$2F,$BA,$E6,$74,$F9,$EB,$87,$CF
	.dc.b	$E7,$70,$FC,$76,$C7,$B3,$D1,$C4,$F9,$19,$1F,$03,$E2,$AE,$5E,$FF
	.dc.b	$8E,$7A,$DF,$C7,$5B,$5F,$1B,$EF,$71,$3E,$E6,$47,$CA,$F4,$AD,$FD
	.dc.b	$BF,$B4,$7C,$CF,$DB,$6D,$7D,$0F,$0F,$13,$F2,$1F,$E5,$14,$E1,$FD
	.dc.b	$65,$7E,$EA,$31,$3A,$9C,$C7,$CE,$EF,$FA,$27,$D2,$FA,$BB,$7F,$C7
	.dc.b	$7E,$23,$FC,$36,$DF,$6B,$EB,$F3,$7F,$3B,$9C,$E1,$F9,$3C,$7E,$F3
	.dc.b	$67,$EE,$FC,$0F,$0C,$3F,$41,$AF,$C6,$AF,$81,$C1,$F8,$2F,$80,$F8
	.dc.b	$0D,$AF,$C7,$C7,$E1,$B7,$C5,$7E,$3F,$5F,$0F,$81,$FC,$81,$B7,$F8
	.dc.b	$6C,$3F,$A8,$46,$2B,$F0,$F1,$F9,$03,$F2,$3E,$ED,$AF,$CE,$B8,$7F
	.dc.b	$3F,$EE,$DA,$FD,$87,$E1,$6F,$8F,$FD,$83,$6B,$F4,$BE,$1E,$27,$F3
	.dc.b	$AC,$8F,$D4,$FC,$C5,$BF,$D4,$79,$3F,$E2,$3D,$7F,$9C,$B6,$BF,$3E
	.dc.b	$7F,$69,$71,$3F,$A0,$8F,$D4,$9F,$C0,$7E,$FB,$F9,$E4,$66,$3F,$87
	.dc.b	$7E,$49,$F3,$EF,$FC,$0A,$D8,$FF,$05,$4A,$9F,$01,$FC,$2E,$B5,$3F
	.dc.b	$E0,$F9,$CD,$7E,$35,$B1,$FD,$21,$AF,$F7,$66,$BD,$9F,$F6,$C4,$7F
	.dc.b	$21,$6C,$7F,$1B,$5E,$4F,$AF,$E2,$3F,$C1,$1F,$59,$FB,$DB,$FE,$E0
	.dc.b	$DB,$FE,$4B,$E8,$BE,$D1,$F1,$FC,$3C,$9F,$F0,$87,$E1,$3F,$CD,$3C
	.dc.b	$D3,$49,$E1,$7F,$58,$FC,$21,$F5,$BF,$E1,$76,$EB,$FE,$02,$CD,$F4
	.dc.b	$3E,$1E,$4F,$ED,$F5,$AD,$BF,$9A,$78,$BE,$AC,$3F,$9F,$36,$33,$F0
	.dc.b	$FF,$CD,$E3,$17,$FE,$38,$5F,$37,$FB,$C6,$DF,$EE,$DB,$7F,$DD,$9A
	.dc.b	$FF,$89,$1F,$C9,$3E,$2E,$BE,$DF,$31,$DD,$FF,$19,$7E,$11,$AC,$7F
	.dc.b	$C0,$ED,$C3,$FB,$5F,$E0,$0F,$C0,$77,$7F,$E3,$9F,$65,$F8,$67,$F7
	.dc.b	$E6,$DF,$F8,$23,$97,$FC,$FD,$39,$3F,$C0,$7F,$82,$3E,$37,$FC,$84
	.dc.b	$DA,$F9,$9E,$77,$F4,$1D,$7F,$C3,$33,$9D,$FF,$AF,$7C,$7F,$F9,$E9
	.dc.b	$B7,$FD,$7F,$F0,$3F,$CF,$C7,$FB,$AE,$31,$F2,$1F,$C9,$27,$FC,$BF
	.dc.b	$1F,$C8,$DB,$3F,$E1,$E6,$BF,$E4,$C5,$F4,$3D,$4E,$DF,$29,$6B,$FC
	.dc.b	$80,$7F,$81,$B2,$FF,$17,$B4,$FF,$09,$9F,$E4,$C3,$7F,$0B,$FE,$61
	.dc.b	$87,$D1,$FF,$89,$9B,$3F,$25,$B8,$FF,$8B,$DA,$FE,$DB,$B1,$F8,$4F
	.dc.b	$F2,$E9,$FF,$60,$46,$3D,$6F,$F1,$6C,$7F,$92,$4C,$CE,$EF,$F7,$FF
	.dc.b	$82,$3E,$97,$FC,$10,$D8,$FF,$0A,$3F,$09,$B1,$AE,$BF,$E9,$F2,$69
	.dc.b	$27,$FA,$87,$CA,$7C,$1C,$FF,$C2,$CD,$C7,$FD,$2A,$FF,$9C,$A3,$FD
	.dc.b	$19,$F2,$9F,$E4,$4F,$D3,$DF,$81,$53,$FC,$8F,$F3,$BF,$E2,$3C,$3D
	.dc.b	$F3,$AC,$7F,$69,$3F,$92,$C3,$FD,$0B,$BA,$31,$1F,$25,$AF,$F9,$21
	.dc.b	$AF,$F9,$2B,$CD,$FF,$15,$4F,$F1,$4C,$3E,$6F,$E0,$AF,$A5,$CD,$6B
	.dc.b	$C6,$99,$7F,$8E,$F0,$FC,$6E,$1F,$EB,$27,$FD,$59,$8F,$F8,$DD,$FC
	.dc.b	$7B,$F6,$FA,$7F,$BC,$C6,$23,$FC,$B7,$1F,$EA,$38,$7C,$33,$F0,$9F
	.dc.b	$F3,$4B,$70,$F0,$71,$62,$7D,$2C,$8F,$E3,$13,$FB,$B7,$67,$F1,$BF
	.dc.b	$C1,$BF,$8D,$B7,$CB,$FF,$25,$67,$FE,$BD,$CF,$F9,$6D,$69,$FF,$28
	.dc.b	$46,$4F,$F6,$79,$6B,$FC,$E7,$2F,$C1,$39,$FC,$EF,$FA,$03,$F0,$3F
	.dc.b	$F3,$E9,$FE,$7E,$FC,$1B,$FC,$F5,$8F,$FA,$E2,$31,$F3,$FF,$E8,$69
	.dc.b	$CE,$E7,$39,$7F,$33,$5F,$35,$F1,$DB,$1F,$E9,$A9,$E1,$39,$3F,$C4
	.dc.b	$FE,$B3,$E0,$BF,$CD,$0D,$AF,$F3,$A6,$5F,$9D,$7F,$CE,$D1,$98,$FF
	.dc.b	$38,$19,$2F,$0F,$F9,$C5,$FC,$A1,$89,$E6,$BC,$6F,$F1,$A3,$FE,$F9
	.dc.b	$6B,$93,$2F,$F8,$E7,$D8,$FF,$79,$7C,$C5,$DC,$D2,$57,$FA,$E2,$5F
	.dc.b	$83,$7F,$B4,$A3,$0D,$7F,$DD,$31,$AE,$3E,$D3,$99,$FF,$55,$FE,$0C
	.dc.b	$F8,$BF,$F0,$2D,$7F,$0E,$8C,$4F,$F6,$36,$BF,$DC,$97,$C4,$70,$D3
	.dc.b	$9D,$DA,$FF,$7A,$D7,$FA,$15,$E5,$E3,$E7,$89,$3D,$31,$FF,$6D,$FE
	.dc.b	$34,$FB,$EF,$F2,$E7,$A6,$8C,$63,$FD,$45,$AF,$FB,$5D,$AF,$FB,$59
	.dc.b	$FF,$59,$BF,$EF,$96,$BF,$E9,$96,$D7,$82,$BD,$15,$DF,$E7,$2D,$7F
	.dc.b	$62,$BE,$5E,$A6,$27,$DD,$ED,$3E,$5D,$D7,$86,$D8,$FF,$5E,$CE,$E7
	.dc.b	$5D,$5B,$78,$79,$52,$B8,$6A,$E2,$E5,$87,$19,$CC,$63,$F0,$4F,$14
	.dc.b	$60,$79,$9F,$E9,$8C,$FE,$A3,$63,$56,$AE,$31,$FE,$D2,$9D,$37,$E3
	.dc.b	$73,$8F,$8E,$3A,$C6,$31,$F1,$63,$F7,$7E,$61,$ED,$7F,$A5,$79,$51
	.dc.b	$8C,$3F,$70,$D7,$F9,$DB,$5F,$EC,$5B,$3E,$5F,$F1,$0D,$8F,$F1,$B3
	.dc.b	$5F,$F2,$2C,$F3,$CF,$CD,$B3,$F1,$97,$99,$F6,$1B,$7F,$09,$AF,$D6
	.dc.b	$7F,$CC,$AF,$C6,$FC,$AF,$F1,$AF,$D2,$6A,$A4,$61,$BD,$1F,$E3,$B8
	.dc.b	$7C,$A7,$E2,$3F,$1E,$FC,$65,$FE,$11,$B9,$7A,$2B,$73,$FE,$53,$3F
	.dc.b	$C1,$65,$F6,$3F,$54,$E0,$FC,$6F,$D6,$BD,$1F,$CD,$C3,$E6,$3F,$56
	.dc.b	$D3,$FC,$74,$53,$FD,$0A,$3B,$BF,$31,$1F,$EC,$DF,$C2,$D9,$7F,$8F
	.dc.b	$6B,$F1,$D1,$8F,$63,$F8,$03,$63,$33,$B7,$8F,$47,$E8,$E5,$F8,$3B
	.dc.b	$C9,$E2,$BE,$96,$98,$8F,$D7,$FD,$87,$EB,$30,$FF,$31,$35,$FF,$4E
	.dc.b	$F2,$23,$11,$FA,$86,$BF,$A4,$FD,$AB,$FC,$22,$FE,$9F,$BD,$18,$8F
	.dc.b	$D1,$37,$4F,$D0,$66,$3F,$CD,$13,$F0,$BF,$4D,$FC,$3B,$EE,$34,$BE
	.dc.b	$AB,$E2,$B6,$38,$7C,$BF,$F1,$0F,$C1,$FF,$C6,$D1,$FE,$6D,$6B,$F5
	.dc.b	$38,$D1,$88,$FF,$5D,$C6,$3E,$0B,$FC,$67,$F0,$BF,$E0,$99,$F6,$35
	.dc.b	$FD,$E3,$73,$F5,$FE,$11,$F8,$17,$F7,$9B,$7D,$3F,$A4,$FE,$0F,$C4
	.dc.b	$69,$F7,$BE,$97,$FD,$63,$8F,$E0,$BF,$69,$FC,$43,$F8,$A3,$E3,$38
	.dc.b	$7F,$1B,$C2,$8C,$47,$FB,$52,$31,$AF,$E3,$FF,$1D,$FF,$58,$CF,$E3
	.dc.b	$9F,$F1,$A3,$7E,$13,$F1,$F1,$A8,$FA,$15,$F1,$B1,$FC,$9F,$03,$E2
	.dc.b	$A9,$88,$FE,$7F,$83,$EC,$7F,$A7,$B7,$0F,$F4,$F3,$5F,$A4,$D7,$FD
	.dc.b	$D7,$B3,$F4,$EF,$CA,$70,$7C,$A6,$F6,$7F,$41,$9D,$79,$3F,$A6,$FA
	.dc.b	$4F,$C1,$61,$FC,$BF,$F4,$ED,$BF,$CE,$1B,$47,$3B,$FF,$9A,$3E,$6F
	.dc.b	$F5,$FF,$7F,$FD,$CA,$FE,$84,$A3,$11,$F4,$1A,$FC,$F6,$BF,$B3,$FC
	.dc.b	$7F,$F8,$43,$FB,$C7,$FD,$EA,$DF,$A1,$FD,$EC,$6A,$3F,$3C,$3F,$05
	.dc.b	$C6,$FF,$D1,$63,$1E,$3F,$F8,$0D,$A7,$F5,$51,$F8,$13,$E8,$F5,$F7
	.dc.b	$6F,$A8,$FF,$14,$B6,$BF,$18,$1C,$3F,$8F,$6B,$FC,$69,$FF,$2B,$F1
	.dc.b	$FF,$71,$8E,$0F,$63,$BC,$FF,$18,$FC,$7F,$EF,$65,$FB,$E6,$CF,$F1
	.dc.b	$76,$D7,$F0,$17,$FC,$62,$FF,$22,$73,$B9,$FE,$39,$19,$EF,$ED,$D1
	.dc.b	$FE,$0E,$FC,$C5,$ED,$DE,$79,$7F,$DE,$7B,$D5,$A8,$C2,$9E,$5F,$19
	.dc.b	$27,$31,$F3,$F9,$BF,$AD,$ED,$F4,$DF,$B7,$3F,$61,$B1,$9F,$5F,$87
	.dc.b	$3C,$AF,$AD,$5A,$E0,$F9,$4F,$46,$F2,$F7,$24,$F0,$AF,$F1,$5F,$E7
	.dc.b	$5F,$E0,$58,$FE,$98,$DD,$4D,$1F,$51,$C3,$E5,$38,$FB,$92,$FE,$25
	.dc.b	$7F,$A9,$9F,$EA,$3F,$85,$FF,$BE,$9B,$E5,$FF,$52,$8F,$E5,$3B,$FF
	.dc.b	$38,$75,$3F,$C1,$97,$D3,$6C,$91,$C8,$FC,$B7,$E3,$9F,$DC,$9F,$CF
	.dc.b	$B6,$BF,$3C,$DD,$FF,$1D,$EF,$FF,$44,$AF,$9D,$E3,$FF,$A0,$B8,$BF
	.dc.b	$95,$47,$E6,$9A,$FF,$27,$6B,$FD,$C9,$FE,$F9,$87,$D9,$F6,$BE,$E1
	.dc.b	$F7,$6B,$FB,$7F,$E7,$BF,$22,$FF,$B9,$AF,$B6,$C9,$AE,$2B,$3A,$FB
	.dc.b	$8E,$26,$1F,$EB,$1C,$8F,$C9,$BF,$C0,$79,$FF,$8F,$E3,$13,$FE,$66
	.dc.b	$9C,$ED,$FE,$5D,$9E,$1E,$CF,$F9,$AB,$F3,$B7,$C4,$DF,$8A,$98,$FE
	.dc.b	$76,$72,$FF,$03,$F2,$1E,$B7,$F0,$76,$C6,$63,$FC,$1B,$F8,$8F,$F1
	.dc.b	$CD,$7F,$39,$EC,$7C,$D2,$57,$CD,$7E,$87,$DE,$7E,$8C,$91,$F9,$C8
	.dc.b	$FF,$3F,$FE,$17,$E2,$8F,$82,$DA,$F8,$0D,$DF,$DD,$FA,$27,$C0,$AF
	.dc.b	$F1,$34,$FF,$B5,$1F,$DC,$3F,$CE,$B8,$0D,$35,$1F,$D5,$BE,$CF,$D5
	.dc.b	$7F,$E2,$36,$C7,$F8,$35,$BE,$61,$7A,$5F,$E2,$E7,$FB,$C5,$71,$FB
	.dc.b	$CC,$19,$F8,$6D,$7D,$EF,$F8,$E6,$8D,$26,$3F,$DD,$ED,$FA,$CF,$FA
	.dc.b	$CD,$FF,$4C,$18,$F9,$8F,$F5,$E3,$67,$FC,$83,$7D,$1B,$E3,$F9,$48
	.dc.b	$EB,$C8,$AF,$C1,$F9,$2F,$51,$8E,$2F,$F1,$6D,$FE,$6E,$7F,$C0,$ED
	.dc.b	$7F,$C0,$CD,$7E,$47,$77,$F4,$4C,$5F,$87,$7E,$B2,$FF,$D2,$D1,$8F
	.dc.b	$0B,$D3,$39,$8E,$37,$EA,$F0,$7E,$31,$1F,$ED,$78,$FE,$AD,$F8,$DB
	.dc.b	$7D,$27,$FD,$5D,$5F,$23,$8B,$E2,$CE,$67,$EF,$5F,$D7,$F0,$7F,$DC
	.dc.b	$1F,$37,$FC,$66,$7F,$C8,$DF,$27,$FE,$FA,$E0,$D4,$7F,$AE,$63,$0D
	.dc.b	$9E,$1B,$7F,$AC,$67,$33,$F8,$BB,$FF,$38,$FE,$F3,$FE,$2C,$8F,$F3
	.dc.b	$73,$5F,$F2,$C3,$79,$CE,$75,$FC,$0F,$2F,$A8,$95,$FE,$7E,$9F,$F1
	.dc.b	$D6,$3F,$C0,$FC,$CF,$77,$CF,$FF,$AE,$FA,$D1,$8F,$96,$FE,$08,$DC
	.dc.b	$3F,$D1,$8B,$AF,$FB,$A5,$F1,$8D,$5F,$2E,$BC,$E1,$C9,$B8,$72,$E4
	.dc.b	$B5,$B8,$FF,$77,$61,$CE,$B3,$87,$FA,$76,$3B,$5F,$0F,$98,$FF,$45
	.dc.b	$47,$A5,$B8,$FF,$1B,$6E,$1F,$EA,$37,$F8,$34,$EF,$E0,$B9,$78,$D2
	.dc.b	$BC,$37,$C6,$71,$0A,$ED,$58,$79,$47,$03,$B4,$66,$BC,$4F,$9E,$BD
	.dc.b	$FF,$50,$FB,$CB,$EF,$4C,$39,$6C,$AD,$46,$1F,$F1,$3E,$5F,$5D,$B9
	.dc.b	$7F,$39,$6C,$FF,$B0,$5E,$2F,$A6,$35,$7D,$E3,$06,$DF,$2E,$1F,$5A
	.dc.b	$1F,$84,$7F,$86,$5D,$98,$F6,$BF,$D0,$F8,$FF,$C0,$19,$46,$30,$FF
	.dc.b	$1C,$47,$19,$D4,$E6,$87,$3C,$BE,$53,$4E,$F5,$C6,$F9,$D7,$E8,$E3
	.dc.b	$BD,$7F,$93,$1B,$5F,$E9,$96,$BF,$EE,$47,$FC,$61,$E6,$EF,$DA,$FF
	.dc.b	$80,$13,$B6,$BB,$6B,$EA,$77,$8F,$BD,$B8,$FB,$B2,$73,$FA,$B1,$EF
	.dc.b	$87,$AA,$72,$FB,$2F,$8B,$F0,$A7,$F0,$95,$C6,$73,$93,$F5,$6C,$DF
	.dc.b	$D8,$8C,$61,$D3,$2F,$DD,$57,$F9,$C7,$0F,$C5,$BC,$23,$33,$BC,$BF
	.dc.b	$C0,$13,$96,$EB,$9C,$58,$F6,$C7,$9C,$E6,$31,$3E,$3B,$F0,$37,$0D
	.dc.b	$7D,$37,$F5,$28,$F4,$37,$2F,$E2,$1F,$40,$E2,$DD,$4C,$FF,$47,$7B
	.dc.b	$D3,$FC,$2D,$7B,$9C,$D7,$8E,$30,$F6,$31,$7C,$DF,$91,$1A,$C3,$C3
	.dc.b	$3C,$7D,$1D,$DF,$EC,$BF,$39,$7D,$3F,$BD,$B1,$9F,$B8,$C4,$E5,$B1
	.dc.b	$93,$94,$65,$C6,$5F,$66,$9F,$5B,$98,$FD,$9C,$BF,$77,$FE,$66,$FF
	.dc.b	$56,$6C,$7F,$B9,$BB,$DF,$7B,$F8,$CF,$03,$FB,$3D,$69,$FC,$7F,$85
	.dc.b	$FD,$1A,$7F,$D7,$97,$FD,$31,$AF,$F4,$C6,$C6,$F0,$FF,$58,$4F,$ED
	.dc.b	$BC,$47,$E2,$4F,$43,$C1,$59,$8C,$3F,$D3,$FE,$5B,$36,$BE,$37,$8C
	.dc.b	$3F,$AC,$D6,$A7,$37,$C4,$FE,$D5,$87,$4C,$19,$7B,$BD,$5E,$B5,$F1
	.dc.b	$DE,$9A,$B9,$B3,$B2,$73,$1E,$E9,$D5,$7F,$BC,$1B,$87,$C0,$D3,$FB
	.dc.b	$6D,$7C,$6A,$71,$3E,$0F,$C2,$D8,$FE,$DD,$F5,$D8,$9F,$1B,$DC,$91
	.dc.b	$DD,$F1,$7D,$87,$A3,$E5,$3D,$2D,$9F,$EF,$0E,$1F,$EF,$5F,$17,$73
	.dc.b	$F8,$71,$FE,$3E,$D9,$F5,$DF,$00,$F8,$7E,$7D,$9F,$F0,$13,$F0,$48
	.dc.b	$FC,$17,$E2,$7C,$6F,$CA,$F9,$FF,$96,$D9,$FF,$05,$6D,$F3,$8F,$9F
	.dc.b	$D3,$63,$FD,$C1,$F8,$3B,$6F,$C0,$F8,$88,$6D,$87,$55,$C2,$31,$BC
	.dc.b	$8E,$7C,$17,$83,$DE,$7F,$48,$F0,$FF,$8A,$CF,$F4,$AE,$7F,$E5,$B3
	.dc.b	$FD,$B7,$2F,$C8,$AF,$BA,$7A,$78,$C3,$8C,$F4,$35,$1F,$A4,$6E,$1F
	.dc.b	$A2,$7F,$AB,$D7,$28,$C4,$68,$D7,$B6,$FA,$8F,$0B,$67,$CC,$7F,$60
	.dc.b	$E9,$7C,$0B,$DE,$DD,$FC,$81,$BD,$E2,$C3,$70,$E7,$97,$F3,$6A,$70
	.dc.b	$71,$39,$75,$DA,$DE,$6F,$CD,$3F,$9B,$F1,$7F,$06,$7F,$A6,$B7,$CC
	.dc.b	$FD,$B4,$62,$BD,$0B,$CD,$CC,$E9,$B3,$E4,$CF,$E8,$7B,$3F,$4B,$88
	.dc.b	$DE,$0C,$36,$BE,$5C,$FC,$36,$CF,$D3,$8C,$57,$D3,$FC,$11,$F0,$7F
	.dc.b	$79,$B7,$FE,$D3,$FB,$AC,$4F,$F0,$FF,$C0,$7F,$A2,$ED,$FC,$91,$B5
	.dc.b	$F4,$D7,$AF,$BF,$3F,$87,$9F,$C6,$2D,$7C,$AA,$DB,$F4,$BB,$1F,$90
	.dc.b	$7C,$C8,$FF,$76,$3F,$27,$5F,$ED,$6E,$5B,$DB,$FA,$68,$FF,$04,$CF
	.dc.b	$36,$CF,$FB,$F3,$DE,$7D,$2E,$EB,$E6,$7B,$D1,$FA,$1C,$5F,$EC,$E6
	.dc.b	$C7,$D1,$F7,$64,$6B,$EC,$EE,$3F,$A1,$B4,$63,$87,$F5,$71,$8A,$EB
	.dc.b	$7F,$BC,$FC,$E7,$C7,$1F,$07,$E1,$BC,$F7,$FE,$1E,$DD,$5A,$CB,$F7
	.dc.b	$30,$AD,$DF,$F9,$0A,$37,$F2,$FD,$75,$EC,$D9,$FC,$CF,$E3,$7F,$D8
	.dc.b	$1B,$5F,$D2,$E3,$18,$7F,$06,$EE,$47,$EF,$B3,$FE,$C6,$E9,$BE,$BF
	.dc.b	$1F,$5B,$93,$F9,$7F,$C2,$66,$5F,$33,$F0,$9B,$97,$C7,$7F,$28,$48
	.dc.b	$FC,$CB,$FC,$27,$B5,$F3,$FF,$95,$3F,$89,$FC,$DB,$EA,$71,$7F,$B4
	.dc.b	$F8,$57,$D7,$7F,$65,$5F,$25,$63,$0F,$F6,$9F,$A1,$FC,$8E,$9F,$D5
	.dc.b	$DA,$FF,$9B,$07,$12,$33,$1F,$C4,$5A,$BE,$95,$FA,$6F,$E5,$DB,$E0
	.dc.b	$7F,$77,$FC,$9B,$EC,$69,$FC,$42,$72,$DE,$EF,$B7,$87,$36,$E4,$FC
	.dc.b	$23,$DB,$5B,$3F,$A2,$FC,$9F,$F9,$22,$FD,$1D,$48,$C4,$7E,$15,$AF
	.dc.b	$F9,$51,$B6,$FC,$21,$F1,$DF,$E0,$DF,$45,$E4,$FF,$83,$E7,$39,$7F
	.dc.b	$22,$9C,$BF,$C9,$3F,$07,$FF,$6E,$37,$C4,$FF,$0B,$46,$33,$FC,$63
	.dc.b	$FA,$D7,$F1,$DF,$A2,$BF,$09,$FF,$14,$B6,$FF,$A4,$68,$8C,$57,$F8
	.dc.b	$35,$AF,$D7,$E1,$7E,$9A,$F0,$7F,$C5,$31,$87,$FC,$51,$F8,$5F,$F2
	.dc.b	$2F,$0F,$FD,$17,$39,$CB,$EA,$7E,$19,$F8,$D6,$F7,$7F,$8B,$FF,$4D
	.dc.b	$78,$FD,$2B,$5A,$7E,$D3,$8B,$FA,$7E,$1F,$D7,$F9,$FF,$C5,$5E,$DF
	.dc.b	$F9,$19,$B1,$99,$DD,$7F,$89,$FD,$8F,$CF,$CF,$F4,$E7,$FC,$26,$DF
	.dc.b	$13,$FC,$81,$18,$9F,$EF,$7A,$BE,$DD,$F4,$BF,$01,$B8,$7F,$58,$FC
	.dc.b	$57,$E0,$3F,$EC,$8A,$E1,$5F,$E4,$BB,$FF,$52,$9F,$A1,$E6,$FF,$03
	.dc.b	$57,$D2,$8C,$9F,$DC,$4A,$7E,$9B,$F0,$DE,$AD,$F9,$DF,$F2,$B4,$62
	.dc.b	$8F,$E0,$F9,$3F,$CD,$BB,$FF,$91,$30,$FE,$DA,$BE,$27,$FC,$BC,$E7
	.dc.b	$1F,$E2,$8E,$67,$FA,$45,$3A,$FE,$73,$63,$E9,$36,$BF,$CE,$75,$FE
	.dc.b	$62,$AF,$F3,$CF,$C9,$7F,$99,$3A,$7F,$CD,$57,$F8,$9C,$3F,$C0,$BE
	.dc.b	$07,$F4,$8F,$91,$CD,$7F,$9D,$21,$FA,$4F,$E1,$2B,$7F,$A6,$FD,$0D
	.dc.b	$3F,$C5,$F5,$AE,$AE,$13,$9C,$BF,$DB,$73,$CB,$B3,$FD,$16,$DC,$3F
	.dc.b	$C5,$0D,$C3,$FD,$16,$D7,$F8,$B2,$FC,$AD,$F1,$FE,$56,$D7,$A9,$F9
	.dc.b	$58,$78,$1C,$DF,$F9,$13,$67,$E1,$B8,$AF,$F7,$7C,$63,$0F,$F4,$77
	.dc.b	$E3,$AF,$A3,$FF,$48,$5E,$F2,$FF,$4F,$4E,$5F,$F2,$05,$A3,$18,$FF
	.dc.b	$A9,$23,$17,$F4,$A7,$33,$FB,$BF,$CA,$3F,$9C,$DF,$F2,$76,$C7,$CC
	.dc.b	$FC,$5F,$F9,$3F,$5A,$73,$A7,$CC,$9C,$C6,$2D,$7F,$C7,$A7,$3E,$97
	.dc.b	$DB,$FB,$0F,$54,$FF,$22,$7F,$1C,$FD,$A7,$3C,$07,$38,$FD,$A7,$24
	.dc.b	$F9,$5E,$9C,$1F,$E7,$AF,$82,$FC,$95,$7F,$7D,$7F,$9D,$E1,$FE,$15
	.dc.b	$E9,$D7,$AD,$B6,$6B,$FE,$25,$AF,$55,$3A,$D1,$D2,$7F,$C0,$4F,$C6
	.dc.b	$9F,$E8,$6D,$73,$AF,$F1,$38,$C4,$7F,$B3,$7E,$87,$D7,$5F,$ED,$77
	.dc.b	$E1,$F2,$1C,$CE,$F4,$FF,$68,$46,$27,$5D,$B8,$F9,$7C,$BF,$E8,$AB
	.dc.b	$FA,$3F,$41,$FE,$80,$69,$5C,$27,$FD,$0B,$F8,$67,$F0,$3F,$C3,$6D
	.dc.b	$1A,$4A,$FF,$47,$4A,$FF,$DB,$F1,$86,$BF,$EF,$48,$D7,$0F,$69,$CC
	.dc.b	$FF,$BD,$FE,$8D,$F8,$6F,$F8,$0E,$BF,$47,$18,$9F,$F4,$4E,$68,$C4
	.dc.b	$7E,$AD,$FF,$7F,$3A,$73,$3F,$AC,$FC,$33,$33,$FB,$16,$C7,$6B,$F0
	.dc.b	$6F,$FE,$3D,$FC,$35,$F5,$5F,$EA,$8E,$44,$63,$1F,$EA,$0D,$7F,$8F
	.dc.b	$B5,$FF,$02,$AC,$7A,$C7,$E2,$E3,$0F,$F9,$CD,$AF,$F9,$41,$E9,$8F
	.dc.b	$F9,$55,$CB,$FE,$57,$FC,$05,$F0,$DD,$30,$46,$2F,$A6,$1E,$0E,$B7
	.dc.b	$CE,$7F,$83,$A7,$F9,$7B,$5F,$E5,$2D,$C7,$F9,$3B,$FE,$8E,$8D,$FB
	.dc.b	$9C,$FF,$18,$FE,$C3,$1F,$CE,$1A,$F2,$8F,$7F,$82,$FB,$DF,$0C,$FF
	.dc.b	$8C,$DA,$FF,$6C,$6E,$1F,$CD,$BF,$0E,$BC,$3F,$C2,$B3,$FF,$01,$FC
	.dc.b	$77,$AF,$CB,$FF,$3A,$63,$FD,$21,$AF,$3D,$7E,$7F,$37,$FA,$9B,$E6
	.dc.b	$3F,$DC,$2D,$9F,$E4,$FF,$2B,$FE,$DB,$6C,$EA,$3F,$D8,$31,$F9,$06
	.dc.b	$BF,$CE,$5B,$A7,$F3,$87,$94,$7F,$52,$FC,$2F,$FD,$D8,$FF,$5C,$6B
	.dc.b	$9F,$9E,$FA,$9F,$87,$7F,$87,$A3,$5F,$87,$7F,$BD,$B6,$5A,$30,$D9
	.dc.b	$FE,$AE,$D7,$FA,$BF,$E1,$DF,$86,$E0,$F9,$1C,$9F,$A2,$6F,$67,$E0
	.dc.b	$E7,$55,$FE,$52,$FC,$39,$F5,$7F,$87,$6C,$7F,$74,$6B,$FD,$64,$FB
	.dc.b	$FB,$FF,$AA,$5F,$F5,$99,$FE,$33,$29,$FE,$02,$29,$FE,$68,$29,$FD
	.dc.b	$1B,$D9,$47,$FA,$32,$3F,$E0,$0A,$FA,$AF,$D4,$CE,$9B,$1B,$9F,$F5
	.dc.b	$57,$E1,$DF,$DD,$1F,$EA,$30,$CF,$B4,$BC,$EE,$6D,$AD,$CE,$BC,$AF
	.dc.b	$D4,$7E,$1C,$FA,$0F,$A8,$DA,$FF,$03,$B5,$FF,$03,$B5,$FF,$1E,$66
	.dc.b	$FA,$CF,$4B,$63,$B9,$FE,$A0,$70,$FF,$93,$BE,$0B,$D9,$A7,$A6,$3F
	.dc.b	$D4,$BF,$87,$7F,$BB,$1E,$34,$FD,$2B,$51,$FE,$CB,$6D,$71,$95,$71
	.dc.b	$AF,$F6,$EF,$E1,$AF,$A5,$FE,$3C,$D8,$FF,$36,$35,$FF,$1D,$B5,$FF
	.dc.b	$2C,$7A,$8F,$84,$FA,$AD,$8F,$2B,$5F,$F3,$4D,$7B,$B1,$F2,$7C,$17
	.dc.b	$F1,$EE,$1F,$C1,$36,$BF,$D8,$6D,$F1,$FF,$CA,$DE,$2F,$E0,$BF,$0F
	.dc.b	$7C,$FF,$FB,$26,$FF,$CD,$51,$87,$FD,$36,$D7,$5B,$A3,$11,$FE,$71
	.dc.b	$FC,$3D,$F5,$BF,$8E,$5F,$61,$4C,$47,$E4,$B8,$7F,$97,$D7,$D2,$5F
	.dc.b	$75,$93,$5E,$87,$F3,$5A,$7C,$87,$F4,$B6,$DF,$FA,$65,$AF,$F9,$A1
	.dc.b	$AF,$FA,$17,$F5,$BF,$B7,$C3,$FA,$76,$C7,$FB,$A1,$BC,$0F,$C4,$3F
	.dc.b	$9C,$4F,$EB,$2B,$F5,$DF,$02,$FA,$AF,$D5,$B6,$7F,$16,$D7,$FD,$3F
	.dc.b	$F8,$87,$2E,$6F,$C5,$7E,$10,$F9,$CF,$A1,$5F,$B9,$8C,$3F,$EC,$A6
	.dc.b	$BF,$41,$AF,$FA,$7F,$F1,$0F,$E1,$75,$F9,$57,$F9,$14,$79,$C7,$32
	.dc.b	$33,$3B,$F2,$3F,$B5,$D3,$E9,$5F,$2D,$86,$C7,$EB,$8D,$47,$0F,$C4
	.dc.b	$1F,$51,$BE,$0F,$F1,$35,$FE,$AA,$FF,$18,$D7,$FC,$B6,$77,$8F,$F1
	.dc.b	$67,$E4,$4F,$A9,$F1,$B6,$7C,$0D,$7D,$87,$13,$F7,$3F,$88,$FF,$AC
	.dc.b	$5B,$F1,$3F,$C7,$A3,$11,$F0,$F7,$47,$E5,$23,$67,$F0,$2F,$CA,$3F
	.dc.b	$83,$E3,$FC,$AD,$B1,$F2,$A7,$9E,$FF,$C2,$FE,$0D,$F8,$93,$ED,$FF
	.dc.b	$C1,$4F,$9E,$74,$DC,$7F,$CF,$75,$C2,$B7,$AC,$C7,$F8,$DC,$F4,$8E
	.dc.b	$73,$F5,$B9,$3E,$FD,$FC,$A6,$DF,$FA,$19,$A9,$FB,$F1,$BE,$0F,$F2
	.dc.b	$D4,$7D,$89,$FC,$89,$B8,$C9,$FE,$5F,$FC,$2F,$FA,$44,$79,$3F,$46
	.dc.b	$FC,$1B,$F8,$DC,$91,$8D,$3F,$94,$D7,$0A,$D7,$E2,$7F,$DF,$B2,$F8
	.dc.b	$2F,$F7,$EA,$F9,$4D,$8C,$FD,$D7,$F8,$CF,$E4,$BF,$97,$57,$CF,$6C
	.dc.b	$FC,$A6,$DF,$FA,$36,$F9,$3F,$A0,$FC,$4D,$F8,$4F,$F6,$05,$F8,$9D
	.dc.b	$31,$1F,$D9,$3E,$4F,$F9,$B6,$5F,$CF,$DB,$97,$F6,$D6,$E1,$D6,$FE
	.dc.b	$A7,$9D,$23,$F4,$DC,$3F,$E2,$AC,$3F,$A5,$FE,$27,$FE,$0D,$7F,$A6
	.dc.b	$C7,$F8,$FA,$7F,$98,$35,$FB,$9F,$4B,$F5,$8F,$F4,$87,$CD,$3F,$09
	.dc.b	$FD,$BB,$67,$F6,$8D,$7F,$4B,$5F,$E0,$CD,$3F,$C0,$9F,$89,$FF,$B8
	.dc.b	$67,$FA,$FA,$FC,$13,$FC,$06,$D8,$CF,$D0,$E2,$64,$70,$77,$1F,$E6
	.dc.b	$7F,$C4,$BF,$CB,$D7,$F9,$C7,$FB,$D6,$9F,$A5,$6C,$67,$E2,$3F,$CD
	.dc.b	$DF,$97,$BD,$9A,$4C,$BF,$43,$7C,$1F,$CE,$7B,$5F,$A1,$23,$FC,$4F
	.dc.b	$A9,$ED,$FE,$97,$EA,$7F,$3A,$49,$FC,$EF,$02,$3F,$D7,$5E,$5B,$F0
	.dc.b	$3F,$E4,$2D,$AF,$E3,$CD,$7E,$5C,$FF,$A1,$7F,$12,$F2,$78,$27,$F9
	.dc.b	$E7,$07,$EF,$1B,$5F,$ED,$F7,$7F,$1D,$FD,$53,$8B,$F8,$43,$6B,$F8
	.dc.b	$1B,$6B,$87,$C9,$7F,$8D,$BE,$DF,$FA,$55,$FF,$84,$5F,$F0,$13,$7E
	.dc.b	$0B,$5E,$B7,$F8,$F3,$F0,$AF,$C1,$4F,$F8,$5D,$B9,$7F,$91,$5B,$8F
	.dc.b	$F4,$4C,$39,$CA,$3F,$C7,$DF,$88,$7F,$AF,$9F,$E6,$F0,$8C,$46,$E3
	.dc.b	$1F,$86,$FF,$0A,$6C,$7F,$5C,$F4,$18,$7F,$84,$DB,$F4,$3E,$73,$FD
	.dc.b	$77,$5C,$2B,$F9,$B6,$3F,$E9,$87,$CD,$F3,$5F,$0C,$F8,$9F,$33,$63
	.dc.b	$FC,$73,$8F,$F9,$CB,$1E,$FF,$3F,$99,$FE,$C5,$8B,$F1,$1F,$F4,$EC
	.dc.b	$7B,$1A,$FA,$DA,$FF,$97,$DB,$5E,$AF,$9B,$F9,$C4,$9F,$F2,$8D,$75
	.dc.b	$F2,$3D,$7B,$FF,$76,$9F,$8C,$D7,$E6,$FB,$BF,$1E,$31,$87,$C6,$E5
	.dc.b	$FB,$44,$AE,$B3,$FE,$02,$E2,$FF,$2D,$7C,$E3,$E9,$3F,$96,$52,$31
	.dc.b	$E9,$7F,$AE,$27,$38,$B8,$FF,$7D,$7F,$D8,$F8,$7F,$2F,$3F,$5B,$A3
	.dc.b	$F1,$47,$1C,$3D,$7F,$4F,$E7,$25,$FF,$86,$38,$3F,$09,$1F,$D2,$AF
	.dc.b	$EF,$FC,$A7,$E4,$C9,$5E,$A9,$FF,$77,$7A,$3F,$E1,$CF,$A7,$7D,$A7
	.dc.b	$F6,$48,$FF,$14,$46,$29,$DF,$FC,$EB,$B7,$FC,$13,$39,$8F,$E7,$BC
	.dc.b	$B8,$C3,$95,$7E,$62,$E9,$F9,$72,$9F,$E7,$EA,$E8,$E4,$EF,$87,$8A
	.dc.b	$BF,$49,$F0,$EF,$8F,$7C,$FF,$7B,$4E,$63,$18,$74,$96,$7E,$7E,$FF
	.dc.b	$F2,$5B,$76,$DC,$7F,$A8,$63,$5C,$58,$99,$5F,$B5,$9E,$0D,$CF,$FC
	.dc.b	$17,$7D,$1B,$5C,$27,$C3,$18,$78,$1F,$32,$78,$36,$7D,$99,$7F,$86
	.dc.b	$FC,$FB,$EE,$7E,$23,$1F,$0D,$F4,$3E,$AF,$C7,$3E,$FF,$E0,$BF,$E0
	.dc.b	$69,$FA,$CD,$CF,$FC,$42,$EE,$BC,$7A,$FF,$A0,$3E,$1F,$89,$23,$33
	.dc.b	$B7,$C1,$3B,$7D,$7C,$BF,$C3,$BD,$2F,$A8,$D8,$E3,$7F,$5D,$7E,$22
	.dc.b	$FA,$5C,$3F,$E5,$1C,$3E,$2C,$EB,$E0,$BF,$22,$FB,$CE,$35,$C2,$33
	.dc.b	$A7,$E5,$E7,$85,$72,$BE,$91,$C5,$F0,$39,$AE,$AF,$FA,$C6,$36,$D8
	.dc.b	$CB,$FE,$68,$8C,$B7,$6F,$C9,$BF,$E8,$A8,$FF,$29,$72,$7F,$8F,$1B
	.dc.b	$19,$9D,$BD,$DE,$38,$7A,$67,$6D,$DF,$FD,$D0,$E6,$77,$BF,$8E,$72
	.dc.b	$F8,$F0,$F8,$11,$8C,$3F,$A6,$61,$D2,$BF,$CC,$3F,$88,$FE,$1F,$AF
	.dc.b	$EC,$FE,$09,$FB,$5C,$FF,$79,$39,$C3,$F1,$DC,$0A,$D6,$1C,$B2,$ED
	.dc.b	$B3,$0E,$18,$F2,$E0,$CC,$62,$FC,$31,$8B,$CB,$73,$FC,$9F,$C1,$76
	.dc.b	$D3,$F7,$BF,$9F,$3E,$53,$BC,$E7,$9D,$39,$BE,$37,$A3,$CB,$F5,$8E
	.dc.b	$9C,$4E,$89,$3F,$9C,$9D,$46,$36,$FE,$3D,$18,$9F,$D0,$E3,$E5,$DD
	.dc.b	$39,$AF,$E4,$93,$9F,$C4,$BF,$C9,$18,$7F,$00,$CF,$F8,$D7,$A9,$F2
	.dc.b	$27,$39,$7E,$EE,$73,$5F,$B4,$9C,$BF,$EB,$8C,$91,$8A,$FF,$4D,$B5
	.dc.b	$FE,$6E,$D8,$DF,$D1,$77,$F8,$1F,$03,$6F,$F0,$3C,$E7,$FC,$25,$E5
	.dc.b	$5C,$7F,$DB,$3C,$DF,$CC,$27,$FB,$7F,$E2,$3F,$E7,$C7,$67,$F8,$67
	.dc.b	$A3,$F0,$D8,$71,$C3,$94,$AB,$86,$1F,$ED,$23,$C0,$E1,$FF,$0F,$B5
	.dc.b	$FF,$0F,$35,$F1,$FB,$5D,$67,$C1,$8F,$F9,$49,$F1,$CF,$F7,$3F,$C4
	.dc.b	$9E,$B6,$E3,$8C,$EB,$CE,$56,$A3,$0D,$9F,$EE,$DE,$FF,$1A,$DD,$F3
	.dc.b	$D7,$F4,$9F,$63,$72,$3D,$F2,$F8,$1E,$DE,$63,$1A,$FF,$98,$E7,$FB
	.dc.b	$D3,$73,$F8,$91,$C0,$CD,$7F,$BE,$70,$FD,$97,$CB,$77,$3B,$70,$B9
	.dc.b	$B1,$7F,$E2,$48,$C3,$FE,$07,$6E,$1D,$9A,$FF,$93,$9A,$FF,$95,$FE
	.dc.b	$67,$D9,$F4,$DE,$FF,$49,$18,$BF,$F2,$0B,$93,$5F,$37,$63,$E9,$E7
	.dc.b	$FC,$27,$F0,$2F,$8C,$7C,$2C,$2C,$FE,$31,$FC,$86,$3F,$82,$6B,$F8
	.dc.b	$26,$BF,$88,$7E,$27,$03,$EF,$7E,$91,$E7,$FA,$7B,$7F,$E5,$6F,$97
	.dc.b	$C3,$4F,$F2,$CC,$7E,$17,$B1,$FA,$1F,$39,$FC,$5C,$62,$F5,$F8,$9B
	.dc.b	$E8,$B8,$5E,$F4,$FF,$7A,$4F,$C6,$C3,$8B,$B6,$F0,$72,$BF,$CA,$57
	.dc.b	$D9,$E5,$BC,$9F,$83,$ED,$F8,$7E,$4F,$F8,$CB,$1F,$F5,$2D,$70,$8C
	.dc.b	$5F,$EA,$CF,$F0,$16,$BF,$19,$B5,$96,$9B,$FC,$3B,$FD,$A5,$1F,$A8
	.dc.b	$84,$61,$61,$F5,$3E,$15,$F0,$5E,$8A,$D4,$FF,$07,$C1,$EE,$FF,$8F
	.dc.b	$E3,$0D,$C7,$FD,$3D,$3C,$B6,$FA,$13,$9C,$BF,$C9,$F8,$BA,$3E,$A3
	.dc.b	$F2,$E3,$C7,$87,$28,$D3,$F5,$9D,$63,$FC,$67,$A1,$F8,$8F,$F5,$B8
	.dc.b	$C3,$6B,$F8,$C2,$F3,$BF,$78,$E7,$05,$F2,$9C,$EF,$A6,$DF,$C9,$F9
	.dc.b	$1F,$F7,$63,$7C,$3F,$8D,$AF,$DE,$9C,$E3,$EA,$3F,$D8,$1C,$1F,$22
	.dc.b	$FE,$B3,$63,$EB,$57,$DB,$8C,$47,$F2,$95,$F0,$FF,$4C,$FC,$5F,$E1
	.dc.b	$8F,$8E,$C6,$1F,$B2,$FC,$4F,$FB,$D3,$7B,$BF,$8D,$7E,$25,$FD,$77
	.dc.b	$8B,$E4,$56,$B3,$F3,$70,$A9,$EF,$F0,$F4,$F8,$B7,$2F,$D3,$FA,$AF
	.dc.b	$C5,$3F,$58,$D8,$FD,$5F,$E2,$7F,$37,$AF,$FB,$E5,$7E,$C6,$3E,$94
	.dc.b	$EB,$9F,$1C,$F9,$7F,$4F,$F1,$5E,$5E,$0F,$C8,$7C,$EB,$E2,$33,$C7
	.dc.b	$FB,$B8,$C6,$3F,$9D,$E7,$7C,$3F,$CB,$24,$7D,$07,$E7,$7C,$47,$E7
	.dc.b	$4F,$CF,$7E,$1D,$67,$F8,$1F,$C6,$FF,$67,$69,$FC,$27,$F1,$37,$E2
	.dc.b	$7F,$C2,$E7,$5D,$3F,$7F,$91,$5A,$BE,$27,$F0,$E7,$13,$BA,$E3,$BF
	.dc.b	$E6,$FF,$13,$FD,$FB,$F9,$DF,$9B,$F8,$F5,$F5,$A7,$38,$7F,$73,$9C
	.dc.b	$D7,$1C,$7F,$AC,$56,$AF,$F3,$AB,$EB,$F3,$3F,$3B,$E2,$66,$3F,$27
	.dc.b	$EF,$AE,$1F,$99,$3F,$83,$BF,$95,$24,$7E,$69,$FC,$C7,$1F,$F0,$73
	.dc.b	$F2,$BE,$85,$9F,$E8,$BF,$15,$FD,$E6,$FF,$1E,$FF,$6A,$8C,$FB,$DF
	.dc.b	$DB,$E3,$13,$F9,$4F,$A9,$74,$FF,$65,$6D,$FF,$23,$F0,$91,$8A,$FF
	.dc.b	$08,$46,$2B,$FB,$67,$E2,$4F,$A0,$FE,$F8,$DC,$3F,$C1,$7F,$0A,$FC
	.dc.b	$4F,$F8,$26,$0A,$D6,$3F,$A6,$FC,$4F,$CB,$CB,$FC,$45,$B8,$FF,$B6
	.dc.b	$BA,$BE,$DB,$FE,$27,$8C,$D7,$FB,$BB,$E5,$BF,$CE,$0F,$F8,$FD,$AF
	.dc.b	$F2,$86,$BB,$F0,$7F,$C0,$9F,$1A,$F9,$1F,$F0,$A7,$E2,$7F,$CA,$F5
	.dc.b	$FF,$7E,$C6,$2F,$FC,$31,$F0,$AF,$06,$F8,$58,$6E,$BF,$DE,$3F,$13
	.dc.b	$7C,$0F,$F8,$45,$B1,$F3,$5B,$87,$CC,$6A,$3F,$C8,$9F,$89,$7E,$C9
	.dc.b	$FE,$16,$DE,$2F,$F2,$77,$2B,$F1,$3F,$E4,$B1,$FC,$E6,$75,$87,$F9
	.dc.b	$37,$F1,$3F,$F0,$CC,$FF,$3D,$6C,$77,$B7,$E2,$6F,$C3,$FF,$C6,$D5
	.dc.b	$AC,$9E,$1B,$93,$FA,$64,$7F,$9B,$5A,$FF,$56,$6B,$C1,$BC,$7F,$D1
	.dc.b	$F7,$FF,$21,$79,$AF,$87,$B1,$87,$FC,$8D,$F8,$9F,$AB,$0F,$E9,$6D
	.dc.b	$D5,$F7,$DF,$8A,$E0,$FE,$B1,$3B,$CB,$E5,$65,$FE,$41,$C3,$F7,$E7
	.dc.b	$CD,$3F,$AA,$98,$3F,$CE,$45,$3F,$AC,$FE,$26,$FA,$5F,$F2,$C3,$D4
	.dc.b	$C7,$17,$F8,$D9,$F0,$6E,$F8,$BF,$F2,$F2,$F8,$BB,$18,$7F,$C5,$1A
	.dc.b	$3E,$3B,$FC,$6B,$5F,$E0,$86,$C6,$70,$FF,$0B,$4E,$5F,$F0,$C7,$C9
	.dc.b	$FF,$28,$F2,$BF,$CE,$CD,$8F,$F3,$B5,$7E,$11,$FC,$64,$67,$4F,$ED
	.dc.b	$4B,$F1,$2F,$95,$3A,$FC,$F3,$FA,$E3,$C2,$33,$4A,$56,$A3,$1F,$8A
	.dc.b	$3F,$0B,$FF,$27,$D2,$31,$87,$A1,$B5,$FE,$9B,$9D,$57,$09,$FC,$95
	.dc.b	$7F,$A5,$27,$97,$57,$BB,$F1,$4F,$99,$39,$AD,$E1,$FE,$05,$F9,$77
	.dc.b	$E2,$8F,$96,$FF,$49,$36,$FF,$D3,$4F,$2A,$E0,$E5,$BC,$5F,$E6,$8B
	.dc.b	$E6,$D9,$E1,$3F,$8B,$9C,$D7,$D4,$FD,$23,$F0,$B3,$F3,$9B,$C1,$FB
	.dc.b	$8F,$C5,$3F,$9C,$3F,$EC,$7D,$93,$AB,$FD,$A5,$69,$BE,$A7,$DE,$FC
	.dc.b	$53,$F9,$1E,$2E,$E5,$EE,$3E,$B7,$C6,$FE,$4F,$0F,$95,$F0,$DF,$DE
	.dc.b	$9F,$99,$19,$9F,$C1,$CE,$5F,$C2,$C6,$74,$F3,$36,$BF,$16,$FC,$98
	.dc.b	$CF,$57,$FA,$0F,$E0,$FF,$31,$7F,$DE,$7E,$19,$E9,$E7,$B7,$FC,$3A
	.dc.b	$DC,$BF,$D0,$F7,$C3,$1F,$F3,$23,$FC,$FB,$B7,$FC,$2D,$7D,$7C,$EF
	.dc.b	$D3,$CF,$F5,$CC,$BF,$DF,$CD,$7F,$CC,$ED,$5C,$8F,$9F,$F4,$65,$FC
	.dc.b	$BE,$1F,$5C,$FC,$3B,$FD,$B1,$5A,$BE,$33,$9B,$FF,$71,$4E,$6F,$A7
	.dc.b	$4F,$0E,$0E,$2F,$FA,$5B,$0F,$E7,$9F,$45,$FE,$8B,$69,$5C,$27,$FD
	.dc.b	$21,$F7,$3F,$C2,$FF,$14,$B6,$69,$2B,$FD,$35,$2F,$AF,$FF,$48,$C6
	.dc.b	$1A,$FF,$21,$11,$BE,$2E,$F3,$A9,$FF,$47,$FE,$28,$FA,$EF,$E5,$15
	.dc.b	$FE,$97,$8C,$4F,$FA,$F9,$AF,$F8,$DD,$FF,$4F,$B9,$9E,$3B,$74,$DE
	.dc.b	$CF,$F4,$CA,$FA,$2E,$CF,$F6,$4F,$E3,$1F,$D4,$DF,$E4,$2D,$7D,$5C
	.dc.b	$A8,$C4,$7F,$3D,$8C,$78,$9F,$EB,$97,$5F,$3F,$E0,$6C,$F6,$8C,$57
	.dc.b	$5C,$FF,$D6,$8A,$33,$E3,$4A,$E0,$7F,$15,$FB,$7F,$99,$7F,$BB,$6E
	.dc.b	$9F,$CC,$1B,$3F,$CA,$5F,$E5,$4F,$13,$E1,$F8,$68,$78,$58,$6C,$EC
	.dc.b	$FE,$49,$F8,$A7,$FA,$1D,$FF,$17,$FC,$4B,$EE,$FE,$6F,$88,$8C,$4F
	.dc.b	$0F,$94,$CC,$62,$3F,$D5,$93,$BA,$C9,$E4,$BE,$D3,$FD,$23,$F1,$3F
	.dc.b	$BA,$FF,$39,$6B,$FD,$B7,$34,$62,$3F,$A8,$B5,$FE,$94,$FF,$94,$9F
	.dc.b	$E7,$CD,$7B,$7C,$47,$C4,$8E,$FB,$7F,$27,$FC,$53,$FD,$0F,$5F,$D1
	.dc.b	$5A,$FF,$5C,$6C,$7F,$2D,$FD,$35,$C1,$F3,$9F,$F6,$C5,$7F,$6A,$6B
	.dc.b	$FD,$4D,$AF,$AD,$AF,$F6,$AF,$D5,$3E,$45,$7F,$39,$F8,$4F,$E3,$B3
	.dc.b	$FD,$5D,$AF,$F6,$E9,$DD,$FF,$8B,$FF,$57,$FE,$1D,$3F,$E6,$66,$BF
	.dc.b	$49,$B7,$FD,$17,$F1,$6F,$ED,$93,$FD,$8F,$F1,$5F,$F6,$5B,$FE,$02
	.dc.b	$6B,$FD,$AD,$AF,$D3,$6B,$FE,$4D,$F6,$D5,$FE,$31,$8F,$F2,$83,$FF
	.dc.b	$00,$47,$B1,$FB,$69,$DB,$63,$84,$FF,$B7,$BF,$16,$7E,$2D,$B5,$F6
	.dc.b	$74,$C4,$73,$FD,$2B,$FD,$DD,$5F,$E8,$F6,$BF,$E8,$E6,$BF,$E3,$7C
	.dc.b	$51,$88,$FF,$2C,$7E,$E1,$FE,$B1,$7F,$D1,$ED,$7F,$36,$DC,$7F,$03
	.dc.b	$0A,$98,$7B,$7F,$18,$FC,$0D,$6A,$50,$F8,$FD,$1B,$F8,$2F,$89,$F8
	.dc.b	$BF,$F9,$93,$1F,$F5,$43,$5F,$F5,$3B,$5F,$53,$5F,$F3,$87,$D3,$7F
	.dc.b	$20,$7F,$62,$DC,$7F,$CF,$AD,$E3,$FB,$1E,$0F,$FB,$DE,$BF,$D5,$BE
	.dc.b	$6B,$F0,$DF,$DF,$B6,$33,$1F,$67,$EA,$F9,$47,$C1,$FA,$3F,$F5,$03
	.dc.b	$FE,$BF,$6B,$FE,$BE,$6B,$F4,$5A,$FE,$27,$F0,$1F,$E9,$CF,$E2,$31
	.dc.b	$C6,$1F,$87,$9F,$F3,$14,$E4,$FA,$0F,$07,$F2,$33,$F9,$3F,$C5,$FD
	.dc.b	$11,$83,$EB,$3F,$CC,$5E,$5C,$28,$C4,$6E,$31,$97,$E5,$FF,$08,$FD
	.dc.b	$8C,$FE,$25,$AF,$E8,$5A,$FE,$0D,$BE,$2B,$F2,$FF,$E6,$ED,$9F,$E5
	.dc.b	$71,$FC,$EA,$DF,$8C,$7E,$23,$2F,$F5,$57,$EB,$FC,$09,$7F,$A9,$F7
	.dc.b	$BF,$D1,$4F,$AB,$F5,$7F,$B4,$FF,$BB,$1A,$FE,$7D,$AF,$ED,$DA,$FE
	.dc.b	$67,$F1,$8F,$EA,$D7,$FB,$B7,$F8,$FB,$7E,$AF,$F7,$11,$A9,$FF,$66
	.dc.b	$7E,$21,$FC,$B7,$9F,$F4,$70,$8C,$5F,$FA,$95,$B1,$FE,$D3,$F8,$5F
	.dc.b	$F3,$53,$F6,$C7,$03,$EA,$FF,$C3,$8D,$8F,$F0,$71,$FE,$DB,$70,$E1
	.dc.b	$AF,$F8,$4D,$7D,$27,$F8,$12,$3A,$B6,$33,$1E,$17,$9E,$3F,$B7,$FC
	.dc.b	$57,$FD,$5D,$3F,$86,$6B,$F8,$56,$C6,$FE,$3B,$F8,$E4,$63,$3F,$E4
	.dc.b	$8F,$F1,$AA,$E5,$E2,$6B,$E0,$BF,$91,$7D,$67,$F6,$2C,$7F,$83,$3F
	.dc.b	$CF,$2B,$F8,$63,$79,$4E,$F1,$FE,$60,$E4,$7E,$15,$D0,$EB,$EA,$FB
	.dc.b	$FF,$12,$F8,$1F,$86,$4C,$7E,$04,$57,$F4,$C7,$D2,$8C,$65,$F7,$37
	.dc.b	$57,$C9,$FC,$42,$5F,$C9,$77,$9A,$3F,$DF,$BE,$F7,$F8,$5B,$DF,$FF
	.dc.b	$16,$47,$F8,$90,$FF,$12,$FB,$3F,$E2,$9F,$C4,$7F,$83,$E7,$FC,$B2
	.dc.b	$90,$FB,$5D,$1B,$F5,$FF,$A5,$7E,$27,$F6,$B1,$FD,$C3,$F6,$B3,$FE
	.dc.b	$A8,$D8,$FE,$6C,$EF,$1F,$E6,$8B,$E4,$79,$92,$7F,$82,$C7,$17,$FC
	.dc.b	$55,$E5,$FE,$2F,$E0,$7F,$C5,$31,$F2,$1B,$7F,$CF,$FE,$E3,$FB,$BB
	.dc.b	$F2,$7C,$EF,$CD,$92,$BF,$36,$7E,$81,$FC,$FF,$E0,$3F,$95,$24,$7E
	.dc.b	$56,$3F,$B4,$53,$F0,$1B,$F0,$92,$BF,$8D,$CE,$5B,$1C,$3A,$24,$60
	.dc.b	$FE,$55,$C4,$FC,$4B,$E8,$F0,$7D,$98,$C7,$0F,$F9,$76,$73,$49,$7E
	.dc.b	$41,$F7,$67,$EB,$4F,$47,$FC,$13,$B7,$DB,$6E,$1D,$5D,$CF,$F3,$5B
	.dc.b	$F1,$77,$F7,$24,$7F,$36,$FA,$F7,$DA,$7E,$E7,$E9,$BF,$82,$92,$7E
	.dc.b	$5F,$6F,$F8,$29,$FE,$05,$F4,$FF,$BF,$23,$FB,$F9,$FF,$09,$5F,$D6
	.dc.b	$CF,$FC,$34,$D9,$FE,$F8,$DF,$08,$BE,$BB,$8F,$D7,$F2,$FA,$79,$7E
	.dc.b	$78,$FF,$38,$FC,$5E,$E3,$FA,$C7,$80,$8C,$47,$F5,$58,$C5,$F0,$AF
	.dc.b	$F1,$37,$DA,$BA,$F9,$FC,$E4,$C3,$FA,$CC,$E5,$BD,$9F,$8C,$D5,$3F
	.dc.b	$E1,$D9,$FF,$0B,$36,$33,$3B,$87,$89,$E0,$25,$74,$3B,$F8,$DE,$42
	.dc.b	$3E,$2F,$1A,$BF,$DA,$6B,$F5,$DC,$3F,$E4,$56,$C7,$8A,$BD,$BE,$6F
	.dc.b	$84,$91,$FC,$FB,$2F,$4C,$79,$7C,$4F,$86,$49,$F8,$31,$F1,$BF,$11
	.dc.b	$F5,$12,$39,$60,$8F,$77,$B3,$F3,$09,$3F,$2A,$3E,$7F,$C9,$7D,$0F
	.dc.b	$92,$93,$F3,$A2,$BE,$D8,$BD,$FF,$E3,$DC,$3F,$C1,$2D,$F9,$3F,$98
	.dc.b	$4A,$F1,$BF,$3F,$D6,$FB,$A4,$8F,$B7,$1F,$EC,$BF,$96,$FF,$38,$FC
	.dc.b	$D3,$EB,$3F,$69,$18,$6F,$D0,$FF,$2F,$BF,$F7,$F4,$6B,$C7,$C5,$F1
	.dc.b	$9F,$DC,$5F,$16,$F9,$FA,$BE,$AD,$8D,$E1,$7C,$6F,$F7,$C3,$6F,$F7
	.dc.b	$38,$75,$9F,$F7,$E5,$E6,$B8,$70,$7F,$97,$F0,$FE,$89,$7B,$8F,$F5
	.dc.b	$2B,$7E,$1F,$7F,$11,$C1,$95,$C3,$3F,$E2,$97,$C1,$B5,$A9,$F8,$27
	.dc.b	$F1,$59,$D3,$6B,$AB,$CC,$C1,$72,$FC,$34,$73,$C3,$CD,$7F,$07,$F0
	.dc.b	$0F,$F6,$1C,$7B,$FE,$FF,$FC,$3E,$DE,$AD,$70,$CC,$BA,$56,$B4,$FE
	.dc.b	$58,$D8,$FE,$55,$87,$49,$FE,$60,$D8,$FE,$5D,$3B,$CC,$D9,$D4,$BF
	.dc.b	$10,$FE,$4B,$5E,$D8,$C7,$95,$FE,$8B,$CF,$F0,$51,$89,$D5,$7F,$BE
	.dc.b	$E2,$D6,$F1,$FC,$64,$F4,$9F,$9A,$71,$2B,$C5,$B3,$FE,$93,$74,$7E
	.dc.b	$54,$A7,$47,$50,$FC,$45,$E2,$FF,$64,$8C,$36,$33,$5F,$E4,$29,$CB
	.dc.b	$63,$38,$FB,$A7,$2D,$DB,$FD,$6E,$E6,$77,$8F,$B9,$B7,$FE,$19,$BE
	.dc.b	$33,$F0,$AF,$8B,$BC,$3F,$11,$F9,$FF,$A7,$7B,$9E,$B1,$FE,$CD,$39
	.dc.b	$7F,$2B,$A2,$3F,$AD,$D7,$2C,$F8,$4E,$6E,$D7,$1D,$3F,$5C,$F0,$C3
	.dc.b	$9B,$9C,$BF,$BC,$39,$9D,$E1,$CA,$FA,$FE,$25,$FE,$D1,$CB,$F4,$0D
	.dc.b	$C7,$CA,$B8,$77,$AC,$AD,$BC,$A3,$15,$C2,$2E,$3F,$C0,$EF,$95,$F0
	.dc.b	$3B,$DE,$B3,$FF,$07,$4E,$B0,$E5,$5F,$B3,$FC,$47,$FC,$80,$FE,$D7
	.dc.b	$E9,$7F,$C9,$8E,$79,$BE,$3C,$62,$BE,$6D,$70,$7E,$7C,$67,$2F,$E2
	.dc.b	$D3,$97,$F8,$C4,$67,$1F,$F1,$04,$E5,$B7,$F5,$21,$D1,$63,$EF,$78
	.dc.b	$2F,$CD,$59,$CC,$FF,$D9,$51,$88,$FF,$67,$BE,$FE,$CD,$57,$19,$E1
	.dc.b	$18,$9F,$89,$5F,$CF,$6B,$B6,$DF,$EA,$E8,$FF,$37,$46,$2F,$FB,$45
	.dc.b	$70,$9C,$BF,$76,$FA,$6F,$DA,$B3,$C3,$6D,$87,$AE,$B0,$FB,$66,$E1
	.dc.b	$CE,$F8,$9E,$CA,$CF,$0C,$8F,$7F,$D0,$5C,$7E,$1B,$E9,$7B,$DA,$31
	.dc.b	$8F,$CB,$6C,$66,$BD,$BD,$DE,$23,$C7,$F7,$F6,$3E,$6B,$C0,$8F,$C4
	.dc.b	$79,$FE,$21,$EA,$FA,$9F,$5B,$67,$EA,$B8,$7E,$BF,$D5,$DC,$FE,$DF
	.dc.b	$9F,$F0,$8F,$87,$E7,$DA,$FB,$AF,$A0,$8F,$D4,$7D,$7E,$8F,$CA,$F3
	.dc.b	$FE,$5B,$67,$F0,$BF,$83,$FC,$E3,$E7,$F8,$36,$3F,$D5,$9C,$BF,$88
	.dc.b	$FC,$3D,$F2,$BF,$59,$7C,$3D,$8C,$39,$F8,$99,$7C,$8F,$C4,$D3,$DB
	.dc.b	$CE,$FF,$EC,$2E,$9F,$F0,$85,$7F,$BD,$5B,$19,$F9,$7F,$C4,$28,$CE
	.dc.b	$7F,$A1,$8C,$4F,$DF,$6D,$7F,$A8,$29,$39,$6C,$7F,$BF,$BC,$57,$E1
	.dc.b	$CF,$9B,$FC,$B4,$61,$FC,$AF,$B6,$E8,$CD,$6E,$FF,$D3,$AB,$F0,$DC
	.dc.b	$FE,$1D,$A6,$CE,$BF,$0E,$E2,$72,$FC,$8A,$C7,$C7,$DB,$F4,$BF,$9E
	.dc.b	$BE,$07,$F4,$AB,$E9,$2C,$61,$CF,$C0,$E2,$F9,$78,$AB,$FF,$3E,$46
	.dc.b	$1E,$8F,$CF,$E1,$F9,$BC,$5F,$EA,$F8,$F1,$64,$F9,$FF,$F7,$FC,$6C
	.dc.b	$CB,$88,$FE,$E0,$6F,$E4,$7E,$94,$66,$3F,$BA,$35,$FF,$25,$7C,$63
	.dc.b	$F1,$0F,$BA,$DB,$FF,$66,$F7,$E2,$7F,$82,$FE,$1F,$9B,$75,$7F,$1C
	.dc.b	$8C,$37,$EC,$37,$5C,$3E,$23,$8C,$5C,$7E,$2D,$A3,$EF,$D3,$96,$F3
	.dc.b	$61,$B7,$FE,$DA,$97,$37,$03,$8F,$A7,$67,$E2,$FC,$6F,$E6,$38,$0D
	.dc.b	$F8,$EC,$36,$7E,$5C,$6F,$87,$1C,$1E,$E9,$FF,$38,$5F,$FB,$CD,$D9
	.dc.b	$F4,$A7,$7C,$1F,$CF,$BF,$10,$FE,$81,$3F,$CE,$1F,$E5,$57,$FC,$C1
	.dc.b	$CB,$79,$3F,$98,$C7,$F8,$26,$B9,$7A,$78,$F4,$F2,$FF,$36,$F8,$CF
	.dc.b	$5E,$3F,$8E,$5A,$67,$5D,$36,$75,$1C,$79,$F1,$A7,$B4,$1F,$8F,$F2
	.dc.b	$7E,$41,$F2,$7A,$2D,$FE,$41,$FC,$69,$1F,$C8,$CF,$E5,$E3,$F8,$17
	.dc.b	$BC,$F8,$CF,$E1,$2D,$AF,$CE,$FC,$8F,$FB,$49,$FC,$FC,$66,$FC,$9D
	.dc.b	$CF,$5F,$FC,$8F,$5F,$AC,$FD,$E3,$FB,$AB,$FA,$E8,$CD,$70,$9C,$BC
	.dc.b	$BA,$1F,$0B,$FD,$41,$B7,$FE,$D1,$6C,$FF,$96,$70,$FC,$6D,$7F,$59
	.dc.b	$FA,$27,$D0,$3E,$17,$FB,$54,$61,$FE,$D3,$F8,$53,$E4,$33,$5B,$CF
	.dc.b	$F8,$65,$BE,$8D,$F1,$9C,$9B,$3F,$C2,$3F,$0A,$7C,$C6,$6B,$FB,$AF
	.dc.b	$E1,$8F,$8E,$FE,$D2,$DB,$FF,$75,$37,$83,$D4,$D7,$FC,$19,$E2,$7F
	.dc.b	$89,$FC,$0F,$F0,$9A,$F2,$BF,$C0,$2E,$6F,$FC,$1E,$E5,$FE,$37,$F7
	.dc.b	$4F,$80,$FF,$03,$AF,$81,$B1,$87,$FC,$17,$E3,$3E,$43,$FB,$FB,$6F
	.dc.b	$FE,$00,$37,$3F,$A0,$D7,$FD,$BD,$C4,$FC,$DF,$F8,$F6,$BF,$E2,$F8
	.dc.b	$C7,$85,$FE,$18,$FC,$25,$F0,$3F,$52,$F7,$97,$E7,$2D,$F8,$4B,$E5
	.dc.b	$7E,$CB,$67,$F9,$BF,$E1,$37,$4F,$D2,$5F,$0B,$FF,$12,$FE,$17,$FC
	.dc.b	$E2,$7F,$0C,$D7,$FC,$70,$D7,$FC,$6E,$D7,$FC,$5F,$E6,$3E,$AB,$FC
	.dc.b	$5E,$FD,$EC,$7F,$A3,$B9,$6F,$57,$E3,$BE,$C3,$EA,$5B,$24,$63,$C3
	.dc.b	$7E,$13,$F0,$B0,$FC,$7C,$E7,$E5,$FF,$D1,$58,$7F,$93,$5A,$FF,$93
	.dc.b	$1A,$FF,$91,$1B,$6F,$C2,$5F,$19,$FE,$46,$7F,$05,$87,$F5,$E7,$2D
	.dc.b	$EA,$FF,$25,$FB,$8F,$BE,$BE,$DE,$1B,$5F,$D9,$B3,$FF,$4A,$1F,$E6
	.dc.b	$12,$9F,$E6,$02,$9F,$E5,$82,$EC,$FD,$7F,$FC,$05,$1F,$61,$BD,$7F
	.dc.b	$E7,$5D,$BF,$71,$F0,$5F,$B8,$9C,$E1,$FC,$82,$5F,$9E,$BE,$13,$FC
	.dc.b	$CD,$0F,$C2,$BF,$45,$97,$F2,$78,$DC,$EA,$73,$97,$7F,$C2,$B9,$63
	.dc.b	$FD,$D2,$D1,$8C,$3F,$DC,$D1,$8C,$3F,$CC,$CD,$AE,$15,$F2,$5F,$F0
	.dc.b	$A4,$72,$D7,$CB,$3F,$E4,$49,$CB,$BB,$F3,$7E,$15,$FC,$31,$BE,$57
	.dc.b	$B2,$3F,$61,$97,$C5,$74,$DF,$13,$C7,$1F,$B3,$AF,$8C,$D7,$FD,$F2
	.dc.b	$DE,$07,$E0,$FF,$EB,$5B,$FD,$B3,$72,$F9,$DF,$95,$FF,$23,$C3,$F5
	.dc.b	$B3,$95,$AA,$75,$7C,$A7,$E9,$C6,$2F,$F7,$97,$C6,$DF,$65,$F3,$70
	.dc.b	$FC,$03,$FE,$43,$AF,$92,$E5,$FC,$03,$9C,$3E,$E3,$97,$C6,$DA,$EB
	.dc.b	$B7,$E1,$23,$5C,$AF,$DF,$9F,$33,$FE,$58,$5F,$33,$63,$0D,$CB,$FC
	.dc.b	$9D,$5C,$67,$87,$CD,$E7,$3F,$EE,$93,$D6,$3F,$D5,$4D,$AF,$F3,$46
	.dc.b	$1F,$EA,$7A,$FE,$F2,$7F,$7A,$9D,$38,$0B,$E6,$5E,$77,$F9,$B9,$FE
	.dc.b	$21,$3E,$07,$37,$E6,$73,$B7,$FB,$7E,$F9,$5F,$13,$97,$85,$6F,$FC
	.dc.b	$B3,$87,$F8,$33,$E8,$3F,$C0,$AD,$2B,$84,$FF,$82,$BF,$07,$FF,$4F
	.dc.b	$FE,$11,$7E,$70,$D2,$57,$F9,$82,$7F,$87,$67,$C7,$8B,$FC,$AB,$39
	.dc.b	$6B,$F0,$E7,$6F,$C5,$D1,$F8,$47,$F9,$4E,$3E,$AB,$5F,$A6,$DB,$F8
	.dc.b	$4F,$F8,$32,$37,$F2,$1F,$4A,$7E,$37,$D5,$7E,$0A,$7E,$FB,$5E,$73
	.dc.b	$BD,$7E,$C7,$D9,$BE,$AB,$FC,$57,$92,$31,$7F,$EA,$76,$BE,$96,$BF
	.dc.b	$82,$8E,$1D,$DD,$63,$F0,$33,$FE,$10,$7F,$22,$D7,$F1,$EF,$4C,$BE
	.dc.b	$F3,$97,$F0,$1F,$8F,$7F,$B4,$27,$F3,$2D,$78,$D7,$7E,$A7,$C4,$7F
	.dc.b	$86,$A7,$F8,$CB,$5F,$E5,$4D,$BF,$E6,$CF,$E5,$DB,$C9,$CF,$F2,$87
	.dc.b	$D6,$D3,$B5,$71,$DB,$FB,$F7,$57,$F9,$FB,$1F,$F1,$7B,$5F,$EF,$6D
	.dc.b	$C7,$FA,$27,$E1,$5F,$E6,$89,$FE,$5B,$A3,$8B,$D4,$D8,$CC,$7C,$BF
	.dc.b	$05,$87,$3C,$7F,$A5,$7E,$15,$FE,$2B,$BF,$E8,$0D,$7F,$9F,$B5,$FD
	.dc.b	$BE,$E8,$C4,$7F,$53,$FC,$A3,$F1,$D5,$79,$7E,$64,$62,$7F,$93,$D7
	.dc.b	$5D,$7F,$AF,$7E,$14,$F5,$3E,$93,$6B,$FA,$4B,$5F,$EB,$5F,$85,$FF
	.dc.b	$96,$DF,$E9,$BF,$85,$FC,$77,$CA,$D7,$FB,$33,$5F,$E6,$ED,$D1,$F5
	.dc.b	$BF,$E4,$59,$FE,$AE,$BE,$AF,$F0,$8D,$8C,$FA,$7F,$87,$FC,$2B,$FC
	.dc.b	$8D,$1F,$D7,$9A,$FF,$6F,$6E,$9F,$E4,$1F,$C2,$DF,$4D,$F8,$FF,$C2
	.dc.b	$DF,$61,$FE,$92,$D1,$18,$C7,$FB,$93,$5F,$EE,$2D,$7F,$94,$FB,$7F
	.dc.b	$ED,$CF,$A7,$FD,$43,$FE,$DE,$FA,$7B,$99,$B3,$AF,$C2,$DF,$57,$7E
	.dc.b	$5E,$63,$FE,$AC,$C1,$D5,$FE,$AE,$C9,$87,$0A,$D4,$E7,$F0,$B7,$DD
	.dc.b	$7F,$AB,$2D,$18,$AF,$F7,$FC,$62,$75,$39,$9F,$F4,$D4,$FF,$15,$7F
	.dc.b	$D5,$6D,$C3,$C7,$08,$98,$FE,$77,$E0,$DF,$8A,$FC,$2F,$C3,$1F,$21
	.dc.b	$7E,$9A,$63,$E8,$39,$4E,$9C,$62,$FC,$33,$9C,$7A,$1A,$FF,$AB,$DA
	.dc.b	$FD,$26,$BF,$1F,$CB,$7D,$1F,$D4,$6C,$76,$6B,$FE,$78,$FA,$3F,$D3
	.dc.b	$5B,$EC,$5F,$59,$F7,$1B,$5E,$66,$BF,$E8,$0F,$63,$FD,$0C,$FA,$B5
	.dc.b	$FE,$B3,$7E,$D8,$FF,$4C,$9F,$56,$33,$F8,$62,$BF,$12,$FF,$25,$E2
	.dc.b	$F9,$F6,$8C,$72,$FE,$0D,$E1,$87,$C8,$FC,$33,$F9,$94,$FE,$1D,$FE
	.dc.b	$64,$DF,$3B,$F2,$B1,$A9,$FB,$5F,$80,$7E,$B5,$FF,$6D,$35,$FC,$EB
	.dc.b	$5F,$C9,$35,$FC,$17,$E1,$9E,$DC,$BC,$ED,$C3,$F3,$0D,$9F,$F4,$97
	.dc.b	$D0,$FF,$37,$9F,$CE,$7C,$9F,$F8,$36,$7F,$02,$D7,$FD,$F6,$DB,$FF
	.dc.b	$5D,$7D,$17,$19,$FC,$97,$E1,$9F,$69,$FD,$9B,$5F,$D4,$35,$DF,$2F
	.dc.b	$E9,$7F,$0C,$FE,$35,$0B,$F9,$FF,$BD,$7E,$7A,$31,$97,$EC,$63,$13
	.dc.b	$F5,$2B,$8B,$BD,$7F,$A7,$DF,$E7,$43,$87,$98,$33,$EC,$FF,$6F,$5F
	.dc.b	$86,$FF,$35,$C7,$FC,$6B,$4F,$C3,$7F,$C4,$6D,$AF,$F1,$B1,$CD,$C3
	.dc.b	$86,$BF,$E3,$BF,$8B,$FE,$10,$F8,$8F,$F3,$0F,$8B,$E1,$8C,$E5,$EB
	.dc.b	$BF,$E3,$7F,$86,$BE,$F2,$93,$0F,$DB,$C7,$E5,$1F,$E0,$8E,$DF,$DD
	.dc.b	$FE,$48,$FB,$5F,$E4,$AD,$8E,$35,$FC,$A3,$F0,$EF,$EC,$91,$FC,$35
	.dc.b	$7D,$36,$4D,$7D,$AD,$9F,$E0,$F3,$A6,$CE,$9F,$A5,$B3,$F0,$9F,$9C
	.dc.b	$D8,$CC,$63,$EB,$F3,$E6,$FC,$E2,$4F,$CE,$3E,$DB,$FE,$2E,$E7,$FE
	.dc.b	$51,$C1,$FE,$5D,$FD,$1D,$F8,$17,$F8,$A7,$E2,$D1,$8C,$3F,$CF,$37
	.dc.b	$C3,$9B,$FC,$E1,$F8,$7F,$F6,$29,$FF,$3E,$BF,$D3,$F1,$FF,$34,$36
	.dc.b	$77,$EE,$75,$F8,$BF,$E2,$D7,$FE,$2C,$F2,$70,$78,$6F,$84,$7F,$37
	.dc.b	$FC,$3F,$F2,$58,$FF,$0C,$7F,$C2,$D1,$F9,$4F,$21,$18,$9D,$77,$7F
	.dc.b	$5C,$F9,$0F,$45,$FF,$41,$6D,$9B,$7C,$74,$FC,$CF,$9B,$F4,$09,$7F
	.dc.b	$A5,$F8,$CF,$F1,$93,$F9,$9F,$67,$F3,$A4,$7E,$87,$3F,$F6,$97,$E9
	.dc.b	$7C,$7F,$89,$FF,$30,$73,$7E,$A6,$31,$3B,$9F,$67,$E1,$FF,$E3,$76
	.dc.b	$F6,$7F,$76,$8F,$C7,$EC,$75,$C1,$FC,$C1,$C9,$27,$87,$DE,$7D,$DF
	.dc.b	$C3,$BE,$6F,$49,$7D,$AE,$4D,$7C,$0F,$E0,$FE,$D1,$F8,$0D,$26,$BF
	.dc.b	$EB,$CB,$E2,$FE,$63,$EA,$3F,$2E,$47,$5E,$27,$E6,$BC,$6F,$E0,$64
	.dc.b	$8F,$8E,$FF,$A0,$AB,$F2,$11,$FE,$C4,$6E,$FF,$EB,$36,$F3,$FF,$0E
	.dc.b	$AF,$F0,$CF,$E0,$AF,$AC,$BF,$67,$31,$FE,$EB,$39,$7F,$BB,$FE,$0D
	.dc.b	$FC,$23,$0F,$F0,$B3,$63,$FC,$0C,$DE,$55,$2F,$FC,$4F,$F4,$5F,$5F
	.dc.b	$4F,$F0,$6F,$C4,$B8,$7F,$A0,$67,$FC,$9E,$31,$F4,$3C,$3B,$3F,$D1
	.dc.b	$15,$CF,$5F,$F0,$65,$71,$CB,$26,$AF,$BF,$67,$F8,$97,$3F,$F1,$05
	.dc.b	$FF,$8E,$9A,$FF,$8B,$DA,$FF,$87,$FF,$08,$7E,$07,$FE,$3E,$6C,$7F
	.dc.b	$86,$1A,$FE,$AE,$7F,$4A,$7F,$34,$D7,$FB,$DB,$75,$ED,$E8,$FC,$2F
	.dc.b	$37,$E1,$BD,$7B,$26,$1F,$E2,$4B,$E0,$FF,$8C,$5F,$DC,$BE,$9F,$9B
	.dc.b	$F6,$11,$FF,$1F,$E5,$F9,$E2,$77,$7D,$09,$FA,$3F,$67,$F0,$C9,$5F
	.dc.b	$E4,$6B,$FF,$0D,$BF,$D5,$CF,$F2,$95,$FF,$2F,$7F,$84,$37,$4F,$F0
	.dc.b	$BB,$74,$F9,$57,$F7,$BD,$F9,$7E,$B7,$F0,$3E,$1F,$F9,$A3,$F6,$C7
	.dc.b	$85,$49,$5B,$3E,$17,$DF,$7D,$92,$3E,$77,$FC,$C1,$3F,$E4,$6C,$FF
	.dc.b	$C5,$2D,$CB,$FC,$D4,$DA,$F9,$55,$FC,$DF,$E3,$3F,$63,$0F,$A5,$7D
	.dc.b	$DF,$F7,$7C,$95,$FE,$83,$9C,$D7,$A2,$31,$5B,$79,$CE,$AB,$8F,$8B
	.dc.b	$F2,$E3,$13,$C2,$F9,$4E,$6F,$97,$8D,$67,$27,$F8,$26,$F8,$D6,$70
	.dc.b	$EB,$39,$BE,$2D,$3E,$6F,$E1,$FF,$EB,$BB,$F2,$36,$7D,$CD,$9D,$FC
	.dc.b	$47,$58,$C5,$9A,$48,$D7,$CE,$4A,$F0,$56,$E3,$94,$E6,$7E,$75,$FF
	.dc.b	$08,$F5,$3D,$DE,$1B,$C4,$FE,$C9,$5A,$3D,$38,$7B,$FF,$04,$FF,$1C
	.dc.b	$E9,$F3,$7F,$04,$FF,$1C,$37,$8F,$5C,$53,$0F,$1D,$6B,$4F,$B8,$D8
	.dc.b	$FC,$8E,$1D,$27,$37,$F3,$6B,$5B,$1B,$5B,$BF,$F1,$E7,$E1,$EF,$C1
	.dc.b	$7F,$37,$1E,$91,$19,$BF,$A9,$18,$9D,$E7,$F8,$C8,$DF,$D2,$7E,$05
	.dc.b	$D5,$70,$BE,$51,$D5,$B5,$C5,$57,$17,$73,$FB,$B3,$99,$FE,$4A,$72
	.dc.b	$FE,$FD,$AF,$B3,$E4,$DE,$3F,$F8,$8E,$51,$89,$E4,$D7,$B4,$6A,$78
	.dc.b	$4E,$EF,$CC,$DD,$7C,$1C,$7F,$EC,$E9,$CD,$7B,$EB,$58,$78,$E3,$E1
	.dc.b	$56,$EF,$D1,$5A,$C9,$F8,$7F,$D3,$DF,$F2,$6D,$C7,$F1,$93,$AA,$E5
	.dc.b	$1F,$C4,$A7,$59,$7E,$E5,$F7,$DF,$2D,$77,$87,$18,$C1,$DB,$0C,$D7
	.dc.b	$2C,$7F,$85,$3C,$B1,$F0,$B9,$7F,$90,$FE,$1F,$FE,$E8,$BF,$DC,$72
	.dc.b	$3C,$0F,$F4,$44,$7E,$B2,$75,$5F,$A5,$AD,$3E,$48,$CE,$9F,$C4,$5B
	.dc.b	$8F,$F1,$B5,$C9,$BD,$A6,$3D,$5E,$51,$8C,$3E,$3D,$72,$7C,$F3,$AA
	.dc.b	$E0,$7F,$00,$FC,$3F,$CF,$3F,$39,$B8,$FF,$23,$FC,$3D,$D5,$F2,$A7
	.dc.b	$39,$FF,$2F,$6D,$FF,$44,$6F,$02,$F8,$E3,$DE,$BE,$9C,$61,$F0,$18
	.dc.b	$C3,$B3,$F7,$B5,$76,$FF,$BF,$EF,$FD,$69,$DD,$FD,$23,$DD,$47,$F7
	.dc.b	$D2,$DF,$F9,$9E,$B8,$3F,$D8,$A3,$FD,$90,$FF,$66,$6B,$FE,$09,$6D
	.dc.b	$BC,$3E,$B1,$FE,$BA,$8F,$1E,$BD,$B1,$FE,$B2,$D9,$FE,$E3,$BD,$C1
	.dc.b	$7C,$27,$FC,$25,$39,$8E,$CD,$C3,$F2,$18,$7F,$76,$8D,$62,$FC,$22
	.dc.b	$9B,$E6,$B8,$FF,$87,$5F,$4E,$1F,$A9,$CB,$FC,$09,$97,$AE,$72,$7B
	.dc.b	$B0,$FF,$2F,$DF,$CC,$3F,$B8,$FA,$DE,$3E,$5F,$0B,$63,$FC,$2F,$1D
	.dc.b	$88,$FF,$86,$BE,$7F,$CF,$F7,$BE,$B6,$C7,$F8,$C5,$C3,$FE,$33,$FC
	.dc.b	$16,$E7,$F5,$FE,$EF,$DE,$7C,$3F,$A6,$B5,$FE,$39,$7D,$A4,$7F,$C5
	.dc.b	$2F,$AF,$EC,$BE,$39,$F2,$FD,$3B,$3F,$E4,$1F,$53,$0D,$3F,$C8,$91
	.dc.b	$F4,$F4,$7D,$1D,$3E,$B7,$8E,$FC,$3F,$17,$FE,$6C,$F9,$CC,$C6,$9B
	.dc.b	$1F,$2F,$5F,$F1,$3F,$E1,$FF,$E1,$48,$C7,$72,$31,$3F,$EF,$FE,$9F
	.dc.b	$F0,$DD,$6A,$73,$D5,$F9,$CA,$56,$9F,$E8,$C6,$3A,$FF,$36,$DB,$FF
	.dc.b	$72,$4E,$E3,$FA,$63,$FA,$4A,$E8,$D9,$FF,$15,$5B,$E4,$FF,$35,$7E
	.dc.b	$0E,$DF,$CE,$64,$FC,$7B,$E3,$E1,$EC,$6D,$32,$FD,$9D,$7B,$F9,$78
	.dc.b	$D6,$A7,$39,$3A,$3A,$AA,$B1,$87,$F5,$1F,$88,$71,$C9,$97,$EB,$A3
	.dc.b	$0D,$F8,$3D,$F1,$F0,$6D,$7D,$5E,$E5,$A7,$37,$E9,$AF,$F3,$67,$A6
	.dc.b	$F7,$BE,$1B,$63,$EF,$4F,$D4,$DF,$35,$F2,$38,$BF,$CC,$F5,$F4,$A7
	.dc.b	$EA,$C6,$2B,$EC,$B5,$FF,$00,$7E,$1E,$F9,$2F,$A2,$DB,$EF,$EF,$3E
	.dc.b	$87,$35,$FC,$2F,$E1,$FF,$C1,$75,$ED,$6E,$1E,$B3,$F8,$9F,$E2,$1C
	.dc.b	$67,$5E,$9F,$F1,$38,$7D,$9D,$C7,$0D,$AF,$CC,$FA,$BF,$D2,$3A,$3E
	.dc.b	$EB,$63,$FC,$E5,$5F,$6D,$DE,$3F,$88,$8F,$E1,$EF,$8F,$7F,$57,$3E
	.dc.b	$7C,$7D,$3F,$8E,$FC,$D5,$E4,$7E,$43,$F2,$F7,$C2,$66,$30,$FC,$6F
	.dc.b	$82,$FD,$0F,$29,$A4,$D7,$F4,$FA,$25,$F0,$7F,$38,$F9,$FC,$D6,$BF
	.dc.b	$0D,$F8,$97,$F7,$AE,$2F,$E9,$BD,$A8,$C4,$FE,$4A,$31,$EB,$FF,$3D
	.dc.b	$9C,$E5,$FD,$09,$FE,$F7,$5D,$3A,$B5,$5B,$7F,$A2,$7C,$57,$92,$7F
	.dc.b	$2D,$F3,$3F,$D4,$B0,$FB,$58,$AB,$51,$89,$D6,$5E,$69,$CB,$74,$EB
	.dc.b	$B7,$67,$F7,$BF,$39,$FC,$07,$0F,$D5,$B5,$F8,$DD,$1A,$F1,$BF,$26
	.dc.b	$FC,$E2,$69,$F3,$38,$51,$F4,$BE,$5F,$F3,$27,$F0,$AF,$80,$7C,$96
	.dc.b	$2F,$F6,$5F,$89,$3E,$13,$FB,$27,$B1,$F8,$BD,$3F,$B4,$CE,$6B,$F7
	.dc.b	$13,$9E,$3F,$EC,$9E,$6F,$F7,$8F,$6B,$F8,$C5,$7F,$77,$6E,$1F,$E6
	.dc.b	$BA,$DC,$FF,$7B,$FC,$4B,$F1,$6F,$F0,$36,$CF,$E9,$FE,$E5,$F4,$8F
	.dc.b	$87,$FE,$F1,$5A,$E7,$FC,$1F,$E2,$5F,$E7,$0C,$7F,$86,$37,$1E,$5F
	.dc.b	$4E,$F8,$DF,$F0,$04,$E6,$FF,$17,$F8,$93,$E4,$FF,$C2,$2D,$C7,$FC
	.dc.b	$20,$DC,$7D,$4D,$7F,$BE,$FB,$19,$BF,$CA,$AF,$8E,$FF,$07,$39,$BF
	.dc.b	$CB,$39,$7F,$91,$7E,$02,$F8,$3F,$F0,$A3,$FB,$08,$CC,$63,$C8,$FF
	.dc.b	$0C,$78,$9E,$B9,$FF,$12,$37,$2F,$9C,$D7,$FC,$22,$D7,$F4,$DF,$89
	.dc.b	$3E,$5F,$FC,$44,$DA,$FE,$D1,$CA,$FC,$4B,$F9,$05,$FE,$82,$B4,$FE
	.dc.b	$8B,$F1,$1F,$F6,$F5,$FF,$3B,$6C,$7D,$9F,$C4,$9F,$2D,$FE,$2F,$9C
	.dc.b	$F0,$FE,$EB,$F1,$2F,$D5,$61,$FE,$3D,$6E,$1F,$87,$6B,$FE,$2A,$6B
	.dc.b	$FB,$FF,$31,$F6,$5F,$E3,$B5,$C5,$63,$0F,$F8,$FF,$F1,$2F,$4E,$3F
	.dc.b	$C2,$9B,$8F,$F5,$AD,$DF,$6F,$FE,$65,$AD,$70,$FF,$91,$FE,$5F,$FD
	.dc.b	$29,$A7,$F8,$F1,$AF,$F1,$16,$BF,$E5,$06,$F8,$0F,$C4,$9F,$7D,$FE
	.dc.b	$51,$5F,$21,$63,$0F,$F9,$57,$DB,$FE,$75,$87,$F1,$07,$FB,$A6,$BF
	.dc.b	$DA,$5C,$E7,$FC,$91,$CC,$FF,$16,$3F,$8F,$9F,$B1,$29,$FE,$54,$29
	.dc.b	$FC,$A4,$A7,$FB,$6B,$C3,$FA,$73,$FE,$7C,$CF,$F6,$5E,$9F,$F9,$B7
	.dc.b	$77,$E2,$5F,$BE,$AD,$5F,$0C,$BF,$C0,$B8,$3F,$11,$FD,$16,$5F,$EE
	.dc.b	$67,$F4,$4F,$F6,$58,$D7,$85,$FE,$79,$78,$78,$37,$F1,$2F,$6D,$7F
	.dc.b	$84,$0F,$F7,$7B,$87,$BB,$5C,$FC,$57,$FA,$3A,$9F,$89,$7C,$6D,$3D
	.dc.b	$8D,$CF,$CE,$BE,$53,$C4,$E7,$0F,$F7,$EF,$E3,$0F,$A3,$F3,$36,$FF
	.dc.b	$CA,$18,$BF,$12,$FD,$A6,$5F,$26,$B5,$87,$BE,$7F,$68,$AB,$87,$97
	.dc.b	$E7,$9C,$C6,$30,$FF,$27,$AF,$C5,$FF,$33,$7C,$5B,$C8,$FC,$4B,$F9
	.dc.b	$D5,$6F,$B3,$EC,$47,$F3,$D9,$D5,$FC,$FF,$C4,$FF,$97,$63,$FC,$0A
	.dc.b	$B8,$FB,$BF,$8E,$8F,$B7,$C9,$F9,$87,$DF,$19,$BF,$54,$E5,$FB,$91
	.dc.b	$9C,$8D,$3E,$63,$8F,$86,$FF,$2E,$FE,$24,$FC,$4F,$FC,$BF,$F6,$9F
	.dc.b	$8D,$CB,$FC,$9F,$DC,$AD,$47,$F9,$F2,$75,$92,$B5,$7D,$0E,$BF,$37
	.dc.b	$6F,$FD,$07,$82,$3F,$DA,$75,$D5,$B1,$99,$FF,$3F,$CE,$5B,$F3,$8F
	.dc.b	$BE,$FF,$3A,$FD,$5F,$EF,$A7,$FD,$1F,$39,$7F,$D2,$31,$9C,$BF,$C6
	.dc.b	$53,$96,$D7,$F9,$FA,$3F,$8D,$D7,$28,$FF,$41,$CF,$F9,$C3,$E9,$7F
	.dc.b	$C5,$6D,$2B,$84,$FF,$8C,$3E,$CD,$F3,$CF,$C4,$F0,$DA,$FF,$4D,$CE
	.dc.b	$A7,$27,$1E,$FF,$F4,$64,$E9,$AF,$EC,$C4,$FE,$A7,$B7,$81,$F2,$3F
	.dc.b	$30,$7D,$7F,$F9,$0B,$74,$62,$F5,$39,$8C,$47,$5E,$5F,$F5,$07,$D1
	.dc.b	$7F,$92,$9B,$1F,$A4,$FA,$2F,$F6,$33,$67,$B7,$DA,$3E,$1A,$92,$7E
	.dc.b	$DC,$65,$BF,$5B,$C0,$8E,$0D,$FB,$BD,$08,$E0,$FD,$AC,$FF,$52,$6F
	.dc.b	$AB,$E7,$8C,$47,$AD,$AF,$F1,$BF,$05,$FA,$3F,$E4,$E3,$FD,$9B,$18
	.dc.b	$EB,$FF,$34,$B6,$33,$3B,$7E,$37,$E2,$5F,$E7,$17,$F8,$3B,$5C,$FC
	.dc.b	$9F,$C9,$8C,$47,$D9,$E5,$FD,$83,$FE,$76,$6B,$FE,$DE,$6B,$BE,$EF
	.dc.b	$C8,$FE,$A5,$F5,$A3,$F0,$4D,$7F,$9B,$35,$FF,$63,$B7,$A1,$FA,$77
	.dc.b	$FA,$F6,$7F,$DD,$71,$8F,$A3,$FF,$43,$B6,$33,$3B,$7F,$21,$1F,$9E
	.dc.b	$FB,$7D,$F6,$9A,$49,$FF,$48,$E1,$FE,$62,$6F,$C8,$FA,$88,$FE,$FB
	.dc.b	$E3,$FF,$85,$37,$B3,$F3,$11,$88,$FD,$04,$7F,$BD,$CF,$CA,$7E,$1F
	.dc.b	$FE,$57,$8F,$E6,$ED,$7F,$34,$D7,$FA,$D7,$5A,$31,$1F,$E1,$3F,$C3
	.dc.b	$FF,$C6,$CF,$E0,$9A,$FE,$BF,$91,$18,$8F,$C5,$35,$FF,$0E,$7E,$DF
	.dc.b	$E7,$7F,$1C,$D7,$38,$7E,$32,$31,$0A,$7E,$EB,$FE,$D4,$8F,$C9,$B5
	.dc.b	$F1,$35,$FC,$D3,$5F,$F1,$63,$FE,$E7,$78,$FD,$1F,$E6,$A3,$11,$CA
	.dc.b	$3F,$3B,$F0,$4F,$8E,$A4,$9F,$D1,$C6,$5F,$D2,$FE,$72,$FC,$0B,$F4
	.dc.b	$AD,$8C,$C6,$0F,$D4,$FE,$80,$E0,$A4,$9F,$D7,$C6,$5F,$D8,$F0,$FF
	.dc.b	$9E,$DB,$E9,$7E,$CA,$31,$1F,$4D,$AF,$EE,$FC,$A7,$E3,$9F,$B8,$6C
	.dc.b	$7E,$F1,$AF,$DB,$6B,$FC,$0F,$E4,$5E,$2F,$F0,$06,$C7,$F0,$46,$BF
	.dc.b	$83,$6B,$FC,$2F,$F7,$C7,$D7,$7F,$08,$6C,$7F,$0C,$6B,$F8,$B6,$BF
	.dc.b	$C4,$FF,$4C,$FF,$54,$3F,$C4,$9A,$FE,$45,$AE,$FD,$DF,$E2,$BE,$5B
	.dc.b	$F0,$7A,$49,$FE,$37,$19,$7F,$8E,$FC,$CF,$F2,$27,$FC,$E4,$D7,$F2
	.dc.b	$4D,$E8,$37,$E6,$E7,$99,$FC,$91,$B5,$FC,$A3,$EA,$3F,$04,$FE,$9D
	.dc.b	$AF,$F2,$F6,$BF,$E8,$16,$BF,$A1,$7F,$04,$F1,$E5,$FE,$63,$18,$8E
	.dc.b	$51,$FC,$5B,$E3,$9F,$61,$49,$3F,$C7,$23,$2F,$F1,$2F,$B1,$FE,$52
	.dc.b	$47,$07,$F6,$B0,$3F,$77,$FE,$30,$E5,$AF,$F7,$16,$BF,$E6,$56,$AF
	.dc.b	$92,$7E,$15,$E9,$8E,$B1,$8C,$7C,$11,$B9,$D3,$63,$33,$CB,$F0,$2B
	.dc.b	$46,$92,$BD,$52,$FB,$0F,$E1,$AD,$F1,$79,$C6,$09,$F8,$A6,$09,$5C
	.dc.b	$A1,$1F,$E0,$9F,$AF,$F6,$D1,$A4,$EE,$F5,$FE,$42,$FD,$5B,$FC,$0B
	.dc.b	$8F,$F8,$5E,$31,$CF,$A9,$CC,$FF,$A3,$67,$2F,$F9,$02,$FF,$A8,$3F
	.dc.b	$E9,$1C,$3C,$1E,$97,$8C,$91,$F0,$3E,$DB,$F3,$CF,$F8,$97,$54,$62
	.dc.b	$3F,$D3,$51,$8E,$AF,$9B,$39,$9F,$AA,$FF,$A3,$DE,$9F,$21,$A2,$3F
	.dc.b	$3B,$F0,$97,$E0,$DF,$59,$B1,$98,$C1,$9F,$A2,$91,$C5,$AF,$E4,$7F
	.dc.b	$24,$7C,$C7,$F9,$39,$B1,$F8,$66,$BB,$8F,$F0,$1F,$74,$8F,$ED,$ED
	.dc.b	$7F,$39,$EA,$FF,$BE,$1F,$EE,$7F,$41,$FE,$0A,$23,$F8,$8F,$BA,$FD
	.dc.b	$3B,$F2,$5A,$FC,$86,$BF,$9A,$6A,$F8,$B7,$EA,$DF,$EB,$B8,$FD,$64
	.dc.b	$63,$DF,$FC,$FB,$63,$33,$B7,$FB,$EF,$E8,$CF,$C1,$3F,$CB,$0D,$8F
	.dc.b	$F4,$1B,$5D,$C6,$B4,$91,$A6,$BF,$E1,$8F,$0F,$FC,$70,$FF,$B6,$FD
	.dc.b	$F4,$62,$3F,$80,$C6,$3E,$17,$F8,$71,$FE,$80,$FC,$5B,$FB,$4C,$FF
	.dc.b	$9C,$5A,$FF,$B9,$5A,$FF,$0A,$6C,$BF,$30,$7C,$AF,$F8,$B9,$B5,$FE
	.dc.b	$32,$6B,$FE,$2E,$6B,$F9,$8F,$C3,$1F,$27,$FC,$11,$B1,$FC,$01,$AF
	.dc.b	$E8,$5A,$FF,$BE,$77,$7E,$66,$FA,$FF,$A6,$D9,$FA,$ED,$7F,$D1,$D3
	.dc.b	$C2,$7F,$89,$E9,$FC,$7B,$BB,$FD,$75,$E2,$E0,$97,$FC,$4A,$7F,$CA
	.dc.b	$53,$FE,$BF,$F9,$2E,$24,$9F,$F3,$7F,$91,$FD,$89,$FE,$63,$E9,$7F
	.dc.b	$19,$F5,$BF,$CE,$33,$99,$FF,$13,$37,$E1,$BF,$4F,$F7,$1F,$E9,$5E
	.dc.b	$63,$49,$5F,$C9,$3F,$07,$FF,$59,$B7,$DC,$FE,$5B,$18,$8F,$E1,$0D
	.dc.b	$7F,$D0,$DF,$9A,$7F,$B0,$33,$34,$91,$FE,$86,$7F,$5C,$FF,$91,$DA
	.dc.b	$F2,$EF,$D1,$E0,$F3,$A4,$79,$5A,$F8,$BE,$07,$CA,$47,$D4,$FA,$3E
	.dc.b	$2A,$47,$B5,$AF,$AB,$D4,$F6,$91,$F8,$4F,$C0,$F4,$E4,$7C,$56,$BF
	.dc.b	$0B,$F0,$2F,$8A,$47,$E5,$3F,$23,$E1,$E4,$7C,$D6,$BF,$DA,$FF,$0E
	.dc.b	$FF,$61,$7A,$5F,$EE,$9F,$88,$FE,$DB,$29,$CB,$FE,$E8,$8C,$C6,$3E
	.dc.b	$37,$FC,$A1,$3B,$F4,$B8,$C6,$27,$55,$C3,$E5,$7F,$D1,$33,$96,$C7
	.dc.b	$4A,$E2,$EF,$0F,$26,$1F,$E0,$0C,$39,$B6,$F8,$B9,$BE,$2E,$6B,$8B
	.dc.b	$EB,$9D,$B6,$33,$5F,$D4,$A2,$CE,$A3,$BC,$6C,$F8,$47,$73,$73,$1C
	.dc.b	$9B,$C3,$E1,$5F,$92,$31,$87,$B5,$DD,$F6,$3C,$4E,$30,$F1,$13,$0F
	.dc.b	$3D,$F3,$AE,$13,$9A,$E7,$39,$8C,$61,$F3,$EF,$8D,$79,$E3,$0F,$78
	.dc.b	$CB,$49,$1A,$3D,$6F,$BC,$FC,$39,$F1,$4B,$7F,$88,$6D,$7F,$AA,$A3
	.dc.b	$FC,$F3,$CF,$FE,$2C,$6C,$FE,$4F,$E1,$53,$96,$C6,$63,$EB,$C6,$2F
	.dc.b	$F1,$DF,$2B,$FD,$76,$BE,$5C,$E7,$3F,$CB,$B6,$3F,$14,$D7,$F3,$D3
	.dc.b	$C6,$BF,$31,$39,$C3,$CB,$18,$C3,$F0,$F5,$BC,$3A,$C7,$90,$98,$7D
	.dc.b	$C9,$D3,$E0,$8C,$D7,$38,$D7,$83,$8B,$E7,$1E,$AA,$EF,$39,$8E,$46
	.dc.b	$4E,$25,$AE,$84,$AF,$DB,$DF,$8A,$7F,$79,$8F,$B4,$98,$76,$8E,$55
	.dc.b	$E2,$BE,$51,$FB,$1B,$E5,$8F,$C4,$AD,$12,$75,$8F,$9D,$EF,$85,$A5
	.dc.b	$F2,$C7,$EA,$77,$3A,$3F,$C5,$B8,$FF,$43,$EB,$FB,$99,$7F,$91,$27
	.dc.b	$2F,$F4,$28,$CB,$FE,$78,$87,$C4,$62,$3F,$CF,$D1,$8A,$FE,$9D,$39
	.dc.b	$BF,$03,$67,$FC,$39,$3F,$D4,$FC,$BC,$11,$FE,$AD,$F7,$5B,$23,$83
	.dc.b	$FC,$FB,$FB,$5F,$FA,$59,$B8,$F7,$8C,$3F,$E6,$16,$BF,$E6,$07,$F4
	.dc.b	$31,$CB,$6E,$04,$7A,$F5,$F3,$23,$83,$FD,$4D,$DB,$FD,$8B,$E7,$3E
	.dc.b	$13,$63,$3D,$C9,$CB,$63,$3F,$05,$82,$38,$3D,$BE,$AF,$AC,$8E,$0F
	.dc.b	$83,$5E,$F2,$38,$3C,$3E,$E7,$80,$8E,$0F,$F5,$2F,$E1,$6F,$AF,$FA
	.dc.b	$CD,$8C,$C6,$0C,$FD,$2C,$8E,$2D,$DF,$FC,$4B,$F9,$9B,$EC,$3F,$09
	.dc.b	$48,$C3,$FE,$62,$6C,$F0,$9F,$C3,$7E,$16,$F6,$A9,$2F,$FD,$77,$19
	.dc.b	$7F,$DA,$DE,$B3,$E4,$FD,$18,$7F,$26,$8C,$67,$F9,$56,$D7,$0A,$DD
	.dc.b	$7F,$2C,$FD,$05,$E4,$7E,$6E,$11,$8A,$FF,$6A,$B6,$3F,$DC,$CD,$7F
	.dc.b	$CD,$5F,$0D,$FE,$F2,$9F,$2B,$63,$C8,$D7,$7F,$17,$FA,$9F,$19,$F4
	.dc.b	$EE,$3C,$37,$0F,$F7,$A4,$EA,$72,$6B,$F0,$B9,$3A,$A4,$63,$0F,$F2
	.dc.b	$CF,$C7,$79,$DF,$D4,$35,$FF,$4C,$36,$8E,$7C,$27,$E1,$7F,$62,$BF
	.dc.b	$D9,$2D,$7F,$DF,$6D,$7F,$E0,$03,$5F,$97,$EF,$7C,$DD,$BF,$D0,$38
	.dc.b	$7C,$C8,$C1,$9F,$56,$47,$D0,$6D,$7F,$62,$1F,$9F,$3E,$7F,$EC,$36
	.dc.b	$B5,$3A,$9C,$9F,$C4,$FF,$0C,$F6,$13,$8F,$06,$7D,$39,$7D,$F5,$52
	.dc.b	$73,$87,$FB,$B7,$F4,$9F,$B8,$85,$23,$31,$86,$C6,$79,$3F,$93,$7D
	.dc.b	$6B,$96,$93,$2F,$14,$65,$FC,$1F,$E7,$7F,$CB,$F1,$FF,$02,$D2,$31
	.dc.b	$1F,$E4,$E8,$C7,$99,$F8,$F9,$CC,$F9,$FA,$1F,$21,$FE,$FA,$C3,$E7
	.dc.b	$46,$3E,$3B,$F9,$F3,$63,$33,$BA,$F0,$DF,$F4,$87,$C0,$D7,$E2,$63
	.dc.b	$E5,$F3,$FC,$47,$F4,$5F,$A6,$7D,$BD,$26,$3F,$46,$32,$FD,$3F,$C3
	.dc.b	$3F,$DD,$CD,$3F,$A7,$38,$C3,$FC,$4F,$C0,$8C,$47,$F9,$C2,$3F,$8B
	.dc.b	$7C,$07,$F5,$5F,$17,$FD,$4F,$7F,$7D,$B5,$C9,$AF,$15,$F2,$9C,$49
	.dc.b	$3C,$5F,$E1,$5F,$A7,$3E,$5E,$92,$7F,$2B,$19,$7F,$86,$FE,$17,$FA
	.dc.b	$52,$38,$3F,$85,$FC,$3F,$F1,$BE,$33,$49,$1F,$D5,$3C,$B7,$A7,$FD
	.dc.b	$51,$B1,$98,$C1,$9F,$3E,$47,$16,$EB,$FE,$50,$F4,$DF,$36,$F9,$DF
	.dc.b	$EB,$ED,$BF,$EC,$6D,$AE,$0D,$7F,$D3,$F4,$F9,$D7,$CC,$7F,$68,$6D
	.dc.b	$FF,$6C,$6D,$78,$5A,$F2,$FD,$ED,$F0,$7F,$DC,$1B,$5F,$DD,$1B,$3E
	.dc.b	$96,$BE,$2F,$DD,$3E,$CC,$FF,$7A,$6C,$7B,$DA,$EF,$93,$FB,$EF,$EC
	.dc.b	$2F,$93,$A4,$BF,$F0,$1C,$65,$FF,$02,$FE,$C1,$FB,$7B,$F9,$4D,$7E
	.dc.b	$1B,$5F,$F7,$FB,$56,$AF,$C1,$DF,$88,$7F,$83,$9B,$7F,$E1,$26,$D7
	.dc.b	$D7,$6B,$F4,$3F,$07,$7D,$BF,$F8,$59,$B5,$FE,$1A,$6C,$FD,$F6,$BF
	.dc.b	$63,$E6,$5F,$2B,$FE,$1E,$6D,$7F,$88,$9B,$3F,$87,$6B,$F8,$0F,$D2
	.dc.b	$6B,$FC,$97,$34,$62,$3D,$ED,$8C,$DF,$F8,$B3,$EB,$3F,$85,$57,$C4
	.dc.b	$6C,$9A,$63,$F7,$27,$D2,$FF,$8D,$1B,$97,$F8,$E1,$B7,$F9,$D6,$BF
	.dc.b	$96,$7F,$91,$C6,$FC,$BE,$44,$7F,$C8,$3F,$5E,$F9,$1F,$F2,$0B,$63
	.dc.b	$31,$83,$FC,$8B,$F8,$23,$E6,$29,$27,$FC,$97,$19,$7F,$AF,$7C,$67
	.dc.b	$F2,$4F,$03,$FC,$DC,$D8,$CF,$9B,$FE,$4B,$D3,$FD,$8B,$39,$7F,$A1
	.dc.b	$FE,$D9,$FE,$09,$6F,$C1,$FF,$83,$E3,$11,$FD,$15,$AF,$F5,$91,$F9
	.dc.b	$47,$F5,$86,$FA,$FF,$E2,$58,$C1,$FD,$73,$E1,$9F,$76,$FC,$83,$FC
	.dc.b	$60,$DD,$3F,$C6,$ED,$C3,$FC,$6C,$D7,$C1,$4F,$DD,$DF,$43,$FE,$39
	.dc.b	$6D,$FF,$90,$1B,$5F,$E3,$F6,$BF,$EE,$6C,$5F,$75,$E7,$25,$7F,$81
	.dc.b	$23,$2F,$F9,$EB,$DE,$FF,$49,$D7,$F1,$F6,$BF,$C6,$1A,$FE,$65,$AF
	.dc.b	$FB,$AB,$F0,$17,$DA,$FD,$66,$D7,$DA,$6C,$FD,$66,$BF,$C9,$3D,$AF
	.dc.b	$F0,$2F,$69,$A4,$E0,$FC,$0F,$C6,$3E,$1E,$93,$1F,$C3,$C6,$5F,$F0
	.dc.b	$2F,$B1,$FD,$FC,$8E,$0F,$F0,$4F,$AF,$FE,$03,$23,$83,$FC,$1B,$EF
	.dc.b	$7F,$81,$C8,$E0,$FC,$AF,$E8,$BF,$EA,$CD,$3F,$C4,$1D,$28,$C4,$23
	.dc.b	$0F,$E6,$A3,$34,$FB,$FF,$D8,$CF,$F8,$19,$AF,$F8,$8D,$B1,$BF,$0F
	.dc.b	$FC,$ED,$3F,$E2,$FF,$9F,$FF,$1F,$B4,$AE,$13,$FE,$44,$FC,$01,$F4
	.dc.b	$CE,$AC,$36,$BF,$C9,$D3,$A9,$C9,$C7,$E5,$7F,$3F,$3A,$6B,$FA,$91
	.dc.b	$3F,$E9,$BE,$CE,$07,$EF,$3F,$00,$7D,$3F,$EB,$E1,$C1,$F2,$A7,$37
	.dc.b	$FE,$4B,$9C,$BF,$2E,$32,$FE,$C2,$7E,$47,$97,$CD,$FF,$17,$FE,$00
	.dc.b	$FB,$AC,$57,$F9,$36,$30,$F5,$9D,$CE,$B8,$D8,$76,$C7,$C3,$C2,$C5
	.dc.b	$F4,$7F,$E6,$E8,$C3,$63,$39,$7F,$A6,$1A,$FF,$95,$97,$D2,$79,$2B
	.dc.b	$A3,$63,$33,$E5,$3D,$15,$BE,$4F,$F5,$34,$66,$FF,$15,$E3,$64,$E5
	.dc.b	$87,$C0,$ED,$FD,$67,$C1,$DE,$EF,$F1,$4E,$1D,$67,$2F,$7D,$BD,$DF
	.dc.b	$3D,$F0,$FF,$35,$FE,$D3,$39,$6B,$F4,$4F,$CF,$38,$DF,$EB,$DF,$49
	.dc.b	$D1,$FE,$DD,$F9,$4F,$A6,$D3,$DC,$53,$A6,$0F,$C0,$7F,$8D,$C7,$97
	.dc.b	$E1,$5F,$21,$F5,$DB,$19,$8E,$BD,$5F,$EA,$58,$F9,$5D,$6F,$C0,$7E
	.dc.b	$4C,$7F,$B9,$DA,$FF,$32,$6B,$FE,$9B,$8E,$1C,$5F,$7F,$3F,$F7,$F4
	.dc.b	$FD,$16,$BF,$0F,$8F,$FD,$69,$F1,$9F,$90,$BE,$33,$EE,$FB,$6F,$E7
	.dc.b	$AD,$F3,$BE,$1C,$62,$3E,$64,$FE,$E3,$EB,$79,$67,$DA,$7C,$D8,$7F
	.dc.b	$AF,$AB,$D8,$D9,$D4,$FF,$A4,$3D,$1F,$F6,$DB,$F8,$76,$BE,$76,$F1
	.dc.b	$FE,$7B,$E6,$3F,$D6,$53,$F8,$3F,$DC,$FF,$62,$FD,$46,$CF,$ED,$39
	.dc.b	$11,$8F,$03,$F9,$1F,$E0,$5F,$E9,$CB,$FC,$67,$E6,$2F,$AA,$FF,$6A
	.dc.b	$E0,$8C,$4F,$F8,$01,$AF,$F7,$C6,$BF,$E9,$9C,$7E,$31,$FC,$13,$E5
	.dc.b	$7F,$B8,$F1,$7E,$F9,$B3,$F9,$DF,$77,$D6,$FE,$5E,$1F,$98,$BE,$C7
	.dc.b	$83,$67,$F6,$13,$A9,$FD,$57,$E0,$79,$9E,$5F,$98,$FC,$A7,$FD,$58
	.dc.b	$D7,$FD,$C5,$48,$C4,$7F,$B9,$1B,$1F,$EE,$6A,$F4,$3D,$72,$F6,$37
	.dc.b	$7F,$F2,$47,$C2,$7F,$8B,$F0,$F2,$FE,$63,$FE,$82,$6F,$C9,$7C,$18
	.dc.b	$C4,$7E,$D7,$C2,$7D,$47,$F8,$C6,$3D,$BF,$98,$BD,$6F,$E7,$11,$E8
	.dc.b	$8C,$3F,$F0,$05,$AF,$FB,$31,$AF,$FB,$2F,$E6,$1F,$39,$F4,$5B,$1F
	.dc.b	$29,$AF,$8F,$91,$F4,$1F,$5F,$07,$CF,$D2,$57,$F1,$5F,$5F,$FC,$1A
	.dc.b	$FD,$4F,$3F,$FB,$63,$E8,$E2,$FE,$57,$D5,$EA,$84,$61,$D4,$F0,$7F
	.dc.b	$DC,$BF,$05,$F0,$DF,$C2,$37,$97,$F1,$8D,$E5,$FF,$3C,$7E,$0A,$FB
	.dc.b	$7F,$C7,$7D,$DF,$93,$0F,$C6,$37,$0F,$CA,$36,$FF,$D0,$1F,$82,$FF
	.dc.b	$B2,$6B,$F1,$BF,$94,$BE,$C7,$D1,$AA,$31,$3E,$D6,$BF,$EF,$C6,$BF
	.dc.b	$EF,$BF,$CC,$7F,$A6,$E5,$FA,$36,$E3,$F9,$26,$CF,$FA,$B3,$E8,$1E
	.dc.b	$B7,$E9,$FE,$3D,$F8,$FD,$25,$7F,$45,$FA,$3D,$BF,$A4,$FC,$C7,$ED
	.dc.b	$BF,$E6,$7D,$51,$88,$FD,$0B,$5F,$A6,$D8,$7E,$62,$FC,$63,$F7,$0D
	.dc.b	$AF,$D8,$B5,$FC,$D7,$D9,$BC,$FF,$E4,$FF,$71,$27,$F9,$63,$FC,$BF
	.dc.b	$F0,$37,$C8,$7F,$4E,$FC,$2F,$F9,$65,$7F,$85,$5A,$FF,$9F,$DA,$83
	.dc.b	$8F,$69,$F4,$FF,$EC,$29,$FE,$E0,$D7,$FA,$03,$79,$BF,$93,$D7,$F1
	.dc.b	$6E,$6F,$E0,$7F,$85,$BE,$CB,$F8,$B3,$70,$FE,$1E,$D7,$F8,$B3,$EF
	.dc.b	$7F,$89,$FE,$5E,$D7,$F8,$F3,$63,$31,$D2,$BF,$91,$7C,$AB,$E5,$29
	.dc.b	$2B,$F8,$EC,$7F,$24,$CF,$F9,$57,$D8,$FC,$89,$FE,$42,$D7,$F9,$43
	.dc.b	$63,$F9,$84,$F5,$9F,$F0,$B7,$D8,$3F,$0C,$FF,$14,$6B,$F2,$E3,$1F
	.dc.b	$27,$FE,$32,$6D,$7C,$AF,$37,$FC,$45,$C3,$FD,$D3,$2F,$F1,$7B,$70
	.dc.b	$FA,$9C,$5F,$4B,$E1,$B9,$67,$DB,$1F,$F3,$37,$95,$FE,$25,$23,$F5
	.dc.b	$A7,$FA,$27,$E0,$7F,$EE,$48,$FE,$7A,$D7,$FA,$23,$72,$FE,$98,$BE
	.dc.b	$93,$FA,$67,$DA,$FF,$9A,$2B,$FA,$2B,$5F,$E9,$8D,$AE,$3F,$33,$FD
	.dc.b	$4F,$F0,$C7,$D9,$D2,$5F,$F4,$E8,$FF,$0C,$BF,$D5,$7E,$6B,$FB,$9E
	.dc.b	$FF,$D2,$9B,$1F,$C7,$1B,$E1,$2B,$AC,$FF,$83,$7C,$7F,$F0,$59,$1F
	.dc.b	$F1,$6F,$E1,$9F,$CB,$9F,$F0,$97,$AF,$FE,$88,$F8,$D3,$49,$1F,$CA
	.dc.b	$FE,$5B,$F2,$4F,$F8,$E9,$BB,$7F,$81,$1B,$A7,$FB,$A6,$F9,$D7,$F9
	.dc.b	$66,$1F,$85,$BE,$47,$FC,$E7,$18,$6C,$66,$7F,$B8,$FE,$19,$F8,$FA
	.dc.b	$FE,$D7,$F7,$DF,$55,$BC,$5F,$DF,$E3,$15,$FE,$8A,$BE,$4F,$F8,$CF
	.dc.b	$F0,$CF,$80,$DF,$73,$FC,$E1,$18,$D7,$FC,$FA,$BF,$06,$CF,$BF,$2F
	.dc.b	$FC,$DD,$3F,$D6,$1B,$7F,$E5,$AF,$65,$F8,$4D,$24,$FF,$94,$3C,$DF
	.dc.b	$F3,$7B,$FE,$3C,$FB,$17,$E0,$1F,$E3,$96,$C7,$F7,$2F,$96,$46,$23
	.dc.b	$FD,$59,$18,$9F,$F4,$86,$5F,$D5,$65,$E1,$A3,$0D,$FA,$FF,$F2,$96
	.dc.b	$1C,$F7,$FF,$55,$F1,$BF,$0B,$7E,$17,$F5,$1B,$97,$D9,$6B,$FE,$29
	.dc.b	$FC,$31,$F7,$9F,$E2,$FF,$C1,$FF,$8C,$63,$FE,$33,$6E,$3F,$E2,$26
	.dc.b	$DF,$FA,$FB,$0E,$75,$FE,$39,$FA,$4F,$E5,$17,$FE,$28,$6D,$7F,$7A
	.dc.b	$6F,$84,$7F,$96,$47,$F7,$EF,$C2,$3F,$9A,$BF,$E3,$86,$BF,$E3,$F6
	.dc.b	$DF,$7F,$AA,$FF,$5F,$73,$FF,$55,$F2,$FF,$C2,$3F,$1B,$FE,$54,$8C
	.dc.b	$13,$E0,$76,$4C,$7F,$D4,$D1,$FE,$1C,$C3,$FD,$3F,$F8,$4F,$FB,$85
	.dc.b	$B3,$FE,$B1,$8C,$6B,$FE,$44,$FC,$2F,$F8,$33,$FE,$46,$FC,$2F,$FD
	.dc.b	$13,$7F,$E5,$B6,$C6,$7E,$8B,$96,$1F,$E5,$DB,$FF,$70,$BF,$DC,$5A
	.dc.b	$F5,$F5,$BF,$C1,$71,$88,$FF,$11,$DF,$F4,$77,$FC,$C4,$DC,$3C,$1E
	.dc.b	$DF,$5B,$ED,$F4,$7C,$89,$7F,$E2,$FC,$11,$FE,$18,$C1,$F8,$57,$F9
	.dc.b	$C5,$B5,$FE,$6A,$6B,$FE,$71,$AE,$2F,$F9,$B9,$FC,$2F,$21,$A4,$F8
	.dc.b	$2E,$9E,$B4,$AE,$8F,$C3,$F7,$BE,$27,$89,$F0,$09,$3E,$77,$CD,$ED
	.dc.b	$79,$49,$1E,$18,$FF,$69,$FE,$18,$F9,$9F,$F5,$0F,$D8,$7F,$9B,$E9
	.dc.b	$39,$F1,$3F,$C0,$D3,$98,$C4,$A3,$1E,$DF,$F8,$2E,$31,$F7,$76,$B7
	.dc.b	$5C,$70,$EF,$39,$6D,$7F,$84,$E9,$7C,$EF,$86,$1E,$2C,$38,$36,$F8
	.dc.b	$61,$C9,$F4,$DF,$1A,$DC,$79,$0D,$D7,$17,$DF,$5C,$70,$F0,$C6,$5F
	.dc.b	$4D,$72,$C3,$C7,$3C,$CE,$E6,$EB,$91,$2F,$96,$1E,$39,$F0,$9C,$70
	.dc.b	$EA,$E2,$FE,$5B,$CE,$A3,$BB,$F3,$5F,$38,$F4,$CE,$AB,$CA,$72,$8E
	.dc.b	$11,$9C,$3D,$47,$D5,$AE,$58,$78,$AB,$99,$C0,$F3,$46,$B0,$F6,$46
	.dc.b	$70,$EC,$7C,$47,$F0,$47,$22,$D7,$E2,$1B,$3F,$ED,$67,$F0,$DF,$5D
	.dc.b	$FE,$97,$6F,$6F,$FA,$76,$93,$99,$FC,$33,$5F,$CB,$BF,$EA,$6B,$FC
	.dc.b	$B3,$71,$FC,$97,$CB,$3A,$BE,$E4,$E6,$F9,$E3,$83,$F0,$15,$B8,$FA
	.dc.b	$D3,$B7,$BC,$ED,$FB,$53,$BC,$3C,$33,$96,$CE,$AF,$18,$7B,$A3,$AD
	.dc.b	$78,$27,$31,$E4,$32,$71,$2D,$74,$25,$7B,$A7,$8C,$F9,$8D,$C6,$5A
	.dc.b	$75,$9E,$2F,$EC,$6F,$D5,$3E,$59,$CC,$F3,$7F,$81,$4E,$A3,$94,$EC
	.dc.b	$8F,$3A,$F1,$E3,$F6,$B0,$B7,$C6,$BA,$5F,$F4,$5E,$FF,$A1,$D5,$FD
	.dc.b	$15,$B8,$FF,$4B,$ED,$B7,$FC,$FA,$6F,$37,$F2,$B9,$FB,$13,$FD,$1D
	.dc.b	$AF,$1A,$FE,$79,$18,$8F,$E8,$FF,$13,$CB,$A0,$F4,$E8,$FF,$13,$5F
	.dc.b	$F3,$D8,$C7,$A1,$23,$F9,$E7,$D2,$5C,$3F,$A7,$FC,$07,$33,$A3,$67
	.dc.b	$FA,$33,$E1,$CD,$97,$FA,$CE,$7F,$DF,$CD,$9F,$D5,$CE,$9E,$10,$F5
	.dc.b	$7D,$07,$A7,$DA,$B7,$FD,$4D,$D9,$1E,$AF,$2F,$B9,$F8,$07,$CC,$F9
	.dc.b	$D7,$8B,$FE,$B1,$AF,$EB,$93,$98,$C6,$2E,$2F,$F0,$1F,$89,$F2,$7D
	.dc.b	$ED,$3F,$2F,$EF,$E6,$3F,$D9,$7E,$4A,$E1,$FD,$D1,$F5,$92,$3E,$0B
	.dc.b	$F1,$79,$1F,$8A,$FD,$33,$EF,$F9,$8F,$C3,$71,$7F,$E2,$37,$26,$BE
	.dc.b	$62,$C7,$FB,$7F,$5F,$F1,$96,$9F,$88,$F3,$FD,$71,$FE,$46,$FC,$21
	.dc.b	$E7,$66,$BF,$1F,$69,$CB,$FE,$7F,$6E,$DF,$0E,$3F,$C8,$5F,$03,$C2
	.dc.b	$73,$87,$F0,$BF,$1D,$F7,$0E,$2F,$CD,$C6,$1F,$CE,$FB,$8F,$A2,$F7
	.dc.b	$65,$E9,$3F,$D8,$B7,$C5,$C6,$6A,$E1,$C1,$F9,$D6,$CF,$F9,$BE,$F9
	.dc.b	$D7,$F9,$B9,$AF,$F3,$8F,$C3,$5E,$87,$EB,$A1,$18,$AE,$F6,$F8,$E7
	.dc.b	$CA,$EE,$B5,$7F,$0C,$FF,$3B,$E6,$F4,$3D,$ED,$C7,$D3,$B7,$C4,$F4
	.dc.b	$B3,$5A,$64,$EA,$BF,$C6,$2D,$8F,$E0,$3E,$DB,$CB,$A7,$D8,$F2,$3C
	.dc.b	$7C,$7F,$CE,$FC,$55,$6F,$0F,$F5,$C5,$69,$FD,$93,$74,$53,$E1,$7E
	.dc.b	$DD,$6B,$8B,$E3,$7E,$1A,$EC,$A7,$C3,$F8,$2B,$5F,$5B,$1F,$97,$1F
	.dc.b	$C4,$5B,$1F,$DA,$11,$83,$1C,$5F,$EA,$A8,$D4,$7D,$4F,$91,$FA,$07
	.dc.b	$8B,$EC,$6C,$FF,$B3,$3D,$5F,$E4,$BF,$13,$FE,$3F,$8F,$C3,$EC,$FC
	.dc.b	$37,$DC,$A7,$58,$FD,$98,$D6,$DF,$CB,$AB,$63,$DA,$C4,$7F,$32,$F8
	.dc.b	$E7,$93,$4F,$E6,$BF,$7D,$F8,$6C,$FC,$33,$A7,$C7,$19,$CB,$C0,$DE
	.dc.b	$6F,$B1,$F2,$5F,$EC,$0A,$DF,$6F,$EA,$3E,$DF,$FB,$E4,$7E,$29,$B8
	.dc.b	$7E,$47,$B9,$F8,$6B,$E3,$B8,$5F,$EC,$23,$0F,$E4,$9B,$5F,$99,$F0
	.dc.b	$5E,$AB,$ED,$B8,$61,$FD,$4A,$73,$16,$75,$87,$3F,$67,$13,$FD,$5B
	.dc.b	$F5,$4F,$F7,$4C,$FD,$26,$E3,$F3,$FE,$E9,$F8,$1E,$FE,$13,$FA,$EC
	.dc.b	$62,$FE,$CB,$6B,$EB,$7C,$DF,$C0,$E3,$FC,$F4,$E5,$FD,$07,$CE,$DB
	.dc.b	$FC,$6F,$5A,$31,$87,$E7,$BD,$27,$C5,$FF,$6C,$9C,$D7,$F2,$48,$FE
	.dc.b	$DF,$3A,$BF,$C5,$AF,$9C,$CC,$FC,$DD,$25,$7E,$87,$F0,$E7,$CF,$63
	.dc.b	$3F,$E0,$C2,$33,$3F,$9E,$6C,$7E,$95,$AF,$F9,$39,$AF,$FA,$47,$C5
	.dc.b	$7D,$6F,$F1,$33,$F9,$5F,$37,$FB,$AE,$FF,$65,$3A,$7F,$47,$F9,$17
	.dc.b	$F7,$EB,$FD,$03,$78,$3F,$C8,$5F,$01,$FD,$07,$E0,$7F,$C1,$F3,$9B
	.dc.b	$F9,$7F,$91,$7F,$B1,$9F,$DD,$71,$23,$18,$FF,$7A,$6B,$FE,$A7,$8D
	.dc.b	$7C,$7F,$F8,$0F,$F2,$27,$C4,$FF,$87,$1B,$87,$F3,$EC,$DF,$50,$FB
	.dc.b	$ED,$5F,$F8,$8B,$F2,$2F,$F0,$46,$FF,$C7,$DB,$87,$F0,$CC,$1F,$56
	.dc.b	$FA,$5E,$1B,$7F,$8A,$BF,$22,$FC,$9B,$E5,$6B,$F8,$86,$BF,$C9,$1B
	.dc.b	$BB,$E4,$3F,$37,$9F,$F2,$F6,$E5,$E1,$FC,$4B,$F0,$CF,$F1,$28,$CE
	.dc.b	$1F,$CB,$BC,$E7,$A9,$FE,$3C,$6D,$7C,$1F,$C4,$9F,$3B,$9A,$FF,$20
	.dc.b	$FB,$CF,$B2,$FF,$16,$36,$FE,$9B,$71,$FF,$15,$B5,$FA,$5F,$91,$7F
	.dc.b	$64,$D7,$FA,$33,$6B,$E4,$7E,$22,$FC,$4B,$FC,$9F,$39,$BF,$E8,$BF
	.dc.b	$91,$7F,$73,$7F,$94,$37,$0F,$C4,$FE,$27,$71,$FD,$8C,$DA,$FE,$9B
	.dc.b	$F9,$13,$E5,$3F,$C8,$8D,$AF,$C5,$B6,$7F,$C8,$6D,$7F,$15,$F6,$8F
	.dc.b	$9C,$FF,$31,$B6,$BF,$11,$F8,$A7,$E2,$AB,$F9,$E4,$E5,$FE,$7D,$F9
	.dc.b	$17,$AF,$A0,$FB,$7B,$1F,$80,$3D,$6F,$F3,$7C,$EF,$3F,$ED,$58,$7F
	.dc.b	$95,$9A,$FE,$75,$AF,$F1,$06,$BF,$DB,$3E,$B5,$F8,$77,$F9,$E9,$BA
	.dc.b	$FE,$7F,$F1,$5F,$E6,$2F,$F5,$88,$CD,$7F,$55,$F7,$5F,$11,$FE,$82
	.dc.b	$6D,$7E,$9F,$F0,$7F,$CA,$3F,$D8,$23,$35,$FD,$74,$FE,$8E,$7E,$CC
	.dc.b	$A7,$F2,$62,$9F,$E0,$22,$E5,$FC,$DB,$BF,$FD,$26,$DA,$FF,$19,$BF
	.dc.b	$CE,$B9,$33,$ED,$7F,$5F,$BF,$EA,$9F,$B3,$F8,$BA,$7F,$C8,$8D,$C3
	.dc.b	$FC,$03,$93,$F0,$C7,$C9,$7F,$91,$AB,$58,$7F,$8F,$FF,$15,$F8,$AF
	.dc.b	$FB,$C3,$E5,$91,$88,$FF,$05,$9B,$F8,$2F,$F2,$24,$66,$33,$F8,$1F
	.dc.b	$F6,$FC,$EF,$C3,$FF,$94,$B5,$7D,$1F,$F9,$0E,$B5,$6D,$9F,$84,$FE
	.dc.b	$97,$1F,$F1,$DB,$67,$FC,$85,$A7,$93,$E6,$7F,$CB,$F3,$9C,$BF,$DD
	.dc.b	$D3,$9B,$FF,$11,$36,$FD,$9A,$7F,$91,$DB,$87,$FA,$E7,$8D,$F8,$EF
	.dc.b	$EC,$DB,$F3,$9F,$E9,$BF,$8E,$FE,$6D,$7F,$A5,$AF,$CE,$46,$3C,$03
	.dc.b	$AB,$7E,$28,$FC,$57,$97,$D5,$FD,$EF,$8E,$7E,$47,$4F,$F5,$47,$C6
	.dc.b	$7C,$4C,$3F,$C1,$D1,$8E,$79,$C5,$F8,$4F,$9F,$FB,$3A,$B0,$F9,$7C
	.dc.b	$CF,$BD,$FE,$08,$DF,$A2,$FF,$0B,$7C,$57,$FA,$1A,$73,$F2,$7F,$D9
	.dc.b	$3A,$7F,$D1,$8D,$CF,$F1,$2F,$F5,$DE,$5F,$F1,$13,$97,$EC,$39,$7F
	.dc.b	$0B,$8F,$F8,$9B,$E2,$BF,$DC,$8D,$AF,$F2,$A3,$67,$F1,$67,$E0,$7F
	.dc.b	$14,$FE,$76,$FF,$8B,$DB,$82,$BF,$CB,$7C,$2F,$C5,$BF,$A9,$DF,$1F
	.dc.b	$90,$FD,$4C,$62,$75,$3E,$FF,$AD,$CB,$FE,$93,$3F,$CC,$BF,$63,$FE
	.dc.b	$62,$69,$5C,$27,$FC,$CF,$F8,$0B,$F0,$BF,$F5,$34,$FF,$9B,$E3,$1E
	.dc.b	$0E,$A7,$33,$EE,$9C,$BB,$FA,$AF,$F7,$03,$75,$EB,$18,$8E,$0D,$7F
	.dc.b	$77,$DF,$FC,$A5,$FF,$4F,$B5,$FF,$2B,$36,$7F,$D8,$DE,$E7,$F9,$AA
	.dc.b	$7F,$85,$FE,$2C,$FB,$EF,$DD,$D7,$CD,$9F,$E8,$13,$97,$73,$AE,$1F
	.dc.b	$0F,$AD,$67,$F7,$1F,$8B,$78,$5F,$C0,$B5,$FE,$22,$D7,$71,$FE,$90
	.dc.b	$E5,$4B,$E5,$FA,$46,$1B,$19,$AF,$F5,$EB,$ED,$BE,$BF,$88,$68,$93
	.dc.b	$AA,$F3,$72,$39,$FF,$B2,$4F,$F2,$C6,$BF,$EC,$D8,$D7,$A5,$C2,$7F
	.dc.b	$DF,$33,$BF,$AF,$C1,$1E,$53,$DB,$2F,$F6,$CC,$7F,$38,$E0,$FE,$38
	.dc.b	$DD,$35,$E1,$79,$27,$33,$FE,$DD,$F0,$7F,$84,$F7,$FC,$D6,$C7,$13
	.dc.b	$FA,$1F,$E6,$FF,$E9,$17,$FD,$7D,$AA,$31,$1F,$AC,$8C,$78,$BF,$EB
	.dc.b	$E6,$C7,$FB,$16,$BF,$D6,$D9,$77,$FA,$7F,$1F,$37,$F9,$59,$BC,$7F
	.dc.b	$B0,$CB,$AC,$FE,$1F,$17,$E6,$FF,$EE,$7C,$BF,$AB,$AF,$BA,$FA,$2D
	.dc.b	$8C,$FC,$46,$3E,$AB,$C9,$26,$92,$7D,$BF,$73,$FE,$E3,$7F,$B1,$B5
	.dc.b	$FB,$0D,$C7,$FC,$DB,$F6,$97,$4F,$E0,$FF,$8B,$BE,$1B,$E6,$7E,$3D
	.dc.b	$FE,$40,$C7,$FC,$C4,$D7,$F0,$53,$B7,$FB,$8F,$EC,$3F,$C1,$77,$34
	.dc.b	$91,$FE,$EA,$F8,$86,$7F,$E7,$08,$FF,$72,$B5,$FF,$71,$34,$FD,$84
	.dc.b	$70,$FB,$4F,$D6,$2F,$C5,$FF,$41,$5F,$85,$F6,$FF,$40,$B9,$FF,$B3
	.dc.b	$F8,$5C,$2B,$78,$F4,$FC,$5F,$FD,$9F,$7F,$E8,$A6,$BF,$CC,$31,$46
	.dc.b	$23,$FD,$14,$D7,$FD,$7B,$1D,$38,$2F,$85,$E6,$8C,$49,$98,$FA,$BF
	.dc.b	$17,$7E,$03,$49,$43,$E3,$E1,$FE,$CA,$97,$27,$C2,$FC,$5F,$FD,$05
	.dc.b	$5F,$EA,$26,$BE,$96,$BF,$EA,$16,$BF,$ED,$4F,$A6,$FF,$65,$3D,$F5
	.dc.b	$F9,$F1,$FE,$71,$8C,$61,$BF,$07,$E8,$CF,$09,$F2,$9E,$DF,$4F,$F4
	.dc.b	$35,$FE,$E0,$6C,$67,$C1,$FB,$30,$F6,$BE,$DF,$73,$0F,$C3,$36,$FF
	.dc.b	$D1,$71,$83,$F1,$5F,$02,$FA,$4F,$C1,$AF,$B3,$A6,$23,$F0,$FF,$85
	.dc.b	$FF,$BF,$AF,$D6,$7D,$37,$0F,$AB,$91,$18,$DF,$F2,$3F,$86,$FF,$41
	.dc.b	$AD,$7E,$2A,$FC,$B7,$F3,$8D,$8F,$F6,$9B,$5F,$C5,$35,$FF,$69,$7E
	.dc.b	$2D,$FB,$B7,$FD,$9D,$F4,$BF,$CA,$9F,$E3,$CD,$C3,$F4,$33,$B9,$F3
	.dc.b	$79,$EF,$B0,$FD,$1B,$63,$DB,$A7,$FA,$76,$73,$EF,$7E,$A3,$F1,$6F
	.dc.b	$B1,$5F,$EE,$36,$BF,$8F,$A4,$62,$3F,$30,$D8,$7E,$2D,$FC,$6E,$7F
	.dc.b	$9A,$7E,$C4,$FC,$43,$F9,$A3,$63,$C7,$3B,$D3,$F7,$7F,$B2,$7A,$DF
	.dc.b	$ED,$9E,$9B,$F5,$DF,$C8,$5F,$E3,$23,$1D,$DF,$C0,$2F,$87,$77,$F9
	.dc.b	$BB,$F1,$9F,$F3,$BD,$FF,$89,$8F,$D4,$38,$70,$D7,$FC,$56,$D7,$97
	.dc.b	$E1,$FF,$DD,$D4,$23,$0D,$8F,$C4,$9A,$F9,$6F,$E2,$7F,$8C,$BD,$EB
	.dc.b	$E4,$CA,$FA,$73,$96,$C7,$0F,$26,$46,$0F,$B3,$F3,$8F,$17,$F8,$FB
	.dc.b	$63,$31,$8F,$AC,$CF,$DA,$C9,$E2,$F0,$9F,$F0,$3F,$17,$E1,$67,$E6
	.dc.b	$2F,$C4,$B2,$6B,$AE,$CF,$E9,$BE,$E5,$FB,$5B,$E3,$7F,$C3,$91,$87
	.dc.b	$73,$FC,$0A,$30,$FF,$8C,$3E,$2F,$FC,$2F,$F1,$3F,$E4,$38,$E9,$7C
	.dc.b	$DF,$CB,$FA,$5F,$EA,$06,$FC,$DF,$F3,$E8,$C4,$7E,$D5,$7C,$96,$7F
	.dc.b	$00,$93,$C9,$E1,$5F,$D1,$BF,$03,$7D,$7D,$FB,$89,$7F,$A6,$9C,$B6
	.dc.b	$3F,$A6,$FC,$8F,$F4,$2F,$0B,$FA,$8B,$7B,$FF,$4D,$6A,$FE,$A7,$F1
	.dc.b	$2F,$9D,$FC,$57,$E3,$3F,$D0,$32,$3F,$D0,$BC,$95,$6E,$7F,$17,$E9
	.dc.b	$3E,$1F,$FC,$DC,$D8,$FC,$E1,$FC,$52,$32,$6B,$D9,$FC,$C4,$FF,$2B
	.dc.b	$9E,$71,$89,$FE,$CF,$F8,$3F,$F8,$5A,$7F,$A9,$37,$03,$47,$C6,$F0
	.dc.b	$EE,$95,$FC,$32,$7F,$24,$FF,$27,$7A,$79,$7F,$B9,$E7,$FC,$79,$27
	.dc.b	$F9,$7C,$7E,$F7,$DE,$FC,$9F,$C5,$49,$CF,$83,$FB,$C7,$FD,$BC,$FF
	.dc.b	$64,$7F,$8D,$FE,$17,$5F,$E0,$9C,$3F,$45,$EA,$D7,$F9,$D4,$6F,$9F
	.dc.b	$F8,$6C,$FD,$E9,$FF,$55,$E3,$F5,$3A,$3F,$CD,$D3,$99,$FF,$5A,$B5
	.dc.b	$FE,$E5,$DE,$FD,$53,$E7,$47,$E7,$DB,$87,$FA,$16,$75,$C7,$FE,$76
	.dc.b	$BF,$04,$FF,$86,$B2,$FD,$86,$1E,$CF,$0F,$FC,$37,$18,$8F,$4F,$87
	.dc.b	$8F,$52,$47,$C2,$77,$5F,$E2,$2E,$07,$C3,$D2,$57,$F4,$28,$FC,$C3
	.dc.b	$FE,$11,$7F,$60,$FF,$5F,$6E,$3F,$E2,$E6,$F6,$FF,$43,$7A,$CF,$F4
	.dc.b	$CF,$C4,$7F,$74,$FF,$89,$5A,$FF,$85,$CF,$F0,$F3,$87,$FA,$63,$5F
	.dc.b	$F1,$A7,$E2,$5F,$ED,$CB,$FF,$56,$7C,$42,$31,$87,$F8,$86,$31,$5F
	.dc.b	$E4,$0B,$F0,$DF,$F9,$B7,$CC,$FF,$09,$37,$D4,$FF,$4D,$C7,$F4,$A8
	.dc.b	$D6,$B8,$9F,$F4,$47,$9A,$FA,$1A,$67,$1F,$ED,$3F,$89,$BC,$8F,$F2
	.dc.b	$C3,$63,$FA,$77,$80,$8C,$47,$F9,$1D,$AF,$C3,$D7,$D9,$F4,$FF,$E9
	.dc.b	$7E,$0F,$A7,$18,$9E,$F8,$7F,$97,$2D,$F7,$BE,$02,$4F,$82,$32,$DE
	.dc.b	$BE,$04,$7D,$33,$CF,$C7,$91,$EC,$7B,$69,$DF,$EA,$E4,$FC,$07,$6F
	.dc.b	$F8,$8F,$D8,$FF,$10,$3F,$77,$6F,$EC,$9C,$BF,$E7,$B9,$CC,$FF,$82
	.dc.b	$5A,$FF,$97,$DF,$F4,$33,$CF,$E4,$FF,$CF,$31,$8E,$4F,$F3,$EB,$FE
	.dc.b	$2D,$7F,$C5,$DF,$8A,$7F,$A7,$FC,$1F,$F3,$17,$D0,$7F,$97,$23,$18
	.dc.b	$23,$1E,$27,$FB,$29,$D7,$8B,$F9,$37,$37,$CB,$D6,$FC,$B6,$1C,$A7
	.dc.b	$38,$73,$9C,$DF,$17,$78,$75,$6E,$1C,$EB,$FC,$ED,$87,$39,$E3,$19
	.dc.b	$11,$97,$75,$CD,$B5,$B9,$FE,$3D,$87,$90,$97,$E0,$AF,$31,$E0,$C3
	.dc.b	$CC,$F2,$32,$78,$30,$EC,$DC,$3B,$4F,$08,$C1,$C3,$0F,$33,$8C,$3E
	.dc.b	$5B,$E5,$BE,$73,$A2,$57,$37,$11,$A3,$71,$97,$17,$CC,$F1,$D5,$3C
	.dc.b	$51,$A7,$C9,$3C,$67,$57,$F1,$27,$E1,$1C,$4F,$E2,$47,$B0,$B7,$F8
	.dc.b	$86,$DA,$BF,$B6,$CF,$FB,$CF,$E4,$EF,$CB,$9A,$4B,$FF,$27,$4E,$DF
	.dc.b	$91,$3B,$7E,$57,$C6,$B7,$FB,$93,$9C,$BF,$34,$DD,$3F,$26,$DB,$FC
	.dc.b	$5D,$F1,$7A,$E3,$C8,$8F,$6C,$2E,$1D,$26,$E1,$F9,$EB,$E5,$39,$BE
	.dc.b	$64,$7F,$59,$1A,$9F,$1D,$72,$AF,$CE,$E1,$E5,$7F,$37,$5B,$73,$3B
	.dc.b	$7F,$3F,$3B,$C5,$5B,$9C,$D6,$E7,$E5,$61,$E2,$C7,$18,$71,$39,$61
	.dc.b	$C6,$B7,$8F,$9B,$1F,$53,$EF,$BE,$38,$F5,$BE,$18,$73,$8C,$4F,$B7
	.dc.b	$83,$FC,$FD,$5F,$3F,$6F,$E2,$6D,$9F,$E3,$97,$71,$FE,$39,$36,$FF
	.dc.b	$D0,$D3,$FC,$89,$B5,$FD,$1B,$74,$E5,$B1,$9D,$B9,$41,$BB,$E1,$B3
	.dc.b	$FE,$BC,$94,$F0,$F0,$33,$97,$F3,$7F,$BA,$E6,$74,$F8,$3B,$87,$17
	.dc.b	$C4,$47,$CC,$FF,$3A,$EB,$FE,$E5,$EA,$FB,$DB,$3E,$1F,$B2,$C7,$72
	.dc.b	$BD,$31,$ED,$24,$FC,$17,$E0,$7D,$47,$0E,$FF,$7B,$67,$D9,$EA,$E1
	.dc.b	$A7,$BE,$3B,$FC,$F4,$8F,$EE,$BF,$5F,$76,$FF,$7C,$78,$48,$FA,$4D
	.dc.b	$9F,$ED,$11,$83,$EB,$7E,$26,$CF,$ED,$36,$7F,$DD,$2F,$FB,$6D,$B1
	.dc.b	$AD,$E4,$6B,$2B,$8B,$A3,$F7,$1C,$EF,$C5,$FF,$CD,$B7,$FE,$F4,$6D
	.dc.b	$6E,$3F,$DE,$10,$F1,$BF,$18,$EA,$F9,$B9,$9E,$53,$AF,$C5,$76,$FE
	.dc.b	$57,$D2,$7E,$22,$FC,$5B,$0D,$BF,$F3,$9F,$E8,$4F,$99,$CD,$7E,$77
	.dc.b	$C5,$4E,$6B,$FD,$D9,$18,$AF,$EF,$ED,$8C,$CE,$F5,$FE,$63,$B3,$D4
	.dc.b	$F9,$51,$FE,$FB,$97,$A1,$FA,$99,$CF,$07,$F1,$FC,$7C,$33,$E3,$AF
	.dc.b	$45,$F1,$CF,$1F,$09,$96,$E7,$EC,$8E,$1F,$53,$8A,$F9,$C3,$47,$E2
	.dc.b	$AF,$9B,$FD,$D3,$70,$57,$0A,$DF,$4E,$AF,$6C,$AD,$ED,$F8,$0C,$1F
	.dc.b	$27,$9C,$7C,$23,$F0,$BF,$5D,$7F,$3D,$8C,$6B,$FC,$35,$D6,$B6,$B7
	.dc.b	$8F,$F9,$B5,$47,$F8,$B9,$B1,$FE,$C6,$6C,$FF,$62,$39,$C7,$F9,$63
	.dc.b	$E1,$1F,$81,$F7,$FC,$55,$EE,$FF,$1B,$9C,$D7,$B0,$FF,$3E,$4E,$9C
	.dc.b	$5F,$CB,$FC,$55,$F5,$BC,$7C,$EF,$E4,$91,$8A,$DF,$1F,$FB,$82,$73
	.dc.b	$F3,$DA,$BB,$5A,$BE,$18,$7D,$1F,$B6,$7B,$BF,$CB,$DB,$5D,$17,$BD
	.dc.b	$99,$D3,$27,$5F,$91,$5B,$FC,$1F,$80,$FD,$B7,$8D,$7F,$B5,$BE,$B7
	.dc.b	$FD,$56,$FB,$A7,$58,$7E,$96,$73,$7F,$8C,$EF,$7D,$C3,$2F,$D9,$65
	.dc.b	$D2,$B5,$39,$CB,$A4,$E7,$A7,$F4,$12,$FA,$FF,$D9,$3B,$BE,$31,$8C
	.dc.b	$3F,$A8,$7D,$BB,$E0,$E9,$8F,$82,$B9,$FD,$09,$46,$63,$1F,$05,$71
	.dc.b	$FC,$F7,$E2,$2F,$A5,$FE,$B8,$FE,$0A,$33,$18,$ED,$CF,$C4,$EA,$2E
	.dc.b	$1F,$7A,$F9,$3F,$86,$B7,$E2,$2F,$82,$FC,$F3,$67,$4B,$E4,$33,$3A
	.dc.b	$7F,$D1,$71,$99,$FC,$E5,$74,$9C,$9F,$DC,$7E,$83,$F9,$D4,$7D,$0E
	.dc.b	$7F,$EE,$D1,$8E,$4F,$9F,$1F,$DE,$7F,$11,$FE,$31,$1F,$ED,$2F,$A3
	.dc.b	$FC,$6D,$FF,$01,$9C,$E3,$FE,$D3,$9C,$D7,$F3,$EF,$9D,$FC,$BD,$FF
	.dc.b	$A9,$5A,$FF,$0A,$F0,$91,$88,$FE,$24,$D8,$7E,$22,$FA,$2F,$F0,$63
	.dc.b	$71,$FD,$4A,$FB,$DC,$CF,$F8,$47,$F4,$D7,$CD,$7F,$77,$FA,$97,$CF
	.dc.b	$E6,$30,$FF,$85,$FF,$0F,$7D,$37,$F8,$5D,$B9,$FF,$B7,$1B,$B7,$F3
	.dc.b	$A6,$BF,$EC,$1E,$67,$E3,$1F,$E2,$16,$D7,$EF,$9E,$3C,$3F,$D8,$9C
	.dc.b	$D7,$F2,$BF,$55,$F1,$9F,$E2,$97,$FC,$37,$7F,$C1,$9C,$B7,$E3,$7F
	.dc.b	$C5,$DF,$21,$FE,$E4,$8F,$F7,$8B,$6F,$FA,$7B,$5F,$F6,$B3,$5F,$A5
	.dc.b	$F8,$6F,$E0,$66,$FC,$37,$F8,$E6,$1F,$0B,$63,$FC,$79,$F8,$67,$F8
	.dc.b	$76,$BF,$8C,$AF,$A7,$CF,$C7,$69,$FF,$21,$7E,$19,$FE,$0F,$CB,$FB
	.dc.b	$0B,$5F,$F7,$43,$5F,$C2,$35,$FE,$A9,$FB,$9F,$DA,$C3,$F9,$3B,$63
	.dc.b	$FA,$97,$CC,$FF,$54,$9C,$F3,$FF,$94,$7F,$0B,$F6,$57,$F5,$D6,$F7
	.dc.b	$1F,$6D,$7F,$32,$73,$F3,$0C,$3F,$CB,$0D,$9F,$E9,$15,$FE,$A1,$9C
	.dc.b	$9F,$E5,$DF,$C2,$BF,$BB,$CF,$F6,$6F,$96,$7E,$17,$F9,$58,$FE,$7F
	.dc.b	$3A,$C9,$5A,$F0,$7F,$CD,$1E,$B7,$F0,$19,$FE,$DC,$DE,$BF,$F1,$4B
	.dc.b	$FA,$8C,$3F,$A4,$B9,$7F,$BA,$E8,$FB,$C7,$CF,$7F,$6B,$C3,$31,$8A
	.dc.b	$FE,$C5,$5B,$AD,$57,$FA,$F6,$B4,$FF,$B2,$27,$9F,$C4,$FF,$98,$7F
	.dc.b	$26,$7C,$5F,$FA,$06,$73,$8F,$F6,$2D,$55,$A8,$FE,$CB,$2F,$CC,$1F
	.dc.b	$61,$C5,$B3,$C7,$E5,$FF,$BB,$EF,$FE,$BF,$BE,$17,$DF,$F2,$FF,$F2
	.dc.b	$06,$BE,$56,$BF,$E1,$63,$FD,$AC,$E1,$F8,$AD,$73,$F5,$7C,$FC,$EE
	.dc.b	$8D,$CB,$D5,$F1,$5E,$4A,$FF,$13,$EC,$8C,$67,$E5,$FC,$BB,$FA,$85
	.dc.b	$F0,$CF,$E2,$46,$30,$F9,$4D,$7C,$DA,$23,$11,$F6,$9A,$FF,$BB,$FE
	.dc.b	$07,$96,$1F,$EE,$1C,$9F,$0D,$FE,$48,$CB,$E6,$D6,$9F,$C9,$46,$6F
	.dc.b	$FC,$7F,$8F,$F7,$4F,$7B,$EC,$5E,$E7,$B4,$63,$E0,$3E,$E7,$E1,$7F
	.dc.b	$D3,$67,$8B,$73,$FA,$8F,$E4,$61,$82,$35,$97,$E5,$1D,$46,$7C,$DE
	.dc.b	$BC,$3D,$9F,$F3,$77,$C3,$BE,$06,$98,$EB,$B8,$FF,$8B,$BB,$3F,$C4
	.dc.b	$31,$98,$C5,$7F,$7C,$6C,$74,$DB,$FC,$EB,$39,$D3,$F0,$C2,$FF,$C5
	.dc.b	$73,$A3,$F4,$79,$BF,$07,$FE,$21,$B7,$E4,$DB,$97,$F9,$92,$DF,$0F
	.dc.b	$FC,$7E,$B5,$E2,$FF,$89,$7D,$09,$7E,$6F,$A8,$BA,$7F,$98,$8F,$F3
	.dc.b	$6F,$D3,$FF,$7A,$69,$5C,$27,$FB,$FF,$D9,$7F,$98,$3F,$08,$BF,$28
	.dc.b	$69,$2B,$FC,$C1,$3F,$B8,$CF,$8F,$87,$FE,$B5,$9C,$B5,$F8,$B3,$B7
	.dc.b	$F8,$16,$FF,$E5,$C7,$F7,$6D,$7F,$1A,$DA,$FF,$4C,$7D,$AF,$FA,$72
	.dc.b	$7E,$AF,$8B,$FB,$59,$FF,$4E,$72,$7F,$A3,$E7,$27,$4F,$90,$D6,$FF
	.dc.b	$37,$F0,$7F,$E0,$F3,$FE,$B0,$6B,$F6,$B9,$11,$88,$FE,$05,$18,$F0
	.dc.b	$7F,$C4,$AE,$BD,$CE,$B3,$F8,$1A,$FF,$00,$BF,$8D,$6B,$FE,$5B,$7B
	.dc.b	$E5,$EC,$9E,$34,$D3,$73,$9B,$E5,$5D,$B9,$DF,$84,$3E,$97,$0D,$AF
	.dc.b	$E4,$AD,$7F,$6E,$FF,$AA,$1F,$E5,$9F,$1C,$FA,$4F,$4F,$8A,$8C,$4E
	.dc.b	$A7,$AD,$F8,$78,$7F,$3D,$3F,$CD,$7C,$57,$DA,$72,$6C,$7F,$36,$9D
	.dc.b	$BF,$C8,$23,$2F,$EA,$31,$FE,$23,$5F,$E5,$06,$C7,$FA,$B5,$AE,$FC
	.dc.b	$8E,$6F,$F3,$58,$FE,$55,$C1,$C6,$73,$3F,$DA,$5A,$FF,$0C,$FD,$0B
	.dc.b	$E8,$F2,$7F,$5E,$6C,$67,$93,$E1,$D7,$A3,$8B,$FA,$47,$E1,$2F,$99
	.dc.b	$FA,$4B,$EB,$A9,$88,$F9,$7F,$06,$FA,$5F,$63,$72,$F6,$35,$F8,$4D
	.dc.b	$7F,$9D,$F8,$9F,$E8,$77,$FA,$FB,$5F,$E6,$2D,$C7,$FA,$E3,$CD,$FE
	.dc.b	$79,$F8,$4F,$FB,$4A,$7F,$B3,$FE,$12,$F8,$5F,$C5,$36,$3F,$B7,$4E
	.dc.b	$DF,$EA,$AF,$FB,$61,$FA,$4D,$7E,$C3,$5F,$A4,$D7,$FC,$31,$FA,$37
	.dc.b	$E4,$6F,$FA,$73,$78,$73,$CF,$FA,$8F,$75,$4F,$C6,$7F,$27,$F8,$5F
	.dc.b	$EF,$F1,$CE,$1F,$E7,$FF,$C2,$F1,$68,$C7,$D4,$F0,$3F,$37,$F2,$3F
	.dc.b	$B9,$BF,$F1,$53,$5F,$F1,$3B,$7A,$BF,$D0,$9D,$1F,$91,$FC,$4D,$F5
	.dc.b	$BF,$A4,$6D,$FF,$84,$DA,$F9,$06,$D2,$BD,$9F,$85,$BE,$D7,$FD,$57
	.dc.b	$28,$C5,$7F,$90,$9A,$FF,$4A,$8D,$4F,$E8,$FF,$2E,$FE,$13,$3F,$9F
	.dc.b	$6B,$FE,$45,$6B,$FE,$42,$6C,$BC,$BF,$F7,$EC,$FF,$94,$1A,$F7,$6B
	.dc.b	$FE,$1F,$8E,$1E,$1F,$A3,$E7,$3F,$D4,$33,$FE,$58,$6B,$E7,$6B,$C5
	.dc.b	$AF,$F9,$67,$E3,$DF,$25,$FB,$D6,$C7,$EF,$5A,$FF,$03,$6B,$E3,$FC
	.dc.b	$2D,$F8,$5F,$CA,$6C,$7F,$9A,$9A,$FC,$96,$BF,$C1,$7F,$28,$FF,$2C
	.dc.b	$BF,$C2,$DA,$FF,$11,$6B,$BF,$73,$F8,$97,$D7,$3E,$4E,$92,$7F,$8A
	.dc.b	$46,$5F,$E2,$DF,$34,$FA,$2F,$E3,$4D,$8C,$C6,$0F,$E3,$DF,$8E,$66
	.dc.b	$3F,$91,$35,$FE,$38,$D7,$FD,$04,$D7,$FC,$FF,$F7,$FE,$E7,$FD,$0C
	.dc.b	$D7,$3A,$7F,$2B,$8C,$74,$BE,$83,$F9,$ED,$7F,$31,$6B,$FC,$AD,$AF
	.dc.b	$FB,$CF,$D2,$FE,$BF,$F8,$1F,$E2,$3C,$0F,$D8,$48,$7E,$11,$FE,$ED
	.dc.b	$71,$1A,$73,$DF,$FD,$7A,$B9,$78,$3F,$E6,$DB,$7D,$F7,$F8,$5D,$B7
	.dc.b	$FE,$2C,$3B,$38,$70,$D7,$FB,$AD,$3E,$97,$B1,$23,$C7,$7D,$EF,$F8
	.dc.b	$87,$E1,$1F,$D0,$3B,$0F,$F0,$C9,$F6,$46,$3C,$4D,$F8,$5C,$1F,$E3
	.dc.b	$1F,$8E,$F8,$F6,$F8,$74,$8C,$5F,$2C,$3A,$65,$FC,$1B,$C5,$7C,$67
	.dc.b	$F2,$16,$C7,$98,$D4,$67,$C0,$FE,$4D,$F8,$FF,$F5,$3C,$3C,$AD,$AF
	.dc.b	$E2,$2D,$CB,$F8,$CD,$FC,$FF,$AC,$FF,$0A,$FD,$8B,$F0,$AF,$F7,$76
	.dc.b	$1D,$E3,$1F,$45,$B8,$C4,$E9,$FF,$2E,$CE,$9F,$A8,$D7,$31,$B7,$FC
	.dc.b	$4F,$F0,$6F,$8A,$FF,$0E,$B6,$3F,$CE,$0D,$7F,$95,$57,$3E,$1F,$97
	.dc.b	$E3,$FE,$B6,$7F,$97,$69,$FC,$2B,$F0,$1F,$E9,$12,$6F,$89,$87,$FA
	.dc.b	$67,$E4,$2F,$16,$99,$C3,$E7,$60,$FC,$2D,$F2,$5F,$D4,$DB,$3A,$6B
	.dc.b	$E2,$5F,$2D,$FD,$67,$E3,$5F,$23,$49,$5C,$6F,$FC,$8A,$BF,$08,$FF
	.dc.b	$07,$AF,$C2,$FF,$95,$F0,$7E,$9D,$BD,$3F,$ED,$1A,$DC,$3F,$0A,$FF
	.dc.b	$3E,$36,$FF,$4C,$73,$8C,$9F,$E8,$1F,$C5,$5F,$33,$E2,$6C,$FE,$3C
	.dc.b	$38,$7F,$AE,$1C,$CF,$8F,$F3,$AE,$EF,$EF,$0D,$8C,$C6,$3E,$2F,$FB
	.dc.b	$E7,$E0,$1F,$67,$5F,$D2,$B7,$5F,$D6,$B6,$FF,$B4,$E1,$CD,$FE,$D9
	.dc.b	$6F,$23,$FA,$2E,$1F,$C4,$23,$1F,$13,$FC,$2C,$FE,$19,$D5,$6F,$F1
	.dc.b	$BE,$0F,$E4,$0C,$3D,$38,$7E,$DE,$1C,$DF,$B7,$1F,$EC,$36,$BB,$E2
	.dc.b	$FD,$4F,$E1,$9F,$3E,$3F,$46,$D7,$F9,$10,$ED,$46,$63,$F9,$34,$63
	.dc.b	$8F,$FD,$13,$6F,$C2,$FF,$D0,$2D,$AF,$14,$78,$27,$FC,$4D,$F8,$23
	.dc.b	$C4,$BE,$64,$BF,$CD,$CE,$5B,$1F,$E2,$DF,$C1,$FF,$D1,$7C,$BF,$C7
	.dc.b	$1B,$3F,$E1,$8A,$FE,$50,$FE,$93,$E2,$DE,$A7,$F8,$D5,$FE,$05,$C1
	.dc.b	$FC,$79,$B8,$7F,$26,$77,$5C,$7C,$8F,$F1,$E7,$D2,$7F,$3E,$8F,$F1
	.dc.b	$2B,$7C,$25,$72,$97,$AA,$FC,$2B,$FD,$B0,$D8,$FF,$42,$46,$0F,$F6
	.dc.b	$E5,$3F,$09,$FE,$D5,$9F,$F3,$26,$E7,$FC,$E9,$B7,$DA,$FF,$DE,$F5
	.dc.b	$FD,$6B,$E6,$1E,$97,$9F,$2F,$F2,$0F,$D1,$7F,$A9,$3D,$13,$FE,$B1
	.dc.b	$9D,$46,$31,$FF,$4B,$35,$FF,$4E,$35,$FF,$4B,$35,$FE,$D1,$5E,$FF
	.dc.b	$1B,$DA,$48,$FF,$50,$65,$FC,$31,$FE,$C5,$AF,$F6,$6F,$91,$FF,$34
	.dc.b	$7A,$DF,$E7,$FD,$FF,$AB,$B6,$BD,$D8,$7F,$9B,$77,$FC,$DC,$7F,$8F
	.dc.b	$5A,$FC,$5F,$2B,$FC,$AB,$18,$8F,$F2,$DD,$7F,$A3,$6F,$AF,$DE,$78
	.dc.b	$0E,$97,$D7,$E0,$25,$FF,$2C,$C3,$FC,$5B,$C5,$F4,$FC,$0F,$3F,$DE
	.dc.b	$CA,$FE,$63,$9F,$FA,$3E,$BE,$C6,$1E,$CF,$AF,$FF,$33,$C9,$A4,$AF
	.dc.b	$87,$1F,$9F,$7F,$A2,$B6,$BF,$B9,$B7,$4F,$8B,$5F,$CC,$BF,$0B,$FF
	.dc.b	$5C,$78,$FF,$E1,$9F,$A4,$FE,$41,$18,$C5,$18,$F2,$3F,$DB,$EE,$BC
	.dc.b	$8F,$DF,$B9,$BE,$5E,$77,$F0,$2C,$39,$4E,$70,$E7,$39,$BE,$2E,$F0
	.dc.b	$EA,$DC,$39,$D7,$F8,$B7,$0E,$73,$C6,$32,$23,$2E,$EB,$9B,$6B,$73
	.dc.b	$FE,$6E,$C3,$C8,$4B,$F0,$57,$98,$F0,$61,$E6,$79,$19,$3C,$18,$76
	.dc.b	$6E,$1D,$A7,$84,$60,$E1,$87,$99,$C6,$1F,$2D,$F2,$DF,$39,$D1,$2B
	.dc.b	$9B,$88,$D1,$B8,$CB,$8B,$E6,$78,$EA,$9E,$28,$D3,$E4,$9E,$33,$AB
	.dc.b	$F8,$93,$F0,$8E,$27,$F3,$E3,$D8,$5B,$FC,$43,$6D,$5F,$DD,$25,$F2
	.dc.b	$EE,$CC,$36,$FF,$CB,$D3,$A9,$C9,$23,$E4,$CE,$F0,$FC,$BF,$CB,$7F
	.dc.b	$9C,$9F,$CC,$C6,$73,$FC,$93,$6B,$F1,$4D,$7F,$11,$5C,$67,$77,$F8
	.dc.b	$6A,$D6,$1E,$3B,$FB,$37,$B7,$9D,$6C,$8E,$B5,$B8,$F9,$26,$E3,$FA
	.dc.b	$5B,$E5,$7F,$99,$C7,$B9,$2B,$F2,$F3,$AC,$7C,$58,$73,$BF,$87,$5A
	.dc.b	$C7,$AC,$79,$0C,$9E,$32,$CF,$12,$63,$E3,$9E,$33,$E3,$AB,$6C,$3C
	.dc.b	$58,$F9,$6B,$9E,$1C,$5A,$78,$71,$EA,$FE,$E6,$B8,$E1,$FB,$CE,$8F
	.dc.b	$F4,$FF,$E0,$BF,$EA,$AE,$1F,$E2,$ED,$D3,$F9,$DD,$DC,$3F,$99,$CD
	.dc.b	$BF,$F5,$3C,$FF,$AB,$67,$75,$FC,$92,$31,$1F,$CE,$9B,$3F,$EA,$EF
	.dc.b	$17,$97,$17,$FA,$9E,$FF,$99,$F0,$62,$72,$C8,$FE,$6F,$C3,$71,$FE
	.dc.b	$5F,$F1,$7C,$8E,$7F,$1D,$6B,$F9,$EB,$E0,$C4,$D2,$47,$F3,$AF,$C0
	.dc.b	$3C,$67,$BF,$BF,$FC,$05,$ED,$E2,$7F,$D4,$91,$87,$FB,$2C,$7B,$A3
	.dc.b	$18,$FE,$D5,$AF,$F8,$D1,$FF,$06,$7B,$7B,$F0,$7F,$D1,$17,$FD,$BA
	.dc.b	$30,$79,$FE,$DF,$C5,$3A,$BF,$E9,$B1,$89,$FA,$5D,$CC,$69,$C3,$E5
	.dc.b	$AE,$3E,$8C,$51,$E1,$24,$7F,$7E,$7E,$AF,$90,$F8,$FF,$BD,$F8,$63
	.dc.b	$D7,$FC,$04,$E6,$FF,$D3,$33,$97,$FD,$37,$0F,$89,$DD,$F0,$6C,$FF
	.dc.b	$A8,$3D,$0E,$1B,$FF,$AB,$3C,$D7,$C6,$FE,$36,$73,$DB,$FE,$3F,$8C
	.dc.b	$79,$99,$D6,$E3,$CD,$FC,$B7,$88,$FC,$27,$FD,$2D,$3F,$ED,$B8,$C7
	.dc.b	$A9,$F9,$BF,$5D,$E7,$FF,$AD,$6B,$57,$C3,$6F,$F6,$1C,$E5,$FF,$68
	.dc.b	$B5,$FF,$37,$D7,$2F,$81,$FC,$FB,$6F,$F9,$B4,$FF,$32,$BF,$F1,$EC
	.dc.b	$76,$FC,$2B,$CF,$18,$9F,$F4,$7A,$8C,$DF,$B3,$EB,$3F,$05,$BF,$FA
	.dc.b	$46,$BC,$8F,$FA,$4E,$33,$83,$0F,$07,$C3,$72,$8F,$6B,$63,$FD,$B7
	.dc.b	$EC,$7F,$87,$67,$FC,$F3,$E7,$7E,$F7,$DF,$7C,$A7,$EF,$EB,$57,$FE
	.dc.b	$E2,$9C,$DF,$FB,$69,$B1,$9F,$AC,$E5,$7F,$C2,$61,$F8,$43,$D7,$C3
	.dc.b	$70,$FF,$61,$7E,$13,$EC,$7F,$D3,$71,$9A,$FF,$61,$CF,$D1,$7F,$D3
	.dc.b	$ED,$7E,$03,$53,$61,$E7,$F8,$E1,$AF,$8A,$30,$FF,$1D,$FA,$3F,$EC
	.dc.b	$F8,$7F,$BD,$9B,$83,$81,$3F,$3F,$CF,$7E,$10,$FA,$CE,$5E,$B7,$F2
	.dc.b	$78,$C5,$7F,$C0,$06,$E1,$CE,$39,$43,$F0,$9E,$17,$FC,$BF,$EA,$5E
	.dc.b	$8F,$F3,$1F,$8D,$F9,$FA,$78,$B2,$68,$BD,$E7,$F8,$3A,$D4,$AF,$8C
	.dc.b	$FA,$6B,$F9,$95,$F8,$CE,$6F,$27,$15,$D7,$D5,$F2,$63,$F9,$5F,$83
	.dc.b	$98,$77,$3F,$D9,$FF,$84,$F4,$CF,$F2,$D3,$FB,$18,$EF,$8F,$EA,$9A
	.dc.b	$F0,$C3,$F5,$71,$88,$FF,$0D,$FD,$9B,$D4,$E5,$5F,$C0,$23,$18,$F8
	.dc.b	$6B,$FD,$C5,$7E,$18,$FA,$DF,$26,$FA,$EF,$EB,$4D,$8E,$1A,$BE,$B5
	.dc.b	$F0,$9A,$BF,$EC,$11,$8A,$DB,$71,$F8,$AB,$E1,$37,$7F,$D9,$FC,$C7
	.dc.b	$C2,$7F,$69,$F9,$7F,$F5,$DE,$3E,$09,$CB,$E2,$8C,$E3,$F4,$27,$F4
	.dc.b	$5C,$5F,$69,$DC,$67,$BF,$B6,$1F,$0F,$92,$63,$F0,$E9,$F8,$43,$E1
	.dc.b	$33,$1F,$7C,$CB,$FD,$E1,$B3,$FC,$69,$A7,$F1,$C9,$E5,$1F,$C7,$BC
	.dc.b	$E7,$27,$CF,$AF,$E2,$54,$F9,$1F,$AD,$CD,$FC,$0E,$31,$87,$F6,$CF
	.dc.b	$9C,$FC,$75,$FF,$67,$6B,$FE,$D6,$6B,$FE,$F6,$6F,$BE,$FC,$8F,$F8
	.dc.b	$8C,$FF,$6F,$6B,$FE,$DE,$6B,$FE,$FE,$6C,$3E,$7D,$F2,$1F,$DE,$DB
	.dc.b	$87,$F3,$A9,$D4,$E4,$FF,$00,$7D,$03,$F0,$5A,$4A,$E1,$19,$79,$7E
	.dc.b	$5D,$FE,$3B,$6F,$C5,$7F,$82,$63,$19,$FF,$82,$5A,$FF,$82,$7F,$20
	.dc.b	$7C,$A7,$F8,$3D,$B7,$E5,$6C,$FF,$84,$5A,$FF,$43,$3F,$82,$FA,$0F
	.dc.b	$8B,$FF,$19,$B6,$BF,$DF,$AD,$9F,$F4,$77,$82,$8C,$47,$F8,$6F,$E0
	.dc.b	$7F,$5E,$FF,$87,$DA,$E7,$C5,$FF,$12,$46,$23,$F9,$37,$E0,$BF,$E0
	.dc.b	$5C,$3F,$B2,$E0,$8C,$4F,$81,$AF,$A6,$35,$F0,$5F,$E4,$AF,$C1,$5F
	.dc.b	$2D,$49,$87,$48,$CB,$F6,$7F,$26,$7D,$2F,$F8,$DD,$B1,$98,$C1,$FE
	.dc.b	$56,$FC,$17,$FC,$05,$97,$85,$AF,$F3,$F6,$BF,$CC,$1B,$CE,$FA,$CF
	.dc.b	$E1,$93,$FC,$AE,$31,$E5,$FF,$9A,$7F,$05,$FF,$15,$4E,$6B,$7E,$47
	.dc.b	$F3,$BB,$7E,$F5,$FD,$EF,$3F,$C9,$36,$FF,$C9,$CD,$7F,$C9,$CD,$7F
	.dc.b	$C0,$52,$FC,$19,$F8,$2F,$F9,$E7,$F0,$6F,$F3,$1E,$1F,$E5,$C9,$CB
	.dc.b	$FC,$8A,$33,$7E,$9F,$B5,$7A,$9F,$E6,$56,$D7,$F2,$8F,$27,$F8,$1D
	.dc.b	$6E,$BF,$CD,$5F,$82,$FF,$6E,$AF,$D5,$36,$7F,$CC,$CD,$7F,$CC,$CD
	.dc.b	$7D,$2B,$F0,$0F,$F4,$97,$CE,$3E,$7F,$FD,$B5,$7F,$E7,$4A,$D3,$FD
	.dc.b	$06,$74,$70,$FC,$17,$E6,$C7,$F8,$96,$11,$8C,$3F,$DC,$8B,$EB,$F8
	.dc.b	$72,$FF,$A0,$FE,$57,$C9,$4E,$24,$61,$BF,$07,$FE,$83,$8D,$E2,$FC
	.dc.b	$15,$F5,$1E,$A5,$F2,$B6,$30,$FB,$BF,$32,$7D,$2D,$3F,$D2,$9F,$11
	.dc.b	$E5,$E0,$F7,$4E,$9F,$F4,$AC,$67,$5F,$F4,$93,$5F,$F4,$93,$5F,$C2
	.dc.b	$B5,$F6,$7C,$1F,$F9,$4D,$FF,$4C,$35,$FC,$5B,$5D,$FD,$37,$FA,$FB
	.dc.b	$4F,$E8,$3F,$01,$FE,$A9,$AF,$F2,$E7,$77,$F8,$56,$73,$8F,$FA,$AA
	.dc.b	$72,$FE,$5B,$E0,$BF,$C1,$D5,$FA,$16,$C7,$D7,$FC,$17,$FA,$3E,$5F
	.dc.b	$AB,$9C,$AC,$13,$AB,$F0,$36,$F9,$7B,$BF,$8A,$8C,$5B,$B1,$F2,$5F
	.dc.b	$E9,$2B,$FF,$44,$36,$7F,$C4,$93,$FE,$32,$FA,$8F,$F3,$A7,$C1,$FF
	.dc.b	$A4,$3E,$07,$FB,$BE,$7F,$E3,$3A,$D6,$3F,$E9,$19,$FC,$C3,$6F,$87
	.dc.b	$77,$FB,$9A,$BF,$C7,$AD,$AD,$E3,$FC,$E4,$F3,$38,$7F,$9E,$4F,$F9
	.dc.b	$13,$E8,$3C,$47,$E9,$BF,$04,$7D,$FE,$32,$FD,$24,$61,$B1,$FE,$8D
	.dc.b	$F5,$78,$3F,$E7,$7C,$3F,$D1,$5F,$53,$FE,$01,$69,$5C,$27,$FC,$0D
	.dc.b	$F6,$1F,$E6,$CF,$C1,$6F,$80,$69,$2B,$FD,$4B,$3F,$BD,$E0,$E3,$E3
	.dc.b	$7F,$9B,$67,$2D,$7E,$14,$ED,$F8,$9B,$7E,$B7,$E0,$3F,$C2,$CD,$8F
	.dc.b	$F0,$A7,$83,$FD,$F1,$B3,$FE,$1B,$E8,$70,$7F,$9B,$E7,$FD,$6F,$18
	.dc.b	$F9,$0D,$4E,$67,$E9,$4E,$5F,$F1,$07,$7B,$E8,$BC,$2D,$8C,$C6,$0C
	.dc.b	$FB,$92,$38,$B5,$F5,$CB,$EA,$F9,$92,$3B,$BE,$9F,$C1,$1F,$59,$49
	.dc.b	$3E,$A8,$CB,$FE,$19,$F0,$1F,$57,$FE,$3B,$F2,$BF,$CB,$CD,$8C,$CE
	.dc.b	$E3,$E3,$76,$3F,$04,$A4,$9F,$8F,$19,$7F,$DB,$FC,$EF,$C8,$9E,$07
	.dc.b	$F9,$26,$30,$D8,$CC,$F0,$6B,$FC,$63,$E3,$BC,$DE,$4F,$F9,$26,$3F
	.dc.b	$C6,$4D,$74,$7D,$8F,$11,$F4,$DF,$75,$B1,$98,$C1,$F8,$1F,$C1,$3F
	.dc.b	$C4,$51,$F8,$06,$BF,$79,$AF,$F8,$A7,$99,$18,$8F,$ED,$BC,$DF,$E7
	.dc.b	$C7,$FC,$8C,$D7,$F7,$1C,$3F,$85,$F9,$2F,$A1,$39,$9E,$9E,$15,$7F
	.dc.b	$AA,$35,$FE,$38,$DC,$3F,$77,$5F,$99,$FC,$11,$F3,$94,$97,$F9,$A8
	.dc.b	$CB,$FE,$59,$FC,$11,$F2,$9F,$81,$6C,$66,$30,$7E,$17,$F2,$0F,$B9
	.dc.b	$1E,$A6,$BF,$E4,$BD,$11,$88,$FF,$07,$B5,$FE,$CD,$F8,$27,$F1,$27
	.dc.b	$FC,$D4,$D7,$E2,$B5,$E8,$D7,$FC,$33,$EF,$7F,$11,$AF,$F0,$D4,$7E
	.dc.b	$66,$30,$67,$E0,$5F,$89,$FE,$75,$B3,$FE,$79,$6B,$F5,$1A,$FA,$FF
	.dc.b	$27,$FF,$62,$3F,$E7,$F6,$BF,$69,$AE,$FD,$BF,$D5,$7E,$08,$F6,$29
	.dc.b	$27,$FD,$11,$19,$7F,$D1,$5F,$91,$FF,$57,$6F,$1F,$EC,$E3,$11,$F8
	.dc.b	$86,$BF,$3F,$AD,$F2,$9F,$B9,$6C,$7F,$A5,$5A,$FE,$41,$AF,$F9,$DF
	.dc.b	$ED,$FF,$80,$7C,$89,$A4,$8F,$F4,$F7,$E4,$9F,$EC,$1E,$13,$49,$EB
	.dc.b	$7F,$09,$F9,$47,$D6,$52,$5F,$FA,$4D,$FE,$91,$F8,$E3,$ED,$A9,$27
	.dc.b	$FD,$99,$19,$6F,$BB,$C0,$8E,$0D,$FE,$2F,$D0,$8E,$0E,$7E,$E7,$80
	.dc.b	$8E,$0F,$F6,$D7,$E0,$1F,$C1,$1F,$ED,$81,$C3,$FD,$BD,$AF,$FA,$A5
	.dc.b	$AF,$F5,$BF,$8D,$F8,$2F,$F2,$06,$BF,$DD,$FC,$B4,$62,$3F,$D6,$51
	.dc.b	$8C,$FF,$B6,$53,$F1,$1A,$4A,$F3,$BF,$DC,$7F,$06,$F8,$D1,$E8,$6B
	.dc.b	$FE,$05,$6B,$FE,$BF,$6A,$F9,$97,$E0,$CF,$B6,$FF,$7E,$46,$1B,$19
	.dc.b	$9F,$F5,$AB,$5F,$F6,$1B,$F0,$A7,$9F,$99,$C0,$93,$FE,$BC,$F8,$BF
	.dc.b	$F2,$EB,$5E,$E6,$1F,$F6,$6B,$5F,$F5,$F2,$FB,$9D,$92,$3A,$3F,$DF
	.dc.b	$BF,$0D,$FF,$32,$C7,$DA,$F8,$14,$62,$3F,$C0,$AD,$7F,$C4,$ED,$43
	.dc.b	$EE,$5F,$BB,$3B,$7F,$0B,$39,$8C,$3F,$88,$9D,$CE,$9B,$1E,$3A,$FF
	.dc.b	$53,$B7,$E5,$7D,$04,$7F,$DC,$3F,$01,$FE,$15,$7F,$DD,$ED,$7F,$DC
	.dc.b	$0D,$7C,$FA,$FF,$BE,$7E,$0F,$EC,$CB,$E3,$E9,$2B,$AB,$F7,$BE,$C3
	.dc.b	$FB,$6C,$7F,$BA,$DA,$F4,$6B,$F8,$76,$D9,$FF,$58,$47,$FB,$E1,$AF
	.dc.b	$F9,$56,$35,$F1,$BD,$CF,$F2,$47,$E0,$5F,$CA,$27,$E7,$87,$0F,$D2
	.dc.b	$6B,$9F,$2F,$F4,$BF,$84,$C8,$FD,$CB,$5F,$DC,$F4,$FF,$8D,$5F,$DC
	.dc.b	$7C,$47,$EB,$08,$FE,$FB,$1F,$C6,$7C,$8F,$E8,$39,$4D,$24,$EE,$3F
	.dc.b	$D6,$ED,$F9,$4E,$04,$7E,$F7,$9B,$F9,$32,$38,$3F,$29,$D3,$FD,$6D
	.dc.b	$FF,$47,$37,$A3,$FC,$2F,$ED,$FF,$96,$23,$07,$F0,$39,$E3,$DD,$92
	.dc.b	$3F,$91,$FD,$81,$F8,$D7,$E4,$5B,$19,$8C,$1F,$E6,$4F,$B9,$7D,$B5
	.dc.b	$24,$FE,$A6,$32,$FC,$7F,$BC,$7C,$3F,$F0,$66,$C6,$63,$07,$E9,$BC
	.dc.b	$E7,$E0,$4F,$97,$C3,$6B,$F9,$7C,$EA,$72,$7F,$12,$F8,$4D,$4F,$E2
	.dc.b	$DA,$FE,$23,$97,$FC,$DD,$3B,$38,$7D,$74,$9F,$BD,$48,$C4,$7F,$8C
	.dc.b	$B0,$7D,$E7,$F9,$85,$B1,$FD,$0B,$D5,$F4,$CE,$4F,$E4,$BF,$81,$3F
	.dc.b	$00,$A4,$AF,$C6,$46,$5F,$C7,$7E,$06,$FC,$13,$F9,$6B,$63,$31,$83
	.dc.b	$C5,$F4,$12,$3C,$59,$7F,$15,$59,$7C,$C9,$FF,$3B,$6F,$27,$FC,$F8
	.dc.b	$D7,$FD,$2B,$F0,$0F,$C6,$BF,$D3,$CD,$8F,$F4,$EB,$5D,$C7,$F4,$2F
	.dc.b	$96,$C4,$7C,$F6,$BF,$C4,$1A,$FF,$4B,$6F,$13,$F7,$CF,$F7,$54,$FE
	.dc.b	$BD,$AF,$F0,$56,$BF,$6D,$AF,$F1,$C7,$FC,$91,$1C,$FE,$53,$FA,$AC
	.dc.b	$60,$9F,$27,$82,$57,$77,$7E,$1C,$8F,$13,$5F,$17,$85,$D0,$8F,$9D
	.dc.b	$FF,$6F,$7E,$0F,$FE,$B1,$F0,$BF,$DC,$35,$FC,$62,$31,$5F,$ED,$F8
	.dc.b	$C7,$CC,$FC,$0F,$A1,$FE,$17,$3F,$D3,$2B,$FD,$17,$18,$84,$72,$8C
	.dc.b	$36,$B9,$DF,$82,$B9,$79,$FF,$CA,$2F,$78,$77,$BD,$D7,$8C,$F0,$46
	.dc.b	$5E,$27,$33,$C2,$E5,$B7,$9C,$3C,$51,$C0,$E8,$68,$AE,$9C,$5F,$AB
	.dc.b	$0F,$36,$1E,$48,$F3,$DF,$5C,$3C,$18,$7C,$4A,$D1,$30,$F4,$D7,$8F
	.dc.b	$0E,$71,$83,$C3,$7C,$8E,$71,$A8,$F1,$D7,$36,$9E,$BC,$3C,$6F,$2B
	.dc.b	$F2,$1D,$DA,$64,$91,$ED,$72,$F2,$24,$FE,$16,$BC,$58,$7E,$1A,$BF
	.dc.b	$AF,$9F,$24,$AE,$5F,$B2,$63,$E4,$BF,$CF,$EF,$E2,$5A,$FE,$21,$B1
	.dc.b	$FE,$C0,$F1,$FF,$C2,$AD,$9F,$AB,$3F,$E8,$DC,$3F,$1B,$F0,$EE,$4F
	.dc.b	$BD,$39,$CB,$F3,$8D,$BF,$06,$3F,$83,$BE,$11,$D2,$38,$91,$FC,$1C
	.dc.b	$6B,$0F,$CD,$D7,$39,$F4,$56,$E7,$C7,$85,$C3,$C3,$36,$BF,$51,$87
	.dc.b	$39,$CC,$78,$49,$7C,$9B,$1A,$8F,$0D,$7A,$6F,$A5,$78,$8D,$9D,$9C
	.dc.b	$E3,$F2,$6B,$8C,$FB,$27,$38,$FA,$EB,$58,$7F,$03,$C7,$CD,$7F,$20
	.dc.b	$EA,$77,$C3,$CC,$74,$24,$F9,$CE,$58,$7C,$1E,$1F,$E5,$5E,$C7,$D0
	.dc.b	$BF,$E4,$CD,$7F,$93,$35,$FF,$3E,$70,$A3,$13,$FC,$CF,$CA,$B3,$FE
	.dc.b	$6B,$BF,$E6,$91,$83,$3F,$19,$23,$F9,$C3,$70,$FE,$71,$F3,$FC,$89
	.dc.b	$3F,$CF,$5E,$BF,$43,$23,$FA,$03,$67,$FA,$07,$81,$E2,$24,$FF,$45
	.dc.b	$7C,$BE,$44,$8F,$E9,$0D,$9F,$EF,$1F,$33,$EF,$EE,$FF,$06,$B7,$BC
	.dc.b	$D3,$1E,$34,$C3,$FB,$5B,$6F,$FC,$AD,$DB,$FE,$7D,$9F,$F2,$23,$63
	.dc.b	$3E,$DF,$CA,$8C,$47,$F8,$47,$C3,$FF,$04,$D7,$F7,$EF,$17,$E7,$C6
	.dc.b	$27,$FC,$C4,$D7,$FC,$BF,$9F,$F8,$93,$E1,$3D,$04,$BF,$EF,$EF,$D8
	.dc.b	$FA,$47,$93,$49,$3F,$E5,$98,$CB,$7E,$2F,$81,$1C,$1F,$E2,$EE,$0F
	.dc.b	$F4,$23,$7D,$AF,$C4,$3F,$E4,$26,$C6,$B2,$FF,$3D,$7E,$0E,$F1,$3F
	.dc.b	$16,$DB,$FF,$35,$36,$77,$3F,$93,$F0,$DF,$95,$7F,$95,$2F,$FD,$68
	.dc.b	$DA,$FE,$47,$3A,$F4,$78,$4E,$67,$FD,$0B,$F0,$1F,$E4,$AF,$C1,$3F
	.dc.b	$3C,$DC,$3F,$D6,$1D,$09,$CB,$63,$FD,$85,$F6,$DF,$84,$6F,$4F,$96
	.dc.b	$31,$87,$89,$AF,$FA,$53,$34,$3E,$D3,$81,$2F,$AB,$CF,$CF,$91,$E0
	.dc.b	$6D,$FF,$A8,$FF,$06,$FF,$66,$CF,$FA,$45,$B1,$93,$F6,$EE,$1F,$F2
	.dc.b	$F2,$F2,$B2,$49,$E4,$FF,$01,$97,$E0,$CF,$47,$0D,$AF,$F6,$CB,$67
	.dc.b	$FA,$99,$C2,$77,$F5,$92,$B6,$D9,$DF,$E1,$6D,$7C,$0D,$AF,$F7,$2B
	.dc.b	$67,$73,$FC,$4B,$F0,$77,$89,$49,$7F,$EB,$38,$CB,$7C,$4E,$04,$70
	.dc.b	$7F,$AE,$7E,$B9,$F7,$7F,$C7,$9B,$19,$8C,$19,$E5,$91,$C5,$B9,$7E
	.dc.b	$07,$C3,$FC,$21,$27,$F1,$2F,$E4,$3F,$01,$FF,$53,$C7,$FB,$FD,$B3
	.dc.b	$FD,$F3,$A1,$18,$8F,$CC,$B5,$1F,$27,$9B,$EE,$57,$E5,$B8,$FF,$28
	.dc.b	$47,$FD,$AD,$F5,$3F,$E8,$96,$FB,$3F,$CF,$23,$17,$FE,$EF,$6B,$FE
	.dc.b	$A5,$5F,$15,$B2,$57,$47,$FA,$2F,$CA,$FE,$BE,$BC,$BE,$82,$31,$08
	.dc.b	$C7,$7F,$03,$8F,$E1,$33,$0F,$F7,$65,$A3,$17,$F8,$ED,$7F,$83,$3F
	.dc.b	$E2,$26,$BF,$10,$F9,$3E,$0F,$E6,$E3,$31,$F2,$3A,$DF,$13,$F9,$C2
	.dc.b	$63,$F9,$C7,$FB,$17,$CA,$7E,$57,$D5,$BD,$7F,$EF,$CA,$FC,$FC,$E5
	.dc.b	$B1,$9C,$3F,$47,$87,$FA,$F6,$7F,$4F,$D3,$FA,$A2,$3F,$AA,$F5,$7F
	.dc.b	$5A,$47,$07,$EB,$B4,$FD,$89,$1C,$1F,$B2,$CB,$F6,$A4,$70,$7E,$DB
	.dc.b	$DC,$FD,$C9,$1C,$1F,$ED,$5F,$99,$7E,$09,$FD,$8D,$B1,$98,$C1,$FD
	.dc.b	$9F,$E6,$DF,$57,$49,$3F,$ED,$F8,$CB,$FE,$E1,$F5,$5F,$4B,$FD,$BD
	.dc.b	$B1,$98,$C1,$FD,$CF,$E7,$9F,$51,$49,$3F,$EE,$B8,$CB,$FE,$ED,$FA
	.dc.b	$07,$CB,$7F,$7B,$6C,$66,$30,$7F,$7F,$FB,$B7,$D0,$D2,$4F,$FB,$DE
	.dc.b	$32,$FF,$BE,$7F,$28,$BC,$DF,$E7,$0F,$6D,$18,$8F,$EE,$6D,$7F,$6B
	.dc.b	$1A,$BF,$F0,$77,$D2,$BE,$5E,$93,$0F,$E5,$34,$9D,$46,$3E,$99,$F7
	.dc.b	$9F,$E1,$76,$C6,$63,$06,$7E,$06,$47,$16,$E5,$E9,$FA,$97,$47,$F8
	.dc.b	$85,$B5,$E2,$6C,$EE,$7F,$C4,$DF,$6F,$FE,$3F,$8F,$84,$D7,$FC,$26
	.dc.b	$D7,$E0,$B5,$FE,$A1,$F3,$BF,$D4,$DF,$E3,$3E,$5A,$31,$1A,$3F,$CA
	.dc.b	$51,$96,$FD,$BB,$E6,$69,$32,$FE,$A3,$49,$D4,$7C,$CF,$83,$7C,$4F
	.dc.b	$F8,$FD,$B1,$98,$C1,$E0,$FB,$4F,$DD,$63,$FC,$2F,$04,$62,$3D,$2D
	.dc.b	$7F,$B6,$36,$1F,$8B,$7F,$28,$BF,$2B,$6B,$FD,$73,$3A,$F3,$FD,$F3
	.dc.b	$99,$FF,$1E,$FD,$97,$A9,$FC,$83,$5C,$E9,$FE,$59,$8C,$47,$F4,$4F
	.dc.b	$29,$C8,$F9,$EF,$E3,$6D,$CF,$F8,$CB,$6B,$FD,$97,$5C,$2B,$E8,$FE
	.dc.b	$0D,$E3,$9F,$AE,$D7,$FC,$14,$DD,$0E,$7A,$1E,$9F,$F8,$A6,$BF,$50
	.dc.b	$D7,$FC,$B2,$D7,$F4,$ED,$7F,$C2,$DF,$72,$F9,$8F,$C1,$B7,$4F,$BE
	.dc.b	$D9,$F9,$6D,$7F,$C6,$1F,$BF,$7D,$E7,$FC,$8A,$D7,$3C,$7F,$E4,$D8
	.dc.b	$C4,$7F,$6C,$F4,$10,$F9,$1D,$13,$0F,$F0,$FC,$EA,$39,$7C,$94,$9C
	.dc.b	$D7,$E8,$6B,$9C,$FE,$73,$EE,$1F,$45,$49,$7F,$9A,$8C,$BD,$BE,$29
	.dc.b	$F2,$FF,$E6,$36,$C6,$63,$07,$F8,$E7,$3F,$F0,$2F,$CB,$FF,$8D,$1F
	.dc.b	$EC,$8D,$72,$57,$5F,$73,$FE,$2C,$E6,$BF,$0E,$69,$2B,$FC,$7F,$3F
	.dc.b	$17,$EE,$B8,$F0,$1A,$49,$FC,$FE,$8F,$A5,$FE,$ED,$2D,$3F,$81,$CE
	.dc.b	$6B,$FC,$3B,$39,$7F,$CB,$11,$97,$B4,$FD,$7E,$FF,$F0,$AC,$70,$8F
	.dc.b	$B5,$F7,$4F,$81,$EF,$18,$6C,$FF,$94,$67,$4F,$13,$BE,$3A,$E3,$4B
	.dc.b	$E7,$BE,$64,$61,$B1,$9A,$FC,$2B,$5F,$F4,$AA,$F9,$FE,$8D,$8F,$05
	.dc.b	$76,$9C,$D7,$87,$27,$D1,$52,$5F,$0C,$39,$3F,$10,$76,$7E,$CB,$D9
	.dc.b	$F7,$B6,$3D,$CD,$78,$77,$BE,$E9,$E6,$FC,$68,$FF,$58,$FC,$3D,$F9
	.dc.b	$63,$4E,$BE,$17,$F8,$13,$0F,$95,$D0,$F6,$3F,$DD,$71,$86,$BE,$78
	.dc.b	$D7,$6F,$7F,$A4,$FE,$01,$3F,$4F,$27,$E1,$3F,$7A,$3C,$ED,$EA,$FC
	.dc.b	$34,$EB,$CC,$FB,$B8,$F4,$BE,$5F,$77,$FB,$07,$FA,$3B,$5F,$E6,$6D
	.dc.b	$7F,$A0,$35,$FE,$AF,$F5,$BF,$EA,$37,$C8,$D7,$E4,$B7,$0F,$1C,$F5
	.dc.b	$7F,$35,$C5,$FE,$F2,$F7,$7E,$9F,$CF,$FF,$87,$B0,$FA,$6D,$7E,$9F
	.dc.b	$DD,$FE,$26,$FD,$BF,$A3,$BE,$A3,$F7,$9B,$23,$17,$FD,$89,$AF,$F4
	.dc.b	$E6,$BF,$D6,$FC,$FF,$F4,$FB,$F3,$5A,$FE,$09,$B7,$F3,$1F,$23,$F3
	.dc.b	$BE,$F3,$F1,$ED,$EF,$FC,$34,$63,$C3,$FE,$FA,$BE,$A7,$F1,$5F,$8C
	.dc.b	$FF,$B4,$71,$FE,$EE,$D7,$FB,$1B,$5F,$EE,$0D,$7F,$D1,$78,$3E,$7F
	.dc.b	$F7,$3F,$29,$FC,$3D,$B1,$99,$FC,$EF,$BD,$F8,$A7,$F4,$94,$FC,$5D
	.dc.b	$F6,$BC,$9B,$3F,$C0,$67,$53,$FA,$8F,$BD,$FE,$D5,$39,$FC,$97,$FD
	.dc.b	$F4,$FF,$B8,$29,$18,$8F,$F4,$91,$FE,$9F,$70,$FF,$A5,$63,$7F,$49
	.dc.b	$FC,$D5,$FF,$52,$5F,$B1,$F1,$CF,$A1,$B9,$7E,$2B,$E1,$7F,$C3,$73
	.dc.b	$E7,$FC,$61,$F8,$3F,$C4,$6C,$66,$3F,$80,$FE,$01,$FE,$A0,$8F,$77
	.dc.b	$E4,$AF,$96,$F1,$36,$3D,$8D,$7D,$79,$23,$11,$EB,$FA,$07,$CF,$FD
	.dc.b	$46,$C7,$CB,$6B,$FC,$57,$C0,$7C,$97,$E0,$76,$FE,$7D,$8B,$F6,$1F
	.dc.b	$93,$18,$6F,$57,$FB,$31,$B8,$7F,$1E,$F2,$BF,$B2,$BF,$67,$DD,$7E
	.dc.b	$01,$F3,$5B,$1F,$5D,$AF,$B1,$AF,$D5,$FC,$6B,$FC,$9C,$FE,$0D,$AF
	.dc.b	$E2,$9B,$8F,$EF,$3E,$63,$FA,$B4,$FE,$3B,$F0,$27,$E0,$3F,$96,$6C
	.dc.b	$7E,$35,$AF,$C3,$F9,$CB,$7C,$BF,$01,$7E,$0D,$FD,$DD,$B5,$FD,$ED
	.dc.b	$AF,$E3,$83,$87,$F2,$3F,$8B,$7F,$36,$7F,$AE,$07,$0F,$F4,$73,$1E
	.dc.b	$B7,$F6,$45,$8F,$FB,$E5,$E3,$3F,$D4,$9B,$EE,$6E,$31,$1C,$E5,$F6
	.dc.b	$DC,$D7,$3F,$F8,$45,$B3,$FE,$15,$6B,$EB,$6B,$FA,$AF,$AE,$7D,$17
	.dc.b	$95,$B1,$EC,$6B,$EB,$FB,$FF,$E6,$1C,$E8,$91,$F6,$23,$10,$8C,$3A
	.dc.b	$F9,$4F,$E9,$3F,$82,$7D,$4A,$DE,$FF,$EC,$8A,$FA,$5F,$3D,$F5,$48
	.dc.b	$FF,$91,$3D,$8F,$E0,$5C,$7F,$E1,$9F,$86,$46,$30,$FF,$27,$36,$BC
	.dc.b	$D8,$75,$7F,$8B,$FE,$09,$EA,$5D,$C8,$C3,$7E,$3F,$F9,$1C,$7F,$06
	.dc.b	$C3,$E0,$4B,$F1,$2F,$F6,$4D,$FF,$1D,$6B,$FC,$91,$B9,$FE,$CA,$FA
	.dc.b	$FD,$3B,$E4,$BF,$97,$B6,$FF,$93,$35,$FE,$5A,$8F,$E6,$BF,$89,$FF
	.dc.b	$31,$BF,$DE,$B7,$2F,$2B,$6B,$DB,$5C,$EB,$FC,$81,$F5,$EF,$C1,$9F
	.dc.b	$53,$FE,$38,$8C,$36,$33,$7F,$E2,$AE,$0F,$E8,$4F,$F8,$8F,$87,$E8
	.dc.b	$FD,$47,$F8,$EE,$4D,$39,$5F,$F8,$BF,$E9,$65,$FF,$8D,$3E,$43,$E2
	.dc.b	$ED,$F6,$2F,$FA,$8F,$C8,$BE,$EE,$92,$FF,$A5,$47,$F8,$56,$BE,$9F
	.dc.b	$E2,$EF,$BC,$FE,$B2,$DA,$FE,$A0,$D7,$FA,$BD,$75,$7F,$AA,$7E,$0F
	.dc.b	$FE,$E3,$7F,$AD,$B7,$87,$F9,$4B,$70,$FC,$0B,$FE,$2E,$8F,$E5,$BF
	.dc.b	$35,$F8,$FE,$83,$4C,$4F,$1B,$FD,$3F,$D5,$7F,$03,$EF,$34,$FE,$0D
	.dc.b	$F8,$3F,$F4,$B9,$FD,$87,$CC,$FF,$A3,$6B,$FC,$ED,$1F,$D8,$9A,$FE
	.dc.b	$C1,$AF,$ED,$7C,$2F,$F3,$E6,$7F,$CA,$9B,$A7,$ED,$9B,$3F,$BA,$BF
	.dc.b	$F0,$0F,$5F,$F5,$5B,$FE,$92,$DB,$FD,$0B,$6B,$F5,$2F,$F0,$69,$FE
	.dc.b	$9F,$F7,$9F,$C6,$1B,$E4,$FF,$83,$23,$17,$CB,$D8,$FF,$41,$78,$3F
	.dc.b	$E0,$2C,$7F,$18,$DC,$3F,$7C,$DA,$FD,$EF,$C5,$7E,$3E,$A1,$F9,$53
	.dc.b	$9F,$C4,$FE,$6F,$F0,$7F,$F4,$8C,$7F,$9E,$7C,$04,$63,$1F,$D1,$B7
	.dc.b	$0F,$F6,$B6,$3F,$C6,$3B,$DF,$83,$FE,$27,$1F,$E3,$B1,$8E,$AF,$F4
	.dc.b	$BC,$E6,$7F,$DB,$CD,$53,$FE,$D4,$E5,$F7,$92,$3F,$D2,$B3,$F9,$5C
	.dc.b	$BF,$D6,$98,$BE,$FD,$F8,$1F,$FA,$B2,$30,$D8,$CD,$FF,$8B,$7F,$07
	.dc.b	$7C,$C7,$F8,$EB,$F0,$27,$E0,$34,$95,$FE,$B3,$C7,$9B,$FE,$35,$F9
	.dc.b	$5F,$AA,$FF,$24,$6F,$79,$A7,$82,$3F,$C8,$3F,$5F,$FE,$12,$D7,$FC
	.dc.b	$1E,$DB,$FE,$4E,$D9,$E9,$D3,$FE,$4E,$FC,$1D,$F7,$3F,$E3,$76,$CF
	.dc.b	$F9,$05,$AF,$F8,$99,$AF,$F8,$FF,$F0,$5F,$E7,$98,$FF,$87,$9B,$8F
	.dc.b	$F3,$66,$CF,$F3,$EA,$E7,$5F,$E8,$8E,$0F,$E3,$6D,$F9,$2F,$F5,$DC
	.dc.b	$63,$0E,$5E,$DF,$F9,$93,$E8,$9F,$7D,$49,$97,$FB,$02,$F8,$2E,$E7
	.dc.b	$E0,$9F,$E8,$68,$FE,$D1,$18,$F9,$BF,$F4,$7C,$E6,$78,$36,$78,$FC
	.dc.b	$5F,$FB,$1E,$3E,$54,$62,$7F,$D7,$51,$FE,$94,$D7,$FC,$7F,$F8,$37
	.dc.b	$E4,$27,$FC,$D4,$D7,$FC,$9E,$D7,$FC,$D0,$D7,$FA,$EB,$FC,$13,$94
	.dc.b	$D2,$76,$FF,$A1,$97,$DC,$D2,$57,$F9,$FE,$3F,$A9,$57,$2E,$EE,$04
	.dc.b	$9D,$57,$3F,$13,$FC,$F1,$7F,$E6,$29,$F1,$C6,$0C,$FB,$BE,$12,$47
	.dc.b	$78,$FF,$17,$7E,$0E,$F9,$DF,$85,$F5,$FF,$E1,$A8,$C6,$9F,$E1,$98
	.dc.b	$C4,$EB,$E0,$7F,$C3,$34,$F4,$BF,$C3,$75,$AB,$E1,$5A,$BE,$18,$FF
	.dc.b	$88,$F0,$EA,$AF,$C1,$87,$2B,$E1,$87,$2A,$E1,$87,$25,$87,$2C,$38
	.dc.b	$C6,$CC,$CE,$E7,$57,$D2,$7C,$0F,$AB,$0F,$04,$76,$C3,$95,$70,$8C
	.dc.b	$1C,$0F,$54,$F0,$7B,$CF,$0B,$F0,$1E,$19,$D5,$73,$AD,$9C,$CB,$87
	.dc.b	$A9,$C4,$6C,$95,$CE,$3B,$5C,$77,$19,$6C,$66,$BE,$A4,$74,$BF,$1D
	.dc.b	$79,$AB,$58,$7C,$3A,$A7,$C0,$BE,$27,$94,$E6,$FF,$B3,$B0,$FB,$6D
	.dc.b	$BF,$D3,$3E,$E3,$5F,$3B,$FE,$EB,$7E,$CB,$67,$FD,$ED,$C5,$FE,$8B
	.dc.b	$9C,$B6,$33,$1F,$12,$7F,$DF,$57,$F9,$7F,$8C,$75,$7D,$89,$CE,$5F
	.dc.b	$96,$6D,$FE,$45,$AF,$E2,$AB,$8B,$D7,$1E,$44,$7E,$9C,$6B,$0F,$D3
	.dc.b	$57,$2C,$7C,$B1,$8B,$FC,$4D,$6F,$0F,$14,$73,$9D,$61,$D3,$1C,$56
	.dc.b	$A3,$C4,$64,$B3,$C0,$91,$FB,$2B,$ED,$3E,$DB,$E6,$D3,$95,$70,$70
	.dc.b	$D7,$E1,$5F,$07,$E8,$9E,$0B,$ED,$19,$C3,$DF,$1E,$7A,$FA,$53,$9C
	.dc.b	$7E,$21,$F1,$71,$F2,$E1,$D2,$BF,$D5,$3C,$5F,$CA,$3F,$00,$FA,$3C
	.dc.b	$9F,$CD,$1B,$8F,$F2,$EF,$7A,$E1,$FE,$92,$E3,$56,$30,$FE,$63,$5F
	.dc.b	$EE,$2A,$E1,$7F,$CF,$23,$11,$C5,$B3,$FC,$E3,$C4,$E5,$87,$F8,$0A
	.dc.b	$FF,$9A,$7D,$96,$27,$2C,$8F,$E7,$1F,$61,$71,$67,$FD,$33,$A7,$99
	.dc.b	$D3,$EC,$6D,$FF,$3F,$7C,$39,$1A,$48,$FE,$A7,$F3,$7E,$43,$CB,$E7
	.dc.b	$58,$FE,$90,$F2,$23,$FD,$69,$F6,$F8,$2E,$EF,$8C,$D9,$FE,$E5,$E6
	.dc.b	$E2,$7E,$6F,$CB,$7F,$8A,$27,$27,$FB,$67,$4F,$D2,$F0,$B0,$FF,$51
	.dc.b	$CE,$B6,$F8,$2F,$F8,$2F,$E7,$3E,$A7,$90,$6F,$C6,$61,$B3,$FE,$E2
	.dc.b	$AF,$F0,$94,$FD,$4A,$FF,$7B,$CE,$67,$FD,$A3,$18,$7F,$C3,$DC,$BF
	.dc.b	$81,$F2,$7F,$D8,$15,$E7,$3F,$C5,$7E,$6E,$67,$F0,$DE,$6F,$F8,$D6
	.dc.b	$75,$7F,$ED,$C8,$C3,$FE,$36,$FC,$1F,$77,$F8,$D1,$F1,$D4,$FC,$9D
	.dc.b	$FF,$AA,$FE,$07,$0F,$F9,$1F,$93,$FD,$68,$D8,$FC,$B7,$A2,$F1,$F8
	.dc.b	$D6,$B2,$FF,$63,$72,$BE,$EF,$F3,$FE,$3F,$D9,$6F,$D7,$3E,$CB,$35
	.dc.b	$BD,$3D,$76,$AD,$47,$C9,$8C,$57,$E8,$9B,$19,$9D,$E1,$F4,$FD,$AF
	.dc.b	$F3,$67,$E0,$9F,$B0,$6D,$76,$C5,$F5,$F9,$F4,$97,$E4,$EE,$97,$BE
	.dc.b	$AB,$97,$B9,$F4,$9E,$AF,$56,$CF,$F9,$D4,$FD,$E7,$B3,$FC,$F1,$FA
	.dc.b	$FD,$CF,$79,$86,$A7,$F8,$14,$61,$DB,$63,$F2,$A2,$79,$79,$FC,$23
	.dc.b	$F8,$68,$F8,$1F,$AD,$7F,$86,$8C,$73,$6B,$F0,$2F,$F2,$BC,$E6,$B7
	.dc.b	$5D,$84,$E9,$FC,$D3,$5F,$D0,$35,$F5,$9C,$3C,$8F,$E2,$EF,$2F,$2F
	.dc.b	$C0,$79,$AB,$FD,$4F,$7E,$5B,$FE,$31,$5E,$EC,$9F,$76,$F3,$FF,$93
	.dc.b	$36,$FE,$5F,$E0,$FF,$C4,$5F,$9B,$19,$BE,$53,$96,$F9,$7F,$CB,$E3
	.dc.b	$FD,$27,$5C,$A3,$F9,$97,$C3,$FF,$9E,$E3,$EB,$B6,$BC,$9F,$83,$BF
	.dc.b	$07,$DF,$0F,$F3,$98,$C5,$FD,$E6,$D7,$0E,$7D,$5F,$E9,$3A,$7F,$2F
	.dc.b	$E6,$7F,$78,$EE,$FC,$9F,$7F,$E9,$3C,$BD,$37,$3F,$CC,$47,$E6,$FF
	.dc.b	$06,$78,$79,$D3,$FD,$C1,$18,$AF,$E0,$BE,$0A,$31,$3F,$B6,$DF,$95
	.dc.b	$FF,$51,$F9,$37,$B7,$4F,$EA,$9F,$6C,$F8,$1C,$67,$F9,$A8,$C3,$63
	.dc.b	$7F,$6B,$C1,$FE,$B9,$F2,$EF,$B6,$A7,$F5,$FF,$C1,$FE,$0D,$FE,$8B
	.dc.b	$B5,$5B,$8F,$D3,$CE,$B6,$FD,$03,$68,$FF,$AD,$A7,$7C,$5F,$C3,$2A
	.dc.b	$47,$E8,$78,$3F,$3B,$C9,$F3,$FC,$9F,$D6,$3F,$E2,$8A,$F8,$DE,$1F
	.dc.b	$E6,$2B,$F4,$71,$FC,$F5,$AF,$E7,$9A,$FE,$EB,$C0,$FE,$F7,$87,$E6
	.dc.b	$FD,$BF,$EF,$FF,$6E,$E9,$7B,$FF,$B1,$9F,$DB,$FC,$53,$F0,$7A,$7E
	.dc.b	$D7,$E2,$6E,$3F,$A5,$F8,$FF,$D5,$E6,$69,$23,$FC,$A1,$7F,$D0,$B7
	.dc.b	$F6,$36,$3F,$6B,$5F,$09,$FD,$BC,$67,$5F,$E2,$6F,$CF,$4D,$9F,$4C
	.dc.b	$FF,$0B,$8D,$6C,$FB,$FF,$90,$49,$FF,$31,$46,$5F,$EC,$5C,$9F,$0B
	.dc.b	$E2,$7F,$C4,$1B,$9F,$EF,$3F,$47,$F0,$EB,$F9,$3C,$E5,$FE,$53,$F4
	.dc.b	$FF,$EB,$F6,$FC,$4F,$F8,$03,$E9,$9F,$77,$FE,$2B,$AD,$6F,$FC,$2F
	.dc.b	$CE,$FC,$BB,$FD,$DF,$4F,$E1,$9D,$DF,$E2,$A9,$CC,$FF,$1A,$9C,$E5
	.dc.b	$E8,$FC,$3D,$EF,$FF,$8D,$DB,$1F,$DC,$BE,$15,$F5,$4F,$BA,$D5,$FF
	.dc.b	$8F,$7F,$0F,$7B,$54,$FF,$0B,$FD,$73,$E7,$F1,$97,$F4,$6F,$C7,$7F
	.dc.b	$C0,$DB,$78,$DA,$FE,$9D,$AF,$F4,$36,$EC,$FC,$41,$F1,$BF,$E4,$D6
	.dc.b	$DF,$9F,$F0,$97,$D5,$66,$BF,$CA,$5F,$88,$3E,$D6,$9F,$E2,$CF,$B7
	.dc.b	$FA,$5A,$7F,$35,$9D,$63,$FD,$73,$D5,$FE,$BD,$3F,$35,$AF,$CB,$6B
	.dc.b	$FB,$F6,$BF,$D6,$FE,$BB,$F2,$5A,$9F,$CA,$73,$3F,$07,$FF,$1C,$61
	.dc.b	$FD,$AA,$77,$E5,$BE,$73,$FC,$D4,$D9,$FA,$9F,$79,$72,$FE,$99,$37
	.dc.b	$C9,$FF,$53,$E8,$FB,$DF,$F0,$1E,$1F,$E3,$78,$C7,$BB,$FE,$70,$6C
	.dc.b	$66,$77,$87,$F9,$06,$5F,$BB,$BE,$9B,$FC,$FF,$18,$6F,$C8,$7F,$10
	.dc.b	$C3,$FC,$03,$39,$7F,$C7,$6B,$F0,$AF,$9F,$9F,$99,$B3,$C1,$FE,$6E
	.dc.b	$F0,$73,$8A,$9F,$9D,$7F,$07,$AF,$63,$67,$D0,$D7,$FB,$CC,$F0,$F4
	.dc.b	$7F,$D1,$FC,$6F,$97,$7C,$F7,$F9,$B9,$B8,$FA,$3E,$BB,$F8,$2E,$3F
	.dc.b	$E1,$6B,$E1,$7F,$0B,$F0,$6F,$CF,$5F,$B9,$B5,$FE,$1C,$C9,$F8,$5B
	.dc.b	$E0,$78,$78,$7F,$2A,$31,$87,$D5,$6D,$FD,$16,$BE,$A6,$BF,$3B,$F1
	.dc.b	$BF,$E0,$B3,$F3,$9B,$1F,$E2,$CF,$73,$F1,$DC,$9F,$62,$73,$C7,$FE
	.dc.b	$EC,$9C,$DF,$F9,$C1,$B7,$E9,$FC,$33,$C7,$5B,$ED,$FC,$15,$FD,$CF
	.dc.b	$0B,$FD,$1F,$7E,$F6,$D7,$FB,$E3,$85,$F0,$9F,$E5,$2B,$E1,$F6,$9F
	.dc.b	$E9,$C8,$C6,$1F,$35,$B7,$F9,$AF,$C3,$3F,$97,$4E,$7E,$37,$FC,$1D
	.dc.b	$EB,$CC,$7F,$CD,$BF,$86,$7D,$8A,$FF,$25,$F8,$3F,$E6,$2A,$FF,$3B
	.dc.b	$CE,$71,$FF,$0A,$36,$FF,$CA,$F4,$87,$D3,$70,$E9,$E2,$FF,$9D,$63
	.dc.b	$35,$FE,$79,$6E,$3F,$EA,$7B,$E5,$5B,$F1,$3C,$27,$EA,$3F,$0C,$7D
	.dc.b	$86,$31,$FD,$2C,$61,$B1,$FE,$2B,$F5,$F8,$3F,$E6,$2C,$7F,$CD,$BF
	.dc.b	$53,$FE,$B1,$7F,$CA,$AD,$72,$57,$5F,$6B,$FE,$DB,$E8,$BF,$2C,$69
	.dc.b	$2B,$FD,$79,$3F,$0B,$83,$8E,$BF,$EE,$A9,$CB,$5F,$F6,$34,$ED,$F8
	.dc.b	$5E,$3F,$FA,$D1,$FF,$4F,$35,$CF,$D2,$7A,$78,$3F,$86,$BE,$BF,$05
	.dc.b	$C1,$FD,$A6,$BF,$8C,$35,$FE,$28,$D7,$F7,$AD,$7E,$DF,$8F,$DD,$FF
	.dc.b	$3C,$DF,$F1,$1B,$F0,$4E,$67,$71,$8F,$76,$47,$FA,$D6,$51,$88,$FB
	.dc.b	$D0,$F9,$4E,$64,$8E,$0F,$AF,$F0,$C7,$DC,$52,$4F,$F3,$38,$CB,$F6
	.dc.b	$F6,$FE,$AC,$DF,$53,$EE,$46,$23,$FD,$E0,$D7,$F8,$EF,$D9,$FF,$47
	.dc.b	$7F,$9F,$35,$FE,$74,$D7,$7D,$1F,$89,$F0,$1F,$33,$FE,$E1,$9F,$F7
	.dc.b	$E4,$63,$E3,$F5,$39,$9F,$F5,$18,$9D,$3C,$3E,$73,$4F,$FA,$A2,$3F
	.dc.b	$D3,$F5,$DE,$73,$3F,$EC,$69,$CB,$FD,$57,$F0,$CF,$E9,$CD,$F2,$7F
	.dc.b	$09,$18,$8F,$EA,$8D,$7F,$9E,$FE,$18,$FA,$AF,$C5,$36,$3F,$B0,$35
	.dc.b	$FE,$B8,$D7,$CD,$1F,$EE,$67,$E0,$B5,$F7,$DF,$5F,$13,$E4,$4E,$67
	.dc.b	$FA,$6E,$15,$FE,$C2,$D7,$FA,$C3,$7A,$3F,$B5,$57,$E7,$FF,$0B,$7D
	.dc.b	$9D,$25,$FF,$77,$8C,$BF,$53,$F1,$17,$CB,$FD,$96,$C6,$63,$07,$DD
	.dc.b	$FC,$3B,$FC,$ED,$1F,$71,$AF,$B1,$AF,$D7,$6B,$F5,$7C,$EF,$EE,$0F
	.dc.b	$E1,$1A,$FC,$46,$BF,$7D,$AF,$1E,$9F,$C7,$47,$E9,$E3,$94,$60,$FC
	.dc.b	$97,$4F,$F8,$F2,$3F,$22,$D7,$E8,$B5,$FC,$63,$5F,$C5,$7E,$12,$FA
	.dc.b	$7F,$CE,$36,$3F,$36,$D7,$ED,$B5,$FC,$B7,$E2,$3F,$CD,$1F,$D0,$B5
	.dc.b	$FC,$0B,$5D,$F9,$1F,$A9,$F8,$8F,$E0,$B1,$FA,$86,$BF,$88,$6B,$FA
	.dc.b	$36,$BF,$A1,$F8,$D7,$9F,$FB,$36,$C7,$EC,$9A,$FE,$49,$AF,$EB,$7D
	.dc.b	$21,$FB,$B2,$BF,$BA,$6B,$F9,$A6,$BF,$B5,$F9,$17,$9D,$FC,$0D,$B1
	.dc.b	$FC,$09,$AF,$E8,$9A,$FE,$F7,$E4,$DF,$6D,$FC,$2D,$B1,$FC,$29,$AF
	.dc.b	$EA,$9A,$FF,$05,$F9,$AF,$F7,$93,$FC,$41,$AF,$EC,$1A,$EF,$E4,$FF
	.dc.b	$8C,$7E,$2A,$FB,$6A,$49,$FE,$33,$19,$7F,$8D,$FE,$2D,$5F,$DD,$B5
	.dc.b	$FE,$2C,$D7,$F8,$A3,$5F,$E3,$DE,$FB,$E7,$FF,$94,$36,$3F,$93,$B5
	.dc.b	$FE,$0E,$D7,$F9,$17,$E3,$1F,$EB,$A7,$F9,$6B,$5F,$E1,$AD,$77,$F3
	.dc.b	$3F,$DF,$7D,$77,$B5,$F8,$89,$FE,$F0,$D8,$CF,$A5,$B8,$D4,$BF,$00
	.dc.b	$7B,$5F,$E0,$56,$C7,$F8,$25,$AF,$F3,$C3,$84,$7F,$82,$FF,$06,$7D
	.dc.b	$25,$24,$FF,$84,$63,$2F,$FB,$8B,$F0,$6F,$E6,$88,$7C,$92,$33,$18
	.dc.b	$6C,$67,$CB,$FF,$0E,$2F,$DD,$BF,$C3,$8D,$AC,$C6,$0C,$FD,$D7,$F9
	.dc.b	$EA,$3F,$DE,$6D,$7F,$DC,$8D,$7F,$DC,$4D,$7C,$D2,$8F,$F2,$0C,$72
	.dc.b	$F5,$FC,$71,$88,$E7,$1F,$E1,$9B,$C4,$7F,$83,$9A,$FF,$93,$1A,$BD
	.dc.b	$CF,$F5,$54,$FF,$8E,$7E,$B5,$F9,$6F,$FB,$CA,$BE,$94,$63,$E7,$F5
	.dc.b	$39,$9F,$EE,$33,$97,$EC,$5F,$FA,$D7,$E1,$BC,$84,$8C,$E3,$EF,$7E
	.dc.b	$9F,$A2,$FA,$BF,$9E,$D8,$FC,$16,$BF,$D5,$27,$26,$BE,$5A,$4E,$A9
	.dc.b	$18,$8F,$F0,$2D,$7F,$BC,$B9,$7E,$DF,$CD,$FE,$6A,$7A,$B5,$DB,$5F
	.dc.b	$C4,$1F,$88,$5F,$86,$72,$24,$72,$7F,$CC,$1F,$83,$3E,$BA,$92,$7F
	.dc.b	$CC,$D1,$97,$F4,$1F,$2F,$FA,$32,$38,$3F,$49,$C9,$FA,$72,$38,$3E
	.dc.b	$2C,$7E,$1D,$B1,$F9,$88,$C4,$7C,$26,$BE,$CC,$5F,$99,$BF,$07,$FF
	.dc.b	$3D,$46,$1B,$19,$9F,$DA,$4E,$5F,$BF,$FB,$13,$E1,$BF,$CD,$4D,$8C
	.dc.b	$C6,$0C,$FB,$B2,$38,$B5,$FF,$1D,$F8,$2F,$C5,$7F,$AF,$47,$EE,$63
	.dc.b	$1E,$A7,$F1,$39,$CC,$FF,$A2,$5A,$FF,$0F,$E8,$FE,$1C,$47,$F8,$73
	.dc.b	$FC,$6F,$D5,$7D,$8D,$24,$FF,$11,$C1,$3A,$8F,$C9,$FD,$DF,$F8,$79
	.dc.b	$FC,$6B,$5F,$C3,$F3,$23,$11,$FB,$C6,$BF,$CA,$BF,$33,$E8,$FF,$96
	.dc.b	$B7,$46,$23,$EB,$35,$FE,$35,$1A,$C7,$F8,$AF,$C1,$FF,$54,$9F,$E1
	.dc.b	$2D,$7F,$8F,$B7,$0F,$F2,$A7,$A1,$C3,$E6,$E5,$7E,$BA,$91,$88,$FD
	.dc.b	$4F,$E0,$6F,$B7,$FE,$46,$D8,$FD,$AB,$5D,$C6,$B9,$A4,$69,$AF,$ED
	.dc.b	$3F,$04,$EE,$FE,$F9,$AE,$6F,$F9,$8C,$62,$3E,$6A,$FC,$13,$24,$8E
	.dc.b	$4F,$F3,$8F,$55,$F8,$3D,$24,$FF,$0C,$8C,$BF,$C3,$BF,$06,$FF,$27
	.dc.b	$BF,$D1,$9B,$B1,$A6,$BE,$2B,$FD,$8B,$CD,$FE,$C7,$F9,$6C,$11,$C7
	.dc.b	$C9,$4C,$3F,$D9,$2D,$7F,$8A,$DF,$E6,$B0,$ED,$D5,$FD,$42,$30,$4F
	.dc.b	$23,$04,$AF,$43,$FD,$5B,$E2,$FF,$AA,$92,$30,$67,$38,$FF,$2C,$8F
	.dc.b	$F3,$2B,$FD,$91,$AB,$F0,$7F,$F2,$1E,$7F,$EF,$EF,$B4,$FE,$47,$18
	.dc.b	$97,$CE,$27,$3F,$0D,$FE,$7E,$8C,$4F,$F7,$1A,$FF,$44,$BC,$2B,$77
	.dc.b	$D6,$35,$7C,$67,$31,$B3,$97,$99,$FC,$D3,$0E,$B1,$E7,$AE,$0F,$2B
	.dc.b	$E8,$E5,$B8,$78,$AB,$8B,$EB,$6D,$72,$AE,$07,$84,$F8,$97,$E1,$C3
	.dc.b	$CB,$5C,$30,$EB,$7C,$70,$ED,$5A,$26,$1E,$53,$A4,$69,$C5,$F3,$73
	.dc.b	$5C,$F0,$F0,$46,$AF,$9B,$E6,$AF,$5C,$EA,$31,$3A,$BF,$73,$F6,$A7
	.dc.b	$51,$89,$E8,$43,$DE,$7B,$30,$8F,$91,$F9,$73,$CB,$0F,$80,$48,$CC
	.dc.b	$F0,$3F,$DE,$F5,$F8,$96,$CF,$EC,$23,$11,$F8,$EF,$82,$FC,$AC,$7F
	.dc.b	$9F,$33,$FF,$22,$4F,$E5,$67,$24,$8F,$BA,$FF,$B1,$B2,$FB,$BF,$09
	.dc.b	$FD,$FD,$FB,$F1,$9C,$FB,$57,$03,$F0,$B5,$B8,$CD,$FD,$6A,$E0,$FE
	.dc.b	$6D,$AF,$E9,$E7,$83,$9A,$E8,$F1,$7A,$F7,$63,$0F,$1C,$5C,$3F,$5B
	.dc.b	$7C,$AF,$DD,$39,$7A,$1F,$88,$8D,$63,$F1,$63,$B1,$2F,$E0,$E1,$C6
	.dc.b	$39,$99,$C3,$BC,$66,$B9,$57,$92,$BB,$4E,$71,$F7,$D6,$AF,$8D,$73
	.dc.b	$3C,$B7,$C4,$F0,$B9,$3C,$2E,$63,$8B,$63,$F8,$97,$C5,$FF,$8C,$5B
	.dc.b	$87,$F1,$78,$C7,$13,$74,$61,$7B,$94,$95,$FC,$7D,$D7,$CC,$C8,$FE
	.dc.b	$44,$DC,$3F,$C8,$F5,$FE,$8E,$9F,$E4,$DE,$DF,$52,$3F,$CA,$7D,$1F
	.dc.b	$01,$1C,$1F,$CB,$7A,$7C,$84,$70,$7F,$4F,$F7,$3D,$7C,$FF,$EA,$48
	.dc.b	$C1,$31,$FE,$8B,$F0,$9A,$24,$7F,$57,$7E,$1D,$BD,$9A,$49,$FE,$BC
	.dc.b	$EB,$E5,$E4,$7F,$61,$6D,$FF,$73,$3E,$6F,$81,$FE,$0E,$71,$3F,$DD
	.dc.b	$5C,$F3,$C8,$FE,$F0,$D8,$FF,$37,$FD,$33,$A7,$EB,$B6,$BF,$CF,$6D
	.dc.b	$9D,$CE,$BA,$E5,$69,$B3,$FE,$25,$FC,$75,$F1,$FF,$85,$A4,$61,$FF
	.dc.b	$2A,$B6,$78,$4F,$E2,$3F,$08,$FE,$AD,$5F,$E8,$C6,$CF,$FA,$19,$AF
	.dc.b	$F8,$E9,$AF,$F9,$47,$EC,$1E,$EF,$E5,$7D,$14,$62,$FF,$C9,$0D,$AF
	.dc.b	$F2,$DB,$5F,$F4,$59,$F9,$6C,$3F,$3C,$D9,$FE,$3B,$D3,$E9,$8D,$B5
	.dc.b	$FF,$31,$7C,$B3,$F1,$1F,$13,$6B,$FC,$B0,$D9,$FF,$35,$B5,$FF,$4D
	.dc.b	$F8,$6F,$D8,$3F,$C7,$95,$FE,$B8,$6C,$F9,$A7,$5E,$6F,$EC,$DB,$5B
	.dc.b	$E8,$C5,$9A,$48,$FF,$43,$FC,$53,$E5,$3E,$C3,$63,$31,$83,$F7,$BE
	.dc.b	$E3,$C9,$A4,$9F,$F4,$FB,$F2,$36,$FD,$FC,$7C,$87,$04,$7E,$5F,$87
	.dc.b	$B2,$38,$3F,$A5,$FE,$09,$FB,$1D,$BE,$A8,$70,$FD,$86,$BF,$EC,$46
	.dc.b	$AF,$15,$F8,$27,$DF,$8F,$F7,$DB,$5D,$36,$37,$CD,$FC,$72,$1F,$0F
	.dc.b	$49,$87,$FB,$8A,$32,$FF,$B9,$3E,$7D,$F8,$07,$F2,$46,$C6,$63,$06
	.dc.b	$7E,$EA,$47,$16,$E5,$F7,$F8,$DC,$9F,$70,$95,$F8,$57,$F9,$A7,$C4
	.dc.b	$F8,$3E,$46,$E2,$69,$2B,$E5,$D3,$F6,$7F,$F3,$9C,$61,$B8,$FE,$3D
	.dc.b	$AF,$E5,$9A,$8F,$9D,$F5,$4F,$C2,$7F,$A1,$B6,$BE,$CF,$0A,$31,$5F
	.dc.b	$B0,$8C,$57,$0F,$C3,$65,$F0,$B4,$62,$F8,$79,$CF,$9A,$FE,$A2,$DB
	.dc.b	$EE,$D9,$DC,$EB,$E5,$65,$69,$B3,$F8,$EF,$C0,$1F,$19,$FC,$09,$B5
	.dc.b	$FE,$1E,$6C,$EE,$7F,$83,$FE,$07,$FA,$5A,$FF,$11,$B6,$7F,$32,$D7
	.dc.b	$F8,$7B,$5F,$E1,$3E,$AF,$F0,$82,$4F,$F0,$87,$FB,$87,$86,$F9,$87
	.dc.b	$E3,$98,$6D,$7E,$9A,$75,$39,$38,$F9,$6F,$3E,$92,$B8,$46,$5E,$5F
	.dc.b	$34,$F8,$7F,$0B,$63,$31,$83,$CB,$F3,$7F,$73,$4F,$C2,$0C,$91,$98
	.dc.b	$FC,$44,$63,$97,$89,$EE,$FC,$7D,$F2,$B4,$98,$FB,$63,$2F,$BF,$E7
	.dc.b	$5F,$87,$FC,$66,$C6,$63,$07,$CA,$F9,$FF,$C2,$65,$E6,$C9,$18,$8F
	.dc.b	$2B,$5F,$53,$63,$F2,$1F,$27,$F5,$3A,$0D,$26,$1F,$4F,$E8,$3F,$80
	.dc.b	$E1,$FD,$3A,$D1,$88,$F1,$35,$F8,$AD,$7E,$57,$D0,$BF,$25,$FC,$1B
	.dc.b	$6F,$F0,$0D,$9E,$CD,$7D,$DF,$8F,$7F,$8A,$5F,$13,$5C,$F6,$FE,$3A
	.dc.b	$31,$1F,$5B,$D7,$AF,$DB,$6E,$5F,$DB,$AB,$F2,$B1,$8F,$5D,$F1,$5F
	.dc.b	$98,$6E,$5F,$95,$6E,$1F,$0D,$AF,$D3,$FA,$57,$6F,$E8,$1B,$5F,$9D
	.dc.b	$6C,$FC,$B6,$BC,$7E,$17,$81,$27,$83,$FA,$AF,$A6,$7C,$6D,$24,$FE
	.dc.b	$A6,$32,$DF,$AC,$E0,$47,$07,$F5,$4E,$F7,$D7,$FA,$B1,$FE,$49,$18
	.dc.b	$D7,$FC,$D0,$DA,$E1,$5B,$AF,$DA,$7D,$7B,$E5,$76,$DA,$EC,$D9,$FF
	.dc.b	$0B,$DA,$31,$1F,$9B,$F8,$5F,$E2,$AD,$F1,$FF,$84,$C6,$23,$F6,$2D
	.dc.b	$7F,$82,$7D,$CF,$C0,$7A,$B5,$FD,$DD,$A3,$11,$EE,$6B,$FC,$37,$EE
	.dc.b	$3F,$83,$3E,$66,$BC,$9A,$FC,$66,$BF,$C1,$FE,$C1,$F8,$37,$F2,$06
	.dc.b	$E7,$FC,$75,$B3,$FA,$B6,$BF,$BB,$F6,$1F,$4B,$FC,$A1,$B5,$FC,$95
	.dc.b	$B3,$FB,$36,$BE,$EF,$B8,$F1,$E6,$69,$3C,$9F,$F5,$17,$DA,$BF,$42
	.dc.b	$63,$86,$BF,$E1,$88,$D3,$93,$F0,$7F,$71,$FD,$F1,$FC,$53,$71,$FB
	.dc.b	$AD,$7E,$8B,$5F,$EE,$1E,$C7,$F7,$9F,$8E,$FF,$21,$7B,$3F,$E7,$99
	.dc.b	$FE,$ED,$18,$33,$F1,$D3,$0F,$EF,$CD,$9F,$EF,$DF,$33,$C8,$93,$FE
	.dc.b	$03,$7F,$33,$F3,$6F,$B9,$A4,$9F,$D1,$46,$5F,$F0,$6F,$5B,$E3,$3F
	.dc.b	$CA,$CD,$8C,$C6,$0F,$F3,$7E,$1F,$E0,$BF,$99,$FF,$1F,$BF,$D2,$9A
	.dc.b	$E4,$AE,$BE,$E1,$F3,$2F,$93,$C3,$6B,$FC,$93,$3A,$9C,$9C,$7E,$25
	.dc.b	$C7,$49,$5C,$23,$2F,$EC,$34,$FC,$AB,$FE,$21,$9F,$C0,$BF,$DB,$63
	.dc.b	$2F,$4C,$FF,$81,$C6,$27,$7C,$5F,$C2,$3C,$8F,$EC,$F1,$8F,$07,$FD
	.dc.b	$15,$87,$5C,$3E,$FC,$62,$3A,$1F,$E4,$47,$13,$E2,$E9,$63,$DE,$F8
	.dc.b	$FC,$6F,$CC,$CB,$F8,$43,$70,$FF,$7A,$E3,$FE,$44,$6B,$FE,$8F,$7F
	.dc.b	$DF,$6F,$8C,$F0,$7C,$6F,$A2,$FC,$AD,$8F,$25,$5B,$FE,$4B,$37,$8B
	.dc.b	$9D,$7A,$A7,$2F,$57,$F8,$E4,$F3,$8F,$F3,$8F,$CB,$5D,$8D,$3C,$1A
	.dc.b	$3E,$E3,$FD,$A7,$5E,$78,$7E,$24,$FA,$2D,$B6,$74,$D7,$E6,$9E,$03
	.dc.b	$E0,$7A,$4F,$05,$E5,$FB,$E3,$0D,$8C,$CF,$87,$EA,$BF,$A7,$BE,$3F
	.dc.b	$AF,$7D,$27,$A1,$B1,$98,$E5,$7E,$CF,$B9,$FF,$58,$4F,$F4,$F6,$BF
	.dc.b	$D0,$9A,$FF,$36,$6B,$FD,$4B,$F1,$5F,$DC,$3F,$0D,$AF,$E2,$B5,$46
	.dc.b	$3A,$3F,$15,$5D,$9F,$CE,$E8,$FA,$8F,$F0,$FE,$1F,$11,$AF,$E5,$23
	.dc.b	$5E,$EF,$D4,$FB,$AE,$13,$F4,$FE,$FB,$FD,$36,$FF,$69,$6B,$FD,$71
	.dc.b	$AF,$F5,$06,$BF,$DB,$3E,$6B,$FC,$C2,$FD,$D6,$BE,$B6,$E3,$FA,$37
	.dc.b	$CA,$FE,$97,$EE,$FC,$73,$F0,$5B,$19,$CB,$F1,$70,$F1,$FF,$1B,$F8
	.dc.b	$07,$E6,$2F,$FB,$1B,$5F,$EF,$ED,$7F,$CD,$CD,$DD,$5F,$98,$9F,$CA
	.dc.b	$35,$F1,$74,$7E,$2F,$D8,$F8,$79,$7F,$0C,$F9,$47,$D8,$FF,$8B,$2F
	.dc.b	$F6,$51,$8F,$6B,$F6,$0D,$8F,$CB,$FB,$BF,$E7,$F7,$7E,$1F,$F2,$A9
	.dc.b	$FE,$5E,$47,$0F,$58,$FE,$1A,$FF,$BA,$6B,$AB,$9C,$3F,$07,$EF,$7E
	.dc.b	$1E,$FF,$D8,$52,$FC,$73,$FD,$69,$87,$99,$AF,$46,$DF,$E3,$3E,$F1
	.dc.b	$F4,$3E,$AF,$C7,$3F,$D6,$37,$FE,$C8,$6B,$FE,$C7,$6B,$FE,$8E,$C1
	.dc.b	$18,$8F,$1F,$8A,$FA,$FF,$8A,$D8,$F9,$2D,$7F,$31,$F0,$FF,$C8,$32
	.dc.b	$FA,$1F,$8E,$7C,$99,$F8,$6D,$8C,$F2,$FF,$01,$F8,$A7,$AF,$F4,$BE
	.dc.b	$C3,$FC,$B3,$87,$FB,$81,$AF,$FB,$7D,$AF,$A5,$AF,$CA,$FC,$73,$FD
	.dc.b	$F0,$FD,$E6,$BF,$51,$B8,$7E,$D3,$E3,$95,$F8,$2F,$C6,$ED,$FE,$19
	.dc.b	$AF,$DC,$6D,$7E,$F3,$E4,$5E,$E7,$E2,$BF,$30,$7D,$27,$F7,$2E,$44
	.dc.b	$63,$1F,$F7,$DB,$5F,$F7,$D3,$5F,$A9,$F9,$87,$EA,$5F,$CB,$B5,$FC
	.dc.b	$6B,$71,$FE,$1B,$F3,$0F,$B0,$FC,$DF,$B6,$F4,$69,$30,$FE,$67,$F3
	.dc.b	$5B,$7F,$47,$F9,$86,$DF,$F6,$89,$F9,$47,$0E,$34,$46,$23,$F1,$4D
	.dc.b	$7F,$37,$F9,$87,$C4,$7F,$5E,$D7,$F4,$AD,$CF,$F9,$2F,$CF,$BE,$8B
	.dc.b	$FA,$C7,$E0,$9F,$E1,$4B,$FE,$D0,$1C,$3F,$D3,$9B,$5F,$E0,$AF,$82
	.dc.b	$FF,$25,$FC,$07,$F6,$1F,$C1,$1F,$33,$FE,$22,$6E,$1F,$E2,$66,$BF
	.dc.b	$00,$38,$7F,$DE,$7F,$02,$F4,$FB,$F9,$3F,$E4,$A6,$CA,$75,$F0,$DE
	.dc.b	$1C,$3F,$86,$FE,$07,$F3,$34,$FE,$28,$DC,$7F,$80,$B7,$07,$13,$F0
	.dc.b	$4F,$E9,$55,$FC,$1B,$F1,$5F,$F0,$BB,$7E,$D3,$F8,$CC,$62,$F8,$E1
	.dc.b	$CF,$07,$CC,$BE,$1F,$F8,$FB,$6B,$F9,$1B,$5F,$E3,$2B,$EC,$3F,$92
	.dc.b	$FE,$2D,$FD,$CA,$7F,$8E,$B7,$5F,$E5,$CD,$D3,$F8,$DD,$F3,$9F,$F1
	.dc.b	$EF,$D7,$BF,$06,$7D,$CF,$F8,$7E,$30,$D8,$CE,$3F,$E1,$BE,$4E,$58
	.dc.b	$FC,$FC,$FF,$C4,$91,$E5,$9D,$47,$1B,$FF,$14,$7B,$8F,$C6,$69,$2B
	.dc.b	$67,$F4,$2F,$7D,$F8,$3F,$F4,$26,$C6,$63,$1F,$15,$FD,$1B,$F1,$EF
	.dc.b	$F6,$74,$FF,$4A,$6B,$FD,$01,$BB,$FF,$51,$BE,$73,$FD,$13,$E9,$1F
	.dc.b	$57,$FD,$41,$B1,$FD,$51,$AE,$F7,$FE,$AB,$F2,$AF,$AC,$A4,$9E,$17
	.dc.b	$FA,$8F,$B8,$91,$FA,$8F,$0D,$1F,$E4,$3F,$C2,$3F,$D7,$B3,$9E,$7F
	.dc.b	$F4,$2D,$7F,$41,$81,$FA,$F8,$C9,$FE,$8A,$F9,$BF,$E4,$39,$EA,$7F
	.dc.b	$D3,$9E,$AF,$F9,$3E,$73,$1C,$AF,$FA,$8E,$FF,$E7,$EB,$7E,$11,$F1
	.dc.b	$9E,$E6,$3E,$A3,$FC,$C4,$DB,$FE,$EF,$A7,$28,$FE,$CF,$F8,$DB,$E8
	.dc.b	$BF,$BC,$B6,$BF,$BE,$B5,$FF,$3F,$CF,$37,$FC,$D1,$F8,$4F,$F8,$72
	.dc.b	$3F,$D8,$6D,$8F,$DD,$37,$C8,$3F,$C8,$B3,$FF,$48,$BF,$E2,$6A,$E3
	.dc.b	$F6,$7F,$E7,$78,$C5,$F2,$F9,$3F,$E3,$78,$48,$FA,$8F,$F9,$FF,$CD
	.dc.b	$FF,$08,$47,$FA,$51,$7E,$12,$FC,$1F,$FD,$77,$18,$6C,$66,$7F,$8C
	.dc.b	$4E,$5F,$E3,$55,$E8,$F4,$3C,$84,$8F,$F4,$94,$ED,$F8,$BF,$84,$73
	.dc.b	$6E,$1F,$4A,$31,$AF,$F8,$AF,$F0,$97,$E1,$5F,$E2,$8F,$C1,$3F,$C2
	.dc.b	$38,$7F,$8B,$1A,$FF,$88,$9B,$5F,$81,$BE,$7C,$1F,$E2,$0F,$97,$FF
	.dc.b	$44,$4F,$F8,$DD,$B1,$9F,$A1,$FF,$1A,$C7,$DB,$9F,$F1,$E7,$D8,$FF
	.dc.b	$84,$5B,$F1,$DF,$E3,$A8,$C5,$F1,$CB,$FC,$7F,$F8,$57,$ED,$9F,$F1
	.dc.b	$03,$5F,$F1,$8B,$5F,$E6,$2D,$F4,$5F,$83,$7F,$67,$9F,$F1,$EB,$71
	.dc.b	$FF,$27,$36,$F9,$F5,$FF,$93,$23,$FD,$2B,$5C,$3E,$8B,$06,$75,$FF
	.dc.b	$28,$FC,$13,$F0,$1A,$4A,$FF,$2B,$DF,$2C,$BF,$CB,$9F,$83,$FF,$9F
	.dc.b	$27,$FA,$3B,$74,$34,$99,$FF,$97,$9F,$D8,$4F,$08,$C7,$47,$FB,$0B
	.dc.b	$EA,$BF,$9F,$4F,$EF,$7E,$47,$FD,$C7,$F6,$DF,$D9,$63,$FD,$EC,$D8
	.dc.b	$CC,$7F,$69,$7F,$DD,$3F,$8C,$73,$24,$7F,$B2,$EB,$FA,$33,$E8,$F5
	.dc.b	$BC,$84,$7F,$CE,$B9,$7F,$99,$F8,$BF,$CC,$BF,$84,$FF,$B5,$27,$FC
	.dc.b	$DA,$DD,$BF,$CF,$CD,$9F,$F3,$5D,$7F,$68,$AF,$ED,$73,$FE,$53,$7F
	.dc.b	$CC,$0D,$7A,$E5,$FE,$37,$8C,$47,$0F,$7B,$FD,$0B,$1F,$DB,$77,$34
	.dc.b	$FF,$2E,$61,$FC,$8F,$F0,$BF,$F2,$DE,$FF,$CA,$BE,$BD,$F8,$5E,$7E
	.dc.b	$5B,$FD,$73,$18,$CF,$F9,$7D,$F0,$51,$95,$1F,$E4,$18,$C3,$B6,$CF
	.dc.b	$4A,$ED,$7E,$17,$C9,$EA,$FF,$37,$AD,$CE,$70,$F0,$DF,$16,$C6,$5F
	.dc.b	$0D,$71,$BE,$B1,$E6,$6D,$F4,$8C,$1E,$C9,$E0,$F4,$32,$7B,$2F,$C0
	.dc.b	$DB,$F0,$61,$EF,$C3,$A9,$27,$96,$1F,$0B,$0E,$B1,$C4,$EB,$5F,$16
	.dc.b	$B8,$47,$A8,$EB,$87,$82,$75,$1D,$27,$E3,$BA,$C2,$3D,$2B,$85,$F8
	.dc.b	$2F,$8C,$F8,$70,$F8,$47,$95,$E2,$E3,$0F,$1C,$FB,$67,$C0,$43,$94
	.dc.b	$78,$9F,$06,$1F,$B1,$9F,$C3,$47,$F9,$D9,$AF,$D4,$31,$F2,$5F,$9A
	.dc.b	$8F,$F4,$AE,$DC,$63,$04,$7E,$7B,$FE,$DF,$C3,$F1,$BE,$83,$4C,$D4
	.dc.b	$C3,$E6,$DE,$F2,$FD,$05,$EE,$73,$5B,$C3,$F2,$F7,$B9,$FC,$FD,$6E
	.dc.b	$73,$5B,$BF,$D5,$5E,$EF,$F4,$37,$B9,$CD,$6E,$7F,$53,$5B,$C7,$FD
	.dc.b	$77,$8E,$67,$F5,$33,$93,$58,$73,$8E,$37,$BB,$F8,$F1,$F8,$A9,$D3
	.dc.b	$CA,$75,$5C,$AE,$C7,$1A,$DD,$F2,$C7,$DB,$87,$C4,$3A,$5F,$07,$11
	.dc.b	$F4,$23,$0F,$03,$11,$AC,$7F,$91,$FD,$EF,$CF,$E2,$FE,$42,$DB,$FE
	.dc.b	$4F,$F3,$76,$FF,$90,$CD,$E1,$FE,$33,$3F,$5E,$7F,$91,$B5,$E3,$5F
	.dc.b	$EA,$39,$DB,$78,$5F,$65,$CD,$B1,$FE,$48,$C7,$F9,$9F,$C8,$E6,$B4
	.dc.b	$C9,$D1,$FC,$F3,$E2,$71,$8F,$F3,$CF,$C5,$BA,$1D,$7D,$CB,$5C,$27
	.dc.b	$F9,$DC,$79,$DB,$1E,$58,$C6,$7F,$CD,$9B,$D0,$F9,$FF,$84,$DA,$FE
	.dc.b	$9F,$F1,$98,$9C,$BC,$09,$1F,$0E,$51,$DB,$C2,$FF,$77,$4F,$97,$0F
	.dc.b	$F7,$97,$C9,$FC,$F9,$FF,$4F,$CE,$5D,$E4,$DF,$F5,$17,$FE,$F9,$F8
	.dc.b	$5E,$07,$1F,$A7,$B8,$7F,$80,$B6,$FB,$5E,$21,$B8,$E1,$B3,$F6,$67
	.dc.b	$FC,$DB,$3F,$EF,$96,$BF,$E7,$9C,$3F,$C3,$AF,$C8,$F2,$1F,$5D,$F8
	.dc.b	$76,$DF,$B7,$CA,$78,$39,$AF,$C6,$46,$27,$53,$8F,$07,$2F,$E4,$7C
	.dc.b	$C7,$C8,$D3,$F2,$D4,$FC,$09,$F4,$B9,$8F,$C0,$19,$77,$1A,$8F,$F2
	.dc.b	$87,$7E,$27,$F4,$1E,$2B,$F0,$87,$C0,$E1,$B7,$FE,$62,$FB,$FF,$F6
	.dc.b	$F3,$F8,$DA,$E1,$5F,$A9,$F6,$15,$AE,$2F,$F5,$00,$8C,$FC,$6F,$E4
	.dc.b	$DB,$87,$F9,$D2,$B8,$57,$0D,$7D,$0F,$E0,$67,$FD,$55,$18,$E9,$FE
	.dc.b	$55,$6F,$AF,$79,$9F,$BB,$C3,$E4,$ED,$EB,$F4,$66,$5E,$9D,$DF,$86
	.dc.b	$62,$FB,$0E,$0F,$89,$1F,$0C,$F8,$9F,$86,$5B,$FF,$54,$FE,$21,$FC
	.dc.b	$1F,$E0,$CF,$D6,$F2,$7F,$51,$8F,$4E,$D5,$6A,$FE,$BD,$6B,$87,$FD
	.dc.b	$80,$D8,$FF,$66,$D7,$FB,$53,$D2,$E9,$E3,$FB,$7F,$0D,$7E,$19,$4F
	.dc.b	$81,$CD,$F8,$8C,$3F,$81,$68,$AD,$C6,$27,$53,$F5,$8F,$C7,$BF,$C1
	.dc.b	$DB,$1F,$E5,$B8,$DC,$7D,$5F,$3B,$B9,$E3,$F2,$6D,$7E,$F7,$F0,$1F
	.dc.b	$F7,$CB,$7E,$37,$F9,$67,$E1,$BF,$85,$E0,$FD,$DD,$6B,$1F,$8D,$39
	.dc.b	$BF,$82,$D8,$6E,$87,$CC,$F4,$EE,$FE,$71,$F8,$DF,$EE,$B2,$FE,$0E
	.dc.b	$DC,$7E,$5F,$DD,$7F,$A2,$5F,$E2,$51,$9C,$BA,$CE,$5F,$B0,$DA,$FE
	.dc.b	$22,$BE,$9F,$FA,$33,$9B,$57,$E2,$7E,$F5,$F3,$1F,$AE,$6C,$FE,$47
	.dc.b	$B3,$13,$FA,$F6,$47,$F3,$9F,$9C,$FF,$4F,$73,$63,$0F,$CA,$C6,$27
	.dc.b	$5F,$05,$F7,$27,$F3,$36,$FB,$F7,$C3,$7F,$5D,$C3,$85,$7E,$4F,$E7
	.dc.b	$9F,$82,$7F,$5F,$6E,$5F,$CD,$7F,$0E,$7C,$D6,$6B,$FB,$24,$62,$75
	.dc.b	$38,$FC,$3B,$2F,$F6,$9F,$A3,$7C,$65,$3F,$B6,$7C,$03,$F0,$AC,$61
	.dc.b	$FD,$0E,$30,$D8,$DF,$C9,$F0,$7F,$8B,$F0,$7F,$BB,$27,$F8,$93,$76
	.dc.b	$FD,$C8,$FC,$39,$F3,$BC,$3E,$4F,$F8,$24,$62,$FF,$D6,$CD,$AF,$D1
	.dc.b	$35,$FE,$32,$D7,$EE,$8F,$C3,$9E,$4F,$F0,$F6,$D7,$F7,$1F,$53,$F8
	.dc.b	$A7,$C3,$E6,$BF,$C1,$9F,$97,$FE,$F7,$0F,$CD,$B7,$D6,$7D,$53,$E7
	.dc.b	$37,$E7,$FF,$84,$FE,$05,$EA,$7F,$78,$6E,$BF,$E1,$16,$E5,$FC,$C3
	.dc.b	$CB,$FE,$AB,$5F,$DF,$BF,$2F,$7C,$47,$F8,$75,$B8,$7F,$81,$F4,$7D
	.dc.b	$73,$E7,$F5,$7F,$E2,$3F,$CB,$DF,$7B,$4F,$F1,$45,$BE,$C7,$F0,$9A
	.dc.b	$FF,$1F,$AD,$E5,$FD,$FF,$F2,$F7,$E1,$BF,$E1,$06,$C7,$F8,$A5,$B5
	.dc.b	$E9,$6B,$FD,$C7,$E1,$5F,$03,$FE,$35,$6D,$7A,$7E,$D9,$F8,$8E,$6B
	.dc.b	$FC,$75,$EE,$BE,$6E,$9E,$8F,$B9,$FC,$7E,$5F,$CD,$67,$58,$FF,$85
	.dc.b	$7E,$6D,$E5,$FF,$89,$9B,$1F,$E3,$F6,$D7,$CF,$6B,$FE,$07,$FC,$75
	.dc.b	$F1,$BF,$E4,$B6,$D7,$CF,$FB,$CF,$E1,$15,$FC,$72,$72,$FF,$47,$FC
	.dc.b	$77,$FC,$8E,$DF,$86,$F9,$FF,$7E,$F9,$DF,$F2,$B5,$6B,$DF,$FF,$12
	.dc.b	$FE,$3A,$FA,$1F,$F1,$EB,$74,$FF,$2B,$37,$2F,$C3,$35,$FF,$0F,$7E
	.dc.b	$3B,$F9,$5F,$0C,$FE,$57,$E4,$9E,$A7,$F9,$96,$35,$F8,$E7,$F8,$E2
	.dc.b	$FF,$AF,$37,$3F,$F4,$F7,$81,$FE,$16,$F0,$7F,$CD,$93,$9D,$3F,$AE
	.dc.b	$FE,$39,$F4,$30,$FC,$C3,$5F,$F1,$83,$5D,$F5,$7F,$96,$21,$F3,$1F
	.dc.b	$E7,$76,$D7,$E6,$FF,$07,$7D,$16,$6B,$FC,$FD,$C4,$F8,$67,$DC,$61
	.dc.b	$B8,$7E,$8F,$F0,$97,$D5,$66,$B8,$FE,$14,$FB,$AF,$F3,$66,$28,$C6
	.dc.b	$1F,$DA,$E3,$13,$A9,$CE,$FF,$E2,$49,$F1,$7C,$EF,$84,$F1,$71,$5C
	.dc.b	$3F,$CF,$52,$FC,$47,$FC,$2F,$18,$6D,$FF,$86,$27,$5E,$37,$BA,$73
	.dc.b	$CD,$FE,$CF,$97,$E1,$9F,$C6,$EF,$F0,$CD,$8C,$FB,$FE,$C8,$D7,$43
	.dc.b	$E8,$BF,$93,$61,$F7,$1B,$9F,$F6,$0C,$1F,$8B,$7C,$88,$FE,$19,$3A
	.dc.b	$CB,$F2,$9F,$8D,$FE,$66,$7D,$AD,$8F,$57,$E2,$FF,$C4,$67,$3E,$1F
	.dc.b	$D8,$8C,$4F,$C9,$6B,$FC,$D3,$14,$62,$3C,$8D,$7F,$49,$F8,$DF,$F0
	.dc.b	$0C,$FE,$23,$70,$FE,$F3,$A3,$E5,$3F,$89,$D6,$BE,$1F,$F8,$FC,$62
	.dc.b	$B6,$DC,$3F,$61,$F8,$C7,$9A,$FF,$43,$39,$CF,$FC,$DB,$F5,$7F,$92
	.dc.b	$E6,$3F,$ED,$1A,$7E,$30,$CB,$53,$98,$C4,$7F,$AF,$67,$6F,$E7,$3F
	.dc.b	$18,$FF,$2C,$39,$9D,$E7,$FE,$04,$EE,$7E,$41,$FE,$13,$3F,$B5,$FD
	.dc.b	$5D,$BF,$ED,$6E,$1F,$ED,$BE,$0D,$BF,$F0,$6F,$C6,$FE,$B9,$FF,$3E
	.dc.b	$FC,$BF,$F8,$3E,$7F,$D9,$CD,$AF,$F2,$63,$67,$7D,$9F,$E6,$17,$FC
	.dc.b	$2E,$DD,$7F,$2D,$F6,$BF,$E9,$09,$CF,$D3,$FE,$E2,$31,$3A,$E1,$FF
	.dc.b	$16,$7A,$59,$7F,$D0,$17,$FE,$32,$FB,$4F,$F2,$E3,$FD,$E1,$AE,$4A
	.dc.b	$EB,$F0,$3F,$F6,$BF,$C7,$DF,$A6,$FF,$4C,$57,$F9,$96,$72,$D8,$CB
	.dc.b	$FE,$9E,$8F,$F2,$9E,$3C,$FE,$8F,$FD,$B1,$3C,$5A,$ED,$AF,$E4,$3F
	.dc.b	$14,$FF,$5F,$BF,$DC,$DA,$FF,$A5,$3B,$51,$8A,$FE,$5B,$8B,$E4,$BF
	.dc.b	$C2,$B1,$F2,$BE,$16,$F1,$FE,$D7,$83,$FB,$8E,$5F,$C3,$1A,$F5,$C7
	.dc.b	$A4,$FE,$8A,$3F,$D8,$67,$FA,$17,$2F,$F2,$EE,$FF,$31,$B1,$93,$67
	.dc.b	$FB,$01,$FF,$31,$FC,$77,$8A,$3F,$D6,$C1,$C4,$6A,$3F,$C2,$47,$2C
	.dc.b	$BF,$A4,$BF,$CD,$8C,$F9,$BC,$1C,$1F,$14,$E8,$7F,$28,$FC,$67,$FD
	.dc.b	$86,$DF,$AF,$F1,$78,$09,$CC,$FF,$29,$7F,$B6,$3C,$79,$77,$06,$98
	.dc.b	$AE,$7F,$6A,$F8,$8A,$4B,$FE,$73,$0F,$91,$DF,$9D,$27,$2E,$DF,$E4
	.dc.b	$27,$E6,$32,$FE,$36,$DF,$27,$F3,$31,$88,$FA,$9F,$52,$FC,$0B,$E5
	.dc.b	$B6,$3F,$C7,$8D,$7A,$9F,$D3,$3F,$16,$7D,$BF,$C4,$7F,$89,$C6,$7C
	.dc.b	$9C,$4E,$7C,$CF,$25,$7F,$9E,$6B,$FA,$F3,$5F,$EB,$8D,$7F,$9A,$7C
	.dc.b	$CB,$6F,$A2,$D8,$FE,$BA,$D7,$F3,$DF,$11,$79,$7F,$AC,$7E,$11,$E6
	.dc.b	$9F,$EE,$9E,$37,$E2,$3D,$BF,$F7,$B4,$7F,$66,$9D,$3F,$D8,$FD,$9F
	.dc.b	$D4,$BF,$D1,$5A,$FF,$57,$6B,$FE,$27,$6F,$C2,$30,$FE,$4F,$3F,$EB
	.dc.b	$C7,$F9,$A3,$FE,$D2,$72,$FF,$BF,$5A,$6A,$39,$7D,$0F,$E7,$BD,$F7
	.dc.b	$D7,$D2,$5F,$5F,$07,$FD,$FF,$E1,$7F,$A3,$23,$07,$FB,$83,$F2,$5F
	.dc.b	$F5,$9B,$FE,$F7,$C5,$18,$8F,$F4,$03,$5F,$F4,$BC,$6B,$1F,$E4,$6F
	.dc.b	$FB,$AB,$83,$F6,$8D,$9E,$ED,$7E,$16,$1F,$ED,$C8,$7E,$ED,$FE,$A2
	.dc.b	$9F,$E2,$9F,$8B,$FE,$DE,$BF,$B4,$B6,$1E,$1A,$31,$6F,$C6,$1F,$49
	.dc.b	$E2,$FC,$63,$F2,$D8,$7A,$1A,$F8,$5A,$A9,$18,$5F,$B5,$7E,$12,$7D
	.dc.b	$26,$DC,$71,$7D,$56,$E7,$ED,$5F,$81,$7F,$14,$F9,$3F,$F2,$6D,$FF
	.dc.b	$9D,$FD,$6C,$1F,$23,$F1,$9F,$B7,$1F,$E8,$7F,$8F,$F9,$BF,$2B,$FE
	.dc.b	$66,$8F,$F4,$34,$60,$FA,$3F,$8C,$FF,$18,$D3,$F0,$CD,$D3,$E7,$36
	.dc.b	$BD,$CF,$F1,$C7,$FA,$17,$CD,$BD,$DF,$E5,$2D,$8F,$F5,$14,$7D,$CF
	.dc.b	$C6,$7F,$D4,$11,$F7,$3F,$19,$7E,$47,$F8,$26,$C7,$E2,$DA,$FF,$B2
	.dc.b	$5A,$FE,$2B,$F1,$AF,$F4,$DB,$F8,$66,$BF,$3B,$8D,$18,$D3,$F4,$4F
	.dc.b	$ED,$1F,$F4,$1F,$D2,$BF,$01,$FE,$82,$DB,$FF,$69,$35,$E9,$ED,$BE
	.dc.b	$BB,$F2,$FF,$8D,$59,$FD,$13,$5F,$F7,$03,$5F,$D0,$B5,$FC,$D7,$E3
	.dc.b	$5F,$1D,$F4,$37,$1F,$DC,$37,$8B,$7C,$DF,$A6,$FC,$CB,$F0,$77,$F8
	.dc.b	$CF,$2B,$FD,$01,$97,$F5,$D6,$C7,$EA,$A7,$78,$FF,$08,$87,$E3,$7F
	.dc.b	$DD,$DF,$F0,$AB,$5F,$D6,$1A,$D3,$FC,$2F,$19,$8F,$D8,$3F,$DD,$91
	.dc.b	$D3,$E9,$3F,$B7,$34,$FE,$12,$73,$3F,$7B,$F8,$D7,$FB,$15,$BE,$B7
	.dc.b	$F0,$28,$C4,$7D,$B7,$F3,$AD,$9D,$11,$E2,$7F,$08,$F6,$1F,$0D,$FC
	.dc.b	$31,$B1,$98,$C7,$D7,$65,$FF,$00,$FE,$F2,$FB,$8A,$66,$BF,$90,$EF
	.dc.b	$FE,$AD,$FB,$1F,$9D,$E9,$3F,$15,$A6,$22,$D7,$09,$FC,$DF,$0F,$F3
	.dc.b	$8B,$FF,$5B,$76,$EA,$BE,$D5,$6B,$9B,$FD,$03,$FB,$1F,$EC,$7C,$D3
	.dc.b	$FB,$3C,$38,$DF,$F2,$6E,$87,$C7,$7F,$86,$BF,$15,$78,$F4,$FE,$5B
	.dc.b	$3A,$AF,$E7,$78,$AB,$71,$86,$BF,$E3,$F6,$EB,$FD,$2E,$B9,$4F,$EB
	.dc.b	$FE,$8B,$3A,$7F,$35,$6C,$7F,$25,$6F,$85,$FE,$E2,$FA,$4F,$E7,$F1
	.dc.b	$F9,$D9,$D7,$8D,$83,$3A,$7F,$44,$F8,$17,$D7,$7E,$33,$F1,$AF,$AD
	.dc.b	$7F,$8C,$6B,$FC,$06,$76,$FE,$43,$F1,$AF,$F1,$03,$FE,$72,$6B,$F9
	.dc.b	$E6,$BF,$D2,$DA,$FE,$70,$3F,$E2,$D7,$A7,$C4,$7E,$7A,$31,$1F,$CF
	.dc.b	$8D,$FC,$47,$F5,$CF,$87,$FF,$10,$70,$FF,$73,$6F,$07,$F5,$B6,$D7
	.dc.b	$F4,$B5,$F8,$37,$17,$97,$C9,$C9,$FE,$CB,$1C,$39,$FF,$E0,$07,$27
	.dc.b	$F9,$77,$F0,$07,$EB,$7F,$5B,$3F,$F0,$3C,$63,$A7,$F3,$35,$BA,$D6
	.dc.b	$9F,$BA,$AD,$65,$FE,$00,$6B,$F9,$6A,$FE,$EF,$F8,$C7,$F2,$DA,$FE
	.dc.b	$03,$93,$F3,$FE,$76,$8C,$65,$C3,$E2,$3F,$8B,$53,$F0,$7F,$F5,$2D
	.dc.b	$71,$6C,$66,$FF,$D7,$2D,$7F,$D7,$EB,$EA,$F9,$92,$3F,$6F,$1B,$DB
	.dc.b	$FC,$45,$9B,$F0,$5F,$8D,$C3,$FB,$9B,$F7,$35,$FE,$95,$3B,$73,$E2
	.dc.b	$70,$23,$FE,$16,$8D,$3F,$D4,$74,$7E,$05,$FE,$1C,$7C,$D1,$FE,$1C
	.dc.b	$6B,$B8,$D6,$72,$34,$F1,$8F,$F1,$2F,$E2,$3F,$9E,$9F,$F1,$5B,$5F
	.dc.b	$F1,$23,$5F,$F0,$A3,$7C,$03,$FE,$AE,$D3,$FC,$88,$DE,$1E,$3F,$37
	.dc.b	$DB,$C8,$FF,$19,$FE,$25,$E6,$BF,$F2,$53,$6B,$FC,$64,$DC,$7F,$C4
	.dc.b	$4F,$CF,$8C,$F2,$72,$24,$FF,$8F,$27,$86,$4F,$C5,$3F,$C3,$F3,$FE
	.dc.b	$4B,$6B,$FE,$3D,$8D,$7C,$17,$F9,$22,$73,$3F,$E4,$C5,$F6,$3F,$E4
	.dc.b	$A6,$C7,$F9,$29,$AF,$F8,$15,$EB,$97,$F9,$95,$7D,$8F,$52,$4F,$8D
	.dc.b	$F0,$7C,$5F,$52,$47,$58,$FF,$5D,$FC,$FF,$F0,$2B,$FF,$3A,$FA,$C8
	.dc.b	$C6,$DF,$EE,$1F,$0F,$7E,$17,$0F,$06,$65,$FE,$E3,$79,$CF,$F9,$93
	.dc.b	$F1,$5F,$F3,$EC,$FF,$9A,$9A,$FF,$98,$DA,$FF,$94,$DB,$C0,$7F,$CE
	.dc.b	$99,$FF,$A0,$DB,$8F,$1F,$8E,$ED,$F4,$DF,$DE,$3F,$16,$FF,$1E,$78
	.dc.b	$5F,$DF,$3E,$97,$F8,$EC,$63,$34,$63,$CD,$FF,$50,$BA,$F1,$7F,$94
	.dc.b	$39,$BE,$5B,$7F,$2B,$C3,$94,$E7,$0E,$73,$9B,$E2,$EF,$0E,$AD,$C3
	.dc.b	$9D,$7F,$84,$30,$E7,$3C,$63,$22,$32,$EE,$B9,$B6,$B7,$3F,$E2,$CC
	.dc.b	$3C,$84,$BF,$05,$79,$8F,$06,$1E,$67,$91,$93,$C1,$87,$66,$E1,$DA
	.dc.b	$78,$46,$0E,$18,$79,$9C,$61,$F2,$DF,$2D,$F3,$9D,$12,$B9,$B8,$8D
	.dc.b	$1B,$8C,$B8,$BE,$67,$8E,$A9,$E2,$8D,$3E,$49,$E3,$3A,$BF,$89,$3F
	.dc.b	$08,$E2,$7F,$AC,$EF,$F0,$CD,$BF,$F5,$13,$F2,$8D,$79,$FF,$E3,$E7
	.dc.b	$F0,$8D,$7E,$E3,$5F,$B4,$D9,$3F,$E4,$9A,$FC,$AB,$71,$FC,$77,$9D
	.dc.b	$FE,$E6,$9C,$F2,$7D,$EA,$D6,$1D,$A3,$F0,$35,$BB,$F1,$C6,$C9,$3F
	.dc.b	$5A,$74,$F1,$3E,$CD,$74,$7C,$F3,$D3,$6E,$38,$78,$A7,$73,$96,$92
	.dc.b	$34,$7E,$C2,$FB,$C7,$43,$58,$79,$9F,$25,$F6,$8C,$57,$27,$E1,$C6
	.dc.b	$67,$D1,$39,$7F,$55,$5D,$B1,$F9,$B8,$6E,$7C,$15,$E3,$BC,$56,$A7
	.dc.b	$C6,$72,$8C,$E3,$F0,$E7,$55,$C6,$B7,$C3,$FC,$57,$EF,$18,$7F,$C0
	.dc.b	$0E,$14,$62,$BF,$D7,$90,$FC,$0B,$1C,$7F,$C6,$EA,$DF,$F2,$39,$FF
	.dc.b	$4F,$5D,$8F,$F4,$DE,$1F,$C8,$E7,$33,$FC,$9D,$B1,$FC,$9F,$F1,$7E
	.dc.b	$73,$FC,$AD,$B1,$FC,$AF,$C2,$B9,$26,$C4,$C3,$D5,$E1,$DC,$3D,$3F
	.dc.b	$01,$C0,$E3,$F1,$96,$BE,$03,$E2,$C4,$D2,$47,$F6,$5F,$3F,$C7,$DF
	.dc.b	$FE,$46,$8F,$EB,$DE,$56,$27,$2F,$B2,$4D,$24,$FC,$4F,$C0,$FF,$DF
	.dc.b	$4F,$FA,$EB,$87,$E8,$6E,$9C,$E9,$FE,$ED,$5D,$5F,$11,$CD,$70,$73
	.dc.b	$7C,$CF,$F6,$2D,$F3,$AE,$1E,$4F,$17,$E6,$F9,$3F,$DC,$AB,$86,$3F
	.dc.b	$E1,$AF,$0B,$3E,$1E,$9B,$3F,$72,$93,$F6,$9B,$19,$8F,$BB,$F6,$9F
	.dc.b	$EA,$48,$EB,$92,$31,$83,$4F,$83,$1F,$C3,$AB,$6E,$A3,$3D,$5F,$84
	.dc.b	$8C,$4F,$B9,$E1,$5C,$E3,$1D,$9F,$8C,$F4,$1F,$91,$BE,$8B,$0D,$AF
	.dc.b	$2F,$E3,$5F,$E1,$C7,$FC,$E5,$19,$C3,$A0,$D1,$5B,$8C,$E1,$DB,$0F
	.dc.b	$CC,$CE,$74,$FF,$41,$E1,$CD,$FD,$77,$B6,$BF,$93,$87,$68,$C6,$DF
	.dc.b	$A4,$75,$F5,$56,$B7,$7C,$94,$7E,$BA,$77,$A7,$FA,$AF,$00,$8C,$B9
	.dc.b	$6B,$FE,$B1,$D1,$E8,$BE,$93,$C1,$F5,$5F,$B4,$8C,$61,$FE,$DC,$C9
	.dc.b	$3F,$3F,$F5,$FF,$D5,$CF,$FA,$6D,$B8,$FB,$69,$F8,$BE,$FB,$7D,$8D
	.dc.b	$BE,$8D,$AF,$A3,$9B,$CD,$EF,$5F,$EB,$4D,$BF,$83,$C6,$2B,$6F,$C4
	.dc.b	$9F,$F5,$CE,$7A,$37,$D1,$63,$F4,$34,$F8,$9F,$3F,$0F,$F2,$38,$C6
	.dc.b	$9F,$C5,$9D,$78,$36,$B7,$7F,$ED,$95,$1F,$C8,$27,$73,$F4,$DF,$A7
	.dc.b	$3A,$8C,$46,$8F,$F3,$97,$BA,$FC,$43,$C1,$F8,$B7,$ED,$E7,$FD,$E1
	.dc.b	$18,$CF,$F9,$3F,$E2,$AF,$8C,$A6,$3D,$4B,$97,$EC,$A5,$19,$8C,$70
	.dc.b	$5C,$7F,$25,$F0,$BF,$AC,$E0,$FC,$BB,$6B,$F8,$6F,$DB,$BE,$17,$F2
	.dc.b	$D3,$9B,$FD,$34,$E6,$BF,$86,$F5,$AB,$5C,$CA,$D5,$F8,$3C,$2F,$CF
	.dc.b	$7B,$5F,$A1,$FA,$4F,$F5,$EE,$5F,$8F,$8C,$79,$DF,$A5,$D7,$FD,$9D
	.dc.b	$39,$3B,$B8,$C3,$F2,$15,$FA,$9F,$5F,$87,$C0,$DB,$DB,$FD,$43,$F1
	.dc.b	$17,$E3,$5F,$B1,$F8,$E7,$C1,$E6,$30,$DB,$FE,$3B,$5C,$1B,$EE,$E7
	.dc.b	$3F,$CC,$F7,$BC,$FF,$EB,$AD,$AF,$E4,$AF,$EB,$27,$73,$AF,$1E,$56
	.dc.b	$BC,$BB,$5F,$C5,$07,$E2,$3F,$D5,$5C,$37,$87,$F3,$4F,$EA,$6B,$F3
	.dc.b	$EE,$5F,$E0,$B3,$CB,$AB,$F3,$13,$B8,$FC,$C3,$5F,$E1,$8D,$7F,$A7
	.dc.b	$FC,$A7,$F1,$DA,$FE,$1F,$D5,$FC,$F7,$EB,$3F,$BD,$CE,$6B,$F8,$96
	.dc.b	$29,$FE,$FF,$F8,$83,$F1,$4A,$63,$ED,$6E,$5F,$D2,$BC,$07,$CA,$FF
	.dc.b	$82,$9B,$1F,$E1,$06,$E5,$FE,$DB,$6B,$EE,$D1,$F3,$8F,$86,$FE,$5C
	.dc.b	$DB,$FE,$AF,$F8,$83,$E4,$F3,$5F,$E1,$4F,$C3,$FF,$87,$70,$7F,$78
	.dc.b	$6D,$FF,$70,$7F,$88,$65,$FC,$61,$CB,$FD,$27,$F0,$F7,$C7,$FF,$88
	.dc.b	$9B,$5F,$E2,$86,$CF,$FB,$CD,$AF,$A7,$E9,$1F,$55,$FD,$15,$B5,$FD
	.dc.b	$BB,$E3,$3F,$96,$BF,$B0,$8C,$D7,$F2,$AF,$49,$F8,$BF,$F8,$B9,$B5
	.dc.b	$FD,$D9,$FE,$DB,$3F,$D7,$DC,$BF,$CC,$BD,$EF,$8C,$FF,$4D,$F3,$11
	.dc.b	$88,$FE,$94,$D7,$FC,$26,$D7,$FA,$97,$D4,$FF,$03,$BF,$E0,$AD,$8F
	.dc.b	$E7,$FF,$87,$FF,$87,$1F,$EC,$91,$9E,$0F,$F1,$3F,$E1,$CF,$B9,$FF
	.dc.b	$23,$B7,$0F,$F0,$7A,$FC,$17,$33,$FE,$4C,$FC,$3B,$FB,$1D,$7F,$95
	.dc.b	$9B,$8F,$AD,$AF,$F6,$36,$BF,$9C,$FC,$3B,$FA,$34,$FF,$1C,$6C,$7F
	.dc.b	$56,$FC,$39,$DD,$FD,$E6,$73,$9F,$F8,$FB,$F0,$EF,$63,$7E,$2B,$FC
	.dc.b	$BF,$1F,$C0,$F0,$FE,$BE,$EB,$1F,$F1,$DF,$DB,$BE,$73,$FC,$D2,$D8
	.dc.b	$FF,$37,$B6,$BE,$A3,$5F,$D0,$7D,$CF,$F4,$F7,$FC,$1C,$DC,$7F,$CA
	.dc.b	$3F,$87,$7F,$93,$2F,$F7,$33,$97,$FC,$97,$EA,$8F,$F3,$A9,$5F,$F1
	.dc.b	$EA,$F5,$B2,$FF,$9F,$3E,$69,$F4,$DF,$CE,$B2,$FF,$0A,$46,$3D,$9F
	.dc.b	$F4,$2B,$70,$FE,$CC,$D9,$DF,$E0,$DF,$AA,$F2,$1F,$7B,$FE,$2D,$BF
	.dc.b	$F0,$ED,$AB,$53,$9B,$4B,$F0,$57,$D6,$79,$5B,$87,$F5,$6F,$93,$78
	.dc.b	$99,$AF,$57,$E1,$9E,$FC,$7D,$2D,$7F,$94,$E2,$8C,$47,$91,$AF,$89
	.dc.b	$F8,$D8,$7F,$9C,$25,$F8,$07,$F5,$7E,$4F,$F5,$94,$73,$E2,$F9,$53
	.dc.b	$9A,$FE,$5B,$F1,$7F,$BE,$C7,$FC,$90,$DC,$38,$FE,$05,$FE,$BC,$AF
	.dc.b	$EE,$33,$97,$E3,$C6,$6B,$E9,$B5,$F5,$B5,$DF,$D0,$7D,$6F,$C2,$9E
	.dc.b	$FF,$DD,$6D,$FA,$DF,$E0,$B8,$FA,$9C,$DF,$F9,$AB,$44,$6A,$BE,$CB
	.dc.b	$67,$EB,$BE,$5A,$E5,$E8,$7E,$2F,$F3,$9F,$E9,$17,$F2,$5B,$7D,$57
	.dc.b	$DA,$6F,$DA,$E0,$CA,$E1,$F0,$D6,$FF,$CE,$1F,$83,$FE,$A7,$2F,$F3
	.dc.b	$AB,$6B,$FA,$DC,$7F,$BA,$1F,$F0,$DC,$66,$7F,$AD,$C7,$F8,$4C,$FF
	.dc.b	$3C,$43,$EB,$71,$1C,$1F,$F3,$B4,$66,$3F,$C8,$9D,$68,$C5,$7F,$9D
	.dc.b	$5A,$FE,$3B,$EA,$7F,$A5,$EE,$7F,$49,$F1,$5F,$C7,$F0,$FC,$64,$63
	.dc.b	$CC,$FF,$15,$1E,$4E,$2B,$3F,$E6,$0C,$3F,$D0,$BF,$4F,$FE,$52,$69
	.dc.b	$5C,$27,$FC,$AF,$F8,$47,$F7,$2C,$EF,$BF,$E9,$AF,$F2,$F4,$E5,$B1
	.dc.b	$97,$FD,$29,$1F,$EE,$6D,$F9,$FD,$17,$FB,$0E,$78,$B5,$DB,$5F,$DF
	.dc.b	$E9,$FE,$34,$7D,$FF,$17,$FE,$1A,$AF,$D4,$CE,$67,$FD,$59,$B7,$F9
	.dc.b	$F5,$FB,$3E,$BE,$F5,$FF,$0B,$E9,$99,$FC,$1C,$E5,$FE,$13,$7C,$F0
	.dc.b	$FF,$5C,$C7,$FA,$80,$FB,$25,$3F,$D5,$85,$3F,$C2,$05,$3F,$C4,$A7
	.dc.b	$F4,$63,$FC,$CA,$53,$FD,$84,$57,$BF,$CC,$78,$F8,$99,$7F,$1C,$AF
	.dc.b	$F3,$6B,$5F,$7F,$16,$DE,$59,$FF,$26,$FC,$21,$F8,$47,$B6,$91,$8A
	.dc.b	$FE,$28,$D7,$F7,$0F,$F6,$F7,$FC,$F7,$F1,$DF,$E8,$16,$FC,$A7,$96
	.dc.b	$31,$1C,$2F,$B7,$57,$FA,$56,$3F,$90,$35,$FE,$08,$DD,$7F,$90,$BF
	.dc.b	$CC,$A3,$FD,$13,$F2,$7F,$EC,$16,$FC,$67,$C3,$8C,$47,$2F,$96,$F8
	.dc.b	$DE,$1E,$23,$FC,$80,$D7,$F9,$4D,$7F,$AC,$38,$13,$99,$FE,$8F,$F9
	.dc.b	$23,$E6,$3E,$33,$63,$F9,$C3,$5F,$E6,$37,$F2,$FC,$67,$C8,$5F,$C2
	.dc.b	$65,$7F,$3F,$9C,$BF,$EA,$4F,$93,$FF,$31,$3F,$C0,$9A,$FF,$BD,$5A
	.dc.b	$FF,$96,$5A,$FF,$60,$FA,$2F,$E3,$EF,$F4,$B6,$BF,$CB,$DB,$C1,$FD
	.dc.b	$31,$EA,$B9,$BF,$CE,$FE,$8F,$F7,$DA,$FE,$A9,$F8,$4F,$99,$FE,$70
	.dc.b	$DA,$E1,$0E,$0F,$EE,$7E,$B7,$E5,$6F,$FC,$05,$3F,$C7,$1A,$FF,$A3
	.dc.b	$8E,$0E,$1F,$F3,$A3,$5F,$F4,$4F,$E6,$2F,$BF,$FD,$03,$63,$FD,$96
	.dc.b	$D5,$87,$93,$E2,$2F,$8F,$FA,$7F,$69,$F8,$57,$EE,$E7,$FD,$F9,$7C
	.dc.b	$36,$28,$CE,$1E,$ED,$58,$72,$AF,$F6,$DE,$6F,$A2,$FC,$53,$63,$F9
	.dc.b	$DB,$5F,$81,$5E,$B9,$FC,$87,$E5,$5F,$E1,$29,$FF,$72,$B5,$F2,$1F
	.dc.b	$35,$C6,$1E,$DC,$5C,$3F,$C0,$FF,$0A,$ED,$3F,$07,$F0,$AF,$F7,$76
	.dc.b	$1E,$C8,$C7,$D0,$EF,$2F,$E1,$5F,$55,$FC,$FF,$43,$49,$7F,$A5,$FC
	.dc.b	$4D,$F4,$1F,$C4,$5B,$1F,$EF,$C6,$BE,$8D,$5F,$85,$7E,$2E,$3F,$CF
	.dc.b	$3F,$13,$F4,$7F,$0A,$FE,$5D,$AF,$D7,$6F,$27,$FB,$23,$E4,$55,$FC
	.dc.b	$7F,$BB,$D7,$87,$CA,$6B,$FE,$BB,$6B,$FE,$D0,$6B,$F9,$5F,$C5,$1F
	.dc.b	$73,$FC,$A5,$B1,$F1,$70,$FF,$6B,$C6,$23,$D5,$F8,$5F,$F2,$87,$F3
	.dc.b	$DF,$31,$9D,$3F,$1D,$F8,$5D,$C3,$F1,$3B,$A3,$17,$C3,$2F,$E7,$1F
	.dc.b	$0C,$F8,$BF,$E2,$0D,$9F,$F0,$0B,$5F,$C3,$B5,$FF,$6C,$7E,$2B,$FC
	.dc.b	$C5,$FC,$2F,$32,$31,$9F,$E5,$DB,$9F,$93,$F0,$B7,$8D,$FD,$27,$E8
	.dc.b	$BF,$51,$7F,$A7,$FA,$F7,$C1,$FE,$95,$B9,$FC,$53,$F7,$D5,$C3,$1F
	.dc.b	$F7,$EF,$A7,$FD,$6E,$7F,$81,$37,$4F,$E9,$BE,$CF,$F2,$D8,$C7,$E2
	.dc.b	$DF,$D1,$33,$FD,$63,$70,$FD,$8D,$A3,$18,$7E,$CF,$F0,$BF,$97,$E6
	.dc.b	$7F,$A2,$BE,$1B,$FB,$A4,$BF,$0B,$6D,$86,$C7,$A6,$B8,$3E,$B3,$87
	.dc.b	$BF,$FE,$1F,$9F,$F0,$04,$7F,$15,$6B,$FC,$10,$D9,$FC,$64,$FF,$51
	.dc.b	$7C,$9F,$F2,$AD,$3F,$84,$DF,$FA,$7B,$DA,$FF,$04,$36,$BC,$17,$F9
	.dc.b	$E7,$F8,$77,$E1,$77,$AC,$D2,$7B,$FF,$C5,$FF,$18,$FF,$6D,$57,$D3
	.dc.b	$A4,$62,$3F,$60,$DE,$2F,$E4,$57,$D7,$DC,$7C,$3F,$F8,$27,$EC,$1F
	.dc.b	$4D,$4F,$E4,$93,$AC,$7F,$01,$CC,$AD,$C7,$87,$F1,$9F,$EE,$8F,$EF
	.dc.b	$DB,$97,$F2,$79,$DC,$FF,$1A,$DF,$FD,$83,$F8,$37,$C4,$E3,$7E,$19
	.dc.b	$4C,$45,$AF,$9D,$5F,$EC,$89,$FA,$11,$F4,$1B,$19,$8F,$F3,$4F,$E1
	.dc.b	$9F,$B7,$CC,$D3,$F7,$53,$B9,$FD,$4F,$A4,$FA,$7F,$F1,$63,$63,$E0
	.dc.b	$B5,$FE,$39,$E6,$7F,$BC,$5F,$A5,$F9,$1F,$BB,$6F,$EA,$0D,$E9,$FE
	.dc.b	$06,$D9,$FE,$CD,$7C,$EB,$FA,$B7,$D6,$DB,$FE,$12,$FF,$08,$E7,$34
	.dc.b	$D5,$FF,$55,$75,$F2,$0F,$97,$FE,$B8,$D9,$FC,$34,$FF,$12,$5F,$86
	.dc.b	$F8,$D7,$E1,$AF,$B1,$FD,$5B,$63,$F4,$D3,$B8,$FB,$F1,$FD,$E5,$AF
	.dc.b	$E9,$5A,$FF,$6A,$6A,$FC,$9F,$D6,$FE,$C5,$AF,$F5,$E3,$57,$E2,$F9
	.dc.b	$FF,$EE,$DF,$15,$FE,$3C,$DB,$FB,$E2,$F9,$6C,$9A,$6D,$FE,$97,$F1
	.dc.b	$E7,$E1,$3F,$DF,$DB,$1F,$A9,$9E,$4F,$F1,$DF,$C0,$1F,$87,$FD,$D8
	.dc.b	$FE,$E3,$18,$FA,$9F,$F4,$13,$63,$F8,$67,$77,$F5,$66,$FA,$FF,$9D
	.dc.b	$23,$D6,$FA,$69,$FE,$28,$EB,$7E,$1F,$8E,$8F,$E1,$2F,$FB,$1E,$FF
	.dc.b	$8C,$35,$FD,$53,$B5,$3C,$BF,$0B,$FD,$1C,$9A,$49,$FD,$5C,$7F,$4D
	.dc.b	$BF,$13,$6F,$C3,$EC,$FF,$18,$8E,$55,$F1,$78,$5E,$17,$C2,$6C,$66
	.dc.b	$3F,$C5,$9F,$82,$FF,$48,$8F,$F0,$17,$E0,$DF,$E1,$4C,$7F,$B2,$B6
	.dc.b	$3F,$C5,$CD,$E0,$FB,$97,$FE,$35,$F6,$3C,$35,$FE,$0B,$6E,$9F,$E3
	.dc.b	$D6,$E1,$FD,$C1,$F5,$4F,$F8,$E7,$F0,$7F,$F0,$85,$7F,$90,$DB,$5F
	.dc.b	$DD,$9B,$3F,$E3,$FA,$E6,$F1,$FA,$0F,$E1,$FE,$02,$BF,$D9,$53,$FD
	.dc.b	$0E,$31,$3A,$9C,$CF,$F8,$0B,$F0,$FF,$EC,$CF,$F9,$2D,$B4,$6C,$FF
	.dc.b	$B2,$F0,$E1,$CD,$FE,$55,$9F,$BD,$5B,$8C,$71,$7F,$96,$7F,$09,$7E
	.dc.b	$3B,$49,$97,$F4,$5C,$38,$FE,$16,$FC,$43,$D6,$D9,$FE,$B5,$92,$31
	.dc.b	$1F,$E7,$58,$C4,$F0,$F4,$3F,$CD,$15,$FC,$6E,$78,$46,$27,$3F,$23
	.dc.b	$D8,$97,$DA,$FF,$CA,$EF,$83,$F0,$EF,$86,$4F,$94,$E3,$1F,$E0,$FF
	.dc.b	$5D,$F4,$BF,$C0,$78,$7F,$DF,$FE,$0F,$FB,$5E,$31,$5F,$EF,$4F,$13
	.dc.b	$1E,$AC,$8E,$53,$FE,$CB,$BE,$31,$FD,$A7,$27,$D5,$FD,$E6,$C7,$96
	.dc.b	$30,$67,$EA,$BF,$D1,$13,$FB,$3E,$AF,$F4,$73,$6F,$FC,$82,$F2,$BF
	.dc.b	$E3,$9F,$88,$7F,$94,$3C,$3F,$F4,$6F,$D5,$7F,$68,$8C,$4A,$31,$E3
	.dc.b	$7F,$AC,$5D,$7A,$7F,$E1,$17,$37,$CB,$7F,$F0,$A6,$1C,$A7,$38,$73
	.dc.b	$9C,$DF,$17,$78,$75,$6E,$1C,$EB,$FD,$43,$87,$39,$E3,$19,$11,$97
	.dc.b	$75,$CD,$B5,$B9,$FE,$49,$1E,$33,$25,$39,$CF,$16,$CF,$1B,$F4,$47
	.dc.b	$0C,$3D,$A4,$C3,$CB,$1C,$F0,$F2,$3C,$8C,$61,$D6,$3C,$75,$C6,$7A
	.dc.b	$57,$C2,$79,$5F,$81,$E5,$3B,$6C,$F8,$23,$15,$E0,$25,$71,$71,$7E
	.dc.b	$13,$C5,$19,$71,$87,$C7,$73,$85,$8F,$F0,$FD,$7E,$1A,$3F,$DD,$0D
	.dc.b	$7D,$06,$3F,$09,$71,$70,$6C,$7F,$95,$E5,$18,$8E,$13,$B3,$EC,$47
	.dc.b	$0C,$7F,$2D,$EA,$FF,$BF,$9F,$CC,$46,$73,$FC,$8B,$6F,$F1,$2D,$7F
	.dc.b	$0F,$5C,$67,$77,$D8,$93,$F6,$67,$4F,$28,$CD,$F8,$1F,$C6,$4E,$A7
	.dc.b	$F2,$77,$DF,$0F,$6D,$70,$C3,$F4,$B7,$B9,$CD,$FE,$1E,$F8,$1C,$67
	.dc.b	$4E,$2F,$94,$FE,$92,$73,$18,$8F,$0B,$E4,$D7,$17,$DB,$1F,$45,$7B
	.dc.b	$67,$38,$75,$8C,$1E,$3B,$C9,$CD,$C1,$2B,$99,$F5,$2B,$CB,$D2,$D9
	.dc.b	$DC,$9F,$E3,$CD,$BF,$F7,$A4,$FF,$91,$E7,$27,$F2,$3F,$4F,$53,$FD
	.dc.b	$76,$74,$FF,$43,$6B,$98,$DB,$FE,$E0,$F9,$FF,$05,$7F,$A5,$5B,$1E
	.dc.b	$0F,$3B,$C5,$C7,$E3,$C9,$5A,$8F,$E5,$73,$AB,$FE,$56,$DB,$5F,$BF
	.dc.b	$CB,$C3,$F0,$66,$E4,$F8,$6A,$63,$E0,$EE,$3F,$D0,$5F,$4E,$3D,$2B
	.dc.b	$FD,$93,$3F,$D5,$67,$38,$7C,$D6,$F6,$BE,$E7,$C2,$DA,$FF,$5A,$2F
	.dc.b	$C3,$F3,$3A,$78,$92,$7F,$B2,$3D,$BC,$57,$E2,$3E,$C6,$CF,$FA,$FD
	.dc.b	$75,$66,$74,$C9,$D7,$5D,$D5,$C3,$F3,$FC,$5F,$89,$5F,$E0,$5E,$0F
	.dc.b	$93,$E4,$3E,$9F,$E5,$D6,$B0,$E3,$7F,$E2,$79,$F8,$75,$B8,$FF,$04
	.dc.b	$B6,$7F,$C1,$0D,$7F,$C4,$CB,$B7,$B9,$F6,$3E,$D7,$FD,$61,$5F,$ED
	.dc.b	$B9,$D3,$FE,$DD,$8C,$D7,$88,$FC,$0B,$8C,$BF,$14,$FF,$1C,$9E,$1D
	.dc.b	$DF,$84,$F3,$1F,$45,$4F,$C5,$79,$EF,$9D,$7D,$3E,$70,$FF,$41,$46
	.dc.b	$1B,$1F,$8E,$F4,$BF,$C5,$51,$83,$FD,$CB,$5F,$07,$3F,$D7,$7C,$12
	.dc.b	$31,$2D,$53,$9F,$37,$F3,$8E,$F8,$6D,$A7,$F2,$B3,$BB,$E5,$72,$30
	.dc.b	$E0,$EB,$5C,$67,$11,$8F,$8F,$EB,$F3,$FD,$9B,$3E,$AF,$C7,$3C,$27
	.dc.b	$C6,$FC,$45,$F4,$38,$9F,$F5,$34,$61,$B1,$E8,$FC,$47,$83,$FB,$4F
	.dc.b	$49,$8F,$ED,$BD,$F7,$A1,$98,$C3,$BC,$F1,$8F,$FA,$D5,$BA,$7F,$B1
	.dc.b	$7F,$11,$FD,$73,$7C,$3F,$E0,$8B,$E4,$BC,$8E,$6F,$E8,$72,$7F,$1C
	.dc.b	$9C,$CF,$C7,$6B,$F3,$1A,$FD,$76,$AB,$FB,$6E,$7C,$6F,$0C,$62,$3C
	.dc.b	$51,$FC,$53,$F1,$07,$AD,$FC,$5B,$EB,$BF,$8F,$65,$F8,$3F,$29,$5A
	.dc.b	$8F,$39,$99,$DE,$1F,$C8,$27,$FD,$ED,$3A,$8C,$5F,$7F,$88,$F0,$3F
	.dc.b	$C9,$3E,$2F,$F0,$53,$FA,$9C,$51,$8D,$7F,$2F,$F1,$7F,$A9,$AD,$4E
	.dc.b	$71,$FC,$C7,$52,$B5,$7C,$3D,$6F,$E6,$3F,$5E,$FB,$1F,$CD,$D6,$B0
	.dc.b	$E3,$F6,$6F,$53,$F3,$ED,$8F,$D4,$FD,$CE,$27,$F4,$2F,$E8,$3C,$03
	.dc.b	$49,$3F,$9C,$D3,$F8,$EF,$CE,$F0,$8F,$89,$18,$7A,$98,$F9,$37,$C3
	.dc.b	$7F,$48,$F9,$47,$89,$FB,$29,$CD,$7C,$DE,$45,$6A,$31,$F6,$6F,$C4
	.dc.b	$38,$63,$FD,$4A,$31,$3D,$E3,$FA,$AF,$CC,$BD,$CA,$7F,$58,$F1,$9E
	.dc.b	$7E,$2F,$C1,$18,$6C,$6F,$F1,$0E,$0F,$F0,$8E,$97,$CE,$D3,$F4,$7F
	.dc.b	$88,$7F,$51,$BF,$87,$3A,$7F,$11,$DA,$9D,$3F,$C1,$9A,$FF,$0B,$6B
	.dc.b	$FC,$41,$AF,$F1,$A9,$7C,$F7,$83,$CB,$F2,$36,$FE,$9F,$D0,$3E,$73
	.dc.b	$F8,$FB,$63,$E7,$7E,$1E,$FC,$3F,$35,$FC,$93,$DD,$FC,$CE,$7F,$C7
	.dc.b	$DA,$FE,$51,$AF,$F2,$C6,$F5,$3E,$8F,$FD,$7B,$8F,$DB,$FC,$3F,$F8
	.dc.b	$2E,$DF,$CD,$A7,$38,$7E,$96,$73,$5F,$A0,$FA,$6F,$B7,$3F,$D4,$70
	.dc.b	$46,$30,$FF,$0D,$7E,$1E,$F4,$7F,$9D,$CE,$70,$FD,$4F,$D4,$7F,$8D
	.dc.b	$1F,$CC,$37,$1F,$E7,$18,$A3,$11,$FC,$F5,$AF,$EC,$BE,$A9,$F8,$77
	.dc.b	$F8,$83,$E0,$F8,$61,$FA,$09,$CB,$F9,$98,$CD,$7A,$7E,$B5,$D5,$FD
	.dc.b	$41,$B5,$F9,$FF,$55,$F8,$76,$6B,$FA,$AF,$D7,$BE,$6F,$FA,$9B,$6F
	.dc.b	$F6,$4D,$C7,$D6,$D7,$FA,$B7,$D8,$7F,$09,$9F,$D6,$7E,$1D,$FE,$1A
	.dc.b	$7F,$6F,$19,$9F,$E2,$D3,$97,$E8,$7D,$A3,$E7,$3F,$B3,$B6,$FF,$6D
	.dc.b	$F8,$73,$F0,$EC,$D7,$F6,$CF,$89,$FE,$89,$5F,$C1,$DB,$87,$D0,$6B
	.dc.b	$FD,$AD,$AF,$F2,$0F,$C3,$37,$D5,$FE,$57,$FC,$37,$F2,$D7,$FC,$3A
	.dc.b	$72,$FF,$3B,$8C,$D7,$E1,$FE,$ED,$F0,$FF,$DF,$9B,$5F,$C3,$7F,$0D
	.dc.b	$7C,$9E,$6B,$FC,$07,$A7,$F8,$A7,$0F,$E6,$2D,$7F,$C1,$2D,$EC,$F8
	.dc.b	$9F,$19,$FE,$06,$FB,$DF,$F2,$CD,$7F,$3F,$FC,$33,$F9,$93,$FE,$10
	.dc.b	$8C,$EF,$FE,$71,$9C,$DB,$F0,$0F,$95,$AF,$F2,$A6,$E1,$FD,$27,$E8
	.dc.b	$C7,$F8,$62,$33,$87,$F8,$83,$E7,$BF,$C7,$35,$F9,$C6,$BF,$E1,$76
	.dc.b	$BF,$D3,$1A,$FF,$89,$1F,$F1,$6B,$FE,$3F,$FA,$27,$E1,$9F,$E0,$0A
	.dc.b	$ED,$F4,$FF,$E8,$39,$CF,$C7,$DF,$C3,$1F,$6F,$4C,$79,$17,$6F,$E8
	.dc.b	$BF,$37,$E0,$E0,$34,$9E,$3F,$FA,$13,$27,$E0,$8F,$84,$F0,$7E,$19
	.dc.b	$F0,$B1,$FE,$E5,$5A,$9C,$F2,$79,$7F,$1B,$F8,$B5,$FE,$41,$A4,$63
	.dc.b	$8F,$FC,$87,$F2,$7F,$84,$7E,$04,$67,$54,$E6,$7F,$C8,$0D,$7F,$BA
	.dc.b	$B5,$E7,$D7,$FE,$48,$F8,$EF,$EB,$11,$FC,$65,$BC,$BF,$EC,$3C,$5F
	.dc.b	$86,$3F,$06,$FF,$29,$D6,$B4,$F4,$4E,$5F,$7B,$70,$FA,$CF,$F2,$AC
	.dc.b	$3A,$39,$7F,$CC,$7F,$0B,$FC,$15,$BE,$67,$D9,$FC,$33,$FA,$4E,$1F
	.dc.b	$DE,$2B,$4F,$C7,$8C,$E1,$F8,$36,$DF,$A5,$FE,$B9,$19,$F7,$BF,$C1
	.dc.b	$BF,$0F,$2F,$FC,$07,$F8,$5F,$F2,$FC,$3F,$CE,$BF,$21,$FC,$D7,$0F
	.dc.b	$F3,$9C,$E6,$BF,$C3,$ED,$8F,$EB,$D1,$FE,$76,$8C,$71,$7F,$A9,$23
	.dc.b	$FC,$25,$7F,$ED,$EA,$FF,$3A,$35,$FF,$14,$B5,$EF,$F5,$7C,$CF,$CD
	.dc.b	$7D,$87,$F4,$9A,$F3,$C6,$3D,$B9,$1F,$E1,$CC,$9F,$86,$63,$0F,$F5
	.dc.b	$21,$FE,$84,$FA,$FF,$F3,$FB,$4A,$1A,$D3,$FD,$0B,$F7,$DF,$A8,$D7
	.dc.b	$8B,$63,$3E,$0A,$72,$D8,$E3,$F6,$FF,$EC,$E6,$ED,$CA,$31,$1F,$E9
	.dc.b	$5B,$46,$23,$FC,$9B,$E8,$3A,$7D,$3C,$28,$C4,$FF,$8F,$DA,$FF,$68
	.dc.b	$6B,$FE,$A4,$97,$0F,$F8,$2E,$3F,$D4,$31,$8F,$85,$F8,$AD,$8C,$CE
	.dc.b	$DF,$EE,$71,$FE,$46,$7F,$C9,$CD,$73,$F1,$FE,$68,$C4,$7F,$AC,$D7
	.dc.b	$DA,$64,$91,$C9,$E1,$F2,$92,$3A,$B5,$FF,$5D,$4B,$F0,$CF,$F2,$74
	.dc.b	$7A,$9A,$E9,$AF,$C1,$6B,$FE,$CF,$E0,$7B,$4F,$9E,$FC,$2C,$61,$B1
	.dc.b	$99,$FF,$63,$CE,$5F,$93,$D4,$F8,$DF,$92,$D8,$CC,$60,$FB,$3E,$07
	.dc.b	$DE,$8F,$F3,$B3,$5F,$F0,$FB,$5F,$F6,$D3,$5F,$E3,$7F,$61,$F8,$2F
	.dc.b	$10,$D2,$47,$DC,$FC,$37,$F0,$2F,$F2,$D6,$BF,$8B,$6E,$A6,$F9,$5F
	.dc.b	$E2,$A9,$FC,$2B,$5F,$A2,$D7,$F1,$8D,$7F,$D4,$7F,$25,$FD,$0D,$FF
	.dc.b	$49,$35,$FF,$1D,$B5,$FF,$7B,$35,$FE,$85,$F5,$15,$FD,$03,$6F,$FE
	.dc.b	$00,$79,$7F,$96,$8C,$7E,$18,$F8,$4F,$CC,$B6,$FF,$3A,$D7,$F1,$2D
	.dc.b	$44,$1F,$80,$7F,$66,$7F,$83,$B5,$D1,$C1,$C3,$86,$BC,$DF,$B0,$EF
	.dc.b	$B3,$F8,$7C,$62,$3B,$C6,$35,$91,$E0,$6B,$E0,$3C,$BC,$1C,$88,$F9
	.dc.b	$9E,$BE,$84,$8F,$4B,$5F,$F3,$5F,$E1,$8C,$FF,$85,$B6,$3F,$D9,$0D
	.dc.b	$77,$1F,$C4,$3E,$07,$F9,$3F,$79,$A4,$9F,$E2,$BF,$05,$F8,$E9,$34
	.dc.b	$9C,$1F,$C6,$BE,$1B,$FC,$DF,$5F,$ED,$76,$BF,$E8,$06,$BF,$E6,$6F
	.dc.b	$87,$E1,$F5,$B2,$38,$35,$FF,$36,$FD,$17,$F2,$6E,$3F,$F7,$84,$7F
	.dc.b	$9C,$A3,$07,$F2,$AF,$8D,$FF,$66,$C7,$FB,$9D,$AF,$FA,$41,$AF,$E1
	.dc.b	$73,$46,$23,$FD,$21,$87,$32,$3C,$DF,$E6,$FE,$D3,$EE,$A9,$27,$FD
	.dc.b	$3E,$FF,$3C,$F9,$3F,$EB,$31,$FE,$F9,$6B,$FE,$A2,$6B,$E3,$6B,$F9
	.dc.b	$AC,$3F,$D7,$9F,$29,$CC,$91,$CD,$FE,$92,$3E,$AE,$92,$3F,$1D,$29
	.dc.b	$D4,$7F,$20,$F4,$3F,$9C,$B7,$C7,$FE,$4F,$18,$8F,$D2,$F8,$68,$C4
	.dc.b	$7E,$A3,$EE,$7F,$DE,$6F,$F7,$26,$BF,$9D,$6B,$BF,$9B,$FE,$EB,$F8
	.dc.b	$2B,$EC,$69,$27,$FB,$DC,$65,$FD,$37,$D1,$3F,$05,$7D,$6F,$F8,$02
	.dc.b	$30,$D8,$CC,$F0,$6B,$FB,$E9,$E5,$F4,$3B,$24,$6D,$E9,$F5,$9F,$E9
	.dc.b	$A8,$F0,$B5,$E4,$D7,$F8,$74,$A3,$11,$FC,$4A,$31,$E8,$74,$24,$74
	.dc.b	$7F,$C0,$3F,$87,$7C,$68,$FC,$03,$57,$12,$30,$DD,$78,$FD,$C4,$AF
	.dc.b	$BF,$48,$C4,$7F,$BC,$21,$E7,$F8,$09,$1C,$1F,$99,$E7,$BD,$CA,$49
	.dc.b	$FA,$71,$96,$FC,$D7,$02,$38,$3E,$CF,$E0,$7F,$EC,$06,$FC,$B7,$D9
	.dc.b	$8C,$47,$D5,$6B,$F3,$1F,$DA,$2D,$91,$86,$FA,$1C,$CF,$F2,$0F,$E0
	.dc.b	$AE,$0A,$4B,$FF,$23,$C6,$5F,$CC,$69,$FE,$48,$E7,$F8,$B1,$82,$1F
	.dc.b	$9B,$DF,$44,$7F,$40,$FE,$4F,$CC,$FE,$8F,$1E,$76,$E5,$FE,$0B,$F7
	.dc.b	$B4,$6F,$C1,$95,$F9,$EA,$46,$23,$EC,$69,$F8,$62,$3F,$79,$FC,$4F
	.dc.b	$DD,$48,$FC,$0B,$5F,$EB,$BC,$BF,$E6,$3F,$B4,$FF,$27,$FC,$09,$A4
	.dc.b	$9F,$EA,$9F,$1A,$FC,$71,$F3,$BF,$E6,$E8,$C3,$63,$33,$C1,$AF,$F6
	.dc.b	$C9,$7E,$82,$FC,$0B,$93,$67,$BB,$5E,$0D,$7F,$5B,$F7,$5F,$EF,$27
	.dc.b	$F8,$1B,$5F,$DE,$35,$DD,$7F,$A1,$3A,$DF,$91,$3E,$3F,$0D,$AF,$EE
	.dc.b	$53,$A9,$C9,$FC,$37,$83,$E3,$4F,$F1,$26,$BF,$C2,$DA,$FF,$02,$6B
	.dc.b	$FC,$05,$7E,$05,$CC,$91,$CD,$FE,$35,$F7,$DF,$E7,$38,$FE,$16,$D7
	.dc.b	$F8,$53,$5F,$D1,$37,$98,$BE,$67,$99,$27,$F9,$4C,$65,$FE,$57,$F8
	.dc.b	$07,$EE,$5F,$E4,$8D,$7F,$8D,$B5,$FE,$34,$D7,$F6,$4B,$C0,$FE,$64
	.dc.b	$D8,$EB,$18,$3F,$9B,$7E,$02,$F7,$69,$27,$F9,$EC,$65,$BF,$0F,$C0
	.dc.b	$8E,$0F,$E8,$7F,$81,$7F,$AF,$B3,$34,$91,$FD,$15,$FF,$32,$C6,$BD
	.dc.b	$BC,$12,$79,$FC,$BE,$C8,$E0,$EF,$F8,$1F,$32,$38,$3F,$D5,$7F,$05
	.dc.b	$FE,$AE,$F8,$DF,$F3,$CC,$60,$87,$FA,$79,$FE,$0B,$7F,$EB,$8F,$0B
	.dc.b	$81,$27,$FB,$57,$81,$F9,$FE,$FF,$89,$0F,$11,$39,$E0,$FF,$2F,$46
	.dc.b	$3A,$BF,$CC,$37,$C2,$3F,$96,$B6,$B8,$5F,$F9,$9A,$78,$3F,$EE,$7B
	.dc.b	$E7,$3A,$8F,$07,$93,$FE,$EF,$9C,$E1,$D9,$B3,$C4,$E9,$5C,$6F,$C3
	.dc.b	$3A,$C3,$C5,$7C,$6F,$6D,$7B,$96,$BC,$B3,$C2,$3B,$1F,$0E,$FB,$61
	.dc.b	$E5,$8E,$66,$AB,$8E,$1E,$13,$E3,$CE,$E7,$96,$1D,$EF,$D0,$74,$8C
	.dc.b	$44,$7A,$46,$5B,$19,$C3,$CC,$7D,$1A,$ED,$87,$8A,$BC,$B3,$AC,$3C
	.dc.b	$71,$C5,$F3,$3A,$72,$F6,$24,$71,$7E,$D5,$76,$C3,$F9,$14,$7C,$23
	.dc.b	$FC,$6C,$E1,$F7,$9F,$E3,$5F,$BC,$FF,$64,$E0,$69,$33,$FF,$1E,$27
	.dc.b	$EA,$36,$7E,$77,$E0,$5F,$E0,$3C,$7F,$21,$39,$6D,$70,$3E,$E4,$EF
	.dc.b	$2F,$CC,$D6,$DB,$3B,$BE,$BD,$97,$0E,$71,$F7,$6F,$83,$F9,$BC,$3A
	.dc.b	$3F,$9D,$8C,$BF,$A9,$6E,$3E,$FC,$3C,$78,$75,$C7,$C3,$7D,$67,$2D
	.dc.b	$26,$3E,$23,$F5,$D8,$F8,$23,$B4,$F3,$7C,$4F,$29,$E1,$5D,$9B,$3E
	.dc.b	$9B,$E0,$F2,$3B,$CF,$EF,$AF,$B6,$1E,$99,$C9,$D8,$F8,$E6,$0E,$57
	.dc.b	$D9,$FE,$15,$3C,$B1,$F9,$3C,$4F,$A2,$FE,$79,$97,$F9,$C1,$B7,$FE
	.dc.b	$69,$6B,$FE,$60,$6B,$FE,$5F,$57,$FD,$26,$BA,$36,$33,$5D,$63,$5C
	.dc.b	$8F,$97,$A4,$9F,$E7,$EE,$BF,$16,$91,$FD,$09,$B8,$FF,$9F,$6B,$FD
	.dc.b	$19,$3F,$D1,$BD,$7E,$A4,$7F,$A4,$FD,$6F,$80,$8E,$0F,$EA,$7E,$A7
	.dc.b	$A7,$AF,$FC,$F9,$18,$26,$1F,$E8,$B8,$79,$9B,$24,$FF,$5C,$7E,$17
	.dc.b	$77,$C0,$AF,$ED,$FA,$E0,$95,$FD,$D3,$83,$64,$70,$7F,$7A,$F9,$9F
	.dc.b	$99,$E0,$1A,$48,$FE,$FF,$EF,$68,$8E,$0F,$F5,$5F,$D0,$FF,$9F,$5A
	.dc.b	$7D,$77,$18,$FF,$A9,$DA,$FF,$A7,$9F,$BE,$FF,$9C,$DB,$C8,$69,$A9
	.dc.b	$7E,$45,$FE,$D1,$7F,$C5,$ED,$A1,$0F,$F4,$83,$97,$11,$CB,$E0,$3F
	.dc.b	$19,$E5,$3E,$F2,$92,$7F,$C8,$AF,$F8,$67,$BF,$FC,$63,$ED,$5F,$24
	.dc.b	$D2,$57,$F9,$83,$E4,$7F,$CC,$8D,$F4,$B9,$46,$33,$FF,$2E,$35,$F4
	.dc.b	$7A,$6F,$AC,$7C,$87,$F8,$A6,$30,$D8,$CE,$1F,$EF,$46,$D7,$F9,$D9
	.dc.b	$F6,$46,$74,$E0,$47,$86,$9C,$C8,$E0,$FF,$45,$7C,$CF,$EC,$7A,$4D
	.dc.b	$24,$7F,$A3,$FC,$FD,$11,$C1,$F4,$F1,$FA,$9D,$AF,$9E,$FF,$21,$36
	.dc.b	$75,$3F,$6A,$30,$E1,$AF,$37,$EE,$3D,$38,$F9,$92,$79,$C3,$F0,$A5
	.dc.b	$7F,$94,$5B,$3A,$6C,$F8,$9A,$FF,$BD,$3F,$02,$7D,$47,$F1,$36,$D7
	.dc.b	$A3,$A5,$18,$AF,$81,$18,$AE,$1E,$4C,$BE,$17,$FE,$5B,$6C,$7F,$1F
	.dc.b	$F8,$17,$D9,$D2,$4F,$FB,$A1,$D7,$E0,$32,$3F,$DD,$6D,$CB,$FD,$6F
	.dc.b	$F8,$DF,$29,$F9,$BC,$E8,$C4,$EA,$FF,$1D,$18,$8F,$C9,$AF,$92,$C9
	.dc.b	$2F,$9C,$65,$BF,$27,$C8,$8E,$0F,$8D,$F4,$CF,$A5,$FE,$78,$D8,$CC
	.dc.b	$60,$CF,$1C,$8E,$2D,$C7,$E6,$FC,$57,$F9,$6A,$7E,$93,$63,$3E,$A7
	.dc.b	$F4,$98,$C4,$7F,$05,$E3,$43,$D5,$D1,$2F,$9B,$FE,$97,$FB,$CF,$D6
	.dc.b	$DF,$E6,$DA,$FF,$0F,$6B,$FC,$49,$AF,$E4,$32,$FC,$F9,$2B,$F2,$BC
	.dc.b	$FF,$06,$7F,$4D,$F2,$0E,$27,$B7,$FE,$A9,$8C,$36,$33,$7F,$A8,$9C
	.dc.b	$BF,$7B,$67,$D8,$FF,$09,$6C,$66,$30,$7F,$0D,$F9,$37,$BD,$49,$3F
	.dc.b	$88,$8C,$B7,$B7,$81,$1C,$1F,$8E,$F8,$8F,$E3,$9A,$7F,$34,$CD,$18
	.dc.b	$8F,$73,$63,$3E,$3F,$F1,$FF,$96,$FC,$65,$7F,$09,$94,$62,$3F,$93
	.dc.b	$B5,$EC,$D7,$E5,$2E,$7E,$84,$C3,$F4,$71,$97,$F4,$BD,$2F,$C6,$3F
	.dc.b	$98,$36,$33,$18,$3F,$9A,$7C,$CF,$E9,$F0,$F7,$35,$FB,$4D,$C4,$D3
	.dc.b	$CC,$E2,$F9,$92,$FF,$6F,$F3,$5F,$E1,$2B,$F8,$4D,$7F,$A3,$B5,$F9
	.dc.b	$CD,$7F,$15,$F3,$76,$CD,$27,$B5,$FD,$35,$FF,$3A,$5F,$F0,$66,$D7
	.dc.b	$D1,$6B,$FD,$59,$AF,$F3,$CE,$AF,$E7,$64,$9F,$CE,$BF,$CF,$AA,$47
	.dc.b	$E8,$1B,$3F,$CF,$7E,$07,$FA,$19,$27,$F4,$AF,$F4,$68,$91,$FA,$86
	.dc.b	$9F,$D2,$3F,$20,$7C,$EA,$7F,$A0,$C6,$1B,$19,$9E,$0D,$7F,$C1,$14
	.dc.b	$FB,$0F,$EB,$F5,$FD,$52,$31,$E5,$EE,$BF,$C1,$F1,$89,$D5,$70,$9F
	.dc.b	$EC,$12,$FB,$0F,$E0,$AD,$F4,$79,$4F,$FA,$8E,$72,$7F,$0A,$DB,$FC
	.dc.b	$45,$E8,$7F,$80,$7D,$7F,$EF,$75,$FA,$36,$BF,$E0,$86,$BE,$4F,$A3
	.dc.b	$FF,$0A,$E7,$E9,$6C,$79,$1A,$FC,$06,$BE,$CF,$A4,$FE,$1A,$FA,$9A
	.dc.b	$FC,$66,$BB,$F0,$7F,$C2,$DF,$4A,$F8,$9A,$4C,$3F,$A8,$C6,$5F,$EA
	.dc.b	$7E,$77,$F0,$46,$F7,$7F,$87,$E3,$18,$7F,$01,$6B,$F5,$BE,$B7,$FC
	.dc.b	$06,$FD,$26,$BF,$EC,$D6,$F0,$9C,$F8,$AF,$B5,$7C,$A7,$F2,$D6,$E5
	.dc.b	$FC,$AD,$B8,$7F,$34,$6B,$F8,$6F,$AE,$7C,$C7,$DA,$C9,$18,$BF,$C4
	.dc.b	$36,$BF,$06,$D7,$FC,$7F,$E7,$7F,$A5,$E7,$FB,$AB,$63,$F9,$03,$5D
	.dc.b	$FC,$5F,$F9,$0E,$1F,$17,$49,$5C,$9D,$71,$48,$EA,$DB,$FC,$AF,$C7
	.dc.b	$BD,$BF,$53,$6B,$F4,$0D,$9D,$CF,$F9,$4F,$EB,$1F,$6B,$49,$7F,$E0
	.dc.b	$E8,$CB,$FE,$32,$AF,$F3,$AF,$C8,$54,$D2,$3F,$E4,$2F,$AE,$FC,$57
	.dc.b	$1F,$F8,$E9,$B1,$9E,$84,$E5,$B1,$C7,$BD,$F5,$FC,$5B,$19,$8C,$1E
	.dc.b	$EF,$0B,$FD,$10,$FF,$05,$E4,$6F,$FE,$60,$AF,$F4,$6C,$E6,$31,$1D
	.dc.b	$B4,$FE,$BD,$AA,$F9,$C7,$F6,$17,$FB,$9C,$F2,$8C,$7A,$5A,$8F,$F0
	.dc.b	$04,$E9,$E8,$73,$BE,$F8,$7F,$19,$F0,$7B,$7D,$37,$F9,$79,$B5,$FE
	.dc.b	$3E,$B4,$E5,$B1,$9F,$A0,$F0,$57,$9D,$B2,$6D,$6B,$F1,$5F,$E5,$B7
	.dc.b	$E9,$3F,$C2,$CE,$1F,$3D,$A3,$18,$79,$F0,$E9,$5C,$78,$1F,$5C,$F9
	.dc.b	$9F,$85,$18,$6C,$66,$7C,$14,$FA,$FF,$F4,$54,$63,$E1,$9F,$82,$FC
	.dc.b	$BA,$FF,$40,$36,$33,$3E,$1B,$E8,$FA,$3C,$3F,$F5,$7B,$ED,$D5,$18
	.dc.b	$8D,$72,$F5,$FA,$3F,$F7,$BC,$79,$3E,$AC,$78,$4A,$E5,$F9,$F5,$CF
	.dc.b	$7F,$C7,$79,$1F,$87,$AF,$E9,$CD,$7F,$98,$6B,$FE,$9A,$11,$98,$FE
	.dc.b	$AF,$F8,$B7,$CE,$E7,$FE,$A6,$D8,$FC,$93,$6B,$89,$F8,$0F,$A0,$7C
	.dc.b	$7D,$25,$7C,$EF,$33,$FD,$22,$FD,$0F,$A4,$FF,$73,$3F,$D8,$DA,$FF
	.dc.b	$5E,$6B,$FD,$A9,$AF,$F6,$6F,$41,$F1,$FF,$55,$B1,$F7,$1A,$FE,$83
	.dc.b	$1F,$EF,$78,$7D,$BF,$B1,$F1,$CE,$5B,$B7,$ED,$DB,$AF,$F7,$47,$FD
	.dc.b	$31,$3F,$97,$1F,$59,$FE,$80,$7F,$D8,$AD,$7F,$DA,$2D,$7F,$DB,$0D
	.dc.b	$44,$FC,$C3,$FC,$CE,$F5,$6B,$FB,$80,$E2,$FF,$C1,$DF,$65,$F8,$1C
	.dc.b	$39,$D6,$BE,$0F,$F8,$13,$4E,$8E,$25,$F4,$97,$E8,$7F,$96,$41,$A4
	.dc.b	$AF,$F6,$E7,$D5,$F4,$EC,$34,$E9,$F2,$7F,$C4,$7D,$0F,$DC,$7D,$8D
	.dc.b	$F1,$0D,$3F,$83,$FD,$97,$E3,$65,$F8,$EF,$93,$2F,$F7,$13,$5F,$F7
	.dc.b	$0B,$5F,$F5,$83,$5F,$97,$0F,$C6,$3F,$DD,$B1,$FC,$36,$31,$F3,$3E
	.dc.b	$59,$CC,$FF,$0E,$F8,$6F,$D3,$71,$FF,$8E,$ED,$F8,$C3,$EC,$FD,$8D
	.dc.b	$BE,$8D,$7F,$8B,$FC,$53,$ED,$7E,$17,$D3,$FC,$69,$FF,$7E,$35,$FF
	.dc.b	$7D,$B5,$FF,$65,$35,$F5,$FE,$30,$FA,$5F,$98,$D8,$F6,$35,$FE,$4F
	.dc.b	$F1,$DF,$39,$F8,$1F,$8C,$7F,$B4,$DB,$F3,$5F,$56,$31,$1F,$CB,$BE
	.dc.b	$49,$E4,$7D,$9F,$C7,$7F,$C3,$73,$F8,$C3,$E6,$B8,$71,$C0,$8C,$47
	.dc.b	$FB,$8D,$AF,$D4,$FC,$67,$FD,$50,$FD,$36,$BF,$D5,$25,$18,$E2,$FE
	.dc.b	$A9,$F1,$5F,$C8,$2B,$5F,$BF,$7E,$F2,$7E,$B3,$7C,$3F,$C0,$B6,$7F
	.dc.b	$DC,$1F,$30,$F9,$CF,$CA,$7E,$3D,$FC,$5A,$7D,$83,$34,$66,$3D,$AD
	.dc.b	$7F,$DF,$CD,$7F,$1F,$F8,$DF,$C8,$7F,$1C,$D7,$D8,$DC,$BF,$B7,$7C
	.dc.b	$EF,$D7,$7F,$D0,$DF,$66,$FA,$9F,$DB,$B6,$FF,$87,$35,$FE,$F3,$EF
	.dc.b	$BE,$F3,$FB,$8C,$7F,$05,$F3,$9F,$35,$FE,$25,$6C,$7F,$14,$8C,$1F
	.dc.b	$DF,$7E,$EF,$DD,$97,$F8,$05,$BC,$5F,$EE,$D6,$D7,$F1,$6B,$E6,$FD
	.dc.b	$7F,$BC,$FE,$3D,$7F,$A7,$6D,$FE,$C9,$B3,$F3,$5F,$D3,$4F,$EB,$7F
	.dc.b	$0E,$6D,$FC,$1D,$B5,$FB,$76,$BD,$FE,$7B,$F8,$67,$9F,$FE,$00,$8F
	.dc.b	$C4,$36,$FF,$70,$DC,$7F,$81,$AF,$BF,$FE,$29,$F8,$7F,$F7,$6A,$FD
	.dc.b	$F3,$6B,$F8,$4B,$6F,$F0,$97,$CD,$7C,$1B,$6F,$F3,$BC,$7F,$7D,$6B
	.dc.b	$FE,$18,$6E,$1F,$37,$8F,$FC,$4B,$0F,$82,$A4,$AE,$5F,$2F,$FE,$25
	.dc.b	$F8,$99,$1F,$E0,$9F,$4B,$EF,$E3,$F5,$EF,$F9,$77,$C1,$BE,$EA,$92
	.dc.b	$FF,$90,$47,$FB,$4E,$BF,$8E,$7E,$24,$FC,$07,$F9,$B3,$6B,$F9,$33
	.dc.b	$5F,$DA,$57,$57,$F9,$57,$B1,$F9,$BE,$2F,$E5,$ED,$8F,$D5,$37,$BF
	.dc.b	$FA,$2F,$93,$FD,$0F,$E4,$FF,$D1,$B5,$FC,$CD,$AF,$F0,$16,$E7,$FC
	.dc.b	$C6,$FA,$4B,$BF,$E4,$3F,$A4,$DB,$F4,$E7,$F5,$EF,$C0,$7E,$B4,$6F
	.dc.b	$8B,$FD,$0D,$5F,$52,$3F,$8D,$35,$FE,$CE,$D7,$FB,$9F,$9D,$EF,$7F
	.dc.b	$B8,$37,$1F,$F7,$E3,$75,$E3,$7F,$D7,$FE,$C9,$E4,$52,$61,$FD,$46
	.dc.b	$3F,$DB,$4F,$F5,$3F,$C5,$3F,$C9,$2F,$F2,$16,$E3,$FD,$45,$B8,$7F
	.dc.b	$5C,$AE,$B5,$F8,$FF,$AE,$FC,$59,$1F,$C9,$79,$BF,$8E,$E1,$91,$F9
	.dc.b	$57,$4F,$FA,$0B,$E1,$DF,$10,$FC,$0B,$FD,$11,$18,$6C,$66,$7F,$CB
	.dc.b	$5F,$32,$FA,$CD,$FE,$05,$FC,$FA,$7F,$C6,$6D,$C0,$D3,$5E,$0F,$FA
	.dc.b	$DB,$E4,$3F,$D0,$F8,$7F,$3D,$E1,$FF,$49,$37,$83,$FC,$4B,$87,$2F
	.dc.b	$A3,$FF,$4E,$43,$F0,$27,$E0,$58,$6E,$9F,$E8,$B9,$D4,$FF,$84,$FE
	.dc.b	$EB,$FA,$27,$71,$FF,$52,$67,$FD,$FD,$B5,$FE,$0A,$6D,$A3,$9B,$FE
	.dc.b	$0B,$E3,$7D,$D7,$F8,$7D,$B1,$98,$C7,$0F,$F8,$8F,$EF,$7F,$CB,$B8
	.dc.b	$FF,$83,$5B,$87,$F8,$61,$B7,$FD,$4F,$0E,$6F,$F6,$58,$FF,$76,$3F
	.dc.b	$D6,$32,$46,$23,$FB,$9B,$5F,$F2,$1C,$6B,$E8,$FF,$AF,$C7,$F5,$FF
	.dc.b	$C5,$BE,$E3,$FD,$C5,$B8,$FF,$88,$5B,$8F,$96,$BF,$C7,$9F,$80,$7E
	.dc.b	$86,$FF,$B4,$37,$03,$4F,$E2,$15,$FE,$34,$5F,$87,$F2,$25,$7F,$8D
	.dc.b	$63,$97,$E0,$5F,$EA,$A8,$F4,$B5,$FF,$1E,$B5,$FE,$F0,$DB,$3F,$C6
	.dc.b	$72,$FF,$1D,$37,$0E,$3E,$47,$F9,$56,$3F,$9F,$53,$F0,$3F,$F5,$16
	.dc.b	$1F,$E4,$A6,$BF,$E0,$56,$C6,$FE,$63,$E5,$BC,$7B,$3F,$CC,$13,$F2
	.dc.b	$2B,$84,$62,$B3,$F5,$32,$7F,$CA,$94,$8E,$3F,$71,$CB,$8E,$56,$9E
	.dc.b	$0F,$FA,$6F,$E9,$7F,$85,$BF,$E3,$5C,$11,$88,$FF,$41,$C6,$3E,$A7
	.dc.b	$FD,$81,$39,$9F,$F4,$8C,$F0,$FA,$1F,$F6,$F4,$FF,$A2,$E5,$39,$7F
	.dc.b	$DC,$F1,$E0,$E0,$FF,$0C,$F8,$7F,$EB,$0E,$E7,$CB,$7F,$2C,$8C,$37
	.dc.b	$BF,$FD,$3F,$3A,$C3,$FD,$43,$E5,$7F,$94,$A3,$11,$FC,$FE,$F8,$BA
	.dc.b	$6B,$9A,$E7,$7F,$CE,$BC,$5F,$F5,$0D,$F2,$C1,$87,$06,$DF,$0C,$3A
	.dc.b	$35,$FE,$9D,$7C,$B0,$F4,$43,$0E,$98,$73,$3A,$56,$E7,$99,$BA,$E3
	.dc.b	$5C,$2F,$B6,$1E,$F9,$C9,$2B,$8D,$7A,$6F,$A4,$7B,$CE,$86,$A7,$83
	.dc.b	$E8,$3A,$E1,$E5,$C2,$9E,$0A,$E7,$87,$C2,$AE,$67,$5C,$3C,$CE,$23
	.dc.b	$CC,$F7,$21,$BC,$3E,$83,$D8,$E1,$87,$85,$E5,$7E,$08,$C5,$F6,$3D
	.dc.b	$D3,$C4,$FE,$EB,$1F,$EF,$87,$E8,$98,$C3,$F8,$CE,$0F,$C0,$FF,$6A
	.dc.b	$DD,$33,$18,$23,$F4,$1F,$CD,$D7,$B7,$E3,$7F,$9E,$3E,$F8,$CE,$3F
	.dc.b	$89,$6E,$7F,$E0,$66,$CF,$E6,$2B,$8D,$6E,$FB,$12,$7B,$74,$59,$D4
	.dc.b	$77,$9F,$C1,$D7,$2C,$3B,$46,$DB,$3C,$8E,$04,$AE,$67,$E7,$EB,$94
	.dc.b	$7E,$B3,$0F,$3B,$FE,$29,$6E,$3F,$1E,$3A,$57,$2B,$EE,$D8,$F7,$63
	.dc.b	$E3,$7A,$63,$5F,$F1,$A6,$3E,$07,$B4,$67,$0E,$D7,$E5,$BC,$63,$E2
	.dc.b	$3C,$78,$71,$AD,$E3,$E2,$26,$3E,$23,$E0,$3F,$EF,$EF,$C1,$3F,$05
	.dc.b	$87,$FB,$F5,$B3,$FE,$CB,$7F,$CC,$91,$9C,$3E,$FE,$DF,$EF,$D9,$CD
	.dc.b	$7F,$B6,$9A,$FF,$68,$6B,$DF,$D4,$FC,$3B,$FD,$B3,$9B,$FC,$03,$87
	.dc.b	$F3,$AF,$07,$AB,$FD,$27,$EB,$AD,$7F,$49,$9B,$D5,$FD,$3E,$6D,$9B
	.dc.b	$F5,$D8,$73,$F1,$13,$1F,$E8,$FF,$09,$72,$FF,$74,$61,$FE,$C0,$9F
	.dc.b	$F0,$23,$63,$FD,$EA,$D7,$D1,$3E,$A7,$FB,$5F,$D6,$F9,$3B,$CD,$F7
	.dc.b	$B0,$DC,$3F,$C2,$12,$9E,$55,$FD,$C9,$BA,$7F,$72,$6F,$84,$F2,$FE
	.dc.b	$0B,$70,$FF,$00,$73,$62,$72,$C8,$FF,$02,$72,$5C,$7F,$C2,$DB,$7C
	.dc.b	$AF,$1B,$FD,$F9,$49,$FF,$13,$7E,$09,$9A,$F9,$FE,$42,$3C,$AD,$AF
	.dc.b	$F1,$53,$6B,$FC,$53,$3A,$F9,$9E,$27,$D8,$FB,$4F,$F2,$15,$F7,$E9
	.dc.b	$FF,$1D,$4E,$62,$19,$3C,$9C,$0A,$FC,$7D,$7F,$10,$8C,$4F,$E1,$3C
	.dc.b	$BF,$ED,$EF,$F9,$37,$BB,$F1,$2B,$F0,$4F,$F7,$3B,$93,$FC,$BC,$E3
	.dc.b	$36,$1F,$C7,$27,$87,$C3,$FE,$43,$CE,$7D,$B7,$4C,$38,$5F,$F2,$2F
	.dc.b	$4D,$F8,$23,$CA,$FC,$CC,$61,$FC,$DF,$8D,$FE,$60,$F8,$5C,$D6,$F4
	.dc.b	$F8,$1A,$AB,$51,$F5,$62,$46,$1C,$1B,$F0,$97,$F8,$C9,$D4,$F0,$FA
	.dc.b	$7F,$F4,$85,$F6,$94,$62,$7E,$0C,$FE,$0F,$2F,$5F,$99,$E8,$AD,$69
	.dc.b	$EF,$CD,$CB,$EA,$3D,$7B,$DC,$7D,$78,$FC,$48,$CC,$62,$BD,$CD,$8F
	.dc.b	$DE,$FB,$4F,$5E,$9F,$C0,$3D,$AF,$D6,$D7,$FB,$06,$31,$E1,$C8,$FF
	.dc.b	$63,$7A,$D7,$7F,$59,$F9,$DF,$A2,$FF,$36,$CF,$CE,$84,$61,$D8,$FB
	.dc.b	$E7,$D3,$7F,$12,$8D,$F8,$B6,$73,$7F,$E8,$46,$CF,$D4,$39,$57,$EF
	.dc.b	$9B,$3B,$EA,$E4,$75,$C3,$C1,$1E,$5B,$FC,$6B,$FA,$2F,$84,$FE,$4F
	.dc.b	$7F,$EF,$FE,$8F,$E4,$7F,$82,$7D,$2A,$FC,$4F,$7A,$F7,$CD,$FA,$7A
	.dc.b	$D6,$1F,$12,$91,$89,$D7,$A1,$D3,$9F,$F9,$7F,$C3,$BE,$CE,$9F,$CC
	.dc.b	$BE,$21,$F8,$1E,$35,$F0,$C6,$1B,$1B,$FC,$13,$83,$FA,$4F,$B6,$7E
	.dc.b	$25,$4F,$D3,$FE,$0B,$C7,$83,$EC,$F7,$AB,$78,$FE,$32,$B5,$BF,$EC
	.dc.b	$5B,$1F,$99,$5F,$43,$D3,$2E,$AC,$9E,$5F,$81,$DC,$BE,$27,$C9,$BE
	.dc.b	$53,$FA,$7B,$6B,$E6,$FE,$0B,$FC,$A2,$B7,$5F,$D5,$23,$19,$FF,$15
	.dc.b	$6D,$7F,$1A,$7E,$DD,$39,$91,$A8,$C7,$CB,$BD,$CF,$EB,$ED,$BF,$BB
	.dc.b	$F2,$FF,$E7,$47,$EF,$C6,$6B,$EE,$4E,$5B,$F0,$5F,$DA,$21,$F0,$76
	.dc.b	$3E,$7F,$D2,$CC,$3F,$49,$F8,$2B,$CD,$C6,$3F,$A4,$8C,$4F,$FB,$89
	.dc.b	$B1,$F9,$F6,$BF,$C5,$1A,$FF,$19,$F0,$7D,$CF,$EB,$7D,$8F,$EF,$BF
	.dc.b	$3A,$E9,$FF,$5D,$61,$FA,$19,$FE,$FF,$F8,$8B,$E3,$69,$FC,$B3,$C7
	.dc.b	$FC,$A5,$7E,$82,$74,$FE,$8B,$EE,$FC,$39,$7C,$F6,$F2,$9A,$63,$F1
	.dc.b	$17,$D6,$7F,$83,$9B,$7F,$C6,$BE,$93,$F7,$BD,$3F,$D4,$67,$38,$FE
	.dc.b	$AF,$F1,$17,$9F,$FE,$16,$6E,$1F,$C8,$BE,$63,$F9,$4B,$FD,$72,$33
	.dc.b	$5F,$B3,$FC,$75,$F7,$FF,$E0,$26,$D7,$F0,$BD,$91,$8A,$FE,$83,$18
	.dc.b	$AF,$F0,$4F,$99,$FD,$2E,$7F,$7C,$DD,$3F,$AB,$7D,$5B,$C3,$FF,$15
	.dc.b	$CE,$72,$FE,$DF,$D1,$FD,$DB,$0F,$E0,$CD,$C3,$FA,$F7,$88,$FC,$27
	.dc.b	$FC,$67,$39,$BF,$EE,$FF,$10,$F9,$5F,$F0,$DB,$63,$D4,$DC,$7D,$2D
	.dc.b	$7F,$AF,$6C,$FC,$43,$FC,$A7,$9F,$F7,$F6,$DF,$F5,$5F,$B0,$FD,$EB
	.dc.b	$FC,$82,$33,$7F,$AD,$FC,$41,$F8,$1F,$F9,$21,$B5,$FD,$53,$ED,$3F
	.dc.b	$C7,$D8,$FF,$82,$E7,$37,$FC,$97,$ED,$9F,$84,$FF,$8D,$1B,$87,$D3
	.dc.b	$6C,$FD,$26,$BE,$AF,$47,$F2,$6F,$F2,$F6,$E1,$FE,$07,$FC,$2F,$FC
	.dc.b	$2F,$39,$F0,$BF,$A7,$7E,$2D,$F8,$4A,$FF,$33,$B7,$1F,$D5,$79,$9A
	.dc.b	$EE,$FE,$A5,$CB,$FD,$34,$FE,$93,$8F,$F5,$16,$BF,$E0,$66,$BF,$E0
	.dc.b	$BB,$FD,$CE,$1D,$BC,$9F,$EB,$9F,$19,$E2,$7C,$79,$CC,$FF,$AA,$61
	.dc.b	$4F,$E0,$9F,$29,$FD,$77,$CB,$FF,$2E,$4F,$F0,$58,$C1,$FE,$5E,$C1
	.dc.b	$F8,$5F,$C2,$9F,$F2,$84,$63,$0F,$F3,$FB,$AF,$83,$B5,$BC,$52,$FC
	.dc.b	$05,$F7,$FC,$1B,$8F,$F8,$AA,$75,$A6,$6B,$AF,$D4,$7F,$01,$AF,$EE
	.dc.b	$D4,$8C,$61,$FE,$2C,$8C,$79,$7E,$99,$CC,$FF,$8C,$FC,$FF,$F2,$A7
	.dc.b	$C5,$7F,$8F,$BE,$4F,$FC,$8B,$8F,$87,$F0,$2F,$F1,$9B,$FE,$22,$8C
	.dc.b	$DF,$F8,$53,$27,$E1,$BF,$90,$E4,$FF,$75,$C6,$3B,$FE,$1F,$E0,$BF
	.dc.b	$C9,$2F,$85,$EF,$1F,$93,$39,$9F,$F2,$35,$A3,$11,$AF,$A6,$FA,$51
	.dc.b	$88,$F9,$7F,$91,$7F,$03,$BF,$90,$D8,$FE,$E3,$F0,$8F,$47,$EA,$CE
	.dc.b	$74,$FF,$30,$4E,$6F,$FC,$7E,$DB,$F7,$7E,$19,$FC,$A2,$7F,$C9,$93
	.dc.b	$9A,$F7,$FD,$7B,$DB,$FC,$0B,$6B,$FC,$C7,$ED,$BF,$0C,$FF,$50,$65
	.dc.b	$F2,$6B,$4F,$F8,$FA,$75,$F4,$1F,$D3,$2B,$78,$FC,$CF,$C3,$1F,$63
	.dc.b	$9A,$FF,$09,$7B,$F3,$3F,$F2,$BF,$A1,$FE,$CB,$7F,$CB,$3E,$57,$F7
	.dc.b	$5E,$BF,$EF,$1F,$4D,$FE,$7C,$AD,$6D,$FE,$16,$7F,$B9,$69,$FE,$83
	.dc.b	$72,$DF,$84,$FF,$5F,$DF,$F8,$05,$B3,$FD,$E6,$BF,$18,$D7,$FC,$E4
	.dc.b	$F8,$4F,$F0,$37,$43,$EA,$FF,$82,$4F,$F6,$99,$CC,$62,$75,$E5,$FF
	.dc.b	$8B,$FB,$3F,$5B,$3C,$2F,$FD,$19,$F5,$7F,$E5,$76,$94,$35,$A7,$F9
	.dc.b	$77,$E4,$BF,$D3,$DC,$BC,$5B,$19,$F9,$E4,$E5,$B1,$C7,$C6,$FF,$57
	.dc.b	$B7,$7E,$51,$88,$FF,$4E,$DF,$F8,$4E,$7C,$DF,$9F,$7F,$AB,$67,$FC
	.dc.b	$F4,$D7,$FC,$F2,$D8,$71,$FF,$AA,$63,$F3,$53,$C3,$0F,$F5,$65,$6A
	.dc.b	$3F,$DC,$53,$FE,$21,$7A,$E1,$AB,$E5,$76,$7F,$DD,$1F,$4C,$EA,$F9
	.dc.b	$0D,$8F,$98,$D7,$FD,$08,$D7,$FD,$72,$FF,$BB,$5F,$F4,$A3,$5F,$F4
	.dc.b	$9B,$71,$F3,$42,$3A,$46,$1D,$5F,$FB,$1D,$FF,$09,$CF,$4D,$FD,$0F
	.dc.b	$1F,$A4,$F5,$70,$BF,$0D,$FE,$F1,$3F,$CA,$57,$87,$A6,$C6,$7D,$CC
	.dc.b	$4E,$7A,$24,$F1,$8E,$BF,$6B,$FC,$DA,$3F,$9A,$3F,$35,$FE,$44,$DD
	.dc.b	$7F,$90,$BB,$8C,$F0,$72,$E0,$91,$CA,$3D,$3E,$2B,$EA,$BF,$35,$AF
	.dc.b	$F2,$16,$C6,$6B,$F9,$FB,$5F,$E7,$CF,$FA,$FD,$FE,$90,$D7,$F9,$B3
	.dc.b	$71,$FE,$6A,$F1,$7F,$11,$F8,$C3,$F4,$B7,$D7,$7E,$16,$30,$D8,$CD
	.dc.b	$FF,$45,$F9,$8C,$72,$FE,$C6,$F9,$36,$BF,$D9,$11,$8A,$FF,$66,$3F
	.dc.b	$EA,$8F,$9D,$ED,$F9,$EB,$EE,$69,$E1,$F9,$B4,$FD,$EF,$C3,$59,$D2
	.dc.b	$63,$FD,$9E,$1F,$85,$FE,$13,$F1,$FF,$F3,$1C,$FF,$66,$6B,$FD,$95
	.dc.b	$AF,$F5,$06,$A3,$FD,$3D,$BB,$9F,$F5,$1E,$22,$3F,$D3,$D3,$FE,$D8
	.dc.b	$6B,$FE,$D5,$78,$3F,$ED,$CE,$0F,$F0,$0B,$79,$39,$C6,$23,$FD,$B8
	.dc.b	$FD,$53,$C5,$C1,$FD,$BB,$A7,$E7,$7A,$5F,$9E,$AF,$31,$A7,$0E,$27
	.dc.b	$83,$FE,$71,$F9,$EE,$D8,$FF,$07,$C9,$1D,$E7,$FD,$78,$D8,$CE,$5F
	.dc.b	$03,$EF,$7F,$CD,$15,$FE,$F6,$6B,$FE,$F3,$6D,$F5,$5F,$75,$F2,$3E
	.dc.b	$1B,$F5,$15,$FE,$66,$6B,$F3,$F0,$46,$23,$FD,$9C,$D8,$7E,$16,$F9
	.dc.b	$0F,$A6,$FF,$A8,$EB,$EC,$63,$ED,$8C,$47,$D7,$DF,$B4,$71,$8D,$C7
	.dc.b	$87,$F0,$B7,$C4,$7D,$E6,$CE,$8F,$F7,$03,$87,$A1,$8E,$9F,$E5,$9F
	.dc.b	$5E,$FD,$8D,$F5,$5F,$D5,$63,$0D,$8C,$D7,$DA,$3F,$21,$3A,$7F,$D5
	.dc.b	$1F,$86,$3E,$9F,$F2,$2F,$F9,$F2,$3C,$9B,$A3,$11,$E4,$8C,$4F,$E6
	.dc.b	$3F,$0C,$7D,$B5,$25,$72,$8F,$F0,$D3,$D3,$F0,$CF,$F0,$7B,$FE,$B9
	.dc.b	$6B,$FE,$B8,$6B,$FE,$A0,$6B,$F6,$FF,$0C,$7C,$D7,$EA,$17,$D5,$D3
	.dc.b	$11,$FA,$CF,$C3,$1F,$13,$49,$7E,$E8,$7D,$77,$CE,$FA,$6F,$EA,$53
	.dc.b	$F2,$1A,$B8,$11,$86,$FA,$DF,$E0,$7E,$07,$E1,$7F,$02,$FE,$43,$FE
	.dc.b	$CD,$7E,$6B,$5C,$FC,$37,$F6,$9F,$80,$FC,$BF,$E0,$37,$BC,$D3,$F8
	.dc.b	$3F,$E4,$7F,$42,$FF,$14,$3F,$0B,$FE,$00,$FE,$05,$AF,$EF,$1A,$FF
	.dc.b	$13,$6B,$C4,$FE,$C6,$FE,$27,$C0,$FE,$DA,$DE,$FE,$5E,$27,$F1,$8F
	.dc.b	$C2,$FF,$97,$5F,$E2,$17,$DB,$64,$D7,$81,$67,$5F,$09,$C4,$C3,$F8
	.dc.b	$DF,$99,$7E,$15,$FC,$91,$7C,$D5,$31,$16,$7F,$2F,$2F,$9A,$E0,$66
	.dc.b	$3F,$57,$C5,$FC,$0B,$EE,$3F,$BF,$FE,$20,$FB,$0A,$7C,$F8,$B5,$F3
	.dc.b	$25,$F1,$DC,$0C,$C7,$ED,$84,$3F,$07,$EA,$48,$E5,$3F,$CF,$BE,$51
	.dc.b	$F5,$14,$93,$FB,$08,$FF,$0D,$37,$EC,$B8,$11,$D4,$FF,$48,$E1,$FC
	.dc.b	$77,$3F,$F3,$16,$F3,$7F,$21,$6D,$3B,$7F,$B2,$7D,$5F,$1F,$16,$4F
	.dc.b	$F0,$0A,$FD,$6B,$7B,$BF,$C4,$BF,$85,$BE,$DE,$92,$7F,$8D,$78,$6F
	.dc.b	$C1,$BF,$03,$1F,$94,$FC,$21,$F3,$1F,$E6,$96,$C7,$F9,$B5,$AF,$F1
	.dc.b	$70,$E1,$FE,$34,$BF,$09,$FF,$1B,$B6,$3B,$46,$27,$F9,$3F,$2F,$F7
	.dc.b	$C9,$F2,$BF,$C1,$9B,$E6,$6B,$E0,$F6,$F0,$DE,$4E,$8F,$6F,$57,$77
	.dc.b	$F9,$AE,$9F,$89,$CB,$F9,$AC,$E7,$93,$FC,$D1,$3F,$8D,$CF,$3E,$D4
	.dc.b	$C3,$9B,$C3,$0F,$F2,$8F,$D3,$FF,$95,$A7,$FC,$0C,$D4,$DF,$2F,$FB
	.dc.b	$0C,$7C,$DF,$B0,$F9,$84,$7E,$63,$F3,$FE,$13,$E3,$92,$38,$1E,$CD
	.dc.b	$FF,$D3,$FE,$EF,$F4,$6D,$BE,$44,$FF,$A7,$A3,$13,$A8,$FE,$45,$3F
	.dc.b	$DD,$BE,$63,$FB,$06,$DF,$AB,$AD,$3F,$AD,$C2,$D7,$F8,$AA,$FF,$C4
	.dc.b	$57,$C2,$3B,$BC,$BC,$6F,$F4,$23,$73,$E7,$18,$CB,$FA,$DB,$FD,$52
	.dc.b	$75,$7D,$27,$3E,$C7,$03,$8D,$FA,$67,$F1,$BD,$FF,$CE,$DF,$71,$FE
	.dc.b	$35,$E6,$46,$70,$FE,$F9,$3C,$34,$EC,$EF,$EE,$3A,$5F,$09,$CC,$EC
	.dc.b	$EA,$5A,$E7,$2A,$E3,$7D,$F2,$FF,$18,$CF,$3B,$E2,$DB,$E3,$7C,$30
	.dc.b	$EF,$39,$BE,$38,$73,$C3,$CC,$69,$C3,$E3,$8D,$D7,$2B,$E9,$3E,$0A
	.dc.b	$DD,$F5,$8E,$66,$6B,$A1,$1F,$3D,$72,$C3,$DF,$7C,$8E,$71,$9C,$3C
	.dc.b	$35,$CC,$E2,$73,$AE,$37,$67,$E2,$D7,$1A,$FB,$17,$D6,$BD,$51,$B3
	.dc.b	$2D,$7F,$86,$CF,$17,$E1,$1A,$3C,$27,$8E,$FC,$2D,$BF,$0B,$CF,$0F
	.dc.b	$88,$F2,$3A,$99,$C3,$C8,$F4,$AE,$64,$AF,$82,$FD,$C8,$E1,$87,$5C
	.dc.b	$3C,$D8,$76,$23,$8C,$3A,$DF,$1B,$F3,$1E,$B7,$B3,$F6,$1A,$F9,$EB
	.dc.b	$9B,$E0,$38,$9E,$A2,$BE,$19,$FB,$31,$F7,$CF,$E6,$AF,$EA,$DB,$7F
	.dc.b	$CF,$63,$11,$F9,$CF,$96,$FF,$08,$D7,$FB,$DF,$CD,$FE,$85,$18,$23
	.dc.b	$F9,$C7,$FD,$3B,$87,$EA,$3C,$27,$85,$9A,$98,$7E,$9A,$F7,$A7,$EA
	.dc.b	$AF,$73,$9A,$DE,$1F,$B2,$BD,$CF,$F0,$1A,$DC,$E6,$B7,$7E,$97,$F5
	.dc.b	$F3,$A8,$C5,$75,$25,$79,$F0,$E3,$1F,$C3,$70,$EE,$FF,$5E,$6E,$3F
	.dc.b	$E0,$49,$DC,$E7,$1B,$39,$8E,$B8,$78,$30,$F3,$47,$F0,$6C,$7C,$8F
	.dc.b	$7B,$E3,$8F,$69,$CC,$F5,$39,$9B,$39,$63,$C2,$BC,$33,$C3,$1F,$B1
	.dc.b	$5C,$0E,$58,$7B,$DC,$46,$A3,$CA,$64,$B8,$FA,$49,$3F,$47,$0F,$03
	.dc.b	$CA,$FC,$18,$FC,$62,$63,$F4,$AB,$78,$74,$BE,$47,$8A,$B6,$FB,$0C
	.dc.b	$57,$B2,$B8,$E1,$E2,$9F,$F0,$8E,$1E,$0A,$CC,$74,$C3,$B3,$F0,$2B
	.dc.b	$86,$3F,$56,$72,$FC,$18,$CD,$F3,$AF,$09,$E3,$AF,$46,$6F,$D3,$7F
	.dc.b	$D6,$99,$7F,$4E,$87,$E9,$CB,$FB,$B1,$BE,$0B,$39,$AE,$17,$CA,$FF
	.dc.b	$AE,$DE,$E7,$3F,$5F,$CA,$B5,$8F,$F7,$AF,$D8,$3F,$C7,$D5,$E1,$7F
	.dc.b	$CE,$6D,$F0,$7C,$31,$8C,$7A,$BC,$1B,$1A,$FB,$6D,$BF,$90,$FB,$07
	.dc.b	$C8,$7E,$D3,$C0,$FD,$AD,$A7,$38,$FF,$A7,$17,$D0,$EE,$3E,$14,$EE
	.dc.b	$F0,$FC,$5A,$DC,$FF,$92,$E3,$A5,$76,$78,$CE,$9F,$F2,$83,$5F,$F7
	.dc.b	$47,$77,$FA,$42,$BF,$D5,$AD,$8F,$F4,$07,$E0,$2F,$92,$F0,$C7,$F8
	.dc.b	$DB,$B3,$C7,$C4,$DD,$F5,$38,$6E,$5F,$E6,$85,$F3,$39,$9D,$32,$75
	.dc.b	$F3,$57,$1F,$F2,$8F,$EE,$4F,$03,$BB,$F7,$F8,$BF,$CA,$2E,$6F,$FC
	.dc.b	$EA,$E5,$FC,$56,$68,$7C,$FF,$2B,$E6,$DA,$FA,$5F,$3D,$EE,$E2,$37
	.dc.b	$E5,$70,$D9,$FA,$12,$9F,$A1,$9F,$D5,$6B,$F7,$DB,$C5,$F4,$F4,$FC
	.dc.b	$1B,$6F,$CF,$E8,$3E,$77,$E7,$CE,$70,$FF,$4E,$7A,$4F,$C1,$BF,$07
	.dc.b	$EA,$A3,$18,$FE,$A7,$2F,$E8,$72,$FC,$0F,$FD,$F9,$3F,$EA,$1B,$E3
	.dc.b	$18,$C1,$7C,$23,$0D,$F8,$AD,$F1,$F8,$3D,$86,$D4,$C7,$B1,$73,$FF
	.dc.b	$5F,$38,$7F,$D8,$1E,$A5,$CF,$F1,$BF,$23,$EE,$3D,$FE,$7D,$AF,$F7
	.dc.b	$FF,$0F,$FB,$51,$B1,$F1,$3D,$AD,$C7,$BC,$FE,$11,$EE,$BE,$53,$E2
	.dc.b	$47,$29,$FD,$4C,$E7,$93,$94,$F1,$CB,$FD,$91,$F8,$0F,$FB,$76,$31
	.dc.b	$CA,$8F,$F7,$FC,$61,$D4,$FC,$88,$FE,$31,$3B,$F8,$4F,$AB,$4F,$84
	.dc.b	$BE,$EE,$1B,$7D,$BE,$55,$EA,$FD,$C5,$F3,$F6,$30,$E7,$D3,$C5,$71
	.dc.b	$6E,$9F,$EF,$7F,$2B,$F6,$2D,$F1,$FF,$73,$0F,$1A,$C6,$1F,$CB,$C6
	.dc.b	$6F,$F8,$63,$6B,$FD,$B0,$D7,$F8,$AF,$0F,$CF,$CF,$E6,$F1,$9F,$CA
	.dc.b	$FC,$4F,$E1,$B0,$F9,$D5,$B7,$E6,$FC,$6B,$D0,$FC,$DB,$FC,$BF,$C0
	.dc.b	$FC,$CB,$99,$DF,$C4,$7F,$10,$FD,$7F,$F8,$B6,$9F,$C9,$BC,$14,$62
	.dc.b	$3F,$80,$B5,$46,$BC,$EF,$DB,$F4,$BE,$FF,$F6,$18,$FE,$01,$B7,$F7
	.dc.b	$44,$EF,$C9,$FD,$8C,$EB,$4F,$D7,$FD,$FF,$F2,$96,$FC,$57,$ED,$A3
	.dc.b	$18,$FE,$19,$47,$EE,$BE,$FF,$FA,$66,$5F,$D0,$32,$46,$23,$F1,$AD
	.dc.b	$7C,$6D,$87,$B8,$FB,$6F,$E0,$AD,$C7,$F3,$6F,$F3,$26,$C7,$F0,$AF
	.dc.b	$C6,$9F,$23,$7D,$7C,$36,$BF,$4D,$C2,$FC,$A3,$FB,$96,$9F,$D5,$8F
	.dc.b	$9A,$E1,$FA,$8D,$7F,$99,$B5,$F9,$7F,$7F,$F9,$EA,$FD,$7B,$63,$E8
	.dc.b	$68,$8F,$F2,$27,$07,$F1,$C7,$4F,$F1,$EF,$BE,$FD,$FD,$70,$7F,$6E
	.dc.b	$DF,$1F,$F9,$3C,$62,$7F,$92,$FC,$AF,$F8,$26,$BE,$83,$5F,$DD,$B5
	.dc.b	$FC,$0B,$70,$7D,$8F,$E5,$67,$CE,$DA,$FE,$62,$FE,$6F,$2F,$E0,$AE
	.dc.b	$5F,$E0,$FF,$84,$FF,$00,$7F,$87,$BF,$EF,$46,$F8,$DF,$CF,$A3,$17
	.dc.b	$F9,$3F,$A4,$FA,$FE,$6F,$F5,$06,$DF,$D3,$94,$61,$FA,$CD,$7F,$1D
	.dc.b	$F8,$F7,$CA,$9F,$E3,$2D,$8F,$CF,$BF,$42,$33,$E4,$FF,$50,$A7,$E1
	.dc.b	$77,$D8,$C3,$6B,$F9,$67,$C5,$3E,$A7,$FC,$3F,$97,$E1,$E3,$1F,$09
	.dc.b	$FD,$A1,$B5,$E9,$6C,$EF,$EF,$FE,$42,$84,$61,$BE,$2F,$F8,$99,$CF
	.dc.b	$8F,$FE,$28,$D1,$F5,$9F,$B7,$C7,$FC,$9D,$0F,$B0,$E9,$1F,$D0,$E3
	.dc.b	$7E,$FF,$19,$CE,$7F,$46,$FF,$35,$5D,$23,$E9,$46,$09,$8F,$85,$E3
	.dc.b	$CD,$FE,$B6,$87,$DE,$3E,$C7,$51,$E9,$FB,$17,$B9,$4C,$7C,$1D,$DB
	.dc.b	$E6,$CF,$F8,$45,$F9,$0D,$CF,$E7,$35,$FF,$25,$DA,$31,$1F,$E6,$F8
	.dc.b	$7D,$D3,$A7,$E5,$3F,$D7,$B4,$F4,$39,$C3,$EE,$B9,$79,$3F,$DD,$EF
	.dc.b	$F9,$45,$F4,$BE,$7E,$EF,$47,$EB,$7D,$CB,$E6,$EF,$CD,$E1,$B9,$7E
	.dc.b	$2A,$5F,$37,$98,$C3,$72,$FE,$6C,$3E,$B7,$F0,$ED,$E8,$FF,$2C,$C6
	.dc.b	$0F,$DA,$47,$FA,$3D,$C7,$46,$BE,$47,$6D,$C3,$FC,$DD,$3F,$CC,$6B
	.dc.b	$B5,$7F,$8F,$A3,$1E,$33,$E5,$3F,$CC,$FF,$B7,$7F,$26,$E3,$FF,$40
	.dc.b	$CE,$71,$FF,$08,$62,$AD,$47,$FA,$02,$75,$3E,$C6,$D7,$2F,$97,$FF
	.dc.b	$39,$C6,$2B,$F8,$07,$CA,$FF,$A0,$EF,$FD,$54,$D4,$D3,$FD,$11,$F6
	.dc.b	$FF,$E0,$3F,$B8,$BF,$06,$69,$2B,$F8,$64,$FF,$9E,$B2,$E3,$F2,$BF
	.dc.b	$E4,$89,$CB,$5F,$E8,$93,$B7,$F8,$A7,$DC,$7F,$A0,$9F,$F4,$C3,$5C
	.dc.b	$CF,$A6,$31,$1F,$DF,$7E,$3D,$F4,$BF,$E9,$FF,$57,$FC,$A2,$D8,$CC
	.dc.b	$FF,$21,$9C,$BF,$EA,$5F,$3F,$FD,$02,$DF,$95,$F8,$D1,$88,$FE,$50
	.dc.b	$D7,$FA,$E2,$FB,$8D,$92,$3A,$3F,$3F,$F3,$87,$D2,$52,$4F,$FA,$D6
	.dc.b	$32,$FF,$1D,$F2,$9F,$6F,$FE,$B9,$F0,$FF,$89,$36,$33,$3B,$8C,$7C
	.dc.b	$EC,$8E,$4D,$7F,$96,$79,$3F,$E3,$D7,$F9,$53,$5C,$F2,$F7,$8C,$47
	.dc.b	$F0,$47,$FD,$72,$F0,$F8,$ED,$11,$FE,$1B,$F0,$DE,$7F,$27,$FD,$B5
	.dc.b	$18,$21,$FC,$43,$F0,$8D,$11,$FE,$2A,$FF,$B5,$FC,$F7,$DD,$7F,$A9
	.dc.b	$9B,$DB,$98,$C1,$9F,$4F,$FD,$A5,$7F,$D7,$6B,$FD,$DB,$F0,$3C,$E7
	.dc.b	$33,$FD,$79,$AF,$F2,$0C,$7F,$CA,$8F,$F2,$6F,$8E,$F5,$11,$FE,$A1
	.dc.b	$F7,$4F,$A8,$FA,$8D,$8C,$C6,$0F,$B1,$EC,$74,$8F,$EB,$8D,$7F,$97
	.dc.b	$B5,$FE,$5C,$D7,$FC,$97,$E0,$B1,$FF,$7D,$7B,$FF,$EB,$26,$C6,$67
	.dc.b	$FB,$E4,$E5,$BB,$7D,$27,$FC,$65,$08,$C4,$7F,$92,$DA,$FF,$A5,$9A
	.dc.b	$FF,$81,$BE,$EB,$FA,$7B,$F3,$5A,$FF,$82,$1A,$8F,$ED,$2E,$1F,$F5
	.dc.b	$0F,$DD,$74,$A3,$49,$C5,$EE,$FB,$AF,$E7,$75,$EC,$6B,$EA,$6B,$FE
	.dc.b	$63,$6B,$FE,$34,$E6,$F8,$FF,$07,$F9,$A8,$C1,$F1,$9C,$3F,$E7,$8F
	.dc.b	$94,$7D,$7F,$CE,$6C,$7C,$C6,$BF,$29,$AF,$F9,$CF,$CF,$7C,$DF,$D6
	.dc.b	$6C,$7D,$46,$BF,$49,$AF,$FA,$57,$EE,$BF,$B6,$3F,$C5,$5A,$FF,$9A
	.dc.b	$7C,$64,$62,$38,$7B,$3F,$84,$F8,$CF,$E7,$9C,$06,$92,$7F,$13,$F7
	.dc.b	$0F,$A0,$A4,$8F,$C4,$C6,$5F,$C5,$FD,$C7,$F8,$45,$BE,$17,$E4,$A3
	.dc.b	$11,$F8,$C6,$BF,$E9,$5F,$B8,$7D,$87,$E6,$1B,$1F,$96,$6B,$F9,$46
	.dc.b	$BF,$E9,$DF,$B8,$FF,$4B,$BF,$9C,$6B,$F9,$A6,$BB,$F9,$4F,$E8,$1F
	.dc.b	$37,$FC,$D6,$3F,$25,$C0,$8C,$47,$FA,$F9,$B1,$9F,$43,$F5,$BF,$6E
	.dc.b	$FA,$5A,$4B,$FD,$6C,$65,$FF,$35,$0F,$B7,$FF,$28,$38,$6B,$FE,$6F
	.dc.b	$6B,$FE,$A3,$6B,$FA,$73,$91,$FD,$D5,$FE,$E8,$47,$FB,$C7,$D0,$FF
	.dc.b	$82,$5F,$F7,$8F,$BF,$82,$3F,$C0,$BE,$DD,$F4,$9F,$C2,$1B,$19,$8C
	.dc.b	$19,$FA,$59,$1C,$5A,$FF,$08,$ED,$FF,$69,$BF,$C3,$9A,$E7,$E7,$BF
	.dc.b	$8A,$C6,$23,$FD,$AC,$FF,$65,$78,$7B,$1A,$23,$F2,$BE,$DD,$F1,$9F
	.dc.b	$E1,$56,$C6,$63,$06,$70,$91,$C5,$AF,$FA,$76,$7D,$ED,$C7,$A1,$1F
	.dc.b	$E4,$FF,$52,$FB,$3F,$E5,$AD,$8C,$C6,$0C,$FD,$A4,$8E,$2D,$7F,$96
	.dc.b	$FC,$EF,$E1,$5B,$C7,$FC,$E2,$31,$1F,$EF,$96,$BF,$C9,$FC,$04,$7F
	.dc.b	$80,$A3,$9F,$BD,$C4,$8F,$E4,$3E,$D3,$FC,$C8,$DF,$90,$FF,$1E,$46
	.dc.b	$23,$FD,$7C,$D7,$FD,$E4,$FF,$26,$7F,$9B,$35,$34,$9F,$23,$D0,$91
	.dc.b	$E1,$7F,$C5,$DB,$FA,$A1,$D4,$8C,$37,$3D,$9F,$E3,$51,$F2,$F4,$95
	.dc.b	$FE,$35,$8C,$BF,$E3,$7F,$19,$87,$F8,$F5,$B1,$98,$C1,$FE,$41,$FB
	.dc.b	$27,$D3,$D2,$4F,$F9,$06,$32,$FC,$BF,$5F,$E4,$91,$C1,$F3,$FE,$D7
	.dc.b	$E6,$91,$C1,$0F,$F3,$4F,$85,$FB,$BD,$BF,$8E,$FC,$6F,$F1,$86,$C6
	.dc.b	$6B,$F8,$04,$FF,$1E,$DB,$FD,$59,$E3,$62,$BF,$8C,$EA,$BC,$63,$F3
	.dc.b	$A3,$0F,$F9,$EA,$77,$5A,$BF,$F5,$B5,$FF,$0D,$F5,$BF,$87,$46,$2F
	.dc.b	$FB,$B4,$63,$0E,$71,$FE,$05,$3C,$35,$C5,$FE,$57,$AB,$EC,$BB,$2F
	.dc.b	$90,$A6,$22,$CF,$94,$FC,$2C,$ED,$C6,$3F,$5A,$B8,$46,$CF,$C9,$F4
	.dc.b	$FE,$31,$F9,$A7,$E6,$1C,$6B,$F3,$5B,$AF,$D1,$E9,$F8,$F5,$CB,$0E
	.dc.b	$9B,$FE,$FD,$B7,$C2,$7F,$47,$5C,$2B,$51,$E0,$39,$15,$CE,$1F,$76
	.dc.b	$78,$57,$3A,$DC,$F9,$6F,$C3,$5A,$7C,$24,$9E,$91,$97,$B5,$73,$BE
	.dc.b	$F8,$7B,$67,$86,$1D,$5D,$E1,$D9,$CC,$FE,$4E,$F9,$1C,$F0,$ED,$7C
	.dc.b	$49,$3D,$6B,$99,$23,$2F,$C6,$3A,$9F,$4E,$EB,$FC,$2A,$BA,$D7,$86
	.dc.b	$73,$82,$73,$5C,$4E,$D5,$E1,$BE,$F1,$F2,$A3,$93,$F1,$8E,$78,$7B
	.dc.b	$30,$F8,$11,$C8,$E9,$5C,$B0,$F0,$47,$09,$F1,$E1,$E2,$8E,$C7,$0C
	.dc.b	$3C,$47,$82,$F9,$61,$F2,$67,$91,$DC,$F8,$06,$23,$C1,$5D,$70,$FB
	.dc.b	$33,$D4,$F9,$47,$17,$2F,$88,$E6,$66,$BD,$CF,$C1,$3E,$A9,$FC,$F1
	.dc.b	$FD,$33,$70,$FE,$8D,$18,$8F,$E2,$9F,$13,$FE,$10,$7F,$9F,$B6,$8D
	.dc.b	$27,$C4,$7E,$A1,$FF,$0E,$E3,$FC,$6F,$EB,$7F,$83,$3F,$C7,$63,$39
	.dc.b	$F6,$AE,$07,$EC,$2B,$71,$9C,$3F,$55,$5C,$1F,$EA,$6D,$7F,$A9,$CF
	.dc.b	$07,$35,$D1,$E2,$F5,$E1,$C6,$1E,$38,$B8,$7F,$33,$BE,$57,$F0,$27
	.dc.b	$2F,$43,$F7,$91,$AC,$7E,$1C,$EA,$78,$1A,$C3,$13,$A8,$F1,$99,$2E
	.dc.b	$1E,$36,$E1,$E3,$AE,$38,$FB,$EF,$95,$FC,$D9,$CE,$1D,$B1,$F2,$5E
	.dc.b	$C9,$1F,$07,$1E,$A7,$AE,$FC,$07,$37,$27,$A9,$CE,$3E,$38,$CE,$3E
	.dc.b	$28,$FA,$D3,$C4,$DC,$F5,$8C,$9C,$63,$31,$D6,$32,$FF,$83,$B0,$F4
	.dc.b	$DF,$B2,$3E,$35,$70,$BF,$04,$75,$38,$63,$B9,$D5,$F2,$9F,$3E,$3F
	.dc.b	$26,$31,$3C,$E3,$15,$C7,$1F,$1D,$FE,$32,$30,$F7,$30,$7C,$AC,$DF
	.dc.b	$6B,$FD,$9A,$3F,$A8,$C6,$1F,$CA,$2F,$B5,$FE,$C5,$5F,$D7,$2B,$86
	.dc.b	$3F,$DF,$67,$37,$CA,$73,$87,$FA,$5A,$73,$5F,$E0,$39,$CD,$7F,$8B
	.dc.b	$39,$9F,$2F,$FD,$D7,$1F,$EF,$71,$87,$CD,$3B,$9D,$3F,$EA,$29,$D3
	.dc.b	$FD,$F4,$E9,$F2,$DE,$A8,$CE,$1F,$E1,$35,$19,$AF,$D4,$FB,$AF,$81
	.dc.b	$BF,$8E,$E1,$B9,$7F,$79,$9D,$3F,$EA,$66,$E5,$FE,$9F,$01,$CB,$96
	.dc.b	$BF,$EA,$55,$F1,$9E,$98,$C3,$5E,$71,$A9,$F0,$4E,$77,$7E,$93,$F1
	.dc.b	$AF,$FD,$8A,$D7,$FC,$B2,$D8,$DC,$71,$D1,$F1,$B4,$98,$7F,$92,$DE
	.dc.b	$DF,$91,$BD,$FA,$49,$F8,$17,$FE,$66,$9D,$46,$30,$7D,$C7,$E1,$AB
	.dc.b	$FC,$6E,$D7,$FD,$B6,$D7,$FC,$DE,$D7,$E5,$7C,$87,$C3,$6D,$7D,$18
	.dc.b	$C6,$A6,$FD,$C4,$9F,$AC,$DA,$F3,$D3,$E8,$38,$92,$BE,$E3,$F4,$7C
	.dc.b	$CF,$F4,$CF,$D2,$DF,$34,$D2,$57,$C2,$9F,$F4,$FF,$D3,$6C,$91,$9A
	.dc.b	$47,$5F,$3D,$F7,$94,$95,$FE,$BA,$75,$F3,$12,$3F,$D7,$AD,$D3,$E8
	.dc.b	$F8,$FF,$D7,$A7,$FD,$F2,$D8,$CF,$81,$F2,$63,$11,$F9,$7F,$B7,$7D
	.dc.b	$2F,$CD,$6D,$FD,$C9,$46,$1F,$BE,$D7,$97,$97,$C4,$93,$C6,$51,$F2
	.dc.b	$FE,$DF,$FD,$87,$3F,$1D,$B5,$FA,$9D,$D1,$88,$FD,$58,$8C,$EF,$CF
	.dc.b	$A2,$5F,$36,$D7,$3F,$94,$4F,$DB,$6B,$F5,$DB,$1B,$8D,$7D,$54,$74
	.dc.b	$D8,$FC,$FF,$BF,$FC,$2A,$7E,$33,$63,$3E,$1F,$E5,$E3,$11,$FC,$15
	.dc.b	$FD,$7C,$F0,$ED,$D1,$1E,$7F,$1F,$C4,$8E,$0F,$C7,$7C,$27,$4B,$FF
	.dc.b	$77,$35,$F8,$CD,$D4,$E7,$B9,$F6,$AF,$7B,$F8,$53,$70,$FD,$C3,$70
	.dc.b	$FE,$39,$5C,$2B,$F4,$9F,$6A,$F3,$E9,$30,$FD,$14,$65,$FD,$27,$DA
	.dc.b	$FF,$36,$C3,$ED,$E6,$8C,$47,$F2,$88,$C7,$95,$FB,$19,$CC,$FE,$07
	.dc.b	$ED,$7E,$4B,$F8,$76,$BD,$1A,$F3,$6B,$F8,$9F,$3D,$F8,$7F,$EF,$5B
	.dc.b	$8F,$EE,$5B,$3F,$93,$6B,$E8,$F9,$07,$D1,$7F,$05,$6D,$7F,$01,$6C
	.dc.b	$FE,$6D,$AF,$FB,$5B,$ED,$1E,$FF,$F3,$96,$D7,$F2,$76,$CF,$D0,$6B
	.dc.b	$F2,$3E,$D3,$F2,$ED,$9F,$E2,$D1,$88,$F9,$4D,$7F,$51,$F5,$2E,$BF
	.dc.b	$E3,$8D,$AF,$E3,$0D,$9F,$D9,$35,$FA,$FF,$31,$FE,$E5,$9F,$E3,$CD
	.dc.b	$8F,$DC,$35,$DF,$87,$FC,$A7,$ED,$3F,$24,$4E,$DC,$1F,$CB,$7E,$D1
	.dc.b	$ED,$52,$63,$FC,$AE,$32,$FF,$4E,$FA,$37,$DA,$FF,$39,$8C,$37,$2F
	.dc.b	$EE,$53,$AF,$3F,$FB,$53,$6B,$F9,$B7,$DA,$FC,$5E,$A3,$49,$1F,$D0
	.dc.b	$65,$F3,$14,$91,$FC,$FE,$32,$FF,$62,$ED,$7D,$17,$F7,$7C,$BF,$C0
	.dc.b	$91,$8F,$6B,$FB,$E3,$6B,$85,$6E,$BF,$8F,$CB,$ED,$7F,$78,$DE,$0E
	.dc.b	$31,$83,$3F,$4A,$F9,$1E,$AD,$C7,$C8,$DC,$3C,$6D,$7F,$96,$E6,$FB
	.dc.b	$5F,$C7,$C7,$F8,$6A,$31,$D7,$F2,$A7,$32,$9C,$DF,$D5,$9F,$F0,$84
	.dc.b	$73,$9E,$04,$9F,$1F,$9F,$E6,$23,$83,$CF,$E8,$FA,$88,$E0,$F9,$DF
	.dc.b	$68,$F9,$2F,$96,$D8,$CC,$60,$CF,$C8,$48,$E2,$DC,$3E,$AF,$DA,$3E
	.dc.b	$3F,$EA,$B6,$BE,$E3,$67,$73,$AF,$90,$95,$A6,$CF,$F8,$93,$F4,$FF
	.dc.b	$F1,$FB,$FE,$44,$6A,$69,$8F,$A3,$FE,$3F,$06,$93,$CC,$FF,$11,$2F
	.dc.b	$B4,$A4,$C3,$FB,$C4,$65,$FF,$19,$7E,$7C,$F8,$FF,$F2,$43,$63,$31
	.dc.b	$83,$FC,$5D,$F6,$BF,$E4,$9C,$4D,$23,$FE,$33,$3F,$CB,$FF,$1B,$FE
	.dc.b	$07,$69,$5C,$27,$FC,$19,$F6,$8F,$61,$F6,$B8,$6D,$7F,$93,$27,$53
	.dc.b	$93,$8F,$DA,$BE,$BA,$92,$B8,$46,$5F,$EB,$3F,$09,$FC,$E1,$FF,$3A
	.dc.b	$CF,$F8,$11,$7B,$49,$D4,$75,$D7,$FC,$F7,$39,$F5,$3F,$C2,$77,$FE
	.dc.b	$8F,$FB,$56,$38,$BD,$FE,$4A,$F9,$8C,$65,$DA,$30,$F8,$A7,$73,$AC
	.dc.b	$BB,$4E,$4F,$F4,$5F,$C3,$FF,$77,$7F,$C6,$8D,$7F,$CE,$2D,$7F,$D2
	.dc.b	$4D,$7F,$C5,$3F,$6C,$F9,$FF,$1C,$23,$0F,$73,$11,$A9,$ED,$A7,$C7
	.dc.b	$F0,$BF,$83,$7C,$37,$B1,$B7,$E7,$6B,$F2,$4F,$F5,$1E,$1D,$9C,$79
	.dc.b	$4F,$B6,$FF,$50,$C7,$BE,$31,$E6,$FC,$86,$C6,$67,$FC,$FC,$EF,$E2
	.dc.b	$BE,$BE,$5F,$6E,$BF,$CA,$FC,$7C,$67,$3C,$1F,$EB,$49,$E7,$7E,$2F
	.dc.b	$3F,$FD,$42,$F8,$5A,$E7,$E1,$BC,$51,$F9,$78,$F3,$61,$FD,$49,$BE
	.dc.b	$A7,$9E,$31,$1C,$4F,$67,$9A,$FB,$6F,$F3,$66,$28,$C4,$FF,$36,$9C
	.dc.b	$C6,$23,$FA,$4F,$D9,$FF,$B3,$75,$F7,$37,$4F,$63,$67,$FC,$69,$5C
	.dc.b	$6B,$EC,$7C,$C3,$ED,$7E,$C3,$6B,$EC,$35,$FB,$1C,$6F,$B6,$FF,$76
	.dc.b	$5F,$E0,$7D,$17,$C7,$7F,$A1,$6B,$FA,$43,$63,$37,$FD,$0E,$3D,$F9
	.dc.b	$FF,$91,$E3,$EF,$35,$FC,$1B,$6F,$ED,$67,$FE,$17,$9F,$C2,$7D,$B3
	.dc.b	$CC,$FC,$23,$63,$31,$8F,$37,$F1,$1E,$C7,$8A,$7F,$B8,$B5,$FE,$DA
	.dc.b	$D7,$F0,$8D,$7F,$D1,$D8,$3E,$73,$F7,$3E,$DF,$F0,$F6,$C6,$67,$F6
	.dc.b	$D3,$FC,$4F,$83,$7C,$5F,$44,$FF,$2E,$16,$7F,$D3,$4D,$9F,$F4,$AF
	.dc.b	$CB,$76,$FC,$73,$F3,$B5,$FE,$A3,$6B,$FE,$68,$6B,$FE,$67,$6B,$FE
	.dc.b	$A5,$FB,$7E,$84,$E6,$F4,$AA,$F5,$FD,$BB,$EA,$29,$2F,$D7,$0F,$B8
	.dc.b	$F8,$3F,$0F,$FA,$49,$FF,$21,$F7,$FF,$B1,$74,$4E,$67,$FC,$F4,$D7
	.dc.b	$FD,$71,$F7,$1F,$E7,$07,$E3,$B5,$FE,$0D,$28,$C7,$07,$F3,$58,$C7
	.dc.b	$D6,$7D,$2F,$5B,$FC,$01,$7F,$41,$AF,$26,$D7,$15,$E3,$FD,$AF,$87
	.dc.b	$B5,$FE,$F7,$6B,$FE,$D1,$6B,$FE,$96,$6B,$FE,$95,$F8,$EF,$CB,$72
	.dc.b	$5F,$0B,$F0,$B1,$89,$FF,$6B,$3F,$B3,$CD,$F6,$CF,$AC,$FC,$5B,$6B
	.dc.b	$F1,$2D,$7E,$1A,$FB,$7F,$C8,$FD,$BB,$E8,$7E,$83,$6B,$FD,$CA,$7F
	.dc.b	$AA,$5C,$3F,$EE,$B6,$BF,$EB,$0F,$B7,$7D,$77,$E7,$1B,$1F,$9A,$6B
	.dc.b	$F5,$17,$D7,$FE,$8B,$ED,$DF,$84,$D2,$57,$E8,$A1,$F3,$3F,$A8,$9F
	.dc.b	$ED,$B3,$E6,$3F,$DF,$4E,$1F,$F8,$00,$D7,$FD,$8A,$D7,$FD,$89,$F6
	.dc.b	$FF,$D2,$79,$BF,$08,$DE,$3C,$FA,$FF,$B6,$8F,$95,$3F,$DF,$69,$F6
	.dc.b	$FF,$C9,$5A,$7F,$62,$70,$7D,$95,$F4,$6F,$B7,$FF,$0E,$36,$3F,$DF
	.dc.b	$F1,$83,$D8,$BE,$F3,$FC,$4F,$DF,$FE,$00,$6C,$7F,$7B,$6B,$FD,$94
	.dc.b	$EF,$3F,$C3,$BE,$E1,$F1,$D4,$93,$FC,$3A,$3F,$3B,$AF,$FB,$9F,$F1
	.dc.b	$17,$C9,$7F,$16,$6D,$7F,$14,$6B,$FA,$EA,$EA,$FF,$18,$F0,$7F,$BE
	.dc.b	$DF,$FB,$C5,$B0,$EC,$46,$1B,$F7,$3F,$C9,$3F,$12,$CD,$7F,$22,$6B
	.dc.b	$FB,$C6,$F0,$FE,$EE,$FA,$CB,$EE,$BC,$6F,$C3,$6C,$FD,$96,$E2,$99
	.dc.b	$1F,$E9,$F7,$FC,$30,$FC,$47,$3F,$29,$FE,$1E,$8C,$47,$7D,$B1,$E2
	.dc.b	$48,$E2,$EE,$BF,$9F,$FD,$73,$EE,$69,$2B,$F9,$FC,$7F,$9A,$DF,$E8
	.dc.b	$3F,$89,$B6,$FE,$8E,$D8,$CC,$71,$7F,$8F,$FB,$5F,$99,$7E,$73,$78
	.dc.b	$3F,$B6,$F5,$7F,$50,$8F,$CD,$ED,$F0,$65,$F1,$CF,$53,$FA,$AB,$70
	.dc.b	$FE,$A6,$D7,$BE,$1F,$E4,$0F,$BA,$FE,$79,$1F,$E3,$F6,$BF,$93,$6F
	.dc.b	$17,$EA,$5F,$E8,$D1,$BF,$67,$FC,$FD,$1F,$DC,$63,$F7,$8D,$7F,$89
	.dc.b	$B5,$FB,$07,$EB,$3E,$2B,$FC,$A3,$DB,$FE,$62,$8F,$D1,$36,$FC,$17
	.dc.b	$FE,$63,$F2,$9F,$37,$49,$7F,$E6,$5F,$8B,$FD,$3C,$FF,$76,$E9,$7D
	.dc.b	$E5,$25,$6E,$79,$BC,$7E,$E1,$F7,$9F,$E7,$16,$C6,$63,$FC,$E9,$F3
	.dc.b	$DE,$FD,$3F,$D1,$1E,$1F,$FA,$93,$EA,$FF,$7D,$F3,$1F,$D6,$39,$FF
	.dc.b	$2D,$39,$C7,$91,$F3,$FC,$EF,$A4,$47,$E9,$7C,$15,$7F,$D4,$DE,$67
	.dc.b	$C8,$E6,$FE,$63,$1F,$EA,$E8,$C7,$9D,$F3,$3E,$63,$F9,$7C,$FE,$96
	.dc.b	$BF,$9C,$C6,$21,$1C,$9E,$A6,$E7,$95,$74,$8E,$C7,$3C,$32,$FF,$A9
	.dc.b	$B6,$EB,$3F,$EA,$46,$C6,$67,$7C,$89,$F0,$63,$E6,$8C,$4F,$4C,$DF
	.dc.b	$83,$7F,$90,$EB,$D7,$18,$F5,$F9,$E1,$DB,$0F,$AF,$F5,$1D,$B8,$FF
	.dc.b	$15,$87,$56,$E9,$FE,$B8,$53,$BE,$B9,$AF,$BB,$0F,$C0,$C7,$39,$F7
	.dc.b	$CF,$C8,$F9,$BF,$EB,$38,$FC,$BE,$0F,$F7,$56,$1F,$1A,$B4,$D8,$FE
	.dc.b	$11,$16,$B8,$E5,$F6,$DB,$1F,$5D,$AF,$D6,$F1,$7C,$35,$DA,$3E,$B9
	.dc.b	$D8,$AE,$70,$F2,$4F,$0A,$E7,$39,$BF,$06,$1E,$5A,$D3,$E5,$26,$1D
	.dc.b	$23,$2F,$6B,$E6,$74,$8C,$E1,$F0,$E3,$35,$D6,$B9,$CE,$62,$DF,$E6
	.dc.b	$EF,$8B,$E4,$AE,$D3,$C5,$A4,$AE,$4F,$94,$E6,$7C,$98,$AF,$C9,$2E
	.dc.b	$1F,$05,$D1,$C4,$E9,$5E,$1A,$EA,$7C,$A8,$E3,$87,$B1,$E1,$1A,$9E
	.dc.b	$04,$8F,$D8,$61,$EC,$7B,$CE,$DC,$9D,$30,$F2,$CE,$E3,$EE,$4E,$DF
	.dc.b	$51,$E6,$39,$D7,$77,$31,$BC,$3D,$D1,$CA,$7E,$73,$DC,$E4,$70,$79
	.dc.b	$47,$02,$57,$8D,$C6,$1F,$7E,$B9,$5F,$7A,$FC,$2C,$7F,$3E,$9F,$F1
	.dc.b	$B3,$F9,$53,$18,$7F,$89,$2D,$F2,$BF,$EB,$F9,$FE,$38,$D7,$F8,$C3
	.dc.b	$5F,$E2,$ED,$8E,$3B,$7E,$B6,$3F,$D9,$58,$FE,$CF,$DB,$72,$E6,$A6
	.dc.b	$1F,$A6,$BD,$E9,$FC,$AE,$F7,$39,$AD,$E1,$FB,$DB,$DC,$FF,$2B,$AD
	.dc.b	$CE,$6B,$77,$EA,$7F,$80,$4E,$A3,$15,$D4,$95,$E7,$C3,$8C,$7C,$0B
	.dc.b	$AF,$F3,$6B,$E4,$E4,$E3,$7D,$2F,$94,$60,$D3,$E5,$7F,$82,$CF,$27
	.dc.b	$E5,$D7,$6C,$7E,$5C,$6B,$0C,$63,$F0,$23,$66,$4E,$C6,$30,$EC,$78
	.dc.b	$8B,$8F,$94,$91,$F2,$6B,$AE,$1E,$1C,$7C,$98,$F8,$B0,$E3,$7E,$AC
	.dc.b	$3E,$B6,$1C,$23,$C3,$4B,$E3,$8F,$9E,$39,$5F,$82,$BE,$55,$F1,$C3
	.dc.b	$AC,$60,$E7,$7C,$4E,$58,$74,$C7,$86,$1E,$2A,$DC,$7A,$70,$CC,$79
	.dc.b	$CF,$71,$A3,$CF,$5C,$DC,$61,$F1,$EF,$CD,$8F,$A2,$79,$1E,$59,$F8
	.dc.b	$79,$3E,$F3,$FC,$C1,$8F,$F5,$76,$D7,$F7,$A9,$7D,$EB,$83,$F1,$31
	.dc.b	$C6,$6C,$E6,$F9,$46,$2B,$77,$D3,$0F,$EE,$DD,$6F,$A0,$FC,$9F,$2F
	.dc.b	$F7,$6F,$89,$EB,$39,$CB,$FC,$0F,$39,$7F,$B6,$46,$72,$F3,$37,$0E
	.dc.b	$BF,$4D,$CB,$1F,$CB,$7C,$3B,$2A,$63,$C8,$B9,$7F,$6F,$9F,$F3,$03
	.dc.b	$67,$4D,$35,$3F,$BE,$F0,$3F,$D8,$55,$C3,$3F,$F5,$82,$F9,$8D,$5F
	.dc.b	$F9,$5A,$35,$5F,$E3,$EF,$C1,$B5,$3F,$EB,$96,$DA,$7F,$85,$F1,$3E
	.dc.b	$FB,$EA,$C7,$1C,$7A,$D2,$B4,$E1,$B8,$7F,$94,$FD,$AC,$61,$CA,$3F
	.dc.b	$D0,$1F,$70,$FA,$FE,$7E,$7A,$31,$3F,$E6,$3F,$BE,$7A,$DA,$9C,$E7
	.dc.b	$E7,$AF,$F5,$56,$6B,$DC,$71,$BF,$27,$DF,$7E,$1E,$F8,$1E,$9A,$DB
	.dc.b	$FE,$78,$FA,$7F,$6B,$FE,$E8,$6F,$0B,$EF,$B3,$E9,$27,$31,$FE,$F5
	.dc.b	$AF,$A0,$D7,$FD,$7D,$5C,$72,$FF,$76,$B6,$3F,$D1,$66,$BE,$FB,$B9
	.dc.b	$E0,$FB,$0B,$5F,$EA,$2F,$9F,$F8,$FE,$B9,$F9,$3E,$C3,$E2,$B3,$18
	.dc.b	$65,$7F,$B3,$FA,$3F,$A3,$F2,$2D,$F7,$DA,$D3,$E2,$CB,$DC,$FF,$27
	.dc.b	$E1,$FE,$C4,$F3,$7E,$BF,$BA,$FB,$EF,$B3,$5A,$C3,$8C,$E6,$FF,$D9
	.dc.b	$B5,$C3,$EA,$78,$BF,$3B,$04,$53,$1E,$75,$AF,$41,$FE,$B0,$9F,$EE
	.dc.b	$2E,$30,$FF,$75,$46,$1F,$BA,$3E,$91,$F0,$FD,$2B,$84,$E6,$3E,$FC
	.dc.b	$EA,$72,$FF,$9F,$DB,$1F,$96,$F8,$57,$83,$4F,$CA,$70,$5A,$F2,$6F
	.dc.b	$F9,$B8,$CC,$63,$BF,$F1,$FC,$5F,$49,$FA,$51,$C3,$E1,$FF,$09,$3F
	.dc.b	$59,$B1,$FE,$E7,$F0,$7F,$C0,$AE,$FA,$FF,$53,$18,$DB,$EC,$B6,$FE
	.dc.b	$FA,$FB,$ED,$DF,$EE,$BC,$C7,$8D,$F0,$70,$FD,$2C,$63,$83,$F3,$FE
	.dc.b	$56,$EB,$59,$7F,$84,$2D,$E0,$AB,$F4,$11,$86,$C6,$71,$FC,$FC,$8D
	.dc.b	$38,$79,$F0,$FE,$B8,$FD,$7F,$D6,$59,$F9,$BF,$7D,$F4,$9B,$3F,$C2
	.dc.b	$A3,$B5,$7E,$69,$D3,$F9,$CF,$8E,$7E,$03,$FB,$5E,$54,$63,$0F,$2C
	.dc.b	$62,$75,$39,$AF,$DB,$CB,$EF,$7F,$8E,$42,$30,$FF,$1C,$87,$B1,$63
	.dc.b	$F8,$FF,$DE,$FF,$09,$C7,$F0,$ED,$8F,$D8,$BF,$D3,$F7,$FD,$BB,$97
	.dc.b	$F7,$3F,$7A,$DB,$D4,$DB,$F4,$B6,$FC,$ED,$7E,$81,$E7,$F6,$5F,$0F
	.dc.b	$FC,$C9,$B5,$F9,$A5,$F0,$B9,$9F,$E6,$FF,$31,$FD,$76,$BF,$3E,$DB
	.dc.b	$FE,$12,$BE,$AB,$7E,$B7,$F3,$EF,$7B,$F1,$D5,$E2,$E1,$FF,$11,$74
	.dc.b	$FD,$39,$CC,$FE,$16,$73,$8F,$D4,$F9,$A7,$D5,$7F,$48,$6C,$7E,$C3
	.dc.b	$EF,$3F,$40,$F1,$8C,$DF,$F1,$CF,$BC,$7D,$27,$F5,$06,$D7,$EE,$1F
	.dc.b	$E9,$BC,$9F,$C9,$9C,$D7,$F2,$4F,$A0,$7C,$7F,$E2,$5B,$7F,$88,$6C
	.dc.b	$FE,$19,$AF,$BF,$EF,$3F,$27,$87,$F3,$26,$D7,$BD,$FC,$CC,$67,$BB
	.dc.b	$FB,$2F,$DE,$7E,$E6,$BF,$86,$B6,$FF,$9B,$BF,$E4,$B8,$CF,$89,$FD
	.dc.b	$B3,$E9,$FF,$A3,$47,$E7,$1B,$5F,$9A,$6B,$F4,$DA,$FD,$18,$F9,$38
	.dc.b	$7F,$48,$69,$F5,$7E,$F3,$E7,$CF,$E8,$63,$2F,$E8,$FE,$B9,$F8,$47
	.dc.b	$F7,$E6,$CF,$F2,$65,$F6,$39,$9F,$F0,$6F,$82,$FA,$CF,$EC,$D8,$7E
	.dc.b	$62,$31,$E6,$7F,$6D,$6E,$3F,$C3,$1B,$3F,$DB,$17,$DE,$FC,$E8,$FD
	.dc.b	$9B,$63,$3E,$57,$82,$31,$F6,$1F,$CD,$2B,$FC,$0F,$48,$C6,$DF,$EB
	.dc.b	$08,$7D,$83,$EF,$7D,$35,$AC,$BF,$8F,$6E,$FB,$4F,$E0,$71,$FC,$8A
	.dc.b	$31,$87,$F8,$1D,$B1,$99,$DE,$5F,$E0,$69,$7D,$9B,$E7,$BF,$C6,$71
	.dc.b	$86,$C7,$F1,$48,$C5,$FF,$67,$C3,$D5,$F1,$3E,$63,$E7,$E9,$FD,$7F
	.dc.b	$1F,$7E,$7F,$D5,$AF,$F9,$B5,$FB,$BE,$63,$E8,$5E,$EF,$EA,$35,$DB
	.dc.b	$5F,$E4,$3F,$85,$7F,$7E,$BF,$13,$63,$E6,$FD,$F3,$E9,$3F,$03,$39
	.dc.b	$E1,$F9,$D3,$9B,$F5,$36,$F8,$FD,$2F,$E2,$23,$15,$F5,$7D,$FF,$F0
	.dc.b	$95,$7D,$A6,$C7,$D9,$FB,$E7,$D0,$FE,$42,$73,$8F,$C0,$9C,$BF,$25
	.dc.b	$B7,$F8,$45,$F4,$9F,$98,$73,$87,$FB,$D3,$8B,$F0,$15,$F5,$B9,$7F
	.dc.b	$A3,$AF,$BE,$FF,$3A,$E3,$F8,$69,$CC,$62,$7F,$B3,$35,$79,$1F,$D1
	.dc.b	$6B,$F9,$DC,$E6,$3F,$D2,$D3,$FE,$4C,$6B,$FD,$01,$AF,$F7,$16,$FC
	.dc.b	$9B,$F7,$9F,$C1,$65,$FE,$82,$6E,$1F,$E6,$EF,$BF,$7D,$07,$F9,$FE
	.dc.b	$73,$8F,$FA,$0E,$73,$5F,$E3,$5C,$11,$89,$FF,$1A,$FD,$37,$F6,$89
	.dc.b	$CE,$1F,$E7,$09,$FF,$02,$7C,$E7,$F8,$0D,$A5,$70,$9F,$F0,$47,$C9
	.dc.b	$7F,$63,$FB,$FB,$EE,$9A,$4A,$FF,$08,$CB,$EF,$BF,$D4,$11,$86,$BF
	.dc.b	$E7,$48,$D6,$DD,$A7,$33,$FD,$AB,$E3,$7F,$A1,$BF,$E4,$C6,$BF,$E5
	.dc.b	$26,$CF,$F8,$9D,$73,$7F,$72,$8C,$7A,$37,$83,$A7,$E0,$37,$0F,$F1
	.dc.b	$36,$08,$C4,$23,$AC,$FF,$7C,$F8,$07,$D4,$FC,$96,$D7,$F8,$B5,$AF
	.dc.b	$F8,$C9,$AF,$FA,$49,$7D,$5F,$66,$C7,$F9,$A1,$AF,$F9,$B9,$E7,$97
	.dc.b	$F9,$0D,$CB,$7E,$AF,$91,$CE,$BD,$9D,$4F,$DF,$FF,$CD,$78,$7F,$9E
	.dc.b	$E3,$1C,$5A,$9F,$F0,$F4,$E5,$DB,$FE,$BA,$9D,$B9,$E0,$E0,$71,$AE
	.dc.b	$FF,$A8,$7F,$AE,$67,$B2,$F9,$CC,$9A,$F3,$EC,$EB,$E7,$38,$98,$7C
	.dc.b	$1F,$7F,$FE,$5E,$C8,$D2,$70,$FF,$96,$E1,$F4,$9F,$E1,$8A,$F7,$35
	.dc.b	$F5,$37,$83,$F8,$45,$FF,$9B,$5F,$E1,$51,$FE,$BF,$9E,$3F,$3D,$F2
	.dc.b	$A3,$11,$F0,$FC,$2F,$F7,$63,$DB,$E1,$7C,$52,$69,$23,$FA,$94,$7F
	.dc.b	$8D,$DF,$A5,$F6,$0F,$27,$EB,$B6,$3D,$3F,$0F,$F6,$BF,$45,$7D,$85
	.dc.b	$24,$FD,$B8,$FE,$5A,$FD,$DF,$49,$C3,$F1,$5B,$1F,$D8,$DA,$FF,$5F
	.dc.b	$6B,$FE,$58,$E0,$7E,$87,$F1,$78,$BF,$B6,$F0,$7C,$96,$FC,$3F,$E3
	.dc.b	$A7,$56,$BF,$F6,$FF,$CA,$FE,$4B,$E8,$7F,$D4,$33,$F9,$16,$BF,$8D
	.dc.b	$5F,$21,$F9,$8C,$FE,$14,$F9,$78,$7D,$44,$7F,$D3,$BE,$48,$8F,$F2
	.dc.b	$3C,$7F,$DD,$ED,$C3,$F7,$D3,$B2,$F9,$1F,$E7,$09,$FE,$54,$D5,$CE
	.dc.b	$DB,$FD,$55,$5F,$C8,$FE,$43,$8F,$CA,$74,$FC,$1F,$FD,$C3,$87,$91
	.dc.b	$AF,$FA,$08,$F3,$38,$7F,$D0,$B4,$8C,$47,$6F,$C1,$9E,$D7,$A1,$FF
	.dc.b	$10,$47,$F3,$F6,$BE,$6C,$78,$4F,$BB,$D2,$FA,$D3,$F1,$5B,$87,$F5
	.dc.b	$CE,$4F,$87,$F1,$DF,$62,$3C,$5E,$D7,$FB,$15,$B7,$EB,$8C,$67,$FE
	.dc.b	$09,$8F,$A3,$F7,$6F,$91,$F9,$AF,$FB,$86,$3F,$AA,$35,$F3,$1F,$5D
	.dc.b	$E0,$FF,$59,$FB,$AF,$C1,$F0,$FC,$F8,$FB,$91,$8D,$BE,$8B,$63,$3C
	.dc.b	$FF,$73,$E6,$7F,$9F,$C7,$E1,$1A,$FD,$1D,$11,$88,$FA,$2D,$7E,$EF
	.dc.b	$CE,$FF,$A6,$1F,$ED,$8B,$E9,$FF,$12,$D8,$CF,$CA,$7E,$33,$EC,$FF
	.dc.b	$81,$C7,$BB,$04,$63,$4F,$64,$63,$E7,$F8,$7D,$0F,$A7,$ED,$3E,$86
	.dc.b	$86,$93,$1F,$CF,$FD,$37,$E2,$F1,$E7,$F8,$DB,$EE,$7F,$3A,$DA,$FF
	.dc.b	$02,$56,$F8,$BF,$DA,$7F,$87,$3E,$47,$FA,$EB,$68,$7D,$60,$E5,$F6
	.dc.b	$7E,$1C,$F8,$EF,$EC,$BF,$70,$FC,$13,$FC,$14,$DA,$FD,$01,$D1,$C3
	.dc.b	$86,$BF,$A5,$5F,$81,$F4,$24,$79,$36,$FC,$22,$9E,$71,$F3,$3B,$7F
	.dc.b	$C8,$B3,$FE,$23,$7F,$91,$BF,$4D,$B1,$BF,$8C,$FE,$07,$18,$D7,$F7
	.dc.b	$AF,$F2,$EA,$E5,$E8,$6B,$E3,$7F,$84,$F9,$7F,$DC,$AF,$F8,$1B,$FD
	.dc.b	$C9,$BE,$F6,$BE,$07,$6F,$0F,$80,$98,$FF,$8D,$EF,$83,$F4,$38,$9E
	.dc.b	$57,$D4,$FA,$49,$3F,$07,$2F,$B3,$38,$8F,$AD,$5F,$CA,$1D,$1C,$FC
	.dc.b	$2D,$70,$7C,$89,$7D,$3F,$E0,$09,$5C,$0F,$BF,$5F,$C9,$BC,$BF,$A7
	.dc.b	$1F,$E4,$D6,$BF,$9D,$6D,$FE,$71,$FA,$78,$FF,$26,$7E,$9C,$EF,$8B
	.dc.b	$A1,$D7,$93,$F9,$9F,$DA,$7E,$BE,$7F,$50,$DC,$3F,$4E,$DC,$3F,$98
	.dc.b	$BF,$AB,$6F,$8B,$C8,$95,$FE,$63,$9E,$0F,$E7,$3E,$A7,$FC,$72,$DF
	.dc.b	$88,$FF,$1E,$C6,$23,$FC,$8F,$97,$E7,$39,$7F,$36,$48,$FC,$D9,$F9
	.dc.b	$17,$F3,$FC,$DF,$A2,$24,$7C,$BE,$8F,$C6,$3F,$D5,$79,$1F,$80,$FE
	.dc.b	$BA,$3F,$CD,$9C,$3F,$8A,$FA,$FC,$7A,$7F,$6F,$C7,$FF,$02,$46,$27
	.dc.b	$F1,$F7,$6F,$A4,$DA,$DC,$7F,$9F,$2B,$A3,$4F,$15,$6F,$93,$FC,$AB
	.dc.b	$87,$76,$D7,$D9,$EB,$D6,$BB,$78,$6B,$31,$E9,$97,$F5,$69,$F2,$47
	.dc.b	$ED,$7A,$AE,$C6,$9A,$E8,$FE,$DD,$A7,$E6,$6B,$F7,$9A,$F0,$23,$E9
	.dc.b	$3A,$16,$FE,$5B,$63,$6F,$CC,$3F,$88,$5F,$C5,$F1,$3F,$A1,$DF,$CD
	.dc.b	$6E,$BF,$DF,$1B,$5F,$E0,$B7,$F9,$53,$CF,$3F,$E6,$9F,$17,$CF,$EE
	.dc.b	$A4,$FC,$2A,$E3,$4F,$B7,$FE,$BB,$3C,$5F,$EC,$58,$FD,$D6,$E3,$FE
	.dc.b	$0C,$77,$19,$FA,$3F,$B9,$1A,$AF,$F0,$F3,$6F,$B7,$A3,$F3,$BE,$47
	.dc.b	$FC,$43,$87,$CB,$DF,$FD,$93,$5F,$1E,$73,$97,$E4,$67,$FA,$AE,$39
	.dc.b	$A8,$FC,$93,$F9,$75,$70,$7F,$9D,$79,$DD,$E7,$78,$77,$AD,$57,$F3
	.dc.b	$5A,$D3,$CE,$FC,$6D,$9D,$C7,$8A,$31,$7D,$30,$F3,$56,$AF,$9E,$1D
	.dc.b	$EB,$57,$E4,$29,$E8,$6E,$1D,$67,$84,$F6,$23,$F1,$EB,$AE,$1E,$0B
	.dc.b	$E3,$18,$AF,$EB,$58,$7A,$0F,$23,$D9,$C4,$FE,$56,$78,$4F,$8C,$EF
	.dc.b	$5F,$46,$33,$87,$C4,$C3,$CA,$78,$EB,$A5,$F8,$9A,$72,$3E,$01,$4E
	.dc.b	$D1,$CA,$FB,$47,$08,$C3,$C1,$AF,$08,$F8,$0F,$D7,$BF,$2E,$1E,$C8
	.dc.b	$C1,$CE,$77,$5E,$78,$F1,$61,$EB,$9C,$9E,$28,$CD,$73,$AE,$07,$8A
	.dc.b	$B8,$4F,$CE,$C3,$E5,$3D,$CC,$1E,$23,$ED,$1F,$E0,$9A,$FF,$40,$B6
	.dc.b	$3F,$9B,$CF,$E4,$23,$FC,$1D,$F1,$CF,$9A,$CB,$63,$FC,$28,$D7,$71
	.dc.b	$A3,$F8,$84,$6F,$0F,$D7,$67,$FE,$57,$7F,$61,$19,$CB,$F5,$CD,$BF
	.dc.b	$F1,$13,$5F,$F0,$D5,$71,$9D,$DF,$62,$4F,$EA,$67,$4F,$28,$CD,$F8
	.dc.b	$1F,$E4,$93,$A9,$FF,$13,$5F,$7C,$0D,$31,$3B,$8C,$D7,$4C,$3C,$77
	.dc.b	$E1,$EF,$C3,$E1,$3F,$79,$87,$28,$CC,$78,$2A,$DF,$FA,$16,$6D,$78
	.dc.b	$67,$CF,$18,$77,$87,$8E,$75,$7D,$A7,$6D,$26,$3F,$05,$F3,$18,$C7
	.dc.b	$D3,$19,$3E,$BE,$1D,$E7,$47,$F9,$57,$0F,$05,$E6,$B7,$3A,$3D,$85
	.dc.b	$C7,$D8,$DC,$7C,$F3,$C0,$F0,$63,$E2,$24,$F7,$C7,$B4,$62,$3F,$07
	.dc.b	$18,$7E,$74,$F3,$9D,$5F,$D6,$9F,$16,$3E,$87,$C5,$8F,$C6,$BF,$0C
	.dc.b	$72,$BE,$35,$B9,$CC,$FB,$1E,$CE,$CF,$F7,$67,$EA,$BE,$E8,$FF,$0A
	.dc.b	$36,$BF,$DC,$6B,$7F,$CA,$E8,$8D,$5F,$29,$CC,$62,$75,$18,$C3,$8F
	.dc.b	$DE,$38,$B9,$FD,$E7,$FA,$53,$6F,$F0,$7C,$E5,$FF,$08,$C6,$72,$FE
	.dc.b	$ED,$39,$6D,$74,$F4,$3A,$BF,$A5,$F0,$DE,$D5,$31,$F0,$77,$2F,$F0
	.dc.b	$84,$E9,$B3,$A6,$9A,$9C,$61,$FD,$A7,$A1,$D9,$8E,$2F,$F1,$E7,$DE
	.dc.b	$7F,$AC,$70,$FF,$14,$AF,$BC,$F9,$2E,$70,$FF,$08,$39,$9D,$FA,$5A
	.dc.b	$37,$F7,$56,$3F,$C6,$7D,$DF,$0F,$1F,$F2,$53,$79,$FF,$D7,$BC,$CF
	.dc.b	$BC,$FF,$64,$5F,$0C,$FE,$CC,$61,$FF,$10,$5F,$15,$5C,$78,$FE,$FC
	.dc.b	$E6,$7F,$DC,$DF,$19,$FE,$F9,$F3,$3F,$D6,$FC,$88,$C6,$5F,$E6,$18
	.dc.b	$FF,$13,$3F,$E9,$17,$4D,$F0,$3F,$D7,$F7,$B6,$BF,$ED,$69,$E2,$FF
	.dc.b	$B3,$9A,$FF,$9B,$47,$95,$FE,$EB,$AF,$F4,$F3,$63,$FD,$BD,$F7,$1F
	.dc.b	$75,$FF,$70,$46,$71,$FF,$7B,$FE,$90,$F8,$6F,$4D,$7C,$28,$C6,$1F
	.dc.b	$EB,$0D,$90,$FA,$AD,$C6,$2E,$CE,$6B,$83,$6D,$87,$83,$EF,$37,$8F
	.dc.b	$2E,$67,$D7,$D3,$1F,$33,$72,$F1,$4A,$33,$18,$F9,$9B,$97,$FB,$F7
	.dc.b	$B1,$E4,$FB,$21,$18,$AF,$F6,$27,$0B,$E4,$B5,$39,$AF,$93,$F2,$9C
	.dc.b	$6B,$57,$C2,$31,$79,$E6,$D4,$59,$D1,$F2,$7A,$7C,$2D,$F0,$BE,$4C
	.dc.b	$63,$1F,$8B,$3F,$9A,$8F,$DF,$71,$23,$1E,$AF,$0D,$38,$B6,$3E,$B7
	.dc.b	$9D,$F8,$8A,$E1,$87,$E7,$23,$F2,$2F,$63,$1D,$5F,$6A,$33,$3C,$3A
	.dc.b	$3F,$D7,$7F,$2F,$9A,$D3,$70,$FC,$E7,$CC,$7C,$DE,$C3,$7E,$53,$0D
	.dc.b	$9F,$F6,$6C,$A7,$F3,$F9,$A3,$13,$A9,$CF,$1F,$AE,$3F,$D9,$8D,$AF
	.dc.b	$DB,$FA,$CF,$A8,$FA,$D9,$FE,$53,$E0,$FF,$C4,$58,$7E,$65,$D3,$7D
	.dc.b	$0D,$57,$E6,$BF,$0C,$71,$7E,$CB,$81,$18,$BF,$D8,$67,$FC,$1A,$30
	.dc.b	$FE,$CE,$77,$5C,$7E,$7D,$ED,$FE,$99,$B8,$7E,$AD,$7D,$56,$67,$F5
	.dc.b	$7F,$76,$FA,$1B,$86,$1B,$87,$EC,$F3,$FD,$F6,$7F,$EE,$FC,$FC,$FE
	.dc.b	$D7,$A6,$73,$3F,$C9,$1A,$FF,$1E,$F8,$BF,$D5,$5F,$EF,$5B,$1F,$85
	.dc.b	$B4,$3C,$1F,$DF,$BA,$E9,$FC,$37,$C7,$3D,$FF,$E0,$6B,$E0,$6C,$61
	.dc.b	$FE,$0F,$F2,$BF,$7F,$3F,$97,$92,$31,$9F,$CB,$6B,$ED,$6B,$FC,$FB
	.dc.b	$AF,$C1,$5F,$C5,$5B,$1E,$97,$C5,$19,$C3,$F8,$BF,$CA,$BE,$0E,$FD
	.dc.b	$B6,$1B,$5F,$C7,$F1,$FE,$E1,$8F,$84,$FC,$23,$87,$FA,$7B,$5F,$E9
	.dc.b	$3F,$2B,$F8,$E7,$E5,$37,$6F,$D6,$3E,$E6,$F4,$FF,$2A,$FB,$07,$D8
	.dc.b	$5F,$3F,$0D,$C9,$8F,$EE,$7E,$69,$EE,$FE,$C2,$7F,$2B,$18,$F1,$7F
	.dc.b	$A5,$B6,$33,$3B,$AF,$E9,$52,$FC,$C9,$F8,$0F,$F4,$28,$C3,$63,$F2
	.dc.b	$97,$E2,$D7,$F8,$6D,$6A,$30,$79,$97,$E7,$3F,$52,$76,$FD,$4C,$3D
	.dc.b	$4E,$6B,$83,$99,$EB,$8F,$D7,$9F,$4F,$0F,$B2,$31,$7E,$A6,$BF,$18
	.dc.b	$60,$F9,$EF,$A1,$7E,$56,$CE,$97,$4F,$C1,$73,$4F,$B1,$7C,$D7,$C5
	.dc.b	$FA,$27,$C6,$FC,$79,$CE,$5F,$0B,$2F,$0D,$6A,$31,$87,$3E,$BF,$AB
	.dc.b	$7F,$9C,$BF,$3B,$67,$55,$F7,$23,$CA,$F2,$6C,$FE,$0F,$CC,$7A,$7E
	.dc.b	$3D,$7C,$9D,$13,$0E,$17,$4F,$C5,$61,$F9,$9C,$7D,$D1,$E4,$24,$FE
	.dc.b	$33,$F5,$7F,$F5,$6B,$EA,$7D,$91,$F9,$37,$2D,$D7,$F1,$51,$FE,$8A
	.dc.b	$E2,$FC,$3B,$A6,$FC,$56,$BE,$3B,$F2,$1E,$7F,$F8,$5E,$7F,$1B,$FA
	.dc.b	$C7,$F6,$1C,$BE,$4C,$E5,$FC,$74,$67,$2F,$D2,$4F,$E9,$27,$7F,$15
	.dc.b	$D3,$9B,$AB,$FE,$03,$FA,$1F,$EF,$77,$FD,$8C,$3C,$5C,$3A,$C7,$FB
	.dc.b	$5C,$6A,$B9,$7B,$73,$0F,$C0,$4F,$2E,$3C,$3C,$9F,$EB,$F8,$7D,$69
	.dc.b	$FE,$97,$3B,$6C,$66,$3F,$A7,$F8,$8F,$88,$FF,$0D,$3F,$E6,$78,$F4
	.dc.b	$39,$AF,$0B,$9C,$BC,$EE,$67,$75,$CB,$E2,$79,$BF,$E2,$58,$FE,$1B
	.dc.b	$F1,$3E,$46,$95,$C2,$7D,$3E,$A7,$F4,$5F,$AB,$BE,$71,$A4,$AF,$ED
	.dc.b	$33,$FE,$39,$CB,$8F,$C0,$FF,$91,$27,$2D,$7F,$CB,$D3,$B7,$FB,$8F
	.dc.b	$27,$F6,$57,$E4,$B5,$CF,$A5,$E9,$8C,$47,$F9,$9F,$F0,$E7,$CD,$7B
	.dc.b	$5B,$1F,$49,$AF,$F7,$E6,$BF,$E6,$FF,$AC,$AF,$F0,$D6,$BF,$CD,$1A
	.dc.b	$9A,$3E,$2F,$E6,$36,$3F,$02,$D7,$FC,$1A,$D7,$FC,$FD,$F8,$EF,$F8
	.dc.b	$51,$FC,$3B,$5F,$F0,$9B,$5D,$FB,$1F,$8C,$EF,$7D,$6F,$F9,$9E,$7F
	.dc.b	$A2,$B5,$FE,$82,$D7,$F9,$44,$70,$DB,$97,$6F,$F9,$AA,$7F,$A6,$B5
	.dc.b	$FE,$92,$D7,$8B,$5F,$E8,$5F,$5B,$FE,$3F,$7F,$AA,$B5,$FE,$A2,$D7
	.dc.b	$C2,$D7,$FA,$57,$81,$F0,$DB,$A7,$B6,$31,$1F,$CF,$9B,$87,$FA,$CF
	.dc.b	$BB,$E1,$7C,$3F,$FB,$8A,$BF,$B2,$B5,$FE,$C2,$D7,$E0,$F8,$6F,$A4
	.dc.b	$F9,$4D,$8F,$E6,$51,$83,$AF,$7C,$8E,$AD,$7A,$F8,$1E,$22,$3E,$27
	.dc.b	$EB,$7D,$5B,$E8,$69,$27,$FB,$DC,$65,$FE,$FB,$E5,$FF,$BC,$1B,$F2
	.dc.b	$BF,$7E,$31,$1F,$DF,$1A,$FD,$CF,$39,$EC,$FE,$15,$B1,$FE,$08,$6B
	.dc.b	$FE,$04,$6B,$F8,$4F,$1F,$FC,$1E,$FF,$82,$9F,$C7,$18,$7F,$B9,$3F
	.dc.b	$91,$F9,$17,$E2,$94,$93,$FE,$16,$8C,$BE,$1F,$84,$F2,$11,$C1,$FE
	.dc.b	$46,$E3,$7C,$8F,$F2,$8F,$7F,$FB,$EB,$63,$33,$B8,$D7,$8E,$FA,$2A
	.dc.b	$49,$FF,$28,$C6,$5F,$F2,$9F,$BD,$FE,$45,$6F,$AF,$DE,$31,$1F,$E7
	.dc.b	$F6,$BF,$E5,$5F,$86,$F2,$7C,$07,$FA,$02,$3F,$CB,$AD,$7F,$CB,$91
	.dc.b	$E9,$F3,$FE,$44,$7F,$9E,$5A,$FF,$8F,$B4,$46,$21,$18,$E0,$F2,$54
	.dc.b	$9F,$4B,$5F,$F4,$A7,$D4,$FF,$AB,$A8,$D2,$7A,$3F,$2F,$EA,$7E,$9D
	.dc.b	$7F,$A6,$DA,$FF,$A3,$5A,$F9,$5A,$FF,$A2,$FC,$6F,$A9,$F0,$BF,$C4
	.dc.b	$E3,$D4,$D7,$47,$FA,$03,$EA,$7F,$D7,$0D,$F6,$7E,$E4,$62,$3F,$D5
	.dc.b	$8D,$7F,$D3,$DF,$27,$FE,$36,$7F,$D1,$2D,$7F,$D6,$4D,$77,$ED,$F9
	.dc.b	$3E,$06,$47,$FA,$43,$0F,$F2,$FB,$63,$F1,$9F,$52,$F9,$4A,$49,$FF
	.dc.b	$4B,$C6,$5F,$EF,$3F,$40,$FA,$7F,$ED,$0D,$8C,$C6,$0C,$FD,$2C,$8E
	.dc.b	$2D,$7F,$D9,$B3,$FE,$A9,$7A,$FC,$67,$42,$3F,$EA,$8F,$19,$F4,$9F
	.dc.b	$A3,$6C,$66,$30,$7E,$9F,$E5,$BF,$8F,$47,$E3,$5A,$FE,$31,$BC,$C6
	.dc.b	$F3,$FF,$2E,$8F,$F6,$DF,$A1,$FB,$18,$C1,$97,$EB,$57,$E6,$5A,$FE
	.dc.b	$61,$B3,$FD,$8F,$93,$FC,$1D,$F4,$8F,$9F,$A4,$BF,$F1,$2C,$65,$BF
	.dc.b	$4D,$C0,$8E,$0D,$FD,$5F,$42,$38,$3F,$D9,$FE,$0F,$CD,$7F,$DF,$0D
	.dc.b	$7F,$DD,$4D,$7F,$64,$D7,$F6,$3F,$3A,$F9,$5F,$E2,$4D,$8F,$F6,$DB
	.dc.b	$5F,$F4,$91,$C2,$3F,$8C,$7D,$3B,$E3,$69,$27,$FD,$C7,$19,$7F,$DC
	.dc.b	$BF,$A4,$FC,$17,$9E,$C8,$C4,$7F,$C0,$16,$BF,$C2,$5A,$FF,$B2,$39
	.dc.b	$7F,$DE,$CF,$E6,$3E,$5B,$F2,$E4,$7F,$37,$C9,$F9,$A2,$38,$21,$FE
	.dc.b	$41,$E0,$FF,$57,$57,$F5,$0F,$23,$CC,$D8,$CD,$7F,$AC,$E7,$ED,$62
	.dc.b	$E4,$7C,$DF,$FA,$DF,$C4,$FF,$5C,$4E,$63,$19,$7F,$23,$AD,$BF,$EF
	.dc.b	$49,$D3,$79,$9F,$35,$DF,$3F,$C9,$36,$7F,$24,$DA,$E3,$1E,$B3,$F9
	.dc.b	$6F,$1F,$A5,$BA,$6E,$3F,$5B,$1B,$94,$6D,$B3,$E8,$9E,$27,$9E,$78
	.dc.b	$E7,$F8,$E9,$D4,$B5,$FD,$83,$FE,$72,$C7,$E6,$F4,$7A,$63,$4D,$9F
	.dc.b	$F2,$46,$DF,$42,$39,$C7,$29,$FF,$4B,$9C,$01,$87,$58,$7E,$59,$E1
	.dc.b	$5F,$6B,$6E,$C7,$7B,$F9,$1C,$6F,$9E,$BD,$F3,$2F,$41,$FD,$5A,$74
	.dc.b	$E3,$6F,$D4,$4E,$E3,$E5,$4F,$F2,$B6,$D7,$D5,$F8,$F7,$D3,$D2,$4F
	.dc.b	$38,$7C,$8E,$23,$EA,$47,$07,$53,$CE,$7E,$47,$C9,$7F,$6A,$9F,$C0
	.dc.b	$1F,$45,$C6,$BF,$9A,$6F,$07,$E7,$BC,$7E,$3A,$FB,$AF,$83,$78,$17
	.dc.b	$C2,$7F,$61,$5C,$1E,$F3,$AA,$E1,$5B,$9F,$C1,$56,$F0,$F1,$56,$8E
	.dc.b	$D1,$9C,$3F,$63,$7C,$70,$E9,$87,$6A,$D3,$FC,$06,$74,$6E,$33,$1E
	.dc.b	$A3,$CB,$19,$7C,$06,$2E,$E1,$FB,$EA,$E3,$7E,$08,$F3,$D7,$84,$DE
	.dc.b	$1E,$6B,$B1,$F1,$0A,$78,$67,$89,$D5,$CD,$75,$AE,$51,$D4,$F2,$12
	.dc.b	$78,$C7,$D7,$8E,$38,$7C,$0B,$F9,$31,$87,$AE,$1E,$A9,$D4,$F5,$BE
	.dc.b	$6F,$C9,$AD,$BC,$4F,$05,$F7,$69,$B2,$9F,$70,$E1,$3D,$1C,$1A,$C3
	.dc.b	$DC,$F1,$8F,$84,$E6,$FC,$B1,$83,$94,$F0,$9F,$95,$5E,$39,$CD,$70
	.dc.b	$8C,$1F,$2A,$7F,$80,$47,$E8,$DB,$7F,$E1,$C8,$C4,$7F,$3E,$F8,$AF
	.dc.b	$E5,$AF,$F8,$59,$AF,$F0,$96,$BB,$E5,$FF,$12,$C7,$0C,$3F,$A2,$B7
	.dc.b	$0F,$E9,$39,$B8,$FF,$55,$39,$C3,$FA,$C3,$5F,$E4,$CD,$CF,$F8,$0C
	.dc.b	$F0,$C6,$57,$EB,$AB,$57,$D6,$31,$3A,$C7,$C6,$FE,$E6,$75,$18,$9D
	.dc.b	$91,$FE,$BB,$1A,$C7,$DD,$39,$6C,$EB,$0E,$D1,$82,$4F,$F7,$3A,$E4
	.dc.b	$FB,$AF,$6D,$3F,$83,$C6,$9F,$83,$1A,$69,$AA,$C6,$2C,$7E,$31,$F0
	.dc.b	$A3,$2E,$2B,$E0,$46,$1E,$37,$DE,$B9,$46,$6F,$B1,$EF,$3C,$D8,$FC
	.dc.b	$73,$E0,$39,$3C,$AE,$71,$FB,$F8,$73,$8C,$4F,$C3,$8C,$4F,$C6,$8F
	.dc.b	$EC,$98,$78,$F1,$F4,$47,$5C,$7F,$02,$D8,$CE,$1E,$23,$AC,$78,$AF
	.dc.b	$A5,$70,$9C,$BE,$E9,$E9,$7D,$F1,$F1,$C7,$94,$EC,$FC,$6C,$73,$1D
	.dc.b	$4D,$5F,$B3,$37,$D4,$77,$CB,$FD,$78,$DA,$FF,$5D,$BF,$D7,$AF,$FB
	.dc.b	$33,$9A,$E3,$8F,$E1,$9B,$3A,$DB,$15,$BF,$A8,$FE,$FB,$3F,$DB,$DB
	.dc.b	$87,$F8,$5F,$EA,$3F,$D3,$93,$98,$E9,$18,$7F,$C3,$93,$C3,$3F,$F1
	.dc.b	$BC,$E5,$FF,$1C,$C6,$73,$FF,$05,$D7,$F9,$6F,$E1,$3F,$11,$87,$E5
	.dc.b	$FB,$D1,$88,$FE,$D1,$18,$7C,$B3,$FE,$14,$8C,$3F,$E1,$86,$C7,$F9
	.dc.b	$A9,$B1,$E0,$77,$87,$F9,$95,$B7,$9F,$47,$C5,$87,$C9,$F0,$5F,$51
	.dc.b	$FC,$0F,$2F,$F2,$7C,$61,$FF,$26,$37,$2F,$E1,$07,$FA,$02,$78,$3F
	.dc.b	$EF,$58,$7C,$AF,$59,$EC,$DA,$FF,$10,$7E,$0E,$F2,$7E,$D3,$63,$FC
	.dc.b	$49,$E0,$FC,$BC,$7F,$DF,$B3,$97,$FC,$9D,$1F,$6A,$7F,$CC,$AD,$8F
	.dc.b	$F5,$71,$E4,$3F,$D5,$C5,$3F,$D1,$DF,$0F,$F4,$FE,$17,$EA,$36,$BE
	.dc.b	$9B,$FE,$46,$9F,$A4,$D7,$8C,$7F,$A6,$9C,$B5,$FF,$4D,$76,$6F,$80
	.dc.b	$D2,$4F,$1E,$67,$EF,$2F,$92,$F0,$46,$1B,$19,$9E,$0D,$C3,$FD,$43
	.dc.b	$F5,$0F,$81,$E2,$DA,$EC,$D9,$DC,$F6,$F2,$FF,$1F,$1F,$E6,$2C,$51
	.dc.b	$88,$46,$1E,$07,$9B,$8D,$F5,$14,$98,$FF,$B0,$1F,$91,$E1,$BE,$B6
	.dc.b	$92,$7E,$44,$65,$BE,$27,$02,$38,$3E,$77,$CE,$3E,$2F,$EA,$36,$33
	.dc.b	$18,$3E,$5F,$33,$EB,$69,$27,$E9,$46,$5F,$A9,$F5,$5F,$C0,$75,$FF
	.dc.b	$58,$B5,$F8,$CD,$7E,$2B,$76,$FF,$5D,$69,$F6,$BC,$FF,$65,$7F,$B5
	.dc.b	$9A,$FF,$AF,$1A,$FD,$1F,$31,$D1,$F8,$26,$E3,$F8,$76,$CF,$FB,$81
	.dc.b	$AF,$F0,$3F,$AA,$FC,$81,$27,$F2,$9D,$28,$FC,$A7,$D1,$C9,$CB,$6B
	.dc.b	$F3,$94,$FA,$AF,$CD,$12,$B9,$BF,$96,$FA,$AF,$F7,$C4,$7E,$6D,$B5
	.dc.b	$FE,$F4,$6B,$FE,$DA,$6B,$F8,$FF,$A9,$F9,$04,$F4,$7F,$42,$FC,$BF
	.dc.b	$9E,$FF,$65,$DF,$E9,$DB,$5F,$F0,$01,$AF,$FB,$99,$AF,$E6,$FE,$AB
	.dc.b	$FD,$69,$3F,$AF,$6C,$7F,$36,$3F,$3A,$E1,$FF,$77,$B5,$F3,$77,$F9
	.dc.b	$09,$3F,$B9,$BF,$D9,$4F,$FB,$DB,$E0,$BF,$47,$3F,$BD,$6C,$77,$E3
	.dc.b	$46,$23,$87,$5F,$F0,$0F,$83,$7C,$05,$26,$1F,$C1,$E3,$2F,$F0,$9F
	.dc.b	$89,$7D,$2F,$F0,$86,$C6,$63,$07,$F0,$CE,$2E,$38,$F8,$DA,$FF,$58
	.dc.b	$CD,$18,$8F,$E0,$8D,$7F,$81,$FC,$57,$E1,$6F,$F8,$D3,$63,$DC,$D7
	.dc.b	$A3,$5F,$E1,$D8,$3E,$0B,$D0,$4A,$F6,$3F,$C8,$BE,$2D,$9D,$24,$FF
	.dc.b	$29,$8C,$BF,$CA,$FE,$A9,$F1,$DF,$CA,$9B,$19,$8C,$1F,$CF,$7E,$3D
	.dc.b	$E9,$7F,$36,$9F,$E7,$91,$8F,$33,$53,$9D,$3F,$11,$39,$AD,$FD,$43
	.dc.b	$E3,$7A,$36,$33,$18,$3C,$1F,$51,$F2,$E8,$D2,$72,$78,$3B,$65,$F9
	.dc.b	$32,$F9,$59,$30,$F4,$D7,$AF,$E5,$3F,$03,$3E,$66,$BF,$E1,$49,$46
	.dc.b	$20,$73,$70,$F5,$F9,$67,$97,$F1,$DB,$8F,$C6,$6C,$FC,$46,$BD,$FE
	.dc.b	$6B,$9E,$06,$93,$A7,$E9,$3F,$C1,$27,$D8,$D9,$E1,$D1,$A3,$EC,$7D
	.dc.b	$83,$E0,$69,$32,$FB,$31,$97,$ED,$DF,$E3,$DB,$D1,$E0,$23,$F8,$0F
	.dc.b	$1F,$EF,$65,$F1,$DA,$FC,$6C,$BF,$9A,$F2,$FE,$26,$73,$3F,$4F,$E7
	.dc.b	$1F,$0B,$F8,$D6,$E1,$F8,$C6,$CF,$E2,$9A,$FD,$5F,$9D,$FF,$0F,$37
	.dc.b	$DA,$FC,$B4,$62,$3C,$4D,$7F,$05,$F3,$DF,$EC,$7B,$FC,$CB,$63,$F2
	.dc.b	$ED,$7F,$04,$D7,$FC,$AF,$F9,$7F,$D8,$6F,$27,$F8,$7E,$31,$1F,$86
	.dc.b	$6A,$FA,$AF,$D4,$DF,$F8,$85,$B1,$FE,$25,$6B,$FC,$BD,$AF,$F9,$99
	.dc.b	$FE,$F0,$EF,$D4,$FF,$17,$46,$23,$84,$7F,$8C,$7E,$DB,$FA,$FC,$7F
	.dc.b	$9B,$9A,$FE,$69,$A9,$BE,$27,$F9,$66,$3F,$A0,$35,$FF,$3B,$B5,$FD
	.dc.b	$1B,$5F,$F3,$DF,$C1,$FF,$8F,$5F,$D3,$B5,$34,$FF,$21,$79,$8F,$AA
	.dc.b	$7D,$56,$1B,$5F,$E4,$99,$D4,$E4,$E3,$F4,$8F,$AA,$A4,$AE,$11,$97
	.dc.b	$F8,$9F,$4F,$F8,$75,$FD,$DC,$FF,$BF,$7A,$3F,$61,$ED,$7F,$A5,$23
	.dc.b	$15,$BA,$ED,$C1,$FE,$98,$FA,$97,$D5,$6E,$3F,$82,$FE,$0C,$FC,$29
	.dc.b	$ED,$E7,$04,$E6,$30,$D8,$F1,$D7,$EA,$71,$F0,$65,$FE,$8A,$FA,$BF
	.dc.b	$E5,$15,$FE,$69,$6B,$FE,$6D,$6B,$FD,$DD,$AF,$FA,$BB,$F3,$C7,$CE
	.dc.b	$78,$E1,$18,$7C,$27,$F1,$E7,$0F,$C0,$3F,$90,$CF,$82,$BF,$D3,$BF
	.dc.b	$5E,$F6,$FA,$36,$B9,$87,$0F,$33,$C3,$87,$C5,$F1,$9E,$37,$F8,$F2
	.dc.b	$7C,$71,$8F,$13,$53,$99,$F2,$72,$3C,$0F,$E6,$31,$C2,$7F,$CC,$0D
	.dc.b	$ED,$E2,$72,$BF,$07,$D8,$7F,$AC,$A7,$BB,$5F,$81,$8A,$31,$AF,$E1
	.dc.b	$FC,$47,$C6,$72,$F8,$E9,$87,$27,$AF,$C7,$7B,$7E,$F7,$FC,$23,$3F
	.dc.b	$EA,$3C,$91,$88,$FF,$17,$46,$3C,$DF,$F5,$33,$63,$87,$9F,$FC,$ED
	.dc.b	$F8,$AD,$7C,$ED,$CD,$D1,$FD,$62,$B9,$57,$C8,$DB,$E2,$BF,$2B,$EB
	.dc.b	$7F,$2E,$FE,$0B,$8D,$18,$8E,$1C,$3F,$9B,$9C,$F9,$7F,$EB,$B6,$CF
	.dc.b	$F8,$12,$BF,$DB,$0D,$9F,$5E,$BF,$9C,$CB,$C8,$D9,$FE,$AC,$DC,$9F
	.dc.b	$8B,$BE,$77,$F0,$DF,$8A,$FF,$90,$2F,$F0,$AB,$E9,$72,$6B,$D7,$FC
	.dc.b	$67,$E0,$3F,$B9,$AF,$EF,$27,$F6,$37,$0F,$C5,$6B,$9F,$1B,$FD,$AD
	.dc.b	$F8,$3F,$F0,$DE,$1F,$DD,$36,$7F,$4E,$D7,$F4,$CF,$F1,$3F,$67,$7B
	.dc.b	$FD,$DE,$4F,$A7,$4A,$FA,$BF,$65,$A3,$7C,$FD,$BF,$03,$FF,$71,$D7
	.dc.b	$FB,$A6,$91,$88,$FF,$3A,$46,$3D,$7F,$3C,$E6,$7D,$F2,$FC,$57,$C5
	.dc.b	$97,$F8,$6A,$31,$87,$B2,$73,$5E,$55,$8F,$C0,$FC,$57,$F7,$53,$F0
	.dc.b	$1F,$D2,$37,$C0,$D7,$89,$F1,$BF,$08,$FD,$54,$F8,$9A,$F8,$5A,$FC
	.dc.b	$36,$E4,$FC,$55,$F7,$BF,$41,$B5,$F4,$1A,$FD,$05,$E8,$FD,$6F,$5B
	.dc.b	$F3,$1C,$5F,$43,$04,$3E,$D7,$46,$FD,$0F,$BB,$F4,$BF,$98,$CB,$E6
	.dc.b	$B5,$F9,$6D,$7E,$53,$5F,$B9,$F8,$5F,$F9,$DD,$FC,$2B,$5F,$C2,$B7
	.dc.b	$1F,$23,$FE,$8A,$8F,$C5,$FE,$2B,$FD,$31,$BF,$75,$F9,$08,$37,$D6
	.dc.b	$FC,$9F,$DB,$BE,$83,$F2,$ED,$BF,$C4,$35,$FC,$3B,$5F,$C2,$FE,$2A
	.dc.b	$FB,$0F,$CE,$B6,$3F,$3A,$D7,$F3,$AB,$ED,$3F,$47,$F7,$0F,$B9,$BF
	.dc.b	$80,$CB,$FD,$2C,$E5,$FD,$3F,$B0,$FB,$BF,$D6,$B6,$3F,$3C,$D7,$F3
	.dc.b	$AD,$7F,$37,$F8,$AF,$C2,$7F,$62,$D7,$F6,$2D,$E2,$FC,$03,$FC,$26
	.dc.b	$3F,$A9,$FD,$97,$F0,$97,$FA,$93,$EB,$73,$9F,$F6,$58,$D5,$7E,$E2
	.dc.b	$5F,$80,$FF,$5A,$7F,$DC,$8D,$7F,$86,$B5,$FF,$43,$B5,$7D,$D3,$E7
	.dc.b	$3F,$B9,$36,$3F,$B3,$B5,$FE,$CD,$3F,$DA,$2B,$F3,$5F,$8A,$7E,$66
	.dc.b	$7F,$88,$B5,$FE,$22,$D8,$FE,$1F,$F3,$AF,$A5,$7D,$87,$F1,$AA,$7D
	.dc.b	$FD,$26,$1D,$3B,$BF,$8F,$F2,$7F,$7F,$F2,$FF,$84,$36,$7F,$8D,$37
	.dc.b	$2F,$EC,$7F,$2B,$C3,$EB,$25,$7F,$26,$A4,$74,$F9,$BF,$83,$F7,$92
	.dc.b	$BE,$0B,$F1,$39,$BE,$2F,$3C,$9F,$8A,$ED,$FF,$05,$71,$FE,$1F,$8B
	.dc.b	$EB,$7B,$DF,$AD,$DB,$F9,$FB,$6B,$FD,$73,$5B,$9F,$F5,$DE,$CF,$AE
	.dc.b	$7D,$CF,$F8,$D3,$F0,$67,$6D,$3F,$C7,$55,$B9,$C6,$3F,$E2,$8C,$1F
	.dc.b	$A5,$FE,$E6,$3F,$C5,$11,$8F,$0F,$71,$FE,$2B,$B7,$DF,$3E,$DF,$FA
	.dc.b	$6F,$A9,$F9,$9C,$3F,$9B,$36,$D7,$C2,$B7,$E3,$F0,$F8,$09,$5F,$D5
	.dc.b	$2F,$82,$F0,$7F,$44,$FF,$9C,$AF,$FB,$1C,$7E,$A7,$E6,$3C,$53,$FE
	.dc.b	$DE,$8C,$4E,$AF,$FC,$FA,$F0,$C3,$FB,$5E,$5F,$DB,$2F,$B4,$61,$D1
	.dc.b	$FE,$E8,$8D,$5F,$ED,$DB,$3D,$27,$C1,$A5,$7D,$0D,$AF,$1C,$7C,$2F
	.dc.b	$8B,$FF,$10,$65,$E8,$D3,$FC,$67,$F0,$9C,$DB,$5F,$E9,$2C,$FE,$FF
	.dc.b	$9E,$FB,$DD,$63,$CA,$7F,$B2,$7E,$27,$EB,$CB,$E1,$FA,$68,$C4,$FE
	.dc.b	$22,$B8,$63,$F5,$3E,$FF,$E6,$E1,$F2,$BE,$5B,$FD,$05,$AF,$B5,$B5
	.dc.b	$E3,$38,$71,$67,$E5,$F9,$CE,$A7,$EB,$1F,$E8,$88,$EF,$BB,$1E,$F9
	.dc.b	$FD,$AF,$27,$E3,$6F,$30,$EC,$7E,$06,$3E,$51,$97,$A9,$F8,$58,$CB
	.dc.b	$C3,$87,$E3,$36,$38,$6B,$F8,$68,$E8,$7A,$F1,$FD,$4F,$CA,$7D,$1E
	.dc.b	$9F,$F3,$FD,$7D,$56,$E5,$FB,$07,$F8,$1D,$F1,$72,$FE,$79,$B7,$D2
	.dc.b	$35,$E2,$7D,$DF,$82,$7D,$F5,$24,$F8,$E1,$EC,$62,$3E,$5D,$A7,$8C
	.dc.b	$6E,$BA,$57,$E3,$FE,$8B,$F5,$F3,$F1,$B6,$46,$34,$F8,$51,$8C,$7E
	.dc.b	$47,$D0,$F1,$7E,$55,$6F,$5F,$D7,$B7,$0F,$D7,$56,$DE,$D1,$97,$F7
	.dc.b	$0D,$BF,$F4,$EC,$6A,$FA,$91,$F3,$57,$3C,$3C,$B7,$C4,$E5,$19,$C3
	.dc.b	$C7,$7C,$6B,$D7,$39,$BF,$E0,$57,$C6,$BD,$B1,$BB,$E2,$F8,$CF,$09
	.dc.b	$93,$8C,$57,$F8,$95,$75,$7D,$B1,$9C,$3D,$B8,$7A,$23,$18,$7A,$2A
	.dc.b	$C6,$EF,$95,$FD,$19,$F1,$5F,$A9,$D9,$8C,$3F,$8F,$D7,$0C,$3D,$93
	.dc.b	$E8,$34,$7A,$0A,$7A,$0E,$75,$E5,$78,$34,$95,$F1,$1F,$14,$72,$BF
	.dc.b	$34,$6A,$3F,$02,$74,$9E,$07,$99,$C1,$E3,$8E,$67,$17,$E4,$9D,$6F
	.dc.b	$D9,$87,$C5,$8E,$26,$A7,$8D,$7D,$88,$FE,$67,$87,$E7,$9B,$5F,$CE
	.dc.b	$DF,$84,$7F,$38,$FB,$FF,$F6,$73,$FD,$21,$AF,$F4,$76,$BF,$D3,$DB
	.dc.b	$89,$FF,$6C,$61,$FA,$86,$E1,$FD,$4B,$2F,$F5,$2C,$E7,$9B,$F8,$95
	.dc.b	$6B,$0E,$D1,$FC,$56,$B7,$7E,$38,$D9,$27,$F7,$13,$A7,$89,$FB,$BA
	.dc.b	$E8,$F9,$E7,$A7,$6F,$1C,$3C,$53,$B9,$CB,$49,$1A,$3F,$81,$DF,$78
	.dc.b	$E8,$6B,$0F,$33,$E4,$BE,$D1,$8A,$E4,$FC,$38,$CC,$FA,$27,$2F,$F7
	.dc.b	$5A,$ED,$8F,$CD,$C3,$73,$E0,$AF,$1D,$E2,$B5,$3E,$33,$94,$67,$1F
	.dc.b	$87,$3A,$AE,$35,$B3,$A6,$3E,$73,$CC,$E6,$78,$46,$63,$A9,$9A,$EA
	.dc.b	$76,$C7,$D2,$E3,$1F,$34,$61,$F2,$1F,$43,$0F,$83,$5E,$23,$E2,$5F
	.dc.b	$8A,$33,$87,$8A,$72,$78,$A3,$38,$77,$8F,$A4,$79,$6F,$C3,$8F,$82
	.dc.b	$3D,$98,$F8,$DD,$1C,$2F,$9D,$70,$8F,$99,$93,$EF,$FF,$A8,$E1,$FD
	.dc.b	$95,$B3,$FD,$49,$FE,$B3,$19,$C3,$F2,$D5,$C5,$B3,$C3,$0F,$F0,$F3
	.dc.b	$C2,$F9,$4F,$0F,$BE,$FF,$17,$47,$F5,$E6,$D7,$F8,$97,$EF,$DC,$9B
	.dc.b	$A9,$97,$F7,$9B,$DE,$DF,$E6,$4B,$DC,$E6,$BF,$1E,$C3,$FC,$1B,$F4
	.dc.b	$FF,$E8,$C7,$FC,$10,$7E,$85,$C4,$7F,$69,$8C,$7A,$9C,$63,$FC,$23
	.dc.b	$3A,$CB,$FB,$5D,$68,$E6,$7F,$89,$FA,$DD,$6F,$BF,$FD,$E4,$61,$F9
	.dc.b	$3E,$4B,$E6,$B3,$5B,$BF,$F0,$B4,$E5,$E9,$19,$CB,$FC,$0B,$39,$7F
	.dc.b	$C6,$38,$75,$AF,$F0,$F7,$DD,$BE,$5F,$F8,$67,$89,$FE,$45,$E6,$FB
	.dc.b	$33,$9A,$FF,$1F,$4E,$5E,$35,$A3,$EF,$F8,$7F,$43,$1F,$F3,$ED,$F1
	.dc.b	$7E,$2E,$BF,$2F,$E4,$3E,$63,$70,$FA,$29,$FC,$0C,$87,$07,$B6,$3F
	.dc.b	$CE,$CD,$8D,$1A,$9E,$4F,$FB,$7C,$E3,$F3,$3E,$0A,$FB,$ED,$8F,$F2
	.dc.b	$AE,$6F,$29,$F1,$7C,$BD,$2E,$73,$9F,$8C,$C6,$3F,$E9,$CA,$B8,$2B
	.dc.b	$53,$E9,$BF,$F6,$CF,$DF,$BE,$DB,$75,$DF,$0E,$3D,$EA,$D6,$1F,$EA
	.dc.b	$19,$EB,$C8,$F6,$3F,$C8,$13,$FE,$B7,$FC,$DB,$F6,$79,$FC,$0A,$D7
	.dc.b	$8F,$E4,$A5,$69,$FF,$63,$FD,$FF,$EB,$AF,$86,$1D,$36,$FF,$5E,$4E
	.dc.b	$6B,$E9,$63,$D1,$F6,$1F,$43,$C9,$73,$FC,$2B,$FF,$72,$FC,$C5,$9C
	.dc.b	$D7,$09,$FB,$90,$9D,$54,$AD,$57,$FB,$BB,$EF,$70,$B3,$FB,$AE,$67
	.dc.b	$78,$2B,$C1,$E5,$78,$5F,$9F,$E8,$FE,$2A,$BC,$5C,$7F,$EA,$3E,$9C
	.dc.b	$FC,$AE,$9B,$3A,$A4,$FD,$6E,$C7,$CC,$6B,$4F,$25,$FE,$5F,$17,$C2
	.dc.b	$3E,$63,$EF,$B6,$7C,$2F,$F0,$CD,$FF,$1A,$E6,$F9,$9F,$8B,$75,$3D
	.dc.b	$5B,$1F,$EE,$AF,$5B,$E0,$4F,$4F,$1F,$E8,$7A,$92,$3E,$5F,$C3,$3F
	.dc.b	$06,$C6,$1D,$84,$66,$7F,$C5,$11,$89,$FD,$53,$FB,$39,$D3,$FE,$2D
	.dc.b	$6B,$FB,$4F,$BD,$FC,$17,$13,$EF,$AC,$61,$FD,$0F,$DD,$BE,$5A,$98
	.dc.b	$FA,$EB,$9F,$D6,$F8,$77,$C1,$F8,$B9,$11,$89,$FF,$1F,$46,$27,$53
	.dc.b	$9D,$3F,$DF,$92,$F9,$07,$E0,$FF,$B2,$5D,$B6,$30,$FE,$DB,$F1,$37
	.dc.b	$C8,$53,$1F,$4F,$74,$FC,$5F,$E1,$1E,$B7,$C6,$DC,$FC,$2D,$7F,$8B
	.dc.b	$E6,$8C,$47,$E2,$7D,$9F,$6D,$7E,$59,$FE,$00,$DD,$7F,$84,$C6,$27
	.dc.b	$F0,$F6,$F9,$AF,$DD,$76,$7F,$18,$6E,$1F,$95,$5E,$8F,$F1,$27,$37
	.dc.b	$FC,$67,$DD,$FF,$7A,$57,$CC,$6C,$7C,$A6,$BE,$A6,$BF,$A1,$F9,$BF
	.dc.b	$CD,$61,$FC,$A1,$FE,$A8,$DF,$03,$F9,$0C,$62,$FF,$51,$F8,$9B,$E4
	.dc.b	$FF,$92,$B6,$3F,$6E,$FE,$E7,$0F,$7B,$97,$E0,$FC,$FF,$F2,$87,$EF
	.dc.b	$B5,$F9,$ED,$7F,$6C,$D7,$F9,$C7,$E2,$6E,$9F,$E6,$AB,$E8,$2C,$61
	.dc.b	$FE,$73,$F4,$5F,$B2,$E1,$37,$F1,$3F,$EB,$59,$FF,$0E,$8D,$7D,$23
	.dc.b	$DD,$FC,$D3,$67,$F3,$2D,$CF,$F2,$ED,$7F,$0F,$1F,$30,$FE,$1C,$7C
	.dc.b	$97,$FA,$83,$96,$FB,$BF,$D3,$BE,$AB,$F9,$74,$FF,$20,$6C,$FE,$21
	.dc.b	$7D,$3E,$FD,$4F,$EB,$7F,$02,$FA,$77,$91,$FC,$0A,$30,$D8,$CE,$3F
	.dc.b	$EA,$F9,$CB,$7F,$13,$F9,$8F,$F4,$18,$47,$CD,$8C,$3A,$CB,$F9,$0C
	.dc.b	$EB,$87,$C7,$F5,$7F,$E0,$BE,$AF,$EF,$B2,$8C,$61,$FE,$07,$87,$D5
	.dc.b	$FF,$7A,$8F,$04,$EB,$06,$CF,$BA,$7B,$3F,$C7,$A3,$0D,$8C,$DF,$FB
	.dc.b	$2E,$72,$DF,$B3,$FE,$8A,$FF,$5D,$8E,$9F,$25,$FE,$04,$53,$AC,$3F
	.dc.b	$9E,$C6,$AB,$E2,$7D,$77,$F8,$45,$BF,$11,$F0,$3E,$8F,$E9,$CE,$6F
	.dc.b	$E2,$E8,$9F,$A8,$DB,$FA,$4D,$9F,$F1,$28,$DB,$E4,$57,$DB,$FC,$4B
	.dc.b	$A7,$07,$CC,$6E,$1F,$2D,$FE,$BE,$FD,$97,$37,$CD,$CB,$F6,$1B,$3F
	.dc.b	$E3,$2C,$FF,$87,$FC,$2F,$E3,$27,$37,$F5,$FF,$12,$7C,$EF,$E4,$1B
	.dc.b	$1F,$87,$5F,$39,$99,$D3,$27,$E8,$7D,$E7,$DD,$D5,$7C,$27,$FC,$17
	.dc.b	$F6,$5F,$E8,$FD,$3F,$C1,$2D,$8F,$F2,$4F,$D3,$7F,$BF,$5D,$FC,$E7
	.dc.b	$F8,$6E,$31,$BF,$F6,$76,$DF,$F8,$4B,$EF,$3F,$0B,$39,$F9,$CF,$F1
	.dc.b	$14,$62,$3E,$8B,$5F,$F2,$BB,$5F,$F2,$F3,$7E,$2F,$FD,$0B,$C7,$FE
	.dc.b	$40,$F9,$6F,$B5,$87,$CB,$FB,$4F,$F1,$A6,$0B,$DE,$9F,$91,$9C,$F2
	.dc.b	$9D,$D3,$EE,$FF,$A0,$D6,$E3,$FC,$70,$7F,$9B,$FE,$6B,$FC,$02,$D2
	.dc.b	$B8,$4F,$F8,$1B,$EE,$FF,$94,$7D,$DD,$F3,$8D,$25,$7E,$48,$4F,$F8
	.dc.b	$07,$4E,$5F,$05,$FE,$3D,$9D,$4E,$5B,$19,$7F,$CF,$FA,$FF,$5F,$FA
	.dc.b	$EF,$F4,$0D,$FF,$98,$7D,$84,$63,$4F,$F2,$CC,$62,$FF,$CA,$57,$FE
	.dc.b	$B2,$F8,$BD,$47,$CB,$97,$D8,$E2,$BF,$B6,$36,$33,$3C,$9E,$8F,$F8
	.dc.b	$AF,$E5,$FF,$DA,$AF,$F8,$A9,$AF,$C1,$6B,$FB,$46,$BF,$E6,$EF,$A3
	.dc.b	$FE,$5E,$F9,$9A,$FF,$91,$2D,$18,$CF,$FC,$C7,$5D,$9F,$F4,$14,$3E
	.dc.b	$F7,$D4,$DB,$E9,$18,$F3,$BE,$0F,$77,$E1,$FD,$1B,$CD,$FE,$D5,$BF
	.dc.b	$E7,$0D,$7F,$9A,$BF,$D6,$DF,$E7,$9F,$84,$FF,$97,$DB,$EE,$F4,$8C
	.dc.b	$47,$0A,$EF,$F3,$DF,$E5,$38,$FE,$88,$D7,$FA,$0B,$74,$FE,$7E,$FF
	.dc.b	$B8,$E3,$FA,$57,$AC,$F9,$1F,$4B,$63,$31,$DB,$E4,$3D,$BE,$5F,$FA
	.dc.b	$F6,$7F,$92,$F9,$BF,$E5,$ED,$BF,$A5,$36,$75,$3F,$CA,$25,$E6,$FF
	.dc.b	$85,$63,$F9,$BF,$BF,$F2,$23,$11,$FC,$D2,$FF,$17,$E0,$BE,$BB,$FD
	.dc.b	$DD,$3F,$D8,$7D,$4F,$E3,$6D,$F3,$77,$3A,$8F,$E7,$BF,$29,$E3,$23
	.dc.b	$FC,$F1,$F4,$7E,$40,$FB,$0F,$B2,$D9,$FE,$DC,$D7,$FB,$5E,$7F,$EE
	.dc.b	$A7,$FB,$97,$DD,$7F,$80,$1B,$D9,$F8,$28,$FF,$70,$C6,$BD,$DF,$C2
	.dc.b	$FB,$9E,$58,$FF,$77,$F0,$FF,$81,$7E,$8F,$E5,$4E,$67,$FB,$93,$5F
	.dc.b	$F7,$27,$47,$D5,$7F,$CA,$ED,$3F,$DD,$A5,$F1,$3F,$D5,$5C,$5F,$EB
	.dc.b	$E8,$FF,$78,$C7,$D4,$6E,$FF,$B8,$F8,$24,$FF,$AD,$2F,$FD,$7A,$EC
	.dc.b	$F1,$F8,$BE,$62,$47,$F8,$5F,$14,$7F,$A8,$2B,$FC,$77,$7D,$7E,$07
	.dc.b	$A5,$FF,$9B,$9B,$8F,$82,$3F,$E0,$04,$7D,$98,$C5,$3E,$E9,$DB,$E7
	.dc.b	$5F,$81,$53,$11,$EB,$F8,$6F,$D1,$63,$FE,$58,$ED,$F2,$46,$23,$FC
	.dc.b	$B3,$1E,$7F,$BB,$FF,$2C,$47,$FB,$71,$AF,$FA,$11,$B9,$79,$1F,$F4
	.dc.b	$DC,$7F,$B8,$BE,$27,$FC,$14,$DF,$AF,$F9,$90,$6F,$AB,$E5,$FA,$9F
	.dc.b	$19,$2B,$FD,$63,$97,$ED,$5F,$F5,$7F,$9A,$FC,$17,$EB,$36,$3F,$DE
	.dc.b	$2D,$7F,$D3,$2F,$FB,$B9,$FF,$7A,$FD,$E7,$F8,$11,$BE,$0F,$DF,$8F
	.dc.b	$D4,$C7,$FA,$81,$D3,$FE,$A2,$F9,$27,$D4,$FD,$B6,$C7,$FA,$D2,$30
	.dc.b	$7D,$FF,$BA,$7C,$67,$E2,$DB,$23,$FD,$5A,$E1,$BD,$FF,$91,$F9,$AF
	.dc.b	$F6,$AC,$6F,$8F,$F0,$31,$FE,$89,$8C,$76,$F3,$E9,$FF,$69,$7C,$67
	.dc.b	$1F,$9B,$FC,$D6,$BF,$B7,$BF,$F6,$2B,$5E,$0D,$55,$FE,$C9,$F1,$FF
	.dc.b	$31,$F7,$5F,$D5,$27,$FD,$86,$DC,$FF,$26,$DC,$BC,$4F,$F5,$88,$FC
	.dc.b	$D4,$BE,$83,$FA,$55,$7C,$18,$7D,$C7,$E2,$15,$70,$8C,$3C,$AB,$8C
	.dc.b	$FC,$97,$FD,$DA,$FF,$80,$4E,$DF,$3B,$FE,$0E,$8C,$C7,$F8,$15,$AF
	.dc.b	$F7,$EF,$A1,$7C,$3F,$F7,$36,$C7,$F7,$A6,$BE,$4C,$7D,$15,$D3,$0F
	.dc.b	$D9,$FD,$1F,$E6,$EB,$E9,$9F,$29,$C3,$FE,$EA,$6C,$70,$FB,$AF,$E1
	.dc.b	$DF,$87,$7F,$69,$AF,$F7,$7B,$67,$F7,$ED,$C3,$EF,$5F,$27,$F0,$1F
	.dc.b	$77,$E3,$6F,$DA,$7F,$19,$8F,$8F,$C2,$75,$8F,$DE,$73,$F8,$4F,$E3
	.dc.b	$A3,$FD,$FE,$DC,$BF,$87,$CF,$D0,$9F,$AF,$C5,$FE,$1C,$7F,$A9,$7A
	.dc.b	$F7,$D5,$FE,$AF,$7F,$66,$B1,$3F,$E4,$09,$FE,$B7,$5F,$E8,$86,$BF
	.dc.b	$83,$6B,$FE,$51,$E0,$7E,$1D,$F5,$49,$1F,$38,$F0,$9F,$43,$57,$DE
	.dc.b	$3E,$EF,$FC,$63,$18,$6C,$66,$7F,$C6,$FF,$7B,$EF,$AF,$D5,$7E,$2E
	.dc.b	$FB,$7F,$E8,$4D,$AF,$BB,$C2,$8C,$47,$E1,$4F,$E8,$F8,$75,$71,$2F
	.dc.b	$77,$F8,$5F,$07,$F3,$A7,$F8,$7B,$7C,$5F,$EA,$11,$AC,$BF,$78,$E9
	.dc.b	$FD,$F7,$C6,$7F,$89,$EB,$FA,$1B,$63,$75,$F5,$7E,$23,$F8,$CF,$C7
	.dc.b	$BE,$9B,$F6,$11,$E5,$8C,$67,$FC,$4E,$31,$3A,$F9,$9F,$DB,$56,$AB
	.dc.b	$FC,$8B,$C3,$FB,$9C,$7F,$D8,$38,$7B,$DB,$19,$9E,$ED,$7B,$3D,$30
	.dc.b	$FE,$05,$E0,$7F,$4C,$C3,$B3,$63,$8D,$7F,$A2,$5F,$E3,$DF,$63,$FE
	.dc.b	$D0,$AF,$4E,$E8,$C4,$EA,$BB,$1F,$CE,$3E,$CF,$FD,$AC,$F3,$3F,$8F
	.dc.b	$CF,$07,$F9,$0F,$5F,$F2,$D7,$FD,$27,$5F,$CB,$63,$18,$78,$2B,$87
	.dc.b	$C7,$3E,$57,$AE,$1C,$0F,$F5,$53,$8A,$F9,$F3,$BF,$81,$F2,$F9,$DF
	.dc.b	$D8,$F8,$BC,$F3,$9C,$3C,$CB,$AD,$F5,$5A,$6C,$EB,$2E,$DA,$FD,$EF
	.dc.b	$BD,$F8,$33,$FE,$7C,$6F,$21,$BF,$63,$87,$EB,$4E,$CF,$EA,$DC,$BF
	.dc.b	$7A,$33,$5C,$67,$27,$73,$E6,$1F,$CF,$35,$F0,$F2,$7F,$66,$8F,$88
	.dc.b	$75,$3E,$34,$3E,$F7,$FD,$37,$C3,$F9,$48,$C7,$89,$F1,$A7,$3A,$7F
	.dc.b	$69,$9E,$1E,$56,$6B,$E4,$FC,$DB,$F0,$0A,$67,$2F,$C0,$5B,$E8,$9F
	.dc.b	$3D,$F4,$9B,$3A,$6B,$E6,$5F,$59,$B9,$FC,$6F,$63,$F0,$F7,$D9,$61
	.dc.b	$B8,$73,$AF,$A3,$F4,$FF,$E0,$8E,$1F,$D6,$71,$A3,$1C,$3F,$B6,$6D
	.dc.b	$FE,$D7,$EA,$78,$ED,$F3,$AF,$83,$6B,$75,$FE,$BA,$AD,$BF,$42,$B6
	.dc.b	$D9,$DC,$27,$78,$79,$27,$26,$A3,$38,$79,$70,$F0,$1E,$38,$CE,$1F
	.dc.b	$C2,$EB,$8C,$E6,$36,$78,$5B,$7F,$0A,$B8,$CF,$FB,$4A,$B8,$BE,$98
	.dc.b	$CE,$1E,$AA,$F3,$5F,$5A,$E8,$48,$CB,$C4,$EF,$87,$A5,$D3,$FC,$92
	.dc.b	$FA,$E1,$E0,$AF,$25,$70,$3D,$97,$5F,$9C,$5C,$3C,$98,$75,$AE,$33
	.dc.b	$D8,$E2,$53,$89,$67,$8B,$A8,$E2,$7A,$89,$3D,$4F,$44,$78,$6F,$D1
	.dc.b	$18,$8E,$91,$87,$F1,$46,$EF,$D1,$87,$C0,$6C,$64,$F2,$46,$5A,$66
	.dc.b	$FC,$6F,$43,$E8,$9F,$D1,$2B,$F8,$03,$F3,$CC,$57,$F8,$DF,$07,$07
	.dc.b	$F2,$89,$FE,$AA,$D7,$FA,$A3,$54,$70,$ED,$E5,$3F,$CF,$30,$FD,$5B
	.dc.b	$72,$FE,$BF,$BF,$FA,$AE,$73,$5F,$A9,$AD,$61,$DA,$3F,$59,$5B,$BF
	.dc.b	$1C,$6C,$93,$FA,$B9,$D3,$C4,$FD,$75,$74,$7C,$ED,$C3,$F8,$4F,$3F
	.dc.b	$2C,$7C,$75,$C3,$0F,$0C,$E7,$1F,$0C,$E4,$F0,$46,$5F,$05,$72,$3F
	.dc.b	$85,$E3,$E4,$8F,$0D,$ED,$A7,$23,$A4,$F1,$BE,$EF,$2B,$E1,$87,$CF
	.dc.b	$AD,$D7,$CC,$9C,$E3,$ED,$3D,$D8,$F8,$6F,$86,$1D,$70,$F3,$E1,$C5
	.dc.b	$E6,$60,$87,$DE,$AE,$57,$D7,$1F,$1C,$61,$F0,$9F,$22,$79,$9E,$EB
	.dc.b	$F0,$3C,$F1,$DB,$AC,$7E,$9B,$70,$EC,$FC,$5A,$E3,$3E,$AC,$7E,$29
	.dc.b	$E9,$8C,$D7,$07,$E4,$57,$3C,$3C,$F8,$78,$27,$9E,$3E,$9A,$D1,$DF
	.dc.b	$0E,$AF,$C2,$C7,$C0,$FF,$8D,$67,$85,$7F,$65,$FB,$FF,$4B,$FE,$22
	.dc.b	$FB,$8F,$EE,$6E,$67,$F2,$BC,$49,$CE,$3F,$DA,$6F,$8B,$6B,$F3,$7A
	.dc.b	$A7,$FB,$2E,$7F,$DA,$1D,$E5,$FE,$E2,$FB,$FE,$6B,$BD,$F6,$AF,$B7
	.dc.b	$E7,$5A,$9C,$C6,$2A,$61,$FD,$D3,$EF,$DD,$58,$9F,$EB,$13,$F9,$BE
	.dc.b	$B7,$D1,$FE,$36,$30,$D7,$FC,$27,$3F,$E5,$B4,$61,$FE,$12,$E1,$FF
	.dc.b	$6A,$98,$75,$C7,$C1,$FE,$DC,$7E,$AF,$C2,$7C,$E7,$7A,$FF,$1F,$69
	.dc.b	$FE,$41,$F9,$EF,$E0,$6B,$E1,$5C,$58,$F1,$55,$A9,$CE,$5F,$E3,$98
	.dc.b	$D1,$F4,$BE,$FD,$F1,$58,$9F,$F2,$C7,$4F,$F9,$7F,$AF,$EE,$F9,$6F
	.dc.b	$A6,$FB,$FE,$62,$73,$18,$65,$FF,$88,$BE,$E9,$F3,$58,$E1,$FF,$1A
	.dc.b	$F1,$BE,$87,$E8,$79,$FF,$05,$B8,$7F,$8F,$A7,$4F,$CD,$6D,$7C,$C6
	.dc.b	$C7,$FB,$FE,$3F,$CB,$53,$F4,$63,$13,$F1,$1E,$91,$F3,$F8,$3C,$11
	.dc.b	$F0,$3D,$57,$CA,$7D,$9F,$5B,$FD,$43,$76,$BF,$D4,$F3,$71,$70,$BD
	.dc.b	$6F,$2E,$1F,$EB,$1F,$9D,$D4,$FF,$AD,$6E,$E4,$9B,$1A,$F6,$5F,$0B
	.dc.b	$4E,$D9,$BE,$FB,$F8,$5E,$5F,$8D,$CB,$FD,$8D,$5A,$9C,$D7,$F9,$B1
	.dc.b	$BA,$7E,$A7,$EF,$BF,$CA,$93,$FE,$DB,$9C,$F6,$BF,$31,$F2,$4A,$FF
	.dc.b	$DA,$13,$9D,$7E,$04,$3C,$27,$C8,$7C,$3A,$D5,$FF,$BA,$67,$2F,$D5
	.dc.b	$AE,$91,$87,$0D,$7E,$D3,$5F,$11,$DF,$DB,$F0,$C7,$1D,$BF,$5D,$39
	.dc.b	$AF,$06,$1F,$43,$AD,$F2,$5F,$4B,$B5,$F5,$99,$8C,$3B,$E9,$C6,$1F
	.dc.b	$EF,$2B,$FF,$52,$36,$BE,$DF,$C4,$BE,$A6,$9F,$7B,$E2,$BF,$03,$5F
	.dc.b	$27,$99,$5B,$8F,$F7,$EC,$EB,$5F,$C4,$B6,$8B,$EA,$3A,$76,$7C,$DE
	.dc.b	$57,$C9,$D3,$E9,$FC,$77,$EE,$B2,$FB,$1C,$7F,$C4,$79,$BE,$E5,$6B
	.dc.b	$C5,$FE,$27,$AF,$DD,$C3,$83,$F5,$5A,$FE,$AD,$AF,$D2,$5F,$37,$F4
	.dc.b	$BE,$0F,$C2,$DC,$BC,$27,$C1,$FB,$D7,$C9,$FE,$89,$B1,$AC,$3F,$53
	.dc.b	$0F,$94,$D5,$7E,$A3,$E5,$37,$5F,$BD,$E3,$46,$23,$F0,$B1,$89,$F6
	.dc.b	$B7,$0F,$C8,$2F,$95,$FC,$C4,$78,$63,$1E,$A7,$ED,$A7,$36,$E1,$FD
	.dc.b	$74,$7F,$48,$BE,$1E,$C7,$F2,$36,$F9,$DF,$BE,$8F,$C6,$C6,$0E,$D9
	.dc.b	$FB,$5F,$1B,$70,$F6,$35,$F5,$B5,$FC,$57,$B6,$F1,$7F,$85,$A2,$B8
	.dc.b	$7F,$87,$FC,$BF,$E3,$E7,$D5,$7F,$93,$8C,$67,$FC,$25,$FE,$D5,$1F
	.dc.b	$0D,$CB,$F9,$9F,$98,$FF,$74,$3F,$29,$B5,$F4,$DA,$FD,$26,$BF,$A1
	.dc.b	$F9,$BF,$E6,$0F,$88,$FE,$4E,$FE,$7C,$CD,$FF,$1B,$7F,$45,$F7,$5F
	.dc.b	$F0,$02,$3F,$8F,$36,$FF,$58,$FF,$51,$7E,$D3,$9A,$FE,$67,$6F,$93
	.dc.b	$FC,$1B,$6B,$D5,$18,$3F,$0F,$F7,$5F,$4F,$0F,$E5,$8F,$DD,$BF,$47
	.dc.b	$12,$35,$7F,$BD,$75,$5F,$CC,$FE,$87,$F7,$44,$EF,$FE,$82,$FE,$FF
	.dc.b	$D6,$7D,$23,$E6,$FF,$67,$18,$6C,$67,$1F,$E5,$13,$97,$F3,$67,$C8
	.dc.b	$3F,$87,$1F,$C2,$4C,$67,$FD,$51,$FE,$19,$F4,$BF,$E1,$C7,$F4,$0D
	.dc.b	$8F,$6A,$FA,$EF,$EA,$EE,$73,$FF,$0D,$5F,$A3,$EB,$BC,$A4,$C3,$CB
	.dc.b	$8A,$3D,$7F,$71,$F1,$B2,$FF,$03,$3F,$1A,$BF,$A6,$5A,$34,$DE,$9F
	.dc.b	$ED,$8B,$EA,$FF,$80,$E5,$E2,$6D,$F2,$FA,$BF,$E7,$AF,$F0,$BA,$E1
	.dc.b	$DD,$CB,$EB,$7F,$C2,$33,$F2,$9B,$9F,$F8,$1E,$91,$88,$FE,$9F,$18
	.dc.b	$F7,$BF,$93,$EE,$FB,$87,$C8,$F9,$BE,$BB,$FB,$3E,$1F,$9D,$AD,$4E
	.dc.b	$7D,$9F,$5E,$AF,$BE,$A7,$97,$EB,$AE,$DE,$9D,$91,$F8,$46,$C7,$A6
	.dc.b	$30,$7E,$23,$F1,$6F,$F3,$84,$7C,$16,$E7,$FA,$FD,$DF,$73,$E1,$AD
	.dc.b	$7C,$9F,$CE,$8C,$56,$DB,$C9,$F8,$05,$F0,$7B,$D3,$EA,$FB,$DF,$0B
	.dc.b	$1F,$A0,$DC,$3C,$1F,$70,$FA,$1D,$F1,$7D,$C8,$C5,$FD,$76,$D7,$0F
	.dc.b	$A2,$D5,$FF,$82,$FE,$43,$FB,$37,$AF,$FE,$F4,$F6,$71,$E2,$65,$B9
	.dc.b	$7F,$4B,$9F,$F7,$CC,$FF,$85,$BE,$E3,$FD,$1B,$97,$F8,$6B,$0F,$ED
	.dc.b	$97,$FD,$46,$13,$A7,$FC,$94,$D7,$FC,$22,$D7,$FC,$48,$DB,$79,$9F
	.dc.b	$E1,$3A,$FF,$14,$46,$3E,$0F,$FC,$49,$3F,$DE,$27,$CC,$EE,$FF,$D3
	.dc.b	$8E,$5F,$F1,$7B,$78,$FF,$BD,$7C,$7B,$EB,$79,$FB,$9F,$E3,$2B,$FF
	.dc.b	$1B,$7C,$CF,$F8,$B5,$FF,$13,$B5,$C9,$5D,$7C,$67,$F8,$8B,$ED,$AF
	.dc.b	$C3,$9A,$4A,$FF,$1C,$4F,$FA,$07,$2E,$3E,$47,$F6,$59,$CB,$5F,$F3
	.dc.b	$94,$ED,$FF,$20,$7A,$EE,$6F,$F1,$DD,$FF,$8E,$23,$13,$F4,$DA,$FF
	.dc.b	$47,$3F,$CC,$DF,$3F,$C3,$83,$FC,$87,$3A,$AE,$5C,$F9,$BF,$F2,$C4
	.dc.b	$FE,$45,$E9,$86,$57,$EB,$3F,$E9,$97,$E4,$35,$FF,$2B,$35,$FF,$46
	.dc.b	$35,$FF,$29,$FE,$09,$FE,$99,$7F,$14,$D7,$D0,$DC,$3F,$CA,$B3,$D5
	.dc.b	$FF,$2B,$C6,$71,$E7,$F2,$3E,$58,$FF,$5C,$DF,$4E,$AF,$67,$84,$FB
	.dc.b	$8F,$F2,$24,$FF,$1F,$7F,$D8,$ED,$F9,$8D,$C6,$A3,$F8,$13,$A7,$F8
	.dc.b	$B7,$E2,$7F,$E2,$F7,$FC,$65,$E5,$A3,$11,$C2,$BC,$3F,$28,$FA,$6B
	.dc.b	$ED,$CA,$FE,$61,$39,$6C,$7D,$3C,$1F,$79,$FD,$7E,$7B,$B5,$FE,$2D
	.dc.b	$5F,$DD,$DF,$1F,$E0,$1E,$97,$FD,$03,$48,$C4,$7F,$A5,$63,$1E,$27
	.dc.b	$E8,$5B,$1F,$E8,$F7,$87,$07,$8E,$BE,$6C,$FF,$37,$6F,$17,$FB,$DE
	.dc.b	$B8,$CF,$F3,$BD,$BF,$D1,$2F,$C6,$6B,$9F,$90,$FA,$BE,$93,$E7,$7E
	.dc.b	$05,$3E,$C3,$C8,$DB,$F2,$35,$FF,$6C,$61,$F7,$67,$7F,$29,$F2,$7C
	.dc.b	$CF,$F4,$EE,$7E,$56,$F4,$7C,$1C,$55,$F3,$FD,$6F,$D0,$65,$FD,$4D
	.dc.b	$B8,$7B,$5B,$7F,$DC,$BE,$EB,$FD,$BF,$5F,$82,$F5,$9F,$6D,$F2,$DB
	.dc.b	$1F,$2D,$AF,$A1,$AF,$F9,$CB,$57,$D4,$7E,$CF,$E1,$10,$FA,$6C,$36
	.dc.b	$BF,$C2,$75,$F8,$7F,$AB,$7D,$1D,$24,$FF,$9D,$F4,$7B,$DF,$8D,$F8
	.dc.b	$3F,$95,$3F,$86,$6B,$F8,$56,$BE,$26,$E6,$F8,$6F,$ED,$F3,$FE,$83
	.dc.b	$7F,$97,$37,$D0,$F5,$46,$A3,$FC,$B5,$8B,$CF,$7E,$03,$FC,$46,$30
	.dc.b	$D8,$CC,$FF,$A3,$FE,$E9,$AE,$FF,$0E,$EF,$3F,$ED,$E3,$E0,$B8,$7F
	.dc.b	$DC,$38,$23,$11,$FE,$F5,$8C,$65,$FD,$07,$37,$DD,$3E,$CB,$E9,$3F
	.dc.b	$EB,$8B,$FF,$50,$35,$FF,$2E,$BB,$7F,$D6,$96,$F8,$FF,$E2,$D1,$F5
	.dc.b	$1B,$19,$F2,$FE,$F4,$7E,$EF,$A9,$C7,$E5,$F7,$3E,$DE,$1F,$F0,$07
	.dc.b	$04,$62,$3E,$04,$60,$FE,$B4,$57,$FD,$73,$EC,$FF,$84,$9F,$F4,$B7
	.dc.b	$DE,$7E,$36,$0D,$F8,$1F,$A5,$F2,$9D,$72,$F1,$DF,$E3,$23,$19,$7C
	.dc.b	$96,$C3,$77,$D4,$FE,$EE,$3E,$1F,$CC,$3E,$D7,$CA,$D8,$F6,$35,$F5
	.dc.b	$B5,$FF,$6A,$FD,$DB,$E3,$3F,$48,$FE,$C2,$3F,$D1,$AD,$CF,$CF,$8A
	.dc.b	$38,$4F,$D4,$F8,$6F,$DF,$B7,$C6,$FC,$84,$62,$3F,$D2,$FA,$3D,$EF
	.dc.b	$CA,$7B,$DF,$CA,$67,$E9,$35,$FA,$2D,$7F,$1E,$D4,$7C,$5F,$BB,$7C
	.dc.b	$27,$EF,$57,$C1,$D3,$11,$FD,$BF,$CD,$FF,$46,$57,$F8,$BD,$AF,$F7
	.dc.b	$D6,$ED,$F5,$A1,$1F,$0F,$8F,$C3,$0F,$9E,$7D,$47,$F8,$75,$B1,$FE
	.dc.b	$21,$6B,$FA,$B3,$84,$7F,$88,$FA,$BE,$E4,$7F,$83,$5B,$87,$85,$B7
	.dc.b	$E0,$BE,$73,$F8,$AE,$AF,$E2,$3F,$01,$FE,$2E,$F9,$9F,$D9,$79,$3F
	.dc.b	$E1,$8F,$1B,$FD,$E9,$9F,$F1,$A8,$C4,$FC,$AE,$77,$D1,$7F,$A9,$EB
	.dc.b	$FC,$73,$3C,$B0,$E9,$1C,$7E,$9E,$E5,$FC,$81,$B8,$F4,$9F,$CC,$7D
	.dc.b	$DB,$E1,$F7,$F2,$4F,$A2,$A7,$4D,$55,$BB,$F0,$9F,$CC,$2B,$83,$8C
	.dc.b	$BF,$D2,$77,$CA,$37,$7C,$B0,$F2,$FD,$67,$F4,$4D,$FE,$0E,$08,$C6
	.dc.b	$DF,$C5,$DB,$87,$EB,$FB,$3F,$C6,$17,$C2,$73,$7C,$73,$EA,$76,$F5
	.dc.b	$6E,$1E,$5A,$FE,$53,$7C,$7C,$BC,$CF,$F3,$5C,$BF,$93,$5F,$E0,$2B
	.dc.b	$CA,$FC,$9C,$7F,$AF,$F0,$FC,$E8,$D5,$70,$8C,$12,$FE,$24,$E9,$FE
	.dc.b	$AF,$F8,$27,$F7,$BC,$3E,$0B,$6F,$F9,$B3,$67,$F8,$77,$8D,$C7,$14
	.dc.b	$E5,$BF,$0F,$F8,$08,$F8,$B7,$CB,$DB,$FC,$2E,$0F,$B7,$7C,$C6,$3B
	.dc.b	$1D,$14,$DC,$EA,$BE,$F9,$FC,$CA,$3F,$5B,$D9,$FD,$46,$B8,$CF,$F0
	.dc.b	$08,$C4,$EA,$3F,$67,$3A,$7F,$70,$73,$3A,$BF,$8A,$7E,$91,$FC,$29
	.dc.b	$C6,$DF,$A6,$8E,$06,$B5,$7E,$87,$E7,$8C,$36,$7F,$1A,$DD,$3F,$59
	.dc.b	$9F,$F5,$9A,$DB,$FA,$89,$D4,$7F,$83,$1B,$CF,$F2,$3D,$1F,$D6,$FC
	.dc.b	$47,$F1,$0A,$FD,$03,$5F,$23,$71,$FC,$EB,$FA,$98,$CF,$83,$C8,$8F
	.dc.b	$2E,$AF,$DC,$FE,$D8,$FA,$5B,$52,$BF,$5D,$39,$6C,$65,$F8,$91,$E1
	.dc.b	$D7,$FC,$00,$EE,$33,$3B,$C7,$F6,$33,$B9,$FF,$15,$BB,$8F,$DA,$FD
	.dc.b	$CF,$F8,$C6,$7E,$83,$FE,$38,$7F,$80,$35,$CF,$BD,$FC,$4A,$38,$4F
	.dc.b	$2F,$77,$06,$67,$F8,$BF,$91,$FE,$8F,$9F,$E0,$7F,$80,$FF,$3D,$6F
	.dc.b	$36,$BE,$36,$65,$FC,$3E,$B5,$AF,$F8,$75,$B1,$FE,$44,$C3,$A4,$FE
	.dc.b	$4B,$EE,$7F,$C1,$B8,$FE,$2F,$9D,$18,$9D,$73,$FE,$53,$C3,$FD,$34
	.dc.b	$E7,$1F,$CA,$35,$FF,$0A,$B6,$D3,$C2,$BF,$3B,$39,$35,$19,$AF,$CF
	.dc.b	$9C,$A7,$67,$85,$A6,$AB,$CF,$18,$C3,$D5,$5C,$DF,$65,$F7,$9E,$57
	.dc.b	$C6,$30,$D8,$CC,$F0,$7A,$61,$F0,$A7,$C1,$7D,$63,$15,$B9,$C9,$C0
	.dc.b	$F8,$B5,$5F,$F1,$CD,$71,$72,$78,$4E,$25,$9E,$84,$AF,$E1,$33,$C6
	.dc.b	$7E,$31,$B8,$CB,$DF,$0F,$8A,$E4,$FE,$15,$3C,$63,$35,$D6,$72,$71
	.dc.b	$C3,$E2,$C7,$CC,$8C,$61,$D6,$F9,$56,$B0,$F0,$D7,$73,$CB,$7D,$5F
	.dc.b	$82,$72,$35,$87,$E1,$9C,$46,$E3,$11,$AA,$EC,$F9,$8C,$61,$E3,$78
	.dc.b	$DF,$D4,$3C,$77,$F5,$63,$F9,$B5,$7F,$9A,$5F,$B8,$62,$BF,$C4,$0F
	.dc.b	$F9,$BB,$EE,$5D,$D8,$C5,$18,$9F,$F1,$34,$62,$75,$39,$9F,$E8,$8E
	.dc.b	$F0,$FD,$73,$71,$FE,$87,$BF,$F9,$BA,$73,$7F,$D0,$AB,$57,$FD,$05
	.dc.b	$AF,$FA,$0A,$B8,$5F,$F4,$C9,$CE,$1E,$3A,$FD,$34,$62,$B7,$1C,$EF
	.dc.b	$B1,$1F,$D6,$46,$9F,$85,$19,$C3,$CB,$1A,$DF,$18,$F8,$6A,$CF,$F5
	.dc.b	$7B,$E5,$8F,$8F,$0E,$24,$BF,$83,$87,$0B,$EE,$7A,$30,$FE,$C7,$87
	.dc.b	$2C,$7D,$A4,$8C,$D7,$1C,$7B,$61,$E3,$C3,$CC,$73,$26,$3E,$17,$CD
	.dc.b	$5C,$70,$E9,$1C,$27,$4F,$91,$B7,$FE,$F3,$AE,$18,$78,$71,$F3,$38
	.dc.b	$BE,$58,$F8,$F0,$E3,$8F,$D3,$C3,$51,$D3,$0E,$37,$E6,$BE,$07,$C6
	.dc.b	$70,$7D,$92,$9F,$83,$AE,$71,$C2,$3D,$A6,$B1,$F8,$AF,$1B,$FE,$81
	.dc.b	$7E,$07,$CB,$8F,$CC,$9E,$18,$FC,$49,$DE,$3D,$A7,$96,$3E,$59,$F9
	.dc.b	$99,$3E,$4B,$F4,$F7,$FE,$63,$7F,$A6,$B7,$0F,$C9,$CE,$F4,$B5,$C3
	.dc.b	$1F,$F0,$E3,$C1,$B5,$A9,$E3,$F7,$2E,$2B,$AE,$1B,$97,$F9,$CE,$74
	.dc.b	$D9,$D3,$4F,$F3,$C7,$ED,$9F,$E0,$F6,$F3,$F9,$A3,$FD,$5B,$18,$26
	.dc.b	$5F,$E3,$18,$E0,$DE,$7E,$15,$C5,$B5,$D9,$B3,$FE,$39,$FB,$8E,$3B
	.dc.b	$FF,$A0,$9B,$C1,$FE,$CA,$E0,$47,$FA,$5F,$0F,$F4,$73,$AB,$4E,$B1
	.dc.b	$F9,$B3,$98,$E3,$D3,$89,$FA,$3F,$BB,$7B,$2F,$FC,$94,$DC,$BF,$D6
	.dc.b	$7D,$BF,$E6,$6E,$CD,$EB,$34,$FF,$4F,$5E,$F5,$FF,$25,$D6,$71,$FE
	.dc.b	$F7,$EB,$3E,$47,$F8,$62,$EC,$B1,$86,$E3,$FE,$69,$AF,$F6,$CD,$71
	.dc.b	$8C,$47,$FB,$B2,$30,$FF,$9A,$9B,$1F,$EE,$08,$FF,$71,$F9,$1D,$1F
	.dc.b	$F0,$6E,$68,$C6,$BF,$EB,$0F,$C2,$BE,$4D,$FF,$AE,$57,$E0,$5F,$96
	.dc.b	$73,$87,$F9,$BD,$CC,$EF,$8B,$47,$F2,$EE,$07,$93,$FE,$FE,$9D,$1F
	.dc.b	$EC,$5F,$BB,$78,$9F,$A2,$6C,$67,$EC,$B1,$39,$6C,$64,$E5,$1F,$AA
	.dc.b	$FD,$AD,$F7,$97,$EB,$70,$DA,$FF,$6A,$4B,$E5,$33,$18,$6E,$DF,$E8
	.dc.b	$B9,$D4,$E4,$FF,$6C,$F8,$AF,$3B,$C1,$7F,$BA,$FC,$5F,$F2,$98,$7F
	.dc.b	$A4,$DB,$C7,$F8,$AF,$BA,$E1,$8F,$FB,$96,$73,$D7,$FE,$FD,$9C,$DF
	.dc.b	$1E,$EF,$95,$18,$C3,$FD,$CB,$7C,$2B,$F3,$FE,$B3,$E2,$BF,$85,$AF
	.dc.b	$9A,$B1,$86,$E7,$FE,$F5,$9E,$0F,$FB,$D3,$DB,$E0,$E6,$BF,$2F,$E7
	.dc.b	$A1,$83,$DC,$C3,$F9,$39,$B1,$88,$7B,$AF,$0F,$37,$F9,$CB,$B5,$D2
	.dc.b	$B5,$3F,$E1,$26,$C7,$E6,$CE,$75,$FE,$C8,$78,$35,$7E,$0D,$FA,$DA
	.dc.b	$F3,$E1,$E3,$8C,$4F,$A6,$7D,$6F,$0F,$13,$C6,$F2,$F1,$3C,$BD,$CF
	.dc.b	$87,$7C,$8E,$17,$BF,$63,$0F,$B3,$C2,$7E,$03,$EA,$C7,$BC,$63,$6F
	.dc.b	$53,$74,$F7,$36,$7F,$74,$BE,$15,$F6,$3D,$5B,$5E,$E3,$EA,$E8,$F8
	.dc.b	$87,$C2,$EE,$FD,$7F,$6F,$FF,$00,$2F,$86,$4C,$12,$FB,$8D,$FB,$FC
	.dc.b	$1F,$67,$E2,$DF,$C9,$EF,$D6,$D7,$F9,$B3,$5F,$E6,$6D,$D7,$E5,$79
	.dc.b	$3F,$17,$8C,$FC,$EF,$0B,$F8,$5D,$FD,$53,$F9,$DC,$EB,$4F,$3E,$8F
	.dc.b	$91,$7D,$AF,$D6,$7E,$1D,$F1,$9D,$4E,$77,$FA,$9F,$B3,$3E,$8B,$E8
	.dc.b	$B7,$1F,$A2,$DC,$3E,$8B,$5F,$B3,$EA,$3E,$47,$EF,$36,$BC,$BF,$2B
	.dc.b	$C7,$BF,$F4,$BC,$2A,$D6,$1E,$5F,$96,$FE,$CF,$5F,$AA,$D9,$1F,$C2
	.dc.b	$A3,$0E,$B1,$FB,$AE,$5F,$EC,$9F,$8B,$BE,$8F,$F1,$6D,$8F,$C5,$B7
	.dc.b	$0F,$C5,$B5,$FC,$9F,$CC,$FF,$36,$9F,$C1,$E6,$8C,$4F,$CB,$FB,$8F
	.dc.b	$C4,$3F,$56,$75,$EB,$FE,$9B,$E7,$7E,$06,$1F,$92,$7F,$85,$D7,$D3
	.dc.b	$72,$FE,$3D,$CE,$9F,$1B,$E7,$DE,$BF,$E9,$1B,$7F,$A4,$6D,$FE,$91
	.dc.b	$AF,$EE,$97,$E2,$DF,$1E,$11,$8F,$80,$FD,$DD,$78,$A3,$0F,$E3,$1F
	.dc.b	$EC,$8D,$F6,$7F,$80,$C6,$2B,$F1,$8F,$F4,$E9,$FD,$4B,$FD,$F7,$F0
	.dc.b	$4F,$E2,$8D,$8F,$E3,$0D,$9F,$EE,$8D,$7D,$1F,$44,$F9,$FF,$E4,$72
	.dc.b	$8C,$57,$ED,$BE,$31,$F7,$5F,$4B,$6F,$E6,$37,$BC,$38,$FA,$1F,$C7
	.dc.b	$A7,$3F,$84,$BE,$DB,$3E,$8F,$68,$C6,$7E,$AB,$FE,$87,$AF,$99,$AF
	.dc.b	$CD,$6B,$FD,$0B,$2F,$E5,$FE,$AF,$C0,$9F,$F0,$9E,$5E,$46,$DF,$F4
	.dc.b	$8D,$5F,$53,$FE,$03,$C7,$F7,$D8,$FF,$31,$AF,$97,$F8,$67,$E9,$DB
	.dc.b	$DF,$F1,$23,$F3,$59,$7C,$07,$57,$C9,$FC,$E3,$E5,$6B,$F8,$26,$D9
	.dc.b	$CE,$2F,$B8,$EB,$8F,$B7,$ED,$DF,$07,$FD,$5A,$B5,$E0,$7D,$69,$CD
	.dc.b	$7D,$5B,$F1,$E7,$FD,$92,$F7,$39,$D3,$95,$76,$D3,$FD,$DB,$F7,$0F
	.dc.b	$9A,$FC,$0F,$E2,$3F,$E7,$69,$FE,$45,$8A,$B5,$8F,$E4,$2B,$59,$77
	.dc.b	$97,$BF,$E7,$77,$87,$27,$37,$F8,$B7,$35,$FD,$DF,$F5,$C7,$DB,$7F
	.dc.b	$62,$7F,$41,$19,$8C,$7C,$76,$59,$7F,$D2,$B0,$7E,$B8,$FC,$3B,$31
	.dc.b	$FC,$EC,$CB,$FD,$DE,$35,$3F,$DF,$67,$2D,$8C,$C7,$F8,$07,$F1,$7F
	.dc.b	$E0,$F3,$FD,$E1,$FC,$BE,$3F,$E0,$F7,$33,$FE,$16,$73,$3B,$F8,$6F
	.dc.b	$EF,$B1,$A9,$F3,$E5,$89,$FF,$00,$C7,$F8,$7B,$E2,$3C,$8D,$2B,$84
	.dc.b	$FA,$7D,$F7,$DD,$BE,$EF,$0D,$AF,$F0,$D4,$EA,$72,$71,$FA,$CF,$F1
	.dc.b	$64,$E9,$AF,$F9,$EE,$51,$88,$FF,$0C,$C6,$34,$F9,$5F,$77,$F8,$B9
	.dc.b	$BC,$DE,$A8,$C4,$7F,$8A,$DA,$FF,$A1,$FF,$15,$FF,$03,$BF,$4D,$AF
	.dc.b	$F8,$C5,$AE,$FE,$0F,$E2,$FD,$E7,$F9,$3E,$3E,$D3,$5F,$F1,$C3,$5F
	.dc.b	$F4,$A3,$5F,$F1,$8B,$FD,$CD,$E3,$ED,$74,$23,$F8,$5F,$B8,$FE,$34
	.dc.b	$DF,$23,$EA,$C6,$23,$FC,$8C,$D7,$FA,$27,$DE,$3E,$17,$F1,$DE,$2A
	.dc.b	$31,$3F,$CD,$9A,$FF,$36,$6B,$FD,$37,$B9,$F4,$3C,$9B,$1F,$CF,$9A
	.dc.b	$FF,$3E,$6B,$FD,$57,$EE,$DF,$59,$E2,$6C,$7F,$46,$6B,$FD,$19,$AF
	.dc.b	$F5,$DF,$17,$F7,$AF,$F4,$C6,$BF,$D3,$1A,$EF,$C2,$F7,$ED,$FE,$85
	.dc.b	$8F,$EA,$6D,$7F,$A9,$B5,$FE,$D4,$D7,$F9,$F7,$98,$FA,$FF,$92,$D8
	.dc.b	$FE,$BA,$D7,$FA,$EB,$5F,$EE,$9B,$FF,$88,$5F,$EC,$6D,$7F,$B1,$B5
	.dc.b	$DF,$8F,$F5,$BE,$7B,$FC,$F9,$C1,$FE,$4B,$9F,$EF,$71,$83,$37,$FD
	.dc.b	$A6,$3C,$7E,$87,$98,$8F,$F7,$2F,$BA,$FC,$FB,$FD,$65,$AF,$F8,$19
	.dc.b	$B6,$6F,$A6,$FA,$FF,$C4,$B6,$BF,$BE,$B5,$FE,$FA,$D7,$FC,$29,$F2
	.dc.b	$7F,$9F,$7F,$C2,$6D,$B3,$4C,$7E,$30,$FB,$8F,$CB,$B6,$BF,$C1,$6D
	.dc.b	$7F,$C1,$6D,$7F,$C4,$5F,$74,$F3,$FF,$3E,$D8,$FF,$09,$B5,$FF,$09
	.dc.b	$B5,$FF,$4A,$E1,$FE,$5D,$7F,$C6,$1F,$0F,$FC,$D7,$E0,$7F,$C5,$6D
	.dc.b	$9D,$4F,$FA,$5B,$C8,$7C,$6E,$DB,$1F,$E9,$C6,$BB,$8E,$9F,$09,$F8
	.dc.b	$78,$FF,$66,$E1,$FE,$51,$E6,$C9,$23,$FC,$B7,$C5,$E3,$F8,$5F,$C0
	.dc.b	$47,$F9,$72,$30,$79,$FE,$23,$E8,$59,$A4,$F3,$7D,$9D,$1F,$26,$FC
	.dc.b	$2D,$55,$FE,$79,$E2,$D9,$F0,$FE,$27,$EC,$CF,$FA,$23,$93,$04,$7F
	.dc.b	$D7,$7E,$C7,$3B,$FF,$3B,$5F,$CB,$8C,$19,$F8,$D7,$DE,$FC,$F6,$DF
	.dc.b	$FB,$11,$AF,$FB,$11,$AF,$FB,$33,$F0,$9F,$EA,$F8,$1A,$48,$FB,$5E
	.dc.b	$E3,$E8,$29,$23,$FD,$A3,$19,$7F,$DA,$7E,$8B,$E6,$FF,$04,$D8,$CC
	.dc.b	$60,$FC,$37,$CC,$7F,$46,$8F,$B6,$D7,$E5,$B7,$1F,$F7,$2F,$9F,$FD
	.dc.b	$BB,$EB,$9F,$43,$49,$7F,$E0,$58,$CB,$7E,$27,$81,$1C,$10,$FE,$EB
	.dc.b	$D7,$FE,$FE,$97,$B5,$FF,$00,$3B,$7F,$27,$18,$AF,$F0,$05,$6E,$F4
	.dc.b	$FE,$53,$ED,$BF,$B4,$DF,$E7,$33,$47,$F8,$A2,$F8,$9F,$88,$70,$F8
	.dc.b	$8C,$7A,$FD,$1D,$7C,$01,$9E,$F5,$72,$60,$F1,$46,$63,$57,$E2,$8C
	.dc.b	$4F,$EA,$BC,$07,$E1,$F8,$8F,$CD,$C6,$1E,$86,$23,$55,$FC,$FD,$B8
	.dc.b	$7F,$07,$CB,$E0,$37,$3F,$DB,$B6,$F9,$1F,$0A,$DF,$6F,$FC,$96,$FE
	.dc.b	$43,$6F,$F7,$ED,$AF,$D7,$BF,$D2,$E7,$7E,$57,$92,$3F,$97,$69,$F0
	.dc.b	$1B,$5C,$A7,$47,$CD,$C3,$FC,$B7,$F2,$B7,$53,$9B,$F0,$4F,$F7,$4E
	.dc.b	$1E,$1E,$A3,$86,$98,$8B,$97,$C8,$AF,$EF,$73,$F2,$2B,$84,$E5,$AB
	.dc.b	$3F,$AF,$F6,$FF,$CC,$5B,$E5,$FD,$08,$C4,$FF,$22,$6B,$FC,$4B,$EA
	.dc.b	$5F,$84,$7B,$5B,$1F,$3D,$AF,$E3,$F2,$57,$D5,$87,$37,$F8,$23,$87
	.dc.b	$F4,$3E,$AF,$E4,$E7,$2D,$9F,$CF,$F1,$FE,$A6,$3F,$15,$3A,$3F,$12
	.dc.b	$53,$DC,$6C,$CC,$F7,$9F,$D9,$72,$F5,$3B,$57,$D7,$FB,$7F,$F1,$8E
	.dc.b	$1F,$AD,$5F,$3D,$93,$5E,$ED,$CB,$F1,$D3,$A6,$CE,$9F,$C9,$7D,$0B
	.dc.b	$D7,$FC,$22,$F9,$4A,$62,$2C,$F0,$8A,$FC,$B9,$E4,$FE,$77,$ED,$EB
	.dc.b	$F3,$DB,$97,$F0,$56,$EB,$FD,$71,$F0,$4F,$1C,$3F,$5F,$1C,$B9,$3F
	.dc.b	$76,$D9,$E1,$3A,$3F,$84,$7E,$04,$F8,$5F,$D3,$61,$FC,$26,$31,$DF
	.dc.b	$FA,$1C,$3C,$38,$7F,$5A,$F1,$1F,$69,$FA,$56,$C7,$ED,$DA,$FE,$25
	.dc.b	$7D,$9E,$E6,$4E,$DE,$51,$FA,$FF,$B7,$7C,$BF,$EC,$7E,$E3,$F0,$77
	.dc.b	$F9,$36,$F0,$7F,$29,$AD,$CF,$F0,$76,$BB,$E0,$E4,$75,$BF,$CA,$7C
	.dc.b	$37,$F7,$49,$FE,$58,$7E,$71,$C5,$FF,$3C,$6E,$FF,$D0,$7C,$CE,$3D
	.dc.b	$9F,$CD,$AF,$83,$7A,$FF,$CF,$AD,$9D,$CF,$69,$CB,$6F,$AB,$FC,$E6
	.dc.b	$13,$C3,$0F,$15,$F5,$BD,$DF,$EC,$AB,$58,$7F,$55,$9C,$E1,$DB,$0F
	.dc.b	$EA,$B5,$AB,$ED,$5A,$37,$19,$8F,$59,$E5,$C3,$D4,$F8,$58,$E4,$F6
	.dc.b	$47,$1B,$F0,$61,$E3,$AE,$35,$BC,$3D,$6D,$9F,$98,$F4,$8C,$92,$33
	.dc.b	$1B,$3D,$33,$B8,$CC,$EE,$72,$6C,$F8,$C6,$CC,$BB,$72,$6E,$BD,$2E
	.dc.b	$A7,$A9,$EB,$75,$87,$C8,$7A,$1F,$22,$BA,$57,$DA,$8C,$BD,$1D,$57
	.dc.b	$47,$B1,$0C,$9F,$83,$C3,$E6,$5F,$D7,$C3,$E5,$E1,$DA,$31,$5F,$48
	.dc.b	$EE,$E2,$3F,$03,$18,$9F,$1D,$74,$6C,$7D,$C3,$FC,$10,$7E,$08,$B8
	.dc.b	$7E,$61,$B7,$FA,$97,$3F,$6F,$FE,$81,$7F,$34,$D7,$FA,$3B,$64,$D3
	.dc.b	$FD,$19,$87,$F8,$A9,$B8,$FF,$8B,$F6,$FE,$87,$39,$F0,$3F,$5D,$5A
	.dc.b	$C3,$BC,$7E,$C6,$B7,$7F,$E3,$7A,$D5,$78,$AB,$89,$BD,$6C,$7E,$D6
	.dc.b	$77,$8F,$9A,$B5,$39,$77,$8F,$8E,$3F,$79,$8F,$81,$F1,$9B,$24,$7F
	.dc.b	$91,$A3,$51,$E3,$33,$8F,$9E,$BB,$D7,$52,$46,$67,$BE,$3E,$28,$C4
	.dc.b	$EA,$7F,$85,$5F,$6C,$3D,$51,$D8,$F2,$CF,$8E,$F9,$46,$2F,$99,$F4
	.dc.b	$63,$31,$E7,$9E,$CF,$F8,$0A,$FB,$63,$F2,$09,$8F,$AB,$0E,$75,$D4
	.dc.b	$87,$1C,$3A,$E3,$E6,$C7,$AB,$FD,$22,$33,$87,$4A,$F8,$27,$AC,$F2
	.dc.b	$5F,$67,$A1,$9A,$E9,$19,$7A,$46,$67,$A5,$FC,$38,$C6,$1E,$4C,$7D
	.dc.b	$AF,$88,$F6,$47,$AA,$FA,$B6,$76,$F2,$3C,$F5,$EC,$BE,$F8,$7C,$D9
	.dc.b	$FA,$99,$BE,$E3,$F4,$B9,$FF,$98,$56,$9F,$8D,$9D,$61,$67,$35,$B8
	.dc.b	$C5,$65,$B1,$FE,$6B,$FB,$8F,$83,$97,$FA,$45,$B9,$7F,$BA,$1F,$EF
	.dc.b	$B3,$BC,$FC,$11,$AC,$FF,$DB,$6D,$CC,$F0,$8F,$F2,$43,$9F,$B8,$FF
	.dc.b	$13,$5F,$FA,$73,$9F,$F8,$0F,$17,$AE,$73,$7F,$E9,$D9,$CD,$7F,$A0
	.dc.b	$67,$2F,$FA,$59,$B3,$FE,$FE,$6B,$FE,$EE,$6E,$07,$AE,$7F,$EA,$F6
	.dc.b	$E7,$C7,$A7,$C5,$D7,$F3,$7F,$0F,$FF,$53,$5F,$FA,$D5,$B5,$FE,$AD
	.dc.b	$C1,$F7,$3C,$B2,$FF,$7E,$D6,$9F,$F7,$FC,$66,$53,$97,$8E,$1D,$EB
	.dc.b	$EE,$7D,$CE,$C9,$96,$27,$2D,$D3,$FD,$49,$3F,$E5,$36,$CF,$F0,$CF
	.dc.b	$73,$FC,$53,$1F,$E1,$33,$1D,$1F,$C0,$DA,$7F,$9C,$23,$78,$7F,$8B
	.dc.b	$83,$88,$FE,$86,$D8,$FF,$66,$3C,$35,$6B,$8B,$F1,$7B,$7D,$AF,$FD
	.dc.b	$15,$F8,$B3,$D4,$FC,$D3,$72,$FF,$48,$7D,$D6,$27,$F3,$DE,$52,$3F
	.dc.b	$D0,$73,$FE,$D6,$8C,$3F,$EA,$96,$DF,$F9,$1B,$81,$18,$9F,$F6,$EF
	.dc.b	$DD,$3F,$0B,$E7,$7F,$6E,$31,$3F,$EB,$36,$C7,$36,$C7,$2F,$BB,$FD
	.dc.b	$94,$6B,$7F,$F7,$D2,$8C,$DF,$ED,$FE,$ED,$F5,$3F,$B9,$F4,$5E,$B6
	.dc.b	$63,$0D,$CB,$FD,$89,$5C,$1B,$3E,$8C,$DA,$AB,$51,$F0,$5B,$A7,$D4
	.dc.b	$8D,$C6,$38,$FF,$34,$FF,$B5,$70,$FF,$77,$39,$9F,$F7,$4F,$B3,$FC
	.dc.b	$3F,$D6,$3F,$C4,$7D,$9F,$CD,$46,$2F,$FD,$ED,$0A,$F3,$64,$8C,$4F
	.dc.b	$FB,$7D,$B3,$F5,$DB,$1C,$63,$DB,$E9,$39,$B9,$B6,$7F,$DD,$0D,$9F
	.dc.b	$C0,$36,$39,$5A,$1F,$09,$FC,$A5,$E1,$AF,$C5,$9D,$4E,$70,$FD,$57
	.dc.b	$C8,$FE,$AE,$2E,$1F,$E7,$FE,$AE,$6D,$9F,$D4,$65,$F0,$65,$18,$8F
	.dc.b	$D5,$B6,$33,$3F,$E2,$E9,$E1,$E6,$E4,$91,$98,$FD,$9C,$76,$C5,$F8
	.dc.b	$D3,$EB,$30,$DB,$E9,$3A,$9C,$9C,$7D,$F7,$C2,$52,$57,$08,$CB,$CB
	.dc.b	$E0,$5F,$87,$78,$5B,$19,$8C,$1E,$5F,$47,$FD,$F7,$A7,$E5,$9B,$97
	.dc.b	$F9,$8E,$BC,$B1,$8F,$B4,$FE,$83,$5F,$61,$AF,$F2,$F6,$BF,$A0,$6D
	.dc.b	$3E,$1B,$F6,$F3,$F8,$06,$BF,$CD,$DA,$FE,$A1,$B0,$F8,$77,$E0,$DF
	.dc.b	$35,$B8,$FC,$A6,$DF,$F9,$92,$B8,$57,$D3,$F8,$AF,$C1,$CE,$5A,$FE
	.dc.b	$49,$AF,$F4,$96,$BF,$BA,$E2,$EA,$FE,$61,$AF,$F4,$F6,$BB,$9F,$C1
	.dc.b	$7C,$5B,$8E,$93,$2F,$C0,$C6,$5F,$C1,$FC,$6B,$E6,$FF,$14,$D8,$CC
	.dc.b	$60,$FC,$77,$C7,$BE,$3E,$92,$7F,$1B,$19,$7F,$1F,$D7,$FC,$CB,$3F
	.dc.b	$D8,$B5,$FE,$D0,$D7,$F8,$CB,$76,$7C,$83,$F0,$0F,$CF,$36,$FF,$36
	.dc.b	$DB,$FA,$4D,$7F,$84,$2F,$AF,$7D,$FF,$ED,$23,$0D,$8C,$D7,$EC,$67
	.dc.b	$2D,$F9,$6F,$9F,$06,$93,$87,$A7,$81,$F3,$71,$FF,$53,$D7,$F1,$7E
	.dc.b	$2C,$93,$0E,$DF,$2F,$FC,$9E,$B6,$D7,$D2,$D7,$F9,$55,$23,$11,$F1
	.dc.b	$BE,$61,$F1,$9E,$D6,$DF,$AD,$B3,$E4,$6B,$F1,$BE,$67,$F0,$0F,$A1
	.dc.b	$AE,$79,$3E,$44,$62,$3C,$7F,$5B,$FE,$06,$7E,$7B,$5C,$FA,$DF,$3E
	.dc.b	$31,$1E,$7F,$82,$F3,$13,$1F,$33,$F5,$7E,$B3,$B4,$FE,$09,$BA,$9A
	.dc.b	$63,$C7,$FE,$53,$3F,$41,$AF,$E0,$9A,$FC,$A6,$BF,$8E,$F9,$DF,$86
	.dc.b	$FC,$E6,$B9,$DB,$F1,$11,$8B,$7C,$F3,$E3,$3F,$18,$DC,$FF,$14,$DC
	.dc.b	$3F,$06,$D7,$F2,$BF,$3F,$FE,$7C,$7F,$0E,$D7,$3F,$0F,$F9,$78,$C4
	.dc.b	$7F,$63,$F7,$1F,$84,$7F,$4B,$6E,$1F,$D1,$5B,$3F,$D2,$DA,$FF,$BE
	.dc.b	$BE,$AF,$FB,$F3,$FD,$3D,$AE,$7E,$5F,$FA,$B4,$62,$3F,$B9,$2F,$98
	.dc.b	$C9,$2F,$FA,$A4,$65,$FE,$AD,$F8,$A3,$CC,$FE,$C0,$D8,$CC,$60,$FE
	.dc.b	$C9,$F5,$5F,$DE,$70,$FF,$00,$07,$0F,$F8,$0D,$AF,$F8,$35,$AF,$F8
	.dc.b	$1F,$C4,$FE,$CC,$FF,$84,$5A,$9A,$7F,$6B,$F3,$DF,$54,$FA,$AC,$36
	.dc.b	$BF,$BA,$4E,$A7,$27,$1F,$AA,$7D,$55,$25,$70,$8C,$BE,$CF,$8E,$7C
	.dc.b	$9F,$F6,$A8,$FF,$13,$53,$5F,$F1,$45,$6B,$D5,$FF,$11,$4E,$69,$39
	.dc.b	$78,$D7,$F8,$9F,$E9,$5F,$17,$C2,$39,$E2,$FA,$0C,$5F,$28,$C3,$C8
	.dc.b	$C4,$6B,$0E,$D1,$EE,$AF,$F1,$A7,$CA,$FC,$C2,$47,$D0,$7E,$8C,$FB
	.dc.b	$BE,$DF,$96,$06,$9E,$2A,$FF,$6F,$1A,$F8,$88,$F8,$EB,$E3,$C7,$08
	.dc.b	$F3,$7C,$BB,$EB,$BE,$24,$79,$44,$67,$2F,$7B,$63,$8C,$BE,$6B,$3E
	.dc.b	$43,$C9,$A4,$AD,$E1,$CB,$D0,$7E,$0B,$49,$3E,$48,$FF,$4E,$3E,$6F
	.dc.b	$C2,$BC,$BC,$9F,$67,$44,$63,$0F,$F3,$34,$61,$E7,$5E,$B8,$7E,$17
	.dc.b	$FD,$C6,$3F,$C8,$D1,$8F,$6B,$FC,$88,$D8,$CC,$ED,$FD,$5C,$7F,$49
	.dc.b	$F7,$7E,$1F,$A7,$FE,$BC,$9F,$96,$D7,$C1,$5C,$72,$FC,$87,$96,$F4
	.dc.b	$71,$9F,$A3,$EC,$FF,$D3,$78,$7D,$46,$BF,$4E,$B8,$69,$F4,$BF,$2A
	.dc.b	$7D,$3F,$A1,$B5,$EC,$6B,$E8,$6B,$FE,$83,$D9,$F8,$6E,$B4,$F9,$8D
	.dc.b	$CF,$E5,$B6,$BF,$06,$FF,$A8,$5F,$99,$F8,$63,$ED,$7F,$0A,$BE,$BE
	.dc.b	$98,$8F,$C4,$FD,$87,$F3,$FA,$FA,$2D,$7E,$6B,$5F,$33,$5F,$F5,$17
	.dc.b	$DE,$FF,$87,$3D,$3F,$D8,$B6,$3F,$60,$D7,$F3,$4F,$F8,$8B,$DA,$DE
	.dc.b	$7D,$F4,$FE,$05,$3F,$C1,$FE,$BB,$26,$BD,$7E,$BF,$8B,$7F,$A4,$31
	.dc.b	$FF,$46,$52,$31,$1F,$CB,$A3,$1E,$DF,$FA,$35,$B1,$C3,$EF,$1E,$1F
	.dc.b	$A1,$B1,$E8,$6B,$E8,$7F,$7C,$FB,$3E,$F7,$FD,$AA,$DF,$A0,$F8,$31
	.dc.b	$FE,$FB,$8D,$7C,$47,$C4,$FC,$3F,$FD,$E9,$1E,$86,$BE,$76,$BF,$EC
	.dc.b	$5C,$51,$88,$F9,$FF,$7F,$C3,$7F,$8E,$DE,$0C,$FC,$FF,$D2,$87,$D7
	.dc.b	$FD,$4F,$8F,$FE,$47,$87,$D2,$7F,$9B,$37,$D3,$D7,$A3,$F6,$BE,$B3
	.dc.b	$EF,$4B,$54,$61,$BC,$BF,$36,$31,$F8,$07,$FB,$AA,$BF,$02,$D7,$F0
	.dc.b	$2D,$DB,$EF,$BF,$81,$DB,$F1,$7F,$86,$FE,$BA,$7F,$06,$FF,$99,$EB
	.dc.b	$F1,$6D,$7F,$16,$EE,$7F,$27,$F6,$3F,$46,$7F,$04,$D7,$E6,$35,$DF
	.dc.b	$27,$E1,$7F,$0D,$FD,$44,$7E,$61,$AF,$E6,$1B,$87,$E5,$9F,$D6,$47
	.dc.b	$E7,$BF,$05,$FE,$DE,$DF,$C6,$3F,$47,$06,$FB,$DF,$A6,$FC,$23,$F8
	.dc.b	$0D,$7E,$A9,$AF,$CF,$6F,$01,$BF,$87,$3D,$DF,$D7,$B6,$BF,$5E,$D7
	.dc.b	$F5,$EF,$FB,$01,$FE,$C4,$7F,$76,$F8,$6F,$F0,$67,$D3,$A2,$7B,$AF
	.dc.b	$A9,$CC,$61,$AF,$13,$FC,$25,$3C,$5F,$F0,$0C,$5E,$2F,$F0,$FD,$73
	.dc.b	$9E,$D1,$89,$D4,$71,$9D,$3F,$D8,$0E,$67,$63,$CD,$F8,$77,$CE,$7F
	.dc.b	$C4,$17,$E1,$8C,$7C,$2F,$F2,$3A,$E6,$A7,$F8,$BF,$E1,$9F,$A0,$8C
	.dc.b	$5F,$E7,$61,$F8,$47,$C0,$DD,$18,$9F,$14,$63,$B3,$53,$99,$F1,$9C
	.dc.b	$EB,$26,$A7,$A7,$E1,$DF,$E4,$2C,$3C,$38,$23,$18,$FF,$2C,$6E,$0F
	.dc.b	$0D,$F5,$3F,$E0,$28,$F1,$CE,$AD,$F8,$9B,$EF,$3D,$4D,$AD,$9F,$73
	.dc.b	$0F,$BD,$E0,$FD,$EF,$85,$7B,$CF,$C2,$F0,$DB,$FA,$57,$C1,$FA,$95
	.dc.b	$C2,$B6,$48,$DF,$27,$F1,$0D,$FE,$7D,$6E,$78,$E1,$D5,$F9,$31,$97
	.dc.b	$E5,$E1,$DD,$CF,$E2,$3F,$05,$BE,$C7,$A6,$31,$87,$DC,$AF,$E7,$BF
	.dc.b	$25,$9A,$F5,$FD,$94,$C3,$D3,$87,$0E,$3F,$D1,$60,$FB,$4F,$EE,$F9
	.dc.b	$7D,$6F,$AC,$FF,$00,$B7,$EC,$FE,$54,$EA,$FF,$B1,$D6,$9B,$3F,$8B
	.dc.b	$8C,$63,$F6,$EB,$83,$FB,$0F,$9C,$FD,$6F,$5F,$FB,$82,$3F,$40,$7F
	.dc.b	$9B,$23,$3F,$61,$A7,$F5,$31,$C4,$FD,$69,$FA,$B7,$2F,$13,$26,$CF
	.dc.b	$CB,$8E,$DF,$EB,$6E,$1B,$9F,$EA,$E7,$F1,$1F,$7F,$F8,$BA,$FC,$FF
	.dc.b	$4B,$F0,$0F,$B5,$7F,$9C,$6E,$DF,$32,$F8,$4E,$79,$FF,$5D,$5A,$C3
	.dc.b	$FA,$73,$78,$17,$C3,$0F,$AB,$E8,$FF,$A0,$AF,$F2,$6D,$7F,$0E,$DC
	.dc.b	$7F,$04,$BE,$E7,$F4,$0E,$6B,$FD,$27,$EE,$FF,$3A,$BF,$C8,$B6,$F8
	.dc.b	$FC,$A7,$E9,$FF,$00,$E5,$5F,$A0,$F4,$3F,$C4,$D5,$F9,$C6,$E1,$F8
	.dc.b	$AA,$DD,$7E,$69,$B5,$BF,$53,$F6,$17,$C2,$BF,$7D,$5A,$BF,$F5,$3C
	.dc.b	$E7,$B0,$D3,$86,$CA,$E1,$7F,$C1,$FE,$97,$FD,$04,$FE,$E1,$AF,$EB
	.dc.b	$1B,$97,$E9,$DF,$EF,$B1,$BF,$93,$FE,$45,$F6,$2F,$AA,$A7,$F0,$9A
	.dc.b	$E3,$A7,$EA,$3E,$FB,$FD,$1D,$87,$F0,$6F,$C0,$3F,$D3,$39,$7E,$CD
	.dc.b	$B5,$BF,$96,$FE,$25,$18,$AF,$E3,$CD,$D4,$DC,$3A,$CF,$E6,$3B,$1F
	.dc.b	$0D,$FE,$32,$CB,$FA,$03,$72,$FE,$98,$DA,$FE,$BB,$F3,$FF,$AE,$8C
	.dc.b	$47,$FB,$61,$B7,$D3,$2F,$D8,$BF,$EF,$F9,$B3,$A2,$4E,$AB,$A3,$A2
	.dc.b	$4F,$9E,$32,$FA,$67,$8B,$93,$78,$7A,$AA,$E1,$FD,$96,$F8,$CE,$6B
	.dc.b	$95,$78,$5D,$9F,$DE,$AF,$AC,$FF,$63,$AE,$37,$F1,$2F,$77,$FE,$F8
	.dc.b	$9C,$BC,$63,$38,$7C,$18,$EE,$4B,$F1,$1D,$5D,$E1,$E0,$C3,$D1,$18
	.dc.b	$BE,$78,$76,$8F,$A1,$18,$74,$75,$9E,$2D,$3C,$85,$3D,$07,$0B,$F1
	.dc.b	$C7,$06,$C7,$06,$BC,$0E,$58,$7D,$F6,$DF,$96,$31,$3C,$1F,$84,$DC
	.dc.b	$3D,$D5,$F8,$C3,$C5,$7F,$35,$FB,$87,$68,$DB,$E1,$9E,$58,$7C,$18
	.dc.b	$F9,$A7,$F8,$A2,$7F,$99,$BC,$CC,$47,$F8,$BD,$FF,$14,$72,$FF,$63
	.dc.b	$6F,$89,$97,$FC,$77,$19,$6C,$65,$62,$9F,$F0,$66,$1F,$E2,$96,$E3
	.dc.b	$FE,$35,$F7,$5E,$8F,$E9,$27,$38,$74,$3F,$68,$E1,$EB,$3C,$1C,$E3
	.dc.b	$C4,$8F,$EE,$E3,$58,$78,$67,$2D,$9D,$5E,$30,$F7,$C7,$43,$25,$9E
	.dc.b	$04,$8F,$54,$F1,$9F,$44,$E5,$FF,$30,$DF,$8A,$BA,$CE,$67,$E3,$E1
	.dc.b	$ED,$C7,$FD,$47,$87,$39,$CE,$36,$73,$1E,$3C,$39,$91,$EF,$5B,$BE
	.dc.b	$87,$5C,$7C,$53,$E6,$AE,$0F,$63,$DF,$8F,$97,$0F,$B5,$8F,$86,$F8
	.dc.b	$D6,$EB,$58,$75,$C7,$C1,$7B,$6D,$79,$CF,$29,$4E,$58,$F8,$B1,$F6
	.dc.b	$E1,$D2,$F9,$56,$E7,$C6,$DC,$7C,$71,$8A,$E1,$3E,$AB,$E9,$8F,$C2
	.dc.b	$C3,$F9,$8E,$1C,$23,$15,$CF,$1F,$3E,$1F,$06,$B5,$78,$C3,$EF,$DF
	.dc.b	$AF,$27,$CD,$FF,$48,$C7,$FC,$B6,$D9,$FF,$27,$BF,$A3,$8C,$E9,$F4
	.dc.b	$AB,$8E,$76,$B8,$67,$FE,$4F,$78,$61,$CA,$F8,$7E,$01,$F1,$B2,$FF
	.dc.b	$57,$78,$0F,$A4,$FF,$1B,$E3,$FE,$57,$AD,$61,$FE,$EF,$9C,$DF,$F9
	.dc.b	$52,$73,$99,$A3,$9F,$BE,$FF,$88,$EA,$FD,$E7,$1D,$DC,$FA,$FF,$00
	.dc.b	$76,$E6,$30,$DE,$2F,$F3,$FC,$E9,$FF,$3F,$D7,$27,$33,$BD,$BF,$CE
	.dc.b	$F5,$B7,$FC,$57,$7F,$27,$EF,$FD,$70,$C5,$F8,$6F,$FA,$C3,$1F,$F4
	.dc.b	$6D,$6A,$73,$EB,$71,$B9,$5F,$E9,$2F,$C0,$7B,$71,$FF,$9F,$F0,$EB
	.dc.b	$C9,$FE,$53,$F8,$7F,$D4,$B6,$FF,$D3,$3F,$80,$F0,$E5,$FA,$7E,$3A
	.dc.b	$FE,$F4,$E7,$3F,$F2,$9C,$63,$87,$9D,$6F,$0F,$F0,$29,$FA,$EF,$19
	.dc.b	$F8,$1F,$E7,$A3,$0D,$7F,$CE,$38,$7F,$30,$7F,$92,$1F,$EE,$62,$BF
	.dc.b	$EE,$36,$BF,$EF,$47,$93,$FE,$78,$F4,$DE,$0E,$37,$FF,$4B,$7D,$A7
	.dc.b	$C0,$F4,$0F,$E6,$7F,$02,$7D,$66,$63,$0D,$CB,$FD,$7B,$3A,$7F,$DB
	.dc.b	$F5,$C9,$FF,$53,$FD,$F7,$E3,$7E,$2B,$FD,$A9,$18,$3F,$D8,$6E,$34
	.dc.b	$FF,$5B,$CF,$E4,$FD,$63,$FA,$BF,$C0,$BE,$54,$EA,$FF,$D9,$7F,$89
	.dc.b	$BE,$67,$F3,$4D,$8F,$F6,$8F,$CA,$3D,$DC,$D4,$9E,$2F,$ED,$E7,$51
	.dc.b	$83,$8C,$EE,$FF,$D9,$B3,$A7,$FD,$A3,$7D,$DF,$F4,$EF,$C0,$BC,$BF
	.dc.b	$C5,$FC,$03,$EA,$BF,$1B,$39,$C9,$7F,$EA,$4A,$D4,$7F,$BD,$AF,$67
	.dc.b	$F0,$BF,$C0,$5F,$81,$62,$78,$F3,$3B,$FF,$83,$FC,$3B,$E6,$3F,$36
	.dc.b	$DA,$DC,$EB,$D5,$CD,$69,$B3,$AD,$1E,$1B,$05,$6A,$7F,$19,$F8,$6F
	.dc.b	$E6,$FE,$17,$FA,$4F,$7A,$31,$8F,$FB,$52,$30,$FE,$71,$B5,$FD,$FC
	.dc.b	$F8,$6E,$23,$FC,$30,$D8,$F8,$E1,$FC,$2C,$75,$9F,$0F,$47,$89,$FD
	.dc.b	$DF,$E0,$0F,$A2,$FF,$06,$7C,$A7,$F8,$37,$54,$E7,$0F,$A7,$39,$8F
	.dc.b	$99,$18,$3F,$99,$BB,$C9,$3C,$FE,$41,$F8,$7F,$AE,$7F,$9C,$C6,$2F
	.dc.b	$FC,$5D,$82,$31,$3F,$E3,$58,$C5,$71,$8F,$A3,$F2,$8F,$C0,$FD,$38
	.dc.b	$7F,$44,$87,$D6,$E2,$B7,$57,$05,$6A,$FD,$55,$F7,$3D,$2F,$14,$62
	.dc.b	$BF,$96,$E5,$C6,$7C,$B9,$FC,$77,$77,$F0,$2F,$FA,$AF,$C1,$FF,$BE
	.dc.b	$DF,$6B,$63,$D8,$D7,$EC,$E1,$FE,$D4,$F3,$3D,$CB,$C5,$FD,$56,$3F
	.dc.b	$57,$A3,$F8,$23,$F5,$E3,$38,$FF,$64,$9C,$BF,$D9,$BE,$65,$F4,$1F
	.dc.b	$05,$B1,$ED,$5E,$1E,$67,$E3,$7B,$3A,$D3,$E9,$35,$F3,$B5,$FE,$E6
	.dc.b	$D7,$F4,$BF,$37,$FD,$4D,$F8,$86,$3A,$FE,$94,$67,$4F,$C8,$3F,$AA
	.dc.b	$CB,$E8,$7C,$E7,$F2,$9B,$E3,$B2,$31,$7F,$97,$5E,$8F,$DD,$73,$7F
	.dc.b	$E0,$6F,$9D,$7C,$57,$CC,$6C,$7E,$41,$B8,$FE,$3D,$AF,$D4,$F9,$EF
	.dc.b	$BB,$8F,$E9,$5F,$C0,$D7,$F8,$49,$CB,$7A,$BF,$1D,$F3,$FE,$D8,$FF
	.dc.b	$0B,$B6,$FF,$84,$BE,$F7,$DA,$E7,$47,$D0,$7E,$EA,$7F,$44,$D7,$F0
	.dc.b	$ED,$7F,$C4,$6D,$7F,$8A,$D7,$33,$B6,$3F,$BD,$71,$BF,$E8,$FE,$9D
	.dc.b	$F3,$34,$C7,$9F,$6B,$F4,$BE,$5B,$E8,$9F,$61,$FB,$38,$C3,$63,$33
	.dc.b	$FE,$A9,$9C,$B7,$F0,$FF,$8F,$A7,$F4,$37,$E5,$B7,$C0,$FD,$C5,$27
	.dc.b	$2B,$E9,$F8,$B6,$77,$1A,$33,$F5,$1F,$C5,$AF,$FD,$7D,$48,$C4,$7F
	.dc.b	$28,$8C,$77,$7A,$E7,$33,$FD,$4B,$67,$DE,$3E,$63,$CB,$F5,$7F,$DE
	.dc.b	$EB,$F3,$B5,$A9,$CF,$BD,$EA,$FA,$E7,$E0,$74,$F6,$FD,$7F,$F5,$8E
	.dc.b	$1F,$9B,$3A,$7E,$7C,$67,$1F,$4B,$5F,$43,$5F,$EC,$FB,$23,$11,$FD
	.dc.b	$BF,$EC,$1F,$2D,$F3,$1B,$7F,$73,$F0,$FF,$EE,$15,$FB,$6C,$15,$A8
	.dc.b	$C4,$EB,$AB,$FD,$A5,$48,$C5,$F8,$97,$E1,$3F,$69,$CD,$FF,$80,$7C
	.dc.b	$4F,$F4,$36,$1C,$9B,$7E,$AF,$B5,$FF,$3E,$3F,$90,$8C,$D7,$F8,$26
	.dc.b	$72,$F9,$DB,$5E,$47,$EB,$57,$E6,$1C,$BF,$E0,$CF,$8F,$FF,$7B,$FC
	.dc.b	$47,$F4,$56,$C7,$F4,$25,$F0,$5F,$EF,$68,$C7,$CC,$FF,$91,$9D,$46
	.dc.b	$71,$FE,$3B,$5C,$27,$FA,$84,$3E,$6B,$ED,$65,$FE,$F9,$AD,$57,$F5
	.dc.b	$0A,$D7,$C0,$7F,$5D,$6C,$7F,$7C,$7F,$BE,$DF,$DC,$72,$66,$98,$FF
	.dc.b	$82,$30,$EC,$DB,$F8,$0B,$F0,$0F,$80,$E6,$FF,$D1,$4E,$67,$DF,$3E
	.dc.b	$AF,$BE,$FF,$52,$46,$0F,$45,$7F,$8B,$8F,$31,$FE,$12,$F8,$2A,$D2
	.dc.b	$87,$3F,$72,$FA,$F7,$DC,$E1,$B5,$D2,$75,$39,$38,$FC,$4F,$F8,$3A
	.dc.b	$74,$D7,$E4,$4F,$F9,$37,$4D,$9E,$6A,$7E,$B7,$FD,$82,$B0,$D7,$E8
	.dc.b	$46,$BD,$7F,$5B,$63,$F8,$E7,$E0,$1F,$E5,$B6,$F4,$FC,$01,$19,$8F
	.dc.b	$E2,$2D,$7E,$E7,$B5,$FD,$D1,$F8,$ED,$7E,$5B,$5D,$FB,$3E,$2C,$7F
	.dc.b	$CB,$F1,$E8,$E5,$C1,$1F,$A7,$F3,$DF,$C5,$FF,$01,$FA,$6D,$8C,$F8
	.dc.b	$89,$CB,$63,$F9,$C7,$5B,$DB,$F6,$36,$33,$18,$3E,$07,$6F,$E6,$76
	.dc.b	$FE,$E7,$CD,$82,$57,$F4,$3F,$BB,$FE,$EE,$DA,$F9,$11,$88,$FE,$6E
	.dc.b	$D7,$F9,$B7,$83,$FE,$F5,$7F,$A6,$35,$FE,$7A,$D7,$7F,$3F,$F9,$0F
	.dc.b	$C9,$3F,$23,$06,$91,$FC,$AF,$E2,$4F,$A6,$A4,$8F,$CD,$C6,$5F,$CE
	.dc.b	$F7,$BE,$83,$F3,$AD,$8C,$C6,$0C,$F9,$72,$38,$B5,$FD,$4D,$3E,$6F
	.dc.b	$C6,$48,$F0,$BF,$8D,$F5,$1C,$BF,$E8,$0E,$0F,$EB,$51,$8A,$FF,$54
	.dc.b	$C6,$27,$51,$FD,$1A,$74,$FF,$A3,$78,$7F,$BE,$74,$6A,$3F,$D6,$DE
	.dc.b	$AF,$E4,$F0,$FE,$CC,$D8,$FE,$9E,$D7,$FD,$47,$E7,$FF,$4E,$7C,$27
	.dc.b	$FA,$EB,$DB,$46,$63,$FD,$39,$18,$F3,$7C,$0D,$8E,$27,$F8,$6B,$C0
	.dc.b	$CE,$1F,$B5,$3F,$CC,$E5,$8E,$CD,$4B,$F0,$C7,$3F,$B6,$30,$D8,$CC
	.dc.b	$F8,$27,$2F,$8A,$DE,$EB,$EC,$BC,$ED,$9D,$35,$F6,$B6,$BF,$C9,$5F
	.dc.b	$1B,$EE,$F8,$1F,$F5,$85,$75,$6B,$F1,$1B,$3F,$EB,$CF,$35,$FB,$93
	.dc.b	$EB,$BE,$B4,$61,$B1,$9B,$F9,$33,$97,$D5,$F0,$DF,$EA,$76,$FC,$6F
	.dc.b	$D0,$8C,$47,$D4,$6E,$1E,$1B,$FE,$6D,$3D,$1A,$FF,$B5,$6F,$CB,$E7
	.dc.b	$78,$8F,$1F,$DA,$49,$E0,$D7,$87,$E0,$3E,$76,$E7,$EC,$8C,$63,$F0
	.dc.b	$5B,$3F,$8B,$F2,$FF,$5C,$FE,$29,$B3,$FE,$AA,$6C,$6F,$BF,$F1,$BF
	.dc.b	$1B,$F3,$23,$FD,$6B,$C1,$82,$3F,$7F,$E3,$BF,$4B,$87,$FA,$D1,$B1
	.dc.b	$F9,$B6,$BF,$EC,$7E,$EF,$CC,$FC,$7B,$ED,$69,$27,$F1,$51,$97,$F1
	.dc.b	$98,$BE,$0F,$CA,$48,$CB,$FA,$4F,$B8,$7D,$ED,$24,$FE,$56,$32,$FE
	.dc.b	$A0,$78,$BF,$EE,$24,$7F,$71,$AA,$EB,$E4,$FF,$BC,$46,$27,$84,$EC
	.dc.b	$72,$B5,$7C,$AF,$09,$EB,$5C,$37,$FF,$0C,$36,$3A,$63,$FC,$35,$F0
	.dc.b	$1E,$78,$EF,$ED,$7C,$38,$F3,$F9,$EF,$8B,$ED,$1F,$EB,$E6,$D7,$0D
	.dc.b	$3A,$CF,$BA,$F9,$3B,$31,$1F,$02,$FA,$46,$0F,$4E,$1D,$0E,$F1,$A8
	.dc.b	$F8,$47,$32,$DF,$85,$F3,$E1,$C0,$8F,$53,$DE,$72,$9E,$98,$79,$9B
	.dc.b	$87,$96,$BC,$55,$C6,$72,$F6,$31,$1A,$7A,$3F,$36,$3D,$18,$FF,$92
	.dc.b	$5B,$3D,$4F,$C3,$3D,$1F,$F2,$4F,$D6,$FC,$EF,$34,$FA,$A3,$A9,$EB
	.dc.b	$8C,$B6,$35,$8F,$AF,$0E,$51,$27,$CE,$6C,$87,$CC,$AE,$6F,$63,$11
	.dc.b	$C5,$A7,$22,$92,$BE,$FC,$F2,$7C,$51,$DB,$0F,$F3,$54,$F6,$6D,$7E
	.dc.b	$4E,$31,$3F,$A4,$79,$63,$7C,$6F,$88,$78,$F0,$EF,$78,$BF,$89,$39
	.dc.b	$AE,$53,$99,$F7,$4E,$5F,$D9,$DF,$F6,$3B,$F4,$CE,$DE,$B3,$D0,$87
	.dc.b	$23,$E4,$BC,$1C,$B5,$F8,$97,$E9,$7F,$82,$57,$F7,$1C,$3E,$1D,$FC
	.dc.b	$12,$4F,$DD,$FA,$EF,$F0,$56,$1F,$31,$B8,$73,$9F,$11,$F6,$63,$2F
	.dc.b	$B1,$AF,$FA,$52,$BD,$93,$F3,$67,$38,$1B,$3F,$91,$72,$4E,$6F,$6D
	.dc.b	$71,$7C,$C4,$BF,$1E,$1E,$78,$C3,$63,$33,$C5,$F4,$4E,$A3,$D9,$3B
	.dc.b	$23,$B7,$E6,$18,$9F,$74,$F1,$7E,$DD,$FA,$63,$11,$FE,$B4,$BF,$6B
	.dc.b	$8C,$3E,$04,$DA,$E1,$3E,$42,$5F,$83,$0F,$11,$23,$C2,$7B,$0B,$1E
	.dc.b	$57,$C1,$5F,$21,$CC,$F1,$72,$71,$2C,$F1,$BF,$2C,$65,$B1,$9C,$3D
	.dc.b	$73,$E6,$AE,$95,$A7,$E5,$98,$9F,$B5,$3C,$DB,$5D,$F0,$F8,$2F,$5C
	.dc.b	$3D,$15,$E4,$BE,$51,$8C,$3B,$4F,$B0,$87,$E3,$CF,$9D,$87,$6A,$F1
	.dc.b	$C6,$A7,$8B,$D4,$FB,$27,$0A,$FA,$8D,$3D,$C6,$8A,$71,$24,$7E,$EE
	.dc.b	$7C,$F1,$A8,$C3,$F6,$0F,$E2,$75,$FE,$72,$7F,$28,$62,$BF,$D1,$BF
	.dc.b	$5B,$FE,$C7,$8F,$E0,$ED,$7F,$DF,$ED,$A3,$4F,$F6,$A6,$1F,$C9,$1B
	.dc.b	$87,$F1,$DF,$53,$FC,$CB,$7F,$CC,$67,$2F,$6F,$37,$F9,$F4,$62,$18
	.dc.b	$7F,$B0,$F4,$CD,$49,$E4,$FF,$45,$BE,$B8,$F8,$63,$4D,$9F,$E8,$13
	.dc.b	$AC,$7F,$98,$63,$DF,$1F,$5E,$1C,$27,$D1,$8D,$C3,$CC,$78,$0E,$6E
	.dc.b	$5F,$8B,$36,$BD,$11,$BC,$7C,$44,$BE,$B1,$97,$C7,$5C,$71,$F3,$4F
	.dc.b	$82,$B8,$36,$B8,$C6,$23,$C2,$73,$3F,$9F,$CF,$27,$31,$BB,$F7,$C7
	.dc.b	$A8,$E0,$6A,$BC,$6E,$67,$86,$3F,$1C,$F9,$17,$D6,$F9,$CE,$B0,$FA
	.dc.b	$98,$FC,$3B,$E1,$18,$C3,$BD,$F0,$C7,$BE,$1F,$DB,$31,$E9,$87,$0B
	.dc.b	$DC,$63,$0E,$87,$3A,$F0,$4E,$B1,$CB,$4E,$F7,$CE,$7B,$4E,$9F,$89
	.dc.b	$7D,$6B,$77,$D7,$0F,$29,$23,$E9,$1F,$3A,$F9,$46,$63,$F3,$35,$CE
	.dc.b	$FB,$C7,$8D,$EF,$5F,$7C,$91,$9A,$E5,$8F,$E2,$A7,$FC,$17,$7F,$AA
	.dc.b	$69,$FD,$B9,$C5,$7E,$12,$A3,$FE,$00,$F4,$FE,$B4,$CA,$F6,$4E,$5A
	.dc.b	$FA,$1B,$6A,$E1,$3F,$E2,$96,$E1,$FE,$22,$6E,$1F,$E1,$B6,$CF,$53
	.dc.b	$F1,$51,$9B,$EB,$3A,$9F,$DF,$BE,$A7,$84,$7F,$8A,$A3,$11,$A8,$FF
	.dc.b	$06,$C7,$E4,$4F,$CF,$70,$7F,$BC,$A3,$FC,$91,$19,$69,$FE,$5D,$31
	.dc.b	$3E,$B7,$2F,$F9,$9D,$CD,$FF,$99,$DB,$7F,$E7,$B6,$CF,$FA,$21,$AA
	.dc.b	$FC,$4F,$53,$A4,$76,$AE,$8D,$AE,$8F,$F9,$98,$F7,$D7,$27,$FC,$D8
	.dc.b	$60,$F7,$9F,$BD,$E1,$78,$98,$8F,$71,$30,$FF,$44,$C6,$5F,$19,$30
	.dc.b	$FF,$4C,$D7,$2A,$E1,$18,$AE,$0D,$8C,$E1,$FE,$B9,$9C,$B6,$33,$7F
	.dc.b	$EB,$69,$CB,$63,$2F,$FB,$22,$3E,$ED,$F8,$63,$FA,$39,$B2,$9B,$9F
	.dc.b	$F5,$D4,$FC,$13,$F9,$3F,$33,$E2,$1E,$8E,$6B,$FD,$61,$E5,$3E,$61
	.dc.b	$F7,$BF,$8F,$8C,$36,$33,$3F,$ED,$A3,$FD,$B3,$3A,$7F,$DD,$7F,$1F
	.dc.b	$C7,$BC,$D2,$47,$FB,$CB,$E5,$BB,$78,$26,$92,$3B,$6D,$FE,$FF,$E0
	.dc.b	$F2,$77,$FF,$BD,$E7,$32,$CB,$FD,$C9,$5A,$3C,$5F,$2B,$2B,$FD,$FE
	.dc.b	$DA,$F9,$74,$F1,$BF,$DF,$B5,$D2,$31,$E1,$FC,$39,$CD,$70,$6D,$7F
	.dc.b	$BE,$B8,$9F,$BC,$7A,$E3,$CB,$18,$D3,$F7,$4D,$8C,$CE,$EF,$EE,$CB
	.dc.b	$F1,$77,$DE,$F0,$6D,$F4,$6D,$70,$6C,$78,$D7,$93,$C4,$91,$C5,$FB
	.dc.b	$DE,$7B,$C6,$A4,$9F,$85,$19,$7C,$1F,$7B,$E3,$23,$83,$E4,$7A,$3F
	.dc.b	$57,$1F,$9A,$D8,$FC,$3B,$60,$7D,$97,$0F,$DB,$F4,$9F,$7B,$F9,$36
	.dc.b	$D7,$D5,$6C,$FD,$A6,$C7,$48,$2B,$FA,$F1,$F9,$96,$BA,$6C,$6F,$87
	.dc.b	$F4,$4F,$49,$F3,$F1,$74,$C3,$24,$9F,$C4,$7B,$0F,$BB,$FD,$53,$63
	.dc.b	$31,$83,$F9,$E7,$8C,$F6,$9E,$D6,$1B,$5F,$9E,$9D,$4E,$4E,$3E,$E7
	.dc.b	$C8,$83,$49,$D5,$FC,$2B,$D5,$FE,$0F,$AF,$D0,$F1,$51,$88,$46,$31
	.dc.b	$F9,$51,$F2,$E3,$0E,$07,$D0,$FD,$27,$6D,$7E,$93,$5F,$BE,$D7,$8F
	.dc.b	$B5,$E1,$83,$49,$5D,$E7,$CB,$EA,$F9,$EC,$D2,$4E,$E1,$F7,$DB,$24
	.dc.b	$65,$F8,$7A,$3E,$0B,$FD,$C5,$3E,$D8,$C7,$A5,$A9,$CE,$3F,$12,$73
	.dc.b	$5B,$F9,$2F,$E2,$73,$F9,$C9,$46,$21,$18,$7C,$31,$98,$FE,$CD,$E9
	.dc.b	$79,$3C,$FF,$F4,$5D,$F8,$DB,$7E,$16,$BF,$9F,$A7,$C2,$3D,$5F,$6B
	.dc.b	$6F,$D8,$DA,$F5,$35,$FD,$47,$AA,$FC,$8D,$F0,$3F,$88,$8C,$36,$33
	.dc.b	$7F,$83,$9C,$BF,$85,$F8,$B7,$EB,$2F,$BC,$FC,$8C,$61,$B1,$99,$E0
	.dc.b	$DC,$7F,$1D,$F5,$DF,$CC,$1F,$C9,$B5,$FD,$F3,$5F,$E1,$6D,$D5,$F1
	.dc.b	$CF,$97,$FB,$ED,$C3,$EF,$36,$FE,$E3,$5F,$E1,$59,$7A,$BE,$BF,$CE
	.dc.b	$4A,$F3,$BF,$A9,$9F,$E3,$7E,$DF,$E5,$DA,$57,$09,$FC,$E7,$C9,$3E
	.dc.b	$1D,$86,$1B,$5F,$AF,$9D,$4E,$4E,$3F,$26,$F9,$3A,$4A,$E1,$19,$7F
	.dc.b	$77,$2F,$93,$7C,$9F,$EB,$63,$0D,$8C,$CF,$DB,$6B,$FC,$13,$8D,$F2
	.dc.b	$4F,$93,$FD,$B4,$61,$B1,$99,$FC,$33,$5F,$E1,$5E,$CF,$E1,$5F,$DF
	.dc.b	$35,$CF,$37,$CB,$8C,$47,$D4,$C3,$93,$6F,$D0,$47,$F2,$9F,$67,$FE
	.dc.b	$FD,$7F,$A7,$E2,$8C,$42,$30,$7A,$A3,$3C,$BF,$AC,$ED,$FE,$24,$DA
	.dc.b	$FD,$5C,$63,$AF,$FE,$00,$C7,$ED,$FE,$17,$7D,$DF,$E8,$78,$C1,$0F
	.dc.b	$E1,$5D,$BF,$E3,$47,$F8,$83,$5C,$FC,$1F,$82,$31,$1F,$BF,$F0,$5F
	.dc.b	$0F,$E4,$6C,$7F,$19,$6B,$FC,$5D,$AF,$F3,$CF,$87,$7C,$97,$E5,$DB
	.dc.b	$1F,$CB,$DA,$FF,$3F,$6B,$C7,$F0,$0E,$04,$78,$3C,$FE,$D2,$47,$36
	.dc.b	$BC,$FE,$AF,$C0,$47,$C0,$F8,$FD,$8F,$EF,$5E,$0F,$9D,$B3,$FD,$F7
	.dc.b	$0F,$EF,$8F,$F3,$06,$BF,$E6,$6E,$2F,$EC,$7F,$13,$C2,$3F,$98,$FA
	.dc.b	$5F,$E5,$E7,$FB,$63,$5E,$0D,$C3,$C3,$F2,$9F,$C8,$A7,$FC,$9F,$08
	.dc.b	$C4,$7F,$9D,$23,$1C,$1D,$DB,$1F,$E0,$4F,$8D,$F1,$7B,$4F,$65,$18
	.dc.b	$6C,$EC,$FF,$29,$53,$E8,$F8,$B6,$33,$18,$3C,$18,$FE,$D2,$3C,$ED
	.dc.b	$7C,$CD,$9F,$6F,$81,$E7,$FD,$57,$FB,$8D,$7B,$9A,$FB,$5B,$3F,$51
	.dc.b	$B1,$FE,$3A,$F8,$57,$CB,$3C,$6F,$A1,$18,$6C,$66,$7F,$CC,$AD,$7F
	.dc.b	$CC,$2B,$E5,$B6,$48,$DB,$CB,$EA,$E4,$72,$6B,$FE,$6A,$F9,$7F,$F0
	.dc.b	$2B,$FE,$72,$6B,$9F,$67,$CB,$18,$CF,$FC,$F4,$FF,$33,$9E,$1E,$6E
	.dc.b	$88,$F3,$F6,$38,$91,$C1,$FE,$89,$F2,$7F,$CF,$2D,$F8,$9F,$C0,$C6
	.dc.b	$23,$FD,$15,$1F,$84,$F9,$07,$BF,$49,$3F,$E9,$38,$CB,$FE,$95,$F8
	.dc.b	$EF,$CB,$3F,$E5,$46,$BF,$EA,$36,$F4,$1B,$F1,$FF,$E4,$09,$FF,$4F
	.dc.b	$35,$FF,$51,$B7,$5F,$F5,$EB,$63,$FC,$C3,$F2,$1E,$27,$FD,$52,$D9
	.dc.b	$FF,$61,$B6,$37,$8F,$E8,$3E,$59,$F6,$54,$93,$FE,$AF,$8C,$BF,$BB
	.dc.b	$EC,$FF,$3D,$7A,$FD,$FC,$2F,$D2,$CF,$F6,$9C,$3F,$81,$56,$9F,$F4
	.dc.b	$07,$96,$F9,$6C,$57,$FA,$3A,$F8,$E9,$FC,$87,$0F,$E0,$92,$9C,$BD
	.dc.b	$0F,$11,$FD,$F3,$0F,$E2,$1E,$FF,$48,$E0,$D9,$F2,$E5,$E7,$BC,$4C
	.dc.b	$AE,$47,$8E,$38,$19,$7D,$87,$9E,$32,$F0,$3C,$D1,$B9,$F8,$93,$A9
	.dc.b	$E2,$68,$E1,$5F,$E4,$36,$C6,$E7,$CD,$18,$21,$CA,$7D,$77,$F0,$E3
	.dc.b	$15,$E1,$8C,$57,$8A,$78,$B5,$E5,$7B,$7D,$86,$A7,$89,$F3,$1C,$1A
	.dc.b	$38,$98,$3E,$79,$6F,$CA,$EE,$BC,$04,$9E,$B1,$97,$B3,$EA,$3E,$2D
	.dc.b	$7D,$1C,$3F,$B2,$61,$F8,$4C,$3E,$45,$7F,$B2,$8E,$33,$A7,$E6,$FC
	.dc.b	$0F,$F8,$25,$F9,$0D,$C7,$C5,$1E,$13,$F2,$51,$9C,$7F,$AD,$36,$34
	.dc.b	$7B,$8E,$7E,$9F,$63,$E0,$63,$F8,$F9,$E1,$3F,$99,$6E,$3E,$8C,$3C
	.dc.b	$33,$9B,$E4,$FC,$12,$47,$EB,$6B,$89,$CA,$FB,$57,$09,$D3,$F4,$67
	.dc.b	$6E,$70,$F9,$B3,$AB,$F4,$91,$F1,$E0,$BE,$33,$9A,$E2,$7C,$4B,$E9
	.dc.b	$1F,$21,$F0,$18,$7D,$87,$AA,$38,$3C,$CE,$06,$4B,$7E,$52,$47,$82
	.dc.b	$32,$F2,$8C,$C7,$D4,$BF,$45,$78,$5B,$1D,$4F,$1C,$F0,$6D,$75,$69
	.dc.b	$D8,$87,$DF,$8E,$13,$D1,$B8,$7D,$29,$F6,$1C,$8E,$15,$E7,$23,$D2
	.dc.b	$BF,$05,$18,$8E,$11,$D6,$3C,$8D,$8E,$18,$7E,$05,$F7,$3F,$5A,$FE
	.dc.b	$CB,$6B,$A5,$76,$AE,$38,$7C,$5C,$3D,$18,$7B,$30,$E9,$5E,$AB,$F0
	.dc.b	$3E,$4A,$DE,$1D,$2B,$77,$C7,$0E,$97,$E4,$6B,$CE,$7C,$8F,$D7,$6B
	.dc.b	$D4,$E0,$78,$6F,$BB,$E8,$3E,$D6,$1E,$42,$4F,$8B,$0F,$C6,$CF,$F9
	.dc.b	$A6,$3F,$D1,$33,$F8,$28,$C5,$FF,$9A,$7D,$2F,$F7,$FC,$7F,$9A,$5A
	.dc.b	$FF,$9C,$5A,$FF,$96,$DA,$FF,$57,$78,$E3,$FD,$85,$B8,$7F,$98,$7C
	.dc.b	$FF,$E2,$93,$9F,$27,$F9,$FD,$6A,$F9,$47,$F4,$2A,$DD,$F7,$8D,$92
	.dc.b	$7F,$A0,$4E,$9E,$27,$F4,$3A,$E8,$FF,$67,$6E,$3F,$E9,$2D,$39,$63
	.dc.b	$E3,$AE,$18,$78,$67,$38,$F8,$67,$27,$82,$32,$F8,$2B,$91,$FE,$01
	.dc.b	$C7,$C9,$1E,$1B,$DB,$4E,$47,$49,$E3,$7D,$DE,$57,$C3,$06,$3E,$A3
	.dc.b	$D7,$8F,$68,$CE,$3B,$9C,$D5,$8E,$55,$B9,$E5,$18,$9D,$63,$E3,$C7
	.dc.b	$EC,$46,$31,$F2,$CF,$F8,$57,$0E,$57,$9B,$DC,$F1,$3E,$23,$EB,$33
	.dc.b	$5C,$A3,$E1,$37,$1F,$A9,$1C,$4C,$CF,$17,$C1,$87,$77,$A9,$83,$84
	.dc.b	$FC,$26,$E1,$E6,$73,$8F,$C4,$6E,$3F,$05,$A7,$0B,$EC,$F8,$8E,$D8
	.dc.b	$F9,$EB,$E4,$C7,$DA,$8C,$57,$36,$C7,$33,$AB,$EF,$BF,$23,$A3,$EF
	.dc.b	$C6,$DD,$5F,$3C,$7E,$14,$FF,$90,$6B,$8C,$7C,$C3,$38,$FC,$AB,$E3
	.dc.b	$5C,$E7,$35,$C2,$3D,$71,$8B,$ED,$1F,$30,$EC,$FF,$B5,$E7,$FC,$18
	.dc.b	$FF,$BC,$DB,$1F,$80,$8C,$1F,$A9,$7F,$C6,$9E,$8F,$E5,$66,$4E,$6F
	.dc.b	$FC,$CD,$38,$AF,$95,$7E,$B7,$FC,$C5,$19,$9F,$8A,$49,$FF,$76,$46
	.dc.b	$5E,$13,$E8,$3C,$35,$DE,$33,$87,$C5,$78,$C6,$67,$BB,$5F,$F3,$75
	.dc.b	$75,$72,$79,$8B,$7E,$36,$D7,$69,$E1,$1C,$AB,$A5,$A7,$2D,$8C,$C6
	.dc.b	$08,$E0,$93,$E1,$3A,$1F,$E3,$F3,$CA,$9F,$69,$93,$6E,$5F,$39,$07
	.dc.b	$02,$3F,$86,$4F,$F2,$DE,$67,$CA,$36,$FC,$0E,$09,$CC,$61,$B7,$E3
	.dc.b	$3C,$0E,$27,$FD,$67,$CB,$B0,$52,$4F,$FB,$03,$D4,$E9,$CE,$69,$23
	.dc.b	$FD,$9D,$F3,$1E,$1E,$93,$49,$1F,$ED,$5B,$F3,$75,$78,$23,$04,$C7
	.dc.b	$E6,$64,$EC,$79,$7E,$D6,$D7,$FB,$A6,$30,$7C,$4F,$19,$F2,$6F,$B5
	.dc.b	$C3,$6B,$E2,$63,$F5,$EB,$75,$A3,$7F,$25,$2B,$74,$AE,$B3,$98,$FC
	.dc.b	$14,$78,$3E,$63,$B9,$2B,$8C,$65,$BF,$2B,$EA,$AE,$2D,$8F,$C3,$4A
	.dc.b	$31,$3E,$76,$BE,$35,$E1,$FC,$16,$D7,$68,$C1,$9F,$5E,$47,$89,$B5
	.dc.b	$F9,$B8,$7E,$EB,$F9,$47,$E0,$98,$E9,$FC,$8B,$4C,$C6,$E7,$92,$F9
	.dc.b	$8D,$B6,$77,$18,$3F,$30,$BE,$32,$91,$FA,$3C,$5F,$A5,$8F,$CF,$7A
	.dc.b	$8F,$4A,$92,$B7,$19,$7D,$3E,$BB,$E4,$FD,$2D,$8C,$C6,$0F,$D6,$F8
	.dc.b	$8F,$5F,$F3,$73,$F8,$28,$C7,$5E,$A7,$3B,$7F,$03,$9C,$D7,$E2,$21
	.dc.b	$F1,$CF,$AF,$FD,$CC,$61,$B1,$99,$E0,$D9,$FE,$14,$BE,$C7,$F8,$64
	.dc.b	$7F,$0F,$8C,$4F,$F0,$D8,$FE,$1D,$18,$70,$78,$A7,$8B,$E4,$FA,$CB
	.dc.b	$26,$92,$B8,$CF,$AB,$EE,$B6,$48,$CD,$23,$9F,$CB,$C9,$CB,$6B,$CD
	.dc.b	$93,$E0,$BF,$C5,$B5,$E6,$8C,$79,$BE,$E6,$C6,$67,$75,$CB,$C0,$7D
	.dc.b	$06,$DB,$5C,$DB,$3B,$9E,$9D,$3F,$E2,$08,$FE,$69,$87,$F3,$2F,$3B
	.dc.b	$24,$C3,$C9,$F0,$BF,$ED,$36,$FA,$1E,$58,$C5,$FB,$1A,$FF,$3F,$F2
	.dc.b	$DF,$AD,$BE,$2B,$EC,$C6,$1B,$19,$C3,$EB,$CE,$5F,$B3,$F1,$6F,$92
	.dc.b	$7C,$27,$E0,$A3,$0D,$8C,$CF,$06,$E3,$F8,$2F,$92,$7B,$7F,$82,$6D
	.dc.b	$7D,$E6,$CE,$E7,$CF,$F2,$7F,$C5,$E3,$D8,$D7,$FB,$06,$48,$C4,$7F
	.dc.b	$57,$6E,$5E,$9F,$8E,$F3,$13,$1F,$33,$F9,$4A,$FE,$D5,$EC,$FE,$41
	.dc.b	$A5,$70,$9F,$CA,$FC,$83,$E0,$18,$61,$B5,$FA,$29,$D4,$E4,$90,$E3
	.dc.b	$7B,$DF,$DD,$EB,$73,$98,$C5,$F4,$AD,$CF,$F8,$06,$B7,$18,$87,$B7
	.dc.b	$FE,$C1,$7F,$D8,$A4,$3F,$C1,$0E,$5B,$CB,$FE,$C8,$26,$DF,$B2,$9D
	.dc.b	$4F,$49,$CB,$FB,$E8,$C9,$23,$31,$FE,$D0,$34,$E1,$F1,$1C,$E7,$B4
	.dc.b	$66,$3C,$71,$82,$3C,$9A,$F5,$21,$D2,$7C,$51,$A8,$C4,$F5,$7C,$07
	.dc.b	$32,$1C,$8F,$09,$83,$8C,$73,$72,$6C,$E6,$59,$F0,$BC,$4D,$9C,$8C
	.dc.b	$1C,$0C,$1C,$67,$CD,$1C,$A3,$13,$E1,$78,$1D,$08,$60,$91,$C5,$EA
	.dc.b	$43,$6E,$5B,$3E,$02,$4F,$79,$D4,$F4,$9C,$BE,$A8,$C9,$23,$31,$E8
	.dc.b	$34,$E1,$F1,$1C,$E7,$B4,$66,$3C,$71,$82,$3C,$9A,$F5,$21,$D2,$7C
	.dc.b	$51,$A8,$C4,$F5,$7C,$07,$32,$1C,$8F,$09,$83,$8C,$73,$72,$6C,$E6
	.dc.b	$59,$F0,$BC,$4D,$9C,$8C,$1C,$0C,$1C,$67,$CD,$1C,$A3,$13,$E1,$78
	.dc.b	$1D,$08,$60,$91,$C5,$EA,$43,$6E,$5B,$3E,$02,$4F,$79,$D4,$F4,$9C
	.dc.b	$BE,$A8,$C9,$23,$31,$E8,$34,$E1,$F1,$1C,$E7,$B4,$66,$3C,$71,$82
	.dc.b	$3C,$9A,$F5,$21,$D2,$7C,$51,$A8,$C4,$F5,$7C,$07,$32,$1C,$8F,$09
	.dc.b	$83,$8C,$73,$72,$7E,$FF,$AD,$CF,$FC,$9D,$B1,$FC,$95,$AF,$F8,$9E
	.dc.b	$D1,$88,$FF,$5B,$46,$27,$91,$23,$93,$5E,$53,$C5,$FE,$76,$D7,$FA
	.dc.b	$0B,$67,$A4,$72,$21,$CA,$7A,$3E,$29,$E0,$48,$CC,$76,$6B,$D9,$B3
	.dc.b	$DA,$3A,$1D,$8E,$91,$A9,$E9,$19,$3A,$14,$E8,$59,$E8,$F2,$31,$1C
	.dc.b	$9E,$E4,$34,$53,$A9,$4E,$A5,$8E,$AE,$8E,$24,$30,$43,$89,$92,$99
	.dc.b	$79,$B8,$23,$82,$47,$A6,$39,$46,$27,$E3,$12,$3D,$ED,$7D,$13,$C5
	.dc.b	$F0,$35,$EE,$D9,$E9,$1C,$88,$72,$9E,$8F,$8A,$78,$12,$33,$1D,$9A
	.dc.b	$F6,$6C,$F6,$8E,$87,$63,$A4,$6A,$7A,$46,$4E,$85,$3A,$16,$7A,$3C
	.dc.b	$8C,$47,$27,$B9,$0D,$14,$EA,$53,$A9,$63,$AB,$A3,$89,$0C,$10,$E2
	.dc.b	$64,$A6,$5E,$6E,$08,$E0,$91,$E9,$8E,$51,$89,$F8,$C4,$8F,$7B,$5F
	.dc.b	$44,$F1,$7C,$0D,$7B,$B6,$7A,$47,$22,$1C,$A7,$A3,$E2,$9E,$04,$8C
	.dc.b	$C7,$66,$BD,$9B,$3D,$A3,$A1,$D8,$E9,$1A,$9E,$91,$93,$A1,$4E,$85
	.dc.b	$9E,$8F,$23,$11,$C9,$EE,$43,$45,$3A,$94,$EA,$58,$EA,$FE,$F7,$1F
	.dc.b	$F3,$5E,$7F,$E7,$0D,$13,$FE,$73,$3F,$CE,$1A,$7F,$9C,$E7,$51,$91
	.dc.b	$5F,$E7,$A7,$FD,$CC,$D7,$66,$27,$FD,$0E,$DA,$EF,$A7,$F9,$EE,$BC
	.dc.b	$73,$E0,$9C,$DF,$82,$BF,$D0,$18,$7A,$AB,$DA,$7C,$22,$1F,$08,$A6
	.dc.b	$8C,$10,$C1,$1E,$84,$C7,$E8,$47,$28,$C6,$5F,$74,$91,$F5,$9A,$FC
	.dc.b	$F9,$E2,$F7,$9E,$2E,$C8,$6E,$78,$BD,$E7,$8B,$E5,$6C,$F2,$8E,$07
	.dc.b	$23,$84,$F2,$7C,$13,$C0,$91,$98,$E8,$D9,$E9,$1C,$8E,$87,$28,$D4
	.dc.b	$F2,$8C,$9C,$8B,$3C,$9E,$06,$23,$83,$E8,$21,$DC,$A7,$02,$9D,$4B
	.dc.b	$1D,$5C,$1C,$48,$71,$29,$A3,$04,$30,$47,$A1,$23,$CF,$1C,$A3,$13
	.dc.b	$F0,$89,$1E,$D6,$BE,$69,$E2,$F7,$9F,$EC,$9A,$7F,$8D,$5F,$F1,$CC
	.dc.b	$7F,$8C,$4F,$F1,$79,$FE,$32,$30,$7F,$8E,$8F,$F1,$C1,$4D,$18,$79
	.dc.b	$B6,$3B,$13,$4F,$F1,$E4,$71,$3B,$90,$C1,$27,$9B,$D8,$86,$DC,$B6
	.dc.b	$7B,$3E,$88,$CB,$CE,$32,$F9,$67,$C7,$3A,$24,$EA,$7B,$9C,$67,$71
	.dc.b	$98,$F0,$46,$08,$F8,$63,$4F,$13,$11,$A7,$C4,$60,$86,$08,$78,$A7
	.dc.b	$CB,$1C,$63,$13,$DA,$73,$1C,$8C,$CE,$DF,$39,$92,$9D,$8A,$70,$34
	.dc.b	$7A,$4D,$14,$D1,$88,$F0,$B6,$7C,$84,$9F,$14,$71,$3B,$90,$C1,$27
	.dc.b	$9B,$FC,$C7,$18,$FF,$35,$3F,$A1,$67,$FC,$E1,$FE,$87,$A7,$F3,$E7
	.dc.b	$FA,$09,$0E,$07,$F4,$E2,$0D,$6B,$86,$BB,$6B,$FD,$59,$DE,$AE,$36
	.dc.b	$EE,$0C,$C6,$1B,$19,$BF,$0D,$71,$8C,$47,$97,$47,$0B,$AB,$33,$FD
	.dc.b	$9A,$73,$18,$84,$63,$0F,$5B,$63,$DB,$C8,$E4,$E1,$08,$C5,$7C,$06
	.dc.b	$C7,$09,$ED,$97,$BF,$96,$E8,$69,$30,$F0,$F3,$39,$FC,$D9,$A3,$13
	.dc.b	$A9,$CC,$63,$2F,$AB,$CC,$E8,$73,$7B,$E3,$0D,$8C,$E1,$F6,$A7,$2D
	.dc.b	$E9,$74,$F1,$6C,$66,$30,$75,$E1,$75,$52,$4F,$58,$CB,$DF,$AF,$FD
	.dc.b	$9E,$DE,$BF,$34,$63,$4F,$5B,$5F,$C8,$F7,$3B,$7E,$9B,$6F,$F2,$2D
	.dc.b	$9F,$CA,$B5,$E3,$DD,$C0,$93,$C1,$E7,$DF,$23,$9B,$67,$9F,$81,$E0
	.dc.b	$24,$F8,$1F,$1F,$83,$F9,$78,$F3,$B6,$7F,$55,$82,$31,$1F,$E3,$A8
	.dc.b	$C4,$FF,$8F,$0F,$D8,$70,$3C,$27,$87,$C2,$30,$D8,$CC,$F1,$BF,$F2
	.dc.b	$2D,$6A,$31,$1F,$AF,$F1,$5E,$2F,$8A,$51,$87,$66,$23,$51,$99,$FD
	.dc.b	$8F,$92,$F1,$DE,$47,$B2,$30,$D8,$CE,$1E,$29,$CB,$47,$93,$C5,$A6
	.dc.b	$5C,$1E,$7F,$2D,$E5,$D2,$4F,$A2,$32,$DE,$AE,$04,$70,$7C,$4F,$3D
	.dc.b	$E7,$BC,$EF,$AB,$18,$6C,$66,$78,$37,$5F,$E2,$75,$CB,$CF,$D9,$2B
	.dc.b	$6F,$2E,$C9,$1C,$9B,$5C,$BD,$37,$A3,$E2,$6D,$79,$DB,$3B,$9D,$7A
	.dc.b	$52,$B4,$D9,$E9,$E9,$F3,$24,$F3,$7E,$C7,$A9,$FC,$A6,$3E,$B3,$67
	.dc.b	$EF,$B6,$3F,$A2,$71,$23,$13,$FE,$94,$87,$AA,$F5,$BF,$08,$DB,$FC
	.dc.b	$13,$6B,$F1,$6D,$8E,$BE,$BF,$E7,$27,$F1,$ED,$8C,$F6,$7E,$3E,$31
	.dc.b	$3E,$1F,$65,$EC,$FE,$4D,$B7,$F9,$16,$CF,$E6,$9B,$1E,$CC,$7C,$53
	.dc.b	$EB,$6A,$CB,$D9,$1F,$9F,$32,$BD,$AF,$E9,$B5,$F9,$F6,$D7,$EA,$1B
	.dc.b	$1F,$29,$B1,$FD,$BF,$DA,$7B,$7F,$D8,$7D,$87,$B9,$98,$C3,$FD,$D7
	.dc.b	$AB,$FB,$56,$3F,$DB,$E3,$4D,$C3,$FD,$8F,$87,$48,$FE,$07,$08,$C3
	.dc.b	$FD,$A1,$B1,$FE,$00,$6C,$74,$3F,$C0,$53,$E7,$F3,$7C,$71,$DB,$0A
	.dc.b	$F5,$8B,$5D,$B1,$F5,$CF,$92,$73,$5A,$6E,$1E,$99,$E4,$DB,$E5,$87
	.dc.b	$86,$31,$7C,$5A,$6E,$F9,$DF,$A0,$F0,$46,$9C,$1D,$8B,$7D,$9D,$98
	.dc.b	$AE,$CD,$9E,$C7,$3C,$3D,$B1,$CD,$B5,$FD,$AA,$3A,$DF,$92,$72,$F4
	.dc.b	$6D,$73,$C3,$DD,$5C,$23,$BB,$E1,$30,$49,$E2,$E5,$A4,$9E,$6F,$93
	.dc.b	$0F,$04,$60,$93,$F6,$EB,$BD,$74,$7F,$09,$7F,$D6,$A7,$E4,$46,$0F
	.dc.b	$C6,$E1,$EF,$AF,$83,$E6,$3C,$4E,$78,$FE,$7E,$31,$87,$A6,$31,$3A
	.dc.b	$8F,$CC,$CE,$AB,$D5,$18,$9D,$3F,$9B,$9D,$57,$AA,$31,$3A,$B5,$6A
	.dc.b	$7E,$ED,$F9,$AB,$53,$99,$FB,$57,$E4,$7C,$C7,$84,$FC,$3C,$6E,$FA
	.dc.b	$D7,$23,$F3,$11,$93,$58,$FE,$91,$C4,$68,$E1,$5D,$1C,$3E,$03,$D6
	.dc.b	$64,$98,$7C,$47,$2F,$AD,$CE,$1F,$0F,$0F,$2C,$60,$F5,$4F,$18,$CB
	.dc.b	$C4,$93,$C6,$FC,$F5,$C4,$F8,$A7,$28,$F8,$D3,$C2,$3E,$6C,$60,$8E
	.dc.b	$CC,$94,$F4,$BA,$7A,$98,$35,$5E,$68,$F2,$5F,$C6,$AF,$05,$70,$AD
	.dc.b	$3C,$2F,$E8,$E3,$BC,$3B,$56,$AB,$85,$7B,$AF,$9C,$FE,$6E,$B7,$39
	.dc.b	$7C,$35,$F0,$E7,$91,$B7,$C1,$87,$77,$DE,$7A,$C9,$1C,$31,$FC,$24
	.dc.b	$EE,$73,$18,$7B,$5F,$A9,$B3,$BA,$F0,$47,$49,$F3,$3E,$08,$F2,$C6
	.dc.b	$23,$DF,$3C,$63,$A6,$1F,$01,$D5,$74,$38,$CF,$7B,$F2,$56,$E3,$F3
	.dc.b	$2F,$E4,$6F,$FA,$55,$FE,$F9,$B3,$FE,$C8,$BF,$C9,$C7,$FB,$8F,$D2
	.dc.b	$FE,$7D,$E8,$7F,$AD,$A3,$F9,$9B,$70,$FE,$57,$EE,$FF,$B3,$F3,$FE
	.dc.b	$D1,$39,$6D,$F2,$3F,$A4,$4E,$EF,$FB,$1D,$6D,$B3,$B7,$FA,$0C,$ED
	.dc.b	$FE,$CD,$3B,$6C,$EE,$FC,$71,$6F,$C2,$FA,$27,$C2,$E4,$EE,$64,$EE
	.dc.b	$64,$EA,$66,$7A,$B6,$7A,$C7,$F4,$CA,$F2,$60,$C3,$D0,$FF,$B0,$B0
	.dc.b	$E8,$DC,$3B,$9C,$CF,$1D,$F3,$9F,$3C,$64,$F1,$C6,$5E,$2D,$26,$1E
	.dc.b	$43,$FC,$35,$7E,$2C,$3C,$A4,$BE,$B5,$C2,$FC,$37,$C5,$FF,$0B,$DF
	.dc.b	$1A,$F4,$DF,$BA,$31,$3E,$5C,$3D,$98,$7C,$63,$A4,$7A,$30,$E8,$FF
	.dc.b	$87,$EB,$8C,$EA,$B8,$D7,$A2,$BB,$E1,$F2,$2F,$73,$CA,$B7,$5E,$63
	.dc.b	$67,$8E,$76,$FB,$08,$78,$4D,$DF,$86,$37,$1E,$73,$73,$F6,$DD,$D7
	.dc.b	$86,$77,$3E,$29,$E0,$75,$3E,$26,$1F,$3D,$FA,$F5,$E9,$8C,$5F,$C7
	.dc.b	$8C,$61,$EF,$AE,$E7,$8B,$0F,$C5,$9E,$0B,$FA,$51,$F6,$AB,$8B,$D8
	.dc.b	$F8,$87,$8A,$B4,$E2,$FE,$85,$72,$8E,$07,$AB,$0F,$61,$F0,$5C,$B8
	.dc.b	$8E,$13,$FA,$4D,$3F,$CA,$D8,$AF,$FC,$22,$D7,$FC,$B8,$D8,$DC,$79
	.dc.b	$0F,$91,$3F,$E3,$6F,$7B,$DF,$19,$C3,$C5,$39,$8C,$5E,$9F,$F2,$3D
	.dc.b	$59,$F9,$91,$EF,$CB,$FC,$F9,$39,$7F,$1D,$1E,$C8,$C3,$B2,$1F,$E7
	.dc.b	$57,$F7,$66,$9F,$3C,$74,$7F,$CC,$91,$9B,$F3,$CE,$67,$CA,$7A,$0D
	.dc.b	$90,$F9,$B7,$D9,$C5,$FA,$9E,$E7,$FA,$2A,$F4,$F0,$37,$5D,$E7,$85
	.dc.b	$F8,$A7,$81,$DF,$0F,$F4,$65,$FD,$38,$F3,$46,$1F,$F2,$4C,$76,$AF
	.dc.b	$A1,$1F,$E7,$6D,$DE,$F7,$F3,$BE,$27,$BE,$F6,$BA,$5F,$FA,$DE,$30
	.dc.b	$FE,$17,$2F,$F6,$CC,$62,$11,$89,$E3,$5F,$ED,$9D,$2F,$2F,$F9,$F2
	.dc.b	$BF,$D8,$D1,$83,$7C,$CF,$96,$A4,$C7,$FD,$A0,$F4,$E9,$7C,$55,$24
	.dc.b	$F9,$4F,$F7,$54,$69,$FF,$73,$DF,$93,$B3,$C5,$1F,$EE,$F6,$BF,$ED
	.dc.b	$56,$E7,$FE,$F5,$F0,$FD,$7D,$C6,$92,$BC,$03,$C4,$7C,$1B,$E0,$FE
	.dc.b	$5C,$61,$B1,$99,$E3,$39,$6C,$3E,$0B,$64,$8C,$BC,$BE,$12,$47,$06
	.dc.b	$BF,$50,$FD,$E5,$3E,$1D,$F0,$DE,$46,$CF,$89,$AF,$13,$D8,$E3,$8F
	.dc.b	$F8,$25,$73,$F8,$7E,$44,$AE,$4F,$5F,$88,$91,$D5,$AF,$5A,$7C,$4B
	.dc.b	$D2,$FC,$8C,$61,$B1,$99,$F9,$F3,$96,$FC,$53,$D5,$E2,$D8,$CC,$60
	.dc.b	$FC,$AC,$BE,$12,$92,$7E,$DF,$27,$E7,$23,$F3,$7E,$B3,$C2,$A4,$AD
	.dc.b	$C6,$5F,$1F,$A5,$E2,$23,$83,$F0,$FE,$43,$D9,$FC,$B4,$7C,$F8,$C7
	.dc.b	$6F,$EC,$9B,$1C,$27,$79,$FF,$35,$87,$8C,$F6,$3F,$73,$18,$6C,$66
	.dc.b	$3F,$D9,$B1,$8F,$8B,$E5,$1F,$CF,$E5,$18,$8F,$E8,$91,$87,$FC,$A1
	.dc.b	$19,$7C,$3F,$19,$64,$D2,$47,$F9,$5F,$E3,$74,$47,$07,$F9,$6F,$D9
	.dc.b	$E4,$47,$07,$9B,$37,$C1,$BD,$1F,$34,$61,$B1,$99,$E0,$D7,$E4,$4B
	.dc.b	$77,$C7,$72,$6C,$F7,$6B,$C1,$AF,$F6,$59,$79,$AF,$4F,$C6,$D9,$F4
	.dc.b	$35,$F0,$B5,$E5,$F0,$DF,$EE,$C7,$DC,$D7,$CE,$D7,$7E,$A7,$D5,$F5
	.dc.b	$1E,$B7,$F9,$FE,$7F,$09,$18,$35,$19,$8F,$EF,$91,$97,$F0,$3F,$14
	.dc.b	$F9,$07,$C8,$7E,$06,$30,$D8,$CC,$F0,$6B,$F8,$AF,$31,$F1,$FF,$85
	.dc.b	$6C,$7E,$2D,$AE,$E3,$D5,$E0,$72,$8F,$2B,$5F,$F0,$A6,$3F,$BD,$C3
	.dc.b	$67,$8F,$E3,$24,$F8,$DA,$FE,$73,$E4,$1E,$33,$E4,$7F,$39,$18,$6C
	.dc.b	$66,$78,$35,$FD,$74,$FF,$33,$D7,$8F,$9A,$69,$27,$F4,$9E,$D7,$E9
	.dc.b	$53,$48,$FE,$E6,$1F,$1E,$F9,$2F,$E6,$13,$FE,$33,$9C,$C6,$23,$C1
	.dc.b	$3B,$F0,$BE,$CB,$63,$F3,$BF,$26,$F9,$4E,$1D,$48,$C5,$FE,$AD,$B5
	.dc.b	$C2,$BB,$2F,$93,$7C,$5E,$1B,$7F,$C1,$5B,$3F,$C2,$9A,$FF,$92,$73
	.dc.b	$F6,$F2,$FA,$5B,$5E,$88,$C6,$7F,$88,$8F,$E2,$7C,$8F,$94,$FD,$B4
	.dc.b	$61,$B8,$7E,$79,$AF,$E7,$1A,$FF,$96,$F5,$75,$FF,$AF,$6B,$F5,$BF
	.dc.b	$05,$F1,$23,$15,$FA,$E8,$F8,$FD,$AE,$DA,$4B,$FD,$DB,$F3,$3E,$57
	.dc.b	$FC,$03,$7D,$DA,$B1,$46,$1B,$1F,$53,$C2,$7C,$25,$26,$1F,$C6,$1F
	.dc.b	$A7,$F2,$DF,$E0,$B8,$FE,$68,$DA,$FE,$72,$D7,$FA,$23,$5F,$F3,$D2
	.dc.b	$F6,$79,$92,$79,$BF,$80,$F8,$FF,$EE,$51,$FD,$05,$B5,$FD,$19,$AF
	.dc.b	$F5,$16,$BF,$E8,$75,$F2,$9C,$C9,$3C,$DF,$5F,$C1,$48,$F5,$B6,$BF
	.dc.b	$9A,$FA,$0F,$81,$FF,$0D,$57,$F2,$F8,$C3,$FD,$8E,$33,$3F,$7E,$72
	.dc.b	$FF,$6B,$F9,$87,$CC,$6A,$3E,$A4,$62,$FF,$B9,$B6,$38,$4E,$FD,$87
	.dc.b	$AC,$F9,$9C,$37,$0F,$F0,$0C,$60,$E0,$FC,$23,$FC,$16,$5B,$FF,$16
	.dc.b	$57,$F9,$A3,$4D,$9E,$59,$FE,$F5,$DD,$7D,$6F,$D6,$C7,$FA,$C5,$B7
	.dc.b	$FE,$73,$6A,$B7,$CD,$77,$AF,$F5,$D4,$63,$0F,$64,$E6,$BF,$D8,$13
	.dc.b	$9A,$FF,$10,$FC,$DF,$CA,$78,$B6,$76,$D7,$C0,$DF,$80,$7C,$9F,$F8
	.dc.b	$4E,$BF,$CB,$AD,$AF,$F2,$64,$EB,$2F,$A8,$DA,$FF,$18,$FC,$E3,$E4
	.dc.b	$BE,$93,$63,$31,$83,$8E,$0F,$9D,$A4,$9F,$0B,$AF,$89,$91,$E4,$6E
	.dc.b	$5E,$4F,$95,$7C,$F7,$91,$B5,$FE,$76,$6C,$EE,$7F,$0D,$EA,$FD,$2D
	.dc.b	$CD,$27,$17,$E2,$FE,$31,$F3,$14,$99,$7F,$94,$23,$2F,$F9,$4B,$E3
	.dc.b	$BE,$76,$06,$93,$AF,$F2,$DF,$1F,$E4,$AF,$BE,$DD,$3E,$EB,$5F,$C3
	.dc.b	$36,$D8,$3E,$43,$CC,$4B,$F2,$3F,$A3,$F9,$4F,$F5,$8C,$7F,$99,$9B
	.dc.b	$87,$E7,$1B,$1F,$90,$6C,$7F,$12,$F9,$7F,$F4,$A7,$AB,$FB,$D8,$0F
	.dc.b	$11,$3A,$D3,$FD,$33,$F3,$FF,$E3,$6C,$77,$87,$1C,$38,$5E,$DE,$78
	.dc.b	$72,$AF,$E6,$AD,$8F,$ED,$AD,$8F,$E4,$F3,$FC,$16,$79,$C7,$6C,$3C
	.dc.b	$B1,$FE,$3E,$8E,$CE,$8F,$F7,$63,$8A,$ED,$7E,$5F,$2F,$FD,$CB,$3C
	.dc.b	$EB,$CB,$87,$29,$CC,$7F,$81,$B0,$F8,$D5,$B8,$F0,$37,$0E,$B3,$E4
	.dc.b	$AE,$38,$77,$AE,$87,$8F,$0A,$F3,$9D,$D7,$18,$D1,$93,$78,$7B,$1F
	.dc.b	$05,$62,$FB,$47,$F6,$0B,$F4,$BF,$E9,$36,$D7,$58,$D1,$F2,$3E,$1F
	.dc.b	$FC,$CC,$F7,$3D,$CE,$23,$C0,$7C,$33,$FD,$23,$F0,$3E,$1C,$7F,$D2
	.dc.b	$CD,$8C,$D7,$37,$81,$FE,$A4,$8F,$01,$23,$58,$7E,$26,$FC,$0D,$9D
	.dc.b	$57,$8A,$31,$5E,$69,$CC,$7C,$29,$EF,$5C,$AF,$CB,$7F,$E0,$48,$F6
	.dc.b	$3F,$EB,$76,$CE,$8F,$F1,$FB,$C5,$FF,$57,$FC,$27,$6A,$FF,$5A,$36
	.dc.b	$33,$5E,$33,$E4,$C6,$63,$FD,$86,$DA,$FF,$53,$4E,$89,$3E,$8C,$3E
	.dc.b	$A5,$49,$CE,$1F,$02,$BD,$B3,$96,$CF,$85,$A4,$7D,$D1,$FA,$BB,$F0
	.dc.b	$CF,$BA,$30,$7E,$0F,$CE,$FF,$13,$47,$2B,$FA,$91,$87,$EB,$4E,$FC
	.dc.b	$DF,$2E,$5F,$ED,$7A,$D3,$F1,$B1,$FC,$5C,$63,$0F,$8E,$7D,$67,$0F
	.dc.b	$FB,$A1,$B8,$FF,$8F,$EF,$AC,$FF,$B8,$27,$24,$8C,$C7,$17,$B9,$1F
	.dc.b	$55,$F9,$63,$18,$FB,$1E,$37,$DE,$BC,$91,$AA,$F5,$46,$09,$5C,$5E
	.dc.b	$A7,$96,$78,$34,$C9,$E5,$38,$CF,$8D,$CD,$78,$6B,$A1,$27,$C2,$D9
	.dc.b	$E8,$D7,$F1,$B8,$78,$1E,$D1,$9A,$FB,$ED,$C3,$E4,$4F,$81,$B1,$BC
	.dc.b	$7E,$E4,$F9,$8E,$F8,$78,$8F,$44,$F3,$C7,$E4,$DF,$29,$C9,$BA,$E2
	.dc.b	$43,$CA,$E5,$A6,$70,$F8,$51,$96,$C6,$67,$F0,$B8,$79,$B0,$F1,$4F
	.dc.b	$AF,$0F,$85,$7C,$23,$E2,$9E,$B3,$B9,$1F,$8B,$87,$A2,$3B,$4F,$B0
	.dc.b	$E7,$87,$83,$1B,$1D,$B1,$F1,$5F,$1B,$E9,$18,$7F,$04,$7F,$0F,$BF
	.dc.b	$D6,$B6,$BF,$AD,$BF,$83,$3F,$8A,$79,$8F,$4B,$31,$FD,$D3,$2F,$F4
	.dc.b	$AC,$62,$11,$87,$6D,$79,$C7,$F3,$C9,$E9,$1F,$C9,$9B,$9F,$F3,$1F
	.dc.b	$43,$FB,$6B,$FC,$CE,$33,$9F,$F7,$0C,$3B,$4E,$70,$FE,$8F,$7C,$2B
	.dc.b	$57,$C3,$0F,$43,$5F,$3B,$70,$57,$19,$DD,$F9,$C8,$FF,$3C,$8D,$3C
	.dc.b	$A3,$38,$78,$E3,$46,$30,$F1,$CD,$8F,$95,$5C,$AF,$D7,$5A,$9F,$89
	.dc.b	$5E,$C7,$81,$DC,$C9,$D8,$B7,$E1,$24,$FC,$7B,$ED,$87,$BC,$97,$D6
	.dc.b	$F8,$D7,$86,$F8,$C7,$98,$F0,$E1,$E6,$7E,$29,$E1,$8D,$BE,$56,$C7
	.dc.b	$CC,$8E,$37,$E0,$78,$57,$72,$3D,$2B,$F0,$4F,$0B,$F0,$E1,$EA,$71
	.dc.b	$1C,$8E,$15,$C9,$F1,$3F,$05,$DD,$79,$2B,$94,$7C,$03,$89,$EF,$30
	.dc.b	$72,$32,$7E,$79,$E3,$1E,$EB,$FC,$54,$62,$BC,$58,$7C,$7B,$E0,$78
	.dc.b	$A3,$33,$B8,$EB,$3B,$3E,$43,$E2,$71,$5D,$A3,$EC,$5F,$87,$0B,$87
	.dc.b	$BE,$78,$56,$DF,$8C,$7A,$C9,$7D,$AF,$E7,$C7,$F8,$43,$0F,$F2,$EE
	.dc.b	$9F,$E1,$56,$C7,$58,$C3,$E0,$3E,$44,$FF,$8A,$7E,$2B,$F3,$95,$27
	.dc.b	$47,$F0,$68,$C3,$E4,$BF,$16,$3F,$BD,$9F,$D2,$1A,$BF,$29,$D6,$BF
	.dc.b	$C8,$B1,$99,$F8,$13,$99,$FF,$18,$CE,$5F,$2C,$F5,$78,$9F,$E3,$38
	.dc.b	$F6,$BF,$C7,$5A,$F9,$CD,$1C,$0E,$A7,$F9,$A7,$E5,$BC,$EE,$6B,$C4
	.dc.b	$7F,$9F,$E3,$31,$C0,$91,$97,$91,$C6,$BC,$CF,$F9,$63,$81,$5F,$C4
	.dc.b	$1B,$1B,$3D,$07,$23,$84,$78,$2B,$C7,$5F,$E8,$58,$E4,$7B,$AB,$9D
	.dc.b	$32,$FF,$3C,$4E,$49,$1D,$6F,$D0,$FF,$9B,$38,$BF,$DE,$8E,$70,$EC
	.dc.b	$D3,$FD,$63,$E1,$78,$1C,$0C,$7C,$37,$FE,$B9,$BE,$73,$D6,$B8,$3E
	.dc.b	$C3,$FD,$40,$FF,$32,$8D,$CF,$90,$97,$FE,$B6,$8C,$B6,$79,$CE,$AB
	.dc.b	$CA,$74,$3F,$D8,$B1,$F8,$2A,$F1,$E3,$FD,$99,$B3,$CC,$E0,$79,$CF
	.dc.b	$F5,$E7,$89,$FE,$F5,$FA,$0F,$E9,$9D,$4F,$97,$77,$FE,$42,$30,$D8
	.dc.b	$CD,$FF,$B8,$B1,$FF,$67,$D6,$A3,$07,$FB,$99,$FF,$7B,$E9,$CF,$B3
	.dc.b	$FD,$C5,$5C,$1B,$5B,$6B,$FF,$00,$3D,$2F,$17,$71,$A4,$8E,$E3,$07
	.dc.b	$7F,$F7,$A8,$FC,$41,$E3,$71,$87,$A1,$B1,$9E,$7F,$37,$CC,$48,$EB
	.dc.b	$6C,$93,$99,$EA,$78,$7C,$27,$8D,$F2,$5B,$1C,$9A,$F0,$B4,$63,$2F
	.dc.b	$F7,$CF,$0B,$E5,$9E,$9F,$F0,$28,$C3,$63,$38,$7C,$29,$CB,$7E,$55
	.dc.b	$F2,$BC,$5B,$19,$8C,$19,$F8,$29,$1C,$5A,$F9,$FC,$D7,$8B,$F8,$56
	.dc.b	$C7,$C6,$6B,$B8,$F0,$65,$23,$C8,$D7,$E1,$79,$FF,$31,$F9,$ED,$73
	.dc.b	$C1,$F9,$38,$C6,$BF,$90,$F4,$1F,$2B,$F9,$76,$DF,$D8,$6B,$F3,$DB
	.dc.b	$3F,$D2,$BE,$51,$F1,$BF,$CB,$A1,$18,$AF,$C6,$C6,$27,$53,$99,$FB
	.dc.b	$52,$F5,$FF,$50,$D8,$FC,$4B,$5D,$C7,$EB,$7D,$7F,$ED,$73,$FA,$A6
	.dc.b	$CF,$EA,$5A,$FE,$75,$BE,$4B,$D8,$7D,$0F,$F4,$68,$C3,$63,$35,$FA
	.dc.b	$69,$CB,$7D,$AF,$D0,$A8,$46,$1B,$D9,$D6,$37,$2F,$11,$F1,$FF,$E6
	.dc.b	$F8,$C3,$63,$31,$FE,$73,$6B,$FE,$72,$F9,$0E,$1E,$61,$A4,$8F,$F3
	.dc.b	$D4,$7E,$A1,$FF,$41,$B5,$FF,$38,$77,$F2,$3C,$1F,$41,$23,$C0,$D7
	.dc.b	$FD,$0F,$F4,$5E,$74,$D2,$3F,$E8,$9D,$1F,$00,$F5,$30,$D9,$FF,$48
	.dc.b	$CE,$A7,$27,$1E,$5F,$E9,$53,$A6,$BF,$E9,$C9,$FC,$8F,$9D,$B3,$C0
	.dc.b	$7F,$0C,$81,$23,$FA,$A7,$26,$89,$1E,$2F,$82,$7A,$BE,$96,$C6,$63
	.dc.b	$07,$FA,$97,$8D,$F2,$0F,$90,$C3,$6B,$FD,$5B,$3A,$9C,$9F,$EA,$DF
	.dc.b	$88,$7C,$8B,$E4,$70,$DB,$FF,$5A,$CE,$A7,$27,$E1,$FC,$AF,$CA,$CF
	.dc.b	$FA,$E9,$AF,$FB,$01,$AF,$C2,$3F,$93,$F5,$FD,$5F,$8D,$FF,$63,$C6
	.dc.b	$77,$46,$23,$E4,$35,$FC,$F3,$70,$F4,$F7,$7A,$09,$3E,$8C,$BF,$70
	.dc.b	$FF,$B4,$BE,$3D,$F2,$0F,$91,$FF,$67,$46,$1B,$19,$9F,$F6,$AC,$E5
	.dc.b	$BE,$53,$1E,$2D,$8C,$C6,$0F,$CB,$7B,$32,$3F,$5C,$D7,$FA,$3F,$6B
	.dc.b	$E9,$3F,$AF,$E5,$FD,$83,$CF,$FF,$70,$C6,$3A,$3E,$57,$D2,$7E,$6D
	.dc.b	$FC,$6D,$6E,$F9,$46,$0F,$C8,$DF,$FB,$AB,$DA,$E5,$8A,$31,$3B,$8C
	.dc.b	$57,$27,$E3,$47,$28,$57,$29,$7D,$2F,$E2,$6B,$FD,$F0,$D8,$FF,$7E
	.dc.b	$B5,$FC,$11,$C3,$9B,$F4,$CF,$8F,$E9,$BF,$C8,$D5,$E3,$C5,$18,$9F
	.dc.b	$19,$FA,$CB,$ED,$97,$C7,$9D,$3F,$0F,$E9,$FD,$F6,$CD,$18,$AD,$E3
	.dc.b	$ED,$8F,$8B,$CC,$ED,$7C,$06,$1B,$5F,$1E,$75,$DB,$9A,$E3,$DE,$F8
	.dc.b	$2A,$66,$F8,$46,$5E,$5F,$21,$FC,$E5,$B8,$78,$A1,$F2,$B6,$3C,$BE
	.dc.b	$2B,$E1,$29,$9B,$F2,$46,$5F,$37,$D4,$7D,$1E,$4F,$E2,$4D,$7F,$A4
	.dc.b	$36,$CD,$FA,$97,$C4,$7C,$46,$DF,$4E,$7C,$4F,$E3,$3D,$27,$D5,$FF
	.dc.b	$14,$AF,$8D,$F5,$7F,$D1,$5F,$C7,$4E,$B8,$FE,$7C,$E7,$4F,$BB,$A3
	.dc.b	$EB,$1F,$0D,$F8,$68,$C3,$63,$38,$FF,$3A,$6D,$7F,$35,$FA,$D7,$C7
	.dc.b	$7D,$C6,$D7,$D9,$5F,$1F,$99,$FC,$0F,$B1,$AB,$F1,$BA,$BF,$C0,$39
	.dc.b	$EF,$FF,$2A,$E7,$FD,$C3,$0F,$C4,$FB,$2F,$A2,$F4,$37,$0F,$E9,$ED
	.dc.b	$C7,$FA,$53,$5F,$F0,$57,$D7,$7E,$DB,$E1,$DC,$48,$C3,$FE,$B8,$BF
	.dc.b	$DC,$4E,$A7,$84,$3E,$B6,$FD,$6E,$1B,$1C,$3E,$BD,$F5,$BF,$B2,$C1
	.dc.b	$18,$9F,$DB,$E5,$FE,$75,$8C,$3F,$BA,$9D,$D7,$1F,$82,$F0,$5F,$91
	.dc.b	$B5,$FB,$8A,$43,$1F,$73,$AE,$0F,$F3,$FC,$BE,$13,$E7,$61,$FE,$93
	.dc.b	$7E,$96,$1C,$DC,$BF,$E8,$97,$35,$CB,$EC,$1F,$2F,$ED,$6D,$FB,$30
	.dc.b	$46,$30,$F6,$35,$FF,$52,$FC,$53,$EB,$7E,$A3,$6F,$E8,$BF,$DF,$27
	.dc.b	$DC,$E5,$F8,$1F,$63,$ED,$9F,$F5,$63,$F0,$5F,$B2,$E6,$3F,$D4,$4E
	.dc.b	$5F,$91,$EF,$FF,$91,$A7,$D0,$D8,$FA,$4D,$7F,$D7,$6D,$7F,$D7,$5F
	.dc.b	$65,$BD,$3E,$6B,$6B,$3E,$DF,$E3,$E3,$15,$FE,$A3,$FA,$FF,$C9,$77
	.dc.b	$BE,$07,$F2,$9F,$00,$9C,$C7,$E5,$FE,$49,$F5,$DF,$8B,$6E,$7F,$3D
	.dc.b	$B8,$FE,$25,$AF,$FA,$EF,$C1,$FD,$07,$CA,$7D,$CE,$4F,$F6,$1F,$81
	.dc.b	$89,$FD,$27,$CB,$7F,$A6,$B1,$FD,$0A,$F9,$9C,$FC,$9E,$9F,$D5,$57
	.dc.b	$E2,$32,$FF,$6F,$B5,$FF,$6F,$35,$FF,$69,$37,$17,$CE,$7F,$B2,$31
	.dc.b	$FD,$7B,$6B,$F3,$8F,$F1,$0E,$4F,$CC,$B9,$7F,$D9,$BF,$47,$FA,$8F
	.dc.b	$81,$FE,$37,$5F,$E4,$48,$FE,$D7,$7F,$E4,$69,$CC,$A7,$87,$27,$F8
	.dc.b	$A5,$DC,$27,$71,$8A,$FE,$DA,$D9,$E1,$EF,$FF,$BA,$A7,$2D,$8C,$C7
	.dc.b	$86,$FF,$C6,$DE,$BB,$EB,$3F,$DE,$38,$F6,$C2,$D6,$A6,$E1,$FE,$F9
	.dc.b	$FB,$3F,$CC,$5F,$BB,$54,$3E,$CB,$71,$8F,$12,$CE,$74,$F1,$61,$CE
	.dc.b	$F2,$6A,$AC,$EA,$DF,$6D,$FD,$A6,$FE,$26,$7F,$E7,$09,$CF,$9B,$E6
	.dc.b	$EA,$7D,$AF,$F8,$57,$5F,$36,$3C,$5B,$5E,$6F,$B6,$FF,$23,$E7,$DE
	.dc.b	$B5,$97,$59,$CE,$7E,$03,$F4,$DE,$07,$F3,$57,$1F,$09,$FE,$1A,$EF
	.dc.b	$F1,$4E,$73,$FC,$1C,$E7,$2F,$82,$DB,$E3,$F0,$7F,$AF,$9F,$8B,$B7
	.dc.b	$FB,$5F,$E8,$BE,$8D,$EF,$0E,$77,$BC,$38,$E5,$E0,$BD,$CE,$6B,$C3
	.dc.b	$F5,$BF,$8E,$BF,$F2,$23,$71,$7D,$BB,$EC,$3B,$CE,$F3,$37,$C5,$FE
	.dc.b	$0D,$AF,$AA,$31,$87,$D1,$FB,$8F,$E4,$99,$FE,$17,$EC,$7F,$C0,$0F
	.dc.b	$E9,$25,$3A,$E2,$FB,$55,$A6,$FC,$DF,$E2,$A3,$1B,$7F,$15,$AE,$0B
	.dc.b	$9D,$3A,$8F,$F2,$16,$7F,$98,$6B,$F9,$07,$F6,$1F,$15,$9F,$8B,$FF
	.dc.b	$28,$69,$FB,$47,$FC,$A5,$5F,$ED,$A3,$F5,$55,$FE,$BE,$8C,$BF,$E9
	.dc.b	$F6,$B9,$8F,$07,$7B,$EC,$FB,$46,$1F,$CB,$7D,$CF,$FB,$D2,$3F,$3D
	.dc.b	$E6,$7D,$17,$F0,$31,$9B,$FC,$86,$9F,$CC,$39,$3F,$5B,$7B,$9C,$B7
	.dc.b	$3F,$9D,$7C,$5F,$DB,$46,$78,$3F,$DA,$FC,$2F,$BA,$FF,$6C,$E7,$F9
	.dc.b	$06,$E3,$F9,$F9,$7D,$DF,$3B,$E3,$5A,$C7,$EE,$36,$3F,$D2,$7E,$6F
	.dc.b	$F8,$12,$73,$87,$E6,$A3,$13,$E2,$FB,$CE,$93,$FE,$97,$FA,$D7,$D0
	.dc.b	$FE,$EE,$73,$D4,$9C,$EB,$FA,$69,$CC,$62,$FD,$11,$8B,$FD,$4C,$E7
	.dc.b	$2F,$D9,$B7,$1F,$17,$D1,$EA,$BF,$85,$7D,$EF,$FB,$DB,$2F,$D9,$B7
	.dc.b	$2F,$88,$FF,$AD,$5B,$EA,$F0,$3F,$90,$FC,$9F,$C3,$8F,$ED,$93,$C7
	.dc.b	$3F,$E5,$13,$9E,$17,$DF,$7B,$71,$FE,$02,$FF,$A2,$70,$FA,$4E,$5B
	.dc.b	$F2,$9F,$C7,$E3,$58,$7E,$B5,$BA,$FD,$E7,$84,$66,$7F,$41,$B3,$C8
	.dc.b	$FD,$27,$5F,$CD,$6E,$DA,$73,$3B,$CB,$F9,$11,$FE,$AF,$71,$87,$EA
	.dc.b	$6B,$84,$EA,$FA,$36,$B7,$97,$EA,$6B,$77,$FB,$19,$CD,$7E,$BD,$AF
	.dc.b	$FA,$F6,$77,$5F,$B1,$9C,$DF,$6B,$F1,$67,$E8,$C6,$63,$E1,$7D,$B1
	.dc.b	$89,$F0,$45,$C3,$C9,$36,$BC,$53,$5F,$14,$5B,$F0,$57,$8E,$BA,$D7
	.dc.b	$72,$5F,$67,$E2,$63,$EA,$7A,$9A,$AF,$0B,$4C,$CF,$13,$DA,$E4,$FA
	.dc.b	$55,$E3,$AE,$91,$83,$57,$F9,$1B,$F3,$46,$31,$F9,$35,$C2,$BD,$73
	.dc.b	$E0,$AF,$1E,$1E,$17,$8B,$63,$F9,$25,$71,$BF,$A6,$74,$C7,$E3,$38
	.dc.b	$C7,$E2,$DF,$1B,$F3,$37,$1F,$34,$78,$48,$6B,$1F,$19,$E2,$23,$A8
	.dc.b	$E2,$EC,$C5,$F8,$DF,$81,$7F,$1F,$0F,$2B,$FE,$A2,$8F,$90,$0F,$8E
	.dc.b	$53,$EB,$94,$53,$2F,$F3,$54,$7F,$AB,$23,$1A,$75,$6C,$66,$76,$F3
	.dc.b	$8C,$E5,$FD,$4E,$79,$B6,$BF,$6F,$F2,$5E,$38,$C5,$FE,$BE,$B7,$9F
	.dc.b	$F5,$56,$C7,$F5,$9A,$E1,$39,$BF,$E0,$35,$AA,$F5,$B7,$05,$F0,$9F
	.dc.b	$EC,$53,$92,$57,$92,$39,$B1,$F2,$39,$69,$30,$F4,$9F,$DC,$6B,$B4
	.dc.b	$78,$A7,$94,$EB,$0F,$EE,$15,$AC,$3E,$3C,$7A,$EF,$A3,$A7,$8B,$9A
	.dc.b	$E5,$1F,$0C,$C9,$EC,$BA,$FD,$CB,$ED,$87,$BA,$3A,$9C,$4A,$79,$CB
	.dc.b	$87,$C8,$79,$BE,$67,$B1,$FC,$96,$7A,$D6,$AB,$E8,$46,$27,$9C,$61
	.dc.b	$D4,$F4,$9D,$5F,$B2,$7C,$4F,$CB,$76,$79,$23,$78,$7C,$32,$47,$AE
	.dc.b	$79,$61,$E7,$8E,$73,$F2,$DF,$1D,$F8,$5E,$B5,$C2,$FA,$C7,$BA,$30
	.dc.b	$EA,$BB,$BF,$34,$FF,$2F,$FC,$B3,$3F,$C7,$60,$8C,$4E,$A7,$3C,$2E
	.dc.b	$EA,$70,$AD,$D7,$F8,$65,$B8,$FF,$94,$6B,$84,$FF,$9B,$BC,$C7,$D1
	.dc.b	$7E,$77,$B9,$F7,$EF,$A5,$C3,$6F,$85,$79,$AF,$8C,$60,$87,$F8,$A5
	.dc.b	$72,$7E,$A7,$A9,$EE,$3E,$E7,$0D,$BF,$F4,$2B,$5F,$F1,$97,$E0,$0F
	.dc.b	$17,$E2,$B6,$37,$5C,$8C,$CC,$8C,$90,$E6,$FC,$EF,$84,$77,$79,$7A
	.dc.b	$FF,$E0,$05,$7F,$A5,$9A,$FF,$90,$E7,$6F,$FA,$59,$AE,$67,$A1,$D6
	.dc.b	$BE,$67,$86,$FA,$CA,$4A,$FF,$4F,$FE,$02,$FA,$EF,$C0,$B6,$3F,$CC
	.dc.b	$B3,$B7,$FC,$A9,$19,$7F,$C1,$DF,$2B,$FC,$23,$F0,$2F,$E5,$7D,$1F
	.dc.b	$EE,$E6,$F6,$FF,$9D,$E7,$F8,$41,$FC,$EF,$B3,$F9,$83,$6F,$47,$F9
	.dc.b	$88,$FF,$40,$39,$69,$FE,$28,$79,$BF,$E2,$AF,$BA,$FC,$57,$9A,$FB
	.dc.b	$DA,$49,$FF,$5E,$FE,$06,$FB,$2F,$CF,$B6,$3F,$DC,$15,$C5,$B5,$AA
	.dc.b	$FC,$57,$A7,$FE,$33,$9F,$F6,$8F,$C0,$BE,$D7,$F1,$CD,$8F,$F5,$B7
	.dc.b	$93,$C0,$8E,$0F,$F6,$7F,$B6,$F8,$1E,$3C,$92,$7F,$DA,$F7,$F6,$5F
	.dc.b	$F6,$AF,$B6,$FB,$3D,$47,$8A,$AF,$2F,$FA,$FE,$B5,$3C,$2B,$3C,$9F
	.dc.b	$DB,$7C,$06,$3D,$6B,$FD,$21,$7C,$1F,$F5,$0F,$BF,$FD,$BF,$16,$1F
	.dc.b	$EF,$5E,$9F,$DC,$7B,$FF,$83,$9F,$F7,$FD,$FE,$EE,$F8,$5F,$F7,$BF
	.dc.b	$21,$18,$9F,$F6,$53,$7C,$C7,$CA,$FC,$DC,$3B,$D7,$FB,$3F,$E0,$DF
	.dc.b	$2B,$FC,$1D,$B1,$FE,$07,$9D,$E2,$9C,$CF,$FA,$E9,$7D,$E7,$13,$95
	.dc.b	$C9,$E5,$9F,$9E,$B7,$1F,$E7,$6D,$BF,$3D,$F5,$52,$30,$43,$F9,$09
	.dc.b	$FE,$14,$8F,$CC,$34,$FC,$E1,$4D,$FA,$6F,$BC,$FE,$D9,$32,$7F,$D0
	.dc.b	$B1,$FC,$97,$EE,$1F,$77,$FE,$23,$F8,$BF,$E6,$BB,$27,$37,$FE,$29
	.dc.b	$9C,$C6,$21,$18,$FB,$4E,$5E,$2F,$D5,$F8,$EF,$D1,$57,$F8,$B7,$04
	.dc.b	$62,$3F,$C6,$51,$8F,$07,$FC,$70,$D8,$FF,$1A,$05,$F2,$2D,$FB,$CE
	.dc.b	$61,$F8,$0E,$2B,$6D,$8E,$13,$EB,$C7,$F1,$5F,$82,$3E,$F3,$83,$67
	.dc.b	$47,$85,$C3,$F0,$27,$8F,$C7,$77,$F5,$5F,$80,$FA,$9B,$5E,$AC,$FF
	.dc.b	$A3,$CE,$4F,$7F,$C6,$FF,$05,$9E,$87,$BD,$7C,$6F,$AD,$B1,$E4,$A4
	.dc.b	$71,$E2,$FB,$3F,$2E,$FA,$9F,$84,$DB,$E0,$D7,$BD,$A3,$FA,$B5,$7E
	.dc.b	$0B,$DB,$FE,$B1,$5F,$59,$AF,$EC,$5A,$FD,$56,$D3,$E6,$7F,$80,$57
	.dc.b	$E3,$1F,$D0,$D7,$81,$B1,$9F,$23,$EC,$7C,$D3,$E3,$E9,$2F,$E0,$43
	.dc.b	$DB,$FB,$DF,$39,$B9,$F9,$CD,$7F,$20,$D7,$F1,$ED,$7F,$85,$7C,$EF
	.dc.b	$F9,$E3,$1F,$D0,$BF,$62,$3E,$2B,$63,$3E,$47,$E3,$7E,$79,$E8,$D2
	.dc.b	$5F,$D5,$87,$D1,$FE,$4F,$E8,$5F,$35,$F9,$E6,$D7,$E1,$5A,$FE,$9D
	.dc.b	$AF,$E9,$BA,$BE,$A9,$FC,$84,$7D,$BF,$E7,$DB,$1F,$85,$76,$FE,$A7
	.dc.b	$E8,$DF,$09,$FA,$56,$C7,$E2,$DA,$FE,$11,$FF,$51,$3F,$A2,$F6,$1F
	.dc.b	$49,$F8,$06,$FC,$37,$FA,$BD,$B3,$FB,$86,$BF,$EA,$D5,$F8,$2F,$EA
	.dc.b	$B8,$7F,$C8,$71,$8F,$07,$F7,$6D,$AF,$05,$6F,$3F,$F5,$D4,$BE,$9F
	.dc.b	$F8,$6E,$1C,$9B,$3A,$6D,$AB,$87,$A3,$DB,$EA,$3F,$89,$4F,$46,$BB
	.dc.b	$6C,$9C,$EA,$FC,$1D,$F8,$3F,$A3,$EA,$FF,$65,$9F,$99,$B8,$AF,$85
	.dc.b	$7B,$BE,$BB,$FB,$3C,$FC,$D6,$BF,$C0,$B2,$46,$35,$F8,$FF,$5F,$E0
	.dc.b	$AF,$A8,$D7,$E7,$35,$FF,$6F,$52,$31,$1F,$29,$AF,$C5,$FB,$0F,$F3
	.dc.b	$7D,$7D,$46,$E7,$F4,$1B,$3F,$2F,$D2,$7D,$4F,$D8,$6C,$7E,$1D,$AF
	.dc.b	$AD,$B7,$F8,$05,$F5,$9F,$80,$FC,$23,$FC,$F7,$5F,$05,$B5,$F8,$66
	.dc.b	$E3,$F8,$2F,$B4,$F1,$E8,$69,$23,$F1,$6F,$F9,$F2,$FF,$96,$FC,$2F
	.dc.b	$FB,$FB,$C0,$FE,$92,$D8,$FE,$7E,$D7,$7E,$E7,$EE,$9C,$B5,$CF,$81
	.dc.b	$CA,$31,$9F,$FA,$63,$E7,$95,$FD,$58,$3F,$D7,$1C,$98,$30,$53,$4E
	.dc.b	$2D,$5F,$CB,$32,$FF,$18,$36,$93,$AF,$27,$C9,$F0,$12,$7D,$2D,$7F
	.dc.b	$A7,$9F,$E0,$2F,$94,$F3,$7D,$C3,$EE,$7F,$BE,$CF,$F7,$28,$C6,$9E
	.dc.b	$96,$C6,$67,$6F,$B0,$47,$F7,$7F,$B9,$E1,$EE,$1A,$49,$F8,$3F,$0B
	.dc.b	$F0,$93,$48,$FF,$81,$3F,$01,$7D,$D7,$F8,$9A,$3E,$7C,$63,$83,$E7
	.dc.b	$36,$38,$4E,$D7,$E1,$2F,$A3,$C3,$6F,$EB,$36,$7F,$C2,$6D,$7D,$DF
	.dc.b	$85,$7F,$8B,$E7,$EE,$36,$3F,$C3,$2D,$77,$F0,$FE,$18,$FF,$3B,$4F
	.dc.b	$0F,$91,$C1,$1F,$C3,$76,$3E,$9B,$D0,$D8,$CC,$60,$F8,$5E,$23,$F0
	.dc.b	$BF,$E7,$F3,$FE,$2A,$8C,$79,$FA,$9C,$E7,$F5,$E7,$35,$BF,$C3,$3F
	.dc.b	$DB,$8D,$F3,$FA,$C6,$2B,$83,$5F,$CD,$D3,$EF,$9F,$39,$E4,$6E,$1E
	.dc.b	$36,$D7,$85,$AF,$2F,$57,$FB,$3C,$FA,$5B,$1E,$76,$BB,$7E,$0F,$E0
	.dc.b	$0F,$C0,$69,$2B,$E1,$46,$5F,$89,$F7,$AF,$AC,$F9,$2D,$8C,$C6,$0F
	.dc.b	$9B,$E4,$BE,$B2,$92,$7E,$74,$65,$FA,$1E,$97,$7C,$FE,$73,$5F,$8E
	.dc.b	$D7,$F8,$3E,$DF,$D3,$73,$F1,$ED,$2F,$CE,$DA,$FB,$1E,$87,$F8,$FE
	.dc.b	$7E,$F3,$63,$3F,$0D,$F8,$48,$C4,$7D,$6E,$C7,$D3,$FE,$21,$B7,$F8
	.dc.b	$76,$CF,$E1,$5A,$FD,$EF,$C3,$5E,$87,$E4,$1B,$5F,$8F,$6C,$FE,$35
	.dc.b	$AF,$E1,$BF,$0E,$7D,$67,$E6,$1B,$5F,$97,$6C,$FE,$55,$AF,$F2,$8F
	.dc.b	$81,$FE,$D9,$F1,$9F,$C0,$5B,$5F,$E8,$86,$CF,$F1,$36,$BC,$7C,$5F
	.dc.b	$F3,$BC,$ED,$B1,$B6,$BB,$F8,$DE,$DF,$87,$FF,$80,$2B,$B3,$67,$B3
	.dc.b	$5F,$33,$5F,$E8,$76,$FC,$43,$F0,$10,$8C,$71,$FA,$E7,$31,$8B,$F6
	.dc.b	$7C,$5F,$D2,$BE,$6D,$79,$36,$7F,$CE,$5C,$3F,$17,$E8,$3E,$8D,$78
	.dc.b	$9A,$F8,$5B,$3F,$EF,$AE,$1F,$99,$F8,$8B,$F0,$1A,$4C,$BE,$74,$65
	.dc.b	$FA,$1F,$81,$F2,$25,$7D,$57,$AF,$E0,$1F,$E7,$D9,$FA,$CD,$CB,$EB
	.dc.b	$35,$FB,$CD,$7E,$4F,$E2,$4F,$57,$F0,$0D,$AF,$C0,$36,$7F,$00,$D7
	.dc.b	$F1,$1E,$7F,$FB,$A1,$FA,$0D,$D3,$FA,$14,$7E,$2E,$31,$E9,$FF,$BF
	.dc.b	$AF,$F1,$8D,$9F,$C6,$35,$FC,$A3,$5F,$BF,$F6,$6F,$C0,$BF,$2C,$DA
	.dc.b	$FC,$B3,$67,$F2,$CD,$7F,$3D,$F8,$9B,$EB,$7F,$3C,$DA,$FC,$F3,$67
	.dc.b	$F3,$CD,$7F,$B9,$71,$7E,$15,$34,$9C,$7F,$D8,$DF,$E5,$71,$FE,$14
	.dc.b	$6B,$FE,$1A,$6B,$CB,$C9,$E0,$2F,$FC,$3F,$D4,$F8,$9E,$5C,$D9,$BF
	.dc.b	$EF,$71,$D6,$F1,$5B,$9E,$D5,$8A,$FD,$3F,$D7,$FE,$36,$78,$9F,$A7
	.dc.b	$7F,$BE,$57,$1E,$14,$6E,$30,$EA,$B9,$F8,$3E,$3C,$BF,$85,$4F,$F8
	.dc.b	$15,$B1,$9C,$37,$1F,$E3,$A3,$F8,$87,$D3,$D8,$F6,$1F,$B9,$71,$1A
	.dc.b	$8E,$46,$4B,$8B,$F0,$2F,$E6,$D9,$7A,$5B,$97,$F0,$6C,$3B,$4E,$7C
	.dc.b	$5F,$89,$87,$F0,$7F,$C5,$1F,$61,$8C,$91,$8C,$FE,$9E,$5F,$C4,$23
	.dc.b	$FC,$1F,$7F,$2A,$B7,$7C,$67,$84,$FB,$27,$37,$EF,$8F,$05,$7A,$AB
	.dc.b	$73,$FC,$CD,$E3,$8F,$D3,$F8,$67,$D4,$75,$BF,$34,$62,$FC,$B8,$7F
	.dc.b	$21,$FC,$09,$F6,$1E,$B7,$F8,$26,$3F,$75,$CD,$70,$CD,$1A,$D3,$F9
	.dc.b	$6B,$FC,$2E,$7F,$A3,$37,$0C,$FC,$EF,$D2,$8C,$47,$F3,$1F,$B2,$FE
	.dc.b	$83,$5F,$D3,$5B,$1F,$CE,$1A,$EF,$C8,$FB,$7F,$81,$7F,$A6,$AB,$C8
	.dc.b	$D9,$F9,$D5,$FD,$33,$3F,$CB,$4E,$5F,$D0,$4F,$ED,$3D,$3F,$BA,$DB
	.dc.b	$F8,$AF,$EE,$39,$73,$D3,$FE,$0D,$8F,$D1,$9B,$3F,$26,$59,$FE,$03
	.dc.b	$5F,$92,$8C,$1F,$BD,$7B,$FC,$57,$EE,$67,$BD,$75,$C3,$CB,$39,$BF
	.dc.b	$C4,$7E,$02,$FB,$8F,$C8,$3F,$CB,$EB,$F6,$4E,$70,$FA,$CE,$78,$3F
	.dc.b	$B0,$B6,$FF,$B0,$37,$84,$F1,$8E,$FE,$2F,$E7,$7F,$01,$D7,$E0,$3F
	.dc.b	$A4,$C3,$FB,$5C,$62,$BF,$45,$39,$7F,$B6,$37,$04,$EF,$1F,$C5,$7E
	.dc.b	$01,$E1,$26,$9F,$AB,$71,$97,$EE,$1F,$E9,$93,$FE,$11,$6E,$3F,$DF
	.dc.b	$1A,$FF,$7B,$7D,$DF,$1E,$F8,$27,$DF,$75,$C3,$F7,$73,$98,$FD,$D7
	.dc.b	$E0,$0F,$A8,$A6,$3E,$FE,$E7,$FA,$19,$D3,$67,$4D,$3F,$4D,$F2,$9F
	.dc.b	$E7,$EE,$3F,$E4,$CF,$F9,$77,$5F,$E0,$4E,$78,$BF,$C2,$AE,$67,$7A
	.dc.b	$FF,$0D,$8C,$4E,$A3,$1F,$03,$F8,$A5,$F2,$FF,$98,$E6,$F8,$0D,$CF
	.dc.b	$4E,$67,$79,$7E,$6A,$B6,$DA,$DB,$99,$DE,$5F,$94,$6E,$3F,$EE,$AB
	.dc.b	$F8,$75,$BC,$3B,$11,$E9,$9D,$C3,$94,$DB,$E5,$8F,$82,$73,$7C,$CE
	.dc.b	$25,$72,$49,$E3,$7C,$E7,$FE,$00,$E1,$EC,$AD,$46,$2F,$B5,$FA,$71
	.dc.b	$F5,$CF,$06,$BF,$F0,$02,$76,$E7,$1E,$CE,$71,$B8,$F7,$C3,$A5,$7D
	.dc.b	$19,$CE,$1D,$2B,$57,$E3,$9C,$D7,$2C,$3B,$CA,$B6,$DB,$E5,$3A,$9C
	.dc.b	$91,$C4,$77,$37,$7D,$AB,$E5,$61,$F2,$1A,$7B,$4A,$7B,$4B,$7D,$AB
	.dc.b	$9C,$61,$ED,$87,$BE,$30,$FA,$EF,$C1,$8F,$D2,$9E,$8D,$BE,$95,$C6
	.dc.b	$31,$7D,$2B,$F0,$31,$89,$44,$3E,$79,$4F,$51,$6B,$F3,$AD,$CB,$FD
	.dc.b	$00,$EC,$F9,$CE,$1D,$72,$65,$CB,$67,$FD,$B0,$FE,$BE,$79,$DF,$F3
	.dc.b	$86,$E5,$FA,$5F,$43,$CD,$39,$AE,$04,$C3,$F6,$53,$AA,$FD,$65,$F2
	.dc.b	$9F,$CD,$B5,$F1,$9C,$0E,$57,$D5,$FD,$E4,$6F,$0F,$F3,$6C,$63,$0D
	.dc.b	$DF,$67,$F8,$46,$1E,$19,$EF,$1A,$69,$DC,$E4,$4C,$3C,$07,$F5,$89
	.dc.b	$E5,$1E,$68,$CD,$77,$8E,$51,$87,$AB,$5F,$D0,$4F,$2B,$F9,$F1,$DE
	.dc.b	$B9,$57,$78,$F3,$1D,$E3,$27,$D1,$73,$3C,$A7,$B3,$F1,$1B,$87,$C7
	.dc.b	$3E,$45,$F8,$5C,$61,$CA,$30,$F4,$8D,$DD,$8D,$34,$F4,$1F,$C5,$27
	.dc.b	$8B,$83,$D0,$64,$E1,$5E,$68,$DB,$E6,$9F,$AF,$3A,$C3,$DE,$F1,$3A
	.dc.b	$61,$EF,$8C,$C7,$CC,$38,$1F,$41,$F8,$CF,$0C,$3D,$EF,$D6,$8F,$F7
	.dc.b	$4F,$03,$A7,$F0,$D8,$3F,$00,$FF,$39,$E5,$FE,$83,$6D,$F0,$AD,$DF
	.dc.b	$1C,$3F,$C9,$55,$A7,$FD,$E4,$DC,$3F,$BB,$D7,$6F,$31,$C7,$49,$7F
	.dc.b	$E5,$38,$FE,$F0,$D8,$CB,$FE,$6F,$8E,$2E,$F8,$7D,$3F,$80,$7F,$8A
	.dc.b	$E3,$FC,$C7,$CA,$FA,$E7,$C5,$7E,$8E,$30,$F0,$EA,$95,$FE,$75,$D1
	.dc.b	$87,$38,$FF,$3D,$37,$9B,$F8,$B4,$EA,$BF,$8C,$46,$27,$87,$0F,$FA
	.dc.b	$AD,$BC,$3F,$C7,$AF,$86,$1F,$E7,$19,$FF,$79,$3F,$E7,$3F,$AA,$AF
	.dc.b	$FA,$35,$B7,$FE,$B5,$6D,$F3,$7A,$CA,$D1,$86,$F5,$7D,$E8,$E3,$3B
	.dc.b	$75,$F7,$EF,$BC,$FC,$1B,$67,$FD,$0C,$D7,$FD,$40,$BF,$00,$D9,$C2
	.dc.b	$64,$F0,$8F,$E0,$1E,$8F,$F0,$1F,$87,$93,$FE,$1D,$8C,$BF,$EA,$53
	.dc.b	$FB,$33,$F3,$1F,$F0,$A3,$5F,$F0,$1B,$5D,$9C,$C8,$7F,$88,$FD,$1F
	.dc.b	$CE,$F9,$0F,$B6,$A4,$9F,$F5,$17,$E1,$9F,$E2,$A7,$FD,$78,$DE,$2F
	.dc.b	$BE,$D8,$FB,$4D,$7E,$83,$77,$79,$EF,$94,$F4,$46,$1B,$19,$AF,$C8
	.dc.b	$35,$FF,$64,$2F,$96,$FD,$AB,$9E,$4F,$F6,$25,$62,$7F,$73,$EB,$3E
	.dc.b	$AF,$F7,$7E,$AB,$E7,$E9,$A8,$99,$7E,$7E,$B8,$36,$3C,$F1,$88,$FF
	.dc.b	$6D,$37,$1F,$F7,$74,$F2,$9F,$F7,$14,$FE,$B9,$FF,$66,$70,$7C,$BA
	.dc.b	$74,$7F,$BC,$FC,$7F,$E1,$F3,$98,$FD,$27,$63,$97,$F8,$9B,$70,$FD
	.dc.b	$A3,$5F,$F7,$63,$78,$BF,$66,$D8,$FD,$77,$B1,$FE,$E8,$7F,$78,$D7
	.dc.b	$FD,$E8,$D9,$FD,$DB,$63,$8E,$1F,$C9,$23,$86,$5E,$A9,$FD,$74,$63
	.dc.b	$1D,$C7,$F9,$DF,$83,$FA,$F7,$04,$8C,$10,$FF,$0B,$78,$FF,$7D,$FF
	.dc.b	$0E,$4F,$F8,$55,$FF,$23,$36,$1F,$81,$D2,$4F,$F8,$65,$FF,$0E,$72
	.dc.b	$3D,$E7,$CC,$61,$B5,$FE,$25,$9D,$4E,$4E,$3F,$02,$F1,$A9,$2B,$84
	.dc.b	$65,$E5,$E6,$3F,$08,$F0,$B6,$33,$18,$3C,$BF,$8A,$FB,$71,$34,$93
	.dc.b	$EA,$F8,$57,$C2,$52,$47,$A6,$32,$FA,$FE,$19,$F5,$BF,$09,$B1,$98
	.dc.b	$C1,$F1,$BE,$25,$EE,$52,$4F,$C5,$8C,$BF,$1F,$DC,$E6,$FF,$94,$F3
	.dc.b	$46,$23,$FA,$1B,$5F,$E7,$CD,$7F,$9A,$7C,$53,$E1,$7E,$BB,$63,$EA
	.dc.b	$35,$EC,$D7,$FA,$7F,$C5,$BE,$A7,$EF,$B6,$3E,$E3,$5F,$33,$5F,$EA
	.dc.b	$FF,$1D,$FB,$57,$D4,$D7,$3E,$07,$E2,$A3,$11,$FD,$5F,$BB,$F4,$0F
	.dc.b	$C1,$6B,$FD,$89,$AE,$FB,$7F,$27,$F2,$8F,$BD,$A4,$9F,$C9,$46,$5F
	.dc.b	$F3,$C2,$FA,$FF,$E6,$51,$FE,$83,$8C,$78,$BF,$E7,$96,$C7,$09,$DB
	.dc.b	$FE,$8A,$87,$CA,$FD,$16,$F2,$71,$8C,$19,$E4,$F9,$D5,$C5,$AA,$BF
	.dc.b	$0F,$C3,$E1,$9C,$C3,$E5,$BF,$93,$57,$9D,$AF,$8D,$AF,$36,$C3,$E6
	.dc.b	$5E,$37,$B5,$B5,$F1,$1A,$FF,$A8,$E7,$84,$FC,$3F,$98,$FB,$D3,$F0
	.dc.b	$5A,$FA,$5B,$26,$FB,$2F,$27,$E6,$36,$BE,$9B,$5F,$03,$5F,$A9,$F8
	.dc.b	$B7,$E8,$9F,$B2,$D7,$CC,$D7,$7D,$3E,$7E,$69,$1F,$80,$C3,$F1,$EF
	.dc.b	$9E,$3E,$F7,$C8,$BF,$08,$A4,$9F,$C3,$C6,$5F,$C4,$FC,$EB,$D7,$FC
	.dc.b	$33,$63,$31,$83,$F1,$5F,$3D,$FD,$DA,$3F,$22,$D7,$E7,$35,$FC,$93
	.dc.b	$5F,$C7,$7C,$FB,$E1,$FF,$28,$D8,$FC,$DB,$5F,$B0,$D7,$F3,$9F,$8B
	.dc.b	$FF,$8A,$5F,$D0,$B5,$FB,$CD,$77,$EE,$7F,$45,$DD,$F3,$74,$93,$FE
	.dc.b	$DD,$8C,$BF,$ED,$FF,$C6,$1A,$FF,$4D,$6C,$66,$30,$67,$A6,$47,$16
	.dc.b	$BF,$EE,$DF,$C6,$7F,$A5,$B7,$D0,$FE,$B1,$18,$8F,$F7,$C3,$5F,$F7
	.dc.b	$E7,$C3,$7E,$69,$FF,$7E,$35,$CF,$4F,$F6,$18,$C4,$07,$A0,$FA,$97
	.dc.b	$D5,$7F,$BE,$E3,$0D,$8C,$CF,$12,$4E,$9C,$1D,$57,$D5,$3E,$AB,$A3
	.dc.b	$67,$67,$89,$C3,$CA,$78,$CF,$F7,$98,$79,$14,$8F,$F7,$D7,$9A,$FA
	.dc.b	$AF,$E8,$9D,$FE,$FC,$BF,$61,$18,$9F,$7C,$FA,$FF,$1A,$FF,$22,$3B
	.dc.b	$3D,$D3,$BA,$E3,$C1,$FB,$B7,$9F,$65,$AD,$CD,$C3,$E7,$5F,$EB,$E7
	.dc.b	$A4,$7D,$1C,$1F,$7F,$FC,$06,$30,$D5,$1A,$9F,$6B,$63,$E4,$FD,$5B
	.dc.b	$E1,$3C,$10,$8C,$57,$82,$39,$DE,$A7,$31,$F0,$8F,$C2,$38,$CF,$F8
	.dc.b	$B3,$F1,$23,$47,$5B,$E5,$EF,$FB,$FC,$C7,$D7,$FF,$07,$CB,$DF,$18
	.dc.b	$E5,$D4,$E7,$0F,$04,$E6,$B7,$3F,$CE,$1F,$8F,$E3,$BB,$FF,$DB,$B1
	.dc.b	$AC,$3E,$74,$E7,$E4,$7C,$5F,$87,$FF,$82,$AB,$C2,$BD,$8C,$F9,$3A
	.dc.b	$64,$F6,$F1,$DE,$EE,$36,$F8,$99,$27,$35,$E4,$FA,$CE,$8F,$F6,$E6
	.dc.b	$D7,$F4,$B6,$BF,$DB,$DA,$FF,$6B,$7C,$CF,$66,$F1,$FF,$47,$B4,$74
	.dc.b	$8C,$3A,$AF,$EE,$9F,$59,$FE,$30,$7E,$3B,$6F,$E6,$FD,$A3,$E6,$3E
	.dc.b	$3C,$E7,$0E,$CF,$E5,$6B,$FB,$13,$5F,$F0,$33,$5E,$FD,$7F,$97,$C9
	.dc.b	$2F,$97,$FA,$7E,$7F,$F2,$AC,$FC,$71,$8F,$27,$EB,$FA,$5F,$E2,$F9
	.dc.b	$FB,$AD,$CB,$F0,$7F,$87,$FF,$B8,$67,$3E,$E7,$DE,$F5,$5F,$83,$F8
	.dc.b	$5B,$3F,$E1,$E6,$DF,$F7,$B6,$BF,$C5,$FE,$FF,$FC,$21,$DD,$FB,$57
	.dc.b	$FC,$25,$F1,$57,$83,$F3,$DC,$5F,$DD,$E7,$38,$FE,$4B,$EB,$DF,$31
	.dc.b	$F9,$06,$C7,$EF,$27,$FC,$01,$E8,$66,$BF,$81,$7C,$1F,$F8,$7F,$EA
	.dc.b	$7F,$7C,$DB,$FF,$1B,$C6,$32,$3F,$C0,$F0,$7D,$E7,$F4,$EA,$EC,$FF
	.dc.b	$13,$7F,$9D,$46,$B1,$F6,$4E,$76,$62,$F8,$37,$D6,$F9,$1B,$8F,$A7
	.dc.b	$EF,$7E,$85,$FF,$03,$BE,$17,$EA,$F8,$6F,$CC,$CF,$46,$C9,$A6,$B3
	.dc.b	$7D,$FF,$C6,$AF,$E1,$D1,$8F,$B3,$FA,$53,$9E,$2F,$A0,$A3,$E4,$F8
	.dc.b	$BF,$E5,$FD,$3F,$94,$E5,$FE,$86,$F8,$7F,$B5,$F1,$3F,$17,$3F,$45
	.dc.b	$ED,$F5,$7C,$56,$5F,$EB,$16,$8C,$47,$F6,$38,$C7,$17,$D6,$9C,$CB
	.dc.b	$57,$E0,$0F,$B8,$FC,$52,$F9,$9B,$18,$7F,$03,$F1,$CF,$C2,$A9,$F9
	.dc.b	$1F,$8F,$7D,$76,$38,$7D,$BF,$68,$F8,$DF,$C3,$B6,$3F,$D9,$CD,$D7
	.dc.b	$FD,$50,$D7,$F3,$DF,$86,$FD,$AE,$7F,$CD,$2F,$A2,$FD,$03,$9B,$FC
	.dc.b	$4B,$97,$E6,$FC,$17,$AD,$BB,$7E,$9F,$F0,$5F,$F7,$0C,$60,$FA,$7F
	.dc.b	$32,$FA,$5F,$CE,$B6,$3F,$DC,$8D,$CF,$FD,$7C,$D7,$F6,$9C,$2F,$C1
	.dc.b	$FF,$68,$BF,$0F,$B1,$87,$FA,$EF,$A5,$FC,$3F,$0F,$EB,$8D,$BF,$EA
	.dc.b	$AF,$F1,$D8,$FF,$0B,$39,$DF,$FC,$E1,$F8,$97,$86,$BF,$C3,$4D,$7F
	.dc.b	$C4,$8D,$4D,$EC,$FF,$7D,$5F,$F6,$B6,$CF,$F6,$49,$F7,$57,$F9,$F6
	.dc.b	$72,$FF,$A0,$7B,$3F,$CD,$5E,$A7,$FA,$0F,$E1,$DD,$1F,$91,$F9,$4F
	.dc.b	$E6,$51,$8C,$BF,$21,$F8,$5F,$F8,$5E,$31,$87,$2D,$3F,$C7,$51,$8B
	.dc.b	$E3,$87,$3F,$89,$EC,$F5,$CA,$E5,$FC,$12,$BA,$4F,$F0,$AF,$C3,$1F
	.dc.b	$3D,$AC,$DF,$8D,$FE,$CF,$4E,$F7,$BA,$D4,$E7,$1F,$F1,$AC,$FE,$AE
	.dc.b	$31,$A7,$A1,$B1,$C6,$F9,$D7,$F1,$69,$CB,$FC,$63,$EE,$DF,$45,$E3
	.dc.b	$6D,$70,$F1,$3F,$BB,$E5,$9A,$93,$D3,$35,$BD,$4C,$E3,$E6,$3B,$3D
	.dc.b	$4E,$DE,$8D,$BE,$55,$FC,$D1,$E9,$F7,$3F,$BB,$6F,$DD,$FC,$78,$C6
	.dc.b	$1D,$71,$42,$74,$4A,$E7,$E3,$5B,$E1,$F5,$BF,$CE,$27,$D4,$D8,$F4
	.dc.b	$6D,$FC,$EB,$E0,$7E,$B4,$E7,$8B,$F0,$AD,$E2,$FE,$38,$FF,$0E,$C3
	.dc.b	$9E,$9C,$3C,$8F,$C1,$E5,$FC,$3F,$D9,$BD,$7F,$E0,$9C,$BF,$25,$39
	.dc.b	$C3,$EB,$A9,$D6,$5F,$E8,$37,$FC,$21,$3F,$E8,$26,$CE,$7D,$AF,$F7
	.dc.b	$14,$E5,$B1,$98,$EB,$3F,$BB,$BF,$0B,$DA,$7F,$D3,$0E,$6F,$F9,$C7
	.dc.b	$DC,$FF,$23,$AF,$D2,$B6,$3F,$B2,$FC,$DF,$F1,$19,$CF,$99,$F9,$DA
	.dc.b	$D6,$1F,$D9,$9B,$C0,$BE,$18,$76,$BF,$D2,$7D,$C7,$F5,$BC,$7F,$66
	.dc.b	$D9,$FE,$94,$FE,$6B,$73,$9C,$BF,$B9,$B6,$33,$EE,$74,$F5,$7A,$BF
	.dc.b	$B1,$FB,$8F,$2C,$9E,$06,$0B,$1B,$38,$65,$77,$FD,$14,$E9,$B1,$9C
	.dc.b	$7F,$2D,$F7,$1F,$E6,$C6,$FD,$AF,$F0,$18,$C6,$BF,$C4,$14,$67,$E3
	.dc.b	$B4,$DC,$3F,$61,$3A,$7F,$65,$F7,$0F,$B8,$FE,$18,$FF,$BF,$63,$31
	.dc.b	$8F,$A3,$CB,$63,$27,$28,$FC,$BF,$D6,$F3,$F9,$6F,$C9,$B7,$4F,$DE
	.dc.b	$07,$C4,$E1,$D6,$9F,$9C,$6F,$1F,$FB,$6D,$B3,$FE,$E5,$78,$CE,$F1
	.dc.b	$5F,$81,$B5,$D5,$B3,$D7,$4F,$3C,$63,$0E,$CD,$AE,$CD,$7D,$13,$C2
	.dc.b	$FB,$12,$3D,$51,$97,$C1,$19,$BF,$06,$57,$0E,$F3,$6B,$D9,$7C,$6F
	.dc.b	$D9,$39,$9F,$0D,$DC,$7E,$1C,$7C,$BC,$7C,$D1,$F1,$31,$EB,$3E,$22
	.dc.b	$61,$E2,$6C,$6A,$3B,$C7,$AA,$FB,$5F,$43,$AE,$3E,$5B,$E0,$4A,$F4
	.dc.b	$D7,$0C,$7C,$D5,$B9,$F9,$71,$89,$D4,$78,$8E,$25,$3D,$91,$B7,$18
	.dc.b	$FC,$02,$57,$58,$D9,$A7,$11,$BC,$3D,$37,$E0,$C7,$DE,$7C,$32,$E3
	.dc.b	$EB,$C3,$9C,$66,$31,$7C,$F1,$FC,$EE,$3D,$B1,$F8,$D1,$8C,$7D,$B8
	.dc.b	$8F,$A6,$53,$E1,$94,$A5,$B6,$5F,$E8,$68,$D1,$FA,$97,$0E,$DA,$FE
	.dc.b	$9D,$AF,$E9,$DE,$59,$7E,$B5,$B9,$6F,$87,$F6,$FF,$23,$E5,$AD,$47
	.dc.b	$EE,$AB,$77,$CE,$36,$E5,$8F,$EE,$E3,$6F,$EF,$EB,$A5,$F7,$8D,$57
	.dc.b	$0B,$FE,$01,$5A,$9E,$C6,$89,$1A,$78,$57,$38,$F3,$C7,$F0,$B9,$E4
	.dc.b	$FC,$7A,$DB,$D2,$79,$3D,$2B,$91,$1F,$E1,$D5,$CB,$0F,$A7,$7C,$A7
	.dc.b	$A1,$DC,$C4,$69,$E8,$72,$33,$87,$C4,$8D,$61,$E2,$3D,$24,$C3,$C5
	.dc.b	$5F,$24,$E3,$7D,$B0,$F2,$61,$CC,$91,$F5,$27,$C2,$79,$AF,$A3,$E3
	.dc.b	$33,$87,$8B,$0E,$EB,$0E,$B7,$E0,$AF,$71,$C3,$0E,$EE,$30,$EA,$F7
	.dc.b	$C3,$9D,$F2,$8E,$17,$E5,$AD,$E1,$E1,$9F,$A9,$7F,$CF,$EF,$85,$6B
	.dc.b	$0F,$91,$18,$9F,$3C,$62,$BC,$D5,$D2,$BF,$15,$83,$B5,$F1,$B8,$FB
	.dc.b	$86,$94,$DC,$EB,$1F,$F2,$B3,$5F,$F2,$44,$F0,$7F,$CB,$4D,$7F,$C9
	.dc.b	$91,$BC,$7F,$D0,$7F,$5D,$FD,$D2,$7F,$CC,$0B,$8F,$CA,$D8,$CF,$0E
	.dc.b	$0C,$CC,$8C,$9F,$E6,$9F,$B7,$FF,$69,$37,$97,$DF,$18,$8F,$F2,$F2
	.dc.b	$E7,$CC,$C9,$E0,$73,$8E,$26,$5E,$31,$87,$FC,$F1,$EB,$F5,$CB,$FD
	.dc.b	$3E,$DE,$2F,$F7,$B3,$65,$C1,$FE,$D1,$E9,$FA,$73,$98,$FF,$6A,$61
	.dc.b	$83,$33,$F6,$3E,$DF,$FC,$7D,$7F,$E8,$76,$BF,$E7,$96,$E3,$FE,$B3
	.dc.b	$ED,$FF,$22,$78,$BC,$7B,$A5,$F1,$7F,$86,$4E,$A3,$FC,$2B,$F3,$1F
	.dc.b	$BD,$F3,$E4,$60,$87,$DD,$3F,$C3,$11,$FC,$E7,$9B,$F9,$B3,$63,$FC
	.dc.b	$2C,$D7,$FA,$07,$E2,$1E,$6E,$0C,$C7,$4C,$FF,$CF,$33,$98,$F8,$FF
	.dc.b	$6F,$FE,$BE,$7F,$D4,$CF,$F8,$69,$FF,$4C,$36,$7F,$D9,$EE,$E3,$3E
	.dc.b	$7F,$E8,$23,$13,$A8,$F8,$3F,$15,$FE,$B8,$AF,$F5,$97,$DB,$BC,$9F
	.dc.b	$D3,$B6,$33,$18,$F4,$64,$60,$87,$FA,$AE,$7A,$3F,$EA,$DF,$B6,$7D
	.dc.b	$77,$EC,$97,$DB,$D3,$11,$9A,$B5,$FE,$C5,$AE,$2D,$8D,$34,$87,$FA
	.dc.b	$C3,$C2,$F8,$9B,$FF,$6F,$E9,$F0,$F1,$FF,$02,$C4,$D3,$F0,$7E,$EB
	.dc.b	$A6,$99,$C7,$FD,$AD,$5C,$1F,$F6,$E4,$ED,$FF,$6D,$7D,$B7,$C7,$7F
	.dc.b	$D9,$8B,$97,$F8,$8B,$63,$7C,$18,$23,$83,$FD,$D1,$ED,$BE,$0F,$FA
	.dc.b	$87,$17,$F8,$1F,$DD,$FE,$1D,$E4,$FC,$E6,$C6,$7A,$BF,$C1,$9E,$F3
	.dc.b	$EB,$31,$97,$F8,$46,$30,$FF,$82,$83,$87,$FC,$1A,$D7,$FC,$28,$D7
	.dc.b	$FC,$31,$E0,$BD,$AE,$B0,$8C,$3F,$E1,$FC,$7F,$92,$46,$21,$18,$FB
	.dc.b	$EF,$2F,$80,$FC,$39,$F8,$06,$1F,$E7,$17,$CF,$4F,$F1,$55,$6F,$CC
	.dc.b	$D5,$FB,$BA,$5F,$10,$F8,$FF,$74,$61,$B1,$9C,$7F,$C5,$73,$96,$F6
	.dc.b	$FF,$BA,$5F,$01,$FC,$69,$C3,$EA,$CB,$F1,$D1,$8D,$7F,$A2,$66,$F8
	.dc.b	$97,$D6,$FB,$BF,$0D,$7C,$9D,$E1,$99,$F6,$C1,$5A,$8F,$0E,$8F,$8B
	.dc.b	$7E,$0B,$EC,$6C,$FA,$9A,$FF,$92,$E7,$84,$FC,$2C,$FF,$C8,$FA,$7F
	.dc.b	$59,$9F,$31,$F2,$63,$3B,$7C,$EF,$C3,$7F,$1C,$FD,$0F,$93,$79,$3F
	.dc.b	$61,$7C,$F5,$31,$1F,$4F,$B7,$FC,$A3,$A7,$F4,$26,$E5,$FE,$C1,$3E
	.dc.b	$AB,$8F,$96,$7C,$6F,$DB,$6D,$7C,$26,$BF,$44,$70,$A1,$F2,$BF,$8C
	.dc.b	$F9,$77,$A9,$7E,$2A,$61,$F3,$E7,$2F,$CE,$F9,$AF,$E2,$AF,$E1,$1A
	.dc.b	$FC,$B6,$BF,$E7,$9D,$11,$88,$FE,$21,$E9,$BE,$4F,$F2,$8D,$8F,$B6
	.dc.b	$D7,$D5,$A3,$F1,$C7,$DD,$7E,$93,$E7,$7F,$90,$D6,$9F,$1F,$47,$E3
	.dc.b	$9B,$CD,$FA,$47,$71,$FE,$15,$F9,$E7,$DD,$FF,$45,$BF,$F4,$84,$63
	.dc.b	$D9,$FE,$04,$D8,$FF,$46,$B5,$FF,$4A,$3E,$4D,$FF,$85,$B6,$3F,$4A
	.dc.b	$D7,$D7,$8B,$D0,$FD,$EC,$FF,$87,$67,$DE,$FF,$88,$DB,$F0,$5A,$F8
	.dc.b	$2E,$BF,$46,$F6,$BC,$AD,$AF,$2B,$5F,$2B,$5F,$6F,$DD,$3E,$BB,$CE
	.dc.b	$D8,$F3,$B5,$FC,$57,$C5,$3F,$0C,$79,$1E,$EF,$B9,$FD,$E6,$FF,$EB
	.dc.b	$9D,$BF,$8F,$3F,$0B,$D2,$FF,$20,$72,$7C,$FA,$DE,$4B,$47,$D1,$8E
	.dc.b	$4D,$74,$D8,$DF,$1F,$D4,$87,$D6,$F8,$C9,$5F,$C5,$34,$7D,$5F,$ED
	.dc.b	$A3,$C1,$E8,$BE,$7B,$EB,$B6,$3E,$BB,$5F,$D5,$65,$FC,$9F,$E5,$F6
	.dc.b	$7E,$12,$4D,$24,$FF,$B4,$B0,$F8,$FF,$7B,$F8,$4F,$BC,$FD,$8C,$FE
	.dc.b	$11,$AF,$CE,$E1,$4F,$F5,$4A,$F9,$FF,$19,$67,$35,$ED,$6B,$C6,$38
	.dc.b	$6D,$FE,$73,$E0,$FC,$F7,$DE,$3E,$97,$F2,$AD,$BF,$8D,$5B,$E4,$FF
	.dc.b	$79,$65,$F5,$35,$E7,$F2,$72,$F7,$1F,$D7,$3E,$F3,$F8,$7E,$7F,$E7
	.dc.b	$D7,$E4,$37,$8B,$5E,$6C,$BF,$E8,$E2,$B8,$E5,$FE,$FF,$6C,$07,$CB
	.dc.b	$FF,$8F,$70,$FF,$45,$3F,$E3,$F8,$D1,$B7,$0E,$1A,$E7,$3F,$EF,$8B
	.dc.b	$EE,$FF,$C2,$95,$FE,$95,$7F,$40,$DE,$CF,$EF,$31,$AC,$7B,$B6,$BC
	.dc.b	$CD,$7A,$7C,$37,$F8,$03,$DA,$FF,$11,$47,$F3,$33,$FB,$8F,$B7,$FD
	.dc.b	$F9,$F4,$9E,$FF,$77,$FB,$1E,$5F,$E1,$16,$C6,$8E,$31,$9F,$BB,$7D
	.dc.b	$DD,$24,$EE,$93,$A8,$F8,$3E,$77,$C2,$4D,$23,$E4,$FC,$05,$ED,$7F
	.dc.b	$B1,$76,$FE,$FF,$18,$F8,$2F,$F5,$F6,$7F,$EB,$F8,$7E,$23,$FE,$5A
	.dc.b	$AD,$DF,$1C,$3F,$AA,$57,$4F,$83,$E3,$5F,$7E,$30,$F0,$C7,$34,$9C
	.dc.b	$BF,$2F,$ED,$BF,$D8,$D5,$FE,$37,$6C,$7C,$76,$BF,$39,$AF,$7F,$73
	.dc.b	$F7,$33,$EC,$6C,$75,$F3,$7C,$36,$9C,$E9,$EF,$7F,$B9,$DF,$1E,$9D
	.dc.b	$C6,$A3,$B4,$7A,$B9,$DF,$73,$FC,$C6,$7F,$8B,$37,$2F,$E5,$EF,$FA
	.dc.b	$3E,$33,$D1,$C7,$CA,$7C,$2D,$3A,$57,$1C,$FF,$DE,$18,$FF,$93,$E7
	.dc.b	$C1,$F7,$0F,$B6,$A6,$3E,$06,$E7,$FC,$9F,$F0,$A7,$E1,$1E,$76,$C7
	.dc.b	$F2,$BB,$E4,$EE,$7D,$9F,$85,$FF,$A6,$A3,$F9,$03,$77,$FC,$EE,$68
	.dc.b	$C4,$7E,$E2,$31,$F2,$9F,$EC,$98,$43,$D5,$EF,$1F,$D4,$9B,$B7,$BA
	.dc.b	$FB,$3F,$02,$33,$87,$F5,$3F,$57,$E3,$BF,$D0,$FD,$2F,$F2,$6B,$FD
	.dc.b	$7B,$83,$EA,$46,$31,$FE,$BF,$8F,$F7,$5A,$F5,$FB,$1E,$5C,$FD,$4D
	.dc.b	$9D,$57,$BB,$F0,$D7,$CE,$FD,$F6,$E1,$FD,$4F,$4F,$EA,$EF,$F5,$58
	.dc.b	$CD,$7F,$6E,$FB,$AF,$E6,$D3,$FD,$65,$FF,$32,$B7,$D0,$FC,$54,$62
	.dc.b	$7F,$BA,$FB,$BF,$EA,$3A,$FE,$A0,$D8,$FA,$2D,$7E,$83,$5F,$F4,$3F
	.dc.b	$AE,$F9,$3F,$D7,$FC,$17,$F7,$56,$C7,$FA,$6E,$FF,$CE,$4E,$DF,$F3
	.dc.b	$A1,$FD,$AC,$BD,$BC,$9C,$5F,$F8,$82,$3F,$45,$87,$F9,$4A,$7C,$FF
	.dc.b	$11,$F3,$63,$15,$E8,$8F,$C2,$62,$F8,$1F,$D7,$CF,$F8,$FF,$0F,$3C
	.dc.b	$7E,$BE,$3F,$CB,$EE,$B9,$BF,$D1,$FF,$71,$F3,$27,$FC,$96,$FC,$98
	.dc.b	$CE,$DF,$02,$31,$9F,$FA,$AB,$F0,$FF,$EE,$6F,$F3,$1C,$91,$8C,$3F
	.dc.b	$9C,$C6,$3D,$BF,$3C,$E6,$7F,$AA,$CB,$F1,$7F,$EB,$97,$FE,$9B,$6C
	.dc.b	$FF,$97,$9F,$D0,$C6,$63,$E9,$FC,$67,$CA,$CF,$FD,$52,$FB,$67,$FD
	.dc.b	$7E,$E5,$BA,$FD,$AC,$5E,$E7,$90,$93,$EB,$8C,$BE,$DF,$31,$F8,$77
	.dc.b	$E0,$9B,$1F,$EB,$15,$F7,$B9,$9F,$C3,$FC,$FB,$F0,$9A,$E1,$B7,$FE
	.dc.b	$B9,$F6,$7E,$96,$DF,$11,$AF,$FB,$61,$AF,$FA,$61,$BC,$0F,$BA,$FA
	.dc.b	$B3,$FE,$9C,$6C,$67,$D5,$FC,$CC,$63,$E5,$3F,$CA,$75,$FE,$8A,$5F
	.dc.b	$4F,$F9,$E7,$3C,$7F,$ED,$47,$39,$79,$FC,$0F,$C8,$E7,$FB,$6C,$3D
	.dc.b	$11,$83,$F2,$BF,$2F,$FB,$AC,$3F,$DB,$AD,$BF,$F7,$9B,$F9,$18,$CC
	.dc.b	$FF,$5C,$A7,$CD,$BE,$53,$1E,$8B,$E8,$B5,$39,$CB,$FC,$53,$39,$7F
	.dc.b	$C5,$70,$FC,$33,$F9,$56,$1D,$1B,$19,$9F,$F0,$54,$E5,$FF,$06,$7A
	.dc.b	$CF,$C4,$FF,$BB,$69,$FD,$ED,$B3,$FE,$2E,$BE,$57,$FD,$EA,$72,$FF
	.dc.b	$B3,$BE,$87,$F2,$8D,$EC,$FE,$17,$1F,$B5,$E5,$3A,$5F,$85,$7E,$3A
	.dc.b	$BF,$D7,$ED,$C9,$97,$F9,$0E,$B8,$E1,$FC,$5A,$B5,$39,$FB,$B7,$DA
	.dc.b	$FF,$1A,$7F,$97,$E3,$FE,$CA,$73,$3F,$E4,$DC,$7A,$CF,$FB,$9A,$77
	.dc.b	$18,$C5,$18,$AF,$17,$E2,$7F,$E2,$C7,$91,$E1,$F9,$2F,$E5,$31,$9C
	.dc.b	$3F,$C8,$1C,$2F,$A6,$FB,$BF,$05,$F7,$9B,$87,$DC,$8F,$B3,$26,$92
	.dc.b	$3E,$BF,$07,$F5,$7D,$3E,$B7,$C4,$72,$C7,$EA,$3F,$E2,$8A,$7E,$3A
	.dc.b	$F9,$8C,$36,$3F,$C7,$7E,$03,$EC,$33,$5F,$D0,$BF,$1E,$7D,$45,$33
	.dc.b	$7F,$03,$27,$1A,$B7,$39,$87,$C0,$F9,$E3,$ED,$3F,$EC,$A7,$4E,$6B
	.dc.b	$9B,$97,$FE,$00,$B7,$5F,$A9,$F1,$7F,$89,$8C,$57,$E1,$63,$F2,$35
	.dc.b	$F8,$2F,$91,$FB,$BF,$18,$F2,$3F,$D0,$FD,$3F,$D2,$A3,$15,$FE,$8F
	.dc.b	$8F,$EA,$F1,$88,$FC,$5F,$5F,$2A,$E2,$FE,$23,$1F,$F3,$CD,$FF,$54
	.dc.b	$6C,$66,$B7,$1F,$CF,$9E,$D1,$CD,$FF,$4F,$73,$3E,$F7,$11,$E0,$AF
	.dc.b	$F5,$16,$3E,$3A,$D4,$63,$1F,$D0,$7B,$DC,$F7,$FC,$BB,$D7,$E7,$5F
	.dc.b	$5B,$FE,$16,$C7,$DF,$5A,$C3,$96,$5F,$A1,$C3,$97,$67,$F8,$36,$F7
	.dc.b	$DD,$FD,$52,$BF,$D0,$F9,$FE,$8B,$E4,$7D,$5E,$1F,$FA,$03,$8A,$FC
	.dc.b	$87,$CA,$AF,$F5,$D6,$EF,$BC,$FE,$13,$9F,$D0,$C3,$86,$DE,$26,$E7
	.dc.b	$D8,$FF,$40,$7D,$3B,$0E,$B9,$F8,$23,$17,$EA,$AF,$F3,$B3,$F2,$BE
	.dc.b	$3F,$F8,$4B,$9E,$67,$CC,$66,$21,$F5,$6B,$8B,$6B,$8C,$7F,$06,$3F
	.dc.b	$DA,$8E,$23,$FD,$72,$EB,$3F,$AD,$F0,$3D,$2B,$D4,$39,$BF,$C5,$0F
	.dc.b	$B7,$47,$E1,$FF,$E0,$2B,$FC,$9C,$62,$7E,$D7,$DD,$7F,$54,$78,$56
	.dc.b	$F8,$7E,$6B,$63,$FD,$4D,$EA,$7F,$76,$D3,$EE,$E1,$CD,$F9,$D1,$99
	.dc.b	$F1,$7D,$CF,$F0,$36,$F8,$5F,$89,$8F,$F5,$5E,$7E,$47,$4D,$8F,$C8
	.dc.b	$7C,$77,$87,$15,$6B,$3F,$E7,$13,$E5,$C3,$FD,$55,$F7,$3F,$E7,$E7
	.dc.b	$E2,$F0,$23,$19,$7E,$C6,$39,$4F,$D0,$9C,$DF,$FC,$00,$9E,$1D,$7F
	.dc.b	$81,$6D,$74,$CF,$F0,$57,$C2,$39,$56,$E3,$11,$FE,$BD,$9D,$CE,$49
	.dc.b	$1C,$F5,$B8,$F9,$6F,$8E,$1E,$08,$F4,$61,$CA,$FD,$B5,$A3,$71,$99
	.dc.b	$E5,$87,$4C,$7D,$93,$97,$DD,$5C,$67,$DD,$39,$BF,$59,$DA,$33,$8F
	.dc.b	$38,$CE,$1D,$9F,$51,$B2,$39,$3D,$31,$C6,$3E,$BC,$DC,$7E,$9C,$60
	.dc.b	$E3,$8F,$9A,$7C,$31,$83,$C3,$19,$7C,$37,$DE,$30,$D8,$FD,$6E,$1E
	.dc.b	$D7,$B4,$65,$FA,$D8,$78,$F1,$F8,$91,$D2,$F9,$91,$F0,$9D,$27,$86
	.dc.b	$1E,$BC,$3B,$63,$F6,$6F,$85,$F6,$3D,$47,$89,$CC,$7C,$23,$84,$76
	.dc.b	$34,$77,$38,$BA,$72,$75,$C7,$D9,$19,$25,$79,$31,$FD,$35,$7F,$0E
	.dc.b	$6B,$F2,$CC,$61,$FE,$3C,$A7,$DC,$FF,$27,$91,$FC,$59,$C3,$75,$E1
	.dc.b	$18,$8F,$D6,$3C,$71,$FD,$93,$72,$CD,$7E,$BF,$E3,$BA,$DF,$EE,$AB
	.dc.b	$4F,$F2,$66,$AC,$3C,$B5,$C6,$3F,$71,$3A,$9C,$C7,$39,$D3,$49,$87
	.dc.b	$92,$32,$D8,$CE,$1E,$67,$F8,$6E,$1E,$69,$D1,$27,$4F,$F0,$F8,$D4
	.dc.b	$60,$93,$EB,$AE,$F1,$88,$F6,$61,$E9,$3D,$55,$CA,$38,$1E,$42,$DF
	.dc.b	$77,$E0,$DC,$74,$7C,$79,$E4,$F6,$8D,$4F,$A6,$3F,$A9,$CE,$B0,$F9
	.dc.b	$33,$99,$E0,$7C,$23,$8E,$1E,$99,$DB,$FC,$AE,$B9,$61,$EB,$C3,$77
	.dc.b	$D8,$A7,$4C,$3B,$60,$BE,$35,$A8,$EB,$5C,$4F,$6B,$87,$96,$1E,$17
	.dc.b	$86,$1E,$17,$D0,$70,$71,$87,$C9,$8C,$5F,$7C,$3C,$31,$83,$8C,$F9
	.dc.b	$AB,$8D,$F7,$9D,$4F,$AA,$B8,$BB,$BF,$BF,$3F,$E1,$1F,$1D,$F2,$DF
	.dc.b	$77,$C0,$7E,$02,$FB,$4C,$36,$FF,$B8,$DF,$06,$DF,$E1,$70,$FF,$38
	.dc.b	$36,$75,$3F,$E0,$57,$FC,$F0,$FF,$7A,$F8,$2F,$F6,$CB,$FE,$27,$E0
	.dc.b	$7D,$47,$E3,$FB,$FF,$B5,$3D,$6B,$75,$FE,$0F,$6E,$BF,$E0,$E6,$EF
	.dc.b	$FE,$36,$6C,$FF,$8B,$DF,$F0,$C3,$FD,$B7,$F0,$FF,$C9,$FD,$C7,$F0
	.dc.b	$AB,$FF,$33,$36,$FF,$BA,$4F,$E4,$CF,$E0,$BC,$DF,$E2,$66,$DE,$8F
	.dc.b	$F1,$73,$87,$FC,$5A,$F1,$7C,$07,$F7,$DF,$6B,$DD,$E2,$7F,$24,$C7
	.dc.b	$FC,$95,$E0,$BE,$BF,$E1,$36,$FF,$C9,$95,$BC,$BF,$91,$EC,$8C,$4A
	.dc.b	$31,$D3,$D8,$FA,$DF,$57,$FE,$75,$9F,$F3,$3F,$3F,$FB,$11,$FC,$17
	.dc.b	$92,$F8,$CA,$7E,$17,$CA,$FE,$4B,$3F,$E5,$BA,$E0,$F2,$8C,$ED,$FE
	.dc.b	$6E,$8F,$C2,$7B,$1E,$47,$F9,$47,$CE,$FD,$0F,$85,$91,$FE,$1F,$8F
	.dc.b	$E0,$07,$C9,$76,$7F,$96,$5C,$78,$06,$FD,$0F,$09,$E4,$7F,$8A,$5F
	.dc.b	$CE,$7A,$7E,$F8,$7A,$9F,$EB,$0F,$3B,$F9,$FB,$63,$FD,$73,$E9,$63
	.dc.b	$AB,$FD,$7B,$85,$9F,$F5,$DC,$DD,$FF,$D8,$6D,$C5,$59,$F5,$5F,$63
	.dc.b	$49,$87,$FB,$3F,$17,$E3,$FE,$B8,$C3,$5F,$3C,$7F,$60,$37,$AF,$8D
	.dc.b	$B1,$FE,$D9,$E7,$7E,$40,$F8,$5F,$44,$3B,$1F,$21,$86,$DF,$16,$BE
	.dc.b	$03,$FB,$7F,$E1,$2F,$AE,$F8,$50,$8C,$3F,$84,$37,$15,$A5,$8D,$BB
	.dc.b	$F0,$DF,$80,$F4,$9C,$F2,$FF,$A5,$5B,$AF,$F7,$4A,$DD,$FF,$BC,$FE
	.dc.b	$EB,$F8,$1D,$AB,$FC,$0B,$CC,$EE,$F8,$57,$FE,$F9,$6B,$ED,$85,$F2
	.dc.b	$AE,$2F,$F0,$69,$D3,$95,$3C,$3E,$3B,$F0,$53,$BC,$3C,$55,$AF,$87
	.dc.b	$7D,$D7,$E1,$DB,$1F,$E1,$A9,$DE,$09,$CC,$F9,$7E,$25,$E1,$FE,$3D
	.dc.b	$B1,$C7,$3F,$7C,$62,$38,$4E,$FE,$63,$EF,$7C,$24,$9F,$BC,$FD,$AF
	.dc.b	$8B,$F4,$61,$FE,$20,$7E,$04,$FE,$F9,$AE,$7D,$DF,$B8,$A3,$2F,$F8
	.dc.b	$D1,$F8,$CB,$85,$F8,$BF,$E2,$1B,$1E,$93,$87,$0F,$EA,$FE,$E9,$F0
	.dc.b	$14,$93,$BE,$04,$BE,$EF,$F6,$5F,$25,$F8,$4A,$FE,$2B,$48,$C4,$7F
	.dc.b	$8F,$23,$1D,$1F,$95,$6C,$7E,$65,$7C,$9F,$E1,$B9,$7E,$AB,$FC,$8E
	.dc.b	$7C,$4D,$7F,$14,$EF,$83,$F2,$3F,$76,$F9,$0F,$E0,$CD,$AF,$53,$5F
	.dc.b	$C9,$2F,$AC,$F5,$FB,$9E,$5D,$3F,$CB,$55,$F9,$C6,$BF,$92,$6E,$6F
	.dc.b	$88,$FE,$23,$3F,$1D,$FE,$41,$A7,$E3,$9B,$9F,$E9,$9D,$CF,$E8,$FE
	.dc.b	$61,$F2,$7F,$C7,$5B,$1F,$45,$AE,$FE,$AB,$F9,$17,$CD,$3E,$5B,$E6
	.dc.b	$36,$7E,$63,$5F,$AA,$D7,$EA,$7C,$DB,$EF,$3F,$96,$3F,$9B,$D3,$F4
	.dc.b	$AD,$D3,$F7,$AE,$EB,$F0,$9F,$76,$F5,$BF,$9A,$36,$33,$18,$EC,$F3
	.dc.b	$FC,$D7,$98,$93,$E6,$7F,$09,$F7,$2F,$B6,$FE,$7C,$BF,$0E,$A6,$23
	.dc.b	$FA,$17,$D1,$BE,$F6,$93,$0F,$CC,$43,$EA,$FF,$A5,$7C,$13,$E9,$9F
	.dc.b	$5D,$FC,$9E,$30,$D8,$CD,$FF,$2D,$9C,$BF,$97,$E1,$7B,$BF,$D4,$DF
	.dc.b	$14,$7E,$9D,$AF,$F1,$F7,$6F,$EA,$F8,$DF,$33,$FD,$6D,$B1,$98,$FE
	.dc.b	$BD,$F5,$1F,$C7,$23,$FA,$26,$1E,$9C,$FC,$C4,$8F,$36,$5E,$F7,$D1
	.dc.b	$F7,$7F,$9D,$7F,$80,$52,$1F,$47,$F1,$1B,$3A,$78,$3E,$0F,$AB,$FD
	.dc.b	$F6,$7F,$CB,$AD,$18,$D3,$FD,$8B,$18,$F1,$FE,$7F,$D6,$7F,$61,$C3
	.dc.b	$C3,$F5,$BF,$11,$FF,$5F,$D2,$31,$1C,$9A,$A3,$5C,$7F,$D7,$77,$FE
	.dc.b	$C5,$BF,$DC,$FA,$EF,$D6,$DE,$DB,$3A,$8F,$83,$8F,$F5,$C7,$A3,$70
	.dc.b	$37,$1F,$AF,$3F,$41,$AF,$A1,$AF,$9F,$8B,$F1,$51,$88,$F4,$7D,$27
	.dc.b	$DC,$7E,$CB,$77,$F9,$4D,$DB,$E4,$F3,$3A,$7F,$2E,$DB,$FB,$2D,$7F
	.dc.b	$06,$DB,$FA,$8F,$EF,$DF,$AD,$F8,$3B,$EE,$3F,$42,$DA,$FC,$1B,$5F
	.dc.b	$C6,$FD,$CB,$E2,$FF,$4E,$D8,$FC,$4B,$5D,$C6,$B3,$FF,$0F,$FC,$67
	.dc.b	$F6,$AC,$7F,$B9,$FC,$27,$FB,$DF,$DE,$C9,$AF,$AD,$98,$FF,$8A,$6D
	.dc.b	$18,$8F,$F1,$11,$F9,$6F,$5F,$FC,$43,$D9,$FE,$92,$8F,$F3,$1B,$51
	.dc.b	$1C,$36,$3F,$C4,$98,$3C,$BF,$EB,$F3,$FE,$39,$6B,$FE,$31,$8D,$7A
	.dc.b	$FF,$E3,$A9,$FE,$C1,$3B,$F5,$B0,$47,$FC,$85,$F2,$5F,$C2,$1F,$4F
	.dc.b	$D4,$BE,$A3,$FC,$C3,$1F,$E3,$78,$C6,$FF,$E3,$76,$C6,$67,$6F,$FA
	.dc.b	$02,$3F,$97,$7D,$BE,$FC,$B3,$49,$3F,$E0,$6F,$93,$FF,$03,$A6,$91
	.dc.b	$FF,$45,$F7,$B9,$BF,$CB,$D1,$FE,$51,$8C,$3F,$E2,$CC,$BF,$89,$7E
	.dc.b	$04,$FB,$DC,$76,$FD,$0B,$DB,$F0,$BD,$1C,$E1,$FE,$A3,$DF,$FA,$74
	.dc.b	$FF,$84,$6B,$73,$F3,$E7,$2F,$E0,$3F,$15,$7D,$27,$F2,$18,$FB,$A2
	.dc.b	$33,$7F,$E5,$F6,$C7,$F9,$8D,$AF,$FA,$A3,$E1,$1F,$41,$E4,$84,$62
	.dc.b	$BF,$C4,$91,$D6,$F5,$39,$8F,$C9,$FD,$D7,$FB,$E2,$B6,$D8,$FC,$BC
	.dc.b	$EF,$6F,$E7,$13,$97,$F9,$47,$DD,$3E,$67,$E2,$F9,$28,$C6,$3F,$17
	.dc.b	$EF,$BF,$AC,$3F,$CE,$23,$39,$7F,$42,$7F,$D3,$F1,$9F,$9A,$E7,$1A
	.dc.b	$C7,$BC,$76,$FB,$AF,$EC,$D5,$FD,$09,$B7,$F7,$F8,$DF,$4B,$BF,$43
	.dc.b	$87,$9B,$30,$E1,$16,$FF,$92,$73,$7D,$57,$F9,$B7,$9B,$F6,$9F,$E4
	.dc.b	$D1,$9D,$FF,$9F,$37,$CE,$FE,$D3,$87,$F3,$4C,$7E,$1E,$1F,$82,$DB
	.dc.b	$FA,$75,$6B,$1F,$CA,$7E,$1A,$F8,$0F,$C6,$F5,$3E,$ED,$F5,$78,$AD
	.dc.b	$69,$F8,$F9,$CB,$7A,$7D,$17,$C5,$B3,$F8,$8C,$7F,$38,$D7,$DB,$EA
	.dc.b	$79,$8F,$B3,$EA,$7F,$6A,$C3,$FA,$EC,$7D,$CF,$BB,$FF,$12,$47,$F7
	.dc.b	$07,$FD,$F8,$DF,$13,$5F,$47,$F8,$2F,$6B,$F9,$0C,$FF,$7A,$F3,$FE
	.dc.b	$44,$63,$2F,$F8,$03,$CD,$FE,$19,$9F,$A1,$F0,$ED,$FF,$CC,$18,$7F
	.dc.b	$9D,$DB,$7F,$EA,$F7,$FC,$BD,$19,$F8,$CF,$DA,$D7,$FA,$E7,$D7,$B3
	.dc.b	$FD,$04,$FF,$0D,$38,$F5,$B7,$F8,$BB,$D7,$F0,$B6,$BF,$CF,$4D,$CB
	.dc.b	$FD,$06,$D7,$FD,$05,$F0,$1F,$E2,$09,$FF,$1A,$FB,$FE,$2F,$C8,$7F
	.dc.b	$D9,$51,$83,$FD,$21,$F7,$BF,$2A,$FF,$D9,$0F,$CA,$6F,$1F,$AA,$31
	.dc.b	$DB,$FE,$BC,$F8,$8F,$F8,$03,$7F,$E6,$2F,$A3,$F8,$F1,$8C,$3F,$CC
	.dc.b	$71,$F2,$BE,$F7,$FD,$DD,$5F,$E9,$F6,$CF,$FB,$5D,$FF,$7D,$46,$7E
	.dc.b	$7B,$E4,$FC,$5F,$F6,$D8,$FF,$53,$3F,$26,$BF,$DB,$CE,$5B,$C3,$F4
	.dc.b	$3E,$C7,$FD,$9B,$5F,$EA,$16,$CF,$FA,$A9,$AF,$FA,$A5,$AF,$FB,$BB
	.dc.b	$DB,$FD,$BC,$FF,$A1,$7C,$9F,$B7,$1F,$CD,$63,$07,$FA,$E7,$EA,$5F
	.dc.b	$81,$7E,$01,$7D,$D5,$8C,$3F,$84,$F9,$3F,$EB,$15,$FE,$B5,$6E,$3F
	.dc.b	$EB,$E6,$BF,$EB,$D6,$BF,$DC,$BA,$9F,$6D,$FE,$D2,$AF,$F6,$53,$65
	.dc.b	$C7,$FE,$F2,$87,$D6,$FE,$51,$D4,$7F,$B0,$BE,$D3,$F2,$FF,$34,$F9
	.dc.b	$EB,$18,$7F,$37,$AB,$EA,$3F,$4A,$DC,$FF,$D9,$CD,$CF,$FD,$AE,$D7
	.dc.b	$FD,$AD,$F7,$BE,$EC,$95,$EB,$8C,$37,$AF,$F5,$31,$8A,$FF,$08,$FD
	.dc.b	$47,$ED,$EF,$FC,$F0,$BE,$FB,$FA,$83,$9C,$3F,$AD,$B9,$7F,$A8,$FD
	.dc.b	$17,$F9,$05,$FE,$BF,$7F,$D8,$23,$0E,$03,$87,$6D,$7F,$B3,$E8,$FB
	.dc.b	$1F,$EC,$0D,$AE,$FF,$27,$89,$FE,$07,$F4,$DF,$B5,$BF,$AB,$5F,$6E
	.dc.b	$31,$9F,$C4,$97,$D0,$7F,$AA,$BE,$DF,$8F,$8B,$CB,$C9,$98,$FF,$8D
	.dc.b	$30,$E1,$D0,$CB,$9C,$F5,$AF,$E2,$7F,$65,$FC,$E3,$3F,$08,$B4,$67
	.dc.b	$0F,$16,$6F,$BC,$FE,$B7,$1D,$A7,$58,$FE,$4D,$FD,$CE,$1D,$30,$E7
	.dc.b	$C3,$E0,$F6,$7C,$2F,$C6,$DA,$61,$FE,$04,$E1,$FF,$06,$E5,$FE,$B3
	.dc.b	$53,$F3,$3E,$CE,$56,$E2,$E5,$FE,$5A,$A5,$7D,$2F,$3B,$38,$73,$7F
	.dc.b	$C4,$FB,$3F,$04,$FF,$19,$CF,$F9,$AB,$C0,$FD,$FE,$3F,$12,$B4,$FF
	.dc.b	$8B,$BE,$53,$F7,$7E,$25,$AD,$DF,$D7,$D5,$F7,$CF,$1B,$57,$FE,$36
	.dc.b	$9C,$C7,$8B,$F1,$6F,$E4,$D7,$F6,$3E,$07,$FA,$F6,$7E,$19,$CD,$FE
	.dc.b	$FE,$73,$5B,$F9,$8F,$E9,$B1,$FE,$FB,$9F,$C1,$1B,$8C,$FC,$E7,$E9
	.dc.b	$3E,$1B,$F1,$9F,$26,$F2,$FF,$4B,$E6,$E1,$FD,$3C,$DD,$3F,$5F,$35
	.dc.b	$74,$7F,$31,$CB,$F5,$2D,$8E,$71,$FE,$5E,$78,$B5,$E3,$3C,$E7,$FA
	.dc.b	$D6,$1E,$39,$D9,$FE,$6B,$C3,$F7,$3E,$BF,$8F,$93,$F2,$5E,$F5,$EE
	.dc.b	$FF,$1C,$DE,$CF,$D3,$F4,$FF,$9B,$B1,$E9,$7C,$34,$FF,$4C,$C6,$2F
	.dc.b	$8B,$E1,$ED,$FD,$0F,$B7,$D7,$2F,$85,$18,$BE,$3A,$7F,$A8,$B1,$7D
	.dc.b	$FB,$E9,$BE,$47,$5B,$E4,$FF,$AB,$69,$E2,$AD,$63,$CB,$7F,$F3,$C7
	.dc.b	$17,$0B,$DE,$7F,$EA,$EC,$79,$4F,$C1,$C3,$8D,$FB,$E7,$38,$73,$8F
	.dc.b	$AB,$E2,$FD,$7F,$9C,$FE,$FF,$1F,$14,$CC,$EE,$33,$9F,$D1,$9C,$9B
	.dc.b	$8C,$E5,$FC,$AA,$FA,$5C,$39,$1F,$6D,$C5,$7D,$3D,$5F,$7D,$F4,$EF
	.dc.b	$E4,$36,$7F,$8E,$FD,$FB,$E7,$BE,$F4,$E6,$F8,$D6,$0F,$C1,$56,$F1
	.dc.b	$F9,$53,$A6,$FD,$FF,$EC,$5B,$DF,$F8,$98,$FF,$45,$6B,$F8,$47,$58
	.dc.b	$79,$DB,$19,$9E,$1F,$7D,$C5,$F9,$9F,$30,$E7,$F9,$4D,$C3,$E4,$4E
	.dc.b	$A7,$24,$8C,$E3,$70,$FA,$1C,$1F,$EA,$8A,$C6,$FF,$EB,$28,$F1,$B7
	.dc.b	$1E,$D5,$C1,$B1,$9A,$FB,$98,$F8,$71,$FB,$55,$A7,$FD,$F3,$87,$77
	.dc.b	$C6,$6A,$BC,$2E,$A7,$82,$9F,$1B,$99,$DB,$99,$DC,$EA,$31,$3A,$8F
	.dc.b	$06,$1F,$02,$FC,$CF,$AE,$75,$18,$C3,$D9,$7E,$53,$CF,$5B,$70,$D3
	.dc.b	$A6,$3D,$F1,$F1,$9E,$4B,$E9,$5E,$93,$8D,$FC,$F8,$CC,$7A,$0C,$9F
	.dc.b	$21,$CE,$3E,$A9,$F2,$E1,$E0,$AD,$9C,$5F,$24,$75,$C3,$C9,$8F,$B9
	.dc.b	$B5,$C2,$FE,$41,$D5,$F3,$5F,$27,$C5,$3C,$23,$11,$F5,$70,$F2,$E3
	.dc.b	$F0,$30,$FC,$1D,$EF,$1F,$39,$E3,$BF,$05,$6D,$F9,$B1,$A7,$C2,$D7
	.dc.b	$B4,$FA,$1F,$F1,$4E,$3F,$8B,$6D,$7E,$8D,$B7,$FE,$2B,$7F,$15,$F1
	.dc.b	$BF,$E1,$C7,$FC,$5D,$5F,$EC,$CD,$74,$D8,$CC,$ED,$EE,$F1,$CF,$F5
	.dc.b	$2D,$CB,$F5,$0D,$9F,$CF,$FC,$61,$FE,$41,$19,$7C,$78,$60,$FE,$49
	.dc.b	$3B,$8F,$E4,$53,$BC,$3C,$98,$78,$2F,$75,$C4,$DB,$63,$F7,$B7,$CD
	.dc.b	$E1,$3A,$23,$A7,$86,$1E,$AA,$B1,$EA,$29,$E8,$38,$C7,$F0,$7A,$EC
	.dc.b	$F1,$8C,$9F,$C0,$30,$F5,$4E,$A3,$15,$D3,$0F,$6D,$F3,$9C,$5F,$C1
	.dc.b	$38,$B8,$AE,$17,$CE,$3C,$98,$7B,$89,$7C,$E3,$C6,$69,$F3,$47,$C1
	.dc.b	$8C,$E1,$E2,$7C,$A7,$F5,$3A,$ED,$19,$8E,$46,$A7,$91,$E0,$71,$5D
	.dc.b	$8F,$14,$66,$38,$92,$55,$DB,$0F,$81,$87,$BA,$BE,$01,$E6,$9D,$46
	.dc.b	$2F,$C3,$87,$7A,$F0,$90,$FE,$55,$5C,$1F,$7D,$FC,$C6,$DF,$76,$9D
	.dc.b	$CF,$EB,$FE,$23,$6F,$B5,$28,$C4,$EA,$73,$1F,$E5,$38,$C1,$0F,$EE
	.dc.b	$CF,$06,$ED,$D3,$EF,$7F,$BA,$47,$F6,$E6,$BF,$DF,$5B,$9F,$F8,$B9
	.dc.b	$FF,$64,$46,$F8,$F9,$11,$E5,$E5,$7D,$AF,$BD,$91,$FE,$13,$7E,$EF
	.dc.b	$A3,$23,$04,$3F,$C2,$8E,$5F,$F0,$B4,$72,$72,$7F,$8B,$7A,$3F,$C6
	.dc.b	$5C,$5F,$9F,$6C,$67,$AF,$F8,$B6,$7F,$E4,$6F,$B4,$7C,$CF,$CC,$6D
	.dc.b	$6F,$0E,$57,$C3,$93,$FC,$85,$5A,$AF,$F1,$07,$85,$97,$FD,$59,$E8
	.dc.b	$FF,$B3,$B2,$FF,$27,$36,$FF,$C9,$55,$BE,$4F,$F4,$CE,$5F,$72,$30
	.dc.b	$67,$EA,$5B,$7E,$8B,$E6,$3F,$47,$F7,$73,$4F,$F1,$E4,$65,$FF,$38
	.dc.b	$75,$23,$F8,$E7,$67,$F3,$86,$CF,$F8,$3D,$AF,$F8,$1B,$87,$93,$FE
	.dc.b	$43,$8F,$D8,$C7,$F0,$9F,$11,$E6,$75,$AF,$F4,$89,$FE,$CB,$7E,$27
	.dc.b	$9B,$FE,$9D,$AF,$E5,$8F,$FA,$7E,$EE,$1F,$E6,$BF,$C2,$BF,$CA,$53
	.dc.b	$9C,$39,$E3,$DA,$F7,$1F,$EE,$4A,$DD,$FF,$99,$FE,$F1,$EF,$EA,$1F
	.dc.b	$7A,$DB,$0D,$C3,$CF,$D2,$C3,$8C,$7A,$B9,$3F,$CF,$BD,$BE,$0A,$D6
	.dc.b	$56,$F7,$7F,$EC,$7F,$21,$7E,$68,$7D,$BB,$EB,$31,$F7,$8F,$8B,$A6
	.dc.b	$E7,$55,$F2,$8E,$35,$92,$39,$87,$B1,$C4,$91,$97,$A7,$C2,$7C,$8E
	.dc.b	$47,$DD,$FE,$DA,$BF,$DE,$CD,$E6,$72,$7F,$67,$AD,$F8,$9C,$69,$F4
	.dc.b	$94,$98,$7D,$68,$CB,$F2,$7B,$BD,$CF,$DA,$F8,$17,$DD,$78,$DB,$19
	.dc.b	$8C,$7D,$04,$8E,$0D,$7F,$DE,$9E,$43,$D3,$F9,$1F,$69,$27,$E2,$3F
	.dc.b	$47,$E1,$BF,$8B,$F0,$7B,$27,$FC,$05,$1F,$C6,$9F,$EF,$C1,$C3,$FE
	.dc.b	$01,$6B,$C3,$CC,$FF,$0D,$3F,$E6,$3F,$82,$FB,$DE,$2B,$EB,$69,$88
	.dc.b	$FA,$1F,$75,$FA,$5A,$63,$E3,$E4,$4E,$63,$0D,$E5,$FA,$4F,$EA,$9F
	.dc.b	$A9,$CD,$F8,$6B,$F4,$DA,$31,$1F,$E1,$98,$C7,$4F,$F8,$C5,$B1,$F0
	.dc.b	$02,$F8,$DF,$C2,$C7,$EC,$1F,$CE,$E3,$E7,$6E,$1F,$85,$77,$3E,$C8
	.dc.b	$FC,$47,$DD,$3E,$C3,$F2,$0D,$9F,$DD,$B5,$E0,$78,$9F,$94,$FF,$00
	.dc.b	$FC,$11,$F8,$07,$A5,$B1,$E9,$C9,$18,$8F,$86,$D7,$E1,$7D,$DF,$EF
	.dc.b	$DE,$F2,$8F,$9D,$1F,$09,$B1,$9F,$27,$F5,$1F,$2D,$F9,$AA,$E2,$DD
	.dc.b	$7E,$D3,$74,$FD,$42,$F8,$5F,$D8,$7B,$0F,$83,$FA,$4D,$C3,$D4,$D7
	.dc.b	$ED,$B5,$FB,$5F,$2E,$F5,$3F,$74,$BB,$A9,$88,$FD,$F7,$DD,$3E,$8E
	.dc.b	$92,$FF,$95,$C3,$ED,$3F,$82,$7C,$EB,$E6,$BF,$0C,$DA,$F9,$CD,$7F
	.dc.b	$1E,$D7,$F1,$DF,$74,$F8,$FF,$E1,$CB,$E3,$E9,$88,$FE,$25,$F4,$2F
	.dc.b	$95,$A4,$BF,$E7,$F1,$FD,$31,$FE,$83,$F4,$5F,$E1,$D7,$F0,$0D,$7F
	.dc.b	$42,$D7,$F4,$0D,$7F,$8D,$72,$F7,$7F,$A5,$BD,$5F,$C1,$B7,$9B,$F6
	.dc.b	$8E,$E3,$F0,$F0,$F9,$1F,$E5,$4D,$8F,$EA,$AD,$77,$F2,$FF,$D5,$BE
	.dc.b	$09,$F4,$AF,$AC,$FF,$51,$46,$1B,$19,$AF,$DA,$4E,$5B,$F7,$2F,$23
	.dc.b	$A2,$FB,$CA,$62,$3C,$1F,$5D,$CE,$BF,$5B,$68,$C6,$7F,$C8,$E3,$19
	.dc.b	$79,$FE,$AF,$F8,$75,$BE,$A5,$EE,$FE,$4E,$3F,$D5,$51,$8F,$03,$FD
	.dc.b	$78,$D8,$FE,$18,$D7,$F9,$6A,$F9,$AF,$8D,$97,$F3,$38,$EB,$DD,$FD
	.dc.b	$F1,$B5,$FE,$0A,$78,$71,$7C,$BF,$AE,$FF,$66,$5F,$A5,$BC,$9F,$EB
	.dc.b	$DD,$91,$8C,$11,$F0,$79,$FE,$29,$F4,$B8,$3E,$F4,$FF,$AF,$E7,$4F
	.dc.b	$D2,$8C,$BD,$7A,$5F,$35,$F7,$5B,$1F,$E2,$0A,$FD,$04,$E7,$B6,$4E
	.dc.b	$6B,$F5,$59,$FD,$16,$E7,$F3,$3E,$C9,$ED,$FE,$27,$ED,$7E,$3C,$FB
	.dc.b	$5B,$87,$B1,$B9,$9F,$9D,$1E,$36,$BF,$82,$6E,$3F,$81,$6D,$7D,$CF
	.dc.b	$C4,$7F,$D4,$35,$FE,$11,$EF,$FE,$D3,$3F,$EE,$36,$C7,$FB,$AD,$AB
	.dc.b	$F2,$2F,$EA,$4F,$F8,$65,$AF,$F3,$E3,$5E,$A7,$F8,$96,$33,$8F,$F7
	.dc.b	$59,$7E,$2D,$FE,$0A,$78,$35,$FE,$EE,$D7,$FD,$EC,$D7,$FD,$FC,$DF
	.dc.b	$8A,$7C,$B7,$F2,$99,$FF,$1F,$B5,$FE,$A3,$1A,$E8,$FC,$97,$D9,$FE
	.dc.b	$9B,$4F,$CF,$B6,$BF,$48,$DE,$93,$98,$7E,$47,$FD,$3F,$3F,$67,$DD
	.dc.b	$FF,$2F,$7F,$B8,$4A,$30,$FF,$73,$6B,$FE,$14,$6A,$22,$FA,$B7,$DD
	.dc.b	$F4,$8C,$36,$33,$3D,$5A,$F8,$BE,$6B,$FC,$36,$9A,$47,$FC,$0D,$E8
	.dc.b	$FE,$83,$4F,$4B,$63,$3D,$7F,$65,$F3,$08,$D3,$FE,$2E,$FB,$BF,$E0
	.dc.b	$55,$F0,$9B,$1F,$E1,$36,$BB,$F0,$7A,$3F,$07,$F1,$7F,$E6,$55,$F2
	.dc.b	$9B,$1F,$E1,$B6,$BF,$E3,$C6,$BE,$45,$E5,$F4,$84,$62,$7A,$4E,$63
	.dc.b	$13,$FE,$27,$FB,$C7,$D1,$71,$6D,$FD,$C6,$CE,$E7,$D3,$C6,$EB,$FF
	.dc.b	$36,$DF,$F2,$66,$D7,$F2,$66,$BF,$2A,$78,$4F,$2F,$BC,$FF,$05,$3F
	.dc.b	$CB,$9B,$3F,$CB,$9A,$F1,$6B,$FC,$37,$E8,$FC,$7E,$09,$A4,$AF,$E7
	.dc.b	$3A,$BE,$8B,$D2,$D8,$CC,$60,$F6,$F4,$7F,$A6,$E3,$FA,$0B,$70,$FE
	.dc.b	$82,$D7,$D8,$D7,$FA,$5F,$81,$FC,$DA,$7F,$A4,$B6,$3F,$A4,$B5,$F8
	.dc.b	$8D,$7F,$6B,$ED,$78,$89,$3E,$27,$CB,$E3,$C8,$F2,$B6,$7F,$AA,$FD
	.dc.b	$D7,$D8,$7E,$83,$5F,$EB,$ED,$7F,$B2,$37,$6F,$E9,$F3,$F7,$BC,$9F
	.dc.b	$F0,$E5,$FF,$62,$6D,$FF,$62,$6B,$F7,$79,$1F,$7B,$F8,$66,$D7,$F6
	.dc.b	$A6,$CF,$F6,$A6,$BF,$85,$FC,$63,$FC,$09,$3F,$DC,$1B,$1F,$DC,$1A
	.dc.b	$EF,$D1,$FC,$9F,$E2,$DF,$EB,$AA,$FE,$EE,$D9,$FE,$EE,$D7,$F2,$4D
	.dc.b	$7F,$C0,$37,$F8,$37,$C9,$C1,$E9,$25,$7F,$A3,$7E,$75,$F9,$23,$F0
	.dc.b	$3F,$DD,$C6,$1B,$19,$9E,$0D,$BF,$F5,$AD,$3F,$0E,$7E,$0B,$C9,$B7
	.dc.b	$DD,$B5,$C1,$AF,$F1,$4A,$7E,$2F,$F9,$18,$46,$3E,$7B,$CB,$39,$8C
	.dc.b	$5F,$AF,$C1,$FA,$37,$FC,$CE,$7F,$8F,$70,$FB,$27,$33,$D1,$B3,$FE
	.dc.b	$83,$E0,$F8,$3C,$DF,$52,$B9,$35,$F1,$B6,$7C,$FC,$9E,$42,$63,$F3
	.dc.b	$72,$FC,$5D,$F8,$F5,$E0,$4C,$7E,$7B,$CF,$EE,$5F,$2D,$F4,$9B,$87
	.dc.b	$D7,$6D,$FD,$06,$BF,$27,$F1,$67,$DF,$FD,$A6,$D7,$DF,$6C,$FD,$86
	.dc.b	$BF,$47,$F0,$97,$D0,$7E,$09,$B5,$F8,$76,$CF,$E0,$1A,$FD,$9F,$31
	.dc.b	$F5,$FF,$8A,$6D,$7E,$3D,$B3,$F8,$86,$BE,$9F,$9A,$F4,$12,$7D,$0F
	.dc.b	$E5,$3F,$1A,$FF,$5C,$47,$E6,$5B,$3F,$92,$6B,$F8,$B6,$BF,$D8,$38
	.dc.b	$FF,$02,$FF,$B0,$79,$7F,$A7,$46,$09,$F8,$26,$09,$7F,$E0,$48,$47
	.dc.b	$0F,$BE,$93,$96,$DF,$E9,$3E,$83,$FB,$3B,$F9,$36,$BF,$9C,$6F,$39
	.dc.b	$BE,$AB,$EF,$7F,$66,$DC,$3F,$70,$DB,$FD,$8B,$5F,$D5,$7E,$2E,$DB
	.dc.b	$F7,$6D,$AF,$E0,$0D,$9F,$DC,$B5,$FD,$97,$DD,$FF,$A8,$E7,$F8,$33
	.dc.b	$63,$F7,$ED,$77,$F1,$1F,$DD,$FF,$09,$FE,$F7,$1F,$E2,$36,$A6,$98
	.dc.b	$E5,$FE,$71,$1F,$90,$F4,$7F,$C0,$11,$82,$69,$82,$63,$FE,$36,$8F
	.dc.b	$81,$5F,$BF,$FA,$CF,$D1,$B7,$AF,$F8,$DC,$62,$3F,$85,$B7,$81,$F7
	.dc.b	$FF,$E3,$9C,$3F,$92,$36,$BF,$8E,$B5,$FE,$2C,$D7,$F8,$77,$E4,$AF
	.dc.b	$9E,$FE,$52,$DA,$FE,$5C,$D9,$FE,$4E,$D7,$F9,$07,$DF,$FE,$C9,$FE
	.dc.b	$30,$D7,$F9,$0B,$70,$39,$87,$3F,$F3,$F7,$FC,$7A,$63,$C9,$FC,$E9
	.dc.b	$30,$FC,$D7,$A5,$FA,$02,$38,$24,$FF,$93,$FE,$E3,$F5,$1F,$7E,$9F
	.dc.b	$F2,$F7,$BA,$9F,$D6,$5E,$6F,$81,$FE,$63,$C5,$86,$6F,$84,$FF,$3C
	.dc.b	$F8,$9F,$9D,$8F,$F4,$16,$C7,$F3,$96,$BB,$3B,$47,$E5,$BE,$3B,$A6
	.dc.b	$5F,$CF,$FA,$3F,$C5,$F8,$76,$DB,$FD,$6F,$86,$8F,$E9,$73,$AA,$FC
	.dc.b	$47,$C6,$66,$9C,$7F,$E6,$9C,$39,$29,$D3,$9A,$FC,$87,$C0,$79,$31
	.dc.b	$B9,$FE,$1F,$1E,$73,$99,$FC,$8B,$5F,$B3,$F5,$5E,$56,$D7,$CE,$BF
	.dc.b	$E3,$1E,$9E,$78,$17,$4E,$4E,$2B,$F2,$8D,$BF,$F4,$03,$63,$F0,$AE
	.dc.b	$74,$FA,$72,$FA,$9F,$F2,$FC,$62,$75,$8F,$FA,$22,$72,$FF,$6D,$FB
	.dc.b	$EF,$E1,$55,$F3,$FE,$FF,$F5,$0F,$D1,$8C,$CF,$F7,$39,$CB,$70,$FA
	.dc.b	$11,$87,$FB,$FC,$F0,$7F,$C0,$31,$9D,$5F,$7E,$FA,$1F,$B0,$DD,$3E
	.dc.b	$45,$FF,$82,$21,$F7,$DA,$A9,$5A,$BE,$71,$CA,$73,$7F,$9A,$3E,$27
	.dc.b	$DF,$BE,$4F,$AC,$61,$BB,$78,$63,$82,$9E,$11,$86,$ED,$F4,$27,$55
	.dc.b	$CB,$0F,$C1,$46,$2F,$84,$F6,$BE,$3B,$7D,$3B,$DC,$62,$BF,$DA,$57
	.dc.b	$C7,$0E,$D3,$99,$FB,$D5,$C3,$0E,$84,$8F,$4C,$65,$F0,$92,$BD,$11
	.dc.b	$97,$AC,$F1,$73,$9D,$C7,$F1,$18,$78,$71,$F2,$47,$A6,$75,$8F,$A2
	.dc.b	$B4,$6F,$1E,$D5,$CB,$1D,$D7,$AA,$EE,$1D,$23,$94,$FC,$B9,$CE,$1D
	.dc.b	$23,$13,$EB,$BE,$8D,$3E,$4E,$1E,$07,$99,$93,$C0,$5C,$3C,$0F,$E0
	.dc.b	$CF,$93,$87,$82,$FC,$B5,$CE,$75,$8F,$C8,$76,$73,$C7,$CC,$E2,$34
	.dc.b	$4C,$7C,$EF,$CE,$8E,$17,$DB,$0F,$04,$79,$70,$F0,$63,$E6,$23,$8C
	.dc.b	$7B,$E1,$CC,$ED,$5F,$01,$AF,$C0,$6D,$7E,$56,$3F,$A7,$B5,$F9,$66
	.dc.b	$27,$F4,$2E,$BE,$0F,$FC,$F6,$FF,$56,$3F,$88,$38,$75,$A6,$5C,$B6
	.dc.b	$BF,$DC,$8F,$EC,$27,$9D,$FE,$AD,$B9,$FF,$17,$F0,$D8,$6E,$E3,$FB
	.dc.b	$6A,$DB,$FA,$89,$DB,$99,$DD,$F5,$7F,$7F,$3A,$8C,$4E,$C8,$FE,$F6
	.dc.b	$35,$7F,$CA,$6F,$BD,$7B,$0D,$C7,$9D,$EE,$EE,$EC,$79,$4A,$71,$BF
	.dc.b	$09,$DD,$CE,$1F,$19,$EA,$71,$29,$C6,$BB,$3F,$C6,$30,$F3,$BF,$C2
	.dc.b	$B0,$F3,$D7,$3B,$EF,$7E,$7A,$E7,$39,$8E,$73,$E3,$9C,$D7,$23,$9D
	.dc.b	$F6,$8F,$11,$DF,$0F,$4C,$67,$0E,$93,$93,$BC,$66,$BA,$57,$09,$F8
	.dc.b	$31,$87,$D8,$62,$34,$F9,$1B,$7E,$C8,$EC,$D7,$FB,$25,$77,$9E,$51
	.dc.b	$D0,$E4,$79,$0C,$CF,$1B,$57,$19,$DC,$FC,$53,$57,$E7,$71,$87,$FB
	.dc.b	$67,$B3,$FA,$CD,$7F,$57,$6B,$FE,$2E,$C5,$18,$AF,$F1,$03,$FD,$96
	.dc.b	$78,$6D,$F7,$2B,$6D,$9E,$17,$DF,$4D,$7D,$B7,$FA,$27,$0F,$EC,$CD
	.dc.b	$7C,$0D,$C7,$FC,$5C,$FF,$B2,$23,$7E,$1F,$2F,$22,$46,$0F,$EC,$FE
	.dc.b	$57,$DD,$FB,$E9,$18,$21,$FE,$4B,$3F,$AD,$C7,$ED,$38,$BF,$87,$B6
	.dc.b	$3F,$C2,$8D,$76,$72,$C7,$FB,$74,$7E,$2F,$BF,$E1,$7A,$2F,$BD,$E1
	.dc.b	$C3,$FE,$D8,$6C,$67,$0F,$F1,$D3,$5F,$24,$FF,$B7,$BE,$F7,$9D,$49
	.dc.b	$FF,$06,$52,$7E,$85,$FF,$22,$C7,$FC,$A8,$DE,$0F,$F2,$F7,$3F,$D5
	.dc.b	$F3,$9B,$53,$CA,$E3,$2F,$F2,$63,$5F,$1B,$76,$FE,$67,$A7,$EA,$73
	.dc.b	$95,$FE,$16,$8F,$D6,$1F,$CB,$5F,$F2,$03,$79,$FD,$D1,$C8,$FE,$61
	.dc.b	$F5,$D9,$23,$83,$F9,$D6,$EF,$BD,$FF,$25,$C7,$F2,$A8,$C6,$BF,$CE
	.dc.b	$DB,$1C,$27,$6F,$F3,$68,$FF,$59,$F7,$6F,$CB,$FF,$4E,$CF,$FA,$D2
	.dc.b	$30,$73,$F0,$DC,$F4,$93,$E4,$8C,$BE,$6E,$47,$D9,$F8,$DB,$19,$8C
	.dc.b	$1F,$17,$EE,$9F,$75,$49,$3F,$D5,$63,$2D,$F9,$BE,$04,$70,$6F,$6E
	.dc.b	$84,$70,$7C,$AE,$77,$C1,$FC,$76,$C6,$63,$07,$CB,$E6,$FF,$4E,$C7
	.dc.b	$FB,$03,$81,$18,$8F,$A4,$D7,$E7,$35,$43,$AB,$FD,$1B,$1F,$71,$AF
	.dc.b	$56,$BF,$75,$AF,$D9,$F2,$FF,$12,$DE,$6F,$C0,$46,$23,$BB,$5F,$F4
	.dc.b	$86,$DF,$91,$23,$FD,$FD,$FC,$A7,$C0,$48,$FF,$01,$35,$FF,$05,$73
	.dc.b	$BE,$8B,$F3,$44,$8F,$F0,$43,$F9,$1F,$C3,$FF,$8D,$47,$E5,$5A,$FE
	.dc.b	$45,$AF,$C4,$6E,$6F,$81,$7E,$1D,$F9,$B6,$D7,$E8,$9A,$FD,$36,$BF
	.dc.b	$A3,$F1,$DF,$07,$FA,$36,$C7,$EA,$9A,$FD,$B6,$BF,$AB,$FC,$3B,$F0
	.dc.b	$6F,$DF,$6B,$9E,$2F,$D7,$46,$23,$F5,$96,$FB,$DF,$41,$23,$D0,$FE
	.dc.b	$D7,$ED,$DF,$71,$49,$3F,$BE,$8C,$BF,$C0,$3E,$DF,$F9,$B3,$FB,$F6
	.dc.b	$BF,$BB,$6B,$F8,$E6,$EA,$F8,$CF,$CB,$CF,$F0,$A6,$BF,$9B,$6B,$FC
	.dc.b	$2D,$AF,$F0,$7F,$8C,$7A,$BF,$C3,$5B,$1F,$C5,$1A,$FE,$95,$AF,$F1
	.dc.b	$5F,$8B,$E6,$FE,$89,$B6,$69,$8B,$78,$1F,$C6,$DB,$5E,$98,$C1,$FC
	.dc.b	$7F,$CB,$7D,$ED,$24,$FF,$28,$8C,$BF,$CA,$BE,$4D,$F2,$1F,$C9,$DB
	.dc.b	$19,$8C,$1F,$E6,$58,$7C,$9B,$ED,$30,$DA,$FF,$37,$4E,$A7,$27,$1F
	.dc.b	$83,$FB,$73,$FE,$60,$12,$8C,$C7,$F9,$92,$31,$C5,$C4,$F1,$70,$7C
	.dc.b	$C9,$FF,$37,$CF,$D4,$8C,$92,$3C,$78,$3E,$F3,$CE,$48,$CB,$ED,$F9
	.dc.b	$6F,$ED,$51,$EA,$6B,$E7,$B4,$62,$3C,$CD,$7C,$7F,$56,$F0,$BE,$43
	.dc.b	$63,$E2,$B5,$E8,$D7,$E0,$7C,$23,$EA,$FE,$83,$63,$E6,$B5,$F2,$35
	.dc.b	$E3,$F0,$BC,$08,$F0,$7E,$D7,$CD,$3F,$01,$A4,$9F,$B3,$19,$7B,$FC
	.dc.b	$27,$88,$8E,$0F,$BF,$F3,$9F,$CF,$75,$34,$9E,$6F,$E2,$7E,$75,$F2
	.dc.b	$74,$97,$F8,$88,$CB,$F8,$AF,$9E,$FF,$2D,$B7,$C6,$FC,$94,$62,$3E
	.dc.b	$A3,$5F,$C4,$7E,$26,$FB,$5F,$CC,$36,$3F,$2A,$D7,$EE,$35,$FC,$A7
	.dc.b	$D5,$7F,$19,$7F,$D8,$8D,$7F,$D9,$4D,$77,$E3,$7F,$87,$7F,$08,$FE
	.dc.b	$BD,$06,$93,$4D,$7D,$D4,$9F,$F6,$84,$A3,$11,$FE,$D7,$FA,$AC,$3F
	.dc.b	$ED,$56,$B9,$8F,$F1,$64,$62,$3F,$59,$F1,$6F,$9F,$FF,$18,$36,$3F
	.dc.b	$DB,$CD,$7F,$DC,$8D,$7F,$69,$F3,$BF,$27,$EA,$BF,$DC,$8D,$8F,$F7
	.dc.b	$53,$5D,$EC,$9D,$FD,$4B,$EA,$A9,$2B,$8C,$65,$FF,$22,$78,$FF,$E4
	.dc.b	$54,$D2,$3F,$EF,$9F,$35,$F8,$33,$E2,$7F,$DD,$F8,$FF,$9E,$67,$31
	.dc.b	$83,$FD,$DA,$BE,$5B,$85,$F9,$E7,$37,$8B,$E4,$D8,$E3,$F2,$D6,$BF
	.dc.b	$DF,$B8,$76,$79,$18,$8D,$2F,$AA,$78,$3F,$D5,$1B,$7F,$C8,$64,$6E
	.dc.b	$74,$6F,$C3,$E7,$5F,$C0,$EB,$D9,$5C,$23,$9B,$E2,$3C,$13,$C0,$E3
	.dc.b	$E4,$DA,$E3,$3E,$58,$F6,$78,$1F,$EB,$C6,$FC,$93,$C2,$B1,$86,$5F
	.dc.b	$8A,$9D,$DF,$4E,$31,$7C,$5D,$46,$70,$F8,$F5,$FC,$1B,$E4,$BC,$3E
	.dc.b	$1B,$A2,$C6,$19,$87,$79,$7C,$A6,$30,$E7,$18,$9F,$27,$8B,$F3,$1F
	.dc.b	$E3,$73,$FE,$06,$C5,$39,$8F,$E7,$8D,$7F,$A8,$3F,$E6,$49,$E7,$ED
	.dc.b	$7C,$68,$E1,$CB,$F8,$AB,$E7,$39,$F6,$E5,$FC,$5F,$9D,$FA,$B7,$BC
	.dc.b	$3F,$C6,$73,$E6,$F8,$CF,$24,$1A,$48,$FE,$67,$F4,$BF,$5B,$CF,$F7
	.dc.b	$69,$F2,$AD,$39,$DB,$ED,$FA,$3C,$F0,$FA,$CD,$BF,$A2,$F8,$9B,$8F
	.dc.b	$E0,$B9,$7F,$C1,$F5,$FD,$E9,$B3,$FD,$85,$AF,$F7,$56,$BF,$DC,$BE
	.dc.b	$A1,$F2,$3F,$8C,$7F,$D4,$35,$F8,$77,$2D,$F8,$6F,$C8,$FD,$47,$F9
	.dc.b	$7A,$3F,$18,$DB,$FC,$3B,$FC,$6E,$33,$F0,$7E,$7F,$A4,$FB,$99,$1A
	.dc.b	$49,$FF,$63,$FD,$43,$EB,$FF,$75,$F0,$48,$7D,$86,$1D,$57,$EB,$A7
	.dc.b	$53,$C3,$E0,$BF,$09,$A7,$F0,$09,$FF,$26,$74,$E7,$87,$4F,$6F,$83
	.dc.b	$FC,$65,$7F,$B3,$DA,$FF,$90,$4F,$F2,$CD,$7F,$90,$5B,$E3,$7D,$0E
	.dc.b	$9F,$E2,$F2,$8F,$F0,$D6,$7E,$57,$4F,$F9,$57,$DD,$F6,$E5,$F3,$B0
	.dc.b	$F3,$37,$0E,$6F,$D4,$8C,$F2,$F9,$E3,$C7,$97,$E1,$E7,$FC,$B1,$18
	.dc.b	$3C,$FD,$9C,$F1,$FF,$75,$3C,$F0,$F3,$B9,$AF,$F5,$FB,$99,$FF,$42
	.dc.b	$F9,$BF,$4B,$E1,$BF,$BB,$47,$CF,$5E,$6E,$67,$EB,$FD,$57,$FB,$C2
	.dc.b	$BF,$E0,$0B,$70,$FF,$69,$35,$FE,$56,$7F,$B3,$9C,$3F,$EF,$FF,$AA
	.dc.b	$FC,$53,$F6,$5F,$F4,$B4,$FD,$F7,$39,$7D,$07,$33,$F7,$FD,$1F,$F6
	.dc.b	$FD,$7E,$15,$B1,$9F,$83,$FC,$6C,$62,$FE,$CE,$1C,$AB,$CB,$C4,$8C
	.dc.b	$47,$F4,$28,$C7,$B1,$F8,$49,$CC,$FF,$BA,$3E,$A9,$F3,$9F,$9A,$7F
	.dc.b	$27,$87,$E1,$9C,$E7,$F9,$97,$2F,$C4,$F1,$9E,$67,$F5,$66,$C7,$E7
	.dc.b	$FE,$73,$78,$FF,$54,$9C,$BF,$D3,$79,$3F,$C7,$35,$FD,$66,$3F,$8C
	.dc.b	$7C,$07,$FA,$C2,$73,$1F,$D7,$5A,$FF,$9F,$DF,$DB,$4F,$F5,$26,$C7
	.dc.b	$59,$FE,$D3,$18,$CB,$F4,$3F,$24,$FA,$AF,$DC,$BE,$48,$CC,$63,$39
	.dc.b	$8F,$EE,$69,$F0,$98,$CB,$F7,$11,$89,$FD,$F7,$D5,$7D,$EC,$7F,$5E
	.dc.b	$BE,$83,$F8,$3B,$99,$FA,$CE,$5E,$31,$AF,$A0,$DB,$63,$FB,$7E,$DD
	.dc.b	$F1,$FE,$EC,$FF,$44,$8F,$A9,$97,$D3,$F6,$24,$70,$F5,$ED,$F0,$7F
	.dc.b	$C1,$B8,$3E,$16,$9F,$6E,$2D,$7F,$19,$F9,$CF,$D6,$DF,$F1,$76,$EB
	.dc.b	$FC,$31,$FD,$7C,$EF,$B3,$F9,$2C,$3E,$A2,$98,$8B,$3F,$C9,$7E,$A3
	.dc.b	$F9,$4B,$7C,$8F,$E5,$B1,$FE,$CC,$8C,$1B,$F8,$EC,$9A,$9B,$1F,$E1
	.dc.b	$4F,$C0,$3F,$79,$97,$F8,$DB,$34,$62,$7F,$35,$EE,$FF,$8B,$A3,$13
	.dc.b	$FE,$2A,$E8,$FC,$BB,$F8,$FE,$7F,$CA,$12,$FF,$31,$3F,$4F,$0E,$CD
	.dc.b	$8F,$E1,$2F,$E3,$5B,$C7,$F8,$2F,$C1,$FF,$C9,$3F,$06,$F9,$DF,$DC
	.dc.b	$7C,$57,$F8,$2E,$31,$9B,$C3,$75,$B0,$E3,$87,$0A,$A6,$F0,$E5,$19
	.dc.b	$CB,$FC,$8B,$8F,$4C,$79,$E1,$C1,$5F,$0C,$3C,$1C,$BF,$E5,$FC,$38
	.dc.b	$E3,$FC,$36,$31,$3F,$E6,$E8,$C7,$C2,$73,$9C,$E5,$DD,$B3,$C7,$E1
	.dc.b	$35,$F0,$BB,$7D,$38,$7F,$86,$7D,$6B,$C0,$73,$D8,$FA,$4C,$C6,$B9
	.dc.b	$A6,$DE,$6C,$3E,$D4,$E1,$F3,$1F,$76,$71,$AF,$CE,$8F,$05,$71,$7F
	.dc.b	$C5,$53,$F1,$7E,$A1,$E0,$D3,$1C,$76,$FC,$1C,$6A,$DC,$63,$EA,$3E
	.dc.b	$64,$E6,$FE,$84,$63,$EA,$1F,$51,$F4,$D7,$C4,$58,$C3,$30,$EF,$3A
	.dc.b	$6C,$E7,$4F,$78,$F9,$FF,$F3,$27,$99,$ED,$6C,$69,$E4,$DF,$89,$FE
	.dc.b	$8A,$D9,$E1,$18,$E0,$F9,$15,$C1,$B1,$9E,$7F,$ED,$33,$DA,$73,$5E
	.dc.b	$1C,$FE,$FC,$E6,$B7,$7C,$70,$FF,$4D,$CE,$5B,$5C,$27,$51,$ED,$6C
	.dc.b	$EE,$3E,$04,$EE,$FD,$24,$8F,$81,$19,$7D,$C4,$C3,$DB,$19,$7D,$31
	.dc.b	$9C,$3D,$B3,$97,$D3,$5D,$6B,$8E,$56,$73,$7C,$4F,$5C,$65,$F7,$E3
	.dc.b	$EA,$3D,$78,$75,$8C,$E1,$B7,$DB,$57,$1F,$80,$4A,$E1,$3F,$16,$B5
	.dc.b	$18,$9D,$DF,$6B,$E4,$F8,$6B,$EB,$46,$1E,$2D,$BF,$51,$E2,$2D,$74
	.dc.b	$9E,$31,$8A,$E8,$FD,$0A,$E5,$8F,$CC,$3E,$3C,$6E,$BB,$B4,$EB,$87
	.dc.b	$A6,$F9,$D7,$03,$96,$3F,$3A,$BF,$C7,$27,$C1,$3F,$DC,$D8,$FE,$81
	.dc.b	$CB,$5F,$CB,$FC,$27,$F2,$F7,$6D,$76,$D7,$7B,$FF,$86,$67,$FD,$96
	.dc.b	$FE,$B6,$79,$5F,$E9,$9B,$7F,$A4,$F1,$FF,$C9,$B3,$9A,$E0,$4C,$3F
	.dc.b	$69,$3A,$AE,$F1,$FB,$7B,$E6,$FE,$99,$AF,$EF,$E7,$83,$9A,$E9,$7D
	.dc.b	$E3,$A9,$8B,$EB,$19,$3F,$7B,$5C,$A7,$F8,$55,$F6,$C3,$DF,$1E,$3C
	.dc.b	$3F,$86,$56,$E7,$37,$CF,$0E,$A7,$C1,$BF,$01,$E5,$39,$1E,$42,$C7
	.dc.b	$17,$4F,$91,$E2,$79,$2E,$D7,$F8,$6E,$2E,$1E,$73,$D1,$87,$95,$C5
	.dc.b	$70,$8C,$3C,$0F,$5C,$D3,$C7,$87,$A5,$A7,$91,$F6,$BF,$C6,$27,$9B
	.dc.b	$A3,$CC,$7B,$70,$F6,$38,$6E,$1E,$98,$E0,$FC,$AA,$EB,$1F,$74,$F2
	.dc.b	$4F,$0A,$EE,$F2,$C3,$DF,$3A,$AE,$37,$D6,$B5,$87,$C1,$AD,$1E,$B3
	.dc.b	$D4,$60,$F3,$57,$C1,$9F,$F0,$57,$81,$FD,$62,$7F,$B0,$B5,$FE,$DF
	.dc.b	$5F,$E8,$6D,$BF,$26,$FF,$A2,$34,$B5,$FD,$7A,$B8,$D6,$0C,$CF,$0F
	.dc.b	$A6,$FF,$35,$4F,$F6,$E7,$FB,$9B,$78,$B5,$D1,$23,$4D,$7C,$73,$FE
	.dc.b	$29,$7F,$B1,$79,$7F,$97,$23,$FD,$97,$BE,$BF,$D9,$62,$E9,$FD,$82
	.dc.b	$7F,$3C,$49,$E0,$7F,$82,$7E,$95,$F8,$27,$C0,$F9,$DF,$F0,$C4,$FF
	.dc.b	$81,$97,$D2,$64,$D7,$74,$9D,$35,$F9,$72,$F7,$7E,$5F,$D2,$BD,$4A
	.dc.b	$4A,$FF,$09,$C3,$5D,$99,$8F,$AB,$1C,$5D,$C7,$EE,$71,$FF,$70,$F9
	.dc.b	$3F,$E6,$9E,$7F,$3D,$04,$66,$77,$D9,$FE,$00,$CA,$D7,$F7,$A7,$FD
	.dc.b	$7C,$FF,$84,$DB,$7E,$B8,$FE,$55,$1A,$F2,$9E,$BF,$29,$FF,$3C,$B6
	.dc.b	$33,$1E,$3C,$7F,$CF,$DC,$8F,$88,$F0,$F5,$3E,$A1,$E3,$E1,$B8,$71
	.dc.b	$D7,$FD,$D1,$18,$8F,$6C,$63,$DF,$E0,$48,$E0,$7F,$9D,$65,$1D,$7E
	.dc.b	$33,$C3,$C9,$E3,$AD,$E2,$8C,$47,$63,$FC,$FC,$E3,$46,$3F,$EF,$B8
	.dc.b	$7D,$37,$42,$57,$48,$CB,$7E,$9B,$E4,$74,$BE,$9E,$92,$7D,$22,$3F
	.dc.b	$CC,$58,$FF,$A9,$D7,$D1,$F2,$F0,$E5,$70,$9C,$B7,$EF,$7B,$D4,$8C
	.dc.b	$1F,$03,$A9,$F3,$7F,$51,$FF,$47,$C6,$7B,$F1,$39,$EE,$B3,$C6,$30
	.dc.b	$FC,$AF,$A2,$FC,$43,$F2,$78,$91,$F9,$68,$C3,$75,$F9,$71,$FE,$AB
	.dc.b	$8E,$31,$89,$E3,$CF,$F8,$89,$CF,$1F,$E7,$E7,$AF,$5E,$6B,$F2,$9D
	.dc.b	$9F,$43,$1F,$B4,$7F,$9B,$F9,$72,$6B,$92,$47,$F8,$1D,$FD,$09,$1F
	.dc.b	$F0,$B9,$FA,$DE,$27,$A5,$F6,$89,$1F,$8A,$F8,$17,$D7,$7E,$7B,$EB
	.dc.b	$BF,$C3,$FC,$1E,$BF,$AA,$FE,$3D,$7F,$15,$B1,$EF,$9D,$F0,$7B,$7E
	.dc.b	$29,$E4,$78,$B0,$FF,$68,$46,$1F,$2F,$12,$31,$1F,$E2,$E8,$C4,$F1
	.dc.b	$5E,$73,$BB,$F6,$6D,$AF,$D1,$B5,$FB,$4B,$E9,$3F,$73,$F5,$5F,$AC
	.dc.b	$9F,$D4,$BF,$B5,$D3,$EC,$37,$3F,$AC,$EE,$7F,$09,$9F,$F4,$87,$CC
	.dc.b	$D7,$F8,$D1,$EE,$F1,$3D,$11,$98,$FC,$07,$D5,$3E,$AF,$F8,$5B,$63
	.dc.b	$F7,$6D,$7F,$24,$FE,$AB,$0F,$C4,$7D,$5F,$F5,$D9,$FC,$B3,$FD,$97
	.dc.b	$2F,$C8,$36,$77,$E8,$7F,$18,$F9,$4F,$CD,$DF,$D0,$6B,$F3,$DA,$FA
	.dc.b	$B8,$11,$88,$FC,$C7,$C9,$BD,$CF,$E4,$6D,$8F,$E2,$6D,$7F,$55,$F4
	.dc.b	$3F,$A6,$7F,$59,$F2,$7F,$C5,$DF,$E3,$2B,$E4,$7F,$96,$B6,$33,$E5
	.dc.b	$FA,$7E,$53,$F0,$59,$1A,$49,$FC,$5F,$AD,$FD,$99,$FE,$4A,$D7,$F7
	.dc.b	$8D,$E0,$FD,$B2,$F2,$FF,$9E,$ED,$FE,$75,$9F,$DD,$7C,$FB,$E2,$7F
	.dc.b	$A0,$B7,$0F,$E6,$35,$BE,$2F,$EB,$52,$F9,$D7,$9B,$FA,$18,$C3,$63
	.dc.b	$37,$FB,$69,$CB,$70,$FC,$16,$3F,$C3,$A7,$C5,$CF,$FD,$89,$B5,$FD
	.dc.b	$21,$F3,$79,$F9,$FC,$9B,$E8,$5F,$1F,$85,$FA,$8D,$E7,$D6,$BC,$F4
	.dc.b	$7C,$9F,$94,$99,$F9,$63,$2F,$9F,$E8,$3F,$91,$57,$C0,$6C,$34,$46
	.dc.b	$1B,$E7,$7A,$7E,$31,$F2,$3E,$BF,$AA,$F8,$78,$7F,$59,$B4,$63,$5F
	.dc.b	$85,$7C,$2B,$E0,$2F,$93,$F9,$CD,$8F,$98,$D7,$FD,$73,$68,$C4,$7E
	.dc.b	$E2,$31,$BF,$94,$91,$F3,$7E,$C3,$F8,$66,$1E,$2E,$0F,$51,$27,$E9
	.dc.b	$7E,$4F,$F4,$37,$FC,$1B,$FA,$18,$EC,$D7,$31,$F8,$58,$C4,$7C,$9F
	.dc.b	$AC,$7D,$07,$DF,$7E,$8C,$7C,$56,$E1,$F8,$57,$71,$F8,$35,$F3,$3C
	.dc.b	$6A,$4E,$63,$F1,$5F,$58,$F9,$EB,$F3,$32,$BF,$19,$39,$6F,$CD,$70
	.dc.b	$A9,$18,$3F,$C0,$3F,$57,$FD,$D1,$BF,$01,$FE,$2F,$8F,$EE,$B1,$FD
	.dc.b	$71,$D3,$7D,$FD,$11,$E7,$F0,$1C,$4F,$F1,$FC,$FF,$98,$23,$FA,$EB
	.dc.b	$5F,$F7,$AB,$5F,$E6,$7F,$21,$FE,$3C,$7F,$B5,$AF,$AA,$FF,$21,$36
	.dc.b	$3F,$C3,$2E,$F8,$3F,$8C,$37,$8B,$9F,$C4,$F0,$38,$FA,$7F,$E4,$BF
	.dc.b	$9F,$70,$FF,$96,$B1,$FE,$E3,$18,$C7,$FB,$13,$63,$33,$B7,$FA,$48
	.dc.b	$8F,$A9,$F5,$5C,$38,$0D,$24,$FF,$98,$BC,$AF,$F3,$1A,$69,$1F,$ED
	.dc.b	$9F,$91,$BE,$53,$FC,$33,$97,$F8,$5E,$31,$AF,$F8,$EB,$F0,$7F,$F3
	.dc.b	$A3,$FE,$4E,$8C,$DF,$F8,$97,$F2,$8F,$F2,$84,$FF,$94,$9B,$1F,$E2
	.dc.b	$99,$DD,$FF,$88,$E7,$2F,$F4,$3F,$55,$F3,$9F,$A0,$6E,$1F,$E0,$76
	.dc.b	$E1,$FE,$1A,$6B,$F1,$BE,$DB,$F7,$53,$FE,$50,$6C,$7F,$8A,$27,$8D
	.dc.b	$FF,$51,$9C,$BF,$CA,$FC,$0F,$6D,$7F,$0F,$6C,$6F,$C3,$F8,$B8,$7F
	.dc.b	$3C,$9C,$BF,$E9,$7F,$99,$7E,$25,$FE,$A9,$8F,$F3,$CC,$63,$C9,$E3
	.dc.b	$8A,$37,$F1,$98,$75,$5C,$39,$DF,$09,$96,$C7,$4C,$7B,$61,$FE,$7D
	.dc.b	$B5,$6A,$31,$F8,$97,$F9,$62,$FC,$6D,$8C,$FA,$9E,$18,$C6,$DF,$E5
	.dc.b	$AC,$1F,$8B,$7F,$B0,$70,$F5,$B7,$0E,$BF,$1F,$E9,$8C,$5F,$1F,$BD
	.dc.b	$FC,$0D,$7F,$88,$B0,$46,$23,$E4,$35,$F8,$8D,$8F,$ED,$1F,$15,$F1
	.dc.b	$7D,$3F,$F2,$9E,$5F,$32,$BF,$12,$FD,$18,$CD,$FF,$47,$FB,$07,$D5
	.dc.b	$FD,$0F,$45,$18,$BF,$EC,$CF,$DA,$AF,$EA,$CE,$5F,$C1,$65,$FC,$D6
	.dc.b	$9C,$BF,$66,$FF,$BB,$36,$BF,$A5,$B5,$F8,$18,$FF,$93,$E7,$F0,$6D
	.dc.b	$8F,$B1,$97,$F8,$7E,$73,$F1,$1F,$83,$FC,$4D,$E6,$53,$1F,$1F,$6B
	.dc.b	$F1,$BF,$80,$FF,$3C,$7F,$CF,$BA,$A3,$19,$7F,$0A,$8C,$7A,$1F,$D3
	.dc.b	$DB,$5F,$E8,$4F,$C2,$3F,$BB,$7D,$87,$E5,$9B,$1F,$ED,$27,$FB,$D7
	.dc.b	$D5,$E7,$85,$5B,$AF,$F2,$11,$B2,$E3,$FE,$D1,$B7,$D8,$7F,$B4,$27
	.dc.b	$5E,$C7,$7F,$C0,$DE,$D7,$F6,$3B,$46,$2B,$CC,$DC,$3F,$92,$CE,$AB
	.dc.b	$87,$E2,$6F,$A6,$F5,$37,$0F,$52,$F8,$EC,$CF,$BF,$F1,$3F,$9F,$5E
	.dc.b	$F6,$E3,$E9,$7F,$D4,$F1,$9E,$4F,$8D,$F8,$43,$D5,$F4,$36,$7E,$63
	.dc.b	$6F,$CE,$D7,$FD,$0B,$A3,$F1,$3F,$F7,$ED,$7C,$F6,$CF,$C9,$7F,$D3
	.dc.b	$D1,$9F,$A1,$FA,$DE,$6F,$CE,$F0,$BF,$DA,$38,$23,$14,$C1,$2F,$92
	.dc.b	$FA,$51,$F5,$6B,$6E,$B7,$FF,$36,$5F,$E1,$5B,$1F,$41,$AF,$A9,$AF
	.dc.b	$7B,$7E,$17,$FE,$33,$AF,$C5,$36,$7F,$08,$FF,$90,$23,$3F,$09,$F9
	.dc.b	$1F,$8A,$7D,$0D,$31,$F5,$36,$BF,$29,$F0,$1F,$E1,$77,$F3,$4D,$C7
	.dc.b	$F0,$AD,$7E,$7B,$5F,$37,$E2,$6F,$B2,$FD,$03,$6B,$F4,$0F,$F4,$87
	.dc.b	$71,$FA,$5F,$CA,$7F,$84,$6A,$69,$FC,$0F,$87,$E9,$BA,$FA,$6F,$F4
	.dc.b	$65,$FE,$BD,$B8,$FE,$69,$AF,$E1,$1A,$FC,$BF,$9C,$FC,$9D,$FD,$E6
	.dc.b	$D6,$7C,$6F,$DC,$46,$2B,$F0,$1F,$53,$53,$4F,$8D,$7F,$EB,$37,$15
	.dc.b	$FE,$50,$7F,$3F,$3F,$E1,$23,$8C,$64,$FF,$2C,$7E,$0D,$FE,$D9,$F9
	.dc.b	$8F,$EB,$AD,$C7,$F2,$18,$75,$AC,$FC,$AF,$E2,$A3,$F2,$FF,$43,$F9
	.dc.b	$26,$D7,$F1,$18,$FD,$C6,$3F,$C2,$5D,$47,$E3,$7E,$83,$F8,$ED,$7E
	.dc.b	$81,$B1,$F9,$19,$DD,$7E,$82,$72,$FE,$8B,$F1,$3F,$E6,$57,$FA,$86
	.dc.b	$DF,$E6,$5F,$E9,$71,$9F,$3F,$F9,$27,$DD,$7C,$DE,$23,$6B,$8E,$1F
	.dc.b	$AE,$8F,$A1,$F7,$3E,$FC,$7F,$65,$28,$C3,$FB,$57,$FA,$DB,$77,$FF
	.dc.b	$03,$BF,$DF,$3E,$36,$FB,$A7,$35,$CA,$75,$7F,$40,$EF,$5F,$DD,$1B
	.dc.b	$F6,$1C,$AF,$9B,$FC,$FF,$F2,$AF,$F0,$46,$DF,$C2,$9B,$5F,$BB,$79
	.dc.b	$61,$FC,$29,$CB,$FC,$33,$F1,$7F,$F0,$8C,$FF,$47,$6C,$7F,$41,$9D
	.dc.b	$EC,$9C,$DF,$F1,$8D,$DF,$82,$FF,$51,$6E,$3F,$D4,$5F,$F7,$83,$63
	.dc.b	$FA,$B7,$C8,$7F,$43,$AF,$E4,$2D,$9F,$E2,$55,$CB,$2F,$EA,$53,$97
	.dc.b	$F9,$27,$D8,$FF,$AB,$2B,$E4,$46,$3E,$63,$E8,$CF,$FB,$8A,$73,$28
	.dc.b	$E0,$B0,$4F,$DE,$9F,$EE,$33,$FE,$7A,$6B,$9F,$63,$FD,$7B,$18,$8F
	.dc.b	$DB,$F7,$7E,$03,$AF,$F8,$94,$75,$BE,$07,$8A,$33,$87,$82,$5F,$42
	.dc.b	$FC,$1F,$F3,$37,$FD,$06,$31,$F0,$5F,$E5,$25,$F7,$8F,$B6,$C5,$F0
	.dc.b	$C3,$F4,$DD,$BF,$E0,$3E,$CF,$EA,$31,$E3,$3B,$CE,$F9,$53,$AD,$3C
	.dc.b	$99,$3E,$D3,$FC,$5C,$DC,$3C,$5F,$79,$FC,$5E,$31,$B7,$D2,$F0,$9E
	.dc.b	$0B,$EE,$3F,$84,$9E,$9C,$7F,$C7,$13,$A8,$FA,$FF,$21,$2F,$EB,$F2
	.dc.b	$F1,$6C,$67,$3F,$F5,$27,$99,$F5,$FC,$DF,$95,$F2,$E8,$FB,$55,$87
	.dc.b	$57,$FD,$AE,$75,$7F,$8A,$BD,$36,$75,$5C,$9B,$19,$3C,$27,$D8,$DB
	.dc.b	$FD,$3B,$7D,$36,$46,$2F,$ED,$3F,$CD,$27,$4E,$6B,$97,$C9,$F8,$B5
	.dc.b	$7E,$15,$FD,$BE,$F7,$39,$8C,$61,$F8,$5F,$BC,$FC,$16,$7E,$6C,$DF
	.dc.b	$79,$FE,$A4,$C7,$C9,$5A,$CF,$D3,$5A,$7E,$0B,$6F,$95,$F1,$C3,$B7
	.dc.b	$46,$A3,$17,$FE,$15,$FB,$9F,$E3,$9F,$0F,$F1,$A1,$18,$7E,$39,$8F
	.dc.b	$5B,$E7,$BA,$F2,$3F,$D5,$8E,$A7,$86,$BF,$5A,$35,$7D,$67,$FC,$85
	.dc.b	$9D,$BE,$37,$B9,$E3,$9F,$F3,$86,$E1,$E7,$C7,$C3,$18,$8E,$A7,$86
	.dc.b	$F9,$D7,$1A,$DC,$62,$7E,$03,$5E,$37,$E1,$9D,$E3,$E5,$24,$FB,$63
	.dc.b	$2F,$78,$CC,$F1,$C6,$E1,$E9,$C3,$B6,$3E,$73,$E0,$E1,$C4,$97,$F0
	.dc.b	$67,$78,$F6,$8C,$63,$D2,$BE,$16,$35,$F4,$E3,$DE,$7E,$75,$F1,$8C
	.dc.b	$4F,$67,$E2,$57,$1C,$7C,$98,$F8,$23,$0F,$A0,$C5,$71,$3E,$4E,$1C
	.dc.b	$70,$ED,$18,$34,$7E,$1E,$B9,$CE,$DB,$1A,$BF,$87,$8F,$BA,$31,$87
	.dc.b	$82,$30,$4C,$7D,$6F,$CF,$9D,$E1,$D5,$A7,$5C,$7C,$AF,$C6,$3C,$D3
	.dc.b	$C4,$E8,$E2,$BA,$9F,$19,$E6,$7F,$82,$8F,$C7,$97,$0F,$D3,$37,$0F
	.dc.b	$D1,$B9,$F9,$BF,$E8,$AF,$F8,$EE,$51,$88,$D6,$BF,$E3,$F7,$F6,$31
	.dc.b	$97,$F4,$4F,$2C,$3F,$4C,$DC,$33,$5F,$AD,$FB,$3F,$F0,$FD,$6A,$3F
	.dc.b	$61,$5B,$BF,$E0,$0D,$7F,$7F,$5C,$2F,$9C,$7E,$EE,$75,$87,$A2,$FB
	.dc.b	$4F,$46,$3B,$72,$D2,$61,$E3,$7F,$83,$61,$DD,$F1,$CF,$29,$D1,$FC
	.dc.b	$0E,$75,$3F,$12,$75,$87,$9A,$B5,$38,$39,$9E,$08,$FE,$15,$5C,$AF
	.dc.b	$D5,$39,$37,$5C,$9C,$1D,$CB,$09,$E5,$87,$C4,$26,$1F,$02,$FD,$38
	.dc.b	$79,$49,$87,$F2,$FB,$E3,$5E,$78,$ED,$5E,$13,$8B,$E4,$AE,$2F,$D6
	.dc.b	$38,$61,$EA,$9D,$C6,$4F,$3E,$1E,$47,$E7,$9C,$1C,$47,$7A,$EB,$5C
	.dc.b	$DB,$87,$9A,$F9,$E1,$ED,$9D,$E1,$C9,$F9,$B3,$D6,$7C,$B5,$C7,$0F
	.dc.b	$2E,$1D,$6B,$E1,$1B,$3E,$6C,$EE,$FA,$DF,$79,$FF,$16,$FD,$A7,$F8
	.dc.b	$9D,$FE,$CC,$D7,$FB,$1B,$70,$FE,$F4,$BC,$CF,$C7,$52,$31,$5C,$B8
	.dc.b	$33,$1C,$70,$FF,$1C,$3A,$9F,$ED,$FF,$60,$FB,$AF,$CC,$73,$7F,$7E
	.dc.b	$E1,$BC,$BF,$C0,$35,$FF,$05,$5E,$F0,$FE,$F8,$D7,$9E,$1E,$48,$C4
	.dc.b	$F0,$FA,$4F,$CB,$12,$BF,$B4,$3F,$99,$A9,$18,$21,$FD,$C5,$CB,$FE
	.dc.b	$47,$3F,$BD,$BF,$EA,$2E,$FF,$F2,$C4,$60,$FF,$36,$F0,$BE,$73,$5C
	.dc.b	$C9,$6A,$FA,$07,$B9,$A9,$CC,$61,$FF,$25,$35,$FF,$2C,$71,$FF,$8B
	.dc.b	$38,$78,$36,$3F,$D0,$2D,$77,$D9,$F5,$DC,$F0,$3E,$77,$C1,$E2,$BD
	.dc.b	$E7,$C2,$E1,$B7,$FE,$8C,$9D,$4F,$83,$E9,$9F,$31,$49,$5E,$28,$CB
	.dc.b	$E4,$F9,$C4,$7B,$E3,$FD,$F7,$D7,$7C,$FF,$C9,$73,$7F,$BE,$E3,$5C
	.dc.b	$DE,$7A,$F9,$BA,$FF,$3D,$7F,$C5,$AD,$8D,$47,$F4,$58,$C7,$DD,$72
	.dc.b	$24,$7D,$6F,$BA,$F8,$3E,$AB,$C9,$A4,$9F,$F5,$C7,$D4,$7C,$6F,$5C
	.dc.b	$D2,$47,$FA,$F7,$6F,$97,$EC,$BF,$00,$A4,$9F,$95,$ED,$FF,$21,$6F
	.dc.b	$8F,$F4,$A3,$11,$FE,$CF,$8C,$79,$1F,$1F,$EB,$5E,$C5,$25,$7F,$AF
	.dc.b	$21,$F4,$1B,$24,$65,$E5,$CF,$23,$83,$5F,$CF,$4B,$ED,$BF,$0B,$F0
	.dc.b	$5F,$85,$9F,$C1,$35,$FF,$55,$76,$23,$1D,$3F,$EB,$B8,$FC,$67,$D5
	.dc.b	$FE,$2C,$95,$F8,$BE,$E4,$7E,$3F,$83,$F2,$7F,$08,$F8,$4A,$4B,$FC
	.dc.b	$97,$E2,$8D,$BF,$32,$D8,$CC,$60,$CE,$FF,$AD,$F2,$3F,$61,$F2,$57
	.dc.b	$C8,$34,$FC,$97,$DF,$CC,$3F,$C1,$5E,$1A,$3F,$20,$D8,$CC,$7E,$4F
	.dc.b	$F2,$4F,$F0,$E7,$17,$F0,$46,$F9,$9F,$C0,$69,$18,$7F,$82,$3F,$E1
	.dc.b	$E9,$F9,$1F,$8A,$FE,$3A,$7F,$60,$D7,$E7,$B6,$37,$DB,$FB,$AF,$8C
	.dc.b	$FF,$0F,$57,$F0,$FE,$64,$62,$11,$8F,$33,$FC,$4C,$D8,$FF,$1A,$1A
	.dc.b	$FB,$CE,$2F,$F0,$16,$BF,$80,$6E,$1F,$7D,$FF,$4B,$3F,$81,$FC,$57
	.dc.b	$F9,$B3,$7D,$2F,$E1,$F1,$8D,$7C,$CB,$D3,$FE,$27,$F7,$AF,$8B,$F2
	.dc.b	$B7,$2F,$2E,$68,$C4,$7F,$92,$E3,$13,$EC,$FC,$57,$F8,$7C,$FF,$19
	.dc.b	$6B,$F9,$86,$E7,$F9,$75,$F8,$17,$F2,$4F,$95,$FA,$9A,$7C,$B6,$DF
	.dc.b	$CA,$6E,$27,$EE,$2F,$7F,$F6,$F3,$E8,$6B,$E7,$6B,$F8,$58,$E1,$E2
	.dc.b	$FD,$4F,$96,$FD,$DC,$FF,$31,$6B,$FB,$06,$E7,$FA,$F7,$AB,$FB,$1F
	.dc.b	$C5,$7F,$A4,$37,$DA,$FE,$7F,$18,$CF,$F0,$CF,$EE,$1F,$C4,$7C,$DF
	.dc.b	$E9,$30,$F9,$ED,$7E,$73,$5F,$43,$5F,$C8,$7E,$2A,$F4,$FF,$A6,$B6
	.dc.b	$3F,$A5,$35,$FE,$12,$BD,$4F,$EA,$7F,$8A,$BE,$72,$92,$BF,$A9,$43
	.dc.b	$E2,$BF,$AC,$FD,$13,$E5,$3F,$36,$DA,$FC,$3B,$5F,$C3,$35,$FA,$1A
	.dc.b	$77,$7F,$AF,$35,$FE,$3A,$DD,$FF,$8E,$3E,$18,$FE,$C9,$0F,$8C,$FE
	.dc.b	$D6,$D8,$CC,$63,$D9,$FE,$FD,$F0,$CF,$A8,$7B,$7F,$E9,$78,$C3,$63
	.dc.b	$37,$FB,$49,$CB,$7F,$15,$FE,$BF,$8B,$2F,$F0,$75,$27,$31,$86,$D2
	.dc.b	$75,$F0,$DE,$4F,$AC,$FC,$EC,$1A,$63,$EB,$1F,$2D,$FE,$17,$FA,$A7
	.dc.b	$D7,$79,$B1,$FF,$5E,$46,$36,$FF,$59,$36,$3F,$86,$B5,$DE,$EF,$A9
	.dc.b	$E8,$D8,$FF,$14,$6F,$ED,$55,$FD,$DE,$BC,$5F,$77,$FE,$32,$AD,$6B
	.dc.b	$E9,$BF,$61,$0E,$7C,$1F,$1B,$5F,$B8,$DE,$7F,$A5,$1F,$81,$9F,$E7
	.dc.b	$14,$8C,$46,$CF,$A9,$F7,$8F,$AB,$FB,$AD,$8E,$E7,$CF,$70,$B1,$F3
	.dc.b	$FD,$77,$EB,$5B,$EA,$FE,$1A,$31,$9F,$DC,$6D,$7D,$95,$F7,$3F,$8C
	.dc.b	$FA,$36,$34,$95,$E2,$D1,$F6,$2F,$9C,$FC,$9B,$67,$4D,$7C,$D9,$7E
	.dc.b	$32,$31,$A7,$CE,$7F,$58,$FD,$5E,$2F,$C1,$F8,$FF,$85,$E6,$FE,$27
	.dc.b	$7F,$E6,$19,$CE,$1F,$10,$7E,$1D,$FD,$6E,$7F,$B3,$36,$3F,$CD,$0D
	.dc.b	$77,$F1,$9F,$E4,$9F,$9A,$FF,$04,$C7,$36,$BF,$D6,$9A,$FF,$71,$0E
	.dc.b	$1F,$EE,$8D,$5E,$B3,$F2,$BF,$F2,$9E,$1F,$E7,$98,$C7,$C9,$7F,$9C
	.dc.b	$27,$35,$C3,$E9,$BE,$8B,$FE,$72,$6B,$FE,$78,$6E,$DF,$9B,$B4,$63
	.dc.b	$0F,$F0,$37,$E5,$8F,$BD,$FF,$33,$7C,$9D,$C3,$FB,$AE,$28,$C3,$FD
	.dc.b	$E1,$AF,$BD,$AF,$F8,$45,$FC,$3F,$DE,$F2,$6C,$71,$E8,$4E,$5B,$1F
	.dc.b	$E6,$FF,$9C,$FF,$38,$26,$91,$FF,$04,$FC,$2F,$FA,$7F,$3F,$3B,$63
	.dc.b	$27,$FA,$A0,$FF,$12,$39,$7F,$C4,$DA,$BE,$F5,$F2,$7B,$9F,$AB,$18
	.dc.b	$BF,$F0,$BD,$6E,$B4,$D8,$F8,$FE,$B3,$EA,$79,$60,$8C,$4E,$A7,$31
	.dc.b	$89,$FF,$11,$AF,$A4,$C9,$2B,$AB,$D3,$EF,$3F,$D1,$51,$F5,$9B,$5F
	.dc.b	$E2,$C6,$BF,$93,$6B,$FC,$E3,$EE,$FF,$D6,$93,$FC,$FD,$B1,$F9,$9E
	.dc.b	$F4,$62,$3F,$8A,$B5,$E5,$E6,$FC,$79,$DB,$67,$FD,$A9,$5D,$63,$5F
	.dc.b	$77,$FE,$6B,$8F,$E9,$CD,$BF,$03,$5F,$E4,$6D,$7F,$A7,$BF,$E9,$08
	.dc.b	$E3,$F1,$DD,$08,$FF,$59,$F2,$3F,$CF,$8D,$F9,$1F,$34,$62,$FD,$CD
	.dc.b	$7F,$98,$AF,$A0,$D9,$2B,$A3,$F8,$5F,$07,$FA,$9D,$5F,$0B,$F1,$8F
	.dc.b	$F1,$16,$BF,$D2,$9B,$1F,$D6,$BE,$7F,$81,$2F,$FA,$BC,$FF,$B7,$9F
	.dc.b	$57,$CC,$FA,$08,$E0,$FE,$ED,$F7,$4F,$AF,$FA,$4D,$8C,$C6,$0F,$AD
	.dc.b	$EB,$FF,$33,$8F,$EF,$CD,$CB,$F0,$0D,$7F,$A9,$B5,$FF,$3F,$64,$F3
	.dc.b	$7F,$4D,$F0,$7F,$D9,$9B,$19,$BF,$F3,$4B,$6B,$FC,$D1,$F7,$3F,$E0
	.dc.b	$99,$FF,$36,$B6,$33,$1D,$23,$11,$FE,$9B,$9F,$EC,$1F,$80,$FD,$1F
	.dc.b	$48,$D2,$5F,$F9,$E3,$EE,$BF,$CC,$8D,$F8,$7F,$3C,$63,$0F,$F3,$D3
	.dc.b	$5F,$F3,$27,$DC,$FC,$97,$FC,$FC,$D7,$3B,$7C,$08,$C4,$7F,$96,$BE
	.dc.b	$AF,$E1,$FC,$39,$A4,$C7,$FD,$17,$F7,$3F,$E1,$C6,$FB,$DF,$2A,$31
	.dc.b	$7F,$E8,$D6,$BF,$E7,$7F,$B9,$F8,$6F,$FA,$BD,$AF,$FA,$B9,$BC,$47
	.dc.b	$3E,$23,$F0,$6F,$C6,$57,$F5,$58,$C7,$4F,$F1,$A9,$CC,$FF,$24,$6B
	.dc.b	$FE,$93,$F6,$7E,$FF,$C6,$FF,$80,$34,$FF,$52,$B6,$7F,$D4,$8D,$7F
	.dc.b	$D1,$DF,$5C,$F8,$EF,$C3,$B6,$BF,$D5,$AD,$9F,$F5,$63,$5F,$F4,$B7
	.dc.b	$DD,$7F,$BA,$27,$FD,$66,$D8,$FF,$59,$35,$DF,$BF,$F9,$4F,$BA,$FE
	.dc.b	$55,$5F,$EB,$A6,$CF,$FA,$E5,$AF,$FA,$8D,$AF,$FB,$6F,$EE,$BF,$CB
	.dc.b	$13,$FE,$C2,$6C,$7F,$B0,$5A,$FF,$AA,$DA,$FF,$79,$FC,$31,$F4,$DF
	.dc.b	$D1,$9B,$5F,$D2,$9B,$3F,$4D,$AF,$FA,$D7,$EE,$FF,$1A,$FF,$BA,$9B
	.dc.b	$C6,$69,$A8,$7D,$17,$F8,$DA,$FF,$BC,$B6,$3F,$DD,$ED,$7F,$C2,$CD
	.dc.b	$44,$7E,$63,$83,$AE,$7F,$D7,$23,$31,$B8,$C7,$54,$AF,$0B,$67,$FD
	.dc.b	$D1,$CE,$FB,$CF,$E1,$2D,$AF,$F7,$7B,$67,$73,$A9,$86,$8B,$1F,$EF
	.dc.b	$3F,$B9,$7C,$27,$F1,$46,$CF,$FB,$E9,$B3,$B9,$D7,$C6,$4A,$D3,$67
	.dc.b	$FC,$74,$3E,$23,$FA,$24,$FF,$89,$1B,$1A,$6B,$FE,$41,$6B,$F2,$5F
	.dc.b	$E9,$CD,$F3,$7F,$C0,$B1,$88,$FA,$2D,$7F,$85,$7C,$27,$F8,$2B,$BF
	.dc.b	$FC,$97,$7F,$E3,$46,$CF,$99,$AF,$FB,$F7,$EB,$BF,$C7,$F3,$EA,$E2
	.dc.b	$46,$27,$FC,$AD,$18,$F8,$3F,$E7,$73,$99,$E5,$F7,$F8,$25,$F3,$94
	.dc.b	$7F,$42,$D1,$F8,$2D,$25,$78,$63,$2D,$F9,$DE,$04,$70,$7F,$94,$3E
	.dc.b	$5B,$F9,$EB,$7C,$AF,$F1,$5C,$63,$2F,$AA,$D7,$FC,$E0,$BE,$FB,$64
	.dc.b	$AE,$8F,$2F,$97,$91,$DD,$B5,$FC,$97,$EF,$9F,$19,$FE,$3E,$6D,$7F
	.dc.b	$90,$9B,$3B,$9F,$F2,$1F,$DF,$7E,$3A,$3F,$C9,$AD,$7F,$CF,$8D,$7F
	.dc.b	$D0,$2D,$F2,$5F,$71,$FA,$C6,$F6,$FF,$93,$A3,$06,$7E,$F7,$FC,$1B
	.dc.b	$26,$92,$FE,$3F,$C9,$CC,$FE,$63,$63,$E4,$FC,$BF,$CC,$24,$FD,$16
	.dc.b	$C7,$F9,$C7,$C8,$FE,$05,$C0,$F9,$AF,$F7,$37,$C5,$A7,$37,$87,$73
	.dc.b	$6F,$FD,$3F,$A7,$F6,$DB,$E1,$18,$F8,$6D,$6B,$F0,$BC,$4B,$7F,$CE
	.dc.b	$2F,$73,$F1,$AB,$FD,$ED,$A7,$F9,$EA,$BF,$9E,$B6,$33,$87,$FA,$AE
	.dc.b	$72,$FF,$AB,$7E,$F7,$EC,$E1,$D3,$EF,$7F,$D1,$0F,$68,$CE,$9F,$EB
	.dc.b	$39,$CB,$6B,$CF,$18,$7F,$D7,$73,$C1,$FF,$5E,$C6,$57,$DD,$7F,$B1
	.dc.b	$3E,$23,$F9,$7E,$1F,$0A,$73,$E1,$27,$38,$FF,$31,$BF,$F0,$C5,$ED
	.dc.b	$E5,$7F,$54,$FE,$72,$E2,$35,$7C,$E7,$47,$8B,$DA,$FA,$8F,$83,$0E
	.dc.b	$EF,$29,$E1,$5C,$A7,$33,$C7,$5B,$87,$7C,$7C,$D8,$70,$6D,$EB,$1E
	.dc.b	$F5,$BC,$3B,$4F,$2B,$E1,$1E,$AA,$DC,$62,$3C,$C7,$AA,$B9,$3E,$33
	.dc.b	$11,$B8,$C9,$27,$9D,$77,$25,$73,$8C,$BC,$A3,$38,$FC,$F9,$CB,$F0
	.dc.b	$E7,$47,$1C,$77,$8F,$C0,$8C,$CF,$19,$CC,$7C,$03,$33,$F7,$6B,$AB
	.dc.b	$C6,$BB,$1C,$67,$67,$17,$31,$C6,$BA,$CF,$C6,$9C,$BF,$2E,$33,$87
	.dc.b	$C4,$C7,$E3,$61,$E0,$AE,$98,$78,$4E,$8F,$80,$E8,$E5,$E8,$66,$BA
	.dc.b	$63,$F7,$F0,$EF,$87,$D3,$3B,$B8,$8F,$1E,$1E,$1B,$EA,$78,$8E,$B8
	.dc.b	$FA,$9C,$93,$0F,$D1,$E3,$E3,$8C,$63,$D9,$F2,$3F,$A1,$7F,$C3,$47
	.dc.b	$E2,$8B,$3F,$AA,$6C,$6D,$FD,$87,$C6,$37,$FF,$01,$C7,$EE,$A3,$0F
	.dc.b	$F8,$1A,$91,$88,$EA,$D7,$9C,$FE,$A2,$38,$E1,$FB,$B6,$E1,$FB,$FF
	.dc.b	$B8,$FF,$18,$4E,$6B,$81,$2F,$F4,$53,$AA,$EB,$1F,$A5,$BE,$6F,$F0
	.dc.b	$46,$BF,$C4,$A7,$83,$9A,$E9,$7D,$E3,$A9,$8B,$EB,$19,$3F,$89,$57
	.dc.b	$29,$F8,$95,$B7,$A1,$C8,$F2,$39,$76,$66,$7A,$57,$62,$47,$F1,$CA
	.dc.b	$ED,$87,$CE,$BE,$4D,$9E,$55,$C1,$EB,$7D,$1E,$47,$0C,$3E,$1D,$71
	.dc.b	$9F,$5C,$E6,$FA,$E1,$E0,$BD,$92,$3F,$90,$C6,$5E,$13,$E2,$3E,$0D
	.dc.b	$78,$5F,$1E,$1F,$75,$DC,$67,$05,$F3,$9F,$8E,$49,$F2,$B6,$BD,$D5
	.dc.b	$E9,$7A,$46,$4F,$0B,$AA,$E9,$5D,$F0,$F8,$31,$89,$E7,$18,$AF,$2C
	.dc.b	$71,$69,$D0,$F9,$67,$73,$DA,$64,$FC,$71,$FD,$33,$07,$DE,$7F,$8B
	.dc.b	$23,$0D,$7F,$AD,$46,$BA,$BF,$2A,$F0,$F1,$AD,$7F,$BB,$6B,$8D,$60
	.dc.b	$CC,$F0,$F8,$67,$D5,$FF,$A6,$78,$7F,$D3,$58,$A7,$38,$7F,$A8,$27
	.dc.b	$31,$88,$46,$2B,$94,$7F,$A9,$AB,$97,$D5,$77,$99,$3E,$FB,$4F,$E6
	.dc.b	$27,$F3,$2C,$9F,$ED,$31,$97,$F8,$6E,$FE,$17,$F8,$AC,$7F,$10,$7F
	.dc.b	$C1,$2D,$87,$C6,$52,$4F,$27,$F8,$BE,$1F,$E4,$CD,$2F,$37,$D2,$AF
	.dc.b	$F4,$EC,$E5,$B1,$97,$F9,$3C,$7C,$7F,$BB,$E7,$99,$A4,$9E,$7F,$5E
	.dc.b	$F7,$BC,$2D,$8C,$C6,$0F,$2F,$DA,$26,$92,$7C,$B1,$97,$E2,$D7,$C1
	.dc.b	$23,$83,$E4,$FD,$3F,$C5,$23,$83,$E6,$FD,$4F,$C9,$23,$83,$FD,$2D
	.dc.b	$F3,$FF,$CF,$BD,$17,$9A,$8C,$36,$76,$7F,$A7,$72,$C9,$1C,$1F,$EA
	.dc.b	$7E,$9E,$9E,$87,$EF,$A3,$04,$3F,$D6,$6F,$F4,$A7,$FD,$5D,$D5,$C0
	.dc.b	$8F,$F5,$2F,$BA,$7C,$97,$E1,$5B,$19,$8C,$19,$DE,$47,$16,$BF,$ED
	.dc.b	$E9,$7C,$87,$32,$47,$77,$E2,$7D,$CB,$DF,$A4,$9F,$F6,$74,$65,$BF
	.dc.b	$73,$C0,$8E,$0F,$F6,$A7,$DC,$31,$FA,$0D,$8C,$C6,$0C,$E5,$23,$8B
	.dc.b	$5F,$F7,$BF,$BE,$FB,$7F,$D4,$36,$3F,$BC,$B5,$DC,$7E,$B7,$ED,$FF
	.dc.b	$5D,$06,$91,$D7,$91,$1F,$F0,$1C,$7F,$81,$D1,$1C,$2E,$57,$DB,$FE
	.dc.b	$6E,$3F,$DE,$2D,$78,$9B,$70,$F4,$8E,$3C,$9F,$BC,$F8,$37,$DE,$52
	.dc.b	$57,$EF,$23,$2F,$FC,$01,$FB,$7F,$F4,$73,$DA,$D1,$88,$EC,$D7,$D0
	.dc.b	$D8,$7D,$BF,$F1,$C9,$FF,$15,$9E,$D7,$0F,$F8,$BA,$37,$E3,$7E,$AE
	.dc.b	$73,$3E,$1F,$B7,$7C,$C7,$EB,$DB,$1D,$2D,$18,$8E,$8D,$7D,$5F,$6E
	.dc.b	$FB,$7F,$DC,$36,$39,$35,$F2,$35,$F8,$5F,$21,$FC,$B6,$CD,$24,$7F
	.dc.b	$2A,$FC,$1F,$FA,$43,$FC,$A1,$AF,$F2,$66,$BF,$D3,$9A,$FD,$9F,$A3
	.dc.b	$7C,$57,$F0,$D6,$C7,$BD,$AF,$C9,$6B,$F5,$FE,$E1,$D9,$FC,$55,$B1
	.dc.b	$F1,$DA,$FD,$16,$BF,$7F,$E6,$1F,$51,$FC,$75,$B1,$F3,$DA,$FD,$96
	.dc.b	$BF,$87,$FB,$87,$93,$FC,$95,$B1,$F5,$DA,$FE,$05,$AF,$E3,$FE,$E1
	.dc.b	$F0,$3F,$CB,$5B,$1F,$7D,$AF,$E2,$5A,$FE,$93,$E7,$DF,$6C,$F9,$1F
	.dc.b	$F3,$9C,$61,$B1,$99,$FC,$AC,$E5,$FC,$67,$B1,$D1,$FD,$63,$5F,$DB
	.dc.b	$37,$8F,$FC,$03,$8F,$F4,$3F,$AE,$7C,$2D,$25,$FE,$5E,$32,$FE,$93
	.dc.b	$C2,$7D,$1B,$E5,$BF,$D1,$31,$86,$C6,$67,$83,$5F,$F4,$BC,$BE,$DF
	.dc.b	$FB,$FC,$75,$6B,$A6,$BF,$C2,$A3,$87,$B7,$E0,$FA,$7E,$93,$96,$BF
	.dc.b	$C6,$25,$18,$84,$63,$0F,$9F,$93,$ED,$E9,$2B,$E7,$C6,$5F,$A3,$F6
	.dc.b	$FF,$E2,$A6,$EF,$F6,$23,$11,$FE,$B4,$E0,$46,$23,$CD,$F6,$FE,$92
	.dc.b	$3F,$37,$34,$7B,$78,$24,$E5,$AF,$CA,$F0,$9F,$1F,$F4,$9B,$1F,$5D
	.dc.b	$AE,$E3,$5D,$52,$34,$D7,$EA,$FD,$BB,$E3,$FE,$EB,$63,$F0,$4D,$77
	.dc.b	$1A,$F9,$39,$1A,$6B,$FC,$B3,$ED,$FF,$C1,$CD,$F7,$7F,$C8,$51,$88
	.dc.b	$FF,$6D,$B5,$FE,$93,$F1,$7F,$DC,$5F,$E6,$AD,$7F,$DC,$2D,$77,$EC
	.dc.b	$7F,$93,$63,$FA,$3C,$70,$F3,$F0,$47,$FC,$8F,$F3,$4F,$A3,$FF,$26
	.dc.b	$B6,$33,$18,$3F,$CB,$1F,$6C,$FB,$6A,$49,$FF,$75,$C6,$5F,$F7,$6F
	.dc.b	$DB,$6F,$C0,$B4,$69,$27,$FA,$8C,$FF,$1C,$FB,$6E,$3C,$46,$92,$7F
	.dc.b	$CD,$5F,$23,$FD,$F8,$8E,$0F,$EB,$7F,$55,$FE,$09,$EC,$FE,$F7,$8F
	.dc.b	$E5,$A3,$13,$E3,$9F,$EE,$F0,$F9,$1F,$CF,$56,$FD,$CB,$5A,$6D,$FF
	.dc.b	$81,$BE,$DB,$FD,$20,$F3,$3F,$C1,$38,$76,$BF,$D2,$CE,$9F,$D3,$FA
	.dc.b	$0F,$98,$FD,$35,$FF,$84,$23,$18,$FF,$74,$84,$61,$FE,$EE,$D7,$FC
	.dc.b	$1E,$6B,$ED,$7C,$AD,$9F,$82,$7F,$85,$27,$95,$F8,$AB,$70,$6F,$CC
	.dc.b	$E1,$B5,$CA,$FF,$85,$FD,$A7,$F4,$A8,$C7,$90,$87,$C6,$FC,$68,$C7
	.dc.b	$A5,$67,$35,$C1,$B7,$F2,$6B,$97,$C7,$6F,$0E,$0F,$3F,$B4,$FF,$2E
	.dc.b	$E1,$E0,$7F,$D8,$4D,$F8,$AD,$7C,$8C,$8D,$7C,$4D,$CB,$B5,$FF,$95
	.dc.b	$E7,$D3,$E5,$3E,$87,$E9,$E9,$FC,$A9,$B1,$99,$FE,$31,$18,$F5,$BD
	.dc.b	$D3,$FD,$6E,$79,$CE,$27,$F0,$FE,$73,$ED,$3F,$99,$D7,$CB,$8C,$61
	.dc.b	$F1,$FF,$17,$FF,$15,$CE,$6B,$79,$FE,$36,$F9,$3F,$88,$6C,$FF,$66
	.dc.b	$9F,$C9,$CF,$F6,$F6,$BF,$E5,$C9,$F1,$7C,$A7,$93,$CC,$EB,$D1,$FC
	.dc.b	$F6,$73,$87,$96,$72,$FF,$4D,$FB,$5F,$C3,$3F,$80,$6E,$1F,$65,$74
	.dc.b	$7E,$05,$CE,$1F,$D2,$74,$FE,$D7,$E9,$FC,$EF,$63,$FB,$9E,$3F,$DA
	.dc.b	$23,$07,$E9,$3D,$96,$8F,$A5,$FD,$8C,$63,$E9,$AC,$E6,$FF,$2D,$39
	.dc.b	$6F,$DA,$BE,$AF,$8B,$63,$3F,$43,$89,$FA,$FF,$6B,$F8,$CE,$1F,$EB
	.dc.b	$8D,$7F,$C3,$4D,$7E,$63,$5F,$F4,$37,$DA,$BD,$2F,$3A,$F9,$9B,$18
	.dc.b	$7D,$7F,$6B,$F1,$B2,$FF,$43,$36,$FF,$D2,$0F,$CC,$8F,$6B,$9D,$1F
	.dc.b	$0D,$FC,$66,$7F,$C8,$5E,$67,$C9,$8C,$5F,$F9,$0E,$3E,$67,$97,$FE
	.dc.b	$E4,$AF,$8E,$FF,$97,$1B,$F3,$BF,$3A,$31,$5F,$C7,$70,$7D,$AF,$C9
	.dc.b	$8C,$37,$A3,$FC,$0D,$3A,$E1,$FA,$F3,$9C,$7F,$D5,$BF,$3E,$FB,$5F
	.dc.b	$CA,$E2,$8C,$3F,$C9,$1B,$3F,$E9,$66,$C7,$FA,$57,$E3,$DC,$7F,$84
	.dc.b	$7F,$C3,$95,$F8,$27,$2D,$F8,$5F,$C4,$FD,$33,$F0,$8A,$63,$D9,$B5
	.dc.b	$F8,$AF,$B5,$FE,$5D,$97,$FA,$E5,$AF,$FB,$0D,$BB,$9B,$F6,$AF,$A6
	.dc.b	$FC,$BA,$FA,$2B,$18,$7F,$37,$F6,$AF,$22,$98,$F8,$AB,$9F,$F3,$1E
	.dc.b	$27,$DA,$F5,$8F,$88,$DD,$7F,$A1,$CE,$B2,$F9,$53,$99,$FF,$64,$7C
	.dc.b	$F7,$E2,$4F,$C6,$3F,$75,$C7,$95,$FD,$7B,$EF,$1F,$2D,$FE,$61,$CB
	.dc.b	$FA,$14,$63,$E4,$7F,$67,$F9,$6B,$E2,$38,$5E,$F2,$FE,$C8,$3E,$93
	.dc.b	$FD,$33,$8F,$FB,$EB,$E1,$BF,$B0,$CE,$5B,$3A,$8F,$F6,$AB,$FE,$40
	.dc.b	$9F,$EA,$51,$CF,$E4,$3F,$AF,$CE,$FC,$1F,$F2,$9B,$B9,$FC,$C6,$6F
	.dc.b	$07,$F8,$4B,$FB,$5C,$3F,$83,$B9,$6F,$91,$A3,$7E,$55,$8D,$BF,$C4
	.dc.b	$BE,$DD,$F4,$D7,$DC,$C3,$6B,$F8,$9C,$FF,$4F,$CB,$F5,$4F,$EB,$34
	.dc.b	$FD,$3B,$96,$FC,$E7,$CA,$F2,$7F,$BD,$FD,$BF,$F6,$0B,$FF,$07,$72
	.dc.b	$FD,$49,$CF,$C0,$4C,$BE,$2D,$FF,$19,$C7,$FC,$EA,$F1,$6F,$C0,$74
	.dc.b	$F7,$65,$70,$8B,$97,$EF,$FE,$AB,$FD,$E9,$3F,$CB,$1B,$19,$FA,$6F
	.dc.b	$E6,$71,$8C,$3F,$80,$BF,$83,$AE,$1D,$3A,$FA,$2D,$BF,$CE,$BF,$2E
	.dc.b	$FE,$01,$3F,$CE,$1B,$8F,$F0,$F5,$F8,$AE,$FC,$2F,$E8,$3C,$0F,$6A
	.dc.b	$99,$C3,$FA,$0C,$65,$F8,$7F,$33,$F2,$8F,$F1,$CD,$7E,$5F,$C6,$C4
	.dc.b	$FE,$33,$C6,$FE,$FD,$9F,$F8,$B0,$FF,$6F,$FC,$27,$E5,$A3,$31,$FE
	.dc.b	$30,$6A,$22,$FB,$8F,$F1,$9B,$6B,$B5,$62,$7C,$7F,$70,$F9,$7B,$F6
	.dc.b	$78,$6D,$FF,$01,$94,$F0,$F6,$B3,$97,$F9,$47,$45,$7F,$40,$E5,$FF
	.dc.b	$5C,$43,$EC,$7F,$CF,$58,$FF,$0B,$6D,$F1,$7F,$55,$1F,$CD,$9C,$D7
	.dc.b	$27,$F5,$8D,$F1,$BE,$5C,$6A,$78,$47,$ED,$3C,$D9,$1F,$94,$F5,$AE
	.dc.b	$1F,$C1,$38,$3F,$D6,$13,$FE,$5F,$6C,$7D,$16,$BF,$3F,$8B,$E9,$7C
	.dc.b	$1F,$CC,$3E,$6B,$6B,$D9,$F2,$5F,$ED,$7F,$95,$7B,$3E,$B8,$C3,$ED
	.dc.b	$F9,$2B,$8F,$F5,$1D,$BF,$CD,$D7,$C2,$31,$7A,$E0,$F9,$5E,$8D,$BD
	.dc.b	$E3,$F1,$E3,$5F,$6B,$FD,$C3,$1E,$56,$8C,$57,$F6,$27,$FD,$27,$3A
	.dc.b	$73,$8F,$F6,$36,$D3,$ED,$7F,$D1,$F5,$BE,$5F,$24,$63,$0F,$C9,$60
	.dc.b	$FB,$4F,$F0,$97,$C5,$FE,$57,$0F,$64,$E7,$CF,$4E,$6F,$8D,$FF,$23
	.dc.b	$A5,$F0,$8E,$71,$89,$DD,$F2,$8C,$1F,$A3,$9D,$D7,$1F,$47,$E7,$CF
	.dc.b	$87,$0F,$03,$BB,$E5,$5D,$DB,$87,$92,$F8,$E1,$D2,$3C,$38,$72,$C3
	.dc.b	$A1,$E2,$3E,$15,$70,$BE,$71,$DE,$78,$5F,$38,$F4,$56,$E7,$35,$C2
	.dc.b	$7A,$E1,$EB,$E1,$CD,$49,$F0,$61,$CD,$F0,$CE,$6B,$D9,$57,$1F,$64
	.dc.b	$DA,$F2,$1C,$B0,$F0,$61,$F4,$F0,$E3,$39,$C3,$83,$D8,$C1,$23,$67
	.dc.b	$9D,$C6,$1F,$21,$A7,$C0,$C3,$9D,$FA,$49,$8F,$C3,$9D,$9F,$0A,$BD
	.dc.b	$2E,$27,$86,$3E,$8A,$E1,$87,$8A,$FC,$0F,$8C,$C5,$FC,$06,$BE,$B6
	.dc.b	$DF,$4C,$7E,$BD,$F4,$BE,$64,$BF,$51,$E3,$C3,$E4,$9E,$3C,$7D,$AE
	.dc.b	$2F,$9C,$61,$F8,$37,$D6,$BC,$97,$D5,$F2,$37,$0F,$24,$7E,$02,$78
	.dc.b	$3F,$92,$3F,$DE,$B1,$FA,$86,$E1,$FA,$76,$DF,$FB,$FE,$7F,$D0,$DA
	.dc.b	$5F,$86,$FF,$32,$57,$F5,$39,$CB,$63,$2F,$67,$79,$FE,$DD,$B9,$7E
	.dc.b	$F1,$B3,$FB,$5F,$B4,$FE,$0F,$5B,$C3,$F4,$75,$AA,$FE,$14,$DC,$17
	.dc.b	$C3,$0E,$97,$C0,$97,$FA,$29,$D3,$FC,$0E,$F9,$12,$7F,$63,$3A,$9E
	.dc.b	$F1,$CB,$0F,$24,$73,$31,$7E,$07,$F5,$F7,$E5,$BE,$58,$7C,$6B,$E2
	.dc.b	$4B,$FE,$3B,$3A,$AF,$15,$F2,$C3,$C6,$F6,$3D,$0D,$7F,$96,$4F,$18
	.dc.b	$C6,$1D,$EF,$AD,$62,$FA,$E1,$FC,$A7,$0E,$13,$DC,$91,$97,$86,$1D
	.dc.b	$27,$73,$E9,$9F,$19,$F0,$CE,$37,$F0,$E3,$51,$E6,$39,$9A,$3F,$00
	.dc.b	$61,$DE,$1F,$72,$31,$3C,$CF,$5B,$C4,$FA,$A5,$8E,$0D,$C3,$DF,$3C
	.dc.b	$2B,$AC,$62,$7C,$26,$F0,$F7,$3E,$03,$DB,$3F,$0D,$D1,$DF,$0F,$C5
	.dc.b	$CF,$FA,$B7,$B2,$F4,$B8,$29,$88,$97,$FD,$3A,$B8,$36,$BF,$C5,$6D
	.dc.b	$F9,$9E,$BE,$53,$EB,$EF,$DA,$CA,$E5,$39,$6C,$F0,$23,$AF,$A7,$FC
	.dc.b	$9D,$7F,$74,$F0,$1C,$BF,$DB,$A3,$0D,$AF,$F4,$EF,$12,$F3,$1F,$DA
	.dc.b	$DF,$F2,$A3,$4F,$E1,$0E,$1F,$F0,$03,$7D,$8F,$F2,$64,$7F,$AA,$1A
	.dc.b	$FF,$AD,$1B,$87,$92,$3F,$CB,$5B,$3E,$5F,$C3,$BB,$5C,$57,$FB,$4D
	.dc.b	$AF,$F9,$45,$B7,$FE,$50,$5F,$35,$C9,$B3,$C6,$30,$67,$E6,$7B,$FB
	.dc.b	$2F,$A3,$BE,$9C,$BF,$F7,$14,$E5,$BC,$5C,$2A,$46,$0F,$97,$87,$CA
	.dc.b	$7F,$C4,$8D,$E0,$F1,$C7,$FB,$59,$63,$93,$E6,$38,$8F,$E6,$D1,$83
	.dc.b	$E8,$7E,$1C,$F0,$BB,$36,$7F,$DA,$ED,$7F,$CE,$DF,$87,$3D,$AF,$BB
	.dc.b	$F1,$0F,$45,$BE,$1B,$7F,$E9,$C8,$C7,$AB,$23,$FD,$3E,$D5,$C5,$F6
	.dc.b	$EF,$A7,$AD,$DB,$D7,$78,$34,$97,$FE,$BD,$F1,$74,$47,$07,$FB,$07
	.dc.b	$C5,$F9,$5F,$6E,$E7,$A4,$9F,$F7,$AC,$3D,$BD,$F4,$F0,$64,$EA,$3F
	.dc.b	$3B,$EE,$3E,$45,$F4,$9C,$2B,$47,$FA,$B7,$83,$F2,$D5,$B9,$F9,$3F
	.dc.b	$5E,$F1,$AF,$C5,$4C,$38,$78,$BF,$EB,$5E,$45,$EE,$72,$23,$3E,$7F
	.dc.b	$4C,$24,$EA,$72,$DF,$B7,$FB,$4D,$F8,$3F,$B7,$06,$F5,$E3,$AE,$D7
	.dc.b	$18,$C3,$E2,$1F,$6F,$FE,$7A,$7E,$DE,$9F,$E7,$C8,$C7,$07,$ED,$E7
	.dc.b	$67,$E7,$63,$FC,$19,$19,$7F,$C1,$D1,$CD,$AF,$3F,$43,$F8,$44,$62
	.dc.b	$1C,$3F,$4F,$E4,$BE,$AF,$19,$A7,$E6,$3E,$29,$F7,$7F,$9A,$FB,$87
	.dc.b	$81,$7C,$09,$97,$C9,$9C,$BF,$43,$CA,$7C,$6F,$F3,$28,$F8,$11,$8F
	.dc.b	$37,$F8,$FB,$63,$FD,$E8,$D7,$E3,$73,$3E,$E1,$F0,$DF,$AC,$6C,$FD
	.dc.b	$96,$BF,$30,$FC,$02,$FC,$1B,$F6,$7F,$73,$FB,$09,$D5,$A3,$F8,$84
	.dc.b	$61,$BC,$BF,$AF,$78,$3F,$B1,$FB,$97,$E0,$FF,$CE,$2D,$18,$BF,$BC
	.dc.b	$7E,$45,$C3,$EC,$6B,$FE,$BC,$87,$DD,$7D,$08,$FC,$A3,$5F,$3B,$6F
	.dc.b	$F8,$13,$F4,$23,$D7,$F8,$3F,$F6,$77,$F3,$8B,$D9,$FE,$26,$D8,$CF
	.dc.b	$B5,$FD,$0B,$EE,$DF,$1F,$F5,$1B,$3F,$51,$AF,$DA,$6B,$E9,$FA,$D7
	.dc.b	$CA,$7F,$20,$6C,$7E,$B5,$AF,$D0,$7F,$92,$3F,$B1,$F9,$6F,$F6,$CE
	.dc.b	$9F,$C7,$57,$E0,$9F,$CA,$9B,$3B,$F8,$3F,$E5,$BD,$EF,$9C,$FA,$4D
	.dc.b	$9F,$A3,$C3,$FC,$46,$72,$7D,$7F,$BC,$FE,$E5,$3F,$C0,$5A,$FE,$11
	.dc.b	$B9,$7F,$2F,$7F,$B0,$F8,$BF,$C3,$FE,$F3,$F6,$6D,$F9,$4F,$E7,$F0
	.dc.b	$6F,$8F,$FD,$FB,$E9,$DF,$78,$F9,$6F,$EF,$F1,$86,$C6,$73,$FC,$4C
	.dc.b	$E5,$F5,$E3,$E0,$7F,$8E,$37,$5F,$F0,$8E,$1F,$D4,$23,$C5,$3F,$C7
	.dc.b	$74,$7D,$C7,$F5,$45,$F2,$14,$C4,$7F,$85,$2D,$F5,$6F,$BA,$FF,$0A
	.dc.b	$C6,$1B,$19,$C3,$F3,$F3,$96,$FD,$E7,$F8,$AD,$FD,$A4,$23,$18,$FF
	.dc.b	$85,$63,$1F,$13,$E1,$97,$C6,$FC,$9F,$94,$7C,$B7,$CB,$FB,$0F,$C4
	.dc.b	$E5,$FB,$FE,$14,$63,$3F,$F1,$04,$2F,$8D,$7F,$44,$87,$DE,$7F,$2F
	.dc.b	$81,$23,$3C,$3F,$5A,$75,$39,$6A,$9E,$7F,$43,$F7,$E4,$D3,$E6,$7D
	.dc.b	$97,$C0,$BE,$FB,$BE,$CD,$F3,$7F,$45,$B9,$7F,$5A,$9D,$4E,$7A,$64
	.dc.b	$FD,$0A,$FB,$EF,$F4,$DA,$E9,$3F,$62,$73,$3E,$E1,$2F,$BC,$FD,$44
	.dc.b	$7D,$65,$F4,$BF,$83,$6C,$78,$DD,$EF,$F6,$57,$D5,$6F,$BE,$56,$E3
	.dc.b	$F1,$9F,$68,$FA,$3B,$D7,$2B,$DD,$39,$6F,$7F,$0A,$91,$83,$FB,$67
	.dc.b	$C6,$7F,$97,$B8,$7D,$8B,$EE,$FF,$C8,$8D,$9D,$FD,$26,$08,$E0,$E7
	.dc.b	$EC,$70,$3F,$CE,$13,$FD,$66,$3F,$BC,$35,$EF,$1F,$E7,$5F,$BB,$F8
	.dc.b	$B1,$FD,$44,$3F,$1B,$0F,$A0,$D8,$FC,$7B,$C3,$83,$FC,$00,$D9,$DF
	.dc.b	$0F,$F9,$76,$31,$5C,$3E,$93,$FC,$C1,$F7,$7F,$C8,$70,$FF,$44,$B6
	.dc.b	$7F,$A8,$B5,$FF,$08,$35,$FC,$13,$F7,$BE,$EF,$8E,$06,$92,$7F,$D2
	.dc.b	$5F,$45,$FE,$93,$4D,$23,$F9,$1F,$CB,$FE,$96,$7F,$99,$6C,$7F,$75
	.dc.b	$F8,$67,$DA,$FF,$8A,$EB,$5C,$3F,$E9,$7D,$3F,$C8,$7E,$57,$F9,$3E
	.dc.b	$7F,$C7,$77,$FE,$3B,$F9,$3D,$47,$EA,$B2,$EB,$7C,$27,$39,$7F,$88
	.dc.b	$30,$FE,$3F,$7F,$DD,$30,$FF,$22,$79,$7F,$BC,$6C,$66,$77,$5F,$E3
	.dc.b	$79,$FF,$25,$74,$74,$6D,$79,$2B,$78,$FA,$2B,$5A,$7A,$70,$F8,$0D
	.dc.b	$9D,$5F,$B2,$30,$FF,$12,$D9,$F9,$87,$FA,$62,$3F,$CB,$71,$8F,$9E
	.dc.b	$F6,$F7,$3E,$F3,$FC,$27,$5A,$BE,$19,$7F,$30,$E7,$FF,$1C,$CF,$F0
	.dc.b	$B6,$C7,$F8,$9F,$1E,$93,$9F,$88,$F7,$7C,$DB,$A2,$9E,$0F,$BC,$FF
	.dc.b	$03,$D7,$D4,$9D,$3F,$C5,$FF,$0A,$7C,$5F,$99,$B1,$BB,$C4,$FA,$BC
	.dc.b	$5F,$E0,$93,$FC,$6F,$C0,$FE,$5F,$C1,$FC,$95,$B3,$AD,$7F,$8F,$7A
	.dc.b	$9F,$89,$7F,$A6,$B7,$93,$FA,$64,$ED,$FE,$9D,$19,$C3,$F9,$FF,$3B
	.dc.b	$F3,$1F,$F2,$6C,$7E,$39,$B8,$7F,$58,$F2,$9F,$7B,$F8,$19,$CE,$3F
	.dc.b	$D6,$7F,$2A,$FF,$0B,$E1,$FD,$89,$B8,$7F,$33,$6B,$FC,$C7,$54,$62
	.dc.b	$3F,$A1,$FE,$41,$FC,$AA,$BF,$A0,$36,$3F,$B4,$BF,$E1,$F8,$CF,$93
	.dc.b	$F8,$2F,$C0,$3F,$D4,$D1,$FD,$1D,$B5,$FD,$C1,$7C,$E6,$FE,$6B,$F4
	.dc.b	$B2,$FA,$BF,$F5,$4D,$FF,$96,$23,$1F,$3B,$FD,$7A,$73,$1F,$E5,$C6
	.dc.b	$BF,$E9,$BF,$CC,$BB,$ED,$FE,$99,$6D,$FF,$4F,$F4,$FF,$B6,$BF,$EA
	.dc.b	$38,$CD,$7F,$70,$7D,$8F,$F2,$96,$D7,$FA,$8E,$BF,$AB,$D7,$1F,$47
	.dc.b	$D3,$1F,$DA,$23,$D0,$DB,$FF,$38,$B5,$FF,$58,$35,$FE,$C5,$7F,$C2
	.dc.b	$26,$F9,$3D,$3E,$1B,$E2,$CE,$6B,$FD,$95,$E2,$FA,$3C,$7F,$E3,$3B
	.dc.b	$7F,$03,$FC,$CF,$CB,$5F,$E8,$CA,$D7,$C1,$FF,$9C,$73,$FF,$29,$B6
	.dc.b	$7F,$D3,$AD,$7F,$D0,$4D,$7F,$DB,$9F,$80,$7E,$5D,$FF,$6D,$37,$AD
	.dc.b	$F9,$A7,$F9,$3E,$FF,$DB,$53,$BC,$DF,$80,$BD,$4F,$E4,$71,$87,$F8
	.dc.b	$D7,$E3,$AF,$AA,$CD,$6F,$0F,$F7,$27,$D4,$FF,$95,$DF,$F3,$5B,$74
	.dc.b	$FF,$5A,$B5,$FF,$4D,$35,$FF,$7A,$72,$27,$C2,$D7,$F9,$DD,$BE,$CB
	.dc.b	$31,$E6,$F8,$BF,$F0,$8B,$FE,$F5,$6D,$FF,$49,$FC,$D7,$F8,$5B,$FE
	.dc.b	$BA,$8C,$F4,$F7,$8F,$F8,$01,$F8,$0B,$EE,$7F,$C7,$3C,$08,$C6,$3F
	.dc.b	$E8,$66,$D7,$FB,$41,$AF,$FA,$D7,$F0,$1F,$C1,$51,$FC,$27,$27,$F2
	.dc.b	$5F,$A4,$F9,$B1,$AF,$C0,$7F,$CD,$18,$FF,$93,$33,$F2,$60,$9C,$E3
	.dc.b	$FD,$8F,$F3,$6F,$E5,$6F,$FB,$52,$33,$5C,$3E,$53,$F4,$B3,$FE,$C7
	.dc.b	$6B,$FE,$58,$BF,$F5,$3F,$81,$EE,$9C,$CA,$7F,$DE,$7D,$BF,$82,$FA
	.dc.b	$3F,$D8,$EC,$FA,$E7,$CD,$66,$BF,$5B,$EF,$CD,$FF,$A8,$DF,$F2,$B7
	.dc.b	$F7,$91,$FC,$2B,$C5,$EF,$9F,$F0,$BD,$7F,$84,$12,$BF,$B8,$7C,$AF
	.dc.b	$F5,$0D,$FE,$B6,$28,$C4,$FF,$95,$70,$FE,$02,$FF,$36,$8C,$E1,$EB
	.dc.b	$F1,$5F,$63,$F9,$C6,$D7,$CF,$3E,$22,$FB,$9D,$4F,$E8,$BF,$11,$FE
	.dc.b	$EF,$5F,$25,$B7,$D6,$75,$F0,$7F,$A8,$FC,$0B,$FB,$AC,$FC,$B6,$F2
	.dc.b	$1F,$95,$87,$86,$75,$C1,$FB,$0F,$C4,$9F,$71,$4E,$91,$6B,$FC,$3B
	.dc.b	$B7,$EC,$DB,$E0,$FF,$83,$E3,$F3,$FC,$5F,$21,$D6,$3F,$E2,$B5,$F7
	.dc.b	$7B,$38,$55,$BF,$F0,$77,$CA,$FF,$85,$0C,$5F,$F8,$51,$FE,$0B,$F0
	.dc.b	$1C,$A7,$D5,$B2,$31,$4C,$BF,$37,$0F,$9A,$E3,$97,$F0,$EF,$C4,$FF
	.dc.b	$17,$8F,$26,$E1,$C6,$3F,$31,$7D,$30,$FD,$2D,$69,$FD,$3F,$E0,$6F
	.dc.b	$AA,$FE,$32,$DB,$F2,$AF,$1B,$33,$FC,$7B,$E4,$3F,$D8,$F5,$EA,$6E
	.dc.b	$5E,$13,$F4,$97,$CD,$C6,$A9,$DE,$3F,$D4,$7B,$FF,$A7,$9F,$DA,$3F
	.dc.b	$04,$FE,$9D,$19,$C7,$F0,$18,$7F,$AE,$BE,$97,$FD,$8B,$E6,$7E,$F2
	.dc.b	$3F,$A8,$C6,$0F,$F6,$4D,$BE,$CF,$A1,$FD,$72,$2D,$FD,$2F,$C3,$FF
	.dc.b	$BC,$B7,$DB,$FE,$77,$1F,$C5,$0C,$E1,$EA,$FC,$DD,$F7,$5F,$D0,$1B
	.dc.b	$7C,$4F,$BA,$E2,$7F,$CE,$9F,$85,$7F,$87,$27,$FC,$4A,$D8,$78,$A8
	.dc.b	$7D,$9F,$F9,$8D,$D4,$67,$93,$06,$7E,$CE,$E1,$FA,$5B,$7E,$0E,$F4
	.dc.b	$7F,$CE,$F1,$87,$A7,$95,$89,$D5,$71,$FA,$5F,$F1,$9B,$FE,$DB,$D3
	.dc.b	$FD,$F3,$82,$73,$19,$8C,$7C,$4F,$FB,$E9,$B5,$FE,$83,$D7,$FA,$4B
	.dc.b	$F0,$3E,$6F,$C7,$87,$2E,$7F,$E2,$F5,$AC,$7F,$AE,$7E,$71,$FC,$01
	.dc.b	$FC,$66,$08,$C5,$7F,$40,$FC,$6F,$E5,$4E,$7A,$BF,$B9,$FD,$C7,$F8
	.dc.b	$22,$BF,$A2,$B6,$7F,$C4,$F7,$C9,$FE,$7D,$19,$C3,$F9,$E7,$C6,$FF
	.dc.b	$6B,$7F,$32,$DB,$FE,$9E,$BE,$97,$FC,$02,$E6,$FF,$A4,$7E,$73,$FA
	.dc.b	$1A,$FE,$AA,$D8,$FE,$99,$3B,$BF,$E9,$53,$96,$FD,$17,$F6,$C8,$34
	.dc.b	$95,$F5,$67,$FA,$97,$E7,$5F,$CA,$0F,$E5,$35,$C2,$7F,$87,$C7,$F2
	.dc.b	$CE,$E7,$C0,$FF,$86,$1B,$3F,$C5,$1F,$F0,$B5,$FE,$E1,$CB,$7E,$7F
	.dc.b	$81,$C6,$AD,$7F,$1C,$FC,$EF,$FC,$B3,$5F,$DA,$5A,$FF,$67,$6F,$8A
	.dc.b	$69,$FA,$ED,$7F,$BD,$B6,$B9,$7C,$77,$F8,$BA,$31,$7F,$DC,$7F,$04
	.dc.b	$FF,$1B,$3F,$97,$C1,$18,$81,$FD,$ED,$C3,$7E,$A7,$FC,$33,$1F,$D4
	.dc.b	$08,$FF,$8F,$1C,$7D,$1D,$31,$EB,$DC,$BF,$9D,$CF,$F1,$C7,$B7,$4E
	.dc.b	$70,$FE,$43,$39,$7F,$C1,$1F,$9E,$7E,$26,$BE,$D6,$48,$C4,$FA,$DA
	.dc.b	$FA,$8E,$3C,$BF,$E3,$08,$FA,$FB,$7F,$88,$9B,$7F,$E1,$97,$9C,$67
	.dc.b	$E2,$FF,$D2,$BF,$83,$FF,$96,$7E,$63,$E2,$FD,$3F,$F3,$FC,$BE,$3C
	.dc.b	$E5,$46,$54,$7E,$12,$30,$ED,$B4,$AE,$BE,$8F,$A6,$B4,$DE,$4F,$F2
	.dc.b	$65,$F0,$AD,$E1,$E2,$C3,$93,$E2,$BE,$51,$8A,$F0,$DF,$29,$CD,$F2
	.dc.b	$BF,$3D,$EE,$73,$1C,$CF,$7D,$71,$7C,$15,$C7,$0F,$3C,$60,$E0,$7C
	.dc.b	$29,$E0,$F6,$9E,$17,$DE,$38,$61,$E6,$CF,$35,$27,$9E,$1D,$9E,$93
	.dc.b	$9A,$F8,$91,$AB,$E8,$E8,$C5,$EF,$1F,$6D,$77,$9C,$91,$E6,$64,$B1
	.dc.b	$C0,$91,$F1,$2F,$A3,$EA,$8C,$BE,$19,$E2,$FC,$53,$FA,$46,$1F,$89
	.dc.b	$6E,$1F,$69,$B8,$7F,$B8,$E7,$FC,$9F,$E0,$DF,$77,$FB,$0D,$7F,$1D
	.dc.b	$9C,$B6,$32,$F6,$77,$87,$E4,$9B,$8F,$E3,$DB,$3F,$9C,$F7,$FF,$55
	.dc.b	$5B,$C3,$EF,$56,$AB,$F3,$AD,$C3,$FC,$BD,$1A,$98,$FD,$89,$D5,$F4
	.dc.b	$C3,$C1,$5A,$C3,$B5,$FD,$EA,$D3,$FA,$EB,$EC,$4A,$FC,$84,$EA,$7C
	.dc.b	$71,$CE,$FA,$C7,$33,$07,$38,$AF,$ED,$A7,$96,$1F,$2B,$0F,$19,$2F
	.dc.b	$F6,$93,$AC,$3C,$57,$ED,$C3,$C6,$F4,$C3,$95,$F8,$B0,$FE,$03,$87
	.dc.b	$1B,$D6,$1F,$13,$0B,$1C,$67,$E0,$E2,$FA,$77,$8F,$FC,$77,$7F,$F6
	.dc.b	$74,$E6,$31,$B7,$F3,$A8,$C6,$3F,$E8,$37,$97,$8B,$A8,$FE,$21,$82
	.dc.b	$3F,$B3,$46,$1B,$3C,$23,$FB,$53,$5F,$E9,$5C,$9F,$E2,$97,$F9,$EF
	.dc.b	$91,$D7,$97,$FA,$6B,$65,$F3,$D4,$95,$FD,$11,$D7,$D4,$EF,$F0,$69
	.dc.b	$3F,$D2,$1E,$58,$C8,$FE,$87,$7F,$A7,$7F,$A0,$63,$FD,$72,$7D,$6D
	.dc.b	$8F,$E9,$3D,$AF,$86,$F3,$36,$3F,$BA,$46,$3E,$26,$47,$07,$9C,$9A
	.dc.b	$76,$F0,$9F,$84,$7C,$16,$D7,$F8,$1E,$31,$DB,$23,$FC,$14,$D7,$FC
	.dc.b	$15,$C5,$F2,$7C,$46,$D4,$93,$FE,$A4,$F1,$5F,$35,$F3,$DB,$1F,$D9
	.dc.b	$9A,$FF,$7F,$6B,$F1,$9A,$FF,$8C,$BD,$0E,$44,$7F,$D6,$3E,$57,$F9
	.dc.b	$6D,$BE,$1F,$DD,$8C,$47,$FA,$61,$BD,$5F,$E4,$AE,$7F,$C1,$79,$BF
	.dc.b	$EF,$CA,$FF,$5D,$B5,$FF,$4F,$B6,$5A,$FF,$84,$87,$E7,$AF,$C2,$BF
	.dc.b	$31,$18,$6C,$66,$FF,$DC,$13,$97,$BF,$7F,$E7,$FD,$4F,$EE,$BE,$B5
	.dc.b	$F0,$FF,$D2,$B5,$FE,$E7,$9C,$BF,$E8,$5F,$CF,$9E,$77,$E9,$F0,$46
	.dc.b	$27,$FC,$99,$18,$9F,$D4,$B6,$33,$3B,$7F,$D0,$5F,$43,$F7,$1F,$F4
	.dc.b	$93,$5F,$F6,$24,$A3,$1A,$FC,$49,$FF,$50,$E3,$FA,$F7,$FD,$95,$8B
	.dc.b	$F1,$CF,$E6,$31,$86,$D7,$19,$D7,$8D,$F0,$E5,$E3,$D2,$5F,$C3,$8C
	.dc.b	$B7,$8F,$F7,$9E,$CB,$F1,$7A,$49,$FF,$63,$7D,$47,$F0,$1F,$70,$D2
	.dc.b	$47,$FB,$2F,$C9,$FE,$24,$BE,$B7,$FD,$0C,$FF,$B6,$9A,$FF,$B9,$CD
	.dc.b	$78,$1C,$3C,$E7,$E0,$54,$FE,$33,$3F,$E3,$06,$C6,$A3,$FB,$54,$63
	.dc.b	$EC,$F9,$12,$3F,$DD,$3F,$86,$BE,$6F,$F8,$D7,$C3,$3D,$7A,$4A,$FF
	.dc.b	$33,$F3,$7D,$5F,$35,$F8,$4B,$93,$F9,$13,$6B,$7D,$FF,$E6,$EA,$DD
	.dc.b	$68,$DF,$47,$F2,$9F,$89,$FE,$29,$6C,$7B,$46,$1B,$E7,$FF,$2F,$F8
	.dc.b	$AF,$8F,$1E,$36,$E5,$DD,$B6,$C9,$18,$6C,$FE,$9F,$EF,$5F,$86,$52
	.dc.b	$5F,$E1,$78,$BF,$14,$47,$07,$E2,$78,$FF,$1C,$47,$07,$E3,$7E,$17
	.dc.b	$FA,$37,$CA,$7F,$2B,$8F,$8A,$79,$5C,$3F,$07,$1F,$91,$18,$D3,$E8
	.dc.b	$7C,$0F,$D7,$AE,$3A,$23,$11,$C4,$F8,$0E,$27,$8F,$2F,$F5,$78,$7C
	.dc.b	$A7,$F5,$9F,$96,$7C,$3D,$26,$1D,$BF,$0C,$7C,$C7,$F5,$F6,$C6,$63
	.dc.b	$06,$7E,$FF,$F6,$1F,$19,$27,$F5,$CF,$E8,$BF,$0B,$7C,$4D,$3F,$36
	.dc.b	$E2,$7F,$BA,$B5,$FF,$6E,$35,$FF,$25,$F1,$7E,$94,$8F,$CC,$FB,$67
	.dc.b	$E7,$FF,$72,$3E,$3F,$E1,$7F,$95,$7E,$37,$1A,$31,$C7,$F9,$D6,$F2
	.dc.b	$FF,$B1,$BC,$DF,$8B,$E3,$7F,$7F,$FD,$01,$BF,$F7,$06,$E5,$FE,$B8
	.dc.b	$6B,$FE,$D1,$6B,$FD,$E3,$F0,$CF,$F2,$BE,$5F,$AB,$6E,$3F,$A7,$6C
	.dc.b	$B1,$7D,$23,$E7,$FF,$C2,$1F,$86,$3E,$E2,$92,$FF,$9F,$79,$DF,$D0
	.dc.b	$1F,$7F,$E1,$9F,$CD,$5F,$F0,$4B,$5F,$DB,$35,$FD,$43,$78,$5F,$86
	.dc.b	$7F,$2E,$D7,$F7,$8D,$AF,$D7,$37,$47,$D5,$FC,$A9,$FD,$97,$E1,$7F
	.dc.b	$C8,$5F,$CC,$B5,$FA,$6D,$E0,$FE,$75,$F4,$8F,$C1,$BF,$C5,$DF,$2D
	.dc.b	$FC,$42,$FF,$DC,$AD,$7F,$DF,$CD,$7F,$C4,$AD,$7F,$89,$7E,$14,$EE
	.dc.b	$FF,$1C,$36,$3F,$56,$D7,$FA,$7F,$D8,$BF,$00,$FF,$1F,$7E,$14,$FA
	.dc.b	$1A,$4A,$FE,$D9,$F6,$BF,$E9,$BD,$7F,$93,$7D,$B3,$EE,$FF,$C6,$AD
	.dc.b	$AF,$F7,$EB,$5F,$F7,$D1,$FC,$5D,$C3,$FE,$3E,$F4,$7E,$13,$FC,$01
	.dc.b	$AF,$E6,$9B,$5F,$D8,$7F,$41,$7E,$1B,$FD,$43,$EA,$7E,$34,$FF,$6D
	.dc.b	$B4,$63,$A1,$12,$30,$E3,$F3,$1F,$F7,$F5,$71,$7F,$88,$47,$F9,$A9
	.dc.b	$AF,$A8,$38,$7D,$AD,$7B,$8F,$C4,$5F,$57,$FD,$99,$B1,$FC,$98,$D4
	.dc.b	$7F,$6A,$9F,$91,$5E,$BA,$7C,$FF,$12,$5F,$1A,$FF,$45,$BC,$FC,$CF
	.dc.b	$E3,$2F,$F0,$CF,$3B,$F8,$89,$1F,$F4,$0F,$E1,$AF,$AF,$FF,$45,$7D
	.dc.b	$02,$31,$3F,$E8,$F6,$BF,$CC,$63,$5D,$BC,$3F,$01,$FD,$DC,$1A,$63
	.dc.b	$87,$F8,$3E,$DF,$E9,$EF,$8F,$7C,$CF,$FA,$51,$B8,$FF,$A5,$1A,$FE
	.dc.b	$23,$0F,$F1,$54,$FF,$A7,$3C,$3F,$EE,$9E,$77,$F9,$0F,$E0,$FA,$F3
	.dc.b	$7F,$1B,$6E,$0C,$1F,$81,$7A,$B3,$FF,$57,$FE,$85,$F8,$AA,$FF,$56
	.dc.b	$35,$F7,$36,$3F,$D6,$BF,$82,$FE,$87,$1F,$F5,$A3,$FE,$97,$AB,$87
	.dc.b	$FA,$BA,$31,$1E,$16,$C3,$E5,$FF,$8C,$5F,$FB,$01,$B7,$FE,$BF,$6D
	.dc.b	$FF,$25,$F8,$4F,$BA,$FF,$91,$9B,$5F,$E4,$36,$D7,$F3,$B6,$CF,$C1
	.dc.b	$F8,$37,$C7,$FD,$EE,$DF,$F2,$04,$FF,$B2,$DA,$FD,$D6,$E1,$FE,$57
	.dc.b	$F6,$9F,$97,$FF,$77,$CB,$F5,$71,$8F,$A4,$FF,$69,$37,$0F,$EA,$D5
	.dc.b	$B7,$E9,$7E,$19,$FA,$07,$FC,$BF,$E1,$FC,$DC,$3F,$A9,$36,$3F,$A2
	.dc.b	$37,$97,$ED,$7C,$2F,$FA,$07,$87,$FC,$01,$7F,$D6,$5B,$7F,$D2,$DB
	.dc.b	$5F,$81,$F8,$8E,$07,$E2,$63,$FB,$44,$FF,$14,$9D,$3F,$BD,$6B,$98
	.dc.b	$FF,$4E,$57,$7F,$BB,$F1,$1F,$A8,$FA,$A7,$DA,$52,$5F,$FB,$DE,$32
	.dc.b	$FF,$A5,$7F,$03,$FC,$53,$FC,$53,$E4,$9F,$5B,$FE,$45,$8F,$E1,$D1
	.dc.b	$8E,$3F,$F7,$DB,$63,$33,$B7,$F9,$6C,$7F,$C0,$1E,$2D,$FC,$B9,$A4
	.dc.b	$9F,$F3,$6F,$CE,$7F,$9B,$93,$48,$FC,$5C,$5F,$89,$FF,$2D,$8F,$F0
	.dc.b	$44,$63,$6F,$90,$D8,$E1,$3B,$7F,$65,$A7,$FB,$7A,$BF,$95,$6E,$8C
	.dc.b	$4F,$EF,$DA,$FF,$2A,$8D,$7D,$3F,$3E,$1F,$F3,$07,$D7,$5F,$7C,$D2
	.dc.b	$63,$F6,$E3,$FD,$00,$DF,$98,$E0,$47,$5F,$65,$CC,$8E,$0F,$3F,$E1
	.dc.b	$9F,$E9,$66,$FC,$B7,$B7,$34,$E7,$4F,$E3,$0D,$7F,$A1,$F0,$BF,$0C
	.dc.b	$FF,$69,$E1,$FD,$2E,$31,$F4,$BF,$1D,$B1,$99,$DD,$7F,$23,$FC,$33
	.dc.b	$FA,$5C,$FC,$A6,$C6,$7D,$7F,$24,$62,$3F,$A9,$F0,$7F,$A1,$BE,$23
	.dc.b	$E5,$78,$DF,$EB,$CC,$3F,$40,$DA,$FD,$23,$5E,$3F,$A1,$BA,$7E,$93
	.dc.b	$6B,$F5,$0D,$9F,$D6,$35,$FB,$5E,$4F,$D6,$9F,$C0,$B6,$3F,$9F,$35
	.dc.b	$DF,$2F,$CD,$F2,$DF,$27,$49,$7F,$8A,$8C,$BF,$B9,$FB,$8F,$E9,$AD
	.dc.b	$F4,$3F,$11,$18,$C3,$F8,$03,$5F,$C7,$7A,$7F,$B2,$AF,$CA,$B6,$3F
	.dc.b	$A9,$35,$DF,$83,$F8,$1F,$C2,$FF,$80,$D7,$E7,$1B,$3F,$D6,$5A,$FE
	.dc.b	$6D,$AF,$F8,$1B,$F1,$AF,$F0,$DC,$FF,$83,$61,$18,$7F,$C2,$0D,$7F
	.dc.b	$C5,$6D,$7F,$C1,$FF,$7B,$FE,$1E,$9E,$4D,$8F,$F0,$C3,$5F,$F1,$9B
	.dc.b	$5E,$3D,$5F,$B7,$51,$FD,$A6,$79,$CE,$63,$13,$D3,$E6,$30,$4A,$F0
	.dc.b	$BB,$F6,$E4,$79,$1B,$3F,$CC,$7F,$0B,$E6,$FE,$59,$AF,$FA,$55,$AF
	.dc.b	$FA,$1D,$BC,$E7,$FD,$5F,$5F,$CE,$5B,$3D,$6F,$F8,$44,$62,$3F,$47
	.dc.b	$F0,$95,$FF,$51,$35,$FF,$4A,$36,$CD,$F8,$87,$B9,$FC,$4D,$B8,$FF
	.dc.b	$49,$6D,$FF,$4D,$6B,$FB,$27,$FC,$61,$1B,$F7,$B9,$11,$FE,$A5,$4F
	.dc.b	$C2,$7F,$90,$36,$33,$18,$3F,$92,$7C,$7F,$FA,$9E,$3F,$AD,$37,$0F
	.dc.b	$EB,$CD,$7F,$81,$B5,$FF,$61,$3E,$66,$D7,$42,$4F,$F6,$5E,$D7,$D8
	.dc.b	$FF,$33,$6C,$66,$30,$7F,$37,$F9,$6F,$E2,$57,$FE,$CF,$6B,$FE,$C1
	.dc.b	$6E,$BF,$EC,$2F,$07,$87,$C2,$CC,$3F,$BA,$57,$F9,$89,$B1,$FD,$0F
	.dc.b	$E7,$5E,$A5,$24,$FF,$7B,$8C,$BF,$CB,$F0,$FE,$1D,$8F,$F8,$69,$AF
	.dc.b	$F6,$76,$F8,$DF,$ED,$5F,$03,$F4,$1F,$2F,$2F,$F5,$AD,$BF,$F4,$27
	.dc.b	$E0,$BF,$C1,$5F,$F4,$0B,$5F,$F4,$8B,$53,$9F,$39,$D2,$FA,$5F,$ED
	.dc.b	$11,$86,$C6,$70,$FF,$49,$4E,$5B,$0F,$AD,$D9,$23,$2F,$4F,$C1,$5B
	.dc.b	$52,$4F,$48,$CB,$73,$E0,$47,$07,$F8,$9F,$F0,$5F,$C5,$69,$FE,$FB
	.dc.b	$6E,$E6,$93,$9B,$FB,$D3,$F8,$F9,$E1,$18,$33,$D3,$2F,$B3,$67,$FD
	.dc.b	$97,$AB,$CB,$7E,$09,$F2,$1B,$7F,$19,$B5,$C1,$AF,$FA,$DA,$B9,$5F
	.dc.b	$12,$57,$17,$E9,$FE,$08,$FB,$FA,$49,$FA,$71,$96,$FC,$A7,$02,$38
	.dc.b	$3F,$9F,$7C,$2F,$F2,$E6,$F8,$9F,$CC,$63,$19,$7F,$B3,$DA,$FF,$38
	.dc.b	$E2,$FE,$64,$4A,$FE,$7E,$FF,$35,$EC,$91,$FD,$09,$B3,$F2,$3E,$1B
	.dc.b	$FC,$5A,$FF,$B9,$35,$FF,$42,$FB,$3F,$90,$9C,$CA,$73,$7F,$90,$9F
	.dc.b	$EB,$33,$96,$CF,$FA,$3B,$E0,$39,$1F,$E3,$7F,$C1,$1F,$69,$49,$87
	.dc.b	$F9,$BE,$32,$DF,$43,$81,$1C,$1B,$F3,$BA,$11,$C1,$FE,$F5,$A7,$C6
	.dc.b	$FF,$14,$CF,$FD,$E6,$78,$9C,$3F,$EF,$68,$DF,$8B,$FE,$57,$6D,$73
	.dc.b	$FB,$9F,$F0,$6D,$72,$94,$62,$39,$35,$1F,$01,$C3,$F2,$FE,$E7,$FC
	.dc.b	$20,$FD,$16,$B9,$CB,$F8,$0C,$62,$DF,$82,$7E,$D2,$BD,$71,$8E,$2F
	.dc.b	$F3,$73,$4F,$7C,$1C,$3F,$A8,$F5,$1F,$69,$FC,$39,$B3,$FC,$31,$B3
	.dc.b	$B9,$D4,$C3,$45,$8F,$D9,$FD,$DB,$EA,$7F,$8C,$B6,$7F,$8B,$B6,$77
	.dc.b	$3A,$FC,$12,$56,$9B,$3F,$BE,$FC,$01,$F8,$1F,$F2,$46,$D7,$F2,$26
	.dc.b	$CE,$E7,$5F,$49,$2B,$4D,$9F,$E1,$3F,$81,$7F,$B9,$27,$F9,$5B,$63
	.dc.b	$3D,$3F,$CC,$A3,$11,$FC,$1D,$7D,$9E,$49,$5C,$9F,$E7,$3F,$81,$FE
	.dc.b	$E2,$BF,$5C,$D7,$F5,$8D,$EB,$34,$FF,$09,$38,$F6,$B9,$92,$FF,$A0
	.dc.b	$CF,$FA,$72,$BF,$91,$B5,$FE,$38,$D7,$C3,$F1,$5E,$4F,$8A,$95,$C9
	.dc.b	$B3,$FC,$B3,$F0,$46,$BF,$D4,$9B,$5F,$D4,$1B,$3B,$9D,$71,$CA,$D3
	.dc.b	$67,$F2,$7F,$77,$F6,$BD,$FF,$D2,$FC,$87,$F2,$B9,$FD,$3C,$E5,$FF
	.dc.b	$69,$F9,$3F,$AA,$51,$E0,$9E,$11,$8E,$35,$F4,$BE,$17,$CE,$F6,$D8
	.dc.b	$CE,$5F,$A5,$C3,$B9,$E2,$C3,$9B,$83,$B6,$1D,$5D,$35,$F2,$CF,$3C
	.dc.b	$3B,$D7,$26,$D6,$F0,$F1,$5F,$C0,$BE,$18,$7A,$30,$E0,$DB,$E1,$18
	.dc.b	$AE,$19,$62,$64,$79,$F0,$EA,$F9,$27,$38,$F8,$A3,$57,$CD,$D1,$8B
	.dc.b	$DE,$1E,$18,$E9,$8F,$99,$F8,$26,$C9,$7F,$3A,$35,$1F,$20,$CC,$F3
	.dc.b	$9D,$61,$DA,$79,$C7,$94,$CB,$C0,$CC,$F1,$3E,$CC,$F1,$8E,$C7,$F3
	.dc.b	$EC,$7F,$14,$DC,$7E,$DB,$71,$FF,$57,$4F,$F8,$9F,$BA,$FA,$5F,$D5
	.dc.b	$AB,$F2,$ED,$7A,$35,$FC,$6B,$BC,$BF,$1A,$DC,$B3,$5F,$7F,$D4,$FC
	.dc.b	$D5,$6A,$3F,$05,$5B,$BF,$D0,$B5,$FC,$FD,$70,$BE,$71,$F7,$67,$58
	.dc.b	$7A,$63,$F0,$15,$B7,$9D,$F6,$23,$A2,$E1,$E3,$7A,$47,$9D,$CE,$1E
	.dc.b	$09,$D1,$FA,$AA,$ED,$3F,$B4,$C3,$C6,$4B,$E5,$1D,$09,$5C,$AF,$F6
	.dc.b	$B7,$CB,$0F,$54,$E6,$BD,$27,$9C,$E3,$5E,$32,$3F,$C2,$2B,$96,$1E
	.dc.b	$18,$D9,$93,$D6,$74,$AE,$CF,$CE,$3F,$79,$8B,$E3,$9F,$13,$FE,$F3
	.dc.b	$8C,$36,$33,$3F,$DB,$A7,$2D,$87,$D2,$EC,$91,$97,$FD,$FD,$CD,$FD
	.dc.b	$03,$D9,$BC,$C6,$92,$BF,$9F,$FB,$9A,$23,$83,$FA,$1E,$3C,$88,$E0
	.dc.b	$FE,$8E,$76,$21,$83,$FB,$0F,$CC,$F9,$3A,$CD,$24,$7F,$66,$E0,$D1
	.dc.b	$1C,$1F,$DC,$F9,$BD,$BD,$BF,$E0,$E8,$C1,$0F,$EF,$9C,$DA,$23,$FD
	.dc.b	$ED,$F8,$BE,$03,$CC,$A4,$9F,$F0,$5B,$AF,$32,$47,$F8,$39,$AF,$FA
	.dc.b	$1B,$E7,$7F,$C6,$EF,$FA,$A1,$AE,$7C,$9F,$A7,$18,$8F,$F4,$23,$FE
	.dc.b	$DE,$78,$79,$5A,$23,$FE,$2E,$C5,$F8,$03,$F0,$1F,$F1,$8C,$FF,$8C
	.dc.b	$63,$0E,$DA,$FF,$8E,$9A,$FF,$A4,$FC,$BF,$E1,$CF,$F9,$17,$BB,$F0
	.dc.b	$F1,$88,$FF,$25,$47,$61,$F8,$0B,$E4,$B0,$D9,$FF,$32,$35,$FF,$2D
	.dc.b	$35,$FF,$31,$7D,$0B,$F0,$2E,$8D,$8F,$F3,$63,$5F,$F3,$13,$5F,$EE
	.dc.b	$DE,$F3,$C0,$F9,$11,$FE,$70,$8C,$6D,$F1,$1B,$1F,$E5,$66,$BB,$FC
	.dc.b	$0B,$FD,$B4,$E5,$AE,$7C,$2E,$51,$88,$F0,$AE,$1C,$92,$39,$3F,$B0
	.dc.b	$F3,$1F,$0F,$49,$3F,$E9,$D7,$5E,$24,$8F,$F5,$03,$5F,$F0,$DF,$CB
	.dc.b	$FD,$E7,$FC,$2A,$DF,$1C,$D3,$1F,$43,$27,$FA,$8B,$5F,$F3,$BF,$E0
	.dc.b	$0F,$9F,$E8,$D8,$FC,$5B,$5D,$C7,$F0,$AF,$C0,$3E,$04,$7F,$A0,$3C
	.dc.b	$84,$62,$3F,$DC,$ED,$7F,$DC,$0D,$7F,$D8,$4B,$CB,$E8,$48,$E9,$19
	.dc.b	$7F,$DD,$9E,$17,$F1,$AF,$82,$FD,$F4,$7F,$BB,$63,$06,$7B,$A4,$7F
	.dc.b	$BC,$DA,$FD,$FC,$FE,$D9,$1F,$C0,$3F,$7B,$49,$1F,$83,$6B,$F8,$7F
	.dc.b	$96,$FC,$19,$1F,$C4,$3F,$CB,$BD,$4F,$91,$1F,$EF,$F6,$BF,$EE,$16
	.dc.b	$BF,$ED,$F6,$E9,$FE,$7E,$FB,$DE,$44,$9F,$9D,$28,$FE,$77,$F7,$8F
	.dc.b	$8B,$A4,$AF,$A9,$19,$6F,$D9,$70,$23,$83,$F8,$4F,$D5,$BD,$9F,$C7
	.dc.b	$36,$33,$18,$3F,$21,$9C,$8F,$F0,$EB,$5F,$F7,$0F,$C7,$BE,$F1,$F4
	.dc.b	$BF,$C5,$A3,$0D,$8C,$CF,$FB,$2A,$72,$FE,$DF,$E0,$3F,$D7,$CD,$F7
	.dc.b	$FF,$79,$18,$8F,$E0,$0D,$7F,$82,$FC,$9A,$3F,$81,$36,$3F,$81,$35
	.dc.b	$FE,$10,$D7,$8F,$C8,$70,$23,$C1,$FE,$19,$DC,$FA,$9A,$49,$FE,$1D
	.dc.b	$19,$7F,$88,$7C,$AB,$EC,$BF,$8A,$36,$33,$18,$33,$E1,$C8,$E2,$D7
	.dc.b	$F8,$F7,$C9,$3E,$D3,$F8,$F3,$63,$F8,$F3,$5D,$C7,$F2,$3F,$99,$7D
	.dc.b	$C5,$24,$FF,$25,$8C,$B7,$0E,$04,$70,$7F,$2B,$F9,$BF,$F2,$FB,$FC
	.dc.b	$D1,$A8,$79,$68,$CC,$61,$B0,$F9,$CF,$E3,$CD,$9F,$E7,$51,$83,$3F
	.dc.b	$64,$FA,$FF,$E7,$CD,$9F,$E7,$CD,$7F,$A2,$35,$FF,$5F,$0F,$B1,$FF
	.dc.b	$5E,$B8,$6B,$FE,$C3,$6B,$EB,$6B,$C7,$CD,$FE,$69,$06,$92,$3F,$BA
	.dc.b	$C7,$F8,$D9,$BF,$0D,$C0,$8E,$BE,$4B,$99,$1C,$1C,$BC,$FE,$E4,$70
	.dc.b	$7F,$91,$7E,$C5,$C5,$FE,$4B,$6C,$66,$30,$7F,$93,$F3,$7C,$9D,$24
	.dc.b	$FF,$93,$E3,$2F,$F6,$9E,$4A,$FF,$6F,$6B,$FD,$D1,$AF,$E0,$77,$FE
	.dc.b	$CD,$5F,$E5,$DF,$A4,$7C,$DD,$25,$7F,$97,$63,$2F,$F9,$7F,$E9,$7F
	.dc.b	$0F,$F1,$26,$91,$FE,$FD,$F4,$DF,$36,$7F,$C0,$4D,$7F,$C0,$ED,$7C
	.dc.b	$8D,$7F,$2B,$E8,$B8,$3F,$C1,$0D,$8F,$F0,$43,$5F,$F0,$73,$5F,$CF
	.dc.b	$0A,$79,$FF,$83,$8F,$F3,$C3,$5F,$C8,$C6,$B9,$3F,$D0,$13,$99,$E7
	.dc.b	$E1,$BE,$C3,$FC,$CB,$1F,$C9,$5A,$FF,$2E,$6B,$C1,$AF,$EC,$6D,$F6
	.dc.b	$0F,$9B,$FF,$47,$36,$7F,$D1,$8D,$7C,$AD,$7F,$C5,$F3,$FD,$63,$87
	.dc.b	$DD,$EB,$FF,$8B,$63,$04,$F2,$70,$4A,$FF,$18,$47,$F3,$E7,$FC,$6C
	.dc.b	$BE,$FF,$91,$23,$6F,$FA,$87,$E2,$24,$7F,$A7,$DA,$FF,$8F,$FE,$BB
	.dc.b	$F8,$49,$2B,$FC,$80,$FF,$0E,$87,$D7,$FD,$DF,$B0,$FF,$33,$B6,$3F
	.dc.b	$81,$9A,$E3,$4E,$9B,$1A,$9F,$F5,$BF,$07,$2E,$D3,$49,$3F,$E3,$7F
	.dc.b	$82,$FF,$1C,$3F,$E5,$73,$04,$3F,$13,$97,$FB,$0F,$E6,$DE,$DF,$F2
	.dc.b	$9C,$ED,$7E,$37,$E5,$95,$89,$CF,$17,$F6,$FF,$3B,$18,$7E,$06,$F6
	.dc.b	$F3,$3F,$69,$9B,$E6,$B7,$97,$F3,$68,$C4,$EE,$3C,$19,$6B,$1F,$7D
	.dc.b	$DC,$BC,$F7,$FE,$6D,$6D,$FF,$9C,$DA,$FF,$9E,$1A,$F8,$BE,$3B,$A6
	.dc.b	$27,$A5,$61,$F0,$7A,$9F,$99,$8F,$26,$3F,$E0,$DE,$8D,$46,$27,$58
	.dc.b	$7D,$A8,$FD,$17,$D7,$FE,$D1,$F6,$6E,$8C,$4E,$CF,$DD,$2F,$5B,$E2
	.dc.b	$CE,$FB,$3E,$17,$D3,$FF,$AC,$70,$F1,$B7,$0E,$55,$B9,$CF,$C8,$72
	.dc.b	$FA,$E7,$CC,$53,$1F,$33,$6B,$DB,$E1,$FF,$9D,$5B,$E0,$67,$2F,$37
	.dc.b	$0A,$B5,$18,$F1,$3F,$C8,$D7,$FE,$9F,$6C,$7F,$24,$F8,$1F,$C5,$BF
	.dc.b	$C9,$1A,$FF,$07,$FC,$37,$FD,$69,$5F,$19,$B1,$D6,$7E,$1C,$63,$9F
	.dc.b	$F2,$BB,$BC,$8F,$E9,$78,$7C,$EF,$D1,$1F,$27,$F2,$A7,$39,$7C,$69
	.dc.b	$CD,$7D,$CF,$A1,$7C,$D7,$92,$BF,$C1,$31,$89,$FF,$5D,$36,$3F,$D8
	.dc.b	$0D,$7F,$57,$F6,$0F,$0F,$EC,$B6,$BF,$02,$FF,$AD,$1B,$1F,$9D,$FA
	.dc.b	$17,$D4,$7E,$03,$EC,$1F,$39,$98,$C3,$BC,$FC,$DC,$AB,$F1,$67,$FD
	.dc.b	$85,$AF,$F8,$89,$B0,$FD,$17,$E5,$67,$FD,$9D,$B4,$F8,$EF,$93,$C3
	.dc.b	$FB,$19,$CD,$BF,$25,$7C,$67,$EE,$AD,$F8,$FB,$E9,$7F,$71,$5A,$CF
	.dc.b	$F7,$95,$A9,$FD,$E7,$DA,$BE,$CB,$D6,$DA,$FF,$74,$37,$5F,$F7,$73
	.dc.b	$5F,$F7,$9F,$DA,$FE,$3B,$AB,$F8,$9B,$6B,$FC,$85,$8B,$EB,$7F,$D8
	.dc.b	$11,$FB,$A9,$D5,$FF,$91,$BE,$15,$F5,$FE,$BF,$59,$EB,$FF,$11,$9C
	.dc.b	$E1,$EA,$9C,$BF,$C1,$3E,$1F,$38,$7F,$8F,$B8,$FF,$CA,$18,$FD,$69
	.dc.b	$CC,$FF,$95,$E3,$EC,$FB,$8D,$FF,$8F,$37,$0F,$8F,$F8,$87,$D0,$8C
	.dc.b	$5F,$EC,$BE,$29,$F6,$5F,$47,$F1,$07,$D0,$FF,$27,$9C,$E3,$F4,$27
	.dc.b	$2F,$F1,$AF,$88,$FE,$11,$87,$F8,$2B,$6F,$F1,$CF,$A1,$F8,$C9,$CC
	.dc.b	$FF,$8A,$DA,$FF,$9B,$3F,$23,$FF,$46,$5F,$F2,$46,$C7,$F1,$BE,$6F
	.dc.b	$97,$F3,$BF,$CE,$67,$3D,$1F,$D2,$3E,$4B,$FC,$67,$87,$E1,$3C,$2F
	.dc.b	$D2,$63,$FC,$3A,$72,$F3,$8C,$D7,$F5,$0F,$93,$FE,$71,$7F,$EA,$0F
	.dc.b	$6F,$F4,$71,$8B,$FA,$4D,$7D,$DF,$A3,$7F,$36,$BF,$EB,$2D,$BF,$C3
	.dc.b	$FE,$23,$FD,$62,$BF,$8F,$CE,$5F,$F4,$5F,$CB,$BD,$6F,$D0,$7E,$23
	.dc.b	$FE,$E3,$C3,$FA,$2C,$E5,$FC,$FC,$66,$BF,$94,$FC,$CF,$FA,$26,$7F
	.dc.b	$D9,$DF,$23,$FB,$A8,$C4,$FF,$B3,$23,$F7,$BF,$9B,$3F,$10,$FE,$C4
	.dc.b	$DB,$FD,$5F,$E2,$5F,$92,$BF,$E6,$B3,$97,$F9,$BE,$DF,$D3,$8F,$E6
	.dc.b	$3F,$79,$FE,$87,$BF,$EB,$71,$97,$FC,$97,$19,$9F,$EA,$81,$7C,$BF
	.dc.b	$F2,$58,$C3,$63,$33,$FE,$1C,$9C,$BF,$DD,$7E,$0F,$FB,$C1,$FE,$00
	.dc.b	$8C,$9E,$67,$F8,$57,$D1,$3E,$5E,$98,$F9,$BB,$7F,$C2,$BF,$0C,$7E
	.dc.b	$25,$FC,$45,$B1,$FC,$3A,$F9,$3B,$9F,$E2,$DF,$0F,$F6,$FC,$17,$E8
	.dc.b	$FF,$E4,$3A,$FE,$2B,$39,$8C,$69,$FD,$81,$BB,$FF,$89,$3C,$07,$E1
	.dc.b	$FF,$D8,$A3,$FC,$47,$3A,$C3,$F8,$8C,$E5,$FF,$4A,$4F,$F5,$23,$FC
	.dc.b	$53,$F8,$CB,$EF,$BF,$C3,$9F,$3D,$FE,$1F,$8C,$57,$F8,$41,$B9,$7F
	.dc.b	$0D,$6B,$FB,$96,$BF,$EA,$16,$BF,$EE,$2F,$8F,$7E,$25,$FE,$1C,$3B
	.dc.b	$61,$6F,$FC,$3B,$E6,$7F,$39,$6F,$B1,$FC,$F7,$E2,$FF,$E0,$04,$60
	.dc.b	$FE,$77,$F8,$97,$E6,$35,$F8,$8D,$8F,$F5,$95,$F3,$CF,$F8,$24,$E5
	.dc.b	$FE,$9D,$F8,$13,$EF,$3F,$A5,$2F,$AD,$B1,$86,$61,$FD,$2B,$E3,$FF
	.dc.b	$C7,$F9,$7D,$3B,$E1,$39,$3F,$1D,$F8,$B7,$F4,$99,$FF,$08,$FB,$1F
	.dc.b	$D4,$F2,$FC,$EC,$E6,$BF,$D5,$73,$9A,$DF,$57,$E7,$E3,$F7,$57,$FE
	.dc.b	$FA,$6B,$C2,$3F,$49,$7F,$DE,$2F,$FC,$B5,$28,$C3,$FA,$CA,$FE,$09
	.dc.b	$7F,$E2,$49,$CB,$F8,$DF,$A2,$FE,$24,$DF,$AB,$FE,$DD,$18,$C3,$FA
	.dc.b	$EA,$8C,$9F,$DC,$FF,$14,$7D,$B6,$2B,$FB,$8C,$61,$FF,$37,$73,$BF
	.dc.b	$48,$F1,$C3,$44,$E7,$5F,$CA,$7C,$AF,$F9,$E2,$31,$5F,$95,$F9,$6F
	.dc.b	$CC,$19,$C3,$9E,$0F,$03,$9C,$FE,$77,$E7,$BF,$3D,$AD,$C3,$F3,$CF
	.dc.b	$E8,$7C,$CF,$F0,$EE,$5F,$C9,$31,$FC,$F4,$60,$CF,$C3,$FE,$9C,$E1
	.dc.b	$7F,$A7,$7C,$DF,$04,$FA,$CA,$7F,$85,$FE,$F3,$FC,$5F,$5F,$0A,$74
	.dc.b	$FF,$85,$BC,$A7,$D8,$7F,$87,$5B,$1B,$BC,$4F,$F8,$87,$E5,$9F,$81
	.dc.b	$53,$1E,$15,$BF,$F1,$0F,$DF,$FF,$92,$7A,$BF,$85,$36,$3F,$C2,$F7
	.dc.b	$C9,$DF,$BB,$FE,$2D,$9F,$F1,$07,$07,$F8,$71,$AF,$F8,$DD,$BE,$59
	.dc.b	$BF,$8B,$3D,$7F,$F1,$B3,$70,$F1,$F1,$62,$7F,$C7,$5F,$80,$FF,$81
	.dc.b	$B1,$FE,$F4,$DD,$3F,$90,$3F,$E2,$A8,$CF,$C3,$F0,$FB,$79,$8F,$F8
	.dc.b	$FA,$B8,$3F,$E3,$25,$1F,$E4,$7F,$C0,$BF,$A3,$DF,$E3,$CF,$F1,$BB
	.dc.b	$87,$1A,$A3,$11,$AF,$63,$FC,$ED,$39,$F9,$8E,$87,$58,$B8,$FF,$94
	.dc.b	$3F,$03,$FE,$01,$7D,$1A,$FF,$91,$9A,$FF,$98,$5B,$67,$E4,$47,$F8
	.dc.b	$75,$BA,$7F,$43,$7F,$84,$57,$3D,$7F,$CC,$91,$FD,$52,$B8,$7A,$D8
	.dc.b	$F4,$72,$FF,$9A,$7E,$67,$F8,$DE,$1F,$E5,$86,$D7,$8E,$75,$3D,$3E
	.dc.b	$A3,$E0,$E1,$FE,$E8,$AF,$94,$7A,$63,$26,$BE,$1B,$C6,$71,$AF,$88
	.dc.b	$FF,$73,$FB,$DF,$F3,$C7,$D1,$7F,$01,$FA,$CF,$F5,$1E,$1F,$C0,$E7
	.dc.b	$2A,$32,$A3,$F3,$0E,$7A,$7E,$E4,$70,$9D,$B6,$33,$D6,$BF,$04,$70
	.dc.b	$EC,$65,$8C,$7C,$57,$B7,$B1,$8B,$D3,$E3,$8D,$63,$E0,$BE,$73,$C0
	.dc.b	$D5,$70,$C7,$C9,$1C,$CD,$3E,$CA,$F4,$DF,$4C,$7D,$37,$C6,$31,$7F
	.dc.b	$1F,$0E,$F7,$C2,$B4,$49,$EB,$3A,$BE,$71,$E0,$BE,$6F,$98,$DE,$3C
	.dc.b	$70,$E9,$1B,$C3,$D5,$87,$38,$F5,$4E,$89,$87,$CD,$BE,$4F,$82,$33
	.dc.b	$3E,$99,$CB,$CA,$74,$71,$7C,$04,$9E,$2F,$E0,$EB,$C4,$FA,$63,$CE
	.dc.b	$FE,$5B,$1F,$C5,$B7,$1F,$B8,$DC,$7F,$C9,$53,$FE,$9C,$F9,$DB,$D3
	.dc.b	$FE,$FC,$AF,$CF,$9F,$91,$70,$F5,$6B,$F9,$27,$86,$1F,$92,$6E,$59
	.dc.b	$AF,$C1,$7D,$E7,$F1,$3A,$D4,$7E,$16,$B7,$7F,$A3,$6B,$FA,$1A,$E1
	.dc.b	$7C,$E3,$F1,$33,$AC,$3D,$51,$F8,$CA,$DB,$CE,$FB,$11,$D1,$70,$F1
	.dc.b	$BD,$23,$CE,$E7,$0F,$04,$E8,$FD,$7D,$76,$9F,$8D,$5B,$7C,$07,$63
	.dc.b	$A1,$27,$95,$7E,$EA,$B9,$5F,$AA,$73,$5C,$9E,$EF,$53,$B9,$CE,$B9
	.dc.b	$39,$9E,$51,$E8,$33,$1B,$3A,$39,$7F,$89,$57,$C8,$E5,$FE,$69,$E7
	.dc.b	$7F,$49,$E6,$FE,$64,$DD,$FF,$9E,$36,$FF,$9E,$7C,$45,$BF,$E7,$8F
	.dc.b	$FA,$5D,$B1,$F8,$C6,$BF,$CB,$1A,$FF,$57,$D9,$F5,$9F,$E5,$78,$FF
	.dc.b	$14,$63,$FC,$3A,$73,$EB,$78,$5B,$3A,$CB,$FC,$5D,$1C,$BD,$2D,$60
	.dc.b	$69,$2B,$81,$FC,$9E,$3F,$98,$E1,$FD,$47,$E5,$BD,$9C,$E6,$9E,$FE
	.dc.b	$E7,$95,$4F,$80,$E3,$0F,$80,$D7,$E0,$35,$FE,$E6,$BE,$C7,$A1,$23
	.dc.b	$CB,$B7,$F7,$7F,$A1,$F0,$1D,$89,$1D,$9F,$A1,$E1,$BE,$53,$EA,$36
	.dc.b	$7F,$C1,$51,$F5,$FC,$67,$6D,$24,$F8,$23,$2D,$FB,$9E,$1A,$C8,$D3
	.dc.b	$F8,$1F,$21,$F6,$54,$FC,$13,$89,$FC,$13,$5F,$C1,$35,$FF,$0A,$BF
	.dc.b	$C4,$38,$7F,$C7,$7D,$BD,$4F,$C7,$79,$AC,$69,$2F,$FC,$9C,$EB,$C5
	.dc.b	$FC,$A7,$9F,$FE,$7B,$9F,$96,$D5,$87,$FB,$46,$30,$FF,$5C,$F4,$1F
	.dc.b	$64,$FC,$0B,$F4,$11,$86,$C6,$6F,$FD,$02,$D7,$AF,$67,$FA,$07,$95
	.dc.b	$FA,$4B,$E1,$BF,$55,$BB,$F3,$2F,$F6,$5E,$1F,$94,$8C,$6D,$A9,$CC
	.dc.b	$F4,$13,$C7,$7D,$1F,$CE,$A3,$33,$F1,$DF,$E7,$B3,$B6,$C6,$5F,$F0
	.dc.b	$FC,$78,$F9,$7C,$1E,$14,$9F,$25,$7E,$11,$F1,$53,$F1,$E7,$91,$EC
	.dc.b	$6C,$FA,$DA,$FA,$78,$5F,$A5,$73,$E3,$FA,$0D,$E1,$FF,$54,$4E,$B4
	.dc.b	$FA,$33,$9B,$F8,$38,$3C,$6F,$E2,$5E,$1F,$D8,$9F,$F5,$AF,$9B,$FC
	.dc.b	$57,$DF,$3A,$E6,$DD,$5A,$FD,$7B,$C5,$F9,$DC,$DF,$C2,$63,$F9,$04
	.dc.b	$3F,$0A,$7E,$05,$86,$C7,$FB,$59,$AF,$D7,$1F,$7D,$FC,$3B,$83,$FC
	.dc.b	$EC,$FF,$90,$23,$FC,$9E,$D7,$3E,$67,$DB,$78,$B7,$F0,$3F,$E1,$11
	.dc.b	$FE,$E4,$9F,$F8,$00,$DF,$88,$F3,$CE,$A3,$D7,$8F,$D2,$E4,$FF,$1C
	.dc.b	$E8,$FB,$B7,$A9,$FC,$ED,$B7,$FC,$4A,$F8,$69,$F9,$B7,$FD,$6B,$19
	.dc.b	$E7,$FE,$85,$1A,$86,$1F,$1B,$D4,$FE,$8D,$F1,$5E,$3C,$FC,$2D,$C7
	.dc.b	$F4,$CD,$BF,$F6,$F7,$C5,$F6,$7F,$41,$F0,$88,$C5,$FF,$94,$A3,$18
	.dc.b	$7F,$55,$6D,$7F,$2A,$E7,$FC,$44,$3F,$4B,$7B,$9B,$6D,$7F,$44,$9D
	.dc.b	$4F,$E6,$FE,$66,$4F,$EA,$BA,$D1,$FA,$FE,$FF,$D1,$36,$33,$3B,$F1
	.dc.b	$B8,$BF,$A8,$8C,$B6,$3D,$1F,$25,$F0,$78,$5F,$25,$F8,$7A,$F9,$6D
	.dc.b	$CD,$5B,$F6,$1F,$63,$CB,$D4,$98,$7F,$4A,$9C,$BF,$D3,$3E,$55,$F7
	.dc.b	$9F,$DE,$5B,$19,$8F,$EF,$9F,$82,$3E,$46,$92,$7F,$AA,$C6,$5B,$EE
	.dc.b	$FF,$01,$FB,$A9,$1F,$C0,$5F,$D4,$FE,$09,$B8,$FD,$23,$5F,$F5,$7B
	.dc.b	$5F,$F6,$D3,$50,$71,$C9,$FC,$47,$E6,$FA,$FD,$9C,$AF,$E2,$11,$F8
	.dc.b	$3F,$9E,$7E,$09,$49,$3F,$DB,$BF,$1B,$FC,$C7,$27,$FB,$F7,$E4,$FF
	.dc.b	$95,$36,$7F,$CC,$AD,$7F,$D8,$AD,$7F,$DE,$0B,$E8,$FF,$C1,$11,$FD
	.dc.b	$EB,$2F,$E8,$0D,$F6,$3F,$2D,$3A,$8F,$CE,$FD,$23,$F0,$8F,$CD,$36
	.dc.b	$3F,$28,$D7,$F8,$AF,$E3,$2F,$7B,$F4,$1F,$51,$FD,$1A,$7C,$6D,$7E
	.dc.b	$4E,$68,$C4,$23,$1E,$EF,$C0,$FA,$9F,$DD,$78,$0F,$C6,$29,$F1,$A3
	.dc.b	$E3,$FD,$5C,$F3,$FF,$60,$F0,$FC,$63,$F5,$79,$7E,$E3,$EB,$BF,$70
	.dc.b	$FC,$A6,$BF,$29,$AF,$CA,$6B,$E8,$FB,$0F,$E7,$CF,$F8,$93,$F1,$8F
	.dc.b	$C5,$79,$FF,$C6,$9B,$E9,$6F,$CF,$FE,$05,$F6,$4F,$A4,$A4,$C3,$FC
	.dc.b	$5B,$F8,$C7,$F8,$3E,$FE,$2F,$D0,$FF,$2F,$E4,$34,$95,$F8,$4F,$C2
	.dc.b	$FF,$C0,$F5,$FD,$F7,$F1,$5F,$5E,$BF,$CB,$1B,$1F,$42,$77,$5F,$3F
	.dc.b	$EE,$BE,$93,$FC,$3D,$AF,$F1,$E6,$DF,$F3,$6E,$6F,$F2,$53,$FC,$7F
	.dc.b	$E4,$3F,$C3,$55,$F9,$06,$BF,$79,$AF,$DD,$6E,$6F,$BC,$7D,$27,$F9
	.dc.b	$E7,$3F,$D6,$DF,$F9,$41,$AF,$F3,$09,$DB,$FC,$97,$EF,$BF,$31,$8F
	.dc.b	$F3,$16,$D7,$E3,$9B,$3F,$D0,$A5,$F8,$27,$F3,$5F,$A1,$7D,$87,$EB
	.dc.b	$9B,$5F,$AE,$6B,$FA,$E6,$BF,$AE,$7E,$C2,$F3,$DF,$86,$3E,$A7,$FD
	.dc.b	$0C,$D9,$FE,$79,$5C,$2B,$FC,$F3,$E7,$7F,$A2,$FE,$78,$D3,$1E,$97
	.dc.b	$F4,$BF,$18,$FF,$04,$6E,$1F,$C1,$1A,$FF,$04,$6B,$FE,$9D,$FC,$13
	.dc.b	$C4,$FF,$8E,$70,$43,$C0,$FF,$4E,$36,$7F,$D2,$B3,$AD,$3F,$D4,$1F
	.dc.b	$8B,$3E,$AA,$92,$BF,$AC,$7D,$97,$FA,$7D,$BF,$21,$FC,$6A,$31,$1F
	.dc.b	$C6,$1A,$FF,$18,$6B,$FC,$1B,$EC,$F7,$C9,$FD,$B5,$BC,$D9,$BF,$F5
	.dc.b	$96,$FF,$EB,$7B,$F6,$36,$FF,$D7,$0D,$AF,$F5,$AB,$67,$EC,$3F,$D6
	.dc.b	$FE,$D7,$F9,$AC,$7F,$71,$6B,$FE,$00,$6D,$7F,$B0,$FF,$0B,$7D,$67
	.dc.b	$FB,$0D,$B5,$F2,$1A,$FC,$F6,$CF,$1F,$4B,$FC,$0D,$F6,$DF,$89,$AF
	.dc.b	$F0,$03,$5F,$F0,$7B,$67,$FC,$1D,$1F,$C7,$1B,$F1,$3F,$E0,$F8,$C6
	.dc.b	$7F,$CB,$9A,$7F,$2D,$1E,$C6,$CF,$E9,$AE,$0F,$E9,$A5,$3F,$A6,$94
	.dc.b	$FE,$9B,$F8,$5F,$F1,$E7,$F9,$C3,$7C,$13,$9E,$5F,$F7,$5F,$D1,$FF
	.dc.b	$B8,$89,$5F,$8C,$FA,$AF,$F5,$96,$7F,$E2,$9D,$3A,$4F,$F9,$0F,$E6
	.dc.b	$7F,$D4,$AF,$F0,$23,$13,$FE,$8D,$F9,$3F,$F5,$2B,$67,$73,$FC,$22
	.dc.b	$7F,$C2,$79,$6F,$E2,$7C,$B3,$CE,$72,$D8,$CB,$FE,$99,$FA,$DF,$F3
	.dc.b	$02,$69,$1F,$EE,$38,$FF,$86,$B7,$FB,$3F,$0E,$7F,$C7,$58,$FF,$10
	.dc.b	$72,$7C,$1E,$0D,$5F,$EC,$F5,$FA,$B3,$A8,$C4,$EB,$5F,$F2,$75,$BE
	.dc.b	$87,$8D,$7D,$46,$BF,$DB,$A3,$57,$FE,$08,$6C,$7F,$82,$2B,$CD,$3F
	.dc.b	$E4,$F6,$C7,$F9,$7A,$77,$1F,$E5,$D3,$27,$8B,$E7,$FC,$95,$DD,$B5
	.dc.b	$FD,$EE,$77,$39,$3F,$85,$FE,$15,$FC,$79,$BE,$1F,$0E,$34,$FF,$B5
	.dc.b	$67,$27,$0E,$AD,$47,$F2,$28,$CE,$1F,$C6,$E3,$0F,$F1,$DF,$C2,$9F
	.dc.b	$5D,$D5,$B5,$FC,$71,$FF,$47,$B6,$35,$E5,$48,$C1,$B7,$CD,$F8,$67
	.dc.b	$0B,$FE,$AB,$92,$31,$1F,$D6,$A3,$1F,$37,$F5,$23,$FA,$34,$E9,$FF
	.dc.b	$11,$F8,$4F,$B8,$F9,$3D,$88,$C6,$5F,$CC,$AF,$C1,$C1,$FE,$8F,$9C
	.dc.b	$E1,$FC,$EF,$F0,$AF,$EB,$77,$FC,$E3,$C8,$F9,$53,$9F,$3B,$E7,$46
	.dc.b	$30,$FC,$8E,$AC,$3F,$DF,$78,$7E,$BE,$31,$F1,$BE,$96,$C6,$67,$78
	.dc.b	$FC,$2F,$C2,$BF,$C7,$93,$FD,$21,$B1,$FE,$A8,$FB,$77,$D1,$FD,$F9
	.dc.b	$CE,$5E,$2F,$C2,$BF,$43,$7F,$EA,$EF,$49,$F6,$5F,$85,$9C,$D7,$FA
	.dc.b	$BE,$73,$97,$F5,$6F,$9C,$7C,$57,$9D,$B1,$EB,$6E,$3F,$61,$AF,$F8
	.dc.b	$43,$DB,$70,$7B,$31,$FF,$6A,$B6,$FF,$D0,$33,$AE,$2F,$CF,$BC,$27
	.dc.b	$EC,$FB,$2F,$90,$A6,$BF,$0A,$7C,$56,$31,$FF,$45,$E5,$FE,$E2,$FC
	.dc.b	$39,$E1,$FE,$1F,$34,$62,$BF,$91,$46,$27,$53,$9D,$7F,$C4,$F2,$FC
	.dc.b	$2D,$F1,$1E,$66,$E1,$E0,$3D,$6F,$F9,$E5,$C5,$7F,$BA,$23,$6F,$FB
	.dc.b	$83,$F0,$AF,$F5,$94,$FF,$B5,$5B,$0C,$BF,$DF,$7F,$85,$7F,$B7,$2B
	.dc.b	$5F,$37,$F1,$BF,$28,$FF,$54,$C7,$A9,$B7,$E8,$6B,$F4,$DA,$FD,$0F
	.dc.b	$C2,$FF,$D2,$93,$E6,$D1,$18,$9F,$14,$EF,$CF,$FA,$B3,$9C,$BF,$DE
	.dc.b	$DF,$1F,$FE,$F9,$AF,$F7,$73,$6F,$88,$C1,$F8,$5F,$C2,$BE,$3F,$51
	.dc.b	$F8,$0F,$A8,$FE,$17,$7F,$41,$AF,$E2,$DA,$FE,$21,$AF,$5D,$5F,$85
	.dc.b	$FE,$29,$B9,$FB,$55,$66,$7F,$82,$FC,$B3,$F0,$FA,$7E,$4B,$C3,$FE
	.dc.b	$07,$9F,$F4,$CF,$29,$5B,$CB,$E4,$7E,$3A,$FA,$6F,$C7,$36,$3F,$1A
	.dc.b	$DD,$7F,$16,$D7,$F4,$DF,$86,$FE,$CA,$7F,$46,$D8,$F9,$0B,$F0,$5F
	.dc.b	$D2,$B9,$C7,$FD,$A9,$E9,$7F,$1E,$FA,$0F,$93,$A7,$F6,$2E,$47,$E1
	.dc.b	$AF,$B0,$D5,$FE,$BB,$EC,$9F,$03,$FA,$56,$ED,$FA,$46,$ED,$FA,$26
	.dc.b	$BF,$BD,$FC,$35,$F8,$77,$EE,$1B,$5F,$8B,$7F,$31,$3F,$82,$CD,$1A
	.dc.b	$8C,$7E,$47,$FE,$7D,$AF,$ED,$1E,$EF,$F0,$28,$C4,$FE,$2B,$83,$F4
	.dc.b	$93,$FD,$92,$1F,$87,$BE,$3B,$FA,$24,$61,$B1,$9C,$3F,$C4,$13,$96
	.dc.b	$C3,$F0,$9E,$8D,$8F,$DC,$CE,$CF,$CD,$9E,$09,$FE,$2F,$F4,$0F,$C3
	.dc.b	$E9,$8F,$95,$B7,$FA,$D9,$FE,$09,$7F,$0D,$BC,$BF,$5D,$E5,$3B,$F9
	.dc.b	$7F,$E4,$7F,$26,$FC,$46,$98,$F8,$4B,$5F,$B9,$9D,$3F,$C4,$38,$1F
	.dc.b	$66,$F9,$1D,$E9,$F2,$E3,$15,$BF,$BF,$FE,$69,$3F,$E0,$FF,$AD,$FE
	.dc.b	$D3,$18,$C3,$F8,$A4,$7F,$6C,$DB,$FB,$05,$AD,$4B,$FE,$D3,$1F,$66
	.dc.b	$B6,$EA,$33,$1F,$AD,$6D,$73,$A7,$C8,$7F,$4E,$AD,$7B,$7F,$DF,$23
	.dc.b	$15,$FD,$2B,$E6,$1F,$5F,$FE,$00,$F3,$5F,$85,$FF,$4C,$9C,$D7,$C5
	.dc.b	$9C,$B6,$78,$1B,$AB,$1F,$C9,$7F,$10,$78,$1F,$D5,$9B,$19,$EB,$C4
	.dc.b	$E7,$ED,$A4,$F1,$8B,$C1,$FA,$3F,$8B,$7E,$31,$F9,$BF,$C4,$7F,$D8
	.dc.b	$57,$FC,$CE,$72,$FE,$4A,$33,$5F,$92,$F4,$FF,$B2,$4F,$E7,$1B,$1F
	.dc.b	$E7,$D6,$BF,$F0,$01,$AF,$D7,$0B,$B3,$F5,$8D,$9F,$D0,$BD,$38,$FF
	.dc.b	$CC,$AE,$6B,$F7,$9A,$7F,$17,$F1,$3F,$65,$F2,$DF,$DE,$A3,$CF,$7F
	.dc.b	$C4,$DD,$13,$CD,$D4,$ED,$FC,$7F,$C8,$7F,$98,$3A,$5F,$88,$F7,$C3
	.dc.b	$F0,$D3,$98,$C6,$9F,$91,$AD,$EB,$93,$9C,$7F,$72,$8C,$BF,$D7,$FF
	.dc.b	$07,$FD,$7E,$BF,$C3,$67,$D5,$18,$C7,$F2,$99,$FE,$8B,$3F,$F2,$87
	.dc.b	$5F,$E7,$5B,$7F,$CD,$71,$FF,$56,$4E,$5B,$0F,$A4,$FD,$59,$CE,$3F
	.dc.b	$60,$FF,$88,$3F,$08,$FF,$18,$E7,$FD,$A2,$9F,$5B,$FC,$8B,$3F,$E7
	.dc.b	$95,$A7,$FD,$3F,$3A,$F1,$BE,$45,$6F,$1F,$88,$DC,$3F,$B9,$61,$C6
	.dc.b	$73,$5F,$E3,$0F,$91,$FD,$1D,$7F,$37,$7F,$B3,$CF,$F8,$15,$CE,$3F
	.dc.b	$D3,$5C,$CE,$FD,$BE,$47,$3E,$DB,$7F,$E0,$DF,$C2,$7E,$13,$FE,$30
	.dc.b	$6D,$7F,$8B,$9A,$FF,$92,$5A,$FF,$91,$9F,$C5,$C5,$FA,$5F,$F2,$3C
	.dc.b	$72,$C7,$FB,$03,$AA,$FF,$08,$7E,$19,$FE,$0F,$BF,$F2,$7B,$63,$3E
	.dc.b	$5F,$F9,$42,$31,$D2,$7F,$C9,$B5,$FE,$25,$6D,$7F,$5B,$76,$63,$0F
	.dc.b	$EE,$1F,$86,$BE,$8F,$FC,$A0,$DB,$FF,$27,$B6,$FF,$C9,$AD,$7F,$CD
	.dc.b	$0F,$F5,$17,$FB,$43,$6B,$AD,$7F,$99,$A3,$15,$FE,$6A,$FC,$3B,$F6
	.dc.b	$97,$FE,$71,$6B,$9E,$9F,$F3,$14,$62,$3F,$C6,$8F,$EF,$67,$97,$C7
	.dc.b	$FF,$9C,$A3,$F5,$B3,$A8,$C7,$C9,$CA,$FF,$24,$53,$F0,$CC,$5F,$F7
	.dc.b	$F8,$7E,$13,$CC,$CE,$9F,$E1,$58,$CD,$F2,$FC,$5B,$CA,$47,$FC,$EB
	.dc.b	$08,$F4,$FC,$FF,$94,$F3,$36,$7F,$BB,$FE,$21,$FD,$03,$E7,$BF,$AA
	.dc.b	$7D,$3F,$FA,$9B,$3F,$EA,$D3,$95,$19,$51,$F9,$C8,$C3,$B6,$F3,$AB
	.dc.b	$AF,$C4,$7F,$2B,$AD,$37,$A5,$8E,$B1,$F0,$E1,$C5,$F0,$D7,$18,$C6
	.dc.b	$3D,$B1,$E1,$5E,$4C,$39,$BC,$0E,$71,$AC,$7C,$31,$83,$47,$1C,$75
	.dc.b	$87,$6C,$7C,$17,$83,$96,$1D,$B1,$F1,$61,$CB,$1E,$AF,$C6,$C3,$93
	.dc.b	$F3,$2B,$86,$1D,$23,$84,$E9,$E1,$87,$6C,$7C,$11,$A8,$B8,$FD,$0C
	.dc.b	$39,$CF,$AE,$3C,$64,$C3,$E7,$5F,$07,$C1,$19,$9F,$24,$E5,$E5,$3A
	.dc.b	$38,$BE,$12,$4F,$17,$E4,$18,$8F,$63,$FD,$A4,$F5,$94,$A6,$71,$FC
	.dc.b	$73,$7E,$5B,$FD,$C6,$EE,$3F,$DB,$3F,$33,$FE,$54,$6C,$66,$76,$F1
	.dc.b	$9F,$BD,$1C,$AF,$F2,$AD,$BF,$C9,$7C,$AF,$F7,$D9,$CD,$70,$25,$FD
	.dc.b	$B9,$D5,$7E,$66,$F9,$4F,$E9,$9A,$FE,$91,$B8,$7F,$B8,$A3,$9E,$1D
	.dc.b	$9F,$C3,$4E,$AF,$AC,$73,$C3,$78,$77,$7F,$31,$3A,$9E,$D1,$A7,$81
	.dc.b	$E7,$30,$68,$EE,$F2,$73,$87,$CA,$C3,$C8,$7E,$EA,$BB,$46,$6F,$C5
	.dc.b	$5C,$8E,$C5,$AE,$31,$A7,$94,$76,$C3,$E1,$9E,$F7,$18,$78,$B0,$F8
	.dc.b	$3D,$2E,$07,$C2,$7F,$8C,$78,$BF,$91,$37,$5F,$E7,$4D,$C7,$F9,$AF
	.dc.b	$B7,$70,$E4,$FF,$A6,$9B,$1F,$C9,$A3,$18,$1F,$F4,$FF,$03,$E9,$FF
	.dc.b	$D4,$D5,$FE,$77,$CB,$F8,$5C,$E7,$CC,$F0,$36,$75,$9F,$F9,$F6,$39
	.dc.b	$70,$6B,$0F,$E5,$18,$72,$6D,$6D,$B7,$FC,$9D,$FF,$34,$BD,$06,$6E
	.dc.b	$BF,$F1,$34,$66,$31,$E3,$7C,$19,$CC,$FF,$64,$F0,$70,$7C,$38,$34
	.dc.b	$91,$FD,$C3,$E7,$B9,$11,$ED,$2F,$67,$B1,$CC,$91,$FD,$E1,$FA,$1F
	.dc.b	$87,$BE,$6B,$EA,$36,$7D,$0D,$7F,$DC,$0F,$F6,$57,$3D,$DC,$08,$F0
	.dc.b	$FA,$8F,$BD,$E3,$7F,$AD,$E3,$FC,$51,$EF,$E0,$CF,$C7,$C8,$FF,$15
	.dc.b	$3F,$83,$93,$4F,$C3,$79,$1E,$19,$FF,$79,$37,$8F,$FA,$5F,$0A,$31
	.dc.b	$C7,$FE,$BB,$8C,$78,$9F,$91,$93,$74,$98,$7D,$79,$CB,$FE,$6F,$F1
	.dc.b	$DF,$83,$7E,$6F,$CE,$FF,$12,$36,$33,$3F,$E7,$3F,$80,$FC,$F7,$A6
	.dc.b	$75,$52,$74,$7E,$27,$C1,$7E,$1F,$E8,$84,$63,$3F,$D5,$36,$38,$4B
	.dc.b	$8F,$2D,$8F,$E7,$53,$D2,$7F,$9E,$47,$49,$D3,$FD,$06,$74,$FF,$41
	.dc.b	$D9,$97,$CB,$8E,$FC,$BD,$8F,$6E,$DF,$0F,$F0,$FF,$C3,$71,$7F,$9E
	.dc.b	$9B,$C5,$FE,$9B,$6D,$FB,$3E,$87,$F9,$82,$AF,$F0,$1E,$69,$CC,$73
	.dc.b	$C7,$B1,$F0,$A3,$38,$7D,$DF,$03,$F8,$1E,$1F,$EB,$AE,$2F,$E1,$D1
	.dc.b	$FE,$59,$8F,$D9,$3A,$C3,$FD,$74,$E8,$91,$AF,$88,$FE,$30,$BF,$0D
	.dc.b	$FE,$02,$DF,$07,$31,$83,$F8,$1C,$7A,$F1,$FF,$6A,$46,$23,$FB,$73
	.dc.b	$5C,$FB,$FF,$51,$EC,$FF,$23,$F7,$BF,$E0,$0C,$7F,$0D,$1F,$86,$FE
	.dc.b	$91,$C3,$5F,$F7,$97,$C1,$FF,$2D,$F9,$17,$99,$49,$3F,$C4,$E1,$E7
	.dc.b	$EC,$91,$97,$97,$A1,$FC,$E7,$DB,$7C,$2F,$0A,$FE,$41,$18,$AD,$4E
	.dc.b	$67,$E1,$1F,$C9,$3D,$0F,$C0,$F5,$9F,$C1,$DF,$19,$E7,$5C,$79,$3F
	.dc.b	$77,$5F,$88,$E8,$FC,$27,$E1,$7E,$3F,$C3,$1F,$63,$4F,$C2,$47,$E1
	.dc.b	$7C,$2B,$97,$EB,$7A,$D3,$98,$C7,$E2,$35,$F3,$E3,$F9,$1B,$FC,$74
	.dc.b	$FF,$57,$F8,$87,$CA,$FC,$3F,$8D,$FA,$1B,$7F,$96,$B8,$11,$8E,$BF
	.dc.b	$E2,$17,$C3,$3F,$E8,$4D,$7E,$2F,$12,$31,$87,$F1,$A6,$D7,$2F,$C1
	.dc.b	$7F,$B4,$7E,$16,$F8,$AA,$4C,$7F,$A7,$47,$F9,$CD,$EA,$DC,$3F,$92
	.dc.b	$B6,$FF,$92,$36,$7F,$AC,$AF,$BC,$FD,$0F,$E1,$7D,$E7,$F2,$4D,$7E
	.dc.b	$E8,$71,$7F,$E7,$C8,$C7,$C1,$72,$3F,$3D,$26,$92,$3F,$88,$7C,$BF
	.dc.b	$EF,$1F,$E0,$DF,$69,$FB,$F3,$FC,$11,$19,$D9,$F8,$57,$E7,$AB,$FB
	.dc.b	$73,$5F,$9A,$DC,$7F,$A2,$BC,$7F,$1E,$61,$FE,$11,$6D,$7F,$21,$6B
	.dc.b	$FE,$65,$B4,$62,$04,$E2,$7C,$F3,$F0,$0F,$D4,$36,$7F,$4E,$D7,$F0
	.dc.b	$3A,$76,$CF,$FA,$CF,$E1,$5F,$CB,$E7,$F0,$AD,$8C,$F9,$FF,$E2,$98
	.dc.b	$7C,$DF,$F8,$AF,$F1,$DF,$CE,$DF,$F3,$96,$BE,$96,$BE,$83,$FC,$2E
	.dc.b	$E1,$F5,$7D,$27,$C7,$E5,$BE,$3F,$F0,$18,$C4,$FF,$87,$BF,$0B,$78
	.dc.b	$FF,$E3,$BF,$C2,$BF,$67,$3F,$E1,$F6,$BF,$9E,$6E,$BF,$DD,$97,$CB
	.dc.b	$7F,$90,$FC,$97,$C8,$7F,$90,$DB,$7F,$D4,$DA,$FD,$36,$BE,$BE,$47
	.dc.b	$D5,$7E,$25,$BE,$57,$F1,$98,$FC,$47,$E1,$5F,$9D,$AF,$F1,$A7,$E1
	.dc.b	$3F,$E9,$4D,$3F,$C1,$2D,$E1,$FF,$04,$36,$7F,$C7,$AF,$F8,$B5,$FF
	.dc.b	$1E,$7E,$3B,$FD,$41,$F9,$9D,$68,$C4,$7F,$84,$5A,$FF,$91,$9B,$9B
	.dc.b	$F0,$97,$CD,$FF,$2F,$6D,$7F,$2E,$6B,$FC,$23,$F0,$A7,$D6,$FF,$9A
	.dc.b	$F1,$FF,$25,$E9,$FE,$22,$6E,$9F,$E2,$16,$DF,$F9,$5B,$F1,$BF,$F6
	.dc.b	$05,$7F,$8A,$3E,$DF,$FC,$F7,$3F,$E0,$16,$BF,$A1,$6B,$EE,$6B,$FA
	.dc.b	$1F,$B9,$F9,$8F,$F4,$0D,$51,$8A,$FE,$4C,$DC,$7F,$C6,$9F,$4F,$E4
	.dc.b	$CB,$FC,$05,$F2,$BF,$E8,$5B,$FF,$40,$63,$FE,$69,$6B,$FC,$A1,$F2
	.dc.b	$9F,$E8,$0F,$53,$F8,$A4,$7E,$77,$E5,$7F,$C1,$5E,$6F,$FA,$22,$73
	.dc.b	$3F,$E8,$16,$A9,$7E,$13,$FA,$C8,$FE,$BB,$F7,$7F,$81,$D8,$FE,$0E
	.dc.b	$B7,$E1,$FF,$A3,$BE,$59,$F5,$B4,$98,$7F,$66,$F1,$F2,$FF,$A1,$ED
	.dc.b	$F7,$EF,$C2,$7F,$D3,$AD,$9F,$F7,$94,$EA,$72,$6B,$1F,$F4,$ED,$3E
	.dc.b	$92,$98,$8F,$EF,$1F,$07,$FE,$1F,$E5,$FE,$2C,$DC,$5B,$23,$19,$3F
	.dc.b	$01,$FC,$2D,$FF,$A9,$DF,$F0,$B4,$66,$FF,$D6,$71,$83,$3A,$FF,$5B
	.dc.b	$8F,$73,$5F,$F0,$1C,$FC,$6F,$63,$FD,$6B,$3F,$EA,$C8,$F9,$3F,$0F
	.dc.b	$82,$57,$2F,$86,$FF,$5D,$FE,$0E,$FC,$22,$92,$7E,$64,$3F,$08,$D9
	.dc.b	$23,$2F,$2F,$BC,$FF,$0E,$47,$F1,$4D,$7F,$9D,$34,$F8,$A5,$DF,$F9
	.dc.b	$FF,$99,$C0,$91,$FE,$22,$3F,$8D,$97,$C5,$EA,$7F,$9C,$27,$FA,$3B
	.dc.b	$63,$F7,$67,$F1,$92,$9F,$EC,$FF,$9A,$FF,$6A,$11,$FF,$17,$F8,$1F
	.dc.b	$DB,$B0,$FE,$A0,$D8,$FB,$6D,$B3,$7F,$07,$FE,$E1,$3F,$E4,$BF,$C2
	.dc.b	$1F,$23,$FE,$71,$8F,$E8,$87,$F5,$C8,$CF,$83,$FE,$88,$6C,$68,$38
	.dc.b	$78,$BF,$E2,$6A,$E1,$F0,$86,$92,$7F,$CC,$1F,$4B,$FE,$61,$4D,$23
	.dc.b	$E4,$E5,$FF,$80,$1B,$35,$FD,$57,$D1,$FF,$28,$7F,$D5,$5F,$82,$7E
	.dc.b	$15,$CC,$EE,$FF,$C7,$BF,$82,$FF,$B7,$E7,$FD,$60,$D8,$FC,$AD,$F3
	.dc.b	$77,$F2,$DC,$76,$FF,$4A,$FE,$99,$FE,$59,$C3,$FD,$5A,$DA,$FF,$55
	.dc.b	$35,$FE,$DD,$F1,$A8,$C4,$7F,$82,$23,$1E,$EF,$5A,$C5,$7F,$9E,$1F
	.dc.b	$F6,$23,$D6,$FC,$33,$BB,$FF,$62,$E8,$F2,$F1,$8A,$31,$8F,$2B,$CC
	.dc.b	$62,$FF,$CE,$B8,$3F,$4D,$F9,$97,$FE,$1C,$8F,$81,$18,$9F,$DC,$FE
	.dc.b	$79,$FE,$D4,$AD,$1B,$9D,$57,$CF,$E0,$FE,$B7,$E2,$F9,$DB,$5C,$30
	.dc.b	$FF,$60,$5F,$08,$FF,$15,$CE,$A7,$FA,$57,$DC,$3A,$BE,$07,$52,$31
	.dc.b	$7F,$CA,$FF,$0A,$7D,$76,$A7,$38,$7F,$10,$FC,$8B,$F9,$75,$FF,$31
	.dc.b	$6C,$76,$C7,$CD,$18,$D7,$D5,$E3,$F4,$BF,$E6,$6D,$BF,$CA,$35,$FE
	.dc.b	$A0,$D7,$D7,$F8,$DB,$C1,$F8,$4D,$AF,$E8,$55,$CE,$FE,$44,$E6,$FF
	.dc.b	$B5,$7D,$A7,$FA,$16,$FF,$D5,$ED,$8F,$E5,$7F,$86,$3E,$BB,$E2,$CE
	.dc.b	$70,$FA,$5F,$8E,$6E,$FF,$B2,$35,$F9,$ED,$7F,$B1,$37,$89,$DA,$FB
	.dc.b	$AF,$B4,$DC,$FF,$AD,$2F,$C7,$33,$3F,$7F,$A7,$F5,$59,$7E,$0D,$B7
	.dc.b	$F7,$97,$D9,$6F,$C1,$FC,$37,$B0,$FA,$AF,$F4,$56,$DF,$DD,$DA,$FF
	.dc.b	$77,$8D,$61,$FD,$99,$B1,$FC,$1F,$0F,$F6,$2C,$FF,$6F,$6C,$7F,$85
	.dc.b	$7D,$8F,$91,$39,$F9,$EF,$D6,$7B,$8F,$C3,$BF,$61,$3F,$E7,$E8,$C7
	.dc.b	$C4,$7E,$4B,$DF,$7D,$17,$0B,$DE,$4B,$E5,$F1,$BE,$86,$E3,$FC,$A5
	.dc.b	$B3,$FE,$8D,$AF,$F3,$5C,$E7,$F0,$EF,$F3,$EB,$FA,$7A,$46,$2B,$FC
	.dc.b	$5D,$F2,$BF,$2A,$B5,$39,$E3,$FE,$41,$6F,$C3,$7F,$CD,$F5,$CD,$B8
	.dc.b	$7F,$8D,$FF,$0E,$FF,$38,$CE,$7D,$6F,$87,$F8,$77,$F8,$96,$3D,$0D
	.dc.b	$AF,$F5,$23,$5F,$F3,$0B,$5F,$91,$F8,$77,$F3,$37,$FD,$BE,$DC,$BF
	.dc.b	$DD,$1F,$89,$FE,$7D,$F3,$C6,$6F,$E8,$7E,$1E,$F8,$6F,$AA,$D8,$FE
	.dc.b	$15,$E6,$7F,$A0,$3F,$05,$CD,$7D,$AF,$94,$7D,$FF,$D1,$6D,$FF,$A2
	.dc.b	$E3,$07,$D5,$F8,$CF,$A3,$AF,$FB,$D1,$BA,$7F,$BF,$7F,$14,$FE,$09
	.dc.b	$7F,$83,$9C,$BF,$85,$F8,$5F,$D1,$3F,$5D,$B7,$FE,$63,$F9,$A7,$D6
	.dc.b	$FE,$3A,$73,$8F,$D6,$FC,$3F,$F8,$BD,$FF,$A2,$9A,$FE,$45,$AF,$E3
	.dc.b	$5B,$CE,$FC,$3F,$F0,$52,$AF,$C0,$C6,$1F,$EC,$5C,$2F,$C5,$5F,$81
	.dc.b	$7E,$82,$B5,$AF,$E7,$FF,$10,$7E,$0D,$FA,$46,$C7,$FB,$6B,$D9,$C4
	.dc.b	$FE,$A3,$E3,$5F,$5B,$F9,$F6,$D7,$E7,$DB,$AF,$FB,$AD,$AF,$FA,$A3
	.dc.b	$F0,$FF,$C9,$ED,$F9,$86,$D7,$FB,$B7,$F1,$67,$D8,$FE,$DA,$73,$7F
	.dc.b	$EF,$2F,$C8,$1F,$33,$FD,$96,$11,$8C,$3F,$A8,$7C,$67,$F9,$83,$EA
	.dc.b	$75,$39,$AF,$91,$5F,$EE,$5F,$C4,$7F,$92,$41,$A4,$F9,$7F,$E2,$7F
	.dc.b	$15,$FD,$F7,$EB,$7F,$B8,$FC,$69,$F1,$57,$F3,$3C,$3F,$BE,$C7,$F0
	.dc.b	$EF,$A7,$FC,$5E,$3F,$C1,$DB,$D7,$FE,$E2,$FA,$1D,$7C,$F7,$F1,$4F
	.dc.b	$C6,$1F,$81,$D3,$39,$7F,$05,$8C,$BF,$C1,$FF,$11,$FB,$AD,$EE,$FE
	.dc.b	$3B,$1F,$A2,$CF,$DE,$7F,$17,$76,$E3,$27,$E3,$2F,$C1,$BF,$92,$B7
	.dc.b	$0E,$87,$F1,$57,$15,$D3,$F0,$67,$9A,$FC,$53,$FC,$31,$18,$7E,$3F
	.dc.b	$E0,$F8,$9D,$61,$FE,$05,$E5,$7E,$71,$FE,$0F,$C7,$CF,$18,$38,$61
	.dc.b	$FE,$20,$9D,$77,$FF,$8F,$E7,$FC,$45,$F8,$47,$F8,$80,$D3,$70,$FE
	.dc.b	$59,$F2,$9E,$66,$E5,$FD,$0A,$1F,$8F,$58,$FE,$8B,$FA,$73,$4A,$66
	.dc.b	$FF,$9B,$46,$5F,$E7,$1F,$8B,$3F,$07,$FE,$9A,$D8,$CF,$D2,$62,$7F
	.dc.b	$A8,$FD,$53,$F1,$8A,$66,$BF,$A1,$46,$5F,$F1,$56,$2F,$55,$F8,$07
	.dc.b	$E8,$5B,$3A,$FC,$5D,$F4,$F9,$AD,$57,$F6,$CB,$E4,$D3,$A5,$FF,$9F
	.dc.b	$E3,$2D,$BF,$AF,$1F,$8A,$0E,$38,$FF,$CE,$11,$C0,$D4,$66,$7E,$37
	.dc.b	$D6,$7E,$E9,$FF,$5E,$7C,$17,$F7,$2F,$C2,$BF,$69,$87,$F9,$E6,$72
	.dc.b	$7F,$63,$FC,$7F,$FC,$F1,$1F,$D7,$9B,$1D,$2B,$FB,$DC,$63,$17,$37
	.dc.b	$E3,$3E,$57,$F1,$07,$E2,$7D,$3B,$87,$E4,$7A,$BF,$9A,$BF,$E7,$76
	.dc.b	$DF,$27,$FC,$44,$FE,$51,$CB,$F4,$FE,$27,$E9,$1F,$E7,$7F,$3F,$FD
	.dc.b	$DF,$F5,$B8,$BF,$E2,$7F,$11,$F8,$3B,$FF,$58,$D2,$31,$1F,$8F,$3F
	.dc.b	$D5,$4E,$1F,$C9,$46,$F9,$FF,$89,$E3,$FE,$2E,$9F,$37,$CA,$F8,$6F
	.dc.b	$B7,$07,$E9,$6B,$58,$BF,$22,$69,$FE,$8D,$8C,$3F,$E2,$4F,$C5,$DF
	.dc.b	$03,$9A,$DD,$7F,$85,$7E,$1F,$F0,$FA,$7F,$09,$6E,$5F,$66,$B7,$39
	.dc.b	$E8,$FF,$17,$FC,$EF,$FA,$7E,$3F,$88,$36,$BF,$C3,$2F,$F8,$B3,$CA
	.dc.b	$39,$C7,$FC,$69,$F8,$8B,$DD,$FF,$1C,$B6,$3F,$C5,$13,$BC,$3F,$C6
	.dc.b	$93,$97,$F8,$F7,$E2,$5F,$B7,$7F,$DC,$38,$23,$11,$F4,$63,$0F,$F8
	.dc.b	$EE,$32,$FF,$91,$57,$E3,$5F,$E4,$A6,$E1,$FE,$39,$7B,$65,$FE,$34
	.dc.b	$73,$5F,$CB,$FE,$03,$FC,$71,$5F,$E4,$F6,$C7,$F9,$2D,$7D,$57,$F9
	.dc.b	$59,$CE,$3F,$DD,$BE,$D3,$06,$7E,$D6,$E1,$FE,$49,$FC,$57,$F8,$0B
	.dc.b	$FE,$51,$6E,$3F,$01,$AF,$BD,$AF,$F9,$85,$FE,$DF,$66,$9C,$B3,$FF
	.dc.b	$27,$BA,$FC,$5B,$FC,$CB,$7F,$E5,$F6,$DF,$CE,$6B,$F3,$5A,$FF,$9B
	.dc.b	$DF,$8B,$5F,$E5,$A6,$CE,$7E,$67,$FC,$EF,$18,$AF,$F3,$82,$F8,$2C
	.dc.b	$9A,$F7,$6D,$FF,$9A,$E7,$FC,$9D,$5C,$3C,$6C,$BF,$D8,$63,$39,$9F
	.dc.b	$E5,$71,$E9,$F2,$3F,$CF,$91,$88,$F5,$46,$3E,$AB,$CA,$79,$9B,$87
	.dc.b	$F3,$8F,$C5,$DF,$36,$FA,$6F,$E9,$9F,$6B,$FA,$EC,$7F,$A7,$56,$A7
	.dc.b	$3D,$DF,$C0,$E1,$F2,$CC,$73,$7C,$31,$E8,$B1,$E8,$AF,$A3,$9C,$3A
	.dc.b	$57,$0A,$DD,$76,$BE,$B7,$C6,$B5,$7C,$31,$CD,$F4,$C3,$8E,$3E,$5C
	.dc.b	$3B,$1E,$5A,$DE,$1C,$A3,$06,$8F,$34,$F1,$AF,$1C,$EE,$7C,$55,$C8
	.dc.b	$F0,$D6,$F0,$F2,$4F,$3C,$3C,$91,$8C,$3C,$B8,$F1,$AD,$63,$D6,$F7
	.dc.b	$1F,$4B,$0E,$8F,$9A,$FC,$78,$F1,$79,$C7,$28,$C6,$3D,$CF,$B1,$5C
	.dc.b	$A7,$D2,$6C,$95,$C8,$EA,$E2,$79,$3E,$3C,$7A,$BE,$38,$CB,$F2,$23
	.dc.b	$2F,$63,$FA,$51,$4F,$8A,$63,$0F,$C1,$B7,$EC,$BF,$B3,$BC,$0F,$C4
	.dc.b	$38,$75,$E5,$FF,$81,$65,$39,$9F,$CB,$3C,$B0,$FC,$5B,$70,$CD,$FE
	.dc.b	$23,$E8,$3F,$AF,$56,$A3,$F1,$75,$BB,$FD,$1B,$5F,$CD,$57,$0B,$E7
	.dc.b	$1F,$7E,$75,$87,$92,$FB,$4F,$46,$3B,$72,$D2,$61,$E3,$7F,$15,$87
	.dc.b	$77,$C7,$3C,$A7,$47,$EA,$27,$53,$F2,$27,$58,$79,$AB,$53,$83,$99
	.dc.b	$E0,$8F,$DF,$57,$2B,$F5,$4E,$4D,$D7,$27,$07,$72,$C2,$79,$61,$F1
	.dc.b	$09,$87,$C0,$BF,$4E,$1E,$52,$61,$FC,$4A,$FE,$07,$07,$F2,$3F,$2B
	.dc.b	$FC,$29,$CD,$FC,$D5,$BA,$7F,$39,$6D,$FF,$34,$DA,$DF,$F2,$B7,$FC
	.dc.b	$98,$D8,$FE,$8D,$18,$E2,$FE,$33,$1F,$D1,$FC,$5E,$31,$FE,$16,$8F
	.dc.b	$E7,$51,$D3,$A3,$FD,$83,$1F,$D7,$FE,$2F,$04,$7F,$B1,$7D,$37,$93
	.dc.b	$07,$DA,$BF,$17,$C3,$6B,$FC,$FD,$B7,$F1,$9A,$DD,$6B,$EB,$25,$7E
	.dc.b	$96,$79,$C7,$F9,$9D,$B1,$9B,$4F,$C0,$C9,$F5,$5E,$92,$57,$F7,$7E
	.dc.b	$4F,$61,$B2,$3F,$DE,$5F,$93,$E0,$FF,$AB,$A3,$CF,$C5,$FE,$09,$F7
	.dc.b	$7E,$7C,$E6,$7F,$C3,$FF,$13,$F4,$BC,$5F,$F6,$5C,$7F,$85,$63,$06
	.dc.b	$7F,$18,$D5,$49,$FF,$0D,$3F,$6B,$C8,$F4,$C7,$80,$FF,$16,$E7,$F7
	.dc.b	$23,$31,$FE,$2C,$6B,$FE,$2B,$6B,$FE,$AB,$E6,$FF,$1E,$FC,$27,$E1
	.dc.b	$BC,$A3,$4F,$C4,$F9,$CC,$A9,$2F,$FC,$A4,$EB,$E5,$7F,$1F,$E8,$3D
	.dc.b	$9A,$4A,$F4,$70,$BF,$22,$7E,$29,$F9,$88,$C3,$63,$33,$C1,$AF,$FB
	.dc.b	$9F,$F1,$8F,$F8,$00,$FF,$B6,$7F,$50,$7C,$8F,$E8,$B9,$D1,$89,$D4
	.dc.b	$FE,$9B,$DB,$7E,$A1,$D3,$BF,$FB,$44,$63,$3F,$F5,$93,$6F,$F9,$A3
	.dc.b	$5F,$F4,$4D,$79,$1F,$E7,$FF,$8C,$FF,$37,$BF,$F5,$9B,$75,$FF,$46
	.dc.b	$36,$56,$FB,$F7,$85,$E9,$D1,$F3,$74,$97,$EF,$FC,$5B,$44,$70,$7C
	.dc.b	$3F,$5F,$F8,$0F,$C1,$B9,$A9,$27,$E9,$FE,$A3,$F8,$8D,$FF,$D8,$0D
	.dc.b	$DF,$FD,$D8,$D9,$FF,$10,$37,$B1,$F0,$AF,$9B,$FE,$23,$4F,$83,$FF
	.dc.b	$13,$BF,$EE,$86,$BF,$ED,$96,$E3,$F1,$A7,$8F,$1F,$3A,$FF,$71,$35
	.dc.b	$FE,$C8,$D7,$E7,$C7,$FB,$9D,$47,$91,$FF,$7B,$B5,$FF,$74,$36,$7F
	.dc.b	$DC,$FF,$8D,$3E,$F3,$F9,$3F,$E0,$4F,$99,$A4,$C3,$F6,$1A,$3D,$F7
	.dc.b	$C7,$7F,$2E,$6C,$E9,$AB,$2F,$F1,$A5,$71,$AE,$1C,$1F,$CD,$BF,$23
	.dc.b	$F9,$84,$F2,$3F,$9C,$FA,$CF,$95,$A4,$CB,$C7,$19,$6F,$DC,$7E,$87
	.dc.b	$F0,$67,$DB,$52,$4F,$F0,$9F,$11,$EB,$BE,$3B,$F5,$0D,$9D,$35,$F5
	.dc.b	$6A,$7F,$DF,$B1,$9E,$0F,$F2,$87,$CD,$3F,$43,$7C,$4F,$F5,$48,$C3
	.dc.b	$63,$37,$FC,$6E,$72,$F8,$3E,$5B,$FA,$CF,$CE,$7F,$3A,$8F,$A2,$D7
	.dc.b	$C6,$39,$51,$9C,$7E,$7F,$62,$FF,$B1,$B6,$FE,$DB,$5E,$77,$F9,$B9
	.dc.b	$DE,$BF,$DA,$BE,$7D,$F0,$34,$97,$F8,$4F,$C6,$7F,$C7,$F9,$7D,$E6
	.dc.b	$F0,$7F,$2F,$6C,$F9,$DB,$93,$F5,$2F,$ED,$73,$EC,$E2,$FA,$BF,$33
	.dc.b	$F5,$09,$1F,$D0,$47,$E3,$4F,$8F,$FD,$8B,$63,$FC,$94,$6A,$3F,$4D
	.dc.b	$19,$8F,$F2,$73,$5F,$E3,$33,$FB,$22,$3F,$D3,$FC,$FF,$F5,$07,$CA
	.dc.b	$7F,$83,$3F,$19,$FF,$55,$CF,$E9,$1A,$FC,$A6,$ED,$FE,$93,$FA,$E7
	.dc.b	$D5,$FF,$33,$AF,$9F,$F8,$CB,$CF,$FF,$0A,$B6,$3F,$CC,$5C,$08,$C4
	.dc.b	$7F,$9B,$23,$13,$FE,$1C,$F9,$5C,$E3,$FA,$E6,$F1,$FF,$5F,$6D,$B2
	.dc.b	$7E,$32,$F8,$3F,$F1,$3F,$E3,$3F,$E3,$9A,$FD,$FB,$5F,$C2,$B7,$1F
	.dc.b	$67,$D8,$F1,$9F,$E0,$BF,$8C,$BE,$9B,$FC,$5E,$D8,$F5,$B5,$F5,$35
	.dc.b	$FD,$61,$F9,$BF,$C6,$7F,$92,$EB,$FD,$DD,$B1,$9F,$97,$FF,$1F,$EA
	.dc.b	$FB,$6F,$9F,$7F,$DF,$7F,$19,$7D,$5F,$F9,$09,$B8,$7F,$1A,$6B,$FA
	.dc.b	$AB,$7D,$C7,$6E,$4F,$F6,$FF,$1F,$F9,$23,$3F,$F7,$F5,$FD,$46,$BF
	.dc.b	$4D,$AF,$AA,$5F,$8C,$7F,$93,$38,$7F,$90,$B7,$4F,$F0,$93,$6B,$EA
	.dc.b	$7E,$33,$FE,$9C,$7F,$97,$FC,$FF,$F8,$51,$BF,$1D,$FE,$5C,$8C,$47
	.dc.b	$A7,$EF,$9D,$BF,$E6,$0F,$C6,$3F,$A3,$4F,$F8,$7D,$AF,$F0,$76,$E8
	.dc.b	$6F,$E2,$EF,$B4,$FF,$34,$B6,$BF,$A0,$B5,$F9,$1F,$8C,$7F,$87,$DF
	.dc.b	$E8,$B7,$FD,$E9,$BF,$13,$FE,$72,$8C,$47,$CF,$F9,$EF,$EC,$CF,$F4
	.dc.b	$CE,$0F,$EB,$53,$FE,$F9,$8F,$D0,$C6,$0C,$FC,$5F,$2E,$2F,$E9,$6D
	.dc.b	$8F,$F1,$E3,$67,$F3,$DF,$83,$7C,$DF,$9E,$7B,$2F,$A8,$FF,$3A,$36
	.dc.b	$BF,$C9,$93,$A9,$FC,$6F,$E0,$FF,$E1,$78,$7E,$12,$E5,$FF,$38,$36
	.dc.b	$FF,$CE,$8D,$7F,$DC,$33,$C2,$7F,$CE,$D8,$3E,$67,$C2,$4A,$F0,$C7
	.dc.b	$EB,$74,$FE,$9D,$7F,$D5,$BE,$43,$FD,$15,$EE,$FF,$89,$2F,$CE,$D7
	.dc.b	$FA,$1C,$FF,$5C,$F9,$0F,$F4,$87,$E1,$7F,$EE,$59,$E6,$D5,$28,$C3
	.dc.b	$7D,$4F,$E5,$3E,$17,$9E,$FF,$5F,$A1,$A7,$FA,$77,$F1,$6F,$66,$3F
	.dc.b	$2B,$F2,$3F,$BA,$F9,$1B,$CB,$FE,$9D,$6E,$DE,$EF,$D4,$FF,$A0,$CF
	.dc.b	$FA,$85,$FE,$87,$5F,$C1,$34,$46,$23,$86,$DF,$EA,$18,$C4,$7B,$3E
	.dc.b	$13,$FD,$1E,$FC,$46,$DF,$FA,$B5,$B8,$7F,$7B,$D5,$F7,$4F,$B1,$FE
	.dc.b	$1A,$DC,$3F,$D2,$13,$A9,$C9,$FE,$BD,$F4,$FF,$B8,$5F,$D8,$FC,$9F
	.dc.b	$FD,$11,$7F,$8C,$6B,$F5,$9B,$97,$E7,$FE,$07,$F2,$53,$FD,$F0,$FE
	.dc.b	$F1,$D5,$FE,$4C,$8C,$CE,$E3,$FD,$99,$F8,$CF,$FB,$56,$BF,$B4,$7E
	.dc.b	$1F,$F9,$1A,$63,$FE,$17,$E6,$4E,$63,$0D,$A4,$BF,$10,$C3,$7A,$FF
	.dc.b	$D1,$91,$82,$78,$F8,$3F,$B7,$CF,$16,$C7,$F1,$4A,$FE,$7A,$78,$DC
	.dc.b	$38,$9F,$07,$C1,$76,$24,$7E,$D3,$EA,$DF,$7F,$FE,$E8,$24,$65,$FF
	.dc.b	$50,$65,$FC,$D3,$4F,$E1,$5F,$29,$FE,$21,$7F,$C0,$6D,$AF,$18,$70
	.dc.b	$F9,$9A,$FF,$54,$5F,$03,$FE,$98,$8E,$31,$8F,$81,$EB,$39,$9E,$4D
	.dc.b	$7B,$FD,$87,$F9,$DD,$34,$8F,$C3,$C1,$F6,$BF,$C8,$E3,$EB,$B5,$FA
	.dc.b	$F1,$AE,$1F,$5E,$DF,$7A,$B5,$DA,$DB,$FD,$29,$5F,$02,$31,$F3,$5A
	.dc.b	$9C,$E3,$FC,$61,$AF,$F7,$4C,$1F,$81,$BE,$D7,$C5,$18,$6C,$67,$1F
	.dc.b	$1B,$6B,$EF,$2F,$93,$D9,$27,$6F,$2F,$B4,$91,$C9,$B5,$FC,$B3,$F1
	.dc.b	$7F,$CB,$CF,$AD,$B1,$9C,$7E,$0C,$62,$3F,$C1,$BC,$6F,$C5,$FF,$4B
	.dc.b	$7F,$82,$8C,$3F,$35,$AE,$63,$73,$FC,$F3,$F1,$7F,$F6,$64,$FC,$B6
	.dc.b	$C6,$7E,$8F,$E8,$C6,$23,$FC,$3B,$E1,$7F,$0E,$F9,$FF,$B4,$DB,$FC
	.dc.b	$9B,$67,$F4,$ED,$78,$FE,$03,$FA,$29,$DB,$63,$6D,$77,$D5,$F8,$2F
	.dc.b	$C5,$BC,$51,$FD,$31,$AF,$F4,$26,$F2,$9B,$F8,$AB,$ED,$FF,$16,$DC
	.dc.b	$7F,$14,$DB,$FE,$BE,$D7,$DF,$F6,$5F,$EE,$59,$F6,$B6,$3D,$AD,$77
	.dc.b	$F3,$3F,$98,$F0,$5F,$41,$49,$7F,$99,$8C,$BF,$9B,$ED,$7B,$9F,$A0
	.dc.b	$6C,$66,$30,$7D,$0F,$63,$F9,$F4,$7D,$06,$E3,$F4,$1A,$FD,$86,$BF
	.dc.b	$D1,$DF,$73,$74,$E8,$49,$FB,$7E,$CF,$FB,$9D,$BF,$47,$F8,$08,$C5
	.dc.b	$7D,$E6,$BF,$86,$5F,$0F,$B2,$57,$47,$F1,$3E,$F7,$BA,$BF,$0E,$D7
	.dc.b	$F1,$ED,$7F,$AF,$35,$FF,$41,$7E,$01,$FE,$A9,$AF,$C8,$B6,$3F,$22
	.dc.b	$D7,$F3,$2D,$7F,$B4,$7E,$47,$5F,$F4,$63,$79,$CD,$31,$F0,$AF,$BE
	.dc.b	$FD,$0B,$70,$FD,$0B,$6F,$F4,$2D,$7F,$53,$F1,$0F,$B9,$FD,$4B,$6B
	.dc.b	$F5,$2D,$9F,$D4,$B5,$FD,$8F,$C3,$FF,$86,$67,$F5,$ED,$8F,$D7,$B5
	.dc.b	$DF,$BD,$FB,$6F,$8C,$FF,$32,$57,$ED,$9B,$3F,$B6,$6B,$FB,$E6,$BF
	.dc.b	$E0,$DF,$8F,$7E,$1F,$FC,$01,$B5,$FC,$01,$B3,$FC,$01,$AF,$F0,$8F
	.dc.b	$D5,$1F,$53,$FC,$21,$B5,$FC,$21,$B3,$FC,$21,$AF,$F1,$0F,$96,$7D
	.dc.b	$57,$F1,$06,$D7,$F1,$06,$CF,$F1,$06,$BF,$C6,$3E,$6B,$F8,$3F,$1F
	.dc.b	$FB,$23,$C8,$FE,$33,$18,$33,$F8,$17,$F9,$33,$0F,$E3,$6D,$BF,$E3
	.dc.b	$6D,$7F,$91,$B5,$FF,$19,$7C,$EB,$E2,$FF,$92,$B6,$BF,$92,$B6,$7F
	.dc.b	$92,$B5,$FF,$64,$FB,$3F,$ED,$DE,$BF,$EC,$CD,$AF,$EF,$ED,$9F,$F0
	.dc.b	$9B,$5F,$F0,$AF,$E1,$8F,$88,$FE,$DD,$08,$C5,$7F,$82,$23,$13,$A9
	.dc.b	$CC,$7F,$83,$3F,$0C,$FE,$38,$FF,$B7,$1A,$83,$86,$F8,$FF,$DF,$7E
	.dc.b	$45,$F4,$4F,$B6,$C3,$72,$FF,$0C,$CE,$A7,$26,$BA,$E4,$EA,$BF,$27
	.dc.b	$82,$FA,$D7,$7F,$C2,$DF,$88,$D2,$61,$DA,$32,$DE,$7E,$04,$70,$7F
	.dc.b	$80,$3E,$67,$FC,$06,$47,$07,$F8,$C3,$EA,$FF,$75,$F3,$DF,$E1,$B6
	.dc.b	$C6,$7C,$2F,$F7,$CB,$67,$FC,$1F,$F8,$3F,$F7,$BD,$7F,$CB,$8D,$7F
	.dc.b	$CB,$2D,$F7,$0E,$76,$FF,$3E,$FD,$17,$CD,$F9,$13,$49,$8F,$CC,$8F
	.dc.b	$E1,$18,$7D,$7A,$FF,$42,$E5,$C4,$95,$F4,$BF,$0B,$7C,$7F,$D9,$6C
	.dc.b	$66,$30,$67,$EB,$24,$71,$6E,$1F,$E5,$4F,$C2,$FF,$C9,$F3,$F8,$F6
	.dc.b	$C6,$7E,$2F,$F2,$B1,$88,$FF,$36,$FC,$77,$FB,$56,$BF,$2C,$D8,$FF
	.dc.b	$2F,$B5,$DF,$C7,$FE,$7A,$05,$24,$F2,$7F,$29,$F6,$DF,$9E,$AF,$CD
	.dc.b	$1F,$8B,$70,$FC,$76,$B9,$EE,$FF,$2C,$57,$F9,$73,$EF,$AF,$C7,$1A
	.dc.b	$48,$FF,$2D,$F1,$BF,$0A,$FC,$09,$23,$2F,$E1,$3F,$0B,$91,$C1,$B8
	.dc.b	$FE,$FB,$CA,$7C,$9F,$F0,$46,$D7,$F0,$46,$CE,$E7,$E2,$7B,$12,$BE
	.dc.b	$5B,$67,$F8,$4F,$B0,$E3,$FE,$1E,$DA,$FE,$1E,$D9,$DC,$FE,$F7,$F0
	.dc.b	$17,$E1,$34,$97,$FB,$C8,$CB,$FE,$96,$1E,$6B,$1F,$96,$D8,$D3,$70
	.dc.b	$FF,$57,$C7,$09,$FE,$19,$F0,$7F,$E7,$18,$FE,$12,$D9,$FF,$5A,$F5
	.dc.b	$A3,$11,$F7,$A3,$1F,$1F,$CB,$E3,$E5,$F2,$6D,$72,$FC,$1B,$F9,$E3
	.dc.b	$FC,$3D,$B6,$69,$23,$F9,$03,$FC,$96,$BF,$8E,$B6,$3A,$1F,$C9,$7E
	.dc.b	$B9,$F3,$34,$95,$FC,$92,$32,$DF,$B7,$E0,$47,$07,$F2,$AF,$C1,$FF
	.dc.b	$1C,$DE,$6F,$E6,$51,$8C,$3E,$2B,$5F,$88,$FE,$C5,$FE,$0C,$D7,$F9
	.dc.b	$73,$72,$37,$F0,$96,$3F,$CF,$DB,$87,$F3,$C6,$DF,$D4,$6B,$F4,$DF
	.dc.b	$F2,$3B,$FC,$55,$AF,$F3,$D6,$D9,$A7,$F8,$DD,$C7,$9D,$E1,$25,$78
	.dc.b	$7D,$4F,$21,$1C,$1E,$5E,$FF,$41,$1C,$1E,$9F,$03,$D8,$47,$07,$F6
	.dc.b	$9F,$C2,$BF,$D9,$5F,$1F,$FD,$8F,$E6,$FF,$47,$A7,$F6,$59,$CB,$FE
	.dc.b	$E6,$E9,$FE,$1F,$0F,$AE,$61,$C6,$73,$F3,$4C,$71,$87,$38,$C3,$CC
	.dc.b	$C4,$69,$61,$D6,$F0,$F8,$B1,$F2,$CE,$AF,$9C,$62,$75,$87,$9B,$1F
	.dc.b	$2D,$F7,$38,$63,$D3,$0E,$35,$E8,$8E,$F5,$C3,$0E,$95,$AB,$F3,$9C
	.dc.b	$F0,$EE,$FC,$7A,$E1,$87,$38,$D5,$71,$7C,$27,$19,$DB,$D2,$FB,$E1
	.dc.b	$E5,$3C,$78,$D9,$E7,$16,$7E,$05,$72,$9E,$B1,$C9,$D9,$DE,$B8,$61
	.dc.b	$E5,$9D,$1C,$A3,$33,$E3,$9D,$1F,$2E,$33,$1D,$A3,$2F,$C1,$8C,$E3
	.dc.b	$F8,$88,$FE,$AA,$7C,$02,$B9,$7E,$31,$8E,$67,$9B,$F9,$56,$C7,$23
	.dc.b	$F0,$AE,$1E,$11,$B3,$F0,$F1,$C3,$0F,$C2,$37,$0F,$CC,$F4,$7F,$68
	.dc.b	$9C,$D7,$02,$5F,$DD,$9D,$57,$E8,$2F,$94,$FE,$95,$AF,$92,$BF,$2F
	.dc.b	$5C,$23,$51,$D2,$FB,$C7,$53,$17,$D6,$6C,$7E,$C6,$B9,$3D,$4F,$C8
	.dc.b	$5F,$83,$03,$67,$71,$8A,$E7,$5C,$70,$FD,$35,$6A,$F9,$12,$BD,$35
	.dc.b	$C2,$F9,$3E,$08,$CB,$49,$7E,$C3,$AB,$C6,$BA,$3E,$F3,$78,$7B,$DA
	.dc.b	$7B,$CF,$E0,$FC,$3F,$C9,$F4,$FF,$3F,$36,$3F,$9A,$46,$38,$BF,$85
	.dc.b	$C7,$F3,$53,$04,$3F,$9D,$1C,$B9,$3F,$D1,$8F,$F4,$FF,$17,$04,$7F
	.dc.b	$A9,$7C,$6E,$C8,$E0,$FC,$95,$BE,$6D,$F1,$BF,$E9,$38,$C3,$63,$33
	.dc.b	$C1,$AF,$E9,$27,$96,$DB,$24,$6D,$F4,$11,$FE,$CC,$D7,$FC,$0F,$D0
	.dc.b	$F4,$FE,$3B,$63,$FB,$94,$60,$F9,$7D,$AF,$C2,$69,$27,$FB,$FB,$AF
	.dc.b	$C2,$24,$7F,$80,$9A,$FF,$85,$BF,$09,$7C,$A7,$D6,$6C,$7F,$A9,$1A
	.dc.b	$EE,$35,$E5,$48,$D3,$5F,$F1,$0F,$D4,$AF,$F0,$2D,$8F,$F5,$7B,$5D
	.dc.b	$C6,$BE,$16,$46,$9A,$FF,$8C,$BF,$09,$7E,$11,$FE,$C9,$ED,$46,$27
	.dc.b	$FC,$7C,$D7,$84,$FE,$4B,$F0,$97,$E1,$B4,$93,$FE,$C2,$8C,$BF,$ED
	.dc.b	$EF,$C2,$B9,$3F,$EE,$46,$BF,$ED,$1F,$87,$46,$21,$18,$BF,$F6,$F4
	.dc.b	$3F,$07,$FE,$40,$FF,$2C,$6B,$FE,$82,$6B,$FE,$DD,$6C,$BE,$CD,$D5
	.dc.b	$FE,$E9,$8C,$36,$33,$5F,$EF,$66,$BF,$CE,$B9,$DF,$85,$3E,$53,$FD
	.dc.b	$E5,$18,$6C,$66,$7F,$9C,$4E,$5B,$0F,$AC,$D9,$23,$2F,$D5,$FC,$27
	.dc.b	$F1,$31,$FE,$9E,$E1,$46,$23,$FC,$AC,$D7,$FC,$26,$DF,$09,$EE,$3E
	.dc.b	$2F,$F8,$5B,$6B,$FD,$F2,$D7,$FC,$E2,$D7,$E7,$47,$D2,$DF,$11,$F3
	.dc.b	$E3,$06,$66,$47,$F5,$73,$EC,$38,$7F,$AD,$7E,$12,$FA,$5F,$BA,$D8
	.dc.b	$E0,$D7,$71,$FC,$8B,$E7,$9F,$2D,$49,$3F,$D9,$72,$4E,$A3,$0F,$F8
	.dc.b	$89,$FF,$46,$B5,$FF,$46,$35,$FF,$77,$35,$43,$CD,$7E,$11,$FC,$D1
	.dc.b	$B3,$E1,$6B,$FE,$9A,$6B,$FE,$99,$5F,$63,$C4,$91,$C5,$FC,$5F,$93
	.dc.b	$FE,$BB,$8F,$C4,$B5,$FC,$93,$5F,$F7,$0F,$1A,$31,$01,$F1,$8F,$C1
	.dc.b	$DE,$CF,$D5,$8C,$36,$33,$3F,$E0,$F9,$CB,$E2,$F8,$8F,$29,$1C,$1F
	.dc.b	$47,$89,$1F,$D6,$1B,$19,$8C,$19,$F8,$09,$1C,$5A,$FF,$B0,$BF,$05
	.dc.b	$FE,$03,$B9,$A4,$8F,$EC,$BF,$82,$FF,$5C,$7F,$D9,$2D,$7F,$D9,$0D
	.dc.b	$7E,$24,$A3,$11,$F1,$FF,$05,$7B,$BF,$C4,$1B,$1F,$9B,$6B,$FE,$33
	.dc.b	$6B,$FC,$4F,$E2,$FF,$4C,$47,$F4,$5D,$28,$FE,$21,$D6,$F7,$69,$2B
	.dc.b	$FC,$7D,$E1,$7F,$A3,$23,$F5,$5F,$0E,$F9,$87,$CB,$61,$B7,$FE,$45
	.dc.b	$9D,$4E,$4E,$3F,$18,$F9,$2A,$4A,$E1,$19,$79,$7C,$D7,$FB,$11,$BE
	.dc.b	$BF,$8A,$31,$1F,$E5,$CC,$11,$88,$FC,$80,$87,$CD,$E6,$3F,$29,$18
	.dc.b	$AF,$6C,$E6,$78,$36,$5F,$0C,$F4,$7E,$1B,$6B,$E0,$35,$F2,$35,$F1
	.dc.b	$CB,$E7,$5F,$0B,$F3,$1B,$3F,$25,$AF,$B5,$AF,$2F,$9F,$FF,$60,$3F
	.dc.b	$3D,$AF,$C3,$6B,$BF,$86,$FB,$3F,$34,$29,$23,$EC,$39,$7E,$D7,$D0
	.dc.b	$7F,$97,$1B,$F0,$DF,$82,$8C,$47,$D1,$6B,$EE,$FA,$1F,$C6,$3F,$83
	.dc.b	$6B,$F5,$DA,$EF,$97,$F1,$BE,$B7,$F5,$E8,$F8,$ED,$7E,$BB,$5F,$B8
	.dc.b	$DD,$5F,$78,$F9,$5F,$CB,$B6,$BF,$28,$D7,$F1,$0D,$7E,$AF,$C8,$3E
	.dc.b	$CB,$F3,$ED,$8F,$CE,$35,$FC,$83,$5F,$BB,$F4,$CF,$8F,$FD,$3B,$63
	.dc.b	$F4,$8D,$7F,$30,$D7,$8F,$C4,$70,$23,$C1,$FD,$8F,$D3,$FF,$83,$33
	.dc.b	$34,$9E,$7F,$ED,$BE,$37,$FC,$43,$7F,$A5,$6B,$F9,$26,$BF,$A2,$6A
	.dc.b	$B7,$DD,$FF,$4F,$8F,$F3,$7B,$5F,$F3,$43,$5F,$D7,$C7,$0F,$4F,$FC
	.dc.b	$E5,$E9,$7F,$84,$67,$FC,$DC,$D7,$6D,$7F,$A3,$4F,$F4,$9F,$77,$F8
	.dc.b	$7F,$C2,$E6,$7F,$74,$DD,$0D,$35,$4F,$BA,$CD,$7F,$A1,$9A,$FF,$9F
	.dc.b	$DA,$FA,$1A,$F8,$BE,$E7,$F9,$9C,$9A,$4C,$7F,$91,$FD,$83,$EE,$69
	.dc.b	$2F,$F9,$14,$65,$FE,$49,$F4,$35,$FE,$30,$D7,$F8,$2B,$5F,$E2,$AD
	.dc.b	$7E,$A7,$C1,$7F,$A6,$1F,$9E,$D7,$E4,$B5,$DF,$C9,$7F,$96,$3D,$77
	.dc.b	$CC,$52,$4F,$F8,$96,$32,$FF,$AB,$3E,$E1,$D3,$FE,$AB,$6C,$66,$30
	.dc.b	$7F,$94,$3E,$BF,$F3,$51,$FD,$FF,$9D,$18,$84,$63,$AB,$81,$C7,$BE
	.dc.b	$4F,$F9,$5A,$BF,$82,$BC,$DA,$F2,$5F,$35,$E1,$24,$71,$7F,$D8,$1F
	.dc.b	$65,$23,$FD,$80,$D7,$FC,$D5,$F6,$1C,$9F,$F3,$2B,$5C,$CF,$F9,$B2
	.dc.b	$31,$1F,$D3,$07,$DC,$7F,$83,$BE,$DB,$FC,$9C,$D8,$D1,$AD,$93,$A6
	.dc.b	$C6,$A7,$FB,$A7,$AD,$CB,$D8,$FF,$48,$CE,$5A,$F8,$67,$6F,$FA,$53
	.dc.b	$E7,$7F,$D2,$A9,$A4,$7F,$DB,$DF,$37,$FE,$90,$F6,$BF,$C6,$99,$1F
	.dc.b	$D8,$C6,$67,$CF,$1F,$6F,$E1,$DE,$7E,$E7,$27,$F0,$7A,$E1,$79,$E3
	.dc.b	$7E,$AB,$FE,$F4,$C7,$B9,$F8,$57,$17,$CB,$BF,$FD,$F5,$F2,$3D,$70
	.dc.b	$E5,$5A,$6F,$07,$F9,$97,$1E,$AD,$9F,$F1,$FE,$3F,$E9,$E6,$E3,$FE
	.dc.b	$A7,$5F,$09,$D6,$7C,$4D,$9F,$CA,$CF,$0C,$3C,$33,$9C,$3F,$CD,$FE
	.dc.b	$8D,$D8,$E6,$FA,$E3,$EE,$C7,$F3,$F5,$A8,$C7,$8A,$FA,$BF,$F5,$C4
	.dc.b	$61,$B5,$F0,$FE,$E1,$F8,$67,$96,$F8,$79,$3E,$7E,$1F,$EA,$DE,$AD
	.dc.b	$F4,$3F,$BE,$67,$CE,$B1,$3E,$DE,$97,$A1,$FE,$75,$AF,$34,$63,$C9
	.dc.b	$E9,$F7,$1F,$EF,$F9,$CD,$6F,$6F,$AD,$F8,$D7,$F9,$02,$7D,$DA,$23
	.dc.b	$13,$F8,$4C,$7A,$56,$A7,$3C,$DF,$89,$B7,$DC,$F7,$C3,$F5,$91,$8F
	.dc.b	$9C,$FF,$51,$37,$0F,$D5,$B6,$77,$F8,$D3,$E0,$FE,$23,$6B,$CB,$87
	.dc.b	$5C,$7C,$13,$9B,$FC,$CF,$13,$E9,$FE,$CC,$61,$B5,$E7,$FB,$A7,$D3
	.dc.b	$FD,$39,$CE,$5F,$D6,$3E,$FF,$FD,$BA,$F9,$9B,$87,$B1,$AF,$F6,$66
	.dc.b	$BF,$D8,$3C,$97,$93,$F7,$5B,$5F,$39,$FF,$6B,$36,$3F,$05,$E4,$7D
	.dc.b	$5C,$FE,$63,$67,$FB,$5F,$E3,$9F,$EB,$17,$EB,$C6,$70,$FE,$FD,$F8
	.dc.b	$0F,$FB,$2A,$7E,$6B,$63,$E9,$B5,$FE,$FA,$D7,$FD,$33,$E1,$7E,$D6
	.dc.b	$6F,$A3,$F9,$7F,$D5,$B1,$5F,$ED,$8A,$BF,$24,$FC,$73,$FB,$CE,$1F
	.dc.b	$A7,$8C,$7B,$7F,$9F,$EC,$7C,$5F,$ED,$67,$35,$BE,$AF,$E3,$0F,$FB
	.dc.b	$BD,$B9,$7F,$35,$9F,$F0,$D4,$FF,$BA,$DA,$FF,$BD,$7F,$1C,$FE,$C6
	.dc.b	$F8,$5B,$3F,$84,$FC,$AB,$E8,$BF,$C7,$A9,$3A,$E2,$FE,$45,$0F,$C7
	.dc.b	$7F,$C9,$18,$7A,$DB,$7F,$8E,$FB,$37,$CF,$7A,$E7,$38,$72,$FC,$77
	.dc.b	$FC,$91,$87,$F4,$8C,$51,$8C,$7D,$8D,$7F,$D3,$B1,$AF,$67,$DD,$F8
	.dc.b	$EB,$F1,$7F,$90,$DC,$3F,$88,$FC,$67,$E8,$72,$F4,$CE,$71,$F7,$7E
	.dc.b	$3B,$FE,$F4,$BF,$A8,$D8,$FD,$3F,$C7,$3E,$F7,$EA,$4E,$70,$F8,$7E
	.dc.b	$D7,$FA,$97,$0F,$53,$70,$FA,$8D,$7D,$2D,$7F,$DC,$7E,$FF,$F2,$77
	.dc.b	$E8,$36,$BF,$A8,$78,$CF,$90,$FE,$17,$87,$0F,$63,$F0,$DE,$3F,$D8
	.dc.b	$BF,$AA,$D9,$FF,$5F,$FC,$8B,$EF,$77,$E3,$7F,$41,$D1,$F8,$07,$F3
	.dc.b	$0B,$FE,$F3,$18,$F5,$FF,$B0,$3F,$E0,$19,$D3,$FD,$F9,$AF,$F8,$17
	.dc.b	$F3,$87,$D7,$75,$6D,$FF,$45,$97,$E0,$5F,$C3,$23,$13,$A9,$FE,$DA
	.dc.b	$1F,$D3,$B8,$6E,$BE,$4A,$FF,$29,$57,$2F,$13,$D9,$1F,$C3,$AC,$78
	.dc.b	$3E,$B3,$D6,$FF,$B9,$A3,$4E,$6B,$E0,$E1,$F1,$DF,$2E,$BD,$3E,$67
	.dc.b	$D8,$F6,$F9,$D9,$3F,$D3,$3C,$9F,$66,$1E,$AC,$7F,$E0,$04,$FC,$1D
	.dc.b	$BC,$F3,$F2,$7E,$43,$C6,$F9,$A3,$2F,$CF,$8F,$F0,$07,$D8,$FF,$1B
	.dc.b	$8F,$F0,$8E,$08,$C6,$FF,$5A,$FA,$BF,$E1,$38,$CE,$1F,$0B,$C1,$7D
	.dc.b	$1F,$CF,$AF,$DD,$B6,$74,$BF,$05,$FE,$24,$E7,$25,$BF,$15,$7D,$67
	.dc.b	$F1,$76,$E1,$FC,$76,$F8,$56,$EB,$F8,$E7,$C2,$FE,$97,$2F,$E1,$6D
	.dc.b	$BF,$33,$FD,$65,$BF,$41,$FC,$8F,$F1,$6F,$E8,$79,$FF,$15,$6D,$F2
	.dc.b	$9D,$7D,$7F,$F8,$77,$EB,$3F,$3B,$F6,$D7,$84,$FF,$87,$78,$7F,$93
	.dc.b	$E5,$FE,$3C,$AD,$E7,$FE,$1B,$E2,$F9,$DE,$5E,$CE,$58,$59,$D7,$C5
	.dc.b	$FC,$5A,$37,$1F,$E3,$5A,$FF,$91,$63,$E1,$FC,$CB,$F2,$5F,$E7,$CD
	.dc.b	$C3,$FA,$23,$FE,$24,$AF,$E5,$EE,$74,$FE,$69,$F8,$DF,$A2,$7F,$99
	.dc.b	$B6,$BE,$F5,$6E,$73,$EB,$7F,$4B,$FC,$15,$F2,$F4,$C7,$7D,$C3,$FA
	.dc.b	$A7,$CE,$FF,$98,$1B,$E6,$E7,$4F,$E8,$93,$97,$F3,$DF,$8C,$FE,$C0
	.dc.b	$FE,$19,$F8,$37,$FA,$82,$FF,$80,$78,$29,$FD,$AF,$D3,$FE,$82,$FF
	.dc.b	$81,$B6,$4D,$35,$7F,$9C,$BF,$DD,$D2,$30,$FE,$FA,$B8,$CF,$F0,$F9
	.dc.b	$CF,$73,$F1,$BF,$D1,$DF,$E6,$9B,$19,$DF,$FB,$84,$63,$EE,$FF,$B6
	.dc.b	$F0,$FF,$60,$6E,$3F,$AB,$9D,$3F,$8B,$FC,$82,$3F,$89,$E7,$FE,$9D
	.dc.b	$7F,$B3,$9F,$C8,$7E,$13,$2B,$51,$73,$FC,$97,$EA,$FF,$81,$9F,$E3
	.dc.b	$3D,$A8,$C4,$FF,$B3,$31,$FF,$0C,$63,$FE,$95,$AD,$7C,$BF,$F9,$E7
	.dc.b	$F1,$2F,$E5,$97,$F2,$3B,$51,$88,$FF,$6C,$35,$FF,$6B,$B6,$3F,$87
	.dc.b	$7C,$5F,$E9,$CE,$9E,$85,$D3,$FD,$00,$7E,$B3,$F0,$3F,$F4,$63,$4F
	.dc.b	$F0,$F7,$E1,$3F,$59,$C6,$3F,$E0,$DF,$4B,$FD,$B7,$1F,$E2,$36,$C7
	.dc.b	$FA,$BA,$37,$AF,$F8,$42,$72,$FF,$84,$BE,$43,$FA,$1C,$FF,$1D,$6D
	.dc.b	$7F,$16,$7F,$BB,$46,$6B,$FC,$59,$F8,$23,$EE,$E9,$FE,$2F,$AE,$3D
	.dc.b	$09,$D5,$FF,$8D,$FF,$05,$FF,$09,$63,$FC,$AD,$AF,$F2,$46,$D1,$A3
	.dc.b	$97,$FC,$78,$DA,$E9,$E4,$62,$7F,$D4,$1C,$CE,$7A,$7F,$90,$FE,$2F
	.dc.b	$FD,$3F,$97,$F9,$EA,$75,$87,$F9,$2E,$7F,$A2,$37,$A3,$8B,$4F,$F2
	.dc.b	$87,$E0,$FF,$C8,$9F,$F1,$83,$6F,$FC,$94,$D7,$FC,$5E,$D7,$FA,$53
	.dc.b	$FE,$3A,$8B,$F3,$FF,$E5,$78,$E5,$C5,$FE,$46,$75,$5F,$D2,$1F,$E2
	.dc.b	$35,$E0,$F1,$7F,$CA,$D1,$89,$F0,$C6,$3C,$1E,$87,$56,$D7,$AB,$E8
	.dc.b	$BC,$07,$87,$E8,$EC,$7B,$BC,$4F,$21,$8C,$3F,$CB,$98,$FF,$13,$9F
	.dc.b	$87,$CD,$E2,$23,$F1,$1F,$2F,$B1,$ED,$3D,$CD,$8F,$F0,$E7,$E1,$0F
	.dc.b	$B6,$78,$BF,$E1,$0F,$A8,$FF,$54,$E5,$FE,$12,$AD,$4E,$7E,$0B,$F8
	.dc.b	$DC,$7F,$2B,$AD,$D2,$36,$A3,$31,$86,$DF,$82,$75,$5D,$3C,$B6,$3D
	.dc.b	$E7,$C5,$5B,$BE,$38,$E6,$7C,$13,$9C,$39,$5E,$71,$F0,$47,$A2,$75
	.dc.b	$5D,$2F,$58,$F8,$23,$C6,$64,$A7,$68,$E9,$3C,$A3,$E0,$4F,$1C,$7C
	.dc.b	$B1,$C3,$0F,$0B,$CE,$3E,$1C,$EF,$0F,$26,$1D,$09,$5C,$67,$4F,$19
	.dc.b	$DB,$D7,$0F,$16,$1F,$4E,$73,$87,$88,$8F,$C4,$AE,$18,$76,$9C,$C7
	.dc.b	$DF,$C3,$BB,$EF,$36,$48,$F4,$E3,$C6,$FB,$92,$79,$46,$5F,$94,$7F
	.dc.b	$62,$3A,$17,$0F,$C3,$B7,$1F,$C2,$39,$F8,$37,$89,$FE,$7E,$A4,$62
	.dc.b	$7A,$1F,$91,$70,$F1,$9E,$11,$BC,$7F,$26,$DC,$7F,$36,$DA,$FD,$07
	.dc.b	$D8,$FF,$78,$AD,$DF,$E2,$EB,$55,$DE,$3F,$1D,$5B,$BE,$71,$B2,$4F
	.dc.b	$E0,$E7,$4F,$EB,$6B,$9E,$1E,$CA,$DE,$1F,$A8,$BD,$DF,$73,$F1,$97
	.dc.b	$C0,$E3,$1A,$BE,$53,$AB,$E7,$87,$EE,$2B,$57,$C6,$73,$5F,$00,$A7
	.dc.b	$7C,$3E,$0B,$E9,$33,$87,$68,$C9,$E7,$8C,$D7,$57,$81,$E3,$24,$70
	.dc.b	$C3,$C9,$1C,$2F,$AE,$1F,$1B,$B9,$C0,$F5,$BF,$C5,$7C,$FF,$CD,$DB
	.dc.b	$B7,$F3,$E6,$E3,$FC,$E7,$DE,$B8,$72,$7F,$D8,$ED,$8F,$E9,$71,$8C
	.dc.b	$0D,$F8,$3E,$30,$69,$8F,$97,$95,$FD,$59,$AF,$F3,$8F,$84,$E3,$F8
	.dc.b	$3C,$8F,$EB,$6F,$4D,$A4,$7F,$5C,$7C,$44,$7F,$3B,$CB,$FE,$4E,$EE
	.dc.b	$F6,$B6,$3F,$D8,$FB,$FF,$19,$8C,$42,$31,$DA,$FC,$23,$8B,$67,$AB
	.dc.b	$5F,$DB,$57,$F9,$72,$7F,$DD,$AD,$8F,$D7,$EE,$F0,$3F,$D6,$11,$FD
	.dc.b	$EF,$E0,$FE,$7C,$62,$3F,$C0,$1F,$21,$F5,$3E,$2B,$FC,$88,$DF,$10
	.dc.b	$D3,$EB,$F9,$0F,$80,$BF,$27,$30,$FF,$0E,$3F,$73,$CB,$7B,$54,$FB
	.dc.b	$AE,$27,$CE,$D7,$F8,$96,$BF,$E2,$79,$FC,$1F,$9B,$FE,$D5,$9F,$97
	.dc.b	$D9,$FE,$35,$E4,$FC,$4C,$E6,$7F,$C8,$5F,$86,$7E,$33,$CF,$34,$91
	.dc.b	$FE,$45,$F3,$7F,$23,$E8,$7F,$A2,$E3,$FC,$9B,$E5,$60,$FC,$BE,$5F
	.dc.b	$E6,$7A,$BC,$5F,$E2,$BA,$FF,$2B,$B5,$FF,$27,$37,$A7,$FC,$FD,$F6
	.dc.b	$5F,$A0,$EB,$7D,$67,$E5,$EF,$FD,$88,$D7,$FD,$E6,$DA,$FE,$C1,$2F
	.dc.b	$8B,$FF,$67,$D7,$FB,$2D,$AF,$F3,$98,$D7,$C2,$7F,$30,$8F,$E6,$B3
	.dc.b	$A7,$FC,$F0,$D7,$FD,$07,$FA,$AB,$D7,$FD,$AF,$B6,$8C,$4F,$D0,$11
	.dc.b	$99,$F6,$65,$FC,$06,$7F,$DB,$DF,$8A,$7F,$4B,$CF,$D2,$DD,$D3,$FD
	.dc.b	$22,$30,$FF,$4A,$A7,$D5,$F4,$24,$7C,$67,$E5,$4F,$F5,$2B,$FF,$02
	.dc.b	$74,$7F,$AC,$F9,$3F,$85,$FB,$DF,$80,$C3,$D1,$39,$6F,$C7,$FF,$0A
	.dc.b	$6F,$97,$FC,$4A,$31,$1F,$B5,$D3,$CF,$49,$CE,$3E,$07,$FC,$5A,$FF
	.dc.b	$85,$6B,$FD,$95,$E3,$FF,$1E,$69,$FE,$C9,$71,$F8,$01,$A7,$D7,$84
	.dc.b	$65,$FB,$4D,$7F,$B5,$B5,$FE,$F0,$FE,$1F,$0F,$F6,$DB,$63,$E7,$B5
	.dc.b	$7E,$0E,$F5,$FF,$98,$FD,$97,$FA,$2E,$7E,$3B,$5F,$84,$33,$46,$72
	.dc.b	$FA,$3F,$12,$F8,$1F,$E7,$6D,$BF,$A2,$D7,$9D,$A3,$15,$CB,$E2,$FF
	.dc.b	$A1,$FE,$3D,$F9,$C8,$FB,$AD,$D7,$E9,$B7,$0F,$27,$AA,$F8,$3F,$E9
	.dc.b	$6D,$8F,$C0,$B5,$F0,$B6,$FF,$12,$BE,$B3,$F6,$DF,$52,$FB,$5A,$4B
	.dc.b	$FC,$EF,$D6,$FE,$48,$8E,$0F,$D3,$76,$BF,$58,$7D,$97,$EB,$1B,$3A
	.dc.b	$6B,$F3,$A7,$84,$FE,$7B,$D1,$FE,$CB,$F2,$8F,$66,$92,$FF,$3B,$F0
	.dc.b	$9F,$95,$6F,$67,$F6,$C8,$C4,$7D,$06,$F0,$7E,$A1,$FF,$6C,$47,$E9
	.dc.b	$BE,$63,$E0,$37,$6F,$EE,$B1,$88,$FB,$3F,$82,$FF,$1A,$7F,$5A,$D7
	.dc.b	$3E,$27,$F7,$C8,$C6,$1F,$B6,$5F,$15,$FB,$AF,$8B,$97,$FC,$21,$FD
	.dc.b	$47,$E0,$AF,$5E,$9F,$A5,$71,$3F,$C7,$5A,$FF,$93,$9A,$FE,$93,$E0
	.dc.b	$FF,$5E,$47,$F8,$D7,$D9,$BE,$AD,$F8,$4F,$ED,$FE,$53,$F7,$7D,$FF
	.dc.b	$C5,$AD,$18,$7F,$8C,$B5,$FD,$BE,$FF,$BF,$FC,$6D,$F2,$5F,$BB,$6C
	.dc.b	$FF,$33,$6B,$FE,$64,$6B,$FE,$71,$FC,$C3,$F8,$C3,$FB,$00,$E3,$8B
	.dc.b	$F8,$73,$71,$F3,$53,$EE,$FF,$88,$4F,$F7,$5B,$FE,$70,$72,$F8,$1C
	.dc.b	$9E,$7F,$A0,$F9,$FF,$88,$FC,$D2,$4F,$CD,$7F,$C6,$78,$7C,$68,$FE
	.dc.b	$56,$D7,$F7,$6D,$CB,$FD,$6D,$F0,$AF,$A3,$C6,$9F,$CA,$BF,$1C,$FE
	.dc.b	$8D,$87,$F0,$26,$DF,$F3,$A6,$CF,$F4,$6B,$7D,$2F,$F2,$DA,$FE,$7F
	.dc.b	$F2,$CF,$A6,$FF,$23,$B7,$1F,$F1,$0B,$5F,$F0,$5B,$5F,$B9,$E7,$BF
	.dc.b	$1C,$FE,$A2,$D8,$FF,$29,$C7,$F5,$4C,$5F,$54,$BF,$F2,$BF,$27,$F8
	.dc.b	$3F,$2F,$EA,$2D,$D3,$FA,$3B,$71,$60,$FA,$CF,$EB,$15,$FD,$33,$F1
	.dc.b	$CF,$E9,$F5,$E4,$6B,$9F,$6B,$FC,$C7,$18,$8F,$E6,$5F,$8E,$3F,$02
	.dc.b	$FF,$34,$B6,$3F,$AF,$B5,$FE,$73,$8B,$F0,$8F,$E8,$11,$FD,$93,$83
	.dc.b	$F6,$6D,$F1,$3F,$CE,$71,$88,$E3,$F6,$AF,$A1,$FF,$3B,$FE,$4B,$FD
	.dc.b	$16,$7E,$5B,$5F,$EC,$ED,$7F,$A4,$36,$9F,$8A,$75,$E2,$FE,$F2,$DE
	.dc.b	$0F,$ED,$ED,$BF,$93,$F9,$F7,$F8,$11,$FE,$F7,$5F,$E5,$46,$FA,$3F
	.dc.b	$E8,$A8,$C4,$7C,$0F,$BA,$FE,$50,$FF,$86,$7E,$69,$F7,$7F,$ED,$A8
	.dc.b	$FD,$F4,$60,$FF,$35,$C6,$63,$FC,$2F,$19,$7F,$CF,$FF,$84,$FF,$B7
	.dc.b	$9E,$8D,$7F,$B3,$C2,$31,$7F,$E8,$15,$F4,$DF,$E8,$6F,$8B,$7D,$85
	.dc.b	$26,$1E,$78,$7D,$B7,$FA,$33,$F4,$3F,$F4,$14,$FA,$1A,$F9,$9A,$B2
	.dc.b	$46,$39,$DF,$33,$D2,$BF,$C7,$91,$8B,$FF,$41,$4E,$67,$D7,$F8,$53
	.dc.b	$E0,$FF,$D3,$BF,$20,$FA,$3A,$4A,$F9,$DF,$85,$7F,$40,$E7,$F9,$2F
	.dc.b	$F4,$58,$F9,$ED,$73,$E8,$FF,$A6,$A3,$11,$E9,$FB,$7F,$EE,$8F,$D9
	.dc.b	$6B,$F3,$9B,$C8,$BF,$F2,$6F,$AD,$FB,$8F,$8A,$FF,$67,$B6,$FF,$A0
	.dc.b	$F9,$29,$CB,$63,$F0,$DF,$33,$F8,$87,$ED,$FD,$7F,$F2,$B6,$FA,$9F
	.dc.b	$EB,$E8,$C6,$5F,$C8,$3E,$DD,$F8,$87,$FA,$AD,$B7,$F3,$9A,$FE,$D1
	.dc.b	$B3,$FE,$C5,$FC,$2F,$FD,$71,$3F,$E6,$FF,$9D,$7C,$6F,$F3,$F8,$46
	.dc.b	$3E,$67,$FC,$2E,$7F,$A3,$7F,$05,$FC,$7F,$0F,$F9,$BA,$51,$8A,$E2
	.dc.b	$D8,$CF,$D8,$71,$E8,$95,$FD,$84,$E5,$5F,$81,$70,$72,$8F,$F4,$75
	.dc.b	$76,$6C,$FF,$A2,$27,$FC,$93,$3E,$2F,$9C,$F3,$1F,$EE,$6F,$C2,$FF
	.dc.b	$CB,$9A,$FF,$B9,$5B,$97,$EC,$1B,$7F,$9B,$6D,$FC,$4F,$93,$FF,$37
	.dc.b	$7C,$C3,$EC,$FF,$DD,$97,$E6,$8C,$63,$FE,$A3,$6C,$66,$76,$FB,$E1
	.dc.b	$F6,$8E,$9E,$0D,$9E,$8D,$78,$35,$F8,$BF,$67,$FE,$7B,$4D,$23,$FE
	.dc.b	$B7,$EA,$FE,$99,$AF,$CC,$D3,$F2,$9E,$DF,$FA,$2E,$BE,$D5,$69,$E1
	.dc.b	$19,$7E,$27,$75,$F1,$9C,$1F,$EC,$EC,$7E,$BC,$62,$75,$39,$AF,$F6
	.dc.b	$47,$E1,$7F,$E9,$DB,$F1,$36,$33,$F2,$5D,$F3,$4E,$67,$FC,$4D,$18
	.dc.b	$E0,$D4,$FD,$3B,$E7,$18,$33,$EA,$48,$EE,$DB,$FC,$86,$6F,$C2,$FF
	.dc.b	$CC,$15,$EA,$6C,$E9,$AF,$F3,$A6,$BF,$E0,$9E,$C7,$DB,$FC,$7F,$29
	.dc.b	$18,$BF,$CD,$C6,$27,$53,$9A,$F1,$FE,$17,$FE,$85,$BF,$96,$D8,$CF
	.dc.b	$47,$D1,$8C,$47,$F9,$27,$99,$F8,$37,$FB,$72,$FF,$53,$18,$F3,$BE
	.dc.b	$D3,$63,$33,$BA,$FD,$96,$6F,$BC,$7C,$87,$AD,$B7,$EA,$6D,$70,$6B
	.dc.b	$FD,$7B,$F0,$AF,$2B,$FD,$71,$AF,$F4,$76,$F0,$1B,$E1,$BE,$EB,$F1
	.dc.b	$CD,$C3,$F1,$AD,$BF,$B2,$D7,$FC,$D5,$93,$EC,$3F,$C1,$55,$F4,$1B
	.dc.b	$3E,$E6,$BC,$1A,$FF,$72,$FC,$27,$D6,$DD,$3F,$41,$18,$8F,$EC,$0D
	.dc.b	$CF,$FD,$A5,$7F,$A2,$FA,$55,$7E,$9D,$B1,$98,$FC,$37,$E1,$BF,$9D
	.dc.b	$7F,$CA,$4D,$7F,$C7,$6D,$C8,$DF,$6D,$F7,$5F,$8F,$6D,$FE,$39,$B7
	.dc.b	$F7,$1A,$FC,$6E,$17,$DC,$7E,$5D,$B5,$F9,$66,$CF,$E1,$1A,$F1,$F9
	.dc.b	$EE,$04,$9E,$0F,$E8,$7E,$11,$F7,$74,$93,$FA,$28,$CB,$FA,$4F,$C2
	.dc.b	$3F,$D6,$4D,$F9,$DF,$D5,$46,$30,$FC,$AB,$5F,$BF,$F8,$ED,$7F,$CF
	.dc.b	$4D,$7F,$CC,$6D,$DC,$DF,$8F,$7D,$47,$ED,$DB,$8F,$ED,$9B,$7F,$A4
	.dc.b	$6B,$F8,$EF,$67,$FC,$4F,$3F,$BC,$6C,$7E,$A9,$AE,$FE,$03,$F8,$27
	.dc.b	$D3,$3F,$1B,$A4,$BF,$E0,$B1,$97,$F8,$3F,$CB,$7F,$7E,$6F,$B5,$FC
	.dc.b	$36,$31,$87,$EE,$1A,$FE,$8B,$E6,$1F,$83,$7F,$13,$6D,$FF,$12,$6C
	.dc.b	$FF,$00,$6B,$FD,$CF,$F0,$3F,$C3,$BF,$ED,$56,$BF,$9C,$6A,$69,$FB
	.dc.b	$37,$87,$83,$FD,$25,$D3,$63,$51,$CF,$E5,$B0,$4B,$FF,$00,$C7,$58
	.dc.b	$E1,$F4,$F2,$73,$48,$C5,$7F,$82,$21,$F7,$9D,$C9,$5C,$1F,$E4,$7E
	.dc.b	$8F,$8B,$C8,$FF,$0F,$E5,$82,$57,$6F,$B0,$F4,$11,$C1,$E6,$FC,$19
	.dc.b	$A7,$BD,$B1,$98,$C1,$FD,$AF,$E5,$1F,$42,$FC,$03,$0D,$AF,$F1,$0C
	.dc.b	$EA,$72,$6B,$E0,$24,$EA,$91,$8D,$FF,$C5,$97,$FD,$4B,$D1,$F0,$7A
	.dc.b	$5F,$CC,$A3,$18,$76,$8C,$0F,$87,$D9,$23,$4F,$F7,$5F,$BE,$FF,$40
	.dc.b	$47,$F8,$2D,$B8,$7F,$8E,$F8,$51,$88,$FE,$F2,$D7,$F9,$2F,$D7,$B7
	.dc.b	$7D,$D7,$F8,$3E,$30,$D8,$CE,$1F,$E4,$49,$CB,$E1,$FA,$0F,$21,$1C
	.dc.b	$1F,$2B,$F0,$6F,$ED,$D8,$7F,$BD,$1A,$FF,$36,$D5,$18,$84,$63,$D9
	.dc.b	$FB,$13,$FE,$AA,$AE,$9F,$2B,$82,$3F,$63,$F0,$6F,$C8,$B7,$A7,$F0
	.dc.b	$11,$8C,$7E,$AB,$5F,$E8,$E7,$81,$FF,$22,$CF,$3F,$86,$E2,$47,$F4
	.dc.b	$BF,$55,$FE,$76,$6F,$0F,$EB,$E3,$17,$FE,$6A,$6B,$FB,$DF,$BF,$FD
	.dc.b	$91,$2B,$F5,$6F,$ED,$7E,$52,$47,$EB,$9B,$3F,$8C,$FB,$BF,$B4,$FC
	.dc.b	$66,$EE,$69,$23,$F4,$0F,$F6,$27,$91,$83,$F1,$9F,$3B,$2F,$F1,$8D
	.dc.b	$8F,$C6,$72,$BE,$C3,$F2,$44,$AF,$C9,$3F,$95,$F8,$B9,$1F,$95,$6D
	.dc.b	$7F,$A3,$F9,$5F,$7D,$F7,$DB,$CB,$F9,$18,$C1,$F8,$3E,$07,$CA,$FE
	.dc.b	$BB,$1F,$F3,$53,$70,$FD,$03,$5F,$F2,$BB,$5F,$C7,$7E,$01,$FC,$11
	.dc.b	$FD,$AB,$5F,$C0,$EE,$8C,$47,$0F,$07,$FC,$E5,$7F,$A6,$FB,$8B,$D8
	.dc.b	$69,$23,$F8,$4F,$C1,$3F,$16,$FE,$20,$D8,$CC,$60,$CF,$CF,$48,$E2
	.dc.b	$DC,$7F,$C8,$BB,$3E,$C7,$FC,$BC,$DA,$FF,$2E,$B6,$77,$3F,$C7,$FF
	.dc.b	$03,$78,$14,$97,$FC,$7A,$32,$FF,$20,$F7,$DE,$97,$F2,$76,$C6,$63
	.dc.b	$07,$8B,$F0,$29,$1E,$26,$E3,$E8,$E7,$F3,$92,$7C,$EF,$F3,$3F,$C1
	.dc.b	$7F,$86,$CF,$F2,$86,$BF,$51,$AF,$F0,$76,$BF,$C1,$9F,$DD,$39,$F0
	.dc.b	$BA,$12,$BF,$9D,$CF,$EF,$27,$B7,$83,$E1,$24,$EB,$E4,$3C,$84,$70
	.dc.b	$72,$F9,$0F,$41,$1C,$1D,$BE,$4B,$D8,$47,$07,$8B,$D3,$F8,$04,$70
	.dc.b	$7F,$70,$FC,$19,$F4,$AF,$8B,$FF,$2D,$FC,$C7,$F9,$D7,$4F,$F2,$ED
	.dc.b	$6A,$73,$EA,$FF,$1A,$8F,$F6,$55,$6E,$91,$B7,$FD,$AF,$3B,$BE,$D1
	.dc.b	$83,$AB,$8A,$E5,$5D,$2B,$85,$F8,$2F,$6D,$7B,$DF,$19,$F2,$7C,$C3
	.dc.b	$1D,$61,$E3,$C7,$AD,$EF,$0E,$67,$8C,$F4,$DF,$17,$15,$C7,$0F,$1C
	.dc.b	$62,$B9,$3C,$30,$EF,$18,$AE,$13,$DB,$0F,$6E,$1D,$5F,$1E,$3C,$09
	.dc.b	$3E,$09,$D6,$3E,$0A,$DC,$F8,$6B,$DB,$5A,$8E,$D5,$B2,$3F,$6A,$34
	.dc.b	$FC,$89,$EF,$5D,$0F,$2E,$38,$BE,$35,$B3,$C9,$19,$7B,$90,$F8,$47
	.dc.b	$94,$87,$F6,$03,$91,$70,$FC,$33,$71,$FC,$2B,$9F,$9D,$78,$7F,$97
	.dc.b	$6C,$73,$3F,$1E,$E1,$E1,$1B,$3F,$03,$1C,$31,$FC,$6B,$71,$FC,$D7
	.dc.b	$CE,$FF,$74,$9C,$D7,$02,$5F,$E0,$A7,$55,$FA,$1B,$E5,$3F,$A6,$6B
	.dc.b	$E4,$AF,$C8,$D7,$08,$D4,$74,$BE,$F1,$D4,$C5,$F5,$9B,$1F,$B2,$AE
	.dc.b	$4F,$53,$F2,$B7,$E0,$C4,$D9,$DC,$62,$B9,$D7,$1C,$3F,$5F,$5A,$BE
	.dc.b	$44,$AF,$4D,$70,$BE,$4F,$82,$32,$D2,$5F,$B0,$EA,$F1,$AE,$8F,$BC
	.dc.b	$DE,$1E,$F6,$9E,$F3,$F8,$07,$86,$E3,$FF,$38,$C7,$F2,$EF,$0B,$FD
	.dc.b	$13,$39,$6C,$EA,$7F,$9F,$56,$08,$FF,$36,$77,$F1,$B2,$3F,$9C,$B5
	.dc.b	$FE,$8D,$F7,$7D,$08,$FF,$3E,$7B,$E7,$23,$FA,$0B,$5F,$F3,$AF,$5B
	.dc.b	$F0,$6F,$3B,$63,$FB,$14,$60,$FF,$51,$F1,$3C,$ED,$4F,$EA,$63,$13
	.dc.b	$A9,$CC,$FF,$93,$27,$2F,$E9,$7E,$21,$EB,$FF,$AD,$5B,$CB,$FE,$B0
	.dc.b	$6C,$EE,$75,$E1,$BC,$FD,$36,$7F,$C0,$31,$83,$E8,$F8,$5F,$E8,$B8
	.dc.b	$FE,$FD,$EB,$E0,$8F,$FA,$7B,$F0,$6F,$F5,$BB,$7C,$1F,$B3,$18,$8F
	.dc.b	$F4,$DB,$5F,$F3,$3E,$5F,$E3,$AF,$17,$C6,$48,$FF,$09,$BF,$82,$F1
	.dc.b	$DF,$43,$49,$3F,$EA,$E8,$CB,$7C,$6E,$04,$70,$7F,$0D,$DF,$FC,$93
	.dc.b	$1F,$ED,$66,$C6,$76,$4E,$5B,$1F,$EB,$CF,$C1,$DF,$49,$F9,$66,$C6
	.dc.b	$63,$07,$FB,$8B,$99,$F8,$3F,$FA,$9A,$7F,$8F,$46,$3C,$FD,$4E,$67
	.dc.b	$FD,$64,$D7,$FD,$1B,$E9,$BF,$07,$FF,$63,$C7,$FA,$AD,$AF,$F9,$95
	.dc.b	$AF,$06,$A3,$F9,$8F,$D1,$3E,$6B,$CC,$D8,$F2,$B5,$FF,$62,$35,$FF
	.dc.b	$4F,$E2,$87,$CF,$72,$24,$F2,$7E,$17,$D4,$FF,$42,$8E,$2D,$7C,$94
	.dc.b	$8C,$47,$F4,$66,$F8,$8F,$33,$F8,$14,$FF,$9C,$1A,$FF,$B0,$DA,$FF
	.dc.b	$B0,$9B,$1C,$7C,$3F,$E1,$B1,$FE,$1D,$9E,$51,$83,$FA,$93,$F1,$FF
	.dc.b	$06,$7E,$0F,$88,$46,$1D,$B5,$FE,$B4,$D7,$FB,$32,$F9,$8E,$4D,$8E
	.dc.b	$31,$83,$B7,$E0,$EF,$8A,$A4,$9F,$2C,$65,$FE,$D1,$93,$F0,$77,$CF
	.dc.b	$FF,$2A,$6C,$E9,$AF,$F7,$99,$E1,$3B,$F5,$E4,$ED,$AE,$FD,$1F,$F4
	.dc.b	$C3,$E0,$6B,$9F,$73,$F9,$CC,$62,$3F,$D4,$AB,$E2,$72,$48,$E4,$FE
	.dc.b	$1B,$E5,$E4,$7E,$9B,$89,$18,$8F,$F6,$17,$E0,$DF,$E8,$E7,$FC,$11
	.dc.b	$D2,$8C,$46,$BD,$CF,$B7,$39,$7F,$27,$E5,$49,$EC,$D7,$E2,$7B,$3F
	.dc.b	$8F,$7E,$6B,$5C,$F3,$FF,$56,$8C,$47,$FB,$09,$19,$23,$C9,$FE,$BB
	.dc.b	$F8,$27,$EA,$63,$EC,$B5,$FA,$8D,$7F,$D9,$AD,$7F,$75,$C2,$87,$DB
	.dc.b	$F4,$24,$F4,$7F,$31,$F8,$23,$E7,$A9,$27,$F3,$F1,$97,$F3,$7D,$FF
	.dc.b	$96,$23,$83,$FC,$6D,$F2,$6F,$C1,$1F,$82,$7E,$A2,$30,$D8,$CC,$F0
	.dc.b	$6B,$FA,$6F,$0B,$F4,$64,$7F,$56,$FE,$BF,$F0,$27,$E0,$34,$93,$FB
	.dc.b	$68,$CB,$7E,$1B,$81,$1C,$1B,$F9,$6E,$84,$70,$73,$D7,$C0,$47,$07
	.dc.b	$F7,$1F,$99,$F4,$BF,$D6,$9A,$FF,$2D,$F4,$D1,$88,$46,$34,$FE,$ED
	.dc.b	$F3,$7F,$DE,$49,$E8,$E0,$FE,$F5,$D1,$FF,$00,$6F,$FC,$00,$D7,$FC
	.dc.b	$E1,$7F,$EE,$EF,$8A,$D9,$FE,$01,$F9,$F7,$C4,$52,$57,$F8,$26,$32
	.dc.b	$FF,$82,$FE,$83,$F9,$7D,$9A,$4F,$7B,$FC,$19,$F8,$8F,$F8,$66,$BF
	.dc.b	$C2,$4D,$7C,$6D,$7F,$BE,$B5,$F8,$9F,$7D,$FC,$5D,$F4,$B5,$FF,$01
	.dc.b	$B5,$DF,$97,$FE,$1B,$FA,$57,$C6,$D2,$4F,$F8,$8E,$32,$FF,$89,$7E
	.dc.b	$9F,$F9,$B6,$66,$92,$3F,$C4,$DF,$79,$FD,$61,$FF,$16,$35,$F9,$AD
	.dc.b	$7F,$C2,$ED,$7E,$F7,$D4,$BE,$27,$FC,$60,$D8,$FF,$1B,$35,$FA,$ED
	.dc.b	$7F,$C4,$3F,$88,$BE,$9B,$FC,$70,$D8,$FF,$1F,$35,$FB,$ED,$7F,$C8
	.dc.b	$9F,$79,$FE,$FE,$7F,$D5,$1C,$08,$C4,$7F,$93,$5A,$F0,$F9,$3F,$F2
	.dc.b	$9F,$B6,$FA,$8A,$49,$E7,$19,$7F,$CA,$1F,$09,$5F,$C8,$B5,$FF,$1D
	.dc.b	$B5,$FD,$13,$56,$0F,$BB,$7A,$9F,$E6,$46,$CF,$A9,$AF,$F9,$95,$AF
	.dc.b	$F9,$8F,$EE,$FF,$91,$3D,$1B,$46,$9A,$E2,$7D,$93,$E4,$BF,$CC,$CD
	.dc.b	$BF,$F3,$63,$5F,$D6,$B5,$FF,$14,$0F,$29,$D8,$F6,$7F,$90,$36,$7F
	.dc.b	$92,$B5,$FF,$73,$4F,$09,$FF,$41,$FC,$2F,$E0,$E7,$FC,$69,$9F,$F9
	.dc.b	$33,$9B,$24,$8F,$F3,$D7,$D5,$FF,$9F,$7E,$40,$D2,$47,$E0,$BE,$E9
	.dc.b	$F2,$9F,$E9,$06,$C6,$63,$07,$F9,$57,$EE,$BF,$D9,$B1,$FE,$54,$6B
	.dc.b	$FE,$FD,$F8,$74,$62,$3F,$C8,$2D,$7F,$C7,$2F,$D6,$6F,$27,$52,$3D
	.dc.b	$79,$7C,$04,$70,$7E,$6B,$EE,$3F,$C5,$0F,$C4,$6B,$F8,$A6,$EA,$73
	.dc.b	$83,$ED,$3F,$C1,$D5,$FE,$6E,$6B,$EB,$6B,$FE,$5A,$6B,$F7,$BE,$E5
	.dc.b	$F3,$CF,$BA,$FF,$21,$C6,$1B,$19,$9F,$C2,$9F,$E6,$D9,$D3,$80,$BE
	.dc.b	$83,$F1,$51,$98,$C7,$2F,$69,$CC,$F4,$9C,$BF,$E5,$1F,$9F,$FF,$29
	.dc.b	$26,$91,$FF,$66,$67,$FE,$74,$F3,$FF,$D0,$19,$1F,$99,$DD,$6B,$DD
	.dc.b	$1F,$E7,$2E,$DF,$ED,$EF,$03,$65,$86,$DF,$EA,$2F,$94,$E5,$87,$38
	.dc.b	$C5,$6F,$83,$EB,$63,$A8,$FF,$3F,$F4,$FF,$AB,$9E,$95,$FE,$AC,$F8
	.dc.b	$9F,$26,$5F,$E8,$EA,$FF,$2D,$D7,$FA,$A1,$B1,$F9,$56,$BB,$EB,$EC
	.dc.b	$F9,$A3,$F6,$D3,$D9,$E3,$97,$9D,$CC,$FC,$23,$DB,$A3,$D6,$E7,$97
	.dc.b	$F0,$D6,$D7,$0B,$F8,$15,$BC,$3E,$07,$DC,$FF,$0F,$6F,$81,$E4,$D9
	.dc.b	$2F,$90,$C3,$BD,$3C,$5E,$17,$FB,$FB,$3F,$5B,$71,$E6,$7D,$37,$13
	.dc.b	$E9,$C1,$F7,$2F,$8C,$E3,$F5,$CF,$9E,$F1,$4E,$71,$E5,$39,$6C,$FD
	.dc.b	$2C,$3F,$C3,$93,$97,$FC,$22,$D7,$FD,$F6,$D7,$FC,$45,$E3,$3E,$AF
	.dc.b	$94,$7F,$88,$A3,$17,$FC,$C2,$73,$9F,$F3,$67,$79,$FF,$10,$F8,$AF
	.dc.b	$91,$E1,$1F,$87,$87,$2C,$7E,$04,$E7,$1F,$C8,$78,$0F,$A2,$FA,$0F
	.dc.b	$F9,$D2,$33,$18,$F8,$AF,$9B,$F4,$BF,$B5,$9F,$DC,$6A,$8C,$63,$FE
	.dc.b	$30,$6B,$FD,$12,$35,$E0,$FA,$3E,$E3,$FD,$71,$7F,$A1,$6D,$7D,$57
	.dc.b	$D1,$D3,$FB,$1D,$D1,$A8,$FD,$A6,$0F,$B8,$FE,$27,$87,$EA,$FB,$BF
	.dc.b	$B5,$BF,$81,$8C,$E3,$F8,$19,$CB,$F8,$A8,$F3,$BF,$E5,$56,$BF,$E5
	.dc.b	$46,$D7,$7C,$BE,$9F,$E6,$3F,$EB,$0C,$3F,$1C,$DB,$FD,$37,$C7,$7F
	.dc.b	$02,$EB,$FC,$AC,$E7,$3F,$4F,$B0,$FC,$1B,$F6,$77,$FE,$87,$F8,$1F
	.dc.b	$CD,$CE,$72,$FE,$19,$39,$7F,$39,$FA,$CB,$3F,$C2,$B6,$3E,$AB,$63
	.dc.b	$FC,$C6,$D7,$EA,$7B,$9F,$2E,$7F,$CE,$8D,$85,$3F,$10,$7C,$27,$EB
	.dc.b	$AB,$52,$FC,$41,$F1,$BF,$9A,$8C,$36,$33,$8F,$FB,$C9,$B8,$FF,$9D
	.dc.b	$BE,$ED,$F7,$BF,$9E,$6D,$7E,$39,$B3,$B9,$ED,$F8,$7B,$EC,$E9,$2F
	.dc.b	$FD,$FF,$19,$7F,$E0,$0F,$E1,$EF,$C2,$3C,$CD,$8C,$C6,$0F,$D7,$EC
	.dc.b	$FB,$6A,$49,$FF,$47,$BE,$FF,$94,$FF,$1B,$C7,$F5,$BD,$FF,$D6,$7E
	.dc.b	$D6,$49,$1F,$EA,$5F,$BA,$FD,$DF,$C9,$1A,$48,$FD,$6F,$E3,$DF,$E9
	.dc.b	$66,$FC,$8F,$F0,$28,$C7,$0F,$FA,$79,$AF,$EB,$FF,$15,$7D,$6F,$F0
	.dc.b	$86,$DF,$EE,$1B,$3F,$EA,$66,$BF,$B7,$FC,$59,$F4,$9F,$C4,$1B,$5F
	.dc.b	$C0,$1B,$3F,$EA,$E6,$BF,$BF,$FC,$75,$F6,$1F,$C6,$1B,$5F,$C2,$1B
	.dc.b	$3F,$EB,$66,$BF,$C1,$FF,$18,$7A,$FF,$C8,$1B,$5F,$C4,$1B,$3F,$EB
	.dc.b	$E6,$BF,$E5,$2F,$73,$F9,$EF,$C8,$FF,$5E,$87,$E0,$1F,$D2,$7C,$6F
	.dc.b	$EE,$5E,$CF,$FA,$37,$1F,$EF,$15,$AF,$73,$85,$FF,$51,$9F,$F0,$B4
	.dc.b	$F2,$F2,$BF,$9B,$B6,$33,$3B,$8F,$F2,$25,$F9,$B0,$EF,$A7,$4C,$3F
	.dc.b	$B3,$CE,$70,$FD,$D7,$47,$F6,$BF,$C1,$3F,$81,$FB,$FF,$E9,$19,$FE
	.dc.b	$02,$A3,$3F,$21,$27,$F8,$2F,$E2,$5F,$DB,$DE,$75,$C7,$4F,$E2,$73
	.dc.b	$97,$F7,$BF,$1D,$FD,$75,$FE,$2E,$DD,$7F,$C4,$19,$21,$F8,$4F,$F1
	.dc.b	$67,$58,$FA,$3F,$13,$FE,$06,$7F,$04,$8C,$E1,$FC,$2A,$3F,$86,$71
	.dc.b	$FF,$70,$F0,$7F,$C2,$3F,$52,$FD,$27,$E9,$5F,$F8,$36,$B5,$8F,$F8
	.dc.b	$AE,$B5,$CD,$C2,$F7,$7F,$E9,$09,$CD,$7F,$92,$DB,$1F,$E5,$06,$BF
	.dc.b	$85,$F8,$6F,$E6,$53,$FE,$3B,$FA,$CE,$98,$7F,$86,$A7,$35,$FC,$8B
	.dc.b	$E1,$3C,$BC,$1F,$11,$FB,$11,$9A,$FE,$73,$18,$E0,$37,$EB,$BF,$3F
	.dc.b	$7C,$7A,$3F,$9F,$C6,$33,$FC,$0F,$E0,$AF,$B6,$FE,$86,$D8,$FE,$82
	.dc.b	$BE,$D3,$33,$FD,$23,$E1,$9F,$37,$4C,$DF,$F4,$88,$CB,$F9,$EF,$B6
	.dc.b	$7E,$07,$FA,$5F,$C2,$3F,$BC,$EB,$FE,$4C,$9C,$D7,$E9,$27,$2F,$FA
	.dc.b	$F9,$7D,$E7,$E8,$9B,$1F,$A6,$6F,$17,$F9,$F5,$AF,$FA,$0A,$BF,$35
	.dc.b	$3E,$3E,$FF,$15,$FE,$DE,$73,$5F,$DA,$BF,$05,$F8,$CF,$F5,$46,$F1
	.dc.b	$9F,$F3,$9C,$6F,$C7,$FE,$D7,$F7,$6F,$AC,$A7,$F6,$EA,$E3,$19,$CF
	.dc.b	$FC,$DB,$FA,$2F,$BB,$0F,$F3,$CB,$EA,$DB,$FC,$F0,$E7,$B4,$E6,$BF
	.dc.b	$B9,$46,$35,$FC,$57,$DF,$FF,$99,$2B,$58,$7F,$9A,$23,$1F,$27,$9F
	.dc.b	$B5,$B8,$F8,$9F,$D2,$FE,$13,$F2,$2F,$FB,$F7,$80,$8C,$42,$31,$97
	.dc.b	$F9,$29,$B7,$FE,$AD,$CF,$CF,$F0,$3E,$33,$AD,$59,$FF,$06,$FE,$0D
	.dc.b	$FE,$41,$C3,$FC,$32,$BE,$E7,$FC,$2A,$E7,$7F,$E0,$AE,$6B,$8C,$FF
	.dc.b	$86,$BE,$8B,$FC,$FD,$7C,$A3,$1E,$5F,$F8,$7B,$E4,$BF,$B8,$4F,$F8
	.dc.b	$99,$FF,$06,$DF,$F8,$79,$CE,$3F,$C4,$9C,$CE,$FD,$EF,$F1,$3F,$E0
	.dc.b	$3E,$18,$C5,$FF,$89,$A3,$17,$FE,$46,$FC,$05,$F6,$BF,$E6,$A6,$D7
	.dc.b	$F9,$69,$B7,$FE,$D6,$F9,$14,$62,$3F,$C5,$C7,$F8,$93,$F0,$2F,$F5
	.dc.b	$A3,$FE,$3E,$6E,$1D,$F0,$FF,$1D,$C6,$30,$FE,$7B,$F4,$EF,$A9,$FF
	.dc.b	$4B,$E2,$E9,$FF,$20,$CE,$6B,$FD,$31,$39,$6C,$FA,$48,$FA,$5F,$67
	.dc.b	$D1,$C8,$F6,$37,$3F,$F7,$8F,$80,$C3,$FC,$39,$5F,$E4,$C6,$CF,$FA
	.dc.b	$B3,$0F,$43,$FE,$42,$8C,$E9,$FD,$11,$7C,$97,$F8,$FD,$B7,$FE,$21
	.dc.b	$6D,$78,$1A,$FC,$AC,$3C,$5E,$67,$C0,$3E,$0D,$DB,$F2,$7E,$0B,$F1
	.dc.b	$0F,$8B,$F0,$F6,$3D,$1E,$87,$C8,$31,$8F,$F9,$C3,$2F,$E6,$F3,$F4
	.dc.b	$3E,$17,$CC,$47,$CC,$FA,$7C,$AF,$A2,$7D,$26,$C7,$F8,$13,$F0,$6F
	.dc.b	$E5,$BF,$37,$FE,$6F,$FA,$7F,$F6,$A6,$3F,$E7,$19,$CA,$8C,$A8,$FF
	.dc.b	$1B,$C6,$1D,$B7,$CF,$57,$5F,$7F,$F1,$F5,$A6,$E2,$C7,$58,$F8,$70
	.dc.b	$E2,$F8,$6B,$8C,$63,$1E,$D8,$F0,$AF,$26,$1C,$DE,$07,$38,$D6,$3E
	.dc.b	$18,$C1,$A3,$8E,$3A,$C3,$B6,$3E,$0B,$C1,$CB,$0E,$D8,$F8,$B0,$E5
	.dc.b	$8F,$57,$E3,$61,$C9,$F9,$95,$C3,$0E,$91,$C2,$74,$F0,$C3,$B6,$3E
	.dc.b	$08,$D4,$5C,$7E,$86,$1C,$E7,$D7,$1E,$32,$61,$F3,$AF,$83,$E0,$8C
	.dc.b	$CF,$92,$72,$F2,$9D,$1C,$5F,$09,$27,$8B,$F2,$0C,$47,$B1,$FE,$1A
	.dc.b	$7A,$CA,$53,$38,$FE,$39,$BB,$3C,$3F,$F2,$AC,$A3,$13,$CA,$31,$3A
	.dc.b	$9C,$CF,$28,$FB,$F1,$C2,$FF,$2C,$DB,$FC,$9E,$FF,$C8,$A7,$35,$C0
	.dc.b	$97,$F7,$27,$55,$F9,$AB,$E5,$3F,$A7,$6B,$FA,$56,$E0,$AE,$33,$BB
	.dc.b	$EB,$87,$82,$3B,$18,$C3,$B4,$D8,$FD,$A5,$72,$7B,$46,$5A,$72,$3F
	.dc.b	$27,$87,$88,$91,$FA,$29,$E5,$87,$CA,$AD,$CF,$38,$F4,$45,$AE,$2F
	.dc.b	$5C,$3D,$04,$C3,$CF,$19,$6C,$66,$3E,$01,$92,$B9,$7D,$2F,$F0,$BC
	.dc.b	$3E,$3F,$17,$F3,$DF,$89,$FE,$AF,$C9,$FC,$DD,$BA,$7F,$52,$6E,$1F
	.dc.b	$CF,$7C,$CB,$7F,$D1,$5F,$F4,$FB,$63,$F9,$E4,$63,$8B,$FC,$59,$1F
	.dc.b	$CF,$FD,$FE,$30,$69,$D3,$AD,$F3,$74,$94,$3F,$B5,$3B,$F7,$F9,$FA
	.dc.b	$52,$7F,$AB,$3C,$FE,$8E,$47,$FB,$BE,$D1,$E8,$6C,$66,$3D,$1E,$67
	.dc.b	$98,$91,$FD,$BB,$E8,$BE,$1F,$87,$FE,$00,$7F,$D7,$ED,$8F,$F0,$34
	.dc.b	$63,$BC,$EB,$E7,$DF,$4B,$49,$5F,$BA,$B4,$EA,$3F,$7B,$D8,$F4,$7F
	.dc.b	$DB,$F1,$F0,$F9,$BF,$9A,$5F,$0C,$3C,$B8,$71,$A5,$7F,$39,$C3,$97
	.dc.b	$85,$F3,$FC,$A9,$5C,$FA,$3F,$D9,$BC,$09,$F7,$ED,$FC,$43,$1F,$F1
	.dc.b	$35,$FB,$48,$7F,$89,$4F,$C1,$F9,$8F,$94,$FC,$3F,$99,$E8,$9F,$F1
	.dc.b	$D6,$7F,$8A,$AD,$47,$84,$75,$21,$F0,$1B,$C2,$56,$EB,$53,$F9,$4F
	.dc.b	$55,$E3,$FE,$5B,$D0,$7D,$A5,$35,$1B,$F0,$2E,$3C,$EF,$8B,$63,$F4
	.dc.b	$3E,$33,$F0,$17,$C7,$61,$B5,$E9,$9D,$4F,$EA,$3D,$67,$E9,$6F,$5F
	.dc.b	$0D,$BF,$F6,$93,$5F,$F6,$93,$5F,$F6,$C3,$5F,$F4,$4E,$1F,$EE,$E7
	.dc.b	$C0,$FE,$69,$FF,$72,$37,$93,$FB,$C6,$3E,$9F,$0B,$D7,$F8,$B7,$F4
	.dc.b	$1B,$F4,$7E,$36,$FA,$6F,$7B,$63,$D9,$3B,$7F,$DE,$0F,$F1,$A6,$F9
	.dc.b	$BD,$0E,$BE,$5F,$F0,$6F,$75,$EA,$52,$57,$D9,$F6,$9F,$1F,$FC,$31
	.dc.b	$B1,$98,$C1,$9D,$7F,$89,$C3,$F0,$BF,$F1,$4E,$FF,$E1,$58,$C7,$A5
	.dc.b	$A9,$CF,$89,$FE,$74,$FC,$1F,$E7,$5E,$DF,$99,$7F,$C3,$5A,$9A,$63
	.dc.b	$8D,$F5,$7E,$28,$7D,$6D,$3C,$B3,$AA,$FC,$B7,$D7,$FF,$02,$6F,$9D
	.dc.b	$FC,$AA,$31,$1F,$B6,$CB,$F8,$F7,$84,$F8,$BF,$BD,$39,$3F,$4A,$E3
	.dc.b	$4F,$F1,$A4,$EF,$A7,$F9,$AC,$E6,$F9,$BF,$83,$7A,$F2,$FF,$A2,$F0
	.dc.b	$FD,$57,$E3,$1F,$EF,$66,$FC,$1F,$F3,$E9,$D4,$78,$71,$FE,$E1,$CD
	.dc.b	$9F,$7E,$4F,$EC,$EB,$EF,$BF,$8C,$FA,$AF,$C5,$91,$FD,$F7,$80,$FD
	.dc.b	$23,$F8,$0C,$7B,$63,$1E,$4F,$F8,$72,$3F,$DB,$13,$A7,$FD,$5B,$3A
	.dc.b	$7F,$C4,$2D,$7F,$C4,$BD,$2E,$E7,$89,$FB,$3B,$EF,$C7,$FE,$5B,$C3
	.dc.b	$83,$FA,$6A,$DE,$1F,$C3,$7E,$4F,$FC,$43,$3F,$39,$BC,$5F,$59,$B3
	.dc.b	$FC,$63,$D2,$FD,$43,$F1,$DB,$5F,$59,$B5,$F7,$9B,$3C,$7B,$BF,$B6
	.dc.b	$FE,$0D,$FC,$FE,$3F,$8F,$7C,$B7,$FA,$CD,$BE,$EE,$BD,$8D,$F8,$16
	.dc.b	$7F,$92,$5F,$17,$EC,$3F,$93,$7F,$73,$E6,$7F,$02,$23,$FC,$C4,$7E
	.dc.b	$0D,$DD,$FC,$81,$8C,$3E,$A3,$4F,$AD,$1B,$7F,$CA,$AD,$7F,$BC,$71
	.dc.b	$FD,$3F,$4B,$F7,$32,$FA,$5A,$7F,$14,$8F,$ED,$BF,$36,$FA,$3F,$ED
	.dc.b	$AD,$C3,$FB,$33,$5F,$F7,$93,$5F,$F7,$8F,$B9,$FD,$23,$F0,$9F,$EE
	.dc.b	$BE,$B1,$A4,$9F,$EE,$1F,$3F,$FD,$7D,$BE,$DF,$F7,$C8,$C4,$7F,$BF
	.dc.b	$DA,$F6,$FA,$0A,$FF,$85,$70,$46,$21,$18,$AF,$F0,$3C,$E6,$1E,$27
	.dc.b	$EA,$2B,$FC,$FB,$18,$F0,$FF,$C1,$B3,$99,$F6,$B5,$F0,$7E,$B5,$FE
	.dc.b	$28,$7F,$C1,$4D,$79,$B5,$F8,$AD,$7F,$87,$FE,$35,$F1,$5F,$F4,$5F
	.dc.b	$8C,$8C,$47,$FA,$42,$31,$A7,$F9,$02,$73,$3F,$43,$E4,$7F,$A2,$37
	.dc.b	$C6,$FF,$12,$C6,$23,$E3,$35,$FE,$FF,$D9,$FC,$A9,$F0,$B5,$F0,$35
	.dc.b	$FF,$21,$1C,$3C,$EF,$F2,$8F,$CA,$7F,$83,$63,$39,$23,$11,$FC,$B1
	.dc.b	$AF,$F8,$39,$BB,$3F,$03,$FF,$03,$CF,$F8,$CD,$AF,$D4,$6B,$F8,$E6
	.dc.b	$BF,$CE,$3F,$18,$7C,$8F,$F9,$91,$B1,$F3,$1A,$FC,$B6,$BF,$98,$FA
	.dc.b	$DF,$F2,$0B,$FE,$23,$B4,$62,$3F,$A1,$46,$3C,$3F,$F3,$9E,$5F,$BD
	.dc.b	$F1,$7F,$C8,$7D,$9F,$BF,$BF,$D1,$B5,$FC,$B3,$5F,$CA,$B6,$DF,$81
	.dc.b	$7F,$09,$AF,$E9,$BF,$8B,$DF,$1F,$FA,$AB,$70,$FC,$ED,$5B,$7D,$A3
	.dc.b	$E5,$FF,$CC,$ED,$AF,$63,$5F,$E9,$98,$BF,$24,$FF,$36,$FE,$06,$F9
	.dc.b	$8F,$F3,$6B,$6B,$FC,$CC,$D7,$F4,$ED,$7F,$80,$BF,$E8,$D7,$FC,$99
	.dc.b	$F1,$AF,$C0,$DF,$31,$FE,$9C,$8C,$3E,$0F,$95,$C7,$4B,$E5,$29,$30
	.dc.b	$E7,$0F,$93,$FF,$41,$FC,$7D,$9F,$F3,$D3,$5F,$DD,$B5,$FE,$2A,$D7
	.dc.b	$FD,$39,$8B,$F0,$37,$CD,$79,$5F,$F1,$87,$07,$F9,$E5,$BC,$DC,$1D
	.dc.b	$DF,$BB,$F0,$EF,$C1,$4F,$46,$C3,$04,$63,$95,$F4,$5F,$D6,$9B,$E8
	.dc.b	$FF,$A7,$23,$FE,$00,$DF,$F2,$C6,$BF,$C7,$DA,$FF,$1E,$FA,$67,$E0
	.dc.b	$7F,$EA,$46,$C7,$FA,$75,$AF,$F1,$E6,$BF,$BC,$C5,$E9,$78,$E7,$7A
	.dc.b	$FF,$A7,$9B,$A7,$C8,$9F,$47,$D5,$FF,$43,$AF,$F5,$1B,$67,$F9,$A3
	.dc.b	$70,$F8,$5F,$81,$FE,$46,$BF,$D4,$8D,$AF,$A6,$D9,$DF,$67,$E1,$E1
	.dc.b	$F8,$57,$F9,$9B,$E5,$E5,$7F,$9A,$9F,$F4,$3F,$EB,$7F,$C2,$31,$FF
	.dc.b	$38,$B7,$1F,$EC,$6D,$A7,$02,$30,$DF,$73,$99,$FE,$23,$71,$3F,$C8
	.dc.b	$8E,$71,$99,$FD,$E4,$FC,$AF,$2B,$FD,$33,$B7,$F2,$3B,$FF,$4B,$35
	.dc.b	$FF,$26,$70,$3F,$03,$7D,$57,$E9,$5B,$3F,$AA,$6B,$C1,$BA,$7F,$A3
	.dc.b	$BE,$57,$F8,$24,$BE,$35,$F8,$2F,$FB,$4A,$3D,$06,$5C,$3E,$B8,$DD
	.dc.b	$7F,$90,$5B,$1C,$17,$C5,$3F,$05,$D3,$67,$4D,$7A,$35,$F0,$FD,$2F
	.dc.b	$F8,$69,$34,$8F,$F9,$2B,$EB,$DF,$70,$E4,$FF,$5F,$46,$1F,$97,$F2
	.dc.b	$79,$AF,$F3,$B7,$1A,$B5,$1F,$1F,$C4,$D6,$BF,$EF,$3E,$0F,$F4,$4E
	.dc.b	$3D,$32,$CB,$FC,$12,$33,$7D,$38,$DF,$27,$FE,$9F,$8F,$F1,$6C,$62
	.dc.b	$BE,$6E,$7F,$AA,$AD,$57,$E3,$AB,$4F,$F9,$99,$7D,$37,$82,$34,$DA
	.dc.b	$E5,$7F,$88,$8C,$53,$E9,$69,$88,$B3,$FC,$C3,$F0,$57,$CD,$75,$E8
	.dc.b	$46,$27,$5F,$3D,$98,$C6,$7F,$94,$7F,$D8,$B3,$CF,$DC,$D6,$1B,$7A
	.dc.b	$7D,$A7,$B3,$87,$E9,$64,$87,$D4,$6B,$83,$6C,$C7,$FB,$77,$8A,$F9
	.dc.b	$0C,$6B,$FD,$BA,$31,$3F,$D2,$BF,$16,$E6,$FE,$6B,$24,$62,$BF,$69
	.dc.b	$18,$C3,$E8,$B6,$BF,$43,$E5,$7F,$30,$F0,$FE,$83,$63,$B5,$63,$CA
	.dc.b	$FF,$67,$C7,$F5,$AF,$C1,$5F,$41,$F2,$DF,$F7,$C4,$66,$31,$C9,$9F
	.dc.b	$7B,$C3,$46,$92,$BD,$5F,$82,$F6,$D7,$F8,$1B,$78,$FF,$39,$A3,$EA
	.dc.b	$7F,$63,$7F,$DB,$2B,$4F,$F5,$AE,$8F,$F1,$83,$FE,$19,$FC,$17,$F8
	.dc.b	$0B,$FD,$CA,$33,$8F,$F8,$6A,$72,$FF,$12,$FC,$67,$F2,$75,$F3,$5B
	.dc.b	$1F,$31,$AF,$F4,$76,$BF,$EE,$8E,$DF,$C4,$E7,$FE,$E4,$6D,$67,$6F
	.dc.b	$CF,$7C,$17,$F3,$0E,$BB,$F8,$7F,$FA,$3E,$BF,$C6,$1F,$83,$3F,$07
	.dc.b	$CF,$CB,$7D,$68,$C4,$B4,$FF,$72,$74,$BE,$C7,$F5,$F9,$FF,$6F,$6C
	.dc.b	$66,$BF,$CD,$AD,$AF,$F3,$67,$E0,$CF,$8D,$F2,$B6,$BB,$70,$7F,$BF
	.dc.b	$A1,$F6,$9A,$AF,$5F,$E0,$DF,$C1,$AB,$F8,$96,$68,$C6,$1E,$78,$F3
	.dc.b	$63,$C3,$CA,$F8,$7F,$8E,$FF,$16,$8F,$F3,$4B,$5E,$0D,$52,$8C,$2F
	.dc.b	$C1,$9F,$2F,$F3,$97,$D3,$D8,$C3,$F4,$BF,$06,$F7,$60,$D3,$D3,$18
	.dc.b	$6E,$DB,$DB,$EC,$FC,$43,$EC,$FE,$7B,$77,$FF,$4C,$B7,$1F,$F4,$C3
	.dc.b	$5F,$F4,$2F,$A2,$F8,$6F,$C1,$AF,$96,$B1,$87,$F0,$FD,$6F,$C4,$69
	.dc.b	$8F,$80,$B9,$7D,$63,$E7,$7E,$5E,$8F,$C1,$B6,$3F,$D5,$AD,$CB,$FD
	.dc.b	$58,$D7,$FD,$33,$F9,$7F,$D8,$7F,$D9,$3F,$3C,$FC,$1F,$F3,$33,$9F
	.dc.b	$27,$FD,$9F,$1F,$9C,$FC,$BF,$FD,$D9,$8F,$8F,$0F,$AF,$18,$7E,$74
	.dc.b	$7E,$0E,$3F,$37,$18,$37,$F3,$1F,$A6,$F1,$7F,$C0,$75,$FE,$C4,$6C
	.dc.b	$FF,$B0,$DA,$FF,$AC,$1A,$FE,$3F,$F2,$F7,$E0,$1F,$C7,$BE,$9B,$FC
	.dc.b	$7B,$7E,$5E,$E5,$6A,$3F,$D4,$13,$AB,$F2,$53,$F3,$2F,$C2,$57,$EC
	.dc.b	$7C,$AF,$E7,$DF,$55,$FB,$C9,$CD,$FF,$8B,$3D,$B4,$FF,$5D,$87,$E1
	.dc.b	$3F,$D3,$1F,$F0,$9B,$5F,$F1,$63,$5F,$F6,$7B,$5F,$E0,$0F,$F0,$C6
	.dc.b	$F3,$F5,$8E,$9D,$FF,$E0,$39,$EF,$3F,$C4,$3E,$9D,$F6,$57,$E6,$30
	.dc.b	$DD,$7E,$24,$FF,$A4,$5B,$F3,$39,$C3,$E1,$CA,$3F,$8D,$FC,$A7,$F6
	.dc.b	$1A,$FD,$F3,$6B,$F6,$EF,$F7,$78,$FA,$0E,$78,$0F,$F5,$9C,$FF,$84
	.dc.b	$B6,$79,$FC,$0F,$6F,$77,$FC,$09,$C0,$FC,$66,$FE,$15,$86,$DF,$E4
	.dc.b	$79,$3F,$B9,$4F,$F8,$37,$EC,$DF,$85,$7D,$4A,$D6,$0A,$D4,$E7,$6F
	.dc.b	$CA,$5F,$0E,$5F,$F0,$A7,$C5,$BE,$BB,$71,$87,$F9,$EF,$C3,$71,$E1
	.dc.b	$FE,$46,$DD,$3F,$91,$3D,$DF,$C9,$39,$C7,$F8,$EB,$F9,$19,$E3,$F1
	.dc.b	$7D,$30,$EA,$FF,$4A,$FC,$29,$F8,$55,$F9,$6C,$37,$2F,$CF,$4F,$EA
	.dc.b	$30,$FE,$6B,$F1,$1F,$EB,$E9,$CD,$FF,$54,$8C,$67,$FA,$07,$EF,$4F
	.dc.b	$EE,$E1,$F8,$4F,$FB,$96,$3F,$55,$3A,$9F,$DD,$1A,$FA,$4F,$D0,$7E
	.dc.b	$15,$FD,$E2,$3F,$CE,$2D,$7F,$D1,$4D,$7F,$AE,$07,$0F,$F5,$F5,$F6
	.dc.b	$3D,$0E,$BF,$2D,$70,$FD,$E7,$E0,$1F,$E6,$0D,$3F,$A9,$B6,$3F,$A9
	.dc.b	$3F,$EA,$A7,$7E,$A7,$F7,$48,$FF,$02,$CF,$0F,$9C,$C4,$E5,$FF,$2A
	.dc.b	$7E,$15,$EB,$9F,$F3,$87,$5A,$31,$87,$F0,$57,$FC,$1B,$3A,$73,$A7
	.dc.b	$FA,$C1,$F5,$CF,$E5,$DB,$1F,$E7,$A7,$75,$D7,$3F,$F0,$4F,$E0,$BF
	.dc.b	$E0,$7A,$D7,$CA,$63,$E2,$3C,$AF,$9B,$CD,$FF,$38,$E1,$E5,$6E,$3F
	.dc.b	$C7,$EF,$FD,$75,$39,$CB,$FD,$4B,$D8,$FA,$7F,$1E,$BF,$E8,$D8,$C6
	.dc.b	$BD,$6B,$F9,$F4,$61,$FE,$59,$F8,$37,$F8,$FD,$BF,$25,$FE,$20,$87
	.dc.b	$E2,$76,$31,$F2,$B2,$B6,$72,$7F,$C4,$FF,$29,$EF,$CB,$FC,$1A,$D9
	.dc.b	$FF,$06,$3F,$E1,$47,$F9,$E3,$9D,$7F,$C0,$EF,$F3,$4F,$2B,$FC,$95
	.dc.b	$F0,$5D,$38,$FA,$BF,$E3,$4F,$C2,$7F,$C1,$64,$E7,$FF,$15,$BF,$E1
	.dc.b	$1F,$69,$F8,$4B,$E9,$FF,$CD,$2D,$D3,$FA,$65,$7E,$3F,$45,$F0,$BF
	.dc.b	$F2,$0F,$DD,$BF,$1C,$BF,$11,$86,$E3,$FD,$56,$7D,$98,$7F,$65,$C3
	.dc.b	$C3,$F8,$16,$A3,$19,$A3,$AE,$BF,$E4,$2B,$FF,$0E,$D7,$EB,$37,$46
	.dc.b	$23,$FC,$68,$D7,$CF,$7F,$E2,$AC,$7F,$BD,$63,$F0,$3E,$33,$AC,$62
	.dc.b	$75,$5F,$D9,$AF,$F2,$8F,$76,$BD,$B8,$3C,$87,$97,$ED,$FE,$19,$F1
	.dc.b	$1B,$7F,$07,$E2,$3E,$39,$F2,$3E,$26,$C7,$C5,$FC,$23,$E5,$98,$C3
	.dc.b	$C4,$FC,$EF,$B0,$F8,$44,$9F,$F8,$03,$B7,$C2,$8F,$8D,$F1,$9F,$50
	.dc.b	$E1,$3F,$45,$FE,$F1,$F8,$5A,$5F,$2D,$FE,$71,$FA,$7F,$F6,$6E,$5F
	.dc.b	$E7,$3A,$D4,$E7,$E0,$BF,$8D,$C3,$0F,$F7,$B5,$F0,$B4,$F0,$51,$98
	.dc.b	$C3,$70,$F0,$CE,$AF,$AD,$F4,$7B,$CE,$AB,$86,$1E,$3B,$F3,$4E,$6B
	.dc.b	$8E,$39,$9F,$1C,$E5,$DD,$67,$1F,$1D,$F0,$9F,$11,$A2,$9C,$A7,$85
	.dc.b	$73,$8F,$7E,$3E,$77,$86,$3E,$67,$8E,$1D,$9F,$99,$3B,$7A,$1A,$AE
	.dc.b	$93,$B8,$EB,$8F,$83,$09,$1B,$7D,$57,$CE,$B8,$12,$F9,$C6,$5E,$A4
	.dc.b	$9E,$37,$D3,$0F,$41,$31,$F8,$8D,$C7,$DB,$1F,$52,$B8,$C7,$8C,$87
	.dc.b	$F6,$03,$DA,$5C,$7F,$08,$DC,$3F,$02,$E7,$37,$89,$C6,$3F,$25,$18
	.dc.b	$71,$48,$C4,$71,$9E,$0F,$E4,$A3,$86,$3F,$91,$6E,$3F,$9A,$6C,$FE
	.dc.b	$7F,$6F,$EE,$F5,$BB,$FC,$35,$6A,$BA,$47,$E2,$6B,$77,$CE,$36,$49
	.dc.b	$FC,$0C,$E9,$F5,$C6,$6F,$B4,$68,$C5,$73,$7F,$13,$7E,$1C,$3F,$D5
	.dc.b	$91,$A3,$99,$AB,$F2,$39,$3C,$E6,$4E,$C5,$BF,$09,$2F,$F7,$B8,$79
	.dc.b	$E7,$8C,$E5,$EE,$FE,$EE,$35,$7F,$4E,$FC,$0D,$AE,$F5,$C5,$E7,$7E
	.dc.b	$D8,$E5,$1C,$30,$F7,$1E,$FC,$3E,$46,$BF,$CF,$BE,$63,$FC,$57,$DD
	.dc.b	$FE,$F4,$E3,$FF,$78,$E1,$FC,$DE,$73,$BF,$F4,$22,$C7,$FB,$BA,$7F
	.dc.b	$D1,$4F,$F3,$B8,$FC,$3D,$FF,$4A,$6C,$EA,$7F,$DF,$DD,$8F,$9E,$E7
	.dc.b	$68,$C4,$FF,$66,$9C,$C3,$DF,$A4,$9F,$ED,$6E,$BE,$97,$7F,$01,$27
	.dc.b	$FB,$7B,$CB,$E4,$E4,$7F,$3E,$BF,$F7,$B3,$63,$D3,$18,$8F,$D8,$76
	.dc.b	$7F,$7C,$FA,$7F,$82,$D8,$FE,$BD,$1E,$6E,$F9,$1D,$5E,$64,$7F,$65
	.dc.b	$D0,$F3,$9E,$D7,$4F,$21,$AD,$3E,$7D,$6E,$6D,$7F,$B2,$2F,$FD,$D5
	.dc.b	$39,$6C,$74,$6F,$B7,$E8,$F0,$7F,$85,$73,$39,$55,$FD,$F0,$D1,$F5
	.dc.b	$3F,$0C,$7A,$B4,$FB,$AE,$23,$FB,$3B,$5F,$E2,$DC,$C8,$C4,$7F,$83
	.dc.b	$7C,$17,$CB,$FE,$01,$B1,$FE,$EF,$6B,$FE,$EC,$7F,$D5,$4E,$7D,$FE
	.dc.b	$04,$78,$71,$FE,$2B,$D5,$F8,$91,$FE,$F4,$6E,$9E,$F8,$64,$9C,$E9
	.dc.b	$F0,$A3,$1F,$11,$CC,$95,$F1,$AF,$94,$EA,$31,$5C,$6D,$ED,$7E,$9B
	.dc.b	$D3,$7E,$16,$FC,$33,$0D,$C3,$FC,$CA,$D5,$97,$FA,$23,$F0,$A7,$77
	.dc.b	$EC,$3C,$84,$62,$BF,$9E,$46,$27,$F9,$BB,$63,$33,$B7,$EA,$E7,$FC
	.dc.b	$81,$EF,$A7,$E0,$AB,$AF,$81,$DE,$FF,$61,$5E,$E3,$E4,$DF,$1C,$7F
	.dc.b	$D3,$BF,$83,$FF,$49,$9F,$F7,$63,$5E,$A7,$F5,$27,$11,$C3,$C5,$F6
	.dc.b	$47,$3C,$BC,$4D,$AE,$13,$AE,$4F,$E0,$FE,$AB,$EA,$A9,$2B,$CD,$C6
	.dc.b	$9F,$EB,$79,$7D,$86,$E5,$F7,$DB,$5F,$84,$6D,$71,$FB,$4F,$E3,$F2
	.dc.b	$FC,$2F,$EB,$61,$F8,$B8,$C7,$1E,$A7,$38,$FF,$B5,$97,$D0,$E4,$FE
	.dc.b	$D7,$5C,$DB,$1A,$8F,$ED,$ED,$7F,$57,$F0,$7F,$E8,$37,$FD,$DD,$C2
	.dc.b	$FA,$BF,$2B,$63,$FD,$DF,$EB,$FF,$2C,$FC,$2F,$EC,$C7,$D2,$D9,$F6
	.dc.b	$2E,$DF,$E6,$2D,$9F,$93,$5C,$37,$FC,$FB,$F8,$96,$F3,$F5,$3B,$74
	.dc.b	$FF,$3A,$E6,$FF,$47,$5F,$D2,$6B,$E1,$6D,$7E,$A9,$7B,$1C,$49,$3C
	.dc.b	$5E,$9F,$39,$FA,$BF,$C2,$BF,$82,$DF,$FB,$B3,$ED,$3F,$88,$37,$C5
	.dc.b	$D7,$89,$F8,$CF,$0E,$E5,$FE,$D2,$C3,$8F,$47,$E3,$A7,$F2,$3E,$1E
	.dc.b	$EB,$F3,$31,$87,$F1,$7E,$63,$E7,$3F,$AB,$B6,$33,$18,$EF,$91,$C1
	.dc.b	$AF,$D1,$9F,$F3,$E6,$9F,$C2,$BF,$08,$FB,$13,$F5,$5F,$E8,$8F,$E6
	.dc.b	$9A,$E7,$AB,$FB,$3C,$3E,$8E,$C4,$9E,$11,$F9,$8F,$C2,$3F,$D7,$71
	.dc.b	$FB,$90,$E1,$FD,$A9,$8F,$5B,$87,$D0,$48,$FF,$20,$CA,$3F,$32,$D8
	.dc.b	$CC,$7F,$80,$3E,$7B,$F7,$BC,$1F,$BA,$F4,$BF,$7A,$7F,$7E,$8C,$D7
	.dc.b	$F3,$2F,$CC,$3F,$D4,$0F,$FB,$2B,$E8,$1F,$05,$FD,$B9,$B1,$FB,$C9
	.dc.b	$DB,$F9,$BF,$CC,$3F,$9F,$BF,$ED,$96,$BF,$CE,$DA,$FF,$40,$6F,$84
	.dc.b	$FB,$AF,$E6,$F3,$FC,$21,$AF,$E9,$DB,$C5,$FD,$01,$FE,$8F,$1F,$D2
	.dc.b	$BF,$30,$F5,$FB,$6F,$A7,$7D,$17,$F8,$19,$B5,$FC,$52,$BE,$0E,$2F
	.dc.b	$CC,$1F,$49,$FE,$28,$6D,$7F,$B4,$9A,$FF,$6C,$6B,$FE,$F7,$F3,$7F
	.dc.b	$C4,$EF,$EF,$5B,$A7,$F5,$66,$F1,$FF,$5D,$FC,$73,$F8,$C4,$FF,$5A
	.dc.b	$FC,$29,$F3,$DF,$E3,$25,$F8,$5D,$31,$1F,$E3,$6F,$8D,$7C,$D7,$F8
	.dc.b	$F9,$B9,$FF,$BA,$1A,$FF,$80,$9A,$FF,$83,$8F,$EE,$3E,$0B,$E6,$BF
	.dc.b	$C8,$2D,$8F,$E7,$2D,$7F,$8C,$3F,$CB,$1F,$EF,$35,$F4,$B8,$3C,$7D
	.dc.b	$7F,$E4,$A8,$FF,$10,$4F,$F1,$F7,$5F,$92,$FF,$BA,$E3,$C5,$48,$C4
	.dc.b	$7F,$80,$5A,$FF,$81,$9B,$0F,$C2,$FF,$AA,$37,$CE,$FF,$2C,$C6,$07
	.dc.b	$B7,$FE,$5C,$9F,$DE,$F0,$7F,$82,$1B,$C5,$FE,$0F,$6E,$3F,$E1,$47
	.dc.b	$F7,$F3,$FE,$0E,$FC,$97,$F9,$A4,$FF,$8B,$1A,$FA,$9A,$FF,$86,$5A
	.dc.b	$FF,$87,$BF,$0B,$EB,$D6,$FC,$2F,$AE,$3F,$E4,$08,$C7,$C4,$7F,$9C
	.dc.b	$AB,$C7,$97,$F8,$4B,$F0,$AF,$F3,$AE,$66,$98,$F8,$6F,$F2,$AF,$DE
	.dc.b	$7F,$5C,$F8,$B3,$49,$AF,$FB,$2A,$1F,$84,$FF,$A3,$B0,$FF,$1A,$68
	.dc.b	$FC,$29,$F6,$1F,$E6,$E6,$E1,$FE,$5E,$BE,$1B,$FF,$96,$31,$FE,$DB
	.dc.b	$F4,$7F,$E6,$6F,$84,$34,$FF,$3B,$F2,$63,$0F,$F3,$3F,$E1,$5F,$E1
	.dc.b	$57,$F2,$96,$8C,$47,$56,$A8,$D7,$CB,$FA,$E7,$F8,$9E,$DF,$E4,$D9
	.dc.b	$E5,$F3,$7F,$E7,$89,$37,$BF,$FD,$13,$F0,$FF,$E1,$0E,$13,$4F,$F0
	.dc.b	$9D,$BE,$B3,$F4,$58,$F5,$7F,$49,$9F,$A1,$AF,$C3,$6B,$E7,$6C,$FF
	.dc.b	$A3,$BE,$3B,$FD,$33,$F5,$8F,$8C,$A7,$FA,$3B,$C2,$F0,$FE,$10,$FB
	.dc.b	$1C,$5C,$CB,$FD,$1F,$39,$7F,$D2,$3F,$84,$BE,$93,$FD,$2E,$D8,$CC
	.dc.b	$63,$D3,$91,$C1,$AF,$F9,$1B,$45,$7F,$28,$8F,$F2,$4E,$0C,$3F,$B0
	.dc.b	$47,$F9,$4E,$31,$EE,$7F,$99,$67,$33,$C1,$FD,$6C,$F1,$F7,$74,$49
	.dc.b	$E5,$3F,$A4,$53,$DF,$C3,$FF,$35,$CF,$F7,$29,$FE,$35,$39,$7F,$CE
	.dc.b	$31,$97,$F2,$9F,$83,$BE,$BF,$FD,$72,$D8,$FF,$5B,$B5,$F6,$FC,$47
	.dc.b	$86,$CD,$26,$1F,$E8,$1E,$07,$D3,$FF,$95,$3E,$3D,$F8,$4F,$FA,$23
	.dc.b	$1F,$E5,$C7,$FA,$F2,$33,$CF,$88,$C4,$ED,$DC,$E9,$D3,$FE,$C0,$FC
	.dc.b	$27,$84,$7F,$92,$3E,$31,$18,$9D,$4F,$09,$CB,$FE,$5E,$FA,$4F,$F2
	.dc.b	$FA,$69,$1F,$3F,$23,$F3,$DF,$F6,$9F,$8B,$FE,$DC,$6C,$FE,$0B,$F1
	.dc.b	$4F,$F6,$4C,$E6,$3F,$D7,$9A,$7F,$AF,$78,$7F,$13,$F4,$18,$AF,$F4
	.dc.b	$E7,$42,$31,$7F,$BB,$C7,$AE,$1C,$6B,$57,$F8,$DF,$CE,$7F,$D1,$CF
	.dc.b	$FA,$B9,$B7,$F3,$B1,$46,$23,$F0,$AD,$7F,$D5,$8F,$EB,$E7,$A7,$6F
	.dc.b	$8A,$B8,$FA,$F6,$B5,$7E,$FF,$92,$D6,$A7,$58,$61,$F5,$F8,$6F,$B9
	.dc.b	$FD,$C6,$7F,$EB,$78,$C7,$83,$F0,$38,$5F,$86,$3E,$3F,$8E,$1C,$34
	.dc.b	$FD,$97,$B8,$FA,$EE,$0D,$CF,$A5,$6F,$8F,$F9,$04,$E6,$FF,$D7,$FA
	.dc.b	$3C,$6F,$F7,$06,$1E,$36,$CF,$EF,$FE,$E9,$F4,$DE,$A9,$CE,$3F,$CA
	.dc.b	$FC,$17,$CE,$FA,$5B,$8F,$3F,$9F,$C4,$FD,$CE,$F7,$EB,$2F,$C2,$3F
	.dc.b	$D6,$DA,$8F,$CC,$C6,$70,$FD,$25,$70,$AD,$D6,$9B,$D7,$FD,$A7,$C4
	.dc.b	$FB,$1E,$F1,$F0,$63,$CE,$FF,$8B,$CE,$71,$FF,$38,$7C,$37,$C7,$7F
	.dc.b	$C4,$4D,$9F,$D5,$ED,$FE,$A2,$9D,$F4,$F4,$FA,$0F,$1E,$BE,$C6,$E5
	.dc.b	$FE,$B9,$AF,$4B,$70,$FC,$A7,$E5,$BF,$76,$7F,$93,$36,$3F,$95,$FE
	.dc.b	$B7,$FE,$7B,$9C,$F5,$7E,$17,$EB,$BF,$86,$C7,$FB,$FF,$B5,$18,$BF
	.dc.b	$E6,$CF,$F9,$A6,$74,$E7,$54,$7C,$FF,$82,$FD,$DB,$FC,$45,$B5,$FC
	.dc.b	$59,$AF,$F8,$D0,$E3,$1B,$E2,$F4,$57,$49,$57,$F3,$6C,$7F,$43,$F9
	.dc.b	$03,$EC,$7F,$7B,$39,$C7,$F0,$18,$71,$87,$E1,$8F,$B5,$DB,$6B,$F5
	.dc.b	$1F,$86,$7C,$BB,$E1,$39,$CF,$FD,$D5,$B3,$DB,$FC,$3E,$5F,$E0,$56
	.dc.b	$C6,$71,$F8,$0D,$AF,$81,$F8,$67,$E6,$9F,$F7,$A3,$71,$FF,$11,$FE
	.dc.b	$26,$F5,$3D,$73,$9C,$3E,$67,$DB,$7F,$84,$DF,$69,$8C,$7F,$81,$FE
	.dc.b	$51,$FE,$37,$8C,$E5,$F2,$2B,$82,$CB,$E2,$7E,$1B,$F6,$E3,$EA,$35
	.dc.b	$FF,$33,$B5,$FB,$8D,$7E,$9F,$E1,$BF,$C5,$6F,$BE,$C8,$C4,$FF,$04
	.dc.b	$FC,$8E,$E1,$F4,$E6,$DF,$E1,$FF,$0E,$F7,$5F,$DE,$6F,$92,$FC,$8B
	.dc.b	$F5,$93,$E9,$AE,$1E,$2F,$E0,$38,$DF,$5F,$F5,$1B,$3F,$E9,$58,$C1
	.dc.b	$F6,$3D,$C7,$DF,$7E,$21,$B3,$FC,$A7,$F2,$4F,$F2,$9B,$EC,$8C,$E3
	.dc.b	$E1,$3F,$27,$D6,$F8,$2F,$C8,$B6,$BF,$99,$7E,$41,$FD,$32,$B8,$71
	.dc.b	$FE,$5E,$FF,$2F,$F8,$86,$E7,$F4,$8D,$7F,$D2,$ED,$7F,$58,$D7,$F4
	.dc.b	$7F,$9F,$7F,$5A,$C3,$F6,$0D,$87,$0F,$FB,$57,$F0,$1F,$D2,$CE,$71
	.dc.b	$7E,$21,$F8,$7B,$F2,$B7,$9B,$FC,$EB,$F8,$E3,$EE,$3F,$5B,$39,$CF
	.dc.b	$F7,$DF,$67,$FE,$CA,$7F,$7C,$DC,$3F,$7C,$D7,$FD,$62,$D7,$F8,$37
	.dc.b	$E2,$3F,$D0,$26,$FA,$9F,$D7,$3F,$20,$FF,$67,$C7,$E3,$27,$55,$FC
	.dc.b	$EF,$F0,$DF,$EE,$7A,$7F,$05,$84,$62,$BF,$A0,$7A,$2F,$C0,$BF,$D6
	.dc.b	$B1,$FA,$4A,$DF,$C2,$7F,$81,$17,$E2,$5F,$DC,$9F,$F1,$5B,$5F,$F6
	.dc.b	$D3,$5F,$E9,$8D,$F3,$FF,$C0,$7C,$FF,$DE,$CE,$91,$FD,$F3,$3F,$F0
	.dc.b	$B7,$C2,$FD,$F7,$F8,$A7,$D7,$BE,$EB,$F8,$CB,$73,$FD,$5B,$F8,$99
	.dc.b	$DC,$FF,$1E,$FC,$7B,$F9,$AD,$7E,$C1,$B9,$FE,$3A,$F9,$3F,$90,$8C
	.dc.b	$DF,$F2,$3F,$83,$B7,$FC,$DD,$B5,$FC,$BB,$F5,$CF,$E6,$6F,$E6,$23
	.dc.b	$3D,$2F,$C8,$3D,$57,$FC,$FD,$B5,$FC,$DA,$B7,$18,$E1,$7B,$6F,$C2
	.dc.b	$7F,$C6,$71,$87,$FC,$45,$E9,$62,$75,$8F,$F8,$CF,$E1,$BF,$C4,$C6
	.dc.b	$6A,$C7,$FA,$0F,$F2,$AF,$3E,$DF,$E4,$36,$C7,$F8,$CA,$BF,$C6,$11
	.dc.b	$8D,$3F,$A5,$FC,$2F,$F7,$3D,$BF,$C3,$1F,$01,$FD,$2B,$F2,$2F,$F4
	.dc.b	$1E,$3F,$C7,$2B,$5F,$82,$7F,$5F,$8F,$EB,$CD,$C3,$9D,$FF,$52,$8C
	.dc.b	$60,$FC,$55,$F6,$5F,$D5,$9B,$87,$F2,$85,$F8,$CE,$67,$FA,$DF,$D0
	.dc.b	$3E,$8E,$99,$BF,$E5,$F1,$97,$F3,$9E,$0F,$F6,$87,$FC,$41,$A7,$E9
	.dc.b	$3E,$0F,$F3,$DF,$8B,$7F,$1A,$E3,$FD,$45,$EE,$BF,$59,$F8,$4F,$F8
	.dc.b	$89,$FD,$5B,$60,$D3,$55,$FE,$F0,$C3,$F5,$8D,$AF,$E4,$C7,$28,$CF
	.dc.b	$D3,$FF,$9C,$97,$D6,$FC,$29,$FE,$59,$18,$D7,$FB,$DF,$E2,$CF,$03
	.dc.b	$85,$EF,$4F,$EF,$9F,$86,$7F,$6D,$6F,$B5,$AC,$3F,$C0,$73,$9C,$7F
	.dc.b	$C8,$5E,$3F,$F9,$8B,$F0,$3F,$F3,$97,$CB,$FF,$88,$F5,$FF,$67,$AF
	.dc.b	$C1,$73,$3A,$F9,$49,$87,$FB,$46,$B8,$69,$FE,$91,$7F,$C8,$4F,$CC
	.dc.b	$CF,$F8,$67,$1F,$9E,$BF,$89,$F8,$7F,$E4,$1C,$FF,$8B,$D7,$F9,$67
	.dc.b	$F2,$7F,$F0,$24,$7D,$8F,$39,$18,$8F,$E3,$6D,$7F,$D6,$6D,$DB,$FC
	.dc.b	$FB,$F3,$5E,$C3,$A7,$C0,$DC,$BF,$C0,$63,$D2,$F9,$8E,$1B,$87,$F5
	.dc.b	$6F,$C5,$7F,$D9,$B3,$9E,$6F,$F1,$6F,$D7,$7F,$BF,$E3,$FB,$FB,$6B
	.dc.b	$EC,$DF,$2E,$0F,$F1,$84,$E5,$FF,$19,$7E,$29,$FD,$FB,$C2,$34,$FC
	.dc.b	$44,$6B,$3F,$F1,$DF,$E2,$2F,$C5,$69,$FE,$3E,$AE,$51,$9C,$FF,$C2
	.dc.b	$3F,$8A,$7F,$85,$71,$34,$9F,$5D,$FE,$38,$7F,$C2,$3A,$FF,$93,$1B
	.dc.b	$3C,$27,$FC,$97,$1A,$F8,$17,$E2,$9E,$BC,$3F,$D1,$91,$8F,$B0,$DF
	.dc.b	$C3,$FF,$AD,$2B,$59,$75,$FC,$47,$99,$9A,$B8,$7A,$3F,$00,$F2,$11
	.dc.b	$FF,$34,$65,$FD,$8A,$FF,$CD,$0F,$FB,$89,$FF,$05,$B7,$87,$FC,$BC
	.dc.b	$ED,$BF,$88,$FA,$A7,$E0,$12,$BA,$BF,$0F,$D5,$F7,$9B,$8F,$96,$FD
	.dc.b	$2F,$81,$91,$FE,$3E,$E1,$7D,$BE,$2F,$C7,$E3,$FC,$83,$19,$79,$1F
	.dc.b	$B7,$EC,$7F,$9A,$2F,$FC,$BD,$87,$FA,$15,$B7,$FE,$85,$6B,$FC,$6B
	.dc.b	$93,$ED,$FD,$E7,$D4,$39,$D7,$E0,$1F,$EF,$5F,$8A,$BE,$E1,$F3,$3F
	.dc.b	$EC,$1F,$B0,$FF,$43,$63,$FE,$C3,$AD,$4E,$7E,$43,$F9,$3C,$7F,$9D
	.dc.b	$EB,$74,$8D,$A8,$CC,$61,$B7,$E0,$9D,$57,$4F,$3B,$FD,$35,$5A,$C3
	.dc.b	$C1,$5E,$3B,$E1,$39,$8E,$27,$9E,$B8,$DF,$9A,$B6,$DB,$E7,$5E,$79
	.dc.b	$CB,$C7,$2C,$E3,$E7,$BE,$13,$D8,$E1,$7A,$C3,$BC,$79,$63,$8E,$1D
	.dc.b	$23,$67,$1B,$F9,$17,$D3,$1F,$51,$1F,$85,$8F,$2C,$25,$F1,$C3,$A6
	.dc.b	$1C,$A3,$C0,$66,$77,$87,$D0,$C7,$CB,$3E,$28,$EC,$F0,$75,$87,$0C
	.dc.b	$7C,$84,$9F,$88,$75,$24,$F2,$7F,$01,$5D,$B0,$F9,$11,$E4,$3F,$B6
	.dc.b	$9F,$08,$AF,$E1,$9B,$8F,$E0,$DC,$FA,$AF,$17,$8C,$7E,$4E,$30,$E2
	.dc.b	$BF,$C0,$13,$A8,$FC,$94,$E9,$E7,$1F,$72,$38,$E1,$F9,$46,$E3,$F9
	.dc.b	$CF,$C0,$7F,$C0,$73,$9A,$E0,$4B,$FB,$53,$AA,$FD,$1D,$F2,$9F,$35
	.dc.b	$6E,$FF,$43,$7B,$9C,$9F,$83,$AE,$0F,$A2,$BF,$23,$3B,$8C,$1B,$8E
	.dc.b	$97,$E7,$8A,$74,$BF,$43,$E6,$32,$72,$72,$79,$1C,$E1,$F2,$9E,$C7
	.dc.b	$12,$C7,$09,$F7,$D7,$9A,$78,$E1,$EA,$BE,$57,$D6,$78,$DF,$06,$9F
	.dc.b	$C3,$70,$F7,$57,$09,$D5,$77,$AD,$1F,$C0,$B6,$FE,$AF,$E6,$FF,$8E
	.dc.b	$38,$FF,$9D,$B7,$65,$7F,$57,$8C,$2A,$70,$FF,$B4,$E3,$F9,$B7,$C3
	.dc.b	$FE,$26,$72,$D9,$D4,$FF,$2F,$8D,$74,$BC,$FE,$B7,$FD,$A9,$B1,$9A
	.dc.b	$FE,$85,$EC,$6B,$0F,$F2,$CC,$60,$87,$FB,$81,$FF,$30,$BF,$DD,$BD
	.dc.b	$1E,$64,$7F,$A6,$7E,$07,$CC,$F3,$11,$FE,$8C,$F9,$BB,$FC,$64,$8F
	.dc.b	$EA,$9F,$8B,$78,$CF,$71,$1F,$E9,$CF,$BB,$CB,$F5,$92,$3E,$6F,$03
	.dc.b	$83,$FB,$EF,$87,$FB,$E8,$C4,$FF,$BB,$23,$13,$A8,$FA,$7E,$2F,$F8
	.dc.b	$19,$F8,$9F,$69,$F5,$BB,$1D,$CF,$C4,$B0,$DA,$F0,$E1,$F7,$AB,$75
	.dc.b	$AF,$89,$95,$FC,$22,$79,$C7,$F0,$56,$C6,$6F,$FD,$98,$FF,$0E,$E5
	.dc.b	$FF,$17,$FD,$AF,$DC,$F2,$CD,$3E,$FF,$9C,$F0,$E9,$2F,$FC,$7E,$EB
	.dc.b	$AF,$F0,$BE,$87,$E0,$27,$FC,$91,$D3,$83,$F2,$BB,$BF,$33,$7E,$21
	.dc.b	$86,$D7,$F9,$51,$AF,$FA,$09,$BC,$BF,$E7,$EF,$C0,$BF,$3F,$E2,$BF
	.dc.b	$5D,$7D,$EE,$1B,$87,$CA,$3F,$97,$38,$57,$F6,$BD,$87,$E4,$FF,$E9
	.dc.b	$C9,$E1,$18,$F8,$2F,$E6,$0D,$8F,$E6,$ED,$7F,$A0,$B5,$FE,$87,$F8
	.dc.b	$5F,$EB,$2F,$F0,$F0,$8C,$6D,$FE,$B4,$6C,$29,$F5,$4E,$4F,$67,$C6
	.dc.b	$CA,$F6,$6D,$F0,$63,$E0,$74,$C9,$CE,$DF,$8E,$6C,$7C,$5F,$4B,$F7
	.dc.b	$7F,$06,$F5,$69,$2B,$FD,$0B,$F9,$9B,$EF,$FE,$CB,$63,$31,$83,$F8
	.dc.b	$2F,$C4,$BE,$33,$F8,$9C,$BE,$F3,$F7,$EF,$FB,$09,$BC,$9F,$11,$BC
	.dc.b	$9F,$ED,$6F,$C3,$1E,$E7,$26,$DF,$F6,$16,$BF,$DA,$1A,$FF,$72,$6B
	.dc.b	$CF,$DE,$E4,$47,$F0,$83,$F1,$8F,$8C,$FE,$12,$FF,$96,$F8,$AF,$C0
	.dc.b	$7F,$DA,$B3,$F8,$B8,$C7,$BB,$FC,$92,$73,$3C,$4F,$E6,$5F,$20,$FA
	.dc.b	$2F,$E5,$2D,$8E,$4D,$78,$5A,$31,$B7,$F0,$E8,$C7,$CA,$FF,$32,$F9
	.dc.b	$47,$9F,$49,$87,$97,$DE,$FB,$8D,$E7,$FE,$73,$18,$8E,$ED,$C7,$B2
	.dc.b	$F8,$3F,$D6,$F2,$7F,$85,$E7,$F8,$FF,$77,$DB,$F3,$F2,$7E,$CB,$CE
	.dc.b	$FE,$85,$8F,$C1,$9F,$E1,$7E,$DE,$49,$1F,$B3,$F8,$C7,$E9,$2F,$63
	.dc.b	$FA,$CC,$61,$B1,$99,$E0,$D7,$B6,$0F,$47,$FA,$AF,$CC,$FE,$46,$7E
	.dc.b	$B3,$5F,$9A,$DD,$3E,$67,$CA,$3F,$00,$FE,$BA,$DA,$FB,$CD,$7E,$AB
	.dc.b	$67,$8E,$BF,$D9,$3E,$6D,$EE,$D2,$5F,$E1,$FF,$18,$FE,$49,$7F,$6D
	.dc.b	$BD,$DF,$8E,$6C,$FF,$79,$6F,$51,$FD,$2C,$7F,$91,$F6,$FE,$68,$FF
	.dc.b	$02,$6B,$F1,$3D,$8F,$D8,$FE,$31,$F1,$63,$F3,$ED,$7F,$C9,$ED,$7F
	.dc.b	$CB,$2D,$7F,$C1,$A1,$C3,$FE,$10,$F0,$FD,$EF,$EA,$9A,$FE,$84,$38
	.dc.b	$CB,$F9,$77,$9F,$F4,$73,$FD,$6F,$E2,$CF,$B2,$FF,$06,$B7,$0F,$D7
	.dc.b	$35,$FD,$67,$BE,$F1,$7E,$1E,$EF,$C5,$BF,$AE,$D7,$F9,$CA,$31,$EF
	.dc.b	$FF,$87,$E7,$33,$FE,$C9,$6B,$FB,$EF,$90,$FE,$E0,$FE,$C9,$BA,$FE
	.dc.b	$C5,$BA,$7F,$BF,$BE,$A1,$E5,$FF,$89,$3F,$15,$FF,$67,$61,$FC,$15
	.dc.b	$AF,$EF,$1B,$87,$F8,$23,$57,$E2,$DC,$AB,$F8,$67,$E2,$AF,$9E,$FF
	.dc.b	$1B,$36,$3F,$BC,$35,$F7,$35,$FF,$72,$7B,$8F,$C1,$7F,$C6,$ED,$8F
	.dc.b	$E3,$0D,$7E,$5F,$E5,$0F,$AD,$FF,$1E,$FE,$07,$FD,$97,$4F,$E2,$0D
	.dc.b	$E0,$FE,$46,$DB,$6A,$FD,$79,$F6,$BF,$E4,$6F,$99,$FF,$4F,$57,$FB
	.dc.b	$DD,$AF,$F2,$C6,$BF,$E5,$76,$F3,$BE,$B7,$F9,$9E,$7F,$C8,$5B,$9F
	.dc.b	$F3,$46,$DF,$8B,$D2,$F6,$57,$F2,$7F,$67,$F0,$6D,$EA,$FF,$2C,$C6
	.dc.b	$30,$FF,$34,$EB,$FC,$27,$E0,$BF,$CB,$BF,$8B,$F8,$72,$FF,$13,$35
	.dc.b	$FC,$63,$5F,$D2,$E8,$8C,$47,$F4,$8F,$C5,$DF,$37,$FE,$68,$6C,$7F
	.dc.b	$49,$6B,$F9,$CF,$A3,$F9,$97,$FE,$79,$FC,$43,$FC,$DE,$9A,$7F,$93
	.dc.b	$3F,$00,$7C,$EF,$F9,$D3,$F1,$7F,$EA,$38,$79,$1A,$FF,$5A,$6B,$FE
	.dc.b	$85,$6D,$BF,$17,$ED,$3F,$D7,$5A,$FF,$54,$6F,$17,$E6,$7F,$0A,$BC
	.dc.b	$3F,$D9,$A3,$F8,$61,$FD,$5C,$AE,$7D,$5F,$F3,$1F,$E0,$DF,$ED,$79
	.dc.b	$FF,$34,$FB,$6F,$CF,$DF,$4B,$FE,$59,$6E,$7F,$E5,$D6,$BF,$EE,$E9
	.dc.b	$E1,$3F,$E5,$FF,$C5,$BF,$C5,$B7,$D1,$B9,$7F,$9A,$5B,$6B,$7E,$82
	.dc.b	$FC,$2B,$FC,$97,$F8,$5F,$F7,$AC,$7F,$AF,$B7,$1F,$F5,$7F,$B4,$8C
	.dc.b	$63,$FE,$F9,$FA,$FF,$65,$7F,$AD,$3E,$59,$FA,$DB,$E8,$FF,$C9,$2D
	.dc.b	$BF,$F2,$4B,$5F,$F2,$4B,$5F,$F0,$D7,$17,$F8,$84,$8F,$F9,$53,$EA
	.dc.b	$5F,$4B,$FE,$62,$FC,$0F,$96,$DF,$DD,$9B,$9F,$A1,$B7,$FE,$4D,$C7
	.dc.b	$FB,$BD,$7F,$98,$5F,$EF,$31,$FD,$03,$3F,$4F,$C4,$7A,$09,$1F,$E8
	.dc.b	$FF,$31,$F1,$DF,$E9,$B6,$C7,$F9,$BF,$93,$F8,$64,$FF,$90,$BE,$55
	.dc.b	$9D,$25,$7F,$8C,$A3,$2F,$1F,$C5,$9F,$91,$FF,$A9,$FE,$E9,$F6,$94
	.dc.b	$95,$FE,$7D,$F2,$9F,$79,$FE,$67,$8C,$35,$FE,$92,$DE,$1F,$E6,$4D
	.dc.b	$E2,$FF,$57,$7E,$2B,$FC,$CE,$BF,$C3,$5F,$23,$F1,$A7,$F0,$2D,$87
	.dc.b	$5F,$F9,$77,$8F,$FD,$27,$6F,$4A,$92,$BF,$CE,$F1,$96,$FD,$97,$03
	.dc.b	$FB,$7B,$FD,$C2,$D3,$FD,$2A,$30,$D7,$F2,$6D,$7C,$6B,$F0,$FF,$19
	.dc.b	$23,$EA,$7E,$2F,$FE,$29,$D7,$F6,$2D,$D3,$FD,$9C,$D9,$FD,$33,$70
	.dc.b	$FF,$3D,$7C,$57,$F9,$F3,$E6,$FF,$D0,$98,$7A,$2C,$7F,$99,$F9,$7F
	.dc.b	$D5,$AE,$67,$6D,$8C,$BF,$02,$3F,$CB,$7D,$5C,$E3,$FC,$E1,$EB,$7F
	.dc.b	$A8,$DE,$4D,$78,$35,$F1,$77,$71,$7C,$6D,$4D,$3D,$19,$7F,$A5,$7E
	.dc.b	$63,$FD,$6B,$3F,$79,$AF,$DE,$E3,$FF,$53,$6D,$C1,$B1,$FC,$4F,$F1
	.dc.b	$7F,$83,$3F,$EF,$36,$C6,$4E,$7D,$A8,$CC,$7D,$FF,$C0,$7F,$17,$5E
	.dc.b	$46,$C7,$E1,$31,$46,$23,$84,$79,$8C,$FB,$56,$7F,$B5,$57,$38,$C1
	.dc.b	$9F,$94,$91,$DD,$B7,$FC,$B7,$EF,$1F,$2B,$E8,$6D,$7D,$FC,$D1,$8A
	.dc.b	$E1,$5F,$23,$B3,$F6,$7F,$41,$7B,$7F,$D8,$98,$FE,$1E,$72,$D8,$CD
	.dc.b	$FF,$3D,$FC,$5B,$FC,$35,$3D,$DB,$19,$F8,$5F,$97,$18,$8F,$F2,$BD
	.dc.b	$BF,$01,$7D,$37,$8D,$B8,$78,$5B,$5F,$B0,$AE,$15,$F8,$08,$7C,$2F
	.dc.b	$F1,$2B,$FC,$A3,$63,$F4,$0D,$7F,$72,$D7,$FA,$17,$E2,$9F,$E9,$59
	.dc.b	$FD,$37,$4A,$31,$3F,$C0,$23,$1E,$6F,$E4,$1B,$5C,$3B,$5F,$77,$96
	.dc.b	$C6,$63,$07,$A7,$1F,$97,$87,$93,$83,$04,$AF,$D0,$FA,$2F,$1B,$F1
	.dc.b	$4D,$8C,$C6,$0F,$CC,$FA,$4F,$90,$A4,$9F,$C2,$46,$5F,$BD,$F9,$33
	.dc.b	$EB,$FF,$08,$D8,$CC,$60,$FC,$47,$B5,$FE,$1C,$8F,$C3,$37,$3F,$AC
	.dc.b	$D7,$EF,$B5,$FF,$1D,$FE,$29,$FE,$D9,$9F,$C7,$36,$3E,$F3,$5F,$C3
	.dc.b	$B5,$E3,$E1,$7E,$66,$3F,$7B,$1A,$8C,$1F,$9D,$F8,$17,$D6,$52,$5F
	.dc.b	$E7,$63,$2F,$E7,$FF,$13,$79,$5F,$A5,$6C,$66,$30,$7E,$A7,$EC,$73
	.dc.b	$8F,$E5,$5A,$FE,$79,$AF,$F9,$69,$AF,$F9,$D3,$F1,$2F,$EC,$F5,$FA
	.dc.b	$F6,$C7,$E7,$DA,$FE,$A1,$AF,$F9,$93,$F1,$07,$E2,$DF,$BB,$6D,$7E
	.dc.b	$E5,$B3,$FA,$96,$BF,$B2,$FD,$7F,$F8,$B3,$FE,$6E,$6B,$FE,$88,$6F
	.dc.b	$71,$BF,$89,$3E,$DF,$F8,$53,$70,$FE,$10,$DB,$FD,$C3,$5F,$E0,$3F
	.dc.b	$89,$3E,$97,$F8,$93,$6B,$F8,$83,$67,$F8,$03,$5F,$E1,$3F,$97,$7F
	.dc.b	$9E,$27,$F8,$B3,$63,$F8,$33,$5D,$FC,$8F,$F1,$FF,$9C,$7D,$95,$25
	.dc.b	$FF,$1F,$8C,$BF,$C8,$7E,$83,$FD,$60,$DF,$84,$FE,$51,$18,$C3,$F8
	.dc.b	$A3,$5F,$E3,$BF,$84,$7E,$B5,$FE,$30,$D7,$3D,$FF,$CC,$63,$11,$FE
	.dc.b	$9E,$5F,$31,$92,$61,$FC,$CE,$32,$FF,$16,$FA,$F7,$DC,$FF,$13,$6C
	.dc.b	$66,$30,$7F,$27,$FC,$27,$FA,$7E,$1F,$E0,$36,$BF,$E1,$C6,$BF,$E0
	.dc.b	$60,$E1,$FF,$04,$FE,$1C,$F5,$BF,$99,$36,$FF,$C0,$8D,$9F,$F0,$83
	.dc.b	$5F,$F5,$9F,$C6,$FD,$B7,$FC,$27,$F2,$BF,$E7,$1E,$5F,$EF,$6D,$BF
	.dc.b	$F0,$C3,$67,$FC,$62,$FF,$9D,$FE,$77,$76,$69,$2B,$8F,$D8,$1E,$AD
	.dc.b	$72,$60,$F0,$FA,$FF,$E8,$F7,$C2,$DC,$3C,$2D,$7F,$D9,$AD,$7F,$D2
	.dc.b	$3E,$E7,$F0,$C7,$FC,$08,$D7,$FC,$15,$E2,$FF,$80,$4F,$F0,$CF,$03
	.dc.b	$EC,$FF,$CE,$F8,$7F,$8D,$E3,$1F,$1F,$A9,$CD,$FF,$94,$A7,$35,$FE
	.dc.b	$13,$F9,$FF,$F0,$61,$1C,$1F,$2B,$ED,$7F,$7B,$B9,$A4,$F8,$1F,$A3
	.dc.b	$ED,$FF,$54,$C7,$EA,$B5,$57,$F7,$9F,$53,$67,$D8,$FC,$35,$F8,$05
	.dc.b	$26,$5F,$6A,$32,$DF,$96,$E0,$47,$07,$F9,$37,$F0,$DF,$F4,$9B,$7E
	.dc.b	$57,$FC,$01,$18,$CB,$FC,$B8,$D7,$F3,$5F,$07,$F9,$F2,$57,$F9,$5F
	.dc.b	$B3,$FC,$45,$3F,$9F,$FB,$FF,$D2,$12,$74,$FE,$4B,$EF,$BF,$D8,$91
	.dc.b	$F9,$26,$CF,$E4,$9A,$FE,$9F,$64,$62,$3F,$C9,$FF,$5E,$D3,$FC,$AF
	.dc.b	$68,$C3,$FE,$75,$31,$1A,$8C,$CF,$F9,$37,$E1,$DF,$83,$7F,$92,$DB
	.dc.b	$7F,$E6,$66,$CE,$E7,$5E,$F4,$AD,$36,$7F,$C3,$9C,$FF,$77,$E0,$CD
	.dc.b	$27,$27,$F0,$9C,$E6,$1F,$2D,$B7,$FE,$25,$FA,$57,$E1,$9F,$E2,$56
	.dc.b	$D7,$F8,$B1,$B3,$B9,$FD,$FF,$E0,$8F,$93,$A4,$BF,$E0,$31,$97,$E3
	.dc.b	$79,$3F,$36,$FF,$78,$D8,$34,$D7,$13,$F0,$57,$D8,$7F,$0E,$6D,$7F
	.dc.b	$0E,$6D,$FF,$0E,$6B,$F7,$5F,$4B,$72,$E4,$49,$FF,$34,$7E,$0F,$FE
	.dc.b	$35,$9F,$F2,$C3,$5F,$F3,$0B,$5F,$F5,$9F,$B2,$8C,$47,$EB,$BF,$09
	.dc.b	$FE,$8A,$DF,$43,$F9,$2C,$63,$33,$47,$DA,$68,$98,$FF,$2A,$8C,$BE
	.dc.b	$6F,$BC,$FF,$41,$62,$69,$23,$D7,$86,$88,$E0,$E9,$E9,$79,$08,$E0
	.dc.b	$F0,$7C,$9F,$A0,$8E,$0F,$26,$7E,$C2,$38,$3F,$A0,$FE,$12,$FA,$8F
	.dc.b	$E8,$AD,$8C,$C6,$0C,$FC,$9C,$8E,$2D,$AF,$E9,$34,$F2,$B9,$92,$BB
	.dc.b	$BD,$BC,$D9,$1E,$26,$D7,$FB,$AF,$F0,$B7,$D0,$33,$FF,$6D,$FD,$77
	.dc.b	$F6,$99,$CC,$62,$BF,$DE,$30,$FA,$2F,$EA,$91,$FE,$E0,$AE,$17,$CE
	.dc.b	$30,$70,$71,$F3,$A9,$DE,$1D,$69,$87,$5C,$B0,$F8,$A3,$99,$AC,$39
	.dc.b	$DF,$28,$CC,$F1,$6C,$E9,$F1,$5F,$48,$C4,$78,$31,$F5,$E3,$E5,$8D
	.dc.b	$9C,$AB,$81,$E4,$9D,$E1,$E1,$AE,$15,$E5,$AD,$BE,$4A,$DB,$D2,$77
	.dc.b	$87,$B5,$B8,$79,$23,$D5,$18,$BE,$6D,$AE,$98,$E3,$1F,$11,$23,$EB
	.dc.b	$E3,$E0,$7C,$87,$AC,$95,$E2,$8D,$61,$E3,$AE,$55,$C2,$B5,$7D,$E3
	.dc.b	$D5,$3A,$AF,$3B,$71,$FC,$14,$FF,$8C,$CF,$09,$67,$EE,$36,$36,$FE
	.dc.b	$47,$E0,$1E,$87,$18,$FC,$8C,$61,$FC,$D3,$5E,$8D,$79,$47,$E2,$A3
	.dc.b	$86,$3F,$41,$B8,$7E,$7F,$F0,$2F,$F2,$44,$E6,$B8,$12,$FF,$1B,$3A
	.dc.b	$AE,$B1,$F9,$0B,$E6,$F9,$AB,$F3,$35,$B8,$C1,$B8,$E9,$7F,$82,$9D
	.dc.b	$4F,$E8,$AB,$9B,$6B,$77,$E7,$8A,$74,$BF,$43,$E6,$32,$71,$3C,$06
	.dc.b	$4A,$F0,$24,$75,$8C,$BE,$E8,$CC,$7F,$02,$BF,$3E,$2C,$3C,$ED,$C3
	.dc.b	$C2,$A7,$73,$A2,$4F,$C6,$AD,$C6,$1F,$35,$FE,$CF,$A1,$C0,$FC,$2F
	.dc.b	$F0,$95,$FE,$3C,$8C,$3F,$CC,$23,$33,$C1,$B2,$FB,$1A,$4A,$FE,$6E
	.dc.b	$EB,$D2,$91,$FC,$E9,$AF,$F3,$2F,$13,$99,$1F,$E7,$EF,$6F,$AB,$91
	.dc.b	$FD,$09,$AF,$F3,$AF,$94,$F1,$91,$FE,$8E,$F9,$BF,$03,$91,$FD,$29
	.dc.b	$AF,$F7,$4E,$AF,$83,$E0,$7F,$71,$6F,$77,$FA,$CA,$7F,$BD,$C6,$18
	.dc.b	$7F,$7E,$29,$FD,$33,$E8,$FE,$29,$1F,$EB,$2F,$CE,$E9,$75,$7F,$A5
	.dc.b	$E7,$F6,$91,$8F,$7B,$53,$99,$FE,$01,$39,$7F,$75,$8B,$D4,$7D,$AF
	.dc.b	$A9,$B3,$A6,$BF,$C2,$67,$84,$FD,$CF,$84,$7C,$95,$25,$7F,$AC,$23
	.dc.b	$2D,$F9,$EE,$04,$70,$7F,$96,$7F,$09,$7E,$15,$F8,$CF,$05,$18,$9D
	.dc.b	$4E,$63,$11,$FE,$BC,$FC,$35,$F2,$3F,$90,$6C,$7F,$B0,$9A,$EE,$3F
	.dc.b	$33,$EA,$BE,$59,$EE,$61,$B5,$FC,$86,$75,$39,$3F,$43,$0F,$BB,$FD
	.dc.b	$0B,$FC,$9D,$AF,$F3,$16,$BF,$EE,$56,$A9,$7E,$03,$FE,$F4,$8F,$03
	.dc.b	$5F,$E7,$51,$AD,$BD,$33,$99,$FF,$78,$78,$8F,$C2,$FF,$0F,$1F,$CE
	.dc.b	$63,$1C,$DF,$B8,$6C,$66,$76,$FF,$47,$8E,$5E,$66,$E4,$D2,$4F,$FA
	.dc.b	$47,$F0,$BF,$F3,$93,$7D,$8F,$B5,$18,$8F,$EA,$5C,$68,$C4,$7F,$A8
	.dc.b	$7F,$0B,$7B,$7F,$C2,$1B,$1F,$EF,$A6,$BF,$EF,$A6,$BF,$D7,$7F,$05
	.dc.b	$FD,$83,$FD,$5D,$AE,$7B,$FE,$BC,$62,$04,$7F,$CC,$2F,$F1,$C6,$BC
	.dc.b	$FD,$6E,$27,$E0,$3F,$0B,$F8,$F1,$FD,$B5,$AF,$40,$E1,$EE,$DF,$2D
	.dc.b	$F8,$5B,$E5,$FF,$93,$B6,$BF,$B8,$4E,$A7,$26,$BE,$36,$4E,$A9,$18
	.dc.b	$8F,$F0,$05,$7F,$41,$FA,$1F,$E6,$7E,$91,$A4,$9E,$F0,$F8,$ED,$92
	.dc.b	$32,$FE,$27,$E4,$DF,$0B,$49,$3F,$8F,$8C,$BF,$E0,$9F,$85,$7E,$16
	.dc.b	$F3,$BF,$A4,$C6,$1B,$19,$9E,$0D,$7F,$29,$4F,$1B,$C4,$48,$F0,$3F
	.dc.b	$D4,$3C,$A7,$CB,$52,$4F,$CC,$8C,$B7,$B7,$81,$1C,$1F,$4B,$F0,$A7
	.dc.b	$A9,$FD,$71,$B1,$98,$C1,$9F,$52,$47,$16,$BF,$E2,$BF,$C2,$BD,$4F
	.dc.b	$F6,$A6,$B9,$CB,$F8,$6C,$62,$3F,$3D,$F0,$5F,$9A,$23,$F9,$77,$F8
	.dc.b	$1F,$6C,$8F,$E1,$2D,$7F,$9D,$7C,$63,$0F,$EB,$2D,$8F,$E8,$AD,$77
	.dc.b	$1A,$F5,$A4,$69,$AF,$F4,$5F,$99,$FD,$3B,$FA,$1F,$4D,$18,$8F,$F6
	.dc.b	$BB,$54,$6B,$BB,$FB,$5F,$CD,$E6,$0D,$24,$7F,$70,$E5,$FC,$B5,$7F
	.dc.b	$5C,$6B,$FD,$5D,$AF,$8D,$AF,$FB,$A3,$F0,$57,$CE,$7F,$7A,$6C,$7F
	.dc.b	$66,$6B,$FD,$85,$AF,$B7,$F0,$5F,$EC,$AD,$F5,$FF,$C0,$71,$88,$FF
	.dc.b	$7A,$B5,$F7,$FD,$BF,$FB,$01,$FE,$E6,$D7,$FB,$73,$5F,$94,$D7,$FD
	.dc.b	$FF,$ED,$BE,$CB,$FC,$22,$D8,$FE,$FA,$D7,$FB,$C3,$5F,$A9,$CB,$FE
	.dc.b	$D9,$7F,$C0,$6D,$7F,$BF,$35,$DF,$C4,$7F,$87,$BF,$2E,$FC,$24,$7D
	.dc.b	$D6,$BF,$CA,$B8,$11,$88,$FB,$AD,$87,$C9,$38,$FF,$C5,$2D,$AF,$F0
	.dc.b	$AB,$5F,$F0,$83,$5F,$C6,$7D,$4F,$FA,$F9,$BF,$13,$FE,$32,$8C,$47
	.dc.b	$8D,$AF,$E3,$BF,$02,$FE,$DF,$81,$A4,$F8,$7F,$F1,$DB,$FE,$55,$9F
	.dc.b	$F1,$3B,$5F,$F1,$13,$5F,$CF,$35,$F8,$5F,$59,$FE,$A1,$6F,$81,$FE
	.dc.b	$46,$8C,$47,$E8,$1B,$6F,$C7,$9F,$29,$FE,$4C,$6D,$7F,$8E,$1A,$FF
	.dc.b	$8C,$DA,$FF,$4F,$F8,$57,$E0,$2F,$B0,$FF,$24,$36,$7F,$C8,$ED,$7F
	.dc.b	$C8,$AD,$7F,$C1,$3D,$FE,$6C,$FF,$83,$1F,$E0,$77,$07,$AA,$1E,$A7
	.dc.b	$F1,$57,$FB,$03,$5F,$ED,$9C,$BF,$DC,$0F,$F1,$7F,$C6,$BE,$2E,$92
	.dc.b	$7F,$C5,$D1,$97,$FC,$61,$F8,$0F,$F4,$A7,$FC,$42,$D7,$FB,$77,$C1
	.dc.b	$23,$11,$FD,$F2,$31,$E8,$7F,$80,$3E,$2A,$4F,$F7,$D6,$BF,$ED,$DF
	.dc.b	$79,$F3,$FF,$E3,$C6,$C7,$FA,$79,$AE,$E3,$FC,$E9,$F8,$0B,$F0,$1A
	.dc.b	$49,$FF,$3A,$C6,$5F,$F3,$8F,$D3,$7F,$8D,$DF,$D4,$35,$FE,$46,$DE
	.dc.b	$73,$9E,$57,$DF,$FF,$29,$6F,$95,$FE,$85,$8C,$1F,$CB,$79,$77,$F2
	.dc.b	$FF,$E3,$F8,$34,$98,$FF,$91,$63,$F8,$73,$FF,$00,$1B,$3E,$0F,$97
	.dc.b	$E4,$73,$F9,$69,$5C,$5A,$FF,$A4,$3E,$FF,$FB,$C3,$F9,$56,$BF,$E3
	.dc.b	$0C,$91,$88,$46,$3C,$EF,$F2,$B7,$DF,$7F,$A2,$A7,$FC,$A8,$D7,$FC
	.dc.b	$BE,$D7,$FC,$77,$8A,$31,$1F,$E3,$EF,$BF,$71,$BE,$EF,$FD,$B3,$18
	.dc.b	$6C,$66,$7F,$C7,$47,$17,$0E,$25,$E5,$FF,$9E,$63,$62,$33,$D1,$DA
	.dc.b	$75,$3C,$5A,$F0,$EA,$FF,$45,$A6,$91,$FF,$3A,$7C,$9B,$E7,$5F,$19
	.dc.b	$FE,$5A,$CE,$D3,$5F,$F5,$1D,$62,$73,$1F,$EA,$6E,$5C,$5F,$C8,$C3
	.dc.b	$A3,$CC,$C5,$EB,$0E,$9F,$80,$3F,$1C,$C6,$3D,$AF,$85,$6A,$73,$FA
	.dc.b	$83,$E5,$3C,$B3,$FE,$96,$8C,$57,$FA,$55,$B1,$FE,$91,$6B,$BE,$EE
	.dc.b	$67,$4A,$B5,$E0,$C1,$7F,$BA,$8C,$7E,$B9,$FE,$1F,$75,$19,$C3,$F0
	.dc.b	$51,$F8,$89,$D4,$60,$FC,$9F,$D4,$7F,$B1,$27,$DB,$CE,$8C,$4E,$E3
	.dc.b	$1F,$89,$7C,$77,$5F,$33,$6B,$7C,$5F,$5A,$5E,$B3,$E9,$75,$7C,$63
	.dc.b	$19,$7C,$CF,$09,$F5,$FF,$E9,$88,$C3,$74,$F1,$7E,$05,$FD,$06,$73
	.dc.b	$1E,$3F,$C7,$BF,$A6,$C7,$A9,$B1,$EF,$BE,$AE,$FD,$2F,$57,$E5,$DF
	.dc.b	$E2,$6B,$FC,$EE,$A8,$C5,$7E,$9A,$31,$F0,$3F,$7A,$3F,$7D,$3A,$7F
	.dc.b	$AB,$7E,$65,$FE,$91,$BF,$96,$D8,$F9,$53,$D3,$83,$E3,$4E,$5F,$91
	.dc.b	$F9,$7B,$8F,$E7,$37,$0F,$AA,$FE,$F9,$B1,$F4,$FF,$3A,$D6,$3E,$76
	.dc.b	$BF,$DA,$5A,$FF,$15,$6B,$FC,$53,$F3,$07,$D0,$7D,$D7,$EC,$DF,$E1
	.dc.b	$1C,$B7,$8F,$F0,$5E,$C3,$EE,$69,$FA,$BE,$27,$E0,$9F,$90,$DB,$F0
	.dc.b	$95,$BC,$3F,$57,$F3,$3F,$ED,$6F,$C5,$BE,$66,$A8,$C4,$FF,$B1,$A1
	.dc.b	$3A,$9C,$C7,$FB,$0B,$BB,$F2,$35,$FE,$E0,$9F,$E0,$DC,$1F,$BD,$7F
	.dc.b	$84,$75,$5A,$F8,$F9,$7F,$0C,$97,$D2,$7F,$01,$C3,$CA,$FE,$5B,$1F
	.dc.b	$D4,$4E,$BC,$2E,$93,$9B,$E5,$D0,$F4,$3C,$CD,$C7,$FD,$BE,$DC,$7F
	.dc.b	$DB,$2D,$7F,$DD,$98,$BF,$22,$FC,$CE,$1E,$56,$DF,$F9,$07,$E9,$3D
	.dc.b	$91,$8B,$F9,$1B,$BE,$8B,$E0,$2F,$2A,$C6,$1F,$89,$F9,$17,$D3,$AF
	.dc.b	$F7,$5B,$5F,$1B,$5F,$98,$D7,$EB,$7E,$45,$FE,$F4,$CB,$ED,$B6,$3F
	.dc.b	$61,$F8,$37,$FC,$01,$9C,$FD,$17,$D3,$FC,$8B,$D1,$B9,$F5,$74,$BF
	.dc.b	$06,$7C,$C6,$FC,$5F,$B5,$EE,$7E,$BF,$4F,$BC,$D7,$F1,$0D,$C7,$FC
	.dc.b	$07,$E3,$7E,$1F,$81,$F5,$FF,$83,$5F,$27,$63,$0F,$E1,$FE,$3F,$F3
	.dc.b	$F8,$FE,$01,$FD,$1B,$7C,$5D,$78,$9F,$8D,$FC,$8B,$73,$F6,$9A,$FE
	.dc.b	$4D,$AF,$E7,$5B,$9B,$F2,$2F,$F0,$2F,$27,$E9,$1B,$5F,$C6,$7E,$67
	.dc.b	$F8,$44,$FE,$76,$73,$DB,$F9,$9F,$C8,$9F,$39,$F9,$F6,$D7,$F2,$DF
	.dc.b	$9C,$FE,$51,$8F,$E6,$A7,$2F,$E7,$3F,$22,$FE,$8A,$FE,$49,$AF,$EA
	.dc.b	$DA,$FE,$D5,$B9,$BF,$22,$7D,$D7,$EB,$FE,$1F,$F9,$C6,$9F,$AD,$9C
	.dc.b	$E7,$FC,$02,$72,$B1,$9F,$AE,$3F,$22,$FE,$E9,$08,$C4,$FF,$72,$87
	.dc.b	$EA,$0F,$93,$DD,$6A,$7F,$AF,$1C,$7E,$E7,$F8,$6B,$6F,$FC,$36,$D3
	.dc.b	$FC,$3B,$29,$DF,$BA,$FC,$5B,$C6,$DB,$F1,$56,$DF,$24,$66,$FF,$83
	.dc.b	$FA,$AF,$C4,$3F,$89,$36,$BF,$90,$BF,$E6,$D6,$C6,$8F,$E3,$1F,$B0
	.dc.b	$7F,$88,$A3,$27,$F2,$87,$15,$FC,$53,$F0,$CF,$E4,$B3,$FC,$B1,$FF
	.dc.b	$31,$B7,$CD,$FE,$45,$18,$9E,$1E,$1E,$0C,$CD,$AF,$6E,$1F,$30,$F9
	.dc.b	$BF,$5D,$63,$E0,$FC,$AF,$CC,$D7,$D5,$8F,$CA,$8C,$1F,$CF,$3D,$67
	.dc.b	$EC,$3F,$B0,$BF,$F0,$E7,$E8,$5F,$92,$E8,$FF,$1C,$CE,$70,$FF,$0B
	.dc.b	$5A,$B5,$7F,$E6,$2F,$A2,$7D,$7F,$F8,$D9,$FE,$83,$7F,$E4,$57,$33
	.dc.b	$FE,$29,$73,$7F,$76,$7F,$C7,$BC,$BF,$92,$78,$7B,$78,$7F,$A4,$7E
	.dc.b	$18,$FC,$0A,$E1,$86,$E3,$FD,$6B,$EE,$DF,$6F,$FD,$3D,$B1,$FD,$7A
	.dc.b	$D7,$9A,$FC,$BF,$A3,$2B,$F3,$9C,$7A,$7F,$1B,$1F,$A2,$E6,$FF,$30
	.dc.b	$47,$E6,$1B,$9F,$F9,$D1,$AF,$F9,$F1,$AF,$EB,$47,$D7,$3E,$83,$F5
	.dc.b	$AD,$AF,$D6,$57,$4E,$4F,$F2,$DC,$E6,$FF,$B0,$FE,$14,$FB,$1F,$ED
	.dc.b	$8B,$F0,$AB,$18,$66,$3F,$DF,$A7,$4D,$9C,$D7,$DE,$FD,$2F,$EA,$9F
	.dc.b	$DD,$27,$FD,$3E,$E3,$E1,$BF,$CE,$7F,$41,$CA,$BF,$0A,$DC,$FF,$D0
	.dc.b	$35,$C6,$7F,$D0,$D3,$9B,$F8,$5A,$7E,$D6,$FF,$CE,$ED,$8E,$FF,$51
	.dc.b	$F3,$FC,$4F,$A1,$EA,$FF,$08,$A3,$49,$5F,$EA,$4E,$A6,$6F,$B5,$FF
	.dc.b	$44,$B6,$7E,$6D,$62,$7F,$C2,$9F,$85,$FF,$AB,$A7,$FC,$46,$BE,$8F
	.dc.b	$3F,$3F,$A7,$FC,$35,$F3,$5E,$BE,$0F,$F1,$33,$7D,$15,$F1,$C3,$FC
	.dc.b	$5B,$39,$FC,$29,$F9,$6F,$F8,$8D,$7D,$05,$8C,$3F,$E2,$7F,$C2,$7F
	.dc.b	$C9,$D9,$7F,$88,$1B,$C9,$FE,$27,$BE,$38,$FF,$8E,$A7,$2F,$F8,$C6
	.dc.b	$3F,$C5,$15,$FE,$1C,$6B,$9F,$9C,$FF,$1D,$46,$23,$FC,$81,$5F,$67
	.dc.b	$43,$67,$FC,$5D,$7F,$E3,$D8,$C7,$C0,$7A,$CF,$67,$C0,$D9,$FF,$26
	.dc.b	$D7,$F0,$BC,$BF,$C7,$D7,$FC,$1B,$04,$E6,$BA,$78,$DF,$17,$6F,$29
	.dc.b	$23,$FD,$C3,$5C,$23,$FC,$A5,$7F,$C3,$B3,$F8,$DE,$47,$1B,$F8,$8F
	.dc.b	$CF,$F5,$BD,$E4,$8E,$79,$23,$9F,$B3,$F4,$4E,$13,$F3,$5F,$B1,$D3
	.dc.b	$23,$DB,$F2,$17,$1F,$AD,$D1,$F6,$4C,$6F,$F0,$1F,$C0,$57,$D0,$21
	.dc.b	$E2,$3D,$7D,$7F,$81,$36,$FD,$D7,$FD,$C7,$FA,$73,$E4,$3F,$DE,$7F
	.dc.b	$47,$FA,$4C,$7F,$DE,$9F,$A7,$7F,$AD,$E3,$1D,$EB,$E1,$8F,$F7,$F9
	.dc.b	$CD,$6F,$2F,$F0,$15,$A3,$18,$74,$3B,$CE,$B0,$E7,$0B,$E6,$E8,$E1
	.dc.b	$87,$7A,$EB,$7D,$30,$F2,$63,$B7,$C1,$77,$F5,$5F,$EC,$D8,$FA,$9B
	.dc.b	$87,$99,$FE,$DF,$3E,$18,$C6,$DC,$49,$97,$86,$58,$2F,$8E,$1D,$6F
	.dc.b	$75,$CC,$DC,$F8,$1B,$EB,$6A,$B8,$61,$D7,$99,$CD,$ED,$C3,$84,$63
	.dc.b	$0F,$77,$EB,$3F,$E9,$1C,$3E,$AB,$F9,$1C,$3C,$D9,$A3,$5A,$7D,$47
	.dc.b	$57,$E5,$9D,$71,$74,$E4,$EA,$DC,$39,$DF,$0C,$7B,$D6,$A7,$2F,$D7
	.dc.b	$C1,$F8,$77,$FA,$A7,$E5,$BE,$44,$63,$14,$62,$7E,$24,$E6,$31,$3A
	.dc.b	$8F,$94,$64,$B3,$C0,$EC,$E2,$38,$3F,$D9,$CE,$05,$8D,$B6,$36,$FF
	.dc.b	$6D,$8F,$DD,$65,$FE,$4C,$6C,$73,$3D,$B7,$F8,$27,$11,$A7,$B4,$67
	.dc.b	$1F,$D6,$47,$0A,$F8,$13,$9C,$FC,$1E,$7F,$F7,$DB,$DE,$1E,$06,$D6
	.dc.b	$F0,$FF,$37,$4E,$6F,$8C,$7E,$03,$DF,$69,$CE,$7F,$67,$18,$9D,$4E
	.dc.b	$5F,$CD,$FA,$AF,$99,$C6,$9F,$E6,$96,$DF,$1A,$FC,$43,$73,$FE,$0B
	.dc.b	$7C,$67,$F2,$2D,$77,$87,$23,$AC,$66,$7F,$86,$FC,$93,$93,$C9,$C5
	.dc.b	$FD,$8B,$E1,$2F,$8D,$FE,$94,$CB,$F3,$B7,$B6,$DF,$A2,$31,$1F,$9D
	.dc.b	$6E,$1D,$B9,$F5,$52,$B8,$9D,$23,$0F,$18,$B3,$C6,$3F,$5B,$E4,$FF
	.dc.b	$29,$6F,$4F,$C8,$8C,$47,$EB,$57,$C5,$E6,$64,$F0,$39,$C7,$13,$2F
	.dc.b	$18,$C3,$FC,$17,$C2,$7E,$01,$FE,$43,$F0,$3F,$B1,$B7,$87,$FB,$63
	.dc.b	$6B,$FC,$1B,$C7,$78,$7F,$86,$BF,$EA,$86,$C7,$F6,$4E,$DF,$B0,$FF
	.dc.b	$93,$E3,$2F,$EA,$DF,$EF,$7C,$BC,$F2,$FD,$14,$FF,$83,$63,$B5,$3E
	.dc.b	$3E,$92,$7F,$C3,$6E,$BD,$4F,$27,$88,$F5,$A9,$2B,$FC,$43,$F8,$37
	.dc.b	$43,$D1,$81,$A4,$8F,$EE,$7E,$FF,$42,$3F,$E2,$AF,$A8,$F2,$9C,$88
	.dc.b	$FC,$17,$AF,$E0,$BE,$4F,$83,$93,$FE,$32,$8F,$69,$1C,$1F,$E3,$6F
	.dc.b	$8D,$F5,$FB,$B2,$3F,$C3,$2F,$C4,$FB,$29,$1F,$3D,$F5,$91,$FE,$1D
	.dc.b	$E3,$BE,$4B,$FC,$7F,$1F,$EE,$DE,$8F,$DD,$4E,$7D,$5F,$BC,$D9,$D7
	.dc.b	$4F,$FB,$CE,$39,$7A,$BF,$53,$DF,$DE,$3F,$E4,$8B,$D9,$FC,$D1,$C2
	.dc.b	$C7,$F8,$47,$4F,$F5,$4A,$FC,$57,$86,$F6,$A9,$2F,$9D,$3F,$0C,$F4
	.dc.b	$C7,$3D,$0A,$75,$F2,$BF,$95,$AD,$E5,$E0,$6D,$F0,$AD,$DF,$A3,$B1
	.dc.b	$FA,$BB,$EF,$BF,$41,$18,$6C,$66,$7D,$E3,$F2,$97,$C7,$7E,$9E,$30
	.dc.b	$FF,$3F,$7C,$F1,$A7,$0D,$7F,$A2,$35,$FE,$A0,$FF,$AB,$5E,$ED,$7E
	.dc.b	$11,$D3,$E0,$FC,$BF,$71,$E8,$E2,$FF,$71,$CF,$B5,$AF,$FB,$A9,$BB
	.dc.b	$7F,$3B,$CF,$E0,$F8,$FE,$EF,$1F,$DE,$CB,$F8,$DA,$7F,$14,$BF,$B7
	.dc.b	$F1,$2E,$FF,$E0,$8D,$AF,$F7,$14,$63,$BA,$46,$08,$7A,$B8,$54,$9F
	.dc.b	$95,$E2,$3B,$29,$2F,$E1,$F4,$FC,$FE,$2F,$F6,$D4,$7C,$C8,$FA,$0F
	.dc.b	$D5,$6C,$6A,$3E,$7F,$C8,$F1,$23,$F7,$3E,$D5,$F4,$3F,$C7,$FE,$CF
	.dc.b	$C5,$3F,$94,$6B,$E9,$6E,$FF,$9E,$FC,$25,$E2,$FF,$27,$5F,$90,$58
	.dc.b	$97,$F9,$E9,$CB,$F9,$3F,$B1,$7C,$0F,$E4,$9B,$19,$8F,$CA,$FE,$B1
	.dc.b	$F4,$FB,$FF,$CC,$EF,$F5,$D9,$FC,$2F,$52,$35,$1F,$E5,$06,$C6,$7D
	.dc.b	$7F,$CF,$FC,$CF,$AA,$FF,$CE,$3F,$B1,$7E,$7F,$0E,$6D,$DB,$F1,$15
	.dc.b	$B9,$FC,$EF,$2F,$CE,$23,$F9,$E7,$E8,$F9,$7F,$5C,$91,$F9,$CF,$C3
	.dc.b	$3F,$A6,$7C,$D7,$F3,$98,$FE,$02,$D7,$F0,$CD,$C3,$F8,$5B,$F8,$F8
	.dc.b	$CF,$6F,$2F,$46,$47,$28,$FE,$B1,$F8,$53,$C4,$FE,$B5,$F9,$7F,$EC
	.dc.b	$71,$FE,$2A,$DD,$FF,$83,$56,$F0,$FE,$26,$D7,$F3,$0D,$AF,$E3,$AD
	.dc.b	$9E,$3F,$1F,$FC,$0F,$E7,$7F,$8F,$5F,$F0,$86,$BF,$BA,$0E,$2F,$FC
	.dc.b	$4D,$DC,$F4,$FF,$3C,$7E,$68,$95,$FC,$29,$FC,$EF,$5F,$E8,$C9,$1F
	.dc.b	$C5,$BF,$63,$FE,$84,$FF,$2C,$FD,$93,$FC,$59,$87,$F3,$4F,$CC,$7F
	.dc.b	$C6,$8D,$D3,$FB,$F4,$E9,$FE,$51,$F9,$8B,$9F,$FC,$00,$DC,$3F,$DC
	.dc.b	$6D,$7F,$BF,$35,$FF,$01,$7D,$EF,$F4,$4C,$3F,$A1,$B6,$FF,$9A,$36
	.dc.b	$5F,$56,$FA,$6F,$F3,$95,$FF,$4A,$FC,$C5,$F0,$DF,$E1,$36,$DF,$F4
	.dc.b	$86,$BF,$C9,$AD,$F5,$3C,$B2,$FE,$85,$C4,$F9,$9F,$F0,$D3,$6F,$FD
	.dc.b	$FA,$D7,$FC,$30,$D7,$FC,$39,$F8,$B3,$CD,$FF,$12,$3F,$E1,$78,$FE
	.dc.b	$B2,$D7,$F8,$F3,$B7,$FA,$EF,$E5,$0F,$9C,$FF,$16,$2E,$9A,$62,$3F
	.dc.b	$C6,$1B,$7F,$1C,$F4,$BD,$B7,$FE,$2B,$8C,$19,$F2,$1F,$45,$FE,$37
	.dc.b	$6C,$FF,$6C,$6B,$F1,$6F,$FC,$44,$FF,$70,$F3,$BE,$D7,$1F,$F7,$86
	.dc.b	$F0,$7F,$69,$6C,$FD,$6F,$41,$F4,$9F,$E4,$5F,$CC,$7F,$C9,$B3,$FE
	.dc.b	$9F,$D5,$18,$8F,$F2,$13,$5F,$F2,$33,$5F,$EE,$DF,$98,$FD,$AC,$BF
	.dc.b	$C1,$2B,$E5,$FF,$CA,$4D,$9F,$F0,$3B,$BC,$FF,$BF,$60,$F8,$EF,$F2
	.dc.b	$C3,$6B,$FC,$1A,$D7,$F0,$DF,$87,$79,$71,$FF,$0B,$FE,$63,$FC,$FA
	.dc.b	$7F,$CB,$4D,$7F,$C1,$4D,$7F,$C2,$4D,$A7,$E6,$3F,$BB,$C7,$FC,$20
	.dc.b	$F7,$BF,$F1,$1B,$63,$3F,$0B,$FE,$6D,$97,$D4,$5F,$6E,$65,$FE,$24
	.dc.b	$9C,$BF,$E3,$0F,$89,$7E,$63,$FE,$88,$8F,$F1,$29,$FE,$E6,$70,$FF
	.dc.b	$8E,$DA,$E7,$E0,$BF,$C2,$D2,$FC,$39,$F8,$A7,$F9,$11,$B5,$DD,$AF
	.dc.b	$EB,$7F,$10,$7D,$47,$F9,$2B,$DB,$7D,$F5,$F1,$65,$F9,$E7,$2F,$A7
	.dc.b	$F4,$7F,$E7,$4F,$A3,$34,$62,$3F,$CA,$4D,$7F,$E0,$0B,$61,$F9,$B7
	.dc.b	$FA,$E3,$7F,$F2,$EB,$78,$FF,$CB,$4D,$B5,$3F,$65,$FE,$3D,$18,$FD
	.dc.b	$0B,$F3,$6F,$CD,$6B,$FE,$30,$94,$63,$0F,$F3,$5F,$E7,$1F,$CB,$AB
	.dc.b	$4D,$7D,$8D,$7F,$CD,$8D,$7C,$6D,$AF,$F5,$07,$C3,$7F,$9D,$3E,$6B
	.dc.b	$FC,$17,$5F,$7D,$AF,$F1,$DF,$A6,$FE,$C5,$87,$FB,$D7,$CF,$46,$35
	.dc.b	$FF,$33,$C6,$3D,$EF,$F5,$E3,$6F,$F1,$5F,$89,$FF,$8C,$5F,$C6,$7C
	.dc.b	$5F,$F8,$89,$FC,$77,$E6,$BF,$E1,$67,$E1,$37,$2F,$F4,$5C,$EE,$7E
	.dc.b	$33,$63,$3E,$4F,$FA,$3A,$31,$97,$FA,$35,$FE,$69,$4F,$88,$7E,$CD
	.dc.b	$FC,$A7,$0F,$F2,$B7,$C5,$FF,$25,$AF,$F1,$37,$A3,$FE,$76,$AD,$63
	.dc.b	$FE,$84,$C3,$8F,$06,$49,$7F,$E7,$FA,$FE,$64,$FF,$89,$9A,$FE,$A2
	.dc.b	$7C,$3E,$A7,$F9,$E5,$B1,$FA,$C3,$B4,$F0,$7C,$1F,$B3,$FE,$0F,$0F
	.dc.b	$17,$1B,$F1,$2F,$F4,$5B,$6B,$FD,$21,$5B,$7E,$03,$70,$34,$9E,$5F
	.dc.b	$FA,$37,$E3,$DF,$8A,$7F,$AB,$6F,$C8,$23,$3C,$1F,$E9,$16,$C6,$9A
	.dc.b	$FA,$DF,$EA,$5F,$8A,$6F,$D4,$34,$93,$FE,$6C,$FA,$6F,$F3,$6A,$69
	.dc.b	$1F,$8D,$9B,$E7,$9F,$81,$B8,$3F,$D6,$11,$86,$F3,$E6,$FF,$CB,$33
	.dc.b	$96,$D6,$74,$FF,$4C,$4B,$E0,$71,$87,$E0,$63,$84,$EA,$73,$7F,$8D
	.dc.b	$FC,$53,$F4,$B7,$F2,$32,$46,$27,$D2,$D7,$F0,$8D,$7F,$1C,$BE,$93
	.dc.b	$AC,$F2,$6C,$73,$C3,$FD,$5C,$EE,$FF,$DA,$0D,$BF,$F3,$BB,$F8,$78
	.dc.b	$CD,$72,$E4,$F4,$F8,$1F,$E4,$6F,$BA,$BF,$18,$6F,$26,$19,$87,$F4
	.dc.b	$59,$7D,$BE,$33,$FE,$35,$18,$9F,$E9,$5F,$89,$BE,$AB,$E3,$F8,$6F
	.dc.b	$C5,$3F,$DD,$F5,$A9,$CC,$63,$0F,$F1,$73,$78,$3F,$89,$36,$BF,$DE
	.dc.b	$FA,$B1,$ED,$1F,$EF,$AA,$DD,$FB,$BF,$13,$FE,$A3,$3F,$9C,$B4,$62
	.dc.b	$7F,$65,$18,$F4,$FF,$11,$1F,$D2,$67,$4D,$F0,$7A,$1D,$66,$DF,$F6
	.dc.b	$9F,$C4,$BF,$56,$DE,$1F,$9F,$1F,$E3,$58,$C1,$BE,$1E,$F8,$7B,$27
	.dc.b	$E9,$4F,$9D,$B1,$E6,$6B,$FD,$77,$F1,$2C,$E9,$F9,$FD,$1F,$89,$3E
	.dc.b	$93,$EF,$D6,$B1,$F9,$33,$97,$E4,$1F,$CF,$3F,$12,$7B,$BF,$61,$B1
	.dc.b	$FE,$05,$E3,$7D,$36,$6B,$F4,$7F,$00,$D7,$FC,$D7,$7F,$E3,$76,$9F
	.dc.b	$C7,$E3,$5F,$15,$F0,$21,$18,$9F,$F1,$F7,$6F,$CA,$9F,$F7,$E3,$F8
	.dc.b	$D6,$F1,$78,$23,$1C,$1F,$BF,$B7,$E2,$4F,$AB,$F2,$BF,$E0,$AC,$7D
	.dc.b	$F8,$F8,$23,$51,$8A,$E3,$ED,$FB,$38,$DF,$73,$E7,$9F,$EF,$ED,$8C
	.dc.b	$D7,$F9,$55,$B8,$7F,$3D,$EC,$7C,$33,$E0,$7F,$46,$DB,$F0,$7E,$24
	.dc.b	$FA,$FC,$D7,$CC,$FB,$5F,$E2,$95,$E5,$5E,$56,$7C,$8D,$3F,$4B,$F1
	.dc.b	$2F,$AB,$5F,$E6,$76,$E9,$F4,$5A,$FF,$8F,$37,$46,$23,$FB,$EF,$E2
	.dc.b	$2F,$B4,$FD,$CB,$6F,$E3,$7E,$25,$6B,$FD,$49,$37,$9B,$FC,$11,$F8
	.dc.b	$8B,$E9,$BF,$80,$B7,$0F,$9D,$F8,$97,$BD,$FA,$7C,$89,$D6,$2C,$3E
	.dc.b	$E7,$67,$EC,$A3,$F1,$DB,$7F,$AE,$3C,$6F,$A9,$39,$9F,$C5,$B6,$7F
	.dc.b	$A3,$F4,$FE,$5B,$D1,$7E,$0D,$63,$0F,$E6,$BE,$FF,$EA,$61,$E1,$D7
	.dc.b	$BC,$7E,$9E,$30,$7E,$61,$DE,$7F,$91,$74,$FE,$4F,$F1,$27,$DA,$FE
	.dc.b	$25,$B7,$FE,$9B,$6D,$FE,$95,$AF,$F9,$C3,$7F,$F6,$7F,$99,$FC,$AB
	.dc.b	$E7,$CF,$EC,$3F,$12,$FC,$1E,$1F,$2A,$72,$FC,$CF,$A2,$FE,$4B,$87
	.dc.b	$FB,$65,$B9,$FE,$9F,$F1,$27,$D7,$FF,$36,$9C,$DF,$CE,$FC,$23,$FD
	.dc.b	$40,$FF,$AC,$5B,$87,$EE,$DA,$FE,$2D,$AF,$D6,$F1,$7F,$CC,$B3,$FA
	.dc.b	$DF,$C4,$9F,$25,$FC,$22,$73,$5F,$33,$99,$5A,$BF,$DB,$CB,$F0,$47
	.dc.b	$D0,$FF,$72,$8C,$37,$E2,$DF,$69,$9A,$DD,$F2,$FC,$45,$F5,$7F,$C4
	.dc.b	$71,$46,$27,$FC,$43,$18,$9D,$4E,$63,$FC,$4A,$BF,$05,$F1,$36,$BB
	.dc.b	$56,$27,$F9,$2F,$DE,$BE,$4E,$98,$F1,$2D,$FF,$11,$9D,$37,$EE,$F3
	.dc.b	$97,$C5,$94,$7F,$30,$FC,$4B,$FD,$B5,$5F,$C6,$57,$9B,$9F,$A6,$D3
	.dc.b	$27,$5E,$2D,$C7,$E8,$5F,$F3,$6E,$BF,$EC,$3B,$BF,$02,$F6,$E7,$EE
	.dc.b	$6E,$3F,$2E,$2E,$1F,$A4,$3E,$5C,$E2,$F9,$ED,$F1,$8C,$5F,$FB,$33
	.dc.b	$0F,$7D,$FE,$B3,$D7,$7D,$5B,$AF,$FC,$46,$DB,$FE,$73,$3A,$F1,$73
	.dc.b	$5A,$FC,$33,$7F,$11,$6F,$78,$7F,$46,$7F,$84,$FE,$21,$FB,$66,$FD
	.dc.b	$67,$F5,$A8,$C6,$7F,$D5,$14,$66,$A5,$7F,$47,$9C,$B6,$3F,$29,$BF
	.dc.b	$AA,$3F,$32,$DB,$FC,$F7,$E0,$1F,$8B,$E0,$FF,$07,$CE,$5F,$A3,$CF
	.dc.b	$F9,$02,$5F,$E5,$38,$3F,$92,$E1,$FE,$89,$E4,$FF,$4F,$4F,$E9,$9B
	.dc.b	$1F,$4E,$7F,$95,$DF,$F8,$72,$72,$FF,$67,$FC,$41,$F5,$3F,$DE,$9F
	.dc.b	$E8,$F1,$98,$C7,$C1,$4C,$7F,$B3,$53,$EA,$B1,$97,$EE,$63,$13,$F5
	.dc.b	$BF,$0F,$71,$FF,$97,$1B,$19,$E4,$C4,$E7,$93,$5F,$17,$6B,$93,$C7
	.dc.b	$E1,$64,$7F,$77,$FC,$15,$F5,$38,$D3,$F2,$5C,$2F,$B1,$FD,$F3,$6F
	.dc.b	$F7,$AD,$9F,$F4,$FB,$5F,$99,$D5,$FC,$36,$0D,$AC,$3F,$E2,$0F,$C1
	.dc.b	$9F,$83,$53,$1F,$80,$5C,$7F,$C1,$93,$FE,$41,$C7,$F8,$F2,$FA,$3E
	.dc.b	$5F,$3B,$CD,$FF,$15,$FE,$0F,$F8,$5B,$FE,$BE,$DB,$FF,$13,$BF,$C3
	.dc.b	$DF,$E5,$0E,$7C,$F3,$FD,$B6,$33,$F3,$9D,$3E,$27,$AB,$DB,$E3,$BF
	.dc.b	$C3,$95,$FE,$22,$AF,$F1,$B3,$5F,$F0,$33,$5F,$1F,$D7,$79,$8E,$58
	.dc.b	$FF,$8A,$65,$1E,$AF,$7E,$4E,$7E,$02,$DF,$A2,$7D,$FC,$98,$F0,$3F
	.dc.b	$C7,$58,$FD,$CC,$3A,$7C,$DF,$98,$8F,$FB,$5B,$6F,$7C,$7A,$BE,$73
	.dc.b	$E4,$9F,$29,$B3,$E5,$FA,$9F,$79,$1E,$57,$FE,$B3,$CF,$E6,$FB,$3F
	.dc.b	$48,$FA,$75,$6B,$DD,$E8,$FD,$6E,$AF,$AE,$D8,$F8,$51,$6F,$EA,$CD
	.dc.b	$D3,$DB,$E8,$FD,$02,$5F,$89,$FA,$6F,$E0,$8F,$C1,$B5,$FE,$43,$F8
	.dc.b	$7F,$F8,$5B,$E8,$3F,$2B,$83,$D0,$4E,$7B,$7F,$CE,$91,$8F,$73,$FD
	.dc.b	$C9,$7C,$23,$7F,$21,$8A,$DD,$F1,$C3,$A7,$E1,$FF,$EB,$5C,$38,$4E
	.dc.b	$70,$4E,$6F,$8E,$1C,$F0,$E2,$5A,$EF,$97,$FA,$1F,$1F,$16,$1C,$6B
	.dc.b	$D5,$F8,$7F,$F5,$AA,$D6,$3E,$A8,$FF,$45,$DF,$A5,$D5,$23,$1B,$F7
	.dc.b	$9D,$D7,$18,$C6,$B9,$66,$1D,$B0,$F0,$C6,$A3,$19,$30,$E7,$5A,$9C
	.dc.b	$C6,$27,$57,$D2,$BC,$98,$7D,$2F,$C3,$FF,$BB,$C6,$0F,$B0,$FF,$A2
	.dc.b	$32,$FA,$8E,$9B,$C5,$F2,$63,$15,$CE,$31,$C3,$96,$E7,$E8,$AE,$95
	.dc.b	$E1,$AD,$4F,$BF,$07,$E1,$E9,$7C,$BE,$63,$0D,$DB,$EE,$57,$0A,$DB
	.dc.b	$EB,$31,$3C,$63,$39,$FE,$5A,$7B,$D7,$2B,$EB,$1C,$5A,$FF,$B2,$CF
	.dc.b	$F6,$63,$87,$0D,$73,$5E,$27,$81,$EC,$CF,$F4,$8E,$0C,$6B,$FE,$3E
	.dc.b	$AF,$F5,$EB,$67,$73,$CA,$72,$F9,$63,$9E,$1E,$B9,$CB,$79,$3D,$0E
	.dc.b	$67,$7A,$7A,$1B,$7E,$C6,$BF,$4A,$B8,$CF,$EE,$3D,$C7,$CD,$62,$BF
	.dc.b	$7D,$EF,$B7,$C6,$9F,$E6,$56,$DF,$0A,$FC,$CB,$5F,$C8,$4E,$DF,$CE
	.dc.b	$D7,$37,$77,$C8,$EB,$19,$8E,$2D,$CF,$F8,$55,$F5,$7F,$86,$7E,$1F
	.dc.b	$FE,$52,$AF,$D4,$B5,$F0,$B6,$37,$BF,$AA,$31,$9F,$F1,$86,$D7,$F1
	.dc.b	$D6,$E4,$55,$C7,$8A,$57,$81,$E9,$12,$30,$43,$C6,$FC,$AF,$A8,$FF
	.dc.b	$5B,$C7,$F0,$07,$FD,$F0,$DE,$6D,$7C,$74,$8C,$10,$F9,$B3,$C9,$B1
	.dc.b	$B6,$9D,$7A,$7E,$E7,$73,$BB,$EB,$76,$7D,$38,$C7,$17,$E8,$E3,$13
	.dc.b	$AE,$3F,$F7,$1A,$F9,$0D,$46,$1F,$F4,$DD,$F5,$6C,$EF,$1F,$ED,$4D
	.dc.b	$7F,$53,$3E,$3F,$8D,$E7,$08,$C2,$3F,$D7,$13,$FE,$D4,$71,$87,$9B
	.dc.b	$EB,$38,$D7,$B5,$B5,$BA,$E3,$78,$9C,$FC,$B4,$9E,$79,$7B,$A3,$8F
	.dc.b	$14,$9F,$4B,$EC,$23,$EC,$FA,$BE,$E4,$79,$9E,$BC,$5F,$82,$79,$0F
	.dc.b	$8A,$48,$F8,$AF,$C8,$FC,$27,$D3,$F8,$34,$9E,$7A,$FD,$99,$F6,$92
	.dc.b	$33,$68,$F8,$3F,$21,$F0,$BE,$32,$56,$9F,$8B,$F1,$D2,$3E,$AB,$F6
	.dc.b	$08,$FD,$8F,$81,$FA,$BE,$23,$E9,$5E,$8E,$1B,$5E,$2A,$C7,$A1,$23
	.dc.b	$E1,$DF,$18,$E1,$E8,$FD,$8F,$64,$DF,$4E,$5F,$F9,$C6,$35,$E9,$C8
	.dc.b	$FF,$3A,$3C,$6F,$F1,$30,$FC,$05,$F1,$D8,$6C,$7F,$BF,$9A,$FF,$B2
	.dc.b	$D7,$C2,$6C,$FF,$7A,$C7,$36,$C6,$A3,$FD,$F2,$D7,$FA,$1F,$8C,$FC
	.dc.b	$35,$ED,$FE,$72,$30,$D8,$CC,$FF,$A3,$7C,$EC,$78,$F9,$F1,$A4,$FF
	.dc.b	$B4,$AB,$F8,$2B,$FE,$D9,$5E,$37,$32,$47,$FB,$63,$F6,$8F,$EA,$BE
	.dc.b	$2F,$EF,$7E,$2B,$FD,$F7,$3F,$8A,$5F,$1D,$FA,$86,$C6,$7E,$3F,$1F
	.dc.b	$19,$96,$47,$18,$F9,$9D,$32,$3E,$63,$F3,$FB,$24,$7C,$67,$E4,$91
	.dc.b	$FE,$CC,$3E,$49,$F4,$1F,$69,$B1,$FA,$A3,$51,$E7,$E8,$FD,$FF,$E1
	.dc.b	$CF,$1E,$92,$BF,$4F,$1F,$B0,$6F,$8D,$C3,$C7,$91,$83,$FD,$FF,$F8
	.dc.b	$77,$CC,$6F,$37,$D9,$8F,$A3,$C5,$FD,$67,$A5,$1B,$A4,$E7,$ED,$3E
	.dc.b	$8F,$99,$FE,$64,$6C,$FF,$9A,$5B,$67,$FC,$19,$5F,$D8,$7B,$1E,$37
	.dc.b	$E6,$49,$7F,$99,$F2,$3F,$38,$47,$07,$E7,$7C,$8F,$D0,$19,$7F,$3D
	.dc.b	$18,$7F,$85,$FE,$1B,$F8,$07,$D3,$9F,$F2,$FF,$2D,$2F,$92,$FE,$5A
	.dc.b	$DA,$E3,$D9,$8E,$BC,$B2,$3A,$C7,$F3,$4F,$C3,$3F,$DA,$51,$FC,$69
	.dc.b	$FD,$0B,$E0,$3F,$A0,$38,$DB,$FA,$A3,$C2,$35,$8A,$31,$3A,$F8,$6E
	.dc.b	$C7,$7F,$89,$FE,$01,$CD,$30,$FE,$00,$FF,$03,$E6,$91,$FB,$A7,$F7
	.dc.b	$A4,$7F,$2B,$F8,$67,$F1,$F7,$F2,$4D,$7D,$01,$C6,$1F,$E1,$9F,$B3
	.dc.b	$FF,$7B,$CF,$F3,$8F,$C2,$FF,$B0,$4F,$F6,$F7,$FB,$64,$7F,$5F,$6E
	.dc.b	$87,$7C,$3F,$4F,$F0,$BF,$E5,$B5,$FD,$A1,$AF,$F9,$66,$91,$88,$E1
	.dc.b	$E9,$7E,$BB,$E8,$9F,$3F,$F9,$26,$CF,$F7,$06,$BF,$D3,$BE,$A9,$E9
	.dc.b	$67,$E8,$B8,$12,$7F,$B7,$FE,$17,$FE,$2F,$DB,$FC,$15,$DA,$F1,$3F
	.dc.b	$C0,$6D,$AF,$F7,$53,$5F,$F0,$0B,$5F,$57,$E1,$6F,$6F,$FB,$DA,$F7
	.dc.b	$29,$88,$FE,$FF,$F8,$5B,$EC,$6F,$CA,$4C,$3F,$B3,$CE,$5F,$ED,$3F
	.dc.b	$3F,$FB,$77,$E4,$35,$FD,$03,$5F,$F0,$93,$78,$55,$E5,$9F,$EF,$B1
	.dc.b	$8C,$3E,$54,$E6,$7F,$2F,$F2,$EF,$A2,$F9,$DE,$85,$9E,$2D,$7F,$45
	.dc.b	$82,$31,$C1,$FE,$35,$F3,$FF,$02,$FF,$8D,$FF,$0B,$FF,$44,$DF,$E0
	.dc.b	$5A,$FE,$E5,$AF,$F8,$AD,$BD,$4F,$C2,$FE,$C7,$67,$EF,$DF,$F2,$ED
	.dc.b	$7F,$83,$9A,$FE,$C1,$DA,$C9,$F6,$4F,$C4,$3F,$14,$D9,$F8,$AD,$7F
	.dc.b	$73,$83,$E7,$7F,$C6,$FF,$86,$7B,$E7,$FD,$51,$28,$C4,$7F,$91,$9A
	.dc.b	$FE,$7D,$AF,$F1,$7F,$C3,$3F,$84,$67,$6F,$FC,$87,$1D,$E7,$FC,$52
	.dc.b	$E9,$FF,$27,$6B,$FD,$E4,$FF,$31,$0F,$A6,$FF,$25,$B6,$3F,$C6,$2E
	.dc.b	$EF,$FC,$A9,$F0,$EF,$C3,$3F,$CE,$55,$FE,$CB,$8C,$7C,$0F,$F5,$E6
	.dc.b	$C7,$FB,$15,$AF,$FB,$2B,$77,$E1,$6F,$A0,$FD,$A3,$67,$F1,$ED,$7F
	.dc.b	$94,$53,$F0,$CF,$F2,$F6,$7F,$EA,$D8,$7E,$19,$FC,$F6,$31,$F0,$9F
	.dc.b	$B1,$6F,$87,$FE,$3C,$9D,$6D,$EC,$FC,$33,$FD,$55,$97,$AD,$AE,$7E
	.dc.b	$5B,$D9,$18,$8F,$F7,$DF,$E1,$9F,$C7,$33,$34,$C4,$FF,$AF,$72,$F3
	.dc.b	$7E,$17,$FC,$A3,$1F,$96,$DC,$18,$23,$0D,$FA,$EF,$F2,$9C,$1F,$9C
	.dc.b	$D9,$F9,$CD,$7F,$DE,$F3,$C5,$FF,$7B,$70,$BE,$7B,$E0,$C7,$E7,$5B
	.dc.b	$19,$F6,$BF,$CC,$3F,$85,$FF,$8A,$AB,$E8,$B7,$3F,$F6,$E5,$6E,$7F
	.dc.b	$82,$DF,$4C,$17,$BC,$FF,$13,$F0,$FF,$E0,$49,$FF,$0E,$DF,$F9,$9A
	.dc.b	$3F,$C4,$2F,$DE,$74,$47,$FC,$E6,$E2,$3F,$C9,$DF,$05,$FC,$9E,$7F
	.dc.b	$D2,$3F,$07,$FE,$0D,$F9,$7C,$9A,$F3,$64,$EA,$B8,$3F,$DC,$EB,$FC
	.dc.b	$29,$3D,$3E,$62,$4F,$FA,$CC,$93,$FE,$19,$8F,$F1,$DB,$63,$47,$F8
	.dc.b	$86,$33,$F8,$4B,$E6,$FF,$CC,$8F,$F1,$ED,$3F,$D7,$D6,$8C,$5A,$31
	.dc.b	$5C,$09,$1E,$E9,$E8,$DE,$3F,$F4,$1F,$E1,$3F,$D1,$78,$BF,$C9,$ED
	.dc.b	$9F,$28,$70,$FA,$1A,$FB,$DF,$E8,$3F,$84,$F2,$F3,$8D,$24,$FF,$A2
	.dc.b	$3E,$83,$FD,$12,$9A,$47,$E2,$FE,$44,$FA,$87,$C2,$B7,$FF,$64,$CE
	.dc.b	$63,$15,$FE,$27,$8E,$18,$E3,$3C,$C6,$EF,$8E,$1D,$1E,$7E,$8F,$F0
	.dc.b	$D7,$38,$F1,$7F,$DD,$D5,$FD,$5F,$C4,$C6,$5F,$47,$DE,$7D,$DF,$8A
	.dc.b	$3F,$03,$18,$BF,$97,$C7,$FE,$C4,$AD,$57,$DC,$AD,$3F,$EA,$65,$EF
	.dc.b	$76,$AF,$F6,$5B,$63,$7A,$FF,$9B,$DD,$E1,$F9,$C6,$DF,$E5,$FE,$47
	.dc.b	$FC,$21,$39,$BE,$91,$89,$FE,$4F,$F0,$7F,$DE,$AB,$FD,$A6,$D8,$CE
	.dc.b	$BF,$03,$AD,$39,$CB,$F9,$63,$6F,$FA,$0F,$95,$FE,$D8,$79,$F1,$71
	.dc.b	$8C,$63,$FD,$F3,$F0,$9F,$F4,$ED,$7F,$82,$1B,$1F,$CE,$BF,$42,$FC
	.dc.b	$04,$E7,$D7,$F1,$7E,$2F,$FE,$32,$8F,$F7,$4B,$6B,$A4,$EB,$E0,$FE
	.dc.b	$BF,$2B,$F2,$4F,$C0,$57,$E3,$9A,$FE,$E6,$35,$C1,$F5,$A1,$18,$C7
	.dc.b	$F4,$AF,$3F,$8B,$EB,$66,$DE,$1F,$91,$F8,$5B,$C1,$A7,$CB,$FC,$2F
	.dc.b	$C7,$C7,$FD,$56,$75,$93,$3F,$EB,$70,$F1,$7C,$9C,$A8,$C5,$7F,$69
	.dc.b	$6D,$7F,$3C,$6B,$FD,$C7,$F1,$87,$C7,$7D,$BF,$C2,$DF,$31,$F7,$67
	.dc.b	$38,$78,$73,$FE,$D1,$5A,$8C,$1F,$DF,$BE,$A7,$FC,$08,$E3,$1F,$EA
	.dc.b	$B1,$87,$F6,$94,$F8,$4F,$C5,$4E,$63,$FD,$90,$7E,$62,$7F,$94,$B5
	.dc.b	$FE,$42,$D7,$F9,$03,$7D,$FF,$E4,$5D,$7F,$B4,$7F,$92,$71,$DF,$03
	.dc.b	$F2,$96,$9C,$E5,$E7,$5F,$85,$7E,$27,$1F,$E1,$71,$8F,$07,$A4,$E6
	.dc.b	$BF,$D8,$4A,$3C,$16,$F5,$BF,$11,$7F,$65,$B1,$9B,$FF,$1D,$B4,$FE
	.dc.b	$87,$2F,$C2,$BA,$4F,$95,$B3,$FE,$62,$FD,$1B,$EA,$BF,$EC,$08,$CE
	.dc.b	$5E,$8C,$1F,$4D,$F0,$D7,$D3,$58,$C3,$FC,$37,$EB,$9F,$77,$F9,$FD
	.dc.b	$D1,$8D,$3F,$8C,$B7,$3F,$EB,$33,$AA,$F7,$FE,$15,$FE,$93,$BF,$98
	.dc.b	$DA,$FF,$3C,$FE,$15,$F9,$C7,$FD,$AF,$19,$D3,$E7,$FE,$15,$EA,$BF
	.dc.b	$F7,$73,$6B,$FC,$D9,$FB,$4F,$E2,$67,$EC,$4E,$78,$5F,$86,$7C,$09
	.dc.b	$FF,$33,$B6,$FE,$6B,$5F,$F6,$23,$5F,$F7,$E7,$9C,$FA,$EF,$C5,$3F
	.dc.b	$C5,$B2,$FC,$1B,$9B,$FC,$43,$97,$F6,$D8,$7C,$8F,$85,$FC,$97,$EC
	.dc.b	$DF,$B9,$9F,$C7,$70,$A7,$39,$7F,$3D,$FB,$17,$DA,$FD,$16,$DF,$F4
	.dc.b	$86,$CF,$D0,$6B,$FE,$DD,$FC,$2F,$FA,$1C,$F8,$9B,$1F,$E9,$3F,$89
	.dc.b	$FE,$52,$B5,$FC,$EC,$EB,$47,$E1,$7F,$8C,$9F,$D0,$B7,$0F,$E2,$5F
	.dc.b	$0A,$FD,$AB,$C7,$5A,$F1,$FF,$61,$F8,$6B,$ED,$FF,$42,$DC,$3F,$D3
	.dc.b	$2D,$D3,$F3,$ED,$7F,$DE,$DF,$6B,$FD,$03,$0F,$D8,$BE,$53,$15,$F3
	.dc.b	$9F,$6C,$BE,$71,$F7,$1F,$DD,$23,$0D,$FB,$BF,$E1,$D7,$FE,$35,$AD
	.dc.b	$BF,$C5,$EB,$97,$E0,$5F,$CF,$DB,$1F,$E2,$26,$9F,$E2,$49,$4E,$F3
	.dc.b	$FF,$33,$57,$36,$D7,$CF,$8F,$04,$F9,$67,$39,$7F,$0E,$FD,$7F,$FC
	.dc.b	$75,$5F,$C4,$DB,$7C,$4F,$E3,$0F,$FB,$2E,$35,$E0,$FF,$1F,$8D,$D7
	.dc.b	$F1,$8F,$AE,$FE,$18,$F5,$9F,$E4,$91,$8A,$FF,$25,$FE,$D6,$F3,$7F
	.dc.b	$C1,$DF,$5B,$F8,$6C,$3F,$C5,$78,$2B,$57,$FE,$68,$AD,$3F,$E5,$46
	.dc.b	$E5,$FE,$53,$6D,$71,$C3,$AD,$7D,$5D,$7F,$55,$7E,$D3,$ED,$77,$E7
	.dc.b	$A7,$4E,$FE,$F3,$EA,$57,$F8,$22,$BF,$CD,$2D,$7E,$E4,$7D,$3E,$2F
	.dc.b	$F1,$47,$EB,$7F,$C2,$67,$F2,$91,$8F,$37,$FC,$59,$FB,$67,$C3,$9F
	.dc.b	$F3,$53,$F9,$CC,$FF,$CD,$EE,$75,$39,$9F,$F3,$87,$C2,$BE,$CB,$FC
	.dc.b	$56,$74,$F0,$6E,$1F,$D3,$7F,$0E,$7C,$37,$F5,$57,$FB,$56,$1F,$77
	.dc.b	$C1,$46,$AB,$87,$B3,$FD,$6F,$EF,$3F,$2F,$7B,$6F,$47,$F5,$6B,$E6
	.dc.b	$FF,$58,$8C,$DF,$E7,$FF,$10,$FA,$07,$E8,$FE,$A6,$C7,$E8,$79,$5F
	.dc.b	$81,$7C,$E2,$47,$E8,$63,$2F,$E8,$FE,$2F,$F9,$96,$7F,$65,$B9,$7F
	.dc.b	$98,$6B,$73,$9F,$73,$FB,$A7,$E2,$1F,$3F,$73,$F2,$DB,$EC,$6B,$8A
	.dc.b	$67,$FD,$D6,$B4,$71,$3E,$97,$37,$FB,$77,$F0,$DF,$F3,$37,$CA,$FF
	.dc.b	$55,$D7,$F3,$2A,$33,$EC,$7F,$9A,$F8,$3F,$9A,$6D,$FE,$BB,$6D,$F0
	.dc.b	$9D,$4F,$E5,$FE,$E6,$67,$FC,$1F,$F0,$0F,$F7,$63,$F2,$63,$F8,$86
	.dc.b	$BF,$B8,$AF,$F5,$EB,$5F,$F6,$23,$5F,$F1,$E3,$57,$CD,$3F,$00,$FE
	.dc.b	$34,$DB,$FE,$2D,$5F,$BE,$7F,$D9,$D1,$9A,$FF,$6A,$7E,$1F,$F0,$67
	.dc.b	$FD,$8E,$BF,$0E,$FF,$45,$39,$9F,$F6,$43,$99,$FF,$71,$FE,$3D,$FC
	.dc.b	$9A,$7F,$D9,$8D,$E1,$56,$E7,$FD,$C5,$3B,$E4,$7E,$1F,$FE,$2E,$6F
	.dc.b	$C3,$7F,$8C,$E3,$07,$F9,$7B,$1F,$EC,$FE,$27,$13,$94,$67,$0F,$F1
	.dc.b	$B4,$78,$E5,$F9,$9F,$F7,$79,$FF,$61,$36,$33,$F0,$9F,$E7,$76,$9F
	.dc.b	$ED,$AC,$7F,$A2,$FC,$27,$F9,$17,$0F,$F1,$1C,$FF,$95,$CD,$E1,$FB
	.dc.b	$38,$CB,$FB,$5C,$7E,$45,$FB,$B9,$3D,$9C,$FE,$D6,$61,$FE,$4B,$C3
	.dc.b	$FC,$21,$CD,$FE,$4A,$C3,$8C,$62,$7F,$C0,$9F,$21,$FE,$25,$CB,$FC
	.dc.b	$FC,$7F,$B7,$5C,$3A,$E3,$F6,$46,$23,$E2,$FC,$C7,$AC,$FF,$69,$D7
	.dc.b	$FA,$46,$B0,$FB,$77,$FC,$35,$76,$8F,$97,$CD,$F1,$E7,$33,$FE,$17
	.dc.b	$6B,$F4,$BE,$07,$E4,$1D,$A3,$F2,$1A,$7F,$99,$B3,$FF,$07,$FC,$9B
	.dc.b	$F3,$07,$E2,$3F,$EB,$06,$E7,$FE,$6F,$DF,$F8,$3B,$FE,$EC,$8C,$E7
	.dc.b	$F4,$7E,$5F,$EB,$1C,$6E,$D7,$AF,$F0,$3F,$41,$1F,$93,$5E,$98,$FF
	.dc.b	$38,$E4,$F9,$3F,$C0,$1E,$98,$B5,$FE,$97,$EB,$7E,$C1,$FE,$B9,$8F
	.dc.b	$F5,$87,$D0,$7F,$98,$B1,$56,$BE,$CD,$5A,$F8,$9F,$F5,$B4,$F2,$E7
	.dc.b	$CD,$B0,$E5,$3C,$6F,$84,$F4,$C5,$8F,$09,$FC,$1D,$6E,$B9,$9B,$9D
	.dc.b	$3E,$0A,$E3,$5F,$C7,$23,$4F,$6F,$85,$7D,$D7,$F9,$CF,$1F,$1D,$F0
	.dc.b	$BF,$36,$7F,$E3,$EA,$D3,$EC,$FC,$47,$FA,$BD,$FA,$BE,$8F,$F9,$4B
	.dc.b	$EC,$8C,$CF,$C2,$9C,$B6,$FE,$1D,$71,$6C,$75,$F4,$EC,$7C,$EA,$E2
	.dc.b	$D3,$3A,$A3,$2B,$0F,$0C,$EF,$4F,$9B,$3B,$8E,$BA,$3F,$4D,$7A,$FD
	.dc.b	$31,$EF,$39,$8E,$FF,$89,$7F,$AD,$75,$FB,$2F,$F9,$C9,$BE,$7E,$BC
	.dc.b	$A9,$A7,$82,$B5,$BF,$E1,$1E,$4D,$9D,$E9,$F4,$0F,$E3,$E6,$30,$FC
	.dc.b	$87,$89,$FE,$5C,$8F,$8A,$3F,$10,$EA,$8C,$7E,$23,$F9,$F7,$A1,$BB
	.dc.b	$E3,$29,$D5,$FE,$76,$B8,$F4,$7E,$51,$A7,$2A,$EE,$E2,$78,$47,$39
	.dc.b	$FF,$33,$36,$7F,$66,$72,$3F,$6A,$E5,$BF,$13,$FA,$C7,$70,$8C,$46
	.dc.b	$B8,$BA,$C6,$23,$F9,$3C,$EE,$7D,$0D,$8F,$44,$F5,$BF,$3F,$D7,$FF
	.dc.b	$2E,$6C,$F0,$9D,$69,$D1,$B1,$F3,$6D,$7C,$6B,$F7,$FF,$88,$FF,$BF
	.dc.b	$EB,$F7,$7A,$FF,$62,$F9,$2D,$42,$30,$FE,$52,$78,$3F,$81,$CD,$18
	.dc.b	$CF,$F8,$3B,$70,$EF,$C3,$D2,$B7,$BD,$AF,$F6,$4E,$1D,$6B,$06,$67
	.dc.b	$87,$D6,$BA,$3C,$BC,$BF,$83,$F4,$AF,$19,$DC,$CC,$7F,$57,$7C,$5B
	.dc.b	$1D,$3D,$ED,$47,$AE,$74,$D9,$D3,$CA,$77,$B7,$F2,$4A,$DB,$63,$DD
	.dc.b	$DE,$F8,$3F,$3F,$6F,$E7,$70,$FD,$46,$9F,$C7,$39,$F7,$52,$7F,$91
	.dc.b	$70,$B0,$E5,$18,$6B,$FC,$B2,$35,$3F,$D9,$63,$F9,$DC,$BE,$13,$A4
	.dc.b	$7D,$78,$E3,$9F,$F3,$C7,$51,$F8,$2F,$1B,$E0,$F9,$1F,$E0,$1D,$FF
	.dc.b	$BE,$B6,$7F,$C2,$1F,$99,$BE,$13,$3D,$DF,$BB,$11,$99,$FD,$5C,$62
	.dc.b	$7F,$DA,$8D,$8C,$CE,$DF,$EC,$7D,$3F,$E3,$4E,$2E,$B1,$F9,$3C,$BF
	.dc.b	$D2,$6F,$F7,$38,$CF,$94,$FC,$0B,$FD,$07,$3F,$C3,$73,$FE,$B7,$39
	.dc.b	$F6,$B7,$5A,$AF,$F6,$B4,$E5,$E1,$9F,$F1,$5C,$9F,$2B,$F8,$D9,$FF
	.dc.b	$01,$7C,$77,$92,$7F,$C9,$2D,$D0,$FA,$2F,$E1,$F1,$5D,$BE,$7E,$1F
	.dc.b	$E0,$09,$CB,$F0,$A7,$E2,$CE,$A3,$FC,$5B,$87,$F9,$5B,$DD,$EC,$48
	.dc.b	$FF,$21,$7E,$68,$F8,$CF,$03,$63,$E8,$B5,$F0,$35,$FF,$16,$35,$F8
	.dc.b	$1A,$3D,$FF,$67,$D6,$C9,$F8,$A7,$F0,$59,$D3,$8C,$7E,$3D,$7C,$59
	.dc.b	$CC,$FF,$B1,$DB,$C5,$FE,$0E,$AF,$9B,$F0,$DF,$27,$BA,$5F,$F8,$BD
	.dc.b	$F9,$1E,$94,$8F,$C3,$BF,$18,$8F,$F9,$9F,$8B,$FC,$26,$FF,$8D,$FD
	.dc.b	$0F,$A8,$78,$FE,$A7,$C9,$DB,$67,$D1,$A2,$73,$1F,$E2,$DE,$5F,$F5
	.dc.b	$93,$7D,$CF,$C9,$F7,$FD,$E6,$FE,$25,$F9,$49,$70,$7F,$8A,$A9,$3F
	.dc.b	$AC,$E8,$FF,$17,$BB,$9D,$36,$33,$DD,$FA,$0B,$FF,$56,$FE,$1F,$7D
	.dc.b	$C3,$4D,$7E,$0F,$FE,$B1,$9F,$AA,$1C,$3F,$D7,$1A,$FF,$94,$1A,$FF
	.dc.b	$9E,$9A,$FF,$75,$5F,$15,$D8,$91,$DB,$ED,$5E,$DF,$C4,$FC,$79,$F1
	.dc.b	$54,$95,$FE,$57,$FC,$47,$E1,$3F,$E5,$45,$E9,$FC,$C7,$3C,$26,$9A
	.dc.b	$9F,$C7,$7C,$B7,$EC,$DF,$9F,$C1,$FB,$78,$C7,$17,$E3,$5B,$19,$FB
	.dc.b	$9F,$DC,$41,$BE,$84,$BE,$B1,$F4,$7E,$3F,$F1,$71,$D8,$FF,$0C,$38
	.dc.b	$C3,$FC,$BA,$DC,$3E,$57,$CE,$7D,$43,$FC,$17,$5F,$2D,$AF,$D5,$6F
	.dc.b	$11,$BE,$2F,$DC,$24,$FD,$CF,$C3,$FF,$86,$BF,$AF,$FB,$7F,$AA,$DE
	.dc.b	$5F,$E2,$91,$88,$F4,$78,$8F,$C4,$9F,$89,$7F,$19,$7F,$A7,$4F,$EE
	.dc.b	$5C,$B6,$35,$E9,$7E,$83,$E0,$5F,$37,$49,$5F,$E9,$3F,$DB,$5E,$57
	.dc.b	$E9,$9F,$E5,$DC,$7F,$A3,$73,$3B,$9D,$7A,$7F,$9F,$EE,$7C,$87,$EB
	.dc.b	$BE,$6F,$FC,$1F,$8F,$EA,$DB,$CD,$FE,$D0,$AD,$E3,$FA,$0E,$67,$E1
	.dc.b	$9F,$30,$91,$F4,$A3,$2D,$F0,$FF,$52,$47,$97,$D9,$FE,$FB,$F5,$FF
	.dc.b	$F1,$E4,$7F,$AB,$DA,$FE,$0F,$C0,$46,$34,$FE,$00,$FF,$95,$E7,$87
	.dc.b	$BF,$CF,$E1,$A4,$60,$FE,$51,$F8,$A3,$EC,$BF,$A6,$7E,$48,$F9,$5A
	.dc.b	$7F,$4F,$9D,$57,$0A,$C1,$1E,$7F,$37,$FC,$27,$F3,$BF,$27,$1F,$FB
	.dc.b	$D9,$B3,$FE,$71,$6A,$FA,$8F,$F5,$BE,$DF,$EB,$87,$FD,$73,$1F,$ED
	.dc.b	$66,$BF,$F0,$01,$AF,$F9,$ED,$AF,$3F,$89,$FE,$35,$1F,$62,$35,$1F
	.dc.b	$C7,$7F,$45,$FA,$71,$F9,$6F,$C5,$DC,$9F,$DB,$57,$25,$31,$1F,$DC
	.dc.b	$79,$7F,$B3,$57,$F7,$76,$BF,$E5,$96,$F8,$5F,$C8,$FE,$0B,$F6,$6F
	.dc.b	$CD,$F6,$7F,$BD,$71,$FF,$8B,$36,$FE,$A8,$DE,$0F,$F4,$B7,$C4,$BF
	.dc.b	$6E,$FE,$6D,$9F,$F9,$97,$F1,$5F,$8F,$5F,$E0,$76,$C6,$63,$FC,$33
	.dc.b	$4F,$DA,$F8,$F6,$3E,$7B,$FA,$6D,$FF,$A8,$CF,$F7,$9B,$87,$FD,$D6
	.dc.b	$D7,$3E,$DF,$AB,$F1,$6F,$F6,$1C,$7F,$71,$7F,$CE,$6D,$F9,$1F,$F0
	.dc.b	$B4,$6A,$3F,$81,$FC,$87,$85,$FE,$F0,$BA,$FF,$C3,$8D,$8C,$E1,$FE
	.dc.b	$1F,$FC,$59,$F8,$AF,$CA,$6C,$FF,$AA,$39,$11,$88,$FF,$1D,$B5,$FF
	.dc.b	$0D,$FB,$6F,$87,$FF,$15,$36,$3F,$C1,$4D,$7F,$98,$70,$7E,$97,$D5
	.dc.b	$FF,$18,$7B,$DF,$DD,$E7,$FC,$20,$D7,$F8,$FB,$79,$DC,$9F,$6B,$D5
	.dc.b	$7E,$37,$FD,$EB,$8B,$F9,$36,$1F,$47,$95,$18,$8F,$F6,$B4,$62,$BF
	.dc.b	$9D,$B6,$3E,$07,$EC,$7F,$C9,$1F,$F1,$17,$E3,$1F,$82,$E1,$36,$75
	.dc.b	$3F,$CD,$27,$C6,$FF,$8A,$5A,$E7,$1F,$C4,$FD,$F7,$F0,$38,$FF,$18
	.dc.b	$FC,$B3,$F1,$4F,$F4,$BC,$7E,$26,$31,$F0,$DF,$D4,$1B,$1F,$ED,$C6
	.dc.b	$BF,$E6,$EF,$C5,$9F,$53,$FE,$5B,$5F,$47,$4C,$47,$F9,$7F,$8B,$C5
	.dc.b	$5F,$D3,$1F,$E6,$2D,$F5,$B5,$E1,$FF,$89,$A1,$F6,$5F,$21,$A7,$FB
	.dc.b	$BB,$88,$E0,$FD,$11,$FE,$ED,$8F,$57,$E2,$9F,$C1,$23,$E0,$B5,$FE
	.dc.b	$E9,$A2,$31,$8F,$FA,$02,$7F,$B6,$1F,$23,$F2,$87,$CA,$FF,$A3,$2F
	.dc.b	$FC,$23,$F8,$6D,$F8,$C3,$9A,$DD,$FD,$0F,$C4,$BF,$B8,$BE,$6B,$46
	.dc.b	$23,$E9,$35,$F6,$C6,$BE,$47,$EA,$4B,$E6,$7C,$4D,$AE,$D1,$E5,$FC
	.dc.b	$53,$F9,$0C,$7F,$9C,$3E,$71,$F3,$DF,$DB,$3C,$0F,$93,$18,$F8,$3F
	.dc.b	$F5,$1C,$E6,$93,$C3,$E1,$32,$7F,$85,$1C,$47,$D0,$37,$3F,$DD,$9F
	.dc.b	$A7,$1A,$7E,$07,$AD,$FE,$8B,$E2,$FD,$6C,$36,$FE,$6F,$E2,$FC,$DA
	.dc.b	$D6,$7F,$CA,$BE,$8F,$FD,$02,$DF,$9C,$FE,$4B,$18,$C1,$E4,$23,$18
	.dc.b	$BF,$13,$B7,$FE,$43,$BF,$C0,$FE,$51,$F7,$9B,$F2,$3F,$76,$75,$AF
	.dc.b	$9B,$0F,$F4,$E7,$46,$49,$7F,$AC,$AF,$E6,$BF,$89,$FE,$02,$3F,$BF
	.dc.b	$7D,$4B,$EF,$7F,$D0,$2D,$8F,$F5,$1F,$7A,$B8,$65,$FE,$81,$6F,$07
	.dc.b	$F0,$A6,$D2,$B8,$76,$FF,$A3,$27,$F9,$05,$F1,$F4,$3F,$87,$C7,$F2
	.dc.b	$3B,$FE,$2C,$D7,$A4,$7D,$BF,$AC,$FC,$CC,$7E,$1E,$DF,$45,$FE,$E6
	.dc.b	$6C,$7E,$2E,$76,$FF,$B8,$9B,$87,$FB,$A1,$B9,$7E,$39,$B3,$F8,$C7
	.dc.b	$9B,$FE,$65,$1F,$2B,$FE,$F3,$AF,$F3,$0B,$63,$F1,$CD,$7F,$7E,$D7
	.dc.b	$FB,$13,$FE,$23,$FC,$4F,$8F,$C4,$1A,$49,$FF,$35,$FD,$47,$F9,$B1
	.dc.b	$34,$8F,$E6,$BE,$53,$FC,$5D,$97,$EA,$DB,$1F,$E7,$DE,$C4,$3E,$9F
	.dc.b	$F3,$B9,$FE,$7A,$73,$96,$30,$FD,$8D,$F2,$AD,$3F,$CB,$23,$B4,$EF
	.dc.b	$0F,$D9,$CE,$5F,$F2,$57,$D4,$7F,$AA,$AB,$F9,$23,$63,$F9,$1F,$12
	.dc.b	$31,$1F,$D5,$DA,$FF,$56,$5F,$65,$DE,$78,$36,$3A,$63,$FE,$D5,$77
	.dc.b	$7F,$9A,$6D,$F9,$3C,$BE,$17,$C5,$F5,$74,$BE,$73,$D5,$8F,$EB,$5B
	.dc.b	$8F,$F2,$07,$FA,$5B,$FD,$61,$CE,$1F,$C1,$DB,$5C,$F4,$EB,$18,$3E
	.dc.b	$65,$7F,$2A,$FC,$4B,$FB,$DB,$FD,$23,$0F,$E2,$3E,$03,$F1,$3F,$6C
	.dc.b	$EC,$FA,$11,$9F,$72,$CE,$B3,$FE,$1E,$DC,$D8,$F6,$9F,$F0,$8C,$EF
	.dc.b	$85,$F9,$F7,$FB,$27,$2F,$E5,$11,$8F,$A0,$FF,$32,$B6,$33,$3B,$BF
	.dc.b	$E5,$7F,$89,$FC,$2A,$36,$75,$AF,$E3,$E3,$E2,$FE,$25,$3D,$5F,$A2
	.dc.b	$FC,$CC,$3F,$B0,$BF,$2B,$F1,$27,$E2,$7F,$E7,$6D,$D1,$8D,$3F,$C0
	.dc.b	$4D,$BF,$EB,$4D,$7F,$C1,$4F,$F9,$C2,$3F,$C1,$FF,$89,$FF,$B5,$5F
	.dc.b	$F0,$8C,$66,$7F,$B8,$CE,$5F,$D5,$66,$F8,$4F,$D0,$6C,$8C,$61,$FE
	.dc.b	$1A,$FA,$17,$DD,$EA,$73,$87,$FA,$B5,$FC,$9B,$FE,$35,$6B,$FE,$34
	.dc.b	$6B,$FC,$AD,$BF,$05,$FC,$B7,$DD,$FD,$83,$FC,$BF,$A6,$F6,$FC,$9A
	.dc.b	$FE,$B3,$39,$CB,$F1,$87,$F8,$51,$7E,$25,$FE,$DC,$8C,$E2,$9C,$FD
	.dc.b	$77,$6A,$D4,$FF,$BC,$14,$78,$A1,$F6,$9F,$EA,$AA,$FF,$10,$B6,$33
	.dc.b	$7F,$E6,$B6,$9F,$E2,$4E,$C7,$C2,$7C,$FC,$BF,$42,$DA,$F4,$7E,$25
	.dc.b	$E2,$9C,$FA,$3F,$0B,$E6,$DF,$8C,$DF,$84,$C3,$70,$FE,$B7,$F6,$EF
	.dc.b	$94,$FF,$5F,$F1,$23,$13,$FE,$39,$6E,$DF,$C5,$9A,$FF,$A6,$F6,$7E
	.dc.b	$27,$F8,$FD,$7F,$6E,$DB,$F9,$DF,$89,$FE,$2E,$73,$DD,$FB,$8F,$C4
	.dc.b	$FF,$07,$1F,$BE,$6D,$7D,$5F,$C4,$FD,$CF,$FA,$DA,$33,$97,$D5,$FC
	.dc.b	$4D,$F8,$27,$CF,$6D,$7F,$A5,$DB,$3F,$E5,$0E,$74,$62,$3F,$D6,$91
	.dc.b	$FC,$B3,$A3,$F1,$7E,$6B,$E7,$EC,$61,$FC,$7F,$E2,$7F,$E2,$3C,$7F
	.dc.b	$0F,$7F,$27,$DE,$CF,$8D,$F9,$4E,$44,$7F,$37,$FB,$A7,$57,$D0,$6E
	.dc.b	$1F,$E6,$76,$EB,$FD,$11,$AF,$CF,$F8,$FF,$E3,$1C,$7F,$C9,$1B,$5F
	.dc.b	$9B,$FC,$4D,$E9,$FF,$24,$9C,$DF,$F8,$C3,$F1,$2F,$C7,$E1,$FC,$B1
	.dc.b	$B8,$7E,$8F,$F1,$37,$C4,$7F,$2C,$9C,$DF,$E8,$FE,$A7,$FC,$AB,$7F
	.dc.b	$EC,$36,$C7,$CD,$6B,$F3,$1A,$FE,$BF,$EB,$7C,$B3,$F2,$1F,$E6,$DC
	.dc.b	$47,$33,$FC,$2E,$78,$7A,$FF,$DB,$61,$F8,$9F,$E7,$A7,$F6,$F1,$8F
	.dc.b	$0F,$5F,$63,$F9,$18,$FF,$48,$D6,$DF,$F0,$D7,$E1,$AF,$C2,$3F,$CB
	.dc.b	$B1,$FE,$94,$8C,$61,$FE,$67,$84,$61,$FF,$35,$35,$E1,$F1,$3E,$58
	.dc.b	$FF,$25,$4F,$91,$E1,$C3,$D9,$CD,$7B,$4F,$E1,$DF,$89,$FF,$BC,$2B
	.dc.b	$F1,$6D,$8D,$6C,$8F,$D0,$B7,$E8,$FF,$90,$C6,$F0,$FE,$3E,$FF,$94
	.dc.b	$6F,$9C,$EB,$C8,$FF,$04,$7E,$27,$FB,$39,$FF,$70,$AF,$53,$3F,$23
	.dc.b	$A6,$4F,$F9,$8B,$9B,$F1,$76,$BE,$13,$93,$E1,$74,$7C,$3F,$06,$C7
	.dc.b	$C3,$70,$D3,$E3,$46,$36,$FA,$BF,$59,$F2,$CC,$6D,$F2,$5F,$F1,$07
	.dc.b	$E2,$7E,$23,$C7,$F0,$9E,$2A,$F9,$99,$A7,$5E,$07,$0F,$53,$8B,$73
	.dc.b	$FD,$CF,$E2,$5F,$B6,$6F,$CA,$FF,$4C,$8C,$63,$F4,$9F,$CF,$E9,$FD
	.dc.b	$31,$7D,$06,$CE,$15,$6F,$F2,$BD,$4F,$A5,$FC,$D3,$63,$F3,$3F,$70
	.dc.b	$FC,$0F,$35,$F8,$BF,$9A,$FC,$AE,$9F,$EE,$A3,$F2,$6E,$1F,$F7,$6F
	.dc.b	$67,$F5,$47,$F3,$51,$97,$F3,$3F,$37,$F9,$53,$A6,$3F,$95,$7F,$B4
	.dc.b	$FE,$22,$FB,$9B,$C3,$86,$D7,$F1,$E9,$FE,$2B,$C3,$FD,$B1,$FE,$0B
	.dc.b	$DE,$73,$3F,$DB,$5F,$E1,$3C,$BF,$EE,$3F,$C8,$BF,$CB,$1F,$40,$6F
	.dc.b	$9F,$87,$3E,$E6,$BE,$B6,$E3,$FE,$B2,$B4,$7F,$03,$F9,$DF,$F7,$74
	.dc.b	$FF,$02,$6F,$07,$EA,$3F,$13,$FE,$92,$84,$6B,$6F,$DA,$7E,$07,$F1
	.dc.b	$F9,$0D,$26,$1F,$C0,$FB,$3F,$59,$87,$FA,$8D,$B7,$CA,$7F,$89,$C6
	.dc.b	$1F,$06,$EF,$C4,$FF,$B5,$D7,$FB,$B2,$31,$B7,$F9,$EE,$7E,$25,$7F
	.dc.b	$B4,$1D,$BC,$9F,$82,$E7,$EC,$B8,$6D,$C5,$FF,$11,$FE,$15,$FE,$21
	.dc.b	$9F,$F0,$F3,$73,$FF,$0E,$2F,$AE,$DF,$C5,$F0,$F8,$BE,$3F,$25,$70
	.dc.b	$FE,$BB,$7F,$E7,$CE,$83,$FE,$A9,$8F,$CA,$63,$DB,$E6,$BF,$C4,$13
	.dc.b	$98,$FE,$E0,$D7,$F3,$52,$FC,$49,$F1,$5F,$EA,$98,$C3,$EA,$BC,$4E
	.dc.b	$A7,$C7,$C1,$EE,$3D,$FB,$58,$C9,$E8,$DF,$18,$FF,$2A,$CB,$E4,$B7
	.dc.b	$A7,$CE,$4A,$F5,$C7,$C6,$8F,$D7,$43,$77,$DD,$79,$DB,$1F,$1A,$71
	.dc.b	$3F,$4B,$AF,$E1,$92,$7A,$EF,$F2,$A3,$AE,$BF,$E6,$69,$FF,$28,$9F
	.dc.b	$17,$E4,$39,$38,$8D,$65,$EB,$F9,$2F,$F3,$8F,$E2,$4F,$A3,$FC,$0C
	.dc.b	$61,$97,$FE,$5C,$F9,$45,$6E,$3E,$ED,$6A,$FE,$37,$AD,$F6,$89,$3E
	.dc.b	$8E,$0F,$EA,$93,$F6,$B2,$7E,$09,$F5,$8F,$C8,$DD,$BF,$F3,$5F,$5B
	.dc.b	$F1,$2F,$F4,$E7,$D0,$FF,$93,$F1,$4E,$79,$17,$FB,$7A,$D4,$F1,$EC
	.dc.b	$CD,$A7,$38,$74,$E2,$FD,$7B,$67,$AA,$FC,$37,$8D,$70,$CA,$DF,$09
	.dc.b	$B7,$C1,$B8,$FA,$5B,$87,$58,$C6,$5E,$7B,$FD,$FF,$D2,$FF,$44,$8C
	.dc.b	$61,$EE,$8C,$3E,$99,$DF,$C4,$F6,$9C,$E7,$F0,$1B,$3F,$C2,$BE,$8F
	.dc.b	$F8,$D6,$1D,$31,$56,$B1,$F9,$75,$A7,$B1,$83,$D6,$E7,$0E,$DE,$9E
	.dc.b	$26,$61,$D6,$B5,$8F,$B2,$B3,$5D,$31,$7E,$23,$FE,$58,$C7,$EB,$BF
	.dc.b	$E9,$C6,$F9,$BF,$2A,$31,$87,$F2,$46,$DF,$E0,$9B,$7D,$2B,$87,$8D
	.dc.b	$33,$F8,$15,$A3,$C1,$87,$39,$FB,$38,$3F,$11,$EA,$DF,$39,$F1,$BF
	.dc.b	$52,$35,$5C,$2B,$FB,$14,$E5,$B3,$F2,$8E,$51,$99,$E1,$5D,$27,$55
	.dc.b	$C1,$FE,$4A,$D7,$FB,$D3,$5F,$EA,$8D,$7D,$1F,$0F,$FA,$47,$F8,$9C
	.dc.b	$7F,$99,$74,$FF,$2A,$36,$33,$3B,$7B,$BC,$75,$FD,$74,$74,$AF,$64
	.dc.b	$E7,$2F,$1F,$D7,$7F,$83,$6F,$78,$23,$13,$E8,$AE,$18,$7A,$1A,$FC
	.dc.b	$F9,$DB,$FB,$CF,$BC,$FF,$4F,$BF,$97,$6B,$F8,$96,$E1,$FB,$E7,$FD
	.dc.b	$D3,$1B,$DF,$9E,$49,$DE,$3F,$DB,$1B,$5D,$F4,$DC,$72,$8E,$D7,$83
	.dc.b	$33,$D3,$EA,$FF,$C7,$33,$FA,$B6,$BF,$9D,$6E,$3F,$C4,$97,$B5,$C7
	.dc.b	$8A,$57,$17,$A4,$48,$C1,$0E,$8E,$5A,$78,$5F,$27,$2F,$C0,$EB,$7B
	.dc.b	$D4,$95,$FC,$06,$3F,$DA,$CD,$F8,$0C,$BF,$B5,$8E,$2E,$48,$66,$3B
	.dc.b	$1D,$FE,$4B,$EC,$F3,$BF,$11,$7C,$0F,$F9,$47,$D3,$FE,$63,$39,$8C
	.dc.b	$3F,$C5,$27,$78,$7F,$82,$39,$32,$FF,$84,$1F,$F5,$67,$93,$FE,$0A
	.dc.b	$6F,$89,$F9,$99,$FF,$79,$B5,$FF,$1C,$78,$EF,$3D,$ED,$74,$8C,$36
	.dc.b	$33,$3F,$EF,$4F,$92,$FF,$1D,$B7,$D6,$D5,$48,$C1,$FD,$C3,$3F,$96
	.dc.b	$FF,$AD,$74,$F8,$5C,$1E,$A6,$C7,$5F,$A6,$F6,$EC,$C9,$1E,$E6,$CF
	.dc.b	$F8,$01,$AF,$F8,$A2,$78,$4F,$FB,$3E,$9F,$88,$FF,$BC,$AB,$FC,$06
	.dc.b	$D7,$C8,$DE,$1F,$EC,$77,$E1,$FC,$53,$C7,$89,$F1,$12,$FC,$4F,$97
	.dc.b	$DF,$DF,$B9,$27,$E4,$47,$22,$38,$3E,$66,$DF,$0F,$6F,$F0,$84,$7C
	.dc.b	$9F,$47,$07,$8B,$96,$47,$C1,$7E,$21,$1F,$89,$D1,$E8,$E7,$91,$F6
	.dc.b	$63,$D7,$F0,$D2,$3E,$E3,$EE,$E3,$FC,$6F,$43,$F1,$2F,$DC,$F1,$7F
	.dc.b	$9B,$5F,$DB,$DF,$F8,$49,$AF,$FB,$19,$DF,$27,$F9,$D9,$75,$F4,$6C
	.dc.b	$F1,$8E,$FF,$90,$7F,$98,$2F,$FC,$27,$2F,$C1,$5E,$B7,$E8,$23,$0D
	.dc.b	$8C,$C7,$FA,$4E,$78,$C7,$FB,$16,$31,$8F,$FA,$FE,$31,$1F,$E5,$46
	.dc.b	$BF,$E7,$06,$BB,$F8,$4F,$F2,$CB,$FE,$9F,$F8,$67,$C9,$F8,$DB,$1F
	.dc.b	$ED,$08,$FC,$CF,$C2,$38,$3E,$5F,$E2,$5F,$02,$BF,$DD,$6F,$C6,$7F
	.dc.b	$C6,$ED,$E0,$3B,$D3,$FD,$39,$F8,$C7,$C5,$D7,$FD,$42,$DB,$FF,$1F
	.dc.b	$B6,$CF,$F6,$98,$57,$E7,$3F,$13,$FD,$3B,$7C,$1F,$D7,$C7,$EF,$27
	.dc.b	$78,$7F,$AF,$7B,$D3,$C2,$30,$FE,$8E,$7E,$68,$9D,$F8,$3D,$0E,$BD
	.dc.b	$FF,$5F,$E9,$A5,$70,$3E,$F3,$F6,$FE,$42,$47,$D0,$E4,$FD,$B3,$F5
	.dc.b	$63,$E9,$BF,$85,$6F,$E2,$6F,$BC,$FC,$33,$67,$B3,$5F,$69,$FE,$D9
	.dc.b	$FB,$3F,$F4,$93,$EA,$CD,$F8,$9B,$F1,$3F,$E2,$6D,$9D,$35,$E7,$3C
	.dc.b	$3C,$CD,$FC,$8F,$06,$4F,$08,$FD,$6F,$E2,$7F,$CB,$23,$C7,$CC,$87
	.dc.b	$91,$A3,$7E,$47,$0F,$26,$5F,$FA,$0A,$BF,$DE,$2D,$E3,$FD,$27,$CE
	.dc.b	$48,$F9,$DC,$68,$FD,$47,$C4,$49,$CB,$5F,$8A,$7E,$C7,$E6,$7F,$5C
	.dc.b	$67,$AF,$F5,$B3,$89,$FF,$49,$FE,$24,$F7,$BF,$84,$AF,$76,$98,$8C
	.dc.b	$FB,$B6,$B8,$C6,$1F,$A5,$F8,$8B,$E3,$BF,$A1,$AF,$92,$A6,$22,$4F
	.dc.b	$06,$BF,$52,$7F,$AF,$3F,$8F,$FA,$0F,$E1,$BA,$7F,$0E,$6F,$84,$99
	.dc.b	$18,$6F,$BF,$FC,$46,$3F,$C5,$F1,$FE,$F6,$6B,$C4,$38,$78,$9F,$C2
	.dc.b	$BF,$10,$FA,$4F,$E8,$5A,$FD,$06,$DF,$F2,$2F,$2D,$F4,$3F,$D9,$3F
	.dc.b	$10,$CE,$5F,$B5,$7F,$DD,$38,$7E,$65,$AF,$E5,$1D,$BF,$B5,$E5,$FE
	.dc.b	$28,$FF,$BA,$1A,$FF,$69,$6B,$EB,$EE,$FE,$65,$E5,$FF,$22,$FB,$07
	.dc.b	$E2,$1F,$D0,$34,$FF,$64,$C6,$3E,$03,$FC,$2A,$DC,$3F,$59,$F4,$FF
	.dc.b	$E0,$2A,$DF,$E1,$FC,$75,$FA,$16,$8C,$5A,$31,$87,$F8,$13,$99,$F4
	.dc.b	$7F,$D5,$BE,$63,$F0,$D7,$F4,$4F,$B5,$E2,$23,$31,$98,$C7,$47,$F3
	.dc.b	$C6,$C7,$13,$F8,$D7,$E1,$BF,$E0,$3D,$3F,$91,$2F,$C0,$BF,$C2,$AD
	.dc.b	$9F,$E1,$0E,$DF,$E4,$5F,$87,$BE,$73,$FC,$36,$F8,$E3,$F8,$73,$5F
	.dc.b	$E1,$2E,$DF,$E2,$5F,$87,$FF,$8D,$DF,$E5,$8D,$E5,$34,$D7,$07,$E9
	.dc.b	$7D,$1F,$A3,$DF,$E3,$8C,$5F,$F1,$DF,$9B,$FF,$7C,$3F,$CE,$21,$ED
	.dc.b	$BE,$67,$EC,$B6,$74,$D7,$F8,$FF,$B9,$FE,$62,$7F,$8C,$EC,$F4,$3E
	.dc.b	$63,$63,$E6,$71,$23,$11,$F4,$DA,$FE,$3F,$F0,$E7,$DA,$FF,$90,$9F
	.dc.b	$BF,$1F,$CD,$5A,$FF,$30,$76,$FA,$FE,$E5,$DB,$F9,$26,$C6,$63,$FC
	.dc.b	$9B,$C5,$FD,$02,$3F,$A1,$FC,$B3,$F0,$EF,$F0,$DC,$7E,$EE,$31,$DB
	.dc.b	$FE,$02,$6C,$7E,$1D,$AF,$F9,$47,$F0,$EF,$E2,$CF,$F4,$B5,$F4,$9F
	.dc.b	$E5,$C6,$C6,$79,$7F,$CB,$FE,$0B,$ED,$2E,$92,$BF,$AA,$4E,$5F,$ED
	.dc.b	$FC,$3F,$C5,$FE,$2B,$F3,$51,$F4,$CF,$E2,$51,$97,$E8,$99,$3F,$D0
	.dc.b	$23,$DF,$FF,$40,$34,$FE,$F7,$F2,$5F,$C8,$E3,$FD,$1F,$AB,$F0,$EF
	.dc.b	$E1,$6B,$F0,$CF,$ED,$99,$3C,$2F,$E5,$52,$FC,$35,$F8,$4F,$26,$E7
	.dc.b	$FE,$08,$9E,$5F,$09,$DB,$17,$D6,$7C,$F6,$D7,$CF,$6B,$FB,$FA,$F1
	.dc.b	$4F,$D5,$FC,$35,$ED,$FA,$1B,$5D,$DA,$FE,$01,$FC,$4B,$FE,$04,$DF
	.dc.b	$FD,$A7,$F0,$3F,$EA,$3E,$8F,$EE,$73,$F1,$1A,$EF,$E1,$5F,$3F,$FE
	.dc.b	$A9,$6C,$7F,$0B,$B4,$62,$3F,$8B,$46,$27,$FC,$9A,$D8,$FE,$25,$E2
	.dc.b	$BD,$BF,$EF,$B1,$FE,$18,$8C,$7C,$17,$F2,$E9,$CF,$27,$D1,$A6,$EF
	.dc.b	$C3,$73,$CB,$F5,$1B,$3F,$CB,$6E,$E3,$D2,$3E,$A4,$EA,$D8,$FF,$34
	.dc.b	$6F,$C8,$FE,$12,$3F,$CA,$D8,$FE,$01,$D1,$36,$D1,$BF,$9C,$FF,$46
	.dc.b	$FE,$18,$F9,$EB,$D3,$34,$FC,$EC,$E5,$B3,$C0,$8E,$B9,$B9,$9F,$CE
	.dc.b	$E7,$F0,$B1,$FA,$26,$BF,$D1,$1A,$B8,$5F,$4B,$FE,$B0,$8F,$C1,$2F
	.dc.b	$C3,$BF,$DC,$8D,$8F,$D2,$BB,$E3,$FA,$15,$CB,$E6,$F6,$4A,$DF,$B9
	.dc.b	$FE,$99,$FC,$37,$FD,$E1,$7F,$E8,$C6,$CF,$F5,$16,$BF,$D6,$DA,$FF
	.dc.b	$83,$5F,$F1,$17,$E1,$BC,$7C,$63,$49,$3F,$E6,$6F,$A3,$FF,$33,$A6
	.dc.b	$91,$FF,$4E,$FD,$87,$E2,$70,$FF,$58,$C2,$31,$CB,$FD,$B7,$C7,$B3
	.dc.b	$FE,$1F,$9A,$BE,$AF,$FC,$81,$5F,$E7,$D6,$C7,$F9,$30,$E9,$19,$F8
	.dc.b	$AE,$7B,$BE,$A9,$CF,$E2,$BF,$AD,$18,$BF,$D6,$36,$3F,$56,$D7,$FC
	.dc.b	$65,$5C,$FA,$3A,$C9,$B7,$87,$C3,$FB,$7F,$3D,$7E,$DB,$71,$FF,$22
	.dc.b	$61,$CD,$FE,$4B,$19,$C7,$FD,$A1,$C1,$F0,$AF,$FC,$AD,$5F,$CA,$BF
	.dc.b	$2E,$7D,$96,$6B,$DF,$1F,$E0,$59,$D7,$BF,$96,$9F,$EA,$5C,$BE,$3E
	.dc.b	$3F,$E7,$46,$E3,$F9,$0F,$94,$7D,$7F,$D2,$9C,$DF,$A2,$7F,$BE,$3F
	.dc.b	$CF,$1B,$7C,$F8,$BE,$8C,$63,$0F,$F7,$8D,$BE,$B3,$E9,$5F,$E9,$DB
	.dc.b	$19,$C3,$F8,$CC,$63,$E5,$3C,$47,$8D,$B1,$EB,$F2,$FC,$C7,$E0,$3F
	.dc.b	$30,$FF,$1B,$46,$4F,$EA,$3F,$86,$FF,$48,$7F,$95,$37,$1F,$E5,$2D
	.dc.b	$7E,$FB,$5F,$AF,$F9,$5F,$F8,$0A,$7F,$AD,$B6,$3F,$9F,$4F,$F1,$19
	.dc.b	$CF,$4F,$E6,$BC,$97,$8F,$4C,$74,$5A,$FF,$6C,$7E,$C7,$F5,$31,$FF
	.dc.b	$28,$35,$FF,$41,$35,$FF,$46,$35,$3F,$E2,$78,$E7,$F0,$1F,$AA,$87
	.dc.b	$CD,$D8,$FD,$2F,$EE,$1F,$C4,$EB,$FB,$AB,$6F,$FA,$C7,$1F,$FA,$9E
	.dc.b	$73,$E3,$7E,$DB,$D8,$FF,$80,$3E,$C7,$EA,$DB,$3F,$E3,$4A,$FF,$01
	.dc.b	$36,$74,$FF,$97,$CF,$F0,$79,$6F,$FB,$33,$F1,$8C,$FC,$57,$F0,$2F
	.dc.b	$C3,$7D,$31,$FE,$FD,$6C,$FF,$B3,$1F,$F2,$7C,$67,$4E,$1C,$DF,$8E
	.dc.b	$9F,$C6,$57,$F9,$CD,$AF,$F9,$F5,$AF,$FA,$BF,$C5,$FF,$08,$5F,$FA
	.dc.b	$E7,$2F,$EC,$53,$FC,$7E,$3C,$9F,$1D,$FC,$5A,$B7,$3F,$EF,$0F,$D0
	.dc.b	$7F,$C6,$15,$FE,$2C,$6C,$7F,$80,$FE,$ED,$F4,$7F,$2E,$73,$8F,$F6
	.dc.b	$FD,$FF,$C0,$3E,$97,$CD,$F5,$FF,$25,$8A,$73,$97,$F9,$8A,$73,$0F
	.dc.b	$C1,$9F,$84,$FD,$76,$D7,$F8,$CB,$F0,$8F,$DE,$BF,$E4,$08,$CE,$1F
	.dc.b	$EF,$6F,$62,$E7,$F0,$1B,$5F,$E5,$9F,$17,$F9,$AC,$62,$BF,$C6,$BF
	.dc.b	$4B,$FB,$C8,$34,$95,$FE,$5F,$FA,$2E,$27,$2F,$16,$C7,$F9,$4B,$F0
	.dc.b	$17,$DE,$7E,$3D,$B1,$FE,$4A,$F2,$DF,$4B,$9A,$FC,$A7,$DA,$BE,$CB
	.dc.b	$F2,$8D,$BF,$F4,$3B,$75,$FF,$5C,$35,$FF,$2E,$FD,$0F,$49,$FF,$31
	.dc.b	$B6,$3F,$C7,$5E,$CF,$F5,$39,$CE,$3F,$D5,$BF,$0B,$FF,$22,$C7,$E1
	.dc.b	$CF,$F7,$C3,$8A,$71,$7E,$2A,$1B,$F1,$F3,$B9,$7C,$9F,$F2,$49,$FF
	.dc.b	$4B,$E1,$FE,$C9,$C1,$18,$7F,$D9,$8D,$7F,$DA,$DF,$86,$FF,$8F,$67
	.dc.b	$C5,$01,$19,$7C,$71,$E6,$AF,$ED,$F3,$9C,$7F,$3F,$DD,$F9,$C6,$D7
	.dc.b	$BF,$C0,$FE,$EB,$39,$8F,$F7,$2F,$E7,$2F,$4F,$F8,$33,$63,$C7,$92
	.dc.b	$B8,$3F,$9B,$9D,$3E,$43,$D1,$0F,$BC,$EE,$78,$3E,$A6,$DF,$4F,$C1
	.dc.b	$3F,$25,$66,$DF,$19,$D7,$7F,$A3,$DE,$91,$E4,$97,$D1,$E3,$0F,$11
	.dc.b	$E7,$F8,$AF,$79,$9D,$FE,$2B,$FC,$8F,$E0,$FF,$CC,$B8,$7F,$AA,$9B
	.dc.b	$5C,$B0,$F9,$93,$F4,$FC,$45,$6B,$0F,$9F,$F8,$47,$CC,$3E,$6F,$E1
	.dc.b	$36,$FE,$4F,$C8,$F2,$CB,$EE,$35,$F9,$95,$FE,$28,$8C,$65,$FC,$EF
	.dc.b	$0F,$B7,$E9,$7D,$8F,$87,$FB,$2F,$F3,$BF,$D5,$3C,$59,$79,$5B,$97
	.dc.b	$D2,$9D,$78,$FF,$80,$F0,$A5,$FC,$9D,$AD,$FB,$FE,$57,$FB,$EE,$86
	.dc.b	$92,$BF,$39,$AB,$7F,$41,$F8,$C9,$B3,$F5,$7E,$F3,$F1,$E7,$E4,$3E
	.dc.b	$F2,$C7,$FB,$63,$E9,$3E,$27,$19,$F8,$35,$FD,$5D,$F8,$7D,$9F,$98
	.dc.b	$3F,$33,$F2,$96,$FF,$2B,$F2,$8F,$A4,$FF,$28,$B6,$3F,$0D,$58,$9F
	.dc.b	$D0,$F8,$D2,$7F,$15,$F2,$97,$87,$F0,$1E,$B7,$C2,$24,$FE,$EF,$8B
	.dc.b	$FA,$0B,$FB,$59,$FF,$76,$FE,$6B,$F5,$E3,$F0,$CD,$8F,$D1,$D7,$F7
	.dc.b	$18,$C4,$7E,$97,$7F,$F3,$3F,$85,$FD,$16,$FF,$39,$F6,$F8,$9F,$F3
	.dc.b	$5F,$94,$FB,$BA,$66,$BF,$6B,$19,$7F,$6F,$7F,$ED,$26,$FA,$FD,$B7
	.dc.b	$EE,$FF,$80,$FF,$07,$FF,$10,$67,$FB,$9C,$91,$8F,$03,$FB,$4E,$3C
	.dc.b	$EB,$53,$9C,$7F,$8B,$7C,$A3,$C9,$FE,$69,$58,$6B,$FE,$94,$6C,$6F
	.dc.b	$BB,$FC,$25,$CD,$FE,$0B,$CB,$FB,$53,$70,$E5,$3F,$E1,$98,$C6,$1F
	.dc.b	$EB,$7E,$B7,$D0,$7F,$A2,$1B,$7F,$CA,$B5,$7E,$2F,$FE,$7F,$8F,$E5
	.dc.b	$BF,$1D,$F9,$DB,$FF,$12,$4F,$F9,$E6,$B7,$F1,$1C,$E7,$2F,$F8,$A7
	.dc.b	$F1,$8F,$5C,$9A,$4D,$3F,$C5,$9D,$0F,$3F,$FC,$60,$DC,$FC,$95,$FD
	.dc.b	$6F,$7F,$F0,$44,$E7,$6F,$E5,$5C,$0F,$D2,$7F,$C7,$51,$F8,$F3,$FD
	.dc.b	$7A,$E1,$FC,$94,$6F,$EB,$BF,$D4,$CD,$C7,$84,$FF,$84,$7F,$60,$F8
	.dc.b	$4F,$DF,$6E,$29,$FB,$78,$78,$23,$FC,$39,$3A,$9F,$EA,$9F,$1D,$FE
	.dc.b	$E7,$F8,$EF,$F6,$0B,$63,$FA,$CB,$FE,$EB,$F9,$8C,$F2,$7F,$91,$72
	.dc.b	$FE,$8D,$3F,$EE,$FE,$2C,$9A,$C2,$E1,$F6,$78,$3F,$5F,$7F,$07,$8B
	.dc.b	$E6,$70,$7F,$96,$67,$33,$FB,$39,$CD,$7C,$0C,$5F,$8A,$FF,$AC,$AB
	.dc.b	$C2,$DD,$BF,$C7,$18,$76,$7E,$14,$AB,$8F,$CB,$FF,$B7,$3E,$95,$F0
	.dc.b	$74,$C7,$C6,$5C,$FF,$B3,$CE,$9F,$EF,$5F,$16,$F0,$3F,$84,$4F,$F8
	.dc.b	$2A,$73,$18,$BF,$0D,$6F,$2F,$F0,$15,$ED,$FF,$7C,$57,$2F,$0B,$6D
	.dc.b	$C3,$9C,$F0,$C3,$2F,$2A,$AA,$FF,$21,$87,$78,$C6,$BA,$9F,$F0,$C4
	.dc.b	$E5,$DE,$3E,$A9,$E7,$F8,$47,$F9,$AA,$B5,$18,$CA,$1E,$4C,$15,$BB
	.dc.b	$E5,$39,$8C,$4F,$F8,$85,$AF,$FB,$DF,$CE,$E6,$FC,$23,$FC,$4D,$CC
	.dc.b	$8F,$7F,$13,$F6,$C7,$36,$1B,$8F,$C6,$9D,$5E,$6A,$4F,$CA,$E2,$FF
	.dc.b	$06,$F9,$98,$CB,$C1,$18,$9D,$7B,$3F,$E2,$2D,$FE,$C4,$E6,$BC,$D9
	.dc.b	$78,$F1,$FA,$73,$9A,$FA,$1A,$FF,$87,$EF,$D8,$7C,$6C,$1E,$A7,$C5
	.dc.b	$F1,$FE,$7C,$62,$77,$5F,$E2,$2A,$F9,$53,$9B,$EB,$7C,$27,$27,$F9
	.dc.b	$56,$B8,$5F,$F0,$66,$BF,$C3,$DA,$FF,$5F,$5E,$D7,$FA,$6E,$38,$35
	.dc.b	$E0,$D7,$83,$5F,$B8,$F1,$BF,$82,$DE,$1D,$E9,$F8,$3F,$6B,$FD,$79
	.dc.b	$1E,$16,$DF,$29,$FC,$45,$71,$8F,$C4,$4E,$B4,$7D,$D3,$EC,$30,$DA
	.dc.b	$FB,$CD,$7E,$BF,$ED,$8E,$3E,$F9,$23,$13,$C2,$F9,$9A,$8C,$C7,$2A
	.dc.b	$C1,$1F,$C8,$47,$E5,$1F,$D0,$7B,$BE,$C5,$F3,$3F,$2F,$6F,$DA,$37
	.dc.b	$6F,$A5,$5B,$DC,$D2,$5F,$EA,$FD,$ED,$18,$7F,$0F,$18,$74,$43,$F1
	.dc.b	$71,$D7,$DD,$ED,$E5,$49,$FD,$B3,$E1,$F3,$64,$60,$87,$E5,$1C,$BF
	.dc.b	$E7,$4C,$3F,$C9,$8F,$FA,$31,$AE,$7A,$FF,$C5,$9E,$17,$F3,$2E,$36
	.dc.b	$CF,$85,$CF,$5F,$F5,$28,$C6,$1F,$D4,$1A,$FF,$86,$FE,$AB,$FD,$38
	.dc.b	$FF,$80,$DA,$FF,$4B,$6B,$FD,$25,$AF,$F9,$FD,$AF,$FB,$E7,$EB,$3F
	.dc.b	$D8,$2F,$F8,$35,$AF,$F8,$A5,$B8,$FF,$4A,$E6,$47,$92,$30,$DC,$3F
	.dc.b	$D1,$8D,$AE,$D8,$FB,$9F,$07,$C3,$5C,$1F,$B1,$7F,$A4,$A7,$FC,$3C
	.dc.b	$D7,$FC,$74,$DC,$7F,$AD,$F9,$CF,$D0,$3F,$DD,$D5,$FE,$CE,$8C,$71
	.dc.b	$7F,$B5,$DB,$1F,$E0,$86,$BF,$E1,$16,$BF,$DB,$FC,$17,$D2,$F9,$5B
	.dc.b	$1F,$E3,$16,$BF,$E4,$DF,$C7,$5F,$45,$EB,$6C,$7F,$85,$B4,$46,$23
	.dc.b	$FC,$00,$D7,$FB,$F3,$5F,$F5,$A7,$5F,$F7,$47,$FC,$80,$D7,$FC,$B8
	.dc.b	$DD,$3F,$BE,$FA,$0F,$87,$FD,$95,$7F,$B6,$63,$19,$7D,$16,$C7,$F9
	.dc.b	$22,$77,$5F,$76,$FE,$FF,$1F,$FB,$17,$F4,$6F,$44,$FE,$05,$B1,$9D
	.dc.b	$3E,$E6,$CF,$C4,$9F,$75,$86,$E7,$F4,$5A,$FD,$16,$BF,$E4,$06,$BF
	.dc.b	$ED,$5F,$C2,$FB,$AE,$04,$61,$B8,$7E,$36,$5F,$7C,$F9,$9C,$36,$3F
	.dc.b	$E0,$0B,$5F,$F7,$6B,$75,$FF,$13,$C7,$2F,$8E,$E9,$E9,$3E,$7E,$92
	.dc.b	$79,$FC,$27,$62,$38,$3B,$F9,$5F,$9F,$F8,$AE,$51,$FE,$7B,$6B,$FE
	.dc.b	$97,$6E,$3E,$2C,$AF,$17,$FA,$D5,$DD,$7F,$9A,$5A,$E7,$1F,$F3,$2A
	.dc.b	$3E,$1B,$87,$E2,$FD,$1F,$C4,$23,$F1,$3E,$6F,$E5,$7C,$27,$E7,$B1
	.dc.b	$F8,$6D,$C7,$FD,$8C,$DA,$ED,$8B,$E8,$BE,$79,$23,$E7,$BF,$4B,$E3
	.dc.b	$BF,$80,$79,$0F,$0E,$92,$BE,$AF,$D1,$7D,$82,$38,$3E,$CF,$BB,$F7
	.dc.b	$FF,$00,$FB,$1A,$7F,$B6,$5B,$3F,$05,$AF,$E1,$3D,$EF,$C1,$9F,$0F
	.dc.b	$4F,$D6,$47,$CB,$8C,$19,$E5,$FC,$49,$23,$F1,$3E,$47,$F1,$9F,$C3
	.dc.b	$5F,$3B,$49,$3F,$8E,$D7,$F2,$24,$70,$7E,$4B,$DF,$FC,$C7,$E4,$5F
	.dc.b	$E1,$28,$FB,$1C,$9F,$92,$8C,$42,$31,$ED,$FE,$7F,$DA,$FC,$BC,$7E
	.dc.b	$61,$BC,$1F,$EF,$B6,$D7,$E0,$DA,$FF,$82,$BE,$67,$F9,$97,$E5,$EF
	.dc.b	$A0,$A4,$9F,$1D,$F3,$F8,$0F,$01,$23,$27,$57,$C9,$F0,$32,$32,$7F
	.dc.b	$86,$1C,$3E,$B9,$F3,$67,$FB,$9F,$C1,$3F,$D9,$D3,$C1,$AE,$CF,$CE
	.dc.b	$38,$D3,$E0,$7C,$0F,$C3,$24,$F8,$35,$47,$C7,$F9,$9F,$EA,$1F,$B8
	.dc.b	$FE,$5E,$BF,$DD,$8D,$7B,$79,$28,$C6,$1F,$37,$E0,$3E,$81,$2B,$B3
	.dc.b	$F4,$FE,$03,$F8,$97,$E6,$EF,$9C,$A4,$AF,$27,$C1,$7D,$93,$F6,$9A
	.dc.b	$9A,$48,$FD,$DC,$7F,$61,$E0,$FB,$9F,$3B,$F8,$13,$F0,$5E,$DF,$F6
	.dc.b	$BF,$CD,$7F,$0F,$AF,$F7,$76,$D7,$F0,$3D,$D1,$8C,$5E,$CB,$F0,$B7
	.dc.b	$C8,$7F,$75,$FC,$41,$F2,$3F,$C3,$31,$46,$30,$FE,$BC,$D7,$FD,$B8
	.dc.b	$D7,$FD,$B5,$FA,$4F,$F8,$93,$4F,$F0,$53,$74,$F2,$36,$7C,$7F,$8E
	.dc.b	$73,$AF,$E6,$6F,$F7,$E8,$FC,$10,$71,$1A,$DB,$FA,$14,$73,$9D,$7D
	.dc.b	$07,$F2,$18,$34,$91,$FC,$A3,$F2,$87,$CF,$7F,$4B,$6C,$7E,$69,$AF
	.dc.b	$89,$FF,$67,$61,$FE,$25,$E3,$FE,$A8,$4F,$6F,$FB,$87,$E0,$AF,$BC
	.dc.b	$FF,$10,$2F,$C1,$BE,$63,$71,$FF,$04,$35,$FE,$74,$7D,$97,$0E,$38
	.dc.b	$5F,$48,$F9,$FF,$F1,$63,$67,$F6,$B3,$A9,$FF,$01,$7E,$C3,$FC,$62
	.dc.b	$7F,$6B,$E3,$FF,$9C,$1B,$E0,$7F,$6B,$8C,$47,$D0,$5E,$06,$7D,$6F
	.dc.b	$E9,$91,$FE,$D7,$9F,$EA,$2D,$7D,$27,$F5,$37,$0E,$1F,$BD,$CF,$EE
	.dc.b	$6C,$6F,$A7,$C5,$F4,$9F,$E4,$7F,$92,$FF,$7F,$4F,$B1,$AF,$EE,$1B
	.dc.b	$C9,$FE,$24,$FC,$13,$F8,$0C,$FC,$21,$E5,$7E,$6D,$A7,$F8,$DC,$A7
	.dc.b	$A3,$83,$FB,$54,$7F,$6C,$F0,$1F,$24,$FA,$DF,$F1,$63,$6F,$F9,$1C
	.dc.b	$EA,$7F,$C6,$10,$F8,$5F,$F0,$B7,$EA,$8F,$A9,$A4,$9F,$E4,$F1,$F0
	.dc.b	$9F,$E3,$3F,$33,$FD,$E7,$CE,$34,$91,$FE,$00,$87,$D0,$7F,$88,$DB
	.dc.b	$1F,$CC,$9A,$FE,$A9,$FC,$33,$FE,$47,$FC,$0B,$F8,$73,$FC,$11,$B9
	.dc.b	$9B,$F3,$BF,$B0,$E6,$FF,$07,$4F,$FB,$37,$E7,$3E,$34,$FF,$A8,$AF
	.dc.b	$F3,$33,$97,$F0,$67,$F1,$08,$D3,$F9,$EF,$CA,$3C,$29,$A7,$8B,$D9
	.dc.b	$FF,$4F,$F2,$FF,$8D,$A3,$F7,$C4,$C3,$FC,$84,$FE,$46,$35,$D5,$FE
	.dc.b	$21,$8F,$C8,$DF,$78,$C1,$F7,$7F,$72,$FE,$C5,$1F,$69,$AF,$F9,$AD
	.dc.b	$B5,$FE,$46,$F3,$BF,$80,$E9,$FB,$66,$D7,$F8,$C5,$AF,$F2,$DE,$0F
	.dc.b	$F3,$95,$FF,$98,$AF,$FC,$F3,$F4,$5F,$90,$6F,$3F,$F9,$DE,$31,$8F
	.dc.b	$5F,$3F,$FC,$BD,$97,$F3,$6C,$BF,$B6,$B5,$FE,$4A,$DC,$3A,$FB,$1F
	.dc.b	$E8,$1F,$C0,$3E,$FC,$7F,$A2,$1B,$87,$DA,$6D,$FF,$44,$FC,$83,$FB
	.dc.b	$74,$FD,$E6,$C6,$7D,$EF,$F4,$64,$63,$91,$F8,$1F,$F8,$41,$B7,$FA
	.dc.b	$98,$FF,$36,$DF,$E1,$1A,$FE,$29,$AF,$F9,$67,$E7,$7F,$1C,$DF,$4F
	.dc.b	$FC,$FD,$18,$87,$DA,$7C,$E2,$71,$FF,$A0,$DC,$7E,$05,$7F,$1B,$FF
	.dc.b	$15,$57,$E4,$A5,$18,$8F,$E2,$F1,$8D,$FF,$CD,$6D,$8E,$1D,$AF,$9B
	.dc.b	$CB,$63,$31,$83,$FC,$E0,$FF,$C0,$03,$FC,$E5,$F4,$98,$23,$FE,$78
	.dc.b	$FC,$03,$E0,$A6,$93,$7F,$F3,$D7,$CD,$68,$8E,$0F,$D6,$F0,$BF,$04
	.dc.b	$FF,$09,$46,$1B,$7F,$CD,$27,$5F,$03,$F4,$9B,$5B,$FA,$2C,$59,$A4
	.dc.b	$8E,$9E,$16,$88,$E0,$F1,$FE,$08,$E2,$EC,$D8,$CC,$60,$FC,$0F,$0F
	.dc.b	$F1,$A8,$FF,$4D,$B7,$3F,$E1,$2D,$7F,$A6,$B5,$FE,$27,$F5,$7F,$EF
	.dc.b	$49,$F8,$6D,$8F,$80,$D7,$F8,$AB,$5F,$E2,$FA,$FE,$4B,$4F,$F6,$55
	.dc.b	$7F,$41,$8C,$19,$F9,$59,$1F,$D1,$5B,$5F,$4B,$E5,$79,$12,$7F,$A5
	.dc.b	$3F,$3F,$F0,$4F,$F7,$3C,$7D,$86,$CF,$D5,$6B,$FC,$A9,$AF,$F2,$CF
	.dc.b	$92,$F3,$12,$7F,$AB,$3F,$6F,$E9,$1F,$83,$52,$4F,$E1,$A3,$2F,$FB
	.dc.b	$23,$D4,$FB,$CD,$E2,$FD,$9C,$63,$0F,$E7,$2D,$7F,$BE,$FE,$09,$F9
	.dc.b	$7A,$FF,$68,$B6,$3F,$9F,$B5,$DF,$37,$EF,$A3,$FD,$D9,$3C,$3E,$7F
	.dc.b	$04,$7F,$37,$C4,$FB,$3F,$CB,$36,$33,$18,$3A,$7C,$AC,$8F,$1B,$71
	.dc.b	$FD,$04,$7F,$97,$8F,$D3,$95,$CF,$91,$FA,$38,$C4,$7F,$86,$BF,$03
	.dc.b	$7D,$1F,$F1,$66,$CF,$FB,$C1,$B3,$FD,$79,$AF,$1E,$4E,$04,$9E,$0F
	.dc.b	$3F,$1E,$47,$36,$CF,$ED,$FE,$0B,$FC,$E7,$3F,$BE,$6C,$67,$DD,$FD
	.dc.b	$D4,$62,$3F,$C6,$3E,$03,$E0,$3F,$96,$B6,$FF,$B9,$CA,$30,$FF,$76
	.dc.b	$6B,$FD,$EB,$F0,$1F,$D3,$3F,$E8,$86,$F6,$1A,$6A,$1F,$63,$FA,$A7
	.dc.b	$FC,$02,$D7,$FC,$8A,$D7,$FC,$BE,$D4,$72,$FA,$4F,$E8,$B0,$69,$30
	.dc.b	$F5,$C3,$E7,$B6,$48,$CB,$FC,$7F,$E0,$FF,$C6,$51,$FC,$A1,$B7,$FC
	.dc.b	$95,$AF,$F8,$69,$AF,$F9,$D3,$F0,$AF,$F2,$5C,$FC,$A6,$C7,$F8,$79
	.dc.b	$AF,$F9,$89,$AF,$F8,$A3,$CA,$79,$BF,$CC,$5B,$5F,$CE,$9B,$3F,$CD
	.dc.b	$DA,$FF,$9B,$3E,$FD,$F4,$1F,$DA,$1B,$5F,$71,$B3,$FE,$30,$6B,$C7
	.dc.b	$C2,$E0,$49,$E0,$F3,$E9,$91,$CD,$B3,$CF,$97,$C0,$49,$F0,$3F,$DE
	.dc.b	$B9,$11,$48,$7E,$41,$CB,$7E,$7B,$81,$1C,$1B,$F9,$DE,$84,$70,$73
	.dc.b	$F6,$7C,$04,$70,$77,$F5,$FC,$84,$70,$78,$FE,$03,$D0,$47,$07,$9F
	.dc.b	$E7,$3D,$84,$70,$7A,$FA,$BE,$01,$1C,$1E,$FF,$8A,$F8,$84,$70,$7C
	.dc.b	$3F,$9A,$F9,$04,$70,$7C,$7F,$9A,$F9,$84,$70,$7C,$BF,$67,$E8,$11
	.dc.b	$C1,$F3,$FE,$0B,$EA,$11,$C1,$F4,$FE,$07,$EC,$11,$C1,$F5,$FE,$F7
	.dc.b	$EE,$11,$C1,$F6,$FE,$A3,$F0,$04,$70,$7D,$FF,$53,$F0,$84,$70,$7E
	.dc.b	$0F,$E6,$3F,$10,$47,$07,$E1,$FE,$27,$F1,$84,$70,$7E,$2F,$C6,$FC
	.dc.b	$81,$1C,$1F,$8F,$F8,$FF,$CA,$11,$C1,$FC,$8B,$95,$F1,$DF,$E5,$26
	.dc.b	$C6,$63,$06,$7E,$62,$47,$16,$F5,$FF,$A7,$AB,$FD,$6F,$3D,$7E,$0B
	.dc.b	$A1,$1F,$E6,$3F,$82,$7E,$F2,$BF,$DF,$ED,$7F,$90,$B5,$39,$F8,$17
	.dc.b	$DE,$39,$3F,$CD,$71,$86,$C6,$6F,$FD,$63,$39,$7F,$D6,$79,$BE,$F1
	.dc.b	$F0,$3F,$E7,$48,$C3,$63,$33,$C1,$B9,$7F,$92,$F9,$9F,$2B,$FE,$4B
	.dc.b	$6D,$7F,$95,$1B,$3B,$9F,$F2,$87,$D2,$B9,$69,$2F,$FC,$B7,$19,$7C
	.dc.b	$FF,$79,$EB,$BF,$F7,$B3,$5F,$E8,$E7,$DA,$70,$FF,$4A,$CD,$18,$8F
	.dc.b	$E9,$BF,$72,$FA,$EF,$F4,$8B,$6F,$FB,$2B,$67,$FD,$B2,$D7,$BE,$CF
	.dc.b	$BC,$FF,$0A,$47,$88,$FF,$43,$38,$76,$D7,$FA,$EC,$71,$F6,$BF,$D4
	.dc.b	$7E,$FF,$FB,$97,$0F,$F1,$7F,$AD,$F3,$AF,$DB,$39,$7B,$35,$CC,$7F
	.dc.b	$86,$1F,$ED,$30,$8C,$42,$31,$ED,$F8,$27,$3A,$BA,$9E,$67,$E0,$DB
	.dc.b	$8F,$E2,$1B,$8F,$BE,$B8,$57,$E2,$7C,$CF,$F0,$94,$FD,$16,$BF,$AE
	.dc.b	$6C,$9C,$C3,$67,$C2,$7D,$42,$63,$F5,$1F,$AF,$EA,$C8,$FA,$ED,$AF
	.dc.b	$AE,$FF,$52,$7F,$C3,$DD,$9F,$86,$BF,$EC,$BE,$CA,$31,$5B,$8C,$57
	.dc.b	$E1,$BD,$2C,$57,$1C,$AC,$FE,$29,$AF,$A7,$4B,$8F,$B7,$2F,$F0,$DE
	.dc.b	$3D,$2B,$DD,$7E,$59,$FC,$45,$6E,$7F,$DC,$ED,$8C,$CE,$E7,$C5,$1D
	.dc.b	$30,$FD,$0F,$DC,$FF,$6C,$9F,$79,$E5,$9D,$3F,$E2,$C8,$D6,$1E,$37
	.dc.b	$84,$FE,$65,$B3,$D2,$3C,$0F,$16,$EA,$F9,$1F,$F4,$35,$FC,$88,$C7
	.dc.b	$C1,$FC,$67,$5D,$D6,$B7,$97,$ED,$5F,$F4,$76,$1F,$BA,$AE,$2F,$F9
	.dc.b	$08,$C4,$6B,$1E,$18,$78,$A3,$9C,$7E,$77,$D1,$FE,$10,$FC,$2B,$E6
	.dc.b	$FC,$4E,$54,$EB,$3F,$85,$E1,$FB,$A9,$AA,$30,$DC,$FC,$93,$F4,$27
	.dc.b	$37,$FE,$60,$8F,$A2,$72,$71,$7D,$AB,$E9,$CE,$7D,$AB,$87,$91,$FF
	.dc.b	$3A,$B6,$3F,$86,$1B,$2B,$97,$EC,$7A,$5F,$D6,$9C,$B5,$CE,$7D,$23
	.dc.b	$11,$F8,$87,$86,$1F,$83,$6E,$D9,$AF,$2F,$A5,$FC,$BA,$BF,$10,$FF
	.dc.b	$30,$CB,$F1,$0D,$C3,$F2,$6D,$7F,$20,$DA,$7D,$D7,$F5,$09,$FC,$AE
	.dc.b	$CF,$A1,$FF,$57,$C6,$1A,$FD,$B9,$E0,$DE,$2E,$91,$8C,$7F,$14,$D9
	.dc.b	$FE,$84,$D7,$FB,$FB,$CB,$2F,$CC,$37,$3D,$DF,$BD,$E3,$C5,$6F,$F5
	.dc.b	$5E,$9F,$B5,$FC,$3B,$77,$FF,$58,$47,$E1,$8D,$94,$FC,$6F,$A7,$CD
	.dc.b	$E2,$7E,$0A,$71,$29,$A7,$52,$4F,$38,$D3,$63,$6F,$E3,$1E,$7E,$9F
	.dc.b	$4F,$1F,$AF,$95,$2B,$F2,$31,$97,$F2,$2F,$FB,$19,$FD,$1B,$6C,$DF
	.dc.b	$4A,$4F,$E6,$1A,$FF,$45,$F9,$BF,$EF,$CF,$F8,$41,$AE,$7C,$9F,$F6
	.dc.b	$9F,$22,$73,$3F,$C0,$FF,$3D,$FC,$0B,$7E,$0F,$71,$88,$FE,$FC,$D7
	.dc.b	$FC,$77,$F7,$6F,$6F,$AB,$63,$FC,$42,$D7,$FA,$D3,$5F,$E2,$4F,$F6
	.dc.b	$27,$7E,$0F,$22,$3F,$E2,$CF,$BA,$FF,$19,$37,$7F,$F7,$6F,$94,$9C
	.dc.b	$CF,$F8,$16,$31,$3F,$37,$E8,$7F,$67,$3F,$E1,$A6,$BF,$EE,$26,$A7
	.dc.b	$3A,$BE,$E9,$F4,$BF,$29,$B5,$FE,$3E,$6B,$FD,$DD,$AF,$F8,$A3,$CB
	.dc.b	$7E,$C1,$E6,$8C,$35,$FF,$79,$C6,$B1,$FE,$27,$1F,$E0,$D9,$D3,$FE
	.dc.b	$0F,$FB,$B7,$56,$AD,$18,$9F,$E3,$71,$89,$D4,$E6,$7F,$C1,$9F,$1F
	.dc.b	$FE,$DE,$7F,$CA,$ED,$73,$E5,$FE,$22,$31,$1F,$E3,$4F,$B9,$7D,$D7
	.dc.b	$D2,$6C,$7F,$98,$9A,$FD,$36,$BF,$E3,$9D,$DF,$53,$F5,$49,$1F,$02
	.dc.b	$32,$FC,$2F,$B4,$7C,$B7,$E9,$DB,$19,$8C,$19,$FA,$69,$1C,$5A,$FF
	.dc.b	$3E,$9F,$F2,$A3,$D7,$D8,$E8,$47,$C1,$D7,$DC,$8E,$0F,$27,$C9,$F8
	.dc.b	$C8,$E0,$F4,$7C,$7F,$9C,$8E,$0F,$67,$DC,$FA,$C8,$E0,$F8,$1F,$71
	.dc.b	$EF,$23,$83,$E2,$7D,$C7,$C3,$23,$83,$E4,$7C,$A7,$C7,$23,$83,$E6
	.dc.b	$7B,$5F,$2C,$8E,$0F,$A1,$ED,$7C,$F2,$38,$3E,$A7,$2F,$D3,$23,$83
	.dc.b	$EC,$78,$5F,$5C,$8E,$0F,$B9,$E0,$FD,$B2,$38,$3F,$01,$EC,$FD,$F2
	.dc.b	$38,$3F,$4B,$E9,$7F,$94,$5F,$EF,$2D,$7F,$D0,$CD,$7F,$D3,$2D,$47
	.dc.b	$F6,$3F,$8D,$FE,$65,$06,$92,$3B,$FB,$3A,$23,$83,$C7,$EC,$F2,$23
	.dc.b	$83,$CF,$EC,$76,$23,$83,$D7,$EC,$78,$88,$E0,$F7,$FC,$6F,$98,$8E
	.dc.b	$0F,$87,$E9,$FA,$88,$E0,$F8,$FE,$9F,$B8,$8E,$0F,$97,$E9,$FC,$22
	.dc.b	$38,$3E,$7F,$C6,$FC,$62,$38,$3E,$9F,$91,$F2,$88,$E0,$FE,$27,$F4
	.dc.b	$8F,$80,$FE,$30,$D8,$CC,$60,$FE,$C9,$F5,$ED,$69,$27,$F9,$2C,$65
	.dc.b	$FE,$3E,$BE,$2F,$F9,$23,$63,$31,$83,$AF,$AF,$23,$9B,$5F,$F2,$1C
	.dc.b	$BE,$33,$C6,$48,$F0,$BF,$DE,$7E,$BF,$FD,$21,$1F,$ED,$56,$BF,$EE
	.dc.b	$86,$BF,$90,$6F,$9D,$FD,$3F,$C9,$FE,$75,$E4,$FF,$20,$AF,$F6,$F4
	.dc.b	$E5,$FE,$77,$19,$7F,$DC,$5F,$2D,$F9,$B4,$D2,$7C,$07,$F4,$5F,$AE
	.dc.b	$7D,$BD,$25,$7F,$46,$8C,$BF,$D4,$2B,$FC,$3E,$FF,$99,$1A,$F8,$39
	.dc.b	$7F,$C3,$11,$88,$FD,$59,$E2,$F9,$3C,$91,$E8,$F0,$F8,$89,$1D,$DA
	.dc.b	$FC,$0F,$AF,$75,$FF,$5D,$6C,$7F,$5C,$6B,$B8,$FF,$15,$7D,$73,$AA
	.dc.b	$92,$7F,$B5,$C6,$5F,$EC,$A3,$D0,$FE,$D4,$D3,$2E,$0E,$BC,$D2,$39
	.dc.b	$B5,$FF,$44,$DB,$E8,$BF,$C7,$8D,$8F,$EF,$6D,$77,$1A,$F9,$09,$1A
	.dc.b	$6B,$F8,$2F,$86,$FF,$00,$F2,$1A,$4E,$DF,$F2,$57,$D6,$3B,$E9,$2F
	.dc.b	$FC,$17,$19,$7F,$C0,$BF,$5B,$FE,$42,$7F,$C2,$2D,$58,$23,$0D,$F8
	.dc.b	$5E,$FF,$0D,$2B,$AB,$5F,$F5,$5D,$B9,$3C,$A4,$8F,$1B,$FE,$61,$F0
	.dc.b	$BF,$4B,$06,$93,$E2,$3F,$C3,$DF,$5B,$FE,$0A,$9F,$F1,$3B,$5F,$3B
	.dc.b	$72,$39,$8E,$9E,$27,$72,$3F,$E2,$EB,$FE,$24,$FE,$C3,$E3,$BF,$86
	.dc.b	$37,$C5,$FF,$17,$46,$23,$FC,$78,$D7,$FC,$73,$F5,$8F,$37,$FC,$68
	.dc.b	$D8,$FF,$19,$B5,$FD,$EB,$5F,$F2,$37,$D6,$7F,$55,$CC,$D2,$78,$5F
	.dc.b	$E3,$F7,$FA,$B4,$FF,$8F,$9A,$FF,$08,$6B,$D3,$CD,$FF,$47,$7C,$F7
	.dc.b	$CE,$26,$F8,$33,$D3,$2B,$FC,$A3,$3F,$97,$7F,$7A,$1C,$3F,$C0,$17
	.dc.b	$D6,$3E,$93,$8B,$67,$FD,$DB,$3A,$9C,$9D,$BE,$B1,$F1,$74,$95,$C2
	.dc.b	$32,$FF,$A9,$FE,$4B,$FD,$50,$9A,$47,$FC,$D3,$F2,$AD,$BF,$BF,$E1
	.dc.b	$FE,$05,$8C,$4F,$9E,$7F,$8C,$FD,$13,$D2,$CD,$EE,$78,$9F,$0A,$F2
	.dc.b	$FC,$4F,$A7,$B8,$7F,$22,$9A,$FF,$83,$A3,$3C,$4F,$AB,$FF,$30,$DF
	.dc.b	$F9,$D9,$B5,$FD,$01,$AF,$FA,$25,$AF,$FA,$19,$7D,$67,$46,$CF,$9A
	.dc.b	$7F,$73,$1D,$31,$F7,$DF,$86,$F8,$1C,$7E,$36,$E1,$C6,$7C,$D1,$EC
	.dc.b	$F9,$97,$DC,$BF,$0F,$F5,$FD,$73,$9F,$53,$9C,$79,$4E,$5B,$3C,$39
	.dc.b	$F8,$B6,$3D,$18,$3E,$87,$FC,$65,$1E,$3F,$15,$E5,$78,$E7,$3C,$3F
	.dc.b	$64,$FE,$2F,$3A,$AE,$15,$F4,$23,$FD,$73,$18,$3A,$7C,$AF,$D6,$DD
	.dc.b	$F0,$FF,$B1,$24,$FF,$AB,$5F,$DA,$7C,$3E,$CE,$0D,$8F,$91,$F5,$DF
	.dc.b	$DC,$9B,$EB,$7C,$9F,$CA,$FE,$26,$9E,$7C,$57,$B8,$C3,$FE,$AC,$BF
	.dc.b	$F5,$CB,$63,$FC,$06,$D7,$FD,$98,$D7,$FD,$9D,$F0,$DF,$8E,$F9,$47
	.dc.b	$E1,$1E,$D8,$C3,$F9,$4F,$A0,$FE,$95,$87,$E1,$ED,$18,$7E,$67,$17
	.dc.b	$F1,$BF,$91,$E1,$EA,$FE,$73,$EB,$DE,$07,$F9,$AD,$B8,$FF,$B6,$1B
	.dc.b	$87,$F0,$80,$E1,$FF,$03,$FD,$6B,$E7,$3E,$2F,$81,$FE,$7B,$A6,$5F
	.dc.b	$AC,$AD,$77,$78,$63,$F5,$7F,$84,$BE,$06,$9B,$FB,$07,$5E,$35,$FD
	.dc.b	$26,$9F,$C5,$7E,$57,$FD,$08,$D8,$F3,$37,$6F,$E3,$ED,$7C,$FF,$80
	.dc.b	$F4,$9E,$63,$E5,$7C,$D1,$99,$FC,$A5,$7F,$11,$9F,$E0,$78,$BE,$62
	.dc.b	$9F,$C1,$3E,$A3,$FD,$27,$8F,$D3,$9D,$37,$BB,$FD,$31,$1F,$96,$8F
	.dc.b	$99,$8A,$30,$FC,$F6,$BF,$E3,$4E,$C5,$BE,$DB,$C7,$F2,$BE,$16,$E1
	.dc.b	$F1,$BE,$B3,$4D,$8F,$A5,$F6,$3E,$DE,$8F,$E5,$3B,$AF,$79,$FC,$0F
	.dc.b	$1D,$E8,$7C,$E6,$E5,$F3,$9B,$3F,$39,$AF,$E1,$3A,$FF,$C1,$13,$F6
	.dc.b	$3F,$09,$7A,$1F,$83,$9C,$D7,$CE,$9C,$D7,$E6,$FB,$EC,$FF,$05,$0E
	.dc.b	$34,$FE,$83,$DA,$FB,$3F,$4C,$7F,$B2,$E7,$57,$FE,$5C,$5F,$3D,$FE
	.dc.b	$C9,$6C,$7D,$26,$E1,$F9,$F6,$BF,$9E,$C5,$F4,$1F,$D1,$25,$F3,$DF
	.dc.b	$9D,$9C,$DF,$A6,$3F,$A5,$F0,$BF,$09,$FD,$34,$FE,$69,$B8,$FF,$21
	.dc.b	$D5,$F9,$33,$CD,$E3,$5A,$DB,$EB,$FE,$15,$FE,$65,$C7,$F3,$ED,$C7
	.dc.b	$F3,$ED,$7F,$CF,$79,$23,$11,$FE,$E8,$CB,$FD,$69,$B7,$9B,$EE,$BF
	.dc.b	$09,$86,$7E,$13,$D7,$7F,$A2,$9C,$BF,$EF,$0F,$CB,$DF,$1D,$FD,$99
	.dc.b	$B7,$FC,$4D,$7C,$EE,$67,$FB,$5F,$5B,$E8,$BF,$DF,$4D,$BF,$D1,$B7
	.dc.b	$0F,$E2,$0D,$7F,$C3,$2B,$D0,$E6,$74,$AB,$5F,$DA,$75,$7C,$F7,$F5
	.dc.b	$9C,$3F,$77,$F6,$8F,$B7,$FE,$29,$39,$E3,$FD,$1C,$E7,$0F,$E1,$ED
	.dc.b	$BF,$C9,$67,$FE,$45,$9C,$E1,$EC,$F8,$AF,$85,$F1,$FF,$E0,$8F,$C1
	.dc.b	$1F,$37,$8B,$FE,$6D,$18,$6C,$7F,$83,$BE,$E7,$FB,$1E,$9F,$E1,$CC
	.dc.b	$D7,$B7,$F4,$53,$AB,$47,$F3,$1F,$D3,$BF,$9B,$DF,$EF,$3D,$FF,$E9
	.dc.b	$35,$FC,$A2,$72,$F0,$EA,$FE,$E7,$ED,$7E,$5A,$B7,$87,$6F,$A5,$C4
	.dc.b	$F9,$5F,$37,$54,$C3,$FA,$86,$69,$C5,$FE,$DF,$F0,$B7,$5F,$F8,$A9
	.dc.b	$B7,$FD,$5F,$EA,$5E,$6E,$6B,$E1,$7D,$3C,$AF,$EB,$97,$C3,$0F,$83
	.dc.b	$3F,$AC,$FA,$EF,$E5,$F1,$FC,$D9,$BC,$3F,$C1,$3E,$42,$E1,$EF,$9B
	.dc.b	$5F,$23,$5F,$EA,$13,$EA,$F6,$BA,$4F,$D3,$7F,$39,$F3,$D2,$7F,$B7
	.dc.b	$7E,$38,$F8,$CC,$5F,$D4,$F9,$EF,$B0,$FF,$BA,$B1,$FF,$5D,$5F,$F3
	.dc.b	$E6,$C7,$F6,$B6,$BF,$E7,$47,$F5,$2F,$DE,$F0,$FF,$09,$5F,$85,$7F
	.dc.b	$B5,$7E,$7B,$FD,$B2,$FF,$B1,$FB,$4E,$6C,$FC,$0F,$F1,$08,$C4,$EA
	.dc.b	$7F,$1D,$F3,$D9,$6C,$7C,$EA,$FA,$C7,$F9,$83,$F0,$8F,$E4,$7C,$7F
	.dc.b	$93,$E5,$F0,$6B,$F9,$2F,$0B,$F3,$07,$0C,$3F,$06,$FF,$20,$FD,$CF
	.dc.b	$FE,$00,$C7,$F7,$86,$CF,$F8,$FF,$E2,$3F,$9C,$7C,$A6,$FE,$8B,$F4
	.dc.b	$5F,$7F,$F2,$AB,$EB,$35,$72,$23,$0D,$C7,$E7,$7D,$6F,$EA,$CE,$6F
	.dc.b	$E5,$9F,$F3,$DC,$FF,$9A,$E3,$F5,$35,$CF,$EB,$3F,$C2,$DF,$82,$7E
	.dc.b	$FF,$7F,$D9,$5E,$EB,$F1,$1E,$47,$ED,$CE,$77,$70,$FF,$0A,$FE,$1B
	.dc.b	$FD,$93,$0F,$ED,$AD,$8F,$E6,$1E,$AF,$F3,$79,$CF,$C0,$FF,$A3,$BE
	.dc.b	$C1,$F8,$0D,$3F,$A4,$E1,$C7,$3F,$DF,$CE,$B1,$FF,$43,$7C,$FF,$FB
	.dc.b	$FB,$4F,$90,$D8,$FF,$1C,$36,$BF,$D8,$2D,$7F,$D8,$7F,$0B,$D5,$FF
	.dc.b	$01,$65,$FE,$A2,$9E,$19,$FF,$13,$AF,$17,$AB,$FD,$2F,$C4,$FE,$BB
	.dc.b	$F3,$1F,$E9,$8C,$7F,$AD,$6F,$FA,$C7,$FC,$53,$19,$C3,$FC,$1E,$DC
	.dc.b	$FF,$89,$3F,$E5,$78,$CF,$CE,$74,$F9,$4F,$ED,$3E,$3F,$F0,$6B,$FF
	.dc.b	$1E,$B7,$0F,$F1,$EB,$5F,$EB,$ED,$7F,$92,$FC,$5B,$F0,$1F,$EB,$2D
	.dc.b	$F9,$2E,$BE,$7F,$FA,$F2,$7F,$AE,$F0,$FF,$9A,$BC,$3F,$97,$DF,$FA
	.dc.b	$0C,$7F,$0F,$39,$CF,$E6,$CE,$67,$FD,$91,$F1,$FF,$EC,$8F,$8A,$FF
	.dc.b	$20,$C7,$FA,$76,$F8,$3F,$E1,$B8,$CF,$07,$F8,$6B,$8B,$FD,$B1,$8F
	.dc.b	$FB,$07,$16,$DF,$1F,$07,$27,$F7,$BF,$07,$FD,$AF,$5A,$E3,$B7,$B8
	.dc.b	$E3,$C7,$89,$CB,$C3,$E4,$7F,$DF,$F9,$7F,$9F,$69,$18,$CB,$F4,$A7
	.dc.b	$FA,$31,$C3,$FA,$88,$DF,$D0,$FF,$9C,$BA,$BE,$A6,$DF,$E4,$C6,$E1
	.dc.b	$FD,$5A,$B6,$FF,$9E,$A3,$37,$FD,$6B,$D9,$7E,$05,$FE,$F0,$6D,$7F
	.dc.b	$A7,$7E,$47,$FB,$85,$7F,$90,$E7,$2F,$CB,$F8,$AF,$E0,$DC,$7F,$E3
	.dc.b	$4B,$FF,$36,$B6,$7F,$D3,$4D,$7D,$AB,$D8,$FB,$07,$D9,$AB,$5F,$EA
	.dc.b	$D9,$FC,$87,$D0,$FF,$B9,$EF,$FB,$8C,$63,$E4,$7F,$85,$F0,$FF,$3A
	.dc.b	$AD,$57,$FA,$BB,$AD,$87,$F7,$89,$CE,$1E,$C6,$FC,$0E,$AD,$1A,$BC
	.dc.b	$3A,$BE,$58,$FF,$1A,$C3,$93,$FC,$73,$EF,$FB,$CE,$B0,$F3,$E3,$CA
	.dc.b	$B5,$36,$7F,$C1,$75,$C1,$F0,$CE,$A7,$94,$7F,$B2,$2B,$82,$AE,$18
	.dc.b	$7C,$5F,$80,$7D,$3F,$FB,$4E,$30,$FF,$69,$C3,$C9,$3A,$D3,$C4,$F0
	.dc.b	$F6,$6D,$6F,$1F,$19,$FE,$CF,$71,$87,$4B,$DD,$4C,$3E,$07,$27,$F8
	.dc.b	$57,$C1,$C6,$3C,$A3,$13,$EC,$CD,$E3,$FB,$1B,$19,$FA,$5C,$4C,$8E
	.dc.b	$1F,$01,$FE,$94,$C3,$CA,$7C,$39,$E5,$19,$D3,$F0,$78,$7F,$89,$FE
	.dc.b	$47,$FD,$82,$DB,$FF,$77,$35,$FD,$8B,$5F,$B9,$EB,$FF,$BB,$DD,$B5
	.dc.b	$DB,$5D,$F0,$76,$F4,$FF,$AF,$E1,$E2,$AE,$4D,$9F,$85,$3A,$E3,$FC
	.dc.b	$53,$6F,$F0,$AD,$7F,$00,$DC,$1F,$01,$BB,$7E,$00,$E3,$E5,$C2,$FB
	.dc.b	$67,$93,$86,$E7,$F8,$4B,$E0,$DB,$E9,$18,$8F,$C1,$E1,$DA,$73,$CB
	.dc.b	$27,$81,$0D,$D7,$F8,$E5,$FC,$07,$8B,$E9,$23,$F8,$3F,$37,$D8,$61
	.dc.b	$FB,$BA,$A7,$31,$B2,$1A,$9F,$89,$F6,$32,$3F,$1C,$D7,$F0,$FF,$3E
	.dc.b	$FB,$AF,$95,$F2,$1F,$E1,$4F,$1B,$FD,$05,$97,$FA,$AA,$FF,$CF,$F1
	.dc.b	$89,$D7,$57,$EF,$FD,$FB,$5F,$BA,$7F,$9C,$B9,$3F,$1E,$E1,$D1,$7C
	.dc.b	$47,$95,$CA,$11,$87,$C3,$5F,$E2,$88,$C4,$7F,$BD,$78,$5F,$17,$E2
	.dc.b	$E3,$7D,$FF,$DC,$A6,$9F,$C7,$67,$3C,$DB,$AD,$4F,$F2,$1E,$7F,$E4
	.dc.b	$3E,$2F,$B1,$B8,$7F,$57,$EC,$FF,$44,$B6,$7D,$93,$A8,$FE,$B3,$8B
	.dc.b	$EF,$DE,$2F,$95,$E3,$3E,$4D,$FF,$D4,$91,$8C,$BC,$CF,$09,$FF,$66
	.dc.b	$7A,$1F,$9E,$7E,$18,$C3,$B6,$1D,$4F,$A7,$E8,$7F,$BC,$7B,$BE,$67
	.dc.b	$DC,$7E,$DF,$2F,$F5,$F3,$73,$FF,$29,$B7,$17,$A4,$FB,$6F,$CC,$47
	.dc.b	$F8,$2D,$AF,$F8,$51,$AF,$FB,$12,$38,$78,$1F,$7D,$B1,$F8,$4F,$4D
	.dc.b	$F7,$BF,$EF,$A8,$F5,$C6,$3E,$23,$F1,$6D,$8F,$7E,$FF,$EA,$1F,$4B
	.dc.b	$EF,$63,$FD,$E2,$7F,$1F,$39,$6C,$65,$FA,$6D,$7F,$CE,$30,$F0,$3F
	.dc.b	$DF,$91,$D9,$7C,$A7,$E2,$DB,$19,$F9,$5F,$9B,$C8,$F2,$35,$12,$7E
	.dc.b	$BC,$7D,$0F,$80,$91,$F6,$07,$0A,$3F,$53,$C0,$FA,$0F,$BD,$5F,$E3
	.dc.b	$96,$BF,$E4,$26,$BF,$3D,$AF,$D8,$6A,$FB,$C7,$DB,$FE,$B5,$B1,$E3
	.dc.b	$6B,$F9,$33,$F1,$9F,$2A,$FA,$6F,$D6,$36,$3E,$9B,$5E,$1C,$08,$C7
	.dc.b	$47,$F9,$37,$D0,$F0,$78,$52,$FC,$0F,$EF,$FE,$35,$F4,$74,$93,$F1
	.dc.b	$E3,$2F,$E8,$3B,$9F,$12,$F9,$1F,$67,$D2,$CA,$F6,$3E,$FF,$BC,$91
	.dc.b	$FC,$47,$81,$1F,$C4,$E3,$E3,$36,$33,$1F,$8C,$FD,$D3,$B3,$F9,$26
	.dc.b	$BF,$01,$BB,$7F,$09,$FA,$57,$B9,$F3,$FE,$52,$5F,$CF,$7F,$8F,$7C
	.dc.b	$B7,$F1,$AB,$FF,$18,$4A,$31,$E3,$7F,$B1,$9B,$03,$5E,$8F,$F2,$7F
	.dc.b	$B0,$7D,$77,$F2,$1F,$53,$F8,$45,$7D,$B3,$F0,$5F,$09,$F8,$13,$5E
	.dc.b	$6F,$F0,$6E,$47,$C1,$62,$BF,$5F,$7C,$B5,$FF,$22,$73,$7D,$0A,$FC
	.dc.b	$57,$8F,$2F,$F1,$4F,$E3,$7C,$59,$1F,$3D,$F9,$84,$7F,$A1,$7E,$7F
	.dc.b	$FE,$19,$7C,$CD,$C3,$E1,$B7,$87,$FD,$CF,$D5,$F9,$69,$F8,$FC,$FE
	.dc.b	$59,$FF,$75,$B6,$FE,$33,$7B,$FF,$D6,$5F,$25,$FD,$17,$E3,$7F,$DB
	.dc.b	$F5,$FC,$31,$AF,$E6,$9B,$87,$C0,$FB,$6F,$E8,$B8,$FF,$BE,$3E,$0B
	.dc.b	$EE,$DF,$EE,$DA,$FC,$F6,$D7,$DC,$FC,$2F,$FC,$FD,$3F,$C0,$5A,$FD
	.dc.b	$56,$C6,$FE,$2B,$F4,$9F,$3D,$FD,$8E,$DF,$11,$49,$87,$87,$C6,$FE
	.dc.b	$C1,$FA,$63,$E3,$F8,$44,$9F,$E5,$33,$97,$F9,$5F,$D6,$FF,$7D,$7F
	.dc.b	$73,$19,$6D,$7F,$6B,$9D,$67,$F9,$6E,$37,$C9,$79,$7D,$B9,$87,$ED
	.dc.b	$E3,$FB,$D7,$E1,$DF,$9C,$BE,$11,$FB,$F3,$3C,$3F,$DD,$FF,$0D,$78
	.dc.b	$BE,$FF,$8A,$98,$FF,$40,$9C,$BF,$C0,$7F,$1F,$FF,$0C,$39,$AE,$47
	.dc.b	$F6,$08,$CC,$7D,$3F,$37,$E5,$11,$FB,$1E,$EF,$CD,$3F,$81,$72,$7F
	.dc.b	$53,$8F,$F0,$CB,$5F,$F6,$9B,$5F,$D9,$0F,$9B,$7D,$E7,$F8,$65,$B1
	.dc.b	$FC,$FD,$AF,$EC,$F8,$7F,$16,$FF,$BB,$3F,$1F,$79,$1F,$E1,$9F,$C3
	.dc.b	$BF,$A6,$4F,$F6,$06,$BF,$CD,$7C,$65,$70,$8F,$EC,$BE,$43,$E6,$3F
	.dc.b	$C5,$DE,$1F,$F2,$66,$C6,$67,$FB,$3B,$5F,$F0,$BF,$E1,$DF,$E2,$A7
	.dc.b	$FB,$7B,$5F,$E7,$AD,$E7,$FE,$07,$F6,$9F,$C8,$39,$FF,$8A,$67,$F9
	.dc.b	$6F,$89,$FF,$1D,$7D,$6B,$CD,$A7,$F8,$53,$C2,$FF,$0F,$5F,$CA,$97
	.dc.b	$CE,$FF,$33,$6E,$9F,$E4,$76,$BF,$E4,$C6,$BF,$48,$7E,$0C,$FB,$BF
	.dc.b	$F2,$53,$63,$FB,$83,$5F,$95,$F0,$6F,$0F,$FC,$4F,$F8,$EE,$78,$3F
	.dc.b	$C8,$DD,$7F,$D6,$B4,$FE,$61,$CB,$FE,$89,$F8,$0D,$E9,$FE,$4F,$C9
	.dc.b	$F3,$DF,$41,$BA,$7F,$85,$1A,$FF,$91,$1A,$FF,$84,$BE,$9B,$B3,$FE
	.dc.b	$3B,$6E,$07,$3C,$8F,$43,$7F,$18,$FC,$13,$F5,$DA,$FE,$53,$EF,$5F
	.dc.b	$37,$FE,$6B,$6E,$7F,$E2,$EA,$DB,$FD,$2E,$1F,$55,$FD,$8D,$B1,$FE
	.dc.b	$77,$6B,$FE,$7C,$6B,$F4,$7F,$47,$FF,$24,$26,$9F,$E4,$BC,$7C,$DF
	.dc.b	$03,$FE,$54,$D1,$F6,$3F,$E7,$7F,$96,$77,$D3,$51,$FE,$7B,$C5,$EA
	.dc.b	$7E,$71,$A7,$F9,$00,$A7,$F9,$C0,$A7,$ED,$B1,$7E,$19,$FB,$06,$FC
	.dc.b	$57,$F9,$3A,$3F,$67,$97,$F9,$7F,$D5,$FF,$34,$F5,$FF,$84,$30,$FE
	.dc.b	$D5,$F8,$3F,$E9,$E3,$FB,$BB,$79,$3F,$43,$1B,$3F,$BD,$FC,$8B,$8B
	.dc.b	$FC,$1C,$DB,$FF,$51,$35,$FF,$53,$B5,$FC,$E3,$FD,$F1,$FF,$24,$35
	.dc.b	$FE,$F8,$D9,$57,$F8,$17,$E9,$3D,$FE,$67,$FA,$AF,$F0,$7F,$DB,$60
	.dc.b	$DF,$FD,$C9,$F4,$1F,$E6,$EF,$9F,$FE,$7B,$8F,$F1,$CB,$DD,$7E,$37
	.dc.b	$E1,$BF,$D7,$0D,$C3,$FC,$F5,$D5,$FE,$08,$6C,$EA,$3F,$92,$47,$E3
	.dc.b	$FD,$2F,$F3,$DC,$62,$3F,$C9,$DA,$3F,$19,$FF,$36,$C7,$F8,$5F,$1F
	.dc.b	$E1,$D9,$78,$FE,$83,$D4,$7A,$FB,$65,$7F,$B2,$EB,$F2,$6F,$A6,$3F
	.dc.b	$C7,$91,$88,$F8,$1F,$84,$3F,$0E,$FF,$46,$BF,$D1,$B8,$BE,$4C,$EB
	.dc.b	$E9,$73,$5F,$ED,$CD,$1F,$88,$FB,$EB,$FC,$CD,$C2,$C7,$A4,$7F,$A6
	.dc.b	$9A,$FB,$A7,$86,$3D,$A9,$F5,$7F,$CF,$23,$0F,$E6,$9B,$DB,$FE,$A2
	.dc.b	$8D,$6F,$F7,$78,$BF,$C8,$77,$CB,$9F,$47,$9E,$33,$F8,$4F,$F8,$4A
	.dc.b	$BF,$C9,$14,$8F,$E0,$2D,$8F,$F5,$7C,$6F,$3F,$C3,$36,$78,$78,$5A
	.dc.b	$37,$F5,$92,$BF,$D7,$94,$8F,$C3,$B6,$33,$1F,$E6,$FF,$B8,$FD,$06
	.dc.b	$BF,$E6,$07,$F9,$86,$3F,$DB,$1B,$5F,$E5,$F7,$71,$9E,$CF,$EE,$35
	.dc.b	$C5,$B1,$9A,$FF,$57,$7E,$BD,$FD,$16,$BB,$77,$9F,$E9,$13,$D6,$30
	.dc.b	$E3,$0F,$F1,$54,$EA,$11,$87,$5F,$85,$7F,$B9,$9B,$F4,$1B,$94,$E6
	.dc.b	$7F,$AB,$46,$27,$FD,$21,$0F,$A9,$A4,$7F,$D2,$8F,$FB,$9F,$E9,$BF
	.dc.b	$C8,$39,$FF,$B9,$9B,$3F,$92,$7E,$84,$67,$D9,$FC,$DE,$5F,$8F,$9D
	.dc.b	$BF,$91,$FC,$6B,$FD,$B1,$5B,$3F,$DE,$2E,$2F,$F3,$BF,$85,$FA,$A7
	.dc.b	$FD,$F0,$D8,$FE,$E1,$C6,$8C,$47,$F2,$F6,$BF,$90,$7F,$9F,$CF,$FB
	.dc.b	$F9,$B1,$DB,$5F,$3C,$EF,$CA,$B3,$FC,$5E,$B7,$8F,$EA,$7E,$F7,$FA
	.dc.b	$ED,$F3,$BD,$E2,$9C,$E1,$FB,$6F,$D1,$FF,$C6,$D7,$FE,$A9,$C3,$F6
	.dc.b	$51,$89,$D7,$8B,$CA,$31,$3A,$E8,$EB,$F9,$37,$F8,$FB,$0E,$47,$EB
	.dc.b	$2B,$6F,$EB,$A3,$58,$7F,$BA,$ED,$F8,$F3,$E2,$7D,$B1,$87,$D3,$97
	.dc.b	$F8,$67,$0E,$B7,$F9,$7B,$E1,$18,$D3,$FD,$6F,$D1,$C0,$E7,$3A,$AF
	.dc.b	$E2,$B7,$FD,$46,$F8,$7C,$FF,$BE,$7F,$B2,$35,$FE,$0A,$D7,$FA,$33
	.dc.b	$70,$2F,$BC,$F9,$ED,$AE,$BB,$62,$7E,$9F,$E2,$7F,$E3,$C9,$F4,$37
	.dc.b	$0F,$E5,$2F,$16,$FC,$1F,$DA,$FA,$DF,$C2,$C7,$F7,$96,$BF,$C6,$9A
	.dc.b	$FF,$5B,$6E,$47,$FD,$C7,$7F,$1F,$0F,$A7,$BF,$F3,$19,$CD,$FF,$3A
	.dc.b	$9C,$D7,$F2,$6F,$AE,$FD,$5C,$FF,$3E,$6D,$FE,$B3,$D4,$7A,$1F,$90
	.dc.b	$9C,$DF,$F2,$EF,$C3,$9E,$EF,$FB,$F3,$1C,$35,$FE,$E5,$1A,$9F,$E2
	.dc.b	$7A,$23,$13,$FA,$EF,$CE,$FC,$AF,$F7,$9F,$D9,$CF,$E8,$23,$34,$9C
	.dc.b	$DF,$E1,$FF,$0E,$FF,$02,$B7,$C8,$FD,$3C,$63,$0F,$EB,$6A,$3F,$63
	.dc.b	$97,$F2,$7F,$6F,$D0,$DB,$FE,$52,$DB,$FF,$39,$35,$FF,$52,$BF,$E1
	.dc.b	$4F,$98,$E7,$3F,$EC,$83,$F3,$47,$ED,$CF,$CF,$C6,$DB,$F7,$BF,$E4
	.dc.b	$06,$F9,$9E,$48,$C5,$FF,$3B,$9F,$F3,$AC,$7F,$19,$E2,$FE,$F7,$6F
	.dc.b	$DC,$9E,$BF,$A9,$B7,$EA,$6E,$1E,$A6,$BE,$AF,$DB,$BF,$31,$87,$46
	.dc.b	$D7,$F6,$CF,$DA,$DE,$E7,$C8,$9C,$DF,$E0,$BE,$2D,$F6,$BF,$31,$B8
	.dc.b	$7E,$EF,$E1,$F8,$BF,$0A,$33,$5E,$D3,$F1,$7F,$7E,$F7,$3E,$9B,$6B
	.dc.b	$E9,$B6,$7E,$9B,$5F,$A7,$F9,$F7,$E6,$27,$E7,$AE,$BF,$BE,$E6,$7C
	.dc.b	$1C,$28,$D5,$7D,$6F,$97,$FF,$6B,$B7,$E9,$3F,$0F,$18,$BE,$2A,$3F
	.dc.b	$17,$F8,$6F,$F4,$D9,$FC,$2B,$5F,$F5,$6B,$5F,$D4,$B7,$07,$E6,$8F
	.dc.b	$A4,$FC,$AB,$71,$FE,$49,$F4,$1F,$E9,$A7,$F1,$71,$9C,$3D,$7F,$71
	.dc.b	$F7,$B9,$BF,$0A,$DA,$FF,$12,$F9,$BF,$CA,$67,$3D,$3F,$A2,$FB,$AF
	.dc.b	$E0,$78,$7E,$71,$AF,$FB,$11,$AF,$F0,$06,$E8,$FA,$87,$D3,$7E,$B1
	.dc.b	$FE,$AB,$87,$E5,$9C,$E5,$FA,$27,$2F,$EF,$B0,$7E,$18,$F9,$0F,$EC
	.dc.b	$B1,$86,$FE,$1D,$CA,$3F,$10,$66,$BF,$A5,$BC,$BF,$13,$FE,$24,$D8
	.dc.b	$FF,$0F,$5A,$72,$D8,$FF,$11,$7E,$18,$FB,$4F,$0B,$63,$AC,$EE,$FC
	.dc.b	$33,$97,$F1,$7F,$87,$FC,$DA,$FF,$75,$B6,$3F,$93,$AF,$AA,$FE,$22
	.dc.b	$E7,$73,$7F,$4D,$FF,$63,$DF,$F0,$88,$C7,$93,$FC,$63,$F1,$FF,$C9
	.dc.b	$70,$FF,$02,$7F,$25,$5F,$C3,$9C,$BF,$EF,$C7,$39,$7E,$81,$B3,$C7
	.dc.b	$AF,$A7,$CB,$75,$7E,$8E,$CF,$B6,$A6,$3E,$DE,$E5,$F1,$3E,$A3,$E5
	.dc.b	$91,$F9,$EF,$D2,$FC,$93,$F8,$C4,$7D,$FF,$77,$FC,$6E,$FE,$06,$33
	.dc.b	$AF,$F4,$09,$CB,$FD,$3B,$F7,$57,$77,$F3,$87,$F2,$17,$FC,$B1,$CB
	.dc.b	$7A,$BF,$9E,$FE,$EB,$FD,$C7,$83,$F9,$3B,$67,$FA,$6F,$2A,$F9,$E3
	.dc.b	$FC,$12,$B4,$FF,$5E,$FA,$67,$E2,$1F,$D2,$17,$C6,$58,$C3,$FD,$33
	.dc.b	$F5,$6F,$B8,$7D,$D9,$DE,$5F,$D0,$A3,$F0,$96,$F3,$BF,$75,$A7,$FA
	.dc.b	$DA,$31,$E3,$FE,$4B,$6F,$D1,$D6,$B0,$FD,$42,$FC,$37,$F3,$ED,$8F
	.dc.b	$F3,$E3,$6F,$FD,$06,$D7,$F6,$07,$F8,$83,$D6,$FD,$7F,$17,$D2,$8F
	.dc.b	$25,$E2,$7F,$B3,$FE,$73,$F9,$FC,$FF,$A8,$FD,$B7,$F6,$D7,$FA,$9C
	.dc.b	$67,$2F,$9B,$39,$69,$FD,$C1,$C6,$1F,$E0,$A9,$E0,$FF,$83,$23,$3C
	.dc.b	$6F,$C7,$BF,$BC,$5F,$F6,$A7,$FC,$27,$5F,$E1,$27,$2F,$F3,$C7,$32
	.dc.b	$FC,$67,$ED,$B7,$E1,$3F,$C0,$33,$A9,$FB,$F5,$F4,$FE,$8B,$E7,$9F
	.dc.b	$C0,$BC,$9B,$8F,$F0,$EE,$07,$C5,$7D,$02,$47,$D0,$7F,$88,$FE,$04
	.dc.b	$FA,$BF,$E3,$6D,$9E,$53,$89,$FF,$09,$FB,$5F,$E5,$BC,$BF,$AA,$BF
	.dc.b	$D7,$F8,$7F,$BF,$B9,$6F,$B3,$FE,$1A,$BF,$F1,$1D,$7F,$5C,$6C,$EB
	.dc.b	$0F,$F0,$8C,$61,$FD,$0F,$E2,$BD,$5B,$F4,$9F,$E2,$48,$C5,$FF,$8F
	.dc.b	$94,$7F,$8A,$3F,$43,$FD,$3E,$1F,$E1,$26,$DF,$23,$FC,$3E,$E2,$FF
	.dc.b	$C8,$FA,$7F,$89,$5F,$F1,$5B,$5F,$F1,$5B,$5F,$DA,$B5,$E3,$B7,$F8
	.dc.b	$E2,$3F,$A3,$57,$F8,$B1,$E9,$7F,$E2,$47,$35,$FD,$FB,$F1,$5F,$F5
	.dc.b	$6D,$7F,$00,$6C,$7F,$96,$9F,$F6,$05,$7F,$8B,$1C,$F2,$1B,$F9,$5F
	.dc.b	$F2,$CA,$E1,$F2,$FF,$E4,$A8,$C5,$3E,$41,$F6,$9F,$E4,$C6,$E5,$FE
	.dc.b	$4C,$6E,$1F,$E4,$C6,$BF,$E4,$C5,$D3,$FE,$56,$6C,$FF,$92,$9E,$39
	.dc.b	$FF,$90,$9C,$D7,$F8,$73,$F2,$9F,$F4,$0B,$FE,$56,$6B,$FC,$81,$AF
	.dc.b	$F9,$E9,$B8,$97,$85,$FE,$66,$6E,$1F,$E5,$A7,$FD,$1B,$87,$F9,$49
	.dc.b	$CD,$FF,$8E,$FE,$35,$F9,$8B,$E0,$3F,$DB,$2D,$C3,$FD,$83,$F6,$0A
	.dc.b	$E7,$1F,$E8,$07,$8D,$7F,$B9,$A7,$59,$59,$EF,$92,$F8,$61,$D7,$1F
	.dc.b	$F2,$66,$1C,$27,$BF,$C5,$BF,$24,$FF,$4B,$4A,$31,$3F,$C9,$29,$39
	.dc.b	$AD,$D6,$A7,$F9,$4C,$F1,$EC,$F3,$4F,$EF,$67,$C2,$EF,$1F,$34,$74
	.dc.b	$8D,$47,$F9,$EB,$F3,$9F,$C7,$93,$E9,$7F,$D2,$AF,$E4,$BF,$56,$FD
	.dc.b	$ED,$FF,$A4,$1A,$FF,$5A,$6B,$FE,$B4,$6E,$3F,$C4,$7D,$7F,$95,$97
	.dc.b	$E5,$A7,$F8,$15,$7B,$A3,$F3,$37,$FE,$C8,$FA,$47,$C5,$FF,$66,$8F
	.dc.b	$EC,$3B,$7F,$AD,$27,$37,$FE,$C2,$9E,$15,$BF,$43,$DF,$5A,$E5,$57
	.dc.b	$2B,$FE,$BD,$18,$AF,$8B,$87,$FA,$B2,$B4,$D8,$CD,$7F,$B1,$34,$7C
	.dc.b	$63,$D7,$EF,$9E,$63,$18,$79,$E3,$C8,$DB,$FB,$B3,$A5,$9F,$B3,$0E
	.dc.b	$B8,$74,$9F,$C1,$FE,$A5,$FE,$D8,$AD,$5F,$17,$84,$67,$1F,$F0,$6F
	.dc.b	$D5,$7F,$A3,$2F,$AC,$63,$E3,$BA,$6C,$A7,$E0,$78,$78,$63,$F2,$7E
	.dc.b	$93,$F9,$EB,$71,$D6,$1C,$A7,$38,$FF,$87,$1B,$3F,$9A,$F9,$0F,$EC
	.dc.b	$5B,$FC,$1C,$D5,$AD,$FF,$D4,$5E,$77,$BF,$07,$E3,$4F,$AC,$C3,$71
	.dc.b	$F9,$53,$AF,$98,$CD,$49,$CD,$3E,$A5,$F6,$38,$AE,$71,$8C,$3E,$29
	.dc.b	$FF,$00,$2B,$85,$7E,$CE,$3C,$73,$B7,$FC,$33,$3A,$7F,$70,$D7,$3F
	.dc.b	$5E,$D7,$94,$A3,$13,$C1,$AF,$09,$DF,$B1,$FE,$95,$E2,$FC,$FC,$FC
	.dc.b	$4A,$E3,$18,$BE,$97,$F0,$2F,$87,$0F,$FA,$76,$51,$8C,$3D,$CD,$7E
	.dc.b	$7B,$63,$FD,$85,$1F,$5F,$DE,$FF,$6D,$3F,$99,$E0,$F9,$9F,$81,$73
	.dc.b	$6C,$6F,$93,$B6,$5F,$E5,$66,$DF,$29,$F9,$F3,$97,$B7,$D0,$E8,$8E
	.dc.b	$0F,$17,$3F,$7E,$9F,$A5,$B7,$E9,$FE,$ED,$F8,$6F,$8D,$B7,$E7,$AD
	.dc.b	$F0,$7E,$D7,$8F,$91,$27,$F2,$9F,$89,$3E,$77,$D2,$D8,$FE,$43,$AA
	.dc.b	$31,$1F,$C2,$9B,$CF,$FE,$CB,$9D,$57,$FB,$36,$31,$2F,$AA,$EB,$32
	.dc.b	$FC,$31,$F2,$FE,$79,$ED,$FF,$B5,$7C,$8F,$E7,$3C,$A9,$CD,$FF,$5A
	.dc.b	$6A,$9F,$E7,$9C,$9C,$73,$FE,$54,$FE,$6F,$4F,$E4,$6D,$F1,$3F,$D1
	.dc.b	$77,$FD,$46,$72,$DD,$5F,$35,$CE,$3F,$39,$18,$47,$87,$0F,$F4,$EC
	.dc.b	$66,$30,$70,$F7,$F9,$68,$FD,$2B,$F2,$B3,$FD,$D1,$AF,$F6,$A6,$E9
	.dc.b	$F0,$9F,$D5,$46,$FB,$7F,$6D,$5C,$DB,$3A,$9F,$EC,$ED,$7E,$B7,$EC
	.dc.b	$39,$BF,$EC,$BA,$23,$19,$FF,$AD,$9B,$5F,$E0,$7F,$D1,$FF,$DE,$CF
	.dc.b	$C0,$BF,$ED,$BE,$8A,$73,$3F,$E1,$36,$CE,$9B,$7D,$7C,$0F,$99,$8B
	.dc.b	$F3,$07,$D6,$E1,$B8,$F7,$BF,$F6,$4C,$62,$3E,$6D,$71,$F8,$2E,$07
	.dc.b	$DF,$9F,$F4,$3C,$FD,$D6,$BF,$EB,$29,$E0,$FE,$07,$EB,$BF,$D6,$B9
	.dc.b	$FC,$46,$E1,$FE,$D8,$6C,$FF,$1D,$F0,$BF,$A4,$BE,$53,$1E,$0F,$E3
	.dc.b	$E3,$31,$DD,$AF,$E9,$DA,$FF,$B9,$3E,$DB,$E7,$E2,$FD,$21,$F8,$1F
	.dc.b	$C6,$6D,$77,$95,$70,$27,$27,$CC,$71,$C1,$F3,$BF,$0D,$78,$D4,$98
	.dc.b	$73,$87,$C7,$6C,$FD,$43,$88,$FD,$33,$5F,$94,$D7,$F8,$14,$3E,$97
	.dc.b	$EE,$B6,$3D,$0D,$7E,$CF,$E3,$3E,$47,$D8,$FF,$27,$E2,$FC,$AB,$98
	.dc.b	$DE,$1F,$88,$8D,$71,$7E,$CB,$57,$D2,$7F,$B0,$6F,$E3,$35,$FC,$3B
	.dc.b	$70,$FD,$0F,$E7,$7C,$A7,$FA,$C3,$F6,$23,$3C,$5F,$96,$8D,$59,$D6
	.dc.b	$1F,$C0,$AD,$F1,$0F,$63,$F8,$44,$61,$B1,$98,$FC,$40,$FC,$53,$FD
	.dc.b	$32,$E1,$AF,$F9,$95,$AF,$73,$87,$C3,$7F,$77,$69,$FD,$E6,$3F,$D3
	.dc.b	$6F,$3F,$86,$F1,$C6,$23,$F8,$A8,$FC,$5B,$EE,$6B,$FC,$13,$D8,$FD
	.dc.b	$BC,$FF,$7B,$6E,$BF,$E1,$AF,$A6,$FD,$CF,$E7,$4F,$1A,$F8,$18,$66
	.dc.b	$1F,$97,$9F,$D8,$E3,$FC,$43,$E1,$7F,$3E,$4E,$CF,$E0,$6F,$F8,$66
	.dc.b	$BF,$3C,$D7,$F7,$EE,$DC,$FB,$1F,$3B,$F0,$4F,$E2,$F0,$69,$8F,$C2
	.dc.b	$BF,$D0,$71,$F0,$1A,$FC,$B6,$BF,$6D,$AF,$F3,$E6,$A5,$EE,$7D,$E6
	.dc.b	$C7,$DC,$6B,$F3,$BF,$14,$7D,$27,$F2,$0F,$55,$F3,$74,$95,$FB,$DF
	.dc.b	$C5,$9F,$7F,$FC,$99,$7C,$95,$89,$3B,$8F,$E5,$7E,$CF,$68,$FE,$0C
	.dc.b	$D7,$F9,$3B,$78,$7F,$A0,$FE,$C5,$FE,$7B,$9F,$E1,$8B,$E9,$BF,$9A
	.dc.b	$B9,$8F,$E5,$ED,$9E,$33,$F9,$EF,$A8,$FC,$4F,$83,$FA,$3F,$03,$F8
	.dc.b	$BE,$1F,$8E,$6D,$FD,$2A,$DC,$FC,$DF,$C3,$3F,$35,$C4,$69,$23,$FC
	.dc.b	$A9,$0F,$AD,$FA,$04,$8F,$A1,$A7,$F4,$CF,$DD,$BF,$53,$8F,$F5,$E6
	.dc.b	$E7,$FC,$1D,$B5,$FE,$57,$5F,$8E,$71,$39,$78,$92,$7F,$98,$4E,$9F
	.dc.b	$E6,$5F,$8B,$FC,$3B,$FE,$25,$F9,$8F,$F0,$D6,$FC,$8F,$F5,$E9,$D5
	.dc.b	$7F,$99,$5B,$E1,$9A,$4E,$2F,$CE,$7D,$A4,$AF,$E0,$FD,$A8,$FC,$87
	.dc.b	$CD,$C9,$F8,$AF,$CA,$23,$F2,$3E,$33,$F2,$C4,$79,$F2,$7F,$72,$FC
	.dc.b	$5D,$F8,$AD,$EE,$95,$FD,$3E,$72,$FF,$51,$F8,$8F,$E7,$4F,$F9,$05
	.dc.b	$AF,$F3,$36,$BF,$E8,$36,$EC,$FC,$3B,$FD,$03,$CD,$FE,$12,$6F,$17
	.dc.b	$F3,$B6,$DB,$E0,$5F,$8A,$7F,$49,$BF,$F0,$CF,$E7,$BE,$4A,$FF,$01
	.dc.b	$B6,$33,$AF,$F8,$E6,$FF,$D2,$5F,$3B,$FE,$10,$FD,$93,$FC,$5F,$87
	.dc.b	$F9,$65,$AF,$F4,$B6,$BF,$E9,$B6,$E8,$FC,$BB,$F5,$5A,$FF,$51,$5F
	.dc.b	$3B,$FE,$1C,$6D,$FF,$77,$76,$FF,$A9,$7C,$0F,$E9,$67,$F5,$DD,$3F
	.dc.b	$C6,$6D,$BF,$EB,$8F,$F5,$0F,$C5,$DF,$4D,$FE,$5B,$6D,$FF,$96,$5A
	.dc.b	$FF,$87,$5A,$FF,$9D,$7E,$AF,$FA,$C9,$3D,$BF,$89,$F9,$9B,$E8,$BF
	.dc.b	$C6,$7F,$81,$38,$69,$30,$FF,$0F,$71,$3F,$45,$FC,$F4,$7F,$AF,$FF
	.dc.b	$4A,$7D,$8F,$F9,$BD,$B1,$FE,$6D,$6B,$FE,$31,$6B,$FE,$89,$F9,$9F
	.dc.b	$E7,$DC,$5F,$E5,$47,$FD,$C5,$1F,$E1,$F6,$BF,$E4,$07,$7C,$5F,$E0
	.dc.b	$1F,$94,$F8,$35,$FE,$5C,$6C,$67,$8B,$F0,$5F,$93,$FF,$09,$C3,$FC
	.dc.b	$54,$53,$FC,$88,$53,$FD,$38,$53,$FC,$1C,$53,$F2,$9F,$86,$7D,$CD
	.dc.b	$4F,$E0,$23,$FD,$03,$18,$E7,$FC,$94,$EF,$F2,$97,$E2,$5F,$95,$7F
	.dc.b	$D2,$77,$C1,$BA,$7F,$9B,$9D,$C7,$D2,$F1,$7F,$DD,$2F,$F8,$47,$14
	.dc.b	$62,$3F,$D2,$B1,$8E,$BF,$F7,$E3,$63,$FC,$14,$7E,$B7,$C8,$FE,$47
	.dc.b	$1F,$D0,$7C,$DF,$D4,$47,$F2,$6C,$5E,$3F,$FA,$12,$BF,$C9,$7F,$01
	.dc.b	$FA,$DF,$C5,$FF,$D4,$15,$FE,$84,$6F,$17,$8B,$E2,$FF,$C0,$2D,$F8
	.dc.b	$2F,$1C,$7F,$82,$74,$F1,$E3,$FE,$56,$9C,$9E,$8F,$59,$F3,$FF,$BB
	.dc.b	$6D,$7E,$51,$AF,$D0,$7F,$A7,$E1,$F0,$BE,$1B,$F8,$43,$7E,$3B,$F8
	.dc.b	$24,$62,$BF,$87,$3C,$1B,$F2,$7F,$22,$3F,$AD,$C7,$CF,$6B,$D6,$3E
	.dc.b	$6F,$E3,$0F,$AE,$FE,$20,$D9,$FD,$4B,$5F,$F6,$2E,$EF,$D2,$BF,$DF
	.dc.b	$31,$86,$C3,$2F,$F5,$84,$61,$BF,$4B,$DB,$E3,$FF,$8F,$7E,$33,$FE
	.dc.b	$F6,$BF,$F5,$AF,$C8,$FD,$2C,$7F,$95,$37,$89,$5B,$C7,$F0,$ED,$AF
	.dc.b	$E6,$0D,$9D,$FA,$DD,$BD,$1F,$E5,$9F,$59,$FD,$DE,$FF,$82,$F7,$FF
	.dc.b	$55,$6F,$B3,$AF,$66,$47,$F0,$AA,$E0,$FE,$99,$BA,$FF,$B3,$1E,$51
	.dc.b	$9A,$FC,$CC,$62,$3F,$40,$D7,$F3,$ED,$7F,$A4,$7E,$EE,$FC,$07,$FA
	.dc.b	$17,$E3,$1F,$C2,$B0,$FF,$6C,$B7,$7F,$D3,$D6,$EF,$F8,$FB,$5F,$DD
	.dc.b	$B6,$BF,$DC,$2D,$9E,$3F,$15,$FE,$80,$FC,$6B,$F9,$65,$7E,$BA,$D1
	.dc.b	$87,$F6,$2D,$7F,$CF,$8D,$7F,$D1,$8F,$F9,$0B,$F1,$AE,$4D,$8E,$3F
	.dc.b	$16,$9C,$B6,$3F,$D0,$9F,$5B,$FE,$85,$4D,$23,$FE,$8B,$E3,$7D,$53
	.dc.b	$EB,$FF,$21,$3F,$C6,$A3,$1A,$7F,$42,$7F,$CC,$2D,$9D,$E7,$C2,$F9
	.dc.b	$C6,$27,$F8,$9D,$3E,$4F,$18,$7F,$7D,$8E,$AF,$78,$CD,$BE,$63,$FA
	.dc.b	$B5,$FF,$06,$6D,$7A,$5A,$FF,$23,$6B,$FC,$89,$FF,$35,$C7,$F5,$A6
	.dc.b	$C7,$F3,$67,$9E,$3E,$97,$2D,$CF,$C7,$1F,$E7,$2A,$E5,$18,$F9,$0F
	.dc.b	$3F,$E3,$4F,$26,$98,$F1,$ED,$FF,$6A,$9D,$37,$E9,$B3,$9F,$F4,$89
	.dc.b	$47,$C2,$B7,$89,$F2,$34,$7D,$2B,$E4,$35,$39,$8C,$64,$CB,$F8,$7C
	.dc.b	$61,$B5,$C2,$75,$7F,$6A,$B4,$FD,$BF,$A0,$F3,$DF,$E1,$DA,$FF,$51
	.dc.b	$6B,$FD,$41,$AF,$F8,$E5,$F5,$D7,$F8,$29,$B1,$FD,$99,$EB,$3D,$F2
	.dc.b	$FC,$0F,$E3,$3F,$E0,$CB,$FF,$08,$B6,$FF,$B6,$BC,$AB,$F2,$0E,$5B
	.dc.b	$E9,$FF,$AF,$38,$93,$FD,$9A,$FF,$A7,$B6,$FF,$44,$D7,$FB,$67,$E3
	.dc.b	$3F,$7D,$FE,$F6,$BC,$7F,$CA,$B9,$C7,$F4,$0E,$6F,$E5,$9F,$33,$D6
	.dc.b	$7A,$1F,$65,$B1,$FE,$2C,$F8,$3F,$D5,$E3,$FD,$12,$72,$FF,$82,$BF
	.dc.b	$4D,$FE,$E3,$5F,$D9,$9B,$1F,$B9,$6B,$FE,$03,$6B,$FE,$02,$FC,$67
	.dc.b	$9E,$5F,$B5,$5F,$53,$FA,$E7,$37,$B8,$FD,$8F,$D0,$BF,$0B,$FD,$9F
	.dc.b	$E3,$1F,$D4,$EB,$FC,$87,$39,$7F,$C8,$B1,$9C,$58,$FF,$0A,$F7,$2F
	.dc.b	$DA,$FF,$BB,$9B,$7F,$E1,$26,$D7,$E5,$CF,$CC,$38,$7F,$C3,$D1,$FE
	.dc.b	$28,$3F,$CA,$65,$7B,$BC,$0F,$F7,$69,$AF,$37,$F8,$7C,$BE,$F3,$F8
	.dc.b	$7E,$1E,$55,$F4,$9A,$73,$7E,$27,$2F,$DC,$C1,$6F,$C5,$FF,$D5,$33
	.dc.b	$EB,$6C,$FA,$CD,$B8,$7D,$D3,$C7,$D1,$FC,$B7,$C5,$BD,$3F,$E4,$2D
	.dc.b	$C3,$C1,$4F,$B6,$FC,$F4,$7F,$8F,$A7,$58,$7F,$94,$3E,$3E,$EB,$FC
	.dc.b	$CD,$B5,$F3,$3E,$DF,$E4,$4E,$72,$F8,$BF,$00,$F7,$FE,$B3,$6F,$EB
	.dc.b	$37,$1F,$66,$9F,$E3,$4A,$FB,$9F,$74,$FC,$17,$E7,$BF,$8C,$AF,$8E
	.dc.b	$E7,$0F,$5B,$9B,$7E,$2E,$FA,$DF,$AE,$FF,$AB,$2B,$E0,$B9,$6F,$C5
	.dc.b	$7D,$CF,$C0,$3E,$AE,$3E,$46,$BF,$39,$AF,$E3,$1A,$FF,$B3,$7F,$17
	.dc.b	$79,$FF,$D4,$1B,$7F,$43,$F1,$77,$DA,$66,$BF,$AA,$EF,$FC,$9A,$BF
	.dc.b	$CD,$ED,$C3,$F0,$BE,$F7,$E0,$B8,$FF,$B2,$4E,$77,$7E,$2F,$FC,$02
	.dc.b	$FF,$06,$D7,$F4,$2D,$7F,$1E,$D8,$7B,$AF,$C3,$7F,$3A,$BE,$0A,$C6
	.dc.b	$1F,$ED,$F6,$FB,$4D,$69,$F9,$B8,$C4,$EA,$5F,$86,$D8,$E3,$F8,$97
	.dc.b	$FB,$4A,$FF,$C3,$8D,$BF,$F4,$5B,$5F,$D0,$35,$FD,$E2,$FC,$5F,$99
	.dc.b	$D3,$E5,$AD,$7E,$7B,$F1,$57,$D7,$FE,$E1,$B1,$9F,$8A,$C4,$C8,$E1
	.dc.b	$ED,$DC,$BF,$81,$9F,$B6,$BF,$A1,$F6,$1F,$89,$8C,$F0,$A3,$F1,$D1
	.dc.b	$87,$5C,$9F,$C6,$5C,$B7,$AF,$F8,$5C,$3E,$CE,$C6,$1F,$8F,$F8,$3F
	.dc.b	$C8,$F9,$8B,$A7,$CF,$E1,$46,$63,$0E,$B2,$F4,$89,$DC,$F1,$F4,$7E
	.dc.b	$99,$C6,$77,$3F,$CC,$DF,$6F,$E2,$6F,$03,$F9,$0B,$6F,$F8,$5A,$FC
	.dc.b	$57,$33,$AE,$DD,$F5,$DB,$DB,$FC,$AB,$E5,$BF,$AF,$63,$F2,$97,$CA
	.dc.b	$67,$DD,$D3,$BF,$8B,$98,$7F,$1A,$AD,$B6,$3F,$2B,$F8,$BF,$F3,$8C
	.dc.b	$3E,$0B,$79,$7F,$CA,$71,$F3,$F8,$37,$9F,$E7,$3C,$EF,$D1,$63,$FE
	.dc.b	$6F,$6D,$7F,$B7,$9A,$FF,$33,$6B,$FD,$3D,$FC,$FC,$7E,$8D,$B1,$F4
	.dc.b	$9E,$77,$E2,$72,$FF,$5D,$FC,$4F,$EA,$CF,$E3,$17,$B3,$FD,$41,$CC
	.dc.b	$EE,$30,$7F,$54,$F2,$FF,$B4,$46,$70,$FE,$7F,$18,$7F,$B1,$FE,$27
	.dc.b	$70,$FE,$A2,$DE,$B7,$E1,$1E,$0A,$FE,$8F,$39,$3F,$C1,$B8,$7F,$73
	.dc.b	$7F,$AE,$DF,$2F,$BC,$FE,$CB,$18,$86,$FF,$BD,$AF,$A2,$DB,$FE,$ED
	.dc.b	$F9,$1F,$EE,$DF,$E1,$51,$9A,$FF,$28,$6D,$FC,$A9,$FA,$5E,$5F,$DA
	.dc.b	$7E,$83,$5D,$B5,$47,$F9,$12,$BE,$B7,$C6,$7D,$FA,$FC,$03,$FD,$F3
	.dc.b	$D2,$FF,$32,$CF,$F6,$66,$DF,$EF,$97,$A7,$BF,$8C,$E1,$E7,$F1,$FB
	.dc.b	$3B,$87,$17,$FC,$0F,$F8,$A7,$F5,$FD,$FF,$C3,$4D,$9F,$F0,$AB,$DA
	.dc.b	$33,$5C,$2B,$8F,$E2,$96,$FF,$BC,$DA,$3F,$C2,$7F,$43,$F9,$FD,$3F
	.dc.b	$C0,$6D,$7F,$C4,$2D,$7F,$C1,$ED,$E5,$2F,$C6,$79,$9D,$3A,$AE,$3F
	.dc.b	$E0,$8F,$C5,$3F,$D3,$6F,$F8,$B1,$B8,$7F,$86,$9F,$F5,$0C,$67,$E5
	.dc.b	$7F,$C4,$F1,$AC,$4F,$F9,$7E,$74,$75,$F9,$5F,$F1,$6F,$C8,$BE,$FB
	.dc.b	$FC,$7C,$D9,$FF,$1F,$37,$1F,$F0,$4B,$5F,$F1,$1A,$FA,$7E,$67,$4A
	.dc.b	$B5,$FE,$22,$FC,$4F,$FC,$E9,$7F,$E2,$A6,$BF,$E4,$86,$BF,$E3,$66
	.dc.b	$E4,$7F,$8C,$F1,$FF,$95,$5B,$5C,$72,$FF,$23,$C6,$2D,$F5,$6F,$95
	.dc.b	$C0,$FC,$0F,$F9,$64,$FF,$50,$9D,$79,$3F,$ED,$A9,$CE,$5F,$EE,$56
	.dc.b	$E5,$FE,$B5,$C3,$97,$97,$FE,$E4,$C3,$8C,$FF,$93,$E7,$2F,$F9,$46
	.dc.b	$3F,$CB,$5A,$7E,$A5,$B1,$FE,$4F,$CB,$CF,$F1,$5F,$E5,$68,$C4,$EA
	.dc.b	$7F,$CA,$8F,$F9,$8F,$EC,$7C,$14,$6D,$E1,$FF,$34,$FE,$1F,$F8,$6B
	.dc.b	$FF,$30,$B6,$4F,$E1,$CC,$1F,$40,$9B,$FF,$A0,$5F,$A7,$F8,$9F,$FB
	.dc.b	$57,$2F,$F2,$EB,$6F,$FB,$AC,$FE,$3D,$FF,$37,$C6,$76,$FF,$46,$4F
	.dc.b	$F5,$2F,$A1,$FD,$ED,$79,$38,$18,$7F,$B8,$67,$39,$79,$BC,$EC,$DA
	.dc.b	$73,$87,$3F,$03,$FD,$22,$D9,$F7,$F9,$FC,$2B,$79,$5B,$DC,$DA,$DE
	.dc.b	$5E,$19,$FF,$79,$70,$3E,$29,$EB,$6F,$5F,$93,$8F,$1A,$D4,$DA,$DE
	.dc.b	$1C,$F1,$FC,$9C,$61,$F5,$63,$E2,$9D,$5F,$27,$86,$2F,$A6,$7C,$C7
	.dc.b	$85,$B3,$A7,$FD,$8A,$D8,$D1,$E7,$FA,$0F,$AD,$B7,$CC,$AE,$39,$78
	.dc.b	$8F,$F5,$2C,$EB,$2F,$F5,$DB,$63,$27,$4C,$3C,$B1,$98,$7E,$22,$FA
	.dc.b	$3F,$99,$18,$6C,$7F,$BA,$31,$FC,$0C,$E6,$FE,$33,$4F,$21,$D3,$E8
	.dc.b	$DF,$63,$AB,$ED,$56,$73,$5C,$25,$18,$8F,$F5,$E4,$62,$BF,$61,$1D
	.dc.b	$5B,$19,$8F,$DA,$3F,$EC,$ED,$73,$F2,$DF,$EC,$A9,$CB,$5E,$AD,$7F
	.dc.b	$DC,$93,$FE,$92,$7D,$11,$C7,$1F,$83,$3D,$67,$D1,$F1,$BF,$EE,$B9
	.dc.b	$CB,$75,$E6,$D9,$1F,$41,$C3,$F4,$BE,$33,$58,$BD,$BF,$C4,$CF,$E3
	.dc.b	$1A,$FE,$06,$35,$C3,$C2,$5F,$3D,$4C,$44,$AE,$51,$96,$90,$E9,$2F
	.dc.b	$9B,$F4,$72,$7E,$AF,$7B,$CB,$F5,$76,$FE,$07,$39,$C3,$F4,$3F,$88
	.dc.b	$FF,$BB,$5B,$F2,$BC,$30,$91,$83,$7F,$13,$F0,$2B,$84,$7B,$5B,$87
	.dc.b	$F0,$C5,$3B,$24,$F2,$3C,$3F,$31,$F2,$BB,$7E,$54,$7F,$BF,$1A,$FF
	.dc.b	$20,$D7,$E9,$BF,$E9,$6B,$FF,$1E,$BF,$5B,$BF,$F8,$CF,$0A,$73,$3F
	.dc.b	$EF,$46,$BC,$A3,$5E,$EF,$E5,$9B,$1F,$D1,$DE,$5E,$C7,$82,$FA,$DF
	.dc.b	$FB,$92,$BC,$1D,$5F,$86,$BD,$C6,$03,$E3,$F6,$39,$4F,$F2,$0F,$C4
	.dc.b	$5D,$78,$BF,$ED,$71,$87,$FA,$E3,$5F,$ED,$4F,$FB,$75,$CF,$47,$EB
	.dc.b	$B9,$13,$99,$FE,$B6,$D7,$E4,$C6,$A7,$F6,$BE,$43,$F1,$0F,$F3,$AD
	.dc.b	$7F,$6D,$6B,$FE,$C8,$8D,$7C,$0F,$9F,$47,$E2,$3F,$B1,$AF,$F6,$4B
	.dc.b	$77,$FE,$97,$7C,$27,$A5,$F1,$AD,$6B,$F2,$7F,$11,$6F,$D5,$B7,$F2
	.dc.b	$F2,$46,$23,$FD,$AE,$BE,$E3,$81,$F6,$DC,$47,$DB,$6B,$F8,$A3,$EA
	.dc.b	$B8,$7F,$09,$DD,$FA,$B7,$C0,$D7,$FD,$CA,$DD,$3F,$1F,$F5,$3F,$C8
	.dc.b	$5B,$D7,$F8,$B8,$FE,$49,$5F,$8F,$6B,$F8,$76,$BF,$86,$FC,$43,$E1
	.dc.b	$B7,$93,$E7,$47,$E0,$6F,$F3,$AE,$B3,$34,$9E,$0F,$0E,$3F,$AD,$F8
	.dc.b	$83,$EA,$E9,$32,$F9,$91,$FC,$5B,$1F,$D4,$BF,$C6,$23,$F4,$CD,$7F
	.dc.b	$3C,$D7,$F3,$AD,$C5,$F8,$87,$3E,$2F,$D2,$3F,$C7,$AB,$ED,$35,$5F
	.dc.b	$EE,$23,$84,$7E,$A1,$F9,$D3,$CB,$C3,$FC,$84,$62,$7F,$5B,$FA,$27
	.dc.b	$DB,$BF,$E0,$3F,$69,$9C,$4F,$46,$F1,$7E,$66,$35,$F8,$EF,$1C,$FF
	.dc.b	$74,$DC,$7F,$72,$DB,$3F,$7E,$7F,$89,$FE,$25,$FE,$75,$AF,$E0,$01
	.dc.b	$C3,$EE,$8D,$FB,$7F,$BF,$8D,$7C,$6D,$A3,$FD,$A2,$32,$7F,$6A,$8C
	.dc.b	$B6,$35,$F7,$3F,$76,$3C,$2D,$7C,$07,$F5,$A7,$1A,$7F,$8F,$A3,$FD
	.dc.b	$4F,$EA,$7E,$EF,$F1,$2F,$D6,$4F,$F2,$A7,$F8,$F4,$F2,$D1,$18,$9D
	.dc.b	$76,$71,$FB,$29,$7F,$B7,$A4,$75,$ED,$FE,$11,$EE,$FF,$82,$2B,$F7
	.dc.b	$8F,$E7,$77,$FE,$54,$D8,$FE,$52,$ED,$F1,$3F,$A1,$31,$8F,$F4,$06
	.dc.b	$CF,$2E,$9F,$9D,$F8,$27,$BE,$BE,$03,$67,$FA,$5B,$6F,$FB,$FA,$F9
	.dc.b	$1E,$27,$D3,$71,$1F,$E1,$C6,$BF,$E8,$D6,$BF,$C3,$7F,$12,$7C,$D7
	.dc.b	$DF,$6C,$7E,$19,$AF,$D6,$1F,$24,$F8,$8F,$E5,$DF,$89,$3E,$26,$92
	.dc.b	$BF,$92,$43,$E0,$B7,$CB,$C1,$93,$A8,$FE,$71,$F8,$93,$E7,$2F,$C0
	.dc.b	$CA,$FE,$5D,$39,$6F,$CC,$70,$23,$AF,$37,$E6,$FD,$0C,$8F,$D0,$F0
	.dc.b	$A3,$E8,$FD,$0C,$9C,$B5,$F8,$13,$FE,$06,$78,$FB,$59,$E6,$FA,$F5
	.dc.b	$F6,$A3,$FA,$87,$E2,$5F,$EE,$BA,$FE,$88,$FE,$BA,$BF,$C0,$2D,$E7
	.dc.b	$FE,$E8,$EE,$7F,$34,$FE,$A9,$BF,$1B,$D0,$EB,$E0,$FF,$5D,$FC,$43
	.dc.b	$E1,$4F,$E8,$5F,$F4,$AB,$FD,$4D,$AE,$78,$78,$11,$D3,$CF,$7F,$C9
	.dc.b	$FC,$7C,$9F,$8E,$FE,$5B,$F0,$29,$1F,$8D,$7F,$20,$47,$FC,$D7,$F8
	.dc.b	$87,$CD,$7F,$87,$B5,$FC,$53,$75,$FD,$F2,$DF,$FB,$C7,$E2,$1F,$E2
	.dc.b	$49,$FF,$13,$AF,$81,$C9,$AF,$7F,$FB,$F7,$CC,$3E,$5F,$FC,$4C,$DC
	.dc.b	$BF,$C4,$CD,$7F,$B6,$79,$28,$C4,$7F,$8E,$3D,$C7,$E1,$EF,$9A,$FF
	.dc.b	$10,$46,$1B,$19,$9F,$F0,$07,$E2,$1F,$EF,$17,$FE,$00,$7B,$75,$FF
	.dc.b	$0A,$37,$6F,$F0,$7B,$7A,$7F,$C6,$9F,$88,$7B,$DF,$F1,$B7,$E2,$0F
	.dc.b	$AE,$FF,$20,$37,$0F,$F2,$03,$5F,$5B,$5F,$F2,$47,$E2,$1F,$80,$C7
	.dc.b	$FC,$9E,$FE,$E9,$BE,$5F,$F8,$96,$35,$87,$F9,$17,$C0,$FE,$A8,$7F
	.dc.b	$53,$27,$B7,$FE,$2B,$FC,$43,$FC,$B1,$8F,$F9,$45,$AF,$CC,$6B,$FE
	.dc.b	$59,$6B,$FE,$96,$E1,$7E,$20,$FC,$43,$E6,$B6,$7F,$CA,$93,$A9,$FF
	.dc.b	$1D,$7D,$2F,$F6,$B8,$7E,$20,$FC,$47,$EA,$B6,$F8,$B5,$FE,$89,$F7
	.dc.b	$77,$83,$FC,$E3,$F8,$8B,$C1,$FF,$36,$36,$BF,$CD,$8D,$7F,$0E,$D7
	.dc.b	$FC,$E9,$F1,$3F,$D0,$7B,$3F,$DA,$7F,$33,$FE,$4F,$8F,$E7,$B3,$FE
	.dc.b	$6C,$7F,$A3,$68,$FC,$47,$F3,$D3,$EE,$6B,$FE,$59,$6E,$5F,$E8,$6F
	.dc.b	$8F,$7D,$B7,$F9,$6E,$BF,$D3,$67,$FA,$14,$A7,$F9,$F4,$A7,$FB,$57
	.dc.b	$E5,$FF,$CD,$3F,$82,$FD,$E8,$62,$8F,$C2,$74,$7D,$56,$CE,$9E,$12
	.dc.b	$FC,$45,$F5,$BF,$A4,$5F,$81,$53,$11,$FA,$8F,$33,$E9,$61,$FA,$D6
	.dc.b	$BF,$83,$6B,$E4,$6E,$BF,$57,$C2,$FF,$48,$36,$7F,$D3,$0D,$7F,$D0
	.dc.b	$1E,$77,$F2,$78,$FF,$3D,$FC,$EF,$F8,$07,$5F,$C8,$3F,$E0,$86,$FC
	.dc.b	$37,$EF,$23,$57,$F1,$3E,$B1,$F3,$FF,$E9,$06,$C7,$F9,$CC,$D4,$64
	.dc.b	$D7,$D7,$7F,$06,$93,$4C,$79,$7F,$C2,$BE,$2B,$37,$F9,$E6,$BF,$E9
	.dc.b	$96,$E7,$FE,$94,$7F,$8B,$46,$FC,$DF,$F5,$1C,$62,$3F,$D0,$DC,$68
	.dc.b	$C4,$7F,$1B,$6B,$FE,$D1,$FC,$3D,$EF,$FC,$C6,$C7,$CC,$6B,$FE,$AB
	.dc.b	$FC,$41,$F4,$FF,$C8,$9B,$1B,$AE,$46,$67,$F9,$3F,$E2,$1F,$CD,$E7
	.dc.b	$F7,$6F,$FB,$CB,$3F,$F5,$7B,$78,$BF,$82,$3B,$9F,$F5,$D3,$67,$7E
	.dc.b	$AF,$43,$AF,$AB,$FC,$D3,$EA,$1F,$7B,$7E,$06,$63,$FC,$2A,$72,$D9
	.dc.b	$E0,$47,$5E,$D7,$33,$FD,$9F,$3F,$B4,$8F,$E8,$8D,$7F,$8D,$35,$FD
	.dc.b	$5F,$E1,$FF,$E2,$37,$F8,$CA,$FB,$7F,$E9,$2D,$8C,$FC,$56,$08,$E0
	.dc.b	$E3,$F6,$7F,$E7,$1F,$C3,$FF,$93,$6D,$F9,$19,$46,$1F,$CA,$35,$FF
	.dc.b	$64,$35,$11,$FF,$1B,$7E,$1F,$CD,$B1,$CB,$E3,$53,$96,$C7,$F9,$FF
	.dc.b	$EA,$BF,$D0,$09,$A4,$7F,$B6,$C7,$F9,$0F,$8F,$FD,$C8,$D8,$FF,$3B
	.dc.b	$3F,$E4,$38,$CA,$7F,$DD,$5F,$B0,$3E,$A7,$15,$FE,$7D,$8C,$4F,$03
	.dc.b	$73,$AC,$B7,$E8,$7F,$94,$FE,$AB,$FD,$15,$87,$C2,$8C,$5F,$EE,$DB
	.dc.b	$1E,$89,$DA,$8F,$F2,$E7,$47,$5A,$E5,$97,$E9,$1F,$6D,$E6,$F8,$4E
	.dc.b	$5F,$47,$89,$9C,$19,$A3,$15,$BC,$BD,$33,$9C,$7E,$CC,$7C,$2F,$41
	.dc.b	$E1,$EE,$7E,$EC,$63,$4F,$13,$AD,$2D,$6F,$44,$BC,$5F,$F1,$66,$1D
	.dc.b	$8E,$6E,$27,$FD,$6B,$3D,$67,$51,$CA,$75,$8F,$97,$ED,$9F,$5F,$F0
	.dc.b	$FD,$17,$C2,$3E,$EF,$CB,$5A,$CB,$9D,$6A,$78,$56,$E7,$BE,$6F,$C5
	.dc.b	$1E,$06,$AF,$DF,$18,$AD,$F8,$AF,$91,$FB,$78,$23,$13,$FC,$EE,$31
	.dc.b	$3A,$9C,$EF,$FA,$83,$E6,$FE,$9B,$FE,$E0,$BF,$1B,$63,$E2,$CF,$0D
	.dc.b	$FE,$B4,$E5,$FB,$1F,$8F,$FD,$1C,$4B,$8F,$EB,$B9,$31,$8F,$C0,$F5
	.dc.b	$3F,$C4,$CF,$9D,$B9,$79,$4F,$B6,$E1,$F4,$5F,$F7,$7F,$80,$F8,$3F
	.dc.b	$87,$BE,$13,$F0,$AD,$C3,$E6,$FE,$3F,$FC,$19,$F9,$F1,$9C,$3F,$03
	.dc.b	$BF,$F8,$4A,$7E,$9A,$FB,$2F,$C7,$B9,$9D,$C7,$E8,$FF,$22,$FE,$23
	.dc.b	$EF,$FE,$B1,$B7,$FD,$AD,$B5,$EB,$6B,$F4,$5F,$D9,$76,$73,$F0,$3F
	.dc.b	$35,$1F,$B5,$E1,$C4,$F7,$FC,$B7,$FD,$F7,$87,$E5,$A9,$18,$C1,$C7
	.dc.b	$FC,$32,$73,$F4,$5C,$7E,$8F,$D1,$C6,$FA,$2F,$43,$72,$F4,$37,$1F
	.dc.b	$43,$5F,$C4,$76,$FF,$97,$63,$E1,$7D,$E7,$87,$3F,$E3,$73,$9B,$F3
	.dc.b	$4E,$5F,$E6,$F9,$FF,$BF,$7C,$AF,$95,$F4,$66,$FE,$11,$87,$E7,$7C
	.dc.b	$57,$FB,$CE,$FE,$63,$74,$F9,$8D,$7E,$63,$5F,$E8,$B9,$3E,$75,$8F
	.dc.b	$D9,$FD,$35,$F3,$7F,$BB,$9C,$E3,$F3,$67,$2F,$EE,$BF,$4D,$FF,$54
	.dc.b	$3F,$47,$F0,$7F,$66,$4C,$95,$A8,$FA,$75,$F0,$2B,$7B,$7F,$0A,$EB
	.dc.b	$FC,$47,$93,$FB,$4B,$FC,$23,$5F,$8E,$D5,$2F,$C4,$FF,$A3,$C7,$63
	.dc.b	$2E,$38,$FF,$04,$FF,$2F,$8D,$4F,$1F,$43,$F3,$5F,$96,$F3,$CF,$F1
	.dc.b	$8D,$AD,$BF,$C5,$DB,$7F,$A1,$F9,$67,$D8,$7E,$85,$B7,$FA,$16,$D7
	.dc.b	$E8,$5A,$FE,$2B,$DB,$78,$3F,$AB,$FC,$EB,$FC,$19,$7F,$CA,$27,$2F
	.dc.b	$E7,$E3,$35,$EF,$E2,$7E,$9B,$F9,$7D,$FF,$41,$F3,$4F,$C3,$3F,$9B
	.dc.b	$4E,$71,$FC,$8C,$E5,$FD,$57,$C5,$7F,$9F,$9F,$DC,$37,$1F,$DC,$35
	.dc.b	$FD,$C3,$5F,$CD,$FC,$6F,$93,$1F,$D8,$36,$B3,$EF,$7F,$07,$8C,$57
	.dc.b	$F8,$63,$F0,$F4,$7F,$6D,$87,$47,$F1,$77,$EE,$09,$DD,$FF,$56,$8D
	.dc.b	$9C,$67,$FB,$0C,$7F,$8A,$DB,$1F,$E7,$D6,$BF,$CB,$5A,$FF,$43,$FC
	.dc.b	$29,$F1,$FE,$36,$D7,$7A,$E7,$97,$F8,$62,$D5,$A7,$F1,$BF,$95,$7F
	.dc.b	$A1,$AF,$F8,$B3,$63,$3F,$1F,$FC,$A6,$31,$7F,$C5,$FE,$8B,$FB,$45
	.dc.b	$7E,$59,$B1,$FC,$6E,$77,$7F,$C2,$A7,$2F,$F0,$CF,$C4,$BE,$2D,$FE
	.dc.b	$79,$B7,$FC,$8D,$FC,$14,$7E,$A9,$CC,$F1,$D7,$F9,$E7,$E5,$2F,$BC
	.dc.b	$FE,$7D,$5A,$BF,$D8,$CE,$6F,$E2,$7A,$5F,$47,$EB,$3E,$93,$63,$E4
	.dc.b	$47,$D5,$FB,$7C,$5F,$DC,$E0,$46,$6B,$F9,$E7,$9D,$F8,$6A,$FC,$1C
	.dc.b	$FF,$4F,$8F,$C4,$FA,$8F,$C4,$7F,$35,$5F,$C0,$63,$1F,$4D,$FE,$33
	.dc.b	$9C,$F0,$7F,$92,$14,$64,$98,$FF,$88,$2B,$A7,$5E,$30,$FE,$0F,$F1
	.dc.b	$7F,$80,$AF,$E7,$2D,$8F,$E2,$6B,$F3,$1F,$EC,$8E,$7C,$77,$E5,$AF
	.dc.b	$AA,$FE,$D0,$FF,$A7,$30,$FE,$B4,$E5,$BF,$25,$FA,$3F,$BE,$7D,$2F
	.dc.b	$E9,$7E,$A3,$F9,$AE,$9F,$A4,$9C,$DF,$E8,$E7,$3C,$3F,$95,$7F,$17
	.dc.b	$5F,$E0,$B6,$BF,$E2,$16,$BF,$E8,$A6,$BF,$9C,$F9,$4F,$D1,$1E,$1F
	.dc.b	$3E,$DF,$F7,$1F,$CE,$7F,$C0,$98,$7F,$6F,$6C,$7F,$5E,$7F,$C1,$4E
	.dc.b	$FE,$97,$FC,$11,$1A,$C3,$F9,$DF,$E2,$DF,$C7,$9F,$E6,$B1,$9E,$04
	.dc.b	$E7,$0F,$F4,$0F,$E2,$9F,$E3,$3C,$7F,$D0,$1F,$69,$FE,$8D,$7E,$D4
	.dc.b	$67,$1F,$F3,$F4,$E5,$FC,$33,$6B,$FD,$16,$DA,$FF,$45,$57,$1E,$6F
	.dc.b	$F2,$24,$7E,$16,$7F,$45,$6D,$BF,$CC,$B3,$E2,$F0,$7E,$B7,$CE,$7D
	.dc.b	$7F,$90,$FF,$47,$DE,$DB,$7F,$E5,$B8,$FF,$1F,$19,$BF,$F4,$C6,$1F
	.dc.b	$B6,$C4,$FE,$E2,$74,$7A,$BB,$3F,$6B,$EE,$7E,$DB,$8A,$E1,$FE,$25
	.dc.b	$F0,$FF,$82,$DF,$EF,$E7,$53,$9A,$B3,$F4,$39,$3E,$AF,$97,$F5,$9B
	.dc.b	$5F,$7E,$2C,$EE,$30,$7F,$A2,$0F,$F2,$47,$C2,$FF,$AB,$9A,$7B,$8B
	.dc.b	$1F,$E5,$D6,$BF,$E4,$3E,$3F,$3F,$27,$E0,$32,$FF,$58,$57,$0E,$4F
	.dc.b	$F1,$E5,$6B,$97,$F6,$13,$F9,$3E,$DF,$C6,$FA,$DF,$8E,$7F,$CA,$D5
	.dc.b	$FA,$49,$FF,$18,$B7,$2F,$F0,$5D,$F2,$7F,$C0,$B1,$9C,$93,$FD,$4E
	.dc.b	$7F,$CB,$0D,$7F,$3C,$7F,$98,$5C,$3F,$9C,$AF,$09,$FA,$06,$BF,$1F
	.dc.b	$C5,$FE,$55,$1F,$0D,$AF,$FA,$56,$3F,$CE,$5E,$3F,$CB,$9C,$CF,$EB
	.dc.b	$7D,$9F,$D5,$9F,$1A,$39,$E9,$FB,$38,$FF,$3B,$7D,$16,$30,$FF,$28
	.dc.b	$BF,$EC,$57,$FC,$D4,$E6,$FF,$C5,$AE,$5F,$F1,$3B,$6B,$F7,$DE,$F7
	.dc.b	$4F,$94,$FD,$DB,$DB,$E9,$BF,$D0,$55,$FE,$05,$BE,$71,$83,$E3,$7A
	.dc.b	$3F,$C0,$8E,$18,$FF,$0B,$7F,$DD,$5F,$B9,$77,$8F,$F7,$E3,$FE,$2A
	.dc.b	$CB,$FE,$00,$B9,$BF,$F7,$D3,$99,$47,$FB,$F7,$F6,$1F,$F3,$13,$79
	.dc.b	$BF,$E0,$0F,$62,$72,$D8,$FF,$4D,$7E,$5E,$FC,$53,$FD,$3C,$DC,$3F
	.dc.b	$A3,$5F,$F9,$C7,$2F,$F4,$74,$E7,$1F,$EA,$5F,$45,$FE,$19,$F9,$F7
	.dc.b	$D4,$79,$34,$4E,$63,$DB,$F2,$6C,$FD,$37,$C3,$06,$1C,$1D,$08,$D4
	.dc.b	$7A,$FE,$2B,$31,$8D,$3F,$D1,$97,$D3,$1F,$16,$5E,$1C,$7E,$36,$1C
	.dc.b	$27,$CD,$7C,$27,$B4,$7F,$A6,$33,$FF,$15,$E3,$D6,$F8,$57,$F9,$D2
	.dc.b	$73,$5B,$F8,$DC,$F5,$3D,$7F,$F6,$EE,$3E,$D8,$5F,$0C,$3F,$C1,$BF
	.dc.b	$29,$D3,$67,$EE,$DF,$93,$C3,$DE,$DC,$FF,$37,$F2,$7B,$8C,$61,$F2
	.dc.b	$1F,$F3,$8D,$70,$F8,$8D,$75,$ED,$F2,$F5,$3C,$4A,$67,$93,$E4,$63
	.dc.b	$FA,$48,$F7,$AF,$C4,$7F,$B9,$E3,$F5,$E3,$1E,$2E,$A7,$3A,$FE,$52
	.dc.b	$73,$87,$46,$9F,$38,$FC,$CC,$FF,$A1,$38,$F4,$FF,$91,$30,$EC,$E6
	.dc.b	$78,$1E,$3F,$88,$57,$F4,$C8,$C3,$67,$84,$78,$CC,$9F,$D4,$9F,$F3
	.dc.b	$9F,$06,$FC,$FF,$C8,$CE,$5A,$F1,$9D,$BC,$7C,$BF,$76,$3E,$6A,$E2
	.dc.b	$DB,$F5,$CE,$B3,$F8,$98,$3F,$09,$C6,$3F,$0A,$AB,$F3,$DB,$3B,$8F
	.dc.b	$AB,$CC,$E0,$7E,$27,$CA,$30,$D8,$CC,$FD,$CC,$9F,$B3,$BE,$EB,$C3
	.dc.b	$18,$6C,$66,$78,$35,$FB,$13,$CA,$B6,$48,$DB,$E1,$FA,$BF,$27,$57
	.dc.b	$D1,$DF,$F4,$5D,$7C,$DB,$96,$B0,$F1,$F7,$7D,$C7,$DF,$7C,$78,$7F
	.dc.b	$69,$1A,$E3,$8F,$B6,$B8,$3E,$66,$E1,$FC,$23,$E5,$7F,$CA,$93,$F0
	.dc.b	$DA,$FB,$1B,$1B,$F7,$3E,$2C,$3D,$BF,$9D,$F0,$DF,$E8,$7B,$FE,$63
	.dc.b	$E2,$7F,$16,$E4,$4E,$67,$FD,$E2,$D9,$FE,$37,$C5,$FC,$93,$D0,$FD
	.dc.b	$34,$63,$4F,$F3,$65,$FF,$89,$AB,$4D,$8C,$BF,$9B,$8C,$B7,$27,$C7
	.dc.b	$74,$EC,$FE,$65,$1F,$CA,$E7,$FD,$10,$DC,$3C,$6F,$08,$FF,$2F,$7E
	.dc.b	$01,$C5,$FF,$03,$B7,$4F,$F0,$2B,$79,$9F,$80,$D3,$F6,$53,$F9,$56
	.dc.b	$C6,$AB,$FC,$83,$39,$6C,$3C,$5E,$DE,$97,$FA,$62,$7F,$CE,$1F,$07
	.dc.b	$DE,$7C,$15,$67,$FB,$DC,$61,$FF,$3B,$7C,$1F,$22,$3F,$E3,$7F,$CB
	.dc.b	$7F,$46,$B5,$FE,$9B,$B2,$73,$3F,$E1,$36,$BF,$DB,$A3,$5D,$DF,$C3
	.dc.b	$DB,$1E,$CC,$DF,$95,$7A,$89,$1F,$E1,$0F,$8E,$F9,$07,$C4,$23,$F0
	.dc.b	$9F,$93,$3F,$E6,$28,$FF,$48,$7B,$BF,$10,$FA,$1F,$75,$23,$FC,$36
	.dc.b	$D7,$FC,$33,$CB,$F2,$BE,$26,$47,$FA,$65,$FB,$99,$7B,$F7,$BA,$FD
	.dc.b	$1A,$FF,$2C,$35,$FF,$2C,$35,$F8,$91,$F7,$BE,$1D,$F6,$9F,$85,$6C
	.dc.b	$FF,$AF,$A3,$1F,$3D,$23,$04,$3E,$96,$1F,$E1,$AE,$5F,$F6,$4C,$FE
	.dc.b	$3F,$E3,$BF,$D8,$75,$FA,$EA,$D4,$62,$38,$8A,$46,$78,$FF,$68,$F9
	.dc.b	$27,$73,$AC,$3B,$7C,$54,$9E,$C7,$3A,$E1,$7F,$B8,$F8,$77,$BB,$FA
	.dc.b	$5E,$04,$62,$FF,$DD,$8D,$7C,$A3,$F0,$9F,$C2,$BD,$8C,$73,$7F,$59
	.dc.b	$69,$FD,$6E,$36,$FF,$AA,$1A,$FF,$6D,$AF,$0F,$BF,$FB,$78,$ED,$D3
	.dc.b	$68,$F9,$F0,$57,$A7,$F0,$87,$A3,$49,$87,$C2,$C5,$2F,$4B,$81,$A9
	.dc.b	$FA,$33,$AA,$CC,$7C,$AF,$25,$F0,$5F,$C3,$F9,$3F,$CA,$F1,$8F,$13
	.dc.b	$F8,$73,$6B,$85,$6D,$FE,$FA,$D8,$E7,$E2,$F0,$8C,$1F,$CE,$B8,$BF
	.dc.b	$C8,$F4,$FC,$23,$DE,$FC,$F6,$BF,$51,$B1,$C3,$E6,$5E,$A7,$CB,$6D
	.dc.b	$7F,$83,$5A,$FF,$B6,$9A,$FF,$83,$5A,$FF,$88,$7F,$08,$7B,$BC,$FD
	.dc.b	$79,$3F,$81,$9C,$BF,$90,$EC,$7C,$83,$E0,$BF,$98,$3A,$9F,$CC,$B7
	.dc.b	$3F,$EA,$0E,$E3,$97,$E6,$99,$E2,$FD,$63,$63,$F4,$CD,$C0,$BE,$2B
	.dc.b	$F3,$9F,$40,$F1,$A9,$2F,$F0,$90,$F9,$5D,$99,$8F,$C1,$C7,$17,$71
	.dc.b	$F2,$FE,$CB,$FA,$57,$17,$FA,$3E,$7E,$03,$5F,$D3,$F1,$27,$E9,$7D
	.dc.b	$27,$19,$FD,$44,$E5,$B1,$F5,$FC,$6D,$C6,$0D,$BF,$C2,$E3,$FA,$D7
	.dc.b	$C8,$7E,$F3,$7F,$E2,$0D,$AF,$E0,$13,$B7,$EA,$78,$A8,$C4,$7F,$0A
	.dc.b	$87,$D0,$7F,$65,$FC,$0D,$F2,$34,$95,$F7,$7F,$1F,$FF,$02,$B7,$DA
	.dc.b	$FE,$D9,$18,$8F,$E2,$AD,$E7,$FF,$06,$3F,$E6,$F7,$F3,$BF,$64,$F8
	.dc.b	$BF,$E0,$0D,$AF,$E1,$8D,$7F,$88,$F8,$BF,$E3,$9A,$FD,$2B,$FE,$3B
	.dc.b	$8F,$F3,$23,$5F,$F3,$C3,$5F,$F6,$6B,$54,$BE,$47,$F8,$84,$FF,$47
	.dc.b	$AE,$BE,$8F,$92,$7F,$C1,$1F,$90,$7F,$1A,$9F,$F1,$33,$78,$BF,$A0
	.dc.b	$37,$1F,$E7,$4F,$E2,$5F,$E6,$FF,$AB,$FF,$2E,$7F,$C1,$2D,$7F,$CF
	.dc.b	$9D,$A8,$C4,$7F,$A4,$63,$1E,$67,$FA,$17,$E8,$DF,$A7,$7A,$6B,$F6
	.dc.b	$1F,$67,$FC,$3E,$CF,$F1,$1A,$DE,$9F,$E2,$35,$FA,$77,$1B,$FE,$97
	.dc.b	$BB,$E5,$FF,$C4,$CD,$8F,$E9,$75,$C2,$FF,$C5,$5E,$9F,$F8,$CE,$7C
	.dc.b	$CD,$7F,$B1,$35,$FE,$B8,$DC,$9F,$83,$FE,$C6,$7F,$84,$35,$FE,$BA
	.dc.b	$DE,$1F,$F2,$A2,$F9,$FF,$F1,$D7,$E3,$7F,$8B,$E3,$FE,$E0,$DB,$FE
	.dc.b	$D4,$DC,$0B,$F1,$EF,$F2,$17,$CF,$BE,$5B,$FC,$7E,$DB,$FF,$1E,$B5
	.dc.b	$FA,$AD,$7F,$D8,$9B,$3E,$ED,$F3,$DF,$E4,$DF,$82,$7C,$A5,$3E,$84
	.dc.b	$EA,$FE,$5F,$C6,$71,$7F,$94,$7C,$6B,$F0,$EF,$F2,$B3,$63,$75,$FE
	.dc.b	$59,$FC,$9B,$FC,$11,$1F,$85,$6B,$FE,$11,$6B,$FE,$0B,$6F,$03,$F4
	.dc.b	$DF,$B9,$3F,$E0,$A6,$F2,$FF,$A0,$5B,$C7,$FE,$1B,$7F,$CB,$53,$FE
	.dc.b	$16,$DD,$F8,$87,$F0,$FC,$3F,$C3,$19,$7F,$9B,$5F,$65,$6E,$75,$F8
	.dc.b	$EF,$F4,$18,$FC,$E3,$5F,$F1,$43,$5F,$F1,$13,$65,$F8,$5F,$F0,$1C
	.dc.b	$BF,$C5,$CD,$AF,$F1,$43,$67,$7E,$3F,$F9,$FF,$4F,$F5,$59,$FE,$9D
	.dc.b	$2D,$FF,$8D,$DB,$7F,$E3,$1F,$C0,$9F,$47,$FD,$F3,$F3,$CF,$F5,$14
	.dc.b	$8F,$F0,$0B,$86,$FC,$07,$FB,$E7,$E0,$93,$99,$FF,$7E,$FE,$17,$FE
	.dc.b	$63,$7F,$C7,$F2,$8F,$E1,$0D,$F8,$AF,$F0,$5C,$6F,$26,$7F,$E1,$0C
	.dc.b	$BF,$D7,$58,$7F,$85,$5A,$F4,$6C,$F0,$86,$1E,$CF,$99,$FF,$0B,$FE
	.dc.b	$A8,$F5,$FD,$0D,$AF,$71,$29,$19,$8C,$9D,$1C,$3B,$FC,$E3,$FC,$6E
	.dc.b	$FF,$8A,$5F,$C4,$E1,$E3,$A4,$63,$2F,$F1,$5B,$C2,$BE,$2F,$E3,$BF
	.dc.b	$D0,$6F,$83,$6B,$FC,$62,$D9,$F6,$FE,$69,$FC,$6E,$BB,$3F,$C7,$63
	.dc.b	$D4,$D7,$FC,$72,$D7,$8F,$91,$FC,$B7,$DD,$FE,$C7,$1F,$E4,$1F,$2F
	.dc.b	$F9,$D5,$FA,$5B,$8F,$E0,$FA,$FF,$C5,$FE,$C7,$F9,$0B,$F3,$27,$D7
	.dc.b	$FF,$2F,$CB,$9E,$5F,$A9,$6E,$5F,$0E,$B7,$3F,$E4,$FF,$D0,$3F,$90
	.dc.b	$65,$FE,$4E,$6C,$7D,$A6,$F0,$7F,$7B,$FC,$47,$FC,$A3,$EF,$1A,$4A
	.dc.b	$FF,$2D,$47,$C7,$F9,$07,$E1,$7F,$D4,$73,$FF,$7D,$B7,$2F,$CF,$4E
	.dc.b	$BD,$7F,$F4,$37,$EF,$1E,$6C,$7F,$CC,$ED,$C3,$F1,$D4,$8C,$5F,$26
	.dc.b	$D2,$75,$C1,$D0,$FC,$5C,$6D,$AE,$E7,$E4,$BF,$E8,$76,$BE,$6C,$7A
	.dc.b	$D7,$B3,$83,$FD,$39,$18,$AF,$CC,$63,$E6,$6E,$3E,$5E,$8F,$69,$EE
	.dc.b	$F9,$8F,$F4,$97,$C5,$FF,$BE,$EF,$C6,$DC,$FC,$ED,$7F,$46,$D7,$F5
	.dc.b	$2F,$F4,$5F,$07,$8F,$8E,$69,$27,$FD,$51,$F4,$9F,$EA,$94,$D2,$3F
	.dc.b	$B0,$DB,$FA,$87,$D1,$7E,$F7,$1F,$A3,$82,$31,$A7,$DA,$87,$BF,$8C
	.dc.b	$3E,$94,$6E,$7F,$D4,$CF,$F5,$B6,$BB,$C3,$BC,$60,$DF,$D0,$7F,$A5
	.dc.b	$AB,$B3,$5E,$AD,$9F,$EC,$CD,$AF,$F2,$05,$78,$3E,$0B,$B4,$26,$C6
	.dc.b	$3E,$B2,$56,$CF,$3B,$EB,$E8,$FF,$1D,$CD,$F8,$3E,$B9,$7F,$2E,$7F
	.dc.b	$C7,$D7,$FE,$BB,$73,$AF,$F1,$35,$87,$13,$95,$DA,$FF,$61,$7E,$18
	.dc.b	$FA,$DF,$03,$63,$3F,$01,$89,$91,$FE,$28,$FC,$33,$FC,$66,$F2,$8F
	.dc.b	$A5,$E0,$A7,$3A,$7F,$44,$F7,$3F,$9F,$DF,$E9,$36,$FF,$0B,$3F,$35
	.dc.b	$AE,$63,$73,$FC,$D2,$5F,$85,$FF,$B9,$AB,$FD,$B4,$D9,$ED,$1F,$26
	.dc.b	$31,$BF,$F3,$5F,$C2,$BE,$3D,$7F,$56,$7E,$2D,$7F,$B8,$DC,$CE,$F7
	.dc.b	$FA,$51,$F4,$2B,$FA,$43,$5F,$F7,$3B,$5F,$EA,$6D,$C1,$F8,$4F,$F6
	.dc.b	$DC,$3F,$DE,$0D,$AF,$EC,$FF,$85,$7E,$6A,$73,$1F,$81,$FA,$0F,$F5
	.dc.b	$D4,$7F,$BD,$DF,$F2,$33,$7E,$4F,$5F,$0B,$FA,$1F,$5F,$FB,$D7,$CF
	.dc.b	$FE,$5D,$B5,$FC,$49,$B8,$7F,$9D,$04,$EF,$AB,$F1,$35,$FE,$03,$AF
	.dc.b	$E3,$9C,$FD,$67,$E2,$CE,$5D,$DA,$9F,$84,$FF,$71,$AF,$F7,$43,$65
	.dc.b	$C3,$FD,$E6,$B7,$E0,$FA,$6B,$53,$9E,$07,$CE,$7A,$DB,$8F,$AD,$B8
	.dc.b	$FF,$A3,$9A,$FF,$A3,$7F,$09,$E7,$3F,$D6,$FB,$FF,$C0,$EF,$FB,$EE
	.dc.b	$33,$2E,$9F,$E1,$77,$F2,$FE,$ED,$F0,$94,$C7,$15,$C7,$FC,$51,$A7
	.dc.b	$AF,$F0,$97,$D0,$7D,$16,$CF,$D1,$6E,$7F,$EA,$46,$BF,$EA,$3F,$C2
	.dc.b	$7F,$2F,$3F,$DC,$7E,$71,$9F,$DD,$9C,$D7,$9E,$73,$5E,$23,$EC,$7E
	.dc.b	$03,$EC,$D9,$F8,$4B,$E0,$FF,$0D,$39,$8C,$57,$1C,$3E,$D7,$EF,$1F
	.dc.b	$B0,$7F,$D6,$CD,$7F,$08,$D7,$71,$F8,$FF,$9A,$7C,$FF,$E4,$DB,$1D
	.dc.b	$7A,$7F,$DE,$3E,$B6,$6B,$F3,$1F,$85,$7F,$06,$E3,$FC,$23,$6F,$F2
	.dc.b	$4B,$EF,$F7,$F3,$3F,$9F,$FA,$1F,$F2,$DD,$7E,$75,$BA,$7F,$B4,$1A
	.dc.b	$FF,$B3,$DA,$FE,$77,$F0,$A7,$D3,$7F,$18,$FA,$A7,$9F,$FA,$B9,$CE
	.dc.b	$1F,$EC,$69,$CB,$EB,$FC,$25,$F6,$DF,$C8,$3F,$0A,$7C,$27,$EC,$E7
	.dc.b	$35,$FE,$CE,$9C,$BF,$0F,$ED,$BF,$C3,$8F,$ED,$5B,$A7,$FB,$A1,$AF
	.dc.b	$FB,$9D,$AF,$ED,$7C,$AF,$2B,$FA,$C9,$C6,$BF,$C0,$E7,$F1,$FC,$5F
	.dc.b	$AF,$76,$FF,$9A,$69,$F8,$57,$B1,$E4,$7F,$A0,$DF,$E0,$B7,$F8,$58
	.dc.b	$DE,$FF,$DB,$DE,$37,$7A,$BF,$8E,$47,$F9,$A9,$AF,$F9,$C1,$AF,$FA
	.dc.b	$3D,$A9,$FD,$AB,$E2,$31,$E2,$75,$7B,$3F,$C7,$FF,$09,$FC,$CC,$7A
	.dc.b	$3A,$51,$8C,$FF,$87,$2F,$B9,$E1,$DD,$C7,$79,$AF,$E5,$6B,$83,$FC
	.dc.b	$85,$47,$F2,$DF,$AB,$FB,$39,$77,$6E,$3F,$C6,$DF,$E2,$6D,$E3,$E1
	.dc.b	$B4,$C7,$F4,$55,$C1,$FE,$5C,$FC,$EB,$F8,$1B,$7F,$BC,$27,$E8,$36
	.dc.b	$3E,$AF,$9A,$FC,$25,$F5,$BF,$47,$E8,$7E,$97,$A5,$31,$E3,$17,$0E
	.dc.b	$B3,$F5,$EB,$1A,$FE,$0F,$2E,$78,$7F,$03,$CB,$B7,$DC,$FF,$89,$F3
	.dc.b	$7D,$AD,$8C,$32,$B8,$4F,$E8,$33,$F3,$BC,$23,$3D,$7F,$D5,$FE,$15
	.dc.b	$F4,$14,$C7,$8D,$6B,$F8,$44,$FE,$13,$5F,$E9,$1F,$70,$F9,$7E,$5D
	.dc.b	$7F,$D8,$63,$17,$F9,$7E,$DF,$F4,$BB,$F9,$9F,$C2,$5F,$73,$F9,$99
	.dc.b	$CE,$FF,$8F,$9C,$D7,$E7,$FE,$9F,$F9,$6C,$FF,$9E,$9A,$FF,$B1,$9A
	.dc.b	$FF,$B3,$1A,$FF,$BA,$DF,$F1,$53,$F9,$B6,$E7,$FB,$17,$AB,$7E,$13
	.dc.b	$FB,$D7,$DD,$FD,$F8,$FE,$CC,$DB,$FE,$E8,$FF,$81,$2F,$F4,$0E,$5F
	.dc.b	$EE,$4F,$F8,$19,$CF,$AD,$D3,$AB,$AB,$FE,$61,$FC,$25,$F2,$F7,$E0
	.dc.b	$30,$DC,$7F,$05,$3F,$D4,$3A,$9D,$5F,$E3,$4D,$3F,$D4,$B3,$98,$C5
	.dc.b	$79,$BD,$AF,$F3,$74,$7E,$76,$BD,$71,$8F,$03,$F8,$57,$03,$E5,$3F
	.dc.b	$CD,$DB,$A3,$18,$FF,$AA,$1B,$1F,$4E,$77,$1F,$6B,$0F,$99,$F2,$7D
	.dc.b	$27,$AB,$FB,$9F,$27,$F7,$5E,$D5,$C7,$F6,$6E,$1F,$DA,$F2,$5C,$BF
	.dc.b	$7C,$FD,$4F,$73,$EA,$FD,$AD,$AF,$EB,$9C,$C8,$CC,$63,$CD,$B8,$7D
	.dc.b	$67,$EE,$FA,$9F,$E3,$5B,$FE,$B9,$5F,$E3,$56,$D0,$87,$DC,$9F,$F4
	.dc.b	$94,$BA,$3F,$CB,$33,$FC,$E2,$31,$F0,$FD,$9F,$EF,$86,$0F,$3F,$83
	.dc.b	$F8,$E3,$F1,$FE,$AD,$8F,$5F,$87,$F9,$23,$17,$F8,$E7,$F2,$BE,$0F
	.dc.b	$A4,$93,$E9,$7D,$93,$F9,$A3,$71,$F1,$9F,$F2,$B7,$9F,$FD,$C9,$FF
	.dc.b	$24,$37,$D0,$FE,$6D,$F3,$DF,$E5,$98,$D5,$7B,$FE,$8B,$FD,$7C,$D3
	.dc.b	$FC,$19,$1F,$E0,$0E,$84,$7F,$9B,$F6,$57,$80,$91,$C3,$E2,$EE,$5F
	.dc.b	$D2,$70,$DC,$F4,$FC,$17,$38,$FF,$95,$E5,$1E,$AF,$99,$F3,$12,$BF
	.dc.b	$CD,$D3,$FD,$DA,$FD,$BF,$39,$F0,$0E,$13,$E8,$7F,$DA,$5F,$84,$78
	.dc.b	$27,$F9,$EB,$DA,$7F,$99,$39,$C7,$DB,$2F,$0D,$3B,$8E,$39,$FF,$B5
	.dc.b	$63,$FD,$5B,$1F,$EE,$D6,$BF,$8C,$9E,$33,$FE,$DB,$FC,$23,$FC,$0B
	.dc.b	$87,$E4,$65,$18,$AF,$05,$F4,$C3,$C7,$5A,$E8,$F5,$F0,$79,$BE,$5D
	.dc.b	$F8,$17,$3F,$BD,$FF,$50,$46,$32,$E7,$5B,$AD,$45,$3D,$D5,$FE,$A7
	.dc.b	$D3,$D0,$D8,$DC,$7F,$84,$31,$4E,$67,$E3,$69,$FE,$70,$C7,$C1,$5E
	.dc.b	$CE,$DE,$CF,$0A,$B5,$BC,$7F,$D4,$5F,$81,$FF,$85,$30,$F4,$73,$23
	.dc.b	$13,$F1,$EB,$85,$FF,$A8,$E7,$2F,$D7,$FC,$23,$EB,$61,$E2,$6D,$FF
	.dc.b	$AB,$F8,$FF,$DF,$B3,$9E,$5F,$2D,$6A,$F9,$71,$E2,$72,$CD,$FF,$84
	.dc.b	$5F,$9B,$0F,$E0,$AF,$79,$F9,$9A,$7F,$88,$FB,$3E,$76,$2F,$C2,$3F
	.dc.b	$AF,$DF,$D5,$8C,$78,$3A,$9C,$EB,$FC,$51,$FE,$BB,$5B,$6C,$67,$C3
	.dc.b	$76,$79,$23,$85,$F3,$9C,$E0,$C3,$94,$62,$3F,$90,$9E,$57,$0F,$5B
	.dc.b	$CB,$FC,$9F,$0F,$F3,$CC,$62,$3D,$9E,$3F,$E9,$DD,$CF,$FB,$83,$8F
	.dc.b	$4D,$8C,$CE,$DF,$0C,$7F,$9F,$1F,$51,$DE,$3A,$61,$F0,$5B,$97,$C5
	.dc.b	$F6,$FF,$9C,$CF,$C5,$F8,$FB,$39,$BE,$67,$D1,$AC,$BE,$B9,$B8,$BF
	.dc.b	$08,$7E,$0F,$AE,$0F,$F2,$E7,$CE,$5E,$33,$C2,$B7,$8F,$A3,$0E,$55
	.dc.b	$9C,$24,$E6,$BF,$CC,$CD,$9E,$44,$7E,$0E,$1F,$E6,$BF,$B3,$F1,$74
	.dc.b	$7D,$6D,$BF,$4A,$DD,$BF,$53,$7E,$AE,$BF,$B9,$5F,$B2,$6C,$FF,$05
	.dc.b	$E2,$D1,$F0,$3C,$EF,$F1,$F5,$FA,$5A,$FF,$A6,$76,$FF,$7A,$FB,$1F
	.dc.b	$D9,$BA,$7F,$8E,$56,$9B,$19,$AF,$E3,$92,$FC,$1F,$B4,$C9,$ED,$1E
	.dc.b	$7F,$1F,$FB,$8F,$85,$FC,$7B,$4F,$F2,$BD,$7F,$99,$5A,$FF,$9D,$30
	.dc.b	$FE,$14,$D8,$FC,$96,$49,$CC,$F9,$A7,$31,$88,$F2,$FD,$A7,$83,$3F
	.dc.b	$3F,$C4,$59,$FE,$81,$48,$C6,$1D,$6F,$5C,$5F,$E9,$FF,$A1,$BC,$86
	.dc.b	$98,$FA,$19,$97,$F5,$87,$F5,$15,$FE,$A7,$6D,$1A,$F4,$F9,$79,$5F
	.dc.b	$DC,$A3,$B9,$1C,$1F,$DD,$BA,$3D,$5A,$BF,$06,$FC,$BC,$FF,$85,$9A
	.dc.b	$FF,$86,$DB,$C1,$FD,$C7,$E2,$BD,$67,$90,$93,$FD,$C1,$F3,$FC,$4F
	.dc.b	$C0,$24,$7F,$81,$B9,$FE,$19,$EC,$23,$FD,$E1,$F7,$FD,$B7,$C8,$24
	.dc.b	$7F,$83,$B8,$7E,$59,$1C,$1F,$E1,$2F,$B7,$F6,$E4,$DA,$92,$7F,$C7
	.dc.b	$99,$FF,$A6,$BE,$0B,$EB,$FC,$1F,$FA,$52,$7F,$8D,$F1,$23,$11,$FC
	.dc.b	$8A,$31,$3F,$5A,$73,$3F,$E4,$69,$E1,$F4,$DD,$3E,$76,$4F,$83,$87
	.dc.b	$FC,$99,$5A,$8C,$07,$C1,$F2,$BE,$4B,$93,$A9,$CB,$7F,$01,$61,$F9
	.dc.b	$07,$CF,$9F,$F8,$5A,$91,$8C,$7C,$2F,$0B,$E5,$87,$1C,$24,$60,$F2
	.dc.b	$8F,$BB,$7D,$9F,$E8,$E3,$0D,$8C,$C7,$07,$FD,$AE,$DF,$8D,$FE,$88
	.dc.b	$F3,$6C,$6A,$3F,$A5,$B5,$F9,$9A,$BE,$63,$F5,$FF,$06,$F8,$8A,$7D
	.dc.b	$A8,$FD,$9F,$E0,$AF,$42,$F9,$F3,$1F,$F7,$64,$E5,$BE,$7F,$0A,$91
	.dc.b	$83,$EA,$F9,$3F,$E5,$AF,$1B,$F8,$13,$63,$3E,$EF,$F1,$0E,$6F,$F2
	.dc.b	$2C,$FF,$98,$3C,$AD,$CF,$F2,$89,$CB,$79,$BF,$D1,$D5,$BB,$E7,$E5
	.dc.b	$7C,$DF,$98,$FE,$C1,$0C,$3F,$CA,$2A,$33,$18,$6D,$FF,$A6,$17,$A3
	.dc.b	$CC,$FA,$2E,$2B,$FB,$B3,$5F,$EF,$6D,$7F,$C0,$3F,$90,$F8,$2F,$C3
	.dc.b	$F8,$2B,$DB,$FC,$13,$6B,$F0,$F5,$BC,$7F,$3F,$F8,$2E,$67,$D5,$1F
	.dc.b	$A2,$F9,$17,$BF,$FC,$A1,$B3,$F9,$FA,$E1,$97,$E6,$DF,$F5,$F4,$67
	.dc.b	$E2,$3F,$97,$C6,$23,$A9,$FA,$B7,$9B,$8F,$1D,$F4,$0F,$97,$FE,$72
	.dc.b	$BE,$5E,$98,$8C,$FC,$7D,$BF,$DB,$DF,$29,$CC,$7E,$9B,$85,$E3,$FD
	.dc.b	$9E,$D9,$39,$8F,$B8,$47,$07,$1F,$17,$F0,$04,$70,$78,$33,$FD,$36
	.dc.b	$DF,$BD,$F1,$BB,$CF,$F0,$3B,$FE,$04,$7F,$03,$FB,$09,$1F,$C1,$DA
	.dc.b	$FC,$6F,$23,$F8,$2E,$5F,$D6,$3F,$04,$7A,$34,$95,$FC,$46,$3F,$CA
	.dc.b	$99,$7F,$0D,$6F,$37,$F1,$C6,$CF,$F1,$B6,$E0,$5E,$8F,$F0,$9F,$07
	.dc.b	$FC,$4B,$5F,$60,$FD,$C7,$A5,$93,$5E,$74,$8F,$F2,$64,$FF,$3D,$6C
	.dc.b	$7F,$0C,$8C,$47,$F9,$53,$EB,$9F,$15,$FC,$5D,$7C,$55,$31,$1F,$DE
	.dc.b	$3F,$04,$FC,$15,$7F,$2F,$6B,$F1,$DB,$B7,$F2,$D5,$D3,$FE,$00,$FC
	.dc.b	$11,$F8,$AF,$F7,$A6,$DF,$F4,$C6,$BF,$D2,$DA,$FF,$49,$F8,$FF,$F1
	.dc.b	$23,$FD,$13,$E9,$9F,$82,$FA,$23,$0D,$7E,$D4,$F0,$7E,$E6,$CF,$C2
	.dc.b	$7F,$C3,$16,$FC,$1B,$86,$BF,$ED,$A8,$C6,$FF,$E1,$8B,$E1,$87,$F4
	.dc.b	$FF,$57,$A3,$FD,$75,$AF,$F5,$A6,$BF,$D6,$5B,$CC,$FC,$19,$F3,$BF
	.dc.b	$E2,$46,$D7,$F5,$B6,$BF,$93,$5F,$2B,$FE,$2A,$FC,$19,$EF,$D2,$57
	.dc.b	$F6,$38,$7C,$07,$F8,$C3,$E7,$1F,$27,$FE,$2A,$6D,$7F,$7C,$6B,$FD
	.dc.b	$ED,$AF,$F7,$99,$FF,$1B,$3F,$47,$E1,$7C,$6D,$C7,$FC,$7B,$3A,$8F
	.dc.b	$D5,$7C,$57,$F9,$0D,$FB,$1F,$15,$FE,$37,$6D,$FF,$91,$27,$51,$FB
	.dc.b	$1F,$C1,$BF,$B9,$3F,$E0,$D6,$BF,$E0,$B6,$BF,$E0,$86,$BF,$E0,$7F
	.dc.b	$C1,$9F,$47,$FE,$50,$6C,$7F,$81,$DA,$FE,$FD,$FB,$AF,$F0,$1E,$1F
	.dc.b	$D3,$F3,$7E,$0F,$F0,$6F,$1B,$7E,$1F,$FC,$B7,$3A,$F1,$FF,$C3,$3F
	.dc.b	$83,$7F,$2C,$C3,$FC,$40,$D7,$3E,$BF,$F9,$6A,$31,$1F,$E1,$DF,$C1
	.dc.b	$BF,$86,$BF,$E1,$B6,$BF,$C4,$DB,$8F,$F8,$65,$FF,$51,$C7,$F1,$7F
	.dc.b	$BC,$7C,$EF,$F9,$B9,$B1,$FE,$25,$6B,$FC,$76,$BF,$DB,$27,$F7,$4F
	.dc.b	$C1,$BF,$81,$C7,$FB,$63,$E0,$51,$88,$FF,$6F,$C6,$3C,$9F,$F6,$D3
	.dc.b	$63,$87,$E0,$DF,$C0,$B5,$FE,$F2,$FF,$02,$8F,$F0,$3B,$5F,$EA,$72
	.dc.b	$8E,$11,$FE,$F4,$CB,$FC,$04,$DF,$84,$FF,$06,$C6,$29,$F8,$33,$D8
	.dc.b	$FF,$08,$FE,$0D,$F7,$30,$1F,$E0,$A7,$0D,$E6,$F4,$C6,$21,$68,$FF
	.dc.b	$1B,$C7,$99,$FE,$9B,$9F,$3F,$8E,$EF,$F1,$FF,$E2,$0F,$C1,$7F,$84
	.dc.b	$CF,$F8,$61,$B7,$D1,$B8,$17,$DD,$F1,$3E,$5B,$89,$FA,$4D,$7E,$8B
	.dc.b	$5F,$69,$E8,$FC,$15,$F5,$5F,$C7,$3E,$39,$EE,$D3,$FC,$6B,$3A,$AF
	.dc.b	$F1,$CB,$5F,$A6,$DE,$6F,$F1,$BB,$73,$FF,$19,$BD,$27,$FC,$83,$EF
	.dc.b	$7E,$16,$BF,$C7,$AD,$AF,$F1,$CB,$67,$E6,$BE,$B9,$DE,$7F,$E4,$88
	.dc.b	$C5,$70,$8C,$79,$DF,$EC,$AF,$C0,$FC,$58,$3E,$11,$18,$6E,$BB,$D2
	.dc.b	$57,$F8,$52,$BD,$0F,$F8,$7D,$B9,$B2,$FC,$3E,$1D,$3E,$03,$C0,$7F
	.dc.b	$B9,$AB,$FC,$DD,$7F,$40,$DB,$87,$13,$C5,$D1,$E7,$8F,$CB,$3E,$3E
	.dc.b	$5F,$61,$ED,$F0,$64,$79,$DA,$F9,$FE,$57,$FD,$19,$F1,$3F,$EB,$9C
	.dc.b	$FA,$CA,$30,$F8,$1A,$F9,$DA,$FE,$21,$FF,$74,$77,$F2,$6C,$71,$F2
	.dc.b	$13,$96,$C7,$FA,$5B,$E7,$3F,$D2,$E9,$A4,$7C,$BE,$4F,$CD,$7F,$C3
	.dc.b	$41,$C6,$1F,$E7,$88,$FD,$8E,$FF,$EA,$C7,$4F,$FA,$BB,$E8,$FF,$D3
	.dc.b	$0D,$F5,$78,$BF,$4E,$ED,$7F,$00,$AD,$BE,$0F,$67,$B3,$FA,$86,$BF
	.dc.b	$C1,$9A,$9B,$F5,$3D,$23,$F8,$5B,$5D,$5F,$9E,$35,$4F,$95,$A6,$E2
	.dc.b	$CF,$FB,$23,$27,$CD,$F5,$6C,$67,$E1,$F1,$3D,$39,$E4,$F1,$8B,$97
	.dc.b	$F9,$17,$F1,$5F,$8E,$DD,$7D,$11,$F3,$23,$15,$F2,$F9,$3C,$E6,$AA
	.dc.b	$C7,$FB,$72,$BF,$C1,$EF,$E3,$9B,$5F,$9C,$6B,$FA,$36,$BF,$C5,$1F
	.dc.b	$F4,$54,$73,$F8,$CF,$1F,$AB,$E4,$7C,$BE,$1C,$9F,$07,$C1,$DC,$3D
	.dc.b	$9E,$2F,$D6,$C4,$D2,$57,$97,$E0,$3D,$E7,$C0,$F2,$2C,$7F,$BD,$BB
	.dc.b	$9F,$29,$F6,$DB,$19,$F3,$31,$39,$EE,$FC,$2D,$1A,$4A,$FD,$8F,$C0
	.dc.b	$3E,$87,$F0,$CD,$8F,$EA,$3C,$FF,$ED,$48,$7C,$F6,$AB,$7E,$3C,$BD
	.dc.b	$9F,$CF,$29,$D9,$9E,$1F,$F4,$17,$E8,$EE,$CF,$CC,$B6,$FF,$81,$B6
	.dc.b	$7F,$88,$B5,$FE,$30,$FF,$89,$23,$BC,$7F,$66,$8C,$3D,$B8,$BF,$41
	.dc.b	$18,$D3,$FD,$DB,$EA,$3A,$BF,$48,$DC,$39,$E0,$8F,$EE,$91,$87,$51
	.dc.b	$FE,$9B,$8F,$86,$7F,$1A,$2B,$FC,$A1,$AF,$F2,$F6,$BF,$D8,$3E,$FF
	.dc.b	$F3,$97,$FF,$00,$1B,$39,$9F,$DC,$46,$1F,$F0,$1B,$FB,$87,$FD,$58
	.dc.b	$D9,$E5,$F5,$9C,$3C,$1E,$39,$FF,$06,$AF,$C8,$E1,$FC,$C1,$B8,$7F
	.dc.b	$3E,$6B,$FD,$17,$C3,$7C,$CF,$F0,$96,$D7,$C6,$5E,$7E,$67,$5E,$14
	.dc.b	$AD,$79,$97,$0F,$F6,$17,$B1,$FC,$19,$FE,$9A,$D7,$FA,$A3,$5F,$F0
	.dc.b	$2B,$5F,$F1,$1B,$FE,$82,$9F,$AE,$D8,$EB,$F3,$5F,$C7,$E3,$1B,$FF
	.dc.b	$39,$9F,$F1,$29,$FC,$8F,$D3,$FF,$16,$DF,$2F,$5B,$13,$9F,$13,$F9
	.dc.b	$6D,$7F,$02,$8F,$F2,$64,$60,$FE,$63,$ED,$BE,$A3,$F9,$73,$70,$FC
	.dc.b	$5A,$FA,$EC,$CF,$F3,$3F,$C6,$5F,$53,$4C,$77,$DC,$3F,$27,$3F,$AA
	.dc.b	$7F,$B3,$B7,$2F,$EE,$AD,$7F,$BE,$35,$FF,$16,$FE,$1F,$FA,$19,$BE
	.dc.b	$37,$87,$BF,$FA,$24,$E6,$BF,$C9,$F6,$E2,$F3,$1E,$7F,$7A,$E3,$FA
	.dc.b	$6B,$FF,$69,$37,$C8,$CE,$1F,$EF,$F9,$CB,$FF,$00,$7E,$DB,$F3,$49
	.dc.b	$A7,$D2,$75,$C5,$E4,$FB,$39,$87,$ED,$2B,$8D,$A7,$FA,$E4,$7F,$07
	.dc.b	$E5,$FB,$CD,$AF,$F2,$6B,$FA,$56,$F8,$7C,$3E,$7A,$47,$58,$B9,$7A
	.dc.b	$7F,$0D,$F8,$A6,$30,$FC,$56,$48,$F0,$FC,$7C,$C3,$CD,$F1,$37,$0F
	.dc.b	$85,$F4,$BF,$38,$C6,$3C,$9F,$4F,$A5,$2F,$DD,$E9,$5B,$5F,$E6,$7E
	.dc.b	$D7,$FC,$7D,$5F,$C5,$9B,$3F,$36,$FF,$BF,$46,$2B,$E1,$F0,$7C,$63
	.dc.b	$E3,$F1,$5C,$3E,$66,$BF,$E0,$AE,$1F,$64,$7F,$8D,$1B,$97,$F9,$09
	.dc.b	$AF,$F9,$27,$F3,$BF,$E0,$33,$FC,$A1,$B1,$F7,$EB,$FC,$21,$18,$C7
	.dc.b	$F6,$5F,$1D,$FA,$5A,$F6,$DF,$FA,$0A,$31,$3A,$ED,$FF,$0C,$46,$27
	.dc.b	$F3,$7C,$9F,$98,$33,$DB,$63,$F9,$CC,$BD,$4E,$27,$2F,$16,$CF,$F3
	.dc.b	$EE,$AF,$23,$FE,$92,$6E,$7F,$EB,$0B,$E5,$3B,$F1,$BC,$7F,$59,$23
	.dc.b	$BF,$D6,$DB,$F5,$9C,$7E,$67,$FC,$6B,$81,$A4,$AF,$F3,$0E,$3F,$CB
	.dc.b	$2B,$89,$FB,$1C,$BE,$0D,$7F,$8E,$67,$5B,$FC,$AF,$84,$F8,$A7,$C6
	.dc.b	$F2,$AE,$3F,$D6,$BE,$6D,$F2,$5F,$E4,$16,$C7,$C6,$AC,$4F,$F9,$17
	.dc.b	$87,$FD,$29,$3F,$D8,$9B,$A7,$93,$5F,$F0,$54,$E7,$D9,$FF,$26,$74
	.dc.b	$3E,$9E,$9C,$E2,$CF,$F6,$CF,$D1,$3F,$9D,$60,$69,$3E,$03,$FC,$B7
	.dc.b	$F9,$B7,$C7,$AF,$EE,$6D,$AF,$8B,$5F,$D2,$78,$3F,$9E,$4E,$5F,$E7
	.dc.b	$D1,$FD,$C9,$BE,$E7,$F9,$86,$1F,$0B,$63,$FC,$C9,$4F,$AC,$A6,$71
	.dc.b	$F0,$BC,$79,$3F,$CD,$F8,$7F,$25,$9F,$F4,$CB,$72,$FF,$53,$B5,$FF
	.dc.b	$57,$7E,$90,$FA,$3F,$F3,$93,$6B,$E2,$5E,$27,$FC,$ED,$ED,$8F,$F3
	.dc.b	$E9,$67,$FD,$4E,$DB,$FF,$5A,$B5,$EF,$F1,$BD,$8E,$71,$67,$FC,$3A
	.dc.b	$FF,$25,$91,$DF,$CC,$F0,$FB,$9E,$27,$CF,$E7,$CB,$E3,$16,$BF,$2F
	.dc.b	$7F,$D9,$5F,$F6,$63,$5F,$F7,$EB,$5E,$ED,$5F,$87,$3E,$5B,$FD,$2C
	.dc.b	$DB,$F3,$FB,$18,$9F,$8B,$E6,$C9,$F5,$FD,$4D,$C3,$C7,$E7,$7C,$82
	.dc.b	$4F,$C8,$7F,$D4,$FF,$A2,$7F,$70,$8F,$F1,$FB,$67,$8C,$63,$D9,$FA
	.dc.b	$7E,$C4,$8F,$93,$F0,$16,$BE,$0C,$FF,$B1,$7D,$F7,$D9,$7C,$0F,$07
	.dc.b	$0F,$C2,$9B,$5F,$0A,$6B,$F7,$DB,$C3,$FD,$77,$1F,$AB,$87,$4A,$E1
	.dc.b	$18,$C3,$E6,$FA,$7C,$0E,$7F,$3B,$FA,$7B,$E1,$7F,$8B,$9C,$F9,$5D
	.dc.b	$BD,$3E,$FE,$B5,$CB,$A5,$FD,$68,$CC,$62,$1F,$6B,$FD,$43,$1E,$BA
	.dc.b	$2F,$D7,$39,$D3,$F1,$B7,$C3,$0F,$2F,$81,$A6,$9D,$6B,$FC,$4C,$DF
	.dc.b	$27,$9F,$A9,$D1,$91,$BF,$90,$FF,$1D,$E9,$EE,$BF,$E9,$8D,$8C,$E5
	.dc.b	$E2,$94,$7F,$90,$B2,$F2,$9E,$C7,$0E,$DA,$FE,$9A,$38,$FB,$9C,$23
	.dc.b	$1F,$20,$F2,$F1,$1F,$1A,$30,$EE,$BF,$DD,$93,$A6,$C6,$7D,$5F,$8F
	.dc.b	$97,$AE,$F8,$B6,$FE,$34,$EB,$2F,$67,$A7,$70,$F3,$4D,$7E,$BB,$71
	.dc.b	$FF,$3D,$C7,$CF,$F5,$B5,$F9,$C7,$9A,$19,$7E,$22,$30,$FC,$D9,$E5
	.dc.b	$F1,$BC,$D9,$5F,$66,$3B,$7D,$8E,$A3,$07,$DD,$7C,$5E,$C7,$83,$D0
	.dc.b	$93,$F8,$18,$F3,$7B,$3A,$8C,$1F,$85,$7D,$5D,$5A,$8C,$1F,$89,$72
	.dc.b	$FF,$AE,$B4,$FE,$E0,$D8,$FC,$9F,$A5,$FE,$83,$9F,$E8,$13,$FC,$0B
	.dc.b	$ED,$F3,$CD,$F9,$8F,$0D,$18,$9F,$E4,$F1,$87,$B7,$C4,$65,$FD,$2B
	.dc.b	$F8,$26,$9F,$7D,$C5,$9B,$ED,$F1,$83,$4C,$7B,$5D,$7C,$2F,$F4,$55
	.dc.b	$FF,$B5,$9A,$FF,$82,$1B,$9F,$B7,$E6,$3A,$1E,$02,$4F,$E4,$5F,$37
	.dc.b	$9A,$F7,$7D,$4D,$9F,$F7,$1B,$5F,$F0,$AF,$D2,$7F,$0C,$BF,$F5,$63
	.dc.b	$77,$FD,$FB,$67,$FA,$6B,$53,$FD,$8D,$F5,$78,$1E,$D2,$3F,$EF,$28
	.dc.b	$7C,$AF,$CD,$6C,$7E,$B9,$AF,$EC,$DA,$FE,$F1,$AF,$F1,$2F,$5F,$FD
	.dc.b	$5E,$FF,$BF,$9A,$FF,$8C,$1B,$87,$B3,$CC,$F0,$1F,$6E,$B7,$5D,$63
	.dc.b	$07,$DF,$AF,$71,$3B,$FD,$1C,$92,$BD,$8D,$7E,$5F,$C5,$7E,$26,$9F
	.dc.b	$11,$49,$3F,$DD,$B2,$FF,$6D,$7E,$78,$F9,$6F,$C8,$36,$7F,$88,$B5
	.dc.b	$FE,$2C,$D7,$F8,$FB,$5F,$E6,$BE,$73,$E2,$3F,$34,$D8,$F0,$70,$EA
	.dc.b	$7F,$3D,$F4,$DF,$E3,$F9,$C9,$FE,$0A,$70,$FF,$96,$9B,$8F,$FB,$C5
	.dc.b	$B5,$FD,$11,$7C,$4F,$EA,$7E,$1B,$E0,$DF,$F9,$67,$4C,$1E,$0F,$5A
	.dc.b	$47,$80,$5A,$33,$1E,$1F,$5F,$F6,$9F,$66,$F6,$E9,$27,$95,$BD,$7F
	.dc.b	$69,$FB,$97,$11,$FC,$F1,$AF,$F4,$16,$BF,$D2,$96,$FF,$18,$91,$DB
	.dc.b	$D7,$FE,$0D,$F9,$1B,$D1,$A4,$9F,$67,$D9,$38,$7F,$86,$36,$33,$18
	.dc.b	$33,$EA,$FF,$11,$FC,$57,$FB,$94,$FC,$56,$BF,$E8,$C6,$F2,$73,$85
	.dc.b	$BD,$2E,$E4,$8F,$1B,$E2,$F4,$FF,$8F,$7E,$02,$F8,$4A,$4A,$F4,$FA
	.dc.b	$3E,$83,$F9,$16,$06,$92,$3F,$B3,$C3,$D0,$F7,$92,$3E,$1F,$A1,$F6
	.dc.b	$BE,$02,$47,$E0,$79,$D1,$F1,$BC,$F9,$3F,$29,$AF,$FA,$B3,$0F,$E5
	.dc.b	$AF,$E2,$78,$DE,$9F,$D4,$24,$7D,$2F,$96,$FA,$24,$70,$7D,$6F,$9A
	.dc.b	$FC,$C7,$95,$23,$FD,$60,$FF,$43,$F6,$FF,$D6,$31,$F9,$76,$BF,$EB
	.dc.b	$E6,$F2,$7D,$36,$F3,$FF,$8A,$1F,$E5,$04,$F2,$7A,$1D,$7E,$1B,$F5
	.dc.b	$1F,$35,$30,$FF,$60,$3E,$1F,$B1,$91,$E1,$6B,$E1,$FB,$1F,$DF,$FC
	.dc.b	$F4,$8F,$D6,$78,$08,$F5,$79,$F2,$7D,$4D,$7D,$5E,$3F,$F6,$1F,$8C
	.dc.b	$FB,$51,$FE,$B8,$EA,$C1,$F0,$FE,$26,$47,$F8,$9B,$74,$7F,$67,$6C
	.dc.b	$66,$31,$DB,$F2,$C9,$1F,$2F,$D7,$FE,$E1,$F8,$1F,$E6,$63,$F8,$23
	.dc.b	$5F,$F7,$5B,$79,$3F,$40,$FF,$88,$A3,$3E,$C7,$F7,$78,$34,$91,$FE
	.dc.b	$3C,$FC,$55,$C3,$FD,$FD,$B1,$FC,$4D,$AF,$FB,$E9,$FF,$67,$B9,$F9
	.dc.b	$9E,$04,$78,$70,$7F,$82,$7E,$D7,$F9,$A4,$7F,$AD,$1A,$9A,$4F,$3B
	.dc.b	$81,$23,$F9,$24,$3E,$33,$FC,$29,$D7,$FE,$04,$EB,$FF,$39,$B6,$BF
	.dc.b	$D8,$2D,$7F,$B7,$FB,$5C,$3E,$23,$FC,$35,$3F,$E2,$69,$FF,$30,$35
	.dc.b	$FF,$34,$B5,$FF,$3F,$FD,$E7,$F7,$D5,$C3,$FC,$3E,$30,$DE,$DF,$F1
	.dc.b	$44,$7F,$7A,$8D,$76,$60,$95,$FC,$FE,$7E,$6C,$F3,$F6,$FF,$C5,$D1
	.dc.b	$FE,$33,$9F,$F3,$A3,$5F,$F3,$EB,$5F,$F4,$9F,$E0,$9F,$D7,$5F,$2E
	.dc.b	$98,$F6,$3F,$C7,$53,$9C,$3F,$93,$3F,$A3,$AF,$EA,$1F,$8B,$7F,$93
	.dc.b	$1E,$27,$FB,$A9,$C4,$FF,$2D,$6D,$FF,$A8,$9F,$95,$1F,$CC,$7C,$FF
	.dc.b	$F5,$AD,$7F,$A4,$9A,$FF,$AA,$1A,$FF,$BD,$9A,$FF,$37,$F9,$2F,$F2
	.dc.b	$84,$1A,$63,$B3,$3D,$72,$BF,$B4,$52,$38,$EB,$FE,$59,$FD,$03,$FB
	.dc.b	$9D,$7F,$A6,$9A,$FF,$A7,$DA,$FF,$AA,$1A,$FF,$AF,$9F,$E9,$B9,$DF
	.dc.b	$8E,$E8,$7F,$99,$7F,$14,$FE,$A1,$A7,$F5,$15,$E8,$64,$98,$7F,$7D
	.dc.b	$8F,$92,$83,$F8,$F7,$0D,$73,$CB,$FE,$71,$8C,$1F,$E6,$F8,$E9,$E7
	.dc.b	$78,$CF,$27,$C3,$79,$7C,$09,$3E,$17,$D1,$F5,$32,$3F,$D7,$B6,$8E
	.dc.b	$C4,$8F,$F6,$0F,$E5,$CF,$6B,$FD,$08,$D8,$FF,$0C,$35,$FC,$48,$FB
	.dc.b	$D7,$CD,$7F,$A2,$DB,$1F,$E1,$F6,$BC,$1B,$B7,$F7,$B5,$D7,$FE,$93
	.dc.b	$F2,$FC,$D5,$FE,$27,$6B,$E0,$6D,$7F,$80,$FE,$27,$FC,$90,$DF,$85
	.dc.b	$FF,$4E,$46,$39,$3F,$DF,$6D,$BF,$A8,$BE,$53,$C1,$C9,$27,$C0,$FF
	.dc.b	$A9,$7E,$AF,$FD,$04,$4F,$4F,$06,$7C,$29,$5F,$EE,$A9,$FE,$14,$FF
	.dc.b	$BD,$9A,$FE,$0C,$38,$7F,$0D,$A7,$DF,$7D,$DC,$DF,$0C,$F1,$70,$48
	.dc.b	$FF,$25,$57,$CA,$D3,$A3,$63,$F2,$0F,$CC,$77,$E5,$FF,$1C,$7C,$46
	.dc.b	$9C,$3E,$76,$BD,$1A,$F4,$F9,$07,$D5,$FB,$DB,$1D,$0F,$6B,$87,$E1
	.dc.b	$35,$D7,$C9,$FF,$B3,$DF,$F7,$C9,$FE,$D1,$70,$7F,$B4,$FE,$5B,$FC
	.dc.b	$81,$1E,$86,$BF,$2D,$AF,$EB,$DA,$FF,$A3,$F1,$FF,$6E,$71,$60,$8F
	.dc.b	$FB,$7F,$87,$64,$70,$7F,$B8,$FE,$6B,$99,$1C,$1F,$EE,$7F,$88,$EE
	.dc.b	$47,$07,$C3,$BF,$E9,$7E,$57,$9B,$9B,$FD,$8B,$39,$6C,$FC,$96,$BF
	.dc.b	$59,$FD,$6C,$6F,$83,$91,$1E,$DF,$56,$D3,$E1,$B6,$33,$18,$33,$A4
	.dc.b	$8E,$2D,$C3,$ED,$57,$FA,$B6,$7A,$F9,$5D,$08,$F8,$3D,$8E,$E4,$70
	.dc.b	$79,$3D,$0F,$19,$1C,$1E,$8F,$3F,$CE,$47,$07,$B3,$E2,$7D,$64,$70
	.dc.b	$7C,$DF,$98,$7C,$BF,$E0,$9B,$19,$8C,$19,$F7,$E4,$71,$6E,$9F,$99
	.dc.b	$FA,$B7,$CF,$7E,$31,$B5,$F6,$DB,$3B,$9D,$76,$CA,$D3,$67,$F4,$9F
	.dc.b	$56,$E7,$FC,$BB,$6B,$F0,$CD,$9D,$CE,$BE,$76,$56,$9B,$3F,$AF,$FA
	.dc.b	$BF,$E7,$93,$F8,$F6,$C6,$7C,$9F,$D2,$C6,$23,$F8,$0C,$BB,$7F,$52
	.dc.b	$D3,$F3,$05,$8F,$DE,$35,$FE,$13,$F5,$7F,$D7,$63,$F3,$CD,$8F,$E0
	.dc.b	$4D,$76,$7E,$D9,$FE,$61,$1C,$3B,$70,$47,$9F,$1E,$C8,$E0,$FD,$67
	.dc.b	$77,$FB,$05,$BE,$9F,$F0,$28,$C6,$1F,$C3,$A3,$1E,$1C,$AE,$2D,$9F
	.dc.b	$E3,$1F,$57,$FB,$07,$F9,$B3,$5F,$EF,$4D,$4D,$3F,$D9,$93,$FB,$F6
	.dc.b	$C7,$4F,$80,$FE,$2B,$18,$8F,$E6,$2B,$E6,$32,$4A,$E4,$FF,$1C,$EF
	.dc.b	$FF,$4E,$47,$F0,$B6,$D7,$F2,$F6,$BF,$C9,$7B,$3F,$91,$FD,$5B,$E4
	.dc.b	$A9,$2F,$F8,$A4,$65,$FE,$2D,$F5,$6F,$82,$FE,$56,$D8,$CC,$60,$CF
	.dc.b	$C1,$48,$E2,$DC,$7F,$A1,$53,$AB,$99,$2B,$BB,$DB,$DC,$91,$E2,$6D
	.dc.b	$72,$DF,$C6,$49,$F3,$BF,$D0,$3E,$A7,$F1,$D3,$FD,$B5,$AF,$F8,$C5
	.dc.b	$A9,$A7,$F2,$57,$FA,$C3,$5F,$EC,$AD,$7A,$6D,$DF,$AE,$5F,$F3,$DA
	.dc.b	$FD,$C5,$70,$9F,$27,$95,$2B,$4D,$9E,$FE,$8F,$A4,$93,$CD,$F6,$7C
	.dc.b	$CC,$8E,$CD,$9F,$3F,$8F,$F0,$49,$3E,$37,$E2,$78,$F2,$3C,$CD,$9F
	.dc.b	$7F,$57,$C9,$24,$FA,$DF,$99,$F1,$52,$3D,$CD,$9F,$8F,$EB,$FD,$12
	.dc.b	$4F,$C3,$7E,$A7,$D1,$C8,$F8,$CD,$9F,$F0,$37,$D3,$FE,$45,$BD,$1F
	.dc.b	$DE,$A3,$11,$FE,$16,$6B,$FE,$29,$5F,$0D,$B2,$5F,$F6,$F8,$CB,$7E
	.dc.b	$5B,$A1,$1C,$1B,$F2,$FC,$04,$70,$73,$F8,$CF,$21,$1C,$1D,$FD,$9F
	.dc.b	$41,$1C,$1E,$3F,$A3,$F6,$11,$C1,$E7,$F8,$0F,$80,$47,$07,$AF,$E8
	.dc.b	$7E,$21,$1C,$1E,$FF,$91,$F9,$04,$70,$7F,$84,$74,$7A,$5F,$E2,$66
	.dc.b	$C6,$63,$07,$F8,$AB,$E9,$BF,$8F,$13,$D9,$C1,$9F,$62,$57,$F8,$76
	.dc.b	$7F,$C2,$BC,$3F,$E4,$16,$BF,$E4,$E7,$FA,$43,$FE,$61,$6B,$FE,$A5
	.dc.b	$6A,$69,$FD,$B9,$E5,$E9,$F9,$49,$3E,$1F,$07,$D0,$47,$07,$97,$C1
	.dc.b	$F6,$11,$C1,$FE,$35,$F3,$7F,$C8,$D7,$FE,$59,$6B,$FE,$66,$6B,$FE
	.dc.b	$79,$6B,$FE,$BD,$7F,$AD,$BC,$5F,$F2,$99,$87,$93,$8F,$02,$4F,$85
	.dc.b	$B3,$E1,$F0,$38,$92,$7C,$CF,$4F,$22,$47,$A5,$B3,$E9,$F9,$6F,$09
	.dc.b	$27,$DC,$F9,$3E,$76,$47,$C1,$6C,$FC,$1F,$03,$D2,$49,$F8,$CF,$B3
	.dc.b	$C0,$91,$F2,$5B,$3F,$27,$D4,$F8,$24,$9F,$9C,$FC,$4F,$62,$47,$D1
	.dc.b	$6C,$FD,$1F,$09,$3F,$CF,$36,$46,$27,$FD,$07,$18,$9D,$4E,$5F,$B1
	.dc.b	$E1,$5A,$FB,$0D,$7E,$C3,$5F,$F6,$13,$69,$E1,$52,$57,$8D,$D7,$C3
	.dc.b	$C8,$F3,$36,$FF,$D2,$FF,$47,$EA,$9F,$EA,$AD,$8C,$C7,$86,$30,$FF
	.dc.b	$AD,$1F,$E4,$8D,$F1,$34,$47,$9F,$5F,$12,$38,$3B,$FC,$9F,$FA,$0E
	.dc.b	$FF,$D6,$96,$8C,$47,$76,$BF,$ED,$58,$D7,$95,$FE,$99,$5F,$3B,$FC
	.dc.b	$2A,$79,$B5,$F4,$35,$E4,$D4,$21,$EA,$7F,$B8,$F0,$F8,$0D,$9F,$13
	.dc.b	$5E,$33,$C1,$F9,$7E,$3E,$5E,$8D,$83,$4C,$78,$F2,$7E,$9B,$5F,$AF
	.dc.b	$F2,$FC,$48,$FD,$87,$A7,$CD,$C8,$FB,$6D,$7E,$FF,$93,$E1,$23,$F8
	.dc.b	$07,$C9,$E4,$C8,$FC,$1B,$5F,$C3,$F9,$3F,$EC,$77,$E3,$37,$23,$4C
	.dc.b	$7C,$64,$9F,$AC,$D7,$EF,$FA,$FC,$48,$FD,$B7,$A7,$97,$23,$EF,$35
	.dc.b	$FC,$3F,$97,$E1,$23,$F8,$37,$C9,$F1,$92,$3F,$0C,$D7,$F1,$FE,$EF
	.dc.b	$A4,$8F,$E2,$DF,$77,$D1,$3E,$26,$92,$7F,$DD,$11,$96,$FC,$4F,$02
	.dc.b	$38,$37,$E7,$74,$23,$83,$9F,$9D,$E0,$23,$83,$BF,$B1,$E4,$23,$83
	.dc.b	$C7,$EB,$FA,$08,$E0,$F3,$FC,$7F,$B0,$8E,$0F,$5F,$B9,$F0,$08,$E0
	.dc.b	$FD,$29,$07,$D0,$BE,$1F,$F0,$8D,$8D,$1F,$AC,$70,$E1,$AF,$35,$E2
	.dc.b	$F2,$24,$72,$3B,$BF,$91,$FA,$07,$D3,$D2,$4E,$E3,$2D,$F6,$F8,$11
	.dc.b	$C1,$BF,$3F,$A1,$1C,$1C,$FC,$FF,$01,$1C,$1D,$FC,$DF,$21,$1C,$1E
	.dc.b	$3F,$37,$D0,$47,$07,$9F,$CD,$F6,$11,$C1,$F0,$7E,$7B,$FA,$63,$F3
	.dc.b	$1A,$FD,$36,$BF,$59,$AF,$F1,$C7,$FD,$5C,$F3,$F3,$3F,$75,$18,$8E
	.dc.b	$91,$8F,$1E,$47,$85,$AF,$87,$C7,$E2,$47,$CC,$F4,$F1,$64,$7A,$5A
	.dc.b	$FA,$7C,$5F,$09,$1F,$73,$E4,$F1,$64,$7C,$16,$BF,$07,$C2,$F4,$91
	.dc.b	$F8,$CF,$B3,$C2,$91,$F2,$5A,$FC,$9F,$89,$F8,$24,$7E,$73,$F1,$3D
	.dc.b	$B9,$1F,$45,$AF,$E7,$3C,$D7,$D4,$7F,$23,$6C,$7E,$5D,$AE,$E3,$5E
	.dc.b	$1C,$8D,$35,$E9,$E1,$73,$23,$CD,$ED,$F2,$12,$3B,$35,$F2,$7C,$7F
	.dc.b	$8C,$8F,$8D,$F3,$78,$52,$3C,$CD,$7D,$9E,$17,$AC,$8F,$AD,$F7,$7B
	.dc.b	$72,$3D,$CD,$7E,$27,$B7,$F0,$C8,$FC,$37,$E3,$7C,$AC,$8F,$8C,$D7
	.dc.b	$F8,$3F,$CE,$7F,$64,$7F,$80,$35,$CF,$37,$F5,$18,$C4,$7F,$13,$5E
	.dc.b	$56,$49,$1C,$9F,$EB,$1F,$38,$F8,$CA,$49,$FE,$1D,$19,$6E,$9C,$08
	.dc.b	$E0,$FE,$29,$F3,$9F,$C6,$1B,$E3,$7F,$65,$8C,$47,$F2,$26,$BF,$CA
	.dc.b	$57,$B9,$B2,$47,$47,$FB,$77,$9B,$FE,$CA,$8F,$E4,$4D,$7F,$96,$B5
	.dc.b	$FE,$68,$D7,$F9,$BA,$EA,$E6,$48,$E6,$FF,$7B,$F9,$C4,$D2,$4F,$F3
	.dc.b	$18,$CB,$78,$F8,$11,$C1,$BF,$33,$A1,$1C,$1F,$CE,$FE,$71,$F0,$1F
	.dc.b	$E0,$B6,$C6,$63,$07,$F8,$3F,$E7,$1E,$AD,$24,$FF,$45,$8C,$BF,$D1
	.dc.b	$FC,$47,$5F,$F8,$5D,$B1,$98,$C1,$9E,$99,$1C,$5A,$FF,$5A,$97,$A3
	.dc.b	$CC,$91,$DD,$FF,$12,$7C,$DB,$E2,$69,$27,$FA,$D4,$65,$FE,$B9,$F3
	.dc.b	$7F,$CD,$93,$49,$EE,$FF,$8B,$DF,$F1,$6C,$70,$F6,$B0,$47,$9F,$26
	.dc.b	$C8,$E0,$EF,$E3,$F3,$23,$83,$C7,$D9,$DC,$8E,$0F,$3F,$8F,$E3,$23
	.dc.b	$83,$FB,$B7,$CD,$3D,$8F,$F2,$33,$63,$31,$83,$3E,$2C,$8E,$2D,$7F
	.dc.b	$C1,$3F,$19,$FE,$06,$7F,$C2,$2D,$7F,$CE,$ED,$4D,$3F,$5E,$F0,$EE
	.dc.b	$FF,$2C,$46,$23,$8C,$7F,$96,$FE,$6B,$FC,$75,$1F,$E1,$86,$BF,$E1
	.dc.b	$F6,$BF,$E2,$26,$BF,$E8,$87,$F1,$AF,$1E,$5E,$84,$7F,$C3,$51,$FA
	.dc.b	$A7,$B7,$6F,$84,$8E,$B1,$F2,$11,$C1,$CB,$1F,$41,$1C,$1D,$BB,$1F
	.dc.b	$33,$49,$1E,$D8,$CB,$EF,$F0,$DF,$35,$E6,$6C,$66,$30,$7F,$A0,$57
	.dc.b	$CA,$D2,$3F,$E8,$56,$FE,$45,$F5,$23,$FC,$93,$F0,$DF,$EE,$58,$C4
	.dc.b	$7F,$93,$27,$FD,$63,$18,$84,$EF,$83,$FC,$97,$5A,$7F,$CA,$51,$FE
	.dc.b	$EF,$CB,$FC,$43,$39,$AE,$FD,$1F,$EB,$C9,$FF,$14,$F9,$8F,$85,$FF
	.dc.b	$12,$B6,$7F,$C9,$CD,$AF,$F2,$CB,$5F,$F2,$DF,$1F,$79,$FF,$31,$36
	.dc.b	$3F,$D9,$33,$C6,$FF,$E0,$0D,$78,$4D,$96,$7F,$CD,$7F,$1B,$C5,$FF
	.dc.b	$68,$DD,$51,$D7,$27,$D2,$3C,$1E,$19,$7F,$AC,$31,$E9,$5A,$9C,$F6
	.dc.b	$49,$F0,$63,$F8,$4E,$5C,$3F,$86,$8F,$F3,$EE,$9F,$D0,$9E,$7F,$8B
	.dc.b	$FF,$BB,$1E,$39,$3E,$17,$C9,$39,$BE,$47,$C2,$8F,$29,$ED,$F9,$DF
	.dc.b	$B7,$7C,$A3,$1A,$67,$E0,$7F,$CA,$F4,$69,$2B,$FD,$05,$1F,$CB,$5F
	.dc.b	$5B,$6F,$C2,$1E,$51,$AF,$0F,$C5,$F0,$32,$3E,$1F,$91,$FE,$BD,$79
	.dc.b	$C7,$9F,$E0,$7C,$64,$8F,$0B,$E6,$F3,$BD,$9F,$1D,$ED,$6C,$FC,$EF
	.dc.b	$93,$FF,$7E,$B7,$D5,$FB,$5F,$1C,$C7,$F0,$15,$AD,$7D,$2F,$F8,$6A
	.dc.b	$FF,$D4,$0D,$8F,$F5,$5B,$5F,$F5,$93,$50,$F9,$84,$7E,$1E,$7F,$03
	.dc.b	$F3,$BF,$E2,$17,$F2,$53,$AE,$8F,$C9,$4E,$6F,$E5,$AF,$5B,$27,$E7
	.dc.b	$E9,$F3,$6F,$7F,$38,$FE,$3B,$E7,$7F,$D7,$6F,$FA,$E5,$B8,$FF,$B1
	.dc.b	$1A,$FF,$B2,$DA,$FA,$B1,$FE,$33,$E9,$FE,$96,$7F,$C4,$F5,$FA,$99
	.dc.b	$CB,$DF,$83,$F5,$5F,$3B,$F9,$58,$FC,$FB,$70,$FC,$AE,$6F,$85,$FF
	.dc.b	$08,$D6,$BA,$BF,$64,$BC,$9F,$F5,$4B,$4F,$F6,$91,$6F,$FD,$BA,$D7
	.dc.b	$FD,$CB,$6F,$CE,$13,$D3,$DA,$F9,$11,$88,$FD,$98,$8F,$E0,$5C,$1F
	.dc.b	$EE,$B8,$FD,$BB,$6F,$D3,$F3,$6F,$4B,$7F,$1F,$C3,$AF,$FD,$7F,$66
	.dc.b	$93,$0F,$F7,$9E,$7F,$EE,$EF,$53,$F8,$87,$97,$FB,$9F,$17,$CF,$A7
	.dc.b	$E2,$7C,$AF,$C8,$47,$FB,$42,$75,$87,$F3,$7F,$03,$89,$F8,$3F,$C3
	.dc.b	$FF,$B8,$C7,$F0,$E9,$FF,$1B,$BF,$C4,$3E,$7F,$FC,$66,$FF,$BE,$DB
	.dc.b	$8F,$E6,$08,$E1,$C0,$70,$FE,$7B,$37,$E1,$EF,$91,$FE,$55,$5F,$9F
	.dc.b	$AF,$BF,$39,$BF,$17,$7E,$BD,$D9,$1F,$A2,$FC,$3D,$F2,$18,$C3,$F8
	.dc.b	$5F,$97,$FC,$35,$FE,$9F,$E4,$FF,$BB,$23,$11,$FD,$72,$3F,$DE,$3E
	.dc.b	$2B,$F4,$F7,$C7,$7F,$3D,$E0,$FC,$05,$FF,$26,$AD,$4E,$73,$FE,$85
	.dc.b	$F8,$7E,$B9,$7F,$DF,$ED,$8F,$D2,$75,$7C,$89,$CC,$7F,$49,$FA,$27
	.dc.b	$D1,$FF,$C0,$06,$C7,$CA,$CB,$F8,$2C,$61,$C3,$5F,$9E,$FE,$FE,$3F
	.dc.b	$A3,$B6,$3F,$96,$BC,$6B,$F0,$F0,$8D,$47,$A3,$C3,$F5,$FC,$EF,$B1
	.dc.b	$B7,$FD,$4B,$24,$FA,$79,$33,$8F,$67,$E1,$FC,$8F,$C4,$F9,$DB,$87
	.dc.b	$17,$0F,$F5,$CF,$0B,$F9,$0B,$93,$17,$F3,$3D,$0F,$9B,$E6,$59,$F1
	.dc.b	$38,$79,$7B,$36,$BF,$A5,$DF,$F7,$29,$ED,$F4,$1F,$DD,$E0,$DA,$C3
	.dc.b	$F6,$BE,$7A,$57,$BF,$DC,$B8,$74,$F5,$F9,$18,$E2,$F9,$79,$7F,$58
	.dc.b	$9F,$EB,$7C,$3F,$8A,$C3,$E4,$73,$7E,$1B,$E5,$BF,$0E,$F8,$FC,$89
	.dc.b	$3A,$3E,$23,$F8,$CF,$96,$91,$F4,$7D,$DB,$8F,$BB,$C0,$FB,$6F,$EC
	.dc.b	$F3,$FD,$1D,$7E,$26,$11,$87,$F1,$8D,$7F,$6E,$BE,$63,$E3,$1F,$1E
	.dc.b	$AD,$7E,$07,$E2,$BF,$C4,$1F,$65,$FD,$26,$30,$7F,$85,$9C,$61,$FD
	.dc.b	$F1,$BB,$FE,$43,$E9,$3F,$9D,$56,$FD,$9F,$F1,$54,$62,$FF,$C0,$7F
	.dc.b	$66,$FB,$2F,$F1,$73,$6F,$FC,$47,$F3,$AF,$A3,$CD,$72,$EC,$95,$FE
	.dc.b	$28,$BE,$18,$7F,$83,$B1,$FE,$F1,$9F,$ED,$BD,$5F,$DC,$9F,$E3,$BF
	.dc.b	$C9,$3F,$D8,$F8,$7E,$7E,$3C,$BC,$DF,$3B,$0F,$EE,$F8,$7D,$D7,$F6
	.dc.b	$CD,$7F,$90,$B5,$FE,$4C,$B9,$7D,$E7,$4B,$FF,$1E,$C6,$5F,$47,$4F
	.dc.b	$C0,$3F,$C9,$DF,$5B,$FE,$27,$CF,$F7,$3A,$7E,$3D,$F9,$9F,$25,$F2
	.dc.b	$8E,$17,$D1,$FF,$2B,$7B,$9F,$CB,$67,$F8,$0B,$67,$F9,$A7,$E1,$2A
	.dc.b	$DF,$8B,$F5,$39,$65,$7F,$94,$AF,$86,$DF,$0A,$7F,$CC,$5F,$59,$FE
	.dc.b	$6F,$8F,$E8,$8D,$7F,$C3,$AD,$4D,$FA,$AF,$F2,$C4,$7F,$90,$9B,$9F
	.dc.b	$F1,$69,$FF,$3A,$61,$FD,$3E,$72,$FA,$3A,$7C,$3F,$4F,$FE,$1D,$D5
	.dc.b	$18,$9F,$F3,$4D,$AB,$FC,$F1,$7F,$E2,$7C,$3B,$46,$3E,$47,$FC,$F9
	.dc.b	$F6,$3F,$E3,$2C,$3F,$54,$DC,$FF,$9A,$35,$FE,$BE,$D7,$CD,$B7,$63
	.dc.b	$9F,$A1,$6B,$FC,$F7,$5F,$E3,$7A,$E3,$7F,$8D,$51,$97,$FB,$73,$5F
	.dc.b	$F4,$7F,$A1,$FE,$91,$F7,$7F,$C1,$D9,$FF,$3C,$AF,$66,$1F,$DD,$67
	.dc.b	$2F,$DE,$F7,$7D,$BE,$C7,$C7,$6C,$79,$23,$A7,$99,$8C,$BF,$CE,$B7
	.dc.b	$FC,$AD,$FF,$3B,$FD,$97,$F9,$01,$FF,$2C,$B5,$34,$C4,$7F,$8E,$23
	.dc.b	$9F,$CE,$7F,$AA,$67,$FC,$9B,$9F,$F5,$4A,$D3,$FD,$5A,$1F,$19,$B3
	.dc.b	$FD,$71,$BF,$FB,$0F,$6C,$69,$FE,$A6,$C3,$FC,$EC,$FF,$21,$6E,$5F
	.dc.b	$EC,$2F,$8B,$F9,$D1,$88,$FF,$61,$F9,$7F,$2C,$F9,$93,$6B,$FD,$63
	.dc.b	$F6,$6F,$B0,$FF,$60,$36,$3F,$D6,$BA,$3E,$3B,$35,$F5,$75,$EE,$4A
	.dc.b	$F2,$3E,$2F,$27,$FC,$73,$F0,$6F,$A6,$FB,$1E,$6E,$1F,$B5,$37,$6F
	.dc.b	$F6,$7C,$D5,$BF,$F3,$DB,$E5,$1F,$ED,$4F,$3B,$53,$9C,$3F,$09,$1C
	.dc.b	$7E,$97,$F8,$85,$2F,$8B,$D3,$CE,$D7,$93,$C9,$E7,$BC,$DB,$FD,$33
	.dc.b	$7C,$33,$FE,$69,$18,$CF,$F2,$55,$E7,$FA,$CF,$F1,$5D,$F6,$6C,$EA
	.dc.b	$B8,$65,$FE,$A4,$9C,$B7,$DC,$E4,$7A,$66,$9C,$21,$3F,$E1,$68,$C3
	.dc.b	$63,$DD,$5F,$E5,$98,$FE,$A1,$7E,$3F,$73,$CD,$EE,$3E,$0B,$FB,$55
	.dc.b	$23,$18,$7F,$8D,$63,$13,$A9,$CC,$FF,$98,$A5,$EE,$FF,$AF,$23,$83
	.dc.b	$5D,$35,$F2,$35,$F8,$5E,$EF,$F9,$BF,$1F,$90,$DC,$BD,$4D,$9F,$91
	.dc.b	$EA,$57,$E1,$B7,$03,$4D,$71,$7F,$98,$27,$E3,$FB,$7B,$23,$F2,$BD
	.dc.b	$CE,$44,$70,$7C,$EF,$6F,$C1,$F6,$8F,$9C,$F0,$44,$9F,$AD,$39,$7E
	.dc.b	$AF,$3F,$97,$D7,$91,$83,$EC,$FB,$BE,$99,$91,$83,$EE,$F9,$3F,$E1
	.dc.b	$F7,$F0,$2D,$D8,$D3,$5E,$0B,$EA,$3F,$D3,$5E,$3F,$F9,$92,$31,$7F
	.dc.b	$C0,$27,$3D,$3F,$E9,$FE,$1F,$E9,$B3,$FE,$F8,$BF,$F3,$5E,$3F,$DA
	.dc.b	$63,$15,$F8,$DF,$57,$8C,$7E,$6D,$AF,$E5,$BC,$FB,$5F,$D5,$63,$F1
	.dc.b	$F1,$E9,$23,$FE,$3B,$DF,$FD,$0E,$F6,$01,$CA,$C3,$FC,$FF,$18,$7F
	.dc.b	$D8,$FE,$4B,$F5,$7F,$F1,$75,$73,$3F,$53,$F1,$9E,$42,$39,$B7,$CA
	.dc.b	$FA,$9B,$1E,$58,$FA,$51,$88,$FF,$4D,$35,$FF,$62,$AF,$98,$F8,$7F
	.dc.b	$5C,$FA,$1A,$75,$8F,$93,$F1,$2F,$69,$CF,$9C,$7C,$95,$A9,$CC,$7A
	.dc.b	$F0,$FC,$73,$F0,$F8,$BF,$DB,$11,$8D,$FF,$D4,$6D,$8F,$F6,$E6,$7F
	.dc.b	$CF,$EB,$E0,$79,$BE,$C2,$3F,$4D,$FC,$85,$7A,$E3,$C1,$8F,$B0,$F8
	.dc.b	$DF,$65,$23,$ED,$3F,$18,$8F,$DD,$F6,$3E,$1F,$C9,$C8,$FA,$11,$ED
	.dc.b	$EB,$FA,$0C,$8F,$C1,$6F,$F1,$67,$F1,$BF,$5F,$FD,$12,$7E,$E3,$5F
	.dc.b	$84,$7E,$21,$C6,$DF,$86,$EA,$FB,$87,$DA,$24,$FE,$45,$FD,$77,$D7
	.dc.b	$BE,$0F,$F3,$6D,$9F,$C5,$35,$EE,$3E,$0D,$EC,$7E,$89,$B1,$F9,$16
	.dc.b	$BC,$1B,$83,$8F,$FC,$AF,$E0,$F8,$3D,$89,$7E,$07,$F5,$BF,$0D,$FB
	.dc.b	$A8,$F0,$B6,$FF,$DC,$2D,$BF,$D4,$CA,$30,$FE,$B3,$E7,$FF,$51,$3A
	.dc.b	$CD,$EB,$7C,$02,$57,$C0,$F5,$BE,$19,$F1,$88,$FF,$AC,$E7,$FD,$5C
	.dc.b	$FF,$AD,$7F,$44,$FE,$8F,$97,$FB,$D1,$B8,$79,$1B,$3F,$BE,$D5,$F4
	.dc.b	$3F,$C2,$BE,$C1,$EB,$52,$57,$EE,$FE,$D9,$ED,$FF,$10,$6C,$66,$30
	.dc.b	$7D,$3F,$8C,$FA,$9F,$31,$D7,$D3,$95,$F5,$63,$F8,$CF,$E1,$5F,$4E
	.dc.b	$30,$43,$EF,$B9,$7F,$03,$97,$F4,$D6,$F3,$7A,$4F,$57,$5C,$8F,$F6
	.dc.b	$C4,$FF,$A9,$DB,$1F,$CF,$63,$11,$F6,$0F,$E4,$3F,$16,$EF,$FE,$5A
	.dc.b	$D8,$FF,$4C,$47,$E3,$FA,$FE,$47,$9D,$27,$F9,$05,$FF,$39,$7F,$21
	.dc.b	$F0,$5F,$95,$23,$F9,$57,$F9,$CF,$DB,$3D,$8F,$A5,$F1,$92,$B9,$BA
	.dc.b	$F5,$A4,$7F,$2D,$A4,$62,$3F,$4B,$1F,$17,$AB,$E4,$7B,$BF,$E3,$26
	.dc.b	$FB,$7F,$D2,$23,$1D,$9F,$98,$7F,$94,$5F,$F3,$9F,$03,$F8,$D3,$FD
	.dc.b	$12,$32,$FE,$C9,$B7,$F5,$67,$72,$7F,$9C,$46,$7D,$9F,$EA,$F1,$A3
	.dc.b	$37,$FA,$85,$E4,$F9,$7E,$7A,$61,$E3,$8E,$3E,$57,$26,$47,$9E,$3F
	.dc.b	$B1,$F8,$8E,$1F,$77,$97,$2B,$FA,$A4,$E5,$BF,$69,$C2,$A4,$60,$FE
	.dc.b	$B5,$B7,$EE,$9F,$F1,$C7,$95,$F2,$4F,$95,$EC,$C8,$FC,$51,$F1,$DF
	.dc.b	$F0,$54,$9A,$7F,$5C,$3F,$B0,$12,$3F,$B9,$7D,$13,$F1,$77,$C9,$FF
	.dc.b	$80,$3D,$A7,$DD,$FD,$0E,$3F,$E1,$FF,$6A,$F8,$2F,$F0,$1B,$72,$FE
	.dc.b	$F9,$7C,$3D,$AC,$E3,$FE,$1D,$AE,$B5,$E8,$0E,$1F,$5B,$5F,$D3,$7D
	.dc.b	$AB,$ED,$FF,$C2,$2D,$8F,$F0,$03,$5F,$E8,$BE,$BB,$E3,$FC,$91,$8E
	.dc.b	$1F,$E1,$2F,$F3,$CA,$FE,$C0,$DE,$6F,$EE,$F3,$AE,$CF,$E0,$5F,$6D
	.dc.b	$CE,$BD,$0D,$7E,$8B,$5F,$DE,$35,$FE,$6A,$FF,$B0,$9F,$F0,$B3,$5F
	.dc.b	$E9,$AD,$8F,$E2,$98,$BD,$AF,$EA,$51,$8F,$C1,$3F,$CE,$0F,$F1,$5F
	.dc.b	$AD,$FF,$3E,$61,$FE,$21,$6B,$F0,$A7,$7C,$7F,$D1,$A7,$F8,$34,$FF
	.dc.b	$11,$6B,$9F,$1F,$FC,$A5,$18,$8F,$F1,$AF,$DB,$FE,$B8,$9E,$17,$F8
	.dc.b	$73,$EE,$1F,$2F,$FE,$1E,$F7,$BF,$82,$DF,$F8,$CE,$F2,$D9,$D7,$91
	.dc.b	$FE,$4A,$FB,$8F,$F3,$04,$9A,$47,$FC,$C8,$BC,$2E,$04,$8F,$F0,$4F
	.dc.b	$E0,$7F,$27,$6F,$F2,$5F,$81,$FE,$4C,$FA,$4F,$F2,$BF,$DC,$FF,$98
	.dc.b	$2F,$FB,$DB,$78,$FF,$A1,$56,$E7,$FC,$95,$E7,$38,$BF,$CE,$8D,$8F
	.dc.b	$CC,$35,$FD,$8B,$5F,$E7,$8F,$EE,$49,$E2,$FF,$90,$BE,$BB,$F5,$52
	.dc.b	$E8,$7D,$A3,$E7,$3F,$CD,$B8,$7F,$41,$AF,$F0,$8B,$73,$FC,$2D,$6E
	.dc.b	$7F,$CC,$1F,$72,$F9,$DF,$F4,$53,$63,$F5,$CD,$7F,$83,$B5,$FE,$A5
	.dc.b	$F7,$3F,$DC,$9F,$F3,$53,$5F,$F1,$0B,$6F,$FB,$47,$CC,$3F,$01,$FA
	.dc.b	$95,$C3,$E8,$FF,$D0,$B7,$FE,$98,$8E,$7F,$2F,$89,$CF,$7F,$FA,$7E
	.dc.b	$BF,$1D,$46,$93,$E5,$79,$92,$3F,$CF,$DF,$6C,$F9,$5E,$FE,$64,$9F
	.dc.b	$4B,$E2,$F6,$E4,$7F,$0F,$C1,$1E,$12,$47,$FA,$57,$8F,$FD,$3B,$F7
	.dc.b	$4F,$B9,$A7,$FA,$1E,$3F,$D4,$7C,$1F,$C9,$37,$C5,$FF,$A3,$67,$3E
	.dc.b	$9F,$C1,$8F,$F2,$4F,$CA,$7D,$9E,$FF,$F1,$BF,$C6,$7F,$96,$1B,$F4
	.dc.b	$1F,$E9,$3F,$83,$EB,$F0,$9F,$C5,$3A,$3F,$D4,$31,$9C,$7F,$D5,$1A
	.dc.b	$AB,$58,$7F,$90,$BE,$EB,$F2,$98,$7F,$86,$5F,$C7,$D9,$A6,$A7,$87
	.dc.b	$83,$FE,$BC,$8C,$4F,$FA,$D1,$AB,$F3,$0F,$F0,$C5,$FF,$89,$5F,$DC
	.dc.b	$92,$3F,$DB,$4F,$EF,$25,$F0,$3F,$EB,$B9,$7D,$DF,$F9,$09,$ED,$C7
	.dc.b	$E1,$3C,$53,$B9,$D4,$FA,$63,$1B,$FB,$EB,$FC,$C3,$5A,$BF,$B1,$7F
	.dc.b	$EC,$C8,$C5,$F0,$8F,$63,$E3,$7F,$C6,$53,$E6,$C7,$FA,$A3,$63,$FB
	.dc.b	$6B,$5D,$F9,$9F,$B8,$72,$D7,$3E,$0F,$28,$C4,$7F,$76,$FA,$0F,$E3
	.dc.b	$E4,$7F,$E0,$03,$71,$F7,$C3,$83,$FC,$AD,$18,$84,$FB,$74,$FF,$25
	.dc.b	$56,$9E,$D3,$F3,$9C,$EF,$B8,$FF,$28,$55,$7F,$DE,$35,$C5,$7B,$8F
	.dc.b	$DE,$7F,$C8,$D7,$FE,$0E,$8C,$78,$BF,$E6,$5A,$FF,$7C,$C6,$27,$55
	.dc.b	$C2,$7F,$CB,$D3,$CF,$E3,$BF,$CE,$1B,$FF,$BE,$A9,$18,$BE,$71,$C6
	.dc.b	$75,$39,$BF,$2C,$F0,$AF,$F0,$7F,$CE,$3E,$93,$58,$F7,$9C,$C7,$1C
	.dc.b	$FF,$C8,$DE,$5D,$F8,$8F,$F1,$35,$7E,$B1,$46,$7E,$12,$4E,$5E,$11
	.dc.b	$8C,$FF,$DA,$7E,$77,$F1,$9E,$1F,$33,$6F,$F6,$FE,$37,$AD,$FE,$AB
	.dc.b	$E2,$7F,$A8,$9F,$F1,$8B,$6F,$FD,$BB,$9B,$F2,$47,$93,$D2,$B5,$C7
	.dc.b	$FC,$CA,$73,$87,$F0,$4F,$0F,$FD,$56,$FF,$91,$30,$FE,$99,$F2,$BF
	.dc.b	$5F,$2F,$C6,$56,$9F,$C4,$B5,$FF,$65,$CB,$F2,$47,$DE,$7C,$96,$E1
	.dc.b	$DF,$0C,$4F,$CD,$FC,$7D,$F1,$94,$C7,$15,$BF,$E3,$3F,$8F,$7F,$82
	.dc.b	$1F,$33,$73,$FC,$DB,$5F,$1B,$5F,$CB,$FE,$3F,$FA,$59,$FE,$46,$FF
	.dc.b	$BD,$1B,$A7,$DF,$8C,$4F,$CC,$E7,$7D,$07,$E0,$DF,$5C,$66,$31,$A7
	.dc.b	$E3,$FD,$64,$FE,$6C,$FF,$77,$B8,$EE,$F9,$8D,$CB,$FC,$E6,$D9,$FF
	.dc.b	$5F,$3F,$AC,$D3,$A7,$5F,$E3,$AB,$AF,$27,$F9,$02,$3F,$27,$7F,$E5
	.dc.b	$A7,$4F,$F9,$73,$BF,$F6,$B5,$C0,$F2,$E1,$FC,$12,$BD,$12,$FC,$32
	.dc.b	$C7,$AF,$7F,$F3,$C5,$F9,$9B,$87,$99,$AF,$B9,$AF,$6F,$C7,$FF,$D0
	.dc.b	$B3,$E1,$8F,$F0,$D4,$7F,$6C,$8C,$62,$8D,$CF,$F9,$D1,$D3,$FE,$76
	.dc.b	$E4,$FF,$1E,$57,$06,$C6,$7E,$43,$E9,$46,$34,$FF,$32,$FC,$83,$E5
	.dc.b	$BE,$8B,$70,$FA,$2D,$C3,$E8,$B5,$FB,$63,$B9,$F8,$9B,$EB,$3F,$01
	.dc.b	$F8,$17,$E9,$72,$FF,$22,$CE,$71,$F6,$CE,$5F,$F2,$7F,$E3,$DF,$D7
	.dc.b	$9F,$83,$F7,$3E,$EB,$FF,$26,$CE,$72,$FF,$4A,$CE,$6B,$EC,$FE,$3D
	.dc.b	$FB,$19,$BB,$7E,$32,$31,$1F,$E4,$76,$F0,$BD,$5F,$9B,$97,$FA,$21
	.dc.b	$B8,$7E,$11,$ED,$7F,$E4,$F7,$2F,$F9,$47,$F0,$37,$D1,$7E,$7D,$B5
	.dc.b	$F6,$97,$D1,$66,$7F,$49,$F2,$7F,$A4,$AF,$CF,$B6,$FF,$3E,$D7,$F4
	.dc.b	$ED,$7F,$CE,$BF,$43,$FC,$E5,$FF,$5B,$FE,$05,$F5,$F3,$FF,$39,$CE
	.dc.b	$5F,$F5,$1C,$66,$FF,$CF,$BF,$8F,$3E,$CF,$F6,$DF,$81,$BE,$B3,$F9
	.dc.b	$8C,$E6,$BF,$27,$39,$7F,$2B,$F3,$FF,$AC,$CF,$F7,$6D,$7F,$56,$D7
	.dc.b	$FC,$EC,$DE,$1F,$C3,$75,$FF,$05,$E9,$F2,$57,$F9,$ED,$7E,$07,$99
	.dc.b	$FE,$B1,$2F,$C0,$9F,$31,$86,$E5,$FA,$B9,$7E,$AC,$FB,$9D,$5E,$FE
	.dc.b	$9F,$F9,$D5,$73,$6E,$06,$9A,$FB,$F7,$DD,$F8,$9B,$97,$7A,$E0,$F8
	.dc.b	$63,$37,$FE,$EA,$FA,$6F,$F3,$14,$FE,$F1,$B1,$9F,$84,$FE,$4D,$18
	.dc.b	$E7,$FF,$6C,$B6,$BF,$8A,$FD,$13,$AB,$7F,$0D,$FC,$B6,$31,$3F,$C2
	.dc.b	$3E,$73,$F8,$9E,$1F,$EE,$27,$ED,$65,$FE,$FC,$73,$3F,$C8,$1C,$CE
	.dc.b	$FA,$BF,$9C,$7C,$CF,$F2,$A8,$C5,$7F,$14,$8C,$3F,$C0,$B6,$B8,$7F
	.dc.b	$3E,$7F,$D6,$51,$F1,$7E,$03,$EB,$7B,$72,$3E,$47,$C6,$5B,$62,$9E
	.dc.b	$93,$FD,$27,$D4,$FD,$5D,$6A,$7F,$9B,$C6,$2B,$84,$F8,$6B,$87,$85
	.dc.b	$FD,$47,$DE,$FE,$75,$87,$F2,$C6,$D7,$33,$FD,$94,$FE,$29,$C7,$6F
	.dc.b	$FA,$AF,$D3,$FE,$B5,$F2,$0F,$A5,$A7,$58,$B8,$7F,$38,$F8,$BD,$36
	.dc.b	$7F,$B1,$43,$F0,$3B,$1F,$D9,$7E,$DD,$F4,$14,$CD,$FF,$42,$8C,$BF
	.dc.b	$9D,$F4,$3F,$B8,$B7,$DD,$FD,$2F,$E0,$6F,$37,$F4,$F5,$AD,$FD,$C1
	.dc.b	$FE,$EE,$FF,$3A,$31,$EF,$7E,$92,$33,$1F,$CF,$9A,$FF,$7A,$FC,$0F
	.dc.b	$EB,$63,$FA,$F6,$C7,$69,$FF,$01,$F2,$FD,$48,$4E,$A3,$EB,$F1,$7F
	.dc.b	$AA,$FD,$CF,$85,$F5,$7F,$19,$B8,$FC,$18,$F8,$DD,$B8,$C7,$EA,$EF
	.dc.b	$F9,$8A,$FF,$34,$FD,$D3,$EE,$BF,$CB,$2D,$8F,$ED,$DF,$00,$87,$D5
	.dc.b	$EA,$BF,$CB,$FF,$8C,$BE,$CE,$99,$C3,$8C,$65,$FE,$1D,$F8,$27,$87
	.dc.b	$3F,$F0,$EB,$72,$7E,$97,$FC,$EF,$87,$E1,$4E,$E8,$FD,$C7,$E9,$18
	.dc.b	$E8,$FE,$2B,$19,$8F,$AC,$D7,$FD,$0F,$5F,$E3,$97,$F3,$57,$E0,$E1
	.dc.b	$FE,$4F,$5D,$70,$FD,$2C,$E7,$1F,$F0,$6F,$E9,$9F,$B6,$C3,$F3,$ED
	.dc.b	$AF,$AA,$FE,$F3,$8C,$E6,$BF,$C4,$FC,$3F,$D3,$5B,$F2,$DF,$E3,$F9
	.dc.b	$D6,$0B,$5E,$E3,$13,$FE,$5E,$BF,$1F,$C6,$F1,$FB,$4E,$4F,$E9,$7C
	.dc.b	$59,$3A,$3F,$48,$F7,$E4,$FF,$24,$DF,$DC,$9F,$F2,$4B,$70,$FF,$24
	.dc.b	$B5,$FF,$28,$E3,$FE,$13,$7F,$17,$EC,$73,$AF,$D8,$3E,$BF,$AA,$FD
	.dc.b	$B7,$9F,$67,$F6,$8E,$1A,$7C,$C8,$C5,$7C,$CF,$9A,$F2,$92,$7C,$AF
	.dc.b	$F9,$83,$07,$E0,$5F,$C0,$CF,$95,$16,$7E,$BF,$A3,$EE,$23,$EE,$79
	.dc.b	$7C,$2F,$F0,$B3,$E9,$36,$3E,$FF,$CF,$7D,$53,$EB,$7B,$D6,$32,$7F
	.dc.b	$0F,$31,$97,$EB,$9F,$B9,$ED,$7C,$E2,$3F,$39,$FF,$3C,$F9,$5F,$EC
	.dc.b	$98,$FE,$45,$F2,$1C,$1A,$7F,$97,$3E,$63,$FC,$57,$CE,$6F,$1F,$F9
	.dc.b	$D3,$1F,$E6,$31,$A3,$FC,$C2,$FF,$A2,$1A,$FF,$9E,$1A,$FE,$3D,$B2
	.dc.b	$C3,$FC,$05,$7F,$CD,$7D,$DF,$F4,$A5,$7F,$8E,$AB,$51,$FE,$E0,$FD
	.dc.b	$AF,$FC,$7F,$D8,$68,$FA,$6D,$7D,$2F,$FB,$9B,$F7,$9F,$CD,$69,$FE
	.dc.b	$F6,$6B,$FE,$66,$B4,$62,$3F,$DF,$ED,$7D,$1F,$47,$E4,$BF,$2B,$74
	.dc.b	$F9,$11,$FD,$2B,$E2,$31,$5F,$D5,$B6,$FF,$1A,$3F,$C0,$BF,$6D,$F9
	.dc.b	$5F,$2D,$FE,$1B,$86,$69,$FE,$C3,$39,$8C,$1F,$DA,$9C,$46,$BE,$17
	.dc.b	$CF,$F2,$F9,$6E,$7F,$EB,$CA,$DF,$B1,$9C,$7B,$ED,$FE,$B0,$9F,$17
	.dc.b	$17,$DF,$F9,$DD,$37,$0E,$58,$F4,$C1,$7C,$39,$3F,$DE,$B1,$89,$FF
	.dc.b	$48,$FE,$09,$FE,$F1,$AF,$74,$A3,$13,$FE,$96,$F8,$0F,$EA,$B5,$A9
	.dc.b	$CD,$7F,$B5,$1F,$D8,$E1,$E2,$8C,$77,$FA,$DE,$9D,$D6,$B7,$1E,$33
	.dc.b	$97,$84,$F2,$FF,$9D,$C7,$8E,$30,$FF,$3E,$9F,$F1,$D4,$E9,$B1,$9F
	.dc.b	$13,$EB,$B7,$93,$51,$88,$FE,$F1,$3F,$EA,$F9,$E5,$E3,$7F,$AD,$34
	.dc.b	$F7,$37,$6F,$8E,$DB,$F8,$1E,$4D,$7E,$43,$70,$34,$D6,$FF,$E0,$D9
	.dc.b	$F8,$DC,$3F,$B9,$FA,$BE,$8D,$8C,$F1,$27,$C1,$A7,$E4,$7C,$1B,$D1
	.dc.b	$FE,$7A,$BF,$A1,$39,$6C,$65,$F9,$D0,$FC,$F3,$E5,$C6,$17,$DA,$76
	.dc.b	$73,$C1,$F9,$D6,$F0,$2B,$F4,$71,$C1,$F3,$FD,$27,$FB,$73,$CE,$FF
	.dc.b	$0A,$F6,$BE,$EB,$1D,$5C,$78,$7F,$86,$46,$30,$FE,$21,$85,$9F,$E3
	.dc.b	$9E,$A5,$E2,$4D,$31,$1C,$77,$FD,$A6,$BF,$88,$6E,$1F,$E0,$46,$BF
	.dc.b	$E3,$76,$BF,$C3,$75,$FF,$6D,$3F,$E1,$E6,$BF,$E0,$46,$EB,$FE,$02
	.dc.b	$7F,$CB,$51,$FE,$57,$FB,$EF,$46,$C6,$9E,$9F,$76,$4F,$F8,$5B,$5F
	.dc.b	$E7,$11,$F9,$E6,$C6,$75,$FC,$33,$FE,$4D,$8F,$E7,$DE,$AF,$C3,$24
	.dc.b	$7F,$43,$F6,$BC,$3F,$1D,$23,$FC,$D4,$FC,$9E,$99,$1D,$5F,$06,$1F
	.dc.b	$96,$8C,$1E,$1E,$CF,$5F,$BD,$23,$FA,$84,$7D,$32,$38,$3F,$AA,$78
	.dc.b	$3F,$0B,$E0,$24,$7F,$9E,$DF,$B5,$E1,$48,$F7,$3E,$22,$3F,$17,$81
	.dc.b	$F3,$DF,$30,$91,$F9,$1E,$3F,$F3,$0F,$C2,$3D,$0F,$C4,$36,$7E,$73
	.dc.b	$5F,$94,$D7,$F3,$47,$F8,$B5,$C3,$F9,$5B,$7D,$AF,$DC,$24,$78,$7B
	.dc.b	$3F,$35,$5F,$EF,$F8,$C1,$CF,$CB,$DB,$FD,$36,$BC,$B1,$F8,$6F,$B3
	.dc.b	$93,$FE,$9F,$7D,$3D,$12,$3D,$2F,$E1,$6F,$F1,$91,$83,$E3,$7C,$A7
	.dc.b	$E4,$08,$F8,$BA,$12,$E3,$F8,$A4,$9D,$46,$5B,$C3,$F9,$AF,$94,$91
	.dc.b	$FE,$B3,$7E,$6E,$92,$38,$B5,$FF,$5F,$D3,$3F,$DA,$7E,$68,$F3,$A9
	.dc.b	$2B,$E3,$8F,$C0,$DE,$AF,$F0,$38,$C3,$FD,$09,$A7,$F4,$48,$DB,$63
	.dc.b	$51,$F7,$F9,$3F,$D5,$0D,$E2,$FB,$ED,$AF,$EF,$78,$7F,$25,$AF,$F3
	.dc.b	$EF,$86,$F8,$CF,$E1,$8D,$AF,$F5,$0B,$5F,$F4,$3D,$3D,$BF,$79,$23
	.dc.b	$FD,$4F,$19,$6F,$DE,$FF,$06,$BF,$E0,$FF,$37,$27,$FD,$C1,$1F,$9A
	.dc.b	$8D,$46,$0F,$AE,$FC,$5F,$C0,$BF,$2B,$1F,$29,$AF,$FB,$59,$B9,$7F
	.dc.b	$B2,$7F,$00,$FE,$95,$3F,$41,$AF,$FA,$A1,$AF,$FB,$6D,$AF,$E9,$9A
	.dc.b	$FF,$41,$5E,$AF,$D7,$6C,$75,$8F,$37,$6F,$F3,$DF,$51,$EB,$BD,$5C
	.dc.b	$36,$FF,$53,$3A,$9C,$F8,$B2,$73,$5F,$C5,$D6,$4E,$3F,$EA,$98,$76
	.dc.b	$BE,$BE,$B7,$F4,$BF,$90,$7C,$C5,$26,$3F,$A7,$CF,$F3,$0D,$E9,$FE
	.dc.b	$A5,$18,$8E,$F9,$7F,$59,$8C,$67,$FE,$F2,$5E,$07,$0F,$90,$97,$D5
	.dc.b	$FC,$E6,$7F,$83,$8E,$67,$F6,$B5,$CB,$AE,$79,$3F,$A6,$9C,$B7,$E1
	.dc.b	$39,$D4,$8C,$1F,$C0,$7E,$5F,$EE,$E7,$F1,$63,$FD,$95,$F2,$FF,$E7
	.dc.b	$17,$F8,$33,$5F,$1F,$88,$AE,$3B,$7F,$77,$5F,$69,$FD,$D5,$CD,$7F
	.dc.b	$B5,$5B,$5C,$67,$FB,$D7,$E3,$FF,$AD,$BF,$F0,$07,$E0,$1F,$E0,$7C
	.dc.b	$3B,$B7,$2F,$36,$0B,$E1,$5F,$EE,$16,$C6,$7D,$CE,$A7,$6F,$5B,$F8
	.dc.b	$47,$DB,$3D,$5A,$4C,$3F,$87,$FE,$B6,$F9,$8F,$E1,$CD,$8F,$F2,$7B
	.dc.b	$5F,$E8,$A7,$F8,$7D,$C3,$86,$BF,$D2,$9F,$F5,$93,$DB,$D5,$EA,$79
	.dc.b	$3D,$8F,$F0,$E7,$E4,$1F,$B4,$E0,$FF,$11,$36,$BF,$DF,$CD,$D4,$FE
	.dc.b	$32,$3F,$9E,$7E,$40,$F7,$7F,$C1,$0D,$8F,$F0,$63,$5F,$F0,$43,$5F
	.dc.b	$EB,$FF,$56,$FC,$81,$F3,$1F,$E2,$FE,$DF,$D6,$72,$FD,$26,$E9,$FE
	.dc.b	$3F,$BE,$15,$FE,$37,$5F,$90,$72,$C7,$FA,$87,$6B,$DE,$FF,$1D,$36
	.dc.b	$3F,$B2,$D7,$0B,$FF,$1F,$7A,$0F,$7F,$FC,$44,$DA,$FF,$13,$35,$FF
	.dc.b	$11,$35,$FE,$F3,$C2,$FC,$15,$E7,$FF,$92,$5B,$3F,$D8,$DA,$FE,$79
	.dc.b	$FA,$EF,$E8,$3E,$AD,$F1,$DF,$E5,$16,$C7,$F6,$B6,$BB,$F8,$DF,$F2
	.dc.b	$B7,$CC,$7F,$1C,$8F,$F1,$E3,$5F,$F1,$C3,$5F,$3B,$5F,$99,$E3,$F9
	.dc.b	$39,$3F,$43,$E0,$7F,$16,$9F,$EF,$4D,$7F,$83,$F0,$7F,$4F,$D3,$E9
	.dc.b	$7B,$6F,$C1,$3F,$CD,$1F,$14,$F9,$DA,$6A,$3F,$CD,$9F,$90,$7E,$FF
	.dc.b	$1F,$92,$D7,$FC,$E8,$D5,$C4,$8C,$2F,$C1,$3F,$C1,$13,$FC,$21,$BA
	.dc.b	$FE,$E9,$BD,$5F,$E7,$F7,$FC,$77,$3F,$E1,$5A,$F4,$37,$E6,$3F,$D0
	.dc.b	$10,$6E,$3F,$E8,$3F,$C5,$3E,$BD,$7E,$01,$AF,$FA,$2D,$AF,$A1,$B4
	.dc.b	$FC,$17,$FD,$BB,$97,$F0,$E6,$E5,$FE,$93,$6D,$FC,$07,$FC,$30,$FF
	.dc.b	$8A,$FE,$FD,$F2,$9F,$E9,$56,$C6,$63,$1F,$3D,$FD,$7F,$F1,$F7,$D0
	.dc.b	$7F,$B6,$BD,$C4,$62,$FF,$86,$9F,$ED,$F7,$0F,$F1,$08,$DC,$F1,$FC
	.dc.b	$17,$F8,$14,$FF,$48,$94,$7F,$52,$CF,$FB,$0B,$67,$FD,$D4,$F0,$D3
	.dc.b	$FA,$15,$3E,$A3,$FB,$85,$73,$7F,$7E,$FF,$BC,$5B,$1B,$F6,$BF,$BD
	.dc.b	$7E,$3E,$F9,$DF,$63,$6F,$D7,$48,$C4,$7A,$DA,$FF,$9C,$21,$F8,$FB
	.dc.b	$E3,$BF,$C0,$CD,$9F,$F0,$1B,$5F,$80,$FF,$27,$97,$E5,$6F,$A5,$FF
	.dc.b	$07,$36,$BF,$C1,$6D,$7E,$47,$E4,$17,$5F,$86,$E6,$3C,$8D,$7F,$C1
	.dc.b	$6D,$56,$8C,$78,$1F,$E1,$BF,$73,$FC,$3D,$F2,$9E,$53,$F9,$E4,$66
	.dc.b	$FF,$90,$EF,$FD,$A9,$B5,$FC,$AE,$31,$1F,$77,$DA,$FE,$75,$97,$D8
	.dc.b	$E0,$FE,$5D,$3F,$81,$6C,$67,$CD,$FF,$17,$43,$E0,$69,$88,$95,$FE
	.dc.b	$2E,$9D,$3F,$E9,$8F,$71,$FA,$CF,$FB,$DE,$3F,$CB,$FF,$21,$FE,$62
	.dc.b	$3F,$DD,$D1,$98,$FF,$0C,$CF,$0F,$9D,$C9,$23,$A4,$FF,$A6,$1D,$B5
	.dc.b	$FF,$58,$CF,$83,$E5,$BF,$DF,$13,$99,$09,$D3,$FA,$81,$F2,$FF,$E3
	.dc.b	$CD,$37,$F8,$37,$F3,$C8,$FE,$15,$38,$F7,$FF,$D7,$F1,$88,$4E,$FE
	.dc.b	$2B,$FC,$97,$18,$AD,$C6,$27,$A7,$91,$FE,$C8,$6D,$75,$3C,$F7,$E3
	.dc.b	$79,$18,$AE,$AF,$F8,$32,$34,$D8,$CB,$E0,$C3,$DD,$3F,$E8,$DF,$12
	.dc.b	$FA,$06,$92,$BF,$D4,$DF,$3F,$FE,$A7,$4D,$23,$A3,$CA,$7F,$9A,$3C
	.dc.b	$2F,$53,$FE,$E4,$8A,$FE,$82,$2B,$D0,$E2,$64,$A6,$72,$FE,$6F,$D7
	.dc.b	$FE,$51,$7F,$C4,$BB,$3F,$06,$7C,$F7,$F9,$56,$FF,$07,$18,$C3,$F9
	.dc.b	$2D,$6E,$B5,$5F,$84,$8C,$4E,$AB,$FD,$18,$FE,$6D,$B3,$F9,$67,$A3
	.dc.b	$7A,$7F,$CC,$9B,$7F,$99,$6F,$C9,$18,$BE,$D6,$F8,$FC,$61,$F9,$38
	.dc.b	$EB,$87,$17,$55,$FE,$9A,$FC,$19,$F6,$3C,$35,$47,$FA,$8A,$75,$39
	.dc.b	$8C,$61,$FC,$F5,$B8,$7F,$38,$6D,$7F,$A7,$9E,$AF,$7F,$92,$ED,$F8
	.dc.b	$37,$92,$3F,$D4,$6D,$BF,$E3,$2F,$99,$FE,$94,$E7,$13,$BF,$01,$F2
	.dc.b	$6E,$8E,$BB,$FC,$29,$CC,$7B,$75,$7D,$0F,$D6,$CB,$F9,$3C,$63,$3F
	.dc.b	$E8,$AD,$8F,$EB,$CD,$7F,$A2,$2F,$A1,$E8,$75,$AB,$5F,$EC,$6F,$C1
	.dc.b	$3E,$1D,$FF,$34,$6E,$A7,$FB,$BC,$EF,$6F,$A5,$0F,$43,$CD,$5B,$6C
	.dc.b	$66,$BF,$AD,$B6,$FF,$BA,$7E,$09,$FE,$FF,$9F,$F6,$BF,$E0,$9F,$A1
	.dc.b	$E3,$FE,$DB,$39,$AF,$EE,$D3,$9A,$FF,$6A,$FD,$27,$F9,$95,$FF,$6F
	.dc.b	$BF,$D8,$5B,$F1,$1F,$87,$8C,$57,$FA,$D0,$FB,$07,$F9,$80,$A7,$F2
	.dc.b	$6F,$5F,$F9,$3F,$47,$D1,$6D,$FF,$83,$9F,$E5,$5E,$B7,$E6,$9B,$1F
	.dc.b	$CB,$B3,$FF,$58,$F5,$7E,$33,$0F,$F6,$1B,$F2,$9F,$F6,$2F,$6B,$E8
	.dc.b	$79,$36,$F9,$1E,$35,$F0,$3A,$9F,$07,$17,$F9,$56,$BC,$8D,$BF,$F5
	.dc.b	$7E,$5F,$DB,$BD,$EF,$54,$E6,$7F,$CF,$F7,$FD,$0E,$FB,$C7,$F9,$5B
	.dc.b	$F0,$47,$E1,$3E,$DA,$D4,$FF,$B7,$27,$35,$C3,$1F,$C7,$7C,$EF,$C3
	.dc.b	$E9,$37,$D7,$C3,$F1,$FF,$05,$FF,$7E,$DF,$CB,$6E,$3E,$86,$BF,$E9
	.dc.b	$46,$BF,$57,$F0,$4F,$D0,$71,$7F,$BC,$FB,$DF,$82,$FF,$14,$8F,$F7
	.dc.b	$0C,$EB,$0F,$F5,$84,$E5,$FB,$DF,$11,$EC,$AE,$5F,$82,$BC,$DF,$B9
	.dc.b	$39,$AF,$F5,$C4,$E5,$FF,$4D,$78,$3F,$85,$F8,$BF,$D8,$69,$F8,$96
	.dc.b	$BF,$E8,$06,$F0,$BF,$05,$FF,$52,$DF,$C9,$6D,$33,$FF,$7C,$FD,$87
	.dc.b	$86,$3F,$D1,$33,$A9,$FF,$7E,$FE,$0B,$FE,$B6,$BF,$A4,$DA,$E5,$1F
	.dc.b	$9B,$87,$C6,$7E,$5D,$D5,$27,$F4,$1F,$27,$FD,$46,$FF,$D8,$6D,$73
	.dc.b	$D5,$FA,$38,$C4,$7F,$A7,$3F,$05,$FB,$37,$F4,$FE,$6F,$F9,$23,$D6
	.dc.b	$75,$F4,$9F,$C7,$E7,$38,$FE,$BF,$F0,$47,$B9,$FC,$8F,$F0,$5F,$CD
	.dc.b	$63,$C7,$85,$5A,$BF,$F7,$0D,$69,$FD,$C7,$DC,$7E,$B7,$3F,$F6,$DB
	.dc.b	$5C,$F8,$DF,$BB,$8C,$47,$FA,$D3,$D6,$F4,$D7,$F0,$16,$C6,$7E,$A3
	.dc.b	$F7,$D1,$88,$7E,$0B,$E3,$7E,$79,$FE,$2E,$71,$A7,$A7,$F3,$46,$FF
	.dc.b	$D8,$2B,$8E,$F7,$F0,$3F,$DE,$E7,$FE,$33,$6C,$FF,$A7,$1A,$FF,$8E
	.dc.b	$8F,$F1,$DB,$87,$0B,$ED,$BA,$1D,$7C,$AB,$5F,$B7,$FC,$17,$F3,$57
	.dc.b	$E8,$E3,$46,$34,$FA,$0F,$F3,$16,$F8,$1F,$C7,$E3,$57,$FC,$9D,$FF
	.dc.b	$67,$3D,$63,$5F,$03,$FC,$97,$EB,$1F,$53,$4C,$6F,$6F,$F8,$54,$BE
	.dc.b	$87,$81,$9D,$3C,$D1,$99,$FF,$37,$7C,$4B,$EF,$7F,$C1,$7E,$9A,$31
	.dc.b	$3F,$C5,$7D,$EF,$D8,$3C,$2B,$E8,$75,$7F,$6D,$C5,$C9,$F8,$46,$C6
	.dc.b	$63,$07,$F3,$BE,$1F,$F1,$73,$FE,$C1,$BF,$E5,$35,$CB,$DD,$FB,$9F
	.dc.b	$29,$FD,$17,$E0,$3E,$A5,$7F,$27,$7E,$EE,$1F,$9D,$73,$7F,$4D,$CC
	.dc.b	$FF,$53,$6C,$39,$7F,$C1,$7E,$A7,$5E,$7E,$CF,$F5,$3F,$87,$7D,$55
	.dc.b	$31,$F0,$57,$2F,$E7,$12,$FA,$5E,$06,$72,$FC,$24,$66,$7D,$BF,$82
	.dc.b	$7E,$69,$BD,$1F,$9B,$8F,$BB,$18,$3E,$1F,$E0,$9F,$DC,$2F,$F3,$01
	.dc.b	$C3,$F9,$B6,$BF,$E9,$66,$BB,$FB,$9F,$D4,$41,$BE,$CE,$1F,$81,$F4
	.dc.b	$52,$FD,$BE,$85,$C3,$D7,$E7,$7C,$53,$1C,$9F,$C3,$75,$FF,$06,$4F
	.dc.b	$FA,$97,$AF,$FD,$81,$87,$E3,$BE,$63,$F1,$B1,$89,$FF,$5B,$BF,$EE
	.dc.b	$79,$D1,$F9,$4F,$C0,$F7,$5C,$1F,$CA,$FE,$05,$FE,$B2,$9F,$F6,$33
	.dc.b	$6F,$E7,$9A,$3F,$83,$7B,$1F,$DD,$DF,$E0,$0D,$7F,$D7,$6D,$4D,$3E
	.dc.b	$9B,$FC,$50,$CF,$87,$D3,$E0,$7A,$BF,$E1,$9F,$C0,$7F,$35,$3F,$C0
	.dc.b	$DF,$F0,$B6,$BF,$CC,$5C,$B7,$C0,$E0,$71,$EF,$B9,$FF,$80,$EB,$F4
	.dc.b	$CF,$F8,$AD,$FF,$26,$F6,$97,$CF,$F8,$7E,$33,$C4,$CA,$E5,$F1,$97
	.dc.b	$3F,$23,$E7,$E0,$FF,$1A,$60,$69,$2B,$FC,$6D,$1F,$56,$BD,$9C,$3C
	.dc.b	$67,$CC,$FC,$1E,$2F,$85,$F2,$F6,$7D,$4E,$1A,$7B,$63,$15,$F1,$FE
	.dc.b	$D3,$C6,$49,$F0,$BE,$6F,$94,$F9,$C7,$CF,$6C,$7C,$7F,$37,$D6,$47
	.dc.b	$D2,$FB,$BC,$DF,$AC,$7D,$76,$C7,$CF,$F5,$7E,$D1,$F6,$FE,$92,$C6
	.dc.b	$4F,$A9,$E5,$E3,$2F,$BC,$FE,$07,$D1,$F9,$64,$9F,$92,$FC,$EF,$27
	.dc.b	$F0,$C6,$E3,$EF,$3F,$89,$F3,$A4,$7E,$07,$CA,$B8,$7F,$4D,$F2,$7F
	.dc.b	$C3,$AF,$F5,$9E,$8F,$F5,$17,$D4,$FF,$9B,$67,$33,$FE,$71,$6B,$FC
	.dc.b	$7D,$FF,$27,$BC,$FE,$2B,$FC,$E1,$1F,$E2,$3C,$38,$47,$F4,$2F,$C0
	.dc.b	$5F,$55,$4C,$45,$9F,$F2,$FE,$6A,$EE,$DC,$7F,$D6,$0D,$9F,$F7,$4C
	.dc.b	$EB,$F0,$1F,$F4,$D7,$81,$FE,$15,$C9,$F3,$BF,$D4,$A7,$37,$EB,$8F
	.dc.b	$39,$E8,$CE,$C7,$F9,$A6,$BC,$EF,$93,$F0,$0F,$F3,$7F,$C8,$7F,$7B
	.dc.b	$D9,$A2,$73,$C5,$F1,$FF,$0B,$FF,$17,$5E,$EB,$46,$27,$FD,$FD,$5F
	.dc.b	$E8,$DB,$FF,$1A,$1D,$1C,$61,$FE,$0D,$AE,$DC,$5F,$E3,$98,$C4,$EA
	.dc.b	$87,$93,$5F,$F7,$7E,$3E,$7E,$2F,$F6,$F7,$DC,$F8,$6B,$5E,$1D,$9C
	.dc.b	$E3,$CA,$FD,$D4,$F8,$8F,$F6,$C4,$E7,$5F,$F1,$04,$62,$55,$C3,$C6
	.dc.b	$FF,$2D,$5E,$E7,$CB,$5B,$AE,$AF,$FB,$9D,$BF,$15,$E5,$9F,$07,$C8
	.dc.b	$DA,$D6,$3E,$0F,$88,$79,$FF,$D6,$AF,$AC,$61,$C5,$23,$11,$FD,$82
	.dc.b	$31,$3C,$63,$DD,$F0,$7F,$71,$E2,$D7,$83,$5F,$82,$D7,$DB,$EE,$FF
	.dc.b	$90,$F3,$F6,$B7,$1F,$8E,$D9,$F8,$7F,$27,$5F,$9C,$DC,$4D,$35,$B7
	.dc.b	$F9,$62,$7E,$67,$2F,$E4,$7B,$BA,$70,$BA,$1F,$81,$61,$B7,$C3,$9F
	.dc.b	$FD,$B2,$F2,$FC,$0B,$E2,$F2,$FE,$3D,$B1,$F8,$CC,$91,$88,$FD,$0B
	.dc.b	$63,$3E,$B7,$B5,$B3,$D6,$3C,$5F,$49,$E3,$FB,$6F,$F3,$47,$81,$FB
	.dc.b	$7F,$C0,$2E,$5F,$C3,$DB,$87,$F0,$FA,$B8,$FE,$BF,$93,$3B,$FB,$E6
	.dc.b	$EC,$77,$B7,$E6,$B3,$FF,$50,$61,$EA,$F6,$7F,$00,$DC,$BF,$14,$D7
	.dc.b	$FC,$4A,$D7,$E2,$7B,$FE,$2E,$0F,$F1,$B4,$7F,$18,$6B,$FE,$90,$9F
	.dc.b	$8F,$F0,$1E,$4F,$80,$93,$FC,$35,$F3,$FB,$12,$3F,$35,$B7,$D9,$8F
	.dc.b	$CA,$36,$33,$B7,$F9,$E1,$BF,$0B,$F0,$08,$FF,$19,$F6,$BE,$1F,$4C
	.dc.b	$8F,$E3,$8F,$C6,$EB,$91,$D1,$EF,$66,$98,$F0,$7E,$6F,$74,$AF,$E4
	.dc.b	$D1,$F4,$08,$E0,$FE,$55,$E0,$7D,$4F,$06,$47,$F2,$D7,$EB,$F4,$C8
	.dc.b	$F6,$BE,$22,$3F,$13,$F0,$17,$CD,$7E,$01,$B1,$FC,$83,$9F,$FD,$A9
	.dc.b	$18,$8F,$F5,$53,$FE,$DA,$75,$5F,$C9,$B8,$13,$99,$FF,$31,$C6,$3E
	.dc.b	$27,$E8,$CE,$67,$FB,$33,$5E,$4B,$EA,$3C,$84,$8E,$E7,$FA,$16,$1F
	.dc.b	$4F,$E9,$35,$1E,$69,$C3,$C2,$FC,$BF,$4F,$EF,$24,$66,$DF,$3D,$F0
	.dc.b	$CF,$39,$23,$E4,$3E,$AF,$86,$F9,$1F,$0D,$27,$94,$7C,$BF,$9D,$F9
	.dc.b	$8C,$8E,$D1,$F3,$FE,$47,$E8,$7E,$05,$27,$C5,$1F,$4F,$E5,$75,$18
	.dc.b	$3C,$CF,$EE,$97,$DB,$7C,$FE,$2F,$EB,$ED,$E2,$FF,$75,$36,$7F,$AD
	.dc.b	$BF,$6E,$77,$C7,$FC,$DE,$79,$B6,$7F,$01,$1F,$E9,$C6,$BF,$C6,$7F
	.dc.b	$02,$FE,$3E,$FE,$D5,$B5,$FC,$11,$B7,$F7,$3E,$6B,$F8,$7C,$3E,$06
	.dc.b	$FC,$14,$C3,$FC,$2F,$39,$6F,$C2,$70,$A9,$18,$3E,$C7,$DB,$7E,$64
	.dc.b	$8F,$DB,$EF,$FC,$E1,$1C,$1F,$7F,$F0,$2F,$95,$F8,$0F,$F3,$A8,$FF
	.dc.b	$45,$37,$5F,$F0,$4B,$75,$FF,$01,$CA,$1E,$CF,$23,$E5,$38,$8F,$EC
	.dc.b	$4D,$7F,$B6,$35,$FF,$63,$FE,$04,$E5,$FB,$8D,$8F,$F7,$7B,$5F,$F4
	.dc.b	$F7,$E1,$FF,$C2,$DF,$D1,$FB,$4F,$6F,$FA,$EC,$61,$AB,$36,$7F,$E1
	.dc.b	$CA,$E1,$E3,$FF,$44,$79,$78,$97,$0F,$EF,$D8,$73,$8C,$19,$F1,$BF
	.dc.b	$A3,$F8,$EE,$AB,$D7,$2F,$F6,$13,$96,$F4,$F0,$A9,$18,$3F,$7A,$7F
	.dc.b	$83,$79,$BF,$3D,$8B,$A6,$98,$8C,$F4,$7E,$66,$4D,$25,$7F,$52,$E9
	.dc.b	$7B,$3F,$92,$24,$79,$8F,$EC,$3F,$81,$BD,$7A,$48,$FE,$15,$1F,$D8
	.dc.b	$DB,$EA,$F0,$23,$AF,$5F,$FB,$5F,$E0,$5F,$E1,$FE,$7F,$A3,$2F,$21
	.dc.b	$3F,$A6,$8C,$36,$33,$E9,$F2,$24,$FF,$1B,$9E,$9D,$3F,$CC,$57,$BF
	.dc.b	$FC,$17,$F0,$27,$AB,$49,$7F,$3A,$3F,$C3,$8D,$F0,$FF,$81,$46,$23
	.dc.b	$FC,$84,$D7,$F4,$21,$C3,$FA,$4F,$B3,$FF,$15,$3F,$EF,$36,$E9,$FE
	.dc.b	$72,$6E,$5F,$D1,$FE,$7B,$F1,$B5,$F9,$58,$FE,$95,$F8,$17,$E1,$E3
	.dc.b	$07,$C7,$71,$7F,$2B,$64,$62,$B8,$F6,$7F,$86,$BF,$02,$F1,$E1,$FE
	.dc.b	$B3,$6B,$FD,$5D,$AF,$D3,$6D,$BD,$BF,$D7,$4F,$8B,$EA,$9F,$81,$3E
	.dc.b	$4B,$FC,$57,$18,$7F,$C5,$35,$C3,$4F,$37,$63,$D1,$DF,$E0,$5C,$72
	.dc.b	$FE,$B7,$8A,$31,$1F,$03,$3F,$F3,$75,$7F,$8D,$3C,$FF,$DD,$CF,$FB
	.dc.b	$35,$AF,$F7,$16,$BF,$8D,$6C,$BF,$02,$FB,$99,$7C,$96,$F0,$7F,$79
	.dc.b	$6D,$FE,$55,$79,$FF,$E4,$4F,$AA,$FF,$67,$CF,$F6,$B6,$BE,$76,$DF
	.dc.b	$D5,$5F,$21,$FE,$4C,$F9,$C7,$CD,$7F,$8C,$9B,$7F,$E3,$66,$BF,$E0
	.dc.b	$86,$BF,$EE,$BF,$8A,$7C,$BF,$D0,$F2,$5E,$B5,$3F,$CB,$33,$AA,$FF
	.dc.b	$00,$7C,$53,$E1,$3E,$BF,$9C,$F9,$5A,$7F,$98,$27,$55,$FE,$07,$FC
	.dc.b	$0B,$FD,$BE,$FF,$92,$9A,$FF,$87,$1A,$FF,$BF,$DA,$FF,$88,$FF,$02
	.dc.b	$FE,$71,$C3,$F9,$76,$F2,$67,$D9,$FF,$37,$C3,$DE,$FF,$38,$F3,$BE
	.dc.b	$5E,$E9,$32,$FF,$0D,$4E,$5F,$F0,$E7,$E0,$5F,$BE,$7F,$CB,$ED,$7F
	.dc.b	$C6,$2D,$7F,$AC,$C2,$31,$1F,$E3,$9F,$C0,$BF,$84,$BF,$E2,$76,$BF
	.dc.b	$93,$6E,$9F,$B1,$7F,$8B,$7D,$DF,$D8,$8F,$DA,$37,$5F,$F2,$23,$6B
	.dc.b	$F9,$03,$FD,$5D,$FE,$B3,$F8,$13,$E8,$3F,$D9,$1E,$EA,$31,$3F,$ED
	.dc.b	$38,$C4,$EA,$73,$3F,$C5,$25,$F8,$13,$93,$FA,$E3,$FC,$16,$7F,$DB
	.dc.b	$AD,$9D,$CF,$F6,$1F,$C0,$BE,$EE,$7F,$CD,$A5,$F5,$5F,$BB,$D3,$FB
	.dc.b	$43,$71,$FF,$73,$DF,$09,$FE,$D9,$F8,$17,$DB,$9F,$43,$5F,$33,$5F
	.dc.b	$F0,$5D,$A3,$11,$CB,$E2,$3F,$BA,$41,$A6,$39,$BF,$BC,$7E,$02,$FA
	.dc.b	$4A,$4B,$FF,$02,$C7,$F0,$CD,$3D,$8F,$F0,$E8,$F0,$B5,$FF,$02,$35
	.dc.b	$62,$8C,$2F,$4B,$FB,$14,$FF,$25,$F8,$2F,$ED,$0F,$F1,$C6,$BF,$CE
	.dc.b	$A7,$87,$B7,$FE,$0F,$AE,$38,$7C,$F6,$EB,$FE,$15,$6D,$F8,$97,$B7
	.dc.b	$FE,$17,$FA,$0F,$EF,$F3,$FE,$22,$6B,$DD,$B7,$F6,$DF,$F2,$64,$67
	.dc.b	$E0,$B9,$7C,$14,$8E,$51,$FE,$E4,$FC,$05,$F1,$F7,$E4,$65,$7F,$A4
	.dc.b	$BE,$05,$5A,$8C,$35,$FF,$6C,$37,$2F,$F7,$3B,$75,$4A,$B9,$FC,$8F
	.dc.b	$84,$FF,$75,$D7,$FA,$8E,$7F,$DD,$6D,$7F,$DD,$6D,$7F,$BA,$FD,$53
	.dc.b	$E9,$BF,$C8,$10,$FC,$07,$FB,$E4,$FF,$25,$8C,$7C,$17,$F9,$72,$31
	.dc.b	$3A,$AE,$2D,$CF,$CE,$F2,$F3,$FF,$CC,$6D,$C3,$38,$73,$BE,$2D,$BF
	.dc.b	$F0,$4C,$FC,$0C,$3C,$D1,$88,$46,$1D,$B5,$F6,$F9,$EF,$0F,$84,$FF
	.dc.b	$91,$5B,$19,$9E,$2D,$7B,$7D,$0F,$FA,$7D,$34,$8F,$FA,$8D,$F5,$BF
	.dc.b	$B6,$DE,$BF,$E8,$28,$AF,$E9,$A2,$BD,$0F,$9A,$71,$29,$C4,$A7,$13
	.dc.b	$7F,$8F,$7F,$23,$F9,$9F,$F2,$3C,$7F,$A4,$23,$13,$F2,$9B,$1F,$E3
	.dc.b	$46,$BB,$F1,$9E,$EF,$F9,$43,$4F,$F2,$8D,$7F,$2F,$8C,$5F,$E4,$A3
	.dc.b	$9C,$EA,$B8,$CF,$FB,$C2,$78,$37,$F4,$87,$77,$2C,$55,$A7,$FC,$72
	.dc.b	$A7,$75,$FE,$98,$FC,$0B,$F9,$6B,$7C,$6E,$5E,$07,$FB,$F3,$0B,$5A
	.dc.b	$7F,$85,$37,$3F,$E6,$DF,$1A,$FA,$3D,$E1,$CA,$31,$8F,$F8,$BB,$EC
	.dc.b	$7F,$C0,$ED,$F6,$3B,$FE,$05,$FE,$05,$8C,$1F,$EA,$CF,$C3,$1F,$53
	.dc.b	$E4,$6C,$6B,$5F,$E3,$91,$87,$F8,$F7,$D8,$7C,$87,$FA,$A3,$5F,$E7
	.dc.b	$ED,$7F,$D6,$4D,$7F,$97,$FE,$18,$F9,$FF,$86,$DB,$FF,$5F,$D7,$5D
	.dc.b	$FF,$AB,$4E,$6F,$E7,$FC,$CF,$8E,$7E,$FB,$6F,$EF,$3F,$ED,$87,$EE
	.dc.b	$39,$9F,$B2,$BC,$8F,$C8,$C7,$06,$BA,$6F,$2F,$FB,$45,$B2,$FA,$CF
	.dc.b	$B7,$F8,$13,$EA,$7F,$0D,$7F,$ED,$79,$CC,$7E,$03,$E8,$BF,$97,$E5
	.dc.b	$FE,$DC,$6E,$5F,$E7,$1F,$D7,$FF,$01,$3F,$ED,$99,$CE,$9F,$E8,$3A
	.dc.b	$FB,$47,$FA,$5B,$DC,$FE,$5B,$F0,$7E,$86,$DF,$F1,$26,$CE,$FC,$BE
	.dc.b	$AF,$FA,$D7,$E3,$BF,$9A,$D7,$F3,$6C,$3F,$D2,$15,$AB,$56,$AF,$FC
	.dc.b	$F1,$E1,$BE,$B3,$F4,$95,$FC,$3B,$EC,$6C,$E7,$2E,$87,$F4,$29,$D3
	.dc.b	$FA,$FF,$C1,$7F,$A2,$3F,$BF,$6E,$1F,$BF,$6B,$FC,$15,$AF,$3F,$9B
	.dc.b	$FF,$61,$CF,$68,$FF,$68,$46,$1F,$F5,$6E,$9F,$E2,$FF,$A9,$F3,$CE
	.dc.b	$63,$FD,$5D,$1F,$E9,$9A,$E5,$82,$3F,$21,$3A,$F9,$2F,$7C,$62,$BD
	.dc.b	$9F,$81,$BE,$67,$E9,$36,$3E,$93,$72,$FB,$0D,$7F,$CA,$FF,$81,$FE
	.dc.b	$7E,$7E,$33,$63,$E8,$FE,$07,$F8,$0C,$BF,$D8,$93,$9C,$DF,$21,$EB
	.dc.b	$C3,$E7,$37,$0F,$B3,$F8,$1B,$DC,$FA,$F3,$9E,$57,$C7,$FF,$0A,$BF
	.dc.b	$F7,$0C,$63,$CA,$FD,$1B,$71,$FE,$BA,$D9,$FC,$97,$D9,$F7,$CC,$BE
	.dc.b	$4B,$F0,$AE,$4C,$5F,$CE,$FC,$0D,$F4,$FF,$8A,$7E,$37,$07,$FB,$85
	.dc.b	$CC,$EF,$8F,$F1,$FE,$A3,$E5,$3F,$40,$D9,$FD,$03,$74,$FD,$33,$5F
	.dc.b	$F4,$2F,$E0,$7D,$A7,$F2,$8D,$8F,$CF,$FD,$9B,$EC,$FF,$3B,$39,$DD
	.dc.b	$F8,$1B,$C3,$FD,$0B,$71,$FF,$6F,$FE,$06,$F8,$AC,$D7,$F4,$9E,$47
	.dc.b	$D4,$7F,$46,$D7,$E7,$B5,$FF,$13,$9F,$E2,$77,$0F,$F8,$AA,$38,$FB
	.dc.b	$7F,$BF,$FA,$DF,$F3,$46,$5F,$98,$5F,$1F,$FB,$47,$37,$FE,$EF,$73
	.dc.b	$3F,$A7,$8F,$AD,$D3,$FD,$75,$B1,$9F,$C0,$DF,$15,$9E,$BF,$C3,$74
	.dc.b	$FF,$78,$7F,$97,$5F,$F8,$81,$AF,$A8,$FC,$17,$D9,$FF,$2C,$6E,$1F
	.dc.b	$8A,$EF,$7B,$9F,$15,$B1,$E5,$BE,$B8,$23,$E1,$4E,$AD,$39,$5C,$AF
	.dc.b	$C0,$DF,$1D,$FC,$59,$B8,$7E,$F9,$FD,$23,$63,$47,$F1,$CF,$AE,$70
	.dc.b	$62,$7F,$82,$C6,$1F,$E0,$FF,$81,$FD,$F6,$FC,$8F,$F2,$48,$C6,$5E
	.dc.b	$69,$FE,$45,$1F,$E7,$E7,$64,$8D,$1E,$A8,$E3,$9F,$B7,$34,$71,$9D
	.dc.b	$8F,$C0,$DF,$55,$F6,$DB,$87,$DF,$7D,$F3,$C1,$CB,$7E,$5B,$EB,$47
	.dc.b	$FA,$BE,$7F,$CB,$0D,$9F,$F2,$EB,$5F,$BD,$F8,$6F,$F5,$69,$FF,$17
	.dc.b	$7D,$B7,$F4,$7C,$FF,$C7,$B3,$9A,$FF,$14,$CE,$6B,$FA,$27,$C3,$3D
	.dc.b	$5F,$E9,$6B,$E8,$AC,$61,$FE,$9F,$F7,$1F,$AA,$AF,$E6,$ED,$C3,$E4
	.dc.b	$EE,$C3,$95,$6B,$BF,$AF,$3F,$6F,$56,$CF,$F4,$0A,$FE,$35,$8B,$EB
	.dc.b	$7B,$9B,$9E,$EF,$E9,$FF,$03,$FF,$79,$47,$EA,$DB,$9F,$D4,$FC,$11
	.dc.b	$F7,$9B,$FA,$2E,$1E,$8F,$E8,$EF,$F8,$ED,$FF,$A2,$1A,$FE,$A5,$AF
	.dc.b	$F0,$4E,$3F,$E1,$CF,$E0,$1B,$C9,$FE,$6A,$9E,$6F,$D0,$F1,$FF,$88
	.dc.b	$7E,$45,$FD,$5E,$FF,$D6,$4F,$D3,$F2,$0E,$5B,$F3,$DC,$09,$B7,$FB
	.dc.b	$0E,$74,$71,$FD,$11,$F3,$5F,$C7,$57,$D7,$D8,$C3,$FC,$87,$F4,$CF
	.dc.b	$F1,$45,$FF,$B3,$5B,$97,$E8,$32,$61,$CA,$B5,$F1,$1F,$C5,$7D,$B7
	.dc.b	$DE,$53,$F8,$C7,$0E,$F1,$FD,$1C,$EA,$FF,$8F,$3F,$E2,$88,$FE,$28
	.dc.b	$D7,$F9,$7B,$5F,$F6,$BB,$57,$C6,$7F,$37,$E1,$FF,$4B,$B6,$B8,$FC
	.dc.b	$47,$F2,$98,$C5,$7F,$B7,$FF,$45,$FF,$05,$57,$FB,$AD,$7C,$DF,$F4
	.dc.b	$37,$33,$F0,$5C,$CF,$FB,$C7,$E1,$BE,$2E,$3F,$C7,$77,$46,$27,$FA
	.dc.b	$25,$70,$BF,$95,$39,$7F,$DE,$1F,$61,$FA,$27,$E7,$2F,$AE,$FE,$A8
	.dc.b	$E7,$1E,$CE,$70,$FF,$16,$63,$FE,$86,$BF,$EB,$5F,$2D,$FB,$97,$AE
	.dc.b	$18,$7F,$C8,$7F,$8A,$3F,$06,$FF,$2A,$37,$0F,$F2,$A3,$67,$FC,$B4
	.dc.b	$D7,$FB,$0F,$8C,$F0,$BF,$CD,$95,$FB,$EF,$91,$EB,$87,$F0,$27,$BF
	.dc.b	$34,$9F,$F1,$CE,$3F,$E7,$D6,$C6,$71,$FE,$0B,$87,$F8,$17,$6F,$E5
	.dc.b	$97,$F1,$F2,$F0,$7C,$1F,$AB,$D7,$F4,$93,$2F,$4B,$FE,$6F,$C3,$FD
	.dc.b	$B9,$87,$F3,$56,$CF,$F9,$7F,$0F,$EB,$2F,$F9,$86,$33,$8F,$F5,$69
	.dc.b	$7D,$F7,$C6,$25,$7C,$67,$B4,$FF,$20,$3E,$6B,$63,$FC,$C1,$F2,$3F
	.dc.b	$E0,$FF,$85,$FF,$6B,$36,$3E,$1F,$A9,$8F,$80,$FF,$24,$79,$7F,$E7
	.dc.b	$D3,$FD,$07,$5B,$C7,$F9,$4C,$7F,$A0,$72,$7C,$2F,$DA,$26,$3F,$69
	.dc.b	$FF,$48,$4B,$E5,$FF,$9C,$9F,$82,$8B,$3F,$53,$B7,$F0,$A7,$E1,$BC
	.dc.b	$BB,$1F,$63,$C1,$FC,$39,$1F,$C3,$BD,$BD,$DF,$E8,$A7,$E3,$9B,$1F
	.dc.b	$96,$FD,$0B,$FB,$EE,$9F,$C2,$DB,$C7,$FA,$95,$F4,$BF,$E3,$37,$35
	.dc.b	$FE,$2D,$FC,$13,$FD,$3D,$7F,$E3,$2E,$E4,$62,$34,$7F,$99,$1C,$3F
	.dc.b	$EA,$6F,$4F,$F9,$75,$3F,$3C,$FD,$A6,$5F,$E1,$5E,$3F,$F3,$27,$CE
	.dc.b	$7F,$85,$E3,$FC,$31,$5B,$F1,$7F,$B0,$F8,$D2,$3F,$D3,$86,$4F,$EA
	.dc.b	$CF,$A7,$E0,$BF,$C4,$1F,$2D,$FE,$B3,$D1,$B3,$7F,$EF,$F3,$98,$C7
	.dc.b	$BD,$C3,$E5,$FF,$C9,$95,$A6,$FB,$FE,$4C,$2E,$7F,$13,$0E,$93,$99
	.dc.b	$EF,$7F,$E0,$5A,$FF,$2C,$EE,$FC,$63,$FC,$93,$0A,$67,$E7,$A7,$E3
	.dc.b	$BF,$EC,$08,$C5,$F0,$AD,$E7,$C6,$F7,$9F,$A6,$7A,$73,$FF,$91,$72
	.dc.b	$ED,$8F,$C3,$9F,$F4,$9F,$CC,$6D,$FF,$26,$CE,$B1,$FF,$37,$FC,$9F
	.dc.b	$F3,$2B,$E0,$FF,$2B,$6F,$17,$18,$D4,$F0,$8C,$78,$4F,$A4,$FF,$30
	.dc.b	$63,$EF,$B4,$E6,$33,$18,$9D,$C7,$FA,$A2,$74,$FF,$9A,$67,$FC,$A3
	.dc.b	$A6,$9B,$1C,$BE,$1F,$EF,$B6,$75,$F1,$DF,$E8,$AC,$FD,$6D,$CF,$DE
	.dc.b	$DA,$F3,$76,$D7,$FC,$E7,$E0,$5C,$BF,$B7,$C7,$C7,$36,$53,$66,$78
	.dc.b	$7F,$01,$AF,$55,$E6,$7F,$B4,$67,$4D,$7E,$E8,$8E,$2F,$7D,$3F,$DD
	.dc.b	$9E,$37,$FA,$93,$87,$F6,$9F,$26,$8C,$57,$FA,$96,$30,$FE,$5D,$AF
	.dc.b	$FA,$14,$FD,$2B,$FE,$B5,$7B,$35,$F1,$D7,$FB,$EF,$C7,$F8,$CF,$FB
	.dc.b	$F7,$C8,$F9,$1E,$67,$F9,$37,$D1,$FF,$5B,$F8,$1F,$E7,$5F,$25,$18
	.dc.b	$D3,$FC,$BA,$DC,$BF,$D3,$5F,$86,$FF,$2C,$7F,$9C,$FC,$C7,$F7,$37
	.dc.b	$F7,$F1,$97,$FD,$07,$48,$C6,$1F,$C2,$EB,$85,$7F,$1F,$FA,$7F,$C5
	.dc.b	$E9,$FC,$05,$BE,$01,$A6,$B7,$7C,$8F,$96,$91,$8B,$FF,$00,$B5,$FF
	.dc.b	$1B,$2F,$43,$D9,$BF,$F9,$BF,$F0,$DB,$C0,$69,$AF,$41,$D7,$4F,$D0
	.dc.b	$D2,$33,$8F,$F8,$1A,$73,$18,$8F,$F0,$3F,$77,$32,$3F,$E0,$0F,$03
	.dc.b	$C3,$B4,$8F,$F3,$E3,$E1,$F0,$64,$7F,$84,$9E,$A4,$7F,$73,$8F,$FA
	.dc.b	$27,$C5,$F5,$92,$3F,$B7,$F4,$7F,$8E,$9B,$87,$DD,$EA,$FF,$80,$0D
	.dc.b	$9D,$4F,$FA,$53,$47,$C0,$FC,$6F,$1E,$4F,$FA,$45,$F9,$72,$FC,$26
	.dc.b	$9F,$19,$C4,$F9,$9A,$FF,$BB,$F3,$F9,$B3,$F3,$38,$7F,$5D,$D1,$F9
	.dc.b	$5F,$27,$EA,$9D,$BE,$4A,$61,$CD,$EC,$47,$B7,$99,$F7,$3E,$46,$47
	.dc.b	$F3,$BB,$FF,$6A,$66,$AF,$49,$27,$5D,$DF,$D0,$74,$FC,$0C,$FB,$BE
	.dc.b	$EB,$F0,$FF,$57,$2F,$FD,$64,$FC,$2F,$96,$91,$EF,$7E,$11,$1F,$E5
	.dc.b	$FD,$AF,$8B,$7A,$7F,$9B,$6C,$FF,$96,$9A,$FF,$A3,$FF,$3F,$FF,$0C
	.dc.b	$3F,$E5,$E5,$F5,$5F,$A3,$73,$1B,$F6,$B1,$3F,$A7,$FD,$0F,$ED,$CF
	.dc.b	$F9,$99,$AF,$FA,$6D,$BC,$3F,$E7,$9F,$CF,$FF,$51,$3F,$E6,$C7,$FC
	.dc.b	$D8,$FF,$A8,$5C,$B7,$C3,$FD,$A4,$6A,$7F,$8D,$7E,$16,$F0,$BF,$86
	.dc.b	$5A,$30,$FD,$03,$FB,$83,$85,$3F,$C8,$FF,$0C,$FE,$48,$DE,$5F,$EA
	.dc.b	$72,$9F,$E5,$0F,$F7,$93,$FB,$D3,$87,$0D,$73,$E1,$79,$20,$D3,$CF
	.dc.b	$F8,$F3,$E3,$FF,$88,$78,$36,$FF,$D1,$ED,$7F,$D8,$8D,$C7,$FD,$45
	.dc.b	$EE,$3B,$7F,$8C,$2F,$B7,$B1,$2F,$FD,$33,$39,$7F,$D3,$7F,$4C,$E9
	.dc.b	$FE,$40,$D8,$CC,$63,$EE,$A4,$60,$87,$FB,$5E,$51,$F4,$B8,$3F,$DF
	.dc.b	$F3,$F4,$4F,$A9,$F0,$59,$3E,$27,$E0,$52,$3F,$C0,$3C,$28,$FA,$2D
	.dc.b	$8C,$C6,$3E,$23,$E6,$12,$3B,$5B,$E9,$7F,$98,$FC,$2C,$9F,$C1,$78
	.dc.b	$6F,$1F,$FC,$C3,$18,$F1,$7F,$9F,$B7,$E5,$BF,$88,$FA,$52,$30,$7E
	.dc.b	$2F,$DB,$FF,$71,$F5,$7F,$5B,$6C,$67,$E1,$7E,$FB,$7C,$FF,$E3,$7F
	.dc.b	$65,$B9,$FE,$47,$D4,$EE,$FF,$22,$61,$FE,$EE,$BD,$E1,$F2,$7F,$26
	.dc.b	$FE,$BF,$5F,$E4,$9F,$AE,$78,$0F,$96,$FE,$D9,$18,$7F,$47,$5C,$2F
	.dc.b	$FC,$A1,$F0,$DF,$DD,$1F,$F1,$7B,$5F,$F1,$BB,$5F,$CF,$1F,$E3,$8F
	.dc.b	$7B,$F4,$70,$D7,$F5,$ED,$9F,$BC,$D7,$F5,$B9,$7E,$9E,$1F,$0F,$FD
	.dc.b	$9B,$E8,$FC,$9C,$C6,$93,$A3,$46,$FE,$16,$63,$F6,$A7,$EF,$3F,$73
	.dc.b	$F0,$BF,$F7,$73,$F7,$BF,$17,$FF,$24,$F2,$7C,$D8,$62,$9C,$E9,$FC
	.dc.b	$CE,$F8,$D7,$A0,$DB,$8B,$E5,$68,$C5,$71,$ED,$FF,$00,$C3,$E4,$7F
	.dc.b	$C1,$5F,$BD,$7F,$92,$31,$E9,$EA,$23,$19,$79,$63,$1F,$15,$FC,$6B
	.dc.b	$F0,$E7,$B1,$4F,$E2,$6E,$30,$FF,$33,$B5,$FF,$37,$B5,$FB,$2B,$E5
	.dc.b	$BA,$12,$3F,$69,$F9,$47,$DE,$C7,$DB,$F8,$7B,$BF,$FC,$36,$BE,$86
	.dc.b	$98,$8F,$F1,$36,$FF,$E1,$3C,$3F,$D8,$CD,$7F,$D3,$DA,$FC,$7F,$13
	.dc.b	$FC,$67,$39,$9F,$A7,$F4,$5F,$CC,$F2,$F9,$AD,$E3,$FE,$CC,$D9,$FF
	.dc.b	$32,$7E,$1C,$F7,$BF,$C5,$FD,$9F,$07,$0F,$AA,$DC,$3F,$B7,$36,$FF
	.dc.b	$CD,$9F,$87,$71,$7F,$81,$FE,$1D,$E3,$E9,$34,$9A,$FF,$8D,$3F,$0E
	.dc.b	$6B,$FE,$3E,$5F,$51,$4C,$47,$F9,$0B,$F0,$EF,$EA,$DA,$7F,$7F,$5F
	.dc.b	$77,$93,$5F,$43,$FE,$4F,$FC,$0F,$F1,$1A,$FF,$7A,$6B,$FB,$2E,$44
	.dc.b	$62,$3F,$DE,$8D,$7F,$C1,$9F,$0D,$FE,$62,$E2,$FC,$BB,$72,$CF,$D1
	.dc.b	$7F,$8D,$79,$3D,$37,$FC,$93,$E8,$BF,$27,$5F,$E1,$56,$C7,$EA,$1B
	.dc.b	$F1,$FF,$83,$F9,$CF,$F2,$FF,$DB,$FF,$78,$AF,$EC,$9C,$C8,$C4,$7F
	.dc.b	$87,$DA,$FE,$C9,$B0,$FC,$3B,$F9,$8C,$FE,$C6,$31,$EE,$FF,$9B,$27
	.dc.b	$FA,$6C,$FF,$3A,$77,$B7,$ED,$29,$F5,$FF,$E7,$26,$D7,$F4,$06,$BF
	.dc.b	$BD,$97,$E9,$3D,$23,$F8,$07,$A7,$FD,$47,$8F,$FD,$A6,$D8,$FF,$6D
	.dc.b	$B5,$FE,$4A,$D7,$E0,$FE,$57,$FC,$11,$FE,$1F,$F8,$7F,$FB,$51,$F6
	.dc.b	$37,$0F,$E1,$73,$BC,$5F,$83,$7F,$A5,$6F,$FA,$BB,$FE,$89,$6F,$CD
	.dc.b	$FF,$A3,$A3,$51,$FC,$5F,$DA,$7E,$1E,$F9,$AF,$F7,$5C,$61,$B1,$99
	.dc.b	$FF,$7A,$CE,$5B,$F8,$7F,$F0,$17,$FA,$F3,$5E,$CD,$CF,$F9,$67,$2B
	.dc.b	$F2,$6F,$EB,$7D,$1F,$D5,$7F,$26,$7D,$BF,$F6,$5F,$95,$FD,$D6,$7F
	.dc.b	$E4,$CF,$92,$FE,$AF,$5A,$D3,$FC,$6D,$C8,$8F,$AF,$1F,$E2,$F8,$C7
	.dc.b	$3F,$D4,$9C,$CF,$78,$C7,$77,$42,$47,$49,$7E,$19,$FD,$DF,$F1,$D7
	.dc.b	$CF,$D2,$57,$F7,$DF,$C4,$3F,$D0,$FB,$FF,$74,$7F,$DE,$D1,$FE,$41
	.dc.b	$D5,$18,$8D,$7C,$FF,$D0,$8C,$42,$7F,$C6,$9F,$13,$FE,$07,$F9,$5F
	.dc.b	$CC,$E1,$FE,$00,$6D,$7F,$C0,$1F,$3B,$FD,$E5,$CD,$F3,$9B,$49,$D7
	.dc.b	$CB,$7F,$AB,$E7,$31,$DF,$F1,$0F,$F0,$06,$5F,$C4,$3C,$8F,$F0,$B7
	.dc.b	$E5,$1D,$A0,$DF,$37,$F9,$25,$FE,$2D,$B9,$FF,$69,$AE,$53,$FE,$21
	.dc.b	$9C,$E7,$FE,$18,$7E,$6C,$7D,$AF,$CF,$1F,$75,$FE,$14,$FC,$43,$F0
	.dc.b	$B2,$E9,$46,$15,$70,$9E,$55,$FE,$1C,$8C,$47,$ED,$9B,$97,$F7,$DA
	.dc.b	$E3,$F1,$BB,$3F,$73,$3F,$E2,$5A,$FD,$CB,$5F,$DC,$B5,$FF,$0F,$FE
	.dc.b	$22,$F9,$4F,$F4,$2C,$7F,$30,$C7,$FD,$01,$3E,$E8,$C1,$F4,$DC,$7A
	.dc.b	$5E,$67,$96,$5F,$5E,$31,$7D,$1D,$75,$7F,$A2,$DB,$7B,$C3,$BD,$6D
	.dc.b	$B7,$FE,$34,$9F,$04,$F9,$9F,$4B,$5D,$9B,$C3,$FB,$47,$8D,$C6,$11
	.dc.b	$87,$5C,$5F,$C4,$9B,$3F,$EA,$25,$F6,$54,$8F,$FA,$99,$B9,$7D,$68
	.dc.b	$FD,$07,$0F,$AD,$FF,$25,$CD,$7F,$CB,$F1,$5F,$CE,$45,$E2,$FF,$1C
	.dc.b	$57,$39,$F0,$9F,$40,$E2,$7D,$2E,$07,$D4,$FF,$AB,$B0,$FE,$6E,$D9
	.dc.b	$FE,$6E,$D7,$F8,$63,$5F,$F3,$62,$F9,$3F,$F2,$46,$DF,$8D,$8C,$57
	.dc.b	$F1,$67,$A4,$66,$77,$3C,$9B,$5F,$9F,$A7,$BB,$8C,$3C,$F5,$67,$37
	.dc.b	$FE,$8D,$F9,$EF,$DF,$37,$0E,$79,$FF,$AF,$CB,$3A,$8F,$E8,$4D,$CB
	.dc.b	$F7,$1F,$88,$FF,$BE,$AF,$91,$FE,$99,$71,$5F,$D2,$7F,$11,$7D,$C7
	.dc.b	$A3,$CB,$FF,$53,$7C,$27,$82,$73,$87,$FB,$AE,$72,$DF,$9A,$FF,$24
	.dc.b	$36,$B8,$5F,$39,$CE,$1F,$D5,$FE,$9B,$FD,$F1,$7F,$D4,$9B,$1F,$E9
	.dc.b	$0D,$51,$88,$FE,$60,$D7,$FD,$4E,$FC,$87,$A7,$0F,$C6,$8E,$7A,$FF
	.dc.b	$19,$74,$FF,$07,$F9,$67,$D2,$FD,$36,$E3,$FE,$BD,$FC,$37,$C2,$FF
	.dc.b	$54,$8C,$E3,$FE,$06,$5F,$03,$FF,$00,$79,$7F,$D4,$8D,$8C,$D7,$F8
	.dc.b	$21,$B5,$FE,$08,$7F,$C8,$71,$D7,$E4,$BC,$5E,$17,$8D,$FA,$5F,$87
	.dc.b	$7D,$6D,$3F,$B7,$B6,$BF,$D8,$FF,$5D,$FF,$00,$5F,$A9,$19,$C7,$FD
	.dc.b	$3E,$3E,$8F,$FD,$44,$FF,$8D,$8C,$4F,$F0,$BF,$6F,$FA,$64,$EA,$7F
	.dc.b	$A7,$46,$27,$5C,$DF,$BC,$9F,$F1,$B7,$BD,$DE,$93,$9C,$BF,$D4,$F3
	.dc.b	$9E,$EF,$F0,$1E,$1F,$EF,$25,$F0,$7F,$A6,$C3,$F8,$6F,$B1,$67,$35
	.dc.b	$FE,$04,$FD,$29,$F2,$FF,$3D,$B1,$FE,$34,$6E,$9F,$E3,$46,$BF,$D9
	.dc.b	$C6,$7F,$E2,$6F,$A0,$FD,$AB,$6B,$FB,$63,$FE,$BE,$FC,$11,$F1,$3A
	.dc.b	$F8,$0F,$DF,$D7,$C4,$B7,$D4,$FA,$5F,$F5,$74,$72,$35,$19,$F9,$6F
	.dc.b	$77,$E1,$EF,$8A,$F9,$DA,$7F,$B5,$23,$17,$F4,$DB,$1F,$EF,$2E,$14
	.dc.b	$62,$3D,$9C,$8F,$A8,$FD,$AC,$7F,$B6,$63,$1E,$27,$CD,$FB,$5F,$D1
	.dc.b	$E3,$FE,$09,$9C,$BF,$BB,$FC,$3D,$EF,$7F,$01,$6C,$7A,$F0,$FF,$74
	.dc.b	$FE,$03,$F1,$23,$85,$7F,$1E,$EA,$FF,$40,$7D,$CF,$B1,$B3,$FD,$5D
	.dc.b	$B7,$FE,$9B,$F9,$05,$7C,$2D,$DF,$79,$F8,$0A,$FF,$41,$2E,$7F,$C3
	.dc.b	$39,$E8,$FF,$79,$B9,$CB,$CD,$6F,$C3,$FE,$4D,$7F,$BE,$17,$D5,$FE
	.dc.b	$3D,$CD,$7E,$79,$CB,$CB,$CB,$7D,$EF,$95,$B1,$FA,$76,$E7,$F5,$5A
	.dc.b	$FD,$4F,$A7,$7D,$47,$F2,$B6,$D7,$9B,$F0,$FF,$0D,$7E,$A2,$72,$FD
	.dc.b	$9F,$C3,$9F,$0D,$FC,$CD,$B7,$EE,$FC,$5C,$E6,$7E,$97,$CE,$FF,$25
	.dc.b	$B7,$2F,$FB,$5B,$E4,$7E,$A4,$E6,$31,$7F,$C0,$65,$F4,$FF,$B3,$7C
	.dc.b	$BC,$67,$33,$F3,$9F,$4E,$0F,$C3,$FF,$89,$3F,$AE,$9F,$D8,$46,$39
	.dc.b	$3C,$F2,$B7,$D2,$7B,$23,$81,$FE,$5D,$AE,$17,$FB,$9F,$1D,$F7,$DF
	.dc.b	$A8,$AF,$DA,$B6,$33,$3F,$E1,$D6,$CF,$F8,$9D,$FE,$73,$8F,$F2,$26
	.dc.b	$CF,$5A,$FE,$EF,$5F,$C0,$3D,$7F,$E2,$9F,$87,$FE,$8A,$3F,$DA,$8D
	.dc.b	$C3,$D7,$9A,$3F,$D1,$78,$7E,$49,$D3,$FC,$A9,$B5,$F0,$57,$DE,$F6
	.dc.b	$EF,$EE,$FF,$23,$FC,$3F,$F4,$B7,$F9,$C7,$E3,$63,$E0,$73,$87,$FB
	.dc.b	$9D,$CC,$FF,$38,$6C,$7C,$F5,$F8,$1F,$5E,$4E,$CF,$C0,$FC,$3F,$CB
	.dc.b	$47,$D6,$FD,$43,$19,$6F,$E5,$A6,$3F,$C1,$67,$84,$73,$9C,$C7,$1A
	.dc.b	$FF,$6A,$56,$DB,$19,$AF,$17,$CF,$7F,$43,$F8,$17,$CD,$53,$1F,$1F
	.dc.b	$71,$FD,$FC,$FF,$C0,$06,$FA,$FC,$7E,$6F,$93,$CF,$D9,$E9,$CF,$6B
	.dc.b	$8B,$DF,$C8,$91,$83,$AB,$F8,$4F,$BF,$FE,$CF,$B7,$C3,$6F,$27,$F6
	.dc.b	$C7,$F9,$8B,$7D,$BF,$C5,$FE,$1C,$FA,$DF,$CC,$37,$1F,$F3,$93,$70
	.dc.b	$FF,$3E,$35,$FE,$B0,$1F,$E2,$AF,$D3,$6D,$FF,$7C,$79,$37,$CB,$FE
	.dc.b	$69,$F8,$77,$EE,$2B,$F5,$BE,$1A,$1F,$81,$6B,$C4,$DB,$32,$FE,$81
	.dc.b	$5B,$F6,$B1,$87,$2C,$3F,$B0,$D6,$BF,$0D,$78,$FF,$D0,$DB,$19,$FA
	.dc.b	$CC,$4E,$7E,$5F,$5F,$33,$6B,$93,$FB,$2F,$C3,$4D,$3F,$6B,$9B,$F0
	.dc.b	$5C,$57,$E4,$7D,$87,$DB,$7F,$05,$6C,$7F,$A7,$DB,$B7,$FA,$A5,$AF
	.dc.b	$CE,$7E,$CC,$7F,$81,$DB,$1F,$EE,$F7,$8F,$1F,$F8,$A9,$CF,$03,$F0
	.dc.b	$DF,$E6,$77,$F2,$1B,$7F,$43,$CC,$7C,$4F,$F6,$79,$CE,$3F,$C9,$30
	.dc.b	$E5,$F0,$99,$F9,$7D,$37,$2F,$A7,$F7,$9F,$9F,$CF,$ED,$B7,$0E,$4F
	.dc.b	$F6,$59,$4F,$F4,$28,$D4,$63,$04,$7F,$8D,$27,$9F,$8B,$FD,$EE,$31
	.dc.b	$7D,$23,$F6,$95,$FD,$DF,$C4,$FE,$9B,$87,$F9,$25,$B8,$7F,$85,$DA
	.dc.b	$FF,$8C,$7C,$F7,$E0,$FF,$E3,$6C,$7F,$94,$B6,$FF,$C9,$F7,$CA,$73
	.dc.b	$F1,$5F,$E4,$5A,$FF,$33,$E1,$FD,$1B,$0E,$7F,$1D,$9F,$C0,$74,$CC
	.dc.b	$BF,$C0,$76,$F8,$FC,$61,$C6,$31,$3A,$F9,$1F,$F0,$94,$7F,$9D,$A7
	.dc.b	$D0,$D7,$FC,$C6,$D7,$FA,$D3,$FD,$29,$BE,$B7,$9E,$7D,$0F,$A7,$97
	.dc.b	$A9,$2F,$FC,$E1,$7F,$D5,$AF,$BF,$C1,$FB,$CE,$13,$EA,$7E,$17,$A7
	.dc.b	$FE,$B5,$F5,$3F,$D3,$73,$E4,$8C,$7C,$04,$8F,$F5,$B4,$EA,$FC,$F5
	.dc.b	$97,$E1,$47,$C0,$24,$F9,$1F,$87,$EF,$7D,$13,$E9,$36,$3E,$6F,$97
	.dc.b	$F5,$4F,$AD,$F0,$16,$3E,$8F,$C4,$7C,$C2,$3F,$01,$F9,$FD,$FF,$74
	.dc.b	$FB,$CD,$8F,$E2,$3F,$84,$3E,$83,$FC,$9C,$D8,$CF,$CF,$E2,$73,$EE
	.dc.b	$FF,$94,$6B,$FB,$6D,$7F,$94,$5B,$AF,$F2,$CE,$C4,$62,$3F,$D0,$D0
	.dc.b	$FC,$33,$FD,$D5,$5F,$E6,$7F,$57,$F9,$3B,$FC,$92,$33,$5F,$EA,$09
	.dc.b	$CB,$FE,$1A,$97,$E0,$CF,$0F,$FC,$43,$DF,$FE,$22,$9C,$C7,$FB,$BE
	.dc.b	$F8,$CE,$6B,$85,$7C,$2F,$73,$FC,$53,$F2,$9F,$EA,$8C,$7F,$C7,$BA
	.dc.b	$7F,$85,$63,$1A,$7B,$2B,$7E,$CF,$EF,$EB,$4D,$F7,$F3,$87,$2F,$7A
	.dc.b	$E1,$F1,$26,$DF,$F7,$6C,$7C,$1F,$1A,$FC,$2F,$FC,$7F,$18,$6E,$5D
	.dc.b	$F1,$EB,$5B,$F9,$FE,$F9,$BF,$02,$FE,$4B,$8F,$AE,$3F,$D0,$3E,$66
	.dc.b	$A7,$38,$FF,$AE,$7E,$39,$3C,$9F,$8B,$4B,$E5,$18,$6C,$EF,$93,$87
	.dc.b	$80,$F9,$DF,$ED,$B8,$7F,$26,$C1,$39,$8C,$C6,$27,$71,$FE,$96,$9D
	.dc.b	$3F,$DD,$A7,$FC,$CD,$C3,$A6,$C7,$2F,$8A,$FF,$39,$B6,$75,$F5,$3F
	.dc.b	$ED,$5C,$FD,$2D,$D3,$E0,$36,$BC,$DD,$D5,$F9,$5E,$15,$CB,$FB,$0C
	.dc.b	$7F,$40,$C3,$E1,$D6,$F7,$5F,$3E,$1F,$B5,$B5,$E3,$F5,$E1,$F6,$E3
	.dc.b	$C1,$AB,$EA,$FF,$C9,$39,$F5,$DB,$FA,$54,$7F,$A5,$A3,$07,$FA,$41
	.dc.b	$DB,$F5,$63,$7E,$27,$49,$E1,$39,$38,$78,$B9,$F1,$64,$74,$C3,$C5
	.dc.b	$3F,$0F,$AD,$F8,$5F,$B2,$BF,$DA,$D1,$8D,$FF,$E0,$07,$77,$EB,$EB
	.dc.b	$55,$FE,$E5,$F9,$5F,$F2,$6B,$FC,$3B,$E9,$37,$87,$FB,$96,$73,$3B
	.dc.b	$F4,$F2,$FE,$C2,$AE,$7F,$17,$CB,$FC,$2B,$67,$FD,$C1,$B7,$FA,$61
	.dc.b	$B3,$A8,$79,$9C,$9B,$3F,$E3,$06,$BF,$E3,$85,$E6,$F8,$BD,$1F,$ED
	.dc.b	$7D,$77,$CE,$34,$C7,$6C,$C3,$FA,$7B,$F9,$AA,$FE,$E6,$DA,$FF,$5F
	.dc.b	$4F,$F5,$4E,$FE,$84,$7F,$B5,$F7,$F7,$F0,$24,$7F,$6E,$7C,$5C,$52
	.dc.b	$3F,$AE,$BD,$48,$FE,$C3,$B7,$FD,$BB,$7F,$31,$B1,$FD,$62,$3D,$D9
	.dc.b	$FF,$8A,$23,$FD,$FA,$D7,$FD,$8F,$3C,$1B,$D1,$FC,$BD,$DF,$AD,$CB
	.dc.b	$D8,$91,$C9,$EB,$F5,$92,$3F,$81,$EF,$F2,$A3,$E3,$62,$69,$8F,$13
	.dc.b	$CA,$4A,$F2,$FB,$5E,$83,$89,$1E,$B4,$8E,$BF,$07,$EE,$F8,$39,$5E
	.dc.b	$E8,$F8,$24,$70,$7C,$2F,$9B,$F8,$BF,$84,$48,$F8,$AF,$C8,$F9,$C9
	.dc.b	$1E,$07,$D6,$47,$F9,$0F,$5B,$F0,$B7,$D9,$7E,$71,$7D,$2D,$31,$1F
	.dc.b	$A0,$FD,$0F,$FD,$2B,$E2,$7F,$96,$5B,$D1,$FE,$4D,$6E,$2C,$51,$FC
	.dc.b	$A2,$BF,$CC,$1F,$85,$FE,$72,$BF,$D1,$EF,$E8,$23,$3E,$07,$EB,$A3
	.dc.b	$53,$FE,$9A,$5E,$F6,$49,$1C,$BD,$EF,$DB,$AF,$C2,$D9,$E1,$B1,$FE
	.dc.b	$D9,$6B,$FE,$DF,$7F,$A7,$B9,$EE,$FE,$82,$F3,$6C,$6A,$3F,$A5,$B5
	.dc.b	$FF,$5C,$FE,$15,$F7,$DF,$F7,$33,$5F,$F7,$63,$73,$F0,$FE,$21,$FE
	.dc.b	$A0,$9F,$F5,$8F,$B3,$F9,$16,$F5,$7F,$10,$83,$7A,$F1,$DB,$2B,$8B
	.dc.b	$BE,$EF,$E2,$FF,$46,$E8,$BD,$32,$FF,$D7,$F3,$96,$F5,$F0,$A9,$18
	.dc.b	$3E,$47,$C1,$FC,$B2,$3F,$2F,$E1,$3E,$71,$F1,$88,$FF,$6A,$E1,$47
	.dc.b	$CC,$FC,$31,$F0,$9F,$61,$B5,$FE,$9A,$E2,$FA,$B1,$88,$46,$3D,$1F
	.dc.b	$E0,$FE,$64,$9F,$E0,$EF,$F0,$AF,$91,$FE,$17,$F2,$76,$7F,$85,$C6
	.dc.b	$1F,$E1,$BF,$59,$FC,$FB,$D8,$7D,$57,$F0,$B9,$EF,$D9,$FE,$AC,$9C
	.dc.b	$F9,$BF,$C5,$AF,$87,$E5,$5F,$C5,$75,$FC,$6D,$23,$17,$C5,$B8,$7F
	.dc.b	$84,$F1,$FC,$8D,$FE,$4B,$CF,$FF,$5A,$CF,$C1,$BF,$F5,$DB,$FE,$16
	.dc.b	$F8,$CF,$CE,$CE,$A3,$CA,$7F,$B0,$1F,$4B,$C4,$C5,$F0,$23,$93,$FC
	.dc.b	$43,$EF,$74,$32,$F4,$3F,$C4,$B1,$97,$A7,$9E,$FC,$33,$EA,$65,$FA
	.dc.b	$C8,$C7,$17,$F5,$79,$FE,$E3,$39,$77,$9F,$FB,$5A,$B9,$7C,$66,$F9
	.dc.b	$66,$3D,$AD,$3F,$D8,$7E,$87,$FD,$79,$3E,$27,$F1,$6D,$E7,$D7,$8D
	.dc.b	$23,$4D,$7F,$D9,$52,$EA,$FE,$0F,$F8,$53,$B6,$92,$BE,$34,$3D,$0D
	.dc.b	$9F,$C1,$1C,$47,$F9,$19,$AF,$F9,$39,$AF,$DA,$1F,$85,$7F,$26,$69
	.dc.b	$FC,$61,$FE,$53,$B3,$1F,$C9,$70,$7C,$DF,$C2,$BF,$83,$5F,$FB,$8D
	.dc.b	$74,$FF,$81,$5B,$1F,$59,$DC,$FF,$20,$F3,$FA,$7C,$87,$F8,$63,$03
	.dc.b	$49,$7F,$DD,$FF,$0A,$7C,$8F,$F7,$76,$C7,$E3,$7D,$A4,$62,$3F,$DE
	.dc.b	$DF,$33,$FE,$6C,$75,$F8,$56,$73,$FE,$63,$68,$C7,$4A,$93,$9C,$7F
	.dc.b	$C4,$1D,$AF,$27,$FC,$45,$F8,$57,$F8,$62,$FE,$67,$12,$31,$1F,$A8
	.dc.b	$8C,$7B,$FF,$E2,$19,$CC,$FE,$47,$F0,$A7,$D6,$7F,$8B,$56,$F4,$C4
	.dc.b	$7F,$8C,$7F,$0A,$FE,$AD,$B7,$E6,$5F,$EA,$BC,$1F,$A1,$6C,$67,$D9
	.dc.b	$FF,$23,$FD,$CF,$E5,$B2,$FF,$69,$35,$FF,$68,$B5,$F8,$5C,$C8,$C4
	.dc.b	$7C,$4E,$17,$CF,$FF,$8A,$5B,$1F,$4D,$AF,$F0,$4E,$07,$CB,$3E,$97
	.dc.b	$FC,$61,$E7,$7F,$AC,$76,$FA,$6D,$D7,$E9,$73,$23,$18,$B1,$7E,$15
	.dc.b	$FE,$50,$8F,$E0,$7F,$63,$FE,$58,$7E,$83,$5F,$DF,$35,$E4,$D9,$7E
	.dc.b	$14,$F4,$7F,$CB,$CF,$FA,$46,$BF,$87,$B5,$F6,$3B,$7F,$8F,$66,$FC
	.dc.b	$2B,$FA,$D4,$7F,$24,$8C,$7A,$DF,$E7,$A6,$C7,$F1,$AC,$3D,$BF,$73
	.dc.b	$FC,$0F,$83,$FD,$7E,$D9,$FF,$63,$B5,$FD,$D3,$5F,$C7,$7E,$13,$FC
	.dc.b	$2D,$FE,$4A,$FD,$FB,$FE,$48,$DE,$0F,$E4,$0E,$E7,$EB,$7D,$F3,$E6
	.dc.b	$FF,$CF,$CB,$E4,$69,$88,$FE,$AD,$F7,$2F,$9F,$FF,$6A,$FB,$28,$C6
	.dc.b	$1F,$ED,$E8,$C4,$EA,$73,$3F,$C4,$F9,$9F,$83,$FF,$0E,$D3,$F0,$2D
	.dc.b	$9F,$D9,$B7,$07,$E1,$1F,$C3,$2B,$F7,$1B,$BF,$08,$7D,$8F,$FA,$A2
	.dc.b	$31,$D7,$67,$F2,$59,$7F,$55,$9E,$15,$9F,$98,$FA,$35,$FE,$B2,$AF
	.dc.b	$EC,$35,$D6,$72,$6B,$EC,$BA,$12,$7F,$B0,$66,$FC,$5F,$F1,$1C,$BF
	.dc.b	$DD,$3F,$08,$7D,$37,$F7,$F6,$DF,$06,$BF,$DB,$2D,$0F,$A8,$E0,$7D
	.dc.b	$07,$13,$F3,$DA,$FF,$3F,$9E,$33,$FD,$06,$3F,$5B,$F3,$3F,$EE,$2C
	.dc.b	$3F,$91,$B7,$83,$3F,$35,$FE,$13,$FC,$23,$FB,$4C,$7F,$84,$5B,$7F
	.dc.b	$C9,$AB,$83,$6B,$87,$C1,$7F,$86,$FE,$51,$BD,$F1,$E5,$FC,$69,$CB
	.dc.b	$67,$81,$1D,$7D,$57,$DB,$FC,$1B,$F0,$10,$E4,$43,$EC,$74,$6F,$EC
	.dc.b	$25,$7D,$F9,$E3,$5C,$27,$A7,$BB,$2B,$F5,$F5,$FD,$69,$FD,$7B,$5F
	.dc.b	$C1,$B5,$FF,$12,$FE,$0C,$C7,$FC,$8B,$1F,$CA,$71,$CE,$5F,$E2,$AF
	.dc.b	$3F,$CD,$3F,$D3,$EF,$E0,$BB,$31,$7C,$A7,$70,$9E,$DC,$7F,$D5,$E7
	.dc.b	$31,$C9,$B7,$DF,$0F,$2E,$1C,$DB,$7F,$83,$C3,$A4,$FA,$1F,$F0,$36
	.dc.b	$1D,$9F,$C4,$35,$CE,$CF,$13,$94,$FF,$76,$6C,$66,$78,$B5,$F0,$7D
	.dc.b	$4F,$FA,$71,$34,$8F,$FA,$81,$F7,$63,$F9,$BE,$5B,$3F,$E1,$B9,$AF
	.dc.b	$F9,$96,$2B,$FD,$EE,$2F,$3F,$FA,$C3,$5F,$97,$39,$3E,$81,$C4,$FA
	.dc.b	$5C,$3F,$8C,$F9,$3F,$F5,$AC,$7F,$0E,$E0,$46,$23,$F2,$51,$8C,$7F
	.dc.b	$13,$39,$AF,$E5,$51,$FE,$C0,$DF,$FD,$83,$18,$8F,$CA,$4E,$5F,$E3
	.dc.b	$7E,$37,$FB,$21,$FE,$2C,$DC,$3F,$CD,$FE,$93,$F5,$FF,$E6,$55,$A3
	.dc.b	$94,$FF,$B8,$63,$2F,$F9,$DB,$07,$E9,$8E,$AD,$C7,$FB,$9B,$E3,$EC
	.dc.b	$E7,$0F,$E4,$DF,$3C,$9D,$B6,$33,$1C,$2F,$2F,$8B,$57,$E1,$6F,$93
	.dc.b	$EF,$FA,$37,$FC,$00,$C3,$FB,$4C,$E5,$FE,$69,$19,$D3,$D0,$DB,$F9
	.dc.b	$2D,$9F,$F5,$B5,$71,$AD,$FD,$0F,$C2,$5F,$B6,$3E,$F3,$EC,$46,$1B
	.dc.b	$19,$BF,$D2,$36,$7F,$B7,$F8,$0E,$2F,$8E,$DA,$FF,$4D,$56,$F3,$FB
	.dc.b	$B3,$9B,$FF,$01,$7E,$13,$FE,$D0,$BF,$F0,$3B,$63,$EA,$7C,$B3,$EB
	.dc.b	$BE,$44,$E7,$0F,$C3,$CB,$EB,$3F,$18,$DC,$3F,$C0,$AD,$C3,$FC,$0A
	.dc.b	$D7,$FA,$CE,$2F,$B0,$FC,$2F,$63,$F1,$8F,$F2,$7E,$1F,$D8,$EB,$4F
	.dc.b	$F5,$A8,$CD,$FE,$27,$F3,$FF,$F7,$D4,$FF,$5E,$EC,$7D,$1F,$E2,$A7
	.dc.b	$35,$FD,$66,$73,$5F,$DB,$FF,$0A,$FF,$77,$BF,$E1,$C6,$CF,$F8,$71
	.dc.b	$AF,$F7,$06,$BF,$DF,$BF,$42,$FE,$97,$3F,$DE,$5B,$1F,$9D,$5E,$97
	.dc.b	$E7,$1C,$F0,$7E,$C3,$0F,$F4,$BD,$FF,$B2,$BC,$3D,$BF,$E0,$38,$CE
	.dc.b	$1F,$E5,$99,$CB,$FA,$9C,$5F,$35,$FB,$D6,$DF,$AE,$30,$7F,$05,$E6
	.dc.b	$7E,$25,$FE,$FC,$9F,$F7,$24,$63,$8F,$F6,$EE,$BB,$6D,$6F,$1F,$DF
	.dc.b	$FE,$ED,$FE,$C5,$C3,$FD,$C3,$F5,$8F,$A9,$E3,$39,$AF,$F3,$84,$E6
	.dc.b	$BB,$FC,$43,$E5,$3C,$8D,$8F,$F0,$26,$68,$C5,$FF,$B1,$E3,$15,$EB
	.dc.b	$FB,$4F,$AF,$C5,$FE,$BB,$6E,$1F,$E3,$2F,$D1,$D9,$7B,$27,$38,$FE
	.dc.b	$A7,$EE,$3F,$C6,$B7,$FE,$FA,$FB,$FF,$C7,$CF,$F1,$C9,$CB,$FE,$89
	.dc.b	$8C,$E3,$FE,$C9,$D1,$EC,$7A,$89,$87,$AA,$32,$FB,$3F,$4E,$FF,$48
	.dc.b	$3F,$41,$B8,$7F,$23,$EF,$7D,$C7,$D2,$9C,$E3,$FE,$D7,$F9,$FF,$D0
	.dc.b	$AC,$3F,$D8,$DF,$18,$FC,$37,$FD,$89,$1F,$72,$B6,$EA,$33,$97,$F3
	.dc.b	$8F,$DD,$BF,$DF,$53,$F4,$1B,$1F,$3D,$AF,$FB,$4C,$FE,$20,$E1,$FF
	.dc.b	$5A,$FE,$67,$FA,$C7,$FA,$1B,$71,$FE,$9B,$F9,$2F,$DA,$7F,$DE,$11
	.dc.b	$9B,$FF,$5E,$FB,$19,$7D,$7F,$87,$7D,$87,$F2,$B1,$9C,$3F,$D6,$13
	.dc.b	$97,$FA,$8F,$CC,$7F,$2F,$7C,$DD,$3F,$EC,$4F,$4B,$F4,$13,$99,$FF
	.dc.b	$65,$B5,$FE,$55,$FB,$5F,$E5,$5F,$EB,$6D,$CF,$FB,$2F,$E5,$7F,$AC
	.dc.b	$9C,$F7,$7F,$3B,$F9,$4F,$EA,$90,$AF,$A8,$FF,$BE,$63,$31,$88,$FF
	.dc.b	$68,$FD,$0F,$1F,$5F,$97,$E0,$5F,$D0,$AC,$D2,$70,$7E,$93,$F7,$9F
	.dc.b	$E3,$78,$F6,$DD,$18,$9F,$AD,$7C,$9F,$DE,$C6,$72,$F2,$7A,$9E,$93
	.dc.b	$D5,$F8,$1D,$A4,$FA,$DE,$73,$FD,$EB,$A5,$F6,$3E,$C9,$C9,$C3,$3F
	.dc.b	$83,$3A,$D7,$E3,$7D,$7F,$C7,$FB,$0B,$7B,$3C,$EE,$5C,$65,$FE,$ED
	.dc.b	$FC,$CD,$F7,$F9,$8F,$F7,$D4,$62,$77,$0F,$C3,$FF,$4F,$1F,$97,$7F
	.dc.b	$93,$E7,$F4,$5C,$CF,$B9,$CC,$EF,$D4,$FE,$ED,$18,$9D,$7C,$BE,$26
	.dc.b	$57,$F2,$8F,$C4,$FF,$E0,$0B,$E2,$BE,$1C,$89,$CE,$1F,$2C,$F0,$7C
	.dc.b	$F7,$63,$BF,$CF,$5C,$3D,$7F,$86,$3F,$05,$FF,$11,$FE,$03,$F5,$27
	.dc.b	$31,$8C,$3C,$FC,$1E,$8F,$85,$B5,$E3,$7E,$C7,$23,$E5,$69,$B9,$F9
	.dc.b	$FC,$DF,$E3,$18,$FF,$0B,$7E,$1E,$F0,$A9,$8F,$AD,$B7,$F4,$27,$FD
	.dc.b	$3C,$FF,$9E,$5B,$9F,$FA,$F5,$AF,$F0,$90,$E1,$FE,$19,$B3,$F0,$1F
	.dc.b	$F1,$33,$6B,$B7,$8F,$89,$F2,$FD,$37,$9B,$EC,$6D,$78,$DC,$3E,$CD
	.dc.b	$BC,$7D,$5F,$B1,$CF,$F9,$0B,$F9,$68,$CC,$FF,$1D,$9F,$06,$7F,$5C
	.dc.b	$F3,$E3,$F1,$3E,$13,$E2,$BF,$E3,$BF,$CF,$3F,$45,$3F,$A8,$F7,$FF
	.dc.b	$8B,$28,$CA,$8C,$FA,$EF,$23,$FC,$88,$DC,$BF,$36,$EA,$33,$9F,$F9
	.dc.b	$1A,$3F,$87,$DF,$CB,$6B,$FC,$D5,$AF,$F9,$17,$E7,$1F,$13,$FD,$F1
	.dc.b	$B8,$76,$BC,$4F,$8B,$C7,$93,$E2,$F6,$6E,$1F,$0F,$3F,$E1,$8F,$FA
	.dc.b	$03,$F3,$CF,$F1,$F7,$0F,$F1,$D9,$CB,$FC,$56,$33,$7E,$5F,$1F,$DE
	.dc.b	$7C,$0F,$83,$B1,$FE,$12,$FD,$7F,$FB,$AF,$88,$FD,$6D,$F8,$17,$F8
	.dc.b	$52,$73,$18,$AF,$F0,$CF,$E0,$6F,$C0,$BF,$C3,$CD,$8F,$F0,$5B,$76
	.dc.b	$FF,$05,$35,$FE,$8F,$F8,$FF,$FA,$6F,$6F,$E5,$8D,$AF,$E4,$79,$AF
	.dc.b	$95,$6A,$7F,$D2,$FF,$97,$BE,$47,$FC,$53,$F4,$FE,$5C,$BF,$D3,$D3
	.dc.b	$9A,$FF,$17,$CE,$6F,$F9,$D3,$63,$BB,$69,$F3,$6B,$E5,$1F,$EB,$09
	.dc.b	$D6,$4E,$1F,$E1,$F3,$FD,$89,$AF,$A9,$AF,$FA,$25,$B1,$FB,$1E,$4F
	.dc.b	$F2,$0F,$67,$F8,$62,$BF,$D7,$B5,$FC,$6B,$3F,$EA,$13,$97,$FA,$97
	.dc.b	$EA,$3F,$E8,$DA,$FE,$E2,$D7,$E4,$B5,$FF,$4C,$37,$07,$EE,$FF,$B0
	.dc.b	$AF,$EB,$8D,$D7,$8C,$FF,$95,$23,$16,$FB,$0F,$8B,$F3,$5F,$EC,$E7
	.dc.b	$FC,$BB,$EE,$5E,$0F,$F5,$16,$3F,$D7,$27,$2F,$F4,$E6,$DF,$F3,$56
	.dc.b	$E5,$FD,$E9,$EB,$3B,$F9,$0F,$F0,$23,$CF,$E4,$1E,$07,$FB,$C6,$91
	.dc.b	$89,$FE,$41,$87,$9E,$30,$FF,$23,$9D,$D7,$1F,$DC,$FF,$A0,$5F,$95
	.dc.b	$B5,$E5,$BE,$38,$F9,$67,$2F,$F8,$37,$9F,$FB,$F5,$1A,$7E,$2D,$C7
	.dc.b	$91,$FE,$D3,$E9,$FF,$36,$C7,$F3,$56,$BF,$E3,$F6,$BF,$E0,$46,$BF
	.dc.b	$AF,$FC,$03,$91,$CF,$E4,$AE,$7F,$DD,$2B,$F9,$09,$FC,$8B,$0F,$F2
	.dc.b	$05,$F3,$F1,$3C,$5F,$03,$FE,$E2,$AF,$E1,$B3,$E3,$8C,$19,$FC,$1F
	.dc.b	$D1,$F7,$7E,$96,$E3,$F6,$FE,$FB,$F4,$BB,$7D,$96,$C7,$AE,$BF,$0D
	.dc.b	$18,$C3,$FD,$D5,$F8,$03,$C8,$FF,$1C,$FC,$CF,$FB,$56,$73,$F3,$D8
	.dc.b	$F9,$3F,$E0,$77,$BC,$BF,$C4,$F7,$BE,$D4,$FF,$B8,$63,$0D,$AF,$E1
	.dc.b	$7E,$47,$E0,$A7,$38,$75,$C7,$FD,$F9,$39,$AF,$CC,$67,$FE,$F5,$BF
	.dc.b	$27,$4A,$31,$3D,$3A,$79,$4E,$7E,$13,$A6,$3F,$C8,$F5,$F3,$DF,$F9
	.dc.b	$EF,$F1,$2F,$F3,$56,$5E,$FA,$D0,$71,$1A,$9E,$1F,$19,$D2,$0D,$F9
	.dc.b	$8C,$39,$E9,$FE,$A3,$87,$E8,$AB,$FD,$81,$E8,$7E,$92,$11,$87,$53
	.dc.b	$C1,$FF,$04,$47,$F6,$7E,$2E,$0D,$8E,$9E,$92,$72,$D8,$F0,$75,$7F
	.dc.b	$A6,$74,$F5,$B7,$4F,$84,$D9,$FF,$30,$E1,$FE,$72,$7E,$39,$FA,$1E
	.dc.b	$0D,$3E,$8E,$9B,$3B,$8F,$91,$7F,$A9,$AE,$9C,$2F,$C0,$1F,$05,$86
	.dc.b	$DF,$69,$D4,$F8,$F5,$FE,$1B,$EC,$5F,$D1,$DF,$7D,$9C,$93,$4D,$46
	.dc.b	$FF,$00,$B9,$7A,$B0,$61,$CA,$30,$A7,$C9,$F9,$2F,$FA,$FF,$BF,$F9
	.dc.b	$16,$A8,$C7,$47,$F2,$B8,$C6,$1F,$E7,$2F,$CF,$1F,$81,$6A,$3F,$D1
	.dc.b	$7F,$9E,$BC,$DC,$42,$B6,$F1,$AC,$D6,$E3,$FC,$FA,$BE,$4F,$F1,$10
	.dc.b	$8C,$3F,$8B,$6B,$FC,$C1,$AF,$F9,$45,$7D,$EF,$62,$47,$6E,$5F,$2E
	.dc.b	$1F,$E6,$78,$F6,$7C,$36,$0C,$FD,$C4,$8F,$CD,$4A,$3C,$AD,$8C,$C6
	.dc.b	$38,$3A,$92,$3E,$47,$1F,$C7,$E7,$7D,$9B,$E3,$B0,$DA,$FE,$B0,$D7
	.dc.b	$FA,$C7,$E6,$8F,$80,$E4,$D8,$FE,$A3,$68,$C4,$7F,$51,$6B,$FE,$FD
	.dc.b	$8D,$4F,$ED,$57,$E0,$4F,$A7,$F3,$36,$BF,$B5,$35,$FF,$4F,$75,$3F
	.dc.b	$53,$79,$DF,$83,$8C,$3C,$AE,$57,$FB,$02,$72,$DF,$DC,$FA,$76,$7F
	.dc.b	$B7,$DB,$D7,$FE,$34,$6C,$FF,$01,$F9,$07,$DE,$77,$6C,$7F,$72,$C1
	.dc.b	$18,$8F,$F4,$A4,$62,$7F,$8A,$36,$38,$7D,$87,$DB,$7D,$AD,$C7,$FD
	.dc.b	$E4,$DC,$7F,$C8,$FF,$65,$FE,$58,$AF,$63,$5E,$ED,$7F,$DD,$E7,$F1
	.dc.b	$97,$0F,$2F,$43,$C7,$F9,$A3,$EE,$BE,$53,$67,$E4,$35,$FF,$67,$7B
	.dc.b	$CF,$CE,$7F,$82,$C7,$FB,$B2,$31,$E3,$F2,$B9,$39,$AD,$BF,$6B,$F4
	.dc.b	$2F,$E6,$5A,$B4,$FF,$31,$C6,$1B,$E8,$FD,$DF,$89,$F8,$57,$FC,$CD
	.dc.b	$F9,$0F,$BB,$83,$FC,$CD,$DF,$F3,$A7,$33,$FE,$6E,$6B,$FC,$EF,$9B
	.dc.b	$F8,$B3,$F8,$CF,$B9,$7E,$05,$F8,$B6,$C7,$E2,$67,$7A,$79,$E0,$3F
	.dc.b	$0B,$CE,$9F,$41,$B1,$A6,$BB,$E3,$FE,$2F,$F3,$BF,$EE,$38,$FC,$B7
	.dc.b	$C8,$BD,$9F,$CE,$36,$3B,$E3,$E1,$55,$C5,$FF,$66,$61,$FE,$FD,$77
	.dc.b	$8F,$8F,$E7,$7F,$92,$CE,$67,$7F,$11,$FA,$9F,$D8,$9F,$45,$49,$3C
	.dc.b	$8F,$7F,$E2,$BF,$DD,$1C,$07,$0F,$F9,$51,$AF,$FB,$D5,$AF,$E5,$1A
	.dc.b	$F4,$F7,$7F,$9C,$C1,$A6,$3D,$DF,$E7,$BF,$91,$FF,$B7,$EB,$C3,$F9
	.dc.b	$6F,$F3,$06,$FA,$5A,$F9,$8E,$7F,$0F,$27,$CF,$92,$3B,$FB,$9E,$0F
	.dc.b	$B3,$FE,$99,$F4,$4F,$82,$F9,$9C,$DF,$EF,$BF,$A2,$D6,$5C,$BB,$AB
	.dc.b	$FE,$FB,$CD,$8F,$AA,$7F,$A9,$ED,$F3,$2B,$CF,$C5,$83,$FA,$BF,$17
	.dc.b	$F4,$CC,$BE,$E3,$FD,$32,$7F,$72,$D7,$F1,$ED,$7E,$37,$B7,$FD,$62
	.dc.b	$3F,$C5,$D3,$CB,$DA,$EA,$76,$E2,$F6,$EC,$FA,$2A,$4C,$7E,$DE,$1F
	.dc.b	$BF,$C3,$C7,$F8,$3F,$9B,$C8,$95,$A8,$FE,$DF,$F9,$8F,$FA,$5A,$3F
	.dc.b	$81,$7E,$37,$FD,$75,$FE,$0E,$DB,$FC,$C4,$EE,$51,$F0,$3D,$3F,$86
	.dc.b	$48,$F1,$F4,$7F,$7C,$FC,$57,$FB,$DC,$7F,$42,$FC,$13,$FC,$64,$DF
	.dc.b	$2B,$5F,$05,$23,$F2,$B5,$C3,$1F,$D4,$37,$C3,$FE,$4A,$FF,$8A,$67
	.dc.b	$7F,$09,$C5,$F2,$CE,$AF,$C5,$3E,$3F,$80,$FC,$3F,$E4,$FF,$EF,$5B
	.dc.b	$FD,$37,$72,$31,$1C,$8F,$E0,$F8,$FF,$88,$DF,$2F,$C0,$F3,$7C,$F7
	.dc.b	$EF,$8F,$81,$F8,$FF,$E5,$3F,$31,$FE,$77,$5C,$F1,$46,$35,$F6,$D7
	.dc.b	$0A,$F8,$2D,$7C,$43,$1F,$E0,$53,$FC,$6F,$67,$EE,$4F,$95,$FE,$D8
	.dc.b	$DB,$FE,$D4,$D7,$9F,$DE,$7F,$15,$6F,$9F,$FE,$2A,$8C,$47,$A1,$AF
	.dc.b	$F5,$46,$BF,$CA,$9F,$E1,$38,$7F,$1D,$6F,$16,$FE,$47,$FC,$7B,$1F
	.dc.b	$C1,$6B,$FB,$E7,$ED,$9F,$C0,$5F,$F0,$2B,$6F,$D8,$DB,$FE,$6F,$ED
	.dc.b	$BE,$B1,$F8,$DF,$F8,$27,$F1,$AF,$E0,$98,$FF,$7B,$6B,$FD,$E9,$AF
	.dc.b	$F4,$F6,$BF,$D0,$B7,$7E,$4F,$FD,$22,$3E,$B7,$85,$EA,$5C,$BF,$B1
	.dc.b	$DF,$FC,$31,$5B,$DB,$E9,$78,$FF,$E5,$5B,$FF,$0F,$B6,$33,$F2,$1F
	.dc.b	$E6,$1D,$9E,$8F,$F8,$DE,$7E,$D6,$BF,$E3,$0C,$3C,$8D,$7F,$D3,$4D
	.dc.b	$D4,$DF,$BB,$7C,$F7,$F2,$2F,$9A,$FD,$35,$FA,$1B,$6A,$DC,$FF,$8B
	.dc.b	$7C,$8F,$F0,$96,$3F,$8B,$6F,$27,$C5,$6C,$B4,$7E,$6B,$8C,$5F,$43
	.dc.b	$F0,$6B,$FB,$E3,$5F,$9C,$D7,$FD,$68,$D8,$FF,$63,$F7,$FF,$3A,$F3
	.dc.b	$BF,$01,$5F,$41,$B3,$A8,$FF,$22,$7E,$0B,$F4,$73,$FB,$4D,$8F,$F6
	.dc.b	$03,$72,$7E,$61,$F3,$EB,$C7,$E6,$FF,$8F,$33,$FF,$13,$61,$E8,$8C
	.dc.b	$1F,$E4,$0F,$84,$FF,$2B,$C7,$F0,$6F,$C6,$FF,$81,$3F,$BB,$6F,$07
	.dc.b	$EC,$E7,$73,$FE,$60,$F9,$0F,$F0,$04,$FE,$DD,$B8,$FE,$45,$B3,$FE
	.dc.b	$D7,$F9,$17,$E1,$3F,$87,$FA,$BF,$D3,$4F,$F9,$13,$64,$62,$3F,$4A
	.dc.b	$D7,$EB,$36,$1F,$25,$FC,$E7,$2F,$F1,$17,$7F,$F9,$EB,$E5,$3F,$B4
	.dc.b	$E7,$FE,$82,$9F,$F1,$35,$F0,$F7,$BF,$31,$FA,$BF,$F8,$EA,$0D,$27
	.dc.b	$CD,$79,$FE,$67,$FD,$F5,$3E,$66,$E9,$FE,$95,$D7,$D4,$D7,$F7,$BF
	.dc.b	$25,$FA,$47,$FC,$31,$C4,$FC,$67,$F8,$04,$61,$AF,$F9,$72,$78,$3F
	.dc.b	$E5,$EF,$D1,$7F,$87,$3F,$CE,$6B,$FC,$FB,$DD,$FE,$D1,$6C,$7F,$98
	.dc.b	$67,$7B,$7C,$0F,$C6,$9F,$8C,$7A,$9B,$5F,$CE,$8F,$D1,$38,$7F,$9E
	.dc.b	$EA,$8C,$47,$FB,$9F,$F0,$EF,$EA,$CF,$F4,$BB,$46,$1F,$F2,$A1,$8F
	.dc.b	$63,$F8,$35,$7F,$7B,$E9,$FE,$43,$93,$E9,$7F,$85,$37,$1F,$54,$7F
	.dc.b	$BA,$BF,$1A,$FE,$F1,$5F,$E8,$A7,$FB,$E4,$FF,$52,$6B,$9E,$3F,$F7
	.dc.b	$4C,$62,$3F,$7D,$F8,$A7,$F8,$85,$FF,$44,$B7,$3F,$C2,$B6,$FF,$CB
	.dc.b	$BF,$7F,$FD,$6D,$B5,$FC,$7E,$74,$F4,$F3,$7F,$D3,$B7,$F0,$2F,$84
	.dc.b	$FF,$0D,$FC,$9F,$F2,$58,$7F,$40,$6C,$67,$AF,$F9,$57,$CB,$7F,$8A
	.dc.b	$2B,$FA,$2B,$67,$47,$FA,$76,$B8,$F1,$7F,$B4,$3E,$23,$F9,$97,$E3
	.dc.b	$1F,$E7,$EF,$21,$F0,$5F,$C2,$5B,$F3,$3A,$F3,$26,$5F,$EA,$7B,$FF
	.dc.b	$42,$E2,$69,$FC,$0B,$4F,$EA,$EF,$F0,$69,$FD,$C3,$5C,$FC,$0F,$E9
	.dc.b	$A3,$11,$FE,$94,$FC,$6B,$F2,$3C,$3F,$E9,$6F,$A5,$FC,$FC,$7F,$B5
	.dc.b	$5B,$F4,$5F,$EC,$67,$3D,$BF,$E3,$4B,$46,$2F,$97,$B9,$FE,$CC,$C3
	.dc.b	$93,$70,$E4,$E7,$0E,$57,$C7,$0F,$D0,$5F,$91,$B3,$DD,$FF,$0E,$C6
	.dc.b	$5B,$E6,$BF,$1A,$FD,$44,$27,$FD,$B1,$18,$6C,$76,$9E,$0F,$01,$F5
	.dc.b	$B5,$A6,$5C,$11,$F8,$38,$FF,$8C,$3C,$2F,$F6,$06,$5F,$C2,$27,$FD
	.dc.b	$83,$5F,$11,$FF,$77,$CE,$AB,$F7,$D3,$99,$FF,$63,$CE,$5F,$F1,$DC
	.dc.b	$65,$46,$78,$3F,$09,$F7,$1F,$EB,$F9,$FC,$1B,$5F,$ED,$5B,$A3,$13
	.dc.b	$C3,$87,$F8,$73,$F8,$DF,$3B,$F8,$2C,$7E,$31,$B7,$CA,$AD,$6E,$BF
	.dc.b	$DF,$9F,$8D,$7F,$BD,$5A,$6B,$0F,$E9,$AE,$0F,$F6,$34,$BE,$83,$FD
	.dc.b	$75,$18,$7F,$D5,$8D,$F1,$FB,$C6,$31,$FF,$33,$57,$1F,$91,$F5,$CE
	.dc.b	$67,$FA,$C3,$5F,$F1,$DB,$5F,$F0,$67,$E5,$DF,$EB,$2C,$3F,$DB,$1C
	.dc.b	$CF,$A6,$7C,$17,$BA,$B5,$7F,$CA,$E7,$37,$C1,$B1,$FE,$6D,$6B,$FC
	.dc.b	$41,$AF,$FB,$D1,$A8,$79,$EF,$B1,$FB,$11,$86,$C6,$6B,$F3,$AD,$9F
	.dc.b	$E8,$DF,$8D,$3E,$0F,$EB,$C2,$31,$5F,$EF,$2A,$E1,$9A,$3C,$55,$BB
	.dc.b	$FB,$BF,$27,$FE,$3E,$9F,$AF,$F2,$4F,$C3,$33,$F0,$9F,$82,$8C,$4F
	.dc.b	$CA,$FA,$AF,$26,$3F,$CC,$1A,$FB,$F2,$46,$23,$FD,$7A,$DE,$37,$CD
	.dc.b	$3C,$8F,$C7,$B7,$0F,$1F,$DA,$62,$7F,$27,$F8,$CB,$EA,$3F,$3A,$DA
	.dc.b	$FC,$EB,$72,$FE,$D2,$D7,$FD,$4F,$FB,$DB,$EF,$FE,$8B,$6B,$F8,$5B
	.dc.b	$67,$FB,$A3,$5F,$F1,$37,$E3,$4F,$C3,$7F,$5A,$DA,$FD,$6B,$67,$FB
	.dc.b	$CB,$5F,$F1,$77,$82,$F5,$FF,$6A,$DA,$FD,$AB,$67,$FC,$02,$D7,$FA
	.dc.b	$EF,$BC,$E3,$FE,$01,$C7,$FE,$1B,$6C,$66,$FF,$C1,$93,$97,$FC,$1D
	.dc.b	$E8,$3E,$FD,$F3,$DF,$C2,$63,$0D,$8C,$CF,$06,$E3,$FE,$14,$F8,$5F
	.dc.b	$F2,$3C,$FF,$0C,$6C,$67,$E0,$BF,$88,$C6,$23,$FD,$AF,$FA,$AF,$F0
	.dc.b	$27,$FC,$40,$D7,$3E,$37,$E9,$23,$11,$FE,$B6,$F8,$C7,$CC,$7F,$1C
	.dc.b	$6E,$3F,$C7,$1B,$3F,$E2,$86,$BF,$E5,$DF,$C2,$3F,$67,$D8,$69,$3C
	.dc.b	$7F,$E4,$FF,$1B,$F9,$4C,$3F,$93,$B6,$FF,$C6,$6D,$7F,$CD,$0D,$7F
	.dc.b	$D9,$7D,$9F,$C2,$9F,$F3,$63,$5C,$FD,$2F,$F3,$48,$C4,$7F,$A8,$7F
	.dc.b	$33,$FD,$CB,$FE,$40,$6B,$FE,$75,$6B,$FE,$D8,$6B,$FC,$37,$DE,$7C
	.dc.b	$07,$F0,$C6,$E3,$FD,$99,$B3,$FE,$4F,$6B,$FE,$81,$FC,$5B,$E7,$BF
	.dc.b	$EB,$36,$EA,$69,$8F,$82,$F9,$B3,$FE,$EA,$6B,$FE,$BB,$6C,$9B,$F4
	.dc.b	$DF,$C0,$49,$E6,$FF,$56,$7F,$82,$7E,$28,$F8,$4F,$EB,$2D,$CF,$FA
	.dc.b	$CB,$71,$FF,$32,$B5,$FE,$3A,$3F,$33,$7D,$1F,$F2,$6F,$96,$FF,$72
	.dc.b	$DF,$F1,$53,$F9,$1C,$E9,$C5,$7F,$37,$9D,$AE,$FF,$E5,$4E,$7D,$6F
	.dc.b	$E7,$33,$FD,$3E,$0D,$27,$C5,$78,$4F,$14,$DB,$EF,$5C,$3F,$02,$F3
	.dc.b	$1E,$7F,$2A,$CE,$B0,$4F,$AF,$D1,$CD,$F9,$6B,$F9,$ED,$7F,$36,$8C
	.dc.b	$1E,$BF,$07,$E1,$1F,$0F,$7B,$5D,$63,$E3,$70,$E2,$B8,$9E,$C7,$FA
	.dc.b	$7F,$E2,$4E,$2B,$C7,$86,$E1,$EC,$9F,$C0,$3F,$E3,$66,$DF,$FA,$7D
	.dc.b	$AF,$D8,$6B,$FD,$59,$FE,$55,$C1,$F4,$21,$18,$9D,$5F,$93,$CB,$F2
	.dc.b	$FE,$03,$27,$C5,$F1,$37,$0F,$5E,$1F,$DC,$A7,$C9,$EB,$FF,$72,$8C
	.dc.b	$4F,$96,$31,$F4,$B2,$BD,$4D,$9F,$57,$67,$C4,$FB,$BF,$8A,$DA,$FB
	.dc.b	$BE,$E7,$F8,$6F,$6F,$57,$E2,$0F,$A3,$FE,$F3,$39,$AF,$EE,$B3,$97
	.dc.b	$F0,$AB,$EC,$3F,$C0,$ED,$8F,$F3,$0B,$72,$FF,$65,$35,$FD,$1A,$F8
	.dc.b	$6F,$F0,$3B,$67,$AD,$62,$73,$F8,$34,$9F,$0F,$B3,$70,$FC,$1E,$0A
	.dc.b	$E7,$F7,$B9,$D3,$DA,$F6,$FC,$42,$63,$EA,$FB,$7B,$8F,$2F,$88,$F5
	.dc.b	$92,$7D,$6F,$F5,$3F,$B2,$7E,$07,$FE,$1E,$F4,$3F,$A4,$67,$FB,$2F
	.dc.b	$61,$5A,$C3,$F5,$75,$A7,$FC,$3E,$FE,$EB,$1F,$E0,$AD,$73,$F5,$3F
	.dc.b	$E3,$28,$C4,$7F,$8D,$DF,$A8,$DF,$1B,$C5,$3E,$37,$C9,$E3,$4B,$F0
	.dc.b	$FD,$05,$BF,$D4,$E1,$FB,$89,$FF,$4B,$B6,$3F,$E0,$03,$5F,$2F,$8F
	.dc.b	$E3,$F1,$7E,$01,$F0,$7E,$0E,$DF,$BB,$C6,$F4,$11,$EA,$FF,$A0,$39
	.dc.b	$7F,$8E,$61,$FC,$93,$CD,$FB,$1E,$66,$49,$7F,$E0,$4F,$C3,$5E,$0F
	.dc.b	$FA,$0B,$8D,$18,$9F,$F3,$DE,$3D,$39,$3E,$94,$E7,$2F,$E1,$F6,$FC
	.dc.b	$9B,$FC,$07,$5F,$CA,$33,$7E,$1D,$FD,$AE,$3E,$D4,$EB,$2F,$F3,$35
	.dc.b	$6B,$6F,$F4,$8D,$FF,$6D,$C5,$39,$AF,$F4,$DC,$E6,$31,$3E,$FF,$D2
	.dc.b	$1E,$3F,$FA,$89,$B7,$FE,$A2,$6C,$EE,$7F,$C4,$5C,$0F,$B6,$F8,$4D
	.dc.b	$BF,$F4,$DD,$F0,$D3,$F2,$53,$9C,$3F,$D2,$FE,$7F,$E4,$DF,$A0,$D7
	.dc.b	$FC,$E3,$97,$F9,$B7,$AF,$FD,$6B,$39,$9F,$F4,$36,$EF,$C3,$8F,$D3
	.dc.b	$6E,$7F,$EA,$CA,$ED,$97,$E8,$E7,$37,$FD,$03,$EA,$7F,$AC,$7A,$1F
	.dc.b	$EB,$C7,$E4,$FC,$75,$BF,$F4,$2E,$3F,$E9,$59,$CB,$FD,$89,$B7,$FD
	.dc.b	$B1,$AF,$FA,$29,$AF,$1D,$3F,$C2,$70,$FC,$35,$F7,$9F,$ED,$58,$C3
	.dc.b	$63,$37,$FD,$E2,$D5,$A8,$C3,$CF,$B7,$D5,$46,$DE,$1F,$F5,$1F,$E1
	.dc.b	$AF,$A0,$BE,$4E,$1B,$8F,$F7,$09,$FE,$34,$F8,$72,$46,$1F,$23,$5F
	.dc.b	$F0,$6B,$5F,$F5,$5F,$67,$F0,$F7,$FD,$7C,$D7,$9F,$67,$FA,$DA,$31
	.dc.b	$2F,$1B,$CF,$E3,$FA,$1B,$87,$F8,$12,$BF,$C1,$84,$F1,$3F,$D8,$EE
	.dc.b	$35,$E8,$75,$F5,$ED,$71,$F6,$F6,$47,$DD,$48,$FF,$40,$7E,$1B,$FE
	.dc.b	$07,$F9,$AF,$F5,$96,$DF,$90,$8F,$E2,$D1,$8E,$A4,$6F,$DF,$FC,$BB
	.dc.b	$A6,$FC,$8F,$F8,$9B,$2F,$F6,$7D,$70,$8C,$61,$A9,$FF,$72,$1A,$B4
	.dc.b	$EA,$7F,$07,$F6,$5F,$EE,$1C,$7C,$BC,$E8,$C6,$5F,$A8,$E5,$FC,$26
	.dc.b	$5F,$EE,$FB,$DD,$6B,$0F,$AB,$3D,$5F,$E3,$8F,$89,$FB,$6E,$75,$FF
	.dc.b	$74,$39,$8F,$F7,$D7,$E1,$BF,$E8,$EB,$F8,$F0,$87,$C2,$FF,$91,$A7
	.dc.b	$31,$BC,$BF,$DE,$2E,$AF,$CD,$18,$F8,$97,$CE,$7E,$4E,$91,$8C,$7F
	.dc.b	$4D,$18,$9D,$4E,$67,$FC,$C3,$2F,$43,$F8,$4C,$70,$6B,$A6,$BE,$46
	.dc.b	$BE,$EE,$33,$E6,$3E,$0F,$2E,$16,$7E,$E8,$C3,$F0,$26,$E5,$FA,$6E
	.dc.b	$BE,$1A,$7F,$9D,$B6,$FD,$D6,$3F,$34,$FE,$A3,$3A,$78,$1F,$47,$85
	.dc.b	$F5,$4F,$7B,$AC,$FF,$08,$8C,$3E,$79,$DC,$EB,$9B,$F8,$47,$AB,$7F
	.dc.b	$0D,$FE,$0D,$7F,$76,$51,$88,$FF,$28,$46,$3B,$FC,$91,$FC,$D2,$75
	.dc.b	$A7,$E7,$97,$3F,$5F,$A5,$95,$F2,$70,$EB,$3E,$88,$7E,$19,$FD,$3D
	.dc.b	$FE,$24,$D7,$F9,$23,$5F,$F4,$EB,$5F,$A5,$F8,$6B,$C6,$FA,$9E,$37
	.dc.b	$F9,$9E,$31,$3F,$E8,$E6,$BF,$E7,$C7,$FD,$2E,$E7,$E1,$7F,$91,$28
	.dc.b	$FF,$3F,$3F,$C9,$5A,$FF,$37,$6B,$FE,$B0,$7F,$D7,$CF,$7E,$4F,$1E
	.dc.b	$BE,$66,$C7,$E3,$27,$6F,$FA,$3B,$17,$E1,$3A,$8F,$8A,$D8,$CC,$7F
	.dc.b	$AA,$16,$B9,$3D,$2E,$23,$E5,$C6,$0C,$E9,$CF,$79,$3C,$E3,$CB,$C4
	.dc.b	$FC,$37,$FD,$A5,$3F,$D8,$DA,$FF,$6D,$6E,$DF,$EC,$49,$E1,$E9,$F2
	.dc.b	$93,$FD,$36,$33,$5F,$D7,$27,$2D,$9F,$EA,$5C,$2F,$C3,$7F,$9B,$4F
	.dc.b	$F6,$58,$C7,$89,$A9,$CC,$FF,$B8,$27,$2E,$E1,$F1,$DF,$42,$73,$3F
	.dc.b	$E3,$69,$E3,$97,$F9,$AD,$DC,$7F,$B8,$A5,$F0,$AF,$AA,$E4,$D9,$D3
	.dc.b	$5F,$F6,$7F,$9C,$FD,$39,$F4,$1F,$80,$8C,$3F,$DB,$DB,$3F,$E0,$16
	.dc.b	$BF,$E5,$79,$E2,$FC,$1F,$A3,$FD,$6B,$E5,$6B,$FE,$50,$6E,$7F,$E4
	.dc.b	$FF,$DD,$BF,$C2,$33,$FE,$FC,$B4,$62,$3F,$CC,$8D,$7F,$C6,$AD,$7F
	.dc.b	$CA,$6D,$7F,$CA,$3F,$87,$BB,$BD,$4D,$8F,$53,$5F,$F3,$83,$5F,$EC
	.dc.b	$7C,$FF,$43,$AF,$FC,$BF,$3B,$C7,$FC,$CC,$D8,$FF,$21,$CE,$EB,$FC
	.dc.b	$87,$C7,$FA,$27,$FC,$DE,$DB,$FF,$52,$36,$BF,$CD,$7F,$99,$FF,$A9
	.dc.b	$AB,$E1,$35,$FF,$42,$35,$FF,$2A,$B5,$E3,$F0,$BF,$C1,$3F,$0C,$FE
	.dc.b	$49,$1F,$D0,$43,$87,$FA,$23,$5F,$E9,$2D,$77,$D3,$FB,$D8,$FA,$F7
	.dc.b	$C2,$31,$EA,$E7,$F0,$DF,$01,$2F,$FC,$6D,$68,$F2,$7A,$9C,$C9,$3F
	.dc.b	$D6,$34,$63,$E4,$F5,$7F,$86,$7E,$17,$FD,$DA,$B5,$3F,$D6,$E3,$10
	.dc.b	$AE,$2F,$F6,$28,$CC,$62,$11,$FD,$D9,$E6,$6F,$D0,$FE,$57,$19,$8F
	.dc.b	$F2,$F3,$5F,$94,$D7,$FD,$03,$FA,$FB,$EC,$BC,$FE,$8C,$9F,$DA,$4E
	.dc.b	$5E,$E7,$8B,$D8,$EE,$47,$FC,$B9,$7F,$99,$7C,$FE,$77,$C2,$23,$CA
	.dc.b	$1F,$15,$F1,$B9,$7E,$3F,$85,$2B,$51,$96,$90,$F9,$1B,$7C,$BF,$88
	.dc.b	$FE,$75,$B7,$F4,$E9,$C9,$FE,$0C,$71,$BF,$CB,$6D,$7F,$83,$5F,$F1
	.dc.b	$6C,$67,$CB,$FE,$AB,$1F,$E3,$0A,$FC,$93,$5E,$11,$FD,$6B,$F0,$C7
	.dc.b	$CA,$F8,$BC,$E9,$5F,$22,$72,$FC,$AB,$FD,$26,$1E,$3F,$B1,$F4,$9E
	.dc.b	$AE,$BF,$5F,$C3,$F8,$89,$7F,$E8,$9B,$FF,$67,$B7,$E8,$FC,$E4,$70
	.dc.b	$6F,$E9,$7E,$29,$1F,$17,$AF,$FD,$87,$F0,$9F,$F0,$16,$7F,$8F,$6F
	.dc.b	$07,$F8,$C5,$B5,$FE,$36,$7F,$C4,$AF,$E1,$1A,$FF,$AB,$5A,$FE,$E5
	.dc.b	$AF,$3F,$63,$FC,$03,$0F,$96,$F1,$44,$9E,$91,$E5,$F4,$FC,$DE,$A4
	.dc.b	$9F,$04,$65,$A4,$3F,$02,$2F,$F8,$0B,$E9,$F0,$7E,$07,$C7,$C8,$F5
	.dc.b	$47,$C3,$23,$83,$DD,$E1,$7F,$81,$FF,$08,$7D,$1D,$24,$FE,$9A,$3F
	.dc.b	$73,$AF,$F9,$65,$FD,$FC,$7F,$1B,$6B,$9F,$82,$FF,$11,$C6,$23,$FC
	.dc.b	$C7,$F9,$D7,$E9,$9E,$9C,$9E,$13,$FC,$31,$19,$8F,$EC,$7F,$2E,$FC
	.dc.b	$23,$FB,$03,$63,$FB,$03,$5F,$E5,$9E,$8A,$31,$1F,$E6,$F8,$7E,$80
	.dc.b	$E6,$FE,$D4,$D9,$FE,$D4,$D7,$9B,$5E,$5F,$86,$3E,$43,$FB,$93,$63
	.dc.b	$FB,$93,$5F,$1B,$5F,$17,$94,$F9,$EF,$EF,$4D,$8F,$EF,$4D,$7D,$6D
	.dc.b	$7F,$A4,$7D,$87,$F5,$96,$FC,$0F,$F9,$42,$31,$1F,$E8,$D6,$AC,$FF
	.dc.b	$11,$E9,$7F,$82,$7A,$3F,$C0,$33,$F1,$5A,$FF,$A6,$27,$6F,$F8,$37
	.dc.b	$F0,$7C,$BC,$DB,$06,$98,$FC,$19,$E0,$7F,$85,$DB,$5F,$E1,$76,$BF
	.dc.b	$61,$AF,$D7,$FA,$2F,$A0,$F8,$DA,$FA,$DB,$66,$FE,$0F,$F1,$E7,$D4
	.dc.b	$D7,$E1,$36,$4E,$61,$FA,$0B,$EC,$BF,$CF,$6D,$8F,$EF,$07,$FA,$F1
	.dc.b	$C3,$86,$BF,$8F,$F9,$0F,$EF,$57,$FE,$D4,$F0,$7F,$C7,$11,$83,$FB
	.dc.b	$F7,$C2,$3E,$21,$F6,$5F,$E5,$A6,$E1,$FE,$59,$6B,$FE,$57,$6B,$FE
	.dc.b	$98,$F9,$77,$E6,$1E,$68,$FE,$E5,$18,$7D,$3E,$52,$31,$1F,$E1,$38
	.dc.b	$C4,$FF,$93,$7F,$09,$FF,$1C,$68,$69,$3D,$AF,$F2,$97,$E1,$1F,$03
	.dc.b	$0F,$C4,$35,$FC,$83,$68,$E6,$1F,$3B,$FE,$61,$7F,$C4,$8D,$7F,$DC
	.dc.b	$CD,$7F,$DD,$6D,$97,$E1,$1F,$E0,$19,$FC,$F3,$63,$3E,$2F,$F9,$9E
	.dc.b	$31,$F8,$37,$F8,$2F,$4F,$F0,$AD,$23,$1E,$4F,$E8,$5B,$87,$EA,$7F
	.dc.b	$07,$FF,$04,$3F,$DC,$AD,$F8,$47,$F4,$28,$FF,$10,$36,$33,$E9,$7D
	.dc.b	$36,$CF,$F9,$DF,$D7,$7E,$03,$FE,$7D,$C1,$EC,$53,$FD,$01,$3A,$AF
	.dc.b	$F0,$BF,$E1,$1F,$90,$C3,$FC,$54,$DD,$13,$23,$0D,$F1,$7F,$D8,$91
	.dc.b	$FC,$12,$0D,$24,$79,$FE,$BF,$FD,$B7,$66,$92,$BF,$D2,$5F,$65,$F8
	.dc.b	$9C,$3F,$C9,$13,$67,$FC,$84,$DE,$05,$5F,$95,$FF,$08,$57,$F1,$56
	.dc.b	$C9,$A4,$F7,$7F,$D4,$4F,$F8,$73,$0F,$E2,$CD,$AE,$53,$FE,$A7,$FC
	.dc.b	$09,$E0,$52,$4F,$FA,$6E,$3F,$DB,$CF,$FA,$C1,$B9,$FE,$11,$B8,$7F
	.dc.b	$23,$6C,$FF,$2B,$7F,$B8,$E2,$7F,$91,$93,$C9,$FF,$5B,$3F,$CA,$25
	.dc.b	$F8,$27,$F9,$57,$0F,$AF,$F8,$27,$F5,$79,$FB,$33,$FE,$C3,$8C,$47
	.dc.b	$0F,$97,$FB,$AA,$33,$3C,$E3,$1E,$A6,$A3,$13,$AA,$E2,$DA,$F0,$D6
	.dc.b	$FE,$2F,$FC,$F1,$1C,$8F,$1B,$87,$FD,$31,$7D,$DB,$19,$79,$61,$FC
	.dc.b	$46,$7F,$37,$F8,$2D,$E6,$34,$95,$FC,$5C,$7D,$05,$69,$97,$04,$7C
	.dc.b	$C7,$E8,$3C,$8F,$F4,$7E,$7F,$22,$3F,$47,$3E,$E7,$E4,$CE,$B3,$FE
	.dc.b	$77,$39,$9E,$5C,$15,$E2,$76,$15,$C5,$FE,$D9,$F7,$9F,$CA,$E7,$FD
	.dc.b	$E4,$D8,$FF,$09,$B5,$FF,$09,$35,$FC,$67,$D9,$7F,$A9,$B5,$FC,$83
	.dc.b	$6B,$FB,$95,$73,$9C,$C7,$E6,$7C,$9D,$5F,$E4,$DB,$C7,$F9,$3C,$2F
	.dc.b	$17,$F4,$CB,$FF,$7D,$7E,$0A,$F5,$76,$BC,$4B,$18,$64,$F0,$FA,$4B
	.dc.b	$9F,$FA,$4B,$26,$09,$F0,$FD,$B3,$E3,$6F,$C7,$61,$B8,$FF,$AF,$B0
	.dc.b	$52,$BD,$7B,$BE,$6D,$F0,$7C,$34,$FE,$61,$18,$C7,$FD,$73,$F8,$2B
	.dc.b	$D3,$FF,$66,$57,$E9,$A3,$13,$FD,$8D,$B1,$FD,$E9,$AF,$FB,$CF,$DD
	.dc.b	$F9,$D3,$FA,$E6,$C7,$F2,$C9,$E3,$8F,$FB,$66,$73,$A3,$F0,$5F,$EE
	.dc.b	$38,$7B,$57,$CF,$FD,$47,$38,$FC,$97,$33,$F0,$FD,$C7,$D7,$FA,$AB
	.dc.b	$FC,$71,$18,$9F,$F0,$8E,$1F,$E2,$CA,$D3,$63,$32,$FA,$BF,$C1,$B6
	.dc.b	$3F,$DD,$B9,$BF,$05,$F4,$CE,$71,$FB,$9F,$05,$FC,$A2,$FE,$B3,$63
	.dc.b	$FD,$A3,$DB,$FE,$16,$EB,$FC,$7C,$E7,$1F,$ED,$7E,$97,$BA,$FF,$CE
	.dc.b	$0D,$7F,$C9,$8D,$7F,$CA,$8D,$D1,$F8,$23,$EE,$FF,$36,$FF,$B4,$B3
	.dc.b	$FC,$43,$96,$FC,$AF,$E8,$3C,$0F,$F4,$04,$7E,$35,$B8,$FF,$BA,$3F
	.dc.b	$02,$FE,$1B,$E5,$6F,$C5,$FD,$3E,$9F,$97,$C3,$E6,$B6,$FF,$D1,$6D
	.dc.b	$7D,$7D,$DF,$57,$A9,$F8,$1F,$FA,$0E,$FF,$B8,$F0,$1F,$EE,$73,$AF
	.dc.b	$95,$FD,$BB,$C3,$E2,$AD,$A7,$3F,$66,$FA,$5A,$7E,$FB,$7F,$F5,$AF
	.dc.b	$D0,$63,$2E,$E7,$6F,$C9,$7F,$49,$B7,$F4,$63,$F4,$8E,$1F,$E9,$3A
	.dc.b	$23,$11,$C7,$C3,$F6,$7E,$4F,$F9,$6D,$FF,$C4,$BF,$82,$FF,$B1,$31
	.dc.b	$EF,$7F,$AF,$AD,$47,$FA,$1E,$75,$2A,$F8,$4F,$ED,$7F,$27,$7C,$FF
	.dc.b	$C2,$6D,$F4,$D5,$F7,$EF,$67,$57,$F2,$35,$7D,$27,$D1,$6E,$1F,$E7
	.dc.b	$4D,$FD,$93,$93,$EB,$7C,$CF,$39,$FE,$21,$7F,$C3,$DF,$5D,$27,$5A
	.dc.b	$7F,$1D,$AD,$3E,$BA,$FB,$7F,$93,$BE,$17,$ED,$B6,$3D,$5F,$93,$BF
	.dc.b	$30,$CD,$7E,$03,$F2,$8C,$57,$9D,$B8,$7D,$7C,$D1,$88,$FA,$ED,$7F
	.dc.b	$17,$F9,$43,$E8,$3F,$9F,$FE,$13,$FC,$E7,$0F,$5C,$E5,$FE,$5D,$19
	.dc.b	$AF,$51,$FC,$9F,$C1,$FC,$3F,$37,$D6,$CB,$FD,$D1,$D3,$F8,$FF,$C4
	.dc.b	$B3,$8F,$E6,$AE,$E1,$FC,$BF,$F0,$8F,$BE,$FE,$31,$AF,$E8,$1A,$EF
	.dc.b	$E8,$FF,$51,$F9,$43,$E9,$7F,$AB,$FE,$16,$FA,$CF,$D2,$CE,$72,$FE
	.dc.b	$8D,$39,$7E,$BF,$ED,$FF,$AD,$7F,$27,$8F,$F9,$33,$EA,$7F,$5B,$39
	.dc.b	$8F,$79,$99,$DF,$E1,$9F,$D8,$5F,$F5,$33,$63,$FD,$7A,$D7,$F5,$4D
	.dc.b	$7F,$C6,$2F,$F5,$F8,$E7,$F0,$1F,$DC,$A5,$3C,$F8,$BF,$43,$18,$9D
	.dc.b	$4B,$EB,$E9,$FB,$DF,$CE,$DE,$F6,$31,$F3,$FE,$19,$C0,$C6,$3E,$CC
	.dc.b	$11,$D3,$79,$5F,$13,$E4,$9E,$56,$30,$E1,$5E,$2F,$9B,$C6,$DE,$6C
	.dc.b	$BF,$D6,$78,$FE,$93,$EC,$7F,$AF,$BF,$E5,$0F,$A6,$FD,$2D,$7E,$1E
	.dc.b	$73,$8F,$FA,$42,$73,$5B,$F2,$3E,$24,$7F,$37,$9F,$A6,$E9,$E1,$EC
	.dc.b	$E8,$DF,$B1,$6F,$EE,$D7,$F4,$1A,$E3,$F6,$3D,$3F,$01,$EA,$DA,$DC
	.dc.b	$59,$DC,$60,$FD,$FC,$78,$F3,$FD,$03,$63,$FD,$53,$E7,$7F,$86,$FC
	.dc.b	$6F,$E6,$F3,$9B,$F7,$7D,$77,$B3,$E6,$7D,$AD,$C3,$F2,$7F,$16,$FC
	.dc.b	$0B,$F0,$8F,$E1,$EF,$F2,$6E,$67,$7D,$7F,$89,$FC,$35,$F6,$9F,$E0
	.dc.b	$1B,$46,$2B,$FD,$F8,$D8,$FE,$7C,$6A,$78,$43,$C4,$E8,$75,$8B,$5F
	.dc.b	$A8,$97,$CD,$78,$4F,$EA,$BF,$8E,$7F,$51,$9D,$46,$5E,$1F,$73,$71
	.dc.b	$FD,$6D,$A3,$D5,$E7,$7A,$FB,$6D,$F7,$70,$D3,$C5,$18,$BF,$E2,$BE
	.dc.b	$63,$F0,$AF,$E5,$91,$8E,$CF,$ED,$CE,$A3,$3C,$FF,$93,$F8,$0F,$D9
	.dc.b	$7B,$3F,$C7,$EB,$FB,$13,$5F,$F5,$4B,$5F,$F7,$3B,$5F,$EA,$3F,$2B
	.dc.b	$C8,$E7,$F2,$B7,$4F,$E1,$F5,$FD,$42,$BF,$86,$BF,$BB,$8C,$FB,$5E
	.dc.b	$2F,$26,$47,$07,$CB,$18,$AE,$3F,$07,$E8,$F1,$BD,$2D,$BE,$F1,$67
	.dc.b	$71,$83,$FA,$CF,$E1,$CF,$8D,$FF,$0A,$36,$3F,$B2,$36,$7F,$A7,$35
	.dc.b	$FF,$65,$FC,$57,$D9,$F3,$FF,$D1,$3C,$C7,$FC,$2F,$5C,$78,$7F,$BD
	.dc.b	$D6,$F0,$FE,$D3,$F8,$77,$F9,$9B,$94,$FF,$0A,$CB,$ED,$B9,$31,$7F
	.dc.b	$D3,$DB,$49,$5B,$F0,$EF,$F1,$95,$EF,$CC,$FF,$4E,$C6,$2F,$FC,$5B
	.dc.b	$1F,$EB,$47,$FC,$04,$DB,$FE,$D8,$D7,$FC,$42,$D7,$C1,$7F,$8F,$F1
	.dc.b	$FF,$D5,$3C,$E7,$FC,$6F,$5C,$F1,$F8,$33,$9C,$3F,$C1,$3F,$95,$FF
	.dc.b	$AE,$F1,$FF,$49,$37,$43,$4D,$65,$FE,$7E,$E8,$FF,$5B,$FC,$9F,$E7
	.dc.b	$32,$FE,$E5,$5F,$E8,$C8,$C6,$3F,$17,$EC,$3E,$2E,$3F,$EC,$07,$FC
	.dc.b	$5D,$F6,$17,$2F,$E6,$71,$8A,$FF,$1B,$B7,$3F,$F0,$0B,$F3,$BF,$08
	.dc.b	$E5,$9F,$FA,$7A,$31,$5C,$3E,$2F,$FC,$ED,$06,$93,$EC,$7F,$D2,$D1
	.dc.b	$9F,$B0,$F2,$49,$B3,$87,$D1,$F1,$7F,$E7,$4C,$3F,$B2,$FE,$1B,$FE
	.dc.b	$89,$C7,$DD,$39,$7F,$09,$9A,$74,$FE,$1D,$FE,$33,$1F,$C8,$A3,$5E
	.dc.b	$37,$F2,$99,$CC,$7F,$26,$6B,$FE,$65,$7F,$8C,$4F,$F8,$29,$B1,$FE
	.dc.b	$08,$78,$DF,$9B,$AB,$FB,$F4,$7F,$85,$67,$5E,$26,$3D,$EC,$BA,$F1
	.dc.b	$BC,$44,$C7,$CD,$7F,$DD,$1F,$47,$9D,$FD,$0F,$0F,$F0,$BB,$72,$FF
	.dc.b	$04,$BF,$D2,$5B,$E7,$7F,$83,$E3,$E7,$57,$F8,$89,$7D,$4E,$77,$D3
	.dc.b	$FE,$90,$FC,$37,$FD,$95,$8B,$E9,$3F,$CE,$6F,$FB,$A6,$33,$18,$F5
	.dc.b	$7F,$DF,$9F,$87,$FF,$9E,$A3,$7D,$4A,$DF,$C6,$FF,$A7,$9F,$F6,$54
	.dc.b	$E9,$B1,$9F,$7B,$93,$BF,$9D,$B9,$7F,$BE,$79,$BF,$96,$61,$E4,$EA
	.dc.b	$7E,$1C,$EE,$DE,$9F,$D6,$3C,$CE,$35,$A9,$E7,$8B,$8B,$FC,$85,$8F
	.dc.b	$BA,$31,$C7,$AF,$0F,$FA,$5C,$E6,$B7,$B7,$E4,$3F,$0E,$FF,$4D,$CF
	.dc.b	$17,$FD,$5A,$DF,$87,$EF,$18,$9E,$1E,$8B,$E7,$BF,$29,$28,$C6,$7F
	.dc.b	$DE,$23,$13,$A9,$CC,$FF,$AE,$65,$E9,$B8,$39,$B6,$79,$35,$E0,$D7
	.dc.b	$D5,$D5,$FA,$AC,$FD,$5E,$CD,$CF,$DC,$DA,$FE,$65,$8F,$F8,$65,$FE
	.dc.b	$69,$5F,$17,$8B,$51,$F3,$5B,$B7,$FA,$D2,$37,$87,$EE,$37,$FF,$6E
	.dc.b	$46,$0F,$55,$7F,$B7,$5C,$71,$7F,$AA,$FD,$CF,$13,$63,$86,$DF,$EC
	.dc.b	$0F,$96,$F3,$2F,$C3,$9F,$5D,$86,$D7,$06,$BF,$9F,$CF,$FD,$E7,$0F
	.dc.b	$C0,$B8,$74,$CA,$E0,$FC,$3E,$E7,$17,$D4,$E1,$FC,$57,$E1,$CF,$2F
	.dc.b	$F8,$AF,$87,$FE,$7E,$FC,$21,$E1,$63,$B7,$F8,$CC,$63,$4F,$F4,$BE
	.dc.b	$1C,$5B,$F1,$1B,$7F,$B1,$47,$FA,$29,$C9,$C3,$E2,$34,$A3,$F3,$8F
	.dc.b	$FB,$FD,$AF,$F2,$86,$BD,$38,$7F,$93,$3C,$7E,$7F,$9C,$BE,$26,$98
	.dc.b	$8E,$F8,$FC,$7A,$FF,$57,$B5,$FF,$56,$B7,$4F,$85,$C9,$C0,$EF,$1F
	.dc.b	$4A,$7F,$9D,$87,$0F,$C6,$8F,$E7,$5C,$EF,$DA,$7F,$D3,$33,$E0,$E4
	.dc.b	$F1,$9F,$1E,$75,$E5,$DF,$C3,$DE,$FD,$25,$FF,$62,$87,$B1,$B3,$BB
	.dc.b	$88,$FF,$35,$DA,$31,$1E,$88,$C4,$FF,$4F,$EB,$7E,$1E,$FA,$CF,$B5
	.dc.b	$18,$F3,$AC,$FF,$34,$C3,$FC,$3D,$C1,$FE,$EF,$8E,$11,$8E,$7F,$F1
	.dc.b	$15,$F4,$D3,$8E,$32,$B4,$7F,$03,$BF,$7B,$8F,$5D,$F8,$73,$EA,$FF
	.dc.b	$19,$18,$6C,$66,$79,$E2,$FC,$3D,$F6,$DE,$A6,$CF,$FA,$3A,$75,$3F
	.dc.b	$06,$73,$3F,$72,$72,$FF,$A5,$30,$7E,$1F,$FE,$03,$8F,$33,$56,$DF
	.dc.b	$E4,$C8,$C2,$DB,$E1,$FD,$87,$FA,$36,$BC,$ED,$7C,$0D,$7F,$C0,$F8
	.dc.b	$23,$11,$CB,$DA,$FB,$1D,$3F,$4E,$3C,$37,$FE,$42,$FC,$3F,$F8,$A4
	.dc.b	$60,$FF,$29,$B8,$C7,$C9,$1F,$CC,$EB,$94,$63,$1F,$CE,$70,$FE,$0A
	.dc.b	$73,$87,$DC,$6D,$F1,$9F,$AB,$F8,$7B,$E3,$29,$27,$EA,$F2,$7E,$B7
	.dc.b	$0F,$C2,$BF,$B1,$8F,$7B,$5C,$F9,$1F,$95,$8C,$47,$F9,$B7,$F0,$FF
	.dc.b	$F5,$FB,$F8,$A0,$E1,$F8,$06,$3B,$3F,$87,$C7,$CF,$D7,$E0,$CF,$0D
	.dc.b	$F3,$52,$3A,$56,$F1,$FF,$1C,$AF,$77,$99,$27,$7D,$BD,$8C,$3E,$29
	.dc.b	$E0,$E8,$F4,$7E,$1D,$FD,$AD,$BE,$AF,$F2,$48,$C4,$7A,$9F,$ED,$2F
	.dc.b	$0F,$57,$F9,$0C,$62,$3F,$DD,$2D,$7F,$8A,$B5,$FD,$9F,$9E,$F6,$FF
	.dc.b	$7B,$F8,$7F,$E8,$37,$FC,$03,$75,$DC,$F8,$FD,$39,$87,$A1,$F3,$7C
	.dc.b	$74,$8C,$10,$F4,$3E,$BF,$8E,$F6,$71,$C9,$F8,$31,$96,$90,$FF,$74
	.dc.b	$FE,$6B,$FE,$B9,$7E,$F6,$FF,$E2,$1D,$3E,$CB,$73,$FF,$36,$B6,$7F
	.dc.b	$D6,$4B,$CE,$FE,$9E,$D8,$FF,$37,$35,$FE,$76,$7F,$87,$63,$FA,$6C
	.dc.b	$66,$3F,$9F,$FE,$00,$F5,$FF,$AC,$AF,$9A,$A6,$23,$BF,$C6,$78,$3A
	.dc.b	$E5,$FA,$A3,$F9,$6F,$E8,$DF,$E3,$D8,$FE,$32,$D7,$F7,$1E,$02,$31
	.dc.b	$AF,$EE,$3D,$9F,$EF,$75,$EC,$F9,$28,$FC,$67,$DF,$EE,$C8,$C1,$0F
	.dc.b	$D2,$ED,$D3,$83,$FD,$2B,$F8,$6B,$E6,$FF,$BA,$BF,$DF,$EB,$F5,$6D
	.dc.b	$CB,$F5,$2E,$E3,$F6,$4B,$E6,$38,$D4,$9C,$C7,$F4,$CF,$DD,$39,$61
	.dc.b	$E0,$8E,$E6,$7D,$CF,$0F,$87,$E2,$F5,$E6,$1D,$A3,$2D,$21,$FC,$03
	.dc.b	$A5,$F8,$2F,$AB,$E3,$64,$F9,$9F,$6F,$8D,$23,$D2,$D7,$F9,$47,$1B
	.dc.b	$F1,$CF,$EA,$B1,$FC,$33,$EB,$DF,$53,$FE,$1A,$6C,$7F,$13,$6B,$BE
	.dc.b	$4D,$1F,$E1,$88,$FD,$EC,$7F,$82,$9A,$FF,$B1,$9A,$FF,$99,$FD,$DF
	.dc.b	$DA,$21,$F8,$9B,$C3,$FF,$15,$B6,$3F,$B1,$CF,$0E,$B7,$E3,$DF,$E2
	.dc.b	$59,$FE,$DD,$F5,$4F,$C5,$3F,$B6,$B6,$3F,$B5,$CE,$DF,$C9,$EB,$FE
	.dc.b	$65,$7F,$B3,$35,$FE,$AE,$D7,$FC,$02,$D7,$FD,$01,$F8,$F7,$EA,$5F
	.dc.b	$EE,$ED,$7F,$3E,$DE,$CF,$CF,$74,$3E,$3F,$FC,$27,$F0,$9F,$E4,$BB
	.dc.b	$FE,$B9,$ED,$7F,$A3,$3E,$A3,$FC,$0B,$E8,$9D,$FC,$47,$F8,$6B,$4F
	.dc.b	$F1,$0D,$F9,$BE,$0F,$FC,$41,$18,$8F,$3C,$7F,$88,$FE,$B5,$F3,$3F
	.dc.b	$E2,$26,$CF,$F8,$39,$AF,$E1,$BE,$53,$FD,$3F,$BF,$F6,$9E,$9F,$E9
	.dc.b	$12,$D5,$F8,$9F,$F5,$5A,$FE,$E3,$18,$F0,$7F,$C5,$F7,$C3,$9D,$F8
	.dc.b	$C7,$BF,$1F,$F0,$1B,$5F,$F1,$6B,$5F,$F5,$63,$63,$F2,$B5,$FE,$3A
	.dc.b	$ED,$F0,$78,$BF,$E3,$46,$DB,$BD,$C5,$FC,$A7,$07,$CD,$FD,$3C,$3F
	.dc.b	$21,$A3,$F1,$3F,$9D,$8F,$F8,$DD,$AF,$F1,$8A,$E0,$F9,$3C,$FF,$F2
	.dc.b	$76,$BF,$E0,$58,$FF,$19,$35,$FF,$18,$35,$FF,$10,$FC,$BF,$F9,$31
	.dc.b	$FE,$46,$DE,$0F,$F0,$B3,$72,$FD,$26,$AF,$C5,$1F,$86,$7F,$96,$BF
	.dc.b	$1E,$FC,$56,$3F,$E4,$6F,$8F,$7E,$01,$FE,$5E,$6C,$7F,$92,$A7,$6F
	.dc.b	$E3,$BA,$BF,$44,$FE,$C4,$FB,$8E,$1F,$8A,$D7,$3D,$DF,$D3,$1F,$C1
	.dc.b	$7E,$27,$FB,$57,$F3,$2D,$DB,$F0,$6D,$D7,$D7,$87,$F1,$09,$FC,$3F
	.dc.b	$E2,$7F,$E2,$9B,$FC,$62,$FC,$57,$FC,$CA,$D8,$FF,$35,$3B,$E5,$FF
	.dc.b	$43,$F9,$DE,$36,$CF,$F9,$77,$4F,$E4,$ED,$9D,$7E,$29,$FD,$3A,$3F
	.dc.b	$5E,$DC,$7F,$32,$DC,$FF,$8A,$71,$BF,$1A,$7D,$8F,$F9,$E3,$F1,$4F
	.dc.b	$E3,$3A,$1A,$7F,$1C,$C5,$F8,$37,$F4,$BC,$FF,$7B,$2F,$C4,$BD,$0D
	.dc.b	$8F,$F3,$A3,$5F,$F5,$9D,$F5,$9F,$77,$C0,$BE,$B7,$FD,$12,$FB,$AB
	.dc.b	$F8,$4B,$71,$FD,$9B,$B8,$FF,$26,$7E,$29,$F8,$A6,$F4,$FF,$B6,$23
	.dc.b	$11,$FE,$4E,$7E,$A2,$7F,$CA,$ED,$F8,$FF,$F6,$BC,$60,$FF,$2D,$E4
	.dc.b	$FB,$9F,$F6,$5E,$3F,$E7,$41,$F7,$18,$E2,$FF,$57,$B7,$1F,$E1,$95
	.dc.b	$6F,$97,$17,$FA,$9A,$30,$47,$FD,$53,$F3,$BF,$E1,$FC,$7F,$88,$BF
	.dc.b	$E2,$7A,$34,$D6,$3F,$C0,$1F,$E8,$53,$C3,$E2,$39,$9D,$3B,$3F,$D6
	.dc.b	$BF,$88,$7E,$3C,$9B,$63,$C9,$9A,$FF,$AF,$67,$F8,$8E,$66,$FE,$23
	.dc.b	$FE,$D6,$CB,$F9,$2B,$FC,$D6,$BF,$95,$36,$8D,$35,$0F,$C5,$3F,$D8
	.dc.b	$35,$FD,$9F,$F1,$37,$B3,$FB,$99,$7B,$7F,$BB,$9C,$DF,$F6,$D9,$CB
	.dc.b	$C6,$B8,$7C,$F7,$FA,$FA,$15,$C1,$F0,$46,$63,$15,$CA,$77,$5E,$1B
	.dc.b	$F0,$3D,$30,$ED,$3E,$1C,$3B,$3F,$C3,$CE,$B1,$FC,$2E,$FD,$B1,$FE
	.dc.b	$58,$6B,$B7,$F5,$1F,$8A,$6F,$A8,$D2,$4F,$F1,$51,$F5,$15,$A6,$5C
	.dc.b	$11,$F5,$9F,$C8,$F9,$7F,$D9,$7A,$7C,$38,$FF,$0B,$4F,$B1,$FE,$0B
	.dc.b	$CB,$5C,$9C,$8C,$94,$CC,$F3,$AE,$BA,$7F,$88,$7C,$FF,$E0,$9E,$6F
	.dc.b	$FA,$EA,$F1,$9F,$FB,$47,$1F,$F7,$F3,$63,$FD,$F8,$D7,$FD,$DA,$FF
	.dc.b	$0D,$8F,$F1,$66,$BF,$EE,$18,$C4,$FF,$9C,$9E,$51,$97,$F6,$D3,$E3
	.dc.b	$FD,$67,$F5,$2F,$23,$1D,$1A,$B6,$3F,$C6,$EF,$0A,$F8,$CF,$FB,$DF
	.dc.b	$3F,$F6,$77,$C3,$F1,$6E,$3D,$65,$1F,$ED,$3F,$89,$D7,$C2,$AF,$87
	.dc.b	$C5,$F4,$D7,$FD,$FF,$68,$C6,$3F,$CE,$2F,$83,$D6,$75,$AF,$87,$E1
	.dc.b	$3F,$C4,$73,$E1,$F3,$DF,$04,$F8,$9D,$7B,$DE,$F9,$CC,$59,$D5,$75
	.dc.b	$F3,$1F,$25,$9A,$F7,$46,$27,$C5,$E2,$BE,$AB,$EF,$71,$FC,$08,$C4
	.dc.b	$FF,$AE,$36,$FE,$4F,$5A,$6C,$67,$F1,$77,$C5,$FC,$B6,$C7,$6A,$C4
	.dc.b	$FC,$FF,$C4,$DF,$49,$F4,$71,$5E,$63,$0F,$D5,$FC,$4B,$FC,$16,$F8
	.dc.b	$DB,$9F,$EE,$73,$46,$23,$F5,$0D,$7F,$B4,$FD,$CF,$F6,$8A,$FA,$4D
	.dc.b	$8F,$C2,$68,$FC,$2B,$F0,$93,$9E,$8F,$BB,$EE,$78,$1F,$BF,$F9,$AF
	.dc.b	$A3,$1F,$B9,$39,$7F,$BB,$C6,$71,$FF,$4F,$8F,$B8,$F9,$FF,$01,$FE
	.dc.b	$8F,$6C,$66,$BF,$DC,$F3,$96,$F3,$FE,$2A,$7F,$D5,$2D,$C7,$FD,$B0
	.dc.b	$FF,$B3,$67,$F3,$54,$8D,$47,$F8,$EB,$C5,$79,$1F,$A9,$AF,$E2,$AD
	.dc.b	$8F,$DA,$DF,$FB,$8D,$DD,$74,$C1,$F0,$DE,$82,$47,$A2,$32,$FF,$C0
	.dc.b	$0C,$5F,$B5,$7C,$D8,$F2,$FD,$93,$EC,$BF,$39,$39,$BF,$F6,$2C,$E6
	.dc.b	$BD,$7F,$20,$FC,$57,$D0,$D8,$CF,$0E,$27,$D9,$F1,$2F,$A3,$F9,$ED
	.dc.b	$AF,$9A,$DC,$7F,$A1,$70,$FF,$9C,$6B,$F4,$7A,$BE,$81,$F4,$3F,$1F
	.dc.b	$F1,$2F,$D2,$E0,$E3,$FC,$CB,$F3,$EB,$79,$FD,$2A,$DD,$6B,$07,$E8
	.dc.b	$3F,$EE,$AA,$FA,$FF,$82,$3D,$CF,$DF,$4E,$72,$FF,$73,$CE,$6B,$EE
	.dc.b	$ED,$FE,$55,$7D,$7F,$15,$EE,$23,$F8,$AF,$B9,$7B,$FF,$7D,$B1,$F8
	.dc.b	$67,$A6,$1F,$EF,$47,$2F,$FB,$DB,$F1,$0F,$F6,$D4,$FE,$31,$B1,$9F
	.dc.b	$C1,$7F,$15,$18,$DB,$E1,$F1,$FE,$1B,$E6,$FF,$3F,$D6,$69,$31,$FC
	.dc.b	$37,$23,$EA,$3F,$29,$F8,$F7,$D5,$C7,$E4,$4E,$6B,$F3,$B3,$97,$FD
	.dc.b	$F1,$C8,$FC,$C1,$F8,$07,$E7,$7E,$73,$F2,$7C,$1F,$5E,$F7,$5A,$7F
	.dc.b	$4F,$3A,$CB,$D3,$F5,$BF,$E4,$69,$FD,$9B,$6F,$F0,$ED,$7F,$D3,$1B
	.dc.b	$A3,$11,$FE,$A8,$FA,$0F,$89,$57,$E8,$BF,$61,$0E,$3B,$1F,$D6,$EB
	.dc.b	$FC,$B1,$E4,$DE,$CF,$E5,$71,$E9,$06,$B0,$FE,$B0,$A3,$79,$FE,$52
	.dc.b	$31,$EC,$FF,$3E,$6D,$7F,$AA,$69,$39,$6C,$7F,$AB,$7F,$30,$FD,$A5
	.dc.b	$F3,$6C,$7F,$AD,$E7,$77,$E5,$13,$A7,$FD,$C5,$F8,$6F,$E6,$EF,$F0
	.dc.b	$AD,$BF,$F6,$6B,$FE,$8B,$7F,$8C,$B9,$9F,$F7,$57,$E6,$4F,$C0,$BF
	.dc.b	$8C,$36,$B7,$F3,$18,$9F,$E3,$9F,$93,$FD,$DB,$F9,$1C,$48,$C6,$5F
	.dc.b	$96,$7F,$BB,$CF,$FB,$DD,$CE,$1F,$EE,$06,$C7,$FB,$EF,$F3,$2F,$EB
	.dc.b	$33,$FC,$BA,$73,$69,$CD,$7C,$3F,$A2,$7D,$5F,$DD,$5F,$3B,$63,$0C
	.dc.b	$C7,$ED,$CB,$F0,$5C,$5F,$37,$91,$FC,$E7,$0F,$CA,$67,$F7,$EB,$85
	.dc.b	$FF,$42,$9C,$BE,$BF,$AB,$F8,$BD,$FF,$19,$B1,$EB,$8B,$8F,$C8,$D5
	.dc.b	$58,$C7,$E4,$7D,$FF,$CE,$3E,$7F,$AB,$71,$F8,$9F,$09,$E5,$F8,$7F
	.dc.b	$A6,$D8,$FC,$BF,$E1,$EE,$DF,$D0,$7E,$1D,$F9,$C9,$CC,$63,$5C,$FB
	.dc.b	$DF,$99,$AF,$F1,$AD,$7F,$C0,$16,$CF,$F9,$EC,$38,$7F,$CF,$DD,$4F
	.dc.b	$B9,$FD,$73,$6B,$F5,$35,$CB,$6F,$CF,$57,$E2,$B1,$95,$F7,$FE,$12
	.dc.b	$E3,$F1,$67,$E9,$7D,$9E,$30,$FE,$43,$C8,$8F,$F2,$9F,$E1,$FF,$BF
	.dc.b	$C7,$FC,$9D,$F2,$08,$FD,$BC,$7F,$9C,$5D,$63,$FB,$F9,$D7,$C3,$F2
	.dc.b	$25,$72,$AD,$E3,$FD,$4A,$B7,$3F,$C1,$3F,$50,$FC,$1E,$1F,$B5,$F5
	.dc.b	$5E,$E7,$F0,$F9,$CE,$1F,$C1,$E7,$37,$FC,$31,$7D,$6F,$F0,$96,$C7
	.dc.b	$CF,$C1,$18,$7E,$93,$5F,$A3,$F2,$BF,$40,$FE,$3D,$1B,$BF,$E1,$91
	.dc.b	$FE,$78,$FD,$67,$FC,$55,$5F,$E7,$CF,$27,$FC,$55,$5E,$D9,$CE,$9F
	.dc.b	$E0,$BC,$7E,$5D,$69,$FF,$4D,$36,$FF,$D0,$B5,$C2,$31,$E1,$FF,$03
	.dc.b	$D7,$F8,$3E,$9F,$C2,$1B,$87,$EF,$62,$E1,$FA,$39,$B9,$7F,$5E,$FC
	.dc.b	$5F,$FC,$9D,$2C,$11,$86,$FC,$07,$F9,$02,$31,$C6,$FC,$73,$F7,$47
	.dc.b	$F8,$CE,$3E,$0D,$7F,$0B,$8C,$79,$93,$4F,$F1,$ED,$7E,$67,$1F,$E3
	.dc.b	$AF,$F3,$2C,$BF,$23,$F4,$DF,$C1,$4E,$13,$FC,$8B,$2F,$E5,$CF,$0F
	.dc.b	$3F,$C8,$4C,$3F,$CA,$38,$7E,$4A,$7F,$AE,$D7,$F3,$3C,$FF,$8F,$79
	.dc.b	$7E,$7C,$3F,$89,$3E,$9E,$0F,$71,$31,$E6,$FF,$98,$2D,$F6,$9F,$C6
	.dc.b	$4E,$13,$FC,$5D,$F8,$5F,$7F,$FE,$9F,$FD,$E7,$C1,$A2,$31,$8C,$BF
	.dc.b	$E3,$DF,$7B,$75,$F7,$3F,$2F,$E5,$3E,$81,$27,$DA,$F8,$FE,$F7,$F9
	.dc.b	$51,$B8,$FE,$50,$FD,$3F,$A7,$FA,$9F,$07,$67,$E8,$BF,$5F,$AF,$EE
	.dc.b	$12,$3E,$4B,$EF,$F9,$FF,$E6,$C6,$E3,$F9,$A3,$FE,$D6,$FD,$67,$FD
	.dc.b	$91,$1F,$C1,$1F,$E6,$F9,$FF,$B7,$1C,$BF,$C0,$DC,$F6,$BF,$0D,$FF
	.dc.b	$21,$4F,$F8,$10,$FF,$7D,$38,$71,$1B,$C7,$FD,$7E,$DB,$FB,$FF,$3F
	.dc.b	$F9,$13,$F2,$5F,$61,$63,$F1,$FF,$09,$FD,$3F,$E8,$BF,$2C,$D8,$F4
	.dc.b	$7C,$CF,$F9,$8B,$E5,$3F,$CF,$99,$FA,$63,$1F,$29,$FE,$18,$ED,$FD
	.dc.b	$2F,$D0,$E2,$B5,$87,$F8,$73,$0E,$3A,$FB,$7E,$B7,$4D,$9D,$63,$FE
	.dc.b	$22,$8F,$F5,$65,$EB,$81,$7D,$72,$FD,$9F,$A5,$7B,$BF,$D3,$76,$CF
	.dc.b	$CF,$5A,$CF,$AC,$FF,$7F,$F9,$7F,$33,$F5,$A3,$2D,$39,$69,$F4,$7F
	.dc.b	$0E,$FD,$6E,$1F,$0A,$74,$DF,$8F,$FE,$5B,$1C,$0F,$F1,$DB,$FC,$A6
	.dc.b	$35,$E4,$F8,$23,$17,$F7,$5F,$F5,$8E,$9F,$CC,$5B,$7F,$D8,$5A,$FE
	.dc.b	$0C,$38,$79,$7A,$2E,$0E,$4D,$8E,$4D,$79,$4F,$F7,$79,$F0,$7C,$17
	.dc.b	$E3,$F2,$F6,$7A,$77,$2F,$33,$70,$F6,$D5,$5E,$CF,$FA,$B1,$F9,$6D
	.dc.b	$7E,$21,$AF,$8D,$E1,$B7,$F2,$78,$FA,$BC,$3F,$CB,$9E,$CF,$F8,$C7
	.dc.b	$EA,$2F,$19,$CF,$8D,$F0,$F5,$FC,$6F,$B8,$E2,$F3,$70,$7F,$BD,$5B
	.dc.b	$19,$CB,$F0,$72,$F0,$9E,$EF,$05,$EE,$D2,$57,$9B,$E5,$1D,$BE,$65
	.dc.b	$F3,$DF,$B7,$F4,$7F,$C0,$FF,$80,$DE,$CF,$D3,$71,$23,$18,$FF,$4D
	.dc.b	$C3,$8B,$FE,$FE,$7E,$5B,$79,$B7,$C9,$FE,$CB,$6C,$FF,$84,$E7,$FC
	.dc.b	$23,$C0,$F0,$3F,$BD,$70,$A3,$13,$FE,$2B,$8C,$4E,$A7,$33,$F8,$D9
	.dc.b	$F8,$9C,$FE,$67,$E3,$4D,$E1,$FD,$AD,$7F,$86,$DF,$BD,$B7,$F8,$9B
	.dc.b	$1F,$3E,$BF,$EE,$1A,$FF,$2C,$47,$A3,$D4,$91,$FE,$3D,$D7,$FD,$CD
	.dc.b	$1F,$E1,$16,$C6,$67,$F0,$CF,$FB,$F1,$79,$BF,$0B,$D3,$93,$F2,$63
	.dc.b	$C3,$F2,$92,$3F,$CD,$0F,$C8,$F3,$E4,$75,$7C,$16,$69,$FE,$23,$F9
	.dc.b	$17,$A5,$FD,$7A,$BF,$CE,$5E,$0F,$B8,$FB,$73,$AF,$4F,$EA,$F9,$58
	.dc.b	$BE,$17,$FE,$BF,$AD,$3F,$7E,$7E,$3F,$C3,$48,$FF,$3F,$3F,$7B,$C8
	.dc.b	$91,$F0,$DF,$39,$1F,$95,$C2,$F9,$AF,$A2,$48,$FC,$67,$57,$E2,$0F
	.dc.b	$79,$1F,$9C,$FD,$19,$FE,$0D,$C9,$FE,$60,$E7,$FB,$67,$E4,$38,$65
	.dc.b	$FE,$5D,$AF,$FA,$D3,$85,$9F,$E6,$3E,$1B,$FD,$41,$3F,$EA,$DF,$88
	.dc.b	$C1,$F9,$AD,$24,$7D,$F7,$E9,$91,$FC,$37,$C5,$F2,$7F,$D8,$2D,$CD
	.dc.b	$DD,$FD,$4E,$30,$DF,$8A,$FD,$4C,$7F,$AD,$AB,$FD,$DF,$5C,$1F,$F6
	.dc.b	$9E,$1A,$37,$5F,$A1,$FD,$63,$FD,$39,$3F,$E7,$E6,$BF,$E7,$B6,$D7
	.dc.b	$F9,$DD,$7C,$AF,$13,$97,$9D,$2B,$94,$7F,$02,$5F,$97,$3E,$03,$0D
	.dc.b	$8F,$3E,$9F,$E0,$38,$C4,$7F,$BD,$7E,$49,$E8,$73,$6C,$7F,$9E,$DA
	.dc.b	$FF,$AA,$DA,$FF,$4B,$6B,$FE,$0F,$FC,$C3,$FB,$7B,$E0,$02,$3F,$DF
	.dc.b	$0E,$4C,$7A,$3E,$B9,$E3,$1F,$E9,$FF,$9B,$7C,$E7,$F0,$B6,$C7,$FA
	.dc.b	$A1,$AE,$FE,$07,$46,$E6,$46,$E3,$ED,$FD,$2F,$E3,$78,$A4,$F9,$FC
	.dc.b	$67,$2A,$B7,$39,$8C,$36,$3F,$0B,$F5,$1F,$4F,$EC,$FF,$86,$CF,$B5
	.dc.b	$AB,$9B,$F1,$0A,$33,$18,$E4,$FE,$AD,$3F,$15,$AF,$FB,$6D,$AF,$F7
	.dc.b	$B6,$BF,$E3,$F6,$BF,$42,$5E,$47,$E1,$E5,$E0,$D3,$D1,$1F,$9E,$F9
	.dc.b	$BF,$D8,$7C,$B4,$C3,$E0,$47,$E8,$BF,$15,$91,$F0,$DF,$D2,$FC,$B7
	.dc.b	$E9,$BD,$09,$3F,$1D,$FD,$4F,$A1,$23,$04,$3F,$66,$FE,$6F,$F0,$4F
	.dc.b	$ED,$91,$F4,$FC,$2F,$9E,$FC,$9E,$DF,$F2,$44,$63,$83,$E5,$D7,$0A
	.dc.b	$D1,$F9,$05,$F1,$3D,$7D,$29,$3F,$91,$9C,$B7,$EC,$FC,$55,$23,$07
	.dc.b	$E6,$FD,$A7,$E7,$EF,$96,$FE,$E7,$18,$FA,$4B,$26,$CC,$AE,$53,$C5
	.dc.b	$B1,$C3,$E5,$BF,$BB,$FC,$E7,$F8,$1A,$7F,$28,$D7,$DB,$DA,$8C,$6D
	.dc.b	$FC,$4E,$1F,$81,$3F,$0A,$FF,$00,$B6,$FF,$3C,$D7,$9B,$6B,$92,$E7
	.dc.b	$FE,$1D,$FB,$E3,$E4,$A9,$2F,$F7,$5F,$83,$3A,$7F,$8A,$36,$3F,$CA
	.dc.b	$07,$F0,$27,$0E,$1A,$FF,$95,$5A,$FF,$A6,$76,$7C,$7F,$4F,$A0,$93
	.dc.b	$FC,$26,$72,$FE,$E7,$F2,$17,$CA,$FF,$88,$1B,$19,$8C,$7C,$8F,$F8
	.dc.b	$93,$F0,$BF,$F0,$9C,$FF,$A9,$F8,$91,$88,$FF,$4B,$35,$FF,$7E,$37
	.dc.b	$DE,$7E,$81,$FE,$35,$9F,$E1,$57,$FB,$EE,$3F,$A8,$DE,$FF,$E4,$77
	.dc.b	$C3,$DE,$FF,$00,$FE,$2F,$C7,$0F,$BB,$E4,$7F,$B7,$5F,$E2,$8D,$7F
	.dc.b	$C0,$51,$BC,$7F,$01,$9F,$F4,$89,$FF,$12,$35,$FF,$40,$69,$EC,$BF
	.dc.b	$F1,$34,$E6,$7E,$07,$E1,$3F,$E8,$5B,$FC,$53,$73,$FF,$5B,$36,$58
	.dc.b	$7F,$A2,$BE,$2B,$FC,$9B,$FA,$07,$E1,$E7,$F9,$53,$5F,$CB,$B7,$0F
	.dc.b	$CB,$2E,$3F,$F2,$AF,$E4,$9F,$DD,$6B,$E9,$35,$FF,$63,$35,$F0,$B7
	.dc.b	$77,$C2,$7F,$94,$B0,$FD,$17,$C0,$7F,$82,$AF,$F9,$E3,$5F,$F1,$54
	.dc.b	$EF,$5F,$A7,$E3,$FF,$86,$1F,$F6,$A7,$C5,$FF,$74,$6F,$BD,$FE,$36
	.dc.b	$9D,$5F,$8B,$BF,$FC,$A1,$A7,$BE,$3D,$31,$83,$FC,$A9,$F8,$43,$E6
	.dc.b	$BF,$CE,$6D,$9F,$EA,$0D,$7F,$7A,$BE,$6B,$FC,$8B,$F6,$3F,$9E,$9F
	.dc.b	$E0,$3C,$9E,$BE,$0F,$E0,$0D,$D7,$F1,$B5,$B9,$FF,$0B,$79,$4F,$D0
	.dc.b	$3F,$D0,$18,$7F,$B0,$E3,$1F,$4F,$FB,$68,$FF,$51,$4E,$9F,$EF,$26
	.dc.b	$23,$FC,$DD,$F8,$47,$F0,$7C,$7F,$DF,$4D,$9F,$C2,$B7,$23,$F8,$19
	.dc.b	$FF,$80,$1F,$7D,$FE,$67,$AF,$EE,$0D,$7F,$8E,$B7,$2F,$E3,$8B,$E8
	.dc.b	$7F,$97,$E5,$FC,$4E,$BF,$86,$E7,$ED,$F2,$3D,$4D,$8C,$CE,$DF,$81
	.dc.b	$F8,$3F,$C5,$CB,$FC,$2B,$48,$78,$5F,$CF,$1B,$5F,$D0,$5E,$0F,$FB
	.dc.b	$1F,$E5,$DF,$4D,$FD,$09,$B1,$FD,$1D,$AF,$F5,$4B,$7E,$29,$FE,$00
	.dc.b	$C7,$C3,$F9,$CF,$F7,$6A,$F5,$B5,$FA,$CD,$7E,$9D,$23,$11,$EA,$FC
	.dc.b	$1F,$FA,$DB,$DE,$3B,$3F,$ED,$D6,$CA,$77,$C1,$F8,$6C,$3E,$33,$76
	.dc.b	$FE,$BB,$18,$C7,$CF,$DB,$FD,$83,$1F,$D8,$E5,$E8,$8F,$A3,$18,$8F
	.dc.b	$4C,$7F,$A8,$FE,$2F,$FD,$83,$8F,$E5,$3C,$CF,$0E,$1F,$49,$B6,$F0
	.dc.b	$7F,$CA,$BE,$DF,$F6,$EF,$CF,$FF,$DD,$B3,$FE,$4D,$9C,$B7,$1E,$B8
	.dc.b	$7D,$CA,$DC,$FD,$1F,$C7,$7A,$F0,$1A,$63,$E1,$7F,$96,$5F,$E1,$1B
	.dc.b	$4A,$FF,$22,$79,$BF,$E0,$09,$CC,$7F,$91,$FE,$53,$FD,$C1,$F8,$1F
	.dc.b	$E7,$ED,$F2,$C8,$C3,$7C,$4F,$F7,$37,$EB,$0F,$85,$A4,$BF,$F3,$55
	.dc.b	$27,$51,$87,$FD,$28,$FF,$88,$23,$5F,$17,$FE,$FA,$9C,$CF,$EE,$A7
	.dc.b	$2F,$FB,$EE,$32,$FE,$FB,$F0,$6F,$F2,$2F,$2F,$F9,$E3,$F0,$6F,$ED
	.dc.b	$32,$9F,$F2,$9C,$61,$BE,$BF,$1F,$8A,$FF,$2F,$52,$B5,$18,$78,$7D
	.dc.b	$37,$F0,$C6,$CF,$7A,$E5,$85,$8F,$F4,$4E,$1E,$47,$FC,$16,$D7,$F8
	.dc.b	$A9,$E5,$C3,$C2,$E2,$34,$FB,$7A,$9E,$2F,$F1,$F9,$FE,$92,$D8,$CC
	.dc.b	$F6,$6B,$FC,$97,$E9,$7F,$93,$3F,$E9,$C6,$A6,$91,$F8,$58,$FF,$2F
	.dc.b	$E6,$F6,$47,$F2,$C9,$AF,$E3,$62,$BC,$CE,$27,$F0,$07,$31,$C3,$93
	.dc.b	$FD,$A5,$A7,$F8,$7B,$7F,$EC,$7A,$7F,$A3,$B0,$FF,$58,$C6,$30,$FF
	.dc.b	$19,$B6,$3E,$4B,$5F,$C3,$65,$FC,$BE,$7F,$22,$D8,$FC,$73,$FE,$8A
	.dc.b	$8C,$F8,$DF,$EE,$3D,$BF,$DC,$97,$FD,$D2,$B1,$5E,$27,$FB,$BB,$C0
	.dc.b	$EA,$EA,$FF,$5D,$F5,$FF,$E4,$09,$FF,$38,$36,$3F,$95,$3F,$CD,$A3
	.dc.b	$3E,$EF,$1D,$12,$F4,$F0,$DA,$D9,$D2,$38,$7E,$0C,$FA,$6A,$63,$C1
	.dc.b	$B5,$FE,$81,$9D,$3F,$C4,$7F,$06,$FF,$46,$B9,$AF,$57,$84,$9C,$E9
	.dc.b	$FD,$07,$DA,$7C,$C7,$D5,$8F,$E1,$91,$8C,$3F,$AD,$B6,$3F,$BB,$35
	.dc.b	$FE,$82,$FF,$BE,$23,$FD,$28,$D8,$ED,$3F,$0E,$31,$C3,$FE,$02,$F8
	.dc.b	$AF,$F7,$9D,$FF,$A6,$FF,$05,$79,$9F,$26,$73,$8F,$F5,$79,$CD,$76
	.dc.b	$3D,$1C,$2C,$51,$86,$BE,$A3,$7F,$82,$3E,$CB,$EB,$36,$3F,$D5,$9F
	.dc.b	$82,$BE,$7F,$35,$F6,$FF,$04,$7D,$8F,$DD,$9D,$5F,$FA,$CE,$72,$FF
	.dc.b	$AD,$63,$3A,$7F,$A9,$87,$CB,$7C,$FF,$65,$18,$AF,$F4,$AC,$62,$75
	.dc.b	$39,$3F,$CD,$3C,$9F,$91,$E2,$FF,$57,$36,$F9,$DF,$F5,$7F,$5F,$F3
	.dc.b	$4F,$F5,$9E,$BB,$E0,$FE,$83,$1F,$F5,$CB,$71,$FF,$5C,$3F,$AD,$6F
	.dc.b	$6F,$29,$7C,$0F,$90,$99,$7A,$23,$2F,$AB,$E9,$1E,$6F,$E4,$BD,$DF
	.dc.b	$9D,$5F,$EC,$89,$CD,$7F,$69,$9C,$E3,$FE,$C9,$E2,$F9,$2D,$E1,$F5
	.dc.b	$FC,$0F,$DC,$C7,$FD,$A3,$5A,$C3,$E0,$FE,$07,$DA,$7F,$D5,$4D,$73
	.dc.b	$CB,$F4,$63,$11,$FE,$8A,$FB,$EF,$CF,$65,$F8,$A8,$FF,$00,$7D,$0F
	.dc.b	$D1,$8F,$F6,$E4,$EB,$0F,$F5,$4C,$E5,$FA,$3F,$72,$F5,$7F,$63,$F8
	.dc.b	$23,$EB,$FE,$9C,$E6,$BF,$C1,$13,$97,$F0,$78,$FF,$BA,$73,$7D,$B7
	.dc.b	$9C,$93,$F8,$68,$CB,$F8,$8F,$B6,$7D,$67,$E0,$1B,$1D,$0F,$C2,$BE
	.dc.b	$1D,$BF,$DF,$2E,$9F,$F7,$DF,$E0,$8F,$BA,$FC,$4B,$6B,$94,$FF,$AC
	.dc.b	$21,$F3,$BA,$AF,$C7,$FC,$A7,$FA,$1E,$FF,$36,$DB,$FF,$60,$B5,$FF
	.dc.b	$6F,$35,$FC,$CF,$E0,$9F,$82,$9F,$F6,$AF,$85,$FE,$B5,$7C,$3E,$02
	.dc.b	$75,$5F,$E3,$5A,$D2,$DF,$ED,$53,$F0,$3F,$CA,$CE,$70,$FF,$70,$FE
	.dc.b	$08,$F7,$FF,$4F,$5A,$C3,$F0,$5E,$2A,$B5,$7E,$2F,$AE,$FE,$27,$8F
	.dc.b	$FB,$71,$AF,$FB,$E5,$AF,$ED,$1A,$FF,$B6,$3C,$2F,$98,$F8,$CC,$6F
	.dc.b	$FB,$88,$F9,$B5,$F0,$1D,$E3,$EC,$F9,$FF,$7B,$F9,$7B,$7E,$15,$F1
	.dc.b	$36,$56,$B0,$FF,$54,$CF,$C8,$A7,$6F,$67,$FC,$32,$63,$8F,$FA,$23
	.dc.b	$6B,$FD,$69,$3B,$8F,$F0,$FB,$FD,$7A,$3F,$0B,$1D,$3D,$7F,$ED,$4F
	.dc.b	$13,$DC,$66,$3E,$DF,$D1,$3E,$A7,$F8,$BB,$63,$3F,$71,$89,$CF,$7C
	.dc.b	$9E,$31,$6F,$E5,$6A,$FA,$97,$D6,$7F,$22,$6D,$FE,$29,$7D,$36,$67
	.dc.b	$5D,$12,$B5,$17,$0F,$4F,$CE,$7F,$99,$67,$DE,$D8,$F6,$F0,$BE,$3F
	.dc.b	$E0,$CE,$79,$3E,$B2,$F8,$9C,$9A,$F8,$AB,$8F,$48,$E3,$F6,$78,$CB
	.dc.b	$C2,$FA,$FE,$B3,$D9,$F4,$F6,$FD,$8E,$1F,$77,$D0,$5B,$F8,$BC,$3F
	.dc.b	$83,$5C,$5F,$1F,$E1,$3E,$31,$F1,$FC,$2B,$87,$C7,$FB,$AF,$95,$F0
	.dc.b	$9F,$2D,$B1,$F8,$0F,$C1,$1E,$3F,$E0,$1F,$D5,$47,$FB,$E1,$CB,$BF
	.dc.b	$07,$F0,$BF,$81,$FF,$7E,$D3,$FC,$D2,$D7,$FD,$FE,$D7,$FB,$3B,$53
	.dc.b	$FE,$57,$7F,$1A,$DE,$5F,$D2,$3F,$E0,$A6,$FC,$0F,$E0,$FE,$5A,$61
	.dc.b	$F8,$3F,$86,$B8,$7C,$6C,$7F,$10,$63,$0F,$E1,$7C,$3E,$C9,$FF,$55
	.dc.b	$FE,$07,$FC,$4B,$E2,$CF,$ED,$B1,$FF,$2F,$B9,$31,$3C,$3C,$8F,$F2
	.dc.b	$75,$70,$F1,$ED,$6A,$31,$7F,$37,$C0,$FD,$5F,$E0,$6E,$BF,$9D,$18
	.dc.b	$7F,$5F,$EA,$FF,$84,$B3,$FD,$DB,$5F,$F4,$AB,$5F,$F0,$88,$70,$FF
	.dc.b	$85,$17,$CE,$F4,$3A,$FB,$57,$1F,$F6,$1F,$E0,$7F,$9F,$6F,$BB,$EC
	.dc.b	$8C,$61,$FE,$C3,$7F,$BB,$3F,$ED,$26,$E9,$FE,$31,$AE,$27,$FB,$3B
	.dc.b	$DC,$FE,$FD,$C5,$F4,$3A,$7F,$75,$97,$EE,$DB,$5F,$BB,$8F,$DF,$67
	.dc.b	$8C,$BF,$3E,$FF,$01,$9F,$F1,$87,$1F,$F8,$26,$7E,$0F,$77,$F0,$79
	.dc.b	$C9,$FE,$35,$FC,$03,$CF,$06,$D7,$92,$7F,$C5,$F1,$A3,$DF,$97,$F5
	.dc.b	$06,$C6,$BE,$2F,$FC,$67,$18,$C3,$F6,$2F,$F2,$97,$F8,$CB,$6B,$97
	.dc.b	$8F,$C3,$CF,$E3,$CF,$E5,$26,$3F,$E4,$8B,$FE,$69,$87,$93,$A3,$B9
	.dc.b	$C2,$7F,$B4,$CF,$E4,$B4,$F6,$7E,$0B,$EF,$24,$F3,$7E,$17,$0F,$A0
	.dc.b	$F4,$B6,$39,$FD,$BF,$B0,$F6,$FD,$C5,$8C,$9F,$E0,$8F,$5D,$F7,$BF
	.dc.b	$E1,$FD,$3D,$F7,$FE,$96,$8C,$70,$7C,$9F,$B6,$FA,$44,$C3,$DC,$FD
	.dc.b	$5F,$B6,$F9,$67,$CC,$6C,$7B,$BE,$07,$E7,$FD,$6F,$D0,$6C,$7C,$2F
	.dc.b	$AD,$FB,$C4,$7E,$53,$F8,$1F,$9B,$FA,$E7,$D8,$6C,$7F,$88,$BB,$1F
	.dc.b	$82,$3E,$B3,$F7,$6B,$F0,$2B,$18,$73,$F2,$DF,$EA,$FA,$E2,$DA,$FF
	.dc.b	$4D,$5F,$FB,$91,$BC,$3F,$B6,$3F,$0F,$F5,$DF,$8A,$3F,$84,$FE,$08
	.dc.b	$CF,$17,$F7,$BE,$53,$F1,$FF,$2B,$F9,$06,$C7,$E0,$BE,$77,$FC,$BD
	.dc.b	$F2,$CF,$C0,$BF,$DE,$BA,$7F,$87,$23,$1A,$7E,$5A,$B7,$5A,$F9,$CB
	.dc.b	$2D,$DE,$CF,$FA,$E3,$2F,$E1,$B1,$89,$FF,$36,$46,$27,$51,$FE,$A2
	.dc.b	$9D,$4F,$F4,$3E,$FF,$D8,$7A,$9D,$9B,$1F,$EA,$8C,$FC,$B4,$E0,$FF
	.dc.b	$31,$E0,$9F,$4D,$F0,$8C,$67,$FC,$83,$8F,$FD,$15,$5E,$E9,$FB,$18
	.dc.b	$FB,$17,$E0,$9F,$E1,$37,$89,$FE,$B8,$AE,$37,$FC,$AB,$E3,$FF,$98
	.dc.b	$61,$CD,$FE,$53,$87,$9E,$75,$E5,$F5,$9C,$E1,$D0,$3F,$DA,$B5,$FF
	.dc.b	$19,$C6,$3C,$C4,$62,$75,$3F,$E3,$86,$BF,$87,$7F,$D6,$FB,$6F,$D4
	.dc.b	$FA,$D3,$B6,$BB,$6B,$EB,$F5,$7F,$A2,$E9,$EB,$F5,$AE,$9E,$F6,$D7
	.dc.b	$FB,$6B,$15,$E2,$3F,$45,$ED,$DD,$FF,$DC,$31,$F0,$CF,$95,$1B,$E0
	.dc.b	$7C,$3F,$4E,$2F,$E6,$15,$7C,$AF,$D1,$6B,$F7,$1B,$AF,$E2,$79,$7E
	.dc.b	$2D,$75,$E0,$78,$2F,$6F,$FA,$FF,$37,$E5,$23,$15,$FE,$11,$AD,$D6
	.dc.b	$A3,$F2,$93,$A7,$FC,$DD,$9F,$FA,$BB,$A3,$E0,$DF,$6F,$72,$D9,$F5
	.dc.b	$7C,$82,$AF,$17,$F7,$0F,$1F,$FD,$2B,$2F,$0B,$1A,$FF,$BE,$A3,$1A
	.dc.b	$7F,$A8,$1F,$E1,$8D,$F0,$F6,$FF,$8E,$A3,$F8,$93,$93,$F8,$B6,$EE
	.dc.b	$ED,$ED,$F1,$23,$F1,$AD,$9F,$F4,$0B,$5F,$E2,$0D,$7E,$47,$BF,$CB
	.dc.b	$7F,$B7,$1F,$E9,$99,$F9,$FC,$F9,$AF,$2F,$0F,$FB,$53,$9A,$E7,$F7
	.dc.b	$B0,$F6,$46,$3E,$1E,$47,$FA,$4A,$BF,$DB,$71,$FA,$26,$C6,$6B,$FD
	.dc.b	$10,$DF,$C0,$FA,$FC,$4C,$8F,$91,$1E,$0F,$C1,$24,$7C,$B7,$C7,$F5
	.dc.b	$F2,$3A,$3D,$EC,$D3,$1F,$39,$F4,$3D,$1F,$EB,$78,$5F,$40,$FB,$A6
	.dc.b	$61,$F6,$5A,$FF,$A5,$B0,$7D,$1F,$C4,$FA,$49,$3F,$7D,$F8,$FF,$51
	.dc.b	$23,$E0,$3E,$52,$3F,$0F,$E6,$BE,$69,$1F,$C4,$5F,$D0,$3D,$84,$7E
	.dc.b	$53,$EF,$F9,$FF,$AA,$48,$FC,$87,$C7,$7D,$82,$38,$3F,$29,$E1,$7D
	.dc.b	$CF,$9F,$91,$F9,$77,$EF,$F4,$48,$FB,$0F,$CD,$23,$F7,$FF,$00,$FF
	.dc.b	$52,$3F,$EF,$E7,$FC,$BC,$DF,$87,$FC,$5C,$6A,$3F,$D7,$EB,$E4,$33
	.dc.b	$F2,$12,$79,$47,$E8,$3F,$03,$FF,$4A,$C7,$F7,$95,$E7,$E4,$D7,$A3
	.dc.b	$BF,$42,$D6,$E3,$0F,$FB,$7B,$E6,$13,$FB,$28,$46,$1D,$46,$63,$F9
	.dc.b	$2C,$71,$F4,$3F,$9E,$BC,$1B,$1A,$8F,$F7,$9B,$5F,$F0,$17,$E5,$FF
	.dc.b	$EF,$FF,$2B,$FE,$00,$7E,$06,$F7,$7C,$4D,$AF,$F7,$9D,$6F,$CB,$FE
	.dc.b	$DD,$F4,$2F,$81,$FE,$06,$BE,$0A,$98,$8C,$FC,$25,$C3,$FC,$41,$5C
	.dc.b	$5B,$1F,$53,$EB,$E4,$7D,$97,$EB,$FC,$C4,$8F,$B8,$D7,$EC,$FC,$CF
	.dc.b	$C9,$E2,$78,$14,$92,$1C,$0D,$7F,$D3,$34,$F0,$79,$1F,$31,$C5,$7F
	.dc.b	$6A,$6B,$FE,$2B,$3F,$DA,$F1,$F4,$BF,$03,$FE,$29,$1D,$AB,$FD,$4B
	.dc.b	$1F,$C3,$A3,$07,$8D,$C5,$F8,$E3,$1E,$6F,$E5,$BE,$67,$F2,$FF,$29
	.dc.b	$27,$F3,$91,$F9,$AF,$42,$47,$E7,$DF,$CE,$FC,$CF,$E7,$BD,$4F,$CF
	.dc.b	$FA,$72,$BF,$49,$1F,$A2,$F6,$B5,$18,$3F,$50,$FE,$4F,$F0,$37,$C5
	.dc.b	$DF,$5E,$57,$CB,$DB,$D0,$AB,$6E,$78,$3F,$D8,$B1,$88,$F8,$26,$EF
	.dc.b	$B1,$DF,$E2,$E4,$71,$3E,$C4,$A3,$31,$CB,$C4,$7A,$DF,$D6,$BE,$0D
	.dc.b	$F8,$1B,$E4,$70,$DB,$E9,$3A,$9C,$FC,$84,$9C,$D2,$31,$1F,$62,$BA
	.dc.b	$FC,$8F,$F5,$4F,$9F,$76,$D2,$57,$A7,$DD,$4B,$A7,$81,$27,$51,$96
	.dc.b	$F7,$FF,$03,$FC,$09,$D5,$49,$3F,$14,$43,$AB,$81,$FC,$11,$C4,$7F
	.dc.b	$90,$9A,$FE,$65,$AF,$FB,$6A,$DF,$15,$FC,$4F,$81,$F1,$B4,$F8,$11
	.dc.b	$FD,$D3,$F0,$27,$C7,$D2,$61,$F0,$61,$F2,$3F,$DE,$BF,$00,$FF,$0E
	.dc.b	$4F,$F7,$56,$BF,$E9,$5E,$04,$62,$3F,$DD,$AD,$7F,$75,$0F,$A5,$7C
	.dc.b	$DF,$F6,$BF,$C0,$9F,$01,$4F,$F0,$4C,$EA,$FE,$A7,$93,$F8,$B7,$FD
	.dc.b	$99,$F8,$17,$06,$D7,$F8,$42,$75,$1F,$B8,$C7,$CB,$F3,$9F,$E0,$BC
	.dc.b	$CD,$24,$FF,$84,$7E,$53,$FC,$62,$F1,$6F,$89,$FB,$FB,$FF,$3B,$FC
	.dc.b	$77,$F8,$7A,$7F,$C7,$D5,$F8,$9F,$C0,$7E,$BD,$ED,$BE,$37,$C2,$6C
	.dc.b	$B0,$43,$CB,$FF,$14,$FD,$67,$F5,$FA,$FE,$44,$6D,$C3,$F3,$1A,$E7
	.dc.b	$DE,$FF,$14,$7C,$23,$E0,$FF,$C3,$1F,$11,$FE,$43,$AF,$F7,$A3,$5F
	.dc.b	$8D,$3B,$7F,$45,$F1,$FF,$DA,$1F,$E4,$3F,$19,$FE,$6C,$F8,$6F,$F1
	.dc.b	$17,$A1,$FE,$8A,$E0,$F8,$97,$B7,$E3,$69,$EB,$C3,$FC,$78,$D7,$F1
	.dc.b	$0D,$7D,$BB,$A3,$11,$F8,$8F,$C0,$7F,$26,$FE,$D5,$AF,$DD,$6D,$7D
	.dc.b	$55,$F3,$7F,$E5,$2B,$F6,$57,$AB,$EB,$BF,$4A,$FF,$17,$D1,$18,$8F
	.dc.b	$C4,$4F,$0C,$3F,$9F,$79,$8F,$C0,$3F,$A3,$D7,$D4,$8C,$7B,$1F,$EB
	.dc.b	$86,$C7,$F4,$F6,$BF,$A7,$FC,$07,$F8,$43,$F9,$56,$DF,$E3,$5B,$CF
	.dc.b	$FA,$85,$E6,$7F,$9B,$3E,$E3,$FC,$93,$7F,$C5,$DA,$FE,$7D,$B7,$F9
	.dc.b	$77,$FC,$B7,$1F,$CA,$B2,$DB,$E9,$6B,$FD,$BF,$1F,$77,$C8,$7A,$E7
	.dc.b	$33,$FE,$D2,$FC,$07,$F1,$18,$7C,$47,$E4,$37,$CA,$FE,$79,$1A,$D3
	.dc.b	$A7,$E0,$1F,$D0,$2F,$F9,$DB,$5F,$E6,$CD,$BF,$9F,$F8,$07,$F6,$EA
	.dc.b	$F2,$3B,$AF,$A0,$D7,$3E,$BF,$D1,$8C,$47,$FB,$A3,$F0,$1F,$F6,$1B
	.dc.b	$FD,$29,$FE,$B9,$7F,$71,$B7,$F0,$5D,$CF,$F4,$9A,$76,$7F,$57,$6C
	.dc.b	$7F,$54,$6B,$BF,$81,$D1,$E3,$23,$F8,$38,$FF,$09,$BF,$89,$FC,$07
	.dc.b	$F9,$5E,$3F,$E0,$BE,$7F,$F2,$6F,$C0,$7F,$A3,$FE,$38,$FF,$67,$AD
	.dc.b	$E9,$F5,$DB,$83,$8D,$18,$6F,$C0,$F1,$BF,$EE,$1F,$0B,$FC,$A6,$FF
	.dc.b	$B3,$37,$3F,$CF,$B7,$1F,$C2,$2F,$94,$E2,$4A,$FE,$EB,$19,$6F,$CA
	.dc.b	$FF,$B4,$3E,$FF,$FC,$ED,$1F,$E0,$5F,$11,$18,$8F,$F6,$53,$71,$FF
	.dc.b	$16,$4E,$BC,$7E,$44,$AE,$55,$FE,$83,$AD,$BF,$E8,$88,$E7,$F2,$BF
	.dc.b	$ED,$E8,$C4,$7F,$2A,$BF,$0C,$E4,$FF,$39,$FD,$FF,$F9,$2E,$FF,$CA
	.dc.b	$1F,$7F,$FC,$2A,$FF,$CA,$53,$FE,$56,$73,$E2,$7F,$34,$8C,$42,$78
	.dc.b	$28,$E8,$E6,$78,$73,$7F,$97,$23,$07,$6F,$8B,$FE,$7F,$7C,$AF,$AB
	.dc.b	$6B,$83,$70,$FF,$2E,$5F,$79,$E8,$FF,$48,$31,$1A,$AF,$1C,$65,$FC
	.dc.b	$EF,$81,$9E,$93,$49,$3F,$E9,$8F,$A3,$FF,$4C,$A6,$91,$D1,$EC,$3F
	.dc.b	$86,$71,$FC,$87,$C9,$15,$C9,$B3,$FD,$00,$E5,$FC,$E4,$57,$C0,$6F
	.dc.b	$9B,$FC,$E5,$9F,$D7,$DD,$FB,$BF,$FA,$26,$FF,$D0,$11,$8B,$FF,$30
	.dc.b	$9F,$E8,$D7,$11,$A9,$E1,$3F,$26,$7F,$D1,$DF,$45,$FE,$7F,$D7,$F2
	.dc.b	$BB,$23,$18,$73,$C3,$8D,$6A,$73,$83,$07,$EC,$9F,$E6,$2C,$3F,$48
	.dc.b	$D9,$FF,$47,$CF,$82,$73,$DF,$CB,$47,$C3,$FF,$8D,$A7,$FC,$CA,$FF
	.dc.b	$60,$6F,$BB,$B8,$C6,$1F,$E6,$36,$E7,$FD,$73,$E8,$FF,$CC,$77,$CB
	.dc.b	$97,$B4,$62,$FF,$C0,$3F,$80,$7F,$B2,$6B,$FB,$43,$63,$FC,$DD,$F8
	.dc.b	$0B,$E4,$7D,$FE,$1A,$B5,$97,$F6,$AF,$C2,$BF,$DF,$18,$FF,$9F,$1B
	.dc.b	$8F,$6F,$96,$F7,$46,$2B,$FB,$8F,$B6,$FA,$7F,$B3,$7F,$C7,$E3,$18
	.dc.b	$7F,$6F,$6C,$7E,$87,$85,$18,$8F,$EF,$1F,$80,$FC,$AA,$FF,$46,$B6
	.dc.b	$3F,$D1,$13,$CB,$0F,$F0,$0C,$E5,$FF,$01,$FD,$87,$E5,$2B,$FD,$1C
	.dc.b	$BD,$8F,$AC,$E6,$77,$1F,$67,$F5,$57,$A5,$EA,$CD,$18,$CF,$FC,$10
	.dc.b	$DC,$7D,$4D,$7F,$C2,$1F,$85,$FD,$A9,$FE,$F0,$D8,$FE,$84,$BE,$C3
	.dc.b	$F1,$0E,$6D,$F8,$17,$F9,$8A,$FF,$D5,$AD,$BF,$F5,$42,$F8,$5F,$C8
	.dc.b	$39,$D3,$FC,$C5,$3F,$E9,$EA,$FF,$76,$7B,$5F,$EF,$7F,$92,$FA,$AD
	.dc.b	$8C,$CE,$CF,$F4,$47,$37,$E6,$5F,$F3,$B3,$E9,$93,$9B,$FF,$38,$39
	.dc.b	$F1,$1F,$55,$FA,$CA,$FE,$25,$F5,$76,$73,$8F,$FA,$ED,$46,$F5,$FE
	.dc.b	$D7,$3E,$86,$E5,$FE,$9B,$6B,$FE,$54,$6B,$FE,$9A,$F8,$37,$9F,$E1
	.dc.b	$6D,$7F,$B2,$5F,$F4,$DB,$63,$CD,$F0,$3F,$EB,$68,$FF,$67,$3F,$2D
	.dc.b	$FF,$3D,$7C,$86,$BE,$23,$D9,$F8,$0B,$EC,$FE,$5B,$6B,$E7,$B7,$2F
	.dc.b	$F5,$93,$5F,$F3,$5F,$DF,$7E,$66,$7F,$D6,$2D,$8F,$9B,$F4,$5F,$E8
	.dc.b	$F9,$FF,$70,$4E,$78,$1F,$70,$F5,$3E,$73,$6B,$FC,$D1,$F8,$13,$EC
	.dc.b	$B3,$5F,$4F,$E3,$BF,$11,$5F,$84,$6E,$7F,$EC,$76,$BF,$E8,$06,$BF
	.dc.b	$EC,$6F,$8F,$EC,$FF,$AC,$7D,$4F,$BD,$18,$EA,$FF,$59,$28,$FC,$17
	.dc.b	$E0,$5F,$B5,$C3,$FD,$9E,$BE,$67,$3E,$86,$9F,$C4,$78,$1F,$C9,$32
	.dc.b	$FF,$48,$B5,$FF,$6B,$B5,$FF,$5B,$37,$83,$FD,$B5,$C9,$F9,$1F,$95
	.dc.b	$FA,$5A,$FF,$8E,$7C,$EF,$F3,$BE,$8C,$7F,$DA,$55,$A8,$FF,$71,$FE
	.dc.b	$04,$F8,$5F,$CD,$36,$3F,$D2,$9F,$81,$3E,$0B,$35,$F9,$FF,$83,$FE
	.dc.b	$37,$5F,$B0,$6E,$3F,$EF,$36,$BF,$EA,$C6,$BF,$EF,$2F,$AF,$F8,$CF
	.dc.b	$FB,$45,$F8,$D3,$E7,$D3,$FD,$BD,$1A,$D8,$EA,$92,$AF,$F6,$AF,$E0
	.dc.b	$3F,$CF,$9F,$F6,$BF,$5B,$EB,$BF,$B0,$DA,$B4,$FC,$59,$D5,$78,$4E
	.dc.b	$3F,$2B,$FC,$FA,$52,$8C,$36,$7F,$C4,$2D,$9F,$EF,$1F,$00,$F9,$CF
	.dc.b	$EE,$3E,$FF,$F8,$11,$F1,$D6,$F3,$FF,$27,$D6,$F0,$F1,$61,$FE,$F5
	.dc.b	$FC,$0B,$F3,$D1,$8D,$91,$89,$F8,$73,$F7,$E7,$3E,$EF,$F1,$5A,$D6
	.dc.b	$1F,$1E,$FF,$E0,$07,$D9,$FE,$BC,$FD,$84,$EB,$0E,$51,$8F,$81,$E4
	.dc.b	$73,$FB,$7B,$3E,$39,$7D,$9F,$F2,$36,$C7,$9E,$FF,$C1,$98,$F8,$73
	.dc.b	$56,$B2,$F4,$9F,$97,$F9,$B7,$D2,$FD,$86,$E3,$F1,$25,$1F,$CE,$F0
	.dc.b	$F6,$46,$2B,$81,$C7,$E7,$6D,$F1,$BF,$E2,$15,$E0,$39,$CF,$A7,$A3
	.dc.b	$85,$7B,$5F,$95,$EA,$7C,$BF,$6E,$CF,$B5,$C3,$F0,$FA,$6E,$1E,$CD
	.dc.b	$7D,$F1,$F4,$FC,$5F,$A9,$E4,$5B,$F8,$CE,$1F,$99,$E1,$DC,$39,$70
	.dc.b	$7F,$1A,$E0,$F7,$FE,$05,$FC,$E6,$7F,$7D,$F9,$3B,$E7,$7F,$3F,$39
	.dc.b	$94,$E7,$1F,$94,$FE,$26,$3F,$CE,$CD,$7F,$84,$1C,$3C,$CF,$D1,$46
	.dc.b	$63,$F8,$67,$E1,$9F,$BA,$4D,$8F,$C5,$46,$BC,$AF,$C9,$78,$73,$0F
	.dc.b	$C2,$7A,$B6,$BD,$BE,$9F,$E2,$4C,$6B,$F8,$7D,$BD,$DE,$13,$F6,$4F
	.dc.b	$F1,$F6,$3F,$EB,$B7,$F0,$4F,$F9,$AD,$CE,$1F,$E5,$A7,$2F,$F9,$85
	.dc.b	$B2,$FD,$99,$F5,$1F,$B6,$9C,$C6,$2F,$B7,$D0,$7F,$0A,$FC,$1B,$F6
	.dc.b	$D8,$F5,$8C,$78,$FF,$C3,$BD,$6F,$E2,$97,$FE,$A4,$6D,$7F,$A8,$CF
	.dc.b	$EB,$B1,$FC,$36,$33,$1F,$EC,$5C,$7F,$CC,$1A,$FF,$B4,$9B,$5F,$52
	.dc.b	$B9,$CE,$7E,$B7,$F7,$DE,$E7,$EF,$F6,$B7,$FB,$17,$0F,$ED,$FD,$BB
	.dc.b	$97,$E3,$1F,$F0,$D7,$C8,$3E,$66,$FD,$86,$1B,$7F,$C4,$67,$FC,$C8
	.dc.b	$FF,$06,$33,$7F,$96,$79,$E5,$FE,$32,$9F,$F3,$D4,$FE,$33,$C1,$46
	.dc.b	$23,$F8,$C3,$54,$3F,$35,$F9,$FA,$FF,$09,$6E,$1C,$30,$FF,$19,$46
	.dc.b	$2F,$8E,$1F,$BE,$BF,$E2,$0B,$EB,$FD,$FE,$3F,$C0,$66,$3F,$CB,$70
	.dc.b	$7C,$66,$31,$FD,$D4,$62,$7E,$17,$C3,$FB,$48,$FF,$94,$B1,$FF,$27
	.dc.b	$57,$BE,$BF,$AC,$65,$E9,$F7,$3E,$7E,$1E,$87,$E8,$FC,$EF,$C7,$26
	.dc.b	$3C,$DF,$F2,$DD,$FF,$63,$3F,$B2,$63,$FE,$48,$C3,$E2,$77,$FD,$AF
	.dc.b	$3E,$61,$F5,$FE,$42,$E9,$EA,$FB,$0F,$8E,$63,$0F,$98,$FE,$0B,$CB
	.dc.b	$FB,$04,$BF,$63,$E2,$FC,$0F,$E7,$9B,$8F,$90,$FE,$2B,$CA,$FC,$5F
	.dc.b	$45,$9F,$C4,$3F,$8E,$7F,$08,$43,$E4,$1E,$EF,$80,$FB,$66,$DF,$A8
	.dc.b	$FF,$6A,$FA,$05,$DC,$70,$DA,$FE,$D3,$D7,$FD,$DF,$6F,$F5,$BB,$63
	.dc.b	$FD,$E7,$ED,$7F,$51,$DB,$F8,$FC,$E5,$FE,$42,$7E,$4B,$F2,$FF,$C0
	.dc.b	$3F,$CE,$DB,$BF,$F5,$B7,$F2,$1D,$67,$38,$7F,$73,$E4,$7D,$C7,$E5
	.dc.b	$8F,$F4,$44,$FE,$12,$75,$1E,$DE,$7F,$EE,$DF,$1A,$FB,$EF,$F1,$D6
	.dc.b	$8E,$1F,$F6,$34,$E7,$5F,$F7,$E4,$7F,$89,$63,$13,$BE,$7E,$77,$CA
	.dc.b	$BD,$79,$7C,$06,$E1,$9E,$BF,$F0,$34,$E5,$B1,$FE,$0D,$D1,$C7,$D2
	.dc.b	$3E,$33,$77,$F8,$5C,$76,$B7,$87,$F9,$4F,$5F,$46,$0F,$C0,$9E,$DF
	.dc.b	$38,$7B,$BF,$CA,$2B,$75,$AB,$E5,$B7,$F5,$6A,$D3,$FD,$63,$E2,$3F
	.dc.b	$D3,$D8,$30,$EF,$1F,$EC,$D6,$F4,$F7,$8C,$57,$1E,$A7,$C7,$7F,$37
	.dc.b	$C1,$5F,$ED,$38,$C3,$63,$F9,$E3,$5F,$C4,$BF,$E5,$8E,$4D,$F6,$1A
	.dc.b	$49,$E3,$D7,$FD,$C3,$3F,$3F,$B7,$73,$F6,$36,$BF,$CF,$98,$FF,$8C
	.dc.b	$1F,$86,$7F,$9F,$F7,$FF,$01,$38,$3E,$5F,$C4,$71,$69,$98,$DB,$F1
	.dc.b	$EF,$EF,$77,$77,$E5,$FC,$E3,$79,$3F,$D8,$95,$E0,$9F,$D0,$75,$F9
	.dc.b	$76,$FF,$38,$65,$FE,$0A,$E4,$FD,$6F,$C2,$23,$18,$7E,$72,$30,$FF
	.dc.b	$A5,$E7,$73,$A8,$F9,$12,$F0,$FB,$2F,$0E,$98,$8F,$47,$7B,$EE,$FF
	.dc.b	$C3,$7E,$0F,$F4,$96,$E5,$FB,$56,$E3,$FB,$FF,$B8,$BB,$7E,$EF,$E6
	.dc.b	$EF,$01,$A6,$3E,$EB,$8F,$07,$ED,$7D,$0F,$C4,$37,$1F,$F0,$03,$5F
	.dc.b	$E1,$AD,$7E,$2F,$95,$CB,$87,$FC,$C3,$1F,$C6,$9A,$FF,$99,$A7,$FD
	.dc.b	$39,$3E,$6D,$FC,$93,$FD,$67,$1C,$1E,$8F,$1E,$47,$F8,$47,$4F,$E2
	.dc.b	$D1,$F9,$D6,$C6,$74,$FF,$4F,$37,$E5,$FE,$0E,$72,$3F,$D2,$52,$9F
	.dc.b	$16,$D2,$7F,$B5,$3F,$23,$CA,$91,$D5,$F0,$57,$F8,$0A,$30,$79,$73
	.dc.b	$6F,$ED,$24,$78,$3E,$B7,$E9,$91,$C1,$E4,$F0,$3E,$26,$B2,$3F,$BE
	.dc.b	$3F,$69,$87,$BC,$F1,$90,$F1,$69,$F3,$48,$FF,$AC,$7F,$2F,$7C,$EF
	.dc.b	$E1,$9B,$1F,$35,$AF,$F4,$AD,$D1,$88,$FF,$11,$35,$FF,$5C,$2F,$7F
	.dc.b	$EE,$70,$49,$F0,$47,$7F,$81,$F0,$32,$3C,$91,$F8,$3F,$A1,$91,$FE
	.dc.b	$1B,$7C,$DF,$15,$23,$CC,$FE,$07,$03,$4C,$7C,$C7,$E3,$78,$E5,$7C
	.dc.b	$18,$F8,$1F,$0F,$F0,$59,$1F,$16,$3E,$27,$D3,$7E,$57,$DF,$93,$F2
	.dc.b	$63,$E4,$7C,$3E,$A3,$07,$CD,$7F,$57,$FA,$EF,$EE,$63,$FB,$DB,$5F
	.dc.b	$D1,$37,$C0,$FF,$60,$FE,$00,$F6,$3F,$77,$48,$C3,$FC,$ED,$AF,$F9
	.dc.b	$CD,$AF,$F8,$01,$AF,$F7,$0A,$FE,$F2,$FF,$B4,$63,$FB,$93,$7D,$6F
	.dc.b	$3F,$8F,$E8,$F5,$FF,$81,$FD,$2F,$FB,$5A,$BF,$C2,$6D,$7D,$31,$EE
	.dc.b	$F6,$78,$47,$BA,$30,$FF,$94,$FB,$7F,$7C,$47,$F2,$5F,$0D,$F9,$12
	.dc.b	$38,$3F,$2D,$E6,$FC,$2A,$78,$74,$93,$FE,$EB,$F5,$1C,$BF,$21,$B1
	.dc.b	$FD,$85,$AF,$FA,$65,$AF,$F8,$99,$AF,$FB,$22,$BE,$A7,$95,$FE,$CA
	.dc.b	$97,$E0,$5D,$4F,$27,$97,$FA,$9F,$8E,$FE,$35,$F4,$53,$1F,$D7,$47
	.dc.b	$EB,$7C,$CD,$46,$0F,$D9,$3F,$CD,$BE,$ED,$F5,$9F,$C8,$FC,$99,$43
	.dc.b	$FD,$E5,$3A,$71,$D4,$D7,$5E,$32,$9F,$07,$D7,$19,$87,$0B,$DD,$6A
	.dc.b	$30,$77,$FB,$EF,$F2,$AB,$FE,$8B,$3F,$9B,$1F,$94,$72,$B1,$F1,$3C
	.dc.b	$7E,$03,$5F,$01,$23,$A8,$AF,$F1,$8B,$86,$BC,$F0,$FF,$7C,$C6,$2F
	.dc.b	$AC,$F1,$BF,$17,$1F,$F5,$EF,$94,$7D,$55,$25,$7B,$36,$F8,$9F,$0B
	.dc.b	$F3,$C9,$19,$D5,$1F,$4F,$E1,$7F,$B5,$F0,$BE,$FB,$FC,$0F,$7E,$53
	.dc.b	$FC,$5A,$E3,$5F,$83,$3B,$F7,$38,$CF,$C6,$9C,$B7,$83,$EE,$57,$5E
	.dc.b	$1D,$9C,$39,$FF,$B9,$FE,$E1,$FB,$6C,$BF,$B4,$37,$1F,$9F,$CC,$8C
	.dc.b	$63,$F8,$7A,$7D,$FF,$E1,$DF,$41,$8E,$6F,$F2,$23,$4F,$F2,$34,$6D
	.dc.b	$FF,$44,$B5,$E9,$F5,$6F,$86,$FE,$2B,$F4,$1F,$E4,$39,$FB,$AD,$7B
	.dc.b	$CE,$DF,$A5,$F7,$EE,$DF,$F0,$AA,$F1,$29,$88,$FF,$0C,$FD,$0F,$F8
	.dc.b	$96,$BF,$C0,$0D,$7F,$AA,$35,$FF,$7A,$35,$FD,$1E,$1F,$EE,$27,$F0
	.dc.b	$4D,$7E,$D7,$42,$31,$C1,$F9,$AF,$A8,$7D,$FF,$F9,$FA,$3F,$3B,$4F
	.dc.b	$0F,$FB,$EE,$1F,$8D,$6B,$D1,$B3,$CF,$EF,$F9,$71,$7E,$DB,$EF,$BE
	.dc.b	$25,$FF,$6D,$6B,$F6,$70,$46,$23,$87,$0F,$F8,$7F,$E2,$7F,$B7,$47
	.dc.b	$E7,$5F,$F4,$7B,$7E,$3B,$FC,$83,$1A,$8F,$7F,$D3,$FF,$95,$9B,$F1
	.dc.b	$1F,$E4,$68,$FF,$2A,$47,$E9,$DD,$69,$F0,$3E,$53,$FD,$3B,$3F,$E3
	.dc.b	$16,$BF,$E0,$56,$BE,$96,$BF,$C5,$3C,$9F,$C2,$4F,$CA,$6E,$B9,$EB
	.dc.b	$FF,$2C,$F1,$3E,$D7,$F7,$78,$7E,$E7,$99,$F3,$7F,$E5,$C6,$C6,$63
	.dc.b	$8F,$CD,$7F,$1B,$8F,$BB,$F5,$9F,$E5,$07,$E6,$35,$FE,$4E,$D5,$C8
	.dc.b	$8C,$2F,$BE,$FC,$B4,$FF,$0B,$5E,$EF,$F9,$B9,$B1,$9E,$CF,$F3,$8D
	.dc.b	$7B,$A3,$AB,$4E,$FA,$FF,$9D,$BC,$47,$CC,$7F,$9E,$3F,$28,$E7,$7F
	.dc.b	$E3,$16,$BF,$85,$6B,$FD,$01,$AF,$AB,$EF,$BF,$CD,$FC,$9F,$98,$5F
	.dc.b	$21,$FE,$84,$6C,$FF,$25,$77,$C9,$FD,$2B,$ED,$DF,$4B,$FE,$8A,$5F
	.dc.b	$07,$4C,$47,$F2,$6F,$BE,$CE,$1F,$C4,$4F,$E3,$AE,$1C,$6A,$8C,$47
	.dc.b	$F1,$78,$C5,$7F,$B5,$D7,$DF,$7F,$B6,$A7,$FA,$AC,$60,$FE,$6B,$19
	.dc.b	$8F,$E6,$91,$FD,$AB,$D3,$FE,$72,$7E,$BB,$5F,$E6,$AD,$CF,$F9,$8B
	.dc.b	$6B,$FD,$CE,$FF,$4E,$9F,$E7,$7F,$AB,$FF,$44,$9F,$61,$FE,$EC,$70
	.dc.b	$FB,$1A,$FA,$DA,$FF,$3E,$FB,$EF,$F4,$73,$EA,$5F,$23,$FD,$39,$B1
	.dc.b	$9F,$90,$FE,$A3,$F7,$EF,$8B,$A4,$9F,$87,$F7,$EF,$8C,$FE,$AE,$D8
	.dc.b	$F3,$EC,$8C,$47,$D2,$6B,$F4,$5A,$FF,$54,$F8,$6F,$F5,$4B,$F3,$1B
	.dc.b	$AF,$F5,$E6,$EF,$FE,$15,$F0,$DF,$3B,$FE,$54,$AF,$E5,$4D,$9D,$36
	.dc.b	$D5,$C3,$E3,$7F,$B7,$FD,$FB,$E8,$3F,$B8,$FD,$97,$F2,$8B,$6A,$FC
	.dc.b	$47,$C3,$18,$6C,$27,$83,$7C,$AE,$7F,$17,$2F,$F1,$75,$FE,$5B,$DB
	.dc.b	$FB,$5F,$C6,$FF,$8F,$A7,$FC,$BF,$F1,$5F,$E6,$23,$FD,$CB,$DD,$FA
	.dc.b	$48,$FF,$13,$F8,$9F,$A6,$8C,$92,$32,$FE,$86,$FF,$BB,$BF,$A0,$8F
	.dc.b	$F1,$E4,$FE,$BB,$BB,$FD,$DF,$39,$94,$E5,$FC,$88,$FC,$03,$F6,$B7
	.dc.b	$FE,$45,$FC,$01,$F2,$FF,$E0,$19,$FB,$F1,$8A,$FF,$25,$46,$27,$58
	.dc.b	$7F,$8E,$65,$5B,$7B,$46,$63,$15,$C6,$79,$F4,$FF,$83,$E7,$35,$BB
	.dc.b	$EB,$87,$A3,$0F,$3E,$1F,$AA,$6B,$FC,$1D,$AF,$A7,$27,$87,$C5,$B1
	.dc.b	$C5,$AF,$16,$BF,$B6,$FA,$4F,$F0,$EA,$69,$1F,$F1,$0B,$F2,$70,$FE
	.dc.b	$2B,$D3,$67,$F8,$D4,$D7,$FC,$6F,$15,$FF,$3A,$45,$E3,$FF,$30,$E9
	.dc.b	$F2,$67,$27,$CC,$BE,$AD,$3E,$EE,$9F,$EF,$2C,$3F,$86,$ED,$FC,$C3
	.dc.b	$E6,$FF,$CB,$F5,$F8,$6A,$D3,$FC,$B6,$75,$5C,$3E,$CF,$FD,$0C,$FE
	.dc.b	$51,$B3,$FE,$57,$9D,$B6,$33,$BF,$E5,$3C,$4F,$F3,$CB,$6F,$95,$61
	.dc.b	$F1,$4D,$8F,$D9,$FC,$6B,$C2,$C6,$C8,$7C,$C6,$1D,$61,$FC,$E9,$CB
	.dc.b	$63,$5B,$6D,$B1,$FD,$03,$F0,$07,$D2,$F6,$6C,$67,$DA,$C4,$F9,$7C
	.dc.b	$17,$E0,$0F,$A4,$C3,$6F,$A5,$62,$7D,$3C,$8F,$9A,$FA,$3C,$1F,$C4
	.dc.b	$23,$17,$FC,$B1,$B1,$FD,$71,$AF,$F5,$B7,$FD,$F5,$1F,$D6,$1B,$1F
	.dc.b	$D5,$9E,$3C,$5F,$E8,$87,$2F,$FA,$2B,$EA,$BF,$CC,$95,$FD,$81,$FE
	.dc.b	$42,$DF,$57,$E7,$C6,$2F,$FB,$0B,$FE,$6D,$9F,$E8,$2D,$8F,$EE,$2D
	.dc.b	$7F,$B8,$35,$FF,$49,$FD,$F7,$D6,$7F,$B5,$B6,$FF,$D3,$BF,$7F,$FD
	.dc.b	$4E,$FF,$B9,$CE,$5F,$F4,$DF,$DF,$FF,$99,$5F,$EE,$ED,$BF,$EE,$AB
	.dc.b	$EB,$3F,$0E,$E7,$0F,$F3,$6C,$BE,$E3,$E8,$FB,$5F,$EE,$16,$C6,$70
	.dc.b	$FE,$E7,$39,$6F,$17,$8E,$7F,$C7,$6F,$E3,$9F,$F3,$FB,$99,$FE,$A3
	.dc.b	$CB,$F9,$B7,$F8,$37,$89,$7C,$07,$DC,$58,$C3,$BE,$BF,$C8,$E3,$FE
	.dc.b	$1F,$BF,$EA,$BC,$DE,$89,$CC,$FF,$9E,$DA,$FF,$BD,$BE,$C1,$F8,$17
	.dc.b	$85,$BA,$FF,$B5,$DF,$EF,$8D,$8F,$37,$C0,$7C,$78,$FF,$2C,$FC,$4E
	.dc.b	$3E,$33,$2F,$AF,$F0,$07,$CB,$FC,$A6,$D7,$C8,$6E,$5E,$96,$BF,$EC
	.dc.b	$9E,$27,$E0,$1F,$E4,$8A,$4F,$FB,$86,$73,$F0,$3F,$27,$EF,$3F,$37
	.dc.b	$5F,$E0,$3B,$E1,$7B,$FC,$01,$F5,$BF,$B5,$6D,$7F,$BD,$7E,$37,$FC
	.dc.b	$9E,$F6,$EA,$FF,$59,$4E,$AF,$FD,$33,$D9,$F8,$4F,$8E,$FC,$55,$FE
	.dc.b	$09,$B3,$F4,$DA,$FF,$AA,$FE,$3F,$FC,$B3,$3F,$EA,$7F,$53,$F0,$11
	.dc.b	$89,$FF,$61,$BF,$41,$FF,$62,$FE,$00,$F0,$FF,$0C,$FF,$1D,$BF,$31
	.dc.b	$F1,$3A,$D1,$B8,$CD,$7F,$A3,$23,$83,$FE,$C5,$F3,$7F,$3B,$CF,$FA
	.dc.b	$F8,$FF,$60,$35,$E8,$D7,$FC,$4F,$D8,$FB,$8F,$E3,$CD,$C3,$FD,$BB
	.dc.b	$D1,$89,$FE,$45,$F8,$07,$F8,$8E,$7C,$7D,$08,$C6,$1F,$EE,$2F,$98
	.dc.b	$FF,$18,$56,$A7,$3C,$9F,$EE,$0F,$BC,$7A,$5F,$B4,$6D,$7E,$C9,$B8
	.dc.b	$7E,$8D,$AF,$FB,$6F,$97,$E2,$3F,$EF,$07,$E1,$DF,$FB,$B1,$CB,$7D
	.dc.b	$2F,$DA,$7D,$27,$E2,$A1,$87,$E4,$AD,$F8,$07,$F3,$B7,$FD,$F7,$5B
	.dc.b	$9D,$4E,$6B,$8D,$7E,$6B,$B7,$FC,$E3,$5F,$E2,$C6,$C7,$F9,$F9,$AF
	.dc.b	$FA,$6D,$AF,$85,$7D,$9F,$F6,$38,$46,$27,$F0,$31,$8F,$8B,$D4,$E6
	.dc.b	$17,$E4,$E7,$7D,$3B,$E6,$7F,$8A,$37,$1F,$C5,$CE,$BD,$2C,$D6,$BC
	.dc.b	$19,$5A,$BE,$18,$F6,$8F,$CB,$47,$EA,$F1,$F0,$2F,$BA,$EE,$78,$3E
	.dc.b	$D2,$D7,$3F,$B3,$F1,$98,$C3,$A1,$FE,$E8,$B4,$7C,$3F,$86,$7B,$54
	.dc.b	$F5,$FA,$17,$0F,$8B,$3F,$CE,$F0,$E6,$1C,$67,$F6,$23,$17,$D7,$D3
	.dc.b	$ED,$E5,$4C,$39,$C5,$AF,$2F,$0F,$CC,$F4,$7E,$59,$F3,$3E,$BA,$DF
	.dc.b	$C7,$F8,$FF,$9E,$62,$F8,$EF,$FD,$7D,$BF,$03,$F5,$0F,$AB,$F0,$76
	.dc.b	$FE,$87,$97,$F6,$0C,$57,$27,$F0,$DF,$80,$3D,$2B,$F3,$98,$6D,$FE
	.dc.b	$23,$EF,$DC,$FF,$E1,$2C,$51,$89,$FF,$41,$B6,$3F,$DA,$AD,$7F,$DE
	.dc.b	$4F,$DF,$8F,$D4,$36,$3F,$9D,$BC,$79,$3E,$73,$9E,$1F,$C9,$FD,$97
	.dc.b	$E4,$8F,$C9,$FA,$F7,$0F,$C7,$78,$BF,$96,$31,$8F,$C8,$7F,$C8,$DF
	.dc.b	$0C,$FB,$DF,$EC,$F9,$FE,$CA,$31,$E7,$6A,$5F,$79,$63,$1D,$73,$2F
	.dc.b	$F5,$7D,$72,$7F,$D5,$2F,$E3,$9F,$D0,$7C,$77,$3F,$3F,$F5,$2D,$8F
	.dc.b	$D9,$FD,$F3,$EF,$7E,$63,$63,$FD,$5A,$DD,$BF,$DF,$ED,$7F,$88,$0F
	.dc.b	$6B,$F8,$9C,$FE,$F9,$B1,$F9,$A9,$E5,$B7,$E6,$67,$2F,$EC,$3D,$EF
	.dc.b	$DE,$9F,$BE,$F0,$AE,$1F,$B9,$E0,$FE,$02,$63,$1F,$C9,$BF,$E1,$7F
	.dc.b	$81,$7C,$FD,$31,$F5,$96,$FF,$83,$CB,$ED,$78,$19,$C3,$F4,$B1,$99
	.dc.b	$DF,$C9,$FF,$8C,$23,$FA,$2C,$FF,$2F,$E9,$FF,$3D,$CE,$4D,$7D,$9F
	.dc.b	$8C,$F2,$36,$FC,$3C,$7E,$7F,$23,$D0,$D8,$F2,$C7,$AF,$EC,$F1,$8F
	.dc.b	$91,$F7,$79,$3E,$32,$57,$F9,$03,$3F,$E1,$15,$E5,$7F,$04,$FB,$3A
	.dc.b	$38,$CF,$C2,$7E,$47,$4F,$B4,$91,$CD,$F8,$1D,$5F,$34,$DC,$7C,$A7
	.dc.b	$E8,$7C,$CC,$8F,$9B,$F1,$B7,$0F,$97,$F3,$1F,$54,$C7,$47,$F3,$DC
	.dc.b	$FF,$B6,$CF,$D0,$F9,$AF,$9C,$4C,$3D,$8F,$D1,$F6,$BE,$F1,$F7,$DB
	.dc.b	$1F,$73,$C6,$FC,$17,$C0,$7E,$0D,$B1,$F8,$0F,$1F,$ED,$11,$F9,$0F
	.dc.b	$DD,$7F,$14,$7E,$2D,$AF,$F8,$63,$AF,$FD,$45,$F6,$7F,$BA,$DB,$FC
	.dc.b	$8F,$AF,$EE,$54,$67,$E0,$75,$F6,$9F,$EA,$8B,$FF,$55,$36,$8D,$3F
	.dc.b	$CF,$BF,$7B,$FA,$6C,$7F,$C5,$4D,$DD,$7F,$DB,$B8,$7F,$CE,$D3,$9E
	.dc.b	$C2,$F9,$EC,$9A,$F8,$7B,$8F,$AF,$3D,$3F,$EF,$8F,$1F,$FC,$B7,$97
	.dc.b	$FA,$FE,$FF,$C0,$BF,$1E,$9F,$1F,$C1,$61,$DE,$1C,$B5,$56,$DF,$3C
	.dc.b	$7F,$8F,$A3,$18,$7F,$B2,$F2,$E5,$5A,$C2,$CE,$72,$F2,$4B,$4F,$6E
	.dc.b	$FC,$1F,$7E,$F7,$6F,$F3,$CE,$7E,$19,$CE,$1C,$B0,$7D,$F7,$F7,$D8
	.dc.b	$C7,$BB,$FC,$5A,$FD,$15,$AF,$73,$CA,$FF,$83,$EB,$97,$C5,$7F,$57
	.dc.b	$6F,$3E,$A3,$13,$F0,$B8,$BF,$C6,$90,$9D,$74,$3E,$43,$F9,$4D,$FF
	.dc.b	$A2,$A9,$39,$8C,$C6,$27,$70,$8C,$3A,$EB,$FC,$2B,$78,$B7,$3F,$EC
	.dc.b	$D6,$CE,$A7,$F9,$CC,$B9,$FF,$C9,$5A,$FC,$28,$75,$E3,$4F,$5B,$6F
	.dc.b	$D1,$75,$F9,$BE,$A5,$E0,$78,$14,$D7,$17,$F7,$AA,$FB,$BC,$7F,$98
	.dc.b	$FB,$9E,$8F,$E6,$BB,$2F,$37,$EB,$2B,$6A,$BC,$5C,$2F,$83,$FE,$CF
	.dc.b	$5F,$88,$8C,$65,$A9,$F1,$CE,$5D,$BF,$5C,$FD,$6B,$FE,$36,$75,$E1
	.dc.b	$F1,$3C,$7F,$6D,$E4,$FB,$9F,$F3,$87,$83,$FE,$57,$CD,$18,$EA,$FD
	.dc.b	$CB,$70,$FC,$D4,$BE,$63,$1B,$FE,$E3,$EF,$6C,$9A,$71,$D9,$E1,$7F
	.dc.b	$77,$C3,$F8,$A6,$CF,$23,$F0,$B5,$F0,$DB,$19,$9F,$C7,$CE,$5B,$0F
	.dc.b	$27,$B2,$F2,$A9,$88,$F2,$FC,$44,$AF,$F3,$8E,$7F,$E8,$06,$DF,$A8
	.dc.b	$FC,$BE,$9F,$C5,$1B,$3F,$5E,$7F,$43,$39,$6C,$3E,$73,$E1,$FC,$84
	.dc.b	$9E,$F1,$F1,$BC,$B9,$1E,$27,$E4,$FD,$6C,$8E,$6F,$80,$8F,$6F,$9B
	.dc.b	$F9,$E4,$7D,$7E,$27,$D2,$23,$83,$DF,$F5,$FF,$59,$87,$C2,$3E,$DF
	.dc.b	$DE,$FF,$B1,$1F,$F4,$AB,$5F,$F1,$3E,$88,$C4,$7F,$82,$DA,$FF,$BA
	.dc.b	$5A,$FF,$21,$5F,$1B,$D8,$91,$DB,$C0,$F0,$9F,$42,$7F,$D5,$11,$F3
	.dc.b	$9A,$F8,$9A,$F8,$7C,$1F,$4F,$C6,$C8,$F4,$C7,$B3,$CB,$F6,$B2,$3D
	.dc.b	$B1,$F0,$3E,$36,$47,$C0,$7E,$1F,$C6,$C8,$F1,$3E,$52,$3E,$5F,$8C
	.dc.b	$F9,$24,$7E,$4F,$C6,$7C,$C2,$38,$3E,$6F,$C6,$FD,$0F,$8B,$91,$F4
	.dc.b	$1F,$A7,$F1,$72,$3E,$9B,$5F,$F5,$1E,$AF,$8C,$FD,$B4,$3E,$F1,$ED
	.dc.b	$E1,$B1,$FE,$DA,$6B,$FE,$B6,$7F,$6A,$E7,$C2,$FE,$6E,$F3,$6C,$6A
	.dc.b	$3F,$A0,$B5,$FC,$17,$C5,$7E,$FF,$EF,$7F,$8F,$47,$F8,$45,$AF,$A2
	.dc.b	$35,$ED,$48,$D3,$C7,$DC,$91,$C5,$AF,$FB,$5B,$C5,$5B,$DC,$FC,$CF
	.dc.b	$C6,$4A,$FC,$CB,$F9,$CF,$5A,$47,$E7,$1A,$FC,$AF,$8C,$F9,$3E,$DB
	.dc.b	$9A,$92,$7F,$E0,$0C,$39,$F6,$7C,$67,$11,$FD,$79,$AF,$F6,$66,$BF
	.dc.b	$E9,$0D,$7F,$95,$23,$FD,$CE,$BD,$1F,$B6,$D8,$D7,$A3,$FE,$E9,$8E
	.dc.b	$31,$F8,$2F,$87,$FD,$C9,$1F,$DC,$EF,$FB,$C2,$38,$3F,$7B,$F1,$1F
	.dc.b	$CE,$BE,$49,$F1,$14,$93,$E0,$F5,$7C,$DE,$2A,$BF,$D2,$53,$93,$FC
	.dc.b	$3C,$E1,$FF,$7B,$B7,$35,$72,$8C,$37,$E2,$7F,$33,$F7,$AF,$84,$A4
	.dc.b	$C3,$B0,$8F,$F2,$8B,$86,$AC,$3B,$BF,$36,$33,$18,$8E,$5F,$09,$E0
	.dc.b	$DE,$47,$36,$BE,$AF,$88,$F8,$DF,$13,$23,$E0,$6B,$F0,$E3,$E5,$7A
	.dc.b	$D2,$73,$B7,$F5,$56,$C7,$CF,$F8,$9F,$ED,$9E,$7B,$EF,$7E,$05,$7C
	.dc.b	$B8,$C6,$FA,$9C,$CF,$B7,$87,$F1,$15,$C7,$9B,$44,$9E,$55,$F8,$97
	.dc.b	$81,$FE,$39,$7F,$67,$C1,$F0,$FE,$F3,$F0,$33,$F3,$9A,$F0,$0E,$23
	.dc.b	$E5,$7D,$D7,$F8,$CD,$BC,$FF,$C2,$23,$FC,$6D,$5F,$EE,$E6,$BF,$9F
	.dc.b	$0E,$1F,$C4,$E5,$FC,$6B,$EF,$1F,$19,$4F,$CB,$47,$F8,$43,$EF,$3F
	.dc.b	$C9,$D5,$EC,$E3,$43,$D5,$D1,$BD,$FF,$C2,$FF,$37,$FC,$5E,$FF,$36
	.dc.b	$D7,$FC,$EC,$DB,$39,$F9,$C7,$DE,$7F,$9B,$B7,$FC,$6C,$63,$C4,$FF
	.dc.b	$13,$37,$0F,$C2,$EF,$FE,$9B,$7F,$4D,$0F,$BC,$E5,$B7,$F0,$18,$C5
	.dc.b	$F2,$9C,$D7,$FA,$27,$C2,$FF,$0B,$3E,$FF,$BC,$FC,$D3,$FE,$FB,$6B
	.dc.b	$FB,$66,$BF,$E9,$16,$EB,$FE,$71,$F5,$FF,$C7,$3F,$1E,$F6,$69,$88
	.dc.b	$FF,$1E,$FD,$0F,$F9,$3B,$7F,$D5,$DA,$3F,$CA,$78,$FD,$16,$CE,$9E
	.dc.b	$1B,$7F,$08,$F9,$67,$B5,$FE,$2C,$6D,$FF,$8A,$9A,$FF,$83,$9A,$FE
	.dc.b	$1F,$81,$F7,$9F,$E1,$3C,$7E,$8B,$70,$F9,$4D,$AF,$F5,$87,$CF,$7E
	.dc.b	$5E,$7E,$77,$DE,$7F,$10,$96,$08,$C3,$7E,$67,$FC,$BB,$D3,$F9,$2A
	.dc.b	$FD,$FF,$D3,$FF,$9B,$5F,$F1,$1B,$5F,$CE,$6E,$8C,$47,$F2,$86,$BF
	.dc.b	$EC,$3F,$BC,$FC,$B3,$FC,$29,$7C,$27,$F9,$B1,$B1,$9B,$FF,$37,$F6
	.dc.b	$3C,$DA,$49,$FE,$27,$1E,$F7,$FD,$A1,$F7,$9C,$9F,$F1,$CB,$5F,$3B
	.dc.b	$5F,$E7,$AD,$7F,$DB,$1F,$79,$EF,$E2,$FC,$0B,$FE,$73,$8F,$D8,$36
	.dc.b	$33,$F2,$7F,$E8,$4F,$B2,$7C,$15,$F8,$19,$87,$F2,$89,$CB,$FC,$7F
	.dc.b	$EF,$3F,$C3,$4F,$FB,$33,$34,$62,$3F,$84,$1F,$ED,$17,$0F,$F0,$B8
	.dc.b	$DD,$7F,$B3,$9F,$EB,$7F,$4F,$FC,$BB,$D6,$FC,$24,$62,$BF,$97,$7D
	.dc.b	$E7,$4D,$3F,$A3,$CA,$FF,$A3,$46,$1B,$D1,$FC,$C1,$B7,$FE,$E3,$FB
	.dc.b	$C7,$A9,$FC,$F3,$EF,$3F,$D0,$D3,$E8,$C5,$18,$8F,$43,$5F,$E7,$6D
	.dc.b	$7F,$DD,$CF,$FA,$2D,$FE,$88,$BE,$47,$B1,$DF,$E4,$7F,$A5,$FD,$DF
	.dc.b	$CC,$AF,$E7,$F6,$8C,$69,$EE,$6E,$9F,$D2,$97,$C3,$72,$3E,$53,$8B
	.dc.b	$FA,$8D,$7D,$8D,$7F,$6D,$F4,$1F,$E4,$CB,$FE,$4A,$DB,$FE,$3C,$D9
	.dc.b	$60,$FB,$BF,$F7,$D4,$7F,$61,$6B,$EA,$6D,$FD,$66,$DF,$F6,$27,$A5
	.dc.b	$3D,$C7,$C0,$FF,$6E,$7F,$95,$C6,$7C,$FC,$4E,$7E,$06,$E5,$FD,$C6
	.dc.b	$F8,$B6,$3F,$DB,$5C,$DF,$93,$CF,$FC,$1B,$EB,$FF,$9C,$63,$26,$BE
	.dc.b	$36,$5F,$F8,$96,$7F,$CE,$D1,$88,$FF,$2D,$4B,$E3,$FA,$9F,$EE,$27
	.dc.b	$11,$FE,$E0,$6B,$FA,$2A,$FF,$40,$4F,$F9,$DB,$EE,$7F,$D0,$58,$7F
	.dc.b	$92,$FE,$E7,$F1,$16,$9F,$C2,$46,$1B,$C9,$FD,$CD,$E3,$29,$D6,$3F
	.dc.b	$E2,$B8,$C5,$6F,$BB,$FB,$D5,$73,$6C,$73,$AE,$13,$CB,$0F,$DC,$DF
	.dc.b	$F9,$7F,$0E,$EF,$EF,$1A,$F1,$8D,$BE,$9C,$DD,$D8,$A4,$62,$7A,$C6
	.dc.b	$27,$53,$99,$E7,$F4,$9F,$E9,$C4,$D2,$3F,$EA,$07,$E5,$65,$FB,$9E
	.dc.b	$4B,$3F,$C4,$A6,$BF,$E3,$28,$AF,$F9,$EE,$2F,$1F,$F9,$4B,$3F,$A7
	.dc.b	$39,$3E,$59,$FC,$66,$B6,$7F,$AC,$27,$FB,$CC,$7E,$9A,$30,$7F,$AD
	.dc.b	$3C,$4F,$F2,$EE,$9F,$E0,$26,$C7,$9E,$7F,$C7,$07,$F8,$8D,$FF,$34
	.dc.b	$FC,$3B,$D6,$DC,$FF,$A4,$63,$13,$FE,$6E,$8E,$07,$F9,$06,$32,$FF
	.dc.b	$90,$FF,$27,$74,$F2,$84,$62,$BF,$CE,$EB,$E9,$B4,$E6,$7F,$CA,$7F
	.dc.b	$8C,$3F,$38,$7D,$4F,$92,$30,$F0,$C7,$13,$AC,$7A,$FC,$AF,$43,$AF
	.dc.b	$D4,$D8,$FF,$4A,$60,$FA,$F7,$C5,$75,$6C,$F2,$36,$E3,$3F,$77,$D5
	.dc.b	$FF,$8A,$B5,$34,$95,$F9,$CC,$3F,$C5,$58,$FA,$7D,$E3,$F0,$2B,$0F
	.dc.b	$C7,$97,$D2,$D3,$1F,$47,$70,$FF,$4F,$CB,$E9,$FF,$C7,$6D,$8F,$D2
	.dc.b	$B7,$3F,$14,$FF,$90,$3F,$41,$FF,$66,$57,$FA,$A1,$B3,$FE,$7B,$A7
	.dc.b	$D4,$7D,$99,$CF,$47,$FA,$A2,$5F,$89,$BE,$9F,$EA,$37,$1F,$BB,$F5
	.dc.b	$58,$9F,$B1,$F8,$EB,$97,$FC,$A7,$82,$31,$7F,$1D,$BA,$7F,$7C,$9D
	.dc.b	$57,$0F,$CD,$DF,$39,$F8,$36,$E1,$F8,$CF,$67,$13,$F9,$1F,$C4,$DF
	.dc.b	$4F,$4F,$C7,$FD,$6D,$BF,$F6,$4F,$9B,$FE,$DE,$7F,$86,$B7,$4F,$E2
	.dc.b	$AD,$7F,$97,$7B,$7F,$E6,$AF,$DF,$7A,$D7,$FB,$49,$B3,$FA,$3F,$AE
	.dc.b	$FD,$14,$63,$93,$FA,$2F,$E6,$1F,$55,$BD,$1F,$A2,$F9,$DF,$D8,$E6
	.dc.b	$AD,$64,$C9,$C7,$FD,$6F,$C6,$F9,$7E,$A1,$A4,$AF,$4F,$C5,$FC,$CD
	.dc.b	$3F,$AA,$B6,$3F,$D6,$1E,$17,$E7,$67,$3C,$FF,$B9,$FD,$57,$E6,$E7
	.dc.b	$FE,$BF,$6C,$FF,$B9,$EB,$FC,$3B,$7F,$EF,$29,$CB,$FE,$BB,$3F,$AD
	.dc.b	$1F,$CE,$0B,$5F,$D0,$1A,$FF,$8E,$9A,$FF,$B0,$FF,$6F,$FE,$EF,$1F
	.dc.b	$EF,$E6,$C7,$F0,$FC,$FF,$87,$46,$30,$FE,$E3,$F3,$3F,$C9,$6B,$FD
	.dc.b	$BB,$BF,$FB,$0B,$CE,$FE,$1F,$F1,$9F,$3A,$B5,$7F,$DE,$3E,$C5,$F3
	.dc.b	$FF,$EA,$8C,$11,$8D,$3F,$06,$DC,$BF,$B8,$35,$FF,$6D,$4B,$F1,$47
	.dc.b	$D2,$7D,$C6,$DF,$F2,$2F,$8C,$C4,$FF,$27,$FD,$21,$F6,$94,$FE,$4B
	.dc.b	$DB,$6F,$F0,$5A,$9F,$F7,$47,$C7,$7F,$8D,$9F,$EC,$AD,$D3,$FB,$6B
	.dc.b	$5F,$F3,$3B,$5F,$F7,$87,$D2,$7F,$10,$9E,$A6,$3C,$DF,$E7,$5F,$6C
	.dc.b	$DF,$F9,$E5,$6E,$D8,$F8,$DF,$CB,$7E,$C7,$FE,$67,$C7,$9C,$63,$E3
	.dc.b	$BF,$A2,$4B,$ED,$2C,$7F,$47,$ED,$FE,$27,$5F,$E7,$C6,$BF,$A1,$0E
	.dc.b	$1F,$D2,$37,$27,$EE,$BF,$D3,$71,$FF,$17,$36,$BC,$F1,$CE,$BF,$8E
	.dc.b	$BF,$AD,$AE,$3E,$DF,$8F,$D3,$93,$EA,$BF,$0F,$9F,$87,$F1,$BF,$84
	.dc.b	$BE,$EF,$F7,$2D,$AE,$19,$30,$CD,$7B,$FE,$B2,$56,$BA,$2E,$3E,$63
	.dc.b	$F8,$07,$E2,$1F,$84,$9F,$13,$63,$47,$C4,$AF,$EE,$11,$9E,$2F,$A9
	.dc.b	$3F,$E6,$0F,$7B,$FB,$A7,$C6,$7F,$6F,$9C,$BF,$15,$B1,$AF,$6A,$63
	.dc.b	$F0,$6B,$83,$F4,$DF,$EB,$6F,$F9,$6F,$F4,$CF,$E9,$2D,$F6,$F8,$E3
	.dc.b	$FB,$29,$CB,$E8,$AF,$E2,$15,$CF,$DF,$F2,$F9,$FE,$67,$F9,$1F,$DC
	.dc.b	$7F,$29,$9F,$AC,$DB,$FE,$49,$7C,$9F,$CB,$46,$57,$DF,$BE,$43,$FC
	.dc.b	$28,$DB,$FB,$8F,$F0,$97,$F1,$CE,$67,$E6,$EE,$ED,$F9,$A7,$2B,$B7
	.dc.b	$CB,$EC,$BA,$9D,$B1,$B1,$DB,$D1,$FA,$78,$F6,$6B,$FC,$F7,$2F,$E7
	.dc.b	$53,$FE,$5E,$6C,$7F,$99,$B2,$FE,$12,$F9,$39,$3D,$77,$F8,$07,$FC
	.dc.b	$61,$F7,$DF,$C7,$30,$FD,$43,$6B,$FA,$53,$FA,$06,$F9,$BF,$88,$FB
	.dc.b	$79,$5F,$9F,$AE,$14,$C5,$5D,$BC,$49,$7B,$EA,$B7,$F2,$3A,$BF,$26
	.dc.b	$63,$A7,$F2,$78,$FE,$96,$7F,$33,$97,$4C,$3C,$19,$7A,$BA,$BD,$6F
	.dc.b	$B3,$ED,$BE,$3D,$AD,$18,$6D,$FF,$A1,$9B,$3F,$E8,$EF,$D3,$3F,$32
	.dc.b	$FF,$6B,$6B,$CE,$BF,$CB,$11,$8F,$A7,$FE,$CD,$3F,$DC,$5A,$9F,$D0
	.dc.b	$BF,$B2,$73,$F0,$FD,$0E,$5E,$BF,$F9,$8D,$E9,$F3,$9F,$E6,$59,$FE
	.dc.b	$D3,$5C,$7A,$F1,$3F,$E6,$9F,$C5,$7F,$29,$3F,$E0,$16,$BE,$18,$D7
	.dc.b	$67,$F3,$6F,$98,$F8,$BE,$61,$A4,$DF,$FC,$37,$F9,$D3,$EB,$3F,$C1
	.dc.b	$ED,$8F,$E3,$D5,$C5,$FF,$6B,$46,$71,$FE,$1F,$3F,$A5,$E5,$FE,$1A
	.dc.b	$DA,$FE,$6C,$F9,$8C,$4F,$F4,$9F,$CE,$9E,$CF,$F8,$79,$B5,$E0,$F0
	.dc.b	$71,$3F,$D4,$3E,$1B,$D6,$49,$FF,$60,$E1,$FE,$14,$C3,$FB,$7F,$AD
	.dc.b	$F9,$BF,$38,$FE,$76,$13,$FE,$14,$7F,$AD,$46,$61,$F7,$5F,$20,$97
	.dc.b	$D9,$FF,$01,$7E,$19,$D2,$3F,$DF,$0D,$BF,$87,$3F,$E3,$88,$C6,$BF
	.dc.b	$13,$E7,$3F,$B4,$9F,$DA,$BE,$82,$E1,$FD,$93,$EA,$FF,$B4,$E7,$FE
	.dc.b	$B0,$6C,$7D,$2A,$C5,$F4,$8F,$9B,$F3,$DF,$78,$8F,$C4,$7F,$85,$D7
	.dc.b	$F9,$27,$AF,$A7,$C3,$7E,$63,$65,$62,$7F,$BE,$FC,$FC,$C3,$F0,$3F
	.dc.b	$3B,$78,$3E,$FF,$99,$F8,$C2,$4F,$D2,$7B,$F5,$7F,$3C,$D7,$FC,$05
	.dc.b	$AA,$31,$5F,$84,$AF,$74,$62,$FF,$CB,$9F,$53,$FD,$5E,$BF,$BF,$36
	.dc.b	$33,$F0,$7F,$EC,$B8,$C4,$17,$E0,$7F,$C1,$1B,$5F,$E0,$79,$E7,$39
	.dc.b	$3F,$84,$FC,$7B,$EF,$FF,$DA,$CD,$AF,$D0,$FA,$58,$9F,$A9,$F9,$1F
	.dc.b	$ED,$B6,$F6,$B6,$BF,$3F,$E9,$6B,$C3,$F3,$7A,$92,$7F,$57,$F6,$37
	.dc.b	$5F,$EE,$FB,$7F,$3A,$FA,$1F,$EA,$91,$F0,$BC,$FF,$F1,$54,$62,$3D
	.dc.b	$DE,$AE,$3C,$4F,$8B,$56,$7E,$4F,$A1,$5F,$F4,$FF,$C9,$DC,$7F,$96
	.dc.b	$43,$1E,$9C,$6A,$E1,$DA,$FC,$0B,$94,$FF,$93,$76,$FF,$48,$5F,$8E
	.dc.b	$B5,$18,$D3,$FC,$69,$5B,$C7,$B4,$7F,$8D,$FE,$D1,$F7,$5E,$57,$FD
	.dc.b	$ED,$3A,$3F,$D3,$8E,$AF,$A4,$63,$E8,$78,$7D,$13,$DB,$FB,$93,$C2
	.dc.b	$30,$E2,$D1,$88,$FC,$14,$62,$78,$C2,$1F,$0B,$FE,$5D,$8E,$4D,$78
	.dc.b	$B5,$F6,$35,$F4,$FC,$0F,$F9,$57,$6F,$B1,$4F,$82,$C6,$BF,$0D,$B7
	.dc.b	$C0,$F7,$9F,$C5,$BD,$2C,$BF,$ED,$8F,$9D,$BC,$07,$3C,$3F,$CE,$67
	.dc.b	$8E,$39,$79,$FE,$72,$F1,$B9,$B6,$37,$7D,$AB,$F0,$51,$CB,$E7,$30
	.dc.b	$67,$9B,$5F,$37,$FE,$71,$F0,$7F,$C9,$76,$8C,$73,$7D,$F6,$E1,$FE
	.dc.b	$D7,$D1,$ED,$E3,$C9,$FD,$C3,$7A,$F3,$5F,$C8,$DA,$FF,$4E,$E0,$E6
	.dc.b	$B7,$79,$58,$6C,$7F,$9D,$5A,$FF,$56,$F8,$CE,$3A,$BE,$17,$F0,$33
	.dc.b	$F8,$16,$C6,$63,$F0,$AD,$7F,$19,$EE,$72,$23,$FE,$83,$87,$8B,$ED
	.dc.b	$F8,$8F,$0F,$17,$F2,$26,$F1,$7F,$BE,$78,$3F,$D3,$3A,$FB,$38,$5F
	.dc.b	$19,$4F,$2B,$8C,$FF,$AF,$C6,$0F,$4F,$EE,$FF,$C7,$72,$FF,$3E,$B7
	.dc.b	$1F,$F5,$33,$68,$1F,$D2,$4F,$F3,$6F,$8F,$7C,$87,$D5,$6D,$7F,$55
	.dc.b	$6B,$FE,$8E,$A7,$D2,$FD,$46,$C7,$EB,$1A,$FE,$D9,$AF,$F3,$36,$BF
	.dc.b	$EA,$FF,$CE,$7F,$C0,$6F,$FB,$1B,$DD,$46,$23,$FD,$F2,$DC,$7D,$FC
	.dc.b	$FE,$12,$4F,$FA,$D6,$FF,$A2,$37,$F0,$6F,$C9,$E1,$F3,$BB,$AE,$DF
	.dc.b	$A3,$AF,$F5,$74,$E5,$E5,$F8,$3F,$B4,$F9,$E4,$7D,$CF,$F0,$A8,$7E
	.dc.b	$2C,$F5,$BB,$B6,$BD,$8D,$7E,$57,$D5,$FC,$93,$ED,$91,$F9,$4F,$5F
	.dc.b	$CD,$DF,$3D,$EA,$6C,$FC,$76,$BF,$57,$EC,$7E,$A1,$1F,$88,$FD,$7F
	.dc.b	$97,$F8,$74,$F9,$3A,$4A,$F9,$5F,$27,$F7,$4F,$C7,$91,$FB,$CF,$EA
	.dc.b	$BE,$35,$ED,$7E,$D9,$B3,$FE,$E7,$6B,$FE,$D6,$FB,$DF,$85,$CD,$FE
	.dc.b	$D4,$6F,$5F,$FB,$F5,$B3,$F9,$86,$A5,$F3,$7F,$C1,$3E,$3B,$F9,$CC
	.dc.b	$FF,$BD,$5A,$FF,$B9,$1B,$7F,$EE,$2F,$C2,$3F,$DC,$6D,$C3,$F8,$6C
	.dc.b	$61,$FD,$3D,$23,$0F,$EA,$9F,$EB,$6F,$EB,$27,$FD,$DB,$F9,$2F,$F1
	.dc.b	$86,$F9,$7F,$C6,$23,$17,$D2,$DF,$5B,$E0,$3F,$89,$B8,$C3,$FA,$AB
	.dc.b	$5F,$EB,$ED,$7F,$C3,$8F,$F3,$F7,$FE,$00,$B7,$0F,$77,$A1,$EC,$F4
	.dc.b	$3E,$37,$C1,$4B,$F8,$CF,$C0,$F8,$B9,$1E,$BC,$BC,$D1,$88,$EB,$E2
	.dc.b	$FD,$DF,$81,$91,$F7,$5F,$E6,$BE,$DB,$E2,$E9,$27,$A1,$FC,$4A,$34
	.dc.b	$E0,$FE,$EF,$F1,$5F,$CF,$BF,$41,$F5,$65,$F4,$9B,$3F,$C6,$9A,$FF
	.dc.b	$7E,$F6,$3C,$27,$F3,$CB,$FE,$C9,$1F,$E0,$66,$BF,$E4,$36,$BF,$79
	.dc.b	$FA,$AF,$97,$8F,$D6,$7B,$3D,$6F,$EA,$3F,$69,$F4,$F8,$99,$7E,$1A
	.dc.b	$30,$DE,$4F,$EA,$FF,$84,$BC,$6A,$4C,$BE,$04,$65,$BE,$1F,$F1,$6F
	.dc.b	$D6,$5E,$FD,$24,$F0,$EC,$EA,$47,$07,$4F,$8B,$FE,$3F,$FA,$53,$7A
	.dc.b	$49,$FA,$1F,$84,$FE,$69,$70,$7B,$1F,$E8,$F1,$98,$C4,$23,$1B,$FE
	.dc.b	$99,$B1,$C3,$E1,$7F,$BB,$C1,$A4,$8F,$0F,$1F,$F7,$CF,$CA,$DF,$21
	.dc.b	$49,$3E,$6E,$EE,$64,$70,$7A,$BE,$0B,$F9,$EF,$E3,$BF,$E7,$B8,$FE
	.dc.b	$94,$D7,$F9,$EB,$74,$F5,$E7,$FD,$71,$BD,$DE,$82,$57,$E5,$7F,$08
	.dc.b	$FF,$21,$4F,$C5,$F4,$5F,$26,$F8,$DF,$E9,$AD,$BF,$EB,$4D,$7E,$6F
	.dc.b	$0F,$F0,$CC,$3F,$D6,$31,$D7,$C0,$FF,$09,$C6,$23,$FC,$BA,$D7,$FD
	.dc.b	$34,$D7,$97,$AB,$FE,$24,$8F,$EB,$59,$7D,$66,$DF,$7F,$77,$FC,$53
	.dc.b	$F0,$18,$C3,$FC,$06,$D7,$FC,$33,$87,$F8,$BB,$F6,$67,$C8,$7F,$8C
	.dc.b	$3F,$20,$FE,$EF,$5F,$E0,$A6,$A6,$98,$FC,$77,$F3,$31,$FB,$D6,$BF
	.dc.b	$08,$FF,$09,$B8,$D3,$F1,$BF,$B3,$7F,$87,$E7,$FB,$9F,$CD,$BE,$73
	.dc.b	$FB,$CB,$63,$FC,$0C,$D7,$FC,$29,$EB,$FF,$7E,$FC,$4F,$FC,$E9,$0E
	.dc.b	$54,$61,$A7,$F7,$97,$1F,$87,$FE,$EA,$78,$36,$4E,$63,$FC,$55,$F6
	.dc.b	$DF,$33,$E5,$E4,$7F,$1B,$E2,$FF,$55,$3F,$E1,$3F,$AB,$FF,$2C,$FD
	.dc.b	$83,$E4,$7F,$CA,$2D,$9F,$F4,$9B,$5F,$F4,$EB,$5F,$F6,$D7,$0F,$E6
	.dc.b	$79,$FF,$AA,$B6,$3E,$5F,$AD,$FE,$33,$F8,$4F,$F0,$FF,$E6,$CF,$5E
	.dc.b	$92,$BF,$C4,$DE,$67,$F9,$DB,$ED,$DF,$97,$FF,$BA,$A7,$FD,$5B,$18
	.dc.b	$F3,$BF,$D4,$ED,$8C,$CE,$DF,$F7,$77,$27,$FA,$D7,$E0,$FF,$C5,$FF
	.dc.b	$37,$FE,$14,$EC,$FD,$43,$78,$3F,$CA,$B9,$7F,$9F,$3E,$B3,$FC,$AD
	.dc.b	$89,$A6,$3F,$62,$FE,$FF,$1F,$D1,$FF,$56,$7C,$3F,$F9,$E9,$B1,$FE
	.dc.b	$BC,$6B,$FE,$C6,$6B,$F2,$BE,$07,$FD,$17,$CB,$FD,$28,$FE,$97,$6E
	.dc.b	$8F,$89,$F9,$B3,$B2,$93,$0F,$F2,$77,$D1,$7C,$9F,$83,$FE,$97,$0E
	.dc.b	$54,$61,$BF,$03,$F2,$A3,$1F,$8A,$FE,$87,$8B,$FC,$EB,$48,$C5,$F3
	.dc.b	$6B,$FE,$79,$F2,$3E,$47,$C8,$48,$FE,$D5,$C6,$8F,$F5,$3F,$E4,$7F
	.dc.b	$3E,$7F,$D0,$2F,$F9,$AD,$FF,$0E,$87,$0F,$F8,$81,$AF,$0F,$AB,$FF
	.dc.b	$53,$C7,$B2,$3F,$BC,$35,$FD,$E1,$CF,$3F,$0F,$A7,$F8,$DF,$CD,$BF
	.dc.b	$B9,$F5,$7F,$4D,$D5,$18,$DB,$E0,$B7,$0F,$F4,$BF,$CA,$3E,$9B,$FD
	.dc.b	$1C,$7F,$A8,$09,$1F,$E8,$F7,$F2,$BF,$8E,$FF,$A4,$31,$EC,$DC,$3E
	.dc.b	$6B,$6B,$AF,$D9,$7F,$A6,$48,$FF,$A2,$5F,$F4,$E7,$DA,$7E,$97,$E8
	.dc.b	$5F,$5D,$49,$5F,$E1,$9F,$95,$FF,$22,$BF,$E6,$2C,$7F,$DA,$F1,$82
	.dc.b	$7D,$56,$3E,$92,$57,$17,$FD,$BB,$F8,$57,$68,$FF,$12,$35,$F9,$AD
	.dc.b	$D7,$FB,$9B,$70,$FF,$2B,$BF,$E2,$C9,$DF,$C9,$FF,$B7,$A3,$11,$F8
	.dc.b	$06,$BC,$23,$5E,$F7,$F8,$0A,$1F,$84,$BE,$3B,$51,$FE,$CA,$8C,$3F
	.dc.b	$EC,$CB,$F0,$46,$21,$0A,$E9,$39,$8C,$63,$CE,$FA,$DF,$81,$B5,$BB
	.dc.b	$E3,$1E,$E7,$CD,$3E,$B6,$B9,$AF,$CE,$BE,$58,$CB,$E7,$F7,$9F,$85
	.dc.b	$6A,$3C,$F1,$87,$8E,$28,$C4,$7C,$96,$BD,$BE,$B3,$93,$FE,$EE,$6A
	.dc.b	$69,$17,$DA,$7C,$75,$E6,$FF,$8C,$63,$11,$C4,$DE,$7F,$B8,$9D,$3E
	.dc.b	$18,$C9,$FE,$98,$8C,$BF,$BE,$AF,$F8,$03,$5D,$E7,$33,$FE,$9E,$9C
	.dc.b	$D7,$5D,$AC,$F1,$3A,$83,$CE,$E9,$C5,$7F,$91,$9E,$47,$CD,$70,$F9
	.dc.b	$63,$E3,$FD,$26,$23,$A3,$53,$4D,$7E,$04,$FA,$FA,$4A,$FB,$51,$96
	.dc.b	$FD,$77,$02,$38,$37,$F2,$9D,$08,$E0,$E7,$F1,$3E,$02,$38,$3B,$FB
	.dc.b	$7E,$42,$38,$3C,$7F,$31,$E8,$23,$83,$FB,$5F,$A9,$F0,$7C,$3F,$EB
	.dc.b	$71,$82,$6B,$FD,$CB,$E0,$F4,$48,$FE,$E6,$F1,$F4,$64,$7F,$76,$6C
	.dc.b	$FE,$8B,$57,$A9,$F2,$89,$5F,$21,$FA,$5E,$C3,$7A,$49,$FD,$34,$65
	.dc.b	$F9,$FF,$1F,$F5,$08,$E0,$FA,$7F,$1F,$F6,$08,$E0,$E7,$E5,$F8,$08
	.dc.b	$E0,$FB,$7F,$5D,$F8,$02,$38,$3E,$FF,$83,$F8,$42,$38,$3F,$79,$FA
	.dc.b	$3B,$E1,$3F,$1C,$D8,$CC,$60,$FC,$6F,$8F,$23,$F1,$6D,$DF,$E5,$4F
	.dc.b	$E4,$89,$3F,$90,$7F,$34,$FF,$07,$3F,$86,$16,$3F,$89,$B5,$FE,$E2
	.dc.b	$D7,$F5,$83,$77,$BB,$FA,$02,$4F,$E7,$9F,$D1,$FA,$32,$3F,$42,$DA
	.dc.b	$FE,$29,$AF,$EB,$7B,$8D,$25,$7E,$A3,$6F,$D6,$91,$C1,$FA,$CF,$43
	.dc.b	$F6,$24,$70,$7F,$29,$FA,$A7,$C4,$FE,$ED,$B1,$98,$C1,$FB,$FF,$AB
	.dc.b	$FF,$3A,$47,$F3,$16,$E5,$FC,$E1,$AF,$F8,$59,$AF,$C6,$FB,$EF,$F3
	.dc.b	$54,$FF,$39,$6C,$7F,$40,$6B,$FE,$1E,$6B,$C7,$EA,$BF,$88,$41,$A4
	.dc.b	$9F,$E8,$9F,$5D,$FC,$7D,$B5,$FC,$5E,$30,$FF,$4E,$6B,$FE,$2C,$5F
	.dc.b	$51,$B2,$4F,$47,$97,$74,$8E,$ED,$AE,$9E,$DF,$72,$4F,$91,$F1,$7A
	.dc.b	$D2,$3C,$ED,$9F,$EC,$3F,$69,$FC,$8D,$BE,$37,$F2,$F8,$C4,$7F,$90
	.dc.b	$9A,$FE,$49,$FD,$F3,$CB,$BB,$81,$2F,$FB,$44,$BC,$6E,$84,$8C,$BD
	.dc.b	$FB,$24,$70,$6D,$F9,$BA,$FC,$84,$9E,$AF,$F4,$5F,$BB,$61,$49,$3F
	.dc.b	$DE,$63,$2D,$FB,$3E,$04,$70,$6F,$E6,$FA,$11,$C1,$CF,$ED,$7C,$04
	.dc.b	$70,$77,$F5,$FC,$84,$70,$7F,$82,$FE,$EF,$F0,$79,$FE,$F1,$AF,$F6
	.dc.b	$96,$A6,$9F,$DB,$1F,$F0,$CB,$5F,$F3,$BB,$5E,$9E,$77,$F6,$98,$FE
	.dc.b	$93,$1C,$38,$F0,$4A,$F2,$79,$3B,$23,$83,$D1,$C7,$CC,$8E,$0F,$67
	.dc.b	$CD,$F7,$23,$83,$E0,$7B,$9E,$32,$38,$3F,$C5,$DF,$75,$F8,$76,$F2
	.dc.b	$7F,$80,$63,$19,$7F,$8E,$1A,$FF,$A5,$97,$A1,$B2,$57,$47,$97,$24
	.dc.b	$8E,$ED,$AE,$9E,$3F,$72,$4F,$91,$FF,$09,$7E,$C9,$F8,$A9,$FF,$24
	.dc.b	$B5,$FF,$53,$B5,$FE,$6E,$D7,$FC,$49,$DF,$FE,$3C,$AF,$E3,$9E,$A7
	.dc.b	$F1,$F2,$3F,$C7,$FA,$3F,$91,$11,$C1,$FC,$8F,$9F,$F9,$31,$1C,$1F
	.dc.b	$E5,$7C,$1F,$2B,$E8,$24,$65,$FE,$5D,$F3,$12,$3F,$97,$37,$2F,$6F
	.dc.b	$83,$F0,$09,$3D,$5F,$E6,$DE,$1C,$8F,$E6,$CD,$9F,$E6,$DE,$FF,$F3
	.dc.b	$B2,$4F,$F3,$B7,$F9,$F7,$BD,$23,$F9,$F3,$67,$FC,$EB,$F6,$6E,$8F
	.dc.b	$F2,$13,$6B,$FC,$E8,$D9,$DC,$FF,$49,$F7,$65,$7F,$49,$6C,$FF,$49
	.dc.b	$F6,$FF,$A7,$12,$7F,$A7,$3F,$D4,$7E,$76,$47,$F5,$16,$CF,$F5,$1F
	.dc.b	$07,$FA,$B1,$27,$FA,$B3,$E6,$F9,$69,$1E,$66,$CF,$B3,$E5,$3D,$64
	.dc.b	$9F,$5B,$FD,$87,$C3,$91,$FD,$85,$B3,$FD,$87,$C3,$FE,$CC,$49,$FE
	.dc.b	$CC,$FF,$69,$F9,$59,$1F,$DA,$5B,$3F,$33,$C9,$F9,$64,$9F,$96,$FC
	.dc.b	$EF,$05,$1F,$C5,$B8,$11,$8A,$FF,$3D,$46,$27,$53,$97,$FB,$A7,$35
	.dc.b	$AF,$F0,$CF,$5F,$E5,$27,$F9,$0B,$63,$FD,$08,$D8,$FF,$0F,$C3,$ED
	.dc.b	$9F,$0B,$CE,$3F,$D1,$D1,$89,$FE,$56,$D8,$FF,$47,$36,$39,$AF,$A3
	.dc.b	$FF,$55,$CF,$FB,$0A,$31,$E4,$7F,$AF,$DB,$19,$9D,$BC,$BC,$C7,$B7
	.dc.b	$B6,$C7,$36,$BB,$8E,$9E,$37,$F8,$EB,$1F,$1B,$67,$FC,$7B,$87,$F6
	.dc.b	$77,$D7,$19,$7F,$84,$FC,$7D,$5A,$23,$0D,$8F,$9B,$1B,$AF,$F3,$EC
	.dc.b	$7C,$6F,$77,$04,$7E,$4F,$AD,$FE,$D6,$9F,$F2,$1C,$7C,$38,$C1,$FE
	.dc.b	$DA,$B7,$D5,$D2,$4F,$F8,$DD,$D7,$D3,$48,$FF,$1D,$35,$FF,$1D,$7C
	.dc.b	$E7,$32,$3F,$E3,$F7,$C5,$E5,$48,$FA,$8D,$7E,$C7,$95,$E7,$23,$F6
	.dc.b	$5F,$37,$8B,$23,$FC,$94,$D7,$FE,$00,$FC,$6F,$FB,$C8,$8F,$FB,$D9
	.dc.b	$C7,$25,$24,$7E,$50,$FF,$6F,$47,$E1,$1F,$CB,$BF,$ED,$47,$41,$C3
	.dc.b	$E0,$F8,$6E,$C7,$CF,$F6,$E4,$7E,$45,$AF,$9B,$AB,$C8,$47,$AD,$23
	.dc.b	$D1,$F2,$12,$72,$D7,$DF,$F5,$5E,$D2,$3C,$9F,$3A,$FB,$87,$8D,$86
	.dc.b	$C7,$B1,$AF,$C3,$6B,$F0,$BE,$2B,$E5,$11,$F3,$BC,$FC,$19,$1C,$DA
	.dc.b	$FD,$4F,$B3,$FA,$44,$7D,$EF,$D5,$F7,$24,$7C,$26,$BF,$73,$D8,$FB
	.dc.b	$44,$7E,$3B,$EE,$FC,$ED,$85,$24,$FD,$88,$CB,$F6,$BE,$CD,$E5,$7C
	.dc.b	$66,$C6,$63,$06,$7C,$A9,$1C,$5A,$FE,$27,$EC,$7F,$95,$37,$C2,$FA
	.dc.b	$31,$88,$FE,$42,$D7,$FD,$82,$FF,$22,$7F,$1A,$D7,$A7,$8D,$C8,$E7
	.dc.b	$F1,$B2,$36,$D7,$C5,$E1,$78,$08,$F4,$7C,$7F,$17,$23,$BB,$5F,$57
	.dc.b	$47,$A0,$8F,$91,$F5,$FC,$2C,$8F,$3B,$5F,$D5,$79,$2F,$4F,$F2,$2D
	.dc.b	$8F,$D4,$B5,$DC,$6B,$D2,$91,$A6,$BD,$3C,$8E,$64,$79,$BD,$BC,$59
	.dc.b	$1D,$9A,$F9,$3C,$5F,$19,$1F,$1B,$FA,$6F,$AF,$76,$D2,$4F,$F0,$08
	.dc.b	$CB,$79,$F8,$11,$C1,$BE,$4E,$84,$70,$73,$FA,$0F,$01,$1C,$1F,$C3
	.dc.b	$BE,$AB,$F0,$4F,$F6,$86,$A2,$38,$6F,$27,$F0,$08,$7D,$6D,$24,$F5
	.dc.b	$75,$C1,$23,$C0,$D7,$F9,$24,$B8,$FA,$12,$3C,$CF,$7F,$2E,$47,$A5
	.dc.b	$AF,$2E,$3F,$21,$1F,$73,$FC,$5B,$EA,$FF,$B6,$C7,$F8,$19,$AF,$BC
	.dc.b	$38,$7E,$13,$5C,$F1,$F3,$23,$FC,$E2,$11,$DF,$C2,$93,$96,$BF,$D0
	.dc.b	$E5,$EB,$F9,$49,$1C,$DF,$E5,$BF,$56,$F5,$69,$27,$FA,$3C,$65,$FE
	.dc.b	$93,$F5,$7F,$CF,$5B,$E4,$7F,$36,$8C,$47,$F5,$36,$BF,$E2,$8F,$AB
	.dc.b	$79,$BF,$CF,$1B,$1F,$D5,$1A,$FF,$5B,$6B,$FD,$6B,$DB,$FE,$C4,$47
	.dc.b	$FA,$7B,$E1,$C6,$47,$99,$AF,$F6,$CF,$A2,$FE,$D4,$47,$FA,$BB,$FD
	.dc.b	$B7,$E5,$64,$7F,$5A,$6B,$FD,$B7,$EA,$7C,$6F,$DB,$6B,$F9,$B6,$A6
	.dc.b	$F9,$FF,$CD,$23,$FB,$EF,$8F,$FD,$EC,$8F,$F6,$8E,$BF,$EF,$C4,$70
	.dc.b	$7F,$78,$FC,$CF,$93,$FE,$52,$6B,$F8,$D6,$BF,$AD,$6B,$FD,$FF,$B3
	.dc.b	$FC,$14,$47,$FB,$BB,$FD,$BB,$E9,$95,$49,$3F,$E0,$88,$CB,$FE,$0A
	.dc.b	$FA,$8F,$F5,$13,$76,$FE,$EF,$18,$8F,$F0,$A3,$5F,$F3,$4C,$B4,$F0
	.dc.b	$12,$3B,$3F,$E1,$FF,$A6,$91,$FE,$0C,$6B,$FE,$1F,$FA,$4F,$E0,$AF
	.dc.b	$F8,$71,$AE,$7B,$FF,$C1,$31,$88,$FF,$11,$F0,$FF,$8A,$C8,$FF,$86
	.dc.b	$1D,$FC,$C4,$8E,$8D,$7C,$1E,$4F,$42,$3E,$17,$BF,$6C,$8F,$23,$5F
	.dc.b	$F1,$EF,$C7,$FF,$8E,$C8,$FF,$8A,$1F,$F1,$F7,$C1,$48,$FF,$15,$B5
	.dc.b	$FF,$22,$FC,$7F,$F9,$0C,$8F,$F8,$C1,$FF,$23,$7D,$14,$8F,$F1,$9B
	.dc.b	$5F,$F2,$37,$C2,$FF,$79,$7F,$99,$35,$34,$9E,$3F,$F8,$C9,$FE,$6D
	.dc.b	$1A,$8C,$1F,$E3,$6F,$9F,$7C,$4D,$24,$FF,$94,$23,$2D,$F4,$38,$11
	.dc.b	$C1,$BF,$19,$F4,$3F,$D1,$FE,$43,$FC,$43,$2B,$FE,$E5,$18,$6F,$9B
	.dc.b	$FC,$BD,$FE,$4F,$5C,$63,$17,$D2,$D7,$C7,$0F,$E6,$CD,$8E,$07,$F5
	.dc.b	$7C,$39,$BC,$2F,$9E,$1A,$8F,$25,$F2,$9F,$13,$63,$96,$1E,$1C,$3D
	.dc.b	$0F,$8E,$3F,$D9,$B8,$76,$6C,$65,$EE,$7F,$9F,$E3,$FC,$C7,$F4,$3F
	.dc.b	$37,$CA,$34,$93,$FE,$6A,$F9,$BF,$F3,$5A,$69,$1D,$0F,$A2,$F9,$EB
	.dc.b	$E0,$FF,$DD,$D1,$88,$E4,$6F,$4F,$F7,$2C,$E9,$FF,$4A,$57,$F7,$FA
	.dc.b	$E7,$39,$AF,$F7,$5C,$E6,$B8,$FB,$56,$72,$7C,$F3,$FB,$55,$7F,$BF
	.dc.b	$F4,$F1,$5F,$D1,$9F,$35,$57,$CB,$87,$82,$3A,$47,$F7,$39,$AF,$03
	.dc.b	$D4,$78,$27,$83,$F9,$29,$F5,$BA,$3F,$C3,$56,$8F,$F0,$E6,$D6,$7E
	.dc.b	$47,$2B,$DE,$C4,$E7,$E8,$5F,$87,$39,$33,$8F,$4C,$38,$3D,$0F,$F1
	.dc.b	$45,$70,$9D,$E2,$8C,$4E,$E3,$FC,$5D,$F4,$0F,$8E,$DD,$6A,$7D,$86
	.dc.b	$A3,$9E,$8F,$A9,$7A,$BC,$3F,$0E,$7C,$EE,$A7,$39,$F7,$C9,$5A,$8C
	.dc.b	$37,$2E,$E3,$37,$C8,$75,$C7,$D7,$39,$8F,$3F,$BE,$F9,$EF,$F4,$86
	.dc.b	$5F,$EF,$E6,$C6,$72,$FF,$71,$4E,$5F,$F6,$FE,$2F,$A5,$EA,$D8,$F4
	.dc.b	$57,$1C,$7C,$B3,$97,$CF,$FA,$4F,$F8,$62,$BE,$64,$23,$0F,$CA,$DD
	.dc.b	$E1,$3E,$37,$D7,$5A,$E6,$FF,$56,$E8,$F8,$AF,$E6,$4D,$CB,$FD,$EA
	.dc.b	$DC,$BD,$73,$AF,$43,$E2,$CB,$D7,$B1,$87,$EE,$FE,$1C,$F8,$CA,$7D
	.dc.b	$CF,$CA,$3D,$D9,$FE,$46,$75,$9F,$D9,$FC,$3F,$E1,$69,$FE,$57,$CD
	.dc.b	$18,$8F,$C8,$B5,$FC,$83,$7F,$19,$7E,$01,$F9,$89,$FC,$6D,$3F,$10
	.dc.b	$7D,$17,$E6,$2F,$7E,$17,$E5,$27,$39,$7C,$EF,$C4,$1D,$1F,$84,$6C
	.dc.b	$7E,$76,$FF,$8D,$7C,$A6,$6B,$F1,$3F,$38,$FA,$BF,$EB,$CD,$BF,$F3
	.dc.b	$96,$BF,$E7,$38,$C3,$80,$E1,$FF,$3C,$61,$FD,$1B,$CE,$FC,$AF,$E2
	.dc.b	$1E,$29,$FD,$4C,$E6,$FF,$0B,$39,$E3,$FE,$05,$83,$F0,$57,$47,$EE
	.dc.b	$23,$0F,$C3,$FC,$67,$F6,$D8,$7E,$E6,$72,$D8,$FE,$1C,$F9,$E1,$18
	.dc.b	$CB,$DB,$18,$9F,$F7,$B3,$5F,$76,$4F,$AC,$7C,$BF,$8D,$FC,$2C,$FF
	.dc.b	$AD,$67,$53,$9F,$0F,$F6,$BF,$88,$BE,$A6,$9F,$C3,$3F,$07,$7C,$C6
	.dc.b	$31,$FD,$39,$FE,$98,$3E,$76,$9F,$C9,$F8,$BF,$C2,$51,$98,$FA,$4D
	.dc.b	$7F,$25,$F8,$4F,$F9,$52,$3F,$63,$F8,$8F,$F7,$87,$F8,$84,$66,$7F
	.dc.b	$86,$CE,$5F,$E3,$BB,$BF,$11,$FF,$68,$5F,$CD,$6C,$FD,$AF,$C9,$5E
	.dc.b	$FF,$D5,$9C,$F3,$7D,$AF,$C4,$BF,$61,$87,$D1,$6E,$1F,$41,$AF,$E3
	.dc.b	$5A,$FC,$BF,$98,$FF,$57,$4F,$DD,$F9,$27,$C7,$FE,$8E,$73,$5E,$89
	.dc.b	$CD,$7F,$10,$FC,$49,$F2,$5F,$C9,$9B,$1F,$A3,$FC,$01,$F6,$59,$AF
	.dc.b	$E5,$7F,$6E,$FB,$2F,$F2,$03,$6F,$E9,$B7,$2F,$D1,$35,$FD,$0F,$B9
	.dc.b	$F1,$27,$F9,$0F,$C8,$FF,$15,$E0,$FE,$BD,$DD,$FD,$37,$8F,$F9,$CC
	.dc.b	$E6,$BF,$DF,$DC,$AF,$DF,$B9,$65,$F5,$1B,$8F,$F6,$7F,$35,$F5,$5F
	.dc.b	$C0,$AB,$59,$7F,$44,$FB,$87,$D7,$7F,$96,$5B,$9F,$E3,$1B,$5F,$BB
	.dc.b	$6B,$FC,$01,$FE,$65,$1F,$EF,$36,$C0,$3C,$B1,$FE,$F5,$18,$F9,$DF
	.dc.b	$E3,$1F,$91,$3E,$EF,$FA,$BF,$DE,$7F,$A6,$F1,$FE,$29,$39,$7F,$A3
	.dc.b	$C6,$6F,$FB,$0B,$70,$EB,$E4,$7F,$27,$7B,$C6,$6B,$C9,$5F,$B7,$F4
	.dc.b	$BC,$3F,$0B,$E0,$F9,$DB,$5D,$9C,$3F,$D5,$B1,$FC,$5E,$3D,$27,$FB
	.dc.b	$6C,$62,$7E,$AF,$E8,$FF,$C2,$6B,$C3,$CB,$FC,$56,$31,$3E,$2C,$D3
	.dc.b	$F0,$0F,$EE,$FE,$B7,$F1,$8D,$3D,$B1,$8E,$8C,$F9,$93,$0F,$ED,$37
	.dc.b	$BC,$BE,$6B,$FC,$43,$0F,$A5,$8B,$E2,$7E,$69,$F4,$3E,$9A,$E1,$CB
	.dc.b	$E1,$FC,$26,$2F,$FC,$1B,$87,$F8,$3F,$1F,$97,$F0,$7F,$58,$FB,$3F
	.dc.b	$59,$70,$FE,$F7,$F7,$9F,$8F,$AF,$BE,$D8,$F8,$15,$FE,$14,$8C,$61
	.dc.b	$F8,$7F,$8A,$F8,$AD,$AF,$F1,$9F,$6F,$E6,$BC,$2F,$F0,$DD,$6A,$96
	.dc.b	$FB,$8F,$F5,$93,$70,$FE,$4A,$D9,$FE,$A8,$D7,$FA,$CA,$FB,$8F,$21
	.dc.b	$E5,$AB,$5F,$E1,$6B,$7D,$67,$8B,$93,$C6,$D9,$D4,$7F,$8C,$3E,$F3
	.dc.b	$F8,$FC,$7E,$42,$BE,$B3,$F9,$28,$CE,$5F,$07,$E3,$7D,$A7,$BB,$C9
	.dc.b	$B5,$FE,$26,$AF,$E0,$55,$FA,$D6,$BF,$E4,$6A,$F0,$7C,$4F,$C6,$E5
	.dc.b	$97,$E9,$ED,$B7,$FE,$48,$C3,$F0,$93,$FD,$29,$B1,$FD,$E5,$AF,$F7
	.dc.b	$E6,$BE,$9F,$9C,$F7,$1E,$FE,$8B,$5F,$E3,$CF,$BF,$FF,$11,$D7,$F9
	.dc.b	$59,$BC,$DF,$C0,$BE,$87,$FC,$7B,$5B,$F8,$BF,$F2,$BF,$DE,$FF,$96
	.dc.b	$A3,$FC,$8A,$DC,$39,$4E,$BE,$4B,$FC,$BB,$8F,$C2,$3E,$1F,$E6,$5F
	.dc.b	$E0,$4C,$3C,$5C,$1D,$3E,$E6,$46,$0E,$AF,$FA,$03,$E7,$3F,$D0,$18
	.dc.b	$7F,$16,$CB,$F8,$D4,$62,$BF,$C3,$8D,$8F,$F1,$0B,$5F,$F1,$1F,$DF
	.dc.b	$7E,$71,$3F,$9E,$8E,$1B,$7F,$22,$8F,$D0,$FA,$4F,$2B,$FC,$13,$87
	.dc.b	$F8,$5F,$DE,$7D,$2F,$F2,$6F,$85,$FF,$0F,$46,$27,$53,$FE,$7E,$FB
	.dc.b	$BF,$D5,$FA,$CA,$D6,$BF,$E8,$69,$F1,$4F,$F8,$3E,$FD,$6D,$9F,$5B
	.dc.b	$5F,$E4,$FA,$FF,$8E,$BE,$17,$E2,$47,$F3,$DD,$5F,$7B,$FE,$8E,$6E
	.dc.b	$1F,$EA,$7B,$E0,$FF,$9F,$A3,$35,$FE,$A6,$9F,$F2,$EF,$CB,$F4,$6D
	.dc.b	$7C,$EF,$1F,$FD,$00,$D9,$F0,$D7,$CC,$F5,$3C,$07,$D9,$AD,$C7,$FA
	.dc.b	$93,$0F,$EA,$7C,$3F,$37,$C0,$FE,$17,$C7,$FE,$56,$3F,$B7,$F1,$FF
	.dc.b	$A5,$5F,$ED,$06,$73,$FE,$B5,$8A,$31,$8F,$F5,$DE,$07,$D2,$74,$F8
	.dc.b	$EF,$A4,$F8,$FD,$4F,$66,$1F,$E2,$BB,$F9,$6D,$7F,$1A,$D7,$FB,$47
	.dc.b	$D8,$FE,$65,$FF,$5E,$B7,$1F,$F7,$17,$0F,$FA,$12,$FF,$D7,$B3,$9C
	.dc.b	$FC,$FF,$82,$3E,$DF,$F9,$4B,$63,$FB,$5F,$8F,$89,$FC,$47,$D2,$FC
	.dc.b	$CD,$CD,$25,$7F,$93,$A1,$F6,$BF,$8F,$38,$C7,$FA,$EA,$32,$FF,$AF
	.dc.b	$3F,$2E,$FE,$BA,$DF,$63,$FC,$ED,$8F,$F8,$52,$30,$7F,$B0,$B3,$FF
	.dc.b	$18,$EF,$FE,$89,$6B,$FE,$8E,$6B,$FD,$E9,$AF,$C8,$D7,$E1,$FC,$47
	.dc.b	$C4,$6D,$79,$7D,$1F,$F7,$1F,$C7,$7D,$F8,$FF,$01,$E9,$FD,$05,$FF
	.dc.b	$69,$C6,$70,$FF,$0D,$69,$FE,$2C,$9F,$F0,$4B,$63,$3E,$AF,$FB,$BA
	.dc.b	$31,$0C,$FE,$0B,$EB,$D7,$F3,$31,$88,$F6,$46,$3C,$B6,$1F,$E2,$9F
	.dc.b	$A5,$FF,$42,$57,$E1,$3C,$DC,$A8,$CB,$DE,$B3,$18,$C1,$58,$F0,$7D
	.dc.b	$FE,$0D,$C3,$F3,$4D,$E9,$FF,$77,$7C,$0E,$65,$C1,$FB,$8C,$BB,$FD
	.dc.b	$FB,$E9,$34,$DD,$BD,$19,$FF,$8B,$B8,$BF,$D5,$57,$C2,$F7,$EB,$3E
	.dc.b	$8B,$FD,$33,$8F,$F6,$C6,$C6,$73,$FF,$65,$4E,$5B,$EB,$FF,$B3,$1B
	.dc.b	$E4,$72,$8C,$47,$FB,$52,$51,$88,$FF,$20,$C3,$CC,$FF,$21,$E5,$EC
	.dc.b	$6E,$9A,$6D,$7F,$96,$F4,$7B,$78,$8F,$8B,$F7,$57,$80,$D3,$8F,$13
	.dc.b	$EE,$7E,$9D,$7C,$63,$FC,$CD,$A3,$D9,$E8,$F6,$8D,$07,$0F,$E0,$74
	.dc.b	$F5,$CF,$0F,$85,$FF,$61,$CE,$DB,$5F,$E6,$D8,$F9,$F1,$AD,$1F,$01
	.dc.b	$FE,$64,$F1,$FF,$CE,$B1,$8A,$FD,$1C,$E7,$9B,$EE,$79,$36,$7F,$91
	.dc.b	$74,$55,$F2,$BF,$C1,$9F,$F4,$16,$3F,$E8,$08,$C4,$7F,$53,$9E,$19
	.dc.b	$BD,$4B,$94,$AE,$39,$7D,$89,$DF,$DF,$FB,$09,$3F,$8A,$7E,$0E,$3F
	.dc.b	$E4,$38,$E2,$7F,$A4,$DC,$3F,$CE,$EB,$FD,$07,$18,$CB,$F5,$BD,$0F
	.dc.b	$D5,$9E,$D7,$B7,$EE,$3F,$D1,$38,$F1,$3F,$D3,$CE,$2F,$FD,$3A,$DC
	.dc.b	$7F,$D1,$EF,$E6,$E7,$80,$70,$F4,$6B,$C3,$B7,$ED,$C6,$21,$CC,$FD
	.dc.b	$1D,$E3,$7D,$38,$C3,$63,$33,$FE,$AA,$EC,$F9,$FF,$67,$23,$E2,$9F
	.dc.b	$8D,$7E,$E7,$A0,$FC,$0D,$F6,$78,$6D,$7F,$88,$9A,$FF,$88,$9A,$FF
	.dc.b	$82,$1A,$FF,$B6,$3F,$14,$5F,$CB,$F6,$A4,$FD,$6E,$35,$6A,$3F,$D8
	.dc.b	$3F,$17,$F8,$F9,$CD,$DA,$D5,$7D,$3F,$67,$FC,$55,$7F,$93,$FC,$0F
	.dc.b	$FA,$DC,$FB,$9A,$FD,$26,$F2,$7F,$6C,$E4,$FE,$71,$3F,$E1,$96,$BF
	.dc.b	$E3,$E6,$BF,$E2,$46,$BF,$E4,$66,$BF,$EE,$19,$7D,$2F,$8C,$91,$FE
	.dc.b	$25,$FD,$9B,$F9,$53,$7C,$BF,$C6,$BF,$E5,$08,$CB,$FE,$4E,$6B,$FE
	.dc.b	$34,$6B,$FE,$EF,$D5,$F8,$A3,$C1,$FC,$C7,$C9,$4A,$F9,$D3,$97,$FD
	.dc.b	$DB,$E6,$7C,$D8,$C4,$7E,$76,$F9,$DF,$EF,$F7,$FF,$2B,$5F,$EF,$FF
	.dc.b	$05,$7D,$B5,$26,$1E,$DF,$B9,$FC,$5E,$87,$DB,$0E,$84,$E6,$3F,$CD
	.dc.b	$AD,$7E,$04,$FF,$1C,$FC,$15,$EC,$FF,$0F,$6C,$FE,$69,$AF,$EB,$BF
	.dc.b	$07,$7D,$67,$F1,$56,$C7,$E8,$1A,$EE,$3F,$8D,$7E,$29,$FC,$0F,$99
	.dc.b	$33,$F0,$7F,$A7,$D1,$FD,$CA,$11,$89,$DC,$63,$9B,$F9,$27,$E1,$6F
	.dc.b	$91,$A4,$9F,$D8,$C6,$5F,$D9,$FC,$7F,$F1,$07,$F6,$B8,$7F,$40,$C7
	.dc.b	$C0,$DC,$7B,$B6,$7C,$DE,$97,$F3,$38,$FE,$99,$1F,$41,$AF,$E3,$9A
	.dc.b	$FF,$1C,$6B,$EA,$F8,$D7,$C6,$7F,$3D,$6C,$7F,$08,$6B,$FC,$73,$17
	.dc.b	$C7,$F8,$78,$E4,$FA,$9F,$E8,$DF,$8A,$3E,$42,$92,$7F,$89,$C6,$5B
	.dc.b	$EE,$7D,$4F,$7F,$FA,$77,$BE,$FA,$BA,$66,$BF,$8D,$F2,$FF,$72,$6D
	.dc.b	$65,$FE,$63,$19,$7F,$99,$D7,$E0,$B8,$BE,$E3,$78,$BE,$DB,$67,$F0
	.dc.b	$6B,$DA,$F1,$9F,$DC,$1C,$47,$E7,$9A,$FE,$F1,$AF,$F4,$AF,$57,$37
	.dc.b	$FD,$96,$BF,$B5,$46,$2F,$FD,$23,$F9,$7F,$F1,$EA,$FC,$87,$E1,$5F
	.dc.b	$B0,$C3,$F9,$CB,$5F,$E9,$AD,$8F,$CB,$7E,$27,$F8,$59,$34,$9B,$7E
	.dc.b	$F7,$E4,$5A,$7F,$7B,$C1,$F1,$74,$98,$7F,$48,$F9,$4F,$EB,$CF,$E9
	.dc.b	$9A,$FF,$B4,$4F,$F0,$23,$AF,$1F,$FD,$BD,$F8,$27,$F8,$1F,$E1,$0D
	.dc.b	$25,$7F,$54,$8F,$1B,$FD,$5B,$EF,$BF,$33,$B7,$EC,$67,$33,$FD,$B1
	.dc.b	$B1,$FB,$59,$DF,$07,$F2,$5F,$C0,$FF,$11,$7C,$FB,$FF,$C3,$15,$A8
	.dc.b	$79,$AA,$DC,$7F,$78,$F8,$AF,$E8,$78,$F8,$7E,$6F,$D4,$7F,$88,$63
	.dc.b	$39,$7F,$08,$D7,$FD,$F9,$E2,$E7,$E3,$26,$3F,$E1,$49,$CB,$FD,$CB
	.dc.b	$F0,$DF,$E1,$4F,$F1,$0F,$AE,$FF,$2D,$E7,$FC,$3D,$B8,$FF,$0E,$AD
	.dc.b	$DB,$9B,$F8,$A5,$FF,$15,$9E,$78,$1D,$CE,$AF,$F8,$BE,$1F,$D8,$9B
	.dc.b	$F2,$BF,$E3,$B8,$C5,$F8,$FE,$57,$07,$F9,$3A,$3F,$C7,$71,$FE,$24
	.dc.b	$6B,$D2,$3F,$9F,$F0,$7C,$BA,$FE,$F5,$C1,$D2,$31,$5F,$E3,$6F,$82
	.dc.b	$7D,$47,$BA,$1B,$3E,$C3,$F1,$BF,$9D,$7D,$79,$3F,$66,$75,$1F,$E5
	.dc.b	$3F,$C3,$1F,$03,$FE,$64,$6C,$7F,$52,$6B,$FD,$C1,$AF,$F8,$FB,$9F
	.dc.b	$FC,$D0,$FF,$8E,$1B,$C7,$F8,$DF,$90,$FE,$C9,$C2,$F9,$5F,$EC,$DE
	.dc.b	$03,$DE,$FF,$0C,$7E,$1C,$29,$A7,$FC,$39,$F8,$83,$F0,$7F,$8E,$DD
	.dc.b	$3F,$C0,$7C,$88,$C4,$7F,$94,$A3,$13,$F3,$BE,$CB,$F3,$50,$69,$F1
	.dc.b	$7E,$51,$E6,$FF,$9E,$7E,$FF,$FC,$0F,$F6,$1F,$E7,$9E,$6F,$F3,$96
	.dc.b	$5F,$E4,$46,$BF,$E5,$F9,$DB,$FE,$48,$FC,$37,$FD,$1E,$9A,$47,$FD
	.dc.b	$27,$C9,$CB,$93,$F6,$9F,$2D,$FB,$03,$C5,$E8,$3E,$93,$FC,$1D,$D9
	.dc.b	$F9,$5C,$3F,$67,$F4,$7E,$87,$FC,$8C,$D7,$FC,$81,$1B,$E1,$7E,$82
	.dc.b	$FA,$2F,$C8,$B6,$BE,$5B,$5F,$94,$D7,$FB,$CC,$FF,$B5,$7E,$43,$FC
	.dc.b	$A1,$F2,$FF,$CC,$7C,$6F,$F1,$73,$6B,$F2,$BF,$92,$FF,$9F,$77,$FF
	.dc.b	$7A,$F8,$5F,$ED,$DF,$8D,$F9,$98,$CB,$FE,$01,$87,$F8,$BA,$FF,$C1
	.dc.b	$5F,$87,$BF,$07,$FF,$5C,$B6,$BF,$C4,$ED,$7F,$C8,$AD,$7F,$2D,$F0
	.dc.b	$5F,$C6,$88,$FF,$0D,$FC,$97,$FC,$09,$9F,$F8,$BF,$EA,$7F,$01,$F8
	.dc.b	$9B,$F1,$0A,$7F,$8F,$7E,$41,$3F,$85,$DB,$FC,$4B,$8F,$F8,$73,$B3
	.dc.b	$FC,$F3,$F1,$9F,$E3,$A6,$C7,$F9,$61,$AF,$FA,$99,$FE,$21,$D3,$F8
	.dc.b	$FF,$4F,$EF,$4F,$FA,$D2,$72,$FF,$AD,$B8,$BF,$C8,$DF,$47,$FE,$CB
	.dc.b	$FC,$43,$E7,$49,$F9,$1C,$FF,$E2,$E6,$C7,$F8,$D9,$D9,$FE,$49,$7F
	.dc.b	$D6,$ED,$78,$7C,$47,$FB,$9A,$31,$1F,$E3,$EF,$C2,$9F,$5B,$FE,$DB
	.dc.b	$6C,$7F,$A9,$9A,$FF,$B0,$71,$7E,$45,$8E,$9F,$F6,$0F,$B9,$FE,$C7
	.dc.b	$25,$7B,$FE,$4B,$FC,$15,$B1,$F9,$B1,$FE,$4E,$8C,$70,$FC,$D9,$CC
	.dc.b	$A7,$2F,$CF,$FB,$8F,$E9,$EF,$FB,$4B,$C9,$FF,$6B,$1F,$BD,$8C,$E7
	.dc.b	$FE,$00,$F9,$DC,$7B,$D2,$FE,$87,$C4,$27,$F8,$27,$93,$FD,$86,$7F
	.dc.b	$D8,$BF,$41,$FD,$28,$FE,$AB,$9F,$FB,$DF,$DD,$FF,$2F,$F0,$39,$BF
	.dc.b	$CC,$FA,$3C,$1F,$E7,$57,$BC,$7F,$01,$97,$57,$E5,$4F,$F3,$F5,$F8
	.dc.b	$8F,$F8,$F7,$1F,$F3,$8F,$E2,$3F,$E1,$D8,$E1,$1F,$E7,$16,$E5,$FE
	.dc.b	$70,$78,$52,$B8,$7C,$7A,$F9,$FC,$DE,$F5,$FF,$3B,$D7,$56,$C7,$3A
	.dc.b	$E1,$7C,$E3,$17,$FE,$84,$F5,$B8,$4D,$9E,$8D,$3F,$D1,$38,$F9,$27
	.dc.b	$C7,$5E,$29,$CE,$1E,$06,$99,$3E,$1E,$1E,$2C,$78,$C7,$8C,$D9,$4F
	.dc.b	$93,$3C,$DE,$35,$CF,$0F,$8B,$1F,$E9,$63,$27,$FB,$41,$CB,$EE,$31
	.dc.b	$F0,$5F,$EB,$76,$F8,$7F,$EB,$A9,$4E,$67,$A4,$62,$7F,$A2,$FD,$74
	.dc.b	$9F,$F6,$08,$70,$FF,$B1,$1F,$80,$B8,$7F,$18,$FB,$4F,$F5,$9C,$6A
	.dc.b	$33,$97,$F9,$86,$72,$6E,$33,$AF,$FB,$B2,$B8,$6D,$DA,$BC,$47,$F7
	.dc.b	$19,$F8,$95,$E0,$DB,$A4,$E9,$ED,$19,$9E,$93,$DE,$BA,$5F,$68,$F4
	.dc.b	$1D,$B0,$E3,$3A,$8E,$07,$DB,$BF,$F0,$57,$CA,$7C,$0D,$7E,$0C,$62
	.dc.b	$BF,$93,$7B,$CF,$C2,$35,$1F,$B0,$8F,$F5,$8B,$FC,$AF,$D5,$FD,$93
	.dc.b	$78,$F7,$97,$CD,$AD,$BF,$C6,$EB,$6F,$F1,$DF,$59,$F0,$3C,$3E,$17
	.dc.b	$F9,$97,$CE,$72,$D1,$5A,$CF,$FC,$75,$E3,$7F,$BD,$9B,$5F,$EE,$7C
	.dc.b	$7F,$DD,$78,$72,$9F,$C9,$42,$30,$FE,$55,$AF,$FB,$B1,$FE,$AB,$BF
	.dc.b	$38,$E1,$E6,$7A,$9D,$35,$EE,$D7,$F1,$9F,$9A,$BC,$5F,$53,$70,$FF
	.dc.b	$29,$57,$C4,$E2,$FA,$33,$9B,$FE,$81,$F5,$1F,$D1,$AF,$FA,$C2,$E6
	.dc.b	$F8,$8E,$67,$71,$F2,$7F,$6A,$79,$9E,$9D,$51,$8C,$3E,$1C,$62,$75
	.dc.b	$39,$BF,$CA,$FE,$3D,$FE,$4D,$BF,$ED,$0D,$8F,$EA,$8F,$A6,$3F,$99
	.dc.b	$B9,$CF,$F5,$78,$7F,$A7,$71,$FB,$5F,$88,$BE,$AB,$EF,$78,$1E,$BC
	.dc.b	$7F,$BA,$4E,$5F,$F3,$B4,$3E,$B7,$D4,$D8,$F4,$B7,$3F,$B4,$D7,$ED
	.dc.b	$7E,$6F,$FE,$67,$9F,$F4,$03,$63,$F9,$B4,$7F,$AA,$72,$F5,$47,$E1
	.dc.b	$67,$59,$FA,$BF,$21,$7C,$97,$E7,$3D,$9F,$E1,$9F,$0D,$F9,$79,$CE
	.dc.b	$5F,$EA,$FF,$0B,$FC,$C2,$FF,$A3,$FF,$7A,$F9,$19,$FE,$35,$AF,$E5
	.dc.b	$9A,$FE,$55,$BA,$3C,$8F,$99,$5F,$9C,$6D,$7F,$A6,$3C,$5F,$E4,$B3
	.dc.b	$FE,$9D,$9C,$E6,$FC,$3F,$FB,$84,$FE,$8A,$31,$E6,$FE,$75,$B1,$99
	.dc.b	$DE,$1F,$A5,$FC,$AB,$FD,$D3,$3F,$88,$6C,$67,$E6,$7F,$1D,$18,$81
	.dc.b	$FE,$D5,$FC,$19,$F3,$DD,$DB,$7E,$36,$CF,$F4,$7C,$51,$88,$E6,$7F
	.dc.b	$AD,$BE,$FF,$DE,$ED,$AE,$A5,$18,$8F,$F7,$33,$5F,$E0,$5F,$9F,$BE
	.dc.b	$67,$F4,$CD,$C3,$F4,$AD,$9F,$E0,$CD,$7B,$7C,$85,$7F,$DE,$4D,$7F
	.dc.b	$D9,$6D,$C0,$E7,$85,$FB,$FF,$FA,$27,$0F,$D9,$B6,$FF,$8A,$35,$FE
	.dc.b	$28,$D7,$F8,$EF,$EF,$6F,$8A,$FD,$F3,$6B,$F7,$AD,$9F,$E3,$CD,$7F
	.dc.b	$8F,$6D,$5F,$E4,$ED,$7F,$93,$37,$0F,$B3,$7F,$C2,$FF,$36,$7C,$1D
	.dc.b	$26,$3F,$C3,$23,$2F,$F0,$EF,$DD,$FF,$9B,$63,$FC,$BD,$AF,$F2,$F6
	.dc.b	$BF,$CE,$1A,$FF,$37,$FC,$91,$F6,$BF,$C6,$DB,$5F,$C6,$9B,$3F,$CE
	.dc.b	$DA,$FE,$0F,$ED,$1F,$05,$FD,$52,$1F,$80,$FF,$00,$E2,$56,$B0,$FC
	.dc.b	$CF,$85,$FE,$04,$BD,$FA,$1C,$7F,$01,$7C,$9F,$F4,$59,$46,$2F,$FA
	.dc.b	$7C,$62,$75,$39,$8F,$EA,$79,$7F,$6D,$AE,$E7,$E7,$9C,$47,$F1,$63
	.dc.b	$F5,$73,$C7,$0F,$24,$E9,$FF,$61,$E2,$F3,$FF,$8B,$3F,$C6,$B0,$FD
	.dc.b	$9E,$7E,$E9,$CF,$B3,$27,$C3,$92,$BE,$1F,$A7,$98,$F1,$46,$27,$6F
	.dc.b	$C8,$CB,$B9,$E0,$FB,$3B,$3A,$C9,$F5,$3E,$43,$38,$74,$8C,$CF,$8B
	.dc.b	$67,$EA,$0C,$3F,$C1,$7F,$08,$FD,$43,$F3,$9C,$1F,$E3,$8B,$DD,$6A
	.dc.b	$73,$87,$FA,$7D,$FC,$E5,$F1,$E9,$FE,$F8,$D9,$FF,$66,$4F,$E8,$6F
	.dc.b	$C3,$C1,$FE,$1A,$6C,$FF,$B2,$E7,$AE,$9F,$82,$CF,$E7,$3F,$27,$4F
	.dc.b	$CF,$E9,$F5,$3E,$4B,$E9,$FB,$5F,$51,$98,$FC,$5F,$C4,$AE,$1F,$0D
	.dc.b	$FB,$3F,$05,$27,$E5,$7C,$BD,$AF,$91,$D1,$F7,$8C,$65,$F8,$27,$D1
	.dc.b	$F5,$9F,$CB,$23,$F3,$5C,$DF,$EA,$9C,$3E,$74,$E5,$FE,$50,$7E,$96
	.dc.b	$3F,$AD,$DF,$F4,$7F,$4F,$F3,$D1,$88,$FE,$6D,$18,$F2,$F8,$1C,$5B
	.dc.b	$5F,$81,$F5,$BF,$26,$7E,$53,$E8,$AC,$7E,$17,$DA,$FC,$B9,$8C,$3F
	.dc.b	$08,$FD,$1A,$FF,$02,$47,$E8,$F9,$7F,$C0,$9F,$2B,$9F,$6F,$4C,$C3
	.dc.b	$F1,$56,$FB,$2C,$61,$C6,$31,$3F,$95,$F9,$5E,$DF,$80,$FE,$A1,$B1
	.dc.b	$FA,$4E,$CF,$F6,$3F,$C7,$FE,$23,$E3,$4D,$24,$FF,$93,$B5,$FF,$0C
	.dc.b	$77,$FA,$4F,$CC,$D7,$09,$F2,$C7,$ED,$BE,$A7,$F6,$FF,$41,$67,$F5
	.dc.b	$0E,$1F,$D5,$7C,$9D,$DB,$F5,$0F,$EF,$BE,$2F,$FC,$29,$FB,$C3,$ED
	.dc.b	$B1,$7F,$B6,$8C,$36,$3F,$82,$DF,$06,$C7,$F0,$2F,$49,$E8,$32,$FF
	.dc.b	$25,$36,$7F,$9E,$46,$0F,$F0,$56,$1F,$E4,$9E,$DF,$F5,$5F,$62,$31
	.dc.b	$5F,$CF,$6B,$CF,$F1,$3E,$8F,$8F,$FF,$40,$7E,$BD,$FA,$CC,$38,$46
	.dc.b	$38,$66,$1F,$E5,$8A,$FE,$A8,$48,$FE,$65,$F8,$4F,$F7,$1D,$CD,$27
	.dc.b	$B9,$FC,$DF,$F2,$9F,$F4,$54,$FF,$2D,$0E,$36,$FE,$81,$5F,$E0,$17
	.dc.b	$FA,$14,$67,$1F,$F4,$6E,$3F,$E9,$AB,$FF,$44,$35,$CF,$C9,$FF,$9C
	.dc.b	$63,$11,$FC,$B7,$77,$CD,$FD,$83,$E0,$D5,$BF,$B1,$F1,$5F,$70,$FB
	.dc.b	$BF,$84,$58,$FA,$D8,$BE,$67,$E4,$99,$CB,$FA,$47,$2A,$75,$5F,$77
	.dc.b	$CA,$FF,$3F,$57,$FA,$7D,$AF,$FA,$A1,$AF,$FA,$9D,$BC,$7F,$6F,$DB
	.dc.b	$FC,$79,$FE,$B7,$8F,$BB,$3E,$D8,$FC,$9F,$D3,$7E,$53,$BE,$E1,$F9
	.dc.b	$37,$F2,$FF,$87,$7E,$20,$91,$FE,$55,$E1,$47,$E3,$39,$7F,$3A,$70
	.dc.b	$9F,$CE,$3F,$CD,$7F,$07,$FF,$7D,$47,$F4,$97,$F2,$D9,$7F,$07,$72
	.dc.b	$DE,$AE,$1D,$DF,$EB,$DF,$05,$48,$C3,$6F,$F7,$8D,$BF,$F2,$B6,$5F
	.dc.b	$E7,$59,$FD,$77,$C5,$FE,$CB,$E7,$3F,$66,$FE,$D3,$E5,$E4,$FE,$CF
	.dc.b	$E4,$EE,$1F,$DB,$30,$7E,$21,$FB,$AF,$AF,$FD,$DB,$6F,$D9,$E7,$F9
	.dc.b	$08,$F9,$1F,$D8,$FE,$23,$FC,$08,$FE,$07,$F8,$5D,$9F,$E0,$5F,$26
	.dc.b	$7F,$CE,$A7,$F9,$9F,$14,$67,$2F,$E8,$F1,$89,$DC,$7F,$82,$A3,$0E
	.dc.b	$B8,$78,$4F,$36,$C6,$DA,$FE,$85,$AF,$F9,$93,$EC,$9E,$57,$F8,$A3
	.dc.b	$E3,$91,$89,$FF,$7C,$C6,$27,$53,$99,$FF,$42,$9F,$E6,$EF,$7B,$FC
	.dc.b	$6A,$F0,$6B,$AA,$FF,$1F,$DF,$79,$CC,$FF,$89,$BF,$7A,$FF,$2E,$BF
	.dc.b	$00,$FF,$12,$38,$7F,$B5,$B7,$1F,$F4,$83,$63,$97,$D4,$FF,$A6,$27
	.dc.b	$2D,$8C,$9D,$7B,$D4,$9D,$4F,$83,$C5,$FF,$50,$F9,$1F,$CB,$29,$18
	.dc.b	$C3,$D0,$DA,$F3,$F9,$57,$37,$EF,$3F,$C4,$27,$FA,$B3,$5F,$93,$1A
	.dc.b	$F2,$7F,$CB,$2D,$8F,$F6,$46,$3F,$C3,$DF,$F2,$0A,$F5,$3C,$1A,$FA
	.dc.b	$9B,$1B,$9F,$F2,$4A,$FA,$2C,$92,$39,$FE,$1B,$C3,$EE,$E4,$76,$79
	.dc.b	$FD,$DF,$AF,$47,$A5,$E6,$6D,$7B,$23,$07,$AB,$DC,$FC,$2F,$17,$F9
	.dc.b	$1E,$BF,$B3,$57,$F9,$3D,$AF,$F8,$36,$77,$C5,$F8,$3F,$93,$F5,$92
	.dc.b	$7E,$0F,$B9,$EE,$3F,$CF,$FC,$1F,$69,$BB,$FF,$84,$67,$E4,$C6,$3F
	.dc.b	$0F,$F8,$BE,$F4,$9F,$97,$1F,$23,$29,$1F,$39,$FB,$9E,$8B,$D0,$7A
	.dc.b	$F8,$6D,$7E,$88,$FE,$71,$1F,$E9,$29,$E0,$D9,$DC,$FE,$A2,$7F,$A3
	.dc.b	$70,$F6,$F4,$8D,$3E,$DF,$64,$AF,$C0,$B5,$FB,$BF,$39,$F8,$0E,$F9
	.dc.b	$1F,$86,$7A,$70,$C8,$FE,$F4,$F9,$70,$34,$FF,$54,$C3,$F5,$D7,$89
	.dc.b	$F9,$D8,$C3,$63,$31,$FE,$9A,$1F,$85,$FE,$11,$FE,$FC,$D7,$F8,$3B
	.dc.b	$51,$1C,$37,$A3,$CD,$C5,$23,$FC,$24,$D7,$F8,$8F,$2F,$EE,$BE,$13
	.dc.b	$E5,$4C,$9F,$D3,$C7,$E9,$3E,$12,$47,$EA,$9F,$B5,$EA,$C8,$FB,$5C
	.dc.b	$5F,$7E,$3E,$C1,$23,$EF,$60,$FB,$DF,$F6,$EC,$7E,$D3,$AB,$F6,$27
	.dc.b	$ED,$23,$31,$FE,$BA,$7F,$D8,$CF,$FA,$E9,$CF,$67,$F8,$E1,$B3,$FE
	.dc.b	$C8,$77,$3C,$7E,$3B,$A7,$C9,$48,$E5,$1D,$FE,$43,$C0,$C8,$ED,$1F
	.dc.b	$C4,$FE,$9B,$FD,$DF,$1F,$EB,$F6,$BF,$E7,$D6,$E3,$FD,$4F,$E1,$9F
	.dc.b	$B4,$7E,$DA,$30,$FE,$E2,$BE,$C8,$9D,$FC,$2F,$FB,$C9,$B1,$FA,$56
	.dc.b	$BF,$63,$37,$CA,$79,$7A,$E4,$FC,$58,$F4,$78,$FE,$9E,$E9,$3E,$98
	.dc.b	$F6,$78,$FA,$8C,$1E,$D7,$E9,$7C,$A4,$8F,$A0,$FD,$5F,$90,$91,$F4
	.dc.b	$DA,$FD,$7F,$92,$FD,$07,$E7,$EC,$E9,$27,$F4,$10,$F1,$B6,$7E,$A9
	.dc.b	$C4,$7F,$05,$0E,$1F,$E1,$0D,$7E,$5F,$9A,$FA,$EF,$F0,$AC,$76,$F2
	.dc.b	$7C,$67,$F4,$D8,$CC,$7E,$37,$E5,$3F,$20,$47,$F1,$DF,$29,$F9,$32
	.dc.b	$38,$3F,$25,$E5,$7E,$5F,$E4,$BF,$31,$F2,$92,$7F,$2D,$1F,$9B,$F9
	.dc.b	$3D,$46,$0F,$CD,$3F,$D7,$FF,$4F,$F1,$F5,$FF,$86,$F5,$46,$3C,$0F
	.dc.b	$DF,$C6,$3D,$FF,$ED,$1D,$6F,$7D,$CF,$86,$E7,$F8,$96,$BF,$E1,$09
	.dc.b	$E1,$3F,$8F,$9C,$CF,$CA,$C5,$F0,$1C,$EA,$4F,$F1,$0C,$FC,$D5,$A8
	.dc.b	$FD,$A5,$FE,$E4,$8F,$ED,$FE,$4B,$F7,$84,$70,$7E,$EF,$E5,$3F,$80
	.dc.b	$7C,$4F,$F0,$1F,$89,$93,$FB,$F8,$FE,$09,$F0,$3A,$8C,$1F,$C0,$DF
	.dc.b	$E5,$DE,$E3,$E2,$29,$27,$F9,$77,$E5,$8F,$27,$F9,$A3,$63,$F1,$CD
	.dc.b	$7F,$AA,$35,$FF,$1B,$B5,$FE,$1C,$7F,$3D,$FB,$D7,$BF,$4F,$E3,$51
	.dc.b	$FD,$8B,$89,$AF,$F5,$D6,$C7,$F2,$86,$BF,$D7,$5A,$FF,$A7,$FE,$1F
	.dc.b	$FB,$11,$1F,$E5,$FD,$68,$FE,$D1,$F9,$ED,$9F,$E6,$6D,$7F,$B4,$35
	.dc.b	$FE,$23,$E9,$23,$11,$FD,$32,$1F,$9E,$E2,$11,$8A,$FE,$ED,$39,$8C
	.dc.b	$4F,$F4,$2F,$A3,$FE,$3C,$F3,$6B,$C9,$B0,$6F,$EA,$4E,$5F,$F0,$13
	.dc.b	$6B,$FA,$6B,$5F,$F0,$13,$5F,$E5,$7E,$A3,$F3,$E7,$B3,$FE,$1E,$6C
	.dc.b	$FF,$52,$6B,$FD,$89,$AF,$C4,$FD,$2F,$F9,$6B,$FD,$71,$AF,$F8,$45
	.dc.b	$AE,$FD,$1F,$F1,$5F,$D4,$FF,$98,$E3,$FA,$FB,$5F,$ED,$ED,$7C,$ED
	.dc.b	$7F,$BD,$E0,$FC,$2B,$FB,$0C,$7B,$1A,$E9,$AA,$38,$7D,$27,$F8,$A7
	.dc.b	$F7,$8F,$F4,$2C,$FD,$B6,$BF,$69,$BB,$9C,$D3,$F0,$1F,$F0,$DB,$6B
	.dc.b	$FC,$8F,$18,$3F,$C3,$F4,$FD,$0B,$FC,$63,$3F,$C1,$7C,$8F,$C5,$46
	.dc.b	$23,$FD,$FA,$DF,$13,$FD,$71,$87,$FA,$AB,$E3,$FF,$C8,$FE,$5B,$E5
	.dc.b	$BF,$C8,$4D,$C3,$FC,$1A,$D7,$FC,$84,$D7,$F3,$7F,$9D,$FF,$27,$7E
	.dc.b	$27,$E0,$5F,$E5,$B5,$8F,$82,$30,$FF,$B2,$74,$57,$15,$F9,$0F,$F7
	.dc.b	$9D,$7F,$48,$DC,$3F,$46,$DD,$1F,$7B,$FC,$8E,$7B,$D7,$EA,$FF,$7C
	.dc.b	$7C,$27,$F9,$75,$B3,$FE,$28,$6B,$FE,$5D,$6B,$FB,$65,$F0,$9F,$EA
	.dc.b	$7F,$29,$F5,$FF,$E6,$4D,$BF,$DC,$51,$8F,$91,$FF,$38,$DF,$0C,$3E
	.dc.b	$A1,$FE,$D8,$FC,$13,$FC,$4B,$7C,$1B,$9F,$F9,$33,$04,$62,$B8,$74
	.dc.b	$7F,$9F,$3D,$D7,$C5,$7F,$9D,$DB,$3F,$E4,$26,$BF,$E7,$76,$BF,$E2
	.dc.b	$0B,$7E,$04,$F4,$BD,$8F,$EF,$73,$F5,$B7,$3F,$23,$BB,$F5,$DB,$F3
	.dc.b	$AF,$E8,$11,$86,$BF,$48,$FF,$7B,$B8,$BF,$C0,$7D,$5F,$F0,$26,$F8
	.dc.b	$9F,$E9,$18,$FE,$07,$5F,$C8,$DA,$FF,$22,$6F,$3F,$F8,$77,$E0,$BF
	.dc.b	$D3,$07,$F8,$7E,$7F,$CC,$6D,$7F,$D3,$0D,$7E,$17,$C0,$7F,$92,$34
	.dc.b	$FE,$61,$3B,$C7,$F2,$8D,$8F,$A9,$3B,$A6,$6F,$C7,$DD,$BF,$85,$6C
	.dc.b	$FC,$66,$BF,$0E,$5F,$A5,$7F,$15,$CB,$F1,$2D,$C7,$F0,$4D,$AE,$ED
	.dc.b	$AE,$3C,$FF,$E9,$DD,$7F,$71,$87,$FA,$53,$F0,$3F,$F0,$A6,$1F,$E9
	.dc.b	$69,$7D,$17,$F9,$82,$73,$29,$CB,$C7,$DB,$FF,$32,$C7,$48,$4F,$08
	.dc.b	$C3,$6B,$A5,$70,$6D,$78,$7C,$2F,$E0,$2F,$29,$EB,$18,$FA,$2E,$53
	.dc.b	$FE,$AC,$6A,$7C,$11,$B9,$E3,$8F,$9E,$BD,$D7,$E0,$7D,$8D,$C7,$CF
	.dc.b	$5C,$27,$C4,$70,$39,$1C,$0B,$1C,$1F,$55,$7F,$98,$E3,$E2,$47,$FA
	.dc.b	$28,$F9,$57,$EE,$71,$1A,$7E,$5F,$B5,$FC,$59,$FF,$78,$47,$F8,$9F
	.dc.b	$C0,$E9,$39,$9E,$2D,$7F,$DE,$DF,$79,$FE,$DE,$4D,$23,$FE,$E2,$7E
	.dc.b	$56,$3E,$5D,$9C,$18,$AF,$45,$57,$F8,$84,$57,$FD,$BB,$16,$78,$C7
	.dc.b	$D3,$AF,$1B,$4F,$E1,$95,$C5,$F2,$99,$3A,$94,$EA,$59,$EB,$AF,$E5
	.dc.b	$23,$2D,$8C,$D7,$82,$31,$5D,$EF,$CC,$DB,$E8,$FB,$EF,$F9,$CF,$AB
	.dc.b	$EA,$9F,$E4,$5E,$0A,$31,$1F,$CC,$63,$15,$F7,$63,$F7,$F1,$87,$E6
	.dc.b	$78,$2F,$A0,$E3,$1F,$C1,$23,$15,$C7,$04,$62,$78,$D7,$09,$7E,$06
	.dc.b	$F8,$DC,$37,$0E,$4D,$BF,$B8,$D7,$FC,$97,$F8,$1F,$FB,$52,$7D,$1D
	.dc.b	$88,$C4,$FE,$5E,$31,$E8,$FA,$1B,$5C,$3E,$99,$F1,$59,$6C,$66,$30
	.dc.b	$7A,$7F,$03,$7D,$B5,$24,$FC,$18,$CB,$F0,$E1,$F1,$7F,$05,$B1,$98
	.dc.b	$C1,$E5,$F8,$76,$74,$93,$EE,$8C,$BF,$3F,$F0,$3F,$3C,$FE,$85,$AF
	.dc.b	$F6,$16,$BF,$DB,$DA,$FF,$3C,$FC,$0D,$F2,$1F,$55,$B5,$F6,$DB,$3F
	.dc.b	$AD,$6B,$FE,$7C,$EF,$FE,$87,$3F,$49,$B1,$F4,$5A,$EF,$C4,$FA,$5F
	.dc.b	$81,$FF,$85,$AB,$EC,$B6,$7E,$C3,$5F,$F4,$4B,$5F,$F4,$47,$E0,$6F
	.dc.b	$A4,$FB,$8D,$AF,$C1,$B6,$7F,$04,$D7,$FD,$27,$F8,$1B,$E4,$FF,$08
	.dc.b	$DA,$FC,$5B,$67,$F1,$4D,$7F,$D3,$7E,$27,$CA,$E5,$FF,$25,$72,$7E
	.dc.b	$36,$30,$67,$CC,$7A,$1F,$91,$6E,$3F,$98,$6E,$1F,$97,$6B,$FC,$BF
	.dc.b	$E4,$DF,$4D,$FB,$08,$46,$2B,$F7,$11,$89,$D4,$E6,$3F,$45,$FA,$3F
	.dc.b	$EC,$DF,$F5,$9B,$5F,$F5,$93,$5F,$F5,$8B,$73,$7C,$1F,$F8,$0B,$0F
	.dc.b	$F4,$B7,$81,$FA,$B8,$C5,$FF,$A5,$E3,$C3,$E8,$CA,$F2,$E2,$8C,$57
	.dc.b	$ED,$76,$7D,$8F,$ED,$5B,$7F,$BC,$6C,$EE,$7F,$75,$F1,$0F,$A1,$A4
	.dc.b	$BF,$F6,$23,$F1,$FF,$1F,$FF,$24,$C7,$C6,$6D,$FF,$1F,$6B,$FD,$75
	.dc.b	$AF,$FB,$57,$EA,$1F,$5D,$FC,$25,$B5,$FC,$41,$B3,$FC,$3D,$AF,$FB
	.dc.b	$83,$CB,$FA,$F3,$FE,$EA,$D7,$F8,$94,$62,$7F,$8B,$35,$F1,$74,$FA
	.dc.b	$49,$5E,$97,$F8,$D7,$DD,$3E,$7A,$92,$7F,$91,$C6,$5F,$C2,$FE,$50
	.dc.b	$F9,$DF,$C5,$36,$33,$18,$33,$F2,$92,$38,$B7,$1F,$E8,$D4,$F2,$79
	.dc.b	$92,$BB,$BF,$D1,$BE,$17,$C1,$F4,$97,$13,$49,$5F,$D4,$67,$CD,$F5
	.dc.b	$1B,$24,$66,$91,$D7,$F1,$AE,$73,$C5,$B9,$7F,$85,$ED,$18,$8F,$F7
	.dc.b	$CB,$5F,$F7,$1B,$D2,$79,$57,$52,$3F,$D9,$78,$1F,$30,$FA,$9F,$EC
	.dc.b	$71,$86,$C6,$67,$83,$70,$FF,$15,$7D,$0F,$D3,$F8,$0F,$ED,$6D,$AF
	.dc.b	$EE,$4D,$9D,$E9,$FE,$B2,$9E,$7F,$0F,$D8,$8E,$0E,$FE,$17,$88,$8E
	.dc.b	$0F,$1F,$77,$98,$8E,$0F,$9D,$EC,$7D,$72,$38,$3E,$97,$D8,$FD,$B2
	.dc.b	$38,$3F,$89,$F8,$3F,$D3,$B2,$FF,$44,$F8,$3F,$C6,$63,$04,$EB,$C1
	.dc.b	$2B,$8B,$BF,$A3,$91,$D1,$B7,$F7,$75,$7D,$CF,$E3,$49,$5F,$87,$7F
	.dc.b	$21,$F7,$52,$3F,$14,$DA,$FE,$5D,$F4,$3F,$55,$FE,$E2,$7F,$2E,$70
	.dc.b	$FF,$74,$E5,$46,$23,$5C,$BF,$82,$8F,$E1,$31,$C7,$8B,$04,$C3,$F1
	.dc.b	$52,$FB,$1E,$04,$8C,$BF,$CF,$FA,$7F,$AC,$5F,$91,$AF,$F8,$19,$A9
	.dc.b	$CE,$3E,$6F,$A1,$F2,$12,$FC,$2F,$EA,$78,$E4,$7E,$A5,$B7,$FA,$9E
	.dc.b	$A7,$27,$EB,$C9,$5F,$AF,$7F,$79,$F7,$32,$3F,$6C,$DA,$FA,$DE,$37
	.dc.b	$F0,$12,$4F,$EE,$DF,$E3,$7F,$3F,$23,$F8,$CB,$67,$EF,$7D,$CF,$F0
	.dc.b	$92,$4F,$F0,$37,$F8,$67,$95,$23,$F8,$33,$67,$FC,$85,$E2,$3E,$9F
	.dc.b	$F9,$49,$2B,$F9,$3B,$FD,$B7,$EF,$1E,$35,$24,$FF,$6E,$8C,$B7,$C5
	.dc.b	$E0,$47,$07,$F3,$2E,$DF,$E6,$E4,$70,$7F,$1B,$FB,$AF,$E4,$84,$70
	.dc.b	$7F,$1F,$F8,$AF,$E5,$04,$70,$7F,$98,$3B,$FF,$8D,$68,$F4,$3F,$CF
	.dc.b	$DE,$47,$F4,$E2,$57,$F9,$9B,$F5,$D7,$D6,$FF,$98,$DB,$19,$8C,$1F
	.dc.b	$CD,$FD,$69,$1F,$CC,$5B,$87,$F8,$57,$EB,$5C,$FF,$C8,$9B,$5F,$C9
	.dc.b	$DB,$3B,$9D,$78,$72,$B4,$D9,$FD,$E7,$BB,$FD,$1C,$93,$FD,$05,$ED
	.dc.b	$D5,$23,$B3,$67,$C9,$E6,$F8,$C9,$3E,$37,$FA,$97,$1C,$8F,$E9,$8D
	.dc.b	$9F,$D0,$3F,$D3,$49,$1F,$D3,$5F,$E7,$FF,$71,$EC,$9F,$E0,$FE,$02
	.dc.b	$31,$1F,$C1,$DA,$FF,$8E,$1B,$1B,$F1,$B9,$92,$FF,$A7,$4A,$3F,$B4
	.dc.b	$74,$49,$FE,$C4,$DB,$FD,$B7,$8F,$FE,$D8,$9F,$EB,$AD,$8F,$EB,$AD
	.dc.b	$7F,$83,$78,$1F,$E0,$17,$FD,$8D,$E1,$FF,$81,$67,$4D,$AF,$F4,$7C
	.dc.b	$FF,$0F,$9E,$11,$FC,$4D,$B1,$FC,$A3,$EE,$7F,$CE,$9D,$7F,$E5,$46
	.dc.b	$C7,$FB,$35,$AF,$F2,$E0,$E1,$E5,$E6,$7F,$97,$9E,$0D,$78,$47,$F9
	.dc.b	$93,$5E,$D3,$99,$F3,$7D,$BF,$F3,$AC,$7C,$CD,$8C,$9E,$C3,$F9,$46
	.dc.b	$28,$D4,$7C,$66,$E5,$FE,$0A,$6E,$7B,$9E,$53,$96,$D3,$EA,$FF,$CE
	.dc.b	$B3,$FE,$31,$6C,$66,$7F,$C3,$91,$FE,$66,$F9,$37,$C5,$52,$4F,$F8
	.dc.b	$46,$32,$DF,$AE,$E0,$47,$07,$F8,$8F,$E4,$BC,$39,$1A,$48,$FF,$16
	.dc.b	$3E,$42,$38,$3F,$C4,$1F,$75,$F9,$36,$F1,$FF,$9F,$23,$11,$FE,$97
	.dc.b	$6B,$FA,$47,$F8,$2B,$CB,$AB,$81,$1F,$F1,$E7,$0F,$C4,$23,$83,$C3
	.dc.b	$D1,$DC,$8E,$0F,$2F,$97,$E3,$23,$83,$FC,$99,$F8,$8B,$D6,$FA,$6D
	.dc.b	$8C,$C6,$0F,$F5,$5F,$86,$F0,$9E,$26,$1B,$5F,$CA,$E3,$07,$03,$FB
	.dc.b	$3B,$FC,$D7,$1C,$11,$F1,$FB,$9E,$22,$38,$3C,$FF,$25,$E6,$23,$83
	.dc.b	$FA,$06,$FD,$C8,$E0,$FF,$61,$FE,$3F,$FD,$7D,$BE,$5F,$FB,$19,$46
	.dc.b	$5F,$D9,$18,$8D,$43,$E7,$E9,$27,$AB,$EF,$F4,$64,$7F,$53,$6B,$FE
	.dc.b	$71,$DD,$F4,$FF,$40,$91,$F5,$1F,$A7,$F6,$52,$3E,$BB,$5F,$F5,$2F
	.dc.b	$A3,$F3,$48,$FF,$61,$7E,$DF,$DF,$48,$FB,$ED,$7C,$5F,$2F,$F8,$02
	.dc.b	$3F,$84,$7F,$6B,$F8,$F3,$D7,$A4,$9F,$F4,$A4,$65,$BF,$09,$C0,$8E
	.dc.b	$0F,$F4,$37,$E7,$1F,$EC,$86,$F1,$FF,$B8,$A3,$11,$FE,$FD,$6B,$FC
	.dc.b	$8F,$D0,$FF,$7F,$BF,$EF,$C6,$BF,$59,$AE,$FC,$8E,$9D,$52,$3C,$2D
	.dc.b	$7F,$97,$76,$7F,$AD,$FC,$BF,$F7,$9F,$9B,$FE,$1E,$9F,$F4,$E3,$5D
	.dc.b	$B5,$FE,$BA,$BC,$FE,$24,$8E,$2F,$EB,$3E,$DE,$47,$EC,$1A,$FE,$0F
	.dc.b	$EE,$3F,$62,$47,$F6,$AF,$E5,$73,$FC,$3C,$7C,$1E,$5C,$11,$F8,$9D
	.dc.b	$9B,$23,$83,$E4,$6F,$CC,$8E,$0F,$E0,$7C,$5F,$C0,$88,$E0,$FE,$0F
	.dc.b	$C3,$FC,$18,$8E,$0F,$F6,$17,$E3,$0F,$83,$FD,$4C,$38,$13,$98,$DC
	.dc.b	$EA,$72,$D9,$7C,$AD,$25,$76,$75,$EF,$C8,$F0,$B5,$FF,$65,$FE,$B6
	.dc.b	$F1,$BC,$7F,$0C,$8C,$4F,$FB,$31,$AF,$09,$F1,$67,$23,$DA,$D7,$FD
	.dc.b	$D5,$D3,$FE,$97,$7F,$47,$87,$EA,$08,$FE,$9B,$E3,$FF,$56,$47,$07
	.dc.b	$F3,$28,$FE,$62,$47,$07,$F3,$68,$FE,$6A,$47,$07,$CE,$FC,$67,$F8
	.dc.b	$C3,$FE,$09,$6B,$FE,$B8,$6A,$0E,$1B,$E1,$7F,$23,$8F,$E3,$D1,$D6
	.dc.b	$30,$67,$E0,$E4,$78,$5A,$FF,$86,$7C,$AF,$E8,$BE,$97,$F4,$12,$47
	.dc.b	$F9,$D5,$FE,$61,$FA,$8F,$EE,$23,$FD,$9A,$D7,$CC,$D4,$E6,$5F,$75
	.dc.b	$FD,$E6,$7F,$9B,$C6,$3D,$0F,$EA,$F3,$99,$F2,$35,$FF,$6D,$CB,$EE
	.dc.b	$B1,$1F,$16,$31,$5F,$D8,$27,$33,$FE,$8B,$6B,$E4,$E5,$F5,$91,$FD
	.dc.b	$0E,$1F,$9F,$7E,$27,$DD,$7C,$C6,$E9,$FD,$B2,$31,$1E,$46,$BE,$3F
	.dc.b	$DF,$1E,$EF,$F7,$26,$C7,$C0,$6B,$F4,$5A,$FF,$92,$FE,$89,$F1,$7F
	.dc.b	$D7,$DB,$1F,$B6,$6B,$FD,$81,$AF,$C6,$FD,$23,$FB,$CB,$F7,$5A,$E7
	.dc.b	$D3,$FF,$02,$C6,$23,$E5,$7E,$22,$F3,$7F,$C1,$2D,$8F,$AE,$D7,$F1
	.dc.b	$0D,$7D,$1F,$29,$E7,$23,$E7,$7F,$C2,$7F,$A1,$EE,$3E,$C3,$5F,$F6
	.dc.b	$03,$78,$CD,$FB,$9F,$EB,$13,$F9,$76,$BF,$84,$6B,$F8,$36,$BF,$EC
	.dc.b	$BF,$A0,$FE,$80,$FE,$41,$AF,$E8,$5A,$FE,$39,$AF,$E3,$7E,$E3,$FC
	.dc.b	$CC,$FE,$9D,$AF,$E5,$1A,$EF,$E1,$3F,$C6,$3F,$70,$F8,$9A,$49,$FD
	.dc.b	$0C,$65,$FD,$1F,$DB,$FF,$99,$1F,$DA,$35,$FD,$0B,$5F,$D0,$35,$FC
	.dc.b	$7F,$A7,$F6,$DB,$C3,$FC,$F3,$97,$FC,$90,$D9,$D4,$3E,$DD,$F7,$1B
	.dc.b	$6D,$FF,$8B,$27,$53,$93,$FC,$9B,$F4,$2F,$8B,$A4,$AF,$DE,$C6,$5E
	.dc.b	$5F,$6F,$FC,$72,$CD,$27,$C0,$F9,$FE,$DD,$F2,$34,$97,$E3,$8C,$BF
	.dc.b	$C2,$BE,$2F,$FA,$8B,$7D,$0F,$F3,$0C,$62,$3F,$8F,$35,$FE,$17,$E1
	.dc.b	$BA,$BF,$CC,$ED,$8F,$E2,$AD,$7F,$93,$35,$FE,$68,$7E,$FB,$ED,$F0
	.dc.b	$FF,$37,$D1,$18,$8D,$57,$D3,$8C,$4B,$ED,$DF,$5B,$F5,$DB,$7F,$45
	.dc.b	$AF,$FB,$3E,$78,$4F,$DD,$FB,$77,$B3,$49,$5F,$66,$32,$FF,$7E,$CF
	.dc.b	$E6,$FD,$B7,$F4,$DF,$2D,$F6,$BE,$C9,$FD,$2C,$62,$BF,$A9,$46,$27
	.dc.b	$5C,$2B,$E5,$18,$6D,$70,$F8,$0F,$8F,$87,$38,$61,$CD,$EB,$87,$37
	.dc.b	$FD,$7D,$87,$36,$9C,$CE,$A7,$2B,$EB,$1B,$BE,$D7,$EE,$AE,$26,$CE
	.dc.b	$2F,$85,$F1,$DF,$86,$78,$B6,$BE,$25,$F6,$AF,$11,$D2,$FB,$5F,$0C
	.dc.b	$3E,$76,$1E,$BA,$FC,$39,$DA,$32,$7D,$3F,$9A,$7D,$B5,$24,$FF,$A7
	.dc.b	$30,$4E,$A3,$FD,$45,$F6,$DF,$EA,$34,$D2,$3A,$1F,$6F,$F4,$5F,$EA
	.dc.b	$BD,$5F,$E0,$D8,$C4,$72,$37,$AF,$E3,$E7,$4F,$FA,$BA,$AC,$FE,$8B
	.dc.b	$87,$F7,$D7,$CA,$73,$3F,$A3,$BA,$FF,$86,$E3,$EC,$BE,$23,$EA,$99
	.dc.b	$2D,$7F,$09,$D7,$A4,$79,$0F,$05,$74,$7D,$17,$ED,$AF,$1E,$1E,$79
	.dc.b	$E0,$78,$CA,$77,$9E,$15,$EC,$A7,$DB,$BE,$97,$F8,$D6,$DF,$C6,$E7
	.dc.b	$31,$8A,$FE,$01,$4F,$9E,$C6,$3F,$E0,$28,$E5,$3F,$E3,$BC,$39,$46
	.dc.b	$27,$F5,$53,$FC,$27,$DD,$7C,$1F,$2C,$BB,$CF,$13,$51,$9F,$B7,$FE
	.dc.b	$F1,$19,$B7,$9A,$F4,$7D,$73,$9C,$17,$C2,$FC,$27,$0A,$DB,$8B,$E5
	.dc.b	$97,$F3,$B8,$FF,$19,$CE,$9B,$19,$7F,$C6,$91,$97,$FC,$6D,$F1,$2F
	.dc.b	$94,$D4,$79,$A3,$19,$7F,$91,$1B,$1F,$E5,$56,$BF,$8F,$FB,$77,$DC
	.dc.b	$7C,$56,$D7,$C1,$C9,$7C,$B7,$FB,$15,$AC,$3D,$FA,$BE,$E1,$E8,$FC
	.dc.b	$F6,$E5,$F1,$D7,$D9,$E6,$7E,$EE,$6F,$B8,$79,$1E,$C8,$C3,$63,$38
	.dc.b	$7F,$98,$1B,$97,$F3,$BF,$B8,$F3,$4F,$F4,$97,$F0,$9A,$FE,$3D,$CC
	.dc.b	$FF,$93,$1C,$CF,$F4,$FF,$B8,$7C,$2F,$E1,$DB,$1F,$7D,$7C,$36,$67
	.dc.b	$D3,$E1,$FC,$CC,$3F,$CD,$4D,$7F,$12,$DD,$3F,$DD,$1C,$9F,$63,$E8
	.dc.b	$FF,$1B,$F5,$DF,$9E,$7F,$1F,$16,$53,$63,$E3,$FD,$BB,$E5,$A9,$C7
	.dc.b	$C3,$7D,$46,$37,$FB,$5F,$6F,$C9,$FF,$43,$B7,$5F,$F4,$C3,$5F,$F4
	.dc.b	$2B,$5F,$F4,$27,$93,$F2,$35,$78,$CF,$AB,$FB,$33,$98,$C6,$1F,$A9
	.dc.b	$A7,$E2,$6F,$5B,$E0,$36,$7F,$9D,$E2,$AD,$D6,$8F,$C1,$7C,$17,$F8
	.dc.b	$36,$BF,$CD,$3E,$9E,$08,$F2,$A7,$DF,$3E,$5B,$E7,$36,$74,$DC,$7F
	.dc.b	$A5,$D7,$0A,$FA,$9F,$89,$FA,$A7,$2D,$7F,$C2,$F2,$8C,$47,$FB,$36
	.dc.b	$31,$B7,$E6,$7F,$13,$7D,$05,$26,$5F,$EB,$88,$CB,$E9,$FA,$DB,$87
	.dc.b	$91,$AF,$81,$AF,$76,$BF,$C5,$BE,$E5,$F5,$3F,$A9,$6D,$7F,$B1,$9B
	.dc.b	$3F,$ED,$76,$BF,$EC,$4F,$C4,$FF,$93,$CF,$FB,$31,$B1,$FE,$DD,$6B
	.dc.b	$BF,$3F,$F6,$DF,$32,$F8,$4A,$4B,$FF,$6A,$C6,$5F,$F6,$BF,$6F,$E2
	.dc.b	$9B,$CF,$FB,$F8,$C6,$1F,$EE,$B6,$BF,$ED,$85,$F3,$BB,$25,$74,$7F
	.dc.b	$84,$7E,$27,$DE,$BF,$DE,$8D,$7F,$DC,$2D,$7F,$DC,$0D,$43,$F3,$0F
	.dc.b	$EA,$55,$FD,$0E,$31,$EB,$7E,$DF,$EA,$1F,$09,$FD,$5A,$73,$5B,$FC
	.dc.b	$F3,$A4,$1A,$4D,$BF,$9C,$C7,$7F,$BC,$E5,$91,$E9,$5E,$BF,$37,$FA
	.dc.b	$44,$E7,$0F,$E6,$DA,$3E,$8F,$F9,$5B,$63,$F6,$5A,$7E,$BF,$2F,$EA
	.dc.b	$53,$97,$FA,$BF,$52,$5F,$0F,$ED,$39,$E1,$70,$E7,$E3,$F6,$FA,$9F
	.dc.b	$84,$D8,$F1,$C7,$C5,$F0,$71,$97,$6C,$7F,$74,$FF,$66,$E3,$FF,$33
	.dc.b	$FC,$BF,$F7,$E6,$E1,$E8,$8C,$7C,$0F,$FB,$EA,$3F,$BA,$7D,$1B,$EA
	.dc.b	$1F,$75,$FD,$8A,$30,$D8,$CC,$FF,$0E,$9C,$BF,$C4,$30,$FF,$3B,$E7
	.dc.b	$F6,$EF,$BB,$67,$E6,$F8,$BF,$30,$FC,$03,$4F,$E3,$DF,$4B,$FD,$8F
	.dc.b	$97,$FD,$6D,$C6,$8C,$4F,$E3,$AB,$F9,$F4,$63,$6F,$E4,$0F,$F6,$99
	.dc.b	$FB,$7E,$9E,$BD,$BF,$B6,$C8,$EF,$F2,$16,$FE,$CD,$FF,$B4,$27,$F2
	.dc.b	$DF,$79,$FE,$E7,$6F,$CB,$78,$3E,$13,$F2,$D3,$9B,$FF,$0B,$85,$F6
	.dc.b	$FF,$94,$6D,$FF,$36,$9D,$4E,$4F,$CC,$FC,$0F,$EA,$27,$F4,$CD,$9E
	.dc.b	$B5,$C9,$EF,$19,$DF,$E3,$E4,$FB,$4F,$CC,$F6,$7C,$66,$E3,$F9,$E8
	.dc.b	$F9,$3F,$6B,$8C,$7E,$0B,$FE,$46,$E6,$FC,$27,$15,$E7,$FE,$7D,$9F
	.dc.b	$FB,$A7,$9F,$FD,$6B,$F5,$B9,$AF,$C6,$F6,$4D,$3F,$1B,$87,$07,$F8
	.dc.b	$1D,$FF,$AF,$27,$F2,$DF,$89,$FF,$12,$9F,$F2,$34,$A3,$1A,$7D,$4F
	.dc.b	$C2,$37,$3E,$3F,$93,$FC,$A1,$2F,$F2,$98,$FF,$B4,$1F,$CA,$61,$FC
	.dc.b	$8B,$4F,$43,$63,$F8,$AE,$BF,$A1,$BF,$1F,$95,$FC,$03,$BE,$57,$E5
	.dc.b	$3C,$6B,$AF,$E1,$BC,$9F,$C8,$98,$CF,$E8,$BF,$E3,$2F,$B5,$FC,$A7
	.dc.b	$C0,$1F,$CB,$78,$1F,$9B,$F8,$FD,$7C,$67,$E6,$BC,$E9,$A7,$F9,$1E
	.dc.b	$F7,$D7,$FE,$2E,$C7,$97,$83,$FE,$3B,$9F,$CB,$6B,$FE,$BA,$9C,$DF
	.dc.b	$F8,$09,$B1,$FD,$D2,$FF,$C2,$DA,$FE,$1B,$E0,$BC,$D8,$FE,$A9,$FD
	.dc.b	$5F,$C4,$CC,$7F,$67,$F1,$57,$1F,$F3,$86,$7F,$8D,$9E,$CD,$8F,$13
	.dc.b	$5F,$E8,$27,$FA,$01,$C3,$8F,$BA,$7D,$CF,$F3,$86,$D7,$E9,$5B,$3F
	.dc.b	$E1,$C6,$BF,$ED,$8F,$C4,$7F,$99,$69,$FE,$81,$6D,$75,$BF,$E9,$51
	.dc.b	$8A,$F1,$78,$1F,$24,$95,$EA,$7F,$C9,$5B,$3E,$63,$F8,$69,$F6,$E2
	.dc.b	$CF,$D5,$E6,$FB,$C7,$F1,$2F,$36,$C6,$4F,$E3,$7F,$7F,$8C,$7F,$86
	.dc.b	$BF,$85,$E6,$FB,$44,$9F,$96,$F8,$BE,$BF,$F8,$E9,$B8,$FE,$4A,$FF
	.dc.b	$20,$F9,$09,$1F,$CA,$3E,$AE,$E1,$FC,$6B,$C9,$F5,$92,$7F,$C7,$9B
	.dc.b	$7F,$B7,$67,$FC,$62,$FF,$09,$8F,$F1,$83,$63,$F9,$5C,$FF,$2E,$7F
	.dc.b	$A6,$7C,$CF,$FA,$42,$BF,$6E,$D8,$FE,$A4,$FE,$3A,$3F,$7E,$E7,$8B
	.dc.b	$F8,$43,$FE,$7B,$F1,$0D,$27,$5F,$F9,$EF,$F7,$EF,$E3,$D5,$FD,$09
	.dc.b	$7C,$C7,$F9,$F5,$CE,$1F,$E4,$77,$2F,$F9,$9D,$B8,$FF,$74,$7F,$9B
	.dc.b	$57,$3A,$E5,$E8,$7B,$3C,$19,$87,$1F,$A2,$B8,$FB,$39,$7F,$D8,$BA
	.dc.b	$FA,$1B,$AF,$F8,$26,$7F,$3E,$7F,$8B,$1C,$3F,$E5,$ED,$1F,$75,$E9
	.dc.b	$F8,$FF,$53,$72,$F4,$7C,$21,$FE,$81,$7F,$E7,$BB,$FF,$27,$C7,$39
	.dc.b	$D4,$FF,$B8,$7C,$55,$6A,$3E,$27,$DC,$EA,$FC,$4D,$8C,$C7,$6A,$E3
	.dc.b	$39,$9F,$F1,$CF,$C3,$3D,$97,$2F,$F5,$98,$C3,$63,$33,$FE,$92,$9C
	.dc.b	$B7,$DC,$FF,$27,$B7,$83,$94,$62,$3E,$4C,$FF,$AC,$E7,$FD,$25,$9B
	.dc.b	$EF,$FF,$85,$E9,$F2,$DB,$7F,$81,$6B,$F1,$DB,$A7,$C4,$F8,$9B,$7F
	.dc.b	$E8,$AA,$F8,$0D,$7E,$53,$6F,$EC,$7C,$F7,$F1,$AE,$8F,$C9,$7C,$5D
	.dc.b	$9F,$C8,$36,$FE,$B5,$5C,$FF,$C8,$3D,$BF,$DF,$DB,$3F,$E2,$7F,$63
	.dc.b	$FD,$EA,$D9,$D4,$3B,$B9,$36,$7F,$B2,$B5,$FC,$E7,$C8,$F8,$76,$FC
	.dc.b	$0F,$CD,$DD,$9F,$2D,$4C,$46,$7D,$7B,$87,$E9,$E3,$0F,$6F,$0F,$C6
	.dc.b	$47,$D0,$F9,$BC,$5F,$29,$23,$CB,$E3,$78,$7C,$A9,$1F,$B3,$7C,$9F
	.dc.b	$51,$23,$EE,$BF,$E2,$3D,$3F,$32,$DD,$0D,$F5,$7D,$84,$AF,$DF,$7D
	.dc.b	$9F,$BF,$8E,$47,$F0,$17,$EA,$7A,$9F,$DC,$7D,$3B,$E8,$9F,$F1,$94
	.dc.b	$66,$BF,$AD,$C6,$27,$51,$FD,$7B,$D3,$F9,$DC,$1F,$EB,$58,$FF,$7A
	.dc.b	$35,$FE,$CB,$1D,$88,$FF,$7B,$78,$FC,$6F,$2F,$62,$4F,$27,$CB,$E9
	.dc.b	$48,$FE,$DA,$F8,$C8,$FF,$92,$7F,$29,$7B,$9F,$91,$84,$61,$FF,$57
	.dc.b	$35,$FF,$2A,$FD,$E3,$2E,$0D,$8F,$F2,$A3,$5F,$E3,$61,$C3,$FC,$79
	.dc.b	$AF,$FA,$FB,$5F,$F0,$F7,$37,$C9,$24,$7F,$81,$9F,$BD,$4F,$BC,$F9
	.dc.b	$64,$9F,$97,$F1,$5F,$3B,$EC,$24,$7C,$E7,$EC,$F7,$C8,$FF,$08,$BF
	.dc.b	$5C,$8F,$A3,$A3,$DD,$F0,$3F,$60,$91,$FE,$2D,$3F,$06,$43,$07,$DC
	.dc.b	$F8,$3F,$BF,$F7,$72,$3E,$FB,$F8,$CF,$B7,$FE,$A7,$1F,$ED,$EF,$23
	.dc.b	$F9,$9E,$E9,$CC,$A7,$31,$86,$FB,$1F,$8C,$F8,$5F,$E1,$1F,$8E,$31
	.dc.b	$A4,$C3,$FD,$6F,$F7,$8E,$9F,$E2,$1F,$0C,$8C,$4F,$FA,$D5,$AF,$F5
	.dc.b	$48,$D4,$FF,$A4,$5B,$1F,$EB,$9B,$6D,$E2,$24,$7E,$7B,$E5,$7F,$8E
	.dc.b	$7D,$37,$F3,$7D,$7F,$D0,$4D,$F3,$FF,$CF,$AD,$A6,$2F,$D3,$FF,$D9
	.dc.b	$D1,$FE,$D0,$6B,$F1,$6B,$FC,$E7,$18,$BF,$EB,$B5,$C7,$CA,$F0,$F8
	.dc.b	$BF,$47,$D9,$FD,$9F,$D1,$CC,$BE,$96,$5F,$93,$E4,$FE,$DF,$1F,$B6
	.dc.b	$F9,$0E,$0E,$87,$37,$D7,$C3,$F7,$B3,$FA,$BD,$5E,$7D,$25,$F1,$3D
	.dc.b	$9E,$96,$CF,$F5,$E3,$83,$FD,$7D,$28,$CB,$63,$3E,$9F,$72,$3F,$C3
	.dc.b	$3E,$63,$F1,$DE,$B7,$F1,$0F,$9F,$93,$E9,$D3,$F1,$D3,$FC,$53,$D5
	.dc.b	$DC,$E7,$6F,$ED,$0F,$F1,$68,$FE,$33,$EC,$C8,$FE,$32,$FE,$6F,$49
	.dc.b	$1C,$5A,$FD,$51,$E0,$7F,$09,$F6,$BF,$07,$EC,$FF,$24,$F8,$D9,$5F
	.dc.b	$EB,$AE,$AF,$ED,$9D,$5F,$6B,$8B,$E9,$CF,$E4,$DA,$FF,$A1,$BA,$BF
	.dc.b	$0D,$CD,$F9,$59,$CC,$FE,$33,$83,$F2,$24,$7F,$98,$FC,$A7,$EC,$FD
	.dc.b	$9F,$E6,$BD,$32,$7E,$14,$7F,$65,$FC,$89,$F1,$FB,$9C,$9F,$83,$E1
	.dc.b	$4E,$A3,$0B,$E5,$79,$12,$3E,$77,$B3,$FC,$0C,$8E,$0E,$3F,$21,$FC
	.dc.b	$D7,$EA,$5D,$74,$93,$F8,$BF,$D0,$1F,$2D,$FC,$ED,$B1,$FC,$C8,$38
	.dc.b	$7F,$9A,$B5,$FF,$28,$B5,$FF,$58,$D3,$CD,$F1,$12,$3D,$BF,$2F,$FC
	.dc.b	$6F,$DE,$FE,$C7,$DE,$69,$27,$FD,$99,$0F,$1B,$64,$8C,$BF,$DB,$3C
	.dc.b	$7F,$E8,$B1,$FA,$E6,$F8,$46,$98,$FB,$CF,$F2,$14,$FF,$B2,$25,$18
	.dc.b	$8F,$E9,$CD,$7F,$72,$D8,$7C,$FB,$E3,$FF,$BE,$B6,$BF,$DC,$8D,$7F
	.dc.b	$DD,$0D,$7C,$1F,$77,$FE,$34,$7F,$DD,$8D,$73,$F0,$FF,$E0,$88,$C4
	.dc.b	$0F,$D2,$FE,$1C,$F2,$7F,$C3,$8D,$8F,$EA,$8D,$7F,$D0,$FD,$28,$C4
	.dc.b	$7C,$1F,$C2,$3F,$BC,$37,$DF,$FF,$0B,$C6,$23,$FB,$83,$5F,$DA,$FD
	.dc.b	$2B,$E0,$BF,$C5,$8D,$8F,$EC,$8D,$7C,$2D,$7F,$DE,$F8,$3F,$0E,$FD
	.dc.b	$04,$7F,$6B,$6B,$E8,$6B,$C1,$AF,$F8,$0F,$D1,$FF,$12,$3F,$6D,$AE
	.dc.b	$7E,$8F,$FC,$63,$18,$8F,$E1,$BC,$9F,$99,$6F,$5F,$F9,$16,$31,$87
	.dc.b	$E4,$63,$FC,$73,$F7,$6F,$53,$FC,$C5,$F1,$FF,$E0,$DC,$3F,$78,$D7
	.dc.b	$F8,$D7,$32,$B8,$47,$F2,$BF,$87,$78,$DF,$E4,$46,$C7,$F8,$8A,$91
	.dc.b	$88,$FF,$5F,$C6,$27,$FC,$97,$F7,$7F,$A5,$9F,$E6,$CD,$7C,$2D,$F3
	.dc.b	$BF,$D7,$3A,$3F,$0C,$6F,$FE,$7B,$FB,$4F,$F3,$0E,$1F,$C5,$57,$C7
	.dc.b	$E4,$D7,$91,$FE,$5E,$FB,$BF,$C1,$6A,$69,$3A,$7F,$85,$70,$B9,$7F
	.dc.b	$99,$ED,$FE,$CA,$8C,$7A,$3F,$E9,$B6,$E9,$FD,$9D,$E1,$F2,$D9,$FB
	.dc.b	$B7,$D2,$5F,$22,$5E,$E7,$2F,$F5,$7F,$C1,$FF,$0A,$FF,$70,$6E,$C6
	.dc.b	$9A,$C5,$F7,$9F,$EE,$CD,$3C,$CD,$F4,$FF,$61,$3A,$C3,$D1,$3F,$ED
	.dc.b	$27,$FA,$C4,$3E,$F3,$FD,$47,$1E,$B6,$BA,$6D,$FF,$8E,$BE,$73,$F9
	.dc.b	$BC,$FF,$6F,$6B,$F8,$06,$BF,$C5,$DA,$FF,$98,$DA,$FF,$83,$BE,$01
	.dc.b	$EB,$FF,$A4,$1B,$1F,$8D,$6B,$F8,$C6,$BF,$C3,$4F,$F7,$8F,$DE,$3D
	.dc.b	$DF,$F4,$EC,$F0,$C3,$FC,$A2,$DD,$3E,$2D,$6E,$FE,$D7,$DD,$FE,$8E
	.dc.b	$FF,$01,$F7,$8F,$9B,$FC,$03,$6B,$7F,$25,$F8,$3B,$E1,$8F,$D0,$6C
	.dc.b	$B7,$46,$1B,$F2,$5F,$E9,$5D,$FF,$C5,$79,$FF,$27,$FB,$CE,$D8,$27
	.dc.b	$E9,$C6,$1B,$A7,$C1,$78,$CA,$75,$F1,$3F,$E6,$D8,$C5,$6F,$E7,$7E
	.dc.b	$45,$73,$6C,$73,$AE,$13,$CB,$0F,$E7,$5E,$5E,$F0,$B7,$CD,$A7,$F3
	.dc.b	$FC,$7C,$73,$C6,$3C,$26,$71,$EE,$D3,$E0,$57,$86,$EC,$73,$C7,$C7
	.dc.b	$3D,$E3,$99,$F1,$6B,$9B,$F1,$E7,$6F,$1A,$FF,$03,$61,$F1,$A3,$F1
	.dc.b	$06,$E3,$33,$F1,$E3,$33,$FE,$55,$F0,$B8,$FC,$41,$A4,$9F,$F2,$EF
	.dc.b	$DB,$FF,$97,$93,$48,$E8,$F4,$9F,$85,$E0,$F7,$BF,$9B,$8A,$E4,$D9
	.dc.b	$D0,$E3,$BA,$77,$19,$DB,$F1,$2F,$16,$BC,$76,$CD,$71,$8F,$F0,$C6
	.dc.b	$DF,$ED,$CC,$3E,$DC,$7F,$0A,$C2,$BF,$DE,$E3,$CB,$5C,$70,$EB,$18
	.dc.b	$AF,$1B,$6B,$D2,$D9,$FD,$55,$3B,$1F,$45,$E3,$DF,$F9,$84,$63,$1F
	.dc.b	$E5,$CF,$FB,$F1,$B3,$B7,$F8,$FF,$CA,$7F,$39,$7F,$9A,$37,$2F,$E3
	.dc.b	$F5,$CD,$FE,$43,$19,$BF,$E7,$DE,$FF,$FB,$C6,$FF,$D1,$5F,$13,$FE
	.dc.b	$5D,$7F,$9F,$46,$6F,$FB,$3C,$E5,$BF,$09,$F2,$DF,$F1,$B4,$65,$B1
	.dc.b	$97,$FC,$49,$E0,$27,$51,$8F,$39,$F4,$FB,$8F,$F5,$9C,$63,$1F,$EE
	.dc.b	$0D,$8F,$ED,$ED,$7F,$92,$6C,$FD,$FB,$F0,$D5,$FD,$5D,$B3,$FE,$00
	.dc.b	$AE,$5A,$7F,$5F,$9C,$BF,$D8,$73,$7C,$FF,$CF,$BF,$E3,$90,$FA,$7C
	.dc.b	$65,$FE,$83,$AB,$39,$CB,$FD,$69,$F7,$8F,$9C,$F6,$6E,$8C,$4F,$F8
	.dc.b	$2D,$B9,$7F,$82,$9A,$FB,$3E,$F1,$F5,$3E,$06,$D7,$F7,$37,$EF,$6B
	.dc.b	$FD,$E5,$CD,$7F,$87,$BB,$BD,$F5,$FD,$9B,$EF,$1F,$2F,$ED,$CD,$5A
	.dc.b	$BF,$E7,$55,$AB,$FF,$67,$4F,$49,$FF,$3A,$35,$FB,$AD,$7F,$DD,$6D
	.dc.b	$A7,$E1,$BF,$EB,$0A,$FF,$6C,$36,$BF,$D9,$9E,$C7,$FB,$CB,$97,$C9
	.dc.b	$39,$D3,$F4,$DC,$8F,$85,$FA,$9F,$29,$FE,$10,$BF,$AD,$5F,$81,$9C
	.dc.b	$FA,$3E,$B8,$FC,$27,$1F,$E3,$5B,$A7,$F2,$4D,$11,$87,$FC,$1D,$97
	.dc.b	$D2,$7F,$CF,$91,$F8,$DF,$03,$BC,$FF,$67,$AC,$9F,$E8,$1C,$DF,$98
	.dc.b	$7E,$2A,$33,$87,$6A,$E0,$FC,$0F,$BC,$FF,$43,$4F,$8F,$74,$62,$7F
	.dc.b	$D0,$AD,$7F,$CE,$AD,$7F,$5B,$F7,$8F,$9A,$FC,$7B,$6F,$FC,$B0,$D9
	.dc.b	$FF,$2B,$B5,$FF,$3F,$FC,$2B,$E8,$3F,$65,$8A,$31,$7F,$E8,$89,$CC
	.dc.b	$62,$BF,$D3,$3F,$78,$F9,$BE,$2D,$BF,$63,$67,$73,$FA,$4F,$BC,$7C
	.dc.b	$DD,$25,$FF,$9B,$E3,$2F,$C5,$FA,$DB,$7F,$EA,$86,$BF,$E9,$86,$BF
	.dc.b	$EB,$3C,$D1,$88,$74,$FF,$B1,$BD,$EF,$DA,$FC,$89,$A4,$C3,$FD,$01
	.dc.b	$F7,$9F,$CB,$6F,$FD,$A8,$D7,$F8,$EB,$5E,$AD,$B7,$CB,$3E,$B7,$F8
	.dc.b	$1B,$6F,$FD,$D7,$18,$3F,$83,$F8,$3C,$49,$F0,$78,$33,$F4,$92,$BF
	.dc.b	$D2,$92,$8C,$65,$FE,$94,$8F,$F6,$14,$EB,$E4,$7A,$91,$FC,$87,$CF
	.dc.b	$7F,$9E,$9F,$CB,$E2,$FC,$DB,$FA,$96,$1F,$B6,$AD,$3F,$A8,$AF,$F7
	.dc.b	$05,$6F,$E2,$3F,$94,$4B,$EF,$7C,$95,$FA,$F8,$C6,$7F,$CF,$DB,$19
	.dc.b	$9D,$C7,$F8,$62,$FC,$5E,$1F,$F3,$36,$D7,$EB,$27,$89,$E3,$3E,$76
	.dc.b	$5E,$6F,$0B,$FC,$01,$C1,$FC,$83,$EA,$BF,$90,$6E,$7F,$9F,$73,$37
	.dc.b	$7D,$AE,$A7,$31,$F0,$B0,$FF,$5B,$B7,$3F,$EB,$B8,$73,$7F,$7F,$87
	.dc.b	$3E,$0E,$3E,$BF,$83,$BA,$D7,$F0,$4D,$3F,$99,$C6,$63,$8E,$3F,$CE
	.dc.b	$B2,$F4,$E5,$FA,$F7,$C3,$E6,$FC,$3D,$BF,$C1,$97,$FC,$3B,$C5,$7E
	.dc.b	$4C,$FA,$6D,$7B,$3F,$CA,$23,$15,$B6,$F1,$FF,$91,$AF,$8D,$FF,$73
	.dc.b	$6D,$7F,$27,$6C,$FD,$5C,$1F,$49,$E3,$3B,$5D,$BF,$76,$9F,$EF,$56
	.dc.b	$FC,$BF,$CC,$F3,$BE,$9B,$F3,$BD,$79,$3A,$3E,$1B,$F4,$BD,$6F,$B3
	.dc.b	$F4,$96,$7A,$B8,$7E,$4F,$DC,$DD,$7F,$30,$FD,$AF,$98,$FC,$AF,$DD
	.dc.b	$BE,$4F,$17,$F8,$B8,$C3,$7E,$DF,$F2,$D0,$69,$23,$F3,$90,$F8,$1E
	.dc.b	$A7,$69,$B3,$F7,$3E,$CF,$F2,$3C,$3F,$8C,$6C,$7C,$68,$FC,$7F,$AD
	.dc.b	$8D,$3F,$60,$FD,$9F,$02,$57,$ED,$BF,$26,$7C,$86,$2F,$F4,$5E,$22
	.dc.b	$5C,$3F,$A9,$6D,$FE,$07,$0C,$4F,$E7,$BE,$CB,$F7,$3F,$05,$FA,$2A
	.dc.b	$FC,$9C,$63,$C6,$CF,$8F,$FB,$9A,$FE,$07,$9F,$FA,$39,$AF,$F8,$D2
	.dc.b	$35,$DF,$D0,$E3,$87,$EE,$6B,$F3,$35,$FB,$9F,$13,$FC,$21,$3F,$88
	.dc.b	$3F,$63,$ED,$FE,$E1,$FD,$9F,$D9,$47,$47,$EA,$1F,$F0,$E7,$EC,$7F
	.dc.b	$94,$8F,$D7,$36,$FF,$8D,$7B,$DA,$F0,$7F,$7B,$94,$BD,$C5,$AF,$25
	.dc.b	$FB,$1F,$E9,$2D,$C8,$D3,$5E,$5F,$F0,$B9,$FE,$12,$79,$35,$B8,$7F
	.dc.b	$06,$FC,$A1,$F4,$5E,$8F,$37,$F8,$9C,$62,$7D,$31,$97,$FC,$31,$97
	.dc.b	$82,$3F,$C9,$B9,$7F,$36,$C4,$7F,$9C,$1C,$38,$1F,$7A,$F9,$EF,$E6
	.dc.b	$4D,$C3,$F9,$B3,$67,$FC,$4C,$D7,$FD,$97,$F7,$AF,$B6,$FE,$7E,$DA
	.dc.b	$FE,$69,$5C,$7C,$EF,$F4,$34,$E6,$FE,$5F,$A3,$E8,$25,$FB,$32,$FF
	.dc.b	$4C,$4F,$9F,$81,$B3,$EA,$FE,$29,$F7,$B0,$B8,$7C,$BF,$0B,$E4,$9F
	.dc.b	$2B,$C3,$B1,$F8,$2F,$2F,$F0,$A6,$32,$FB,$AF,$F9,$8F,$EC,$FF,$D2
	.dc.b	$B1,$FE,$06,$6F,$19,$A6,$B4,$F2,$F8,$FF,$60,$E5,$F1,$97,$1F,$BD
	.dc.b	$F7,$9F,$73,$DC,$FB,$AD,$8F,$C1,$7D,$E7,$F8,$C3,$84,$D2,$47,$F9
	.dc.b	$16,$3F,$C4,$3C,$9F,$ED,$BC,$BE,$2F,$C0,$FE,$22,$FF,$12,$FF,$67
	.dc.b	$FC,$27,$F8,$DE,$FF,$C1,$1F,$E3,$58,$FF,$7E,$73,$87,$FB,$97,$C8
	.dc.b	$46,$AB,$87,$17,$FA,$D6,$72,$D9,$FE,$41,$7F,$E6,$B9,$CB,$E7,$E0
	.dc.b	$FC,$D9,$F9,$CF,$83,$B1,$FA,$1F,$99,$FC,$F9,$FA,$0F,$98,$B1,$FA
	.dc.b	$5F,$92,$FD,$1F,$C7,$FE,$91,$B1,$FE,$16,$F3,$BF,$3E,$F8,$5B,$AF
	.dc.b	$F8,$25,$AF,$E8,$73,$FE,$4E,$FC,$FF,$A6,$FD,$79,$FB,$0F,$B1,$B7
	.dc.b	$FB,$5F,$08,$F2,$E4,$FC,$9C,$62,$FF,$DC,$51,$89,$D7,$2F,$EA,$5F
	.dc.b	$CD,$2F,$B2,$E1,$18,$6D,$7F,$95,$E7,$53,$CA,$73,$3F,$BC,$F8,$9F
	.dc.b	$EF,$9E,$4F,$F8,$AD,$B1,$FE,$9B,$6B,$FB,$F3,$FC,$C3,$EA,$27,$87
	.dc.b	$C7,$7F,$94,$E7,$83,$5E,$15,$FE,$8B,$DB,$67,$E4,$7E,$6B,$F8,$56
	.dc.b	$5E,$C6,$D7,$A5,$AB,$4F,$DC,$C6,$1F,$DD,$D5,$D3,$C1,$7F,$1A,$30
	.dc.b	$6F,$F2,$9F,$F2,$E6,$1F,$E2,$C6,$BF,$EC,$86,$F5,$7E,$8D,$7C,$4F
	.dc.b	$F1,$F8,$FC,$E7,$DB,$FF,$94,$6B,$FC,$E4,$DB,$34,$D6,$4C,$1F,$3F
	.dc.b	$CF,$2F,$F4,$64,$3E,$2F,$18,$FB,$E3,$0F,$9E,$B3,$7C,$FD,$9B,$5F
	.dc.b	$E5,$1A,$F6,$46,$3E,$87,$3F,$6B,$27,$E0,$3C,$3E,$E3,$BE,$BE,$8F
	.dc.b	$5B,$FC,$D8,$DA,$FD,$96,$DF,$EE,$76,$CE,$A1,$EC,$72,$5E,$C5,$31
	.dc.b	$1E,$6F,$8C,$97,$F4,$5F,$4F,$AB,$23,$FD,$B1,$B7,$F4,$36,$C7,$F9
	.dc.b	$FF,$7F,$94,$D9,$7C,$37,$C2,$F8,$79,$7F,$76,$3E,$2F,$D6,$C8,$FC
	.dc.b	$03,$F3,$BD,$1F,$E6,$5E,$E5,$F2,$BF,$CA,$35,$FD,$91,$AF,$E8,$67
	.dc.b	$6E,$67,$FD,$25,$1F,$DE,$3E,$1B,$A7,$B0,$69,$F5,$BE,$06,$5F,$E4
	.dc.b	$5A,$FE,$47,$ED,$3E,$DF,$93,$23,$F2,$CF,$4F,$85,$91,$FD,$75,$F0
	.dc.b	$E0,$69,$F9,$5F,$BC,$FF,$48,$CF,$FA,$75,$AF,$FA,$74,$38,$F0,$3F
	.dc.b	$89,$43,$EF,$1F,$0B,$F9,$48,$C3,$FA,$18,$CC,$7F,$AC,$DA,$FF,$07
	.dc.b	$36,$7F,$73,$8F,$39,$1C,$1F,$80,$F1,$FE,$4C,$48,$FD,$B4,$7E,$7B
	.dc.b	$C9,$91,$FB,$97,$E7,$F5,$48,$FF,$00,$3F,$10,$8F,$E3,$3E,$AB,$F4
	.dc.b	$E4,$7F,$C7,$67,$D8,$21,$83,$F8,$37,$D5,$FE,$BB,$C6,$91,$FC,$25
	.dc.b	$FC,$57,$DC,$7F,$9D,$E3,$FD,$64,$D7,$FC,$A4,$DD,$3F,$D9,$2D,$9F
	.dc.b	$F1,$6F,$67,$FA,$B7,$EE,$BE,$77,$DE,$7E,$87,$0F,$AC,$DC,$3E,$AB
	.dc.b	$6F,$FC,$64,$BB,$F8,$9F,$C5,$5C,$57,$FB,$9D,$AF,$D5,$6B,$FE,$25
	.dc.b	$D5,$EF,$78,$BC,$69,$3F,$ED,$29,$CB,$FE,$FC,$FC,$2D,$3F,$C2,$9B
	.dc.b	$19,$8F,$E3,$7F,$26,$E2,$79,$F8,$6D,$7F,$B3,$DA,$FF,$BC,$67,$84
	.dc.b	$F5,$F2,$7E,$7F,$C0,$7F,$22,$CB,$F6,$BB,$FF,$86,$5B,$BF,$F9,$63
	.dc.b	$BD,$C6,$BE,$1F,$17,$CB,$4F,$AF,$18,$7F,$CB,$F8,$FE,$8B,$C9,$7B
	.dc.b	$9F,$A1,$6C,$EB,$2F,$F3,$14,$62,$3F,$58,$BD,$FE,$07,$F4,$77,$11
	.dc.b	$FA,$53,$FA,$3C,$64,$FE,$3D,$A3,$E2,$FC,$1F,$03,$2B,$E9,$47,$E4
	.dc.b	$2B,$E1,$7C,$64,$9F,$C8,$C7,$C5,$F8,$BD,$46,$0F,$B4,$FC,$9F,$88
	.dc.b	$91,$F7,$5F,$CD,$FC,$0C,$8F,$CD,$B5,$6D,$F3,$FE,$57,$FA,$4F,$BF
	.dc.b	$F6,$23,$FC,$80,$DD,$FF,$D1,$3E,$AA,$31,$6E,$8F,$C2,$6F,$F8,$33
	.dc.b	$E2,$38,$8F,$E1,$4D,$7F,$C2,$4D,$7F,$C6,$53,$FE,$23,$7F,$15,$3C
	.dc.b	$DF,$AC,$2B,$C5,$EF,$7E,$E3,$E0,$7F,$15,$F1,$12,$BF,$73,$1F,$8D
	.dc.b	$F9,$3D,$46,$0F,$D1,$3F,$91,$F9,$39,$1F,$A5,$7F,$81,$FC,$1C,$8F
	.dc.b	$E0,$6D,$7E,$4E,$BF,$C0,$39,$BF,$A3,$47,$9B,$3F,$8D,$F0,$9F,$30
	.dc.b	$91,$EA,$1F,$99,$31,$FE,$68,$D8,$FC,$63,$5F,$E6,$4D,$7F,$C7,$4D
	.dc.b	$7C,$5F,$29,$F6,$B8,$24,$7E,$FE,$3F,$8E,$EF,$FD,$7B,$BB,$E7,$4F
	.dc.b	$F6,$36,$BF,$EC,$3E,$FF,$D7,$CE,$FC,$19,$3F,$EC,$EA,$FD,$EB,$B6
	.dc.b	$BF,$B7,$7F,$C9,$0E,$74,$F0,$11,$FF,$6B,$E3,$FE,$0E,$6F,$99,$FD
	.dc.b	$CA,$31,$1F,$DD,$DA,$F9,$7E,$73,$F9,$A3,$7E,$2F,$FB,$CC,$62,$3F
	.dc.b	$84,$35,$66,$FB,$CF,$F3,$04,$7F,$BA,$1A,$FF,$81,$5A,$FC,$36,$BF
	.dc.b	$CF,$BE,$F1,$EC,$7F,$86,$DB,$1F,$D9,$FC,$F4,$62,$3F,$CE,$11,$89
	.dc.b	$FF,$07,$FD,$E3,$C0,$A4,$AF,$F7,$DC,$65,$E9,$F4,$5F,$E5,$06,$FC
	.dc.b	$3F,$F8,$AE,$31,$1D,$9A,$FF,$B9,$FE,$F3,$FC,$42,$DF,$7B,$FC,$65
	.dc.b	$18,$8F,$EB,$2D,$7E,$3F,$C0,$7F,$9B,$1F,$C2,$35,$FE,$68,$D7,$7F
	.dc.b	$3F,$FE,$3D,$E6,$FF,$7A,$47,$C3,$6B,$F0,$9A,$FF,$BE,$DA,$FF,$38
	.dc.b	$FC,$A1,$E7,$FF,$9A,$BC,$64,$62,$7F,$81,$C6,$27,$FA,$AF,$CA,$7C
	.dc.b	$F8,$FF,$59,$1C,$3D,$5F,$E6,$CD,$AF,$F0,$E3,$5F,$F5,$23,$5F,$F4
	.dc.b	$37,$E9,$3F,$E2,$A8,$FF,$5E,$79,$2F,$C6,$BF,$B2,$D7,$F5,$D6,$CF
	.dc.b	$82,$B8,$78,$BF,$D6,$7F,$0C,$7C,$C7,$F9,$5D,$B8,$7F,$1C,$EB,$46
	.dc.b	$23,$F8,$69,$FE,$A7,$FB,$4F,$DB,$E3,$FE,$CD,$A7,$DE,$BC,$FF,$F2
	.dc.b	$CB,$6F,$FC,$71,$3A,$9C,$9A,$F3,$BF,$D2,$BF,$2D,$FE,$A4,$F9,$AB
	.dc.b	$F2,$A7,$FC,$D9,$8A,$B7,$F7,$AF,$56,$93,$3F,$E6,$B8,$A7,$FC,$A5
	.dc.b	$C1,$FD,$E7,$EF,$7E,$C5,$7F,$9C,$5A,$FF,$0F,$6B,$BE,$7F,$ED,$96
	.dc.b	$EA,$FC,$4F,$07,$FB,$B6,$31,$F4,$1E,$A9,$37,$E7,$BD,$BF,$78,$F5
	.dc.b	$E9,$32,$FA,$1F,$73,$FF,$00,$1F,$9F,$CA,$8C,$47,$F2,$36,$BF,$E1
	.dc.b	$36,$BF,$C3,$1B,$0F,$A0,$7C,$47,$FA,$31,$B5,$F5,$1A,$FF,$A4,$DA
	.dc.b	$89,$A3,$EF,$1F,$15,$FE,$9F,$AE,$3B,$FE,$0D,$BC,$5C,$2B,$78,$7D
	.dc.b	$4F,$BC,$FF,$3D,$DF,$C9,$6D,$76,$6C,$B1,$7D,$E3,$E1,$FF,$0E,$D9
	.dc.b	$FB,$ED,$7F,$2F,$3C,$27,$7F,$3D,$FE,$8F,$DD,$F4,$7F,$E1,$0B,$7D
	.dc.b	$DB,$EA,$3E,$04,$7D,$1F,$A5,$B3,$98,$F8,$96,$9C,$D7,$08,$EB,$B7
	.dc.b	$C9,$BE,$17,$F6,$F0,$F0,$46,$2B,$94,$EE,$F9,$61,$D2,$3B,$DF,$09
	.dc.b	$EC,$FA,$7C,$BE,$55,$E8,$6D,$7D,$97,$B4,$EE,$B8,$E3,$DE,$BD,$77
	.dc.b	$DD,$F8,$06,$4F,$44,$5C,$3D,$B1,$F2,$2B,$CE,$E6,$B9,$4E,$B1,$F0
	.dc.b	$DF,$F9,$56,$38,$D7,$CE,$3A,$9F,$EB,$D7,$2F,$CB,$F8,$67,$89,$C5
	.dc.b	$B1,$B6,$BC,$E5,$18,$8F,$F6,$47,$DE,$7F,$99,$5F,$CE,$35,$34,$8F
	.dc.b	$FA,$33,$2F,$36,$EE,$2F,$99,$1D,$27,$47,$E2,$27,$4E,$0F,$C5,$D7
	.dc.b	$07,$55,$5C,$9B,$3F,$40,$E5,$F3,$99,$9E,$BB,$7E,$32,$32,$D8,$CD
	.dc.b	$F9,$63,$15,$CE,$BD,$E7,$37,$17,$DE,$F8,$CF,$D5,$8C,$4F,$A6,$7F
	.dc.b	$B1,$FC,$4F,$FB,$F2,$BF,$82,$F7,$23,$11,$FB,$98,$C5,$7F,$BE,$6B
	.dc.b	$8B,$7F,$2C,$FF,$35,$37,$DF,$E3,$29,$CD,$FD,$26,$BF,$3F,$EF,$5E
	.dc.b	$B7,$06,$E1,$DD,$B3,$F6,$9A,$FF,$93,$7F,$1D,$FB,$93,$FD,$59,$B1
	.dc.b	$F9,$5E,$E4,$62,$38,$4F,$BA,$7F,$9E,$79,$77,$C0,$FE,$C5,$3F,$D8
	.dc.b	$1B,$5C,$DA,$FE,$1F,$EE,$3F,$95,$CF,$C5,$6C,$7E,$41,$AE,$FC,$0F
	.dc.b	$8B,$E9,$7F,$63,$AF,$98,$D9,$FC,$BB,$5F,$CA,$35,$FF,$6D,$5F,$F9
	.dc.b	$FE,$7C,$DF,$0F,$E9,$23,$F5,$FE,$F3,$FB,$73,$7C,$BF,$AB,$18,$BF
	.dc.b	$D2,$B5,$FD,$17,$DE,$3E,$9B,$ED,$B6,$FF,$04,$D9,$FD,$6B,$5F,$F4
	.dc.b	$07,$7F,$F9,$4E,$7F,$C1,$2D,$8F,$A8,$D7,$7E,$F7,$D3,$FA,$1F,$AD
	.dc.b	$1F,$EF,$76,$BF,$E7,$06,$F0,$9B,$4F,$A2,$FB,$8D,$C3,$C3,$18,$33
	.dc.b	$F1,$72,$3C,$AD,$C3,$F1,$1E,$27,$DE,$7F,$B5,$63,$FD,$43,$93,$F1
	.dc.b	$13,$99,$FF,$41,$B6,$5E,$EF,$C3,$9F,$F2,$EE,$DF,$8F,$8C,$47,$F9
	.dc.b	$92,$3F,$5E,$FF,$B4,$FE,$3A,$F6,$1A,$4C,$FF,$C6,$7F,$88,$3E,$37
	.dc.b	$F3,$4D,$8C,$C6,$0F,$CF,$7C,$27,$F6,$58,$FF,$41,$F8,$98,$23,$FC
	.dc.b	$EF,$1F,$F7,$AF,$CC,$78,$89,$19,$8F,$F7,$44,$7E,$A7,$A1,$F0,$34
	.dc.b	$95,$FE,$48,$8C,$BF,$E4,$AF,$8A,$B7,$FB,$56,$BF,$EB,$3A,$FD,$A4
	.dc.b	$62,$3F,$BD,$FE,$33,$FA,$A6,$F7,$7C,$88,$C4,$7F,$87,$DB,$C4,$FB
	.dc.b	$87,$BF,$FB,$F6,$E5,$FE,$5C,$6D,$FF,$07,$6B,$FE,$D9,$7F,$C9,$11
	.dc.b	$BF,$6B,$91,$1F,$89,$83,$E7,$FC,$84,$8C,$BE,$1F,$AD,$91,$E1,$6E
	.dc.b	$1F,$DA,$F9,$5F,$9C,$BE,$5F,$F9,$9C,$61,$B1,$9B,$FB,$13,$97,$EB
	.dc.b	$79,$CF,$A1,$FC,$6B,$63,$31,$83,$9F,$6C,$8E,$77,$F9,$DB,$F1,$CF
	.dc.b	$F4,$4F,$8F,$FC,$A7,$D2,$DE,$C3,$49,$87,$FB,$66,$3F,$21,$A7,$FB
	.dc.b	$61,$AF,$F8,$DA,$BF,$3D,$D3,$CC,$E9,$F4,$72,$F9,$36,$B9,$7D,$67
	.dc.b	$84,$93,$DD,$F8,$BE,$34,$8F,$5B,$67,$F3,$DE,$D3,$E0,$BF,$62,$DA
	.dc.b	$FC,$F3,$67,$73,$F4,$3C,$B9,$5F,$19,$B3,$ED,$F9,$7F,$88,$49,$F8
	.dc.b	$4F,$C7,$FA,$D9,$1F,$15,$B3,$F1,$7D,$5F,$98,$49,$F9,$4F,$CF,$89
	.dc.b	$1F,$35,$A7,$EE,$7E,$D1,$F3,$9F,$C3,$1B,$3F,$B9,$6C,$EE,$75,$F3
	.dc.b	$52,$B4,$D9,$E9,$E9,$73,$24,$F3,$7F,$09,$F8,$C7,$FA,$A2,$3F,$85
	.dc.b	$71,$A3,$15,$FE,$47,$8C,$7C,$4F,$03,$8E,$F2,$F8,$DF,$D1,$BF,$C3
	.dc.b	$35,$FF,$4E,$74,$39,$1E,$47,$E9,$9B,$7F,$A4,$6D,$7E,$B5,$AF,$FA
	.dc.b	$1A,$B9,$7D,$5F,$12,$57,$17,$F9,$AF,$07,$F6,$FA,$FF,$11,$79,$38
	.dc.b	$25,$7F,$B3,$BB,$F6,$47,$07,$FB,$4B,$E3,$79,$91,$C1,$FB,$AF,$C6
	.dc.b	$BF,$8D,$E4,$69,$3C,$9F,$E0,$6F,$F0,$F9,$E1,$E2,$60,$8F,$CA,$F9
	.dc.b	$DF,$AA,$47,$07,$CE,$F9,$3F,$B2,$47,$07,$F2,$5E,$0F,$E5,$04,$70
	.dc.b	$7D,$6F,$37,$F0,$24,$70,$7D,$AF,$A6,$FC,$29,$1C,$1F,$CC,$70,$FE
	.dc.b	$68,$47,$07,$F4,$FF,$B5,$FE,$B9,$C1,$FD,$55,$AF,$FA,$45,$AF,$FB
	.dc.b	$01,$AF,$FB,$87,$F2,$77,$C2,$FF,$6E,$6D,$7F,$5B,$6C,$FF,$61,$6B
	.dc.b	$FD,$07,$6F,$E9,$04,$9F,$E9,$0F,$E8,$3E,$56,$47,$E6,$1B,$3F,$9F
	.dc.b	$E7,$7D,$27,$E9,$89,$5F,$9E,$7F,$CA,$5F,$5F,$FE,$FE,$8F,$EE,$5E
	.dc.b	$9A,$31,$7F,$E5,$B6,$BF,$ED,$E8,$D7,$C9,$FF,$96,$FF,$59,$7D,$8D
	.dc.b	$26,$1D,$23,$2F,$EB,$BE,$A3,$F7,$24,$70,$7E,$CB,$EA,$FF,$7A,$47
	.dc.b	$07,$ED,$B2,$FE,$02,$47,$07,$EE,$BE,$3B,$F8,$29,$1C,$1F,$D6,$3E
	.dc.b	$B9,$BF,$F5,$E6,$C6,$63,$06,$79,$A4,$71,$6E,$DF,$3F,$E0,$7F,$89
	.dc.b	$92,$7F,$85,$BF,$C5,$BE,$12,$47,$F0,$E6,$CF,$F7,$1C,$1D,$7E,$32
	.dc.b	$57,$9D,$F3,$78,$92,$3D,$4D,$AF,$C0,$78,$FF,$C9,$09,$3F,$C7,$1F
	.dc.b	$E4,$FE,$47,$FB,$3A,$3F,$8E,$B6,$43,$C8,$FE,$55,$19,$8C,$73,$A3
	.dc.b	$13,$FC,$BF,$C9,$B3,$FC,$FA,$FF,$94,$C6,$0E,$56,$FB,$6F,$F2,$77
	.dc.b	$8F,$FE,$3D,$6C,$7F,$90,$DA,$FF,$BF,$DA,$F9,$8E,$5E,$0B,$C4,$E4
	.dc.b	$D8,$E4,$D7,$93,$5F,$7E,$5F,$D1,$FE,$AF,$A6,$5F,$D0,$1B,$3A,$CD
	.dc.b	$18,$8F,$86,$DA,$FE,$A1,$F0,$36,$BF,$CC,$D5,$F3,$2F,$FA,$5C,$E4
	.dc.b	$D7,$83,$FE,$5D,$BF,$EB,$0D,$BF,$E1,$37,$FD,$52,$3F,$CC,$59,$3F
	.dc.b	$1B,$78,$B8,$6D,$7F,$8D,$1A,$FF,$8F,$5A,$FF,$51,$7F,$CD,$2E,$F9
	.dc.b	$F9,$11,$FE,$C7,$5E,$32,$38,$3F,$B7,$FE,$33,$FE,$5E,$6F,$C0,$FA
	.dc.b	$E3,$11,$FD,$C5,$AF,$87,$C8,$F4,$11,$F4,$3E,$BE,$79,$1E,$B6,$BC
	.dc.b	$FE,$03,$B1,$1F,$1B,$E1,$F7,$E4,$79,$9A,$F8,$FE,$43,$CC,$47,$D6
	.dc.b	$FA,$7E,$DE,$47,$B9,$AF,$FB,$F7,$C4,$FF,$61,$3F,$C7,$7E,$2F,$FD
	.dc.b	$2D,$E4,$27,$33,$B9,$CF,$88,$F9,$3A,$49,$FE,$63,$19,$7E,$6F,$B5
	.dc.b	$F2,$C8,$E0,$FA,$3F,$33,$F3,$C8,$E0,$E7,$F2,$BE,$02,$38,$3F,$D5
	.dc.b	$9F,$5D,$FE,$D9,$6F,$C7,$FF,$AB,$D4,$65,$FF,$5B,$35,$FD,$90,$FB
	.dc.b	$37,$C1,$71,$6C,$7F,$AE,$5A,$FF,$AF,$DA,$FF,$4B,$C9,$F0,$3F,$18
	.dc.b	$91,$F1,$1F,$C6,$7D,$94,$8F,$C8,$B5,$FC,$77,$C4,$7E,$44,$8F,$E5
	.dc.b	$9F,$A3,$ED,$48,$F9,$ED,$7F,$2D,$F1,$FF,$99,$23,$F9,$E7,$F3,$9F
	.dc.b	$6D,$23,$F4,$2D,$7F,$D5,$5E,$67,$F2,$57,$FD,$4A,$D7,$3E,$67,$EE
	.dc.b	$A3,$11,$FE,$43,$5E,$7E,$49,$1C,$9F,$A9,$E7,$BA,$A9,$27,$FD,$CD
	.dc.b	$19,$6F,$5F,$02,$38,$37,$D9,$D0,$8E,$0F,$F5,$EF,$E3,$4F,$80,$FE
	.dc.b	$22,$D8,$CC,$60,$CF,$5C,$8E,$2D,$7F,$81,$7A,$FF,$C1,$48,$FF,$0A
	.dc.b	$7F,$84,$7C,$94,$8F,$E1,$AD,$7F,$1E,$3E,$D5,$F0,$DF,$97,$6C,$68
	.dc.b	$D4,$64,$D7,$C3,$C8,$D4,$A3,$11,$DA,$1F,$13,$D4,$91,$C1,$FE,$3D
	.dc.b	$8C,$8F,$E4,$6D,$7F,$90,$7D,$87,$F2,$32,$3F,$CA,$1F,$E4,$D9,$C8
	.dc.b	$FE,$56,$D7,$F9,$47,$A9,$FC,$AC,$8F,$F3,$07,$E9,$EF,$23,$EB,$B5
	.dc.b	$F8,$FF,$96,$3E,$5B,$F7,$8D,$8F,$8C,$D7,$71,$F7,$B8,$64,$7E,$09
	.dc.b	$AF,$F3,$BF,$8D,$FE,$7C,$47,$FA,$1B,$FD,$07,$E1,$A4,$7F,$46,$6B
	.dc.b	$FD,$0F,$E1,$3F,$A3,$11,$FE,$96,$FF,$49,$F6,$24,$7F,$4E,$6B,$FC
	.dc.b	$43,$D1,$7D,$F7,$F7,$B8,$FF,$5A,$B5,$FF,$02,$36,$36,$7E,$1F,$ED
	.dc.b	$A4,$FE,$09,$AF,$59,$FE,$E2,$DF,$5F,$FB,$8C,$62,$3F,$D7,$4D,$7F
	.dc.b	$C5,$3E,$7B,$E5,$9F,$1D,$FD,$11,$B3,$FE,$6E,$6B,$FC,$AA,$78,$4F
	.dc.b	$F3,$1F,$A7,$FD,$74,$FE,$81,$AF,$F3,$06,$BF,$E2,$36,$A9,$F1,$7C
	.dc.b	$17,$84,$91,$E1,$7C,$9E,$AC,$8F,$23,$5F,$07,$3F,$8C,$8F,$EB,$6D
	.dc.b	$1F,$D8,$3E,$D7,$FA,$46,$06,$92,$BF,$B2,$FE,$04,$F6,$29,$23,$FD
	.dc.b	$23,$19,$7F,$75,$F8,$4B,$97,$FA,$A3,$63,$31,$83,$FB,$8F,$CE,$3E
	.dc.b	$2E,$92,$7F,$D3,$B1,$97,$FD,$3F,$F8,$2B,$CC,$FE,$F2,$D8,$CC,$60
	.dc.b	$F0,$F7,$48,$F2,$B5,$FB,$5F,$8D,$71,$6C,$FF,$81,$23,$11,$FE,$A9
	.dc.b	$6B,$F8,$8F,$C3,$9E,$3F,$F8,$29,$B1,$FE,$B4,$6B,$F8,$76,$BF,$EB
	.dc.b	$3F,$B1,$7B,$5F,$E1,$26,$C7,$FA,$F1,$AF,$E3,$DA,$F1,$EB,$E0,$47
	.dc.b	$83,$FE,$1C,$FB,$2F,$F2,$3C,$7E,$5D,$AF,$FB,$0D,$AF,$E8,$9A,$FE
	.dc.b	$4E,$5F,$2F,$E5,$24,$79,$5F,$E7,$1F,$18,$F8,$4F,$F6,$CC,$FF,$40
	.dc.b	$8C,$7C,$9E,$A7,$33,$FE,$12,$9C,$BB,$FB,$2F,$F0,$5B,$7D,$BE,$B1
	.dc.b	$88,$E0,$D7,$FD,$B9,$F5,$5F,$E3,$86,$F9,$BF,$E3,$C8,$C4,$7E,$F9
	.dc.b	$AA,$DE,$4F,$FA,$EA,$3D,$0D,$7C,$CD,$7F,$D3,$51,$C3,$C0,$F8,$3F
	.dc.b	$63,$FE,$04,$9F,$F5,$9C,$A3,$11,$DD,$B1,$9F,$67,$CD,$CF,$2B,$FD
	.dc.b	$F7,$8A,$31,$1F,$C3,$FC,$27,$DC,$7F,$96,$1B,$1F,$DB,$BC,$24,$62
	.dc.b	$38,$4F,$D6,$FB,$27,$97,$49,$3F,$5E,$32,$FD,$9F,$B2,$FE,$D6,$DF
	.dc.b	$4F,$EF,$C6,$23,$ED,$35,$F8,$BF,$64,$F8,$1F,$C2,$B6,$3F,$08,$D7
	.dc.b	$F0,$4D,$7F,$CA,$BA,$BE,$59,$F6,$3F,$ED,$EF,$43,$FC,$A7,$3F,$E3
	.dc.b	$0A,$C7,$C3,$7F,$B3,$63,$10,$9D,$DF,$28,$C5,$6E,$31,$3D,$3E,$03
	.dc.b	$FD,$AC,$DA,$EA,$7F,$BB,$AF,$C9,$86,$21,$18,$7A,$9B,$AE,$78,$79
	.dc.b	$5B,$1C,$F0,$F0,$CF,$8E,$FA,$5F,$AB,$0F,$15,$F4,$3B,$DF,$07,$95
	.dc.b	$70,$C3,$AC,$79,$1B,$7E,$5B,$F8,$75,$FE,$2F,$8F,$45,$7D,$0B,$F8
	.dc.b	$CD,$9D,$7C,$CB,$EC,$69,$27,$FC,$7B,$69,$D4,$7F,$A4,$FE,$9F,$FD
	.dc.b	$28,$9A,$47,$43,$EC,$BF,$D1,$BD,$1F,$92,$7F,$D1,$F1,$C4,$D6,$BF
	.dc.b	$C2,$78,$D1,$83,$FD,$3B,$78,$AE,$73,$6B,$FC,$A7,$36,$BF,$77,$35
	.dc.b	$E6,$77,$D7,$8B,$F4,$4E,$F3,$F8,$17,$F8,$54,$D7,$CD,$7C,$0F,$4C
	.dc.b	$F5,$8D,$9E,$73,$BB,$9C,$3E,$06,$1E,$69,$D1,$C2,$BF,$D6,$B3,$FE
	.dc.b	$D3,$F1,$3F,$6F,$AF,$EE,$23,$18,$FC,$48,$C7,$C7,$FF,$B6,$63,$07
	.dc.b	$F1,$65,$5C,$A3,$0D,$C3,$F9,$37,$D9,$7E,$AE,$78,$C7,$F2,$9A,$E3
	.dc.b	$18,$C1,$5B,$AF,$DC,$CF,$F2,$4F,$B2,$7B,$FD,$6D,$18,$9E,$4F,$F3
	.dc.b	$46,$CF,$8F,$AD,$F6,$4F,$7B,$F8,$BE,$3F,$62,$73,$18,$8F,$F1,$EC
	.dc.b	$EE,$FF,$93,$36,$3F,$C7,$5F,$13,$FE,$DA,$9F,$E5,$4D,$8C,$F8,$DC
	.dc.b	$DF,$F1,$B4,$65,$FE,$87,$F6,$5C,$F8,$3E,$93,$6F,$FC,$9B,$5C,$E7
	.dc.b	$38,$FC,$1F,$B2,$7D,$65,$31,$95,$C3,$E7,$E8,$FC,$2D,$F5,$BE,$38
	.dc.b	$C3,$63,$33,$FD,$0D,$BA,$FF,$6F,$FB,$3F,$F0,$7C,$FD,$77,$EC,$C7
	.dc.b	$D6,$3F,$B2,$3A,$71,$C4,$8D,$EB,$F8,$FF,$B3,$7C,$9F,$DB,$6E,$1F
	.dc.b	$83,$7F,$A1,$36,$3E,$D7,$81,$E0,$C7,$EC,$35,$FF,$39,$B5,$FF,$32
	.dc.b	$37,$E8,$7F,$CC,$DB,$FE,$55,$FF,$34,$6F,$77,$F9,$78,$FE,$5A,$72
	.dc.b	$FC,$AF,$C6,$7F,$CC,$2D,$F8,$3E,$5D,$DF,$E5,$D8,$C1,$F4,$3F,$19
	.dc.b	$7D,$77,$CF,$6C,$7F,$6F,$6E,$BF,$E8,$36,$BF,$93,$F1,$7F,$01,$3F
	.dc.b	$C5,$B2,$FE,$A9,$F5,$5F,$9B,$9C,$D2,$73,$5E,$9F,$C2,$FF,$C9,$AF
	.dc.b	$FB,$AA,$11,$89,$FE,$2E,$71,$70,$FF,$1A,$8D,$FB,$BF,$87,$F8,$07
	.dc.b	$C8,$52,$61,$FE,$6C,$7F,$3D,$4F,$BD,$FB,$38,$76,$8C,$79,$BA,$9C
	.dc.b	$CF,$76,$BE,$FF,$C2,$FF,$CF,$98,$7A,$5B,$1C,$DA,$F9,$DA,$FF,$AD
	.dc.b	$3E,$21,$EE,$FE,$7D,$B5,$FE,$90,$8C,$1E,$8F,$AB,$CD,$79,$DA,$FC
	.dc.b	$36,$BF,$4D,$AF,$FA,$C7,$E3,$BF,$04,$FF,$AF,$5A,$FF,$0D,$6B,$FE
	.dc.b	$C8,$6B,$C7,$E9,$BF,$69,$06,$92,$FF,$D9,$5F,$51,$A2,$38,$3F,$D9
	.dc.b	$DF,$5B,$C8,$8E,$0F,$F6,$96,$3D,$88,$E0,$FF,$6B,$7C,$5F,$88,$8E
	.dc.b	$0F,$C3,$E2,$F9,$37,$E4,$3F,$DA,$A7,$F9,$EC,$62,$7F,$5D,$F9,$DB
	.dc.b	$EC,$3F,$A2,$4E,$6B,$FA,$1D,$3F,$0C,$FE,$74,$F6,$6B,$F9,$66,$BF
	.dc.b	$DB,$9A,$F8,$F0,$FE,$CB,$E3,$7F,$94,$AF,$CB,$E4,$7F,$53,$AE,$DE
	.dc.b	$87,$8A,$F7,$8F,$EB,$73,$79,$9E,$8F,$4F,$D2,$D9,$FE,$09,$3F,$D8
	.dc.b	$F0,$FE,$41,$D2,$8F,$F3,$6C,$EA,$FE,$0F,$BF,$F0,$BE,$32,$4F,$2F
	.dc.b	$5A,$E3,$CB,$07,$DB,$FC,$83,$3C,$1F,$E6,$CC,$FB,$47,$A7,$E7,$66
	.dc.b	$3E,$4F,$2E,$E3,$FC,$CF,$EF,$DF,$86,$7F,$6B,$8F,$F0,$AC,$63,$DD
	.dc.b	$FE,$A8,$DB,$FB,$AD,$9F,$E0,$BC,$BF,$7B,$E0,$FE,$DF,$A1,$F2,$5B
	.dc.b	$7F,$13,$E9,$BE,$FF,$C7,$FE,$01,$B1,$F2,$23,$F0,$7E,$76,$31,$F8
	.dc.b	$0F,$E1,$BE,$7A,$57,$CE,$FA,$AB,$5F,$17,$DA,$FC,$59,$8E,$4F,$89
	.dc.b	$CB,$FD,$59,$FE,$F7,$E9,$7F,$10,$6F,$97,$F9,$2F,$82,$FC,$14,$E7
	.dc.b	$2F,$F1,$4F,$D5,$FF,$75,$CB,$FC,$9C,$D9,$34,$C7,$D6,$72,$39,$B6
	.dc.b	$BF,$0B,$C3,$FA,$03,$F4,$3F,$71,$63,$F1,$3D,$3F,$A4,$31,$8F,$E0
	.dc.b	$5F,$CA,$74,$FE,$57,$B6,$D7,$EB,$3C,$1F,$ED,$9B,$FE,$75,$B1,$FE
	.dc.b	$5A,$E8,$FD,$9F,$77,$18,$E0,$FE,$87,$C3,$6B,$4F,$DA,$FD,$2D,$AF
	.dc.b	$D9,$CB,$6F,$88,$D8,$FF,$7B,$C6,$0C,$F1,$75,$3B,$36,$3F,$4F,$3F
	.dc.b	$DB,$FE,$D1,$F4,$5F,$C1,$23,$0C,$9F,$D6,$F0,$D3,$F4,$87,$1F,$A2
	.dc.b	$E5,$F4,$36,$7F,$BB,$61,$FC,$C3,$83,$F4,$CD,$8F,$ED,$0B,$E5,$78
	.dc.b	$C7,$07,$FC,$89,$F5,$EF,$AB,$FF,$12,$36,$7F,$DC,$8D,$CB,$FB,$3F
	.dc.b	$C0,$A3,$11,$FD,$9F,$0F,$E7,$15,$E4,$F4,$FA,$E1,$F0,$5F,$85,$E9
	.dc.b	$49,$F4,$FA,$77,$0F,$F1,$56,$5D,$7B,$3C,$64,$AF,$1B,$E6,$EE,$91
	.dc.b	$E6,$6D,$7F,$6F,$CF,$FA,$DE,$DF,$13,$D3,$F7,$D7,$D3,$7E,$0F,$C1
	.dc.b	$FC,$02,$61,$F0,$1F,$EF,$77,$FD,$C2,$FE,$6F,$A9,$C2,$BE,$8B,$F2
	.dc.b	$BE,$82,$47,$D4,$F8,$0B,$9F,$B3,$89,$5F,$83,$FA,$DC,$E3,$F6,$1F
	.dc.b	$A7,$EF,$FD,$22,$63,$F4,$9F,$1F,$C8,$7E,$2C,$DC,$7E,$09,$FC,$77
	.dc.b	$C3,$48,$FC,$2F,$C1,$DC,$FF,$9B,$EC,$F8,$BF,$69,$2B,$FC,$4D,$19
	.dc.b	$6F,$CC,$7E,$64,$FC,$D4,$D8,$FE,$BF,$F6,$4C,$3F,$AE,$B6,$33,$96
	.dc.b	$27,$3A,$7F,$A0,$FC,$55,$A3,$0D,$BF,$ED,$F3,$A9,$FA,$7F,$6B,$FA
	.dc.b	$BF,$B5,$FD,$63,$6B,$F4,$71,$FA,$FD,$31,$AF,$C0,$7C,$BC,$92,$BF
	.dc.b	$53,$EB,$DA,$FF,$3D,$65,$E4,$E7,$F1,$92,$BC,$7A,$7F,$88,$3D,$A7
	.dc.b	$D8,$BE,$87,$F7,$D8,$7F,$9B,$63,$17,$FB,$68,$C7,$CF,$A3,$FC,$11
	.dc.b	$97,$F9,$76,$BF,$D7,$0F,$FA,$D3,$0E,$B5,$B8,$C4,$EA,$BF,$CA,$CF
	.dc.b	$F8,$21,$BF,$03,$B6,$9D,$A7,$FC,$CD,$3B,$7F,$DE,$FF,$0A,$FA,$77
	.dc.b	$95,$FE,$1E,$94,$E6,$30,$D8,$FF,$11,$B5,$FF,$35,$AF,$9F,$FF,$13
	.dc.b	$46,$9A,$E9,$AF,$89,$AF,$9B,$C7,$3E,$9D,$3E,$7C,$62,$3E,$1B,$72
	.dc.b	$F8,$6D,$8F,$99,$F6,$75,$F8,$2D,$AF,$B2,$DC,$CD,$F5,$BB,$77,$FF
	.dc.b	$84,$65,$E2,$E3,$2F,$B0,$DC,$7E,$6D,$DE,$1F,$E9,$FC,$5F,$2D,$B3
	.dc.b	$F9,$CF,$47,$F1,$0D,$9D,$7E,$2D,$F1,$B3,$FF,$1D,$37,$2F,$EB,$CD
	.dc.b	$AF,$B7,$E3,$F8,$76,$FC,$9F,$D9,$DD,$BF,$D9,$B5,$FE,$BB,$6B,$FE
	.dc.b	$38,$51,$BF,$17,$83,$27,$84,$7A,$F0,$75,$79,$89,$3F,$B6,$8C,$B7
	.dc.b	$B7,$81,$1F,$F3,$2C,$3B,$7C,$DF,$63,$27,$CC,$FA,$7C,$69,$1F,$EB
	.dc.b	$9C,$D1,$F9,$5D,$3F,$62,$D8,$CC,$EF,$E4,$FE,$11,$1F,$85,$F4,$BF
	.dc.b	$17,$E4,$64,$7C,$57,$ED,$79,$BF,$D7,$6A,$F9,$67,$F4,$91,$9A,$FE
	.dc.b	$CB,$18,$9D,$47,$F6,$8E,$CF,$A4,$47,$F3,$7C,$BF,$BB,$BF,$ED,$53
	.dc.b	$D8,$93,$AE,$5F,$D0,$CE,$A3,$ED,$77,$FD,$DF,$99,$91,$F7,$5F,$3C
	.dc.b	$48,$FE,$F2,$F9,$48,$FC,$BF,$AF,$FC,$31,$1F,$00,$AF,$EF,$D1,$F0
	.dc.b	$8F,$71,$23,$FC,$0A,$FC,$4F,$AE,$FF,$4F,$47,$FA,$9A,$51,$88,$FF
	.dc.b	$04,$B5,$FD,$D4,$6B,$E4,$3E,$E7,$E4,$9B,$D3,$FD,$00,$DC,$FF,$81
	.dc.b	$F1,$A3,$18,$7F,$1C,$9F,$D1,$47,$FA,$6A,$3D,$73,$BF,$90,$C9,$F7
	.dc.b	$B2,$7C,$7D,$3D,$44,$AF,$54,$69,$FF,$60,$D7,$87,$E4,$3B,$D4,$9E
	.dc.b	$F8,$27,$C7,$F0,$1E,$46,$4E,$A3,$CF,$F2,$1E,$8F,$83,$93,$CA,$3D
	.dc.b	$7F,$09,$A8,$C1,$D9,$FE,$20,$BF,$07,$77,$E1,$B1,$FE,$CE,$6A,$E0
	.dc.b	$FE,$98,$F1,$EE,$E0,$7F,$41,$70,$7F,$43,$9F,$DC,$36,$35,$0F,$4B
	.dc.b	$C0,$49,$F8,$7F,$1F,$FC,$8F,$F1,$86,$B7,$59,$5F,$EE,$89,$CB,$76
	.dc.b	$E1,$52,$30,$7C,$EF,$90,$FB,$E4,$7E,$8F,$C8,$7E,$08,$8E,$0F,$AB
	.dc.b	$F2,$1F,$23,$ED,$1C,$34,$93,$FE,$F4,$8F,$F6,$9B,$78,$BE,$24,$62
	.dc.b	$3F,$A9,$1F,$D5,$9C,$38,$6B,$FD,$5D,$69,$E1,$24,$7E,$23,$E4,$3F
	.dc.b	$33,$F2,$32,$3F,$16,$FE,$73,$E2,$E4,$7E,$39,$AF,$FB,$CB,$99,$F2
	.dc.b	$3F,$A2,$F4,$3F,$47,$F1,$92,$BF,$2B,$1F,$A6,$F8,$FD,$46,$0F,$CC
	.dc.b	$BF,$D6,$7F,$1A,$79,$5F,$AC,$F8,$F9,$5F,$C1,$BD,$55,$6A,$3F,$84
	.dc.b	$FC,$9B,$F0,$C7,$2F,$F8,$2E,$30,$D8,$CC,$FF,$BC,$1A,$FF,$BC,$2B
	.dc.b	$F9,$DB,$FE,$34,$6F,$7F,$F8,$8F,$0F,$0F,$A1,$FC,$03,$DD,$FE,$03
	.dc.b	$70,$FD,$99,$FD,$97,$DD,$7C,$85,$24,$7F,$06,$8C,$B7,$E4,$78,$7C
	.dc.b	$8F,$1F,$91,$93,$B8,$E9,$F2,$3A,$8C,$1C,$DF,$95,$EC,$48,$F9,$1B
	.dc.b	$23,$F7,$3F,$73,$F0,$37,$FD,$9B,$F0,$AB,$F1,$AD,$7F,$C4,$0D,$7F
	.dc.b	$C4,$AD,$DF,$E9,$FB,$BF,$64,$93,$ED,$F9,$1F,$E4,$1D,$1F,$CD,$C9
	.dc.b	$1F,$C7,$9F,$E7,$5D,$12,$3F,$90,$B5,$FE,$73,$CF,$FC,$FC,$8F,$F2
	.dc.b	$67,$FA,$17,$1C,$8F,$E5,$2D,$7F,$A0,$9F,$D1,$C8,$7F,$2E,$3F,$B6
	.dc.b	$FC,$72,$69,$23,$FA,$64,$65,$FE,$67,$F8,$4B,$DC,$FF,$02,$36,$33
	.dc.b	$18,$3F,$BC,$FE,$08,$F3,$29,$27,$FA,$C4,$65,$FE,$7F,$F3,$BF,$E4
	.dc.b	$46,$FC,$3F,$F8,$46,$31,$1F,$C0,$BD,$44,$62,$3F,$CA,$D0,$FC,$27
	.dc.b	$FA,$24,$7F,$4A,$6B,$C5,$AF,$06,$BF,$E0,$2E,$D7,$C5,$7F,$84,$9B
	.dc.b	$1F,$DB,$DA,$FF,$BD,$9A,$F8,$BE,$11,$D9,$FE,$29,$6C,$7F,$5C,$6B
	.dc.b	$EB,$6B,$FD,$DB,$ED,$7F,$C3,$AF,$F4,$9F,$6F,$FB,$9F,$1F,$F9,$92
	.dc.b	$3F,$B7,$B5,$FF,$7D,$CF,$0E,$AF,$F1,$07,$94,$E4,$FF,$15,$B6,$BF
	.dc.b	$C1,$0D,$7F,$96,$FB,$A8,$C4,$7F,$82,$3C,$7F,$F3,$4B,$E3,$3F,$C3
	.dc.b	$0E,$2F,$FC,$5C,$DC,$BF,$C1,$3F,$8A,$FF,$17,$AE,$1C,$2F,$AE,$FF
	.dc.b	$36,$37,$6F,$F3,$64,$7F,$97,$27,$6F,$F9,$C7,$37,$DA,$BC,$FF,$F2
	.dc.b	$2B,$6B,$FC,$44,$D7,$D6,$D7,$FD,$35,$C1,$FE,$C9,$F9,$8F,$F4,$53
	.dc.b	$63,$FC,$66,$D7,$DC,$FF,$B3,$FE,$4F,$7F,$6A,$F6,$2F,$64,$BE,$33
	.dc.b	$97,$FC,$7B,$9B,$F0,$7F,$EA,$B1,$FE,$3F,$6C,$67,$D7,$FF,$2F,$C6
	.dc.b	$B2,$7C,$9B,$CB,$F3,$B7,$0F,$3B,$5F,$3A,$F3,$3D,$BF,$6D,$FE,$07
	.dc.b	$9F,$6B,$5F,$6B,$7A,$3F,$CC,$DF,$65,$FE,$D5,$9F,$4B,$5F,$C2,$B5
	.dc.b	$FF,$25,$B5,$FF,$34,$B5,$FF,$2E,$7D,$B7,$F8,$39,$FC,$6B,$5F,$F2
	.dc.b	$9B,$5F,$F3,$8B,$5F,$F7,$2F,$E1,$9F,$E4,$D7,$E9,$B6,$BF,$DE,$D9
	.dc.b	$A3,$1D,$5F,$ED,$A8,$C7,$C4,$FD,$EF,$B5,$FF,$58,$E1,$F7,$BE,$D9
	.dc.b	$E6,$7E,$09,$B1,$F8,$29,$DB,$F8,$26,$E5,$E0,$6E,$3D,$DB,$3F,$EF
	.dc.b	$C7,$FD,$83,$3F,$E4,$0D,$1F,$3B,$FE,$43,$A7,$DB,$3D,$FF,$F7,$75
	.dc.b	$3E,$87,$FD,$E1,$39,$AF,$F7,$AC,$E5,$FF,$7B,$C6,$54,$76,$C3,$37
	.dc.b	$D3,$0F,$05,$EE,$B9,$5F,$8A,$BC,$38,$78,$1E,$18,$72,$9F,$0E,$1E
	.dc.b	$9F,$03,$95,$A6,$46,$1B,$7C,$9F,$F6,$DB,$71,$D4,$78,$4F,$6D,$78
	.dc.b	$5E,$F8,$78,$5F,$6E,$1E,$AB,$F3,$C6,$71,$F5,$E1,$ED,$AD,$4F,$23
	.dc.b	$FC,$92,$72,$BA,$D3,$35,$F2,$5B,$F4,$1F,$C5,$DB,$E0,$F0,$94,$E6
	.dc.b	$7F,$CA,$B1,$89,$FF,$3E,$43,$E9,$69,$1F,$F4,$03,$71,$FB,$50,$E2
	.dc.b	$6F,$F0,$63,$E7,$55,$9C,$C7,$C9,$3E,$91,$DC,$C1,$CE,$79,$39,$AE
	.dc.b	$F3,$C9,$F0,$C7,$1D,$B3,$5F,$6A,$3F,$CE,$37,$5F,$D7,$45,$7C,$A7
	.dc.b	$82,$FA,$6B,$F4,$2B,$D1,$5E,$26,$D7,$98,$F2,$4F,$F9,$FA,$B8,$C7
	.dc.b	$C3,$F8,$8F,$F6,$4D,$FF,$36,$6C,$7F,$36,$5F,$21,$F1,$36,$FE,$0B
	.dc.b	$1A,$7F,$83,$FD,$B7,$FA,$E2,$FF,$9F,$B6,$F9,$7C,$87,$19,$FE,$15
	.dc.b	$39,$AF,$E1,$FF,$6C,$F9,$DD,$61,$F5,$FF,$46,$7B,$B8,$D5,$7B,$CB
	.dc.b	$FA,$6D,$69,$B5,$FE,$C4,$BF,$E3,$13,$97,$E1,$35,$FE,$B2,$D7,$F6
	.dc.b	$1E,$1B,$E9,$5E,$27,$39,$4E,$63,$0D,$C3,$F4,$6D,$9F,$A1,$F6,$CF
	.dc.b	$98,$F5,$36,$BF,$AD,$E1,$D9,$FE,$CF,$19,$AF,$EC,$FC,$59,$6F,$C6
	.dc.b	$FC,$38,$C7,$07,$F3,$87,$FD,$6B,$7F,$D9,$BE,$DB,$FB,$E3,$E4,$6D
	.dc.b	$FF,$85,$1A,$FE,$7B,$34,$62,$3F,$A0,$7D,$B3,$97,$E9,$36,$FF,$BE
	.dc.b	$BF,$83,$77,$1F,$6A,$FF,$98,$3F,$67,$ED,$9F,$49,$9F,$03,$96,$3F
	.dc.b	$E0,$49,$CE,$1F,$DF,$A7,$F2,$73,$E5,$6B,$F5,$9A,$FD,$46,$E1,$FE
	.dc.b	$C0,$F9,$DF,$17,$AC,$7F,$1B,$EA,$FD,$AE,$4C,$C7,$9F,$ED,$9E,$E6
	.dc.b	$BB,$FF,$C9,$FA,$7F,$84,$27,$2F,$F3,$5E,$3F,$E0,$F8,$FF,$13,$F7
	.dc.b	$3E,$07,$DB,$7F,$86,$EB,$F0,$6D,$DF,$FC,$86,$D7,$E8,$35,$FC,$8F
	.dc.b	$8F,$F6,$9F,$F2,$26,$FF,$D7,$B9,$FD,$3E,$1B,$D6,$F9,$D7,$BD,$3F
	.dc.b	$8C,$E0,$FD,$4F,$A3,$FD,$2D,$AE,$7D,$7F,$CE,$37,$4F,$E7,$3F,$4F
	.dc.b	$FE,$25,$6F,$AB,$A8,$C4,$FF,$40,$6B,$FE,$86,$F8,$2F,$B2,$FF,$A2
	.dc.b	$DA,$E7,$AB,$F5,$5C,$5F,$E8,$E7,$F5,$9F,$26,$FC,$AF,$FC,$13,$97
	.dc.b	$F4,$C8,$C7,$A5,$A9,$CE,$1F,$CD,$04,$EA,$BF,$63,$F6,$AF,$B0,$FD
	.dc.b	$AB,$63,$31,$83,$C7,$F2,$79,$BF,$F5,$13,$5F,$F5,$0B,$5F,$EE,$99
	.dc.b	$A3,$10,$8C,$7D,$77,$42,$5F,$F6,$68,$CB,$FC,$1B,$E7,$7E,$4D,$7F
	.dc.b	$10,$6B,$FC,$39,$BD,$07,$35,$E1,$E6,$F4,$92,$FB,$63,$F8,$56,$FC
	.dc.b	$0F,$B8,$8E,$0F,$13,$FC,$6A,$CD,$24,$74,$FA,$FF,$8E,$47,$07,$D7
	.dc.b	$DD,$F3,$DF,$CE,$B4,$7E,$7C,$EB,$FE,$7B,$C5,$F7,$EB,$53,$9C,$3F
	.dc.b	$58,$7F,$70,$FB,$57,$D8,$7E,$59,$B1,$FE,$0E,$6E,$3F,$E1,$26,$BF
	.dc.b	$85,$AF,$E9,$D3,$FD,$E9,$B1,$FE,$2D,$9E,$3D,$9F,$A8,$9C,$F0,$FD
	.dc.b	$DF,$6F,$F8,$3F,$0F,$F9,$27,$1F,$2E,$3F,$7E,$3F,$C2,$F1,$8A,$FC
	.dc.b	$27,$91,$C0,$E3,$DF,$67,$55,$FE,$58,$6F,$C5,$F6,$3B,$FA,$16,$37
	.dc.b	$F4,$9E,$23,$19,$7F,$B0,$2D,$1F,$DA,$3D,$47,$E2,$58,$BF,$3F,$1B
	.dc.b	$E9,$DF,$25,$BA,$D6,$5F,$ED,$1A,$D4,$FD,$6A,$DD,$FE,$CD,$B5,$FB
	.dc.b	$1A,$DC,$E6,$BF,$B3,$61,$E2,$D3,$B1,$CE,$2C,$FC,$2C,$3E,$6F,$37
	.dc.b	$CE,$6C,$7C,$68,$FA,$3F,$51,$8C,$BE,$03,$FC,$EB,$C1,$FF,$01,$47
	.dc.b	$FB,$89,$BC,$BF,$CD,$37,$FF,$00,$3F,$CD,$A3,$38,$7E,$E9,$FF,$61
	.dc.b	$BF,$65,$B5,$CF,$DB,$ED,$F1,$3D,$FE,$4B,$F3,$9E,$7B,$EA,$F1,$8F
	.dc.b	$E1,$30,$FF,$2A,$46,$BE,$5B,$F9,$6D,$1A,$4A,$FF,$33,$C3,$EB,$7A
	.dc.b	$9C,$63,$A4,$65,$F3,$70,$FC,$13,$E1,$77,$58,$F5,$7D,$8F,$C5,$31
	.dc.b	$97,$C5,$7F,$01,$E6,$FE,$07,$D2,$B5,$F9,$CE,$7F,$C4,$C6,$63,$1D
	.dc.b	$2F,$A2,$C6,$2E,$4F,$F3,$3E,$1F,$90,$9F,$A9,$E6,$FE,$4F,$ED,$9F
	.dc.b	$3B,$8C,$BB,$46,$1F,$CA,$FD,$B7,$F1,$07,$FA,$03,$72,$FC,$73,$5F
	.dc.b	$F3,$9F,$72,$31,$1F,$EA,$7C,$97,$FE,$C1,$BE,$FF,$11,$FC,$42,$FA
	.dc.b	$69,$E8,$AD,$61,$F8,$9F,$4F,$F1,$9F,$41,$F8,$D6,$E3,$F8,$58,$FC
	.dc.b	$87,$DA,$E2,$BF,$20,$FE,$4F,$ED,$3F,$29,$EE,$5B,$FC,$63,$87,$F1
	.dc.b	$DF,$49,$6F,$F3,$0F,$F8,$9F,$F5,$5F,$EC,$F1,$FE,$52,$6B,$FE,$50
	.dc.b	$6F,$AA,$73,$E6,$7F,$BC,$3E,$07,$F4,$87,$3E,$5B,$8F,$E7,$7B,$BF
	.dc.b	$51,$DF,$FA,$96,$C7,$E8,$63,$FB,$AF,$89,$F9,$78,$FE,$B0,$DC,$FF
	.dc.b	$B8,$63,$FE,$FE,$AF,$F3,$0E,$4D,$A9,$30,$FF,$31,$C6,$5E,$7E,$0F
	.dc.b	$EE,$4F,$81,$36,$3F,$96,$EB,$ED,$DB,$A9,$23,$9B,$E6,$F8,$8F,$E0
	.dc.b	$47,$C9,$6C,$7E,$EF,$EC,$BF,$83,$1F,$C1,$FE,$0A,$C7,$6C,$9C,$3F
	.dc.b	$44,$CE,$DF,$E3,$08,$CC,$F9,$30,$72,$FF,$A0,$1B,$1F,$E8,$06,$D7
	.dc.b	$F8,$57,$34,$62,$3F,$D9,$7B,$FF,$87,$AB,$F8,$81,$FE,$67,$9E,$3C
	.dc.b	$FF,$C5,$E3,$FC,$AF,$F8,$4F,$FA,$C9,$FE,$9C,$BE,$57,$3F,$15,$A7
	.dc.b	$7F,$21,$FE,$2E,$9F,$F3,$2E,$26,$93,$E2,$7A,$9C,$30,$EA,$FF,$4B
	.dc.b	$F0,$7C,$71,$FA,$D7,$FC,$7B,$97,$F0,$07,$3A,$7F,$AD,$9C,$CE,$FC
	.dc.b	$6F,$E8,$70,$69,$23,$FC,$49,$87,$F9,$65,$FF,$16,$B6,$BF,$C5,$8B
	.dc.b	$EC,$3E,$1F,$C3,$7C,$4F,$6E,$5F,$C4,$F8,$9B,$87,$F1,$58,$E7,$F0
	.dc.b	$DF,$2B,$E7,$7E,$5B,$63,$F9,$06,$8F,$43,$CE,$48,$FF,$27,$5F,$B6
	.dc.b	$3E,$67,$C1,$7D,$43,$EA,$FD,$65,$AF,$CC,$FD,$37,$F9,$2D,$FF,$56
	.dc.b	$36,$FF,$D9,$AD,$7F,$D9,$6D,$E8,$2F,$9D,$52,$BF,$CF,$B5,$A9,$CC
	.dc.b	$62,$B8,$56,$CF,$0C,$E9,$FF,$34,$7A,$FF,$ED,$3F,$94,$F6,$DF,$F9
	.dc.b	$6F,$C1,$46,$27,$53,$F0,$9A,$F2,$F0,$FF,$DE,$AE,$5A,$E7,$CE,$E5
	.dc.b	$18,$8F,$F7,$6F,$E7,$6F,$86,$F6,$47,$B2,$30,$FA,$4D,$C6,$B2,$FF
	.dc.b	$6B,$4E,$A7,$87,$C4,$65,$EA,$D7,$FD,$EF,$1F,$EA,$77,$A1,$88,$D7
	.dc.b	$6F,$E9,$7E,$93,$C7,$DB,$FE,$BC,$EC,$46,$31,$FC,$AB,$70,$FC,$AF
	.dc.b	$D1,$DD,$7F,$AA,$78,$5F,$31,$B3,$F8,$A6,$BF,$E9,$96,$BF,$C1,$30
	.dc.b	$7A,$7D,$73,$FD,$C7,$D7,$D9,$36,$71,$D5,$2F,$F8,$2B,$BF,$AF,$E1
	.dc.b	$F5,$52,$7F,$84,$BC,$FC,$FF,$47,$E1,$1F,$EC,$2D,$0D,$27,$07,$AD
	.dc.b	$FF,$75,$5F,$F8,$1D,$79,$DC,$8E,$7E,$77,$9F,$A6,$57,$F1,$87,$D4
	.dc.b	$C3,$F8,$09,$F9,$2E,$0F,$E2,$CD,$E0,$FC,$AB,$FE,$AD,$FD,$7F,$FC
	.dc.b	$24,$FF,$9B,$D7,$D8,$7C,$76,$C6,$7D,$BC,$7B,$B2,$38,$3F,$67,$C1
	.dc.b	$FE,$6B,$EC,$5F,$40,$FE,$92,$33,$5F,$CF,$63,$13,$A8,$FE,$BB,$DF
	.dc.b	$CF,$D2,$FF,$22,$47,$F4,$D8,$F2,$7B,$D2,$3D,$0D,$7F,$8B,$1E,$DF
	.dc.b	$13,$D5,$F5,$D2,$3D,$CF,$5E,$29,$1F,$DB,$1F,$36,$1F,$52,$3F,$A8
	.dc.b	$35,$FE,$9F,$B7,$FB,$8D,$F9,$1F,$27,$F1,$CF,$47,$34,$8F,$EE,$EF
	.dc.b	$42,$3F,$ED,$99,$7C,$9F,$D7,$6C,$66,$30,$67,$C3,$FA,$8C,$8F,$B1
	.dc.b	$1F,$5F,$E4,$E4,$7D,$B7,$EA,$7E,$4A,$F5,$69,$F4,$DC,$4F,$F3,$4E
	.dc.b	$34,$62,$3F,$B8,$35,$F8,$3F,$23,$F8,$5F,$26,$47,$E2,$63,$F1,$1F
	.dc.b	$21,$F8,$96,$47,$E3,$63,$F1,$9F,$21,$F8,$DF,$91,$93,$F9,$18,$FC
	.dc.b	$87,$C8,$6A,$30,$7E,$55,$FE,$09,$F8,$5F,$FB,$B2,$3F,$CD,$CD,$7F
	.dc.b	$03,$E1,$A3,$1E,$6F,$F3,$CF,$88,$7D,$A3,$AF,$F8,$84,$61,$FE,$72
	.dc.b	$DC,$3F,$AB,$B5,$FF,$5B,$4F,$17,$ED,$7C,$9F,$93,$92,$47,$EA,$E3
	.dc.b	$CF,$E0,$FA,$3C,$19,$3E,$88,$F5,$F8,$5A,$8C,$1E,$C7,$F6,$5F,$27
	.dc.b	$23,$F6,$CF,$ED,$7E,$4E,$47,$EE,$5A,$FE,$8B,$C6,$76,$3E,$4F,$F9
	.dc.b	$74,$B8,$FF,$87,$BF,$EA,$F6,$BF,$AA,$35,$E4,$F0,$8F,$D2,$71,$7F
	.dc.b	$66,$9E,$5E,$17,$F0,$C9,$FD,$34,$7F,$52,$6B,$FD,$BD,$B8,$F9,$39
	.dc.b	$7A,$7A,$D2,$BF,$89,$47,$F3,$7F,$5D,$F2,$74,$93,$FD,$E7,$2F,$EE
	.dc.b	$F8,$7F,$7E,$F6,$F8,$12,$75,$E2,$7F,$7F,$9D,$46,$03,$FE,$06,$F9
	.dc.b	$3F,$07,$C9,$78,$7E,$4A,$56,$EB,$FC,$17,$3E,$5F,$85,$DC,$E7,$44
	.dc.b	$E6,$3F,$D8,$5B,$3D,$EF,$EB,$8D,$8C,$C7,$E4,$BE,$EB,$11,$FC,$6B
	.dc.b	$B9,$18,$8F,$F5,$BB,$5F,$F0,$D4,$6A,$7F,$33,$39,$97,$0A,$DF,$03
	.dc.b	$DE,$E5,$79,$A3,$E8,$7C,$74,$8F,$63,$F4,$FD,$39,$1E,$F6,$BF,$8F
	.dc.b	$D3,$FC,$F6,$FC,$3F,$8C,$FB,$5F,$21,$23,$07,$C7,$F5,$7E,$F4,$C8
	.dc.b	$C1,$FF,$00,$77,$7A,$7F,$98,$6C,$66,$3F,$37,$F6,$7E,$C8,$FB,$81
	.dc.b	$C3,$FE,$50,$6B,$FE,$26,$35,$A7,$F6,$48,$CC,$7C,$CF,$AA,$FF,$42
	.dc.b	$3D,$3F,$13,$FF,$38,$37,$93,$F8,$4C,$EA,$3E,$17,$C9,$FF,$71,$D8
	.dc.b	$D2,$47,$F2,$FF,$47,$FC,$8A,$DF,$6F,$F9,$A1,$1F,$F3,$3F,$47,$F9
	.dc.b	$6B,$E4,$7A,$12,$33,$87,$FA,$C1,$BF,$27,$E1,$23,$83,$F9,$F7,$CD
	.dc.b	$FF,$98,$DB,$EE,$7F,$75,$8C,$47,$FA,$BD,$AF,$FA,$BA,$9F,$29,$ED
	.dc.b	$24,$78,$9F,$F3,$0E,$4F,$D4,$7E,$5C,$FF,$98,$9B,$19,$E2,$DC,$6B
	.dc.b	$D9,$7D,$AD,$24,$FF,$82,$5E,$7E,$03,$F0,$5F,$F6,$1E,$3F,$CD,$9A
	.dc.b	$FD,$D8,$D7,$57,$F3,$C9,$CC,$FF,$B2,$3E,$DB,$FD,$44,$FC,$E6,$BF
	.dc.b	$EF,$46,$BB,$F9,$5F,$F1,$4F,$EF,$7F,$E1,$48,$FF,$4D,$B5,$FC,$63
	.dc.b	$5F,$F3,$9B,$5F,$F4,$87,$EF,$9F,$B5,$4D,$27,$85,$FE,$27,$FB,$8E
	.dc.b	$DE,$5F,$FA,$53,$74,$63,$CA,$FF,$4D,$46,$3E,$1B,$FC,$BD,$C6,$F1
	.dc.b	$BF,$C6,$9F,$22,$F6,$7F,$C6,$4D,$D7,$FC,$00,$D7,$FD,$DC,$D7,$FD
	.dc.b	$17,$87,$FB,$F3,$CE,$FF,$35,$7D,$C3,$B6,$9F,$E4,$1E,$EF,$F6,$05
	.dc.b	$FF,$2D,$B7,$E1,$1F,$E2,$78,$FF,$1A,$C6,$3E,$23,$FC,$91,$39,$E1
	.dc.b	$FE,$77,$5F,$D0,$0E,$5F,$73,$FE,$76,$7F,$C2,$CD,$7F,$75,$68,$C4
	.dc.b	$7A,$1A,$FC,$67,$FD,$06,$FF,$41,$DF,$FA,$6F,$C3,$7F,$A4,$5B,$1E
	.dc.b	$0F,$91,$C7,$DC,$BE,$62,$92,$79,$47,$F3,$8C,$BE,$B7,$DC,$BE,$6B
	.dc.b	$FC,$C0,$DA,$FF,$17,$35,$F4,$B5,$FF,$5C,$E2,$F6,$3E,$27,$47,$FB
	.dc.b	$5E,$31,$E8,$79,$E7,$F9,$1D,$7F,$98,$27,$87,$0F,$AF,$EE,$5F,$03
	.dc.b	$49,$5F,$57,$ED,$BF,$BD,$E5,$F8,$A7,$FB,$BF,$19,$A4,$F7,$BF,$7B
	.dc.b	$E8,$BE,$82,$93,$0F,$F7,$EE,$89,$D4,$61,$FF,$26,$70,$FF,$BB,$A7
	.dc.b	$55,$F4,$67,$75,$EB,$77,$97,$C1,$FB,$87,$CB,$FD,$76,$D7,$DB,$6B
	.dc.b	$F0,$FE,$E3,$F8,$1E,$5D,$DB,$7F,$2D,$B3,$FE,$F6,$6E,$45,$ED,$7F
	.dc.b	$8A,$35,$7C,$83,$16,$0F,$B7,$FC,$6B,$F2,$7E,$47,$FD,$E7,$18,$B4
	.dc.b	$63,$AF,$FD,$EF,$1D,$2F,$9D,$6F,$0E,$D8,$75,$BD,$E1,$C6,$3C,$38
	.dc.b	$74,$78,$CF,$4A,$E3,$7C,$F0,$F3,$FC,$4F,$8F,$1F,$4E,$08,$C3,$AC
	.dc.b	$39,$55,$8E,$F8,$74,$9F,$1C,$78,$0F,$1D,$F9,$5F,$C5,$4E,$DE,$35
	.dc.b	$B7,$E3,$17,$1F,$53,$EF,$36,$53,$9C,$7F,$79,$72,$7F,$7B,$8F,$94
	.dc.b	$E5,$AE,$7E,$49,$F7,$1C,$63,$F3,$D1,$87,$83,$5E,$11,$FE,$E8,$5F
	.dc.b	$55,$48,$FF,$BB,$1B,$9F,$E0,$A1,$BB,$83,$E5,$47,$D0,$AA,$EA,$2B
	.dc.b	$D8,$FA,$9A,$FF,$85,$EB,$94,$60,$FF,$0D,$57,$28,$CD,$F8,$67,$93
	.dc.b	$E3,$8C,$D7,$C9,$AE,$17,$E0,$6D,$F5,$3C,$C5,$3C,$C5,$3A,$CF,$96
	.dc.b	$79,$46,$EB,$BB,$F0,$BE,$55,$F4,$DF,$E8,$98,$FE,$39,$18,$AF,$E1
	.dc.b	$2D,$8F,$80,$D7,$F9,$2F,$DC,$7F,$1C,$9F,$E1,$ED,$8F,$8C,$D7,$7D
	.dc.b	$5D,$DF,$A7,$F6,$FF,$9F,$AF,$E2,$ED,$8F,$9C,$D7,$F9,$9B,$5F,$ED
	.dc.b	$FF,$84,$7F,$61,$9F,$F7,$43,$63,$FA,$9B,$5F,$ED,$2D,$7F,$BB,$FC
	.dc.b	$CB,$EA,$7C,$9D,$C8,$C5,$FF,$BC,$DB,$5F,$D7,$9A,$FF,$46,$FB,$87
	.dc.b	$A5,$ED,$DD,$18,$BF,$E6,$6D,$AF,$C5,$35,$FE,$9F,$F4,$AF,$63,$93
	.dc.b	$6B,$F9,$DB,$67,$F2,$4D,$7F,$C0,$BF,$71,$FE,$9D,$9F,$DA,$66,$8C
	.dc.b	$4F,$F7,$66,$BC,$3D,$7F,$47,$DC,$30,$A4,$C3,$FA,$5C,$65,$FE,$9B
	.dc.b	$F7,$0F,$8A,$F8,$0D,$8C,$C6,$0F,$9B,$DF,$C7,$1F,$F0,$3B,$5F,$F0
	.dc.b	$EB,$5F,$F1,$5B,$56,$6F,$4F,$FD,$55,$7E,$E6,$CF,$F8,$49,$AF,$F8
	.dc.b	$A1,$AF,$F8,$DB,$EE,$3F,$A5,$CF,$C6,$6C,$7F,$86,$9A,$FF,$8C,$1A
	.dc.b	$FF,$8F,$BE,$E3,$FC,$19,$3F,$39,$B1,$FE,$22,$6B,$FE,$38,$6B,$FE
	.dc.b	$46,$F0,$9D,$5F,$90,$6D,$7D,$86,$CF,$F8,$B1,$AF,$FA,$9F,$F6,$07
	.dc.b	$CB,$7E,$AE,$11,$8A,$FE,$58,$D8,$FE,$6C,$D7,$FC,$E9,$F5,$FC,$7E
	.dc.b	$14,$D2,$57,$F3,$5F,$B1,$7D,$37,$76,$C6,$63,$07,$EA,$BE,$21,$F5
	.dc.b	$D4,$93,$F9,$48,$CB,$FD,$07,$0F,$F0,$D3,$7D,$1F,$21,$1F,$F5,$3F
	.dc.b	$A9,$FB,$6F,$8E,$34,$91,$FE,$AC,$C7,$44,$70,$7F,$4F,$F9,$9F,$D8
	.dc.b	$6B,$FD,$C1,$AF,$F8,$29,$AF,$FB,$E3,$C0,$F0,$7D,$1C,$AF,$F5,$CB
	.dc.b	$67,$FD,$6D,$F4,$FE,$52,$4F,$FA,$F9,$FA,$D7,$FE,$B8,$8F,$17,$C8
	.dc.b	$60,$8F,$FB,$13,$E0,$3D,$C4,$70,$7F,$67,$3F,$C4,$25,$E5,$FC,$13
	.dc.b	$8C,$3F,$BA,$35,$FF,$77,$2F,$43,$64,$7A,$3C,$AA,$47,$76,$CF,$FB
	.dc.b	$DF,$E4,$DF,$25,$F2,$1B,$AF,$F3,$D6,$DF,$FB,$92,$31,$3F,$B2,$A5
	.dc.b	$72,$F9,$BC,$92,$FF,$D8,$5F,$13,$CF,$E3,$CD,$24,$7F,$B1,$FE,$BB
	.dc.b	$44,$70,$7F,$85,$77,$7E,$13,$FB,$AC,$91,$8F,$6F,$FA,$A3,$6B,$85
	.dc.b	$6F,$A1,$F9,$C7,$EA,$70,$FF,$27,$C6,$3B,$FF,$AE,$36,$FF,$C4,$8D
	.dc.b	$9F,$F1,$E9,$E0,$B7,$CE,$F8,$1B,$5E,$08,$C1,$9E,$E9,$1E,$46,$D7
	.dc.b	$4C,$BD,$BF,$21,$CC,$95,$EA,$7E,$8F,$C1,$C8,$FF,$1C,$71,$7D,$A8
	.dc.b	$C5,$FD,$DE,$37,$D1,$7D,$82,$5F,$08,$CB,$FE,$39,$FC,$F9,$F6,$9F
	.dc.b	$C4,$DB,$19,$8C,$19,$F1,$24,$71,$6E,$1F,$E4,$BA,$FE,$DA,$FF,$99
	.dc.b	$1A,$FF,$9F,$9A,$9C,$E5,$F9,$4A,$FC,$79,$0F,$41,$F9,$5C,$BF,$C3
	.dc.b	$55,$FE,$88,$94,$62,$3F,$2A,$D8,$CF,$B7,$F8,$6F,$69,$F4,$94,$98
	.dc.b	$7E,$3E,$32,$DF,$63,$81,$1C,$1B,$F0,$FA,$11,$C1,$FE,$65,$F7,$BF
	.dc.b	$B0,$37,$CE,$FE,$81,$18,$D3,$FC,$95,$1E,$5F,$8A,$95,$E3,$6C,$FE
	.dc.b	$E7,$B9,$EF,$FE,$B4,$95,$F8,$48,$CB,$7E,$C7,$F6,$44,$70,$7E,$2B
	.dc.b	$83,$FB,$B7,$07,$CA,$F5,$FE,$71,$27,$5E,$67,$D1,$23,$83,$FC,$2F
	.dc.b	$E6,$7E,$D1,$BD,$5F,$CF,$C8,$F7,$F4,$3E,$D1,$1C,$1E,$3F,$6B,$EF
	.dc.b	$11,$C1,$FE,$24,$F9,$AF,$F7,$E3,$7E,$33,$FC,$83,$18,$E0,$FD,$BF
	.dc.b	$AC,$8C,$47,$FA,$DA,$31,$E2,$70,$25,$F5,$7F,$1D,$EF,$C8,$F8,$CD
	.dc.b	$AE,$B6,$FA,$2F,$09,$2B,$CC,$FE,$5F,$BA,$47,$CF,$6D,$7F,$33,$E2
	.dc.b	$FD,$7B,$7B,$FF,$3C,$4A,$F8,$1E,$17,$B8,$8E,$0F,$F2,$3F,$E9,$FF
	.dc.b	$C9,$75,$34,$9E,$4F,$EA,$7E,$42,$57,$E0,$5B,$8F,$F1,$3E,$9F,$F7
	.dc.b	$9C,$FF,$31,$6C,$64,$FE,$5E,$E1,$FE,$E4,$BE,$37,$24,$9E,$4F,$EE
	.dc.b	$7E,$76,$47,$E4,$5B,$5E,$1F,$B2,$EA,$49,$F1,$3E,$0E,$C9,$1E,$56
	.dc.b	$CF,$CB,$F8,$2F,$E0,$C4,$9F,$CF,$3F,$D7,$7F,$41,$FF,$08,$C7,$F2
	.dc.b	$A6,$CF,$F6,$56,$BF,$EF,$CE,$E4,$62,$3F,$C1,$DF,$71,$FE,$79,$6F
	.dc.b	$57,$F5,$28,$C4,$7F,$56,$6F,$61,$FF,$21,$C7,$2E,$CE,$04,$C7,$F6
	.dc.b	$9F,$2D,$FC,$7C,$8E,$0F,$F4,$2F,$E8,$9F,$E4,$57,$F3,$8D,$A1,$FE
	.dc.b	$70,$70,$DE,$DF,$F7,$0C,$FE,$8F,$CC,$7D,$8E,$2B,$8F,$47,$F8,$1A
	.dc.b	$31,$1F,$E7,$A9,$DD,$7E,$06,$B4,$FE,$0E,$5F,$70,$F5,$FF,$DF,$31
	.dc.b	$FD,$A8,$CB,$87,$FB,$6C,$6F,$E2,$F5,$1A,$6C,$7F,$AD,$B7,$BE,$CF
	.dc.b	$F8,$0A,$BA,$4E,$5B,$19,$7F,$80,$7D,$C7,$F8,$A5,$F4,$35,$63,$AF
	.dc.b	$73,$E4,$C6,$0F,$87,$3B,$9F,$E1,$5E,$BB,$E1,$F7,$18,$6D,$7C,$A9
	.dc.b	$D3,$F1,$DA,$F5,$3F,$AB,$FC,$27,$C0,$E8,$FE,$B6,$DF,$13,$FC,$E7
	.dc.b	$3F,$CD,$E3,$FC,$C5,$D8,$FA,$DA,$49,$FF,$14,$46,$5F,$F1,$56,$DF
	.dc.b	$0D,$FF,$34,$35,$FF,$34,$35,$FF,$16,$B7,$0F,$D0,$FA,$5E,$5E,$CF
	.dc.b	$F4,$5D,$7E,$A6,$30,$79,$FD,$E9,$1F,$D3,$5A,$FF,$87,$FF,$32,$F5
	.dc.b	$B7,$3F,$F3,$E4,$62,$3F,$CF,$8D,$7F,$CF,$8F,$F9,$29,$FF,$22,$35
	.dc.b	$E9,$3C,$8F,$91,$EF,$C8,$FE,$BC,$D7,$C5,$E0,$78,$08,$F4,$7C,$7E
	.dc.b	$14,$8E,$ED,$7F,$9C,$7A,$5F,$54,$8F,$F6,$B7,$F3,$3D,$5F,$CC,$BE
	.dc.b	$AA,$F9,$26,$92,$BF,$AE,$7B,$DC,$3C,$FF,$F2,$6C,$60,$87,$F4,$7F
	.dc.b	$1F,$F0,$84,$7F,$BD,$3F,$87,$FB,$99,$1F,$DF,$5A,$FF,$58,$9E,$C4
	.dc.b	$7F,$B6,$3F,$8F,$E5,$91,$FE,$05,$6B,$FD,$83,$D7,$F3,$11,$FE,$E8
	.dc.b	$FA,$7E,$12,$47,$F7,$76,$BF,$D7,$FC,$BF,$CE,$11,$FF,$09,$3F,$9F
	.dc.b	$F0,$A4,$7F,$85,$5A,$FF,$70,$F0,$FE,$31,$1F,$F0,$23,$FA,$76,$1F
	.dc.b	$E1,$D2,$9F,$EA,$1F,$C3,$9F,$67,$FA,$F6,$C7,$F9,$B1,$AE,$E3,$F4
	.dc.b	$1F,$2D,$23,$FA,$43,$5F,$E9,$1E,$37,$E9,$48,$FF,$4D,$7F,$DB,$1F
	.dc.b	$6D,$FD,$96,$3F,$D3,$8D,$7F,$D0,$4D,$7F,$DB,$8D,$7F,$89,$FC,$07
	.dc.b	$F0,$77,$FD,$26,$D7,$3E,$2F,$F2,$87,$FC,$B2,$76,$DF,$C2,$47,$07
	.dc.b	$8B,$B7,$CA,$47,$07,$9B,$AB,$D2,$47,$07,$C7,$FB,$5F,$94,$47,$07
	.dc.b	$CB,$FB,$6F,$9C,$47,$07,$F3,$EC,$5F,$0D,$E9,$6C,$66,$30,$67,$D7
	.dc.b	$91,$C5,$AF,$F4,$D9,$7C,$97,$32,$47,$77,$F0,$1F,$5F,$23,$F0,$0D
	.dc.b	$7F,$0D,$E8,$7E,$14,$8F,$E1,$5F,$EA,$BC,$0F,$9E,$7D,$4E,$1B,$5F
	.dc.b	$ED,$96,$BF,$EB,$F6,$BF,$D3,$A7,$97,$97,$C4,$91,$C5,$EF,$F1,$8E
	.dc.b	$5A,$49,$FF,$73,$46,$5F,$F6,$CF,$90,$CF,$FB,$5F,$C6,$23,$13,$A9
	.dc.b	$CC,$62,$3F,$DA,$94,$FB,$5F,$09,$23,$BB,$FD,$5F,$D0,$FE,$9D,$1F
	.dc.b	$EE,$6F,$3B,$FD,$EF,$9E,$49,$1F,$ED,$89,$7C,$1F,$12,$46,$5F,$DA
	.dc.b	$7C,$3C,$8F,$DA,$35,$FD,$D7,$C9,$7E,$E4,$8F,$EE,$5F,$D1,$7E,$46
	.dc.b	$E9,$A4,$9F,$F0,$6C,$65,$BD,$FC,$08,$E0,$FF,$09,$FE,$47,$FA,$D4
	.dc.b	$D2,$3F,$D0,$3E,$55,$F1,$14,$91,$FE,$21,$F3,$D3,$A8,$FF,$0F,$FE
	.dc.b	$29,$FE,$DA,$6F,$B1,$FB,$B8,$C4,$7F,$8A,$5A,$FF,$A6,$E5,$F1,$1E
	.dc.b	$12,$47,$78,$CB,$7E,$23,$CC,$47,$07,$8B,$C9,$F0,$11,$C1,$CF,$E1
	.dc.b	$FD,$C4,$70,$77,$F5,$BE,$11,$1C,$1E,$EF,$88,$F6,$11,$C1,$F1,$7F
	.dc.b	$14,$FF,$1A,$3F,$EB,$76,$BF,$E5,$46,$AD,$51,$88,$7D,$8A,$FF,$92
	.dc.b	$36,$3F,$C9,$AD,$7F,$CA,$8D,$7F,$C8,$23,$57,$8D,$F4,$C9,$1F,$61
	.dc.b	$FE,$61,$F9,$7F,$DB,$8F,$F2,$F3,$5E,$6D,$7F,$C9,$AD,$7D,$5F,$65
	.dc.b	$FE,$67,$7F,$0C,$D7,$FD,$B6,$D7,$FD,$B4,$D7,$FC,$EF,$EE,$BE,$4F
	.dc.b	$FC,$22,$D8,$FC,$83,$5F,$F7,$2B,$5F,$F7,$27,$AC,$F8,$7F,$F0,$CB
	.dc.b	$63,$F3,$0D,$7F,$DD,$AD,$7F,$DD,$9F,$5F,$FE,$44,$7F,$DE,$2D,$73
	.dc.b	$F0,$3F,$E2,$48,$C4,$7F,$A3,$BE,$BD,$EE,$7F,$8A,$1B,$1F,$A7,$6B
	.dc.b	$FE,$FC,$6B,$FE,$76,$F6,$7F,$CC,$5F,$0F,$FE,$88,$6C,$7F,$A3,$DA
	.dc.b	$FD,$C6,$BC,$7E,$BF,$FC,$9C,$A1,$18,$6F,$AF,$D6,$37,$CC,$FA,$F7
	.dc.b	$B1,$FE,$40,$6E,$1F,$BF,$6B,$F8,$C9,$E1,$3E,$7F,$1F,$E1,$CF,$85
	.dc.b	$B8,$FF,$26,$DB,$47,$B7,$EB,$FF,$9A,$57,$B5,$AF,$B5,$AF,$C5,$6B
	.dc.b	$FE,$BC,$FA,$B7,$CE,$7F,$95,$9B,$1F,$C5,$5A,$F9,$DA,$FF,$A9,$32
	.dc.b	$7E,$10,$EF,$FF,$2F,$B6,$7F,$8F,$35,$F8,$0D,$7E,$C7,$D8,$3E,$57
	.dc.b	$EC,$36,$3E,$C3,$5F,$B0,$D7,$F0,$1F,$61,$FE,$C9,$6E,$BF,$81,$8C
	.dc.b	$47,$FB,$8A,$11,$88,$FB,$DF,$61,$FE,$73,$7F,$0C,$D7,$F0,$CD,$7F
	.dc.b	$1C,$D7,$FD,$8B,$9B,$E8,$9F,$61,$FE,$F0,$F6,$9F,$1D,$AF,$92,$FF
	.dc.b	$65,$C6,$30,$FF,$7A,$E1,$FE,$C2,$8C,$42,$3A,$5F,$3B,$EF,$87,$6A
	.dc.b	$FF,$01,$47,$57,$47,$4C,$3B,$C7,$0A,$CC,$F5,$AE,$58,$78,$30,$F1
	.dc.b	$5F,$3A,$EA,$79,$AB,$71,$CE,$32,$79,$F0,$EE,$F8,$4E,$F5,$EA,$76
	.dc.b	$DB,$F0,$4F,$09,$E4,$70,$BE,$EF,$87,$0F,$E6,$6D,$2D,$F5,$7F,$9B
	.dc.b	$B5,$FA,$BD,$2F,$B0,$E3,$1F,$36,$30,$F2,$6B,$C9,$AF,$83,$E8,$3F
	.dc.b	$C0,$04,$7F,$CD,$CD,$F8,$0F,$B3,$1F,$C2,$7E,$20,$FF,$91,$75,$E4
	.dc.b	$D3,$F8,$C4,$6A,$78,$C2,$B1,$5F,$C5,$EA,$CF,$F5,$49,$AF,$E7,$62
	.dc.b	$BE,$08,$FB,$EF,$8C,$FA,$E6,$4B,$C0,$9D,$C7,$84,$F3,$1D,$4A,$71
	.dc.b	$BF,$7C,$F1,$8C,$9C,$63,$33,$E4,$8C,$5F,$AD,$F8,$54,$FC,$93,$C5
	.dc.b	$5F,$23,$4E,$93,$98,$ED,$18,$9F,$E3,$31,$B3,$F7,$CF,$2A,$FD,$9C
	.dc.b	$60,$E2,$7F,$03,$FB,$17,$DA,$62,$7F,$A0,$57,$1B,$F0,$D6,$AF,$B5
	.dc.b	$7F,$45,$FB,$2F,$F3,$24,$63,$14,$62,$7C,$26,$FC,$8E,$D3,$AF,$80
	.dc.b	$BF,$91,$7F,$98,$B0,$FE,$AC,$DB,$FE,$AC,$D7,$FA,$F3,$5F,$C6,$F6
	.dc.b	$3E,$9B,$F9,$55,$7F,$93,$5B,$3F,$D3,$DA,$FF,$39,$6B,$FC,$97,$EC
	.dc.b	$9F,$27,$E6,$6D,$7C,$0B,$ED,$DB,$9A,$F5,$FD,$3B,$E8,$A9,$8F,$A7
	.dc.b	$B8,$71,$FB,$2F,$E8,$CF,$F9,$85,$B9,$FF,$63,$6B,$FD,$35,$AF,$F3
	.dc.b	$9F,$B2,$7D,$57,$CA,$FD,$8B,$EB,$5E,$63,$0F,$CE,$FB,$37,$D5,$D3
	.dc.b	$1E,$AD,$9F,$67,$E8,$DF,$8A,$CB,$FB,$A3,$5F,$EB,$CD,$7E,$7B,$57
	.dc.b	$C7,$7F,$94,$7E,$1B,$F1,$4F,$F9,$4F,$8A,$F1,$FF,$A0,$B0,$FC,$4C
	.dc.b	$E5,$F9,$9F,$68,$FA,$1F,$C9,$52,$31,$3C,$3C,$3F,$F4,$D3,$C2,$B7
	.dc.b	$F6,$8F,$86,$FD,$96,$9F,$DD,$63,$15,$FE,$8F,$6C,$7F,$84,$5A,$FF
	.dc.b	$7B,$F2,$5C,$DE,$AF,$2D,$E9,$F9,$27,$38,$7A,$A7,$2F,$B3,$E4,$1F
	.dc.b	$43,$E2,$C7,$F9,$0F,$02,$73,$5F,$EC,$1A,$D5,$7C,$A9,$CB,$FC,$FB
	.dc.b	$ED,$3F,$CD,$F5,$FC,$E1,$B1,$9F,$83,$E2,$A3,$2F,$F9,$E3,$F0,$BF
	.dc.b	$EE,$35,$FD,$01,$B1,$FD,$29,$AE,$FC,$CF,$1E,$4F,$A2,$F6,$CF,$FB
	.dc.b	$6D,$AF,$F3,$C3,$F0,$2E,$1F,$7B,$5F,$F4,$A7,$C6,$3D,$FF,$7F,$3F
	.dc.b	$FB,$01,$B1,$9C,$BF,$D6,$D1,$F1,$3E,$65,$F8,$53,$E8,$B0,$DA,$FE
	.dc.b	$C7,$3A,$9C,$9E,$AC,$9F,$6D,$49,$5F,$DA,$B4,$F7,$47,$8B,$59,$3F
	.dc.b	$EC,$76,$EB,$F0,$7E,$DB,$A1,$27,$96,$08,$F0,$7D,$B4,$9C,$B6,$BF
	.dc.b	$CC,$1B,$7F,$53,$7F,$DB,$1E,$7F,$C2,$25,$74,$F5,$FD,$44,$70,$7F
	.dc.b	$7A,$FD,$1F,$FD,$3C,$FF,$34,$A4,$61,$FE,$63,$BB,$D4,$7D,$37,$F3
	.dc.b	$F9,$DF,$A9,$7F,$0B,$FE,$0F,$8F,$F8,$19,$B3,$FE,$10,$6B,$FE,$C4
	.dc.b	$6B,$F8,$3C,$7F,$A3,$3F,$D0,$0F,$1F,$B3,$D7,$CA,$F0,$4E,$BD,$8F
	.dc.b	$ED,$1D,$2F,$D5,$DF,$45,$FC,$09,$B9,$FF,$08,$E5,$FF,$27,$3F,$CC
	.dc.b	$E3,$39,$FF,$4E,$E8,$FC,$C7,$BF,$FC,$4A,$CF,$3C,$30,$E7,$D8,$ED
	.dc.b	$F3,$D7,$2E,$8E,$1F,$07,$D3,$DC,$BB,$62,$8F,$ED,$1F,$48,$FC,$17
	.dc.b	$E8,$6F,$FE,$03,$8C,$72,$78,$F8,$F7,$5A,$CB,$FC,$4B,$CF,$FA,$FF
	.dc.b	$9D,$FD,$06,$66,$92,$7F,$C8,$38,$F7,$C3,$DD,$8F,$46,$CF,$C5,$ED
	.dc.b	$F9,$47,$C3,$F7,$EC,$7C,$BF,$27,$E3,$18,$DF,$DA,$FC,$9C,$7E,$97
	.dc.b	$95,$6B,$E8,$38,$7F,$9D,$FD,$93,$E9,$31,$87,$D3,$8C,$4F,$E5,$7E
	.dc.b	$E9,$F5,$BF,$91,$6C,$72,$F9,$0C,$4E,$7C,$8F,$E3,$B5,$FB,$FC,$3F
	.dc.b	$CC,$2D,$7F,$89,$07,$0F,$F1,$57,$FD,$D7,$7E,$1F,$95,$E8,$D3,$D3
	.dc.b	$F5,$3F,$8F,$E8,$FE,$B3,$9A,$31,$3F,$8F,$9C,$FB,$98,$C7,$DC,$FF
	.dc.b	$8A,$79,$FF,$94,$7E,$09,$7A,$0F,$E0,$3E,$A3,$0F,$E0,$BE,$AE,$65
	.dc.b	$FE,$2C,$BE,$2F,$F8,$BB,$93,$F1,$93,$F8,$EF,$C2,$3E,$3E,$7F,$E2
	.dc.b	$C9,$46,$3B,$7F,$CA,$5E,$13,$83,$F7,$31,$F9,$39,$D6,$3F,$90,$D7
	.dc.b	$FC,$55,$8B,$3F,$E8,$1C,$BF,$92,$25,$7E,$42,$FE,$96,$1C,$F9,$3D
	.dc.b	$D5,$E5,$7F,$71,$D3,$2F,$F0,$BF,$5B,$72,$FC,$27,$AB,$F3,$8C,$63
	.dc.b	$E8,$D7,$FA,$EB,$FC,$13,$EC,$BF,$13,$5F,$E4,$AF,$1D,$18,$9E,$1F
	.dc.b	$5F,$FE,$08,$8C,$65,$FD,$C3,$0F,$A2,$FE,$5B,$93,$5D,$9F,$79,$E1
	.dc.b	$D3,$FD,$52,$FF,$01,$5F,$EA,$EC,$BB,$CE,$4F,$EC,$B8,$7F,$85,$33
	.dc.b	$FC,$F7,$BD,$CA,$FF,$46,$FE,$2B,$DA,$91,$FA,$6F,$A7,$B5,$FA,$5F
	.dc.b	$B2,$E6,$49,$FE,$A9,$82,$3C,$9C,$12,$7C,$9A,$7F,$7E,$BF,$F3,$0B
	.dc.b	$5F,$F1,$9D,$BE,$87,$F3,$67,$C5,$BB,$7F,$E5,$BB,$70,$7C,$12,$47
	.dc.b	$67,$DF,$F7,$DF,$A4,$3E,$73,$63,$96,$BF,$44,$FD,$4F,$C7,$D8,$C9
	.dc.b	$FC,$1B,$ED,$F1,$A7,$F0,$57,$EC,$6F,$F4,$49,$3E,$E7,$E7,$FD,$5F
	.dc.b	$ED,$0D,$C7,$F1,$07,$F6,$FF,$3D,$23,$F8,$9F,$99,$70,$FE,$25,$1F
	.dc.b	$60,$87,$CE,$3F,$C3,$FF,$A4,$7F,$9B,$1F,$F2,$AB,$63,$8C,$63,$E3
	.dc.b	$BF,$9F,$F6,$7C,$88,$FD,$B3,$67,$F6,$8F,$E6,$A3,$FD,$C3,$E1,$7F
	.dc.b	$86,$77,$E3,$D9,$FE,$40,$8F,$F0,$E3,$68,$D3,$CB,$F2,$7F,$E6,$CC
	.dc.b	$BF,$D1,$D8,$79,$A3,$1F,$27,$23,$D3,$F1,$17,$3F,$EC,$97,$FE,$58
	.dc.b	$7C,$BE,$47,$A8,$E9,$F1,$B6,$FD,$7D,$BE,$32,$3F,$D4,$34,$FD,$2D
	.dc.b	$FF,$30,$F1,$D8,$BE,$B7,$CD,$F1,$FE,$76,$E1,$FE,$A9,$FB,$AF,$F0
	.dc.b	$5B,$FE,$49,$F0,$91,$8C,$7F,$D7,$6D,$51,$AF,$77,$FD,$DD,$5C,$7E
	.dc.b	$E9,$E7,$62,$FF,$D8,$71,$89,$47,$FA,$12,$D3,$98,$C3,$67,$FA,$5F
	.dc.b	$DD,$3A,$DF,$09,$FD,$3A,$F9,$CE,$63,$10,$8C,$3B,$6B,$D3,$B9,$D3
	.dc.b	$C2,$51,$89,$E0,$D7,$84,$FF,$13,$F1,$7F,$C9,$D1,$EC,$C7,$E5,$46
	.dc.b	$23,$DC,$D8,$CF,$AF,$FE,$77,$7E,$87,$C6,$BE,$17,$17,$FE,$EA,$8C
	.dc.b	$3D,$4C,$46,$9F,$A7,$1A,$7F,$D5,$FC,$BF,$82,$FA,$3F,$F4,$27,$9D
	.dc.b	$FD,$4B,$9F,$FD,$69,$5F,$AC,$F9,$1C,$E5,$FE,$A4,$95,$E2,$BE,$2C
	.dc.b	$7F,$BA,$DF,$F0,$FB,$5F,$EF,$FA,$FE,$43,$BB,$E8,$B6,$37,$DD,$CE
	.dc.b	$3F,$2A,$DA,$FF,$5E,$CF,$EF,$BE,$BF,$C1,$A7,$E6,$BE,$EE,$D3,$E9
	.dc.b	$E9,$88,$CF,$C8,$5C,$3F,$83,$C6,$1F,$EF,$14,$F0,$FC,$AD,$8F,$5C
	.dc.b	$60,$CF,$8D,$E5,$24,$7F,$80,$79,$7F,$DB,$7C,$5F,$2F,$EB,$3C,$A4
	.dc.b	$9F,$E2,$FF,$75,$E8,$3F,$27,$AF,$ED,$1B,$91,$A6,$3E,$EB,$DE,$4A
	.dc.b	$FE,$43,$F2,$3F,$0B,$DF,$91,$FC,$91,$FA,$DE,$07,$F7,$2F,$52,$FA
	.dc.b	$7F,$D3,$EB,$FA,$AC,$7E,$92,$75,$18,$3F,$AC,$FA,$3F,$3B,$C9,$91
	.dc.b	$FD,$4E,$39,$91,$C1,$FC,$D3,$E0,$7E,$AF,$87,$23,$F9,$BB,$E2,$F0
	.dc.b	$A4,$7F,$66,$7C,$59,$7E,$BA,$30,$7F,$B0,$77,$FF,$24,$F1,$7E,$5E
	.dc.b	$3F,$02,$7B,$3D,$08,$FF,$70,$79,$91,$FF,$65,$7B,$4F,$B5,$FA,$2D
	.dc.b	$8C,$C6,$0C,$FC,$7F,$E4,$27,$FD,$67,$3F,$C5,$AF,$FC,$73,$18,$8F
	.dc.b	$F4,$E2,$FB,$7E,$1F,$1B,$27,$87,$89,$FC,$7B,$A1,$7F,$EA,$A9,$C9
	.dc.b	$FC,$AB,$C7,$FC,$93,$FE,$2B,$6B,$FC,$CD,$AF,$F2,$8E,$4F,$E0,$DD
	.dc.b	$BF,$6E,$73,$2B,$EF,$F0,$7E,$B2,$47,$82,$FF,$87,$7A,$7F,$88,$B3
	.dc.b	$49,$3E,$6B,$F4,$F9,$DF,$17,$E2,$E4,$E7,$34,$73,$F2,$A4,$F7,$C7
	.dc.b	$F8,$7B,$ED,$8F,$9D,$F3,$1F,$C0,$FE,$15,$F2,$94,$95,$FE,$71,$E7
	.dc.b	$7C,$57,$F1,$3F,$87,$FE,$BA,$FF,$3E,$6C,$FF,$43,$6B,$FE,$92,$6B
	.dc.b	$FD,$40,$FE,$7F,$6F,$95,$F1,$12,$3D,$8F,$97,$9A,$47,$FA,$06,$7F
	.dc.b	$00,$F9,$7E,$2F,$C9,$89,$A4,$8E,$1D,$5F,$3F,$CC,$FC,$77,$A7,$27
	.dc.b	$E3,$47,$E4,$75,$D4,$60,$F9,$CF,$F3,$35,$F6,$EF,$8D,$C3,$63,$F1
	.dc.b	$A6,$A3,$3F,$19,$23,$FB,$2C,$FF,$B9,$DB,$1F,$DA,$1E,$AD,$87,$C8
	.dc.b	$76,$24,$FE,$07,$93,$F1,$5E,$6F,$EA,$75,$93,$F8,$18,$FD,$67,$C8
	.dc.b	$EA,$30,$7E,$25,$FD,$85,$C8,$FC,$63,$F9,$5F,$36,$47,$E2,$DA,$FE
	.dc.b	$CB,$2F,$E9,$9E,$23,$D1,$FD,$D6,$B2,$BF,$D7,$13,$97,$F1,$9E,$2F
	.dc.b	$E2,$78,$FF,$79,$1F,$85,$E7,$FF,$14,$CE,$4F,$EB,$75,$FD,$46,$79
	.dc.b	$F1,$78,$0F,$E1,$1F,$21,$FA,$BF,$2B,$F8,$57,$74,$AF,$D1,$47,$F0
	.dc.b	$DE,$BD,$46,$0F,$D5,$3F,$C4,$7D,$69,$1F,$AD,$7F,$9E,$FD,$B7,$F1
	.dc.b	$98,$FF,$58,$37,$C9,$79,$1F,$C1,$63,$0D,$EC,$FE,$D7,$F8,$73,$E3
	.dc.b	$FF,$15,$F8,$7F,$FA,$6F,$1F,$C3,$B5,$63,$D1,$FF,$1B,$C6,$63,$10
	.dc.b	$FC,$23,$FC,$0E,$FE,$41,$AF,$F8,$D5,$AF,$EC,$4E,$1E,$EF,$E3,$E3
	.dc.b	$31,$FB,$5A,$FF,$37,$BE,$4F,$88,$F3,$9F,$C0,$79,$3C,$5A,$7F,$2D
	.dc.b	$CD,$F2,$5F,$DB,$9B,$7F,$E4,$5F,$2D,$18,$8F,$F3,$33,$5F,$F3,$06
	.dc.b	$FF,$CD,$5F,$37,$9A,$8C,$46,$BD,$0F,$ED,$51,$88,$F1,$7C,$47,$84
	.dc.b	$8F,$86,$D1,$E5,$F8,$99,$39,$6B,$E0,$AF,$E2,$8F,$FB,$7D,$AE,$7C
	.dc.b	$BF,$09,$EF,$F8,$69,$1D,$DA,$FF,$21,$FC,$15,$F1,$BF,$C9,$DB,$1F
	.dc.b	$C9,$DA,$EE,$3D,$BE,$4C,$8F,$13,$5F,$B5,$EC,$3F,$06,$FD,$24,$61
	.dc.b	$AF,$F2,$F8,$D7,$81,$FE,$74,$6C,$7F,$9D,$BE,$E7,$FD,$2C,$DF,$8C
	.dc.b	$FE,$25,$18,$8F,$F3,$F3,$5F,$C5,$7D,$4F,$E1,$9F,$F4,$13,$5C,$F4
	.dc.b	$7F,$1B,$8C,$47,$F7,$3F,$03,$EC,$3F,$D9,$DB,$CD,$FD,$3B,$97,$FC
	.dc.b	$6F,$1A,$E6,$7D,$CB,$E4,$7F,$CA,$BA,$FF,$BD,$B1,$FE,$CC,$D7,$F5
	.dc.b	$3D,$5F,$DD,$EB,$84,$7F,$71,$F3,$FF,$C0,$6F,$FA,$67,$04,$62,$3F
	.dc.b	$22,$D7,$FD,$68,$D7,$FC,$05,$F0,$3F,$07,$C1,$FD,$B5,$23,$18,$6B
	.dc.b	$7F,$F3,$5F,$E0,$AF,$81,$FF,$36,$E3,$F9,$43,$FB,$F1,$FA,$5C,$7F
	.dc.b	$80,$70,$7F,$AD,$BB,$F7,$DD,$FE,$4E,$FB,$97,$C1,$FF,$92,$DB,$87
	.dc.b	$AD,$AF,$EA,$9A,$FE,$0F,$C6,$7A,$9F,$E8,$DE,$3F,$E2,$8D,$8C,$CF
	.dc.b	$F8,$9A,$7F,$8C,$7A,$BB,$FB,$9F,$F3,$44,$E5,$F9,$2D,$E4,$D7,$C3
	.dc.b	$FF,$99,$3E,$E5,$F1,$7F,$E6,$16,$D7,$D1,$6B,$FC,$01,$AF,$E6,$77
	.dc.b	$FF,$5D,$70,$FF,$23,$CF,$F8,$7F,$B9,$E9,$9C,$F3,$7F,$B7,$27,$F8
	.dc.b	$9F,$8D,$ED,$FB,$9F,$F0,$CE,$3E,$B6,$BF,$86,$3F,$DB,$6E,$31,$FB
	.dc.b	$7F,$6B,$FE,$8E,$C3,$FC,$80,$FF,$30,$AF,$F7,$B3,$5C,$C7,$F9,$F6
	.dc.b	$31,$1F,$EE,$6F,$8F,$F8,$AF,$EE,$1A,$FE,$D5,$AF,$A5,$AF,$FB,$BB
	.dc.b	$EE,$3F,$65,$7F,$05,$B7,$EF,$6C,$EF,$E5,$FE,$E7,$DB,$BE,$6F,$EF
	.dc.b	$7D,$C7,$F9,$F3,$0F,$8C,$DB,$FF,$7D,$D6,$DF,$C0,$35,$F8,$DB,$23
	.dc.b	$19,$7D,$36,$DF,$D1,$7F,$D4,$15,$F3,$46,$9F,$E3,$5A,$43,$ED,$DF
	.dc.b	$25,$FE,$45,$AF,$C8,$CE,$63,$1D,$5F,$E4,$9A,$DC,$F2,$AE,$8E,$CC
	.dc.b	$74,$FF,$6A,$70,$78,$AB,$95,$F4,$8F,$1D,$F4,$7C,$F5,$C6,$31,$7D
	.dc.b	$BC,$6B,$87,$F9,$6D,$B7,$EC,$77,$7E,$16,$E3,$DB,$1F,$05,$79,$30
	.dc.b	$F7,$E3,$A8,$C6,$1E,$1A,$DC,$F6,$9C,$DF,$37,$DE,$FF,$7D,$69,$6F
	.dc.b	$E2,$BF,$E0,$16,$BF,$3F,$CD,$7D,$C7,$18,$FA,$F1,$87,$93,$5E,$4D
	.dc.b	$7C,$1F,$5F,$FE,$59,$4D,$23,$FE,$5D,$7D,$0F,$ED,$F6,$AF,$E1,$22
	.dc.b	$EF,$FE,$21,$CB,$EA,$4E,$4F,$B6,$23,$59,$F8,$A3,$31,$87,$FC,$41
	.dc.b	$5F,$56,$7F,$B4,$DD,$9C,$9F,$E9,$E8,$CB,$4C,$DF,$8A,$FA,$47,$FA
	.dc.b	$6E,$FC,$F7,$EA,$8F,$82,$78,$9C,$46,$A3,$F2,$D2,$FB,$87,$C0,$7F
	.dc.b	$8F,$F5,$FE,$3D,$18,$C3,$F9,$92,$F9,$CD,$47,$F1,$F8,$D6,$1F,$B6
	.dc.b	$FA,$B7,$C3,$EA,$3E,$34,$62,$B9,$E3,$9C,$93,$9A,$FD,$FF,$DC,$7F
	.dc.b	$A7,$6F,$FA,$13,$9C,$FF,$99,$39,$9D,$E3,$CE,$39,$47,$F8,$45,$AF
	.dc.b	$F3,$56,$BF,$65,$BE,$6F,$FA,$7B,$E7,$3F,$96,$37,$0F,$EE,$DF,$16
	.dc.b	$9C,$B6,$3F,$41,$9B,$D3,$FE,$41,$1F,$D6,$5B,$8F,$F3,$FC,$3A,$3F
	.dc.b	$C7,$23,$39,$FF,$4D,$F9,$61,$F1,$47,$CD,$D7,$0F,$CD,$B7,$DB,$BD
	.dc.b	$8F,$4B,$72,$FF,$07,$37,$2F,$F2,$EB,$5F,$F1,$B7,$DB,$7F,$87,$67
	.dc.b	$FB,$83,$63,$F9,$8B,$F7,$B2,$FE,$C4,$E5,$FF,$08,$FD,$AE,$6B,$FB
	.dc.b	$17,$DB,$7F,$98,$1F,$E5,$F3,$AE,$4F,$C0,$4E,$70,$F8,$97,$F8,$DB
	.dc.b	$FF,$12,$36,$3F,$CE,$6D,$7F,$DF,$59,$A3,$11,$FE,$24,$F5,$3E,$27
	.dc.b	$AF,$FE,$1F,$BF,$09,$FD,$03,$D5,$61,$F6,$6F,$EA,$56,$DF,$AD,$F6
	.dc.b	$CF,$A6,$FC,$D5,$FF,$0C,$F8,$AF,$A1,$39,$C7,$AD,$AB,$51,$8C,$5F
	.dc.b	$6D,$F3,$EB,$FA,$8C,$63,$8B,$F1,$F3,$99,$F2,$B6,$5E,$27,$E0,$AF
	.dc.b	$A3,$6B,$FC,$15,$83,$C5,$7C,$DF,$E1,$6B,$5C,$7F,$EB,$8F,$91,$7C
	.dc.b	$17,$F8,$19,$B9,$7F,$31,$6E,$5F,$CE,$DA,$FF,$AE,$7F,$53,$AF,$F6
	.dc.b	$A0,$E1,$FE,$DA,$D7,$35,$FE,$17,$CF,$FD,$67,$EF,$DF,$8A,$34,$98
	.dc.b	$FF,$43,$FB,$6F,$EE,$AD,$F2,$7F,$C3,$B1,$8C,$7F,$A8,$35,$FF,$67
	.dc.b	$72,$BE,$73,$FC,$46,$DB,$FE,$9C,$D9,$FE,$B0,$D7,$D9,$97,$FB,$23
	.dc.b	$EA,$FC,$C4,$AF,$F5,$E3,$E9,$A9,$1F,$EB,$F6,$D7,$FA,$EF,$E5,$3F
	.dc.b	$85,$FC,$69,$A4,$AF,$F6,$37,$D7,$7C,$32,$38,$3F,$D9,$5F,$59,$F1
	.dc.b	$C8,$E0,$FF,$45,$7E,$71,$FA,$BC,$3F,$4A,$D7,$F1,$ED,$7F,$42,$DC
	.dc.b	$FF,$DC,$BF,$5F,$F4,$C9,$5F,$ED,$97,$EB,$7C,$7F,$F3,$BF,$97,$FF
	.dc.b	$75,$F8,$3F,$A4,$E4,$FF,$66,$DF,$E8,$A7,$2F,$E9,$32,$7D,$AF,$DE
	.dc.b	$8F,$EF,$CD,$8F,$F0,$43,$5F,$C6,$1C,$39,$7F,$CB,$D5,$FE,$FA,$F0
	.dc.b	$BC,$58,$7F,$A7,$F0,$ED,$87,$3E,$A4,$7F,$8F,$AF,$83,$FB,$3D,$BF
	.dc.b	$71,$97,$F3,$36,$D7,$FA,$83,$07,$D8,$7F,$2D,$9C,$DF,$F0,$99,$F0
	.dc.b	$5F,$EC,$1B,$7F,$EB,$2A,$DC,$FF,$03,$9C,$D7,$1F,$73,$A6,$FF,$25
	.dc.b	$B1,$E5,$8F,$99,$B6,$2F,$B3,$F0,$32,$FE,$D9,$D8,$FD,$0F,$F9,$96
	.dc.b	$1E,$FC,$BD,$2F,$F1,$49,$D7,$C8,$70,$AF,$E7,$98,$FE,$E6,$73,$5F
	.dc.b	$BF,$9C,$CF,$F9,$9A,$72,$FD,$4D,$98,$7E,$02,$3F,$D3,$51,$8E,$2E
	.dc.b	$D8,$FD,$67,$E5,$70,$CC,$BF,$93,$F1,$F2,$FA,$AC,$63,$F2,$FD,$7F
	.dc.b	$C6,$18,$E3,$FA,$8F,$E3,$FC,$BF,$A9,$57,$0F,$B8,$E1,$EB,$AF,$C4
	.dc.b	$C3,$F3,$F3,$F9,$7D,$3F,$31,$5F,$E1,$59,$FC,$3F,$93,$88,$FC,$C0
	.dc.b	$75,$F6,$7F,$E8,$88,$FF,$2B,$36,$7F,$CC,$0D,$7F,$32,$D7,$F9,$33
	.dc.b	$FE,$47,$8F,$D2,$36,$3F,$D2,$CF,$3C,$FF,$33,$1F,$E9,$68,$D4,$7F
	.dc.b	$BD,$3E,$CB,$F9,$DD,$7E,$F3,$0F,$EA,$B1,$8B,$FD,$EC,$7E,$64,$CF
	.dc.b	$A3,$6B,$E8,$D7,$F8,$39,$FC,$37,$1B,$B3,$F0,$C7,$EB,$BE,$46,$CF
	.dc.b	$EF,$76,$7C,$47,$EC,$8C,$E7,$F8,$28,$CC,$FE,$47,$E0,$DF,$3D,$F2
	.dc.b	$B4,$FC,$CC,$62,$7E,$1C,$7C,$EF,$57,$FD,$B3,$C1,$FD,$77,$1F,$97
	.dc.b	$18,$3F,$DB,$BE,$F7,$EF,$AF,$F8,$73,$67,$E5,$CF,$DD,$8C,$70,$FC
	.dc.b	$CF,$3D,$FA,$9F,$C1,$BF,$F1,$A3,$67,$E7,$7C,$47,$F8,$B6,$31,$8F
	.dc.b	$F8,$27,$8F,$FC,$11,$5F,$4F,$DC,$D6,$BF,$49,$E1,$F5,$72,$FE,$7B
	.dc.b	$72,$FE,$EF,$83,$E2,$3E,$C9,$9C,$7E,$F3,$E1,$F5,$BF,$CA,$79,$FF
	.dc.b	$80,$C8,$FF,$6B,$AF,$F1,$E5,$7E,$16,$38,$63,$F7,$9F,$C2,$FD,$57
	.dc.b	$E1,$BE,$86,$CF,$E0,$9C,$3F,$EB,$9E,$F6,$09,$EC,$49,$FF,$5F,$70
	.dc.b	$7C,$AF,$4F,$CA,$48,$CE,$7F,$E0,$3C,$39,$FD,$87,$E5,$8F,$99,$56
	.dc.b	$BF,$C4,$D4,$FA,$6F,$79,$23,$9B,$EF,$9F,$CF,$9F,$55,$B1,$F9,$CF
	.dc.b	$92,$FD,$19,$FA,$48,$B1,$DB,$E5,$3E,$E1,$8E,$1F,$C0,$BF,$E6,$BF
	.dc.b	$D6,$7F,$29,$1F,$D4,$B4,$46,$23,$FC,$A2,$D7,$FC,$76,$D7,$FC,$85
	.dc.b	$B2,$BF,$6A,$75,$EB,$B8,$7E,$CB,$D8,$FD,$CF,$CA,$FE,$E9,$B1,$FB
	.dc.b	$6F,$A8,$FB,$C4,$7E,$B6,$08,$FC,$17,$D8,$7F,$02,$3F,$81,$B6,$7F
	.dc.b	$BA,$7D,$47,$F8,$F3,$9F,$F8,$6F,$9A,$87,$27,$F7,$D7,$55,$FD,$A5
	.dc.b	$CB,$CB,$9B,$FB,$9C,$7E,$16,$3F,$E0,$05,$73,$9C,$9F,$95,$D7,$F8
	.dc.b	$A9,$FC,$5B,$E7,$2D,$7F,$12,$E0,$FE,$32,$7F,$1A,$F8,$9B,$1F,$D1
	.dc.b	$BC,$87,$D6,$FF,$9C,$9F,$E6,$91,$98,$C7,$97,$E9,$EE,$F2,$92,$7C
	.dc.b	$BA,$FE,$BD,$FC,$F5,$BE,$83,$A1,$D7,$E9,$AC,$F8,$7E,$BF,$F9,$9B
	.dc.b	$FE,$44,$A4,$63,$4F,$F4,$64,$62,$7F,$A1,$77,$FF,$75,$AD,$57,$FA
	.dc.b	$43,$9B,$FD,$41,$5F,$E9,$A6,$C7,$FA,$11,$AE,$FC,$CE,$2B,$EC,$FF
	.dc.b	$DE,$5C,$1F,$E2,$68,$19,$F4,$91,$98,$C3,$67,$8C,$71,$5E,$8B,$9B
	.dc.b	$4D,$9D,$35,$FF,$1A,$CF,$09,$FE,$07,$F6,$6F,$90,$FF,$42,$69,$FE
	.dc.b	$B6,$8C,$5F,$F6,$5F,$05,$18,$8F,$EB,$E7,$FA,$D1,$C4,$FC,$FF,$B1
	.dc.b	$E7,$3F,$39,$AF,$FA,$BF,$85,$18,$E1,$FF,$12,$36,$3F,$D0,$53,$F8
	.dc.b	$BF,$85,$FF,$4C,$76,$FF,$6E,$BF,$F1,$54,$63,$0F,$B7,$F1,$36,$FE
	.dc.b	$27,$E1,$0F,$0F,$E6,$6E,$8C,$4F,$E8,$9A,$FF,$89,$DA,$FE,$27,$E8
	.dc.b	$3A,$65,$FE,$39,$FA,$0B,$A9,$B3,$8F,$9A,$97,$FA,$37,$7F,$63,$C3
	.dc.b	$E7,$A4,$FE,$9D,$E7,$F1,$FE,$9D,$3C,$DE,$6F,$A3,$53,$49,$7E,$9F
	.dc.b	$3B,$84,$FF,$33,$8F,$F3,$E3,$5F,$F1,$FB,$FE,$CA,$7F,$6D,$CF,$EA
	.dc.b	$23,$FB,$9F,$99,$F6,$9F,$93,$E0,$FF,$2C,$37,$80,$D3,$1F,$5F,$F1
	.dc.b	$3E,$C2,$5F,$F0,$38,$F8,$FF,$1B,$23,$F8,$33,$F5,$FA,$FF,$9B,$F7
	.dc.b	$5F,$47,$F4,$75,$FD,$61,$AF,$F3,$D6,$BF,$CF,$23,$F5,$13,$FE,$06
	.dc.b	$F3,$BA,$7A,$7F,$CB,$EB,$FA,$BC,$7D,$8E,$29,$1F,$C5,$DA,$FF,$BD
	.dc.b	$35,$FF,$7A,$3F,$E6,$0F,$35,$F4,$DF,$82,$6C,$66,$3B,$F9,$D2,$3F
	.dc.b	$B5,$BE,$4C,$4D,$3E,$65,$7F,$8D,$27,$B7,$C0,$75,$3D,$3A,$48,$FE
	.dc.b	$EA,$F3,$23,$FE,$01,$97,$8F,$D0,$91,$97,$BF,$D7,$7A,$59,$1E,$D8
	.dc.b	$F6,$7D,$74,$8F,$80,$FD,$5E,$97,$99,$4F,$9E,$E2,$7F,$C1,$AD,$7F
	.dc.b	$B9,$35,$FF,$7A,$E0,$F8,$FF,$94,$48,$FE,$47,$E3,$7F,$8F,$EB,$F8
	.dc.b	$2C,$63,$C5,$FE,$4F,$C9,$FA,$2F,$1F,$C8,$49,$FA,$DF,$27,$F5,$7E
	.dc.b	$4E,$47,$D9,$7D,$5E,$34,$8E,$82,$FF,$5B,$19,$86,$5F,$C3,$F4,$FF
	.dc.b	$37,$47,$FB,$97,$B5,$F1,$3F,$C3,$E3,$F4,$6D,$8C,$C6,$FE,$33,$07
	.dc.b	$F3,$D7,$F5,$2D,$8F,$F6,$CC,$7F,$41,$6B,$FE,$EC,$F8,$C7,$E8,$FF
	.dc.b	$CC,$31,$FE,$63,$4F,$8D,$F0,$1F,$C8,$6B,$7C,$1F,$C2,$FB,$3F,$37
	.dc.b	$8F,$FB,$01,$B8,$7F,$83,$1B,$3F,$EC,$9F,$15,$E0,$F3,$6D,$FF,$B5
	.dc.b	$5A,$FC,$69,$E1,$3D,$CF,$C4,$79,$5F,$A7,$F9,$39,$5F,$43,$8B,$F8
	.dc.b	$DC,$FE,$AF,$E4,$37,$39,$E7,$4E,$63,$EB,$8F,$B2,$79,$9F,$45,$B1
	.dc.b	$A6,$BE,$53,$FB,$7A,$F3,$38,$1F,$4D,$C4,$7F,$71,$6B,$FD,$DD,$AF
	.dc.b	$E6,$B8,$7C,$17,$23,$F8,$34,$7E,$9F,$CB,$FE,$0D,$CB,$27,$D9,$1F
	.dc.b	$C2,$79,$B5,$18,$3F,$87,$3F,$C3,$7E,$4A,$47,$F1,$17,$F6,$5E,$54
	.dc.b	$8F,$8C,$D7,$DD,$E6,$7F,$AE,$3E,$4B,$F0,$1E,$84,$9F,$CF,$F9,$6E
	.dc.b	$84,$FD,$DF,$4E,$57,$F8,$57,$AF,$FB,$34,$7F,$84,$DB,$19,$EA,$FE
	.dc.b	$EA,$FF,$86,$8F,$F0,$DE,$1F,$D9,$5F,$C3,$57,$47,$F2,$1D,$FF,$C2
	.dc.b	$E7,$F9,$57,$AD,$27,$F0,$11,$FC,$B7,$D7,$D4,$60,$FE,$64,$FF,$31
	.dc.b	$F9,$29,$1F,$CD,$5F,$E9,$5F,$90,$FF,$22,$8F,$C7,$35,$FC,$17,$62
	.dc.b	$31,$E0,$7F,$B3,$5B,$E0,$7F,$8D,$38,$BF,$DC,$3F,$27,$FA,$9F,$62
	.dc.b	$63,$F2,$B8,$D1,$FC,$07,$EC,$DF,$13,$4F,$DF,$38,$AF,$F1,$E3,$5F
	.dc.b	$F2,$13,$5F,$E4,$B8,$FE,$DF,$CA,$91,$FD,$46,$3F,$C2,$3C,$8F,$90
	.dc.b	$FE,$E8,$D9,$FF,$61,$35,$FA,$83,$4F,$F1,$04,$FF,$66,$D7,$F9,$94
	.dc.b	$FF,$B1,$1A,$FF,$AA,$5A,$FD,$DD,$FE,$47,$99,$E0,$F8,$E9,$3E,$06
	.dc.b	$BE,$0C,$3F,$C5,$DF,$1F,$E6,$24,$79,$9F,$17,$8F,$23,$C4,$D7,$96
	.dc.b	$6F,$8F,$F7,$92,$3D,$EF,$C2,$F8,$F9,$1F,$09,$AF,$F9,$BF,$F0,$4F
	.dc.b	$F5,$3B,$FE,$6D,$6B,$9F,$8B,$FE,$69,$18,$8F,$93,$F6,$2F,$91,$FE
	.dc.b	$70,$D8,$FF,$3B,$35,$FF,$3E,$35,$FD,$27,$B2,$FB,$2F,$F1,$94,$61
	.dc.b	$AF,$FA,$0A,$35,$F2,$5F,$E8,$16,$C7,$FA,$0F,$F1,$6F,$E2,$2F,$EB
	.dc.b	$DA,$FF,$A2,$9A,$9B,$F8,$2F,$E8,$63,$F0,$6D,$7F,$70,$D7,$FD,$26
	.dc.b	$DE,$47,$D0,$7E,$2F,$BD,$D3,$FD,$CA,$30,$DE,$9F,$F3,$2E,$5F,$D9
	.dc.b	$B1,$FD,$7F,$CA,$FF,$50,$7F,$14,$D7,$FB,$DB,$5F,$EE,$4D,$C7,$FB
	.dc.b	$6F,$B5,$FE,$70,$FB,$3F,$F4,$FD,$7F,$78,$E9,$46,$23,$F7,$BE,$72
	.dc.b	$BF,$C9,$9F,$82,$FF,$1E,$9F,$EA,$0D,$7F,$81,$B7,$6F,$F4,$64,$3E
	.dc.b	$03,$FC,$A9,$F6,$9F,$E4,$E8,$FF,$08,$B5,$FF,$04,$37,$33,$4F,$D8
	.dc.b	$D7,$C8,$9F,$F5,$EF,$C3,$7F,$AA,$63,$1B,$7F,$93,$3D,$3E,$F5,$FE
	.dc.b	$99,$87,$DA,$7F,$96,$63,$FA,$D4,$63,$E2,$35,$39,$BE,$6B,$E6,$7F
	.dc.b	$CD,$9F,$69,$FE,$9A,$C3,$FC,$52,$D7,$FC,$40,$DD,$0E,$71,$7C,$47
	.dc.b	$F7,$2A,$FF,$13,$C6,$3E,$3B,$CD,$26,$FB,$9E,$9F,$B4,$FF,$68,$69
	.dc.b	$EE,$6F,$4F,$E7,$DB,$8F,$F8,$C7,$F3,$8F,$C5,$B7,$AB,$FC,$F7,$1F
	.dc.b	$C6,$E3,$F8,$93,$5F,$F2,$1B,$5F,$F7,$4F,$DA,$FF,$2B,$7F,$58,$D7
	.dc.b	$3E,$7F,$EB,$63,$1A,$97,$CE,$7D,$17,$F9,$B6,$7F,$59,$B9,$F4,$9F
	.dc.b	$AD,$F9,$FB,$E7,$29,$32,$F8,$FF,$6D,$F9,$5C,$7E,$FB,$70,$FE,$00
	.dc.b	$D9,$FF,$80,$2D,$C8,$BD,$FF,$F2,$77,$DB,$7F,$86,$70,$5B,$ED,$7F
	.dc.b	$00,$E2,$7F,$C4,$31,$88,$46,$39,$FF,$BC,$3C,$E3,$94,$63,$E1,$FF
	.dc.b	$CC,$31,$83,$A7,$C4,$7F,$7F,$BE,$97,$CE,$7C,$15,$C2,$72,$F2,$FA
	.dc.b	$CD,$DF,$F8,$1D,$B5,$EA,$77,$8D,$C7,$C4,$FB,$B0,$DD,$78,$CF,$5D
	.dc.b	$F8,$EB,$71,$9C,$7D,$95,$C0,$93,$A7,$BC,$FF,$66,$3E,$2E,$3F,$01
	.dc.b	$FE,$D3,$5F,$15,$B1,$97,$E4,$F3,$3E,$DB,$94,$7C,$E8,$C3,$D6,$BF
	.dc.b	$CB,$73,$A6,$C7,$FA,$0F,$E9,$3F,$D0,$89,$A4,$74,$34,$FB,$8B,$7F
	.dc.b	$E1,$D1,$88,$E4,$6F,$3F,$F0,$1C,$E9,$F8,$15,$F3,$E7,$FA,$BF,$05
	.dc.b	$7F,$C7,$91,$5E,$D1,$5E,$B7,$E2,$BE,$91,$FE,$10,$AE,$11,$88,$E6
	.dc.b	$78,$CE,$E5,$33,$3C,$EB,$B5,$72,$6C,$EC,$F1,$9C,$E9,$F6,$CF,$83
	.dc.b	$FE,$61,$3F,$DC,$A3,$15,$F7,$EB,$75,$AA,$FE,$E5,$5A,$7E,$3F,$DB
	.dc.b	$3C,$2D,$47,$CF,$8C,$61,$CD,$B1,$C2,$77,$83,$ED,$B3,$3F,$E0,$46
	.dc.b	$BF,$E0,$16,$BF,$6D,$BE,$0B,$F0,$F7,$D6,$78,$A3,$0D,$8C,$E5,$F9
	.dc.b	$B9,$CB,$71,$7B,$9C,$5B,$19,$8C,$1E,$8F,$B7,$FF,$6E,$C7,$C7,$6E
	.dc.b	$5F,$8A,$6B,$FE,$1E,$6B,$FE,$1B,$FD,$57,$FC,$0F,$3F,$3D,$B1,$F9
	.dc.b	$26,$BF,$E2,$66,$BF,$EF,$EB,$7D,$4F,$9C,$95,$EB,$7E,$6F,$DB,$BE
	.dc.b	$2A,$92,$7E,$F4,$65,$FC,$07,$DB,$FF,$5C,$6F,$93,$F5,$A3,$18,$FE
	.dc.b	$99,$AF,$F9,$87,$ED,$FF,$08,$DE,$5F,$A5,$18,$8F,$E6,$7B,$A3,$10
	.dc.b	$87,$A6,$F9,$3F,$B4,$DC,$BE,$C3,$6B,$FC,$F3,$5C,$2B,$F0,$1E,$07
	.dc.b	$FA,$06,$FE,$F3,$6B,$6D,$7F,$CF,$0D,$7C,$AB,$E1,$79,$92,$79,$BF
	.dc.b	$8D,$FC,$95,$F1,$B4,$93,$F8,$D8,$CB,$FC,$83,$F5,$2F,$F7,$63,$7C
	.dc.b	$3F,$D3,$E2,$9C,$BF,$CB,$DA,$FF,$A5,$BE,$4D,$F0,$1C,$1B,$7F,$ED
	.dc.b	$18,$C1,$D3,$EE,$1F,$51,$49,$3F,$CC,$A3,$2F,$F3,$4F,$C2,$1F,$45
	.dc.b	$E4,$6C,$66,$30,$78,$7E,$1E,$47,$FB,$85,$B9,$7F,$47,$F7,$DF,$5D
	.dc.b	$EE,$6D,$7F,$42,$6C,$EE,$7F,$71,$F2,$1F,$DE,$EB,$FD,$61,$E8,$60
	.dc.b	$8F,$F4,$AF,$96,$FC,$C5,$FF,$55,$6B,$FE,$BC,$6B,$FE,$F1,$6B,$F8
	.dc.b	$FA,$FC,$8F,$3F,$FB,$7F,$AB,$C2,$4B,$FF,$60,$77,$79,$08,$E0,$F1
	.dc.b	$47,$DA,$C3,$F3,$78,$1A,$76,$C1,$F2,$DE,$E2,$5F,$FB,$35,$FE,$53
	.dc.b	$2F,$AC,$FD,$FD,$F3,$6C,$66,$B7,$1A,$97,$E5,$EF,$A7,$FC,$6B,$6B
	.dc.b	$FB,$8B,$67,$FB,$E3,$5F,$F1,$38,$E6,$FF,$3F,$7C,$DF,$F3,$C6,$D7
	.dc.b	$F8,$0D,$B3,$F9,$5F,$86,$57,$F4,$2F,$89,$7C,$A7,$F1,$6C,$3F,$DE
	.dc.b	$3E,$3E,$63,$18,$7F,$B6,$A3,$1C,$D2,$BF,$DE,$ED,$9F,$F6,$FF,$BF
	.dc.b	$C4,$93,$FE,$FC,$7A,$7B,$D2,$3F,$DF,$ED,$9F,$C2,$6F,$F4,$A7,$F1
	.dc.b	$0D,$8C,$F0,$FC,$C3,$F6,$BF,$AE,$7E,$22,$0D,$27,$2F,$D2,$25,$F4
	.dc.b	$6C,$FF,$BD,$BE,$4F,$E0,$12,$7F,$C5,$9B,$7C,$68,$F8,$9D,$D2,$73
	.dc.b	$BF,$ED,$AF,$FD,$FF,$3F,$27,$BE,$56,$9B,$3F,$68,$61,$F4,$BE,$8B
	.dc.b	$E8,$12,$BA,$BF,$C4,$9F,$E1,$F3,$F8,$26,$BF,$E6,$A6,$BF,$ED,$EF
	.dc.b	$0B,$87,$B9,$27,$FC,$7A,$1C,$4F,$F9,$26,$3F,$D4,$46,$BE,$1B,$C0
	.dc.b	$47,$DF,$F1,$DF,$85,$23,$83,$E1,$FD,$77,$E2,$48,$E0,$FB,$7E,$17
	.dc.b	$E9,$08,$E0,$FB,$FE,$1F,$EA,$08,$E0,$FF,$2A,$FD,$AB,$9F,$F9,$AB
	.dc.b	$63,$31,$83,$3D,$12,$38,$B7,$3F,$F3,$45,$7E,$79,$FA,$8D,$7F,$5E
	.dc.b	$D7,$FD,$2E,$DE,$F7,$D0,$BE,$9F,$F8,$2B,$70,$FE,$40,$DB,$FE,$40
	.dc.b	$D7,$F9,$03,$FE,$28,$AE,$7E,$A7,$22,$57,$E3,$BE,$03,$F6,$84,$70
	.dc.b	$7E,$4B,$E3,$BF,$70,$47,$07,$F8,$AF,$EB,$BF,$A9,$E3,$FD,$D1,$AF
	.dc.b	$F0,$CF,$21,$18,$8F,$F0,$1B,$5F,$CB,$F5,$3E,$D7,$F8,$31,$2F,$F4
	.dc.b	$6F,$F0,$9F,$AE,$91,$FA,$66,$D7,$87,$97,$FC,$85,$81,$A4,$AF,$46
	.dc.b	$7A,$23,$83,$F6,$3F,$59,$FC,$5C,$8E,$0F,$81,$A7,$62,$38,$3F,$73
	.dc.b	$F0,$3F,$C7,$C8,$E0,$FD,$EE,$3F,$C8,$C8,$E0,$F9,$99,$7A,$88,$E0
	.dc.b	$FF,$2D,$7E,$FA,$C3,$FC,$C0,$D8,$CC,$60,$FE,$69,$F8,$9F,$FB,$CE
	.dc.b	$3F,$88,$B6,$FF,$84,$77,$23,$11,$FD,$12,$31,$F1,$BC,$BE,$DE,$5F
	.dc.b	$26,$D7,$2B,$7D,$87,$80,$95,$E0,$7C,$7F,$6D,$23,$C6,$DA,$FE,$F5
	.dc.b	$C7,$FC,$45,$A7,$F6,$67,$0F,$F7,$D6,$BF,$F0,$03,$9B,$FA,$8E,$9B
	.dc.b	$25,$FF,$81,$EB,$E3,$57,$9F,$D4,$F8,$A4,$9F,$8A,$FC,$8F,$A6,$91
	.dc.b	$F2,$1B,$3E,$6E,$6F,$21,$27,$B3,$E7,$E3,$91,$E1,$6C,$FC,$7F,$A8
	.dc.b	$FA,$A4,$9F,$AA,$FF,$A6,$FE,$D9,$F6,$09,$F8,$3D,$49,$CC,$61,$B7
	.dc.b	$FE,$BE,$9D,$57,$18,$C4,$BD,$2C,$5F,$ED,$E3,$0F,$38,$CC,$FF,$9E
	.dc.b	$A7,$2F,$FB,$17,$ED,$BF,$D1,$DD,$3F,$EF,$E6,$C7,$F5,$30,$E1,$FE
	.dc.b	$AC,$D7,$FC,$22,$FF,$B5,$78,$79,$7B,$1F,$E2,$B9,$E5,$39,$6C,$65
	.dc.b	$FE,$1D,$E8,$BE,$03,$FA,$D4,$7F,$A4,$23,$0F,$C8,$6B,$F1,$5B,$9F
	.dc.b	$2F,$AD,$E3,$3F,$E9,$59,$46,$36,$FF,$1B,$36,$33,$3F,$94,$EB,$FD
	.dc.b	$57,$AD,$F0,$63,$FB,$9F,$57,$F8,$99,$FD,$8C,$66,$77,$1F,$E5,$BF
	.dc.b	$C4,$9E,$7F,$F9,$7D,$B1,$98,$C1,$9F,$42,$47,$16,$BF,$B3,$CB,$B6
	.dc.b	$3E,$68,$C1,$FC,$0A,$33,$F5,$F2,$3F,$82,$B5,$FF,$36,$61,$FE,$8C
	.dc.b	$7C,$DE,$97,$94,$8F,$A7,$D3,$F4,$11,$C1,$FC,$43,$D1,$F1,$91,$C1
	.dc.b	$FC,$53,$D2,$F3,$91,$C1,$FC,$63,$D3,$F5,$91,$C1,$FC,$73,$8F,$DE
	.dc.b	$47,$07,$F2,$83,$EA,$71,$3C,$6F,$AD,$1F,$ED,$36,$C6,$63,$7E,$56
	.dc.b	$08,$E0,$F9,$BE,$5C,$7F,$96,$B5,$FE,$49,$E6,$FD,$22,$3F,$CC,$9F
	.dc.b	$1F,$E8,$4F,$26,$92,$7F,$55,$82,$75,$1F,$CE,$BE,$E3,$EE,$91,$C1
	.dc.b	$C2,$1E,$4F,$62,$46,$5F,$0F,$8D,$23,$83,$5F,$E7,$9E,$7F,$E2,$88
	.dc.b	$FF,$4A,$7F,$1B,$F6,$B2,$3F,$A6,$B5,$FF,$7B,$DF,$E8,$5B,$D9,$F0
	.dc.b	$08,$FF,$54,$F9,$6F,$CB,$91,$C1,$FD,$63,$C2,$FC,$D9,$1C,$1F,$EC
	.dc.b	$3F,$C5,$3F,$4E,$FF,$B0,$DA,$FF,$B0,$DA,$FF,$47,$6B,$FE,$B9,$7F
	.dc.b	$3C,$F1,$E8,$FF,$68,$46,$23,$94,$7E,$23,$81,$EB,$D2,$4F,$F0,$18
	.dc.b	$CB,$7D,$7E,$04,$70,$6F,$EC,$7A,$11,$C1,$FE,$DF,$F9,$B7,$DC,$7F
	.dc.b	$B8,$DB,$19,$8C,$19,$FA,$69,$1C,$5A,$FF,$80,$3C,$64,$7F,$13,$6C
	.dc.b	$7F,$A5,$9A,$EE,$35,$F2,$32,$34,$D7,$FA,$07,$DB,$FE,$59,$BD,$3F
	.dc.b	$AE,$8C,$47,$FA,$5D,$AF,$FB,$49,$7A,$FB,$24,$7F,$2B,$8C,$BF,$CB
	.dc.b	$7F,$13,$FD,$CA,$69,$1F,$66,$51,$F6,$35,$F8,$5E,$87,$C1,$23,$F0
	.dc.b	$5F,$89,$F0,$F2,$3E,$23,$5F,$95,$B7,$C9,$23,$F2,$5F,$E8,$5C,$8F
	.dc.b	$C5,$3F,$D4,$D3,$FD,$96,$31,$E2,$6A,$73,$3F,$EB,$B9,$CB,$FD,$2F
	.dc.b	$EC,$5F,$05,$FC,$61,$B1,$98,$C1,$CF,$C7,$91,$DE,$D1,$88,$FC,$17
	.dc.b	$2F,$E0,$48,$FE,$05,$FC,$27,$3C,$8F,$C2,$35,$FC,$57,$CD,$FE,$24
	.dc.b	$8F,$E2,$5F,$C6,$7A,$D2,$3F,$18,$D7,$FD,$B1,$F1,$2F,$D3,$5F,$09
	.dc.b	$FD,$A6,$30,$D8,$CC,$FF,$83,$67,$2F,$E0,$FE,$27,$F9,$DF,$29,$A4
	.dc.b	$8F,$85,$E8,$FC,$F2,$38,$3F,$94,$70,$FF,$34,$23,$83,$FB,$E7,$D8
	.dc.b	$3E,$23,$FA,$4B,$63,$31,$83,$FA,$7F,$DC,$FE,$AA,$3F,$D4,$8D,$43
	.dc.b	$CC,$46,$63,$0D,$8C,$F7,$74,$24,$FF,$46,$94,$79,$3D,$A9,$3E,$86
	.dc.b	$BE,$03,$C3,$5E,$0E,$CF,$39,$23,$A3,$FD,$9B,$F1,$85,$52,$4F,$F8
	.dc.b	$7E,$32,$FC,$6E,$5F,$8A,$47,$07,$F8,$97,$E7,$5D,$9F,$DC,$9B,$19
	.dc.b	$8C,$1F,$DD,$BE,$CD,$EA,$D2,$4F,$F8,$BE,$32,$FD,$6D,$3E,$A9,$1C
	.dc.b	$1E,$1F,$03,$C0,$47,$07,$F7,$0F,$B3,$7C,$CF,$F8,$21,$B1,$98,$C1
	.dc.b	$FE,$0C,$FB,$1F,$F6,$D4,$7F,$78,$6B,$F3,$CF,$BC,$E1,$FF,$73,$35
	.dc.b	$FA,$6B,$E1,$7A,$12,$3A,$3F,$E1,$AF,$B2,$FD,$DC,$7F,$81,$9A,$F7
	.dc.b	$E2,$46,$23,$FD,$E8,$D7,$F0,$3E,$17,$F3,$27,$F2,$4D,$9F,$E1,$FE
	.dc.b	$47,$F8,$A2,$35,$F0,$9F,$D2,$3C,$DF,$F2,$F3,$6F,$FD,$00,$D7,$E0
	.dc.b	$35,$FC,$82,$FB,$2F,$81,$08,$C7,$C4,$74,$9C,$C6,$27,$FC,$47,$F5
	.dc.b	$2F,$9D,$FF,$1D,$B6,$33,$18,$3C,$BC,$5F,$84,$8F,$1B,$5F,$1B,$5F
	.dc.b	$D9,$D7,$EA,$B9,$3D,$BA,$7E,$CE,$3F,$C6,$DE,$7E,$09,$5F,$E3,$6F
	.dc.b	$AF,$FC,$2B,$F4,$5A,$FE,$BF,$74,$62,$3F,$62,$D7,$F8,$17,$E2,$DF
	.dc.b	$CF,$1F,$B4,$D7,$93,$5F,$DC,$B5,$F9,$5F,$60,$F8,$3F,$AA,$D8,$FA
	.dc.b	$AD,$7E,$AB,$5F,$AB,$F6,$1F,$78,$99,$FD,$E7,$F0,$3F,$61,$FC,$7E
	.dc.b	$7E,$EB,$5F,$BA,$D7,$E5,$B5,$FE,$3F,$C1,$FB,$AF,$B0,$FE,$43,$ED
	.dc.b	$BE,$5B,$F8,$A4,$FF,$AD,$A3,$15,$C3,$E1,$FF,$8C,$46,$21,$3C,$D4
	.dc.b	$66,$30,$D9,$E3,$5D,$9B,$5E,$4E,$26,$1E,$4A,$F1,$C6,$27,$AC,$71
	.dc.b	$74,$79,$4A,$79,$23,$85,$F8,$DB,$3C,$67,$99,$C4,$A7,$1A,$E7,$87
	.dc.b	$AA,$78,$5F,$AE,$B3,$7D,$1F,$55,$FF,$B4,$6F,$E2,$47,$F1,$16,$BE
	.dc.b	$C8,$D7,$AC,$FB,$0A,$49,$FF,$7D,$52,$75,$1F,$EA,$8F,$B0,$FF,$18
	.dc.b	$91,$C1,$3D,$DF,$AC,$FE,$F7,$E4,$3F,$AF,$C6,$21,$38,$D3,$F8,$3D
	.dc.b	$59,$FE,$51,$35,$FD,$AC,$57,$FB,$4C,$5E,$74,$E6,$7A,$9F,$30,$F1
	.dc.b	$14,$CE,$BE,$69,$E3,$19,$AE,$95,$CD,$B0,$C3,$D5,$18,$6C,$67,$0F
	.dc.b	$5B,$4F,$E3,$DF,$60,$FA,$2E,$FB,$78,$32,$4E,$72,$FF,$02,$CE,$61
	.dc.b	$F4,$FC,$27,$F8,$BC,$6F,$0F,$CC,$C6,$30,$D4,$7F,$2D,$FB,$07,$16
	.dc.b	$32,$EF,$56,$73,$7E,$03,$F9,$1F,$D8,$63,$0E,$A7,$F3,$77,$19,$FA
	.dc.b	$5F,$EE,$B3,$C6,$75,$8F,$7F,$C5,$BF,$D1,$F3,$FD,$19,$B3,$FD,$19
	.dc.b	$AF,$F4,$66,$BF,$E1,$6F,$8E,$FE,$E7,$3F,$EE,$46,$C7,$F5,$F6,$BF
	.dc.b	$CB,$DA,$FF,$3C,$F3,$7F,$C5,$B3,$E5,$D1,$18,$9F,$6F,$72,$FA,$E7
	.dc.b	$F5,$6B,$4F,$D7,$FA,$FF,$EA,$58,$F3,$6E,$1F,$1B,$14,$3E,$93,$E8
	.dc.b	$3A,$BF,$C7,$E4,$FC,$21,$F7,$5F,$92,$C7,$FA,$44,$63,$2F,$F0,$54
	.dc.b	$7F,$91,$E3,$0F,$F7,$F6,$BF,$DB,$BE,$C7,$FC,$6F,$3E,$F7,$F0,$D8
	.dc.b	$7E,$3D,$CC,$F9,$DC,$CF,$F6,$2F,$B1,$FA,$B5,$F9,$56,$C7,$F6,$6F
	.dc.b	$D6,$5E,$E7,$E2,$E7,$38,$7F,$96,$FE,$C9,$F6,$5E,$B8,$FF,$27,$C6
	.dc.b	$30,$F5,$EA,$8C,$4F,$F8,$65,$AF,$F9,$5F,$E7,$FF,$CC,$DF,$67,$F9
	.dc.b	$C7,$FC,$D1,$EA,$DB,$4D,$8F,$D0,$C3,$F0,$9F,$F2,$7E,$3F,$95,$8C
	.dc.b	$7B,$5C,$FF,$33,$FF,$39,$CE,$6B,$79,$7E,$13,$D9,$7D,$7F,$AE,$3F
	.dc.b	$CF,$11,$8A,$FF,$24,$4E,$71,$4E,$6B,$FC,$69,$F6,$BF,$83,$BF,$D3
	.dc.b	$6B,$FD,$63,$E6,$FC,$F3,$9A,$FD,$7C,$E6,$BD,$1F,$06,$F9,$2F,$6B
	.dc.b	$63,$9F,$CA,$BE,$FB,$35,$FC,$0B,$07,$DB,$7E,$D9,$FC,$5B,$5F,$F2
	.dc.b	$B3,$5F,$F4,$33,$5F,$F4,$BF,$C6,$BE,$63,$E1,$42,$31,$8F,$FA,$E6
	.dc.b	$73,$18,$9F,$F3,$16,$6C,$9F,$80,$FC,$B8,$C3,$63,$38,$7E,$82,$72
	.dc.b	$FE,$8A,$9F,$6C,$F8,$6F,$1B,$67,$4D,$9F,$D4,$B5,$FF,$3A,$7E,$15
	.dc.b	$FE,$27,$9E,$8D,$8F,$D7,$B5,$DF,$B7,$E5,$E0,$95,$E5,$CB,$F8,$55
	.dc.b	$7E,$5D,$AF,$FA,$BA,$5F,$29,$F8,$56,$D7,$FA,$E6,$30,$7C,$1E,$D9
	.dc.b	$1D,$5B,$5F,$2A,$FF,$A1,$CF,$87,$D0,$F9,$04,$7F,$D8,$8F,$72,$38
	.dc.b	$3D,$1E,$7F,$CF,$23,$83,$F8,$7F,$82,$FD,$45,$F5,$3F,$D3,$23,$0F
	.dc.b	$EC,$F0,$FE,$17,$E2,$FF,$6A,$9C,$DF,$F4,$3C,$9F,$85,$3E,$37,$F1
	.dc.b	$2D,$9F,$B8,$D9,$FE,$3E,$D7,$FD,$B1,$5E,$3E,$7F,$16,$7F,$8D,$C7
	.dc.b	$F9,$45,$F1,$7C,$51,$9C,$3F,$DD,$FC,$BF,$91,$F9,$1F,$47,$C3,$7A
	.dc.b	$5B,$3F,$AA,$8F,$67,$9D,$8D,$3F,$67,$B7,$ED,$72,$FF,$03,$E3,$FE
	.dc.b	$93,$FB,$CF,$E2,$98,$1F,$89,$8E,$1C,$F8,$F2,$3F,$80,$CE,$DB,$1E
	.dc.b	$CF,$95,$91,$AF,$26,$E7,$FD,$CB,$B9,$97,$FC,$00,$AF,$E8,$B1,$8F
	.dc.b	$88,$FE,$2C,$D8,$CC,$EE,$BF,$DA,$9C,$7F,$77,$ED,$FD,$3E,$8F,$CC
	.dc.b	$6D,$F9,$BF,$02,$F6,$FC,$D7,$B9,$B1,$EA,$8F,$83,$E5,$E3,$1F,$E3
	.dc.b	$3C,$1F,$C8,$78,$BE,$3A,$E0,$F8,$E7,$C8,$CA,$E3,$F1,$3C,$0F,$96
	.dc.b	$63,$0E,$6F,$F1,$BF,$BA,$FD,$E6,$3F,$D9,$DB,$5F,$7B,$1F,$C0,$58
	.dc.b	$4E,$9C,$2F,$97,$FE,$36,$DA,$FE,$1C,$D9,$FE,$E2,$D7,$FD,$F8,$FC
	.dc.b	$BB,$F1,$71,$79,$6F,$CC,$FE,$13,$EB,$E6,$1F,$80,$EE,$B8,$FC,$6F
	.dc.b	$B8,$FC,$49,$8C,$7E,$3B,$FA,$FF,$DD,$3F,$21,$7E,$BE,$44,$7E,$C2
	.dc.b	$30,$EA,$FF,$BA,$B9,$6F,$CD,$73,$F0,$BA,$36,$BF,$6D,$F7,$4C,$7F
	.dc.b	$A4,$36,$3F,$C7,$9D,$2F,$C2,$7F,$B3,$E3,$FE,$44,$15,$B7,$0B,$6F
	.dc.b	$E8,$ED,$BF,$E7,$4D,$9F,$F0,$EB,$5F,$B0,$7D,$97,$F8,$8C,$79,$B6
	.dc.b	$F2,$D7,$99,$FD,$8F,$CB,$FE,$CB,$E6,$6D,$7E,$B9,$C3,$FB,$0F,$87
	.dc.b	$B8,$FE,$59,$FD,$CF,$CB,$FE,$EB,$EE,$6D,$7E,$D9,$C3,$FB,$8F,$BB
	.dc.b	$B7,$F9,$E7,$FC,$1F,$F6,$5E,$68,$F4,$8F,$69,$19,$8F,$6C,$63,$4F
	.dc.b	$F0,$44,$E6,$7F,$35,$F0,$9F,$C3,$0E,$B3,$70,$FF,$04,$FD,$3B,$EA
	.dc.b	$BF,$C4,$8D,$8F,$E1,$B7,$89,$CF,$CB,$FF,$6F,$A3,$49,$5F,$E3,$4C
	.dc.b	$7F,$DE,$8D,$DB,$CA,$47,$FB,$A5,$3E,$EF,$C2,$7C,$29,$B3,$FE,$2F
	.dc.b	$A7,$07,$B8,$91,$DD,$F2,$7D,$8F,$A8,$F9,$6D,$8E,$5B,$FC,$E3,$DF
	.dc.b	$F7,$56,$32,$7C,$4F,$80,$C6,$DF,$E2,$AD,$11,$F5,$B8,$3E,$79,$2B
	.dc.b	$E0,$3F,$17,$EE,$BE,$59,$B8,$F9,$EF,$CE,$FA,$79,$1F,$4B,$EC,$2E
	.dc.b	$3D,$FE,$DF,$E7,$12,$7E,$3B,$F4,$7E,$BB,$EB,$9F,$61,$B1,$FE,$A9
	.dc.b	$F0,$9F,$6E,$E9,$FE,$E0,$D9,$D2,$E0,$CC,$EB,$AB,$FC,$AD,$7C,$9B
	.dc.b	$46,$9F,$83,$F9,$BF,$C4,$FC,$AF,$E2,$9B,$8F,$E3,$A3,$F1,$BF,$3D
	.dc.b	$8D,$7E,$63,$E4,$E8,$95,$F9,$4F,$95,$B5,$FE,$C8,$CF,$D7,$DF,$E2
	.dc.b	$25,$7F,$83,$B4,$47,$E5,$7E,$DB,$F3,$7F,$4F,$F9,$C6,$D7,$E8,$3E
	.dc.b	$DB,$FD,$EE,$FF,$A1,$7A,$51,$8C,$BF,$BF,$C6,$27,$F8,$BC,$7F,$68
	.dc.b	$8C,$3F,$EA,$FF,$09,$F7,$5F,$E8,$0A,$FD,$AB,$67,$FA,$54,$EB,$E4
	.dc.b	$BB,$47,$EF,$67,$4F,$F4,$FF,$BA,$7C,$03,$B7,$FC,$B1,$18,$6C,$66
	.dc.b	$7F,$D7,$B3,$96,$F9,$9F,$EA,$E6,$F0,$F2,$8C,$47,$F0,$E9,$46,$23
	.dc.b	$CF,$EF,$B9,$7F,$95,$EB,$FC,$42,$31,$3F,$13,$5F,$F3,$A5,$6A,$38
	.dc.b	$B5,$F9,$1F,$71,$7D,$D7,$02,$30,$D9,$F9,$F3,$A9,$CE,$9F,$CF,$3D
	.dc.b	$2F,$F1,$4D,$BE,$8B,$1A,$7D,$86,$E3,$F5,$EE,$8F,$A2,$FB,$F1,$FE
	.dc.b	$56,$6B,$FE,$1C,$6B,$FE,$48,$6B,$FD,$A5,$72,$F4,$6C,$7E,$C5,$AF
	.dc.b	$FA,$CB,$E8,$9F,$55,$E4,$D5,$1F,$E0,$99,$FF,$5D,$AF,$5F,$4D,$8C
	.dc.b	$FB,$18,$F9,$0C,$B2,$38,$C7,$8F,$AF,$FD,$3D,$A7,$BD,$B9,$9A,$63
	.dc.b	$BB,$D4,$4B,$F5,$78,$1E,$2F,$93,$91,$E6,$7C,$BD,$B2,$3F,$D5,$DA
	.dc.b	$23,$FD,$5D,$B7,$B9,$B1,$99,$DF,$DA,$7B,$88,$FC,$2F,$8C,$F8,$3F
	.dc.b	$69,$23,$E2,$BF,$73,$A9,$FA,$DF,$F5,$99,$FF,$7A,$35,$FE,$D3,$1A
	.dc.b	$F5,$BF,$D8,$0D,$8F,$ED,$A7,$F6,$D9,$E0,$FF,$6C,$7F,$AA,$3F,$53
	.dc.b	$E4,$7E,$99,$E1,$8C,$C3,$AD,$D5,$FB,$E9,$CC,$62,$7B,$7A,$DC,$4F
	.dc.b	$BB,$F2,$FF,$7B,$E5,$A4,$FE,$09,$ED,$EB,$C8,$F3,$E7,$F0,$23,$D7
	.dc.b	$89,$A7,$CE,$F9,$4F,$C4,$92,$BE,$2F,$AD,$E9,$38,$91,$EB,$5F,$D7
	.dc.b	$9F,$B7,$F2,$9F,$60,$8E,$0F,$BF,$F0,$5F,$96,$23,$F9,$AF,$94,$FC
	.dc.b	$CF,$91,$23,$F3,$AF,$C9,$F5,$64,$78,$1F,$59,$1F,$C5,$FC,$8F,$E8
	.dc.b	$BC,$C9,$1F,$A6,$8F,$D2,$FC,$3F,$E9,$99,$1F,$AA,$8F,$D7,$7A,$5F
	.dc.b	$EA,$B8,$FD,$6F,$2F,$EB,$0C,$FC,$8E,$BE,$46,$CF,$F9,$FF,$9F,$FD
	.dc.b	$6B,$7C,$63,$EF,$5F,$FB,$0A,$15,$FC,$3A,$7F,$59,$E4,$5F,$13,$F7
	.dc.b	$11,$FE,$5D,$6E,$1F,$83,$8F,$D9,$79,$3B,$3F,$26,$FF,$A3,$E7,$F8
	.dc.b	$3B,$86,$BF,$98,$6B,$FE,$46,$FB,$7F,$E7,$75,$D5,$B9,$7F,$98,$9B
	.dc.b	$2F,$87,$7E,$10,$DB,$F9,$2F,$A9,$FD,$42,$BF,$C1,$4B,$83,$26,$B8
	.dc.b	$77,$C7,$6B,$71,$87,$E5,$67,$F1,$FE,$4F,$E9,$92,$32,$7F,$B4,$B8
	.dc.b	$9F,$25,$F6,$09,$1A,$F2,$3F,$C3,$71,$F7,$3E,$4F,$E8,$7D,$B3,$E2
	.dc.b	$69,$2F,$A8,$87,$CB,$F0,$3E,$7C,$65,$B1,$A8,$FA,$6D,$7F,$83,$AF
	.dc.b	$91,$F0,$92,$3C,$D9,$3E,$53,$F2,$3F,$2B,$27,$2F,$E5,$3E,$56,$47
	.dc.b	$06,$BF,$8E,$3E,$27,$CA,$FE,$6B,$2F,$CD,$E3,$27,$BC,$7E,$7B,$3D
	.dc.b	$46,$0F,$1B,$F7,$3D,$1F,$D1,$FC,$84,$9F,$BB,$1F,$D9,$BA,$7F,$C1
	.dc.b	$D5,$F9,$26,$CF,$F2,$66,$BF,$E9,$46,$BF,$EA,$EF,$5F,$FC,$2D,$EA
	.dc.b	$7E,$28,$91,$F1,$3E,$4F,$F6,$7F,$09,$23,$E3,$BF,$B6,$F8,$F9,$1F
	.dc.b	$29,$AF,$C3,$E6,$FF,$2D,$FC,$37,$EF,$3E,$2B,$F7,$BF,$13,$2B,$E8
	.dc.b	$C7,$F0,$0F,$8A,$D4,$60,$FA,$AF,$E7,$BD,$F9,$1F,$9E,$7F,$75,$F8
	.dc.b	$3B,$9E,$9F,$86,$71,$3F,$B6,$6B,$FD,$29,$AF,$C3,$F0,$BF,$7F,$DF
	.dc.b	$FE,$EF,$6C,$38,$7F,$C7,$9E,$4F,$F1,$69,$FE,$2F,$3F,$DA,$3B,$3F
	.dc.b	$2B,$BF,$F9,$EA,$30,$7F,$5A,$F9,$CF,$ED,$4F,$F5,$E6,$BF,$EC,$46
	.dc.b	$BF,$E5,$06,$A8,$FF,$14,$7B,$5F,$DD,$BE,$13,$FB,$0C,$8D,$38,$6C
	.dc.b	$6F,$E4,$E4,$7F,$6A,$AF,$F1,$94,$62,$3C,$9E,$D7,$9C,$8F,$F5,$AC
	.dc.b	$11,$E1,$F6,$E4,$F3,$C7,$FB,$6B,$FE,$FA,$C3,$FC,$7D,$3F,$B8,$FC
	.dc.b	$1F,$F6,$93,$FE,$E6,$6B,$FF,$00,$1A,$FA,$DA,$BE,$EB,$F6,$CF,$F3
	.dc.b	$06,$BF,$E0,$46,$BF,$EF,$26,$BF,$16,$BF,$CA,$1E,$DF,$C9,$24,$7A
	.dc.b	$5F,$E1,$BF,$73,$FE,$3D,$8F,$F0,$7B,$5F,$F7,$EB,$5F,$A0,$1C,$3F
	.dc.b	$4F,$EA,$5C,$DF,$C6,$1B,$1F,$D2,$1A,$FF,$87,$1A,$F4,$F7,$9F,$94
	.dc.b	$7F,$7D,$8F,$E1,$D1,$8F,$7B,$FC,$59,$39,$F5,$7F,$C3,$1C,$AF,$C2
	.dc.b	$1E,$3F,$F8,$C7,$E8,$DE,$7F,$F9,$6E,$9F,$07,$4D,$47,$F8,$DF,$C6
	.dc.b	$FF,$46,$6B,$FE,$A2,$6A,$D5,$18,$6F,$6F,$F8,$E3,$E0,$3A,$D7,$F4
	.dc.b	$89,$46,$23,$D6,$DD,$7F,$C6,$1C,$8F,$C2,$3F,$AE,$57,$F0,$B3,$F5
	.dc.b	$33,$A5,$E4,$FF,$9D,$5A,$7F,$51,$37,$E0,$FF,$AE,$BC,$9F,$CA,$C5
	.dc.b	$EB,$FF,$26,$C6,$23,$D8,$DE,$0F,$F4,$0F,$CA,$7F,$A5,$1B,$3F,$D8
	.dc.b	$9A,$FF,$41,$7F,$D0,$FF,$1D,$BF,$B8,$FF,$7D,$4E,$DF,$F7,$C3,$7C
	.dc.b	$9D,$79,$3F,$E6,$5F,$C2,$3F,$DE,$D3,$FE,$48,$6B,$FE,$64,$6B,$F8
	.dc.b	$56,$BF,$C0,$70,$FF,$49,$7C,$77,$9F,$C9,$34,$C7,$C7,$7A,$FC,$8F
	.dc.b	$E3,$B4,$69,$FC,$83,$99,$F7,$2F,$6F,$FC,$F4,$DC,$BF,$CB,$AD,$7F
	.dc.b	$CF,$2D,$7F,$34,$D7,$CD,$1F,$F0,$01,$FF,$3D,$DF,$5E,$1F,$F1,$5B
	.dc.b	$63,$33,$B7,$F5,$E7,$FB,$B7,$EE,$3F,$C1,$4F,$D3,$6B,$F4,$DB,$CD
	.dc.b	$FA,$47,$FC,$08,$FE,$9B,$ED,$FF,$C5,$53,$F6,$FE,$E3,$FD,$03,$CF
	.dc.b	$F2,$DB,$1F,$EF,$B9,$DB,$F7,$DB,$5C,$FE,$0B,$F0,$D1,$8B,$FA,$8F
	.dc.b	$F8,$CE,$BF,$C3,$BF,$6F,$FE,$F5,$BF,$E7,$1F,$26,$F5,$BF,$85,$CB
	.dc.b	$8B,$F8,$6C,$E6,$FF,$9F,$4E,$5E,$35,$C3,$E6,$3F,$D3,$91,$FE,$D6
	.dc.b	$8C,$61,$A8,$E3,$87,$49,$EE,$D8,$EB,$5C,$B0,$F0,$43,$0F,$43,$FE
	.dc.b	$F9,$BE,$2F,$F5,$1F,$43,$73,$E5,$8D,$1D,$31,$D5,$79,$23,$CC,$79
	.dc.b	$0F,$31,$DF,$1F,$23,$E1,$C7,$C9,$3C,$71,$EF,$5A,$C3,$C1,$1E,$2A
	.dc.b	$DD,$F8,$4F,$F6,$4C,$FA,$9F,$C5,$B5,$F2,$18,$F9,$C7,$DC,$7F,$9F
	.dc.b	$A5,$18,$9E,$6D,$78,$4F,$F9,$59,$7C,$BD,$23,$FE,$5A,$6E,$1F,$3A
	.dc.b	$1C,$7F,$ED,$18,$C4,$72,$3E,$39,$FB,$FD,$FE,$3B,$FE,$01,$AB,$5C
	.dc.b	$74,$FF,$73,$C6,$2B,$75,$CF,$6F,$05,$FF,$B7,$DE,$F5,$F5,$A7,$FA
	.dc.b	$EC,$E6,$F9,$C7,$88,$F3,$DF,$4A,$ED,$19,$3D,$18,$76,$9D,$47,$1B
	.dc.b	$F3,$3F,$03,$EE,$3F,$A2,$CF,$F2,$D6,$C7,$F2,$D5,$F2,$9F,$EF,$5E
	.dc.b	$0F,$EB,$F1,$A7,$FB,$0F,$DC,$38,$35,$87,$F2,$58,$C3,$D7,$C7,$CE
	.dc.b	$3F,$42,$72,$FF,$3B,$FB,$8F,$BF,$5F,$CF,$DB,$19,$FA,$1E,$3A,$27
	.dc.b	$39,$2F,$FC,$69,$D8,$FC,$1F,$FA,$BC,$7F,$63,$6E,$3D,$F9,$3C,$75
	.dc.b	$BF,$22,$F8,$0F,$94,$FF,$69,$5F,$48,$C6,$3F,$C9,$23,$FB,$A4,$61
	.dc.b	$FE,$4F,$7F,$D0,$AB,$8C,$7E,$3B,$D5,$7C,$CF,$A9,$B8,$7F,$77,$C3
	.dc.b	$96,$9F,$E0,$C9,$CD,$FF,$5F,$FB,$87,$C7,$FC,$25,$F2,$16,$30,$FB
	.dc.b	$3E,$E3,$FD,$E3,$97,$9F,$54,$63,$0F,$3B,$5F,$F1,$13,$5F,$EB,$DF
	.dc.b	$70,$F9,$AF,$A2,$DB,$FE,$E8,$FE,$0D,$DC,$7C,$CF,$B6,$FC,$4C,$7F
	.dc.b	$79,$FB,$87,$81,$9E,$DF,$B7,$18,$9F,$F0,$C5,$7C,$0C,$7F,$C6,$8D
	.dc.b	$7E,$EB,$5F,$F1,$23,$73,$FF,$0F,$DF,$DC,$EB,$F0,$D7,$FA,$96,$7E
	.dc.b	$6D,$FE,$1A,$73,$C3,$F9,$DF,$82,$7C,$CF,$E5,$BD,$77,$DB,$FE,$C3
	.dc.b	$2F,$F6,$85,$FF,$99,$6B,$53,$98,$FC,$94,$FF,$8D,$FD,$DF,$F7,$BE
	.dc.b	$5F,$45,$B1,$F8,$66,$BF,$E5,$16,$BF,$E9,$3E,$3F,$CA,$FC,$3F,$DE
	.dc.b	$AF,$F7,$F7,$0B,$EB,$33,$5A,$F9,$9F,$44,$BE,$13,$17,$C1,$E7,$E7
	.dc.b	$7E,$53,$89,$92,$30,$DA,$FF,$49,$46,$FC,$6F,$93,$A7,$F9,$4D,$BC
	.dc.b	$3F,$73,$EB,$3F,$57,$1A,$C7,$F8,$27,$A5,$E4,$FA,$BC,$B7,$2F,$F2
	.dc.b	$B5,$7F,$99,$9F,$CB,$37,$1F,$D5,$B5,$FF,$3B,$35,$FD,$FA,$F9,$9F
	.dc.b	$09,$F0,$BE,$B2,$D7,$6F,$99,$F8,$BF,$43,$E6,$6C,$78,$BE,$6F,$DC
	.dc.b	$47,$C6,$FE,$EB,$9D,$E2,$7F,$82,$67,$FD,$85,$1F,$C9,$63,$19,$A3
	.dc.b	$F0,$7F,$2E,$FC,$F5,$F3,$3F,$C8,$A7,$9F,$AB,$F2,$3E,$6F,$EC,$FA
	.dc.b	$57,$3F,$BE,$FC,$CB,$FA,$44,$8F,$8E,$FD,$5F,$8A,$FA,$27,$E0,$BD
	.dc.b	$0B,$3F,$88,$E3,$FE,$63,$AB,$F5,$47,$C9,$FE,$22,$73,$18,$AF,$69
	.dc.b	$FE,$B0,$FB,$57,$D4,$FF,$1C,$C1,$18,$9F,$E4,$91,$89,$D4,$E6,$3F
	.dc.b	$94,$53,$E3,$3B,$9E,$49,$D7,$07,$EA,$AB,$CD,$3F,$9C,$F4,$A6,$3F
	.dc.b	$A2,$F7,$ED,$7C,$F8,$FF,$68,$78,$7F,$BA,$7F,$4F,$E4,$EB,$1F,$F4
	.dc.b	$FC,$E6,$B8,$1F,$EF,$AF,$8B,$FE,$87,$F9,$FB,$AF,$F5,$71,$86,$63
	.dc.b	$FB,$2F,$8D,$B7,$CB,$C4,$FE,$3B,$1F,$D7,$BE,$3D,$F6,$BF,$B3,$BF
	.dc.b	$97,$69,$CF,$99,$BA,$D4,$F9,$27,$35,$C1,$B1,$FE,$74,$6B,$FE,$E4
	.dc.b	$6A,$3B,$4F,$F0,$79,$FE,$B1,$3E,$4E,$6F,$01,$E1,$BB,$3E,$6D,$9E
	.dc.b	$9F,$F0,$3C,$7F,$90,$B6,$3E,$5F,$D7,$7E,$6E,$31,$5F,$4B,$83,$F8
	.dc.b	$99,$9F,$2A,$C7,$F1,$AC,$FF,$97,$7D,$AB,$E3,$7F,$8C,$C6,$19,$3F
	.dc.b	$23,$81,$4F,$A8,$CE,$BF,$46,$31,$5F,$EE,$1D,$5F,$55,$FA,$0F,$94
	.dc.b	$7D,$0F,$F8,$02,$73,$5F,$53,$0F,$F7,$14,$EF,$CC,$FC,$F5,$FF,$0D
	.dc.b	$AF,$CE,$B6,$BF,$C0,$4D,$7F,$D4,$EF,$F3,$EB,$FC,$D3,$67,$EA,$4F
	.dc.b	$FB,$BE,$BC,$9F,$4F,$F0,$BE,$BB,$F9,$4F,$C3,$DB,$FE,$4A,$E1,$F8
	.dc.b	$5A,$3E,$77,$18,$FC,$17,$FC,$5B,$E6,$7F,$88,$FD,$6B,$7F,$8C,$D7
	.dc.b	$FA,$AF,$1F,$F8,$9F,$E4,$B3,$5F,$97,$A7,$93,$4C,$BF,$91,$32,$78
	.dc.b	$BE,$1F,$87,$A3,$F6,$9B,$1F,$9E,$CB,$FC,$69,$F4,$3F,$E2,$26,$C7
	.dc.b	$F8,$A5,$BA,$FF,$B3,$9A,$FE,$6E,$3E,$B9,$E1,$E4,$F5,$D7,$9D,$FD
	.dc.b	$BE,$12,$7E,$2F,$D3,$5B,$FC,$27,$D7,$7E,$38,$C6,$FF,$DD,$BA,$7F
	.dc.b	$17,$1F,$E4,$2F,$D6,$7F,$DE,$98,$7E,$DB,$DA,$46,$30,$FC,$77,$9D
	.dc.b	$BF,$98,$FC,$CF,$D7,$CC,$38,$5E,$F3,$ED,$CE,$AE,$DE,$2F,$F7,$1A
	.dc.b	$FE,$5D,$97,$F9,$21,$AF,$E1,$B3,$F0,$CF,$F9,$2B,$5F,$D0,$FC,$FF
	.dc.b	$EA,$0E,$79,$FE,$11,$FD,$5F,$C6,$C8,$FC,$3F,$D2,$DC,$7F,$8F,$E3
	.dc.b	$EC,$E8,$F8,$06,$76,$FF,$0C,$C6,$6F,$D7,$EF,$F9,$C8,$FF,$92,$F5
	.dc.b	$47,$97,$7F,$F1,$5C,$FF,$11,$D7,$F6,$FF,$11,$F3,$6B,$F7,$EF,$F9
	.dc.b	$C3,$AB,$E6,$69,$EF,$6B,$FE,$7A,$6A,$CB,$F2,$BC,$5D,$FE,$23,$F8
	.dc.b	$49,$CD,$B9,$7E,$9B,$C4,$FE,$1A,$7F,$0E,$F8,$AB,$1F,$5B,$E0,$3F
	.dc.b	$00,$63,$1F,$8B,$BA,$3F,$B3,$FC,$F3,$EA,$7F,$B0,$36,$FF,$AF,$B6
	.dc.b	$FF,$D0,$8D,$7F,$7E,$3E,$6F,$F9,$93,$FD,$40,$C7,$A1,$DA,$FF,$BA
	.dc.b	$CE,$B9,$3F,$13,$F2,$1F,$C9,$BE,$2B,$F9,$3B,$72,$FE,$07,$F1,$5F
	.dc.b	$44,$8F,$FA,$57,$9B,$F9,$3C,$7D,$5F,$43,$F9,$79,$FC,$C1,$B3,$FD
	.dc.b	$37,$ED,$7F,$86,$37,$CC,$FE,$B5,$18,$DB,$FA,$A2,$8C,$FA,$3F,$E9
	.dc.b	$EF,$2D,$3F,$CE,$EF,$FA,$7B,$6F,$FA,$C3,$5F,$E4,$79,$3E,$DF,$FC
	.dc.b	$96,$7F,$93,$3B,$7F,$D7,$97,$FD,$06,$33,$F1,$5F,$E7,$54,$DF,$B0
	.dc.b	$C3,$9F,$47,$5F,$49,$6F,$93,$FD,$D7,$E7,$3F,$D2,$19,$FF,$6B,$6B
	.dc.b	$FD,$AD,$AF,$F6,$BD,$51,$88,$69,$E9,$F3,$7C,$87,$97,$D4,$B8,$FA
	.dc.b	$7D,$CF,$F4,$2B,$F0,$36,$FF,$07,$3F,$DD,$CC,$46,$B0,$FD,$54,$E9
	.dc.b	$FF,$12,$4B,$F7,$2F,$F1,$53,$C6,$39,$46,$1D,$B6,$3F,$A8,$9C,$7E
	.dc.b	$0F,$B4,$FF,$81,$FE,$D7,$FD,$9B,$EC,$7F,$8E,$9B,$1F,$ED,$F6,$BF
	.dc.b	$9C,$6B,$FD,$71,$FF,$69,$ED,$C7,$AB,$FC,$3D,$3C,$5A,$F1,$6B,$F0
	.dc.b	$7F,$7D,$7A,$9E,$68,$C3,$5F,$F7,$F4,$69,$F8,$06,$23,$5C,$1F,$8B
	.dc.b	$6E,$BB,$9F,$F2,$64,$E5,$E2,$D7,$30,$F8,$CF,$EE,$FD,$DF,$D1,$E3
	.dc.b	$13,$F9,$39,$CE,$DF,$5B,$EC,$2C,$FF,$8F,$7D,$8F,$EE,$CF,$F8,$C1
	.dc.b	$AF,$F8,$E5,$AF,$F8,$E5,$AF,$E0,$3E,$A7,$9E,$5F,$A5,$F8,$DB,$97
	.dc.b	$FA,$2A,$BF,$49,$58,$F7,$24,$7E,$9D,$DF,$9F,$C3,$C6,$93,$FA,$B7
	.dc.b	$9F,$C7,$FA,$38,$DA,$7F,$8C,$EB,$D1,$E1,$7F,$9F,$27,$2D,$9D,$4F
	.dc.b	$F9,$4E,$7E,$EF,$0F,$F8,$3A,$7F,$C6,$D3,$FE,$59,$6B,$FE,$5D,$9D
	.dc.b	$F0,$7E,$7B,$EC,$3D,$A4,$9F,$E0,$1E,$57,$C0,$3F,$D1,$BC,$3F,$C1
	.dc.b	$DB,$C0,$69,$8E,$8F,$8F,$F5,$32,$FF,$85,$47,$D6,$5F,$72,$DF,$0D
	.dc.b	$8F,$F2,$23,$5F,$E7,$1F,$83,$3E,$27,$93,$63,$FD,$42,$D7,$FA,$08
	.dc.b	$70,$FF,$44,$6B,$FE,$77,$F5,$7F,$C1,$FF,$27,$F8,$A6,$C7,$F8,$32
	.dc.b	$3F,$07,$AB,$E0,$A9,$27,$FC,$CD,$19,$7F,$94,$1C,$3D,$68,$E9,$ED
	.dc.b	$CF,$FE,$B3,$8F,$F1,$03,$5F,$EB,$38,$7F,$91,$3C,$1F,$F5,$EB,$63
	.dc.b	$DA,$D7,$FA,$FD,$FF,$24,$9F,$39,$23,$FC,$4B,$F1,$DE,$DF,$54,$F4
	.dc.b	$6C,$7F,$83,$63,$06,$7E,$3B,$81,$23,$DF,$EA,$FC,$42,$38,$3E,$1F
	.dc.b	$C0,$7C,$8F,$76,$47,$C6,$7C,$5C,$32,$3F,$C3,$8F,$8C,$8F,$F6,$BC
	.dc.b	$DF,$2F,$F4,$89,$1F,$E4,$0F,$8A,$F6,$11,$C1,$F5,$3D,$FF,$B3,$F2
	.dc.b	$F2,$3E,$BB,$F7,$FE,$41,$DB,$49,$3F,$E5,$38,$CB,$F7,$BD,$DF,$DC
	.dc.b	$71,$FE,$86,$3F,$D8,$CD,$7F,$D9,$4D,$F0,$9E,$07,$FB,$17,$C2,$FF
	.dc.b	$6A,$FB,$FF,$C6,$1B,$7F,$E7,$5F,$8F,$FF,$54,$B6,$7F,$DF,$B3,$A8
	.dc.b	$FB,$FC,$0E,$CF,$1E,$B2,$7F,$27,$1E,$6F,$22,$47,$99,$F0,$62,$F8
	.dc.b	$3A,$49,$FF,$45,$46,$5F,$6F,$95,$FA,$5F,$C5,$7F,$88,$ED,$FE,$7B
	.dc.b	$CD,$F2,$DF,$A6,$35,$E5,$7E,$A3,$E5,$AE,$3F,$E7,$DC,$38,$F0,$FF
	.dc.b	$95,$F0,$57,$F0,$5F,$C0,$BF,$CF,$17,$FE,$C5,$6B,$FE,$39,$6E,$5F
	.dc.b	$32,$3F,$DE,$10,$F8,$BE,$47,$F0,$28,$CB,$54,$6D,$D3,$5F,$D8,$9F
	.dc.b	$E5,$33,$C5,$F1,$9E,$53,$F8,$33,$91,$5F,$0F,$E5,$7F,$83,$F8,$52
	.dc.b	$7F,$82,$C7,$F0,$AF,$0B,$F8,$5B,$23,$F8,$4C,$7F,$0E,$F9,$59,$1F
	.dc.b	$C3,$1F,$EB,$B6,$FC,$05,$EA,$61,$B4,$3C,$7B,$DA,$E1,$5B,$3F,$8D
	.dc.b	$F8,$5F,$D4,$7E,$4F,$E5,$5F,$01,$E9,$A3,$37,$FE,$DA,$6E,$1F,$E3
	.dc.b	$5E,$E7,$D1,$FE,$90,$FE,$B0,$E3,$0F,$E6,$78,$7F,$A2,$E7,$27,$F5
	.dc.b	$DB,$7C,$4F,$1A,$FD,$5C,$F9,$63,$F6,$DE,$A7,$F3,$2F,$97,$95,$FB
	.dc.b	$58,$FE,$6B,$E3,$EA,$30,$7F,$34,$7F,$9C,$E3,$23,$F9,$BB,$FD,$FB
	.dc.b	$D4,$FC,$9C,$7F,$93,$1B,$D2,$69,$3E,$5B,$FA,$D7,$CD,$7F,$A3,$B8
	.dc.b	$3E,$4B,$78,$3E,$96,$08,$C3,$F5,$7F,$4A,$7C,$17,$F6,$36,$C7,$E5
	.dc.b	$DA,$FF,$5A,$6B,$FE,$9C,$6B,$FE,$BF,$C7,$FD,$08,$FD,$FB,$F0,$FE
	.dc.b	$4F,$EC,$83,$EA,$F8,$DF,$DD,$7C,$B4,$9F,$DA,$3F,$0F,$E6,$7F,$7D
	.dc.b	$F2,$FF,$D4,$3C,$33,$49,$5F,$D4,$A1,$F2,$FF,$2F,$E6,$64,$FE,$FE
	.dc.b	$3F,$33,$F8,$23,$E3,$A9,$27,$D0,$7F,$88,$BD,$97,$CB,$FF,$0F,$F9
	.dc.b	$9F,$11,$27,$C0,$F9,$63,$ED,$7C,$D4,$9F,$E1,$D1,$FC,$5F,$E3,$24
	.dc.b	$7F,$11,$7F,$01,$E4,$FF,$1B,$F9,$8F,$71,$27,$D8,$FC,$1F,$03,$F0
	.dc.b	$DE,$64,$9F,$E3,$51,$FE,$07,$FA,$0F,$CF,$47,$FB,$F9,$BD,$9F,$EF
	.dc.b	$C6,$F4,$7D,$7E,$6E,$1F,$2D,$FD,$F2,$4D,$25,$FF,$7E,$8F,$90,$FC
	.dc.b	$6E,$FF,$CA,$73,$FE,$64,$F1,$71,$FF,$86,$7E,$D1,$EF,$52,$5F,$CE
	.dc.b	$ED,$7D,$A3,$E1,$7F,$C4,$1F,$82,$7F,$91,$AB,$F3,$3E,$5A,$31,$1F
	.dc.b	$E2,$76,$A8,$D7,$8B,$FC,$BF,$89,$FA,$07,$E5,$2B,$F8,$66,$FF,$C2
	.dc.b	$FE,$77,$FC,$DB,$E6,$9F,$E3,$B5,$BB,$FE,$79,$8B,$F0,$6F,$E8,$B8
	.dc.b	$FE,$1A,$31,$EB,$FF,$8D,$E7,$33,$D8,$75,$FF,$8E,$FE,$49,$E0,$FF
	.dc.b	$8A,$9B,$3F,$E2,$A6,$BF,$E3,$06,$BF,$09,$7C,$3F,$FA,$3B,$8D,$C0
	.dc.b	$F4,$30,$DB,$F3,$5F,$07,$E8,$7D,$BF,$F0,$2B,$34,$8F,$F9,$47,$ED
	.dc.b	$DF,$25,$FE,$40,$6D,$7F,$90,$1A,$FF,$91,$9A,$FF,$10,$97,$BA,$F6
	.dc.b	$7D,$0B,$DA,$A6,$23,$D9,$F1,$DF,$47,$C3,$FE,$95,$68,$C6,$1F,$C8
	.dc.b	$DB,$C8,$FB,$7F,$CA,$CF,$F4,$98,$C7,$81,$FE,$58,$9C,$CF,$E3,$1A
	.dc.b	$FF,$9C,$5A,$FE,$2F,$DF,$FF,$14,$63,$FE,$75,$6C,$67,$D9,$FA,$5F
	.dc.b	$6F,$F9,$98,$FC,$E3,$5F,$EB,$59,$A3,$1D,$5F,$E0,$78,$C7,$81,$CB
	.dc.b	$E1,$7E,$E7,$D1,$BE,$3E,$93,$1F,$D4,$7E,$15,$FB,$A7,$FA,$23,$70
	.dc.b	$34,$C7,$DC,$7F,$66,$9F,$EE,$9F,$71,$FE,$1E,$CF,$FB,$B7,$DB,$FF
	.dc.b	$A7,$63,$FC,$E3,$28,$C7,$C3,$FF,$4A,$75,$E1,$FF,$83,$A7,$5F,$33
	.dc.b	$FD,$3A,$FA,$46,$0E,$F7,$C6,$74,$D8,$F1,$D7,$3B,$EB,$87,$F8,$27
	.dc.b	$E0,$FC,$35,$70,$E2,$D3,$8C,$D5,$8F,$8A,$F8,$C7,$8C,$CE,$3E,$17
	.dc.b	$78,$F8,$70,$F5,$36,$3D,$27,$C6,$8F,$0B,$CE,$FB,$47,$0C,$3F,$C7
	.dc.b	$D1,$F0,$A7,$FB,$89,$DA,$33,$5F,$26,$33,$3F,$E5,$8F,$B8,$E3,$F1
	.dc.b	$46,$92,$7F,$C9,$1F,$49,$FE,$49,$4D,$23,$A1,$A7,$FB,$65,$FF,$05
	.dc.b	$73,$FC,$77,$FC,$6D,$37,$83,$FC,$B5,$5B,$9E,$A7,$2D,$3D,$73,$A7
	.dc.b	$FD,$BB,$5C,$6F,$A5,$70,$BF,$05,$EE,$3E,$1D,$6D,$E8,$60,$A7,$F0
	.dc.b	$0A,$E7,$1B,$AF,$23,$6F,$C3,$5C,$0F,$69,$D2,$DF,$71,$FE,$04,$C3
	.dc.b	$FD,$E7,$18,$F7,$BF,$C9,$F5,$FC,$3A,$B4,$FB,$9A,$FF,$1B,$FB,$8F
	.dc.b	$F4,$6D,$F2,$6C,$7C,$46,$BB,$D3,$AB,$FE,$39,$FB,$87,$CB,$63,$24
	.dc.b	$63,$0F,$2B,$6B,$E7,$B5,$FE,$C7,$F7,$0F,$86,$EB,$E0,$A3,$17,$FE
	.dc.b	$F4,$6D,$7F,$BE,$9A,$FF,$6C,$FD,$1B,$FD,$87,$3F,$EF,$A6,$C7,$FC
	.dc.b	$00,$6B,$B8,$F4,$7D,$97,$FA,$DD,$F9,$6D,$8F,$C2,$B5,$FE,$84,$D7
	.dc.b	$F9,$DF,$DC,$BE,$6B,$E1,$36,$BE,$AB,$67,$F1,$ED,$7F,$C0,$7F,$A8
	.dc.b	$79,$1F,$F4,$76,$48,$C4,$7F,$80,$DB,$19,$AF,$A1,$F7,$2F,$88,$A4
	.dc.b	$8F,$C2,$C6,$5F,$C3,$FD,$9F,$F6,$D6,$FA,$7F,$66,$31,$8F,$E9,$1A
	.dc.b	$FF,$86,$3E,$E7,$E0,$3F,$E1,$56,$E2,$69,$31,$FB,$6F,$86,$7F,$84
	.dc.b	$CE,$E7,$27,$E3,$3B,$9F,$72,$F9,$9C,$37,$3F,$E2,$38,$7F,$B1,$23
	.dc.b	$0F,$FA,$1D,$AF,$FA,$1A,$B9,$FA,$1C,$89,$5C,$A3,$2F,$F9,$0F,$EC
	.dc.b	$7F,$23,$F0,$A6,$92,$3F,$C9,$1F,$61,$A2,$38,$3C,$BA,$3C,$5F,$F2
	.dc.b	$8C,$61,$B9,$7C,$26,$BF,$E4,$46,$BF,$82,$C1,$F6,$3E,$42,$57,$F9
	.dc.b	$65,$F3,$FC,$B4,$8F,$F2,$E3,$6B,$E7,$D3,$EB,$BD,$A4,$AF,$F3,$13
	.dc.b	$F0,$3E,$72,$47,$F9,$95,$B5,$CB,$EB,$BE,$29,$27,$FC,$D4,$FE,$E3
	.dc.b	$2F,$F3,$04,$7B,$FD,$1C,$11,$FA,$1D,$6F,$A3,$FC,$03,$63,$31,$83
	.dc.b	$E9,$FD,$04,$8F,$F3,$C3,$71,$F6,$FA,$7F,$60,$93,$FE,$7D,$7E,$DE
	.dc.b	$12,$3F,$D0,$0D,$9F,$F7,$4F,$83,$FC,$71,$FF,$7A,$B7,$E7,$7F,$86
	.dc.b	$C7,$FB,$41,$A8,$FF,$7D,$FB,$9B,$25,$FE,$5A,$32,$D9,$FF,$2E,$B7
	.dc.b	$E0,$7B,$11,$FC,$EF,$C4,$FF,$B9,$7E,$8B,$F9,$7B,$63,$3E,$47,$EC
	.dc.b	$A3,$18,$7F,$74,$7F,$DD,$9E,$6F,$09,$FE,$7F,$3F,$ED,$B8,$C1,$D1
	.dc.b	$7C,$AD,$25,$F8,$A3,$2D,$F9,$5E,$04,$70,$7E,$D3,$E9,$1F,$AD,$3E
	.dc.b	$77,$FA,$3C,$61,$B1,$99,$E0,$DB,$FF,$80,$1F,$AE,$FF,$91,$E7,$F6
	.dc.b	$CD,$8C,$FB,$BF,$C1,$E3,$11,$F4,$C6,$1E,$2E,$2E,$E4,$BE,$EF,$8B
	.dc.b	$82,$47,$89,$B5,$FE,$1E,$A7,$EC,$0E,$2F,$5B,$6F,$E0,$36,$B8,$35
	.dc.b	$F2,$70,$B8,$7F,$91,$46,$1A,$FE,$2E,$35,$E3,$FD,$93,$ED,$7D,$C4
	.dc.b	$C3,$F0,$2D,$AF,$E4,$7F,$24,$F8,$3F,$CD,$36,$BF,$24,$D9,$DC,$EB
	.dc.b	$AE,$56,$9B,$3D,$3E,$CB,$99,$27,$9B,$F8,$DF,$9D,$91,$F9,$46,$CF
	.dc.b	$E4,$F4,$FE,$BC,$FB,$5A,$B4,$F8,$9E,$9F,$E5,$CF,$F0,$6E,$EF,$2B
	.dc.b	$D1,$5F,$B1,$6B,$FB,$18,$D6,$9C,$27,$33,$F5,$DF,$E1,$8F,$F9,$62
	.dc.b	$7A,$69,$B2,$67,$FD,$32,$3C,$1E,$17,$32,$46,$6D,$1F,$3B,$E6,$A4
	.dc.b	$FC,$E6,$E1,$DE,$BF,$98,$4F,$1F,$17,$CC,$47,$F5,$DF,$2B,$FB,$32
	.dc.b	$38,$3F,$65,$F3,$9F,$B7,$23,$83,$F8,$1F,$77,$EF,$C8,$E0,$FE,$0F
	.dc.b	$DD,$FC,$0C,$8E,$0F,$EA,$FF,$AF,$79,$B2,$FE,$EA,$D7,$F5,$9E,$62
	.dc.b	$31,$1F,$E1,$66,$BF,$6F,$8D,$97,$F1,$12,$5F,$F1,$87,$F8,$A6,$52
	.dc.b	$3F,$8D,$B6,$BF,$AB,$9F,$DC,$89,$1F,$DB,$3C,$AF,$E6,$0F,$A3,$C5
	.dc.b	$FF,$23,$E3,$FD,$42,$BD,$31,$83,$F9,$14,$C3,$F9,$49,$63,$DF,$C5
	.dc.b	$C8,$91,$F1,$1F,$E6,$3F,$72,$D6,$92,$7F,$9B,$C6,$5F,$E7,$3F,$7A
	.dc.b	$FA,$8F,$E7,$2D,$8C,$C6,$0F,$F2,$FE,$2F,$90,$A4,$9F,$F0,$5C,$65
	.dc.b	$FE,$99,$CE,$F9,$3F,$EB,$0D,$8C,$C6,$0C,$C4,$8E,$2D,$D3,$F8,$F7
	.dc.b	$37,$E7,$27,$F9,$D3,$63,$3C,$FF,$CE,$A3,$11,$FC,$BF,$C1,$47,$E0
	.dc.b	$9B,$CF,$B2,$5F,$C4,$9F,$86,$47,$07,$83,$D0,$EC,$47,$07,$93,$D2
	.dc.b	$F1,$11,$C1,$F4,$3E,$E7,$E7,$91,$C1,$FE,$A6,$FD,$69,$E9,$FF,$A5
	.dc.b	$DB,$19,$8C,$1F,$C4,$B2,$7C,$D5,$24,$FF,$07,$8C,$B7,$E9,$38,$11
	.dc.b	$C1,$F8,$4F,$95,$FC,$19,$1C,$1F,$E5,$C1,$F4,$4F,$4D,$1F,$A0,$E0
	.dc.b	$FD,$0C,$E6,$31,$1A,$8C,$3C,$0D,$CD,$9F,$EC,$C3,$EE,$9E,$DE,$6B
	.dc.b	$F5,$B1,$89,$EA,$6E,$75,$3E,$49,$D3,$FE,$D0,$FB,$AF,$D3,$78,$5F
	.dc.b	$E2,$16,$C7,$F9,$E1,$AF,$F7,$70,$E1,$FE,$F4,$FF,$85,$3C,$6E,$5E
	.dc.b	$2F,$F7,$19,$E4,$D7,$94,$ED,$E9,$8B,$F6,$47,$9B,$F0,$27,$F8,$4C
	.dc.b	$61,$F6,$B5,$F9,$07,$D0,$71,$3F,$21,$FF,$6B,$47,$0F,$47,$A3,$B6
	.dc.b	$CF,$85,$CF,$96,$F0,$7F,$D1,$95,$FC,$1B,$DD,$FF,$23,$4E,$5B,$3A
	.dc.b	$9F,$F0,$97,$3F,$F8,$25,$FF,$17,$35,$CE,$DF,$E4,$FE,$54,$E6,$7F
	.dc.b	$6C,$FF,$07,$78,$F7,$EE,$31,$1D,$A3,$FC,$B1,$5C,$BD,$9B,$46,$92
	.dc.b	$BF,$8F,$F5,$FA,$08,$E0,$FF,$22,$7B,$CF,$6F,$CC,$D8,$CC,$60,$CD
	.dc.b	$48,$E2,$D7,$FC,$BD,$9F,$FA,$91,$FE,$5F,$EF,$7C,$82,$3E,$0E,$DE
	.dc.b	$E4,$70,$79,$3C,$1F,$19,$1C,$1F,$E7,$9F,$B8,$FD,$2B,$7A,$BF,$CF
	.dc.b	$8A,$32,$FF,$B5,$1A,$FE,$75,$FF,$6D,$3C,$FE,$03,$8C,$62,$38,$47
	.dc.b	$FA,$1F,$D0,$78,$EF,$B7,$C3,$6B,$FA,$1C,$60,$FC,$3F,$8B,$23,$FA
	.dc.b	$8B,$5F,$F4,$27,$4F,$A3,$A3,$FD,$3D,$1F,$E3,$18,$C1,$EC,$C3,$FB
	.dc.b	$F7,$37,$8F,$CE,$FC,$2D,$7F,$80,$9A,$FF,$80,$63,$F2,$FB,$48,$FE
	.dc.b	$C2,$D7,$FC,$16,$73,$21,$FE,$10,$3D,$BF,$43,$1F,$EB,$ED,$7E,$FF
	.dc.b	$07,$E9,$08,$FF,$6E,$7F,$4F,$E4,$C8,$FE,$E2,$D7,$F8,$76,$2F,$D3
	.dc.b	$57,$F3,$DB,$3F,$EF,$86,$BC,$1A,$FF,$9E,$B9,$3F,$30,$FE,$4B,$9B
	.dc.b	$F6,$E4,$7F,$C0,$3C,$DF,$BA,$23,$83,$F9,$01,$FE,$83,$F8,$5F,$F0
	.dc.b	$03,$8E,$B4,$62,$3F,$93,$46,$3C,$DF,$DA,$CE,$67,$FD,$41,$3C,$39
	.dc.b	$32,$48,$E9,$3F,$89,$7F,$CC,$47,$F8,$6C,$FF,$30,$14,$CF,$A1,$FE
	.dc.b	$D7,$70,$7F,$37,$F9,$C7,$C1,$7E,$89,$B1,$FE,$DF,$6B,$EE,$6B,$FE
	.dc.b	$E9,$3F,$D9,$E7,$83,$C6,$EC,$47,$C7,$F0,$7E,$12,$38,$3C,$FF,$53
	.dc.b	$E5,$23,$83,$FC,$63,$D7,$FC,$A8,$8E,$0F,$F1,$AF,$D1,$7F,$2E,$23
	.dc.b	$83,$FD,$F1,$ED,$BE,$C7,$F7,$ED,$8C,$C6,$0C,$FD,$0C,$8E,$2D,$7E
	.dc.b	$FC,$FF,$04,$7F,$DA,$CD,$41,$C3,$7C,$2F,$E8,$7F,$2D,$23,$FC,$9A
	.dc.b	$D7,$F9,$1F,$C0,$7F,$48,$23,$FE,$52,$7F,$A2,$7A,$3F,$D9,$E3,$FC
	.dc.b	$70,$D7,$C3,$E5,$7F,$83,$FD,$0D,$9C,$3E,$32,$4F,$FB,$F6,$91,$88
	.dc.b	$FF,$73,$43,$3E,$E4,$8E,$0F,$E8,$1F,$F5,$E7,$C2,$BE,$13,$1F,$04
	.dc.b	$8C,$4F,$E9,$A3,$13,$A9,$CC,$F8,$35,$F6,$91,$F3,$3F,$A6,$F8,$79
	.dc.b	$1F,$A2,$6B,$FA,$CE,$9F,$DE,$BF,$4B,$AF,$F6,$3E,$0F,$EB,$89,$1F
	.dc.b	$E7,$CF,$80,$FE,$EA,$47,$07,$FA,$03,$DE,$FE,$F2,$47,$07,$C1,$FD
	.dc.b	$3B,$FB,$53,$7C,$1F,$E9,$71,$88,$FE,$0A,$D7,$FC,$4C,$FF,$10,$6F
	.dc.b	$85,$C0,$8F,$5F,$1B,$99,$1C,$1F,$E8,$AF,$9C,$FE,$96,$47,$07,$FA
	.dc.b	$3B,$E3,$7F,$A7,$91,$C1,$FE,$92,$F4,$3F,$A9,$91,$C1,$F5,$FE,$B7
	.dc.b	$FD,$06,$DF,$82,$FB,$B1,$88,$FF,$20,$72,$A3,$11,$FE,$CE,$FC,$49
	.dc.b	$EB,$FF,$63,$6C,$7F,$28,$F3,$51,$88,$FE,$7A,$D7,$FD,$AF,$E1,$FE
	.dc.b	$B6,$7F,$B5,$F3,$FF,$6A,$24,$FF,$AB,$3A,$BF,$B6,$91,$C1,$FE,$B0
	.dc.b	$F0,$BF,$B8,$91,$C1,$FE,$B4,$ED,$FE,$EA,$47,$07,$FA,$E3,$E7,$3F
	.dc.b	$BC,$91,$C1,$EB,$C9,$CF,$EF,$24,$65,$FF,$00,$F7,$C8,$FF,$61,$B5
	.dc.b	$FE,$BB,$F5,$5F,$A9,$7E,$33,$5C,$F7,$7F,$83,$63,$11,$FC,$40,$70
	.dc.b	$BE,$AB,$FC,$BB,$1F,$59,$AE,$E3,$5E,$EF,$EE,$27,$33,$FC,$6A,$79
	.dc.b	$78,$59,$24,$7E,$EE,$7F,$D3,$8F,$DB,$F1,$DF,$53,$FE,$21,$6C,$66
	.dc.b	$30,$67,$EE,$64,$71,$6B,$FD,$9B,$36,$FF,$E5,$A6,$C7,$F6,$D6,$BB
	.dc.b	$8F,$F2,$FF,$EC,$0F,$BC,$A4,$9F,$EE,$91,$97,$F2,$5F,$57,$FE,$6D
	.dc.b	$7F,$9C,$F3,$A3,$10,$8C,$78,$BF,$E3,$D9,$CB,$FE,$BD,$E6,$FF,$37
	.dc.b	$EF,$FD,$36,$FF,$BF,$B5,$FE,$FA,$D7,$F8,$B7,$A8,$F9,$6F,$F3,$BB
	.dc.b	$63,$FC,$14,$D7,$FC,$12,$D7,$F0,$18,$BE,$AF,$FD,$1D,$18,$6B,$EC
	.dc.b	$6B,$EB,$6C,$3D,$3F,$7B,$72,$FF,$2D,$46,$0F,$7F,$D5,$B8,$FF,$D1
	.dc.b	$AD,$C3,$FC,$40,$D7,$FC,$3E,$D7,$FC,$3B,$F5,$7F,$CD,$95,$A3,$0D
	.dc.b	$F2,$FF,$D3,$31,$8F,$AB,$FE,$1B,$5F,$CF,$5A,$F2,$6D,$1C,$C3,$EA
	.dc.b	$5E,$47,$E9,$A3,$F1,$BE,$9F,$EA,$65,$E4,$FE,$AF,$9D,$39,$8C,$3B
	.dc.b	$BE,$53,$AB,$FF,$00,$DF,$6E,$CF,$F0,$1C,$E6,$7C,$58,$72,$76,$D7
	.dc.b	$CD,$7D,$30,$ED,$7C,$9F,$14,$D3,$89,$4E,$31,$70,$F1,$47,$B8,$D9
	.dc.b	$4F,$19,$BA,$F0,$3E,$A6,$BA,$38,$47,$C6,$78,$4C,$7F,$CC,$71,$E0
	.dc.b	$7F,$8A,$1E,$43,$33,$F4,$5C,$C3,$E9,$5F,$53,$FE,$0D,$8C,$36,$33
	.dc.b	$3D,$9A,$FF,$84,$FE,$A7,$FD,$76,$9A,$47,$FD,$7E,$FF,$8D,$BE,$AB
	.dc.b	$EA,$47,$E4,$BE,$17,$19,$FF,$9F,$2A,$CF,$F9,$D2,$6B,$FC,$BE,$2B
	.dc.b	$C4,$57,$47,$C4,$62,$BF,$D0,$35,$C2,$33,$AF,$8B,$74,$FD,$B9,$FF
	.dc.b	$25,$E1,$67,$31,$F6,$4C,$9D,$4F,$31,$4F,$31,$4F,$36,$1E,$59,$F4
	.dc.b	$CF,$48,$C4,$EE,$31,$87,$F8,$AF,$EA,$9E,$E7,$8B,$7F,$1D,$27,$39
	.dc.b	$7F,$49,$38,$7C,$96,$DE,$57,$71,$FE,$03,$5C,$1F,$EB,$33,$FE,$39
	.dc.b	$FA,$A7,$DC,$EF,$1E,$D8,$59,$CD,$77,$BF,$E1,$7F,$54,$FB,$BE,$DD
	.dc.b	$8F,$AB,$FF,$2A,$5E,$F2,$5E,$E3,$95,$6F,$C6,$E3,$7C,$31,$F2,$9F
	.dc.b	$C6,$3C,$97,$99,$AA,$F3,$46,$30,$E7,$F5,$6F,$A0,$FB,$38,$7F,$38
	.dc.b	$8C,$4F,$F8,$8D,$B1,$FE,$DE,$6B,$FE,$C4,$FA,$77,$1F,$C3,$6D,$76
	.dc.b	$BC,$4F,$C7,$FA,$B7,$D2,$53,$1F,$35,$6F,$DF,$F5,$7D,$B2,$FE,$C2
	.dc.b	$D7,$FB,$03,$78,$0E,$78,$9F,$57,$FE,$CA,$BF,$98,$FD,$BE,$9F,$C2
	.dc.b	$B9,$C3,$E4,$D7,$0E,$7F,$B9,$F0,$9F,$EB,$5C,$3E,$A3,$6B,$6F,$FA
	.dc.b	$4E,$33,$E7,$7D,$CF,$AB,$F8,$35,$FC,$3B,$44,$62,$3F,$B2,$B5,$FE
	.dc.b	$FA,$D7,$FD,$F1,$EC,$FE,$4F,$E5,$9F,$51,$FB,$2C,$7C,$73,$98,$C5
	.dc.b	$7E,$0B,$07,$B2,$F6,$7F,$1E,$DC,$3F,$29,$8B,$E4,$9F,$2D,$AB,$FD
	.dc.b	$17,$13,$E6,$DF,$0B,$F8,$6D,$3F,$C6,$51,$8C,$BF,$D5,$4D,$8F,$F3
	.dc.b	$8B,$5F,$F1,$E2,$F8,$1F,$66,$5F,$C5,$70,$FD,$C4,$E6,$FC,$B3,$97
	.dc.b	$FC,$43,$A3,$E8,$BD,$67,$2C,$6C,$FE,$9B,$D0,$7D,$2F,$9E,$7F,$C4
	.dc.b	$D1,$89,$FF,$3F,$36,$3E,$B3,$5F,$F0,$CD,$BE,$CF,$E9,$79,$BF,$E3
	.dc.b	$9B,$FD,$9C,$E5,$F9,$71,$9A,$FA,$53,$F8,$2D,$3F,$92,$F1,$7D,$4E
	.dc.b	$FE,$7E,$D7,$D4,$7E,$AF,$57,$9E,$7E,$CB,$6F,$F8,$96,$1E,$48,$C7
	.dc.b	$C7,$F9,$3E,$5F,$EE,$B7,$4F,$2F,$83,$F8,$03,$D3,$F2,$16,$3E,$37
	.dc.b	$C7,$F7,$E1,$34,$91,$FE,$E5,$F1,$BF,$0A,$FF,$1B,$F9,$17,$D0,$3E
	.dc.b	$6B,$F0,$F5,$AD,$BF,$89,$D6,$A7,$F2,$3F,$2F,$30,$F8,$FE,$F5,$BF
	.dc.b	$8F,$F3,$3E,$F2,$4F,$FA,$39,$F8,$5E,$67,$D2,$F8,$AF,$CD,$B6,$7F
	.dc.b	$97,$53,$F5,$E7,$07,$F3,$B8,$C3,$F6,$BE,$2F,$13,$AA,$E3,$F8,$13
	.dc.b	$E6,$BF,$11,$82,$31,$3F,$E0,$88,$19,$9D,$CE,$A3,$54,$E6,$EE,$78
	.dc.b	$E7,$5D,$9F,$CC,$27,$58,$7E,$2B,$ED,$3F,$1F,$F6,$1F,$90,$6C,$7E
	.dc.b	$3A,$3F,$27,$ED,$63,$2F,$E8,$3E,$57,$F8,$36,$3F,$2F,$F0,$32,$FF
	.dc.b	$2B,$EE,$DB,$DF,$87,$FE,$50,$9F,$CC,$FC,$26,$75,$FE,$F7,$97,$F8
	.dc.b	$5E,$3F,$B2,$74,$FE,$B3,$E0,$2F,$8F,$FE,$52,$C3,$F4,$3D,$F8,$9C
	.dc.b	$FC,$47,$EC,$6F,$F6,$6D,$BF,$F3,$B5,$FC,$B9,$CB,$61,$F2,$7F,$B0
	.dc.b	$F3,$BB,$B6,$7F,$5F,$E7,$7E,$D3,$D4,$FD,$AB,$63,$F6,$71,$FB,$8F
	.dc.b	$93,$C7,$0F,$C0,$7F,$77,$F0,$B2,$BF,$73,$EA,$DA,$F8,$BF,$01,$FC
	.dc.b	$00,$C7,$1F,$C4,$E4,$FE,$D5,$3F,$DC,$0F,$F5,$97,$3F,$E3,$FC,$DF
	.dc.b	$EF,$11,$F5,$3E,$AF,$FB,$B5,$FF,$6E,$9C,$B7,$AF,$F1,$F0,$69,$33
	.dc.b	$FC,$85,$FF,$AF,$FC,$0E,$C7,$79,$B3,$FC,$F3,$0F,$E7,$D5,$E1,$F5
	.dc.b	$7F,$A5,$43,$E3,$EC,$63,$E3,$E5,$F1,$8B,$7F,$BA,$E6,$7C,$BF,$E0
	.dc.b	$BE,$EF,$FA,$7E,$DF,$E4,$0E,$9F,$E5,$F3,$9C,$61,$FC,$1E,$BF,$C1
	.dc.b	$35,$B7,$8C,$7E,$AA,$91,$A8,$C2,$F4,$7C,$07,$AB,$ED,$EC,$FF,$13
	.dc.b	$BF,$F5,$2B,$F8,$B6,$E5,$FE,$7E,$C1,$18,$8F,$DC,$88,$CF,$77,$E5
	.dc.b	$F1,$FF,$02,$57,$9F,$E8,$3F,$42,$F4,$31,$87,$CF,$F6,$7E,$6F,$B3
	.dc.b	$F3,$9B,$1F,$4E,$3E,$8F,$D6,$63,$3F,$C2,$3F,$53,$CB,$FA,$BE,$BD
	.dc.b	$BF,$B4,$E1,$FB,$9E,$CD,$BF,$C6,$3F,$E4,$AF,$A4,$FE,$65,$3F,$C5
	.dc.b	$7A,$91,$8A,$FF,$09,$35,$FF,$7A,$B5,$FE,$37,$E4,$3A,$7E,$0D,$E1
	.dc.b	$B2,$BE,$BE,$AF,$E3,$38,$3F,$1B,$CB,$E3,$F6,$2E,$3F,$E5,$8C,$FF
	.dc.b	$A6,$B6,$3F,$27,$F0,$BB,$8F,$CC,$3E,$BF,$27,$FB,$15,$7F,$06,$CF
	.dc.b	$D3,$18,$3F,$BE,$6B,$E7,$F6,$BF,$3A,$71,$C3,$F4,$4F,$C5,$F2,$F9
	.dc.b	$92,$3E,$36,$1F,$C6,$70,$FF,$61,$DF,$C1,$FA,$2F,$D3,$9C,$6B,$F5
	.dc.b	$8F,$CF,$F9,$A9,$1F,$AF,$F6,$AD,$FF,$8C,$70,$57,$EC,$3D,$9C,$F5
	.dc.b	$7F,$2C,$E4,$47,$F5,$5F,$A1,$FF,$5B,$61,$FE,$D4,$6B,$F9,$66,$BF
	.dc.b	$D4,$1A,$BE,$57,$F9,$D3,$FC,$4D,$B7,$FE,$7C,$9F,$F1,$54,$FF,$58
	.dc.b	$9C,$E7,$F7,$7D,$AF,$E0,$A7,$F0,$6F,$AF,$B1,$FC,$2C,$F9,$F9,$7F
	.dc.b	$93,$AF,$F5,$3B,$FE,$1E,$72,$7D,$43,$FC,$59,$97,$F4,$A6,$9F,$8B
	.dc.b	$3F,$CB,$86,$27,$F6,$5F,$0E,$F7,$7F,$A7,$B6,$7F,$C2,$0B,$DE,$CC
	.dc.b	$EB,$DE,$FE,$9D,$3F,$DE,$EF,$F8,$03,$5F,$DF,$F8,$A8,$C4,$7E,$83
	.dc.b	$0F,$E3,$A7,$F9,$90,$7D,$6F,$9B,$C8,$F3,$B2,$35,$F5,$D6,$BF,$B8
	.dc.b	$C2,$3D,$3F,$09,$23,$A7,$C1,$DA,$FF,$12,$E3,$FE,$47,$7C,$AD,$7F
	.dc.b	$92,$35,$F8,$BF,$3D,$F3,$7E,$AB,$E4,$FD,$0F,$CA,$6E,$1E,$9F,$A9
	.dc.b	$FF,$79,$BF,$58,$38,$BF,$80,$D5,$C7,$FE,$FE,$8C,$3F,$F0,$03,$EA
	.dc.b	$6D,$1A,$47,$FC,$46,$7F,$2C,$A7,$D6,$FE,$7B,$E4,$3F,$AC,$B6,$3F
	.dc.b	$C5,$4D,$78,$35,$FE,$6A,$BB,$FF,$DB,$71,$C1,$AF,$06,$BC,$1A,$FB
	.dc.b	$2B,$FC,$AC,$F9,$8F,$C0,$79,$1F,$D0,$A3,$E3,$36,$32,$6C,$E5,$F4
	.dc.b	$D9,$69,$9C,$94,$9D,$46,$1B,$1F,$E6,$E9,$FC,$77,$D3,$FF,$52,$EA
	.dc.b	$FD,$33,$67,$F2,$9E,$45,$AF,$EA,$DF,$26,$F9,$8F,$F0,$1F,$57,$FB
	.dc.b	$79,$B1,$99,$FF,$16,$35,$FE,$C5,$1F,$23,$1F,$EF,$0D,$C3,$FB,$93
	.dc.b	$67,$E4,$7C,$2F,$87,$4F,$6F,$91,$6D,$E5,$53,$11,$9F,$A9,$B8,$7E
	.dc.b	$AE,$30,$FF,$A2,$71,$7C,$1F,$A9,$B1,$FE,$43,$6B,$E8,$8D,$7C,$1F
	.dc.b	$C3,$DB,$E1,$36,$8D,$F8,$5F,$37,$97,$30,$FD,$E3,$E9,$F2,$A4,$7F
	.dc.b	$9E,$F6,$FB,$31,$FE,$2E,$6C,$67,$64,$E5,$74,$7C,$32,$4F,$F0,$9F
	.dc.b	$93,$FB,$0B,$E8,$BF,$22,$FF,$91,$DB,$C7,$F8,$88,$FF,$45,$C6,$2F
	.dc.b	$FC,$2B,$F5,$6F,$77,$9B,$6B,$FC,$86,$D7,$FD,$28,$D7,$F6,$4D,$7F
	.dc.b	$D2,$FE,$B7,$F8,$13,$1F,$B5,$10,$FE,$EC,$FE,$3E,$35,$18,$21,$FE
	.dc.b	$8E,$FB,$C7,$87,$F8,$55,$E3,$52,$4F,$F8,$07,$5E,$9E,$97,$C8,$8F
	.dc.b	$EE,$31,$FE,$5C,$C3,$FC,$2E,$D9,$D4,$FF,$74,$BF,$F5,$53,$FE,$B0
	.dc.b	$DB,$FD,$2B,$18,$8F,$27,$C9,$7A,$7E,$B3,$CB,$1B,$DF,$E0,$C6,$23
	.dc.b	$F8,$B6,$48,$C4,$7F,$91,$23,$1A,$79,$67,$33,$CB,$CD,$F8,$04,$7B
	.dc.b	$1F,$E2,$58,$FF,$68,$B8,$E3,$65,$FD,$66,$37,$87,$86,$FD,$07,$CB
	.dc.b	$F3,$FE,$67,$9F,$2B,$95,$FF,$65,$78,$E3,$5F,$F2,$EE,$7E,$EA,$F4
	.dc.b	$12,$3F,$C8,$93,$B7,$FB,$CD,$7F,$A3,$DE,$BF,$0F,$F6,$CF,$B9,$F0
	.dc.b	$FF,$77,$E2,$2C,$F8,$A3,$0F,$93,$E3,$FF,$05,$F2,$32,$30,$7A,$3E
	.dc.b	$47,$F0,$D3,$23,$07,$C8,$F0,$1F,$65,$FC,$32,$30,$D7,$DA,$7F,$9D
	.dc.b	$DC,$72,$FF,$A4,$61,$E1,$FF,$05,$69,$FE,$CD,$29,$FE,$83,$29,$FE
	.dc.b	$04,$29,$FE,$93,$9F,$E6,$CF,$82,$BC,$BE,$47,$F1,$19,$CD,$FE,$4B
	.dc.b	$E4,$B1,$EA,$67,$D3,$98,$FA,$A3,$87,$AD,$A8,$C1,$EE,$7F,$55,$E0
	.dc.b	$C8,$FB,$EF,$EB,$7E,$46,$47,$E0,$9A,$FE,$CF,$E4,$BF,$87,$FB,$FF
	.dc.b	$83,$CF,$F8,$64,$A3,$1D,$3F,$EA,$06,$C0,$D7,$57,$23,$FD,$9E,$E0
	.dc.b	$FF,$68,$C7,$E6,$5F,$EF,$C6,$4A,$FE,$81,$FF,$72,$36,$BC,$6D,$AF
	.dc.b	$F4,$7F,$43,$E4,$BF,$85,$79,$33,$0F,$CF,$BF,$C3,$7C,$C9,$1F,$A2
	.dc.b	$6B,$FC,$4F,$CB,$FE,$AD,$F3,$3F,$D7,$F3,$FE,$0B,$F2,$7F,$C5,$4C
	.dc.b	$F9,$DA,$F9,$2B,$87,$ED,$6F,$86,$BF,$EA,$27,$FD,$CD,$1F,$95,$FD
	.dc.b	$8D,$F1,$5D,$BE,$4A,$4E,$63,$FA,$A4,$BE,$2B,$FB,$AB,$67,$E3,$F6
	.dc.b	$A3,$11,$FE,$D5,$6B,$FE,$56,$5F,$17,$D0,$91,$E7,$F9,$3F,$7F,$C9
	.dc.b	$48,$F5,$3F,$0B,$E3,$E4,$7B,$5A,$FE,$AB,$CD,$6A,$F9,$1F,$93,$F2
	.dc.b	$1F,$2B,$E4,$65,$FC,$58,$F9,$B9,$EA,$30,$7C,$97,$FA,$FF,$E4,$FF
	.dc.b	$1A,$3F,$91,$87,$0F,$EE,$8C,$70,$7F,$66,$FB,$DF,$EC,$11,$F9,$37
	.dc.b	$FB,$3B,$FE,$4B,$6B,$FE,$F8,$6B,$BF,$83,$FC,$BD,$37,$F1,$92,$7E
	.dc.b	$F5,$FF,$45,$EC,$FC,$3F,$1C,$9F,$E5,$71,$FD,$27,$B3,$FA,$01,$23
	.dc.b	$F9,$4B,$FD,$0F,$B3,$FA,$71,$23,$F1,$5F,$27,$FE,$37,$EC,$7D,$0F
	.dc.b	$EC,$9F,$EF,$DD,$0F,$BD,$71,$FF,$86,$DB,$87,$F4,$AA,$E3,$D7,$FD
	.dc.b	$CF,$EB,$DE,$47,$F6,$76,$D7,$FB,$FD,$AF,$F5,$C6,$BF,$D6,$FE,$5B
	.dc.b	$F3,$AF,$FB,$E5,$BE,$C1,$CF,$5F,$E1,$1F,$F0,$10,$EF,$FD,$6F,$CB
	.dc.b	$4B,$FE,$B3,$39,$6F,$37,$F7,$B8,$FA,$51,$FC,$1B,$0E,$D3,$93,$5D
	.dc.b	$9D,$89,$3F,$B3,$F1,$BC,$3C,$1F,$BC,$F8,$A9,$3E,$08,$F8,$9F,$0B
	.dc.b	$F0,$C9,$1F,$0D,$F1,$77,$7A,$89,$1F,$BF,$F9,$6F,$AD,$CD,$E9,$8F
	.dc.b	$D2,$78,$2F,$94,$FE,$16,$6B,$87,$7E,$37,$F2,$0A,$73,$FD,$E8,$C3
	.dc.b	$63,$F9,$2B,$5F,$F0,$7F,$DE,$3D,$8F,$F1,$73,$63,$FC,$06,$D7,$FA
	.dc.b	$FF,$C1,$3E,$F3,$FA,$FF,$5F,$F3,$78,$FF,$41,$77,$FF,$9D,$6B,$F1
	.dc.b	$27,$EF,$4B,$8F,$F9,$36,$77,$1F,$E1,$0F,$8A,$FF,$00,$3F,$E2,$26
	.dc.b	$BF,$E2,$16,$BF,$E4,$86,$BF,$E6,$0F,$8C,$FC,$5C,$FF,$A4,$63,$FC
	.dc.b	$42,$D7,$D3,$E6,$FF,$A2,$E1,$F6,$3F,$E3,$7C,$3C,$8F,$F9,$22,$B8
	.dc.b	$B5,$F8,$2E,$F2,$FF,$2C,$FE,$97,$FE,$32,$97,$62,$30,$DF,$85,$FF
	.dc.b	$26,$46,$F1,$70,$FF,$A3,$32,$F8,$8D,$C7,$E6,$37,$1F,$F3,$EB,$FE
	.dc.b	$8C,$7D,$BF,$10,$F4,$3F,$D6,$98,$7F,$98,$1B,$19,$9F,$F1,$EF,$D8
	.dc.b	$BD,$BF,$F3,$63,$63,$F4,$8D,$7F,$CA,$6D,$7F,$CA,$4D,$7F,$CF,$3F
	.dc.b	$86,$FF,$96,$73,$FB,$9F,$31,$FE,$1C,$6F,$C0,$FF,$9E,$67,$5A,$FE
	.dc.b	$1F,$E3,$B4,$F2,$E0,$D1,$EA,$8C,$E3,$FE,$9C,$FB,$17,$C0,$7B,$BE
	.dc.b	$77,$FC,$0F,$C2,$6C,$EA,$BF,$02,$D7,$F3,$2D,$D3,$F2,$4D,$CB,$F3
	.dc.b	$4B,$5F,$F5,$3F,$D8,$BE,$0F,$F5,$B8,$3E,$C3,$FC,$73,$3F,$E9,$F6
	.dc.b	$FC,$27,$F8,$66,$75,$C3,$6B,$96,$7F,$E1,$DB,$E3,$87,$47,$8D,$E6
	.dc.b	$F8,$61,$CE,$77,$79,$AF,$F1,$36,$1D,$1F,$F6,$CD,$F0,$7F,$C5,$5D
	.dc.b	$BD,$1C,$9C,$31,$F5,$C7,$48,$F3,$9D,$31,$F0,$57,$38,$C8,$BF,$06
	.dc.b	$3C,$4E,$C4,$9F,$81,$87,$B5,$B8,$FF,$17,$AF,$45,$FC,$73,$A5,$7C
	.dc.b	$E7,$11,$A7,$E5,$7C,$43,$EC,$79,$36,$3F,$C8,$32,$8C,$47,$26,$BF
	.dc.b	$E4,$6F,$A3,$FF,$25,$A6,$91,$FF,$27,$BF,$3D,$FE,$3D,$BD,$7F,$CA
	.dc.b	$D1,$78,$7F,$DD,$1A,$7F,$9E,$23,$15,$BA,$EC,$7C,$A3,$FD,$EE,$E6
	.dc.b	$3E,$BE,$9F,$0E,$FF,$DE,$5B,$73,$BF,$B2,$78,$8A,$78,$8E,$91,$E2
	.dc.b	$9F,$74,$62,$FB,$34,$F1,$1E,$4C,$3F,$D2,$35,$C6,$0B,$EC,$5F,$00
	.dc.b	$BF,$E5,$B1,$86,$F9,$B9,$BF,$E8,$F3,$97,$FA,$4F,$D8,$FF,$4F,$AF
	.dc.b	$E6,$AD,$8D,$DF,$1E,$1F,$E8,$D3,$97,$FA,$47,$D8,$BE,$3B,$58,$7F
	.dc.b	$51,$8C,$3F,$D0,$17,$C8,$6B,$1F,$5C,$69,$FF,$5C,$F8,$7F,$EB,$6E
	.dc.b	$0D,$36,$F9,$61,$8E,$05,$6B,$EC,$3F,$BA,$57,$F8,$2D,$B8,$7F,$BE
	.dc.b	$9A,$FF,$BE,$5A,$FF,$67,$FA,$AF,$89,$A7,$F6,$36,$D7,$FA,$87,$25
	.dc.b	$F4,$AD,$6D,$EB,$E1,$7D,$83,$E0,$30,$FF,$9F,$F1,$FE,$C2,$E5,$BF
	.dc.b	$07,$EE,$F2,$1F,$2F,$FE,$B4,$DB,$FA,$DB,$5F,$EB,$71,$AA,$FF,$14
	.dc.b	$36,$3F,$2D,$E6,$7F,$A1,$27,$FB,$7B,$63,$FC,$0E,$FB,$1F,$EE,$DE
	.dc.b	$B7,$CE,$FA,$FF,$F2,$7D,$7F,$76,$FA,$57,$C9,$67,$AF,$DD,$18,$9F
	.dc.b	$F6,$27,$0F,$E3,$78,$D8,$3E,$33,$D1,$18,$6C,$66,$BF,$C8,$0D,$CF
	.dc.b	$F8,$85,$7F,$93,$F7,$FC,$0F,$C0,$7C,$99,$FF,$18,$FD,$46,$27,$C7
	.dc.b	$F2,$1F,$8B,$F8,$4F,$F4,$BE,$58,$3F,$D7,$07,$B7,$97,$C2,$47,$C2
	.dc.b	$F9,$B1,$FA,$98,$FE,$49,$BA,$FE,$A6,$FF,$C1,$7F,$55,$E4,$BF,$06
	.dc.b	$9F,$D5,$71,$F8,$67,$FB,$9E,$BF,$C2,$3F,$51,$AF,$2B,$6F,$B3,$E1
	.dc.b	$BC,$04,$AF,$F3,$16,$68,$F2,$7C,$27,$C2,$3B,$F9,$F6,$BB,$EC,$CB
	.dc.b	$FC,$3F,$EA,$79,$3E,$23,$CA,$DA,$FF,$39,$FC,$1F,$BC,$8F,$81,$F8
	.dc.b	$5F,$03,$F4,$0F,$6F,$D4,$59,$EF,$87,$F8,$B3,$E9,$FE,$0F,$D4,$7C
	.dc.b	$26,$CF,$FA,$00,$F9,$84,$3D,$A7,$E1,$3F,$01,$72,$FE,$F9,$B1,$DB
	.dc.b	$91,$F2,$79,$AF,$9F,$F4,$FF,$43,$EA,$2D,$ED,$E5,$CF,$D4,$91,$FE
	.dc.b	$8D,$C7,$EB,$71,$7E,$39,$FF,$0C,$C7,$CA,$C9,$18,$9F,$F7,$14,$3E
	.dc.b	$77,$CD,$C9,$F8,$0F,$A8,$FE,$35,$F2,$6F,$D8,$1D,$D9,$CF,$EA,$CE
	.dc.b	$63,$F9,$06,$4F,$AE,$FD,$A3,$FD,$BD,$AF,$F8,$5D,$A8,$E1,$E3,$FF
	.dc.b	$03,$AF,$07,$CB,$F8,$B1,$3D,$AF,$F8,$07,$83,$FE,$CC,$9C,$E5,$F3
	.dc.b	$7E,$2B,$F3,$3F,$23,$F9,$A6,$C7,$D1,$8F,$CE,$FA,$98,$D3,$E1,$BF
	.dc.b	$A0,$FA,$99,$5F,$5F,$EA,$AD,$7C,$8F,$A9,$FD,$29,$8C,$FF,$55,$C8
	.dc.b	$8F,$47,$9D,$E9,$FA,$2B,$8F,$C8,$D9,$19,$8F,$F0,$A6,$2E,$1F,$A1
	.dc.b	$3F,$0E,$B7,$5A,$9C,$B6,$1F,$09,$F1,$CE,$97,$FE,$B6,$9F,$9D,$5F
	.dc.b	$6E,$7F,$7B,$EA,$FF,$0D,$FA,$07,$C4,$E2,$FF,$47,$18,$6C,$7F,$04
	.dc.b	$C3,$83,$63,$7E,$B7,$33,$A7,$C6,$58,$E7,$F5,$1D,$CC,$5F,$94,$FD
	.dc.b	$F6,$08,$F2,$7D,$1F,$E3,$38,$7F,$C5,$D7,$E4,$8C,$7C,$36,$7E,$97
	.dc.b	$FA,$56,$3F,$EA,$46,$D9,$A6,$3E,$5B,$A1,$C7,$0F,$E1,$F5,$FE,$AC
	.dc.b	$8F,$89,$F0,$92,$7E,$1F,$D4,$5A,$F4,$79,$8D,$7F,$BF,$74,$7F,$2B
	.dc.b	$6D,$FF,$27,$7F,$C0,$AF,$CC,$F8,$4F,$CE,$70,$FE,$77,$8A,$CF,$E1
	.dc.b	$5C,$3F,$AE,$FC,$D3,$F4,$CE,$4C,$65,$F6,$7E,$A3,$F6,$3F,$37,$FB
	.dc.b	$B9,$FB,$31,$8F,$23,$3E,$67,$C6,$A3,$49,$5F,$EF,$D8,$79,$9D,$0E
	.dc.b	$31,$F1,$A3,$2F,$E1,$B8,$BF,$15,$CB,$F8,$B6,$C7,$E2,$A3,$F1,$DC
	.dc.b	$98,$CF,$E4,$F0,$7D,$5A,$FC,$7F,$C7,$FE,$4F,$BB,$F2,$8D,$8F,$C9
	.dc.b	$C7,$E5,$FB,$31,$7C,$9F,$F1,$5F,$E0,$2F,$A3,$FE,$EA,$DB,$FD,$6F
	.dc.b	$52,$31,$7F,$DD,$1A,$FF,$AA,$FA,$3E,$E7,$83,$FA,$53,$AE,$17,$0F
	.dc.b	$D2,$BF,$A9,$3F,$55,$E4,$D8,$FE,$61,$E7,$7C,$6C,$3F,$C8,$2D,$EB
	.dc.b	$FC,$35,$FF,$45,$E4,$F4,$F9,$9E,$82,$61,$E8,$D3,$F9,$8B,$FC,$1B
	.dc.b	$4F,$C8,$4F,$EE,$4E,$31,$FB,$87,$FC,$9B,$FB,$97,$0C,$FF,$C2,$2D
	.dc.b	$7F,$D9,$1B,$23,$11,$FE,$43,$6B,$FE,$74,$FD,$AD,$E0,$FF,$90,$DB
	.dc.b	$87,$F4,$DA,$ED,$AF,$F9,$CF,$C4,$56,$B0,$FB,$FC,$BF,$C3,$CF,$E2
	.dc.b	$1F,$19,$71,$F8,$FD,$9F,$D9,$CC,$57,$C8,$7F,$A6,$F9,$5F,$A2,$8F
	.dc.b	$E4,$ED,$7F,$D7,$8D,$7F,$DC,$8D,$7F,$A7,$7C,$AF,$EC,$9F,$E2,$07
	.dc.b	$FB,$EF,$C5,$EA,$FF,$5A,$8D,$69,$FE,$25,$F2,$1F,$35,$FE,$33,$5F
	.dc.b	$39,$63,$0E,$7E,$7F,$FC,$B9,$CB,$FC,$ED,$B4,$69,$8F,$A3,$E8,$71
	.dc.b	$DB,$FC,$27,$28,$FE,$E7,$F1,$9F,$E5,$AD,$3F,$94,$4A,$3F,$C9,$D9
	.dc.b	$FF,$A8,$E3,$1F,$19,$C3,$E2,$F8,$FD,$1F,$F6,$DB,$34,$99,$7E,$36
	.dc.b	$F2,$F8,$A3,$E0,$7C,$F7,$C1,$7E,$16,$12,$7C,$F8,$5C,$BF,$92,$E1
	.dc.b	$FD,$DA,$7F,$C9,$0F,$F9,$D7,$5F,$EB,$DE,$CF,$CB,$F9,$1F,$6E,$FE
	.dc.b	$52,$5F,$F5,$FD,$7E,$43,$E5,$F8,$CF,$A4,$7D,$3F,$A6,$B1,$F2,$3E
	.dc.b	$37,$FD,$88,$FE,$EF,$3F,$F3,$0B,$63,$3A,$79,$1A,$FF,$93,$E5,$F1
	.dc.b	$98,$D7,$FB,$A3,$63,$37,$F9,$19,$CB,$F9,$1F,$AF,$7C,$4F,$FB,$73
	.dc.b	$0F,$EF,$B1,$8E,$FF,$EF,$F0,$8C,$3A,$9E,$0F,$F8,$94,$FF,$13,$43
	.dc.b	$E9,$7F,$A8,$47,$46,$BB,$3A,$38,$79,$47,$1E,$FF,$BD,$F1,$DD,$71
	.dc.b	$F8,$CD,$AF,$AE,$D7,$E2,$37,$2F,$F0,$B7,$C7,$5D,$0D,$25,$7C,$58
	.dc.b	$F8,$6F,$E7,$33,$F6,$74,$FE,$B5,$B8,$FE,$1B,$4B,$5F,$E2,$3F,$8F
	.dc.b	$7C,$37,$C8,$C5,$18,$9F,$F1,$23,$5F,$85,$1A,$9F,$9F,$9F,$F8,$B7
	.dc.b	$E4,$3A,$BF,$E2,$FF,$8E,$B4,$6C,$EB,$4F,$F3,$DE,$1F,$E8,$66,$BF
	.dc.b	$E0,$26,$9F,$C1,$F3,$E1,$F1,$32,$78,$3C,$F3,$F6,$E3,$EA,$7D,$87
	.dc.b	$B2,$7D,$6E,$0F,$81,$F2,$8B,$E3,$1E,$07,$FC,$7A,$D9,$FF,$25,$4E
	.dc.b	$DF,$46,$9E,$B2,$3E,$BF,$91,$F7,$1F,$1B,$23,$49,$1F,$D4,$61,$F2
	.dc.b	$1F,$17,$E4,$64,$FC,$58,$FB,$2B,$E7,$DF,$23,$86,$C7,$F8,$C1,$AF
	.dc.b	$F3,$4F,$AF,$7B,$7C,$9B,$1F,$A8,$6B,$FE,$68,$6B,$FA,$16,$BF,$E8
	.dc.b	$DF,$4B,$FD,$43,$F2,$1F,$8C,$6C,$7F,$76,$8F,$BD,$84,$8F,$BC,$D7
	.dc.b	$EF,$7C,$9F,$E5,$3B,$5F,$56,$F0,$30,$DA,$FF,$6C,$CE,$A7,$C1,$EA
	.dc.b	$7E,$DB,$83,$EA,$67,$FD,$C7,$BF,$F8,$1B,$70,$FF,$0B,$D7,$EE,$5F
	.dc.b	$EE,$93,$F5,$7E,$53,$F2,$84,$8F,$F5,$F7,$B0,$F5,$7D,$CD,$8F,$F0
	.dc.b	$54,$76,$8C,$47,$F8,$86,$3C,$3E,$DB,$E4,$3E,$33,$67,$FC,$37,$1F
	.dc.b	$A3,$F8,$59,$1F,$A3,$6B,$FA,$3F,$73,$F5,$1E,$CC,$8F,$D4,$3E,$4C
	.dc.b	$E4,$7F,$87,$5F,$21,$1F,$C7,$7C,$97,$EC,$79,$64,$7E,$C6,$3F,$69
	.dc.b	$8F,$ED,$59,$1F,$B5,$8F,$DC,$7C,$8F,$EE,$7E,$4A,$4F,$EE,$63,$F7
	.dc.b	$9E,$E6,$A3,$07,$EF,$5F,$CC,$FD,$C7,$F7,$78,$FF,$42,$B5,$FF,$04
	.dc.b	$B7,$93,$F5,$D0,$8F,$F0,$3B,$C3,$DB,$FE,$7E,$E5,$B1,$A8,$FF,$60
	.dc.b	$B5,$FF,$29,$4F,$FB,$2D,$F0,$FB,$7E,$43,$F8,$A7,$C9,$FF,$0A,$FA
	.dc.b	$57,$C8,$DF,$80,$97,$FE,$F6,$9C,$B7,$E0,$B8,$54,$8C,$1F,$C7,$7B
	.dc.b	$7F,$90,$11,$FF,$7B,$65,$FC,$2F,$CF,$7D,$17,$FB,$22,$73,$E6,$7F
	.dc.b	$BE,$E7,$3F,$21,$FC,$BA,$B5,$5F,$EA,$86,$DF,$33,$FA,$C1,$F2,$BD
	.dc.b	$47,$C8,$FF,$0A,$6D,$7F,$AD,$9A,$FF,$8E,$97,$CA,$6C,$F9,$EE,$23
	.dc.b	$FB,$7B,$5F,$F7,$7B,$5F,$F4,$05,$E6,$FC,$58,$78,$4F,$55,$FA,$F0
	.dc.b	$97,$F0,$06,$3F,$12,$3E,$06,$B2,$75,$3F,$22,$31,$1F,$0F,$E5,$3E
	.dc.b	$37,$07,$C7,$DE,$4F,$0D,$13,$F2,$FE,$53,$73,$98,$C3,$E3,$F4,$DF
	.dc.b	$81,$BB,$7F,$A7,$B6,$74,$D7,$FC,$96,$BB,$76,$7F,$6F,$71,$1F,$E0
	.dc.b	$D1,$5D,$A7,$46,$FB,$F8,$CF,$68,$F0,$C7,$E0,$FE,$52,$4F,$B5,$FC
	.dc.b	$37,$C8,$C8,$F8,$0D,$7F,$11,$F2,$7F,$8B,$F4,$3F,$19,$F2,$72,$7E
	.dc.b	$34,$7E,$3F,$E4,$B5,$18,$3E,$53,$F0,$BC,$69,$1F,$09,$FE,$AF,$F8
	.dc.b	$8B,$D6,$A7,$EE,$1C,$4F,$F8,$E1,$AF,$C3,$6B,$F8,$AF,$94,$F9,$FE
	.dc.b	$74,$8F,$B5,$1F,$CD,$3B,$24,$7F,$38,$7F,$4D,$F2,$9F,$CE,$3B,$7F
	.dc.b	$9B,$12,$7F,$D8,$4F,$F8,$6B,$A3,$FD,$BD,$1F,$85,$F9,$4F,$D7,$1F
	.dc.b	$AF,$F9,$4F,$E8,$3D,$92,$7F,$A2,$BF,$B3,$F9,$4F,$E8,$BD,$DF,$D0
	.dc.b	$C9,$3F,$EC,$F7,$B7,$9D,$FB,$9F,$46,$4F,$E4,$E3,$FA,$6F,$74,$8F
	.dc.b	$EA,$2F,$F8,$B1,$FF,$52,$9F,$D5,$3B,$FF,$A8,$11,$FF,$6E,$3F,$E3
	.dc.b	$47,$F9,$57,$8B,$FC,$25,$FE,$49,$E2,$7F,$0F,$6C,$67,$E2,$FF,$AE
	.dc.b	$F5,$C9,$FE,$C2,$F8,$7E,$4B,$FB,$0F,$6F,$F5,$F2,$4F,$FB,$AD,$F0
	.dc.b	$7C,$8F,$9F,$E4,$A4,$FE,$D8,$FF,$52,$9F,$DB,$3B,$BF,$B5,$91,$D7
	.dc.b	$C8,$FF,$82,$FC,$3F,$E5,$B1,$FE,$71,$F6,$7F,$55,$E4,$7F,$7C,$9C
	.dc.b	$CF,$FB,$C1,$A8,$79,$9F,$EC,$CE,$4F,$95,$F2,$52,$7F,$C1,$7E,$87
	.dc.b	$F7,$8E,$3F,$F0,$4F,$D4,$BC,$7F,$F0,$E3,$6B,$F8,$A6,$88,$C4,$2B
	.dc.b	$FD,$ED,$F5,$3F,$DF,$63,$1C,$DF,$E2,$7A,$71,$FF,$85,$1B,$3F,$C0
	.dc.b	$70,$E9,$39,$35,$C3,$FE,$25,$F9,$2F,$AF,$5E,$C6,$D7,$8E,$7F,$C5
	.dc.b	$5E,$93,$B2,$92,$FF,$C5,$30,$EE,$FF,$1C,$7C,$FF,$F8,$B6,$7E,$73
	.dc.b	$5F,$F5,$36,$08,$C4,$7F,$90,$5A,$FC,$EF,$A0,$FC,$AB,$F0,$0F,$A3
	.dc.b	$EB,$FF,$8D,$76,$FA,$9C,$BF,$F0,$03,$C0,$FF,$20,$5F,$0E,$2F,$F3
	.dc.b	$64,$FF,$50,$EB,$FF,$22,$78,$BF,$88,$CB,$A4,$E5,$FA,$7F,$53,$FE
	.dc.b	$68,$6F,$C3,$FF,$90,$23,$11,$F8,$96,$BE,$C6,$BF,$E5,$BF,$A9,$FE
	.dc.b	$AD,$A5,$C3,$FC,$B9,$18,$9F,$C5,$FC,$43,$DD,$FF,$2C,$36,$3F,$CA
	.dc.b	$AD,$7C,$ED,$D3,$DA,$BD,$EF,$F3,$07,$CC,$FF,$52,$AF,$F2,$EB,$5F
	.dc.b	$7B,$6B,$E2,$FD,$4F,$F2,$66,$FA,$7F,$E6,$98,$C6,$9F,$39,$B7,$FA
	.dc.b	$27,$FD,$22,$FF,$6D,$1F,$53,$FE,$84,$BF,$D4,$7D,$47,$E6,$61,$EB
	.dc.b	$FF,$B9,$E3,$0D,$F8,$3F,$F7,$54,$F1,$B5,$71,$ED,$FD,$24,$62,$B7
	.dc.b	$7C,$63,$F6,$D7,$D9,$F0,$CF,$4A,$E3,$7C,$F0,$FD,$E7,$8B,$E2,$BB
	.dc.b	$7D,$1A,$74,$BF,$0C,$F1,$8F,$11,$E0,$C7,$C4,$FC,$26,$E3,$EB,$C7
	.dc.b	$81,$DE,$78,$9B,$3C,$E7,$7C,$6B,$F3,$30,$F8,$2F,$F1,$33,$47,$F7
	.dc.b	$FA,$F9,$0E,$4E,$13,$FE,$03,$7E,$AF,$87,$FE,$44,$78,$35,$E1,$3F
	.dc.b	$C5,$FD,$2E,$0D,$8F,$F3,$2F,$D4,$FF,$99,$93,$48,$E8,$FF,$4F,$69
	.dc.b	$E3,$E8,$FF,$6F,$6B,$FC,$6E,$7F,$96,$57,$CE,$7F,$77,$BD,$72,$72
	.dc.b	$3E,$05,$72,$BF,$F0,$F5,$6E,$31,$3F,$06,$31,$5C,$E3,$CE,$73,$3B
	.dc.b	$94,$EF,$7E,$5A,$E7,$1F,$3E,$B9,$BB,$39,$CE,$EB,$B4,$EA,$3F,$D6
	.dc.b	$94,$FA,$97,$97,$FE,$92,$C3,$F8,$CC,$63,$0F,$F1,$C5,$6E,$B5,$5F
	.dc.b	$D4,$2B,$4F,$CB,$FA,$97,$B9,$AB,$46,$30,$E6,$DA,$E1,$5B,$FA,$9F
	.dc.b	$F4,$9D,$78,$1B,$3F,$5D,$AF,$D0,$6B,$FE,$F7,$D1,$EA,$F9,$BC,$4F
	.dc.b	$F6,$A3,$63,$37,$FD,$4D,$B5,$FD,$13,$E6,$78,$78,$A6,$92,$BE,$1F
	.dc.b	$86,$F3,$BD,$CD,$8C,$C6,$0F,$85,$F2,$8F,$9D,$A4,$9F,$9B,$19,$7F
	.dc.b	$B3,$60,$7C,$64,$32,$7C,$DF,$33,$C1,$97,$E6,$DA,$FE,$51,$AF,$F5
	.dc.b	$46,$BF,$E0,$8F,$A8,$7B,$DF,$5D,$B5,$F8,$06,$CF,$E9,$DA,$FF,$68
	.dc.b	$FA,$8E,$7C,$06,$93,$4F,$A4,$FE,$1A,$78,$73,$60,$98,$7F,$81,$65
	.dc.b	$F4,$5C,$09,$19,$79,$ED,$23,$83,$6F,$FC,$55,$4F,$A1,$F0,$92,$BB
	.dc.b	$3E,$4F,$A1,$91,$E1,$6D,$72,$F9,$FF,$49,$27,$CC,$FB,$38,$E4,$7A
	.dc.b	$5B,$3E,$2F,$9E,$F8,$24,$9F,$73,$F9,$5F,$11,$F3,$B4,$93,$FE,$27
	.dc.b	$8C,$B7,$E7,$B8,$11,$C1,$BF,$9D,$E8,$47,$07,$3F,$9B,$F0,$11,$C1
	.dc.b	$DF,$E7,$3C,$84,$70,$78,$FE,$33,$D0,$47,$07,$9F,$E4,$3D,$84,$70
	.dc.b	$7A,$FE,$03,$E0,$11,$C1,$FE,$C5,$F8,$F7,$E0,$6F,$94,$FE,$1F,$18
	.dc.b	$6C,$66,$78,$37,$83,$FD,$9D,$ED,$D7,$FA,$83,$5F,$EA,$CD,$7F,$CD
	.dc.b	$4D,$BF,$0F,$27,$62,$5F,$67,$F9,$0F,$C9,$FF,$9C,$7E,$8E,$E0,$69
	.dc.b	$2B,$FD,$C1,$3F,$E7,$7E,$AD,$92,$33,$5F,$A0,$6F,$57,$42,$38,$37
	.dc.b	$D3,$E0,$23,$83,$E1,$FC,$EF,$C6,$23,$83,$FB,$87,$ED,$3D,$EF,$FD
	.dc.b	$E8,$D7,$FB,$B3,$5F,$EF,$8D,$7F,$DF,$3B,$7E,$B6,$7E,$8D,$FD,$42
	.dc.b	$3E,$CB,$F8,$A4,$70,$7C,$0D,$3E,$49,$1C,$1F,$13,$E8,$FE,$69,$1C
	.dc.b	$1F,$23,$83,$E8,$91,$C1,$FE,$0D,$E0,$7C,$6F,$F5,$8C,$D1,$89,$D4
	.dc.b	$E6,$31,$3F,$8A,$F8,$2C,$12,$BA,$3B,$F8,$99,$1D,$DB,$3F,$E1,$FE
	.dc.b	$1F,$F5,$B7,$7F,$E1,$89,$5F,$80,$7F,$13,$DF,$23,$F0,$6D,$AF,$F1
	.dc.b	$47,$2F,$E9,$30,$FD,$09,$2B,$F4,$2F,$E9,$BF,$04,$F6,$C7,$EA,$DB
	.dc.b	$5C,$9A,$FF,$8D,$E9,$18,$8F,$F1,$D7,$D3,$BE,$77,$E4,$36,$FF,$C7
	.dc.b	$0D,$9F,$F1,$E0,$70,$FF,$8F,$DF,$F4,$54,$70,$F8,$1E,$64,$79,$FC
	.dc.b	$07,$62,$38,$3F,$77,$ED,$F4,$BF,$DB,$B5,$FB,$6D,$E0,$FF,$6B,$63
	.dc.b	$FE,$16,$7F,$DB,$18,$59,$F1,$D7,$F0,$46,$BF,$C1,$1A,$FF,$BB,$77
	.dc.b	$FF,$77,$7D,$2F,$EF,$89,$8F,$EE,$1F,$E0,$3E,$34,$8F,$DD,$B6,$BF
	.dc.b	$DF,$9F,$44,$F9,$2F,$F0,$A3,$6B,$FD,$C8,$D9,$DC,$FF,$0C,$98,$7F
	.dc.b	$06,$2C,$7F,$96,$7F,$01,$7D,$67,$F8,$BD,$B3,$FE,$38,$6C,$EE,$7C
	.dc.b	$3F,$01,$2B,$C2,$D9,$F3,$FA,$3E,$62,$4F,$99,$ED,$E6,$48,$EC,$D9
	.dc.b	$F7,$FC,$C7,$B8,$93,$EE,$7C,$DE,$64,$8F,$33,$67,$D9,$F4,$FE,$B2
	.dc.b	$4F,$AD,$F9,$3F,$33,$23,$E4,$B6,$7F,$27,$F9,$93,$E7,$3F,$80,$36
	.dc.b	$BF,$D2,$0D,$9D,$CF,$F0,$5F,$D3,$DF,$3B,$49,$7F,$E9,$78,$CB,$F3
	.dc.b	$7D,$2F,$A0,$47,$07,$D1,$F8,$BF,$A8,$47,$07,$D5,$F9,$1F,$B0,$47
	.dc.b	$07,$F6,$3D,$DE,$97,$F6,$26,$C6,$63,$06,$7D,$39,$1C,$5B,$A7,$E1
	.dc.b	$FD,$4F,$C3,$12,$7F,$0C,$FE,$26,$64,$7E,$25,$B3,$FC,$4A,$D8,$78
	.dc.b	$C9,$5E,$77,$CD,$8C,$8F,$53,$6B,$F3,$1F,$3D,$F9,$72,$4F,$E5,$DF
	.dc.b	$CD,$65,$FE,$61,$98,$3E,$15,$D1,$F0,$A3,$31,$89,$E1,$7F,$EB,$4A
	.dc.b	$D3,$FE,$D5,$3F,$C9,$BE,$4E,$AC,$7F,$9B,$FD,$34,$66,$30,$D9,$E2
	.dc.b	$70,$3F,$DB,$9F,$55,$FE,$87,$C7,$FD,$B8,$DA,$FE,$4A,$D7,$7E,$57
	.dc.b	$AD,$CF,$97,$FD,$51,$BE,$2F,$28,$C4,$76,$6B,$E1,$FC,$13,$FA,$EB
	.dc.b	$E3,$31,$E6,$FB,$DA,$64,$FF,$46,$47,$07,$E4,$EA,$FA,$7E,$11,$F3
	.dc.b	$E3,$0F,$1C,$FE,$8C,$E9,$B1,$99,$7D,$1F,$FB,$F6,$7E,$BB,$63,$33
	.dc.b	$FC,$56,$3F,$A4,$AF,$39,$F7,$FE,$08,$FE,$C2,$D7,$FA,$C3,$5F,$B3
	.dc.b	$1C,$30,$E6,$F2,$E2,$B5,$FA,$AA,$FF,$27,$35,$EB,$1E,$5F,$6E,$47
	.dc.b	$F2,$96,$BF,$CB,$FD,$BF,$51,$1F,$E5,$CF,$6E,$59,$1D,$9A,$FA,$BA
	.dc.b	$7C,$64,$7C,$6F,$C5,$F8,$69,$1F,$CE,$5A,$FF,$3F,$F4,$3E,$51,$1F
	.dc.b	$E7,$CF,$CD,$F3,$E4,$7F,$41,$6B,$FD,$DB,$E9,$13,$F4,$21,$18,$7E
	.dc.b	$9B,$5E,$11,$BF,$AC,$79,$54,$93,$F6,$A3,$2F,$F5,$CF,$8E,$FB,$BF
	.dc.b	$08,$F3,$D1,$86,$CE,$CF,$EB,$1E,$A7,$E1,$C8,$E0,$FE,$B9,$F5,$1F
	.dc.b	$8B,$23,$83,$FC,$27,$6F,$D7,$7F,$77,$18,$6B,$FC,$00,$FE,$03,$1B
	.dc.b	$F0,$7C,$F3,$A9,$FE,$15,$D5,$F9,$92,$3F,$DB,$5D,$FD,$2C,$8E,$CD
	.dc.b	$7B,$0B,$F2,$7D,$2F,$62,$47,$95,$FD,$33,$0F,$EF,$45,$3F,$DE,$5E
	.dc.b	$67,$EB,$7E,$27,$FA,$F4,$7F,$BD,$E3,$07,$AB,$E1,$64,$7C,$16,$BF
	.dc.b	$EF,$BF,$9C,$E4,$47,$FE,$00,$3D,$7E,$A6,$47,$F8,$FB,$36,$09,$CC
	.dc.b	$FF,$AB,$87,$C3,$FF,$9C,$5A,$7F,$AD,$0A,$6D,$D6,$12,$34,$D7,$FC
	.dc.b	$9B,$F0,$2F,$53,$F3,$0D,$8F,$E1,$2D,$77,$1A,$F5,$E4,$69,$AF,$4F
	.dc.b	$53,$99,$1E,$6F,$F0,$AE,$2F,$E5,$F1,$FC,$39,$AF,$F1,$76,$F8,$7F
	.dc.b	$D2,$3C,$AE,$1F,$21,$2B,$FD,$B9,$48,$C4,$7E,$13,$7F,$97,$F4,$1F
	.dc.b	$6C,$91,$F3,$F8,$11,$F7,$FE,$8A,$4E,$5A,$FF,$9C,$BF,$05,$FE,$5A
	.dc.b	$DF,$2B,$F7,$D1,$88,$FC,$9B,$5F,$D1,$3F,$CA,$DF,$EC,$6D,$4D,$27
	.dc.b	$91,$D0,$91,$FC,$EA,$3F,$96,$3F,$AD,$C7,$F0,$FF,$4F,$F9,$72,$47
	.dc.b	$B5,$FC,$D7,$D4,$C8,$F8,$0D,$7E,$36,$6F,$A9,$FD,$01,$23,$E3,$3F
	.dc.b	$A3,$F9,$89,$1F,$25,$AF,$FA,$AB,$7E,$4F,$D4,$9F,$BC,$47,$FA,$9C
	.dc.b	$3E,$43,$F9,$33,$63,$31,$83,$F0,$D8,$48,$FB,$ED,$7D,$3F,$07,$FB
	.dc.b	$72,$3F,$75,$FD,$D7,$D3,$C8,$FC,$03,$5F,$83,$E3,$FE,$FC,$8F,$E1
	.dc.b	$5F,$E0,$5F,$3F,$23,$F1,$0D,$7E,$07,$17,$E6,$C8,$FE,$51,$FE,$85
	.dc.b	$F8,$9B,$E2,$29,$27,$FB,$AC,$65,$FC,$DF,$2F,$E9,$48,$E0,$FE,$63
	.dc.b	$F6,$CF,$86,$FE,$82,$D8,$CC,$60,$CF,$BF,$23,$8B,$5F,$E5,$3E,$32
	.dc.b	$BF,$CA,$31,$DB,$E0,$3A,$91,$F0,$FA,$7E,$02,$38,$3C,$BE,$C7,$90
	.dc.b	$8E,$0F,$4F,$67,$A0,$8E,$0F,$6F,$AD,$EC,$23,$83,$E0,$FA,$9F,$00
	.dc.b	$8E,$0F,$8B,$E7,$FC,$42,$38,$3E,$4F,$6F,$C8,$23,$83,$FA,$ED,$3C
	.dc.b	$AF,$ED,$AD,$8C,$C6,$0C,$F9,$52,$38,$B5,$F9,$B2,$E4,$E6,$48,$EE
	.dc.b	$F6,$E9,$91,$E2,$6B,$CB,$B3,$C6,$47,$CE,$FF,$80,$BE,$A9,$D5,$49
	.dc.b	$3F,$DE,$23,$2D,$E9,$E0,$47,$07,$F5,$3F,$99,$FF,$2D,$3F,$C3,$DA
	.dc.b	$FF,$9F,$35,$46,$23,$FD,$14,$D8,$7D,$53,$C2,$FF,$2C,$36,$BF,$B0
	.dc.b	$35,$FE,$70,$D7,$FC,$15,$CE,$FA,$97,$C3,$FF,$88,$1B,$3F,$E1,$46
	.dc.b	$BF,$9C,$6B,$FC,$9F,$E7,$BF,$C2,$0F,$A5,$AF,$FA,$85,$B8,$9B,$C6
	.dc.b	$F9,$BF,$F3,$83,$6B,$FB,$C3,$5F,$EA,$0D,$7D,$36,$E9,$F4,$12,$3D
	.dc.b	$0F,$F8,$EB,$EA,$7F,$C0,$11,$FE,$2E,$6B,$FB,$66,$BF,$11,$AF,$D3
	.dc.b	$FA,$97,$47,$FA,$15,$B1,$FE,$0D,$6B,$FD,$95,$AF,$F4,$BF,$A9,$FE
	.dc.b	$72,$FF,$69,$6B,$9E,$5F,$F4,$84,$62,$3F,$05,$F5,$3F,$C2,$1F,$EE
	.dc.b	$0D,$7F,$AB,$35,$DF,$77,$F9,$97,$E9,$DC,$CF,$96,$FF,$63,$FB,$CF
	.dc.b	$82,$FF,$25,$57,$F4,$5C,$3F,$CB,$B1,$8F,$2F,$FC,$9D,$5B,$9F,$F6
	.dc.b	$7D,$74,$BE,$77,$CB,$0E,$D5,$C8,$DE,$1C,$A7,$A4,$70,$39,$CF,$0C
	.dc.b	$3B,$43,$0E,$CF,$33,$17,$CD,$EA,$73,$38,$57,$3B,$F2,$61,$9B,$F7
	.dc.b	$61,$D9,$B3,$AB,$ED,$1E,$09,$D6,$1D,$E7,$C2,$73,$C3,$B3,$EA,$C3
	.dc.b	$FD,$87,$7E,$88,$FF,$34,$B5,$EC,$63,$CA,$7D,$47,$06,$C7,$F9,$C0
	.dc.b	$F8,$4E,$1E,$11,$FE,$A1,$5F,$51,$48,$FF,$A9,$5B,$F5,$3F,$32,$1D
	.dc.b	$2F,$8F,$42,$74,$F4,$31,$B7,$F8,$FE,$77,$3A,$BB,$39,$AD,$9F,$E8
	.dc.b	$88,$CB,$FE,$2B,$AA,$FF,$A6,$63,$EC,$3E,$48,$AF,$87,$6F,$46,$1E
	.dc.b	$78,$FC,$11,$E8,$AE,$6F,$13,$C2,$75,$29,$9C,$3D,$0E,$27,$A4,$60
	.dc.b	$EF,$7E,$88,$7D,$4B,$EA,$BC,$5B,$78,$E7,$31,$8A,$FE,$71,$4F,$81
	.dc.b	$C6,$3F,$42,$39,$4F,$D0,$C3,$94,$62,$7F,$6D,$3F,$D0,$7E,$AB,$FC
	.dc.b	$90,$EC,$EB,$87,$4C,$53,$AA,$E2,$7D,$EF,$AA,$FE,$C9,$7C,$63,$FD
	.dc.b	$75,$B7,$FA,$4E,$31,$6A,$DD,$72,$F0,$7F,$83,$7B,$FE,$A6,$DF,$0F
	.dc.b	$B6,$FF,$71,$BD,$63,$35,$E2,$C1,$5A,$8C,$78,$0F,$A5,$FB,$91,$FE
	.dc.b	$6F,$8C,$61,$FD,$02,$73,$3F,$D8,$E7,$2F,$FA,$47,$C3,$FE,$91,$5E
	.dc.b	$A6,$C7,$6A,$F8,$D1,$8E,$0F,$B1,$F5,$6F,$97,$F9,$4D,$C3,$E1,$AF
	.dc.b	$85,$CC,$FC,$FF,$21,$ED,$79,$33,$FD,$5C,$63,$0F,$ED,$F3,$9D,$BF
	.dc.b	$BA,$B7,$CB,$F1,$4F,$F7,$D7,$F0,$19,$FE,$2D,$CE,$1F,$DF,$B9,$7E
	.dc.b	$87,$D4,$EA,$FF,$DF,$ED,$9F,$DD,$FB,$DF,$E7,$F9,$FA,$53,$9C,$BF
	.dc.b	$DB,$FF,$3D,$E3,$A3,$49,$5F,$C3,$F9,$16,$FA,$AF,$C7,$5B,$E1,$BC
	.dc.b	$71,$87,$3C,$F3,$0F,$C5,$D3,$E1,$F1,$A7,$E0,$E3,$FD,$95,$3F,$EC
	.dc.b	$5F,$85,$F2,$11,$F1,$64,$FA,$8F,$0F,$85,$F0,$9B,$3F,$98,$FB,$F7
	.dc.b	$D3,$FE,$8D,$B1,$E2,$F8,$9C,$4E,$7E,$1B,$EA,$51,$A4,$AF,$0F,$C3
	.dc.b	$73,$F9,$AF,$A2,$D8,$F0,$78,$DD,$CF,$07,$C7,$D8,$F2,$74,$76,$23
	.dc.b	$FC,$EF,$A1,$97,$EF,$F8,$BF,$C6,$3E,$73,$E8,$BD,$5E,$CF,$EB,$E3
	.dc.b	$18,$7B,$E3,$DD,$E0,$C8,$F8,$5D,$57,$0F,$D8,$63,$F9,$B7,$FD,$EC
	.dc.b	$D7,$FD,$B8,$D7,$FC,$9C,$D5,$D7,$FE,$C9,$F9,$7F,$99,$F2,$1E,$36
	.dc.b	$E1,$F4,$79,$BE,$81,$F4,$7C,$6B,$1F,$57,$EA,$7E,$79,$1F,$85,$92
	.dc.b	$3F,$55,$8F,$FB,$5F,$CB,$FB,$5F,$53,$F0,$5B,$7F,$C6,$7E,$45,$FA
	.dc.b	$73,$E1,$BF,$96,$C6,$1F,$C1,$7C,$9E,$27,$55,$FC,$87,$81,$F6,$EE
	.dc.b	$1F,$DD,$46,$1B,$19,$9F,$EE,$53,$97,$FB,$A6,$3E,$2F,$9C,$EC,$7F
	.dc.b	$3E,$9D,$1F,$CC,$A3,$53,$F9,$7F,$A1,$FC,$B1,$F9,$7F,$A1,$B1,$F9
	.dc.b	$BE,$EF,$CD,$18,$BF,$B6,$FE,$77,$EA,$3F,$A0,$7D,$40,$C5,$7F,$3A
	.dc.b	$70,$D7,$F4,$8F,$06,$BF,$2F,$57,$C8,$FC,$73,$AF,$D2,$D9,$D7,$CD
	.dc.b	$7F,$1F,$D7,$E8,$B6,$CD,$31,$D5,$E6,$F9,$6F,$A4,$DB,$FE,$8D,$F3
	.dc.b	$CF,$9C,$FE,$98,$D8,$F5,$78,$58,$9F,$7F,$C7,$49,$E3,$17,$87,$FA
	.dc.b	$76,$EF,$98,$F8,$BF,$4D,$F1,$9B,$7F,$16,$3E,$4F,$C4,$62,$F8,$BF
	.dc.b	$67,$CE,$77,$53,$E7,$FC,$7D,$C3,$F1,$53,$F9,$97,$F0,$AD,$BF,$F1
	.dc.b	$C3,$5F,$F1,$69,$FE,$2E,$70,$E1,$FE,$3D,$1D,$3C,$4F,$17,$85,$E3
	.dc.b	$7E,$FF,$0F,$E0,$39,$ED,$7D,$F7,$0F,$E0,$79,$2E,$3F,$11,$FF,$08
	.dc.b	$FB,$2F,$BD,$7C,$BE,$1B,$7F,$AD,$F8,$17,$66,$6B,$F5,$B0,$ED,$A6
	.dc.b	$6B,$83,$FA,$1F,$1B,$9E,$7F,$E1,$23,$F8,$5D,$FE,$63,$E6,$74,$EF
	.dc.b	$0F,$D8,$D1,$A4,$C3,$F6,$31,$F8,$5B,$E8,$DC,$FF,$C1,$2F,$19,$FD
	.dc.b	$F3,$97,$FC,$BB,$DD,$FE,$A1,$F5,$FF,$83,$D7,$F5,$2B,$FC,$4F,$87
	.dc.b	$89,$FD,$6F,$8F,$30,$FE,$21,$7B,$3F,$C8,$1D,$7F,$C3,$30,$FF,$13
	.dc.b	$77,$FF,$8C,$67,$FC,$AD,$9A,$31,$7F,$EF,$BB,$FF,$58,$3F,$A2,$F8
	.dc.b	$FF,$DD,$FD,$14,$8C,$1F,$BB,$7F,$7F,$F3,$5F,$E0,$BA,$34,$93,$FD
	.dc.b	$36,$3F,$DA,$AF,$F0,$3F,$93,$E2,$D3,$F8,$37,$D0,$FE,$C3,$E4,$7F
	.dc.b	$85,$B6,$3F,$6B,$1F,$C3,$BD,$5C,$6D,$FA,$E7,$F8,$8B,$E3,$24,$F8
	.dc.b	$5F,$F1,$07,$DB,$7D,$17,$FD,$1E,$D9,$FE,$31,$1F,$E5,$98,$C1,$F6
	.dc.b	$5F,$D3,$9C,$39,$39,$92,$BF,$B5,$4F,$E3,$B8,$3F,$CE,$4D,$8F,$F4
	.dc.b	$DD,$A3,$A4,$EB,$97,$C1,$F3,$DF,$C4,$39,$2E,$1F,$C5,$1C,$3F,$E6
	.dc.b	$4B,$FF,$31,$DF,$DC,$E2,$FC,$14,$F6,$9C,$B7,$E6,$7E,$09,$1F,$3B
	.dc.b	$F1,$3E,$73,$E0,$1F,$05,$B1,$DF,$E8,$3E,$21,$F1,$79,$6C,$78,$FE
	.dc.b	$27,$E6,$91,$F8,$6F,$D0,$F9,$DF,$98,$7C,$D6,$C7,$F6,$9F,$A5,$FC
	.dc.b	$2F,$17,$F5,$56,$EB,$FE,$AD,$EC,$47,$E3,$AB,$87,$37,$F6,$E8,$74
	.dc.b	$FF,$43,$6C,$FF,$64,$6E,$1F,$6A,$F9,$D7,$E1,$3E,$5F,$F0,$1E,$F7
	.dc.b	$E0,$5B,$87,$CF,$8F,$C2,$74,$63,$0F,$F4,$57,$12,$3C,$5D,$B2,$FE
	.dc.b	$B7,$C5,$DB,$F6,$77,$F9,$49,$3E,$57,$F2,$BF,$13,$F9,$1F,$8B,$FC
	.dc.b	$93,$67,$F0,$1F,$17,$FE,$CF,$7F,$9C,$F6,$FE,$2E,$31,$87,$E0,$5A
	.dc.b	$FE,$79,$B1,$F9,$1E,$BB,$AF,$F0,$FB,$E6,$DB,$FC,$54,$F2,$5F,$50
	.dc.b	$F6,$BF,$DE,$F1,$FA,$43,$2E,$1F,$D3,$C6,$F0,$FF,$0B,$B6,$38,$2F
	.dc.b	$69,$E7,$E9,$B3,$A6,$BD,$1A,$FE,$7F,$E2,$BF,$AD,$E7,$EA,$6C,$64
	.dc.b	$1F,$EB,$CC,$3F,$65,$1A,$8F,$E0,$77,$67,$F6,$CD,$CF,$C3,$7F,$1A
	.dc.b	$35,$F6,$DF,$E7,$7B,$FF,$6B,$BF,$E8,$06,$FC,$5F,$F6,$EE,$BF,$F7
	.dc.b	$A4,$EE,$7F,$C6,$5F,$77,$FE,$CF,$7F,$C7,$2D,$7F,$2C,$D7,$FB,$BF
	.dc.b	$52,$31,$1F,$D2,$9F,$F2,$43,$FE,$53,$6B,$DB,$E0,$3C,$59,$7F,$12
	.dc.b	$B7,$C5,$E2,$9A,$A5,$F1,$79,$8C,$7C,$5D,$93,$67,$1F,$17,$6F,$E1
	.dc.b	$C6,$1F,$27,$C8,$FF,$96,$35,$F5,$37,$5F,$F7,$04,$FC,$28,$C7,$C5
	.dc.b	$F9,$6B,$E0,$36,$BF,$DC,$55,$EF,$F8,$BF,$27,$C6,$48,$FA,$6F,$9E
	.dc.b	$24,$7F,$A7,$F4,$FF,$65,$36,$3F,$CD,$7A,$FC,$16,$CF,$92,$5C,$3F
	.dc.b	$6B,$3F,$84,$7C,$3C,$BE,$27,$0C,$AF,$C1,$BF,$1F,$E3,$24,$78,$5E
	.dc.b	$A4,$7F,$DB,$DE,$6F,$E7,$B0,$FC,$3F,$B9,$F3,$CF,$AD,$EA,$FF,$79
	.dc.b	$E7,$FA,$5E,$EE,$2F,$87,$3F,$E7,$AB,$58,$FE,$52,$BE,$C7,$BF,$23
	.dc.b	$F2,$EF,$DB,$F8,$A9,$1F,$15,$F8,$24,$7E,$47,$6F,$E0,$48,$FE,$83
	.dc.b	$BF,$F0,$87,$C8,$23,$F3,$DF,$F1,$3F,$D3,$39,$BF,$15,$ED,$CA,$FD
	.dc.b	$AF,$07,$F8,$5A,$B5,$1F,$AC,$F7,$3E,$0C,$62,$3E,$9E,$DE,$F8,$C5
	.dc.b	$7E,$BA,$BF,$2D,$32,$BF,$6A,$FE,$6F,$DD,$FE,$FF,$1C,$8F,$D0,$B8
	.dc.b	$7F,$96,$B5,$FE,$F2,$D7,$7E,$DF,$E2,$21,$F1,$5D,$49,$3E,$CA,$F3
	.dc.b	$72,$FA,$89,$19,$3F,$DB,$B9,$7F,$01,$C9,$5E,$0F,$89,$F4,$7C,$4C
	.dc.b	$BE,$1D,$BF,$E6,$98,$FD,$34,$3E,$12,$91,$FF,$6B,$19,$38,$E8,$F8
	.dc.b	$9F,$E0,$50,$E4,$7A,$78,$6C,$7F,$B5,$DA,$FF,$9A,$D7,$B1,$B3,$FD
	.dc.b	$AB,$1C,$DB,$1A,$8F,$F6,$CB,$5F,$E9,$4B,$D8,$F0,$12,$3C,$FF,$11
	.dc.b	$FC,$0F,$5F,$54,$7F,$B6,$DB,$BF,$F8,$1D,$BB,$FE,$33,$CA,$FE,$9D
	.dc.b	$5F,$3E,$91,$88,$FC,$14,$63,$A7,$F9,$23,$63,$FD,$FD,$6F,$88,$FC
	.dc.b	$51,$F8,$BF,$88,$93,$F3,$23,$F1,$DC,$3A,$8C,$1F,$41,$F9,$FF,$4D
	.dc.b	$F9,$2C,$3E,$FC,$FF,$56,$8C,$67,$FE,$B6,$6C,$7F,$5A,$FA,$6D,$BF
	.dc.b	$E9,$B6,$BF,$E5,$66,$BF,$EB,$E6,$BC,$67,$EA,$C7,$FA,$B2,$3F,$24
	.dc.b	$1C,$3D,$7A,$BF,$51,$F1,$5F,$A9,$F8,$AF,$D5,$7C,$54,$AF,$C5,$C7
	.dc.b	$EB,$7E,$2B,$51,$83,$F1,$EF,$EC,$7D,$A9,$1F,$92,$7F,$7B,$E0,$48
	.dc.b	$FD,$27,$8A,$EE,$FF,$5F,$46,$1A,$FE,$E7,$BB,$FD,$95,$EE,$7F,$47
	.dc.b	$FC,$1F,$EE,$71,$7C,$C9,$FF,$67,$7C,$7F,$F0,$13,$5D,$1B,$F1,$BF
	.dc.b	$AA,$5F,$F8,$77,$1F,$F0,$E7,$A4,$8C,$47,$F9,$25,$AF,$F7,$47,$F5
	.dc.b	$0F,$EE,$E3,$87,$77,$8C,$FE,$21,$F0,$3F,$C4,$7E,$0F,$F8,$97,$C6
	.dc.b	$4B,$FD,$9C,$7F,$15,$F8,$CD,$46,$0F,$DB,$BF,$C6,$75,$91,$FB,$A7
	.dc.b	$FB,$B7,$80,$F8,$DA,$49,$FE,$0F,$19,$7F,$80,$9F,$1F,$CD,$8F,$A7
	.dc.b	$54,$7F,$04,$FA,$9F,$E0,$D3,$F8,$66,$BF,$E4,$16,$BF,$EA,$66,$BF
	.dc.b	$E3,$00,$E1,$6D,$F3,$7D,$2F,$B0,$48,$F3,$FC,$77,$F6,$0F,$A8,$79
	.dc.b	$BE,$EF,$8F,$95,$F8,$69,$CB,$EE,$F8,$FF,$EF,$7A,$9A,$48,$FF,$2C
	.dc.b	$6F,$FC,$0B,$E3,$BE,$49,$23,$E3,$FD,$47,$87,$C3,$F9,$BF,$21,$27
	.dc.b	$C1,$1F,$43,$E3,$FB,$92,$3A,$E4,$8F,$1F,$9F,$F5,$89,$3F,$53,$E4
	.dc.b	$3D,$DD,$DF,$6B,$D1,$93,$ED,$8F,$BB,$E9,$7B,$09,$1D,$9F,$7F,$C7
	.dc.b	$FE,$08,$91,$F8,$0F,$8F,$F9,$4F,$E1,$BE,$3E,$3F,$24,$FC,$4F,$C7
	.dc.b	$FC,$82,$3E,$A7,$E5,$FC,$77,$E3,$89,$1F,$8C,$F8,$FF,$F0,$AF,$90
	.dc.b	$E4,$FC,$9F,$C7,$CA,$FD,$DC,$E5,$FC,$9F,$C7,$7D,$62,$3F,$21,$FF
	.dc.b	$19,$FC,$BF,$F9,$AE,$3F,$7C,$DF,$2F,$F7,$8D,$F5,$7F,$DD,$BC,$DF
	.dc.b	$E2,$2A,$79,$CF,$A8,$FF,$5B,$47,$F6,$8B,$FF,$39,$61,$CB,$E5,$BF
	.dc.b	$C5,$B5,$AA,$EF,$F2,$1E,$22,$3F,$88,$BF,$F3,$4B,$FE,$A1,$C3,$B7
	.dc.b	$17,$22,$47,$F7,$6F,$B4,$FF,$6A,$BF,$DE,$3E,$6B,$F0,$8D,$EE,$FF
	.dc.b	$37,$41,$BD,$BF,$E7,$1F,$9B,$7C,$97,$F9,$59,$B7,$F9,$A6,$BF,$AF
	.dc.b	$3F,$CF,$6E,$1F,$F6,$3F,$CE,$7F,$9E,$1F,$F0,$4B,$5F,$E5,$4D,$E0
	.dc.b	$FE,$50,$FB,$A3,$F9,$77,$91,$FE,$68,$6F,$C8,$FF,$A1,$63,$15,$FC
	.dc.b	$BF,$E7,$5D,$9F,$E8,$96,$DF,$EC,$1A,$F9,$FC,$64,$62,$3F,$DB,$AD
	.dc.b	$7F,$CB,$9F,$50,$F1,$FF,$D2,$9F,$0B,$F3,$27,$FA,$13,$71,$FE,$D7
	.dc.b	$C8,$BE,$15,$FE,$28,$FA,$87,$A9,$FE,$6C,$9C,$CF,$F9,$0E,$B8,$63
	.dc.b	$FE,$A1,$BD,$CF,$F8,$C3,$E9,$FF,$C1,$AF,$FA,$A3,$EA,$1E,$57,$FA
	.dc.b	$9D,$B1,$BF,$86,$FF,$55,$57,$0D,$CD,$27,$C3,$FF,$8C,$BE,$A1,$E4
	.dc.b	$FF,$A4,$30,$7D,$33,$CD,$FE,$59,$4F,$33,$F9,$74,$E6,$BF,$DB,$F3
	.dc.b	$97,$FD,$C3,$19,$51,$DB,$8F,$37,$D3,$0F,$05,$EE,$B9,$5F,$8A,$BC
	.dc.b	$38,$78,$1E,$18,$72,$9F,$0E,$1E,$9F,$1F,$95,$FF,$43,$6C,$FF,$BD
	.dc.b	$1E,$D3,$BB,$E7,$3D,$CF,$5D,$77,$69,$DC,$F5,$CF,$A2,$78,$E3,$DA
	.dc.b	$79,$1F,$02,$B9,$3C,$30,$E5,$8F,$C3,$8F,$D5,$19,$2B,$97,$F5,$CD
	.dc.b	$7E,$9F,$9A,$FA,$8E,$31,$F3,$A3,$0F,$26,$BC,$9A,$F8,$3E,$A3,$FC
	.dc.b	$D2,$9A,$47,$FC,$DA,$FC,$17,$FD,$F3,$E5,$57,$81,$F0,$0C,$94,$CE
	.dc.b	$5F,$83,$72,$D7,$39,$FE,$D6,$7A,$57,$D7,$8C,$4B,$5E,$13,$98,$FF
	.dc.b	$00,$70,$7F,$97,$27,$31,$E3,$8F,$F2,$FE,$18,$3C,$4E,$2F,$E1,$4E
	.dc.b	$F0,$F0,$B4,$FC,$24,$BE,$A1,$F3,$DF,$E6,$AD,$BF,$C8,$11,$8B,$FF
	.dc.b	$53,$AF,$27,$50,$C1,$87,$3C,$33,$18,$FA,$8F,$E6,$95,$FE,$B0,$6C
	.dc.b	$7F,$94,$70,$EC,$FF,$94,$E3,$38,$FF,$94,$7E,$A3,$FC,$25,$7F,$EB
	.dc.b	$86,$D7,$F0,$FC,$FF,$93,$CE,$7D,$7F,$0E,$7F,$C9,$3D,$1B,$E0,$9C
	.dc.b	$E3,$CB,$2F,$E6,$11,$C3,$C3,$7C,$DF,$FA,$0B,$04,$63,$2F,$F1,$0B
	.dc.b	$74,$FE,$64,$D7,$FA,$3F,$91,$E2,$F1,$FF,$A8,$D7,$FB,$42,$B8,$65
	.dc.b	$FD,$1A,$BE,$3F,$BA,$FC,$DD,$ED,$7C,$AF,$9C,$7E,$55,$C3,$2F,$A1
	.dc.b	$7B,$AD,$4E,$74,$FA,$7E,$87,$F2,$CB,$FF,$1B,$B6,$3F,$A4,$35,$FC
	.dc.b	$7B,$5F,$B9,$B3,$D3,$7D,$17,$CC,$6D,$FF,$80,$FA,$71,$39,$F9,$CD
	.dc.b	$6F,$6B,$FC,$10,$F1,$F4,$3E,$66,$06,$92,$7F,$C0,$DF,$41,$DC,$F0
	.dc.b	$7D,$05,$8F,$F0,$97,$CE,$78,$F9,$FC,$8D,$8F,$F0,$B7,$C2,$F6,$23
	.dc.b	$FE,$11,$7C,$3F,$0F,$EB,$3D,$9F,$3F,$67,$FC,$3D,$F2,$1E,$FF,$8F
	.dc.b	$F8,$0D,$8F,$F1,$17,$C8,$7A,$88,$FF,$87,$5F,$6F,$CB,$FC,$73,$E4
	.dc.b	$7C,$9D,$9F,$D4,$76,$FF,$AB,$7E,$ED,$F4,$1F,$36,$7F,$33,$18,$7F
	.dc.b	$C6,$06,$3E,$83,$F2,$90,$69,$23,$FC,$EB,$1F,$E8,$07,$B9,$F7,$1C
	.dc.b	$75,$7F,$B0,$57,$BB,$E0,$F8,$FF,$0F,$CF,$78,$BE,$7E,$E1,$CB,$0F
	.dc.b	$DF,$E0,$9F,$3F,$CF,$F9,$09,$5E,$27,$CF,$ED,$7B,$0F,$6F,$B3,$67
	.dc.b	$A7,$82,$C1,$F3,$BF,$0B,$C5,$F8,$6D,$AD,$CB,$E7,$7E,$19,$23,$E0
	.dc.b	$BF,$1B,$E7,$3E,$59,$F3,$3E,$76,$CF,$FB,$E3,$C9,$E4,$DC,$3F,$8C
	.dc.b	$EF,$FE,$AE,$C2,$D7,$92,$B8,$64,$7F,$AB,$C7,$EE,$1B,$19,$AF,$F7
	.dc.b	$1B,$5F,$F2,$77,$93,$FC,$A6,$7E,$64,$7E,$F8,$EF,$3E,$29,$D4,$7D
	.dc.b	$7F,$8F,$FC,$29,$FC,$7B,$CA,$7C,$9E,$30,$F8,$BE,$A7,$E2,$8C,$5F
	.dc.b	$DF,$7F,$92,$FD,$33,$8A,$9F,$CA,$3E,$9F,$F2,$14,$EC,$F5,$72,$2F
	.dc.b	$86,$3F,$4B,$E7,$BF,$2E,$72,$70,$7F,$BE,$FD,$77,$9D,$FC,$79,$B1
	.dc.b	$F8,$BC,$1F,$68,$F9,$4D,$5F,$F2,$4F,$49,$F2,$F4,$CE,$1F,$93,$E6
	.dc.b	$FE,$B3,$19,$8F,$C2,$FC,$FD,$DF,$F3,$CF,$86,$FC,$5E,$F5,$AF,$33
	.dc.b	$87,$F1,$BF,$3F,$6F,$F4,$CF,$CF,$F8,$3F,$DA,$F1,$5A,$FC,$9B,$87
	.dc.b	$F2,$BF,$1F,$6F,$F5,$CF,$E2,$BC,$2F,$C5,$F8,$96,$BF,$24,$FE,$3B
	.dc.b	$C4,$FC,$35,$FF,$C0,$1D,$FF,$5D,$5F,$8E,$9C,$CF,$F4,$26,$A9,$FD
	.dc.b	$EF,$1F,$E3,$DB,$5C,$EB,$F3,$7E,$2F,$F0,$EB,$93,$FE,$AD,$F8,$8B
	.dc.b	$1F,$81,$8F,$E8,$7F,$78,$FA,$0F,$ED,$AD,$8F,$D9,$CE,$27,$FA,$DF
	.dc.b	$C1,$FE,$52,$7E,$B2,$F9,$CC,$F8,$1A,$64,$FD,$BE,$4E,$98,$78,$4F
	.dc.b	$65,$F1,$9D,$61,$FB,$4F,$2F,$BF,$CD,$D9,$EE,$F8,$BE,$5F,$E3,$F0
	.dc.b	$30,$FF,$08,$E3,$F3,$5A,$FF,$6C,$6B,$F2,$FF,$3A,$75,$7F,$76,$6E
	.dc.b	$3E,$CC,$31,$3F,$A9,$CB,$F5,$BE,$75,$AF,$E3,$BE,$12,$33,$1F,$C8
	.dc.b	$38,$7F,$C1,$AE,$A3,$E6,$FC,$1F,$EF,$7E,$3B,$E7,$F8,$76,$B8,$38
	.dc.b	$79,$7C,$FD,$DB,$E4,$76,$B2,$4F,$F8,$2F,$F2,$37,$A3,$F8,$AF,$85
	.dc.b	$46,$31,$FF,$3C,$C6,$27,$53,$9C,$7F,$DC,$D2,$FC,$8F,$FC,$63,$1F
	.dc.b	$5E,$31,$F3,$3D,$EF,$FC,$35,$5A,$EF,$FE,$17,$9F,$FB,$4F,$2F,$85
	.dc.b	$F2,$FF,$E4,$8F,$98,$7E,$B6,$F1,$BF,$1F,$3F,$90,$EC,$97,$C0,$F9
	.dc.b	$4F,$E5,$3A,$7F,$C7,$18,$FD,$AE,$2F,$85,$18,$33,$F3,$9F,$9A,$38
	.dc.b	$5F,$D3,$7F,$3B,$F3,$3C,$89,$1F,$E9,$CA,$F9,$91,$DB,$C1,$FD,$19
	.dc.b	$C2,$7E,$E3,$FA,$6F,$0A,$47,$F6,$9D,$BD,$BE,$5E,$39,$7E,$BE,$9E
	.dc.b	$FA,$FC,$26,$F9,$E1,$F8,$EF,$EC,$3C,$4F,$69,$27,$C9,$5F,$EE,$06
	.dc.b	$F5,$7E,$D8,$E0,$FE,$3D,$FF,$27,$FC,$C7,$FB,$EA,$3F,$95,$7E,$C5
	.dc.b	$FE,$9E,$C7,$FA,$C7,$89,$FE,$EE,$F2,$FF,$CA,$B3,$9A,$FF,$22,$FA
	.dc.b	$5F,$DA,$35,$FF,$41,$18,$F6,$7F,$CD,$11,$98,$FF,$4D,$B5,$FF,$26
	.dc.b	$F8,$7F,$EF,$3A,$FD,$2F,$AB,$FD,$46,$FF,$AA,$BE,$6F,$53,$FC,$CB
	.dc.b	$F5,$8F,$83,$C5,$FF,$41,$8C,$37,$E1,$3F,$7F,$06,$92,$3E,$D6,$7F
	.dc.b	$9B,$DF,$FA,$EB,$63,$FC,$C3,$F7,$7F,$E9,$49,$CF,$85,$FC,$83,$E8
	.dc.b	$65,$7E,$F3,$CC,$BC,$1F,$C5,$F8,$DF,$8A,$FF,$6A,$C7,$E5,$FC,$FB
	.dc.b	$83,$FC,$FF,$39,$C7,$FC,$F7,$39,$7D,$9F,$13,$EA,$23,$FE,$EC,$C1
	.dc.b	$C0,$9F,$CF,$FD,$2F,$F3,$43,$F9,$AF,$93,$6F,$FC,$1B,$F7,$3F,$E7
	.dc.b	$DD,$7F,$D7,$0D,$4D,$31,$ED,$7F,$91,$A7,$EE,$CA,$30,$FE,$01,$AF
	.dc.b	$FB,$0D,$AB,$EC,$5F,$19,$FE,$C6,$8E,$66,$5C,$3D,$A3,$7E,$3E,$A3
	.dc.b	$4D,$8F,$C1,$79,$D7,$A4,$D2,$57,$69,$7C,$97,$E5,$B2,$EC,$D9,$FC
	.dc.b	$CC,$E9,$FF,$17,$9F,$ED,$D7,$0F,$F7,$4C,$FF,$D8,$6F,$95,$AF,$B1
	.dc.b	$AF,$C8,$6B,$FD,$AB,$E6,$7F,$4F,$CB,$FC,$89,$B8,$66,$7F,$A2,$78
	.dc.b	$D8,$7F,$BE,$72,$3E,$17,$FC,$9E,$D8,$FC,$23,$5F,$AF,$CA,$8C,$47
	.dc.b	$F8,$2F,$C1,$FB,$4F,$F8,$A7,$D3,$FF,$2A,$67,$FC,$01,$B9,$7F,$B3
	.dc.b	$AB,$75,$FC,$5B,$E7,$BC,$9C,$5F,$D3,$5B,$E3,$9C,$F8,$9C,$32,$FD
	.dc.b	$6E,$BC,$3C,$1F,$34,$63,$1F,$F4,$0F,$82,$F9,$4F,$81,$C1,$E9,$F0
	.dc.b	$BD,$E4,$AF,$7B,$FE,$23,$F6,$77,$C4,$F8,$BA,$72,$F8,$0F,$F4,$ED
	.dc.b	$FF,$72,$E1,$F2,$91,$FF,$48,$EB,$FB,$56,$F8,$7F,$91,$94,$E6,$7F
	.dc.b	$47,$18,$9D,$4E,$67,$FC,$5D,$F1,$DE,$AF,$2D,$E4,$D3,$C9,$1F,$23
	.dc.b	$BA,$57,$F8,$0D,$F9,$BE,$FF,$F9,$8E,$3F,$D4,$AD,$7C,$2D,$7F,$98
	.dc.b	$E5,$FE,$35,$F8,$6F,$44,$E6,$7F,$DF,$6B,$E9,$FA,$92,$3A,$FC,$17
	.dc.b	$80,$8E,$0F,$EB,$1F,$24,$8F,$2F,$C3,$C9,$F2,$BE,$8F,$99,$91,$FD
	.dc.b	$9F,$04,$79,$09,$19,$F9,$9F,$79,$1F,$7F,$C5,$7C,$23,$D6,$47,$A3
	.dc.b	$EE,$F9,$8F,$8E,$48,$F8,$FD,$5F,$28,$8E,$0F,$5B,$FA,$DD,$FF,$C4
	.dc.b	$3F,$23,$7E,$13,$FD,$A3,$5F,$ED,$F9,$CB,$7D,$FF,$E1,$ED,$8F,$E2
	.dc.b	$C2,$33,$1F,$E2,$66,$BF,$D9,$A3,$47,$D6,$F9,$8F,$BF,$E6,$49,$FB
	.dc.b	$F1,$FA,$EF,$A5,$FB,$11,$82,$1F,$E3,$E7,$2F,$F9,$03,$F2,$E7,$C2
	.dc.b	$7E,$2B,$E6,$A4,$FF,$BB,$A7,$2F,$FB,$C3,$E7,$DF,$15,$FB,$C9,$CC
	.dc.b	$F3,$BF,$DF,$D6,$A3,$FD,$61,$F9,$AB,$2F,$DB,$B6,$33,$1F,$BA,$F3
	.dc.b	$FF,$94,$F9,$3F,$C3,$75,$FF,$73,$C7,$F2,$CA,$FF,$4F,$35,$1F,$E9
	.dc.b	$B7,$0E,$3E,$BB,$B3,$FD,$BE,$3F,$84,$3F,$C6,$F8,$FF,$94,$29,$DD
	.dc.b	$FF,$1D,$E7,$F3,$F3,$4A,$FE,$45,$1E,$7E,$7F,$43,$23,$F9,$34,$7E
	.dc.b	$D3,$E7,$64,$7E,$D1,$FD,$BF,$B7,$23,$F6,$ED,$7E,$B7,$67,$93,$2F
	.dc.b	$87,$E9,$7F,$8A,$1B,$3E,$DD,$B5,$AC,$8F,$EC,$53,$EE,$7F,$B4,$B5
	.dc.b	$FE,$EA,$D7,$A6,$9D,$FA,$E4,$7C,$CD,$98,$EA,$3F,$86,$FC,$07,$F1
	.dc.b	$0E,$F9,$18,$38,$4F,$D4,$F0,$9D,$4F,$03,$F8,$BE,$12,$F6,$2B,$73
	.dc.b	$FC,$F3,$F6,$17,$B9,$FC,$7B,$DA,$97,$FE,$FB,$9C,$B7,$DD,$E1,$52
	.dc.b	$30,$7F,$C0,$1F,$C5,$7F,$C1,$AF,$F2,$6F,$77,$F9,$49,$FD,$2A,$33
	.dc.b	$1F,$E3,$56,$EF,$FE,$5E,$6E,$DF,$90,$F4,$D1,$8A,$FF,$76,$64,$F8
	.dc.b	$6F,$51,$23,$D0,$FF,$77,$C1,$EA,$FF,$61,$C7,$F9,$8F,$CF,$7F,$3C
	.dc.b	$25,$FE,$CF,$E3,$7F,$60,$47,$07,$ED,$7C,$3F,$E8,$7F,$3F,$23,$FA
	.dc.b	$1B,$FD,$1A,$E4,$7F,$46,$6B,$FC,$17,$C0,$FD,$1F,$E2,$FF,$ED,$A8
	.dc.b	$FA,$A1,$C3,$FE,$EF,$31,$F0,$DF,$97,$E5,$91,$F9,$27,$FC,$68,$FE
	.dc.b	$B5,$BD,$9F,$83,$8F,$CB,$F9,$FF,$9F,$F0,$24,$78,$27,$F8,$BF,$07
	.dc.b	$E4,$A3,$F0,$DF,$82,$3E,$87,$FC,$1A,$D8,$FC,$5F,$5A,$31,$1F,$ED
	.dc.b	$E5,$E9,$F0,$F6,$FF,$BC,$47,$F8,$02,$7F,$BB,$B5,$FE,$CE,$D7,$F3
	.dc.b	$8B,$D9,$F0,$FC,$CC,$9F,$0C,$E5,$BF,$37,$E6,$23,$AF,$9B,$F4,$FC
	.dc.b	$EF,$88,$91,$FE,$C9,$A4,$79,$BE,$83,$DF,$F3,$72,$BB,$C7,$C2,$F9
	.dc.b	$09,$1E,$27,$E2,$FB,$5F,$1B,$E0,$7D,$E4,$9E,$CF,$C2,$F8,$4F,$97
	.dc.b	$F0,$92,$7D,$71,$F3,$BE,$0E,$47,$B9,$FA,$3F,$03,$F4,$BA,$FE,$59
	.dc.b	$27,$D4,$FC,$EF,$91,$FA,$FF,$1D,$27,$E3,$C7,$DA,$61,$F2,$8F,$BB
	.dc.b	$CF,$FE,$4A,$F9,$F7,$C1,$FF,$8B,$DB,$3F,$EE,$F6,$BF,$E3,$06,$BF
	.dc.b	$E2,$05,$ED,$73,$F9,$C9,$3F,$88,$9C,$B7,$E6,$BC,$04,$75,$F0,$3E
	.dc.b	$3F,$23,$C4,$48,$F1,$3E,$5F,$84,$F4,$79,$52,$7B,$47,$F9,$93,$EB
	.dc.b	$7F,$80,$47,$F2,$16,$EA,$69,$FB,$0F,$0F,$F4,$51,$FC,$B3,$DA,$FF
	.dc.b	$28,$7B,$06,$92,$FF,$D8,$5F,$49,$FE,$B4,$BF,$D5,$7C,$33,$E8,$BF
	.dc.b	$DC,$15,$FE,$19,$6B,$FE,$D5,$AE,$1E,$7F,$F3,$BE,$2F,$E9,$98,$78
	.dc.b	$9B,$19,$F7,$BF,$C0,$3E,$D4,$C3,$FD,$73,$AB,$E1,$53,$FE,$67,$6C
	.dc.b	$E9,$AF,$F9,$A3,$CA,$F8,$18,$7F,$4B,$6C,$7F,$47,$6D,$FF,$A6,$4F
	.dc.b	$F4,$7F,$C6,$7F,$B1,$B3,$F8,$ED,$BF,$EA,$8D,$B7,$CD,$3E,$B7,$FC
	.dc.b	$69,$1F,$C2,$BE,$B5,$F3,$DF,$E9,$A6,$C6,$63,$FD,$3B,$F1,$FF,$E4
	.dc.b	$28,$65,$F5,$A3,$0D,$F1,$F6,$70,$F2,$7F,$C2,$3F,$5B,$FD,$1E,$7F
	.dc.b	$C3,$7F,$57,$FE,$82,$C3,$FC,$3B,$2F,$9E,$FF,$0F,$C2,$74,$A7,$4F
	.dc.b	$2F,$1F,$71,$D6,$38,$46,$3D,$FF,$3C,$60,$EF,$7D,$AF,$AD,$F4,$9D
	.dc.b	$1D,$EB,$8C,$E9,$E7,$F2,$BC,$2F,$CE,$DA,$FF,$1B,$BB,$C6,$E3,$E3
	.dc.b	$7D,$13,$E2,$29,$ED,$AF,$24,$EE,$3C,$78,$FB,$71,$ED,$18,$AD,$BF
	.dc.b	$2B,$1F,$3D,$F1,$8F,$96,$7A,$F0,$AE,$DA,$E4,$CE,$0F,$9F,$7D,$6E
	.dc.b	$A3,$0D,$8C,$CF,$06,$BD,$3E,$B7,$FB,$2A,$69,$1F,$ED,$2F,$C6,$7F
	.dc.b	$D7,$3C,$B5,$FC,$1C,$57,$FC,$55,$15,$E8,$7F,$BB,$F1,$FF,$69,$C6
	.dc.b	$67,$87,$37,$F6,$DD,$BF,$17,$7F,$E6,$C9,$CD,$6E,$39,$1E,$22,$9E
	.dc.b	$22,$9E,$22,$9E,$29,$EB,$5D,$EF,$E7,$DF,$6A,$DB,$6B,$F4,$54,$FA
	.dc.b	$DF,$F6,$A6,$1F,$C3,$1B,$7F,$EE,$E6,$BF,$ED,$E6,$BF,$E5,$37,$FD
	.dc.b	$A9,$1C,$7D,$DF,$F0,$64,$ED,$AE,$DA,$FF,$95,$7E,$B7,$FA,$4C,$FF
	.dc.b	$1C,$6C,$7F,$BF,$9A,$FF,$BB,$9A,$FF,$A5,$BA,$3F,$DF,$BF,$15,$D1
	.dc.b	$B5,$F9,$06,$CF,$E5,$1A,$FE,$5B,$E2,$5E,$8F,$92,$11,$8A,$FC,$BB
	.dc.b	$63,$F3,$6D,$7F,$D9,$9F,$88,$3D,$5F,$6F,$8C,$8C,$5F,$E8,$A3,$13
	.dc.b	$A9,$CD,$7E,$A2,$5F,$73,$FD,$D2,$F8,$36,$74,$D7,$FD,$B8,$D7,$FC
	.dc.b	$D5,$E3,$3A,$FE,$7E,$48,$C5,$FF,$5A,$9C,$C6,$2B,$FB,$F7,$D3,$FD
	.dc.b	$6E,$27,$A2,$8C,$37,$0F,$F0,$34,$63,$C0,$91,$FE,$09,$6D,$FD,$1F
	.dc.b	$C1,$7F,$D4,$33,$F5,$5B,$19,$8F,$B5,$18,$8F,$E6,$BC,$2F,$C2,$7F
	.dc.b	$03,$1F,$D3,$9B,$19,$F5,$3F,$05,$18,$FC,$49,$F1,$7F,$85,$6E,$3F
	.dc.b	$85,$6E,$1F,$85,$6B,$E0,$7C,$4F,$E3,$DA,$FF,$5E,$6E,$06,$FE,$3D
	.dc.b	$FE,$6B,$AF,$C9,$36,$BF,$24,$D7,$DA,$D7,$F3,$8B,$E1,$3F,$38,$D9
	.dc.b	$E7,$18,$3F,$41,$F5,$4F,$84,$A4,$9F,$D1,$46,$5F,$1F,$C2,$F4,$23
	.dc.b	$83,$CF,$F0,$7E,$02,$38,$3D,$7E,$57,$90,$8E,$0F,$7F,$C1,$7A,$08
	.dc.b	$E0,$F8,$7F,$05,$EC,$23,$83,$CF,$F0,$7E,$C2,$38,$3E,$5F,$7F,$C4
	.dc.b	$23,$83,$E7,$FC,$0F,$C8,$23,$83,$E9,$EF,$F3,$08,$E0,$FE,$11,$F5
	.dc.b	$8F,$63,$F8,$63,$63,$31,$83,$F8,$47,$73,$3A,$49,$FE,$12,$EB,$29
	.dc.b	$1F,$C3,$1B,$87,$F0,$CD,$39,$92,$7F,$88,$BF,$B3,$F8,$96,$74,$93
	.dc.b	$FE,$44,$8C,$BF,$E7,$9F,$AC,$FD,$B9,$1C,$1F,$E4,$AC,$FF,$37,$DD
	.dc.b	$ED,$8C,$1F,$C0,$63,$3F,$47,$FE,$BE,$8F,$F2,$8B,$72,$FE,$FF,$48
	.dc.b	$C4,$7F,$8B,$23,$1F,$17,$FC,$4B,$E1,$1F,$33,$49,$87,$F2,$C8,$CB
	.dc.b	$FE,$08,$F4,$DF,$60,$FA,$4F,$E9,$31,$86,$C6,$67,$83,$72,$FE,$7D
	.dc.b	$7F,$E7,$B9,$ED,$F0,$5E,$32,$3E,$1F,$93,$F3,$11,$C1,$E5,$F9,$2F
	.dc.b	$51,$1C,$1F,$E7,$4C,$1F,$15,$FC,$A5,$B1,$98,$C1,$FC,$AB,$D7,$91
	.dc.b	$FE,$AF,$6E,$5F,$D3,$BE,$3B,$F1,$A4,$9F,$EA,$6F,$E4,$3B,$A4,$7F
	.dc.b	$56,$6C,$FF,$56,$F0,$3F,$2A,$49,$FE,$B6,$F9,$7E,$2A,$47,$A5,$B3
	.dc.b	$FE,$BF,$F7,$3F,$9F,$92,$7F,$D8,$8F,$F4,$CF,$5F,$FC,$07,$3F,$DA
	.dc.b	$1B,$DD,$FE,$0D,$F9,$2D,$1F,$D4,$BE,$C1,$D7,$49,$8F,$F6,$C8,$CB
	.dc.b	$FD,$BB,$E0,$9E,$37,$F5,$A6,$C6,$63,$07,$EC,$BE,$16,$47,$F7,$86
	.dc.b	$E7,$FD,$FE,$DF,$1B,$E3,$25,$78,$5F,$EC,$1D,$DF,$CC,$E9,$DD,$F9
	.dc.b	$DF,$03,$24,$BF,$D1,$E3,$C0,$8E,$0F,$C2,$7C,$17,$E0,$C8,$E0,$FF
	.dc.b	$77,$3F,$DB,$08,$E0,$FF,$33,$FE,$77,$F9,$BC,$3F,$CD,$AD,$7F,$D3
	.dc.b	$2D,$41,$C7,$C1,$3F,$1F,$FE,$FD,$3F,$EA,$08,$C7,$B3,$FE,$34,$6D
	.dc.b	$FF,$9E,$1B,$3F,$BE,$EC,$7B,$DF,$9D,$25,$7E,$85,$FE,$4F,$D9,$23
	.dc.b	$F9,$5B,$6B,$FC,$65,$E5,$FF,$AC,$67,$F9,$7E,$9F,$CB,$88,$FF,$B7
	.dc.b	$BD,$9F,$BF,$5F,$C3,$9A,$FF,$0D,$6B,$F0,$38,$3F,$C2,$FC,$9F,$CE
	.dc.b	$32,$97,$FC,$F1,$B5,$CD,$FF,$0D,$92,$3C,$0F,$EE,$7D,$C9,$1F,$BD
	.dc.b	$6C,$FA,$BE,$AF,$FA,$29,$27,$FA,$4B,$FD,$23,$E0,$A4,$7F,$4C,$6C
	.dc.b	$FA,$FE,$2F,$FC,$56,$49,$F8,$0F,$F8,$E3,$E2,$3E,$DC,$7F,$94,$1B
	.dc.b	$3F,$E5,$96,$BF,$ED,$AE,$D4,$62,$3F,$07,$18,$F3,$BA,$12,$BA,$3F
	.dc.b	$E4,$6F,$AB,$FF,$47,$57,$36,$BC,$9A,$FF,$99,$9B,$5E,$0F,$93,$F4
	.dc.b	$12,$FD,$0F,$AF,$E5,$24,$7A,$DB,$7F,$E7,$3F,$AF,$7D,$07,$F9,$61
	.dc.b	$B5,$FE,$70,$6C,$EE,$7F,$6F,$F0,$F2,$BF,$AA,$B6,$7F,$CF,$DF,$60
	.dc.b	$EE,$FF,$31,$B6,$BF,$CF,$6D,$9D,$CF,$F9,$A3,$E4,$19,$52,$5F,$FA
	.dc.b	$0A,$32,$FE,$5F,$8B,$FC,$CA,$47,$07,$FB,$83,$47,$99,$FE,$8F,$6C
	.dc.b	$66,$30,$67,$E6,$64,$71,$6E,$5F,$EE,$CF,$B0,$7D,$0F,$F1,$68,$C3
	.dc.b	$5F,$F7,$9C,$6B,$AB,$FD,$E2,$D8,$E5,$F4,$4E,$9C,$B6,$33,$18,$38
	.dc.b	$FC,$B7,$FB,$EA,$14,$8C,$36,$36,$78,$3E,$9F,$24,$70,$7E,$FF,$EB
	.dc.b	$FF,$8C,$BF,$BF,$6B,$FB,$F6,$BF,$BF,$EF,$47,$FA,$DB,$E2,$BF,$C6
	.dc.b	$64,$DB,$07,$FA,$EB,$AB,$FC,$53,$7F,$EB,$A6,$BF,$EC,$36,$BF,$A1
	.dc.b	$EB,$EE,$FF,$42,$F2,$9F,$07,$8E,$0F,$F1,$03,$63,$33,$FE,$26,$8F
	.dc.b	$F6,$4F,$37,$FB,$16,$3F,$11,$D5,$82,$3F,$EC,$CE,$4F,$CB,$37,$A7
	.dc.b	$FD,$A3,$18,$8F,$F6,$C3,$5F,$E0,$0F,$FA,$1D,$E5,$F0,$BC,$08,$FF
	.dc.b	$09,$F9,$4F,$F6,$DF,$80,$69,$23,$C5,$F0,$1E,$02,$38,$3F,$DC,$3F
	.dc.b	$BD,$FF,$98,$09,$C3,$FE,$E6,$7F,$DD,$5B,$FF,$B2,$3C,$4F,$8D,$CA
	.dc.b	$69,$2F,$FD,$ED,$F4,$DF,$E3,$D6,$FC,$07,$D0,$F8,$B4,$E6,$7F,$DF
	.dc.b	$4D,$7F,$D8,$AB,$E5,$78,$36,$3A,$C6,$0F,$B1,$F4,$8F,$AD,$A4,$9F
	.dc.b	$C6,$E6,$9D,$47,$F5,$3E,$FF,$59,$1C,$1E,$AF,$82,$F0,$91,$C1,$D3
	.dc.b	$E7,$BD,$0F,$F5,$76,$BF,$E0,$E6,$D9,$CE,$7F,$0B,$E1,$3D,$A4,$AF
	.dc.b	$92,$FC,$0F,$0A,$47,$CC,$6B,$FE,$76,$C9,$E1,$FC,$22,$47,$C0,$8C
	.dc.b	$BF,$E0,$1F,$CC,$BF,$90,$B7,$C9,$FD,$1F,$AA,$9C,$CF,$ED,$A3,$13
	.dc.b	$FC,$17,$F4,$2F,$F5,$24,$FE,$A9,$AF,$EE,$9A,$FE,$F9,$AF,$C5,$F7
	.dc.b	$FF,$5D,$A9,$A4,$8F,$47,$C9,$7D,$C2,$38,$3E,$DF,$CF,$BE,$27,$F7
	.dc.b	$0D,$8C,$C6,$0C,$FC,$44,$8E,$2D,$7F,$89,$FA,$8F,$2B,$F8,$F4,$61
	.dc.b	$AF,$F7,$36,$BF,$DC,$9A,$FF,$8D,$BE,$95,$F5,$BF,$CA,$DB,$1F,$C4
	.dc.b	$1A,$FF,$20,$6B,$FD,$FF,$F2,$57,$47,$F1,$6C,$11,$89,$FE,$4F,$18
	.dc.b	$9D,$4E,$67,$D1,$E7,$FF,$29,$23,$C3,$14,$75,$FD,$8D,$EA,$D2,$57
	.dc.b	$68,$CB,$F9,$FE,$EF,$B2,$47,$07,$EB,$BF,$5B,$79,$5F,$CA,$DB,$19
	.dc.b	$8C,$1F,$17,$E4,$24,$7C,$56,$BF,$17,$B7,$E5,$11,$F9,$4F,$CD,$CA
	.dc.b	$47,$CD,$6B,$FE,$DA,$E6,$7C,$7F,$E4,$09,$1F,$B9,$7C,$DE,$BC,$8F
	.dc.b	$53,$5F,$C2,$FD,$63,$DD,$FA,$8D,$8F,$A6,$D7,$71,$AF,$7E,$46,$9A
	.dc.b	$F4,$F8,$DE,$64,$79,$BF,$D5,$BE,$47,$F8,$C7,$BB,$7C,$33,$49,$5E
	.dc.b	$2F,$84,$F2,$91,$C1,$FA,$8F,$8A,$FD,$39,$1C,$1F,$AC,$F8,$FF,$D5
	.dc.b	$91,$C1,$FC,$7F,$AB,$F7,$24,$70,$7E,$07,$D8,$FC,$01,$1C,$1F,$A1
	.dc.b	$E7,$FD,$21,$1C,$1F,$A5,$F0,$FF,$50,$47,$07,$EA,$7C,$FF,$D6,$11
	.dc.b	$C1,$FD,$43,$EB,$3F,$A3,$91,$C1,$FD,$53,$D4,$FE,$96,$47,$07,$F8
	.dc.b	$4B,$CD,$7E,$3A,$F8,$DF,$EB,$0D,$9D,$35,$FF,$0D,$CF,$09,$DF,$B5
	.dc.b	$27,$6D,$77,$F4,$AF,$B0,$FE,$C8,$D8,$F1,$35,$DC,$78,$BE,$56,$47
	.dc.b	$89,$AF,$8B,$17,$CD,$FA,$09,$1E,$87,$D7,$F3,$12,$3D,$6D,$79,$7C
	.dc.b	$D7,$C0,$23,$F0,$1F,$87,$F0,$72,$3E,$1B,$5F,$6F,$A3,$EC,$23,$EC
	.dc.b	$7E,$5F,$CD,$C8,$F9,$6D,$7D,$5F,$0D,$F4,$08,$FD,$07,$E9,$FB,$52
	.dc.b	$3E,$9B,$5F,$A7,$F3,$CF,$AC,$FF,$08,$36,$3E,$F3,$5D,$C7,$F8,$57
	.dc.b	$E8,$1F,$63,$49,$3F,$85,$8C,$BF,$87,$FC,$2B,$F8,$EB,$F8,$26,$BF
	.dc.b	$E6,$0E,$64,$62,$3F,$DF,$71,$8F,$3B,$FB,$1F,$99,$27,$FC,$18,$D7
	.dc.b	$F8,$D7,$9B,$FD,$A0,$8F,$F8,$45,$FE,$8D,$F4,$BF,$E7,$9F,$54,$D2
	.dc.b	$4F,$F9,$83,$E0,$7F,$0B,$E1,$DE,$7F,$E9,$13,$FE,$76,$9C,$BF,$E7
	.dc.b	$46,$BF,$E7,$85,$F4,$FF,$E9,$38,$E0,$D4,$7F,$A4,$5C,$37,$8B,$FD
	.dc.b	$1D,$F8,$EF,$F2,$78,$FF,$41,$B5,$FF,$43,$B5,$E6,$DD,$9F,$58,$FA
	.dc.b	$FF,$F2,$63,$6B,$F7,$4D,$7F,$6A,$D7,$FD,$27,$6F,$8F,$FF,$50,$36
	.dc.b	$3F,$D0,$51,$83,$3F,$1C,$F9,$CF,$F4,$33,$67,$FC,$9A,$D7,$FD,$3B
	.dc.b	$5F,$EF,$89,$DF,$D2,$C9,$DB,$5F,$F5,$0F,$86,$F4,$9E,$97,$CA,$F8
	.dc.b	$CF,$F3,$E5,$7C,$CA,$7C,$DF,$FA,$B2,$73,$29,$CB,$C7,$E4,$7F,$D6
	.dc.b	$31,$D6,$F1,$3C,$F0,$EE,$FF,$9E,$EB,$78,$75,$6C,$70,$3B,$4F,$0C
	.dc.b	$3C,$38,$75,$9E,$47,$9E,$F9,$34,$E5,$1B,$9E,$35,$CE,$FB,$D7,$23
	.dc.b	$D9,$7C,$9F,$0E,$1C,$AB,$84,$75,$38,$61,$F0,$9B,$1A,$78,$9F,$41
	.dc.b	$F8,$2F,$FB,$52,$3D,$4D,$8C,$BF,$2B,$D3,$7C,$4F,$FB,$2A,$91,$89
	.dc.b	$FB,$6D,$7A,$35,$FF,$6F,$FD,$67,$FB,$BD,$34,$8F,$FB,$CD,$FD,$5F
	.dc.b	$D6,$FD,$E8,$6E,$F3,$F3,$AF,$49,$56,$EB,$51,$89,$D4,$E6,$B8,$CF
	.dc.b	$FB,$D2,$EB,$FB,$AB,$E3,$5E,$18,$F3,$9F,$54,$F1,$D7,$E1,$5F,$E0
	.dc.b	$35,$6B,$C3,$B7,$18,$EB,$3C,$0F,$21,$E1,$29,$E1,$AD,$BE,$0B,$F4
	.dc.b	$3A,$3B,$96,$FC,$F8,$7A,$7E,$95,$E9,$7C,$2D,$FE,$1D,$3E,$91,$D7
	.dc.b	$9C,$F8,$5D,$CB,$E7,$D6,$A7,$F8,$EE,$1C,$A3,$13,$FB,$89,$F0,$7A
	.dc.b	$9C,$71,$ED,$8D,$CB,$FD,$41,$5C,$8F,$9D,$F5,$AF,$A8,$E0,$FF,$59
	.dc.b	$8D,$63,$F8,$08,$D4,$63,$04,$6E,$7C,$46,$FC,$9E,$7F,$0F,$D1,$B7
	.dc.b	$D3,$27,$E1,$4F,$53,$F0,$F8,$7F,$43,$8C,$57,$F4,$36,$C7,$F4,$36
	.dc.b	$BF,$EA,$AF,$3F,$E4,$2F,$1D,$1F,$2F,$0E,$F8,$6A,$73,$87,$8F,$EB
	.dc.b	$BD,$B8,$7A,$AD,$18,$E2,$FB,$3F,$0A,$F8,$6E,$93,$9C,$3F,$A8,$DF
	.dc.b	$0D,$7D,$78,$E2,$BE,$33,$FD,$7D,$AE,$63,$6F,$A7,$CA,$FE,$81,$7F
	.dc.b	$1F,$EB,$FF,$D7,$AF,$D4,$D3,$FB,$7F,$85,$ED,$9C,$DF,$D2,$E3,$FF
	.dc.b	$80,$1F,$69,$F8,$3D,$8D,$F9,$6C,$39,$F8,$FF,$0D,$2B,$9D,$FF,$58
	.dc.b	$6B,$FB,$56,$BF,$C1,$1F,$F0,$CE,$7F,$85,$5F,$2F,$E1,$F5,$7C,$4F
	.dc.b	$8F,$E6,$25,$6F,$DE,$B8,$7A,$70,$7D,$1F,$91,$B5,$FD,$FA,$7C,$93
	.dc.b	$93,$5F,$45,$EF,$F9,$AF,$80,$DA,$F2,$7B,$BF,$0C,$F8,$9F,$09,$63
	.dc.b	$D0,$77,$21,$DF,$0F,$80,$FC,$6F,$84,$FF,$3B,$3F,$A3,$5F,$3B,$FB
	.dc.b	$17,$33,$F4,$DC,$CE,$FE,$17,$46,$FE,$7A,$C7,$3F,$7B,$B1,$1F,$15
	.dc.b	$FF,$32,$BF,$F6,$FC,$63,$8F,$FD,$11,$E0,$78,$7E,$B3,$C4,$D9,$F3
	.dc.b	$7B,$9E,$53,$CD,$F2,$36,$3D,$5D,$9E,$D2,$3D,$71,$FC,$9E,$5F,$E0
	.dc.b	$E5,$EB,$7B,$FE,$1F,$E0,$36,$BE,$27,$C5,$7C,$33,$E2,$7C,$45,$8F
	.dc.b	$91,$F1,$1F,$2C,$8F,$77,$E7,$7C,$E7,$CB,$F4,$FE,$63,$67,$F8,$AE
	.dc.b	$6F,$B7,$FE,$25,$E5,$FF,$0A,$6C,$FF,$4B,$D7,$F7,$F5,$FC,$58,$E1
	.dc.b	$E4,$7F,$03,$CF,$F6,$31,$FC,$25,$BE,$39,$A6,$36,$FE,$14,$71,$C3
	.dc.b	$C7,$3F,$11,$FC,$27,$C7,$FE,$0F,$9B,$F0,$8D,$8F,$C4,$47,$E1,$FE
	.dc.b	$2F,$18,$FE,$1D,$FC,$53,$FC,$EF,$EA,$7D,$6F,$E0,$4C,$67,$2F,$F9
	.dc.b	$A7,$A7,$FD,$D5,$FE,$6D,$5C,$30,$FE,$73,$1F,$C2,$7D,$4F,$DB,$E1
	.dc.b	$FC,$E3,$34,$7F,$B1,$67,$5E,$47,$F5,$39,$CF,$8D,$6B,$57,$C7,$E7
	.dc.b	$3F,$92,$61,$D6,$B8,$46,$27,$FA,$2F,$1B,$3F,$EC,$B7,$FD,$3A,$31
	.dc.b	$EC,$70,$F4,$BF,$B1,$46,$27,$F9,$06,$0F,$8A,$EA,$7C,$5D,$EC,$EB
	.dc.b	$E2,$7C,$26,$33,$F3,$3F,$2E,$E4,$FC,$6F,$5E,$D7,$F2,$9D,$1F,$39
	.dc.b	$F4,$4C,$E7,$FD,$EA,$33,$7F,$80,$F0,$7F,$A6,$1F,$D4,$B5,$7C,$D6
	.dc.b	$27,$F0,$BE,$17,$E3,$08,$FE,$F3,$2F,$EB,$11,$F9,$0F,$0F,$FA,$B1
	.dc.b	$FD,$5D,$B3,$F7,$3E,$4F,$F0,$9F,$13,$F8,$56,$C7,$E0,$23,$F1,$1F
	.dc.b	$27,$8C,$3D,$8F,$D9,$F9,$39,$5F,$83,$F9,$8B,$5F,$83,$E8,$FF,$03
	.dc.b	$B7,$E0,$33,$C3,$FE,$3C,$EA,$FC,$3C,$7E,$17,$E0,$5F,$51,$4F,$CC
	.dc.b	$7D,$8D,$CF,$FA,$C4,$BE,$63,$FA,$FE,$1F,$E2,$36,$C6,$67,$F8,$FC
	.dc.b	$E5,$F8,$CB,$D0,$FE,$F4,$78,$6A,$CF,$E8,$3E,$BB,$F5,$07,$EA,$7E
	.dc.b	$C6,$C7,$F8,$67,$E9,$BF,$9A,$EB,$FE,$34,$6C,$7F,$94,$7A,$97,$FE
	.dc.b	$2B,$9C,$EF,$FE,$62,$9F,$F5,$45,$7E,$BB,$E3,$BF,$6D,$F1,$9F,$B7
	.dc.b	$7F,$71,$F1,$F2,$74,$7E,$D1,$FF,$1F,$7C,$7F,$FA,$C3,$AB,$FC,$02
	.dc.b	$D9,$FE,$7E,$DC,$7F,$A9,$35,$FE,$A9,$F5,$DF,$AA,$7F,$8E,$37,$5F
	.dc.b	$F3,$8C,$F1,$7F,$C5,$B1,$9A,$F7,$6F,$FC,$37,$E7,$7F,$87,$36,$3F
	.dc.b	$84,$C7,$F1,$1E,$5C,$63,$FA,$57,$F8,$A7,$77,$F8,$23,$7F,$CF,$DF
	.dc.b	$F5,$36,$D7,$FB,$51,$AF,$F7,$9F,$95,$7D,$6F,$FB,$32,$BF,$C9,$F1
	.dc.b	$8D,$FA,$C6,$27,$55,$FB,$5F,$AA,$F0,$12,$7C,$0F,$F8,$AA,$FA,$E5
	.dc.b	$FE,$6B,$6C,$FF,$24,$9C,$5F,$E6,$23,$F9,$3F,$C9,$FF,$9A,$7E,$7F
	.dc.b	$E3,$63,$F7,$E3,$1E,$64,$AF,$E5,$BF,$1F,$6B,$F8,$43,$C7,$BB,$FC
	.dc.b	$57,$87,$EA,$A7,$FE,$00,$B5,$FF,$4F,$35,$67,$F1,$31,$F2,$9C,$F0
	.dc.b	$F8,$4F,$F3,$3F,$96,$FF,$4D,$72,$7E,$CD,$B3,$D3,$C2,$D7,$A3,$BC
	.dc.b	$7F,$5F,$46,$92,$BF,$5F,$0F,$4F,$B1,$C6,$3F,$CA,$11,$96,$FC,$8F
	.dc.b	$8C,$FE,$8D,$F5,$8F,$83,$C4,$F0,$F5,$FC,$24,$7B,$BE,$4F,$73,$FA
	.dc.b	$77,$D6,$BE,$27,$DD,$18,$73,$F1,$5F,$57,$CC,$B8,$7F,$A4,$F9,$11
	.dc.b	$E2,$F7,$A4,$E7,$D7,$BC,$3F,$D7,$F6,$FF,$19,$BF,$E3,$DF,$2D,$18
	.dc.b	$8F,$F2,$74,$63,$E1,$FF,$8B,$CE,$67,$E5,$4B,$6F,$EC,$7C,$DF,$EB
	.dc.b	$ED,$3F,$AB,$65,$FE,$77,$9C,$BF,$80,$F6,$BF,$D2,$FC,$2F,$5B,$FC
	.dc.b	$22,$D9,$FE,$88,$DC,$BF,$C8,$2D,$7F,$CA,$8B,$E1,$B8,$92,$78,$BF
	.dc.b	$E1,$CF,$AD,$7C,$5F,$FB,$EA,$FF,$C9,$71,$8E,$AF,$EA,$4D,$8C,$CE
	.dc.b	$DF,$F3,$14,$7D,$5E,$FD,$F3,$7F,$B6,$A7,$9B,$5D,$B5,$FF,$3C,$4F
	.dc.b	$FA,$A5,$EF,$EC,$FF,$61,$51,$97,$CB,$19,$87,$25,$24,$FA,$1F,$E8
	.dc.b	$BF,$34,$F8,$6F,$F4,$37,$1F,$D3,$F9,$EB,$BF,$EC,$1B,$87,$FB,$1E
	.dc.b	$F8,$3F,$EA,$DF,$AD,$70,$FF,$50,$6C,$7F,$54,$6B,$F8,$96,$BF,$88
	.dc.b	$7F,$DA,$2F,$3E,$EE,$FC,$E9,$CC,$FF,$AD,$F9,$FE,$06,$7F,$DB,$71
	.dc.b	$7C,$4F,$9B,$91,$4F,$94,$C3,$70,$FF,$62,$5F,$07,$FD,$8D,$F1,$BF
	.dc.b	$D9,$DC,$FC,$4F,$8A,$7F,$CB,$91,$AF,$6B,$CB,$5F,$B9,$8F,$EE,$2D
	.dc.b	$7F,$4C,$D7,$F4,$AD,$7F,$46,$BB,$F9,$92,$3E,$67,$85,$DB,$C0,$91
	.dc.b	$E6,$7E,$2F,$D6,$B9,$69,$FE,$00,$7F,$CE,$64,$71,$F2,$1C,$09,$5F
	.dc.b	$C0,$FE,$57,$EE,$F9,$6F,$93,$79,$D8,$6D,$7F,$50,$6B,$F4,$FE,$57
	.dc.b	$E6,$9F,$E3,$5E,$4F,$A2,$DE,$33,$4C,$7C,$67,$D4,$F9,$59,$7F,$C5
	.dc.b	$63,$EB,$FC,$6F,$D8,$64,$7F,$19,$8F,$27,$81,$23,$C8,$FD,$EC,$A4
	.dc.b	$7F,$AF,$F1,$F6,$36,$3C,$87,$F9,$2F,$4F,$81,$DF,$EF,$24,$FF,$27
	.dc.b	$F3,$3F,$12,$7A,$C8,$F5,$7D,$DE,$37,$E3,$BC,$59,$3F,$CB,$E3,$F2
	.dc.b	$3F,$29,$F9,$26,$47,$F3,$38,$FD,$6A,$FA,$BF,$C7,$BF,$EB,$16,$F8
	.dc.b	$1F,$7E,$3F,$AA,$46,$39,$BF,$AB,$7B,$CF,$9D,$7C,$3F,$EE,$23,$0F
	.dc.b	$F9,$A5,$B7,$FE,$28,$9D,$B6,$7E,$C7,$4F,$EA,$09,$1F,$D8,$F1,$F1
	.dc.b	$73,$BE,$9F,$FD,$F7,$39,$CB,$FB,$44,$E7,$B3,$F6,$95,$AA,$5F,$28
	.dc.b	$C3,$FD,$0E,$FF,$75,$F5,$7F,$E8,$CA,$E8,$30,$46,$63,$9C,$F8,$23
	.dc.b	$15,$D7,$C1,$F0,$92,$B9,$DF,$A6,$3C,$BE,$07,$D2,$FA,$BF,$CB,$D7
	.dc.b	$FB,$A1,$AF,$FB,$8D,$B7,$FE,$7D,$FA,$BF,$E5,$F3,$E6,$F2,$51,$88
	.dc.b	$FE,$D0,$D7,$FB,$3B,$5F,$EC,$AD,$51,$CB,$D5,$FE,$47,$3F,$DE,$27
	.dc.b	$B3,$5F,$82,$7C,$5D,$7F,$89,$5F,$FA,$47,$EA,$FF,$D5,$B3,$E5,$6B
	.dc.b	$FE,$DB,$6C,$6F,$E0,$FF,$98,$C1,$A4,$9E,$5F,$01,$F7,$3D,$4F,$BB
	.dc.b	$E4,$C9,$F9,$B8,$FF,$6F,$6F,$C8,$ED,$CD,$7E,$12,$3E,$87,$74,$9F
	.dc.b	$A0,$FD,$3F,$5E,$47,$FB,$07,$83,$FD,$B8,$D8,$FE,$C9,$18,$8F,$F1
	.dc.b	$C1,$1F,$A5,$AF,$EA,$61,$18,$E5,$D7,$87,$FB,$2F,$61,$EE,$FE,$63
	.dc.b	$E1,$E6,$1E,$C7,$F3,$7F,$37,$23,$DE,$D7,$E9,$F4,$FE,$97,$DB,$FE
	.dc.b	$B7,$F5,$7F,$EB,$B9,$FC,$33,$5F,$B3,$D2,$8C,$6D,$FE,$C9,$5F,$0B
	.dc.b	$C8,$95,$C9,$FD,$6F,$CE,$7F,$76,$AF,$EC,$3E,$D5,$F3,$3F,$B8,$C7
	.dc.b	$F0,$93,$37,$E2,$F8,$CF,$BF,$F2,$5E,$66,$47,$E1,$E3,$F0,$7F,$29
	.dc.b	$A8,$C1,$F8,$B7,$F7,$BF,$35,$FB,$89,$93,$F8,$28,$FE,$03,$ED,$6A
	.dc.b	$30,$7E,$19,$FE,$0B,$EF,$FF,$06,$F7,$64,$FE,$2A,$3F,$84,$FC,$E6
	.dc.b	$A3,$07,$E3,$9F,$DC,$7C,$4C,$8F,$DC,$3F,$BB,$F6,$A4,$7E,$25,$FC
	.dc.b	$39,$1F,$C3,$9F,$88,$F9,$17,$BD,$FC,$D1,$B1,$FC,$48,$38,$7E,$FF
	.dc.b	$87,$78,$3F,$8E,$71,$7F,$20,$F8,$29,$5F,$A4,$8F,$E4,$7F,$13,$FD
	.dc.b	$9E,$FF,$D1,$53,$E2,$3F,$75,$19,$34,$FF,$2A,$21,$FA,$EE,$AF,$F1
	.dc.b	$06,$8F,$9A,$BF,$37,$2B,$F8,$FC,$E5,$BF,$35,$FD,$DE,$0D,$24,$78
	.dc.b	$ED,$3F,$CB,$3C,$3F,$E7,$AD,$9F,$E6,$8D,$9D,$FC,$DF,$93,$9F,$CB
	.dc.b	$D1,$2F,$C7,$1E,$8F,$9A,$F1,$12,$3C,$18,$23,$CD,$F3,$1E,$E2,$4F
	.dc.b	$B3,$D7,$F8,$3D,$9F,$0B,$E6,$24,$FC,$08,$F8,$BF,$31,$EF,$24,$76
	.dc.b	$7E,$17,$CB,$FC,$A2,$47,$C8,$F0,$BE,$6F,$A5,$F3,$BD,$29,$3F,$32
	.dc.b	$3E,$8F,$CB,$FC,$B2,$47,$A9,$F9,$DF,$29,$F5,$89,$1F,$53,$E5,$BF
	.dc.b	$CB,$1F,$7F,$FE,$2C,$EE,$FE,$DB,$C5,$FE,$85,$6F,$C9,$EB,$C2,$DC
	.dc.b	$FF,$8B,$EB,$EA,$E1,$F5,$DA,$FE,$73,$A1,$18,$8F,$E4,$B1,$8F,$6B
	.dc.b	$C2,$48,$FE,$CB,$3F,$C8,$9F,$EB,$9F,$41,$FA,$6C,$BF,$C2,$59,$3E
	.dc.b	$81,$F1,$1F,$E8,$68,$C3,$FE,$7A,$AE,$1D,$BF,$EE,$B9,$FF,$5D,$7A
	.dc.b	$BF,$9F,$C5,$18,$7F,$44,$D7,$F6,$AD,$7F,$9E,$BF,$EC,$3C,$FC,$1E
	.dc.b	$8F,$F6,$3C,$FC,$AD,$C3,$FC,$27,$F6,$7F,$AE,$C3,$FB,$38,$FA,$8F
	.dc.b	$EB,$CF,$EF,$5A,$E9,$B7,$F0,$39,$DE,$07,$BD,$B5,$E8,$6B,$E8,$6B
	.dc.b	$E8,$6B,$FE,$26,$FB,$17,$89,$FE,$99,$F9,$4F,$8F,$3F,$E1,$F6,$F2
	.dc.b	$2B,$73,$FE,$1A,$7F,$78,$FF,$50,$72,$DA,$FF,$52,$46,$B1,$FF,$1C
	.dc.b	$FC,$DF,$F9,$26,$BF,$C5,$8B,$D2,$FF,$55,$36,$3F,$C4,$EE,$F0,$3F
	.dc.b	$EA,$98,$FE,$BE,$E4,$9E,$8F,$F9,$73,$EA,$1E,$0F,$F9,$02,$9F,$50
	.dc.b	$F4,$BF,$BB,$D7,$F5,$E9,$1A,$8C,$EF,$FD,$F2,$B8,$4E,$AB,$AE,$9C
	.dc.b	$27,$32,$9C,$BC,$CE,$15,$CE,$FA,$C7,$17,$C7,$87,$9E,$F9,$DF,$2E
	.dc.b	$FB,$87,$F8,$39,$B8,$7A,$1D,$DF,$BD,$B8,$F5,$9F,$09,$4F,$35,$78
	.dc.b	$DF,$65,$78,$71,$F1,$D6,$F0,$F8,$D5,$EB,$38,$99,$AE,$0F,$29,$FD
	.dc.b	$74,$FC,$28,$FE,$BE,$D7,$A9,$FD,$45,$78,$8F,$8B,$E2,$D9,$D3,$5E
	.dc.b	$6D,$7F,$AD,$7C,$5F,$EF,$5F,$B6,$D7,$FD,$2C,$65,$C1,$F0,$DF,$F3
	.dc.b	$F7,$3D,$77,$AF,$DC,$72,$7C,$13,$A9,$78,$3F,$D4,$D1,$B8,$E3,$A7
	.dc.b	$F2,$3A,$E9,$7F,$5E,$31,$2D,$B8,$4E,$63,$F9,$3D,$57,$FC,$9B,$1E
	.dc.b	$7C,$3B,$DF,$C0,$9F,$15,$76,$69,$E7,$3C,$67,$43,$F2,$92,$FA,$1F
	.dc.b	$14,$FF,$A5,$9B,$3F,$E4,$C5,$E1,$FF,$8E,$E7,$FD,$5F,$1A,$DB,$FD
	.dc.b	$6B,$C9,$FC,$5A,$7F,$83,$73,$EF,$0F,$F5,$55,$EE,$7F,$89,$4E,$7E
	.dc.b	$47,$F6,$9C,$57,$4F,$DC,$D7,$FA,$A3,$E2,$BF,$6A,$FF,$AD,$2B,$6F
	.dc.b	$FA,$B3,$E2,$DE,$3F,$7D,$FF,$D6,$EF,$3C,$71,$3A,$AF,$F5,$9F,$A4
	.dc.b	$F9,$7F,$F3,$B6,$DF,$EC,$A8,$C4,$FF,$84,$DB,$1F,$D7,$5A,$FF,$5C
	.dc.b	$F1,$DE,$27,$85,$B5,$DA,$F1,$3F,$0E,$BF,$D5,$FE,$BD,$F7,$CF,$C7
	.dc.b	$F1,$9F,$4E,$F8,$8D,$4E,$7E,$A9,$E0,$FA,$31,$FE,$B7,$18,$D3,$FA
	.dc.b	$DB,$63,$FA,$DB,$5F,$47,$C4,$79,$CF,$91,$BB,$E6,$71,$7F,$DD,$3E
	.dc.b	$2B,$D8,$6B,$E6,$6C,$7F,$78,$F9,$EF,$17,$3F,$F8,$01,$BC,$86,$98
	.dc.b	$F9,$CF,$8C,$7E,$07,$EB,$1F,$2F,$8C,$3F,$C0,$7C,$FF,$2C,$D7,$CB
	.dc.b	$58,$FF,$03,$EB,$D8,$8F,$F8,$4D,$F0,$E9,$F4,$CF,$C7,$7B,$4F,$96
	.dc.b	$C5,$7F,$84,$3E,$73,$EC,$1A,$F9,$7B,$1F,$E1,$4F,$9C,$F5,$91,$FF
	.dc.b	$10,$3E,$EF,$9C,$FC,$01,$DB,$E8,$6C,$FF,$87,$3D,$BF,$C2,$1E,$2F
	.dc.b	$98,$B1,$FE,$20,$F9,$DF,$8E,$47,$FC,$58,$FC,$AF,$9D,$FC,$61,$EA
	.dc.b	$EF,$B3,$FE,$28,$E3,$FC,$81,$EE,$F9,$6B,$1F,$E2,$CE,$3F,$A6,$47
	.dc.b	$FC,$70,$FD,$67,$F3,$07,$C6,$F9,$6A,$FF,$64,$F3,$CF,$F6,$6E,$B7
	.dc.b	$DC,$FC,$B8,$D1,$97,$1E,$02,$B8,$AF,$2D,$CE,$F9,$6F,$DC,$5F,$0B
	.dc.b	$DE,$3D,$B9,$BF,$1B,$26,$93,$2F,$F4,$44,$3C,$DE,$67,$E0,$3E,$56
	.dc.b	$CF,$FA,$AF,$EA,$9D,$5F,$B3,$6C,$7A,$B6,$79,$6F,$8F,$FF,$27,$CF
	.dc.b	$1B,$E1,$DD,$FB,$1B,$60,$8C,$36,$BF,$CA,$0D,$9F,$F6,$0E,$1D,$7B
	.dc.b	$3F,$93,$36,$BC,$D7,$8C,$55,$FB,$6E,$D9,$5F,$C6,$B2,$B8,$FC,$F8
	.dc.b	$FA,$5D,$BF,$30,$CF,$0F,$CC,$DD,$1F,$C0,$3B,$65,$7F,$24,$CA,$E1
	.dc.b	$CC,$71,$7E,$03,$B7,$EB,$9A,$CB,$A3,$FC,$77,$E3,$9F,$13,$4F,$E4
	.dc.b	$1F,$53,$FE,$70,$CB,$FD,$B9,$3A,$7F,$DB,$D1,$FD,$17,$1F,$F1,$9B
	.dc.b	$5F,$F5,$A5,$72,$F6,$BE,$3F,$C3,$F3,$9F,$37,$22,$31,$9F,$E4,$25
	.dc.b	$87,$5A,$D7,$A3,$FC,$83,$C2,$99,$FF,$44,$F0,$EE,$1F,$07,$7F,$CF
	.dc.b	$FC,$EF,$E5,$8C,$E7,$F2,$DF,$E5,$3E,$0C,$C3,$FA,$6F,$87,$70,$E5
	.dc.b	$CD,$FC,$EE,$7F,$1D,$8F,$EA,$BC,$4D,$65,$F5,$5F,$E8,$9F,$07,$2B
	.dc.b	$F2,$98,$7E,$53,$BB,$0F,$FA,$33,$83,$F5,$53,$FE,$9D,$6C,$7F,$BF
	.dc.b	$5A,$FE,$43,$C1,$FC,$C4,$3C,$4F,$EA,$47,$F5,$38,$B5,$E6,$F9,$6F
	.dc.b	$DA,$3F,$C0,$5B,$7F,$84,$F3,$FF,$82,$CE,$FC,$8F,$DD,$F9,$12,$30
	.dc.b	$7C,$0E,$04,$7E,$77,$A2,$4F,$EC,$74,$FD,$87,$AB,$8C,$BE,$EF,$86
	.dc.b	$F9,$BF,$D2,$19,$E3,$FE,$1B,$7F,$B7,$8F,$ED,$9C,$73,$1F,$93,$3E
	.dc.b	$06,$C7,$F7,$1E,$7F,$86,$48,$F9,$4F,$C6,$EF,$FE,$EC,$6E,$39,$BF
	.dc.b	$B8,$F7,$5F,$21,$4F,$DD,$FD,$43,$E3,$F1,$7E,$DA,$7C,$9F,$13,$3B
	.dc.b	$FE,$6A,$33,$3F,$47,$CC,$FF,$0E,$73,$FF,$1D,$8F,$A9,$E6,$E2,$7F
	.dc.b	$94,$FC,$74,$9E,$07,$CF,$7F,$C6,$5F,$0E,$FB,$27,$77,$F8,$26,$30
	.dc.b	$D8,$CD,$7F,$A9,$33,$95,$A8,$FF,$73,$FE,$33,$F4,$DF,$E5,$1E,$32
	.dc.b	$31,$5F,$C2,$6B,$83,$FC,$2E,$75,$CD,$FE,$41,$FC,$61,$F5,$94,$F7
	.dc.b	$F9,$F7,$2F,$C4,$78,$7F,$D1,$8C,$63,$FA,$C7,$E2,$F3,$FF,$84,$B8
	.dc.b	$8D,$25,$7F,$46,$F0,$FF,$A7,$9C,$1F,$CA,$3F,$3B,$A7,$91,$23,$E1
	.dc.b	$61,$FA,$18,$FE,$A7,$E2,$7F,$58,$38,$4F,$E7,$5F,$67,$3C,$8F,$D5
	.dc.b	$75,$DB,$FE,$07,$C9,$FD,$80,$C7,$6F,$F3,$57,$EF,$F8,$3E,$B2,$5F
	.dc.b	$8A,$BF,$99,$EB,$F9,$B7,$CB,$E2,$FF,$74,$F7,$FF,$B5,$9C,$AB,$F6
	.dc.b	$0F,$E2,$B8,$24,$7E,$F7,$A6,$DF,$F1,$FC,$3F,$83,$BF,$BF,$E0,$F9
	.dc.b	$84,$AF,$ED,$FF,$3D,$FD,$E0,$FC,$AC,$58,$FD,$D7,$C7,$BE,$6F,$FA
	.dc.b	$EB,$63,$37,$89,$FE,$C3,$F4,$FF,$EA,$E8,$FC,$87,$88,$8C,$47,$FB
	.dc.b	$DA,$31,$B7,$F9,$85,$B7,$FE,$02,$F8,$9F,$F0,$59,$FA,$8E,$5B,$1F
	.dc.b	$C4,$7E,$7F,$FC,$1E,$7E,$B3,$8E,$C7,$A3,$E7,$FC,$E7,$A3,$E1,$EC
	.dc.b	$7F,$15,$FA,$27,$A5,$E1,$6C,$7F,$2F,$6E,$BF,$DF,$9A,$FE,$AF,$5F
	.dc.b	$F1,$4D,$FF,$88,$0F,$DE,$F3,$5B,$FE,$7F,$F7,$BF,$C0,$AF,$F0,$0D
	.dc.b	$7F,$26,$D7,$F6,$CD,$43,$2F,$EA,$0F,$E2,$5A,$FE,$71,$AF,$EF,$DA
	.dc.b	$F2,$FA,$97,$AB,$FE,$02,$8F,$C9,$46,$32,$FC,$A4,$23,$0E,$A7,$83
	.dc.b	$FA,$A8,$F9,$3E,$27,$06,$C7,$4D,$53,$96,$C7,$83,$E2,$3F,$DF,$D4
	.dc.b	$69,$23,$C7,$93,$E5,$FC,$B0,$8C,$3A,$8C,$C6,$23,$F5,$7E,$9F,$C9
	.dc.b	$7F,$AC,$69,$FC,$8F,$BF,$FC,$54,$DD,$BF,$B2,$75,$DA,$FF,$24,$FC
	.dc.b	$C7,$F2,$E7,$FB,$93,$5F,$F2,$0B,$5F,$CF,$B5,$FC,$F7,$89,$FE,$36
	.dc.b	$7F,$B2,$35,$FF,$0F,$B7,$0F,$EC,$2F,$F9,$96,$3F,$B7,$75,$F8,$F8
	.dc.b	$BF,$CE,$ED,$CF,$F6,$93,$FD,$F7,$A7,$15,$23,$FB,$B3,$C7,$B9,$D3
	.dc.b	$ED,$EB,$76,$B0,$FF,$08,$37,$0F,$F0,$83,$5F,$F2,$FB,$5F,$DE,$E7
	.dc.b	$FE,$B3,$7D,$BE,$AF,$58,$C4,$7F,$8F,$3E,$A7,$FC,$75,$B7,$F9,$66
	.dc.b	$3D,$44,$3F,$C2,$DD,$7E,$73,$F2,$7A,$B5,$FF,$76,$CF,$F2,$98,$C4
	.dc.b	$EA,$73,$3F,$CA,$9F,$C0,$3F,$C8,$F8,$FE,$11,$F0,$F8,$3E,$67,$DE
	.dc.b	$7E,$1F,$1F,$F4,$BB,$79,$3F,$D4,$CD,$BF,$E6,$F9,$7C,$B3,$C0,$47
	.dc.b	$C9,$8F,$E9,$9F,$E0,$D2,$F6,$3B,$92,$3F,$D2,$DF,$0C,$F3,$3F,$D5
	.dc.b	$71,$86,$BF,$E7,$98,$D7,$C3,$7E,$55,$B1,$FE,$3D,$3D,$7F,$47,$1F
	.dc.b	$F2,$AE,$08,$FC,$57,$C8,$FA,$F3,$3F,$A7,$AF,$E3,$8D,$7F,$D8,$CD
	.dc.b	$7F,$D1,$2D,$85,$7F,$AB,$67,$BF,$B5,$F3,$0E,$F9,$C8,$FB,$17,$F0
	.dc.b	$DB,$1F,$7A,$31,$1F,$D8,$AB,$FC,$16,$FA,$BD,$0F,$AE,$7C,$1D,$3F
	.dc.b	$85,$C7,$C5,$6E,$FF,$ED,$6D,$3E,$67,$85,$B3,$E0,$EB,$FB,$D5,$F0
	.dc.b	$DF,$A4,$C3,$FC,$42,$D7,$FC,$3C,$DB,$FF,$70,$FA,$6F,$B5,$78,$FF
	.dc.b	$C1,$E3,$0F,$F2,$A6,$DF,$F7,$F6,$BF,$E1,$26,$BF,$86,$F9,$5F,$C9
	.dc.b	$91,$FF,$46,$EF,$F3,$5F,$F4,$D3,$70,$FF,$76,$B7,$97,$EB,$F1,$7D
	.dc.b	$EE,$0F,$DF,$6C,$F7,$A9,$31,$FF,$3B,$43,$E4,$76,$48,$CB,$FC,$1F
	.dc.b	$EA,$5F,$2D,$FC,$96,$5F,$52,$F9,$2C,$36,$3F,$D3,$ED,$7F,$DA,$47
	.dc.b	$0F,$93,$D9,$FC,$F2,$0D,$24,$7F,$54,$8F,$E8,$CF,$4F,$33,$B9,$E0
	.dc.b	$C7,$F8,$E7,$D4,$BE,$96,$92,$BF,$D2,$90,$FA,$5D,$92,$32,$F2,$EF
	.dc.b	$FE,$0B,$F4,$F2,$7F,$56,$FE,$AB,$9A,$47,$D6,$6B,$F5,$A3,$E1,$FC
	.dc.b	$4C,$8F,$86,$FC,$6F,$81,$91,$ED,$7E,$11,$1F,$81,$F0,$FF,$30,$8F
	.dc.b	$CC,$F8,$EF,$E3,$3F,$1F,$23,$F7,$EF,$EF,$88,$FE,$29,$AF,$E2,$BE
	.dc.b	$9F,$F9,$09,$1F,$E0,$EF,$F0,$6F,$83,$91,$F9,$26,$BF,$92,$F8,$CF
	.dc.b	$E9,$5F,$76,$F9,$2A,$49,$FF,$6F,$FD,$4B,$D9,$FC,$43,$63,$31,$83
	.dc.b	$3E,$24,$8E,$2D,$7F,$13,$3F,$DC,$7C,$3F,$E6,$F9,$7F,$3A,$24,$FE
	.dc.b	$A3,$E8,$FF,$8F,$91,$C1,$FA,$CF,$9F,$FE,$83,$A4,$8F,$E4,$EF,$F2
	.dc.b	$6F,$A5,$91,$FB,$26,$BF,$B2,$E6,$FC,$7F,$87,$23,$F1,$AF,$E9,$7F
	.dc.b	$04,$FE,$F9,$E3,$7F,$AC,$5F,$F0,$04,$FF,$A4,$9A,$E7,$E3,$7F,$C3
	.dc.b	$B1,$88,$FF,$20,$AF,$90,$E8,$48,$FC,$DF,$91,$FD,$6B,$59,$1F,$D0
	.dc.b	$1F,$E8,$DF,$4C,$F9,$AF,$EB,$FB,$FF,$6C,$F4,$8D,$25,$FF,$6D,$86
	.dc.b	$FF,$D9,$89,$1F,$D6,$2F,$F0,$0F,$F4,$ED,$FF,$B5,$91,$D7,$CD,$7F
	.dc.b	$6E,$F0,$FC,$04,$8E,$CF,$8F,$C2,$FE,$E6,$48,$ED,$F0,$FF,$DD,$BC
	.dc.b	$A9,$1F,$D7,$1F,$6F,$CB,$FF,$7A,$FA,$4F,$61,$27,$D4,$FB,$FE,$6F
	.dc.b	$FB,$F9,$23,$D5,$CD,$FE,$02,$F5,$64,$7F,$68,$7F,$B7,$7D,$AF,$F8
	.dc.b	$6B,$B1,$5F,$E7,$4F,$6B,$FB,$E7,$8B,$FE,$16,$F6,$8D,$26,$3F,$E1
	.dc.b	$78,$78,$FF,$E3,$B6,$C7,$F9,$FB,$0E,$AA,$FF,$C1,$15,$E2,$EB,$E1
	.dc.b	$E4,$4A,$E1,$97,$8F,$14,$FA,$3B,$BB,$79,$3E,$52,$5F,$8D,$F4,$7A
	.dc.b	$9E,$32,$47,$F8,$53,$67,$CA,$F9,$FC,$F9,$3E,$03,$CD,$97,$F4,$97
	.dc.b	$87,$CA,$FB,$7D,$1F,$88,$48,$F8,$4F,$C7,$F4,$FE,$11,$23,$FA,$FF
	.dc.b	$5F,$C5,$F5,$24,$77,$E3,$FF,$15,$47,$F8,$6B,$B3,$FB,$D7,$D1,$FF
	.dc.b	$8B,$EB,$FD,$4F,$D0,$8C,$47,$FA,$49,$B1,$9B,$FF,$45,$FD,$47,$F4
	.dc.b	$3A,$F6,$07,$0F,$F5,$86,$F4,$7D,$6A,$FD,$2F,$A9,$FE,$75,$FA,$9F
	.dc.b	$F5,$BD,$F9,$DA,$AB,$EE,$E0,$9C,$C6,$1F,$F5,$6B,$63,$FC,$C3,$1E
	.dc.b	$A7,$81,$83,$FC,$CB,$D3,$FE,$5E,$9E,$AD,$E9,$FC,$B3,$6D,$EF,$3E
	.dc.b	$D5,$E7,$FF,$92,$DB,$3F,$E5,$46,$BF,$E5,$16,$E1,$EA,$F9,$DF,$F2
	.dc.b	$64,$FD,$5F,$A8,$7A,$3F,$E9,$76,$C6,$63,$FD,$35,$F1,$5F,$D0,$63
	.dc.b	$E2,$69,$FA,$B8,$C6,$1F,$19,$B1,$9F,$57,$8F,$C5,$FF,$84,$7C,$CF
	.dc.b	$E3,$37,$FE,$E6,$FA,$AF,$F4,$B4,$7F,$18,$C7,$15,$F5,$23,$10,$9D
	.dc.b	$FA,$1F,$56,$31,$5B,$8C,$4F,$4F,$3F,$EE,$B6,$BA,$9F,$EF,$6B,$F1
	.dc.b	$BC,$8C,$57,$5F,$27,$84,$1A,$78,$B1,$FF,$7F,$B6,$39,$0A,$F4,$DF
	.dc.b	$23,$44,$C7,$DD,$8F,$99,$B5,$E6,$39,$63,$EC,$8F,$76,$3F,$F0,$03
	.dc.b	$0F,$4D,$7E,$61,$AE,$8F,$CD,$9E,$3F,$49,$F0,$5C,$5B,$1A,$94,$62
	.dc.b	$3A,$35,$E7,$F5,$5F,$EF,$C4,$D2,$3F,$F0,$01,$F8,$B8,$7E,$B3,$9F
	.dc.b	$F9,$3E,$BF,$ED,$3A,$E2,$7C,$D9,$D6,$FC,$23,$F6,$77,$5F,$E7,$51
	.dc.b	$51,$CA,$BF,$DB,$D7,$DA,$FC,$F3,$D4,$FD,$FC,$65,$EA,$64,$EA,$53
	.dc.b	$AD,$F7,$8E,$11,$FC,$22,$31,$7E,$38,$F1,$57,$A9,$E4,$60,$F0,$C7
	.dc.b	$B6,$9C,$4F,$37,$FA,$5D,$79,$A3,$15,$FE,$E9,$6C,$7C,$B6,$BB,$C7
	.dc.b	$8B,$FD,$7E,$3F,$31,$3C,$67,$33,$FD,$5E,$72,$FD,$5E,$8F,$97,$5F
	.dc.b	$EF,$76,$C6,$4E,$38,$7F,$87,$5E,$7D,$0C,$BF,$C2,$B7,$F9,$08,$C7
	.dc.b	$06,$A7,$35,$FC,$BA,$73,$5B,$FA,$87,$A5,$D1,$B1,$98,$C1,$E2,$F1
	.dc.b	$9F,$51,$49,$3F,$DE,$5F,$7F,$7B,$EA,$9E,$46,$1B,$5F,$A6,$9D,$4E
	.dc.b	$4F,$8F,$9B,$E4,$5F,$1D,$86,$DF,$EB,$A7,$53,$93,$CD,$F5,$32,$7F
	.dc.b	$C1,$8D,$9F,$D2,$66,$FA,$8F,$61,$2B,$FC,$24,$FC,$3F,$62,$47,$C7
	.dc.b	$6D,$7C,$BD,$DE,$BF,$C9,$25,$7C,$D7,$E6,$7D,$3C,$8F,$A0,$DA,$E5
	.dc.b	$1F,$44,$87,$D5,$3E,$FF,$B4,$F9,$4A,$48,$FC,$04,$65,$BF,$4D,$C0
	.dc.b	$8E,$0D,$F4,$F4,$23,$83,$9F,$B1,$E0,$23,$83,$BF,$CA,$F9,$08,$E0
	.dc.b	$F1,$FA,$DE,$82,$38,$3C,$FC,$9E,$C2,$38,$3D,$7F,$49,$F0,$08,$E0
	.dc.b	$F7,$FD,$1F,$C4,$23,$83,$E1,$FD,$27,$C8,$23,$83,$E3,$FD,$27,$CC
	.dc.b	$23,$83,$E5,$FD,$1F,$D0,$23,$83,$E7,$FD,$1F,$D4,$23,$83,$FA,$0F
	.dc.b	$53,$E1,$BF,$57,$C7,$FE,$1E,$8C,$77,$FF,$04,$6D,$70,$AD,$D7,$F8
	.dc.b	$93,$EA,$7F,$D2,$93,$F9,$F6,$C7,$F4,$E6,$BB,$F8,$7F,$D4,$C7,$E7
	.dc.b	$67,$87,$5E,$08,$FF,$9B,$E7,$F7,$84,$70,$7F,$9C,$FE,$5F,$F8,$01
	.dc.b	$1C,$1F,$E7,$7F,$8D,$FE,$08,$47,$07,$F9,$EF,$3F,$E1,$04,$70,$7E
	.dc.b	$E7,$EA,$5E,$17,$F0,$06,$C6,$63,$07,$F0,$4F,$91,$FF,$44,$C7,$F0
	.dc.b	$16,$E9,$FD,$C1,$AF,$F9,$31,$AF,$F9,$42,$BF,$98,$BF,$E5,$06,$B9
	.dc.b	$F2,$FC,$47,$F1,$FF,$A4,$97,$FE,$95,$6C,$FF,$0A,$F0,$BF,$92,$12
	.dc.b	$7F,$D3,$4F,$F2,$7F,$06,$47,$FA,$75,$B3,$FC,$4B,$C1,$FE,$58,$49
	.dc.b	$FF,$51,$3F,$C7,$F9,$24,$7F,$1F,$E4,$46,$2B,$F1,$FC,$BF,$C9,$49
	.dc.b	$5B,$7F,$94,$73,$48,$E4,$D9,$FE,$47,$1F,$CF,$48,$7F,$AC,$CF,$EA
	.dc.b	$3F,$37,$FC,$36,$3F,$C1,$8D,$7F,$BE,$B5,$FF,$60,$35,$FE,$75,$CE
	.dc.b	$F8,$7F,$D0,$36,$BF,$41,$18,$33,$F1,$12,$3F,$48,$DA,$FD,$6F,$67
	.dc.b	$F4,$32,$4F,$C5,$7F,$A3,$76,$C8,$F9,$0D,$9F,$DA,$F7,$7F,$4B,$24
	.dc.b	$FC,$D7,$F6,$9F,$53,$FE,$D2,$8F,$DD,$B6,$7F,$74,$D7,$FD,$37,$BA
	.dc.b	$31,$1F,$E2,$68,$7B,$9F,$E8,$0A,$FE,$0A,$D9,$FE,$08,$D7,$9B,$5F
	.dc.b	$97,$E2,$7D,$22,$4F,$F9,$3B,$D0,$6E,$9F,$AF,$E2,$CA,$D4,$A3,$17
	.dc.b	$FE,$56,$FB,$3F,$3B,$FE,$3C,$6A,$69,$AF,$5D,$F5,$5F,$92,$C3,$FC
	.dc.b	$BF,$18,$E6,$FF,$0F,$B6,$33,$3B,$BF,$C4,$F9,$1F,$8C,$24,$F9,$9F
	.dc.b	$F0,$EF,$CF,$3E,$4E,$92,$7F,$C3,$D1,$96,$FC,$C7,$02,$38,$3E,$0F
	.dc.b	$97,$F9,$A2,$38,$39,$F8,$5E,$02,$38,$3E,$4F,$2F,$E8,$88,$E0,$F9
	.dc.b	$BC,$DF,$A6,$23,$83,$CF,$F0,$7E,$C2,$38,$3D,$7E,$DF,$C0,$23,$83
	.dc.b	$EC,$F4,$FE,$C8,$8E,$0F,$BB,$E6,$FE,$D8,$8E,$0F,$C0,$F9,$BF,$BA
	.dc.b	$23,$83,$F0,$BD,$FF,$BE,$23,$83,$FA,$DF,$C0,$FF,$95,$38,$BF,$3E
	.dc.b	$D7,$F3,$CD,$7F,$C7,$0D,$EC,$2F,$9F,$FE,$C0,$DA,$EB,$18,$3F,$87
	.dc.b	$F8,$52,$3F,$32,$DB,$FE,$23,$D1,$FC,$50,$93,$F9,$E7,$AF,$AB,$23
	.dc.b	$D2,$D9,$FE,$E5,$83,$E8,$BC,$64,$AF,$7B,$FC,$83,$E6,$24,$7E,$A9
	.dc.b	$B5,$FE,$BE,$FA,$5F,$DB,$3F,$EA,$46,$A6,$98,$FA,$27,$0F,$F9,$79
	.dc.b	$B7,$FE,$5B,$6D,$7F,$B2,$9A,$FF,$AC,$0F,$F5,$8F,$D3,$BC,$B7,$89
	.dc.b	$FE,$63,$8C,$36,$33,$3F,$ED,$49,$CB,$7C,$CF,$EE,$2D,$F3,$F9,$46
	.dc.b	$23,$C0,$25,$19,$8F,$24,$3E,$4F,$FC,$E9,$1B,$6C,$66,$3C,$D1,$AF
	.dc.b	$9A,$FF,$6E,$12,$30,$7F,$9F,$71,$FE,$3B,$F1,$37,$C8,$34,$93,$FE
	.dc.b	$30,$F0,$3F,$D4,$CF,$F8,$BD,$AF,$F3,$66,$BF,$CD,$5A,$FF,$A3,$DF
	.dc.b	$E5,$4F,$3F,$27,$FB,$84,$62,$3A,$47,$8F,$E0,$64,$7F,$2E,$6B,$FD
	.dc.b	$9B,$93,$D0,$47,$F9,$9B,$D3,$DB,$91,$E9,$6B,$FE,$47,$FA,$4F,$F5
	.dc.b	$01,$38,$FF,$C0,$4F,$FA,$9B,$4F,$E8,$3D,$1F,$14,$95,$FD,$01,$F9
	.dc.b	$1C,$92,$3F,$A1,$B5,$FE,$F7,$D7,$F3,$48,$FF,$48,$7E,$85,$48,$FE
	.dc.b	$96,$D7,$FD,$99,$E3,$3E,$27,$FD,$69,$F0,$68,$C4,$FF,$38,$9C,$C6
	.dc.b	$23,$8F,$BD,$82,$3F,$CF,$9F,$C0,$30,$FE,$B6,$53,$F2,$B6,$FA,$37
	.dc.b	$D3,$FE,$25,$B3,$FD,$9D,$AF,$06,$BF,$AD,$C7,$D1,$F2,$7E,$62,$47
	.dc.b	$F5,$07,$A7,$C1,$C8,$E8,$D7,$E1,$FC,$4F,$B8,$8F,$F5,$87,$E0,$F6
	.dc.b	$C8,$FE,$B6,$D7,$C9,$F4,$9E,$92,$3E,$97,$FD,$C1,$E9,$BE,$95,$E2
	.dc.b	$61,$B5,$FD,$D1,$AF,$EE,$A7,$84,$FE,$DB,$EC,$1E,$45,$25,$7F,$BB
	.dc.b	$63,$2F,$C9,$F7,$3E,$61,$1C,$1F,$37,$D4,$FA,$04,$70,$7F,$A9,$32
	.dc.b	$FE,$04,$FF,$A8,$DA,$FF,$8A,$9A,$FF,$8A,$5A,$BC,$B7,$D3,$3E,$83
	.dc.b	$FC,$C8,$D9,$FF,$58,$B5,$FF,$58,$35,$FE,$37,$F4,$CF,$9C,$FE,$2E
	.dc.b	$D8,$FE,$3E,$03,$97,$2D,$79,$FC,$EF,$22,$3C,$8E,$2F,$6F,$26,$46
	.dc.b	$67,$DA,$F6,$8F,$F3,$F7,$D3,$39,$A9,$27,$FD,$9B,$19,$6F,$5F,$02
	.dc.b	$38,$3F,$41,$F3,$5F,$A3,$23,$83,$F4,$9E,$DF,$E9,$C8,$E0,$F3,$7D
	.dc.b	$07,$C6,$23,$83,$D5,$F4,$3F,$28,$8E,$0F,$77,$D1,$7C,$E2,$38,$3F
	.dc.b	$69,$EF,$7E,$DC,$8E,$0F,$DC,$7D,$37,$EE,$C8,$E0,$FD,$E7,$BB,$FB
	.dc.b	$F2,$38,$3F,$80,$7C,$E7,$F0,$32,$38,$3E,$B7,$D2,$FF,$4C,$6F,$A5
	.dc.b	$FD,$6A,$31,$1F,$84,$6B,$FD,$91,$FD,$6B,$7B,$78,$11,$EB,$ED,$F3
	.dc.b	$23,$83,$F8,$A7,$BF,$FC,$5C,$8E,$0F,$E3,$1F,$31,$FC,$6C,$8E,$0F
	.dc.b	$E3,$9E,$5F,$F1,$F2,$38,$3F,$90,$71,$FF,$23,$23,$83,$F2,$DE,$BB
	.dc.b	$E0,$FF,$BA,$12,$32,$FF,$77,$F8,$69,$1C,$1A,$FF,$81,$BB,$DD,$0F
	.dc.b	$0B,$F3,$64,$9F,$CC,$BF,$D1,$BE,$6D,$F2,$74,$93,$FB,$2F,$31,$3A
	.dc.b	$8C,$2F,$93,$E4,$48,$CB,$D7,$E4,$E4,$70,$6B,$FE,$1C,$AE,$BF,$27
	.dc.b	$E3,$24,$76,$7C,$DF,$01,$23,$C2,$D7,$96,$1F,$E3,$37,$CB,$F0,$3E
	.dc.b	$D2,$3E,$8F,$91,$F8,$04,$70,$7B,$3E,$47,$E2,$11,$C1,$F0,$3D,$CF
	.dc.b	$90,$47,$07,$C7,$FA,$47,$BD,$FD,$C9,$B1,$98,$C1,$9F,$81,$91,$C5
	.dc.b	$AF,$C5,$9F,$F1,$2E,$A6,$93,$93,$B7,$2C,$AF,$13,$5E,$5F,$05,$E3
	.dc.b	$23,$E7,$7C,$DF,$19,$23,$D4,$D7,$C5,$F1,$9E,$B2,$3E,$F7,$DD,$F1
	.dc.b	$92,$3E,$13,$5F,$EF,$DF,$49,$FE,$03,$23,$FE,$03,$7F,$C0,$DE,$6C
	.dc.b	$8F,$F0,$33,$5F,$E8,$7F,$45,$FC,$BD,$B1,$FE,$5E,$8C,$47,$F9,$99
	.dc.b	$AF,$F4,$9F,$6F,$FC,$49,$E7,$FF,$43,$6C,$7F,$4E,$6B,$FE,$6E,$6B
	.dc.b	$FE,$7F,$F8,$07,$D1,$7F,$9E,$A3,$FC,$3E,$D7,$4D,$7F,$AC,$47,$03
	.dc.b	$FC,$67,$F8,$1F,$92,$1C,$48,$C3,$72,$D9,$FE,$3A,$FA,$37,$9B,$49
	.dc.b	$7F,$B5,$8C,$BF,$DA,$30,$7C,$A3,$DB,$FF,$3D,$36,$74,$D7,$FC,$72
	.dc.b	$D7,$E0,$7C,$87,$F2,$E7,$D6,$D7,$FD,$3E,$DA,$FF,$66,$F9,$FF,$E8
	.dc.b	$78,$7C,$65,$25,$79,$63,$2F,$F7,$5F,$01,$E6,$3E,$8B,$FC,$01,$F3
	.dc.b	$3A,$AF,$E2,$3E,$77,$FA,$C6,$31,$7F,$E5,$96,$DF,$1A,$46,$2F,$9E
	.dc.b	$3F,$EB,$9B,$E0,$DC,$3A,$39,$C3,$95,$F1,$C3,$9C,$ED,$E6,$7F,$85
	.dc.b	$6F,$9B,$6B,$8C,$F3,$C3,$73,$EA,$2D,$70,$7F,$96,$D7,$07,$94,$EB
	.dc.b	$0F,$2C,$77,$6B,$C2,$39,$CF,$C1,$8F,$F6,$3D,$7C,$C9,$F9,$D3,$99
	.dc.b	$F2,$1F,$E2,$E9,$79,$EF,$83,$FA,$8D,$9D,$35,$EA,$D7,$FC,$6F,$F4
	.dc.b	$5F,$ED,$04,$D2,$3F,$ED,$47,$FC,$79,$F4,$7F,$76,$3F,$DD,$3E,$2E
	.dc.b	$34,$E7,$09,$D7,$02,$B5,$3F,$EE,$4B,$AF,$EC,$22,$BF,$E4,$98,$B3
	.dc.b	$C6,$3A,$5F,$6D,$BA,$DF,$D6,$9E,$07,$E0,$0F,$E0,$4E,$5F,$69,$C8
	.dc.b	$C9,$4F,$F2,$C1,$AC,$3A,$B9,$3A,$B8,$3E,$13,$BA,$FF,$55,$CF,$BA
	.dc.b	$DF,$47,$FE,$6F,$8F,$F3,$26,$DE,$B9,$CF,$C1,$FB,$63,$15,$FE,$59
	.dc.b	$AD,$9F,$C5,$1E,$58,$7E,$BE,$3F,$8B,$F1,$BE,$13,$94,$61,$B2,$C7
	.dc.b	$BC,$F4,$8C,$57,$18,$C4,$BC,$5F,$BB,$5D,$CF,$E4,$39,$BD,$9E,$B5
	.dc.b	$B3,$BF,$5F,$F0,$A8,$B3,$BC,$3F,$C2,$8D,$BF,$F0,$53,$5F,$F6,$53
	.dc.b	$5F,$F4,$55,$7F,$14,$E4,$FC,$14,$69,$AE,$A7,$F8,$9D,$76,$9C,$CF
	.dc.b	$C1,$F8,$AF,$93,$B7,$D1,$6D,$71,$AF,$5C,$62,$BE,$47,$DB,$7F,$B9
	.dc.b	$2B,$E2,$BF,$CD,$5B,$F1,$9F,$0A,$31,$3F,$1F,$2F,$E5,$7E,$27,$89
	.dc.b	$B7,$F0,$A3,$19,$9F,$3E,$BF,$EF,$3F,$8B,$F1,$9D,$7D,$6B,$7F,$67
	.dc.b	$0F,$E6,$FC,$9E,$8E,$6F,$9E,$D9,$D6,$4C,$3B,$92,$3E,$9B,$E2,$AF
	.dc.b	$AE,$7A,$FD,$FB,$3D,$FC,$77,$C0,$7C,$6F,$1F,$EF,$C6,$27,$F1,$3C
	.dc.b	$49,$CF,$C1,$7C,$1A,$34,$95,$FE,$57,$87,$CE,$F4,$3E,$67,$75,$9F
	.dc.b	$CA,$FE,$0B,$DF,$93,$F6,$E7,$E7,$5C,$3F,$E0,$16,$BF,$DF,$1A,$FF
	.dc.b	$7A,$B7,$D3,$F9,$49,$5E,$37,$E9,$F3,$7D,$58,$FE,$EC,$DA,$F0,$F2
	.dc.b	$A3,$11,$FE,$11,$6B,$E1,$F8,$5F,$BB,$89,$A4,$BF,$85,$ED,$FC,$72
	.dc.b	$38,$3C,$3B,$79,$08,$E0,$F9,$5E,$C7,$CF,$23,$83,$E7,$1F,$4C,$86
	.dc.b	$0F,$F4,$47,$C4,$3D,$7F,$CC,$36,$33,$18,$33,$EB,$48,$E2,$DC,$FF
	.dc.b	$D0,$D4,$F5,$79,$92,$BB,$BD,$BE,$62,$47,$89,$B5,$FE,$A5,$FD,$91
	.dc.b	$F1,$7F,$C7,$21,$F4,$6D,$BF,$8F,$FC,$22,$F7,$8F,$F1,$FB,$DD,$71
	.dc.b	$BF,$F5,$14,$FE,$2A,$47,$4F,$82,$FE,$12,$D8,$D3,$4F,$27,$C6,$78
	.dc.b	$30,$3D,$6B,$0F,$F2,$4D,$5F,$3B,$4F,$E5,$3F,$5A,$FA,$6C,$65,$F9
	.dc.b	$EE,$DF,$E5,$BC,$FF,$A8,$8F,$E0,$18,$72,$74,$7F,$38,$BF,$F6,$1B
	.dc.b	$AF,$93,$EB,$F1,$1E,$16,$47,$EA,$BE,$8E,$EF,$FC,$F7,$B1,$92,$7D
	.dc.b	$9E,$07,$F4,$AF,$35,$F5,$4E,$CC,$E5,$FD,$2E,$73,$1B,$9D,$4F,$CD
	.dc.b	$9C,$B6,$33,$97,$EE,$A7,$2F,$FB,$8F,$F6,$CF,$E6,$0F,$F5,$1E,$84
	.dc.b	$62,$7F,$B4,$57,$27,$FC,$0D,$3A,$F3,$BE,$87,$C8,$4C,$7E,$37,$3D
	.dc.b	$CB,$E0,$6D,$FC,$42,$7F,$86,$EB,$9C,$BF,$BB,$4E,$6B,$E3,$71,$3E
	.dc.b	$33,$EE,$79,$9F,$69,$B8,$FD,$68,$FB,$3D,$38,$BE,$EF,$E6,$7C,$AF
	.dc.b	$9D,$5F,$95,$DF,$FB,$7E,$7F,$D7,$A7,$37,$FD,$C4,$4E,$AB,$FC,$2E
	.dc.b	$BE,$6B,$F2,$ED,$8F,$CB,$B6,$7F,$3C,$D7,$F9,$4D,$7F,$13,$9F,$D3
	.dc.b	$36,$3C,$13,$C6,$FC,$9E,$2F,$E6,$39,$3F,$39,$F3,$97,$1F,$CA,$B8
	.dc.b	$7F,$9B,$F8,$37,$1F,$9E,$FE,$17,$EA,$A4,$FF,$84,$F2,$FE,$83,$F0
	.dc.b	$98,$BF,$C0,$71,$FF,$2F,$D3,$EE,$B6,$FF,$C5,$7C,$EC,$7F,$80,$CE
	.dc.b	$6F,$E3,$BF,$E0,$FB,$E3,$F0,$9A,$F5,$F6,$FF,$0B,$CD,$F3,$FF,$28
	.dc.b	$99,$FC,$A7,$FC,$17,$1F,$D5,$BE,$5B,$B9,$C6,$2C,$FE,$DB,$8F,$F8
	.dc.b	$01,$FB,$E6,$9F,$D6,$BE,$63,$F7,$A6,$34,$FC,$9B,$FC,$1D,$91,$FE
	.dc.b	$34,$FC,$11,$F2,$18,$9F,$CC,$FA,$3F,$C1,$8C,$6D,$F9,$87,$FC,$67
	.dc.b	$F8,$4B,$E7,$BF,$C2,$4D,$C7,$F9,$07,$B6,$8C,$61,$FD,$B2,$31,$5F
	.dc.b	$D6,$38,$FF,$89,$9D,$1B,$7E,$5F,$47,$BF,$B9,$FE,$44,$D5,$18,$9F
	.dc.b	$F0,$C6,$4A,$F1,$92,$3C,$71,$97,$F9,$A7,$C2,$BF,$1C,$FE,$07,$A7
	.dc.b	$F9,$DA,$31,$E3,$7F,$8F,$9B,$5C,$2B,$75,$FE,$01,$F3,$BF,$95,$9F
	.dc.b	$E6,$4A,$DC,$7F,$02,$8F,$8F,$F3,$9F,$18,$97,$E4,$7F,$BC,$FD,$5F
	.dc.b	$F3,$43,$71,$FE,$14,$7E,$07,$C9,$48,$FB,$1E,$B5,$C3,$FC,$89,$A3
	.dc.b	$A7,$EA,$99,$E5,$FF,$4F,$72,$7D,$38,$FB,$3D,$9F,$60,$98,$FC,$67
	.dc.b	$FC,$17,$E8,$7F,$46,$37,$1F,$E2,$67,$E9,$7D,$3F,$D3,$F9,$DB,$3F
	.dc.b	$8B,$7E,$B7,$9D,$FE,$50,$E7,$34,$93,$FD,$97,$D1,$FE,$8B,$E7,$FF
	.dc.b	$54,$38,$C7,$F8,$D1,$FE,$E3,$F8,$DB,$E6,$AF,$CD,$E1,$B5,$FB,$FF
	.dc.b	$99,$FF,$02,$3F,$C5,$1B,$C7,$FE,$3C,$F1,$51,$88,$FE,$3E,$D7,$FB
	.dc.b	$BE,$4D,$1E,$87,$E4,$4F,$C9,$7C,$9D,$C7,$FC,$93,$EE,$FF,$6C,$3F
	.dc.b	$2D,$E4,$58,$FD,$77,$9D,$EA,$F4,$FF,$34,$D8,$F4,$69,$FC,$08,$F2
	.dc.b	$BE,$97,$07,$F8,$FA,$7F,$BC,$9F,$A3,$CE,$CF,$F9,$6B,$7F,$F2,$19
	.dc.b	$1F,$9D,$EA,$A3,$FC,$8F,$C1,$27,$2D,$DF,$FC,$31,$F4,$FF,$3F,$0F
	.dc.b	$F0,$AB,$63,$33,$FE,$C1,$6C,$EA,$FF,$CE,$D1,$C9,$AF,$F9,$B2,$BC
	.dc.b	$5F,$03,$B3,$CB,$E9,$FF,$88,$AF,$FC,$0E,$DC,$BF,$C0,$97,$A3,$D7
	.dc.b	$F3,$99,$23,$83,$DF,$9F,$FB,$C3,$BB,$FB,$7B,$4F,$F3,$C1,$63,$FD
	.dc.b	$B5,$EE,$FF,$B7,$DF,$E9,$1C,$EE,$CF,$EB,$ED,$8F,$EB,$ED,$7F,$18
	.dc.b	$D7,$FD,$FF,$9F,$FA,$53,$E6,$BB,$74,$BF,$7E,$FF,$1B,$D7,$F5,$56
	.dc.b	$BF,$E2,$7C,$55,$E5,$BF,$F7,$47,$A1,$7E,$3F,$F9,$BD,$FF,$BB,$1A
	.dc.b	$F1,$F0,$B8,$60,$F6,$A9,$27,$FD,$E7,$0F,$57,$CB,$AB,$C9,$F7,$92
	.dc.b	$BD,$EF,$F8,$03,$57,$91,$EC,$5E,$4D,$31,$1E,$7F,$85,$97,$E7,$7C
	.dc.b	$BE,$44,$8F,$E0,$59,$FF,$C0,$16,$C7,$E0,$E3,$11,$FE,$A4,$5F,$1F
	.dc.b	$EE,$F8,$99,$3E,$C8,$F8,$DC,$12,$3E,$33,$F7,$FC,$BF,$EA,$93,$7C
	.dc.b	$CE,$75,$FD,$1A,$3F,$CE,$73,$A8,$C1,$FD,$2B,$E8,$38,$11,$F9,$7F
	.dc.b	$21,$F4,$08,$E0,$F9,$FF,$21,$F6,$7D,$69,$1F,$65,$F1,$7B,$D2,$3F
	.dc.b	$AB,$BE,$2C,$8D,$3F,$4D,$9F,$F1,$A9,$FC,$17,$8F,$F8,$53,$F3,$9E
	.dc.b	$A3,$F4,$0F,$F5,$84,$FF,$A4,$BC,$17,$83,$FE,$94,$AD,$72,$7F,$A4
	.dc.b	$9B,$5F,$D9,$AF,$8B,$FE,$E4,$8C,$BF,$DE,$F3,$7C,$87,$9C,$91,$FD
	.dc.b	$C1,$FB,$5E,$37,$CF,$64,$7E,$4A,$3F,$39,$F1,$92,$3F,$38,$FE,$D5
	.dc.b	$FB,$FE,$95,$E5,$3F,$EA,$88,$C9,$FE,$AA,$8C,$B6,$3E,$3F,$CF,$7E
	.dc.b	$A4,$91,$F0,$F5,$FF,$5A,$F3,$7F,$9C,$7E,$13,$F0,$DE,$77,$FB,$AE
	.dc.b	$B4,$27,$4E,$1A,$FF,$A7,$FF,$07,$75,$FE,$DF,$B6,$4F,$ED,$DE,$BF
	.dc.b	$13,$23,$96,$1F,$EA,$56,$C7,$F0,$1B,$C4,$7F,$A5,$FF,$7C,$76,$7F
	.dc.b	$0C,$6C,$7F,$97,$DA,$FF,$9F,$7E,$45,$DB,$F2,$DB,$1F,$ED,$06,$BF
	.dc.b	$ED,$76,$BF,$ED,$83,$FD,$B2,$E1,$FF,$6D,$F9,$6A,$7C,$DF,$8F,$C0
	.dc.b	$95,$E3,$8F,$37,$93,$23,$CC,$FF,$31,$ED,$7D,$5F,$F1,$39,$FF,$60
	.dc.b	$35,$47,$E6,$1B,$19,$8C,$1C,$BC,$9F,$89,$E5,$7D,$AF,$8B,$FE,$51
	.dc.b	$E4,$4C,$3E,$7C,$7C,$DF,$8B,$D4,$60,$FB,$FE,$73,$8D,$3F,$93,$FB
	.dc.b	$B7,$8F,$49,$7F,$86,$1C,$6E,$1F,$A0,$7E,$5D,$C4,$7D,$18,$C1,$F9
	.dc.b	$BF,$98,$FF,$36,$47,$39,$F9,$AD,$E2,$FD,$24,$EA,$3E,$2F,$6F,$E6
	.dc.b	$BA,$5F,$15,$FD,$1F,$E4,$A5,$7E,$3A,$3F,$19,$F1,$5F,$9E,$64,$7C
	.dc.b	$0D,$FF,$D4,$0F,$F4,$DF,$84,$FE,$A0,$47,$F2,$F5,$FA,$62,$38,$39
	.dc.b	$7C,$37,$F6,$0F,$90,$FC,$9F,$47,$FB,$A9,$BA,$7E,$29,$B5,$F7,$BE
	.dc.b	$2F,$FD,$F1,$3F,$C5,$DF,$DF,$57,$BE,$7F,$DF,$91,$83,$F8,$17,$6B
	.dc.b	$F6,$7E,$33,$CC,$FC,$74,$E2,$75,$1F,$BE,$F5,$BF,$B6,$7C,$94,$8F
	.dc.b	$D9,$C7,$EC,$7D,$BD,$46,$0F,$8C,$FF,$73,$F8,$99,$1F,$DC,$DF,$E1
	.dc.b	$3E,$5C,$8F,$DE,$B5,$FE,$0D,$F0,$BF,$E0,$4C,$FF,$DD,$D1,$FE,$D7
	.dc.b	$6B,$F7,$0F,$F7,$AB,$8D,$3F,$45,$0F,$A6,$FE,$6B,$18,$7F,$9B,$C7
	.dc.b	$EE,$4D,$7A,$BF,$C9,$E3,$31,$FB,$D6,$BD,$BE,$B3,$F9,$84,$7F,$83
	.dc.b	$2B,$A7,$BD,$E7,$9F,$E4,$BD,$3F,$D9,$81,$49,$7F,$EC,$27,$FC,$F0
	.dc.b	$DF,$85,$E0,$47,$5F,$03,$CC,$8E,$0F,$F6,$5F,$E3,$BF,$E1,$E7,$F8
	.dc.b	$A7,$A2,$8C,$47,$F3,$A6,$BF,$CD,$DB,$1B,$F0,$BF,$BE,$C7,$EE,$A7
	.dc.b	$AC,$60,$CF,$B7,$23,$C2,$D7,$E6,$78,$CF,$B9,$E5,$18,$6B,$FE,$42
	.dc.b	$6B,$FE,$8C,$6B,$E5,$F2,$7C,$64,$7D,$EF,$F7,$EF,$8D,$91,$FE,$A7
	.dc.b	$C3,$FD,$56,$D8,$FF,$34,$57,$F8,$D2,$3F,$D5,$8D,$7C,$9E,$9E,$8F
	.dc.b	$83,$F1,$72,$3E,$6B,$5F,$6D,$A3,$FC,$1E,$48,$F1,$3F,$2B,$83,$FC
	.dc.b	$59,$A1,$A7,$F9,$67,$0F,$3D,$7E,$6F,$5F,$B1,$A7,$D2,$24,$7D,$E7
	.dc.b	$FC,$63,$F4,$AF,$91,$FF,$30,$7C,$74,$AF,$EE,$13,$96,$F8,$1F,$E3
	.dc.b	$D8,$FE,$49,$CD,$FE,$E0,$6F,$83,$F4,$7E,$37,$5F,$4B,$D8,$98,$7E
	.dc.b	$9C,$FE,$FD,$48,$F1,$FD,$07,$F9,$C7,$E4,$25,$78,$A3,$FC,$E9,$72
	.dc.b	$3F,$CE,$8F,$C1,$D7,$C9,$EA,$7B,$09,$3D,$F7,$65,$FE,$77,$8F,$F7
	.dc.b	$27,$9D,$ED,$7F,$81,$5F,$E3,$DF,$98,$7F,$A1,$63,$5A,$FF,$65,$9F
	.dc.b	$EA,$BC,$5F,$E8,$72,$57,$C8,$F5,$7E,$31,$F3,$88,$F8,$F0,$FE,$E0
	.dc.b	$FE,$BB,$E9,$9F,$43,$FE,$64,$6C,$7A,$5A,$FF,$0F,$CD,$18,$8F,$F5
	.dc.b	$1F,$D4,$F9,$B8,$BF,$C5,$F3,$C3,$4F,$EF,$97,$C3,$0F,$F5,$05,$EF
	.dc.b	$06,$8F,$A3,$7C,$67,$FA,$95,$B3,$FD,$E1,$AF,$C8,$FA,$77,$27,$FA
	.dc.b	$AD,$B1,$FD,$F9,$AE,$E3,$5C,$1F,$E5,$1F,$A7,$FF,$47,$EA,$A7,$D3
	.dc.b	$3A,$7F,$C4,$D5,$FD,$DA,$73,$18,$F6,$3F,$C5,$55,$B9,$E5,$5D,$1D
	.dc.b	$98,$F8,$7F,$F0,$FB,$81,$1D,$67,$94,$62,$BA,$DF,$8A,$F8,$D6,$AF
	.dc.b	$B7,$C5,$5B,$FF,$1E,$37,$0F,$63,$BB,$F0,$B7,$1F,$81,$86,$EB,$C9
	.dc.b	$87,$8F,$1F,$05,$69,$B3,$AC,$7C,$AF,$C0,$6B,$EC,$C7,$CF,$3C,$23
	.dc.b	$FA,$C5,$FC,$56,$CF,$C4,$6D,$FE,$DD,$F8,$BB,$3E,$33,$17,$F8,$66
	.dc.b	$C6,$67,$F7,$D3,$97,$FD,$15,$F4,$FF,$E8,$B4,$D2,$3A,$3E,$A1,$F9
	.dc.b	$4E,$6F,$80,$FF,$6F,$8A,$FE,$1F,$8B,$FC,$B1,$AF,$D8,$8F,$C5,$4E
	.dc.b	$9E,$47,$43,$C2,$76,$9E,$3A,$7F,$15,$9F,$AE,$FF,$52,$BB,$5E,$13
	.dc.b	$FD,$67,$19,$69,$9B,$F1,$5F,$48,$C4,$EE,$7C,$D1,$E2,$77,$5E,$78
	.dc.b	$C4,$EC,$F7,$71,$7F,$B5,$A7,$F6,$5F,$1F,$FE,$74,$73,$F0,$F6,$FF
	.dc.b	$77,$5B,$7F,$8E,$BF,$E9,$69,$FD,$D7,$C3,$EF,$2F,$E0,$18,$5A,$D5
	.dc.b	$FF,$24,$F1,$9F,$3D,$B7,$E5,$F0,$FF,$23,$73,$97,$F9,$F1,$CB,$F5
	.dc.b	$5B,$5F,$EB,$DD,$1F,$25,$CF,$3F,$47,$5F,$EF,$E3,$B5,$7F,$8C,$8E
	.dc.b	$45,$3F,$8B,$DF,$F0,$7C,$3F,$BC,$46,$6B,$F8,$47,$27,$4C,$4F,$4A
	.dc.b	$FF,$66,$46,$1F,$F4,$9F,$D2,$BC,$5F,$88,$FF,$65,$8C,$C6,$3C,$6C
	.dc.b	$F8,$DE,$3D,$57,$FD,$6E,$FF,$AF,$46,$27,$53,$9C,$FF,$C3,$AB,$E7
	.dc.b	$BB,$9E,$7E,$4B,$87,$93,$C7,$F5,$72,$F9,$1B,$1E,$8F,$97,$EA,$47
	.dc.b	$C3,$82,$3C,$3E,$3F,$B0,$F6,$FC,$ED,$AF,$83,$F1,$9F,$4B,$E1,$DF
	.dc.b	$4F,$F0,$A3,$0F,$C4,$F2,$BD,$64,$AE,$EF,$BB,$C0,$F9,$27,$CA,$E9
	.dc.b	$B3,$F3,$BC,$2F,$9A,$7C,$EE,$8B,$1F,$4B,$E2,$7E,$39,$1F,$5B,$F2
	.dc.b	$BB,$FE,$AF,$1F,$D6,$6C,$FD,$AF,$27,$EC,$9F,$6B,$92,$C7,$DE,$F1
	.dc.b	$7E,$99,$1F,$8E,$FD,$BF,$A3,$F8,$31,$F8,$4F,$95,$FD,$34,$63,$A7
	.dc.b	$FC,$0F,$CE,$9C,$FC,$94,$BF,$F6,$ED,$BE,$4F,$18,$7F,$83,$A1,$F2
	.dc.b	$7D,$C9,$5F,$88,$8C,$B7,$E8,$BF,$4B,$F4,$8F,$81,$F1,$C6,$1F,$2F
	.dc.b	$C9,$EB,$E6,$AD,$FF,$6B,$F2,$7F,$C0,$38,$27,$D7,$F2,$9E,$F2,$57
	.dc.b	$85,$F8,$5F,$25,$E0,$F7,$7E,$0B,$67,$DB,$F4,$7F,$C3,$FE,$39,$F4
	.dc.b	$BF,$EF,$A9,$C9,$E1,$3F,$BE,$46,$B0,$FE,$2D,$C2,$FA,$67,$C5,$FE
	.dc.b	$F6,$30,$D8,$CD,$FF,$53,$6C,$FF,$92,$D7,$C6,$74,$6C,$F8,$67,$FC
	.dc.b	$19,$1C,$37,$FF,$5E,$BA,$8F,$A9,$F2,$5F,$78,$FB,$FF,$27,$6F,$97
	.dc.b	$A6,$E2,$7C,$B7,$D4,$35,$87,$F9,$C6,$33,$5F,$82,$F9,$1F,$C4,$9F
	.dc.b	$8A,$F9,$2B,$19,$3A,$CB,$E4,$F3,$7C,$DF,$9D,$C5,$F8,$48,$FF,$6C
	.dc.b	$71,$7D,$0E,$9C,$F5,$E9,$98,$FC,$1E,$17,$D3,$3E,$8B,$38,$FF,$B8
	.dc.b	$67,$31,$FD,$25,$B5,$F2,$71,$7C,$B7,$F2,$3B,$E1,$87,$3A,$D4,$E6
	.dc.b	$FF,$CC,$DD,$DF,$35,$FC,$8F,$C6,$FE,$AB,$9F,$F3,$6C,$BF,$82,$E5
	.dc.b	$A7,$EA,$BE,$4B,$F6,$27,$EC,$BA,$6C,$7C,$8F,$73,$F4,$A6,$34,$FA
	.dc.b	$4F,$EE,$3E,$93,$F1,$FE,$B7,$C9,$CF,$CB,$18,$E2,$FC,$9F,$07,$F8
	.dc.b	$1B,$0F,$F7,$E3,$5F,$F2,$A0,$70,$FE,$73,$DB,$FC,$C4,$7F,$62,$85
	.dc.b	$B2,$E3,$39,$8C,$4F,$19,$FC,$B3,$D7,$DF,$FE,$1F,$C7,$27,$F8,$5F
	.dc.b	$C9,$DA,$FE,$15,$3F,$C1,$3E,$37,$18,$7E,$F9,$FD,$A7,$37,$F8,$06
	.dc.b	$3B,$9F,$82,$EE,$CF,$9F,$F7,$FA,$11,$49,$9F,$E0,$6F,$93,$FC,$2E
	.dc.b	$76,$FF,$1E,$F9,$3E,$FE,$AF,$F2,$66,$C7,$1F,$5F,$FB,$C7,$03,$04
	.dc.b	$61,$B3,$FE,$C5,$9D,$4F,$CC,$F8,$9F,$AA,$7B,$1B,$3F,$CB,$8F,$E4
	.dc.b	$BC,$3F,$CD,$9A,$7E,$B1,$BC,$1F,$CC,$62,$ED,$FE,$BB,$FA,$57,$CA
	.dc.b	$FF,$2E,$AF,$F1,$24,$62,$BF,$6B,$19,$69,$FC,$37,$91,$3B,$9F,$E0
	.dc.b	$47,$2F,$3B,$FC,$0D,$1F,$CE,$6B,$FC,$0A,$D9,$FC,$69,$FE,$B2,$70
	.dc.b	$FE,$43,$C6,$7D,$07,$F8,$56,$BF,$89,$B6,$7F,$DB,$35,$CF,$2F,$F5
	.dc.b	$FC,$E5,$F7,$7C,$37,$B3,$CF,$F6,$B7,$0F,$55,$FF,$45,$7A,$FC,$1E
	.dc.b	$71,$FC,$DC,$7F,$AD,$72,$F1,$CF,$C6,$F9,$6F,$69,$27,$DA,$FF,$8D
	.dc.b	$FE,$31,$F4,$DF,$C1,$67,$FD,$7D,$18,$F8,$8E,$51,$FC,$DE,$30,$F9
	.dc.b	$3E,$87,$E4,$91,$F9,$2F,$F8,$FF,$77,$C3,$7F,$92,$9B,$3C,$4F,$8F
	.dc.b	$1F,$73,$D6,$91,$F0,$3E,$86,$ED,$F2,$BE,$67,$F0,$26,$38,$FF,$0C
	.dc.b	$FE,$13,$D3,$FB,$64,$BF,$B6,$F8,$FB,$BC,$46,$E3,$EB,$BF,$8C,$F4
	.dc.b	$3F,$1B,$F3,$B6,$7E,$8B,$F9,$0F,$9D,$FC,$39,$23,$F0,$EF,$E2,$BC
	.dc.b	$EF,$CA,$9B,$8F,$B2,$FF,$73,$FA,$4F,$E6,$9A,$7F,$46,$5E,$4E,$7C
	.dc.b	$9D,$3B,$F2,$FF,$6B,$26,$93,$0F,$CB,$FC,$E7,$E9,$0F,$F0,$07,$4B
	.dc.b	$9B,$13,$F8,$AF,$13,$F5,$27,$EA,$BC,$EB,$1F,$8E,$F9,$5F,$D6,$F9
	.dc.b	$BF,$AE,$6C,$7F,$4F,$CF,$FB,$A6,$DF,$CA,$1A,$FF,$82,$9A,$FF,$33
	.dc.b	$6F,$89,$FB,$23,$F7,$27,$EE,$BC,$6B,$5F,$9E,$AF,$D9,$91,$FD,$9B
	.dc.b	$FE,$7B,$F8,$0F,$EB,$CF,$F4,$0A,$46,$23,$FD,$A0,$D5,$1A,$EF,$FF
	.dc.b	$0D,$7D,$37,$D0,$8F,$F7,$53,$67,$F7,$E1,$C3,$FC,$09,$AF,$F5,$07
	.dc.b	$EA,$7C,$17,$2E,$2F,$F2,$84,$F2,$9C,$B6,$32,$FA,$70,$FF,$4E,$3F
	.dc.b	$E9,$46,$BF,$C4,$2F,$DB,$F2,$5F,$EF,$F6,$C6,$FE,$4B,$14,$69,$23
	.dc.b	$A7,$C9,$7F,$66,$EF,$FE,$BB,$7F,$22,$31,$AF,$C8,$6D,$7C,$8C,$3F
	.dc.b	$D4,$4F,$F8,$3F,$E9,$5D,$9F,$1A,$3F,$B8,$46,$1F,$E8,$CD,$7F,$BB
	.dc.b	$B5,$F8,$1E,$1F,$6C,$7E,$CF,$C9,$DC,$3F,$D4,$95,$FC,$E5,$FE,$12
	.dc.b	$DF,$92,$D1,$1D,$7B,$1C,$7D,$19,$1C,$5F,$5F,$A4,$E7,$7C,$06,$35
	.dc.b	$78,$FE,$BA,$46,$30,$F5,$CE,$63,$11,$EC,$F2,$39,$79,$32,$3E,$34
	.dc.b	$7A,$FC,$89,$1E,$B7,$DD,$F2,$32,$3F,$00,$FD,$A2,$3F,$83,$F9,$1F
	.dc.b	$89,$F2,$32,$3E,$24,$7D,$6F,$47,$FA,$8C,$DF,$3B,$FD,$D7,$5F,$D4
	.dc.b	$A3,$E6,$7C,$8C,$8F,$C7,$3F,$88,$23,$F9,$1F,$91,$FA,$44,$7E,$97
	.dc.b	$C8,$7D,$52,$38,$3E,$B7,$BD,$E1,$23,$FD,$89,$FB,$9F,$23,$23,$F3
	.dc.b	$CF,$E6,$08,$FE,$C3,$D4,$7D,$2F,$FA,$AE,$3F,$A8,$35,$FE,$BE,$23
	.dc.b	$7E,$DF,$E2,$7E,$4A,$4F,$EA,$DF,$D2,$91,$FD,$77,$91,$F1,$E2,$47
	.dc.b	$E3,$E3,$F2,$9F,$45,$FE,$EE,$8C,$10,$F0,$F8,$08,$D4,$79,$06,$FF
	.dc.b	$EA,$07,$FC,$92,$D7,$4D,$F1,$7F,$67,$3F,$ED,$39,$FF,$0C,$35,$FE
	.dc.b	$04,$D7,$F9,$03,$51,$FC,$79,$C2,$E2,$7C,$2F,$3A,$FF,$2E,$C6,$3E
	.dc.b	$4B,$F1,$53,$9F,$1F,$9E,$5E,$DF,$0F,$E1,$9A,$A9,$8F,$9A,$38,$FC
	.dc.b	$2C,$8F,$F5,$9E,$1F,$EE,$A7,$D7,$1F,$30,$D3,$F6,$B1,$F2,$C6,$63
	.dc.b	$FC,$DF,$F2,$AE,$5F,$E2,$70,$8C,$3A,$8F,$DE,$FD,$2B,$A2,$9F,$BB
	.dc.b	$71,$3F,$E5,$96,$BF,$E8,$A6,$BF,$33,$E1,$39,$F5,$C8,$FA,$71,$FC
	.dc.b	$93,$81,$F4,$BF,$E6,$69,$FF,$5D,$35,$FF,$5C,$36,$3F,$D1,$7E,$1B
	.dc.b	$07,$8F,$E8,$3A,$56,$F8,$FE,$DD,$F6,$F1,$F5,$8F,$E1,$3E,$4B,$EB
	.dc.b	$FC,$AC,$8C,$10,$FC,$AF,$3A,$7F,$36,$7E,$72,$7E,$5F,$0F,$C2,$C3
	.dc.b	$FD,$3F,$7F,$37,$83,$26,$B8,$24,$7F,$3F,$A4,$7F,$41,$8F,$A8,$D8
	.dc.b	$CC,$7F,$6A,$E7,$FD,$37,$CA,$FD,$AF,$A0,$FE,$BB,$8F,$81,$BB,$7E
	.dc.b	$43,$55,$7E,$50,$98,$7E,$31,$FD,$AF,$C6,$48,$FD,$3B,$5F,$83,$F2
	.dc.b	$DF,$9C,$89,$1F,$94,$7F,$7B,$F2,$B2,$30,$43,$E4,$BF,$7F,$D5,$FF
	.dc.b	$40,$E7,$F6,$BE,$49,$E3,$FF,$4C,$6D,$FF,$BE,$EB,$7A,$7E,$D3,$A1
	.dc.b	$F4,$98,$AF,$F6,$8D,$5C,$7F,$BF,$C7,$F1,$CA,$DB,$FD,$62,$BF,$C3
	.dc.b	$2F,$FB,$FD,$E5,$1A,$8C,$70,$7F,$18,$F9,$6F,$DD,$FA,$32,$7F,$85
	.dc.b	$C6,$5A,$43,$F0,$F1,$FC,$0B,$E4,$E4,$7E,$E9,$FE,$47,$F2,$72,$3F
	.dc.b	$8A,$B5,$FC,$7C,$7E,$DB,$E9,$3F,$9B,$67,$F8,$C6,$EB,$FE,$F6,$6D
	.dc.b	$7F,$18,$7F,$D7,$AF,$FB,$B9,$AF,$F8,$C9,$AF,$FA,$C9,$AF,$3F,$92
	.dc.b	$FF,$12,$72,$FF,$1E,$8F,$8D,$F0,$3F,$24,$FE,$77,$E2,$FF,$3C,$F7
	.dc.b	$BF,$8F,$B2,$7F,$97,$C7,$F4,$0F,$88,$E6,$48,$FC,$D5,$FF,$9D,$9B
	.dc.b	$F1,$3F,$CA,$3E,$4E,$47,$F3,$88,$FE,$93,$1F,$DD,$7E,$93,$D7,$B7
	.dc.b	$F3,$06,$ED,$F9,$F9,$FE,$9E,$FA,$89,$7E,$07,$FC,$81,$F4,$9F,$23
	.dc.b	$90,$D2,$6F,$FE,$2B,$5F,$2B,$7E,$56,$61,$FE,$E3,$9C,$B7,$E3,$7A
	.dc.b	$91,$FD,$C4,$23,$C3,$EE,$75,$24,$F3,$F8,$DF,$07,$C0,$78,$7D,$F9
	.dc.b	$3D,$E3,$C9,$F2,$5E,$A2,$47,$67,$DB,$F2,$5E,$92,$47,$9F,$E1,$FD
	.dc.b	$9F,$23,$23,$D4,$FB,$FE,$4B,$E0,$7C,$97,$C6,$24,$FA,$9F,$93,$F2
	.dc.b	$3F,$E2,$7F,$A4,$7C,$85,$3E,$44,$7C,$9F,$92,$97,$F1,$DF,$99,$F2
	.dc.b	$7F,$37,$C1,$FA,$64,$9F,$8E,$FD,$6F,$93,$ED,$F2,$32,$7E,$84,$7D
	.dc.b	$5E,$A9,$1F,$4D,$F2,$7B,$5F,$66,$3E,$F9,$0F,$A6,$7F,$A0,$7D,$3F
	.dc.b	$E4,$EF,$FA,$71,$F5,$BF,$D2,$3B,$7F,$BC,$C6,$23,$FC,$2E,$F0,$F7
	.dc.b	$DB,$FF,$4F,$9F,$F0,$33,$63,$3E,$D7,$F9,$D6,$BF,$D1,$95,$D4,$F0
	.dc.b	$7C,$97,$8C,$8F,$F3,$F8,$DB,$FE,$99,$FA,$3F,$17,$89,$FE,$10,$FA
	.dc.b	$47,$C5,$FF,$96,$5B,$5D,$76,$5F,$0D,$7F,$C2,$6F,$C0,$8C,$F2,$FF
	.dc.b	$A2,$63,$55,$FE,$BE,$75,$F4,$8E,$4F,$F4,$73,$67,$B3,$5F,$F1,$DA
	.dc.b	$E4,$D9,$23,$6F,$2E,$5F,$F0,$9F,$D2,$7F,$8F,$B6,$FF,$18,$FD,$1B
	.dc.b	$A3,$FC,$67,$82,$BF,$B1,$A9,$D3,$FE,$38,$8C,$C6,$23,$87,$C4,$AF
	.dc.b	$AD,$F4,$BF,$05,$73,$8C,$7B,$FE,$0C,$39,$57,$49,$E1,$87,$3B,$E0
	.dc.b	$78,$5C,$7C,$5F,$0C,$3D,$2D,$9F,$F2,$6B,$D2,$77,$5D,$31,$F2,$D7
	.dc.b	$AE,$FC,$53,$E4,$8C,$9E,$C9,$B8,$7A,$B1,$ED,$3E,$43,$E4,$5F,$27
	.dc.b	$CD,$5C,$30,$F5,$47,$CE,$3A,$1F,$EF,$8B,$F8,$AE,$5A,$FC,$9F,$91
	.dc.b	$FF,$44,$3B,$6B,$B6,$BB,$F8,$EF,$F0,$F2,$FA,$4A,$47,$FC,$44,$DC
	.dc.b	$BE,$D4,$38,$DC,$9F,$0A,$3F,$CE,$F5,$67,$31,$F0,$4C,$95,$CD,$74
	.dc.b	$8F,$11,$9A,$E8,$E3,$6F,$0D,$74,$78,$5F,$DA,$DB,$E4,$4E,$5F,$95
	.dc.b	$5D,$30,$E3,$87,$4B,$F3,$36,$7C,$07,$9C,$E1,$5D,$23,$F0,$76,$FA
	.dc.b	$4F,$F3,$A4,$7E,$F5,$B5,$FE,$AA,$7F,$C9,$71,$9F,$85,$F8,$7B,$FC
	.dc.b	$4B,$74,$BE,$73,$37,$FC,$79,$B8,$FC,$FC,$7A,$E1,$C6,$F8,$5F,$87
	.dc.b	$E3,$37,$8F,$95,$B6,$F9,$3B,$1A,$B7,$87,$4C,$E1,$FC,$5E,$32,$FB
	.dc.b	$78,$3F,$CD,$79,$FF,$AA,$B1,$46,$23,$CC,$DC,$BE,$33,$6D,$F3,$3E
	.dc.b	$13,$CB,$17,$0F,$95,$F2,$5E,$AF,$B3,$7C,$9F,$AA,$30,$E7,$E4,$FB
	.dc.b	$49,$A4,$AF,$E7,$1F,$37,$EF,$3A,$7C,$A5,$8F,$E6,$94,$F9,$0F,$89
	.dc.b	$E5,$78,$5B,$3F,$0F,$E6,$FD,$44,$7F,$A2,$3E,$DF,$27,$E6,$1E,$9F
	.dc.b	$8F,$B3,$DB,$E3,$BE,$87,$CC,$FB,$5B,$1F,$3F,$C0,$F8,$C4,$7F,$A8
	.dc.b	$3F,$27,$C2,$FB,$07,$C5,$F8,$7B,$3E,$AF,$86,$FB,$9E,$2F,$C9,$6C
	.dc.b	$7D,$BF,$03,$E9,$11,$FE,$B8,$FD,$5F,$8F,$FC,$21,$F4,$7E,$12,$CF
	.dc.b	$C6,$3F,$11,$E0,$7D,$56,$9F,$87,$F0,$3E,$F1,$1F,$ED,$0F,$E0,$7C
	.dc.b	$FF,$C8,$1F,$96,$F8,$17,$C0,$E2,$BF,$B3,$47,$F8,$D5,$FC,$97,$6E
	.dc.b	$FE,$03,$F3,$0F,$E6,$7B,$BF,$18,$49,$FE,$EC,$FE,$47,$D3,$7D,$2F
	.dc.b	$E9,$1B,$3D,$BD,$FC,$4F,$8F,$DE,$D7,$85,$6B,$FB,$C5,$FE,$49,$BE
	.dc.b	$17,$52,$3F,$EC,$38,$FC,$93,$7C,$1F,$01,$EF,$F6,$EC,$7F,$83,$FE
	.dc.b	$35,$FB,$43,$E1,$BF,$7F,$E5,$3E,$95,$F1,$7F,$C0,$AF,$75,$A9,$CE
	.dc.b	$1F,$E1,$9B,$E1,$D9,$FB,$36,$C7,$F2,$26,$9F,$C9,$25,$3F,$B9,$BF
	.dc.b	$1F,$57,$7C,$8F,$79,$C3,$F6,$7D,$9F,$B5,$E9,$5C,$BE,$CB,$87,$E4
	.dc.b	$FC,$CD,$CF,$FC,$01,$07,$FA,$57,$02,$7F,$09,$F0,$12,$BF,$07,$ED
	.dc.b	$5B,$E2,$70,$F2,$FE,$E1,$9F,$53,$F9,$66,$3F,$BB,$BF,$9D,$B3,$E3
	.dc.b	$3D,$9B,$FB,$5B,$3A,$8F,$85,$3F,$12,$3F,$A4,$F0,$7D,$69,$C9,$F1
	.dc.b	$FF,$70,$FE,$4B,$9F,$FB,$75,$B9,$7F,$60,$AE,$17,$F7,$E7,$2F,$C1
	.dc.b	$F5,$3F,$D2,$EF,$F9,$17,$EB,$BF,$CF,$2F,$E8,$B0,$FE,$77,$18,$7E
	.dc.b	$0C,$A3,$18,$7C,$0F,$99,$DF,$C0,$FE,$7D,$E3,$ED,$4C,$7B,$7C,$AD
	.dc.b	$CB,$9E,$2E,$0E,$E6,$72,$FE,$AD,$19,$BF,$C7,$7A,$AF,$4F,$F2,$F4
	.dc.b	$F3,$3F,$33,$39,$AF,$CA,$4E,$5B,$E6,$FE,$56,$3F,$B3,$E5,$FE,$1F
	.dc.b	$6B,$FE,$4E,$6B,$F8,$7A,$FB,$8F,$E0,$83,$88,$FF,$07,$C7,$F9,$36
	.dc.b	$7C,$7C,$BF,$2A,$A5,$FC,$FF,$5E,$DF,$CC,$D1,$F3,$1F,$C1,$CC,$E5
	.dc.b	$F8,$C7,$F4,$9D,$5F,$A5,$F0,$AE,$3F,$8E,$70,$FD,$FE,$3B,$8F,$E1
	.dc.b	$5F,$DD,$7E,$E0,$F3,$EF,$A1,$86,$DF,$C8,$F5,$3E,$31,$1F,$DD,$68
	.dc.b	$8F,$DC,$5F,$CD,$C7,$F4,$BD,$48,$C5,$7C,$43,$85,$FF,$11,$8C,$BF
	.dc.b	$93,$E5,$7A,$DF,$90,$C7,$F2,$2D,$9F,$CB,$47,$F2,$BE,$6C,$65,$F2
	.dc.b	$5F,$CB,$FA,$7F,$98,$E6,$B7,$F9,$D7,$0F,$F2,$CF,$99,$B7,$F4,$5F
	.dc.b	$F0,$FF,$C3,$BF,$31,$FE,$6F,$3F,$E6,$38,$C7,$9F,$FE,$0F,$6D,$FF
	.dc.b	$8F,$9B,$3F,$E4,$2E,$77,$B3,$FD,$04,$EB,$87,$4E,$1F,$E0,$F1,$FA
	.dc.b	$EF,$5E,$63,$FB,$3E,$4B,$8F,$F1,$0C,$BC,$9E,$7F,$23,$38,$FF,$97
	.dc.b	$A3,$37,$FE,$CB,$7D,$5D,$3E,$7C,$FF,$8F,$4F,$9D,$B8,$7B,$5A,$FF
	.dc.b	$AB,$BC,$2F,$DA,$7C,$BF,$F5,$93,$D3,$76,$FF,$DB,$18,$FD,$F7,$FC
	.dc.b	$64,$D7,$FC,$C2,$DB,$FD,$77,$3F,$F9,$E2,$7F,$AC,$7C,$C7,$F6,$53
	.dc.b	$FC,$9F,$5B,$3F,$95,$BF,$E4,$CF,$C0,$FF,$AE,$DF,$F1,$26,$CF,$F9
	.dc.b	$6D,$FE,$CD,$19,$D7,$E7,$F8,$B3,$0F,$F7,$15,$70,$7F,$A1,$6D,$F7
	.dc.b	$E7,$FC,$BF,$F0,$CF,$CA,$FE,$1E,$3F,$D1,$E3,$1A,$FF,$93,$DB,$87
	.dc.b	$F9,$F9,$B3,$F8,$1F,$77,$FC,$00,$74,$BF,$9F,$A7,$F4,$68,$F6,$61
	.dc.b	$ED,$C6,$E3,$E6,$7E,$07,$85,$FE,$5B,$D3,$F8,$2C,$FF,$A3,$38,$51
	.dc.b	$8C,$FF,$D2,$2D,$7E,$D7,$33,$E6,$BF,$C2,$47,$4B,$FF,$06,$BF,$DF
	.dc.b	$7E,$51,$F2,$FF,$DF,$FF,$70,$FA,$D8,$7E,$F2,$73,$7F,$E0,$99,$CE
	.dc.b	$FF,$D3,$FE,$91,$F2,$9F,$DE,$DB,$5F,$D6,$9B,$87,$FB,$37,$C5,$46
	.dc.b	$23,$F2,$1A,$BB,$FF,$00,$7E,$07,$C2,$B8,$7F,$89,$3D,$0F,$F1,$61
	.dc.b	$F8,$5F,$02,$C7,$DE,$ED,$F5,$7B,$3F,$89,$6C,$7F,$88,$73,$7B,$7E
	.dc.b	$72,$47,$9D,$FF,$1D,$7B,$7F,$E3,$93,$F2,$5C,$96,$7F,$C6,$BE,$AF
	.dc.b	$F8,$FC,$8F,$F8,$CD,$FF,$21,$7B,$52,$3F,$C6,$CD,$E0,$FF,$62,$FD
	.dc.b	$37,$F8,$17,$A7,$FC,$4C,$D8,$CD,$FF,$B7,$DB,$3A,$C5,$F0,$14,$93
	.dc.b	$D2,$32,$FF,$13,$DB,$FC,$65,$5F,$DB,$DB,$D5,$FE,$41,$9F,$F2,$54
	.dc.b	$63,$39,$1F,$E4,$D6,$BF,$ED,$EF,$C1,$BF,$0E,$FF,$8A,$FA,$3F,$D9
	.dc.b	$91,$88,$FF,$03,$B6,$3F,$6D,$3F,$DB,$4F,$F1,$97,$3B,$A7,$FC,$9F
	.dc.b	$7F,$E2,$06,$BF,$87,$6B,$FE,$39,$11,$C7,$3F,$6F,$E0,$5D,$EF,$AF
	.dc.b	$EF,$EF,$37,$C1,$87,$08,$C3,$E9,$9F,$F1,$25,$71,$86,$CD,$1F,$4D
	.dc.b	$CA,$30,$D8,$CC,$F3,$38,$D3,$E1,$F8,$3F,$E9,$08,$D5,$A3,$11,$FE
	.dc.b	$2E,$6E,$DF,$EF,$75,$F0,$FE,$AD,$BE,$47,$5F,$F9,$B5,$B7,$F2,$63
	.dc.b	$1C,$1F,$E6,$E8,$FF,$25,$ED,$FD,$E5,$FF,$41,$2F,$89,$E8,$D8,$CF
	.dc.b	$A5,$ED,$F8,$99,$1E,$8D,$FC,$BC,$3F,$23,$C3,$F6,$9F,$48,$93,$FC
	.dc.b	$5F,$64,$7D,$6F,$88,$F8,$7F,$3B,$2B,$E1,$D2,$7E,$67,$1C,$9F,$03
	.dc.b	$F6,$BC,$C7,$E8,$4F,$67,$0D,$AF,$E6,$8D,$7F,$10,$27,$8C,$FF,$A5
	.dc.b	$5B,$1F,$D9,$7C,$5F,$C3,$F9,$FF,$E0,$36,$FA,$7F,$E5,$19,$F3,$FC
	.dc.b	$E7,$DC,$23,$83,$D7,$E2,$FE,$17,$A6,$47,$C9,$7F,$26,$FD,$1F,$4E
	.dc.b	$F3,$9F,$F5,$44,$64,$FF,$55,$46,$5B,$1F,$BC,$E4,$F9,$B1,$F5,$78
	.dc.b	$FF,$28,$7A,$3C,$99,$1F,$E0,$17,$FD,$69,$AF,$F8,$7D,$BA,$7F,$86
	.dc.b	$27,$FC,$0D,$87,$C0,$23,$F5,$BE,$1F,$F3,$64,$70,$7D,$AF,$97,$FD
	.dc.b	$27,$34,$8F,$C4,$3F,$2B,$C9,$91,$FE,$13,$7C,$24,$7F,$11,$F1,$1F
	.dc.b	$AD,$E7,$91,$F8,$68,$FD,$5B,$FA,$C6,$3F,$8A,$3F,$6D,$EF,$BD,$FF
	.dc.b	$DB,$7C,$F4,$9F,$D8,$CE,$5B,$E4,$F0,$A9,$18,$3F,$DD,$1F,$BE,$FF
	.dc.b	$AE,$DB,$E6,$7E,$CE,$31,$1F,$E8,$5F,$A6,$E8,$DA,$FB,$71,$88,$FC
	.dc.b	$9B,$5F,$E3,$8D,$7F,$8E,$CF,$EE,$5F,$F3,$6B,$FC,$93,$C7,$FF,$5F
	.dc.b	$37,$C1,$FE,$56,$EE,$33,$D9,$FC,$1F,$C5,$7A,$F4,$F4,$47,$A7,$D7
	.dc.b	$B8,$FB,$A3,$0F,$EF,$FE,$3F,$F8,$F1,$1F,$93,$F2,$FF,$38,$8E,$0F
	.dc.b	$9B,$F0,$BF,$9E,$F8,$67,$BF,$49,$3F,$C9,$C7,$53,$9B,$F4,$AD,$8F
	.dc.b	$DC,$9F,$EF,$97,$0E,$1A,$FE,$F9,$AF,$F4,$7F,$96,$FA,$7B,$FF,$AD
	.dc.b	$A5,$F1,$DD,$8F,$D9,$CE,$B8,$BF,$DC,$DC,$EF,$94,$FC,$59,$31,$FC
	.dc.b	$4F,$CA,$7E,$38,$8E,$0F,$C6,$FC,$4F,$E4,$BD,$0F,$E8,$DE,$54,$9F
	.dc.b	$C8,$C7,$E5,$BD,$6D,$46,$0F,$CA,$BF,$67,$82,$47,$F8,$C5,$FE,$85
	.dc.b	$F1,$7F,$EB,$B8,$FE,$A1,$EA,$A3,$11,$FE,$2F,$6B,$FD,$A6,$35,$F0
	.dc.b	$FF,$D2,$67,$33,$FD,$D1,$FE,$51,$C4,$DF,$15,$E0,$3C,$3E,$87,$F6
	.dc.b	$1F,$A0,$98,$7F,$45,$7F,$63,$F1,$72,$3F,$5E,$D5,$87,$57,$FB,$7F
	.dc.b	$17,$83,$1F,$A3,$8F,$F6,$DF,$46,$57,$EE,$23,$F7,$BF,$15,$FB,$E6
	.dc.b	$47,$EF,$23,$E2,$E3,$23,$E2,$E8,$8F,$CA,$61,$27,$F1,$A7,$FB,$F1
	.dc.b	$FE,$63,$D3,$FE,$FC,$8F,$E9,$2D,$7E,$37,$07,$E7,$7D,$D9,$1D,$A3
	.dc.b	$FC,$09,$F0,$B2,$3F,$B4,$3F,$E0,$9F,$C1,$1F,$2B,$FE,$09,$FA,$1F
	.dc.b	$E7,$44,$AF,$F3,$F7,$AA,$8F,$F1,$CD,$3E,$2B,$FC,$21,$DF,$2F,$FC
	.dc.b	$17,$39,$7F,$BB,$7C,$AF,$F8,$54,$8E,$BC,$7F,$F0,$C7,$A5,$FD,$24
	.dc.b	$91,$E0,$7B,$FC,$4F,$F8,$77,$C3,$93,$DA,$3F,$C4,$1E,$24,$8F,$F0
	.dc.b	$23,$ED,$F8,$9F,$F1,$27,$C1,$FF,$55,$24,$FB,$1F,$5F,$A9,$FE,$29
	.dc.b	$88,$7A,$9F,$F1,$77,$CD,$7E,$CE,$3F,$C6,$6D,$93,$75,$FE,$45,$3A
	.dc.b	$F9,$3F,$F2,$FC,$9A,$48,$FF,$55,$4F,$F9,$85,$FC,$9F,$41,$EF,$EA
	.dc.b	$78,$8F,$B1,$D1,$E4,$F9,$49,$3F,$E3,$AC,$51,$FE,$45,$F9,$EF,$F2
	.dc.b	$56,$7E,$6F,$3B,$C6,$4B,$FF,$5D,$5A,$3F,$CE,$B5,$FE,$47,$9E,$9F
	.dc.b	$37,$F0,$8E,$9F,$09,$F1,$3D,$09,$3E,$17,$C1,$F0,$1F,$23,$C0,$F8
	.dc.b	$24,$9E,$EF,$6F,$6B,$E6,$FA,$32,$7F,$CA,$79,$27,$E8,$FA,$32,$7D
	.dc.b	$CF,$B7,$D0,$F7,$7A,$7E,$82,$4F,$FB,$56,$D1,$EC,$F2,$BE,$27,$9B
	.dc.b	$2B,$BC,$7D,$EF,$4A,$47,$CB,$7E,$57,$89,$23,$A3,$F1,$88,$FC,$AF
	.dc.b	$4B,$F0,$E4,$7D,$7E,$1F,$D3,$23,$83,$DF,$E0,$FF,$9E,$FE,$7F,$EC
	.dc.b	$C7,$F9,$D1,$AF,$E1,$0F,$F4,$1B,$8F,$43,$F8,$C7,$4A,$1D,$9C,$C9
	.dc.b	$5C,$DE,$DD,$1F,$E4,$0F,$9F,$FF,$38,$61,$F8,$FF,$9F,$FE,$C3,$7F
	.dc.b	$91,$9F,$F4,$AA,$8C,$A8,$CC,$61,$E7,$1B,$F2,$3F,$D4,$D1,$89,$D7
	.dc.b	$9B,$FE,$A9,$6D,$78,$6B,$7F,$09,$FC,$D2,$39,$1E,$37,$1F,$3B,$C2
	.dc.b	$0D,$3C,$51,$A9,$E7,$8F,$5A,$F6,$61,$E1,$9F,$2B,$63,$B4,$ED,$F7
	.dc.b	$57,$8F,$0F,$01,$2F,$AC,$F9,$8D,$D7,$F3,$AC,$7E,$1C,$7F,$A0,$1A
	.dc.b	$F0,$31,$C1,$FE,$83,$7F,$C1,$B0,$8C,$46,$BE,$1F,$A4,$62,$3F,$C2
	.dc.b	$7F,$3F,$FE,$76,$4D,$23,$FE,$7A,$7D,$8F,$FB,$33,$C5,$AF,$03,$D4
	.dc.b	$64,$A6,$72,$FC,$0B,$96,$B9,$CB,$FD,$3B,$C3,$FD,$AE,$73,$7F,$C6
	.dc.b	$F6,$E3,$3F,$DB,$DC,$DF,$DA,$BE,$38,$6D,$F6,$C7,$92,$BA,$46,$30
	.dc.b	$E7,$5E,$AC,$3D,$B5,$AB,$FF,$57,$C6,$27,$F0,$53,$FC,$E3,$6F,$F0
	.dc.b	$06,$BF,$AA,$F0,$FF,$DE,$D5,$FC,$AF,$7C,$BF,$CC,$EF,$83,$FC,$BF
	.dc.b	$D2,$CE,$5F,$D1,$1B,$3A,$8F,$DC,$3F,$D1,$A3,$37,$FE,$C7,$AF,$E7
	.dc.b	$1E,$BE,$3E,$6F,$2E,$BD,$79,$3A,$3F,$9F,$3C,$BD,$7F,$F0,$16,$3F
	.dc.b	$D4,$9B,$81,$A6,$38,$7C,$27,$1B,$FE,$90,$FA,$3E,$7D,$F3,$7E,$6F
	.dc.b	$93,$C3,$6B,$FA,$A7,$07,$F1,$0A,$FF,$04,$C7,$F9,$93,$24,$63,$7F
	.dc.b	$C1,$B5,$FE,$B8,$23,$7E,$C7,$AA,$5E,$DF,$8C,$F2,$7B,$B7,$0F,$09
	.dc.b	$F4,$E7,$FC,$9C,$E3,$0E,$91,$87,$E8,$BF,$E1,$78,$D6,$1E,$6F,$73
	.dc.b	$E0,$7B,$3E,$12,$4F,$AE,$BF,$D2,$ED,$F7,$FE,$31,$F1,$FE,$02,$C7
	.dc.b	$49,$7C,$57,$CB,$3E,$67,$B7,$67,$5E,$D7,$BC,$8F,$77,$E1,$7C,$CF
	.dc.b	$D3,$3E,$A7,$C5,$59,$ED,$F2,$5F,$5C,$FB,$1E,$ED,$8F,$17,$CA,$7C
	.dc.b	$B2,$3E,$B7,$E7,$7C,$B7,$DF,$3F,$01,$F1,$76,$7D,$47,$E0,$CF,$C2
	.dc.b	$7B,$D5,$F9,$3E,$E3,$E6,$7F,$22,$D8,$FC,$13,$79,$BF,$81,$B5,$FC
	.dc.b	$0F,$32,$1F,$29,$FA,$86,$D7,$6F,$96,$C4,$E7,$E4,$E4,$F8,$AA,$E1
	.dc.b	$FA,$E5,$EF,$F9,$09,$3E,$47,$CF,$F2,$1D,$CF,$07,$C8,$D9,$F6,$FB
	.dc.b	$7E,$33,$C9,$EF,$D8,$F8,$3E,$F7,$C0,$23,$F0,$1F,$87,$EE,$7A,$CF
	.dc.b	$67,$BB,$67,$E4,$FC,$5F,$F0,$8F,$8C,$7C,$F3,$97,$E1,$4E,$63,$15
	.dc.b	$FE,$D5,$F9,$F7,$4F,$FA,$07,$89,$18,$9F,$F4,$4B,$63,$FD,$AE,$D7
	.dc.b	$F6,$27,$0E,$BF,$E3,$2D,$9E,$D5,$8A,$57,$D5,$F8,$0F,$AD,$F0,$16
	.dc.b	$FF,$06,$E1,$FC,$2F,$B3,$71,$FA,$8F,$DD,$F8,$09,$3F,$8A,$F8,$CB
	.dc.b	$5F,$67,$E3,$3F,$04,$63,$CF,$FF,$0E,$F5,$7F,$BC,$EF,$FD,$9B,$BF
	.dc.b	$B6,$7F,$D7,$1C,$5F,$2B,$C0,$F9,$1C,$3F,$25,$97,$C3,$82,$E7,$FE
	.dc.b	$2B,$BF,$A3,$1F,$3C,$FA,$9F,$3F,$FE,$63,$AE,$07,$FB,$71,$C5,$7E
	.dc.b	$3C,$FC,$C1,$F9,$9F,$02,$CF,$F0,$EF,$9F,$7C,$AF,$F4,$46,$C7,$EA
	.dc.b	$AB,$13,$9F,$8F,$D7,$C7,$5A,$E4,$E1,$E9,$F0,$D6,$FF,$A6,$E1,$FE
	.dc.b	$40,$9F,$07,$C7,$7D,$CF,$8D,$7A,$5E,$18,$C3,$F8,$0E,$27,$BF,$F8
	.dc.b	$B6,$DF,$FB,$49,$B8,$FE,$20,$FF,$73,$CE,$FE,$03,$91,$CD,$B5,$F0
	.dc.b	$7E,$3B,$E0,$1F,$07,$E2,$6C,$7C,$5C,$FE,$21,$8C,$3D,$6F,$E8,$7E
	.dc.b	$42,$4F,$E0,$3D,$9B,$5F,$67,$A7,$FA,$3C,$FE,$13,$DB,$CE,$BF,$84
	.dc.b	$9C,$D7,$C6,$F8,$FF,$D8,$7C,$FB,$E1,$3E,$A4,$61,$CF,$C2,$FF,$BB
	.dc.b	$7B,$15,$FD,$DE,$FF,$53,$18,$9D,$4E,$67,$F5,$B8,$FF,$8E,$73,$FE
	.dc.b	$A9,$08,$C3,$F9,$19,$E9,$09,$FE,$CA,$E8,$FF,$3E,$F2,$BE,$3B,$F1
	.dc.b	$47,$E2,$FE,$42,$D7,$E3,$FE,$33,$F1,$C6,$31,$FC,$2B,$F9,$1F,$8A
	.dc.b	$93,$F9,$5F,$8A,$B5,$F8,$AE,$8F,$CB,$18,$CF,$C2,$FE,$67,$A3,$FC
	.dc.b	$1F,$A7,$E1,$2B,$F9,$DB,$5F,$F1,$33,$57,$CC,$3E,$63,$FC,$1B,$97
	.dc.b	$F7,$78,$C7,$C4,$75,$8C,$4E,$AB,$F3,$BC,$7F,$A9,$3F,$55,$C5,$6B
	.dc.b	$F5,$DF,$17,$FA,$D3,$15,$FA,$47,$F6,$1F,$17,$E8,$24,$FF,$85,$F3
	.dc.b	$FF,$44,$BF,$9D,$CF,$CD,$D5,$FE,$3C,$6C,$7E,$EE,$B1,$9F,$F8,$2F
	.dc.b	$1F,$D8,$FC,$5F,$C4,$25,$75,$7E,$E4,$7E,$13,$9B,$EF,$4F,$F8,$5F
	.dc.b	$87,$F1,$AF,$F0,$5A,$FF,$25,$C6,$23,$FC,$2F,$F3,$DF,$E2,$B8,$FF
	.dc.b	$5F,$B6,$3F,$CA,$8F,$C7,$8C,$FC,$67,$F9,$4A,$1F,$19,$4C,$45,$AE
	.dc.b	$BD,$BF,$E4,$AA,$FD,$3E,$5F,$DD,$9A,$F6,$8F,$5B,$FE,$3A,$F9,$07
	.dc.b	$CF,$FC,$3C,$BD,$F1,$8E,$0F,$15,$FF,$99,$A3,$18,$FB,$3C,$0F,$35
	.dc.b	$79,$DB,$1F,$03,$E3,$BC,$84,$7F,$C9,$D9,$FF,$8D,$33,$FF,$47,$AF
	.dc.b	$8E,$F8,$07,$C1,$AB,$5F,$E4,$5F,$03,$FC,$0B,$F0,$BF,$DC,$7E,$0F
	.dc.b	$FC,$2D,$B7,$F4,$A5,$19,$F8,$6F,$EE,$DF,$31,$F6,$34,$FE,$02,$D7
	.dc.b	$FC,$B7,$82,$31,$01,$19,$C7,$E7,$79,$5F,$54,$FF,$01,$7C,$FB,$DE
	.dc.b	$C4,$FE,$03,$E3,$FE,$D1,$F6,$FE,$3E,$C7,$E1,$3E,$37,$EF,$7C,$7F
	.dc.b	$DF,$6C,$7B,$F4,$7B,$3F,$C6,$DB,$1F,$CE,$F4,$F4,$4E,$4F,$C1,$7C
	.dc.b	$77,$E2,$4F,$C5,$7C,$7D,$AF,$C9,$FB,$7F,$87,$23,$F5,$1F,$C5,$7B
	.dc.b	$92,$3E,$BB,$78,$7F,$A5,$7C,$FF,$93,$DC,$FE,$99,$08,$C3,$AC,$BF
	.dc.b	$D0,$CD,$9D,$CF,$F8,$8B,$C2,$B6,$69,$2B,$F0,$B7,$DA,$DD,$7F,$12
	.dc.b	$7F,$17,$39,$8C,$47,$E3,$3E,$43,$04,$7F,$22,$EF,$2F,$E6,$E7,$EC
	.dc.b	$8B,$1F,$E9,$D9,$FD,$3F,$57,$EE,$3E,$1F,$3C,$0F,$9F,$7A,$1F,$0A
	.dc.b	$30,$D8,$CE,$1F,$5D,$AF,$FB,$BB,$E7,$FF,$2E,$FF,$A7,$1F,$F3,$A3
	.dc.b	$7B,$3C,$7B,$27,$73,$FE,$D4,$7F,$CE,$4F,$1F,$82,$DC,$62,$35,$F2
	.dc.b	$58,$F9,$29,$1D,$DF,$5F,$47,$F9,$1A,$3F,$5D,$93,$D5,$F5,$12,$3D
	.dc.b	$4F,$B7,$D9,$F8,$72,$F6,$29,$88,$F1,$F8,$92,$FD,$CF,$9B,$E4,$64
	.dc.b	$7F,$B4,$74,$47,$DF,$6C,$66,$31,$F2,$1E,$EF,$92,$93,$F2,$63,$EC
	.dc.b	$F9,$8F,$9F,$7B,$18,$6D,$7F,$B3,$E7,$53,$F2,$3E,$4E,$4F,$8D,$F1
	.dc.b	$11,$F3,$7C,$9F,$CE,$23,$F6,$7E,$4F,$E8,$91,$C1,$F7,$7E,$4F,$F1
	.dc.b	$AF,$CD,$F1,$EF,$19,$A7,$DA,$F2,$E4,$FC,$57,$E2,$11,$FE,$C1,$EC
	.dc.b	$7F,$B6,$1F,$EE,$3E,$67,$E6,$BD,$AF,$EF,$0F,$E1,$3E,$4A,$4F,$CF
	.dc.b	$7E,$71,$1F,$A5,$F2,$3F,$8A,$23,$F9,$5F,$90,$FC,$69,$1C,$1F,$99
	.dc.b	$F9,$2F,$06,$2F,$3E,$92,$7F,$3D,$F2,$5F,$96,$3E,$F1,$1F,$C1,$3F
	.dc.b	$81,$F9,$2F,$CD,$92,$3F,$4C,$7E,$78,$86,$0F,$D5,$78,$7F,$A8,$F6
	.dc.b	$FF,$CF,$31,$FC,$EF,$C9,$FE,$F7,$29,$CC,$FF,$8F,$7C,$BF,$E8,$13
	.dc.b	$9F,$91,$E5,$F1,$F2,$B9,$56,$AD,$39,$9E,$42,$FB,$7C,$8F,$E0,$3E
	.dc.b	$7F,$BE,$5F,$DC,$9B,$8F,$F4,$FB,$46,$1F,$EA,$3F,$3F,$FE,$C8,$7F
	.dc.b	$B1,$35,$FF,$2D,$B5,$FE,$B6,$D7,$FA,$D7,$9C,$8C,$47,$97,$37,$C9
	.dc.b	$FF,$1C,$9F,$E0,$53,$FE,$AD,$6E,$3F,$22,$B7,$3E,$2F,$73,$81,$1F
	.dc.b	$2F,$BF,$CC,$8E,$0F,$4F,$BF,$F5,$7E,$4E,$47,$CC,$3C,$8F,$D9,$F9
	.dc.b	$39,$19,$E0,$FB,$11,$88,$F9,$F1,$F6,$FE,$4F,$F5,$BD,$4E,$5A,$4A
	.dc.b	$FF,$3D,$C3,$9B,$67,$F5,$A8,$46,$5F,$EB,$AD,$7F,$D5,$4D,$7E,$C5
	.dc.b	$FE,$2F,$E4,$FF,$7B,$F3,$FF,$66,$4F,$DB,$9D,$47,$AF,$77,$C9,$7E
	.dc.b	$5F,$E4,$E4,$E6,$3F,$35,$E8,$FE,$6D,$91,$C6,$3F,$3D,$E9,$48,$E8
	.dc.b	$FE,$87,$E4,$E4,$77,$6B,$F2,$30,$FE,$7C,$FF,$8B,$7E,$23,$FD,$58
	.dc.b	$FF,$84,$7E,$33,$FA,$29,$FD,$82,$33,$1F,$EF,$17,$F9,$5B,$7E,$47
	.dc.b	$46,$FC,$FE,$1F,$27,$67,$B4,$61,$F8,$5E,$F7,$ED,$C8,$FC,$58,$FD
	.dc.b	$D1,$1C,$1F,$26,$7F,$7D,$F2,$5F,$BF,$B9,$3F,$36,$3F,$81,$7C,$8E
	.dc.b	$A3,$07,$D1,$7F,$99,$F1,$FF,$7E,$F0,$BF,$64,$D9,$FF,$21,$35,$2F
	.dc.b	$81,$E2,$7E,$25,$C4,$7F,$89,$1A,$FF,$BC,$5A,$FE,$23,$6F,$F1,$03
	.dc.b	$F4,$77,$F9,$C7,$F1,$9F,$85,$FF,$00,$FC,$F3,$E4,$69,$2B,$F4,$90
	.dc.b	$F9,$4D,$F1,$FF,$89,$E3,$FD,$13,$3F,$F0,$07,$A1,$18,$8F,$E2,$07
	.dc.b	$F1,$77,$0F,$E7,$17,$C8,$F8,$89,$1E,$0C,$FC,$BE,$0C,$8F,$1B,$E8
	.dc.b	$E6,$F4,$F8,$5E,$32,$4F,$8F,$0F,$F4,$03,$7E,$3F,$DE,$47,$DB,$E2
	.dc.b	$7C,$2F,$8F,$91,$F0,$1F,$8B,$E9,$FC,$6F,$5B,$E0,$12,$7B,$3F,$0F
	.dc.b	$D0,$F9,$64,$8F,$93,$F1,$FF,$3B,$E3,$64,$7C,$C7,$E8,$FC,$7F,$D2
	.dc.b	$F8,$EF,$98,$49,$F5,$3F,$3F,$AB,$EB,$92,$3E,$AF,$C7,$7F,$8E,$BE
	.dc.b	$7B,$FC,$7B,$1F,$C6,$1B,$E7,$FF,$54,$6F,$77,$F5,$2E,$2F,$C9,$74
	.dc.b	$27,$FC,$83,$7F,$C4,$1A,$FF,$58,$6B,$FD,$45,$B1,$C7,$E3,$7C,$44
	.dc.b	$9F,$E5,$92,$F8,$DF,$37,$C5,$49,$CB,$E9,$F8,$DF,$57,$C5,$7F,$9C
	.dc.b	$EF,$FC,$6B,$5F,$8A,$C7,$CD,$39,$35,$F1,$3F,$04,$93,$E0,$F8,$AF
	.dc.b	$89,$F0,$F2,$3C,$6F,$C7,$F8,$8F,$91,$F0,$FD,$C9,$3F,$9B,$C9,$1E
	.dc.b	$3F,$86,$F9,$DF,$11,$2B,$DB,$1F,$47,$E1,$A4,$7C,$07,$EA,$7A,$F2
	.dc.b	$3A,$3E,$D2,$3D,$FE,$17,$ED,$11,$F9,$3E,$CF,$DD,$23,$83,$E6,$FC
	.dc.b	$37,$F9,$EB,$E7,$BF,$C0,$F1,$FD,$6D,$AF,$F5,$A6,$F4,$FF,$85,$97
	.dc.b	$BB,$C4,$93,$C5,$E9,$EF,$FF,$93,$3E,$7B,$FC,$F1,$7F,$ED,$9F,$9E
	.dc.b	$FE,$B7,$7F,$ED,$B9,$7A,$7F,$E0,$F9,$CC,$A7,$2F,$1D,$7F,$C2,$91
	.dc.b	$D2,$13,$C2,$30,$DA,$E9,$5C,$1B,$5E,$1F,$8B,$61,$D6,$30,$FF,$BB
	.dc.b	$FD,$6E,$51,$6F,$93,$4D,$88,$F1,$57,$48,$F2,$CF,$06,$9F,$06,$FC
	.dc.b	$6F,$BE,$78,$61,$CB,$0F,$24,$FC,$3A,$DE,$3F,$12,$39,$9C,$67,$71
	.dc.b	$F1,$63,$F9,$5B,$5F,$93,$7E,$57,$2D,$7E,$6F,$B8,$F8,$CF,$F7,$5D
	.dc.b	$7F,$AA,$DB,$19,$9F,$E6,$E7,$D7,$70,$FF,$BC,$BE,$7B,$FC,$34,$9A
	.dc.b	$47,$FC,$3C,$FB,$9F,$F2,$B7,$87,$5E,$07,$B8,$C9,$4C,$E7,$F8,$87
	.dc.b	$2D,$73,$A7,$F0,$39,$E9,$5F,$66,$7F,$CD,$B1,$51,$E0,$D7,$A7,$72
	.dc.b	$75,$3D,$0F,$F1,$7B,$D1,$F3,$46,$DF,$35,$7B,$63,$18,$79,$63,$15
	.dc.b	$C6,$3B,$B7,$0F,$45,$F2,$3A,$C6,$5F,$09,$99,$E2,$D9,$E3,$1D,$E3
	.dc.b	$89,$93,$8B,$9C,$38,$96,$7C,$2F,$18,$EA,$58,$EA,$7E,$0A,$FD,$75
	.dc.b	$D4,$F9,$A7,$A6,$BA,$CE,$E3,$F5,$0F,$63,$15,$D9,$A7,$68,$FF,$02
	.dc.b	$D7,$69,$F1,$47,$AA,$7A,$57,$68,$D9,$D8,$F8,$EF,$07,$CB,$35,$FC
	.dc.b	$4C,$F8,$1E,$87,$28,$F2,$3D,$27,$B3,$FE,$A7,$AF,$C5,$61,$F8,$19
	.dc.b	$E8,$FE,$02,$FA,$57,$81,$AF,$80,$EC,$FB,$DE,$87,$A4,$F9,$73,$F5
	.dc.b	$5F,$5C,$72,$9E,$D1,$C8,$F4,$9C,$AF,$D2,$D9,$EC,$FE,$82,$FD,$38
	.dc.b	$7E,$79,$B1,$E1,$34,$7E,$2A,$BD,$B3,$F9,$38,$F6,$1E,$12,$9E,$AC
	.dc.b	$3E,$3B,$C9,$AF,$EA,$CD,$9D,$A7,$CE,$DB,$F3,$BB,$3E,$41,$D5,$FC
	.dc.b	$53,$F5,$5A,$FB,$CF,$86,$73,$29,$CF,$0F,$DC,$6B,$FB,$98,$E0,$7E
	.dc.b	$69,$C3,$FE,$BB,$C3,$FD,$6D,$F0,$78,$8F,$F7,$43,$5F,$E1,$6D,$F6
	.dc.b	$9E,$47,$6A,$FF,$3E,$B6,$B8,$C7,$FB,$45,$FF,$3E,$C6,$7D,$A7,$0B
	.dc.b	$E2,$F0,$DB,$FF,$6C,$4B,$EF,$FF,$8D,$C6,$27,$58,$7F,$60,$8F,$E3
	.dc.b	$F4,$FC,$FD,$F0,$FD,$DB,$3C,$71,$C4,$F8,$F8,$C5,$31,$F0,$96,$BC
	.dc.b	$7F,$A7,$3D,$AF,$53,$63,$CF,$5C,$9D,$CF,$C3,$F8,$B7,$ED,$EF,$96
	.dc.b	$C3,$6B,$FD,$ED,$3A,$F9,$AC,$D7,$1C,$FF,$9C,$56,$9B,$BF,$C1,$17
	.dc.b	$DA,$B8,$54,$3E,$5F,$ED,$0F,$8C,$C5,$F1,$E1,$4E,$6B,$B7,$AB,$E1
	.dc.b	$E5,$F1,$36,$3C,$51,$ED,$FD,$E1,$EC,$67,$1F,$26,$89,$CD,$FD,$CF
	.dc.b	$DE,$ED,$78,$57,$B9,$F8,$57,$3D,$9F,$D4,$75,$46,$AF,$95,$7E,$2F
	.dc.b	$E0,$3E,$56,$1C,$E3,$1C,$1F,$93,$8F,$EC,$7E,$55,$5E,$55,$8C,$39
	.dc.b	$F2,$FF,$41,$26,$93,$1F,$F6,$97,$B9,$D7,$C0,$3C,$63,$F4,$11,$87
	.dc.b	$F0,$AF,$E3,$CF,$C8,$73,$58,$FC,$6F,$C2,$FE,$30,$C6,$1F,$57,$B3
	.dc.b	$FD,$8F,$3F,$86,$FD,$43,$F1,$D3,$F3,$97,$77,$E3,$DC,$F1,$FF,$79
	.dc.b	$E5,$46,$AF,$95,$7E,$57,$CD,$7C,$17,$18,$C3,$CF,$E4,$BA,$7C,$A5
	.dc.b	$BF,$13,$DF,$E5,$7F,$84,$65,$D9,$B9,$1A,$6B,$F4,$35,$7E,$95,$B7
	.dc.b	$ED,$9C,$4F,$F0,$7F,$89,$73,$5E,$6C,$36,$FF,$83,$CF,$DC,$7F,$84
	.dc.b	$FB,$CF,$87,$D7,$17,$E7,$E3,$13,$FC,$3B,$E7,$FF,$19,$3F,$A3,$5F
	.dc.b	$0F,$FC,$5D,$CF,$27,$E4,$DC,$D7,$19,$FE,$39,$F7,$4E,$9E,$11,$87
	.dc.b	$F9,$07,$CF,$FF,$9B,$6B,$FC,$9F,$E0,$A3,$19,$7F,$8E,$0F,$E3,$6E
	.dc.b	$1C,$46,$FD,$4F,$E6,$CB,$EA,$1A,$FF,$2D,$6E,$1F,$CA,$B0,$EB,$AF
	.dc.b	$EA,$E7,$38,$FF,$01,$F8,$F7,$97,$FC,$D5,$B8,$7F,$32,$7F,$5B,$7F
	.dc.b	$B3,$73,$5F,$C2,$7E,$FF,$F4,$D5,$FC,$E1,$B1,$D6,$FF,$A0,$46,$2F
	.dc.b	$F7,$9F,$87,$FF,$AA,$2F,$E1,$B6,$FE,$10,$D1,$19,$8F,$86,$D7,$FA
	.dc.b	$0E,$1F,$AF,$7F,$8E,$36,$FB,$FB,$9F,$D4,$63,$18,$7F,$48,$FC,$7B
	.dc.b	$85,$FC,$E6,$B9,$F1,$3E,$7C,$62,$3C,$9F,$92,$BE,$27,$FA,$E3,$6F
	.dc.b	$FA,$CD,$76,$AD,$D7,$F6,$5F,$A0,$FF,$4A,$79,$37,$DF,$F9,$B8,$7F
	.dc.b	$9E,$27,$FA,$E3,$FE,$7A,$D7,$FC,$E5,$1C,$23,$06,$FF,$31,$7C,$4F
	.dc.b	$56,$C6,$B2,$57,$17,$FA,$42,$9D,$C6,$3F,$2D,$FC,$2B,$F9,$39,$46
	.dc.b	$23,$CC,$D7,$AB,$61,$F9,$FB,$E2,$BE,$03,$70,$F2,$4F,$4E,$9C,$D7
	.dc.b	$F3,$7F,$D2,$BF,$A9,$63,$FD,$74,$3F,$AD,$BF,$F0,$63,$AD,$3F,$C3
	.dc.b	$CE,$67,$87,$E4,$2F,$90,$FF,$0D,$36,$FF,$D7,$F2,$8C,$3F,$EC,$26
	.dc.b	$BF,$B3,$F2,$9F,$4F,$FE,$12,$8F,$EB,$AD,$C7,$FC,$5D,$5C,$9F,$E6
	.dc.b	$11,$9B,$FD,$E7,$E9,$EF,$94,$FE,$A8,$FE,$AA,$BF,$C6,$CE,$67,$FC
	.dc.b	$50,$E5,$FE,$09,$F5,$1F,$DA,$31,$FD,$2B,$5C,$FB,$7F,$E3,$78,$C4
	.dc.b	$7F,$B0,$BF,$00,$7C,$CF,$F6,$36,$DF,$F8,$D7,$D2,$FF,$1D,$BF,$E3
	.dc.b	$78,$CD,$7E,$1B,$EB,$3C,$8D,$D3,$F3,$51,$E5,$8F,$C6,$BA,$D3,$FC
	.dc.b	$A1,$FA,$7F,$F2,$E9,$FB,$4B,$E7,$7F,$BA,$39,$9D,$C7,$F7,$7C,$BF
	.dc.b	$99,$D7,$F0,$26,$BF,$C9,$5A,$FF,$B7,$9B,$CC,$C3,$F5,$73,$E4,$EC
	.dc.b	$FD,$87,$E5,$DF,$E2,$8C,$7F,$57,$5A,$F1,$5F,$8C,$BC,$8F,$E5,$AD
	.dc.b	$E0,$FE,$6E,$DC,$BF,$8A,$D7,$0A,$FE,$67,$F8,$AB,$E6,$29,$30,$FE
	.dc.b	$7F,$19,$7F,$23,$8F,$F9,$AB,$E4,$6F,$C3,$1A,$4C,$BF,$91,$9E,$DF
	.dc.b	$23,$F9,$23,$8A,$FF,$08,$E7,$D2,$73,$3F,$CA,$A7,$2F,$A3,$D0,$CB
	.dc.b	$FD,$31,$AF,$F9,$A9,$AF,$F3,$16,$D7,$F6,$4F,$82,$FD,$4B,$67,$FA
	.dc.b	$DF,$A5,$FE,$27,$6C,$FE,$BB,$8F,$EA,$7C,$DF,$F5,$DE,$2F,$A2,$DD
	.dc.b	$BF,$90,$66,$8C,$61,$F8,$CF,$22,$BF,$B2,$C7,$F7,$2D,$3F,$77,$DF
	.dc.b	$E1,$7E,$3B,$5F,$E9,$7F,$AB,$7E,$C5,$FF,$4A,$2F,$9F,$EA,$D8,$CF
	.dc.b	$67,$8B,$5F,$C6,$79,$97,$4F,$E8,$15,$FC,$69,$AF,$E9,$FE,$7B,$CF
	.dc.b	$F4,$7F,$E1,$E8,$FF,$51,$35,$FE,$38,$DE,$0F,$F5,$B7,$DA,$3E,$AB
	.dc.b	$DF,$F8,$23,$C0,$F7,$B7,$0F,$E3,$2D,$7F,$0A,$D7,$FB,$4F,$EA,$BF
	.dc.b	$E2,$17,$FD,$64,$FF,$8F,$9B,$F0,$DF,$32,$35,$1F,$E7,$8F,$A3,$7D
	.dc.b	$17,$D2,$F3,$51,$89,$FF,$4B,$B5,$FE,$6D,$F7,$5C,$B5,$FF,$68,$3D
	.dc.b	$23,$FB,$DB,$5F,$F0,$F3,$50,$70,$FF,$B3,$B3,$FC,$0F,$80,$FA,$2A
	.dc.b	$62,$3E,$B7,$D0,$7F,$B2,$EB,$FA,$6B,$5F,$EB,$ED,$C3,$FB,$5F,$85
	.dc.b	$FE,$A3,$9F,$EA,$6D,$7F,$C2,$3C,$DF,$B3,$F9,$7F,$CD,$4E,$67,$FB
	.dc.b	$23,$5F,$EE,$4F,$E1,$FD,$8F,$DD,$47,$F5,$E6,$BF,$DC,$0F,$51,$FD
	.dc.b	$88,$A6,$72,$FC,$83,$8C,$3F,$C0,$2F,$DF,$7F,$D4,$9E,$DB,$E7,$FF
	.dc.b	$4F,$E2,$23,$15,$FE,$EA,$6B,$FE,$BF,$FC,$9E,$DF,$FA,$B5,$CC,$7F
	.dc.b	$93,$9A,$FF,$BB,$1A,$FF,$8C,$DB,$4F,$C8,$3F,$CA,$33,$FE,$F8,$6B
	.dc.b	$FE,$CD,$6D,$7F,$AE,$FF,$1D,$FF,$53,$CF,$FB,$F9,$AF,$F2,$C6,$BF
	.dc.b	$E6,$06,$BF,$EF,$E6,$BF,$E4,$5F,$C4,$5F,$27,$F0,$9B,$1F,$E2,$1C
	.dc.b	$11,$88,$FF,$70,$7E,$09,$70,$FF,$67,$37,$C9,$FC,$9D,$A3,$15,$FE
	.dc.b	$53,$6C,$67,$6F,$A3,$F0,$FF,$D0,$6B,$C0,$D7,$FD,$DC,$DC,$3F,$DB
	.dc.b	$9E,$3B,$E7,$3E,$CB,$6B,$FD,$0A,$D7,$FA,$7B,$5F,$F4,$1B,$5F,$3F
	.dc.b	$DF,$7F,$1D,$7F,$DF,$9F,$19,$FE,$C4,$7E,$13,$5F,$F8,$01,$1B,$BF
	.dc.b	$F7,$73,$FC,$86,$33,$E4,$7E,$32,$35,$AF,$6F,$2B,$F1,$FF,$3B,$F0
	.dc.b	$BE,$0F,$FD,$4F,$87,$FA,$C3,$D0,$CF,$1E,$BE,$6A,$CF,$D3,$BF,$92
	.dc.b	$23,$F8,$77,$CB,$79,$75,$FF,$6C,$37,$3F,$E0,$BE,$22,$31,$82,$1F
	.dc.b	$BC,$3D,$BF,$E3,$0D,$9F,$ED,$33,$A9,$FF,$20,$B6,$33,$3B,$7F,$D2
	.dc.b	$30,$FD,$3B,$FC,$53,$1F,$E6,$B8,$C6,$1F,$C9,$64,$DF,$80,$C7,$97
	.dc.b	$FC,$A7,$67,$D7,$D3,$38,$F3,$C3,$2D,$AC,$AD,$7F,$DF,$53,$A8,$C6
	.dc.b	$3F,$96,$F9,$C7,$D8,$7F,$02,$6C,$FE,$85,$AF,$03,$FD,$59,$FA,$93
	.dc.b	$E6,$3F,$84,$B6,$3F,$50,$D7,$87,$72,$31,$97,$C0,$9E,$7F,$4B,$FD
	.dc.b	$DB,$DD,$FF,$1F,$47,$FA,$44,$FF,$72,$7C,$26,$4F,$EF,$7F,$5A,$FB
	.dc.b	$5A,$7F,$7C,$71,$3F,$E7,$66,$BF,$E2,$26,$BF,$EC,$65,$F6,$DD,$3D
	.dc.b	$89,$3E,$19,$CB,$FC,$0F,$F7,$AF,$CF,$67,$E9,$D5,$F4,$DF,$EE,$2A
	.dc.b	$FE,$10,$D7,$E0,$D7,$0E,$2F,$E0,$AF,$FA,$36,$3A,$78,$1F,$CC,$A3
	.dc.b	$51,$FE,$D9,$74,$FD,$1E,$97,$D4,$BE,$BF,$FB,$77,$81,$FB,$FD,$3E
	.dc.b	$65,$23,$19,$7E,$8A,$F8,$61,$F9,$37,$FB,$95,$F0,$73,$1D,$DB,$19
	.dc.b	$F6,$BF,$BE,$47,$FA,$D6,$7F,$93,$7D,$53,$EE,$7F,$A5,$B6,$3B,$7C
	.dc.b	$B7,$02,$38,$3C,$5F,$67,$D0,$8E,$0F,$37,$C4,$FE,$57,$AE,$47,$E5
	.dc.b	$5F,$C8,$FC,$6C,$8F,$C8,$B5,$FC,$8F,$63,$ED,$FF,$AF,$FD,$83,$B6
	.dc.b	$92,$BF,$A0,$7D,$88,$FE,$CA,$57,$2E,$0C,$FD,$87,$E6,$72,$93,$F9
	.dc.b	$9A,$FC,$FF,$1F,$F8,$5A,$1D,$6F,$B9,$FC,$D9,$F9,$92,$47,$E6,$5F
	.dc.b	$CE,$63,$FB,$B8,$7C,$0F,$D6,$A7,$D1,$7E,$90,$FD,$2D,$6E,$74,$DE
	.dc.b	$AF,$15,$70,$BF,$F1,$4D,$73,$BF,$33,$C3,$A3,$37,$C3,$0F,$17,$D3
	.dc.b	$6F,$0F,$F1,$8B,$53,$C5,$B3,$C6,$B9,$C7,$F8,$E3,$1F,$13,$EC,$9F
	.dc.b	$2E,$3E,$27,$84,$64,$E1,$87,$AD,$E5,$3B,$69,$D8,$F9,$F1,$B6,$DF
	.dc.b	$A2,$BB,$C5,$38,$9E,$83,$BC,$65,$E2,$78,$CF,$B7,$5E,$16,$9E,$1B
	.dc.b	$FC,$54,$DB,$F1,$47,$C0,$C7,$E4,$1F,$1C,$F9,$18,$FC,$B2,$63,$F0
	.dc.b	$9D,$3E,$8C,$3C,$55,$CE,$3E,$3D,$6A,$7D,$6F,$68,$C9,$CC,$E5,$5D
	.dc.b	$A7,$DE,$7C,$13,$DF,$3C,$5E,$73,$C9,$DC,$65,$F8,$27,$92,$FD,$E7
	.dc.b	$E2,$27,$B9,$23,$BD,$79,$DE,$35,$CE,$FE,$4C,$73,$C3,$E6,$63,$F3
	.dc.b	$E7,$E0,$4F,$8B,$0F,$63,$F6,$B0,$F6,$3F,$61,$ED,$3C,$23,$BC,$FA
	.dc.b	$DF,$7B,$F3,$70,$F1,$63,$EA,$C7,$F0,$57,$F9,$8C,$3A,$BE,$67,$AE
	.dc.b	$3E,$53,$CD,$5C,$71,$FA,$F1,$97,$F0,$97,$1F,$A1,$7D,$71,$F8,$D1
	.dc.b	$E9,$7C,$D1,$E9,$25,$F5,$C7,$E4,$61,$E0,$C3,$A4,$63,$0B,$87,$2A
	.dc.b	$F1,$47,$CD,$24,$ED,$B3,$B7,$E8,$9C,$8F,$B2,$F9,$B0,$F4,$46,$0F
	.dc.b	$A8,$7E,$20,$87,$CE,$32,$7D,$23,$2F,$E0,$27,$F8,$54,$7F,$77,$C3
	.dc.b	$F9,$13,$70,$FE,$FD,$18,$8F,$E3,$BB,$3E,$87,$FC,$04,$D8,$FF,$01
	.dc.b	$35,$FF,$01,$35,$FE,$6C,$7F,$B3,$3E,$D2,$B8,$6B,$FC,$ED,$AE,$CA
	.dc.b	$FF,$11,$5F,$07,$FE,$1F,$C7,$FB,$6C,$6D,$B1,$9C,$FB,$D7,$18,$C5
	.dc.b	$78,$4F,$E2,$07,$4D,$FB,$29,$FE,$2C,$E4,$FD,$D1,$E5,$DB,$BC,$EA
	.dc.b	$38,$1D,$2B,$BB,$ED,$31,$1D,$AB,$C8,$D8,$C9,$E4,$39,$E1,$D9,$C4
	.dc.b	$F0,$69,$C2,$3E,$1B,$C3,$0F,$A3,$7E,$28,$F8,$D1,$8B,$F0,$57,$53
	.dc.b	$8D,$F5,$6D,$75,$AF,$96,$FC,$2B,$FF,$28,$61,$A3,$CC,$70,$AE,$95
	.dc.b	$C1,$B5,$F4,$5A,$FC,$56,$BF,$36,$3C,$CF,$D0,$3E,$6C,$F6,$8E,$86
	.dc.b	$A7,$C0,$D3,$33,$FE,$07,$AF,$84,$FC,$53,$18,$7D,$1B,$E9,$1F,$85
	.dc.b	$6B,$E5,$AE,$8F,$A5,$AE,$4E,$67,$D4,$BF,$5B,$E8,$8F,$5C,$FC,$76
	.dc.b	$9E,$68,$1F,$8F,$9F,$5B,$D1,$AE,$AF,$D7,$87,$D7,$AF,$AB,$7E,$8A
	.dc.b	$F9,$47,$91,$E2,$F5,$8F,$03,$4F,$61,$DD,$CD,$76,$8F,$C0,$61,$F9
	.dc.b	$FB,$FD,$2C,$FE,$0A,$B7,$3F,$8C,$7E,$2D,$7E,$40,$E8,$76,$3D,$51
	.dc.b	$F4,$63,$E1,$C7,$D1,$6B,$F3,$4D,$9F,$35,$EA,$FD,$B3,$38,$7B,$AB
	.dc.b	$D1,$87,$C3,$9F,$A4,$7C,$3A,$F3,$BC,$CF,$29,$DC,$F8,$85,$9F,$D5
	.dc.b	$1F,$EA,$F7,$F8,$DB,$5F,$CB,$98,$BF,$E3,$7F,$05,$FE,$B5,$D7,$4D
	.dc.b	$8F,$CE,$C6,$1F,$C0,$BF,$EA,$3F,$B4,$FC,$1C,$60,$FC,$2F,$88,$8C
	.dc.b	$DF,$FB,$13,$77,$D6,$F5,$85,$6D,$C3,$FE,$73,$FB,$4B,$47,$37,$D3
	.dc.b	$1F,$F5,$A4,$E7,$1F,$16,$1D,$38,$1F,$63,$D1,$B3,$AC,$91,$FE,$81
	.dc.b	$8C,$3A,$AF,$F6,$DB,$A8,$CB,$CE,$BA,$3E,$BF,$13,$FD,$6F,$F5,$17
	.dc.b	$C3,$39,$F4,$1F,$71,$98,$D7,$AF,$32,$FF,$5B,$45,$CF,$FD,$85,$EC
	.dc.b	$A7,$B3,$5F,$F5,$E5,$61,$CF,$C0,$47,$E1,$E6,$A7,$B1,$9C,$3A,$FC
	.dc.b	$6F,$43,$1C,$5F,$EC,$CA,$47,$D4,$FB,$57,$D1,$D3,$1F,$39,$73,$F9
	.dc.b	$92,$FB,$5E,$06,$72,$65,$FE,$EB,$9D,$57,$FC,$01,$97,$D8,$FE,$15
	.dc.b	$FE,$1E,$DF,$15,$F0,$39,$9D,$D7,$1F,$B2,$FC,$F4,$7F,$AC,$5A,$FF
	.dc.b	$18,$6A,$6F,$CF,$73,$3A,$36,$7F,$07,$BF,$E2,$4F,$C5,$7C,$95,$8F
	.dc.b	$C3,$E3,$F8,$D3,$18,$7E,$21,$FC,$27,$DA,$7E,$26,$3F,$DF,$8D,$D7
	.dc.b	$FD,$F2,$FE,$46,$33,$D5,$C3,$AE,$61,$F6,$AB,$83,$FC,$CB,$95,$1F
	.dc.b	$97,$C3,$9F,$CA,$F8,$0F,$0F,$C6,$DB,$E5,$E1,$7C,$6C,$38,$B7,$0F
	.dc.b	$DA,$CF,$F8,$8A,$72,$D8,$7D,$A7,$B0,$F6,$D5,$9F,$D6,$79,$DF,$53
	.dc.b	$5F,$BF,$1F,$C1,$23,$19,$FE,$6E,$7F,$80,$C3,$C5,$FE,$02,$EA,$B8
	.dc.b	$71,$60,$CF,$2D,$8F,$DD,$7D,$9F,$E2,$B0,$EE,$7F,$0C,$71,$8F,$7F
	.dc.b	$0D,$1F,$9A,$AE,$3D,$3F,$C4,$23,$F1,$B1,$D3,$A3,$13,$97,$F8,$AF
	.dc.b	$D9,$BE,$E7,$F1,$F8,$23,$17,$FE,$4A,$8C,$4E,$A7,$39,$7F,$94,$27
	.dc.b	$87,$97,$D0,$EB,$56,$FF,$8B,$7A,$AF,$0F,$F9,$2B,$63,$3F,$1D,$89
	.dc.b	$CF,$8B,$27,$8C,$5C,$BE,$2E,$4F,$B3,$FC,$84,$7F,$7C,$3F,$99,$B8
	.dc.b	$7F,$9D,$B5,$CF,$95,$F0,$21,$E9,$FF,$38,$6E,$1F,$CB,$5F,$CE,$71
	.dc.b	$7D,$17,$37,$E2,$FB,$3F,$A4,$FC,$36,$BE,$3C,$51,$88,$FE,$9A,$D7
	.dc.b	$FA,$4A,$F1,$7F,$A4,$B7,$0F,$E7,$CF,$6C,$3F,$04,$E6,$BF,$0F,$F3
	.dc.b	$7F,$96,$AF,$E9,$2D,$8C,$FA,$9F,$CE,$A7,$37,$FC,$C6,$73,$AF,$F5
	.dc.b	$BB,$FF,$49,$36,$74,$7E,$7E,$33,$5D,$BE,$CD,$F7,$1F,$C6,$25,$18
	.dc.b	$9F,$F4,$9C,$62,$75,$39,$C3,$FD,$33,$3C,$3E,$4B,$C5,$26,$D6,$1F
	.dc.b	$E5,$9F,$66,$F2,$29,$8F,$82,$B9,$7F,$6A,$C9,$4F,$27,$F9,$5B,$67
	.dc.b	$FD,$4A,$D9,$FF,$54,$B5,$FF,$56,$FD,$93,$EC,$FF,$9E,$36,$BB,$57
	.dc.b	$1D,$BF,$C3,$13,$9B,$FD,$D8,$FB,$37,$4F,$F4,$56,$E1,$E7,$5F,$13
	.dc.b	$99,$D7,$57,$F4,$29,$FB,$57,$FE,$BE,$6B,$FE,$C2,$6B,$FA,$3F,$2F
	.dc.b	$FB,$0E,$1F,$0D,$B1,$E4,$FB,$3F,$C6,$4E,$FD,$3F,$C1,$7D,$57,$6D
	.dc.b	$7F,$00,$D9,$FD,$77,$D9,$FF,$72,$7F,$1F,$19,$C3,$FC,$DB,$F6,$6F
	.dc.b	$B6,$FE,$C4,$DA,$FA,$9F,$3A,$F8,$7C,$D6,$BE,$43,$FA,$FD,$9A,$4B
	.dc.b	$FC,$5F,$DA,$7E,$51,$FF,$3D,$F6,$A3,$19,$7F,$9F,$B9,$5F,$67,$F9
	.dc.b	$D9,$CF,$81,$FE,$3D,$F8,$DF,$ED,$ED,$C7,$FD,$D2,$DC,$3F,$DD,$AD
	.dc.b	$7F,$87,$3F,$E4,$5F,$B2,$E3,$F0,$86,$92,$BF,$DE,$3C,$3F,$E1,$2F
	.dc.b	$86,$BC,$BF,$EE,$EC,$7F,$DE,$F3,$96,$C6,$5F,$F7,$EC,$7F,$05,$F1
	.dc.b	$F9,$F8,$FF,$E5,$09,$E2,$D7,$6D,$7F,$D3,$F9,$BF,$1A,$7B,$FF,$C9
	.dc.b	$E7,$D7,$18,$7E,$0B,$5F,$82,$D7,$7C,$1C,$5E,$B1,$F3,$9E,$31,$95
	.dc.b	$5E,$18,$C7,$13,$2F,$91,$5F,$8A,$DB,$F9,$D7,$4F,$E4,$1B,$BF,$F3
	.dc.b	$DF,$C5,$FD,$13,$FE,$85,$FB,$4F,$DA,$5F,$76,$BE,$56,$BE,$F6,$BF
	.dc.b	$D1,$DF,$D9,$BF,$C1,$DA,$FE,$E1,$B1,$FE,$93,$9F,$A8,$FE,$93,$EC
	.dc.b	$FC,$B4,$CB,$D9,$18,$6C,$7F,$13,$FB,$3F,$F4,$4B,$FC,$57,$EC,$DF
	.dc.b	$35,$CA,$11,$87,$51,$9F,$B3,$78,$5B,$FB,$37,$D1,$FC,$FB,$46,$1F
	.dc.b	$A4,$D7,$EC,$35,$FC,$13,$FB,$77,$F9,$43,$5F,$E3,$2D,$D3,$FD,$68
	.dc.b	$BE,$83,$D1,$F6,$59,$E1,$FE,$E4,$DB,$FF,$77,$B5,$FF,$39,$FC,$3F
	.dc.b	$BF,$AC,$FA,$5B,$1F,$89,$6B,$F8,$F6,$BF,$99,$7E,$1B,$FC,$FD,$AF
	.dc.b	$F3,$06,$E3,$FE,$CB,$7F,$C8,$31,$FD,$9B,$B3,$FA,$53,$7D,$0F,$B1
	.dc.b	$18,$8F,$ED,$AF,$F4,$97,$87,$9F,$F9,$A8,$C4,$7F,$5A,$6B,$FE,$13
	.dc.b	$E3,$46,$23,$F8,$B9,$EC,$C3,$FB,$63,$73,$CE,$5F,$89,$8F,$4C,$70
	.dc.b	$DF,$04,$8F,$EF,$E7,$B0,$FF,$12,$FD,$87,$F9,$4D,$FF,$63,$35,$CF
	.dc.b	$C2,$7E,$63,$C0,$4F,$FA,$62,$77,$F0,$5F,$B9,$9C,$B6,$75,$3F,$E4
	.dc.b	$16,$BF,$E1,$FF,$AF,$FF,$0D,$3F,$ED,$56,$BF,$E3,$86,$E7,$FE,$5B
	.dc.b	$7F,$D6,$71,$BF,$8E,$F0,$46,$23,$FC,$9C,$D7,$B4,$79,$7E,$B9,$F5
	.dc.b	$3E,$B6,$CF,$FB,$95,$AF,$F9,$21,$F9,$CF,$FB,$E1,$B9,$1A,$4D,$3E
	.dc.b	$37,$D6,$BE,$4E,$92,$FF,$DE,$51,$FE,$CE,$6F,$2F,$C9,$8C,$47,$F9
	.dc.b	$9D,$AF,$F9,$15,$A8,$60,$FA,$DF,$E2,$3C,$39,$8C,$7A,$DF,$66,$73
	.dc.b	$5C,$1F,$F0,$4B,$CF,$7F,$F4,$44,$62,$3F,$CD,$2D,$8C,$FA,$FC,$FD
	.dc.b	$7F,$C2,$7D,$27,$C8,$9F,$F4,$83,$72,$F1,$37,$0F,$3B,$FE,$F5,$7C
	.dc.b	$11,$AF,$93,$E6,$48,$C9,$C3,$9F,$F9,$4F,$C8,$BE,$B3,$FD,$0D,$3F
	.dc.b	$F0,$01,$AF,$FA,$92,$35,$F0,$5C,$3E,$63,$FD,$65,$39,$7F,$DC,$F2
	.dc.b	$69,$31,$FE,$A9,$19,$F9,$7E,$44,$AF,$2C,$71,$E4,$ED,$E2,$49,$CE
	.dc.b	$1F,$CA,$F6,$FF,$61,$5F,$7F,$AC,$F2,$12,$B6,$F9,$FD,$AF,$D8,$7D
	.dc.b	$13,$E4,$E9,$2B,$E2,$F3,$A5,$F5,$5C,$09,$3A,$8C,$B7,$EE,$7F,$A5
	.dc.b	$7D,$13,$EA,$E9,$27,$FC,$C9,$1F,$D3,$5F,$F2,$AB,$5F,$F0,$23,$5F
	.dc.b	$F5,$63,$5F,$CB,$35,$2F,$AB,$F0,$92,$3F,$CE,$5F,$84,$FF,$78,$E4
	.dc.b	$FC,$D7,$D6,$7E,$45,$BD,$9F,$C4,$E3,$15,$FA,$27,$FB,$DF,$17,$FB
	.dc.b	$85,$B5,$CB,$E2,$B8,$1C,$7B,$FF,$17,$F4,$4F,$A4,$A4,$C7,$EE,$C3
	.dc.b	$EB,$36,$48,$CB,$CB,$DD,$E7,$F5,$72,$78,$BD,$BE,$12,$47,$46,$BF
	.dc.b	$97,$B7,$D5,$F9,$09,$1E,$27,$F3,$5F,$15,$F8,$4F,$98,$93,$F7,$5F
	.dc.b	$C3,$C4,$8F,$C0,$35,$FF,$71,$F3,$BE,$9B,$F9,$E7,$CC,$8A,$49,$FC
	.dc.b	$6A,$FA,$5D,$92,$32,$F2,$FA,$5F,$C5,$FD,$54,$9F,$C4,$BF,$8E,$F9
	.dc.b	$C9,$1F,$E3,$FE,$BF,$F7,$5B,$63,$FC,$29,$18,$8F,$F2,$43,$E5,$E8
	.dc.b	$FE,$41,$F5,$3F,$5E,$7F,$1B,$5F,$EF,$B8,$C4,$7E,$3A,$31,$CB,$F9
	.dc.b	$4C,$D7,$CA,$30,$DA,$E6,$53,$91,$70,$F1,$2C,$3B,$C6,$3E,$7E,$E3
	.dc.b	$FC,$D6,$91,$87,$F9,$C3,$C3,$1F,$1C,$6B,$1F,$1B,$D5,$FD,$5D,$7A
	.dc.b	$A3,$CA,$64,$FD,$6C,$F2,$C3,$A6,$3E,$7A,$E1,$5E,$A3,$81,$FB,$2A
	.dc.b	$E1,$7E,$0C,$3C,$D3,$E3,$3B,$BD,$1C,$1C,$6F,$B3,$0E,$B8,$FC,$6C
	.dc.b	$78,$1D,$09,$7E,$E7,$A5,$F8,$09,$1E,$33,$47,$C2,$AE,$D3,$D2,$33
	.dc.b	$1F,$12,$3A,$BF,$12,$FE,$C6,$1F,$3A,$FA,$CF,$B2,$FB,$3B,$3D,$38
	.dc.b	$FC,$8C,$7F,$25,$8F,$79,$D3,$D9,$F9,$8E,$5F,$88,$7C,$68,$FC,$D5
	.dc.b	$F8,$23,$2E,$E3,$AD,$FB,$67,$E4,$E1,$F5,$1E,$F1,$9B,$F0,$57,$CD
	.dc.b	$8F,$4C,$FD,$DB,$ED,$3F,$93,$9F,$8B,$3E,$67,$91,$88,$E5,$1F,$26
	.dc.b	$32,$F8,$0F,$89,$3D,$30,$FC,$E3,$F6,$27,$94,$7C,$73,$E4,$61,$E9
	.dc.b	$AF,$87,$87,$A2,$73,$8F,$D4,$3F,$47,$7F,$07,$0F,$4E,$1F,$5E,$3A
	.dc.b	$E1,$F8,$07,$D1,$1D,$71,$F9,$33,$E9,$3D,$07,$82,$7D,$6F,$C1,$33
	.dc.b	$1F,$01,$C1,$1C,$13,$0F,$64,$79,$CF,$AB,$87,$8D,$F0,$1D,$CF,$8F
	.dc.b	$5E,$B2,$38,$3C,$33,$D1,$E6,$4C,$7E,$33,$DD,$FD,$83,$FD,$BB,$0F
	.dc.b	$E0,$ED,$BF,$E5,$51,$88,$FE,$E3,$A7,$FB,$A9,$FE,$52,$D7,$FB,$BB
	.dc.b	$5D,$FC,$5F,$FB,$0D,$7D,$55,$23,$FE,$C7,$6E,$9F,$CB,$A3,$FD,$2F
	.dc.b	$D7,$F6,$9F,$E4,$07,$23,$F8,$39,$A2,$99,$8E,$5E,$EA,$79,$4F,$F2
	.dc.b	$9A,$AF,$1A,$F0,$3C,$75,$EF,$7C,$63,$AD,$70,$8C,$57,$08,$C5,$F4
	.dc.b	$AE,$47,$F3,$7A,$DD,$FA,$A3,$27,$F9,$BE,$32,$FB,$AB,$9B,$E0,$3D
	.dc.b	$F1,$A7,$D4,$62,$BC,$13,$C2,$FC,$93,$93,$93,$F1,$27,$71,$98,$D9
	.dc.b	$97,$77,$F0,$A7,$73,$93,$71,$99,$D9,$EF,$9D,$C6,$67,$73,$93,$67
	.dc.b	$C1,$9D,$B9,$8D,$B9,$36,$7D,$79,$F9,$55,$EF,$6C,$7A,$DB,$3E,$76
	.dc.b	$9E,$32,$C7,$76,$BD,$E3,$C2,$F7,$3E,$A1,$D4,$AF,$56,$C7,$56,$9D
	.dc.b	$4A,$75,$7B,$BD,$4F,$B7,$1C,$DB,$1C,$DA,$73,$29,$CC,$EA,$73,$3F
	.dc.b	$D7,$53,$F1,$5C,$DF,$E2,$5C,$9E,$B3,$99,$C4,$FC,$A9,$63,$93,$A8
	.dc.b	$E4,$E9,$B1,$A8,$EB,$5E,$B6,$D7,$AD,$A7,$AC,$A7,$AC,$ED,$7F,$84
	.dc.b	$70,$68,$F2,$1E,$93,$17,$E2,$70,$79,$63,$93,$C2,$BC,$8E,$4F,$D3
	.dc.b	$1F,$B5,$9F,$F5,$DB,$F9,$D3,$13,$FB,$9F,$13,$FC,$B9,$F4,$DA,$9F
	.dc.b	$F7,$3E,$88,$C4,$EA,$B8,$4E,$5E,$5F,$23,$FE,$EF,$6D,$F0,$8C,$3F
	.dc.b	$C8,$29,$18,$8F,$F2,$14,$7F,$9A,$6B,$BE,$BF,$E6,$FB,$F3,$36,$B9
	.dc.b	$35,$E2,$B8,$F8,$92,$78,$BE,$DF,$2B,$FD,$B5,$CB,$7C,$BF,$F3,$F5
	.dc.b	$7F,$1B,$9C,$B6,$33,$87,$F3,$89,$75,$BD,$4E,$8D,$BE,$4D,$AE,$0D
	.dc.b	$7F,$CC,$15,$CB,$D0,$E2,$4A,$E2,$F9,$FD,$37,$A3,$49,$3E,$78,$CB
	.dc.b	$7D,$1E,$04,$70,$7D,$FF,$A9,$7C,$97,$DC,$6C,$66,$30,$67,$DD,$91
	.dc.b	$C5,$B9,$FF,$41,$F8,$37,$D2,$7E,$26,$11,$8A,$FE,$8D,$18,$9D,$4E
	.dc.b	$63,$3F,$05,$C5,$34,$95,$CA,$1F,$4F,$B2,$46,$5F,$C4,$FC,$34,$8F
	.dc.b	$C4,$B6,$7F,$13,$AF,$E3,$89,$3F,$8E,$7F,$0D,$F5,$3F,$8D,$90,$E1
	.dc.b	$46,$63,$0D,$8C,$F4,$72,$E8,$98,$7E,$36,$BF,$D8,$CD,$8E,$FE,$E4
	.dc.b	$8D,$37,$0F,$C8,$53,$E0,$7C,$A4,$AE,$8F,$A3,$B6,$47,$76,$D7,$EC
	.dc.b	$7D,$97,$C3,$7E,$F1,$B5,$FC,$0D,$B3,$B9,$D7,$04,$AD,$36,$7F,$79
	.dc.b	$F5,$0E,$4F,$E0,$ED,$AF,$E1,$CD,$9D,$CE,$BE,$A2,$56,$9B,$3F,$C1
	.dc.b	$FE,$A1,$E0,$FF,$12,$6D,$7F,$19,$6C,$EE,$75,$F5,$12,$B4,$D9,$FE
	.dc.b	$25,$E8,$3D,$6F,$E3,$AD,$AF,$E4,$8D,$9D,$CE,$BB,$E5,$69,$B3,$FC
	.dc.b	$77,$EA,$1E,$87,$F2,$86,$D7,$F2,$F6,$CE,$E7,$5C,$F2,$B4,$D9,$FE
	.dc.b	$51,$F5,$0F,$88,$FE,$66,$DA,$FE,$74,$D9,$DC,$EB,$82,$56,$9B,$3F
	.dc.b	$E5,$2F,$8D,$79,$FF,$C5,$63,$F3,$F1,$8F,$1F,$FA,$23,$6F,$FC,$80
	.dc.b	$D9,$E5,$E8,$7F,$95,$67,$2D,$8C,$C7,$28,$C4,$7F,$C0,$19,$7D,$43
	.dc.b	$E5,$BB,$B6,$FB,$B6,$BD,$0D,$79,$2F,$A8,$E2,$49,$E2,$FF,$15,$FA
	.dc.b	$67,$D3,$D2,$4F,$F1,$68,$CB,$7C,$5E,$04,$70,$6F,$C7,$E8,$47,$07
	.dc.b	$F2,$0F,$A7,$7C,$1F,$F2,$56,$C6,$63,$06,$7A,$24,$71,$6E,$9F,$E6
	.dc.b	$FF,$A7,$FD,$53,$FE,$8A,$6B,$FC,$58,$38,$7F,$8C,$B5,$7D,$3F,$F4
	.dc.b	$87,$FC,$F4,$D7,$FD,$2A,$D7,$A3,$5F,$E4,$7F,$4E,$F7,$FF,$A0,$37
	.dc.b	$0F,$E7,$ED,$9F,$F4,$33,$5F,$F5,$0F,$D3,$BE,$67,$FA,$43,$6B,$FA
	.dc.b	$3B,$67,$FD,$1C,$D7,$FD,$53,$F4,$CF,$86,$FE,$A0,$DA,$FE,$9E,$D9
	.dc.b	$FF,$4B,$35,$FF,$58,$FE,$3D,$E3,$7E,$43,$5F,$E7,$ED,$4D,$3F,$C0
	.dc.b	$6F,$2F,$6B,$FB,$04,$62,$39,$C7,$F6,$3F,$A4,$FD,$C7,$BD,$FD,$89
	.dc.b	$B3,$FE,$A9,$6B,$FE,$C4,$6B,$F6,$DF,$EE,$FE,$3F,$1F,$5F,$F9,$1C
	.dc.b	$F1,$6B,$C5,$AF,$E0,$FF,$1B,$7C,$47,$F6,$96,$C7,$F7,$46,$BF,$EB
	.dc.b	$B6,$BF,$ED,$65,$F1,$3C,$49,$1C,$5F,$EE,$9F,$47,$FE,$E5,$24,$60
	.dc.b	$FE,$F1,$F4,$9F,$A2,$7F,$C0,$6D,$7F,$D9,$4D,$7F,$DC,$AD,$7F,$39
	.dc.b	$0C,$FB,$92,$3F,$C1,$B1,$97,$FC,$1F,$F4,$9F,$E5,$96,$FA,$BF,$E0
	.dc.b	$98,$C4,$7F,$B6,$5A,$FF,$BD,$3E,$8B,$FB,$73,$FE,$18,$6B,$FE,$DF
	.dc.b	$6B,$BF,$5F,$FC,$29,$F4,$4F,$97,$A4,$9F,$F1,$0C,$65,$BD,$1C,$08
	.dc.b	$E0,$FF,$13,$F9,$8F,$6F,$FC,$40,$D8,$CC,$60,$CE,$12,$38,$B5,$FF
	.dc.b	$7B,$4B,$EA,$F9,$92,$3B,$BF,$E3,$6F,$2B,$E7,$78,$97,$83,$FC,$77
	.dc.b	$5F,$D0,$1A,$FF,$49,$6B,$FC,$ED,$7B,$FC,$49,$1C,$5E,$CF,$CA,$3F
	.dc.b	$A7,$94,$57,$F8,$C3,$1D,$9C,$30,$93,$FD,$48,$F6,$15,$FF,$27,$F3
	.dc.b	$BE,$8B,$FC,$97,$3F,$E9,$F6,$BF,$E9,$36,$BF,$E9,$26,$BF,$E4,$08
	.dc.b	$E5,$F0,$3D,$27,$FC,$A3,$3C,$E3,$07,$83,$E8,$BF,$C4,$51,$FE,$AD
	.dc.b	$6B,$FE,$A1,$6B,$FE,$A0,$6B,$DD,$7B,$7C,$C9,$1C,$DF,$6F,$D0,$FE
	.dc.b	$0E,$3F,$D5,$8D,$7F,$D5,$6D,$7D,$4D,$7F,$C5,$6F,$CF,$C0,$D2,$65
	.dc.b	$F2,$BE,$87,$FC,$21,$5F,$EC,$76,$BF,$EB,$B6,$BF,$EB,$A6,$BF,$25
	.dc.b	$FF,$47,$BC,$7B,$7A,$11,$FF,$69,$7D,$0F,$EF,$DF,$F1,$FB,$70,$34
	.dc.b	$9E,$A7,$DE,$7F,$D6,$53,$CA,$30,$7E,$0B,$E8,$7F,$27,$1F,$ED,$F6
	.dc.b	$BF,$ED,$36,$BF,$ED,$26,$BF,$81,$7E,$45,$1A,$4E,$5F,$F4,$EF,$C3
	.dc.b	$3E,$33,$FD,$44,$DF,$1F,$31,$82,$7C,$5E,$09,$8F,$FB,$3D,$DF,$C5
	.dc.b	$C8,$FF,$69,$35,$FF,$69,$4B,$E7,$FA,$92,$3F,$DB,$0F,$83,$BA,$47
	.dc.b	$FB,$6D,$AF,$FB,$CF,$E5,$9F,$3F,$FA,$86,$C7,$F8,$2C,$FD,$13,$87
	.dc.b	$84,$6F,$E7,$E4,$6C,$FF,$3B,$CA,$33,$19,$F9,$87,$D2,$7F,$AE,$DB
	.dc.b	$1F,$EC,$66,$BB,$8D,$7C,$FC,$8D,$35,$F0,$7C,$FF,$FB,$15,$34,$8F
	.dc.b	$F0,$4F,$9F,$7C,$FD,$24,$7A,$79,$53,$A8,$C3,$FE,$87,$6F,$B5,$CC
	.dc.b	$8F,$FC,$00,$F9,$FF,$C1,$FC,$B9,$A4,$8F,$8D,$CC,$9F,$F0,$D4,$61
	.dc.b	$AF,$D1,$8D,$7A,$FD,$0E,$BE,$9C,$9E,$0D,$78,$53,$E7,$FC,$64,$8E
	.dc.b	$CF,$E5,$3D,$B9,$1E,$4E,$7F,$F1,$E3,$63,$F0,$84,$7E,$37,$26,$9F
	.dc.b	$C5,$47,$E2,$7D,$79,$1A,$6B,$F8,$BD,$DF,$14,$F9,$6F,$E6,$ED,$9F
	.dc.b	$A4,$D7,$83,$5F,$F2,$F4,$F2,$F9,$9E,$24,$8E,$2F,$E3,$7E,$7E,$47
	.dc.b	$E3,$5A,$FE,$4E,$9F,$15,$FE,$FE,$F9,$C7,$CE,$FF,$9A,$FA,$D8,$7F
	.dc.b	$9B,$67,$38,$04,$E9,$E5,$5C,$70,$EB,$0A,$E2,$F0,$8C,$C6,$2B,$94
	.dc.b	$EE,$BC,$57,$E1,$7A,$61,$DA,$7C,$55,$F3,$23,$11,$E4,$AE,$0E,$9A
	.dc.b	$F8,$63,$73,$C6,$3C,$06,$CC,$CF,$08,$D5,$F9,$F0,$F4,$E1,$E6,$BD
	.dc.b	$9D,$23,$2F,$99,$AF,$94,$D9,$E0,$3C,$85,$AE,$F3,$D6,$31,$7E,$3A
	.dc.b	$E7,$5A,$9F,$41,$A3,$C5,$84,$70,$4C,$3D,$51,$97,$E9,$5F,$BC,$ED
	.dc.b	$19,$C3,$C5,$3A,$AC,$1E,$D9,$F5,$C6,$63,$D5,$87,$98,$F3,$DF,$58
	.dc.b	$E2,$7C,$09,$E0,$4B,$FB,$F1,$CA,$FA,$C7,$88,$CC,$FA,$A7,$D9,$19
	.dc.b	$7C,$87,$13,$27,$C6,$8E,$04,$8E,$11,$F5,$4E,$11,$97,$84,$6B,$0F
	.dc.b	$A1,$1E,$98,$C5,$FC,$88,$E3,$18,$7D,$F3,$C0,$FC,$FD,$7B,$E3,$CE
	.dc.b	$F7,$7C,$4E,$5E,$C6,$A7,$A3,$D8,$C1,$E6,$21,$C0,$93,$E0,$8F,$56
	.dc.b	$1F,$57,$0F,$9B,$5C,$2B,$CD,$3F,$62,$32,$72,$70,$FC,$C3,$13,$A9
	.dc.b	$ED,$1E,$43,$C3,$87,$E0,$5F,$41,$83,$85,$78,$DE,$93,$C9,$EA,$62
	.dc.b	$7E,$76,$1F,$81,$8F,$EF,$13,$FB,$B6,$D7,$F1,$26,$DF,$F6,$FC,$DF
	.dc.b	$1B,$49,$3F,$DD,$E3,$2F,$F7,$D3,$F9,$07,$CF,$7F,$6F,$23,$90,$E3
	.dc.b	$E7,$FF,$8F,$C7,$FA,$E7,$C5,$FE,$51,$18,$8D,$87,$7E,$AA,$77,$B7
	.dc.b	$F8,$36,$ED,$FF,$31,$3A,$1C,$63,$4D,$35,$5C,$DC,$1C,$67,$F9,$3B
	.dc.b	$CB,$5F,$24,$62,$78,$46,$27,$89,$E0,$C3,$07,$A3,$0F,$2C,$75,$37
	.dc.b	$5E,$19,$DE,$1F,$E2,$D8,$C3,$E8,$8E,$53,$D8,$F8,$AE,$0E,$2F,$B2
	.dc.b	$BB,$BF,$50,$EE,$E1,$E3,$3D,$E3,$18,$70,$BF,$0D,$7F,$8F,$E3,$13
	.dc.b	$A8,$E6,$78,$E3,$E2,$47,$C8,$8F,$B3,$87,$A0,$F2,$BF,$D9,$6B,$AB
	.dc.b	$F0,$F0,$F5,$61,$EF,$3C,$91,$D4,$EF,$1F,$3E,$BA,$4F,$DC,$6C,$F9
	.dc.b	$CD,$C7,$37,$E2,$1C,$E7,$BB,$B6,$BF,$56,$79,$B6,$7C,$71,$EA,$8F
	.dc.b	$BD,$1D,$5F,$65,$FC,$5A,$E7,$1F,$45,$AF,$C4,$9E,$67,$A4,$AF,$8D
	.dc.b	$F8,$31,$DE,$FE,$3C,$F3,$9C,$9E,$17,$C6,$7C,$1B,$F4,$BF,$81,$6B
	.dc.b	$F8,$03,$E1,$D7,$5B,$F4,$57,$49,$F1,$1A,$7C,$CF,$CA,$AF,$A5,$1E
	.dc.b	$88,$FB,$4F,$16,$9C,$4E,$67,$C6,$3E,$93,$B7,$CC,$F3,$69,$DD,$F7
	.dc.b	$5F,$C2,$8F,$51,$CE,$FD,$4F,$43,$C0,$7D,$93,$91,$F1,$67,$B3,$C8
	.dc.b	$F9,$27,$A0,$E8,$F8,$2F,$E2,$C7,$43,$CA,$7A,$EF,$E1,$BD,$8E,$47
	.dc.b	$43,$EC,$1C,$4F,$05,$FA,$F0,$FD,$6C,$7A,$4F,$09,$F2,$0E,$47,$0A
	.dc.b	$F1,$BA,$3B,$CF,$37,$F4,$27,$17,$18,$7E,$4A,$BB,$BF,$20,$98,$7B
	.dc.b	$EB,$C4,$78,$27,$71,$99,$DC,$F5,$BF,$1B,$B3,$C2,$79,$1C,$DF,$F2
	.dc.b	$FC,$FF,$98,$4E,$4F,$F3,$7C,$FF,$B3,$7E,$6B,$31,$63,$FA,$54,$62
	.dc.b	$3F,$9F,$35,$7A,$EF,$63,$B5,$7F,$43,$8C,$3F,$E7,$5B,$EB,$7A,$BF
	.dc.b	$F3,$C5,$69,$F1,$8E,$3F,$F4,$A6,$DB,$6E,$1C,$CF,$F4,$03,$FE,$99
	.dc.b	$E3,$D4,$FF,$A6,$E7,$55,$CF,$EB,$FF,$DD,$B1,$8B,$46,$27,$C3,$1D
	.dc.b	$9B,$19,$CB,$FA,$C7,$6B,$D9,$79,$5E,$B8,$C6,$D6,$73,$97,$8A,$72
	.dc.b	$D8,$F8,$17,$D3,$CD,$7C,$97,$1C,$3A,$46,$2F,$D7,$9F,$FA,$7F,$E6
	.dc.b	$FE,$53,$FE,$A1,$F6,$2A,$8B,$18,$FA,$9F,$B1,$26,$92,$BF,$C1,$9C
	.dc.b	$4F,$53,$AB,$63,$C3,$5C,$5F,$1C,$66,$3E,$CE,$CF,$9E,$FA,$51,$85
	.dc.b	$C9,$F1,$DC,$EF,$F3,$1C,$E1,$F1,$DB,$5D,$3E,$95,$F4,$9C,$70,$FA
	.dc.b	$11,$8B,$FB,$9F,$41,$F0,$5B,$B7,$E2,$63,$18,$7D,$E5,$19,$3F,$1D
	.dc.b	$F5,$8F,$53,$17,$F8,$78,$C3,$F8,$CE,$B7,$DB,$FF,$B4,$63,$FC,$45
	.dc.b	$8F,$F6,$F8,$C3,$FE,$0D,$6B,$FE,$22,$6B,$F9,$AF,$6D,$F0,$DF,$9F
	.dc.b	$6D,$77,$AE,$7C,$1F,$E2,$9E,$C5,$6B,$0F,$D4,$E2,$FD,$F9,$F5,$FF
	.dc.b	$AF,$8C,$3F,$AB,$FB,$DF,$87,$8F,$F8,$CE,$73,$9F,$E9,$7E,$89,$F3
	.dc.b	$DF,$B1,$6C,$7C,$1C,$7A,$BB,$9E,$D0,$F8,$CF,$E1,$17,$E3,$6C,$66
	.dc.b	$BE,$3E,$6A,$D4,$63,$57,$D0,$BE,$8F,$F8,$13,$67,$C5,$8E,$27,$C3
	.dc.b	$F1,$5D,$89,$3D,$B0,$F8,$91,$FC,$6F,$07,$C5,$3E,$8F,$F8,$73,$6F
	.dc.b	$DD,$58,$9F,$6F,$C4,$49,$FE,$1B,$F1,$8F,$B9,$7C,$9E,$77,$FE,$37
	.dc.b	$3F,$E7,$CE,$9F,$E4,$0D,$BF,$95,$86,$3D,$25,$7C,$AF,$89,$F8,$E4
	.dc.b	$AF,$57,$1A,$3F,$9A,$60,$FA,$2F,$0D,$79,$1B,$97,$E1,$2F,$84,$E7
	.dc.b	$3F,$E1,$3F,$A5,$FD,$AA,$FE,$62,$3F,$53,$FC,$7D,$7F,$9F,$67,$5D
	.dc.b	$3F,$D2,$A7,$35,$FC,$DF,$EF,$BF,$81,$3F,$E8,$3E,$44,$62,$3F,$9B
	.dc.b	$B5,$E1,$E4,$FD,$3F,$C9,$1F,$01,$FC,$FD,$B9,$FF,$37,$BF,$E1,$D9
	.dc.b	$7E,$86,$73,$87,$A3,$E9,$1F,$49,$FC,$91,$FE,$AB,$87,$F2,$87,$2D
	.dc.b	$F3,$7F,$94,$FC,$F3,$E2,$BF,$AE,$36,$7F,$D5,$CD,$C3,$F2,$ED,$7F
	.dc.b	$D4,$1C,$8F,$9E,$F9,$FA,$7F,$A7,$63,$1C,$7F,$D1,$7C,$6F,$8F,$18
	.dc.b	$C7,$F1,$3F,$9D,$7F,$4D,$58,$A5,$39,$8C,$2F,$90,$FC,$5B,$98,$7D
	.dc.b	$2B,$E9,$7F,$A2,$AF,$8E,$B1,$87,$FA,$4F,$D2,$FF,$4D,$AF,$D9,$B5
	.dc.b	$FE,$36,$D7,$FD,$7C,$D7,$CD,$F4,$5D,$73,$FD,$1B,$63,$F4,$2B,$E6
	.dc.b	$3F,$42,$E7,$7F,$E2,$1E,$7B,$E0,$BF,$83,$B7,$0F,$DC,$37,$0F,$E5
	.dc.b	$53,$FB,$9E,$97,$9D,$49,$7F,$C2,$63,$2F,$FB,$9F,$1F,$F0,$B7,$95
	.dc.b	$7D,$DF,$E2,$18,$FF,$12,$9C,$B6,$32,$FF,$3C,$8F,$EC,$9D,$FC,$FB
	.dc.b	$FF,$C2,$D3,$C5,$AE,$DA,$FF,$85,$BE,$B3,$FB,$83,$FC,$75,$AF,$F4
	.dc.b	$89,$46,$23,$F5,$4D,$7F,$A6,$61,$FF,$00,$4F,$EA,$25,$D7,$CF,$1F
	.dc.b	$D3,$63,$F9,$6B,$F5,$BE,$FF,$FD,$ED,$BF,$E5,$72,$46,$34,$FC,$8F
	.dc.b	$3A,$73,$7F,$73,$D1,$D3,$F9,$6F,$A1,$79,$7D,$A1,$18,$7C,$2D,$7F
	.dc.b	$9F,$B5,$FE,$CF,$F4,$2F,$A1,$EB,$08,$FE,$E4,$FF,$88,$1A,$FF,$16
	.dc.b	$78,$3F,$AA,$FB,$AF,$2E,$5F,$A8,$6F,$61,$BE,$DE,$BC,$17,$CA,$D2
	.dc.b	$61,$FE,$37,$E3,$7C,$BF,$93,$EC,$FC,$33,$FD,$F9,$AF,$EE,$03,$87
	.dc.b	$F7,$6D,$7F,$CC,$DB,$FB,$BA,$7F,$9D,$47,$F8,$F1,$AF,$F3,$27,$E4
	.dc.b	$72,$FF,$7F,$FC,$63,$FB,$64,$FF,$96,$5A,$FF,$96,$1B,$1F,$E0,$5E
	.dc.b	$8F,$E0,$13,$FE,$5F,$6B,$F8,$F6,$BF,$8E,$F0,$3F,$0F,$D5,$F9,$27
	.dc.b	$EE,$7E,$B4,$F8,$DF,$AA,$BE,$4A,$98,$8F,$A3,$DB,$E2,$EC,$FC,$5B
	.dc.b	$71,$FE,$38,$D3,$FC,$2D,$F0,$7F,$8C,$6C,$7F,$5E,$8F,$CF,$7A,$66
	.dc.b	$9F,$9C,$E9,$7A,$FF,$E5,$FA,$FE,$E6,$D7,$FB,$8B,$77,$FF,$0F,$5F
	.dc.b	$F9,$E6,$7F,$C3,$5F,$19,$E1,$8C,$47,$F6,$78,$FC,$5F,$B3,$C7,$3F
	.dc.b	$F4,$7F,$B9,$F8,$8A,$FF,$45,$35,$FF,$44,$B7,$0E,$FE,$8F,$53,$D6
	.dc.b	$49,$FE,$F6,$FD,$0F,$A5,$FC,$DC,$7F,$A4,$9B,$3F,$E3,$F6,$DF,$F4
	.dc.b	$6D,$BF,$C7,$DE,$A7,$A4,$F8,$E4,$BF,$F0,$3B,$F7,$3C,$9F,$E5,$D1
	.dc.b	$FE,$3D,$F2,$BF,$77,$CD,$FC,$A9,$F8,$5F,$59,$27,$FC,$7C,$D7,$FD
	.dc.b	$4B,$F7,$FF,$94,$CB,$FC,$B1,$F1,$3C,$BD,$3F,$F3,$AD,$7F,$AB,$5A
	.dc.b	$FF,$84,$7E,$27,$E7,$1F,$70,$8F,$F8,$71,$FC,$8F,$C7,$7D,$F8,$FF
	.dc.b	$38,$76,$F9,$BE,$03,$F9,$2C,$7F,$AE,$67,$4F,$FA,$E1,$FB,$3C,$5F
	.dc.b	$E6,$57,$2D,$EA,$F7,$C6,$B3,$FE,$DF,$F1,$8F,$9B,$F3,$DF,$FA,$FA
	.dc.b	$DF,$47,$FC,$82,$BF,$24,$DA,$DD,$F2,$77,$8F,$FA,$FD,$CA,$AE,$DB
	.dc.b	$71,$BF,$E3,$FC,$FF,$72,$BF,$DC,$7E,$76,$0F,$E4,$7F,$30,$FB,$5A
	.dc.b	$7F,$2F,$71,$3F,$CF,$9A,$FF,$9B,$1A,$FF,$A5,$17,$D1,$74,$FA,$39
	.dc.b	$3F,$EC,$C9,$CB,$FE,$D7,$FA,$07,$CD,$7C,$1E,$2F,$F6,$FC,$FF,$B6
	.dc.b	$DA,$FF,$B0,$27,$6F,$1F,$37,$E4,$46,$23,$55,$89,$FC,$8F,$CC,$BD
	.dc.b	$2A,$4A,$FF,$7B,$7D,$07,$F7,$E6,$F7,$FE,$62,$31,$1F,$EF,$86,$F3
	.dc.b	$FF,$A9,$57,$D0,$7C,$EB,$95,$FE,$D8,$7E,$6F,$D8,$C8,$F9,$4F,$CD
	.dc.b	$23,$F3,$7C,$C6,$1F,$E3,$6F,$9D,$F6,$70,$3E,$83,$EB,$9F,$62,$B7
	.dc.b	$6A,$DC,$61,$E3,$3F,$D2,$5B,$3D,$77,$FE,$99,$7F,$CB,$F0,$C4,$FF
	.dc.b	$91,$A3,$0F,$59,$DD,$F3,$6C,$70,$13,$C2,$FC,$77,$98,$F5,$4D,$BE
	.dc.b	$4F,$7C,$3D,$18,$79,$1B,$1E,$39,$DB,$D9,$B5,$DE,$B8,$5F,$5A,$87
	.dc.b	$A4,$EC,$F8,$EF,$AC,$73,$37,$87,$A1,$E0,$7B,$23,$77,$EA,$8C,$B6
	.dc.b	$78,$12,$FE,$2C,$EB,$0F,$3D,$F2,$39,$BE,$A3,$38,$79,$4F,$1D,$72
	.dc.b	$BE,$97,$EA,$3A,$1A,$30,$E9,$F6,$1A,$3C,$F7,$E1,$AF,$F4,$1F,$D0
	.dc.b	$7E,$12,$4D,$3E,$11,$CF,$1F,$76,$1E,$33,$E7,$CF,$4A,$F1,$3E,$0C
	.dc.b	$38,$61,$F9,$6C,$3A,$5F,$C4,$7A,$BC,$A7,$C4,$47,$04,$AF,$C1,$1D
	.dc.b	$B0,$FC,$03,$F0,$EB,$AB,$A3,$D7,$3E,$68,$F5,$63,$F3,$E7,$E7,$1E
	.dc.b	$93,$E3,$5F,$85,$F9,$75,$C4,$F4,$11,$F0,$3E,$57,$E0,$3F,$15,$E2
	.dc.b	$43,$27,$96,$3E,$2E,$3F,$4A,$7A,$4F,$D8,$BF,$8A,$F1,$C3,$E8,$BC
	.dc.b	$67,$C1,$1F,$0A,$3A,$91,$C1,$FB,$A9,$F8,$B1,$D6,$7D,$8F,$5C,$3E
	.dc.b	$23,$F0,$F0,$F8,$EF,$CF,$C7,$F1,$0F,$CC,$3E,$75,$F4,$C3,$B6,$1E
	.dc.b	$4C,$39,$E1,$EE,$BF,$9B,$3F,$15,$FD,$66,$3F,$4B,$0E,$57,$FE,$5F
	.dc.b	$7E,$0C,$E8,$E2,$7C,$13,$E2,$63,$FA,$A7,$CD,$5C,$EF,$E5,$E3,$FB
	.dc.b	$7B,$E6,$FC,$4C,$3C,$B3,$F2,$B0,$E5,$87,$63,$C1,$8F,$C2,$71,$7F
	.dc.b	$32,$77,$3F,$87,$9D,$D7,$C1,$7C,$67,$CB,$7E,$05,$7D,$A8,$D9,$F8
	.dc.b	$A3,$E5,$C7,$A2,$BF,$50,$F3,$8F,$0C,$7D,$2A,$F0,$57,$33,$C9,$8F
	.dc.b	$F4,$B2,$57,$F7,$7A,$FE,$8E,$FF,$77,$F3,$FF,$CB,$AF,$F7,$B6,$B9
	.dc.b	$F3,$BF,$C0,$11,$88,$FF,$49,$01,$F4,$1F,$EB,$56,$9F,$EB,$77,$04
	.dc.b	$7F,$95,$8F,$29,$E0,$7F,$49,$8E,$93,$67,$31,$C0,$EB,$9F,$F7,$F7
	.dc.b	$95,$71,$73,$A7,$9A,$B8,$D6,$E7,$CB,$B7,$19,$EA,$E7,$0F,$D9,$4F
	.dc.b	$FA,$BE,$73,$7F,$02,$3A,$99,$C3,$C1,$5E,$76,$C7,$BA,$BC,$35,$C6
	.dc.b	$B5,$3E,$B6,$E1,$E1,$8E,$17,$CA,$3C,$CD,$AF,$63,$67,$C4,$73,$38
	.dc.b	$96,$78,$B6,$78,$BB,$C3,$27,$84,$B1,$E7,$8E,$73,$E5,$39,$96,$39
	.dc.b	$B6,$3C,$6F,$38,$EF,$36,$7C,$EE,$DF,$F4,$75,$7A,$DA,$79,$CE,$73
	.dc.b	$E3,$AF,$0B,$E5,$AF,$3C,$FE,$52,$FE,$CB,$CA,$3B,$3F,$86,$3A,$1D
	.dc.b	$AB,$38,$7D,$68,$E4,$7C,$0B,$EC,$7A,$8A,$FA,$9F,$21,$F0,$8A,$FA
	.dc.b	$0E,$73,$E8,$6C,$6D,$EF,$5E,$A7,$FD,$FD,$7E,$A7,$D1,$1E,$6C,$3E
	.dc.b	$34,$7A,$EF,$0F,$B0,$C5,$F9,$9F,$84,$D7,$9B,$E6,$7E,$13,$5E,$27
	.dc.b	$98,$F5,$1E,$43,$E2,$1E,$87,$E2,$34,$F8,$87,$13,$A9,$4E,$A7,$38
	.dc.b	$F8,$8D,$8F,$88,$F7,$AF,$88,$FF,$04,$3F,$A8,$1E,$C3,$A6,$1F,$65
	.dc.b	$E4,$7A,$0E,$15,$D9,$B5,$D9,$E9,$5D,$9F,$E0,$B1,$EC,$79,$9E,$A2
	.dc.b	$9E,$A3,$99,$DC,$FE,$0A,$6C,$C9,$4C,$9C,$4B,$5E,$97,$66,$70,$FE
	.dc.b	$0B,$3F,$56,$75,$3C,$30,$FC,$1C,$62,$B7,$3F,$12,$72,$47,$F0,$D3
	.dc.b	$B7,$ED,$9E,$C7,$18,$7C,$78,$EC,$70,$9E,$51,$F9,$02,$5F,$96,$3F
	.dc.b	$8E,$CF,$F2,$7A,$FE,$72,$D8,$FE,$54,$FE,$F8,$FE,$59,$E4,$7F,$3E
	.dc.b	$D7,$70,$8C,$4F,$F1,$89,$CC,$62,$5E,$4B,$47,$8D,$98,$C3,$61,$BD
	.dc.b	$8C,$44,$C7,$FA,$E6,$3F,$D4,$2F,$8C,$60,$FE,$C0,$F7,$9F,$17,$84
	.dc.b	$F3,$9F,$41,$86,$DF,$1C,$FF,$DE,$D0,$FA,$6D,$56,$FD,$69,$7B,$C3
	.dc.b	$A4,$E3,$0F,$F6,$2C,$FF,$72,$FA,$0F,$85,$1F,$DC,$DB,$E2,$BC,$BC
	.dc.b	$CA,$6F,$CF,$7C,$E8,$FE,$F4,$1C,$3F,$DF,$5A,$E6,$FE,$CA,$FA,$0E
	.dc.b	$A7,$48,$F7,$47,$F8,$16,$3E,$2F,$D0,$7D,$59,$F6,$36,$7D,$63,$34
	.dc.b	$7D,$69,$DE,$FC,$B8,$A4,$60,$E6,$FD,$FE,$EF,$74,$7C,$FE,$2F,$F8
	.dc.b	$01,$A6,$79,$74,$EF,$59,$5B,$BE,$3A,$79,$30,$47,$C1,$8C,$7D,$FB
	.dc.b	$E1,$FE,$55,$23,$17,$FE,$07,$8C,$4E,$A7,$35,$FE,$15,$F7,$FF,$C5
	.dc.b	$57,$F8,$5E,$94,$62,$7B,$DF,$E7,$E7,$31,$F9,$DF,$9E,$F3,$F4,$FE
	.dc.b	$5D,$B8,$7E,$5B,$C5,$7D,$CE,$F7,$E5,$ED,$BE,$82,$99,$BF,$14,$65
	.dc.b	$F9,$5F,$1B,$F9,$AC,$FF,$C8,$3C,$08,$C4,$7E,$2A,$31,$D7,$F1,$27
	.dc.b	$33,$FE,$47,$D9,$83,$E7,$FC,$67,$ED,$EB,$B6,$FF,$E4,$0A,$DC,$FE
	.dc.b	$EF,$F0,$17,$CC,$FD,$06,$E9,$F3,$DB,$97,$99,$AF,$F8,$EF,$27,$CF
	.dc.b	$FC,$23,$A5,$DB,$FD,$B7,$C1,$3E,$09,$F2,$DF,$C3,$E3,$0D,$F2,$FE
	.dc.b	$DE,$DF,$E5,$69,$CD,$FC,$AF,$87,$F9,$E7,$1C,$39,$C6,$4E,$FF,$81
	.dc.b	$BE,$9F,$F6,$96,$8C,$4F,$F9,$46,$31,$3A,$9C,$ED,$FE,$5F,$D9,$F4
	.dc.b	$1F,$78,$F4,$4E,$B3,$F9,$93,$AC,$7F,$6F,$F3,$FF,$7F,$0F,$D9,$87
	.dc.b	$15,$FC,$E9,$FB,$13,$C3,$9F,$F8,$8F,$CF,$FE,$5E,$FF,$D0,$AD,$7F
	.dc.b	$86,$B5,$FF,$44,$37,$A1,$F2,$EF,$97,$FE,$70,$DC,$3F,$85,$D7,$E8
	.dc.b	$F2,$FE,$69,$39,$C3,$CF,$F3,$CF,$8A,$FE,$46,$DC,$3F,$89,$AF,$8A
	.dc.b	$CC,$EB,$CE,$FE,$3F,$26,$92,$BF,$D0,$1E,$63,$E2,$7C,$51,$FE,$A1
	.dc.b	$8C,$63,$FC,$CB,$E7,$5F,$3D,$FD,$12,$F7,$2E,$77,$CB,$3E,$7B,$F1
	.dc.b	$8D,$CB,$F2,$3F,$2D,$F8,$6B,$FE,$4B,$7C,$2F,$F9,$FF,$CF,$3E,$7A
	.dc.b	$9F,$D0,$BE,$C9,$F1,$98,$CB,$F9,$9D,$7E,$DD,$FE,$E2,$DC,$7F,$D3
	.dc.b	$8D,$7F,$D7,$AD,$7F,$D7,$D8,$BE,$7B,$87,$37,$E7,$DB,$7F,$D2,$BE
	.dc.b	$CF,$FA,$E3,$FA,$A8,$CD,$BC,$37,$DC,$BE,$03,$FA,$94,$61,$B1,$9B
	.dc.b	$FF,$58,$4E,$5B,$0F,$9C,$D9,$23,$2F,$F5,$AC,$1E,$F7,$FB,$3F,$1F
	.dc.b	$F5,$D4,$63,$CD,$FF,$09,$B6,$33,$3B,$7F,$DB,$F1,$F9,$5F,$03,$7E
	.dc.b	$07,$F8,$42,$79,$B5,$DB,$5F,$F7,$36,$6F,$BC,$FD,$F4,$7B,$1A,$FF
	.dc.b	$BB,$23,$5E,$3F,$F8,$99,$B1,$BF,$2F,$15,$FC,$8A,$30,$43,$FD,$EB
	.dc.b	$C5,$F5,$FE,$87,$FB,$EF,$1F,$D7,$6F,$07,$DC,$D5,$18,$BF,$F7,$E7
	.dc.b	$D4,$7F,$35,$7F,$BD,$FD,$0B,$E5,$BF,$C0,$2D,$8F,$F1,$DB,$5F,$F7
	.dc.b	$33,$5F,$F0,$68,$7F,$2A,$F4,$E8,$F0,$39,$6D,$7F,$55,$8F,$E6,$7D
	.dc.b	$BF,$DC,$9B,$EE,$79,$F5,$FD,$5B,$7B,$BF,$09,$3F,$E2,$9F,$6D,$F2
	.dc.b	$1E,$DF,$13,$FB,$6B,$63,$33,$FE,$18,$FA,$07,$D4,$F9,$3A,$BF,$95
	.dc.b	$CF,$F9,$61,$AF,$B0,$FC,$6F,$A3,$FE,$24,$8C,$C7,$C0,$5E,$9F,$32
	.dc.b	$47,$F8,$9A,$3F,$A9,$BF,$E1,$7F,$86,$7D,$07,$D3,$EE,$FF,$1D,$36
	.dc.b	$33,$3F,$E2,$EF,$A0,$FF,$2D,$B7,$E2,$FF,$C6,$B1,$88,$F3,$35,$FA
	.dc.b	$6D,$7F,$C8,$2F,$F9,$A5,$E9,$F0,$FF,$52,$31,$1F,$D1,$1F,$F3,$6B
	.dc.b	$FE,$06,$F9,$CF,$C4,$78,$7F,$DF,$1B,$EB,$9F,$AD,$EC,$7E,$AF,$C0
	.dc.b	$FE,$CC,$FE,$B6,$FF,$CA,$3D,$1F,$D7,$F0,$FF,$42,$B6,$3F,$B0,$4E
	.dc.b	$DF,$F0,$56,$1F,$E6,$F7,$FC,$2F,$F1,$9E,$03,$F0,$9E,$D7,$F9,$7E
	.dc.b	$3F,$BD,$B5,$FE,$56,$DC,$7F,$BC,$37,$C2,$FE,$3F,$E2,$FF,$98,$EB
	.dc.b	$FC,$47,$E9,$FB,$08,$FF,$93,$7D,$DD,$3F,$E4,$76,$BF,$E2,$36,$BF
	.dc.b	$E8,$86,$BF,$E9,$86,$EB,$FE,$24,$F5,$BE,$49,$27,$FC,$B5,$F8,$17
	.dc.b	$F4,$C6,$FA,$BF,$EA,$AE,$74,$E6,$7F,$C6,$AD,$7F,$C6,$8D,$7E,$4F
	.dc.b	$B1,$F5,$88,$FF,$99,$FE,$7F,$F2,$ED,$EB,$FC,$94,$62,$3F,$CC,$8D
	.dc.b	$E0,$FF,$53,$7C,$2F,$C9,$96,$9F,$E3,$EF,$8B,$FC,$19,$23,$FC,$E9
	.dc.b	$F1,$3F,$E2,$C6,$FC,$0F,$AE,$31,$1F,$E4,$E6,$BF,$E4,$D6,$BF,$EB
	.dc.b	$2F,$99,$FC,$73,$FE,$7F,$6B,$FE,$79,$6E,$3F,$EB,$B7,$F1,$11,$9E
	.dc.b	$6F,$D7,$46,$27,$51,$8F,$07,$F8,$47,$CF,$7F,$1D,$97,$3F,$F8,$06
	.dc.b	$30,$DF,$3B,$7F,$3D,$32,$FF,$5D,$52,$31,$1F,$EA,$0F,$25,$3F,$C4
	.dc.b	$23,$FD,$53,$F2,$AF,$9E,$FE,$4A,$D8,$CC,$63,$E7,$64,$7F,$BB,$25
	.dc.b	$1F,$EB,$66,$C6,$63,$1F,$3D,$D0,$91,$FE,$B8,$FA,$EF,$E6,$FA,$FB
	.dc.b	$A3,$FD,$69,$F3,$FF,$DD,$DF,$F6,$A3,$5F,$53,$63,$86,$68,$FE,$5B
	.dc.b	$3A,$F3,$FF,$99,$46,$2B,$6D,$DF,$FC,$58,$FF,$64,$8F,$F7,$97,$CF
	.dc.b	$3A,$3F,$92,$37,$0F,$F5,$F3,$5F,$F5,$C2,$F9,$ED,$92,$36,$F2,$FA
	.dc.b	$DF,$9B,$F3,$F2,$7E,$8B,$F4,$3E,$7E,$47,$CD,$7E,$61,$1F,$A1,$E7
	.dc.b	$32,$43,$E7,$7E,$37,$1A,$FF,$DD,$F1,$86,$FC,$27,$FB,$CA,$58,$73
	.dc.b	$BE,$53,$98,$C5,$F1,$9E,$ED,$76,$78,$AB,$9E,$1A,$BE,$4D,$9D,$61
	.dc.b	$E2,$69,$91,$5C,$9E,$A6,$30,$F1,$DF,$A6,$B9,$4F,$06,$D7,$1C,$3E
	.dc.b	$1D,$F0,$AE,$E7,$49,$F8,$CF,$85,$E1,$7E,$88,$D5,$F4,$3A,$D7,$38
	.dc.b	$F2,$61,$C6,$75,$1C,$E7,$57,$D2,$3C,$38,$7C,$93,$CB,$7B,$39,$44
	.dc.b	$3C,$D5,$C8,$E7,$7D,$A3,$33,$C5,$F3,$9C,$6B,$AB,$A3,$24,$72,$4C
	.dc.b	$3E,$64,$71,$7E,$14,$72,$F9,$FF,$C0,$12,$3E,$33,$C6,$7D,$B1,$C9
	.dc.b	$F0,$4E,$5F,$AC,$75,$3F,$11,$87,$B6,$FF,$37,$7D,$67,$EB,$BE,$1B
	.dc.b	$F1,$9D,$6F,$E5,$38,$9E,$75,$D5,$EE,$78,$A3,$CE,$79,$6B,$B4,$7D
	.dc.b	$A0,$F9,$F1,$FC,$2C,$7C,$79,$EC,$EE,$FC,$2F,$C0,$3C,$27,$CF,$AE
	.dc.b	$B5,$C9,$F5,$46,$A7,$8B,$E6,$7C,$18,$7D,$E8,$F5,$57,$E2,$5F,$35
	.dc.b	$74,$C7,$F8,$0C,$7A,$30,$F6,$BD,$5F,$8C,$47,$04,$3C,$0E,$5F,$63
	.dc.b	$C9,$F9,$33,$EA,$BF,$E0,$CF,$C5,$AF,$C9,$61,$E8,$C7,$CF,$1F,$A6
	.dc.b	$7F,$29,$3E,$7B,$FB,$B8,$75,$C7,$D8,$7C,$EC,$3E,$1B,$8C,$7E,$15
	.dc.b	$7D,$13,$D1,$1F,$C1,$6F,$A4,$7C,$13,$E8,$BE,$57,$E4,$CF,$E3,$5E
	.dc.b	$47,$C0,$39,$BE,$CB,$FC,$21,$23,$9B,$F8,$0A,$F3,$3F,$1D,$F3,$61
	.dc.b	$F0,$27,$F6,$58,$78,$5E,$F3,$A6,$E3,$F1,$5F,$CA,$1F,$89,$3C,$87
	.dc.b	$D2,$3D,$EF,$CE,$BF,$16,$1E,$79,$FD,$2C,$79,$DF,$C2,$BE,$79,$E1
	.dc.b	$1E,$EC,$3C,$CF,$80,$FD,$01,$FE,$A8,$BF,$E6,$0D,$BF,$E8,$8D,$9F
	.dc.b	$F0,$07,$A3,$FE,$ED,$8F,$F4,$A5,$23,$11,$CD,$AF,$F4,$F6,$BF,$EB
	.dc.b	$6F,$9F,$FF,$4E,$3F,$9A,$6B,$FD,$71,$A9,$A4,$76,$D7,$9B,$5F,$F5
	.dc.b	$43,$5F,$01,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08
	.dc.b	$7B,$CE,$A5,$39,$14,$E2,$53,$89,$F1,$4E,$25,$38,$94,$E2,$53,$89
	.dc.b	$F2,$CE,$25,$38,$94,$E2,$53,$8B,$F2,$DF,$21,$0E,$C5,$38,$15,$E0
	.dc.b	$E8,$86,$8A,$E9,$C1,$1C,$11,$F4,$B5,$F4,$B5,$F4,$B5,$F4,$9C,$08
	.dc.b	$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$87,$C3,$3A,$94,$E4
	.dc.b	$53,$89,$4E,$27,$C6,$38,$94,$E2,$53,$89,$4E,$27,$CB,$38,$94,$E2
	.dc.b	$53,$89,$4E,$2F,$CB,$7C,$84,$3B,$14,$E0,$57,$83,$A2,$1A,$2B,$A7
	.dc.b	$04,$70,$47,$D2,$D7,$D2,$D7,$D2,$D7,$D2,$70,$21,$C0,$A7,$02,$9C
	.dc.b	$0D,$10,$D1,$4D,$18,$21,$82,$1F,$0C,$EA,$53,$91,$4E,$25,$38,$9F
	.dc.b	$18,$E2,$53,$89,$4E,$25,$38,$9F,$2C,$E2,$53,$89,$FB,$89,$FE,$79
	.dc.b	$E6,$7F,$A3,$BE,$5B,$FD,$91,$3F,$E7,$6B,$FE,$85,$18,$AD,$CA,$31
	.dc.b	$3A,$F5,$FF,$A1,$5F,$36,$D6,$DB,$5C,$27,$53,$FE,$2C,$3F,$A9,$1F
	.dc.b	$D4,$32,$FE,$AA,$FF,$9D,$CC,$90,$C1,$0E,$45,$3F,$B0,$97,$0F,$EC
	.dc.b	$2D,$8E,$AF,$02,$1C,$0B,$1C,$1B,$1C,$1D,$12,$34,$D8,$D4,$60,$8E
	.dc.b	$09,$97,$FB,$00,$93,$F0,$5A,$FB,$CF,$0B,$87,$D4,$D7,$27,$12,$1C
	.dc.b	$4F,$23,$B6,$9A,$39,$94,$E6,$71,$9F,$4C,$71,$6C,$6C,$E6,$6E,$7D
	.dc.b	$AE,$63,$6E,$67,$94,$EA,$30,$6D,$C3,$E4,$32,$43,$04,$3C,$E5,$3C
	.dc.b	$65,$9F,$0B,$63,$AB,$C0,$87,$02,$C7,$06,$C7,$07,$44,$8D,$36,$35
	.dc.b	$18,$23,$82,$4F,$C2,$24,$FC,$16,$BE,$F3,$C2,$E1,$F5,$35,$C9,$C4
	.dc.b	$87,$13,$C8,$ED,$A6,$8E,$65,$39,$9C,$67,$D3,$1C,$5B,$1B,$39,$9B
	.dc.b	$9F,$6B,$98,$DB,$99,$E5,$3A,$8C,$1B,$70,$F9,$0C,$90,$C1,$0F,$39
	.dc.b	$4F,$19,$67,$C2,$D8,$EA,$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D
	.dc.b	$8D,$46,$08,$E0,$93,$F0,$89,$3F,$05,$AF,$BC,$F0,$B8,$7D,$4F,$F9
	.dc.b	$E2,$A4,$7F,$9E,$31,$FF,$39,$BF,$CE,$E3,$2F,$F9,$E5,$AF,$F3,$63
	.dc.b	$07,$17,$28,$E4,$64,$87,$02,$1C,$C8,$71,$BF,$E9,$2E,$8C,$4F,$9A
	.dc.b	$33,$1C,$0E,$E6,$CA,$65,$DB,$CA,$7C,$EE,$8C,$1E,$63,$84,$F7,$24
	.dc.b	$70,$9E,$73,$C4,$8E,$09,$3C,$23,$A1,$D8,$F3,$46,$A7,$A4,$EE,$72
	.dc.b	$D7,$C7,$1A,$23,$D0,$F0,$C6,$DE,$C7,$03,$53,$DD,$B1,$B7,$8C,$F4
	.dc.b	$7C,$2D,$79,$18,$38,$B9,$7B,$90,$E6,$68,$87,$12,$1D,$08,$71,$9F
	.dc.b	$33,$A3,$13,$E7,$8C,$C7,$03,$B9,$B2,$99,$76,$F2,$9F,$43,$A3,$07
	.dc.b	$98,$E1,$3D,$C9,$1C,$27,$9C,$F1,$23,$82,$4F,$08,$E8,$76,$3C,$D1
	.dc.b	$A9,$E9,$3B,$9C,$B5,$F1,$C6,$88,$F4,$3C,$31,$B7,$B1,$C0,$D4,$F7
	.dc.b	$6C,$6D,$E3,$3D,$1F,$0B,$5E,$46,$0E,$2E,$5E,$E4,$39,$9A,$21,$C4
	.dc.b	$87,$42,$1C,$67,$CC,$E8,$C4,$F9,$E3,$31,$C0,$EE,$6C,$A6,$5D,$BC
	.dc.b	$A7,$D0,$E8,$C1,$E6,$38,$4F,$72,$47,$09,$E7,$3C,$48,$E0,$93,$C2
	.dc.b	$3A,$1D,$8F,$34,$6A,$7A,$4E,$E7,$2D,$7C,$71,$A2,$3D,$0F,$0C,$6D
	.dc.b	$EC,$70,$35,$3D,$DB,$1B,$78,$CF,$47,$C2,$D7,$91,$83,$8B,$97,$B9
	.dc.b	$0E,$66,$88,$71,$21,$D0,$87,$19,$F3,$3A,$31,$3E,$78,$CC,$70,$3B
	.dc.b	$9B,$29,$97,$6F,$29,$F4,$3A,$30,$79,$8E,$13,$DC,$91,$C2,$79,$CF
	.dc.b	$12,$38,$24,$F0,$8E,$87,$63,$CD,$1A,$9E,$93,$B9,$CB,$5F,$1C,$68
	.dc.b	$8F,$43,$C3,$1B,$7B,$1C,$0D,$4F,$76,$C6,$DE,$33,$D1,$F0,$B5,$FE
	.dc.b	$ED,$F9,$D0,$80

;--------------------------------------------------------------------------------
;	FNEG.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fneg_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#63,-(sp)
	peamsg	'FNEG.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FNEG.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fneg_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fneg.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FNEG.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fneg.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FNEG.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fneg_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$E6,$3A,$29,$F7,$3F,$80,$9C,$C6,$1C,$4E,$E7,$51
	.dc.b	$E6,$9D,$3F,$88,$9D,$3C,$23,$2F,$96,$78,$CE,$A3,$84,$E9,$F0,$CE
	.dc.b	$A3,$11,$DE,$38,$E1,$E2,$8F,$8E,$DA,$34,$C5,$7A,$FE,$BB,$FC,$DF
	.dc.b	$F0,$5F,$AE,$6E,$7F,$D0,$1F,$F3,$9D,$FA,$9B,$5C,$7D,$2F,$D1,$E3
	.dc.b	$FD,$6A,$7F,$CF,$31,$FC,$B6,$11,$C3,$2D,$4F,$D7,$9F,$83,$87,$9C
	.dc.b	$8F,$BE,$76,$FF,$07,$8F,$24,$FF,$0D,$6C,$7C,$63,$BC,$7C,$D6,$9F
	.dc.b	$DB,$9F,$25,$79,$BD,$6E,$33,$F2,$CD,$9F,$90,$DB,$E5,$11,$FA,$EF
	.dc.b	$CA,$C3,$FB,$CF,$C3,$7C,$89,$B8,$79,$5A,$79,$63,$E1,$3F,$DA,$63
	.dc.b	$E4,$B4,$F3,$5F,$F7,$09,$D9,$F4,$B8,$3E,$E1,$27,$F1,$2F,$8D,$F4
	.dc.b	$EB,$F6,$EF,$C5,$1E,$CD,$7E,$D4,$7C,$5C,$FE,$14,$FC,$E1,$1F,$18
	.dc.b	$D1,$FB,$07,$F2,$FC,$3F,$8D,$23,$F5,$CF,$0E,$BF,$33,$0F,$14,$7D
	.dc.b	$5B,$F8,$27,$81,$F9,$F7,$DD,$D3,$FC,$0F,$8B,$F3,$44,$8F,$D3,$BE
	.dc.b	$F9,$FA,$11,$E0,$3D,$58,$FB,$9E,$67,$CC,$3D,$D5,$DD,$FC,$04,$FE
	.dc.b	$7B,$3F,$D4,$12,$3F,$6A,$F7,$9F,$97,$3E,$03,$D6,$FC,$69,$EC,$7B
	.dc.b	$34,$FC,$6D,$F8,$E3,$F8,$DF,$C3,$FE,$D4,$8F,$F0,$27,$B3,$F9,$F3
	.dc.b	$EA,$65,$F2,$9F,$3D,$F7,$7E,$BD,$FE,$6A,$3D,$97,$DF,$C6,$72,$FF
	.dc.b	$05,$6D,$FF,$18,$6B,$FC,$7D,$AF,$F1,$2E,$0F,$0E,$DF,$6C,$F5,$F1
	.dc.b	$7F,$01,$9F,$9A,$78,$8F,$9B,$96,$AB,$F1,$96,$E6,$F1,$12,$BC,$4F
	.dc.b	$C0,$7F,$C0,$D8,$77,$7E,$DE,$1D,$DF,$82,$7C,$ED,$BF,$14,$7D,$27
	.dc.b	$FA,$EF,$3F,$BC,$8F,$BD,$ED,$B7,$CB,$3E,$66,$BF,$54,$FA,$DB,$7E
	.dc.b	$14,$F9,$78,$78,$73,$74,$7C,$B2,$63,$F2,$DE,$B9,$FA,$63,$F6,$15
	.dc.b	$E9,$C3,$AE,$DF,$5C,$F8,$D3,$FC,$43,$07,$4F,$D7,$24,$7D,$77,$AB
	.dc.b	$F5,$0F,$8B,$BC,$7F,$20,$7D,$23,$1B,$7E,$F1,$FF,$00,$F5,$7E,$0C
	.dc.b	$8F,$E0,$DE,$AF,$EA,$CF,$4C,$F2,$D7,$F8,$BD,$7D,$07,$F2,$58,$3A
	.dc.b	$FF,$1C,$4B,$FC,$73,$D2,$BF,$51,$5F,$99,$CF,$E1,$D7,$9C,$EE,$7C
	.dc.b	$3E,$CF,$CC,$11,$FC,$C3,$CE,$7F,$66,$73,$D7,$F0,$CF,$D9,$D7,$E6
	.dc.b	$E1,$F0,$6D,$93,$B7,$F4,$84,$8F,$D2,$3D,$5F,$D9,$1E,$49,$FC,$73
	.dc.b	$F9,$F3,$E7,$99,$BF,$F1,$23,$FA,$DE,$EF,$D8,$11,$FE,$4D,$3F,$09
	.dc.b	$EA,$F4,$CF,$E1,$C7,$48,$EF,$83,$BF,$F7,$44,$9F,$DD,$3E,$77,$F5
	.dc.b	$27,$D7,$3F,$6B,$87,$C3,$70,$7C,$2F,$03,$F8,$21,$1F,$E0,$8F,$3E
	.dc.b	$4F,$D2,$EB,$F9,$93,$E6,$EB,$F9,$9A,$E7,$8B,$C1,$FE,$1C,$48,$FE
	.dc.b	$1C,$F3,$E4,$FA,$79,$7F,$19,$3C,$93,$F2,$4F,$F2,$AF,$85,$FC,$5C
	.dc.b	$8F,$F1,$77,$93,$EA,$DB,$E3,$ED,$F9,$38,$FE,$27,$8B,$C3,$FE,$40
	.dc.b	$48,$FE,$40,$F2,$7C,$07,$F0,$C3,$EE,$4F,$37,$12,$F1,$3F,$95,$12
	.dc.b	$3F,$95,$3C,$E7,$ED,$6B,$F1,$A7,$F3,$91,$F6,$BC,$5F,$E6,$44,$8F
	.dc.b	$E6,$4F,$1C,$FF,$35,$CD,$C1,$FD,$79,$FA,$EF,$1B,$F9,$D1,$2B,$F9
	.dc.b	$D3,$C5,$FE,$67,$A7,$C4,$9E,$31,$F1,$3E,$17,$FA,$11,$1F,$E8,$4F
	.dc.b	$17,$E1,$4F,$17,$F9,$74,$FE,$FB,$E2,$FF,$A5,$12,$3F,$A5,$3C,$67
	.dc.b	$F9,$56,$7F,$06,$7F,$35,$5F,$C1,$BE,$3B,$FA,$91,$23,$FA,$93,$C6
	.dc.b	$7F,$1D,$97,$C1,$9F,$09,$F9,$AF,$2B,$FC,$BD,$66,$92,$3F,$5B,$1F
	.dc.b	$C1,$1F,$D6,$BF,$31,$F1,$9D,$0F,$C0,$2F,$2F,$A9,$23,$FC,$FB,$19
	.dc.b	$7A,$9D,$08,$7C,$8B,$F8,$1B,$7E,$27,$2F,$E9,$7E,$67,$A0,$95,$D1
	.dc.b	$F8,$19,$7E,$D3,$0F,$D1,$DF,$F0,$E5,$E6,$FC,$02,$5F,$91,$E2,$FE
	.dc.b	$6A,$38,$9F,$AD,$F3,$BE,$39,$23,$D6,$F0,$AE,$DD,$1F,$BB,$9E,$5E
	.dc.b	$7F,$CE,$23,$F0,$9E,$17,$FB,$99,$E1,$5E,$7F,$43,$EA,$92,$FE,$4B
	.dc.b	$C1,$FE,$27,$3C,$2B,$F8,$17,$A3,$F7,$09,$3F,$41,$E0,$FD,$69,$E0
	.dc.b	$FF,$1C,$F4,$BF,$06,$47,$EB,$BC,$36,$FC,$D4,$7E,$F2,$FF,$88,$7A
	.dc.b	$7F,$8A,$24,$7D,$E7,$87,$07,$DA,$9F,$D4,$CF,$E1,$BD,$4F,$C8,$91
	.dc.b	$FC,$2B,$C3,$83,$F3,$11,$DC,$FE,$63,$EA,$FE,$60,$8F,$E3,$1E,$0F
	.dc.b	$E5,$0E,$07,$E5,$3D,$6F,$CF,$91,$FC,$9B,$C3,$7F,$E2,$D8,$7F,$85
	.dc.b	$AF,$FC,$BB,$EB,$FE,$98,$97,$F9,$A7,$84,$FB,$0E,$11,$FE,$72,$F6
	.dc.b	$3F,$5A,$47,$F4,$2F,$07,$F5,$11,$C2,$FF,$D7,$EF,$ED,$09,$1F,$A8
	.dc.b	$78,$3E,$C7,$80,$7F,$D6,$9E,$CF,$EF,$08,$FE,$C1,$E3,$1F,$9C,$8F
	.dc.b	$EB,$7E,$D7,$F0,$22,$3F,$B6,$77,$C3,$F8,$68,$FD,$EF,$B7,$FC,$20
	.dc.b	$8F,$EF,$1F,$C6,$D7,$FB,$5F,$37,$5F,$FA,$D1,$B5,$FE,$7C,$6B,$F9
	.dc.b	$33,$F8,$2D,$FF,$14,$23,$FC,$18,$FE,$E6,$7F,$8A,$F2,$FD,$7B,$C0
	.dc.b	$87,$C3,$3F,$D8,$65,$3F,$D6,$7E,$E7,$F2,$02,$3F,$C5,$1F,$49,$87
	.dc.b	$81,$C0,$D1,$FC,$68,$FE,$52,$43,$07,$FA,$E7,$D5,$FF,$55,$93,$1F
	.dc.b	$F7,$53,$FE,$FC,$8E,$07,$68,$EE,$FA,$5A,$FD,$DF,$21,$EE,$E1,$B1
	.dc.b	$C1,$AF,$A1,$AF,$FA,$44,$F9,$13,$FC,$ED,$FE,$ED,$1A,$7E,$43,$5F
	.dc.b	$04,$F7,$3C,$1E,$F4,$8F,$03,$5F,$8F,$CF,$FE,$B5,$AF,$F4,$27,$93
	.dc.b	$FE,$16,$9F,$55,$6D,$FA,$92,$8C,$47,$8B,$DF,$F7,$11,$F7,$3F,$C4
	.dc.b	$FE,$1F,$EF,$6F,$F8,$AC,$5F,$0F,$8E,$1F,$C1,$46,$30,$FD,$35,$6E
	.dc.b	$F5,$1F,$37,$E0,$BE,$61,$1F,$98,$F7,$78,$DF,$4C,$38,$C3,$0E,$CB
	.dc.b	$0E,$CF,$03,$ED,$7C,$1F,$D9,$23,$F6,$5F,$E5,$DF,$0F,$D2,$8D,$31
	.dc.b	$F0,$FE,$0D,$9F,$0D,$F8,$32,$61,$F8,$37,$9B,$E1,$3C,$D5,$CD,$C5
	.dc.b	$AF,$A7,$C3,$FE,$34,$93,$F8,$D7,$FA,$11,$E5,$B9,$1C,$A3,$C9,$3F
	.dc.b	$95,$24,$7E,$55,$E2,$F8,$27,$8B,$F1,$F4,$FA,$F5,$ED,$C3,$F3,$C4
	.dc.b	$9F,$CF,$3F,$D4,$F2,$F8,$DA,$C9,$F5,$C7,$C5,$D7,$F4,$C4,$8F,$D3
	.dc.b	$3C,$5F,$83,$3C,$5F,$11,$FC,$AF,$35,$7C,$FD,$BF,$60,$49,$FD,$83
	.dc.b	$FD,$97,$7F,$B0,$48,$F9,$BC,$1F,$5F,$E1,$FF,$70,$48,$FD,$C3,$C5
	.dc.b	$FA,$73,$C5,$F1,$9F,$8D,$9F,$C0,$78,$3F,$C0,$49,$1F,$C0,$5F,$E0
	.dc.b	$9C,$7F,$85,$24,$7D,$E8,$FF,$80,$1C,$47,$8C,$7F,$07,$6B,$FB,$4F
	.dc.b	$86,$FE,$1C,$47,$F8,$73,$D9,$FC,$64,$AB,$BC,$7A,$34,$FE,$05,$18
	.dc.b	$E4,$FE,$7F,$5B,$AE,$38,$70,$7F,$81,$D7,$57,$F8,$D3,$5C,$9F,$C7
	.dc.b	$DC,$3F,$C6,$1A,$FF,$B3,$8D,$9F,$C8,$FE,$1B,$F9,$41,$1F,$E5,$0D
	.dc.b	$74,$60,$86,$08,$7F,$18,$78,$3F,$C6,$BE,$1B,$F9,$A1,$1F,$E6,$8D
	.dc.b	$7F,$9A,$35,$E0,$68,$86,$8A,$7F,$26,$F8,$6F,$E8,$04,$7F,$A0,$35
	.dc.b	$FE,$80,$D7,$81,$A2,$1A,$29,$FC,$CB,$E1,$BF,$A6,$11,$FE,$98,$D7
	.dc.b	$FA,$63,$5E,$06,$88,$68,$A7,$F3,$EE,$BF,$EB,$04,$7F,$AC,$3F,$D7
	.dc.b	$B4,$FE,$B3,$A7,$F8,$13,$EE,$FF,$BC,$36,$BF,$92,$75,$FF,$30,$97
	.dc.b	$6D,$EF,$95,$FC,$7E,$72,$FE,$5B,$E1,$FF,$B4,$11,$FE,$D0,$F8,$6F
	.dc.b	$A4,$43,$F9,$2B,$DE,$FA,$C7,$ED,$2B,$FC,$3D,$D9,$FE,$0F,$7B,$C7
	.dc.b	$EE,$3E,$25,$F1,$1F,$DE,$9B,$1F,$DF,$9A,$FF,$83,$9A,$FF,$85,$D7
	.dc.b	$C4,$7F,$4E,$6C,$7F,$3B,$6B,$E3,$F1,$35,$32,$34,$FC,$7F,$17,$11
	.dc.b	$DE,$3C,$84,$7C,$8D,$74,$60,$86,$08,$79,$DF,$77,$C4,$48,$F7,$35
	.dc.b	$E1,$BF,$DD,$F1,$FE,$D1,$E4,$7F,$C8,$78,$FB,$1F,$C1,$78,$DF,$7C
	.dc.b	$D6,$BF,$E3,$D9,$7C,$47,$CD,$26,$1F,$35,$F5,$5F,$78,$FC,$26,$1C
	.dc.b	$9F,$F3,$45,$77,$BF,$B5,$87,$79,$F9,$27,$CF,$F8,$8F,$B4,$47,$ED
	.dc.b	$3F,$77,$E1,$FE,$B1,$23,$F2,$5E,$5D,$7F,$09,$17,$2F,$CF,$79,$9F
	.dc.b	$77,$E2,$3F,$0C,$4A,$FC,$33,$D5,$FC,$05,$71,$CB,$F2,$18,$59,$F5
	.dc.b	$CD,$5C,$6C,$3C,$51,$F8,$D9,$E8,$FE,$45,$AE,$4F,$CB,$38,$7E,$83
	.dc.b	$5E,$E6,$8F,$CE,$7C,$47,$E7,$C8,$E0,$D1,$FA,$9F,$3F,$F7,$47,$9F
	.dc.b	$65,$7E,$F7,$D0,$95,$F0,$0F,$D0,$7C,$47,$EA,$C8,$FE,$AD,$F5,$ED
	.dc.b	$FC,$27,$0E,$B3,$C4,$FA,$27,$F0,$5A,$EA,$E0,$FD,$C7,$C4,$7E,$DC
	.dc.b	$8F,$ED,$DE,$BB,$7E,$FE,$BA,$CF,$1B,$FC,$CE,$3D,$1F,$E3,$58,$7B
	.dc.b	$7E,$23,$F8,$11,$2B,$F8,$13,$5F,$E0,$4D,$7F,$81,$35,$FD,$03,$5F
	.dc.b	$CC,$BF,$C3,$7E,$22,$47,$F0,$D6,$BF,$C3,$5A,$FF,$0D,$6B,$FA,$B6
	.dc.b	$BF,$A4,$6A,$D1,$F1,$1F,$C6,$C9,$1F,$C6,$DA,$FF,$1B,$6B,$FC,$6D
	.dc.b	$AF,$EE,$5A,$FE,$C9,$FE,$4D,$F1,$12,$3F,$93,$35,$FE,$4C,$D7,$F9
	.dc.b	$33,$5F,$E0,$8D,$7F,$7A,$D7,$D5,$F1,$1F,$CC,$C8,$FF,$33,$6B,$FC
	.dc.b	$CD,$AF,$F3,$36,$BF,$C4,$5A,$FF,$0A,$7F,$9F,$7C,$3C,$8F,$E7,$CD
	.dc.b	$7F,$9F,$35,$FE,$7C,$D7,$F8,$E3,$5F,$E2,$AD,$7E,$97,$C3,$FF,$4B
	.dc.b	$23,$FD,$2D,$AF,$F4,$B6,$BF,$D2,$DA,$FF,$29,$6B,$FC,$89,$FE,$AD
	.dc.b	$F0,$F2,$3F,$AB,$35,$FE,$AC,$D7,$FA,$B3,$5F,$E6,$8D,$7F,$96,$B5
	.dc.b	$FC,$57,$C3,$FF,$63,$23,$FD,$8D,$AF,$F6,$36,$BF,$D8,$DA,$FF,$41
	.dc.b	$6B,$FC,$E9,$FE,$DD,$F0,$F2,$3F,$B7,$35,$FF,$7D,$FC,$53,$7F,$F5
	.dc.b	$CB,$63,$FD,$4C,$D7,$FB,$F7,$BD,$FE,$0F,$F7,$F7,$52,$7F,$C1,$D1
	.dc.b	$FD,$D7,$97,$FB,$F1,$23,$FB,$F3,$FE,$FA,$BE,$9C,$68,$4C,$9C,$C6
	.dc.b	$21,$18,$77,$D4,$C7,$C5,$1F,$E8,$69,$D3,$C3,$1F,$43,$EC,$F8,$8F
	.dc.b	$11,$1E,$C7,$79,$47,$9E,$7A,$E3,$E7,$AF,$34,$7F,$8C,$6F,$38,$F1
	.dc.b	$8B,$3E,$6C,$BE,$6F,$C4,$7C,$12,$47,$80,$F8,$0F,$77,$E8,$CF,$27
	.dc.b	$FA,$CD,$DC,$BD,$A7,$BB,$36,$1F,$6B,$E2,$3E,$71,$2F,$B1,$F2,$9E
	.dc.b	$CF,$C7,$C3,$B5,$7C,$B8,$F0,$90,$D1,$F7,$8F,$C2,$FC,$47,$DB,$23
	.dc.b	$DB,$04,$76,$D3,$E4,$4F,$D3,$3A,$3F,$E5,$DC,$BD,$98,$78,$B2,$FC
	.dc.b	$9F,$C4,$7E,$1C,$97,$D5,$E9,$AF,$CF,$8E,$6F,$F9,$46,$F9,$C6,$0F
	.dc.b	$A4,$7E,$73,$E2,$3F,$24,$47,$D0,$F4,$7E,$E9,$C4,$FA,$25,$CF,$E8
	.dc.b	$E3,$F8,$C8,$FC,$87,$C4,$7E,$74,$8F,$C3,$7A,$3F,$04,$F9,$67,$EB
	.dc.b	$0E,$91,$F0,$5F,$D8,$FC,$47,$E9,$C8,$FC,$D7,$9D,$78,$A3,$FC,$FD
	.dc.b	$87,$C0,$7A,$E9,$FE,$89,$D3,$F7,$BF,$11,$FB,$12,$4F,$D7,$79,$E3
	.dc.b	$F8,$2B,$E5,$7F,$63,$0F,$C7,$61,$FC,$18,$FD,$D1,$30,$FC,$03,$C9
	.dc.b	$F9,$1B,$FE,$2D,$F1,$EB,$F5,$DA,$7D,$43,$F7,$18,$7F,$05,$23,$F8
	.dc.b	$B7,$A3,$C0,$FA,$65,$BF,$84,$FF,$A4,$EF,$F8,$1B,$FB,$C8,$FE,$34
	.dc.b	$DC,$7F,$42,$D3,$F3,$0E,$1F,$31,$93,$F9,$3B,$97,$F0,$A7,$F2,$AE
	.dc.b	$FF,$E3,$84,$7F,$44,$FE,$03,$16,$7F,$EB,$73,$E6,$E5,$FC,$77,$A1
	.dc.b	$EE,$FF,$1F,$D7,$F4,$F1,$8C,$7F,$21,$E2,$7F,$29,$23,$FA,$F7,$B6
	.dc.b	$1D,$63,$F1,$67,$F2,$CA,$E7,$7D,$70,$F0,$3F,$C2,$FC,$5F,$E6,$84
	.dc.b	$AF,$DD,$3D,$1F,$F6,$55,$FA,$1F,$E6,$F8,$F9,$56,$1D,$9F,$D3,$79
	.dc.b	$1F,$CF,$C8,$FF,$05,$7A,$3F,$EC,$9A,$E8,$AB,$A5,$73,$8F,$BD,$E4
	.dc.b	$FF,$49,$23,$FC,$3D,$E7,$1F,$EE,$7A,$E7,$5C,$B2,$F7,$DF,$EE,$7C
	.dc.b	$DF,$EA,$44,$7F,$8C,$3C,$A3,$FD,$B5,$5C,$B3,$F9,$15,$F0,$1F,$E6
	.dc.b	$5E,$77,$F5,$B2,$47,$F2,$17,$93,$FD,$5A,$7F,$9D,$61,$E3,$8E,$47
	.dc.b	$D5,$BF,$2B,$F9,$BF,$4F,$FB,$31,$1F,$E5,$8F,$47,$FD,$37,$5D,$30
	.dc.b	$F2,$C7,$C9,$6D,$F7,$8F,$E2,$19,$7E,$EF,$D1,$FE,$E4,$48,$FE,$70
	.dc.b	$F5,$E0,$FC,$BC,$7F,$21,$AE,$44,$BF,$EA,$51,$FD,$EC,$87,$F4,$13
	.dc.b	$96,$1E,$57,$DC,$7C,$03,$F0,$F9,$7C,$0C,$7F,$87,$71,$7A,$3D,$2F
	.dc.b	$EA,$F5,$F8,$36,$E2,$69,$DA,$3C,$D8,$FC,$02,$38,$3F,$B0,$CF,$E5
	.dc.b	$DF,$EF,$ED,$7E,$89,$FB,$FB,$FC,$53,$FA,$03,$91,$FD,$DE,$38,$3E
	.dc.b	$B3,$9C,$F9,$BD,$E9,$1E,$06,$BF,$C2,$8E,$87,$E8,$4E,$B5,$FD,$9B
	.dc.b	$37,$47,$DC,$6C,$FC,$A3,$EF,$17,$07,$C1,$7C,$62,$47,$C0,$7E,$4F
	.dc.b	$B5,$F8,$42,$47,$3A,$75,$FE,$23,$E0,$A4,$F5,$7C,$5D,$B8,$8F,$85
	.dc.b	$1F,$50,$8F,$CE,$BE,$AF,$DE,$3F,$83,$BE,$BA,$E5,$1F,$4A,$78,$BD
	.dc.b	$0F,$AF,$F0,$DF,$80,$23,$83,$B1,$F0,$6B,$BE,$9F,$26,$FC,$CF,$F2
	.dc.b	$A8,$CF,$0F,$F2,$FC,$3F,$4B,$F0,$DF,$8C,$26,$1F,$86,$7E,$3B,$D4
	.dc.b	$87,$AC,$67,$F2,$1F,$8C,$6B,$0F,$03,$83,$86,$AF,$87,$FC,$D9,$23
	.dc.b	$F2,$AF,$81,$DE,$5E,$87,$FC,$BF,$9F,$B3,$3F,$E9,$13,$99,$FE,$51
	.dc.b	$9F,$F0,$0F,$87,$FD,$39,$23,$F4,$2F,$5C,$BC,$B1,$D8,$FC,$16,$1D
	.dc.b	$5F,$AE,$7A,$F0,$F9,$0F,$E3,$FE,$1F,$F6,$64,$7F,$5A,$F5,$8F,$2C
	.dc.b	$71,$3D,$27,$C3,$AF,$09,$23,$AC,$7C,$6F,$87,$FD,$F9,$1F,$DC,$BD
	.dc.b	$5F,$29,$E9,$3E,$A6,$BF,$4D,$C5,$75,$76,$FF,$6C,$CB,$F9,$2F,$C3
	.dc.b	$FF,$0C,$24,$FF,$06,$7B,$3F,$D4,$F4,$F9,$D7,$D6,$FA,$4E,$4F,$71
	.dc.b	$F4,$8F,$DA,$7C,$2F,$F1,$82,$3F,$C4,$9E,$BB,$AF,$AE,$BF,$6E,$31
	.dc.b	$5D,$16,$5F,$CF,$3E,$1F,$F9,$19,$27,$F8,$EB,$D2,$F9,$C7,$17,$A4
	.dc.b	$FF,$44,$F8,$6F,$E5,$64,$9F,$E4,$AF,$17,$D7,$9F,$BF,$6F,$B7,$1F
	.dc.b	$31,$BA,$7C,$B8,$FE,$25,$F0,$DF,$CD,$88,$FF,$30,$7A,$3C,$4F,$C1
	.dc.b	$94,$F9,$45,$CF,$DE,$FE,$C0,$FE,$6F,$97,$E9,$5A,$7F,$02,$3F,$9A
	.dc.b	$39,$7C,$E6,$8C,$9B,$30,$7F,$42,$F8,$6F,$E9,$A4,$7F,$A3,$BF,$82
	.dc.b	$C5,$E5,$7F,$33,$3E,$B5,$7F,$70,$F8,$B7,$C1,$53,$ED,$7C,$3F,$F5
	.dc.b	$82,$63,$FD,$49,$E9,$7C,$E3,$EB,$1F,$CF,$6B,$9B,$6B,$F0,$9F,$0F
	.dc.b	$FD,$88,$91,$FD,$71,$E6,$FF,$3B,$AF,$2B,$C4,$87,$CD,$F8,$6F,$ED
	.dc.b	$64,$7F,$B2,$BC,$9F,$D5,$C7,$F0,$DB,$F1,$C7,$F2,$16,$ED,$FE,$BC
	.dc.b	$6E,$FF,$84,$3F,$DF,$0B,$E1,$BF,$BD,$B6,$BF,$DA,$6D,$7F,$DC,$4D
	.dc.b	$7E,$C5,$7F,$63,$BF,$8D,$1F,$87,$9E,$E8,$38,$6E,$5E,$93,$C2,$47
	.dc.b	$C2,$FB,$EB,$B4,$7E,$EA,$7D,$EF,$53,$B3,$87,$DB,$3C,$DF,$4F,$C3
	.dc.b	$7B,$88,$FB,$9E,$D8,$F3,$DB,$F1,$73,$DE,$BF,$15,$F0,$72,$3E,$BE
	.dc.b	$3F,$37,$E1,$BE,$49,$23,$E4,$BD,$1F,$85,$7D,$1F,$8D,$7C,$DF,$B1
	.dc.b	$7F,$95,$7F,$1D,$F0,$DF,$50,$8F,$D4,$7A,$3F,$D7,$B4,$FC,$14,$F3
	.dc.b	$C7,$F0,$0F,$39,$FC,$27,$C3,$7D,$E2,$3F,$79,$E7,$8F,$96,$BF,$94
	.dc.b	$EA,$BE,$9A,$7E,$21,$E9,$3F,$8D,$F8,$4F,$C4,$93,$1F,$C4,$BD,$32
	.dc.b	$F2,$C7,$EA,$6B,$9B,$F6,$34,$F8,$4F,$DB,$3F,$0F,$1F,$94,$23,$FC
	.dc.b	$2E,$FF,$06,$FE,$88,$FC,$24,$69,$FC,$BB,$5E,$98,$BE,$1B,$F3,$E4
	.dc.b	$8F,$CF,$BE,$77,$99,$FE,$31,$F6,$3F,$C7,$C7,$A1,$FC,$DF,$C3,$7E
	.dc.b	$A0,$8F,$EA,$1F,$F3,$2E,$AF,$6B,$FC,$8C,$49,$FF,$1E,$7B,$7F,$E4
	.dc.b	$9F,$72,$47,$F8,$FD,$FF,$43,$F6,$B8,$7F,$8D,$7C,$37,$ED,$89,$5F
	.dc.b	$B6,$7B,$E1,$D6,$3F,$B7,$BF,$E5,$5C,$3A,$C7,$F7,$2C,$2E,$FD,$63
	.dc.b	$F6,$D8,$D7,$F8,$57,$C3,$7F,$04,$23,$FC,$11,$AF,$F0,$46,$BF,$39
	.dc.b	$AF,$F0,$E6,$BF,$B6,$7F,$87,$FC,$2C,$8F,$E1,$ED,$7F,$87,$B5,$FB
	.dc.b	$2D,$7F,$8C,$B5,$FE,$02,$D7,$F8,$7F,$C2,$FF,$1C,$23,$FC,$71,$AF
	.dc.b	$F1,$C6,$BF,$86,$6B,$FC,$99,$AF,$F0,$E7,$F9,$3F,$C2,$C8,$FE,$4E
	.dc.b	$D7,$F9,$3B,$5F,$C8,$B5,$FE,$62,$D7,$F8,$CB,$5F,$E4,$FE,$0F,$F3
	.dc.b	$42,$3F,$CD,$1A,$FF,$34,$6B,$F9,$E6,$BF,$CF,$9A,$FF,$26,$7F,$9F
	.dc.b	$FC,$2C,$8F,$E7,$ED,$7F,$9F,$B5,$FD,$4B,$5F,$E9,$2D,$7F,$98,$B5
	.dc.b	$FE,$7F,$F0,$BF,$D3,$08,$FF,$4C,$6B,$FD,$31,$AF,$ED,$9A,$FF,$56
	.dc.b	$6B,$FC,$F9,$FE,$AF,$F0,$B2,$3F,$AB,$B5,$FE,$AE,$D7,$F8,$0B,$5F
	.dc.b	$EC,$2D,$7F,$A4,$B5,$FE,$AF,$F0,$BF,$D9,$08,$FF,$64,$6B,$FD,$91
	.dc.b	$AF,$F0,$E6,$BF,$DB,$9A,$FF,$56,$7F,$B7,$F1,$48,$FE,$DE,$D7,$FB
	.dc.b	$7B,$5F,$E3,$2D,$7F,$BC,$B5,$FE,$C2,$D5,$F1,$8E,$0F,$F7,$BB,$63
	.dc.b	$F9,$71,$FD,$D5,$C3,$86,$BF,$CC,$DA,$FF,$6F,$6B,$C6,$3A,$91,$EA
	.dc.b	$78,$9C,$3D,$5F,$1F,$0C,$8D,$CA,$31,$1E,$36,$BE,$36,$BE,$36,$BE
	.dc.b	$B6,$BF,$0B,$8B,$E0,$91,$EE,$D7,$E0,$35,$F8,$0D,$7E,$03,$5F,$90
	.dc.b	$FC,$CE,$69,$1E,$A6,$BF,$2D,$AF,$CB,$6B,$F2,$DA,$FD,$36,$BF,$6B
	.dc.b	$AF,$EC,$91,$F8,$ED,$7E,$C3,$5F,$B0,$D7,$EC,$35,$FC,$03,$F8,$4F
	.dc.b	$4E,$47,$D2,$6B,$F8,$36,$BF,$83,$6B,$F8,$36,$BF,$8B,$6B,$F9,$2F
	.dc.b	$2F,$F2,$24,$7E,$FB,$5F,$C8,$35,$FC,$83,$5F,$C8,$35,$FC,$C3,$F9
	.dc.b	$CF,$56,$47,$E2,$9A,$FE,$6D,$AF,$E6,$DA,$FE,$6D,$AF,$E8,$DA,$FE
	.dc.b	$AB,$DA,$FD,$49,$1F,$CB,$B5,$FD,$43,$5F,$D4,$35,$FD,$43,$5F,$D8
	.dc.b	$3F,$B4,$F8,$59,$1F,$A2,$6B,$FB,$6F,$8D,$7B,$7F,$E5,$FE,$BF,$EE
	.dc.b	$D1,$89,$FF,$37,$46,$27,$FC,$13,$C5,$BE,$2F,$DC,$7C,$3D,$BF,$E0
	.dc.b	$0D,$7F,$83,$BF,$E1,$0E,$4C,$71,$A3,$F8,$49,$27,$F6,$EF,$EF,$96
	.dc.b	$5E,$35,$97,$8D,$F2,$E1,$FE,$77,$9D,$4F,$F8,$3F,$B1,$97,$83,$0E
	.dc.b	$D8,$7F,$13,$6B,$FC,$4F,$DD,$FE,$36,$47,$F8,$4B,$E3,$C3,$AE,$3D
	.dc.b	$F4,$F7,$46,$2B,$FC,$3F,$87,$77,$CA,$7F,$88,$B9,$FF,$C6,$26,$5F
	.dc.b	$5D,$FF,$21,$F8,$6F,$E5,$64,$8F,$E3,$2F,$87,$0E,$E4,$8E,$A3,$3F
	.dc.b	$7B,$8D,$5F,$0D,$F4,$09,$1C,$25,$D5,$88,$FE,$6D,$3F,$CF,$08,$FF
	.dc.b	$30,$D3,$DC,$FD,$4C,$3C,$4F,$A8,$F8,$27,$86,$78,$BD,$0F,$E6,$6F
	.dc.b	$F4,$8F,$86,$91,$FC,$D9,$AF,$E5,$67,$AF,$07,$DF,$8E,$C7,$BF,$6F
	.dc.b	$44,$7D,$FE,$CF,$F2,$E9,$8D,$BF,$A6,$7C,$37,$F5,$62,$5F,$F4,$47
	.dc.b	$E6,$3F,$2B,$3F,$69,$EE,$D7,$E5,$C7,$83,$67,$C3,$7E,$68,$91,$F2
	.dc.b	$FB,$EB,$FD,$7E,$3F,$B2,$91,$F9,$5A,$FB,$32,$FC,$54,$7D,$3D,$78
	.dc.b	$BF,$D9,$6B,$83,$CC,$FE,$B7,$F0,$DF,$DC,$08,$E0,$EA,$7A,$EF,$B3
	.dc.b	$F5,$0F,$54,$FF,$A0,$BC,$19,$1F,$87,$8F,$DA,$FC,$37,$F7,$C2,$47
	.dc.b	$F6,$A7,$E3,$BF,$4B,$5F,$8F,$7F,$C1,$23,$F2,$1F,$0D,$FB,$E2,$3F
	.dc.b	$8F,$F1,$2B,$FE,$03,$8F,$F0,$49,$1F,$49,$D6,$7D,$2F,$DF,$3A,$9F
	.dc.b	$DF,$23,$83,$CC,$FE,$FF,$F0,$DF,$E1,$42,$38,$3A,$F1,$FE,$2C,$9B
	.dc.b	$7F,$BF,$F8,$7E,$FC,$7F,$A5,$F8,$58,$7F,$AA,$3C,$69,$87,$88,$FC
	.dc.b	$21,$FE,$22,$21,$FE,$10,$3E,$3D,$76,$47,$EA,$4F,$C0,$1F,$5C,$87
	.dc.b	$C1,$3F,$6D,$E3,$FE,$B9,$FF,$6D,$6B,$FC,$8D,$BA,$7D,$46,$9F,$28
	.dc.b	$A7,$EA,$CC,$1F,$90,$30,$43,$AF,$C3,$7F,$1F,$23,$FC,$05,$F5,$E9
	.dc.b	$F1,$FC,$AF,$F6,$59,$F5,$AF,$F5,$1E,$5F,$FB,$38,$C7,$C3,$7F,$29
	.dc.b	$26,$1F,$C2,$DE,$6F,$FB,$1A,$B9,$BF,$0B,$3F,$EA,$7F,$0D,$FC,$B8
	.dc.b	$91,$FC,$67,$CF,$AF,$F5,$B8,$FE,$6C,$47,$EF,$E5,$E7,$7F,$90,$5F
	.dc.b	$9E,$75,$1C,$4F,$E3,$CF,$F3,$FF,$76,$47,$F2,$16,$BF,$23,$0F,$E9
	.dc.b	$D5,$E8,$8F,$B0,$7E,$96,$BF,$DD,$FE,$94,$8F,$E1,$D1,$FD,$23,$E1
	.dc.b	$BF,$A6,$12,$3F,$97,$3F,$11,$FE,$79,$5D,$6F,$F1,$51,$FD,$93,$E1
	.dc.b	$BF,$A8,$91,$FE,$79,$EA,$D7,$FB,$F4,$7F,$59,$23,$F1,$0F,$A0,$7C
	.dc.b	$43,$E8,$67,$FD,$AE,$37,$1C,$8F,$E8,$4F,$F6,$3F,$86,$91,$FD,$15
	.dc.b	$AF,$C5,$9F,$A0,$FF,$38,$E7,$63,$F8,$7B,$FF,$03,$43,$D7,$A7,$A7
	.dc.b	$2F,$F0,$9F,$C3,$7F,$71,$26,$1F,$D4,$DF,$8D,$5D,$23,$D4,$7D,$29
	.dc.b	$FF,$0D,$FC,$37,$F7,$62,$47,$F6,$1F,$6A,$BF,$E2,$38,$FE,$FC,$47
	.dc.b	$E3,$1F,$48,$EB,$3F,$4B,$95,$3B,$9E,$47,$F6,$77,$FC,$11,$F0,$D2
	.dc.b	$3F,$B5,$35,$F8,$C7,$D2,$3E,$36,$9F,$AB,$9F,$8B,$EE,$49,$FE,$AD
	.dc.b	$1F,$E1,$0F,$86,$FF,$0A,$12,$3F,$BB,$3F,$12,$BF,$11,$BF,$EF,$F2
	.dc.b	$FF,$1E,$FC,$37,$F8,$60,$97,$FE,$03,$F7,$AB,$FE,$45,$8F,$F1,$01
	.dc.b	$1F,$87,$B7,$CF,$7C,$E7,$E2,$0F,$E3,$BA,$A7,$84,$73,$3F,$C1,$4F
	.dc.b	$F8,$AF,$E1,$A4,$7F,$83,$5A,$FC,$43,$E8,$1D,$8F,$B9,$8F,$F8,$FE
	.dc.b	$7F,$13,$F0,$12,$7F,$BD,$47,$F8,$D3,$0F,$F1,$C1,$23,$FC,$34,$FF
	.dc.b	$93,$7E,$19,$D3,$FE,$57,$6C,$FF,$8B,$1A,$FF,$84,$3E,$F1,$F0,$7F
	.dc.b	$E4,$09,$93,$FE,$09,$7F,$CC,$12,$F8,$4C,$CF,$F9,$BE,$7F,$C9,$44
	.dc.b	$7F,$DB,$6F,$6D,$DA,$7F,$46,$94,$E6,$30,$D8,$FE,$97,$7E,$47,$C4
	.dc.b	$7F,$8D,$E3,$B3,$E1,$3F,$C7,$5F,$0D,$FE,$5B,$23,$83,$C7,$97,$C3
	.dc.b	$23,$FE,$4E,$AE,$CF,$F8,$06,$BD,$FF,$0D,$23,$C2,$7F,$81,$7E,$1F
	.dc.b	$FC,$D0,$47,$FC,$94,$FD,$1B,$EB,$1F,$E3,$CA,$F6,$2B,$EF,$97,$FA
	.dc.b	$53,$0F,$F3,$79,$27,$C9,$85,$7F,$D3,$91,$FE,$77,$23,$EC,$3E,$E6
	.dc.b	$1E,$0A,$F9,$F1,$F2,$B4,$E4,$FF,$6E,$AE,$2F,$43,$FC,$C5,$A7,$FA
	.dc.b	$0C,$8E,$0E,$C7,$AE,$FB,$69,$F6,$F2,$F5,$BF,$E9,$7B,$FC,$3E,$92
	.dc.b	$FF,$CA,$51,$FE,$8C,$C7,$FD,$20,$48,$FF,$39,$3F,$25,$F5,$90,$FC
	.dc.b	$B3,$F9,$BD,$E4,$7E,$6D,$FC,$AE,$B8,$8F,$C7,$1F,$E1,$45,$1F,$A5
	.dc.b	$24,$7B,$34,$F8,$EF,$E5,$33,$F6,$BF,$24,$EC,$7E,$72,$78,$BD,$8F
	.dc.b	$CD,$BF,$B1,$E2,$91,$FA,$26,$BB,$BF,$C1,$92,$BF,$5B,$87,$A7,$0F
	.dc.b	$1E,$9F,$EA,$1B,$FC,$B6,$F9,$D7,$F4,$FC,$DF,$BF,$24,$FE,$C1,$F9
	.dc.b	$4F,$D7,$3E,$55,$F4,$C3,$F8,$27,$44,$9F,$E0,$8F,$EF,$B8,$F1,$1E
	.dc.b	$78,$FE,$18,$47,$E4,$61,$FB,$8C,$FE,$3C,$7E,$76,$F8,$BF,$4B,$45
	.dc.b	$F1,$8E,$87,$F0,$6E,$AF,$E3,$04,$70,$76,$3E,$D5,$F6,$D7,$F1,$F1
	.dc.b	$F7,$6B,$F8,$AF,$24,$8F,$59,$FC,$7F,$B7,$F9,$21,$1F,$E2,$AF,$C6
	.dc.b	$D7,$F0,$39,$7D,$39,$FE,$09,$9F,$B2,$31,$1F,$C9,$FA,$3F,$7A,$FB
	.dc.b	$73,$F9,$ED,$3F,$4E,$53,$F0,$0E,$9F,$90,$68,$A6,$8F,$E5,$3E,$0F
	.dc.b	$F3,$B2,$38,$3E,$21,$FC,$2F,$AF,$F9,$A9,$E7,$9F,$E6,$FD,$72,$7F
	.dc.b	$03,$C1,$FC,$BB,$C3,$FE,$8C,$48,$FE,$6E,$FC,$38,$FD,$E5,$F4,$9E
	.dc.b	$27,$EA,$FB,$3F,$A6,$11,$FE,$55,$DB,$5F,$CF,$47,$F5,$32,$3E,$8E
	.dc.b	$25,$75,$8F,$CB,$1F,$C9,$E7,$6F,$23,$FA,$5B,$FD,$77,$C6,$91,$FD
	.dc.b	$39,$AF,$C9,$3F,$AE,$4F,$5E,$0F,$BB,$A7,$95,$FD,$6F,$81,$FD,$70
	.dc.b	$C6,$7F,$D4,$BC,$7F,$ED,$44,$C3,$FA,$DB,$F1,$5F,$EA,$B8,$75,$9F
	.dc.b	$B3,$7F,$8A,$F2,$FF,$B8,$12,$BF,$A7,$78,$55,$FE,$05,$1F,$DD,$C8
	.dc.b	$FC,$5B,$E8,$FE,$80,$FE,$F1,$1A,$78,$9F,$DB,$9F,$F0,$07,$9D,$23
	.dc.b	$FB,$8B,$5F,$89,$7F,$C3,$2B,$AC,$7F,$1F,$3A,$9F,$D8,$FC,$4F,$EF
	.dc.b	$E7,$9E,$7F,$4B,$E8,$FF,$83,$49,$3F,$DF,$9F,$88,$FE,$4C,$FE,$D9
	.dc.b	$5F,$A9,$F4,$FF,$C2,$44,$8F,$EE,$1E,$45,$7F,$92,$47,$F8,$64,$8F
	.dc.b	$C3,$39,$9F,$92,$3F,$C1,$2E,$9E,$27,$F8,$41,$FF,$11,$7A,$D2,$3F
	.dc.b	$C2,$6D,$7E,$1E,$BF,$9E,$7A,$9F,$92,$C7,$F8,$1D,$7F,$88,$3C,$99
	.dc.b	$3F,$E2,$98,$FF,$16,$7B,$1F,$E2,$F2,$47,$F8,$81,$F8,$7C,$BF,$9E
	.dc.b	$8E,$B7,$F9,$27,$FC,$65,$ED,$FF,$8D,$48,$FF,$84,$FC,$CA,$FF,$47
	.dc.b	$8F,$F1,$E1,$1F,$88,$7E,$78,$F8,$9D,$A8,$D4,$71,$3F,$C6,$6F,$F9
	.dc.b	$17,$D5,$91,$FE,$36,$6B,$F1,$0F,$CE,$9E,$8D,$3F,$3A,$7F,$8D,$A7
	.dc.b	$FC,$87,$E7,$49,$F5,$9F,$E4,$D9,$FF,$29,$11,$FF,$21,$B5,$FE,$C4
	.dc.b	$D7,$FA,$33,$5F,$F2,$1B,$5F,$EF,$4F,$F6,$F4,$3E,$3D,$F0,$D9,$8C
	.dc.b	$36,$33,$1C,$5A,$FF,$79,$6B,$FE,$08,$6B,$FE,$57,$6B,$E0,$6B,$D8
	.dc.b	$E9,$1D,$C8,$F7,$F8,$6F,$11,$1C,$10,$F1,$39,$69,$93,$C8,$E5,$A7
	.dc.b	$AF,$E1,$BD,$C4,$70,$43,$DC,$E5,$A6,$4F,$80,$E5,$A7,$C7,$F8,$6F
	.dc.b	$94,$47,$04,$3E,$53,$96,$99,$3E,$63,$96,$9F,$4F,$E1,$BE,$B1,$1C
	.dc.b	$10,$FA,$CE,$5A,$64,$FB,$0E,$5A,$7D,$FF,$86,$FC,$11,$1C,$1F,$8B
	.dc.b	$E1,$7A,$DF,$E9,$5E,$2F,$F2,$1E,$8F,$6B,$1B,$7E,$0D,$FF,$51,$D6
	.dc.b	$30,$FC,$27,$C3,$57,$F2,$8D,$B3,$9F,$BD,$FF,$2A,$E3,$DA,$3F,$2C
	.dc.b	$4A,$FC,$B3,$F9,$9C,$39,$12,$3F,$D3,$18,$75,$9E,$11,$F9,$B6,$BF
	.dc.b	$9B,$F8,$7F,$D1,$11,$FD,$13,$D5,$FD,$5D,$FC,$0B,$ED,$1F,$EC,$5B
	.dc.b	$EC,$EB,$2F,$F5,$D6,$2F,$77,$D8,$6A,$58,$7C,$AF,$87,$FD,$91,$23
	.dc.b	$F6,$4F,$85,$DD,$FC,$D7,$B1,$FA,$FC,$7E,$7B,$83,$F8,$07,$C3,$FC
	.dc.b	$D2,$39,$C5,$EF,$62,$3F,$7D,$3F,$C0,$C8,$FF,$03,$AF,$7E,$3F,$06
	.dc.b	$3E,$69,$DF,$65,$70,$8E,$67,$F0,$B7,$F8,$77,$C3,$C8,$FE,$1C,$D7
	.dc.b	$F8,$75,$F6,$CB,$DD,$1F,$34,$ED,$5F,$85,$99,$18,$3C,$5F,$0F,$FC
	.dc.b	$6C,$8F,$F1,$B7,$E5,$63,$E9,$8F,$96,$7C,$93,$D6,$60,$FE,$51,$F0
	.dc.b	$FF,$95,$23,$F2,$7E,$12,$BF,$C9,$63,$F9,$51,$1F,$90,$7C,$5B,$F6
	.dc.b	$3F,$EF,$8E,$0E,$2F,$F2,$B9,$E0,$F3,$3F,$99,$7C,$3F,$F3,$52,$38
	.dc.b	$3A,$9E,$7D,$7F,$64,$F7,$3D,$78,$FE,$DA,$BF,$59,$F0,$B2,$7E,$C4
	.dc.b	$7E,$C7,$E1,$FF,$A1,$12,$3F,$A1,$3F,$23,$6F,$A7,$7F,$24,$FF,$58
	.dc.b	$E5,$FB,$9F,$87,$97,$F8,$37,$F7,$46,$1F,$E9,$4F,$F4,$F2,$3E,$A9
	.dc.b	$E9,$5D,$8F,$F7,$45,$71,$9F,$EA,$15,$C1,$E6,$7F,$57,$C7,$FA,$D1
	.dc.b	$1C,$1D,$6B,$A3,$F7,$CF,$C6,$1F,$26,$7F,$88,$61,$23,$D4,$7E,$77
	.dc.b	$5F,$EC,$C4,$7F,$B3,$3F,$15,$E9,$8F,$A9,$FB,$C7,$D3,$3E,$B6,$BF
	.dc.b	$82,$5C,$4D,$BF,$85,$C7,$F6,$D9,$FE,$00,$D3,$EF,$94,$FD,$39,$83
	.dc.b	$F4,$E6,$48,$60,$FE,$ED,$C5,$FD,$FC,$8F,$F7,$F7,$D4,$BE,$45,$EA
	.dc.b	$7F,$37,$6C,$7A,$F1,$F7,$3E,$AE,$0F,$E7,$07,$B7,$87,$F9,$04,$7F
	.dc.b	$82,$F9,$BF,$C1,$C4,$7E,$A1,$FE,$16,$BE,$AF,$EC,$5B,$7D,$AB,$FA
	.dc.b	$1F,$07,$F8,$3E,$BF,$A4,$1F,$D1,$F8,$EB,$FE,$19,$3F,$C3,$84,$3B
	.dc.b	$9E,$DC,$BD,$87,$9D,$D1,$FE,$1F,$29,$FE,$2A,$3C,$DC,$7F,$E2,$9C
	.dc.b	$7F,$B1,$C7,$F8,$BD,$AE,$CD,$9E,$07,$D7,$7C,$DB,$1F,$88,$8F,$63
	.dc.b	$FC,$D7,$97,$FB,$11,$EC,$7E,$27,$81,$E1,$23,$E1,$7E,$CD,$74,$8F
	.dc.b	$71,$F5,$EB,$C1,$E2,$7D,$B2,$47,$F4,$3E,$7A,$FA,$4F,$C1,$47,$BC
	.dc.b	$8F,$AF,$2F,$92,$F5,$3E,$D4,$69,$E5,$68,$C4,$7C,$58,$78,$FF,$2C
	.dc.b	$91,$F2,$DF,$9D,$5D,$A3,$F0,$8F,$EA,$67,$A4,$7F,$5B,$E9,$FF,$04
	.dc.b	$1E,$AD,$D9,$7C,$FF,$27,$EC,$12,$BE,$C3,$D9,$FE,$D1,$7F,$24,$F9
	.dc.b	$5C,$1F,$CB,$23,$F2,$9E,$67,$E8,$C8,$FF,$6B,$EC,$AF,$C0,$8F,$C3
	.dc.b	$11,$FC,$36,$7E,$BC,$BC,$B1,$FA,$18,$F6,$4E,$E3,$91,$F8,$D7,$F2
	.dc.b	$1E,$7C,$8F,$C8,$35,$F9,$F3,$D7,$3F,$C7,$3D,$23,$FC,$01,$DD,$FE
	.dc.b	$2D,$3C,$F8,$78,$A3,$F8,$07,$A3,$F9,$D2,$3F,$9D,$7E,$33,$F4,$8F
	.dc.b	$8F,$5F,$4C,$F0,$3E,$B3,$FC,$1D,$E0,$FD,$97,$F0,$11,$C9,$A7,$C4
	.dc.b	$21,$EA,$29,$A3,$04,$30,$7E,$2F,$D6,$FD,$A1,$1F,$DA,$3E,$8B,$F7
	.dc.b	$76,$BC,$3F,$F2,$71,$E8,$AF,$D3,$9F,$E0,$18,$78,$9F,$E5,$33,$5C
	.dc.b	$3F,$ED,$79,$FC,$EF,$9D,$FC,$0C,$8F,$F0,$3F,$63,$F9,$41,$1F,$F1
	.dc.b	$87,$8F,$5F,$E3,$D1,$72,$F2,$F9,$19,$9F,$D3,$4F,$F0,$F2,$3D,$48
	.dc.b	$7E,$B9,$CB,$4C,$9C,$8A,$7F,$11,$F6,$FF,$8D,$11,$FE,$34,$F2,$AF
	.dc.b	$C2,$7C,$93,$E0,$3F,$CF,$0D,$AF,$D3,$DF,$A7,$93,$FC,$BF,$E6,$7F
	.dc.b	$9E,$4C,$4F,$F3,$DF,$80,$FE,$50,$47,$F9,$43,$DE,$7F,$97,$5F,$59
	.dc.b	$FF,$19,$37,$1F,$CB,$6C,$C7,$FA,$A7,$C0,$CB,$FD,$CB,$FE,$8D,$FB
	.dc.b	$EF,$CE,$5F,$08,$FE,$70,$49,$FE,$70,$F8,$5F,$F7,$3E,$1D,$F1,$F1
	.dc.b	$E9,$EA,$9F,$EC,$95,$C5,$FE,$3A,$D7,$F8,$E7,$A5,$FD,$14,$8F,$F4
	.dc.b	$57,$B2,$BE,$CF,$82,$BB,$66,$BF,$57,$A5,$2B,$DA,$7F,$16,$F8,$6F
	.dc.b	$EA,$44,$7F,$A9,$3D,$B0,$EB,$1D,$F0,$EB,$F1,$1F,$DE,$C9,$1F,$E8
	.dc.b	$7D,$5E,$A6,$23,$F9,$CC,$FF,$5F,$23,$FD,$14,$E8,$7F,$A3,$EF,$B9
	.dc.b	$E0,$CF,$EB,$47,$F2,$C9,$E2,$F4,$3F,$B4,$3F,$DA,$FE,$1E,$47,$F6
	.dc.b	$B6,$BF,$D4,$EB,$AE,$3F,$92,$3D,$71,$E9,$F5,$BF,$00,$7A,$78,$3F
	.dc.b	$0F,$F1,$1F,$DD,$C9,$3F,$DD,$DF,$53,$EA,$3F,$1F,$5D,$5F,$CE,$9F
	.dc.b	$D1,$3E,$0F,$FC,$44,$47,$E4,$7B,$55,$FE,$BF,$1F,$E0,$52,$3F,$1F
	.dc.b	$1F,$4B,$F1,$EB,$EF,$CE,$A3,$89,$FE,$0B,$7C,$B2,$F8,$8A,$49,$E8
	.dc.b	$EA,$BD,$11,$EA,$3D,$27,$F8,$AE,$BF,$49,$ED,$C8,$F6,$1F,$67,$E2
	.dc.b	$3B,$91,$F4,$3D,$5F,$41,$0F,$E3,$25,$CF,$F8,$3D,$F8,$9F,$E9,$9C
	.dc.b	$2F,$77,$CB,$5F,$E2,$5C,$FF,$7C,$DB,$EC,$FD,$F6,$B9,$3E,$53,$87
	.dc.b	$F6,$A6,$08,$7E,$1F,$E1,$FE,$49,$1F,$A4,$F6,$C3,$F8,$EF,$C2,$FF
	.dc.b	$1C,$3D,$1D,$3D,$2B,$FD,$85,$93,$E0,$3F,$86,$1A,$8F,$E0,$FF,$11
	.dc.b	$F5,$88,$FD,$F7,$B5,$7F,$1C,$BE,$D3,$F1,$E3,$F8,$FF,$C4,$7F,$2C
	.dc.b	$24,$7E,$CF,$E0,$EB,$FA,$18,$FC,$19,$1F,$7E,$5E,$A7,$F6,$57,$E5
	.dc.b	$7F,$87,$4E,$DE,$47,$E4,$DF,$C6,$FC,$2C,$8F,$CB,$35,$FD,$84,$FA
	.dc.b	$1F,$E6,$6F,$5D,$95,$FC,$BF,$E1,$65,$7F,$03,$8F,$E8,$FF,$11,$F9
	.dc.b	$A2,$47,$E8,$DF,$CE,$DF,$5C,$7D,$53,$F9,$93,$B9,$EC,$21,$F5,$3E
	.dc.b	$1F,$C8,$FE,$2E,$79,$BF,$AF,$6B,$EA,$30,$43,$04,$39,$19,$3F,$2D
	.dc.b	$F1,$1F,$B2,$23,$FB,$B7,$AD,$7F,$66,$E3,$FE,$CA,$7A,$F6,$FC,$05
	.dc.b	$7F,$2D,$AF,$EA,$86,$5F,$E9,$F8,$7E,$F8,$8F,$F0,$77,$AC,$FF,$64
	.dc.b	$AE,$B3,$F0,$AF,$FB,$AE,$7F,$DD,$09,$1F,$CE,$F1,$AF,$F2,$98,$FE
	.dc.b	$16,$47,$F8,$A5,$74,$7F,$B7,$BE,$57,$FA,$B4,$6D,$E4,$7F,$1B,$7F
	.dc.b	$8A,$ED,$23,$F8,$F3,$5F,$E7,$13,$D6,$FE,$34,$7A,$70,$FE,$FF,$7F
	.dc.b	$DD,$F3,$93,$FD,$2E,$3F,$C1,$FC,$1F,$C8,$89,$1F,$CA,$DF,$E4,$B8
	.dc.b	$75,$BF,$8D,$3D,$4F,$EF,$71,$96,$9F,$D7,$F7,$F8,$0F,$F1,$38,$E6
	.dc.b	$FC,$42,$1E,$A2,$9B,$29,$E0,$34,$43,$F8,$F7,$2F,$F3,$A2,$3F,$D0
	.dc.b	$DE,$EF,$F0,$ED,$FF,$C3,$07,$9D,$FF,$03,$4F,$F8,$6F,$86,$47,$F6
	.dc.b	$F8,$FF,$19,$F4,$7F,$46,$24,$7F,$4F,$6B,$FD,$A5,$AF,$F3,$16,$BF
	.dc.b	$E3,$06,$BF,$E2,$27,$FA,$8F,$54,$8F,$EB,$4D,$7F,$BA,$35,$FE,$78
	.dc.b	$D7,$FC,$76,$D7,$FC,$5A,$D7,$FC,$45,$DD,$FD,$78,$8F,$F6,$76,$BF
	.dc.b	$E0,$16,$BF,$D2,$5A,$FF,$92,$1A,$FF,$8E,$9F,$ED,$3E,$04,$8F,$EE
	.dc.b	$4D,$7F,$C1,$8D,$7F,$AA,$35,$FF,$29,$B5,$FF,$22,$B5,$FF,$1D,$69
	.dc.b	$FD,$D8,$8F,$F7,$F7,$FC,$1B,$CC,$E7,$FF,$2A,$74,$4A,$FF,$18,$47
	.dc.b	$F9,$5F,$A6,$47,$F8,$CD,$FF,$2D,$75,$A2,$72,$FF,$AD,$FE,$1D,$E3
	.dc.b	$7F,$86,$9B,$3F,$E3,$56,$BF,$EF,$B6,$BC,$CE,$99,$79,$5F,$F2,$16
	.dc.b	$5E,$57,$FB,$0E,$5E,$47,$FB,$1E,$57,$1F,$EF,$B9,$73,$7C,$9E,$47
	.dc.b	$8C,$95,$E3,$6B,$E3,$6B,$E3,$19,$23,$31,$E7,$8C,$3C,$C8,$7B,$7C
	.dc.b	$9F,$84,$47,$E1,$35,$D1,$82,$18,$21,$E3,$84,$71,$8C,$79,$7F,$3C
	.dc.b	$8F,$CF,$7E,$97,$9B,$F0,$F7,$F4,$36,$7E,$D5,$FF,$B4,$5F,$B7,$3F
	.dc.b	$4F,$91,$E0,$EE,$7E,$1D,$DB,$E1,$0C,$FF,$05,$3B,$B7,$85,$B8,$F1
	.dc.b	$D7,$E0,$DA,$FE,$0D,$AF,$BD,$AF,$A0,$E0,$43,$81,$F8,$DF,$43,$F2
	.dc.b	$04,$70,$6E,$D9,$7B,$7C,$4F,$31,$E5,$9F,$82,$7C,$4B,$F5,$78,$B2
	.dc.b	$32,$7C,$22,$E6,$F4,$BF,$40,$48,$FD,$03,$FA,$3F,$57,$E4,$12,$39
	.dc.b	$EA,$F2,$71,$1F,$2F,$3F,$C7,$56,$34,$F4,$CF,$EC,$6D,$E5,$F0,$8F
	.dc.b	$C3,$57,$77,$F6,$0D,$7F,$60,$D7,$66,$88,$68,$A7,$E4,$7D,$8F,$DE
	.dc.b	$11,$FD,$E3,$D6,$FC,$B1,$FE,$6D,$F3,$3F,$04,$7A,$71,$F5,$C7,$C0
	.dc.b	$F3,$BF,$0C,$65,$FC,$9F,$B7,$FC,$24,$8F,$F0,$97,$F8,$67,$BB,$F9
	.dc.b	$12,$47,$CA,$F4,$6B,$F2,$4F,$CE,$C6,$31,$F2,$EB,$FE,$C2,$9D,$1F
	.dc.b	$BE,$9E,$AF,$F1,$76,$BF,$C5,$DA,$FE,$CD,$AF,$F0,$36,$BE,$03,$91
	.dc.b	$FC,$87,$DF,$FE,$48,$47,$06,$8F,$1C,$BD,$3F,$D5,$1E,$9C,$FB,$4F
	.dc.b	$E2,$3D,$5F,$D7,$19,$7F,$77,$F0,$3F,$CC,$48,$FF,$31,$7F,$9A,$7C
	.dc.b	$27,$EE,$49,$1F,$92,$F6,$2B,$F1,$39,$93,$8D,$7E,$4F,$B3,$AA,$FE
	.dc.b	$29,$3D,$1F,$E7,$CD,$7F,$9F,$35,$D9,$A2,$1A,$29,$FC,$7F,$E1,$BF
	.dc.b	$A5,$11,$FE,$94,$F5,$AF,$B7,$ED,$FF,$0F,$3C,$DA,$7F,$35,$BF,$E2
	.dc.b	$5E,$E4,$C3,$F6,$F1,$FC,$67,$E1,$FF,$AB,$12,$3F,$AB,$3F,$D6,$73
	.dc.b	$FE,$3C,$48,$FD,$BF,$BF,$5F,$8D,$17,$2F,$C6,$FC,$06,$67,$F9,$C4
	.dc.b	$F3,$7F,$B1,$35,$FE,$C4,$D7,$66,$88,$68,$A7,$F4,$3F,$87,$FE,$DC
	.dc.b	$47,$FB,$73,$D6,$BF,$3F,$F0,$5F,$CC,$CF,$36,$9F,$D6,$6F,$F9,$B7
	.dc.b	$C1,$CC,$3F,$8D,$C7,$F3,$DF,$87,$FE,$F4,$48,$FE,$F4,$FF,$7D,$F8
	.dc.b	$7F,$E8,$44,$8F,$E3,$7F,$0D,$5F,$C5,$45,$CB,$F8,$2F,$DF,$FF,$6D
	.dc.b	$7F,$C1,$11,$CE,$7F,$C0,$CD,$7F,$C0,$CD,$7F,$B1,$35,$FE,$E4,$D7
	.dc.b	$FA,$DB,$5F,$EF,$3F,$11,$FE,$13,$23,$FE,$13,$7B,$4F,$91,$FE,$BD
	.dc.b	$47,$D1,$9F,$6A,$FE,$83,$2B,$D9,$A8,$FE,$D6,$7F,$88,$48,$7F,$88
	.dc.b	$4F,$F0,$61,$FE,$1F,$F9,$56,$3F,$DB,$5B,$1F,$D3,$9A,$FF,$16,$A6
	.dc.b	$5B,$A9,$3B,$7D,$3B,$62,$3B,$DF,$27,$FC,$38,$D7,$C5,$3E,$47,$FC
	.dc.b	$79,$19,$6C,$65,$46,$5B,$19,$8C,$11,$C1,$9D,$64,$F7,$94,$62,$3C
	.dc.b	$F3,$F4,$B5,$91,$ED,$21,$F0,$F8,$7E,$B9,$EF,$DB,$FA,$8E,$0D,$FC
	.dc.b	$84,$AF,$03,$E7,$E0,$FA,$04,$8F,$BB,$CB,$5F,$E8,$B4,$9C,$5F,$E6
	.dc.b	$B8,$F5,$5F,$32,$7A,$3E,$D6,$BF,$49,$AE,$CD,$10,$D1,$4F,$AF,$C7
	.dc.b	$F3,$48,$FC,$97,$AD,$7E,$B3,$A3,$F3,$27,$9B,$8F,$F1,$D7,$F9,$CE
	.dc.b	$C9,$87,$E3,$63,$F2,$1C,$DF,$64,$91,$F5,$5F,$B9,$CF,$F9,$52,$47
	.dc.b	$E8,$BB,$6B,$F1,$A2,$E5,$FB,$EE,$FC,$CF,$E4,$27,$9B,$F8,$16,$BF
	.dc.b	$96,$6B,$B3,$44,$34,$53,$F3,$7D,$5F,$91,$23,$F8,$D7,$AD,$7F,$10
	.dc.b	$F0,$BF,$7A,$79,$B4,$FD,$95,$FF,$00,$F0,$E6,$1F,$B1,$8F,$DA,$76
	.dc.b	$FE,$74,$91,$F9,$97,$F4,$1D,$DF,$B9,$24,$7F,$06,$F1,$AB,$F8,$A8
	.dc.b	$B9,$7F,$22,$F2,$B3,$3F,$B4,$9E,$6F,$E9,$5A,$FE,$E9,$AE,$CD,$10
	.dc.b	$D1,$4F,$DF,$F8,$3F,$B5,$23,$FB,$17,$AD,$7F,$34,$F2,$7F,$90,$9E
	.dc.b	$6D,$3F,$8B,$5F,$F2,$4F,$36,$61,$FC,$56,$3F,$8C,$7C,$3F,$F0,$12
	.dc.b	$47,$EF,$5F,$E0,$9E,$37,$F1,$D2,$47,$F2,$AF,$3E,$BF,$AE,$8B,$97
	.dc.b	$F4,$6F,$43,$33,$FC,$62,$79,$BF,$C2,$5A,$FF,$1E,$6B,$B3,$44,$34
	.dc.b	$53,$F9,$1F,$93,$FC,$64,$8F,$F1,$57,$AD,$7F,$58,$F5,$BF,$A2,$9E
	.dc.b	$6D,$3F,$9D,$5F,$FA,$97,$F0,$0F,$F3,$0D,$7F,$98,$9B,$3A,$C3,$F9
	.dc.b	$E7,$99,$FC,$9C,$8F,$F2,$37,$F9,$57,$9B,$FD,$0C,$91,$FE,$90,$A7
	.dc.b	$B1,$88,$FE,$2B,$37,$0F,$07,$B3,$99,$FE,$7F,$3D,$1F,$E6,$0D,$7F
	.dc.b	$A2,$B5,$D9,$A2,$1A,$29,$FD,$2B,$D0,$FE,$80,$47,$F9,$E3,$D6,$BE
	.dc.b	$17,$B5,$F1,$4F,$36,$7F,$D8,$6F,$E4,$7B,$73,$0F,$EC,$11,$FD,$8F
	.dc.b	$D3,$FE,$98,$48,$FE,$90,$FF,$4F,$F5,$3F,$B4,$12,$3E,$37,$BD,$5F
	.dc.b	$E7,$31,$72,$FA,$9E,$AE,$67,$FB,$1C,$F3,$7F,$AA,$35,$FE,$D2,$D7
	.dc.b	$66,$88,$68,$A7,$F6,$EF,$5B,$FB,$21,$1F,$EC,$0F,$5A,$FC,$37,$C0
	.dc.b	$7E,$24,$F3,$69,$FE,$05,$BF,$C6,$7C,$1C,$C3,$FC,$09,$1F,$E0,$7F
	.dc.b	$67,$FB,$81,$23,$FB,$63,$FD,$DF,$E5,$DE,$DF,$FB,$E5,$B3,$FE,$DA
	.dc.b	$6B,$FE,$A2,$E5,$7C,$1E,$EA,$4E,$DF,$CF,$7C,$36,$23,$FC,$37,$7C
	.dc.b	$91,$03,$97,$0D,$76,$D7,$73,$E9,$23,$FE,$15,$8D,$12,$33,$1E,$2F
	.dc.b	$7B,$E0,$11,$E0,$EB,$6F,$DA,$7C,$3F,$ED,$CF,$53,$FE,$D8,$BF,$86
	.dc.b	$FE,$E7,$E1,$FF,$78,$6B,$7F,$4F,$C1,$7C,$D2,$63,$E5,$7E,$87,$C1
	.dc.b	$7C,$F2,$47,$EE,$7E,$1E,$BF,$ED,$0C,$53,$8C,$7F,$84,$7C,$3E,$AB
	.dc.b	$E5,$1F,$2A,$7A,$BF,$0D,$AF,$D2,$6B,$C2,$BA,$BB,$21,$B2,$9B,$F8
	.dc.b	$6F,$C4,$11,$FA,$0F,$6B,$FE,$39,$F0,$FF,$C7,$CF,$46,$7F,$73,$0F
	.dc.b	$E4,$55,$2F,$F1,$D1,$F9,$1A,$FC,$A1,$23,$EE,$3F,$97,$C7,$F2,$C4
	.dc.b	$8F,$E4,$13,$5F,$91,$17,$0F,$E5,$D9,$E6,$7E,$04,$F3,$7F,$10,$D7
	.dc.b	$F2,$ED,$76,$68,$86,$8A,$7C,$BC,$FF,$58,$47,$F2,$8F,$5A,$FE,$81
	.dc.b	$B7,$F4,$33,$CD,$9F,$E8,$2F,$FA,$36,$F3,$0F,$D9,$47,$ED,$76,$FD
	.dc.b	$C1,$23,$F4,$0F,$EE,$F7,$FD,$D1,$23,$F5,$F8,$B0,$FD,$87,$E0,$2E
	.dc.b	$4D,$57,$E3,$EE,$DE,$E3,$F6,$F4,$E8,$DC,$7E,$26,$BC,$0F,$EC,$5A
	.dc.b	$FF,$00,$6B,$B3,$44,$34,$53,$F2,$9C,$7F,$C5,$48,$FE,$F5,$EB,$7E
	.dc.b	$58,$5B,$9B,$F8,$64,$E6,$31,$3F,$C6,$EF,$DB,$97,$C2,$EB,$DB,$67
	.dc.b	$47,$F2,$1E,$6F,$E4,$C4,$7F,$87,$3F,$CA,$7A,$3F,$94,$12,$3E,$7C
	.dc.b	$FC,$EE,$FC,$47,$EE,$E6,$E5,$E5,$ED,$CC,$FE,$CE,$7A,$3F,$C6,$DA
	.dc.b	$FF,$2A,$6B,$B3,$44,$34,$53,$F7,$DD,$7F,$CF,$C8,$FF,$27,$7A,$D7
	.dc.b	$C7,$F0,$BE,$1C,$FE,$01,$F3,$E3,$FC,$D2,$F3,$3F,$43,$C5,$BB,$1D
	.dc.b	$C7,$F3,$8E,$EF,$E9,$C4,$7F,$9B,$35,$FE,$8A,$D7,$F8,$A3,$5F,$E8
	.dc.b	$6D,$7A,$1F,$CF,$FB,$FF,$AD,$11,$C1,$0F,$EA,$2E,$5A,$64,$FE,$44
	.dc.b	$E5,$A7,$F4,$BF,$03,$FB,$31,$1C,$10,$FE,$BA,$E5,$A6,$4F,$E5,$CE
	.dc.b	$5A,$7F,$57,$F0,$7F,$B9,$11,$C1,$0F,$ED,$2E,$5A,$64,$FE,$74,$E5
	.dc.b	$A7,$F6,$3F,$0B,$FB,$F1,$1C,$10,$FE,$EA,$E5,$A6,$4F,$E8,$CE,$5A
	.dc.b	$7F,$6F,$F0,$FF,$C1,$44,$70,$43,$FC,$02,$E5,$A6,$4F,$EA,$4E,$5A
	.dc.b	$7F,$7B,$D3,$FC,$2C,$47,$07,$F6,$03,$FC,$17,$F3,$2F,$1B,$F2,$AD
	.dc.b	$8F,$D9,$9A,$8F,$CB,$F8,$DA,$64,$68,$4F,$15,$C3,$FC,$2F,$C9,$DC
	.dc.b	$78,$67,$4F,$F7,$26,$BF,$E1,$D6,$BE,$13,$D0,$76,$AA,$D3,$31,$C1
	.dc.b	$C1,$DE,$37,$19,$27,$91,$82,$4F,$27,$73,$FB,$6F,$36,$4F,$BE,$BD
	.dc.b	$D8,$7F,$2B,$6E,$5F,$D1,$BC,$BF,$DE,$9F,$02,$75,$E5,$F8,$89,$7E
	.dc.b	$87,$CB,$E6,$7D,$22,$47,$CD,$E5,$7A,$18,$8F,$9B,$8F,$C7,$AC,$61
	.dc.b	$FC,$E7,$D4,$D5,$7C,$D9,$EA,$FC,$66,$BF,$4D,$AE,$CD,$10,$D1,$4F
	.dc.b	$B1,$E7,$FC,$C2,$3F,$49,$EB,$5F,$D2,$FD,$2F,$E3,$E7,$9B,$0F,$45
	.dc.b	$7B,$7D,$7F,$E4,$66,$5F,$CA,$FA,$5F,$60,$8F,$DE,$7E,$DF,$A7,$F9
	.dc.b	$82,$47,$C7,$F6,$EB,$F1,$8F,$CF,$C6,$31,$FE,$BD,$EC,$EA,$7F,$25
	.dc.b	$3D,$1F,$C5,$B5,$FC,$C3,$5D,$9A,$21,$A2,$9F,$9D,$F5,$BF,$22,$47
	.dc.b	$F2,$EF,$5A,$FE,$E1,$EF,$FF,$44,$3C,$D9,$FE,$8E,$FF,$A3,$FC,$1C
	.dc.b	$C3,$F2,$51,$FB,$8F,$67,$F3,$A4,$8F,$D1,$BF,$A0,$F6,$BF,$7A,$48
	.dc.b	$FC,$67,$C1,$D7,$E3,$1F,$BA,$8C,$65,$FE,$07,$F8,$1F,$F3,$3B,$F9
	.dc.b	$F8,$E9,$3F,$B0,$6B,$FB,$D6,$BF,$BD,$6B,$FB,$D6,$BF,$C4,$9A,$FF
	.dc.b	$0A,$F7,$7F,$6E,$47,$F7,$AF,$69,$F2,$7C,$47,$F6,$CB,$3E,$8D,$3B
	.dc.b	$5F,$F1,$59,$7C,$46,$CD,$47,$F2,$2F,$80,$FE,$0A,$47,$F8,$5B,$FC
	.dc.b	$23,$E0,$7F,$92,$92,$3A,$7C,$3D,$7E,$4F,$7A,$71,$9F,$C9,$F8,$7D
	.dc.b	$57,$F1,$F9,$E8,$FF,$18,$6B,$FC,$95,$AE,$CD,$10,$D1,$4F,$E5,$5F
	.dc.b	$09,$FC,$78,$8F,$F2,$47,$AD,$7D,$BF,$87,$F9,$07,$9B,$5F,$E6,$17
	.dc.b	$F2,$FE,$1E,$61,$FC,$7E,$3F,$A2,$ED,$FC,$A8,$91,$FC,$C1,$FE,$5B
	.dc.b	$3F,$D2,$89,$1F,$46,$2B,$F1,$A2,$E1,$F8,$DC,$73,$3F,$D0,$E7,$9B
	.dc.b	$FC,$F1,$AF,$F4,$96,$BB,$34,$43,$45,$3F,$A7,$63,$FD,$08,$8F,$F4
	.dc.b	$87,$AD,$7E,$7F,$0F,$C6,$1E,$6C,$7F,$AA,$5F,$E3,$F5,$98,$7F,$43
	.dc.b	$8F,$ED,$39,$FF,$4E,$24,$7F,$54,$7F,$A8,$EB,$FD,$B8,$91,$F9,$5E
	.dc.b	$0A,$FE,$2A,$2E,$5F,$B1,$E1,$CC,$FF,$67,$9E,$6F,$F6,$06,$BF,$DB
	.dc.b	$5A,$EC,$D1,$0D,$14,$FE,$E5,$C1,$FD,$98,$8F,$F6,$C7,$AD,$7F,$03
	.dc.b	$E3,$FD,$81,$E6,$D3,$FB,$C5,$FE,$CF,$9E,$61,$FD,$9E,$3F,$C1,$BC
	.dc.b	$7F,$DC,$89,$1F,$DE,$1F,$EE,$AF,$F8,$48,$87,$EE,$79,$6B,$FA,$E2
	.dc.b	$E3,$FC,$57,$AB,$31,$FE,$0B,$39,$9F,$E0,$42,$9F,$E1,$12,$9B,$34
	.dc.b	$43,$45,$04,$1E,$63,$9F,$FC,$26,$D8,$FF,$0B,$B5,$FF,$18,$35,$F0
	.dc.b	$1F,$CC,$39,$9D,$DF,$C6,$CF,$6E,$BF,$E2,$39,$FE,$3D,$DF,$1F,$F0
	.dc.b	$91,$FE,$8C,$B7,$57,$88,$91,$E2,$7C,$BD,$7B,$24,$7F,$25,$F0,$EB
	.dc.b	$FD,$0A,$93,$8D,$7F,$A1,$78,$FA,$AF,$84,$7C,$29,$EA,$FC,$06,$BF
	.dc.b	$01,$AF,$0A,$EA,$EC,$86,$CA,$6F,$BB,$E6,$91,$F9,$AF,$6B,$FE,$AD
	.dc.b	$E3,$FF,$43,$3D,$1A,$7D,$7B,$EF,$7F,$D7,$BF,$02,$FF,$71,$5F,$FB
	.dc.b	$8D,$B5,$BC,$7E,$67,$83,$F7,$09,$7F,$71,$FB,$FE,$17,$D1,$24,$7F
	.dc.b	$98,$ED,$E5,$E2,$3E,$64,$DC,$7C,$1E,$66,$67,$E2,$4F,$47,$F1,$2D
	.dc.b	$7F,$12,$D7,$66,$88,$68,$A7,$CF,$F1,$7F,$2A,$47,$F2,$AF,$5A,$F8
	.dc.b	$5E,$6F,$FB,$5B,$13,$E7,$D3,$E5,$61,$9D,$66,$3F,$E8,$15,$1F,$93
	.dc.b	$F1,$FF,$46,$49,$FD,$1B,$FA,$6F,$23,$F3,$04,$8F,$91,$E7,$57,$F2
	.dc.b	$51,$73,$FA,$DE,$7E,$67,$F1,$53,$CD,$FD,$7B,$5F,$D7,$B5,$D9,$A2
	.dc.b	$1A,$29,$F9,$5F,$33,$F7,$64,$7F,$76,$F5,$AF,$C4,$7A,$7F,$1C,$F3
	.dc.b	$69,$F8,$FB,$F9,$5E,$94,$C3,$E3,$47,$ED,$FC,$FF,$E0,$E4,$8F,$E0
	.dc.b	$EF,$F0,$AF,$43,$F7,$84,$8F,$C6,$7A,$D5,$FD,$B4,$5C,$BF,$35,$ED
	.dc.b	$66,$7F,$5D,$3C,$DF,$E2,$6D,$7F,$89,$B5,$D9,$A2,$1A,$29,$FB,$9F
	.dc.b	$4F,$F8,$F9,$1F,$E3,$EF,$5A,$FD,$67,$B3,$F8,$B3,$CD,$A7,$EC,$EF
	.dc.b	$F1,$DE,$EC,$C3,$F1,$51,$FC,$6F,$D6,$FE,$56,$48,$FE,$56,$FF,$2E
	.dc.b	$F5,$FF,$90,$12,$3F,$61,$F0,$35,$FE,$35,$17,$2F,$DF,$7B,$F9,$9F
	.dc.b	$E2,$53,$CD,$FE,$6E,$D7,$F9,$BB,$5D,$9A,$21,$A2,$9F,$C7,$7D,$AF
	.dc.b	$E8,$64,$7F,$A1,$BD,$6B,$F8,$87,$C2,$7E,$BC,$F3,$69,$FC,$5E,$FF
	.dc.b	$65,$F0,$B3,$0F,$D7,$47,$F3,$FF,$77,$FA,$79,$23,$FA,$7B,$FD,$4B
	.dc.b	$DE,$FE,$88,$48,$FE,$29,$F0,$B5,$FE,$7D,$17,$2F,$E4,$5F,$0B,$99
	.dc.b	$FE,$6D,$3C,$DF,$EB,$6D,$7F,$AD,$B5,$D9,$A2,$1A,$29,$FD,$07,$E0
	.dc.b	$7F,$B3,$91,$FE,$CE,$F5,$AF,$E6,$9F,$0B,$FC,$4C,$F3,$69,$FC,$EE
	.dc.b	$FF,$8B,$7C,$2C,$C3,$F8,$94,$7F,$63,$F8,$4F,$EE,$64,$8F,$EE,$6F
	.dc.b	$F7,$6F,$85,$FE,$D0,$48,$FE,$FB,$8B,$0F,$EF,$DF,$81,$AB,$55,$FD
	.dc.b	$95,$B5,$B8,$FF,$02,$CA,$F7,$1F,$D7,$EB,$C0,$FF,$81,$1A,$FF,$81
	.dc.b	$1A,$EC,$D1,$0D,$14,$FE,$E3,$F3,$8C,$BF,$B9,$C7,$F8,$56,$30,$E0
	.dc.b	$38,$76,$D7,$C0,$7F,$B0,$AD,$A7,$F8,$7F,$1F,$74,$A7,$2F,$F4,$1C
	.dc.b	$3E,$1E,$5F,$23,$3E,$0D,$9D,$1F,$39,$F3,$69,$E5,$8F,$2B,$72,$34
	.dc.b	$C6,$BD,$27,$FC,$67,$28,$C4,$7B,$9B,$8B,$7C,$D6,$EB,$C4,$72,$AC
	.dc.b	$9C,$38,$EC,$7C,$99,$E0,$FC,$89,$F1,$B9,$6B,$93,$04,$30,$43,$EC
	.dc.b	$BE,$FE,$09,$1D,$DA,$F0,$D7,$E8,$E2,$E1,$F6,$9E,$69,$FF,$7C,$E1
	.dc.b	$E6,$AD,$F3,$7C,$23,$4F,$DA,$87,$1F,$CF,$25,$7B,$DF,$A5,$C9,$F2
	.dc.b	$09,$1F,$20,$FC,$17,$56,$1F,$0C,$5B,$F3,$F4,$66,$7F,$2F,$3D,$1F
	.dc.b	$98,$D7,$F2,$6D,$76,$68,$86,$8A,$7E,$8B,$9F,$F1,$04,$7E,$C3,$D6
	.dc.b	$BE,$4F,$6F,$E1,$0F,$37,$1F,$E9,$EF,$F0,$FE,$0C,$C3,$DD,$A2,$7F
	.dc.b	$13,$D5,$F9,$52,$4F,$E1,$5F,$CC,$75,$FE,$38,$91,$F1,$F8,$5D,$F8
	.dc.b	$8F,$AF,$37,$4F,$BB,$E2,$66,$7F,$7B,$3D,$1F,$C9,$35,$FD,$CB,$5D
	.dc.b	$9A,$21,$A2,$9F,$C2,$3B,$BF,$5C,$47,$F3,$CF,$5A,$FC,$7F,$93,$FA
	.dc.b	$A3,$CD,$B7,$F0,$DB,$FD,$6F,$8D,$30,$F8,$F1,$FA,$EF,$07,$F7,$44
	.dc.b	$8F,$D5,$35,$FE,$22,$D7,$F8,$DB,$5F,$E3,$4D,$7F,$7E,$FF,$05,$F0
	.dc.b	$A4,$7E,$D5,$AF,$F1,$C6,$BF,$C9,$9A,$FF,$25,$6B,$FC,$29,$AF,$F2
	.dc.b	$6F,$0F,$F8,$91,$1F,$E0,$4D,$7F,$94,$B5,$FE,$66,$D7,$F9,$93,$5F
	.dc.b	$E2,$EF,$F1,$DF,$12,$47,$F0,$D6,$BF,$CD,$3C,$5F,$E2,$E4,$7F,$15
	.dc.b	$E6,$57,$F7,$F1,$75,$FE,$0B,$E7,$E6,$7F,$A2,$CF,$F2,$B2,$3F,$C6
	.dc.b	$DA,$F7,$30,$43,$04,$39,$9F,$C8,$FC,$8F,$E6,$C4,$70,$71,$AF,$E2
	.dc.b	$FE,$77,$F3,$23,$CD,$BF,$F5,$5B,$FE,$6B,$E9,$4B,$FD,$7C,$7F,$36
	.dc.b	$F2,$FF,$A1,$12,$3F,$99,$3F,$D1,$F9,$BF,$DD,$98,$FF,$54,$FC,$13
	.dc.b	$FD,$A6,$DA,$FE,$83,$EB,$FF,$44,$97,$B5,$49,$3F,$C6,$E1,$EF,$7F
	.dc.b	$46,$F4,$3F,$A9,$12,$7F,$A1,$3E,$1B,$E9,$10,$FD,$F3,$D1,$61,$E5
	.dc.b	$C3,$0F,$9E,$FF,$A9,$FA,$5F,$D7,$C9,$5F,$D3,$DF,$EC,$98,$BD,$CF
	.dc.b	$51,$27,$D1,$F0,$3E,$DF,$85,$F7,$7C,$1C,$9F,$64,$7F,$60,$F5,$FF
	.dc.b	$B6,$12,$3F,$AE,$3D,$AF,$A4,$7A,$38,$CC,$C7,$C4,$F8,$0E,$A7,$C9
	.dc.b	$C3,$C0,$F1,$38,$61,$FD,$C2,$BF,$BE,$12,$3F,$B6,$3F,$E0,$1F,$9E
	.dc.b	$7B,$3F,$DE,$9B,$3F,$D3,$DA,$F5,$18,$3D,$EE,$15,$27,$83,$F8,$1F
	.dc.b	$87,$C4,$7F,$89,$AF,$9B,$EE,$3D,$EE,$1C,$35,$CC,$F9,$C8,$FF,$8B
	.dc.b	$63,$49,$A4,$8D,$C7,$87,$DC,$F8,$04,$79,$3A,$E1,$FC,$E7,$C3,$FE
	.dc.b	$64,$F4,$BF,$E3,$2B,$F8,$6F,$E8,$7E,$1F,$F3,$C6,$B4,$F4,$FB,$FF
	.dc.b	$34,$98,$F9,$DF,$C6,$FC,$3F,$E9,$7E,$1E,$4F,$E8,$E3,$F5,$1F,$0F
	.dc.b	$23,$F4,$CF,$EA,$FE,$1F,$EB,$7C,$27,$D9,$24,$FC,$37,$AB,$FA,$DB
	.dc.b	$EA,$FD,$9E,$26,$1D,$63,$18,$5D,$3F,$79,$8F,$E0,$FE,$17,$F0,$C4
	.dc.b	$AF,$A0,$FE,$9B,$E1,$FF,$7C,$48,$FD,$E7,$C3,$FF,$01,$9F,$E0,$5A
	.dc.b	$49,$FE,$01,$1F,$8F,$9F,$C9,$12,$3E,$E3,$D5,$FE,$0D,$5D,$5F,$54
	.dc.b	$75,$7F,$85,$57,$57,$DD,$5F,$9B,$AF,$CF,$12,$3F,$10,$FE,$86,$FF
	.dc.b	$40,$48,$FE,$29,$AD,$7F,$41,$C0,$9C,$65,$FC,$87,$8B,$55,$F9,$B3
	.dc.b	$F2,$53,$D5,$FC,$B3,$5F,$D1,$35,$E1,$5D,$5D,$90,$D9,$4D,$E9,$FB
	.dc.b	$92,$3F,$A3,$7B,$5F,$F3,$5E,$3F,$E6,$47,$A3,$87,$F5,$B8,$7F,$9E
	.dc.b	$3D,$BF,$F4,$6D,$7F,$A3,$9B,$3A,$C3,$F8,$17,$17,$F0,$72,$3F,$B1
	.dc.b	$7F,$D1,$52,$EA,$E9,$D9,$2B,$FC,$B1,$39,$6F,$81,$E1,$23,$AF,$0B
	.dc.b	$F8,$97,$2F,$F1,$52,$47,$EF,$DE,$F8,$75,$8F,$E2,$96,$C3,$D9,$3D
	.dc.b	$27,$38,$5C,$FD,$D8,$FF,$1F,$E7,$FE,$44,$4A,$FE,$18,$FB,$3C,$6F
	.dc.b	$8C,$48,$F5,$F8,$9F,$27,$C9,$F9,$5E,$54,$9F,$7C,$7F,$2B,$EC,$FE
	.dc.b	$5C,$48,$FE,$30,$F5,$7E,$2D,$F5,$7D,$77,$D5,$F9,$37,$D5,$F7,$DF
	.dc.b	$F3,$7E,$DF,$E7,$44,$8F,$E4,$8F,$F3,$DE,$EF,$E7,$84,$8F,$A7,$E5
	.dc.b	$D7,$F3,$B1,$75,$FC,$17,$99,$99,$FC,$AC,$F3,$7F,$98,$35,$FE,$78
	.dc.b	$D7,$66,$88,$68,$A7,$E8,$3C,$2F,$EA,$44,$7F,$9E,$3D,$6B,$F2,$9E
	.dc.b	$87,$E5,$CF,$37,$3F,$F4,$6B,$FC,$D7,$9D,$30,$FE,$B1,$1F,$D6,$FC
	.dc.b	$4F,$EB,$C4,$8F,$E9,$0F,$E6,$3D,$3F,$D1,$7B,$12,$7F,$2F,$1F,$A5
	.dc.b	$F5,$64,$7E,$69,$FD,$47,$B7,$FD,$9F,$CA,$FE,$D4,$49,$FE,$A8,$F5
	.dc.b	$7F,$43,$7D,$5F,$CB,$DF,$47,$F3,$37,$75,$FD,$96,$1F,$DC,$FC,$AF
	.dc.b	$EE,$C4,$AF,$EC,$0F,$EC,$3D,$DF,$DD,$12,$3F,$5F,$EF,$7E,$F7,$E0
	.dc.b	$7F,$7D,$F0,$B2,$7F,$67,$1F,$DF,$FC,$FF,$F0,$11,$23,$FB,$63,$D5
	.dc.b	$FD,$CD,$F5,$7F,$5F,$7D,$5F,$DE,$DF,$57,$F6,$77,$FE,$0B,$F4,$3F
	.dc.b	$C1,$C4,$8F,$EF,$0F,$F8,$45,$FF,$08,$10,$FF,$6F,$7E,$0A,$F8,$5C
	.dc.b	$47,$F3,$22,$EA,$E5,$FF,$05,$53,$E1,$F7,$5F,$CB,$0E,$C7,$F8,$1C
	.dc.b	$A7,$F8,$4C,$A6,$CD,$10,$D1,$4F,$E7,$DF,$40,$F5,$7F,$A0,$47,$F8
	.dc.b	$6E,$30,$E0,$38,$76,$D7,$C0,$7F,$8B,$69,$F1,$1F,$E2,$5C,$7D,$D3
	.dc.b	$FE,$44,$9C,$BF,$E3,$53,$E1,$E3,$F0,$BE,$23,$83,$63,$47,$F9,$32
	.dc.b	$B9,$BE,$4F,$67,$CC,$D3,$D1,$8A,$32,$D8,$F6,$60,$F8,$8E,$F1,$E5
	.dc.b	$F8,$8A,$A5,$18,$78,$47,$DE,$F8,$8C,$11,$FC,$0F,$9D,$CB,$07,$BB
	.dc.b	$E2,$24,$F9,$9F,$4D,$F4,$8F,$95,$87,$73,$C1,$8F,$58,$F9,$D8,$DC
	.dc.b	$FE,$1D,$B2,$F0,$7B,$DF,$00,$97,$F1,$1E,$FF,$11,$EB,$24,$7E,$63
	.dc.b	$E2,$3D,$DF,$11,$EF,$F8,$89,$3F,$9C,$8F,$67,$C1,$FC,$C2,$47,$D0
	.dc.b	$7A,$BF,$A2,$BE,$AF,$E1,$6F,$AB,$FA,$6B,$EA,$FE,$26,$FE,$47,$C2
	.dc.b	$7D,$82,$47,$DC,$7E,$DF,$C2,$FC,$D2,$47,$EC,$3E,$1E,$BF,$E7,$FE
	.dc.b	$34,$E3,$7F,$8F,$F0,$FA,$AF,$D1,$CF,$47,$F1,$2D,$7F,$3E,$D7,$66
	.dc.b	$88,$68,$A7,$EB,$A3,$F2,$24,$7F,$2A,$F5,$9F,$B3,$3F,$8D,$3C,$DC
	.dc.b	$DF,$B3,$7F,$21,$8C,$BF,$D1,$4E,$5A,$7F,$AC,$30,$63,$FA,$02,$47
	.dc.b	$E9,$1F,$C7,$ED,$F9,$BD,$E4,$FF,$18,$8F,$CF,$72,$C8,$FE,$36,$FE
	.dc.b	$87,$9B,$F3,$9B,$FE,$B0,$93,$FB,$07,$AB,$FC,$8A,$FA,$BF,$C1,$6F
	.dc.b	$A3,$FC,$1E,$EE,$BF,$9A,$C3,$F5,$1B,$FE,$E0,$95,$FB,$C7,$F4,$FD
	.dc.b	$3F,$B3,$24,$7C,$21,$F8,$37,$FB,$F3,$6F,$F5,$B3,$FE,$FA,$8D,$3F
	.dc.b	$EF,$86,$F8,$5F,$C0,$23,$51,$F8,$9E,$3F,$E1,$24,$7F,$86,$BE,$27
	.dc.b	$F9,$FD,$F8,$9F,$F6,$74,$F8,$5F,$F7,$1B,$D3,$2F,$E3,$2F,$86,$37
	.dc.b	$7E,$1E,$EF,$E1,$1C,$BF,$C6,$09,$3F,$C7,$1F,$E3,$7C,$DF,$C3,$49
	.dc.b	$1E,$CF,$0E,$BF,$A1,$8B,$8F,$83,$C8,$CC,$FF,$42,$9E,$6F,$F2,$86
	.dc.b	$BF,$CF,$9A,$EC,$D1,$0D,$14,$FE,$9B,$D5,$FC,$C0,$8F,$F3,$47,$AD
	.dc.b	$7C,$2F,$27,$E8,$1E,$6D,$FF,$A9,$5F,$D3,$F3,$66,$1F,$77,$A3,$F8
	.dc.b	$F3,$7B,$BF,$9E,$91,$FE,$82,$FF,$44,$E0,$FF,$7F,$47,$E0,$3D,$0F
	.dc.b	$BC,$7D,$FF,$33,$F0,$1E,$AC,$9F,$C2,$BF,$83,$F5,$BF,$9D,$78,$5F
	.dc.b	$D2,$89,$3F,$D3,$9E,$D5,$D2,$3B,$1F,$8E,$7A,$3F,$84,$AB,$BF,$E0
	.dc.b	$AF,$FA,$37,$87,$FD,$58,$95,$FD,$69,$F5,$7B,$7F,$95,$24,$7E,$6B
	.dc.b	$DC,$FC,$C7,$BF,$F9,$9F,$7A,$4F,$E7,$A3,$FA,$97,$91,$FD,$88,$91
	.dc.b	$FD,$99,$EA,$FE,$92,$FA,$BF,$9A,$BE,$AF,$EA,$2F,$AB,$F9,$EB,$FE
	.dc.b	$BD,$E4,$FF,$6E,$24,$7F,$72,$7F,$B8,$F9,$5F,$D8,$C9,$1F,$B2,$F8
	.dc.b	$3A,$FF,$37,$8B,$AF,$E9,$BE,$1B,$33,$FE,$18,$9E,$6F,$F7,$E6,$BF
	.dc.b	$E1,$46,$BB,$34,$43,$45,$3F,$DC,$1C,$2F,$3B,$FC,$16,$D8,$FF,$11
	.dc.b	$35,$FF,$13,$B5,$EC,$7F,$00,$F8,$5F,$E1,$47,$AB,$AB,$FC,$59,$7F
	.dc.b	$C3,$7E,$1A,$57,$E6,$E3,$FC,$1F,$6F,$43,$C2,$49,$F0,$BF,$9B,$F8
	.dc.b	$6F,$E2,$DF,$0D,$27,$F8,$DC,$7F,$19,$F8,$59,$1F,$C7,$9F,$F0,$DF
	.dc.b	$E0,$FF,$EE,$48,$1F,$E2,$1F,$57,$DE,$47,$DE,$F7,$7F,$93,$DF,$77
	.dc.b	$F8,$ED,$F6,$7F,$8F,$DD,$E0,$FF,$1A,$E1,$DE,$7F,$C5,$DE,$B7,$CC
	.dc.b	$23,$F3,$1F,$F1,$CC,$7F,$BB,$61,$3C,$BE,$17,$D2,$7A,$BE,$1B,$D7
	.dc.b	$F0,$DE,$CF,$86,$97,$FD,$07,$04,$FC,$CF,$6F,$ED,$92,$7E,$DB,$E0
	.dc.b	$C3,$AC,$71,$7B,$3C,$E7,$AB,$E4,$E3,$F9,$2A,$D8,$79,$23,$FA,$97
	.dc.b	$0B,$DC,$FC,$59,$23,$F1,$6F,$E3,$BD,$EF,$C0,$92,$3D,$5F,$0D,$5F
	.dc.b	$01,$F6,$29,$39,$DB,$E7,$FC,$36,$EB,$F3,$73,$D5,$FC,$CB,$5F,$CC
	.dc.b	$B5,$D9,$A2,$1A,$29,$FA,$6F,$80,$FD,$29,$1F,$D2,$BD,$6B,$F0,$3F
	.dc.b	$0B,$F9,$03,$CD,$AF,$C8,$C7,$D3,$7F,$5F,$E1,$BF,$2A,$69,$F7,$7C
	.dc.b	$1F,$EC,$88,$FE,$C9,$FC,$3F,$C3,$7E,$70,$91,$F8,$1F,$86,$FC,$FF
	.dc.b	$C3,$7E,$82,$A4,$FE,$16,$3F,$5E,$7E,$F8,$8F,$EF,$9E,$AF,$E2,$EB
	.dc.b	$AB,$F4,$8F,$C0,$57,$67,$F2,$15,$D9,$D7,$12,$FF,$7B,$5F,$C3,$09
	.dc.b	$3F,$C3,$1F,$D0,$EB,$FB,$6D,$A4,$FE,$6E,$3F,$73,$C3,$23,$F3,$CF
	.dc.b	$EF,$39,$7F,$84,$E7,$FC,$60,$93,$FC,$61,$EA,$FE,$96,$FA,$BE,$6B
	.dc.b	$E8,$FA,$2E,$F0,$7E,$EB,$0F,$E2,$BA,$FF,$24,$25,$7F,$24,$7F,$93
	.dc.b	$F0,$7F,$1A,$24,$7E,$CF,$9E,$BF,$27,$81,$38,$C3,$F8,$3F,$5E,$AB
	.dc.b	$F9,$9C,$F4,$7F,$98,$B5,$FE,$62,$D7,$66,$88,$68,$A7,$F3,$EE,$3F
	.dc.b	$E7,$A4,$7F,$9E,$BD,$6B,$F8,$CF,$67,$F2,$83,$CD,$C9,$F9,$9B,$FE
	.dc.b	$57,$E0,$4C,$3F,$7F,$1F,$CE,$B9,$BF,$A4,$92,$3F,$A4,$BF,$ED,$68
	.dc.b	$77,$7F,$B3,$9B,$3F,$E5,$96,$BF,$8F,$5E,$1E,$FC,$8E,$1E,$4C,$AE
	.dc.b	$11,$FD,$2B,$AF,$FA,$C9,$23,$FA,$CB,$E1,$C3,$C0,$48,$F0,$5F,$57
	.dc.b	$C5,$7D,$5F,$15,$FF,$56,$EC,$FE,$CA,$48,$FE,$CA,$FB,$78,$7D,$9E
	.dc.b	$7C,$9F,$64,$7B,$FD,$49,$1E,$F7,$E1,$7A,$7F,$D8,$BC,$1F,$EE,$24
	.dc.b	$9F,$EE,$2F,$57,$E3,$5F,$57,$E3,$5F,$47,$E4,$5D,$E3,$71,$7F,$38
	.dc.b	$E0,$FF,$24,$D7,$E4,$FC,$1F,$EF,$E4,$7F,$BF,$BF,$E0,$29,$FE,$EA
	.dc.b	$48,$FA,$DE,$BD,$7F,$39,$16,$FB,$C7,$F6,$EB,$7B,$7B,$8F,$F0,$7D
	.dc.b	$75,$7F,$C1,$AD,$7F,$C1,$AD,$76,$68,$86,$8A,$7F,$54,$1F,$46,$F1
	.dc.b	$71,$1F,$D6,$23,$0E,$0D,$C5,$69,$DB,$0F,$5B,$FB,$8F,$77,$F3,$07
	.dc.b	$BE,$FF,$9E,$1F,$9B,$F8,$29,$3F,$D8,$A3,$FB,$3D,$72,$C1,$E4,$F7
	.dc.b	$6D,$78,$DA,$FA,$1A,$FB,$73,$FF,$6E,$BF,$A7,$F8,$3F,$D5,$1F,$AB
	.dc.b	$F8,$6F,$D6,$7C,$37,$EB,$7E,$1A,$57,$EB,$63,$C7,$9F,$C3,$F3,$7C
	.dc.b	$44,$9F,$13,$FB,$7C,$3A,$C7,$73,$F6,$D8,$76,$7F,$71,$87,$67,$F7
	.dc.b	$18,$76,$7F,$C5,$F9,$3C,$EF,$80,$48,$F8,$0F,$F0,$5F,$85,$FE,$09
	.dc.b	$F0,$D2,$7F,$82,$47,$F0,$7F,$86,$91,$FC,$1D,$FE,$15,$F0,$DD,$E3
	.dc.b	$F0,$3E,$9F,$CD,$23,$F3,$5E,$CF,$F0,$FB,$EC,$FF,$0F,$BE,$AF,$F1
	.dc.b	$1B,$BC,$5F,$82,$E1,$63,$DF,$37,$A9,$F6,$C9,$87,$DB,$7E,$F7,$AB
	.dc.b	$B2,$47,$F1,$FF,$86,$AF,$F6,$DA,$4E,$31,$FC,$8F,$E1,$1F,$EE,$BA
	.dc.b	$FC,$EC,$F4,$7F,$10,$D7,$F1,$0D,$7F,$49,$5D,$1D,$90,$D9,$4D,$C7
	.dc.b	$E5,$7D,$99,$1F,$95,$6B,$C0,$FF,$02,$57,$FB,$F7,$9B,$FC,$8A,$D8
	.dc.b	$FC,$0D,$7A,$E7,$3F,$0F,$F5,$E8,$E6,$BF,$0F,$ED,$7E,$94,$8F,$E9
	.dc.b	$5F,$DB,$4B,$E1,$B9,$12,$7F,$CF,$B9,$FC,$8F,$86,$ED,$F0,$DD,$FE
	.dc.b	$1A,$56,$B3,$F3,$CF,$DB,$F7,$FF,$68,$49,$FD,$A3,$E0,$C7,$B4,$7C
	.dc.b	$07,$D5,$87,$68,$E8,$7F,$A3,$B2,$63,$E0,$8F,$8F,$87,$E2,$7E,$03
	.dc.b	$F8,$11,$23,$F8,$13,$EE,$F8,$3F,$8D,$F0,$92,$7A,$8C,$3F,$DF,$AF
	.dc.b	$D2,$F8,$4F,$98,$47,$70,$F8,$4F,$CC,$1F,$C4,$49,$1F,$C4,$5F,$03
	.dc.b	$C2,$BC,$0F,$DE,$BE,$8F,$73,$AD,$FD,$BA,$FD,$23,$FC,$74,$91,$FC
	.dc.b	$75,$FE,$41,$5F,$A9,$24,$79,$B0,$AF,$E8,$E2,$EF,$FA,$2D,$B3,$3F
	.dc.b	$9C,$9E,$6F,$F2,$96,$BF,$CA,$5A,$EC,$D1,$0D,$14,$FD,$3E,$3F,$CD
	.dc.b	$48,$FF,$35,$7A,$D7,$ED,$38,$3F,$34,$79,$B8,$3F,$A7,$5F,$E7,$78
	.dc.b	$E6,$1F,$63,$91,$3F,$C2,$78,$3F,$A1,$12,$7F,$A1,$3F,$9E,$E4,$FD
	.dc.b	$31,$23,$ED,$F3,$7E,$A7,$B3,$F5,$5D,$B2,$7E,$FC,$7F,$15,$E4,$FE
	.dc.b	$9C,$48,$FE,$9C,$F5,$7F,$0B,$7D,$5F,$DD,$DF,$57,$F1,$37,$D5,$FD
	.dc.b	$FD,$FF,$21,$E5,$FE,$B4,$48,$FE,$B4,$FE,$DB,$C2,$FD,$F7,$87,$27
	.dc.b	$F2,$71,$FC,$07,$C6,$91,$F9,$67,$F9,$D7,$E1,$3F,$EE,$9D,$FF,$7D
	.dc.b	$3F,$D9,$7A,$BF,$B4,$11,$FE,$F5,$D4,$C3,$B4,$7F,$14,$C7,$F8,$74
	.dc.b	$76,$8F,$E2,$F8,$DE,$5F,$EF,$B9,$77,$EB,$FE,$E8,$4A,$FE,$E8,$FF
	.dc.b	$77,$F8,$8F,$E8,$04,$8F,$D9,$CF,$FB,$E6,$BF,$BF,$4E,$67,$F9,$E3
	.dc.b	$EC,$F3,$F1,$1F,$CD,$23,$A3,$FE,$01,$6B,$FE,$01,$6B,$B3,$44,$34
	.dc.b	$53,$F9,$F9,$FE,$0D,$21,$FE,$0D,$3A,$CF,$C0,$F4,$7E,$79,$E7,$7D
	.dc.b	$31,$EC,$F5,$BE,$99,$9E,$9F,$F8,$03,$F4,$CF,$07,$FC,$50,$DA,$FF
	.dc.b	$16,$35,$FF,$18,$B5,$FF,$1B,$66,$F6,$3E,$F9,$23,$E5,$FB,$7F,$82
	.dc.b	$F1,$7F,$07,$EF,$49,$F9,$F1,$FD,$FF,$0F,$EE,$71,$E0,$F1,$BC,$64
	.dc.b	$7F,$C7,$D8,$F7,$7E,$BE,$3D,$DF,$F6,$E4,$31,$F0,$47,$DC,$C7,$C0
	.dc.b	$EB,$1F,$07,$8F,$F0,$09,$1F,$01,$FF,$25,$5B,$E0,$FF,$37,$F0,$92
	.dc.b	$BF,$0F,$1F,$9E,$F8,$59,$1F,$8A,$7F,$43,$F0,$FE,$DF,$37,$E6,$92
	.dc.b	$7E,$6B,$DA,$BA,$47,$AA,$FF,$53,$1D,$23,$DB,$77,$8F,$F3,$F8,$7C
	.dc.b	$9F,$37,$EC,$92,$BE,$CB,$F7,$3C,$FF,$9C,$48,$FC,$FF,$C2,$D5,$7F
	.dc.b	$E6,$6E,$35,$66,$3F,$5B,$F0,$DB,$AF,$F3,$B5,$69,$FC,$3D,$77,$69
	.dc.b	$0F,$C4,$B9,$69,$93,$9B,$97,$AF,$A3,$F9,$42,$3F,$94,$79,$5F,$F0
	.dc.b	$0F,$86,$FE,$1E,$79,$FA,$3F,$3B,$7F,$C4,$BE,$1A,$61,$FD,$C3,$1F
	.dc.b	$43,$F0,$2B,$F4,$7E,$A7,$E8,$89,$1F,$A2,$7F,$07,$F0,$DF,$C7,$09
	.dc.b	$1F,$C2,$3E,$1B,$F8,$FF,$C3,$7F,$20,$F8,$69,$3F,$C3,$23,$F3,$DE
	.dc.b	$CF,$EB,$89,$1F,$AE,$7A,$BF,$C4,$AF,$AB,$F9,$CB,$EA,$FF,$16,$BE
	.dc.b	$AF,$E8,$2F,$F5,$5E,$D7,$EE,$89,$1F,$BA,$7F,$D4,$9F,$85,$7F,$BD
	.dc.b	$38,$3F,$D4,$6F,$FB,$D2,$7F,$C2,$6D,$7E,$20,$0F,$1D,$31,$18,$F8
	.dc.b	$4F,$F3,$C4,$9A,$63,$E1,$7F,$7D,$EF,$7F,$0E,$25,$7F,$0E,$7C,$BF
	.dc.b	$0F,$CB,$E1,$E4,$F9,$23,$9B,$E6,$BA,$F9,$8F,$2D,$FF,$0B,$F8,$1F
	.dc.b	$E3,$64,$8F,$E3,$6F,$F1,$EF,$82,$FE,$20,$48,$F8,$3F,$0B,$5E,$9C
	.dc.b	$A9,$C6,$DF,$23,$E1,$B5,$5F,$A2,$9E,$8F,$F2,$86,$BF,$CA,$1A,$EC
	.dc.b	$D1,$0D,$14,$FD,$6F,$C3,$FF,$34,$23,$FC,$D1,$EB,$5F,$6B,$E1,$FE
	.dc.b	$89,$E6,$E1,$FE,$7B,$7F,$52,$A5,$FE,$86,$3F,$98,$57,$F4,$02,$47
	.dc.b	$F4,$07,$F0,$F1,$F7,$09,$1F,$86,$CF,$EF,$F0,$7E,$03,$8A,$4F,$E2
	.dc.b	$A3,$F9,$E6,$7F,$D3,$09,$1F,$D3,$1E,$AF,$E4,$2F,$AB,$F6,$F8,$98
	.dc.b	$76,$8F,$CA,$E1,$D9,$D6,$1F,$D2,$76,$FE,$B2,$48,$FE,$B2,$FE,$83
	.dc.b	$9F,$F2,$BD,$12,$7F,$3F,$1F,$98,$EC,$91,$FA,$27,$F3,$7D,$7F,$D5
	.dc.b	$78,$FF,$B2,$92,$7F,$B2,$BD,$5F,$D4,$DF,$57,$D5,$7D,$1F,$65,$DE
	.dc.b	$2F,$D8,$E1,$FD,$87,$93,$FB,$89,$2B,$FB,$8B,$FD,$D3,$9B,$FB,$31
	.dc.b	$23,$F6,$FD,$F5,$FC,$BC,$5B,$FD,$EF,$89,$99,$FE,$6F,$3C,$DF,$EF
	.dc.b	$AD,$7F,$BE,$B5,$D9,$A2,$1A,$29,$FD,$0B,$A3,$FC,$12,$47,$FC,$12
	.dc.b	$F5,$AF,$E1,$FE,$37,$F0,$63,$CD,$D1,$FE,$12,$BF,$E1,$3E,$5C,$C3
	.dc.b	$F9,$B4,$7F,$0E,$B5,$7F,$10,$FA,$87,$6F,$F1,$2B,$91,$87,$07,$F1
	.dc.b	$97,$0E,$1A,$F1,$E7,$FF,$0F,$5F,$F2,$4F,$37,$F8,$D9,$E1,$C3,$F8
	.dc.b	$FF,$C0,$7F,$BE,$4F,$F7,$25,$FF,$A0,$BD,$3D,$3F,$C9,$E3,$FD,$C6
	.dc.b	$FF,$AF,$9D,$E2,$F0,$3C,$B3,$E5,$6A,$8D,$3A,$6B,$EC,$CB,$CF,$87
	.dc.b	$F3,$26,$C7,$F3,$76,$BF,$E9,$63,$CA,$7F,$A6,$3D,$2B,$87,$9A,$3F
	.dc.b	$95,$37,$3F,$9F,$1E,$B9,$E3,$83,$C2,$F3,$B6,$7C,$D1,$83,$D7,$7F
	.dc.b	$E8,$08,$EB,$ED,$78,$0F,$0F,$B3,$E2,$F7,$E4,$FC,$77,$CB,$F0,$1E
	.dc.b	$0F,$1F,$B9,$27,$E2,$3D,$AB,$A4,$7C,$FB,$EA,$F3,$21,$EE,$BE,$AF
	.dc.b	$FB,$CB,$27,$91,$EC,$24,$7D,$17,$DF,$E4,$FC,$22,$47,$DD,$F8,$3A
	.dc.b	$FF,$53,$A4,$E3,$7F,$97,$3F,$EC,$8D,$76,$FE,$62,$7A,$BF,$7D,$AF
	.dc.b	$E5,$DA,$EC,$D1,$0D,$14,$FD,$1F,$97,$F5,$88,$FE,$2D,$EB,$5E,$4A
	.dc.b	$7C,$37,$E1,$0F,$47,$1F,$EB,$F0,$FC,$3F,$C3,$CC,$7E,$44,$7E,$13
	.dc.b	$CE,$FC,$19,$23,$F3,$0F,$E0,$BE,$1F,$F1,$E4,$8F,$D2,$7C,$3F,$E4
	.dc.b	$BE,$1F,$F2,$7F,$0F,$27,$F5,$11,$F9,$0F,$4F,$F1,$E4,$8F,$D2,$3D
	.dc.b	$5F,$D7,$5F,$57,$F3,$77,$D5,$FD,$95,$F5,$7F,$3F,$7F,$9C,$F5,$3F
	.dc.b	$36,$48,$FD,$83,$F9,$AF,$87,$FD,$3F,$C3,$C9,$FD,$E4,$7E,$AB,$E1
	.dc.b	$A4,$7E,$FD,$FD,$6F,$C3,$FE,$A3,$DA,$FD,$39,$27,$F7,$8F,$57,$F8
	.dc.b	$35,$F5,$7F,$67,$7D,$1F,$DA,$DD,$E4,$FB,$18,$7E,$D3,$DA,$FD,$99
	.dc.b	$2B,$F8,$43,$FB,$6F,$77,$F7,$24,$8F,$E2,$9F,$09,$5F,$CC,$45,$BF
	.dc.b	$C2,$7C,$2E,$67,$F9,$6C,$F3,$7F,$8A,$35,$FE,$58,$D7,$66,$88,$68
	.dc.b	$A7,$F3,$8F,$7F,$F8,$59,$1F,$E4,$0F,$5A,$FC,$B7,$C2,$FF,$18,$3C
	.dc.b	$DD,$1F,$D0,$EF,$F8,$EF,$E1,$7F,$AE,$AF,$D7,$B6,$75,$87,$EF,$FE
	.dc.b	$0B,$F8,$C9,$1F,$E5,$CF,$F3,$29,$7C,$4F,$42,$4F,$F9,$8E,$FA,$FC
	.dc.b	$3F,$83,$E1,$FC,$3F,$0F,$2B,$59,$7F,$99,$B1,$57,$F2,$AF,$87,$FE
	.dc.b	$52,$4A,$FE,$7C,$F8,$72,$EF,$1F,$E5,$DA,$F1,$BD,$0F,$F3,$A2,$C7
	.dc.b	$C5,$19,$10,$C7,$C9,$1C,$F0,$FE,$77,$1F,$CE,$89,$1F,$D3,$DF,$85
	.dc.b	$87,$CC,$CE,$4F,$5B,$FF,$79,$3E,$5D,$BE,$89,$1D,$6B,$FD,$23,$1F
	.dc.b	$E8,$E4,$8F,$EB,$8F,$67,$A5,$F6,$7E,$06,$1C,$DE,$C7,$86,$FE,$FD
	.dc.b	$FF,$54,$CB,$FA,$99,$27,$FB,$43,$FD,$5B,$3F,$EB,$24,$8F,$57,$0D
	.dc.b	$7F,$9C,$45,$DB,$F1,$3C,$D9,$9F,$F0,$6C,$F3,$7F,$BA,$35,$FF,$06
	.dc.b	$35,$D9,$A2,$1A,$29,$FE,$18,$DF,$FB,$59,$1F,$F0,$03,$D6,$BF,$37
	.dc.b	$CF,$F9,$E3,$CD,$1F,$E2,$3A,$FD,$0F,$64,$BF,$EB,$91,$FE,$00,$F1
	.dc.b	$DC,$7F,$E0,$E6,$CF,$F8,$49,$AF,$F8,$A1,$AF,$F8,$B7,$89,$DD,$FA
	.dc.b	$E2,$47,$E0,$FC,$3F,$D8,$F8,$1F,$B2,$F1,$A4,$FE,$1E,$3C,$5D,$1E
	.dc.b	$12,$47,$85,$FD,$85,$F4,$8F,$D0,$3F,$BC,$BE,$B1,$F9,$1B,$EA,$FE
	.dc.b	$96,$FD,$DD,$3E,$D2,$47,$B5,$FD,$E7,$93,$FC,$1B,$CB,$93,$F9,$B8
	.dc.b	$FE,$13,$E7,$48,$FC,$F3,$FC,$33,$D1,$F9,$5D,$DF,$24,$93,$F2,$5E
	.dc.b	$AF,$E9,$6F,$AB,$FB,$9B,$E8,$FE,$EE,$EF,$2F,$F1,$6C,$3E,$B7,$7F
	.dc.b	$D5,$25,$7D,$57,$EC,$78,$3F,$6C,$91,$FB,$FF,$C3,$3F,$DE,$8F,$FB
	.dc.b	$DF,$8D,$3A,$8F,$F7,$34,$61,$E1,$E9,$E5,$FF,$35,$46,$5F,$C1,$4F
	.dc.b	$81,$FC,$33,$5C,$98,$21,$82,$1D,$1F,$C8,$78,$72,$3F,$20,$D7,$86
	.dc.b	$5F,$E2,$6C,$1E,$CF,$F9,$61,$B3,$FC,$77,$0F,$35,$7F,$8B,$1B,$D4
	.dc.b	$99,$4F,$7B,$91,$F0,$EF,$87,$8F,$FA,$32,$57,$E8,$DF,$F3,$9D,$3D
	.dc.b	$FF,$11,$27,$E8,$7C,$27,$97,$E1,$7C,$DF,$0D,$27,$EA,$47,$EC,$7C
	.dc.b	$BF,$D8,$12,$3F,$60,$F6,$BE,$90,$CB,$E1,$47,$C1,$C3,$B4,$FD,$FC
	.dc.b	$3B,$3B,$D3,$E0,$2C,$99,$7F,$01,$F3,$3F,$80,$12,$7F,$80,$3E,$EF
	.dc.b	$86,$F9,$DF,$0D,$27,$F1,$91,$F4,$7E,$16,$47,$E3,$DF,$A9,$F0,$DF
	.dc.b	$C3,$7D,$1F,$E1,$84,$9F,$E1,$8F,$57,$F2,$D7,$D5,$F4,$1F,$5A,$FA
	.dc.b	$B8,$3D,$57,$8D,$FF,$09,$87,$F1,$BF,$4B,$F8,$D1,$2B,$F8,$D3,$FC
	.dc.b	$77,$D3,$FE,$40,$48,$FD,$37,$C2,$D7,$F4,$71,$6F,$F2,$1F,$0D,$99
	.dc.b	$FD,$1C,$F3,$7F,$93,$35,$FE,$4C,$D7,$66,$88,$68,$A7,$EB,$BD,$6F
	.dc.b	$E6,$44,$7F,$99,$3D,$6B,$F4,$5F,$0D,$FA,$13,$CD,$C5,$F7,$BA,$7F
	.dc.b	$CE,$F7,$FC,$07,$E1,$BF,$4C,$69,$FE,$05,$ED,$7F,$3F,$23,$FC,$FD
	.dc.b	$FC,$EF,$C3,$7E,$B4,$91,$FC,$2F,$E1,$BF,$61,$F0,$BF,$B1,$F8,$59
	.dc.b	$3F,$C3,$E3,$FA,$67,$C0,$FF,$4B,$24,$7F,$4B,$7A,$BF,$C5,$6F,$AB
	.dc.b	$F7,$7A,$98,$76,$8F,$E3,$58,$76,$75,$87,$F5,$9F,$80,$FE,$B0,$48
	.dc.b	$FE,$B0,$FE,$DB,$E1,$7F,$DB,$BF,$86,$FE,$8A,$7F,$14,$DE,$1F,$E1
	.dc.b	$93,$AF,$84,$95,$FE,$27,$AF,$F7,$D9,$FE,$E1,$E1,$F9,$DC,$3F,$D9
	.dc.b	$C9,$3F,$D9,$DF,$03,$C4,$F0,$9F,$C5,$EF,$AB,$E2,$BB,$87,$FB,$0B
	.dc.b	$0E,$D3,$FC,$DF,$E1,$3F,$BA,$11,$FE,$E8,$FF,$77,$F8,$6F,$EF,$44
	.dc.b	$8F,$44,$57,$F9,$EC,$5A,$ED,$8E,$67,$F9,$EC,$F3,$7F,$C0,$0D,$7F
	.dc.b	$C0,$0D,$76,$68,$86,$8A,$7F,$48,$C3,$FC,$18,$47,$FC,$18,$F5,$AF
	.dc.b	$81,$9F,$D6,$3C,$D9,$FE,$6E,$FE,$CF,$0C,$C3,$F9,$F4,$7F,$86,$5F
	.dc.b	$F0,$C1,$0F,$F0,$C1,$F7,$CF,$E4,$39,$B8,$BF,$06,$48,$FB,$DC,$BF
	.dc.b	$86,$E6,$FC,$3F,$4C,$9F,$C1,$47,$F2,$AD,$3F,$DE,$FF,$58,$E1,$ED
	.dc.b	$5F,$D3,$63,$0E,$1A,$F8,$DA,$FE,$4F,$2F,$23,$F8,$FC,$BC,$8F,$E5
	.dc.b	$B2,$F2,$3F,$93,$CB,$C8,$FE,$6B,$2F,$26,$BF,$CF,$B1,$71,$F9,$29
	.dc.b	$18,$7C,$91,$98,$C4,$7C,$7C,$9D,$FF,$A8,$F0,$24,$FE,$8E,$3F,$57
	.dc.b	$E2,$C8,$FD,$33,$FA,$EF,$1B,$C7,$1F,$0F,$9F,$B1,$1F,$88,$F9,$70
	.dc.b	$EF,$1F,$AF,$C3,$BB,$FB,$6C,$3B,$3F,$B8,$C2,$F0,$FE,$9F,$1E,$F3
	.dc.b	$EE,$E8,$F6,$11,$FA,$4F,$BF,$AB,$7A,$1A,$7F,$06,$F2,$6D,$7F,$A9
	.dc.b	$E7,$4D,$9F,$E1,$7E,$56,$A3,$EE,$47,$57,$F0,$0D,$72,$60,$86,$08
	.dc.b	$7F,$15,$9E,$EE,$0F,$B9,$DB,$F5,$C8,$FE,$35,$ED,$97,$F0,$9F,$33
	.dc.b	$F8,$F9,$E5,$7F,$2F,$7F,$C8,$BD,$19,$7E,$B3,$DD,$E0,$7E,$0C,$8F
	.dc.b	$E6,$5F,$AD,$E9,$7F,$2A,$24,$7F,$25,$F5,$FF,$96,$FB,$3F,$4B,$C8
	.dc.b	$FF,$76,$4F,$F8,$BD,$AB,$97,$FD,$17,$9F,$EA,$BC,$5F,$C9,$12,$7F
	.dc.b	$50,$F8,$1F,$E6,$58,$F8,$1F,$E7,$38,$F8,$1F,$F6,$E6,$1D,$8E,$F8
	.dc.b	$76,$E9,$FD,$27,$8D,$F9,$F2,$57,$ED,$5F,$D3,$57,$FB,$AA,$3F,$C1
	.dc.b	$CF,$FB,$A5,$EB,$EF,$FB,$7D,$BF,$77,$C2,$49,$F0,$BF,$07,$E1,$7F
	.dc.b	$1B,$E5,$FE,$B4,$93,$FC,$0D,$EF,$5F,$1E,$E4,$FA,$A3,$A3,$E2,$BB
	.dc.b	$BB,$3F,$F1,$DE,$5F,$EF,$9A,$FC,$B7,$99,$FB,$B2,$3F,$C4,$5F,$DF
	.dc.b	$79,$BF,$A0,$24,$7C,$8F,$84,$AF,$F8,$B7,$A5,$38,$C7,$C1,$3F,$6F
	.dc.b	$E1,$76,$FE,$5E,$7A,$BF,$C7,$DA,$FF,$1F,$6B,$B3,$44,$34,$53,$F4
	.dc.b	$58,$FF,$15,$23,$FC,$AD,$EB,$5E,$48,$43,$E1,$7F,$24,$7A,$73,$FE
	.dc.b	$49,$7F,$95,$F8,$59,$7F,$BD,$8F,$E0,$7E,$9F,$F2,$32,$47,$F3,$97
	.dc.b	$F0,$BF,$0B,$F9,$DF,$85,$93,$F9,$08,$FD,$07,$C2,$C8,$FC,$9B,$FA
	.dc.b	$3F,$85,$FE,$1F,$EB,$FF,$2F,$24,$FF,$45,$7A,$BF,$9A,$BE,$AF,$E4
	.dc.b	$AF,$A3,$F9,$4B,$BB,$7C,$1D,$98,$FF,$1C,$F6,$3F,$9E,$12,$FF,$A9
	.dc.b	$3F,$9E,$F8,$4F,$DB,$12,$3F,$53,$F0,$9F,$B9,$F8,$3F,$DD,$7C,$24
	.dc.b	$9F,$D6,$C7,$F2,$8F,$77,$FA,$41,$23,$FA,$F3,$D5,$FD,$9D,$F5,$7F
	.dc.b	$57,$7D,$5F,$DB,$DF,$57,$F5,$F7,$FC,$D3,$DE,$FE,$A8,$48,$FE,$D4
	.dc.b	$FF,$57,$F7,$FF,$9C,$92,$3F,$80,$FC,$1D,$7F,$35,$17,$6F,$8D,$F0
	.dc.b	$99,$9F,$E5,$D3,$CD,$FE,$EC,$D7,$FB,$B3,$5D,$9A,$21,$A2,$9F,$CE
	.dc.b	$7E,$0B,$FB,$61,$1F,$F0,$13,$D6,$BE,$C7,$C2,$7F,$27,$3C,$DD,$3F
	.dc.b	$DE,$6F,$F9,$57,$C2,$4C,$3F,$A9,$47,$F5,$8F,$88,$FE,$F0,$48,$FF
	.dc.b	$07,$3F,$DF,$FF,$0E,$FF,$7F,$C1,$65,$FD,$99,$B1,$FE,$E3,$CB,$F5
	.dc.b	$51,$C0,$FF,$75,$C7,$FA,$BD,$B1,$AB,$E5,$AF,$F7,$0F,$AE,$65,$FE
	.dc.b	$0F,$6E,$1F,$E2,$46,$BF,$E3,$36,$BF,$E7,$3C,$3C,$18,$F9,$4F,$36
	.dc.b	$5E,$7C,$65,$79,$F3,$FF,$6D,$DF,$91,$B9,$78,$F4,$F1,$12,$BC,$4F
	.dc.b	$C5,$E3,$F8,$44,$8E,$80,$F9,$1C,$9F,$1F,$AB,$E4,$75,$C9,$D9,$F3
	.dc.b	$63,$E1,$70,$FC,$12,$47,$C1,$7C,$0F,$0C,$D8,$78,$63,$E7,$61,$DD
	.dc.b	$D6,$1D,$DF,$A5,$87,$CF,$E2,$F9,$C4,$8F,$9C,$FD,$1E,$4F,$A8,$48
	.dc.b	$F0,$74,$57,$FD,$D5,$C4,$9C,$72,$7E,$0B,$B3,$55,$FE,$A7,$AD,$3F
	.dc.b	$6E,$BB,$34,$87,$DD,$72,$D3,$27,$37,$2F,$5E,$6F,$C4,$11,$FC,$43
	.dc.b	$CA,$FF,$2B,$E1,$7E,$64,$F3,$F1,$FD,$7A,$61,$E6,$7F,$31,$E3,$7E
	.dc.b	$7C,$DB,$F7,$61,$D3,$F9,$72,$47,$E5,$DF,$D5,$79,$1F,$A7,$F3,$24
	.dc.b	$FD,$98,$FD,$57,$9D,$23,$EE,$3F,$AD,$F4,$BF,$49,$DD,$FA,$32,$4F
	.dc.b	$E8,$DE,$AF,$E0,$EF,$AB,$FA,$FB,$E8,$FE,$C6,$EE,$BF,$B6,$C3,$F6
	.dc.b	$1D,$FF,$AF,$25,$7E,$BD,$FE,$05,$E8,$FE,$FC,$91,$F9,$1F,$57,$F8
	.dc.b	$17,$B5,$FC,$0F,$DB,$93,$F9,$58,$FD,$E7,$89,$FB,$B2,$47,$EE,$DE
	.dc.b	$AF,$E6,$EF,$AB,$FC,$1E,$FA,$BF,$9F,$BE,$AF,$F0,$BB,$FE,$11,$E3
	.dc.b	$7F,$07,$24,$7F,$07,$7F,$85,$78,$DF,$C3,$49,$1F,$A9,$F8,$0A,$FE
	.dc.b	$6A,$2E,$BF,$E0,$3F,$C3,$FF,$DF,$9C,$DF,$85,$E5,$57,$4B,$FE,$28
	.dc.b	$D7,$F8,$A3,$5F,$D0,$35,$E4,$D7,$F1,$4D,$7F,$23,$E4,$FF,$21,$23
	.dc.b	$FC,$85,$ED,$87,$97,$E1,$7F,$D2,$F8,$9F,$4E,$3D,$AF,$F8,$6C,$BE
	.dc.b	$0F,$66,$A3,$F3,$5E,$67,$F2,$F2,$3F,$CB,$DF,$F6,$04,$BE,$1F,$BF
	.dc.b	$C3,$CA,$EB,$1E,$2F,$87,$91,$E0,$7C,$BF,$0F,$FC,$F7,$D1,$FE,$78
	.dc.b	$49,$FE,$78,$F6,$AE,$90,$D7,$E9,$61,$EE,$9E,$B3,$9D,$3F,$71,$93
	.dc.b	$2C,$E1,$F6,$72,$FE,$99,$E9,$FF,$4B,$26,$1F,$D2,$DF,$6F,$C3,$FC
	.dc.b	$A2,$47,$C8,$F8,$7F,$9B,$F0,$DF,$3B,$E1,$E4,$FC,$C8,$FE,$B1,$ED
	.dc.b	$FF,$57,$24,$7F,$57,$7A,$BF,$4A,$FA,$BE,$93,$FA,$36,$8C,$3B,$C7
	.dc.b	$D8,$C3,$BB,$B1,$2C,$3F,$B4,$7B,$3F,$D9,$C9,$3F,$D9,$DF,$ED,$5E
	.dc.b	$DF,$F6,$D2,$47,$E1,$3E,$1A,$BF,$CE,$A2,$ED,$FA,$1F,$87,$CC,$FF
	.dc.b	$39,$9E,$6F,$F7,$36,$BF,$DC,$DA,$EC,$D1,$0D,$14,$FE,$89,$EE,$FF
	.dc.b	$7F,$23,$FD,$FD,$EB,$5F,$B3,$F8,$7F,$CF,$9E,$6E,$3F,$DA,$5F,$E8
	.dc.b	$BE,$1E,$61,$FD,$B3,$B7,$FD,$60,$FF,$56,$AF,$F0,$61,$1F,$F0,$63
	.dc.b	$F9,$EF,$87,$FD,$5F,$C3,$C9,$FD,$4C,$7E,$BB,$E1,$E4,$7E,$B1,$FD
	.dc.b	$8F,$C3,$FF,$49,$D3,$FD,$15,$F6,$0F,$88,$FE,$99,$72,$30,$E0,$FF
	.dc.b	$03,$38,$70,$D7,$8F,$23,$1F,$24,$7E,$0E,$19,$7F,$02,$9F,$2C,$E7
	.dc.b	$2B,$8F,$F1,$BC,$BC,$D9,$FF,$5F,$C9,$F0,$FE,$3C,$3C,$6D,$51,$A7
	.dc.b	$4D,$7D,$59,$BE,$23,$F8,$99,$23,$F8,$8D,$7F,$16,$D7,$C3,$1F,$E2
	.dc.b	$5F,$C4,$3F,$DF,$67,$FB,$F2,$FF,$DD,$D1,$F1,$2F,$C5,$87,$58,$F0
	.dc.b	$C6,$0F,$E2,$A7,$77,$E2,$3F,$C9,$30,$F0,$47,$F8,$BA,$FB,$E3,$88
	.dc.b	$FD,$65,$B2,$EE,$48,$F7,$BE,$2D,$3D,$31,$F1,$1B,$27,$37,$FF,$00
	.dc.b	$74,$FC,$1C,$68,$95,$E3,$E1,$AF,$E0,$23,$B3,$F3,$9A,$E4,$C1,$0C
	.dc.b	$10,$FC,$73,$C1,$FC,$86,$FF,$28,$8F,$DA,$7A,$D7,$FB,$86,$DC,$DE
	.dc.b	$A3,$CB,$1E,$AA,$F9,$5D,$7E,$E3,$3A,$7E,$2F,$87,$EB,$92,$7F,$0E
	.dc.b	$FE,$2F,$07,$67,$C6,$F0,$65,$7D,$58,$F9,$3E,$14,$8F,$B0,$FC,$CF
	.dc.b	$17,$E7,$F3,$FE,$10,$93,$F9,$47,$B5,$74,$8F,$E5,$18,$30,$FA,$B3
	.dc.b	$D6,$73,$85,$D3,$E1,$63,$F7,$3A,$3F,$22,$4A,$FD,$0B,$EB,$F1,$BE
	.dc.b	$F9,$23,$F1,$DE,$47,$E0,$BC,$DF,$C1,$F9,$D2,$7F,$25,$1F,$88,$EE
	.dc.b	$FC,$E9,$23,$F5,$AF,$57,$F3,$17,$D5,$F5,$DF,$57,$F3,$97,$D5,$F7
	.dc.b	$DF,$E5,$3B,$FF,$52,$48,$FD,$CB,$FA,$CF,$07,$F3,$24,$8F,$D3,$7A
	.dc.b	$75,$FD,$2C,$5D,$7F,$0D,$EA,$66,$7F,$8E,$4F,$37,$F8,$2B,$5F,$E3
	.dc.b	$6D,$76,$68,$86,$8A,$7F,$27,$F0,$FF,$80,$91,$FE,$22,$F5,$AF,$CC
	.dc.b	$7A,$FF,$B0,$3C,$DC,$FF,$C7,$6F,$F6,$7E,$C4,$C3,$FD,$5F,$DC,$9F
	.dc.b	$DC,$F8,$DF,$C3,$89,$3F,$C7,$9F,$CC,$7B,$7F,$BC,$F8,$19,$3F,$C2
	.dc.b	$E3,$F7,$FF,$05,$23,$F8,$73,$FC,$0B,$E1,$FF,$82,$F9,$7F,$C6,$89
	.dc.b	$3F,$CA,$9E,$AF,$F1,$5B,$EA,$FE,$62,$FA,$3F,$9A,$BB,$B7,$ED,$F0
	.dc.b	$FE,$23,$E6,$7F,$26,$25,$7F,$36,$7F,$9D,$75,$7F,$BE,$89,$F0,$DF
	.dc.b	$C6,$9F,$F7,$CC,$FF,$83,$5F,$F7,$13,$FE,$D9,$5F,$0B,$B3,$87,$C3
	.dc.b	$7F,$20,$F4,$3F,$9A,$92,$BF,$A2,$BE,$0B,$EC,$48,$FF,$6F,$DF,$83
	.dc.b	$27,$1F,$FA,$C6,$3C,$32,$7C,$2F,$0C,$7C,$2F,$DA,$F4,$BF,$A2,$91
	.dc.b	$FE,$AA,$FF,$48,$F4,$BF,$9A,$92,$3C,$3F,$0B,$5F,$E8,$31,$71,$F1
	.dc.b	$C7,$C1,$F8,$5D,$3F,$DE,$67,$A3,$FD,$89,$AF,$F7,$86,$BB,$34,$43
	.dc.b	$45,$3F,$C0,$9E,$AF,$F5,$E2,$3F,$DB,$9E,$B5,$E4,$8F,$89,$F0,$9F
	.dc.b	$64,$F4,$69,$FD,$EE,$FE,$E7,$C2,$CA,$FC,$FC,$7F,$56,$F6,$3F,$B5
	.dc.b	$92,$3F,$BD,$BF,$86,$F8,$5F,$C2,$7C,$2C,$9F,$AF,$1F,$87,$F8,$59
	.dc.b	$1F,$69,$FC,$57,$C2,$FF,$62,$8F,$EE,$E4,$3F,$C0,$E7,$53,$F0,$35
	.dc.b	$D5,$FC,$85,$74,$7F,$25,$57,$4F,$81,$C4,$C3,$FB,$8F,$A4,$F7,$7F
	.dc.b	$C2,$CD,$C3,$FC,$34,$D7,$FC,$5C,$D7,$FC,$67,$4F,$84,$FD,$11,$23
	.dc.b	$F2,$BF,$0B,$FA,$5F,$85,$FD,$37,$C2,$C9,$FC,$CC,$78,$BE,$0F,$C2
	.dc.b	$48,$F0,$BF,$B0,$BE,$91,$FA,$D7,$F6,$17,$D6,$3F,$47,$7D,$5F,$D9
	.dc.b	$5F,$BB,$E0,$FD,$A4,$8F,$6B,$F0,$3E,$13,$E1,$92,$3F,$5B,$F0,$B5
	.dc.b	$FF,$44,$68,$9C,$6F,$F3,$1D,$4F,$F9,$5A,$74,$FC,$A8,$EC,$D2,$1F
	.dc.b	$31,$CB,$4C,$9C,$DC,$BD,$6B,$EB,$91,$FA,$EF,$2A,$FB,$F7,$FC,$54
	.dc.b	$F3,$F3,$2A,$61,$E6,$7F,$8C,$6B,$FC,$6C,$DB,$AC,$FF,$0A,$47,$F0
	.dc.b	$AF,$F2,$4D,$FF,$91,$12,$3B,$63,$FA,$FF,$C4,$7F,$E0,$0C,$78,$5F
	.dc.b	$F7,$73,$AC,$7C,$8A,$3F,$DD,$CE,$4E,$1D,$3F,$E4,$88,$E3,$1F,$53
	.dc.b	$6F,$CC,$91,$FC,$CB,$E5,$7F,$93,$E1,$E5,$7F,$D8,$53,$E4,$7C,$1C
	.dc.b	$DD,$6E,$47,$84,$C9,$74,$70,$7E,$A0,$91,$FA,$87,$FD,$01,$7F,$EF
	.dc.b	$F8,$E4,$BB,$3D,$47,$AF,$A3,$D9,$DF,$2B,$6F,$BF,$C1,$FD,$BF,$2F
	.dc.b	$ED,$89,$3F,$B6,$7B,$D7,$48,$EF,$7D,$5E,$64,$3E,$0D,$F5,$7D,$9C
	.dc.b	$FF,$C0,$88,$FF,$02,$7F,$82,$F4,$7F,$08,$24,$7B,$3C,$6A,$FE,$9A
	.dc.b	$2F,$1F,$E0,$7C,$7C,$CF,$E9,$67,$9B,$FC,$39,$AF,$F0,$E6,$BB,$34
	.dc.b	$43,$45,$3F,$61,$D9,$FC,$68,$8F,$F1,$A7,$AD,$7E,$4F,$CC,$FC,$51
	.dc.b	$E6,$E6,$FC,$8F,$13,$0F,$C7,$79,$33,$1F,$C0,$47,$F2,$8E,$EF,$E4
	.dc.b	$E4,$8F,$E4,$EF,$E0,$3C,$FF,$CB,$12,$3F,$03,$EA,$7E,$67,$D5,$FC
	.dc.b	$D7,$B5,$27,$F0,$B1,$FC,$D3,$C3,$FE,$66,$48,$FE,$66,$F5,$7F,$17
	.dc.b	$7D,$5F,$CD,$DF,$57,$F1,$F7,$D5,$FC,$FD,$FF,$40,$F1,$3F,$9F,$92
	.dc.b	$3F,$9F,$BF,$98,$F6,$FF,$5D,$EE,$C9,$FC,$CC,$7E,$C7,$DE,$91,$F9
	.dc.b	$C7,$F6,$9F,$01,$FD,$33,$C9,$FE,$96,$49,$FE,$96,$F5,$7F,$47,$7D
	.dc.b	$5F,$D9,$DF,$47,$F6,$B7,$79,$3F,$85,$61,$FD,$63,$CA,$FE,$AE,$4A
	.dc.b	$FE,$AE,$FF,$5A,$F2,$FF,$AE,$92,$3F,$63,$F0,$55,$FE,$77,$16,$FF
	.dc.b	$8B,$7C,$16,$67,$F9,$D4,$F3,$7F,$B1,$B5,$FE,$C6,$D7,$66,$88,$68
	.dc.b	$A7,$F4,$5F,$3F,$FB,$79,$1F,$ED,$EF,$5A,$FE,$51,$F0,$9F,$C7,$CF
	.dc.b	$37,$4F,$E5,$AF,$FD,$D3,$F8,$97,$F8,$6E,$BF,$C3,$8D,$9D,$61,$FC
	.dc.b	$5F,$E2,$7F,$BE,$91,$FE,$FA,$FF,$30,$64,$FF,$BB,$E7,$F9,$9B,$63
	.dc.b	$FA,$43,$52,$F8,$5F,$1F,$67,$93,$E1,$65,$7F,$29,$AF,$F7,$1B,$FC
	.dc.b	$A7,$D4,$FF,$07,$91,$FF,$07,$BE,$37,$C3,$F1,$1E,$E2,$3C,$F0,$EE
	.dc.b	$F5,$BE,$EB,$3F,$87,$8F,$F4,$C7,$FC,$81,$F6,$8F,$5F,$B5,$7F,$63
	.dc.b	$8C,$38,$6B,$E3,$6B,$F1,$BE,$17,$E8,$FC,$34,$8F,$74,$7D,$4F,$86
	.dc.b	$91,$F0,$5F,$AF,$ED,$78,$F8,$3F,$AF,$E0,$F6,$FC,$76,$8C,$3E,$38
	.dc.b	$CC,$62,$3E,$37,$0B,$1F,$14,$7C,$1B,$FB,$39,$7E,$1A,$7C,$73,$93
	.dc.b	$ED,$E5,$8D,$3F,$13,$9F,$96,$7E,$27,$BB,$E0,$23,$F1,$5F,$1F,$BB
	.dc.b	$B8,$F4,$37,$43,$F9,$6F,$86,$C3,$27,$FB,$64,$E9,$FC,$37,$C3,$65
	.dc.b	$FA,$13,$D9,$FA,$8D,$72,$60,$86,$08,$7E,$76,$7B,$38,$3E,$8F,$C0
	.dc.b	$FC,$C2,$3F,$81,$7B,$65,$FA,$CF,$85,$FD,$79,$E5,$7F,$17,$7F,$B2
	.dc.b	$F8,$69,$7E,$C8,$F8,$5F,$09,$F6,$09,$1F,$8D,$7E,$FF,$13,$E1,$BF
	.dc.b	$76,$49,$FD,$47,$C3,$7E,$FA,$BF,$7F,$9C,$9F,$D6,$47,$CF,$9F,$C2
	.dc.b	$92,$3F,$34,$F6,$AE,$91,$F8,$13,$F7,$2F,$F0,$A8,$ED,$1F,$B7,$AE
	.dc.b	$CE,$ED,$7F,$76,$FF,$26,$49,$FD,$43,$EC,$C3,$F8,$AE,$F2,$7F,$81
	.dc.b	$47,$F1,$8E,$59,$1F,$C1,$5F,$E3,$7C,$DF,$89,$D3,$F3,$E4,$9F,$DA
	.dc.b	$3D,$5F,$E1,$97,$D5,$F3,$5F,$47,$D1,$77,$5F,$E4,$B8,$7E,$57,$6F
	.dc.b	$D5,$92,$BF,$80,$3F,$AE,$E0,$FC,$D1,$23,$FC,$85,$F8,$9B,$B3,$11
	.dc.b	$F2,$F8,$95,$8C,$7F,$D9,$D7,$FE,$F5,$E8,$FD,$FC,$F5,$BF,$E1,$CD
	.dc.b	$7F,$87,$35,$FC,$C3,$5E,$27,$02,$1C,$0F,$E3,$1C,$9F,$C1,$88,$E0
	.dc.b	$DD,$F9,$3C,$0F,$DC,$60,$7D,$18,$76,$AF,$C3,$4B,$BF,$66,$A3,$F9
	.dc.b	$47,$3F,$F1,$42,$3F,$CA,$5F,$E3,$72,$F0,$FB,$F9,$12,$BE,$5C,$78
	.dc.b	$BC,$99,$1F,$39,$F2,$F9,$DF,$C2,$FB,$3F,$90,$92,$7F,$9B,$3D,$AB
	.dc.b	$A4,$7A,$2F,$DB,$1D,$23,$D7,$77,$5F,$AB,$87,$F1,$7E,$EF,$E5,$A4
	.dc.b	$AF,$E8,$4F,$AB,$CF,$F8,$E4,$8F,$C1,$FA,$7F,$2B,$D1,$F9,$7E,$B4
	.dc.b	$9F,$C3,$C7,$F2,$3F,$0B,$F9,$C9,$23,$FA,$73,$D5,$FC,$6D,$F5,$7E
	.dc.b	$8D,$F5,$7F,$23,$7D,$5F,$AB,$7F,$CB,$FC,$4F,$E8,$A4,$8F,$EB,$4F
	.dc.b	$F4,$8F,$13,$F9,$A1,$23,$F3,$7E,$D5,$7F,$35,$17,$5F,$CC,$FB,$B9
	.dc.b	$9F,$E5,$53,$CD,$FE,$CC,$D7,$FB,$33,$5D,$9A,$21,$A2,$9F,$CD,$7C
	.dc.b	$7F,$EB,$A4,$7F,$B9,$3D,$6B,$F5,$7F,$03,$F9,$83,$CD,$D1,$FD,$AA
	.dc.b	$FF,$37,$F0,$53,$0F,$E9,$B1,$FD,$4F,$CA,$FE,$D2,$48,$FE,$FC,$FE
	.dc.b	$5B,$E1,$7F,$47,$F0,$D2,$7F,$7F,$1F,$A6,$F8,$69,$1F,$C0,$9F,$D4
	.dc.b	$FC,$37,$F5,$FC,$BF,$BA,$92,$7F,$C1,$4F,$57,$F8,$4D,$F5,$7F,$5B
	.dc.b	$7D,$1F,$D7,$DD,$D3,$F8,$4E,$1F,$DB,$7E,$D9,$E8,$FF,$81,$9B,$7F
	.dc.b	$E1,$A6,$BF,$E2,$96,$BF,$E3,$DE,$E7,$C3,$FF,$00,$24,$7F,$B7,$7D
	.dc.b	$EF,$F7,$11,$FE,$EC,$C5,$AF,$FB,$D2,$3F,$DE,$73,$FD,$25,$B3,$A8
	.dc.b	$08,$E5,$3F,$EF,$CC,$1E,$A7,$A4,$95,$E9,$7F,$83,$E1,$E4,$8F,$E2
	.dc.b	$2F,$F8,$36,$FC,$BF,$0F,$7E,$1E,$57,$A6,$3E,$37,$AF,$F1,$49,$1F
	.dc.b	$15,$F9,$1E,$C7,$CB,$24,$7C,$1F,$85,$AF,$F6,$5C,$D3,$8D,$7D,$D3
	.dc.b	$ED,$F8,$5D,$BF,$E6,$9A,$DB,$F4,$EB,$BB,$48,$7D,$57,$2D,$32,$73
	.dc.b	$72,$F5,$F6,$FF,$00,$47,$F0,$0F,$2B,$F2,$E8,$F8,$5F,$90,$7A,$78
	.dc.b	$3E,$96,$1F,$2F,$E1,$A6,$3F,$63,$0F,$6E,$5E,$A8,$F8,$5E,$FF,$E3
	.dc.b	$C8,$FE,$3D,$FF,$2B,$E4,$F8,$6F,$AC,$49,$FC,$67,$C3,$7D,$99,$FB
	.dc.b	$5F,$0D,$27,$F2,$11,$F9,$DF,$82,$FC,$E1,$23,$F3,$8F,$6B,$E9,$1F
	.dc.b	$91,$7F,$09,$7D,$63,$F3,$57,$D5,$FC,$B5,$FE,$A7,$E1,$3F,$50,$48
	.dc.b	$FD,$43,$EA,$F8,$6F,$C7,$7C,$34,$9F,$D2,$47,$E4,$7E,$1A,$47,$E9
	.dc.b	$DF,$CA,$57,$ED,$5F,$DA,$12,$3F,$68,$F5,$7F,$5D,$5D,$5F,$D5,$57
	.dc.b	$47,$F5,$93,$76,$FB,$77,$FC,$06,$FF,$80,$12,$BF,$80,$3F,$C0,$EF
	.dc.b	$F8,$31,$23,$F7,$9A,$57,$F3,$91,$6F,$F0,$FA,$E6,$7F,$35,$3C,$DF
	.dc.b	$E1,$8D,$7F,$86,$35,$D9,$A2,$1A,$29,$FA,$5C,$FF,$8C,$11,$FE,$30
	.dc.b	$F5,$AF,$CC,$F1,$FE,$FC,$F3,$71,$7F,$09,$BF,$E0,$5C,$53,$0F,$CC
	.dc.b	$4E,$5A,$7F,$12,$E1,$FE,$24,$FF,$25,$DF,$F9,$41,$1F,$F6,$0A,$E6
	.dc.b	$FE,$1D,$D5,$27,$F9,$24,$7F,$11,$EC,$91,$FC,$9D,$FF,$50,$7E,$29
	.dc.b	$FE,$EF,$8F,$07,$27,$F3,$52,$3F,$CD,$5F,$D3,$D7,$58,$FE,$55,$7F
	.dc.b	$C7,$63,$AC,$7F,$2D,$BB,$B7,$F1,$0C,$3B,$4B,$2F,$D8,$3F,$D0,$79
	.dc.b	$7F,$A2,$11,$C1,$0F,$E8,$8E,$5A,$64,$FE,$34,$E5,$FE,$39,$CD,$FD
	.dc.b	$3C,$8F,$F4,$F7,$FD,$AF,$4F,$0B,$E1,$F8,$D2,$BE,$98,$9D,$38,$5E
	.dc.b	$27,$CA,$24,$67,$C9,$FE,$C1,$D7,$FD,$7C,$91,$FD,$7D,$F0,$E1,$DE
	.dc.b	$3F,$D3,$96,$C3,$EA,$CF,$49,$CE,$17,$4F,$99,$8F,$F6,$DE,$CF,$ED
	.dc.b	$84,$AF,$ED,$8D,$7F,$B6,$35,$FE,$80,$D7,$FA,$F3,$5F,$C9,$3F,$DD
	.dc.b	$FB,$64,$7F,$77,$6B,$FD,$DD,$AF,$F4,$B6,$BF,$DA,$5A,$FE,$75,$AF
	.dc.b	$E0,$FB,$FF,$C0,$84,$7F,$C0,$8D,$7F,$C0,$8D,$7F,$AC,$35,$FE,$EC
	.dc.b	$D7,$F5,$4F,$F8,$3F,$BE,$47,$F8,$3D,$AF,$F8,$3D,$AF,$F6,$36,$BF
	.dc.b	$E0,$16,$BF,$B5,$6B,$F9,$B3,$FC,$38,$43,$FC,$38,$53,$FC,$38,$53
	.dc.b	$FB,$81,$4F,$F0,$71,$4F,$E0,$47,$F8,$50,$87,$F8,$5B,$EE,$3F,$C4
	.dc.b	$6E,$0C,$78,$35,$A5,$8D,$BB,$6B,$D1,$AF,$76,$BC,$8E,$11,$D5,$46
	.dc.b	$9B,$19,$8C,$0F,$0F,$04,$9F,$23,$5F,$23,$5F,$23,$5F,$23,$5F,$84
	.dc.b	$F7,$F1,$24,$7B,$DA,$FB,$DA,$FB,$DA,$FB,$DA,$FC,$D6,$BE,$EF,$1B
	.dc.b	$D8,$47,$E6,$35,$F9,$8D,$7E,$63,$5F,$98,$D7,$ED,$3F,$1F,$C6,$91
	.dc.b	$F5,$DA,$FD,$76,$BF,$5D,$AF,$D7,$6B,$F8,$56,$BF,$5B,$C7,$FA,$84
	.dc.b	$7F,$08,$D7,$F0,$8D,$7F,$08,$D7,$F0,$8D,$7F,$24,$FD,$FF,$22,$47
	.dc.b	$E3,$DA,$FE,$3D,$AF,$E3,$DA,$FE,$3D,$AF,$E7,$5A,$FE,$3B,$C9,$FC
	.dc.b	$61,$1F,$CE,$35,$FC,$E3,$5F,$CE,$35,$FC,$E3,$5F,$D5,$3F,$97,$F2
	.dc.b	$A4,$7E,$9D,$AF,$E9,$DA,$FE,$9D,$AF,$E9,$DA,$FE,$D5,$AF,$E9,$BC
	.dc.b	$CF,$D2,$11,$FD,$A3,$5F,$DA,$35,$FD,$A3,$5F,$DA,$35,$FE,$04,$FE
	.dc.b	$BF,$CC,$91,$FB,$F6,$BF,$BF,$6B,$FB,$F6,$BF,$BF,$6B,$FC,$35,$AF
	.dc.b	$EF,$BC,$DF,$DE,$11,$FE,$18,$D7,$F8,$63,$5F,$E1,$8D,$7F,$86,$35
	.dc.b	$FE,$34,$FF,$07,$F3,$A4,$7F,$17,$6B,$FC,$5D,$AF,$F1,$76,$BF,$C5
	.dc.b	$DA,$FF,$25,$6B,$FC,$5B,$D1,$FE,$28,$47,$F9,$23,$5F,$E4,$8D,$7F
	.dc.b	$92,$35,$FE,$48,$D7,$F9,$93,$FC,$7F,$D2,$91,$FC,$BD,$AF,$F2,$F6
	.dc.b	$BF,$CB,$DA,$FF,$2F,$6B,$FC,$F5,$AF,$F2,$EF,$53,$F9,$61,$1F,$E7
	.dc.b	$8D,$7F,$9E,$35,$FE,$78,$D7,$F9,$E3,$5F,$E9,$4F,$F3,$7F,$52,$47
	.dc.b	$F4,$76,$BF,$D1,$DA,$FF,$47,$6B,$FD,$1D,$AF,$F5,$56,$BF,$D1,$BD
	.dc.b	$7F,$E8,$84,$7F,$AA,$35,$FE,$A8,$D7,$FA,$A3,$5F,$EA,$8D,$7F,$B1
	.dc.b	$3F,$D3,$FD,$89,$1F,$D7,$DA,$FF,$5F,$6B,$FD,$7D,$AF,$F5,$F6,$BF
	.dc.b	$DB,$5A,$FF,$5E,$F8,$8F,$EB,$84,$7F,$B6,$35,$FE,$D8,$D7,$FB,$63
	.dc.b	$5F,$ED,$8D,$7F,$BD,$3F,$D9,$FD,$D9,$1F,$DD,$DA,$FF,$77,$6B,$FD
	.dc.b	$DD,$AF,$F7,$76,$BF,$E0,$56,$AA,$7B,$FF,$DE,$5B,$1F,$E0,$76,$BF
	.dc.b	$E0,$B6,$BF,$E1,$06,$BF,$E1,$36,$A3,$B4,$75,$23,$D5,$AF,$57,$C5
	.dc.b	$F0,$12,$3C,$4D,$7C,$4D,$7C,$4D,$7C,$4D,$7C,$52,$8C,$47,$C1,$F8
	.dc.b	$2F,$80,$47,$E0,$35,$F8,$0D,$7E,$03,$5F,$80,$D7,$BB,$F2,$FE,$0A
	.dc.b	$47,$CB,$6B,$F2,$DA,$FC,$B6,$BF,$2D,$AF,$A9,$AF,$D9,$F8,$7F,$B0
	.dc.b	$47,$EC,$35,$FB,$0D,$7E,$C3,$5F,$B0,$D7,$E3,$BF,$83,$F8,$69,$1F
	.dc.b	$83,$6B,$F8,$36,$BF,$83,$6B,$F8,$36,$BF,$49,$AF,$E4,$7E,$23,$F2
	.dc.b	$04,$7F,$20,$D7,$F2,$0D,$7F,$20,$D7,$F2,$0D,$7E,$FB,$F9,$BF,$87
	.dc.b	$91,$F9,$B6,$BF,$9B,$6B,$F9,$B6,$BF,$9B,$6B,$F8,$A6,$BF,$A9,$F8
	.dc.b	$7F,$D4,$11,$FD,$43,$5F,$D4,$35,$FD,$43,$5F,$D4,$35,$FC,$BB,$FB
	.dc.b	$3F,$87,$91,$FB,$36,$BF,$B3,$6B,$FB,$36,$BF,$B3,$6B,$FA,$26,$BF
	.dc.b	$C0,$7E,$1B,$F8,$01,$1F,$E0,$0D,$7F,$80,$35,$FE,$00,$D7,$F8,$03
	.dc.b	$5F,$D7,$BF,$C2,$FE,$1A,$47,$F0,$B6,$BF,$C2,$DA,$FF,$0B,$6B,$FC
	.dc.b	$2D,$AF,$EE,$9A,$FF,$19,$F8,$6F,$E3,$04,$7F,$8C,$35,$FE,$30,$D7
	.dc.b	$F8,$C3,$5F,$E3,$0D,$7F,$83,$BF,$C8,$FE,$1A,$47,$F2,$36,$BF,$C8
	.dc.b	$DA,$FF,$23,$6B,$FC,$8D,$AF,$F1,$26,$BF,$CC,$7E,$1B,$F9,$81,$1F
	.dc.b	$E6,$0D,$7F,$98,$35,$FE,$60,$D7,$F9,$83,$5F,$E3,$EF,$F3,$BF,$86
	.dc.b	$91,$FC,$ED,$AF,$F3,$B6,$BF,$CE,$DA,$FF,$3B,$6B,$FC,$A9,$AF,$F4
	.dc.b	$9F,$37,$FA,$41,$1F,$E9,$0D,$7F,$A4,$35,$FE,$90,$D7,$FA,$43,$5F
	.dc.b	$E6,$EF,$F5,$3F,$3E,$47,$F5,$36,$BF,$D4,$DA,$FF,$53,$6B,$FD,$4D
	.dc.b	$AF,$F4,$26,$BF,$D8,$7D,$1F,$EC,$04,$7F,$B0,$35,$FE,$C0,$D7,$FB
	.dc.b	$03,$5F,$EC,$0D,$7F,$A7,$BF,$DA,$FD,$19,$1F,$DA,$DA,$FF,$6B,$6B
	.dc.b	$FD,$AD,$AF,$F6,$B6,$BF,$D6,$9A,$FF,$79,$F7,$3F,$BC,$11,$FE,$F0
	.dc.b	$D7,$FB,$C3,$5F,$EF,$0D,$7F,$BC,$35,$FE,$CE,$FF,$BE,$FE,$E5,$EC
	.dc.b	$D2,$4F,$F7,$D2,$08,$DB,$83,$FC,$04,$E1,$C3,$5E,$2D,$7A,$35,$E3
	.dc.b	$1E,$07,$8B,$5E,$3E,$DF,$9C,$8F,$84,$F2,$38,$7C,$2D,$78,$9B,$94
	.dc.b	$66,$3C,$6D,$7D,$7E,$DF,$C5,$23,$C5,$AF,$B9,$AF,$B9,$AF,$B9,$AF
	.dc.b	$43,$C5,$EF,$7D,$12,$38,$21,$F2,$9C,$B4,$C9,$F3,$1C,$B4,$F7,$7C
	.dc.b	$07,$DD,$23,$82,$1F,$59,$CB,$4C,$9F,$61,$CB,$4F,$95,$F0,$5F,$89
	.dc.b	$23,$82,$1F,$82,$72,$D3,$27,$E1,$1C,$B4,$FA,$DF,$07,$F9,$52,$38
	.dc.b	$21,$F8,$E7,$2D,$32,$7E,$41,$CB,$4F,$C1,$7C,$2F,$E8,$48,$E0,$87
	.dc.b	$E6,$9C,$B4,$C9,$F9,$C7,$2D,$3F,$1D,$F0,$DF,$AD,$23,$82,$1F,$A6
	.dc.b	$72,$D3,$27,$EA,$1C,$B4,$FC,$D7,$C3,$7E,$E4,$8E,$08,$7E,$C9,$CB
	.dc.b	$4C,$9F,$B4,$72,$D3,$F4,$DF,$0B,$FC,$14,$8E,$08,$7E,$F9,$CB,$4C
	.dc.b	$9F,$C0,$1C,$B4,$FD,$97,$C3,$7F,$11,$23,$82,$1F,$C2,$9C,$B4,$C9
	.dc.b	$FC,$31,$CB,$4F,$DF,$7C,$37,$F1,$D2,$38,$21,$FC,$59,$CB,$4C,$9F
	.dc.b	$C6,$1C,$B4,$FE,$15,$F0,$DF,$CA,$48,$E0,$87,$F2,$27,$2D,$32,$7F
	.dc.b	$24,$72,$D3,$F8,$B7,$C2,$FF,$35,$23,$82,$1F,$CB,$9C,$B4,$C9,$FC
	.dc.b	$C1,$CB,$4F,$E4,$5F,$0B,$FD,$04,$8E,$08,$7F,$3A,$72,$D3,$27,$F3
	.dc.b	$C7,$2D,$3F,$97,$7C,$2F,$F4,$D2,$38,$21,$FD,$19,$CB,$4C,$9F,$D2
	.dc.b	$1C,$B4,$FE,$75,$F0,$7F,$D6,$48,$E0,$87,$F5,$27,$2D,$32,$7F,$54
	.dc.b	$72,$D3,$FA,$37,$C1,$FF,$65,$23,$82,$1F,$D7,$9C,$B4,$C9,$FD,$81
	.dc.b	$CB,$4F,$EA,$5F,$07,$FD,$C4,$8E,$0E,$BF,$87,$7F,$BE,$5F,$F7,$77
	.dc.b	$DD,$BE,$0B,$FB,$A4,$E6,$53,$97,$FD,$E9,$19,$E9,$61,$C3,$CF,$FE
	.dc.b	$E2,$FF,$7D,$F8,$67,$6F,$F7,$4F,$86,$C3,$71,$FF,$7F,$4E,$9F,$EF
	.dc.b	$F5,$A7,$FC,$03,$5F,$E0,$B2,$3F,$DB,$DF,$8B,$8F,$9F,$2F,$34,$A3
	.dc.b	$19,$F9,$63,$A9,$EA,$CB,$C4,$FF,$83,$2F,$AB,$4F,$EF,$87,$F8,$6C
	.dc.b	$86,$0F,$F0,$35,$F8,$2F,$BC,$75,$3E,$6D,$F8,$5F,$F0,$F6,$17,$3F
	.dc.b	$F0,$74,$7F,$87,$8E,$A7,$F8,$78,$CD,$BE,$1A,$92,$3C,$2E,$AF,$B4
	.dc.b	$F8,$F2,$FF,$17,$36,$BE,$09,$F1,$63,$33,$F3,$E3,$EF,$B6,$7E,$5E
	.dc.b	$7F,$56,$7F,$C6,$C6,$67,$C0,$FC,$26,$BF,$E3,$99,$FA,$7A,$7F,$8F
	.dc.b	$9B,$1E,$00,$75,$3C,$26,$23,$84,$F4,$8A,$77,$3E,$8F,$C2,$FC,$62
	.dc.b	$38,$21,$D6,$91,$FE,$51,$AF,$23,$67,$E3,$B5,$FD,$2B,$D1,$F6,$35
	.dc.b	$F6,$7C,$2F,$D6,$23,$F8,$46,$BE,$17,$A3,$FA,$C6,$D1,$A7,$08,$EA
	.dc.b	$73,$3F,$1F,$F0,$BF,$84,$23,$82,$1E,$F7,$A9,$1F,$98,$7E,$11,$E2
	.dc.b	$FD,$76,$BF,$5F,$E1,$7F,$22,$47,$F3,$ED,$7E,$63,$EA,$23,$D8,$FE
	.dc.b	$1A,$ED,$E8,$72,$3F,$53,$F0,$BF,$9E,$23,$82,$1F,$65,$F8,$E4,$7B
	.dc.b	$1F,$C5,$9E,$2F,$E4,$5A,$FE,$47,$E1,$7F,$56,$47,$F7,$2D,$7F,$0C
	.dc.b	$FD,$42,$3D,$B0,$46,$E3,$A1,$C8,$FE,$07,$F0,$BF,$B8,$23,$83,$47
	.dc.b	$EC,$4F,$D0,$FD,$E3,$CB,$7B,$9F,$C2,$65,$39,$8C,$61,$E2,$9B,$87
	.dc.b	$ED,$FE,$1A,$47,$F2,$CD,$FF,$80,$CE,$23,$F6,$93,$FC,$30,$8F,$93
	.dc.b	$9B,$FA,$06,$3D,$E3,$CC,$78,$B2,$F2,$B8,$C7,$C4,$DB,$E6,$F6,$3F
	.dc.b	$92,$BF,$C6,$BE,$1A,$47,$F2,$86,$BF,$EB,$FB,$F0,$61,$E2,$C3,$BB
	.dc.b	$73,$F7,$D7,$C2,$BE,$2F,$F0,$D6,$BF,$C8,$3E,$1B,$F9,$49,$1F,$E6
	.dc.b	$EF,$F3,$0C,$3B,$F2,$FF,$B4,$DB,$3F,$53,$1F,$0C,$FC,$A8,$FA,$35
	.dc.b	$CD,$FE,$3C,$D7,$F8,$F7,$C3,$7F,$38,$23,$FD,$19,$F0,$57,$48,$ED
	.dc.b	$8F,$DF,$C3,$C3,$1F,$DC,$35,$FE,$77,$7C,$A3,$F9,$63,$5F,$E5,$9F
	.dc.b	$0D,$FD,$18,$8F,$F5,$47,$BD,$F5,$C3,$C1,$3F,$6E,$3E,$5B,$F6,$E7
	.dc.b	$8B,$FC,$E5,$AF,$F3,$9F,$86,$FE,$A6,$47,$FB,$0B,$D9,$FC,$3E,$3E
	.dc.b	$A3,$D7,$BF,$E1,$B6,$FE,$ED,$5C,$6B,$FA,$33,$5F,$E8,$DF,$0D,$FD
	.dc.b	$80,$8F,$F6,$E7,$B4,$FE,$0E,$3B,$E5,$EC,$BE,$DB,$7F,$60,$C3,$FB
	.dc.b	$1B,$2B,$F5,$31,$FD,$9B,$E1,$BF,$B6,$92,$3F,$BD,$BD,$A7,$F0,$97
	.dc.b	$F2,$32,$F5,$ED,$CE,$FF,$45,$57,$2F,$6C,$7E,$DA,$3F,$BC,$91,$FF
	.dc.b	$03,$BD,$5F,$C0,$9E,$AD,$7E,$9C,$FF,$86,$E7,$FA,$29,$27,$F8,$93
	.dc.b	$FD,$23,$E1,$BF,$C0,$A4,$8F,$F0,$9B,$D5,$FB,$C7,$53,$F7,$17,$FE
	.dc.b	$06,$8E,$51,$FC,$75,$AF,$F6,$7F,$86,$FF,$08,$91,$FF,$0F,$BD,$71
	.dc.b	$FB,$D9,$7D,$1B,$FF,$08,$CF,$F5,$C2,$57,$F2,$97,$FA,$FE,$5F,$E1
	.dc.b	$C2,$47,$F8,$A9,$E9,$5F,$3C,$E9,$1F,$C8,$23,$8B,$FC,$CD,$AF,$F8
	.dc.b	$26,$DF,$0D,$FB,$D2,$47,$F8,$38,$87,$F2,$C7,$BC,$FE,$94,$FA,$39
	.dc.b	$7F,$03,$8F,$EE,$F5,$D2,$BF,$A7,$B5,$F0,$46,$BE,$1B,$F8,$61,$23
	.dc.b	$B9,$FC,$F9,$EE,$FD,$3D,$3F,$CB,$19,$FF,$91,$6D,$5C,$EB,$BB,$5F
	.dc.b	$77,$C3,$7F,$19,$23,$D8,$FE,$8E,$F6,$BF,$D3,$1D,$A3,$F8,$74,$7B
	.dc.b	$08,$FB,$1F,$7F,$C3,$7F,$25,$24,$75,$B4,$75,$CB,$F4,$B9,$FC,$8A
	.dc.b	$FE,$B7,$7E,$BD,$3E,$D4,$F8,$27,$F8,$F7,$C3,$7F,$31,$23,$D5,$E8
	.dc.b	$FE,$93,$2F,$43,$F1,$F0,$EE,$D3,$B9,$FC,$A7,$E1,$BF,$9E,$11,$F4
	.dc.b	$3D,$1F,$D1,$CF,$43,$E2,$CF,$46,$C7,$43,$F8,$DF,$C3,$7F,$47,$23
	.dc.b	$F0,$DE,$9A,$7F,$04,$8F,$C5,$1E,$53,$BD,$7F,$4D,$CF,$EF,$7C,$37
	.dc.b	$F5,$12,$4F,$CD,$79,$E9,$FA,$08,$E5,$7F,$AE,$D3,$E5,$47,$F2,$EF
	.dc.b	$86,$FE,$B4,$47,$EB,$3C,$A3,$EA,$47,$88,$F9,$98,$FB,$CF,$23,$FD
	.dc.b	$47,$E1,$BF,$3E,$47,$F3,$EF,$E8,$B0,$FB,$71,$ED,$E7,$FE,$5E,$F4
	.dc.b	$21,$DC,$FC,$99,$FD,$77,$E1,$BF,$56,$47,$F5,$6F,$5C,$3E,$4E,$DF
	.dc.b	$4A,$7F,$B1,$18,$C7,$D0,$F6,$CD,$AF,$EA,$7E,$1B,$F6,$E4,$9F,$DB
	.dc.b	$BD,$5F,$97,$9F,$CD,$C3,$B5,$FC,$D9,$FE,$B3,$7F,$B3,$F8,$6F,$E0
	.dc.b	$24,$8F,$E0,$2F,$3A,$E3,$7C,$F1,$F2,$D7,$C1,$C3,$FB,$16,$1F,$BD
	.dc.b	$F8,$6F,$E1,$64,$9F,$E1,$6F,$3A,$E3,$5F,$8B,$9F,$3E,$7F,$05,$FC
	.dc.b	$A6,$3F,$E3,$6F,$86,$FE,$2A,$48,$FE,$2A,$F3,$7E,$36,$5E,$56,$D7
	.dc.b	$37,$EA,$9F,$DE,$3E,$1B,$F8,$F9,$1F,$E3,$EF,$37,$D9,$3C,$DB,$3C
	.dc.b	$F0,$FA,$71,$FA,$23,$F9,$7B,$FE,$4B,$6C,$FF,$92,$9B,$3F,$C3,$63
	.dc.b	$11,$FB,$E3,$27,$F9,$15,$CB,$FD,$A0,$C9,$0E,$07,$F8,$EF,$E1,$9E
	.dc.b	$8F,$F1,$96,$C7,$F9,$3D,$AF,$99,$AF,$18,$E0,$47,$83,$5E,$0D,$78
	.dc.b	$1F,$E5,$BF,$86,$78,$FF,$D2,$1B,$1F,$D5,$DB,$1F,$19,$B1,$CB,$4F
	.dc.b	$F3,$9D,$FA,$16,$BF,$D6,$FE,$F5,$E6,$74,$9D,$E5,$FE,$4B,$BE,$17
	.dc.b	$CC,$FE,$C5,$E9,$BD,$CE,$11,$9C,$BF,$D2,$7F,$0F,$E2,$6E,$BF,$D8
	.dc.b	$23,$11,$FE,$94,$6F,$03,$1E,$F3,$FE,$9C,$6A,$D7,$CF,$7E,$5C,$E4
	.dc.b	$FF,$92,$2B,$79,$F9,$6B,$3A,$7F,$A5,$1E,$E4,$3F,$B9,$9E,$29,$EF
	.dc.b	$8F,$8F,$0F,$F5,$56,$3E,$0C,$7B,$47,$4C,$FE,$04,$32,$F0,$3A,$3F
	.dc.b	$D4,$2B,$E1,$FE,$0B,$72,$FF,$68,$B6,$7E,$E8,$3A,$4F,$96,$FC,$11
	.dc.b	$9C,$3C,$99,$FD,$88,$F0,$34,$FB,$EF,$3B,$F2,$C6,$71,$FB,$7F,$0F
	.dc.b	$E4,$26,$5E,$47,$C1,$7D,$27,$F1,$1B,$FE,$02,$3E,$B6,$5E,$47,$FD
	.dc.b	$9F,$AF,$D2,$DF,$51,$89,$FF,$6A,$66,$F8,$7F,$86,$4C,$FE,$1B,$E0
	.dc.b	$D3,$CD,$7E,$FC,$B9,$EF,$F5,$63,$C8,$78,$8F,$9F,$96,$B3,$FC,$1F
	.dc.b	$C3,$FD,$02,$65,$F4,$1E,$DA,$FD,$AD,$FE,$5B,$F7,$31,$ED,$9F,$8B
	.dc.b	$2F,$BD,$97,$5D,$BD,$7F,$0F,$F6,$27,$F4,$AD,$7E,$83,$5F,$C3,$36
	.dc.b	$FA,$E1,$E1,$D7,$A5,$7B,$0F,$90,$7B,$34,$F9,$DB,$BE,$1F,$A3,$63
	.dc.b	$FC,$74,$D9,$FF,$24,$B5,$FE,$1E,$75,$D7,$DE,$FC,$9D,$7D,$D9,$79
	.dc.b	$23,$F1,$1A,$F9,$27,$F1,$71,$F9,$FF,$87,$F0,$93,$1F,$0B,$ED,$E0
	.dc.b	$FD,$59,$FB,$AC,$7B,$D7,$B2,$FB,$CF,$E1,$EF,$DB,$F0,$FE,$C2,$3E
	.dc.b	$C7,$A7,$07,$E3,$F0,$E7,$9F,$9F,$1E,$9B,$FE,$D3,$5F,$CF,$B7,$E1
	.dc.b	$FE,$39,$32,$F8,$EF,$4A,$F8,$9C,$1E,$FA,$FC,$16,$9E,$59,$FA,$98
	.dc.b	$3E,$1F,$E8,$93,$1F,$A2,$F3,$DB,$F1,$1A,$79,$B0,$E5,$1F,$8B,$BF
	.dc.b	$2E,$1F,$47,$1F,$E5,$BC,$1F,$70,$99,$7D,$C7,$A4,$7E,$1E,$3F,$07
	.dc.b	$1F,$C3,$08,$FF,$0C,$7F,$87,$FC,$3F,$E1,$89,$1F,$86,$79,$D7,$84
	.dc.b	$FD,$74,$7E,$3A,$38,$BF,$C5,$9B,$1F,$9F,$B7,$C3,$FE,$48,$95,$F9
	.dc.b	$27,$AC,$FD,$43,$A9,$F2,$9F,$E4,$64,$8F,$E4,$6F,$F2,$6F,$87,$FC
	.dc.b	$E9,$23,$F3,$AF,$4E,$0F,$C0,$6B,$F3,$78,$7F,$86,$CF,$1A,$FE,$5A
	.dc.b	$DC,$3D,$1F,$0F,$FA,$82,$4F,$EA,$1E,$9B,$FE,$5F,$87,$F0,$55,$FC
	.dc.b	$A6,$7F,$9C,$92,$BF,$9C,$BF,$CF,$3E,$1F,$F6,$64,$8F,$D9,$BD,$31
	.dc.b	$FE,$97,$7D,$2F,$FA,$44,$F1,$AF,$E8,$6D,$8F,$9D,$F0,$FF,$BE,$24
	.dc.b	$FE,$F9,$E9,$0E,$2F,$DA,$E9,$F0,$E3,$FA,$8C,$FF,$50,$25,$7F,$50
	.dc.b	$7F,$A9,$FC,$3F,$F0,$A2,$47,$F0,$A7,$AD,$F4,$C7,$D1,$C0,$BF,$E7
	.dc.b	$AD,$BF,$E7,$BF,$0F,$89,$CC,$FF,$16,$24,$FF,$16,$7A,$F2,$7F,$4C
	.dc.b	$8F,$E3,$B5,$C5,$E6,$7F,$48,$AE,$F1,$FC,$80,$91,$FC,$81,$E7,$97
	.dc.b	$E3,$B1,$FD,$74,$6A,$3F,$91,$B5,$FF,$7D,$E1,$FE,$3E,$AF,$E5,$3A
	.dc.b	$FF,$2C,$23,$FE,$39,$BF,$DD,$E5,$F0,$A7,$37,$FC,$B9,$AF,$43,$F9
	.dc.b	$A9,$2B,$F9,$AB,$E6,$3F,$C3,$93,$F3,$1D,$1F,$CE,$0A,$7F,$80,$CF
	.dc.b	$F0,$07,$C3,$7F,$4B,$D4,$D2,$57,$9F,$4F,$E0,$D5,$FC,$FE,$72,$F3
	.dc.b	$39,$CF,$FC,$01,$AE,$6D,$F8,$5E,$C4,$70,$75,$9F,$E1,$07,$A9,$FE
	.dc.b	$69,$1B,$8E,$C7,$F8,$E5,$C3,$E1,$97,$C2,$FA,$C9,$5E,$37,$DD,$8F
	.dc.b	$E9,$E7,$F3,$99,$FE,$7A,$77,$7D,$A7,$9D,$F1,$3D,$9F,$0B,$F1,$C8
	.dc.b	$E0,$F8,$18,$75,$23,$FE,$31,$F8,$6F,$F1,$9D,$74,$B4,$62,$BF,$CE
	.dc.b	$6E,$B0,$B7,$F9,$BA,$F2,$57,$AA,$3B,$92,$3F,$C1,$17,$FE,$4D,$8F
	.dc.b	$89,$3F,$E6,$68,$C4,$F3,$33,$3C,$5A,$7C,$1F,$86,$FC,$21,$23,$EE
	.dc.b	$36,$3B,$BD,$49,$1F,$E4,$F3,$EA,$3B,$8B,$1C,$8F,$C3,$7C,$37,$E4
	.dc.b	$48,$E0,$91,$EC,$8F,$51,$23,$B4,$7F,$A4,$63,$8B,$F6,$5A,$FE,$17
	.dc.b	$DA,$FC,$F1,$27,$F2,$CD,$8F,$92,$FC,$72,$47,$68,$FF,$33,$46,$DE
	.dc.b	$91,$C8,$FD,$07,$C3,$7E,$AC,$8E,$09,$1F,$5A,$3E,$A1,$23,$B4,$7F
	.dc.b	$A5,$63,$8B,$F9,$36,$BF,$9F,$F8,$6F,$DC,$12,$7F,$60,$D8,$FC,$1B
	.dc.b	$F8,$12,$47,$6C,$51,$B9,$E9,$3C,$8F,$DD,$7C,$37,$F0,$52,$38,$24
	.dc.b	$7E,$42,$3F,$1C,$48,$EC,$F0,$8F,$D6,$35,$FD,$C7,$C3,$7F,$11,$24
	.dc.b	$FF,$09,$78,$70,$7F,$52,$AF,$F4,$0E,$0F,$84,$B1,$2F,$F5,$C7,$FB
	.dc.b	$4D,$FD,$4E,$1F,$E9,$DF,$86,$AF,$FB,$06,$2D,$A7,$F9,$11,$2B,$F8
	.dc.b	$D3,$E2,$E5,$E9,$1F,$EB,$AC,$3C,$18,$7F,$67,$9F,$F7,$15,$71,$8F
	.dc.b	$E1,$4D,$7F,$8B,$7C,$37,$F2,$F2,$5F,$F2,$77,$F9,$95,$F4,$CB,$FB
	.dc.b	$8B,$70,$FF,$77,$4F,$0B,$FE,$2C,$D7,$F9,$17,$C3,$7F,$3B,$26,$1F
	.dc.b	$CC,$DE,$AF,$FB,$4B,$3F,$97,$87,$5A,$FF,$61,$3F,$12,$FE,$2B,$6B
	.dc.b	$E6,$CE,$9F,$E7,$D1,$D2,$3F,$94,$35,$FE,$68,$D7,$FA,$86,$7F,$66
	.dc.b	$78,$92,$38,$B5,$FE,$79,$F0,$DF,$D5,$49,$1F,$D3,$5B,$1F,$CD,$DA
	.dc.b	$F0,$8D,$11,$D3,$5F,$E9,$19,$7F,$61,$24,$7F,$59,$6C,$7F,$43,$6B
	.dc.b	$C2,$34,$47,$4D,$7F,$AA,$7C,$37,$F6,$D2,$47,$F6,$56,$C7,$F4,$F6
	.dc.b	$BC,$23,$44,$74,$D7,$FB,$07,$C3,$7F,$79,$24,$7F,$71,$6C,$7F,$5B
	.dc.b	$6B,$C2,$34,$47,$4D,$7F,$B6,$7A,$BF,$E0,$52,$47,$F7,$D6,$C7,$F6
	.dc.b	$76,$BC,$23,$44,$74,$D7,$FC,$11,$F7,$CF,$87,$FF,$6E,$36,$7F,$BF
	.dc.b	$CA,$30,$FF,$80,$9A,$F2,$9E,$24,$9E,$33,$A9,$FF,$10,$35,$1D,$67
	.dc.b	$83,$D5,$F1,$FC,$3C,$AE,$6D,$9F,$13,$5F,$5B,$5E,$ED,$7B,$D2,$31
	.dc.b	$1E,$DF,$73,$E0,$92,$BD,$2D,$8F,$80,$D7,$E4,$B5,$F6,$B5,$EE,$FC
	.dc.b	$CF,$87,$93,$F1,$1B,$3F,$2D,$AF,$D4,$6B,$F2,$1A,$FA,$9A,$FD,$5F
	.dc.b	$87,$FB,$24,$9F,$A2,$D8,$FB,$0D,$7F,$02,$D7,$EA,$B5,$F8,$EF,$E1
	.dc.b	$3E,$1E,$4F,$DC,$6C,$FE,$0D,$AF,$E3,$1A,$FE,$01,$AF,$D2,$6B,$F8
	.dc.b	$DF,$87,$FC,$89,$27,$F1,$2D,$8F,$C8,$35,$FC,$CB,$5F,$C6,$B5,$FB
	.dc.b	$EF,$E7,$3E,$1E,$4F,$E5,$1B,$3F,$9B,$6B,$FA,$46,$BF,$98,$6B,$F8
	.dc.b	$A6,$BF,$A5,$F8,$7F,$D4,$92,$7F,$42,$D8,$FD,$43,$5F,$D8,$B5,$FD
	.dc.b	$2B,$5F,$CB,$BF,$B4,$F8,$79,$3F,$AC,$6C,$FE,$CD,$AF,$EF,$1A,$FE
	.dc.b	$C1,$AF,$E8,$9A,$FE,$F7,$E1,$BF,$80,$92,$7F,$72,$D8,$FE,$00,$D7
	.dc.b	$F8,$4B,$5F,$DE,$B5,$FD,$7B,$FC,$33,$E1,$A4,$FF,$04,$6C,$FF,$0B
	.dc.b	$6B,$FC,$51,$AF,$F0,$86,$BF,$BA,$6B,$FC,$57,$E1,$BF,$8C,$92,$7F
	.dc.b	$88,$B6,$3F,$8C,$35,$FE,$42,$D7,$F8,$AB,$5F,$E0,$EF,$F2,$4F,$86
	.dc.b	$93,$FC,$71,$B3,$FC,$8D,$AF,$F2,$C6,$BF,$C8,$1A,$FF,$12,$6B,$FC
	.dc.b	$B7,$E1,$BF,$98,$92,$7F,$94,$B6,$3F,$98,$35,$FE,$72,$D7,$F9,$6B
	.dc.b	$5F,$E3,$EF,$F3,$CF,$86,$93,$FC,$D1,$B3,$FC,$ED,$AF,$F4,$46,$BF
	.dc.b	$CE,$1A,$FF,$2A,$6B,$FD,$17,$E1,$BF,$A4,$92,$7F,$A0,$B6,$3F,$A4
	.dc.b	$35,$FE,$A2,$D7,$FA,$2B,$5F,$E6,$EF,$F5,$4F,$56,$4F,$F4,$C6,$CF
	.dc.b	$F5,$36,$BF,$D7,$1A,$FF,$50,$6B,$FD,$09,$AF,$F5,$DF,$86,$FE,$C2
	.dc.b	$49,$FE,$B2,$D8,$FE,$C0,$D7,$FB,$4B,$5F,$EB,$AD,$7F,$A7,$BF,$DB
	.dc.b	$3E,$1A,$4F,$F6,$46,$CF,$F6,$B6,$BF,$DD,$1A,$FF,$68,$6B,$FD,$69
	.dc.b	$AF,$F7,$58,$FE,$F2,$43,$FB,$89,$5F,$EF,$0D,$7F,$C0,$2D,$7F,$BA
	.dc.b	$B5,$FE,$CE,$8F,$F0,$1F,$DF,$BE,$1F,$13,$FD,$E8,$91,$99,$DC,$EA
	.dc.b	$74,$27,$4E,$0C,$46,$DE,$6D,$7F,$C0,$F1,$C1,$E6,$D7,$73,$E0,$23
	.dc.b	$D6,$34,$F2,$F8,$7E,$8F,$8D,$B0,$69,$27,$E0,$B6,$7A,$35,$F1,$B5
	.dc.b	$F8,$6F,$4F,$87,$93,$F2,$9B,$3E,$56,$BE,$C6,$BF,$31,$AF,$C2,$6B
	.dc.b	$F2,$FE,$1F,$D2,$49,$FA,$ED,$8F,$84,$D7,$E4,$B5,$FB,$2D,$7E,$7B
	.dc.b	$F1,$3E,$1E,$4F,$E0,$9B,$3F,$35,$AF,$D4,$6B,$F8,$46,$BF,$69,$AF
	.dc.b	$E0,$FE,$1F,$E8,$92,$7F,$1E,$D8,$FB,$4D,$7F,$02,$D7,$F2,$2D,$7F
	.dc.b	$0E,$FD,$CF,$87,$93,$F9,$A6,$CF,$E1,$5A,$FE,$31,$AF,$E7,$1A,$FE
	.dc.b	$49,$AF,$E6,$FE,$1F,$F1,$24,$9F,$D3,$B6,$3F,$24,$D7,$F3,$2D,$7F
	.dc.b	$52,$D7,$F3,$EF,$E5,$3E,$1E,$4F,$EC,$9B,$3F,$9D,$6B,$FA,$46,$BF
	.dc.b	$B4,$6B,$FA,$A6,$BF,$B3,$F8,$7F,$D0,$92,$7F,$7E,$D8,$FD,$53,$5F
	.dc.b	$D8,$B5,$FE,$02,$D7,$F6,$EF,$EB,$3E,$1E,$4F,$F0,$A6,$CF,$ED,$5A
	.dc.b	$FE,$F1,$AF,$F0,$C6,$BF,$C0,$9A,$FF,$0B,$F8,$7F,$DC,$92,$7F,$8B
	.dc.b	$B6,$3F,$81,$35,$FE,$12,$D7,$F8,$CB,$5F,$E1,$EF,$F0,$4F,$87,$93
	.dc.b	$FC,$89,$B3,$FC,$35,$AF,$F1,$46,$BF,$C9,$1A,$FF,$1A,$6B,$FC,$8F
	.dc.b	$E1,$BF,$88,$92,$7F,$97,$B6,$3F,$8D,$35,$FE,$42,$D7,$F9,$8B,$5F
	.dc.b	$E4,$EF,$F1,$CF,$86,$93,$FC,$E9,$B3,$FC,$95,$AF,$F2,$C6,$BF,$CF
	.dc.b	$1A,$FF,$32,$6B,$FC,$EF,$E1,$BF,$94,$92,$7F,$A3,$B6,$3F,$99,$35
	.dc.b	$FE,$72,$D7,$FA,$4B,$5F,$E7,$EF,$F3,$4F,$85,$93,$FD,$49,$B3,$FC
	.dc.b	$F5,$AF,$F4,$46,$BF,$D5,$1A,$FF,$4A,$6B,$FD,$4F,$E1,$BF,$A0,$92
	.dc.b	$7F,$AF,$B6,$3F,$A5,$35,$FE,$A2,$D7,$FB,$0B,$5F,$EA,$EF,$F4,$CF
	.dc.b	$85,$93,$FD,$A9,$B3,$FD,$55,$AF,$F5,$C6,$BF,$DB,$1A,$FF,$62,$6B
	.dc.b	$FD,$AF,$E1,$BF,$AC,$92,$7F,$BB,$B6,$3F,$B1,$35,$FE,$D2,$D7,$FB
	.dc.b	$CB,$5F,$ED,$EF,$F6,$4F,$86,$93,$FE,$02,$6C,$FF,$6D,$6B,$FD,$D1
	.dc.b	$AF,$F8,$11,$AF,$F7,$A6,$BF,$E0,$33,$FB,$89,$23,$FC,$1E,$D3,$FB
	.dc.b	$D1,$4F,$F0,$09,$4F,$F0,$89,$4F,$F0,$39,$FD,$FB,$AD,$F0,$D4,$93
	.dc.b	$FE,$11,$94,$EA,$30,$D8,$FF,$0B,$35,$FF,$0D,$B5,$FF,$11,$35,$1D
	.dc.b	$27,$BB,$C9,$AF,$2F,$86,$F3,$12,$BC,$2D,$8F,$0B,$5E,$11,$A2,$3A
	.dc.b	$6B,$E6,$F8,$4F,$84,$48,$F6,$B6,$3D,$AD,$78,$46,$88,$E9,$AF,$C2
	.dc.b	$F8,$6F,$9C,$48,$F9,$2D,$8F,$92,$D7,$84,$68,$8E,$9A,$FC,$EF,$85
	.dc.b	$FB,$44,$8F,$AA,$D8,$FA,$AD,$78,$46,$88,$E9,$AF,$DA,$F8,$5F,$C3
	.dc.b	$12,$3F,$02,$D8,$FC,$0B,$5E,$11,$A2,$3A,$6B,$F8,$6F,$85,$FC,$91
	.dc.b	$23,$F1,$AD,$8F,$C6,$B5,$E1,$1A,$23,$A6,$BF,$92,$F8,$5F,$CF,$12
	.dc.b	$3F,$32,$D8,$FC,$CB,$5E,$11,$A2,$3A,$6B,$F9,$EF,$85,$FD,$51,$23
	.dc.b	$F4,$AD,$8F,$D2,$B5,$E1,$1A,$23,$A6,$BF,$AA,$F8,$5F,$DB,$12,$3F
	.dc.b	$62,$D8,$FD,$8B,$5E,$11,$A2,$3A,$6B,$FB,$6F,$81,$FE,$04,$48,$FD
	.dc.b	$EB,$63,$F7,$AD,$78,$46,$88,$E9,$AF,$F0,$2F,$85,$FE,$1C,$48,$FE
	.dc.b	$12,$D8,$FE,$12,$D7,$84,$68,$8E,$9A,$FF,$0E,$F8,$5F,$E3,$44,$8F
	.dc.b	$E2,$AD,$8F,$E2,$AD,$78,$46,$88,$E9,$AF,$F1,$AF,$83,$FE,$4C,$48
	.dc.b	$FE,$42,$D8,$FE,$42,$D7,$84,$68,$8E,$9A,$FF,$26,$F8,$4F,$E6,$44
	.dc.b	$8F,$E5,$AD,$8F,$E5,$AD,$78,$46,$88,$E9,$AF,$F3,$2F,$84,$FE,$7C
	.dc.b	$48,$FE,$72,$D8,$FE,$72,$D7,$84,$68,$8E,$9A,$FF,$3E,$F8,$3F,$E9
	.dc.b	$44,$8F,$E8,$AD,$8F,$E8,$AD,$78,$46,$88,$E9,$AF,$F4,$AF,$83,$FE
	.dc.b	$AC,$48,$FE,$A2,$FF,$B8,$7F,$0F,$7E,$28,$F8,$2F,$F7,$8F,$C4,$7F
	.dc.b	$BC,$A4,$BE,$13,$38,$7F,$52,$9E,$35,$8C,$51,$C2,$B1,$F8,$A3,$F2
	.dc.b	$B7,$B3,$86,$E3,$FD,$73,$2F,$F8,$03,$5B,$F6,$B5,$7D,$72,$FE,$D7
	.dc.b	$F1,$1F,$DC,$49,$87,$F6,$C7,$D9,$AF,$AE,$3A,$1E,$7C,$7D,$4E,$2B
	.dc.b	$CF,$3B,$7D,$F1,$BC,$FC,$2E,$F2,$F5,$63,$E9,$CB,$37,$C0,$FE,$EB
	.dc.b	$E0,$3E,$23,$FC,$0C,$4C,$3F,$C0,$2D,$9F,$F0,$0B,$5F,$F0,$0B,$5F
	.dc.b	$F0,$0B,$5E,$AF,$F8,$47,$19,$3F,$E0,$C6,$CF,$F8,$18,$FA,$D9,$3E
	.dc.b	$1B,$EB,$6F,$F8,$0F,$87,$B3,$9C,$3E,$24,$E5,$F8,$DB,$3F,$60,$78
	.dc.b	$5F,$7E,$30,$D8,$78,$76,$3E,$C7,$E0,$BF,$37,$4F,$20,$70,$F9,$DA
	.dc.b	$FF,$87,$9B,$8F,$E5,$B4,$F3,$E9,$E8,$6E,$7E,$6C,$BC,$AD,$C7,$C9
	.dc.b	$97,$8E,$3F,$01,$97,$87,$04,$7F,$8D,$F2,$46,$27,$CE,$6E,$75,$19
	.dc.b	$8F,$5C,$FC,$97,$F0,$52,$F9,$F7,$A3,$F0,$B3,$FD,$34,$E6,$30,$FE
	.dc.b	$37,$F2,$67,$1E,$30,$FC,$67,$2A,$73,$5E,$AC,$BC,$15,$E2,$C7,$C8
	.dc.b	$FC,$B6,$BF,$2D,$AF,$D1,$6C,$72,$8E,$13,$F0,$E7,$24,$70,$48,$F8
	.dc.b	$B2,$F8,$BE,$64,$9F,$98,$F1,$9F,$D3,$1E,$18,$FD,$97,$94,$FD,$71
	.dc.b	$DB,$F0,$79,$3E,$76,$3F,$35,$B9,$7F,$0C,$FC,$9D,$E7,$EF,$2F,$C5
	.dc.b	$7C,$6F,$B0,$8F,$DD,$7F,$82,$9F,$A0,$D7,$F4,$2D,$8F,$CF,$FE,$A9
	.dc.b	$FC,$0F,$1F,$E4,$6D,$8F,$E2,$DE,$E7,$F0,$AD,$FF,$80,$52,$B5,$1F
	.dc.b	$99,$F8,$EF,$9A,$4B,$FC,$73,$E2,$7F,$5F,$D7,$F2,$5B,$8F,$8E,$BA
	.dc.b	$DF,$4C,$3B,$78,$5F,$B1,$CB,$C5,$1E,$68,$FD,$7F,$C7,$7D,$C2,$63
	.dc.b	$F9,$D7,$F9,$29,$FC,$00,$FE,$03,$9D,$8F,$E3,$DE,$D7,$FA,$39,$BC
	.dc.b	$9F,$CE,$A3,$19,$FE,$51,$47,$EF,$BE,$3B,$F1,$64,$AF,$D6,$3E,$07
	.dc.b	$DF,$87,$81,$A7,$82,$FB,$57,$76,$C7,$E7,$FE,$3B,$F2,$E4,$7F,$70
	.dc.b	$FE,$6B,$E3,$BF,$A9,$57,$E2,$FC,$8F,$E2,$31,$8E,$84,$62,$FF,$8E
	.dc.b	$FE,$C2,$F7,$F1,$97,$DC,$8C,$4F,$EB,$F0,$EC,$FF,$07,$6C,$FE,$89
	.dc.b	$AF,$F0,$46,$BF,$C7,$5A,$FF,$1D,$6B,$FC,$8F,$E3,$FF,$66,$49,$FE
	.dc.b	$28,$F6,$E1,$FD,$1D,$7F,$6C,$FC,$01,$E6,$FA,$27,$37,$E7,$3F,$92
	.dc.b	$7B,$39,$7F,$97,$7C,$87,$F0,$02,$3F,$C8,$5F,$ED,$3C,$1F,$D1,$0F
	.dc.b	$E8,$BC,$56,$7F,$B1,$7E,$38,$F5,$7F,$BE,$AF,$5A,$C6,$1F,$E8,$3E
	.dc.b	$2F,$F0,$E2,$57,$F2,$F7,$BB,$FA,$2C,$3B,$B4,$EF,$5C,$DA,$78,$AB
	.dc.b	$FA,$6F,$A1,$FC,$68,$91,$FC,$ED,$FF,$01,$F0,$7F,$5E,$E0,$FE,$BE
	.dc.b	$D9,$FD,$FF,$85,$FD,$6F,$B7,$F8,$74,$7F,$84,$DF,$E8,$6E,$8F,$E6
	.dc.b	$1F,$23,$FC,$A0,$8F,$F4,$97,$BB,$FB,$CB,$EE,$D3,$97,$27,$EF,$2F
	.dc.b	$C1,$E5,$78,$5F,$EC,$FF,$23,$FC,$D0,$95,$FD,$55,$FE,$6F,$F2,$3F
	.dc.b	$9B,$F5,$3F,$B6,$B6,$7F,$A2,$3F,$E3,$0F,$35,$F2,$3A,$9F,$F0,$D4
	.dc.b	$61,$D9,$FD,$86,$7A,$B4,$99,$7F,$3E,$8C,$B6,$32,$FF,$62,$8C,$B6
	.dc.b	$3F,$B6,$57,$F4,$E2,$38,$3A,$F4,$7F,$3E,$AF,$DB,$7A,$9E,$7B,$FE
	.dc.b	$8F,$E5,$A7,$15,$FE,$46,$C5,$EA,$E3,$1F,$F0,$53,$FD,$70,$8F,$F7
	.dc.b	$D7,$FC,$8B,$CF,$FE,$20,$3F,$C4,$3D,$56,$7F,$C1,$5E,$F7,$F8,$6F
	.dc.b	$DF,$FF,$0E,$36,$3F,$B8,$7C,$8F,$F6,$82,$3F,$E0,$97,$AB,$FC,$F2
	.dc.b	$FA,$B4,$EA,$7F,$85,$EB,$AB,$4F,$EF,$87,$F7,$42,$1F,$E1,$53,$FC
	.dc.b	$BB,$CB,$FE,$38,$E8,$FF,$1C,$B6,$3F,$CA,$DF,$6A,$C3,$FC,$E2,$D8
	.dc.b	$CE,$78,$9F,$F3,$67,$E1,$8F,$93,$FF,$02,$B6,$BF,$C4,$3C,$48,$C3
	.dc.b	$FE,$24,$6B,$FE,$4F,$F6,$BF,$D1,$79,$78,$E3,$FC,$C9,$18,$AF,$EB
	.dc.b	$F9,$F8,$5F,$3E,$3E,$6C,$B1,$7F,$E0,$6C,$5F,$25,$E4,$25,$F9,$23
	.dc.b	$2D,$F9,$2F,$F2,$A7,$D2,$FE,$EA,$C7,$A6,$33,$D7,$F0,$1F,$F2,$D7
	.dc.b	$D2,$3E,$1F,$18,$7F,$98,$23,$0F,$F9,$23,$15,$7C,$62,$47,$84,$98
	.dc.b	$7C,$58,$CB,$63,$2F,$F9,$8A,$79,$C6,$23,$FC,$F1,$18,$7E,$77,$CA
	.dc.b	$7D,$42,$4F,$C1,$7A,$F4,$7F,$94,$EB,$83,$FE,$AA,$F2,$9D,$DE,$CC
	.dc.b	$3E,$66,$36,$BF,$C3,$BE,$BD,$CB,$C3,$F2,$9F,$82,$23,$F3,$DF,$F5
	.dc.b	$6F,$6F,$FA,$23,$C1,$FF,$44,$B6,$7F,$D5,$9F,$8E,$3F,$02,$FF,$0A
	.dc.b	$C7,$D2,$8C,$7C,$27,$FB,$1D,$D7,$87,$6B,$7A,$FD,$8E,$07,$B3,$F9
	.dc.b	$42,$63,$F8,$17,$C6,$FF,$8A,$B3,$F1,$B4,$F1,$DF,$4C,$7C,$31,$8C
	.dc.b	$7C,$19,$7E,$8B,$E5,$BF,$42,$49,$FC,$73,$FE,$CC,$E9,$FF,$53,$9F
	.dc.b	$EA,$8F,$1A,$CF,$FA,$47,$E3,$3F,$D4,$5D,$1F,$EA,$36,$C7,$E2,$7E
	.dc.b	$5B,$F5,$A4,$7F,$34,$F5,$7F,$C8,$77,$D5,$A7,$53,$FD,$3D,$5D,$5A
	.dc.b	$7E,$57,$E5,$BF,$72,$47,$F4,$CF,$EF,$3D,$6F,$F1,$16,$7F,$C0,$1B
	.dc.b	$3F,$9D,$7F,$DE,$FF,$5F,$FE,$6C,$8C,$1F,$EE,$17,$1C,$7F,$B5,$9E
	.dc.b	$93,$FB,$36,$C7,$F0,$56,$BF,$9F,$6B,$F9,$E6,$BF,$7D,$AF,$F0,$FF
	.dc.b	$86,$FE,$26,$49,$FE,$00,$F6,$E8,$FD,$05,$7F,$90,$3F,$5D,$FF,$38
	.dc.b	$DF,$F1,$78,$F7,$5F,$AA,$3F,$D5,$7F,$67,$F8,$C8,$C5,$FE,$A7,$DF
	.dc.b	$FE,$42,$48,$FE,$1C,$FF,$AA,$BC,$67,$C1,$FF,$B2,$27,$FC,$F3,$E3
	.dc.b	$7F,$BA,$9F,$F7,$5B,$FE,$0B,$F8,$19,$1F,$E8,$CF,$83,$B9,$FF,$1D
	.dc.b	$F9,$9F,$E5,$E4,$9F,$E3,$8F,$7C,$3B,$4F,$F2,$A7,$F9,$D3,$67,$B9
	.dc.b	$FE,$98,$BE,$B9,$FF,$A6,$30,$B3,$FC,$33,$E6,$7F,$9E,$91,$FE,$54
	.dc.b	$FB,$6F,$F9,$59,$FC,$E0,$B5,$FE,$E3,$CF,$F9,$5F,$9B,$FE,$E5,$8C
	.dc.b	$4E,$AF,$F7,$DF,$81,$7F,$8A,$21,$3F,$EE,$BE,$D4,$FF,$A9,$9B,$EB
	.dc.b	$7F,$03,$8F,$B3,$E1,$3E,$6B,$FA,$99,$31,$FE,$80,$F9,$FC,$EF,$14
	.dc.b	$7F,$3E,$D3,$F9,$FC,$63,$34,$7A,$27,$FA,$1E,$7E,$8A,$F0,$65,$E1
	.dc.b	$F8,$7B,$5F,$C6,$B3,$F1,$FC,$D7,$F6,$42,$47,$F5,$17,$FB,$3F,$CD
	.dc.b	$7E,$46,$FF,$B5,$B6,$7F,$A4,$3E,$3C,$5F,$0A,$EB,$DD,$7F,$A4,$9B
	.dc.b	$AF,$F6,$DB,$EB,$3F,$D7,$DB,$19,$8C,$11,$C1,$0F,$E9,$D1,$C1,$FE
	.dc.b	$A1,$F1,$5F,$DF,$89,$3F,$DA,$DE,$B8,$7F,$4B,$AF,$D6,$7A,$5E,$48
	.dc.b	$C5,$7B,$0F,$96,$7C,$CF,$91,$7C,$BE,$E3,$FA,$DF,$CD,$7F,$83,$08
	.dc.b	$FF,$79,$7F,$21,$EA,$7D,$7E,$3F,$EE,$FE,$95,$D5,$36,$3F,$0D,$E9
	.dc.b	$4C,$3E,$B6,$96,$FF,$C1,$1F,$35,$FE,$1A,$24,$FF,$81,$DE,$EF,$A2
	.dc.b	$3F,$0A,$7F,$68,$7F,$14,$F8,$1D,$63,$E4,$7F,$09,$7D,$F5,$FC,$E7
	.dc.b	$27,$F8,$3F,$A7,$F3,$B3,$FE,$2C,$F8,$DF,$F1,$59,$2F,$FC,$30,$FE
	.dc.b	$BF,$D2,$F7,$1E,$FF,$52,$CF,$97,$17,$C1,$BE,$27,$FD,$2D,$F1,$DF
	.dc.b	$E9,$76,$D7,$F8,$57,$6F,$F1,$C9,$1F,$F1,$33,$DD,$F5,$BF,$C8,$4F
	.dc.b	$DF,$61,$E0,$7F,$6F,$1F,$EA,$19,$EB,$18,$3F,$77,$7E,$39,$F3,$2B
	.dc.b	$FD,$84,$BF,$0F,$7C,$DA,$FF,$C9,$71,$86,$C6,$67,$FC,$6E,$49,$D3
	.dc.b	$83,$FC,$76,$BE,$57,$FA,$06,$DE,$08,$C5,$FD,$DD,$9E,$6A,$B5,$7F
	.dc.b	$C1,$7E,$55,$F3,$7E,$BC,$BD,$C7,$A6,$75,$19,$E0,$F3,$4E,$72,$F5
	.dc.b	$D6,$23,$D0,$D7,$CE,$D7,$81,$EB,$29,$ED,$CF,$D2,$E5,$AE,$63,$94
	.dc.b	$62,$3B,$CF,$88,$F9,$BF,$2F,$A2,$38,$35,$1E,$E9,$EF,$7F,$47,$4F
	.dc.b	$8F,$E4,$FC,$6B,$FD,$35,$FA,$2B,$E0,$DF,$09,$FF,$36,$E6,$FA,$4F
	.dc.b	$EE,$33,$9B,$FA,$FF,$25,$E8,$25,$FE,$09,$FE,$A1,$EC,$7E,$FF,$DB
	.dc.b	$FE,$00,$D9,$FD,$17,$82,$F9,$0F,$E5,$4D,$8C,$FC,$86,$27,$F0,$5F
	.dc.b	$13,$F1,$89,$3F,$90,$7B,$BF,$A2,$C3,$BB,$4E,$F3,$CD,$EA,$60,$F8
	.dc.b	$79,$BE,$2F,$E9,$92,$3F,$3A,$FF,$6B,$F6,$3F,$8D,$1F,$C6,$FD,$BB
	.dc.b	$3F,$D3,$7D,$4F,$EC,$1C,$7F,$D8,$5B,$1D,$BD,$CF,$BE,$47,$F5,$2F
	.dc.b	$57,$F7,$77,$D5,$A7,$53,$FA,$D5,$75,$69,$EA,$F7,$BF,$16,$47,$F6
	.dc.b	$AF,$E3,$BC,$AF,$D3,$7E,$00,$F9,$8F,$D7,$46,$1B,$5F,$C0,$BE,$4E
	.dc.b	$F3,$7F,$07,$E9,$FE,$33,$18,$8F,$CF,$5F,$5A,$FE,$06,$D8,$FD,$03
	.dc.b	$5D,$C6,$88,$E9,$AF,$EA,$79,$3F,$4A,$48,$FE,$1E,$F5,$BF,$CC,$D7
	.dc.b	$F8,$17,$F1,$C7,$E0,$4F,$8D,$F4,$D6,$B0,$FB,$F7,$FD,$E7,$C5,$7C
	.dc.b	$A6,$A7,$F8,$B7,$3F,$ED,$08,$FF,$1E,$7F,$C5,$9F,$03,$FD,$6C,$FE
	.dc.b	$B9,$EA,$D9,$FF,$07,$79,$3F,$E2,$1F,$27,$FC,$44,$D8,$FC,$7F,$8D
	.dc.b	$FC,$00,$8F,$F2,$A7,$AB,$FC,$F2,$FA,$B4,$EA,$7F,$86,$EB,$AB,$4F
	.dc.b	$CD,$FC,$FF,$F0,$C2,$3F,$CD,$9F,$F2,$27,$BF,$FD,$CF,$E0,$7F,$BA
	.dc.b	$36,$7F,$6B,$3F,$B4,$3F,$6B,$85,$8F,$D3,$FC,$FF,$F1,$82,$3F,$D0
	.dc.b	$9E,$AF,$F5,$4B,$EA,$D3,$A9,$39,$7F,$19,$E1,$FF,$38,$8C,$71,$7F
	.dc.b	$14,$F9,$FF,$E4,$A4,$8F,$E9,$EF,$F2,$8F,$9A,$FF,$43,$78,$1F,$D1
	.dc.b	$DB,$3F,$BC,$7F,$A2,$7D,$8B,$2D,$4F,$F8,$1F,$CA,$4E,$63,$81,$FD
	.dc.b	$72,$BB,$34,$99,$FF,$3B,$8C,$B6,$32,$FF,$41,$8C,$B6,$3F,$B3,$F6
	.dc.b	$FF,$40,$23,$83,$AE,$7F,$CD,$2B,$FD,$45,$C3,$E8,$BF,$E7,$D1,$67
	.dc.b	$FB,$76,$2E,$17,$87,$9C,$7F,$B8,$7C,$E7,$F4,$E2,$3F,$DE,$1F,$F4
	.dc.b	$37,$C0,$7F,$8D,$4F,$F1,$B7,$C1,$D9,$FF,$01,$FC,$C7,$E9,$FE,$6F
	.dc.b	$F5,$0D,$8F,$E5,$3F,$27,$FD,$68,$8F,$F8,$11,$EA,$FF,$85,$AF,$AB
	.dc.b	$4E,$A7,$F8,$42,$BA,$B4,$FE,$6B,$F2,$1F,$D9,$88,$FF,$84,$1F,$F4
	.dc.b	$F7,$C1,$FF,$93,$7E,$0F,$FC,$9C,$D9,$FF,$40,$78,$1F,$E9,$33,$FD
	.dc.b	$29,$E3,$D8,$FE,$83,$F1,$DF,$DC,$88,$FF,$87,$1E,$AF,$F8,$BA,$FA
	.dc.b	$B4,$EA,$4E,$4F,$F1,$67,$BB,$FE,$A4,$BE,$D1,$FE,$12,$D5,$F0,$FF
	.dc.b	$E0,$96,$DF,$F8,$26,$51,$87,$FC,$18,$D7,$FC,$50,$FF,$AC,$F9,$7F
	.dc.b	$BF,$7B,$7F,$E1,$BA,$FF,$17,$CE,$55,$3D,$8F,$F1,$4F,$89,$FE,$3C
	.dc.b	$AD,$4F,$FA,$F3,$E1,$9E,$2E,$7E,$1F,$FD,$3D,$18,$C7,$FC,$13,$7F
	.dc.b	$EC,$3B,$F1,$D7,$AC,$93,$EB,$9D,$4F,$92,$72,$FB,$1A,$F2,$9E,$11
	.dc.b	$88,$F5,$FC,$27,$C7,$24,$F5,$79,$E9,$FE,$3A,$BF,$5E,$7F,$E4,$6D
	.dc.b	$BD,$57,$E8,$C6,$E3,$FC,$43,$C5,$B8,$F1,$F7,$BE,$A1,$2F,$D4,$FF
	.dc.b	$BA,$FE,$13,$FD,$34,$7F,$A6,$FE,$1E,$CF,$FB,$37,$AF,$FD,$C1,$D5
	.dc.b	$FE,$E1,$6C,$7C,$8F,$57,$F0,$04,$7E,$33,$D5,$FF,$3D,$DF,$56,$9D
	.dc.b	$4F,$F6,$D5,$75,$69,$F5,$3C,$7F,$C6,$11,$FA,$4F,$F9,$67,$E1,$7F
	.dc.b	$CE,$71,$FE,$88,$E0,$FF,$5A,$6D,$FE,$B5,$7F,$B2,$7A,$32,$7F,$D0
	.dc.b	$7F,$33,$6B,$F3,$5E,$3F,$E6,$49,$3F,$7D,$ED,$1F,$90,$8E,$C7,$FA
	.dc.b	$59,$EC,$FF,$A3,$AB,$A7,$07,$FA,$1C,$FF,$52,$DE,$27,$F3,$7D,$1F
	.dc.b	$A7,$23,$F8,$D7,$F5,$5C,$9C,$63,$F3,$F9,$7F,$A1,$BB,$BF,$3D,$39
	.dc.b	$CB,$F4,$5D,$5F,$E0,$0B,$FC,$DC,$FF,$A2,$7C,$2F,$F7,$D4,$60,$FE
	.dc.b	$8C,$E2,$BF,$75,$3E,$09,$FC,$F3,$63,$F6,$ED,$7F,$6E,$D7,$F6,$AD
	.dc.b	$7F,$6B,$D0,$8C,$47,$F0,$2C,$3F,$83,$12,$BF,$58,$F7,$D3,$F4,$15
	.dc.b	$ED,$CB,$FD,$4B,$C3,$F0,$2F,$D5,$87,$A1,$F5,$F5,$E6,$3F,$8A,$4F
	.dc.b	$F1,$32,$4F,$EE,$5F,$F5,$02,$FC,$43,$FD,$04,$DF,$83,$F9,$8E,$7E
	.dc.b	$12,$CF,$FA,$1F,$37,$9E,$F2,$FF,$CF,$FE,$7F,$FA,$01,$B8,$7F,$0C
	.dc.b	$EB,$FE,$48,$47,$F8,$53,$E2,$C3,$C2,$4C,$7B,$E5,$CE,$BF,$D1,$B8
	.dc.b	$5D,$3F,$C1,$AF,$F1,$FE,$5F,$E6,$04,$9F,$E2,$CF,$5F,$82,$F4,$FC
	.dc.b	$27,$A9,$B3,$FC,$AB,$6F,$F6,$E7,$8A,$F8,$7F,$F2,$74,$E4,$FE,$5F
	.dc.b	$19,$EA,$FC,$83,$BC,$6D,$70,$CF,$FA,$25,$7F,$43,$25,$FF,$29,$7C
	.dc.b	$8F,$58,$FF,$78,$6C,$CF,$CD,$39,$3E,$B3,$DA,$A4,$71,$C7,$C4,$DC
	.dc.b	$3C,$59,$7F,$29,$F9,$9F,$EA,$84,$9F,$E7,$4F,$F5,$7F,$95,$FC,$77
	.dc.b	$5B,$8B,$FA,$54,$61,$B5,$E7,$B7,$B2,$C3,$3B,$7F,$A2,$E3,$15,$FD
	.dc.b	$33,$0E,$CF,$F4,$96,$CF,$F6,$26,$BF,$D8,$9A,$FF,$60,$6B,$FA,$C6
	.dc.b	$BF,$DB,$3E,$5B,$FB,$81,$27,$FA,$B3,$DB,$0F,$E9,$55,$EE,$E0,$F3
	.dc.b	$C6,$2B,$CC,$7B,$70,$56,$A3,$FB,$8F,$CA,$FF,$7D,$23,$FD,$8D,$FC
	.dc.b	$D7,$C2,$7E,$98,$FD,$3F,$C3,$D9,$FA,$95,$F8,$6D,$3F,$0E,$D8,$FE
	.dc.b	$E3,$F2,$BF,$E0,$92,$3F,$DB,$DE,$AF,$E0,$F9,$3F,$DB,$8D,$BE,$DB
	.dc.b	$7F,$85,$B8,$7D,$33,$F8,$3C,$7B,$B4,$FE,$FF,$F2,$7F,$E1,$72,$3F
	.dc.b	$DF,$5F,$DA,$7C,$2F,$F0,$0F,$87,$FE,$02,$D9,$FF,$7F,$FA,$FF,$EF
	.dc.b	$E3,$FD,$FF,$ED,$D8,$FF,$05,$9F,$E2,$32,$1F,$E0,$93,$A9,$EB,$AE
	.dc.b	$AD,$3A,$93,$2F,$F4,$5C,$F5,$5C,$49,$FF,$11,$C3,$F1,$17,$C9,$E6
	.dc.b	$32,$7F,$8A,$23,$33,$FE,$35,$6C,$68,$E5,$3C,$5E,$30,$F9,$2F,$D3
	.dc.b	$F5,$FF,$51,$9F,$04,$62,$FF,$79,$18,$7F,$69,$D1,$BB,$E5,$18,$36
	.dc.b	$E2,$7C,$4D,$7C,$35,$E3,$7D,$F1,$96,$C6,$5F,$73,$5F,$11,$D1,$C4
	.dc.b	$71,$8D,$CF,$C3,$9C,$9A,$F9,$19,$1F,$1D,$B3,$CE,$3D,$33,$FC,$2B
	.dc.b	$83,$DC,$C8,$C9,$FC,$57,$9F,$15,$F6,$A3,$D7,$F2,$5E,$42,$3F,$55
	.dc.b	$FE,$8D,$F1,$5F,$D4,$8F,$EA,$7F,$11,$67,$F7,$17,$FF,$00,$74,$FF
	.dc.b	$02,$6F,$FC,$5A,$31,$3F,$3B,$E4,$7E,$09,$1F,$C1,$3D,$9F,$D7,$61
	.dc.b	$D9,$A7,$63,$F8,$35,$F6,$EC,$5F,$6C,$FD,$87,$E3,$FE,$4B,$E7,$92
	.dc.b	$FF,$20,$FF,$61,$F8,$7F,$ED,$BF,$15,$FD,$B9,$B3,$FC,$EF,$A7,$FA
	.dc.b	$69,$FD,$3B,$B6,$C7,$E2,$3E,$43,$ED,$91,$FC,$E3,$D5,$FE,$11,$7D
	.dc.b	$5A,$75,$26,$3F,$C2,$6B,$AB,$EA,$3F,$59,$F2,$3F,$88,$24,$FE,$A5
	.dc.b	$FC,$5F,$C3,$7F,$4A,$F0,$7F,$3C,$D9,$FC,$EB,$FE,$90,$F8,$AC,$C7
	.dc.b	$F4,$BF,$81,$FF,$7B,$61,$DF,$AB,$FD,$BF,$3F,$B5,$BE,$AD,$24,$FE
	.dc.b	$9A,$32,$D8,$CB,$FB,$88,$CB,$63,$F8,$07,$C8,$7E,$88,$8E,$0E,$BC
	.dc.b	$DF,$C0,$6F,$C7,$5F,$73,$D3,$79,$2E,$6F,$5C,$FA,$72,$B8,$7F,$98
	.dc.b	$7D,$1F,$F6,$0E,$5F,$08,$FD,$FF,$C6,$FE,$C8,$8F,$F1,$37,$FC,$49
	.dc.b	$F0,$3F,$E3,$03,$FC,$63,$F1,$B6,$7F,$C0,$1F,$83,$7F,$18,$C3,$FC
	.dc.b	$33,$E4,$FF,$23,$9C,$F8,$BF,$E1,$47,$F8,$87,$AF,$FC,$00,$8F,$F2
	.dc.b	$17,$BB,$FD,$37,$1E,$ED,$39,$9F,$E0,$D3,$FC,$43,$87,$85,$FF,$12
	.dc.b	$DF,$87,$5F,$B5,$C9,$FC,$38,$93,$FC,$BD,$FF,$20,$7A,$DF,$E4,$CF
	.dc.b	$53,$FC,$9A,$D9,$FE,$E5,$D9,$FA,$D3,$F5,$DD,$56,$3F,$93,$FC,$67
	.dc.b	$F1,$A2,$3F,$CE,$DE,$AF,$F6,$7B,$EA,$D3,$A9,$31,$FF,$0C,$57,$57
	.dc.b	$F2,$5E,$E7,$F2,$62,$4F,$F4,$77,$F9,$4F,$C6,$7F,$8A,$3C,$2F,$E7
	.dc.b	$8D,$9F,$E7,$6F,$C6,$F2,$33,$1F,$61,$B7,$FC,$FA,$79,$CF,$F5,$36
	.dc.b	$C6,$63,$04,$70,$43,$FA,$AC,$70,$7F,$AB,$FC,$4F,$F3,$E2,$4F,$F5
	.dc.b	$F7,$AF,$27,$E7,$6B,$FC,$83,$97,$92,$31,$5E,$A3,$F1,$FE,$5E,$5F
	.dc.b	$EA,$3F,$17,$FD,$28,$8F,$F6,$B7,$FC,$F9,$E8,$7F,$A2,$0F,$F4,$4F
	.dc.b	$9F,$67,$FC,$95,$D6,$F8,$BF,$E6,$EB,$E3,$2C,$61,$FE,$D1,$F1,$1F
	.dc.b	$D6,$09,$5F,$DE,$5E,$EF,$F8,$97,$0E,$ED,$3B,$D7,$36,$9E,$2A,$FE
	.dc.b	$E9,$F0,$FF,$D9,$09,$1F,$E0,$57,$FD,$2D,$1F,$EA,$2F,$83,$FF,$51
	.dc.b	$B6,$7F,$67,$E8,$7E,$90,$FD,$2F,$A5,$63,$FB,$EF,$8D,$FD,$C0,$8F
	.dc.b	$F8,$45,$EA,$FF,$8F,$6F,$AB,$4E,$A4,$E4,$FE,$9B,$EE,$3B,$D8,$77
	.dc.b	$9F,$E5,$47,$F7,$E2,$57,$F8,$7D,$FF,$02,$7E,$24,$F8,$1F,$F4,$DF
	.dc.b	$E1,$AF,$82,$FF,$0E,$4E,$6F,$FD,$1D,$39,$7F,$C3,$B1,$9B,$FF,$40
	.dc.b	$47,$0B,$FE,$C8,$D3,$FC,$59,$09,$E0,$FF,$8D,$1A,$FF,$8E,$9A,$8E
	.dc.b	$13,$E8,$24,$F4,$9D,$CE,$A3,$13,$A8,$FF,$1E,$4E,$8F,$53,$8B,$F8
	.dc.b	$64,$AE,$27,$BA,$BF,$D4,$BE,$07,$AE,$BF,$B2,$5F,$C1,$E8,$FF,$23
	.dc.b	$36,$FF,$D9,$9F,$AB,$7F,$25,$AF,$54,$E5,$BF,$05,$F4,$89,$5E,$B7
	.dc.b	$FD,$EB,$FC,$02,$F9,$0F,$F7,$8F,$37,$FB,$D2,$31,$5F,$E8,$38,$FE
	.dc.b	$67,$EB,$48,$FE,$8B,$EC,$5C,$7C,$FF,$0B,$F8,$02,$4F,$C9,$7B,$E3
	.dc.b	$DA,$3F,$DA,$18,$F6,$AF,$F0,$DC,$3E,$0B,$F7,$0F,$F5,$2C,$7B,$6D
	.dc.b	$FE,$EA,$E4,$67,$8A,$FB,$9F,$05,$F9,$02,$3F,$65,$F7,$FC,$87,$B4
	.dc.b	$F7,$7C,$85,$9F,$EB,$FE,$A7,$F5,$0F,$6F,$FA,$8B,$63,$F2,$7F,$0B
	.dc.b	$F9,$C2,$3F,$85,$7A,$BE,$CB,$EA,$D3,$A9,$FD,$B2,$BA,$B4,$FC,$FF
	.dc.b	$C2,$7E,$A0,$8F,$E4,$5F,$D5,$FC,$2F,$FB,$AA,$3F,$52,$D7,$F4,$EF
	.dc.b	$F6,$8F,$73,$31,$FD,$E9,$BC,$3F,$96,$9E,$73,$F9,$D6,$C6,$63,$04
	.dc.b	$70,$43,$F5,$91,$C1,$FD,$77,$C1,$FF,$00,$24,$FE,$A5,$EB,$C1,$F9
	.dc.b	$AA,$FF,$54,$3F,$EA,$9F,$05,$FA,$F7,$E6,$F0,$FD,$D5,$7A,$2B,$86
	.dc.b	$1F,$E1,$4F,$6F,$0F,$EE,$FE,$13,$F8,$79,$1F,$DC,$3F,$9E,$F9,$0F
	.dc.b	$CE,$7C,$8F,$E7,$5B,$3F,$E1,$EF,$67,$FC,$1E,$7F,$84,$3D,$CB,$1F
	.dc.b	$C4,$BE,$13,$F8,$D9,$1F,$E0,$8F,$57,$F3,$77,$D5,$A7,$52,$69,$FC
	.dc.b	$B3,$A5,$7D,$A3,$F7,$FF,$0B,$FC,$A0,$95,$FC,$45,$FD,$BF,$C7,$FE
	.dc.b	$D4,$FD,$B7,$C8,$D9,$FF,$1B,$FB,$7F,$E2,$8F,$77,$FC,$52,$D8,$FE
	.dc.b	$57,$F0,$5F,$CD,$08,$FF,$1D,$7A,$BF,$B4,$BE,$AD,$3A,$9F,$E4,$4A
	.dc.b	$EA,$D3,$F9,$BF,$C0,$FF,$40,$23,$FC,$A5,$FE,$87,$EF,$FE,$02,$3F
	.dc.b	$98,$73,$FE,$03,$DE,$FE,$87,$39,$D5,$39,$AF,$E8,$73,$9F,$92,$7C
	.dc.b	$9B,$2E,$35,$FC,$EF,$A3,$FD,$E9,$5A,$9C,$E3,$FC,$DB,$0F,$1D,$FF
	.dc.b	$3D,$6C,$66,$30,$47,$04,$3F,$A6,$47,$07,$FA,$77,$C1,$7F,$60,$24
	.dc.b	$FF,$49,$7A,$E1,$FC,$EA,$BD,$D9,$7E,$83,$DE,$F2,$CE,$6F,$DE,$7A
	.dc.b	$CF,$EA,$D9,$7B,$FB,$B7,$1F,$D6,$FD,$DF,$ED,$E4,$7F,$AC,$3F,$CF
	.dc.b	$BE,$43,$F9,$E7,$C9,$7F,$3D,$6C,$FF,$9F,$BE,$0F,$FC,$DE,$7F,$9C
	.dc.b	$3E,$0E,$C7,$F7,$2F,$7F,$FB,$D9,$1F,$EC,$8F,$57,$F9,$DD,$F5,$69
	.dc.b	$D4,$9C,$1F,$A1,$AE,$AF,$F5,$EF,$73,$FC,$0E,$49,$FE,$E0,$FF,$56
	.dc.b	$F9,$1F,$EA,$87,$F5,$5F,$92,$B3,$FE,$96,$F8,$2F,$F4,$3F,$C2,$7F
	.dc.b	$A2,$1B,$1F,$E0,$AF,$82,$FF,$09,$91,$FE,$F8,$F5,$7F,$A9,$DF,$56
	.dc.b	$9D,$4F,$F5,$1D,$75,$69,$FE,$16,$F6,$3F,$C3,$E4,$7F,$C1,$09,$FF
	.dc.b	$0A,$FE,$2D,$FD,$7F,$93,$F9,$07,$97,$FE,$2C,$9F,$BD,$AF,$F8,$4E
	.dc.b	$B5,$1F,$A6,$9D,$4F,$FB,$9E,$72,$BE,$0B,$FC,$6D,$F0,$9F,$A9,$84
	.dc.b	$EA,$73,$7D,$A7,$3E,$D7,$95,$FF,$11,$12,$BF,$BA,$38,$37,$3E,$A9
	.dc.b	$E1,$19,$9D,$BC,$63,$2D,$9E,$04,$7A,$1D,$0B,$3C,$9F,$85,$3C,$9F
	.dc.b	$89,$EA,$D7,$DA,$D9,$34,$C4,$FC,$38,$F8,$FD,$7F,$D3,$BE,$13,$13
	.dc.b	$30,$F8,$95,$F1,$FC,$9F,$E2,$31,$89,$E5,$EA,$77,$24,$7D,$C7,$FC
	.dc.b	$57,$F2,$5F,$E2,$9F,$93,$FF,$15,$36,$7E,$EE,$AE,$BF,$F1,$B3,$7E
	.dc.b	$0F,$FD,$BF,$BF,$F2,$18,$C5,$79,$3C,$EF,$71,$1F,$C5,$BE,$07,$FC
	.dc.b	$57,$8F,$81,$A7,$83,$0E,$E7,$83,$4F,$F3,$04,$67,$1F,$8D,$E8,$FC
	.dc.b	$B2,$47,$E6,$1F,$F2,$3F,$C9,$7F,$91,$4F,$F2,$37,$C9,$D9,$ED,$F0
	.dc.b	$5E,$9C,$7F,$CC,$3C,$FF,$8D,$D7,$E0,$4E,$7E,$1B,$EA,$7A,$3F,$64
	.dc.b	$98,$7E,$99,$EE,$FF,$91,$B2,$EE,$D3,$B9,$ED,$9F,$05,$75,$8F,$B7
	.dc.b	$E7,$FE,$14,$8F,$EC,$9F,$C4,$78,$7F,$CC,$BA,$FF,$C9,$73,$F8,$28
	.dc.b	$C7,$BF,$F8,$97,$FA,$87,$CE,$BC,$2C,$57,$EE,$3C,$A4,$E6,$39,$1F
	.dc.b	$C0,$2B,$BB,$49,$CF,$FA,$08,$CB,$63,$2F,$EB,$63,$2D,$8F,$E1,$7E
	.dc.b	$47,$E8,$48,$E0,$EB,$C3,$FB,$6E,$C5,$FB,$FE,$07,$D3,$87,$F0,$B8
	.dc.b	$B5,$FD,$A3,$E7,$5F,$03,$8C,$7F,$82,$79,$1F,$AF,$23,$FC,$71,$FF
	.dc.b	$45,$FC,$87,$FA,$27,$E5,$3F,$D1,$4D,$9F,$C9,$7C,$27,$E1,$4F,$C3
	.dc.b	$7C,$3D,$8F,$D8,$F9,$7F,$BB,$23,$FC,$A1,$EA,$FF,$A2,$2F,$AB,$4E
	.dc.b	$A4,$D7,$FD,$B3,$D8,$BE,$D1,$FC,$A7,$C1,$FE,$10,$4A,$FE,$6A,$FF
	.dc.b	$A9,$3E,$3B,$FD,$44,$7F,$A8,$FE,$3E,$CF,$EA,$FE,$17,$F3,$DF,$0F
	.dc.b	$F9,$F6,$C7,$F0,$AF,$1F,$F8,$A1,$1F,$E8,$2F,$57,$FD,$43,$7D,$5A
	.dc.b	$75,$3F,$71,$5D,$5A,$7F,$16,$F0,$FF,$90,$11,$FE,$9A,$FF,$23,$F0
	.dc.b	$7F,$39,$F0,$FF,$C8,$5B,$3F,$C7,$9F,$DB,$7C,$36,$63,$F8,$33,$75
	.dc.b	$FE,$67,$3C,$E7,$FA,$CB,$63,$31,$82,$38,$21,$FD,$16,$38,$3F,$D1
	.dc.b	$F8,$3F,$9C,$12,$7F,$B2,$BD,$79,$FF,$31,$5F,$B5,$E6,$F2,$46,$2B
	.dc.b	$D4,$7F,$1D,$F8,$7C,$BF,$D6,$79,$BF,$A2,$11,$FE,$E2,$FF,$B8,$1F
	.dc.b	$F6,$F7,$C6,$7F,$B7,$DA,$7F,$27,$F8,$5F,$E3,$47,$F1,$BF,$87,$B1
	.dc.b	$FD,$1B,$93,$FA,$81,$1F,$EF,$AF,$57,$FD,$BB,$5D,$5A,$75,$26,$BF
	.dc.b	$96,$AE,$AF,$F7,$DE,$2F,$EB,$84,$9F,$F0,$4B,$FD,$8D,$7F,$00,$BE
	.dc.b	$0B,$FC,$D1,$18,$6C,$3E,$4A,$C7,$F3,$5F,$85,$FE,$6D,$F0,$F6,$FF
	.dc.b	$A3,$3F,$D9,$F3,$FE,$D8,$48,$FF,$0C,$BE,$1C,$3C,$11,$FE,$18,$3C
	.dc.b	$05,$7C,$0D,$BE,$D9,$7E,$D2,$EF,$07,$F4,$5E,$C4,$7F,$77,$BF,$EF
	.dc.b	$84,$3F,$C5,$27,$F8,$07,$D8,$EB,$6B,$FF,$15,$FC,$07,$F8,$86,$73
	.dc.b	$18,$AF,$F1,$04,$61,$FF,$35,$FC,$3B,$E2,$78,$6F,$FE,$8D,$8C,$4F
	.dc.b	$08,$F0,$11,$FF,$0C,$E2,$BF,$1B,$FE,$27,$6B,$FE,$2C,$6A,$38,$4F
	.dc.b	$37,$FC,$47,$1A,$73,$3D,$A3,$04,$D3,$E1,$12,$34,$F3,$C7,$FC,$6B
	.dc.b	$3E,$FA,$FF,$55,$F9,$AF,$8A,$F7,$E1,$ED,$24,$FC,$B7,$E2,$FC,$3E
	.dc.b	$A3,$07,$A6,$DF,$0B,$F4,$C9,$5E,$77,$F0,$DF,$21,$F9,$27,$F2,$6D
	.dc.b	$8F,$F0,$A7,$C3,$FF,$81,$CF,$F0,$47,$C5,$58,$FA,$F3,$F7,$C8,$FC
	.dc.b	$37,$AB,$F8,$FA,$EA,$D3,$A9,$30,$FE,$87,$5D,$5F,$59,$F6,$7D,$FF
	.dc.b	$C6,$12,$7E,$83,$F9,$FF,$53,$F5,$67,$EB,$1B,$1F,$E2,$DF,$88,$FF
	.dc.b	$10,$7C,$4F,$F8,$85,$B1,$F9,$0A,$FC,$C1,$1F,$B8,$F5,$7F,$50,$F5
	.dc.b	$69,$D4,$FF,$1D,$4F,$56,$9F,$9C,$F4,$7F,$48,$47,$F1,$0F,$E9,$FC
	.dc.b	$6F,$F0,$F7,$C5,$FE,$95,$B3,$FA,$47,$FC,$71,$F1,$59,$8F,$F2,$43
	.dc.b	$73,$FC,$AC,$F3,$9F,$CA,$36,$33,$18,$23,$82,$1F,$A5,$8E,$0F,$E9
	.dc.b	$FC,$5F,$DE,$12,$7F,$40,$F5,$E5,$FC,$DD,$7F,$8D,$FA,$7C,$91,$8A
	.dc.b	$F5,$1F,$E5,$FF,$89,$CB,$FB,$1E,$BF,$E1,$04,$7F,$58,$FF,$1B,$BF
	.dc.b	$E4,$FA,$7F,$28,$6C,$FF,$9A,$7E,$1F,$FC,$B8,$7F,$97,$7E,$26,$C7
	.dc.b	$F0,$CE,$9F,$E2,$84,$7F,$70,$F5,$7F,$92,$5F,$56,$9D,$49,$AF,$F8
	.dc.b	$62,$BA,$BF,$96,$E8,$70,$7F,$21,$25,$7F,$05,$7F,$9A,$67,$FD,$00
	.dc.b	$FE,$83,$BD,$9F,$F4,$8F,$CD,$7F,$A9,$9B,$F0,$FF,$E8,$F8,$C6,$3F
	.dc.b	$BE,$F8,$24,$7F,$B3,$A3,$F8,$F7,$17,$F3,$12,$5F,$F1,$57,$C2,$FF
	.dc.b	$41,$C7,$C2,$D3,$C3,$F1,$3D,$F1,$F0,$36,$3F,$9A,$F6,$FF,$3D,$23
	.dc.b	$FC,$85,$FE,$81,$E0,$FF,$9D,$DF,$E7,$CD,$8F,$E7,$AF,$FB,$6F,$97
	.dc.b	$FD,$B9,$EF,$3F,$0D,$7C,$2F,$0B,$FF,$3C,$46,$1E,$27,$F2,$FB,$EE
	.dc.b	$D2,$6B,$FC,$DA,$32,$D8,$CB,$FD,$06,$32,$D8,$FE,$75,$F1,$9F,$D6
	.dc.b	$48,$E0,$EB,$8F,$F3,$8A,$FF,$4A,$1E,$7A,$FC,$13,$63,$DB,$F8,$77
	.dc.b	$FB,$2E,$BD,$35,$F8,$7F,$23,$FB,$31,$23,$FA,$53,$FD,$E7,$93,$FC
	.dc.b	$0B,$D1,$FE,$06,$6C,$FE,$1B,$E2,$BF,$08,$7E,$17,$E2,$EC,$7F,$6A
	.dc.b	$F8,$DF,$EE,$44,$7F,$AB,$3D,$5F,$F0,$15,$F5,$69,$D4,$98,$FE,$3F
	.dc.b	$99,$7D,$A3,$F4,$1F,$0D,$FD,$FC,$95,$FD,$8D,$FF,$09,$71,$FF,$87
	.dc.b	$8F,$F0,$FF,$3D,$9F,$CF,$FC,$57,$E7,$79,$3F,$3C,$D8,$FF,$01,$F5
	.dc.b	$FF,$82,$C8,$FF,$6F,$7A,$BF,$E1,$BB,$EA,$D3,$A9,$FA,$CA,$EA,$D3
	.dc.b	$FC,$1F,$F1,$BF,$E1,$72,$3F,$DE,$DF,$F0,$D6,$5F,$ED,$1E,$5F,$F0
	.dc.b	$C3,$67,$FC,$2E,$FE,$D7,$E2,$73,$1F,$BA,$6C,$7F,$82,$0E,$71,$FE
	.dc.b	$07,$69,$97,$04,$70,$43,$FC,$30,$70,$3F,$C2,$C7,$F8,$5F,$F1,$87
	.dc.b	$C6,$A7,$FC,$43,$18,$6C,$66,$78,$9C,$5F,$1F,$3F,$F8,$52,$3F,$DE
	.dc.b	$1C,$DE,$A8,$C5,$7C,$03,$F8,$BF,$C4,$65,$F1,$E0,$F8,$DE,$D8,$76
	.dc.b	$8C,$4F,$92,$73,$3C,$1F,$F2,$C6,$7F,$E6,$4E,$AF,$F3,$2B,$6B,$F9
	.dc.b	$2F,$C4,$7F,$23,$3F,$92,$7C,$5D,$8F,$F1,$C6,$7E,$0C,$9F,$1B,$DB
	.dc.b	$3F,$8C,$D5,$1A,$AF,$8F,$3E,$27,$FC,$CD,$97,$89,$A7,$88,$95,$FA
	.dc.b	$5C,$3C,$4F,$F8,$62,$FF,$DB,$93,$F4,$BE,$23,$C1,$18,$9F,$A2,$78
	.dc.b	$A3,$FC,$F1,$D7,$FE,$84,$3F,$D0,$BB,$DB,$FE,$8B,$F1,$1F,$D0,$FE
	.dc.b	$27,$FA,$23,$63,$E0,$9F,$4B,$E2,$BB,$91,$FC,$1B,$D9,$FF,$41,$5F
	.dc.b	$66,$9D,$8F,$EA,$75,$D9,$A7,$6C,$7F,$21,$F1,$1E,$C2,$47,$E4,$1F
	.dc.b	$7F,$C5,$7F,$3F,$8F,$F3,$67,$3B,$D9,$63,$9C,$3F,$35,$76,$B5,$39
	.dc.b	$9F,$AF,$3F,$D8,$FE,$2B,$1B,$A3,$18,$F9,$6B,$6F,$E8,$A3,$C6,$D2
	.dc.b	$47,$D2,$9E,$51,$88,$FC,$C4,$F0,$74,$D7,$31,$FA,$EF,$89,$FB,$A4
	.dc.b	$70,$76,$E4,$FC,$A5,$7B,$73,$78,$1E,$A8,$95,$E9,$C7,$C7,$8F,$F3
	.dc.b	$1F,$8A,$CB,$F8,$DF,$88,$FC,$59,$2B,$F7,$EF,$FB,$07,$C0,$FF,$66
	.dc.b	$E3,$FE,$CE,$6C,$FF,$83,$BE,$2B,$FC,$18,$7F,$83,$7E,$2A,$C7,$87
	.dc.b	$E1,$FF,$2E,$47,$F8,$5B,$D5,$FF,$65,$5F,$56,$9D,$49,$8F,$F4,$9A
	.dc.b	$EA,$FE,$97,$E2,$7F,$46,$49,$FE,$2E,$FF,$B6,$63,$FD,$CA,$7F,$B9
	.dc.b	$B8,$EC,$FF,$89,$BE,$27,$FC,$49,$F1,$5F,$E2,$56,$C7,$C9,$F8,$6F
	.dc.b	$D7,$91,$FE,$46,$F5,$7F,$DC,$57,$D5,$A7,$53,$FC,$67,$5D,$5A,$7D
	.dc.b	$5F,$89,$FD,$D9,$1F,$E5,$EF,$EF,$BE,$0B,$EF,$6F,$F9,$66,$CF,$DD
	.dc.b	$7F,$C7,$9F,$15,$98,$FF,$20,$B7,$0F,$CA,$4F,$39,$FE,$76,$D8,$CC
	.dc.b	$60,$8E,$08,$7F,$31,$8E,$0F,$F3,$3F,$86,$FE,$26,$49,$FE,$8E,$F5
	.dc.b	$E7,$FE,$55,$5F,$97,$D7,$C9,$18,$AF,$51,$FE,$58,$F8,$AC,$BF,$D0
	.dc.b	$BE,$0B,$F8,$F9,$1F,$EA,$6F,$F3,$35,$F1,$3F,$D1,$32,$FF,$6C,$C6
	.dc.b	$3C,$BF,$E0,$8E,$BC,$AB,$5F,$D6,$EB,$FC,$C5,$F8,$77,$FB,$A7,$DD
	.dc.b	$37,$D4,$FE,$C5,$87,$F1,$79,$CD,$FF,$08,$F8,$6F,$E6,$64,$9F,$EC
	.dc.b	$EF,$96,$3C,$9A,$FF,$67,$9F,$21,$FC,$84,$F2,$16,$3C,$91,$FD,$B6
	.dc.b	$BB,$CF,$87,$0F,$23,$FC,$7E,$33,$8F,$F4,$7F,$85,$FE,$8A,$49,$FE
	.dc.b	$F2,$F8,$FC,$AF,$82,$7C,$2F,$36,$CF,$5F,$88,$E9,$8B,$E2,$AC,$61
	.dc.b	$FE,$4F,$F0,$FF,$D4,$89,$5F,$E0,$67,$B3,$EB,$C3,$B3,$4E,$C7,$AE
	.dc.b	$7A,$B4,$FE,$31,$F0,$9F,$D7,$88,$FF,$84,$9F,$EC,$3F,$0D,$FC,$87
	.dc.b	$4F,$E7,$2D,$9F,$E4,$0F,$F4,$3F,$81,$77,$3D,$2D,$D7,$F5,$19,$CC
	.dc.b	$62,$FF,$9D,$5F,$6A,$FF,$10,$B6,$33,$18,$23,$82,$1F,$E1,$78,$E0
	.dc.b	$FF,$89,$BF,$19,$7C,$27,$F8,$25,$B5,$FD,$EA,$51,$87,$FB,$EB,$5E
	.dc.b	$91,$EE,$3F,$B5,$75,$FF,$A8,$3A,$7D,$5C,$BF,$E1,$9C,$6D,$2D,$9B
	.dc.b	$E8,$7F,$D8,$77,$BA,$FE,$C1,$97,$FB,$1A,$FD,$1F,$0B,$E7,$24,$FF
	.dc.b	$84,$0F,$D8,$FA,$5F,$AC,$F5,$3F,$5A,$D8,$FC,$C7,$C5,$FE,$30,$FC
	.dc.b	$6F,$C6,$58,$EF,$F0,$5F,$0C,$8F,$C3,$CB,$FD,$0E,$F7,$ED,$61,$D7
	.dc.b	$3F,$DF,$C6,$27,$F5,$59,$75,$9F,$DE,$E5,$E0,$9F,$19,$EE,$F8,$4F
	.dc.b	$A4,$47,$C0,$FE,$E7,$D5,$FE,$0E,$7F,$08,$F5,$2C,$FE,$CF,$E2,$BF
	.dc.b	$4F,$F1,$9F,$A8,$6C,$7C,$AF,$82,$FB,$C4,$7D,$8F,$57,$F8,$05,$F5
	.dc.b	$69,$D4,$FE,$01,$5D,$5A,$7D,$6F,$84,$FC,$51,$1F,$90,$FE,$37,$E0
	.dc.b	$FF,$C4,$57,$FC,$4A,$7F,$CF,$FC,$09,$CF,$91,$FE,$25,$8C,$4F,$F0
	.dc.b	$FF,$8C,$C5,$7F,$04,$6C,$7D,$58,$EC,$D2,$6F,$F9,$88,$CB,$63,$2F
	.dc.b	$E8,$E3,$2D,$8F,$C0,$7C,$27,$E9,$48,$E0,$EB,$CF,$FE,$84,$C3,$CF
	.dc.b	$48,$FF,$4A,$76,$33,$FF,$3A,$4E,$6F,$E0,$E3,$EA,$8C,$1F,$A5,$C3
	.dc.b	$9E,$1F,$CC,$7C,$57,$C6,$6E,$3C,$1E,$07,$EE,$49,$8F,$E5,$DF,$E8
	.dc.b	$1E,$C7,$F4,$EF,$6B,$FA,$7B,$67,$F9,$D7,$C4,$7F,$2E,$3F,$97,$FC
	.dc.b	$65,$8F,$D8,$FC,$37,$F0,$52,$3F,$A3,$7A,$BF,$D1,$EF,$AB,$4E,$A4
	.dc.b	$D7,$FA,$ED,$75,$7D,$E7,$F0,$EF,$83,$FE,$24,$49,$FD,$83,$FD,$67
	.dc.b	$C7,$FE,$CE,$7F,$68,$F6,$EC,$FF,$53,$F8,$9F,$E8,$FF,$15,$FD,$21
	.dc.b	$B1,$FC,$2B,$E1,$BF,$8F,$11,$FD,$E3,$D5,$FE,$C1,$7D,$5A,$75,$3F
	.dc.b	$B0,$57,$56,$9F,$C5,$BE,$17,$F9,$51,$1F,$E1,$0F,$F2,$DF,$83,$FD
	.dc.b	$0F,$A1,$F9,$D6,$CF,$F1,$E7,$FA,$FF,$C5,$66,$3F,$B7,$37,$5F,$E6
	.dc.b	$33,$CE,$7F,$8A,$36,$33,$18,$23,$82,$1F,$D0,$63,$83,$FD,$0F,$E0
	.dc.b	$BF,$A1,$12,$7F,$90,$3D,$79,$7F,$3D,$5F,$BA,$F1,$BC,$91,$8A,$F5
	.dc.b	$1F,$E0,$5F,$89,$CB,$FB,$EF,$85,$FE,$9C,$47,$F9,$63,$FE,$15,$F7
	.dc.b	$BF,$C4,$7F,$05,$FE,$24,$6C,$FF,$83,$FE,$27,$FC,$06,$7F,$81,$3E
	.dc.b	$2A,$C7,$F4,$6F,$85,$FE,$B4,$47,$F9,$C3,$D5,$FF,$0E,$5F,$56,$9D
	.dc.b	$49,$B7,$F8,$CA,$BA,$BF,$96,$F0,$1F,$0D,$FD,$9C,$95,$FD,$15,$FF
	.dc.b	$17,$7C,$17,$F8,$F0,$FF,$1E,$FC,$25,$9F,$F1,$47,$C3,$FF,$87,$3E
	.dc.b	$2B,$FC,$3A,$D8,$FE,$BF,$F0,$FF,$DC,$C8,$FF,$51,$7A,$BF,$E3,$5B
	.dc.b	$EA,$D3,$A9,$FE,$35,$AE,$AD,$3F,$B5,$FC,$3F,$F7,$F2,$3F,$D7,$5F
	.dc.b	$F0,$13,$FC,$DB,$D6,$FE,$64,$D3,$FB,$99,$FE,$34,$F8,$8C,$BF,$E4
	.dc.b	$36,$EB,$FE,$06,$8E,$67,$F6,$92,$B9,$70,$47,$04,$3F,$C2,$CF,$07
	.dc.b	$FC,$10,$7F,$82,$7F,$1A,$FF,$37,$C7,$F8,$5D,$B1,$A3,$5E,$C2,$74
	.dc.b	$F8,$B9,$FF,$9C,$47,$F5,$4F,$5B,$D3,$18,$8F,$79,$FE,$5F,$F3,$F2
	.dc.b	$F8,$B0,$FF,$2A,$3E,$1F,$82,$ED,$8F,$84,$4E,$9C,$35,$F5,$CF,$FA
	.dc.b	$23,$E1,$BF,$D0,$7F,$11,$FE,$84,$6C,$7F,$9F,$FF,$0F,$FF,$39,$BF
	.dc.b	$E2,$9F,$89,$46,$2D,$1F,$E8,$87,$FC,$E9,$19,$7C,$7F,$11,$E6,$7E
	.dc.b	$36,$28,$C4,$7B,$5A,$FC,$F7,$FD,$15,$87,$92,$3F,$C3,$ED,$D3,$FD
	.dc.b	$0D,$7E,$3C,$3C,$BF,$0B,$FE,$7A,$BF,$1B,$F1,$AB,$C5,$8B,$E0,$7B
	.dc.b	$D7,$C8,$6C,$67,$3F,$BD,$1F,$E9,$BF,$87,$FF,$54,$9F,$EA,$AF,$88
	.dc.b	$B3,$F1,$FE,$27,$E0,$FB,$BF,$09,$B1,$FE,$51,$CB,$C3,$F0,$7D,$89
	.dc.b	$1F,$8B,$7B,$3F,$EA,$3C,$3B,$34,$EC,$7D,$5B,$EC,$D3,$B1,$F8,$7F
	.dc.b	$80,$F5,$91,$FC,$C3,$EE,$F8,$2F,$F1,$2F,$C1,$7C,$59,$FC,$74,$61
	.dc.b	$E7,$B7,$E8,$63,$FD,$19,$F7,$BF,$E8,$B9,$CC,$FE,$42,$73,$EE,$73
	.dc.b	$8E,$13,$FA,$79,$F0,$34,$91,$F9,$68,$CB,$63,$2F,$E2,$23,$2D,$8F
	.dc.b	$D9,$7C,$17,$D8,$23,$83,$AE,$BF,$80,$AF,$F2,$1F,$B5,$E5,$BF,$45
	.dc.b	$7A,$B8,$95,$96,$C7,$FA,$9F,$D3,$7C,$4E,$6B,$5E,$FF,$E2,$09,$5F
	.dc.b	$C1,$5F,$F7,$B7,$F0,$0F,$FA,$6F,$3F,$F7,$13,$67,$FD,$28,$7F,$BE
	.dc.b	$AF,$FD,$EF,$19,$CB,$F5,$BD,$2F,$4B,$F5,$BE,$B7,$EB,$9B,$87,$A7
	.dc.b	$DD,$FC,$C9,$1F,$E2,$CF,$8B,$E2,$FA,$9F,$56,$3C,$07,$FB,$CF,$1F
	.dc.b	$07,$6F,$EF,$AF,$AE,$5F,$E4,$68,$C5,$7F,$05,$C7,$E3,$7C,$07,$EA
	.dc.b	$09,$1F,$C9,$5F,$83,$F1,$1E,$CF,$89,$F6,$B6,$7F,$87,$78,$DF,$C2
	.dc.b	$0F,$E1,$3E,$4D,$8F,$A9,$EE,$7E,$D0,$8F,$F3,$17,$AB,$E9,$BE,$AD
	.dc.b	$3A,$93,$3F,$57,$07,$FA,$9D,$B8,$7D,$FF,$7F,$F8,$09,$2B,$F9,$F3
	.dc.b	$FC,$13,$DD,$FF,$45,$FC,$1F,$E1,$5B,$3F,$89,$7F,$90,$F7,$66,$3F
	.dc.b	$95,$37,$0F,$E6,$73,$CE,$7F,$A5,$36,$33,$18,$23,$82,$1F,$CA,$A3
	.dc.b	$83,$FC,$B3,$DC,$FE,$32,$49,$FE,$AC,$F5,$E7,$FE,$43,$5F,$EA,$4F
	.dc.b	$81,$F2,$46,$2B,$D4,$7F,$41,$ED,$CB,$FD,$47,$D8,$FE,$4A,$47,$FB
	.dc.b	$13,$F9,$EF,$86,$FC,$C9,$F9,$AF,$89,$B3,$FD,$2F,$0F,$E8,$BA,$FF
	.dc.b	$46,$6C,$7E,$C7,$DD,$FE,$62,$47,$FB,$73,$D5,$FC,$B5,$F5,$69,$D4
	.dc.b	$FE,$B1,$5D,$5A,$7E,$F7,$D5,$FE,$7A,$47,$FB,$D3,$FB,$6F,$87,$FD
	.dc.b	$8F,$C4,$FE,$C9,$B3,$FD,$8F,$DE,$FE,$BA,$7F,$5E,$F8,$4B,$1F,$C2
	.dc.b	$7D,$CF,$E9,$24,$7F,$C0,$CF,$57,$F5,$D7,$D5,$A7,$52,$71,$FE,$5A
	.dc.b	$BA,$BF,$E9,$8F,$1D,$EC,$FF,$56,$25,$7F,$84,$DF,$EB,$3E,$C7,$FB
	.dc.b	$82,$FF,$8D,$36,$7F,$8F,$3F,$DD,$BE,$07,$31,$FD,$FD,$B8,$7F,$82
	.dc.b	$27,$9C,$FF,$87,$DB,$19,$8C,$11,$C1,$0F,$EF,$F1,$C1,$FF,$00,$9F
	.dc.b	$DB,$89,$1F,$E2,$B7,$AF,$2F,$F7,$69,$FF,$7C,$E9,$E4,$8C,$57,$A8
	.dc.b	$FF,$0B,$7B,$D9,$7F,$C6,$FE,$E3,$D7,$FE,$FA,$D8,$FF,$13,$62,$8C
	.dc.b	$3F,$E2,$96,$BF,$E2,$6F,$1F,$FA,$F3,$FE,$E6,$94,$63,$0F,$E7,$7F
	.dc.b	$13,$FC,$E3,$D0,$FF,$10,$FB,$7F,$E2,$2F,$7A,$E1,$FE,$2B,$7F,$A8
	.dc.b	$64,$F5,$7C,$A4,$9F,$2C,$65,$FF,$1E,$1E,$0B,$EC,$FF,$37,$C2,$E7
	.dc.b	$FC,$73,$1F,$04,$FF,$8E,$5B,$1F,$EE,$0B,$7B,$1F,$10,$95,$E0,$7F
	.dc.b	$AB,$FC,$3F,$F5,$2F,$88,$FE,$A6,$D9,$FF,$22,$7B,$7F,$E3,$E3,$FC
	.dc.b	$7F,$EE,$D8,$FF,$01,$E0,$F3,$FE,$89,$23,$DA,$F6,$7F,$A8,$61,$D9
	.dc.b	$A7,$62,$65,$FC,$E2,$FB,$3E,$BF,$47,$EE,$92,$7E,$4B,$F8,$0F,$3B
	.dc.b	$FC,$25,$CE,$F0,$5D,$3F,$4A,$73,$18,$BF,$4C,$7F,$96,$BD,$BD,$47
	.dc.b	$F9,$8D,$B8,$FE,$1E,$7B,$4F,$D8,$6C,$66,$30,$47,$04,$3F,$D1,$1A
	.dc.b	$FF,$53,$6C,$FE,$63,$CC,$95,$F8,$46,$CF,$2E,$7F,$F3,$65,$FF,$5B
	.dc.b	$7F,$D0,$79,$FB,$FA,$BC,$F1,$AC,$6E,$3F,$E7,$BF,$37,$FC,$36,$FC
	.dc.b	$03,$EC,$F9,$9F,$A8,$23,$F9,$47,$FC,$37,$EC,$7F,$85,$8F,$F0,$BF
	.dc.b	$C4,$59,$FF,$23,$FA,$8F,$85,$7A,$5F,$E4,$38,$C7,$AD,$67,$35,$C1
	.dc.b	$FF,$10,$70,$7E,$D7,$D0,$FD,$C9,$27,$F4,$AF,$89,$FF,$0C,$65,$E2
	.dc.b	$69,$CF,$1E,$D8,$77,$6C,$7E,$5B,$CC,$FE,$0A,$47,$F6,$2F,$F8,$E3
	.dc.b	$E2,$3F,$C6,$7F,$11,$FE,$34,$6C,$F9,$FD,$1F,$51,$EB,$F5,$EC,$7E
	.dc.b	$8B,$D2,$FE,$22,$47,$F7,$AF,$57,$FC,$61,$7D,$5A,$75,$26,$BF,$E3
	.dc.b	$1D,$97,$DA,$3F,$79,$E5,$FF,$1E,$25,$7F,$0A,$7F,$90,$F9,$7F,$9E
	.dc.b	$F2,$7F,$88,$B6,$7F,$6E,$FF,$9D,$FE,$2D,$F2,$CF,$5F,$75,$FE,$82
	.dc.b	$9C,$C6,$31,$FE,$51,$7D,$AB,$F8,$CB,$63,$31,$82,$38,$21,$FA,$28
	.dc.b	$E0,$FE,$93,$C1,$FE,$72,$49,$FE,$4A,$F5,$EB,$FD,$0D,$7E,$EB,$C9
	.dc.b	$F2,$46,$2B,$DA,$7A,$7D,$77,$B5,$A8,$FE,$8F,$DF,$FD,$18,$8F,$F3
	.dc.b	$27,$FC,$F3,$C3,$FE,$74,$3F,$CE,$BF,$15,$67,$F3,$1E,$A7,$E6,$FC
	.dc.b	$6F,$CE,$36,$3F,$91,$F1,$7F,$52,$23,$FC,$F9,$EA,$FF,$9C,$6F,$AB
	.dc.b	$4E,$A7,$E9,$6B,$AB,$4F,$E5,$FA,$7F,$5E,$23,$FD,$29,$FF,$48,$FC
	.dc.b	$47,$FA,$33,$E2,$7F,$D1,$AD,$9F,$D8,$7A,$1F,$B3,$3F,$69,$EC,$58
	.dc.b	$FE,$77,$E6,$7F,$6A,$23,$FD,$59,$EA,$FF,$A2,$6F,$AB,$4E,$A4,$E4
	.dc.b	$FC,$D5,$75,$7F,$AA,$79,$FF,$DD,$89,$3F,$D8,$9F,$EF,$3E,$67,$F3
	.dc.b	$6C,$3F,$B4,$B6,$7F,$A7,$BF,$C2,$FD,$2C,$C7,$F0,$B6,$DF,$F7,$D9
	.dc.b	$E7,$3F,$DB,$9B,$19,$8C,$11,$C1,$0F,$E6,$F1,$C1,$FE,$73,$EF,$FF
	.dc.b	$83,$89,$3F,$DE,$9E,$BC,$DF,$CD,$AB,$FA,$7E,$3E,$48,$C5,$7A,$8F
	.dc.b	$E4,$5E,$86,$56,$13,$F1,$C7,$8C,$9F,$F0,$DC,$61,$B1,$99,$E2,$49
	.dc.b	$D3,$83,$97,$AD,$CE,$BF,$D8,$47,$FB,$4B,$D0,$FF,$69,$BF,$CD,$FC
	.dc.b	$CF,$E7,$1E,$A5,$8F,$E7,$0F,$FB,$AF,$0F,$E7,$78,$75,$0E,$1F,$03
	.dc.b	$5F,$27,$95,$FE,$E6,$AE,$E7,$FB,$4F,$0B,$3F,$C8,$70,$F0,$CF,$F4
	.dc.b	$76,$C3,$4F,$80,$F8,$3C,$5A,$FC,$16,$BF,$2D,$AE,$EB,$FC,$5F,$FC
	.dc.b	$02,$F8,$AF,$F7,$32,$F8,$BB,$18,$7F,$AD,$79,$32,$BF,$AD,$79,$D7
	.dc.b	$0F,$24,$7D,$1F,$33,$BE,$E6,$92,$3C,$38,$F5,$7F,$03,$A7,$DB,$59
	.dc.b	$F9,$5C,$9F,$86,$33,$9F,$86,$7F,$05,$E8,$7F,$73,$D3,$FB,$86,$99
	.dc.b	$BF,$C2,$77,$F9,$08,$FE,$31,$F3,$F8,$5F,$E2,$6E,$8F,$BC,$D9,$F0
	.dc.b	$EF,$F9,$58,$FE,$F9,$E7,$6A,$7F,$BE,$36,$7F,$09,$C0,$AE,$B5,$F9
	.dc.b	$A6,$CE,$9A,$FE,$21,$AF,$E6,$DA,$F8,$63,$F4,$7D,$BF,$38,$8E,$0E
	.dc.b	$C7,$A2,$3F,$2B,$3F,$E4,$2E,$BF,$36,$DF,$8A,$C3,$CD,$7B,$6D,$FF
	.dc.b	$73,$F1,$37,$3F,$AD,$F0,$3E,$E9,$1F,$DA,$BF,$90,$7E,$CF,$C5,$7E
	.dc.b	$30,$FC,$6F,$C5,$D8,$FF,$11,$78,$5F,$E2,$5F,$17,$FC,$4C,$D8,$F1
	.dc.b	$78,$5F,$8A,$23,$FC,$09,$ED,$CE,$C3,$A6,$7F,$E9,$98,$63,$8D,$7F
	.dc.b	$0B,$93,$2E,$D7,$FE,$0F,$6C,$FC,$3F,$07,$F3,$24,$CF,$F8,$8B,$E0
	.dc.b	$F8,$AF,$D3,$FC,$5F,$EA,$1B,$3F,$E3,$DE,$FF,$F2,$01,$FE,$41,$F1
	.dc.b	$AC,$7C,$DF,$0B,$F4,$A4,$7F,$8E,$BD,$5F,$C7,$9F,$A2,$BE,$CE,$36
	.dc.b	$47,$4D,$3F,$C5,$18,$77,$AF,$11,$F6,$FC,$1F,$D9,$91,$FE,$56,$FE
	.dc.b	$DB,$AF,$FD,$13,$FB,$0F,$F9,$53,$99,$18,$F1,$7F,$90,$3F,$E5,$BB
	.dc.b	$A7,$F9,$6C,$B8,$FE,$82,$3A,$47,$F3,$86,$C6,$63,$04,$70,$43,$F9
	.dc.b	$4C,$70,$7F,$95,$F7,$7F,$0C,$24,$FF,$44,$7A,$F0,$FE,$6E,$BD,$7E
	.dc.b	$27,$92,$31,$5F,$9D,$9F,$F3,$B7,$E2,$7F,$C7,$E3,$07,$E8,$7B,$BF
	.dc.b	$8C,$91,$FE,$A2,$FE,$43,$E3,$3F,$93,$9F,$CA,$3E,$32,$CF,$FA,$12
	.dc.b	$BF,$22,$DF,$17,$BF,$57,$81,$FE,$95,$DD,$FC,$98,$93,$FD,$79,$EC
	.dc.b	$FD,$FC,$3B,$34,$ED,$5C,$DA,$77,$3F,$75,$D7,$FC,$C8,$8F,$F6,$A7
	.dc.b	$F5,$3F,$17,$FD,$03,$E3,$3F,$A0,$B6,$7D,$5C,$FF,$18,$F8,$FD,$D6
	.dc.b	$3F,$83,$77,$7F,$3E,$23,$FD,$D9,$EA,$FE,$5E,$FA,$B4,$EA,$4E,$4F
	.dc.b	$E9,$23,$C6,$6F,$FB,$D9,$D7,$1F,$F1,$7E,$DF,$E9,$A4,$8F,$F0,$2B
	.dc.b	$FD,$43,$AF,$FD,$C7,$F9,$FB,$E1,$BF,$BC,$46,$1B,$5F,$E7,$FF,$9A
	.dc.b	$D3,$FC,$86,$7F,$D0,$D1,$8C,$FC,$A7,$F7,$E8,$ED,$3F,$E1,$46,$C7
	.dc.b	$F4,$06,$BF,$E0,$26,$BF,$E1,$56,$BF,$E1,$F6,$BF,$E0,$A8,$FE,$D2
	.dc.b	$43,$FC,$42,$76,$C7,$F9,$C4,$FE,$E3,$CC,$F5,$D7,$9B,$B5,$38,$AF
	.dc.b	$F1,$37,$B0,$F9,$97,$26,$72,$FF,$14,$FE,$3D,$FA,$3C,$BF,$C3,$8D
	.dc.b	$7F,$C5,$4D,$77,$E7,$7F,$83,$23,$11,$FE,$3A,$6A,$69,$23,$81,$27
	.dc.b	$FC,$55,$3B,$8E,$9D,$72,$73,$5F,$E0,$3A,$F2,$CE,$5B,$3A,$8C,$47
	.dc.b	$A1,$AF,$B9,$AF,$9F,$A7,$D4,$49,$EE,$D8,$F8,$6D,$7E,$1B,$5F,$92
	.dc.b	$D7,$A4,$7A,$39,$FE,$31,$1C,$1E,$CE,$8F,$F2,$D6,$BF,$E5,$BF,$AA
	.dc.b	$7C,$CF,$F9,$5F,$0F,$C8,$E7,$FD,$03,$0F,$F0,$9D,$69,$E2,$7D,$AC
	.dc.b	$3F,$8E,$F9,$18,$BF,$A9,$18,$9F,$F0,$25,$F3,$9F,$BF,$3F,$6C,$93
	.dc.b	$E5,$24,$69,$AF,$E1,$23,$6E,$48,$7E,$22,$BF,$0C,$49,$FB,$6F,$0C
	.dc.b	$7F,$CB,$77,$E3,$7E,$17,$E3,$16,$BD,$87,$A3,$0F,$66,$1F,$8D,$7F
	.dc.b	$9D,$F8,$18,$C3,$84,$FE,$4E,$3F,$2C,$4A,$FC,$5B,$F9,$9E,$FF,$EC
	.dc.b	$3E,$97,$16,$CF,$C6,$7F,$A3,$99,$7F,$9F,$36,$BF,$4E,$F3,$8F,$CB
	.dc.b	$B6,$33,$18,$23,$82,$1F,$AE,$8E,$0F,$EC,$3B,$BF,$5C,$49,$FD,$1B
	.dc.b	$D7,$3F,$93,$5F,$DE,$BD,$2F,$24,$62,$BD,$47,$F5,$5A,$CB,$FC,$03
	.dc.b	$93,$F7,$44,$7F,$5E,$FE,$F7,$D9,$FF,$02,$FA,$3F,$35,$B3,$F8,$A7
	.dc.b	$FB,$5C,$E6,$3F,$B1,$37,$2F,$E1,$73,$CE,$7F,$76,$D8,$CC,$60,$8E
	.dc.b	$08,$7F,$12,$8E,$0F,$F1,$4F,$93,$FE,$24,$49,$FE,$0E,$F5,$C7,$F1
	.dc.b	$B5,$FE,$1E,$F4,$FC,$91,$8A,$F5,$1F,$DE,$67,$2F,$F2,$4F,$93,$FE
	.dc.b	$3C,$47,$F8,$9B,$FC,$87,$DC,$FF,$14,$FA,$9F,$91,$6C,$FE,$B9,$FF
	.dc.b	$07,$E3,$98,$FF,$03,$37,$2F,$E5,$F3,$CE,$7F,$8F,$B6,$33,$18,$23
	.dc.b	$82,$1F,$CD,$A3,$83,$FC,$E3,$DF,$FE,$6C,$49,$FE,$56,$F5,$CB,$F6
	.dc.b	$35,$FE,$42,$F5,$7C,$91,$8A,$F5,$1F,$E1,$DC,$B2,$FF,$48,$F9,$3F
	.dc.b	$E8,$44,$7F,$9B,$BF,$D1,$7E,$4F,$FC,$9B,$EB,$FE,$D5,$B3,$FC,$49
	.dc.b	$FF,$19,$E3,$98,$FF,$15,$37,$3F,$EA,$73,$CE,$7F,$A1,$B6,$33,$18
	.dc.b	$23,$82,$1F,$D6,$A3,$83,$FD,$73,$E4,$FF,$AD,$12,$7F,$A7,$BD,$72
	.dc.b	$FE,$2B,$5F,$E6,$6F,$63,$C9,$18,$AF,$51,$FE,$41,$D7,$2F,$F6,$CF
	.dc.b	$93,$FE,$CC,$47,$FA,$DB,$FD,$A7,$CA,$FF,$38,$FB,$1F,$C6,$5B,$3F
	.dc.b	$CD,$9F,$F2,$BE,$D9,$8F,$F2,$73,$74,$FE,$EF,$3C,$E7,$FB,$3B,$63
	.dc.b	$31,$82,$38,$21,$FD,$FA,$38,$3F,$E0,$0F,$8C,$FE,$FC,$49,$FE,$E6
	.dc.b	$F5,$CB,$F9,$CD,$7F,$A2,$BD,$5F,$24,$62,$BD,$47,$F9,$97,$5C,$BF
	.dc.b	$E1,$0C,$3F,$C1,$44,$7F,$BF,$A7,$FC,$0B,$F9,$03,$E4,$FF,$D2,$FE
	.dc.b	$CF,$F4,$28,$C4,$FF,$BD,$B3,$7B,$6A,$D5,$FF,$A0,$38,$AD,$FF,$9D
	.dc.b	$DB,$88,$85,$C7,$AB,$83,$15,$E5,$70,$23,$6E,$DA,$F4,$6B,$CA,$BD
	.dc.b	$B3,$CE,$38,$35,$E0,$D7,$81,$E0,$9E,$6F,$87,$E4,$F4,$DA,$F9,$31
	.dc.b	$89,$36,$BE,$0E,$9E,$BC,$FE,$4F,$B5,$8C,$3C,$27,$C5,$9F,$36,$1F
	.dc.b	$11,$F8,$5E,$BF,$F9,$E3,$8F,$86,$7F,$27,$E4,$FC,$04,$C3,$ED,$BE
	.dc.b	$3F,$93,$FF,$68,$FB,$9F,$6D,$B3,$DC,$F9,$0F,$FA,$E3,$93,$51,$FE
	.dc.b	$AF,$6D,$FC,$39,$E9,$3F,$88,$6C,$66,$30,$47,$04,$3E,$C4,$70,$7E
	.dc.b	$D7,$C9,$FC,$92,$4F,$E5,$1E,$B9,$FC,$4A,$FF,$75,$FB,$BE,$48,$C5
	.dc.b	$7A,$EF,$F1,$95,$FE,$D2,$DF,$53,$F8,$5F,$92,$FA,$C4,$7F,$42,$FD
	.dc.b	$9F,$93,$FF,$7B,$FB,$FF,$A1,$6C,$F5,$C3,$F2,$51,$FE,$DE,$75,$1F
	.dc.b	$ED,$56,$9F,$89,$8E,$91,$FA,$E6,$C6,$63,$04,$70,$43,$F3,$D1,$C1
	.dc.b	$FD,$07,$C9,$7E,$3C,$93,$FB,$A7,$AE,$5F,$8A,$AF,$F2,$B6,$3F,$B8
	.dc.b	$F7,$FC,$B3,$9B,$F6,$71,$AF,$FC,$B9,$C6,$FC,$57,$F2,$55,$FA,$D9
	.dc.b	$FE,$09,$F2,$7F,$9F,$24,$7F,$0A,$7F,$45,$F2,$7E,$0E,$CF,$F5,$3C
	.dc.b	$F6,$8C,$7B,$DF,$C3,$5E,$F8,$7E,$A2,$74,$7D,$BC,$7C,$5D,$7B,$8F
	.dc.b	$E2,$F5,$D9,$A4,$E0,$FD,$C4,$65,$B1,$97,$F8,$44,$65,$B1,$FC,$8B
	.dc.b	$E4,$FF,$74,$47,$07,$5D,$BF,$6B,$5E,$D8,$F2,$57,$FB,$6B,$64,$E2
	.dc.b	$BF,$73,$93,$A7,$18,$7F,$17,$F9,$3F,$E1,$04,$7F,$99,$BF,$C2,$FE
	.dc.b	$47,$EC,$FE,$C3,$FE,$C9,$C3,$F8,$2C,$63,$E0,$3F,$9B,$3F,$2B,$A6
	.dc.b	$E1,$F2,$72,$46,$35,$FE,$39,$5D,$6B,$FA,$0B,$63,$31,$82,$38,$21
	.dc.b	$FC,$C6,$38,$3F,$CC,$FE,$4F,$F9,$11,$27,$FA,$6B,$D7,$1F,$4C,$7F
	.dc.b	$4B,$CB,$FD,$A1,$18,$AF,$95,$7E,$78,$F9,$3D,$96,$FF,$9F,$7C,$9F
	.dc.b	$F2,$F2,$3F,$D6,$9F,$E6,$5F,$27,$F2,$3E,$0B,$FA,$D3,$67,$F3,$EF
	.dc.b	$E4,$3B,$B3,$1F,$1D,$B9,$7F,$41,$9E,$73,$FD,$99,$B1,$98,$C1,$1C
	.dc.b	$10,$FE,$AD,$1C,$1F,$EB,$1F,$27,$FD,$1C,$93,$FD,$C9,$EB,$97,$F4
	.dc.b	$2A,$FC,$0F,$C1,$F9,$23,$15,$EA,$3E,$F7,$66,$5F,$EC,$5F,$27,$FD
	.dc.b	$4C,$8F,$F7,$E7,$FA,$B7,$C9,$FE,$33,$E0,$7F,$BF,$36,$7F,$81,$BF
	.dc.b	$B4,$F0,$B3,$1F,$8B,$6E,$9F,$D9,$67,$9C,$FF,$82,$9B,$19,$8C,$11
	.dc.b	$C1,$0F,$EF,$51,$C1,$FE,$F9,$F2,$5F,$DA,$C9,$3F,$E1,$67,$AE,$5F
	.dc.b	$D9,$AB,$F4,$BF,$09,$E4,$8C,$57,$A8,$FD,$17,$85,$97,$FC,$0D,$F2
	.dc.b	$5F,$DD,$C8,$FF,$88,$9F,$EF,$B8,$3E,$47,$F6,$3F,$09,$FE,$E4,$8C
	.dc.b	$57,$E6,$63,$13,$FC,$6B,$BE,$CF,$EC,$5B,$A7,$F8,$6A,$F8,$B6,$BB
	.dc.b	$92,$BF,$C3,$F0,$9D,$BF,$E2,$56,$BF,$E2,$A6,$BF,$E2,$D6,$BC,$67
	.dc.b	$A3,$DE,$7A,$CB,$E4,$7D,$E4,$8C,$BC,$33,$FF,$14,$4F,$A1,$FF,$0D
	.dc.b	$75,$3E,$13,$DB,$1A,$C6,$E1,$FE,$3F,$F1,$7F,$A5,$78,$98,$CF,$CD
	.dc.b	$F1,$DF,$38,$95,$E4,$7E,$8F,$C3,$7F,$20,$F8,$5F,$A0,$D9,$FE,$E5
	.dc.b	$AA,$3F,$A4,$F8,$BA,$9F,$E4,$0D,$C3,$E8,$4F,$49,$F8,$2D,$8C,$C6
	.dc.b	$08,$E0,$87,$D1,$8E,$0F,$D3,$F8,$6F,$C3,$92,$7E,$6B,$D7,$5F,$93
	.dc.b	$5F,$CE,$BE,$17,$C9,$18,$AF,$5C,$FF,$62,$F0,$F3,$1F,$77,$E1,$BF
	.dc.b	$26,$47,$EC,$BF,$96,$F8,$6F,$E8,$BF,$0D,$F9,$56,$CF,$C6,$7F,$B6
	.dc.b	$F8,$F9,$8F,$E8,$8D,$D3,$F2,$93,$CE,$7F,$0A,$D8,$CC,$60,$8E,$08
	.dc.b	$7E,$56,$38,$3F,$97,$F8,$6F,$D5,$92,$7F,$22,$F5,$CF,$F1,$B5,$FD
	.dc.b	$7B,$E1,$BC,$91,$8A,$F5,$1F,$D6,$FC,$8C,$BF,$A1,$F8,$6F,$DB,$91
	.dc.b	$FC,$EB,$FB,$AF,$85,$FE,$D3,$F0,$DF,$B9,$6C,$FE,$29,$FF,$08,$F9
	.dc.b	$19,$8F,$ED,$0D,$D3,$F7,$13,$CE,$7F,$52,$D8,$CC,$60,$8E,$08,$7E
	.dc.b	$E6,$38,$3F,$BB,$F8,$5F,$E1,$E4,$9F,$DA,$BD,$72,$FD,$8D,$7F,$80
	.dc.b	$BE,$1B,$C9,$18,$AF,$51,$FD,$FF,$CB,$CB,$FC,$17,$E1,$7F,$8D,$91
	.dc.b	$FE,$02,$FF,$1E,$F8,$4F,$F0,$6F,$C3,$7F,$1D,$6C,$FE,$B9,$FF,$1A
	.dc.b	$F9,$99,$8F,$F0,$63,$74,$FE,$39,$3C,$E7,$F8,$6B,$63,$31,$82,$38
	.dc.b	$21,$FC,$76,$38,$3F,$C7,$FE,$13,$F9,$99,$27,$F8,$CB,$D7,$2F,$E2
	.dc.b	$B5,$FE,$26,$F8,$6F,$24,$62,$BD,$47,$F8,$8F,$CC,$CB,$FC,$A7,$E0
	.dc.b	$FF,$9F,$91,$FE,$4A,$DC,$7F,$9C,$35,$FE,$64,$D7,$94,$70,$23,$FC
	.dc.b	$FF,$E0,$BF,$A5,$92,$3F,$98,$B6,$35,$18,$23,$82,$1F,$D0,$63,$83
	.dc.b	$FD,$0F,$E0,$7F,$AB,$92,$7F,$9E,$B6,$3F,$A8,$35,$E1,$1A,$23,$A6
	.dc.b	$BF,$D3,$FE,$07,$FB,$19,$23,$FA,$4B,$63,$FA,$E3,$5E,$11,$A2,$3A
	.dc.b	$6B,$FD,$6F,$E0,$3F,$B7,$92,$3F,$AA,$B6,$3F,$B4,$35,$E1,$1A,$23
	.dc.b	$A6,$BF,$D9,$FD,$FF,$EF,$64,$8F,$EC,$2D,$8F,$EE,$8D,$78,$46,$88
	.dc.b	$E9,$AF,$F7,$37,$FC,$0E,$43,$FB,$68,$3F,$BD,$FE,$45,$FE,$79,$50
	.dc.b	$F8,$27,$9C,$F8,$7E,$12,$DE,$AC,$39,$3C,$6B,$86,$1F,$ED,$59,$CF
	.dc.b	$9B,$FE,$B9,$F2,$B9,$E1,$CA,$73,$EF,$7F,$83,$6F,$87,$BF,$DB,$2F
	.dc.b	$29,$DD,$C3,$E8,$A4,$62,$35,$3E,$E8,$C4,$EA,$30,$49,$EA,$73,$9D
	.dc.b	$92,$35,$1A,$97,$C6,$74,$24,$FC,$C7,$84,$FC,$78,$F4,$CF,$C6,$D7
	.dc.b	$E6,$FC,$47,$C2,$9C,$E1,$6F,$C5,$1F,$08,$FF,$54,$7A,$F9,$8F,$69
	.dc.b	$D4,$F8,$5E,$FF,$B8,$93,$F7,$9F,$83,$EF,$7F,$AE,$3E,$1F,$96,$68
	.dc.b	$C5,$7D,$77,$FD,$DF,$F8,$B7,$F8,$B6,$3D,$31,$8F,$5F,$8B,$D2,$3F
	.dc.b	$16,$DC,$7E,$59,$F8,$57,$0F,$C9,$6B,$D2,$38,$91,$E3,$7E,$58,$FC
	.dc.b	$BF,$BF,$F6,$88,$E0,$E1,$C3,$F8,$BB,$FF,$6A,$FC,$47,$9A,$7D,$84
	.dc.b	$C3,$E4,$72,$33,$FF,$28,$4E,$70,$FC,$A7,$BF,$F8,$82,$47,$EA,$1F
	.dc.b	$C5,$FB,$DF,$EE,$2E,$3F,$96,$D9,$FD,$03,$F2,$BD,$AC,$C7,$EB,$AF
	.dc.b	$F5,$78,$F8,$67,$39,$7C,$0B,$EB,$7F,$B6,$6C,$66,$30,$47,$04,$3F
	.dc.b	$2D,$1C,$1F,$CC,$7B,$9F,$A2,$24,$FF,$02,$7A,$E1,$FB,$3A,$FF,$7E
	.dc.b	$D7,$92,$31,$5E,$C3,$D4,$7F,$9A,$F3,$FF,$30,$39,$9F,$CA,$1F,$C1
	.dc.b	$7D,$BF,$D8,$92,$3F,$88,$BF,$B4,$F6,$7F,$41,$A3,$F6,$17,$C5,$7E
	.dc.b	$5A,$73,$18,$BF,$E1,$91,$F9,$6F,$73,$51,$E4,$C5,$BF,$DD,$9C,$C7
	.dc.b	$E3,$2B,$C1,$5F,$C8,$9B,$19,$8C,$11,$C1,$0F,$E0,$51,$C1,$FE,$09
	.dc.b	$EC,$FF,$0E,$24,$FF,$2E,$7A,$EB,$FC,$7E,$BD,$7F,$13,$E4,$8C,$57
	.dc.b	$B0,$F5,$63,$F9,$5F,$6F,$51,$FC,$2B,$D8,$FE,$36,$47,$F9,$DB,$FC
	.dc.b	$7B,$D6,$F8,$FF,$11,$FB,$E6,$CF,$F3,$37,$F7,$BE,$D6,$63,$BF,$1A
	.dc.b	$31,$B7,$EC,$AB,$A5,$7F,$48,$6C,$66,$30,$47,$04,$3F,$94,$47,$07
	.dc.b	$F9,$57,$AB,$FC,$D0,$93,$FD,$51,$EB,$A7,$F4,$5A,$FC,$0F,$C4,$79
	.dc.b	$23,$15,$EB,$3D,$3E,$E6,$5F,$E5,$FE,$AF,$F4,$02,$3F,$D8,$1F,$E8
	.dc.b	$7E,$97,$E3,$3E,$23,$F9,$1B,$67,$FA,$C3,$FC,$8B,$DC,$CC,$7C,$66
	.dc.b	$E9,$FC,$5A,$79,$CF,$F6,$C6,$C6,$63,$04,$70,$43,$FA,$64,$70,$7F
	.dc.b	$A7,$79,$BF,$D6,$09,$3F,$DE,$1E,$B9,$7F,$69,$AF,$D2,$FC,$47,$92
	.dc.b	$31,$5E,$A3,$EE,$FB,$79,$7F,$A9,$F9,$7F,$D9,$08,$FF,$81,$1F,$EC
	.dc.b	$FE,$47,$EC,$3E,$23,$FA,$3B,$67,$FB,$E3,$FD,$1B,$DD,$CC,$7E,$29
	.dc.b	$BA,$7F,$3A,$9E,$73,$FE,$10,$6C,$66,$30,$47,$04,$3F,$B8,$47,$07
	.dc.b	$FB,$97,$8D,$FD,$F0,$93,$FE,$1C,$7A,$E5,$FE,$0D,$AF,$E1,$3F,$11
	.dc.b	$E4,$8C,$57,$A8,$FD,$0F,$BB,$97,$FB,$BF,$87,$FE,$08,$23,$FE,$28
	.dc.b	$7F,$C1,$78,$FF,$14,$F8,$8F,$ED,$6D,$9F,$F1,$03,$FD,$AB,$DC,$CC
	.dc.b	$7E,$B9,$B1,$FD,$78,$E7,$1F,$E3,$46,$99,$70,$47,$04,$3F,$C2,$87
	.dc.b	$03,$FC,$37,$F9,$27,$F2,$77,$FB,$9B,$5C,$F8,$DF,$E2,$78,$C4,$7F
	.dc.b	$85,$8F,$2E,$5F,$E4,$0C,$7F,$9A,$7C,$4E,$30,$D1,$EE,$3F,$84,$FB
	.dc.b	$F9,$BF,$F2,$5E,$0F,$13,$B1,$2B,$FC,$0B,$29,$D4,$61,$78,$3F,$D0
	.dc.b	$3E,$23,$FC,$41,$18,$9F,$E7,$D9,$7F,$47,$9C,$D7,$F8,$63,$DF,$B5
	.dc.b	$FC,$71,$B8,$7C,$29,$F8,$04,$9F,$03,$63,$51,$82,$38,$21,$F3,$A3
	.dc.b	$E5,$77,$C9,$F5,$36,$78,$6B,$FE,$68,$CB,$D5,$1F,$E3,$9C,$BF,$B0
	.dc.b	$7C,$47,$9A,$B5,$8D,$C7,$FB,$77,$C0,$5C,$3E,$6F,$7F,$D9,$24,$FC
	.dc.b	$97,$EE,$76,$7F,$68,$F8,$7F,$92,$D9,$FE,$CD,$5F,$46,$3F,$C7,$FF
	.dc.b	$03,$A9,$FE,$86,$D9,$FC,$4C,$F4,$9F,$AC,$D8,$CC,$60,$8E,$08,$7E
	.dc.b	$52,$38,$3F,$96,$EC,$FC,$91,$27,$F0,$4F,$5C,$FE,$45,$7F,$80,$BE
	.dc.b	$1F,$C9,$18,$AF,$5E,$EA,$FF,$2C,$FC,$16,$A7,$FD,$13,$2E,$BF,$D0
	.dc.b	$12,$3F,$20,$FE,$8F,$AB,$FC,$1F,$F0,$FF,$90,$6C,$F6,$7F,$CC,$BF
	.dc.b	$07,$98,$FE,$D2,$DD,$7F,$5F,$3C,$E7,$F3,$8D,$8C,$C6,$08,$E0,$87
	.dc.b	$EE,$A3,$83,$FB,$CE,$AF,$DC,$12,$7F,$50,$F5,$D3,$F1,$D5,$FE,$28
	.dc.b	$F8,$7F,$24,$62,$BD,$47,$F8,$03,$E1,$32,$FC,$6E,$AF,$E0,$84,$7F
	.dc.b	$68,$FF,$07,$E7,$FF,$19,$FC,$3F,$ED,$1B,$3F,$49,$FF,$44,$B9,$3F
	.dc.b	$C1,$A5,$CF,$F8,$9C,$73,$8F,$E0,$0D,$8C,$C6,$08,$E0,$87,$F1,$E8
	.dc.b	$E0,$FF,$20,$E5,$FE,$38,$49,$FE,$18,$F5,$C7,$F6,$55,$FE,$4C,$F8
	.dc.b	$7F,$24,$62,$BD,$47,$F8,$CB,$0F,$F6,$EE,$1F,$C5,$EF,$F8,$9F,$27
	.dc.b	$F2,$92,$57,$F1,$97,$F9,$67,$07,$F9,$63,$E1,$FF,$8C,$B6,$7F,$2E
	.dc.b	$FF,$AF,$FD,$36,$3F,$EC,$C9,$C9,$E1,$F8,$4E,$2F,$F2,$5A,$EA,$D2
	.dc.b	$65,$FC,$FA,$32,$D8,$CB,$FD,$12,$32,$D8,$FE,$61,$C1,$FD,$0C,$8E
	.dc.b	$0E,$BA,$7F,$19,$AF,$F3,$9F,$C3,$79,$2F,$FC,$FF,$CC,$9C,$57,$FA
	.dc.b	$7B,$27,$C1,$E3,$1F,$D9,$C7,$F5,$12,$3F,$D0,$5F,$EA,$9C,$1F,$E8
	.dc.b	$0F,$84,$FE,$82,$D9,$FD,$FB,$F2,$67,$FD,$8D,$97,$8A,$31,$F0,$3C
	.dc.b	$2B,$FA,$64,$74,$69,$27,$FB,$2C,$65,$B1,$97,$FB,$5C,$65,$B1,$FD
	.dc.b	$5F,$83,$FB,$51,$1C,$1D,$72,$FE,$81,$5F,$E9,$EE,$EF,$F7,$65,$FF
	.dc.b	$66,$8F,$33,$E9,$AF,$C8,$7C,$16,$2F,$FA,$46,$FF,$DD,$C9,$3F,$D9
	.dc.b	$DF,$EF,$5A,$75,$F8,$9F,$EC,$ED,$9F,$E4,$AF,$C7,$F8,$3C,$C7,$E7
	.dc.b	$5B,$8F,$F8,$1A,$79,$CF,$F7,$36,$C6,$63,$04,$70,$43,$FC,$23,$1C
	.dc.b	$11,$07,$E4,$AF,$86,$57,$F8,$12,$30,$D8,$CC,$F1,$39,$CE,$9F,$0E
	.dc.b	$5F,$DB,$6F,$E7,$7C,$2F,$B2,$31,$87,$C2,$3F,$79,$F0,$99,$7C,$98
	.dc.b	$33,$EF,$8F,$78,$C4,$F9,$67,$33,$C1,$AF,$93,$0F,$BD,$F1,$5E,$8C
	.dc.b	$79,$E0,$9C,$E1,$FD,$CA,$30,$FF,$0F,$F8,$6E,$0F,$92,$7F,$3D,$18
	.dc.b	$C3,$B4,$FC,$79,$F0,$C7,$CC,$6D,$74,$8F,$90,$66,$78,$11,$CC,$6D
	.dc.b	$F8,$13,$C8,$93,$B8,$F8,$98,$76,$24,$7D,$97,$70,$C3,$E0,$D7,$E8
	.dc.b	$7E,$13,$D5,$3E,$33,$57,$EF,$3E,$24,$FE,$F7,$E1,$B7,$8F,$AF,$0F
	.dc.b	$97,$87,$B8,$97,$F8,$97,$E0,$FC,$1F,$EB,$CE,$25,$8F,$A0,$FF,$09
	.dc.b	$F8,$4C,$C7,$F3,$F6,$E5,$F1,$27,$9C,$FE,$55,$B1,$98,$C1,$1C,$10
	.dc.b	$FC,$54,$70,$7F,$19,$F0,$5F,$58,$93,$FA,$17,$AE,$1F,$88,$AF,$E1
	.dc.b	$11,$E4,$8C,$4F,$A8,$FE,$AD,$F0,$B9,$7F,$37,$F0,$3F,$82,$23,$FA
	.dc.b	$D7,$F0,$BF,$01,$FC,$4D,$F9,$AD,$8F,$CA,$3F,$CB,$7E,$17,$31,$FD
	.dc.b	$8D,$B9,$7E,$22,$79,$CF,$EE,$5B,$19,$8C,$11,$C1,$0F,$D7,$47,$07
	.dc.b	$F6,$1E,$FF,$E6,$89,$3F,$C1,$5E,$B8,$FE,$B2,$BF,$96,$1E,$47,$13
	.dc.b	$EA,$3F,$BD,$7C,$1E,$5F,$DF,$FB,$FF,$A6,$23,$FC,$45,$B8,$7E,$91
	.dc.b	$AF,$F0,$36,$BC,$A3,$81,$1F,$D5,$FB,$BF,$B2,$24,$7F,$1D,$6C,$6A
	.dc.b	$30,$47,$04,$3F,$89,$47,$07,$F8,$A7,$BB,$FB,$E2,$4F,$F2,$97,$F8
	.dc.b	$9A,$F9,$76,$9F,$C2,$63,$1B,$59,$CC,$7F,$36,$CF,$F9,$7B,$A3,$FC
	.dc.b	$45,$F0,$9F,$DB,$CF,$EE,$1F,$0B,$63,$F8,$D7,$BF,$FC,$40,$8F,$F3
	.dc.b	$B7,$C7,$7D,$B3,$FD,$F9,$FD,$01,$63,$E3,$CB,$C8,$F5,$9D,$F9,$9F
	.dc.b	$E2,$6C,$BC,$38,$F9,$0F,$E5,$5E,$F7,$F2,$02,$3F,$D2,$DB,$5F,$C7
	.dc.b	$9A,$FF,$3D,$6B,$FD,$09,$AF,$F4,$E7,$F9,$5F,$B7,$27,$FA,$B3,$67
	.dc.b	$F9,$4B,$5F,$E9,$0D,$7F,$A6,$B5,$FE,$B2,$D7,$FB,$1F,$B7,$FC,$E0
	.dc.b	$93,$FD,$8D,$B1,$FC,$D9,$AF,$F5,$56,$BF,$D6,$9A,$FF,$66,$7F,$A1
	.dc.b	$FB,$32,$7F,$B7,$36,$7F,$A0,$B5,$FE,$C0,$D7,$FB,$2B,$5F,$EE,$2D
	.dc.b	$7F,$BD,$FB,$1F,$D4,$09,$3F,$DE,$DF,$EA,$7E,$B7,$F4,$0F,$2B,$F9
	.dc.b	$D4,$FF,$70,$8C,$78,$1F,$CF,$9F,$EB,$DC,$6F,$79,$E1,$E6,$FF,$2C
	.dc.b	$7E,$5E,$32,$FF,$6B,$AE,$EF,$F8,$2D,$B9,$7F,$5E,$6B,$FD,$F5,$AF
	.dc.b	$8E,$38,$11,$E0,$FF,$85,$4F,$ED,$A4,$30,$6E,$FF,$BF,$CF,$F5,$EF
	.dc.b	$79,$E5,$FA,$6F,$F4,$27,$F8,$8A,$71,$5F,$E0,$7C,$1E,$26,$31,$FF
	.dc.b	$14,$62,$F8,$9F,$EF,$ED,$8F,$EF,$B8,$A3,$0F,$F8,$01,$AF,$BE,$3E
	.dc.b	$31,$F2,$9F,$F1,$AC,$79,$A3,$31,$E6,$8F,$F4,$A3,$FB,$2F,$41,$F0
	.dc.b	$BE,$DF,$5B,$FC,$E4,$DB,$F1,$DF,$0B,$FF,$1C,$35,$F5,$D7,$D4,$CF
	.dc.b	$E3,$D7,$8E,$7D,$5A,$FC,$17,$CB,$5F,$00,$93,$D9,$F8,$7E,$9C,$8F
	.dc.b	$0B,$67,$E1,$57,$37,$AC,$FB,$AF,$FC,$7F,$87,$B9,$B8,$7F,$A3,$7E
	.dc.b	$1B,$D9,$7F,$0F,$D4,$FA,$24,$C3,$DE,$FD,$4C,$7B,$63,$F2,$EB,$EF
	.dc.b	$F9,$6E,$1F,$C2,$57,$E7,$B8,$5E,$06,$32,$FA,$9A,$FE,$86,$31,$8F
	.dc.b	$AA,$7E,$6D,$F7,$69,$23,$44,$3E,$8C,$6D,$C1,$F8,$C7,$F1,$7E,$8C
	.dc.b	$9F,$B2,$D9,$E7,$7F,$1A,$BF,$39,$DD,$E4,$8C,$5F,$D9,$D3,$E0,$47
	.dc.b	$E6,$FB,$75,$3F,$9A,$F3,$FF,$32,$47,$F0,$EF,$96,$79,$E3,$F9,$1C
	.dc.b	$38,$BF,$96,$E7,$FF,$60,$7C,$37,$EB,$5B,$87,$85,$FD,$29,$1F,$C9
	.dc.b	$BF,$A8,$AE,$75,$FA,$76,$E5,$FC,$2A,$3F,$51,$3C,$5B,$5F,$9C,$E8
	.dc.b	$FD,$89,$1C,$1F,$A0,$BE,$67,$4A,$F3,$47,$EB,$FC,$0F,$F7,$FF,$C2
	.dc.b	$E3,$0F,$D7,$79,$DF,$BC,$25,$7E,$A9,$F4,$3F,$37,$7F,$A2,$D3,$E8
	.dc.b	$63,$F8,$BF,$86,$CD,$FF,$6E,$DB,$F9,$54,$62,$3F,$67,$3D,$1F,$E0
	.dc.b	$CD,$73,$5E,$B2,$3A,$6B,$A3,$07,$F0,$FF,$13,$F8,$A9,$23,$F8,$1B
	.dc.b	$DE,$FF,$5F,$5F,$E8,$2F,$C9,$9F,$0F,$E8,$9C,$DF,$B0,$FD,$1F,$37
	.dc.b	$F1,$D7,$18,$FF,$15,$F3,$3F,$91,$92,$3F,$88,$BD,$DF,$E5,$38,$F7
	.dc.b	$C3,$B4,$7F,$2C,$E7,$FF,$56,$70,$3B,$DB,$7F,$39,$9C,$C6,$2B,$F3
	.dc.b	$D1,$FC,$C8,$8F,$F2,$07,$C1,$97,$F2,$CC,$7C,$13,$E8,$D3,$F8,$7D
	.dc.b	$70,$AF,$E6,$CD,$7F,$93,$F3,$FF,$3E,$25,$7F,$2C,$7A,$C7,$E6,$63
	.dc.b	$F9,$46,$DF,$29,$EC,$7B,$4F,$39,$F9,$1C,$7E,$54,$FF,$3D,$BF,$EA
	.dc.b	$DD,$FF,$D3,$09,$5F,$CE,$9E,$FC,$DF,$E3,$68,$EF,$B7,$CA,$BF,$E9
	.dc.b	$3D,$BF,$DE,$3F,$64,$FE,$F1,$5F,$C0,$E3,$1C,$5F,$D8,$DF,$EB,$24
	.dc.b	$8F,$E8,$EF,$6D,$FE,$3D,$FF,$01,$9E,$46,$88,$FF,$5D,$9F,$EC,$7D
	.dc.b	$32,$7F,$A8,$B6,$7F,$DC,$3B,$7C,$57,$A1,$FD,$C9,$C7,$07,$E4,$23
	.dc.b	$FB,$05,$F9,$EF,$F9,$9E,$4F,$0F,$15,$FD,$97,$A3,$FB,$99,$2B,$FB
	.dc.b	$0B,$FE,$F4,$C7,$AC,$7E,$48,$F9,$27,$D6,$9F,$EE,$3C,$7F,$8E,$7F
	.dc.b	$C1,$0D,$8F,$E5,$8F,$F8,$00,$8F,$F6,$E7,$B6,$DF,$4E,$55,$E0,$7F
	.dc.b	$25,$5C,$8E,$04,$7F,$C0,$D5,$FE,$0C,$DE,$4F,$F7,$A6,$CF,$CA,$BE
	.dc.b	$D7,$D6,$3F,$24,$7F,$81,$EF,$D3,$7F,$C7,$FF,$1A,$D6,$1F,$A9,$AF
	.dc.b	$F0,$61,$FE,$1A,$24,$FF,$82,$1F,$5B,$F1,$5F,$5E,$1E,$03,$F1,$DF
	.dc.b	$09,$F8,$6D,$95,$FE,$99,$D5,$3F,$8A,$B1,$0E,$73,$FE,$28,$CF,$F0
	.dc.b	$91,$93,$27,$BE,$35,$87,$92,$30,$43,$84,$F2,$7B,$35,$F0,$B5,$FA
	.dc.b	$BF,$13,$F8,$52,$57,$8D,$ED,$8F,$B2,$78,$DD,$8C,$43,$EB,$5E,$DF
	.dc.b	$C3,$C3,$C3,$76,$96,$CD,$F0,$B9,$9F,$C3,$FC,$57,$E5,$C9,$87,$C5
	.dc.b	$7F,$35,$CD,$E1,$8F,$95,$FA,$43,$DE,$FE,$75,$39,$BF,$9F,$39,$AF
	.dc.b	$E2,$B3,$FE,$7C,$FD,$8F,$E9,$C6,$0F,$F0,$FB,$8A,$FD,$A4,$F8,$27
	.dc.b	$EB,$36,$3F,$76,$D7,$E8,$B5,$FA,$2D,$7E,$8B,$5F,$D2,$F3,$7E,$D4
	.dc.b	$93,$F8,$37,$B6,$9F,$3B,$87,$F8,$DD,$7F,$07,$F8,$5F,$7D,$FA,$63
	.dc.b	$CE,$FF,$8E,$3C,$2F,$D2,$57,$0B,$E5,$F1,$7F,$C0,$C9,$5F,$91,$7B
	.dc.b	$BE,$83,$F9,$DE,$1D,$B0,$FE,$2D,$F1,$9F,$D5,$BA,$1C,$DF,$DA,$63
	.dc.b	$0D,$BF,$CE,$47,$F1,$12,$3F,$28,$F4,$63,$ED,$C3,$B3,$EF,$9D,$E1
	.dc.b	$FC,$A1,$AF,$E7,$BE,$33,$F8,$E1,$2F,$F5,$4F,$AA,$3E,$7C,$7A,$8E
	.dc.b	$E7,$E9,$27,$CD,$1F,$EF,$CC,$5C,$58,$C3,$E8,$7C,$67,$F2,$92,$61
	.dc.b	$FB,$77,$B5,$FF,$1E,$E0,$FF,$05,$67,$EA,$AF,$D8,$F4,$FF,$B2,$B8
	.dc.b	$1B,$FF,$83,$E3,$0D,$9F,$E0,$51,$FC,$DC,$8F,$F0,$5A,$E9,$87,$EE
	.dc.b	$9A,$FF,$2E,$9D,$BF,$D3,$9A,$FF,$04,$E5,$FE,$84,$4A,$FE,$20,$FF
	.dc.b	$45,$AE,$7C,$26,$DF,$96,$FF,$C7,$58,$B7,$C6,$BF,$95,$F8,$BF,$E9
	.dc.b	$C4,$CF,$F8,$E3,$D5,$FA,$37,$E9,$6B,$FD,$1B,$E3,$3F,$19,$B7,$F8
	.dc.b	$CD,$B3,$FE,$0C,$7F,$AD,$11,$FE,$50,$F5,$9F,$D0,$9F,$A5,$8E,$BA
	.dc.b	$27,$84,$FF,$72,$6B,$FC,$A3,$E3,$3F,$B3,$12,$BF,$9A,$3D,$71,$FD
	.dc.b	$96,$1F,$7A,$7A,$DF,$9D,$FE,$9D,$CB,$70,$FE,$F9,$F1,$7F,$DC,$49
	.dc.b	$3F,$CF,$DE,$8F,$EB,$E3,$D0,$7A,$B0,$FD,$24,$7E,$EF,$3F,$04,$FD
	.dc.b	$26,$BF,$D0,$37,$F8,$4D,$9F,$C4,$11,$FF,$08,$C6,$9F,$E8,$AD,$7E
	.dc.b	$A1,$A8,$FE,$A9,$E3,$7F,$82,$48,$E0,$E9,$7E,$69,$FE,$7B,$1F,$D7
	.dc.b	$7F,$49,$FE,$86,$FE,$C6,$3D,$B3,$B3,$FA,$57,$07,$D5,$9D,$47,$F1
	.dc.b	$FF,$8B,$FF,$0C,$12,$3F,$B3,$3F,$22,$BC,$F3,$FD,$4A,$FB,$C7,$6A
	.dc.b	$FE,$C3,$F1,$9F,$C5,$F9,$7F,$CF,$CD,$8F,$F2,$BB,$FE,$25,$23,$FD
	.dc.b	$CD,$ED,$3F,$E1,$AC,$FF,$A6,$2B,$EF,$87,$AE,$FB,$9C,$49,$3F,$E4
	.dc.b	$0B,$FF,$19,$7C,$64,$9F,$F0,$03,$67,$E8,$1E,$93,$FC,$0F,$7D,$A7
	.dc.b	$EB,$9E,$82,$47,$F1,$9E,$4F,$6C,$7F,$9C,$BB,$BF,$C8,$04,$7F,$C1
	.dc.b	$CF,$B3,$97,$F7,$D1,$FE,$14,$CB,$E5,$78,$3F,$8E,$7F,$BE,$FE,$19
	.dc.b	$E1,$C7,$FB,$CF,$03,$83,$E1,$CE,$69,$39,$C3,$A1,$FE,$1F,$9F,$AB
	.dc.b	$7F,$E1,$D6,$B9,$C3,$FC,$16,$D9,$FC,$69,$F0,$1F,$C6,$9F,$12,$3C
	.dc.b	$D1,$B8,$FF,$14,$09,$EA,$FF,$8C,$1A,$E6,$7B,$B9,$F8,$DA,$F1,$6C
	.dc.b	$9A,$7C,$9E,$0F,$C1,$ED,$F1,$6B,$F0,$D8,$7C,$07,$C1,$87,$C6,$3F
	.dc.b	$A1,$70,$59,$D4,$F9,$FD,$CE,$E4,$8F,$5C,$65,$F8,$6F,$E8,$1F,$D7
	.dc.b	$57,$C2,$8F,$77,$C6,$FF,$75,$F0,$3E,$AB,$6B,$BE,$88,$F6,$92,$3B
	.dc.b	$C6,$5E,$F5,$EC,$3E,$89,$FA,$1B,$F0,$3F,$3E,$B9,$3F,$4A,$B9,$36
	.dc.b	$3E,$5F,$C6,$CB,$EE,$D9,$FD,$0E,$FF,$AB,$7E,$F4,$FF,$04,$3F,$15
	.dc.b	$5E,$C7,$FC,$F9,$DF,$67,$CB,$83,$E3,$7E,$C9,$2B,$DA,$FD,$97,$E8
	.dc.b	$ED,$F6,$78,$FF,$47,$7C,$BE,$37,$FC,$95,$9B,$F0,$67,$07,$E8,$27
	.dc.b	$31,$8C,$7E,$6C,$7E,$20,$8F,$CC,$7C,$17,$F2,$B1,$F0,$71,$73,$AD
	.dc.b	$12,$7E,$55,$7E,$4B,$E3,$A4,$FD,$66,$CF,$E8,$2F,$9C,$FB,$38,$7E
	.dc.b	$AE,$CC,$3B,$5F,$EA,$B4,$F6,$70,$61,$FD,$FF,$C6,$FE,$80,$98,$FE
	.dc.b	$11,$B5,$F7,$5A,$FE,$ED,$AF,$EF,$DA,$FE,$C5,$FD,$5F,$C7,$49,$FC
	.dc.b	$7B,$67,$F1,$0D,$7F,$83,$35,$FE,$14,$D7,$F7,$8D,$7F,$80,$7C,$77
	.dc.b	$EE,$09,$3F,$9C,$6C,$7E,$55,$AF,$F1,$36,$BF,$C5,$DA,$FF,$09,$7F
	.dc.b	$81,$FC,$74,$9F,$D3,$B6,$7F,$40,$D7,$F8,$F3,$5F,$E4,$4D,$7F,$8A
	.dc.b	$35,$FE,$31,$F1,$DF,$C4,$09,$3F,$B4,$6C,$7E,$B5,$AF,$F2,$B6,$BF
	.dc.b	$CB,$DA,$FF,$21,$7F,$8D,$FC,$74,$9F,$DF,$B6,$7F,$70,$D7,$F9,$B3
	.dc.b	$5F,$E7,$4D,$7F,$96,$35,$FE,$61,$F1,$BF,$CA,$09,$3F,$C3,$1F,$E5
	.dc.b	$BF,$74,$F8,$3F,$F5,$B4,$BE,$3D,$F0,$1F,$DB,$AB,$53,$FC,$AC,$D4
	.dc.b	$7F,$50,$F8,$FF,$DA,$FD,$63,$CE,$C5,$7F,$31,$E2,$4E,$63,$F9,$0D
	.dc.b	$7F,$3C,$23,$FC,$81,$F2,$E5,$E4,$CB,$C7,$3F,$CD,$B2,$F1,$E9,$FD
	.dc.b	$4A,$3C,$67,$43,$91,$1F,$E4,$57,$FD,$2B,$E4,$24,$FF,$2E,$6C,$FF
	.dc.b	$58,$AE,$B9,$7F,$BB,$EB,$AC,$FF,$30,$9C,$C7,$FB,$63,$0F,$2C,$FF
	.dc.b	$20,$C3,$E0,$BF,$BD,$C9,$E6,$63,$1F,$EF,$7F,$21,$FD,$70,$98,$FF
	.dc.b	$40,$7E,$0B,$F3,$F1,$EB,$3F,$EF,$EC,$BC,$38,$AB,$CB,$3F,$E0,$6F
	.dc.b	$90,$FF,$62,$7A,$5F,$E0,$46,$CF,$F5,$07,$FB,$59,$1F,$E9,$EF,$82
	.dc.b	$BE,$66,$1E,$09,$44,$C3,$C3,$1E,$03,$FC,$19,$5C,$E7,$FA,$83,$5F
	.dc.b	$F0,$77,$C5,$FF,$7A,$24,$FF,$5E,$7C,$18,$FC,$FC,$3C,$15,$DF,$7F
	.dc.b	$C1,$70,$FA,$F0,$F5,$1F,$DF,$7E,$19,$EB,$BD,$6D,$61,$FD,$4A,$BF
	.dc.b	$C1,$24,$3F,$B6,$9E,$03,$ED,$5F,$E7,$F0,$EF,$1F,$8F,$F9,$0F,$CE
	.dc.b	$FA,$3F,$A3,$6C,$64,$FF,$02,$EE,$8F,$D6,$92,$33,$5D,$30,$FC,$53
	.dc.b	$75,$FB,$95,$B9,$FC,$93,$5F,$A2,$7D,$1F,$8F,$FD,$C9,$2F,$C2,$FA
	.dc.b	$E3,$F0,$9B,$7C,$A5,$8F,$D8,$8C,$3E,$7A,$F5,$4F,$E5,$FD,$1F,$74
	.dc.b	$7E,$BF,$E4,$3F,$83,$91,$F7,$BD,$DF,$C1,$4F,$D7,$7F,$31,$AF,$F0
	.dc.b	$9F,$90,$FC,$B4,$FF,$BF,$7D,$7C,$57,$D9,$E1,$46,$27,$F6,$55,$FC
	.dc.b	$58,$95,$F2,$8E,$A7,$F0,$3D,$FE,$53,$F1,$B3,$4F,$0A,$E7,$3F,$5A
	.dc.b	$3F,$90,$FA,$72,$3E,$CB,$67,$F8,$15,$F4,$7D,$87,$E9,$32,$F3,$70
	.dc.b	$7E,$CF,$0F,$6B,$EE,$E3,$FC,$9F,$B3,$AC,$BC,$9F,$1F,$FC,$C0,$8F
	.dc.b	$E2,$1F,$95,$C3,$F5,$33,$F0,$67,$F6,$1F,$F1,$0F,$A4,$3F,$2D,$7F
	.dc.b	$CB,$5C,$3F,$0C,$F8,$D3,$EA,$6B,$F8,$A3,$A3,$87,$E4,$4F,$F0,$17
	.dc.b	$5B,$7F,$2E,$6C,$65,$DC,$FE,$83,$E3,$BF,$A5,$91,$C1,$E2,$D3,$F8
	.dc.b	$C5,$7F,$3B,$F8,$BF,$4D,$FE,$8A,$BD,$0F,$F8,$8F,$1F,$F2,$44,$62
	.dc.b	$FF,$90,$7C,$7F,$F5,$82,$47,$EC,$1F,$D0,$57,$B0,$FE,$33,$7D,$AB
	.dc.b	$F9,$47,$C7,$FD,$6B,$FE,$37,$BF,$F0,$8F,$0F,$FA,$F4,$E7,$04,$E6
	.dc.b	$BF,$A8,$4F,$F6,$82,$3F,$C0,$1F,$0E,$BF,$AD,$C7,$C1,$3F,$93,$9D
	.dc.b	$DF,$F4,$D6,$BF,$3B,$E3,$3F,$B9,$92,$7F,$85,$BD,$35,$F9,$11,$F0
	.dc.b	$CE,$87,$90,$FD,$57,$83,$EA,$7F,$B3,$7C,$7F,$F7,$E2,$3F,$C5,$9E
	.dc.b	$8F,$F4,$DC,$7C,$F8,$7F,$3D,$E5,$FC,$7F,$95,$FE,$01,$6D,$7F,$71
	.dc.b	$7F,$C1,$24,$7F,$90,$BD,$27,$E8,$C7,$9D,$E8,$E8,$8F,$F6,$F8,$FF
	.dc.b	$09,$D4,$9F,$E5,$6D,$9F,$CD,$1F,$87,$3A,$97,$7F,$EC,$95,$E7,$AF
	.dc.b	$D2,$79,$56,$BF,$BC,$7C,$7F,$F8,$80,$93,$FC,$E1,$F4,$5F,$F5,$1C
	.dc.b	$3E,$1D,$7F,$28,$F8,$FF,$C2,$70,$FF,$4E,$F5,$7F,$C4,$91,$8E,$DF
	.dc.b	$F4,$BE,$1F,$DD,$63,$FC,$5A,$4C,$BF,$A2,$BD,$9F,$F0,$46,$3D,$B8
	.dc.b	$D3,$BC,$BF,$C2,$4D,$7F,$67,$F1,$DF,$E3,$82,$5F,$F5,$07,$A4,$7F
	.dc.b	$40,$BE,$95,$E6,$AF,$E5,$DF,$05,$6B,$FC,$35,$F1,$FF,$E4,$22,$4F
	.dc.b	$F5,$A7,$9B,$FE,$43,$BF,$37,$C7,$FF,$12,$F3,$BF,$C8,$8D,$9F,$F1
	.dc.b	$A3,$FE,$4C,$23,$FD,$91,$E7,$DA,$AE,$73,$92,$61,$FE,$31,$AF,$F2
	.dc.b	$AF,$C7,$C9,$FE,$DA,$D9,$FC,$1E,$DF,$C3,$27,$F3,$47,$E3,$E3,$FA
	.dc.b	$FD,$79,$A3,$F9,$37,$9B,$67,$FC,$75,$F1,$FF,$E6,$22,$4F,$F7,$A7
	.dc.b	$FA,$BC,$7F,$27,$8F,$EA,$A7,$F9,$87,$17,$C7,$7F,$07,$9F,$F1,$97
	.dc.b	$8D,$FE,$6A,$9C,$D7,$F9,$5E,$7F,$DB,$3F,$27,$FE,$C1,$8C,$1F,$E6
	.dc.b	$57,$11,$FE,$4D,$3F,$CE,$A4,$8F,$F0,$6B,$E4,$3F,$C0,$6F,$91,$FF
	.dc.b	$31,$4F,$81,$67,$EA,$9D,$3F,$E5,$B9,$D3,$FE,$4B,$9E,$C7,$F9,$84
	.dc.b	$A7,$C1,$C3,$C0,$FF,$90,$AB,$B3,$97,$FA,$63,$98,$DB,$F1,$23,$E4
	.dc.b	$47,$C7,$7C,$67,$3A,$F1,$BF,$DE,$67,$FD,$20,$7C,$8F,$76,$C7,$F9
	.dc.b	$4A,$FD,$9F,$19,$DA,$0D,$23,$E4,$AF,$11,$F1,$B0,$F1,$65,$F5,$DF
	.dc.b	$27,$C6,$7E,$EF,$E2,$2D,$FF,$85,$5B,$7E,$E8,$4F,$84,$8F,$73,$DB
	.dc.b	$1F,$E7,$5C,$FD,$AF,$C9,$C3,$C1,$3E,$B3,$ED,$D7,$29,$EC,$D7,$F2
	.dc.b	$5F,$19,$EF,$24,$F7,$7E,$15,$74,$9F,$C0,$CF,$E1,$A3,$19,$FE,$0E
	.dc.b	$FD,$6F,$F8,$5F,$A5,$F1,$18,$BF,$C6,$FC,$67,$CD,$23,$ED,$7B,$BF
	.dc.b	$70,$F5,$63,$ED,$70,$F8,$8F,$C4,$7C,$67,$E8,$BD,$DF,$C1,$36,$78
	.dc.b	$CF,$FA,$6E,$3E,$E1,$23,$E6,$3E,$08,$FB,$51,$C0,$F9,$E7,$F0,$DD
	.dc.b	$FA,$BC,$09,$1F,$66,$7F,$11,$F1,$92,$7E,$C3,$67,$EF,$6B,$F2,$27
	.dc.b	$E5,$1E,$CD,$3F,$59,$7D,$AB,$DD,$0E,$37,$BD,$98,$FD,$97,$C5,$FE
	.dc.b	$5C,$98,$7E,$1D,$F6,$47,$ED,$73,$FC,$AC,$F7,$3F,$7F,$F1,$7F,$D3
	.dc.b	$FD,$9F,$D5,$36,$3E,$13,$FA,$32,$3F,$93,$7A,$BF,$AF,$9E,$B9,$7C
	.dc.b	$A3,$EB,$1F,$EC,$B9,$E2,$FE,$7D,$AF,$F2,$3F,$77,$F6,$04,$AF,$D0
	.dc.b	$3D,$A3,$F4,$15,$DB,$0F,$65,$79,$89,$3F,$85,$F1,$7F,$D7,$3E,$C7
	.dc.b	$B2,$7E,$E7,$B7,$FB,$D2,$3F,$AD,$7B,$3F,$7C,$FC,$17,$07,$D7,$BF
	.dc.b	$C9,$1F,$53,$77,$BB,$E3,$AF,$E1,$51,$87,$89,$FB,$9A,$FC,$75,$7F
	.dc.b	$41,$6B,$F8,$23,$F0,$B1,$B2,$38,$21,$C5,$FE,$09,$EB,$FF,$17,$23
	.dc.b	$83,$8E,$7F,$CD,$EB,$FC,$A5,$0F,$09,$E5,$7A,$70,$F6,$DE,$EB,$FA
	.dc.b	$57,$BB,$99,$43,$D4,$FE,$4C,$47,$F8,$B3,$F5,$E3,$F9,$2D,$7F,$BC
	.dc.b	$6F,$95,$3D,$1F,$F1,$0F,$E3,$8F,$87,$FE,$75,$18,$6C,$FF,$21,$8F
	.dc.b	$E6,$84,$7D,$A7,$C9,$BE,$CF,$D0,$C3,$F2,$2F,$73,$91,$1F,$E6,$13
	.dc.b	$FC,$FF,$CF,$93,$FC,$BD,$B3,$F4,$6B,$E2,$3F,$C4,$67,$E5,$63,$EC
	.dc.b	$BF,$3C,$FE,$47,$17,$C6,$E3,$0F,$ED,$1E,$6F,$F4,$D2,$63,$FC,$F5
	.dc.b	$F9,$0F,$F1,$0D,$BE,$AB,$EB,$9F,$EE,$1E,$2F,$FA,$2D,$78,$1F,$ED
	.dc.b	$49,$FE,$31,$39,$8C,$65,$F0,$A3,$FA,$E1,$27,$FA,$63,$E0,$E1,$FC
	.dc.b	$2E,$1E,$0A,$F9,$31,$FD,$B6,$78,$47,$F5,$76,$BF,$E1,$BF,$07,$FB
	.dc.b	$41,$2F,$FA,$C3,$D5,$FE,$76,$7C,$83,$D8,$7F,$84,$27,$B3,$FD,$03
	.dc.b	$1F,$6F,$BB,$8B,$FB,$7D,$BF,$DD,$49,$5F,$D9,$5E,$D1,$F5,$72,$F5
	.dc.b	$5F,$D8,$3F,$C3,$1D,$3F,$90,$F8,$BF,$F0,$43,$70,$FE,$80,$FF,$80
	.dc.b	$48,$FF,$71,$7A,$CF,$E3,$AF,$E2,$CF,$03,$44,$7F,$BE,$4F,$F8,$63
	.dc.b	$E7,$3F,$3D,$3F,$E0,$96,$CF,$F8,$79,$AF,$FB,$AD,$AF,$E7,$62,$3F
	.dc.b	$B2,$C3,$DB,$18,$8F,$DA,$D7,$B2,$3F,$B4,$F6,$38,$9F,$11,$9B,$FE
	.dc.b	$6B,$8F,$E5,$AF,$C9,$C7,$E7,$23,$FA,$E9,$FE,$0A,$FF,$8D,$F2,$F0
	.dc.b	$C6,$00,$A6,$9F,$93,$F6,$FF,$CB,$91,$8B,$FB,$DC,$69,$CD,$7C,$72
	.dc.b	$4F,$C7,$7E,$8B,$E1,$E6,$FE,$29,$7C,$F0,$F9,$55,$C1,$B5,$F3,$F3
	.dc.b	$FA,$44,$70,$7B,$63,$F8,$BB,$D4,$FC,$AB,$EC,$AE,$DB,$FF,$9E,$9B
	.dc.b	$3F,$E4,$9C,$9E,$DF,$AA,$3F,$C2,$32,$F7,$7F,$04,$48,$FC,$13,$70
	.dc.b	$FC,$13,$5F,$C1,$35,$F9,$A7,$CD,$70,$F9,$5F,$C7,$7C,$74,$9F,$C7
	.dc.b	$36,$7F,$1C,$D7,$F1,$CD,$7A,$D2,$31,$1F,$05,$AF,$DD,$F8,$AF,$CE
	.dc.b	$12,$BF,$38,$D8,$FC,$E3,$5F,$CE,$35,$EE,$D7,$E7,$3F,$A7,$F4,$24
	.dc.b	$FE,$9D,$B3,$FA,$76,$BF,$A7,$6B,$EA,$6B,$F6,$5A,$FE,$57,$E2,$BF
	.dc.b	$68,$49,$FD,$A3,$63,$F6,$8D,$7F,$68,$D7,$E3,$B5,$FC,$33,$FB,$FF
	.dc.b	$8A,$93,$FB,$F6,$CF,$EF,$DA,$FE,$FD,$AF,$D2,$6B,$F9,$16,$BF,$AD
	.dc.b	$F7,$BF,$86,$12,$7F,$86,$36,$3F,$86,$35,$FE,$18,$D7,$EF,$B5,$FC
	.dc.b	$F3,$FC,$5F,$DD,$93,$FC,$5D,$B3,$FC,$5D,$AF,$F1,$76,$BF,$8A,$6B
	.dc.b	$FA,$96,$BF,$C1,$7C,$FF,$E4,$84,$9F,$E4,$8D,$8F,$E4,$8D,$7F,$92
	.dc.b	$35,$FC,$BB,$5F,$DB,$3F,$CB,$FC,$D9,$3F,$CB,$DB,$3F,$CB,$DA,$FF
	.dc.b	$2F,$6B,$FA,$26,$BF,$C0,$5A,$FF,$1D,$F8,$AF,$E7,$84,$9F,$E7,$8D
	.dc.b	$8F,$E7,$8D,$7F,$9E,$35,$FD,$7B,$5F,$E1,$CF,$F4,$7F,$8A,$93,$FD
	.dc.b	$1D,$B3,$FD,$1D,$AF,$F4,$76,$BF,$BA,$6B,$FC,$65,$AF,$F3,$5F,$6F
	.dc.b	$FA,$A1,$27,$FA,$A3,$63,$FA,$A3,$5F,$EA,$8D,$7F,$83,$B5,$FE,$4C
	.dc.b	$FF,$5F,$F6,$64,$FF,$5F,$6C,$FF,$5F,$6B,$FD,$7D,$AF,$F1,$26,$BF
	.dc.b	$CC,$5A,$FF,$4D,$F3,$3F,$B6,$12,$7F,$B6,$36,$3F,$B6,$35,$FE,$D8
	.dc.b	$D7,$F8,$FB,$5F,$E7,$CF,$F7,$7F,$2A,$4F,$F7,$76,$CF,$F7,$76,$BF
	.dc.b	$DD,$DA,$FF,$2A,$6B,$FD,$25,$AF,$F6,$58,$FF,$02,$10,$FF,$02,$15
	.dc.b	$FF,$02,$35,$FF,$02,$35,$FE,$6E,$D7,$FA,$B3,$FE,$76,$7F,$A3,$9F
	.dc.b	$DD,$3F,$2B,$7C,$56,$63,$31,$89,$DC,$6E,$74,$FF,$50,$9D,$3A,$6B
	.dc.b	$CA,$36,$F1,$9E,$6A,$76,$F3,$9D,$B9,$6C,$6A,$39,$1D,$FE,$2B,$24
	.dc.b	$AE,$B2,$8C,$4E,$E7,$31,$88,$F7,$B5,$F6,$35,$F8,$DF,$15,$E2,$24
	.dc.b	$F6,$7D,$AF,$FA,$A7,$E0,$BF,$CC,$E7,$F9,$A3,$E0,$AC,$7F,$8C,$3C
	.dc.b	$4F,$ED,$4D,$E9,$FF,$4F,$F5,$FF,$A8,$1F,$AD,$DD,$F4,$F4,$FF,$3B
	.dc.b	$E7,$F5,$3E,$2B,$E1,$93,$0F,$AA,$78,$4F,$F3,$5F,$88,$D3,$C5,$18
	.dc.b	$7B,$3E,$13,$C9,$96,$2F,$FD,$EB,$83,$C9,$FA,$24,$AF,$9A,$FD,$4F
	.dc.b	$83,$FF,$42,$9E,$88,$FC,$23,$87,$FC,$83,$9F,$AF,$E2,$B5,$87,$F9
	.dc.b	$13,$C0,$79,$F8,$8F,$C2,$4D,$9C,$DF,$23,$F2,$11,$F8,$42,$3F,$80
	.dc.b	$7C,$8F,$C2,$C3,$A6,$47,$C9,$7E,$EA,$E2,$FE,$65,$AF,$E8,$BE,$23
	.dc.b	$F2,$24,$BF,$C6,$BD,$A3,$FD,$19,$5D,$8E,$B9,$7B,$EB,$D1,$3F,$AF
	.dc.b	$BF,$1F,$C4,$79,$1F,$2F,$C5,$7E,$78,$95,$F9,$A7,$B5,$FE,$E6,$3F
	.dc.b	$DE,$D8,$77,$7D,$C7,$6C,$3E,$4B,$72,$61,$E0,$8C,$3F,$13,$E2,$BF
	.dc.b	$5A,$48,$FD,$4B,$E0,$7F,$CF,$97,$DE,$FD,$F1,$C8,$8F,$E2,$32,$F6
	.dc.b	$47,$87,$CC,$FD,$C9,$1F,$DA,$BD,$67,$FD,$31,$5D,$72,$F7,$71,$7D
	.dc.b	$7B,$F5,$61,$D2,$FB,$47,$82,$BF,$4D,$E4,$7F,$06,$24,$7F,$02,$7B
	.dc.b	$3F,$20,$F8,$98,$7D,$67,$F8,$D5,$F5,$AF,$D8,$36,$32,$7F,$0C,$7F
	.dc.b	$89,$91,$FE,$1F,$3C,$EF,$F9,$63,$5F,$E3,$D3,$B7,$F9,$33,$5F,$E5
	.dc.b	$FF,$15,$FC,$78,$95,$FC,$69,$F4,$E3,$F0,$E7,$F6,$B7,$D7,$56,$1F
	.dc.b	$76,$73,$87,$C8,$3F,$05,$F0,$BF,$CA,$88,$FF,$26,$7A,$DF,$D1,$8F
	.dc.b	$BE,$71,$9F,$9E,$74,$2E,$5F,$C5,$7C,$EF,$E6,$A4,$9F,$E6,$2F,$47
	.dc.b	$E7,$9F,$96,$9E,$53,$F3,$A3,$31,$D1,$FE,$99,$D5,$FD,$00,$8F,$F3
	.dc.b	$C7,$A4,$72,$8F,$3B,$4F,$EB,$98,$FA,$DB,$1F,$A3,$E7,$FE,$96,$47
	.dc.b	$FA,$3B,$D2,$FF,$98,$63,$F3,$4F,$98,$E2,$7E,$23,$6B,$FB,$04,$FF
	.dc.b	$03,$7F,$B4,$B5,$FE,$47,$3F,$86,$23,$C1,$AF,$06,$BF,$DB,$7E,$23
	.dc.b	$FB,$09,$23,$FA,$EB,$FA,$98,$FE,$D5,$3F,$CB,$25,$AA,$BE,$15,$6A
	.dc.b	$7E,$7D,$FF,$7F,$F4,$5F,$03,$AD,$DF,$13,$FD,$C0,$97,$FD,$B1,$F0
	.dc.b	$47,$E4,$EF,$F6,$F8,$F1,$CF,$BE,$3D,$30,$EE,$7F,$81,$FE,$27,$FB
	.dc.b	$D9,$1F,$EE,$EF,$49,$F5,$69,$E9,$9E,$3A,$7E,$B1,$AA,$FB,$3F,$81
	.dc.b	$3F,$C1,$87,$F8,$20,$91,$FE,$04,$7B,$3F,$D9,$B0,$F4,$3D,$27,$9E
	.dc.b	$3F,$74,$F8,$DF,$03,$FE,$14,$83,$49,$3F,$72,$78,$2B,$3F,$52,$FB
	.dc.b	$C7,$9E,$FB,$CF,$E1,$A1,$5F,$57,$E2,$7C,$04,$9F,$F1,$4D,$23,$D7
	.dc.b	$5C,$27,$C0,$77,$3E,$3E,$9F,$5D,$ED,$7F,$81,$DB,$EB,$FC,$47,$B4
	.dc.b	$93,$D9,$EB,$3F,$7A,$3E,$34,$FD,$E3,$EF,$FC,$4F,$C5,$24,$79,$9E
	.dc.b	$2F,$93,$5F,$B9,$8F,$C1,$AE,$53,$F4,$7D,$5F,$9C,$47,$DE,$F2,$9F
	.dc.b	$97,$5C,$B0,$F9,$F1,$F3,$F4,$F9,$6F,$E1,$CF,$88,$67,$0F,$CA,$FC
	.dc.b	$3F,$DB,$26,$3F,$31,$EC,$FC,$99,$F3,$3F,$00,$FC,$37,$C4,$7E,$0C
	.dc.b	$8F,$D4,$78,$E3,$E5,$8F,$9E,$7E,$0E,$7F,$31,$7F,$01,$EA,$7F,$9B
	.dc.b	$B0,$FC,$DB,$FE,$76,$6B,$C5,$AF,$B7,$1F,$83,$5E,$77,$E1,$C7,$68
	.dc.b	$FC,$3D,$7E,$64,$8F,$E2,$BE,$23,$F3,$84,$70,$7E,$6C,$F4,$CF,$D0
	.dc.b	$E0,$F4,$5F,$E2,$67,$FC,$21,$FA,$8F,$F9,$72,$BE,$79,$E1,$2E,$2F
	.dc.b	$89,$FD,$51,$2B,$F3,$6F,$FA,$03,$DE,$7B,$DF,$C0,$BA,$3F,$C8,$91
	.dc.b	$8A,$FF,$2E,$C7,$F9,$1E,$DF,$13,$FE,$49,$F8,$E7,$BB,$98,$C3,$6F
	.dc.b	$F5,$31,$FB,$B2,$3F,$E9,$4C,$3B,$67,$F8,$88,$FE,$57,$8F,$8B,$2F
	.dc.b	$24,$7F,$8B,$F2,$F0,$BD,$09,$5F,$6F,$0F,$E1,$1F,$09,$27,$F7,$2D
	.dc.b	$9F,$F5,$6E,$3E,$0D,$BF,$C4,$39,$77,$BF,$79,$DF,$0F,$55,$FC,$EB
	.dc.b	$E1,$3F,$85,$9F,$89,$F1,$3F,$C5,$C9,$1F,$C2,$1F,$9D,$9F,$AA,$F9
	.dc.b	$37,$1F,$E8,$97,$D9,$FF,$62,$61,$DA,$BF,$5D,$3F,$47,$E2,$7F,$92
	.dc.b	$13,$1F,$E2,$AF,$67,$D3,$3D,$9B,$7D,$9F,$61,$FE,$59,$BE,$D3,$FC
	.dc.b	$F6,$FE,$F7,$C3,$FF,$31,$25,$7F,$22,$7B,$4F,$D7,$E5,$FE,$B3,$3C
	.dc.b	$F2,$F6,$3F,$14,$EF,$3E,$33,$F2,$B7,$F8,$BF,$88,$FE,$7C,$49,$FE
	.dc.b	$5E,$F6,$CB,$EA,$C7,$53,$D6,$75,$AE,$6E,$5A,$7E,$8B,$CE,$FE,$94
	.dc.b	$48,$FE,$76,$F5,$7F,$93,$6D,$F8,$19,$F6,$36,$BE,$44,$7A,$5A,$7F
	.dc.b	$7B,$F8,$7F,$EA,$C4,$AF,$E8,$EF,$59,$FE,$53,$5C,$E7,$B9,$E7,$AE
	.dc.b	$6F,$F1,$7D,$7E,$9F,$0F,$E7,$AF,$F7,$7F,$11,$FD,$8C,$95,$FD,$51
	.dc.b	$EC,$FE,$A0,$E8,$76,$8F,$43,$F1,$EF,$F1,$BA,$FE,$3E,$3F,$8F,$7C
	.dc.b	$47,$F6,$F2,$3F,$D8,$1E,$BB,$FD,$EB,$E6,$7B,$27,$BB,$F9,$66,$BF
	.dc.b	$8B,$7F,$71,$97,$F7,$B2,$3F,$DB,$1E,$B9,$FC,$59,$E8,$76,$39,$17
	.dc.b	$3F,$89,$97,$E9,$BE,$27,$F8,$D1,$30,$FE,$18,$F4,$7C,$A7,$C5,$AE
	.dc.b	$5C,$1D,$DE,$B5,$F1,$27,$F8,$A6,$9F,$9E,$9F,$E9,$43,$2F,$E2,$EE
	.dc.b	$1A,$FD,$98,$E0,$D9,$E6,$47,$F8,$8C,$69,$FF,$0C,$4B,$E2,$7F,$9A
	.dc.b	$12,$BF,$93,$3F,$8B,$9F,$15,$7F,$8E,$3C,$6F,$55,$FE,$E8,$FD,$DF
	.dc.b	$43,$E1,$F3,$F1,$3F,$CF,$C9,$97,$F3,$17,$A5,$7F,$0F,$8F,$F2,$DD
	.dc.b	$F4,$C3,$E3,$61,$CE,$7E,$ED,$3D,$7F,$E9,$44,$8F,$E7,$8F,$4A,$F3
	.dc.b	$47,$07,$9E,$3E,$8F,$67,$FA,$91,$27,$FA,$23,$C5,$FF,$7D,$9E,$43
	.dc.b	$91,$F8,$EC,$3C,$99,$FE,$1F,$17,$AF,$FD,$74,$95,$FD,$3D,$E9,$3F
	.dc.b	$C6,$73,$F7,$4F,$3C,$7E,$5F,$6F,$F6,$52,$47,$F5,$77,$8E,$FF,$EE
	.dc.b	$8A,$E3,$7F,$97,$CB,$C7,$3F,$8B,$8F,$E9,$18,$7F,$6F,$26,$1F,$D8
	.dc.b	$5E,$6F,$F8,$DE,$F9,$D7,$13,$FC,$B7,$5F,$E5,$DB,$EB,$5F,$8D,$8F
	.dc.b	$EA,$9F,$11,$FD,$EC,$91,$FD,$B5,$EB,$1F,$E3,$AB,$EB,$3C,$6F,$AC
	.dc.b	$AF,$AB,$F9,$4D,$7F,$C0,$7F,$11,$FE,$07,$25,$FF,$79,$7A,$BF,$56
	.dc.b	$7E,$94,$FD,$E8,$FC,$3F,$C4,$7F,$83,$C8,$FF,$80,$5E,$3C,$EC,$BC
	.dc.b	$B7,$E1,$AF,$F3,$E6,$5E,$6C,$BC,$B3,$FA,$29,$FE,$B5,$F1,$1F,$E1
	.dc.b	$B2,$3F,$E0,$D7,$AD,$F4,$9E,$07,$F8,$A2,$B9,$F0,$7E,$A7,$E2,$3F
	.dc.b	$C4,$84,$C3,$FC,$2C,$F2,$7F,$23,$5F,$7F,$7F,$D8,$FC,$47,$F8,$AC
	.dc.b	$93,$FE,$1D,$78,$4E,$E3,$F1,$C7,$EA,$BE,$23,$FC,$64,$47,$FC,$48
	.dc.b	$D8,$FF,$0B,$B5,$FE,$20,$D7,$FA,$43,$5F,$E1,$6F,$F8,$F7,$E2,$24
	.dc.b	$FF,$8B,$DB,$3F,$E2,$26,$BF,$C6,$DA,$FF,$53,$6B,$FC,$59,$AF,$F2
	.dc.b	$3F,$88,$FF,$25,$12,$7F,$C7,$8D,$8F,$F1,$7B,$5F,$E5,$0D,$7F,$B0
	.dc.b	$35,$FE,$46,$FF,$95,$7E,$22,$4F,$F9,$1D,$B3,$FE,$3A,$6B,$FC,$CD
	.dc.b	$AF,$F6,$B6,$BF,$CB,$9A,$FF,$3B,$F8,$8F,$F3,$11,$27,$FC,$A8,$D8
	.dc.b	$FF,$23,$B5,$FE,$80,$D7,$FB,$C3,$5F,$E7,$6F,$F9,$B6,$21,$FE,$5F
	.dc.b	$2C,$7F,$94,$9A,$FF,$4B,$6B,$FE,$03,$6B,$FD,$19,$AB,$E8,$5F,$13
	.dc.b	$FB,$A8,$FF,$50,$46,$27,$FA,$CC,$7F,$B9,$63,$0E,$1A,$FF,$5E,$6B
	.dc.b	$CA,$7A,$3C,$5A,$F8,$5B,$1C,$23,$53,$E0,$74,$D8,$D4,$7A,$7E,$27
	.dc.b	$89,$27,$4D,$8F,$33,$5F,$33,$5F,$7B,$5F,$73,$E4,$F8,$99,$3E,$16
	.dc.b	$CF,$C1,$6B,$F0,$5A,$FC,$A6,$BF,$25,$AF,$D1,$F8,$8F,$82,$49,$F7
	.dc.b	$3F,$13,$E1,$FF,$9C,$F3,$7F,$3A,$E5,$7E,$88,$F6,$F3,$87,$D8,$BB
	.dc.b	$8F,$F2,$3A,$D4,$FF,$22,$34,$58,$60,$FD,$85,$F0,$FC,$A7,$F0,$D8
	.dc.b	$74,$AD,$4E,$76,$7C,$57,$DA,$26,$3F,$5D,$F5,$3D,$F4,$F4,$E9,$C7
	.dc.b	$1F,$3C,$62,$7C,$99,$78,$1F,$36,$5E,$07,$38,$AF,$C9,$9F,$A2,$73
	.dc.b	$F1,$3F,$8C,$23,$F8,$97,$C7,$87,$C0,$83,$F1,$F2,$F2,$47,$9F,$5F
	.dc.b	$93,$3F,$EE,$6C,$7C,$13,$E4,$C3,$C1,$39,$C7,$C7,$AF,$28,$FD,$34
	.dc.b	$78,$1F,$D3,$35,$FA,$66,$AB,$F3,$D1,$98,$FD,$83,$5F,$88,$7E,$C7
	.dc.b	$E2,$BF,$52,$49,$FD,$33,$EE,$D3,$FB,$AE,$3E,$8A,$F8,$07,$F6,$7F
	.dc.b	$75,$E1,$F9,$EB,$78,$DC,$7F,$B4,$E6,$F8,$AC,$63,$FC,$0F,$4F,$DC
	.dc.b	$92,$BF,$6C,$F8,$5F,$6E,$1F,$99,$CB,$BC,$7D,$AC,$79,$DF,$B2,$BF
	.dc.b	$3F,$8F,$7A,$D1,$F8,$0C,$7F,$83,$91,$FE,$08,$FF,$0A,$F8,$5F,$CB
	.dc.b	$7E,$EF,$EF,$BF,$83,$18,$C7,$F3,$0F,$E6,$FE,$61,$CD,$8C,$7F,$1F
	.dc.b	$18,$7F,$1F,$F1,$9F,$C5,$88,$FF,$13,$7C,$0F,$E5,$B1,$F0,$3F,$CA
	.dc.b	$AF,$C0,$FE,$61,$F0,$5F,$7D,$3E,$9E,$4C,$7C,$11,$F8,$FA,$EA,$FF
	.dc.b	$2F,$6B,$FC,$BD,$AF,$7B,$FE,$4D,$18,$8F,$E7,$0D,$7F,$A0,$3F,$CC
	.dc.b	$3E,$32,$4F,$F2,$D6,$CF,$78,$F4,$1F,$A3,$F8,$4F,$46,$1F,$32,$33
	.dc.b	$3E,$AC,$1D,$3A,$9F,$E9,$7F,$19,$FC,$F8,$8F,$F3,$B7,$FA,$2A,$FE
	.dc.b	$01,$7C,$4F,$F1,$A8,$79,$5F,$B6,$7F,$CF,$73,$A7,$FC,$FB,$19,$C7
	.dc.b	$F3,$BE,$B3,$E3,$9E,$87,$F2,$48,$C7,$A3,$67,$38,$7F,$23,$9C,$BF
	.dc.b	$C3,$BE,$3B,$FA,$A9,$1F,$EA,$4F,$A3,$2F,$16,$3F,$AC,$9F,$3B,$E9
	.dc.b	$8C,$EF,$FE,$72,$9C,$9D,$F0,$F0,$D7,$A1,$FF,$48,$C6,$72,$F3,$65
	.dc.b	$FC,$BF,$E3,$BF,$B3,$92,$7F,$B2,$36,$FF,$BA,$35,$FE,$E8,$D7,$FB
	.dc.b	$EB,$5F,$EF,$8F,$F7,$2F,$8E,$93,$FD,$BD,$B3,$FD,$FD,$AF,$F7,$F6
	.dc.b	$BF,$E0,$86,$BF,$E0,$76,$BF,$E0,$FF,$8E,$FE,$FE,$49,$FE,$F8,$D8
	.dc.b	$FF,$06,$35,$FF,$06,$35,$FF,$0A,$B5,$FF,$0A,$3F,$E0,$AF,$88,$93
	.dc.b	$FE,$07,$6C,$FF,$85,$DA,$FF,$85,$DA,$FF,$88,$1A,$FF,$87,$DA,$FF
	.dc.b	$89,$FD,$DF,$F0,$B9,$27,$FC,$28,$D8,$FF,$12,$35,$FF,$12,$35,$FF
	.dc.b	$16,$B5,$FF,$16,$3F,$E2,$2F,$1A,$4F,$F8,$7D,$B3,$D2,$3F,$B0,$CF
	.dc.b	$F7,$58,$FE,$B3,$F2,$9F,$D6,$BE,$89,$9F,$6B,$D9,$2B,$FA,$D1,$FD
	.dc.b	$E6,$FF,$BD,$3F,$E3,$22,$47,$F8,$BD,$F6,$6F,$FD,$2A,$7C,$10,$C3
	.dc.b	$C1,$7D,$E3,$AD,$F1,$23,$FE,$47,$3F,$C8,$37,$FA,$87,$FB,$89,$FE
	.dc.b	$B6,$E4,$C9,$27,$FC,$85,$1E,$E1,$8F,$B9,$F9,$31,$E2,$79,$9E,$53
	.dc.b	$77,$E5,$9E,$1C,$7F,$C1,$7C,$3F,$E0,$47,$F0,$3F,$4A,$E1,$FE,$12
	.dc.b	$CB,$FC,$03,$8F,$46,$BE,$86,$BF,$EC,$EF,$23,$FC,$BA,$F9,$F1,$F3
	.dc.b	$3F,$89,$AF,$33,$F0,$34,$BE,$33,$6F,$F7,$BD,$F9,$B9,$7F,$B3,$DF
	.dc.b	$F0,$5D,$3C,$D8,$FF,$09,$7F,$C6,$D9,$BE,$3F,$E6,$92,$7F,$CD,$51
	.dc.b	$97,$E5,$D7,$C5,$C3,$CB,$97,$86,$7C,$B3,$E9,$DF,$E4,$1F,$E2,$6D
	.dc.b	$FF,$96,$7B,$1F,$CA,$7D,$AF,$E5,$4D,$BF,$91,$DF,$F7,$88,$F8,$9F
	.dc.b	$8B,$87,$87,$0E,$AF,$D8,$E0,$F8,$7C,$1F,$0A,$7C,$1C,$0C,$7B,$5F
	.dc.b	$F2,$76,$EC,$CB,$EA,$79,$1F,$8E,$23,$F0,$9F,$0E,$5D,$A3,$DF,$7E
	.dc.b	$33,$F1,$67,$E2,$06,$BF,$D2,$FD,$6F,$E9,$1E,$CF,$F4,$96,$D7,$E1
	.dc.b	$FE,$3F,$F3,$A4,$7E,$8B,$E0,$BE,$87,$E0,$CA,$7C,$F3,$C3,$C5,$F8
	.dc.b	$1A,$EB,$1F,$D1,$5B,$9F,$BE,$3F,$CB,$2B,$E3,$FF,$5A,$4C,$3F,$02
	.dc.b	$F8,$6F,$EF,$F0,$7C,$DB,$FA,$07,$93,$8F,$F0,$4F,$C1,$D3,$FB,$57
	.dc.b	$B1,$FD,$9C,$FE,$D1,$EC,$DC,$BF,$4D,$F1,$FF,$BC,$23,$F9,$07,$C1
	.dc.b	$C9,$F7,$B0,$FD,$01,$DF,$2F,$16,$3D,$AF,$FB,$26,$3D,$9A,$7E,$CF
	.dc.b	$E3,$FF,$84,$91,$FC,$EB,$DA,$FE,$4E,$1F,$ED,$8E,$2F,$BD,$1E,$BE
	.dc.b	$57,$B5,$FE,$01,$F6,$FF,$C0,$4D,$C7,$F8,$07,$C7,$7F,$16,$23,$FA
	.dc.b	$A7,$B5,$FD,$9E,$1F,$DE,$BF,$64,$C6,$5F,$98,$BE,$91,$FD,$F5,$BC
	.dc.b	$7F,$66,$3F,$D6,$3F,$09,$FC,$90,$98,$FE,$E1,$EF,$C7,$F5,$31,$EF
	.dc.b	$5F,$97,$3B,$9F,$82,$9E,$2F,$FB,$E2,$D1,$88,$FE,$35,$F1,$DF,$CC
	.dc.b	$89,$87,$F0,$67,$BF,$17,$E1,$67,$EB,$E1,$F8,$8A,$FE,$33,$82,$F8
	.dc.b	$D7,$76,$BF,$C9,$FE,$3B,$F9,$F9,$2B,$F8,$9B,$FA,$3F,$88,$E9,$27
	.dc.b	$DB,$C3,$FE,$B4,$72,$7F,$2A,$C1,$5C,$EB,$DC,$D7,$F9,$B7,$C7,$7F
	.dc.b	$4E,$26,$1F,$C8,$9F,$6B,$FC,$03,$0F,$C2,$E5,$D6,$77,$5F,$25,$AF
	.dc.b	$F4,$1F,$8E,$FE,$B2,$4B,$FE,$5A,$F4,$8F,$DD,$9F,$57,$6F,$B1,$C1
	.dc.b	$FD,$1E,$78,$4F,$D5,$6B,$FD,$37,$E3,$BF,$B2,$92,$FF,$9C,$BD,$71
	.dc.b	$F8,$B3,$FA,$99,$FA,$75,$FD,$08,$FE,$EB,$96,$23,$FA,$D7,$C7,$7F
	.dc.b	$70,$23,$FD,$11,$E9,$8F,$E9,$EF,$A1,$EA,$AE,$4D,$3F,$8C,$EA,$C3
	.dc.b	$F0,$CB,$E3,$BF,$BE,$93,$1F,$EA,$2F,$6A,$FE,$1F,$97,$E0,$0F,$59
	.dc.b	$F7,$6F,$F2,$11,$FA,$BB,$ED,$7F,$90,$F8,$EF,$F0,$49,$1F,$EB,$AF
	.dc.b	$59,$FC,$BE,$5F,$C3,$DE,$57,$D1,$FD,$45,$F4,$7D,$7C,$1F,$E1,$42
	.dc.b	$61,$FD,$A1,$E9,$1F,$BD,$DB,$ED,$BE,$AC,$3E,$E5,$7E,$2E,$3E,$2E
	.dc.b	$1F,$13,$E1,$FF,$C3,$E4,$9F,$EE,$6F,$49,$FC,$37,$1F,$97,$6F,$9F
	.dc.b	$AF,$C4,$C7,$AD,$7F,$48,$FC,$15,$F1,$FF,$C2,$63,$FC,$53,$18,$CB
	.dc.b	$FC,$43,$01,$19,$72,$D7,$67,$8A,$3D,$57,$F5,$6F,$F0,$8F,$D5,$8C
	.dc.b	$C7,$92,$3F,$C3,$B4,$F8,$BF,$25,$F8,$63,$17,$D9,$B1,$E2,$9F,$85
	.dc.b	$8F,$EC,$B0,$EE,$75,$2C,$FF,$5E,$D3,$F8,$A5,$BE,$3F,$BC,$FB,$9B
	.dc.b	$19,$C7,$C1,$1D,$A7,$F3,$B9,$7C,$D7,$A4,$FB,$1A,$7C,$FF,$8F,$E6
	.dc.b	$48,$F5,$3C,$E7,$E3,$4F,$E0,$CE,$2F,$55,$C3,$F8,$F8,$F8,$FF,$1F
	.dc.b	$E6,$25,$FC,$57,$A6,$7F,$99,$BF,$CB,$E7,$F8,$57,$F1,$B5,$FE,$4A
	.dc.b	$CD,$F1,$FF,$00,$91,$F3,$DE,$73,$FD,$90,$F3,$1E,$4A,$E9,$8F,$DD
	.dc.b	$8F,$C7,$7A,$DF,$28,$97,$F6,$5E,$71,$FD,$40,$FD,$79,$F0,$31,$FA
	.dc.b	$EF,$59,$F3,$7A,$FF,$50,$8F,$E0,$DE,$7C,$9F,$AC,$9E,$67,$13,$F8
	.dc.b	$86,$1E,$53,$F1,$FE,$97,$DE,$25,$FE,$35,$E7,$C9,$EC,$C7,$F1,$57
	.dc.b	$F0,$37,$FA,$51,$FA,$8F,$47,$F0,$E4,$C3,$F2,$CF,$27,$F0,$95,$E3
	.dc.b	$C7,$EE,$5F,$28,$FC,$F7,$A1,$F9,$02,$57,$E7,$DE,$51,$EA,$C3,$C6
	.dc.b	$FB,$AB,$93,$FB,$6F,$3F,$F3,$24,$BF,$D4,$3C,$B0,$F0,$EF,$ED,$AF
	.dc.b	$A0,$7E,$BF,$CE,$FD,$09,$2F,$F6,$0F,$1C,$7E,$AE,$3E,$0A,$FE,$4B
	.dc.b	$5F,$8B,$F3,$7F,$52,$47,$F7,$0F,$18,$F9,$58,$78,$5F,$89,$7F,$C4
	.dc.b	$BC,$CF,$D8,$92,$FF,$80,$3C,$67,$E0,$57,$1C,$3E,$0F,$23,$CA,$FD
	.dc.b	$C9,$30,$FE,$10,$F1,$BF,$39,$F2,$CF,$82,$7F,$1E,$F2,$3F,$80,$92
	.dc.b	$7F,$88,$3C,$63,$F5,$71,$F7,$2F,$C3,$E3,$FF,$08,$24,$7F,$17,$78
	.dc.b	$75,$FF,$34,$D3,$E4,$E7,$FC,$4B,$E3,$BF,$87,$92,$FF,$8F,$3C,$31
	.dc.b	$FD,$B9,$E7,$3E,$39,$FC,$97,$E3,$BF,$8B,$92,$7F,$93,$3C,$78,$7F
	.dc.b	$41,$87,$82,$7F,$80,$1F,$C7,$88,$7F,$2D,$38,$3F,$8E,$AF,$8F,$3F
	.dc.b	$E5,$6F,$8E,$FF,$6A,$92,$BF,$DD,$AF,$09,$F9,$67,$0A,$FF,$75,$FC
	.dc.b	$77,$FB,$70,$95,$FE,$F4,$78,$4F,$B0,$FB,$B1,$FE,$CA,$F8,$4F,$F7
	.dc.b	$19,$1F,$F7,$DB,$C3,$6F,$C9,$4F,$D4,$3F,$D4,$7F,$03,$FE,$EA,$25
	.dc.b	$7F,$56,$84,$71,$D7,$F1,$47,$F8,$23,$E3,$BF,$DE,$24,$AE,$0E,$DF
	.dc.b	$C0,$47,$F8,$3B,$E3,$BF,$DE,$E4,$8E,$CE,$EF,$EC,$61,$FA,$7F,$89
	.dc.b	$FF,$7E,$92,$BC,$8E,$F6,$FD,$AC,$7F,$BA,$BE,$37,$FA,$AD,$9A,$48
	.dc.b	$FF,$79,$F1,$7F,$61,$9F,$E6,$BD,$FF,$C8,$DF,$A4,$7E,$C2,$7F,$DE
	.dc.b	$BF,$1B,$D0,$8F,$C5,$7C,$5C,$3F,$87,$9F,$F6,$6F,$C6,$78,$89,$5F
	.dc.b	$2D,$DC,$75,$3F,$DA,$7F,$19,$E8,$24,$FD,$17,$71,$F7,$8F,$F6,$CF
	.dc.b	$C5,$FB,$89,$3F,$5D,$DC,$7A,$8F,$F0,$A7,$C5,$FC,$42,$4F,$DD,$77
	.dc.b	$1F,$83,$D7,$FD,$5B,$F1,$5F,$28,$95,$F8,$37,$7A,$FE,$D6,$BF,$C4
	.dc.b	$03,$E2,$7E,$89,$2B,$F1,$4F,$08,$FD,$AD,$FF,$AE,$B4,$FA,$E4,$AF
	.dc.b	$C8,$3B,$8F,$D9,$CF,$6F,$88,$FB,$A4,$AF,$CB,$3B,$AF,$D9,$7C,$47
	.dc.b	$E0,$89,$3F,$9B,$75,$5F,$B0,$F8,$8F,$C3,$92,$7F,$40,$FF,$A9,$EB
	.dc.b	$DB,$C3,$FE,$F6,$F4,$FF,$D0,$67,$FA,$9A,$3F,$CF,$4F,$F7,$B3,$2F
	.dc.b	$F9,$29,$FF,$6B,$F9,$3F,$94,$25,$FE,$B5,$F1,$E1,$FB,$B1,$F1,$1F
	.dc.b	$9A,$24,$FE,$CD,$DF,$3F,$EE,$B7,$FC,$DD,$7E,$DD,$AF,$F9,$39,$AF
	.dc.b	$FA,$3C,$C1,$B2,$18,$21,$F7,$3F,$2E,$FF,$68,$78,$BF,$E5,$86,$C7
	.dc.b	$F9,$81,$AF,$E0,$83,$87,$97,$B9,$FE,$64,$78,$35,$E1,$1B,$7F,$CD
	.dc.b	$4D,$79,$C6,$DE,$73,$E3,$8C,$47,$07,$5F,$74,$20

;--------------------------------------------------------------------------------
;	FREM.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
frem_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FREM.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FREM.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	frem_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	frem.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FREM.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	frem.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FREM.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
frem_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3D,$61,$F3,$7F,$DA,$BD,$E7,$99,$C6,$30,$DA,$FE
	.dc.b	$50,$D9,$DC,$7F,$72,$F8,$17,$97,$FD,$A2,$FF,$B9,$CA,$73,$F2,$7E
	.dc.b	$08,$FE,$EF,$18,$9D,$B6,$33,$5C,$DB,$5C,$5F,$E2,$6D,$7C,$51,$B7
	.dc.b	$35,$D2,$31,$1A,$9E,$0F,$96,$32,$D8,$CD,$78,$E3,$99,$E2,$AE,$93
	.dc.b	$AA,$E3,$5B,$AE,$D5,$A2,$3C,$48,$74,$21,$C0,$C9,$4C,$9E,$57,$9B
	.dc.b	$88,$F0,$39,$38,$94,$D4,$F9,$89,$5E,$92,$57,$49,$F0,$10,$E4,$5A
	.dc.b	$EE,$4A,$E7,$3C,$67,$64,$9D,$57,$68,$F2,$C6,$27,$83,$63,$35,$D1
	.dc.b	$B5,$C5,$F5,$B5,$F2,$46,$DC,$D7,$48,$C4,$6A,$78,$3E,$58,$CB,$63
	.dc.b	$35,$E3,$8E,$67,$8A,$BA,$4E,$AB,$8D,$6E,$BB,$56,$88,$F1,$21,$D0
	.dc.b	$87,$03,$25,$32,$79,$5E,$6E,$23,$C0,$E4,$E2,$53,$53,$E6,$25,$7A
	.dc.b	$49,$5D,$27,$C0,$43,$91,$6B,$B9,$2B,$9C,$F1,$9D,$92,$75,$5D,$A3
	.dc.b	$CB,$18,$9E,$0D,$8C,$D7,$46,$D7,$17,$D6,$D7,$C9,$1B,$73,$5D,$23
	.dc.b	$11,$A9,$E0,$F9,$63,$2D,$8C,$D7,$8E,$39,$9E,$2A,$E9,$3A,$AE,$35
	.dc.b	$BA,$ED,$5A,$23,$C4,$87,$42,$1C,$0C,$94,$C9,$E5,$79,$B8,$8F,$03
	.dc.b	$93,$89,$4D,$4F,$98,$95,$E9,$25,$74,$9F,$01,$0E,$45,$D3,$FD,$25
	.dc.b	$1F,$D1,$A3,$FA,$63,$5F,$EA,$B9,$7F,$3F,$C7,$FA,$3C,$62,$75,$5C
	.dc.b	$A3,$FA,$69,$C8,$A6,$CA,$EE,$79,$E2,$F8,$0F,$F3,$95,$FF,$4A,$6D
	.dc.b	$FF,$4E,$6B,$FE,$59,$F8,$EF,$E7,$B1,$88,$FE,$B4,$D7,$FA,$C3,$6B
	.dc.b	$B4,$F0,$25,$77,$6D,$77,$6B,$DE,$34,$47,$4D,$74,$60,$86,$08,$79
	.dc.b	$A3,$AB,$63,$93,$4E,$25,$38,$9E,$C8,$E2,$D8,$E2,$D3,$89,$4E,$33
	.dc.b	$F0,$A3,$6D,$8D,$B5,$DD,$7C,$92,$57,$82,$75,$3E,$52,$1D,$CA,$E9
	.dc.b	$C1,$1C,$12,$7A,$D7,$91,$B3,$E4,$6B,$E4,$AF,$14,$62,$3B,$B5,$EC
	.dc.b	$DA,$ED,$3C,$09,$5D,$DB,$5D,$DA,$F7,$8D,$11,$D3,$5D,$18,$21,$82
	.dc.b	$1E,$68,$EA,$D8,$E4,$D3,$89,$4E,$27,$B2,$38,$B6,$38,$B4,$E2,$53
	.dc.b	$8C,$FC,$28,$DB,$63,$6D,$77,$5F,$24,$95,$E0,$9D,$4F,$94,$87,$72
	.dc.b	$BA,$70,$47,$04,$9E,$B5,$E4,$6C,$F9,$1A,$F9,$2B,$C5,$18,$8E,$ED
	.dc.b	$7B,$36,$BB,$4F,$02,$57,$76,$D7,$76,$BD,$E3,$44,$74,$D7,$46,$08
	.dc.b	$60,$87,$9A,$3A,$B6,$39,$34,$E2,$53,$89,$EC,$8E,$2D,$8E,$2D,$38
	.dc.b	$94,$E3,$3F,$0A,$36,$D8,$DB,$5D,$D7,$C9,$3F,$96,$61,$FC,$E7,$4F
	.dc.b	$F4,$51,$27,$FA,$06,$3F,$D1,$A7,$FA,$24,$0F,$F4,$91,$4F,$E8,$C5
	.dc.b	$3F,$A3,$15,$EC,$F4,$9E,$CF,$F4,$B9,$ED,$97,$F5,$59,$CD,$70,$AD
	.dc.b	$46,$27,$51,$CC,$86,$8B,$3C,$49,$1C,$23,$44,$8C,$E3,$FD,$8A,$75
	.dc.b	$18,$9F,$EB,$F1,$82,$4E,$C9,$1A,$6D,$FB,$1B,$7E,$96,$CF,$9A,$B9
	.dc.b	$BC,$49,$5C,$5B,$5C,$5A,$71,$AE,$6F,$AE,$B9,$D7,$12,$57,$16,$CF
	.dc.b	$13,$99,$C6,$B9,$DF,$C1,$9D,$CE,$4D,$C6,$6B,$A1,$23,$A4,$F0,$9D
	.dc.b	$46,$27,$4F,$39,$D5,$FA,$27,$35,$C2,$B5,$18,$9D,$47,$32,$1A,$2C
	.dc.b	$F1,$24,$70,$8D,$12,$33,$7F,$02,$75,$18,$9F,$7C,$60,$93,$B2,$46
	.dc.b	$9B,$7E,$C6,$DF,$A5,$B3,$E6,$AE,$6F,$12,$57,$16,$D7,$16,$9C,$6B
	.dc.b	$9B,$EB,$AE,$75,$C4,$95,$C5,$B3,$C4,$E6,$71,$AE,$77,$F0,$67,$73
	.dc.b	$93,$71,$9A,$E8,$48,$E9,$3C,$27,$51,$89,$D3,$CE,$75,$7E,$89,$CD
	.dc.b	$70,$AD,$46,$27,$51,$CC,$86,$8B,$3C,$49,$1C,$23,$44,$8C,$DF,$C0
	.dc.b	$9D,$46,$27,$DF,$18,$24,$EC,$91,$A6,$DF,$B1,$B7,$E9,$6C,$F9,$AB
	.dc.b	$9B,$C4,$95,$C5,$B5,$C5,$A7,$1A,$E6,$FA,$EB,$9D,$71,$25,$71,$6C
	.dc.b	$F1,$39,$9C,$6B,$9D,$FC,$19,$DC,$E4,$DC,$66,$BA,$12,$3A,$4F,$09
	.dc.b	$D4,$62,$74,$F3,$9D,$5F,$A2,$73,$5C,$2B,$51,$89,$D4,$73,$21,$A2
	.dc.b	$CF,$12,$47,$08,$D1,$23,$37,$F0,$27,$51,$89,$F7,$C6,$09,$3B,$24
	.dc.b	$69,$B7,$EC,$6D,$FA,$5B,$3E,$6A,$E6,$F1,$25,$71,$6D,$71,$69,$C6
	.dc.b	$B9,$BE,$BA,$E7,$5C,$49,$5C,$5B,$3C,$4E,$67,$1A,$E7,$7F,$06,$77
	.dc.b	$39,$37,$19,$AE,$84,$8F,$E9,$87,$F8,$06,$D7,$FE,$05,$8F,$F0,$39
	.dc.b	$FE,$0A,$3F,$C1,$19,$BE,$0F,$51,$FE,$0F,$8C,$38,$6B,$CD,$AF,$FB
	.dc.b	$B1,$EE,$FF,$84,$FC,$D6,$5D,$6F,$C5,$18,$BD,$D7,$F8,$6E,$31,$3A
	.dc.b	$AE,$13,$CF,$CA,$FF,$11,$7C,$DF,$26,$C6,$44,$65,$A7,$F8,$97,$67
	.dc.b	$03,$3C,$CF,$F8,$AE,$73,$18,$8F,$F1,$84,$62,$BE,$53,$63,$DD,$87
	.dc.b	$C9,$9F,$46,$7C,$65,$9F,$BE,$72,$7F,$8D,$FE,$51,$F8,$17,$FB,$E6
	.dc.b	$9F,$82,$66,$30,$DE,$2F,$F2,$13,$67,$67,$F9,$0F,$17,$DF,$3C,$3F
	.dc.b	$AB,$3F,$76,$31,$3F,$83,$37,$3A,$69,$98,$79,$5F,$4A,$BF,$CA,$53
	.dc.b	$DA,$31,$1D,$5B,$19,$3E,$EC,$32,$F7,$F8,$FF,$8E,$6C,$7E,$55,$AF
	.dc.b	$C1,$8F,$C9,$AD,$FE,$AB,$F4,$B8,$7E,$D3,$78,$7F,$33,$3A,$8F,$C0
	.dc.b	$4E,$9B,$3A,$8C,$47,$E0,$5A,$BE,$71,$F9,$C3,$B7,$F4,$DF,$11,$FE
	.dc.b	$6F,$9C,$C6,$2B,$FC,$E9,$18,$D1,$5B,$BE,$51,$8A,$F1,$9E,$92,$18
	.dc.b	$3F,$CF,$9F,$4C,$F4,$3F,$6D,$3F,$BB,$8C,$3F,$79,$AE,$63,$73,$F8
	.dc.b	$5F,$C0,$3F,$41,$3F,$81,$6C,$7F,$A2,$61,$18,$8E,$1A,$7E,$29,$FA
	.dc.b	$5A,$7C,$66,$C7,$CB,$23,$CF,$E2,$9D,$BF,$C3,$30,$46,$27,$F2,$4D
	.dc.b	$78,$4E,$F0,$7C,$A7,$F1,$5F,$A4,$F3,$46,$30,$46,$27,$57,$EF,$6B
	.dc.b	$F7,$71,$F2,$BF,$34,$C1,$E7,$F2,$FF,$D5,$D5,$63,$A9,$FA,$DC,$93
	.dc.b	$A8,$C3,$5E,$67,$53,$F8,$4E,$7F,$1F,$8B,$F2,$91,$A2,$1A,$C3,$CA
	.dc.b	$E0,$F9,$A7,$E1,$0D,$E5,$F5,$BE,$53,$F7,$1A,$1A,$48,$E7,$3D,$4E
	.dc.b	$45,$3F,$09,$1B,$7E,$6B,$5F,$F6,$67,$D1,$BF,$4C,$FD,$FC,$23,$15
	.dc.b	$FB,$E8,$C4,$EA,$73,$E1,$78,$27,$D6,$7C,$52,$9F,$ED,$B8,$E7,$F9
	.dc.b	$17,$F5,$76,$C7,$8A,$BF,$AB,$4E,$5B,$1F,$9D,$E7,$F8,$64,$C3,$F3
	.dc.b	$CF,$E6,$74,$FA,$33,$F0,$4F,$9C,$E3,$A7,$F9,$9F,$87,$FC,$57,$C0
	.dc.b	$34,$95,$FC,$5E,$1F,$23,$FD,$9E,$3F,$DE,$95,$FC,$F2,$31,$1F,$EF
	.dc.b	$98,$C3,$FC,$41,$BC,$1F,$17,$0F,$2C,$FC,$43,$CC,$43,$04,$7D,$79
	.dc.b	$7E,$E3,$E2,$FF,$BD,$7D,$17,$E9,$9B,$5F,$C0,$5A,$FF,$3A,$E2,$46
	.dc.b	$23,$F9,$B6,$7F,$C5,$7D,$9F,$EA,$5D,$A9,$46,$1A,$7C,$09,$E0,$DF
	.dc.b	$86,$E1,$D5,$FD,$6D,$B1,$93,$F7,$A5,$8F,$EC,$1D,$7D,$FE,$12,$C9
	.dc.b	$A4,$7F,$8E,$7C,$67,$F8,$67,$1F,$E0,$AD,$8C,$ED,$E6,$7E,$04,$65
	.dc.b	$FC,$0F,$57,$ED,$49,$C3,$FD,$97,$F5,$4F,$E2,$F2,$F4,$3F,$93,$46
	.dc.b	$1B,$DD,$F1,$5B,$3B,$3E,$27,$C4,$7C,$07,$F0,$5C,$08,$C4,$7D,$06
	.dc.b	$F3,$7F,$81,$1C,$C7,$DF,$CB,$CA,$FE,$EB,$4F,$E6,$FF,$2F,$FE,$39
	.dc.b	$FA,$8F,$E4,$AD,$9F,$E8,$0D,$7F,$00,$D7,$FC,$81,$CC,$F8,$EF,$F2
	.dc.b	$17,$E2,$3F,$CB,$9B,$3F,$D2,$5A,$FE,$25,$AF,$E2,$3A,$FF,$0D,$F3
	.dc.b	$BF,$E1,$96,$C7,$E1,$5A,$FC,$26,$BF,$9D,$EC,$FD,$09,$34,$F7,$C7
	.dc.b	$F0,$78,$C7,$1F,$EC,$7A,$7F,$29,$DD,$F8,$DA,$FE,$EC,$7E,$35,$C1
	.dc.b	$FD,$23,$E6,$BF,$CB,$DF,$79,$78,$4D,$25,$7D,$83,$99,$F6,$38,$BB
	.dc.b	$B9,$BF,$F1,$B4,$62,$3F,$17,$8F,$F3,$07,$F7,$5C,$7F,$C0,$3E,$7B
	.dc.b	$E2,$41,$A4,$7D,$DA,$FC,$07,$99,$FC,$15,$C3,$F8,$73,$FA,$A9,$E5
	.dc.b	$29,$9F,$23,$CA,$E0,$FC,$55,$F9,$76,$FE,$41,$CD,$F9,$39,$FF,$3B
	.dc.b	$9F,$07,$D5,$E2,$47,$CB,$5C,$70,$F1,$C7,$43,$CD,$19,$3F,$A8,$E5
	.dc.b	$FE,$5E,$F2,$7F,$85,$7C,$FF,$F9,$75,$B5,$FC,$25,$AF,$E9,$9A,$FF
	.dc.b	$A5,$7F,$05,$7A,$BF,$E9,$7F,$C8,$FF,$C3,$11,$8E,$34,$62,$75,$5F
	.dc.b	$E6,$5A,$D4,$F9,$A2,$1E,$C2,$9F,$40,$C1,$0C,$1F,$E6,$28,$7A,$5F
	.dc.b	$EA,$9F,$4B,$C6,$D9,$F0,$35,$E0,$D7,$D9,$E5,$7F,$9C,$6B,$FA,$7E
	.dc.b	$3F,$E7,$56,$CF,$BA,$7F,$AD,$6D,$F4,$38,$7F,$47,$3E,$83,$F5,$51
	.dc.b	$FE,$5A,$DB,$FD,$1B,$F1,$3F,$CE,$7E,$67,$FD,$73,$18,$A4,$63,$8D
	.dc.b	$5B,$CF,$F9,$E3,$6B,$E0,$9D,$B8,$BE,$F3,$E1,$C3,$A7,$BF,$CE,$BF
	.dc.b	$A4,$36,$4D,$3F,$D9,$39,$FA,$A3,$E7,$1F,$7F,$AF,$FD,$A9,$3F,$ED
	.dc.b	$0C,$5B,$27,$31,$F7,$5B,$5F,$55,$AF,$E2,$7C,$9F,$F6,$DC,$FE,$0D
	.dc.b	$B1,$F6,$DA,$EE,$3F,$D3,$EF,$F3,$8F,$85,$FF,$57,$34,$C9,$FD,$A1
	.dc.b	$CB,$4C,$9F,$A5,$3D,$B1,$EA,$9E,$ED,$39,$96,$3E,$05,$FF,$9A,$A3
	.dc.b	$9E,$3F,$1E,$7E,$99,$D4,$BE,$F3,$DE,$FF,$79,$F4,$FF,$9B,$63,$1E
	.dc.b	$A7,$F7,$B8,$C4,$EB,$2F,$EF,$D6,$BD,$BF,$EF,$A8,$CC,$62,$B8,$DF
	.dc.b	$39,$CD,$F3,$BF,$F0,$45,$EE,$7F,$CE,$47,$FA,$DC,$91,$FE,$B4,$6B
	.dc.b	$A3,$04,$3E,$B6,$3F,$0D,$B1,$F1,$84,$66,$3E,$D4,$7E,$3B,$8F,$FE
	.dc.b	$00,$3D,$A3,$FC,$E0,$DE,$0F,$F7,$9E,$BF,$E1,$F8,$FD,$23,$6F,$FC
	.dc.b	$3C,$DA,$FC,$14,$F8,$67,$C0,$7D,$B7,$04,$7F,$2C,$3C,$6F,$BB,$BF
	.dc.b	$A1,$B1,$93,$6F,$A7,$8B,$FD,$4D,$7E,$83,$E2,$14,$FF,$06,$F1,$7F
	.dc.b	$8D,$E7,$F6,$04,$AF,$DD,$CF,$53,$E7,$10,$E4,$43,$FD,$7B,$87,$C7
	.dc.b	$69,$E0,$CB,$F8,$57,$C2,$7E,$F4,$95,$D9,$F4,$57,$1E,$16,$BF,$4B
	.dc.b	$0F,$49,$D0,$F1,$14,$F1,$DF,$F0,$0E,$6F,$BB,$7F,$C8,$E3,$E1,$99
	.dc.b	$2F,$1F,$F2,$A8,$FE,$67,$0F,$CA,$29,$2F,$F9,$AC,$65,$F1,$E3,$E6
	.dc.b	$DB,$E7,$C7,$8C,$91,$9B,$F6,$65,$EB,$9F,$4C,$E5,$E7,$19,$7E,$64
	.dc.b	$65,$DC,$65,$F8,$17,$FC,$7B,$AF,$F3,$7F,$C0,$3F,$D6,$E0,$D2,$4F
	.dc.b	$F6,$91,$F4,$3F,$DB,$5A,$65,$C1,$E8,$CB,$DB,$8F,$8E,$7C,$B8,$F8
	.dc.b	$E3,$11,$F0,$23,$04,$9E,$2D,$CB,$FC,$1F,$F0,$52,$BF,$C5,$4D,$7D
	.dc.b	$F5,$C0,$93,$FE,$16,$FA,$29,$3F,$E3,$16,$BC,$23,$D0,$47,$E6,$CF
	.dc.b	$E9,$38,$7F,$D5,$BC,$BF,$E2,$E2,$4F,$F9,$44,$3C,$5E,$87,$F8,$FF
	.dc.b	$CF,$FF,$13,$91,$C1,$FC,$93,$67,$9A,$F2,$7F,$87,$FE,$9D,$FE,$4B
	.dc.b	$25,$6E,$75,$49,$FF,$2B,$5C,$AF,$F5,$13,$5F,$19,$0F,$A7,$97,$C6
	.dc.b	$7F,$61,$DF,$FE,$D0,$5E,$27,$F9,$30,$8F,$BD,$FF,$3D,$60,$FD,$C2
	.dc.b	$92,$7F,$CF,$91,$97,$FD,$0E,$BF,$33,$FF,$45,$B6,$33,$18,$3E,$53
	.dc.b	$E6,$DF,$FC,$9F,$D9,$23,$FC,$D4,$D7,$FD,$95,$4F,$4F,$FD,$5B,$F3
	.dc.b	$BF,$EA,$82,$4F,$CF,$8C,$BF,$7A,$FB,$D7,$D2,$F5,$7F,$9F,$FE,$AF
	.dc.b	$FE,$C0,$83,$49,$5F,$ED,$5A,$FF,$6D,$57,$FB,$30,$91,$90,$FF,$A2
	.dc.b	$F1,$91,$FE,$D7,$6B,$FE,$BA,$F6,$BF,$DC,$97,$FE,$E3,$24,$7F,$B8
	.dc.b	$E3,$2F,$FB,$C7,$07,$E5,$FF,$EF,$56,$C6,$63,$07,$CC,$7F,$61,$C1
	.dc.b	$F8,$17,$F0,$87,$F2,$5E,$67,$0F,$E0,$FD,$0F,$F4,$69,$27,$F4,$0F
	.dc.b	$F7,$C7,$FD,$FB,$E0,$FF,$5D,$7F,$DF,$0D,$7F,$DA,$AD,$76,$78,$4F
	.dc.b	$BC,$7F,$50,$C7,$F4,$7E,$8F,$F3,$48,$FD,$03,$5F,$BF,$CE,$8C,$47
	.dc.b	$FB,$41,$AB,$E0,$1E,$FF,$F7,$AF,$8B,$F4,$4E,$63,$0F,$AE,$77,$82
	.dc.b	$B7,$18,$7E,$57,$81,$FC,$68,$FE,$37,$1C,$3F,$2B,$DE,$1F,$E0,$86
	.dc.b	$D9,$A7,$C9,$7F,$80,$FC,$1F,$A9,$A7,$F5,$2C,$D1,$96,$C7,$C9,$CB
	.dc.b	$ED,$F3,$7F,$84,$63,$8B,$D6,$76,$DA,$E4,$7C,$6C,$BD,$5E,$17,$F2
	.dc.b	$FC,$3F,$48,$43,$F0,$B1,$AE,$2F,$B1,$8B,$B7,$F8,$3D,$3C,$1F,$64
	.dc.b	$61,$BD,$C9,$D1,$FC,$86,$74,$F7,$3F,$06,$7D,$EF,$80,$FF,$18,$D7
	.dc.b	$AD,$B1,$FC,$8D,$AE,$EF,$AB,$FC,$12,$9A,$7E,$46,$7D,$0D,$8C,$CF
	.dc.b	$F8,$01,$B1,$F7,$B3,$F8,$93,$F9,$0C,$BE,$91,$2F,$EB,$3F,$90,$F8
	.dc.b	$57,$C2,$FF,$92,$3E,$B7,$F1,$F1,$8A,$46,$27,$57,$FA,$BA,$D4,$F8
	.dc.b	$6F,$F4,$DE,$B3,$E1,$FF,$41,$8B,$E2,$BF,$65,$39,$E8,$FF,$2B,$C6
	.dc.b	$2B,$6D,$8C,$E5,$F3,$CF,$C4,$97,$8B,$EF,$6D,$FB,$DE,$CF,$E0,$7E
	.dc.b	$C7,$EE,$5B,$5F,$D9,$7B,$7F,$DA,$ED,$9F,$9D,$F1,$4F,$8A,$FD,$14
	.dc.b	$FF,$5F,$8C,$6D,$FA,$38,$C4,$E9,$FE,$77,$3A,$EA,$FD,$33,$EF,$AF
	.dc.b	$F1,$AB,$72,$CE,$BF,$A9,$8C,$5F,$FA,$43,$3F,$E3,$7E,$7F,$E4,$7F
	.dc.b	$68,$F6,$6B,$FF,$00,$29,$18,$9D,$47,$19,$D4,$FF,$08,$F3,$9E,$CF
	.dc.b	$FA,$13,$F2,$DC,$36,$FF,$92,$4E,$B5,$56,$9B,$3F,$CA,$78,$7F,$C8
	.dc.b	$93,$FC,$51,$B1,$9E,$97,$37,$08,$FF,$55,$CE,$9F,$F5,$6E,$0F,$0F
	.dc.b	$F5,$D8,$79,$DB,$3E,$46,$BC,$1A,$FE,$D6,$7D,$1C,$5E,$72,$47,$B3
	.dc.b	$3F,$AC,$FE,$EF,$0F,$F2,$E4,$FA,$72,$F3,$B7,$2F,$BD,$5E,$89,$CB
	.dc.b	$FC,$CA,$3F,$9B,$BE,$EC,$FF,$CC,$AD,$E2,$FE,$6E,$D3,$FA,$87,$B1
	.dc.b	$FE,$CA,$6F,$B2,$F6,$B7,$39,$9F,$F6,$4C,$72,$5F,$07,$FE,$BC,$7A
	.dc.b	$B5,$FE,$8A,$D7,$FA,$F3,$5F,$8D,$BF,$F1,$B9,$F9,$D1,$E1,$3E,$71
	.dc.b	$4F,$9C,$53,$E4,$67,$F7,$23,$E3,$12,$3F,$01,$9E,$FE,$D5,$FA,$33
	.dc.b	$CE,$FE,$55,$F2,$1F,$DC,$67,$31,$8B,$FF,$77,$C6,$14,$EE,$B9,$46
	.dc.b	$0F,$3C,$7A,$88,$E0,$F1,$BF,$6C,$8E,$0F,$D3,$7C,$AF,$F3,$DF,$9F
	.dc.b	$FD,$6B,$63,$22,$32,$D3,$FC,$C5,$EE,$FB,$78,$BF,$BA,$87,$11,$FD
	.dc.b	$9D,$AF,$0C,$BE,$A7,$5F,$F8,$63,$C6,$FE,$9D,$F2,$7F,$BE,$6D,$7E
	.dc.b	$ED,$AF,$F5,$39,$E0,$A7,$F8,$2F,$D6,$7E,$42,$7F,$C4,$6D,$ED,$34
	.dc.b	$FC,$67,$4F,$FA,$F6,$FF,$40,$43,$ED,$9F,$E1,$0F,$A6,$FE,$9F,$F2
	.dc.b	$97,$AB,$FB,$23,$64,$D3,$FC,$2F,$F7,$1F,$E8,$0F,$87,$BE,$FB,$85
	.dc.b	$18,$6D,$FF,$8D,$6B,$83,$E2,$21,$FB,$3E,$7F,$D6,$E7,$E9,$8C,$9F
	.dc.b	$96,$8C,$F4,$FF,$12,$8F,$61,$FE,$70,$E3,$FF,$01,$61,$F6,$1B,$1F
	.dc.b	$5F,$8B,$F5,$13,$F6,$DF,$D5,$43,$0F,$03,$E3,$3A,$9E,$13,$FC,$4D
	.dc.b	$FA,$D3,$E9,$BF,$81,$DA,$31,$3F,$88,$6B,$C2,$77,$83,$E1,$BE,$47
	.dc.b	$C6,$7E,$2D,$B5,$F9,$26,$BE,$D6,$BE,$33,$E2,$9E,$9C,$FE,$23,$F2
	.dc.b	$31,$F2,$79,$3F,$E1,$FB,$F4,$B6,$7F,$D8,$19,$7F,$79,$53,$A6,$9F
	.dc.b	$E6,$DE,$3F,$F0,$AF,$77,$E1,$49,$19,$E4,$FE,$24,$F9,$36,$F9,$6F
	.dc.b	$9C,$F9,$86,$CF,$C6,$97,$93,$F1,$CD,$3F,$CF,$9F,$72,$F9,$7F,$AE
	.dc.b	$DD,$7F,$99,$B6,$7F,$CD,$6D,$7C,$5B,$7C,$47,$C2,$43,$FC,$85,$1B
	.dc.b	$7F,$C8,$81,$C3,$AA,$F0,$9F,$CC,$7D,$07,$D0,$3C,$FF,$D1,$78,$3F
	.dc.b	$E2,$28,$C6,$5F,$C7,$70,$FF,$28,$5F,$0B,$DC,$78,$23,$13,$B8,$F9
	.dc.b	$04,$3E,$F7,$3F,$E8,$8F,$F4,$4C,$F0,$F8,$EF,$05,$FE,$CD,$B6,$69
	.dc.b	$FC,$67,$EA,$7F,$47,$CE,$69,$27,$B6,$3F,$A2,$3C,$27,$E9,$31,$78
	.dc.b	$6F,$AE,$FE,$CB,$18,$6C,$66,$7F,$B7,$35,$FF,$50,$FD,$73,$EB,$7F
	.dc.b	$29,$1F,$4A,$30,$FE,$48,$C4,$69,$FF,$3D,$46,$BA,$3C,$DC,$BF,$90
	.dc.b	$9F,$39,$FA,$6F,$C8,$FE,$9F,$37,$FA,$C1,$B1,$FE,$C9,$6B,$F8,$26
	.dc.b	$BF,$C7,$7F,$14,$7C,$F7,$F1,$FF,$CA,$BF,$D5,$11,$8D,$D1,$89,$D5
	.dc.b	$7E,$32,$55,$BA,$E3,$F5,$FF,$CA,$78,$2D,$7E,$5C,$DC,$66,$F9,$5A
	.dc.b	$31,$1F,$D2,$7C,$7F,$E1,$8B,$EC,$7F,$C9,$BE,$17,$FA,$7D,$B1,$93
	.dc.b	$FC,$99,$19,$7F,$C9,$DF,$4D,$FE,$A5,$F4,$BF,$CA,$AD,$8C,$88,$CB
	.dc.b	$4F,$2F,$2F,$F4,$3F,$C1,$7E,$2B,$63,$3A,$27,$2D,$8F,$F7,$D7,$DB
	.dc.b	$BE,$2F,$FA,$67,$E2,$DF,$EE,$68,$C5,$23,$13,$AA,$E2,$D7,$FD,$7D
	.dc.b	$2F,$D6,$99,$FF,$B0,$7F,$10,$FF,$70,$46,$29,$18,$9D,$47,$D9,$9D
	.dc.b	$4F,$F9,$E3,$B3,$FD,$95,$F1,$56,$BF,$DC,$A7,$F9,$DE,$32,$D3,$27
	.dc.b	$F0,$EF,$D7,$1F,$5D,$FD,$A7,$AB,$F8,$8C,$FF,$B5,$67,$39,$FF,$26
	.dc.b	$8E,$15,$B7,$73,$AA,$FC,$9F,$B9,$FD,$CF,$F2,$DB,$81,$A4,$BF,$F4
	.dc.b	$A7,$D6,$FF,$83,$D7,$9D,$FE,$AF,$8C,$37,$D8,$FF,$74,$36,$76,$7A
	.dc.b	$3C,$2F,$DB,$FC,$07,$EF,$3C,$4F,$F7,$73,$67,$FD,$2E,$6E,$32,$FF
	.dc.b	$36,$EF,$FE,$1E,$FF,$5C,$AF,$F0,$4E,$BF,$C1,$BD,$6F,$F0,$AC,$F7
	.dc.b	$6B,$FC,$1D,$AF,$ED,$FE,$D7,$FC,$29,$E4,$7E,$D9,$B1,$FA,$A6,$BF
	.dc.b	$86,$F0,$91,$88,$7B,$4F,$0B,$F0,$DF,$75,$FA,$02,$4F,$E9,$1F,$C1
	.dc.b	$73,$48,$FC,$03,$5F,$B3,$C9,$FC,$CD,$FB,$31,$FC,$DD,$B1,$97,$7E
	.dc.b	$5F,$C9,$FA,$49,$1F,$24,$38,$7F,$AE,$1F,$C8,$0B,$E5,$7F,$23,$E0
	.dc.b	$FF,$19,$D7,$CB,$9E,$C7,$F2,$AD,$CE,$04,$F0,$7F,$35,$D1,$F9,$F2
	.dc.b	$38,$3E,$A0,$FC,$C7,$B1,$1C,$BF,$A7,$A9,$1F,$A5,$6B,$F6,$C7,$E6
	.dc.b	$1E,$72,$3D,$9F,$E7,$FE,$AF,$F7,$D8,$FE,$CD,$97,$F8,$43,$47,$E7
	.dc.b	$D9,$9F,$F2,$B3,$63,$3C,$DF,$E3,$A8,$C7,$27,$FB,$F7,$F2,$7F,$E7
	.dc.b	$BE,$DF,$F9,$75,$B6,$69,$FD,$0B,$E3,$7F,$CC,$5F,$53,$64,$D2,$65
	.dc.b	$F9,$F3,$D3,$97,$FA,$3B,$27,$D5,$78,$C9,$39,$7F,$CF,$BA,$FA,$F5
	.dc.b	$FE,$BC,$47,$E1,$7C,$3B,$8B,$FA,$1D,$FF,$AD,$5A,$FF,$AD,$1A,$FE
	.dc.b	$02,$38,$79,$FF,$E5,$CF,$2F,$FD,$C5,$C9,$FC,$53,$6F,$E6,$1A,$FF
	.dc.b	$9F,$E3,$13,$AC,$3F,$AF,$35,$FE,$62,$FF,$B7,$AB,$FD,$56,$D8,$E3
	.dc.b	$F1,$3F,$E7,$48,$C4,$7C,$52,$1F,$CE,$B6,$FE,$D2,$FC,$7E,$3F,$B0
	.dc.b	$47,$F9,$49,$FB,$8F,$13,$FC,$11,$87,$F1,$EF,$5B,$FB,$94,$FF,$AB
	.dc.b	$E7,$2D,$8C,$BF,$8A,$DF,$E7,$13,$4F,$F4,$AF,$CE,$EA,$7F,$D1,$0D
	.dc.b	$E1,$34,$E7,$DE,$C7,$BD,$79,$71,$CC,$62,$75,$1C,$F1,$F0,$57,$E1
	.dc.b	$88,$7C,$41,$F8,$33,$F0,$DF,$F2,$9E,$9F,$D8,$27,$F7,$73,$9F,$77
	.dc.b	$FD,$6D,$10,$E3,$5C,$1C,$4E,$EB,$94,$7C,$BF,$67,$CE,$47,$39,$7B
	.dc.b	$63,$D1,$B7,$F3,$A9,$FE,$23,$97,$ED,$70,$E9,$1D,$B1,$F0,$63,$FE
	.dc.b	$70,$9C,$B6,$33,$3C,$AF,$A3,$F5,$5A,$FF,$1F,$3F,$CC,$DE,$B7,$E8
	.dc.b	$7E,$0B,$FA,$8F,$C1,$7F,$9F,$23,$12,$F2,$D3,$9E,$A5,$F0,$AF,$F1
	.dc.b	$14,$61,$E1,$3B,$9D,$47,$F8,$8E,$75,$5D,$23,$8C,$EA,$FC,$33,$99
	.dc.b	$FF,$3D,$DF,$E6,$38,$DA,$B2,$D5,$7F,$B8,$6B,$53,$96,$CE,$A7,$99
	.dc.b	$FE,$0A,$AE,$71,$96,$CF,$F8,$3B,$5F,$BC,$FD,$3E,$0F,$F7,$D3,$76
	.dc.b	$F8,$A7,$D3,$8C,$BD,$71,$F1,$63,$71,$ED,$18,$8F,$F7,$E5,$FF,$82
	.dc.b	$F7,$FE,$9A,$FE,$47,$83,$F1,$13,$D8,$FD,$A4,$EB,$0F,$F0,$AB,$63
	.dc.b	$F3,$7C,$1F,$9E,$BF,$31,$2B,$CC,$73,$9F,$90,$79,$63,$E3,$91,$C1
	.dc.b	$0F,$39,$4F,$DB,$1C,$EF,$F6,$AD,$E2,$FF,$72,$CF,$AD,$B5,$F4,$78
	.dc.b	$FD,$15,$FD,$B3,$6F,$E0,$AF,$94,$FE,$12,$57,$37,$FB,$58,$FB,$67
	.dc.b	$90,$87,$B0,$FE,$30,$78,$F8,$FB,$E3,$E6,$9F,$24,$78,$B0,$FE,$C7
	.dc.b	$C9,$F4,$AF,$FD,$29,$CB,$FC,$B2,$FF,$9F,$36,$33,$3B,$7C,$27,$90
	.dc.b	$F1,$12,$BF,$CE,$9C,$FF,$8A,$8F,$E8,$6D,$7F,$A1,$B5,$FF,$3E,$B6
	.dc.b	$3F,$A6,$8F,$E0,$1F,$E9,$BC,$28,$C5,$7F,$4E,$8C,$4E,$A7,$33,$FD
	.dc.b	$6F,$9F,$FD,$47,$97,$F6,$1E,$AF,$54,$BA,$FF,$AF,$BF,$AE,$6C,$FE
	.dc.b	$16,$3D,$FB,$FE,$A1,$FC,$26,$1F,$E0,$CE,$CE,$6D,$8F,$DB,$B6,$BE
	.dc.b	$5C,$F8,$27,$E4,$12,$BA,$3F,$72,$BF,$D8,$BD,$DE,$FF,$A9,$FF,$05
	.dc.b	$12,$7F,$D8,$31,$97,$FD,$87,$E0,$FC,$5F,$DC,$BF,$C2,$A4,$8C,$DA
	.dc.b	$3E,$EE,$FE,$8E,$2F,$E1,$D5,$E8,$6E,$9A,$8F,$DD,$CC,$8F,$89,$1F
	.dc.b	$80,$3F,$C8,$BE,$27,$F9,$23,$F8,$03,$FC,$98,$48,$FF,$21,$BF,$EA
	.dc.b	$5E,$09,$1F,$EA,$36,$BF,$E3,$FE,$2F,$F2,$A1,$1F,$F2,$5B,$FE,$AD
	.dc.b	$F8,$B9,$1F,$EA,$B6,$BF,$E4,$7F,$AC,$FF,$2E,$11,$FF,$29,$BE,$87
	.dc.b	$E8,$76,$FF,$15,$7D,$5E,$43,$D1,$FF,$30,$FD,$77,$FA,$5C,$93,$FE
	.dc.b	$6F,$7F,$C1,$BE,$B7,$FA,$D3,$0B,$E7,$7E,$55,$BC,$66,$9F,$EC,$5F
	.dc.b	$5E,$61,$FE,$C3,$6B,$FE,$F9,$F6,$FF,$CE,$3E,$9F,$FB,$B0,$91,$FE
	.dc.b	$F0,$7F,$D3,$56,$FD,$9E,$92,$7F,$D3,$91,$97,$FC,$E7,$F4,$BF,$E7
	.dc.b	$F2,$38,$3F,$8F,$1F,$13,$9B,$EF,$3F,$E8,$DF,$96,$FF,$5C,$11,$C1
	.dc.b	$FE,$C8,$F6,$BF,$D0,$FF,$A7,$FF,$B0,$C9,$19,$0F,$FB,$4B,$F7,$F9
	.dc.b	$1F,$E8,$E3,$F7,$9E,$2F,$F9,$5F,$24,$EA,$7F,$D4,$F9,$7C,$C2,$3F
	.dc.b	$EA,$67,$FD,$E1,$F1,$32,$3F,$DD,$CD,$7F,$E0,$0F,$C0,$FF,$BF,$7F
	.dc.b	$5A,$FF,$7E,$41,$A4,$7F,$DE,$3F,$08,$F8,$4F,$E2,$1F,$B4,$FF,$93
	.dc.b	$23,$19,$A3,$13,$AA,$E2,$D7,$FD,$99,$F0,$FF,$C6,$7E,$63,$FC,$D5
	.dc.b	$B2,$31,$3F,$0F,$04,$E5,$57,$0A,$DD,$7E,$1C,$7E,$C7,$F2,$DF,$F6
	.dc.b	$BF,$C1,$39,$FF,$D7,$73,$9F,$47,$FD,$B5,$18,$AF,$A5,$18,$79,$98
	.dc.b	$A7,$87,$FE,$C4,$FD,$4F,$94,$62,$51,$89,$D4,$71,$6B,$FE,$ED,$F8
	.dc.b	$CF,$E7,$5B,$7C,$26,$C7,$C1,$D9,$18,$8E,$1B,$7D,$37,$5E,$BF,$F4
	.dc.b	$3F,$DF,$FE,$5B,$63,$32,$9C,$B6,$3E,$37,$C7,$BE,$43,$FA,$67,$47
	.dc.b	$FA,$22,$31,$48,$C4,$EA,$B8,$B5,$FE,$8B,$E3,$FF,$0B,$7F,$55,$E6
	.dc.b	$7F,$A4,$BC,$8F,$E1,$90,$FA,$8C,$79,$3F,$EE,$E6,$C6,$4F,$F7,$E3
	.dc.b	$F6,$FC,$27,$77,$E1,$E7,$FC,$87,$18,$E9,$FE,$B6,$D8,$CC,$EF,$1F
	.dc.b	$E4,$1E,$9F,$F4,$E9,$FE,$70,$DD,$9F,$12,$F1,$FF,$B9,$36,$3F,$AD
	.dc.b	$B6,$FF,$D8,$4D,$BF,$DD,$7C,$93,$DE,$F9,$3E,$57,$E4,$E3,$1C,$5F
	.dc.b	$C9,$6B,$75,$A6,$C6,$5F,$97,$AF,$F1,$27,$FC,$3F,$5C,$B2,$7C,$FF
	.dc.b	$5C,$3E,$BB,$70,$FF,$03,$36,$77,$7F,$A5,$F9,$67,$CB,$7F,$84,$61
	.dc.b	$18,$9F,$F3,$53,$75,$FF,$65,$B6,$7F,$93,$EF,$F2,$DF,$BE,$75,$BE
	.dc.b	$93,$B8,$FE,$3C,$D4,$78,$23,$C5,$F1,$FF,$E8,$7E,$7F,$F6,$CB,$67
	.dc.b	$F7,$8D,$7F,$94,$35,$FB,$7C,$FF,$91,$C3,$C8,$79,$67,$FD,$25,$F3
	.dc.b	$FD,$32,$F0,$B6,$CD,$3F,$DF,$DF,$30,$F6,$BF,$2D,$3F,$C5,$63,$0F
	.dc.b	$E5,$CF,$5E,$A8,$CF,$47,$2A,$E3,$5F,$73,$0F,$C9,$E7,$FD,$C7,$0F
	.dc.b	$C8,$F8,$3F,$E3,$DC,$3F,$1A,$DB,$FE,$49,$CD,$FC,$4A,$BC,$11,$FE
	.dc.b	$43,$8C,$57,$A3,$E6,$7F,$C6,$79,$9A,$4A,$FC,$61,$F2,$39,$AE,$1F
	.dc.b	$A2,$F9,$AF,$D0,$91,$FD,$0B,$F8,$7E,$0F,$9B,$A7,$CF,$25,$74,$BF
	.dc.b	$C1,$9F,$E6,$6F,$9B,$7B,$1F,$E5,$9D,$BF,$9B,$4A,$73,$97,$F2,$B8
	.dc.b	$FE,$C9,$18,$9D,$B6,$3F,$07,$BF,$E2,$B8,$FE,$D1,$23,$27,$F1,$93
	.dc.b	$F2,$F9,$79,$A7,$FC,$73,$E4,$3C,$87,$CE,$7F,$99,$23,$0D,$8C,$DF
	.dc.b	$FB,$4A,$72,$D8,$6F,$F9,$CF,$C3,$3F,$A2,$B6,$7F,$A5,$35,$FD,$0D
	.dc.b	$78,$63,$11,$F6,$FE,$7B,$F8,$89,$36,$FE,$22,$FF,$A1,$EB,$D1,$E0
	.dc.b	$FE,$AB,$F6,$1F,$EA,$6D,$BF,$EB,$0D,$7B,$E5,$FC,$A9,$4F,$FB,$A3
	.dc.b	$B6,$47,$FB,$65,$AF,$F8,$57,$D7,$7D,$0F,$EE,$27,$EC,$C6,$27,$F8
	.dc.b	$69,$B9,$D3,$FD,$CE,$35,$3F,$DD,$7E,$1B,$FB,$DE,$7A,$AF,$E0,$71
	.dc.b	$87,$99,$FE,$95,$F4,$93,$B8,$C3,$57,$47,$FB,$FB,$BB,$F1,$AD,$8F
	.dc.b	$F7,$A4,$63,$91,$D3,$FC,$02,$7C,$DA,$7F,$04,$69,$93,$D8,$53,$F8
	.dc.b	$CE,$9F,$06,$FF,$37,$0F,$88,$FE,$58,$DE,$27,$36,$A7,$27,$F3,$1F
	.dc.b	$8F,$7C,$4B,$E4,$39,$C6,$1B,$19,$C3,$F5,$33,$97,$C3,$F5,$5F,$EE
	.dc.b	$77,$FC,$98,$D7,$F9,$7B,$53,$98,$FF,$07,$C7,$53,$FC,$A8,$E2,$BB
	.dc.b	$FA,$5F,$CE,$1B,$3B,$3E,$1F,$DC,$E5,$FC,$03,$5F,$C5,$35,$D9,$C4
	.dc.b	$FB,$3C,$1F,$52,$7E,$B7,$06,$78,$53,$9A,$FF,$1F,$46,$27,$97,$03
	.dc.b	$E3,$BF,$DC,$D8,$7E,$59,$B7,$F9,$36,$BF,$9D,$6B,$F7,$3E,$4A,$AA
	.dc.b	$46,$1B,$C3,$E0,$8E,$3D,$6F,$90,$F9,$7F,$92,$7F,$93,$DB,$87,$F9
	.dc.b	$39,$AF,$F0,$26,$BF,$EF,$DF,$94,$7C,$67,$DB,$C5,$18,$7F,$94,$98
	.dc.b	$8D,$46,$67,$FC,$C5,$3C,$3C,$D2,$FC,$2B,$35,$FD,$FD,$B3,$A6,$CE
	.dc.b	$EB,$FC,$E9,$4F,$89,$EE,$49,$F1,$C6,$5F,$D1,$F0,$FD,$5A,$FB,$64
	.dc.b	$3E,$CF,$17,$FB,$C1,$FF,$3C,$7E,$0C,$FA,$FF,$C3,$42,$30,$FE,$1D
	.dc.b	$BC,$7F,$D2,$9B,$1F,$E8,$1F,$AE,$7D,$7F,$C0,$D3,$FA,$7C,$63,$3F
	.dc.b	$55,$6E,$B4,$D8,$CB,$FE,$B5,$CF,$D4,$FA,$39,$BD,$BA,$7A,$63,$4D
	.dc.b	$7F,$C3,$B8,$77,$BF,$0E,$1F,$49,$A7,$ED,$E3,$53,$FE,$99,$FB,$DF
	.dc.b	$E0,$EB,$E8,$3F,$B3,$C6,$1B,$DF,$FE,$2A,$6C,$EC,$F3,$1F,$3C,$F2
	.dc.b	$E5,$E8,$3F,$8B,$F9,$AF,$B8,$FC,$2D,$7F,$6E,$6D,$7F,$AA,$E7,$59
	.dc.b	$75,$8F,$EE,$B3,$A7,$FD,$B5,$F8,$43,$EF,$7F,$C4,$71,$FE,$B3,$8C
	.dc.b	$3F,$C5,$5B,$5F,$E0,$26,$C7,$A0,$87,$A0,$F5,$1F,$33,$5F,$94,$FD
	.dc.b	$79,$F9,$64,$7F,$D8,$58,$F8,$1F,$F0,$7B,$5C,$DF,$F3,$87,$E7,$37
	.dc.b	$7F,$BF,$1E,$13,$C6,$7F,$96,$1C,$BF,$47,$AF,$F2,$F5,$FC,$1A,$3F
	.dc.b	$8E,$37,$8F,$C3,$3F,$9C,$F6,$7F,$CD,$D5,$FE,$6C,$E8,$FE,$6D,$18
	.dc.b	$8F,$F3,$56,$5F,$5D,$F8,$27,$F2,$6D,$F0,$E8,$E5,$1F,$D4,$A7,$94
	.dc.b	$7F,$9C,$2B,$A9,$27,$4D,$8E,$B9,$7B,$1E,$C7,$82,$B9,$BF,$E7,$86
	.dc.b	$BF,$E7,$9A,$E6,$7E,$33,$E9,$FF,$A8,$70,$1A,$48,$F4,$CF,$CD,$21
	.dc.b	$F3,$79,$3F,$54,$FF,$91,$67,$83,$FE,$42,$6B,$FE,$FF,$F8,$2F,$F6
	.dc.b	$47,$C8,$FF,$B3,$FD,$9F,$F4,$6B,$64,$D3,$DF,$7D,$B3,$F8,$07,$C1
	.dc.b	$F8,$D7,$C4,$FF,$69,$6E,$7F,$DF,$9B,$87,$F9,$69,$AF,$F2,$6F,$B9
	.dc.b	$FE,$5F,$F9,$47,$F2,$E6,$C6,$69,$39,$6C,$7F,$45,$E7,$FF,$52,$FD
	.dc.b	$17,$F4,$96,$C6,$44,$65,$A7,$F3,$6E,$8F,$E7,$FF,$94,$FF,$3E,$6C
	.dc.b	$66,$D3,$96,$C7,$F5,$5E,$8F,$F5,$AF,$D5,$7F,$59,$6C,$64,$46,$5A
	.dc.b	$7F,$5D,$F8,$9F,$F5,$EF,$E7,$1F,$D8,$5B,$19,$11,$96,$9F,$EC,$43
	.dc.b	$FD,$8B,$B7,$D0,$F1,$3E,$77,$07,$F6,$59,$D0,$9D,$38,$6B,$FE,$B4
	.dc.b	$97,$E2,$1F,$D8,$BF,$0B,$FE,$C2,$D9,$FE,$BA,$D7,$83,$5F,$F6,$BF
	.dc.b	$3F,$F4,$8F,$8A,$FF,$4C,$B6,$3F,$25,$E2,$FF,$B7,$27,$2D,$8F,$26
	.dc.b	$5F,$8B,$26,$FF,$8B,$7F,$C7,$1E,$67,$F8,$27,$CB,$FF,$72,$FE,$33
	.dc.b	$FE,$09,$6D,$FF,$82,$5A,$FF,$82,$5A,$FF,$82,$7E,$13,$FB,$FF,$E1
	.dc.b	$1F,$DF,$9B,$1F,$DE,$9A,$FC,$56,$BE,$EE,$EF,$E7,$2F,$D3,$D5,$18
	.dc.b	$8D,$1F,$E4,$F8,$CB,$FE,$51,$EC,$FD,$5F,$6F,$EB,$49,$E1,$FF,$99
	.dc.b	$BE,$5F,$F8,$25,$7F,$90,$5B,$E3,$7F,$B1,$1B,$1F,$79,$AF,$DB,$EC
	.dc.b	$FF,$0D,$FE,$29,$F9,$92,$47,$E7,$5F,$CA,$7E,$71,$23,$F3,$0D,$7D
	.dc.b	$9C,$87,$FB,$C6,$9F,$E0,$4C,$BE,$CC,$62,$75,$1F,$ED,$8F,$D9,$5F
	.dc.b	$55,$FE,$16,$6D,$7F,$96,$DB,$A6,$E7,$FC,$37,$87,$C6,$7F,$8A,$12
	.dc.b	$BF,$8D,$6F,$E3,$8F,$E2,$B8,$7F,$26,$9F,$E4,$8F,$FA,$3B,$F6,$CF
	.dc.b	$F2,$77,$3F,$F9,$A5,$B8,$FF,$BB,$1B,$5F,$A6,$6B,$F0,$0F,$F4,$B1
	.dc.b	$F9,$82,$99,$F7,$7E,$09,$E9,$5F,$71,$48,$FE,$CB,$4F,$83,$1F,$23
	.dc.b	$EB,$24,$F5,$6B,$FB,$DB,$F4,$FC,$27,$9F,$A3,$FC,$AA,$DC,$CD,$3E
	.dc.b	$1D,$43,$F8,$34,$3F,$2E,$7C,$A7,$F9,$9D,$B1,$FE,$66,$6C,$FF,$99
	.dc.b	$1A,$F2,$7F,$DB,$B0,$FB,$ED,$7C,$2F,$F9,$A1,$B1,$9A,$E1,$39,$79
	.dc.b	$7E,$17,$E1,$24,$F8,$5F,$F5,$97,$5F,$E9,$38,$BE,$E5,$7F,$6E,$3C
	.dc.b	$5F,$01,$E9,$25,$FA,$5E,$3D,$9F,$AA,$9F,$84,$7F,$BA,$3E,$19,$F5
	.dc.b	$FF,$EC,$1D,$D1,$8C,$3F,$93,$D7,$CC,$8C,$3F,$EF,$36,$BF,$E9,$5F
	.dc.b	$C3,$BF,$49,$B7,$FA,$65,$B1,$93,$E1,$1F,$E9,$97,$2F,$EA,$EB,$EA
	.dc.b	$7B,$7F,$4C,$8F,$D3,$7D,$7A,$7C,$2D,$7E,$25,$7D,$5F,$17,$ED,$12
	.dc.b	$FE,$D3,$FE,$A3,$FD,$73,$F5,$FE,$D7,$FA,$72,$BC,$11,$8A,$FC,$2B
	.dc.b	$63,$3D,$9F,$BF,$AE,$11,$FA,$C3,$E5,$94,$FE,$A5,$EB,$3F,$62,$7C
	.dc.b	$CF,$7C,$73,$18,$7F,$93,$CE,$E7,$55,$FE,$12,$AD,$3E,$0A,$63,$FC
	.dc.b	$23,$D3,$C3,$63,$D3,$97,$F9,$F5,$B5,$FE,$83,$7F,$CA,$6F,$E2,$4F
	.dc.b	$F1,$49,$4F,$F5,$FF,$CC,$BE,$23,$FC,$FB,$3F,$A8,$8C,$3C,$E7,$EC
	.dc.b	$46,$30,$F3,$36,$7F,$BA,$7D,$A3,$ED,$1F,$3B,$EE,$C1,$39,$8C,$36
	.dc.b	$BF,$A5,$B6,$7E,$8A,$F2,$9F,$1F,$FB,$C9,$FE,$69,$39,$8C,$47,$69
	.dc.b	$DB,$D1,$AF,$DC,$3E,$11,$38,$FF,$1A,$FA,$B1,$F4,$E1,$E4,$6C,$79
	.dc.b	$1B,$1B,$C7,$DD,$3C,$FA,$FF,$64,$FD,$2B,$EA,$FD,$27,$FC,$D5,$F1
	.dc.b	$DE,$CF,$23,$F2,$1D,$1F,$E1,$D8,$C5,$39,$BF,$D7,$EA,$33,$18,$6C
	.dc.b	$F1,$D5,$7D,$2F,$FA,$D5,$69,$B3,$A7,$B5,$71,$AE,$11,$FE,$8B,$8F
	.dc.b	$E0,$B8,$31,$74,$79,$2B,$FA,$8D,$6A,$72,$D9,$D4,$F3,$3F,$DC,$D1
	.dc.b	$93,$FC,$6F,$BF,$E0,$DF,$9D,$A3,$54,$61,$A7,$D1,$2D,$7F,$B6,$A3
	.dc.b	$A3,$D8,$87,$E2,$37,$FC,$2C,$7E,$39,$B3,$F9,$A6,$CE,$F4,$FF,$01
	.dc.b	$57,$84,$95,$E1,$6C,$FE,$9B,$8B,$D1,$F5,$FF,$A6,$6D,$7E,$B5,$B3
	.dc.b	$B8,$4F,$E5,$75,$F9,$DA,$7C,$86,$E1,$EC,$BF,$71,$FC,$02,$BF,$82
	.dc.b	$F2,$7F,$85,$DB,$1D,$30,$D1,$FC,$53,$9F,$E6,$5F,$F1,$EE,$8F,$E2
	.dc.b	$93,$F2,$AB,$FD,$47,$D1,$FC,$66,$B4,$D9,$F9,$73,$F8,$02,$9F,$74
	.dc.b	$FE,$56,$3E,$3B,$C6,$48,$FE,$2E,$FF,$1F,$F0,$FF,$D5,$37,$FC,$F6
	.dc.b	$7F,$18,$7F,$79,$EB,$F3,$DF,$F7,$36,$BF,$08,$87,$8C,$B1,$FD,$38
	.dc.b	$E7,$5F,$88,$AF,$EF,$C7,$29,$FC,$74,$65,$FE,$B5,$F2,$FF,$D7,$53
	.dc.b	$48,$FF,$A4,$FC,$0F,$ED,$7E,$0F,$36,$E1,$FA,$56,$E1,$FD,$AE,$7C
	.dc.b	$15,$F1,$CF,$41,$0C,$57,$F8,$67,$C2,$F3,$34,$F5,$B8,$C4,$F4,$E1
	.dc.b	$FC,$3D,$7D,$46,$BF,$12,$5E,$1F,$12,$5F,$C6,$7A,$78,$9F,$B1,$D3
	.dc.b	$FC,$3A,$4B,$FF,$0E,$D7,$C4,$7F,$C4,$DE,$2F,$8C,$91,$F4,$1F,$27
	.dc.b	$5F,$F0,$49,$F2,$35,$FB,$23,$C7,$FF,$2C,$FE,$87,$FE,$3F,$24,$7F
	.dc.b	$95,$FC,$8F,$E0,$91,$FE,$72,$F8,$89,$3F,$E5,$76,$BF,$EC,$7F,$33
	.dc.b	$FC,$E3,$F2,$BF,$E5,$72,$47,$F9,$57,$AF,$FC,$FD,$1F,$E7,$CF,$8E
	.dc.b	$93,$FE,$61,$6B,$FE,$55,$C7,$F8,$A7,$99,$FE,$53,$24,$79,$76,$FF
	.dc.b	$16,$BE,$EE,$EF,$DD,$6D,$EF,$8F,$BD,$E8,$FF,$A9,$3D,$1F,$F6,$39
	.dc.b	$FE,$8F,$2C,$7F,$AA,$0F,$F5,$73,$97,$D0,$76,$3D,$DA,$7F,$A5,$F1
	.dc.b	$FF,$3A,$12,$3F,$CF,$5D,$7F,$E9,$82,$38,$3F,$DF,$1F,$BD,$7F,$BD
	.dc.b	$C8,$E0,$FF,$35,$FB,$CF,$5B,$F9,$A7,$EA,$9F,$EE,$ED,$93,$95,$19
	.dc.b	$8C,$36,$78,$D7,$FA,$E7,$E2,$7D,$0D,$3D,$87,$FA,$C7,$D9,$FF,$61
	.dc.b	$E7,$FE,$C1,$9D,$CF,$E2,$3B,$7F,$3A,$FE,$A5,$AB,$7F,$F6,$8F,$E0
	.dc.b	$FF,$30,$91,$FE,$AC,$0F,$FB,$8F,$47,$E1,$D4,$93,$FE,$E5,$8C,$BF
	.dc.b	$EE,$F1,$FA,$87,$FB,$CD,$A6,$5C,$1F,$E6,$EF,$84,$FE,$3F,$1F,$F0
	.dc.b	$02,$7F,$B2,$C6,$23,$FD,$A0,$1C,$3F,$ED,$3C,$1E,$E7,$FB,$57,$F1
	.dc.b	$2F,$F6,$99,$27,$FD,$70,$FD,$1F,$84,$FE,$D5,$FB,$3D,$C4,$D2,$57
	.dc.b	$8F,$83,$CD,$CB,$FD,$C5,$B4,$69,$E6,$97,$C2,$3A,$DF,$11,$FE,$E9
	.dc.b	$FD,$EF,$53,$98,$C5,$7B,$A3,$0A,$77,$5C,$A3,$06,$D7,$B9,$FE,$F5
	.dc.b	$FB,$4D,$B6,$74,$D7,$FD,$F5,$7E,$58,$C4,$7F,$83,$7A,$BF,$31,$FB
	.dc.b	$5F,$E5,$74,$34,$93,$E1,$F8,$F2,$FB,$AF,$3C,$61,$B1,$93,$81,$4F
	.dc.b	$F5,$15,$BE,$4B,$FC,$3D,$FB,$5F,$E9,$DB,$3F,$B0,$6B,$C1,$AF,$8B
	.dc.b	$9F,$F0,$AF,$E1,$F8,$BF,$A6,$C6,$0C,$F5,$FE,$90,$9E,$1F,$F1,$3A
	.dc.b	$FE,$A4,$7E,$9F,$C3,$F6,$FD,$2F,$F1,$22,$47,$E9,$DF,$F6,$3F,$CA
	.dc.b	$BC,$0F,$EB,$6D,$E1,$FF,$1F,$B7,$D4,$DC,$FF,$42,$F1,$FF,$91,$92
	.dc.b	$3F,$C3,$98,$7C,$CA,$E0,$FE,$35,$AF,$E9,$1A,$EE,$BF,$B5,$FC,$B3
	.dc.b	$E5,$E9,$31,$FF,$28,$C6,$5F,$05,$7F,$35,$C3,$C2,$A7,$51,$FE,$E4
	.dc.b	$6C,$7E,$7D,$AF,$ED,$9A,$FE,$11,$AF,$3A,$F1,$4F,$37,$E1,$3E,$0D
	.dc.b	$7E,$9E,$5E,$69,$DD,$77,$6B,$FC,$13,$9B,$FD,$08,$FF,$26,$3F,$47
	.dc.b	$F0,$9F,$C9,$1F,$CA,$C6,$BC,$6F,$F7,$EB,$63,$E0,$10,$E9,$E7,$7F
	.dc.b	$97,$9F,$DE,$42,$31,$1A,$3F,$CF,$71,$97,$FC,$FB,$F3,$5F,$D0,$7C
	.dc.b	$8F,$E0,$90,$8C,$3F,$C1,$DA,$FF,$1E,$6B,$FC,$E7,$8F,$F2,$9A,$7C
	.dc.b	$4D,$79,$F5,$77,$C2,$E5,$FB,$26,$E1,$FC,$E8,$FE,$29,$3B,$7F,$15
	.dc.b	$F3,$3F,$E7,$46,$F0,$FF,$9C,$1B,$EF,$FE,$72,$75,$1F,$D3,$73,$F1
	.dc.b	$93,$87,$FD,$37,$F3,$5F,$E9,$49,$E4,$7F,$38,$71,$7F,$09,$B1,$FC
	.dc.b	$9D,$AF,$C1,$F3,$3F,$D3,$D9,$FC,$8D,$FE,$1C,$FC,$99,$CB,$F2,$5A
	.dc.b	$F9,$70,$FC,$DB,$FE,$2A,$9F,$EE,$5F,$5B,$CE,$BF,$D5,$4D,$D0,$D3
	.dc.b	$EB,$FC,$E3,$D2,$FC,$56,$08,$C3,$F8,$F6,$F2,$70,$9D,$F1,$FD,$7A
	.dc.b	$FC,$61,$FC,$B7,$CC,$FF,$2C,$F8,$3F,$D6,$32,$FF,$3E,$B7,$97,$FC
	.dc.b	$0D,$3A,$E8,$FF,$60,$7E,$D3,$FE,$09,$6E,$1F,$E1,$26,$BF,$CA,$9A
	.dc.b	$FF,$3B,$F9,$FE,$E4,$C7,$BB,$FE,$7F,$F1,$BF,$AB,$7D,$0F,$F3,$AF
	.dc.b	$83,$FE,$9E,$DB,$FE,$9C,$D7,$FA,$4F,$94,$8C,$47,$F4,$2F,$83,$FF
	.dc.b	$7B,$63,$FB,$36,$F0,$FE,$89,$B3,$FD,$9D,$AF,$F7,$E7,$FC,$F3,$3C
	.dc.b	$7E,$07,$FC,$9B,$3B,$6B,$B6,$BF,$D4,$B8,$7F,$D8,$F7,$FE,$EF,$24
	.dc.b	$7F,$75,$8C,$BE,$1F,$86,$FB,$B8,$1A,$48,$FE,$19,$F1,$9E,$F6,$E5
	.dc.b	$F8,$48,$C6,$1F,$BE,$6C,$FF,$00,$FA,$2F,$F4,$EA,$69,$33,$FD,$44
	.dc.b	$7F,$87,$A5,$F1,$2F,$8D,$FE,$13,$29,$CC,$61,$B1,$E1,$9E,$0F,$0F
	.dc.b	$93,$5F,$28,$46,$1D,$46,$63,$11,$FC,$61,$AF,$F8,$75,$AF,$F8,$D1
	.dc.b	$AF,$F8,$09,$BA,$3C,$DF,$F0,$2E,$3F,$81,$6D,$7E,$05,$AF,$E0,$5A
	.dc.b	$FE,$2F,$9E,$BF,$EE,$96,$D7,$F0,$FA,$FC,$64,$6B,$77,$84,$EF,$F8
	.dc.b	$78,$7F,$92,$A7,$31,$88,$FC,$44,$EE,$FF,$9B,$D6,$9F,$CC,$71,$7C
	.dc.b	$A6,$E9,$C6,$FF,$DF,$93,$97,$F2,$D1,$D7,$97,$FC,$85,$B7,$F8,$9B
	.dc.b	$F2,$EF,$F1,$DB,$6B,$FC,$74,$D7,$F5,$AD,$7F,$DE,$7F,$0B,$F4,$2D
	.dc.b	$F1,$1F,$EF,$27,$FA,$69,$88,$D4,$66,$7F,$C9,$3F,$81,$7F,$8D,$BF
	.dc.b	$20,$B0,$69,$1F,$F2,$67,$8B,$F9,$0C,$FF,$CB,$4D,$EA,$FE,$99,$3F
	.dc.b	$E5,$48,$FC,$66,$FF,$32,$3E,$4F,$07,$D1,$6F,$0F,$E6,$A7,$51,$FB
	.dc.b	$39,$D3,$67,$51,$88,$FB,$0D,$5F,$18,$FA,$3F,$D6,$6F,$FB,$99,$FF
	.dc.b	$43,$CE,$7A,$7F,$43,$D9,$FE,$75,$8C,$56,$EB,$FC,$B3,$18,$9F,$3C
	.dc.b	$43,$C8,$53,$FD,$27,$F9,$63,$F0,$4F,$D6,$FC,$87,$F7,$E8,$C6,$08
	.dc.b	$C4,$EA,$BF,$D0,$D5,$A9,$FF,$64,$71,$BE,$EB,$F6,$FE,$8E,$1B,$7F
	.dc.b	$EA,$2C,$7F,$03,$18,$8F,$F3,$8C,$60,$FF,$41,$BE,$82,$1F,$2B,$7F
	.dc.b	$9A,$7E,$23,$F1,$77,$D9,$7F,$89,$A7,$B4,$61,$E6,$76,$71,$87,$FB
	.dc.b	$7D,$B1,$E5,$FA,$EF,$81,$87,$F0,$80,$E1,$FF,$73,$36,$38,$67,$FC
	.dc.b	$5A,$7E,$74,$7E,$D3,$1F,$3B,$88,$5B,$E5,$7E,$0C,$FA,$1A,$F9,$DA
	.dc.b	$FF,$22,$8E,$0F,$F1,$AE,$4F,$C9,$BF,$14,$F0,$87,$2F,$8C,$CB,$87
	.dc.b	$5C,$9F,$95,$6E,$1E,$87,$E8,$9C,$4F,$8D,$F7,$3F,$BB,$CC,$D2,$65
	.dc.b	$C6,$7C,$87,$23,$A9,$F2,$A3,$6D,$8F,$E0,$73,$D1,$FF,$28,$F4,$7F
	.dc.b	$B4,$FF,$01,$FE,$5A,$DA,$FE,$2D,$D2,$9C,$B6,$3F,$8F,$67,$EF,$26
	.dc.b	$7E,$F7,$E2,$DF,$FB,$7F,$C2,$FE,$0C,$4A,$F0,$DA,$3F,$AD,$E3,$F8
	.dc.b	$4E,$0F,$84,$47,$D2,$FF,$4C,$C3,$BE,$BF,$0F,$0F,$94,$7F,$84,$BF
	.dc.b	$60,$FE,$28,$47,$BB,$FC,$67,$5F,$EA,$50,$FA,$7F,$F7,$8C,$61,$BE
	.dc.b	$2F,$FB,$69,$B3,$B3,$E8,$61,$E2,$7F,$03,$B7,$C1,$26,$3E,$5D,$BF
	.dc.b	$E0,$07,$E2,$DF,$EE,$4F,$C8,$7F,$DF,$2D,$BF,$F7,$C3,$5F,$F1,$5B
	.dc.b	$5F,$EE,$F9,$FF,$80,$B6,$7E,$81,$FA,$8F,$13,$FB,$DB,$63,$35,$FB
	.dc.b	$19,$CB,$FE,$F4,$18,$B8,$9F,$3D,$FE,$F8,$FC,$63,$55,$A9,$CA,$8C
	.dc.b	$D7,$1A,$E1,$3D,$63,$0E,$CC,$57,$F0,$9F,$0B,$8B,$63,$3D,$67,$F9
	.dc.b	$D1,$82,$9F,$E4,$EF,$C9,$75,$FA,$0F,$26,$C7,$66,$BB,$11,$F5,$B8
	.dc.b	$7F,$11,$1F,$A8,$3C,$BD,$9E,$1F,$CD,$7C,$AD,$8C,$88,$CB,$4F,$6F
	.dc.b	$85,$FE,$F5,$68,$F8,$5F,$7C,$66,$3F,$CC,$ED,$7F,$32,$43,$C8,$7E
	.dc.b	$F8,$F8,$9A,$BF,$3D,$FB,$5F,$9D,$FD,$26,$CF,$D1,$6B,$C1,$AF,$E2
	.dc.b	$B8,$3F,$C8,$3F,$A4,$FF,$30,$6C,$7F,$37,$6B,$FD,$2B,$9B,$F0,$BC
	.dc.b	$AA,$FD,$1E,$D2,$BF,$01,$5F,$AC,$8C,$47,$A4,$F6,$75,$7E,$9D,$FD
	.dc.b	$4F,$57,$E9,$9F,$E7,$BA,$7E,$23,$C9,$FC,$2E,$7F,$83,$6C,$FF,$99
	.dc.b	$CF,$C5,$C6,$5F,$C3,$78,$5F,$8C,$E5,$FC,$53,$63,$F9,$59,$A3,$EA
	.dc.b	$D7,$D4,$7A,$1F,$50,$A7,$E8,$78,$3B,$3E,$B3,$EA,$9E,$CC,$7F,$C2
	.dc.b	$58,$FF,$5C,$EE,$FC,$D7,$D6,$FF,$08,$25,$7C,$68,$CB,$F0,$B5,$FC
	.dc.b	$9E,$BF,$16,$3B,$35,$F1,$0F,$AC,$F4,$11,$FC,$7B,$F9,$D6,$1F,$92
	.dc.b	$F9,$EF,$71,$1C,$1E,$13,$67,$83,$2F,$A8,$E4,$87,$DB,$CB,$F2,$58
	.dc.b	$7E,$DF,$F1,$AF,$C9,$36,$7F,$24,$D7,$F2,$4D,$7F,$D2,$3D,$7F,$E9
	.dc.b	$0E,$47,$E2,$DF,$C4,$1F,$E2,$9E,$BF,$FA,$5A,$31,$3B,$6C,$66,$BC
	.dc.b	$A7,$D3,$BF,$97,$97,$AA,$3D,$91,$82,$3F,$00,$FD,$8E,$1F,$C8,$FE
	.dc.b	$37,$FD,$4B,$E8,$78,$1B,$3F,$E2,$FC,$91,$88,$FF,$17,$C6,$2B,$FC
	.dc.b	$A7,$1B,$DB,$FC,$ED,$3F,$66,$7F,$CE,$AD,$9F,$6C,$FF,$5D,$C7,$D1
	.dc.b	$1E,$B3,$CC,$7E,$40,$FC,$8E,$BF,$C0,$F8,$7F,$D5,$D8,$7F,$01,$6C
	.dc.b	$FF,$01,$6B,$FD,$03,$A5,$18,$8E,$9C,$3F,$8F,$6E,$1E,$26,$EB,$FE
	.dc.b	$08,$BE,$30,$F7,$7F,$D8,$B9,$78,$E3,$19,$78,$DA,$78,$CA,$7B,$4F
	.dc.b	$B2,$FF,$53,$F7,$BF,$DA,$BF,$19,$FE,$D0,$8C,$07,$1B,$27,$73,$F5
	.dc.b	$5B,$5F,$53,$E0,$7F,$DB,$93,$F6,$5B,$19,$8F,$F2,$29,$FD,$8E,$33
	.dc.b	$1F,$1F,$A3,$FC,$DC,$FF,$2B,$6B,$93,$FC,$9B,$8F,$F3,$58,$CC,$7C
	.dc.b	$83,$F1,$D5,$D4,$EF,$82,$FB,$C6,$67,$DB,$1E,$27,$91,$F9,$E2,$9F
	.dc.b	$9C,$3F,$DF,$DE,$DF,$F6,$9C,$9E,$E7,$F7,$FF,$75,$E4,$AB,$53,$98
	.dc.b	$C5,$FC,$D8,$C4,$EA,$3F,$BA,$57,$28,$C4,$23,$0F,$13,$75,$D2,$FC
	.dc.b	$71,$8A,$E3,$C5,$FE,$F0,$C3,$FD,$86,$49,$FF,$60,$35,$FF,$4B,$35
	.dc.b	$FF,$41,$3F,$C1,$31,$65,$F0,$0F,$BC,$59,$FB,$D3,$F3,$23,$11,$C8
	.dc.b	$EE,$43,$B9,$FD,$AF,$4F,$F0,$A4,$FE,$7B,$83,$FA,$93,$63,$35,$BA
	.dc.b	$EA,$FD,$F9,$F0,$11,$F2,$BF,$9A,$DF,$D3,$E8,$59,$FF,$32,$1F,$9B
	.dc.b	$2D,$7F,$0E,$9D,$C7,$9E,$3D,$47,$A0,$87,$EF,$F9,$3F,$C5,$9F,$80
	.dc.b	$FA,$5B,$3F,$B5,$8C,$49,$F9,$6F,$23,$FD,$33,$87,$D1,$69,$E1,$29
	.dc.b	$BD,$3E,$E3,$FC,$07,$E3,$7A,$92,$F0,$71,$9D,$CF,$1E,$67,$37,$E2
	.dc.b	$23,$0D,$7D,$E4,$AF,$F2,$D7,$3F,$E0,$B1,$FE,$57,$1B,$32,$5E,$5F
	.dc.b	$E6,$51,$FC,$F6,$1F,$49,$49,$7F,$CF,$FA,$7F,$1B,$1F,$07,$3F,$3C
	.dc.b	$F8,$8F,$82,$57,$2E,$08,$E0,$99,$FB,$B0,$E8,$D9,$F8,$8D,$7E,$DE
	.dc.b	$1F,$D9,$07,$E0,$3F,$D8,$9B,$1F,$D5,$7B,$3F,$47,$39,$3F,$BA,$C3
	.dc.b	$EF,$E9,$1F,$EE,$E6,$4F,$0E,$DF,$3D,$F4,$E1,$E3,$7D,$CD,$73,$1B
	.dc.b	$CB,$DA,$D9,$D4,$7C,$8E,$BF,$F4,$87,$85,$FE,$7F,$F9,$3F,$F1,$91
	.dc.b	$30,$FF,$34,$BF,$13,$0E,$44,$D7,$FC,$71,$F0,$32,$BF,$CD,$ED,$7F
	.dc.b	$0F,$1C,$48,$FB,$1F,$F5,$DF,$C4,$C8,$FF,$1A,$07,$0F,$23,$81,$FE
	.dc.b	$7C,$FB,$EF,$F2,$41,$1C,$1F,$E9,$7F,$96,$FF,$28,$11,$C1,$FE,$58
	.dc.b	$11,$FE,$5C,$23,$97,$CE,$F8,$48,$7D,$0C,$3F,$45,$9F,$F9,$3E,$3F
	.dc.b	$DB,$67,$FA,$1B,$CB,$FF,$6E,$F9,$1F,$ED,$82,$47,$F9,$F7,$8D,$FB
	.dc.b	$F7,$FA,$05,$B1,$98,$C1,$FE,$8D,$7F,$D6,$9F,$88,$54,$D2,$3F,$18
	.dc.b	$F3,$E1,$F7,$97,$EA,$BF,$E6,$F2,$3F,$EA,$67,$FD,$57,$FB,$2C,$8F
	.dc.b	$F5,$7B,$5F,$82,$7E,$3B,$7F,$F6,$02,$FD,$DA,$91,$FF,$61,$19,$3F
	.dc.b	$D9,$5F,$39,$FE,$83,$23,$83,$EA,$7E,$DF,$FE,$8E,$23,$83,$FD,$D5
	.dc.b	$A3,$F7,$6F,$F7,$71,$23,$2F,$FB,$F1,$7E,$27,$48,$FF,$BF,$8C,$9F
	.dc.b	$1C,$FB,$98,$FE,$01,$A7,$E5,$3E,$A3,$FD,$DF,$C2,$E1,$FF,$73,$46
	.dc.b	$1B,$3B,$3F,$DD,$FF,$07,$FD,$61,$A7,$E2,$3C,$14,$66,$3F,$D9,$B1
	.dc.b	$87,$F5,$C7,$94,$FE,$3B,$9B,$07,$C2,$7F,$2F,$9F,$F7,$C1,$FD,$CA
	.dc.b	$33,$3F,$CC,$5B,$97,$F9,$F5,$B3,$C3,$89,$83,$C8,$F2,$4A,$73,$18
	.dc.b	$6C,$7A,$E3,$6F,$E1,$4F,$99,$1D,$3E,$15,$0F,$91,$F3,$BD,$6D,$3A
	.dc.b	$9C,$D7,$1A,$E0,$79,$EB,$6E,$0F,$F7,$37,$AF,$F0,$DB,$9F,$C1,$8C
	.dc.b	$47,$C5,$6B,$E7,$3F,$7B,$1F,$36,$36,$D3,$91,$33,$FC,$34,$F1,$3F
	.dc.b	$0E,$7E,$9B,$C0,$FE,$63,$1F,$8B,$97,$C3,$BD,$FF,$87,$F1,$BF,$EA
	.dc.b	$F8,$C6,$08,$C4,$EA,$3E,$ED,$72,$8C,$47,$7F,$13,$F9,$E9,$FD,$16
	.dc.b	$7F,$0D,$F1,$3F,$E3,$4B,$FC,$03,$6B,$9F,$0A,$31,$1F,$5A,$B8,$9E
	.dc.b	$37,$C3,$8B,$DC,$FB,$7F,$37,$E9,$6C,$E9,$AF,$39,$E0,$A7,$E4,$EB
	.dc.b	$F1,$A3,$DB,$F1,$8F,$8C,$FF,$25,$D7,$FB,$4E,$31,$8F,$1A,$DD,$69
	.dc.b	$B1,$97,$F4,$7F,$1A,$F2,$BF,$CA,$BF,$3F,$F9,$09,$4E,$69,$39,$8C
	.dc.b	$36,$B8,$D7,$EC,$F2,$F4,$3F,$9C,$F7,$9F,$12,$F6,$3F,$CC,$77,$F9
	.dc.b	$E9,$CC,$62,$BF,$6F,$18,$CB,$CD,$18,$BE,$51,$88,$FB,$DE,$4F,$F0
	.dc.b	$46,$F3,$7F,$9C,$FA,$7F,$73,$CB,$FB,$5E,$9F,$B1,$3F,$AA,$6D,$7E
	.dc.b	$A6,$75,$F2,$3F,$B4,$FD,$BB,$4D,$9F,$D7,$35,$FE,$2A,$D7,$FC,$FF
	.dc.b	$C6,$F3,$FF,$4E,$DD,$3F,$C8,$0D,$CB,$ED,$4F,$EC,$B3,$FE,$1F,$EF
	.dc.b	$FF,$04,$FD,$9F,$CC,$DC,$7F,$7C,$D7,$F9,$2B,$5F,$E1,$7E,$0B,$E5
	.dc.b	$7F,$D2,$3E,$0F,$DD,$8C,$69,$FC,$16,$B7,$5A,$7F,$88,$CE,$AB,$F0
	.dc.b	$9C,$CE,$FF,$F3,$0C,$61,$B5,$FC,$62,$75,$C7,$D6,$11,$8A,$F1,$6D
	.dc.b	$FC,$6E,$7B,$B5,$FE,$6F,$87,$FA,$D2,$77,$1C,$E3,$FA,$73,$7C,$5E
	.dc.b	$0F,$C0,$6B,$FC,$DD,$AB,$D6,$FE,$3A,$DD,$3F,$C9,$4D,$C3,$F4,$4D
	.dc.b	$3F,$49,$3C,$09,$3C,$1F,$E4,$FF,$05,$FE,$CE,$9F,$E7,$2D,$7F,$DA
	.dc.b	$E7,$FB,$88,$EB,$EC,$FF,$A6,$9F,$D8,$35,$FE,$82,$D7,$FA,$CB,$5F
	.dc.b	$E8,$ED,$4D,$25,$7A,$A3,$2D,$8C,$C7,$8A,$3D,$F8,$BE,$D3,$FA,$9F
	.dc.b	$C1,$7E,$1D,$B5,$F8,$46,$BF,$59,$AF,$E4,$32,$AF,$F6,$76,$D7,$FB
	.dc.b	$BA,$BE,$D9,$F8,$9D,$BF,$DF,$B5,$FD,$DB,$6F,$C4,$CE,$49,$1F,$DC
	.dc.b	$FE,$07,$ED,$2F,$A3,$FE,$C7,$18,$6E,$DF,$DB,$1B,$3B,$3F,$A3,$F5
	.dc.b	$BD,$8F,$9D,$B7,$F7,$E6,$C6,$63,$F9,$21,$93,$F9,$37,$3B,$1F,$A3
	.dc.b	$6F,$CC,$7C,$7C,$3F,$DF,$1B,$19,$3D,$51,$97,$D9,$EF,$FE,$42,$FA
	.dc.b	$36,$CD,$3E,$E6,$DF,$AE,$AF,$94,$43,$F6,$5C,$DF,$ED,$67,$F9,$E7
	.dc.b	$80,$FC,$B5,$A3,$E2,$33,$18,$6C,$66,$BF,$9B,$4E,$5F,$F0,$AF,$C7
	.dc.b	$3E,$E3,$9F,$6F,$F4,$98,$C5,$A3,$13,$AA,$E2,$D7,$F0,$19,$79,$DF
	.dc.b	$4F,$37,$EC,$B4,$F7,$46,$0F,$71,$C2,$35,$3C,$63,$07,$F8,$AB,$0F
	.dc.b	$F1,$CF,$95,$F0,$3D,$2C,$72,$7F,$6F,$9F,$F2,$3D,$7E,$12,$B4,$D8
	.dc.b	$CB,$D0,$F1,$10,$C1,$FE,$05,$FC,$AD,$F9,$3F,$F8,$F3,$0F,$14,$62
	.dc.b	$79,$9F,$49,$C4,$6A,$78,$4F,$E5,$BE,$C3,$EB,$DF,$A1,$B1,$98,$FE
	.dc.b	$10,$7F,$95,$A3,$31,$E4,$3F,$C9,$BA,$7C,$A7,$07,$F3,$4C,$FE,$43
	.dc.b	$E3,$EE,$FF,$0F,$3F,$E2,$CC,$BF,$82,$3D,$9B,$87,$B2,$77,$3F,$E6
	.dc.b	$EF,$CC,$1F,$4B,$FC,$5D,$B5,$FC,$B1,$B3,$FE,$0F,$6B,$F2,$CF,$3E
	.dc.b	$1E,$47,$CD,$BF,$DE,$3F,$9A,$FE,$6E,$FB,$BF,$CA,$4F,$F9,$FA,$31
	.dc.b	$87,$E6,$63,$13,$A7,$FC,$DF,$3A,$E6,$FF,$10,$7C,$9B,$F2,$4F,$E5
	.dc.b	$CD,$C3,$FA,$1B,$6F,$73,$FC,$CF,$7F,$99,$1E,$8C,$BC,$CF,$E5,$8F
	.dc.b	$F3,$ED,$71,$7F,$CF,$AD,$7F,$D2,$AD,$77,$5F,$CF,$FE,$B7,$F5,$7E
	.dc.b	$1F,$F8,$D5,$B3,$FE,$86,$6B,$FE,$9E,$6B,$E3,$3C,$47,$CA,$2B,$FB
	.dc.b	$6C,$BA,$73,$FC,$59,$4E,$5E,$D1,$F2,$73,$FF,$51,$FE,$0D,$FC,$98
	.dc.b	$8F,$F2,$6E,$84,$7C,$5E,$0F,$B5,$B7,$C6,$C3,$C4,$7D,$FE,$FF,$E5
	.dc.b	$5F,$8C,$FF,$1F,$24,$7F,$2E,$F0,$11,$FD,$6F,$ED,$FF,$9C,$4F,$F6
	.dc.b	$56,$E7,$FE,$59,$6C,$7F,$AA,$9A,$FE,$83,$0F,$11,$31,$FC,$77,$E4
	.dc.b	$EE,$AF,$ED,$0D,$AF,$F0,$1B,$5F,$F3,$33,$5F,$F4,$8E,$5F,$87,$FC
	.dc.b	$47,$F8,$83,$63,$3C,$7F,$CC,$27,$2D,$8E,$3F,$31,$FC,$D3,$EB,$78
	.dc.b	$B6,$33,$AA,$72,$D8,$FE,$C9,$F3,$5F,$EC,$AF,$AD,$FE,$C6,$D8,$C8
	.dc.b	$8C,$B4,$FE,$D9,$F3,$5F,$C1,$9A,$7F,$8E,$0B,$BF,$F6,$E8,$D4,$7F
	.dc.b	$B6,$38,$BF,$B1,$7C,$DF,$78,$C6,$C8,$C4,$7F,$B9,$E3,$0F,$84,$DC
	.dc.b	$69,$6F,$F4,$7F,$16,$F9,$FC,$3F,$EE,$66,$C9,$A7,$D0,$F0,$7F,$AE
	.dc.b	$FC,$3D,$DC,$D2,$57,$F8,$2F,$E0,$3F,$DE,$3F,$81,$FF,$82,$9B,$19
	.dc.b	$11,$96,$9F,$EF,$6C,$BF,$04,$FF,$BE,$8F,$4E,$5F,$8A,$23,$F9,$BF
	.dc.b	$97,$FF,$7F,$FB,$BF,$E1,$86,$C7,$F8,$75,$AF,$CA,$6B,$FD,$E3,$67
	.dc.b	$D5,$7F,$7A,$E3,$FC,$7B,$67,$F2,$6D,$7F,$DB,$13,$C3,$5F,$A1,$01
	.dc.b	$D5,$FA,$CF,$0B,$F5,$5E,$9F,$F1,$89,$FD,$53,$5F,$74,$ED,$FE,$70
	.dc.b	$7E,$F0,$FD,$77,$67,$F3,$9F,$87,$FF,$26,$61,$FC,$CE,$7C,$11,$8F
	.dc.b	$0F,$FD,$A0,$D8,$F9,$47,$FB,$AF,$93,$F9,$C7,$3F,$F8,$BF,$AB,$D1
	.dc.b	$C7,$F9,$A2,$57,$E7,$9F,$CB,$2F,$66,$91,$FC,$EB,$F6,$FD,$49,$1F
	.dc.b	$7D,$AF,$DB,$F9,$57,$27,$F2,$56,$EB,$EB,$94,$63,$5F,$F7,$53,$63
	.dc.b	$E4,$8E,$FF,$9A,$D3,$2E,$0F,$E6,$FD,$DF,$C5,$7C,$8F,$E9,$73,$F1
	.dc.b	$3F,$02,$7E,$11,$F6,$71,$FF,$7C,$46,$27,$C8,$D8,$CC,$EE,$7F,$39
	.dc.b	$E0,$7E,$6C,$8E,$0F,$A3,$97,$BB,$F2,$5F,$5E,$BF,$E4,$86,$EA,$69
	.dc.b	$FA,$5F,$52,$5F,$EA,$5A,$FE,$97,$3F,$37,$E2,$5E,$48,$34,$8F,$E3
	.dc.b	$36,$F8,$4F,$F6,$BC,$1F,$AE,$39,$FF,$D6,$F1,$86,$C6,$67,$FC,$F2
	.dc.b	$D7,$FD,$0F,$D5,$FA,$1F,$B9,$B0,$69,$1F,$E7,$BF,$1B,$F0,$5B,$87
	.dc.b	$C1,$6E,$BF,$D0,$27,$51,$FA,$2E,$3F,$B2,$FF,$8C,$EB,$D3,$F1,$9E
	.dc.b	$62,$47,$99,$FF,$3E,$F6,$7F,$68,$25,$72,$97,$B9,$F4,$7D,$AF,$F6
	.dc.b	$BB,$63,$FC,$FF,$92,$72,$D8,$FE,$D9,$F8,$D7,$F0,$1F,$C5,$7F,$D5
	.dc.b	$2D,$8F,$F5,$73,$5F,$E7,$6D,$7F,$D1,$3F,$99,$FF,$2C,$D3,$FD,$4A
	.dc.b	$D8,$FF,$38,$35,$FE,$49,$08,$C4,$7C,$53,$F3,$F9,$FF,$94,$A5,$F2
	.dc.b	$AF,$A7,$FB,$AD,$AF,$F5,$9B,$5F,$F3,$F3,$5E,$FF,$B5,$FF,$B1,$BE
	.dc.b	$03,$FD,$DD,$68,$C3,$AF,$86,$FF,$3B,$36,$95,$FE,$F3,$E7,$FC,$8C
	.dc.b	$7E,$4F,$2F,$F3,$3F,$CA,$F3,$FA,$FF,$6C,$FF,$2B,$8C,$3F,$E6,$46
	.dc.b	$BF,$EF,$F6,$E1,$DE,$3C,$B9,$7F,$B7,$1B,$97,$47,$FD,$6F,$19,$9F
	.dc.b	$17,$1F,$E0,$E7,$F9,$81,$FE,$3F,$32,$7F,$94,$7F,$0C,$FE,$CD,$F2
	.dc.b	$BF,$EC,$08,$82,$33,$5F,$E7,$A6,$C6,$9B,$1F,$6B,$C0,$7C,$CF,$96
	.dc.b	$3F,$90,$B6,$33,$5F,$E8,$49,$CB,$FE,$86,$3C,$E4,$DB,$E6,$BF,$D4
	.dc.b	$EF,$BB,$7D,$66,$3E,$9B,$FE,$DD,$5A,$AF,$55,$6A,$7B,$5F,$49,$F8
	.dc.b	$87,$E3,$CF,$8A,$7F,$58,$F2,$3F,$DA,$DF,$01,$FD,$8F,$C8,$7C,$6F
	.dc.b	$F7,$B9,$CD,$7F,$39,$8C,$78,$AB,$E1,$18,$6C,$FF,$6D,$AE,$37,$D2
	.dc.b	$31,$7C,$5E,$86,$2B,$8B,$FD,$66,$35,$3C,$63,$FD,$47,$7F,$D9,$9A
	.dc.b	$7F,$BA,$1C,$60,$E4,$70,$71,$AF,$F4,$B5,$6A,$72,$D9,$D4,$F4,$3E
	.dc.b	$96,$CD,$3E,$29,$F4,$B5,$FA,$17,$F8,$76,$C6,$67,$84,$FE,$AF,$0E
	.dc.b	$F7,$E7,$21,$FE,$FB,$8F,$F7,$C6,$FF,$8A,$7F,$72,$DB,$FC,$97,$07
	.dc.b	$F8,$66,$72,$D8,$F0,$1E,$6A,$F4,$B5,$FC,$3E,$1E,$48,$DD,$79,$72
	.dc.b	$F8,$39,$FC,$96,$E5,$E8,$C3,$CC,$7A,$0F,$96,$53,$E4,$1F,$B8,$E2
	.dc.b	$F2,$3F,$E2,$FC,$3F,$89,$46,$23,$47,$96,$32,$F9,$F8,$FF,$85,$E3
	.dc.b	$FC,$27,$8D,$C9,$C6,$73,$7F,$E5,$16,$DF,$98,$87,$98,$B1,$DB,$B3
	.dc.b	$F8,$7C,$7F,$47,$8F,$E4,$51,$83,$F9,$4F,$3F,$F3,$1C,$7F,$97,$F4
	.dc.b	$7D,$1B,$FE,$7B,$39,$6C,$65,$F0,$9F,$10,$9B,$7C,$DA,$FE,$8F,$EC
	.dc.b	$7E,$1A,$7F,$A3,$F5,$7F,$4D,$9C,$E5,$F9,$93,$FA,$55,$6E,$BF,$AA
	.dc.b	$8F,$BA,$FE,$B6,$D3,$FA,$D9,$4D,$BF,$E4,$EE,$FF,$ED,$17,$7B,$FF
	.dc.b	$B3,$37,$5F,$55,$FF,$70,$9C,$BF,$7F,$3F,$53,$78,$BF,$0C,$FE,$9F
	.dc.b	$C4,$78,$3C,$B5,$FF,$00,$CF,$AE,$72,$D9,$EE,$FC,$6A,$E6,$DE,$5F
	.dc.b	$85,$E9,$BC,$3F,$DA,$7E,$39,$FE,$11,$C8,$D2,$5F,$FA,$57,$D1,$5F
	.dc.b	$FB,$AB,$E1,$FF,$C4,$24,$9D,$46,$5F,$EF,$3A,$7F,$8A,$F8,$3F,$CE
	.dc.b	$B8,$FF,$8B,$E7,$D5,$E4,$FB,$6B,$51,$F0,$F4,$FD,$49,$2F,$FC,$81
	.dc.b	$F8,$04,$9F,$F2,$E7,$9F,$FE,$93,$8C,$47,$FB,$2B,$DC,$FF,$4B,$7C
	.dc.b	$17,$FA,$C4,$91,$FE,$B5,$7F,$CA,$7B,$C8,$FF,$2C,$FA,$1F,$11,$BA
	.dc.b	$FC,$47,$FD,$79,$F7,$BF,$EB,$B2,$38,$3C,$63,$EA,$FF,$CA,$C4,$72
	.dc.b	$FF,$27,$D2,$47,$BD,$FF,$3E,$FC,$94,$8F,$F4,$0B,$5F,$F4,$66,$1F
	.dc.b	$CA,$EB,$FC,$D6,$48,$F3,$BF,$EC,$FF,$C6,$A4,$7F,$B4,$DA,$FF,$BB
	.dc.b	$BE,$0B,$FD,$75,$F9,$07,$FA,$80,$91,$FE,$FA,$7F,$DF,$FF,$90,$48
	.dc.b	$FB,$BE,$E7,$FB,$06,$31,$1F,$ED,$9E,$5F,$F5,$BE,$3F,$EB,$4F,$77
	.dc.b	$CD,$3F,$EB,$49,$CB,$63,$2F,$EF,$BB,$7F,$46,$DE,$CF,$CC,$FB,$CF
	.dc.b	$63,$FD,$A1,$BF,$FA,$88,$95,$FE,$A6,$7F,$DB,$7C,$6F,$C2,$29,$27
	.dc.b	$FD,$BB,$19,$7F,$DD,$03,$E2,$FF,$DD,$8D,$32,$E0,$FE,$D1,$F0,$0F
	.dc.b	$85,$FD,$7B,$67,$FB,$0B,$79,$B7,$3F,$C0,$BC,$1F,$D4,$F5,$FF,$BA
	.dc.b	$7F,$0C,$FF,$6B,$C6,$30,$46,$3C,$15,$6E,$7F,$5A,$D7,$FD,$EB,$C5
	.dc.b	$FD,$B7,$7F,$D6,$D7,$FB,$A1,$B1,$99,$FF,$35,$CE,$5F,$F7,$3E,$7E
	.dc.b	$9F,$C8,$BC,$F0,$69,$1F,$DE,$FC,$0F,$FC,$01,$E0,$F3,$36,$32,$7C
	.dc.b	$D0,$34,$FE,$13,$1B,$8F,$C9,$0F,$D9,$BF,$DE,$24,$7F,$85,$F5,$F2
	.dc.b	$8F,$89,$D5,$C6,$7A,$66,$8C,$61,$FE,$FF,$9F,$11,$A8,$E1,$C7,$FC
	.dc.b	$6B,$D8,$FF,$0A,$7E,$41,$F2,$1B,$5F,$19,$AF,$F7,$1C,$3F,$DE,$54
	.dc.b	$AF,$9F,$F1,$9F,$E1,$DF,$C1,$6C,$9A,$4B,$F9,$FC,$BF,$88,$23,$F8
	.dc.b	$87,$F9,$CF,$5F,$E8,$A3,$FB,$81,$0F,$EC,$5D,$5F,$A3,$5F,$47,$FE
	.dc.b	$3B,$8C,$37,$9B,$E8,$B6,$76,$7F,$36,$F9,$0F,$CA,$6A,$69,$23,$F9
	.dc.b	$81,$FC,$54,$FF,$0C,$66,$F9,$17,$8F,$FD,$AE,$7F,$C7,$D1,$89,$FE
	.dc.b	$E0,$DE,$CF,$E8,$2D,$9F,$C2,$9F,$86,$F9,$2F,$D9,$57,$F2,$98,$C8
	.dc.b	$8C,$F8,$5F,$E5,$96,$C6,$9A,$FF,$09,$F2,$FD,$2F,$F8,$2A,$3A,$FC
	.dc.b	$F7,$1F,$07,$FC,$A8,$DC,$0D,$3F,$CC,$DF,$1A,$E5,$F7,$60,$8C,$57
	.dc.b	$F9,$81,$BA,$70,$AD,$F1,$FD,$BB,$FE,$E5,$3F,$D0,$1F,$83,$5C,$1F
	.dc.b	$7B,$5C,$D7,$82,$31,$1F,$AC,$F9,$47,$CA,$FF,$2A,$8C,$35,$46,$B1
	.dc.b	$F9,$91,$F4,$23,$1B,$78,$E3,$F1,$D9,$79,$CF,$47,$CB,$3D,$2F,$82
	.dc.b	$DC,$3F,$CF,$ED,$D3,$FD,$18,$DB,$FF,$7C,$FC,$BB,$C5,$FC,$9C,$61
	.dc.b	$AF,$F9,$76,$34,$FE,$58,$C4,$69,$F2,$F9,$3F,$E1,$CC,$BC,$CF,$F8
	.dc.b	$A3,$1F,$CA,$75,$FF,$A3,$FA,$7F,$1E,$DB,$FC,$7C,$E8,$EC,$FE,$62
	.dc.b	$34,$F8,$FE,$63,$F8,$CF,$B1,$FE,$18,$CD,$18,$9F,$6B,$5E,$02,$7E
	.dc.b	$63,$F8,$03,$F5,$3D,$1F,$E6,$78,$FD,$19,$FE,$A5,$2E,$1C,$67,$CF
	.dc.b	$3B,$7F,$64,$7E,$77,$6E,$BA,$7C,$2E,$8F,$CF,$CF,$F3,$0F,$99,$FD
	.dc.b	$5F,$57,$FA,$FD,$B1,$FA,$C6,$BF,$B2,$6B,$FC,$E3,$DA,$7C,$D7,$E2
	.dc.b	$DB,$9F,$E2,$A7,$53,$93,$F2,$18,$F8,$F2,$F2,$F0,$FF,$97,$3C,$5F
	.dc.b	$F2,$4F,$CD,$FF,$10,$26,$1F,$EC,$D8,$CB,$7C,$07,$CA,$FF,$4A,$E3
	.dc.b	$FE,$C8,$D9,$FF,$27,$B5,$FE,$9F,$5F,$E5,$68,$C4,$7F,$14,$F0,$BF
	.dc.b	$8B,$6B,$FC,$65,$BC,$5F,$E0,$D8,$FE,$2F,$1F,$C6,$EF,$FC,$BF,$F0
	.dc.b	$DF,$D6,$BF,$2F,$FE,$E8,$DC,$3F,$CB,$ED,$7C,$39,$A3,$11,$FD,$8B
	.dc.b	$7F,$F6,$27,$83,$FE,$E7,$24,$7F,$1E,$8C,$BF,$C8,$3D,$7F,$F8,$01
	.dc.b	$3F,$EF,$76,$C6,$78,$7F,$B6,$CE,$5B,$1C,$7A,$EB,$FC,$BA,$BF,$5B
	.dc.b	$5C,$E7,$33,$FD,$DA,$72,$FF,$30,$3E,$8E,$BF,$EF,$C6,$DE,$5A,$6F
	.dc.b	$7F,$E4,$AE,$0F,$E7,$2E,$09,$3C,$DB,$3C,$DB,$3F,$D0,$3E,$39,$F1
	.dc.b	$AF,$8E,$F0,$46,$1B,$19,$9F,$44,$E5,$F5,$7E,$30,$F9,$16,$5E,$98
	.dc.b	$C3,$63,$33,$C1,$AF,$F5,$29,$DB,$99,$E0,$47,$04,$38,$4F,$27,$FA
	.dc.b	$FC,$F2,$73,$3C,$5A,$71,$37,$3C,$5F,$F1,$5C,$F1,$73,$3C,$5A,$71
	.dc.b	$37,$3C,$54,$F1,$7F,$29,$CE,$EC,$F9,$51,$FD,$D3,$81,$39,$CF,$F0
	.dc.b	$91,$8A,$DD,$72,$6A,$D9,$F6,$9F,$42,$FF,$34,$D9,$FC,$F3,$5F,$CB
	.dc.b	$35,$FF,$1E,$F8,$7F,$E4,$AF,$CA,$7F,$CC,$8D,$8F,$F6,$7B,$5F,$D8
	.dc.b	$B5,$FB,$FF,$13,$F8,$AD,$9E,$AF,$FB,$4A,$30,$DE,$1F,$C2,$B6,$77
	.dc.b	$5F,$B5,$DB,$E3,$3F,$09,$AE,$75,$FD,$26,$DF,$16,$3F,$CF,$5E,$17
	.dc.b	$D8,$7F,$DC,$AD,$7F,$83,$B5,$FF,$2D,$B6,$D9,$7F,$8F,$9F,$DA,$18
	.dc.b	$F1,$7F,$08,$D3,$31,$BA,$F3,$61,$E4,$C7,$CF,$1F,$75,$AF,$D9,$3F
	.dc.b	$D2,$1E,$07,$EF,$7B,$3F,$DF,$ED,$EB,$FE,$87,$3F,$C1,$67,$2D,$8F
	.dc.b	$F5,$17,$2F,$FA,$67,$C5,$FF,$40,$B6,$7F,$64,$7F,$90,$29,$5C,$23
	.dc.b	$E5,$1E,$82,$18,$3F,$AC,$7E,$17,$FB,$AA,$FE,$B8,$D8,$CE,$9D,$9F
	.dc.b	$EB,$91,$97,$F9,$AB,$EB,$23,$E9,$7F,$8E,$2D,$FE,$D3,$E1,$3E,$46
	.dc.b	$9F,$2C,$F9,$FF,$82,$7D,$2D,$8D,$26,$1F,$17,$83,$FD,$47,$F1,$3F
	.dc.b	$84,$C0,$D2,$3D,$CF,$A8,$53,$C9,$C1,$F6,$36,$F9,$35,$F8,$96,$CF
	.dc.b	$EF,$8F,$C7,$94,$FC,$79,$4F,$F5,$4F,$8F,$FD,$FD,$FA,$9B,$FF,$9F
	.dc.b	$17,$8D,$FE,$D3,$9D,$47,$F5,$E6,$BE,$63,$E2,$9F,$60,$A7,$F8,$C3
	.dc.b	$47,$C9,$7F,$32,$EC,$FD,$93,$67,$4D,$7F,$D9,$3A,$FE,$8A,$31,$1F
	.dc.b	$BC,$F8,$AF,$E7,$7F,$37,$FB,$B6,$C7,$76,$BC,$B0,$46,$23,$D0,$43
	.dc.b	$E6,$F0,$FD,$D7,$A6,$1F,$C7,$8F,$D9,$CF,$F1,$56,$99,$AE,$AF,$F2
	.dc.b	$4D,$7F,$3B,$E9,$7F,$B8,$BD,$8F,$F7,$E3,$70,$F9,$ED,$7F,$C1,$2D
	.dc.b	$7C,$9C,$9F,$92,$7E,$59,$0F,$EB,$7F,$A5,$7F,$39,$C3,$FD,$C5,$1F
	.dc.b	$E5,$E8,$C7,$B5,$FE,$AA,$94,$62,$75,$A7,$F6,$4A,$FF,$57,$E5,$FC
	.dc.b	$42,$FF,$81,$67,$FD,$26,$31,$7F,$C0,$63,$FA,$63,$F2,$0F,$C0,$1E
	.dc.b	$8E,$4F,$CB,$47,$FA,$A6,$3E,$23,$7C,$1E,$ED,$8F,$E0,$8D,$7F,$C7
	.dc.b	$6B,$54,$3F,$30,$E3,$18,$6C,$64,$E7,$19,$7A,$DF,$2F,$C2,$3C,$2D
	.dc.b	$8C,$88,$CB,$4F,$2F,$A3,$E2,$FD,$27,$D2,$D8,$C8,$8C,$B4,$F6,$FA
	.dc.b	$DE,$AF,$CB,$7E,$0B,$63,$22,$32,$D3,$EE,$7B,$7F,$E1,$AF,$C9,$3E
	.dc.b	$E7,$1F,$FB,$E7,$85,$18,$9D,$57,$09,$CB,$FC,$2F,$93,$F1,$8F,$D7
	.dc.b	$E4,$FB,$9E,$77,$CE,$E8,$FA,$AD,$9F,$AA,$D7,$EA,$B5,$FC,$9F,$C0
	.dc.b	$FF,$54,$CC,$D2,$63,$FD,$6A,$3F,$B9,$BF,$AD,$6F,$2F,$B3,$C7,$FD
	.dc.b	$74,$62,$FF,$0B,$F8,$DF,$E9,$7A,$3F,$B2,$B7,$A0,$D3,$F5,$58,$FE
	.dc.b	$83,$E2,$3F,$C8,$3C,$9F,$21,$B8,$7E,$9D,$AF,$F1,$0C,$51,$88,$63
	.dc.b	$FA,$27,$F3,$37,$FA,$62,$3F,$C2,$4F,$9D,$CF,$FD,$B9,$FD,$31,$F9
	.dc.b	$EF,$7B,$F9,$C6,$7F,$A5,$EF,$FD,$1C,$60,$FF,$01,$64,$F9,$EF,$F2
	.dc.b	$CF,$E2,$1F,$80,$6E,$1F,$BF,$6B,$E9,$6B,$E8,$C1,$D3,$FA,$AF,$A4
	.dc.b	$FC,$69,$27,$F2,$2F,$E3,$38,$FE,$BC,$5D,$7E,$C4,$7F,$94,$F6,$FF
	.dc.b	$0A,$E9,$FD,$09,$B5,$FD,$20,$F7,$14,$FD,$DC,$6F,$1F,$E1,$84,$BF
	.dc.b	$93,$3F,$C2,$8F,$8E,$7F,$1F,$29,$EB,$9F,$E0,$7F,$35,$F2,$20,$D2
	.dc.b	$3F,$C8,$F0,$F5,$B6,$BD,$67,$F1,$BA,$FE,$05,$1E,$63,$F8,$01,$7B
	.dc.b	$7B,$1F,$0D,$87,$32,$9C,$7A,$3C,$84,$7E,$1B,$ED,$E9,$FD,$8C,$7B
	.dc.b	$0F,$6D,$FF,$36,$F1,$3F,$54,$47,$F5,$0F,$EB,$7C,$2F,$F4,$DF,$C2
	.dc.b	$7F,$02,$6D,$7F,$41,$6B,$F9,$B6,$BF,$83,$FA,$DF,$E5,$7F,$8E,$7F
	.dc.b	$32,$6C,$7F,$32,$6B,$FC,$C9,$AF,$F0,$73,$D5,$CF,$FC,$FB,$8B,$FA
	.dc.b	$39,$23,$FA,$33,$FE,$2C,$CF,$DE,$49,$FD,$23,$FC,$4B,$3F,$DC,$F8
	.dc.b	$7F,$E5,$CE,$BF,$E8,$11,$FE,$60,$6E,$1F,$D6,$DA,$FF,$44,$3F,$D7
	.dc.b	$AF,$FA,$62,$1F,$86,$6F,$0F,$F6,$0B,$63,$35,$C5,$FE,$13,$3A,$8C
	.dc.b	$53,$C4,$FF,$25,$47,$66,$D7,$F1,$06,$CE,$E3,$FC,$A5,$3F,$1C,$FE
	.dc.b	$07,$C1,$FD,$A7,$DA,$FE,$DA,$48,$C1,$FD,$B7,$EB,$FC,$8F,$F6,$26
	.dc.b	$BF,$BB,$6F,$47,$FA,$19,$CC,$7E,$2F,$5F,$9D,$7F,$C4,$E3,$FC,$57
	.dc.b	$8F,$F2,$6C,$7F,$C5,$B1,$88,$FF,$70,$C7,$C0,$6C,$66,$36,$F8,$4B
	.dc.b	$87,$85,$BD,$0C,$7C,$AD,$8E,$27,$97,$4F,$C0,$93,$1F,$F0,$67,$E0
	.dc.b	$5C,$09,$3D,$1E,$7E,$CC,$8E,$ED,$7F,$DE,$F8,$FF,$AA,$3F,$05,$F0
	.dc.b	$92,$3C,$AF,$C6,$D7,$E2,$C7,$E6,$B0,$FF,$0E,$5F,$3D,$3F,$C2,$6D
	.dc.b	$9F,$E2,$55,$C1,$FB,$67,$13,$F8,$C7,$B8,$F6,$5E,$37,$F9,$07,$87
	.dc.b	$F2,$D3,$98,$C5,$7F,$87,$23,$0F,$F8,$7E,$77,$5C,$A3,$06,$DC,$72
	.dc.b	$27,$77,$D2,$97,$D2,$B5,$1C,$67,$58,$7F,$BE,$EF,$FD,$8E,$49,$FF
	.dc.b	$6E,$35,$FF,$50,$B5,$FF,$80,$0F,$F8,$EB,$1F,$F1,$AB,$F4,$B8,$3F
	.dc.b	$E0,$0C,$F9,$73,$F9,$6D,$3F,$C7,$B5,$C2,$3B,$C4,$3B,$94,$FE,$B5
	.dc.b	$AE,$DF,$F2,$2B,$5C,$9D,$4B,$7E,$3A,$D6,$DF,$EE,$C6,$E1,$E6,$9D
	.dc.b	$57,$1A,$FF,$78,$EF,$A2,$46,$69,$1E,$18,$FF,$44,$D7,$C5,$69,$E9
	.dc.b	$AF,$D3,$47,$C5,$E1,$F0,$92,$7D,$19,$7F,$86,$B8,$8E,$1F,$DC,$56
	.dc.b	$27,$F6,$8D,$7F,$90,$C7,$08,$C4,$78,$8F,$96,$43,$EC,$9F,$E7,$0A
	.dc.b	$F1,$47,$96,$3D,$D1,$F1,$A3,$EE,$C7,$F9,$D9,$AF,$F9,$D9,$AF,$F8
	.dc.b	$3D,$AF,$84,$EE,$78,$4C,$57,$F2,$91,$E9,$7F,$3D,$E4,$F8,$2D,$8C
	.dc.b	$CF,$F3,$49,$CB,$FC,$DB,$93,$E3,$FD,$F7,$F3,$D2,$46,$69,$1F,$D2
	.dc.b	$78,$BE,$34,$FF,$3E,$6B,$FC,$99,$AF,$F4,$C6,$BE,$B2,$70,$FE,$A5
	.dc.b	$FB,$DB,$B9,$FE,$1E,$E8,$C4,$FE,$C8,$FF,$4E,$CE,$A3,$3A,$FE,$26
	.dc.b	$FA,$57,$C7,$6B,$FD,$E4,$7E,$C9,$FD,$E5,$A7,$F7,$92,$9F,$DE,$4A
	.dc.b	$7F,$B2,$3A,$FF,$1B,$FB,$47,$F8,$27,$53,$49,$3F,$EC,$BE,$EF,$87
	.dc.b	$3F,$E1,$53,$D5,$1F,$ED,$28,$CF,$8C,$7E,$3E,$3E,$89,$F8,$FC,$0E
	.dc.b	$BE,$31,$F8,$5D,$FF,$C9,$3C,$9F,$E3,$B2,$47,$F9,$29,$FF,$29,$6D
	.dc.b	$23,$FC,$74,$D7,$FC,$71,$F7,$DF,$E4,$32,$3F,$E5,$27,$4F,$FA,$5B
	.dc.b	$4F,$49,$FE,$E4,$F0,$FD,$27,$E0,$30,$FF,$2D,$FC,$AF,$F9,$34,$95
	.dc.b	$FE,$5C,$7E,$B6,$7E,$BC,$6E,$7F,$12,$FF,$DC,$F1,$FE,$86,$BC,$71
	.dc.b	$7E,$E6,$30,$7F,$96,$7F,$07,$93,$FE,$68,$6B,$FE,$93,$CB,$CB,$9F
	.dc.b	$A3,$83,$FD,$5A,$D8,$F4,$61,$FE,$7E,$A7,$AB,$FE,$CA,$23,$FE,$CE
	.dc.b	$7F,$D1,$54,$FD,$86,$92,$7F,$D1,$91,$97,$FC,$F7,$F8,$F7,$F9,$BC
	.dc.b	$8E,$0F,$E6,$C7,$DE,$D7,$EE,$F1,$7E,$63,$6F,$87,$3F,$EF,$EB,$FD
	.dc.b	$01,$1F,$67,$9D,$FC,$83,$F1,$FF,$F5,$B4,$9A,$49,$FF,$61,$8F,$85
	.dc.b	$FF,$21,$3F,$E9,$FE,$F3,$93,$71,$AC,$3F,$D9,$31,$93,$FD,$81,$E9
	.dc.b	$7F,$30,$F8,$7F,$F6,$96,$26,$92,$7F,$22,$43,$EB,$47,$D3,$3E,$FF
	.dc.b	$A9,$E2,$F9,$0F,$F7,$5C,$9A,$47,$E4,$EB,$FA,$72,$73,$FE,$36,$3F
	.dc.b	$DF,$5F,$8C,$7F,$BF,$70,$34,$91,$FE,$63,$CF,$F8,$87,$A5,$F0,$BF
	.dc.b	$2E,$F8,$1C,$DF,$85,$6C,$66,$B8,$4E,$5B,$E2,$FC,$AE,$2E,$2D,$8C
	.dc.b	$9F,$54,$B1,$DB,$3F,$A3,$DB,$64,$D2,$3E,$0F,$13,$EB,$FE,$E1,$E6
	.dc.b	$6C,$64,$46,$5A,$7F,$B3,$B8,$BF,$D8,$44,$70,$7C,$CB,$F9,$4F,$F2
	.dc.b	$B6,$BF,$EE,$5D,$FF,$25,$DD,$FB,$48,$FC,$B1,$FA,$FE,$8F,$EE,$7F
	.dc.b	$87,$7E,$3A,$9D,$DF,$C6,$D4,$66,$30,$D9,$E3,$5E,$37,$F4,$B8,$78
	.dc.b	$A3,$F9,$17,$C1,$3C,$5F,$CC,$F5,$7E,$5E,$31,$3F,$97,$1C,$A8,$CC
	.dc.b	$EE,$B8,$D7,$FA,$77,$8F,$F8,$1E,$9F,$9B,$8C,$77,$FE,$7D,$B1,$97
	.dc.b	$F5,$2F,$F5,$98,$D4,$7C,$9C,$FD,$24,$DB,$E2,$5F,$6E,$9B,$7F,$E1
	.dc.b	$A6,$E1,$FA,$1C,$3B,$C7,$0F,$81,$FF,$0F,$F7,$DA,$34,$95,$F2,$71
	.dc.b	$FE,$5E,$7C,$5F,$17,$F7,$FE,$1D,$93,$48,$FE,$97,$7F,$F0,$5D,$FC
	.dc.b	$DF,$2B,$F8,$3B,$67,$C2,$6E,$32,$F8,$48,$7F,$0E,$E1,$FA,$51,$FC
	.dc.b	$43,$E9,$39,$12,$79,$07,$E3,$FC,$2F,$F6,$CE,$03,$49,$AF,$CD,$AE
	.dc.b	$E7,$F8,$A8,$FC,$0E,$4F,$86,$FF,$07,$B6,$3F,$A9,$37,$6F,$C5,$4F
	.dc.b	$F8,$D3,$E5,$38,$5F,$F8,$B2,$FF,$DD,$51,$8C,$FF,$95,$C2,$30,$FF
	.dc.b	$2E,$6B,$F8,$3B,$FA,$6F,$CD,$F8,$9F,$E6,$35,$F9,$96,$C6,$4F,$01
	.dc.b	$EA,$72,$FC,$BC,$5F,$15,$EC,$24,$7B,$1F,$F2,$6F,$89,$FC,$9B,$9B
	.dc.b	$E3,$4F,$F9,$77,$3F,$F1,$87,$C5,$FC,$D2,$3F,$E0,$0D,$D1,$FD,$FF
	.dc.b	$1F,$C1,$D7,$D6,$E2,$FC,$11,$1F,$A2,$7F,$5D,$D7,$F3,$FE,$BF,$F9
	.dc.b	$DB,$E9,$2E,$46,$92,$BF,$21,$F1,$BF,$D5,$3F,$31,$FD,$E3,$63,$22
	.dc.b	$32,$D3,$FA,$EF,$6B,$D7,$FD,$9C,$7F,$06,$6B,$FD,$8D,$B1,$BC,$3F
	.dc.b	$79,$3F,$D9,$38,$1D,$FF,$E3,$3F,$0B,$EF,$C6,$38,$B7,$48,$C4,$EA
	.dc.b	$B8,$4F,$F7,$0F,$84,$7C,$77,$F9,$99,$B8,$FF,$8F,$9B,$C1,$F8,$C6
	.dc.b	$D7,$1A,$E0,$4A,$E0,$DA,$E0,$D9,$E1,$3A,$24,$E9,$B1,$A8,$C1,$1C
	.dc.b	$10,$FF,$03,$8A,$F2,$B8,$6B,$FE,$0C,$6B,$FE,$EE,$E0,$FF,$7B,$3F
	.dc.b	$E1,$2F,$73,$FD,$F6,$D8,$CB,$FE,$F6,$3E,$17,$E3,$DF,$EB,$67,$A3
	.dc.b	$5F,$13,$68,$D2,$61,$FB,$69,$DB,$67,$6D,$9D,$CF,$82,$30,$47,$04
	.dc.b	$3E,$E4,$70,$6C,$6D,$A6,$CF,$8D,$1B,$6C,$79,$76,$7E,$03,$FE,$34
	.dc.b	$F7,$BF,$86,$C6,$2B,$E2,$9F,$E3,$B9,$D3,$FA,$06,$BF,$5B,$7F,$CB
	.dc.b	$57,$E4,$F8,$38,$C6,$27,$FC,$86,$D7,$F9,$B7,$43,$E9,$FF,$C8,$BC
	.dc.b	$3F,$E2,$46,$CF,$F8,$89,$AF,$06,$BF,$EF,$4F,$51,$EF,$FF,$BD,$BD
	.dc.b	$5F,$F0,$24,$63,$64,$62,$75,$5F,$8E,$95,$6E,$BD,$53,$0F,$4E,$1F
	.dc.b	$EA,$77,$07,$F7,$4F,$DB,$7E,$46,$BF,$E1,$36,$C7,$56,$BB,$DB,$E4
	.dc.b	$BE,$9C,$FF,$21,$97,$89,$B9,$7C,$58,$FE,$82,$64,$FC,$76,$BF,$C1
	.dc.b	$6B,$C6,$7F,$0B,$CB,$CC,$FA,$36,$FE,$93,$E2,$FF,$B8,$72,$FC,$B3
	.dc.b	$67,$3B,$FE,$05,$FF,$2A,$CF,$E6,$D7,$CA,$7F,$B2,$DF,$CF,$35,$FE
	.dc.b	$AA,$D7,$FD,$02,$D7,$F0,$D8,$79,$1B,$87,$AA,$31,$1E,$98,$C1,$1F
	.dc.b	$69,$FC,$86,$37,$87,$FB,$EF,$F2,$DF,$5E,$26,$92,$B9,$1F,$0C,$E7
	.dc.b	$97,$C9,$7F,$D2,$FF,$A1,$7E,$1D,$FE,$DE,$D7,$3C,$5F,$C0,$E3,$11
	.dc.b	$F6,$C7,$C7,$7E,$07,$5F,$F4,$FB,$63,$27,$E0,$E3,$2F,$8C,$87,$C4
	.dc.b	$3E,$DE,$CE,$1F,$C4,$C2,$31,$3F,$8A,$6C,$7F,$1C,$6F,$17,$FA,$CB
	.dc.b	$E2,$5F,$04,$FC,$2F,$97,$4E,$63,$17,$FC,$BF,$0F,$E6,$37,$C2,$F7
	.dc.b	$18,$AD,$DF,$CA,$E2,$FB,$33,$F5,$78,$FE,$7B,$FE,$C0,$C1,$F3,$4F
	.dc.b	$B6,$FE,$71,$3D,$A3,$0F,$84,$E0,$E3,$0F,$53,$63,$F7,$1F,$A8,$3E
	.dc.b	$8B,$F5,$2D,$D7,$F4,$ED,$9F,$F4,$33,$5F,$9E,$7E,$3C,$F5,$65,$C2
	.dc.b	$B9,$92,$B0,$74,$FB,$3F,$A5,$ED,$F4,$6D,$F9,$1A,$FF,$9E,$30,$46
	.dc.b	$23,$C0,$78,$8F,$90,$7E,$02,$37,$4F,$99,$EC,$D9,$EC,$D7,$B5,$7E
	.dc.b	$16,$7F,$D9,$BE,$D7,$E9,$09,$3F,$BE,$7F,$79,$D1,$FA,$18,$FC,$81
	.dc.b	$E3,$70,$FF,$71,$FC,$F7,$F0,$AF,$F8,$35,$AF,$F7,$36,$BF,$DC,$DA
	.dc.b	$FF,$A9,$3A,$DF,$66,$87,$C2,$FF,$B8,$A3,$0D,$8C,$9F,$EE,$28,$CB
	.dc.b	$FE,$E4,$F3,$79,$7D,$AF,$FB,$A9,$B1,$91,$19,$69,$F8,$8F,$1B,$F5
	.dc.b	$51,$F8,$46,$C6,$76,$DB,$F6,$63,$2F,$FA,$37,$DF,$7C,$AF,$AF,$F5
	.dc.b	$1F,$F0,$BC,$FF,$B0,$A7,$2A,$33,$E2,$7F,$B5,$AB,$84,$FF,$87,$A3
	.dc.b	$0E,$CC,$57,$E7,$57,$CA,$72,$8C,$37,$9B,$E6,$36,$76,$79,$3F,$07
	.dc.b	$7C,$7F,$F0,$9F,$C0,$BF,$D1,$71,$8D,$51,$89,$D5,$71,$6B,$F9,$BF
	.dc.b	$3B,$F8,$97,$E9,$FF,$AC,$6C,$7E,$D1,$AE,$ED,$3F,$B2,$FC,$13,$F6
	.dc.b	$E4,$8F,$DB,$BF,$E3,$1D,$7F,$DF,$7D,$5F,$D9,$9B,$7F,$E3,$86,$F8
	.dc.b	$BF,$0E,$75,$BB,$F2,$2F,$C7,$7E,$9B,$F1,$DB,$87,$CC,$6B,$F6,$DA
	.dc.b	$FD,$9F,$C1,$3F,$2D,$F5,$DF,$3D,$B1,$F5,$1A,$FE,$0D,$AF,$E7,$FB
	.dc.b	$BF,$5E,$FE,$C8,$FD,$81,$FC,$90,$FE,$02,$7E,$C0,$87,$E2,$7D,$7F
	.dc.b	$59,$1F,$81,$8F,$F9,$F1,$BD,$1E,$8D,$53,$FE,$54,$96,$7F,$87,$AD
	.dc.b	$4E,$63,$F1,$5C,$FF,$B3,$F4,$6D,$FF,$7E,$3F,$23,$19,$69,$93,$F8
	.dc.b	$E7,$07,$D9,$7E,$E7,$0F,$DA,$3F,$98,$FC,$27,$F5,$7E,$2F,$E6,$2D
	.dc.b	$9F,$E7,$2D,$7F,$93,$35,$F9,$47,$C8,$3E,$51,$0F,$E2,$DE,$57,$F5
	.dc.b	$0F,$0B,$F8,$BC,$7F,$2A,$3F,$53,$BF,$F4,$E7,$F7,$0D,$7E,$71,$FD
	.dc.b	$3C,$BE,$47,$A9,$87,$F4,$EC,$BD,$CF,$41,$F4,$DF,$B3,$AF,$E3,$8D
	.dc.b	$8C,$9B,$79,$BF,$23,$3F,$8C,$E0,$F1,$F9,$7C,$48,$F7,$7F,$AA,$B0
	.dc.b	$FE,$C3,$F1,$3F,$D8,$09,$3F,$D8,$1F,$D5,$70,$BE,$AF,$F9,$87,$97
	.dc.b	$FC,$25,$B7,$FC,$39,$AF,$F1,$D6,$BF,$C6,$FE,$EF,$F5,$A4,$7F,$26
	.dc.b	$FF,$96,$9F,$A0,$7E,$68,$A7,$E7,$8F,$E4,$9D,$BF,$E5,$0B,$34,$91
	.dc.b	$FA,$53,$89,$37,$FE,$A7,$97,$F2,$BF,$13,$FC,$CF,$E2,$7F,$3E,$8F
	.dc.b	$F3,$2B,$6B,$F9,$9C,$6C,$FF,$30,$E0,$E0,$FE,$8F,$3F,$CE,$DA,$FF
	.dc.b	$39,$6F,$4F,$EB,$1C,$D7,$F3,$BF,$AB,$F0,$91,$F0,$BF,$EC,$7C,$BF
	.dc.b	$26,$4E,$4F,$CB,$1F,$E5,$8C,$5F,$09,$F0,$C9,$97,$99,$F8,$DF,$7D
	.dc.b	$FE,$95,$8F,$EA,$0D,$7F,$A7,$37,$3F,$E0,$2E,$63,$E4,$5F,$83,$87
	.dc.b	$F8,$86,$7F,$D4,$FE,$9F,$FC,$95,$50,$FF,$26,$6C,$8C,$CF,$FA,$96
	.dc.b	$30,$FF,$79,$31,$1A,$9F,$29,$92,$9D,$71,$F2,$B9,$3D,$25,$3C,$F7
	.dc.b	$F8,$1D,$3F,$C2,$33,$FC,$73,$DE,$F7,$6A,$69,$2F,$A4,$FF,$22,$F5
	.dc.b	$76,$48,$CD,$7F,$98,$5B,$D5,$D0,$8E,$0F,$26,$1F,$67,$1F,$26,$BF
	.dc.b	$67,$0E,$95,$ED,$C3,$A3,$63,$E9,$5F,$18,$FD,$C1,$FA,$57,$0F,$F9
	.dc.b	$07,$DF,$FD,$1F,$67,$F8,$97,$D3,$7C,$07,$F8,$9E,$95,$A5,$3A,$7F
	.dc.b	$C5,$91,$98,$C4,$70,$E7,$5F,$5B,$E9,$7B,$6C,$EA,$BA,$C6,$2B,$84
	.dc.b	$76,$9D,$5F,$6C,$19,$B4,$F0,$DF,$FA,$EF,$1F,$F1,$2B,$FE,$85,$6D
	.dc.b	$70,$9D,$CE,$4F,$F6,$C4,$EA,$BA,$36,$7A,$ED,$F7,$EB,$E3,$F0,$FF
	.dc.b	$8A,$35,$FF,$27,$5F,$D3,$9C,$9F,$E3,$98,$CC,$F6,$88,$7F,$B5,$27
	.dc.b	$B1,$FE,$2F,$DF,$71,$FE,$41,$6B,$FE,$3F,$6D,$FF,$72,$6E,$FF,$8F
	.dc.b	$AF,$11,$2F,$CD,$7E,$43,$FC,$91,$C1,$C0,$8E,$70,$FD,$13,$B3,$E3
	.dc.b	$1E,$73,$D2,$7C,$F3,$FC,$93,$C7,$E2,$23,$CF,$04,$7A,$AF,$F8,$21
	.dc.b	$E3,$2E,$1F,$C1,$5A,$7F,$00,$3C,$5A,$FD,$98,$F8,$F7,$E2,$3C,$A7
	.dc.b	$EF,$9F,$D3,$57,$89,$B1,$E3,$E4,$FA,$EF,$F9,$81,$AF,$F9,$BD,$AF
	.dc.b	$F9,$B5,$AF,$62,$57,$8D,$F8,$77,$FC,$DC,$FC,$99,$6B,$FA,$2F,$23
	.dc.b	$97,$FD,$D9,$18,$6D,$7F,$9C,$B9,$BF,$D3,$1F,$41,$FE,$71,$6C,$64
	.dc.b	$46,$5A,$7F,$29,$E8,$E6,$FF,$29,$C3,$FA,$C6,$1E,$29,$CC,$FF,$32
	.dc.b	$6B,$EC,$E3,$FD,$0B,$F3,$73,$F7,$3F,$E9,$3E,$5F,$8B,$C1,$FE,$42
	.dc.b	$6F,$07,$F6,$E9,$FD,$3B,$71,$FA,$F1,$F8,$92,$1F,$89,$3F,$D3,$DD
	.dc.b	$3F,$EB,$7F,$82,$FF,$48,$36,$3F,$D2,$AD,$76,$23,$FC,$1C,$FF,$A4
	.dc.b	$0F,$F0,$71,$4F,$F0,$71,$4F,$F5,$47,$5F,$FB,$17,$CB,$FF,$7B,$F6
	.dc.b	$FC,$4E,$37,$AC,$8C,$36,$3E,$EB,$72,$F6,$E1,$EB,$21,$F9,$4D,$7F
	.dc.b	$CC,$FE,$0F,$FB,$4F,$E5,$3F,$C7,$04,$9F,$F1,$F3,$FE,$5C,$A9,$1F
	.dc.b	$E4,$3C,$3B,$F2,$FD,$8C,$39,$CB,$E3,$FF,$C8,$44,$7F,$C9,$0F,$27
	.dc.b	$FC,$85,$75,$5F,$BA,$FA,$70,$7F,$9F,$FE,$23,$FC,$98,$4A,$FF,$29
	.dc.b	$3E,$E7,$F0,$DD,$DF,$CA,$B9,$3F,$0B,$1F,$E6,$AC,$7F,$D2,$44,$7F
	.dc.b	$DA,$1E,$BB,$CB,$FD,$6E,$7F,$3B,$BF,$C7,$CB,$FE,$78,$8F,$B9,$7F
	.dc.b	$E7,$86,$BF,$E7,$86,$BE,$AE,$2F,$F1,$74,$7C,$62,$63,$EF,$E0,$FF
	.dc.b	$6F,$FC,$27,$FA,$40,$97,$FE,$8E,$7F,$D3,$0B,$F2,$CA,$47,$FD,$CB
	.dc.b	$E6,$9D,$1F,$E7,$5C,$BF,$CE,$44,$70,$7E,$A7,$AF,$F5,$6F,$E8,$0F
	.dc.b	$C2,$6D,$F0,$DF,$E9,$AD,$CB,$E4,$47,$7C,$93,$98,$FF,$62,$0F,$D7
	.dc.b	$FF,$D9,$0D,$32,$E0,$FF,$6A,$3E,$83,$FD,$A4,$51,$3F,$EB,$CF,$47
	.dc.b	$DD,$9F,$03,$FD,$C1,$3F,$42,$7F,$DC,$B9,$60,$91,$F9,$7C,$7F,$1D
	.dc.b	$AF,$E3,$4F,$A7,$1F,$EF,$75,$F0,$3F,$EF,$B6,$C6,$63,$07,$F7,$AF
	.dc.b	$59,$D9,$F9,$8F,$B7,$C3,$6B,$FC,$30,$27,$76,$AD,$B6,$7F,$C3,$D9
	.dc.b	$7F,$B6,$BD,$9F,$B9,$F0,$3C,$A3,$12,$8C,$75,$2B,$73,$CA,$30,$EC
	.dc.b	$FF,$6C,$FE,$EB,$FE,$DE,$23,$83,$FC,$5D,$87,$F0,$EE,$4F,$D9,$F8
	.dc.b	$DF,$E8,$AE,$74,$62,$B7,$18,$AF,$E7,$F1,$FE,$65,$8C,$57,$A7,$95
	.dc.b	$EE,$7F,$48,$AF,$C5,$46,$1E,$27,$E0,$1C,$4F,$19,$E1,$39,$F7,$7F
	.dc.b	$41,$F7,$9D,$5B,$1F,$BC,$6B,$D9,$AB,$D5,$7C,$03,$D0,$FE,$A3,$18
	.dc.b	$6C,$67,$4F,$F1,$DC,$E5,$FE,$1D,$D9,$FA,$AB,$FE,$5B,$E0,$7F,$10
	.dc.b	$7F,$57,$E0,$7F,$54,$DB,$F4,$D1,$F0,$23,$07,$F1,$0D,$DE,$E7,$F2
	.dc.b	$28,$FA,$0D,$9F,$9E,$D7,$E7,$35,$F0,$E1,$E3,$7F,$91,$F5,$F2,$CB
	.dc.b	$FD,$11,$1C,$9A,$F2,$6B,$FC,$66,$BF,$99,$FA,$3F,$81,$E6,$FE,$9E
	.dc.b	$4A,$FE,$9E,$FF,$76,$D3,$E0,$13,$1F,$13,$FC,$CA,$7F,$AC,$CF,$1C
	.dc.b	$7A,$47,$23,$FC,$CF,$EB,$3E,$09,$F0,$7F,$C7,$7C,$7F,$D5,$CE,$63
	.dc.b	$1A,$FF,$A1,$63,$0A,$77,$5C,$A3,$07,$F2,$1F,$1F,$EC,$93,$1E,$BD
	.dc.b	$9F,$7E,$EF,$2F,$F5,$46,$E7,$CB,$0F,$E6,$19,$7E,$CE,$B5,$1F,$AA
	.dc.b	$F0,$5F,$09,$FE,$05,$8F,$EB,$B1,$87,$FA,$3D,$FF,$10,$9D,$36,$3F
	.dc.b	$A5,$9F,$68,$F4,$65,$FE,$FC,$AE,$5F,$2D,$6F,$FA,$13,$68,$D3,$FD
	.dc.b	$23,$F0,$EF,$88,$FF,$16,$B7,$1F,$F4,$8B,$71,$DC,$FD,$DE,$3F,$B6
	.dc.b	$4D,$FE,$46,$3F,$DA,$BE,$27,$F3,$C4,$7F,$A4,$BF,$E5,$6E,$4F,$C2
	.dc.b	$4F,$09,$FE,$A1,$18,$C3,$FB,$D3,$FA,$9F,$80,$7C,$B7,$F9,$79,$B5
	.dc.b	$FE,$A6,$6D,$7F,$AB,$1B,$1F,$DF,$36,$FD,$81,$1F,$EB,$6F,$8F,$2F
	.dc.b	$55,$78,$1F,$DD,$6A,$F8,$B7,$9D,$FE,$31,$6C,$7F,$9A,$9B,$5F,$CF
	.dc.b	$A3,$84,$FF,$8D,$F8,$7E,$F7,$97,$F4,$C9,$97,$E1,$72,$FC,$23,$F7
	.dc.b	$4F,$87,$97,$8B,$2F,$A9,$E1,$F5,$6C,$FF,$36,$6B,$FE,$01,$CB,$D1
	.dc.b	$C3,$FE,$2A,$AF,$E6,$19,$FF,$81,$79,$2D,$FF,$BD,$8F,$E5,$D1,$97
	.dc.b	$FC,$46,$D7,$FD,$E1,$C1,$FE,$D5,$D3,$FD,$A0,$48,$FE,$BB,$19,$7F
	.dc.b	$AF,$F5,$FC,$06,$EB,$EE,$8C,$47,$FB,$ED,$AF,$F8,$67,$C0,$FE,$95
	.dc.b	$C1,$FE,$E6,$6C,$7F,$6A,$6B,$FE,$1F,$3F,$9D,$78,$BF,$E8,$68,$CC
	.dc.b	$73,$8C,$13,$6F,$F7,$6C,$EA,$78,$C6,$27,$55,$E1,$6C,$F8,$5A,$F8
	.dc.b	$6B,$91,$23,$31,$E1,$AE,$84,$AE,$8F,$FB,$FB,$DF,$7B,$0F,$7F,$FC
	.dc.b	$0F,$7F,$E3,$B9,$CC,$62,$11,$87,$6D,$7F,$C1,$FF,$02,$E8,$FF,$35
	.dc.b	$C7,$48,$C6,$1C,$DB,$19,$9D,$BC,$9A,$FC,$F8,$DB,$99,$E5,$3A,$8F
	.dc.b	$F2,$94,$E9,$E1,$19,$51,$97,$71,$98,$C4,$6A,$30,$49,$EE,$48,$F4
	.dc.b	$35,$F2,$18,$21,$82,$1E,$47,$83,$4D,$94,$D9,$E6,$7E,$96,$CE,$5F
	.dc.b	$9D,$8F,$D8,$8C,$5F,$DB,$BF,$D2,$D6,$9B,$19,$7E,$67,$C6,$7F,$A8
	.dc.b	$2F,$EB,$36,$3E,$DB,$5D,$E1,$C5,$FF,$40,$61,$E0,$AF,$16,$1D,$E7
	.dc.b	$F7,$2D,$76,$DF,$35,$E6,$BF,$1E,$FD,$95,$7F,$92,$5A,$FF,$9F,$E3
	.dc.b	$4F,$F8,$A4,$D1,$4F,$F7,$E7,$87,$FE,$BA,$FB,$1F,$E4,$9E,$FF,$F7
	.dc.b	$BD,$11,$89,$D5,$7F,$94,$2B,$53,$EA,$DB,$A4,$FA,$1A,$B5,$F9,$FC
	.dc.b	$9B,$3F,$CC,$7E,$77,$F7,$49,$FF,$31,$35,$F5,$35,$FF,$46,$E6,$8C
	.dc.b	$66,$C9,$F2,$3F,$9D,$8C,$36,$33,$1F,$B8,$32,$7E,$1B,$7F,$D9,$61
	.dc.b	$EB,$21,$F8,$23,$FC,$F3,$F8,$1F,$EF,$78,$3E,$9B,$72,$34,$FE,$05
	.dc.b	$7E,$5F,$94,$FC,$25,$FF,$BA,$31,$46,$27,$47,$E1,$CB,$C5,$E2,$AF
	.dc.b	$C2,$EF,$E7,$23,$F6,$3E,$57,$FD,$E7,$F2,$56,$CD,$24,$FF,$8C,$79
	.dc.b	$1F,$55,$F7,$23,$FC,$4B,$18,$AF,$C0,$C6,$E7,$4B,$CA,$FC,$94,$62
	.dc.b	$3E,$F7,$07,$CA,$7E,$41,$0F,$49,$FE,$44,$B7,$AB,$F9,$88,$E9,$18
	.dc.b	$7F,$17,$7F,$9F,$9D,$36,$3F,$1F,$87,$E0,$7C,$82,$FA,$5F,$CD,$C6
	.dc.b	$1B,$19,$9F,$CD,$4E,$5F,$31,$F9,$93,$E8,$14,$FF,$5B,$7C,$1F,$FB
	.dc.b	$E7,$EB,$3F,$D7,$6D,$8C,$88,$CB,$4F,$F5,$BF,$EA,$9E,$6B,$FC,$93
	.dc.b	$63,$FC,$28,$D7,$71,$FC,$CF,$AB,$F8,$BC,$79,$0F,$9F,$C3,$F8,$27
	.dc.b	$FD,$B3,$FA,$03,$E0,$3F,$9D,$36,$FF,$AA,$36,$7F,$AC,$B5,$F2,$E3
	.dc.b	$FA,$9C,$7D,$55,$E6,$8C,$1E,$3E,$0F,$BD,$5F,$70,$F9,$27,$90,$F5
	.dc.b	$FC,$6F,$32,$57,$C0,$7F,$A9,$75,$7F,$1C,$F9,$F5,$7F,$27,$8C,$37
	.dc.b	$AD,$3A,$E8,$FF,$75,$B6,$7C,$3B,$FC,$66,$E5,$E6,$58,$7F,$91,$DE
	.dc.b	$07,$81,$C3,$E1,$F3,$BF,$C3,$18,$7F,$BE,$DB,$1F,$F0,$01,$FE,$0F
	.dc.b	$79,$E0,$FE,$7B,$39,$6B,$D6,$76,$F1,$F8,$BF,$7F,$B9,$D5,$B1,$91
	.dc.b	$19,$69,$F8,$AF,$A4,$FC,$FF,$BD,$F8,$86,$C6,$78,$3F,$DC,$ED,$9F
	.dc.b	$57,$B5,$F3,$BE,$4E,$C9,$A4,$7D,$DF,$1D,$F5,$3F,$08,$F8,$4D,$8C
	.dc.b	$88,$CB,$4F,$F2,$00,$FC,$01,$E1,$71,$8F,$F2,$24,$61,$E0,$62,$34
	.dc.b	$FE,$12,$35,$3F,$4F,$D9,$FC,$3F,$E5,$16,$8D,$23,$F5,$FE,$EF,$F9
	.dc.b	$C7,$09,$A4,$8F,$D9,$FA,$5F,$CC,$FB,$0D,$27,$07,$F3,$58,$7C,$4F
	.dc.b	$E9,$C9,$1F,$BC,$7F,$55,$E2,$7F,$14,$E7,$FD,$EB,$70,$FE,$06,$D7
	.dc.b	$FD,$09,$7F,$C9,$2F,$FB,$44,$7F,$9A,$7A,$7F,$47,$E6,$FE,$82,$3F
	.dc.b	$B7,$C6,$1D,$76,$FE,$99,$FE,$E3,$8F,$E8,$DF,$DB,$10,$FE,$38,$7E
	.dc.b	$54,$FE,$65,$3F,$A2,$AF,$D0,$46,$35,$FF,$35,$3F,$A1,$F9,$5F,$DB
	.dc.b	$FE,$A9,$FA,$56,$C6,$44,$65,$A7,$F2,$2F,$6B,$F9,$BF,$DD,$7F,$20
	.dc.b	$6C,$67,$44,$E5,$B1,$F6,$CF,$F3,$8E,$5F,$EC,$1E,$1F,$B5,$1F,$B7
	.dc.b	$7E,$FB,$FE,$5E,$3F,$70,$53,$E0,$1F,$E5,$E3,$25,$3F,$C5,$5E,$0F
	.dc.b	$F1,$4F,$98,$FA,$04,$8F,$9F,$19,$7F,$87,$F8,$1F,$CC,$DF,$9D,$7F
	.dc.b	$C3,$2B,$E5,$C1,$A4,$7F,$A2,$FC,$57,$F7,$A6,$E3,$FD,$E9,$BE,$77
	.dc.b	$F4,$69,$D4,$7A,$BC,$CE,$A4,$7D,$6E,$EB,$FB,$DC,$7E,$3A,$7F,$BD
	.dc.b	$C3,$91,$18,$6C,$70,$9F,$E5,$3C,$3F,$EB,$BC,$FD,$64,$8F,$92,$FF
	.dc.b	$55,$E3,$FC,$84,$7E,$38,$94,$F9,$F7,$C4,$7F,$8D,$DB,$1F,$E3,$56
	.dc.b	$CF,$FB,$DD,$AF,$F3,$17,$F9,$CB,$E4,$6E,$CC,$7B,$4F,$F3,$B9,$DD
	.dc.b	$78,$7C,$BF,$F0,$3F,$E5,$3D,$5B,$5F,$ED,$79,$D0,$9D,$36,$3E,$F9
	.dc.b	$EF,$F2,$7F,$A7,$63,$E1,$F3,$FF,$87,$92,$7F,$2F,$97,$F8,$45,$E1
	.dc.b	$AF,$E1,$DF,$8B,$9B,$E5,$BE,$21,$32,$F8,$8F,$F6,$8B,$7B,$5F,$DA
	.dc.b	$3C,$BF,$EA,$CD,$C3,$FA,$B3,$5F,$EA,$4D,$7F,$C4,$F8,$3E,$5F,$FC
	.dc.b	$53,$87,$CF,$6C,$FC,$39,$D6,$3E,$58,$C4,$E9,$FE,$18,$7F,$80,$4F
	.dc.b	$8A,$71,$F9,$8F,$BC,$49,$FB,$CF,$AF,$EC,$3F,$C0,$31,$FE,$DC,$DB
	.dc.b	$ED,$46,$23,$FB,$83,$6B,$FB,$63,$5F,$C7,$1E,$13,$C7,$C7,$F9,$29
	.dc.b	$FC,$87,$CB,$FF,$BB,$9F,$63,$67,$C3,$92,$31,$1F,$19,$B8,$79,$23
	.dc.b	$F2,$6D,$E4,$F4,$6E,$CB,$FC,$37,$18,$6E,$3F,$12,$FC,$BC,$5F,$88
	.dc.b	$B7,$D7,$FF,$88,$BC,$BE,$6D,$BF,$F0,$83,$6F,$FC,$18,$D7,$F6,$BE
	.dc.b	$FF,$E6,$5F,$F0,$83,$5C,$9F,$38,$C9,$4F,$DE,$FD,$B6,$F4,$FD,$53
	.dc.b	$71,$34,$E9,$3F,$3C,$FA,$3B,$FF,$00,$71,$5F,$B3,$BE,$44,$8F,$04
	.dc.b	$65,$F1,$1C,$C9,$3C,$C9,$9F,$F8,$EE,$32,$D8,$F0,$7C,$13,$E0,$FF
	.dc.b	$DF,$9E,$B7,$F7,$AB,$FF,$26,$CE,$7D,$6F,$F2,$7C,$E6,$53,$95,$19
	.dc.b	$8C,$36,$BA,$F9,$9F,$E5,$8A,$E3,$18,$7B,$B5,$CD,$F8,$27,$FD,$76
	.dc.b	$7F,$BA,$49,$1F,$E2,$8C,$18,$BA,$BC,$D3,$FE,$ED,$AD,$BC,$8E,$13
	.dc.b	$B8,$FF,$2C,$69,$FE,$6C,$7F,$2A,$DE,$0F,$F2,$D3,$6B,$F0,$73,$FE
	.dc.b	$AA,$7A,$10,$F0,$1D,$AF,$C7,$AF,$E1,$A3,$FC,$4F,$7F,$E6,$5D,$BA
	.dc.b	$47,$E3,$0C,$9F,$E6,$37,$33,$F8,$1B,$F2,$CE,$8F,$CB,$B8,$3F,$46
	.dc.b	$BE,$2A,$92,$FF,$4B,$19,$7C,$D3,$F1,$A3,$D2,$7A,$EB,$F7,$5C,$3F
	.dc.b	$B5,$8F,$D8,$D7,$F0,$E6,$9F,$C3,$8A,$7F,$03,$7C,$C7,$F0,$C2,$F3
	.dc.b	$7F,$09,$6F,$37,$FB,$53,$0F,$99,$5E,$22,$1F,$EB,$A7,$CD,$87,$F0
	.dc.b	$DF,$9C,$F2,$12,$FC,$8F,$27,$FB,$57,$47,$5B,$FE,$5D,$1F,$86,$32
	.dc.b	$5C,$53,$B9,$FB,$87,$84,$F8,$E4,$3F,$91,$F4,$FF,$BB,$3E,$9F,$FA
	.dc.b	$25,$9A,$49,$FE,$9E,$7F,$BE,$4F,$E9,$E5,$3F,$C1,$3D,$9F,$EB,$EF
	.dc.b	$3B,$FA,$BC,$66,$BF,$C0,$9D,$BE,$58,$FE,$AA,$D8,$C9,$EB,$75,$1E
	.dc.b	$93,$D7,$CB,$FA,$8E,$3F,$D3,$C6,$4F,$F0,$CF,$7F,$E8,$DF,$F5,$D3
	.dc.b	$5F,$F5,$B3,$5D,$9C,$A7,$A7,$07,$E2,$E3,$A1,$D7,$0F,$F0,$3F,$E2
	.dc.b	$DF,$E0,$62,$47,$F6,$9F,$09,$29,$FF,$66,$70,$7F,$B8,$3E,$2E,$D9
	.dc.b	$A4,$7F,$B9,$78,$3F,$0E,$7F,$C4,$F8,$FF,$89,$63,$07,$C4,$CB,$CF
	.dc.b	$5F,$6C,$FC,$7B,$87,$FC,$C3,$E8,$FF,$95,$08,$FF,$98,$9F,$F2,$47
	.dc.b	$D5,$C8,$FF,$29,$B5,$FF,$34,$FC,$FF,$F9,$70,$8F,$F9,$A9,$FF,$28
	.dc.b	$7C,$9C,$8F,$F2,$DB,$5F,$F3,$8F,$8B,$FE,$64,$23,$FE,$72,$7E,$76
	.dc.b	$1E,$78,$F4,$F7,$FF,$1D,$7F,$CE,$27,$F6,$F2,$CF,$FA,$94,$8F,$F9
	.dc.b	$F3,$E7,$FF,$CE,$24,$70,$7F,$A2,$F4,$FF,$44,$3F,$C7,$3C,$5F,$F6
	.dc.b	$33,$7C,$FF,$C2,$71,$7F,$99,$FD,$1F,$B7,$F5,$1F,$E9,$52,$57,$FA
	.dc.b	$65,$FF,$50,$7E,$E3,$23,$FD,$48,$D7,$FD,$23,$F2,$FF,$E8,$42,$3F
	.dc.b	$E9,$27,$E0,$EB,$EF,$8F,$CE,$B5,$78,$2F,$63,$FD,$81,$5F,$EC,$52
	.dc.b	$4F,$C6,$7F,$DA,$39,$3F,$04,$A4,$9F,$F6,$9C,$65,$FF,$31,$7A,$1F
	.dc.b	$ED,$E8,$FE,$18,$DD,$FF,$DF,$B1,$FE,$89,$E6,$FD,$B4,$E6,$B8,$D7
	.dc.b	$E8,$3A,$77,$3F,$D1,$9B,$19,$9D,$DF,$E8,$5F,$A8,$40,$7F,$A4,$B8
	.dc.b	$3F,$6D,$A7,$FB,$06,$3E,$C7,$B7,$AF,$C5,$BF,$BA,$47,$FA,$62,$30
	.dc.b	$FF,$BA,$F8,$D1,$88,$EB,$5F,$E8,$98,$C4,$78,$3D,$E7,$BF,$E2,$FC
	.dc.b	$1B,$AC,$A7,$2A,$33,$18,$6C,$F1,$9F,$17,$A7,$5F,$0E,$28,$C4,$7F
	.dc.b	$AC,$1B,$8F,$FA,$75,$B2,$F8,$2F,$8B,$F1,$FF,$02,$BF,$D4,$51,$8B
	.dc.b	$F8,$0D,$7B,$B5,$FE,$2B,$C2,$F8,$5F,$E2,$FF,$A2,$7D,$86,$CF,$E1
	.dc.b	$1A,$FF,$9E,$27,$87,$0A,$BC,$59,$3D,$8F,$B3,$FA,$1E,$1B,$87,$D3
	.dc.b	$6B,$F0,$9A,$FF,$B4,$3C,$9F,$C0,$F4,$74,$6C,$7D,$B6,$BF,$29,$AF
	.dc.b	$FB,$60,$FE,$5C,$7F,$BF,$B9,$FF,$DE,$64,$7F,$93,$1F,$CF,$3C,$1F
	.dc.b	$E2,$0F,$F4,$9D,$FF,$5D,$EF,$F2,$8F,$C3,$F6,$23,$11,$FB,$48,$C3
	.dc.b	$F9,$B3,$71,$AD,$BF,$DE,$1E,$CF,$EA,$08,$FE,$B9,$FE,$75,$DD,$FB
	.dc.b	$FA,$F2,$37,$1F,$E4,$95,$FE,$F9,$F8,$9F,$D8,$91,$FD,$BB,$FD,$9B
	.dc.b	$E2,$9E,$65,$24,$FF,$66,$8C,$BF,$D7,$EF,$B1,$36,$C1,$FC,$0A,$3F
	.dc.b	$7E,$43,$07,$F9,$13,$47,$C5,$FE,$E3,$E9,$FF,$29,$18,$D1,$18,$9D
	.dc.b	$47,$F9,$AF,$7F,$E3,$91,$88,$FE,$43,$87,$8B,$C1,$FF,$1E,$62,$F9
	.dc.b	$17,$C7,$7F,$6E,$F8,$2F,$E2,$B1,$89,$46,$27,$58,$7F,$BF,$EB,$53
	.dc.b	$FC,$23,$E0,$DF,$1E,$F5,$BF,$88,$F9,$39,$8C,$60,$8C,$4E,$A3,$FD
	.dc.b	$09,$5C,$A3,$11,$EA,$CF,$F9,$EE,$9F,$E7,$CA,$EF,$F2,$1F,$E0,$6F
	.dc.b	$4B,$B3,$6F,$9E,$48,$C4,$7F,$24,$AE,$3B,$2F,$E0,$F8,$BF,$E1,$0F
	.dc.b	$46,$C9,$A4,$BF,$D0,$F4,$FE,$80,$9C,$DF,$D3,$EB,$E7,$7C,$8F,$F3
	.dc.b	$7A,$FF,$7B,$F1,$23,$11,$FE,$1E,$8C,$4F,$E8,$63,$13,$AE,$5F,$EB
	.dc.b	$1F,$25,$FC,$EB,$A0,$D2,$57,$DF,$3C,$B8,$FE,$4F,$8B,$EE,$F8,$FF
	.dc.b	$E4,$5C,$CD,$26,$7F,$93,$F9,$3F,$E9,$3E,$AF,$E4,$9B,$8F,$E3,$F3
	.dc.b	$46,$23,$F7,$AD,$7F,$A8,$ED,$FE,$4D,$E7,$FE,$82,$D8,$FF,$52,$46
	.dc.b	$34,$FF,$5C,$3F,$92,$F8,$4F,$AF,$3F,$15,$B9,$7E,$65,$BA,$1D,$EF
	.dc.b	$FD,$F7,$9B,$F1,$37,$FC,$35,$B1,$98,$FC,$91,$EB,$8D,$AC,$BF,$CC
	.dc.b	$9E,$17,$97,$A3,$EA,$4E,$74,$FF,$67,$4E,$4F,$EB,$BF,$2A,$F7,$7E
	.dc.b	$ED,$23,$1A,$F2,$6D,$70,$AD,$E3,$E0,$BF,$F6,$44,$F5,$69,$FD,$DA
	.dc.b	$BF,$DB,$D7,$FE,$78,$6F,$1F,$8B,$2F,$F3,$FC,$62,$B8,$46,$09,$5C
	.dc.b	$5B,$5C,$63,$04,$70,$49,$E7,$19,$6C,$65,$FF,$7F,$57,$85,$B5,$D9
	.dc.b	$D9,$0C,$D7,$36,$D7,$37,$FC,$37,$5C,$DB,$5C,$DD,$9C,$CF,$F1,$0F
	.dc.b	$C4,$3E,$21,$F1,$1F,$E1,$29,$4E,$63,$0D,$8F,$F0,$BB,$5F,$A1,$E4
	.dc.b	$FF,$B5,$1F,$6B,$5F,$2B,$5F,$C0,$37,$0F,$EF,$99,$78,$5B,$3F,$88
	.dc.b	$9D,$12,$74,$D8,$D4,$60,$8E,$08,$79,$9C,$B4,$C9,$E8,$72,$D3,$27
	.dc.b	$A9,$CB,$4C,$9E,$C7,$2D,$33,$F8,$B7,$F4,$87,$E7,$9F,$E4,$C7,$18
	.dc.b	$7F,$90,$9B,$19,$F0,$7F,$CA,$F3,$FA,$F2,$47,$F9,$53,$6F,$F6,$B4
	.dc.b	$62,$3F,$CB,$A7,$EC,$F4,$FB,$93,$FE,$DB,$8F,$DC,$10,$FF,$2D,$74
	.dc.b	$F6,$69,$E6,$3F,$C9,$6E,$5A,$64,$FF,$69,$FB,$DF,$D7,$B2,$7C,$37
	.dc.b	$F9,$F2,$30,$DE,$4F,$F6,$CB,$67,$75,$EB,$E4,$65,$E8,$6C,$79,$CF
	.dc.b	$6E,$5E,$B7,$FC,$E5,$F0,$7F,$E7,$CF,$8C,$FF,$23,$B6,$32,$23,$2D
	.dc.b	$3E,$3E,$0C,$FF,$A4,$BF,$E7,$43,$1D,$DF,$9B,$69,$98,$DC,$F9,$4F
	.dc.b	$E7,$B8,$FA,$1F,$66,$5E,$B3,$FC,$FF,$D0,$F4,$3F,$B0,$4E,$1A,$FF
	.dc.b	$98,$3C,$0F,$DA,$4E,$DF,$DB,$43,$F4,$BF,$E9,$F9,$7D,$38,$C7,$7A
	.dc.b	$31,$3A,$E3,$FF,$47,$B6,$BC,$A7,$A4,$FC,$41,$5E,$BF,$1B,$C2,$FA
	.dc.b	$36,$CD,$3F,$D3,$7F,$07,$FE,$0D,$8C,$DF,$FA,$AA,$31,$3A,$F2,$39
	.dc.b	$CE,$5A,$FD,$ED,$BE,$53,$76,$F9,$4F,$9F,$ED,$FF,$BF,$BF,$05,$BA
	.dc.b	$7F,$AB,$1B,$1F,$EB,$A6,$BF,$EC,$2C,$9F,$96,$5E,$EF,$F6,$9B,$67
	.dc.b	$F0,$AD,$78,$35,$FB,$F8,$79,$1F,$96,$43,$FD,$67,$E4,$7F,$B4,$35
	.dc.b	$FF,$4D,$B6,$3F,$D9,$4D,$76,$23,$FD,$8B,$FA,$AB,$D2,$FC,$0B,$67
	.dc.b	$F9,$C3,$5D,$C7,$E1,$BA,$7F,$4D,$1F,$24,$FB,$CE,$1F,$49,$8D,$BE
	.dc.b	$7B,$83,$FD,$BD,$F6,$DF,$92,$E6,$34,$91,$F0,$4E,$C7,$C1,$39,$1F
	.dc.b	$C5,$31,$57,$F4,$36,$ED,$FD,$45,$B3,$F8,$A9,$FE,$95,$CD,$FA,$7F
	.dc.b	$3F,$FD,$F3,$F2,$BF,$EC,$46,$D7,$FB,$CD,$AF,$FB,$89,$AF,$88,$87
	.dc.b	$E7,$8F,$34,$3F,$57,$7D,$3F,$F5,$98,$C3,$63,$39,$7F,$BE,$27,$2F
	.dc.b	$F0,$7E,$27,$C1,$7E,$C4,$FC,$59,$70,$FC,$5B,$5F,$C5,$B5,$FD,$C4
	.dc.b	$3E,$13,$F7,$5C,$BF,$94,$6C,$FE,$51,$AF,$E5,$1A,$F2,$F9,$7F,$E0
	.dc.b	$5F,$2F,$F9,$C6,$C7,$E7,$1A,$FE,$71,$AB,$6F,$E2,$DF,$70,$87,$E3
	.dc.b	$FF,$EF,$48,$C3,$63,$27,$FA,$C7,$91,$3A,$8F,$D5,$FC,$87,$F1,$5E
	.dc.b	$CF,$D7,$B6,$33,$BF,$F4,$D3,$F6,$B1,$98,$FC,$69,$0F,$B7,$CF,$F9
	.dc.b	$B9,$FD,$DF,$A1,$B9,$F9,$6D,$8C,$9D,$78,$3F,$CA,$38,$7F,$00,$F8
	.dc.b	$FF,$E7,$D0,$69,$27,$F9,$CF,$D5,$7F,$2D,$C3,$FB,$16,$C8,$C6,$1F
	.dc.b	$DC,$A7,$E5,$46,$1F,$F6,$A3,$5F,$E1,$7F,$2D,$FC,$DF,$F4,$EF,$E2
	.dc.b	$2D,$8F,$E2,$2D,$77,$82,$7F,$8B,$FA,$BF,$CF,$FF,$31,$B2,$69,$2B
	.dc.b	$F4,$C7,$F6,$13,$F4,$C5,$3F,$01,$C3,$FE,$E2,$71,$9F,$EC,$5D,$1F
	.dc.b	$A9,$3E,$91,$4F,$D7,$3F,$0F,$BF,$FC,$11,$1F,$09,$B1,$98,$F9,$39
	.dc.b	$FF,$1B,$D3,$F2,$D8,$FE,$2A,$7F,$28,$D7,$F3,$8D,$3F,$9C,$F2,$7F
	.dc.b	$5F,$E4,$FE,$73,$E7,$FF,$61,$FD,$56,$D1,$A4,$BF,$E7,$BC,$3F,$75
	.dc.b	$FC,$14,$FE,$EE,$3F,$07,$E7,$7E,$27,$8C,$D2,$4F,$E0,$B7,$F9,$0F
	.dc.b	$F1,$5B,$FF,$32,$6D,$F2,$63,$07,$F2,$FF,$17,$F9,$44,$FC,$8D,$FF
	.dc.b	$9E,$76,$FE,$26,$FE,$33,$63,$33,$F8,$AE,$1F,$E3,$B5,$F9,$1F,$84
	.dc.b	$F9,$B4,$69,$2B,$E7,$D7,$FA,$21,$E4,$7F,$AC,$CB,$EA,$F7,$8D,$47
	.dc.b	$F6,$0F,$17,$E1,$13,$3F,$D5,$3F,$AF,$C7,$EE,$7B,$FC,$68,$D2,$3F
	.dc.b	$D1,$4F,$C4,$13,$AF,$F7,$5C,$1F,$C2,$30,$FE,$59,$DB,$FB,$42,$3F
	.dc.b	$E3,$5F,$CC,$7F,$C3,$CB,$E7,$7F,$60,$A3,$31,$86,$E1,$FE,$40,$6B
	.dc.b	$FE,$68,$FD,$3B,$FC,$B5,$D5,$FB,$98,$46,$1F,$DE,$35,$E1,$EB,$7F
	.dc.b	$0F,$7D,$5E,$4F,$F6,$A6,$F2,$FF,$08,$8F,$F1,$DF,$27,$F6,$29,$C1
	.dc.b	$0F,$C2,$10,$F3,$72,$7F,$94,$72,$FB,$9A,$7C,$18,$C1,$FD,$9A,$9E
	.dc.b	$6F,$F3,$7F,$63,$F1,$0D,$C3,$F8,$8B,$5F,$E7,$9C,$7F,$C6,$63,$11
	.dc.b	$FC,$63,$E3,$7F,$A0,$7C,$CF,$E4,$5B,$1F,$C6,$9A,$F2,$C1,$18,$8F
	.dc.b	$E6,$BA,$7C,$3E,$4F,$F0,$84,$7F,$07,$D3,$EB,$57,$B4,$8F,$DD,$7D
	.dc.b	$9F,$77,$FE,$29,$8F,$E5,$2D,$9F,$03,$5F,$E5,$ED,$7E,$89,$F2,$CF
	.dc.b	$F2,$06,$DF,$E7,$19,$FF,$21,$46,$3A,$FF,$9E,$47,$E1,$5B,$19,$8D
	.dc.b	$BD,$8F,$A6,$78,$8A,$67,$7F,$E8,$CD,$8F,$D0,$63,$F0,$89,$97,$F8
	.dc.b	$C3,$6F,$E9,$58,$7F,$6F,$C3,$8C,$63,$0E,$AD,$8C,$E9,$FE,$8B,$9C
	.dc.b	$AF,$4B,$FD,$3B,$1F,$55,$AF,$F5,$46,$D7,$F5,$96,$CB,$F0,$5A,$4A
	.dc.b	$F8,$35,$FD,$7E,$39,$C7,$CE,$3D,$B8,$FA,$EB,$C5,$1C,$1A,$FA,$2F
	.dc.b	$B9,$CB,$2F,$87,$C5,$FC,$92,$7A,$1F,$D3,$A3,$2F,$ED,$7D,$E7,$9B
	.dc.b	$FE,$EC,$D1,$EF,$FF,$B8,$A7,$3D,$1F,$2E,$31,$5B,$6C,$66,$FF,$C3
	.dc.b	$D5,$C1,$FE,$55,$3A,$69,$AA,$E4,$DA,$E4,$FF,$BC,$67,$F8,$56,$1F
	.dc.b	$E6,$0C,$7C,$B3,$FE,$C4,$9C,$E9,$FE,$E0,$6C,$EA,$7F,$D4,$11,$82
	.dc.b	$3F,$E0,$AC,$FF,$C6,$91,$F2,$E3,$E0,$F0,$65,$A7,$E0,$4A,$7F,$A6
	.dc.b	$23,$9B,$D8,$F2,$1F,$E1,$6D,$BA,$CD,$8E,$3B,$FE,$05,$BB,$FE,$1D
	.dc.b	$B8,$70,$7C,$18,$79,$49,$7F,$E2,$2E,$0F,$F2,$67,$D4,$FE,$69,$B1
	.dc.b	$9A,$4E,$5B,$1E,$67,$D0,$43,$D8,$7F,$06,$36,$7E,$D6,$3F,$4F,$C7
	.dc.b	$DA,$31,$3F,$BF,$3E,$29,$FD,$26,$BF,$DD,$57,$FB,$A8,$C1,$AE,$3F
	.dc.b	$E1,$51,$FB,$EA,$FC,$01,$E5,$3C,$67,$97,$3F,$F2,$36,$1F,$E6,$BF
	.dc.b	$2F,$C7,$26,$93,$0F,$EA,$DC,$BE,$29,$B1,$F8,$57,$FD,$FD,$8F,$F3
	.dc.b	$08,$C1,$FC,$B0,$D4,$79,$0F,$8C,$53,$CA,$7F,$3F,$3F,$C3,$A7,$F3
	.dc.b	$D2,$9F,$CF,$67,$FD,$05,$E1,$FF,$43,$8C,$9F,$E5,$EE,$8F,$F4,$37
	.dc.b	$E2,$7F,$D4,$20,$D2,$3F,$D7,$2F,$CF,$87,$F5,$CD,$3F,$BB,$74,$7B
	.dc.b	$EB,$F9,$BB,$5F,$39,$E9,$27,$2F,$EB,$67,$FC,$4D,$CB,$FA,$49,$FF
	.dc.b	$37,$B5,$FE,$AC,$D7,$FC,$EA,$D7,$9B,$FB,$37,$ED,$1D,$FB,$1D,$8C
	.dc.b	$9F,$05,$FE,$07,$9C,$C6,$2B,$FC,$0F,$5B,$AD,$36,$3F,$C2,$53,$CF
	.dc.b	$E7,$7F,$C2,$CD,$8C,$A8,$CB,$63,$FD,$68,$7F,$B2,$3C,$0F,$8D,$0F
	.dc.b	$07,$FC,$55,$9F,$F8,$91,$B5,$C2,$75,$5F,$36,$3E,$0F,$5F,$E2,$35
	.dc.b	$F9,$91,$F9,$BE,$6F,$F3,$C7,$C3,$FF,$8F,$49,$1F,$E4,$A7,$B7,$89
	.dc.b	$ED,$8F,$F2,$46,$3F,$0D,$BB,$6A,$31,$1F,$E8,$51,$F3,$BF,$E4,$92
	.dc.b	$3F,$E5,$67,$FC,$C5,$C5,$FE,$D3,$AB,$3C,$B1,$FF,$31,$37,$B7,$FC
	.dc.b	$C4,$DF,$1B,$FD,$B5,$F0,$FF,$E5,$52,$47,$F9,$89,$FE,$46,$FE,$3B
	.dc.b	$5F,$82,$FE,$E7,$B7,$FD,$49,$F4,$1F,$E7,$D2,$47,$FA,$B1,$F9,$DD
	.dc.b	$7F,$06,$3F,$DF,$6D,$DB,$EA,$37,$8F,$FD,$17,$1F,$08,$FB,$1E,$5F
	.dc.b	$DA,$8F,$F6,$94,$7E,$F1,$BC,$BF,$E9,$63,$FD,$3B,$FB,$7C,$C3,$EE
	.dc.b	$4B,$CE,$FC,$B4,$E6,$7F,$D5,$7F,$B4,$7F,$AB,$88,$F0,$F4,$7F,$39
	.dc.b	$1F,$E7,$9E,$39,$3F,$E8,$66,$BF,$C2,$B7,$F8,$AF,$C7,$ED,$FD,$43
	.dc.b	$FE,$CC,$D7,$FD,$44,$47,$FD,$A2,$FF,$B5,$BF,$4D,$91,$E8,$F5,$3F
	.dc.b	$59,$18,$8F,$F6,$FF,$E3,$7F,$EE,$4E,$03,$49,$1F,$EE,$C8,$FC,$23
	.dc.b	$4F,$C1,$38,$7F,$CA,$CD,$7F,$75,$BF,$CC,$7F,$D8,$71,$F3,$F0,$FA
	.dc.b	$47,$D3,$F5,$FE,$C1,$23,$EC,$52,$3F,$01,$5F,$AA,$E1,$FC,$5B,$FC
	.dc.b	$BE,$BF,$69,$EC,$7E,$04,$8F,$27,$F1,$1D,$2E,$67,$7A,$DC,$5F,$EC
	.dc.b	$59,$CC,$60,$FE,$07,$5C,$6B,$85,$6D,$B3,$F7,$F0,$7D,$07,$FB,$C0
	.dc.b	$91,$97,$F8,$47,$B7,$FC,$7F,$DF,$B4,$69,$2B,$F2,$3E,$EF,$F2,$5F
	.dc.b	$A2,$FE,$20,$D8,$CD,$27,$2D,$8F,$E3,$7E,$CB,$DF,$7C,$07,$F4,$18
	.dc.b	$C3,$63,$35,$FD,$8E,$72,$FF,$19,$E8,$FC,$9D,$FE,$57,$C7,$C1,$FD
	.dc.b	$A7,$83,$E7,$69,$FC,$12,$7E,$4F,$0F,$B5,$78,$3F,$EF,$B2,$3F,$CC
	.dc.b	$07,$C0,$B1,$FE,$54,$DC,$BF,$B8,$B7,$97,$FD,$98,$DA,$FE,$9B,$D9
	.dc.b	$FA,$A7,$F1,$07,$8B,$C7,$FE,$AD,$A7,$E0,$DB,$99,$A7,$F5,$9A,$E4
	.dc.b	$4C,$BE,$0D,$F4,$3F,$70,$E2,$3C,$1F,$04,$F1,$7F,$B1,$36,$BF,$C1
	.dc.b	$2D,$CF,$FD,$C8,$D8,$F8,$FF,$07,$FE,$1B,$69,$FE,$1B,$2F,$A5,$F9
	.dc.b	$78,$D4,$78,$B8,$7E,$9D,$FF,$30,$DF,$FA,$C4,$FE,$3C,$FE,$36,$53
	.dc.b	$F4,$4E,$8F,$AC,$E3,$B7,$3C,$5F,$EF,$9F,$45,$F1,$3F,$BF,$AF,$F7
	.dc.b	$2B,$63,$31,$FD,$48,$C9,$76,$F9,$59,$79,$A3,$FC,$07,$C7,$FE,$AD
	.dc.b	$C9,$F2,$1F,$73,$D7,$FF,$02,$36,$33,$5F,$87,$AE,$13,$F8,$98,$7C
	.dc.b	$43,$E2,$38,$C7,$E3,$63,$17,$CA,$3F,$BF,$C6,$2B,$93,$63,$F1,$7C
	.dc.b	$9F,$74,$99,$7F,$90,$A7,$9F,$91,$C3,$1F,$E2,$F1,$8B,$F0,$47,$F1
	.dc.b	$2E,$77,$C4,$FE,$40,$93,$FA,$98,$CB,$F1,$4F,$B9,$C7,$F6,$FE,$2B
	.dc.b	$E8,$EB,$FE,$54,$6C,$7E,$A6,$11,$89,$FF,$0A,$35,$FE,$41,$D5,$FC
	.dc.b	$8F,$CE,$FE,$45,$E4,$7F,$21,$6C,$FE,$E1,$AF,$0C,$7F,$CB,$73,$FC
	.dc.b	$CB,$8F,$ED,$E5,$E5,$F5,$7F,$C7,$5E,$37,$8F,$3F,$F1,$CB,$63,$38
	.dc.b	$F9,$1A,$F5,$F3,$FF,$A5,$69,$FC,$DD,$B9,$FF,$71,$9F,$F0,$A3,$78
	.dc.b	$7C,$5C,$7F,$53,$3B,$C7,$F8,$F8,$34,$93,$F8,$7F,$81,$7C,$6F,$E6
	.dc.b	$DB,$8F,$E6,$5B,$3B,$9F,$D0,$7C,$2B,$E3,$BF,$C7,$CD,$E4,$FF,$42
	.dc.b	$B7,$3F,$CA,$37,$1F,$E3,$CD,$7F,$C9,$ED,$7F,$AA,$35,$34,$95,$E0
	.dc.b	$25,$78,$1B,$5E,$06,$D7,$81,$AF,$81,$AF,$81,$AF,$81,$B5,$E0,$AE
	.dc.b	$C4,$AE,$CD,$AE,$CD,$7F,$DF,$C7,$F3,$DF,$83,$7C,$1B,$E0,$FF,$71
	.dc.b	$29,$CC,$61,$B1,$F2,$BD,$1F,$F6,$23,$F1,$9A,$FC,$66,$E3,$FE,$09
	.dc.b	$C3,$AB,$67,$94,$62,$76,$47,$04,$8D,$90,$FB,$31,$B7,$07,$DA,$70
	.dc.b	$EA,$39,$34,$D9,$E9,$72,$F1,$6B,$93,$67,$B5,$CB,$4C,$CF,$B5,$EC
	.dc.b	$43,$A1,$A9,$F2,$B8,$3F,$B8,$73,$7F,$6F,$DF,$EA,$46,$38,$3F,$53
	.dc.b	$97,$F2,$DA,$D3,$63,$33,$FD,$E7,$F2,$3F,$ED,$59,$72,$8F,$F5,$5C
	.dc.b	$62,$BF,$9B,$36,$38,$79,$1F,$EC,$19,$FE,$75,$7F,$CE,$65,$E5,$63
	.dc.b	$B7,$FC,$70,$D8,$C9,$FE,$09,$9E,$11,$E5,$CF,$33,$FE,$BC,$CF,$DD
	.dc.b	$39,$24,$7F,$84,$7F,$51,$7C,$B7,$FB,$4E,$7F,$A6,$46,$35,$E7,$5F
	.dc.b	$E4,$98,$C4,$EA,$B8,$4F,$F8,$6F,$CA,$FF,$0D,$3F,$9E,$6B,$FE,$00
	.dc.b	$CB,$F4,$2A,$32,$FD,$1C,$BF,$4B,$97,$D5,$8F,$61,$0F,$EC,$5E,$2F
	.dc.b	$E8,$BC,$43,$49,$A7,$85,$EF,$86,$AB,$04,$3A,$1F,$1F,$1F,$E5,$2E
	.dc.b	$0F,$F3,$46,$1E,$37,$FB,$94,$78,$49,$7F,$BC,$C3,$FC,$71,$3F,$80
	.dc.b	$DF,$FD,$FD,$F1,$F6,$8D,$24,$FC,$13,$C0,$30,$F0,$3E,$5C,$FF,$84
	.dc.b	$F0,$FE,$2E,$3F,$1F,$87,$E2,$DF,$C6,$E1,$F9,$37,$FD,$01,$3B,$7F
	.dc.b	$D0,$5F,$18,$F9,$95,$3E,$4B,$FC,$19,$39,$8C,$57,$F9,$5E,$31,$3A
	.dc.b	$AE,$2F,$A3,$2F,$3D,$7C,$8D,$7E,$5B,$ED,$39,$57,$83,$E4,$B9,$62
	.dc.b	$69,$27,$FC,$CD,$F3,$0F,$99,$FE,$42,$DA,$FE,$56,$D9,$FF,$12,$35
	.dc.b	$F0,$1F,$24,$F3,$61,$E4,$7F,$D4,$82,$38,$38,$6B,$FE,$A9,$6B,$BA
	.dc.b	$FE,$67,$E1,$BE,$36,$93,$0F,$E7,$F1,$97,$F3,$B8,$78,$6F,$E3,$70
	.dc.b	$60,$8F,$F8,$F6,$B8,$36,$FA,$36,$FF,$A5,$FE,$AB,$FD,$F1,$FC,$6C
	.dc.b	$7F,$A3,$63,$1C,$5F,$E4,$76,$C6,$7C,$1F,$CA,$BF,$18,$F1,$9F,$1B
	.dc.b	$1F,$EA,$FF,$AE,$FF,$15,$EC,$FF,$28,$36,$7F,$D3,$4D,$7E,$5B,$56
	.dc.b	$6F,$85,$78,$5F,$ED,$3F,$D6,$FF,$D4,$53,$98,$C5,$7F,$B4,$AB,$6A
	.dc.b	$B7,$18,$7F,$C8,$5F,$18,$FB,$4F,$CD,$C6,$1B,$5F,$7B,$75,$6D,$E9
	.dc.b	$19,$8C,$57,$4F,$B8,$FD,$2F,$EA,$D6,$CD,$25,$F5,$FC,$09,$F0,$3F
	.dc.b	$AE,$8C,$35,$FF,$48,$C6,$97,$67,$F9,$6E,$30,$FF,$9F,$7E,$0F,$F6
	.dc.b	$B2,$FA,$8C,$63,$FE,$FB,$6C,$64,$FF,$66,$4F,$08,$FF,$43,$0F,$75
	.dc.b	$2F,$B2,$CC,$61,$B1,$93,$FD,$AD,$3C,$5C,$1F,$A8,$FC,$2F,$F7,$0F
	.dc.b	$F3,$A0,$E3,$AF,$FD,$2C,$D8,$FF,$70,$35,$FD,$8C,$F5,$7C,$24,$3C
	.dc.b	$37,$FD,$23,$0F,$87,$CD,$F0,$09,$3F,$03,$2F,$6C,$7E,$1B,$EB,$3D
	.dc.b	$FF,$A5,$5A,$34,$8F,$E0,$7E,$1B,$F9,$1F,$E8,$9F,$8C,$6C,$67,$04
	.dc.b	$E5,$B1,$FA,$B5,$F7,$3F,$AF,$ED,$FE,$5A,$D8,$CC,$EE,$3F,$5C,$C3
	.dc.b	$F8,$98,$F9,$7F,$E4,$93,$FC,$EE,$31,$E2,$7F,$3F,$7F,$8C,$CE,$B4
	.dc.b	$FF,$62,$36,$3F,$39,$CF,$FC,$FB,$F3,$CF,$CD,$12,$33,$82,$3F,$25
	.dc.b	$CF,$F8,$6F,$B8,$B4,$69,$1F,$3D,$BD,$CF,$E1,$71,$FC,$B2,$31,$87
	.dc.b	$F1,$48,$C4,$E9,$FC,$9C,$EB,$C4,$FC,$DF,$A3,$27,$EE,$1F,$E1,$21
	.dc.b	$C9,$FA,$48,$F9,$9E,$5C,$9F,$A4,$D7,$FA,$01,$FC,$74,$FE,$D9,$D3
	.dc.b	$FC,$C5,$BE,$2F,$F9,$D3,$7F,$9F,$7F,$2B,$87,$FC,$3B,$5F,$39,$AF
	.dc.b	$EA,$A7,$6F,$E9,$BD,$EF,$D6,$E7,$FB,$06,$C6,$4F,$84,$7E,$E5,$CB
	.dc.b	$D7,$6F,$F5,$0B,$F8,$2A,$FC,$0F,$AD,$E3,$9C,$CE,$E3,$F0,$EF,$EF
	.dc.b	$7E,$3F,$FB,$DF,$E5,$58,$EC,$46,$27,$F1,$B1,$8B,$56,$EB,$55,$E3
	.dc.b	$FA,$1F,$01,$1F,$26,$9F,$E0,$88,$FC,$C4,$7C,$E2,$1F,$E0,$CF,$15
	.dc.b	$F7,$DF,$E9,$AC,$7F,$98,$B6,$33,$1E,$A3,$27,$F6,$1F,$21,$F5,$5F
	.dc.b	$C6,$3E,$8B,$FA,$0B,$70,$FE,$8C,$D7,$FD,$FF,$3C,$14,$FE,$B2,$9F
	.dc.b	$AD,$BD,$7F,$F1,$0C,$E6,$31,$3C,$2B,$FA,$44,$62,$75,$5C,$27,$F1
	.dc.b	$9E,$AF,$F8,$81,$FC,$59,$FE,$49,$BE,$FF,$81,$FC,$1C,$0D,$24,$7F
	.dc.b	$80,$27,$9F,$47,$E5,$6F,$FC,$57,$C6,$FC,$F1,$F7,$5F,$EC,$48,$C3
	.dc.b	$63,$38,$7F,$BC,$9B,$87,$B7,$D6,$FF,$1F,$FE,$31,$68,$D2,$4F,$C0
	.dc.b	$F0,$7D,$04,$9F,$AD,$AF,$F0,$37,$D1,$3F,$80,$3E,$09,$FE,$0B,$7E
	.dc.b	$7E,$FF,$04,$8F,$27,$E2,$7E,$03,$FE,$52,$F8,$2F,$EB,$F5,$FE,$54
	.dc.b	$B4,$E6,$FE,$9B,$67,$FA,$DC,$7D,$C8,$C4,$7F,$86,$8F,$97,$DF,$E6
	.dc.b	$7F,$B0,$FD,$B3,$E2,$FF,$B6,$DF,$F3,$08,$C3,$D2,$7F,$C8,$B1,$87
	.dc.b	$CA,$1C,$47,$90,$FF,$7F,$97,$E1,$BD,$6F,$94,$CC,$FA,$31,$F6,$C6
	.dc.b	$23,$D0,$7C,$BE,$2F,$C6,$1F,$E6,$C8,$E5,$CB,$D2,$B8,$B7,$3F,$F3
	.dc.b	$DB,$67,$F3,$77,$DF,$EE,$FD,$ED,$8F,$73,$F3,$E3,$2D,$8C,$9F,$A8
	.dc.b	$8C,$B4,$CD,$A3,$2D,$8E,$B3,$F8,$43,$D9,$CB,$F9,$D3,$FC,$C5,$1C
	.dc.b	$E3,$D9,$5C,$89,$1C,$B1,$F8,$75,$C7,$5F,$B9,$7D,$4F,$92,$43,$03
	.dc.b	$E1,$DA,$7F,$C0,$0F,$A3,$FF,$26,$CE,$63,$1E,$DF,$FA,$12,$B7,$E6
	.dc.b	$FF,$8A,$2F,$71,$8B,$46,$2F,$8C,$7F,$8B,$A7,$53,$9A,$E5,$7D,$6F
	.dc.b	$77,$FD,$4A,$3F,$DF,$44,$7F,$DC,$8D,$7F,$D8,$4D,$7F,$D1,$2F,$F1
	.dc.b	$9D,$3F,$D7,$6F,$CA,$D3,$FC,$DE,$7F,$1A,$8C,$B4,$CC,$F4,$3B,$90
	.dc.b	$C1,$F8,$79,$E8,$48,$E4,$D8,$D4,$60,$8F,$E7,$F7,$FF,$67,$FC,$BF
	.dc.b	$E5,$49,$3F,$A9,$7F,$5B,$B3,$66,$DF,$5B,$0F,$79,$E5,$29,$D6,$DF
	.dc.b	$2D,$FA,$12,$3F,$B5,$7F,$73,$C5,$FB,$EC,$EF,$1F,$C4,$BF,$DF,$B5
	.dc.b	$FE,$0E,$D7,$EB,$1E,$12,$9F,$18,$FE,$31,$A7,$FB,$B3,$C0,$FD,$91
	.dc.b	$31,$FE,$0C,$FE,$E7,$9B,$EB,$D7,$F4,$E6,$BF,$C1,$5A,$FF,$23,$6B
	.dc.b	$E0,$3E,$09,$0C,$4F,$F5,$EE,$7E,$2D,$3E,$51,$71,$FE,$C1,$1A,$E5
	.dc.b	$FF,$76,$7E,$31,$FD,$14,$97,$FD,$8B,$A9,$82,$7F,$CA,$FD,$BF,$D4
	.dc.b	$A3,$EB,$37,$1C,$F6,$7F,$54,$6F,$6F,$AB,$93,$EF,$71,$FD,$DC,$FD
	.dc.b	$84,$8C,$D7,$F7,$5E,$EF,$31,$32,$F5,$EB,$FD,$B1,$E5,$5C,$2F,$EB
	.dc.b	$35,$FA,$D1,$FE,$02,$EC,$F5,$FC,$97,$F7,$C8,$34,$93,$FE,$0E,$32
	.dc.b	$57,$FC,$26,$E0,$FF,$08,$F8,$3F,$EA,$E8,$FF,$6A,$F1,$FD,$2F,$0B
	.dc.b	$FC,$3F,$A6,$30,$FA,$71,$8C,$3C,$FD,$3F,$A0,$AF,$DF,$47,$C4,$F2
	.dc.b	$3F,$D0,$7F,$4D,$FE,$43,$24,$7F,$8D,$30,$47,$F8,$ED,$7E,$4F,$49
	.dc.b	$3F,$E4,$2F,$23,$ED,$C7,$F9,$37,$E9,$E4,$F3,$6B,$FE,$B6,$F3,$3F
	.dc.b	$D9,$3F,$31,$FE,$5D,$83,$48,$F8,$FE,$77,$FC,$B0,$47,$07,$BB,$9B
	.dc.b	$EB,$ED,$EF,$8F,$D6,$7C,$3B,$D9,$FC,$7F,$83,$FE,$A9,$24,$FF,$AB
	.dc.b	$DF,$F7,$0E,$12,$3F,$D0,$ED,$7F,$D1,$FB,$79,$9F,$F4,$BF,$AD,$F9
	.dc.b	$AC,$FF,$CD,$3E,$B7,$FA,$59,$B1,$99,$FF,$7C,$8F,$96,$FF,$51,$91
	.dc.b	$CB,$FE,$4D,$E5,$5F,$E9,$BF,$38,$C3,$6B,$FD,$5F,$18,$C4,$FF,$9F
	.dc.b	$7E,$56,$4F,$DB,$6B,$FC,$34,$FD,$26,$FF,$1D,$C6,$FF,$1E,$DF,$6B
	.dc.b	$FE,$CF,$24,$FF,$B5,$5F,$F6,$C7,$E8,$92,$3F,$DB,$8D,$7F,$CD,$3F
	.dc.b	$04,$5F,$BE,$7F,$BA,$48,$E4,$7C,$63,$D6,$57,$1A,$FF,$4A,$FB,$DF
	.dc.b	$C2,$2D,$AE,$33,$EC,$D8,$C9,$FE,$C0,$7F,$D3,$BF,$0A,$F7,$3C,$9E
	.dc.b	$06,$1B,$5F,$EE,$89,$D3,$FC,$2D,$B3,$FE,$A6,$F8,$67,$B9,$ED,$FC
	.dc.b	$1F,$0D,$BF,$F7,$8C,$E9,$4E,$9A,$77,$C3,$E3,$7D,$FF,$8E,$0D,$23
	.dc.b	$FE,$DA,$FC,$A7,$DC,$47,$07,$E3,$BE,$A3,$FD,$DE,$47,$07,$FB,$DB
	.dc.b	$F6,$9F,$C4,$11,$C1,$F4,$F6,$75,$3E,$47,$F9,$BF,$EF,$5F,$95,$6D
	.dc.b	$6E,$75,$4A,$D3,$67,$EF,$73,$7D,$DE,$0F,$C2,$57,$E0,$67,$F0,$3C
	.dc.b	$C9,$CC,$70,$8C,$3C,$CF,$F5,$AC,$69,$FC,$57,$C9,$BE,$77,$F0,$71
	.dc.b	$D6,$30,$FE,$10,$DC,$69,$A6,$4F,$EC,$5E,$87,$F3,$F6,$F6,$FE,$9D
	.dc.b	$FC,$E7,$6F,$E9,$FC,$2F,$F7,$4B,$63,$D8,$D7,$F3,$CD,$7F,$49,$A3
	.dc.b	$E5,$7F,$B9,$B7,$6F,$EB,$4D,$D3,$FD,$EA,$DA,$FE,$41,$1F,$CF,$32
	.dc.b	$F3,$47,$EA,$AF,$99,$2F,$9B,$FE,$00,$F9,$67,$B3,$49,$3F,$DA,$63
	.dc.b	$2F,$F2,$CF,$26,$63,$DF,$5F,$85,$8A,$F7,$3F,$6E,$31,$87,$48,$C4
	.dc.b	$EA,$7F,$01,$F2,$6F,$97,$FF,$01,$FD,$0F,$18,$C6,$68,$C4,$EA,$B8
	.dc.b	$B5,$FE,$39,$AF,$C2,$26,$7E,$6C,$39,$F9,$7C,$B1,$FF,$5C,$5F,$26
	.dc.b	$BF,$C4,$E7,$6F,$E5,$7E,$65,$F3,$5F,$E1,$8F,$0F,$DD,$18,$BF,$C4
	.dc.b	$52,$31,$3A,$AE,$13,$F6,$08,$7F,$21,$9E,$5C,$9B,$BF,$E9,$D3,$9A
	.dc.b	$FE,$94,$D8,$CB,$FC,$B7,$D8,$FF,$2F,$47,$F8,$85,$BC,$1F,$4D,$B3
	.dc.b	$FA,$F6,$BF,$23,$4F,$B7,$AF,$C8,$E2,$FF,$15,$7C,$DB,$E7,$3F,$92
	.dc.b	$DF,$F3,$78,$C3,$A8,$CB,$FC,$FE,$33,$8F,$86,$7F,$C9,$F8,$F3,$CB
	.dc.b	$CD,$18,$3F,$C7,$59,$78,$89,$8F,$E4,$31,$FE,$0D,$8F,$F9,$ED,$F2
	.dc.b	$CF,$81,$46,$71,$FC,$93,$63,$F8,$77,$CF,$3E,$7B,$FC,$A3,$E9,$7E
	.dc.b	$5E,$31,$9A,$31,$3A,$AE,$2D,$7F,$AC,$6B,$FE,$50,$9F,$D0,$36,$3F
	.dc.b	$8E,$35,$DE,$5F,$E9,$29,$FD,$33,$FC,$51,$BD,$7F,$D9,$DB,$3E,$33
	.dc.b	$F9,$34,$ED,$FC,$A6,$FF,$8C,$7F,$09,$C5,$F8,$92,$3E,$29,$FF,$50
	.dc.b	$C7,$99,$B1,$9D,$BF,$33,$8F,$F9,$AA,$BF,$75,$F0,$2F,$8E,$FF,$43
	.dc.b	$B7,$5F,$F3,$73,$78,$37,$3A,$24,$E9,$B1,$A8,$C1,$1C,$13,$1F,$F4
	.dc.b	$1C,$F0,$6C,$ED,$B3,$B7,$FC,$15,$5E,$37,$24,$33,$5C,$DB,$5C,$DB
	.dc.b	$5C,$DF,$F0,$C5,$73,$72,$73,$33,$5C,$DB,$5C,$DB,$5C,$D5,$73,$72
	.dc.b	$73,$1F,$1C,$F8,$C7,$C6,$FF,$2B,$9C,$C6,$1F,$E5,$D3,$B9,$D3,$63
	.dc.b	$EC,$F5,$FF,$89,$DF,$F5,$6B,$5F,$8E,$D7,$DE,$D7,$CE,$DB,$FE,$65
	.dc.b	$8F,$8D,$B3,$DA,$31,$3B,$23,$82,$46,$C8,$79,$1C,$B4,$C9,$E6,$72
	.dc.b	$D3,$27,$A1,$CB,$4C,$9E,$A7,$2D,$33,$3E,$A7,$C0,$43,$B1,$4E,$85
	.dc.b	$39,$17,$E6,$FF,$C2,$B8,$7F,$61,$6C,$7F,$87,$5A,$FF,$97,$DA,$FF
	.dc.b	$99,$75,$FF,$6C,$CF,$ED,$76,$FB,$D3,$FC,$21,$AE,$F7,$46,$21,$FA
	.dc.b	$E3,$E4,$3F,$CE,$1F,$2B,$FD,$CE,$31,$28,$C4,$EA,$3F,$C6,$53,$A9
	.dc.b	$F5,$70,$66,$7D,$0D,$7F,$CF,$99,$FF,$8D,$27,$71,$D4,$FE,$2B,$AD
	.dc.b	$7F,$AC,$47,$F5,$D2,$3F,$E4,$FF,$9D,$F5,$BF,$E0,$99,$46,$23,$FC
	.dc.b	$8D,$18,$CB,$F4,$8D,$8F,$D3,$5F,$FC,$01,$7C,$B8,$7E,$A5,$46,$5B
	.dc.b	$1C,$BE,$1F,$F6,$15,$E1,$6C,$F8,$1A,$FF,$99,$5A,$FF,$8E,$BC,$87
	.dc.b	$E3,$3F,$B7,$F1,$FF,$D3,$4D,$8C,$CF,$F8,$E9,$AF,$C8,$21,$F8,$7C
	.dc.b	$3D,$18,$FA,$F1,$62,$FB,$2F,$DC,$46,$1B,$19,$8F,$54,$7E,$EF,$89
	.dc.b	$C3,$F8,$8F,$98,$FC,$71,$32,$FE,$84,$F8,$9F,$BE,$7C,$CC,$BC,$67
	.dc.b	$CB,$AF,$CB,$C7,$F4,$9E,$3F,$CC,$11,$FF,$3E,$7C,$5B,$ED,$59,$BE
	.dc.b	$6B,$FC,$8D,$39,$8C,$57,$0A,$DD,$69,$B1,$F1,$4F,$21,$0F,$41,$FE
	.dc.b	$CF,$AE,$9E,$CF,$1B,$FF,$59,$4F,$FB,$3E,$72,$F7,$8C,$BF,$ED,$7F
	.dc.b	$2F,$F2,$31,$FC,$C5,$B1,$FE,$5A,$E5,$FE,$4D,$18,$8F,$21,$E2,$3F
	.dc.b	$1A,$79,$4F,$F4,$F3,$B7,$FD,$C0,$D7,$C1,$1D,$0F,$B7,$E1,$4A,$FC
	.dc.b	$A3,$67,$B7,$17,$E6,$C9,$B7,$E6,$E7,$53,$F9,$76,$BF,$EE,$FF,$81
	.dc.b	$7D,$FF,$FB,$D7,$87,$FC,$F1,$18,$94,$62,$75,$1F,$EB,$79,$D4,$F8
	.dc.b	$AF,$F0,$B3,$EF,$3F,$8C,$1F,$C6,$C8,$7F,$83,$CF,$DD,$FE,$1A,$F8
	.dc.b	$64,$3E,$AB,$31,$86,$C6,$4F,$F6,$B4,$F1,$70,$7E,$CF,$E4,$79,$FD
	.dc.b	$2F,$FB,$37,$B1,$D6,$9C,$D7,$FB,$35,$AF,$11,$5F,$ED,$5E,$47,$C7
	.dc.b	$3E,$47,$CF,$F4,$B9,$8C,$5A,$31,$3A,$8F,$68,$AE,$71,$98,$F8,$30
	.dc.b	$FC,$4F,$F4,$95,$E1,$6C,$F2,$9D,$74,$FE,$B2,$B4,$D9,$FF,$76,$7B
	.dc.b	$AE,$AF,$D8,$FE,$AB,$FE,$B5,$8C,$74,$A3,$13,$AA,$E2,$D7,$FA,$97
	.dc.b	$C3,$7F,$55,$23,$FE,$49,$7F,$D8,$1F,$8E,$BD,$5F,$E8,$B3,$FE,$FD
	.dc.b	$8C,$3F,$CC,$8C,$46,$9F,$8B,$1A,$EC,$E7,$3F,$90,$9C,$B5,$FC,$3C
	.dc.b	$F4,$E9,$DC,$7E,$60,$FC,$71,$0F,$C9,$94,$72,$FE,$13,$EC,$BF,$23
	.dc.b	$F2,$9F,$87,$E7,$46,$2B,$EC,$D6,$D5,$6E,$75,$1F,$9A,$DB,$F9,$BF
	.dc.b	$A9,$FC,$83,$EA,$7F,$A5,$36,$7F,$A5,$35,$E1,$E4,$23,$11,$FA,$4E
	.dc.b	$2F,$F7,$9F,$D2,$7D,$9F,$7F,$F9,$5B,$71,$34,$FE,$79,$5F,$E0,$9D
	.dc.b	$AB,$FE,$09,$F3,$3E,$33,$F1,$08,$64,$FD,$B9,$E4,$FA,$FF,$E7,$3F
	.dc.b	$8B,$FF,$63,$6C,$7F,$63,$6B,$BC,$13,$F9,$45,$F5,$14,$93,$F9,$BE
	.dc.b	$CF,$E8,$51,$F6,$BE,$A6,$4F,$E1,$83,$87,$CE,$3C,$DF,$EB,$B9,$7F
	.dc.b	$99,$DB,$1E,$A3,$E5,$6B,$FD,$4B,$F2,$6F,$90,$48,$FD,$3C,$65,$FF
	.dc.b	$52,$5F,$F8,$CC,$8E,$0F,$C5,$1E,$1F,$3F,$FC,$2C,$F9,$2F,$C2,$FA
	.dc.b	$7C,$EF,$CA,$91,$C1,$CA,$36,$F1,$9E,$6F,$F6,$8E,$3F,$E6,$FA,$BF
	.dc.b	$4F,$FF,$0A,$79,$9F,$DB,$1B,$19,$AF,$F0,$B4,$E5,$ED,$7D,$1F,$46
	.dc.b	$BF,$1E,$31,$8F,$8A,$7F,$04,$70,$9D,$11,$F9,$87,$23,$F5,$1C,$9F
	.dc.b	$CE,$BE,$8E,$C1,$A4,$7F,$57,$D6,$F3,$78,$46,$1A,$FF,$8B,$5A,$FE
	.dc.b	$E2,$38,$3D,$4F,$31,$E4,$C3,$D4,$F9,$4C,$90,$F9,$D8,$7A,$DC,$1F
	.dc.b	$00,$FD,$1E,$6F,$C3,$BF,$CD,$EF,$FB,$8E,$3F,$9D,$B9,$9F,$F0,$EF
	.dc.b	$B9,$FE,$56,$AD,$B6,$33,$5C,$25,$F5,$FF,$E3,$A6,$DF,$F9,$85,$B3
	.dc.b	$CE,$7F,$C7,$5A,$FD,$62,$5F,$A6,$7D,$67,$C1,$8F,$C4,$DF,$A7,$E9
	.dc.b	$FF,$C6,$3F,$15,$FE,$C3,$8C,$07,$18,$A7,$73,$E7,$6D,$F6,$3E,$39
	.dc.b	$F3,$4C,$1F,$22,$79,$4F,$F7,$B3,$91,$63,$F9,$34,$6B,$D0,$F0,$D7
	.dc.b	$F9,$19,$FF,$6B,$69,$F6,$89,$8F,$C0,$BF,$9D,$F8,$67,$F9,$33,$1F
	.dc.b	$4B,$5F,$9C,$D8,$EE,$D7,$FC,$A7,$8B,$EB,$BF,$CA,$B1,$F0,$9B,$5F
	.dc.b	$5D,$B1,$E8,$6B,$CB,$EE,$2C,$FE,$1D,$B4,$69,$C2,$7E,$41,$F8,$F2
	.dc.b	$18,$9F,$C8,$57,$22,$3F,$D3,$63,$C4,$D7,$8E,$3F,$1B,$5F,$C2,$46
	.dc.b	$32,$F9,$0F,$F1,$8F,$81,$7C,$13,$DB,$FF,$18,$FC,$03,$D9,$FF,$0B
	.dc.b	$D6,$A7,$31,$8A,$E1,$7F,$EC,$F8,$C4,$EA,$3F,$C7,$75,$C5,$D1,$8A
	.dc.b	$E3,$7D,$E3,$15,$C5,$57,$19,$D4,$7F,$18,$BF,$F7,$31,$1F,$F1,$69
	.dc.b	$92,$E5,$FD,$12,$35,$1F,$EE,$B6,$BF,$82,$D3,$FD,$2B,$1F,$81,$D3
	.dc.b	$E1,$1F,$81,$2C,$FD,$08,$FC,$AE,$7F,$EF,$27,$9C,$78,$4E,$C7,$EF
	.dc.b	$F5,$EB,$54,$E7,$1F,$89,$69,$F8,$92,$ED,$F9,$56,$E1,$CA,$7C,$51
	.dc.b	$E5,$BF,$D0,$BF,$EC,$18,$FC,$FB,$5F,$CF,$EF,$FE,$B2,$F7,$76,$7A
	.dc.b	$B8,$3E,$6C,$FF,$81,$6F,$D7,$A7,$C9,$8F,$DC,$F5,$7E,$DB,$1F,$EA
	.dc.b	$71,$FB,$AE,$3F,$F5,$EC,$E4,$E9,$7C,$B6,$F8,$D5,$F5,$1A,$FC,$62
	.dc.b	$1E,$13,$58,$F9,$30,$FE,$2F,$F1,$BE,$12,$3F,$C3,$1E,$F3,$C2,$FC
	.dc.b	$15,$EC,$8F,$A8,$75,$3B,$98,$AF,$E7,$DC,$DF,$82,$8F,$E6,$CD,$7F
	.dc.b	$8B,$B5,$FE,$12,$D8,$FE,$84,$3E,$97,$FA,$03,$67,$FA,$05,$7F,$B8
	.dc.b	$A7,$27,$9A,$E4,$FA,$48,$79,$B4,$FA,$4D,$CB,$E4,$BE,$B2,$1E,$E2
	.dc.b	$47,$07,$5E,$6F,$FA,$8B,$AF,$FD,$5F,$EB,$FF,$97,$DB,$3F,$DC,$4F
	.dc.b	$F4,$D9,$FD,$C4,$A7,$F7,$1E,$ED,$BF,$EB,$8A,$FF,$56,$C6,$23,$47
	.dc.b	$08,$CB,$C8,$F2,$E1,$EA,$6E,$7F,$5F,$1F,$44,$63,$2F,$8A,$FC,$19
	.dc.b	$D5,$7C,$2D,$1E,$07,$F8,$9B,$F4,$6F,$F1,$19,$2F,$FC,$47,$28,$E7
	.dc.b	$1F,$23,$1E,$FE,$87,$FA,$5B,$E3,$FF,$D1,$64,$9F,$F2,$93,$C1,$F2
	.dc.b	$1F,$96,$F9,$FF,$E2,$2F,$F8,$B1,$B5,$FE,$54,$6E,$4C,$D1,$FE,$79
	.dc.b	$FC,$86,$5F,$F9,$4B,$34,$62,$3E,$06,$BE,$1C,$FF,$D5,$77,$FE,$A0
	.dc.b	$24,$FF,$99,$DF,$F3,$87,$D0,$C8,$FF,$37,$7B,$4C,$53,$99,$FF,$3C
	.dc.b	$79,$BF,$E7,$D9,$FF,$69,$47,$F9,$F5,$AE,$8F,$CA,$9F,$27,$4F,$B5
	.dc.b	$AF,$D8,$3F,$CE,$7F,$21,$F0,$36,$34,$95,$EF,$FA,$6F,$F4,$D1,$1C
	.dc.b	$1F,$EA,$6C,$1F,$41,$FE,$AA,$6C,$66,$30,$7F,$AC,$D7,$C2,$52,$3F
	.dc.b	$E5,$0F,$60,$E8,$F5,$9F,$70,$FE,$0B,$C1,$FE,$D8,$C7,$FD,$9A,$48
	.dc.b	$FF,$4A,$3F,$ED,$0F,$9C,$91,$FE,$97,$6B,$FF,$00,$34,$FF,$4A,$FC
	.dc.b	$E7,$C3,$24,$7F,$1E,$7F,$DE,$39,$BF,$50,$A4,$9F,$F7,$9C,$65,$FB
	.dc.b	$6B,$E7,$3F,$D5,$C4,$8C,$BF,$D9,$3E,$11,$F0,$AE,$FF,$83,$39,$8C
	.dc.b	$57,$33,$E2,$8A,$E3,$5C,$23,$31,$87,$81,$FC,$0F,$D1,$7C,$3F,$CE
	.dc.b	$F9,$BF,$F7,$63,$67,$FD,$02,$D7,$83,$5F,$CB,$7C,$4F,$D5,$F1,$FF
	.dc.b	$BE,$E8,$8C,$4F,$F9,$C9,$AF,$FB,$C5,$AF,$E5,$79,$FF,$83,$93,$B7
	.dc.b	$F6,$FF,$09,$FE,$11,$69,$F8,$42,$CF,$F4,$98,$D3,$FE,$0E,$F1,$58
	.dc.b	$FE,$2B,$97,$F8,$0C,$62,$7F,$3E,$6E,$74,$D3,$27,$FC,$00,$F8,$AF
	.dc.b	$E9,$5E,$21,$A4,$D3,$C5,$97,$96,$BF,$C3,$F9,$79,$1F,$AE,$D7,$F3
	.dc.b	$AD,$9F,$E4,$B5,$CE,$BE,$5D,$F7,$9F,$ED,$CD,$8F,$E3,$7F,$18,$F8
	.dc.b	$CF,$EC,$18,$A3,$17,$FD,$B5,$B5,$C2,$B7,$8F,$8E,$CD,$24,$FA,$E3
	.dc.b	$F1,$F5,$C9,$FF,$1D,$B5,$CD,$78,$A3,$11,$FB,$AF,$89,$78,$7F,$C7
	.dc.b	$7E,$47,$F9,$9C,$63,$24,$62,$75,$5F,$E4,$8A,$D4,$FE,$03,$19,$3F
	.dc.b	$39,$F3,$3F,$31,$AA,$7B,$78,$FD,$70,$34,$93,$FE,$1C,$F9,$2F,$E4
	.dc.b	$84,$70,$77,$3E,$F9,$2F,$C6,$FF,$0B,$F1,$3F,$A5,$71,$7F,$2A,$8C
	.dc.b	$71,$FE,$31,$FE,$3B,$DD,$FC,$C9,$FF,$6B,$F0,$BD,$0F,$C7,$93,$4F
	.dc.b	$C0,$BF,$63,$2F,$6F,$17,$D6,$BF,$11,$FE,$F8,$F8,$FF,$87,$E6,$3A
	.dc.b	$D1,$86,$CF,$F9,$32,$75,$3F,$2B,$1F,$EE,$FE,$CF,$E8,$08,$FE,$49
	.dc.b	$F1,$57,$80,$FC,$EE,$3E,$43,$97,$37,$F6,$B8,$FF,$23,$36,$3F,$83
	.dc.b	$B6,$3F,$DE,$ED,$7F,$8E,$FC,$C3,$E5,$3F,$B2,$7C,$AF,$F7,$A8,$C6
	.dc.b	$28,$C4,$EA,$BF,$D1,$F5,$A9,$FE,$0E,$48,$FB,$4D,$7F,$95,$FC,$A3
	.dc.b	$DB,$FE,$A3,$3F,$E0,$88,$C4,$F7,$3F,$D3,$6E,$23,$53,$C2,$7F,$0E
	.dc.b	$7E,$03,$9B,$F1,$4D,$3F,$14,$FF,$1A,$23,$83,$FD,$4F,$F2,$BF,$A4
	.dc.b	$CB,$FC,$CC,$DD,$3F,$C3,$6D,$9D,$F5,$7F,$02,$8C,$47,$F4,$66,$BF
	.dc.b	$DB,$1A,$9A,$48,$E4,$4A,$F0,$B6,$BC,$2D,$AF,$0C,$E8,$93,$A6,$C6
	.dc.b	$A3,$04,$70,$43,$D1,$3D,$5B,$3C,$9B,$3C,$5A,$71,$3D,$D3,$C5,$B3
	.dc.b	$C5,$B3,$C5,$FF,$6E,$9F,$E4,$6F,$77,$FD,$C8,$FF,$B6,$1A,$FF,$BA
	.dc.b	$8F,$A3,$F8,$97,$F8,$DD,$B5,$FE,$4E,$7E,$64,$65,$FB,$6D,$7E,$A3
	.dc.b	$5F,$A8,$D7,$8C,$73,$C1,$39,$6C,$66,$30,$47,$04,$9F,$29,$0F,$1B
	.dc.b	$96,$99,$3C,$AE,$5A,$64,$F3,$B9,$69,$93,$D2,$59,$F3,$B6,$7C,$CD
	.dc.b	$9E,$4F,$03,$C6,$43,$FC,$3F,$F6,$CF,$AB,$FF,$22,$FC,$0F,$F9,$F2
	.dc.b	$31,$28,$C4,$EA,$3F,$C4,$E2,$77,$3F,$E8,$4F,$C3,$5E,$E6,$EB,$FC
	.dc.b	$5B,$18,$7A,$1C,$9C,$4F,$19,$E1,$3F,$CF,$3B,$FF,$D1,$53,$FE,$6D
	.dc.b	$24,$7F,$BA,$1F,$E2,$F8,$F4,$8F,$E3,$FC,$9F,$E8,$19,$C9,$23,$D3
	.dc.b	$F3,$5F,$E0,$38,$FF,$4D,$B6,$33,$E2,$F1,$6D,$F5,$5F,$6A,$FC,$03
	.dc.b	$D9,$1F,$12,$31,$9F,$E6,$5B,$1F,$EA,$86,$BE,$82,$1F,$CB,$B2,$FC
	.dc.b	$C7,$47,$F3,$48,$FF,$21,$FC,$23,$EB,$3F,$3D,$92,$30,$FF,$AD,$0C
	.dc.b	$46,$A3,$33,$FE,$4C,$F3,$1F,$01,$FA,$DB,$FF,$27,$36,$33,$3C,$DA
	.dc.b	$F9,$8F,$9B,$3F,$B5,$C3,$CD,$1E,$98,$F1,$5B,$E2,$7C,$44,$8C,$BF
	.dc.b	$B5,$F8,$AF,$E8,$75,$FE,$CC,$6B,$F8,$E6,$BF,$E7,$66,$BF,$E9,$CB
	.dc.b	$F0,$BE,$43,$CC,$67,$3F,$E0,$5F,$B9,$3E,$1B,$F1,$B2,$8C,$3F,$74
	.dc.b	$C4,$6A,$33,$C7,$FE,$1F,$F8,$E9,$5F,$E2,$BF,$3F,$FD,$9D,$18,$8F
	.dc.b	$11,$F0,$CF,$18,$C7,$CE,$FF,$8E,$E7,$FD,$AD,$F3,$DF,$E2,$B2,$47
	.dc.b	$5D,$51,$FE,$56,$F9,$4F,$E3,$9E,$91,$A4,$9F,$A9,$AF,$E2,$88,$E0
	.dc.b	$FE,$59,$3C,$5B,$3C,$6F,$B4,$7D,$AF,$07,$F1,$44,$70,$73,$8F,$6E
	.dc.b	$5E,$C7,$DD,$86,$DF,$CC,$D5,$7F,$D5,$F5,$F9,$98,$FE,$6D,$E7,$3A
	.dc.b	$FF,$C3,$D5,$FC,$3A,$31,$E7,$FF,$43,$6C,$7F,$BB,$CF,$0E,$1E,$E7
	.dc.b	$E7,$FE,$49,$FE,$BC,$E0,$FE,$2F,$08,$C3,$FC,$69,$AF,$FB,$09,$AB
	.dc.b	$CC,$FF,$20,$7D,$D7,$FB,$AB,$F6,$3F,$E4,$31,$89,$46,$3C,$F5,$6E
	.dc.b	$7F,$D9,$B5,$BA,$FD,$DF,$7B,$7E,$35,$F1,$3A,$BF,$86,$46,$23,$FD
	.dc.b	$F2,$FE,$FE,$78,$57,$28,$D4,$F3,$AF,$EB,$7F,$87,$7F,$08,$F1,$BB
	.dc.b	$36,$8D,$3F,$DC,$9E,$E2,$DF,$9E,$E1,$A7,$FB,$8B,$DE,$FF,$74,$C6
	.dc.b	$4F,$9B,$F6,$CF,$BF,$E5,$F9,$EE,$1B,$7F,$D2,$27,$4A,$74,$D3,$F0
	.dc.b	$1C,$EE,$7C,$47,$FA,$02,$7E,$DF,$3F,$E1,$1B,$5B,$7E,$FF,$E7,$0F
	.dc.b	$C2,$3E,$74,$61,$B7,$FD,$6A,$75,$BF,$68,$F8,$51,$8A,$FC,$4F,$D2
	.dc.b	$FF,$A7,$A7,$F8,$33,$7A,$73,$EF,$7F,$0C,$8C,$4F,$5E,$CF,$E9,$4F
	.dc.b	$CD,$18,$7F,$94,$EF,$CF,$3A,$9F,$0B,$71,$FE,$7D,$C5,$F8,$7F,$C4
	.dc.b	$3F,$A1,$B6,$BF,$10,$D7,$F0,$FC,$C8,$C4,$7E,$6F,$AD,$F1,$9F,$7F
	.dc.b	$EA,$FF,$30,$D9,$FC,$BC,$E9,$4E,$9A,$7D,$6E,$3F,$DA,$71,$7E,$E2
	.dc.b	$3F,$D6,$A7,$ED,$4F,$EE,$C4,$3F,$D7,$3A,$FF,$7A,$7F,$60,$7E,$47
	.dc.b	$C3,$FD,$BF,$7F,$C7,$6C,$64,$FD,$C1,$4F,$7F,$5F,$EA,$FE,$B3,$F3
	.dc.b	$C4,$8F,$D6,$46,$5F,$D7,$7D,$8F,$EC,$7F,$5E,$FE,$DA,$D8,$CF,$12
	.dc.b	$72,$D8,$FB,$5F,$14,$FB,$4F,$EB,$D3,$FE,$05,$8C,$79,$DF,$6D,$B1
	.dc.b	$99,$DE,$BF,$80,$8F,$E7,$B7,$F8,$76,$BF,$87,$E7,$FF,$4C,$6B,$B3
	.dc.b	$E6,$47,$CE,$3E,$91,$83,$FC,$51,$C1,$C9,$B1,$FA,$68,$C4,$3B,$3F
	.dc.b	$96,$CF,$F2,$6F,$CF,$1F,$47,$FC,$57,$DB,$C3,$6F,$FC,$83,$3A,$CB
	.dc.b	$FD,$41,$18,$9D,$3F,$C2,$F5,$F8,$38,$7F,$92,$1B,$1F,$2B,$5F,$93
	.dc.b	$39,$6C,$7F,$93,$BF,$22,$76,$7F,$15,$FC,$EF,$F9,$DC,$62,$51,$89
	.dc.b	$D4,$71,$6B,$E3,$AF,$AF,$87,$7E,$1F,$E6,$4D,$CC,$D3,$E0,$AF,$CD
	.dc.b	$69,$2F,$E0,$46,$5F,$A6,$43,$EB,$10,$FE,$6D,$E6,$7F,$8D,$49,$1F
	.dc.b	$E2,$87,$FC,$71,$F8,$1F,$F4,$5F,$83,$FF,$20,$B6,$BF,$A3,$35,$FD
	.dc.b	$FB,$5F,$E8,$5E,$1D,$BF,$F2,$33,$7C,$13,$4E,$13,$F8,$3E,$DF,$DD
	.dc.b	$3F,$8A,$3F,$18,$67,$9B,$FC,$9D,$F8,$67,$F5,$9C,$29,$FE,$54,$71
	.dc.b	$3F,$E8,$68,$46,$23,$E2,$1D,$4A,$75,$3F,$DE,$FF,$70,$FA,$2F,$F5
	.dc.b	$E7,$42,$31,$7F,$EF,$86,$D7,$0A,$FF,$61,$7A,$0F,$0D,$F4,$5F,$D7
	.dc.b	$7F,$5B,$FE,$EC,$DB,$FE,$3E,$D7,$FB,$15,$71,$8C,$47,$F7,$7C,$DC
	.dc.b	$3E,$F2,$67,$F1,$1F,$F0,$DE,$DF,$2A,$BD,$87,$BB,$B7,$FC,$32,$FF
	.dc.b	$88,$AF,$FC,$41,$7F,$32,$BE,$51,$1F,$E5,$BB,$CA,$FA,$77,$C9,$FC
	.dc.b	$61,$D8,$AE,$5F,$CD,$FB,$0F,$95,$FF,$77,$5F,$F1,$A8,$C3,$FE,$0E
	.dc.b	$9F,$DC,$46,$1F,$F7,$87,$3A,$31,$3E,$BC,$7D,$D8,$77,$BF,$11,$F7
	.dc.b	$4F,$1C,$68,$FC,$6C,$BE,$45,$F8,$2F,$F8,$8F,$ED,$3F,$DF,$D3,$98
	.dc.b	$C3,$FE,$48,$9D,$A9,$DC,$61,$E3,$CE,$FA,$7F,$7E,$3F,$D9,$1A,$FF
	.dc.b	$89,$DB,$7F,$EA,$B6,$EF,$FD,$82,$3E,$24,$FE,$07,$CB,$C9,$23,$31
	.dc.b	$F0,$09,$A7,$C3,$3E,$94,$75,$39,$96,$3C,$F7,$E2,$C3,$D1,$3D,$2B
	.dc.b	$96,$9F,$27,$83,$F0,$58,$7D,$16,$BF,$82,$F8,$B7,$C8,$FF,$BB,$B8
	.dc.b	$DF,$1B,$FE,$11,$9C,$F9,$3F,$E3,$F9,$CD,$FF,$90,$67,$2F,$E7,$AB
	.dc.b	$85,$F3,$6D,$F1,$6C,$EC,$E5,$3B,$6C,$ED,$FE,$AB,$3F,$D4,$C8,$FF
	.dc.b	$82,$DA,$FF,$89,$F2,$FF,$40,$C6,$23,$89,$FE,$67,$D7,$FC,$D5,$1F
	.dc.b	$2B,$4F,$79,$92,$CF,$E0,$5A,$F2,$3B,$69,$FE,$4F,$9E,$D1,$FC,$7D
	.dc.b	$FE,$35,$BF,$E0,$DF,$CA,$B6,$BF,$75,$BF,$89,$A7,$63,$F0,$13,$E5
	.dc.b	$23,$E5,$7F,$33,$C5,$FE,$83,$FC,$1A,$C9,$A4,$7E,$36,$10,$F4,$F0
	.dc.b	$7D,$A6,$DF,$AB,$4F,$93,$1B,$3E,$01,$FC,$13,$9B,$D0,$FF,$00,$6B
	.dc.b	$9B,$F8,$FC,$DF,$C0,$67,$33,$F1,$4E,$85,$3C,$26,$2B,$C2,$E0,$87
	.dc.b	$F2,$BD,$7E,$CB,$FC,$8D,$AF,$FA,$83,$5F,$E4,$B8,$FD,$29,$CC,$FA
	.dc.b	$8F,$01,$6B,$BB,$FC,$DF,$8B,$79,$7F,$52,$6E,$1F,$CE,$A3,$E4,$4E
	.dc.b	$DB,$D1,$FD,$06,$1F,$A0,$7F,$48,$69,$FD,$20,$A7,$F4,$8E,$BF,$E9
	.dc.b	$EE,$07,$77,$F5,$9E,$DF,$4B,$70,$F0,$B7,$5F,$EB,$51,$E4,$AF,$69
	.dc.b	$34,$F6,$BF,$02,$3F,$B6,$9E,$C2,$1E,$C3,$27,$38,$E2,$4D,$7F,$41
	.dc.b	$3F,$85,$9F,$F0,$38,$FC,$57,$FB,$D7,$4F,$F8,$22,$31,$3F,$ED,$7F
	.dc.b	$03,$FD,$B5,$5A,$6C,$7F,$85,$A7,$6F,$F8,$59,$AF,$FB,$5F,$C1,$DB
	.dc.b	$9A,$EF,$E1,$7D,$8C,$0D,$25,$FA,$F1,$F4,$DF,$E3,$0F,$8C,$E1,$FF
	.dc.b	$6E,$79,$7F,$E5,$7F,$D6,$FF,$C7,$64,$8F,$F1,$CB,$FE,$41,$F2,$24
	.dc.b	$7F,$95,$7C,$7F,$F7,$B4,$62,$3F,$D7,$5E,$57,$F9,$47,$F6,$6F,$F2
	.dc.b	$84,$1A,$47,$96,$4F,$94,$FF,$27,$12,$32,$FF,$93,$BF,$07,$91,$FE
	.dc.b	$4C,$6B,$FB,$E8,$FF,$3D,$11,$FF,$59,$3F,$BF,$D3,$E0,$47,$C2,$ED
	.dc.b	$FC,$57,$A1,$FE,$6A,$C3,$C5,$3F,$E8,$FB,$FF,$49,$11,$F3,$FC,$AF
	.dc.b	$F9,$8C,$8E,$0F,$F5,$2F,$ED,$9F,$EA,$42,$38,$3F,$D5,$F4,$FC,$0B
	.dc.b	$F1,$3E,$B2,$31,$3A,$9C,$C6,$23,$FD,$86,$79,$8B,$87,$FB,$09,$C0
	.dc.b	$1E,$C7,$FB,$B6,$3F,$D3,$D8,$C9,$FF,$6B,$35,$FC,$37,$4F,$E8,$74
	.dc.b	$FC,$51,$FC,$03,$B7,$FD,$C1,$F3,$3F,$ED,$F2,$57,$FB,$93,$DA,$F0
	.dc.b	$47,$FB,$CB,$37,$E2,$94,$95,$FE,$F4,$8C,$BF,$EB,$DC,$BF,$D8,$04
	.dc.b	$70,$7A,$CF,$67,$BD,$FE,$76,$51,$97,$3D,$DF,$06,$31,$09,$E1,$18
	.dc.b	$6C,$71,$E2,$7C,$07,$D3,$FD,$D3,$8C,$63,$14,$62,$75,$5C,$5A,$FF
	.dc.b	$50,$8F,$F6,$61,$1F,$36,$EE,$57,$07,$49,$F5,$7C,$13,$E0,$FE,$D6
	.dc.b	$1E,$A8,$C4,$EC,$F8,$8E,$23,$53,$C2,$7F,$DE,$7F,$0A,$F8,$2F,$C2
	.dc.b	$7E,$91,$FE,$CF,$8C,$4A,$31,$3A,$8E,$2D,$7F,$19,$AB,$C3,$75,$7E
	.dc.b	$67,$F0,$7F,$CB,$92,$BF,$3E,$FF,$26,$F8,$AF,$CF,$7E,$FF,$70,$34
	.dc.b	$95,$F7,$FC,$DF,$F4,$B7,$C6,$7E,$6E,$7E,$F4,$62,$78,$9B,$EC,$56
	.dc.b	$E3,$0D,$7F,$53,$D1,$FA,$72,$3F,$89,$EE,$FC,$74,$7E,$23,$E3,$7F
	.dc.b	$96,$E5,$FD,$3D,$BE,$66,$7C,$AF,$ED,$F8,$27,$2F,$F7,$DE,$8F,$E1
	.dc.b	$FD,$3F,$AB,$BF,$E8,$FE,$3F,$F3,$92,$63,$FD,$4F,$8F,$F8,$94,$F5
	.dc.b	$8F,$06,$5D,$B2,$EB,$97,$E8,$DA,$FF,$09,$BE,$2F,$E9,$EB,$EB,$61
	.dc.b	$DB,$0F,$0C,$7F,$69,$F8,$F7,$C7,$FA,$9B,$19,$8C,$1E,$1C,$A4,$79
	.dc.b	$09,$5F,$AE,$AE,$2D,$AE,$33,$CC,$FE,$F1,$F4,$6F,$91,$FE,$41,$E3
	.dc.b	$7E,$DA,$31,$3F,$C6,$B3,$46,$27,$55,$C2,$7F,$80,$6B,$F0,$9F,$A5
	.dc.b	$9F,$9B,$2F,$94,$70,$C5,$F6,$34,$95,$FC,$27,$2F,$0C,$7F,$08,$F9
	.dc.b	$37,$CA,$7F,$0B,$AF,$E1,$D1,$87,$F8,$89,$88,$D3,$FC,$5E,$35,$C7
	.dc.b	$E2,$8F,$3E,$5E,$56,$E3,$FE,$1B,$F9,$5F,$F7,$4C,$5F,$67,$F9,$D6
	.dc.b	$9F,$C6,$A7,$2D,$8C,$CF,$DF,$E3,$7A,$3F,$E0,$52,$61,$F8,$D7,$C1
	.dc.b	$3F,$89,$C6,$BF,$D1,$F7,$FC,$5D,$3E,$59,$F2,$FD,$AB,$F8,$0C,$61
	.dc.b	$FF,$15,$37,$3F,$E6,$55,$C6,$B8,$57,$E5,$7E,$63,$FC,$30,$49,$FD
	.dc.b	$03,$E2,$8F,$B3,$1F,$8C,$7E,$2D,$7F,$3D,$F9,$97,$80,$F9,$AE,$B1
	.dc.b	$86,$C6,$6B,$FC,$B3,$39,$7F,$C9,$1F,$38,$F9,$CF,$EE,$1C,$7F,$D1
	.dc.b	$A3,$19,$7D,$CA,$FE,$A1,$18,$9D,$57,$09,$FB,$31,$FB,$72,$38,$3F
	.dc.b	$B6,$90,$FF,$49,$BF,$DB,$DF,$EE,$DF,$3B,$E7,$8F,$EA,$ED,$8F,$EB
	.dc.b	$8D,$7F,$B1,$35,$FE,$FD,$F3,$CF,$9F,$FC,$8C,$61,$AA,$34,$FB,$4D
	.dc.b	$14,$FF,$39,$F9,$1F,$C7,$3E,$03,$F9,$EE,$7F,$E6,$F6,$EF,$FD,$B1
	.dc.b	$B5,$FD,$C5,$AF,$F7,$76,$BF,$E1,$36,$BD,$A7,$A5,$77,$23,$82,$57
	.dc.b	$78,$CB,$63,$33,$E1,$9C,$B6,$32,$F8,$E3,$2D,$8C,$BE,$58,$CB,$63
	.dc.b	$2F,$9E,$7C,$04,$7B,$35,$E8,$D7,$93,$5E,$47,$03,$FC,$31,$F1,$8F
	.dc.b	$8C,$7C,$67,$F8,$62,$53,$98,$C3,$63,$FC,$3A,$D7,$E9,$7E,$17,$FE
	.dc.b	$4B,$7E,$B3,$5C,$E1,$D7,$24,$E6,$7E,$C3,$5F,$B0,$D7,$FD,$6E,$D8
	.dc.b	$E3,$3C,$E3,$04,$70,$49,$F1,$B5,$F1,$98,$21,$82,$1E,$47,$83,$4D
	.dc.b	$94,$D9,$E8,$76,$D3,$65,$36,$7B,$1D,$B4,$D9,$4D,$CF,$C0,$72,$D7
	.dc.b	$33,$F0,$5F,$F2,$8E,$7F,$B0,$C2,$E9,$FE,$75,$6C,$FE,$00,$FF,$2D
	.dc.b	$4E,$DF,$F2,$E7,$D1,$B4,$E1,$18,$6B,$FE,$65,$8D,$3C,$CC,$46,$9F
	.dc.b	$F4,$27,$8F,$FE,$51,$F5,$75,$18,$F2,$7F,$CA,$78,$FF,$9A,$2B,$4D
	.dc.b	$8C,$CF,$F5,$88,$FF,$0B,$11,$C1,$FE,$73,$FC,$1B,$FA,$DB,$FE,$72
	.dc.b	$6B,$9F,$03,$F3,$CA,$32,$FF,$9F,$EF,$CC,$F8,$70,$FD,$24,$62,$3F
	.dc.b	$CD,$78,$7F,$9A,$9F,$F3,$FF,$C8,$7F,$82,$9B,$E3,$FE,$66,$31,$1F
	.dc.b	$E8,$76,$BF,$E8,$DC,$3F,$CE,$8F,$EC,$5A,$FA,$6B,$CF,$18,$8F,$0E
	.dc.b	$1E,$49,$F1,$46,$0F,$41,$E4,$F8,$FF,$11,$38,$BF,$7D,$82,$3F,$5D
	.dc.b	$E2,$3E,$33,$E2,$B6,$FF,$82,$35,$FF,$02,$47,$CA,$D3,$E2,$47,$C9
	.dc.b	$C7,$C6,$DC,$BC,$C7,$2D,$7F,$09,$0C,$3F,$C2,$3B,$7F,$83,$63,$F0
	.dc.b	$A6,$67,$7D,$7F,$D9,$BE,$4B,$FB,$1E,$C6,$92,$BE,$11,$0F,$11,$F1
	.dc.b	$0F,$F5,$C4,$FF,$70,$F9,$9E,$64,$8C,$F6,$FF,$B9,$63,$FD,$77,$F9
	.dc.b	$F3,$F0,$4F,$E2,$8D,$C7,$F9,$43,$6F,$73,$FC,$67,$7F,$8B,$1F,$9F
	.dc.b	$DB,$E1,$BE,$33,$FD,$9D,$5C,$5F,$F6,$63,$5C,$D7,$80,$FE,$49,$EA
	.dc.b	$FE,$D7,$0F,$E7,$0D,$9F,$F6,$8B,$5F,$F4,$CF,$37,$8F,$87,$EA,$C7
	.dc.b	$8C,$F1,$14,$DD,$74,$9E,$E4,$7F,$A3,$73,$FF,$00,$9F,$F6,$FB,$5F
	.dc.b	$E0,$C6,$A7,$F0,$D5,$E0,$86,$7F,$11,$B1,$98,$C7,$E5,$CF,$C3,$3F
	.dc.b	$A1,$37,$1F,$EA,$CD,$BF,$F7,$7B,$5F,$F4,$97,$C9,$3F,$15,$53,$EF
	.dc.b	$BF,$C8,$F3,$98,$C5,$70,$AD,$D6,$9B,$1F,$EC,$5F,$C8,$5F,$2C,$87
	.dc.b	$D4,$FF,$B1,$A7,$31,$83,$84,$EE,$74,$D8,$E9,$DB,$CF,$EA,$79,$36
	.dc.b	$32,$23,$2D,$3C,$9E,$BE,$DF,$27,$1B,$1F,$F4,$DC,$6C,$CB,$8A,$F4
	.dc.b	$57,$19,$F6,$7D,$AB,$0F,$77,$E6,$5E,$F9,$93,$95,$19,$78,$CF,$07
	.dc.b	$13,$B9,$FA,$FC,$3F,$51,$BB,$F2,$AE,$5C,$3F,$7E,$72,$75,$87,$C1
	.dc.b	$7F,$83,$FE,$AB,$0D,$8F,$13,$5F,$43,$5F,$B1,$F1,$8F,$80,$FE,$86
	.dc.b	$D8,$FF,$11,$35,$F0,$35,$F2,$E5,$F9,$42,$65,$F9,$47,$FC,$35,$C2
	.dc.b	$FD,$6F,$FC,$2F,$F9,$9F,$C5,$8C,$79,$9F,$94,$8C,$4E,$B0,$F9,$D5
	.dc.b	$AC,$15,$F9,$8C,$3F,$37,$F5,$96,$4D,$25,$FD,$3F,$C3,$DF,$49,$FA
	.dc.b	$2F,$C2,$7E,$54,$62,$D1,$89,$D5,$71,$6B,$FC,$A8,$FE,$58,$7F,$07
	.dc.b	$30,$7E,$17,$E6,$DF,$49,$FC,$F9,$B1,$FC,$F1,$BE,$5F,$DB,$6C,$7E
	.dc.b	$E3,$D2,$FD,$BF,$E1,$DF,$B6,$6C,$67,$24,$E5,$B1,$FA,$21,$F5,$FF
	.dc.b	$9C,$69,$97,$07,$DC,$E4,$FC,$09,$23,$F0,$2F,$F8,$1B,$A7,$F3,$D9
	.dc.b	$7E,$45,$B3,$F9,$CE,$AF,$EB,$71,$88,$FD,$36,$FF,$37,$CD,$F9,$99
	.dc.b	$7F,$57,$9F,$98,$D7,$7B,$FF,$16,$9F,$E6,$71,$FC,$93,$CA,$FE,$8E
	.dc.b	$FD,$01,$F4,$0F,$C5,$FF,$8E,$E9,$FA,$58,$C6,$3F,$DA,$9B,$19,$9D
	.dc.b	$D7,$E3,$BA,$DF,$45,$F7,$A3,$F8,$EC,$63,$0F,$C2,$C6,$E7,$5D,$3F
	.dc.b	$B2,$78,$43,$6F,$E0,$9E,$77,$C7,$E0,$FE,$4C,$D9,$34,$FD,$6F,$37
	.dc.b	$F2,$BE,$7B,$3E,$03,$F8,$87,$6F,$E2,$A3,$27,$AF,$37,$B1,$49,$5F
	.dc.b	$C4,$F8,$3E,$1C,$7C,$CF,$2B,$FA,$81,$3C,$BF,$EC,$76,$F0,$7F,$C6
	.dc.b	$B8,$7F,$48,$6C,$67,$D7,$FF,$4B,$34,$FE,$DF,$E3,$BE,$BB,$FA,$27
	.dc.b	$A9,$FD,$0D,$B8,$7F,$49,$6B,$FC,$C9,$AF,$F5,$6F,$3B,$FC,$65,$7F
	.dc.b	$E5,$80,$E2,$BF,$AC,$FA,$DF,$D7,$1B,$3F,$E5,$AD,$3D,$9D,$5F,$88
	.dc.b	$C3,$F6,$A7,$AF,$F0,$6F,$4E,$A6,$92,$F8,$4F,$EF,$48,$7E,$4C,$FF
	.dc.b	$78,$7B,$EF,$9C,$FF,$5C,$70,$A3,$15,$FE,$F7,$6D,$70,$AF,$93,$AB
	.dc.b	$EF,$BF,$BA,$FE,$21,$FD,$D1,$B5,$FD,$E9,$AF,$F6,$FC,$3E,$1C,$62
	.dc.b	$3E,$47,$C3,$F4,$24,$FF,$47,$3E,$4B,$FE,$21,$D7,$DB,$3E,$B3,$F8
	.dc.b	$DC,$FE,$07,$CA,$F8,$A4,$79,$6D,$FE,$0D,$9F,$17,$C6,$3F,$06,$FF
	.dc.b	$06,$FE,$13,$FE,$DF,$B4,$E5,$46,$63,$0D,$9E,$33,$FB,$DE,$2F,$B3
	.dc.b	$7F,$E3,$4F,$05,$FA,$03,$EF,$3F,$DD,$D7,$FD,$1A,$30,$FC,$5A,$FF
	.dc.b	$0A,$57,$0A,$DC,$62,$B6,$F8,$69,$F2,$14,$8F,$F7,$F3,$D7,$8F,$F8
	.dc.b	$AA,$3F,$DF,$18,$F9,$AB,$D1,$D9,$FA,$AD,$3E,$06,$1D,$48,$73,$72
	.dc.b	$FF,$8D,$BE,$F7,$CB,$5F,$A1,$6C,$7E,$81,$BC,$86,$92,$3E,$04,$F1
	.dc.b	$69,$C4,$CB,$F7,$CF,$D1,$94,$FC,$F3,$B7,$ED,$4F,$78,$E6,$D8,$CC
	.dc.b	$FA,$1E,$44,$C3,$E2,$3D,$9A,$FE,$AB,$E1,$DF,$11,$FE,$F1,$EB,$FF
	.dc.b	$35,$46,$3C,$AF,$E1,$3E,$FF,$F8,$BE,$31,$6A,$DD,$F1,$AD,$36,$7F
	.dc.b	$CA,$95,$C2,$B7,$7D,$27,$35,$B7,$FA,$75,$FF,$B7,$C8,$FF,$B3,$9A
	.dc.b	$FF,$AE,$DA,$FF,$A6,$DF,$F3,$5E,$3F,$E6,$E7,$E2,$E7,$FE,$91,$3E
	.dc.b	$93,$83,$F1,$3A,$B9,$BF,$D1,$17,$E2,$3C,$04,$3F,$80,$6B,$F4,$EB
	.dc.b	$95,$FE,$24,$ED,$19,$71,$1A,$9E,$38,$79,$67,$27,$E4,$F8,$3F,$D0
	.dc.b	$BF,$79,$FA,$32,$46,$43,$E8,$C7,$CF,$BB,$3F,$56,$5F,$13,$2F,$51
	.dc.b	$FC,$1B,$1F,$F4,$8F,$9D,$FA,$F2,$65,$D5,$FF,$4A,$CF,$ED,$79,$3C
	.dc.b	$CD,$9F,$F5,$D7,$2F,$EE,$DB,$CB,$FC,$03,$1F,$21,$E1,$CF,$C9,$A7
	.dc.b	$9B,$4F,$2E,$3E,$BE,$CF,$E0,$04,$C7,$07,$EF,$FA,$3F,$D5,$D8,$FF
	.dc.b	$9E,$E3,$F9,$21,$FE,$BD,$2B,$98,$DD,$7D,$73,$F0,$85,$3E,$19,$FC
	.dc.b	$DA,$79,$70,$BA,$BF,$11,$3F,$EB,$86,$BF,$80,$6B,$FD,$97,$AF,$FD
	.dc.b	$77,$F7,$DF,$D1,$EC,$D2,$4F,$FA,$5F,$BF,$FA,$A6,$7C,$9B,$5F,$ED
	.dc.b	$76,$E7,$FD,$56,$3E,$46,$B2,$3E,$66,$BF,$2B,$0F,$EA,$5A,$FF,$69
	.dc.b	$BF,$88,$DA,$F8,$9E,$07,$A2,$7F,$C1,$8F,$AA,$BE,$19,$37,$FA,$F5
	.dc.b	$FA,$CA,$FE,$FF,$E0,$FF,$B6,$3C,$CF,$EF,$A4,$BF,$F6,$EC,$65,$FF
	.dc.b	$06,$1F,$E3,$A2,$C7,$F8,$45,$C1,$FE,$0F,$29,$FE,$1D,$F1,$7C,$D9
	.dc.b	$1A,$4B,$F8,$53,$F6,$F1,$F3,$BF,$77,$C6,$FF,$27,$FD,$5F,$F8,$CC
	.dc.b	$91,$FE,$31,$F3,$FF,$DE,$31,$FE,$3C,$FB,$E9,$3F,$EC,$7F,$2F,$FD
	.dc.b	$0B,$18,$8F,$F6,$6F,$9D,$FE,$ED,$F9,$9F,$F2,$41,$23,$FC,$E0,$1F
	.dc.b	$F5,$67,$A3,$FE,$57,$AB,$BF,$FC,$00,$6E,$A6,$9F,$E4,$9F,$2E,$61
	.dc.b	$E7,$9F,$F0,$D3,$63,$F0,$3C,$BF,$0A,$3F,$91,$1F,$E7,$AF,$99,$FF
	.dc.b	$3E,$91,$C1,$E6,$9F,$F3,$01,$1C,$1F,$07,$D5,$3F,$C8,$B5,$FF,$69
	.dc.b	$72,$FF,$A5,$63,$13,$A8,$FF,$4E,$7D,$5C,$9F,$F5,$03,$5F,$E4,$FE
	.dc.b	$CB,$37,$DF,$FF,$A9,$C9,$3F,$EA,$B7,$FD,$65,$EA,$78,$B3,$B3,$F0
	.dc.b	$8F,$F5,$E4,$65,$FF,$7F,$7B,$7F,$EB,$C9,$F8,$B9,$49,$FF,$4E,$EF
	.dc.b	$FE,$A0,$23,$83,$FC,$19,$E1,$3D,$1F,$37,$D5,$7F,$B6,$C9,$3A,$8C
	.dc.b	$BF,$EE,$4F,$7D,$F0,$1E,$DF,$98,$FF,$74,$B6,$74,$D7,$FC,$47,$3C
	.dc.b	$37,$57,$D0,$99,$F0,$7F,$C6,$9B,$87,$ED,$E3,$11,$F5,$9A,$B9,$3E
	.dc.b	$BB,$F6,$CF,$AE,$4D,$7B,$3F,$7B,$A3,$F9,$FF,$D3,$E2,$11,$87,$6D
	.dc.b	$7C,$CD,$7C,$58,$7D,$DE,$1E,$58,$FC,$BC,$D1,$89,$D5,$78,$AB,$53
	.dc.b	$FE,$B3,$24,$7E,$B1,$AF,$E5,$3E,$1D,$ED,$BE,$1B,$D5,$3F,$EE,$FC
	.dc.b	$BF,$DD,$C4,$7E,$65,$70,$EE,$DC,$73,$B4,$62,$13,$B7,$E1,$9C,$A7
	.dc.b	$6E,$0E,$5E,$77,$EA,$3E,$8F,$E6,$D7,$9A,$31,$3C,$5A,$F0,$13,$FD
	.dc.b	$77,$95,$F1,$5F,$A7,$FB,$1C,$36,$BF,$2E,$D7,$FC,$26,$D7,$E9,$F0
	.dc.b	$7F,$86,$B7,$FC,$29,$38,$3F,$A0,$78,$DF,$CA,$2B,$F8,$49,$0F,$EC
	.dc.b	$E7,$E4,$FE,$35,$C9,$FB,$F8,$FC,$4C,$61,$FF,$13,$B6,$3F,$14,$D8
	.dc.b	$F8,$71,$FD,$3F,$6E,$78,$FF,$8B,$9B,$1F,$91,$6C,$FF,$3D,$F3,$3F
	.dc.b	$AF,$3E,$42,$1F,$5A,$7A,$12,$BA,$3F,$D7,$7E,$3D,$F1,$F4,$93,$FE
	.dc.b	$3E,$8C,$BF,$CA,$EB,$AE,$5F,$15,$EF,$3A,$BF,$D3,$36,$3F,$56,$D7
	.dc.b	$F7,$4D,$7F,$95,$FC,$8B,$E4,$7F,$91,$79,$3F,$AF,$8C,$63,$FD,$26
	.dc.b	$BF,$6B,$18,$9D,$57,$09,$F8,$9E,$47,$9D,$FE,$F3,$5C,$BC,$8D,$E0
	.dc.b	$69,$27,$F8,$0F,$C9,$7F,$9E,$5F,$E5,$87,$F4,$07,$1E,$27,$F0,$36
	.dc.b	$C7,$0F,$7F,$F9,$83,$F8,$73,$FC,$11,$C1,$F5,$9F,$F0,$5C,$BD,$1F
	.dc.b	$C4,$47,$F1,$D8,$C3,$FE,$6E,$6E,$FD,$E7,$73,$FD,$4B,$B1,$F2,$7F
	.dc.b	$E1,$76,$E9,$FE,$76,$6D,$FF,$19,$6D,$7C,$68,$FE,$84,$FF,$2D,$F9
	.dc.b	$4F,$F3,$33,$FE,$7E,$6C,$7F,$21,$6C,$7F,$26,$6B,$FC,$D6,$7F,$C2
	.dc.b	$95,$E1,$F4,$BC,$84,$7E,$A7,$27,$E2,$6B,$E2,$F0,$F7,$DB,$DB,$0C
	.dc.b	$7E,$96,$5E,$19,$CC,$62,$7F,$9B,$FC,$E3,$E7,$7F,$B5,$70,$FF,$37
	.dc.b	$8C,$67,$F6,$A9,$18,$9D,$57,$09,$FE,$83,$9F,$F9,$D2,$7F,$A0,$36
	.dc.b	$33,$C3,$FE,$A3,$9C,$B6,$3F,$80,$E7,$F8,$27,$FC,$0F,$D3,$FC,$0E
	.dc.b	$BF,$80,$CE,$65,$39,$7F,$A8,$FC,$BB,$CF,$D3,$63,$D4,$DB,$FE,$A8
	.dc.b	$D8,$FE,$B3,$EC,$7F,$69,$C3,$FA,$C3,$63,$3E,$67,$F9,$76,$31,$1F
	.dc.b	$E7,$18,$C1,$36,$FF,$5F,$CE,$A7,$FC,$42,$D4,$D2,$4F,$1A,$F0,$B6
	.dc.b	$7C,$35,$D2,$31,$1E,$18,$C1,$27,$8B,$63,$8C,$60,$8E,$08,$73,$72
	.dc.b	$D3,$27,$AA,$7C,$2D,$9E,$CE,$C8,$66,$79,$B4,$E6,$7C,$29,$E6,$BE
	.dc.b	$15,$F0,$8F,$85,$FF,$05,$4E,$63,$0F,$F8,$36,$77,$3A,$6C,$7D,$AF
	.dc.b	$A8,$FF,$68,$BF,$79,$AF,$FA,$A5,$AF,$D6,$6B,$F5,$9B,$91,$A4,$CB
	.dc.b	$CB,$19,$6C,$66,$7C,$51,$82,$38,$21,$E5,$72,$D3,$27,$9D,$CB,$4C
	.dc.b	$9E,$97,$2D,$32,$7A,$DC,$B4,$CC,$FA,$5B,$3E,$97,$C4,$43,$C0,$53
	.dc.b	$B1,$4E,$85,$8E,$8D,$8E,$8F,$F9,$57,$77,$E1,$FF,$E8,$6C,$7F,$C9
	.dc.b	$11,$89,$FB,$E6,$E7,$4F,$F9,$42,$35,$3F,$C1,$A3,$67,$F0,$AD,$36
	.dc.b	$FF,$84,$5A,$E4,$D8,$E1,$53,$9B,$31,$86,$C6,$6F,$F8,$DC,$78,$33
	.dc.b	$7A,$4F,$1F,$F3,$31,$86,$C6,$71,$FF,$33,$4E,$5F,$CF,$F2,$7F,$88
	.dc.b	$DF,$61,$0F,$E4,$E7,$F9,$B7,$E7,$1F,$98,$BE,$BB,$F3,$71,$86,$C6
	.dc.b	$6F,$83,$6B,$FC,$F3,$1C,$72,$F3,$D7,$C1,$9F,$2B,$5D,$B5,$F1,$9E
	.dc.b	$53,$E7,$1E,$AD,$BE,$0F,$DB,$FE,$F2,$FD,$ED,$9F,$0B,$5F,$F4,$C4
	.dc.b	$23,$11,$FE,$9B,$FB,$AF,$F4,$F7,$C9,$7F,$A3,$5B,$1F,$E9,$26,$BB
	.dc.b	$E7,$FF,$63,$4F,$96,$3F,$00,$53,$FA,$5E,$1F,$C3,$97,$E6,$9F,$DD
	.dc.b	$1F,$F4,$D3,$5F,$F4,$FB,$5F,$F5,$73,$5F,$F6,$FF,$6F,$F6,$AF,$96
	.dc.b	$FE,$CF,$99,$A4,$9F,$92,$43,$E4,$9E,$59,$FE,$E7,$F2,$BC,$89,$1F
	.dc.b	$DE,$BB,$FF,$DC,$51,$FE,$4A,$F8,$7F,$E3,$3E,$61,$A4,$C7,$EA,$5C
	.dc.b	$3E,$17,$27,$E3,$DF,$F2,$8C,$ED,$E9,$3F,$CC,$AB,$A3,$F9,$8F,$49
	.dc.b	$F1,$BF,$CA,$A1,$18,$75,$19,$8C,$57,$83,$6F,$A1,$8F,$86,$3E,$09
	.dc.b	$FE,$EA,$AE,$04,$BE,$67,$F4,$4F,$97,$7D,$97,$DE,$AF,$F6,$9C,$62
	.dc.b	$5C,$3F,$ED,$38,$C3,$6B,$85,$78,$08,$7E,$18,$FA,$80,$FF,$72,$7D
	.dc.b	$FF,$E1,$FD,$E3,$49,$1F,$E9,$1F,$69,$F8,$D7,$1F,$A2,$FF,$5D,$C6
	.dc.b	$25,$18,$9D,$47,$FB,$15,$AF,$87,$37,$E6,$1F,$EC,$E7,$8C,$7E,$A2
	.dc.b	$31,$5F,$BC,$6C,$67,$CF,$57,$08,$FF,$6C,$7A,$5E,$DF,$BC,$F2,$B6
	.dc.b	$3F,$DA,$CD,$76,$23,$B7,$C0,$74,$6D,$66,$7C,$93,$96,$C6,$5F,$98
	.dc.b	$3E,$49,$F0,$1E,$6F,$B5,$CC,$FD,$19,$CD,$27,$33,$C6,$B8,$3C,$2B
	.dc.b	$73,$AE,$0F,$AF,$2F,$A3,$C7,$0F,$DD,$6C,$64,$F0,$CF,$08,$EF,$F6
	.dc.b	$CF,$BC,$F9,$71,$86,$D7,$A2,$75,$9F,$38,$F7,$C6,$2B,$EB,$7D,$07
	.dc.b	$E4,$A7,$EA,$F0,$A3,$17,$F8,$06,$C6,$63,$F3,$B3,$E6,$CB,$F2,$2F
	.dc.b	$A8,$87,$CE,$D5,$FB,$9F,$EE,$21,$F4,$DF,$3A,$73,$D5,$F9,$F8,$C5
	.dc.b	$6D,$B1,$9A,$E9,$F8,$D3,$8B,$F1,$B5,$CE,$31,$97,$6A,$DD,$69,$B1
	.dc.b	$97,$F0,$7D,$7F,$B9,$F1,$3F,$6D,$3F,$CA,$67,$F7,$BD,$7F,$9A,$9C
	.dc.b	$9F,$DC,$BB,$24,$FF,$A4,$3C,$AF,$8A,$FE,$3B,$D1,$FE,$F5,$3F,$88
	.dc.b	$F4,$9F,$93,$7D,$1A,$FF,$09,$B6,$33,$97,$29,$FE,$05,$18,$AF,$D3
	.dc.b	$C3,$EB,$A9,$1F,$D4,$99,$3E,$97,$A3,$FE,$D8,$FA,$8F,$A0,$D8,$CF
	.dc.b	$C0,$FF,$0F,$9C,$B6,$3F,$7D,$D1,$F0,$5B,$97,$C5,$7F,$83,$74,$7E
	.dc.b	$17,$7F,$C6,$CE,$67,$F8,$D7,$57,$E5,$34,$FC,$C4,$FB,$65,$39,$7F
	.dc.b	$35,$D7,$F1,$78,$FF,$96,$7D,$2F,$D4,$24,$FD,$4B,$47,$F8,$93,$19
	.dc.b	$3F,$4D,$AF,$F0,$0F,$9A,$7D,$27,$F2,$DC,$FF,$77,$18,$9F,$23,$63
	.dc.b	$33,$B9,$F8,$37,$EE,$23,$83,$D3,$87,$C1,$7D,$B9,$7C,$97,$F6,$6D
	.dc.b	$7E,$2B,$7E,$29,$E4,$7F,$1B,$FD,$A7,$F9,$CF,$8B,$FD,$0D,$B6,$69
	.dc.b	$E6,$5F,$8B,$D2,$3F,$C7,$35,$F7,$47,$F9,$2A,$21,$F9,$92,$9F,$D0
	.dc.b	$7F,$24,$FF,$32,$3F,$E3,$86,$BF,$E2,$86,$BB,$E8,$FF,$54,$E5,$EA
	.dc.b	$F9,$EF,$EB,$5F,$0F,$FD,$49,$B5,$FC,$CA,$74,$A7,$4D,$35,$A7,$D5
	.dc.b	$8F,$EB,$7C,$1C,$A7,$FB,$35,$7C,$88,$C4,$23,$0F,$33,$F6,$9E,$FF
	.dc.b	$F9,$2D,$FE,$DD,$DB,$FE,$6F,$7E,$09,$88,$D7,$0B,$4F,$F3,$54,$FC
	.dc.b	$3F,$7B,$FC,$DF,$18,$8F,$89,$1F,$E3,$DB,$E9,$1F,$34,$87,$8B,$BF
	.dc.b	$FE,$00,$F3,$3E,$EB,$FD,$7A,$D9,$CC,$60,$FF,$3E,$F5,$7E,$6B,$27
	.dc.b	$A9,$FD,$C7,$6F,$ED,$ED,$8C,$DF,$F7,$79,$CB,$FD,$E7,$E1,$3F,$C5
	.dc.b	$57,$FE,$60,$8F,$E8,$71,$83,$FD,$AB,$9F,$AA,$3D,$07,$C6,$3F,$D3
	.dc.b	$5B,$7C,$6F,$8D,$FE,$21,$1F,$E7,$E6,$BF,$E6,$E6,$BF,$E0,$D6,$BF
	.dc.b	$E0,$6F,$7B,$FB,$F6,$4F,$A9,$FF,$75,$BF,$E7,$5E,$AF,$F0,$5C,$62
	.dc.b	$76,$D8,$CC,$FC,$EA,$FB,$5C,$1F,$EF,$CC,$BD,$51,$83,$FA,$97,$E8
	.dc.b	$CF,$B7,$FF,$41,$C7,$F8,$C7,$CE,$4E,$63,$FC,$48,$65,$C3,$FF,$00
	.dc.b	$23,$73,$CA,$3D,$98,$FF,$88,$B6,$F3,$65,$FE,$91,$3D,$85,$D3,$E0
	.dc.b	$C7,$F8,$A2,$BE,$79,$EB,$3E,$7E,$7E,$BD,$BF,$35,$18,$3E,$A6,$9E
	.dc.b	$E7,$07,$FA,$7B,$DA,$7C,$77,$5C,$B0,$D9,$FF,$20,$4E,$A7,$DF,$1F
	.dc.b	$E3,$A8,$C3,$F5,$38,$9F,$3B,$C6,$3F,$C7,$F1,$8A,$E2,$7F,$30,$9D
	.dc.b	$34,$EB,$3E,$63,$DE,$43,$EB,$1E,$67,$9D,$78,$9C,$47,$E1,$A3,$0F
	.dc.b	$13,$A1,$C8,$C6,$1F,$14,$98,$FC,$A7,$E4,$7C,$53,$47,$C4,$7F,$1A
	.dc.b	$F7,$BF,$C9,$33,$98,$C6,$1F,$E7,$78,$C3,$FC,$6E,$77,$5C,$A3,$06
	.dc.b	$DC,$76,$FF,$BC,$6B,$8B,$63,$35,$CA,$FB,$DF,$E0,$A3,$FC,$A7,$83
	.dc.b	$16,$FF,$E9,$D8,$C3,$63,$33,$C1,$B5,$FE,$19,$8C,$4E,$CF,$F5,$B6
	.dc.b	$9F,$EA,$ED,$FF,$CF,$73,$F3,$DB,$1F,$56,$30,$9F,$CD,$57,$EC,$9E
	.dc.b	$A4,$7C,$07,$FB,$CB,$7F,$C4,$47,$E3,$77,$EB,$3F,$E9,$39,$CD,$6E
	.dc.b	$73,$3F,$95,$BB,$3E,$73,$C4,$79,$49,$1F,$CC,$6F,$C4,$D3,$C6,$52
	.dc.b	$1E,$93,$27,$EF,$21,$F8,$17,$EF,$08,$FF,$A3,$8F,$DE,$BF,$C1,$79
	.dc.b	$7F,$85,$55,$E4,$FF,$5F,$36,$FC,$E7,$F0,$D8,$CD,$F8,$23,$F0,$05
	.dc.b	$D7,$EE,$D7,$FA,$62,$7F,$85,$FA,$9F,$C3,$88,$F8,$F9,$3F,$96,$C7
	.dc.b	$F2,$CE,$8F,$F2,$35,$7F,$37,$F3,$FC,$78,$FF,$37,$6B,$FC,$DD,$AF
	.dc.b	$C3,$3E,$29,$E4,$3E,$3E,$1F,$EC,$8E,$3F,$C3,$C7,$F4,$F8,$FC,$5F
	.dc.b	$57,$EB,$5B,$3A,$9F,$F2,$5F,$0F,$FB,$23,$F6,$FF,$EB,$44,$8F,$EB
	.dc.b	$54,$8F,$F5,$BF,$C6,$3B,$FF,$B4,$67,$86,$DF,$FB,$02,$75,$DF,$EA
	.dc.b	$6C,$7E,$7E,$BF,$06,$FB,$49,$87,$F8,$07,$C0,$F8,$2D,$8F,$D2,$F8
	.dc.b	$29,$CB,$63,$A1,$F5,$FA,$7F,$5E,$FE,$33,$C2,$FF,$0A,$FD,$1F,$F8
	.dc.b	$72,$0D,$23,$FE,$22,$F1,$4E,$8C,$14,$FF,$11,$F8,$A8,$DC,$FF,$8B
	.dc.b	$0F,$F1,$D7,$8D,$E6,$25,$7A,$9F,$BA,$FD,$B3,$E1,$14,$9E,$47,$F9
	.dc.b	$9F,$F1,$1F,$F2,$81,$23,$FC,$9D,$C7,$E6,$8F,$F2,$BF,$F0,$04,$9F
	.dc.b	$F1,$DF,$8D,$EA,$8C,$47,$F9,$93,$81,$F2,$BF,$E8,$62,$47,$F9,$B1
	.dc.b	$FF,$46,$FC,$34,$8F,$F4,$9B,$5F,$F3,$F6,$3F,$1B,$E8,$BF,$CC,$84
	.dc.b	$8F,$28,$7F,$8F,$D4,$8F,$F5,$D3,$5F,$F6,$2F,$B2,$C1,$87,$FB,$18
	.dc.b	$93,$FE,$93,$7F,$D3,$B7,$23,$FD,$06,$D7,$FD,$4F,$EA,$F3,$C3,$FD
	.dc.b	$4E,$48,$FF,$56,$BF,$EB,$25,$F4,$74,$8F,$FA,$E0,$C9,$FE,$C0,$F9
	.dc.b	$5F,$F6,$29,$1C,$1F,$88,$3F,$D9,$7E,$DF,$AF,$0F,$F6,$A3,$63,$47
	.dc.b	$FB,$53,$2F,$8C,$E3,$87,$FD,$AB,$5F,$D1,$89,$1F,$EE,$16,$BF,$22
	.dc.b	$7E,$DB,$FE,$B7,$C7,$FD,$DA,$47,$FD,$E7,$F4,$5F,$EF,$9F,$77,$E7
	.dc.b	$36,$8D,$3C,$BE,$8B,$E0,$1E,$37,$E9,$27,$31,$8C,$38,$9F,$06,$B8
	.dc.b	$56,$E3,$13,$BA,$F8,$30,$F8,$37,$C1,$7E,$91,$B1,$FD,$74,$F8,$EE
	.dc.b	$3C,$0E,$2D,$8F,$F7,$3E,$BF,$EE,$42,$38,$3F,$D8,$FF,$43,$FE,$E4
	.dc.b	$23,$83,$AD,$FF,$00,$8E,$AD,$8C,$88,$CB,$4F,$B7,$F0,$CD,$3F,$35
	.dc.b	$F5,$FE,$E8,$C6,$C8,$C4,$EA,$B8,$B5,$FF,$7F,$EF,$FA,$3F,$96,$FE
	.dc.b	$A1,$E5,$23,$13,$FD,$86,$31,$82,$B7,$5A,$AF,$E2,$FF,$13,$FC,$72
	.dc.b	$7F,$23,$E1,$FF,$B8,$A3,$13,$FE,$C2,$6B,$C0,$E8,$77,$F8,$AF,$DB
	.dc.b	$78,$FE,$46,$C6,$71,$FF,$5B,$0C,$93,$A9,$FE,$53,$F1,$9F,$CB,$AF
	.dc.b	$C7,$E3,$23,$13,$FE,$D4,$6B,$C0,$FF,$61,$1F,$89,$E1,$FC,$41,$1F
	.dc.b	$C4,$1F,$B3,$E3,$FC,$33,$FB,$5F,$8D,$FC,$67,$2F,$E4,$9B,$1F,$B4
	.dc.b	$6B,$B8,$FE,$6F,$F0,$8F,$07,$F7,$53,$FB,$1E,$7F,$CD,$CE,$77,$FD
	.dc.b	$DC,$E4,$FF,$42,$75,$7F,$71,$7F,$CF,$CD,$87,$BB,$FA,$E7,$FC,$86
	.dc.b	$DC,$7C,$C7,$F4,$8C,$BC,$B7,$FD,$59,$A9,$CD,$74,$BE,$D7,$CE,$BC
	.dc.b	$73,$DA,$7C,$91,$FE,$DE,$F0,$AF,$BF,$FC,$CA,$BE,$5C,$FF,$64,$8C
	.dc.b	$63,$FB,$A6,$C7,$C9,$8F,$31,$1E,$27,$95,$FE,$C3,$1C,$23,$F9,$43
	.dc.b	$5E,$D5,$E2,$3C,$BF,$20,$F9,$0F,$E8,$DF,$39,$FC,$2A,$31,$8A,$31
	.dc.b	$3A,$BF,$E5,$F5,$A9,$F9,$71,$E4,$D3,$E4,$93,$2F,$EE,$B5,$D3,$C1
	.dc.b	$E7,$8F,$F2,$EB,$FE,$7D,$39,$7A,$E3,$FD,$3D,$FE,$7B,$F2,$4F,$9B
	.dc.b	$F9,$AD,$BF,$86,$DC,$7F,$91,$B6,$BE,$51,$E5,$8F,$94,$7A,$0F,$E5
	.dc.b	$3F,$2A,$F9,$5F,$C9,$CF,$F8,$62,$31,$7F,$92,$8C,$4E,$94,$EB,$8F
	.dc.b	$FC,$B9,$ED,$FF,$91,$F0,$BE,$CF,$E5,$63,$17,$FE,$5D,$8F,$C3,$EF
	.dc.b	$F1,$E3,$CC,$4A,$FF,$33,$7C,$B7,$FA,$9A,$3F,$D3,$AD,$8C,$FC,$1F
	.dc.b	$F8,$FA,$31,$3F,$DA,$7E,$61,$F3,$09,$FD,$44,$61,$B1,$9D,$BF,$2B
	.dc.b	$39,$7C,$55,$F2,$B0,$F9,$0D,$3F,$D1,$1F,$25,$FE,$49,$8F,$F3,$E3
	.dc.b	$63,$3E,$07,$F9,$4A,$31,$1F,$DF,$7E,$07,$FD,$7F,$F3,$5F,$EB,$A2
	.dc.b	$47,$F7,$C8,$CB,$F6,$A3,$F1,$E4,$70,$7F,$80,$BD,$B7,$A1,$FE,$50
	.dc.b	$9F,$EC,$B1,$89,$46,$27,$51,$C5,$AF,$F9,$C7,$C9,$FD,$DF,$BB,$FD
	.dc.b	$0F,$4F,$D8,$B7,$4F,$EE,$2D,$AF,$F3,$ED,$70,$25,$70,$6D,$70,$6C
	.dc.b	$F0,$9D,$12,$74,$D8,$D4,$60,$8E,$08,$7A,$63,$83,$63,$6D,$36,$7F
	.dc.b	$9D,$AB,$D4,$E4,$86,$6B,$CE,$E0,$47,$07,$55,$E0,$6D,$78,$1A,$78
	.dc.b	$0E,$35,$D1,$C9,$D6,$BA,$3A,$AE,$8D,$AF,$F1,$0F,$C5,$38,$DF,$13
	.dc.b	$FE,$24,$A4,$E6,$30,$D8,$E3,$3C,$1F,$F1,$5F,$9C,$FA,$0F,$F3,$54
	.dc.b	$61,$AF,$F8,$F1,$AF,$F8,$DD,$AF,$FA,$82,$39,$63,$E3,$6C,$73,$8C
	.dc.b	$4F,$08,$FF,$10,$36,$77,$1D,$E7,$64,$79,$10,$D1,$4F,$23,$B7,$24
	.dc.b	$32,$6C,$F4,$4F,$84,$91,$C6,$36,$71,$36,$E5,$A7,$33,$D6,$E5,$A6
	.dc.b	$67,$DA,$F4,$21,$C8,$EC,$70,$2B,$C1,$D1,$C0,$D1,$5D,$38,$36,$43
	.dc.b	$04,$76,$49,$F8,$31,$EB,$DD,$BB,$97,$F1,$13,$FC,$6E,$31,$9F,$F9
	.dc.b	$D2,$B7,$5A,$AF,$F7,$1C,$62,$7F,$B1,$92,$3F,$B0,$E3,$B5,$DD,$FD
	.dc.b	$D2,$7F,$CF,$FF,$3A,$F8,$6F,$F5,$F4,$61,$B5,$FE,$A4,$9D,$5F,$36
	.dc.b	$D7,$F7,$8F,$9B,$95,$FD,$99,$AF,$F7,$4F,$07,$FD,$95,$1F,$DE,$09
	.dc.b	$1E,$67,$FD,$23,$E9,$7F,$73,$6F,$85,$C2,$7F,$D6,$B3,$96,$C6,$51
	.dc.b	$FE,$D7,$F5,$DE,$97,$EE,$73,$FF,$54,$36,$33,$3F,$EA,$09,$CB,$FB
	.dc.b	$FD,$7F,$C3,$2F,$F8,$38,$87,$CC,$3F,$D0,$3F,$5F,$FB,$69,$FF,$36
	.dc.b	$37,$D3,$34,$D7,$E2,$8F,$A8,$F2,$36,$FC,$4D,$7F,$DA,$76,$8C,$47
	.dc.b	$84,$87,$CD,$8F,$46,$DE,$1F,$84,$91,$ED,$C5,$18,$AF,$F6,$FF,$82
	.dc.b	$FD,$11,$F7,$1F,$EF,$FF,$BB,$CC,$63,$14,$62,$75,$7F,$E0,$5A,$D4
	.dc.b	$FC,$68,$F2,$14,$FA,$BB,$7C,$C7,$85,$2D,$F2,$7D,$5B,$3A,$6B,$F7
	.dc.b	$B1,$F4,$57,$1A,$FF,$6C,$FE,$B6,$FC,$87,$F9,$13,$70,$FE,$74,$DC
	.dc.b	$37,$3F,$C9,$F1,$F2,$47,$A0,$F9,$A7,$A3,$6E,$73,$2F,$9B,$67,$FD
	.dc.b	$86,$FE,$04,$91,$F4,$5E,$D3,$23,$EB,$92,$7F,$D5,$FD,$FF,$9A,$E8
	.dc.b	$B1,$FB,$FF,$75,$EE,$7F,$55,$8C,$36,$7F,$D6,$FF,$35,$FA,$2F,$BB
	.dc.b	$B4,$69,$1F,$D3,$8B,$EB,$CF,$C6,$33,$18,$79,$18,$C3,$F3,$AE,$0F
	.dc.b	$D8,$C6,$EB,$FD,$93,$DD,$FB,$4F,$7A,$D7,$FB,$2C,$E5,$19,$69,$93
	.dc.b	$F7,$7A,$3F,$17,$F8,$15,$FA,$46,$CF,$AA,$75,$AF,$38,$E3,$3A,$79
	.dc.b	$FC,$AF,$E7,$97,$D8,$FE,$BA,$30,$DE,$5F,$E1,$6D,$9D,$9F,$E4,$BF
	.dc.b	$17,$F9,$C4,$7F,$35,$6C,$67,$C5,$DB,$FC,$ED,$AF,$E7,$FF,$36,$F6
	.dc.b	$D7,$EF,$9B,$1E,$B6,$BF,$09,$AF,$E9,$F1,$7D,$1F,$E1,$E3,$E1,$46
	.dc.b	$1F,$E8,$DA,$7C,$99,$D4,$7C,$46,$BF,$81,$E1,$FE,$0A,$FE,$33,$B7
	.dc.b	$3C,$5F,$90,$6C,$EB,$97,$D1,$F2,$FF,$B9,$F2,$6E,$5F,$E1,$FC,$7D
	.dc.b	$53,$96,$C6,$5F,$67,$E0,$39,$6C,$65,$F8,$D1,$96,$C6,$7C,$E7,$EF
	.dc.b	$6F,$27,$D7,$18,$6C,$66,$BF,$B4,$4E,$5F,$EC,$9E,$67,$F6,$97,$EC
	.dc.b	$67,$F6,$F8,$F3,$C9,$FD,$CA,$7F,$4F,$6F,$B6,$7C,$DF,$F6,$69,$FF
	.dc.b	$30,$46,$34,$EE,$DC,$7F,$5F,$3F,$E5,$E8,$C3,$FA,$0F,$B6,$FD,$29
	.dc.b	$1C,$1F,$73,$AB,$FA,$EC,$7E,$85,$B1,$91,$19,$69,$F8,$BF,$C7,$3C
	.dc.b	$CF,$E4,$5B,$83,$A1,$18,$6C,$7E,$0E,$5E,$7F,$AB,$9F,$F0,$91,$83
	.dc.b	$F8,$27,$3F,$F7,$29,$D7,$3F,$E3,$1F,$83,$BF,$D1,$6F,$1F,$D5,$6F
	.dc.b	$27,$1C,$3E,$7C,$E5,$FD,$64,$FD,$52,$38,$3E,$AE,$FF,$46,$3B,$B6
	.dc.b	$33,$C3,$FE,$07,$7E,$93,$73,$FE,$5B,$E1,$FF,$A6,$67,$F8,$F3,$5F
	.dc.b	$E3,$B9,$FF,$4D,$7F,$D2,$E6,$8A,$7E,$0A,$BF,$AE,$67,$E2,$6C,$78
	.dc.b	$B5,$DB,$E2,$6E,$1E,$28,$78,$FF,$2B,$1F,$EC,$11,$DE,$30,$E8,$E8
	.dc.b	$F7,$AC,$47,$8D,$AF,$78,$FC,$F3,$0F,$E3,$05,$3F,$88,$7D,$0F,$F8
	.dc.b	$77,$80,$D2,$57,$F8,$A6,$3F,$C3,$8B,$EB,$BF,$8E,$79,$DF,$DF,$5B
	.dc.b	$19,$BF,$E3,$D3,$97,$F9,$07,$D2,$F7,$24,$F5,$7F,$D3,$9D,$5F,$E6
	.dc.b	$29,$F3,$1F,$E6,$2E,$9F,$F3,$53,$8A,$FC,$E7,$E0,$5E,$92,$3E,$77
	.dc.b	$9C,$F0,$9F,$82,$43,$F9,$C1,$FC,$D7,$C2,$F8,$44,$AF,$80,$FC,$5F
	.dc.b	$83,$FF,$11,$CF,$F3,$76,$BF,$CF,$DB,$3F,$AF,$6E,$8F,$9F,$F9,$C4
	.dc.b	$9F,$98,$F9,$BB,$BF,$DA,$59,$FC,$1C,$3C,$C7,$C1,$8F,$A5,$ED,$7D
	.dc.b	$72,$5E,$0F,$F2,$F7,$CA,$3E,$5B,$FD,$99,$1F,$DA,$63,$16,$F6,$BF
	.dc.b	$DA,$B2,$9C,$C6,$1B,$1E,$29,$63,$E5,$7E,$49,$E6,$8F,$E1,$DE,$7F
	.dc.b	$FB,$89,$FE,$BC,$D8,$FE,$BC,$D7,$FB,$33,$63,$C5,$87,$F8,$11,$F4
	.dc.b	$36,$7F,$C0,$F7,$EA,$9F,$34,$E6,$7E,$31,$F8,$83,$CF,$BF,$E0,$5F
	.dc.b	$39,$FD,$BF,$E2,$BF,$CA,$95,$F0,$B6,$46,$23,$FB,$B3,$67,$84,$70
	.dc.b	$7F,$C3,$3E,$17,$2F,$67,$E5,$B6,$7F,$C3,$0D,$9D,$C7,$F9,$8E,$7F
	.dc.b	$CC,$91,$FC,$43,$E0,$FF,$C3,$CD,$3F,$C4,$A5,$36,$7F,$7B,$76,$FC
	.dc.b	$03,$EB,$6B,$F5,$5C,$1E,$0A,$E6,$FF,$9C,$EF,$D5,$1E,$63,$E9,$38
	.dc.b	$8F,$8E,$72,$21,$F3,$38,$3F,$2F,$9F,$CD,$8C,$DF,$85,$D1,$FD,$90
	.dc.b	$7B,$AF,$19,$E3,$FF,$AF,$3D,$67,$BD,$FE,$58,$AD,$4E,$63,$15,$C2
	.dc.b	$FF,$D1,$71,$89,$D4,$7F,$B1,$AB,$8B,$A3,$15,$C6,$FB,$C6,$2B,$8B
	.dc.b	$E6,$9D,$5F,$F8,$F4,$FF,$04,$12,$3F,$C8,$4D,$74,$60,$87,$D1,$7F
	.dc.b	$0B,$A7,$FA,$8D,$AF,$C0,$3E,$91,$63,$99,$0F,$E0,$51,$DD,$FC,$11
	.dc.b	$B2,$BF,$86,$F0,$3F,$D1,$7B,$7F,$BA,$23,$1E,$1A,$31,$7E,$0E,$8F
	.dc.b	$19,$1C,$1F,$A2,$3F,$DB,$E5,$F5,$3F,$51,$C7,$FE,$BD,$6C,$6B,$2F
	.dc.b	$41,$F8,$7C,$7D,$0D,$3F,$7A,$6C,$FF,$63,$73,$7F,$16,$AF,$DE,$C6
	.dc.b	$6B,$F8,$2B,$79,$5C,$FF,$2E,$7F,$87,$9F,$0C,$FE,$26,$58,$DB,$F1
	.dc.b	$E7,$C4,$79,$B8,$9C,$6E,$4F,$C2,$CF,$D1,$AF,$0F,$77,$90,$8F,$A9
	.dc.b	$FE,$15,$D5,$F7,$67,$FC,$FC,$D7,$FC,$FC,$D7,$FD,$0C,$D7,$C6,$7D
	.dc.b	$E3,$F1,$25,$3F,$D0,$DB,$F1,$AF,$F1,$E3,$72,$FE,$9D,$1E,$89,$FF
	.dc.b	$49,$4E,$7A,$FF,$DE,$5F,$83,$7F,$A4,$DB,$3F,$E9,$36,$BF,$D7,$69
	.dc.b	$3F,$D9,$5E,$87,$F9,$2F,$B7,$C0,$47,$C1,$3E,$A2,$6B,$F3,$34,$CB
	.dc.b	$78,$7F,$D4,$9D,$DE,$92,$5F,$6D,$11,$F1,$6B,$86,$1F,$10,$EB,$BF
	.dc.b	$D7,$C3,$FD,$5D,$DF,$F0,$C9,$8F,$37,$E3,$70,$21,$F8,$27,$F8,$60
	.dc.b	$98,$7F,$86,$23,$2F,$FA,$DF,$C2,$F9,$C4,$8F,$33,$E5,$C7,$C5,$3E
	.dc.b	$5C,$BC,$EF,$EE,$7C,$6F,$F4,$67,$8D,$E1,$FC,$27,$FC,$74,$49,$FF
	.dc.b	$24,$3F,$E4,$02,$3F,$E5,$16,$BF,$E3,$FF,$BB,$FF,$21,$11,$FF,$28
	.dc.b	$3F,$E4,$82,$3F,$E5,$96,$BF,$E4,$7F,$8E,$FF,$25,$11,$FF,$2C,$3F
	.dc.b	$EA,$52,$3F,$ED,$F6,$A1,$7F,$EA,$F8,$FF,$3E,$C1,$A4,$7F,$D1,$26
	.dc.b	$4A,$FF,$A2,$BC,$EE,$8D,$8D,$1F,$EE,$5F,$3F,$EC,$7C,$2F,$FB,$AD
	.dc.b	$B1,$FE,$A1,$D9,$39,$6C,$7F,$AA,$7E,$FB,$FD,$50,$47,$07,$AC,$FF
	.dc.b	$73,$7B,$5E,$D8,$FF,$4E,$B6,$35,$E1,$BC,$34,$E6,$7F,$D1,$2D,$7F
	.dc.b	$D0,$18,$7F,$9D,$C8,$FF,$A1,$5F,$F6,$91,$1F,$F6,$4F,$07,$E9,$E3
	.dc.b	$11,$F3,$F5,$FF,$6E,$47,$CE,$24,$7F,$AA,$BD,$2F,$A9,$1F,$ED,$2C
	.dc.b	$FF,$D8,$1E,$9F,$E7,$F8,$B1,$D4,$8C,$61,$F0,$6B,$67,$FB,$81,$C4
	.dc.b	$EE,$31,$1F,$81,$7F,$D9,$9E,$A7,$28,$C4,$A3,$11,$FA,$F8,$C4,$F2
	.dc.b	$8C,$4E,$B6,$FF,$67,$E3,$F8,$12,$4F,$E2,$1F,$67,$AE,$F5,$FD,$F3
	.dc.b	$E9,$8C,$3F,$80,$3C,$CE,$27,$8C,$F0,$9F,$39,$FC,$0F,$EE,$BE,$2B
	.dc.b	$4F,$82,$53,$78,$A3,$FC,$39,$BF,$F1,$D6,$F4,$7F,$49,$6F,$7E,$EF
	.dc.b	$EE,$CE,$5F,$A1,$ED,$3C,$1F,$D6,$5F,$D9,$8C,$4F,$13,$E8,$38,$8D
	.dc.b	$4F,$09,$F8,$5E,$5F,$ED,$7C,$4F,$8E,$D8,$C8,$8C,$B4,$FC,$EF,$3F
	.dc.b	$E7,$1C,$1F,$9C,$29,$FA,$62,$F0,$7E,$8F,$9F,$B4,$7F,$BA,$BC,$8F
	.dc.b	$E0,$DF,$73,$FA,$D8,$57,$EA,$67,$31,$E0,$6B,$C4,$4F,$19,$FD,$87
	.dc.b	$BD,$A9,$E7,$82,$31,$3A,$9E,$27,$B2,$33,$3F,$E5,$18,$FC,$31,$CB
	.dc.b	$2F,$C2,$34,$CD,$7E,$FE,$33,$0F,$7F,$FC,$17,$E3,$7F,$97,$5B,$1E
	.dc.b	$F6,$BF,$C3,$5A,$FF,$51,$F8,$37,$B3,$FD,$ED,$BC,$BF,$DB,$9B,$7F
	.dc.b	$BB,$6D,$7C,$BF,$81,$AF,$F1,$96,$BF,$CD,$5A,$9B,$87,$8A,$3D,$5E
	.dc.b	$77,$F8,$46,$30,$66,$78,$93,$1F,$F0,$B3,$DB,$DA,$7C,$1F,$E6,$73
	.dc.b	$78,$DF,$7E,$73,$CB,$FD,$BE,$31,$5B,$6C,$67,$3F,$91,$9F,$8C,$98
	.dc.b	$7F,$83,$33,$F9,$38,$F4,$8F,$F0,$AD,$7F,$39,$8C,$79,$DD,$27,$33
	.dc.b	$FD,$5E,$72,$FF,$1F,$D7,$CF,$3F,$3F,$2F,$A4,$47,$C1,$8F,$8B,$6F
	.dc.b	$BC,$DB,$FE,$51,$F0,$8F,$37,$FC,$DD,$18,$6B,$FC,$9E,$34,$FF,$8A
	.dc.b	$9B,$5F,$77,$9B,$F0,$44,$C3,$F0,$4F,$7A,$FC,$56,$15,$FE,$E5,$AF
	.dc.b	$E3,$29,$F0,$AF,$6B,$C3,$5F,$E1,$F8,$C3,$FE,$3A,$6E,$3F,$D1,$DB
	.dc.b	$1F,$B7,$F1,$DF,$0D,$FE,$27,$F3,$3F,$D7,$51,$8C,$7F,$BA,$D7,$F8
	.dc.b	$0E,$31,$3A,$AE,$13,$F7,$2B,$6F,$E4,$1B,$97,$F9,$D2,$BE,$6C,$6B
	.dc.b	$E3,$3F,$D6,$B1,$FE,$C9,$6D,$7F,$0B,$6B,$FD,$E5,$AF,$F8,$57,$E2
	.dc.b	$3F,$BA,$D7,$EF,$A1,$18,$F1,$BF,$DA,$8D,$8F,$E2,$CD,$7E,$7C,$F1
	.dc.b	$C3,$FD,$13,$5E,$62,$4F,$F3,$CF,$13,$B7,$8B,$FD,$E6,$31,$3F,$DE
	.dc.b	$9F,$F1,$47,$BE,$F8,$4F,$F4,$03,$71,$FF,$34,$B7,$83,$FD,$D0,$DA
	.dc.b	$FE,$54,$D7,$FC,$30,$D7,$AD,$74,$25,$74,$6D,$74,$6C,$F4,$6B,$D1
	.dc.b	$AF,$49,$E4,$49,$E4,$D8,$E4,$D7,$93,$5E,$51,$C0,$8F,$06,$BC,$1A
	.dc.b	$F0,$34,$43,$45,$34,$60,$86,$08,$2B,$E4,$36,$BD,$ED,$AF,$43,$6B
	.dc.b	$C6,$D3,$C0,$53,$B9,$4E,$F5,$D8,$95,$DB,$DE,$77,$BD,$DF,$EC,$D7
	.dc.b	$FE,$6D,$9C,$C6,$21,$18,$76,$D7,$F0,$BF,$2D,$FE,$9D,$7F,$12,$D7
	.dc.b	$39,$73,$C5,$39,$9F,$C5,$B5,$FC,$BB,$5F,$90,$D8,$E3,$3C,$E3,$04
	.dc.b	$70,$49,$F1,$B5,$F1,$98,$21,$82,$1E,$47,$83,$4D,$94,$D9,$E8,$76
	.dc.b	$D3,$65,$36,$7B,$1D,$B4,$D9,$4D,$CF,$C0,$72,$D7,$33,$F0,$5C,$B5
	.dc.b	$CC,$F6,$9E,$11,$E9,$21,$E5,$29,$E1,$2B,$C1,$D1,$07,$03,$3F,$F5
	.dc.b	$B6,$3F,$CA,$5B,$3F,$CA,$5A,$FF,$29,$6B,$FE,$BD,$D7,$FD,$F1,$3F
	.dc.b	$E5,$A2,$47,$F1,$F7,$F9,$B9,$1F,$E7,$CD,$7F,$D5,$9E,$53,$E2,$BF
	.dc.b	$D9,$1F,$0B,$F9,$E8,$C4,$A3,$13,$A8,$FF,$24,$4E,$A7,$FC,$D6,$48
	.dc.b	$FE,$60,$D7,$FD,$75,$C7,$FD,$02,$3F,$9E,$F9,$3F,$EF,$F6,$EB,$DE
	.dc.b	$BF,$9F,$CE,$5F,$F2,$B7,$EC,$AC,$3F,$61,$08,$C3,$FB,$36,$BF,$B8
	.dc.b	$6B,$FD,$0C,$FF,$40,$90,$C1,$FE,$5B,$FA,$BF,$D6,$79,$06,$93,$0E
	.dc.b	$CF,$8A,$B5,$58,$21,$D4,$FF,$6D,$6B,$E9,$E6,$FE,$10,$D8,$D1,$FE
	.dc.b	$F9,$C3,$DD,$87,$F0,$D6,$C7,$F1,$26,$BF,$E7,$1C,$7F,$81,$7E,$80
	.dc.b	$FA,$C5,$BE,$3B,$F8,$CC,$E6,$31,$7F,$E8,$18,$C4,$EA,$B8,$BE,$49
	.dc.b	$63,$E7,$7E,$39,$EC,$C1,$97,$F2,$3C,$3B,$7C,$37,$3C,$8D,$25,$F8
	.dc.b	$7E,$B7,$D2,$49,$E4,$FF,$1D,$C7,$C1,$3D,$F0,$F9,$38,$F8,$23,$CD
	.dc.b	$8A,$F9,$47,$C0,$6B,$FE,$9F,$FC,$91,$EE,$7E,$29,$B1,$FC,$ED,$AF
	.dc.b	$F4,$06,$BF,$00,$FD,$95,$F8,$5C,$1F,$19,$C2,$26,$6F,$17,$F9,$C3
	.dc.b	$74,$FE,$94,$D9,$FE,$9E,$D7,$91,$F7,$3E,$27,$FC,$7D,$F7,$1F,$EB
	.dc.b	$9A,$46,$27,$71,$8F,$0F,$F0,$75,$BA,$D5,$79,$6B,$8F,$D2,$6A,$38
	.dc.b	$CA,$31,$1B,$8C,$4A,$BA,$56,$AB,$CD,$E0,$FE,$4B,$F3,$2F,$03,$63
	.dc.b	$AB,$5E,$8D,$7D,$D8,$5C,$3D,$AD,$3D,$EA,$BD,$91,$9C,$3F,$43,$F7
	.dc.b	$1E,$F6,$DF,$B1,$AF,$AD,$AF,$D7,$FB,$D7,$DE,$E2,$11,$8C,$57,$FE
	.dc.b	$04,$8C,$36,$B9,$CF,$D4,$F0,$2E,$3F,$D0,$63,$8E,$1F,$3A,$72,$77
	.dc.b	$E1,$FE,$63,$1F,$91,$EF,$FC,$B1,$1F,$C3,$F0,$FE,$28,$97,$83,$FC
	.dc.b	$2E,$7E,$13,$E4,$3F,$81,$FE,$69,$F8,$16,$D7,$DE,$6B,$F7,$5A,$FE
	.dc.b	$3F,$F1,$BB,$5F,$8E,$6E,$66,$9C,$25,$97,$EA,$1F,$CC,$B5,$FD,$23
	.dc.b	$5D,$F3,$7F,$3C,$EA,$FD,$FC,$7E,$BB,$AB,$F7,$E4,$7F,$4B,$A7,$C0
	.dc.b	$6E,$5F,$05,$FC,$7A,$F4,$E9,$23,$F2,$31,$97,$F3,$F9,$FF,$C0,$16
	.dc.b	$C6,$5F,$D0,$46,$5B,$19,$7D,$7B,$7D,$57,$EC,$1F,$C2,$DF,$EA,$FD
	.dc.b	$1F,$BB,$8F,$DD,$B5,$FE,$0B,$87,$AD,$FE,$2D,$9F,$F1,$52,$3F,$C5
	.dc.b	$5F,$8D,$D3,$FC,$3A,$3E,$31,$FC,$DF,$4F,$92,$FF,$30,$AE,$8F,$81
	.dc.b	$AF,$A0,$F3,$10,$F4,$9F,$30,$87,$F1,$23,$A6,$3F,$3D,$FE,$59,$8F
	.dc.b	$C1,$69,$CC,$E1,$8F,$BD,$E2,$74,$3C,$93,$F7,$5F,$E5,$77,$EA,$3F
	.dc.b	$07,$1F,$C7,$BC,$7F,$F3,$13,$4F,$CF,$E7,$FE,$31,$71,$5F,$C7,$FE
	.dc.b	$8B,$FB,$1F,$CA,$59,$34,$93,$FC,$A3,$E5,$FF,$D8,$0F,$F2,$D6,$B9
	.dc.b	$EB,$F0,$C6,$23,$FA,$A1,$F4,$76,$FA,$8F,$F4,$3E,$1F,$D9,$FE,$0D
	.dc.b	$FE,$9D,$C6,$F9,$7E,$A8,$C5,$7F,$3B,$6B,$F0,$AF,$9E,$DF,$0D,$FC
	.dc.b	$05,$A9,$F5,$1C,$C9,$1E,$07,$FD,$11,$E2,$BD,$F7,$C2,$FF,$55,$FC
	.dc.b	$F3,$EE,$4E,$63,$19,$FF,$50,$8C,$29,$DD,$72,$8C,$1F,$D4,$31,$7D
	.dc.b	$77,$B4,$99,$7C,$27,$FB,$E6,$7E,$C2,$6D,$F8,$1B,$FE,$A9,$6F,$9C
	.dc.b	$F8,$C4,$C3,$E5,$BF,$82,$F8,$87,$C6,$7F,$60,$BF,$EB,$91,$87,$FA
	.dc.b	$F1,$88,$D3,$FE,$26,$8D,$3F,$EA,$FE,$0F,$9D,$8F,$8E,$3F,$5A,$7B
	.dc.b	$7E,$C7,$FD,$75,$8D,$3A,$C7,$F8,$C3,$24,$EA,$3D,$6D,$7C,$18,$F8
	.dc.b	$A9,$8F,$89,$FE,$F2,$D9,$F4,$1D,$8F,$C6,$1F,$8F,$C3,$CD,$3F,$83
	.dc.b	$FB,$0E,$CD,$3C,$A5,$9F,$F6,$83,$4F,$B1,$EB,$FF,$92,$3D,$DC,$E7
	.dc.b	$FE,$E2,$94,$62,$75,$1C,$0C,$96,$B4,$F8,$2B,$94,$60,$F9,$DB,$7E
	.dc.b	$01,$F5,$1F,$A1,$AF,$4C,$F0,$AE,$84,$9F,$D7,$9F,$A5,$C7,$EB,$3F
	.dc.b	$10,$C1,$FE,$59,$F6,$3F,$DE,$9D,$BF,$EB,$18,$FF,$7D,$61,$FE,$FD
	.dc.b	$7F,$CA,$BE,$37,$F8,$BA,$B6,$D9,$D5,$71,$AF,$F7,$FC,$62,$75,$EB
	.dc.b	$FF,$8F,$9B,$3D,$0E,$F1,$97,$FA,$E4,$7F,$2B,$7F,$83,$37,$87,$FD
	.dc.b	$0C,$DD,$3F,$CD,$AD,$CB,$C5,$5C,$8F,$E6,$95,$C1,$B5,$C1,$FA,$D5
	.dc.b	$F5,$09,$9E,$0D,$DF,$49,$F1,$98,$9F,$01,$F5,$67,$FC,$35,$1F,$B6
	.dc.b	$C7,$FD,$1F,$3D,$23,$13,$A7,$F4,$93,$FE,$94,$F9,$8F,$C2,$92,$7F
	.dc.b	$14,$FE,$43,$7E,$71,$FC,$2D,$B5,$FC,$11,$AF,$EB,$DA,$FA,$0F,$F4
	.dc.b	$35,$FA,$1F,$51,$FA,$CE,$1F,$F3,$A4,$7E,$8B,$87,$F8,$33,$63,$37
	.dc.b	$BA,$F5,$3F,$CB,$F8,$BC,$2D,$33,$C3,$FC,$01,$B1,$E4,$DB,$EA,$57
	.dc.b	$90,$F1,$94,$FD,$EE,$1E,$17,$BE,$3E,$4B,$F1,$4F,$ED,$38,$FE,$7D
	.dc.b	$7F,$2F,$6B,$FC,$A1,$AF,$F1,$36,$BF,$14,$FC,$41,$E1,$29,$FC,$67
	.dc.b	$93,$8D,$7F,$4D,$6E,$5F,$CF,$23,$F2,$26,$4F,$C9,$F3,$FF,$46,$FD
	.dc.b	$BB,$FA,$2F,$4F,$FA,$85,$B8,$1A,$7F,$5F,$EA,$EA,$4B,$EB,$87,$F5
	.dc.b	$C7,$F2,$9A,$FC,$AB,$F3,$9E,$B3,$D9,$1F,$D8,$FA,$FC,$E4,$79,$BF
	.dc.b	$0A,$79,$4F,$16,$9D,$65,$D8,$ED,$F8,$6D,$8F,$82,$DD,$FF,$C0,$D3
	.dc.b	$F2,$3C,$07,$81,$FE,$F2,$F9,$7F,$F0,$81,$32,$FF,$56,$C6,$5F,$F5
	.dc.b	$7F,$85,$C4,$99,$7C,$EB,$47,$A7,$AF,$F7,$95,$F6,$4F,$53,$87,$FC
	.dc.b	$97,$5F,$BB,$ED,$FF,$67,$67,$FE,$66,$24,$FF,$8D,$9F,$F2,$B9,$1C
	.dc.b	$1F,$E3,$8F,$13,$F1,$33,$CA,$31,$5D,$67,$F1,$BE,$3B,$C5,$FD,$6D
	.dc.b	$72,$3F,$D0,$2E,$1A,$71,$A9,$87,$92,$3F,$C8,$6F,$EE,$5F,$77,$07
	.dc.b	$D6,$CB,$DD,$1F,$E8,$83,$F8,$C6,$1F,$EF,$AA,$FF,$7C,$78,$DD,$E7
	.dc.b	$FD,$1E,$D7,$FC,$B3,$1D,$DF,$8E,$D7,$FD,$33,$8E,$EB,$D9,$1D,$A3
	.dc.b	$07,$5C,$7F,$DA,$A3,$F3,$8F,$F6,$9B,$67,$FD,$17,$18,$3F,$D2,$A4
	.dc.b	$7F,$D2,$6D,$7E,$09,$F3,$AF,$C9,$A7,$D1,$8F,$87,$97,$C2,$CF,$E2
	.dc.b	$78,$3F,$0B,$4F,$41,$2B,$07,$E2,$FC,$96,$6F,$DE,$7F,$CF,$8F,$FA
	.dc.b	$F4,$C4,$8F,$F8,$01,$3C,$27,$7C,$3F,$EC,$1A,$DD,$7F,$B2,$5A,$F6
	.dc.b	$87,$97,$C6,$3F,$DA,$13,$A3,$FD,$AC,$70,$7E,$A9,$F0,$48,$7C,$43
	.dc.b	$F1,$3A,$FE,$5D,$A7,$E3,$CF,$03,$EE,$FB,$2F,$3B,$63,$FD,$E1,$29
	.dc.b	$CB,$63,$FD,$3D,$E8,$7F,$BF,$E7,$FD,$F9,$39,$9F,$88,$27,$4F,$0F
	.dc.b	$3F,$B3,$F2,$65,$8B,$D1,$CC,$FC,$56,$CE,$9B,$3B,$8F,$04,$FC,$B8
	.dc.b	$FB,$3F,$9A,$7E,$38,$91,$FE,$D8,$7F,$DC,$24,$7F,$DB,$ED,$7C,$90
	.dc.b	$F5,$3F,$07,$F4,$DF,$71,$B3,$EE,$BF,$4C,$62,$38,$B5,$FB,$DE,$B5
	.dc.b	$9F,$7B,$6C,$D3,$84,$FD,$E3,$EE,$1F,$B2,$3E,$F9,$FC,$FB,$8F,$F3
	.dc.b	$BB,$7F,$03,$8F,$DD,$9F,$9C,$7F,$7E,$D3,$FD,$AD,$EC,$3D,$8F,$EC
	.dc.b	$2D,$C7,$FC,$3C,$DA,$FF,$19,$36,$BF,$BA,$9E,$1E,$6F,$D7,$1F,$C4
	.dc.b	$23,$51,$F8,$26,$BF,$A7,$6A,$68,$F6,$BF,$63,$F1,$3E,$06,$CF,$A1
	.dc.b	$AF,$56,$BE,$3E,$8F,$EC,$77,$E4,$7D,$73,$FB,$CF,$8B,$E9,$91,$A4
	.dc.b	$BF,$F5,$FD,$FF,$81,$3C,$93,$49,$5D,$B0,$99,$FD,$1B,$FE,$E1,$EE
	.dc.b	$BD,$EF,$E5,$8D,$8F,$F2,$53,$67,$FC,$BC,$D8,$FF,$67,$CF,$84,$97
	.dc.b	$FE,$51,$C7,$A3,$E0,$9F,$A6,$7F,$5E,$C3,$AE,$3E,$02,$46,$0F,$F2
	.dc.b	$D7,$C0,$BE,$03,$F8,$FF,$C2,$FF,$9B,$E3,$18,$23,$13,$AA,$FD,$E8
	.dc.b	$AD,$CF,$F9,$C4,$91,$FE,$68,$6B,$FC,$12,$FF,$A3,$C7,$F9,$F5,$B1
	.dc.b	$D1,$AE,$FC,$2F,$E5,$D3,$FE,$83,$F8,$57,$C4,$7F,$36,$F8,$6C,$36
	.dc.b	$FC,$B3,$AC,$D5,$A6,$CF,$F9,$F2,$3F,$D1,$04,$70,$7B,$BE,$1B,$FD
	.dc.b	$49,$1F,$E9,$36,$C6,$78,$7F,$A1,$4E,$5B,$1F,$1B,$CE,$FF,$0A,$DF
	.dc.b	$FA,$25,$B3,$FE,$9D,$6C,$EF,$CA,$FF,$29,$C6,$23,$F9,$5B,$5F,$EA
	.dc.b	$0D,$4D,$24,$72,$25,$78,$5B,$5E,$16,$D7,$86,$74,$49,$D3,$63,$51
	.dc.b	$82,$38,$21,$E8,$9E,$AD,$9E,$4D,$9E,$2D,$38,$9E,$E9,$E2,$D9,$E2
	.dc.b	$D9,$E2,$D3,$89,$F1,$27,$8B,$67,$8B,$67,$8B,$4E,$35,$F2,$A7,$6D
	.dc.b	$9D,$B6,$77,$5F,$E8,$AF,$71,$ED,$FF,$BA,$67,$F4,$F1,$8F,$73,$FD
	.dc.b	$B9,$1F,$EE,$66,$C6,$63,$FD,$C1,$F8,$EB,$E4,$7F,$C2,$71,$86,$BF
	.dc.b	$B2,$8D,$67,$D1,$B1,$B8,$C4,$F0,$8F,$EE,$B3,$A7,$94,$65,$FF,$0E
	.dc.b	$46,$5D,$C6,$54,$65,$DC,$66,$31,$1A,$8C,$12,$7C,$64,$8F,$31,$0F
	.dc.b	$1B,$96,$99,$3C,$6E,$5A,$64,$F1,$B9,$69,$93,$C8,$E5,$A6,$67,$D2
	.dc.b	$49,$F4,$46,$5A,$F9,$08,$78,$8A,$78,$0A,$76,$2C,$75,$9F,$13,$63
	.dc.b	$B4,$62,$7E,$87,$17,$F9,$BB,$0F,$E5,$1C,$5F,$E5,$16,$DF,$E3,$CF
	.dc.b	$F2,$A4,$ED,$FE,$2C,$64,$6E,$DB,$94,$6A,$BF,$8D,$3B,$EE,$FE,$C3
	.dc.b	$CD,$70,$34,$95,$FD,$12,$9F,$90,$3C,$DF,$F3,$37,$C4,$E6,$31,$28
	.dc.b	$C4,$EA,$3F,$A6,$CE,$A7,$C6,$48,$F5,$B5,$FF,$27,$F2,$FF,$3F,$8F
	.dc.b	$E7,$C4,$8F,$E8,$7C,$DF,$E7,$C8,$FF,$3D,$FC,$B3,$E0,$BF,$67,$C4
	.dc.b	$8C,$5F,$AE,$31,$3A,$9C,$CF,$C6,$21,$F3,$8A,$7F,$97,$7D,$87,$A9
	.dc.b	$FA,$DB,$46,$1E,$CD,$9F,$F4,$7B,$5F,$F2,$77,$BC,$FC,$17,$CE,$D8
	.dc.b	$F3,$35,$FF,$4B,$B5,$F2,$10,$FF,$3A,$F2,$7E,$05,$FE,$26,$D7,$C3
	.dc.b	$E2,$F9,$EC,$D2,$61,$F7,$FE,$41,$F2,$4F,$A1,$FF,$54,$FC,$77,$F8
	.dc.b	$1A,$73,$18,$C7,$FD,$0B,$18,$B5,$6E,$F9,$46,$2B,$D0,$7F,$AB,$F2
	.dc.b	$F3,$38,$3F,$37,$D3,$FC,$B3,$0E,$9C,$BC,$F2,$FF,$5D,$B7,$0F,$C8
	.dc.b	$DF,$FA,$E6,$B6,$FF,$99,$3E,$8B,$F8,$F7,$96,$69,$2B,$E3,$E3,$FC
	.dc.b	$EA,$BC,$26,$08,$7F,$9A,$27,$83,$67,$FA,$0D,$75,$7F,$03,$F2,$DF
	.dc.b	$ED,$0F,$C0,$7F,$02,$DA,$FF,$47,$70,$23,$11,$F8,$D6,$BF,$17,$1F
	.dc.b	$26,$BF,$1A,$3E,$FC,$60,$FE,$A3,$CC,$FC,$03,$FC,$A9,$3F,$E3,$C8
	.dc.b	$C7,$97,$A9,$CC,$FF,$AF,$27,$2F,$E9,$7F,$77,$7D,$BF,$FB,$8F,$D7
	.dc.b	$FF,$73,$C6,$27,$FD,$78,$7F,$A9,$DC,$46,$A7,$84,$FF,$B2,$3C,$E7
	.dc.b	$C7,$3C,$5C,$CF,$FB,$BE,$73,$18,$84,$61,$EE,$6E,$79,$38,$36,$BE
	.dc.b	$83,$FD,$F5,$F2,$BF,$E5,$CA,$F0,$C6,$2B,$C9,$5B,$55,$B8,$C3,$7E
	.dc.b	$07,$84,$F8,$5A,$F8,$1B,$7E,$E7,$3C,$8F,$9D,$7D,$F2,$DF,$2F,$FE
	.dc.b	$70,$AD,$4E,$63,$15,$C2,$B6,$7F,$B5,$5C,$47,$F3,$5F,$B8,$FA,$8D
	.dc.b	$DB,$95,$7F,$9F,$27,$2D,$8C,$BF,$EF,$7A,$7C,$8E,$21,$8F,$A6,$30
	.dc.b	$D8,$F1,$9C,$63,$3F,$B6,$FF,$4A,$97,$57,$B6,$30,$DF,$33,$F3,$B2
	.dc.b	$9C,$DF,$FB,$C5,$7D,$16,$9B,$3D,$27,$53,$93,$F4,$FB,$FE,$3B,$E0
	.dc.b	$7E,$77,$CD,$7D,$B6,$D7,$E0,$1A,$FE,$2D,$AF,$E4,$3F,$3C,$FA,$CA
	.dc.b	$7F,$15,$18,$6C,$70,$A4,$E6,$3F,$59,$FC,$02,$F8,$9F,$C0,$6B,$FB
	.dc.b	$38,$C4,$FA,$CF,$ED,$6E,$23,$53,$C2,$7F,$D6,$BE,$A7,$F7,$89,$FE
	.dc.b	$E8,$D8,$CF,$A9,$FD,$EA,$72,$D8,$FD,$F7,$DC,$7F,$A6,$27,$F8,$23
	.dc.b	$63,$3E,$D7,$F1,$0E,$F4,$E6,$7F,$31,$C1,$F5,$FE,$0F,$F0,$BF,$75
	.dc.b	$F7,$5B,$3F,$65,$AF,$0E,$34,$62,$3F,$1B,$0F,$A9,$C5,$FE,$0D,$B1
	.dc.b	$98,$5F,$E2,$E7,$3C,$BF,$96,$FC,$0F,$F1,$94,$69,$30,$FE,$25,$C5
	.dc.b	$FE,$1E,$7E,$81,$F7,$EB,$E7,$3F,$C4,$E3,$5B,$7E,$02,$73,$3F,$E1
	.dc.b	$DE,$2F,$B3,$7F,$5A,$3C,$07,$E0,$38,$BF,$7D,$19,$3E,$9D,$6A,$7F
	.dc.b	$1B,$C5,$FE,$32,$9C,$92,$3E,$9E,$5B,$7F,$C5,$9B,$FF,$B4,$27,$EC
	.dc.b	$F0,$FC,$19,$DC,$7F,$BE,$BE,$A9,$E6,$FC,$D9,$FE,$88,$7E,$72,$33
	.dc.b	$BF,$F1,$3F,$1B,$85,$6D,$B1,$9A,$F1,$C7,$83,$2E,$71,$FE,$42,$CB
	.dc.b	$FD,$C7,$39,$38,$C3,$F2,$CA,$47,$C5,$87,$96,$3A,$12,$3E,$0B,$5F
	.dc.b	$8F,$CD,$F6,$5F,$CF,$ED,$FE,$B0,$DB,$FC,$B7,$1F,$38,$FD,$BF,$57
	.dc.b	$FA,$D3,$A3,$F6,$AD,$E4,$FE,$1B,$3F,$E5,$D9,$CB,$61,$F3,$BF,$CD
	.dc.b	$FE,$87,$F8,$0B,$67,$F1,$6D,$7F,$10,$D7,$F3,$BF,$4D,$FE,$45,$E9
	.dc.b	$34,$9C,$3E,$9D,$7D,$95,$EE,$D7,$04,$7F,$88,$FE,$25,$E9,$23,$E7
	.dc.b	$7F,$C3,$1B,$7C,$28,$F8,$A7,$C3,$29,$FA,$2E,$BF,$84,$47,$E0,$3F
	.dc.b	$E6,$8C,$9F,$51,$FD,$0B,$F4,$0F,$E7,$4D,$AF,$E6,$CD,$7F,$7C,$D7
	.dc.b	$F7,$BE,$17,$C1,$26,$9F,$E7,$1B,$FE,$C8,$FB,$67,$CA,$7C,$13,$D0
	.dc.b	$7E,$CF,$9F,$EC,$11,$FA,$CF,$F9,$FB,$E4,$9F,$80,$FF,$6D,$F8,$5F
	.dc.b	$F7,$5D,$A7,$37,$FE,$E5,$8C,$56,$DF,$E7,$E6,$23,$FD,$CD,$AF,$CB
	.dc.b	$7C,$87,$98,$FE,$F4,$E5,$FE,$91,$EF,$72,$7F,$85,$35,$C9,$FC,$A8
	.dc.b	$C9,$4F,$ED,$F7,$E6,$FB,$1F,$4B,$6F,$CA,$D9,$FE,$66,$D7,$C7,$C5
	.dc.b	$F8,$A6,$F1,$7C,$D7,$D2,$7F,$33,$F9,$1F,$EF,$B8,$7C,$6D,$91,$8C
	.dc.b	$7F,$B9,$36,$B8,$42,$BF,$C0,$B1,$FE,$33,$FB,$5F,$F0,$E7,$C2,$62
	.dc.b	$3F,$C6,$71,$8A,$FF,$17,$B6,$38,$1F,$EB,$78,$FA,$3F,$29,$FE,$10
	.dc.b	$7E,$20,$71,$1F,$71,$B1,$C0,$F1,$9F,$56,$1B,$7D,$56,$E9,$F1,$5B
	.dc.b	$A7,$F0,$1F,$49,$8F,$D7,$EC,$FF,$2B,$B6,$33,$5F,$5E,$72,$FD,$93
	.dc.b	$99,$F9,$83,$E8,$94,$C9,$E5,$C3,$C8,$79,$7E,$2D,$F1,$8F,$8B,$FF
	.dc.b	$23,$7C,$0F,$FA,$E6,$73,$18,$F7,$3F,$C2,$B1,$8C,$3F,$DE,$91,$8B
	.dc.b	$E5,$18,$3F,$CA,$13,$BB,$E9,$3A,$6D,$F5,$AD,$CE,$6F,$F8,$A1,$FD
	.dc.b	$68,$91,$FC,$C9,$AE,$8C,$10,$FB,$EF,$E1,$71,$F9,$2D,$CB,$E6,$B7
	.dc.b	$3F,$F3,$33,$F8,$BA,$FA,$EF,$67,$F3,$AD,$DB,$F0,$C7,$F9,$61,$FC
	.dc.b	$33,$4F,$C3,$07,$2D,$3F,$2B,$0F,$A7,$F1,$36,$33,$18,$3F,$51,$C7
	.dc.b	$FE,$66,$F9,$EB,$26,$91,$F8,$55,$E8,$29,$EC,$3F,$7B,$D1,$FE,$85
	.dc.b	$9F,$E0,$2D,$9C,$CF,$56,$CF,$CB,$F0,$DD,$3F,$22,$7F,$C3,$ED,$AF
	.dc.b	$E1,$F9,$6E,$31,$1E,$33,$EC,$10,$C1,$FC,$53,$0F,$F4,$C7,$DA,$F9
	.dc.b	$20,$D2,$5F,$F8,$BB,$B3,$EC,$CD,$3F,$DF,$BB,$FF,$99,$A3,$31,$FC
	.dc.b	$F4,$FF,$6E,$72,$39,$9C,$D8,$6E,$9F,$81,$AF,$90,$7F,$93,$7B,$7F
	.dc.b	$00,$FF,$A2,$5A,$E7,$2F,$C0,$39,$6B,$FD,$61,$7E,$7F,$FD,$79,$B1
	.dc.b	$FD,$79,$AE,$E3,$FD,$07,$E0,$FA,$A3,$FD,$10,$DD,$3F,$D0,$6D,$E0
	.dc.b	$FF,$30,$CF,$AF,$39,$97,$AF,$4F,$81,$1F,$1A,$9E,$17,$12,$5E,$5F
	.dc.b	$03,$F0,$8F,$AE,$53,$19,$FF,$AB,$7C,$3F,$87,$1F,$E1,$48,$F8,$2D
	.dc.b	$C7,$FD,$6D,$3A,$CF,$FD,$37,$E2,$7F,$B5,$FE,$87,$FC,$43,$66,$91
	.dc.b	$FF,$36,$F8,$FE,$F2,$3E,$B7,$CF,$AF,$BB,$1F,$81,$C3,$F7,$2B,$F4
	.dc.b	$BE,$9F,$FA,$43,$CF,$FF,$68,$FE,$59,$FE,$4F,$24,$FF,$99,$1F,$F2
	.dc.b	$69,$1F,$F2,$5F,$2F,$FB,$B6,$31,$1F,$E5,$59,$FF,$6D,$FE,$0D,$FE
	.dc.b	$59,$24,$7F,$9B,$9F,$F2,$B9,$1E,$61,$C3,$CF,$83,$FD,$4A,$FF,$9D
	.dc.b	$BC,$FF,$F3,$21,$1F,$F2,$AB,$FE,$80,$23,$83,$FD,$0C,$72,$3F,$D8
	.dc.b	$3E,$91,$C1,$FE,$8F,$71,$1F,$E8,$F3,$FD,$29,$0F,$BF,$A4,$7F,$D3
	.dc.b	$FC,$DC,$E3,$FD,$4B,$32,$7F,$DD,$D3,$FE,$FE,$8C,$47,$FA,$DF,$3F
	.dc.b	$F8,$03,$FB,$EF,$FA,$BF,$13,$49,$3F,$EB,$F8,$CB,$63,$FD,$89,$18
	.dc.b	$8F,$F6,$17,$47,$47,$FD,$3B,$8F,$FA,$2C,$8E,$0F,$F3,$EB,$FE,$A8
	.dc.b	$23,$83,$E4,$BF,$EB,$12,$38,$3F,$D7,$E7,$CB,$FA,$2F,$F7,$BA,$69
	.dc.b	$1F,$F7,$FF,$33,$C3,$7B,$5C,$67,$31,$87,$91,$88,$D3,$F6,$67,$94
	.dc.b	$62,$3F,$DB,$FC,$CC,$BE,$C4,$7F,$94,$3C,$C4,$E7,$2F,$C1,$4F,$FB
	.dc.b	$A6,$30,$ED,$AF,$FB,$B7,$7A,$FF,$BB,$1B,$6B,$FE,$56,$F8,$27,$83
	.dc.b	$5F,$27,$73,$DB,$FC,$97,$D1,$78,$A7,$31,$87,$FC,$CF,$3B,$53,$BA
	.dc.b	$4E,$63,$94,$BD,$E7,$BD,$CE,$3D,$F1,$87,$99,$E1,$71,$3C,$9B,$1E
	.dc.b	$2F,$80,$E4,$FA,$9B,$87,$89,$B8,$1D,$EF,$F7,$FE,$07,$F6,$1E,$9F
	.dc.b	$BE,$31,$87,$FA,$24,$FC,$0C,$E9,$A6,$4F,$C0,$71,$3E,$0F,$51,$86
	.dc.b	$AA,$FC,$74,$61,$B1,$CC,$D4,$BE,$13,$F8,$27,$27,$96,$31,$3F,$93
	.dc.b	$3C,$33,$A7,$F2,$ED,$7F,$86,$68,$F8,$6F,$9D,$7F,$5D,$B3,$FE,$A2
	.dc.b	$6B,$C1,$AB,$BD,$EE,$BE,$1F,$FB,$14,$61,$B1,$99,$FE,$9F,$39,$7F
	.dc.b	$A8,$8F,$88,$FE,$EB,$E2,$7E,$19,$B1,$F9,$16,$BF,$A4,$6B,$FA,$BF
	.dc.b	$8A,$7C,$57,$E8,$1B,$A7,$E7,$43,$8E,$CF,$CB,$36,$BF,$3D,$F1,$6F
	.dc.b	$8C,$FE,$CF,$08,$C5,$FA,$1B,$97,$E7,$1B,$3E,$78,$FE,$EB,$97,$9A
	.dc.b	$3D,$D1,$83,$FB,$CC,$F0,$8F,$DF,$B5,$FE,$84,$D7,$C5,$7F,$DE,$BE
	.dc.b	$0D,$F1,$CF,$8E,$FE,$19,$F2,$1F,$B7,$9C,$C6,$2F,$F8,$5C,$63,$35
	.dc.b	$6E,$F9,$46,$2B,$D4,$7D,$6C,$FD,$44,$CF,$FC,$41,$87,$3F,$9E,$E5
	.dc.b	$97,$F1,$D6,$E0,$69,$F4,$32,$FF,$09,$C7,$D4,$6E,$BE,$48,$FF,$0A
	.dc.b	$99,$3C,$58,$7C,$9C,$A9,$1F,$E2,$1F,$25,$FE,$62,$8E,$2D,$8C,$FB
	.dc.b	$1F,$E3,$7A,$4E,$6B,$FE,$00,$DB,$E4,$FC,$64,$AF,$1C,$65,$F0,$4F
	.dc.b	$C6,$C3,$F3,$07,$94,$FE,$45,$F2,$8F,$84,$F0,$36,$7D,$AD,$9D,$CF
	.dc.b	$F8,$F3,$E5,$9F,$2D,$FD,$1F,$E6,$FF,$94,$C6,$31,$46,$27,$58,$FF
	.dc.b	$37,$AD,$4F,$92,$36,$7D,$C7,$04,$7F,$9F,$7C,$C3,$E1,$BF,$C2,$53
	.dc.b	$FC,$E2,$31,$3C,$8F,$EA,$AE,$23,$53,$C2,$7F,$A4,$7B,$1F,$E5,$CF
	.dc.b	$9B,$FE,$82,$D8,$CF,$B3,$FE,$7E,$9C,$B6,$3F,$23,$1E,$92,$38,$3F
	.dc.b	$AA,$47,$F2,$28,$FE,$9A,$D8,$CF,$85,$FE,$2B,$9C,$B6,$3F,$AE,$7C
	.dc.b	$D3,$E3,$3F,$7E,$DD,$BF,$8A,$B7,$83,$FA,$CB,$6B,$8D,$70,$25,$70
	.dc.b	$6D,$70,$6C,$F0,$9D,$12,$74,$D8,$D4,$60,$8E,$08,$7F,$81,$AB,$C8
	.dc.b	$DA,$EE,$DA,$E8,$DA,$E6,$D3,$99,$FE,$51,$AE,$6D,$AE,$6D,$AE,$6D
	.dc.b	$AE,$6D,$39,$8A,$E6,$DA,$E6,$DA,$E6,$DA,$E6,$D3,$9D,$72,$25,$72
	.dc.b	$6D,$72,$6D,$72,$6D,$7F,$86,$FE,$29,$F1,$2F,$8A,$FF,$0A,$CA,$73
	.dc.b	$18,$6C,$7F,$86,$9A,$FE,$2B,$A7,$FC,$78,$FE,$19,$AE,$71,$EB,$87
	.dc.b	$F1,$D8,$D4,$62,$77,$1F,$6D,$AF,$D3,$6B,$F2,$DB,$3C,$E3,$B1,$D0
	.dc.b	$86,$08,$75,$9F,$01,$1D,$99,$29,$E1,$74,$F2,$30,$68,$F2,$3A,$79
	.dc.b	$1B,$32,$6E,$3D,$0E,$88,$74,$38,$3A,$70,$68,$E6,$6A,$7E,$19,$27
	.dc.b	$C5,$1C,$08,$70,$39,$4F,$57,$06,$9C,$12,$7A,$47,$A5,$AF,$F9,$97
	.dc.b	$E7,$DF,$59,$FE,$B9,$C7,$FC,$BF,$18,$9F,$C7,$1B,$9D,$34,$C9,$FE
	.dc.b	$80,$FC,$95,$F4,$BA,$8C,$35,$FF,$63,$46,$9E,$26,$23,$4F,$F8,$FF
	.dc.b	$E9,$2F,$96,$69,$23,$88,$E4,$72,$3D,$9C,$C6,$1B,$19,$D3,$F9,$CC
	.dc.b	$72,$E9,$FE,$AD,$1F,$D5,$78,$31,$D5,$FE,$A1,$8C,$1F,$E8,$4F,$A1
	.dc.b	$7D,$16,$73,$FF,$48,$46,$2F,$CA,$DA,$E1,$5B,$AF,$F2,$EF,$D9,$3E
	.dc.b	$BB,$F7,$78,$23,$13,$FE,$92,$8C,$4E,$A7,$33,$E7,$21,$FD,$AF,$1F
	.dc.b	$D8,$E5,$FD,$C6,$3F,$D3,$DF,$B8,$7F,$67,$C3,$A3,$63,$3D,$BF,$BA
	.dc.b	$94,$E6,$7F,$D7,$B1,$E5,$F0,$3C,$44,$8F,$36,$3F,$E0,$77,$C6,$7F
	.dc.b	$B1,$30,$F1,$3E,$5A,$FE,$33,$18,$8F,$F3,$FF,$D9,$7F,$85,$DF,$F5
	.dc.b	$A3,$5F,$0B,$5D,$F8,$9F,$C4,$3C,$97,$EA,$1F,$EE,$2C,$3F,$93,$D7
	.dc.b	$F9,$2A,$31,$DD,$FE,$4F,$6C,$67,$D4,$FF,$6F,$4F,$A2,$03,$2F,$33
	.dc.b	$86,$B9,$DF,$C5,$7E,$1F,$AF,$ED,$87,$FB,$A5,$B8,$9A,$7D,$4F,$57
	.dc.b	$FD,$DD,$BD,$9F,$95,$8F,$FB,$C7,$83,$FD,$31,$18,$3E,$AD,$70,$9F
	.dc.b	$98,$7A,$4F,$24,$7C,$EC,$BE,$E5,$72,$BE,$C4,$AF,$BB,$19,$6F,$77
	.dc.b	$02,$5F,$E2,$1F,$C6,$DC,$8E,$C4,$9F,$F6,$A7,$C8,$3E,$9F,$F8,$17
	.dc.b	$B2,$8C,$4F,$F0,$28,$C4,$EA,$73,$D1,$CB,$F1,$4F,$A7,$FA,$A5,$C0
	.dc.b	$D2,$3F,$60,$FF,$6C,$53,$E7,$38,$C6,$71,$F1,$46,$23,$EF,$C6,$1E
	.dc.b	$A6,$E7,$93,$83,$FD,$AB,$AB,$EA,$5A,$3F,$1C,$FD,$7C,$61,$B1,$9A
	.dc.b	$F9,$F3,$97,$E8,$F9,$55,$EA,$D7,$F0,$4D,$9F,$DA,$B6,$FF,$93,$47
	.dc.b	$FB,$56,$7E,$DF,$87,$E4,$F5,$BF,$80,$B6,$33,$3B,$7E,$47,$D7,$72
	.dc.b	$AF,$95,$1F,$A2,$8C,$38,$6B,$C2,$BA,$3F,$13,$E4,$2B,$F9,$96,$DF
	.dc.b	$F0,$C6,$E0,$77,$AF,$E1,$7E,$5B,$F0,$44,$BF,$C3,$6C,$8F,$C3,$7D
	.dc.b	$FC,$9C,$B7,$8F,$F2,$3E,$D7,$EE,$BF,$67,$B4,$69,$1F,$BD,$F4,$5B
	.dc.b	$69,$B2,$C7,$E7,$23,$51,$FC,$1B,$E0,$9F,$B4,$3B,$FF,$B0,$46,$1B
	.dc.b	$19,$9F,$EC,$33,$97,$FA,$AF,$97,$32,$FE,$AA,$48,$FC,$E7,$D3,$87
	.dc.b	$5F,$F4,$B8,$C4,$63,$8F,$FB,$6B,$B9,$D3,$63,$FB,$89,$FC,$FB,$A3
	.dc.b	$F3,$3E,$97,$E5,$89,$7F,$96,$9F,$F0,$23,$F9,$46,$BF,$94,$69,$F9
	.dc.b	$42,$F2,$BA,$7A,$CF,$F0,$EF,$9F,$F4,$57,$F4,$E6,$D1,$A7,$B7,$9B
	.dc.b	$F8,$47,$27,$CF,$BF,$DA,$35,$FD,$CF,$32,$31,$1F,$C0,$9B,$C1,$F3
	.dc.b	$CF,$CE,$96,$3F,$84,$35,$FB,$9C,$3F,$5E,$3F,$C0,$E4,$8F,$F6,$1F
	.dc.b	$0A,$31,$1F,$D2,$37,$F9,$D1,$FA,$A6,$C7,$DA,$7E,$B4,$65,$B1,$FC
	.dc.b	$3F,$E3,$39,$34,$FC,$01,$63,$F9,$6C,$6A,$3C,$11,$F8,$13,$A7,$0F
	.dc.b	$F1,$46,$ED,$CA,$77,$39,$7B,$79,$1D,$13,$48,$FA,$63,$F8,$FC,$74
	.dc.b	$24,$66,$78,$BF,$18,$FF,$17,$90,$C1,$FD,$43,$E8,$3F,$AA,$92,$3F
	.dc.b	$C2,$AF,$F8,$4E,$9F,$5F,$FC,$BB,$F2,$4F,$E2,$B5,$FA,$28,$C5,$FF
	.dc.b	$85,$AB,$72,$AD,$C6,$1B,$E4,$75,$24,$FF,$8B,$E7,$FB,$B3,$FE,$AE
	.dc.b	$F5,$3F,$C9,$57,$E9,$3F,$A7,$38,$7D,$67,$F8,$22,$3F,$6B,$F7,$7E
	.dc.b	$92,$3D,$5F,$F5,$2C,$F1,$9E,$7E,$AF,$F9,$65,$EC,$7F,$84,$A3,$4F
	.dc.b	$F4,$EB,$FF,$24,$5F,$9F,$D3,$F8,$C4,$7F,$9C,$CF,$F6,$2F,$A5,$FF
	.dc.b	$0D,$E7,$FD,$D1,$B1,$FD,$75,$AF,$EA,$9A,$FF,$5C,$F8,$3E,$EF,$D0
	.dc.b	$6B,$FE,$44,$6B,$F0,$5A,$FA,$72,$F6,$3F,$18,$C7,$1F,$CD,$8C,$C7
	.dc.b	$C9,$C9,$E6,$FE,$04,$99,$7D,$57,$FC,$AD,$F7,$8F,$86,$FF,$1B,$F0
	.dc.b	$FF,$B6,$A9,$39,$C3,$F0,$67,$D5,$71,$3B,$6C,$F9,$67,$FD,$19,$8F
	.dc.b	$91,$F3,$9F,$52,$3F,$98,$FC,$97,$FB,$A9,$FC,$83,$63,$D4,$D7,$EC
	.dc.b	$36,$3C,$B8,$7E,$1D,$FF,$36,$DF,$9D,$F2,$46,$6B,$D0,$D7,$C6,$79
	.dc.b	$F8,$3F,$0A,$FA,$34,$FC,$D7,$A9,$FE,$14,$BF,$89,$AA,$31,$1F,$A7
	.dc.b	$6D,$7D,$16,$BF,$8F,$F8,$1F,$F0,$EF,$B3,$C9,$B5,$FA,$F6,$CF,$D9
	.dc.b	$6B,$FE,$32,$F8,$AF,$F3,$5F,$07,$F8,$9A,$3F,$9B,$C6,$2B,$FC,$5C
	.dc.b	$D8,$C9,$F8,$28,$F7,$6F,$F4,$E3,$DC,$7D,$3C,$98,$7F,$9A,$AB,$C8
	.dc.b	$71,$9F,$23,$5F,$63,$5E,$67,$6C,$FF,$28,$47,$F6,$66,$8F,$DB,$BF
	.dc.b	$AE,$5F,$05,$F3,$7C,$6F,$F4,$4E,$9F,$D8,$67,$FD,$17,$39,$51,$97
	.dc.b	$8E,$BF,$EF,$08,$C4,$ED,$B1,$9F,$13,$FC,$BD,$39,$6C,$66,$BB,$B5
	.dc.b	$FC,$24,$7F,$B8,$08,$FF,$5A,$6B,$EC,$6B,$FB,$67,$F9,$4E,$5F,$E5
	.dc.b	$EE,$5F,$C1,$66,$69,$1E,$95,$0E,$85,$3F,$08,$71,$7F,$02,$D3,$F0
	.dc.b	$31,$F8,$77,$6F,$83,$E0,$64,$FD,$E6,$CF,$E6,$63,$FD,$90,$7F,$89
	.dc.b	$B0,$FF,$3A,$FB,$5F,$A2,$6C,$7E,$84,$FF,$66,$71,$38,$1B,$FD,$5C
	.dc.b	$3E,$11,$E9,$3E,$59,$FB,$8E,$3F,$F6,$3E,$3F,$B8,$8F,$0F,$27,$F1
	.dc.b	$18,$C1,$FB,$2E,$5F,$F4,$3E,$F7,$3F,$F1,$31,$FC,$28,$B7,$FC,$29
	.dc.b	$AF,$88,$F9,$64,$30,$7F,$A4,$AF,$F9,$07,$9B,$E3,$24,$F9,$E7,$FC
	.dc.b	$D4,$FF,$67,$E6,$F1,$C5,$8F,$80,$7F,$93,$A3,$31,$FC,$D5,$AF,$80
	.dc.b	$F8,$67,$8C,$A7,$F4,$9E,$AD,$CB,$A7,$FA,$CE,$3F,$D2,$E7,$31,$8A
	.dc.b	$FE,$A8,$3F,$7C,$FE,$B0,$D3,$2E,$0F,$EF,$B3,$DB,$A3,$94,$FF,$4B
	.dc.b	$8C,$1E,$87,$CA,$4E,$1F,$C0,$70,$7D,$F8,$F4,$E5,$FD,$E8,$F8,$76
	.dc.b	$EC,$EC,$48,$E1,$7C,$B3,$FD,$4D,$F5,$E2,$FC,$5C,$7F,$84,$2B,$9F
	.dc.b	$6F,$9C,$8F,$6E,$CF,$F6,$5E,$BF,$E1,$86,$E1,$FE,$1D,$6B,$FD,$9D
	.dc.b	$AF,$FB,$4B,$C8,$77,$FD,$06,$DF,$DC,$6E,$3F,$E2,$A6,$D7,$E7,$73
	.dc.b	$F5,$37,$3F,$4B,$FE,$3D,$FC,$0F,$FC,$96,$47,$07,$F9,$5D,$FF,$25
	.dc.b	$11,$C1,$FE,$45,$FC,$0F,$FC,$A6,$47,$07,$F9,$7D,$FF,$29,$11,$C1
	.dc.b	$FE,$4D,$F8,$8F,$F2,$D9,$1C,$1F,$E7,$47,$FC,$43,$7F,$71,$B7,$FB
	.dc.b	$C6,$F0,$9F,$F4,$1F,$8F,$FE,$B5,$6F,$9C,$C3,$47,$FA,$1A,$74,$D8
	.dc.b	$D4,$FF,$BA,$3C,$BF,$F4,$BC,$F0,$3F,$D2,$BE,$56,$9F,$F4,$C9,$93
	.dc.b	$FD,$2C,$5D,$FF,$D4,$83,$E2,$3E,$B7,$9A,$8C,$4F,$FA,$AD,$AF,$F9
	.dc.b	$47,$83,$FD,$55,$18,$8F,$F5,$B7,$97,$F7,$F8,$B9,$36,$4D,$3F,$09
	.dc.b	$E7,$EA,$BF,$D8,$51,$89,$DC,$60,$FF,$40,$63,$FE,$8E,$23,$83,$FD
	.dc.b	$4E,$FD,$02,$38,$3E,$51,$E7,$F3,$7F,$DC,$91,$F4,$A3,$1B,$FF,$BA
	.dc.b	$63,$FD,$E5,$0F,$D8,$E9,$1F,$F7,$CB,$FF,$00,$08,$FD,$B1,$B3,$CF
	.dc.b	$7A,$7F,$7E,$B7,$5F,$60,$F9,$55,$F7,$2F,$EF,$5F,$FB,$3E,$72,$D8
	.dc.b	$CF,$83,$F8,$96,$CE,$A3,$0F,$8E,$BC,$51,$88,$FB,$5E,$A7,$DF,$FC
	.dc.b	$CF,$F6,$CD,$8F,$56,$AE,$14,$E6,$78,$CA,$BF,$D0,$7E,$BB,$D7,$D5
	.dc.b	$7E,$76,$31,$5B,$8F,$8C,$23,$35,$CA,$B8,$D7,$85,$B8,$6A,$FC,$AD
	.dc.b	$7F,$29,$1A,$E3,$FC,$D7,$E3,$FF,$91,$24,$F0,$8C,$BC,$B9,$2B,$F9
	.dc.b	$D6,$BF,$C7,$DB,$87,$C3,$73,$AF,$DD,$F6,$FF,$75,$F9,$0F,$D8,$6D
	.dc.b	$FD,$1E,$14,$62,$3F,$80,$57,$15,$5F,$80,$F1,$75,$1F,$8A,$B4,$62
	.dc.b	$BF,$18,$D8,$CC,$F2,$3C,$1E,$F7,$DF,$79,$36,$BF,$28,$D9,$3B,$97
	.dc.b	$C1,$BE,$03,$FA,$E3,$74,$FE,$B4,$DD,$7F,$D8,$6D,$AF,$AB,$F0,$3F
	.dc.b	$D1,$49,$5F,$D1,$5F,$E9,$1F,$05,$FE,$3F,$9F,$B6,$D7,$F8,$73,$5F
	.dc.b	$EA,$0D,$7C,$B8,$79,$09,$5E,$47,$AB,$FE,$49,$8E,$AD,$7A,$B5,$EA
	.dc.b	$DF,$6D,$F0,$6F,$0B,$F9,$27,$83,$F9,$89,$CC,$62,$FF,$80,$46,$2F
	.dc.b	$F7,$11,$FC,$F6,$31,$5F,$1B,$0F,$3E,$96,$FD,$07,$F9,$0E,$B9,$7D
	.dc.b	$6F,$7C,$BC,$52,$8C,$65,$FC,$DD,$B1,$9C,$BF,$73,$F0,$AF,$53,$FC
	.dc.b	$4A,$D9,$FF,$18,$B7,$2F,$D9,$C2,$31,$5F,$B6,$DF,$EE,$3F,$C7,$5A
	.dc.b	$FF,$59,$6B,$F5,$27,$FC,$7B,$F0,$DF,$EE,$E8,$BE,$9F,$F5,$FC,$3B
	.dc.b	$CE,$5B,$19,$9F,$DE,$FB,$9F,$E0,$D2,$5F,$F8,$35,$F9,$31,$E5,$DE
	.dc.b	$BE,$57,$FC,$87,$83,$D3,$E9,$3F,$DB,$ED,$39,$8F,$3C,$E9,$FE,$C5
	.dc.b	$19,$C7,$F8,$BC,$3E,$21,$F1,$1F,$D9,$FC,$4C,$C6,$31,$FE,$C9,$5B
	.dc.b	$AD,$3F,$C8,$27,$55,$F8,$12,$47,$AE,$B8,$EB,$FA,$F9,$FE,$B7,$3F
	.dc.b	$EC,$46,$BF,$E0,$9F,$3B,$FC,$9B,$1E,$6F,$89,$FE,$65,$3E,$26,$C6
	.dc.b	$7C,$1F,$DF,$46,$27,$F9,$9F,$C5,$BE,$2F,$F8,$13,$63,$D6,$DB,$F8
	.dc.b	$6D,$8F,$E7,$7D,$6F,$86,$FE,$14,$D8,$F8,$6D,$AF,$96,$D8,$FE,$87
	.dc.b	$F1,$8F,$80,$FF,$35,$37,$6F,$F3,$43,$78,$7E,$9B,$6B,$8D,$70,$25
	.dc.b	$70,$6D,$70,$6C,$F0,$9D,$12,$74,$D8,$D4,$60,$8E,$08,$7F,$80,$6B
	.dc.b	$C8,$DA,$EE,$DA,$E8,$DA,$E6,$D3,$99,$FE,$45,$AE,$6D,$AE,$6D,$AE
	.dc.b	$6D,$AE,$6D,$39,$8A,$E6,$DA,$E6,$DA,$E6,$DA,$E6,$D3,$9D,$72,$25
	.dc.b	$72,$6D,$72,$6D,$72,$6D,$7F,$80,$BE,$0D,$F0,$2F,$84,$FD,$7C,$A7
	.dc.b	$31,$86,$C7,$ED,$1A,$FF,$35,$FA,$BF,$F7,$E3,$FC,$D1,$AF,$F3,$36
	.dc.b	$AF,$01,$18,$6E,$5E,$26,$CF,$F0,$26,$BD,$E3,$53,$C4,$8E,$09,$1C
	.dc.b	$63,$2D,$8C,$BC,$CC,$10,$C1,$0F,$63,$C1,$A6,$CA,$6C,$F8,$0E,$DA
	.dc.b	$6C,$A6,$CF,$89,$1E,$67,$13,$F0,$1E,$A4,$38,$94,$E3,$3D,$DA,$77
	.dc.b	$29,$D2,$7B,$3C,$67,$B3,$D0,$87,$03,$51,$C1,$CC,$70,$74,$62,$34
	.dc.b	$E0,$91,$B3,$86,$FF,$C7,$F0,$FE,$3B,$C1,$FE,$B5,$6D,$7E,$54,$FE
	.dc.b	$4F,$19,$7F,$CB,$7F,$16,$D7,$11,$FE,$83,$8C,$3C,$4C,$46,$9F,$F6
	.dc.b	$14,$6A,$BF,$D9,$24,$8F,$F3,$3F,$23,$D9,$7B,$3C,$E7,$FA,$14,$62
	.dc.b	$BF,$C1,$84,$8F,$E9,$8D,$7F,$D2,$DF,$74,$EE,$FD,$7C,$23,$0F,$EC
	.dc.b	$9A,$FF,$62,$6B,$E4,$AF,$1B,$FE,$E0,$21,$9B,$FD,$DF,$92,$E5,$FD
	.dc.b	$8B,$78,$73,$18,$3F,$6B,$87,$BE,$FC,$58,$79,$A3,$11,$E3,$8F,$4E
	.dc.b	$7F,$17,$1F,$E1,$2D,$9F,$19,$FE,$B5,$70,$FF,$BA,$FC,$7F,$19,$31
	.dc.b	$F3,$BF,$C1,$BE,$55,$F5,$DF,$EE,$5E,$6F,$F0,$1C,$62,$BF,$8C,$DF
	.dc.b	$FA,$72,$31,$3A,$AE,$13,$F0,$CF,$19,$F6,$31,$F3,$E5,$F2,$A3,$94
	.dc.b	$FF,$BC,$FE,$0B,$96,$26,$92,$7C,$1E,$1F,$FB,$E7,$C4,$F2,$B6,$33
	.dc.b	$5F,$E3,$37,$FD,$AD,$9F,$F9,$EE,$BC,$70,$D7,$E2,$BF,$1F,$2F,$23
	.dc.b	$73,$FE,$6B,$7C,$63,$AD,$F8,$30,$ED,$01,$3B,$7F,$DC,$DF,$7A,$F9
	.dc.b	$CF,$C8,$B6,$3F,$A0,$B5,$FF,$18,$B5,$F9,$44,$3C,$87,$AB,$8F,$FC
	.dc.b	$64,$FF,$8E,$1A,$FF,$B4,$1A,$FF,$BD,$DA,$AF,$FD,$5F,$F6,$17,$EC
	.dc.b	$FF,$CC,$BE,$8A,$31,$5F,$EE,$68,$C5,$2B,$75,$AA,$FF,$28,$53,$F0
	.dc.b	$1C,$47,$78,$C5,$F1,$8E,$93,$A8,$FF,$60,$BC,$39,$DE,$93,$E9,$3E
	.dc.b	$77,$CA,$FF,$BB,$27,$31,$8C,$3F,$DA,$4D,$7F,$DA,$35,$C5,$57,$18
	.dc.b	$FF,$60,$46,$0F,$F6,$D7,$27,$17,$FD,$A8,$DB,$F1,$B9,$C5,$F3,$3F
	.dc.b	$E6,$6A,$76,$FF,$A3,$A3,$0D,$F3,$3F,$90,$CF,$E1,$CF,$A5,$3A,$70
	.dc.b	$3E,$8F,$D1,$5F,$E7,$46,$C6,$9A,$F3,$6B,$D3,$D1,$EE,$FF,$A0,$EB
	.dc.b	$FD,$4C,$F9,$8F,$16,$1E,$A9,$DC,$71,$34,$7E,$6B,$6F,$CC,$91,$FC
	.dc.b	$CB,$FC,$B3,$8B,$F1,$B1,$FC,$BF,$2F,$F0,$8C,$60,$FC,$8D,$BE,$6F
	.dc.b	$FD,$B1,$F9,$FF,$FA,$D1,$B7,$E4,$6B,$FE,$9B,$6A,$DB,$F0,$BE,$A7
	.dc.b	$EC,$6B,$EE,$36,$7E,$74,$E9,$4E,$9A,$7D,$0E,$6F,$D8,$3F,$B8,$E7
	.dc.b	$FD,$B9,$1F,$E3,$3D,$51,$FE,$17,$D9,$FC,$29,$FE,$05,$A7,$BF,$5F
	.dc.b	$82,$4D,$7F,$D5,$A3,$F3,$AF,$F1,$8F,$DF,$7F,$B5,$1B,$7F,$51,$AF
	.dc.b	$FA,$ED,$AF,$EE,$57,$D6,$7E,$EE,$79,$36,$79,$35,$E4,$D7,$F8,$C6
	.dc.b	$9F,$B7,$7E,$D9,$0F,$E2,$7C,$3F,$77,$1F,$54,$7F,$0F,$6B,$F2,$75
	.dc.b	$F8,$CF,$F1,$A6,$BF,$C6,$4F,$E8,$47,$C8,$EF,$FE,$5E,$FC,$89,$E6
	.dc.b	$FF,$17,$6B,$E6,$AF,$C8,$3E,$43,$CC,$7F,$1F,$8D,$BF,$91,$CF,$FC
	.dc.b	$DB,$A7,$E5,$31,$F7,$4F,$12,$1F,$81,$BF,$75,$70,$BE,$77,$F9,$CD
	.dc.b	$7E,$F5,$FF,$49,$C7,$D2,$DD,$BF,$08,$FD,$9F,$77,$FC,$E7,$8F,$DF
	.dc.b	$6E,$E6,$98,$F8,$4F,$E6,$9F,$33,$F8,$36,$DF,$E0,$DA,$FE,$0D,$AF
	.dc.b	$E0,$FD,$DE,$A4,$C3,$F1,$F8,$23,$FD,$07,$9F,$A8,$99,$FA,$9F,$CD
	.dc.b	$57,$E5,$7E,$1F,$CE,$47,$9B,$CA,$78,$13,$4F,$83,$8F,$E5,$FE,$6F
	.dc.b	$DE,$4B,$F1,$BF,$0B,$E2,$FF,$A9,$7C,$DF,$E8,$5B,$5F,$A1,$6B,$FA
	.dc.b	$16,$BF,$D6,$7E,$A9,$1F,$39,$B7,$FA,$E6,$CF,$F2,$C8,$46,$23,$F2
	.dc.b	$87,$D9,$3F,$29,$87,$C8,$F9,$AF,$AE,$4A,$C1,$FB,$AE,$E7,$CA,$7F
	.dc.b	$2A,$BF,$E5,$CD,$7F,$9E,$B6,$7F,$A3,$B9,$8F,$01,$E2,$DB,$E9,$E1
	.dc.b	$F6,$3C,$5F,$EE,$D3,$0F,$E0,$0E,$2F,$F8,$33,$63,$F8,$33,$5F,$09
	.dc.b	$FE,$15,$8F,$E7,$F8,$7A,$23,$07,$A0,$A7,$03,$FD,$9B,$7E,$47,$E2
	.dc.b	$9F,$C5,$FB,$BB,$57,$F1,$B6,$F1,$1A,$6B,$F6,$77,$C4,$7F,$84,$7A
	.dc.b	$3D,$96,$9C,$BF,$E7,$73,$FA,$AB,$88,$DB,$63,$33,$FE,$43,$68,$F3
	.dc.b	$37,$19,$C7,$E1,$57,$69,$F6,$F2,$7E,$66,$7D,$A7,$28,$E3,$7C,$CC
	.dc.b	$12,$3F,$C6,$67,$D9,$9E,$4D,$DB,$F4,$6F,$F2,$03,$FB,$EF,$B3,$FE
	.dc.b	$FC,$CB,$FD,$4F,$1F,$F0,$07,$D0,$FF,$58,$46,$27,$5E,$AF,$F8,$1D
	.dc.b	$B5,$FE,$4F,$6B,$C7,$D5,$FE,$A3,$3B,$BE,$6D,$8C,$CF,$F9,$CE,$3F
	.dc.b	$C1,$EF,$F1,$96,$F0,$7F,$A3,$5B,$97,$FA,$C9,$B9,$7A,$2B,$91,$FE
	.dc.b	$A1,$8C,$9F,$AF,$D7,$ED,$F2,$7C,$A2,$6B,$F5,$9F,$A3,$7C,$EF,$C4
	.dc.b	$73,$21,$F6,$37,$FF,$2C,$4F,$28,$C5,$23,$11,$FE,$5D,$BF,$C5,$FB
	.dc.b	$7F,$68,$95,$F8,$67,$F2,$3C,$5F,$97,$8F,$F6,$E1,$FD,$C7,$8B,$FC
	.dc.b	$CD,$9E,$9B,$19,$37,$87,$FA,$1E,$7E,$0B,$E8,$3E,$29,$FA,$D7,$F3
	.dc.b	$FC,$7F,$B1,$6F,$2F,$FA,$25,$B3,$D0,$FD,$FD,$53,$FC,$11,$CB,$E2
	.dc.b	$8F,$DF,$C6,$71,$FF,$33,$3C,$63,$E5,$9E,$53,$C8,$60,$FE,$0B,$5E
	.dc.b	$02,$47,$F0,$6B,$F2,$3F,$E1,$3D,$3E,$1D,$58,$FC,$11,$FE,$1B,$8C
	.dc.b	$C7,$F8,$5D,$AF,$C2,$21,$E1,$3C,$51,$FC,$D3,$9F,$6D,$3A,$F3,$FF
	.dc.b	$3C,$C3,$F9,$D3,$6B,$FA,$34,$BE,$4F,$FA,$53,$63,$31,$83,$FC,$5D
	.dc.b	$D5,$DB,$33,$49,$5F,$93,$E4,$FD,$15,$FC,$86,$BF,$2C,$FF,$1A,$75
	.dc.b	$F3,$24,$FC,$4B,$47,$C2,$9F,$81,$5C,$9F,$86,$DA,$FF,$22,$76,$79
	.dc.b	$48,$F3,$7D,$1D,$BF,$EB,$9F,$96,$FF,$03,$49,$A4,$BF,$F0,$94,$78
	.dc.b	$FC,$0F,$80,$47,$DD,$D3,$FB,$D9,$F7,$35,$F7,$1F,$E3,$58,$FD,$B7
	.dc.b	$6B,$C1,$F6,$36,$33,$1E,$FF,$0F,$FC,$99,$C5,$FE,$37,$25,$FF,$8B
	.dc.b	$CD,$D4,$8F,$F2,$63,$63,$51,$C1,$FF,$49,$1F,$C1,$E7,$AB,$FE,$41
	.dc.b	$9E,$0F,$F9,$25,$AF,$CE,$C3,$D8,$DE,$CE,$5E,$27,$EC,$63,$FC,$B4
	.dc.b	$60,$98,$79,$08,$F7,$6B,$E4,$7C,$CF,$F9,$5C,$F3,$78,$9D,$5F,$F3
	.dc.b	$8E,$0F,$A9,$FF,$3B,$B6,$3F,$CE,$ED,$77,$1F,$E8,$5F,$77,$FD,$F0
	.dc.b	$D9,$D4,$FF,$A0,$E7,$2F,$FA,$25,$AF,$C3,$BF,$3B,$F6,$DA,$FE,$53
	.dc.b	$2F,$8A,$DC,$BE,$0B,$F1,$27,$FC,$F8,$E0,$FF,$76,$F6,$97,$E7,$BF
	.dc.b	$EA,$D2,$3F,$EB,$2F,$47,$FC,$08,$E8,$FF,$5F,$76,$71,$9F,$F6,$13
	.dc.b	$63,$26,$CF,$F6,$13,$97,$D8,$7D,$32,$18,$3E,$A1,$F8,$A2,$18,$3F
	.dc.b	$DC,$9E,$77,$A7,$E3,$FF,$DF,$0D,$8F,$F1,$3B,$5D,$88,$FF,$7A,$78
	.dc.b	$19,$8F,$85,$E8,$23,$11,$FE,$FB,$6C,$66,$39,$1F,$E3,$4E,$4E,$5C
	.dc.b	$0F,$0B,$F5,$D3,$A3,$FD,$F5,$3A,$8C,$E1,$F4,$67,$E9,$79,$ED,$5E
	.dc.b	$97,$D5,$F8,$ED,$57,$FB,$A6,$72,$A3,$35,$C6,$B8,$4F,$D9,$6B,$DA
	.dc.b	$BD,$55,$C1,$B1,$F7,$F4,$DB,$DD,$B9,$BD,$57,$3A,$1F,$6B,$FE,$77
	.dc.b	$9C,$C6,$2F,$DC,$F1,$AE,$15,$B6,$CE,$A9,$EB,$E2,$3F,$CC,$51,$8B
	.dc.b	$E9,$18,$9D,$3E,$39,$D5,$FE,$9B,$B3,$F8,$F7,$35,$E1,$FE,$79,$E0
	.dc.b	$FF,$3B,$C3,$F5,$13,$FE,$02,$7E,$A7,$6B,$B7,$F8,$14,$7F,$BC,$1A
	.dc.b	$A3,$4F,$EB,$4F,$D6,$7B,$3F,$6B,$5F,$F0,$F4,$7D,$A8,$C7,$47,$F8
	.dc.b	$E9,$B1,$9E,$3F,$F1,$3B,$FC,$F3,$AF,$F9,$F9,$32,$E1,$3A,$9F,$C5
	.dc.b	$57,$26,$C6,$5F,$1B,$54,$3D,$BF,$B5,$E1,$F8,$DB,$3E,$76,$BD,$9A
	.dc.b	$FF,$33,$BF,$90,$FF,$81,$AB,$A6,$5F,$C2,$23,$83,$5F,$8B,$5F,$89
	.dc.b	$ED,$F0,$4E,$65,$D0,$F7,$3F,$B9,$36,$7F,$CA,$0D,$C7,$FC,$C6,$DA
	.dc.b	$F9,$11,$E4,$7F,$C3,$1E,$EB,$C0,$FB,$2D,$8F,$F2,$D3,$67,$FC,$DA
	.dc.b	$D8,$FD,$6F,$BD,$E2,$26,$1E,$27,$F1,$77,$FC,$CB,$8F,$F1,$84,$7F
	.dc.b	$9A,$DB,$97,$C2,$4C,$3C,$2F,$E3,$3E,$1F,$F1,$1F,$09,$FE,$7E,$6C
	.dc.b	$FF,$02,$6B,$FC,$25,$AF,$F3,$BC,$3D,$DC,$6A,$FF,$40,$92,$7F,$85
	.dc.b	$CE,$A3,$FD,$17,$EE,$B5,$FE,$89,$18,$6D,$7F,$12,$9D,$3D,$4C,$46
	.dc.b	$9F,$E3,$9F,$04,$F8,$2C,$47,$F4,$F8,$C3,$C4,$DC,$69,$46,$AF,$FD
	.dc.b	$3F,$F0,$6F,$23,$FC,$68,$DE,$1F,$F4,$93,$71,$DC,$E8,$93,$A6,$C6
	.dc.b	$A3,$04,$70,$4C,$3F,$98,$D7,$56,$D7,$26,$D7,$16,$D7,$17,$F9,$F5
	.dc.b	$71,$6D,$71,$6D,$71,$6D,$71,$7F,$BB,$57,$16,$D7,$16,$D7,$16,$D7
	.dc.b	$15,$5C,$5B,$5C,$5B,$5C,$5B,$5C,$6B,$81,$2B,$83,$6B,$83,$67,$84
	.dc.b	$E8,$93,$A6,$C6,$A3,$04,$70,$7F,$B7,$BD,$97,$B2,$F6,$7F,$DC,$38
	.dc.b	$FF,$BA,$4D,$46,$5B,$1A,$C7,$FD,$9D,$85,$EC,$FE,$B1,$5F,$8C,$6B
	.dc.b	$F7,$9A,$FD,$86,$BF,$49,$AF,$CC,$6C,$F6,$8E,$93,$DC,$8E,$08,$77
	.dc.b	$72,$D3,$27,$85,$CB,$4C,$9E,$37,$2D,$32,$79,$5C,$B4,$CC,$79,$DF
	.dc.b	$01,$0E,$C5,$3A,$14,$E4,$57,$93,$C0,$87,$02,$9C,$0A,$F0,$74,$43
	.dc.b	$45,$74,$E0,$8E,$09,$3F,$48,$93,$F0,$23,$2D,$8C,$CF,$C2,$6B,$EB
	.dc.b	$6B,$E6,$7E,$FF,$23,$E8,$FF,$CD,$18,$7F,$8D,$2B,$FA,$64,$61,$FF
	.dc.b	$1B,$CE,$DF,$CA,$1F,$E3,$F7,$13,$FC,$B6,$32,$7F,$9C,$F8,$3F,$96
	.dc.b	$C6,$49,$1F,$CE,$25,$C5,$C2,$1C,$5F,$D0,$2B,$C8,$D8,$E1,$3D,$B2
	.dc.b	$FF,$2B,$7C,$5F,$EA,$9B,$3E,$B6,$BF,$E5,$06,$BF,$CF,$8F,$D8,$E3
	.dc.b	$FD,$09,$F5,$10,$FF,$1F,$FC,$0F,$F9,$16,$7A,$36,$3A,$35,$DC,$7E
	.dc.b	$B8,$FF,$35,$7B,$BE,$5F,$30,$D2,$4F,$8B,$8B,$F8,$1A,$C3,$C8,$F9
	.dc.b	$CF,$21,$FE,$9F,$B7,$CE,$3C,$FF,$03,$63,$E1,$36,$7E,$13,$5F,$E7
	.dc.b	$7F,$1E,$FA,$BF,$F3,$17,$2F,$F4,$68,$C5,$FF,$18,$AD,$D6,$9B,$19
	.dc.b	$7F,$D0,$58,$FA,$1B,$97,$B2,$31,$1F,$85,$8F,$E3,$97,$DB,$E2,$3C
	.dc.b	$18,$7F,$9C,$DB,$81,$A7,$F5,$AF,$C4,$71,$1F,$33,$0F,$EA,$B1,$8D
	.dc.b	$75,$5F,$E7,$D6,$D6,$EB,$CA,$7D,$12,$1F,$38,$FE,$9F,$5C,$DE,$B5
	.dc.b	$CD,$B3,$CD,$DC,$65,$FA,$DB,$FF,$89,$E5,$F7,$DF,$88,$EF,$FF,$48
	.dc.b	$36,$33,$5F,$90,$9C,$BF,$81,$21,$E8,$3C,$DF,$16,$FB,$67,$E3,$FF
	.dc.b	$D3,$23,$0D,$8C,$CF,$EF,$E7,$2F,$F9,$43,$D4,$FF,$35,$2F,$8D,$FF
	.dc.b	$56,$79,$5F,$EE,$D6,$C6,$4F,$F7,$64,$65,$FF,$52,$FC,$1B,$E8,$71
	.dc.b	$1F,$97,$8C,$3C,$4C,$46,$9F,$F5,$94,$6A,$BF,$D8,$7F,$2D,$FD,$7B
	.dc.b	$F2,$0B,$B7,$FB,$FD,$B6,$69,$FB,$4F,$C1,$BF,$D7,$8D,$9D,$4F,$EF
	.dc.b	$27,$2D,$8C,$BF,$AD,$F9,$67,$C5,$7F,$BB,$23,$FD,$83,$18,$5E,$FF
	.dc.b	$F0,$D8,$C3,$67,$84,$FF,$B9,$7C,$2F,$F2,$DC,$FF,$1C,$F1,$3F,$90
	.dc.b	$B7,$C3,$E7,$5F,$C7,$A3,$F8,$6F,$C6,$7F,$B5,$5B,$96,$A7,$F1,$B3
	.dc.b	$97,$F8,$BB,$5F,$C8,$F7,$7E,$29,$B3,$F8,$E6,$F4,$7F,$1A,$9D,$47
	.dc.b	$FB,$8F,$A5,$F2,$B8,$B4,$61,$E5,$3B,$9D,$4E,$6F,$EC,$FE,$86,$A7
	.dc.b	$C4,$E1,$A6,$4F,$F5,$3F,$1A,$78,$4E,$CF,$C1,$F6,$62,$B8,$7A,$8B
	.dc.b	$FF,$4A,$47,$43,$C1,$19,$9E,$11,$89,$DD,$78,$61,$F1,$FF,$E4,$EF
	.dc.b	$A7,$F4,$65,$98,$C4,$FC,$06,$C6,$7C,$E5,$70,$AF,$EA,$F3,$F7,$48
	.dc.b	$E0,$F8,$14,$FC,$5F,$E0,$6F,$D5,$B5,$E1,$6D,$70,$6B,$EB,$9F,$9D
	.dc.b	$F9,$E7,$E9,$09,$1F,$1B,$3F,$F2,$3C,$7E,$8C,$91,$FA,$16,$BF,$86
	.dc.b	$FC,$0F,$F0,$94,$69,$23,$EC,$43,$EA,$5E,$E7,$C1,$9F,$EA,$D1,$87
	.dc.b	$F0,$F3,$B9,$D4,$23,$0F,$CD,$D7,$E5,$BE,$13,$E3,$B8,$BF,$A1,$AF
	.dc.b	$E3,$5B,$3B,$3F,$61,$D1,$FE,$0E,$8F,$D7,$12,$3F,$C1,$6F,$CD,$E2
	.dc.b	$F1,$37,$0E,$53,$F8,$4D,$7E,$AC,$E4,$FE,$2B,$7F,$9D,$FB,$4B,$46
	.dc.b	$91,$FB,$18,$7B,$E3,$B1,$23,$38,$71,$7E,$C5,$7E,$A7,$F0,$3E,$A4
	.dc.b	$8C,$87,$D0,$47,$9B,$5F,$87,$D5,$FE,$90,$EB,$FF,$76,$C7,$F1,$12
	.dc.b	$1F,$87,$F7,$BF,$D2,$BC,$FF,$AD,$6F,$19,$A6,$3D,$EF,$DF,$7C,$BF
	.dc.b	$F3,$86,$DF,$E4,$5A,$FE,$C5,$AF,$F0,$FF,$27,$FC,$51,$D7,$FD,$6F
	.dc.b	$6F,$F1,$7C,$60,$FF,$4B,$E7,$E8,$AF,$F2,$19,$E8,$3D,$A7,$F4,$8F
	.dc.b	$C2,$3C,$E4,$7F,$CB,$DA,$7F,$07,$8F,$F1,$7C,$F3,$9E,$27,$42,$1F
	.dc.b	$C1,$FC,$3F,$82,$47,$9E,$A8,$F8,$BF,$81,$FF,$40,$9F,$E3,$6D,$C3
	.dc.b	$F8,$B3,$68,$EE,$7F,$CC,$3C,$3F,$E6,$47,$D8,$7F,$29,$C3,$CC,$D8
	.dc.b	$D3,$E6,$31,$C1,$F7,$63,$31,$EC,$3F,$90,$78,$3F,$68,$8F,$A9,$FF
	.dc.b	$35,$7C,$FB,$E4,$BF,$9C,$7D,$7F,$FB,$33,$81,$39,$B4,$E6,$30,$DA
	.dc.b	$E3,$5E,$5A,$F6,$E3,$E5,$23,$FD,$FB,$C4,$FE,$16,$FF,$35,$6B,$FD
	.dc.b	$11,$BB,$7F,$3F,$73,$5E,$4B,$FE,$9C,$78,$6B,$D5,$1F,$C5,$B1,$F4
	.dc.b	$4E,$4F,$F7,$2E,$9F,$32,$BC,$A7,$D0,$29,$FC,$7B,$EC,$5F,$75,$FC
	.dc.b	$5E,$BE,$24,$63,$0E,$F5,$FE,$75,$8C,$4E,$AB,$85,$7F,$58,$F3,$DF
	.dc.b	$65,$FD,$97,$C7,$FF,$7C,$EA,$9C,$E5,$FE,$49,$3F,$AD,$B8,$9D,$B6
	.dc.b	$FF,$99,$FC,$7F,$D4,$7F,$D1,$E6,$38,$3F,$CC,$4D,$33,$7F,$E8,$B9
	.dc.b	$E1,$EB,$7C,$4E,$3F,$C4,$CF,$E3,$0F,$89,$BF,$E4,$B3,$F3,$69,$FD
	.dc.b	$F6,$FC,$6D,$C7,$DC,$DB,$F4,$B5,$E8,$7E,$16,$FE,$91,$30,$F0,$C7
	.dc.b	$D2,$3F,$34,$70,$3F,$C1,$1F,$04,$F6,$DF,$01,$F2,$F9,$33,$18,$A7
	.dc.b	$07,$FA,$FE,$F8,$5E,$E7,$FC,$2D,$18,$77,$7C,$A7,$50,$9D,$57,$3B
	.dc.b	$EC,$DA,$DE,$2C,$7F,$DB,$F8,$FF,$8F,$C9,$8F,$FB,$71,$AF,$F3,$66
	.dc.b	$BF,$BC,$7F,$C2,$DA,$FD,$08,$7C,$17,$D0,$6E,$BF,$41,$A7,$D0,$29
	.dc.b	$CC,$F0,$10,$F0,$1F,$85,$8F,$E0,$1B,$FE,$2D,$BC,$1F,$DD,$DB,$5D
	.dc.b	$2F,$C9,$E8,$78,$C9,$1E,$37,$F4,$4B,$EC,$69,$1F,$D7,$72,$7F,$80
	.dc.b	$E3,$F2,$5D,$1F,$EC,$BD,$BE,$53,$76,$F9,$7A,$FA,$4F,$F0,$57,$47
	.dc.b	$F1,$5B,$FE,$04,$4C,$7F,$81,$73,$FF,$94,$E3,$F5,$FD,$6E,$1F,$F2
	.dc.b	$B1,$82,$D7,$F8,$A5,$B8,$7F,$15,$38,$47,$85,$FE,$3B,$1F,$61,$A7
	.dc.b	$80,$7A,$FF,$C5,$49,$39,$7C,$7D,$7E,$29,$FF,$1C,$35,$FE,$72,$79
	.dc.b	$CC,$9F,$0F,$49,$7E,$1D,$7E,$3B,$75,$F8,$F5,$F7,$88,$7F,$56,$1F
	.dc.b	$9C,$FF,$AC,$3B,$7F,$CF,$0D,$8C,$9F,$E4,$B8,$CB,$FE,$4D,$EE,$F3
	.dc.b	$3F,$E7,$A6,$F1,$FF,$9D,$5B,$B7,$F6,$79,$F3,$9F,$27,$2F,$3B,$F6
	.dc.b	$EB,$FB,$B3,$EF,$3F,$CF,$1D,$FC,$88,$F2,$3D,$67,$D0,$3D,$BB,$FF
	.dc.b	$A6,$3C,$0F,$F2,$76,$BF,$E7,$F8,$C1,$D6,$7F,$C2,$4D,$33,$1F,$E1
	.dc.b	$B8,$7C,$CD,$23,$FE,$1D,$32,$7F,$8A,$4E,$98,$7F,$8A,$5B,$97,$F8
	.dc.b	$A5,$BE,$2F,$C8,$7D,$07,$ED,$4F,$5E,$7F,$11,$FF,$55,$7A,$3F,$E8
	.dc.b	$7F,$27,$FC,$D1,$F8,$17,$F9,$5C,$93,$FE,$4A,$7F,$CA,$64,$7F,$C8
	.dc.b	$ED,$79,$F9,$9F,$E9,$0F,$C0,$BF,$CC,$04,$8F,$F2,$9B,$FE,$5C,$23
	.dc.b	$FE,$50,$6B,$CA,$BF,$CF,$5F,$81,$7F,$9A,$49,$1F,$E5,$C7,$FD,$BC
	.dc.b	$47,$FD,$61,$8F,$EF,$23,$11,$EE,$F4,$3F,$D0,$31,$FE,$7C,$24,$7F
	.dc.b	$A0,$BD,$0F,$F6,$64,$7D,$BF,$4B,$FD,$1D,$F7,$77,$D3,$FF,$68,$36
	.dc.b	$CD,$3F,$D3,$B5,$23,$FD,$37,$E7,$7E,$F2,$31,$1F,$ED,$9E,$5F,$F7
	.dc.b	$0F,$47,$FA,$5F,$1F,$C5,$5F,$F9,$06,$7F,$D4,$55,$CA,$B4,$D8,$CB
	.dc.b	$FE,$B7,$29,$FE,$E0,$7F,$DC,$D5,$FC,$FB,$8F,$FD,$49,$3A,$E4,$FF
	.dc.b	$6F,$56,$A3,$FD,$0F,$97,$FA,$5C,$8E,$0F,$9B,$D6,$EC,$F8,$31,$86
	.dc.b	$BF,$E8,$49,$FF,$71,$3F,$E3,$67,$FD,$56,$47,$E8,$1C,$0F,$F7,$3E
	.dc.b	$FF,$EB,$68,$FF,$7A,$C3,$F3,$AA,$47,$FD,$A5,$E4,$9D,$1F,$ED,$5E
	.dc.b	$3C,$4F,$C1,$6C,$64,$D9,$FE,$F9,$72,$FD,$5D,$DE,$C7,$FB,$2F,$0F
	.dc.b	$AA,$D9,$FB,$AD,$78,$35,$FF,$69,$4F,$F3,$C9,$FF,$70,$92,$3F,$0B
	.dc.b	$D7,$BD,$54,$F0,$3F,$DB,$5E,$07,$E1,$1B,$7F,$8C,$6B,$E7,$6B,$FE
	.dc.b	$78,$8F,$A3,$CF,$FD,$79,$BC,$1F,$65,$BC,$1F,$CC,$AB,$F2,$9E,$D3
	.dc.b	$C7,$FE,$67,$18,$6B,$FE,$ED,$8D,$3C,$4C,$46,$B3,$FC,$E8,$FC,$7B
	.dc.b	$F3,$BC,$FF,$EF,$56,$C6,$6B,$F3,$91,$F9,$78,$FB,$8D,$8D,$1F,$9F
	.dc.b	$70,$7F,$31,$F7,$9E,$FF,$E0,$FE,$63,$0D,$AF,$C3,$0B,$57,$08,$57
	.dc.b	$06,$9F,$E6,$DF,$1F,$93,$FE,$68,$6A,$E9,$FF,$37,$46,$1B,$1C,$8F
	.dc.b	$CB,$7C,$0B,$C6,$7C,$17,$F5,$E8,$C3,$63,$33,$FD,$2E,$72,$FF,$4D
	.dc.b	$1F,$0F,$FD,$E7,$9F,$F5,$ED,$8F,$79,$A8,$FE,$89,$19,$8F,$CF,$76
	.dc.b	$7E,$47,$A3,$FC,$B7,$B7,$EC,$CF,$C9,$C6,$4F,$CE,$76,$BE,$15,$F0
	.dc.b	$BF,$E3,$9C,$3E,$6C,$E6,$31,$1F,$B5,$8C,$5F,$F9,$3A,$37,$5C,$A3
	.dc.b	$11,$F5,$B1,$F3,$3F,$BD,$6B,$EB,$3E,$13,$AA,$EB,$E8,$FC,$06,$CF
	.dc.b	$6C,$3C,$6A,$32,$F9,$7E,$1D,$F0,$FF,$85,$F9,$4E,$51,$8C,$91,$89
	.dc.b	$D5,$7C,$0A,$D4,$FD,$2D,$BE,$79,$27,$CE,$FF,$87,$30,$57,$5F,$1F
	.dc.b	$A6,$26,$92,$FF,$33,$0F,$89,$7C,$57,$F8,$4A,$73,$18,$9F,$F0,$83
	.dc.b	$71,$FF,$0B,$36,$7F,$0B,$B7,$CC,$24,$F9,$CF,$F0,$E7,$C5,$BD,$EF
	.dc.b	$F1,$9E,$08,$C5,$FF,$87,$5B,$5C,$2B,$7C,$7F,$AF,$BF,$F1,$13,$67
	.dc.b	$FA,$FC,$7A,$6F,$F9,$83,$5F,$C0,$47,$8D,$F3,$B5,$FE,$D5,$6F,$8C
	.dc.b	$F2,$B6,$3C,$91,$83,$FC,$49,$E2,$7F,$A0,$3E,$3F,$F9,$09,$27,$FA
	.dc.b	$D3,$F6,$53,$48,$F8,$8C,$9F,$AE,$F1,$5F,$1D,$FD,$4E,$7F,$B0,$C6
	.dc.b	$27,$67,$F1,$27,$11,$A9,$E1,$3F,$99,$E6,$F8,$4F,$E5,$DB,$C7,$FE
	.dc.b	$B4,$CF,$FB,$6B,$67,$86,$9F,$C1,$63,$11,$FB,$4A,$E2,$49,$D3,$6B
	.dc.b	$9B,$67,$9C,$62,$76,$47,$04,$8D,$90,$F0,$46,$DC,$1F,$C0,$E7,$B3
	.dc.b	$A9,$E4,$D3,$67,$C1,$9E,$4E,$48,$73,$72,$D8,$E6,$D3,$9D,$7C,$47
	.dc.b	$07,$23,$B4,$70,$69,$C2,$BC,$EF,$02,$1A,$39,$4F,$57,$06,$9C,$12
	.dc.b	$39,$C6,$89,$19,$AF,$5C,$62,$77,$39,$AF,$33,$DF,$E0,$1F,$00,$F8
	.dc.b	$0F,$F1,$8E,$09,$CC,$61,$B1,$FE,$36,$6B,$FD,$23,$F1,$07,$DE,$FF
	.dc.b	$92,$23,$0D,$7F,$B5,$C6,$B3,$EE,$D8,$FF,$20,$B5,$E3,$1A,$9E,$51
	.dc.b	$FE,$39,$9D,$3D,$23,$2A,$32,$EE,$33,$18,$8D,$46,$09,$3D,$C9,$1E
	.dc.b	$48,$CB,$63,$2F,$98,$C1,$0C,$10,$F3,$3C,$1A,$6C,$A6,$CF,$43,$B6
	.dc.b	$9B,$29,$B3,$D6,$ED,$A6,$CA,$6E,$7E,$11,$27,$DB,$19,$6B,$E4,$21
	.dc.b	$E0,$29,$D0,$B1,$CE,$7C,$4D,$8F,$13,$5E,$D1,$A9,$EF,$1D,$48,$75
	.dc.b	$2C,$F1,$7B,$C7,$07,$44,$8D,$3C,$B7,$6F,$FE,$93,$C7,$FC,$B6,$49
	.dc.b	$FF,$4A,$3F,$E9,$2F,$94,$7C,$BF,$F9,$06,$30,$D9,$FF,$13,$CE,$B2
	.dc.b	$FD,$54,$62,$74,$FF,$92,$BE,$B9,$F3,$78,$8F,$F4,$94,$62,$B8,$C6
	.dc.b	$E7,$4A,$75,$87,$F4,$FF,$B7,$93,$FD,$35,$AF,$F9,$63,$1F,$24,$7A
	.dc.b	$30,$34,$93,$FE,$40,$F2,$9F,$17,$FE,$5F,$9F,$F5,$9C,$62,$B8,$D6
	.dc.b	$EB,$4D,$8C,$BF,$E6,$8F,$7D,$F3,$9F,$C3,$21,$18,$7F,$87,$B5,$FF
	.dc.b	$60,$B5,$F9,$53,$FC,$53,$1F,$9D,$1E,$B2,$1F,$E5,$8F,$11,$F0,$7F
	.dc.b	$C0,$9B,$5F,$C7,$1B,$3F,$EC,$E6,$BF,$E2,$6F,$53,$CA,$FF,$B4,$A5
	.dc.b	$18,$8F,$F3,$2C,$63,$0F,$E4,$ED,$8F,$37,$07,$F1,$2C,$7C,$F1,$EA
	.dc.b	$3C,$E7,$FB,$EF,$E3,$3F,$C6,$0F,$46,$BF,$E3,$86,$BB,$F0,$FF,$90
	.dc.b	$FC,$13,$EB,$BF,$11,$3F,$E7,$88,$C7,$26,$A7,$33,$FE,$9A,$9C,$E5
	.dc.b	$E6,$24,$79,$0F,$F7,$26,$3E,$87,$FD,$CD,$E7,$BD,$DF,$F6,$EC,$7E
	.dc.b	$58,$FF,$4D,$38,$E5,$FF,$44,$CE,$FD,$BF,$F7,$74,$62,$75,$5F,$EA
	.dc.b	$CA,$E9,$F6,$BF,$EF,$3F,$BC,$D3,$67,$54,$8C,$47,$92,$30,$A7,$B7
	.dc.b	$CE,$6A,$3C,$AD,$BF,$53,$68,$EF,$37,$03,$EF,$FF,$D8,$1B,$FD,$56
	.dc.b	$E3,$F5,$5A,$FD,$B6,$BF,$EC,$79,$7D,$0E,$2F,$AB,$63,$31,$F6,$E7
	.dc.b	$F0,$13,$9F,$A2,$FF,$68,$79,$BF,$85,$6C,$FE,$15,$AF,$E2,$DA,$87
	.dc.b	$DB,$B8,$73,$18,$6C,$67,$0F,$F6,$2C,$FF,$B0,$A3,$17,$F9,$5F,$27
	.dc.b	$AD,$7C,$66,$E5,$FE,$5A,$8E,$37,$F9,$D9,$DF,$E7,$4F,$88,$FC,$1C
	.dc.b	$FF,$A5,$23,$0F,$E2,$E9,$18,$E4,$FD,$7B,$67,$F0,$7E,$D7,$27,$F0
	.dc.b	$4D,$7F,$26,$DE,$3F,$CC,$39,$BF,$BB,$EB,$7F,$BE,$FD,$FF,$D8,$36
	.dc.b	$FF,$60,$D7,$F6,$CD,$7E,$E7,$DB,$5A,$FC,$3B,$75,$34,$E1,$D4,$F8
	.dc.b	$BF,$CB,$EB,$FB,$76,$E1,$FB,$56,$BF,$66,$78,$75,$7B,$61,$E4,$FF
	.dc.b	$5A,$F8,$4F,$D9,$46,$3C,$AF,$EB,$8F,$F5,$A2,$1C,$FC,$4F,$F5,$66
	.dc.b	$5F,$C1,$DB,$1F,$C1,$5A,$EF,$5F,$EA,$C7,$E9,$3A,$8B,$ED,$7F,$48
	.dc.b	$D9,$FD,$23,$5F,$D2,$35,$FD,$27,$5F,$29,$FE,$40,$D8,$FD,$0C,$EC
	.dc.b	$FD,$39,$7C,$FF,$CE,$3F,$87,$6E,$3F,$CD,$1B,$3F,$EE,$06,$BF,$86
	.dc.b	$3F,$AA,$9F,$86,$2F,$2F,$E1,$9B,$DE,$E6,$FD,$57,$97,$F5,$F8,$6D
	.dc.b	$7D,$76,$BF,$5F,$8F,$F8,$9C,$EE,$3B,$9F,$57,$86,$F1,$7D,$99,$FD
	.dc.b	$B4,$C8,$C3,$83,$84,$7D,$7D,$7F,$B7,$C7,$D5,$7D,$7C,$6E,$2F,$F7
	.dc.b	$45,$6E,$73,$F3,$AF,$7B,$F1,$98,$7F,$43,$8C,$2C,$3F,$A5,$46,$1B
	.dc.b	$3C,$27,$C1,$1D,$48,$FE,$26,$BF,$0B,$1C,$21,$E9,$52,$3E,$1A,$FE
	.dc.b	$2D,$1E,$72,$47,$3A,$E2,$D8,$F8,$8F,$CB,$2E,$5F,$A3,$6F,$33,$4F
	.dc.b	$F4,$DC,$E6,$7F,$09,$F4,$5F,$E8,$AC,$3F,$D4,$2D,$E3,$34,$C7,$D9
	.dc.b	$FE,$83,$E6,$3F,$1C,$DB,$FC,$73,$5F,$CA,$D2,$31,$1F,$A8,$87,$E0
	.dc.b	$DD,$C9,$87,$F8,$3E,$32,$FA,$33,$F5,$57,$B3,$3C,$11,$FD,$07,$81
	.dc.b	$E9,$25,$73,$7F,$D4,$7A,$7C,$19,$FC,$75,$F4,$7F,$1D,$19,$7F,$A4
	.dc.b	$36,$BF,$D8,$BC,$DF,$14,$8F,$9D,$FF,$59,$F3,$FF,$81,$7F,$2A,$FE
	.dc.b	$A1,$FA,$17,$F7,$F6,$E1,$FD,$E9,$AF,$F7,$96,$BF,$17,$E6,$3E,$91
	.dc.b	$30,$F8,$8F,$AB,$2F,$44,$FA,$B4,$F8,$4F,$AA,$32,$FC,$FF,$B0,$FB
	.dc.b	$84,$BF,$9C,$FE,$07,$F0,$A7,$CD,$7F,$67,$FB,$6F,$F6,$C5,$27,$36
	.dc.b	$9C,$C6,$1B,$5C,$6B,$C5,$8F,$92,$FC,$C7,$D9,$3F,$8B,$7D,$9F,$F2
	.dc.b	$A7,$BF,$12,$31,$5F,$C6,$5B,$3F,$8C,$6B,$FC,$CB,$1F,$2C,$F8,$A3
	.dc.b	$F9,$AB,$FC,$30,$F5,$61,$FC,$A1,$C4,$F0,$7C,$F3,$E5,$29,$0F,$CD
	.dc.b	$FC,$7F,$48,$FE,$58,$D8,$CC,$78,$1A,$FB,$3E,$07,$CE,$B1,$46,$1B
	.dc.b	$1F,$CF,$27,$F5,$33,$C3,$E4,$BE,$84,$64,$FE,$A2,$E1,$FA,$ED,$9F
	.dc.b	$80,$D8,$FA,$27,$AC,$FC,$EE,$DF,$86,$7F,$85,$1F,$D1,$EB,$C8,$4A
	.dc.b	$F0,$1E,$97,$E8,$C7,$03,$A9,$0E,$BA,$7E,$05,$A7,$CD,$34,$3E,$07
	.dc.b	$E2,$78,$9F,$E3,$7E,$D7,$C1,$7F,$44,$9C,$D7,$F3,$98,$C6,$9F,$EC
	.dc.b	$EA,$DC,$FF,$8F,$EB,$71,$88,$ED,$3B,$AE,$37,$DD,$B3,$D4,$E5,$3B
	.dc.b	$7D,$71,$FC,$44,$8F,$E7,$9A,$FE,$CD,$AF,$03,$E6,$65,$EC,$2C,$7E
	.dc.b	$15,$C4,$7E,$01,$AF,$13,$A1,$DC,$87,$F8,$0B,$4F,$C2,$CF,$E0,$B8
	.dc.b	$7F,$CC,$0D,$9E,$74,$AD,$BE,$22,$47,$86,$7C,$87,$ED,$36,$FF,$0F
	.dc.b	$FB,$16,$BF,$D6,$A7,$E6,$E3,$2F,$E9,$0F,$F5,$BF,$4F,$F8,$D6,$BE
	.dc.b	$1D,$FF,$80,$5F,$8B,$3F,$B4,$8E,$7C,$3F,$C3,$B0,$FD,$D1,$27,$F8
	.dc.b	$11,$EA,$CF,$F8,$33,$78,$FF,$C6,$78,$7F,$0A,$6C,$7C,$6C,$3E,$51
	.dc.b	$0F,$11,$FC,$0B,$8F,$FD,$0D,$C9,$FC,$24,$93,$99,$47,$F1,$63,$BB
	.dc.b	$FE,$2D,$6B,$FD,$81,$AF,$F3,$18,$FC,$93,$E1,$CF,$E2,$3E,$53,$C6
	.dc.b	$7F,$68,$E5,$E3,$3F,$C1,$DB,$91,$A7,$6E,$8F,$EA,$14,$FD,$D3,$FA
	.dc.b	$83,$4F,$EA,$05,$3F,$A8,$14,$FE,$E9,$D5,$D8,$98,$7A,$0F,$EC,$27
	.dc.b	$AA,$3C,$E7,$AF,$A3,$FC,$1D,$B7,$DC,$CB,$FD,$3D,$5C,$5B,$1F,$DD
	.dc.b	$A3,$1D,$7F,$1A,$F9,$93,$4F,$8A,$FE,$80,$FF,$05,$F6,$7C,$22,$57
	.dc.b	$57,$E2,$F6,$FF,$9D,$3E,$3F,$FC,$2B,$26,$92,$BF,$C4,$31,$E5,$F0
	.dc.b	$3E,$69,$1F,$76,$1E,$29,$F7,$76,$63,$5F,$7C,$7F,$8D,$7F,$02,$FF
	.dc.b	$27,$11,$C1,$FE,$3D,$0F,$F9,$0C,$8E,$5F,$F2,$B7,$DB,$C8,$FF,$20
	.dc.b	$35,$FF,$51,$78,$BF,$E5,$68,$FF,$26,$12,$3B,$3F,$E5,$FF,$C0,$A4
	.dc.b	$7F,$92,$5A,$FF,$81,$DF,$F1,$51,$1F,$EE,$EF,$F9,$DC,$8F,$F9,$F4
	.dc.b	$38,$7F,$D1,$4F,$F9,$D7,$93,$FD,$10,$DE,$3F,$F4,$44,$F0,$3F,$DA
	.dc.b	$E5,$3F,$C8,$A7,$FA,$AB,$E2,$3F,$D3,$0D,$3F,$D3,$05,$37,$AA,$3F
	.dc.b	$D4,$EB,$B2,$92,$7F,$D5,$2F,$FA,$D0,$8F,$FA,$D1,$AF,$FA,$26,$7F
	.dc.b	$D0,$C4,$7F,$CE,$0F,$F9,$3C,$8F,$F8,$71,$AF,$FA,$7F,$CB,$3E,$53
	.dc.b	$E4,$47,$FA,$84,$8F,$F8,$4F,$D1,$FF,$72,$DB,$E0,$FF,$D3,$CF,$FB
	.dc.b	$A8,$C4,$EF,$D2,$FF,$67,$CE,$AB,$8B,$FE,$F8,$AD,$D7,$FB,$D9,$AF
	.dc.b	$FB,$5A,$5E,$9F,$28,$FF,$7E,$CE,$B5,$FF,$60,$63,$F6,$7D,$0F,$5B
	.dc.b	$6B,$D8,$7F,$B2,$5F,$3F,$AA,$EA,$DC,$FF,$8B,$E3,$0F,$03,$FD,$CE
	.dc.b	$E2,$78,$CF,$09,$FF,$6D,$E4,$F5,$BF,$DD,$BF,$6B,$EF,$6C,$FC,$96
	.dc.b	$BC,$1A,$FD,$2F,$65,$ED,$62,$11,$8D,$57,$FE,$AA,$8C,$36,$B9,$CB
	.dc.b	$07,$B7,$86,$C7,$2C,$3E,$E4,$E4,$D7,$AE,$97,$DA,$FE,$60,$95,$FE
	.dc.b	$FE,$7E,$E7,$B5,$B6,$CF,$09,$FC,$44,$E5,$B1,$97,$F2,$3C,$2E,$77
	.dc.b	$BD,$F9,$4F,$BE,$FD,$A3,$6B,$FD,$DF,$5B,$8C,$47,$2A,$E2,$AB,$F7
	.dc.b	$9D,$1A,$85,$F7,$8C,$36,$38,$1D,$A9,$3A,$7E,$AC,$7E,$16,$7D,$1B
	.dc.b	$7E,$0E,$3A,$CE,$B8,$FF,$86,$4E,$65,$F0,$2F,$83,$FE,$94,$DD,$3F
	.dc.b	$AB,$37,$C7,$FF,$73,$B6,$BF,$6E,$7F,$BF,$FE,$05,$F0,$7F,$DE,$3D
	.dc.b	$C4,$63,$0F,$EF,$31,$89,$D4,$E7,$0F,$DA,$7A,$3F,$DE,$39,$8D,$26
	.dc.b	$1E,$4C,$FC,$E4,$70,$7F,$6F,$F8,$57,$C2,$FF,$8E,$B4,$FF,$4D,$46
	.dc.b	$1D,$46,$5F,$E2,$32,$9D,$63,$EE,$9F,$E2,$BF,$00,$F8,$6F,$EE,$0D
	.dc.b	$BF,$EF,$8D,$AE,$14,$8C,$3C,$F7,$F9,$4F,$CA,$6B,$FC,$89,$AF,$D6
	.dc.b	$CB,$FB,$F6,$5F,$C7,$2F,$AB,$67,$CC,$D7,$EA,$B5,$FE,$5B,$F1,$0F
	.dc.b	$88,$FD,$67,$8F,$E6,$8C,$67,$FD,$36,$B7,$5A,$7E,$14,$EA,$BF,$07
	.dc.b	$8E,$67,$D4,$D7,$D4,$D7,$D5,$7D,$57,$C3,$FF,$A4,$2B,$E3,$B6,$3E
	.dc.b	$83,$5F,$C7,$B5,$FE,$8F,$F1,$3F,$8D,$E6,$F9,$ED,$8F,$B0,$D7,$F2
	.dc.b	$ED,$7C,$B7,$FC,$B5,$F2,$CF,$B2,$31,$AF,$E8,$E3,$FA,$BC,$FF,$AD
	.dc.b	$1A,$BC,$F7,$9C,$F3,$FA,$7C,$53,$E2,$FF,$C6,$CD,$CB,$FC,$84,$DE
	.dc.b	$2F,$C3,$B7,$1F,$C8,$35,$FD,$7B,$5F,$EC,$ED,$4D,$25,$78,$09,$5E
	.dc.b	$06,$D7,$81,$B5,$E0,$6B,$E0,$6B,$E0,$6B,$E0,$6D,$78,$2B,$B1,$2B
	.dc.b	$B3,$6B,$B3,$5E,$CD,$7B,$35,$EC,$D9,$ED,$3D,$09,$3D,$1A,$F4,$6B
	.dc.b	$D1,$AF,$46,$C7,$48,$E4,$43,$91,$4E,$45,$39,$15,$E4,$F0,$21,$C0
	.dc.b	$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$57,$D6,$F8,$1F,$F8,$03
	.dc.b	$3F,$C6,$DA,$FF,$BE,$0C,$7A,$EF,$80,$FE,$33,$1F,$DC,$63,$FB,$93
	.dc.b	$FC,$73,$14,$62,$3F,$C6,$2D,$7F,$A2,$C6,$8E,$65,$79,$46,$DF,$E6
	.dc.b	$2D,$7C,$11,$C2,$30,$F1,$8E,$EE,$48,$F4,$6B,$A3,$04,$30,$43,$CF
	.dc.b	$1D,$89,$1C,$A3,$2D,$7C,$ED,$7C,$E7,$53,$65,$36,$53,$67,$84,$C9
	.dc.b	$4C,$9C,$8C,$3F,$05,$C1,$0E,$25,$38,$98,$21,$82,$3C,$DC,$B5,$CC
	.dc.b	$7C,$27,$D1,$19,$24,$66,$3C,$AD,$7C,$51,$A3,$E3,$15,$E7,$1B,$7C
	.dc.b	$2D,$7C,$31,$C2,$30,$F1,$8F,$E8,$B8,$FF,$A2,$2B,$FA,$4B,$63,$27
	.dc.b	$E6,$CB,$1F,$DD,$34,$FE,$6F,$9E,$67,$FC,$37,$18,$9D,$69,$FC,$ED
	.dc.b	$B5,$BA,$FE,$93,$1F,$D1,$C8,$FF,$7D,$E2,$FF,$2C,$47,$F8,$0A,$BF
	.dc.b	$CE,$92,$F8,$7C,$79,$7F,$EA,$06,$C6,$4F,$F0,$84,$F0,$8F,$09,$23
	.dc.b	$FC,$29,$8F,$AA,$31,$1F,$D7,$BB,$BF,$CD,$11,$FD,$91,$B1,$FD,$91
	.dc.b	$AE,$F4,$F0,$3F,$E3,$CF,$DB,$5F,$0F,$FC,$0A,$11,$87,$F8,$2B,$5F
	.dc.b	$F0,$FB,$5F,$2E,$3F,$D8,$DF,$92,$43,$3A,$FF,$0E,$EB,$7B,$FF,$C0
	.dc.b	$5B,$8E,$63,$07,$F0,$5F,$B2,$7C,$F7,$95,$B8,$7F,$18,$6B,$FE,$32
	.dc.b	$6B,$E0,$3C,$47,$E0,$31,$F3,$39,$3F,$C8,$18,$3E,$97,$C8,$4C,$7D
	.dc.b	$0F,$F1,$7F,$90,$7B,$DF,$1D,$B7,$FC,$A9,$AF,$F9,$11,$AF,$73,$C5
	.dc.b	$A7,$F3,$0E,$1F,$CA,$C7,$98,$FF,$72,$FE,$9E,$F8,$4F,$E4,$CD,$CB
	.dc.b	$F9,$BB,$67,$FC,$A2,$D7,$F8,$EF,$CB,$7F,$AA,$F6,$FE,$6E,$D8,$CD
	.dc.b	$27,$2D,$8F,$E4,$BF,$37,$B7,$F9,$BB,$5F,$E8,$CD,$8F,$E8,$AE,$7C
	.dc.b	$EF,$F5,$67,$E0,$9F,$EF,$EF,$33,$FC,$D9,$18,$BF,$F4,$39,$FE,$9F
	.dc.b	$9D,$34,$C9,$FE,$76,$F6,$BF,$19,$5F,$EE,$AD,$B7,$39,$6B,$FE,$E8
	.dc.b	$6E,$2F,$89,$7C,$62,$BF,$CA,$67,$F5,$88,$CB,$83,$FD,$0A,$E3,$C7
	.dc.b	$FF,$42,$57,$3A,$E5,$1F,$E8,$0F,$B3,$7D,$CE,$30,$FF,$45,$46,$1E
	.dc.b	$87,$85,$C4,$F1,$9E,$13,$E0,$E9,$AF,$FB,$8F,$1F,$CF,$D7,$89,$B1
	.dc.b	$E9,$9D,$AE,$57,$E4,$9F,$95,$24,$7E,$55,$FD,$F7,$27,$E1,$E3,$FD
	.dc.b	$7A,$D7,$F2,$3C,$BF,$90,$3F,$BC,$6A,$FD,$0B,$FD,$13,$F8,$9F,$C4
	.dc.b	$6D,$7F,$80,$9A,$FF,$AB,$9A,$FF,$BF,$5F,$D7,$4A,$79,$75,$7F,$C0
	.dc.b	$16,$C6,$6B,$A4,$E5,$FF,$5B,$65,$FA,$C7,$FD,$AA,$D7,$EF,$35,$F8
	.dc.b	$9D,$5F,$C7,$7A,$BF,$5F,$1F,$B3,$3F,$08,$E1,$FD,$A9,$FA,$CC,$BE
	.dc.b	$16,$1F,$B0,$6C,$7C,$1F,$98,$4B,$F0,$2C,$36,$7E,$24,$EA,$72,$7C
	.dc.b	$9F,$4F,$13,$F2,$5A,$FC,$96,$BF,$25,$AF,$DF,$DF,$EE,$BF,$C3,$DA
	.dc.b	$FF,$B6,$F7,$FB,$89,$A4,$C7,$F8,$E6,$7F,$C6,$C9,$3F,$C6,$DF,$91
	.dc.b	$DF,$FC,$A2,$3E,$29,$F2,$4F,$F5,$23,$97,$F1,$39,$FC,$72,$3F,$2D
	.dc.b	$FC,$93,$0F,$41,$4F,$45,$7C,$E6,$C7,$27,$F0,$38,$7B,$9F,$C0,$B5
	.dc.b	$F0,$9D,$C8,$72,$3F,$97,$1E,$8E,$7F,$E6,$57,$E9,$8C,$1F,$9D,$F7
	.dc.b	$3F,$43,$C4,$69,$30,$FF,$33,$43,$E1,$3F,$07,$E2,$7E,$95,$B7,$FA
	.dc.b	$56,$BF,$A5,$6B,$F8,$AF,$C5,$BF,$DE,$AF,$F1,$06,$BF,$E6,$E6,$BF
	.dc.b	$C6,$ED,$18,$8F,$CD,$E3,$EB,$E2,$FC,$E4,$63,$3F,$79,$E8,$CF,$D5
	.dc.b	$1E,$FC,$30,$69,$F8,$F5,$C9,$F8,$A7,$EF,$FE,$5B,$E2,$92,$7E,$1B
	.dc.b	$F2,$3B,$FF,$4B,$FA,$17,$F0,$76,$D7,$F0,$76,$BF,$C1,$DA,$FF,$07
	.dc.b	$DF,$E8,$92,$FE,$7B,$FE,$89,$C3,$C8,$4C,$FE,$01,$F9,$39,$FE,$31
	.dc.b	$E8,$7D,$A2,$3F,$61,$FE,$3B,$F0,$2F,$C0,$BF,$D3,$78,$FC,$D8,$C3
	.dc.b	$F3,$CC,$46,$9F,$E0,$31,$A7,$FD,$49,$87,$8F,$6F,$95,$1F,$C3,$FE
	.dc.b	$95,$DE,$FC,$4B,$FD,$5F,$9F,$F5,$5A,$56,$A7,$31,$FD,$66,$31,$3E
	.dc.b	$38,$DD,$73,$8C,$0C,$7C,$8F,$FA,$D9,$B3,$EA,$36,$E1,$FC,$31,$F8
	.dc.b	$0D,$3E,$5E,$3F,$EC,$3F,$DC,$DF,$35,$E4,$AF,$E4,$31,$8C,$3C,$2D
	.dc.b	$8F,$F6,$44,$23,$0F,$FB,$21,$79,$9C,$E3,$4D,$8C,$C7,$48,$D7,$E5
	.dc.b	$6E,$DF,$A5,$28,$C5,$FC,$26,$BE,$26,$C7,$F1,$9D,$FE,$B4,$FF,$8C
	.dc.b	$63,$C2,$77,$23,$EA,$C3,$C6,$FD,$EC,$FE,$2B,$F1,$F3,$FB,$EF,$F2
	.dc.b	$62,$1F,$30,$FF,$78,$FA,$7F,$E1,$3F,$21,$EA,$FE,$07,$D7,$FF,$77
	.dc.b	$CE,$63,$1E,$BF,$FA,$76,$B7,$7F,$E1,$BB,$DB,$FE,$1E,$AD,$DF,$18
	.dc.b	$D5,$F1,$9C,$C6,$CC,$CE,$DF,$E4,$91,$FE,$A2,$23,$FE,$7A,$6B,$FE
	.dc.b	$5A,$6B,$FC,$35,$FF,$13,$E5,$F6,$1F,$91,$BF,$F6,$66,$E9,$F0,$0F
	.dc.b	$C1,$46,$5E,$D5,$0E,$C5,$3F,$B7,$ED,$F7,$AB,$EE,$6F,$FE,$3D,$6C
	.dc.b	$F5,$3F,$93,$4E,$EB,$B9,$98,$FC,$CD,$F8,$5F,$E5,$98,$7A,$67,$77
	.dc.b	$F8,$1E,$0D,$57,$E8,$1B,$3F,$9D,$6D,$7E,$8E,$BF,$0D,$B7,$D8,$BF
	.dc.b	$59,$0F,$90,$7E,$D7,$97,$D5,$85,$8F,$33,$FB,$B8,$CC,$7E,$E6,$BF
	.dc.b	$C2,$2F,$F0,$5E,$3F,$E0,$51,$88,$FE,$0B,$CB,$E8,$E5,$FE,$13,$39
	.dc.b	$6E,$1F,$36,$BF,$02,$5E,$4F,$F7,$4D,$7F,$98,$B9,$7F,$91,$BF,$BC
	.dc.b	$6D,$79,$E3,$19,$79,$AF,$FC,$81,$CD,$F8,$17,$FC,$8C,$D7,$FC,$8C
	.dc.b	$D7,$75,$F1,$DF,$8A,$7C,$73,$CA,$7C,$97,$FC,$A9,$CF,$C5,$A7,$12
	.dc.b	$DF,$F9,$5E,$35,$DE,$E9,$FF,$1C,$FD,$17,$F5,$5B,$34,$98,$7F,$5F
	.dc.b	$8E,$94,$EC,$E6,$48,$F4,$4F,$98,$F4,$90,$C3,$FE,$6B,$ED,$F1,$91
	.dc.b	$EA,$FC,$38,$E4,$4D,$BE,$9B,$FA,$39,$FF,$00,$FB,$FF,$07,$BC,$D2
	.dc.b	$47,$A3,$C0,$FF,$2B,$F8,$7F,$E1,$58,$FF,$6A,$9F,$E7,$D8,$CD,$FF
	.dc.b	$85,$E3,$39,$FF,$88,$5F,$29,$37,$FF,$04,$1F,$C0,$A7,$CD,$E0,$39
	.dc.b	$FE,$1E,$9E,$89,$FF,$47,$F8,$7E,$DF,$13,$FD,$B9,$87,$F9,$14,$93
	.dc.b	$FE,$3E,$3E,$5F,$8F,$FE,$9C,$6F,$27,$EF,$DD,$4F,$07,$AE,$7F,$4A
	.dc.b	$FA,$C7,$F9,$2A,$7F,$C9,$E4,$78,$90,$E2,$F3,$7F,$CB,$6D,$7D,$B9
	.dc.b	$7B,$33,$F7,$4F,$C1,$C5,$B7,$96,$FF,$D2,$AD,$7F,$D2,$AD,$7F,$D2
	.dc.b	$A7,$9F,$3F,$88,$4C,$B0,$7C,$08,$97,$E9,$7F,$CC,$27,$E5,$3B,$FF
	.dc.b	$D2,$7F,$82,$FF,$A3,$68,$D2,$4F,$FA,$52,$32,$D8,$FF,$4C,$91,$F9
	.dc.b	$04,$3E,$29,$4F,$CE,$6D,$F7,$5B,$9F,$C4,$7F,$D6,$F9,$7B,$F3,$F8
	.dc.b	$B1,$F7,$23,$07,$C5,$C3,$FD,$9D,$E4,$FF,$AB,$3F,$13,$FF,$61,$92
	.dc.b	$BF,$D8,$72,$8F,$F6,$97,$1E,$E7,$FD,$A4,$D7,$FD,$A4,$D7,$FD,$A4
	.dc.b	$D7,$FD,$49,$B7,$CF,$DB,$EF,$70,$7E,$22,$7E,$6C,$7C,$F7,$FD,$9B
	.dc.b	$C3,$EB,$3D,$1E,$6B,$CD,$4B,$E3,$BF,$DF,$11,$86,$C6,$4F,$F7,$D1
	.dc.b	$4F,$F2,$17,$A0,$F4,$38,$47,$F9,$14,$46,$5E,$07,$81,$C4,$EE,$37
	.dc.b	$3C,$9F,$F4,$45,$BC,$BC,$47,$FB,$6A,$30,$F4,$DD,$18,$BF,$14,$62
	.dc.b	$B8,$AF,$4B,$F8,$05,$BE,$33,$71,$86,$F4,$FC,$66,$CE,$CF,$F3,$17
	.dc.b	$A8,$F8,$AF,$E6,$D1,$93,$E6,$C6,$7C,$2E,$51,$C2,$77,$5F,$1E,$B5
	.dc.b	$3C,$4F,$F7,$BF,$9C,$F5,$BF,$DF,$3F,$57,$EC,$8C,$68,$8C,$4E,$AB
	.dc.b	$FC,$ED,$2A,$DD,$71,$F6,$75,$5F,$EB,$59,$46,$23,$C2,$DA,$E4,$E6
	.dc.b	$3F,$27,$1F,$C0,$39,$7F,$27,$CF,$FD,$0F,$BB,$F5,$D5,$FC,$4A,$7F
	.dc.b	$88,$76,$7E,$AA,$3F,$58,$48,$FC,$17,$9E,$ED,$FE,$98,$DC,$7F,$AB
	.dc.b	$B6,$77,$3D,$FB,$FF,$75,$7F,$DC,$71,$F0,$9E,$2F,$67,$F4,$11,$FD
	.dc.b	$81,$B7,$F9,$16,$C7,$E6,$5A,$FF,$A7,$F0,$7B,$5E,$42,$65,$E4,$7C
	.dc.b	$5C,$BF,$CB,$C9,$87,$F4,$FF,$6D,$EB,$FF,$2B,$6E,$3F,$DC,$9B,$3F
	.dc.b	$A3,$6C,$7E,$AF,$E0,$DE,$EB,$27,$C2,$7E,$A6,$73,$18,$AE,$15,$BA
	.dc.b	$D3,$63,$F6,$DB,$37,$FE,$C5,$5F,$B3,$8C,$69,$CE,$BF,$D8,$11,$89
	.dc.b	$D5,$70,$9F,$E0,$0B,$E0,$3F,$86,$CF,$EF,$63,$13,$CC,$FE,$22,$E2
	.dc.b	$35,$3C,$27,$F8,$57,$C2,$3E,$0B,$E0,$7C,$27,$F0,$78,$C6,$88,$C4
	.dc.b	$EA,$B8,$B5,$F4,$E3,$FD,$FE,$7D,$39,$F1,$9F,$87,$39,$6C,$65,$FF
	.dc.b	$70,$7B,$1F,$E9,$CA,$FE,$30,$D8,$FE,$40,$D7,$7C,$1F,$CB,$A7,$F8
	.dc.b	$F6,$3F,$B3,$3F,$C6,$25,$8C,$F4,$FF,$8A,$E3,$15,$FB,$F7,$71,$92
	.dc.b	$38,$24,$F2,$6C,$F1,$8C,$11,$C1,$0F,$E0,$F3,$D2,$35,$18,$9E,$6D
	.dc.b	$8C,$BF,$C2,$63,$2D,$8C,$BF,$C6,$A3,$2D,$8C,$C7,$BA,$3A,$1E,$33
	.dc.b	$91,$4E,$05,$78,$3A,$38,$90,$C1,$0E,$26,$4A,$65,$E6,$E0,$8E,$09
	.dc.b	$3F,$16,$3C,$4D,$7A,$46,$8F,$79,$67,$DF,$3B,$9F,$03,$5E,$ED,$9E
	.dc.b	$11,$D9,$E7,$19,$23,$FE,$7A,$F7,$7F,$DD,$B3,$FE,$F8,$3F,$B6,$38
	.dc.b	$7F,$DD,$67,$F7,$5F,$C9,$BF,$C0,$33,$FD,$C6,$11,$87,$FB,$A3,$5E
	.dc.b	$07,$FA,$DC,$FF,$04,$9C,$0C,$1F,$E0,$C7,$83,$A6,$B9,$36,$71,$8F
	.dc.b	$14,$6A,$30,$78,$8D,$B9,$23,$82,$47,$26,$9C,$5C,$11,$C1,$0F,$6B
	.dc.b	$96,$99,$3D,$47,$43,$24,$32,$F1,$69,$C4,$F8,$0E,$5A,$64,$F7,$9C
	.dc.b	$CC,$94,$C9,$D0,$C4,$7C,$28,$E9,$1A,$38,$9B,$21,$82,$1B,$24,$73
	.dc.b	$7C,$8D,$7B,$9F,$04,$E6,$70,$30,$74,$78,$3A,$6B,$93,$67,$18,$F4
	.dc.b	$C6,$A3,$FA,$A6,$DF,$E5,$6F,$7B,$FA,$2C,$1A,$47,$F9,$8E,$3B,$69
	.dc.b	$B2,$C7,$F3,$88,$D4,$7F,$44,$D3,$4D,$9F,$E8,$A4,$7F,$CB,$9F,$88
	.dc.b	$7F,$94,$3E,$17,$FC,$B0,$D8,$CF,$8D,$FE,$6C,$9C,$B6,$3C,$71,$E1
	.dc.b	$4D,$23,$FD,$63,$3C,$B6,$3F,$AE,$C6,$23,$F6,$3C,$5F,$D8,$A7,$FC
	.dc.b	$11,$E0,$FE,$EE,$5D,$1F,$E4,$97,$F7,$CD,$76,$7F,$01,$E2,$6B,$C5
	.dc.b	$B1,$98,$C1,$D7,$F1,$D7,$C5,$FE,$E9,$B9,$FF,$0B,$6D,$7F,$9F,$5A
	.dc.b	$FF,$95,$3D,$DF,$F4,$23,$FE,$76,$6B,$9F,$33,$F8,$A5,$27,$33,$FD
	.dc.b	$B0,$F2,$10,$F9,$87,$FA,$03,$E0,$1F,$55,$FC,$52,$91,$87,$FB,$9B
	.dc.b	$67,$84,$FF,$1A,$F8,$27,$E1,$1F,$C9,$BE,$73,$FD,$39,$18,$AF,$F1
	.dc.b	$94,$62,$75,$1F,$C9,$27,$58,$2B,$CB,$3F,$94,$29,$EB,$3F,$D5,$1F
	.dc.b	$20,$F1,$5E,$AF,$F3,$58,$C3,$63,$33,$FE,$A9,$9C,$BF,$EA,$BF,$37
	.dc.b	$FD,$29,$5F,$EB,$EF,$23,$FD,$1D,$E7,$F0,$6C,$66,$BF,$D2,$8D,$7F
	.dc.b	$D3,$32,$F8,$2C,$42,$FF,$D4,$11,$86,$C2,$FB,$CE,$6B,$FD,$D7,$F8
	.dc.b	$2F,$8B,$CD,$FF,$68,$B6,$CD,$3E,$D7,$87,$FE,$5F,$6E,$DA,$9F,$F4
	.dc.b	$EF,$17,$F8,$AE,$72,$7C,$AF,$C6,$7F,$CC,$F3,$FE,$D2,$3F,$AB,$EC
	.dc.b	$8C,$FA,$FB,$F2,$FF,$DA,$95,$A9,$CB,$FE,$42,$9F,$F2,$39,$1C,$1D
	.dc.b	$3E,$B9,$F6,$1F,$E7,$59,$FB,$71,$8C,$FC,$55,$FE,$F0,$C5,$39,$AD
	.dc.b	$DF,$1A,$F6,$F5,$FE,$39,$B8,$7E,$2D,$BC,$BF,$9F,$9D,$47,$F7,$CD
	.dc.b	$7F,$D0,$73,$FA,$3D,$33,$3F,$E8,$99,$CB,$F2,$A3,$3C,$7F,$ED,$BF
	.dc.b	$41,$2F,$8E,$FC,$4C,$61,$B1,$93,$FD,$B9,$19,$7F,$DC,$BF,$A3,$BA
	.dc.b	$71,$1F,$EE,$48,$C3,$C4,$C4,$69,$FC,$74,$6A,$BF,$4B,$FA,$AB,$EA
	.dc.b	$FC,$BA,$61,$B7,$F7,$A7,$4F,$C7,$3E,$83,$89,$FF,$40,$92,$3F,$CF
	.dc.b	$ED,$7E,$B7,$E8,$4F,$A6,$F9,$3C,$88,$C7,$0F,$5A,$4E,$6B,$75,$AB
	.dc.b	$FC,$F7,$E1,$12,$FF,$C5,$3B,$BA,$9D,$5F,$42,$B5,$5F,$6F,$AB,$FD
	.dc.b	$81,$1F,$A2,$24,$7E,$8A,$32,$FE,$1B,$F0,$8F,$C2,$6A,$69,$23,$EA
	.dc.b	$F7,$FC,$87,$EA,$35,$FA,$2D,$C9,$C5,$FC,$56,$7E,$06,$EE,$0F,$9F
	.dc.b	$87,$D3,$9F,$01,$F4,$F8,$7F,$17,$19,$3F,$D6,$24,$8F,$F5,$7B,$5F
	.dc.b	$A3,$C3,$FE,$DF,$8F,$E4,$BC,$3F,$41,$B1,$99,$FD,$EC,$E5,$FD,$FF
	.dc.b	$0F,$FB,$9D,$7E,$11,$FC,$BA,$30,$DD,$BF,$3E,$D9,$D9,$D2,$FF,$DE
	.dc.b	$11,$E0,$9C,$CF,$99,$AF,$E0,$AF,$F0,$B3,$CF,$F0,$89,$3E,$4A,$FD
	.dc.b	$6C,$62,$39,$3D,$48,$F8,$61,$1F,$ED,$88,$87,$FB,$58,$A7,$EB,$7E
	.dc.b	$6B,$FC,$F9,$BF,$D8,$6E,$E6,$98,$F8,$9F,$DF,$FB,$9F,$E7,$CB,$FD
	.dc.b	$8C,$61,$FD,$DB,$5F,$DD,$35,$FE,$19,$F4,$FF,$DB,$FB,$0D,$26,$7F
	.dc.b	$E8,$FF,$71,$ED,$3D,$6C,$37,$0F,$59,$FC,$4A,$77,$3F,$EE,$AF,$D1
	.dc.b	$FF,$D1,$76,$69,$2B,$FD,$1B,$5C,$1E,$61,$C3,$F2,$4F,$E2,$FF,$6B
	.dc.b	$F1,$88,$FF,$6B,$D9,$1F,$2B,$CC,$FE,$4B,$F8,$B7,$C7,$E1,$FD,$2C
	.dc.b	$62,$FF,$90,$B5,$FE,$40,$D7,$F8,$EF,$E0,$DF,$50,$8F,$77,$FC,$75
	.dc.b	$8F,$9A,$3F,$CA,$C7,$A9,$C3,$FC,$9A,$FE,$E9,$1F,$43,$FE,$2A,$F8
	.dc.b	$93,$FC,$F3,$F0,$7F,$D2,$4B,$55,$6A,$13,$A9,$CD,$71,$7F,$D8,$26
	.dc.b	$27,$6F,$CB,$7C,$E5,$3E,$D6,$FF,$75,$FE,$8B,$F9,$0F,$27,$FA,$23
	.dc.b	$5F,$E8,$2D,$D3,$F8,$8B,$9A,$F0,$7D,$A7,$97,$2F,$43,$71,$FE,$B2
	.dc.b	$DC,$0F,$FB,$67,$3F,$93,$1F,$83,$CB,$C8,$FA,$08,$7C,$5E,$0F,$E6
	.dc.b	$D8,$F2,$3E,$0B,$8B,$F8,$CD,$8C,$F8,$9F,$36,$73,$CD,$FC,$4E,$78
	.dc.b	$B6,$FF,$DE,$76,$8C,$47,$85,$B3,$F5,$B4,$F8,$5A,$1A,$4A,$FF,$78
	.dc.b	$E7,$EE,$26,$7E,$E8,$FA,$7C,$7F,$AB,$C7,$AE,$5E,$1C,$BF,$E0,$01
	.dc.b	$FA,$D8,$C9,$E4,$AF,$91,$5F,$B8,$23,$F9,$5B,$EC,$E8,$C4,$7F,$0D
	.dc.b	$7F,$C1,$BF,$02,$F0,$FF,$D6,$DE,$EF,$FA,$42,$53,$9F,$7F,$FC,$27
	.dc.b	$39,$8C,$4A,$31,$5C,$BD,$9F,$F0,$CC,$F2,$8C,$3E,$06,$B9,$BE,$B3
	.dc.b	$C2,$FA,$DF,$8F,$5F,$E2,$33,$E9,$6D,$FF,$27,$6B,$FE,$CB,$8D,$65
	.dc.b	$FC,$75,$B9,$7B,$67,$E8,$C6,$0F,$E6,$1A,$7C,$F3,$FD,$66,$7E,$18
	.dc.b	$B1,$F8,$6D,$3F,$71,$3F,$8C,$9C,$CF,$43,$C9,$7E,$16,$DF,$85,$FB
	.dc.b	$35,$E0,$DF,$F1,$51,$CA,$FF,$1C,$D8,$CC,$FE,$BF,$7F,$F2,$2B,$E1
	.dc.b	$25,$79,$DB,$5F,$9C,$E1,$FE,$1D,$E5,$FE,$99,$B5,$FA,$61,$19,$69
	.dc.b	$E8,$AF,$8D,$54,$87,$F9,$5A,$3F,$89,$4F,$EF,$C8,$FF,$16,$C3,$F8
	.dc.b	$E4,$7F,$20,$E4,$FE,$17,$5F,$21,$B7,$9E,$9F,$7E,$1F,$C5,$AB,$E6
	.dc.b	$9F,$20,$F0,$9F,$3C,$FC,$01,$DF,$97,$FC,$D3,$EE,$7F,$11,$24,$E6
	.dc.b	$51,$DE,$7E,$5C,$7F,$9A,$1A,$FF,$99,$DA,$FF,$3F,$3E,$D7,$07,$E4
	.dc.b	$1F,$11,$0F,$31,$FE,$6F,$E8,$E3,$97,$F8,$E9,$B8,$9A,$63,$A7,$FD
	.dc.b	$03,$EC,$FF,$5D,$9F,$F7,$5D,$FF,$BD,$AF,$FA,$FB,$63,$33,$BE,$BE
	.dc.b	$A4,$9F,$D0,$57,$F9,$2A,$3D,$71,$E6,$3D,$18,$FA,$DC,$69,$FD,$F5
	.dc.b	$FC,$F9,$FC,$C0,$A7,$F2,$82,$9F,$CD,$CA,$75,$24,$75,$3E,$44,$3B
	.dc.b	$7F,$C2,$EF,$ED,$4C,$76,$F4,$69,$98,$DE,$FF,$17,$BB,$FD,$4F,$E5
	.dc.b	$DC,$7F,$CF,$67,$F8,$A0,$A7,$F8,$A0,$A7,$FA,$7B,$C0,$FA,$64,$7E
	.dc.b	$3E,$DF,$E3,$77,$EF,$F4,$7E,$ED,$FC,$79,$0F,$C0,$F9,$3F,$E9,$1C
	.dc.b	$DF,$83,$7F,$95,$C9,$3F,$E4,$B7,$FC,$BC,$47,$FC,$A8,$70,$2B,$FE
	.dc.b	$5F,$FC,$1A,$47,$F9,$4D,$AF,$FB,$6F,$C8,$FF,$37,$47,$F9,$A4,$91
	.dc.b	$D1,$FF,$34,$7E,$0D,$23,$FC,$B8,$D7,$F9,$BF,$0F,$F8,$FA,$3F,$93
	.dc.b	$92,$3F,$8C,$BF,$E8,$D2,$3F,$E8,$D0,$E1,$FF,$4D,$8F,$C1,$3F,$35
	.dc.b	$E7,$7F,$B7,$1B,$19,$3F,$D4,$65,$3F,$DD,$5E,$9F,$F8,$96,$BF,$D5
	.dc.b	$64,$8F,$F7,$9B,$FC,$E3,$8F,$FA,$D6,$8F,$C1,$B1,$E3,$7F,$8B,$5B
	.dc.b	$19,$BF,$F5,$C3,$5F,$F7,$7F,$47,$FB,$0F,$93,$FD,$E3,$5F,$E4,$DF
	.dc.b	$03,$F1,$6D,$8C,$D7,$FB,$2A,$72,$FF,$A1,$F4,$FF,$4B,$11,$C1,$FD
	.dc.b	$09,$FE,$A6,$47,$07,$CB,$7F,$D7,$44,$70,$7F,$BF,$3D,$32,$F9,$6F
	.dc.b	$E0,$D1,$86,$C6,$4E,$05,$3D,$F5,$FD,$01,$C0,$C7,$FD,$C9,$E0,$F3
	.dc.b	$8F,$73,$5E,$2D,$7F,$CB,$5A,$BD,$5F,$F2,$E6,$5F,$55,$B3,$FE,$58
	.dc.b	$9D,$6B,$F7,$E3,$8C,$E9,$FC,$0C,$64,$FE,$0D,$C1,$98,$EE,$D8,$C9
	.dc.b	$CF,$99,$EC,$3D,$7F,$F3,$77,$D4,$FF,$AB,$67,$31,$8C,$3F,$C2,$F1
	.dc.b	$8B,$56,$EF,$94,$62,$BF,$BD,$CF,$F8,$30,$8E,$0E,$7E,$D7,$C3,$78
	.dc.b	$1F,$62,$D1,$99,$F2,$36,$78,$E5,$F7,$A7,$F3,$3F,$81,$C9,$FC,$CB
	.dc.b	$5F,$B7,$C7,$F8,$48,$FC,$B4,$1A,$47,$F3,$1E,$EF,$E2,$BE,$47,$F3
	.dc.b	$CD,$8C,$88,$CB,$4F,$D0,$7C,$06,$DF,$C5,$B5,$FC,$57,$22,$31,$C8
	.dc.b	$8C,$4F,$33,$E2,$75,$63,$C3,$FA,$CD,$8F,$AA,$D7,$77,$FD,$6B,$E1
	.dc.b	$15,$49,$3F,$E2,$F8,$CB,$FE,$41,$E9,$7C,$17,$F8,$67,$A3,$FD,$E1
	.dc.b	$18,$9F,$C7,$37,$93,$F2,$CD,$9F,$E2,$1E,$4F,$F5,$5E,$93,$49,$5E
	.dc.b	$1C,$BC,$67,$F9,$C7,$2F,$1B,$4F,$F0,$6C,$3E,$0D,$E5,$61,$B9,$FF
	.dc.b	$65,$6D,$7F,$96,$1B,$1F,$E7,$01,$F0,$8E,$CF,$EE,$6D,$C3,$FC,$A4
	.dc.b	$DB,$FF,$42,$36,$B9,$63,$E6,$6C,$FA,$9A,$7F,$96,$FE,$17,$FC,$56
	.dc.b	$DF,$3F,$FC,$1B,$18,$9F,$F4,$83,$67,$CD,$F0,$EF,$8C,$FF,$71,$7C
	.dc.b	$6F,$FA,$56,$31,$9A,$31,$3A,$AE,$2D,$7F,$D1,$F9,$7D,$E9,$F9,$91
	.dc.b	$C4,$F3,$B8,$3F,$D4,$5F,$15,$FE,$99,$8D,$1F,$E9,$87,$11,$AF,$63
	.dc.b	$F9,$7B,$67,$67,$FA,$EB,$E3,$7F,$D1,$A4,$7F,$CF,$EF,$88,$8F,$C8
	.dc.b	$6B,$FE,$A6,$F3,$9F,$19,$FC,$F2,$7F,$D6,$D1,$89,$46,$27,$51,$F7
	.dc.b	$E7,$53,$FE,$BD,$F8,$DF,$F3,$A6,$F6,$BF,$C6,$71,$87,$F0,$CD,$7F
	.dc.b	$D6,$ED,$7F,$A6,$B5,$34,$8F,$42,$6B,$FE,$9D,$8C,$B6,$33,$3E,$38
	.dc.b	$C1,$1C,$12,$3C,$A7,$02,$C6,$DB,$1B,$7D,$26,$CB,$1B,$6C,$6D,$F6
	.dc.b	$9B,$2C,$6D,$B1,$B7,$E0,$9B,$2C,$6D,$B1,$B9,$F8,$2D,$8E,$11,$87
	.dc.b	$D2,$43,$CA,$53,$C2,$53,$A9,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1,$D1
	.dc.b	$0D,$15,$D3,$82,$38,$23,$EC,$3F,$CF,$FE,$E3,$CC,$7B,$BF,$C4,$E7
	.dc.b	$FE,$00,$35,$FF,$80,$06,$67,$84,$66,$9F,$46,$C7,$0D,$BF,$CF,$35
	.dc.b	$FC,$DB,$5F,$CD,$B5,$FC,$DB,$5F,$CD,$B6,$7B,$CF,$62,$4F,$66,$BD
	.dc.b	$9A,$F6,$6B,$D9,$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$85,$3A,$14,$E8
	.dc.b	$57,$A3,$C8,$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78
	.dc.b	$3A,$21,$A2,$BA,$70,$47,$04,$9F,$A8,$49,$F9,$2D,$9F,$80,$D7,$D2
	.dc.b	$D7,$C8,$D7,$C2,$D7,$C0,$D9,$F0,$4F,$62,$4F,$E6,$7A,$5F,$3B,$FE
	.dc.b	$70,$C3,$0D,$9F,$EE,$13,$A7,$F4,$26,$23,$4F,$F9,$E0,$FF,$03,$F0
	.dc.b	$61,$FF,$3D,$CF,$F5,$3A,$F0,$4E,$67,$67,$43,$97,$87,$FA,$67,$07
	.dc.b	$F8,$EF,$2E,$11,$9A,$FD,$93,$67,$FC,$0D,$E8,$B8,$1E,$F7,$29,$4E
	.dc.b	$63,$0D,$BF,$DE,$36,$7F,$B9,$E9,$C7,$9F,$FB,$CD,$7F,$78,$8C,$73
	.dc.b	$FE,$EE,$3F,$82,$7C,$87,$FA,$09,$BE,$5E,$70,$FF,$0D,$4E,$5B,$19
	.dc.b	$7E,$57,$E6,$EF,$AE,$FE,$19,$08,$C3,$FC,$3D,$AF,$F1,$36,$BE,$83
	.dc.b	$CC,$43,$07,$F8,$AF,$BB,$FC,$69,$97,$F1,$96,$C7,$F8,$A3,$C0,$FE
	.dc.b	$19,$18,$8F,$F1,$97,$D6,$3E,$D7,$CA,$D8,$FE,$44,$D7,$F9,$33,$5F
	.dc.b	$09,$F8,$7D,$BF,$C4,$6F,$94,$C6,$DF,$C6,$FC,$97,$D0,$52,$67,$FC
	.dc.b	$BE,$32,$FF,$33,$F4,$9F,$3B,$E5,$8C,$37,$0F,$F5,$33,$5F,$F5,$2B
	.dc.b	$5F,$0C,$7F,$3B,$C7,$CB,$1E,$83,$FD,$AA,$E1,$FF,$76,$FB,$BF,$D1
	.dc.b	$5F,$CB,$F9,$E8,$C7,$27,$F9,$66,$31,$DF,$FA,$56,$D7,$0A,$7C,$4F
	.dc.b	$FB,$8B,$F1,$AE,$11,$89,$46,$27,$51,$FB,$1A,$F0,$46,$23,$CB,$E5
	.dc.b	$57,$FD,$C5,$7F,$E6,$97,$99,$C5,$C5,$F8,$27,$86,$6F,$89,$7D,$BF
	.dc.b	$F5,$9F,$CE,$3F,$CE,$93,$98,$C6,$3F,$C0,$E3,$02,$77,$5C,$A3,$13
	.dc.b	$B8,$7C,$97,$28,$1F,$D9,$E3,$33,$C2,$32,$F9,$67,$84,$6D,$BC,$2F
	.dc.b	$9B,$7D,$BF,$4F,$A6,$FF,$74,$46,$2D,$18,$9D,$47,$FB,$AA,$75,$3F
	.dc.b	$C1,$BC,$2F,$E1,$AD,$F0,$B8,$D7,$F1,$29,$CB,$63,$2F,$FB,$9F,$EE
	.dc.b	$9F,$15,$FE,$03,$BF,$F7,$D4,$63,$4D,$D3,$34,$E6,$B7,$7C,$6B,$ED
	.dc.b	$78,$5F,$8D,$E3,$FF,$7B,$73,$FE,$2A,$30,$67,$F0,$79,$5F,$91,$6C
	.dc.b	$FE,$43,$C1,$FA,$3F,$69,$68,$D2,$3F,$89,$FB,$7D,$B4,$D9,$63,$F3
	.dc.b	$71,$A8,$FB,$BA,$BE,$5F,$F0,$31,$FD,$02,$31,$C9,$FD,$2D,$B1,$99
	.dc.b	$DE,$3F,$44,$91,$F3,$DA,$F6,$FA,$7F,$F7,$B7,$0F,$56,$CF,$FB,$36
	.dc.b	$D1,$8A,$FE,$B9,$18,$F7,$FF,$9B,$47,$EA,$BE,$E3,$F2,$B8,$FE,$D9
	.dc.b	$B1,$FA,$B6,$BF,$C1,$9A,$FE,$CB,$D5,$BD,$FF,$C1,$DB,$C5,$FE,$A5
	.dc.b	$BF,$E0,$31,$C3,$4F,$D9,$7B,$5E,$26,$9F,$D9,$8B,$AF,$EB,$A3,$CF
	.dc.b	$87,$F1,$3E,$2F,$D5,$FD,$17,$F0,$C6,$E7,$FC,$05,$AF,$F1,$C6,$BF
	.dc.b	$91,$CB,$EA,$CF,$C7,$6C,$7C,$66,$BF,$71,$AF,$EE,$BC,$EF,$EE,$13
	.dc.b	$FA,$72,$47,$E9,$5F,$A5,$C1,$F5,$9B,$BF,$F9,$9D,$BA,$1A,$7F,$26
	.dc.b	$F2,$BF,$17,$7F,$C6,$DA,$FF,$36,$6B,$FC,$89,$BC,$3F,$76,$BF,$5D
	.dc.b	$3F,$80,$C3,$F0,$F3,$E0,$24,$75,$1F,$81,$78,$5A,$65,$C1,$E8,$23
	.dc.b	$D7,$1F,$F4,$14,$62,$3F,$C5,$59,$7F,$85,$5F,$E1,$84,$3F,$85,$1F
	.dc.b	$CC,$3F,$08,$FF,$0F,$63,$FE,$93,$0E,$23,$5D,$BF,$E8,$79,$CB,$65
	.dc.b	$E6,$7F,$A4,$FF,$34,$E8,$DB,$FE,$75,$3A,$B5,$69,$B3,$AF,$AC,$E8
	.dc.b	$48,$F2,$52,$3F,$D2,$D8,$FB,$6B,$D6,$43,$E0,$E3,$FE,$38,$7D,$DF
	.dc.b	$4F,$E7,$26,$1D,$1E,$73,$B9,$E4,$7E,$3D,$C3,$D3,$C1,$F1,$FD,$BF
	.dc.b	$C1,$25,$E0,$F3,$F9,$FF,$EB,$5F,$A4,$F9,$CD,$BF,$86,$D7,$FA,$8E
	.dc.b	$9F,$E1,$29,$FA,$8F,$CD,$25,$7C,$27,$CF,$87,$FB,$1F,$3F,$85,$7E
	.dc.b	$53,$EF,$1D,$3D,$DF,$AC,$4B,$F9,$6F,$DB,$FA,$C7,$57,$F6,$6F,$87
	.dc.b	$FF,$64,$52,$73,$69,$CC,$61,$B5,$C6,$BE,$4D,$7F,$99,$B6,$F9,$2F
	.dc.b	$CB,$CB,$D1,$18,$9F,$C7,$FA,$FF,$ED,$97,$F0,$ED,$AF,$8A,$D7,$FD
	.dc.b	$C1,$A7,$F9,$76,$7F,$BC,$C7,$F7,$AC,$BC,$F1,$E4,$8F,$F7,$33,$FE
	.dc.b	$0B,$8C,$B7,$1F,$F0,$0B,$E8,$D3,$E5,$D7,$9C,$F9,$67,$83,$DD,$E7
	.dc.b	$3F,$9C,$6C,$7D,$A6,$BB,$8F,$0B,$F7,$BE,$C7,$D5,$1F,$AE,$6E,$26
	.dc.b	$9A,$F8,$EF,$A7,$5F,$E1,$66,$BF,$F0,$05,$B4,$73,$1F,$47,$5F,$9E
	.dc.b	$47,$D7,$3E,$13,$E4,$BF,$32,$BC,$31,$83,$50,$7E,$36,$BF,$72,$3B
	.dc.b	$1F,$18,$A6,$AB,$C0,$FF,$AF,$47,$C0,$BD,$5F,$F6,$1F,$BE,$E2,$FF
	.dc.b	$61,$4E,$7C,$3F,$F1,$CC,$E6,$FF,$C7,$73,$95,$19,$7A,$DF,$4B,$E6
	.dc.b	$D8,$E2,$D7,$77,$DA,$76,$D8,$FC,$C4,$7F,$8D,$48,$FF,$8C,$DA,$FF
	.dc.b	$05,$6B,$C0,$FE,$43,$9F,$C9,$8F,$D6,$EB,$F7,$F4,$F8,$06,$8B,$5F
	.dc.b	$87,$6B,$CC,$EC,$78,$48,$7E,$15,$FF,$2F,$CF,$EE,$B5,$FB,$13,$CA
	.dc.b	$7F,$0C,$68,$FF,$37,$46,$A7,$CB,$7E,$79,$F2,$46,$1E,$57,$E9,$8F
	.dc.b	$5E,$DF,$85,$F5,$2D,$1A,$47,$CF,$8F,$AC,$87,$E5,$EB,$FD,$07,$C5
	.dc.b	$FC,$42,$BF,$74,$4A,$FD,$C5,$75,$7F,$83,$9E,$4E,$3F,$44,$7F,$08
	.dc.b	$8C,$1F,$29,$EC,$78,$8A,$7C,$C3,$F8,$57,$1F,$FA,$23,$D0,$FE,$14
	.dc.b	$49,$F0,$BF,$6B,$1F,$B9,$5F,$D7,$9A,$FF,$31,$9F,$CA,$3F,$C6,$23
	.dc.b	$2F,$C0,$E4,$72,$39,$37,$3E,$43,$E3,$5F,$F7,$4D,$B8,$B4,$EA,$5C
	.dc.b	$3F,$A4,$B8,$8F,$F6,$EF,$CA,$7F,$52,$83,$49,$5F,$DD,$BA,$7A,$D1
	.dc.b	$A4,$AF,$2E,$5E,$8C,$6E,$5E,$9C,$BF,$43,$8F,$A6,$FF,$C1,$7D,$5D
	.dc.b	$08,$F9,$5F,$87,$1E,$F2,$61,$D6,$BF,$57,$5F,$E1,$4E,$BF,$29,$1F
	.dc.b	$73,$E8,$ED,$FE,$2D,$F9,$47,$F8,$34,$98,$7F,$84,$A3,$2F,$F8,$53
	.dc.b	$2F,$79,$2F,$E4,$3E,$4C,$7C,$53,$F8,$76,$BF,$87,$3F,$D5,$9E,$17
	.dc.b	$FA,$13,$D2,$FF,$26,$92,$3F,$C6,$D7,$FE,$B0,$7F,$C8,$11,$FE,$3F
	.dc.b	$23,$83,$7F,$7F,$FE,$56,$23,$83,$FC,$90,$FF,$91,$C8,$E0,$EF,$F0
	.dc.b	$DF,$E5,$E2,$38,$3F,$D9,$4F,$FB,$3C,$8E,$0F,$F3,$97,$9D,$FE,$E8
	.dc.b	$8F,$F3,$B3,$63,$22,$32,$D3,$FD,$04,$53,$FD,$04,$51,$E5,$FF,$A2
	.dc.b	$DC,$36,$3F,$D2,$18,$3E,$1A,$92,$BF,$DF,$DC,$FF,$D4,$E3,$FD,$3B
	.dc.b	$0F,$CF,$A9,$1F,$F8,$02,$7F,$A9,$CD,$1F,$EA,$D2,$9F,$EA,$D2,$9F
	.dc.b	$EB,$52,$E0,$F3,$DE,$82,$77,$E5,$FF,$A0,$70,$FF,$46,$92,$3F,$02
	.dc.b	$FF,$BF,$48,$FF,$BE,$DA,$FF,$A7,$F4,$E9,$1F,$30,$91,$F4,$1F,$F7
	.dc.b	$0A,$FC,$72,$91,$FF,$74,$3F,$EE,$F2,$3F,$0B,$1F,$57,$A9,$E9,$9C
	.dc.b	$CF,$D6,$AF,$F5,$E3,$F4,$4F,$B6,$E0,$FE,$F5,$5F,$0F,$D5,$F3,$CF
	.dc.b	$5C,$D1,$88,$F8,$D3,$B7,$A1,$89,$DB,$81,$93,$85,$EB,$7F,$84,$7F
	.dc.b	$28,$FF,$77,$46,$38,$51,$89,$D5,$FF,$BC,$2B,$53,$FE,$11,$8F,$F1
	.dc.b	$23,$79,$7F,$8D,$37,$9B,$F9,$93,$7B,$3A,$D7,$FB,$EB,$C0,$7B,$3F
	.dc.b	$81,$AF,$F8,$01,$18,$9D,$9F,$85,$D1,$19,$9D,$D7,$1A,$FC,$AE,$8F
	.dc.b	$69,$C9,$FE,$2F,$FC,$DF,$FC,$51,$39,$8C,$57,$0A,$DC,$AB,$71,$86
	.dc.b	$FB,$BC,$2B,$F3,$15,$DE,$31,$1F,$E3,$DA,$46,$27,$47,$59,$FB,$DE
	.dc.b	$37,$E8,$7F,$50,$FC,$1F,$1F,$F7,$F9,$46,$27,$51,$C0,$C9,$F8,$7F
	.dc.b	$80,$CB,$F8,$36,$BF,$C1,$A9,$18,$E5,$FD,$93,$63,$FC,$95,$3F,$89
	.dc.b	$9F,$DB,$79,$DF,$91,$8E,$4D,$8C,$C7,$E3,$C7,$C0,$BE,$13,$FA,$9B
	.dc.b	$73,$FE,$E6,$DF,$1B,$F1,$4D,$AF,$E1,$BC,$8F,$83,$7B,$5F,$DF,$A3
	.dc.b	$0D,$8C,$E1,$FE,$04,$9C,$BF,$E0,$0F,$84,$7B,$1F,$D9,$1B,$19,$8C
	.dc.b	$1E,$6C,$E4,$79,$9B,$9F,$E7,$FC,$8F,$1E,$7F,$E0,$C6,$C6,$73,$FF
	.dc.b	$23,$CF,$F9,$D6,$72,$FF,$9D,$FE,$15,$E1,$FF,$76,$6E,$3F,$E1,$66
	.dc.b	$E1,$FA,$D6,$D7,$F9,$FF,$6F,$8C,$FF,$32,$6B,$FC,$A5,$AF,$D2,$CB
	.dc.b	$FC,$07,$F0,$CF,$47,$CB,$3F,$CE,$A3,$0F,$F8,$99,$B8,$7E,$F5,$B1
	.dc.b	$FE,$8F,$F8,$7F,$E2,$79,$7D,$F6,$F8,$BF,$C0,$DB,$3B,$F1,$7F,$A6
	.dc.b	$3E,$9C,$AB,$E9,$6B,$E9,$6B,$E9,$BF,$C6,$78,$7F,$E2,$5C,$3F,$9E
	.dc.b	$CF,$FB,$85,$AB,$E2,$1F,$13,$F9,$78,$C3,$63,$31,$F9,$76,$CF,$F1
	.dc.b	$BF,$8A,$FF,$50,$CF,$E6,$DB,$19,$F6,$FA,$46,$27,$FD,$64,$BD,$8C
	.dc.b	$92,$39,$3E,$3F,$8B,$F2,$E7,$FE,$54,$6F,$07,$F2,$B6,$D7,$FA,$F5
	.dc.b	$AF,$F7,$06,$BF,$D9,$DA,$9A,$4C,$FF,$C8,$6D,$AF,$04,$60,$8E,$09
	.dc.b	$5E,$38,$CB,$63,$35,$E5,$9C,$B6,$32,$F9,$E3,$2D,$8C,$BE,$98,$CB
	.dc.b	$63,$2F,$AE,$32,$D8,$CD,$7A,$EB,$C8,$49,$F1,$35,$F0,$35,$EC,$D7
	.dc.b	$A3,$63,$A4,$72,$21,$C8,$A7,$22,$9C,$8A,$F2,$78,$10,$E0,$53,$81
	.dc.b	$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$2B,$E4,$BB,$F7,$7F,$DF,$D3,$FE
	.dc.b	$3A,$3F,$DE,$E7,$F7,$B7,$2F,$FB,$E0,$FC,$A7,$85,$FE,$E0,$7F,$28
	.dc.b	$D7,$F2,$8D,$7F,$28,$D7,$F2,$8D,$9F,$EC,$78,$7F,$82,$67,$2D,$8C
	.dc.b	$C7,$82,$31,$3C,$48,$E0,$91,$B2,$1E,$67,$2D,$32,$7A,$1C,$B4,$C9
	.dc.b	$EA,$72,$D3,$27,$A0,$B3,$E8,$7B,$10,$E8,$53,$91,$4E,$45,$79,$3C
	.dc.b	$08,$70,$29,$C0,$AF,$07,$44,$34,$57,$4E,$08,$E0,$93,$F3,$A3,$DA
	.dc.b	$D7,$CC,$D7,$C2,$D7,$B3,$5E,$AD,$8E,$B3,$F3,$27,$2D,$8C,$C7,$82
	.dc.b	$31,$3C,$48,$E0,$E5,$97,$F3,$8F,$77,$F9,$B5,$1A,$47,$F4,$1F,$1A
	.dc.b	$CF,$FD,$21,$3F,$E5,$B8,$C4,$F1,$37,$3A,$69,$93,$FD,$47,$C1,$FE
	.dc.b	$F9,$AF,$F2,$F1,$23,$F5,$D1,$97,$FB,$16,$1C,$5F,$EC,$0D,$7F,$B0
	.dc.b	$4F,$F8,$93,$85,$39,$E9,$7D,$0F,$EA,$6D,$18,$7F,$58,$D7,$FC,$31
	.dc.b	$1B,$C1,$E6,$BC,$DF,$D9,$46,$1B,$19,$AF,$F3,$C1,$FE,$7F,$FC,$65
	.dc.b	$F8,$37,$EF,$5B,$97,$EE,$DB,$3B,$9F,$E0,$5F,$22,$FA,$9F,$43,$67
	.dc.b	$F8,$13,$5F,$F4,$63,$5F,$93,$87,$8D,$B8,$F9,$88,$FF,$A4,$FC,$37
	.dc.b	$CF,$7F,$11,$6E,$7F,$C3,$DB,$3B,$9F,$E2,$DA,$BE,$67,$E6,$36,$7F
	.dc.b	$8B,$35,$FF,$50,$35,$F1,$9F,$18,$A4,$3F,$DC,$FE,$E3,$E7,$BF,$92
	.dc.b	$37,$1F,$E4,$4D,$9F,$F5,$63,$5F,$F4,$DF,$C1,$7A,$BF,$16,$FF,$4C
	.dc.b	$B6,$33,$97,$E1,$A7,$2D,$8F,$F5,$1F,$E5,$CF,$89,$C4,$7F,$A6,$23
	.dc.b	$18,$AF,$FD,$4B,$18,$6D,$70,$AF,$C9,$FD,$CF,$E5,$17,$CB,$7F,$AB
	.dc.b	$23,$0D,$C3,$FD,$F8,$D9,$D9,$FE,$DD,$FC,$23,$2D,$32,$7F,$B6,$DC
	.dc.b	$B4,$C9,$FA,$4C,$7F,$47,$4F,$CA,$3D,$3B,$FE,$A1,$B1,$9A,$F8,$73
	.dc.b	$97,$FB,$17,$E1,$2F,$C2,$31,$02,$46,$6B,$94,$64,$46,$6F,$FD,$C8
	.dc.b	$D8,$FA,$3F,$59,$D1,$F4,$36,$98,$A3,$0D,$8E,$F3,$E3,$C5,$F7,$BF
	.dc.b	$96,$24,$7E,$59,$FE,$09,$C5,$F8,$98,$FC,$8F,$1F,$E4,$09,$7F,$EF
	.dc.b	$91,$BF,$F0,$A9,$7E,$23,$F6,$FB,$BE,$13,$63,$27,$FB,$BA,$78,$47
	.dc.b	$FB,$63,$E1,$37,$3F,$17,$D8,$7B,$49,$CC,$FC,$D9,$CC,$62,$3F,$DA
	.dc.b	$CD,$7F,$8F,$F4,$FE,$D0,$8F,$ED,$1F,$E3,$9D,$9F,$C1,$63,$F9,$9B
	.dc.b	$5F,$B5,$E0,$FF,$1D,$7F,$6C,$7C,$4D,$3F,$63,$AF,$C4,$8F,$D8,$BF
	.dc.b	$E1,$B8,$F8,$A7,$E0,$FF,$08,$FC,$D1,$27,$E1,$46,$5F,$CD,$47,$FB
	.dc.b	$EB,$9F,$11,$F9,$98,$C2,$7F,$85,$F0,$FD,$A8,$A7,$DB,$3C,$B5,$1F
	.dc.b	$2B,$5F,$F1,$D7,$17,$F1,$37,$F2,$8D,$7F,$11,$9F,$C8,$7E,$09,$F1
	.dc.b	$33,$F8,$CB,$3F,$8C,$FF,$19,$3F,$45,$C5,$F8,$B6,$E9,$F1,$DF,$F2
	.dc.b	$54,$65,$FE,$AB,$97,$AD,$F2,$B5,$FA,$07,$A8,$87,$29,$FE,$67,$1B
	.dc.b	$8F,$D0,$35,$4F,$72,$4F,$8E,$3F,$03,$DD,$FD,$11,$F5,$1E,$DC,$3F
	.dc.b	$09,$9F,$EC,$3E,$73,$EF,$72,$9A,$48,$FD,$34,$3E,$23,$FB,$9F,$CB
	.dc.b	$7F,$26,$6E,$1F,$CB,$9A,$FF,$2D,$6B,$FC,$5F,$E7,$3B,$92,$FA,$BF
	.dc.b	$E6,$7D,$BE,$8C,$FA,$8F,$AE,$7A,$73,$F1,$E5,$E5,$8F,$E6,$ED,$7D
	.dc.b	$F3,$C9,$FB,$D3,$F0,$DF,$E3,$96,$F9,$F7,$DB,$FE,$7D,$B1,$FC,$15
	.dc.b	$B3,$FE,$87,$8E,$13,$FE,$7F,$E6,$7E,$87,$FC,$8F,$F5,$2F,$F4,$64
	.dc.b	$62,$FF,$95,$46,$27,$51,$FC,$15,$AF,$F5,$3E,$7F,$A6,$4C,$3E,$8B
	.dc.b	$F1,$78,$7F,$2F,$A7,$B6,$BE,$21,$F9,$CA,$FE,$57,$D1,$F7,$08,$FD
	.dc.b	$A7,$F7,$9F,$83,$7D,$F9,$F8,$67,$C4,$70,$FF,$A7,$E5,$18,$7F,$A0
	.dc.b	$18,$DE,$CF,$D7,$3E,$8E,$1E,$63,$F1,$FF,$B7,$FF,$8B,$EB,$BC,$FF
	.dc.b	$8F,$23,$0F,$76,$C7,$66,$C7,$96,$BB,$BF,$E4,$4B,$F2,$39,$3A,$D7
	.dc.b	$95,$EE,$7C,$63,$CB,$C1,$F8,$37,$C5,$EA,$FF,$7B,$9F,$03,$63,$F0
	.dc.b	$F8,$23,$11,$C2,$7B,$CF,$E1,$9F,$47,$06,$94,$66,$79,$CE,$70,$FD
	.dc.b	$64,$7A,$1A,$FD,$16,$D7,$D0,$6C,$7C,$F2,$1F,$9E,$3F,$05,$5D,$DF
	.dc.b	$9D,$5E,$38,$C5,$79,$23,$D7,$3F,$14,$E6,$53,$A1,$F2,$0F,$D9,$E9
	.dc.b	$F9,$1F,$55,$F0,$BF,$E8,$7F,$63,$FD,$13,$5F,$EE,$C6,$CE,$BD,$2F
	.dc.b	$F0,$64,$62,$B8,$7A,$7F,$EE,$F8,$C4,$EA,$30,$4B,$FD,$DB,$FE,$F9
	.dc.b	$7F,$82,$37,$0F,$F3,$D3,$78,$3F,$CF,$CD,$DF,$CB,$5C,$8F,$F3,$5C
	.dc.b	$64,$FD,$F6,$BF,$C0,$9F,$CA,$69,$FE,$92,$6E,$BF,$EA,$93,$EF,$96
	.dc.b	$BF,$98,$C7,$37,$A9,$0F,$D7,$74,$7F,$8B,$63,$EF,$36,$7E,$F3,$67
	.dc.b	$67,$53,$ED,$C5,$3E,$D1,$6F,$D0,$F9,$2B,$F5,$9C,$1E,$98,$FC,$E9
	.dc.b	$B7,$15,$F9,$E6,$BF,$14,$F5,$9E,$93,$07,$EC,$78,$7C,$CD,$3C,$9C
	.dc.b	$3F,$C3,$6F,$F3,$53,$AB,$FD,$1E,$7D,$9A,$7A,$5C,$63,$FC,$1E,$BF
	.dc.b	$CD,$D8,$78,$0F,$BE,$78,$CA,$7F,$01,$7B,$BE,$2B,$F0,$61,$E3,$C3
	.dc.b	$FD,$07,$C5,$F6,$EA,$C7,$E1,$0F,$E5,$25,$EA,$FF,$1C,$3F,$88,$8E
	.dc.b	$E7,$8F,$7F,$C7,$BF,$16,$3F,$CB,$7C,$9D,$63,$F9,$E4,$73,$8C,$47
	.dc.b	$F9,$E5,$BC,$FF,$D1,$E1,$FB,$E7,$F4,$77,$FC,$E4,$D7,$27,$F4,$93
	.dc.b	$27,$F3,$4E,$9F,$CC,$37,$8B,$BE,$1F,$D6,$0F,$74,$EA,$7E,$E9,$0F
	.dc.b	$67,$1F,$EA,$1F,$F5,$4E,$DC,$B4,$FF,$03,$5F,$C5,$8C,$1E,$C9,$F8
	.dc.b	$A4,$D3,$E2,$BD,$67,$FC,$FF,$D5,$CC,$8F,$CF,$7B,$72,$7F,$C0,$0F
	.dc.b	$96,$FF,$07,$49,$A4,$AF,$2C,$78,$C9,$C5,$FE,$04,$7E,$C6,$1E,$29
	.dc.b	$F6,$F2,$63,$4F,$74,$7F,$98,$EF,$FC,$9A,$47,$FC,$69,$7F,$81,$F0
	.dc.b	$7F,$D2,$31,$7C,$0F,$DD,$3F,$E3,$E3,$07,$E1,$FC,$0F,$F6,$F5,$F3
	.dc.b	$8E,$C7,$47,$FC,$BE,$47,$93,$5D,$90,$FF,$28,$DF,$5F,$03,$E8,$67
	.dc.b	$ED,$9F,$A1,$97,$B1,$FD,$46,$5F,$EF,$BC,$3F,$DF,$3E,$0F,$A2,$7F
	.dc.b	$CC,$ED,$7F,$CC,$E7,$5C,$64,$FA,$63,$C2,$F9,$5F,$F2,$A9,$F0,$1C
	.dc.b	$72,$FF,$A8,$39,$BF,$D9,$5F,$82,$7F,$A0,$09,$5F,$ED,$F7,$FD,$C8
	.dc.b	$47,$FD,$16,$1C,$3F,$04,$FF,$30,$E9,$F8,$FD,$7E,$6C,$7C,$B3,$F8
	.dc.b	$26,$BF,$0C,$9A,$FC,$38,$FF,$55,$9F,$6F,$4F,$80,$E3,$3F,$8A,$FF
	.dc.b	$46,$CF,$FD,$6A,$BF,$20,$FF,$5F,$34,$FF,$5F,$78,$BF,$EB,$D7,$07
	.dc.b	$FA,$F5,$C4,$7F,$B2,$49,$1F,$EC,$9F,$23,$FD,$DD,$18,$8F,$76,$7E
	.dc.b	$C6,$E9,$FA,$DE,$0F,$C8,$E5,$F4,$67,$FD,$1E,$7D,$13,$FD,$DF,$C0
	.dc.b	$FC,$A3,$FC,$B1,$D1,$FE,$EB,$6F,$94,$9D,$57,$09,$FF,$7C,$F9,$B9
	.dc.b	$6C,$EA,$7F,$E0,$07,$98,$AD,$40,$CB,$87,$FC,$DF,$83,$2C,$47,$86
	.dc.b	$31,$7D,$63,$13,$A8,$4E,$B0,$FF,$64,$FC,$7C,$BF,$A3,$CD,$E1,$8C
	.dc.b	$47,$F9,$FB,$CD,$F7,$28,$FF,$3D,$46,$1B,$87,$19,$FC,$8C,$7C,$8F
	.dc.b	$43,$E3,$7E,$37,$68,$D2,$57,$F9,$93,$D4,$7A,$9C,$23,$0D,$7F,$D2
	.dc.b	$31,$AA,$E7,$1A,$6B,$F8,$07,$F6,$2F,$E7,$4F,$E6,$C7,$EF,$BB,$7F
	.dc.b	$8D,$1F,$9B,$67,$17,$E6,$88,$FD,$53,$EB,$7A,$BE,$18,$FF,$15,$36
	.dc.b	$34,$1C,$3C,$5A,$FF,$2A,$8F,$F0,$F3,$E4,$F5,$9E,$BF,$EB,$1B,$1F
	.dc.b	$DA,$5B,$3F,$E3,$96,$C7,$83,$D9,$7B,$3F,$D2,$DB,$97,$F7,$16,$DF
	.dc.b	$F9,$05,$B5,$FC,$03,$BF,$F7,$DE,$07,$F0,$0D,$3F,$BC,$76,$FE,$46
	.dc.b	$AF,$3F,$F0,$08,$C6,$3F,$E4,$B6,$CF,$C2,$F8,$67,$B7,$FA,$7F,$81
	.dc.b	$FF,$28,$46,$35,$46,$27,$55,$C5,$AF,$FB,$D2,$3F,$CA,$44,$7E,$7E
	.dc.b	$6D,$1A,$74,$9F,$A7,$EE,$B4,$FD,$76,$1F,$E6,$18,$C4,$EC,$FB,$CE
	.dc.b	$23,$53,$C2,$7E,$EF,$BC,$F5,$31,$1F,$85,$8C,$3C,$CF,$DE,$38,$AF
	.dc.b	$F3,$83,$63,$F0,$FF,$00,$F8,$0F,$F0,$13,$76,$FF,$14,$37,$1F,$F3
	.dc.b	$C3,$6B,$8D,$70,$25,$70,$6D,$70,$6C,$F0,$9D,$12,$74,$D8,$D4,$60
	.dc.b	$8E,$08,$7E,$96,$BC,$8D,$AE,$ED,$AE,$8D,$AE,$6D,$39,$9F,$C8,$EB
	.dc.b	$9B,$6B,$9B,$6B,$9B,$6B,$9B,$4E,$62,$B9,$B6,$B9,$B6,$B9,$B6,$B9
	.dc.b	$B4,$E7,$5C,$89,$5C,$9B,$5C,$9B,$5C,$9B,$5C,$AB,$81,$2B,$83,$6B
	.dc.b	$83,$67,$84,$E8,$93,$A6,$C6,$A3,$04,$70,$7F,$16,$F6,$3F,$D9,$8F
	.dc.b	$FB,$0D,$AE,$7D,$87,$B1,$FE,$C7,$84,$E9,$FE,$49,$F8,$BB,$E7,$3F
	.dc.b	$D1,$F1,$86,$BF,$D7,$63,$58,$F3,$6C,$7F,$B4,$DA,$FF,$2A,$6B,$CE
	.dc.b	$38,$4F,$48,$4E,$9E,$D1,$98,$C4,$6A,$30,$49,$E6,$48,$F1,$46,$5B
	.dc.b	$19,$7C,$91,$96,$C6,$5F,$31,$C0,$86,$8A,$68,$C1,$0C,$10,$F5,$BD
	.dc.b	$5A,$72,$29,$C4,$A7,$13,$E0,$3C,$5A,$71,$29,$C4,$A7,$19,$F9,$24
	.dc.b	$9F,$81,$19,$6B,$E0,$21,$C8,$B1,$C6,$7C,$4D,$8F,$13,$5F,$13,$5E
	.dc.b	$D1,$C2,$7B,$C7,$52,$4E,$DE,$F1,$A8,$C4,$6A,$30,$49,$E8,$48,$FE
	.dc.b	$F2,$7F,$4B,$F4,$BF,$C6,$98,$63,$89,$18,$9F,$EF,$F1,$87,$F4,$E6
	.dc.b	$23,$4F,$F4,$CC,$EB,$FE,$3F,$E1,$FE,$A3,$7F,$D3,$A3,$89,$B2,$9F
	.dc.b	$E4,$5E,$3F,$EB,$DC,$78,$84,$62,$7D,$2D,$8F,$EC,$B8,$FF,$64,$9F
	.dc.b	$EE,$DE,$9F,$EE,$E5,$CE,$8C,$36,$3F,$80,$35,$F3,$10,$FE,$DB,$97
	.dc.b	$3C,$BF,$DF,$D1,$FE,$5B,$E9,$7E,$3B,$FB,$76,$ED,$FC,$2D,$B3,$FE
	.dc.b	$13,$9F,$DE,$76,$3E,$6F,$CC,$D9,$FE,$24,$D7,$FC,$45,$1E,$1C,$64
	.dc.b	$7F,$17,$C7,$C6,$9B,$F1,$6F,$92,$FE,$14,$DC,$FF,$8F,$36,$BF,$06
	.dc.b	$D7,$FC,$85,$0F,$75,$F4,$DF,$E8,$3F,$B9,$FF,$20,$CE,$63,$15,$C2
	.dc.b	$B7,$2A,$DC,$61,$F9,$93,$F8,$12,$9F,$87,$CB,$F9,$97,$D4,$71,$EB
	.dc.b	$B5,$FE,$09,$AF,$CC,$CE,$5B,$19,$7F,$CA,$BE,$B3,$E0,$BF,$C4,$F0
	.dc.b	$F8,$9C,$C6,$1B,$D3,$FE,$95,$6C,$EC,$FF,$32,$57,$3F,$CA,$75,$18
	.dc.b	$B4,$62,$3F,$D3,$D3,$B7,$91,$89,$DB,$83,$FD,$5D,$A3,$E7,$3F,$D6
	.dc.b	$1F,$7B,$F2,$1B,$3F,$35,$AF,$FA,$AF,$7F,$AD,$18,$8E,$7E,$C5,$9F
	.dc.b	$B4,$DB,$34,$D7,$EA,$CF,$A8,$FF,$5B,$7C,$6F,$DE,$8C,$60,$8C,$4E
	.dc.b	$A3,$F0,$F3,$A9,$FE,$1B,$F8,$43,$EC,$B5,$5F,$E8,$D8,$C3,$CC,$FF
	.dc.b	$45,$B8,$9E,$33,$C2,$7F,$DA,$54,$F9,$1C,$66,$8C,$4F,$33,$E2,$38
	.dc.b	$BF,$55,$70,$CF,$FC,$A1,$C7,$F9,$4E,$5F,$CC,$E1,$F2,$E7,$2F,$E6
	.dc.b	$E3,$2F,$D6,$F9,$FB,$97,$E9,$5B,$46,$9C,$38,$BF,$25,$F1,$BF,$EE
	.dc.b	$1F,$C8,$7F,$22,$DC,$3F,$14,$D7,$F0,$CD,$53,$F9,$BF,$8B,$C3,$61
	.dc.b	$5F,$9C,$9C,$F2,$9E,$5F,$39,$FC,$AB,$AF,$FD,$E5,$B7,$FA,$EA,$31
	.dc.b	$87,$ED,$CF,$F6,$13,$88,$D7,$91,$FE,$8A,$7F,$98,$F9,$1F,$E9,$26
	.dc.b	$F6,$71,$BF,$F6,$55,$7F,$3A,$9C,$9D,$F9,$9F,$86,$7C,$CC,$51,$8E
	.dc.b	$2E,$58,$27,$35,$BA,$D5,$FE,$96,$FF,$57,$7F,$67,$A9,$18,$8F,$D6
	.dc.b	$C6,$3B,$38,$1F,$C5,$3A,$F1,$3F,$C6,$5A,$FE,$0D,$AF,$EB,$2B,$A1
	.dc.b	$FC,$97,$9B,$BF,$E1,$1F,$8B,$83,$49,$3F,$CB,$3E,$BB,$E5,$37,$87
	.dc.b	$E9,$B4,$F8,$34,$9D,$C0,$F8,$9B,$FF,$AF,$2F,$EA,$DE,$5B,$3E,$23
	.dc.b	$ED,$CE,$DF,$EC,$FC,$19,$7E,$BF,$0F,$FB,$32,$BE,$D9,$23,$F0,$7B
	.dc.b	$FD,$58,$FB,$5C,$1F,$02,$72,$D7,$CD,$C3,$F6,$34,$ED,$1F,$9A,$E3
	.dc.b	$FF,$7C,$AF,$C2,$3F,$99,$FB,$3E,$D6,$C6,$4F,$E5,$D1,$97,$BD,$65
	.dc.b	$EF,$97,$E2,$6B,$CB,$39,$9F,$C5,$4E,$5E,$63,$F0,$8E,$E4,$72,$F2
	.dc.b	$23,$E0,$6B,$F2,$35,$FF,$44,$6D,$FE,$18,$8F,$D1,$90,$FE,$8D,$F8
	.dc.b	$BF,$FA,$37,$A7,$FA,$A3,$74,$34,$C7,$D3,$FD,$FF,$98,$FE,$AE,$DB
	.dc.b	$FE,$A0,$D7,$FA,$63,$5F,$D0,$FE,$96,$FA,$5F,$F4,$AE,$C8,$C3,$FE
	.dc.b	$42,$31,$1A,$8C,$CF,$A4,$F7,$14,$FF,$1B,$72,$7F,$6C,$7F,$C0,$F0
	.dc.b	$F7,$39,$12,$FB,$46,$5F,$8F,$5C,$6B,$A9,$F9,$43,$B6,$9F,$DE,$7C
	.dc.b	$AF,$8C,$48,$FF,$2C,$C6,$5E,$BE,$97,$F3,$4C,$3F,$C7,$F8,$7E,$A9
	.dc.b	$BD,$DF,$E1,$86,$D7,$F7,$9F,$80,$FA,$44,$7A,$F0,$23,$FC,$95,$C3
	.dc.b	$F7,$67,$CC,$43,$D2,$7F,$7D,$FC,$2B,$EE,$11,$E6,$FF,$B3,$71,$FE
	.dc.b	$B1,$97,$EE,$FE,$9F,$F8,$BD,$7F,$58,$8C,$4F,$F6,$C6,$C6,$44,$FF
	.dc.b	$B4,$AF,$C9,$1E,$23,$EA,$F2,$7C,$E3,$D7,$E0,$FF,$8F,$23,$FC,$24
	.dc.b	$D7,$F6,$8D,$DB,$FC,$50,$E7,$1F,$3C,$B2,$F3,$BF,$DD,$32,$FE,$29
	.dc.b	$18,$9D,$B7,$C9,$FF,$38,$D7,$19,$F9,$0F,$98,$FC,$61,$FE,$7E,$E1
	.dc.b	$FC,$23,$F9,$AE,$EF,$E7,$77,$D7,$81,$18,$BF,$C8,$B6,$BF,$C4,$93
	.dc.b	$A9,$F0,$46,$3E,$37,$E7,$C7,$E6,$1B,$19,$8F,$F1,$63,$4F,$57,$B5
	.dc.b	$F0,$F2,$34,$91,$C0,$F6,$1E,$E3,$D9,$B7,$D3,$7F,$25,$7E,$07,$F6
	.dc.b	$ED,$BF,$DB,$37,$03,$B9,$E6,$7C,$8D,$FF,$02,$F8,$0F,$C1,$1A,$3C
	.dc.b	$4E,$97,$B8,$F7,$7F,$D7,$7C,$FF,$32,$7F,$A7,$4E,$7D,$BF,$EA,$13
	.dc.b	$99,$E3,$5C,$36,$FE,$A9,$18,$AD,$C6,$2F,$94,$FF,$05,$8C,$57,$18
	.dc.b	$EE,$D7,$91,$B3,$F3,$71,$FE,$0E,$23,$FD,$69,$AF,$F3,$36,$BC,$0F
	.dc.b	$BD,$9F,$E0,$DF,$BD,$1F,$8B,$CB,$FC,$0E,$D9,$D4,$F1,$3A,$1D,$C8
	.dc.b	$7F,$93,$B5,$FE,$E9,$3F,$86,$8E,$33,$FE,$55,$9C,$BF,$E6,$66,$C7
	.dc.b	$8A,$3C,$B3,$E1,$23,$F9,$C1,$F0,$DF,$A1,$8F,$F0,$EB,$63,$26,$DF
	.dc.b	$E0,$9C,$2D,$FE,$C5,$FA,$C8,$7E,$52,$FF,$CF,$DC,$5F,$E0,$DB,$FD
	.dc.b	$C3,$70,$FE,$25,$5D,$8F,$E2,$53,$FC,$0B,$93,$F8,$3C,$7C,$86,$DE
	.dc.b	$79,$FF,$D3,$93,$96,$C7,$9B,$1F,$2C,$F8,$CF,$9A,$79,$F0,$FE,$23
	.dc.b	$EC,$79,$23,$FD,$D1,$1E,$48,$C1,$DF,$97,$E9,$CF,$FA,$D5,$AF,$F3
	.dc.b	$53,$E1,$E3,$FE,$C5,$8C,$C7,$C4,$E3,$6B,$F2,$72,$FC,$5C,$F9,$4F
	.dc.b	$F5,$D7,$47,$0C,$3F,$D9,$CD,$D3,$FA,$44,$F6,$6D,$7F,$51,$DB,$FC
	.dc.b	$AB,$F7,$97,$53,$49,$7F,$EC,$DE,$BE,$84,$7A,$3E,$87,$C8,$7A,$73
	.dc.b	$F4,$3F,$2A,$BF,$C0,$1C,$DF,$9E,$8F,$EE,$ED,$7F,$93,$B5,$FF,$04
	.dc.b	$B5,$EA,$4D,$7E,$2F,$07,$D8,$6E,$7F,$DB,$7B,$39,$61,$FE,$0C,$3F
	.dc.b	$22,$58,$D3,$F1,$7B,$7F,$CB,$9F,$2F,$FE,$1E,$C4,$D2,$3F,$E2,$A3
	.dc.b	$E6,$78,$1F,$3C,$8F,$8F,$4F,$82,$FE,$E0,$87,$E8,$7C,$5F,$F2,$A7
	.dc.b	$E0,$DF,$E4,$62,$47,$F9,$31,$FF,$28,$91,$FF,$23,$B5,$FF,$27,$F8
	.dc.b	$FF,$E8,$EF,$C1,$BF,$C9,$E4,$8F,$F2,$AB,$FE,$5A,$23,$FE,$50,$6B
	.dc.b	$CB,$37,$E0,$DF,$E5,$82,$47,$F9,$79,$FF,$56,$91,$FF,$3C,$B5,$FF
	.dc.b	$37,$EF,$FC,$E6,$3F,$D0,$24,$8F,$F4,$0E,$DF,$E8,$98,$FE,$6F,$1F
	.dc.b	$EA,$6F,$2F,$FA,$07,$26,$30,$FF,$48,$5F,$FA,$3E,$31,$3A,$AF,$F4
	.dc.b	$B4,$7F,$A7,$EE,$47,$FB,$9E,$FD,$6D,$8F,$F5,$74,$3E,$A2,$92,$7F
	.dc.b	$D6,$11,$97,$FA,$5A,$F5,$1E,$2F,$FA,$F2,$BF,$D8,$04,$AD,$CE,$AB
	.dc.b	$94,$7C,$DC,$3F,$D0,$64,$7F,$D1,$2F,$FB,$60,$8F,$FA,$91,$AF,$CC
	.dc.b	$CB,$E1,$3F,$28,$FD,$B9,$63,$FD,$5E,$7F,$BC,$57,$E4,$14,$8F,$F6
	.dc.b	$6E,$CF,$F7,$C4,$7D,$25,$D6,$F5,$3F,$D7,$D3,$98,$C5,$74,$39,$38
	.dc.b	$8D,$4F,$09,$F8,$7E,$AF,$F0,$E6,$E5,$E8,$7F,$D9,$4F,$CD,$8D,$46
	.dc.b	$33,$46,$23,$FD,$D7,$0F,$47,$15,$FE,$E1,$3A,$4E,$A3,$33,$B9,$C8
	.dc.b	$F2,$5E,$B7,$FB,$6F,$E4,$FF,$21,$18,$C1,$18,$9D,$47,$E1,$27,$53
	.dc.b	$F9,$A8,$7A,$0F,$67,$71,$FE,$14,$8C,$3D,$0F,$D0,$B8,$9E,$4D,$8E
	.dc.b	$DE,$D7,$25,$8F,$A6,$30,$D8,$E8,$7C,$CC,$13,$AC,$9E,$B7,$2F,$82
	.dc.b	$FD,$19,$2F,$F4,$2F,$EC,$67,$F3,$4D,$E6,$E1,$3F,$9E,$6B,$F5,$DA
	.dc.b	$FF,$BB,$7D,$82,$F8,$FF,$C5,$46,$1B,$19,$3F,$17,$19,$7F,$1B,$E2
	.dc.b	$D7,$9F,$37,$E5,$A3,$11,$FE,$FB,$6C,$66,$3A,$1E,$2F,$7D,$EA,$FE
	.dc.b	$26,$3F,$C2,$71,$8E,$4F,$EE,$2D,$8C,$CE,$FA,$BE,$F7,$C1,$7F,$95
	.dc.b	$5A,$7F,$61,$71,$3F,$C7,$1A,$FE,$6F,$CA,$FC,$99,$38,$FF,$27,$C0
	.dc.b	$8F,$CF,$4B,$E0,$5F,$05,$FD,$85,$B7,$FE,$16,$6E,$3F,$91,$6D,$78
	.dc.b	$63,$C8,$47,$FC,$3B,$8F,$91,$FA,$BF,$06,$F4,$7F,$BD,$36,$3F,$C4
	.dc.b	$6D,$AF,$CF,$36,$3F,$9C,$FC,$23,$97,$FB,$B3,$72,$FF,$15,$B6,$FF
	.dc.b	$4C,$DA,$F3,$1E,$48,$F3,$1F,$31,$FF,$19,$FA,$0F,$85,$FF,$07,$36
	.dc.b	$33,$18,$3F,$C2,$5E,$3F,$F8,$13,$E2,$BF,$68,$D9,$FE,$9E,$D7,$FD
	.dc.b	$0E,$D7,$F2,$FA,$FF,$A4,$67,$C4,$48,$F3,$BF,$BF,$F3,$5F,$0F,$F9
	.dc.b	$E8,$C3,$67,$FA,$EC,$E9,$4E,$9A,$7F,$62,$F8,$8F,$F5,$DE,$BF,$E5
	.dc.b	$26,$E9,$FC,$35,$B3,$BF,$63,$FC,$49,$18,$8F,$F5,$0B,$5F,$D8,$B5
	.dc.b	$34,$91,$C8,$95,$E1,$6D,$78,$5B,$5E,$19,$D1,$27,$4D,$8D,$46,$08
	.dc.b	$E0,$87,$A2,$7A,$B6,$79,$36,$78,$B4,$E2,$7B,$A7,$8B,$67,$8B,$67
	.dc.b	$8B,$4E,$27,$C4,$9E,$2D,$9E,$2D,$9E,$2D,$38,$D7,$CA,$9D,$B6,$76
	.dc.b	$D9,$DD,$76,$8C,$47,$AC,$87,$94,$A7,$72,$BC,$1D,$10,$D1,$5D,$38
	.dc.b	$23,$82,$47,$8C,$95,$EA,$6D,$7A,$9B,$5F,$E3,$8F,$6F,$FD,$D6,$DF
	.dc.b	$6D,$EE,$7F,$63,$8F,$F7,$49,$97,$18,$FF,$70,$C6,$F8,$FF,$D7,$2B
	.dc.b	$F3,$CD,$7F,$48,$D7,$F3,$AD,$7F,$3A,$D7,$F3,$AD,$9E,$D1,$D2,$7B
	.dc.b	$91,$C1,$0E,$EE,$5A,$64,$F0,$B9,$69,$93,$C6,$E5,$A6,$4F,$2B,$96
	.dc.b	$99,$8F,$3B,$E0,$21,$D8,$A7,$42,$9C,$8A,$F2,$78,$10,$E0,$53,$81
	.dc.b	$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$27,$E9,$12,$7E,$04,$65,$B1,$99
	.dc.b	$F8,$4D,$7D,$6D,$7C,$CD,$7C,$6D,$7C,$4D,$9E,$D1,$D2,$7B,$91,$C1
	.dc.b	$0E,$EE,$5A,$64,$F0,$B9,$69,$93,$E3,$EF,$FE,$D4,$BF,$EA,$3A,$BB
	.dc.b	$BF,$4B,$5F,$D4,$A3,$13,$A9,$CC,$FF,$8A,$07,$C6,$3E,$33,$FC,$83
	.dc.b	$5F,$E4,$F8,$C6,$3C,$2B,$FC,$5F,$18,$9D,$57,$09,$FF,$57,$FD,$43
	.dc.b	$EA,$31,$1D,$E3,$18,$73,$84,$62,$FF,$CB,$CD,$8F,$EF,$3E,$07,$F7
	.dc.b	$72,$38,$3F,$D7,$B8,$F9,$A3,$C7,$66,$92,$7F,$D8,$7F,$84,$7F,$62
	.dc.b	$8F,$F3,$63,$63,$3C,$5F,$E6,$29,$CB,$63,$FD,$6F,$F0,$4F,$C4,$BF
	.dc.b	$83,$60,$8C,$3F,$E7,$73,$11,$A8,$CC,$F9,$88,$7A,$B5,$FE,$1D,$97
	.dc.b	$D0,$8F,$F5,$27,$93,$FE,$5E,$D6,$C7,$D3,$F1,$BF,$D5,$73,$98,$FF
	.dc.b	$53,$B5,$FF,$50,$E1,$FE,$76,$C9,$F8,$47,$FB,$86,$30,$DF,$2F,$FD
	.dc.b	$02,$D9,$DD,$7C,$0F,$A1,$CB,$63,$2F,$68,$CB,$63,$2F,$FA,$32,$DF
	.dc.b	$82,$FF,$A3,$BE,$A7,$C4,$D9,$F9,$AD,$7D,$2D,$79,$7D,$15,$BF,$A0
	.dc.b	$DB,$34,$C7,$E0,$FF,$E9,$CF,$07,$E0,$36,$FE,$DB,$5F,$8E,$D7,$FD
	.dc.b	$1B,$F3,$EF,$8A,$C4,$23,$18,$AF,$FD,$6D,$18,$6D,$73,$9F,$C3,$7C
	.dc.b	$55,$C8,$D2,$47,$3C,$3F,$6C,$BF,$20,$F1,$47,$EC,$E3,$17,$F7,$E3
	.dc.b	$FD,$73,$3A,$E3,$FD,$F4,$62,$75,$C7,$FE,$F2,$FC,$72,$FB,$A6,$92
	.dc.b	$BF,$1D,$0F,$AB,$FF,$57,$F9,$5F,$EF,$76,$CF,$FB,$D9,$AF,$FA,$AA
	.dc.b	$78,$29,$F1,$E7,$EB,$FB,$3D,$CF,$A1,$B1,$98,$E0,$7C,$2D,$13,$C3
	.dc.b	$E1,$FE,$86,$9F,$05,$B5,$FC,$5D,$BA,$FC,$26,$C7,$F2,$5F,$0B,$F8
	.dc.b	$97,$89,$FC,$AE,$31,$5F,$C9,$CE,$1F,$8C,$3C,$5F,$E3,$72,$C1,$39
	.dc.b	$D3,$C0,$DB,$E3,$7C,$27,$F8,$37,$47,$FB,$BB,$E9,$7F,$43,$06,$93
	.dc.b	$1F,$E1,$FD,$3B,$6C,$ED,$B3,$FB,$B9,$D4,$7F,$B9,$25,$F2,$0F,$9E
	.dc.b	$F1,$36,$3F,$14,$DC,$BE,$AB,$6B,$FA,$37,$37,$E1,$DF,$E5,$86,$39
	.dc.b	$3F,$10,$E0,$FC,$07,$37,$FB,$8D,$D3,$5F,$F7,$BE,$FF,$5B,$C5,$FA
	.dc.b	$9D,$5F,$9C,$6F,$26,$EB,$E7,$4E,$5F,$B1,$3F,$D6,$C8,$E0,$FE,$B9
	.dc.b	$5F,$53,$87,$F9,$C4,$78,$38,$BE,$B5,$7D,$4D,$BD,$13,$F6,$E7,$2F
	.dc.b	$B7,$DA,$F3,$B4,$FC,$01,$70,$FD,$6E,$AD,$7F,$E0,$04,$75,$24,$F9
	.dc.b	$2F,$F0,$D1,$88,$FD,$35,$97,$E0,$5D,$2D,$39,$8C,$36,$7C,$91,$EB
	.dc.b	$23,$DF,$36,$5D,$23,$0D,$7F,$DA,$B1,$FE,$09,$7F,$C0,$C4,$3F,$4C
	.dc.b	$7E,$13,$F0,$DF,$EB,$FC,$06,$92,$BF,$87,$CB,$4F,$89,$F0,$9F,$6D
	.dc.b	$B1,$F4,$1B,$B7,$BE,$7F,$01,$5F,$98,$F3,$FB,$91,$FE,$D9,$8A,$3F
	.dc.b	$CE,$9D,$1F,$B5,$E9,$FF,$42,$37,$2F,$51,$EE,$8E,$3F,$29,$EA,$26
	.dc.b	$18,$38,$EB,$F0,$EB,$E0,$90,$E4,$7E,$9B,$DB,$F8,$64,$9F,$2B,$F1
	.dc.b	$BC,$FF,$A3,$FB,$4F,$EB,$1B,$5F,$CF,$34,$FF,$11,$CE,$6C,$FE,$CB
	.dc.b	$F0,$6F,$B3,$F4,$56,$4D,$25,$F9,$4F,$C5,$6B,$F0,$5C,$07,$0F,$F0
	.dc.b	$5C,$1F,$6F,$F7,$09,$7F,$39,$FF,$0E,$FE,$84,$FA,$AF,$CB,$7D,$47
	.dc.b	$FA,$A6,$93,$9C,$13,$98,$C3,$6B,$8D,$79,$31,$F1,$DF,$A3,$7F,$A0
	.dc.b	$FC,$E3,$E3,$7D,$87,$FA,$E6,$A9,$FC,$76,$3D,$51,$96,$C6,$5F,$D0
	.dc.b	$E3,$E5,$8F,$14,$7E,$91,$FE,$DB,$19,$6E,$1F,$CA,$5F,$35,$F9,$2B
	.dc.b	$CD,$C5,$F8,$63,$F2,$7F,$73,$CE,$BE,$E3,$67,$E7,$35,$FD,$9E,$A8
	.dc.b	$C7,$E3,$2F,$3F,$FA,$55,$7F,$B5,$63,$18,$FB,$61,$18,$C3,$DC,$D8
	.dc.b	$E1,$F8,$9B,$ED,$BF,$C1,$31,$FC,$36,$53,$98,$E2,$D7,$8C,$F0,$7D
	.dc.b	$E7,$EA,$8A,$7E,$3E,$7C,$15,$FE,$8B,$6C,$77,$C3,$D7,$39,$6C,$7D
	.dc.b	$D2,$1D,$0A,$7E,$5F,$0F,$A0,$DC,$FE,$83,$E3,$1E,$F7,$F0,$3F,$1B
	.dc.b	$F9,$57,$CF,$BE,$03,$FD,$3D,$39,$A4,$E5,$FE,$61,$19,$8C,$47,$4E
	.dc.b	$CF,$F5,$5C,$62,$75,$5D,$27,$2D,$BF,$0B,$6F,$BB,$FC,$44,$FF,$59
	.dc.b	$11,$FF,$09,$35,$D1,$82,$1F,$2F,$1F,$E7,$71,$F8,$5C,$BF,$C2,$0D
	.dc.b	$8C,$E7,$F3,$DB,$5D,$08,$7E,$6A,$7C,$0F,$E3,$36,$FC,$24,$7E,$0A
	.dc.b	$3E,$55,$7D,$D9,$CB,$76,$FE,$B1,$5E,$18,$DD,$79,$DA,$FE,$63,$0F
	.dc.b	$3C,$79,$4C,$9F,$E1,$6D,$FF,$AB,$F5,$FF,$0D,$E1,$FD,$34,$A3,$13
	.dc.b	$A8,$FD,$79,$93,$E3,$DF,$B4,$A7,$E2,$70,$FF,$28,$F1,$FF,$17,$BF
	.dc.b	$E0,$64,$AF,$E0,$35,$CC,$FE,$17,$5E,$3E,$4F,$97,$1F,$BA,$6C,$FF
	.dc.b	$8B,$9C,$C7,$73,$C3,$CB,$FD,$56,$BE,$7D,$7F,$28,$E5,$FF,$00,$7C
	.dc.b	$77,$F1,$42,$47,$89,$F8,$5D,$1F,$80,$C3,$FA,$43,$5F,$E7,$4D,$7F
	.dc.b	$86,$B5,$F1,$1F,$80,$29,$F1,$4F,$E8,$5D,$5F,$E0,$6E,$EF,$E9,$C4
	.dc.b	$8F,$EA,$D1,$97,$FA,$C7,$6F,$F8,$33,$F6,$2F,$EB,$04,$8C,$E2,$8F
	.dc.b	$EC,$CF,$EB,$23,$F9,$0E,$DE,$78,$FF,$2A,$B7,$87,$ED,$6B,$F9,$4A
	.dc.b	$F3,$93,$6F,$94,$FF,$69,$EF,$E0,$4C,$3B,$1F,$E6,$07,$E1,$4F,$09
	.dc.b	$F8,$4D,$7E,$11,$FC,$D3,$A3,$C4,$4A,$E7,$87,$F8,$3D,$F5,$78,$3F
	.dc.b	$E1,$AF,$C0,$FF,$C3,$16,$69,$1F,$EF,$DE,$1F,$B4,$95,$D1,$F2,$F5
	.dc.b	$3B,$BE,$35,$7B,$E3,$1A,$FE,$DB,$C7,$FF,$2E,$60,$FC,$07,$FC,$B4
	.dc.b	$4B,$FF,$1E,$BF,$E4,$67,$47,$F9,$65,$C1,$1F,$F2,$1F,$E0,$9F,$E6
	.dc.b	$32,$38,$3F,$CC,$0F,$32,$38,$3F,$C9,$79,$FF,$9A,$C8,$E0,$FF,$57
	.dc.b	$3F,$E4,$4D,$FF,$C6,$CD,$CB,$E0,$37,$3F,$95,$5F,$E7,$AF,$2F,$FD
	.dc.b	$BF,$3F,$E8,$48,$C9,$FE,$8A,$8C,$BB,$F3,$3F,$CD,$53,$FE,$69,$F3
	.dc.b	$6C,$FF,$A4,$8E,$27,$F9,$57,$CC,$FF,$4C,$CE,$E3,$3A,$FF,$AA,$3C
	.dc.b	$EF,$F3,$4F,$C3,$7F,$A9,$1B,$3F,$E3,$96,$BF,$E2,$B6,$BF,$E7,$0F
	.dc.b	$61,$E8,$FF,$9C,$BD,$0F,$F5,$B6,$1F,$E7,$CA,$FF,$5D,$C6,$0F,$F6
	.dc.b	$19,$1F,$F8,$03,$4F,$4F,$94,$61,$AF,$FA,$77,$83,$C3,$C9,$F8,$6C
	.dc.b	$FF,$D3,$C4,$9F,$F7,$03,$F3,$C8,$E0,$F9,$41,$FA,$04,$72,$FF,$BE
	.dc.b	$BD,$5F,$F7,$9C,$3E,$23,$1E,$B7,$FA,$6D,$B1,$93,$FD,$F5,$19,$7F
	.dc.b	$D9,$5E,$C5,$7F,$D4,$5D,$7F,$EA,$B8,$C4,$7D,$76,$C6,$63,$A1,$FE
	.dc.b	$D0,$3D,$07,$DF,$C3,$EE,$11,$FF,$57,$1F,$EA,$4C,$1C,$9F,$EB,$19
	.dc.b	$F6,$36,$77,$BA,$B7,$1D,$8F,$F7,$73,$8A,$EB,$1F,$81,$F6,$7F,$D6
	.dc.b	$3F,$89,$E2,$DC,$09,$CD,$7D,$E6,$BF,$EF,$98,$D4,$AB,$FD,$5B,$ED
	.dc.b	$FA,$23,$15,$E0,$8C,$4E,$B2,$E7,$3F,$EC,$D9,$CB,$F8,$CF,$73,$37
	.dc.b	$E4,$6E,$06,$98,$E4,$FF,$65,$FD,$CF,$E6,$89,$7F,$96,$8C,$BF,$95
	.dc.b	$B4,$FC,$1B,$FC,$D4,$62,$75,$39,$E4,$E3,$1F,$07,$DE,$FF,$6C,$FE
	.dc.b	$6D,$C5,$B1,$FA,$46,$BB,$B4,$F7,$F8,$1C,$CF,$EC,$DB,$3F,$49,$B8
	.dc.b	$9D,$F1,$7F,$BE,$FE,$0B,$F0,$57,$F9,$4E,$C4,$62,$3F,$85,$36,$78
	.dc.b	$47,$E1,$4F,$F7,$E7,$C1,$BE,$13,$FB,$A3,$72,$FE,$7E,$DF,$17,$FD
	.dc.b	$D2,$DA,$FE,$03,$F0,$BF,$D9,$F4,$34,$97,$FD,$23,$E1,$DF,$0F,$FF
	.dc.b	$00,$3C,$2F,$F7,$9C,$62,$7F,$90,$E0,$8C,$4E,$AB,$84,$FE,$6B,$1F
	.dc.b	$34,$F9,$33,$CC,$FE,$AA,$72,$EE,$32,$87,$89,$FE,$6F,$AF,$D2,$C6
	.dc.b	$34,$FE,$6A,$FE,$D2,$74,$FE,$A9,$AF,$EE,$3E,$21,$E8,$7F,$2E,$84
	.dc.b	$63,$4F,$A1,$18,$9D,$4E,$7B,$7D,$55,$AC,$FE,$94,$60,$97,$FC,$8C
	.dc.b	$FE,$0F,$E4,$3C,$4F,$EE,$D3,$FC,$A2,$73,$18,$7F,$80,$FC,$4B,$E2
	.dc.b	$BF,$C6,$4D,$C7,$FB,$EB,$75,$F8,$6D,$AE,$35,$C0,$95,$C1,$B5,$C1
	.dc.b	$B3,$C2,$74,$49,$D3,$63,$51,$82,$38,$21,$FC,$9A,$BC,$8D,$AE,$ED
	.dc.b	$AE,$8D,$AE,$6D,$39,$9F,$83,$AE,$6D,$AE,$6D,$AE,$6D,$AE,$6D,$39
	.dc.b	$8A,$E6,$DA,$E6,$DA,$E6,$DA,$E6,$D3,$9D,$72,$25,$72,$6D,$72,$6D
	.dc.b	$72,$6D,$72,$AE,$04,$AE,$0D,$AE,$0D,$9E,$13,$A2,$4E,$9B,$1A,$8C
	.dc.b	$11,$C1,$0F,$A0,$43,$E4,$14,$F8,$05,$AF,$4B,$6B,$C8,$DA,$FF,$29
	.dc.b	$FB,$FF,$EE,$98,$FE,$BA,$D7,$FD,$BC,$63,$D0,$7B,$FF,$BE,$87,$E4
	.dc.b	$4F,$84,$61,$98,$C3,$63,$33,$FC,$1A,$72,$D8,$CB,$FD,$BA,$32,$D8
	.dc.b	$CB,$FD,$FA,$32,$D8,$CB,$FC,$56,$32,$D8,$CB,$E8,$6C,$6A,$31,$3E
	.dc.b	$72,$3E,$56,$BE,$36,$BE,$16,$BD,$DA,$F4,$39,$10,$E4,$53,$91,$4E
	.dc.b	$45,$39,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$91
	.dc.b	$F1,$C9,$3F,$28,$93,$F0,$23,$2D,$8C,$CF,$AE,$72,$D8,$CB,$E8,$8C
	.dc.b	$B6,$32,$F9,$E3,$2D,$8C,$BE,$88,$CB,$63,$2F,$A9,$B1,$A8,$C4,$F9
	.dc.b	$C8,$F9,$5A,$F8,$DA,$F8,$5A,$FF,$65,$E2,$69,$FD,$9F,$0F,$EC,$B1
	.dc.b	$89,$FD,$59,$FD,$8E,$74,$D3,$27,$F5,$2F,$3E,$BF,$DC,$78,$3F,$AC
	.dc.b	$46,$21,$18,$BF,$F5,$81,$FD,$90,$91,$FD,$EF,$0E,$1C,$3F,$E1,$79
	.dc.b	$FE,$F3,$C0,$F8,$CF,$EF,$DF,$0F,$86,$E1,$FE,$62,$9D,$29,$D3,$4F
	.dc.b	$F3,$3E,$3E,$68,$FF,$61,$12,$33,$3F,$EE,$58,$FF,$04,$F2,$7F,$7F
	.dc.b	$6F,$2F,$F0,$C6,$E7,$D6,$FF,$C1,$13,$97,$FC,$EB,$E8,$7F,$9E,$9F
	.dc.b	$F0,$9B,$5C,$F9,$1F,$C6,$54,$65,$F9,$C7,$FB,$5C,$87,$9B,$6F,$E4
	.dc.b	$1F,$73,$FE,$3D,$8F,$F0,$FB,$5F,$F4,$2B,$5F,$F4,$9B,$5F,$F6,$A7
	.dc.b	$C7,$FF,$8A,$7E,$3B,$FD,$17,$E0,$BC,$34,$E6,$BF,$D3,$73,$98,$C4
	.dc.b	$7F,$95,$DA,$FF,$B7,$FE,$6E,$CF,$36,$D1,$A6,$3B,$FF,$C7,$5F,$0B
	.dc.b	$E2,$9F,$F2,$D5,$A3,$13,$A8,$F1,$CE,$A7,$FD,$8D,$F1,$F9,$9F,$F5
	.dc.b	$3B,$5F,$F5,$93,$78,$3F,$D5,$CD,$F8,$57,$DA,$FE,$9E,$BF,$D6,$71
	.dc.b	$87,$FC,$9A,$62,$34,$FF,$94,$69,$3B,$EB,$FF,$28,$FC,$DF,$BE,$B8
	.dc.b	$9F,$EC,$47,$11,$AD,$39,$CF,$F9,$6E,$72,$F8,$BE,$37,$38,$76,$6E
	.dc.b	$06,$9C,$A7,$F1,$7F,$6D,$F8,$DE,$43,$49,$1F,$8A,$DF,$FA,$9F,$1F
	.dc.b	$E1,$60,$D3,$F1,$1C,$3F,$85,$FA,$5F,$F3,$87,$E0,$5F,$5D,$B5,$F5
	.dc.b	$1A,$FF,$A6,$DA,$A7,$F1,$DF,$2F,$86,$C2,$BF,$05,$39,$E4,$3C,$A7
	.dc.b	$F4,$FD,$1E,$D2,$3F,$E0,$3D,$3F,$54,$FF,$15,$EC,$74,$FF,$81,$FB
	.dc.b	$7F,$8B,$CF,$EB,$0F,$D7,$9F,$A8,$DB,$E1,$93,$8B,$E3,$1F,$E8,$C1
	.dc.b	$F9,$FB,$E1,$3F,$C1,$FF,$8F,$FF,$BC,$63,$12,$8C,$4E,$AF,$FD,$FF
	.dc.b	$5A,$9E,$1F,$19,$99,$FC,$64,$A3,$11,$FE,$1C,$8C,$5F,$58,$E7,$3A
	.dc.b	$7F,$02,$7C,$3E,$4F,$BE,$FF,$12,$7F,$88,$72,$7D,$DC,$BD,$D3,$FC
	.dc.b	$29,$FF,$0A,$ED,$F2,$A3,$E7,$9F,$37,$7F,$9E,$FC,$93,$E7,$6D,$F3
	.dc.b	$77,$F9,$F1,$93,$F1,$77,$D9,$F9,$D9,$FC,$C7,$C8,$79,$C8,$60,$F2
	.dc.b	$3B,$78,$98,$34,$7F,$30,$CB,$DE,$D3,$81,$FC,$8A,$7C,$4D,$9F,$45
	.dc.b	$C8,$F4,$6F,$F8,$67,$EB,$FC,$5F,$F8,$F2,$BF,$02,$DD,$8D,$31,$DD
	.dc.b	$FE,$52,$F1,$3F,$08,$DB,$FC,$0B,$5F,$BE,$D7,$EF,$75,$F5,$26,$1F
	.dc.b	$91,$C3,$F9,$53,$E7,$9F,$E8,$87,$A3,$5F,$A2,$FA,$23,$86,$3E,$56
	.dc.b	$9C,$63,$F4,$24,$3E,$D9,$F9,$BD,$3D,$C4,$8F,$13,$FE,$5E,$E7,$7E
	.dc.b	$75,$FE,$6C,$FA,$FF,$E9,$6D,$BF,$E9,$2D,$7F,$DF,$13,$C2,$FF,$DF
	.dc.b	$55,$FA,$9E,$7F,$F3,$CF,$95,$67,$FC,$F9,$7F,$A8,$9C,$B6,$32,$FA
	.dc.b	$4F,$F0,$8E,$7F,$09,$C7,$07,$DB,$7F,$9C,$7D,$8F,$FA,$1A,$BF,$78
	.dc.b	$D8,$10,$FF,$45,$B9,$71,$08,$C4,$7F,$43,$F5,$1E,$BE,$A3,$0D,$7F
	.dc.b	$D2,$B1,$A7,$FA,$51,$A2,$9F,$48,$FB,$A4,$3E,$B5,$78,$7E,$57,$FD
	.dc.b	$36,$F9,$2D,$18,$9F,$7B,$5F,$5B,$65,$5E,$39,$FF,$50,$C6,$2F,$D0
	.dc.b	$D8,$F2,$B5,$FF,$51,$E1,$F3,$5F,$41,$F4,$31,$FA,$4F,$CA,$F1,$7C
	.dc.b	$19,$1A,$4B,$F8,$3F,$2D,$FE,$01,$F5,$BD,$0D,$8C,$F9,$09,$CB,$63
	.dc.b	$8F,$CB,$3E,$57,$E8,$5A,$31,$96,$A7,$31,$88,$F5,$E7,$E8,$7F,$17
	.dc.b	$D7,$FA,$77,$DA,$2F,$C8,$FF,$B2,$1B,$87,$E2,$5B,$1F,$8C,$6B,$C8
	.dc.b	$EA,$7E,$E6,$7B,$38,$3F,$00,$6F,$D4,$7A,$9F,$E7,$8E,$2F,$F3,$D4
	.dc.b	$A7,$3D,$BF,$9A,$9C,$C6,$27,$F0,$BE,$9F,$E7,$63,$13,$AB,$E9,$39
	.dc.b	$6D,$7E,$92,$B7,$39,$BE,$CF,$FA,$14,$FD,$44,$FF,$44,$6B,$F2,$9B
	.dc.b	$C5,$FE,$88,$6E,$9F,$DD,$9B,$97,$B2,$B9,$1F,$E6,$18,$EB,$9F,$E0
	.dc.b	$DF,$A1,$BF,$E0,$5B,$C0,$DB,$FD,$42,$7D,$39,$D3,$CE,$E1,$E9,$29
	.dc.b	$F7,$77,$FC,$3D,$FD,$D6,$CF,$DD,$6C,$FD,$D6,$BF,$85,$AE,$EF,$E0
	.dc.b	$5A,$E6,$FC,$D7,$E8,$BF,$3C,$FF,$16,$E1,$CC,$7E,$F5,$B1,$98,$DB
	.dc.b	$FA,$D6,$BF,$3C,$F5,$94,$F7,$1F,$C3,$E7,$F7,$FC,$9F,$C4,$DB,$C9
	.dc.b	$FC,$89,$B3,$DA,$BF,$88,$17,$93,$F8,$7B,$78,$FC,$B3,$FC,$3A,$3C
	.dc.b	$7C,$1F,$7E,$3C,$6D,$7C,$67,$F2,$4C,$3C,$13,$FC,$BB,$1F,$23,$63
	.dc.b	$33,$E9,$8F,$E4,$07,$CA,$6B,$F8,$93,$F9,$B1,$97,$FA,$0B,$5F,$C6
	.dc.b	$9E,$23,$E3,$18,$3F,$C6,$1C,$BF,$8B,$EB,$FE,$A8,$48,$FE,$A4,$FF
	.dc.b	$60,$9F,$E9,$3F,$03,$64,$D2,$3F,$E4,$2E,$CF,$14,$FF,$93,$1B,$DA
	.dc.b	$69,$E7,$9F,$29,$33,$F9,$4F,$9E,$BF,$C2,$A7,$C1,$AF,$F0,$AB,$70
	.dc.b	$EA,$73,$AE,$35,$CF,$97,$F6,$4F,$E8,$AB,$FC,$BF,$DD,$E4,$23,$D5
	.dc.b	$F3,$F7,$FE,$BF,$D8,$FF,$0F,$C9,$A4,$AF,$F0,$A7,$17,$F9,$9F,$2F
	.dc.b	$1D,$79,$4F,$E0,$13,$E3,$E8,$B9,$F9,$8F,$3F,$89,$F1,$6B,$6D,$C3
	.dc.b	$FC,$8E,$4B,$FF,$21,$3F,$E4,$E2,$38,$3F,$CA,$6E,$DF,$F2,$23,$5F
	.dc.b	$F0,$1F,$89,$FE,$AA,$7C,$AD,$8E,$71,$FE,$07,$8C,$3D,$08,$74,$29
	.dc.b	$F8,$D3,$DD,$1F,$C8,$0F,$E4,$F0,$F1,$3F,$65,$3F,$CB,$23,$0F,$8A
	.dc.b	$32,$F8,$5A,$FF,$86,$AB,$E2,$3E,$03,$FC,$E7,$7F,$23,$83,$F1,$97
	.dc.b	$E8,$9F,$57,$89,$C9,$FF,$45,$E3,$EE,$7F,$D2,$5E,$2F,$F2,$DF,$C7
	.dc.b	$7F,$D3,$38,$9A,$49,$FF,$4F,$F9,$1B,$6C,$ED,$B3,$FE,$A6,$8C,$47
	.dc.b	$C7,$21,$EB,$CB,$F4,$59,$FB,$23,$F7,$1A,$FE,$17,$8F,$F3,$31,$FB
	.dc.b	$BD,$7E,$3E,$9F,$33,$6F,$64,$62,$7E,$64,$7C,$8F,$1B,$FD,$A7,$F9
	.dc.b	$D7,$FB,$4A,$4D,$23,$FE,$DF,$32,$57,$FD,$C8,$E0,$FF,$71,$14,$F5
	.dc.b	$69,$F4,$49,$87,$E3,$E3,$E7,$9F,$95,$3E,$99,$F9,$0E,$0F,$BB,$2F
	.dc.b	$39,$E7,$7F,$96,$3E,$17,$FB,$8C,$E6,$31,$7F,$E5,$8A,$DA,$AD,$C6
	.dc.b	$1F,$EE,$FE,$83,$A7,$8D,$7F,$98,$E3,$0F,$13,$FB,$FB,$89,$E3,$3C
	.dc.b	$27,$FB,$95,$3D,$2C,$63,$E9,$6C,$66,$7A,$1F,$E7,$29,$D3,$83,$FB
	.dc.b	$E7,$C3,$FF,$92,$57,$C2,$7C,$08,$C3,$7C,$4F,$F3,$DB,$67,$67,$F9
	.dc.b	$F7,$81,$E8,$FF,$2F,$8C,$9F,$E9,$67,$13,$FE,$1C,$9D,$F3,$F4,$AE
	.dc.b	$55,$A6,$E9,$FE,$8A,$F8,$2E,$13,$F2,$1B,$19,$9D,$88,$DF,$7E,$63
	.dc.b	$C4,$D5,$87,$A2,$30,$D8,$E4,$7E,$FF,$A7,$F4,$CF,$E4,$39,$3F,$1D
	.dc.b	$D3,$FA,$7E,$DF,$E5,$BD,$5F,$A8,$AF,$E2,$9D,$5F,$A5,$7F,$86,$1F
	.dc.b	$C3,$C9,$7F,$E3,$CF,$65,$EC,$7F,$8B,$FD,$FC,$36,$BF,$C8,$73,$AE
	.dc.b	$55,$69,$B3,$FE,$48,$CD,$93,$3C,$CF,$FA,$CE,$73,$18,$8F,$F5,$BC
	.dc.b	$62,$FF,$C9,$27,$F9,$42,$9E,$DF,$FA,$F2,$30,$DA,$FF,$2B,$CE,$9E
	.dc.b	$87,$8A,$34,$FF,$B0,$FD,$CF,$EA,$1E,$27,$F9,$71,$B1,$FE,$60,$6B
	.dc.b	$BE,$5F,$EC,$8F,$B6,$FD,$95,$FE,$5E,$24,$67,$2F,$ED,$0F,$F9,$AB
	.dc.b	$DA,$6F,$FE,$D4,$AF,$F3,$5C,$62,$76,$7F,$B5,$DC,$46,$A7,$84,$FF
	.dc.b	$9D,$3D,$D7,$AD,$FE,$0B,$6E,$FF,$DC,$1B,$5F,$E7,$86,$D7,$1A,$E0
	.dc.b	$4A,$E0,$DA,$E0,$D9,$E1,$3A,$24,$E9,$B1,$A8,$C1,$1C,$10,$FF,$7A
	.dc.b	$57,$91,$B5,$DD,$B5,$D1,$B5,$CD,$A7,$33,$FD,$FF,$5C,$DB,$5C,$DB
	.dc.b	$5C,$DB,$5C,$DA,$73,$15,$CD,$B5,$CD,$B5,$CD,$B5,$CD,$A7,$3A,$E4
	.dc.b	$4A,$E4,$DA,$E4,$DA,$E4,$DA,$E5,$5C,$09,$5C,$1B,$5C,$1B,$3C,$27
	.dc.b	$44,$9D,$36,$35,$18,$23,$82,$1F,$40,$87,$C8,$29,$F0,$0B,$5E,$96
	.dc.b	$D7,$91,$B5,$FE,$46,$F5,$3F,$DB,$91,$FE,$D0,$6B,$FE,$D5,$3F,$DA
	.dc.b	$2E,$1F,$BF,$B7,$FB,$15,$FC,$F3,$5F,$D3,$35,$FC,$7B,$5E,$33,$FE
	.dc.b	$DC,$7E,$FC,$65,$B1,$9B,$FE,$22,$D9,$D4,$62,$7C,$33,$92,$38,$21
	.dc.b	$E4,$72,$D3,$27,$99,$CB,$4C,$9E,$82,$9E,$62,$C7,$8D,$E8,$43,$91
	.dc.b	$4E,$45,$39,$15,$E4,$F0,$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38
	.dc.b	$23,$82,$4F,$CE,$24,$FC,$49,$E5,$18,$7E,$0B,$5F,$53,$5F,$2B,$5E
	.dc.b	$31,$CD,$F0,$C6,$5B,$19,$9F,$1B,$67,$51,$89,$F0,$CE,$48,$E0,$87
	.dc.b	$91,$CB,$4C,$9E,$67,$2F,$CE,$D3,$FA,$E7,$A9,$FD,$82,$0D,$23,$FD
	.dc.b	$3F,$23,$A3,$05,$3F,$A7,$3A,$8F,$E9,$D1,$93,$FB,$9C,$FF,$79,$8E
	.dc.b	$71,$FD,$E6,$30,$FF,$5F,$8C,$BC,$35,$E4,$BF,$15,$7A,$1E,$4F,$7B
	.dc.b	$73,$98,$C3,$CA,$76,$A7,$71,$87,$CE,$48,$F6,$B5,$FE,$F1,$97,$17
	.dc.b	$FB,$E7,$0A,$31,$1A,$9F,$F0,$14,$E5,$FF,$38,$FA,$4F,$C3,$B8,$7D
	.dc.b	$26,$1B,$7F,$13,$05,$6D,$56,$DB,$3F,$48,$F9,$E4,$30,$7C,$EF,$61
	.dc.b	$EF,$BD,$AF,$E2,$D1,$86,$C6,$67,$E5,$CE,$5F,$9B,$5E,$4F,$8C,$F8
	.dc.b	$07,$B0,$B1,$E9,$C5,$18,$8F,$F3,$C4,$62,$73,$1B,$F3,$74,$FC,$B9
	.dc.b	$46,$27,$E1,$B6,$33,$1C,$A7,$F2,$BE,$0B,$DE,$7D,$A7,$E6,$3F,$1D
	.dc.b	$FF,$3B,$46,$29,$18,$9D,$47,$FA,$06,$75,$3F,$63,$D2,$FB,$CD,$C3
	.dc.b	$8D,$7B,$E7,$2D,$8C,$BF,$E9,$7F,$DE,$7F,$CC,$6B,$D1,$FF,$59,$C6
	.dc.b	$1B,$E2,$FE,$B9,$B3,$B3,$FC,$84,$FE,$7B,$C4,$FF,$2F,$B7,$AF,$FC
	.dc.b	$C8,$DE,$AE,$15,$F4,$3D,$CF,$CB,$4F,$E4,$9B,$1F,$5D,$AE,$F2,$E2
	.dc.b	$FD,$3F,$85,$FC,$AB,$F6,$9A,$E6,$3F,$26,$E5,$AF,$DB,$F7,$77,$1F
	.dc.b	$9C,$6F,$11,$A7,$4E,$2F,$B5,$DB,$FC,$43,$ED,$FE,$CB,$70,$FA,$CD
	.dc.b	$7F,$03,$C7,$FB,$E9,$FD,$7F,$C6,$7E,$D5,$B9,$1A,$4A,$FE,$27,$E0
	.dc.b	$3F,$09,$7C,$A7,$F2,$0E,$EF,$E2,$73,$98,$C5,$70,$AD,$BF,$02,$74
	.dc.b	$FC,$21,$C9,$FE,$96,$BF,$AC,$48,$FF,$3F,$3D,$BD,$27,$DC,$7C,$EB
	.dc.b	$FF,$69,$B6,$33,$9F,$76,$D7,$E6,$ED,$F8,$A5,$23,$F9,$DE,$CF,$E7
	.dc.b	$71,$F9,$B2,$47,$E6,$5A,$FD,$BF,$BC,$FE,$33,$CB,$FC,$7D,$B8,$9A
	.dc.b	$7E,$BB,$A7,$E2,$D7,$D4,$6B,$FE,$D6,$E2,$FF,$5D,$BF,$86,$E4,$FD
	.dc.b	$C4,$EB,$6F,$93,$7F,$43,$83,$E6,$B7,$87,$FA,$9B,$70,$F1,$1F,$EA
	.dc.b	$77,$F4,$44,$70,$7D,$ED,$FE,$4C,$7C,$FE,$0F,$64,$E5,$AF,$F5,$B9
	.dc.b	$DB,$F7,$FE,$27,$FB,$1F,$DD,$F9,$5B,$19,$11,$96,$9D,$76,$AF,$5B
	.dc.b	$FE,$95,$3E,$39,$CC,$FF,$74,$BF,$BF,$3C,$FA,$24,$F7,$6B,$CE,$BF
	.dc.b	$A4,$BD,$E3,$A1,$1F,$F6,$61,$FC,$1C,$86,$0F,$E4,$5E,$CF,$F9,$6F
	.dc.b	$7F,$E5,$8D,$E0,$34,$FF,$30,$5B,$F7,$67,$BB,$FD,$97,$87,$FB,$DC
	.dc.b	$E6,$31,$8F,$F5,$FA,$DC,$FF,$81,$CD,$15,$7C,$87,$F9,$BE,$3B,$36
	.dc.b	$33,$1F,$D5,$9B,$3E,$C3,$FC,$1D,$DF,$FE,$75,$E9,$72,$A3,$0D,$9F
	.dc.b	$3D,$3D,$7E,$C4,$C3,$B3,$FE,$7D,$C3,$9E,$1C,$89,$7E,$13,$F4,$1F
	.dc.b	$13,$F2,$48,$FF,$86,$37,$FE,$B1,$1F,$2B,$1F,$F7,$8F,$D5,$78,$BD
	.dc.b	$4F,$F0,$73,$73,$34,$FF,$13,$DB,$F0,$26,$1F,$E3,$8B,$CC,$63,$1F
	.dc.b	$F7,$F9,$FE,$32,$AE,$15,$B8,$C4,$EE,$BD,$51,$FC,$6B,$5F,$8C,$FE
	.dc.b	$1C,$F9,$A7,$F6,$EF,$3B,$F0,$C4,$7D,$DC,$7F,$E2,$17,$FC,$17,$FB
	.dc.b	$D3,$A7,$FC,$9B,$F4,$3F,$EA,$FD,$93,$9A,$4E,$63,$0D,$AE,$35,$E3
	.dc.b	$AF,$F2,$41,$E8,$E1,$FA,$AF,$F8,$0B,$F1,$4F,$F1,$73,$FD,$21,$AF
	.dc.b	$F1,$56,$EF,$FE,$0B,$73,$5E,$4B,$FF,$31,$FC,$4F,$86,$CD,$24,$F8
	.dc.b	$E3,$E4,$71,$FE,$11,$F3,$47,$E9,$F2,$FF,$25,$57,$C1,$E3,$F7,$CF
	.dc.b	$43,$F4,$D3,$A6,$99,$AF,$F3,$AF,$A0,$F8,$F6,$7F,$ED,$BA,$FF,$21
	.dc.b	$46,$1F,$F3,$D9,$88,$D5,$72,$9D,$3F,$EE,$1A,$7C,$83,$EA,$3E,$3C
	.dc.b	$61,$B1,$99,$EA,$79,$1C,$3F,$EE,$9D,$FE,$9B,$F1,$5A,$FE,$28,$F2
	.dc.b	$63,$C9,$5F,$B7,$1F,$3C,$63,$1F,$3D,$70,$8E,$DC,$7F,$96,$7A,$10
	.dc.b	$FC,$29,$E1,$74,$E0,$D0,$F7,$3F,$CD,$1D,$9F,$EA,$2F,$2D,$ED,$7F
	.dc.b	$A6,$A7,$35,$FE,$94,$8C,$7A,$FF,$EA,$7A,$DC,$FF,$AA,$AB,$71,$88
	.dc.b	$ED,$3B,$AE,$37,$DD,$B3,$D4,$E5,$3B,$7F,$D4,$D1,$FE,$41,$23,$FE
	.dc.b	$AA,$D7,$F8,$AC,$62,$3F,$91,$B5,$E2,$7D,$07,$F0,$19,$FE,$1D,$BA
	.dc.b	$7F,$A7,$DB,$A7,$C2,$3B,$69,$FC,$AE,$3A,$5F,$E8,$1E,$DA,$FE,$1E
	.dc.b	$38,$1F,$85,$DF,$FD,$9B,$3F,$8F,$DF,$F1,$4D,$8C,$E1,$E5,$25,$7E
	.dc.b	$5F,$1F,$24,$62,$3F,$D6,$9C,$3F,$AA,$F9,$CF,$D6,$71,$7F,$AD,$DB
	.dc.b	$46,$9F,$88,$E4,$FA,$97,$F1,$0F,$8E,$7B,$9C,$CF,$F0,$38,$E4,$FF
	.dc.b	$AE,$79,$FF,$8C,$57,$F0,$52,$47,$FA,$37,$1F,$5D,$7F,$11,$3E,$19
	.dc.b	$FC,$44,$BC,$FF,$C4,$5C,$9F,$2C,$F1,$69,$F3,$B8,$7E,$EC,$78,$79
	.dc.b	$FF,$DE,$FF,$25,$FC,$54,$98,$E6,$51,$FE,$D9,$7F,$CD,$3D,$5F,$49
	.dc.b	$FF,$51,$35,$C9,$F1,$31,$FF,$57,$46,$63,$E3,$10,$F3,$1F,$20,$FF
	.dc.b	$5A,$F6,$7F,$51,$F9,$0F,$EA,$7D,$C9,$46,$1A,$7F,$52,$8D,$47,$F6
	.dc.b	$1F,$DA,$24,$FF,$82,$7B,$FF,$DE,$51,$88,$FF,$41,$D7,$97,$1F,$ED
	.dc.b	$B1,$E9,$EF,$FE,$C5,$BF,$FB,$12,$BD,$53,$E8,$3D,$5B,$FC,$C7,$D4
	.dc.b	$7E,$33,$C1,$FD,$13,$FE,$19,$6B,$FC,$B3,$9B,$FC,$11,$C1,$F8,$66
	.dc.b	$E1,$F5,$FA,$BF,$4D,$5F,$68,$87,$1F,$0B,$9B,$F9,$A6,$DF,$F8,$80
	.dc.b	$E8,$E2,$FF,$C5,$3F,$03,$FE,$18,$F1,$14,$8C,$36,$FF,$C6,$13,$A9
	.dc.b	$FF,$72,$F8,$BF,$1F,$6F,$F1,$ED,$7C,$78,$C1,$F1,$B5,$F8,$91,$F1
	.dc.b	$B3,$F5,$BF,$1B,$C1,$FF,$4B,$E4,$FB,$EF,$F2,$E1,$27,$FC,$C4,$FF
	.dc.b	$99,$48,$FF,$9A,$5A,$FF,$96,$FE,$FB,$FC,$C8,$47,$FC,$D4,$FF,$9B
	.dc.b	$48,$FF,$9C,$5A,$FF,$98,$FE,$F7,$FC,$D8,$47,$FC,$E4,$FF,$B9,$C8
	.dc.b	$FE,$F5,$AF,$E5,$6F,$FD,$77,$1F,$E8,$F2,$47,$FB,$07,$CE,$FC,$D4
	.dc.b	$7F,$A8,$61,$F7,$94,$8F,$F8,$BF,$5F,$F5,$0C,$7F,$AA,$A6,$4F,$E8
	.dc.b	$CF,$F7,$63,$87,$FB,$BF,$96,$F7,$3E,$EF,$E2,$DF,$E6,$C8,$C6,$A8
	.dc.b	$C4,$EA,$BA,$4A,$B7,$5F,$EC,$72,$4F,$FB,$14,$E4,$57,$FD,$29,$84
	.dc.b	$8F,$F4,$D3,$5F,$F3,$6B,$E9,$7F,$79,$BF,$F9,$92,$31,$1F,$E2,$26
	.dc.b	$C6,$63,$E5,$1F,$20,$B1,$FE,$BD,$23,$EA,$A7,$B1,$F9,$EF,$C0,$7F
	.dc.b	$DF,$A4,$9D,$46,$5F,$A1,$E6,$7F,$0F,$9F,$F3,$3E,$A8,$C4,$E8,$4E
	.dc.b	$A3,$31,$F4,$BD,$AF,$D6,$E3,$F4,$5B,$1E,$56,$BB,$D3,$D8,$FD,$AF
	.dc.b	$5B,$2F,$ED,$B3,$46,$27,$EB,$36,$33,$1C,$A7,$87,$B7,$FB,$EF,$C0
	.dc.b	$3F,$04,$D8,$F8,$CD,$7F,$12,$D7,$EA,$F3,$7F,$A8,$67,$E9,$B6,$3F
	.dc.b	$AF,$46,$3C,$A3,$FE,$BA,$F7,$DE,$FF,$F3,$98,$C3,$6B,$F1,$F3,$A7
	.dc.b	$89,$88,$D3,$F9,$CE,$4F,$E2,$FF,$8E,$FE,$63,$33,$49,$3F,$95,$7E
	.dc.b	$CF,$BB,$FD,$25,$B1,$C0,$D1,$F8,$7F,$82,$FE,$47,$F6,$3F,$A7,$6C
	.dc.b	$66,$D3,$96,$C7,$E3,$FE,$13,$6F,$EC,$9A,$FD,$96,$C7,$EB,$5C,$FA
	.dc.b	$AF,$85,$7B,$1F,$D9,$E3,$0D,$8C,$DF,$F8,$42,$72,$FF,$84,$87,$C5
	.dc.b	$FF,$92,$BC,$8F,$D5,$B6,$3F,$60,$D7,$F8,$53,$5F,$C7,$7A,$1F,$8B
	.dc.b	$26,$5F,$8B,$3F,$82,$3F,$94,$87,$C6,$3E,$27,$FA,$5F,$C6,$FE,$EA
	.dc.b	$31,$68,$C4,$EA,$BF,$81,$56,$A7,$CF,$8F,$F7,$F6,$CF,$13,$D3,$19
	.dc.b	$69,$93,$F9,$27,$C5,$FE,$DD,$C1,$FE,$09,$71,$3F,$C3,$9B,$1F,$C4
	.dc.b	$DA,$FF,$2D,$F0,$BF,$DE,$FE,$57,$E7,$49,$1F,$E3,$D7,$CA,$47,$C6
	.dc.b	$D7,$F9,$B7,$53,$E3,$7F,$C3,$53,$FC,$7A,$31,$3F,$DB,$CD,$CE,$9F
	.dc.b	$E4,$D1,$A9,$FE,$4B,$F1,$BC,$F5,$FC,$03,$73,$CE,$1F,$E0,$4C,$13
	.dc.b	$97,$F9,$83,$5F,$E9,$4D,$7F,$C5,$CD,$78,$C7,$38,$C1,$1C,$12,$BC
	.dc.b	$6D,$AF,$1C,$60,$8E,$09,$5E,$49,$E0,$D9,$DB,$67,$6F,$A2,$76,$D9
	.dc.b	$DB,$67,$6F,$B2,$76,$D9,$DB,$67,$6F,$C0,$8C,$B6,$33,$5F,$06,$72
	.dc.b	$D8,$CD,$76,$AE,$07,$A4,$87,$94,$A7,$84,$A7,$03,$44,$34,$53,$46
	.dc.b	$08,$60,$93,$E5,$7A,$B5,$E4,$D7,$8B,$5E,$31,$EB,$8C,$11,$C1,$2B
	.dc.b	$E1,$B6,$BE,$1C,$60,$8E,$09,$5E,$49,$E0,$FF,$81,$7E,$03,$FD,$CE
	.dc.b	$DE,$37,$95,$FE,$06,$8F,$F7,$1B,$67,$71,$F9,$CF,$B2,$FF,$4C,$3F
	.dc.b	$A0,$6B,$9C,$79,$64,$9C,$CF,$E8,$9A,$FE,$A9,$B1,$B9,$E3,$18,$23
	.dc.b	$82,$4F,$EE,$23,$2D,$8C,$BE,$56,$BE,$53,$44,$34,$53,$46,$08,$60
	.dc.b	$87,$AD,$EA,$D3,$91,$4E,$25,$38,$9F,$05,$E2,$D3,$89,$4E,$25,$38
	.dc.b	$CF,$C7,$76,$D7,$6D,$77,$3F,$19,$CB,$5C,$CF,$8A,$78,$47,$9C,$87
	.dc.b	$84,$AE,$9C,$11,$C1,$27,$B4,$70,$6B,$B6,$C6,$E7,$C5,$18,$23,$82
	.dc.b	$4F,$C0,$8C,$B6,$32,$FB,$9A,$FB,$8D,$10,$D1,$FE,$66,$FC,$BD,$F4
	.dc.b	$FF,$E9,$18,$FF,$36,$46,$1F,$D3,$98,$8D,$3F,$E8,$98,$D6,$3C,$F9
	.dc.b	$33,$3F,$EE,$F9,$46,$21,$18,$7A,$9C,$E3,$5F,$A7,$3E,$5A,$92,$76
	.dc.b	$7F,$BE,$E3,$4E,$07,$13,$89,$E1,$E6,$30,$D8,$CE,$5F,$E5,$13,$FC
	.dc.b	$1B,$A7,$3E,$6F,$F0,$8C,$60,$98,$7F,$A4,$7E,$4F,$FC,$AB,$5E,$16
	.dc.b	$CF,$FA,$49,$B3,$BF,$0B,$FD,$CF,$3F,$E9,$9A,$7A,$BC,$AB,$0D,$7F
	.dc.b	$D5,$6D,$7F,$E0,$05,$7F,$0C,$C3,$FC,$4B,$1E,$82,$1F,$E2,$4E,$AF
	.dc.b	$E2,$7E,$CB,$CE,$78,$1F,$5A,$30,$D8,$CD,$7C,$E9,$CB,$61,$E1,$21
	.dc.b	$F8,$6F,$FA,$7A,$30,$D8,$C9,$FE,$D3,$8C,$BF,$ED,$5E,$CA,$F3,$BF
	.dc.b	$34,$62,$11,$87,$99,$BB,$4E,$F5,$F9,$7F,$0A,$5F,$89,$FC,$C8,$C3
	.dc.b	$63,$37,$F3,$A7,$2F,$D0,$EE,$AF,$BE,$7E,$BC,$62,$3B,$70,$7F,$8A
	.dc.b	$EF,$EC,$CE,$69,$F3,$4F,$84,$F9,$1F,$8E,$FE,$06,$31,$8A,$31,$3A
	.dc.b	$8F,$F7,$5C,$EA,$7F,$D4,$C4,$8F,$F7,$17,$47,$E7,$23,$11,$E7,$CB
	.dc.b	$FD,$1F,$87,$58,$FC,$5F,$91,$F1,$6B,$F1,$ED,$8C,$CE,$DF,$9D,$E5
	.dc.b	$FE,$14,$8F,$E0,$DF,$C4,$7E,$0D,$23,$F0,$CD,$E8,$FC,$F7,$C4,$FC
	.dc.b	$9D,$6D,$7D,$56,$E1,$F0,$A3,$E7,$4E,$DF,$C7,$1F,$6D,$FD,$DC,$6E
	.dc.b	$31,$D9,$CD,$DC,$27,$75,$E2,$AD,$57,$E6,$6D,$FA,$4B,$F0,$2F,$E3
	.dc.b	$31,$86,$C6,$67,$F8,$8C,$E5,$FE,$15,$52,$3F,$85,$47,$D4,$EE,$FE
	.dc.b	$1A,$FA,$87,$AC,$FC,$07,$F8,$1C,$66,$31,$A7,$F1,$B6,$DF,$19,$E3
	.dc.b	$5C,$1F,$D1,$7A,$9F,$B3,$FC,$77,$B7,$5F,$EF,$E3,$13,$F2,$A3,$0A
	.dc.b	$77,$D6,$AD,$CF,$EB,$7D,$7A,$FE,$CA,$51,$8B,$FE,$18,$D8,$CC,$F5
	.dc.b	$3E,$9F,$0F,$CD,$D3,$F9,$C9,$F5,$78,$3D,$53,$F7,$5B,$19,$3F,$17
	.dc.b	$0F,$9C,$F3,$F1,$FF,$11,$6C,$64,$F7,$60,$9D,$47,$E2,$B6,$FB,$15
	.dc.b	$F4,$DB,$19,$E2,$FB,$D3,$96,$C7,$DB,$E2,$EC,$FF,$AF,$78,$BF,$A6
	.dc.b	$D7,$D5,$E8,$FE,$9A,$D8,$CC,$FD,$DE,$2F,$E2,$93,$F6,$F8,$FE,$FB
	.dc.b	$63,$33,$BD,$7B,$BF,$E5,$8E,$1F,$F5,$FF,$DA,$7A,$5B,$1E,$D6,$BF
	.dc.b	$9D,$6B,$FC,$DF,$0F,$C6,$CF,$7A,$34,$93,$CF,$AF,$F9,$F7,$DF,$77
	.dc.b	$6C,$64,$46,$5A,$7A,$5E,$C4,$70,$7C,$B3,$E3,$6B,$FE,$73,$D7,$FB
	.dc.b	$AC,$60,$FE,$13,$F8,$4F,$F9,$B3,$1F,$DE,$37,$90,$D3,$FA,$FF,$1B
	.dc.b	$DC,$FE,$B8,$4B,$FE,$B4,$FF,$85,$AD,$F5,$5F,$E0,$F7,$F5,$C6,$3C
	.dc.b	$BF,$04,$66,$BF,$BC,$35,$F5,$10,$FC,$26,$7E,$BE,$3F,$F0,$8C,$7F
	.dc.b	$22,$E2,$F4,$13,$0E,$FA,$7F,$00,$9F,$80,$73,$3F,$B7,$6B,$F1,$1C
	.dc.b	$4F,$F1,$DF,$90,$F8,$44,$79,$BF,$17,$E5,$7F,$B1,$7E,$83,$FC,$C9
	.dc.b	$B5,$FB,$46,$BF,$CC,$1A,$FF,$2F,$FA,$BF,$ED,$5F,$05,$FC,$E9,$B1
	.dc.b	$FB,$C6,$BF,$CE,$1A,$FA,$34,$F8,$2F,$AC,$FF,$1A,$B8,$7E,$39,$FD
	.dc.b	$1B,$EC,$FE,$E1,$32,$F8,$AF,$E0,$7E,$05,$F1,$DF,$E1,$1F,$AA,$FF
	.dc.b	$5A,$D2,$73,$69,$CC,$61,$B5,$C6,$BC,$9B,$7C,$DB,$F1,$9F,$3B,$77
	.dc.b	$DB,$7F,$82,$E3,$BC,$62,$BF,$C0,$91,$FD,$62,$75,$1F,$C8,$1B,$1F
	.dc.b	$E0,$9C,$7C,$B3,$E2,$8F,$F0,$B3,$FD,$62,$32,$DC,$3F,$C2,$0F,$9B
	.dc.b	$3F,$91,$5F,$37,$8B,$F1,$C7,$E2,$BE,$53,$9D,$7F,$30,$6C,$FF,$6C
	.dc.b	$6B,$FD,$AD,$BF,$04,$FA,$47,$DF,$7A,$29,$39,$8C,$36,$FE,$0C,$EA
	.dc.b	$B8,$C6,$25,$FA,$EB,$D1,$FE,$EB,$7F,$E4,$99,$CC,$62,$3B,$CE,$DE
	.dc.b	$4D,$7F,$DE,$3C,$1F,$5D,$F8,$47,$EA,$0F,$B2,$75,$C3,$C6,$FF,$BE
	.dc.b	$AF,$D0,$D8,$CE,$1F,$E0,$E9,$CB,$F3,$F3,$F9,$B5,$D4,$EE,$64,$C0
	.dc.b	$F8,$07,$C0,$FF,$AA,$39,$BF,$CA,$B3,$FE,$67,$9C,$F8,$FF,$E6,$99
	.dc.b	$CC,$F1,$AE,$1E,$A7,$F8,$82,$F7,$2A,$DC,$61,$B5,$C2,$7B,$35,$E9
	.dc.b	$7D,$EF,$F8,$A9,$FC,$C0,$91,$FD,$75,$AE,$8C,$10,$F9,$9B,$FF,$9B
	.dc.b	$63,$E5,$37,$2F,$95,$A7,$E0,$A7,$31,$FE,$76,$39,$57,$E6,$63,$BB
	.dc.b	$4F,$A5,$B7,$E0,$63,$EC,$36,$73,$BF,$FA,$46,$72,$D8,$F0,$E1,$E3
	.dc.b	$BF,$3C,$78,$5B,$5E,$43,$F7,$7C,$1F,$E8,$7F,$BE,$B4,$69,$1F,$4D
	.dc.b	$7F,$1A,$DF,$EC,$BE,$93,$E4,$1F,$BF,$E3,$FF,$53,$57,$EF,$C9,$5F
	.dc.b	$BD,$9E,$CF,$F0,$73,$CC,$FF,$97,$F9,$7D,$4D,$8C,$9F,$E5,$D8,$CC
	.dc.b	$78,$F1,$F9,$6D,$3E,$7C,$FF,$25,$BF,$F5,$C7,$4F,$F1,$12,$47,$8D
	.dc.b	$FE,$3B,$D3,$F7,$6B,$FC,$72,$D7,$FC,$EA,$D7,$FC,$E8,$D7,$C0,$7C
	.dc.b	$23,$E2,$12,$7F,$CE,$93,$FE,$B5,$D3,$FA,$64,$7F,$10,$9F,$E9,$8D
	.dc.b	$7F,$A6,$35,$FF,$41,$75,$7F,$AF,$7F,$2A,$FE,$B9,$26,$92,$7F,$CA
	.dc.b	$3D,$BE,$79,$FE,$D0,$7D,$82,$E7,$FE,$96,$8F,$57,$43,$A5,$C1,$F7
	.dc.b	$B0,$F6,$9E,$C2,$E1,$FE,$8F,$EE,$ED,$D9,$FE,$7A,$CB,$FC,$01,$3D
	.dc.b	$E3,$15,$F8,$AC,$7B,$61,$D5,$B7,$FA,$4B,$FF,$09,$1E,$7E,$FF,$F0
	.dc.b	$B4,$7F,$84,$BC,$0F,$64,$63,$0F,$F3,$0F,$83,$FE,$F5,$F8,$4F,$F3
	.dc.b	$EB,$63,$22,$32,$D3,$FC,$EF,$E2,$7A,$89,$8F,$A5,$F8,$BB,$7C,$4C
	.dc.b	$BF,$1F,$9F,$A2,$FF,$4B,$E4,$7F,$A6,$BE,$FB,$FC,$9C,$48,$FF,$2E
	.dc.b	$3F,$E4,$72,$3F,$E4,$B6,$BF,$EC,$08,$FF,$51,$FD,$F7,$F9,$5C,$91
	.dc.b	$FE,$65,$7F,$CA,$04,$7F,$CA,$8D,$79,$70,$B3,$7C,$47,$F9,$84,$93
	.dc.b	$FE,$6F,$7F,$CA,$64,$7F,$D0,$0D,$7F,$D9,$1E,$77,$FB,$0E,$3F,$D0
	.dc.b	$DE,$82,$7A,$9F,$E8,$B7,$11,$A8,$C9,$BF,$39,$E5,$FF,$A5,$E3,$27
	.dc.b	$FA,$BF,$D0,$E2,$FF,$A7,$1A,$F1,$8E,$15,$FE,$EE,$BF,$77,$15,$CB
	.dc.b	$EE,$5F,$FB,$B2,$72,$FF,$AD,$1A,$FF,$AF,$C7,$E3,$FF,$EC,$4F,$4B
	.dc.b	$D8,$D8,$C9,$FE,$AE,$8C,$BF,$EB,$0F,$4E,$BF,$ED,$43,$65,$9F,$F6
	.dc.b	$7C,$6A,$3F,$D2,$9B,$78,$9B,$87,$A1,$B8,$FF,$A5,$CF,$F4,$39,$1F
	.dc.b	$F5,$A3,$5F,$AB,$8F,$E7,$DA,$73,$F2,$3F,$DF,$13,$F4,$CF,$F7,$A5
	.dc.b	$3A,$FF,$7E,$BE,$DB,$31,$86,$F6,$7E,$2D,$B3,$BC,$3F,$DB,$E6,$4A
	.dc.b	$E5,$FC,$33,$DD,$D1,$8A,$CC,$FF,$B7,$2E,$4F,$E0,$4F,$89,$C3,$F8
	.dc.b	$8E,$1F,$74,$7F,$BD,$7D,$7B,$3F,$8C,$6F,$11,$A7,$0C,$9E,$D3,$D9
	.dc.b	$FD,$37,$E3,$DF,$12,$31,$28,$C4,$EB,$1F,$D0,$1F,$AD,$AD,$CF,$FA
	.dc.b	$F0,$91,$FE,$E9,$6B,$F0,$BD,$C7,$B7,$FA,$99,$FA,$71,$89,$F0,$1D
	.dc.b	$B0,$46,$67,$75,$C6,$B5,$6F,$63,$F8,$07,$E2,$BE,$B6,$CE,$9A,$FD
	.dc.b	$49,$E1,$8A,$BE,$17,$9F,$AA,$F7,$36,$7E,$4B,$6B,$FC,$BD,$1C,$67
	.dc.b	$E2,$ED,$FB,$EF,$03,$F8,$37,$E8,$1F,$22,$53,$95,$19,$8C,$37,$83
	.dc.b	$FC,$96,$DF,$80,$E7,$5D,$A5,$18,$8F,$E0,$CD,$7B,$C6,$9E,$E7,$1F
	.dc.b	$82,$FC,$39,$1F,$C7,$C2,$3F,$1F,$0F,$84,$7C,$27,$F8,$05,$B9,$FF
	.dc.b	$67,$6E,$7E,$FA,$E1,$5F,$DC,$F4,$7C,$2D,$26,$1F,$DD,$63,$2F,$F7
	.dc.b	$1F,$86,$BD,$1C,$DB,$87,$CB,$6B,$F9,$46,$BF,$CD,$72,$F2,$CF,$8D
	.dc.b	$B1,$F4,$DA,$FE,$71,$AF,$FA,$BE,$3F,$87,$34,$FE,$22,$7D,$BF,$87
	.dc.b	$79,$BF,$19,$B1,$FC,$6C,$38,$9F,$88,$D8,$FC,$27,$C4,$BE,$27,$E7
	.dc.b	$B6,$39,$36,$BE,$63,$63,$8F,$C5,$70,$23,$C1,$FB,$3F,$18,$F8,$CF
	.dc.b	$F1,$C3,$75,$FF,$0F,$37,$87,$EC,$36,$FF,$30,$D7,$F8,$0B,$5F,$EC
	.dc.b	$4D,$4D,$25,$78,$09,$5E,$06,$D7,$81,$B5,$E0,$6B,$E0,$6B,$E0,$6B
	.dc.b	$E0,$6D,$78,$2B,$B1,$2B,$B3,$6B,$B3,$5E,$CD,$7B,$35,$EC,$D9,$ED
	.dc.b	$3D,$09,$3D,$1A,$F4,$6B,$D1,$AF,$46,$C7,$48,$E4,$43,$91,$4E,$45
	.dc.b	$39,$15,$E4,$F0,$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82
	.dc.b	$57,$D4,$25,$7C,$96,$D7,$C0,$6D,$7A,$5A,$F9,$1A,$F8,$5A,$F8,$7E
	.dc.b	$07,$FD,$D2,$FF,$2F,$6B,$FE,$ED,$F8,$07,$8B,$FC,$56,$3F,$DC,$E7
	.dc.b	$F1,$7F,$15,$EC,$FF,$92,$63,$0D,$7F,$9D,$46,$B2,$E4,$D8,$DC,$62
	.dc.b	$78,$47,$F6,$89,$D3,$CA,$32,$FF,$92,$E3,$2E,$E3,$2F,$FA,$51,$B1
	.dc.b	$A8,$E1,$18,$24,$F8,$49,$1E,$42,$1E,$17,$2D,$32,$78,$5C,$B4,$C9
	.dc.b	$E0,$72,$D3,$27,$9D,$CB,$4C,$CF,$A5,$C1,$E1,$21,$DC,$A7,$52,$9C
	.dc.b	$0D,$10,$D1,$4D,$18,$21,$82,$3E,$32,$4F,$CA,$8F,$3B,$5F,$0C,$6A
	.dc.b	$7D,$AD,$8D,$C6,$27,$84,$78,$E7,$4F,$28,$CB,$E4,$8C,$BB,$8C,$BE
	.dc.b	$66,$C6,$A3,$84,$60,$93,$E1,$24,$79,$08,$78,$5C,$B4,$C9,$E1,$72
	.dc.b	$D3,$FA,$8E,$FF,$EA,$1B,$B1,$FA,$C3,$FA,$DC,$65,$A6,$4F,$F2,$4F
	.dc.b	$C7,$34,$D4,$61,$AF,$FA,$B6,$34,$F1,$31,$1A,$5E,$97,$FB,$B2,$FF
	.dc.b	$C2,$0D,$8F,$F1,$83,$5F,$F2,$CB,$5F,$F0,$37,$47,$FB,$1E,$7F,$C0
	.dc.b	$CD,$8F,$F0,$33,$5F,$F0,$33,$5F,$F6,$17,$B9,$E6,$7F,$C4,$0D,$7F
	.dc.b	$C8,$0D,$7F,$CD,$2D,$7F,$D9,$54,$F5,$1A,$FF,$15,$8C,$36,$33,$3E
	.dc.b	$E6,$BF,$E4,$DF,$B1,$7C,$87,$F0,$86,$ED,$FC,$7D,$B3,$FE,$35,$6B
	.dc.b	$FE,$6F,$F3,$3F,$C7,$7E,$07,$FA,$8B,$4F,$F7,$53,$6F,$FC,$95,$5F
	.dc.b	$E9,$99,$CB,$C7,$0A,$FF,$A7,$5A,$FF,$BC,$EF,$C7,$7F,$EA,$E9,$CD
	.dc.b	$BC,$F7,$D2,$7F,$BB,$7E,$3F,$FD,$6F,$18,$C5,$18,$9D,$47,$FA,$1E
	.dc.b	$75,$3F,$E6,$9F,$BF,$CC,$FF,$BA,$5A,$FF,$99,$1B,$E3,$7F,$BD,$1C
	.dc.b	$E0,$FA,$A7,$CA,$A1,$F8,$A7,$FB,$2A,$73,$18,$C3,$FD,$31,$18,$9D
	.dc.b	$57,$17,$FC,$C9,$F8,$0F,$FA,$D6,$F8,$5F,$FA,$B2,$31,$3A,$F4,$B8
	.dc.b	$4F,$FA,$26,$72,$F6,$F8,$BC,$BF,$EC,$EC,$51,$88,$F4,$46,$30,$EF
	.dc.b	$18,$9D,$2E,$7F,$F4,$03,$FD,$95,$AF,$E7,$3C,$AF,$ED,$0E,$DB,$CB
	.dc.b	$F8,$A8,$FF,$62,$B5,$07,$0D,$E2,$FE,$C3,$87,$CF,$FA,$1F,$9D,$F9
	.dc.b	$47,$CD,$6D,$7C,$C6,$BE,$F6,$BE,$EF,$A5,$B5,$F6,$5B,$A1,$A7,$09
	.dc.b	$F8,$1D,$BE,$CD,$3F,$69,$18,$3E,$14,$7C,$0E,$CF,$D9,$C7,$EC,$7B
	.dc.b	$3F,$68,$47,$F5,$9A,$7E,$C4,$99,$7E,$C5,$F7,$F3,$FE,$6F,$EB,$3F
	.dc.b	$35,$46,$91,$FF,$5F,$73,$7C,$A8,$FC,$E4,$6C,$C9,$4F,$CF,$14,$FE
	.dc.b	$2B,$C3,$ED,$23,$F7,$5F,$B3,$7E,$68,$F4,$E1,$E7,$3E,$0B,$0F,$E3
	.dc.b	$B9,$FC,$73,$E2,$76,$FF,$12,$8F,$E5,$AD,$7F,$8E,$B5,$FE,$6E,$DE
	.dc.b	$2F,$C1,$92,$FC,$53,$FC,$A3,$0F,$3B,$F9,$13,$D0,$79,$4F,$39,$0F
	.dc.b	$E4,$51,$FE,$28,$DB,$F3,$31,$CC,$E2,$73,$38,$1F,$98,$8F,$0B,$71
	.dc.b	$FB,$8F,$F4,$AC,$3D,$5B,$7E,$0A,$31,$B7,$FB,$7E,$BF,$84,$6D,$F9
	.dc.b	$CE,$0F,$E3,$91,$96,$D7,$F0,$99,$FF,$20,$F4,$3C,$F7,$CF,$FF,$62
	.dc.b	$8C,$36,$33,$5F,$D8,$E7,$2F,$F6,$5C,$1E,$7B,$E9,$BF,$5A,$D8,$F3
	.dc.b	$36,$7F,$6A,$7E,$DA,$38,$CF,$F9,$43,$4F,$74,$FD,$CD,$3E,$1B,$4F
	.dc.b	$55,$7B,$73,$F8,$04,$77,$1F,$82,$9A,$FE,$09,$F8,$9F,$0D,$F1,$C9
	.dc.b	$38,$3C,$B3,$FC,$2B,$2F,$8D,$E5,$FF,$0C,$FB,$4C,$C6,$25,$18,$D5
	.dc.b	$5B,$9F,$7D,$6E,$BF,$9D,$6D,$FC,$F7,$AE,$E3,$FC,$AA,$FF,$9D,$CE
	.dc.b	$5B,$19,$7E,$11,$F8,$23,$F8,$21,$0F,$E8,$78,$7E,$0C,$91,$F8,$37
	.dc.b	$F9,$0F,$B5,$FE,$8D,$8F,$A6,$1C,$4F,$D5,$6B,$FD,$4D,$AF,$CB,$3C
	.dc.b	$07,$90,$87,$F4,$2F,$0B,$FD,$39,$3F,$D5,$9B,$1F,$D2,$5A,$EE,$3F
	.dc.b	$08,$F8,$5F,$2B,$4C,$9E,$37,$2D,$32,$79,$4F,$8C,$43,$CA,$BE,$C5
	.dc.b	$F5,$1E,$28,$C3,$63,$31,$E0,$32,$78,$BE,$57,$FA,$77,$9B,$EA,$6C
	.dc.b	$67,$AD,$39,$6C,$71,$F5,$5E,$CF,$AA,$D1,$8C,$B5,$39,$8C,$47,$FA
	.dc.b	$FB,$9B,$F4,$EF,$E7,$CF,$A6,$53,$F5,$F5,$E4,$7F,$D9,$8D,$BC,$E1
	.dc.b	$E4,$9C,$B6,$3E,$44,$73,$3E,$E9,$F3,$4B,$5E,$27,$06,$CF,$EE,$7E
	.dc.b	$87,$C2,$F3,$7F,$BA,$F1,$7F,$96,$63,$13,$FE,$0A,$F0,$3F,$DD,$15
	.dc.b	$A6,$C6,$4E,$7A,$FF,$A0,$6F,$93,$67,$71,$8A,$E5,$5F,$C0,$E7,$FB
	.dc.b	$51,$1F,$E5,$6D,$7F,$88,$35,$E0,$7C,$6A,$FF,$1A,$47,$C6,$6E,$39
	.dc.b	$3F,$D1,$C5,$8F,$AD,$DB,$FE,$6E,$7A,$57,$F1,$37,$B6,$9F,$80,$DB
	.dc.b	$FC,$33,$1F,$5B,$6F,$C1,$36,$33,$87,$DC,$9C,$BD,$C9,$7F,$E2,$2B
	.dc.b	$FC,$9D,$F9,$E7,$CB,$3F,$CC,$77,$FC,$EC,$FF,$03,$6C,$FF,$04,$6C
	.dc.b	$EF,$83,$DE,$FC,$63,$E4,$90,$C1,$FB,$7A,$FE,$1D,$C5,$FE,$5B,$8F
	.dc.b	$09,$FB,$78,$CD,$FB,$23,$F8,$23,$A3,$FC,$C9,$C5,$E0,$8F,$E0,$D3
	.dc.b	$FC,$67,$B7,$FC,$F3,$5D,$27,$E6,$1E,$33,$EC,$9E,$43,$C0,$FF,$20
	.dc.b	$BF,$25,$78,$A7,$C1,$18,$9F,$E5,$67,$FA,$B4,$FF,$3B,$72,$7E,$0A
	.dc.b	$73,$5F,$E7,$58,$C4,$77,$3C,$65,$E3,$FF,$56,$CF,$F4,$0F,$8F,$93
	.dc.b	$FC,$F9,$AF,$F5,$1E,$9F,$E9,$6B,$E8,$FF,$A7,$61,$FE,$5E,$6C,$64
	.dc.b	$FE,$9F,$3C,$23,$FD,$11,$D7,$E4,$9F,$F3,$AB,$72,$FF,$48,$37,$43
	.dc.b	$B9,$F4,$9F,$28,$86,$27,$FC,$22,$7C,$59,$FF,$08,$E5,$FD,$D8,$93
	.dc.b	$F4,$CE,$64,$CF,$E1,$BF,$D4,$BB,$FF,$C1,$13,$FD,$CA,$FF,$C0,$A7
	.dc.b	$D1,$8C,$CE,$FB,$FD,$9D,$BF,$E1,$4F,$5B,$FB,$97,$6F,$FA,$09,$B4
	.dc.b	$69,$E4,$7F,$D3,$5E,$07,$C1,$26,$3F,$EE,$9C,$FC,$B3,$F6,$7A,$7F
	.dc.b	$74,$FA,$FB,$BF,$DB,$FA,$FF,$96,$89,$1F,$E4,$4B,$FF,$79,$78,$5F
	.dc.b	$E9,$E9,$FF,$26,$12,$7F,$C8,$CF,$07,$FC,$84,$D7,$FD,$37,$5D,$9D
	.dc.b	$90,$FF,$7E,$F8,$7F,$C2,$1C,$1C,$3C,$3F,$B5,$3C,$63,$A9,$E5,$AF
	.dc.b	$E2,$79,$7B,$A3,$EE,$B5,$FF,$4A,$47,$F9,$C9,$F5,$BD,$DF,$F1,$DE
	.dc.b	$08,$C4,$78,$71,$C3,$DC,$87,$88,$F4,$9F,$1B,$1E,$2F,$FA,$FC,$C1
	.dc.b	$C8,$FF,$47,$FD,$C7,$FA,$50,$8F,$FA,$51,$FF,$4E,$91,$FF,$7B,$74
	.dc.b	$FF,$B1,$23,$11,$E1,$3C,$D7,$F3,$DF,$F6,$35,$FE,$07,$4F,$BF,$1F
	.dc.b	$81,$F1,$30,$7C,$26,$57,$C5,$23,$FE,$F4,$F1,$FF,$DF,$5F,$93,$FF
	.dc.b	$BE,$5B,$19,$C1,$39,$6C,$7F,$B4,$5A,$FF,$B4,$5A,$FF,$BB,$BC,$9E
	.dc.b	$6E,$67,$D5,$BF,$DF,$8F,$69,$F4,$0F,$F5,$A9,$F3,$30,$FC,$93,$F4
	.dc.b	$9F,$D8,$73,$FF,$BE,$7F,$11,$BE,$33,$24,$61,$B7,$FE,$4A,$AE,$0F
	.dc.b	$F9,$AB,$CC,$79,$BC,$6F,$FC,$DB,$29,$CD,$F2,$8F,$F4,$2C,$62,$76
	.dc.b	$D8,$ED,$E7,$F3,$6D,$71,$7C,$D8,$27,$51,$FE,$8F,$8C,$4E,$FA,$BF
	.dc.b	$D4,$5E,$6A,$9E,$FF,$FB,$80,$8F,$FB,$81,$FF,$74,$91,$FC,$69,$C0
	.dc.b	$AF,$F1,$9F,$4D,$D5,$F2,$7D,$8C,$36,$BF,$CE,$33,$AC,$55,$A6,$CE
	.dc.b	$BD,$5E,$0F,$FA,$A6,$91,$88,$FA,$D1,$89,$E3,$1B,$9D,$3F,$9E,$E1
	.dc.b	$FC,$3E,$1F,$C1,$8F,$E8,$1D,$9F,$C0,$49,$D1,$FA,$2A,$FE,$83,$D9
	.dc.b	$FB,$CB,$FC,$FC,$7F,$02,$7F,$73,$19,$8F,$F5,$5F,$97,$E3,$EF,$FC
	.dc.b	$0B,$63,$39,$7E,$36,$72,$D8,$FF,$6F,$C7,$E2,$31,$62,$EB,$FE,$79
	.dc.b	$5A,$21,$FE,$CE,$F5,$BF,$88,$DF,$E4,$1B,$1F,$EB,$E6,$BB,$CF,$FA
	.dc.b	$43,$FE,$C2,$F6,$9E,$C6,$23,$FD,$B3,$18,$7C,$26,$23,$4F,$E2,$E3
	.dc.b	$57,$FA,$0C,$1E,$CF,$F6,$D6,$F0,$7F,$1D,$BF,$C9,$46,$31,$E1,$5B
	.dc.b	$25,$6E,$BE,$14,$62,$77,$39,$AE,$55,$FE,$DC,$6B,$FE,$F2,$6B,$E0
	.dc.b	$AE,$4E,$DA,$ED,$AE,$E7,$B4,$62,$3C,$71,$82,$57,$90,$91,$D6,$32
	.dc.b	$D8,$CB,$DE,$32,$D8,$CB,$E1,$6B,$E0,$6D,$7B,$23,$07,$52,$1C,$CA
	.dc.b	$73,$29,$C0,$D1,$0D,$14,$D1,$82,$18,$23,$E3,$25,$7C,$76,$D7,$BA
	.dc.b	$78,$C6,$2B,$D7,$5B,$25,$6E,$BA,$C6,$27,$73,$9A,$E5,$5E,$C6,$BE
	.dc.b	$96,$BE,$0A,$E4,$ED,$AE,$DA,$EE,$7B,$46,$23,$C7,$18,$25,$79,$09
	.dc.b	$1F,$CB,$CF,$F5,$DF,$A5,$FD,$11,$7A,$7F,$EC,$47,$46,$3D,$34,$EE
	.dc.b	$75,$3C,$FA,$3F,$9E,$BF,$D3,$9A,$F2,$6B,$BE,$AF,$F4,$A3,$96,$B9
	.dc.b	$C7,$C9,$3C,$23,$FA,$33,$53,$49,$3C,$A3,$6D,$76,$D8,$DC,$F6,$8C
	.dc.b	$11,$C1,$27,$D9,$19,$6C,$65,$F5,$35,$F5,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$3D,$6F,$56,$9C,$8A,$71,$29,$C4,$F8,$2F,$16,$9C,$4A,$71,$29
	.dc.b	$C6,$7E,$3B,$B6,$BB,$6B,$B9,$F8,$CE,$5A,$E6,$7C,$53,$C2,$3C,$E4
	.dc.b	$3C,$25,$74,$E0,$8E,$09,$3D,$A3,$83,$5D,$B6,$37,$3E,$28,$C1,$1C
	.dc.b	$12,$7E,$04,$65,$B1,$97,$DC,$D7,$DC,$68,$FF,$18,$F0,$3E,$03,$FC
	.dc.b	$41,$8E,$1B,$3F,$B0,$9D,$3F,$B0,$31,$1A,$7F,$B4,$19,$2C,$FF,$6F
	.dc.b	$CF,$51,$FE,$2D,$E1,$FF,$41,$CF,$F7,$BB,$FE,$F4,$DD,$B9,$57,$F8
	.dc.b	$CF,$8B,$FB,$FC,$62,$11,$89,$FF,$63,$7A,$4E,$6F,$F2,$B7,$DB,$E1
	.dc.b	$B7,$FC,$3F,$55,$6D,$56,$DB,$3F,$E0,$5D,$3F,$8A,$2D,$3F,$B8,$3E
	.dc.b	$B8,$CC,$63,$E8,$3F,$D4,$5E,$E7,$F1,$B6,$D7,$F9,$81,$AF,$F1,$C6
	.dc.b	$BF,$EF,$8F,$BF,$FF,$29,$4F,$F1,$7F,$49,$18,$9D,$09,$D4,$66,$3F
	.dc.b	$CB,$58,$BE,$75,$F8,$3E,$67,$FC,$F5,$18,$79,$1F,$15,$C5,$FF,$BC
	.dc.b	$5B,$1F,$27,$5E,$76,$FC,$5F,$E8,$E1,$FE,$85,$6C,$66,$B8,$4E,$5E
	.dc.b	$5F,$94,$57,$CF,$68,$C4,$7F,$A3,$E3,$15,$CE,$37,$3A,$87,$B0,$F8
	.dc.b	$DF,$F3,$D7,$C0,$FF,$48,$9C,$C6,$2F,$F0,$55,$B5,$5B,$8C,$3F,$E9
	.dc.b	$6F,$A0,$7C,$37,$1A,$FE,$B1,$18,$78,$9F,$D5,$9C,$4F,$19,$E1,$3F
	.dc.b	$EA,$9A,$FF,$2C,$3C,$0F,$AE,$E2,$7C,$5D,$1C,$31,$F6,$B7,$37,$CF
	.dc.b	$7E,$73,$0F,$92,$D7,$E4,$36,$BF,$3E,$D9,$7C,$4D,$24,$FE,$82,$BE
	.dc.b	$9C,$73,$EE,$72,$3D,$6F,$F5,$37,$C2,$FE,$55,$B7,$FD,$C6,$74,$A7
	.dc.b	$4D,$3F,$BB,$70,$BE,$87,$8C,$7F,$6C,$8C,$3B,$31,$1A,$7F,$B7,$C6
	.dc.b	$A7,$FC,$0B,$F4,$8F,$D2,$1E,$E7,$FA,$EF,$BB,$31,$8C,$3D,$C2,$B8
	.dc.b	$56,$E3,$FC,$29,$18,$76,$7F,$10,$F0,$BF,$D2,$17,$FC,$5C,$91,$FE
	.dc.b	$16,$F0,$BF,$8B,$C7,$93,$A9,$EB,$7D,$5C,$3F,$50,$D8,$CF,$17,$92
	.dc.b	$72,$F9,$8F,$E2,$FD,$7F,$EF,$1E,$BF,$D6,$93,$2F,$D6,$CF,$E6,$DF
	.dc.b	$D2,$B5,$FD,$56,$FC,$5F,$D8,$C6,$4F,$E4,$71,$98,$FC,$5C,$7F,$26
	.dc.b	$7F,$80,$73,$7E,$CD,$B1,$97,$7C,$DF,$C1,$FA,$7E,$54,$7F,$12,$EC
	.dc.b	$FD,$AB,$EB,$6E,$DB,$8F,$81,$C5,$F5,$6F,$EB,$F1,$7F,$BC,$DB,$3D
	.dc.b	$CF,$B1,$3B,$7E,$87,$17,$FB,$0E,$7F,$A5,$12,$33,$D3,$C7,$83,$E6
	.dc.b	$CF,$D5,$E2,$F6,$57,$FC,$01,$9C,$BF,$82,$E2,$FA,$B5,$FE,$49,$E0
	.dc.b	$7C,$22,$9E,$A6,$63,$0D,$8C,$9F,$A0,$8C,$BD,$71,$70,$EA,$3C,$58
	.dc.b	$FE,$12,$FC,$D3,$99,$FC,$C4,$E5,$E8,$3E,$B3,$C0,$47,$2F,$22,$3E
	.dc.b	$19,$E0,$D8,$FE,$E4,$C3,$E4,$94,$FD,$B7,$C0,$7F,$99,$32,$34,$91
	.dc.b	$FE,$78,$3F,$9D,$7C,$9F,$F9,$8E,$FF,$B8,$B7,$43,$4F,$F3,$37,$47
	.dc.b	$F8,$02,$BC,$07,$F8,$0F,$83,$F9,$4B,$E4,$21,$FE,$8B,$8E,$0E,$0E
	.dc.b	$1C,$1E,$3C,$3F,$C0,$3F,$25,$E1,$23,$9C,$51,$E2,$C3,$BD,$F5,$6B
	.dc.b	$D0,$F2,$F7,$7F,$59,$F2,$3E,$21,$23,$FD,$25,$AF,$F8,$55,$F1,$73
	.dc.b	$7F,$76,$F9,$1F,$EA,$CD,$8F,$F0,$B3,$5F,$EF,$6D,$7F,$90,$FE,$1F
	.dc.b	$F9,$1E,$4F,$CA,$36,$3F,$96,$B5,$FE,$58,$D7,$D0,$7C,$53,$F8,$E6
	.dc.b	$DF,$EB,$1D,$FE,$4C,$7F,$33,$FB,$4F,$BC,$47,$D6,$FF,$AD,$F8,$1F
	.dc.b	$9E,$3F,$0B,$FD,$27,$B5,$FE,$AC,$9C,$C6,$1F,$F1,$94,$EE,$D5,$B8
	.dc.b	$C3,$FE,$B2,$C3,$C7,$3E,$83,$E9,$63,$E7,$E1,$FA,$7C,$3F,$E0,$9E
	.dc.b	$7F,$F1,$F5,$7F,$4B,$6B,$FD,$21,$B1,$B8,$FF,$13,$9F,$E2,$8C,$7C
	.dc.b	$9F,$49,$E3,$D3,$F7,$CD,$93,$4F,$36,$1F,$84,$8F,$96,$7E,$60,$A7
	.dc.b	$DC,$FA,$1F,$DF,$CF,$C5,$E1,$46,$36,$D1,$D8,$B3,$FE,$D6,$F8,$F7
	.dc.b	$99,$F3,$E7,$D3,$82,$73,$97,$F0,$C8,$E1,$5B,$6C,$66,$BF,$DE,$54
	.dc.b	$FA,$87,$E1,$DF,$22,$30,$D8,$CD,$F8,$27,$2F,$88,$F7,$9F,$58,$F8
	.dc.b	$5C,$3F,$6A,$3C,$B7,$C7,$1F,$23,$E2,$6D,$7F,$2D,$9D,$BF,$3B,$87
	.dc.b	$F0,$04,$7A,$ED,$F4,$70,$FC,$3C,$78,$8F,$C4,$1F,$CD,$FD,$77,$B1
	.dc.b	$FE,$9B,$E7,$7A,$BF,$CE,$E9,$5A,$F5,$BF,$5D,$39,$8C,$4F,$F7,$C8
	.dc.b	$C5,$F3,$6D,$70,$9F,$04,$E5,$B1,$97,$BC,$65,$FD,$9C,$FF,$A8,$C8
	.dc.b	$FF,$5A,$6B,$FA,$26,$BF,$C7,$9F,$53,$F8,$3D,$3F,$0D,$1F,$80,$D7
	.dc.b	$DE,$7C,$FE,$0F,$EA,$31,$93,$E7,$1D,$2B,$C6,$78,$4F,$C3,$EF,$F8
	.dc.b	$28,$FC,$A3,$6B,$3C,$1F,$D9,$67,$2D,$8F,$01,$E2,$AF,$BD,$87,$9A
	.dc.b	$33,$5E,$37,$F4,$7E,$7F,$E9,$F8,$FF,$B3,$B6,$3F,$CA,$86,$8F,$35
	.dc.b	$FC,$52,$9F,$14,$FE,$23,$97,$F7,$99,$FD,$E1,$27,$33,$D0,$FE,$14
	.dc.b	$79,$CF,$E1,$45,$E6,$FE,$15,$87,$EB,$E3,$2F,$CA,$7C,$27,$CA,$3C
	.dc.b	$87,$F0,$FB,$FF,$05,$7D,$CF,$F0,$D2,$5F,$89,$E9,$7F,$EA,$5E,$AF
	.dc.b	$99,$7F,$E6,$A6,$BF,$CD,$CF,$CD,$99,$29,$E0,$3E,$11,$4F,$21,$FD
	.dc.b	$30,$FF,$1B,$9F,$D2,$0A,$67,$CA,$FE,$9E,$E2,$FF,$DC,$E7,$F7,$A7
	.dc.b	$FA,$C9,$FE,$1C,$F0,$BF,$AE,$B4,$CC,$6D,$FF,$40,$78,$1E,$A7,$FC
	.dc.b	$FA,$DC,$3F,$B6,$47,$A8,$C9,$4F,$49,$39,$5C,$CE,$7D,$46,$1E,$58
	.dc.b	$7F,$80,$8B,$E1,$7F,$80,$1B,$E1,$7E,$93,$0F,$F7,$03,$5F,$91,$7C
	.dc.b	$F8,$3F,$4C,$D8,$FB,$ED,$77,$8F,$8E,$7F,$DA,$98,$7F,$88,$30,$F0
	.dc.b	$E9,$D6,$FF,$C4,$F1,$83,$E0,$67,$FE,$E5,$F0,$FE,$27,$87,$FE,$30
	.dc.b	$AC,$B7,$1F,$F2,$B4,$FC,$B9,$F8,$99,$78,$5F,$2B,$5F,$F2,$7F,$5F
	.dc.b	$F9,$4B,$ED,$FF,$C9,$E4,$8F,$F2,$33,$FE,$5B,$23,$FE,$50,$6B,$CB
	.dc.b	$27,$DB,$FF,$96,$09,$1F,$E4,$F7,$FC,$C8,$47,$FC,$B2,$D7,$97,$DB
	.dc.b	$7F,$98,$08,$FF,$95,$DF,$F2,$66,$FF,$6D,$B8,$FD,$F6,$F0,$9A,$7F
	.dc.b	$A2,$89,$5F,$EB,$EF,$2F,$FD,$89,$18,$8F,$F4,$C7,$27,$F9,$F7,$E9
	.dc.b	$BF,$D2,$C4,$8F,$E8,$D9,$BC,$EF,$F4,$EC,$FD,$CF,$3F,$FC,$F7,$85
	.dc.b	$AF,$15,$FF,$AA,$23,$07,$E0,$B8,$BF,$D0,$3F,$41,$68,$D2,$57,$FA
	.dc.b	$F9,$FC,$4F,$A5,$C2,$30,$43,$E9,$63,$FE,$88,$23,$FE,$7F,$7E,$29
	.dc.b	$1F,$F5,$2B,$5F,$F5,$16,$9F,$EE,$78,$FF,$58,$92,$3E,$47,$A3,$F1
	.dc.b	$5F,$F7,$DF,$A9,$FE,$A3,$F8,$DF,$F5,$33,$63,$22,$32,$D3,$FB,$87
	.dc.b	$43,$D6,$78,$3C,$67,$FD,$5D,$39,$8C,$42,$30,$F3,$37,$3C,$9C,$1F
	.dc.b	$EA,$EE,$C7,$AF,$FE,$B0,$CF,$F6,$6D,$9F,$BB,$3A,$DD,$5A,$6C,$FF
	.dc.b	$B8,$A3,$EF,$11,$C1,$F8,$6E,$4F,$E2,$4B,$E2,$FF,$BA,$D2,$73,$18
	.dc.b	$6C,$F1,$E6,$FF,$67,$4F,$FA,$AF,$C5,$7B,$58,$AF,$EF,$F1,$87,$B1
	.dc.b	$88,$D3,$FE,$1D,$8D,$4F,$7F,$63,$33,$FE,$D2,$C1,$18,$C7,$CC,$D8
	.dc.b	$CC,$F6,$3F,$DC,$52,$CD,$CD,$FE,$DC,$FC,$0B,$F3,$84,$AF,$9B,$19
	.dc.b	$7E,$7F,$35,$79,$D2,$31,$1F,$65,$B7,$DD,$BD,$5B,$97,$C4,$FC,$E8
	.dc.b	$C3,$67,$FD,$F0,$D9,$D9,$F9,$2E,$27,$BD,$A8,$C3,$5F,$C6,$B5,$FC
	.dc.b	$94,$70,$79,$81,$EF,$FF,$93,$BD,$FF,$EB,$CD,$8F,$EB,$8D,$F1,$7F
	.dc.b	$C7,$2D,$8F,$F1,$DF,$C0,$7F,$7C,$9F,$F7,$63,$63,$31,$FC,$FE,$31
	.dc.b	$1F,$2B,$C9,$FC,$E3,$75,$FC,$CB,$79,$FF,$A9,$CE,$A3,$FD,$CF,$F0
	.dc.b	$9F,$CA,$88,$FF,$93,$3C,$4F,$ED,$51,$E2,$C7,$DC,$D9,$34,$8F,$F9
	.dc.b	$53,$C4,$7C,$17,$08,$C3,$5F,$F3,$94,$69,$E6,$68,$A7,$F0,$8F,$83
	.dc.b	$FE,$D0,$FF,$03,$6F,$71,$A4,$D3,$FC,$11,$19,$6C,$67,$1F,$F3,$4D
	.dc.b	$73,$9C,$92,$33,$5C,$5B,$5C,$6B,$FC,$E3,$39,$6C,$65,$F4,$57,$37
	.dc.b	$2D,$73,$5D,$23,$11,$FC,$8A,$79,$36,$32,$F9,$E7,$81,$1E,$AD,$7A
	.dc.b	$98,$21,$82,$1D,$5E,$0D,$36,$53,$67,$84,$C9,$4C,$D7,$C3,$6D,$7A
	.dc.b	$9D,$1C,$C8,$60,$91,$C5,$CB,$5C,$C7,$38,$C1,$1C,$12,$BE,$3C,$65
	.dc.b	$B1,$9A,$F7,$D7,$49,$C9,$23,$35,$C5,$B5,$C6,$BE,$0C,$E5,$B1,$97
	.dc.b	$E0,$D7,$37,$2D,$73,$5D,$23,$11,$F0,$27,$93,$63,$2F,$FB,$2B,$D4
	.dc.b	$FC,$CB,$4F,$CD,$FA,$BF,$ED,$87,$FD,$8A,$D7,$FD,$C1,$F3,$EF,$27
	.dc.b	$FD,$35,$18,$6B,$FC,$6E,$35,$7D,$9B,$1F,$A8,$6B,$FE,$D2,$6B,$CE
	.dc.b	$38,$4F,$48,$4E,$9E,$D1,$98,$C4,$6A,$30,$49,$E6,$48,$F1,$46,$5B
	.dc.b	$19,$7C,$91,$96,$C6,$5F,$31,$C0,$86,$8A,$68,$C1,$0C,$10,$F5,$BD
	.dc.b	$5A,$72,$29,$C4,$A7,$13,$E0,$3C,$5A,$71,$29,$C4,$A7,$19,$F9,$24
	.dc.b	$9F,$81,$19,$6B,$E0,$21,$C8,$B1,$C6,$7C,$4D,$8F,$13,$5F,$13,$5E
	.dc.b	$D1,$C2,$7B,$C7,$52,$4E,$DE,$F1,$A8,$C4,$6A,$30,$49,$E8,$48,$F1
	.dc.b	$C6,$5B,$19,$7C,$B1,$96,$C6,$5F,$39,$C0,$86,$8A,$68,$C1,$0C,$10
	.dc.b	$FF,$01,$6D,$FE,$D1,$C3,$FD,$2A,$48,$FE,$D9,$A7,$FB,$5E,$3F,$A8
	.dc.b	$FE,$89,$FE,$12,$B4,$63,$99,$18,$BF,$F0,$71,$FE,$0F,$AE,$15,$B9
	.dc.b	$D4,$EB,$E3,$5C,$5C,$AB,$FA,$FC,$62,$B7,$5F,$E1,$98,$C4,$EA,$B8
	.dc.b	$4F,$80,$4B,$CF,$CC,$77,$6B,$CD,$AF,$06,$C3,$E0,$E9,$30,$F1,$D7
	.dc.b	$68,$DF,$93,$27,$34,$8C,$63,$FD,$EB,$F6,$DF,$F4,$5D,$72,$6C,$67
	.dc.b	$C9,$F1,$C6,$31,$F6,$3F,$29,$BA,$E8,$97,$CF,$E6,$B8,$91,$C1,$DF
	.dc.b	$E5,$FA,$91,$C1,$EC,$FD,$47,$D1,$7F,$E4,$63,$E6,$38,$ED,$FF,$09
	.dc.b	$B6,$33,$A7,$F9,$36,$7F,$C1,$FE,$27,$F9,$85,$BE,$16,$6B,$FD,$5B
	.dc.b	$39,$6C,$65,$FC,$4F,$C8,$FE,$31,$F1,$18,$C7,$4F,$F9,$6E,$35,$7F
	.dc.b	$E2,$66,$C7,$6B,$DF,$67,$FA,$09,$A7,$FA,$12,$3A,$9F,$EB,$97,$15
	.dc.b	$FE,$F8,$24,$7E,$93,$CC,$DA,$E8,$74,$27,$3D,$1F,$A7,$BF,$D0,$12
	.dc.b	$FF,$2D,$19,$7F,$31,$CB,$F9,$F9,$FC,$E6,$9D,$A3,$13,$AE,$8F,$D7
	.dc.b	$D7,$E9,$B7,$B3,$C9,$AF,$EB,$5A,$FE,$A5,$AB,$93,$F2,$73,$F9,$1C
	.dc.b	$3D,$B1,$89,$FC,$8F,$2F,$0E,$37,$AD,$FC,$03,$E2,$7F,$CA,$B1,$8C
	.dc.b	$11,$89,$D5,$FE,$D6,$B5,$3E,$E2,$47,$B6,$B8,$B6,$3F,$CC,$3C,$1F
	.dc.b	$3B,$7F,$CA,$CF,$F0,$E8,$C6,$9A,$E0,$FA,$4D,$AD,$D7,$F1,$3E,$7F
	.dc.b	$E2,$F3,$FC,$4C,$91,$F5,$23,$2F,$8E,$3E,$99,$1C,$1E,$BC,$FF,$90
	.dc.b	$FD,$9F,$BF,$33,$49,$3E,$D7,$EE,$91,$C1,$CE,$3F,$DE,$AF,$F2,$AD
	.dc.b	$3E,$66,$3D,$49,$1F,$1D,$FC,$39,$1C,$1F,$97,$3E,$11,$0C,$1F,$B2
	.dc.b	$F8,$0F,$D1,$91,$FD,$1F,$2F,$FB,$A2,$3F,$5F,$6F,$83,$7D,$37,$F4
	.dc.b	$07,$F9,$84,$66,$31,$7F,$D0,$A3,$07,$17,$13,$CA,$31,$1F,$EF,$88
	.dc.b	$7D,$77,$F8,$FA,$7C,$0D,$8C,$C7,$FA,$90,$C9,$FE,$4E,$E0,$FC,$35
	.dc.b	$7B,$CF,$C3,$70,$7C,$C6,$E7,$C6,$DF,$0B,$D4,$98,$65,$FF,$22,$61
	.dc.b	$D2,$FE,$49,$C8,$EC,$7F,$09,$F8,$8F,$E1,$A4,$7E,$5F,$52,$3E,$5E
	.dc.b	$0F,$8F,$FE,$D3,$F8,$1F,$CA,$8C,$64,$8C,$4E,$B0,$F9,$2D,$7E,$47
	.dc.b	$D8,$FC,$52,$57,$B5,$F4,$EB,$F0,$67,$FC,$02,$D7,$F8,$D3,$73,$F7
	.dc.b	$47,$D3,$FC,$43,$E8,$12,$FE,$43,$D3,$6F,$A1,$7C,$48,$E0,$95,$F1
	.dc.b	$8E,$EF,$D7,$F2,$BF,$A4,$5F,$FB,$ED,$B3,$FB,$D6,$BF,$E1,$86,$BC
	.dc.b	$8E,$98,$7A,$1F,$6D,$78,$1F,$AE,$60,$FA,$F4,$F0,$7F,$17,$F0,$5F
	.dc.b	$D3,$E7,$FC,$37,$18,$9F,$CD,$36,$06,$B8,$BF,$C4,$35,$C6,$9C,$2F
	.dc.b	$C6,$FF,$AD,$47,$E8,$E3,$0F,$23,$11,$AB,$C4,$E9,$E9,$FB,$9F,$EC
	.dc.b	$1E,$07,$17,$0E,$B9,$7F,$60,$E5,$AF,$C1,$3E,$19,$F2,$4F,$8B,$7E
	.dc.b	$8B,$F1,$47,$F8,$F2,$FC,$53,$E6,$3C,$31,$99,$DB,$CC,$FD,$21,$DB
	.dc.b	$4F,$A8,$E4,$F1,$1A,$21,$F3,$3C,$C7,$3F,$F9,$2F,$36,$FF,$E4,$C9
	.dc.b	$CF,$91,$F4,$A7,$37,$C6,$F8,$5E,$DB,$19,$AE,$13,$96,$C6,$45,$74
	.dc.b	$7F,$DB,$0D,$C3,$FA,$95,$7F,$4D,$24,$FF,$81,$5A,$FF,$0E,$6B,$F9
	.dc.b	$47,$F9,$3E,$7F,$82,$7F,$10,$DC,$FE,$74,$63,$3F,$F3,$6B,$FE,$E2
	.dc.b	$BF,$0D,$F7,$21,$DC,$FE,$61,$AF,$E1,$E7,$F0,$91,$F2,$AF,$3A,$FE
	.dc.b	$6A,$72,$77,$8D,$4F,$E4,$B0,$F0,$C7,$90,$91,$F9,$9D,$FF,$99,$FB
	.dc.b	$1F,$A1,$6C,$FE,$A3,$83,$FD,$0B,$39,$6C,$7A,$2B,$E2,$14,$FE,$0B
	.dc.b	$87,$C8,$69,$9C,$FE,$0B,$FC,$14,$7D,$7F,$C5,$9F,$DA,$B6,$33,$5F
	.dc.b	$C3,$5A,$7C,$72,$1E,$22,$9E,$02,$D7,$81,$D1,$FD,$87,$93,$EA,$BF
	.dc.b	$D7,$1A,$FF,$29,$9F,$C5,$3B,$6B,$D4,$E6,$53,$0F,$F6,$FE,$5E,$71
	.dc.b	$FD,$BD,$B1,$93,$A9,$70,$FE,$8E,$BF,$21,$A4,$AF,$E9,$31,$97,$FC
	.dc.b	$E7,$D5,$F2,$1A,$7C,$C2,$E5,$FE,$7C,$8D,$47,$C4,$E7,$FF,$30,$61
	.dc.b	$F5,$F3,$F8,$D8,$7F,$81,$FA,$FC,$FD,$7F,$DB,$A3,$CE,$DA,$FE,$C1
	.dc.b	$3E,$EC,$3E,$15,$FC,$F6,$E1,$D9,$FF,$04,$ED,$D0,$98,$F4,$7B,$F1
	.dc.b	$7F,$B5,$3D,$3F,$ED,$8D,$AF,$F0,$8B,$5F,$F0,$83,$5F,$F4,$17,$47
	.dc.b	$B9,$A7,$C3,$3F,$A0,$F6,$FF,$89,$5C,$1A,$DF,$E1,$65,$E7,$8F,$8D
	.dc.b	$D7,$F8,$0F,$23,$FD,$B9,$F3,$BF,$E7,$E2,$4F,$F9,$85,$FF,$25,$11
	.dc.b	$FF,$22,$35,$FF,$74,$79,$3F,$EF,$0F,$B3,$FF,$43,$92,$3F,$CD,$4F
	.dc.b	$F9,$4C,$8F,$F9,$35,$AF,$2D,$FF,$C2,$4F,$F9,$C7,$C0,$FF,$48,$91
	.dc.b	$FF,$32,$46,$9F,$F7,$EF,$96,$F1,$FF,$E0,$0D,$7F,$BF,$49,$3F,$EA
	.dc.b	$77,$FC,$E4,$47,$07,$F9,$FF,$EC,$FF,$D3,$E4,$70,$7F,$98,$DF,$F3
	.dc.b	$61,$1C,$1F,$E9,$4B,$FA,$2F,$FA,$6F,$CA,$46,$3C,$BF,$F7,$7F,$07
	.dc.b	$FA,$74,$FF,$54,$47,$0D,$BF,$DE,$CD,$F2,$B8,$CE,$8F,$F5,$99,$30
	.dc.b	$FF,$4B,$35,$F8,$79,$F0,$7E,$3C,$7F,$81,$9C,$D7,$9B,$4F,$90,$FA
	.dc.b	$39,$FF,$D3,$95,$FE,$97,$25,$7F,$B3,$DF,$F7,$5F,$87,$F9,$3E,$4F
	.dc.b	$F6,$D4,$FE,$41,$B3,$A8,$F8,$7E,$6F,$FB,$A6,$3F,$DC,$98,$7E,$6E
	.dc.b	$53,$97,$CC,$62,$3B,$5F,$F8,$96,$3B,$36,$BF,$DF,$71,$8F,$0B,$A7
	.dc.b	$9C,$D6,$92,$7E,$AC,$65,$FA,$9E,$57,$D6,$F5,$BC,$31,$FE,$D3,$8C
	.dc.b	$3E,$4E,$14,$62,$3E,$64,$63,$35,$72,$87,$A1,$BA,$FF,$1B,$C6,$2B
	.dc.b	$71,$D6,$74,$FF,$B8,$DB,$3F,$5A,$BA,$70,$E0,$97,$CA,$FC,$7C,$7C
	.dc.b	$09,$19,$8F,$0C,$76,$E3,$FC,$7D,$FE,$48,$93,$FE,$F9,$AE,$0A,$5E
	.dc.b	$8F,$12,$4E,$A3,$2D,$F4,$BB,$11,$C1,$BE,$9F,$11,$1C,$1E,$0E,$6F
	.dc.b	$81,$87,$F0,$0C,$7E,$AC,$62,$3F,$37,$1F,$EE,$D2,$3F,$8A,$E6,$FF
	.dc.b	$7B,$46,$23,$F4,$5E,$A3,$D4,$FE,$7C,$DC,$3F,$B1,$B7,$C4,$FC,$1B
	.dc.b	$6B,$F2,$AD,$79,$D7,$22,$57,$26,$D7,$26,$CF,$26,$BC,$A7,$81,$27
	.dc.b	$83,$63,$83,$5E,$11,$A2,$3A,$6B,$A3,$04,$30,$43,$FD,$17,$5E,$D6
	.dc.b	$D7,$99,$B5,$E1,$6D,$76,$69,$D4,$A7,$51,$5D,$5B,$5D,$5B,$5D,$5B
	.dc.b	$5D,$5A,$75,$29,$D6,$BA,$12,$BA,$36,$BA,$36,$BA,$36,$BA,$35,$E9
	.dc.b	$5C,$89,$5C,$9B,$5C,$9B,$3C,$9A,$F2,$9E,$04,$9E,$0D,$8E,$0D,$78
	.dc.b	$46,$88,$E9,$AE,$8C,$10,$C1,$0F,$A6,$43,$E4,$96,$BE,$03,$6B,$D2
	.dc.b	$DA,$FF,$23,$F9,$FF,$EA,$08,$FE,$4E,$D7,$F9,$CB,$5F,$F4,$E1,$FA
	.dc.b	$AF,$82,$FF,$6E,$3F,$B2,$6B,$F9,$06,$BF,$86,$6D,$9A,$4A,$FE,$6F
	.dc.b	$3B,$6C,$ED,$B3,$B9,$F1,$46,$08,$E0,$87,$EF,$A3,$83,$63,$6D,$36
	.dc.b	$7A,$DC,$B4,$C9,$ED,$72,$D3,$27,$A4,$B3,$E9,$6C,$F8,$9E,$87,$84
	.dc.b	$86,$08,$75,$32,$53,$27,$73,$25,$33,$1E,$17,$2D,$73,$1E,$37,$A1
	.dc.b	$0E,$45,$8E,$0D,$8E,$0E,$89,$1A,$6C,$6A,$30,$47,$04,$9F,$97,$3C
	.dc.b	$1B,$3B,$6C,$EE,$7E,$5C,$60,$8E,$08,$7C,$08,$E0,$D8,$DB,$4D,$9F
	.dc.b	$11,$CB,$4C,$9F,$05,$CB,$4C,$9E,$92,$CF,$A5,$CF,$C8,$3C,$74,$3D
	.dc.b	$1F,$F0,$BC,$E6,$30,$7F,$8A,$9C,$46,$A7,$8B,$FE,$1E,$D5,$E2,$3C
	.dc.b	$4E,$38,$7F,$91,$E7,$31,$88,$F0,$C6,$2B,$95,$71,$7F,$BA,$F2,$FF
	.dc.b	$98,$1B,$9F,$1B,$70,$AB,$51,$FE,$40,$8C,$4E,$FF,$4B,$79,$BA,$AF
	.dc.b	$F2,$D4,$62,$75,$39,$9F,$8B,$39,$C3,$FC,$76,$BC,$4D,$4A,$31,$7D
	.dc.b	$E7,$31,$8A,$E3,$AE,$09,$5E,$57,$7E,$54,$8F,$43,$67,$C3,$E4,$F4
	.dc.b	$24,$FB,$5F,$1F,$D0,$3C,$9A,$49,$F9,$11,$97,$E5,$77,$7F,$9E,$2F
	.dc.b	$FC,$87,$18,$EE,$F6,$CE,$75,$FF,$35,$DB,$E5,$FB,$92,$FA,$BF,$13
	.dc.b	$C2,$FF,$32,$7B,$3F,$E6,$26,$DF,$F9,$D1,$AF,$FA,$2D,$AF,$E5,$3C
	.dc.b	$87,$B5,$F9,$6A,$FF,$35,$C6,$25,$18,$9D,$47,$FA,$02,$75,$3F,$E8
	.dc.b	$68,$D9,$9C,$36,$FF,$A6,$E7,$FD,$5A,$E8,$C4,$69,$FD,$3F,$4F,$E9
	.dc.b	$6B,$F4,$51,$B3,$3C,$3B,$53,$B7,$FC,$95,$CF,$FA,$5A,$FC,$FF,$4F
	.dc.b	$19,$74,$FE,$9A,$72,$7E,$80,$91,$F9,$DE,$9F,$F6,$1C,$62,$3F,$37
	.dc.b	$8E,$1F,$F5,$0F,$37,$EF,$F3,$EC,$DA,$FC,$9C,$63,$9B,$85,$7E,$91
	.dc.b	$AF,$E8,$9A,$FE,$7D,$BA,$39,$BF,$85,$FB,$5F,$C2,$9B,$5F,$C2,$5A
	.dc.b	$FF,$B2,$67,$82,$9F,$DE,$91,$F7,$5E,$5B,$1F,$43,$C7,$91,$FC,$57
	.dc.b	$3F,$E3,$71,$88,$FE,$21,$BF,$F1,$38,$F9,$E4,$8F,$A7,$BF,$FA,$E5
	.dc.b	$FF,$80,$3B,$7F,$BB,$A7,$F8,$79,$23,$C6,$FB,$57,$B3,$48,$FB,$CC
	.dc.b	$9E,$A7,$F8,$C9,$1C,$1B,$AE,$84,$70,$7C,$27,$F9,$09,$1C,$1E,$A3
	.dc.b	$F1,$64,$30,$7F,$BD,$E5,$FB,$9F,$FA,$5E,$3F,$99,$46,$3C,$1F,$E7
	.dc.b	$71,$B9,$D7,$57,$FB,$AA,$B5,$3F,$CE,$75,$7A,$4F,$B9,$FF,$18,$C6
	.dc.b	$1B,$19,$8F,$F7,$19,$92,$FD,$BB,$E2,$78,$B6,$33,$18,$3F,$C7,$7C
	.dc.b	$FF,$D2,$B0,$F7,$1F,$7B,$5F,$F1,$CC,$66,$3F,$AE,$8A,$F3,$FC,$37
	.dc.b	$90,$8F,$93,$14,$76,$E1,$FE,$B5,$5D,$8F,$EA,$EE,$1F,$9A,$7E,$43
	.dc.b	$F6,$C7,$91,$F2,$A3,$0D,$5B,$27,$71,$F4,$9B,$BF,$A2,$3D,$18,$FC
	.dc.b	$3F,$C4,$B1,$9F,$D0,$DB,$D9,$18,$9D,$57,$C0,$AD,$09,$E3,$F8,$A3
	.dc.b	$CC,$FA,$ED,$9F,$AA,$DA,$DC,$FC,$3E,$0F,$9B,$B7,$C1,$9F,$A0,$7C
	.dc.b	$5B,$FC,$1F,$D2,$7E,$00,$8F,$CD,$7F,$0B,$F0,$1F,$87,$9F,$A9,$3E
	.dc.b	$78,$C3,$FC,$CA,$3E,$C4,$62,$35,$C9,$FE,$18,$7F,$D0,$3C,$7F,$7E
	.dc.b	$BC,$67,$D4,$3E,$BE,$3F,$84,$F1,$FF,$C2,$F5,$FE,$FB,$6C,$FF,$85
	.dc.b	$5A,$FF,$85,$1B,$0C,$3D,$1E,$8F,$96,$31,$8F,$F8,$AA,$31,$3A,$9F
	.dc.b	$3D,$A7,$CF,$9F,$C9,$8F,$29,$F8,$D3,$FA,$0B,$99,$7C,$3B,$E7,$7F
	.dc.b	$35,$7E,$18,$C4,$F5,$3F,$00,$E2,$35,$3C,$27,$F0,$9F,$98,$BD,$6F
	.dc.b	$EB,$F1,$86,$E3,$F3,$67,$53,$F3,$E3,$13,$A7,$F5,$9F,$72,$FA,$6F
	.dc.b	$C8,$47,$D5,$8C,$3C,$4D,$C6,$9A,$64,$F8,$3B,$FE,$38,$8F,$E3,$70
	.dc.b	$F1,$BF,$45,$B8,$7A,$4E,$C1,$CC,$7E,$10,$87,$D0,$3F,$0C,$78,$0F
	.dc.b	$A5,$C1,$F8,$B7,$47,$FB,$43,$E1,$BF,$6E,$BD,$5F,$F6,$C4,$61,$BD
	.dc.b	$9B,$24,$FF,$9B,$A3,$2D,$F8,$0F,$F3,$6D,$71,$6D,$7F,$1B,$3F,$C0
	.dc.b	$A4,$9F,$EA,$ED,$74,$60,$87,$D6,$C7,$F8,$99,$F1,$63,$E4,$6B,$FC
	.dc.b	$8D,$B1,$A9,$E6,$43,$F5,$31,$DD,$FA,$EF,$E8,$75,$FB,$AD,$D7,$FD
	.dc.b	$16,$D9,$E8,$7F,$31,$D3,$F1,$71,$E0,$25,$F9,$4F,$CC,$ED,$FC,$DF
	.dc.b	$D3,$FC,$CC,$7F,$9D,$6F,$F8,$07,$0F,$E8,$27,$27,$EB,$38,$3E,$AC
	.dc.b	$FA,$0F,$86,$53,$F8,$47,$1F,$F1,$19,$FD,$BF,$27,$FA,$89,$B3,$DC
	.dc.b	$FE,$23,$19,$C3,$F8,$44,$3D,$7F,$E2,$3C,$DF,$EA,$A6,$C6,$67,$71
	.dc.b	$E5,$D3,$E5,$4F,$E7,$9F,$99,$3F,$C2,$A1,$E1,$F8,$89,$1F,$C2,$9F
	.dc.b	$E2,$1D,$8E,$9F,$B5,$5F,$CF,$E3,$15,$F0,$8F,$E6,$65,$73,$7E,$2E
	.dc.b	$7F,$F5,$9B,$7A,$7F,$D7,$0F,$F4,$0A,$E5,$D7,$FD,$16,$30,$4C,$3C
	.dc.b	$1D,$BF,$DB,$7F,$14,$FE,$A4,$D8,$CD,$A7,$2D,$8F,$F0,$A7,$81,$FD
	.dc.b	$8B,$6F,$F0,$73,$63,$38,$7C,$D6,$9F,$44,$FB,$A7,$CB,$2E,$7E,$C7
	.dc.b	$E2,$78,$5F,$19,$FE,$91,$AF,$F7,$EC,$7E,$24,$E6,$7F,$B7,$B5,$E8
	.dc.b	$7E,$8C,$A7,$E2,$CF,$F0,$67,$87,$FE,$17,$C3,$FB,$AB,$63,$FB,$93
	.dc.b	$5F,$EB,$0D,$7F,$BE,$F1,$78,$6F,$FC,$49,$5F,$DE,$E3,$18,$7F,$8A
	.dc.b	$63,$FC,$65,$E2,$7F,$8E,$74,$BC,$5F,$17,$1F,$F3,$0C,$E5,$FF,$1D
	.dc.b	$46,$5F,$41,$EA,$26,$9F,$1D,$FF,$30,$E5,$FE,$C9,$7F,$CE,$3C,$7F
	.dc.b	$EF,$02,$3F,$E5,$D7,$FC,$C4,$47,$07,$F9,$E3,$E0,$7F,$DE,$E4,$70
	.dc.b	$7F,$99,$5F,$F3,$51,$1C,$1F,$E8,$0F,$5B,$FD,$FE,$47,$07,$F9,$E1
	.dc.b	$FF,$54,$91,$C1,$FE,$83,$3F,$DE,$C5,$7F,$D6,$E4,$7F,$D1,$BF,$03
	.dc.b	$FE,$2F,$5E,$C3,$CA,$FF,$24,$4E,$63,$18,$A3,$13,$AA,$FF,$40,$47
	.dc.b	$FA,$14,$8E,$0F,$F5,$9F,$9F,$FE,$53,$F7,$FC,$1E,$92,$31,$3A,$C1
	.dc.b	$5A,$9C,$CF,$FA,$FA,$72,$FF,$AF,$BC,$EC,$A1,$7F,$EC,$1B,$FF,$5F
	.dc.b	$9E,$8A,$E3,$3F,$80,$8C,$11,$FF,$52,$BF,$ED,$62,$38,$3E,$73,$FE
	.dc.b	$E2,$23,$83,$FD,$DC,$76,$3F,$DF,$91,$E5,$8F,$F7,$A1,$23,$FD,$F1
	.dc.b	$EA,$65,$A7,$11,$EA,$7F,$C0,$19,$FF,$80,$11,$9F,$57,$FC,$DA,$DF
	.dc.b	$43,$93,$FE,$82,$8C,$CF,$DF,$6C,$7F,$9B,$FD,$77,$AF,$FE,$87,$E7
	.dc.b	$FC,$24,$63,$3F,$E2,$15,$FE,$8C,$8C,$4E,$AB,$84,$FF,$9E,$7D,$87
	.dc.b	$B3,$88,$FF,$49,$C6,$30,$E7,$1F,$5E,$31,$7F,$93,$6C,$79,$79,$B9
	.dc.b	$3F,$77,$14,$62,$11,$8B,$EB,$18,$9D,$6F,$FE,$F7,$9C,$B7,$0F,$C9
	.dc.b	$FB,$1A,$37,$C5,$2B,$6D,$9F,$CC,$F1,$74,$24,$F4,$7B,$F3,$48,$EE
	.dc.b	$D9,$FD,$0F,$8D,$E4,$24,$F9,$1F,$67,$3F,$EC,$27,$F3,$04,$D3,$FD
	.dc.b	$4D,$19,$7F,$DF,$71,$FB,$BE,$97,$8A,$F2,$3E,$1C,$61,$B1,$99,$FE
	.dc.b	$05,$EC,$34,$FE,$51,$82,$30,$FC,$C6,$CF,$09,$DF,$B4,$F6,$FF,$A5
	.dc.b	$37,$5F,$F0,$AE,$08,$C7,$83,$FC,$45,$B1,$FE,$C1,$6B,$FE,$DA,$6B
	.dc.b	$FB,$D6,$BC,$EB,$B9,$2B,$BC,$E8,$91,$9A,$F0,$36,$BC,$0D,$7C,$0D
	.dc.b	$7C,$0D,$7C,$15,$DA,$31,$5D,$A3,$04,$AE,$ED,$AE,$ED,$7B,$B5,$EE
	.dc.b	$D7,$BC,$72,$23,$C9,$AF,$26,$BC,$9A,$F2,$38,$10,$E0,$53,$81,$4E
	.dc.b	$06,$88,$68,$A6,$8C,$10,$C1,$27,$E3,$47,$91,$B1,$DD,$AF,$46,$BC
	.dc.b	$DA,$F3,$AF,$A2,$4A,$F9,$13,$A2,$46,$6B,$C0,$DA,$F0,$35,$F0,$35
	.dc.b	$F0,$35,$F0,$57,$68,$C5,$76,$8C,$12,$BB,$B6,$BB,$B5,$EE,$D7,$FD
	.dc.b	$4D,$EC,$7F,$75,$7F,$D6,$1E,$9A,$31,$EA,$A3,$13,$FD,$EE,$31,$3C
	.dc.b	$70,$7C,$33,$0C,$36,$FF,$64,$D7,$F6,$4D,$7F,$64,$D7,$F6,$4D,$7F
	.dc.b	$78,$D9,$EF,$3D,$89,$3D,$9A,$F6,$6B,$D9,$AF,$66,$BD,$9B,$1D,$A3
	.dc.b	$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8
	.dc.b	$AF,$27,$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$7E
	.dc.b	$A1,$27,$E4,$B6,$7E,$03,$5F,$4B,$5F,$23,$5F,$0B,$5F,$03,$67,$C1
	.dc.b	$3D,$89,$3D,$9A,$F6,$6B,$D9,$AF,$66,$BD,$9B,$1D,$A3,$A1,$0E,$85
	.dc.b	$3A,$14,$E8,$53,$A1,$5E,$8F,$22,$1F,$E4,$D1,$FB,$37,$F8,$A1,$FD
	.dc.b	$89,$97,$17,$FE,$6D,$6C,$71,$38,$F6,$7F,$95,$E7,$FB,$8F,$35,$9D
	.dc.b	$70,$7F,$99,$5E,$A7,$F9,$75,$C4,$F0,$8C,$8E,$3F,$EF,$D1,$FE,$03
	.dc.b	$E2,$FF,$01,$37,$8F,$73,$AA,$F2,$64,$F5,$3F,$83,$5F,$A5,$B3,$A6
	.dc.b	$BE,$F6,$BF,$E7,$1E,$97,$3E,$23,$FC,$E6,$D8,$CE,$FE,$86,$DF,$C7
	.dc.b	$FA,$DE,$18,$9A,$4A,$E1,$CD,$CC,$8E,$0F,$87,$2F,$AD,$E6,$48,$CB
	.dc.b	$E3,$EB,$91,$E3,$6E,$1F,$E6,$2E,$CF,$57,$6F,$F9,$36,$31,$5F,$42
	.dc.b	$31,$AF,$CA,$E1,$F4,$92,$7C,$6F,$C3,$F8,$EF,$E6,$1F,$01,$FE,$B1
	.dc.b	$6D,$7E,$15,$AF,$E3,$DA,$FF,$A7,$7F,$52,$7C,$77,$FA,$7E,$BF,$D7
	.dc.b	$51,$89,$46,$27,$51,$F9,$19,$D4,$F0,$F8,$FC,$CF,$E7,$65,$18,$8F
	.dc.b	$03,$54,$69,$EE,$7E,$2B,$A7,$5C,$1F,$A5,$BF,$D1,$46,$0E,$5D,$1F
	.dc.b	$9D,$8F,$CD,$12,$BF,$43,$D3,$DF,$A3,$FC,$B1,$3F,$99,$24,$FE,$AE
	.dc.b	$32,$FE,$6B,$97,$FA,$F4,$FE,$2F,$7F,$2C,$FE,$D6,$72,$EE,$32,$FE
	.dc.b	$67,$97,$2F,$FB,$3E,$BF,$CF,$F3,$A8,$FF,$66,$CE,$9F,$F6,$87,$CA
	.dc.b	$FF,$9B,$D7,$C2,$7F,$BB,$63,$0D,$EB,$FF,$6B,$36,$76,$7A,$F8,$FF
	.dc.b	$CF,$11,$FB,$72,$46,$67,$6F,$F0,$FD,$BF,$DC,$1E,$4F,$CE,$8F,$F6
	.dc.b	$19,$92,$9F,$EE,$5E,$0F,$E3,$11,$FC,$24,$91,$F4,$0D,$9F,$E8,$1D
	.dc.b	$FA,$BE,$18,$FE,$14,$47,$07,$F1,$DC,$BE,$17,$C2,$7B,$49,$19,$C7
	.dc.b	$FD,$ED,$1F,$C4,$C9,$1E,$D6,$BF,$CB,$E7,$E6,$4F,$C7,$24,$73,$3E
	.dc.b	$81,$FC,$D7,$83,$FD,$3D,$1F,$C7,$C9,$1F,$25,$FE,$44,$C3,$E6,$14
	.dc.b	$E5,$A7,$46,$C7,$F8,$CE,$7F,$D5,$51,$B1,$F1,$4F,$8A,$7A,$5F,$CE
	.dc.b	$E4,$6A,$32,$FF,$3C,$9E,$13,$B8,$D3,$63,$A7,$8D,$FE,$43,$6C,$FF
	.dc.b	$4B,$3A,$46,$6B,$FD,$65,$1E,$7C,$7D,$4F,$8B,$0F,$4B,$F0,$08,$7B
	.dc.b	$8F,$61,$FD,$2E,$38,$B7,$1E,$84,$77,$3D,$1E,$47,$43,$C1,$3C,$A3
	.dc.b	$51,$F0,$FC,$3F,$8B,$F8,$1F,$F6,$5C,$BE,$54,$62,$7F,$B8,$46,$14
	.dc.b	$EF,$1F,$F2,$EC,$60,$FE,$ED,$C8,$F1,$DC,$7F,$E5,$98,$C3,$63,$31
	.dc.b	$F2,$CC,$9E,$D9,$F5,$1F,$04,$F7,$E7,$EB,$26,$9B,$8F,$BF,$2F,$D4
	.dc.b	$3F,$B6,$47,$FB,$CE,$3C,$B1,$89,$FF,$38,$46,$FC,$5F,$C3,$C6,$27
	.dc.b	$53,$E2,$DB,$E4,$DF,$8C,$C6,$1E,$27,$CB,$7E,$BB,$F0,$FC,$BF,$3C
	.dc.b	$7F,$26,$DA,$34,$FF,$40,$61,$D8,$98,$75,$9F,$9B,$A3,$F2,$5F,$F4
	.dc.b	$66,$7E,$A8,$C3,$89,$F8,$71,$88,$D4,$F0,$7E,$BE,$9F,$5D,$F9,$35
	.dc.b	$F2,$A3,$0F,$C6,$6C,$70,$7F,$D3,$4F,$E1,$A9,$F4,$DF,$82,$AF,$D8
	.dc.b	$F0,$7E,$D6,$31,$1F,$B3,$6C,$66,$3C,$27,$E6,$36,$FB,$8D,$3F,$80
	.dc.b	$5F,$E3,$70,$F0,$C7,$9D,$AF,$FA,$BD,$F3,$B9,$8E,$AF,$D9,$3F,$4E
	.dc.b	$43,$B1,$F7,$27,$C2,$EC,$FF,$62,$8F,$88,$FC,$F3,$86,$BF,$EC,$8F
	.dc.b	$13,$F8,$8C,$63,$CB,$FF,$62,$CF,$1C,$3F,$CB,$91,$89,$D7,$99,$FE
	.dc.b	$CC,$6D,$F2,$AD,$CE,$5F,$E4,$A7,$F9,$34,$97,$FD,$5D,$AE,$8C,$10
	.dc.b	$F8,$8F,$DD,$3E,$46,$3F,$74,$FC,$0C,$7C,$B8,$CB,$CC,$87,$D7,$7B
	.dc.b	$BF,$74,$E3,$97,$DD,$8F,$F6,$F3,$67,$F1,$19,$FF,$36,$9F,$14,$78
	.dc.b	$0F,$BC,$57,$F2,$C6,$8F,$CD,$BF,$D0,$0F,$D2,$EB,$FE,$EF,$E7,$C9
	.dc.b	$23,$E0,$61,$E7,$29,$FC,$B6,$BF,$D0,$D1,$FC,$0E,$1F,$3D,$FB,$72
	.dc.b	$47,$F0,$AE,$3F,$E9,$B1,$FB,$8E,$17,$07,$E0,$AF,$B1,$F1,$8E,$C7
	.dc.b	$83,$89,$3C,$31,$F9,$33,$A3,$F8,$DF,$27,$F1,$AF,$23,$F8,$41,$33
	.dc.b	$F1,$4A,$3C,$58,$F8,$63,$91,$E2,$21,$82,$63,$E2,$8E,$0D,$C3,$E1
	.dc.b	$9F,$CD,$8B,$CD,$FA,$DA,$FE,$12,$D7,$F8,$43,$5F,$E9,$C3,$EF,$FF
	.dc.b	$A4,$74,$FF,$74,$6C,$64,$FE,$15,$19,$7E,$36,$9F,$15,$B8,$F9,$A3
	.dc.b	$11,$F3,$71,$F4,$3A,$C7,$E5,$C6,$23,$53,$C1,$F8,$71,$97,$E3,$71
	.dc.b	$7F,$7E,$AF,$EB,$DD,$5F,$95,$9F,$E3,$CD,$7F,$8E,$F4,$7F,$80,$27
	.dc.b	$FB,$27,$17,$F8,$43,$CF,$B4,$69,$1F,$31,$0F,$2E,$7E,$D2,$57,$FB
	.dc.b	$27,$79,$3F,$E5,$36,$BF,$E3,$79,$FF,$0F,$C7,$F9,$74,$91,$FE,$37
	.dc.b	$7F,$DA,$5E,$D4,$8F,$F2,$E3,$5E,$41,$FF,$33,$11,$FF,$20,$3F,$ED
	.dc.b	$BF,$4A,$47,$F9,$95,$AF,$FB,$DF,$C7,$FF,$75,$47,$FB,$9C,$91,$FE
	.dc.b	$5F,$7F,$CD,$04,$7F,$D3,$41,$C3,$FE,$42,$F6,$7F,$DD,$44,$7F,$CE
	.dc.b	$8F,$F9,$EC,$8F,$F9,$55,$AF,$F9,$08,$FF,$44,$F2,$3D,$7F,$F4,$69
	.dc.b	$23,$46,$9F,$F4,$D1,$1F,$F4,$C3,$5F,$F2,$AB,$FE,$FD,$23,$FE,$87
	.dc.b	$7F,$C8,$D6,$F1,$3F,$CA,$71,$86,$CF,$FA,$06,$30,$F9,$4F,$86,$FF
	.dc.b	$A3,$48,$E0,$FF,$66,$79,$7C,$9F,$F6,$53,$5F,$F6,$49,$CC,$C9,$4F
	.dc.b	$F6,$E7,$99,$B7,$FD,$9A,$DF,$33,$FD,$B9,$5F,$ED,$B9,$E0,$3C,$D6
	.dc.b	$5C,$5B,$1A,$6C,$FF,$BB,$E7,$B7,$93,$FC,$81,$FD,$25,$7F,$BE,$5A
	.dc.b	$FD,$78,$D7,$17,$F9,$BD,$B1,$F5,$BC,$EF,$E4,$15,$C2,$3F,$CF,$F1
	.dc.b	$8A,$FF,$3B,$36,$38,$70,$FA,$27,$FC,$F5,$E8,$BD,$0C,$57,$E0,$A3
	.dc.b	$18,$75,$84,$62,$B8,$D7,$09,$F4,$7A,$35,$FF,$60,$35,$64,$8C,$37
	.dc.b	$0E,$D8,$3C,$FE,$0D,$8C,$C6,$0E,$3E,$8C,$8F,$35,$23,$19,$7B,$BD
	.dc.b	$2E,$64,$AE,$EF,$8B,$C5,$91,$C9,$B3,$FE,$9F,$EF,$FE,$8C,$DF,$03
	.dc.b	$FA,$FB,$7A,$FE,$34,$EB,$6F,$F6,$E4,$FE,$70,$8E,$0F,$DF,$BF,$EE
	.dc.b	$92,$38,$3F,$67,$E9,$BC,$2F,$E9,$90,$8C,$65,$FD,$7D,$BD,$DF,$EB
	.dc.b	$56,$CF,$2A,$E3,$18,$8F,$E0,$8D,$7F,$D8,$ED,$7B,$47,$02,$57,$76
	.dc.b	$D7,$76,$CF,$79,$D1,$27,$4D,$8D,$46,$08,$E0,$87,$9A,$7A,$B6,$79
	.dc.b	$36,$78,$B4,$E2,$7B,$27,$8B,$67,$8B,$67,$8B,$4E,$27,$C2,$8D,$B6
	.dc.b	$36,$D3,$75,$F2,$49,$5E,$08,$D1,$E5,$21,$DC,$A6,$8C,$10,$C1,$23
	.dc.b	$AD,$79,$1B,$5E,$46,$CF,$92,$BC,$51,$88,$EE,$D7,$B3,$5E,$D1,$C0
	.dc.b	$95,$DD,$B5,$DD,$B3,$DE,$74,$49,$D3,$63,$51,$82,$38,$21,$E6,$9E
	.dc.b	$AD,$9E,$4D,$9E,$2D,$38,$9E,$C9,$FD,$4F,$99,$FE,$92,$8F,$EB,$ED
	.dc.b	$7F,$D1,$BE,$5F,$F6,$27,$FD,$3E,$7E,$CB,$EE,$7F,$9E,$B7,$0D,$3F
	.dc.b	$B3,$8C,$BF,$AC,$6B,$FA,$C6,$BB,$8E,$38,$A7,$2D,$8C,$C6,$08,$E0
	.dc.b	$93,$FB,$F8,$CB,$63,$2F,$9C,$C1,$0C,$10,$F4,$3C,$1A,$6C,$A6,$CF
	.dc.b	$63,$B6,$9B,$29,$B3,$E0,$19,$29,$99,$F8,$0E,$5A,$E6,$7D,$AD,$9F
	.dc.b	$03,$C8,$F2,$90,$F0,$94,$D1,$82,$18,$24,$F8,$1E,$0D,$76,$D7,$73
	.dc.b	$E4,$76,$D7,$6D,$77,$3E,$89,$E4,$49,$D3,$63,$51,$82,$38,$24,$FC
	.dc.b	$78,$CB,$63,$2F,$C7,$30,$43,$04,$3D,$2F,$06,$9B,$29,$B3,$DA,$ED
	.dc.b	$A6,$CA,$6C,$7E,$A2,$F9,$04,$30,$FD,$84,$E6,$30,$7F,$95,$1C,$46
	.dc.b	$A7,$8B,$C3,$E3,$AD,$6D,$AF,$F9,$86,$91,$88,$F0,$C6,$1E,$E6,$FD
	.dc.b	$5D,$3E,$86,$C1,$A6,$B7,$FF,$65,$5F,$F8,$0C,$95,$FE,$6C,$8C,$BF
	.dc.b	$E6,$EF,$61,$EF,$7F,$92,$AB,$F8,$84,$63,$3F,$1D,$7C,$68,$C4,$EA
	.dc.b	$B8,$4F,$F8,$DF,$96,$BF,$E7,$FE,$94,$62,$3E,$7C,$63,$0E,$B1,$FE
	.dc.b	$3C,$9D,$3F,$E3,$EF,$1A,$D7,$FA,$4E,$B6,$DC,$3F,$C9,$33,$AE,$A9
	.dc.b	$5A,$6C,$FF,$93,$BC,$4E,$64,$9E,$6F,$83,$E4,$7E,$64,$79,$1B,$3F
	.dc.b	$CC,$DB,$8F,$FA,$59,$AF,$FA,$77,$C0,$FF,$31,$CF,$F9,$83,$C0,$F5
	.dc.b	$C6,$27,$FC,$C3,$7F,$03,$DB,$7B,$9F,$01,$B1,$98,$C1,$F1,$3F,$52
	.dc.b	$76,$FE,$63,$DD,$FE,$8D,$18,$C1,$18,$9D,$47,$FA,$B2,$75,$3C,$23
	.dc.b	$27,$FA,$D6,$3F,$D6,$D1,$98,$E6,$71,$36,$7E,$63,$A2,$BF,$AC,$6F
	.dc.b	$3F,$EA,$67,$F4,$D3,$C8,$FD,$17,$46,$88,$E4,$FD,$1B,$CB,$A2,$46
	.dc.b	$4F,$D1,$16,$BF,$44,$E7,$93,$F1,$97,$FE,$6F,$D3,$1C,$BF,$EA,$68
	.dc.b	$C1,$F8,$88,$E1,$CB,$9B,$FF,$48,$35,$FC,$D3,$5F,$E0,$FE,$DB,$DA
	.dc.b	$FE,$13,$EE,$FF,$77,$8C,$4A,$31,$3A,$8F,$F6,$EC,$EA,$7F,$87,$E9
	.dc.b	$99,$FE,$00,$71,$3F,$88,$39,$7F,$83,$B5,$FA,$07,$C2,$3F,$87,$6D
	.dc.b	$FC,$72,$3E,$89,$23,$F8,$A1,$B3,$E7,$94,$FE,$4B,$B7,$17,$71,$E6
	.dc.b	$75,$B7,$D0,$8C,$12,$3F,$DF,$3E,$DF,$EA,$BE,$AF,$FC,$34,$D8,$CC
	.dc.b	$FF,$B6,$27,$2D,$8F,$78,$F7,$BC,$04,$72,$FF,$89,$7F,$2D,$FF,$5D
	.dc.b	$43,$F0,$0C,$79,$7F,$CF,$1B,$19,$3F,$9F,$9F,$D0,$27,$8C,$62,$3F
	.dc.b	$DD,$7E,$33,$F0,$3F,$F7,$6C,$E7,$8D,$39,$7A,$57,$F4,$A7,$35,$C2
	.dc.b	$3B,$C6,$0D,$BF,$D4,$3C,$37,$93,$FB,$98,$FE,$6C,$D7,$CF,$1A,$EA
	.dc.b	$FF,$1F,$B6,$3D,$1E,$CB,$D8,$7D,$6F,$F5,$CF,$82,$CC,$62,$91,$89
	.dc.b	$D4,$7C,$4A,$E5,$18,$8F,$66,$8F,$13,$FC,$97,$A7,$C2,$6C,$66,$3F
	.dc.b	$CB,$06,$4F,$F2,$D5,$F9,$7C,$9F,$ED,$FE,$9F,$91,$B8,$78,$73,$FF
	.dc.b	$29,$29,$D3,$4F,$8B,$AF,$C2,$DB,$E3,$47,$CB,$6A,$DB,$E3,$3D,$7D
	.dc.b	$6E,$EF,$F8,$03,$2F,$91,$18,$84,$61,$F0,$1B,$C5,$3B,$8E,$D7,$E0
	.dc.b	$BE,$F1,$F6,$23,$07,$9A,$FE,$B7,$AD,$F3,$67,$BE,$08,$C4,$EA,$7E
	.dc.b	$B9,$F1,$63,$86,$9F,$92,$F2,$CB,$F4,$2F,$CB,$6F,$FD,$E5,$B3,$A6
	.dc.b	$CE,$EF,$F2,$BE,$8F,$F6,$0F,$B5,$FC,$F3,$63,$3D,$5D,$A7,$2D,$8F
	.dc.b	$A6,$7F,$A5,$F6,$F8,$2F,$D3,$E0,$FA,$C7,$E1,$2F,$FC,$43,$83,$0E
	.dc.b	$FD,$DC,$EF,$F5,$55,$A9,$CD,$6E,$7C,$FD,$FF,$03,$ED,$7F,$D4,$72
	.dc.b	$BF,$F1,$E2,$8C,$C6,$1B,$3C,$67,$D3,$50,$FB,$18,$FA,$1F,$E0,$99
	.dc.b	$BC,$7F,$9F,$1F,$AA,$AF,$E1,$73,$97,$F6,$06,$27,$7D,$3F,$EB,$89
	.dc.b	$D5,$71,$3F,$19,$F5,$DF,$E5,$2F,$C7,$7F,$D7,$BC,$9F,$BB,$8C,$63
	.dc.b	$FA,$56,$D7,$01,$5F,$EC,$0E,$0F,$A3,$1F,$94,$CF,$D4,$FB,$4F,$C7
	.dc.b	$FD,$37,$F9,$83,$F4,$EE,$FC,$E8,$C6,$5F,$B3,$6D,$7E,$85,$AF,$E9
	.dc.b	$BE,$6B,$F9,$5F,$5F,$A1,$B5,$FC,$2D,$B3,$FA,$F6,$BF,$98,$E3,$7C
	.dc.b	$D7,$E3,$6B,$F2,$6D,$9F,$C1,$B5,$E0,$D7,$DC,$7E,$47,$AF,$F5,$FA
	.dc.b	$7C,$08,$FC,$5F,$D0,$71,$BF,$C3,$37,$03,$4E,$9B,$7C,$A2,$61,$D2
	.dc.b	$FB,$61,$D5,$FC,$F1,$83,$FD,$F1,$EC,$BE,$33,$FD,$13,$E4,$FF,$48
	.dc.b	$9F,$F4,$7C,$E7,$D6,$FF,$48,$CE,$65,$39,$51,$98,$C3,$6B,$AF,$7F
	.dc.b	$F5,$C9,$CB,$63,$35,$E0,$6B,$F8,$58,$F5,$91,$FE,$5A,$D7,$F4,$6D
	.dc.b	$7F,$78,$FF,$06,$CF,$F8,$8F,$9B,$FA,$E6,$ED,$FC,$31,$B5,$F4,$A1
	.dc.b	$3F,$97,$D3,$FC,$B7,$7D,$6B,$C8,$78,$8F,$C1,$D7,$02,$D7,$EF,$37
	.dc.b	$EB,$3F,$84,$9D,$3F,$8F,$3F,$D7,$BF,$8E,$FE,$55,$B1,$F9,$56,$9B
	.dc.b	$11,$FA,$4E,$3F,$F6,$2F,$EB,$F6,$4D,$25,$79,$E3,$E3,$94,$F8,$E7
	.dc.b	$EE,$47,$E3,$FF,$B0,$E9,$FF,$68,$B6,$33,$3B,$8F,$E1,$1D,$BF,$C4
	.dc.b	$3E,$2A,$CF,$FB,$61,$B4,$69,$F3,$79,$3F,$19,$1E,$23,$CA,$43,$F6
	.dc.b	$5D,$FF,$D2,$BE,$BB,$F9,$2D,$9A,$47,$FA,$3F,$87,$F8,$0A,$B1,$F7
	.dc.b	$1B,$C1,$FC,$E0,$FB,$9D,$1F,$E1,$9E,$2B,$C7,$F8,$0A,$FE,$9B,$7C
	.dc.b	$5F,$E8,$3E,$07,$56,$C6,$5F,$E1,$11,$97,$FC,$6F,$E2,$FF,$62,$FC
	.dc.b	$23,$FA,$E9,$23,$FC,$63,$87,$FB,$6E,$3F,$C4,$7E,$3F,$C9,$9F,$ED
	.dc.b	$87,$C9,$2F,$06,$9C,$1F,$1C,$FC,$A9,$73,$F6,$3F,$E0,$D3,$F1,$25
	.dc.b	$E6,$FF,$18,$F9,$1F,$92,$8C,$1F,$85,$39,$9F,$54,$9E,$27,$F8,$9D
	.dc.b	$E6,$7F,$84,$7C,$9E,$4D,$32,$7F,$82,$9C,$BF,$E5,$AF,$2B,$FC,$B6
	.dc.b	$FC,$26,$BF,$E3,$C6,$BB,$DF,$F0,$F5,$FE,$42,$8F,$82,$7F,$85,$BC
	.dc.b	$BF,$90,$FF,$88,$DA,$E4,$D9,$EC,$26,$5F,$19,$FC,$85,$7F,$98,$FF
	.dc.b	$3F,$FF,$0B,$F9,$BF,$E1,$A6,$C1,$A7,$F9,$E8,$93,$FE,$77,$F3,$FF
	.dc.b	$D0,$31,$88,$FF,$54,$FC,$5F,$FA,$5C,$8F,$FA,$31,$FF,$47,$D3,$F5
	.dc.b	$5A,$49,$FF,$4A,$C6,$5F,$F4,$9F,$B1,$FE,$A1,$23,$83,$FD,$56,$FF
	.dc.b	$A2,$5F,$F5,$48,$71,$EB,$3A,$93,$93,$89,$FE,$91,$87,$66,$AB,$FD
	.dc.b	$D2,$49,$D4,$65,$FF,$53,$FC,$67,$FA,$D4,$8E,$0F,$F6,$7F,$B3,$E0
	.dc.b	$8F,$F6,$9C,$1A,$47,$FD,$C3,$8B,$E2,$BF,$DC,$C4,$8C,$BF,$EE,$D2
	.dc.b	$3F,$EA,$9F,$05,$EE,$7F,$95,$27,$33,$FE,$DC,$7F,$D9,$44,$7F,$D7
	.dc.b	$EF,$C6,$D3,$E2,$B6,$3C,$47,$FB,$61,$C1,$F3,$32,$DC,$7C,$43,$67
	.dc.b	$FB,$88,$A7,$D0,$D3,$E6,$B8,$1E,$EF,$FB,$12,$BE,$D3,$63,$FD,$88
	.dc.b	$D7,$FD,$92,$D7,$ED,$AF,$0F,$C7,$1A,$6B,$A6,$BE,$06,$FB,$EF,$7F
	.dc.b	$FC,$49,$3F,$ED,$38,$C6,$BF,$E5,$D8,$C4,$E9,$FF,$6D,$CE,$BD,$0F
	.dc.b	$C9,$7C,$03,$E0,$75,$5F,$90,$8C,$61,$CE,$3F,$DC,$71,$8A,$E3,$5C
	.dc.b	$27,$FD,$81,$E2,$DE,$D3,$49,$87,$39,$FC,$EF,$13,$E0,$B9,$36,$75
	.dc.b	$3F,$E6,$B9,$C9,$AF,$83,$93,$AA,$46,$32,$FD,$74,$7E,$B6,$75,$CF
	.dc.b	$D8,$8F,$1E,$5F,$09,$1C,$1E,$2F,$84,$79,$1E,$F6,$C6,$63,$07,$C3
	.dc.b	$EF,$FF,$79,$63,$FE,$42,$26,$9F,$ED,$47,$F7,$FA,$FE,$F9,$B1,$F0
	.dc.b	$39,$BE,$4B,$67,$4F,$F0,$1E,$8A,$FF,$06,$3F,$85,$D7,$E8,$C8,$FE
	.dc.b	$7D,$FD,$13,$D7,$A3,$8B,$F9,$F6,$AF,$85,$78,$7F,$D0,$23,$0D,$8C
	.dc.b	$E9,$FE,$08,$9C,$B6,$33,$E4,$FF,$1F,$6C,$F9,$CD,$C6,$6B,$A1,$2B
	.dc.b	$93,$63,$51,$82,$38,$24,$F5,$AE,$EF,$C5,$6B,$FC,$B1,$AF,$88,$E1
	.dc.b	$5D,$DB,$5D,$9B,$5D,$9E,$93,$B6,$CE,$DB,$3B,$7C,$13,$B6,$CE,$DB
	.dc.b	$3B,$7C,$87,$23,$C2,$43,$04,$38,$99,$29,$93,$99,$82,$18,$25,$7C
	.dc.b	$42,$57,$9E,$32,$D8,$CD,$7A,$67,$2D,$8C,$D7,$AD,$B3,$E5,$37,$19
	.dc.b	$AE,$84,$AE,$4D,$8D,$46,$08,$E0,$93,$D6,$BB,$BF,$05,$AF,$C0,$6B
	.dc.b	$E2,$38,$57,$76,$D7,$66,$D7,$67,$A4,$ED,$B3,$B6,$CE,$DF,$04,$FE
	.dc.b	$AF,$D4,$FF,$5C,$47,$E8,$DA,$FF,$AC,$FD,$6F,$F0,$33,$FE,$BE,$F5
	.dc.b	$51,$8C,$5E,$43,$2C,$36,$FF,$60,$D7,$F4,$8D,$7F,$48,$D7,$F4,$8D
	.dc.b	$7F,$81,$36,$7B,$CF,$62,$4F,$66,$BD,$9A,$F6,$6B,$D9,$AF,$66,$C7
	.dc.b	$68,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8,$87,$22,$9C,$8A
	.dc.b	$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA,$70,$47,$04
	.dc.b	$9F,$A8,$49,$F9,$2D,$9F,$80,$D7,$D2,$D7,$C8,$D7,$C2,$D7,$C0,$D9
	.dc.b	$F0,$4F,$62,$4F,$66,$BD,$9A,$F6,$6B,$D9,$AF,$66,$C7,$68,$E8,$43
	.dc.b	$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8,$87,$F6,$9F,$9B,$7D,$07,$F9
	.dc.b	$C3,$0F,$F3,$B4,$62,$7F,$62,$7F,$59,$71,$1A,$9E,$13,$9F,$9F,$AF
	.dc.b	$56,$BF,$E7,$CA,$46,$21,$18,$7B,$9F,$DF,$78,$BF,$D0,$8E,$0F,$F2
	.dc.b	$B7,$0F,$18,$CE,$1F,$E6,$33,$FD,$1D,$F1,$FF,$E2,$DA,$E0,$1C,$46
	.dc.b	$BA,$FC,$51,$89,$FE,$ED,$CA,$F4,$7F,$CB,$17,$FE,$99,$8C,$78,$5D
	.dc.b	$1B,$19,$9D,$D7,$F7,$FF,$8D,$7C,$67,$F1,$7A,$FF,$51,$46,$38,$7C
	.dc.b	$AD,$8C,$CE,$EB,$FD,$11,$D9,$5F,$F1,$07,$77,$FA,$2B,$0F,$6B,$6F
	.dc.b	$FC,$83,$5B,$AF,$F5,$5E,$0F,$8D,$C4,$23,$18,$78,$A7,$31,$8B,$FF
	.dc.b	$25,$F8,$1C,$31,$FF,$27,$DF,$82,$30,$67,$79,$1E,$46,$D7,$F9,$53
	.dc.b	$E3,$B9,$12,$7D,$2F,$93,$C4,$FF,$38,$CF,$F9,$BA,$7E,$CC,$63,$3F
	.dc.b	$F1,$53,$5F,$F2,$17,$43,$DE,$FF,$64,$FC,$6F,$FB,$3E,$31,$28,$C4
	.dc.b	$EA,$3F,$C6,$93,$A9,$FF,$6B,$7D,$8F,$E4,$DB,$D3,$BA,$FF,$28,$4E
	.dc.b	$5F,$F1,$E3,$5F,$D6,$75,$6D,$FD,$03,$5F,$CF,$B7,$AB,$F5,$D7,$FB
	.dc.b	$0B,$C1,$2B,$D0,$7E,$B4,$A7,$E9,$7A,$BA,$75,$7E,$9A,$BF,$D8,$84
	.dc.b	$9F,$C9,$73,$7F,$67,$8F,$C6,$ED,$CA,$7F,$0B,$39,$77,$19,$5C,$7F
	.dc.b	$8F,$AF,$C6,$CE,$63,$13,$F9,$68,$E1,$EC,$BF,$2C,$FF,$7D,$FC,$7F
	.dc.b	$FC,$01,$8C,$52,$31,$3A,$8F,$F3,$34,$EA,$7F,$DB,$C4,$8F,$F4,$11
	.dc.b	$C4,$AF,$F1,$37,$FD,$12,$7D,$22,$9F,$48,$A7,$D2,$2E,$BF,$C4,$7C
	.dc.b	$5F,$A5,$18,$D8,$D3,$8E,$BF,$C5,$27,$FD,$E4,$48,$F2,$BF,$E7,$FC
	.dc.b	$9E,$2B,$C1,$FF,$0C,$7C,$7E,$63,$12,$8C,$4E,$AF,$FD,$3D,$3C,$A3
	.dc.b	$11,$F8,$11,$F1,$FF,$E8,$D5,$8F,$F2,$88,$C3,$63,$E4,$C7,$FA,$4F
	.dc.b	$D9,$7C,$2F,$FB,$D6,$5F,$1D,$98,$C3,$7C,$4F,$86,$D9,$D9,$FE,$9D
	.dc.b	$FD,$4D,$F0,$FF,$E3,$59,$F2,$46,$34,$DD,$7F,$AE,$63,$13,$AA,$E1
	.dc.b	$3F,$CC,$BC,$5F,$23,$70,$FF,$22,$F8,$DF,$CC,$A3,$F9,$84,$7C,$1C
	.dc.b	$7E,$7F,$D6,$7F,$91,$7E,$3F,$E7,$36,$BE,$5F,$12,$31,$1F,$7A,$B8
	.dc.b	$8A,$F8,$5E,$CF,$F5,$AD,$FE,$3B,$77,$34,$C7,$4F,$E2,$3E,$2B,$E5
	.dc.b	$37,$1F,$F6,$C5,$FD,$08,$C4,$7E,$3A,$30,$A7,$DA,$C3,$E1,$14,$E1
	.dc.b	$6F,$99,$E7,$0A,$FB,$B1,$86,$C7,$53,$C7,$3C,$5C,$1F,$EE,$DD,$BE
	.dc.b	$73,$E0,$21,$9D,$FE,$AB,$F2,$76,$7C,$D7,$CC,$8F,$F3,$B4,$63,$6F
	.dc.b	$99,$1F,$4A,$75,$1F,$A6,$6C,$BE,$0B,$FC,$FD,$FA,$57,$E9,$DB,$7F
	.dc.b	$69,$AF,$85,$AF,$83,$D2,$E6,$4C,$3C,$31,$E7,$8F,$32,$CB,$DE,$FB
	.dc.b	$78,$FF,$1F,$9F,$BA,$3F,$81,$CE,$9F,$6F,$8D,$E8,$26,$1D,$A3,$2D
	.dc.b	$3F,$81,$FC,$37,$FA,$67,$F4,$7F,$F4,$B4,$60,$38,$D5,$3B,$9F,$CC
	.dc.b	$B6,$BC,$C7,$E0,$CF,$2E,$3E,$77,$D8,$FA,$7B,$9F,$83,$BF,$05,$FD
	.dc.b	$36,$1F,$C7,$AD,$5A,$9C,$C7,$F2,$28,$C4,$FE,$94,$E0,$E3,$A3,$FD
	.dc.b	$61,$C9,$FC,$33,$EC,$FF,$94,$FE,$D1,$8E,$64,$63,$0F,$C8,$B7,$1F
	.dc.b	$CD,$B5,$F5,$F1,$7D,$22,$3E,$FC,$3F,$8C,$FE,$01,$FC,$DB,$87,$B3
	.dc.b	$70,$FC,$EB,$6B,$F4,$ED,$7F,$4D,$8B,$F2,$2F,$DC,$57,$E9,$1B,$3F
	.dc.b	$AE,$6B,$C1,$B0,$F8,$5F,$E1,$5A,$FD,$C6,$C7,$F3,$A6,$D7,$F0,$F6
	.dc.b	$BE,$E3,$ED,$1F,$63,$3F,$7B,$C5,$FA,$6F,$FB,$86,$11,$89,$F1,$36
	.dc.b	$33,$1F,$5E,$7F,$0E,$7E,$E8,$F0,$14,$F0,$1E,$2E,$5F,$CF,$38,$3F
	.dc.b	$DF,$1E,$27,$FA,$13,$D0,$7B,$9F,$EF,$99,$FF,$19,$36,$33,$1F,$DA
	.dc.b	$BD,$CF,$EC,$5E,$A7,$F6,$C9,$CC,$FF,$BF,$E3,$15,$C4,$91,$C6,$FC
	.dc.b	$11,$FB,$53,$FA,$09,$23,$FA,$33,$5D,$18,$21,$F4,$73,$64,$F3,$7E
	.dc.b	$C4,$61,$B1,$93,$E9,$16,$7A,$90,$FC,$BD,$78,$5F,$AC,$73,$8F,$C1
	.dc.b	$34,$FC,$11,$4D,$BD,$5F,$C7,$0F,$8C,$FD,$56,$FF,$E1,$46,$C6,$67
	.dc.b	$F3,$0D,$3F,$3E,$3F,$7D,$FD,$2B,$63,$F4,$AD,$37,$1F,$11,$F6,$1F
	.dc.b	$93,$BF,$63,$4F,$DC,$8F,$DA,$7F,$7D,$CD,$FE,$28,$6C,$66,$F7,$1D
	.dc.b	$63,$F8,$7F,$57,$F1,$39,$F8,$1D,$3F,$C3,$B4,$D3,$63,$F8,$7C,$EE
	.dc.b	$79,$57,$8C,$F3,$F1,$7F,$AD,$B2,$FA,$57,$FC,$B7,$3F,$E5,$9D,$9F
	.dc.b	$80,$6C,$7C,$1B,$FE,$AF,$39,$8F,$F0,$3F,$77,$E0,$E6,$C7,$E1,$DB
	.dc.b	$8F,$F8,$EE,$75,$1E,$2E,$DF,$DA,$E3,$F2,$32,$F2,$4F,$F8,$4B,$BF
	.dc.b	$6F,$F5,$7C,$7A,$3F,$D5,$5A,$E7,$1F,$EC,$50,$FD,$4E,$92,$7F,$B2
	.dc.b	$46,$5F,$ED,$CF,$23,$FB,$71,$7C,$0F,$EE,$95,$FE,$53,$F0,$FC,$75
	.dc.b	$F2,$CF,$CD,$9F,$A8,$29,$FE,$10,$F1,$3F,$24,$FF,$82,$9A,$E4,$FC
	.dc.b	$C1,$92,$9F,$17,$9B,$F8,$0B,$D0,$9E,$07,$F8,$8B,$9B,$FC,$53,$1C
	.dc.b	$0F,$E6,$FE,$37,$66,$99,$9F,$F1,$AB,$FE,$70,$3F,$C3,$BE,$3F,$F9
	.dc.b	$EB,$3F,$F1,$DC,$64,$FF,$14,$C6,$5F,$F2,$2F,$91,$F5,$35,$FF,$21
	.dc.b	$EB,$FE,$4A,$7E,$44,$7C,$6E,$0F,$B7,$5E,$BE,$1F,$D2,$F9,$7F,$E5
	.dc.b	$FF,$8D,$FF,$2E,$60,$69,$2B,$FC,$63,$BF,$FA,$3A,$3F,$CD,$84,$8C
	.dc.b	$9C,$4F,$F5,$0F,$C6,$FF,$9F,$3D,$7F,$F4,$AB,$60,$D3,$FD,$17,$6F
	.dc.b	$8D,$A4,$9F,$F3,$1F,$81,$FC,$8A,$39,$7C,$3C,$9F,$F5,$7B,$5F,$F0
	.dc.b	$37,$AA,$F5,$7F,$D4,$B3,$FE,$A8,$24,$FF,$B5,$DF,$F5,$A1,$1F,$F0
	.dc.b	$71,$FE,$B1,$70,$FF,$A6,$FE,$37,$D4,$47,$FD,$7E,$FF,$B1,$93,$48
	.dc.b	$FF,$B2,$39,$BF,$D6,$F1,$FE,$D6,$F2,$24,$FF,$B5,$5A,$FF,$87,$FE
	.dc.b	$05,$ED,$FF,$36,$8C,$35,$FF,$72,$EB,$FD,$F5,$FF,$72,$D7,$CF,$23
	.dc.b	$FE,$D4,$6B,$F1,$B2,$FF,$64,$C7,$FB,$30,$91,$FE,$FB,$7F,$DB,$04
	.dc.b	$7E,$2E,$6F,$73,$8C,$61,$AF,$F8,$DF,$DD,$FA,$4F,$94,$C5,$E9,$FF
	.dc.b	$6C,$C6,$AF,$FD,$8C,$D8,$FA,$7E,$13,$CB,$E1,$08,$C3,$D4,$FF,$71
	.dc.b	$B8,$AF,$F6,$83,$63,$EE,$FB,$CF,$8A,$FF,$40,$C6,$1A,$FF,$93,$E3
	.dc.b	$5A,$7E,$CE,$3F,$DD,$F1,$8E,$EF,$F6,$E7,$BC,$F8,$0E,$35,$FE,$F4
	.dc.b	$8C,$5F,$18,$46,$2B,$8D,$70,$9C,$FC,$0D,$A3,$49,$7C,$E1,$F0,$5B
	.dc.b	$24,$65,$E9,$C6,$F8,$3A,$49,$F2,$46,$5E,$9F,$09,$E1,$23,$83,$7C
	.dc.b	$BD,$08,$E0,$F7,$78,$2F,$2F,$FC,$83,$B7,$FB,$88,$9A,$FF,$B9,$DF
	.dc.b	$D7,$AF,$85,$F7,$BF,$1D,$B1,$A6,$C7,$ED,$9A,$FF,$A1,$7E,$09,$E0
	.dc.b	$7F,$27,$8C,$35,$08,$DB,$F5,$5B,$5E,$4E,$AF,$CC,$B7,$A7,$95,$27
	.dc.b	$2F,$E9,$23,$31,$FA,$58,$FC,$03,$67,$A7,$56,$8E,$7F,$0D,$F9,$BD
	.dc.b	$7F,$A7,$DA,$31,$E2,$FC,$66,$C7,$F1,$66,$BF,$EA,$56,$BE,$E6,$BC
	.dc.b	$FA,$7F,$C9,$73,$CD,$B1,$CD,$AF,$36,$BC,$DA,$F3,$AF,$3D,$76,$24
	.dc.b	$E9,$B1,$A8,$C1,$1C,$12,$BC,$8D,$AF,$23,$5F,$23,$5F,$23,$5F,$24
	.dc.b	$72,$23,$C9,$AF,$26,$BC,$9A,$F2,$38,$10,$E0,$53,$81,$4E,$06,$88
	.dc.b	$68,$A6,$8C,$10,$C1,$2B,$E4,$C7,$91,$B1,$DD,$AF,$46,$BC,$DA,$F3
	.dc.b	$AF,$A1,$3C,$DB,$1C,$DA,$F3,$6B,$CD,$AF,$3A,$FA,$15,$D8,$93,$A6
	.dc.b	$C6,$A3,$04,$70,$4A,$F2,$36,$BC,$8D,$7C,$8D,$7C,$8D,$7C,$91,$C8
	.dc.b	$8F,$27,$EE,$FA,$FF,$EB,$E8,$FE,$22,$D7,$FD,$93,$EB,$FF,$14,$7F
	.dc.b	$D8,$07,$F1,$6F,$C2,$BF,$84,$DE,$5B,$19,$3F,$CF,$E7,$F0,$E7,$2F
	.dc.b	$F8,$31,$AF,$F1,$20,$E1,$E0,$72,$32,$57,$4E,$23,$C4,$E4,$8E,$08
	.dc.b	$F9,$5A,$F8,$CD,$99,$21,$97,$6F,$9E,$3C,$24,$78,$9B,$38,$9B,$32
	.dc.b	$58,$E6,$FB,$9A,$FA,$CE,$26,$4A,$64,$EC,$43,$A1,$A8,$F2,$47,$46
	.dc.b	$C6,$4F,$19,$D4,$E4,$53,$99,$83,$C0,$60,$86,$C8,$68,$AF,$81,$E4
	.dc.b	$D7,$89,$B2,$39,$78,$9E,$63,$25,$74,$E2,$3D,$CE,$48,$E0,$8F,$C1
	.dc.b	$6B,$E4,$36,$64,$86,$5D,$BE,$88,$F0,$91,$E2,$6C,$E2,$6C,$C9,$63
	.dc.b	$9B,$EE,$7F,$5D,$F5,$0F,$A9,$FE,$EB,$7F,$E0,$48,$C5,$23,$13,$A8
	.dc.b	$FF,$3B,$CE,$A7,$85,$66,$7F,$5D,$28,$C4,$7F,$80,$A3,$0F,$53,$9C
	.dc.b	$69,$FE,$FD,$F8,$2D,$DC,$D2,$47,$13,$FC,$1E,$53,$FC,$1F,$C3,$FE
	.dc.b	$0E,$C3,$FD,$44,$48,$FF,$47,$C3,$D8,$F0,$36,$39,$1F,$E1,$78,$C9
	.dc.b	$E5,$E6,$FF,$17,$79,$7A,$6D,$FF,$8B,$1B,$5F,$E9,$C6,$BF,$C4,$7E
	.dc.b	$35,$F1,$BF,$16,$BF,$C6,$71,$8D,$3C,$2D,$8C,$CE,$EB,$FD,$4F,$DE
	.dc.b	$EC,$C4,$7F,$1D,$8C,$61,$EA,$6D,$FF,$9A,$2B,$75,$FE,$67,$F2,$6B
	.dc.b	$FE,$66,$6D,$1A,$6B,$E3,$9F,$1D,$49,$8F,$C8,$8C,$BE,$1F,$0F,$99
	.dc.b	$1C,$1B,$DF,$A1,$1C,$1F,$4B,$C5,$FF,$41,$E9,$FE,$58,$6E,$7F,$E8
	.dc.b	$99,$D7,$8D,$EA,$F8,$F7,$C8,$52,$4F,$DC,$8C,$BF,$7B,$9F,$F3,$FF
	.dc.b	$21,$F9,$EC,$0D,$23,$F9,$63,$25,$7F,$42,$E0,$FD,$0C,$FF,$98,$23
	.dc.b	$FC,$B9,$CF,$67,$FD,$76,$D7,$FD,$D2,$D7,$FC,$CF,$D1,$FA,$07,$A1
	.dc.b	$FE,$6A,$E6,$FC,$8D,$7E,$3F,$C4,$74,$6A,$95,$A7,$F2,$53,$AA,$FC
	.dc.b	$5E,$3B,$E8,$FC,$84,$60,$95,$FE,$F4,$F3,$1F,$77,$FE,$7A,$F9,$1F
	.dc.b	$F6,$3B,$67,$4D,$7F,$DA,$4D,$7D,$7A,$D7,$F7,$A2,$7E,$2B,$4F,$8A
	.dc.b	$53,$FB,$C7,$AF,$FE,$02,$8F,$F4,$CB,$63,$FB,$E0,$70,$F0,$3F,$89
	.dc.b	$47,$FB,$23,$A6,$C7,$EA,$9A,$FE,$A5,$AF,$E9,$73,$F1,$3F,$C6,$FC
	.dc.b	$E7,$B1,$A9,$46,$1D,$98,$8D,$46,$6F,$FD,$63,$EB,$BE,$3B,$F9,$37
	.dc.b	$C8,$E1,$B8,$FB,$27,$58,$AB,$4D,$9F,$83,$F6,$CF,$82,$E3,$1F,$E1
	.dc.b	$F8,$C4,$EE,$31,$3A,$7F,$D6,$B3,$AA,$EB,$F9,$54,$9F,$16,$3F,$CE
	.dc.b	$E3,$11,$CB,$04,$7F,$B3,$2F,$BB,$70,$E1,$5E,$66,$BF,$E3,$3E,$5F
	.dc.b	$E5,$6F,$E6,$F9,$BF,$C8,$51,$88,$FC,$CC,$7F,$B6,$30,$F9,$FE,$C7
	.dc.b	$F4,$DF,$91,$FC,$0C,$60,$38,$DD,$3B,$BF,$F2,$15,$6E,$B5,$F0,$3F
	.dc.b	$E4,$A8,$FB,$6D,$B3,$4C,$7C,$47,$F5,$DF,$8A,$EC,$DC,$3A,$D2,$31
	.dc.b	$1F,$E5,$A8,$C5,$AB,$E0,$4C,$3D,$E5,$38,$7C,$67,$27,$F2,$4D,$9F
	.dc.b	$C8,$B6,$8E,$E7,$F8,$1F,$67,$F0,$67,$A1,$FC,$2D,$C3,$E0,$3D,$3F
	.dc.b	$17,$F1,$DF,$43,$5F,$D1,$36,$FF,$42,$E7,$E1,$1F,$18,$F7,$BF,$E0
	.dc.b	$0C,$FE,$DA,$73,$18,$BF,$F3,$A4,$61,$4E,$EB,$94,$60,$DA,$F2,$51
	.dc.b	$F8,$03,$2E,$37,$E4,$D8,$C9,$B3,$F2,$59,$7C,$26,$F0,$FE,$4A,$31
	.dc.b	$1F,$1E,$3F,$87,$AC,$7D,$FE,$F7,$58,$F1,$C6,$1E,$C6,$F1,$4E,$E3
	.dc.b	$0D,$D1,$FB,$8B,$EA,$BD,$BF,$BB,$FF,$8A,$2F,$FC,$59,$39,$51,$9A
	.dc.b	$E3,$5C,$1F,$CF,$D6,$E7,$95,$7E,$13,$2F,$71,$1F,$C2,$F9,$AF,$C0
	.dc.b	$3F,$D6,$3B,$F8,$63,$13,$FB,$1C,$7F,$D5,$91,$87,$F6,$26,$2B,$8B
	.dc.b	$FA,$4F,$4B,$FC,$89,$F9,$27,$9F,$97,$FC,$2E,$D8,$CC,$EC,$47,$A3
	.dc.b	$AB,$F8,$26,$3F,$58,$FC,$D9,$EB,$3F,$80,$78,$DF,$E5,$1F,$C3,$BB
	.dc.b	$36,$FF,$6A,$D9,$FD,$A3,$5F,$E7,$DF,$8D,$FE,$F6,$7F,$76,$D8,$FD
	.dc.b	$D3,$5D,$C7,$EE,$CF,$CA,$F8,$DF,$91,$6C,$7C,$43,$F0,$BC,$9F,$E2
	.dc.b	$D7,$07,$E1,$67,$83,$ED,$3F,$68,$52,$1F,$8B,$CF,$6D,$39,$F2,$FE
	.dc.b	$CA,$32,$D8,$EE,$7C,$73,$A9,$F9,$D3,$E5,$E7,$F3,$9C,$1A,$38,$1F
	.dc.b	$EE,$BF,$6B,$FD,$F9,$F0,$CF,$73,$1F,$13,$FE,$7D,$8C,$57,$ED,$AB
	.dc.b	$75,$AF,$5B,$F7,$35,$AA,$FF,$80,$2D,$AD,$C7,$1B,$E9,$7C,$61,$5C
	.dc.b	$70,$FF,$3A,$CF,$F4,$B2,$4F,$F4,$A6,$BF,$BE,$6B,$FB,$D7,$F8,$86
	.dc.b	$5F,$C3,$3E,$37,$F6,$19,$FC,$96,$EB,$FE,$71,$B4,$62,$B8,$6D,$DE
	.dc.b	$61,$DC,$A7,$DF,$38,$BF,$86,$9E,$BB,$72,$9F,$C3,$B5,$FC,$7F,$9B
	.dc.b	$FA,$9E,$0F,$F5,$73,$6F,$FC,$44,$7E,$60,$B3,$F9,$E5,$FB,$EF,$E9
	.dc.b	$1A,$7E,$90,$B1,$B8,$FD,$87,$77,$F9,$3B,$5F,$8F,$3F,$C2,$5F,$93
	.dc.b	$5F,$E2,$2C,$FF,$D8,$1F,$9A,$FE,$EE,$0D,$23,$FC,$16,$BF,$D8,$DA
	.dc.b	$7F,$0D,$E8,$FE,$77,$3F,$12,$7F,$9E,$F4,$FF,$0D,$AD,$47,$09,$F9
	.dc.b	$6F,$8C,$BC,$3F,$82,$7F,$90,$F7,$7F,$46,$FD,$3B,$F9,$37,$7F,$F2
	.dc.b	$96,$C6,$6F,$74,$9F,$BF,$E0,$7E,$0E,$3F,$A8,$B5,$FD,$DB,$5F,$E7
	.dc.b	$67,$E1,$0F,$1E,$FF,$25,$BD,$FF,$9F,$7F,$A2,$78,$3C,$79,$FF,$B4
	.dc.b	$B7,$9F,$FA,$F4,$F3,$3F,$84,$C6,$7C,$3F,$F1,$CC,$7F,$64,$6B,$FC
	.dc.b	$39,$AF,$F5,$16,$E7,$FD,$E0,$E9,$1F,$DF,$1A,$67,$7F,$F1,$57,$87
	.dc.b	$F5,$A7,$D3,$CB,$F8,$2E,$1F,$9D,$1F,$E1,$6B,$FF,$39,$3F,$E2,$2F
	.dc.b	$13,$F2,$AD,$8C,$BB,$E8,$E5,$BF,$C5,$8F,$D3,$10,$FF,$38,$78,$DC
	.dc.b	$9F,$F1,$53,$5C,$D7,$46,$D7,$F8,$F2,$3E,$71,$FE,$3C,$2F,$87,$FE
	.dc.b	$3C,$6F,$8E,$73,$E4,$7B,$EF,$FC,$70,$D7,$FD,$04,$D8,$FF,$28,$6D
	.dc.b	$EA,$E3,$FB,$95,$F9,$33,$D4,$3C,$BF,$F3,$9F,$EC,$FF,$E6,$5B,$34
	.dc.b	$93,$FE,$70,$8C,$B6,$3F,$D0,$24,$7F,$CD,$90,$F8,$EE,$84,$8C,$BF
	.dc.b	$E8,$82,$3D,$5A,$FF,$95,$7D,$2F,$85,$F1,$DF,$E6,$96,$C7,$F9,$71
	.dc.b	$AE,$E9,$3F,$EA,$D6,$C7,$FA,$B5,$AF,$FA,$B5,$AF,$FA,$B6,$3F,$D7
	.dc.b	$6F,$43,$FC,$77,$B7,$F0,$28,$CC,$F0,$9F,$C6,$B7,$D8,$FF,$73,$4F
	.dc.b	$FB,$28,$93,$FE,$CA,$7F,$DA,$EF,$1F,$5B,$FD,$B6,$DF,$2B,$FC,$93
	.dc.b	$3A,$3F,$DC,$BC,$32,$7F,$D9,$BE,$C3,$E0,$7F,$D8,$51,$86,$BF,$E4
	.dc.b	$89,$FF,$4D,$11,$FF,$4C,$BF,$ED,$E2,$3F,$ED,$E6,$BF,$E5,$A1,$ED
	.dc.b	$7E,$4D,$FF,$66,$9F,$ED,$E7,$11,$FE,$52,$6C,$67,$3F,$F3,$1C,$E7
	.dc.b	$E0,$EC,$F1,$6B,$DE,$91,$88,$FF,$35,$C6,$1F,$09,$E5,$F7,$1E,$EF
	.dc.b	$E9,$FE,$23,$FC,$C7,$18,$C5,$18,$9D,$57,$FB,$DA,$B5,$3F,$E6,$AF
	.dc.b	$80,$7C,$06,$AB,$E1,$46,$30,$E7,$1F,$EF,$E8,$C5,$71,$AE,$13,$FE
	.dc.b	$82,$F8,$17,$95,$88,$46,$30,$E7,$1F,$3E,$31,$7F,$E7,$C6,$C7,$D3
	.dc.b	$1F,$07,$FE,$8F,$AE,$31,$FE,$90,$8C,$57,$FA,$11,$B1,$C5,$AF,$7F
	.dc.b	$83,$F2,$59,$A4,$BE,$B3,$E0,$F4,$BD,$56,$69,$27,$70,$F4,$36,$48
	.dc.b	$CB,$F0,$F8,$FF,$35,$1D,$9B,$97,$FA,$71,$B8,$7E,$76,$7C,$7E,$23
	.dc.b	$C3,$77,$E1,$B8,$7F,$BE,$49,$7F,$27,$E1,$3E,$9B,$EC,$6C,$7F,$AA
	.dc.b	$9B,$1B,$DF,$FD,$77,$39,$F8,$AB,$3C,$DA,$FE,$9E,$91,$88,$FF,$61
	.dc.b	$C6,$1F,$97,$D8,$F8,$4D,$42,$30,$EA,$33,$18,$BF,$D7,$69,$C3,$9F
	.dc.b	$F6,$91,$8B,$FA,$F1,$87,$F5,$C6,$3A,$38,$11,$FB,$BF,$0D,$FE,$1E
	.dc.b	$F2,$7F,$D9,$6D,$8F,$C8,$35,$FE,$08,$D7,$FD,$BE,$D4,$D2,$3F,$D6
	.dc.b	$E3,$AB,$4E,$A5,$3A,$94,$EA,$53,$A9,$63,$AE,$1F,$D7,$9B,$3D,$A3
	.dc.b	$13,$B2,$38,$24,$6C,$91,$E4,$8C,$B6,$32,$F9,$A3,$2D,$8C,$BE,$88
	.dc.b	$CB,$63,$2F,$AA,$32,$D8,$CC,$FA,$A3,$C0,$43,$B1,$4E,$85,$39,$15
	.dc.b	$E4,$F0,$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$3F,$42
	.dc.b	$3D,$AD,$3C,$C5,$3C,$25,$3B,$14,$EA,$58,$EB,$3F,$29,$B3,$DA,$31
	.dc.b	$3B,$23,$82,$46,$C9,$1E,$48,$CB,$63,$2F,$9A,$32,$D8,$CB,$E8,$8F
	.dc.b	$BB,$EC,$BD,$97,$B5,$FC,$EE,$BF,$D9,$4D,$53,$B8,$C3,$63,$F6,$BC
	.dc.b	$5F,$DC,$DF,$DA,$B5,$FD,$AB,$5F,$DA,$B5,$FD,$AB,$71,$FE,$53,$87
	.dc.b	$85,$B3,$DA,$31,$3B,$23,$82,$46,$C8,$79,$1C,$B4,$C9,$E6,$72,$D3
	.dc.b	$27,$A1,$CB,$4C,$9E,$A7,$2D,$33,$3E,$A7,$C0,$43,$B1,$4E,$85,$39
	.dc.b	$15,$E4,$F0,$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$4F
	.dc.b	$D0,$8F,$6B,$5F,$33,$5F,$0B,$5E,$CD,$7A,$B6,$3A,$CF,$CA,$6C,$F6
	.dc.b	$8C,$4E,$C8,$E0,$91,$B2,$1E,$47,$2D,$32,$79,$9C,$B4,$C9,$E8,$72
	.dc.b	$D3,$27,$A9,$CB,$4C,$CF,$A9,$F0,$1F,$77,$EC,$9F,$02,$8F,$D9,$9F
	.dc.b	$E2,$D8,$CB,$83,$FC,$C4,$E3,$1F,$DC,$36,$38,$7C,$47,$39,$CB,$63
	.dc.b	$27,$33,$C1,$28,$D4,$7F,$98,$70,$AF,$F9,$81,$AF,$F9,$C9,$BB,$7F
	.dc.b	$9C,$EF,$F8,$57,$03,$E2,$F6,$DB,$F0,$1F,$E0,$D8,$C9,$D3,$8F,$FC
	.dc.b	$29,$5F,$E1,$D9,$DC,$61,$FF,$0F,$1C,$BE,$33,$FE,$00,$CF,$F1,$96
	.dc.b	$C7,$F9,$55,$AF,$FA,$25,$AF,$CF,$E8,$FF,$1C,$6F,$E4,$6D,$7F,$23
	.dc.b	$6C,$FF,$97,$DA,$FF,$A4,$FE,$39,$F1,$DF,$EE,$5A,$FE,$55,$18,$D7
	.dc.b	$DE,$D8,$CC,$EE,$BF,$CC,$5D,$95,$FF,$4F,$76,$FF,$95,$F0,$F8,$ED
	.dc.b	$BF,$F3,$55,$6E,$BF,$CD,$DD,$D5,$FF,$32,$5F,$08,$C5,$FF,$99,$63
	.dc.b	$9F,$B3,$82,$38,$3B,$FA,$FC,$08,$E0,$F1,$FA,$FD,$08,$E0,$FB,$BE
	.dc.b	$2F,$FA,$57,$3F,$F4,$94,$63,$C5,$F3,$4E,$73,$FC,$F7,$47,$E8,$7E
	.dc.b	$33,$F3,$6D,$8C,$88,$CB,$4F,$CD,$BF,$9E,$70,$7E,$9B,$A3,$93,$4C
	.dc.b	$9F,$9B,$29,$F9,$B2,$F4,$7F,$B3,$67,$FC,$B4,$7E,$37,$4E,$73,$FB
	.dc.b	$7E,$54,$62,$11,$87,$AB,$67,$F1,$9E,$1F,$FA,$19,$BC,$DE,$2E,$6F
	.dc.b	$C9,$C6,$0F,$F5,$47,$0B,$7F,$3F,$12,$31,$3A,$9C,$C6,$32,$FF,$70
	.dc.b	$0F,$8C,$FF,$3A,$11,$CB,$FE,$BB,$97,$94,$FB,$AF,$04,$E6,$31,$3C
	.dc.b	$EB,$FD,$E8,$F0,$AD,$C6,$27,$75,$F9,$8A,$E1,$3F,$9E,$E7,$DB,$63
	.dc.b	$2F,$FB,$26,$32,$FF,$25,$FA,$4F,$D6,$57,$F8,$5D,$B1,$FE,$D3,$6B
	.dc.b	$FC,$AC,$38,$7F,$97,$7C,$77,$F8,$85,$7C,$77,$F8,$86,$30,$DF,$0F
	.dc.b	$F9,$A3,$67,$67,$FB,$87,$F6,$0F,$F7,$44,$7F,$8A,$5B,$19,$F2,$BE
	.dc.b	$C4,$E5,$B1,$FE,$E9,$BF,$17,$C7,$75,$24,$67,$0F,$E8,$91,$C7,$43
	.dc.b	$49,$3D,$E3,$2F,$F1,$7E,$4F,$F6,$9B,$F8,$E2,$19,$E4,$FF,$20,$7C
	.dc.b	$0B,$E0,$5E,$AF,$F5,$3F,$8C,$FF,$7F,$4E,$63,$17,$F3,$63,$0A,$77
	.dc.b	$5C,$A3,$06,$D7,$C4,$FF,$4C,$AF,$03,$63,$33,$FD,$56,$72,$FF,$57
	.dc.b	$CB,$F2,$BF,$3D,$FA,$3F,$8F,$FC,$A3,$67,$F2,$0D,$7F,$B6,$4F,$01
	.dc.b	$3F,$0A,$3F,$0D,$D7,$FD,$09,$C1,$FD,$17,$5F,$92,$FE,$6F,$D1,$E8
	.dc.b	$FE,$CF,$A3,$FD,$D1,$87,$48,$D3,$5F,$0C,$6B,$0E,$B0,$C3,$C0,$FD
	.dc.b	$33,$C3,$97,$D3,$FC,$A3,$E7,$C7,$F8,$16,$30,$FC,$C3,$1A,$27,$70
	.dc.b	$9D,$ED,$FB,$DF,$69,$E9,$7F,$0B,$FC,$E3,$0D,$BF,$DD,$CE,$B5,$56
	.dc.b	$9B,$3F,$6F,$B7,$BE,$3F,$89,$6E,$7F,$C2,$67,$CC,$78,$DC,$65,$F0
	.dc.b	$B4,$F8,$33,$F4,$B8,$3E,$7B,$79,$3F,$2B,$6F,$3F,$A9,$27,$AE,$5F
	.dc.b	$99,$8F,$83,$9B,$F7,$C7,$C3,$7F,$8B,$7F,$07,$FF,$4B,$4E,$63,$0F
	.dc.b	$F1,$A9,$DE,$0A,$DC,$61,$FF,$4E,$69,$E8,$9F,$C3,$67,$FE,$A7,$8C
	.dc.b	$47,$DD,$D3,$D6,$BB,$3F,$96,$71,$7E,$B7,$2F,$E2,$D1,$89,$FD,$54
	.dc.b	$63,$83,$EB,$C6,$0F,$F2,$34,$EF,$83,$D7,$C4,$F9,$7F,$E8,$DC,$58
	.dc.b	$6E,$3F,$71,$B8,$7E,$92,$B8,$4F,$F0,$13,$F2,$51,$0F,$A4,$7E,$91
	.dc.b	$47,$F8,$77,$F7,$AF,$0F,$A3,$FD,$55,$B2,$69,$E6,$FB,$BF,$DC,$47
	.dc.b	$E3,$9B,$7E,$36,$BF,$D7,$ED,$18,$E5,$FE,$2D,$EA,$FC,$9A,$FD,$0B
	.dc.b	$5F,$C7,$36,$37,$1F,$24,$FA,$E7,$E4,$78,$3E,$AB,$F5,$F8,$BF,$1F
	.dc.b	$F3,$3C,$76,$34,$93,$CC,$E0,$73,$9F,$C1,$57,$4E,$3F,$C1,$91,$C1
	.dc.b	$E0,$F6,$7F,$DD,$5F,$28,$F6,$FF,$AC,$46,$1B,$E2,$7F,$A1,$27,$58
	.dc.b	$FF,$BD,$83,$87,$FD,$EF,$1B,$9E,$3E,$AF,$FA,$36,$B5,$7D,$6F,$94
	.dc.b	$71,$C7,$FC,$C7,$7F,$E1,$42,$4F,$F4,$76,$BF,$B5,$6B,$FA,$27,$FA
	.dc.b	$0E,$7F,$E9,$BF,$8A,$FF,$0C,$4F,$F1,$68,$C7,$17,$F4,$36,$C6,$44
	.dc.b	$FE,$BB,$46,$BE,$D9,$EC,$78,$C8,$7E,$08,$E3,$7F,$D3,$F8,$3A,$C7
	.dc.b	$E1,$0C,$9B,$BF,$CA,$4B,$E1,$A9,$1F,$CB,$19,$3F,$92,$F2,$FF,$8C
	.dc.b	$FF,$55,$FC,$F9,$23,$21,$F5,$63,$E9,$C3,$F1,$DA,$7C,$86,$9F,$B5
	.dc.b	$A7,$EB,$1F,$B9,$69,$97,$07,$CC,$8F,$E1,$9D,$7F,$C1,$B3,$C7,$5F
	.dc.b	$F0,$B6,$C6,$6F,$F8,$67,$67,$C4,$BF,$09,$F6,$CF,$96,$7C,$E3,$FB
	.dc.b	$4F,$77,$F2,$4F,$C1,$7F,$91,$F8,$1F,$ED,$76,$C9,$A7,$97,$AB,$F9
	.dc.b	$A4,$7D,$C3,$FC,$05,$E0,$FE,$15,$A6,$4F,$EB,$6E,$5F,$21,$0F,$21
	.dc.b	$23,$FA,$83,$C4,$FE,$9D,$97,$16,$99,$8D,$E7,$FD,$6F,$F5,$89,$3F
	.dc.b	$D8,$33,$FF,$37,$46,$23,$FB,$6F,$89,$CE,$FF,$B8,$DF,$F8,$7A,$31
	.dc.b	$DB,$FD,$C6,$BC,$47,$CC,$3C,$E5,$31,$B7,$F9,$BB,$C7,$FF,$3E,$73
	.dc.b	$7F,$9B,$5B,$3F,$E0,$A8,$C4,$FE,$56,$3E,$A9,$CF,$AB,$F6,$EF,$D8
	.dc.b	$8F,$F0,$F4,$FF,$89,$1E,$11,$FE,$12,$F2,$3A,$B6,$32,$FF,$75,$F2
	.dc.b	$4B,$E0,$BF,$C7,$71,$86,$C6,$4F,$F1,$E4,$65,$FF,$1F,$79,$7F,$10
	.dc.b	$8F,$D1,$CF,$FC,$89,$AF,$E8,$73,$F4,$CF,$DD,$3E,$37,$9F,$FD,$9F
	.dc.b	$27,$C6,$FF,$98,$49,$3F,$E5,$F7,$FC,$E8,$47,$FC,$D0,$70,$2B,$FE
	.dc.b	$69,$F8,$D9,$1F,$E6,$76,$BF,$E8,$7F,$43,$FD,$09,$1F,$E7,$F2,$47
	.dc.b	$47,$FD,$21,$DF,$FD,$56,$3F,$D2,$0D,$7F,$D2,$1E,$97,$FC,$01,$F2
	.dc.b	$77,$8F,$F6,$89,$FF,$7F,$B5,$FF,$50,$47,$FA,$94,$8F,$FA,$A7,$A3
	.dc.b	$A3,$FE,$4F,$6C,$FF,$AB,$CE,$51,$96,$BF,$ED,$B9,$FF,$51,$3F,$ED
	.dc.b	$F6,$B9,$3F,$D8,$AF,$FB,$27,$D0,$91,$C3,$D7,$FF,$67,$57,$FB,$12
	.dc.b	$31,$EB,$7F,$98,$5F,$F4,$55,$7F,$B6,$88,$E0,$F0,$FB,$1F,$EE,$38
	.dc.b	$FF,$75,$47,$33,$FD,$6D,$19,$77,$EC,$FE,$4A,$7C,$FE,$D7,$F4,$A6
	.dc.b	$E1,$9A,$FF,$6E,$C6,$1E,$87,$E6,$BE,$0D,$EE,$6E,$3F,$D8,$91,$87
	.dc.b	$61,$C4,$FC,$66,$C7,$CA,$5E,$EF,$38,$EC,$D7,$93,$63,$71,$DB,$DE
	.dc.b	$7B,$FF,$EF,$BE,$CF,$A7,$18,$CB,$F5,$95,$FE,$D9,$8C,$4E,$AB,$84
	.dc.b	$FD,$5F,$80,$AF,$FB,$7F,$81,$18,$8F,$DD,$C6,$30,$EB,$18,$9D,$29
	.dc.b	$6D,$86,$D7,$09,$D4,$E4,$D7,$04,$9D,$57,$E2,$31,$FC,$7B,$6F,$F1
	.dc.b	$9F,$03,$D4,$95,$C1,$F0,$7C,$04,$8E,$6D,$BF,$CC,$FC,$13,$9F,$D2
	.dc.b	$DA,$F3,$36,$77,$3E,$EE,$C7,$A5,$FE,$0C,$9F,$D1,$C6,$3D,$0F,$73
	.dc.b	$63,$33,$BC,$FF,$2A,$7C,$3E,$7F,$8E,$DD,$FF,$91,$3F,$9C,$6B,$FC
	.dc.b	$B3,$E0,$DF,$0B,$88,$46,$1F,$9C,$DC,$3F,$6E,$D8,$FD,$DF,$98,$F5
	.dc.b	$39,$36,$3E,$B3,$6B,$F7,$ED,$8F,$E0,$7D,$7F,$E3,$49,$FD,$41,$23
	.dc.b	$3C,$1E,$5E,$8F,$D3,$4F,$E8,$FA,$F0,$4A,$FC,$24,$FE,$7A,$7F,$51
	.dc.b	$D5,$C4,$93,$F8,$9F,$85,$FF,$1D,$E1,$FE,$23,$6F,$A9,$FC,$51,$B3
	.dc.b	$FC,$61,$AF,$F5,$40,$E1,$FE,$AE,$D7,$3C,$6A,$EC,$DA,$EC,$D9,$EC
	.dc.b	$D7,$B5,$73,$24,$66,$3B,$B5,$EF,$5D,$89,$5D,$9B,$1D,$9A,$F6,$9E
	.dc.b	$55,$DA,$73,$3D,$9A,$F6,$AE,$84,$7A,$35,$E9,$1C,$27,$A4,$E6,$7A
	.dc.b	$35,$E8,$72,$21,$C9,$D4,$72,$8C,$9C,$8A,$72,$38,$18,$78,$38,$21
	.dc.b	$C4,$A7,$13,$04,$30,$47,$9B,$96,$B9,$AF,$96,$4A,$F8,$4D,$AF,$5B
	.dc.b	$67,$CC,$D7,$C7,$5D,$09,$19,$8F,$03,$5F,$05,$76,$25,$76,$6C,$76
	.dc.b	$6B,$DA,$79,$57,$69,$CC,$F6,$6B,$DA,$BA,$11,$E8,$D7,$A4,$70,$9F
	.dc.b	$D2,$7B,$0F,$69,$ED,$7F,$80,$AF,$FD,$8E,$D9,$D3,$67,$73,$FB,$5F
	.dc.b	$90,$FF,$48,$3F,$A6,$6B,$FB,$26,$BF,$B6,$6D,$9A,$4C,$3C,$51,$96
	.dc.b	$C6,$67,$C1,$18,$23,$82,$1F,$C1,$E3,$83,$63,$6D,$36,$7A,$5C,$B4
	.dc.b	$C9,$EB,$72,$D3,$27,$B5,$CB,$4C,$CF,$AD,$B3,$E4,$7B,$1E,$32,$18
	.dc.b	$21,$DC,$C9,$4C,$9D,$CC,$94,$CC,$78,$5C,$B5,$CC,$78,$DE,$84,$39
	.dc.b	$16,$38,$36,$38,$3A,$24,$69,$B1,$A8,$C1,$1C,$12,$7E,$4C,$65,$B1
	.dc.b	$99,$F9,$31,$82,$38,$21,$EE,$8E,$0D,$8D,$B4,$D9,$F0,$9C,$B4,$C9
	.dc.b	$EF,$72,$D3,$27,$B9,$CB,$4C,$CF,$AD,$FF,$37,$FD,$6B,$3F,$F5,$6E
	.dc.b	$38,$6C,$FF,$9D,$A7,$4F,$EC,$0F,$EC,$0E,$27,$9C,$70,$32,$53,$FD
	.dc.b	$73,$1F,$D9,$E3,$71,$CC,$FF,$32,$7D,$AE,$DF,$F3,$03,$77,$34,$D7
	.dc.b	$0F,$F6,$B9,$E4,$1C,$57,$F9,$C4,$91,$FE,$16,$F4,$7B,$71,$9A,$48
	.dc.b	$E7,$CB,$FE,$21,$AF,$F0,$FF,$37,$96,$31,$3F,$E9,$A6,$BF,$E0,$35
	.dc.b	$EC,$EC,$91,$D1,$F7,$75,$FF,$8F,$6F,$FC,$77,$B7,$AA,$31,$5F,$EA
	.dc.b	$66,$BF,$E0,$FF,$89,$7C,$77,$F8,$2E,$FF,$D5,$31,$8E,$2F,$86,$D8
	.dc.b	$CC,$EE,$BF,$D6,$32,$F8,$EC,$47,$56,$BF,$EF,$08,$D5,$F4,$9C,$CF
	.dc.b	$F9,$D7,$C2,$BE,$19,$A4,$AF,$AB,$F1,$AF,$8D,$E2,$D8,$CC,$60,$E7
	.dc.b	$E0,$48,$FB,$B7,$FE,$80,$D3,$FC,$F3,$3C,$B7,$95,$D9,$B3,$FE,$95
	.dc.b	$8F,$C2,$F9,$3F,$E9,$B8,$C1,$33,$FC,$F0,$1F,$1B,$F9,$62,$3A,$32
	.dc.b	$7F,$AC,$79,$F8,$47,$FA,$D5,$B1,$93,$67,$FA,$F9,$CB,$F8,$7E,$4A
	.dc.b	$FF,$9C,$DA,$FF,$9C,$9B,$C9,$FE,$73,$AF,$F3,$8E,$58,$25,$76,$3F
	.dc.b	$CF,$8F,$2D,$3E,$A6,$79,$9F,$B0,$D8,$CD,$7F,$A9,$37,$FF,$2B,$3F
	.dc.b	$EF,$4D,$FE,$B3,$77,$CE,$1F,$EF,$A9,$CB,$63,$2F,$FB,$AB,$C9,$F9
	.dc.b	$9F,$1B,$FE,$FE,$F1,$7F,$D8,$72,$8C,$4E,$A3,$FD,$EA,$D7,$F4,$1C
	.dc.b	$BF,$E6,$99,$FC,$DE,$9C,$67,$F3,$73,$96,$C6,$5F,$B8,$BC,$D7,$CC
	.dc.b	$6E,$73,$18,$9E,$11,$FE,$77,$8C,$5F,$F9,$FE,$31,$5C,$5F,$F3,$E7
	.dc.b	$E1,$1F,$E9,$FF,$8F,$D6,$08,$C4,$FF,$A1,$E3,$1E,$4F,$E3,$EB,$75
	.dc.b	$AA,$F3,$FE,$14,$F8,$9D,$47,$18,$C2,$AE,$F1,$86,$CF,$09,$F6,$F9
	.dc.b	$1F,$E9,$1F,$D0,$7C,$BA,$A3,$13,$E5,$8C,$66,$AD,$D6,$AB,$E3,$E9
	.dc.b	$5F,$87,$5B,$79,$46,$63,$E8,$3B,$87,$7F,$FB,$75,$FE,$42,$43,$F9
	.dc.b	$07,$9B,$FD,$41,$FF,$57,$7C,$0B,$DC,$7B,$FF,$D3,$63,$0D,$8C,$CF
	.dc.b	$F4,$D9,$CB,$63,$FA,$03,$F6,$C4,$75,$F1,$7F,$CA,$4D,$8D,$35,$F9
	.dc.b	$3F,$89,$FF,$93,$B3,$FC,$63,$76,$34,$FF,$28,$E7,$EF,$AF,$6F,$8D
	.dc.b	$FE,$5F,$8C,$47,$FB,$2D,$AF,$E4,$7E,$D3,$A1,$30,$F4,$1F,$9B,$7A
	.dc.b	$EB,$F1,$E7,$93,$5F,$E3,$95,$D1,$FA,$31,$9C,$9F,$63,$FD,$A2,$7E
	.dc.b	$7B,$5F,$E4,$91,$AC,$3E,$84,$E6,$7E,$4A,$F8,$6F,$E5,$5F,$77,$FA
	.dc.b	$7B,$EB,$18,$AE,$AD,$7A,$D7,$15,$5F,$96,$F9,$9F,$98,$4C,$3E,$0E
	.dc.b	$C8,$F8,$3A,$FC,$BC,$FE,$06,$1E,$E8,$C1,$C3,$F0,$EF,$AC,$4C,$30
	.dc.b	$7F,$A4,$7F,$55,$7D,$77,$F3,$C8,$FE,$01,$18,$BF,$E8,$B1,$89,$D4
	.dc.b	$7F,$49,$13,$BF,$0F,$FD,$2B,$C1,$F4,$49,$97,$99,$FB,$FD,$4F,$95
	.dc.b	$79,$BF,$E3,$4C,$FA,$4E,$63,$19,$7F,$09,$AD,$BF,$EA,$AA,$DC,$63
	.dc.b	$CC,$FF,$52,$E7,$E9,$F2,$6E,$BF,$D7,$DB,$87,$F8,$63,$0F,$63,$4F
	.dc.b	$C5,$EF,$F2,$9F,$DB,$18,$3F,$85,$F8,$69,$FC,$11,$FC,$2D,$C7,$2F
	.dc.b	$F8,$DE,$7F,$96,$46,$27,$47,$08,$FE,$6F,$D5,$F7,$63,$F9,$CC,$63
	.dc.b	$8F,$FC,$81,$1F,$CA,$E7,$53,$D6,$3F,$83,$4E,$A3,$FA,$0F,$83,$FD
	.dc.b	$41,$D3,$67,$F9,$A3,$63,$7E,$27,$F3,$77,$DC,$7E,$E0,$F8,$27,$EB
	.dc.b	$38,$FF,$5D,$5E,$03,$CF,$1F,$0F,$0F,$2C,$79,$F0,$F2,$4E,$5A,$FE
	.dc.b	$5C,$EA,$53,$C0,$43,$C6,$43,$C5,$9F,$FB,$97,$7F,$F0,$D7,$C9,$FF
	.dc.b	$3D,$9F,$F0,$E7,$B1,$FE,$EA,$51,$97,$FD,$DE,$63,$C1,$FC,$D4,$EE
	.dc.b	$74,$D8,$CF,$B1,$FE,$F2,$9C,$B6,$33,$7E,$27,$F1,$27,$F9,$78,$97
	.dc.b	$FC,$89,$AE,$8C,$10,$FC,$06,$5F,$D5,$1B,$1F,$7F,$3F,$C3,$37,$3F
	.dc.b	$C9,$3F,$B0,$D3,$FD,$3C,$F6,$AF,$8E,$DB,$FE,$ED,$AF,$F9,$06,$3F
	.dc.b	$0A,$D8,$FC,$2B,$67,$F0,$A7,$8C,$FB,$51,$4F,$C1,$4F,$8D,$C6,$1E
	.dc.b	$67,$F3,$FF,$43,$F9,$7E,$BF,$F3,$C3,$78,$7F,$54,$D9,$3E,$BE,$0F
	.dc.b	$B1,$8F,$A8,$F8,$C5,$3F,$89,$72,$FF,$17,$F3,$3F,$7A,$4B,$FD,$EB
	.dc.b	$FB,$3E,$9F,$E1,$94,$F6,$F1,$D3,$FC,$31,$BD,$5F,$E6,$56,$CE,$F0
	.dc.b	$F0,$BF,$CB,$F7,$F9,$CD,$3F,$85,$FC,$17,$F0,$E2,$4E,$0F,$F0,$87
	.dc.b	$77,$D8,$9B,$1E,$E3,$FC,$71,$19,$8F,$F1,$BB,$5F,$80,$43,$C0,$7E
	.dc.b	$2C,$FF,$1E,$F2,$6E,$7F,$C8,$4D,$E1,$FC,$4C,$FF,$88,$63,$A7,$83
	.dc.b	$FD,$62,$1F,$29,$FD,$6F,$C2,$FE,$AC,$DE,$DF,$F4,$43,$67,$67,$F6
	.dc.b	$AC,$7E,$43,$4F,$9C,$5D,$7A,$46,$A3,$D0,$7A,$FC,$07,$4F,$E6,$A3
	.dc.b	$EF,$CF,$F8,$04,$F8,$DA,$3C,$3F,$61,$23,$F0,$F3,$F1,$0E,$95,$CD
	.dc.b	$EF,$1F,$E1,$A3,$9F,$89,$F1,$48,$FE,$43,$5F,$F1,$27,$0F,$F7,$DE
	.dc.b	$AF,$F1,$5B,$7C,$5F,$D5,$61,$FE,$38,$9C,$B7,$C6,$F7,$3F,$E3,$08
	.dc.b	$FF,$1B,$ED,$CE,$73,$87,$F8,$E2,$3F,$17,$B7,$C2,$EA,$F5,$B7,$2F
	.dc.b	$DD,$6F,$FE,$79,$FD,$83,$FC,$BB,$E4,$7F,$53,$6E,$C6,$9F,$E5,$C7
	.dc.b	$1A,$FF,$98,$6F,$83,$63,$2F,$F9,$A9,$AF,$F5,$FF,$2F,$FC,$E7,$F0
	.dc.b	$FF,$E7,$08,$34,$8F,$F4,$EF,$37,$FD,$9D,$1F,$E8,$06,$C6,$4E,$25
	.dc.b	$8F,$4E,$52,$3F,$CD,$B7,$FE,$9B,$6C,$7F,$A9,$08,$FF,$A8,$9A,$FF
	.dc.b	$AC,$CE,$05,$7F,$D6,$2E,$0F,$F7,$17,$9F,$FE,$B6,$8F,$F6,$04,$BA
	.dc.b	$E9,$2B,$FD,$81,$19,$7F,$D4,$DE,$97,$FB,$42,$7F,$D9,$93,$99,$FF
	.dc.b	$13,$4E,$5D,$F8,$1F,$E0,$9A,$FF,$6E,$C6,$CF,$F6,$D9,$47,$A6,$F2
	.dc.b	$FF,$DF,$13,$C0,$FF,$3E,$43,$D4,$CD,$7F,$AE,$63,$0F,$36,$CF,$09
	.dc.b	$DC,$FF,$BD,$CF,$07,$B6,$F5,$B3,$01,$19,$77,$1A,$5E,$E7,$F9,$16
	.dc.b	$FF,$D2,$0D,$8F,$F6,$63,$5F,$F2,$53,$5E,$DE,$C3,$E1,$7F,$C5,$DD
	.dc.b	$3F,$E9,$98,$C6,$FF,$C3,$2B,$FC,$A5,$18,$9D,$57,$09,$FF,$7C,$6B
	.dc.b	$5F,$F2,$C7,$B3,$FE,$41,$AF,$F8,$01,$39,$9F,$74,$E5,$FB,$BE,$3E
	.dc.b	$57,$0F,$F9,$63,$2F,$14,$7C,$58,$C6,$1F,$EA,$F6,$C7,$E2,$23,$F1
	.dc.b	$A4,$70,$7E,$12,$79,$BD,$B8,$7C,$14,$69,$E3,$F6,$DE,$DD,$25,$F2
	.dc.b	$8C,$B7,$CE,$E0,$47,$07,$83,$B7,$F8,$86,$5F,$EC,$76,$E5,$FE,$BA
	.dc.b	$8C,$78,$8E,$FF,$3F,$3F,$E5,$6A,$F0,$7F,$20,$7F,$32,$D8,$D5,$FE
	.dc.b	$46,$78,$F3,$E4,$8F,$C2,$F7,$3F,$DB,$33,$FE,$D9,$6C,$7F,$A4,$43
	.dc.b	$87,$FD,$28,$D7,$F3,$1E,$EF,$F9,$DD,$F9,$AD,$9F,$F7,$23,$63,$A3
	.dc.b	$5F,$D2,$47,$FB,$A7,$A7,$F4,$0D,$E9,$F7,$46,$27,$F3,$A7,$C1,$E8
	.dc.b	$FF,$53,$57,$E8,$A3,$06,$74,$8F,$E0,$27,$E6,$36,$3E,$57,$BC,$F7
	.dc.b	$FF,$C0,$52,$8C,$70,$FF,$60,$6F,$83,$FE,$FF,$6C,$FC,$76,$BF,$EB
	.dc.b	$96,$BF,$BD,$6B,$D6,$BB,$C6,$08,$E0,$95,$E1,$6D,$78,$5B,$5E,$16
	.dc.b	$BE,$16,$BE,$16,$BE,$19,$E8,$49,$E8,$D8,$E8,$D7,$A3,$5E,$8D,$7A
	.dc.b	$47,$22,$3C,$9A,$F2,$6B,$C9,$AF,$23,$81,$0E,$05,$38,$14,$E0,$68
	.dc.b	$86,$8A,$68,$C1,$0C,$12,$BE,$9C,$FB,$5B,$3E,$66,$CF,$85,$AF,$66
	.dc.b	$BD,$5A,$F5,$AF,$A7,$18,$23,$82,$57,$85,$B5,$E1,$6D,$78,$5A,$F8
	.dc.b	$5A,$F8,$5A,$F8,$67,$A1,$27,$A3,$63,$A3,$5E,$8D,$7A,$35,$F9,$3E
	.dc.b	$BF,$FB,$15,$FE,$14,$D7,$FD,$76,$7F,$0D,$3F,$D8,$E7,$EA,$7E,$4B
	.dc.b	$FD,$F8,$FE,$A5,$AF,$EA,$5A,$FE,$D5,$B8,$1A,$4A,$ED,$3B,$6C,$ED
	.dc.b	$B3,$B9,$FD,$E3,$5F,$24,$68,$8E,$9A,$E8,$C1,$0C,$10,$F6,$3C,$1A
	.dc.b	$6C,$A6,$CF,$80,$ED,$A6,$CA,$6C,$F7,$19,$29,$99,$F8,$0E,$5A,$E6
	.dc.b	$7E,$03,$E2,$21,$D8,$A7,$42,$C7,$26,$C7,$27,$81,$0E,$05,$8E,$0D
	.dc.b	$8E,$0E,$89,$1A,$6C,$6A,$30,$47,$04,$9F,$91,$3C,$1B,$3B,$6C,$EE
	.dc.b	$7E,$23,$5F,$89,$1A,$23,$A6,$BA,$30,$43,$04,$3E,$2B,$C1,$A6,$CA
	.dc.b	$6C,$F8,$8E,$DA,$6C,$A7,$F7,$7F,$82,$7B,$3F,$E4,$BF,$63,$F5,$D1
	.dc.b	$89,$46,$27,$51,$C5,$AF,$F9,$4C,$7C,$EB,$C6,$CC,$7F,$95,$E3,$0F
	.dc.b	$43,$FC,$10,$E2,$BF,$7C,$D8,$FE,$F2,$6C,$FF,$2A,$9D,$0F,$F2,$E9
	.dc.b	$77,$F1,$4F,$F8,$17,$0D,$70,$9A,$4A,$E8,$7F,$85,$32,$E9,$C7,$FE
	.dc.b	$15,$BF,$F0,$9C,$60,$F0,$72,$79,$6B,$FC,$3B,$9E,$08,$F4,$3F,$D0
	.dc.b	$FC,$BE,$7E,$6F,$F4,$0E,$1F,$E2,$F2,$4F,$F8,$A7,$C5,$78,$1E,$D6
	.dc.b	$C7,$36,$CF,$F2,$36,$C7,$F8,$C7,$AF,$FD,$AD,$3F,$E4,$4E,$7F,$14
	.dc.b	$62,$75,$D7,$FE,$47,$AF,$E5,$9D,$AE,$D6,$FE,$98,$C3,$63,$33,$C1
	.dc.b	$B7,$FE,$B0,$F8,$6A,$FF,$AB,$9A,$FF,$9C,$DA,$FF,$9E,$1B,$E0,$FF
	.dc.b	$9A,$EF,$E4,$36,$FF,$D7,$24,$AF,$F5,$EF,$C3,$F0,$25,$70,$79,$ED
	.dc.b	$23,$9B,$6B,$FD,$05,$ED,$F8,$09,$3E,$07,$CB,$D1,$F9,$6F,$7F,$F2
	.dc.b	$7C,$9F,$EA,$F6,$E4,$69,$FE,$4B,$70,$7E,$61,$C1,$1F,$F3,$37,$2F
	.dc.b	$56,$CF,$56,$9C,$8F,$C6,$94,$FF,$60,$72,$7F,$B2,$AB,$FC,$F0,$48
	.dc.b	$FF,$67,$72,$F3,$7E,$96,$5C,$5F,$A0,$DD,$9B,$FE,$02,$FF,$D4,$CD
	.dc.b	$AF,$F5,$14,$7F,$BA,$88,$E0,$FF,$40,$7C,$87,$F9,$9B,$E1,$3F,$BB
	.dc.b	$36,$33,$CD,$FE,$D7,$9C,$B6,$3F,$26,$A3,$F3,$04,$8C,$BF,$E0,$1F
	.dc.b	$8B,$7E,$21,$F9,$68,$C3,$67,$FD,$25,$3A,$CF,$A4,$62,$74,$FC,$4F
	.dc.b	$1B,$DF,$3F,$EA,$78,$C1,$F9,$B7,$0F,$FB,$BF,$1F,$F6,$02,$F7,$BD
	.dc.b	$FE,$4F,$FB,$CD,$B1,$93,$E2,$C6,$5F,$E4,$D4,$F8,$0F,$F0,$94,$3E
	.dc.b	$17,$FD,$55,$39,$F1,$3F,$DF,$D1,$8A,$DB,$63,$35,$ED,$AE,$4D,$9F
	.dc.b	$8B,$18,$8F,$E6,$DA,$7C,$59,$FE,$0D,$10,$FA,$65,$07,$F0,$4F,$85
	.dc.b	$7B,$9F,$E2,$EB,$46,$1F,$F1,$B1,$88,$D4,$66,$7D,$B9,$35,$E4,$DB
	.dc.b	$E2,$D9,$DC,$EB,$92,$56,$9B,$3E,$73,$CF,$E0,$3C,$0F,$F5,$CC,$7B
	.dc.b	$08,$7B,$29,$EC,$FF,$20,$7D,$2D,$8F,$A6,$D7,$EC,$B5,$FE,$C1,$5D
	.dc.b	$F9,$3F,$C9,$2D,$8F,$20,$70,$FF,$67,$E4,$7E,$6C,$F3,$BF,$CB,$31
	.dc.b	$86,$C6,$71,$FF,$31,$4E,$5F,$8F,$F1,$1F,$CE,$3E,$E7,$E6,$B7,$0F
	.dc.b	$C7,$EE,$8C,$47,$C1,$8C,$52,$BE,$AF,$47,$D4,$24,$7D,$47,$FC,$D7
	.dc.b	$9F,$AA,$FE,$0E,$DF,$39,$BA,$FC,$28,$C4,$7F,$4C,$DB,$EA,$61,$F3
	.dc.b	$A7,$F1,$51,$89,$F9,$F1,$FE,$7A,$E7,$7D,$5F,$F5,$49,$F3,$5F,$F8
	.dc.b	$3E,$72,$FE,$5B,$3D,$4F,$ED,$6B,$84,$62,$B7,$3C,$8F,$39,$F5,$B2
	.dc.b	$F6,$3F,$8A,$C3,$D3,$C4,$FC,$59,$F0,$3E,$3C,$FF,$C4,$13,$98,$C5
	.dc.b	$FE,$A2,$31,$3F,$8F,$8C,$57,$28,$C4,$79,$DF,$D7,$EB,$E7,$D3,$E0
	.dc.b	$36,$32,$7C,$18,$CB,$F9,$C9,$F4,$4F,$D1,$23,$FE,$35,$3F,$8C,$7A
	.dc.b	$4F,$B2,$FE,$FF,$3F,$16,$31,$48,$C4,$EA,$3E,$84,$EA,$7E,$7C,$3D
	.dc.b	$38,$FB,$0C,$3F,$12,$FF,$C1,$AE,$9E,$2D,$8F,$E4,$BF,$14,$F1,$7F
	.dc.b	$93,$D7,$ED,$E3,$0F,$56,$BE,$26,$CF,$C3,$8F,$D8,$6D,$F9,$26,$E1
	.dc.b	$F1,$1F,$E6,$78,$78,$5F,$E0,$8D,$7C,$EF,$A6,$BD,$67,$D8,$8E,$67
	.dc.b	$E2,$6B,$B6,$FF,$8A,$8C,$9E,$37,$06,$7D,$47,$C5,$7F,$98,$FD,$4F
	.dc.b	$95,$5F,$CC,$27,$3E,$5F,$F3,$29,$CC,$A7,$2F,$F9,$EB,$2F,$F7,$3C
	.dc.b	$EA,$72,$DB,$ED,$3A,$25,$7E,$21,$FF,$24,$11,$FE,$EA,$D7,$F7,$AD
	.dc.b	$78,$1F,$29,$FB,$B9,$53,$EF,$47,$CA,$69,$F2,$A3,$FD,$F3,$8F,$F2
	.dc.b	$97,$A4,$FD,$67,$B4,$B4,$FC,$0C,$7D,$FD,$7F,$AA,$B6,$BA,$CE,$AB
	.dc.b	$C5,$3F,$8E,$3C,$B3,$F7,$9A,$F9,$5F,$CF,$C3,$D9,$A4,$7F,$43,$C1
	.dc.b	$FE,$66,$8F,$4E,$DF,$A7,$AF,$4E,$5E,$BC,$3E,$49,$FB,$7C,$3F,$B5
	.dc.b	$DF,$EE,$E3,$B4,$65,$A7,$F8,$2F,$87,$F9,$04,$FF,$1C,$E3,$FE,$17
	.dc.b	$85,$E3,$FE,$14,$DE,$4F,$5C,$EB,$1F,$0E,$3F,$60,$A7,$D8,$3B,$3F
	.dc.b	$DF,$BE,$C3,$C8,$47,$C8,$FE,$FB,$A3,$F9,$CE,$1F,$CD,$A7,$F2,$75
	.dc.b	$FC,$F7,$5F,$DE,$BF,$D9,$9A,$F8,$48,$78,$49,$0E,$AF,$F1,$C5,$7F
	.dc.b	$52,$8E,$A6,$7A,$B9,$D7,$FA,$5A,$B6,$FF,$5E,$EB,$FF,$4C,$E5,$FD
	.dc.b	$8A,$CD,$24,$FF,$71,$ED,$FE,$D9,$8F,$F2,$5E,$DF,$ED,$8D,$8C,$CE
	.dc.b	$FB,$BC,$2F,$E5,$4F,$96,$53,$DB,$87,$F8,$30,$F3,$94,$F5,$EB,$FE
	.dc.b	$0F,$BF,$F0,$6C,$74,$E3,$FC,$45,$7C,$63,$A1,$F1,$B4,$77,$F4,$6D
	.dc.b	$7C,$66,$DF,$F8,$BA,$75,$C1,$2B,$4D,$9F,$F2,$7F,$81,$FE,$39,$6F
	.dc.b	$83,$E8,$8C,$63,$FE,$58,$6B,$E5,$CB,$CE,$FD,$B3,$CE,$3C,$3F,$F3
	.dc.b	$97,$C0,$7F,$94,$89,$1F,$E5,$77,$FC,$B4,$9A,$47,$FC,$CF,$E3,$7F
	.dc.b	$99,$A3,$FC,$DB,$EF,$C9,$FF,$36,$87,$0F,$FA,$83,$6F,$F3,$E9,$1F
	.dc.b	$F3,$E3,$FE,$64,$F7,$E4,$7F,$9A,$5A,$FF,$86,$9F,$F7,$51,$1F,$F4
	.dc.b	$BB,$FE,$9A,$4D,$23,$FE,$A0,$32,$7F,$71,$F3,$3F,$DB,$DE,$FF,$F7
	.dc.b	$38,$46,$1D,$57,$FB,$CA,$75,$39,$9F,$F5,$DC,$7F,$A6,$08,$FF,$A5
	.dc.b	$DE,$9E,$77,$E1,$A1,$18,$E2,$FF,$3D,$B6,$32,$7F,$B2,$0A,$7F,$B4
	.dc.b	$1F,$37,$82,$F3,$F8,$CF,$FB,$5A,$73,$18,$8F,$F4,$38,$F6,$AC,$FF
	.dc.b	$B9,$49,$1F,$E1,$38,$CB,$FE,$15,$F4,$7B,$3E,$23,$FD,$22,$E2,$7F
	.dc.b	$DE,$24,$8F,$99,$E9,$39,$DE,$E7,$FB,$F9,$B3,$A6,$BE,$6D,$55,$C6
	.dc.b	$BE,$7F,$A9,$6B,$EC,$35,$E0,$D7,$FD,$4B,$48,$C4,$7D,$CF,$56,$C9
	.dc.b	$A4,$C3,$C7,$3C,$B8,$36,$48,$CC,$7E,$1A,$39,$FA,$B2,$73,$7F,$EF
	.dc.b	$A6,$C7,$8B,$D7,$7A,$F4,$93,$D2,$32,$F6,$E9,$FF,$4D,$69,$F9,$56
	.dc.b	$E5,$FE,$F6,$8C,$78,$BF,$E4,$57,$F3,$5E,$C7,$C1,$9F,$3B,$63,$3B
	.dc.b	$7F,$B2,$9B,$3E,$CE,$6F,$CB,$47,$E4,$C9,$1F,$93,$7F,$1A,$FC,$2E
	.dc.b	$5F,$C2,$C6,$09,$1F,$E5,$DF,$67,$FD,$0D,$1F,$AD,$6C,$67,$3F,$13
	.dc.b	$4F,$9B,$ED,$BD,$BC,$C7,$E6,$A3,$0E,$A3,$2F,$FB,$92,$33,$7F,$BB
	.dc.b	$E7,$FD,$0B,$67,$67,$E8,$CB,$1A,$7F,$41,$D1,$AE,$8F,$6C,$60,$FC
	.dc.b	$0B,$87,$F5,$B8,$3D,$1E,$AD,$8F,$C3,$B7,$0F,$E1,$CD,$8F,$F4,$17
	.dc.b	$B8,$F6,$BF,$AD,$B7,$4F,$EF,$72,$8C,$78,$9F,$C5,$DB,$1C,$AB,$89
	.dc.b	$2B,$8C,$EA,$7F,$81,$B5,$1D,$67,$85,$75,$6C,$75,$AE,$71,$88,$E6
	.dc.b	$D7,$9D,$75,$7A,$4F,$02,$57,$56,$C7,$58,$C1,$1C,$10,$EF,$1C,$1B
	.dc.b	$1B,$69,$B3,$C7,$5E,$57,$C7,$1B,$73,$3C,$DB,$19,$76,$73,$29,$D8
	.dc.b	$87,$03,$55,$F0,$49,$5E,$16,$D7,$79,$E6,$71,$39,$98,$24,$70,$AF
	.dc.b	$23,$63,$BD,$75,$25,$75,$9D,$4F,$46,$BD,$C8,$76,$9E,$35,$D9,$B1
	.dc.b	$DA,$BA,$46,$23,$A3,$5E,$75,$D9,$E9,$3C,$09,$5D,$5B,$1D,$63,$04
	.dc.b	$70,$43,$BC,$70,$6C,$6D,$A7,$FB,$B7,$D4,$FF,$59,$B7,$D4,$7A,$BF
	.dc.b	$DF,$E3,$FD,$71,$18,$7F,$C0,$53,$C1,$FD,$3F,$D3,$FF,$5B,$7F,$60
	.dc.b	$D7,$F7,$2D,$77,$87,$5C,$53,$99,$FD,$2B,$5F,$DE,$36,$37,$3C,$63
	.dc.b	$04,$70,$49,$F1,$B5,$F1,$B5,$F1,$9A,$21,$A2,$9A,$30,$43,$04,$3D
	.dc.b	$2F,$56,$9C,$8A,$71,$29,$C4,$F7,$BC,$5A,$71,$29,$C4,$A7,$19,$F8
	.dc.b	$AE,$DA,$ED,$AE,$E7,$E4,$3B,$6B,$B6,$BB,$9F,$24,$F2,$8F,$49,$0F
	.dc.b	$19,$5D,$38,$23,$82,$4F,$68,$E0,$D7,$6D,$8D,$CF,$8A,$30,$47,$04
	.dc.b	$9F,$63,$5F,$63,$5F,$61,$A2,$1A,$29,$A3,$04,$30,$43,$D2,$F5,$69
	.dc.b	$C8,$A7,$12,$9F,$D9,$BE,$3B,$FA,$7E,$3F,$D8,$9B,$1F,$E5,$63,$FC
	.dc.b	$00,$E1,$E0,$7E,$D4,$FF,$2D,$F0,$BD,$8D,$47,$F5,$A8,$C3,$B3,$11
	.dc.b	$A7,$FC,$C4,$DE,$1F,$F2,$ED,$FF,$A0,$5B,$3F,$DF,$E3,$1C,$5C,$79
	.dc.b	$1F,$11,$49,$5D,$C4,$69,$C7,$2F,$F8,$62,$FF,$D1,$CD,$8F,$80,$D7
	.dc.b	$FC,$3D,$C7,$E7,$E7,$7A,$54,$93,$DE,$32,$F8,$7A,$3D,$2F,$FA,$01
	.dc.b	$AF,$F7,$E6,$A6,$F5,$7F,$98,$E3,$D4,$DC,$7F,$C9,$7E,$17,$C9,$8C
	.dc.b	$4F,$D7,$7F,$C9,$D3,$FE,$4A,$24,$67,$AF,$E8,$76,$BB,$7E,$C5,$7F
	.dc.b	$96,$38,$BE,$44,$E5,$B3,$AB,$FC,$27,$C8,$BE,$47,$FD,$19,$87,$FA
	.dc.b	$DE,$31,$CF,$F4,$5B,$19,$9D,$D7,$FA,$FF,$C2,$AF,$FA,$3D,$AF,$FA
	.dc.b	$21,$AF,$FB,$09,$B4,$FA,$0A,$4C,$3C,$0F,$0F,$13,$FD,$19,$1F,$83
	.dc.b	$6E,$3F,$EC,$8D,$F4,$74,$FC,$1E,$57,$A5,$B3,$FE,$A7,$26,$5F,$EA
	.dc.b	$76,$BF,$E5,$0C,$FC,$FC,$9F,$02,$3F,$54,$D7,$57,$F8,$7E,$4F,$F1
	.dc.b	$FF,$C1,$FF,$B5,$1B,$3F,$B1,$E5,$FC,$5C,$E6,$0F,$FA,$00,$8F,$ED
	.dc.b	$F9,$7F,$D8,$B1,$88,$FF,$67,$EF,$F5,$23,$E9,$61,$CA,$7E,$DE,$5F
	.dc.b	$EC,$59,$C9,$FE,$EC,$E0,$C4,$FD,$7A,$E0,$FD,$76,$BF,$5D,$BF,$12
	.dc.b	$F8,$DF,$F3,$87,$B1,$FC,$2E,$31,$E3,$FF,$97,$A3,$13,$AB,$FF,$72
	.dc.b	$D6,$94,$FF,$A9,$08,$FE,$6C,$C9,$5F,$E3,$0B,$D2,$7A,$7D,$E3,$F8
	.dc.b	$E4,$61,$E2,$7F,$BD,$1C,$4F,$26,$C7,$C6,$C5,$EA,$FC,$A2,$46,$5F
	.dc.b	$91,$1A,$8F,$91,$18,$23,$EE,$C7,$F6,$4B,$E1,$3E,$1F,$4F,$FA,$09
	.dc.b	$B1,$93,$84,$65,$F5,$E5,$5F,$53,$5F,$F4,$5E,$3E,$FA,$F6,$4E,$65
	.dc.b	$D7,$FB,$F7,$F8,$11,$0F,$E8,$4E,$5F,$F4,$A8,$C9,$FA,$FF,$FB,$0A
	.dc.b	$3F,$A4,$46,$39,$BF,$C7,$8D,$8F,$AA,$D7,$F0,$3D,$1F,$D2,$37,$FC
	.dc.b	$43,$75,$34,$C7,$5F,$F9,$1E,$7F,$1C,$D9,$FE,$09,$5F,$82,$9C,$BF
	.dc.b	$E4,$DC,$BD,$4F,$A9,$AF,$FA,$C1,$AF,$C4,$DB,$FA,$CB,$F0,$31,$F7
	.dc.b	$FB,$3E,$CD,$0D,$24,$F7,$AE,$2F,$76,$BD,$DA,$F7,$ED,$FC,$CF,$13
	.dc.b	$DE,$7C,$AF,$F9,$96,$30,$D8,$CD,$70,$6B,$F4,$BD,$27,$9B,$F5,$FF
	.dc.b	$50,$C3,$67,$D9,$3A,$D5,$5A,$6C,$EB,$ED,$7F,$CD,$10,$A4,$61,$B1
	.dc.b	$F7,$9B,$3F,$8B,$3E,$16,$7F,$14,$8F,$E2,$3D,$B7,$D5,$7E,$21,$B1
	.dc.b	$F8,$86,$BB,$8F,$C6,$FC,$77,$E0,$3E,$A7,$FD,$05,$8A,$30,$FD,$96
	.dc.b	$D7,$FB,$BA,$35,$CD,$FE,$F0,$AF,$F4,$47,$27,$E0,$89,$C5,$F5,$9F
	.dc.b	$C3,$F6,$3E,$95,$F3,$5F,$85,$CB,$FD,$F3,$39,$8C,$47,$F0,$89,$DB
	.dc.b	$E1,$36,$E1,$F4,$E7,$F9,$AF,$3A,$E6,$9F,$F0,$FE,$9E,$B8,$C4,$EA
	.dc.b	$72,$7C,$E3,$E8,$1F,$8F,$DF,$E7,$BF,$E3,$DF,$99,$7D,$E7,$EE,$F4
	.dc.b	$FF,$54,$46,$31,$FE,$3E,$6E,$74,$D3,$35,$F2,$FE,$13,$F6,$F1,$ED
	.dc.b	$8F,$E0,$D1,$87,$EA,$B6,$3F,$58,$D8,$FA,$99,$FF,$83,$5E,$53,$F6
	.dc.b	$FC,$6F,$96,$D8,$CC,$ED,$FB,$3C,$1F,$5C,$99,$FC,$08,$F3,$77,$FF
	.dc.b	$B2,$6F,$C2,$D8,$CC,$78,$31,$F2,$47,$C9,$3E,$59,$FB,$72,$F0,$FE
	.dc.b	$1A,$FB,$46,$CF,$F6,$F7,$97,$F6,$7C,$57,$B3,$FC,$AB,$D9,$FD,$AC
	.dc.b	$E6,$31,$D9,$FE,$E1,$AD,$DF,$E0,$6F,$6A,$32,$FF,$BA,$2F,$93,$BA
	.dc.b	$E5,$7D,$27,$33,$B3,$94,$7E,$D0,$FF,$21,$11,$FE,$83,$AF,$F9,$F6
	.dc.b	$31,$1B,$33,$8F,$89,$FC,$74,$7F,$42,$CF,$F0,$6F,$E0,$35,$FE,$90
	.dc.b	$DD,$7E,$19,$F4,$8B,$5D,$48,$74,$3B,$C7,$F4,$ED,$BF,$0C,$FE,$A3
	.dc.b	$7F,$D3,$36,$79,$9F,$D5,$E7,$75,$DE,$7C,$44,$70,$7E,$77,$AB,$F4
	.dc.b	$7C,$3F,$D7,$DB,$1F,$C8,$4D,$1E,$4B,$87,$A0,$FE,$79,$5E,$7E,$5F
	.dc.b	$DD,$75,$7E,$DC,$95,$FB,$BA,$47,$F6,$D7,$F8,$2F,$47,$B2,$7F,$B8
	.dc.b	$F3,$7F,$0D,$9F,$6B,$70,$E1,$1F,$C3,$EB,$83,$E3,$E5,$FF,$3E,$4F
	.dc.b	$DA,$8F,$98,$7E,$C3,$1F,$E1,$DF,$3B,$FC,$2C,$98,$F9,$5F,$E5,$B9
	.dc.b	$7F,$34,$BF,$E6,$57,$F8,$1A,$F4,$F5,$FD,$28,$FE,$D1,$5B,$7E,$29
	.dc.b	$E0,$21,$F1,$8F,$E9,$D3,$CF,$4F,$E9,$77,$FD,$6F,$B3,$F1,$35,$A8
	.dc.b	$FE,$AD,$FA,$74,$8F,$EB,$7C,$FF,$E1,$58,$C4,$7F,$93,$FB,$FF,$B3
	.dc.b	$D7,$FB,$77,$C1,$FE,$D7,$3F,$99,$6C,$66,$77,$3E,$83,$CC,$7C,$BE
	.dc.b	$8F,$D2,$3F,$E0,$5F,$0F,$FC,$06,$FC,$36,$D7,$F8,$91,$B1,$BF,$0F
	.dc.b	$E3,$3C,$FA,$BF,$54,$F6,$3E,$37,$89,$FE,$1A,$BF,$F3,$3F,$1F,$09
	.dc.b	$4E,$5F,$F2,$64,$66,$3F,$C5,$70,$F7,$7F,$C5,$ED,$3F,$CA,$5D,$7F
	.dc.b	$D3,$63,$27,$9F,$C7,$ED,$C7,$FE,$39,$C7,$17,$ED,$9F,$13,$E7,$27
	.dc.b	$3F,$E3,$9F,$91,$E5,$7F,$9B,$FE,$17,$FC,$A3,$06,$91,$FF,$2E,$99
	.dc.b	$2B,$FE,$62,$23,$FE,$93,$E6,$FF,$37,$91,$C1,$FE,$77,$7F,$D3,$E4
	.dc.b	$70,$7F,$9A,$07,$C2,$79,$08,$E5,$FF,$45,$FA,$CF,$5B,$FB,$94,$61
	.dc.b	$B3,$FE,$93,$8C,$3D,$0F,$F3,$F7,$89,$A7,$FB,$D7,$07,$FA,$76,$7F
	.dc.b	$D4,$33,$99,$FB,$7E,$67,$F3,$D7,$91,$FE,$AF,$F8,$3F,$F4,$69,$1F
	.dc.b	$F5,$A3,$FE,$BC,$23,$FE,$97,$6B,$FE,$78,$F4,$DE,$D7,$F8,$2E,$7F
	.dc.b	$D4,$F1,$89,$46,$27,$51,$FE,$CD,$6B,$FE,$82,$F5,$5E,$B6,$23,$FC
	.dc.b	$2D,$18,$79,$9F,$ED,$B8,$D2,$8D,$57,$FA,$E3,$89,$E4,$7F,$A3,$79
	.dc.b	$B0,$DB,$FF,$75,$B5,$FF,$75,$C7,$FB,$03,$D8,$7C,$1E,$67,$FC,$51
	.dc.b	$18,$E8,$FD,$04,$7F,$A6,$A3,$15,$C6,$B8,$4B,$D8,$FF,$0D,$CF,$FB
	.dc.b	$31,$B1,$FE,$A2,$35,$D5,$C2,$33,$1D,$7D,$9E,$78,$1A,$4C,$7C,$73
	.dc.b	$FE,$3D,$A7,$67,$32,$4E,$A7,$C9,$1D,$3D,$89,$39,$C1,$18,$CB,$FD
	.dc.b	$FE,$4B,$FF,$6D,$35,$FF,$39,$BF,$EE,$8E,$C7,$82,$EF,$F0,$46,$1B
	.dc.b	$19,$9F,$CF,$92,$BF,$DE,$8D,$7F,$D8,$FD,$7F,$05,$F5,$36,$3F,$5B
	.dc.b	$AF,$F1,$19,$F9,$93,$97,$F2,$07,$C1,$E6,$FC,$B5,$7E,$56,$31,$CD
	.dc.b	$C5,$FC,$A7,$35,$7F,$26,$D7,$F2,$0D,$E6,$FC,$B5,$7D,$B9,$7B,$4E
	.dc.b	$6E,$0D,$8F,$A4,$D9,$E0,$D8,$FC,$27,$B6,$F5,$3C,$0D,$8F,$B4,$DA
	.dc.b	$F0,$36,$3F,$19,$EE,$7B,$9B,$9F,$A6,$31,$1F,$EF,$26,$BF,$E8,$5F
	.dc.b	$77,$EC,$3F,$87,$6D,$FB,$5B,$1F,$99,$6B,$FE,$FC,$C9,$EF,$7C,$76
	.dc.b	$C7,$E4,$5B,$5F,$B7,$9F,$99,$F0,$0F,$80,$FE,$EB,$7F,$E9,$71,$19
	.dc.b	$E8,$FE,$BA,$DF,$1F,$E8,$36,$7F,$58,$D7,$93,$5F,$F5,$13,$5E,$B5
	.dc.b	$E0,$72,$D7,$35,$E0,$8C,$4F,$12,$38,$24,$6C,$95,$E6,$8C,$B6,$33
	.dc.b	$5E,$89,$CB,$63,$2F,$AA,$32,$D8,$CB,$E8,$6B,$E8,$9E,$C4,$7A,$35
	.dc.b	$E4,$D7,$93,$5E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82
	.dc.b	$18,$25,$7C,$E9,$F6,$B6,$7C,$CD,$9F,$0B,$5E,$CD,$7A,$B5,$EB,$5F
	.dc.b	$31,$CB,$5C,$D7,$82,$31,$3C,$48,$E0,$91,$B2,$57,$9A,$32,$D8,$CD
	.dc.b	$7A,$27,$2D,$8C,$BE,$A8,$CB,$63,$2F,$A1,$AF,$A2,$7B,$1F,$6F,$D8
	.dc.b	$FF,$60,$C7,$FB,$30,$FE,$30,$E1,$FF,$5F,$1F,$AC,$F1,$7F,$D0,$6F
	.dc.b	$EA,$9A,$FE,$A9,$AF,$ED,$9B,$87,$F8,$7E,$FF,$92,$4E,$5B,$19,$8E
	.dc.b	$D1,$89,$E2,$47,$04,$8D,$90,$FD,$FC,$6D,$C1,$E7,$8E,$0E,$9A,$E4
	.dc.b	$D9,$EB,$72,$D3,$27,$A4,$B3,$E9,$7A,$18,$9F,$13,$82,$1D,$CA,$75
	.dc.b	$29,$D4,$B1,$D5,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA,$70,$47
	.dc.b	$04,$8F,$5C,$75,$6B,$C9,$AF,$16,$C7,$19,$F9,$73,$96,$C6,$63,$E4
	.dc.b	$46,$27,$89,$1C,$12,$36,$43,$E3,$C6,$DC,$1E,$78,$E0,$E9,$AE,$4D
	.dc.b	$9E,$B7,$2D,$32,$7A,$4B,$3E,$97,$A1,$89,$FD,$1F,$E1,$EF,$85,$4B
	.dc.b	$DA,$FF,$03,$4E,$63,$07,$F9,$0D,$C4,$6A,$78,$BF,$E0,$DF,$9D,$FF
	.dc.b	$7E,$37,$1C,$DF,$F9,$2E,$72,$D8,$CB,$FD,$AE,$DE,$C6,$23,$CB,$18
	.dc.b	$7C,$66,$E3,$4A,$35,$7F,$E1,$FE,$2F,$F4,$03,$78,$F7,$18,$AF,$F0
	.dc.b	$89,$D0,$FF,$07,$1D,$38,$FF,$BF,$57,$F8,$36,$30,$78,$39,$BF,$C5
	.dc.b	$11,$FE,$26,$E4,$C1,$27,$A0,$E5,$F8,$2F,$F8,$25,$AF,$D2,$6B,$BD
	.dc.b	$3D,$BD,$3E,$78,$FF,$4F,$67,$82,$3D,$BA,$3E,$AB,$FE,$79,$6B,$FE
	.dc.b	$9B,$E0,$F8,$F1,$88,$FF,$4D,$EF,$82,$3C,$DF,$9B,$DD,$FE,$65,$D3
	.dc.b	$FC,$D5,$DB,$FE,$69,$6F,$77,$BA,$75,$7F,$E8,$9F,$85,$7C,$33,$6F
	.dc.b	$87,$18,$6C,$67,$0F,$F2,$B4,$E5,$FF,$50,$78,$39,$BF,$F4,$A7,$85
	.dc.b	$FE,$82,$C3,$E6,$B6,$B7,$18,$F3,$78,$57,$FA,$8D,$AF,$FA,$A1,$AA
	.dc.b	$91,$87,$FD,$39,$3D,$BC,$3C,$11,$F0,$F5,$6C,$8E,$0F,$C0,$F2,$FA
	.dc.b	$F0,$FF,$56,$35,$FF,$2A,$69,$E2,$8C,$47,$ED,$39,$BF,$CE,$9F,$0B
	.dc.b	$F8,$D8,$34,$8F,$F9,$D2,$FF,$77,$1F,$E7,$06,$C6,$4E,$25,$8F,$B2
	.dc.b	$47,$FD,$81,$BF,$FA,$8E,$31,$1F,$E5,$FD,$FE,$C4,$7D,$6C,$7E,$E3
	.dc.b	$6F,$AD,$6E,$72,$FF,$B2,$FE,$51,$F5,$DF,$C4,$3D,$6F,$F4,$A4,$62
	.dc.b	$51,$89,$D4,$71,$6B,$F9,$B3,$67,$E7,$F9,$F2,$FF,$BD,$9A,$FF,$BC
	.dc.b	$DC,$E0,$FA,$37,$DB,$61,$B5,$D0,$FE,$4B,$3B,$9D,$1D,$35,$FF,$64
	.dc.b	$74,$7E,$88,$95,$F2,$BA,$3F,$D9,$D1,$F1,$49,$1F,$A5,$D7,$F4,$91
	.dc.b	$88,$FF,$46,$57,$FB,$4F,$0F,$D6,$92,$3A,$74,$7F,$87,$B8,$3F,$55
	.dc.b	$3F,$44,$93,$D5,$FF,$11,$D7,$FB,$76,$3F,$6E,$43,$99,$FE,$FF,$F4
	.dc.b	$DE,$8B,$D0,$FE,$8B,$F8,$4E,$63,$14,$8C,$4E,$A3,$FD,$49,$5C,$A3
	.dc.b	$11,$F2,$BA,$9F,$29,$FE,$36,$CB,$F5,$ED,$8C,$C7,$FB,$5C,$C9,$FE
	.dc.b	$D9,$BF,$2F,$95,$FD,$63,$C6,$F2,$37,$0F,$0E,$7F,$EB,$C9,$CC,$1F
	.dc.b	$80,$FC,$EE,$1F,$F6,$5B,$F0,$88,$7F,$64,$BF,$87,$F7,$9D,$63,$16
	.dc.b	$8C,$42,$76,$F7,$3C,$71,$A5,$B7,$D3,$7B,$9F,$DC,$9C,$3E,$5E,$0F
	.dc.b	$81,$6F,$9D,$F9,$73,$DE,$30,$FD,$33,$E0,$C6,$B0,$FB,$6D,$97,$97
	.dc.b	$FE,$00,$FD,$EB,$EE,$36,$FE,$4B,$5F,$7B,$5F,$E7,$5B,$BF,$06,$FE
	.dc.b	$79,$F6,$5F,$EE,$39,$CC,$61,$FF,$73,$70,$FF,$2C,$8C,$47,$E0,$DA
	.dc.b	$FC,$1D,$FE,$9B,$F0,$4F,$AE,$67,$5F,$C2,$60,$F7,$78,$E1,$FC,$E1
	.dc.b	$B1,$98,$63,$E5,$74,$62,$FC,$BF,$25,$FB,$EF,$D2,$FF,$AB,$47,$F0
	.dc.b	$08,$C4,$EA,$3F,$AD,$4E,$9F,$EB,$93,$A7,$D0,$7D,$C3,$EF,$E5,$EC
	.dc.b	$7D,$47,$C7,$E9,$FD,$1E,$9F,$AE,$C3,$FC,$67,$3F,$E2,$2E,$6F,$9C
	.dc.b	$FE,$BE,$33,$5C,$23,$07,$F8,$E9,$C7,$47,$97,$8F,$F7,$7F,$33,$FE
	.dc.b	$3E,$FB,$2C,$72,$23,$18,$FE,$A1,$B8,$FE,$79,$AF,$A0,$F6,$73,$7E
	.dc.b	$59,$C1,$FA,$4F,$C4,$7F,$C9,$7F,$8B,$F6,$6D,$FE,$D1,$B3,$B1,$3E
	.dc.b	$7F,$8F,$7D,$BF,$EF,$6F,$F8,$5C,$62,$BF,$75,$18,$9D,$47,$ED,$E7
	.dc.b	$5D,$9F,$56,$9C,$6F,$8C,$FC,$BC,$7E,$01,$B7,$F7,$DA,$FD,$C6,$BF
	.dc.b	$6B,$A7,$FD,$8F,$A7,$C1,$8C,$1F,$96,$A7,$C1,$2B,$9C,$61,$B1,$98
	.dc.b	$FC,$A9,$FA,$67,$1D,$1F,$ED,$CB,$EC,$4E,$7F,$C5,$61,$E0,$3E,$9B
	.dc.b	$88,$D6,$1F,$ED,$CF,$65,$F1,$5F,$E7,$AF,$84,$FE,$03,$87,$FA,$02
	.dc.b	$73,$E9,$7F,$A0,$A7,$32,$9C,$A8,$CC,$61,$B5,$D7,$BF,$FD,$EB,$18
	.dc.b	$BE,$91,$83,$F8,$E1,$FE,$48,$25,$FF,$5C,$6B,$A3,$04,$3F,$B9,$E9
	.dc.b	$F2,$7C,$CF,$BC,$DD,$BE,$56,$7F,$E4,$F9,$CB,$63,$95,$7F,$0B,$BE
	.dc.b	$ED,$3F,$09,$0A,$FC,$57,$07,$E2,$63,$06,$A3,$3C,$1D,$6B,$F2,$5C
	.dc.b	$7F,$EA,$CF,$AE,$FC,$E3,$6F,$F3,$12,$9C,$B6,$3F,$4C,$BF,$5F,$FD
	.dc.b	$53,$63,$31,$83,$F0,$DC,$5F,$8F,$CF,$D7,$5F,$97,$7D,$BD,$7F,$C0
	.dc.b	$A1,$F6,$7F,$C0,$BA,$BF,$D8,$2D,$8C,$E3,$FC,$0D,$A7,$F0,$DE,$FF
	.dc.b	$E2,$7F,$5F,$FC,$87,$C0,$FE,$24,$D8,$CD,$7F,$16,$6D,$7D,$B2,$1F
	.dc.b	$3C,$B1,$FB,$67,$FD,$0D,$F8,$3F,$D0,$6C,$7F,$31,$11,$96,$9F,$D3
	.dc.b	$BC,$4F,$E8,$1C,$5F,$CF,$DB,$59,$F1,$7F,$08,$DA,$FB,$CF,$E1,$0F
	.dc.b	$90,$5E,$B7,$07,$CD,$8E,$DE,$27,$27,$FA,$D5,$72,$6E,$3E,$19,$D4
	.dc.b	$7F,$77,$E1,$FF,$77,$FD,$27,$F7,$08,$34,$8F,$F7,$A1,$B7,$F8,$C7
	.dc.b	$C8,$FF,$7B,$B6,$32,$6D,$FC,$0E,$7E,$8E,$9F,$DD,$CF,$CF,$27,$27
	.dc.b	$F8,$5B,$9B,$FC,$79,$E4,$FE,$5E,$7F,$C2,$E7,$E5,$4B,$1D,$37,$F8
	.dc.b	$51,$F6,$8F,$8E,$7F,$89,$3C,$AE,$4D,$39,$16,$7F,$C6,$67,$D5,$E4
	.dc.b	$FF,$62,$6D,$78,$7C,$E7,$F9,$01,$C1,$FE,$4E,$AF,$79,$FE,$32,$F2
	.dc.b	$FE,$1B,$FE,$40,$6B,$93,$67,$C1,$3C,$E5,$D7,$E3,$4F,$FA,$67,$F1
	.dc.b	$29,$3F,$E6,$96,$BF,$E3,$37,$FD,$BE,$47,$FC,$E6,$FF,$A8,$79,$64
	.dc.b	$7F,$9C,$5A,$FF,$8E,$DF,$F7,$39,$1F,$F3,$DB,$FE,$15,$F4,$3F,$D2
	.dc.b	$3F,$19,$7D,$15,$23,$0D,$3F,$C5,$B3,$C1,$FF,$51,$19,$3F,$D4,$D1
	.dc.b	$FE,$35,$7F,$D4,$46,$4F,$F5,$31,$4F,$E4,$1E,$9F,$82,$3F,$D6,$51
	.dc.b	$9E,$8F,$F5,$B4,$60,$FF,$60,$78,$52,$7F,$D2,$0D,$7F,$CB,$4F,$87
	.dc.b	$4F,$CA,$37,$8B,$FB,$33,$79,$7F,$B4,$D7,$FB,$5E,$3F,$DB,$4E,$0F
	.dc.b	$F1,$2F,$BF,$FE,$36,$8C,$9F,$EE,$52,$47,$F9,$4F,$D7,$7B,$1F,$2A
	.dc.b	$73,$3F,$EF,$0E,$37,$C2,$FC,$5A,$7A,$C9,$CD,$7F,$BE,$23,$15,$B6
	.dc.b	$C7,$F9,$87,$DC,$7B,$78,$8F,$F1,$DC,$62,$F9,$C6,$27,$4F,$DE,$9D
	.dc.b	$5F,$D7,$5E,$E6,$A5,$18,$BE,$71,$C6,$75,$39,$AE,$5C,$B8,$25,$F0
	.dc.b	$BE,$71,$C3,$DD,$93,$9B,$46,$2F,$FD,$FC,$4B,$FF,$67,$35,$FF,$4C
	.dc.b	$3F,$6C,$8F,$FB,$61,$FC,$97,$BC,$F1,$BF,$3B,$18,$6C,$FF,$A2,$E7
	.dc.b	$4F,$9D,$B3,$CB,$DC,$DC,$FF,$9A,$25,$18,$8F,$CE,$CE,$DF,$81,$F0
	.dc.b	$0F,$80,$D4,$23,$0E,$A3,$31,$89,$FC,$99,$27,$FD,$41,$C7,$FE,$88
	.dc.b	$8C,$47,$F9,$EB,$8F,$F2,$2F,$CF,$6C,$7E,$CD,$B1,$F9,$7E,$2F,$1F
	.dc.b	$C0,$FF,$A0,$AB,$F7,$0D,$9F,$F5,$73,$5F,$F4,$3B,$5F,$E0,$FF,$05
	.dc.b	$E7,$7F,$D6,$8D,$7F,$D1,$AD,$77,$97,$C2,$F8,$3F,$9B,$1F,$EB,$B6
	.dc.b	$BF,$E9,$46,$BF,$C5,$1A,$FF,$9D,$0F,$C6,$7B,$1F,$0D,$FF,$62,$35
	.dc.b	$FF,$4E,$B5,$E5,$AF,$84,$8F,$E5,$67,$ED,$BF,$97,$F8,$4F,$D5,$78
	.dc.b	$5F,$C6,$DB,$1F,$ED,$16,$BF,$EA,$A6,$BF,$CA,$9A,$9A,$4E,$7F,$EA
	.dc.b	$71,$CD,$AF,$36,$BC,$DA,$F3,$6C,$73,$E5,$FE,$E3,$39,$6C,$66,$BD
	.dc.b	$55,$C0,$93,$A6,$C6,$A3,$04,$70,$4A,$F4,$46,$5B,$19,$7D,$51,$96
	.dc.b	$C6,$5F,$64,$65,$B1,$97,$D4,$DA,$F5,$47,$62,$1D,$0A,$72,$29,$C8
	.dc.b	$AF,$27,$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$BE
	.dc.b	$7C,$79,$1A,$F7,$6B,$D1,$AF,$36,$C7,$3A,$F9,$53,$96,$C6,$6B,$E5
	.dc.b	$57,$02,$4E,$9B,$1A,$8C,$11,$C1,$2B,$D1,$19,$6C,$65,$F5,$46,$5B
	.dc.b	$1F,$77,$D8,$FF,$5F,$C7,$F8,$40,$FF,$63,$7B,$0F,$5F,$FD,$98,$D8
	.dc.b	$FF,$06,$4F,$07,$F6,$1D,$5F,$DA,$5F,$DB,$35,$FD,$B3,$5D,$E1,$FE
	.dc.b	$15,$7F,$69,$19,$7F,$70,$D7,$31,$BC,$BF,$B1,$47,$68,$C4,$EE,$73
	.dc.b	$3E,$09,$C9,$1E,$04,$34,$53,$CA,$53,$C6,$F1,$76,$43,$65,$36,$71
	.dc.b	$3D,$25,$8F,$2B,$C0,$87,$03,$91,$C0,$AF,$07,$47,$03,$45,$74,$E0
	.dc.b	$D9,$0C,$11,$D9,$27,$E5,$12,$7E,$04,$F5,$8C,$3E,$B6,$BE,$66,$BB
	.dc.b	$8E,$2F,$86,$7B,$46,$23,$C5,$3C,$1D,$35,$CC,$6E,$78,$C7,$86,$31
	.dc.b	$3B,$9C,$CF,$8E,$72,$47,$81,$0D,$14,$F4,$14,$F1,$BC,$5D,$90,$D9
	.dc.b	$4D,$9C,$4F,$49,$63,$CA,$F0,$21,$FE,$80,$FD,$87,$FC,$27,$8F,$F9
	.dc.b	$E5,$B1,$FB,$50,$E1,$FD,$C3,$5E,$5C,$DF,$E8,$A9,$E4,$D8,$E5,$3F
	.dc.b	$E2,$17,$F7,$87,$28,$D3,$FD,$E2,$EB,$FE,$94,$DF,$F7,$D7,$FE,$8D
	.dc.b	$6C,$7F,$80,$27,$6B,$89,$DD,$86,$C7,$80,$FF,$4F,$B8,$BF,$F1,$0C
	.dc.b	$F2,$E3,$F6,$51,$A4,$E3,$EF,$CE,$F8,$7A,$4A,$EC,$23,$4E,$3A,$3F
	.dc.b	$C6,$F8,$FF,$8D,$7A,$3D,$31,$89,$FF,$1B,$1C,$BA,$5F,$11,$EE,$6C
	.dc.b	$66,$30,$7C,$2E,$0F,$49,$33,$C1,$C7,$A6,$57,$56,$E5,$FE,$B9,$F8
	.dc.b	$9F,$F4,$6C,$FA,$9B,$19,$E9,$F9,$F1,$89,$FC,$37,$86,$F0,$FE,$8D
	.dc.b	$FF,$AB,$C9,$3A,$8C,$BF,$8D,$F8,$B7,$97,$FE,$A3,$AF,$F6,$54,$63
	.dc.b	$9B,$E5,$B6,$33,$3B,$AF,$F5,$27,$83,$5F,$F5,$1B,$5F,$F5,$03,$5F
	.dc.b	$F4,$EB,$69,$F1,$74,$98,$78,$1E,$1E,$1F,$FA,$8F,$1F,$F5,$47,$6E
	.dc.b	$09,$5F,$80,$FA,$C7,$D6,$77,$6C,$66,$30,$78,$FA,$3F,$0B,$F1,$9F
	.dc.b	$EE,$96,$EB,$FE,$80,$EA,$FF,$7A,$CE,$5B,$1F,$86,$8F,$C5,$26,$91
	.dc.b	$FD,$D7,$17,$DC,$8F,$B5,$1B,$33,$C7,$D1,$FE,$E9,$3B,$7E,$C7,$66
	.dc.b	$DF,$EF,$1C,$9F,$7E,$31,$01,$19,$BF,$F7,$81,$FC,$2F,$D4,$7C,$33
	.dc.b	$EB,$F8,$7C,$6E,$63,$18,$23,$13,$A8,$FF,$04,$57,$28,$C4,$7F,$82
	.dc.b	$FA,$FF,$D5,$B5,$FC,$66,$3F,$D6,$1D,$7A,$7C,$D3,$A7,$F8,$F3,$5F
	.dc.b	$F4,$CF,$C8,$6D,$FE,$04,$DF,$04,$D3,$5B,$7E,$A3,$CB,$F9,$A4,$AF
	.dc.b	$9A,$7F,$B3,$8F,$93,$DB,$FC,$5A,$3C,$DB,$FC,$B6,$C6,$67,$E8,$C7
	.dc.b	$D8,$E5,$91,$D1,$AF,$4E,$1F,$F0,$E5,$FD,$59,$E7,$18,$BF,$E4,$B1
	.dc.b	$8A,$87,$DE,$29,$F6,$D7,$C2,$3E,$8D,$0F,$87,$FE,$8C,$EA,$33,$18
	.dc.b	$9E,$13,$BA,$FF,$20,$3B,$30,$FE,$27,$8B,$F0,$BF,$51,$FD,$09,$B1
	.dc.b	$FC,$E5,$BD,$9F,$87,$9D,$72,$BF,$2A,$FE,$91,$7F,$86,$8C,$2F,$27
	.dc.b	$F1,$D3,$A8,$F0,$4E,$A7,$E0,$C7,$E6,$F7,$FC,$F3,$83,$F3,$BB,$7F
	.dc.b	$4C,$7D,$7F,$BB,$BE,$DB,$A4,$7C,$68,$C3,$E1,$37,$1A,$D5,$3A,$C3
	.dc.b	$B6,$1E,$0B,$FA,$84,$3E,$A6,$DF,$92,$BF,$93,$1F,$52,$31,$BF,$D1
	.dc.b	$E4,$FA,$73,$F4,$27,$2D,$EB,$FB,$7F,$A3,$FD,$76,$C6,$70,$4E,$5B
	.dc.b	$1C,$FC,$1E,$AF,$91,$AF,$8D,$B1,$F8,$A7,$31,$F2,$B9,$7F,$CE,$99
	.dc.b	$FB,$E3,$F3,$47,$D2,$AF,$C8,$FA,$9C,$B2,$34,$91,$F0,$8F,$CB,$F8
	.dc.b	$DF,$96,$FC,$13,$FD,$09,$18,$0E,$29,$3B,$9F,$8C,$DB,$F1,$9F,$44
	.dc.b	$FC,$27,$17,$E0,$5C,$9F,$C8,$7A,$5F,$83,$3F,$16,$FD,$1E,$1F,$E9
	.dc.b	$F9,$CC,$62,$B8,$56,$EB,$F5,$86,$DC,$71,$7F,$A7,$7A,$BF,$94,$F8
	.dc.b	$BF,$CB,$FF,$5B,$C7,$4F,$F5,$0C,$51,$89,$D6,$3F,$A3,$6C,$FA,$0F
	.dc.b	$5F,$17,$D2,$D7,$D9,$18,$3F,$9A,$7D,$C7,$F3,$DF,$C5,$BB,$B7,$0F
	.dc.b	$D7,$36,$7F,$56,$D5,$9F,$F4,$0F,$BA,$FF,$1E,$57,$ED,$5B,$3F,$B2
	.dc.b	$6B,$C3,$44,$62,$5E,$DB,$F1,$4F,$95,$7F,$C5,$23,$13,$F2,$A3,$F8
	.dc.b	$D4,$63,$7F,$C1,$B6,$7E,$E4,$FE,$4C,$FD,$8E,$9F,$0A,$3F,$90,$FD
	.dc.b	$F7,$66,$9D,$67,$F6,$11,$9E,$53,$E4,$8E,$87,$6E,$8F,$CF,$3F,$34
	.dc.b	$C5,$FD,$07,$67,$D5,$1E,$D3,$DB,$FF,$72,$7C,$AB,$DD,$FF,$75,$4E
	.dc.b	$7C,$6F,$EE,$53,$9B,$FF,$76,$C6,$2B,$77,$C9,$B1,$99,$D8,$AE,$4F
	.dc.b	$8A,$B8,$3A,$8C,$E3,$FE,$03,$BF,$E9,$E4,$9F,$F0,$4B,$5F,$EA,$CD
	.dc.b	$7F,$22,$FF,$47,$CB,$FA,$6F,$C1,$7E,$D5,$BA,$FF,$19,$6D,$7F,$22
	.dc.b	$6B,$C8,$EE,$43,$B9,$FD,$43,$6F,$C2,$D7,$E0,$B8,$38,$DF,$E3,$1B
	.dc.b	$3A,$8F,$CB,$53,$E2,$3F,$5F,$C5,$FE,$1F,$6C,$66,$77,$1F,$A6,$87
	.dc.b	$E2,$54,$93,$FA,$88,$CB,$E7,$7D,$47,$A0,$87,$E8,$FA,$7F,$C5,$DF
	.dc.b	$AD,$7E,$F6,$0D,$24,$FF,$09,$ED,$F8,$BD,$DF,$C4,$21,$AF,$F7,$AB
	.dc.b	$FE,$23,$5A,$EE,$FE,$25,$39,$9F,$19,$0F,$E7,$1C,$1F,$66,$FF,$96
	.dc.b	$7E,$13,$FC,$AF,$C0,$FF,$02,$B6,$0D,$3F,$98,$78,$5F,$7A,$FF,$AA
	.dc.b	$35,$FE,$BA,$D7,$FA,$E3,$5F,$09,$F1,$CA,$7C,$69,$FE,$9C,$7B,$9F
	.dc.b	$EA,$CD,$7F,$AB,$35,$FE,$AC,$D3,$FC,$3F,$E2,$FF,$87,$BF,$19,$FE
	.dc.b	$C0,$49,$FE,$C8,$FF,$6D,$9E,$1D,$9F,$DC,$F3,$C1,$2B,$CD,$CF,$FA
	.dc.b	$56,$F6,$7F,$B5,$B5,$FF,$14,$F9,$3F,$E0,$6E,$5B,$E5,$7E,$36,$FF
	.dc.b	$C0,$E7,$E4,$4B,$1F,$A9,$7E,$09,$F5,$70,$EC,$FF,$87,$3C,$BE,$4D
	.dc.b	$39,$16,$7F,$C4,$6E,$0F,$E6,$1E,$67,$F9,$EF,$AB,$FC,$5D,$1F,$CC
	.dc.b	$0F,$F1,$64,$65,$A6,$4F,$F0,$AF,$9B,$F6,$77,$FF,$1F,$6D,$FE,$43
	.dc.b	$8C,$6F,$FE,$44,$8F,$56,$7E,$8B,$F5,$6F,$F1,$5F,$F4,$8F,$C5,$FF
	.dc.b	$96,$88,$E0,$FF,$36,$BF,$E9,$C2,$38,$3F,$D3,$3C,$7F,$E6,$22,$38
	.dc.b	$3F,$CE,$AF,$FA,$90,$8E,$0F,$F4,$E1,$FE,$9D,$F4,$3F,$D0,$2B,$E2
	.dc.b	$3F,$D1,$D1,$86,$F9,$FF,$EA,$36,$CE,$CF,$F4,$7F,$A7,$96,$99,$3F
	.dc.b	$D3,$4E,$5A,$64,$FE,$C9,$EA,$E5,$A6,$7C,$AF,$F5,$23,$7C,$AF,$F5
	.dc.b	$5D,$FF,$A0,$BE,$22,$4F,$FA,$F1,$AF,$FB,$03,$AF,$FD,$75,$1F,$EC
	.dc.b	$62,$47,$FA,$69,$FF,$67,$91,$F8,$A3,$E0,$9F,$07,$CA,$33,$18,$53
	.dc.b	$FE,$DB,$69,$FE,$DF,$29,$F4,$3D,$B7,$B7,$CE,$30,$D7,$FD,$A1,$EE
	.dc.b	$7F,$BB,$BA,$BF,$D8,$0D,$8F,$2B,$5F,$B6,$D7,$FD,$B1,$EE,$BD,$7F
	.dc.b	$F0,$54,$7B,$23,$1D,$7F,$9D,$8D,$CE,$9F,$C2,$CE,$AF,$FD,$A1,$EF
	.dc.b	$BD,$FD,$57,$FB,$96,$31,$87,$38,$46,$2B,$8D,$70,$95,$FF,$95,$AF
	.dc.b	$83,$6B,$FD,$BC,$D9,$F4,$78,$DC,$70,$7C,$05,$26,$1E,$18,$CB,$E3
	.dc.b	$EC,$FF,$7E,$63,$FE,$C8,$26,$1F,$7A,$32,$FF,$B7,$23,$FD,$DD,$EE
	.dc.b	$B8,$FF,$82,$4F,$E2,$CF,$74,$65,$F8,$1F,$06,$F8,$2D,$C0,$46,$5D
	.dc.b	$C6,$9C,$B7,$E0,$1D,$74,$93,$F3,$63,$2F,$83,$E0,$F9,$59,$A4,$8F
	.dc.b	$A9,$CF,$F9,$7C,$FF,$3B,$1F,$9D,$E8,$F1,$4E,$63,$F3,$B1,$E5,$E7
	.dc.b	$FC,$BC,$7D,$D6,$D7,$E7,$C9,$1F,$A8,$EE,$F2,$C7,$F0,$AC,$FD,$D1
	.dc.b	$89,$D4,$7C,$1F,$84,$7B,$D4,$93,$F8,$B8,$CB,$F8,$DF,$82,$FE,$FD
	.dc.b	$87,$A9,$B1,$FB,$86,$BF,$C7,$9A,$FD,$3F,$85,$6F,$F4,$5B,$1F,$99
	.dc.b	$6D,$7C,$66,$C7,$F0,$47,$D4,$FF,$2A,$6B,$D7,$1E,$87,$DD,$F8,$6F
	.dc.b	$D7,$F9,$FF,$41,$B3,$FC,$39,$AF,$F3,$46,$BF,$89,$6A,$69,$3B,$1C
	.dc.b	$FF,$E5,$C9,$E1,$18,$AF,$03,$63,$C0,$D7,$C0,$D7,$8C,$F3,$7C,$11
	.dc.b	$96,$C6,$6B,$C4,$D9,$D4,$62,$BC,$33,$92,$38,$25,$79,$23,$2D,$8C
	.dc.b	$BE,$68,$CB,$63,$2F,$A1,$AF,$99,$B3,$E3,$8E,$84,$39,$14,$E4,$53
	.dc.b	$91,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$25
	.dc.b	$7C,$E2,$57,$C4,$9E,$51,$8A,$F8,$2D,$8F,$53,$5F,$2B,$5E,$33,$CD
	.dc.b	$F0,$C6,$5B,$19,$AF,$1B,$67,$51,$8A,$F0,$CE,$48,$E0,$95,$E4,$8C
	.dc.b	$B6,$32,$F9,$A3,$2D,$8C,$BE,$87,$F0,$3E,$F7,$FB,$26,$3F,$8F,$B5
	.dc.b	$FE,$FA,$D7,$FD,$8E,$7E,$D7,$CF,$FF,$43,$3F,$B3,$6B,$FB,$36,$BF
	.dc.b	$B3,$6D,$9A,$4B,$FE,$5F,$3B,$6C,$ED,$B3,$B9,$F1,$46,$08,$E0,$87
	.dc.b	$F0,$98,$E0,$D8,$DB,$4D,$9E,$B7,$2D,$32,$7B,$5C,$B4,$C9,$E9,$2C
	.dc.b	$FA,$5B,$3E,$27,$A1,$E1,$21,$82,$1D,$4C,$94,$C9,$DC,$C9,$4C,$C7
	.dc.b	$85,$CB,$5C,$C7,$8D,$E8,$43,$91,$63,$83,$63,$83,$A2,$46,$9B,$1A
	.dc.b	$8C,$11,$C1,$27,$E5,$CF,$06,$CE,$DB,$3B,$9F,$97,$18,$23,$82,$1F
	.dc.b	$02,$38,$36,$36,$D3,$67,$C4,$72,$D3,$27,$C1,$72,$D3,$27,$A4,$B3
	.dc.b	$E9,$73,$F8,$D7,$FB,$C7,$DF,$FF,$06,$B6,$3F,$CD,$6D,$7F,$D2,$2D
	.dc.b	$7F,$BA,$7E,$03,$FE,$27,$9F,$EE,$0D,$8F,$EE,$0D,$7F,$B8,$35,$FE
	.dc.b	$F9,$F8,$0D,$5B,$23,$0D,$85,$7F,$7D,$9E,$1C,$3E,$39,$34,$99,$75
	.dc.b	$F8,$A7,$B5,$49,$5D,$0F,$F4,$4C,$69,$C1,$FE,$B0,$E6,$FF,$1C,$E3
	.dc.b	$FE,$25,$8C,$9F,$EB,$3E,$2F,$54,$EA,$3F,$D2,$7C,$BF,$EB,$8A,$FF
	.dc.b	$1C,$B6,$32,$7C,$17,$0F,$FA,$FB,$A7,$FD,$2D,$3F,$E4,$06,$C7,$F8
	.dc.b	$F8,$F0,$9F,$E4,$32,$F4,$7C,$32,$67,$F3,$5F,$8D,$CE,$E0,$FA,$0D
	.dc.b	$AE,$6D,$C3,$FC,$9E,$D8,$FF,$56,$F5,$FF,$98,$E7,$FC,$A6,$D8,$CF
	.dc.b	$67,$DB,$8C,$47,$FB,$61,$F9,$93,$C3,$7D,$12,$7D,$7D,$EF,$3F,$FC
	.dc.b	$9B,$5F,$E6,$08,$C7,$66,$A7,$35,$FE,$83,$6B,$FE,$E6,$FA,$47,$C6
	.dc.b	$7F,$8B,$B0,$FF,$34,$C6,$37,$F9,$2D,$8C,$CE,$EB,$FD,$33,$5F,$3B
	.dc.b	$C6,$FF,$50,$57,$FA,$7B,$C6,$D5,$7F,$AC,$1A,$FF,$AE,$BD,$0A,$FF
	.dc.b	$AA,$1B,$46,$9A,$EA,$7D,$CD,$26,$3F,$7A,$32,$F8,$FC,$6E,$64,$70
	.dc.b	$7E,$13,$C4,$F6,$63,$FE,$E4,$6B,$FE,$9D,$3F,$DC,$BA,$79,$A3,$31
	.dc.b	$FE,$A1,$E2,$FF,$50,$7C,$6F,$DE,$E4,$FF,$52,$36,$4D,$3E,$E3,$83
	.dc.b	$F0,$FC,$9C,$1A,$64,$FC,$23,$97,$E5,$63,$97,$FD,$C7,$BF,$F9,$9A
	.dc.b	$BE,$84,$E6,$7F,$DA,$73,$97,$FD,$CD,$B5,$7F,$DC,$6D,$7E,$A4,$F1
	.dc.b	$DF,$E9,$3C,$F6,$F9,$B8,$7F,$A2,$77,$E1,$7F,$C5,$1B,$3A,$8F,$CF
	.dc.b	$9F,$C1,$CF,$CF,$67,$FE,$C9,$CF,$E3,$12,$3E,$0B,$F1,$88,$E0,$FA
	.dc.b	$98,$3E,$81,$F3,$1F,$E8,$6F,$D6,$BE,$C5,$7F,$A2,$27,$34,$9C,$BC
	.dc.b	$6B,$85,$74,$AD,$D6,$BC,$67,$B5,$AA,$FF,$70,$60,$9C,$BD,$63,$CB
	.dc.b	$18,$AE,$55,$C6,$7F,$DD,$1D,$3F,$B6,$7F,$71,$D1,$FE,$C2,$7F,$6D
	.dc.b	$D3,$FE,$2C,$9D,$47,$C0,$F6,$5F,$71,$FE,$A4,$F1,$3D,$71,$8D,$3F
	.dc.b	$9E,$D6,$EB,$4D,$8C,$BF,$EF,$9F,$59,$EB,$3F,$02,$FF,$1D,$C6,$1B
	.dc.b	$19,$9F,$F7,$6C,$E5,$FF,$77,$D3,$0F,$F7,$EA,$F8,$CF,$2F,$8F,$FE
	.dc.b	$BA,$6C,$66,$BD,$93,$97,$E0,$9F,$13,$5F,$85,$B7,$C5,$8F,$DA,$6D
	.dc.b	$F1,$1F,$A7,$8F,$F0,$0C,$7E,$1F,$3F,$39,$CB,$5F,$F6,$5C,$ED,$EE
	.dc.b	$43,$AD,$F7,$7E,$BB,$72,$F7,$79,$3F,$32,$7C,$2D,$7A,$B6,$36,$7D
	.dc.b	$08,$69,$F4,$3B,$D0,$F4,$7F,$33,$39,$8C,$1F,$C4,$DC,$46,$B7,$FE
	.dc.b	$73,$3A,$3E,$E7,$E3,$3F,$DD,$78,$F9,$B6,$39,$35,$DF,$4F,$8A,$7E
	.dc.b	$84,$43,$E2,$1F,$0B,$87,$F0,$34,$EA,$E1,$3D,$1A,$F8,$9A,$FC,$08
	.dc.b	$E0,$FC,$3D,$1F,$B6,$7A,$BF,$60,$FF,$7F,$CF,$EB,$A3,$13,$FC,$01
	.dc.b	$B1,$9D,$BE,$4D,$7F,$A7,$39,$3F,$09,$3E,$73,$D4,$7D,$CC,$BF,$82
	.dc.b	$BF,$E1,$EF,$9E,$F5,$63,$FE,$24,$3F,$A0,$6A,$8C,$EB,$E4,$AE,$1A
	.dc.b	$FE,$C6,$F7,$5A,$7F,$8A,$FE,$01,$FD,$D5,$7D,$5F,$15,$19,$8C,$36
	.dc.b	$38,$F2,$1E,$27,$D2,$C1,$A7,$B9,$B3,$F9,$76,$E9,$FB,$4B,$7D,$77
	.dc.b	$3F,$98,$EE,$DC,$7C,$ED,$BF,$E3,$B5,$C2,$55,$E9,$F8,$07,$CF,$7E
	.dc.b	$EE,$7F,$99,$C6,$1F,$E0,$06,$23,$57,$FC,$A2,$75,$D7,$F2,$FA,$1C
	.dc.b	$CF,$6F,$F2,$58,$FE,$59,$B8,$7E,$71,$AF,$E2,$DA,$FC,$03,$EF,$9F
	.dc.b	$81,$3F,$33,$4F,$A9,$FF,$04,$C7,$DF,$3F,$91,$38,$E4,$F2,$36,$FF
	.dc.b	$60,$D9,$FC,$51,$D8,$F0,$1F,$57,$87,$EA,$13,$87,$EA,$3F,$5F,$DA
	.dc.b	$FF,$7C,$7C,$AF,$F4,$9A,$FE,$03,$18,$F1,$7F,$DE,$F3,$99,$4E,$5D
	.dc.b	$FC,$B7,$F9,$CE,$75,$39,$9D,$CE,$6B,$99,$FA,$13,$F9,$61,$23,$FA
	.dc.b	$7B,$5D,$18,$21,$F1,$BC,$1F,$E2,$FF,$15,$F1,$9B,$A7,$C6,$6C,$EC
	.dc.b	$4F,$ED,$73,$73,$FB,$67,$F8,$66,$1E,$16,$9F,$81,$38,$E1,$6F,$F1
	.dc.b	$1C,$3F,$87,$9C,$9F,$85,$29,$FA,$9E,$3F,$CA,$7C,$4F,$E6,$79,$7F
	.dc.b	$58,$DC,$0D,$3F,$4A,$BE,$36,$91,$FD,$39,$93,$E2,$DF,$A0,$F9,$24
	.dc.b	$3F,$75,$D1,$FC,$0F,$E9,$3F,$7F,$B7,$EF,$5B,$46,$9F,$C3,$FB,$BF
	.dc.b	$89,$77,$7F,$03,$DB,$1D,$DF,$C6,$1B,$19,$BF,$E2,$6D,$CB,$C4,$7F
	.dc.b	$04,$E0,$F9,$CD,$3F,$96,$78,$1F,$D4,$BF,$51,$FE,$2D,$E1,$7F,$4C
	.dc.b	$6C,$67,$0D,$D2,$7E,$9E,$1F,$84,$8F,$E8,$47,$E0,$0B,$8F,$F5,$28
	.dc.b	$D4,$79,$3A,$7F,$D2,$F9,$79,$1B,$1F,$D5,$B4,$DB,$FD,$25,$BD,$9F
	.dc.b	$C2,$1B,$CB,$FE,$D3,$C7,$FB,$26,$9F,$DB,$3F,$4F,$B8,$1A,$47,$FC
	.dc.b	$59,$E3,$7C,$D6,$ED,$E9,$7F,$BE,$B5,$FE,$FA,$D7,$E5,$1E,$BE,$2F
	.dc.b	$CE,$EB,$F9,$B9,$FF,$12,$74,$FE,$42,$BF,$C7,$4D,$7F,$C2,$B1,$F9
	.dc.b	$27,$2D,$7A,$74,$FE,$9D,$EE,$7E,$3C,$FF,$0D,$F9,$1C,$DB,$3C,$5B
	.dc.b	$D1,$FE,$30,$3F,$C1,$1E,$4F,$F8,$43,$1B,$3F,$B3,$E1,$FF,$1E,$46
	.dc.b	$0F,$F2,$6F,$95,$F1,$67,$FC,$5E,$D7,$FC,$56,$DC,$3F,$C8,$4E,$63
	.dc.b	$CE,$4E,$3F,$BA,$FE,$96,$7F,$CD,$5E,$5F,$F5,$FF,$D8,$BF,$CC,$F0
	.dc.b	$69,$27,$FC,$ED,$E7,$ED,$B3,$B6,$CF,$F9,$FB,$CF,$FF,$40,$BF,$EB
	.dc.b	$0F,$4E,$4F,$FA,$B9,$AF,$FA,$50,$FF,$68,$1F,$E9,$87,$FD,$74,$47
	.dc.b	$FD,$1B,$E8,$FF,$B3,$BE,$2F,$FD,$45,$81,$A4,$9F,$F5,$7C,$65,$B1
	.dc.b	$FE,$12,$F5,$3F,$D5,$AD,$BF,$F0,$03,$FE,$BA,$24,$FF,$BB,$9A,$FF
	.dc.b	$B1,$0E,$85,$7F,$D9,$A4,$7F,$D6,$2F,$FB,$4C,$8E,$0F,$F7,$07,$47
	.dc.b	$F8,$3F,$E1,$BF,$C4,$BE,$C7,$FC,$01,$6C,$66,$77,$8A,$7D,$5E,$CE
	.dc.b	$67,$FD,$D1,$92,$31,$1F,$E5,$E8,$C5,$71,$8F,$F5,$BC,$62,$BE,$59
	.dc.b	$23,$E7,$F4,$72,$7F,$D9,$D5,$C5,$FF,$6C,$B5,$FF,$6C,$CF,$CB,$8C
	.dc.b	$42,$BF,$DB,$73,$FE,$E6,$24,$78,$BD,$8E,$D2,$7C,$1E,$DF,$C3,$9F
	.dc.b	$F6,$6B,$5F,$F3,$C8,$70,$FF,$9F,$1A,$FF,$8D,$2B,$FD,$38,$47,$FD
	.dc.b	$7E,$FF,$95,$3D,$A7,$83,$FE,$84,$8C,$36,$7F,$DB,$B3,$A7,$DA,$D9
	.dc.b	$E5,$E1,$6E,$7F,$39,$18,$77,$19,$7F,$D2,$3E,$E3,$DD,$C4,$23,$0F
	.dc.b	$CA,$6D,$7F,$9A,$9B,$1F,$EF,$0F,$7B,$FC,$F3,$3F,$E6,$E6,$C6,$75
	.dc.b	$EB,$18,$8F,$93,$EF,$B8,$BC,$2D,$8F,$B2,$DB,$FF,$3C,$B6,$3F,$DF
	.dc.b	$FF,$01,$F4,$A7,$FC,$FA,$D8,$C9,$EC,$70,$FD,$8F,$81,$78,$5F,$01
	.dc.b	$B1,$F8,$86,$D7,$FA,$21,$B1,$F8,$01,$F0,$5F,$45,$C3,$5F,$C2,$B5
	.dc.b	$FF,$5F,$B5,$FF,$29,$4F,$FA,$53,$D6,$F8,$91,$FE,$60,$8C,$69,$F5
	.dc.b	$3A,$9E,$3D,$25,$7E,$7A,$32,$DE,$0E,$04,$70,$7E,$93,$E0,$FB,$A6
	.dc.b	$92,$FF,$07,$F0,$8F,$85,$FF,$4A,$78,$BF,$EA,$A8,$C4,$FF,$55,$6C
	.dc.b	$66,$77,$7F,$1E,$72,$D8,$CB,$FA,$A8,$CB,$63,$2F,$FB,$9A,$32,$D8
	.dc.b	$CA,$8C,$B6,$33,$18,$23,$82,$57,$A5,$B5,$E9,$8C,$11,$C1,$2B,$D5
	.dc.b	$19,$6C,$65,$F5,$46,$5B,$19,$7D,$51,$96,$C6,$5F,$54,$65,$B1,$9A
	.dc.b	$F5,$CE,$5B,$19,$AF,$5D,$79,$08,$78,$8A,$78,$0A,$76,$2C,$F4,$6C
	.dc.b	$74,$79,$10,$E4,$53,$91,$63,$93,$63,$93,$C0,$87,$02,$C7,$06,$C7
	.dc.b	$07,$44,$8D,$36,$35,$18,$23,$82,$57,$C9,$6D,$7C,$98,$C1,$1C,$12
	.dc.b	$BE,$1C,$65,$B1,$97,$DD,$19,$6C,$65,$F5,$C6,$5F,$83,$EB,$BD,$77
	.dc.b	$C2,$FF,$0F,$9F,$F6,$0B,$5F,$F5,$FB,$67,$83,$9C,$5F,$34,$C3,$0D
	.dc.b	$BF,$DD,$35,$FD,$B3,$5F,$DA,$35,$FD,$93,$5F,$D8,$36,$7B,$CF,$62
	.dc.b	$4F,$66,$BD,$9A,$F6,$6B,$D9,$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$85
	.dc.b	$3A,$14,$E8,$57,$A3,$C8,$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81
	.dc.b	$4E,$05,$78,$3A,$21,$A2,$BA,$70,$47,$04,$9F,$A8,$49,$F9,$2D,$9F
	.dc.b	$80,$D7,$D2,$D7,$C8,$D7,$C2,$D7,$C0,$D9,$F0,$4F,$62,$4F,$66,$BD
	.dc.b	$9A,$F6,$6B,$D9,$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$85,$3A,$14,$E8
	.dc.b	$57,$A3,$C8,$87,$F8,$FB,$E6,$FF,$C2,$97,$FB,$36,$C7,$F8,$0D,$AE
	.dc.b	$F9,$CF,$EE,$8F,$F0,$47,$13,$E2,$B3,$7F,$E0,$16,$CE,$9A,$FF,$93
	.dc.b	$A7,$83,$CC,$FF,$28,$F0,$FF,$8D,$AB,$F8,$33,$63,$FD,$2F,$C3,$C9
	.dc.b	$F1,$C6,$5E,$3D,$59,$51,$FE,$41,$7B,$1C,$23,$51,$9A,$FF,$0A,$CE
	.dc.b	$85,$FF,$0D,$8C,$92,$3F,$C6,$FE,$7F,$7E,$93,$49,$1C,$FA,$3F,$C6
	.dc.b	$75,$FE,$31,$E9,$F3,$46,$27,$FC,$64,$D7,$FC,$E7,$D6,$F8,$DF,$63
	.dc.b	$63,$CC,$DB,$FF,$2B,$4F,$C1,$EC,$FB,$35,$F6,$1A,$9A,$62,$BD,$D1
	.dc.b	$FE,$5D,$3A,$F1,$F8,$0F,$99,$B7,$96,$3F,$CD,$F9,$60,$8F,$C9,$EE
	.dc.b	$77,$7D,$36,$C6,$63,$07,$D7,$F8,$6F,$F5,$74,$7D,$16,$E9,$FD,$1D
	.dc.b	$B1,$FE,$57,$6B,$FE,$98,$F3,$5E,$97,$D2,$AF,$E9,$D1,$8E,$0F,$B8
	.dc.b	$D8,$CC,$EE,$BF,$CC,$5F,$17,$F9,$BD,$7F,$04,$DA,$FE,$B0,$D9,$DF
	.dc.b	$37,$F5,$39,$FF,$71,$78,$B6,$7F,$D6,$57,$F8,$BF,$1B,$FD,$64,$D9
	.dc.b	$D4,$FE,$CF,$F0,$9A,$FE,$BB,$AD,$18,$84,$62,$BF,$2C,$DC,$3F,$DE
	.dc.b	$5F,$19,$BC,$4D,$25,$7E,$77,$E3,$9F,$33,$D5,$B1,$98,$C1,$E1,$E5
	.dc.b	$FF,$45,$FC,$77,$FA,$9F,$9B,$F8,$1B,$74,$34,$FF,$5B,$11,$FF,$5F
	.dc.b	$F2,$F0,$69,$C0,$F9,$78,$57,$FD,$A7,$C3,$F4,$EB,$E8,$F0,$FF,$80
	.dc.b	$5B,$5B,$AF,$F7,$A7,$0E,$5F,$A3,$3F,$52,$75,$1F,$EE,$99,$D3,$F4
	.dc.b	$63,$3C,$1F,$36,$BF,$92,$70,$EA,$BE,$A3,$69,$F5,$8F,$BB,$FF,$0A
	.dc.b	$7D,$17,$FA,$AE,$73,$18,$BF,$F7,$6C,$60,$4E,$EB,$94,$62,$7E,$2F
	.dc.b	$9B,$FE,$DF,$23,$83,$9F,$E6,$7F,$C7,$7E,$4B,$C0,$D8,$CF,$17,$F3
	.dc.b	$89,$CB,$63,$FD,$81,$F6,$CF,$A7,$FF,$16,$57,$F8,$D6,$31,$8A,$31
	.dc.b	$3A,$AE,$2D,$7F,$65,$D5,$FC,$89,$FD,$6F,$67,$F8,$F7,$B3,$F6,$51
	.dc.b	$83,$FD,$A1,$EE,$BE,$09,$F8,$37,$F4,$EF,$1F,$31,$8A,$FE,$66,$7F
	.dc.b	$52,$15,$C6,$B8,$46,$63,$0F,$03,$FA,$CE,$8F,$88,$7C,$57,$F9,$0E
	.dc.b	$30,$D8,$CC,$7F,$95,$CC,$95,$7D,$57,$F9,$4F,$E4,$3F,$01,$18,$94
	.dc.b	$62,$75,$7F,$82,$AD,$4F,$C3,$8F,$F7,$96,$9F,$19,$BB,$7C,$B6,$ED
	.dc.b	$E1,$FA,$CE,$71,$EC,$C1,$18,$8F,$F3,$34,$63,$2F,$1C,$78,$67,$4B
	.dc.b	$0F,$03,$F6,$AF,$EC,$61,$E1,$8C,$1F,$53,$F3,$0F,$9E,$FF,$81,$35
	.dc.b	$FB,$51,$88,$FA,$BB,$A3,$13,$A9,$FA,$87,$9B,$EC,$3F,$CF,$9F,$BB
	.dc.b	$FC,$A6,$C7,$C7,$6B,$EF,$6B,$DB,$F3,$0E,$B8,$1A,$48,$EE,$7B,$F1
	.dc.b	$F6,$3F,$E1,$CC,$FE,$0B,$63,$2E,$F3,$FF,$4C,$65,$F9,$7F,$A9,$EE
	.dc.b	$49,$F5,$C6,$5F,$6E,$CF,$D0,$BF,$9D,$C3,$F5,$0F,$F4,$B4,$E7,$C0
	.dc.b	$FF,$16,$C6,$2B,$6D,$8C,$CF,$E0,$AB,$F1,$07,$D4,$30,$7E,$97,$B5
	.dc.b	$FA,$3F,$EF,$63,$FC,$37,$39,$8C,$74,$FF,$A9,$EB,$7D,$5F,$B1,$AD
	.dc.b	$36,$FF,$1D,$E4,$FF,$AC,$7F,$54,$F4,$F4,$7F,$48,$6C,$66,$FF,$3C
	.dc.b	$DA,$F3,$10,$F9,$DC,$1F,$8E,$3F,$59,$E2,$FF,$AF,$76,$EA,$DB,$FD
	.dc.b	$43,$6B,$F4,$0D,$7F,$3F,$F5,$BF,$B3,$9F,$D7,$36,$3F,$4C,$D7,$71
	.dc.b	$FB,$63,$F8,$47,$CD,$3E,$2B,$E0,$4F,$F0,$E8,$C3,$F2,$8F,$D0,$38
	.dc.b	$DF,$FD,$A1,$C6,$8C,$5F,$D4,$AF,$80,$7E,$2C,$FB,$27,$E9,$B7,$EF
	.dc.b	$5F,$C8,$DB,$D1,$F5,$AB,$F8,$E3,$6B,$9F,$3F,$E6,$1E,$87,$73,$E6
	.dc.b	$9D,$8C,$9E,$2F,$7B,$FD,$DA,$BE,$69,$F0,$3F,$CB,$67,$31,$89,$FF
	.dc.b	$7B,$9B,$9D,$7C,$0F,$FA,$0E,$75,$3F,$EF,$A3,$47,$07,$35,$D2,$B8
	.dc.b	$FA,$BF,$E8,$C9,$E3,$87,$F8,$DE,$7F,$A4,$92,$FF,$90,$B5,$FE,$2C
	.dc.b	$D7,$F1,$6F,$F4,$8C,$FF,$D2,$FF,$29,$F5,$33,$FD,$13,$63,$3A,$7C
	.dc.b	$C6,$DF,$43,$B9,$0C,$1F,$80,$E0,$DB,$F8,$56,$ED,$F8,$16,$EF,$F8
	.dc.b	$8C,$3F,$4F,$C2,$C5,$F2,$78,$69,$F9,$82,$E3,$F9,$86,$BF,$CF,$38
	.dc.b	$F9,$7E,$BD,$FA,$69,$34,$8F,$A6,$7E,$29,$4F,$E3,$9B,$7F,$58,$E7
	.dc.b	$F2,$7E,$E9,$FB,$AA,$34,$8F,$F0,$4A,$67,$FC,$27,$B3,$E6,$CE,$63
	.dc.b	$1C,$1F,$1A,$BB,$9E,$1E,$2F,$BC,$F9,$6B,$F9,$13,$EA,$7F,$C4,$FD
	.dc.b	$DF,$CA,$34,$FE,$45,$39,$3F,$BD,$C6,$67,$F4,$BD,$FF,$80,$9A,$79
	.dc.b	$8B,$8F,$F8,$32,$35,$1F,$70,$F8,$A5,$3C,$E7,$F5,$DE,$4F,$E9,$8F
	.dc.b	$33,$1E,$0F,$27,$FA,$91,$92,$D7,$F5,$BB,$7B,$34,$93,$FD,$76,$32
	.dc.b	$FE,$FB,$C5,$E9,$CF,$FD,$9E,$FF,$B8,$4F,$58,$C3,$F2,$FC,$3F,$F6
	.dc.b	$8F,$1F,$E8,$E3,$D2,$7F,$80,$BC,$6F,$C8,$73,$7F,$85,$9B,$1F,$E0
	.dc.b	$3C,$BF,$26,$D9,$E5,$C3,$F5,$A3,$E2,$35,$FF,$06,$F9,$1C,$9F,$F0
	.dc.b	$93,$5C,$D7,$47,$2D,$7F,$C2,$BE,$4F,$E8,$F1,$FF,$14,$B6,$3F,$C4
	.dc.b	$CD,$7F,$C3,$CD,$7F,$C1,$BE,$57,$C6,$E1,$FF,$23,$6D,$FE,$48,$8C
	.dc.b	$1F,$56,$7E,$19,$32,$F5,$38,$F3,$3F,$CB,$FF,$21,$FE,$5E,$CC,$D2
	.dc.b	$4F,$F9,$B0,$C9,$5F,$F3,$6B,$83,$FC,$9F,$E7,$BC,$F4,$64,$BE,$E7
	.dc.b	$FB,$93,$C1,$FF,$2E,$46,$3D,$8D,$C6,$21,$3B,$7F,$D5,$31,$FE,$89
	.dc.b	$F5,$3F,$86,$7C,$E5,$CC,$D2,$57,$6F,$0B,$FD,$95,$F2,$3F,$E9,$EA
	.dc.b	$34,$93,$FE,$DA,$7F,$D5,$44,$70,$7F,$AE,$1F,$F5,$99,$1C,$1F,$EB
	.dc.b	$E7,$D5,$8B,$2F,$87,$18,$6C,$EC,$F8,$79,$BD,$BC,$C7,$FA,$06,$30
	.dc.b	$FA,$0D,$C6,$9F,$9F,$1A,$AF,$95,$EE,$7F,$B7,$BE,$32,$F8,$DF,$CC
	.dc.b	$DB,$99,$A7,$FB,$B1,$C6,$1F,$5B,$BF,$FD,$DB,$18,$8F,$F4,$A4,$62
	.dc.b	$78,$1F,$34,$91,$F4,$7D,$3F,$03,$6B,$93,$FE,$DA,$6B,$FE,$D7,$6D
	.dc.b	$3D,$BE,$33,$E0,$6B,$F8,$EF,$75,$EE,$E2,$11,$8B,$EA,$DC,$7F,$DD
	.dc.b	$67,$FA,$E0,$93,$FE,$90,$6B,$FE,$B1,$F7,$DF,$21,$F9,$C9,$FC,$B4
	.dc.b	$62,$7D,$2D,$8C,$CE,$E7,$FE,$00,$C7,$E0,$5F,$7F,$17,$E1,$5B,$C5
	.dc.b	$B7,$F0,$1C,$55,$FB,$AD,$7F,$5C,$DE,$2F,$C2,$DF,$EA,$7E,$07,$FB
	.dc.b	$85,$7C,$C6,$C6,$74,$E5,$18,$9F,$D9,$FC,$13,$E2,$3B,$36,$3E,$AB
	.dc.b	$6F,$F6,$CD,$8F,$DD,$D3,$E1,$3A,$92,$3C,$4F,$A7,$E1,$7C,$F1,$FC
	.dc.b	$09,$AF,$FB,$7D,$AF,$F0,$A6,$AF,$86,$FF,$78,$61,$FC,$1D,$B1,$FC
	.dc.b	$31,$AF,$FB,$A5,$AF,$F9,$F6,$3F,$87,$7B,$3F,$06,$3F,$DD,$CD,$74
	.dc.b	$7F,$16,$F8,$7F,$E3,$8F,$E5,$9B,$3F,$C5,$DB,$1B,$D7,$EB,$43,$E2
	.dc.b	$29,$27,$93,$F6,$EF,$F3,$31,$FA,$26,$DF,$9B,$97,$47,$E0,$7B,$3E
	.dc.b	$85,$7F,$2E,$0E,$1F,$E6,$2D,$73,$BF,$E2,$7E,$25,$F1,$5F,$D3,$65
	.dc.b	$18,$E2,$FF,$2E,$C7,$F3,$A8,$C7,$95,$FC,$D5,$B1,$FC,$E9,$AF,$8D
	.dc.b	$AF,$1A,$ED,$18,$8E,$F1,$82,$57,$82,$32,$D9,$E0,$49,$DD,$78,$DB
	.dc.b	$5E,$36,$BE,$36,$BE,$3A,$E9,$18,$8E,$F3,$CE,$72,$47,$04,$8E,$D1
	.dc.b	$96,$C6,$5F,$04,$65,$B1,$97,$C4,$D7,$C2,$72,$21,$C0,$A7,$02,$9C
	.dc.b	$0D,$10,$D1,$4D,$18,$21,$82,$57,$CD,$9F,$44,$6A,$31,$3E,$76,$C7
	.dc.b	$89,$AF,$76,$BC,$6B,$E7,$46,$23,$BC,$60,$95,$E0,$8C,$B6,$78,$12
	.dc.b	$77,$5E,$36,$D7,$8D,$AF,$8D,$AF,$8E,$BA,$46,$23,$BC,$F3,$9C,$91
	.dc.b	$C1,$23,$B4,$65,$B1,$F6,$FD,$CF,$F6,$7C,$7F,$87,$9A,$FF,$B0,$8F
	.dc.b	$F0,$F3,$87,$F5,$DC,$BF,$CE,$1F,$D6,$35,$FD,$63,$5D,$D7,$F8,$9D
	.dc.b	$FD,$CC,$65,$B1,$9C,$3F,$C1,$8D,$9D,$46,$27,$F7,$91,$82,$4E,$C9
	.dc.b	$1A,$6B,$E6,$6B,$E6,$6B,$E6,$38,$10,$E0,$53,$81,$4E,$06,$88,$68
	.dc.b	$A6,$8C,$10,$C1,$0F,$86,$75,$29,$C8,$A7,$12,$9C,$67,$E3,$3B,$6B
	.dc.b	$B6,$BB,$9F,$92,$F7,$8C,$BD,$08,$E0,$91,$C2,$3B,$35,$EC,$D7,$B4
	.dc.b	$FC,$E7,$9C,$65,$B1,$99,$EA,$D9,$D4,$62,$7C,$D1,$82,$4E,$C9,$1A
	.dc.b	$6B,$E6,$6B,$E6,$6B,$E6,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6
	.dc.b	$8C,$10,$FC,$07,$EB,$FF,$E1,$9F,$77,$FD,$18,$D8,$FF,$48,$B5,$FE
	.dc.b	$D4,$D7,$FB,$4F,$CA,$7F,$A6,$A7,$FD,$29,$AA,$31,$3F,$E8,$D6,$BF
	.dc.b	$DC,$E3,$4A,$7F,$C0,$DC,$3A,$8F,$F5,$2F,$17,$F8,$15,$D1,$FE,$0F
	.dc.b	$8D,$28,$D4,$FF,$83,$FE,$3F,$30,$69,$27,$99,$FE,$1C,$CB,$9F,$1F
	.dc.b	$F9,$12,$FF,$C4,$31,$83,$BF,$C8,$3E,$1A,$92,$F9,$9F,$EB,$C8,$D3
	.dc.b	$81,$D1,$FE,$96,$BF,$F5,$D3,$63,$FC,$7D,$BF,$B2,$31,$1F,$E3,$53
	.dc.b	$DF,$D1,$FE,$1B,$AF,$F1,$E4,$60,$F8,$7D,$6F,$B8,$A4,$8F,$14,$65
	.dc.b	$F2,$76,$7B,$DF,$F2,$AE,$9F,$32,$31,$1F,$E2,$96,$BF,$EA,$AE,$0F
	.dc.b	$A3,$D7,$F4,$23,$FD,$6B,$18,$3E,$AF,$77,$F9,$E7,$2F,$F3,$E1,$27
	.dc.b	$FD,$BB,$19,$6E,$FC,$08,$FC,$98,$F9,$73,$FE,$E1,$F3,$9E,$B3,$E1
	.dc.b	$7E,$74,$61,$B1,$99,$E0,$DB,$FF,$76,$FC,$93,$E5,$1B,$7D,$78,$C3
	.dc.b	$63,$37,$FE,$F1,$9C,$BF,$A0,$F3,$B3,$7F,$EE,$B6,$BF,$9F,$6B,$FE
	.dc.b	$50,$6A,$F2,$5E,$7E,$1B,$8F,$E2,$5B,$7F,$F0,$05,$B5,$FE,$71,$F9
	.dc.b	$4E,$7D,$66,$92,$FE,$F8,$97,$C8,$FF,$A7,$DA,$7E,$08,$A7,$E0,$63
	.dc.b	$84,$7F,$9A,$E3,$F0,$46,$8A,$E5,$71,$FE,$E6,$30,$D8,$E5,$C1,$F5
	.dc.b	$E7,$EA,$F0,$73,$94,$E5,$DF,$07,$DF,$AF,$A5,$C1,$F6,$1B,$8F,$1E
	.dc.b	$0F,$B3,$18,$3E,$B4,$75,$ED,$70,$F1,$C7,$EC,$4E,$63,$11,$FE,$89
	.dc.b	$F5,$1E,$AF,$E1,$BE,$47,$DD,$18,$B4,$62,$75,$5F,$1E,$B5,$3E,$22
	.dc.b	$47,$81,$AF,$79,$FC,$9F,$E2,$9C,$EC,$D2,$4F,$A1,$F6,$91,$C1,$F2
	.dc.b	$37,$FF,$11,$6F,$FA,$58,$F9,$E4,$3F,$D4,$DE,$F3,$DA,$7E,$1F,$FC
	.dc.b	$CE,$30,$D8,$CC,$FF,$20,$9C,$B6,$1F,$17,$FE,$31,$9E,$8D,$8C,$CF
	.dc.b	$E1,$E7,$2F,$E2,$72,$7A,$2F,$47,$FC,$79,$18,$6C,$66,$38,$36,$7D
	.dc.b	$87,$FB,$07,$2F,$80,$FC,$33,$27,$F0,$0F,$8E,$FD,$FB,$F9,$DA,$46
	.dc.b	$23,$FD,$95,$18,$E5,$F1,$CE,$67,$B7,$37,$F1,$37,$A9,$F9,$A7,$0F
	.dc.b	$84,$F9,$F9,$BF,$07,$F9,$84,$8F,$94,$FF,$63,$C1,$EE,$FF,$51,$5F
	.dc.b	$95,$79,$E3,$0D,$F0,$7F,$8D,$B6,$77,$7E,$AA,$F8,$CD,$DB,$E6,$C6
	.dc.b	$23,$F8,$34,$F8,$27,$FC,$D1,$AF,$CA,$9F,$7F,$17,$E3,$34,$F6,$C6
	.dc.b	$0F,$27,$D1,$F4,$26,$1F,$77,$14,$74,$AE,$35,$FC,$57,$0E,$CD,$8C
	.dc.b	$BF,$2F,$0F,$01,$1F,$9F,$3C,$1F,$E1,$3F,$03,$FE,$15,$C7,$F8,$7B
	.dc.b	$63,$F0,$8D,$7C,$0D,$7B,$1D,$2F,$C3,$5D,$48,$E0,$97,$F9,$9F,$9A
	.dc.b	$7C,$47,$F4,$CD,$3F,$D3,$19,$27,$2F,$D2,$31,$3B,$7F,$1F,$1A,$9E
	.dc.b	$3E,$A0,$FE,$AF,$96,$1B,$5F,$96,$74,$FD,$A8,$FB,$71,$87,$C7,$F8
	.dc.b	$4B,$EE,$BF,$3B,$1F,$80,$8C,$3C,$4F,$EC,$AE,$1E,$31,$C1,$FA,$D9
	.dc.b	$7D,$37,$F3,$C7,$E0,$CF,$25,$79,$9F,$1B,$5F,$EE,$2F,$96,$F6,$78
	.dc.b	$63,$E9,$9D,$0E,$E7,$D5,$9F,$03,$DC,$87,$F5,$FF,$5B,$F8,$9F,$8F
	.dc.b	$FD,$87,$E4,$BE,$54,$E7,$AF,$FC,$9B,$18,$85,$70,$51,$98,$C3,$67
	.dc.b	$8E,$3F,$ED,$98,$C5,$F7,$8C,$1F,$A8,$3F,$AF,$92,$FF,$9B,$35,$D1
	.dc.b	$82,$1F,$23,$1F,$B4,$FC,$88,$F9,$31,$F2,$1F,$C2,$3A,$8F,$D4,$47
	.dc.b	$57,$EB,$B6,$19,$FE,$02,$3E,$EC,$F6,$9C,$E7,$FE,$09,$9C,$9D,$89
	.dc.b	$1E,$12,$4F,$F0,$AD,$7F,$3F,$EF,$FE,$89,$B3,$F9,$ED,$FF,$C6,$33
	.dc.b	$96,$C7,$C0,$AF,$39,$4F,$59,$78,$7F,$C8,$18,$7B,$63,$27,$EF,$87
	.dc.b	$B9,$FA,$87,$F7,$E6,$2B,$F8,$2B,$4E,$11,$B7,$99,$E2,$3E,$A9,$0C
	.dc.b	$10,$D9,$C2,$3F,$9C,$F2,$7E,$7A,$3F,$99,$35,$FE,$5F,$CB,$FC,$92
	.dc.b	$7F,$25,$1F,$E5,$48,$C5,$FB,$70,$F7,$CF,$C1,$E4,$FF,$31,$4F,$F4
	.dc.b	$3E,$4E,$93,$C5,$B8,$7F,$3B,$3A,$96,$7F,$A3,$7E,$53,$2B,$FA,$37
	.dc.b	$47,$F9,$92,$31,$1F,$E4,$9E,$EF,$F0,$37,$57,$F9,$2F,$6F,$F2,$24
	.dc.b	$62,$7E,$79,$FD,$7A,$74,$D3,$27,$C9,$3E,$3E,$9E,$C7,$E6,$65,$FD
	.dc.b	$BC,$F3,$F6,$7F,$9D,$33,$FE,$EC,$D9,$FE,$EA,$D7,$A5,$73,$7E,$A3
	.dc.b	$6B,$FC,$14,$7C,$2E,$EF,$F0,$55,$FF,$80,$5B,$5F,$E0,$06,$BF,$07
	.dc.b	$B7,$FC,$33,$3F,$01,$B1,$FD,$F3,$A7,$F4,$CD,$3E,$DE,$7F,$E2,$6D
	.dc.b	$38,$77,$FC,$49,$E8,$D5,$3B,$7E,$F7,$67,$E1,$1F,$8F,$C5,$F8,$07
	.dc.b	$FC,$B3,$B7,$F9,$78,$8E,$0F,$F2,$97,$83,$FE,$1F,$6E,$DF,$E1,$E6
	.dc.b	$CF,$F9,$3C,$FF,$33,$FC,$2C,$8F,$F3,$2B,$5F,$F2,$B4,$A3,$FC,$AC
	.dc.b	$48,$EA,$FF,$9C,$3E,$16,$47,$F9,$B9,$AF,$FA,$4C,$79,$3F,$EB,$08
	.dc.b	$FF,$3D,$92,$3F,$D6,$FE,$1F,$FA,$DA,$3F,$D0,$E4,$8F,$F2,$F3,$5E
	.dc.b	$9E,$53,$D0,$FF,$50,$4F,$FA,$49,$B3,$FE,$D8,$11,$96,$9F,$EA,$37
	.dc.b	$FD,$4C,$47,$07,$FA,$C7,$1F,$CD,$BF,$EA,$DF,$37,$FD,$65,$18,$84
	.dc.b	$63,$87,$FD,$4C,$D8,$FF,$56,$79,$DF,$EC,$09,$FF,$5F,$79,$FF,$ED
	.dc.b	$46,$DF,$39,$D5,$7F,$A6,$27,$FD,$3A,$47,$07,$C8,$9F,$16,$9F,$ED
	.dc.b	$C9,$FF,$53,$92,$7F,$D0,$CF,$FA,$C8,$8E,$0F,$F7,$A9,$EB,$F4,$1D
	.dc.b	$9F,$97,$84,$61,$D4,$66,$3F,$CF,$51,$C7,$B7,$04,$7E,$36,$DF,$5D
	.dc.b	$BB,$75,$8C,$61,$F5,$4F,$F6,$C1,$27,$FD,$B0,$D7,$F1,$9E,$93,$D3
	.dc.b	$F3,$36,$3E,$A3,$67,$FD,$B0,$D8,$FF,$6F,$7A,$9F,$95,$9F,$F6,$E3
	.dc.b	$63,$27,$C1,$70,$FE,$6F,$D5,$7A,$DF,$15,$B1,$F8,$36,$D7,$FB,$9D
	.dc.b	$B1,$FE,$ED,$F5,$FE,$54,$FF,$BA,$DB,$19,$DB,$E8,$46,$23,$F5,$5E
	.dc.b	$C7,$F9,$12,$BF,$DE,$4D,$8F,$F7,$C3,$5D,$F0,$FD,$9E,$47,$B3,$49
	.dc.b	$3F,$99,$8C,$B7,$83,$81,$1C,$1F,$9F,$F5,$9E,$D7,$E1,$5B,$19,$8C
	.dc.b	$1F,$89,$F6,$DE,$F5,$24,$FE,$A6,$32,$FE,$AF,$DD,$F9,$B9,$FF,$23
	.dc.b	$6F,$8D,$FC,$2E,$42,$32,$E5,$AF,$F1,$16,$BC,$1A,$F1,$E2,$FE,$D6
	.dc.b	$DA,$F0,$46,$27,$64,$70,$48,$D9,$2B,$CD,$5C,$DB,$5C,$67,$53,$E7
	.dc.b	$AE,$55,$A6,$C6,$67,$6F,$AE,$32,$D8,$CB,$ED,$8C,$B6,$32,$F8,$CE
	.dc.b	$93,$DE,$33,$5E,$58,$C1,$27,$64,$3C,$0E,$5A,$64,$F0,$39,$69,$93
	.dc.b	$C4,$70,$21,$A2,$9A,$30,$43,$04,$8F,$43,$D5,$AF,$26,$BC,$5A,$F1
	.dc.b	$AF,$8E,$DA,$F8,$71,$89,$D9,$1C,$12,$36,$4A,$F8,$15,$CD,$B5,$C6
	.dc.b	$75,$3E,$7A,$E5,$5A,$6C,$66,$76,$FA,$E3,$2D,$8C,$BE,$D8,$CB,$63
	.dc.b	$2F,$8C,$E9,$3E,$6F,$53,$FD,$45,$1F,$AD,$6B,$FD,$E5,$AF,$FA,$AF
	.dc.b	$D3,$FC,$EB,$FB,$1F,$73,$FC,$D4,$FE,$B5,$AE,$6F,$78,$FF,$68,$8F
	.dc.b	$D1,$CE,$A3,$11,$FB,$98,$C1,$27,$8B,$63,$8C,$60,$8E,$08,$73,$72
	.dc.b	$D3,$27,$F0,$58,$E8,$E4,$86,$63,$8B,$4E,$27,$BD,$CB,$4C,$9F,$04
	.dc.b	$B3,$EE,$79,$1D,$8E,$05,$38,$15,$E0,$E8,$E2,$43,$04,$38,$99,$29
	.dc.b	$97,$9B,$82,$38,$24,$FC,$48,$E5,$18,$9F,$7C,$70,$9D,$3D,$5A,$E6
	.dc.b	$37,$3C,$63,$E1,$CE,$A3,$11,$EF,$8C,$12,$78,$B6,$38,$C6,$08,$E0
	.dc.b	$87,$37,$2D,$32,$7B,$23,$A3,$92,$19,$8E,$2D,$38,$9E,$F7,$2D,$32
	.dc.b	$7C,$12,$CF,$B9,$E4,$76,$3F,$AD,$E5,$FD,$8A,$FF,$A4,$B6,$3F,$B0
	.dc.b	$35,$FF,$36,$35,$FE,$C7,$C2,$F7,$FF,$B2,$D7,$F8,$19,$B3,$FE,$06
	.dc.b	$6B,$FE,$0C,$6B,$FE,$5D,$E0,$AF,$F9,$71,$BC,$3F,$E3,$9A,$FF,$2E
	.dc.b	$C6,$3A,$A5,$27,$FC,$C3,$18,$7F,$C2,$F9,$F4,$C4,$D2,$57,$13,$FC
	.dc.b	$6B,$E3,$78,$BA,$4D,$24,$6C,$FF,$1A,$70,$FA,$3A,$7F,$C6,$78,$FF
	.dc.b	$8C,$63,$07,$39,$87,$13,$FC,$17,$7F,$E9,$79,$FF,$21,$C7,$C2,$EA
	.dc.b	$F9,$C9,$A4,$E4,$F8,$FE,$F7,$B2,$7A,$07,$15,$FE,$04,$6C,$7F,$86
	.dc.b	$DA,$FF,$A7,$30,$FA,$3A,$FA,$A3,$FC,$40,$D7,$47,$6E,$EF,$F3,$8D
	.dc.b	$FF,$84,$1B,$1F,$E2,$76,$BF,$EB,$36,$BF,$E2,$AF,$81,$79,$BF,$80
	.dc.b	$6C,$7A,$9B,$5F,$E1,$C6,$C7,$F8,$CF,$CE,$79,$5F,$5A,$BF,$C4,$31
	.dc.b	$8E,$0F,$86,$D8,$CC,$EE,$BF,$C7,$9F,$8B,$FF,$A4,$38,$BE,$4B,$6B
	.dc.b	$FC,$56,$D9,$DF,$83,$F8,$A9,$FF,$17,$F8,$AF,$57,$15,$FE,$B0,$6B
	.dc.b	$FE,$D3,$8D,$61,$F5,$9B,$5F,$E3,$8F,$01,$F0,$9C,$2F,$FC,$9F,$18
	.dc.b	$BF,$BA,$DA,$E1,$5B,$A7,$C3,$3E,$1B,$0D,$C3,$F0,$AD,$BF,$F2,$3B
	.dc.b	$6B,$F0,$93,$FE,$5E,$F8,$3F,$C4,$36,$32,$23,$2D,$3F,$CC,$DC,$3F
	.dc.b	$6A,$3E,$C7,$17,$19,$FF,$7A,$4E,$5D,$C6,$5F,$F3,$6F,$0F,$D9,$AF
	.dc.b	$AF,$C5,$D6,$7F,$DF,$93,$97,$F0,$5C,$5F,$B6,$AF,$AF,$C3,$FB,$E6
	.dc.b	$E5,$C6,$BF,$CF,$3C,$5F,$7E,$72,$7F,$9A,$3E,$09,$E9,$FF,$9F,$BE
	.dc.b	$0B,$0D,$BF,$F4,$14,$E9,$4E,$9A,$7E,$1B,$0A,$FF,$A2,$06,$1F,$8F
	.dc.b	$9F,$29,$23,$AD,$FF,$A3,$BE,$0B,$A1,$23,$21,$F6,$91,$F0,$35,$F8
	.dc.b	$F9,$FF,$1A,$7E,$59,$0F,$93,$AF,$FA,$23,$17,$E2,$8F,$4B,$FC,$5B
	.dc.b	$18,$6C,$66,$38,$36,$BF,$D4,$D0,$F6,$DF,$1B,$C9,$B1,$DD,$B3,$C1
	.dc.b	$AF,$C7,$87,$E7,$8F,$7B,$C6,$D8,$F4,$36,$7C,$2D,$7D,$A7,$BC,$F8
	.dc.b	$79,$7F,$92,$B8,$FF,$83,$47,$A3,$E0,$38,$3F,$EB,$EB,$46,$23,$FA
	.dc.b	$4B,$63,$38,$7F,$94,$2B,$9D,$78,$08,$78,$0F,$F6,$57,$03,$F5,$B7
	.dc.b	$D1,$7F,$96,$A3,$0D,$8C,$E1,$FE,$BC,$9C,$BF,$EB,$EF,$75,$EF,$BE
	.dc.b	$D3,$FD,$AB,$FB,$77,$FB,$5E,$73,$18,$CB,$F2,$51,$85,$3B,$AE,$51
	.dc.b	$83,$E9,$7E,$C4,$F8,$CF,$C1,$B7,$1F,$C4,$36,$7F,$02,$D7,$E0,$9F
	.dc.b	$BC,$E6,$FC,$93,$81,$B7,$C8,$7E,$87,$03,$E5,$3F,$22,$DC,$3F,$2C
	.dc.b	$D9,$FB,$13,$F9,$6F,$B9,$FE,$93,$DD,$FD,$8E,$7F,$BF,$C6,$2F,$F1
	.dc.b	$8D,$8F,$BC,$D7,$CD,$C9,$F8,$5C,$BC,$F1,$F6,$4F,$47,$C4,$BD,$2F
	.dc.b	$EE,$18,$7E,$F6,$53,$98,$FC,$54,$7F,$01,$8C,$3E,$53,$73,$C9,$F1
	.dc.b	$F8,$3F,$91,$AF,$3B,$63,$27,$F8,$6A,$32,$D3,$CC,$43,$D6,$7E,$84
	.dc.b	$FF,$10,$F7,$3F,$1A,$E5,$3F,$AF,$8C,$3C,$E7,$73,$A7,$F4,$93,$A9
	.dc.b	$F9,$C3,$E7,$5F,$63,$EB,$AF,$F1,$7C,$62,$7E,$1C,$62,$75,$1C,$5B
	.dc.b	$F5,$CF,$B8,$FE,$05,$5F,$B3,$8C,$3C,$A9,$18,$8F,$1B,$63,$F1,$DC
	.dc.b	$7F,$8D,$9F,$86,$7E,$52,$3C,$07,$3C,$3C,$D1,$EC,$6C,$7F,$13,$6C
	.dc.b	$7C,$D3,$F3,$3B,$7E,$19,$EA,$66,$BC,$0E,$B7,$FC,$61,$1F,$C4,$0F
	.dc.b	$75,$F1,$7F,$E5,$3F,$90,$7C,$07,$FB,$8E,$30,$DF,$35,$6A,$DB,$FE
	.dc.b	$E7,$31,$3B,$69,$98,$DB,$E1,$AE,$AE,$67,$F9,$39,$FD,$C8,$8F,$F4
	.dc.b	$86,$BA,$30,$43,$F9,$56,$7F,$81,$8F,$99,$BF,$F9,$C9,$BA,$FB,$8F
	.dc.b	$9D,$19,$79,$57,$ED,$27,$BB,$4F,$E8,$DB,$7E,$12,$3F,$03,$BF,$C4
	.dc.b	$6C,$F5,$3F,$15,$19,$AE,$E6,$63,$F2,$8D,$3F,$0B,$3E,$63,$F3,$FC
	.dc.b	$1F,$A6,$EE,$BC,$5F,$E8,$C6,$D7,$F3,$33,$F5,$91,$97,$D7,$7F,$83
	.dc.b	$BF,$5B,$4F,$DD,$D7,$EB,$39,$BF,$CC,$6D,$E4,$FB,$ED,$9E,$C7,$F0
	.dc.b	$DE,$5F,$90,$DC,$3D,$8F,$F0,$E6,$F3,$7F,$15,$C7,$F8,$97,$2F,$F9
	.dc.b	$9D,$F1,$B6,$BE,$8D,$32,$FF,$3A,$FC,$47,$84,$91,$FC,$49,$FE,$07
	.dc.b	$D5,$FC,$DF,$0F,$E6,$B3,$F8,$EB,$FE,$D3,$D5,$F8,$17,$FB,$F5,$6D
	.dc.b	$F8,$B8,$F8,$5F,$C6,$10,$FF,$5C,$F5,$FF,$51,$C3,$FA,$7F,$67,$39
	.dc.b	$FE,$F1,$39,$7F,$AC,$C6,$57,$7F,$F6,$1F,$DB,$BF,$B0,$70,$7E,$A9
	.dc.b	$B2,$69,$FD,$E3,$C2,$FE,$E7,$8D,$F0,$FA,$78,$7F,$DD,$23,$1B,$FB
	.dc.b	$F0,$FB,$87,$A4,$A7,$E6,$4F,$EB,$FE,$27,$96,$3F,$C1,$67,$C7,$2C
	.dc.b	$FF,$85,$63,$D5,$3C,$89,$B7,$C7,$7A,$1F,$ED,$EF,$17,$FC,$4D,$5F
	.dc.b	$E2,$4A,$ED,$18,$9F,$F1,$73,$5F,$F1,$07,$8D,$FE,$36,$AF,$F1,$A7
	.dc.b	$8F,$E5,$8C,$4F,$F8,$D5,$AF,$F8,$9F,$C8,$FF,$22,$57,$F9,$09,$B1
	.dc.b	$FE,$3B,$6B,$BF,$27,$D7,$97,$A6,$3D,$7E,$37,$FB,$C2,$FF,$4B,$E5
	.dc.b	$FF,$AD,$7E,$33,$FC,$F6,$4A,$FF,$37,$3F,$E7,$12,$3F,$E6,$66,$BF
	.dc.b	$E7,$18,$FF,$5F,$7C,$67,$FA,$10,$91,$FE,$77,$7F,$CF,$44,$7F,$CD
	.dc.b	$ED,$79,$79,$4E,$07,$C6,$7F,$A3,$89,$3F,$E8,$17,$FD,$32,$47,$FC
	.dc.b	$DE,$D7,$FD,$47,$E0,$7F,$3A,$8F,$F5,$3C,$74,$3F,$D4,$A5,$36,$53
	.dc.b	$FA,$0F,$B7,$FF,$00,$63,$FD,$66,$48,$FF,$5B,$BF,$EB,$EF,$53,$D1
	.dc.b	$F1,$17,$03,$49,$5F,$E1,$BF,$5F,$FD,$A1,$1F,$EC,$E9,$DC,$E7,$D7
	.dc.b	$E6,$D8,$CB,$FE,$DF,$F6,$72,$D8,$CB,$FD,$63,$D9,$FF,$71,$DF,$FB
	.dc.b	$AA,$B7,$3F,$EB,$6D,$6B,$FE,$A8,$6E,$FF,$EF,$7A,$FF,$5E,$9F,$EC
	.dc.b	$62,$4F,$FA,$71,$AF,$C8,$7E,$69,$1F,$9A,$FF,$72,$F6,$9F,$17,$FD
	.dc.b	$B2,$30,$D9,$FF,$1F,$4E,$9F,$58,$71,$3F,$DF,$3D,$C7,$BB,$C6,$BF
	.dc.b	$D8,$51,$87,$8B,$67,$84,$EE,$7E,$EE,$95,$70,$FE,$45,$F0,$B6,$BF
	.dc.b	$DE,$67,$F8,$33,$DF,$7B,$FC,$5B,$5E,$86,$CE,$E7,$B7,$C0,$7F,$B5
	.dc.b	$E3,$FC,$2A,$DA,$FF,$2F,$35,$FF,$6E,$35,$FF,$09,$FC,$0F,$AD,$FF
	.dc.b	$32,$35,$FF,$71,$35,$DE,$5E,$0F,$6A,$47,$CC,$AF,$F6,$C5,$F8,$A7
	.dc.b	$E1,$7C,$17,$BE,$BF,$CE,$0D,$7F,$DD,$CD,$7F,$C4,$AD,$5F,$06,$DB
	.dc.b	$E6,$36,$3E,$DB,$6F,$FC,$6A,$D8,$FF,$3F,$3F,$E8,$17,$7E,$97,$22
	.dc.b	$3F,$84,$B7,$C2,$7A,$09,$19,$7E,$E7,$C2,$FD,$28,$FF,$45,$35,$FF
	.dc.b	$1E,$87,$0F,$F9,$01,$AB,$E1,$9F,$0F,$FD,$19,$BC,$7F,$D4,$DB,$E0
	.dc.b	$FF,$94,$1B,$5F,$E9,$96,$BE,$06,$BF,$E4,$E6,$BD,$EB,$B1,$2B,$B3
	.dc.b	$6B,$B3,$67,$B3,$5E,$CD,$7B,$35,$ED,$3D,$09,$3D,$1B,$1D,$1A,$F4
	.dc.b	$6B,$D1,$AF,$48,$E4,$47,$93,$5E,$4D,$79,$35,$E4,$70,$21,$C0,$A7
	.dc.b	$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$57,$D4,$25,$7C,$96,$D7,$C0
	.dc.b	$6D,$7A,$5B,$5E,$46,$BE,$16,$BE,$06,$BE,$0A,$EC,$4A,$EC,$DA,$EC
	.dc.b	$D9,$EC,$D7,$B3,$5E,$CD,$7B,$4F,$42,$4F,$46,$C7,$46,$BD,$1A,$F4
	.dc.b	$6B,$D2,$39,$11,$E4,$D7,$93,$5F,$F5,$61,$FE,$FF,$F8,$0F,$F5,$BF
	.dc.b	$C0,$7F,$AE,$A3,$FD,$54,$D8,$CC,$7E,$BF,$E2,$7F,$E0,$03,$FA,$B6
	.dc.b	$BF,$B9,$6E,$26,$93,$0E,$D1,$96,$C6,$67,$B4,$60,$8E,$08,$7E,$FE
	.dc.b	$38,$36,$36,$D3,$67,$F0,$38,$DB,$63,$6D,$36,$7B,$5C,$B4,$C9,$EF
	.dc.b	$72,$D3,$33,$ED,$6C,$F8,$1E,$47,$94,$87,$84,$A6,$8C,$10,$C1,$0F
	.dc.b	$01,$C0,$A6,$CA,$6E,$3C,$8E,$DA,$ED,$AE,$E3,$D0,$F2,$24,$69,$B1
	.dc.b	$A8,$C1,$1C,$12,$7E,$3C,$65,$B1,$99,$F8,$F1,$82,$38,$21,$E9,$8E
	.dc.b	$0D,$8D,$B4,$D9,$ED,$8D,$B6,$36,$D3,$67,$C1,$72,$D3,$27,$C0,$72
	.dc.b	$D3,$33,$ED,$6C,$F8,$1E,$47,$94,$FB,$1F,$34,$F7,$50,$C7,$FB,$4B
	.dc.b	$89,$D3,$93,$2E,$23,$70,$8C,$3F,$E0,$EF,$B8,$EB,$3A,$DD,$18,$9D
	.dc.b	$1D,$4F,$F0,$9B,$97,$FB,$DD,$3E,$A7,$11,$E5,$6B,$E1,$6C,$6E,$3A
	.dc.b	$1F,$E1,$CF,$98,$DD,$9A,$48,$D9,$EB,$FC,$3B,$99,$1C,$D2,$3C,$3D
	.dc.b	$5F,$E0,$A9,$E2,$DC,$BE,$2B,$63,$FC,$7B,$C3,$E6,$FB,$B9,$1D,$1B
	.dc.b	$5F,$33,$07,$C0,$FB,$09,$5E,$27,$C3,$F8,$2C,$8F,$0B,$6B,$97,$C1
	.dc.b	$7C,$42,$4F,$A9,$F8,$FF,$85,$48,$F6,$B6,$7C,$5D,$BF,$30,$93,$F0
	.dc.b	$9F,$A3,$F1,$FF,$80,$8F,$F1,$8B,$5F,$C6,$B5,$37,$D2,$7A,$4A,$FA
	.dc.b	$31,$86,$C6,$6F,$F1,$F3,$97,$E3,$5F,$FB,$35,$BD,$1F,$5A,$30,$7E
	.dc.b	$60,$A7,$FA,$B6,$3F,$D5,$51,$F8,$0F,$2D,$E5,$EA,$BF,$D6,$0D,$50
	.dc.b	$FA,$EC,$36,$B9,$46,$0E,$1C,$6D,$FF,$5F,$F2,$1F,$84,$6E,$1F,$E8
	.dc.b	$06,$BF,$E8,$06,$BF,$B3,$D6,$BF,$AF,$6F,$17,$C9,$85,$7D,$86,$9F
	.dc.b	$BC,$DF,$F7,$71,$FB,$86,$C6,$7A,$BA,$C6,$23,$FC,$D7,$BF,$D3,$AF
	.dc.b	$A3,$C1,$F5,$DB,$8F,$6A,$DC,$E5,$FE,$17,$F2,$0F,$8C,$FE,$1F,$EE
	.dc.b	$FF,$08,$8C,$4A,$31,$3A,$8E,$2D,$7A,$DE,$DF,$E2,$F8,$FE,$1E,$BC
	.dc.b	$64,$8E,$23,$E4,$3A,$B4,$CB,$83,$D6,$47,$AE,$3F,$95,$8C,$47,$88
	.dc.b	$F8,$24,$37,$7F,$E1,$DB,$FC,$D4,$64,$E0,$68,$FE,$5F,$C2,$F6,$DF
	.dc.b	$4B,$FE,$20,$8C,$36,$33,$1F,$CC,$0C,$9F,$CC,$BE,$11,$F2,$0F,$9B
	.dc.b	$FF,$17,$46,$1B,$19,$C3,$83,$5F,$31,$FD,$13,$1F,$4B,$E7,$21,$EA
	.dc.b	$21,$BD,$3F,$24,$E8,$C6,$7F,$0E,$33,$1C,$04,$F6,$70,$D7,$89,$E4
	.dc.b	$C3,$DB,$F8,$0F,$FB,$5A,$3F,$D7,$F1,$85,$E1,$7F,$5A,$8C,$37,$15
	.dc.b	$71,$AF,$EC,$39,$FD,$1F,$9A,$F9,$C4,$7E,$76,$A8,$FF,$28,$E5,$E5
	.dc.b	$AF,$4E,$7E,$C7,$CE,$7C,$13,$3A,$7C,$12,$3F,$3B,$F1,$C7,$E3,$8C
	.dc.b	$BF,$B9,$F0,$DA,$FE,$EB,$5A,$9C,$9F,$ED,$E7,$11,$F9,$D8,$C5,$71
	.dc.b	$AE,$53,$E4,$3E,$6E,$1E,$77,$CA,$7E,$00,$A7,$A7,$1F,$A5,$87,$6C
	.dc.b	$79,$B7,$1E,$E7,$F8,$2A,$33,$8F,$A5,$F0,$DF,$43,$C2,$FC,$27,$FC
	.dc.b	$1F,$F8,$C3,$F1,$7F,$85,$EE,$70,$8C,$60,$8C,$4E,$A3,$8B,$6B,$ED
	.dc.b	$EC,$F8,$CF,$95,$5F,$E1,$E9,$F9,$D3,$98,$FC,$0C,$70,$9D,$B6,$32
	.dc.b	$B8,$FF,$5B,$F8,$5F,$39,$FC,$8F,$1A,$31,$5F,$E2,$CA,$DB,$C9,$AF
	.dc.b	$C4,$DF,$F1,$44,$E0,$FA,$8F,$E8,$70,$FE,$09,$8F,$92,$BD,$2D,$8C
	.dc.b	$9E,$62,$C7,$5A,$E6,$4D,$BE,$84,$7D,$23,$F0,$17,$F9,$BF,$3F,$FD
	.dc.b	$63,$F3,$0F,$47,$E4,$46,$1B,$D7,$FE,$43,$9D,$5F,$F9,$2A,$B4,$D9
	.dc.b	$D5,$70,$F1,$BF,$C9,$F7,$C2,$BF,$D8,$21,$C5,$F2,$8F,$D4,$D7,$F4
	.dc.b	$E2,$3F,$CB,$9A,$FF,$44,$6B,$FC,$41,$FE,$5B,$C1,$F7,$DF,$D1,$E9
	.dc.b	$F2,$1B,$19,$8F,$B8,$D9,$E8,$77,$21,$83,$EF,$70,$7E,$05,$F9,$13
	.dc.b	$F8,$4E,$0F,$F3,$6D,$FE,$CB,$83,$F0,$6D,$BE,$0F,$8B,$83,$F8,$13
	.dc.b	$6F,$9E,$1F,$93,$8C,$1F,$A2,$87,$C9,$D2,$3F,$A4,$32,$7A,$27,$E2
	.dc.b	$14,$F9,$A7,$EE,$87,$D8,$FE,$EA,$3F,$D0,$AD,$8C,$C6,$E3,$F1,$DC
	.dc.b	$3F,$6E,$7A,$9F,$14,$A4,$3B,$99,$28,$E7,$FE,$37,$F1,$1D,$C9,$7E
	.dc.b	$57,$BE,$3F,$7A,$B8,$93,$8F,$FC,$91,$19,$6C,$7C,$1A,$FE,$7C,$79
	.dc.b	$CB,$3F,$86,$EA,$FA,$33,$F8,$26,$BF,$7B,$AF,$FD,$4F,$F5,$1F,$D2
	.dc.b	$DB,$1F,$C0,$9A,$EC,$47,$8F,$9F,$FC,$BD,$77,$4F,$91,$5E,$99,$CB
	.dc.b	$F5,$72,$F4,$BF,$1D,$AF,$CB,$6B,$C6,$79,$C6,$0D,$F0,$FF,$80,$67
	.dc.b	$EB,$36,$FF,$B2,$77,$7E,$5A,$7F,$8D,$B5,$FE,$57,$DF,$F0,$9F,$E5
	.dc.b	$2D,$D7,$FC,$23,$5F,$AE,$8D,$4F,$8E,$3D,$C4,$7D,$99,$7B,$9F,$F1
	.dc.b	$DE,$DF,$E4,$A2,$3F,$E2,$A7,$FC,$6E,$47,$FC,$76,$D7,$FC,$87,$F1
	.dc.b	$3F,$E5,$22,$3F,$E3,$27,$FC,$7E,$47,$FC,$86,$D7,$FC,$97,$F1,$1F
	.dc.b	$E5,$A2,$3F,$E3,$A7,$FC,$CD,$E3,$7F,$B9,$9B,$06,$91,$FF,$72,$ED
	.dc.b	$B6,$9B,$2C,$7F,$9D,$4F,$F4,$13,$93,$FD,$00,$53,$FD,$EB,$E4,$FF
	.dc.b	$A2,$5C,$07,$11,$FE,$F2,$F2,$FF,$D2,$5E,$E5,$F3,$3F,$C1,$8D,$DC
	.dc.b	$D3,$FD,$42,$4C,$3F,$E0,$0D,$7F,$A8,$A7,$71,$8F,$37,$9B,$FE,$AB
	.dc.b	$8C,$BF,$EF,$63,$25,$72,$87,$99,$FE,$30,$BF,$F4,$49,$23,$FC,$D2
	.dc.b	$D7,$FD,$97,$A7,$18,$FA,$24,$8F,$F4,$C1,$D8,$FF,$52,$16,$7F,$D4
	.dc.b	$91,$FE,$A3,$24,$7F,$80,$CF,$1F,$9D,$D5,$C4,$A3,$11,$FE,$09,$8C
	.dc.b	$3D,$63,$2F,$B6,$1E,$7E,$23,$DE,$D8,$CC,$79,$63,$5A,$FD,$DB,$FC
	.dc.b	$19,$27,$FD,$88,$7A,$FD,$0E,$44,$70,$7B,$FD,$1E,$C4,$70,$7C,$5A
	.dc.b	$7A,$5E,$02,$46,$5F,$37,$A7,$23,$E5,$37,$0F,$91,$1F,$E6,$EF,$53
	.dc.b	$DD,$E7,$9A,$48,$FA,$5E,$AF,$C1,$23,$83,$EC,$4B,$D6,$E6,$48,$CB
	.dc.b	$F2,$78,$24,$7D,$D6,$E1,$E4,$F6,$1E,$C7,$F1,$26,$F1,$FF,$2E,$6F
	.dc.b	$66,$E7,$44,$9D,$36,$35,$18,$23,$82,$61,$F8,$BA,$EA,$DA,$E4,$DA
	.dc.b	$E2,$DA,$E2,$FF,$95,$6B,$8B,$6B,$8B,$6B,$8B,$6B,$8B,$FA,$DA,$E2
	.dc.b	$DA,$E2,$DA,$E2,$DA,$E2,$AB,$8B,$6B,$8B,$6B,$8B,$6B,$8D,$70,$25
	.dc.b	$70,$6D,$70,$6C,$F0,$9D,$12,$74,$D8,$D4,$60,$8E,$09,$5F,$2C,$87
	.dc.b	$C5,$29,$EF,$29,$E9,$2D,$79,$1B,$5E,$16,$D7,$81,$B5,$E0,$7B,$10
	.dc.b	$EC,$53,$B1,$6B,$B3,$6B,$B3,$6B,$B3,$6B,$B3,$D0,$87,$42,$D7,$46
	.dc.b	$D7,$46,$D7,$46,$D7,$47,$91,$2B,$93,$6B,$93,$6B,$93,$6B,$F8,$1F
	.dc.b	$96,$F3,$9E,$77,$F0,$6A,$FF,$3E,$9F,$CB,$E7,$51,$96,$C7,$F0,$CF
	.dc.b	$9C,$FE,$15,$87,$16,$C6,$78,$3F,$D9,$6F,$F4,$B8,$CB,$FC,$ED,$AF
	.dc.b	$FA,$FD,$AE,$E3,$8A,$8C,$B6,$33,$18,$23,$82,$4F,$95,$B3,$E5,$8C
	.dc.b	$11,$C1,$0F,$34,$70,$6C,$6D,$A6,$CF,$54,$6D,$B1,$B6,$9B,$3D,$CE
	.dc.b	$5A,$64,$F7,$39,$69,$99,$F7,$C6,$5B,$19,$9F,$13,$D0,$F3,$90,$F1
	.dc.b	$94,$D1,$82,$18,$21,$E2,$38,$14,$D9,$4D,$C7,$91,$DB,$5D,$B5,$DC
	.dc.b	$7A,$1E,$44,$8D,$36,$35,$18,$23,$82,$4F,$C5,$6C,$FC,$58,$C1,$1C
	.dc.b	$10,$F3,$C7,$06,$C6,$DA,$6C,$F5,$C6,$DB,$1B,$69,$B3,$DE,$E5,$A6
	.dc.b	$4F,$73,$96,$99,$9F,$7C,$65,$B1,$C7,$DD,$7B,$BF,$E3,$BC,$7F,$C0
	.dc.b	$71,$89,$FE,$02,$7F,$8B,$5C,$46,$A7,$84,$E7,$DE,$AF,$F9,$14,$52
	.dc.b	$33,$1F,$E4,$98,$C3,$D8,$E0,$E2,$BF,$C2,$BC,$4C,$78,$D2,$30,$F1
	.dc.b	$3F,$CA,$8E,$27,$8C,$F0,$F9,$E7,$AB,$49,$7C,$44,$69,$C7,$43,$E0
	.dc.b	$7C,$2D,$8C,$C6,$0F,$2F,$C0,$FC,$1C,$7F,$C5,$AD,$9F,$F2,$B3,$5F
	.dc.b	$8C,$D7,$FC,$D9,$F0,$6F,$83,$F7,$B6,$3C,$CD,$AF,$F1,$CB,$63,$FC
	.dc.b	$C5,$F0,$9F,$05,$BA,$7C,$88,$C4,$7D,$56,$BF,$E7,$BD,$FD,$97,$FE
	.dc.b	$44,$6C,$7F,$9B,$5A,$FD,$D6,$BF,$E8,$3F,$85,$7C,$37,$D6,$6C,$7C
	.dc.b	$D6,$D7,$F9,$41,$B1,$FE,$76,$F5,$BE,$EF,$88,$69,$23,$EB,$7A,$2F
	.dc.b	$39,$E6,$61,$B8,$FF,$97,$27,$53,$93,$EF,$7C,$3F,$E0,$B6,$BC,$BF
	.dc.b	$8D,$BF,$F3,$1C,$E5,$B1,$9C,$7F,$D1,$DF,$12,$F8,$9C,$47,$E7,$E3
	.dc.b	$18,$7E,$2D,$B7,$FE,$6E,$6C,$FF,$AD,$34,$E2,$FF,$AA,$FC,$DF,$81
	.dc.b	$7F,$93,$6D,$FF,$B0,$EB,$75,$FA,$BE,$6F,$F5,$97,$C2,$FD,$76,$C7
	.dc.b	$D7,$6B,$F5,$DA,$FD,$7E,$2A,$B8,$BE,$BD,$7D,$59,$CC,$62,$7E,$B0
	.dc.b	$9F,$F4,$4F,$17,$D7,$8C,$1A,$8C,$E1,$DB,$A1,$E2,$FF,$B2,$7E,$13
	.dc.b	$FD,$23,$18,$B4,$62,$75,$1F,$EB,$D9,$D4,$F4,$8D,$9E,$47,$04,$7F
	.dc.b	$D8,$99,$BE,$2F,$F8,$57,$B7,$FE,$A0,$8C,$70,$23,$13,$AA,$E2,$D7
	.dc.b	$F8,$E7,$C3,$3E,$87,$11,$FE,$E3,$8C,$5F,$38,$FF,$6A,$46,$2F,$FD
	.dc.b	$5E,$D8,$FF,$6C,$74,$7E,$59,$FC,$C7,$4F,$E7,$1B,$D3,$FE,$31,$9D
	.dc.b	$47,$F2,$DF,$69,$F0,$0F,$84,$FF,$7A,$7E,$23,$FE,$E3,$9C,$C6,$2B
	.dc.b	$85,$6D,$56,$E3,$0F,$FB,$AF,$47,$C2,$3E,$13,$FC,$6F,$18,$6C,$66
	.dc.b	$3F,$D9,$66,$4F,$F6,$6E,$3E,$6E,$DF,$E6,$B8,$F8,$9A,$84,$6D,$FE
	.dc.b	$7C,$6C,$A7,$F4,$DC,$FE,$1B,$75,$F9,$0F,$CC,$6B,$9D,$7C,$7A,$7F
	.dc.b	$2B,$F1,$BE,$07,$89,$E2,$AF,$0C,$E5,$F8,$F1,$95,$87,$81,$F4,$35
	.dc.b	$CE,$1E,$07,$2D,$C7,$EB,$FD,$3F,$D3,$9F,$81,$18,$7E,$99,$F5,$A3
	.dc.b	$4F,$91,$B1,$F8,$0F,$A1,$F9,$1F,$33,$F6,$5B,$5F,$79,$AF,$6C,$91
	.dc.b	$88,$FE,$D9,$BE,$37,$7D,$A7,$E0,$63,$0D,$AF,$EE,$4D,$9D,$D7,$CF
	.dc.b	$DF,$F2,$DA,$7C,$18,$F8,$71,$83,$F0,$B8,$3E,$8F,$9C,$FF,$81,$EF
	.dc.b	$F1,$F1,$88,$FC,$24,$60,$EE,$E2,$79,$46,$23,$FC,$0D,$F1,$3C,$3F
	.dc.b	$74,$FF,$02,$C6,$0F,$F0,$53,$89,$4E,$DF,$13,$67,$F1,$07,$EC,$88
	.dc.b	$7B,$CF,$F0,$F1,$FE,$0C,$FA,$17,$CC,$FE,$CA,$BF,$C4,$5E,$4A,$73
	.dc.b	$5F,$B0,$9C,$C6,$3A,$3C,$F5,$BB,$E5,$7F,$67,$3F,$D3,$7C,$7F,$B3
	.dc.b	$8F,$F8,$0C,$63,$2F,$DE,$B6,$B8,$0A,$F8,$6F,$D4,$3D,$67,$B8,$F6
	.dc.b	$1F,$C3,$7E,$8D,$EF,$7E,$D7,$F2,$CF,$04,$62,$51,$89,$D6,$1C,$5B
	.dc.b	$3F,$9E,$FC,$11,$F6,$9F,$E3,$6A,$FE,$35,$18,$9F,$F1,$E4,$62,$75
	.dc.b	$1C,$5A,$FD,$EF,$86,$FC,$9C,$FD,$86,$C7,$E2,$38,$51,$88,$E1,$3F
	.dc.b	$9E,$3D,$EF,$C1,$21,$FB,$8E,$4F,$E3,$7F,$82,$BE,$8B,$95,$23,$0F
	.dc.b	$3A,$FC,$BC,$63,$A3,$F1,$CD,$9E,$F3,$D4,$FA,$25,$32,$7D,$52,$F6
	.dc.b	$FF,$0D,$F7,$7F,$DD,$BF,$28,$F8,$0F,$F3,$74,$61,$BF,$01,$FE,$78
	.dc.b	$9D,$5F,$F2,$DA,$D3,$67,$55,$C3,$D4,$FE,$65,$7C,$29,$7C,$1B,$7C
	.dc.b	$1F,$8D,$7F,$D5,$88,$FF,$03,$6B,$FC,$C5,$AF,$03,$F8,$06,$5F,$2B
	.dc.b	$D2,$F8,$FA,$FF,$9F,$DB,$A1,$A7,$E6,$B4,$FF,$1E,$E1,$D3,$0F,$E1
	.dc.b	$EF,$6A,$FD,$EE,$FC,$DF,$C3,$1C,$CB,$7F,$C0,$F8,$3F,$0F,$19,$BF
	.dc.b	$C8,$0F,$91,$FD,$87,$1F,$E5,$DB,$C7,$FD,$65,$B3,$BA,$FD,$12,$FC
	.dc.b	$AF,$F4,$CD,$8C,$C6,$0F,$67,$83,$FE,$5C,$D3,$D9,$AF,$BA,$FE,$61
	.dc.b	$FB,$E1,$F9,$37,$F0,$3F,$13,$F9,$64,$63,$1D,$4E,$7B,$3F,$B7,$3F
	.dc.b	$C3,$BC,$1F,$E2,$7F,$88,$DF,$03,$FB,$93,$70,$34,$F2,$F5,$FE,$16
	.dc.b	$3E,$D1,$F7,$4F,$29,$FB,$13,$FD,$71,$E6,$FC,$D6,$E5,$FC,$B9,$A6
	.dc.b	$D4,$7F,$35,$F1,$7F,$9E,$F2,$FF,$53,$D3,$F0,$97,$FD,$1E,$72,$D8
	.dc.b	$CB,$E3,$21,$E6,$25,$FF,$87,$3C,$6D,$C7,$F6,$26,$C6,$4F,$C8,$47
	.dc.b	$78,$DF,$87,$FD,$C6,$1F,$BF,$7F,$71,$69,$FD,$C4,$A7,$F7,$19,$FF
	.dc.b	$0F,$47,$F8,$BF,$C9,$F9,$F3,$FD,$FC,$F9,$E5,$D7,$4E,$0F,$59,$F9
	.dc.b	$E3,$E7,$9E,$DB,$FF,$21,$F9,$5F,$93,$8F,$F0,$C9,$F9,$92,$F3,$FF
	.dc.b	$88,$1C,$1D,$39,$3F,$1C,$DE,$6F,$C9,$3F,$E3,$19,$FF,$23,$79,$9D
	.dc.b	$63,$FC,$18,$D7,$FC,$52,$DD,$7F,$C8,$5C,$FF,$E0,$BF,$37,$FD,$F4
	.dc.b	$D8,$FC,$51,$AA,$FF,$2A,$F9,$DF,$11,$FF,$0C,$35,$FF,$1A,$B7,$4F
	.dc.b	$F2,$EC,$7C,$53,$EF,$1F,$14,$F5,$72,$FE,$5D,$FC,$6F,$A0,$BF,$E7
	.dc.b	$51,$FE,$76,$8C,$78,$3F,$E7,$66,$C6,$67,$6F,$F9,$5C,$7A,$5F,$E8
	.dc.b	$AA,$FF,$44,$CE,$A7,$FC,$B9,$39,$7F,$D1,$71,$97,$F8,$E7,$AD,$FE
	.dc.b	$61,$F9,$2F,$F4,$E1,$23,$FD,$5E,$FF,$A8,$D3,$48,$FF,$A9,$7D,$7E
	.dc.b	$4F,$FA,$DB,$E4,$BF,$D4,$C4,$70,$7F,$86,$3D,$AF,$F7,$AC,$7F,$AF
	.dc.b	$E0,$D2,$3F,$EC,$B3,$25,$7F,$DA,$0E,$0F,$F6,$71,$4F,$F6,$E5,$FC
	.dc.b	$FF,$8A,$FF,$6D,$62,$69,$27,$FD,$DB,$C5,$B6,$CE,$DB,$3F,$EE,$B9
	.dc.b	$D4,$7D,$3F,$7F,$FD,$F9,$3F,$EF,$8A,$E1,$39,$F4,$F6,$FF,$A3,$EB
	.dc.b	$6F,$F9,$66,$BF,$D5,$E4,$70,$7F,$B8,$1F,$F6,$81,$1C,$1F,$2F,$DD
	.dc.b	$FF,$4D,$47,$CB,$24,$64,$F1,$9F,$E9,$FF,$80,$FC,$4B,$83,$FD,$F8
	.dc.b	$57,$31,$F3,$8F,$F2,$7C,$3E,$07,$F3,$4F,$16,$C6,$9B,$1F,$E5,$66
	.dc.b	$BF,$99,$F8,$27,$67,$A2,$3F,$CB,$71,$87,$C8,$DA,$F0,$B6,$3F,$CC
	.dc.b	$3F,$07,$FA,$39,$F2,$B6,$32,$75,$3F,$50,$E5,$FD,$57,$C0,$7B,$9F
	.dc.b	$F7,$F7,$26,$DF,$F6,$24,$65,$DC,$65,$7C,$23,$6C,$36,$3E,$63,$70
	.dc.b	$F9,$2D,$8F,$F3,$CF,$BD,$F1,$1F,$F2,$7C,$FC,$E8,$C4,$7F,$95,$0F
	.dc.b	$AD,$F0,$AD,$BC,$4D,$8C,$C6,$0C,$FC,$34,$8E,$2D,$BF,$B7,$F0,$EF
	.dc.b	$88,$FE,$4B,$5F,$E8,$C8,$C7,$27,$F1,$96,$F8,$1C,$2B,$75,$FC,$4E
	.dc.b	$72,$F1,$8C,$BF,$EE,$D8,$CB,$B8,$CA,$8C,$BB,$8C,$C6,$23,$51,$82
	.dc.b	$57,$95,$B5,$E5,$8C,$11,$C1,$27,$CD,$19,$6C,$65,$F3,$46,$5B,$19
	.dc.b	$7C,$D1,$96,$C6,$5F,$3C,$65,$B1,$9A,$F6,$12,$BD,$33,$E2,$21,$E0
	.dc.b	$29,$D8,$A7,$42,$C7,$3A,$F0,$36,$BC,$15,$DA,$BA,$10,$E8,$53,$A1
	.dc.b	$67,$93,$DA,$39,$3C,$08,$70,$2C,$70,$79,$47,$07,$44,$8D,$3C,$23
	.dc.b	$51,$88,$D4,$60,$95,$EE,$6D,$7B,$A3,$04,$70,$49,$F5,$46,$5B,$19
	.dc.b	$7C,$F1,$96,$C6,$5F,$34,$65,$B1,$97,$CF,$19,$6C,$66,$BD,$84,$AF
	.dc.b	$4C,$F8,$88,$78,$0A,$76,$3F,$03,$EC,$3D,$87,$B5,$FC,$DA,$7F,$D4
	.dc.b	$ED,$9D,$36,$77,$3F,$C0,$BD,$AF,$F7,$AB,$FB,$76,$BF,$B5,$6E,$06
	.dc.b	$93,$2F,$04,$65,$B1,$99,$ED,$18,$23,$82,$1F,$C4,$63,$83,$63,$6D
	.dc.b	$36,$7F,$11,$8D,$B6,$36,$D3,$67,$B5,$CB,$4C,$9E,$F7,$2D,$33,$3E
	.dc.b	$D6,$CF,$81,$E4,$79,$48,$78,$4A,$68,$C1,$0C,$10,$F0,$1C,$0A,$6C
	.dc.b	$A6,$E3,$C8,$ED,$AE,$DA,$EE,$3D,$0F,$22,$46,$9B,$1A,$8C,$11,$C1
	.dc.b	$27,$E3,$C6,$5B,$19,$9F,$8F,$18,$23,$82,$1E,$98,$E0,$D8,$DB,$4D
	.dc.b	$9E,$D8,$DB,$63,$6D,$36,$7C,$17,$2D,$32,$7C,$07,$2D,$33,$3E,$D6
	.dc.b	$CF,$81,$E4,$79,$48,$78,$4A,$68,$C1,$0C,$10,$FA,$9E,$7B,$E3,$11
	.dc.b	$FB,$83,$FD,$05,$19,$70,$69,$C6,$3F,$E3,$86,$C7,$F8,$43,$3E,$73
	.dc.b	$FE,$0F,$6C,$64,$E6,$38,$7F,$C0,$CE,$3C,$B7,$D5,$F0,$8F,$F4,$A4
	.dc.b	$61,$D4,$65,$FF,$7E,$B5,$73,$BE,$3B,$0D,$8F,$0B,$71,$FF,$28,$B6
	.dc.b	$3F,$D4,$43,$27,$C7,$F5,$6C,$7A,$9B,$5F,$0E,$75,$F2,$12,$B4,$D9
	.dc.b	$EF,$F3,$BE,$42,$4F,$91,$ED,$F2,$52,$3B,$36,$7E,$87,$C0,$78,$C9
	.dc.b	$3E,$37,$E1,$78,$EF,$0E,$92,$7E,$8C,$65,$FA,$7F,$3A,$F2,$7E,$53
	.dc.b	$63,$31,$83,$E7,$7A,$6F,$4D,$E7,$FD,$C8,$C3,$63,$39,$FF,$9F,$A7
	.dc.b	$2F,$FA,$03,$E3,$DF,$29,$FE,$D9,$8C,$37,$1F,$A7,$3A,$EC,$FC,$4B
	.dc.b	$6B,$FD,$51,$1F,$8C,$F9,$17,$9B,$C6,$FF,$DB,$91,$89,$D4,$E6,$7F
	.dc.b	$39,$C3,$FA,$2F,$91,$FC,$E3,$63,$F0,$7C,$C8,$C4,$7E,$99,$AF,$E9
	.dc.b	$78,$EB,$F8,$96,$C9,$A6,$B8,$BE,$95,$FF,$B5,$C9,$5F,$EA,$D7,$E1
	.dc.b	$E2,$69,$23,$E2,$EB,$FF,$00,$63,$FD,$C1,$F2,$32,$7F,$D6,$CD,$7F
	.dc.b	$7F,$4F,$98,$FC,$DF,$C8,$FE,$82,$31,$D3,$FC,$12,$31,$3A,$AF,$E0
	.dc.b	$E2,$B6,$A7,$FD,$77,$F0,$0F,$98,$D4,$61,$B3,$FA,$E9,$D5,$71,$8C
	.dc.b	$4E,$9E,$FC,$9F,$8C,$7F,$15,$CB,$F9,$06,$F3,$7F,$85,$67,$51,$E6
	.dc.b	$BF,$F7,$A7,$CD,$79,$FE,$33,$DD,$E5,$23,$15,$B8,$C5,$2B,$75,$AA
	.dc.b	$F9,$3F,$17,$F8,$3A,$FF,$6C,$87,$11,$F2,$DA,$F8,$DA,$FF,$30,$F7
	.dc.b	$1F,$69,$F3,$3C,$6F,$E2,$31,$89,$FE,$6B,$18,$9D,$47,$D2,$9D,$3F
	.dc.b	$CA,$CF,$E3,$39,$FC,$27,$E4,$EB,$F1,$DC,$B5,$CE,$BF,$EF,$2A,$7C
	.dc.b	$83,$C3,$F0,$D7,$78,$C3,$DC,$F5,$38,$CB,$D2,$D8,$EF,$7F,$57,$0F
	.dc.b	$04,$7F,$50,$C7,$CC,$D8,$CE,$8D,$DF,$01,$F4,$6B,$F2,$11,$89,$FA
	.dc.b	$D1,$F6,$67,$51,$C5,$BB,$7F,$3A,$87,$E2,$3F,$57,$9B,$F3,$4D,$8C
	.dc.b	$DF,$D7,$9C,$BF,$67,$E4,$39,$92,$7C,$31,$F8,$48,$F3,$2C,$FD,$AD
	.dc.b	$CF,$E0,$B7,$4F,$88,$7E,$2B,$47,$B5,$EA,$6E,$3D,$AB,$E0,$CE,$4F
	.dc.b	$7F,$C0,$7E,$F3,$F0,$9F,$DA,$4F,$EF,$A3,$15,$FA,$36,$C6,$4F,$DD
	.dc.b	$4F,$07,$E9,$61,$E4,$7F,$5A,$43,$F8,$0F,$B3,$F9,$8F,$98,$FD,$66
	.dc.b	$18,$E6,$46,$2B,$FC,$1B,$5B,$AF,$D6,$1B,$9D,$4B,$99,$F2,$2F,$9E
	.dc.b	$FF,$0A,$FE,$2D,$E5,$9C,$C6,$30,$FD,$15,$6E,$D7,$B8,$C3,$ED,$8F
	.dc.b	$A0,$7A,$CF,$D0,$94,$F1,$43,$C8,$FF,$12,$FE,$25,$CD,$B7,$E0,$6D
	.dc.b	$7E,$EE,$B8,$2A,$F4,$69,$FC,$5F,$AA,$ED,$FB,$CC,$7F,$75,$39,$6C
	.dc.b	$66,$FF,$87,$F9,$9F,$1A,$7F,$87,$37,$A7,$FB,$C4,$7C,$C9,$DB,$5F
	.dc.b	$B6,$7B,$4F,$82,$7E,$D0,$FE,$2D,$F3,$7C,$C9,$1C,$4F,$E3,$8F,$D6
	.dc.b	$8E,$27,$C8,$39,$10,$C1,$0F,$09,$93,$FC,$9F,$ED,$7F,$B2,$1F,$F6
	.dc.b	$07,$B7,$FD,$32,$31,$EC,$A3,$1E,$DF,$FB,$32,$31,$5C,$A3,$13,$AF
	.dc.b	$43,$FC,$B7,$5A,$AF,$F6,$B9,$D8,$E1,$1A,$9E,$11,$93,$F6,$8F,$F4
	.dc.b	$C2,$3F,$C6,$9A,$FF,$28,$6B,$FA,$87,$E5,$E3,$FE,$49,$3F,$A0,$63
	.dc.b	$FD,$9F,$CE,$FD,$74,$7C,$C3,$55,$FD,$32,$75,$1D,$E2,$1D,$CA,$7F
	.dc.b	$71,$CF,$F0,$35,$D1,$B3,$FD,$3F,$4E,$75,$F8,$4A,$FC,$8A,$F8,$AA
	.dc.b	$4A,$FC,$8C,$65,$FC,$F2,$FC,$4B,$F4,$4D,$8C,$C6,$0F,$D7,$77,$FF
	.dc.b	$90,$B3,$F8,$B5,$F3,$0F,$4F,$47,$EE,$97,$E2,$9F,$B7,$E9,$FF,$49
	.dc.b	$B6,$33,$8F,$EF,$A7,$2F,$F0,$93,$E4,$C3,$AF,$F8,$76,$9F,$C6,$DB
	.dc.b	$3B,$8F,$2F,$81,$FE,$58,$CB,$CB,$9F,$9F,$0F,$9B,$6E,$EF,$F1,$3F
	.dc.b	$E1,$7F,$61,$B1,$FC,$AE,$7F,$C4,$53,$96,$C7,$F8,$2F,$C0,$FC,$04
	.dc.b	$58,$FF,$46,$37,$2F,$F0,$9C,$EA,$3E,$F1,$0F,$C6,$1E,$4E,$27,$85
	.dc.b	$B8,$FC,$13,$6F,$FC,$24,$D9,$DC,$78,$27,$FC,$33,$E1,$FF,$8E,$7F
	.dc.b	$13,$FE,$C0,$48,$FE,$C1,$28,$FF,$13,$F8,$DD,$7C,$5F,$EE,$19,$E2
	.dc.b	$BA,$5F,$F7,$4A,$F2,$3F,$7C,$F5,$14,$F5,$9F,$E0,$8F,$23,$FD,$A3
	.dc.b	$CF,$FE,$08,$6C,$7F,$83,$DA,$EC,$FC,$C1,$FA,$23,$A1,$38,$3E,$D3
	.dc.b	$FE,$2D,$EF,$FF,$2B,$D7,$F8,$A6,$39,$9F,$D3,$E3,$2F,$F3,$23,$38
	.dc.b	$7F,$8E,$FD,$4F,$F2,$F7,$1F,$F8,$ED,$B3,$FE,$3B,$6B,$FE,$44,$6B
	.dc.b	$FE,$16,$F2,$BE,$41,$2B,$E3,$9F,$E5,$43,$E4,$C7,$A7,$97,$F2,$8F
	.dc.b	$B7,$2F,$5B,$FE,$F6,$F3,$7F,$CC,$FC,$DF,$E6,$BB,$FC,$B4,$FF,$9A
	.dc.b	$1A,$FF,$9D,$A7,$6F,$F9,$8F,$CF,$FF,$6C,$D7,$F9,$F3,$8F,$FE,$00
	.dc.b	$B6,$78,$9F,$E7,$D9,$DB,$FE,$E4,$F4,$FF,$D9,$1F,$19,$FE,$E6,$84
	.dc.b	$61,$D7,$B2,$53,$B9,$D4,$7F,$A2,$BD,$8F,$F5,$E7,$C6,$7F,$A7,$89
	.dc.b	$1F,$E9,$CD,$11,$C7,$0E,$8D,$9C,$9E,$02,$9F,$E9,$6B,$FF,$6F,$47
	.dc.b	$FB,$A2,$32,$7F,$AD,$CA,$7F,$AD,$CB,$CD,$FE,$B7,$6F,$0F,$81,$FF
	.dc.b	$61,$B5,$FF,$61,$1A,$39,$F9,$9F,$EC,$FE,$8F,$F6,$A9,$23,$FD,$52
	.dc.b	$7F,$83,$7D,$9F,$F2,$DB,$EF,$8F,$F5,$F9,$1C,$1F,$E6,$A7,$FD,$3E
	.dc.b	$47,$07,$4F,$6B,$FD,$F1,$1F,$55,$B1,$9F,$17,$EA,$4E,$5B,$1F,$26
	.dc.b	$3E,$61,$1C,$1F,$EE,$2F,$6B,$6F,$FB,$0B,$2F,$8D,$3F,$34,$91,$F1
	.dc.b	$8F,$F3,$D1,$5F,$F7,$6B,$83,$E4,$BF,$8B,$F7,$1E,$E7,$69,$FC,$54
	.dc.b	$61,$E0,$62,$34,$A3,$53,$F9,$5A,$FF,$4B,$B7,$B7,$71,$8A,$FF,$7E
	.dc.b	$C6,$39,$A4,$76,$12,$8C,$D7,$9F,$DD,$F0,$E0,$69,$2F,$97,$BE,$F7
	.dc.b	$FC,$ED,$8C,$C6,$0F,$5F,$A4,$F8,$0A,$49,$F2,$46,$5F,$36,$6F,$81
	.dc.b	$F6,$92,$32,$FC,$6F,$82,$7A,$5F,$C5,$A5,$18,$EA,$FF,$14,$47,$EE
	.dc.b	$63,$1E,$1F,$EF,$DB,$1F,$9A,$6B,$FB,$F3,$9D,$75,$8C,$C7,$38,$C1
	.dc.b	$2B,$BC,$70,$6C,$F0,$8C,$57,$86,$B8,$36,$B7,$18,$9F,$23,$5F,$25
	.dc.b	$73,$24,$66,$3C,$93,$C8,$95,$D4,$91,$D2,$32,$D8,$CB,$DA,$7A,$3E
	.dc.b	$83,$44,$35,$3C,$9B,$3C,$9F,$1B,$5E,$46,$8E,$46,$8A,$68,$C1,$B2
	.dc.b	$18,$21,$B2,$57,$C4,$9F,$24,$6A,$31,$3E,$36,$C7,$81,$AF,$63,$9D
	.dc.b	$7C,$28,$CC,$73,$8C,$12,$BB,$C7,$06,$CF,$08,$C5,$78,$6B,$83,$6B
	.dc.b	$71,$89,$F2,$35,$F2,$57,$32,$46,$63,$C9,$3C,$89,$5D,$49,$1D,$23
	.dc.b	$2D,$8C,$BD,$A7,$A3,$E8,$34,$43,$53,$C9,$B3,$C9,$F1,$B5,$E4,$68
	.dc.b	$E4,$68,$FC,$0F,$A9,$FE,$AB,$7F,$A6,$B5,$FF,$4E,$98,$F5,$1E,$A7
	.dc.b	$F7,$98,$FE,$03,$F5,$1F,$EE,$57,$F8,$03,$5C,$E1,$BC,$93,$96,$C6
	.dc.b	$63,$04,$70,$49,$FE,$15,$19,$6C,$65,$FE,$19,$19,$6C,$65,$FD,$DC
	.dc.b	$65,$B1,$97,$D2,$D7,$D2,$76,$21,$D0,$A7,$22,$9C,$8A,$72,$38,$10
	.dc.b	$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$27,$E7,$BE,$46,$BD
	.dc.b	$DA,$F4,$6B,$CD,$AF,$39,$F9,$4E,$5A,$E6,$7E,$54,$F0,$24,$E9,$B1
	.dc.b	$A8,$C1,$1C,$12,$7D,$11,$96,$C6,$5F,$54,$65,$B1,$97,$D9,$19,$6C
	.dc.b	$65,$F5,$35,$F5,$1D,$88,$74,$29,$C8,$A7,$22,$9C,$8E,$04,$38,$14
	.dc.b	$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$49,$F9,$EF,$91,$AF,$76,$BD
	.dc.b	$38,$3F,$C1,$D7,$FE,$77,$6C,$7F,$94,$9A,$FF,$79,$6B,$FE,$30,$E1
	.dc.b	$FF,$3D,$4F,$F8,$7E,$3F,$84,$9F,$E5,$AE,$2F,$F1,$03,$83,$FC,$04
	.dc.b	$E2,$3F,$C6,$5C,$9C,$A7,$FC,$C3,$C1,$FE,$2B,$78,$1F,$E8,$77,$13
	.dc.b	$C2,$32,$7F,$8C,$B9,$79,$3F,$E5,$16,$BC,$5A,$EC,$E3,$7F,$E4,$99
	.dc.b	$FF,$27,$73,$6A,$31,$3E,$B8,$E3,$F1,$6D,$69,$27,$C7,$19,$7C,$B8
	.dc.b	$FF,$A1,$F0,$FF,$47,$12,$3F,$CD,$07,$AB,$BD,$E0,$79,$9B,$19,$8C
	.dc.b	$1E,$AF,$19,$E3,$BD,$7F,$87,$18,$6C,$67,$0F,$F5,$A4,$E5,$FF,$5B
	.dc.b	$7C,$63,$D3,$F6,$46,$1B,$8F,$FA,$52,$75,$AF,$D0,$6D,$7F,$AF,$FC
	.dc.b	$67,$C2,$F0,$BF,$F5,$D3,$5F,$F6,$04,$6A,$FE,$CB,$6B,$FD,$91,$E2
	.dc.b	$BC,$5E,$17,$FE,$A3,$BF,$BF,$39,$6C,$EB,$9F,$FA,$7D,$FE,$6F,$A7
	.dc.b	$04,$AF,$87,$BB,$E1,$7E,$1F,$07,$C9,$6C,$67,$3F,$F5,$9F,$07,$CB
	.dc.b	$9F,$89,$79,$9F,$F2,$E6,$FF,$12,$72,$48,$E3,$87,$D5,$E0,$EA,$48
	.dc.b	$CE,$3F,$45,$F9,$D1,$F8,$D3,$B9,$1C,$E1,$B7,$FC,$D9,$8F,$DF,$7D
	.dc.b	$24,$3A,$9F,$EC,$CF,$61,$F0,$3F,$EA,$7F,$0B,$0D,$9F,$F3,$B4,$EB
	.dc.b	$A3,$FD,$ED,$5A,$6C,$FF,$9E,$F0,$7A,$8F,$A3,$FF,$03,$46,$1B,$19
	.dc.b	$8F,$F6,$EC,$FE,$A6,$30,$FF,$B8,$23,$9F,$C7,$7E,$B7,$E1,$31,$48
	.dc.b	$C5,$78,$63,$14,$AD,$D6,$AB,$E2,$EF,$FA,$57,$E2,$18,$DF,$FC,$3D
	.dc.b	$B7,$C9,$8F,$67,$9C,$F8,$8E,$51,$FB,$F8,$C6,$5D,$E3,$73,$A8,$4E
	.dc.b	$B2,$EF,$41,$7D,$DF,$11,$AD,$FE,$AF,$CE,$7C,$E8,$E1,$18,$5C,$1F
	.dc.b	$62,$75,$09,$D4,$FC,$8E,$BF,$E2,$DF,$B8,$DC,$CD,$25,$FE,$03,$D7
	.dc.b	$79,$FD,$31,$46,$27,$C2,$6E,$75,$19,$C7,$DF,$3E,$EE,$0F,$A0,$FE
	.dc.b	$48,$87,$B3,$95,$F1,$7D,$DB,$1D,$8D,$46,$4F,$7E,$6F,$C3,$5F,$59
	.dc.b	$F3,$7F,$10,$FE,$61,$7F,$CD,$67,$2A,$33,$5C,$6B,$83,$F9,$BA,$DC
	.dc.b	$F2,$AF,$57,$2F,$E1,$B2,$F5,$C7,$D8,$3D,$67,$F2,$9F,$C2,$FF,$DE
	.dc.b	$1A,$78,$BA,$D1,$8C,$BF,$DC,$51,$8F,$4B,$F5,$B1,$89,$D5,$FE,$1E
	.dc.b	$1F,$53,$FC,$F9,$7C,$67,$AA,$30,$DC,$BF,$AA,$36,$76,$79,$AB,$D6
	.dc.b	$7D,$6E,$7F,$C9,$3F,$D6,$ED,$E2,$FF,$5D,$FC,$1B,$B3,$6F,$F8,$04
	.dc.b	$E9,$4E,$9A,$7E,$D7,$F0,$AF,$DD,$34,$FD,$B1,$7A,$FE,$9F,$7A,$77
	.dc.b	$8F,$E8,$39,$DE,$DF,$C6,$8C,$35,$FD,$04,$69,$FD,$2E,$1F,$9B,$9D
	.dc.b	$47,$C0,$3E,$16,$9F,$01,$C1,$DF,$CF,$E9,$D7,$FA,$06,$E0,$69,$D2
	.dc.b	$B9,$13,$83,$F0,$B7,$FB,$F8,$C7,$2F,$E2,$49,$1F,$A6,$F5,$1E,$A7
	.dc.b	$F8,$A3,$D6,$7A,$DF,$C1,$E9,$5A,$F2,$FF,$85,$56,$A7,$F8,$64,$E5
	.dc.b	$46,$5E,$37,$D6,$FA,$56,$EB,$4D,$8C,$CE,$EB,$AB,$5F,$D1,$C7,$F6
	.dc.b	$62,$3F,$D6,$9A,$FF,$2D,$6B,$F0,$F2,$FF,$0B,$47,$F5,$3D,$3E,$64
	.dc.b	$BE,$07,$F7,$3B,$7D,$18,$C7,$07,$F1,$5D,$FF,$95,$56,$AF,$8B,$FB
	.dc.b	$7D,$FF,$C4,$18,$F8,$30,$F3,$9E,$53,$F0,$87,$43,$F1,$65,$3F,$17
	.dc.b	$C1,$DB,$1D,$DF,$E5,$B8,$7F,$C8,$1E,$95,$93,$48,$FE,$9B,$9F,$FC
	.dc.b	$8D,$F8,$2F,$E9,$49,$19,$0F,$A7,$7F,$8B,$97,$9C,$F9,$45,$3F,$6C
	.dc.b	$FF,$94,$7E,$1F,$F7,$C4,$AF,$DA,$C6,$5F,$E1,$9D,$9E,$BA,$FE,$A5
	.dc.b	$DB,$FC,$45,$F8,$0D,$D3,$F8,$C7,$6A,$B8,$57,$D9,$E8,$FC,$63,$78
	.dc.b	$3E,$77,$0F,$D0,$8F,$E5,$5E,$0F,$F5,$FF,$C3,$BF,$97,$F8,$5F,$CC
	.dc.b	$9B,$19,$D7,$74,$9F,$C0,$F8,$9F,$D0,$B5,$FF,$0A,$F8,$9F,$80,$AF
	.dc.b	$F0,$AC,$E5,$B1,$97,$E3,$90,$F3,$1F,$27,$A7,$FA,$DF,$8B,$B8,$FE
	.dc.b	$12,$D8,$CC,$7E,$4C,$FE,$BC,$53,$F9,$A7,$7F,$F9,$67,$D5,$FE,$D8
	.dc.b	$48,$FF,$0A,$79,$1F,$E8,$58,$FE,$ED,$1C,$4F,$EE,$D8,$BC,$9F,$03
	.dc.b	$63,$58,$7F,$7F,$CF,$E6,$F3,$7E,$12,$7D,$87,$E7,$0F,$E8,$DE,$57
	.dc.b	$F8,$53,$AF,$FC,$54,$D8,$FF,$0A,$F0,$7E,$59,$B3,$D3,$97,$ED,$C7
	.dc.b	$33,$E3,$9E,$EF,$33,$F3,$EF,$F8,$BD,$AE,$6B,$A3,$6B,$FC,$6F,$E6
	.dc.b	$FF,$A9,$79,$ED,$7E,$20,$FF,$22,$17,$87,$FC,$89,$1F,$E4,$DB,$F8
	.dc.b	$3B,$3C,$EF,$8A,$48,$F9,$47,$9C,$9D,$1F,$91,$7E,$2F,$A1,$FE,$96
	.dc.b	$D3,$FC,$EE,$48,$FF,$34,$BF,$E7,$E2,$3F,$E7,$26,$BF,$E7,$CE,$CF
	.dc.b	$F2,$03,$FE,$70,$DF,$FD,$02,$47,$FC,$F2,$FF,$A2,$08,$E0,$FF,$75
	.dc.b	$7A,$3F,$EC,$7F,$77,$FD,$29,$B1,$A4,$9F,$E8,$07,$FA,$7D,$FF,$51
	.dc.b	$FA,$9C,$63,$3E,$17,$FA,$86,$73,$5C,$E7,$8F,$AB,$FE,$AD,$9F,$F5
	.dc.b	$A7,$57,$19,$FF,$59,$D6,$DB,$3F,$E9,$C2,$3F,$E8,$E6,$BF,$E6,$77
	.dc.b	$FD,$4A,$47,$FD,$32,$FF,$9F,$3A,$BF,$DA,$ED,$F5,$BF,$CC,$0D,$9F
	.dc.b	$01,$FE,$63,$9D,$BF,$EE,$28,$FF,$42,$F3,$7F,$93,$DB,$D9,$CA,$74
	.dc.b	$7C,$6F,$5F,$FD,$DD,$3F,$E8,$C6,$C6,$71,$E0,$FF,$6B,$8C,$BF,$DB
	.dc.b	$7D,$8F,$C7,$3E,$40,$E2,$BF,$D2,$ED,$8F,$AA,$D7,$FB,$B6,$1F,$D6
	.dc.b	$A7,$D3,$E5,$78,$27,$FD,$01,$39,$7E,$EB,$5E,$3E,$CE,$DF,$E9,$6D
	.dc.b	$83,$4D,$5B,$DA,$A4,$AF,$54,$65,$F6,$7B,$4F,$77,$B3,$63,$31,$83
	.dc.b	$C5,$EE,$39,$7F,$75,$3F,$94,$8C,$6D,$FC,$21,$BD,$BF,$EB,$C6,$CE
	.dc.b	$EB,$8B,$FE,$1B,$8C,$B6,$32,$FF,$A9,$63,$2D,$8C,$A8,$CB,$63,$31
	.dc.b	$82,$38,$25,$79,$DB,$5E,$76,$D7,$9E,$74,$49,$D3,$63,$51,$82,$38
	.dc.b	$21,$ED,$8E,$0D,$8D,$B4,$D9,$EF,$8D,$B6,$36,$D3,$67,$C1,$8D,$B6
	.dc.b	$36,$D3,$75,$F1,$27,$6D,$9D,$B6,$77,$5F,$11,$F3,$10,$F1,$14,$EC
	.dc.b	$58,$E4,$D8,$E4,$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D,$8D,$46
	.dc.b	$08,$E0,$95,$F0,$1B,$5F,$01,$B5,$F0,$27,$44,$9D,$36,$35,$18,$23
	.dc.b	$82,$1F,$06,$38,$36,$36,$D3,$67,$C0,$8D,$B6,$36,$D3,$67,$C1,$8D
	.dc.b	$B6,$36,$D3,$75,$F1,$27,$6D,$9D,$B6,$77,$5F,$47,$D1,$FF,$46,$47
	.dc.b	$F5,$46,$BF,$E8,$6F,$4F,$FA,$BB,$FE,$86,$6B,$FB,$7F,$C0,$7F,$DD
	.dc.b	$4F,$EF,$5A,$E7,$1D,$E0,$9C,$B6,$33,$18,$23,$82,$4F,$F1,$18,$CB
	.dc.b	$63,$2F,$F1,$98,$CB,$63,$2F,$F1,$58,$CB,$63,$2F,$A5,$AF,$A4,$EC
	.dc.b	$43,$A1,$4E,$45,$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1
	.dc.b	$4D,$18,$21,$82,$4F,$CF,$7C,$8D,$7B,$B5,$E8,$D7,$9B,$5E,$73,$F2
	.dc.b	$9C,$B5,$CC,$FC,$A9,$E0,$49,$D3,$63,$51,$82,$38,$24,$FA,$23,$2D
	.dc.b	$8C,$BE,$A8,$CB,$63,$2F,$B2,$32,$D8,$CB,$EA,$6B,$EA,$3B,$10,$E8
	.dc.b	$53,$91,$4E,$45,$39,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46
	.dc.b	$08,$60,$93,$F3,$DF,$23,$5E,$ED,$7A,$35,$E7,$F3,$2F,$11,$0C,$3F
	.dc.b	$82,$CE,$63,$07,$F9,$79,$C4,$6A,$78,$BF,$E3,$0F,$99,$B5,$FE,$30
	.dc.b	$6B,$FE,$39,$6B,$E0,$DD,$18,$87,$2F,$FB,$6D,$F0,$9F,$E4,$3F,$3B
	.dc.b	$51,$9B,$FF,$25,$47,$1F,$9A,$7C,$65,$25,$F2,$3D,$F1,$A7,$07,$F9
	.dc.b	$2B,$C5,$F1,$65,$FE,$4E,$C3,$FC,$97,$18,$3C,$DE,$6F,$A2,$B8,$B6
	.dc.b	$7F,$96,$B6,$3F,$D0,$AD,$7F,$CA,$3D,$7E,$FE,$93,$49,$1E,$4F,$84
	.dc.b	$7C,$27,$C5,$6C,$66,$30,$7C,$9F,$21,$E6,$7D,$DC,$BF,$9F,$C6,$38
	.dc.b	$BD,$ED,$8C,$CE,$EF,$FC,$BE,$39,$7E,$2E,$7F,$86,$6C,$66,$BF,$D0
	.dc.b	$2D,$3F,$D3,$7E,$15,$7F,$D3,$31,$FE,$9C,$9F,$9F,$E1,$AB,$55,$F9
	.dc.b	$2F,$69,$E4,$62,$11,$8B,$FA,$ED,$C3,$FA,$DB,$67,$FD,$67,$B7,$C6
	.dc.b	$F8,$3F,$EB,$ED,$8C,$EF,$FA,$09,$CB,$63,$FD,$77,$B7,$C7,$9F,$8B
	.dc.b	$B9,$A4,$9E,$67,$A0,$B9,$F4,$70,$7E,$B8,$61,$F6,$A3,$A9,$23,$C3
	.dc.b	$BF,$ED,$63,$11,$C7,$0F,$B4,$F9,$08,$70,$C3,$FB,$EE,$8F,$D4,$9E
	.dc.b	$DF,$EF,$BE,$07,$FD,$AF,$39,$8C,$5F,$96,$30,$A7,$75,$CA,$30,$6F
	.dc.b	$D0,$7B,$1F,$E0,$8B,$46,$1F,$1B,$67,$C2,$D7,$97,$37,$E3,$FD,$EC
	.dc.b	$B6,$33,$49,$CB,$63,$E0,$1F,$08,$F8,$DB,$7F,$19,$72,$79,$FB,$BA
	.dc.b	$3F,$9B,$C3,$F8,$8D,$73,$8C,$4E,$9F,$D1,$4E,$AF,$B4,$7F,$28,$DB
	.dc.b	$E9,$BD,$8F,$0E,$DF,$4B,$47,$C7,$7C,$E9,$FD,$54,$61,$F9,$66,$23
	.dc.b	$4F,$EC,$63,$5B,$7E,$26,$1E,$B7,$DC,$F2,$BF,$9C,$FE,$5B,$98,$C4
	.dc.b	$A3,$1C,$5F,$ED,$48,$C1,$F9,$39,$DC,$EB,$E3,$FC,$31,$F8,$86,$D1
	.dc.b	$A7,$9A,$9A,$7C,$27,$E0,$9F,$18,$CF,$0F,$D3,$7E,$0F,$DC,$FB,$B8
	.dc.b	$BF,$2C,$DD,$4D,$35,$F9,$5B,$E3,$3F,$AB,$FD,$6F,$F1,$5C,$3F,$8A
	.dc.b	$4E,$54,$65,$FE,$35,$19,$8C,$60,$8C,$3F,$66,$7E,$E6,$3E,$76,$F1
	.dc.b	$7F,$99,$B3,$F7,$47,$F1,$6F,$61,$F8,$D7,$EB,$70,$FD,$F7,$52,$73
	.dc.b	$5F,$AB,$9C,$C6,$39,$3C,$D5,$BB,$E5,$7F,$CB,$FD,$9F,$E4,$D8,$3F
	.dc.b	$19,$F4,$C6,$1B,$E1,$7F,$37,$6C,$EC,$F6,$57,$D2,$D3,$D5,$AF,$B6
	.dc.b	$30,$7E,$D3,$E3,$7F,$A0,$7C,$67,$C6,$EC,$46,$30,$FE,$00,$DA,$FD
	.dc.b	$53,$5F,$E9,$1F,$8E,$7F,$49,$9F,$DC,$1F,$C3,$5C,$76,$FD,$26,$C6
	.dc.b	$4F,$E0,$47,$E5,$7E,$B5,$F5,$7F,$12,$FF,$43,$18,$7E,$61,$FE,$1A
	.dc.b	$EC,$46,$78,$FF,$05,$18,$AE,$55,$F0,$0F,$8B,$CB,$F7,$1C,$1C,$FB
	.dc.b	$1F,$23,$E1,$BF,$E5,$6D,$8C,$CF,$88,$FD,$AB,$8C,$3F,$25,$1D,$0F
	.dc.b	$C5,$9F,$A5,$BF,$01,$38,$3E,$A3,$78,$7F,$D6,$BE,$AB,$E3,$7F,$C6
	.dc.b	$7E,$4B,$D8,$FF,$5E,$63,$FD,$99,$B1,$98,$FF,$5F,$19,$2B,$95,$5D
	.dc.b	$7C,$BF,$EE,$55,$C8,$93,$FA,$27,$FB,$19,$1F,$EA,$0D,$7F,$90,$B5
	.dc.b	$E0,$7F,$0F,$D3,$E3,$FC,$2F,$D4,$6F,$16,$77,$FF,$29,$CE,$5B,$1D
	.dc.b	$AA,$1D,$8A,$7E,$13,$87,$6F,$E0,$F8,$BF,$0D,$7F,$82,$6D,$7E,$36
	.dc.b	$B7,$5F,$98,$8F,$F1,$07,$9D,$F9,$56,$9F,$95,$0E,$5A,$7E,$90,$7C
	.dc.b	$3F,$EA,$79,$FF,$C5,$4D,$8C,$9B,$7F,$0B,$C5,$FE,$35,$D7,$E4,$61
	.dc.b	$FB,$77,$E5,$53,$B3,$F8,$1F,$E3,$3F,$BE,$24,$7F,$79,$8C,$BF,$C3
	.dc.b	$BB,$BD,$BD,$FF,$C3,$70,$6F,$FC,$43,$C0,$FE,$2A,$DA,$F8,$73,$E1
	.dc.b	$3E,$69,$4F,$A0,$7F,$30,$F0,$FF,$D1,$3F,$5D,$FC,$B3,$C3,$FE,$64
	.dc.b	$D8,$CE,$5F,$E0,$D9,$CB,$FE,$0F,$F1,$BF,$9F,$DF,$90,$C7,$8D,$F8
	.dc.b	$26,$99,$3F,$A4,$94,$FC,$69,$E4,$29,$3B,$FF,$AD,$3B,$6E,$7F,$D7
	.dc.b	$E3,$1D,$1F,$D7,$5B,$E4,$FF,$A9,$3F,$27,$FE,$D2,$D8,$FC,$D3,$5F
	.dc.b	$E0,$CD,$7F,$BC,$79,$7E,$97,$FB,$E1,$E7,$CF,$FC,$69,$19,$8D,$9E
	.dc.b	$B3,$E7,$1E,$A2,$98,$8F,$F1,$6F,$99,$F9,$68,$FF,$0E,$47,$E5,$1B
	.dc.b	$D3,$FE,$2E,$9D,$47,$C5,$39,$F3,$FE,$3D,$FD,$4D,$7F,$8E,$3C,$DD
	.dc.b	$C7,$F8,$81,$B1,$93,$B4,$7F,$8A,$63,$3E,$3F,$F8,$FF,$A7,$FC,$7B
	.dc.b	$E7,$7F,$A2,$5B,$7F,$C3,$EF,$FC,$57,$39,$7F,$05,$B7,$C3,$9F,$F2
	.dc.b	$8E,$FF,$E5,$08,$C7,$17,$F9,$46,$3F,$4B,$9F,$A6,$FD,$8D,$7F,$BB
	.dc.b	$F3,$FF,$9E,$08,$FF,$99,$5F,$F3,$A1,$1F,$F3,$CB,$5F,$F1,$CF,$4F
	.dc.b	$FA,$00,$8F,$F9,$B5,$FF,$3E,$11,$FF,$40,$B5,$FF,$3F,$78,$5F,$E9
	.dc.b	$7F,$82,$FF,$4A,$F5,$FF,$AA,$DB,$A1,$A7,$FA,$94,$95,$FE,$AA,$8D
	.dc.b	$36,$3F,$D5,$F1,$88,$FF,$46,$71,$7F,$AD,$67,$24,$8F,$F4,$23,$FE
	.dc.b	$9A,$23,$83,$FD,$32,$FF,$A5,$C8,$E0,$FF,$28,$47,$FB,$59,$F0,$03
	.dc.b	$F9,$1F,$A9,$67,$FC,$AD,$18,$36,$57,$FD,$CB,$C7,$B8,$FF,$72,$B5
	.dc.b	$FF,$53,$71,$FF,$BB,$67,$71,$CB,$D6,$7A,$DB,$8F,$14,$61,$D4,$65
	.dc.b	$FF,$7E,$1E,$2F,$5F,$FE,$00,$57,$99,$B1,$FE,$6D,$6B,$FE,$B7,$6B
	.dc.b	$EF,$F6,$9E,$D7,$56,$C7,$B9,$B5,$F0,$9B,$1F,$E7,$9C,$3D,$8D,$CB
	.dc.b	$CD,$18,$8F,$F6,$33,$5F,$9D,$ED,$BC,$CF,$63,$63,$E6,$B7,$0F,$A2
	.dc.b	$D8,$FF,$44,$7B,$8F,$77,$F6,$11,$FE,$D3,$8C,$70,$FE,$D6,$11,$8E
	.dc.b	$BF,$B6,$D8,$E7,$5C,$9F,$C1,$46,$5E,$91,$97,$9B,$63,$51,$C2,$30
	.dc.b	$4A,$F1,$4E,$A7,$8C,$62,$75,$5E,$46,$CF,$92,$74,$49,$D3,$63,$51
	.dc.b	$82,$38,$21,$EA,$8E,$0D,$8D,$B4,$D9,$EB,$8D,$B6,$36,$D3,$67,$B1
	.dc.b	$CB,$4C,$D7,$C1,$25,$7B,$A3,$C0,$43,$A1,$4E,$45,$8E,$35,$E4,$8C
	.dc.b	$9C,$08,$60,$91,$C2,$BC,$AD,$8F,$2D,$78,$DE,$31,$97,$C1,$19,$79
	.dc.b	$B6,$35,$1C,$23,$04,$AF,$34,$EA,$78,$C6,$27,$55,$E4,$6C,$F9,$27
	.dc.b	$44,$9D,$36,$35,$18,$23,$82,$1E,$A8,$E0,$D8,$DB,$4D,$9E,$B8,$DB
	.dc.b	$63,$6D,$36,$7B,$1C,$B4,$CD,$7C,$12,$57,$BA,$3C,$04,$3A,$14,$E4
	.dc.b	$58,$E3,$5E,$48,$C9,$C0,$FF,$42,$9F,$CE,$CA,$7F,$A3,$0C,$1F,$E8
	.dc.b	$6F,$39,$E6,$FF,$3E,$8F,$E0,$F8,$7F,$A7,$9F,$E0,$6D,$73,$5B,$BF
	.dc.b	$F4,$C3,$67,$51,$89,$FE,$1F,$18,$24,$EC,$91,$A6,$BF,$C7,$A3,$6E
	.dc.b	$48,$66,$36,$FF,$21,$8D,$B9,$36,$66,$36,$FB,$1A,$FB,$0D,$99,$29
	.dc.b	$93,$A1,$0E,$86,$A7,$E0,$39,$39,$14,$E4,$70,$30,$F0,$70,$43,$89
	.dc.b	$4E,$26,$08,$60,$8F,$37,$2D,$73,$3F,$19,$E9,$19,$24,$66,$38,$B5
	.dc.b	$E3,$3F,$22,$76,$D9,$D4,$62,$7E,$24,$60,$93,$B2,$46,$9A,$FC,$48
	.dc.b	$DB,$92,$19,$8D,$BE,$A8,$DB,$93,$66,$63,$6F,$B1,$AF,$B0,$D9,$92
	.dc.b	$99,$3A,$10,$E8,$6A,$7E,$03,$93,$91,$4E,$47,$03,$0F,$07,$04,$38
	.dc.b	$94,$E2,$60,$86,$08,$F3,$72,$D7,$33,$F1,$9E,$91,$92,$46,$63,$F1
	.dc.b	$BC,$8F,$03,$FC,$8D,$87,$F9,$12,$31,$3F,$C2,$8F,$F2,$6C,$E9,$A6
	.dc.b	$4F,$E2,$99,$FF,$15,$AF,$F0,$EC,$6C,$C9,$4F,$F1,$39,$4F,$F2,$0B
	.dc.b	$FE,$54,$DB,$FC,$8C,$DE,$3D,$46,$27,$F8,$FF,$0B,$4E,$4D,$8F,$21
	.dc.b	$FE,$43,$8C,$9D,$FE,$0D,$F0,$94,$93,$C0,$46,$9C,$68,$EB,$F4,$36
	.dc.b	$33,$18,$3D,$9D,$8F,$47,$FC,$89,$AF,$F9,$C2,$31,$CB,$DD,$B1,$99
	.dc.b	$DD,$FF,$3C,$F8,$77,$67,$BE,$BF,$CF,$31,$8D,$7D,$0D,$8C,$CE,$EB
	.dc.b	$FA,$2F,$CB,$BE,$A3,$11,$FD,$2E,$31,$87,$C1,$6D,$FF,$A1,$9B,$3F
	.dc.b	$D3,$B6,$F8,$9E,$FF,$FA,$29,$B1,$9E,$0F,$F4,$84,$E5,$B1,$FE,$96
	.dc.b	$DE,$BF,$EA,$46,$BF,$E9,$7E,$0F,$99,$5F,$93,$9C,$E4,$FC,$09,$F5
	.dc.b	$BF,$81,$FA,$8F,$EC,$33,$98,$C6,$3F,$D9,$E3,$02,$77,$5C,$A3,$13
	.dc.b	$F1,$B6,$E2,$FC,$56,$BF,$EA,$BD,$7E,$3D,$FC,$59,$CC,$B8,$BE,$EB
	.dc.b	$F6,$F8,$FF,$B9,$46,$23,$FB,$BB,$5F,$F5,$B6,$0F,$C1,$5F,$15,$FD
	.dc.b	$DE,$30,$D8,$CC,$7F,$AE,$8C,$9F,$EB,$CF,$39,$E5,$3E,$17,$FC,$05
	.dc.b	$18,$6C,$67,$0E,$0D,$7E,$5F,$BA,$F2,$7C,$4D,$C3,$C0,$D9,$DC,$FF
	.dc.b	$82,$F3,$F6,$47,$FB,$45,$AF,$C5,$EA,$FF,$52,$3F,$09,$AF,$C2,$F8
	.dc.b	$EF,$DE,$61,$ED,$6C,$7D,$E6,$BF,$75,$AF,$F8,$96,$BB,$92,$BA,$C7
	.dc.b	$8D,$BC,$0F,$BF,$7D,$87,$F8,$96,$30,$D8,$CE,$3F,$EE,$89,$CB,$F2
	.dc.b	$3E,$93,$FD,$DD,$E3,$FF,$8C,$B8,$3E,$8C,$63,$2F,$F1,$C4,$61,$FF
	.dc.b	$5F,$18,$DD,$3B,$8F,$C1,$EE,$F8,$DF,$A2,$4C,$7E,$8B,$F3,$F5,$F6
	.dc.b	$D7,$CD,$E4,$FC,$7C,$63,$7F,$A7,$6E,$0F,$E2,$8F,$CD,$9F,$C7,$46
	.dc.b	$2F,$E7,$47,$5B,$E6,$48,$FC,$E1,$0F,$B4,$E5,$A7,$F9,$53,$BD,$F8
	.dc.b	$FB,$D7,$FE,$6D,$8F,$F9,$6E,$73,$18,$A1,$FC,$82,$B8,$57,$E9,$A3
	.dc.b	$35,$C2,$31,$1F,$76,$FC,$4F,$E0,$F0,$F8,$44,$8C,$C6,$CF,$51,$FB
	.dc.b	$2E,$67,$C8,$7E,$4A,$FF,$85,$46,$1F,$C8,$9F,$B7,$8D,$4F,$16,$BF
	.dc.b	$D5,$7E,$C9,$EB,$FB,$63,$0D,$A4,$E9,$F5,$98,$8D,$7E,$0C,$F9,$5F
	.dc.b	$E3,$97,$FC,$22,$30,$F0,$6B,$FD,$8C,$FE,$CC,$E2,$BE,$AC,$43,$E3
	.dc.b	$65,$F1,$23,$C0,$78,$B0,$F2,$4F,$38,$CB,$63,$31,$CE,$3E,$61,$4F
	.dc.b	$CA,$1F,$38,$C9,$E0,$30,$7F,$A6,$7C,$6F,$C0,$F5,$BC,$DF,$F5,$07
	.dc.b	$C3,$3D,$0F,$F0,$A4,$E6,$FF,$0F,$39,$BE,$37,$C2,$F6,$D8,$CD,$70
	.dc.b	$9C,$B6,$32,$F8,$CF,$E4,$53,$F1,$8F,$E7,$84,$7F,$42,$D7,$46,$08
	.dc.b	$7F,$3E,$C7,$EF,$47,$C9,$D7,$FC,$34,$DC,$BF,$D1,$E7,$D2,$7E,$D5
	.dc.b	$7E,$F5,$EC,$FE,$7D,$B0,$D3,$F0,$71,$F5,$DB,$1F,$D8,$DB,$3F,$89
	.dc.b	$D7,$F5,$F5,$E1,$25,$7F,$69,$C3,$CE,$DC,$3C,$EF,$E9,$36,$FF,$19
	.dc.b	$47,$E1,$77,$FD,$33,$63,$33,$B3,$F1,$78,$FA,$4F,$8A,$53,$E4,$90
	.dc.b	$C1,$0F,$84,$7F,$09,$E2,$FE,$0B,$8F,$E7,$F8,$BF,$82,$36,$33,$5B
	.dc.b	$E3,$F4,$DF,$53,$99,$79,$7F,$C6,$B7,$F3,$A3,$27,$86,$32,$FF,$51
	.dc.b	$E4,$FB,$2D,$BF,$CA,$3F,$D6,$E3,$2D,$8C,$BC,$E1,$C9,$88,$EE,$78
	.dc.b	$25,$CF,$FD,$0D,$E2,$63,$9F,$AC,$FF,$98,$E7,$58,$FF,$7A,$9C,$F5
	.dc.b	$7F,$51,$FB,$2F,$EA,$DD,$7F,$D4,$30,$46,$3A,$FF,$98,$B6,$32,$7F
	.dc.b	$34,$3F,$C5,$5D,$DF,$D9,$31,$C4,$63,$BF,$C7,$DF,$FD,$9A,$72,$D9
	.dc.b	$FB,$E7,$BB,$83,$E5,$BF,$7F,$6F,$EA,$C7,$B0,$A7,$C0,$EE,$FF,$02
	.dc.b	$DF,$F8,$0E,$3E,$0C,$F3,$E1,$F9,$73,$F6,$6F,$AC,$FC,$5E,$EE,$9C
	.dc.b	$3F,$E1,$BA,$FF,$0C,$E1,$F3,$23,$15,$FD,$D7,$C0,$FD,$2B,$4F,$B3
	.dc.b	$C3,$FE,$2B,$6F,$0F,$F8,$AF,$0F,$F2,$9F,$4F,$F8,$D9,$BE,$0F,$AE
	.dc.b	$31,$5F,$D3,$23,$D3,$A7,$9A,$3D,$5C,$BF,$83,$8C,$78,$7F,$E7,$6F
	.dc.b	$A2,$FF,$28,$6C,$69,$2F,$FC,$C1,$E2,$FF,$97,$63,$FC,$BF,$E3,$7F
	.dc.b	$A9,$DB,$5C,$6B,$73,$97,$FC,$EC,$3D,$9F,$F3,$93,$4C,$B8,$3F,$CF
	.dc.b	$CE,$4F,$C3,$79,$9F,$EA,$98,$C9,$23,$FC,$DD,$4F,$53,$D6,$48,$CB
	.dc.b	$FE,$9A,$8D,$47,$FA,$8F,$BF,$FD,$81,$39,$24,$7F,$84,$9F,$F3,$81
	.dc.b	$1C,$1E,$C7,$D6,$47,$07,$FA,$64,$EE,$7A,$BC,$DF,$F0,$9C,$7B,$89
	.dc.b	$1F,$ED,$2A,$CB,$4E,$27,$FB,$44,$A3,$CF,$FF,$6B,$4F,$FB,$76,$33
	.dc.b	$E8,$3D,$1E,$71,$FA,$F8,$C3,$A8,$CB,$FE,$EA,$3A,$FA,$5F,$B5,$AF
	.dc.b	$F7,$8C,$60,$D8,$F4,$7F,$DF,$B1,$FE,$FF,$F4,$F3,$18,$9F,$F7,$EC
	.dc.b	$FE,$F1,$7A,$9F,$DC,$F3,$FE,$2C,$DE,$8D,$36,$7F,$81,$B5,$FE,$10
	.dc.b	$D7,$FC,$7A,$D7,$AC,$7E,$EA,$30,$47,$04,$9F,$0B,$4F,$09,$67,$C2
	.dc.b	$D7,$C2,$D7,$C2,$D7,$C3,$1D,$09,$1D,$1B,$1D,$1A,$F4,$6B,$D1,$AF
	.dc.b	$48,$E4,$47,$93,$5E,$4D,$79,$35,$E4,$70,$21,$C0,$A7,$02,$9C,$0D
	.dc.b	$10,$D1,$4D,$18,$21,$82,$4F,$D3,$9F,$6B,$67,$CC,$D9,$F0,$B5,$EC
	.dc.b	$D7,$AB,$5E,$B1,$F4,$E3,$04,$70,$49,$F0,$B4,$F0,$96,$7C,$2D,$7C
	.dc.b	$2D,$7C,$2D,$7C,$31,$D0,$91,$D1,$B1,$D1,$AF,$46,$BD,$1A,$F4,$8E
	.dc.b	$44,$79,$35,$E4,$D7,$93,$5E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D
	.dc.b	$14,$D1,$82,$18,$24,$FD,$39,$F6,$B6,$7C,$DE,$7F,$F9,$C6,$BF,$84
	.dc.b	$35,$FF,$02,$B5,$FF,$39,$1F,$C2,$3D,$DF,$F2,$EB,$FC,$11,$AF,$F1
	.dc.b	$06,$D1,$A4,$BF,$E2,$D3,$B6,$CE,$DB,$3B,$9F,$04,$60,$8E,$08,$7F
	.dc.b	$1A,$8E,$0D,$8D,$B4,$D9,$FC,$7A,$36,$D8,$DB,$4D,$9E,$F7,$2D,$32
	.dc.b	$7A,$CB,$3E,$B6,$CF,$67,$81,$E3,$21,$DC,$A6,$8C,$10,$C1,$0F,$01
	.dc.b	$C0,$A6,$CA,$6E,$3C,$8E,$DA,$ED,$AE,$E3,$D0,$F2,$24,$69,$B1,$A8
	.dc.b	$C1,$1C,$12,$7E,$4C,$F0,$6C,$ED,$B3,$B9,$F9,$31,$82,$38,$21,$EB
	.dc.b	$8E,$0D,$8D,$B4,$D9,$EF,$8D,$B6,$36,$D3,$67,$C3,$72,$D3,$27,$AC
	.dc.b	$B3,$EB,$6C,$F6,$78,$1E,$32,$1D,$CA,$68,$C1,$0C,$10,$F0,$1C,$0A
	.dc.b	$6C,$A6,$E3,$C8,$ED,$AE,$DA,$EE,$3D,$0F,$22,$46,$9B,$1A,$8F,$C2
	.dc.b	$62,$C9,$4F,$3F,$FC,$34,$D3,$FD,$27,$3B,$70,$7F,$A6,$5C,$47,$F8
	.dc.b	$77,$D8,$7C,$9F,$48,$C3,$6B,$FD,$3B,$3A,$79,$9E,$27,$13,$FE,$3B
	.dc.b	$24,$7F,$A9,$38,$3F,$C7,$31,$88,$F1,$7B,$7D,$5A,$74,$3F,$C6,$CE
	.dc.b	$5F,$7E,$08,$C4,$39,$9C,$98,$6C,$73,$6D,$FF,$9D,$8F,$F1,$E7,$C9
	.dc.b	$3C,$2E,$6D,$AF,$13,$67,$73,$E0,$ED,$78,$BF,$ED,$89,$FF,$23,$46
	.dc.b	$3B,$FD,$4D,$8C,$CE,$F0,$FF,$63,$F7,$BE,$B9,$CB,$F0,$63,$0D,$8C
	.dc.b	$DF,$F9,$4E,$72,$FF,$95,$7C,$67,$6E,$B0,$FF,$69,$C6,$30,$F9,$8D
	.dc.b	$AE,$15,$BA,$FF,$3F,$0F,$67,$FD,$36,$47,$FD,$08,$FF,$98,$FE,$71
	.dc.b	$E7,$FE,$4B,$CB,$C3,$6B,$FD,$C7,$3A,$53,$A6,$9F,$0B,$4F,$83,$1E
	.dc.b	$D6,$C6,$73,$DB,$FE,$F1,$D3,$E2,$D7,$CC,$DF,$FB,$0E,$FF,$46,$B0
	.dc.b	$43,$FD,$F1,$6F,$B8,$7D,$E7,$F7,$08,$C3,$63,$31,$FE,$81,$32,$5F
	.dc.b	$31,$F3,$1C,$5B,$19,$8C,$1D,$7C,$C7,$E2,$D4,$93,$D6,$32,$FA,$B9
	.dc.b	$5A,$52,$67,$EB,$3F,$80,$CF,$AE,$BD,$04,$7D,$19,$23,$FC,$17,$A7
	.dc.b	$C5,$AF,$E1,$27,$F0,$A7,$0E,$1A,$E7,$0F,$1B,$C0,$FB,$7E,$23,$E3
	.dc.b	$7F,$C3,$3B,$FF,$B6,$DB,$19,$CB,$FD,$55,$39,$7F,$D5,$9E,$7F,$97
	.dc.b	$DA,$FE,$37,$F2,$7D,$63,$18,$A3,$1A,$2B,$73,$EA,$AD,$D7,$96,$DC
	.dc.b	$3F,$05,$FB,$18,$7D,$72,$3F,$32,$FD,$9A,$FC,$04,$D3,$E1,$C7,$E2
	.dc.b	$7C,$6F,$C3,$11,$F9,$EF,$F8,$E3,$EF,$DF,$3B,$FC,$DB,$AB,$F7,$B8
	.dc.b	$7D,$09,$CB,$FA,$FB,$F9,$91,$89,$DC,$FF,$3E,$8C,$4F,$9B,$87,$E9
	.dc.b	$CD,$E3,$F5,$3F,$66,$32,$FF,$46,$EB,$7D,$5B,$D0,$F1,$E1,$F8,$29
	.dc.b	$CC,$62,$B8,$56,$EB,$F1,$F1,$89,$D3,$E5,$CF,$B4,$FA,$9B,$1E,$F6
	.dc.b	$BB,$3F,$8A,$1E,$7C,$7D,$2F,$E3,$DA,$B2,$7D,$B7,$F2,$6C,$BE,$2C
	.dc.b	$E6,$31,$87,$CA,$6C,$7C,$F6,$BF,$CB,$B8,$1F,$87,$FA,$AB,$F0,$ED
	.dc.b	$8C,$C7,$C0,$9E,$13,$BF,$8F,$7D,$37,$F0,$C9,$FD,$9C,$61,$E1,$48
	.dc.b	$C4,$7D,$C8,$C4,$F1,$8F,$80,$7E,$F4,$A7,$DD,$3F,$39,$7E,$47,$93
	.dc.b	$67,$F8,$EB,$63,$F8,$2B,$5E,$67,$CF,$3B,$1F,$9E,$DF,$F0,$EE,$88
	.dc.b	$67,$4F,$A7,$E7,$FF,$A7,$7C,$47,$A1,$FE,$9A,$6D,$7F,$85,$27,$53
	.dc.b	$92,$46,$7E,$13,$FC,$2D,$5C,$21,$5C,$1D,$78,$DF,$E1,$E9,$D7,$3F
	.dc.b	$F6,$DC,$3F,$A0,$12,$BF,$44,$D7,$F1,$8D,$7F,$18,$FC,$AD,$FF,$B7
	.dc.b	$BF,$29,$BA,$FC,$A6,$CF,$CA,$6B,$C8,$EE,$43,$B9,$FD,$3B,$7F,$C1
	.dc.b	$57,$DF,$6D,$67,$7F,$F1,$CC,$E5,$B1,$E1,$3E,$69,$63,$CC,$E2,$BC
	.dc.b	$8D,$7F,$83,$70,$7F,$82,$BE,$1F,$F4,$BC,$7F,$A8,$A4,$62,$75,$1F
	.dc.b	$A7,$32,$7A,$EB,$DC,$7B,$08,$7E,$E6,$7A,$F2,$7F,$94,$9B,$1C,$8F
	.dc.b	$E0,$05,$3F,$C2,$BC,$DF,$C2,$F1,$FE,$3B,$CB,$FC,$43,$0F,$53,$63
	.dc.b	$33,$B9,$F2,$57,$E4,$E7,$EE,$34,$FE,$20,$3D,$6F,$19,$27,$2F,$EC
	.dc.b	$7C,$1F,$F7,$25,$7F,$47,$6D,$7F,$81,$9A,$FF,$6E,$6B,$F1,$08,$78
	.dc.b	$4F,$8B,$5F,$D4,$3A,$FF,$1B,$1F,$B4,$6B,$FB,$46,$BF,$C4,$DB,$DB
	.dc.b	$FD,$83,$E5,$3F,$AF,$F7,$FF,$9F,$9B,$46,$9F,$DC,$FC,$0F,$ED,$99
	.dc.b	$5F,$07,$FB,$83,$7C,$2E,$F3,$AB,$F9,$7C,$FF,$AB,$6F,$6F,$F8,$42
	.dc.b	$FE,$26,$9E,$6D,$7F,$C1,$C7,$ED,$EB,$FC,$AF,$B7,$F6,$DA,$DC,$E6
	.dc.b	$7E,$21,$F1,$B0,$EC,$FC,$6B,$FF,$0F,$D7,$F7,$76,$C7,$F8,$D9,$AF
	.dc.b	$F8,$79,$AF,$F3,$9F,$13,$C9,$DF,$FE,$30,$AF,$E4,$B1,$83,$D1,$8F
	.dc.b	$F8,$EA,$3E,$F7,$8B,$82,$65,$E7,$CB,$CB,$8F,$9F,$3F,$84,$FE,$43
	.dc.b	$C6,$FF,$39,$FB,$1F,$E5,$5B,$34,$93,$FE,$60,$F2,$BF,$D6,$B1,$FE
	.dc.b	$5F,$24,$64,$E2,$7F,$8A,$FC,$DF,$F4,$BF,$37,$F8,$B6,$3F,$D3,$9E
	.dc.b	$77,$FA,$FA,$31,$3A,$9E,$13,$97,$FC,$FF,$D3,$FE,$8A,$E5,$FF,$3E
	.dc.b	$E4,$69,$27,$FC,$D8,$FA,$C8,$E0,$FF,$06,$3F,$E6,$D2,$38,$3F,$C8
	.dc.b	$5E,$77,$FA,$9E,$3F,$D4,$B1,$DC,$CF,$9D,$FE,$D6,$8C,$9E,$AC,$31
	.dc.b	$3E,$A3,$81,$FE,$92,$8D,$38,$3F,$D8,$3E,$7F,$07,$FD,$8F,$1F,$1F
	.dc.b	$D0,$FF,$62,$46,$CF,$F6,$49,$7D,$17,$AF,$C6,$3F,$DD,$31,$87,$51
	.dc.b	$95,$19,$AF,$F7,$1C,$75,$F4,$FF,$DC,$51,$82,$47,$FB,$C3,$D5,$E5
	.dc.b	$87,$FB,$C6,$7F,$DE,$A4,$9F,$F3,$57,$AD,$FC,$31,$34,$9E,$8F,$F0
	.dc.b	$68,$CB,$63,$3B,$FE,$BD,$BD,$3F,$89,$6D,$76,$AE,$04,$AE,$0D,$AE
	.dc.b	$0D,$9E,$13,$A2,$4E,$9B,$1A,$8C,$11,$C1,$0F,$F4,$7D,$79,$1B,$5D
	.dc.b	$DB,$5D,$1B,$5C,$DA,$73,$07,$C4,$8E,$0E,$2B,$B3,$6B,$B3,$67,$6F
	.dc.b	$1A,$EC,$E5,$AE,$4F,$01,$4F,$19,$0E,$06,$AB,$C2,$EC,$C9,$B3,$04
	.dc.b	$76,$4A,$F3,$C6,$5B,$19,$AF,$4B,$6B,$CE,$DA,$F1,$57,$02,$57,$06
	.dc.b	$D7,$06,$CF,$09,$D1,$27,$4D,$8D,$46,$08,$E0,$87,$BE,$BC,$8D,$AE
	.dc.b	$ED,$AE,$8D,$AE,$6D,$39,$9F,$28,$87,$C5,$8E,$2E,$2B,$BB,$6B,$BB
	.dc.b	$67,$6F,$1A,$EE,$E5,$AE,$4F,$09,$4F,$21,$0E,$06,$AB,$C2,$EC,$C9
	.dc.b	$B3,$04,$76,$7E,$23,$CD,$79,$CF,$3B,$FA,$25,$7F,$99,$A3,$0F,$F4
	.dc.b	$79,$DC,$E9,$B1,$FC,$0F,$E4,$FF,$D4,$EF,$E9,$5A,$FF,$10,$6D,$9A
	.dc.b	$4C,$7C,$31,$96,$C6,$67,$B4,$60,$8E,$08,$7F,$20,$8E,$0D,$8D,$B4
	.dc.b	$D9,$FC,$22,$36,$D8,$DB,$4D,$9E,$D7,$2D,$32,$7B,$DC,$B4,$CC,$FB
	.dc.b	$5B,$3E,$07,$91,$E5,$21,$E1,$29,$A3,$04,$30,$43,$C0,$70,$29,$B2
	.dc.b	$9B,$8F,$23,$B6,$BB,$6B,$B8,$F4,$3C,$89,$1A,$6C,$6A,$30,$47,$04
	.dc.b	$9F,$8F,$19,$6C,$66,$7E,$3C,$60,$8E,$08,$7A,$63,$83,$63,$6D,$36
	.dc.b	$7B,$63,$6D,$8D,$B4,$D9,$F0,$5C,$B4,$C9,$F0,$1C,$B4,$CC,$FB,$5B
	.dc.b	$3E,$07,$91,$E5,$21,$E1,$29,$A3,$04,$30,$43,$C0,$70,$29,$B2,$9B
	.dc.b	$8F,$23,$B6,$BB,$6B,$B8,$F4,$3C,$89,$1A,$6C,$6A,$30,$7E,$17,$A9
	.dc.b	$F0,$BF,$DB,$B1,$FF,$2A,$46,$2F,$F8,$89,$FD,$C8,$39,$8D,$CF,$19
	.dc.b	$D1,$FE,$2C,$FA,$6C,$3C,$78,$91,$88,$FF,$1A,$B5,$E0,$78,$0F,$F0
	.dc.b	$0D,$BE,$B3,$4D,$8D,$1E,$48,$C9,$CE,$9C,$F4,$93,$C2,$32,$F2,$AF
	.dc.b	$E6,$6D,$DB,$B9,$1D,$7B,$FE,$22,$38,$3C,$7E,$03,$C0,$7A,$98,$6E
	.dc.b	$DF,$E8,$19,$D4,$E4,$F5,$FD,$0B,$E7,$9C,$78,$6D,$FF,$A2,$67,$53
	.dc.b	$93,$D9,$B7,$F8,$A3,$D8,$FF,$47,$6F,$EE,$6C,$67,$4F,$AF,$39,$7E
	.dc.b	$F5,$3D,$57,$5F,$F8,$CB,$CB,$FB,$B3,$98,$C5,$70,$AD,$AA,$DC,$61
	.dc.b	$F8,$1B,$FC,$46,$ED,$BA,$FC,$4E,$FF,$12,$72,$6A,$21,$AA,$F9,$AE
	.dc.b	$0F,$CB,$4F,$CC,$70,$7F,$92,$9C,$3C,$6B,$FD,$65,$8B,$CA,$75,$7F
	.dc.b	$6A,$8C,$36,$33,$1C,$1B,$5F,$A0,$F3,$DE,$63,$D1,$FE,$F7,$18,$6C
	.dc.b	$66,$7F,$D8,$53,$97,$FC,$03,$97,$99,$F5,$E7,$EE,$7D,$07,$B0,$46
	.dc.b	$DF,$74,$6D,$E3,$9F,$48,$C4,$6A,$30,$47,$A4,$F3,$7A,$CF,$06,$CE
	.dc.b	$DF,$C3,$E3,$FE,$07,$FD,$4F,$F0,$AD,$8C,$F6,$FE,$6A,$72,$D8,$E3
	.dc.b	$F8,$DF,$CD,$26,$BF,$35,$FF,$0C,$E5,$E3,$26,$5E,$36,$E5,$E3,$BF
	.dc.b	$79,$F8,$0F,$B7,$7E,$A2,$FB,$7F,$CD,$65,$FC,$6E,$73,$18,$AF,$F4
	.dc.b	$7C,$61,$FC,$D4,$EE,$B9,$46,$0E,$E7,$D4,$21,$82,$1F,$3C,$F9,$B8
	.dc.b	$FD,$2A,$ED,$F8,$27,$EF,$A7,$C3,$5D,$A3,$10,$9D,$DF,$8F,$0F,$F5
	.dc.b	$2D,$6A,$BE,$A9,$DB,$1F,$88,$FE,$87,$E5,$BF,$85,$F4,$EF,$44,$63
	.dc.b	$1F,$D8,$B6,$BF,$56,$D7,$F8,$9F,$03,$F1,$DF,$45,$7E,$09,$B3,$FE
	.dc.b	$B8,$8F,$D6,$46,$23,$8B,$78,$BF,$4F,$DB,$FC,$42,$7F,$D7,$F1,$87
	.dc.b	$91,$8E,$4F,$E4,$8E,$04,$6D,$F6,$70,$FE,$14,$9A,$FC,$2C,$7C,$67
	.dc.b	$12,$CF,$CE,$69,$EB,$3B,$3F,$30,$FB,$A7,$DB,$29,$93,$EF,$94,$D1
	.dc.b	$FD,$A7,$C5,$FF,$3B,$FB,$FF,$C3,$A3,$FD,$BB,$18,$E8,$D7,$83,$FE
	.dc.b	$80,$AD,$57,$FA,$0F,$C0,$FF,$71,$C6,$2F,$84,$F2,$AE,$17,$CA,$F6
	.dc.b	$D9,$5F,$4E,$1F,$EA,$33,$FC,$C8,$93,$FC,$95,$AF,$F0,$A6,$BF,$C2
	.dc.b	$5F,$DF,$EF,$FD,$59,$F9,$5C,$1F,$85,$CF,$F9,$2D,$7E,$01,$A7,$D2
	.dc.b	$9F,$AF,$A7,$BA,$7A,$1E,$12,$1F,$81,$E0,$3D,$37,$FC,$35,$7E,$99
	.dc.b	$B7,$FE,$17,$6C,$FE,$4F,$0F,$25,$79,$B1,$F4,$5F,$DF,$8C,$3E,$8A
	.dc.b	$FE,$5B,$C3,$FD,$CB,$D1,$FD,$43,$67,$F4,$E2,$32,$D3,$D3,$3E,$C3
	.dc.b	$E5,$18,$E6,$FE,$1D,$F6,$1F,$BE,$6C,$FE,$F6,$53,$9E,$8F,$E2,$2F
	.dc.b	$C8,$C7,$E9,$CF,$F3,$46,$EF,$F7,$23,$C2,$64,$F0,$57,$F1,$AF,$4F
	.dc.b	$A7,$4F,$CC,$6C,$9A,$77,$79,$D7,$57,$D6,$43,$DC,$43,$C0,$7D,$1E
	.dc.b	$AF,$C3,$3F,$D0,$32,$FC,$01,$FD,$C6,$33,$1B,$8C,$EF,$FE,$D3,$FA
	.dc.b	$EF,$E9,$30,$69,$27,$CB,$D1,$FE,$30,$D3,$E1,$B6,$33,$C9,$F8,$F7
	.dc.b	$CA,$79,$F2,$F6,$36,$78,$3E,$D3,$64,$8D,$BC,$B8,$FF,$BD,$5F,$F8
	.dc.b	$0F,$A7,$FB,$F3,$7B,$BF,$32,$DA,$F6,$BF,$01,$FE,$5A,$DC,$FF,$C2
	.dc.b	$26,$8F,$46,$1E,$02,$63,$E6,$AF,$46,$5A,$8C,$47,$9D,$B3,$FE,$50
	.dc.b	$DB,$FC,$62,$49,$FF,$12,$3F,$E3,$8F,$07,$F0,$91,$FE,$3D,$6B,$FC
	.dc.b	$F7,$C1,$FF,$1F,$CE,$E3,$F1,$07,$F9,$33,$C3,$FD,$BF,$37,$F9,$42
	.dc.b	$3F,$4C,$64,$A7,$FA,$16,$36,$FF,$A2,$BC,$4F,$F2,$EF,$99,$FE,$5F
	.dc.b	$EB,$FF,$31,$B7,$63,$4F,$F2,$59,$2B,$FC,$7C,$D7,$FD,$35,$E3,$BC
	.dc.b	$7F,$E1,$73,$FE,$A0,$6C,$FF,$A6,$A7,$4F,$63,$11,$A7,$9F,$81,$5F
	.dc.b	$F4,$BB,$5F,$F5,$3B,$67,$FD,$1B,$5F,$0C,$93,$F0,$DA,$F9,$BC,$8E
	.dc.b	$04,$7C,$46,$4A,$7F,$AA,$8A,$7F,$AA,$A3,$FD,$03,$E4,$F8,$FC,$9F
	.dc.b	$F5,$8C,$EA,$FF,$D7,$D5,$FE,$BA,$8C,$79,$5E,$73,$D5,$E5,$BC,$9F
	.dc.b	$CB,$4A,$31,$AF,$EA,$DB,$D1,$FE,$C2,$9D,$57,$09,$CC,$7F,$B1,$9A
	.dc.b	$FF,$B3,$5A,$FF,$3E,$6B,$CA,$BB,$4F,$32,$38,$25,$78,$6B,$89,$27
	.dc.b	$71,$96,$D7,$8D,$AF,$8D,$AF,$8D,$AF,$8E,$39,$57,$82,$39,$35,$E4
	.dc.b	$D7,$93,$5E,$47,$08,$E6,$47,$04,$39,$B9,$69,$93,$AB,$96,$99,$3B
	.dc.b	$9C,$08,$68,$A6,$8C,$10,$C1,$2B,$E6,$4F,$AD,$B3,$E1,$9D,$57,$C8
	.dc.b	$9C,$C7,$8D,$AF,$81,$AF,$76,$BC,$AB,$B4,$F3,$23,$82,$57,$86,$B8
	.dc.b	$92,$77,$19,$6D,$78,$DA,$F8,$DA,$F8,$DA,$F8,$E3,$95,$78,$23,$93
	.dc.b	$5E,$4D,$79,$35,$E4,$70,$8E,$64,$70,$43,$9B,$96,$99,$3A,$B9,$69
	.dc.b	$93,$B9,$C0,$86,$8A,$68,$C1,$0C,$12,$BE,$64,$FA,$DB,$3E,$19,$D5
	.dc.b	$7C,$89,$CC,$7E,$4F,$C7,$6C,$F1,$FF,$84,$4F,$F9,$7C,$FE,$15,$3A
	.dc.b	$8C,$B6,$3F,$7D,$E3,$7F,$A8,$5F,$E1,$0D,$7F,$8B,$36,$FF,$CA,$91
	.dc.b	$C3,$2E,$F1,$88,$FE,$29,$3B,$24,$E9,$B1,$C4,$87,$08,$D1,$1C,$9F
	.dc.b	$C7,$63,$8B,$B3,$27,$A2,$38,$39,$21,$C5,$CB,$4E,$27,$B9,$CB,$CC
	.dc.b	$CC,$FB,$5E,$07,$43,$45,$39,$1C,$27,$CE,$F0,$21,$C0,$A7,$03,$94
	.dc.b	$77,$74,$43,$47,$07,$4E,$0D,$38,$24,$78,$A3,$2D,$78,$12,$37,$3F
	.dc.b	$0A,$31,$3B,$9C,$C7,$B6,$75,$18,$8F,$4C,$EC,$93,$A6,$C7,$12,$1C
	.dc.b	$23,$44,$72,$7B,$E3,$8B,$B3,$27,$B2,$38,$39,$21,$C5,$CB,$4E,$27
	.dc.b	$B9,$CB,$CC,$CC,$FB,$5E,$07,$43,$45,$39,$1C,$27,$CE,$F0,$21,$C0
	.dc.b	$A7,$03,$94,$77,$74,$43,$47,$07,$4E,$0D,$38,$24,$78,$A3,$2D,$78
	.dc.b	$12,$37,$3F,$0A,$31,$3B,$9C,$C7,$B6,$75,$1F,$92,$D5,$F8,$2F,$FB
	.dc.b	$3B,$1F,$F0,$D4,$62,$FF,$8B,$1F,$E4,$59,$D3,$4C,$9F,$E3,$59,$E3
	.dc.b	$C3,$A8,$C1,$FE,$4D,$70,$FF,$B7,$38,$1C,$B8,$84,$61,$EA,$6E,$35
	.dc.b	$19,$AF,$F1,$87,$AA,$F4,$F9,$36,$F8,$87,$13,$C2,$7C,$1D,$8E,$D7
	.dc.b	$77,$08,$C3,$63,$35,$FE,$3F,$9C,$BF,$23,$E1,$9C,$1E,$0D,$D1,$89
	.dc.b	$D4,$E6,$31,$9F,$FB,$F8,$FF,$7F,$DF,$83,$CC,$ED,$8F,$91,$B9,$1A
	.dc.b	$74,$25,$7F,$9E,$EF,$C7,$18,$8E,$D7,$EC,$79,$90,$F1,$07,$4F,$97
	.dc.b	$0F,$F4,$8E,$FE,$EF,$A7,$DC,$60,$38,$E9,$4E,$EB,$94,$62,$B7,$1D
	.dc.b	$7C,$67,$D6,$FF,$5D,$94,$61,$FF,$37,$18,$8D,$46,$67,$FD,$A5,$EA
	.dc.b	$97,$D4,$61,$A7,$CF,$F2,$1E,$42,$B8,$56,$CF,$89,$5F,$23,$87,$F0
	.dc.b	$AE,$0F,$AE,$E1,$E9,$F2,$0F,$97,$E7,$18,$6B,$E5,$C5,$3B,$8F,$29
	.dc.b	$97,$17,$DE,$3E,$A5,$F7,$7C,$9C,$7F,$E0,$AE,$1F,$AF,$1F,$07,$6F
	.dc.b	$F6,$3F,$DA,$FD,$18,$C7,$02,$31,$1F,$EC,$E6,$CF,$03,$EA,$C7,$9F
	.dc.b	$6F,$F5,$5F,$EF,$7F,$EF,$FF,$13,$E7,$36,$4D,$3F,$0B,$C2,$FB,$1E
	.dc.b	$AD,$C3,$AD,$7E,$A2,$31,$CB,$C2,$BC,$B3,$F0,$35,$F7,$6D,$F0,$63
	.dc.b	$E8,$C6,$0F,$C2,$EA,$F9,$0F,$5B,$72,$EE,$7F,$BA,$A3,$27,$C2,$F5
	.dc.b	$1F,$87,$BE,$87,$F7,$7F,$09,$FB,$CC,$BF,$19,$5A,$9C,$E1,$FB,$98
	.dc.b	$C2,$AE,$14,$9C,$C6,$23,$D8,$7D,$DE,$3F,$B2,$FB,$0F,$D9,$FC,$17
	.dc.b	$E5,$BD,$6F,$D8,$61,$FC,$1E,$30,$FE,$C0,$C7,$22,$77,$5F,$A3,$AD
	.dc.b	$F0,$F9,$2D,$E9,$3E,$1B,$F5,$1F,$21,$98,$C6,$08,$C4,$EA,$BF,$82
	.dc.b	$56,$A7,$D9,$9F,$AA,$FD,$A7,$F0,$C7,$07,$BA,$9E,$BB,$E0,$3F,$71
	.dc.b	$F6,$BE,$58,$C5,$A3,$13,$AA,$FE,$29,$5A,$9F,$9F,$DD,$FB,$CF,$B3
	.dc.b	$FD,$DD,$7E,$FF,$81,$18,$C7,$F3,$8D,$55,$C5,$FE,$13,$5F,$C8,$7F
	.dc.b	$24,$7D,$6F,$CC,$9F,$E7,$F1,$89,$FA,$51,$FB,$18,$C7,$07,$CF,$6C
	.dc.b	$FC,$3E,$AF,$CD,$57,$C6,$3F,$76,$7F,$0D,$84,$BE,$93,$C1,$18,$6C
	.dc.b	$66,$3C,$33,$E4,$9F,$33,$D4,$FC,$61,$E1,$EB,$FD,$43,$F9,$1B,$F1
	.dc.b	$F0,$7D,$67,$CB,$E9,$7F,$A4,$7E,$09,$EB,$7F,$57,$8C,$37,$D5,$FF
	.dc.b	$03,$CE,$B0,$FE,$65,$5A,$6C,$EA,$B8,$79,$BF,$CD,$EF,$85,$7F,$A8
	.dc.b	$DB,$7C,$63,$F8,$15,$7F,$62,$23,$EB,$6B,$FC,$91,$AF,$F0,$67,$FA
	.dc.b	$BE,$BF,$1F,$E1,$BE,$A6,$BF,$DE,$1B,$B1,$A7,$E7,$34,$FF,$02,$DF
	.dc.b	$4C,$3C,$67,$84,$FC,$37,$07,$3B,$A7,$32,$DF,$E0,$E3,$53,$F9,$21
	.dc.b	$F4,$FF,$9A,$E4,$FC,$BC,$62,$FF,$35,$C9,$FD,$76,$B4,$D8,$FD,$1F
	.dc.b	$3F,$F6,$1F,$98,$FD,$3D,$9A,$49,$F4,$47,$C7,$29,$F1,$CF,$DE,$F5
	.dc.b	$FF,$6B,$F8,$FF,$E0,$30,$69,$27,$F8,$8F,$77,$B2,$5D,$DF,$C5,$B6
	.dc.b	$FE,$46,$DA,$D9,$F1,$67,$C4,$43,$C4,$5F,$03,$F9,$6F,$DA,$7F,$2C
	.dc.b	$6C,$FF,$31,$F0,$FF,$9C,$29,$D4,$61,$FE,$41,$E2,$FF,$43,$C7,$FC
	.dc.b	$2D,$E2,$FE,$07,$1F,$F0,$AB,$5F,$EC,$ED,$7C,$64,$3F,$02,$79,$F0
	.dc.b	$FF,$0B,$F8,$DB,$69,$F8,$B8,$FF,$0F,$46,$7C,$7F,$EC,$57,$FD,$A7
	.dc.b	$F7,$E9,$5F,$C5,$FC,$AF,$F0,$D4,$62,$3F,$BC,$F9,$7D,$73,$FE,$F5
	.dc.b	$97,$59,$CB,$5F,$EF,$8D,$7C,$E7,$A7,$C1,$FF,$2E,$F5,$7E,$9B,$C4
	.dc.b	$FF,$09,$F9,$DF,$8D,$AF,$F0,$C1,$F9,$42,$F5,$7F,$91,$A7,$71,$93
	.dc.b	$EC,$1D,$89,$7D,$9F,$E6,$5E,$7F,$57,$FC,$18,$D7,$F9,$8F,$67,$F8
	.dc.b	$CD,$E8,$62,$35,$97,$F9,$04,$FE,$0F,$1F,$E0,$9F,$43,$FC,$E3,$9F
	.dc.b	$F8,$41,$B1,$FE,$4D,$6B,$FE,$22,$F4,$7E,$51,$2B,$E4,$1F,$E5,$BE
	.dc.b	$1F,$DF,$65,$E9,$AF,$6B,$5F,$F2,$77,$A9,$FE,$7E,$A7,$C0,$FF,$9F
	.dc.b	$89,$3F,$E7,$38,$CB,$FC,$C6,$3F,$CF,$64,$70,$6F,$83,$FD,$0E,$47
	.dc.b	$07,$F3,$97,$FD,$06,$47,$07,$F9,$EB,$D7,$FF,$51,$47,$FA,$7E,$0D
	.dc.b	$23,$FE,$80,$F6,$3F,$D6,$11,$FE,$82,$8F,$F5,$9F,$B3,$FE,$6B,$8C
	.dc.b	$4E,$A7,$84,$E5,$FF,$5F,$1D,$8B,$3F,$EC,$47,$07,$FA,$7F,$1F,$F4
	.dc.b	$2B,$83,$E5,$FB,$5C,$DD,$35,$C9,$B1,$1F,$ED,$79,$F4,$7B,$6F,$73
	.dc.b	$EF,$D7,$FA,$0E,$31,$A7,$E2,$21,$18,$E1,$E3,$5C,$27,$2D,$9D,$47
	.dc.b	$FA,$99,$AF,$FA,$E5,$AF,$78,$E0,$4A,$F0,$4E,$A7,$AC,$62,$75,$5E
	.dc.b	$28,$C1,$1C,$12,$7C,$53,$C1,$B3,$B6,$CE,$DF,$34,$ED,$B3,$B6,$CE
	.dc.b	$DF,$54,$65,$B1,$97,$D2,$DA,$F6,$12,$BB,$4F,$03,$C6,$43,$B9,$4D
	.dc.b	$18,$21,$82,$1E,$03,$81,$4D,$94,$DC,$78,$EB,$D5,$19,$24,$66,$BD
	.dc.b	$55,$E8,$9C,$B6,$75,$1E,$06,$BE,$06,$BE,$08,$E0,$4A,$F0,$CE,$A7
	.dc.b	$AC,$62,$75,$5E,$28,$C1,$1C,$12,$7C,$53,$C1,$B3,$B6,$CE,$DF,$34
	.dc.b	$ED,$B3,$B6,$CE,$DF,$54,$65,$B1,$97,$D2,$DA,$F6,$12,$BB,$4F,$03
	.dc.b	$C6,$43,$B9,$4D,$18,$21,$82,$1E,$03,$81,$4D,$94,$DC,$78,$EB,$D5
	.dc.b	$19,$24,$66,$BD,$55,$E8,$9C,$B6,$75,$1E,$06,$BE,$07,$F7,$9E,$5B
	.dc.b	$CE,$79,$DF,$CA,$E9,$39,$8C,$36,$3F,$98,$35,$FE,$1D,$F8,$5F,$FB
	.dc.b	$1D,$FE,$34,$D7,$F8,$CB,$6C,$D2,$65,$E1,$8C,$B6,$33,$3F,$CA,$1A
	.dc.b	$F8,$23,$44,$74,$D7,$46,$08,$60,$87,$F2,$58,$EA,$D8,$E4,$D3,$89
	.dc.b	$4E,$27,$C0,$76,$D3,$65,$36,$7C,$47,$6D,$36,$53,$73,$F1,$1C,$B5
	.dc.b	$CC,$FA,$1E,$C7,$94,$86,$08,$75,$32,$53,$31,$D5,$CB,$5C,$C7,$77
	.dc.b	$81,$23,$4D,$8D,$46,$08,$E0,$93,$F0,$A3,$2D,$8C,$CF,$C0,$6B,$F0
	.dc.b	$23,$44,$74,$D7,$46,$08,$60,$87,$BA,$3A,$B6,$39,$34,$E2,$53,$89
	.dc.b	$F1,$1D,$B4,$D9,$4D,$9F,$15,$DB,$4D,$94,$DC,$FC,$47,$2D,$73,$3E
	.dc.b	$87,$B1,$E5,$21,$82,$1D,$4C,$94,$CC,$75,$72,$D7,$31,$DD,$E0,$48
	.dc.b	$D3,$63,$51,$82,$38,$24,$FC,$28,$CB,$63,$F8,$57,$D5,$3D,$85,$FF
	.dc.b	$18,$29,$88,$D3,$93,$2E,$31,$FF,$24,$36,$38,$9F,$EE,$BF,$8F,$EB
	.dc.b	$3F,$E4,$DE,$64,$62,$77,$18,$7B,$1E,$47,$13,$FE,$52,$E2,$E1,$3E
	.dc.b	$6E,$34,$62,$3D,$71,$87,$99,$FE,$5C,$2F,$13,$E0,$B8,$47,$F9,$31
	.dc.b	$B1,$98,$DC,$6B,$9D,$E8,$3D,$4E,$B1,$86,$C6,$72,$FF,$36,$4E,$5F
	.dc.b	$F3,$76,$1D,$7D,$EB,$87,$D2,$6E,$26,$9C,$B0,$F0,$37,$1D,$4F,$D0
	.dc.b	$C7,$C5,$39,$35,$10,$D5,$7A,$5C,$1F,$82,$E1,$7D,$D3,$E1,$D0,$F6
	.dc.b	$3E,$EB,$A8,$CC,$62,$78,$4E,$EB,$F1,$6E,$CC,$3D,$FD,$77,$E1,$9F
	.dc.b	$D7,$E5,$18,$7F,$D2,$86,$23,$51,$99,$FF,$6A,$78,$6F,$AE,$FC,$DF
	.dc.b	$C0,$E1,$B3,$F9,$CF,$19,$5B,$85,$6D,$A7,$C6,$E0,$FE,$F5,$C1,$F7
	.dc.b	$63,$E3,$9F,$A8,$70,$FB,$BE,$AD,$E6,$78,$23,$E6,$46,$1E,$A6,$23
	.dc.b	$4F,$CE,$8D,$5F,$68,$57,$67,$F6,$AD,$7C,$5C,$3F,$6D,$F9,$FB,$FE
	.dc.b	$93,$7F,$9F,$1F,$6A,$31,$C1,$F5,$F9,$FE,$CC,$F0,$6B,$F5,$7F,$13
	.dc.b	$FC,$0F,$E9,$FF,$80,$E2,$46,$27,$EE,$C6,$31,$56,$EB,$55,$F7,$FE
	.dc.b	$1B,$AB,$F8,$CB,$FF,$75,$63,$D6,$73,$87,$E0,$9D,$D7,$C2,$CF,$DC
	.dc.b	$FC,$26,$BF,$0B,$83,$EB,$C6,$2B,$F3,$7F,$1B,$D3,$53,$49,$1F,$18
	.dc.b	$7C,$57,$E8,$FF,$0B,$FD,$0E,$3F,$A5,$8C,$4F,$E7,$1B,$19,$3F,$4F
	.dc.b	$3C,$1F,$21,$F6,$F8,$3E,$C3,$EB,$CB,$FD,$E1,$EC,$BF,$22,$79,$7F
	.dc.b	$AC,$CB,$F9,$9C,$E6,$31,$7F,$CE,$23,$13,$FB,$53,$75,$C9,$C4,$FF
	.dc.b	$3E,$E8,$7D,$BB,$EA,$BF,$A0,$FD,$D7,$9E,$73,$18,$BF,$D4,$56,$ED
	.dc.b	$7B,$8C,$3E,$F8,$F6,$1E,$FE,$1F,$AA,$E0,$FE,$9B,$EE,$FF,$2A,$F6
	.dc.b	$BF,$A8,$F0,$7C,$78,$C4,$A3,$14,$AD,$E3,$CA,$31,$5B,$8F,$DF,$76
	.dc.b	$BC,$8F,$E0,$71,$86,$CF,$C3,$9D,$3F,$C0,$A7,$4D,$3F,$8B,$79,$BF
	.dc.b	$41,$FC,$C3,$63,$F8,$BB,$7A,$7F,$D0,$51,$F4,$27,$F0,$27,$F0,$B3
	.dc.b	$E2,$71,$FF,$1C,$97,$A7,$C4,$95,$C4,$FE,$46,$F8,$6B,$E5,$1F,$A1
	.dc.b	$AF,$CF,$91,$EA,$43,$FB,$9F,$9D,$FE,$8F,$F8,$67,$A3,$FD,$FA,$30
	.dc.b	$DC,$BF,$C0,$53,$AA,$FF,$02,$D6,$9B,$3A,$AE,$1E,$57,$F8,$2A,$F8
	.dc.b	$57,$FA,$82,$B4,$F0,$9D,$57,$A4,$FE,$B4,$48,$FE,$60,$D7,$46,$08
	.dc.b	$7F,$49,$C7,$FA,$AF,$99,$FA,$FC,$FE,$38,$71,$A7,$CE,$8C,$1F,$C6
	.dc.b	$27,$8B,$D2,$BF,$85,$5F,$85,$A7,$F1,$EC,$76,$FE,$12,$79,$B7,$0E
	.dc.b	$F3,$A8,$73,$7F,$50,$DF,$FC,$5F,$EB,$FE,$BF,$83,$F2,$2D,$8C,$E3
	.dc.b	$F9,$28,$FD,$0F,$17,$F8,$DB,$CC,$B8,$9A,$4A,$F5,$47,$F0,$DD,$3D
	.dc.b	$4D,$3F,$77,$D1,$FE,$40,$FB,$7F,$DD,$36,$B3,$29,$CB,$63,$F8,$6F
	.dc.b	$67,$B7,$B3,$F8,$8F,$07,$F1,$F6,$CE,$8F,$8B,$5F,$28,$FB,$64,$30
	.dc.b	$7F,$83,$BB,$BF,$C2,$59,$FD,$B6,$CF,$F2,$F9,$FF,$09,$CE,$5B,$1F
	.dc.b	$D8,$3B,$FF,$07,$16,$3F,$04,$DC,$3F,$C2,$D3,$A8,$F1,$1F,$14,$BD
	.dc.b	$FF,$E3,$BD,$7F,$AA,$5F,$29,$FE,$B4,$D7,$FA,$D7,$83,$CD,$B1,$93
	.dc.b	$FC,$37,$E1,$FF,$8A,$BF,$7E,$FE,$C9,$06,$92,$7F,$B9,$47,$1F,$0F
	.dc.b	$FB,$D6,$7F,$DD,$E3,$1E,$37,$7A,$F4,$1F,$28,$A6,$21,$E4,$7F,$84
	.dc.b	$79,$7F,$C7,$2D,$9F,$F0,$C7,$93,$F9,$46,$89,$FC,$4E,$DF,$0E,$3F
	.dc.b	$1A,$D7,$F9,$57,$95,$C5,$A7,$13,$F9,$9B,$97,$F4,$7D,$3F,$E3,$1A
	.dc.b	$FF,$06,$79,$7F,$E4,$EE,$1B,$9F,$F0,$AB,$FF,$03,$4E,$5F,$F0,$7B
	.dc.b	$5F,$F2,$C6,$DF,$E5,$0D,$BF,$CA,$5E,$67,$D3,$BF,$F2,$84,$61,$FB
	.dc.b	$15,$F1,$B4,$F5,$3F,$25,$AF,$F7,$5F,$5F,$FC,$DC,$47,$FC,$88,$FF
	.dc.b	$9D,$88,$FF,$A8,$39,$7F,$D1,$11,$88,$FE,$6D,$EB,$7F,$9D,$C8,$FF
	.dc.b	$93,$5F,$F3,$F9,$1E,$5E,$87,$FA,$D2,$31,$1F,$E9,$FE,$3C,$3F,$EA
	.dc.b	$EF,$47,$FD,$3D,$3F,$EA,$58,$34,$FF,$58,$7A,$5C,$23,$FD,$7B,$A7
	.dc.b	$FA,$BE,$73,$1C,$0F,$F5,$9C,$ED,$FF,$61,$7A,$75,$FE,$D8,$DF,$07
	.dc.b	$FD,$8B,$5F,$EC,$28,$D7,$AA,$F5,$7E,$B5,$FF,$B3,$A3,$1C,$1F,$76
	.dc.b	$11,$8E,$4F,$F6,$4C,$62,$B8,$B6,$B8,$56,$E3,$FD,$9E,$D7,$FD,$B2
	.dc.b	$D7,$C1,$3C,$A3,$11,$CE,$BB,$C6,$09,$3B,$25,$78,$1B,$3C,$1B,$5E
	.dc.b	$26,$BE,$26,$BE,$29,$E4,$49,$E4,$D8,$E4,$D7,$93,$5E,$51,$C0,$8F
	.dc.b	$06,$BC,$1A,$F0,$34,$43,$45,$34,$60,$86,$08,$7C,$93,$81,$4D,$94
	.dc.b	$DD,$7C,$C9,$F5,$46,$49,$19,$9F,$3C,$62,$BE,$4B,$6B,$8D,$6E,$3C
	.dc.b	$2D,$7C,$2D,$7C,$33,$CA,$31,$1C,$EB,$BC,$60,$93,$B2,$57,$81,$B3
	.dc.b	$C1,$B5,$E2,$6B,$E2,$6B,$E2,$9E,$44,$9E,$4D,$8E,$4D,$79,$35,$E5
	.dc.b	$1C,$08,$F0,$6B,$C1,$AF,$03,$44,$34,$53,$46,$08,$60,$87,$C9,$38
	.dc.b	$14,$D9,$4D,$D7,$CC,$9F,$54,$64,$91,$99,$F3,$C6,$2B,$E4,$B6,$B8
	.dc.b	$D6,$E3,$C2,$D7,$C2,$D7,$C3,$3F,$8A,$F2,$9E,$33,$C9,$FE,$1F,$5F
	.dc.b	$E5,$76,$CE,$9B,$3B,$9F,$E0,$FE,$37,$FA,$7D,$FE,$38,$D7,$F8,$DB
	.dc.b	$70,$FF,$4C,$63,$D9,$B3,$CA,$31,$3B,$23,$82,$46,$C8,$7F,$20,$8D
	.dc.b	$B8,$3F,$90,$B8,$75,$1C,$9A,$6C,$F4,$B9,$78,$B5,$C9,$B3,$DA,$E5
	.dc.b	$A6,$67,$DA,$F6,$21,$D0,$D4,$F9,$5C,$1A,$3B,$94,$EA,$58,$EA,$F0
	.dc.b	$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$47,$A6,$38,$35
	.dc.b	$DB,$63,$73,$F1,$DB,$3F,$0A,$31,$3B,$23,$82,$46,$C8,$7C,$38,$DB
	.dc.b	$83,$DE,$E1,$D4,$72,$69,$B3,$D2,$E5,$E2,$D7,$26,$CF,$6B,$96,$99
	.dc.b	$9F,$6B,$D8,$87,$43,$53,$E5,$70,$68,$EE,$53,$A9,$63,$AB,$C0,$87
	.dc.b	$02,$9C,$0A,$F0,$74,$43,$45,$74,$E0,$8E,$09,$1E,$98,$E0,$D7,$6D
	.dc.b	$8D,$CF,$C7,$6C,$FC,$28,$C4,$EC,$8E,$09,$1B,$3E,$96,$6F,$A1,$47
	.dc.b	$F1,$52,$88,$D3,$93,$6E,$31,$FF,$17,$B6,$38,$8E,$67,$9A,$F3,$79
	.dc.b	$4E,$63,$17,$FE,$38,$AD,$D6,$9B,$1D,$BE,$C1,$97,$2C,$51,$89,$D4
	.dc.b	$E6,$31,$8F,$92,$FA,$BD,$F1,$F2,$37,$1F,$E7,$73,$A8,$F8,$7E,$5B
	.dc.b	$BF,$E2,$43,$D6,$F7,$CE,$74,$F9,$71,$8A,$DB,$63,$35,$F3,$B3,$79
	.dc.b	$CF,$8E,$FE,$89,$18,$6C,$66,$3F,$CB,$A6,$4A,$BC,$CF,$8B,$EB,$7D
	.dc.b	$29,$EB,$18,$BF,$05,$6E,$55,$B8,$C3,$F0,$E6,$1F,$27,$6F,$8E,$DD
	.dc.b	$BC,$3E,$F7,$58,$FC,$5E,$1E,$C8,$C4,$7D,$8C,$11,$89,$D1,$E2,$9E
	.dc.b	$87,$80,$FC,$9E,$FF,$4D,$B8,$78,$9F,$CC,$EF,$F5,$7E,$BB,$E6,$C6
	.dc.b	$38,$11,$88,$FF,$64,$B6,$78,$47,$DB,$3E,$FC,$BE,$17,$F4,$BF,$B5
	.dc.b	$7D,$B6,$CE,$9A,$F7,$6B,$FE,$D6,$F8,$8E,$D4,$69,$23,$C0,$7B,$36
	.dc.b	$FC,$8E,$5E,$D8,$FD,$AE,$7F,$05,$B1,$9C,$9F,$9F,$FE,$92,$79,$88
	.dc.b	$CE,$BE,$F9,$D7,$27,$E6,$1B,$5F,$C0,$3D,$37,$E0,$7F,$6A,$BA,$4A
	.dc.b	$72,$FE,$CA,$33,$18,$BF,$D2,$1C,$EB,$93,$F7,$2A,$1E,$F3,$EF,$1E
	.dc.b	$7B,$7C,$F7,$EA,$A7,$F6,$31,$89,$EA,$7E,$01,$C6,$1E,$6C,$7E,$1D
	.dc.b	$F1,$AF,$E3,$B8,$F9,$7E,$53,$CF,$97,$E5,$9B,$3F,$C3,$1B,$91,$F8
	.dc.b	$31,$FC,$5F,$5F,$AA,$47,$D3,$6F,$D5,$DE,$CF,$F1,$6D,$BE,$2C,$63
	.dc.b	$1F,$E4,$67,$CF,$71,$1A,$9E,$13,$FB,$67,$F7,$6D,$3F,$6E,$5E,$7F
	.dc.b	$AA,$FF,$0A,$8D,$E1,$F1,$7D,$BF,$84,$FC,$D6,$ED,$F8,$86,$F0,$1D
	.dc.b	$CF,$BC,$FA,$A5,$3E,$C9,$F5,$F1,$E2,$47,$EA,$99,$29,$C4,$FC,$01
	.dc.b	$C8,$C1,$F9,$CD,$3E,$63,$9E,$1F,$D0,$3F,$E6,$D3,$CD,$E2,$7F,$56
	.dc.b	$F0,$FF,$CD,$D1,$83,$FA,$DB,$8C,$13,$BF,$0F,$FC,$E5,$18,$AD,$C6
	.dc.b	$27,$5E,$07,$F6,$4A,$D5,$78,$5B,$7D,$6B,$F7,$07,$F4,$92,$4F,$F2
	.dc.b	$56,$BA,$30,$43,$F3,$59,$7F,$72,$F4,$7F,$35,$A7,$F6,$89,$46,$27
	.dc.b	$5B,$FE,$91,$B3,$D4,$87,$CB,$BF,$0B,$F8,$13,$9C,$FE,$B3,$1E,$31
	.dc.b	$9B,$F0,$C6,$27,$F1,$83,$C9,$FC,$7F,$77,$E6,$DB,$C3,$FD,$D1,$B3
	.dc.b	$B3,$F3,$D1,$FD,$E3,$BB,$F3,$D1,$FD,$08,$C9,$71,$FD,$BF,$27,$E9
	.dc.b	$E7,$F6,$3C,$3F,$E0,$4D,$3D,$75,$FB,$37,$DB,$CB,$FC,$47,$E2,$FF
	.dc.b	$72,$4C,$7F,$89,$57,$F8,$22,$3E,$6F,$4F,$C0,$E6,$FE,$0F,$3F,$10
	.dc.b	$FE,$19,$19,$E2,$FE,$21,$C3,$E2,$E2,$FF,$0D,$67,$E2,$D3,$C9,$87
	.dc.b	$DB,$3F,$93,$74,$FF,$62,$BF,$E1,$64,$D3,$C6,$77,$9F,$E4,$0F,$E0
	.dc.b	$5F,$E3,$ED,$7F,$8F,$35,$D9,$F8,$43,$C0,$7C,$32,$D7,$91,$FE,$C5
	.dc.b	$D7,$C7,$4F,$F0,$D3,$7B,$3F,$AA,$4F,$33,$65,$3F,$A5,$3F,$D8,$D7
	.dc.b	$E4,$7F,$D8,$DA,$7F,$63,$29,$FD,$8C,$A7,$F7,$03,$A5,$7F,$70,$C7
	.dc.b	$FC,$4D,$E0,$78,$CF,$74,$FA,$CF,$CD,$94,$FC,$E1,$FE,$1A,$E2,$FC
	.dc.b	$83,$73,$FC,$B3,$79,$7F,$C3,$93,$A8,$E7,$A7,$C5,$79,$9F,$A8,$AF
	.dc.b	$F1,$39,$C9,$FF,$0D,$F3,$7F,$89,$EF,$FC,$4D,$E0,$F3,$9F,$F0,$E3
	.dc.b	$6B,$FC,$69,$C1,$FE,$2B,$F0,$BF,$CA,$ED,$8F,$E1,$75,$FE,$23,$9C
	.dc.b	$BF,$E1,$0F,$0F,$EA,$11,$F8,$FC,$1F,$E5,$0D,$BF,$21,$C7,$F9,$29
	.dc.b	$FB,$ED,$7F,$CF,$35,$FE,$64,$C1,$E4,$FF,$99,$09,$3F,$E7,$69,$D4
	.dc.b	$7F,$98,$49,$1F,$E3,$D6,$BD,$3C,$6F,$F4,$3C,$7F,$A0,$1B,$1C,$DA
	.dc.b	$EC,$F0,$9F,$E8,$4F,$21,$E4,$7C,$08,$FF,$47,$C6,$30,$F8,$B0,$8C
	.dc.b	$6B,$FE,$8C,$8C,$57,$17,$FD,$33,$3A,$7B,$46,$5D,$D7,$49,$D4,$78
	.dc.b	$27,$51,$88,$E7,$18,$24,$F1,$6D,$78,$EB,$85,$76,$8C,$4E,$A3,$C7
	.dc.b	$5C,$49,$3A,$6C,$6A,$30,$47,$04,$3D,$11,$C1,$B1,$B6,$9B,$3D,$71
	.dc.b	$B6,$C6,$DA,$6C,$F5,$16,$BD,$A4,$AF,$64,$76,$21,$C8,$A7,$22,$C7
	.dc.b	$1A,$F1,$46,$4E,$04,$30,$48,$E1,$5E,$78,$C5,$78,$9E,$13,$A7,$D1
	.dc.b	$19,$77,$5E,$09,$D4,$79,$27,$51,$88,$E7,$18,$24,$F1,$6D,$78,$EB
	.dc.b	$85,$76,$8C,$4E,$A3,$C7,$5C,$49,$3A,$6C,$6A,$30,$47,$04,$3D,$11
	.dc.b	$C1,$B1,$B6,$9B,$3D,$71,$B6,$C6,$DA,$6C,$F5,$16,$BD,$A4,$AF,$64
	.dc.b	$76,$21,$C8,$A7,$22,$C7,$1A,$F1,$46,$4E,$04,$30,$48,$E1,$5E,$78
	.dc.b	$C5,$78,$9E,$13,$A7,$D1,$19,$77,$5E,$09,$D4,$79,$27,$51,$88,$E7
	.dc.b	$18,$24,$F1,$7F,$86,$77,$FF,$94,$24,$7F,$92,$0F,$E2,$CE,$5C,$1A
	.dc.b	$3F,$C9,$9D,$9F,$E8,$C6,$D7,$F1,$D7,$FC,$E5,$19,$6C,$67,$BF,$FD
	.dc.b	$64,$D9,$D4,$62,$7F,$93,$35,$F0,$1C,$8C,$C7,$02,$4F,$32,$46,$9A
	.dc.b	$F2,$9E,$8F,$AD,$AF,$A8,$E0,$43,$84,$F6,$70,$71,$70,$EA,$7B,$B9
	.dc.b	$6B,$D4,$D1,$C8,$C1,$B3,$07,$BC,$D9,$0D,$14,$E2,$43,$81,$A2,$19
	.dc.b	$9F,$7B,$A3,$13,$EA,$76,$64,$91,$C0,$91,$B8,$E3,$3E,$E7,$71,$96
	.dc.b	$C6,$67,$8B,$67,$51,$89,$F8,$2D,$7D,$C7,$23,$31,$C0,$93,$CC,$91
	.dc.b	$A6,$BC,$A7,$A3,$EE,$6B,$EA,$38,$10,$E1,$3D,$9C,$1C,$5C,$3A,$9E
	.dc.b	$EE,$5A,$F5,$34,$72,$30,$6C,$C1,$EF,$36,$43,$45,$38,$90,$E0,$68
	.dc.b	$86,$67,$DE,$E8,$C4,$FA,$9D,$99,$24,$70,$24,$6E,$38,$CF,$B9,$DC
	.dc.b	$65,$B1,$99,$E2,$D9,$D4,$62,$7E,$0B,$5F,$71,$C8,$CC,$70,$24,$F3
	.dc.b	$24,$69,$FC,$A7,$C7,$BE,$9B,$F8,$FF,$85,$FE,$33,$8C,$52,$31,$3A
	.dc.b	$8F,$E2,$33,$A9,$FF,$1D,$73,$B8,$5E,$27,$1B,$FE,$2F,$39,$8C,$47
	.dc.b	$F2,$98,$C5,$72,$AE,$2D,$AF,$E5,$53,$C6,$B0,$43,$F9,$96,$2F,$BB
	.dc.b	$72,$FF,$3A,$F2,$3F,$93,$CE,$63,$15,$C2,$B6,$AB,$71,$87,$8F,$C2
	.dc.b	$FB,$2F,$FC,$B0,$1C,$56,$B1,$FE,$85,$39,$6C,$BC,$F4,$BC,$9E,$B1
	.dc.b	$86,$C6,$6F,$DB,$5C,$63,$11,$F2,$75,$F9,$EF,$C3,$31,$87,$C8,$8C
	.dc.b	$C7,$C6,$F7,$DD,$3D,$E3,$0D,$7F,$B0,$DA,$76,$F8,$23,$D3,$18,$C7
	.dc.b	$C1,$1F,$DA,$36,$FA,$CF,$E0,$30,$F0,$BE,$43,$E6,$FC,$97,$D2,$7B
	.dc.b	$EC,$8C,$47,$C6,$8C,$3F,$54,$FB,$11,$AC,$7E,$97,$D7,$7C,$CF,$C8
	.dc.b	$FE,$BB,$6B,$E4,$B5,$F6,$EC,$8C,$43,$C4,$79,$1F,$A2,$E2,$FE,$DF
	.dc.b	$18,$E6,$F2,$B6,$BC,$87,$F8,$21,$C4,$FC,$48,$6D,$F1,$1F,$84,$7C
	.dc.b	$5E,$2F,$C5,$EC,$FB,$8F,$63,$67,$A9,$F0,$E3,$27,$C1,$FC,$A5,$F0
	.dc.b	$5F,$E1,$8E,$2F,$C6,$D7,$DA,$9C,$BF,$B5,$E1,$FD,$3C,$62,$77,$3F
	.dc.b	$94,$8C,$4F,$96,$3D,$59,$7A,$5F,$BD,$5F,$71,$75,$3E,$11,$F4,$1F
	.dc.b	$B1,$C3,$FC,$4D,$39,$8C,$57,$F8,$BA,$30,$FE,$C6,$77,$5C,$A3,$07
	.dc.b	$F8,$D3,$E2,$5F,$4A,$F9,$9F,$D5,$FC,$5F,$AA,$73,$18,$C3,$FC,$6D
	.dc.b	$18,$53,$BA,$E5,$18,$3E,$05,$C3,$EB,$90,$F7,$FA,$8F,$B3,$FD,$DF
	.dc.b	$27,$CA,$D1,$39,$BF,$4D,$F0,$BD,$C7,$C4,$9D,$5F,$D2,$8F,$E1,$9F
	.dc.b	$07,$FC,$0D,$FE,$02,$DB,$FC,$CB,$5F,$E1,$7D,$48,$C7,$3F,$E6,$BE
	.dc.b	$B7,$E7,$D7,$E7,$7A,$91,$89,$D4,$F9,$9B,$8F,$E1,$23,$F3,$3C,$7F
	.dc.b	$7D,$BD,$3F,$C3,$48,$FE,$9B,$07,$D7,$75,$6C,$75,$6D,$FE,$43,$0F
	.dc.b	$D7,$D7,$9A,$BA,$9F,$90,$C3,$C1,$CD,$F9,$18,$C9,$F5,$4F,$C9,$E3
	.dc.b	$E2,$7F,$CC,$BE,$3B,$DA,$FE,$A5,$E4,$3C,$BF,$F3,$E3,$6B,$FC,$FF
	.dc.b	$5B,$F2,$3F,$D0,$33,$99,$4E,$54,$66,$30,$DB,$EF,$7D,$89,$5E,$67
	.dc.b	$FA,$E1,$1F,$E0,$4D,$7F,$9C,$B5,$E0,$7F,$A3,$F4,$F8,$EB,$CE,$F9
	.dc.b	$5B,$FC,$E6,$ED,$FE,$9A,$6C,$EF,$8F,$BB,$0E,$E5,$3F,$0D,$C3,$C6
	.dc.b	$3C,$17,$F8,$89,$E4,$7F,$1F,$71,$5E,$2A,$E1,$3F,$96,$A7,$A3,$F9
	.dc.b	$DE,$4F,$CB,$46,$0F,$CE,$96,$78,$4F,$EA,$F9,$7F,$55,$85,$E6,$FE
	.dc.b	$D2,$DB,$F1,$DF,$EB,$5A,$FE,$5B,$5F,$94,$D3,$DD,$D5,$FC,$0F,$E1
	.dc.b	$7F,$81,$76,$7F,$74,$6D,$9A,$7E,$F7,$B7,$F8,$A6,$37,$BB,$DF,$8F
	.dc.b	$F2,$29,$CD,$FF,$1A,$9C,$CF,$84,$F9,$E4,$3E,$E9,$FC,$DF,$C0,$FF
	.dc.b	$68,$FA,$9F,$C0,$FC,$1F,$E7,$2D,$8C,$DF,$F3,$A6,$D7,$F6,$AF,$0B
	.dc.b	$F0,$0F,$F8,$75,$AE,$4F,$C4,$19,$29,$E3,$3C,$C7,$90,$9C,$DF,$D7
	.dc.b	$F3,$FD,$F7,$89,$D2,$7F,$86,$36,$75,$1F,$DC,$30,$7D,$DF,$F7,$46
	.dc.b	$C6,$63,$07,$F8,$11,$F1,$9F,$E0,$1F,$23,$B4,$FF,$8F,$5B,$E3,$7F
	.dc.b	$81,$31,$FB,$BE,$07,$F9,$37,$92,$F8,$3F,$E5,$27,$FC,$83,$E5,$7F
	.dc.b	$86,$BA,$6F,$97,$FE,$32,$6E,$7F,$96,$D3,$FC,$46,$D7,$E2,$63,$D5
	.dc.b	$FD,$31,$F1,$A3,$FC,$85,$E6,$73,$8F,$F1,$BB,$5F,$F0,$7B,$5F,$E6
	.dc.b	$ED,$8F,$F2,$31,$FC,$3E,$3F,$9D,$F8,$9F,$E4,$6C,$BF,$C9,$93,$93
	.dc.b	$6F,$F9,$5B,$CD,$F8,$64,$AF,$93,$3F,$E5,$F7,$E5,$9F,$70,$98,$7A
	.dc.b	$DF,$6E,$9F,$E7,$8F,$4B,$FC,$E0,$48,$FF,$0B,$3F,$E8,$24,$D2,$3F
	.dc.b	$E8,$4E,$33,$A2,$9F,$CB,$3C,$9F,$63,$83,$F9,$5B,$87,$7E,$2F,$F8
	.dc.b	$33,$1F,$89,$1A,$6C,$67,$3E,$53,$96,$C6,$63,$B3,$5E,$B5,$DD,$F2
	.dc.b	$57,$48,$C4,$73,$38,$B8,$9E,$51,$C2,$BB,$CE,$49,$5C,$C9,$1A,$6C
	.dc.b	$F4,$6B,$E4,$6B,$D4,$D4,$71,$23,$82,$1C,$5C,$B4,$C9,$CC,$C1,$0C
	.dc.b	$10,$F8,$06,$4A,$66,$BE,$0B,$82,$38,$25,$7C,$29,$F1,$CF,$18,$C3
	.dc.b	$E4,$38,$B8,$7C,$13,$CA,$BE,$1C,$69,$B1,$9A,$EB,$39,$6C,$66,$3B
	.dc.b	$35,$EB,$5D,$DF,$25,$74,$8C,$47,$33,$8B,$89,$E5,$1C,$2B,$BC,$E4
	.dc.b	$95,$CC,$91,$A6,$CF,$46,$BE,$46,$BD,$4D,$47,$12,$38,$21,$C5,$CB
	.dc.b	$4C,$9C,$CC,$10,$C1,$0F,$80,$64,$A6,$6B,$E0,$B8,$23,$82,$57,$C2
	.dc.b	$9F,$1C,$F1,$8C,$3E,$43,$8B,$87,$C1,$3C,$AB,$E1,$C6,$9B,$19,$AE
	.dc.b	$B3,$96,$C6,$63,$B3,$5E,$B5,$DD,$F2,$57,$48,$C4,$73,$38,$B8,$9E
	.dc.b	$51,$FE,$48,$F1,$5E,$3F,$F1,$28,$FF,$23,$C6,$1F,$E2,$BE,$3F,$F9
	.dc.b	$25,$FF,$2A,$C6,$5F,$E4,$5F,$7A,$FC,$6B,$FC,$E1,$1F,$CA,$A3,$19
	.dc.b	$75,$6C,$7F,$20,$6B,$CA,$78,$A8,$CB,$FC,$C9,$AF,$58,$DC,$62,$38
	.dc.b	$46,$09,$3E,$16,$C7,$85,$AF,$86,$34,$47,$4D,$74,$60,$86,$08,$79
	.dc.b	$E3,$AB,$63,$93,$4E,$25,$38,$9E,$D7,$6D,$36,$53,$67,$C0,$76,$D3
	.dc.b	$65,$37,$3F,$1C,$93,$E6,$8E,$87,$8C,$86,$08,$73,$32,$53,$31,$D2
	.dc.b	$7C,$6D,$8F,$1B,$5F,$1C,$F8,$5E,$44,$8D,$3D,$9A,$F6,$8E,$11,$88
	.dc.b	$E1,$18,$24,$F8,$9B,$1E,$26,$BE,$28,$D1,$1D,$35,$D1,$82,$18,$21
	.dc.b	$E7,$8E,$AD,$8E,$4D,$38,$94,$E2,$7B,$5D,$B4,$D9,$4D,$9F,$01,$DB
	.dc.b	$4D,$94,$DC,$FC,$72,$4F,$9A,$3A,$1E,$32,$18,$21,$CC,$C9,$4C,$C7
	.dc.b	$49,$F1,$B6,$3C,$6D,$7C,$73,$E1,$79,$12,$34,$F6,$6B,$DA,$38,$46
	.dc.b	$23,$84,$60,$93,$E2,$6C,$78,$9A,$F8,$A3,$F5,$3B,$B7,$70,$FF,$1F
	.dc.b	$AF,$F7,$BE,$7F,$E3,$8A,$D4,$E6,$51,$89,$FF,$1E,$FA,$2F,$B6,$63
	.dc.b	$B8,$C3,$63,$33,$C1,$B7,$FE,$4F,$A7,$1E,$23,$FC,$AB,$18,$BE,$93
	.dc.b	$99,$FF,$2D,$CE,$6F,$F9,$87,$DA,$66,$7F,$97,$37,$53,$4D,$65,$EE
	.dc.b	$9F,$81,$A7,$C4,$6E,$9F,$D2,$67,$51,$E7,$D3,$F9,$DE,$5E,$E8,$F6
	.dc.b	$69,$FD,$3E,$31,$1F,$E6,$BA,$E2,$FF,$9C,$1A,$FF,$9D,$1A,$FC,$EC
	.dc.b	$3C,$4F,$F9,$CF,$F1,$A7,$4F,$B5,$B8,$7D,$56,$EB,$FE,$70,$6D,$7F
	.dc.b	$9F,$BB,$9E,$63,$EF,$FF,$D0,$7F,$84,$FF,$9F,$E7,$31,$8A,$E1,$5B
	.dc.b	$55,$B8,$C3,$FE,$87,$C5,$E6,$BE,$2F,$FB,$3C,$61,$B1,$98,$FF,$42
	.dc.b	$99,$3E,$76,$BF,$03,$0F,$8C,$D7,$E3,$35,$F8,$DC,$3F,$DF,$74,$F8
	.dc.b	$71,$F0,$76,$FE,$DF,$18,$8F,$95,$1E,$4F,$92,$7C,$4F,$FA,$93,$AF
	.dc.b	$F3,$35,$FE,$A6,$C9,$5A,$8F,$F5,$01,$97,$13,$C2,$3F,$43,$18,$9F
	.dc.b	$C0,$65,$E7,$E2,$FB,$A4,$71,$D7,$FA,$49,$FC,$33,$6F,$F0,$CD,$7F
	.dc.b	$0C,$D8,$F2,$E3,$E7,$7F,$D7,$66,$27,$F6,$6E,$30,$F5,$3E,$73,$47
	.dc.b	$A6,$3E,$59,$FA,$02,$1E,$AD,$3E,$BF,$C6,$BE,$7F,$AD,$78,$23,$0F
	.dc.b	$82,$BE,$54,$62,$3C,$0D,$7C,$FB,$FD,$27,$E8,$4F,$D1,$8C,$3E,$76
	.dc.b	$C7,$0C,$7E,$BB,$ED,$F7,$BE,$0D,$7D,$56,$BF,$51,$B1,$F7,$1C,$C7
	.dc.b	$C0,$3D,$C7,$C0,$3F,$23,$8F,$E5,$AF,$C3,$1E,$5B,$F0,$B6,$FD,$2F
	.dc.b	$F2,$29,$DB,$F9,$8A,$FC,$E9,$1E,$C6,$4A,$78,$CF,$09,$FE,$56,$ED
	.dc.b	$FF,$2C,$68,$EF,$FF,$7C,$F7,$A7,$31,$86,$D7,$F9,$8F,$A7,$FE,$00
	.dc.b	$DF,$18,$C0,$70,$FF,$99,$A3,$75,$CD,$B1,$C2,$77,$C3,$FD,$1A,$BF
	.dc.b	$93,$92,$7F,$10,$D7,$E5,$B5,$F8,$8F,$F0,$9C,$BF,$04,$FD,$FD,$3F
	.dc.b	$A5,$B7,$4F,$80,$7E,$04,$B5,$D0,$EE,$43,$07,$E0,$CD,$9F,$83,$2B
	.dc.b	$F8,$36,$F0,$36,$46,$B0,$F2,$D7,$86,$FC,$84,$7F,$45,$C1,$FD,$8B
	.dc.b	$8F,$F3,$ED,$8C,$9F,$C4,$C7,$0F,$EA,$27,$F6,$5A,$7C,$BC,$7D,$47
	.dc.b	$C6,$2D,$7E,$F6,$7F,$75,$9F,$F1,$56,$E9,$E8,$6D,$7F,$21,$3F,$BC
	.dc.b	$EB,$FD,$DB,$C1,$EE,$DC,$3F,$88,$88,$CB,$4F,$1C,$FF,$30,$D3,$ED
	.dc.b	$34,$FE,$1D,$87,$C8,$F3,$BC,$24,$AC,$CA,$3F,$88,$73,$7F,$5A,$BF
	.dc.b	$E7,$15,$F9,$2A,$FA,$33,$97,$F9,$CC,$65,$F0,$1F,$78,$F8,$C5,$7F
	.dc.b	$A6,$4F,$E3,$79,$FF,$17,$3F,$C4,$1A,$FF,$0F,$6A,$EB,$FE,$C6,$BE
	.dc.b	$77,$FB,$07,$67,$E1,$1B,$19,$3F,$B3,$C6,$5F,$3E,$9E,$C2,$38,$3E
	.dc.b	$5F,$07,$DF,$7D,$47,$B0,$FF,$0E,$F7,$79,$17,$77,$F8,$B3,$3F,$F0
	.dc.b	$79,$23,$9F,$5F,$F8,$F2,$3E,$AD,$F5,$9F,$F6,$C7,$7F,$E0,$5F,$F6
	.dc.b	$EB,$5F,$F6,$93,$5D,$ED,$F8,$37,$FC,$6D,$E0,$E5,$F1,$B7,$1F,$F1
	.dc.b	$B6,$7F,$E4,$48,$C6,$72,$F9,$5F,$F8,$F2,$31,$5F,$E2,$E2,$47,$F4
	.dc.b	$F6,$BF,$EC,$DC,$BF,$DA,$E4,$7F,$DB,$8D,$74,$60,$86,$08,$7F,$B4
	.dc.b	$5E,$0D,$36,$53,$67,$FB,$45,$DB,$4D,$94,$DF,$82,$F0,$BE,$C3,$70
	.dc.b	$FC,$C3,$6B,$9B,$6B,$9B,$5E,$75,$C8,$95,$C9,$B5,$C9,$B3,$C9,$AF
	.dc.b	$29,$E0,$49,$E0,$D8,$E0,$D7,$84,$68,$8E,$9A,$E8,$C1,$0C,$10,$F9
	.dc.b	$84,$3E,$41,$6B,$E0,$36,$BD,$2D,$AF,$23,$6B,$C2,$D3,$C0,$53,$C0
	.dc.b	$76,$25,$76,$6D,$76,$6D,$76,$6D,$76,$69,$D8,$A7,$6A,$E8,$4A,$E8
	.dc.b	$DA,$E8,$DA,$E8,$DA,$E8,$D7,$A5,$72,$25,$72,$6D,$72,$6C,$F2,$6B
	.dc.b	$CA,$78,$12,$78,$36,$38,$35,$E1,$1A,$23,$A6,$BA,$30,$43,$04,$3E
	.dc.b	$A1,$0F,$92,$5A,$F8,$0D,$AF,$4B,$6B,$C8,$DA,$F0,$B4,$F0,$14,$F0
	.dc.b	$1D,$89,$5D,$9F,$F0,$F1,$FE,$2D,$EC,$FE,$FE,$C3,$25,$EE,$FF,$17
	.dc.b	$46,$63,$FC,$0B,$2F,$33,$FC,$D5,$18,$6A,$8D,$61,$C6,$31,$3A,$9C
	.dc.b	$C7,$56,$BE,$36,$BE,$66,$BD,$E7,$A4,$62,$3A,$C6,$27,$B1,$1C,$12
	.dc.b	$3C,$11,$96,$C6,$5F,$14,$65,$B1,$97,$C9,$19,$6C,$65,$F3,$1D,$08
	.dc.b	$72,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$87,$C4,$3A,$94,$E4
	.dc.b	$53,$89,$4E,$33,$F3,$49,$3E,$F8,$CB,$5E,$04,$8D,$CF,$C3,$8C,$4E
	.dc.b	$E7,$31,$E4,$6B,$E4,$6B,$E4,$6B,$E0,$9E,$91,$88,$EB,$18,$9E,$C4
	.dc.b	$70,$48,$F0,$46,$5B,$19,$7C,$51,$96,$C6,$5F,$24,$65,$B1,$97,$CC
	.dc.b	$74,$21,$C8,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$10,$EA
	.dc.b	$53,$91,$4E,$25,$38,$CF,$CD,$24,$FB,$E3,$2D,$78,$12,$37,$3F,$0E
	.dc.b	$31,$3B,$9C,$C7,$91,$AF,$91,$AF,$91,$AF,$82,$7A,$46,$23,$AC,$62
	.dc.b	$7B,$11,$C1,$23,$C1,$1F,$99,$D5,$F3,$BF,$EA,$2C,$3F,$90,$46,$29
	.dc.b	$18,$9D,$47,$F2,$99,$D4,$FF,$27,$E0,$70,$BC,$1E,$37,$FC,$BE,$73
	.dc.b	$18,$8F,$F7,$8D,$71,$AE,$0F,$8B,$CD,$67,$BA,$FF,$5B,$E2,$9C,$D6
	.dc.b	$EB,$55,$FE,$6A,$9C,$BF,$E5,$C8,$CD,$38,$B3,$3F,$D1,$23,$13,$A9
	.dc.b	$CC,$F8,$33,$FE,$78,$FF,$9D,$9B,$5F,$1E,$3E,$31,$FF,$00,$4F,$81
	.dc.b	$8F,$C3,$7F,$E0,$0E,$7F,$09,$B1,$97,$79,$F1,$7F,$AB,$CE,$DC,$9C
	.dc.b	$23,$FA,$B7,$33,$C5,$F6,$4B,$C4,$FE,$BD,$39,$CF,$FD,$BF,$18,$AD
	.dc.b	$B6,$33,$5F,$D9,$B4,$78,$5F,$EB,$B5,$F9,$F7,$FB,$EF,$B7,$EA,$B6
	.dc.b	$33,$5C,$6B,$84,$F9,$E2,$1E,$A2,$1E,$ED,$7E,$13,$74,$FE,$E5,$EA
	.dc.b	$3E,$6F,$F0,$72,$FB,$3A,$A7,$59,$7F,$BF,$27,$31,$89,$E1,$3F,$E0
	.dc.b	$28,$C4,$FF,$BD,$30,$F0,$D0,$70,$7D,$17,$C8,$7A,$8A,$7A,$8E,$0E
	.dc.b	$DF,$26,$1E,$8C,$7C,$6D,$C7,$C4,$D9,$F0,$63,$E7,$AF,$21,$F8,$38
	.dc.b	$EB,$87,$99,$F2,$B7,$EE,$9F,$75,$FA,$8E,$0F,$8B,$49,$CE,$5F,$85
	.dc.b	$3F,$C4,$4E,$27,$6D,$8F,$B5,$F6,$3F,$41,$A7,$C5,$2D,$7D,$27,$F0
	.dc.b	$F1,$B8,$F9,$5C,$7F,$95,$EE,$FC,$64,$63,$91,$18,$8F,$F7,$A4,$61
	.dc.b	$EC,$6E,$34,$FC,$0D,$3E,$23,$4F,$E1,$4F,$F8,$E7,$0F,$19,$CA,$3F
	.dc.b	$85,$F1,$31,$F3,$46,$1B,$3C,$23,$E8,$9F,$7C,$FA,$35,$E0,$E0,$FB
	.dc.b	$D1,$E1,$3F,$04,$7F,$94,$3C,$57,$6F,$F9,$43,$B7,$FC,$9B,$29,$CB
	.dc.b	$F8,$F8,$CC,$62,$39,$4F,$1F,$2F,$F9,$5C,$E6,$B7,$39,$EC,$FC,$C4
	.dc.b	$E6,$B7,$3F,$AB,$3E,$29,$2F,$F8,$7B,$5D,$18,$21,$F7,$F4,$F9,$19
	.dc.b	$FB,$0F,$C1,$CF,$E0,$63,$2D,$D3,$F9,$CB,$F9,$A3,$B5,$7C,$A6,$EF
	.dc.b	$F8,$58,$E8,$7D,$F2,$9F,$84,$DB,$F6,$93,$FC,$02,$78,$3D,$F0,$D9
	.dc.b	$F8,$07,$04,$7F,$3F,$C5,$FD,$33,$A3,$F3,$ED,$8C,$9F,$C7,$F0,$FF
	.dc.b	$43,$F1,$7E,$A2,$BD,$51,$F1,$0F,$51,$79,$7F,$87,$7A,$7F,$BE,$BF
	.dc.b	$E1,$4D,$93,$4E,$6F,$D0,$21,$D0,$A7,$E7,$F1,$F9,$2D,$3F,$84,$CF
	.dc.b	$F1,$7F,$B5,$EC,$47,$30,$8E,$EF,$52,$1C,$CE,$FC,$7F,$DD,$AB,$BB
	.dc.b	$63,$F9,$E6,$3F,$DD,$3F,$41,$FE,$80,$48,$CE,$08,$FE,$99,$09,$A4
	.dc.b	$7F,$D6,$7A,$FF,$55,$74,$7C,$D3,$C8,$79,$AF,$C4,$F9,$BA,$3F,$C1
	.dc.b	$D8,$F9,$9B,$1C,$67,$6F,$18,$CB,$FD,$FF,$BB,$F1,$6F,$F8,$6D,$AE
	.dc.b	$4F,$D9,$1F,$E0,$12,$8E,$EF,$F0,$3F,$87,$FE,$28,$EF,$FF,$0B,$B7
	.dc.b	$53,$4F,$F0,$B1,$2B,$FC,$55,$E0,$70,$6C,$F0,$7F,$C0,$73,$FC,$70
	.dc.b	$8E,$0F,$41,$B3,$D0,$FF,$84,$88,$FF,$AE,$B3,$FF,$5F,$91,$C1,$0F
	.dc.b	$F5,$DB,$96,$99,$3F,$DA,$4E,$5A,$64,$FF,$69,$B9,$69,$9F,$0B,$FD
	.dc.b	$CB,$19,$6C,$67,$C2,$FE,$0E,$DB,$F9,$F3,$A8,$F1,$90,$F0,$94,$EE
	.dc.b	$57,$83,$A2,$1A,$2B,$A7,$04,$70,$49,$F2,$C7,$56,$BC,$9A,$F1,$6C
	.dc.b	$71,$9F,$5C,$60,$8E,$09,$5F,$16,$32,$D8,$CD,$7C,$58,$C1,$1C,$10
	.dc.b	$F3,$47,$06,$C6,$DA,$6C,$F5,$46,$DB,$1B,$69,$B3,$DD,$1B,$6C,$6D
	.dc.b	$A6,$EB,$E1,$46,$5B,$19,$AF,$7B,$6B,$A4,$EA,$3C,$E4,$3C,$65,$3B
	.dc.b	$95,$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$7C,$B1,$D5,$AF,$26,$BC,$5B
	.dc.b	$1C,$67,$D7,$18,$23,$82,$57,$C5,$8C,$B6,$33,$5F,$16,$30,$47,$04
	.dc.b	$3C,$D1,$C1,$B1,$B6,$9B,$3D,$51,$B6,$C6,$DA,$6C,$F7,$46,$DF,$D4
	.dc.b	$F6,$7F,$87,$9F,$F3,$51,$FE,$1D,$EA,$FF,$5A,$A8,$D3,$FE,$19,$36
	.dc.b	$7F,$9C,$70,$4F,$F9,$6A,$30,$D5,$1A,$AE,$31,$A6,$BA,$9E,$13,$B8
	.dc.b	$DB,$5F,$0B,$5F,$4B,$5E,$87,$08,$E6,$49,$EF,$3B,$9F,$04,$62,$75
	.dc.b	$1C,$63,$04,$70,$43,$C7,$1C,$1B,$1B,$69,$B3,$CF,$1B,$6C,$6D,$A6
	.dc.b	$CF,$59,$C8,$86,$8A,$68,$C1,$0C,$10,$F7,$9D,$4A,$72,$29,$C4,$A7
	.dc.b	$19,$F9,$04,$9F,$54,$65,$AF,$02,$46,$E7,$DD,$1B,$6B,$A9,$E3,$3B
	.dc.b	$8F,$33,$5F,$33,$5F,$33,$5E,$87,$08,$E6,$49,$EF,$3B,$9F,$04,$62
	.dc.b	$75,$1C,$63,$04,$70,$43,$C7,$1C,$1B,$1B,$69,$B3,$CF,$1B,$6C,$6D
	.dc.b	$A6,$CF,$59,$C8,$86,$8A,$68,$C1,$0C,$10,$F7,$9D,$4A,$72,$29,$C4
	.dc.b	$A7,$19,$F9,$04,$9F,$54,$65,$AF,$02,$46,$E7,$DD,$1B,$6B,$A9,$E3
	.dc.b	$3B,$8F,$33,$5F,$33,$5F,$33,$5E,$87,$08,$E6,$49,$EF,$3B,$9F,$04
	.dc.b	$62,$75,$1F,$A5,$D1,$F0,$E8,$76,$FF,$1A,$9C,$C6,$0F,$E4,$AE,$23
	.dc.b	$53,$C5,$E1,$F7,$96,$B6,$D7,$F9,$9D,$23,$11,$E1,$8C,$3D,$CF,$E7
	.dc.b	$3C,$4E,$27,$7E,$67,$FD,$C7,$87,$FB,$C0,$8F,$9B,$B1,$A6,$E3,$FC
	.dc.b	$D1,$18,$9F,$73,$63,$84,$EF,$14,$F2,$F1,$37,$26,$92,$BF,$A7,$C3
	.dc.b	$BD,$F3,$DC,$67,$31,$89,$E5,$18,$9D,$3F,$ED,$F6,$BF,$F0,$00,$FE
	.dc.b	$9B,$A7,$FC,$00,$26,$9F,$12,$3F,$AC,$F4,$3C,$2F,$F7,$EF,$D2,$E1
	.dc.b	$B3,$FE,$A0,$6B,$F8,$19,$E1,$CB,$F8,$3A,$E1,$DD,$F8,$5F,$9B,$B6
	.dc.b	$69,$1E,$9E,$F7,$DC,$C3,$C4,$DC,$4D,$3F,$03,$2C,$FD,$7C,$7F,$DA
	.dc.b	$1B,$1F,$DD,$36,$F9,$9E,$CF,$F7,$36,$9F,$DC,$0B,$1B,$E0,$4F,$F7
	.dc.b	$9C,$BE,$27,$DB,$FB,$6F,$E1,$36,$FE,$16,$68,$C4,$2B,$8A,$DB,$F1
	.dc.b	$AF,$E5,$B8,$3E,$93,$FE,$0A,$CF,$D2,$D8,$CB,$DB,$B7,$AD,$FD,$76
	.dc.b	$E2,$69,$8C,$7F,$6D,$E4,$78,$2F,$C9,$5E,$75,$19,$BF,$3B,$67,$4F
	.dc.b	$19,$FA,$18,$79,$1F,$C7,$1E,$7D,$BE,$94,$66,$FF,$19,$E0,$7F,$07
	.dc.b	$9E,$8D,$D7,$F1,$8D,$BF,$96,$D7,$CF,$F3,$7F,$A6,$7E,$7B,$5C,$9F
	.dc.b	$C4,$E7,$F2,$D1,$9B,$FC,$E7,$D3,$DC,$CD,$25,$F1,$3E,$A6,$3E,$A2
	.dc.b	$3F,$AA,$CB,$D9,$4C,$3A,$12,$79,$1E,$43,$E4,$6D,$F7,$1E,$A7,$3B
	.dc.b	$EA,$FC,$93,$04,$3F,$C9,$BD,$FF,$E4,$CE,$BF,$E6,$1E,$53,$A3,$FC
	.dc.b	$A4,$DA,$FF,$2C,$4E,$A7,$24,$8C,$89,$E0,$FF,$97,$7B,$FF,$9C,$46
	.dc.b	$2F,$C1,$3B,$AF,$8A,$7A,$48,$FE,$05,$AE,$8C,$10,$FB,$27,$A7,$7F
	.dc.b	$AE,$D3,$31,$F9,$1C,$FF,$A7,$3F,$2B,$5F,$E7,$6F,$69,$FE,$A5,$A7
	.dc.b	$F3,$FD,$3D,$71,$8A,$DB,$FC,$67,$6E,$78,$D8,$F0,$1F,$A4,$8C,$BF
	.dc.b	$BC,$6B,$F8,$EC,$3C,$8F,$0D,$FF,$94,$65,$A6,$C6,$77,$70,$7F,$67
	.dc.b	$E9,$FD,$0C,$9A,$4C,$BE,$34,$7A,$74,$F5,$BF,$B0,$6E,$BE,$FC,$FF
	.dc.b	$9B,$71,$FF,$73,$F1,$3F,$70,$DC,$3F,$8B,$E0,$9C,$B6,$3F,$9E,$72
	.dc.b	$FF,$7B,$F9,$EF,$E0,$0D,$8C,$CA,$72,$D8,$F3,$BE,$5C,$3E,$A3,$4F
	.dc.b	$0C,$F8,$09,$38,$3F,$AE,$F4,$7F,$70,$AF,$E6,$4D,$AC,$CF,$DC,$3F
	.dc.b	$9C,$CF,$6C,$BE,$14,$F8,$9A,$FF,$34,$EA,$FF,$0B,$78,$9F,$D6,$DB
	.dc.b	$1F,$D4,$04,$65,$A7,$F5,$FD,$BF,$C3,$DF,$AB,$FF,$5A,$24,$64,$3F
	.dc.b	$D9,$F0,$F9,$51,$FD,$9D,$AF,$F4,$66,$BF,$CC,$9A,$FC,$73,$F5,$04
	.dc.b	$3F,$04,$FF,$55,$EE,$FF,$0A,$F8,$F7,$33,$49,$5F,$E1,$47,$FC,$22
	.dc.b	$47,$07,$F8,$63,$1F,$F0,$A9,$1C,$1F,$CE,$DF,$E2,$C4,$70,$7F,$8D
	.dc.b	$8F,$8E,$7F,$8D,$BC,$3F,$9A,$D3,$27,$F8,$DD,$CB,$FE,$39,$F1,$2A
	.dc.b	$F1,$3F,$C9,$35,$FE,$46,$8E,$26,$67,$7E,$2F,$FB,$A7,$3F,$F6,$A1
	.dc.b	$27,$FD,$9E,$D7,$FD,$96,$D7,$FD,$DC,$D7,$FD,$BE,$D8,$E9,$1C,$88
	.dc.b	$72,$29,$C8,$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$2B
	.dc.b	$A7,$04,$70,$4E,$EF,$D5,$47,$56,$BC,$9A,$F1,$6C,$71,$CF,$F8,$8B
	.dc.b	$6B,$E3,$4E,$89,$3A,$6C,$6A,$30,$47,$04,$AF,$8F,$3C,$1B,$3B,$6C
	.dc.b	$ED,$F5,$46,$5B,$19,$7D,$91,$96,$C6,$5F,$74,$65,$B1,$9A,$F7,$47
	.dc.b	$80,$87,$62,$9D,$0A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21
	.dc.b	$A2,$BA,$70,$47,$04,$AF,$99,$1D,$5A,$F2,$6B,$C5,$B1,$C6,$BE,$33
	.dc.b	$6B,$E3,$4E,$89,$3A,$6C,$6A,$30,$47,$04,$AF,$8F,$3C,$1B,$3B,$6C
	.dc.b	$ED,$F5,$46,$5B,$19,$7D,$91,$96,$C6,$5F,$74,$65,$B1,$9A,$F7,$47
	.dc.b	$DC,$ED,$FF,$10,$BF,$E3,$2B,$FF,$62,$3F,$E2,$7E,$FF,$F7,$13,$96
	.dc.b	$B8,$A7,$D3,$7F,$9A,$63,$FD,$D1,$18,$C7,$71,$89,$D2,$9D,$3A,$6B
	.dc.b	$D9,$AF,$78,$E3,$18,$8E,$93,$D9,$E9,$3A,$24,$66,$78,$4E,$A3,$13
	.dc.b	$A8,$F1,$35,$F1,$35,$F1,$47,$02,$3C,$1A,$F0,$6B,$C0,$D1,$0D,$14
	.dc.b	$D1,$82,$18,$21,$F0,$0E,$05,$36,$53,$67,$BC,$D9,$4D,$94,$DC,$FC
	.dc.b	$62,$4F,$C1,$8E,$84,$38,$16,$37,$3F,$02,$31,$3C,$5D,$92,$76,$F9
	.dc.b	$1A,$F9,$1A,$F9,$23,$94,$62,$3A,$CF,$77,$A4,$E8,$91,$99,$E1,$3A
	.dc.b	$8C,$4E,$A3,$C4,$D7,$C4,$D7,$C5,$1C,$08,$F0,$6B,$C1,$AF,$03,$44
	.dc.b	$34,$53,$46,$08,$60,$87,$C0,$38,$14,$D9,$4D,$9E,$F3,$65,$36,$53
	.dc.b	$73,$F1,$89,$3F,$06,$3A,$10,$E0,$58,$DC,$FC,$08,$C4,$F1,$76,$49
	.dc.b	$DB,$E4,$6B,$E4,$6B,$E4,$8E,$51,$88,$EB,$3D,$DE,$93,$A2,$46,$67
	.dc.b	$84,$EA,$31,$3A,$8F,$13,$F9,$7F,$B4,$7C,$AA,$58,$7F,$1C,$9C,$C6
	.dc.b	$0F,$E4,$AE,$23,$53,$C5,$FF,$20,$12,$7F,$95,$EC,$D7,$51,$86,$BF
	.dc.b	$CC,$A3,$FD,$5E,$F7,$31,$D9,$A7,$CD,$1A,$AF,$E6,$2D,$8F,$E7,$3C
	.dc.b	$DF,$EC,$B9,$FF,$7B,$12,$33,$E0,$BB,$7B,$47,$FA,$DF,$C7,$FF,$7E
	.dc.b	$11,$DC,$23,$FA,$2F,$DD,$3F,$22,$7C,$6B,$87,$58,$7F,$4C,$AD,$4E
	.dc.b	$6F,$8D,$F0,$BD,$DF,$D1,$BD,$C6,$23,$FD,$94,$7F,$4E,$CF,$FD,$CC
	.dc.b	$DC,$BD,$4D,$CB,$FA,$9F,$86,$F0,$DF,$8C,$7C,$98,$C3,$63,$35,$F0
	.dc.b	$A7,$2D,$87,$57,$E1,$7F,$02,$FE,$CB,$5D,$63,$15,$FD,$AA,$30,$A7
	.dc.b	$75,$CA,$30,$6D,$FC,$23,$E8,$6B,$C3,$8F,$FB,$05,$7A,$DF,$EC,$51
	.dc.b	$FE,$EB,$3E,$01,$0F,$AE,$7F,$7A,$D3,$FD,$ED,$EC,$FF,$79,$69,$FD
	.dc.b	$E0,$B1,$BF,$01,$3F,$DF,$ED,$F8,$D3,$CB,$FD,$31,$F7,$EF,$F5,$13
	.dc.b	$A8,$CE,$1F,$0E,$B8,$57,$18,$C4,$EA,$3F,$C2,$3C,$1F,$71,$F4,$90
	.dc.b	$CF,$07,$DE,$7F,$C2,$FE,$9B,$ED,$5F,$3F,$FB,$1C,$73,$18,$AF,$DD
	.dc.b	$C6,$27,$51,$FB,$FA,$E5,$18,$86,$7E,$B7,$F3,$95,$E8,$69,$F0,$33
	.dc.b	$F6,$C7,$C3,$8E,$31,$88,$F5,$47,$E6,$8A,$7D,$93,$E3,$7D,$DF,$F1
	.dc.b	$18,$E5,$5F,$42,$31,$1A,$9F,$09,$DA,$32,$F8,$F1,$FF,$1D,$CF,$D1
	.dc.b	$6C,$64,$EC,$53,$BD,$BE,$DF,$71,$EA,$6E,$1F,$82,$6C,$EE,$79,$9E
	.dc.b	$D3,$EB,$F2,$FF,$8D,$F5,$FB,$91,$83,$F5,$33,$F6,$5E,$37,$F8,$EC
	.dc.b	$3A,$92,$3A,$9F,$1C,$E4,$7C,$AA,$EA,$F8,$48,$6C,$BE,$13,$C2,$FE
	.dc.b	$75,$E8,$3C,$4F,$F3,$0C,$E7,$BB,$FA,$14,$E6,$FF,$A2,$4E,$54,$65
	.dc.b	$FF,$35,$18,$AE,$77,$D9,$B7,$D1,$B3,$B3,$AC,$ED,$FE,$A5,$83,$8B
	.dc.b	$F9,$B5,$7F,$1D,$24,$FF,$34,$6B,$F4,$1A,$F2,$8E,$2F,$E0,$CF,$81
	.dc.b	$BD,$9F,$C0,$E7,$FD,$85,$B3,$A3,$A1,$DC,$86,$0F,$D1,$6F,$C6,$BF
	.dc.b	$80,$36,$32,$73,$3F,$18,$5D,$7F,$00,$DD,$BF,$09,$7E,$22,$5F,$E8
	.dc.b	$7E,$0F,$F2,$FC,$3F,$DD,$DB,$66,$9F,$B3,$5A,$7C,$BE,$1F,$B1,$1F
	.dc.b	$15,$AE,$B5,$F8,$EF,$B4,$FD,$DF,$27,$7C,$7F,$81,$61,$FA,$88,$C4
	.dc.b	$FF,$03,$6A,$E4,$FE,$1A,$BC,$6F,$E0,$D1,$86,$F0,$7F,$17,$6F,$2F
	.dc.b	$F8,$4D,$F5,$69,$E9,$DB,$EE,$E1,$E7,$D7,$D3,$39,$8F,$E2,$DE,$77
	.dc.b	$95,$34,$9A,$FF,$06,$E8,$FE,$7E,$FC,$43,$FA,$2E,$1F,$D0,$FA,$3F
	.dc.b	$BE,$36,$3B,$F0,$7C,$6C,$3E,$54,$F9,$31,$F1,$47,$F4,$4E,$AF,$EB
	.dc.b	$B3,$FD,$6E,$3E,$8F,$27,$F5,$B6,$CE,$9F,$EC,$9D,$9F,$E3,$3F,$CA
	.dc.b	$FF,$B4,$B6,$32,$23,$2D,$3F,$B6,$77,$7C,$96,$9F,$44,$BC,$1F,$DD
	.dc.b	$23,$51,$F7,$CF,$99,$8F,$AF,$3F,$77,$7F,$F8,$53,$C9,$FF,$0C,$F8
	.dc.b	$7F,$E1,$F6,$EE,$69,$FE,$13,$26,$1F,$E2,$A6,$BF,$C9,$B2,$FF,$0C
	.dc.b	$11,$FF,$17,$BF,$E3,$CF,$17,$CB,$1F,$E4,$06,$AF,$17,$FC,$8F,$3B
	.dc.b	$8F,$71,$FE,$4B,$F0,$B7,$3F,$E0,$E9,$FF,$23,$C6,$0E,$7E,$3F,$F9
	.dc.b	$7A,$3F,$CB,$75,$C8,$FF,$32,$46,$5F,$F3,$0F,$8F,$E2,$F2,$5E,$4F
	.dc.b	$E1,$A7,$FD,$C9,$C1,$FE,$E7,$23,$82,$1F,$EE,$E7,$2D,$32,$7F,$C0
	.dc.b	$07,$2D,$32,$7F,$BF,$5C,$B4,$CC,$F9,$35,$FD,$EB,$67,$6D,$9D,$D7
	.dc.b	$A5,$F1,$10,$F0,$14,$EC,$58,$E4,$D8,$E4,$F0,$21,$C0,$B1,$C1,$B1
	.dc.b	$C1,$D1,$23,$4D,$8D,$46,$08,$E0,$95,$F0,$1B,$5F,$01,$B5,$F0,$27
	.dc.b	$44,$9D,$36,$35,$18,$23,$82,$1F,$06,$38,$36,$36,$D3,$67,$C0,$8D
	.dc.b	$B6,$36,$D3,$67,$C1,$8D,$B6,$36,$D3,$75,$F1,$27,$6D,$9D,$B6,$77
	.dc.b	$5F,$11,$F3,$10,$F1,$14,$EC,$58,$E4,$D8,$E4,$F0,$21,$C0,$B1,$C1
	.dc.b	$B1,$C1,$D1,$23,$4D,$8D,$46,$08,$E0,$95,$F0,$1B,$5F,$01,$B5,$F0
	.dc.b	$27,$44,$9D,$36,$35,$18,$23,$82,$1F,$06,$38,$36,$36,$D3,$67,$C0
	.dc.b	$8D,$B6,$36,$D3,$67,$C1,$8D,$BF,$E3,$03,$FD,$F0,$7F,$8A,$BC,$4A
	.dc.b	$FF,$8B,$8B,$E3,$7F,$8E,$5C,$E0,$F8,$0F,$F0,$34,$7F,$C0,$18,$C5
	.dc.b	$F2,$8C,$4E,$94,$E9,$D3,$5E,$B1,$B8,$C4,$79,$9A,$F4,$9E,$CF,$49
	.dc.b	$E0,$48,$CC,$F0,$9D,$46,$27,$51,$E2,$6B,$E2,$8D,$11,$D3,$5D,$18
	.dc.b	$21,$82,$1E,$98,$EA,$D8,$E4,$D3,$89,$4E,$27,$BC,$C9,$4C,$9E,$C3
	.dc.b	$25,$33,$3F,$04,$93,$E2,$8E,$47,$72,$18,$21,$CC,$C9,$4C,$C7,$49
	.dc.b	$F5,$C6,$27,$B3,$B2,$4E,$DF,$3B,$5F,$3C,$70,$8C,$47,$81,$AF,$59
	.dc.b	$EE,$F4,$9E,$04,$8C,$CF,$09,$D4,$62,$75,$1E,$26,$BE,$28,$D1,$1D
	.dc.b	$35,$D1,$82,$18,$21,$E9,$8E,$AD,$8E,$4D,$38,$94,$E2,$7B,$CC,$94
	.dc.b	$C9,$EC,$32,$53,$33,$F0,$49,$3E,$28,$E4,$77,$21,$82,$1C,$CC,$94
	.dc.b	$CC,$74,$9F,$5C,$62,$7B,$3B,$24,$ED,$F3,$B5,$F3,$C7,$08,$C4,$78
	.dc.b	$1A,$F5,$9E,$EF,$49,$E0,$48,$CC,$F0,$9D,$46,$27,$51,$E2,$6B,$E2
	.dc.b	$8D,$11,$D3,$E0,$E9,$FF,$52,$61,$FC,$7D,$B3,$FC,$A9,$AF,$FB,$F2
	.dc.b	$3F,$92,$7C,$1B,$C4,$5A,$73,$18,$6C,$66,$BF,$9A,$D7,$18,$C4,$7F
	.dc.b	$BF,$08,$7F,$BE,$AF,$FD,$FB,$CA,$FD,$71,$EF,$FF,$38,$8C,$36,$33
	.dc.b	$7F,$CE,$67,$2F,$F3,$B1,$DE,$E6,$F4,$FB,$9F,$D2,$27,$31,$8C,$BF
	.dc.b	$D4,$CD,$47,$C0,$AE,$52,$AE,$50,$F5,$BF,$A6,$CF,$81,$B1,$98,$F2
	.dc.b	$36,$BE,$97,$5B,$C3,$FE,$8B,$AF,$FB,$66,$73,$18,$9D,$47,$83,$D7
	.dc.b	$EA,$4C,$3A,$BE,$0E,$EF,$AF,$30,$F5,$38,$9F,$F6,$23,$63,$DF,$1F
	.dc.b	$22,$31,$1F,$56,$BC,$2F,$E1,$DA,$FD,$AF,$1D,$FA,$DB,$C7,$E3,$18
	.dc.b	$6C,$67,$0E,$F5,$F3,$AB,$51,$88,$F3,$B4,$F2,$17,$46,$3E,$68,$CC
	.dc.b	$FF,$BA,$B1,$F2,$BE,$9C,$FE,$51,$F4,$AB,$85,$F2,$BE,$91,$83,$3E
	.dc.b	$1C,$8F,$03,$6F,$E6,$47,$E3,$A3,$0F,$FB,$BD,$B1,$F9,$E3,$FD,$F4
	.dc.b	$7D,$C2,$1B,$8E,$A4,$7A,$D2,$7E,$9C,$72,$3E,$95,$FB,$1E,$67,$7D
	.dc.b	$7F,$14,$E8,$C9,$FE,$29,$1E,$73,$AF,$FC,$5B,$D6,$F3,$BF,$75,$39
	.dc.b	$AE,$0D,$53,$BE,$9F,$D3,$CE,$63,$13,$DA,$FA,$DF,$46,$DF,$06,$CE
	.dc.b	$CF,$D8,$E0,$DB,$F4,$D3,$FE,$F7,$CB,$C5,$5A,$9C,$BF,$12,$74,$7E
	.dc.b	$A8,$FD,$BE,$7F,$C6,$9F,$9A,$DD,$7E,$94,$61,$F8,$87,$39,$87,$F9
	.dc.b	$4B,$0F,$26,$1E,$28,$46,$2B,$BD,$72,$9F,$D6,$EB,$F6,$E3,$11,$FC
	.dc.b	$37,$0F,$13,$A6,$B9,$8D,$9F,$8F,$2C,$F3,$AF,$41,$F8,$8D,$B8,$CF
	.dc.b	$EA,$23,$0F,$08,$EB,$5E,$8C,$BE,$4B,$ED,$6B,$F3,$72,$F7,$65,$ED
	.dc.b	$8F,$E1,$37,$CC,$98,$76,$3F,$88,$6F,$FB,$C6,$F0,$7F,$01,$6F,$0F
	.dc.b	$BA,$75,$5F,$3A,$FC,$15,$F6,$9E,$EF,$58,$A4,$8F,$9B,$C5,$FE,$F4
	.dc.b	$8F,$E8,$2D,$D4,$D3,$FD,$E5,$97,$AE,$BE,$06,$7E,$ED,$3E,$0C,$7C
	.dc.b	$58,$FC,$7E,$1F,$CF,$4B,$87,$F3,$DC,$7F,$A1,$F1,$F7,$3C,$1C,$BF
	.dc.b	$DA,$3E,$5B,$FA,$3B,$71,$FE,$9F,$82,$72,$D8,$FF,$4F,$F3,$FF,$5C
	.dc.b	$C3,$FD,$3E,$D8,$CF,$47,$F6,$29,$CB,$63,$E5,$1E,$98,$F9,$4F,$FA
	.dc.b	$AB,$AB,$FB,$A7,$83,$FD,$E0,$91,$FD,$E3,$B3,$FB,$FC,$7F,$52,$24
	.dc.b	$7F,$10,$6B,$FE,$85,$AF,$F4,$79,$1F,$F4,$A3,$5D,$18,$21,$82,$1F
	.dc.b	$EA,$77,$83,$4D,$94,$D9,$FE,$A6,$76,$D3,$65,$37,$DD,$FE,$C0,$8D
	.dc.b	$B6,$36,$D7,$7D,$DF,$C1,$DC,$B5,$CE,$7F,$C0,$67,$84,$79,$48,$78
	.dc.b	$4A,$E9,$C1,$1C,$12,$7B,$47,$06,$BB,$6C,$6E,$7C,$51,$82,$38,$25
	.dc.b	$7C,$08,$CB,$63,$35,$EE,$6B,$EE,$8D,$11,$D3,$5D,$18,$21,$82,$1E
	.dc.b	$B8,$EA,$D8,$E4,$D3,$89,$4E,$27,$C1,$8E,$2D,$8E,$2D,$38,$94,$E3
	.dc.b	$5F,$1E,$36,$D8,$DB,$5D,$D7,$C6,$72,$D7,$35,$E2,$9E,$11,$E7,$21
	.dc.b	$E1,$2B,$A7,$04,$70,$49,$ED,$1C,$1A,$ED,$B1,$B9,$F1,$46,$08,$E0
	.dc.b	$95,$F0,$23,$2D,$8C,$D7,$B9,$AF,$BA,$34,$47,$4D,$74,$60,$86,$08
	.dc.b	$7A,$E3,$AB,$63,$93,$4E,$25,$38,$9F,$06,$38,$B6,$38,$BF,$33,$AF
	.dc.b	$FC,$22,$FF,$9E,$8F,$F1,$01,$4F,$F0,$CF,$5F,$FA,$51,$FF,$08,$99
	.dc.b	$C1,$F7,$1F,$E6,$98,$E5,$18,$C3,$94,$7F,$A1,$A3,$0E,$1A,$F3,$9E
	.dc.b	$4E,$DA,$A3,$6F,$68,$CC,$62,$38,$4F,$81,$E1,$3A,$24,$66,$78,$4E
	.dc.b	$A3,$13,$A8,$F3,$35,$F3,$46,$88,$E9,$AE,$8C,$10,$C1,$0F,$53,$C1
	.dc.b	$A6,$CA,$6C,$F6,$99,$29,$93,$D4,$64,$A6,$67,$DE,$49,$F6,$47,$62
	.dc.b	$1D,$0A,$72,$2C,$71,$9F,$44,$6C,$86,$CB,$1B,$7C,$8D,$7C,$93,$E2
	.dc.b	$76,$D7,$91,$23,$83,$E2,$8D,$46,$23,$8C,$F9,$1E,$13,$A2,$46,$67
	.dc.b	$84,$EA,$31,$3A,$8F,$3B,$5F,$3C,$68,$8E,$9A,$E8,$C1,$0C,$10,$F5
	.dc.b	$3C,$1A,$6C,$A6,$CF,$69,$92,$99,$3D,$46,$4A,$66,$7D,$E4,$9F,$64
	.dc.b	$76,$21,$D0,$A7,$22,$C7,$19,$F4,$46,$C8,$6C,$B1,$B7,$C8,$D7,$C9
	.dc.b	$3E,$27,$6D,$79,$12,$38,$3E,$28,$D4,$62,$38,$CF,$91,$E1,$3A,$24
	.dc.b	$66,$78,$4E,$A3,$13,$A8,$F3,$B5,$F3,$C6,$88,$FC,$FD,$1F,$0A,$8F
	.dc.b	$E5,$02,$32,$E1,$BD,$BF,$C9,$9B,$1F,$ED,$88,$FE,$53,$CF,$FE,$AA
	.dc.b	$6F,$3B,$9F,$AC,$E6,$BF,$DB,$44,$3F,$DB,$D3,$FE,$FE,$7F,$DC,$86
	.dc.b	$27,$FD,$C8,$FF,$41,$F2,$9E,$83,$EB,$FD,$5F,$80,$FF,$43,$9C,$C6
	.dc.b	$2B,$85,$6D,$56,$E3,$0F,$F4,$BD,$5F,$BF,$3D,$2F,$E7,$F1,$86,$C6
	.dc.b	$63,$FA,$49,$93,$CF,$7E,$93,$FD,$EE,$7B,$0F,$F5,$91,$8D,$3F,$DF
	.dc.b	$AF,$D3,$8F,$0F,$39,$C4,$66,$93,$98,$86,$E3,$33,$C2,$31,$3B,$8F
	.dc.b	$A1,$1E,$22,$9E,$52,$57,$91,$D1,$FE,$AC,$3E,$35,$76,$7C,$33,$E8
	.dc.b	$3C,$05,$73,$1B,$8E,$F3,$D6,$7B,$EB,$F8,$77,$AF,$A5,$E7,$7E,$03
	.dc.b	$5F,$73,$6B,$EB,$B6,$DE,$A5,$24,$F9,$23,$2D,$D7,$81,$1C,$1F,$26
	.dc.b	$FC,$BB,$FD,$C7,$E1,$9F,$3E,$FA,$BE,$E6,$CF,$BB,$1F,$CE,$65,$EF
	.dc.b	$23,$82,$5F,$5B,$FF,$0D,$F0,$7E,$D7,$C6,$74,$7F,$88,$7A,$1D,$0A
	.dc.b	$D4,$E6,$31,$6B,$E2,$FE,$FE,$F8,$DF,$08,$C5,$70,$AD,$C7,$88,$E5
	.dc.b	$7D,$6B,$84,$66,$77,$97,$FB,$C6,$3F,$DF,$E7,$FB,$61,$C3,$FE,$D9
	.dc.b	$31,$97,$FB,$D2,$38,$67,$FB,$0D,$3F,$41,$8F,$CB,$8F,$82,$7F,$90
	.dc.b	$4D,$18,$BF,$D2,$4F,$08,$EB,$A7,$FC,$01,$25,$7E,$63,$5D,$5F,$EA
	.dc.b	$DA,$FE,$A5,$B5,$FC,$09,$BB,$7E,$22,$BC,$84,$9F,$24,$65,$BB,$F0
	.dc.b	$23,$E7,$BF,$CC,$B5,$F8,$87,$F9,$77,$2F,$8E,$4C,$FF,$95,$61,$FB
	.dc.b	$0E,$0F,$19,$2F,$A1,$FB,$18,$FF,$2F,$57,$1C,$FE,$0E,$1F,$12,$7F
	.dc.b	$9A,$C6,$23,$C1,$87,$67,$D7,$87,$5A,$FD,$CB,$70,$E5,$30,$E3,$7C
	.dc.b	$DF,$C0,$98,$AF,$E0,$FC,$3F,$EE,$2A,$FE,$4D,$18,$7F,$1F,$3F,$C6
	.dc.b	$67,$FD,$CF,$C9,$FC,$EA,$1F,$7D,$FC,$DF,$6F,$EB,$0D,$8C,$9F,$CF
	.dc.b	$A3,$2F,$94,$F1,$90,$F9,$E5,$D3,$E3,$63,$F1,$3A,$E4,$FF,$A3,$65
	.dc.b	$AF,$F5,$B9,$CC,$FF,$A4,$F9,$FF,$B3,$C7,$F6,$5E,$8F,$ED,$8D,$93
	.dc.b	$4F,$F3,$96,$12,$3F,$CE,$ED,$7F,$CD,$ED,$7F,$D0,$ED,$7F,$D0,$4D
	.dc.b	$EA,$FF,$4D,$37,$AB,$F5,$33,$D0,$87,$42,$9D,$0A,$74,$2C,$F4,$6C
	.dc.b	$74,$79,$10,$E4,$53,$91,$63,$93,$63,$93,$C0,$87,$02,$C7,$06,$C7
	.dc.b	$07,$44,$8D,$36,$35,$18,$23,$82,$6F,$F9,$09,$E0,$D9,$DB,$67,$75
	.dc.b	$F3,$63,$04,$70,$4A,$F8,$F1,$96,$C6,$5F,$85,$19,$6C,$65,$F7,$C6
	.dc.b	$5B,$19,$7C,$ED,$AF,$3B,$6B,$CF,$3E,$02,$1D,$8A,$74,$29,$D0,$B3
	.dc.b	$D1,$B1,$D1,$E4,$43,$91,$4E,$45,$8E,$4D,$8E,$4F,$02,$1C,$0B,$1C
	.dc.b	$1B,$1C,$1D,$12,$34,$D8,$D4,$60,$8E,$09,$5F,$36,$78,$36,$76,$D9
	.dc.b	$DD,$7C,$D8,$C1,$1C,$12,$BE,$3C,$65,$B1,$97,$E1,$46,$5B,$19,$7D
	.dc.b	$F1,$96,$C6,$5F,$3B,$6B,$CE,$DA,$F3,$CF,$80,$87,$62,$9D,$0A,$7E
	.dc.b	$B3,$A3,$FC,$0A,$FF,$A1,$4F,$F0,$01,$4F,$F0,$3F,$3D,$69,$7E,$89
	.dc.b	$F2,$9F,$E5,$48,$E5,$18,$CB,$84,$23,$0F,$19,$E0,$FF,$A5,$63,$2F
	.dc.b	$85,$AF,$48,$DB,$CD,$B1,$A8,$E5,$18,$24,$F8,$A7,$53,$C6,$31,$3A
	.dc.b	$8F,$24,$60,$8E,$08,$79,$63,$83,$63,$6D,$36,$7A,$5C,$B4,$C9,$E8
	.dc.b	$72,$D3,$27,$9C,$B3,$EA,$24,$F8,$A3,$A1,$E1,$21,$82,$1D,$4C,$94
	.dc.b	$C9,$DC,$C9,$4C,$C7,$82,$7C,$F1,$93,$91,$0C,$12,$38,$3E,$A9,$F4
	.dc.b	$3B,$8C,$BE,$06,$BE,$08,$DB,$CD,$B1,$A8,$E5,$18,$24,$F9,$A7,$53
	.dc.b	$C6,$31,$3A,$8F,$24,$60,$8E,$08,$79,$63,$83,$63,$6D,$36,$7A,$5C
	.dc.b	$B4,$C9,$E8,$72,$D3,$27,$9C,$B3,$EA,$24,$F8,$A3,$A1,$E1,$21,$82
	.dc.b	$1D,$4C,$94,$C9,$DC,$C9,$4C,$C7,$82,$7C,$F1,$93,$91,$0C,$12,$38
	.dc.b	$3E,$A9,$F4,$3B,$8C,$BE,$06,$BE,$08,$DB,$CD,$B1,$A8,$E5,$18,$24
	.dc.b	$F9,$A7,$53,$C6,$31,$3A,$8F,$24,$60,$8E,$08,$79,$63,$83,$63,$6F
	.dc.b	$EA,$37,$FF,$25,$F2,$7F,$AD,$A3,$FC,$D3,$87,$F9,$D0,$8F,$F2,$FD
	.dc.b	$DE,$BB,$0E,$12,$13,$A8,$CC,$62,$78,$57,$F3,$7A,$D3,$7C,$DC,$D7
	.dc.b	$76,$BF,$E5,$D6,$BD,$E5,$18,$8F,$E7,$5E,$BD,$D0,$D2,$63,$E4,$CF
	.dc.b	$FD,$C3,$9F,$FB,$7D,$F8,$77,$E5,$CF,$F9,$DE,$5E,$B8,$C5,$FF,$48
	.dc.b	$8C,$47,$FA,$C3,$8B,$FE,$00,$12,$7F,$DF,$CF,$C7,$8F,$F5,$1F,$1F
	.dc.b	$FB,$B9,$FF,$50,$90,$FF,$3F,$76,$7C,$6A,$A7,$FB,$B4,$BB,$7F,$AB
	.dc.b	$23,$53,$F4,$65,$D6,$C3,$C5,$3E,$13,$EC,$46,$67,$CA,$71,$8F,$F5
	.dc.b	$49,$E3,$29,$FE,$91,$C1,$A7,$F6,$F9,$EF,$7D,$9F,$84,$74,$AF,$02
	.dc.b	$FD,$A5,$D7,$F0,$E3,$FD,$D5,$18,$7C,$53,$F8,$D8,$C3,$C9,$B1,$C2
	.dc.b	$79,$4F,$C1,$9C,$B6,$32,$F3,$6A,$DB,$F0,$2F,$BC,$FC,$D3,$F8,$27
	.dc.b	$B5,$79,$0F,$54,$F4,$69,$9A,$F0,$CF,$2C,$7E,$F3,$F2,$71,$FC,$74
	.dc.b	$62,$3E,$69,$0F,$F0,$CE,$6E,$5F,$F0,$FB,$63,$31,$82,$38,$20,$E4
	.dc.b	$FF,$14,$CF,$18,$FF,$67,$93,$3F,$F6,$7B,$FE,$D5,$31,$7F,$EF,$17
	.dc.b	$E5,$65,$FE,$FC,$CB,$EB,$63,$F4,$F3,$FE,$04,$D9,$F4,$1C,$F0,$EB
	.dc.b	$7C,$A7,$44,$9F,$9A,$D8,$D4,$60,$8F,$E7,$75,$FC,$C4,$FE,$31,$B3
	.dc.b	$F9,$08,$C6,$BE,$48,$F1,$65,$EB,$BF,$F7,$8C,$7E,$60,$F3,$BD,$67
	.dc.b	$83,$63,$2F,$EA,$23,$2F,$E1,$B3,$FE,$45,$CF,$FB,$3A,$FF,$76,$9F
	.dc.b	$B8,$70,$7F,$00,$DB,$FD,$7D,$9F,$C4,$BF,$1E,$7E,$A7,$C1,$9F,$AB
	.dc.b	$0F,$09,$E7,$7F,$CB,$5C,$1F,$CB,$5F,$E2,$64,$9F,$E6,$DC,$3F,$ED
	.dc.b	$5C,$3B,$47,$DC,$C7,$FD,$9D,$C7,$FC,$93,$16,$7F,$E6,$7E,$3F,$5D
	.dc.b	$FF,$97,$5A,$FF,$9A,$1A,$FF,$31,$E4,$AF,$F4,$06,$E9,$FC,$D2,$FA
	.dc.b	$CF,$F3,$B6,$BF,$E7,$B8,$FE,$8D,$7F,$CB,$09,$1E,$A9,$DB,$FA,$BC
	.dc.b	$6B,$FB,$9E,$0F,$E2,$75,$FE,$81,$CF,$FA,$DB,$63,$33,$FE,$61,$D7
	.dc.b	$FC,$D2,$47,$04,$3F,$CC,$4E,$5A,$64,$FF,$31,$39,$69,$93,$FC,$F0
	.dc.b	$E5,$A6,$74,$FF,$48,$46,$5B,$19,$E5,$FD,$5B,$72,$FA,$F1,$88,$F1
	.dc.b	$90,$F0,$94,$EE,$59,$EA,$D8,$E4,$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1
	.dc.b	$23,$4D,$8D,$46,$08,$E0,$93,$EC,$25,$7C,$38,$CB,$63,$35,$F0,$E3
	.dc.b	$04,$70,$43,$DD,$1C,$1B,$1B,$69,$B3,$DB,$1B,$6C,$6D,$A6,$CF,$74
	.dc.b	$6D,$B1,$B6,$9B,$AF,$83,$3B,$6C,$ED,$B3,$BA,$F8,$2D,$AE,$11,$88
	.dc.b	$F4,$90,$F2,$94,$F0,$96,$7A,$B6,$39,$3C,$08,$70,$2C,$70,$6C,$70
	.dc.b	$74,$48,$D3,$63,$51,$82,$38,$24,$FB,$09,$5F,$0E,$32,$D8,$CD,$7C
	.dc.b	$38,$C1,$1C,$10,$F7,$47,$06,$C6,$DA,$6C,$F6,$C6,$DB,$1B,$69,$B3
	.dc.b	$DD,$1B,$6C,$6D,$A6,$EB,$E0,$CE,$DB,$3B,$6C,$FE,$37,$91,$DD,$FE
	.dc.b	$68,$8F,$EF,$4D,$8C,$F3,$7F,$9B,$E6,$1A,$6B,$FE,$73,$A1,$FE,$41
	.dc.b	$78,$38,$C3,$71,$A8,$D3,$C5,$AE,$63,$71,$C6,$3B,$B5,$F1,$B5,$EE
	.dc.b	$F2,$30,$F5,$8E,$71,$92,$38,$24,$76,$8C,$B6,$32,$F8,$23,$2D,$8C
	.dc.b	$BE,$26,$BE,$13,$91,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C
	.dc.b	$10,$F9,$87,$52,$9C,$8A,$71,$29,$C6,$3E,$5B,$82,$38,$21,$F1,$1E
	.dc.b	$4E,$23,$E1,$C7,$08,$D3,$D5,$AE,$63,$71,$C6,$3D,$0D,$7D,$0D,$7B
	.dc.b	$BC,$8C,$3D,$63,$9C,$64,$8E,$09,$1D,$A3,$2D,$8C,$BE,$08,$CB,$63
	.dc.b	$2F,$89,$AF,$84,$E4,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$3E,$61,$D4,$A7,$22,$9C,$4A,$71,$8F,$96,$E0,$8E,$08,$7C,$47
	.dc.b	$93,$88,$F8,$71,$C2,$34,$F5,$6B,$98,$DC,$71,$8F,$43,$5F,$43,$5E
	.dc.b	$EF,$23,$0F,$58,$E7,$19,$23,$82,$47,$68,$CB,$63,$2F,$82,$32,$D8
	.dc.b	$CB,$E2,$7F,$3F,$B3,$F0,$A4,$AF,$F9,$4C,$E6,$30,$7F,$2E,$71,$1A
	.dc.b	$9E,$2F,$F9,$3F,$7D,$D7,$F3,$5D,$DB,$7F,$BA,$2B,$44,$9E,$9D,$0C
	.dc.b	$B8,$C7,$93,$24,$E6,$BC,$AD,$9E,$35,$C1,$FF,$39,$36,$BF,$CE,$55
	.dc.b	$EC,$E1,$E1,$5F,$E7,$16,$BD,$F3,$FE,$86,$AB,$E1,$3E,$F3,$FA,$3B
	.dc.b	$87,$FD,$B8,$7F,$BA,$4F,$F3,$A6,$7F,$D3,$33,$FF,$75,$47,$FA,$E7
	.dc.b	$BF,$FD,$D1,$CA,$69,$33,$FF,$80,$33,$C8,$E9,$5F,$EF,$67,$B9,$0F
	.dc.b	$F7,$41,$B3,$FB,$24,$2B,$B3,$FE,$F9,$6C,$7F,$B2,$6B,$C0,$FF,$AD
	.dc.b	$4E,$FA,$FF,$BD,$5F,$A9,$EC,$3E,$23,$FE,$00,$37,$0F,$C4,$9E,$07
	.dc.b	$13,$F6,$CC,$B8,$8F,$2E,$16,$7F,$DB,$95,$F1,$09,$3E,$7F,$B7,$F8
	.dc.b	$E4,$70,$7C,$6C,$BE,$25,$7C,$27,$E7,$BF,$8B,$BE,$CD,$BE,$AF,$FB
	.dc.b	$A5,$F3,$57,$CB,$8F,$F5,$71,$F0,$B3,$FB,$59,$7C,$38,$F9,$C7,$D0
	.dc.b	$E2,$FF,$09,$AE,$97,$2F,$F8,$56,$73,$18,$9F,$D7,$C6,$27,$5C,$BF
	.dc.b	$AE,$AD,$57,$F8,$6C,$E9,$3A,$E3,$FD,$BC,$EA,$F9,$57,$4B,$E5,$7F
	.dc.b	$B1,$C3,$FE,$00,$4F,$FB,$86,$3F,$6C,$47,$FC,$53,$AF,$FB,$98,$FD
	.dc.b	$BE,$5E,$28,$F9,$19,$FF,$BC,$B5,$FC,$05,$FC,$98,$C6,$BF,$B9,$8F
	.dc.b	$F7,$F1,$D6,$BC,$57,$FF,$00,$31,$3D,$6F,$F2,$DB,$71,$7F,$0A,$DA
	.dc.b	$FC,$6E,$28,$C4,$7D,$D6,$CF,$E7,$F1,$B5,$E6,$8F,$C7,$C6,$0D,$EF
	.dc.b	$9B,$FC,$AC,$60,$93,$F6,$8F,$8B,$8F,$CA,$CF,$E3,$3E,$AC,$FF,$CA
	.dc.b	$33,$FA,$FE,$0F,$09,$1F,$90,$FA,$30,$FA,$47,$4D,$3E,$21,$2B,$E4
	.dc.b	$47,$FB,$FD,$B5,$D4,$E6,$47,$F7,$4F,$9E,$3E,$7C,$72,$7B,$9F,$E6
	.dc.b	$7E,$1F,$E7,$91,$FC,$6C,$91,$FC,$F7,$8B,$FA,$1C,$7F,$27,$A7,$3D
	.dc.b	$25,$7F,$43,$E4,$FE,$71,$1F,$CD,$27,$73,$FC,$D2,$30,$47,$E0,$3F
	.dc.b	$D1,$C8,$E0,$F1,$1C,$8F,$45,$FF,$56,$8F,$E6,$44,$8F,$F2,$F5,$FF
	.dc.b	$98,$C8,$E0,$87,$F9,$85,$CB,$4C,$9F,$E7,$37,$2D,$32,$7F,$A0,$5C
	.dc.b	$B4,$CF,$37,$FA,$2E,$32,$D8,$CB,$FA,$56,$DF,$EB,$5E,$47,$8C,$87
	.dc.b	$84,$A6,$8C,$10,$C1,$23,$C0,$F0,$6B,$B6,$BB,$7C,$8E,$DA,$ED,$AE
	.dc.b	$E3,$D1,$1C,$88,$68,$AE,$9C,$11,$C1,$27,$E3,$C6,$5B,$19,$9F,$8F
	.dc.b	$18,$23,$82,$1E,$98,$E0,$D8,$DB,$4D,$9E,$D8,$DB,$63,$6D,$36,$7C
	.dc.b	$17,$2D,$33,$3F,$02,$32,$D8,$CB,$ED,$6C,$F8,$1E,$47,$94,$87,$84
	.dc.b	$A6,$8C,$10,$C1,$23,$C0,$F0,$6B,$B6,$BB,$7C,$8E,$DA,$ED,$AE,$E3
	.dc.b	$D1,$1C,$88,$68,$AE,$9C,$11,$C1,$27,$E3,$C6,$5B,$19,$9F,$8F,$18
	.dc.b	$23,$82,$1E,$98,$E0,$D8,$DB,$4D,$9E,$D8,$DB,$63,$6D,$36,$7C,$17
	.dc.b	$2D,$33,$3F,$02,$32,$D8,$CB,$ED,$7F,$CD,$E7,$F7,$7E,$5A,$FF,$7D
	.dc.b	$29,$FE,$01,$33,$C9,$FE,$0E,$77,$2F,$94,$FF,$17,$46,$1A,$F6,$8D
	.dc.b	$61,$D2,$11,$87,$BB,$5E,$51,$B8,$C4,$73,$9E,$AF,$FA,$22,$32,$D8
	.dc.b	$CB,$D9,$B1,$A8,$C4,$F5,$9C,$91,$C1,$23,$CB,$19,$6C,$65,$F3,$98
	.dc.b	$21,$82,$1E,$97,$83,$4D,$94,$D9,$EB,$76,$D3,$65,$36,$7B,$0C,$94
	.dc.b	$CC,$FC,$12,$4F,$AE,$32,$D7,$C4,$43,$B1,$4E,$85,$8E,$73,$E6,$8C
	.dc.b	$9C,$08,$60,$91,$C1,$F3,$B5,$F1,$C7,$38,$C4,$77,$9F,$33,$D2,$32
	.dc.b	$D8,$CB,$D9,$B1,$A8,$C4,$F5,$9C,$91,$C1,$23,$CB,$19,$6C,$65,$F3
	.dc.b	$98,$21,$82,$1E,$97,$83,$4D,$94,$D9,$EB,$76,$D3,$65,$36,$7B,$0C
	.dc.b	$94,$CC,$FC,$12,$4F,$AE,$32,$D7,$C4,$43,$B1,$4E,$85,$8E,$73,$E6
	.dc.b	$8C,$9C,$08,$60,$91,$C1,$F3,$B5,$F1,$C7,$38,$C4,$77,$9F,$33,$D2
	.dc.b	$32,$D8,$CB,$D9,$B1,$A8,$C4,$F5,$9C,$91,$C1,$23,$CB,$19,$6C,$65
	.dc.b	$F3,$98,$21,$FB,$4F,$C5,$5D,$28,$63,$FC,$BA,$73,$18,$3F,$99,$B8
	.dc.b	$8D,$4F,$17,$86,$F6,$B6,$D7,$F9,$ED,$7F,$A7,$9E,$86,$23,$4F,$8F
	.dc.b	$8B,$FD,$5F,$5F,$D0,$1B,$1F,$E8,$58,$C7,$17,$FA,$A1,$FE,$8B,$C2
	.dc.b	$EB,$67,$DA,$FD,$D3,$98,$C4,$23,$15,$CA,$B8,$B6,$9F,$01,$49,$7F
	.dc.b	$D5,$AF,$A4,$7F,$A8,$39,$DF,$2B,$E2,$8C,$36,$BF,$DF,$ED,$7F,$D7
	.dc.b	$6D,$5E,$23,$C8,$78,$1F,$3F,$2F,$EC,$F3,$98,$C5,$7D,$B8,$C3,$F7
	.dc.b	$27,$75,$CA,$30,$7F,$B4,$8F,$F7,$E1,$F3,$37,$FE,$DF,$5E,$1E,$AF
	.dc.b	$C3,$7C,$4F,$F6,$C8,$C5,$23,$16,$AD,$E1,$E0,$6D,$7F,$B6,$37,$F9
	.dc.b	$B3,$FE,$00,$1A,$FD,$27,$0D,$9E,$11,$E8,$7F,$0D,$C1,$F8,$B7,$FD
	.dc.b	$90,$7A,$8E,$3F,$55,$F4,$66,$9F,$68,$BB,$FF,$81,$9F,$0C,$6E,$7E
	.dc.b	$56,$5F,$EE,$D7,$F4,$D3,$F2,$BB,$BD,$11,$A6,$BF,$62,$78,$3D,$71
	.dc.b	$7C,$F7,$E5,$23,$0D,$BF,$7B,$63,$FD,$FA,$D7,$DF,$BF,$F8,$5F,$91
	.dc.b	$A7,$C1,$9F,$B4,$7C,$23,$CC,$7A,$0F,$31,$4F,$36,$5E,$8E,$0F,$C7
	.dc.b	$3E,$A8,$57,$6B,$F0,$5F,$79,$F1,$9F,$3C,$F0,$90,$C9,$FE,$36,$F0
	.dc.b	$3F,$85,$78,$2E,$CF,$F1,$E3,$67,$FC,$80,$1C,$2E,$C5,$72,$AE,$24
	.dc.b	$9E,$35,$C2,$B7,$D1,$FC,$76,$39,$E5,$EB,$BF,$CF,$12,$FF,$C9,$F1
	.dc.b	$96,$E2,$CB,$FD,$A9,$39,$9F,$F2,$BB,$5F,$AB,$5F,$E5,$8D,$3E,$5C
	.dc.b	$7B,$F6,$FA,$73,$FA,$6D,$7F,$98,$36,$74,$74,$C7,$C1,$1E,$58,$C6
	.dc.b	$3E,$29,$CC,$7E,$22,$F8,$BF,$C8,$37,$FC,$7C,$60,$FD,$6E,$FF,$C9
	.dc.b	$30,$FC,$8C,$70,$CB,$CF,$1E,$32,$47,$A0,$FD,$4F,$BF,$F9,$BE,$2F
	.dc.b	$E5,$AD,$83,$4F,$F3,$7E,$BF,$17,$2F,$C5,$E7,$E8,$7D,$78,$7B,$5F
	.dc.b	$47,$27,$F3,$8E,$EF,$D4,$37,$1F,$E0,$6D,$9F,$E0,$73,$FC,$43,$9B
	.dc.b	$FA,$D7,$C4,$D9,$34,$97,$F7,$38,$3E,$51,$31,$F1,$6D,$FC,$A3,$1F
	.dc.b	$E3,$7D,$3F,$D9,$5B,$89,$A7,$EF,$7A,$BE,$1C,$7F,$3C,$C7,$E1,$B7
	.dc.b	$7F,$E8,$31,$88,$FF,$4F,$F5,$FF,$53,$9F,$E9,$C4,$8F,$E9,$9B,$74
	.dc.b	$76,$FC,$57,$35,$E3,$6E,$1F,$D5,$CF,$E7,$44,$7E,$03,$5F,$EE,$07
	.dc.b	$12,$BF,$DE,$9C,$1F,$E0,$5E,$CF,$EF,$11,$FE,$00,$8C,$88,$FF,$03
	.dc.b	$C6,$49,$1F,$EA,$8D,$3F,$D2,$A4,$70,$43,$FD,$24,$E5,$A6,$4F,$F4
	.dc.b	$83,$96,$99,$3F,$D5,$2E,$5A,$67,$BF,$FD,$7F,$19,$6C,$67,$BF,$F4
	.dc.b	$8D,$C3,$F8,$8B,$D8,$F1,$90,$C1,$0E,$E6,$4A,$66,$7B,$B9,$6B,$99
	.dc.b	$F0,$B9,$6B,$99,$F1,$BD,$09,$3C,$9B,$3C,$1B,$3C,$27,$44,$9D,$36
	.dc.b	$35,$18,$23,$82,$57,$C9,$8C,$B6,$33,$5F,$26,$30,$47,$04,$3D,$D1
	.dc.b	$C1,$B1,$B6,$9B,$3E,$13,$96,$99,$3D,$EE,$5A,$66,$BD,$D1,$96,$C6
	.dc.b	$6B,$D6,$DA,$F2,$3D,$8F,$19,$0C,$10,$EE,$64,$A6,$67,$BB,$96,$B9
	.dc.b	$9F,$0B,$96,$B9,$9F,$1B,$D0,$93,$C9,$B3,$C1,$B3,$C2,$74,$49,$D3
	.dc.b	$63,$51,$82,$38,$25,$7C,$98,$CB,$63,$35,$F2,$63,$04,$70,$43,$DD
	.dc.b	$1C,$1B,$1B,$69,$B3,$E1,$39,$69,$93,$DE,$E5,$A6,$6B,$DD,$19,$6C
	.dc.b	$66,$BD,$6F,$E2,$FA,$BF,$C1,$2F,$FA,$94,$8E,$1F,$F0,$E7,$56,$5C
	.dc.b	$F5,$FF,$A3,$15,$BE,$97,$FC,$81,$1B,$8C,$65,$C6,$3F,$D6,$51,$87
	.dc.b	$0D,$78,$4F,$27,$C9,$19,$6C,$65,$E6,$D7,$C6,$D8,$E1,$1A,$9F,$13
	.dc.b	$A6,$C6,$A3,$07,$92,$79,$12,$76,$D8,$D4,$60,$8E,$08,$7B,$1C,$B4
	.dc.b	$C9,$EC,$72,$D3,$27,$A4,$A7,$98,$E4,$43,$81,$4E,$05,$38,$1A,$21
	.dc.b	$A2,$9A,$30,$43,$04,$9F,$90,$49,$F8,$91,$DE,$31,$3E,$C8,$E5,$18
	.dc.b	$7D,$6D,$78,$4F,$27,$C7,$19,$6C,$65,$E6,$D7,$C8,$D8,$E1,$1A,$9F
	.dc.b	$13,$A6,$C6,$A3,$07,$92,$79,$12,$76,$D8,$D4,$60,$8E,$08,$7B,$1C
	.dc.b	$B4,$C9,$EC,$72,$D3,$27,$A4,$A7,$98,$E4,$43,$81,$4E,$05,$38,$1A
	.dc.b	$21,$A2,$9A,$30,$43,$04,$9F,$90,$49,$F8,$91,$DE,$31,$3E,$C8,$E5
	.dc.b	$18,$7D,$6D,$78,$4F,$27,$C7,$19,$6C,$65,$E6,$D7,$C8,$D8,$E1,$1A
	.dc.b	$9F,$13,$A6,$C6,$A3,$07,$92,$79,$12,$76,$D8,$D4,$60,$8E,$08,$7B
	.dc.b	$1F,$C8,$F3,$BD,$A4,$B1,$FE,$4D,$39,$8C,$1F,$EB,$89,$DC,$E9,$B1
	.dc.b	$FC,$D7,$67,$83,$88,$FE,$79,$18,$79,$98,$8D,$3F,$E6,$16,$AE,$27
	.dc.b	$85,$86,$C7,$23,$D2,$E3,$0F,$E7,$CD,$8F,$F4,$94,$7B,$3A,$BF,$D9
	.dc.b	$CD,$E8,$E5,$3F,$EA,$86,$E7,$FE,$F3,$F0,$74,$49,$FF,$4E,$3F,$D4
	.dc.b	$FD,$86,$CF,$5A,$CE,$63,$01,$C4,$6A,$BF,$B0,$57,$28,$C4,$7D,$4E
	.dc.b	$B7,$86,$EB,$FA,$38,$FF,$67,$9C,$C6,$2B,$85,$6D,$FB,$95,$B8,$C6
	.dc.b	$9F,$EC,$8D,$7F,$DE,$15,$F0,$C8,$7F,$AE,$4F,$77,$A2,$F3,$BB,$43
	.dc.b	$44,$E6,$3B,$9F,$8C,$8C,$CF,$06,$C6,$8F,$F7,$57,$0F,$CA,$C3,$FB
	.dc.b	$CE,$4F,$0F,$F0,$2E,$CC,$4E,$DF,$F0,$41,$83,$A5,$7D,$B3,$D6,$7F
	.dc.b	$BF,$0A,$78,$1F,$A8,$72,$9E,$FB,$F8,$9B,$19,$AF,$BA,$FC,$BE,$8F
	.dc.b	$0D,$FD,$06,$AD,$FE,$CC,$61,$B1,$FB,$09,$F0,$7D,$8F,$E2,$A3,$C0
	.dc.b	$D8,$C9,$B8,$E1,$82,$75,$1E,$D3,$EC,$10,$C1,$E5,$39,$1E,$5D,$D8
	.dc.b	$79,$5B,$1A,$C3,$CE,$FE,$1F,$8F,$FC,$33,$C6,$E1,$FC,$4C,$F7,$3F
	.dc.b	$85,$C6,$67,$73,$A9,$F1,$9A,$70,$D8,$FF,$12,$F5,$7F,$91,$3C,$67
	.dc.b	$6F,$F9,$03,$87,$F8,$BC,$62,$FF,$8C,$D6,$EB,$50,$9D,$29,$D3,$C3
	.dc.b	$4F,$E3,$B3,$A9,$CC,$EE,$73,$5F,$E5,$5C,$BD,$8F,$F0,$92,$4F,$F9
	.dc.b	$56,$32,$DC,$99,$7F,$B3,$27,$33,$CE,$73,$5C,$A3,$EB,$3F,$E5,$AC
	.dc.b	$FF,$0E,$DD,$3E,$AB,$71,$FE,$72,$FD,$53,$95,$7C,$B6,$EF,$FA,$BD
	.dc.b	$BF,$8B,$C7,$E3,$A3,$E6,$9F,$8E,$70,$75,$9C,$9E,$48,$FD,$03,$83
	.dc.b	$F1,$55,$E2,$8F,$D3,$0F,$03,$F5,$5C,$3F,$C9,$9B,$19,$36,$FA,$6B
	.dc.b	$E4,$6B,$F1,$F4,$F8,$B1,$8C,$7F,$D0,$57,$FB,$FE,$FE,$64,$7C,$2F
	.dc.b	$FA,$0B,$93,$F8,$7B,$79,$BF,$8B,$37,$9B,$FB,$0B,$70,$3E,$37,$EB
	.dc.b	$65,$E3,$26,$3F,$E9,$3E,$0F,$E6,$33,$C3,$2F,$E5,$DC,$FD,$1B,$19
	.dc.b	$3F,$9A,$C6,$5F,$E4,$D1,$FD,$1B,$5F,$E8,$AF,$C9,$8F,$E8,$AE,$5B
	.dc.b	$D1,$F2,$DF,$8F,$3F,$D6,$C8,$E0,$F9,$79,$7F,$5F,$23,$83,$FD,$59
	.dc.b	$D3,$FD,$B2,$3F,$B4,$F0,$7C,$49,$CB,$5F,$F5,$B4,$ED,$FF,$00,$75
	.dc.b	$65,$FE,$F4,$7C,$A3,$25,$3F,$C0,$05,$3F,$83,$F6,$71,$5C,$DF,$E1
	.dc.b	$1C,$3F,$C1,$D1,$B3,$33,$BE,$EF,$F5,$EE,$FF,$EA,$22,$4F,$FA,$7D
	.dc.b	$AF,$FA,$75,$AF,$FA,$CD,$AF,$FA,$CD,$B1,$D2,$39,$10,$E4,$53,$91
	.dc.b	$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF,$07,$44,$34,$57,$4E,$08,$E0
	.dc.b	$9D,$DF,$C0,$23,$C8,$D7,$BB,$5E,$8D,$79,$B6,$39,$ED,$FC,$B2,$76
	.dc.b	$D7,$6D,$8D,$D7,$CC,$AE,$44,$9D,$36,$35,$18,$23,$82,$57,$AA,$32
	.dc.b	$D8,$CB,$EC,$8C,$B6,$32,$FA,$1A,$FA,$1B,$5E,$88,$E8,$43,$91,$4E
	.dc.b	$45,$39,$15,$E4,$F0,$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23
	.dc.b	$82,$57,$CD,$8F,$23,$5E,$ED,$7A,$35,$E6,$D8,$E7,$5F,$32,$76,$D7
	.dc.b	$6D,$8D,$D7,$CC,$AE,$44,$9D,$36,$35,$18,$23,$82,$57,$AA,$32,$D8
	.dc.b	$CB,$EC,$8C,$B6,$32,$FA,$1A,$FA,$1B,$5E,$88,$E8,$43,$91,$4E,$45
	.dc.b	$39,$15,$E4,$F0,$3E,$FF,$4F,$F8,$C3,$AF,$FC,$3F,$D1,$FE,$92,$9F
	.dc.b	$F0,$84,$60,$FF,$5A,$B0,$53,$CC,$FF,$24,$46,$63,$18,$EE,$3F,$D7
	.dc.b	$B1,$87,$0D,$78,$4F,$27,$BC,$65,$B1,$97,$9B,$5F,$1B,$63,$84,$6A
	.dc.b	$7C,$4E,$9B,$1A,$8C,$1E,$49,$E4,$49,$DB,$63,$51,$82,$38,$21,$EC
	.dc.b	$72,$D3,$27,$B1,$CB,$4C,$9E,$92,$9E,$63,$91,$0E,$05,$38,$14,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$12,$7E,$41,$27,$E2,$47,$78,$C4,$FB,$23
	.dc.b	$94,$61,$F5,$B5,$E1,$3C,$9F,$1C,$65,$B1,$97,$9B,$5F,$23,$63,$84
	.dc.b	$6A,$7C,$4E,$9B,$1A,$8C,$1E,$49,$E4,$49,$DB,$63,$51,$82,$38,$21
	.dc.b	$EC,$72,$D3,$27,$B1,$CB,$4C,$9E,$92,$9E,$63,$91,$0E,$05,$38,$14
	.dc.b	$E0,$68,$86,$8A,$68,$C1,$0C,$12,$7E,$41,$27,$E2,$47,$78,$C4,$FB
	.dc.b	$23,$94,$61,$F5,$B5,$E1,$3C,$9F,$1C,$65,$B1,$97,$9B,$5F,$23,$63
	.dc.b	$84,$6A,$7C,$4E,$9B,$1A,$8C,$1E,$49,$E4,$49,$DB,$63,$51,$82,$38
	.dc.b	$21,$EC,$72,$D3,$27,$B1,$CB,$F4,$B5,$FF,$6C,$5F,$F2,$96,$C7,$FA
	.dc.b	$52,$31,$D3,$FE,$BC,$7F,$97,$77,$BC,$06,$09,$CC,$61,$B1,$9B,$FE
	.dc.b	$79,$5C,$63,$11,$FE,$CF,$F8,$AF,$E5,$2F,$F9,$91,$B8,$9A,$63,$1F
	.dc.b	$F7,$5C,$FF,$9A,$9B,$3F,$EF,$38,$F4,$F4,$FF,$45,$EC,$7E,$AA,$F7
	.dc.b	$BD,$35,$FE,$A9,$8C,$64,$8C,$4E,$A3,$FA,$AD,$72,$8C,$4F,$FB,$9F
	.dc.b	$C1,$F8,$7E,$1F,$F4,$DC,$BF,$DE,$98,$23,$13,$A8,$FF,$75,$36,$BF
	.dc.b	$D6,$27,$A0,$87,$FB,$18,$FA,$CF,$CD,$8E,$27,$D1,$2E,$DD,$9C,$4F
	.dc.b	$AD,$7D,$CF,$FB,$D9,$C4,$66,$30,$FF,$6E,$8D,$E9,$EE,$8F,$BF,$3A
	.dc.b	$9F,$EE,$E7,$4F,$9A,$B3,$CD,$AF,$06,$C6,$E3,$C2,$7C,$FC,$BE,$C3
	.dc.b	$EF,$21,$F1,$0E,$F5,$E1,$7D,$17,$E5,$AF,$39,$D0,$B1,$F4,$F6,$FC
	.dc.b	$29,$27,$FB,$BD,$F8,$27,$B1,$0F,$24,$ED,$FF,$07,$72,$FF,$86,$D6
	.dc.b	$6E,$6F,$F0,$CC,$E6,$31,$3F,$8B,$8C,$4E,$B9,$BF,$49,$5A,$AF,$F1
	.dc.b	$19,$D2,$75,$AF,$E4,$E7,$57,$CA,$BA,$5F,$2B,$FF,$7C,$9F,$A0,$24
	.dc.b	$7F,$8A,$F1,$62,$CB,$FD,$FB,$39,$8C,$57,$08,$E6,$7D,$09,$FD,$FC
	.dc.b	$7E,$17,$4F,$C5,$36,$33,$A6,$CF,$88,$75,$7B,$D7,$42,$1F,$85,$36
	.dc.b	$7E,$24,$A6,$67,$C1,$AF,$ED,$5F,$05,$68,$FC,$0B,$82,$3F,$A6,$E9
	.dc.b	$FD,$21,$1C,$1E,$58,$F8,$15,$F8,$0B,$F3,$BF,$95,$D7,$F8,$FF,$6F
	.dc.b	$EB,$49,$7F,$B2,$7F,$83,$BE,$8E,$2F,$E0,$CD,$E3,$FE,$1B,$8E,$8F
	.dc.b	$CA,$63,$DB,$1F,$E3,$4D,$DB,$F9,$04,$7F,$B3,$DB,$3F,$CD,$F9,$3E
	.dc.b	$14,$FD,$D6,$BF,$E6,$AE,$5F,$E5,$DA,$FF,$32,$24,$7F,$32,$E6,$FE
	.dc.b	$73,$1F,$C9,$89,$1F,$5D,$AF,$F9,$26,$FF,$CA,$64,$7F,$CA,$0D,$74
	.dc.b	$60,$86,$08,$7F,$99,$DE,$0D,$36,$53,$67,$F9,$AD,$DB,$4D,$94,$DF
	.dc.b	$47,$FA,$12,$36,$D8,$DB,$5D,$F4,$7F,$BF,$C9,$AF,$FB,$ED,$B5,$C9
	.dc.b	$B3,$C9,$AF,$29,$E0,$49,$E0,$D8,$E0,$D7,$84,$68,$8E,$9A,$E8,$C1
	.dc.b	$0C,$10,$F9,$04,$3E,$21,$6B,$E0,$36,$BD,$2D,$AF,$23,$6B,$C2,$D3
	.dc.b	$C0,$53,$C0,$76,$25,$76,$6D,$76,$6D,$76,$6D,$76,$69,$D8,$A7,$6A
	.dc.b	$E8,$4A,$E8,$DA,$E8,$DA,$E8,$DA,$E8,$D7,$A5,$72,$25,$72,$6D,$72
	.dc.b	$6C,$F2,$6B,$CA,$78,$12,$78,$36,$38,$35,$E1,$1A,$23,$A6,$BA,$30
	.dc.b	$43,$04,$3E,$A1,$0F,$92,$5A,$F8,$0D,$AF,$4B,$6B,$C8,$DA,$F0,$B4
	.dc.b	$F0,$14,$F0,$1D,$89,$5D,$9B,$5D,$9B,$5D,$9B,$5D,$9A,$76,$29,$DA
	.dc.b	$BA,$12,$BA,$36,$BA,$3E,$1E,$5F,$EE,$D5,$FE,$3B,$3F,$B3,$F3,$7F
	.dc.b	$BD,$A3,$31,$FD,$B8,$D1,$FE,$41,$97,$C1,$7F,$BF,$C9,$7F,$EF,$C6
	.dc.b	$CF,$01,$3C,$27,$44,$8D,$46,$5B,$3C,$9F,$0C,$65,$B1,$97,$D5,$19
	.dc.b	$6C,$65,$F4,$35,$F1,$B6,$38,$46,$A7,$C8,$F6,$6B,$D1,$AE,$CF,$3B
	.dc.b	$87,$AB,$63,$AC,$60,$D9,$0E,$C5,$3B,$14,$EC,$68,$86,$8A,$68,$C1
	.dc.b	$0C,$10,$F1,$1C,$0A,$6C,$A6,$DF,$49,$27,$E4,$12,$7D,$CD,$9F,$41
	.dc.b	$C8,$93,$C6,$76,$48,$DC,$65,$B3,$CD,$F3,$46,$5B,$19,$7D,$11,$96
	.dc.b	$C6,$5F,$53,$5F,$1B,$63,$84,$6A,$7C,$8F,$66,$BD,$1A,$EC,$F3,$B8
	.dc.b	$7A,$B6,$3A,$C6,$0D,$90,$EC,$53,$B1,$4E,$C6,$88,$68,$A6,$8C,$10
	.dc.b	$C1,$0F,$11,$C0,$A6,$CA,$6D,$F4,$92,$7E,$41,$27,$DC,$D9,$F4,$1C
	.dc.b	$89,$3C,$67,$64,$8D,$C6,$5B,$3C,$DF,$34,$65,$B1,$97,$D1,$19,$6C
	.dc.b	$65,$F5,$35,$F1,$B6,$38,$46,$A7,$C8,$F6,$6B,$D1,$AE,$CF,$3B,$F8
	.dc.b	$1D,$1F,$35,$FE,$E7,$C7,$F9,$14,$62,$BF,$DF,$A7,$FB,$AD,$C4,$6A
	.dc.b	$78,$4F,$FA,$16,$3F,$DE,$E4,$70,$7F,$34,$EF,$7B,$FF,$EE,$78,$FF
	.dc.b	$7B,$46,$30,$FF,$77,$1B,$9D,$4A,$75,$D7,$FE,$70,$8F,$F7,$69,$FE
	.dc.b	$BF,$3F,$CE,$26,$0F,$F4,$11,$4C,$9F,$EA,$BF,$8A,$3E,$03,$FA,$77
	.dc.b	$1A,$33,$2C,$38,$56,$CF,$F7,$04,$EA,$3F,$DB,$F1,$E0,$29,$E0,$3A
	.dc.b	$C7,$CF,$8E,$78,$F5,$9C,$9F,$F0,$00,$EA,$58,$FF,$80,$3A,$7F,$B9
	.dc.b	$FA,$BF,$AB,$C7,$FB,$9D,$B3,$FE,$BD,$9F,$F8,$01,$E2,$3C,$87,$A5
	.dc.b	$FE,$FC,$8C,$36,$33,$5F,$EF,$C9,$CB,$FE,$FE,$3D,$47,$B4,$FA,$1A
	.dc.b	$7F,$BC,$5F,$8D,$87,$8B,$1F,$F7,$FE,$3E,$72,$3C,$EB,$85,$74,$23
	.dc.b	$82,$77,$7F,$BE,$63,$E5,$F1,$FF,$BA,$5F,$F7,$7F,$6F,$FB,$B3,$BF
	.dc.b	$FD,$D5,$1F,$E0,$63,$FD,$E4,$58,$D6,$DF,$EF,$3E,$7F,$F0,$44,$62
	.dc.b	$76,$7F,$AE,$A3,$FD,$96,$D3,$FD,$9C,$43,$FD,$C2,$64,$FF,$71,$E5
	.dc.b	$F5,$8F,$F6,$E6,$9F,$5A,$7E,$A4,$63,$4F,$F7,$7D,$72,$D3,$FD,$FB
	.dc.b	$D7,$FE,$C9,$D7,$FD,$F7,$3F,$E2,$1D,$BE,$9E,$3F,$87,$84,$63,$1D
	.dc.b	$6F,$F6,$AB,$53,$9A,$FC,$FD,$79,$67,$C6,$47,$EB,$9F,$EF,$88,$63
	.dc.b	$FA,$35,$D3,$E1,$73,$C5,$FA,$A8,$C4,$ED,$B1,$9A,$F9,$15,$FE,$F5
	.dc.b	$BF,$9B,$7F,$01,$FF,$19,$3D,$FB,$B8,$C9,$A4,$AE,$71,$F4,$4E,$86
	.dc.b	$27,$A3,$73,$F9,$6F,$D8,$E5,$FE,$21,$7E,$B6,$DF,$F1,$13,$C0,$58
	.dc.b	$FF,$7F,$E5,$FE,$EC,$AF,$E1,$38,$7D,$88,$C4,$3A,$3F,$95,$47,$F2
	.dc.b	$78,$F9,$27,$F9,$5A,$32,$DE,$8F,$94,$FF,$8E,$F3,$FF,$1B,$91,$C1
	.dc.b	$0F,$F1,$F3,$96,$99,$3F,$C8,$CE,$5A,$64,$FF,$20,$B9,$69,$9E,$0F
	.dc.b	$F3,$14,$65,$B1,$9E,$9F,$E4,$93,$96,$C6,$73,$4E,$5B,$19,$8C,$11
	.dc.b	$C1,$0F,$59,$0F,$49,$4F,$39,$6B,$CA,$DA,$F1,$B6,$BC,$2D,$AF,$03
	.dc.b	$4F,$01,$D8,$87,$62,$D7,$66,$D7,$66,$D7,$66,$D7,$66,$9D,$8E,$84
	.dc.b	$AE,$8D,$AE,$8D,$AE,$8D,$AE,$8D,$3A,$57,$22,$57,$26,$D7,$26,$D7
	.dc.b	$26,$D7,$2A,$E0,$4A,$E0,$DA,$E0,$D9,$E1,$3A,$24,$E9,$B1,$A8,$C1
	.dc.b	$1C,$10,$FA,$84,$3E,$49,$4F,$80,$5A,$F4,$B6,$BC,$8D,$AF,$0B,$6B
	.dc.b	$C0,$D3,$C0,$76,$21,$D8,$B5,$D9,$B5,$D9,$B5,$D9,$B5,$D9,$A7,$63
	.dc.b	$A1,$2B,$A3,$6B,$A3,$6B,$A3,$6B,$A3,$4E,$95,$C8,$95,$C9,$B5,$C9
	.dc.b	$B5,$C9,$B5,$CA,$B8,$12,$B8,$36,$B8,$36,$78,$4E,$89,$3A,$6C,$6A
	.dc.b	$30,$7F,$BC,$79,$BF,$DF,$71,$FD,$E4,$FF,$7D,$B8,$4D,$B7,$DE,$FF
	.dc.b	$53,$9F,$F8,$00,$D7,$FD,$F6,$D8,$DD,$FF,$BF,$E7,$2D,$8C,$C7,$FB
	.dc.b	$FA,$78,$12,$74,$D8,$D4,$60,$8E,$08,$7F,$93,$8F,$2B,$A7,$07,$88
	.dc.b	$D9,$B9,$F1,$BA,$6B,$99,$F1,$39,$6B,$99,$EC,$D3,$A9,$E2,$29,$E1
	.dc.b	$39,$3C,$1D,$18,$37,$3E,$27,$27,$53,$44,$30,$43,$64,$3C,$06,$4A
	.dc.b	$65,$EC,$E0,$8E,$09,$3F,$02,$38,$35,$DB,$63,$73,$ED,$9C,$B6,$33
	.dc.b	$1E,$E9,$E0,$49,$D3,$63,$51,$82,$38,$21,$F0,$CF,$2B,$A7,$07,$88
	.dc.b	$D9,$F0,$E7,$C6,$E9,$AE,$67,$C4,$E5,$AE,$67,$B3,$4E,$A7,$88,$A7
	.dc.b	$84,$E4,$F0,$74,$60,$DC,$F8,$9C,$9D,$4D,$10,$C1,$0D,$90,$F0,$19
	.dc.b	$29,$97,$B3,$82,$38,$24,$FC,$08,$E0,$D7,$6D,$8D,$CF,$B6,$72,$D8
	.dc.b	$CC,$7B,$A7,$81,$27,$4D,$8D,$46,$08,$E0,$87,$C3,$3C,$AE,$9C,$1E
	.dc.b	$23,$67,$C3,$9F,$1B,$A6,$B9,$9F,$13,$96,$B9,$9E,$CD,$3A,$9E,$22
	.dc.b	$9E,$13,$F2,$7A,$7F,$A4,$B9,$BF,$DF,$51,$FE,$E4,$E0,$72,$27,$33
	.dc.b	$FE,$62,$FB,$3F,$F7,$34,$7F,$A0,$9B,$87,$F1,$EC,$11,$8A,$FF,$77
	.dc.b	$B5,$FF,$79,$F1,$7F,$BB,$EB,$FD,$C5,$1F,$F0,$05,$D3,$5C,$C6,$EF
	.dc.b	$FD,$23,$8F,$9B,$D4,$FF,$45,$B6,$FC,$4D,$7D,$99,$27,$FD,$23,$8F
	.dc.b	$FB,$FD,$F5,$63,$93,$D5,$19,$69,$9A,$FE,$7D,$F3,$0F,$99,$D4,$61
	.dc.b	$A8,$46,$E7,$94,$66,$31,$7F,$EB,$FF,$0D,$E0,$BD,$9E,$71,$86,$C6
	.dc.b	$70,$E7,$39,$7A,$E1,$FE,$BC,$FA,$9F,$17,$BF,$FE,$B8,$6C,$FD,$76
	.dc.b	$CF,$F5,$9E,$14,$62,$3F,$01,$2A,$F8,$E4,$7E,$2B,$E9,$C3,$C8,$DC
	.dc.b	$53,$C4,$FE,$CD,$8F,$9E,$75,$19,$BF,$AF,$C1,$F9,$07,$FD,$AB,$C1
	.dc.b	$F9,$48,$C6,$7F,$EE,$AD,$3F,$03,$C1,$F7,$6F,$F0,$2D,$7F,$03,$1F
	.dc.b	$E1,$1F,$03,$F3,$90,$F0,$3F,$D0,$FC,$C9,$CC,$61,$BC,$FF,$DF,$67
	.dc.b	$59,$7D,$DA,$D3,$67,$55,$C2,$FF,$05,$7C,$1F,$04,$EA,$B8,$47,$FB
	.dc.b	$3B,$3F,$D8,$FA,$BE,$C2,$47,$B1,$FF,$68,$E9,$FE,$D3,$8F,$7F,$67
	.dc.b	$FB,$5F,$83,$F0,$5E,$2F,$CC,$27,$07,$D0,$E3,$FC,$44,$7E,$80,$91
	.dc.b	$F9,$E6,$DF,$E0,$F9,$3F,$35,$1F,$89,$83,$49,$3D,$79,$3E,$FE,$7F
	.dc.b	$70,$91,$90,$FE,$12,$BF,$51,$18,$3F,$C4,$FC,$BF,$A7,$9D,$4F,$E9
	.dc.b	$F9,$BA,$D6,$9B,$1F,$C3,$A3,$93,$67,$F7,$A4,$8F,$95,$10,$FE,$0B
	.dc.b	$1F,$39,$FF,$1D,$74,$7F,$15,$D7,$C3,$1F,$C4,$DB,$19,$AF,$F1,$EB
	.dc.b	$67,$C5,$3F,$C3,$09,$3F,$C2,$9F,$E3,$3C,$C7,$ED,$F0,$7F,$30,$7F
	.dc.b	$36,$7F,$94,$38,$A5,$FF,$8E,$1A,$FF,$90,$5A,$FF,$94,$1A,$FF,$98
	.dc.b	$1B,$D5,$FE,$64,$6F,$57,$FB,$1C,$98,$FF,$B1,$DB,$1A,$8C,$11,$C1
	.dc.b	$2B,$C8,$43,$C8,$53,$C8,$53,$C8,$5A,$F2,$36,$BC,$2D,$AF,$03,$6B
	.dc.b	$C0,$F6,$21,$D8,$A7,$62,$D7,$66,$D7,$66,$D7,$66,$D7,$67,$A1,$0E
	.dc.b	$85,$AE,$8D,$AE,$8D,$AE,$8D,$AE,$8F,$22,$57,$26,$D7,$26,$D7,$26
	.dc.b	$D7,$2A,$E0,$4A,$E0,$DA,$E0,$D9,$E1,$3A,$24,$E9,$B1,$A8,$C1,$1C
	.dc.b	$12,$BE,$A1,$0F,$92,$53,$E0,$14,$F4,$96,$BC,$8D,$AF,$0B,$6B,$C0
	.dc.b	$DA,$F0,$3D,$88,$76,$29,$D8,$B5,$D9,$B5,$D9,$B5,$D9,$B5,$D9,$E8
	.dc.b	$43,$A1,$6B,$A3,$6B,$A3,$6B,$A3,$6B,$A3,$C8,$95,$C9,$B5,$C9,$B5
	.dc.b	$C9,$B5,$CA,$B8,$12,$B8,$36,$B8,$36,$78,$4E,$89,$3A,$6C,$6A,$30
	.dc.b	$47,$04,$AF,$A8,$43,$E4,$9F,$EF,$CE,$6F,$ED,$ED,$E5,$FF,$7A,$73
	.dc.b	$7F,$BF,$CF,$F7,$F7,$C9,$7F,$BE,$49,$87,$FB,$DD,$B5,$FE,$EA,$19
	.dc.b	$23,$31,$FE,$E9,$6B,$FE,$ED,$6B,$FE,$33,$8E,$13,$D4,$91,$D6,$32
	.dc.b	$D9,$E2,$F8,$1B,$1A,$8F,$04,$F2,$69,$C5,$C3,$E3,$6B,$E3,$9E,$64
	.dc.b	$8C,$C7,$76,$CF,$78,$E4,$75,$21,$E0,$21,$A2,$9D,$E3,$9C,$F7,$8E
	.dc.b	$26,$C8,$6C,$C9,$4E,$67,$94,$C9,$4C,$CF,$9D,$E0,$74,$34,$57,$4E
	.dc.b	$0D,$90,$C1,$1D,$92,$3D,$51,$E0,$6B,$D5,$AF,$38,$E1,$3F,$08,$91
	.dc.b	$EA,$8C,$B6,$78,$BE,$16,$C6,$A3,$C1,$3C,$9A,$71,$70,$F8,$DA,$F8
	.dc.b	$E7,$99,$23,$31,$DD,$B3,$DE,$39,$1D,$48,$78,$08,$68,$A7,$78,$E7
	.dc.b	$3D,$E3,$89,$B2,$1B,$32,$53,$99,$E5,$32,$53,$33,$E7,$78,$1D,$0D
	.dc.b	$15,$D3,$83,$64,$30,$47,$64,$8F,$54,$78,$1A,$F5,$6B,$CE,$38,$4F
	.dc.b	$C2,$24,$7A,$A3,$2D,$9E,$2F,$85,$B1,$A8,$F0,$4F,$26,$9C,$5C,$3E
	.dc.b	$36,$BE,$39,$E6,$48,$CC,$77,$6C,$FE,$9B,$9F,$FD,$6B,$3F,$EE,$86
	.dc.b	$E1,$FE,$75,$25,$7F,$A8,$9F,$F5,$3E,$BF,$EA,$9A,$C1,$FE,$F7,$E8
	.dc.b	$FF,$77,$DF,$FA,$8E,$FA,$46,$0C,$F3,$7F,$B6,$FA,$7F,$D0,$44,$9F
	.dc.b	$F6,$55,$FF,$B4,$2F,$FD,$68,$FF,$AC,$A7,$FD,$6A,$47,$FD,$D6,$FF
	.dc.b	$BC,$09,$38,$3F,$CF,$87,$FB,$6B,$97,$FD,$CA,$E2,$3F,$CE,$71,$89
	.dc.b	$FF,$79,$FC,$17,$FB,$5D,$FF,$77,$B6,$0D,$35,$6C,$9C,$7F,$ED,$DB
	.dc.b	$FF,$64,$F5,$FF,$B3,$08,$F9,$1F,$0F,$3F,$FB,$71,$B9,$FF,$B6,$67
	.dc.b	$FD,$2C,$D7,$FD,$B3,$5F,$D6,$5A,$7F,$B6,$08,$7F,$B6,$0A,$7F,$BC
	.dc.b	$EF,$CA,$E0,$FF,$65,$F1,$7F,$B3,$57,$17,$FB,$37,$7F,$F6,$9C,$7F
	.dc.b	$B3,$8D,$F0,$7F,$B3,$9D,$D7,$1A,$E1,$C7,$FE,$D3,$9E,$06,$AB,$86
	.dc.b	$BF,$1F,$25,$7F,$BC,$09,$5F,$EF,$06,$BE,$98,$DB,$92,$1F,$EC,$0E
	.dc.b	$9F,$92,$4D,$3F,$02,$FF,$A9,$8F,$F7,$E9,$23,$FD,$FA,$FF,$80,$F8
	.dc.b	$3E,$EC,$7F,$B5,$27,$FD,$93,$C1,$FE,$D8,$6C,$64,$FC,$39,$4F,$AF
	.dc.b	$D1,$F6,$89,$1F,$92,$7F,$D9,$11,$FE,$CB,$83,$48,$7F,$BE,$70,$FF
	.dc.b	$6E,$BF,$EC,$D3,$13,$FE,$CD,$78,$E5,$FB,$A8,$FF,$76,$12,$BF,$56
	.dc.b	$FF,$B4,$23,$F6,$9C,$F7,$75,$7F,$B2,$4F,$DD,$16,$FC,$A2,$B8,$C6
	.dc.b	$23,$F2,$75,$FE,$D0,$C3,$FD,$A5,$19,$AF,$BD,$3F,$E2,$F7,$F8,$7B
	.dc.b	$FF,$00,$49,$5F,$C0,$1F,$F7,$2F,$2F,$F8,$B3,$5F,$F1,$01,$2F,$FC
	.dc.b	$4A,$D7,$FC,$44,$D7,$FC,$28,$D7,$FC,$40,$D8,$E9,$1C,$88,$72,$29
	.dc.b	$C8,$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04
	.dc.b	$70,$4E,$6F,$B3,$3E,$D6,$CF,$99,$AF,$85,$AF,$66,$BD,$5B,$3D,$6F
	.dc.b	$F5,$F3,$C5,$AF,$16,$BC,$5B,$1C,$6B,$C9,$18,$AE,$A4,$70,$48,$D9
	.dc.b	$2B,$D5,$19,$6C,$65,$F2,$35,$F2,$35,$F2,$36,$BC,$91,$C8,$87,$22
	.dc.b	$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA,$70
	.dc.b	$47,$04,$AF,$95,$3E,$D6,$CF,$99,$AF,$85,$AF,$66,$BD,$5B,$3D,$6B
	.dc.b	$E8,$4F,$16,$BC,$5A,$F1,$6C,$71,$AF,$24,$62,$BA,$91,$C1,$23,$64
	.dc.b	$AF,$54,$65,$B1,$97,$C8,$D7,$C8,$D7,$C8,$DA,$F2,$47,$22,$1C,$8A
	.dc.b	$72,$29,$C8,$AF,$27,$81,$0E,$05,$38,$15,$E0,$E8,$86,$8B,$C1,$FE
	.dc.b	$DB,$E1,$FF,$6F,$C7,$F6,$B3,$FD,$C0,$7F,$B3,$7E,$A3,$F9,$CB,$70
	.dc.b	$FF,$02,$5A,$73,$3F,$CD,$E3,$13,$A9,$CC,$FF,$3E,$9C,$BC,$C8,$71
	.dc.b	$8D,$BF,$EE,$16,$BF,$EE,$A6,$C7,$28,$E1,$3D,$C8,$F4,$3C,$13,$C0
	.dc.b	$91,$A8,$E2,$6C,$F0,$39,$78,$B5,$C9,$B3,$C8,$E5,$A6,$67,$CC,$F2
	.dc.b	$31,$3E,$26,$CF,$19,$DC,$76,$3C,$A4,$39,$14,$D1,$82,$18,$21,$E4
	.dc.b	$38,$14,$D9,$4D,$BE,$27,$04,$70,$49,$F5,$47,$48,$C4,$FA,$E7,$33
	.dc.b	$DA,$72,$FC,$02,$1C,$63,$6F,$89,$AF,$85,$B1,$CA,$38,$4F,$72,$3D
	.dc.b	$0F,$04,$F0,$24,$6A,$38,$9B,$3C,$0E,$5E,$2D,$72,$6C,$F2,$39,$69
	.dc.b	$99,$F3,$3C,$8C,$4F,$89,$B3,$C6,$77,$1D,$8F,$29,$0E,$45,$34,$60
	.dc.b	$86,$08,$79,$0E,$05,$36,$53,$6F,$89,$C1,$1C,$12,$7D,$51,$D2,$31
	.dc.b	$3E,$B9,$CC,$F6,$9C,$BF,$00,$87,$18,$DB,$E2,$6B,$E1,$6C,$72,$8E
	.dc.b	$13,$DC,$8F,$43,$C1,$3C,$09,$1A,$8E,$26,$CF,$03,$97,$8B,$5C,$9B
	.dc.b	$3C,$8E,$5A,$66,$7C,$CF,$EB,$BA,$7F,$D1,$BC,$7F,$C9,$1B,$5F,$E5
	.dc.b	$F6,$CF,$F9,$91,$AF,$F9,$CB,$E4,$3F,$CF,$75,$FE,$6F,$6D,$7F,$98
	.dc.b	$DB,$3F,$E6,$CD,$91,$88,$FF,$40,$57,$FB,$BB,$7F,$E8,$31,$FE,$C0
	.dc.b	$8F,$F3,$82,$C3,$FD,$47,$D1,$E7,$DB,$FD,$D9,$97,$9D,$B5,$E7,$6B
	.dc.b	$E7,$3F,$DD,$D8,$E1,$FF,$3F,$37,$2F,$2C,$7F,$BC,$F8,$7F,$D0,$F8
	.dc.b	$7F,$A2,$70,$C1,$1F,$90,$E8,$F9,$05,$9F,$90,$D7,$FD,$43,$8F,$FA
	.dc.b	$4A,$13,$FD,$63,$B5,$F5,$6F,$B1,$FF,$53,$F5,$66,$31,$8F,$FB,$9E
	.dc.b	$B7,$5A,$C3,$FD,$33,$18,$9F,$F7,$56,$7E,$CC,$3E,$B4,$63,$4F,$F7
	.dc.b	$D9,$FF,$00,$73,$F2,$4F,$B3,$0F,$F7,$94,$61,$F6,$9C,$6B,$FD,$76
	.dc.b	$FD,$F6,$9F,$7E,$BF,$D5,$71,$98,$FF,$5A,$75,$38,$1E,$57,$FA,$9E
	.dc.b	$30,$D8,$CC,$FF,$AB,$67,$2F,$F8,$13,$7F,$F7,$B9,$38,$3F,$DE,$4D
	.dc.b	$8F,$F7,$C9,$FE,$F4,$E5,$FF,$7B,$B8,$8F,$F3,$FE,$BF,$EF,$5F,$17
	.dc.b	$C0,$49,$F0,$62,$8F,$F5,$E6,$9F,$ED,$AC,$7E,$59,$34,$F8,$4F,$C9
	.dc.b	$EC,$91,$F2,$1B,$87,$C5,$3F,$DD,$47,$E4,$8B,$8F,$E4,$A3,$EB,$9F
	.dc.b	$81,$E1,$FF,$7A,$4F,$FB,$AB,$87,$EF,$4E,$63,$FD,$E9,$18,$BF,$AF
	.dc.b	$D5,$F6,$C8,$E0,$FF,$70,$BF,$9D,$23,$83,$EE,$3F,$88,$23,$83,$F6
	.dc.b	$12,$8F,$DA,$36,$33,$18,$3F,$03,$A7,$FB,$A2,$7F,$82,$6F,$F8,$3E
	.dc.b	$AF,$F7,$61,$F3,$A3,$FD,$E8,$7E,$E5,$FC,$C9,$2F,$C2,$F8,$08,$E0
	.dc.b	$FF,$19,$E9,$FE,$47,$24,$E0,$87,$F8,$B1,$CB,$4C,$9F,$E3,$27,$2D
	.dc.b	$32,$7F,$8C,$1C,$B4,$CF,$3F,$F9,$4A,$32,$D8,$CF,$47,$FB,$48,$95
	.dc.b	$FE,$CC,$6D,$70,$6C,$F0,$9D,$12,$74,$D8,$D4,$60,$8E,$08,$7B,$C8
	.dc.b	$7B,$4A,$7A,$CB,$5E,$96,$D7,$91,$B5,$E1,$6D,$78,$1A,$78,$0E,$C4
	.dc.b	$3B,$16,$BB,$36,$BB,$36,$BB,$36,$BB,$34,$EC,$74,$25,$74,$6D,$74
	.dc.b	$6D,$74,$6D,$74,$69,$D2,$B9,$12,$B9,$36,$B9,$36,$B9,$36,$B9,$57
	.dc.b	$02,$57,$06,$D7,$06,$CF,$09,$D1,$27,$4D,$8D,$46,$08,$E0,$87,$D4
	.dc.b	$21,$F2,$4A,$7C,$02,$D7,$A5,$B5,$E4,$6D,$78,$5B,$5E,$06,$9E,$03
	.dc.b	$B1,$0E,$C5,$AE,$CD,$AE,$CD,$AE,$CD,$AE,$CD,$3B,$1D,$09,$5D,$1B
	.dc.b	$5D,$1B,$5D,$1B,$5D,$1A,$74,$AE,$44,$AE,$4D,$AE,$4D,$AE,$4D,$AE
	.dc.b	$55,$C0,$95,$C1,$B5,$FE,$22,$E2,$FF,$73,$74,$7F,$89,$2D,$18,$7F
	.dc.b	$C4,$F8,$3E,$EB,$FD,$AA,$4C,$7F,$DC,$ED,$9D,$B5,$FF,$74,$35,$FF
	.dc.b	$73,$35,$09,$ED,$3D,$09,$1D,$1A,$F4,$6B,$D1,$AF,$48,$E6,$F1,$21
	.dc.b	$C4,$A7,$12,$9C,$67,$CC,$F9,$5B,$1D,$23,$91,$0E,$45,$39,$4F,$91
	.dc.b	$E4,$D8,$E5,$1C,$08,$70,$3C,$67,$02,$BC,$1D,$1D,$CD,$15,$D3,$83
	.dc.b	$A9,$4E,$26,$1D,$92,$7D,$E4,$9F,$43,$67,$C6,$D7,$C0,$D7,$A3,$5E
	.dc.b	$E4,$3C,$44,$9F,$04,$F6,$24,$76,$6B,$D9,$AF,$66,$BD,$63,$9B,$C4
	.dc.b	$87,$12,$9C,$4A,$71,$9F,$43,$E5,$6C,$74,$8E,$44,$39,$14,$E5,$3E
	.dc.b	$47,$93,$63,$94,$70,$21,$C0,$F1,$9C,$0A,$F0,$74,$77,$34,$57,$4E
	.dc.b	$0E,$A5,$38,$98,$76,$49,$F7,$92,$7D,$0D,$9F,$1B,$5F,$03,$5E,$8D
	.dc.b	$7B,$90,$F1,$12,$7C,$13,$D8,$91,$D9,$AF,$66,$BD,$9A,$F5,$8E,$6F
	.dc.b	$12,$1C,$4A,$71,$29,$C6,$7D,$0F,$95,$B1,$D2,$39,$10,$F9,$BC,$2F
	.dc.b	$33,$FD,$A9,$08,$C2,$C5,$87,$F2,$28,$FE,$47,$3A,$9C,$C7,$FB,$7E
	.dc.b	$B9,$CF,$FB,$A2,$3A,$65,$82,$3D,$6F,$FD,$1F,$7F,$EE,$33,$FD,$44
	.dc.b	$7F,$B8,$A9,$CB,$E2,$8F,$F6,$A1,$27,$C9,$5E,$99,$D4,$62,$9E,$26
	.dc.b	$2B,$FD,$E8,$D7,$D6,$D7,$E1,$37,$A3,$FD,$B1,$87,$FB,$14,$97,$FE
	.dc.b	$C4,$7F,$D9,$BD,$72,$3F,$D8,$8D,$BF,$91,$87,$AA,$7F,$DD,$47,$97
	.dc.b	$1F,$F7,$34,$FA,$63,$FD,$CA,$7F,$B4,$2B,$FD,$A8,$47,$07,$13,$FD
	.dc.b	$A3,$1F,$52,$7E,$9C,$66,$3F,$DF,$47,$FA,$4A,$FE,$73,$F4,$0F,$F6
	.dc.b	$66,$5F,$EC,$A2,$3F,$EC,$76,$BF,$DC,$3B,$3F,$08,$FF,$B3,$5B,$7F
	.dc.b	$ED,$18,$FC,$57,$17,$FA,$75,$FF,$65,$F1,$7F,$A7,$9F,$45,$FE,$21
	.dc.b	$F4,$62,$8C,$47,$C3,$6E,$7F,$25,$FF,$5D,$C7,$E4,$F4,$FC,$AD,$7F
	.dc.b	$AC,$63,$0F,$E6,$39,$37,$1D,$9F,$F5,$FB,$F5,$32,$D4,$FF,$B0,$63
	.dc.b	$FD,$AD,$3F,$EC,$06,$C7,$FB,$05,$EA,$FF,$AD,$95,$78,$B1,$F4,$B6
	.dc.b	$7F,$2A,$D9,$FC,$8C,$71,$C7,$F2,$AD,$DB,$F5,$F7,$F0,$1B,$1F,$7E
	.dc.b	$3E,$1E,$38,$BF,$54,$F3,$8C,$47,$ED,$B3,$FA,$4F,$FB,$43,$93,$FD
	.dc.b	$A4,$E5,$AF,$D8,$6E,$3F,$4E,$3C,$DA,$7F,$88,$B1,$FA,$6D,$C7,$F5
	.dc.b	$CD,$9F,$E0,$47,$B7,$83,$0F,$F0,$86,$BF,$C3,$DB,$C3,$FE,$2C,$9F
	.dc.b	$F1,$51,$27,$FC,$54,$D7,$FC,$54,$D7,$FC,$32,$D7,$FC,$32,$D8,$E9
	.dc.b	$1C,$88,$72,$29,$C8,$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2
	.dc.b	$1A,$2B,$A7,$04,$70,$4E,$2F,$E6,$73,$ED,$6C,$F9,$9A,$F8,$5A,$F6
	.dc.b	$6B,$D5,$B3,$D6,$FF,$01,$3C,$5A,$F1,$6B,$C5,$B1,$C6,$BC,$91,$8A
	.dc.b	$EA,$47,$04,$8D,$92,$BD,$51,$96,$C6,$5F,$23,$5F,$23,$5F,$23,$6B
	.dc.b	$C9,$1C,$88,$72,$29,$C8,$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83
	.dc.b	$A2,$1A,$2B,$A7,$04,$70,$4A,$F9,$53,$ED,$6C,$F9,$9A,$F8,$5A,$F6
	.dc.b	$6B,$D5,$B3,$D6,$BE,$84,$F1,$6B,$C5,$AF,$16,$C7,$1A,$F2,$46,$2B
	.dc.b	$A9,$1C,$12,$36,$4A,$F5,$46,$5B,$19,$7C,$8D,$7C,$8D,$7C,$8D,$AF
	.dc.b	$24,$72,$21,$C8,$A7,$22,$9C,$8A,$F2,$78,$10,$E0,$53,$81,$5E,$0E
	.dc.b	$88,$68,$AE,$9F,$E0,$3C,$7F,$ED,$D7,$FC,$3A,$D7,$FC,$65,$0F,$B9
	.dc.b	$FE,$CF,$1F,$EC,$B6,$BF,$ED,$76,$C6,$EF,$FD,$D4,$D9,$FF,$74,$CE
	.dc.b	$89,$3A,$6C,$6A,$30,$47,$04,$3F,$C6,$73,$D5,$B3,$C9,$B3,$C5,$A7
	.dc.b	$13,$DB,$1B,$6C,$6D,$A6,$CF,$80,$E5,$A6,$4F,$84,$E5,$A6,$67,$CC
	.dc.b	$F4,$3C,$64,$30,$43,$A9,$92,$99,$3A,$99,$29,$97,$BB,$C0,$86,$8A
	.dc.b	$E9,$C1,$1C,$12,$7E,$34,$70,$6B,$B6,$C6,$E7,$E1,$36,$7E,$14,$E8
	.dc.b	$93,$A6,$C6,$A3,$04,$70,$43,$E0,$4F,$56,$CF,$26,$CF,$16,$9C,$4F
	.dc.b	$8D,$1B,$6C,$6D,$A6,$CF,$80,$E5,$A6,$4F,$84,$E5,$A6,$67,$CC,$F4
	.dc.b	$3C,$64,$30,$43,$A9,$92,$99,$3A,$99,$29,$97,$BB,$C0,$86,$8A,$E9
	.dc.b	$C1,$1C,$12,$7E,$34,$70,$6B,$B6,$C6,$E7,$E1,$36,$7E,$14,$E8,$93
	.dc.b	$A6,$C6,$A3,$04,$70,$43,$E0,$4F,$56,$CF,$26,$CF,$16,$9C,$4F,$8D
	.dc.b	$1B,$6C,$6D,$A6,$CF,$80,$E5,$A6,$4F,$84,$E5,$A6,$73,$FF,$6D,$47
	.dc.b	$FB,$1F,$2F,$F6,$C9,$27,$FD,$AC,$7F,$AE,$BA,$7F,$DB,$A4,$72,$7F
	.dc.b	$B9,$1F,$F7,$0E,$CD,$BF,$D4,$F3,$FE,$71,$E4,$FF,$58,$BF,$E7,$87
	.dc.b	$FD,$B4,$DC,$3F,$D7,$91,$FE,$81,$7A,$C7,$FB,$51,$B8,$FF,$A2,$27
	.dc.b	$47,$FA,$9B,$C2,$57,$FB,$28,$FF,$68,$B8,$E6,$D4,$64,$FF,$7D,$C6
	.dc.b	$6F,$FD,$D1,$8F,$A6,$BF,$DE,$4D,$AC,$E2,$9C,$B6,$3F,$DE,$1C,$7F
	.dc.b	$ED,$CC,$BF,$D3,$B5,$EB,$26,$1C,$AB,$89,$0C,$10,$FF,$4C,$19,$29
	.dc.b	$FE,$D8,$3F,$DA,$C4,$3F,$DA,$D9,$7C,$2E,$3F,$F6,$A7,$03,$9B,$FD
	.dc.b	$AF,$C5,$FE,$DC,$8C,$56,$E3,$FD,$4C,$7B,$09,$7F,$01,$AE,$8C,$10
	.dc.b	$FF,$5E,$F3,$7D,$92,$63,$F1,$0F,$F6,$B4,$7F,$AC,$63,$E4,$91,$F9
	.dc.b	$AF,$DA,$7E,$C9,$C0,$66,$DF,$EF,$CF,$DC,$25,$77,$E5,$FC,$51,$1C
	.dc.b	$1F,$EC,$B2,$5F,$FB,$2A,$3F,$D9,$B1,$FE,$D5,$E0,$FF,$64,$CF,$FB
	.dc.b	$1E,$7F,$DF,$4F,$E8,$CF,$F6,$D7,$0F,$FA,$F3,$13,$4C,$71,$49,$FF
	.dc.b	$62,$12,$BF,$DF,$FC,$9F,$E1,$5D,$3F,$C1,$C4,$9F,$F0,$3B,$5F,$F0
	.dc.b	$1B,$5F,$F0,$0B,$5F,$EF,$ED,$8E,$91,$C8,$87,$22,$9C,$8A,$72,$2B
	.dc.b	$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA,$70,$47,$04,$E4,$FF
	.dc.b	$5B,$93,$3F,$F5,$8B,$6B,$E0,$36,$BD,$2D,$7C,$8D,$7C,$2D,$7C,$0D
	.dc.b	$AF,$05,$76,$25,$76,$6D,$76,$6B,$D9,$AF,$66,$BD,$9B,$3D,$A7,$A1
	.dc.b	$27,$A3,$5E,$8D,$7A,$35,$E8,$D8,$E9,$1C,$88,$72,$29,$C8,$A7,$22
	.dc.b	$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70,$4A,$FA
	.dc.b	$84,$AF,$92,$DA,$F8,$0D,$AF,$4B,$5F,$23,$5F,$0B,$5F,$03,$6B,$C1
	.dc.b	$5D,$89,$5D,$9B,$5D,$9A,$F6,$6B,$D9,$AF,$66,$CF,$69,$E8,$49,$E8
	.dc.b	$D7,$A3,$5E,$8D,$7A,$36,$3A,$47,$22,$1C,$8A,$72,$29,$C8,$AF,$27
	.dc.b	$81,$0E,$05,$38,$15,$E0,$FF,$74,$DF,$FD,$74,$DE,$0F,$F0,$2B,$FC
	.dc.b	$FE,$3F,$A9,$BF,$D0,$4F,$EF,$72,$F9,$96,$1F,$DF,$8F,$F5,$7B,$89
	.dc.b	$FE,$56,$D8,$C9,$B8,$FF,$64,$C6,$5B,$19,$7B,$C6,$5B,$19,$53,$E3
	.dc.b	$8C,$36,$3C,$31,$89,$E5,$1C,$1B,$3B,$9F,$2C,$F0,$23,$82,$3E,$63
	.dc.b	$25,$72,$F9,$8C,$95,$CB,$E1,$7C,$84,$76,$64,$B1,$E5,$7B,$10,$E8
	.dc.b	$53,$84,$76,$8E,$11,$A2,$1A,$38,$3A,$70,$69,$C1,$27,$E3,$92,$7D
	.dc.b	$07,$77,$13,$E8,$6C,$64,$DC,$78,$23,$2D,$8C,$BE,$18,$CB,$63,$2F
	.dc.b	$98,$93,$E5,$8C,$B6,$3C,$51,$89,$E7,$1C,$1B,$3B,$9F,$34,$F0,$23
	.dc.b	$82,$3E,$73,$25,$72,$F9,$CC,$95,$CB,$E1,$7C,$84,$76,$64,$B1,$E5
	.dc.b	$7B,$10,$E8,$53,$84,$76,$8E,$11,$A2,$1A,$38,$3A,$70,$69,$C1,$27
	.dc.b	$E3,$92,$7D,$07,$77,$13,$E8,$6C,$64,$DC,$78,$23,$2D,$8C,$BE,$18
	.dc.b	$CB,$63,$2F,$98,$93,$E5,$8C,$B6,$3C,$51,$89,$E7,$1C,$1B,$3B,$9F
	.dc.b	$34,$F0,$23,$82,$3E,$73,$25,$72,$F9,$CC,$95,$CB,$E1,$7C,$84,$76
	.dc.b	$64,$B1,$E5,$7B,$10,$E8,$7F,$01,$D7,$FC,$E4,$FF,$AD,$38,$3F,$D6
	.dc.b	$3C,$3F,$EA,$E2,$5F,$F3,$BA,$FF,$32,$D7,$FA,$EB,$7F,$F6,$55,$7F
	.dc.b	$39,$8F,$F6,$51,$FE,$BE,$E0,$FF,$65,$47,$FB,$22,$3F,$D7,$B4,$9C
	.dc.b	$C7,$FB,$22,$E3,$FE,$CB,$21,$FE,$C4,$E7,$6D,$FE,$BF,$8F,$F4,$74
	.dc.b	$61,$FF,$62,$07,$1C,$9E,$58,$C4,$F1,$7C,$F1,$9C,$3D,$AD,$CB,$53
	.dc.b	$FE,$98,$6C,$B9,$3F,$D8,$C4,$C7,$C3,$19,$6C,$FF,$A8,$DF,$94,$DA
	.dc.b	$FF,$49,$1D,$CF,$EA,$D3,$F1,$79,$9D,$BF,$ED,$62,$4F,$FB,$69,$FE
	.dc.b	$DC,$FF,$B2,$0F,$26,$5F,$EC,$5E,$2F,$F6,$3C,$7F,$B3,$78,$BF,$BA
	.dc.b	$E7,$FE,$CC,$9C,$CF,$18,$F9,$4F,$FA,$DB,$66,$2E,$DF,$F7,$39,$27
	.dc.b	$FD,$D6,$F8,$B3,$FC,$2E,$5E,$B2,$6B,$F0,$2B,$EC,$BF,$8B,$C7,$FC
	.dc.b	$07,$CF,$FE,$B8,$24,$64,$FF,$64,$5F,$E4,$23,$E3,$12,$3A,$3F,$77
	.dc.b	$83,$F0,$4D,$C3,$FD,$9B,$3F,$E1,$39,$CB,$63,$38,$7F,$B2,$63,$FD
	.dc.b	$96,$FE,$5B,$0F,$F0,$BF,$37,$FB,$30,$FC,$36,$9F,$EC,$C9,$F8,$DA
	.dc.b	$27,$FD,$9B,$1F,$A2,$7E,$E1,$2F,$E1,$3F,$8F,$E3,$FF,$68,$12,$79
	.dc.b	$4F,$EB,$E1,$18,$3F,$C3,$F8,$FF,$85,$89,$3F,$E2,$66,$C6,$A3,$04
	.dc.b	$70,$43,$FC,$4B,$1C,$1B,$1B,$69,$B3,$FC,$25,$1B,$6C,$6D,$A6,$F9
	.dc.b	$3F,$C5,$D3,$B6,$CE,$DB,$3B,$E5,$FF,$22,$D6,$DB,$5B,$6C,$EE,$3F
	.dc.b	$16,$D9,$E1,$18,$8F,$49,$0F,$29,$6B,$C2,$DA,$EA,$D9,$E4,$F0,$24
	.dc.b	$F0,$6C,$F0,$6C,$F0,$9D,$12,$74,$D8,$D4,$60,$8E,$09,$5E,$C2,$3F
	.dc.b	$0C,$C9,$5C,$D7,$C3,$8C,$11,$C1,$0F,$74,$70,$6C,$6D,$A6,$CF,$6C
	.dc.b	$6D,$B1,$B6,$9B,$AF,$74,$ED,$B3,$B6,$CE,$EB,$E0,$D6,$DB,$5B,$6C
	.dc.b	$EE,$3E,$0B,$67,$84,$62,$3D,$24,$3C,$A5,$AF,$0B,$6B,$AB,$67,$93
	.dc.b	$C0,$93,$C1,$B3,$C1,$B3,$C2,$74,$49,$D3,$63,$51,$82,$38,$25,$7B
	.dc.b	$08,$FC,$33,$25,$73,$5F,$0E,$30,$47,$04,$3D,$D1,$C1,$B1,$B6,$9B
	.dc.b	$3D,$B1,$B6,$C6,$DA,$6E,$BD,$D3,$B6,$CE,$DB,$3B,$AF,$83,$5B,$6D
	.dc.b	$6D,$B3,$B8,$F8,$2F,$97,$87,$FD,$AA,$FF,$63,$3F,$DA,$4F,$FB,$51
	.dc.b	$FF,$6A,$5B,$B7,$FD,$76,$4A,$FF,$61,$B6,$7F,$C1,$13,$B9,$C9,$23
	.dc.b	$33,$B8,$EA,$D8,$FE,$72,$7F,$B4,$1C,$3F,$EE,$46,$B9,$8E,$87,$22
	.dc.b	$BC,$A7,$C4,$F8,$23,$2F,$36,$BC,$CD,$11,$E8,$D3,$A4,$F8,$A7,$AC
	.dc.b	$E8,$93,$A7,$AC,$6A,$30,$73,$3C,$45,$3C,$51,$C9,$CB,$5C,$9C,$9C
	.dc.b	$47,$58,$D9,$92,$18,$21,$C8,$AF,$17,$04,$70,$49,$F8,$04,$9F,$6B
	.dc.b	$67,$D5,$3B,$9C,$92,$33,$3B,$8F,$4B,$63,$CE,$78,$5C,$3E,$66,$B9
	.dc.b	$8E,$87,$22,$BC,$A7,$C4,$F8,$23,$2F,$36,$BC,$CD,$11,$E8,$D3,$A4
	.dc.b	$F8,$A7,$AC,$E8,$93,$A7,$AC,$6A,$30,$73,$3C,$45,$3C,$51,$C9,$CB
	.dc.b	$5C,$9C,$9C,$47,$58,$D9,$92,$18,$21,$C8,$AF,$17,$04,$70,$49,$F8
	.dc.b	$04,$9F,$6B,$67,$D5,$3B,$9C,$92,$33,$3B,$8F,$4B,$63,$CE,$78,$5C
	.dc.b	$3E,$66,$B9,$8E,$87,$22,$BC,$A7,$C4,$F8,$23,$2F,$36,$BC,$CD,$11
	.dc.b	$E8,$D3,$A4,$F8,$A7,$AC,$E8,$93,$A7,$AC,$7E,$B7,$6F,$F5,$87,$37
	.dc.b	$F9,$BE,$8D,$25,$7F,$96,$E3,$FD,$9C,$48,$FF,$66,$BF,$C9,$B8,$7F
	.dc.b	$D9,$B1,$FE,$A3,$24,$FF,$A2,$5F,$F3,$8E,$DD,$AB,$FD,$A9,$BF,$FB
	.dc.b	$29,$E4,$D7,$FD,$9E,$D5,$E1,$3C,$2F,$F5,$EB,$63,$FD,$92,$DE,$8F
	.dc.b	$F6,$43,$71,$F5,$47,$F9,$E6,$7F,$D7,$FC,$3F,$EC,$43,$FD,$8B,$87
	.dc.b	$FA,$A2,$7F,$D4,$8D,$9F,$3E,$29,$CB,$63,$FD,$78,$43,$D6,$FF,$B0
	.dc.b	$9C,$1E,$C5,$F8,$6F,$22,$47,$67,$FD,$8B,$C1,$FE,$C0,$E2,$FF,$62
	.dc.b	$6B,$FE,$8C,$E4,$FF,$61,$36,$33,$7F,$EC,$A3,$FD,$23,$5C,$5F,$07
	.dc.b	$E9,$9E,$82,$3E,$C7,$FD,$88,$FA,$39,$EB,$E7,$7D,$FC,$F2,$3E,$1B
	.dc.b	$5F,$C1,$6D,$FE,$B4,$7F,$D4,$3C,$48,$C4,$23,$0F,$FB,$1A,$33,$C3
	.dc.b	$F6,$EB,$EC,$F1,$60,$95,$FE,$CC,$E1,$FF,$5D,$61,$FE,$EF,$E7,$FF
	.dc.b	$5C,$CF,$FB,$FC,$FF,$58,$1F,$EB,$BE,$3F,$F5,$C9,$1F,$39,$A3,$C9
	.dc.b	$97,$FA,$EE,$7F,$0C,$7E,$94,$93,$FE,$D1,$9E,$1C,$8D,$FF,$D7,$F5
	.dc.b	$FD,$F5,$95,$FE,$03,$6D,$7F,$73,$6B,$FE,$01,$6B,$FE,$0B,$6F,$2F
	.dc.b	$F8,$65,$BC,$DF,$E9,$22,$57,$FA,$5D,$B1,$A8,$C1,$1C,$11,$F2,$12
	.dc.b	$BC,$8D,$7C,$8D,$7C,$8D,$AF,$23,$6B,$C2,$D3,$C0,$5A,$F0,$57,$62
	.dc.b	$1D,$8A,$76,$2D,$76,$6D,$76,$6D,$76,$6D,$76,$7A,$10,$E8,$5A,$E8
	.dc.b	$DA,$E8,$DA,$E8,$DA,$E8,$F2,$25,$72,$6D,$72,$6D,$72,$6D,$72,$AE
	.dc.b	$04,$AE,$0D,$AE,$0D,$9E,$13,$A2,$4E,$9B,$1A,$8C,$11,$C1,$1F,$A8
	.dc.b	$4A,$F9,$2D,$7E,$03,$5F,$4B,$6B,$C8,$DA,$F0,$B4,$F0,$16,$BC,$15
	.dc.b	$D8,$87,$62,$9D,$8B,$5D,$9B,$5D,$9B,$5D,$9B,$5D,$9E,$84,$3A,$16
	.dc.b	$BA,$36,$BA,$36,$BA,$36,$BA,$3C,$89,$5C,$9B,$5C,$9B,$5C,$9B,$5C
	.dc.b	$AB,$81,$2B,$83,$6B,$83,$67,$84,$E8,$93,$A6,$C6,$A3,$04,$70,$47
	.dc.b	$EA,$12,$BE,$4B,$5F,$80,$D7,$D2,$FF,$AD,$3A,$3F,$BD,$45,$E3,$FF
	.dc.b	$53,$BF,$EA,$8E,$3F,$F5,$DE,$2F,$BA,$60,$AF,$F5,$AC,$61,$B1,$93
	.dc.b	$FD,$7F,$19,$69,$98,$FF,$56,$46,$5B,$19,$7B,$4F,$86,$31,$08,$C3
	.dc.b	$E0,$24,$76,$69,$D0,$A7,$49,$F0,$36,$3A,$CF,$47,$91,$23,$93,$5E
	.dc.b	$07,$43,$C9,$3D,$1B,$1D,$23,$13,$D6,$79,$90,$E7,$3D,$5E,$27,$58
	.dc.b	$E0,$E8,$E0,$6A,$38,$46,$A3,$06,$9C,$12,$7C,$64,$9F,$31,$27,$C1
	.dc.b	$19,$6C,$64,$EF,$19,$69,$98,$F0,$46,$5B,$19,$7D,$44,$3C,$E5,$9F
	.dc.b	$34,$6E,$32,$48,$CB,$E4,$24,$78,$9A,$78,$0A,$75,$9F,$23,$63,$B4
	.dc.b	$F4,$79,$12,$39,$35,$E0,$74,$3C,$93,$D1,$B1,$D2,$31,$3D,$67,$99
	.dc.b	$0E,$73,$D5,$E2,$75,$8E,$0E,$8E,$06,$A3,$84,$6A,$30,$69,$C1,$27
	.dc.b	$C6,$49,$F3,$12,$7C,$11,$96,$C6,$4E,$F1,$96,$99,$8F,$04,$65,$B1
	.dc.b	$97,$D4,$43,$CE,$59,$F3,$46,$E3,$24,$8C,$BE,$42,$47,$89,$A7,$80
	.dc.b	$A7,$59,$F2,$36,$3B,$4F,$47,$91,$23,$93,$5E,$07,$43,$C9,$3D,$1B
	.dc.b	$1D,$23,$13,$D6,$79,$90,$E7,$3D,$5E,$27,$58,$E0,$E8,$E0,$7E,$93
	.dc.b	$67,$CA,$7F,$35,$C7,$F9,$A4,$FF,$9E,$A3,$17,$FC,$FA,$B7,$3F,$EC
	.dc.b	$88,$C4,$EA,$3F,$D8,$9E,$75,$7F,$D0,$6D,$7F,$D0,$4D,$A3,$9F,$C4
	.dc.b	$5E,$35,$25,$77,$3F,$A5,$08,$DB,$93,$FA,$77,$0B,$E6,$3C,$0D,$8D
	.dc.b	$07,$19,$FC,$08,$C4,$F9,$79,$3F,$DA,$59,$1A,$4E,$CF,$EC,$75,$FE
	.dc.b	$A0,$EB,$FE,$BF,$87,$3B,$46,$36,$F9,$4D,$AF,$F7,$53,$5F,$F5,$4F
	.dc.b	$87,$5F,$F5,$43,$68,$D3,$13,$FD,$26,$3B,$D7,$F6,$06,$D7,$8E,$75
	.dc.b	$8F,$A7,$0F,$F6,$D9,$27,$FB,$E1,$FE,$CB,$1C,$5F,$EC,$B6,$F1,$FF
	.dc.b	$BE,$EF,$FD,$A7,$1B,$9D,$12,$75,$C9,$FE,$E4,$E6,$FF,$5B,$E5,$FE
	.dc.b	$B9,$BF,$87,$9F,$FA,$AC,$B5,$E7,$71,$1F,$19,$AF,$C6,$7F,$D7,$53
	.dc.b	$F1,$1B,$A7,$F8,$42,$FE,$93,$63,$BC,$75,$23,$83,$FD,$FE,$F9,$B4
	.dc.b	$6A,$8F,$F0,$FE,$1F,$86,$3F,$D8,$7B,$7F,$B0,$AF,$FC,$47,$BF,$E3
	.dc.b	$EB,$F1,$8D,$8F,$F1,$3F,$07,$E6,$27,$F2,$0D,$8C,$F0,$F3,$C3,$F3
	.dc.b	$F3,$97,$F5,$A7,$C0,$E2,$D3,$FA,$E3,$07,$C4,$E3,$CB,$FE,$F1,$6E
	.dc.b	$DF,$BA,$C7,$F2,$D1,$AE,$4F,$F1,$47,$07,$F8,$D0,$99,$FF,$8C,$9B
	.dc.b	$87,$F8,$C1,$AF,$F8,$8D,$AF,$F8,$69,$B3,$D2,$79,$12,$79,$35,$E4
	.dc.b	$D7,$93,$63,$94,$70,$21,$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23
	.dc.b	$82,$72,$FF,$97,$2B,$DA,$DA,$F3,$36,$BC,$2D,$7B,$35,$EA,$DA,$EB
	.dc.b	$3F,$9E,$9C,$B6,$33,$5E,$08,$C5,$71,$23,$82,$46,$C8,$F9,$8C,$95
	.dc.b	$CD,$7A,$23,$2D,$8C,$BE,$A8,$CB,$63,$2F,$A1,$B5,$E8,$9E,$C4,$9E
	.dc.b	$8D,$79,$35,$E4,$D8,$E5,$1C,$08,$70,$29,$C0,$AF,$07,$44,$34,$57
	.dc.b	$4E,$08,$E0,$95,$F3,$AB,$DA,$DA,$F3,$36,$BC,$2D,$7B,$35,$EA,$DA
	.dc.b	$EB,$3F,$32,$72,$D8,$CD,$78,$23,$15,$C4,$8E,$09,$1B,$23,$E6,$32
	.dc.b	$57,$35,$E8,$8C,$B6,$32,$FA,$A3,$2D,$8C,$BE,$86,$D7,$A2,$7B,$12
	.dc.b	$7A,$35,$E4,$D7,$93,$63,$94,$70,$21,$C0,$A7,$02,$BC,$1D,$10,$D1
	.dc.b	$5D,$38,$3F,$05,$C6,$E3,$FE,$FF,$1F,$ED,$B8,$C7,$2F,$F4,$48,$FE
	.dc.b	$CA,$D8,$CC,$7F,$48,$A7,$CA,$FF,$AF,$09,$87,$FB,$29,$B3,$FE,$13
	.dc.b	$9D,$CE,$49,$19,$9D,$C7,$FB,$CA,$77,$39,$8D,$99,$9D,$C7,$F8,$AA
	.dc.b	$77,$39,$37,$19,$9D,$BF,$EE,$C6,$BF,$EE,$B6,$CF,$28,$E1,$3C,$9E
	.dc.b	$0D,$78,$36,$38,$46,$8F,$41,$23,$C8,$D3,$A9,$4E,$0E,$88,$E9,$AE
	.dc.b	$8C,$10,$C1,$0F,$29,$D4,$AF,$26,$9C,$4A,$71,$8F,$5B,$82,$38,$24
	.dc.b	$FC,$82,$4F,$C3,$6C,$F8,$67,$73,$92,$46,$67,$71,$E5,$9D,$CE,$63
	.dc.b	$66,$67,$71,$E9,$9D,$CE,$4D,$C6,$67,$6F,$B5,$AF,$B1,$B3,$CA,$38
	.dc.b	$4F,$27,$83,$5E,$0D,$8E,$11,$A3,$D0,$48,$F2,$34,$EA,$53,$83,$A2
	.dc.b	$3A,$6B,$A3,$04,$30,$43,$CA,$75,$2B,$C9,$A7,$12,$9C,$63,$D6,$E0
	.dc.b	$8E,$09,$3F,$20,$93,$F0,$DB,$3E,$19,$DC,$E4,$91,$99,$DC,$79,$67
	.dc.b	$73,$98,$D9,$99,$DC,$7A,$67,$73,$93,$71,$99,$DB,$ED,$6B,$EC,$6C
	.dc.b	$F2,$8E,$13,$C9,$E0,$D7,$83,$63,$84,$68,$F4,$12,$3C,$8D,$3A,$94
	.dc.b	$E0,$FF,$B5,$76,$FF,$66,$12,$3F,$D7,$FB,$7F,$B3,$9F,$F6,$0E,$FF
	.dc.b	$EC,$8C,$BF,$CE,$3C,$1F,$EC,$E6,$CF,$0F,$0B,$FD,$AC,$47,$07,$4E
	.dc.b	$5F,$F5,$53,$FE,$E5,$6B,$FE,$97,$9F,$F4,$D1,$32,$FF,$4F,$BF,$EC
	.dc.b	$59,$FF,$46,$CF,$FA,$33,$1F,$F6,$77,$2F,$F4,$B6,$C6,$67,$6F,$79
	.dc.b	$D4,$62,$3F,$DC,$B3,$C1,$D1,$FE,$98,$DB,$FD,$69,$68,$FF,$5C,$6F
	.dc.b	$FE,$B6,$7F,$AE,$35,$FE,$B6,$6A,$38,$73,$FF,$B0,$07,$0F,$83,$1F
	.dc.b	$F5,$3B,$E7,$CB,$37,$E7,$8C,$11,$C1,$0E,$51,$C1,$FF,$60,$72,$7F
	.dc.b	$AD,$9B,$A7,$FA,$DD,$FF,$66,$1F,$EB,$82,$18,$3E,$1E,$BF,$EA,$02
	.dc.b	$3F,$EC,$87,$FD,$4B,$BF,$FA,$B6,$3E,$23,$53,$4C,$6D,$F8,$BB,$FF
	.dc.b	$7F,$70,$7F,$AB,$09,$5F,$EC,$7C,$3F,$B6,$65,$FD,$CC,$8E,$09,$1F
	.dc.b	$DA,$23,$2D,$8C,$BF,$DA,$A3,$2D,$8C,$BF,$DB,$23,$2D,$8C,$F0,$FF
	.dc.b	$82,$A7,$2D,$8C,$F1,$7F,$A4,$48,$FF,$A3,$DB,$3C,$1B,$1C,$0D,$12
	.dc.b	$34,$D3,$4E,$08,$E0,$95,$EF,$21,$ED,$29,$EB,$2D,$7A,$5B,$5E,$46
	.dc.b	$CF,$85,$B5,$E0,$6C,$78,$1E,$C4,$3B,$16,$BB,$36,$BB,$36,$7B,$36
	.dc.b	$BB,$36,$3B,$3D,$09,$5D,$1B,$5D,$1B,$3D,$1B,$5D,$1B,$1D,$2B,$91
	.dc.b	$2B,$93,$63,$93,$6B,$93,$63,$94,$F0,$23,$C1,$B3,$C1,$B1,$C0,$D1
	.dc.b	$23,$4D,$34,$E0,$8E,$09,$5F,$50,$87,$C9,$29,$F0,$0B,$5E,$96,$D7
	.dc.b	$91,$B3,$E1,$6D,$78,$1B,$1E,$07,$B1,$0E,$C5,$AE,$CD,$AE,$CD,$9E
	.dc.b	$CD,$AE,$CD,$8E,$CF,$42,$57,$46,$D7,$46,$CF,$46,$D7,$46,$C7,$4A
	.dc.b	$E4,$4A,$E4,$D8,$E4,$DA,$E4,$D8,$E5,$3C,$08,$F0,$6C,$F0,$6C,$70
	.dc.b	$34,$48,$F3,$70,$FF,$A8,$2F,$FA,$71,$FD,$33,$83,$FD,$49,$C3,$FE
	.dc.b	$AA,$8F,$E9,$A6,$86,$8C,$24,$43,$FA,$C4,$66,$3F,$AD,$1C,$5C,$46
	.dc.b	$CE,$71,$C8,$D1,$5F,$EC,$ED,$7A,$B5,$E2,$F6,$76,$D7,$6D,$76,$78
	.dc.b	$63,$C7,$1E,$03,$4E,$23,$99,$B8,$F0,$46,$63,$A9,$98,$E2,$68,$86
	.dc.b	$08,$EC,$8F,$85,$A7,$81,$C1,$1C,$10,$F0,$B9,$69,$93,$DE,$E5,$A6
	.dc.b	$63,$D4,$E0,$E2,$43,$04,$76,$48,$F6,$C7,$A0,$E1,$19,$8F,$79,$E1
	.dc.b	$71,$1B,$3A,$47,$23,$45,$7C,$CD,$7A,$B5,$E2,$F6,$76,$D7,$6D,$76
	.dc.b	$78,$63,$C7,$1E,$03,$4E,$23,$99,$B8,$F0,$46,$63,$A9,$98,$E2,$68
	.dc.b	$86,$08,$EC,$8F,$85,$A7,$81,$C1,$1C,$10,$F0,$B9,$69,$93,$DE,$E5
	.dc.b	$A6,$63,$D4,$E0,$E2,$43,$04,$76,$48,$F6,$C7,$A0,$E1,$19,$8F,$79
	.dc.b	$E1,$71,$1B,$3A,$47,$23,$45,$7C,$CD,$7A,$B5,$E2,$F6,$76,$D7,$6D
	.dc.b	$76,$78,$63,$C7,$1E,$03,$4E,$23,$99,$B8,$F0,$46,$63,$A9,$98,$E2
	.dc.b	$68,$86,$08,$EC,$8F,$85,$A7,$81,$C1,$1C,$10,$F0,$B9,$69,$93,$DE
	.dc.b	$E5,$A6,$63,$D4,$FD,$DE,$9F,$F4,$85,$FF,$25,$6D,$7F,$39,$6C,$FF
	.dc.b	$A6,$9F,$F5,$84,$FF,$2B,$DE,$D7,$F3,$F7,$75,$FD,$13,$87,$FD,$62
	.dc.b	$DE,$15,$FF,$A4,$CF,$F5,$75,$7F,$54,$BE,$55,$FC,$D2,$30,$FF,$A2
	.dc.b	$0F,$E6,$F6,$B7,$85,$FE,$9D,$24,$FF,$AC,$DF,$F5,$76,$7F,$EA,$9D
	.dc.b	$3F,$D5,$F3,$FE,$A9,$27,$1F,$FA,$A7,$6F,$EC,$B2,$AF,$F5,$86,$DF
	.dc.b	$EC,$2E,$EF,$F5,$51,$30,$FF,$60,$3E,$A7,$FD,$63,$AF,$AB,$0F,$8D
	.dc.b	$18,$3D,$4F,$FB,$3F,$37,$1F,$FA,$D4,$93,$FE,$A8,$3F,$B6,$ED,$FE
	.dc.b	$A6,$8E,$64,$BF,$AD,$3F,$E9,$D3,$EB,$14,$FE,$EB,$C1,$FE,$CD,$9F
	.dc.b	$BD,$19,$3F,$B5,$E1,$FD,$6C,$95,$FD,$95,$B1,$FD,$B9,$AF,$F6,$E6
	.dc.b	$BC,$A3,$81,$1E,$0D,$78,$35,$E0,$68,$86,$8A,$68,$C1,$0C,$13,$2F
	.dc.b	$F0,$94,$F9,$1B,$3D,$DB,$3D,$1A,$F3,$6B,$CE,$3F,$C4,$13,$CD,$A7
	.dc.b	$32,$C7,$36,$BC,$DA,$F3,$7F,$04,$E5,$AE,$63,$B4,$62,$7C,$64,$79
	.dc.b	$B5,$D1,$82,$18,$24,$74,$23,$E9,$32,$57,$33,$EB,$8C,$B6,$32,$FA
	.dc.b	$5A,$FA,$63,$91,$1E,$0D,$78,$35,$E0,$68,$86,$8A,$68,$C1,$0C,$12
	.dc.b	$7E,$24,$F9,$1B,$3D,$DB,$3D,$1A,$F3,$6B,$CE,$3E,$5C,$F3,$69,$CC
	.dc.b	$B1,$CD,$AF,$36,$BC,$DF,$96,$E5,$AE,$63,$B4,$62,$7C,$64,$79,$B5
	.dc.b	$D1,$82,$18,$24,$74,$23,$E9,$32,$57,$33,$EB,$8C,$B6,$32,$FA,$5A
	.dc.b	$FA,$63,$91,$1E,$0D,$78,$35,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$7E
	.dc.b	$24,$F9,$1B,$3D,$DB,$3D,$1A,$F3,$6B,$CE,$3E,$5C,$F3,$69,$CC,$FB
	.dc.b	$BA,$7F,$AB,$67,$FA,$4E,$9F,$EA,$B8,$C4,$7F,$AB,$CF,$F5,$61,$FE
	.dc.b	$AC,$EA,$FF,$46,$92,$BF,$D2,$6D,$8F,$F4,$C3,$63,$FD,$58,$D8,$FF
	.dc.b	$5D,$E0,$8C,$47,$FA,$D1,$AF,$79,$EA,$47,$AB,$4E,$A5,$7A,$C7,$29
	.dc.b	$ED,$39,$9E,$CD,$7B,$1D,$08,$F4,$69,$D1,$E1,$1D,$23,$27,$42,$9D
	.dc.b	$23,$91,$1E,$46,$A3,$94,$66,$39,$35,$E4,$70,$71,$1C,$23,$04,$38
	.dc.b	$94,$E2,$60,$86,$08,$F3,$72,$D7,$33,$F2,$C9,$3F,$09,$B1,$EB,$6C
	.dc.b	$79,$9B,$1E,$39,$E8,$48,$CC,$78,$1A,$F8,$27,$B1,$1E,$CD,$3B,$15
	.dc.b	$ED,$1C,$A7,$B4,$E6,$7B,$35,$EC,$74,$23,$D1,$A7,$47,$84,$74,$8C
	.dc.b	$9D,$0A,$74,$8E,$44,$79,$1A,$8E,$51,$98,$E4,$D7,$91,$C1,$C4,$70
	.dc.b	$8C,$10,$E2,$53,$89,$82,$18,$23,$CD,$CB,$5C,$CF,$CB,$24,$FC,$26
	.dc.b	$C7,$AD,$B1,$E6,$6C,$78,$E7,$A1,$23,$31,$E0,$6B,$E0,$9E,$C4,$7B
	.dc.b	$34,$EC,$57,$B4,$72,$9E,$D3,$99,$EC,$D7,$B1,$D0,$8F,$46,$9D,$1E
	.dc.b	$11,$D2,$32,$74,$29,$D2,$39,$11,$E4,$6A,$39,$46,$63,$93,$5F,$AB
	.dc.b	$C1,$FE,$94,$52,$8C,$37,$57,$83,$FC,$E2,$30,$D9,$FF,$49,$36,$77
	.dc.b	$43,$FA,$1F,$15,$E1,$FF,$47,$47,$F4,$36,$AC,$FF,$D2,$B0,$D7,$F9
	.dc.b	$EE,$3F,$E9,$FE,$EE,$C4,$BE,$C7,$FA,$B5,$FF,$44,$BC,$31,$FF,$46
	.dc.b	$BF,$E8,$B3,$DF,$3F,$EA,$9D,$FD,$9B,$7F,$93,$09,$B7,$F9,$45,$B8
	.dc.b	$FF,$46,$24,$7F,$46,$6D,$FF,$3B,$6B,$FD,$11,$AF,$F5,$16,$ED,$FD
	.dc.b	$29,$B5,$DA,$BA,$12,$BA,$35,$E8,$D7,$A3,$6B,$A3,$67,$A4,$F2,$21
	.dc.b	$C8,$A7,$22,$CF,$26,$C7,$27,$81,$0E,$05,$8E,$0D,$8E,$0E,$89,$1A
	.dc.b	$6C,$6A,$30,$47,$04,$9F,$A8,$4A,$F9,$2D,$3E,$01,$6B,$D2,$D7,$C8
	.dc.b	$D7,$C2,$DA,$F0,$36,$BC,$15,$D8,$91,$D9,$B5,$D9,$AF,$66,$BD,$9B
	.dc.b	$5D,$9B,$5D,$AB,$A1,$2B,$A3,$5E,$8D,$7A,$36,$BA,$36,$7A,$4F,$22
	.dc.b	$1C,$8A,$72,$2C,$F2,$6C,$72,$78,$10,$E0,$58,$E0,$D8,$E0,$E8,$91
	.dc.b	$A6,$C6,$A3,$04,$70,$49,$FA,$84,$AF,$92,$D3,$E0,$16,$BD,$2D,$7C
	.dc.b	$8D,$7C,$2D,$AF,$03,$6B,$C1,$5D,$89,$1D,$9B,$5D,$9A,$F6,$6B,$D9
	.dc.b	$B5,$D9,$B5,$DA,$BA,$12,$BA,$35,$E8,$D7,$A3,$6B,$A3,$67,$A4,$F2
	.dc.b	$21,$C8,$A7,$22,$CF,$26,$C7,$27,$F8,$06,$9F,$E7,$C7,$F9,$C6,$FF
	.dc.b	$E7,$F2,$6D,$FE,$7E,$8F,$E9,$BB,$31,$FE,$88,$7F,$38,$8C,$B8,$7F
	.dc.b	$CF,$71,$B6,$1F,$D2,$F1,$46,$67,$94,$61,$E5,$3B,$9D,$3D,$23,$2F
	.dc.b	$80,$EA,$57,$9C,$70,$8E,$51,$C2,$7B,$47,$06,$C7,$03,$53,$C6,$7A
	.dc.b	$1E,$13,$83,$99,$E3,$19,$9E,$4E,$9E,$64,$70,$47,$64,$79,$10,$F4
	.dc.b	$9B,$30,$79,$0C,$BC,$5A,$6A,$3C,$AE,$08,$71,$23,$A6,$CF,$B4,$93
	.dc.b	$E9,$3B,$46,$9C,$3D,$E3,$6F,$9C,$EF,$3D,$49,$1A,$9E,$71,$87,$9C
	.dc.b	$F0,$9D,$3D,$63,$2F,$84,$EC,$57,$9C,$70,$8E,$51,$C2,$7B,$47,$06
	.dc.b	$C7,$03,$53,$C6,$7A,$1E,$13,$83,$99,$E3,$19,$9E,$4E,$9E,$64,$70
	.dc.b	$47,$64,$79,$10,$F4,$9B,$30,$79,$0C,$BC,$5A,$6A,$3C,$AE,$08,$71
	.dc.b	$23,$A6,$CF,$B4,$93,$E9,$3B,$46,$9C,$3D,$E3,$6F,$9C,$EF,$3D,$49
	.dc.b	$1A,$9E,$71,$87,$9C,$F0,$9D,$3D,$63,$2F,$84,$EC,$57,$9C,$70,$8E
	.dc.b	$51,$C2,$7B,$47,$06,$C7,$03,$53,$C6,$7A,$1E,$13,$83,$99,$E3,$19
	.dc.b	$9E,$4E,$9E,$64,$70,$47,$64,$79,$10,$F4,$9B,$30,$79,$0C,$BC,$5A
	.dc.b	$6A,$3C,$AE,$08,$71,$23,$A6,$CF,$B4,$93,$F2,$74,$FF,$46,$9F,$E9
	.dc.b	$1C,$7F,$CF,$F9,$FF,$A1,$AF,$F9,$EB,$6F,$FD,$0B,$3A,$9F,$E8,$55
	.dc.b	$FE,$79,$7F,$CF,$98,$7F,$A5,$F4,$FF,$40,$C7,$F3,$8C,$3F,$95,$13
	.dc.b	$1F,$E6,$CD,$8F,$E8,$8D,$78,$46,$88,$E9,$AE,$8C,$10,$C1,$0F,$EB
	.dc.b	$B1,$D5,$B1,$C9,$A7,$12,$9C,$6B,$FA,$E4,$F1,$6C,$F1,$6C,$F1,$6B
	.dc.b	$C7,$1F,$EE,$15,$C5,$B5,$C5,$B3,$C5,$AF,$18,$FF,$59,$1B,$2B,$B6
	.dc.b	$BB,$9E,$D1,$88,$F5,$92,$BC,$AD,$AE,$ED,$9E,$13,$A2,$4E,$9B,$1A
	.dc.b	$8C,$11,$C1,$2B,$C6,$47,$D4,$DA,$F5,$35,$F5,$46,$88,$E9,$AE,$8C
	.dc.b	$10,$C1,$0F,$64,$75,$6C,$72,$69,$C4,$A7,$1A,$F7,$CF,$16,$CF,$16
	.dc.b	$CF,$16,$BC,$6B,$E2,$57,$16,$D7,$16,$CF,$16,$BC,$63,$E5,$1B,$2B
	.dc.b	$B6,$BB,$9E,$D1,$88,$F5,$92,$BC,$AD,$AE,$ED,$9E,$13,$A2,$4E,$9B
	.dc.b	$1A,$8C,$11,$C1,$2B,$C6,$47,$D4,$DA,$F5,$35,$F5,$46,$88,$E9,$AE
	.dc.b	$8C,$10,$C1,$0F,$64,$75,$6C,$72,$69,$C4,$A7,$1A,$F7,$CF,$16,$CF
	.dc.b	$16,$CF,$16,$BC,$6B,$E2,$57,$16,$D7,$16,$CF,$16,$BC,$63,$E5,$1B
	.dc.b	$2B,$B6,$BB,$9F,$F3,$5E,$3F,$E8,$52,$67,$FE,$84,$9F,$F4,$00,$C1
	.dc.b	$93,$5F,$F4,$44,$61,$B1,$99,$E3,$39,$6C,$65,$FF,$41,$B5,$1F,$D2
	.dc.b	$27,$99,$1E,$2D,$78,$B5,$E2,$79,$E3,$A3,$CC,$E2,$43,$89,$4E,$33
	.dc.b	$E7,$72,$43,$31,$D1,$ED,$3E,$17,$B1,$B2,$9C,$1D,$3C,$0F,$01,$A2
	.dc.b	$19,$39,$14,$E4,$70,$30,$70,$9F,$43,$A3,$8B,$87,$53,$E9,$8C,$47
	.dc.b	$22,$38,$24,$6C,$93,$EA,$8C,$B6,$33,$3E,$59,$CB,$63,$2F,$91,$AF
	.dc.b	$9C,$87,$9A,$7A,$11,$E4,$D7,$93,$5E,$27,$AA,$3A,$BC,$CE,$24,$38
	.dc.b	$94,$E3,$3E,$87,$24,$33,$1D,$1E,$D3,$E1,$7B,$1B,$29,$C1,$D3,$C0
	.dc.b	$F0,$1A,$21,$93,$91,$4E,$47,$03,$07,$09,$F4,$3A,$38,$B8,$75,$3E
	.dc.b	$98,$C4,$72,$23,$82,$46,$C9,$3E,$A8,$CB,$63,$33,$E5,$9C,$B6,$32
	.dc.b	$F9,$1A,$F9,$C8,$79,$A7,$A1,$1E,$4D,$79,$35,$E2,$7A,$A3,$AB,$CC
	.dc.b	$E2,$43,$89,$4E,$33,$E8,$72,$43,$31,$D1,$ED,$3E,$17,$B1,$B2,$9C
	.dc.b	$1D,$3C,$0F,$01,$A2,$19,$39,$14,$E4,$70,$30,$70,$9F,$43,$A3,$8B
	.dc.b	$87,$53,$E9,$8C,$47,$23,$F8,$15,$FF,$9E,$53,$49,$6D,$3F,$9B,$E6
	.dc.b	$CB,$FD,$04,$7F,$1F,$3F,$9B,$92,$FF,$9B,$B5,$D1,$82,$18,$21,$FD
	.dc.b	$0D,$E0,$D3,$65,$36,$7F,$34,$76,$D3,$65,$37,$87,$F5,$98,$DB,$63
	.dc.b	$6D,$77,$8F,$F6,$19,$DB,$63,$6D,$77,$97,$C5,$72,$D7,$33,$CA,$31
	.dc.b	$1E,$B2,$1E,$72,$D7,$8D,$B5,$DD,$B3,$C9,$E0,$49,$E0,$D9,$E0,$D9
	.dc.b	$E1,$3A,$24,$E9,$B1,$A8,$C1,$1C,$12,$BD,$84,$AF,$82,$D7,$E0,$98
	.dc.b	$21,$82,$1E,$C7,$83,$4D,$94,$D9,$EB,$76,$D3,$65,$37,$5E,$C8,$DB
	.dc.b	$63,$6D,$77,$5E,$F9,$DB,$63,$6D,$77,$5F,$0D,$CB,$5C,$CF,$28,$C4
	.dc.b	$7A,$C8,$79,$CB,$5E,$36,$D7,$76,$CF,$27,$81,$27,$83,$67,$83,$67
	.dc.b	$84,$E8,$93,$A6,$C6,$A3,$04,$70,$4A,$F6,$12,$BE,$0B,$5F,$82,$60
	.dc.b	$86,$08,$7B,$1E,$0D,$36,$53,$67,$AD,$DB,$4D,$94,$DD,$7B,$23,$6D
	.dc.b	$8D,$B5,$DD,$7B,$E7,$6D,$8D,$B5,$DD,$7C,$37,$2D,$73,$3C,$A3,$11
	.dc.b	$EB,$21,$E7,$2D,$78,$DB,$5D,$DB,$3F,$C0,$72,$FF,$3D,$E3,$FE,$7A
	.dc.b	$9F,$E7,$79,$7F,$9F,$5F,$F3,$39,$FE,$7D,$FA,$A7,$4F,$F0,$26,$E7
	.dc.b	$FE,$7A,$6C,$F0,$CD,$18,$8F,$F3,$4B,$5F,$F3,$DC,$6A,$BA,$12,$BA
	.dc.b	$36,$7A,$57,$17,$C9,$19,$7A,$B5,$EB,$1B,$AE,$D5,$D0,$93,$D2,$B8
	.dc.b	$3D,$23,$2F,$46,$BD,$23,$73,$D2,$39,$4E,$9E,$46,$08,$73,$29,$CD
	.dc.b	$E1,$1C,$E3,$44,$34,$70,$3A,$99,$29,$A3,$0F,$37,$04,$36,$47,$4D
	.dc.b	$AF,$90,$4A,$F7,$B6,$BC,$ED,$9F,$25,$72,$24,$66,$3C,$6D,$7C,$51
	.dc.b	$AA,$EA,$4A,$EA,$D9,$EB,$5C,$5F,$34,$65,$EA,$D7,$AC,$6E,$BB,$57
	.dc.b	$42,$4F,$4A,$E0,$F4,$8C,$BD,$1A,$F4,$8D,$CF,$48,$E5,$3A,$79,$18
	.dc.b	$21,$CC,$A7,$37,$84,$73,$8D,$10,$D1,$C0,$EA,$64,$A6,$8C,$3C,$DC
	.dc.b	$10,$D9,$1D,$36,$BE,$41,$2B,$DE,$DA,$F3,$B6,$7C,$95,$C8,$91,$98
	.dc.b	$F1,$B5,$F1,$46,$AB,$A9,$2B,$AB,$67,$AD,$71,$7C,$D1,$97,$AB,$5E
	.dc.b	$B1,$BA,$ED,$5D,$09,$3D,$2B,$83,$D2,$32,$F4,$6B,$D2,$37,$3D,$23
	.dc.b	$94,$E9,$E4,$60,$87,$32,$9C,$DE,$11,$CE,$34,$43,$47,$03,$A9,$92
	.dc.b	$9A,$30,$F3,$70,$7F,$01,$C7,$FC,$EF,$8F,$F9,$DA,$BF,$CE,$B2,$3F
	.dc.b	$A1,$C6,$5C,$4F,$05,$A7,$F9,$5E,$FF,$9D,$47,$F3,$82,$1F,$CB,$4A
	.dc.b	$7F,$36,$29,$FC,$AC,$A7,$F3,$C2,$BD,$1E,$44,$39,$14,$E4,$53,$91
	.dc.b	$5E,$4F,$02,$1C,$0A,$70,$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$26,$3F
	.dc.b	$DA,$A3,$DA,$D7,$CC,$D7,$C2,$D7,$B3,$5E,$AD,$8E,$B8,$FF,$7B,$6C
	.dc.b	$F6,$8C,$4E,$C8,$E0,$91,$B2,$1E,$47,$2D,$32,$79,$9C,$B4,$C9,$E8
	.dc.b	$72,$D3,$27,$A9,$CB,$4C,$CF,$A9,$F0,$10,$EC,$53,$A1,$4E,$45,$79
	.dc.b	$3C,$08,$70,$29,$C0,$AF,$07,$44,$34,$57,$4E,$08,$E0,$93,$F4,$23
	.dc.b	$DA,$D7,$CC,$D7,$C2,$D7,$B3,$5E,$AD,$8E,$B3,$F2,$9B,$3D,$A3,$13
	.dc.b	$B2,$38,$24,$6C,$87,$91,$CB,$4C,$9E,$67,$2D,$32,$7A,$1C,$B4,$C9
	.dc.b	$EA,$72,$D3,$33,$EA,$7C,$04,$3B,$14,$E8,$53,$91,$5E,$4F,$02,$1C
	.dc.b	$0A,$70,$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$24,$FD,$08,$F6,$B5,$F3
	.dc.b	$35,$F0,$B5,$EC,$D7,$F9,$A9,$FE,$76,$29,$FE,$78,$BF,$F3,$EB,$99
	.dc.b	$71,$FF,$3D,$C9,$B7,$F9,$DE,$30,$DA,$4E,$9F,$E8,$82,$B9,$BA,$9E
	.dc.b	$13,$B8,$FF,$3E,$D7,$47,$8D,$70,$9D,$47,$39,$D3,$C0,$E8,$70,$72
	.dc.b	$EC,$C9,$B8,$E7,$3D,$6B,$C6,$49,$E4,$70,$39,$4F,$0A,$EB,$5C,$67
	.dc.b	$73,$97,$84,$65,$F0,$57,$56,$9C,$1C,$3D,$9B,$1C,$63,$06,$C8,$72
	.dc.b	$21,$A2,$BC,$49,$5E,$82,$47,$99,$B5,$E9,$25,$78,$A3,$2D,$9E,$04
	.dc.b	$9D,$BE,$43,$C4,$4A,$F0,$BB,$9E,$53,$C2,$3C,$F5,$D9,$E5,$5C,$27
	.dc.b	$51,$CE,$74,$F0,$3A,$1C,$1C,$BB,$32,$6E,$39,$CF,$5A,$F1,$92,$79
	.dc.b	$1C,$0E,$53,$C2,$BA,$D7,$19,$DC,$E5,$E1,$19,$7C,$15,$D5,$A7,$07
	.dc.b	$0F,$66,$C7,$18,$C1,$B2,$1C,$88,$68,$AF,$12,$57,$A0,$91,$E6,$6D
	.dc.b	$7A,$49,$5E,$28,$CB,$67,$81,$27,$6F,$90,$F1,$12,$BC,$2E,$E7,$94
	.dc.b	$F0,$8F,$3D,$76,$79,$57,$09,$D4,$73,$9D,$3C,$0E,$87,$07,$2E,$CC
	.dc.b	$9B,$8E,$73,$D6,$BC,$64,$9E,$47,$03,$94,$F0,$AE,$B5,$C6,$77,$39
	.dc.b	$78,$46,$5F,$05,$75,$69,$C1,$C3,$D9,$B1,$C6,$30,$6C,$87,$22,$1A
	.dc.b	$2B,$C4,$95,$E8,$24,$79,$9B,$5E,$92,$57,$8A,$32,$FF,$14,$C1,$8B
	.dc.b	$7F,$E9,$11,$86,$C6,$70,$FE,$86,$7F,$43,$9D,$3F,$D4,$75,$6B,$FE
	.dc.b	$89,$8F,$EA,$FA,$FF,$4C,$7F,$D1,$73,$A7,$FD,$17,$9F,$FA,$49,$DB
	.dc.b	$75,$4E,$EB,$95,$68,$FF,$4C,$6A,$C7,$D7,$8F,$A0,$93,$E8,$9D,$65
	.dc.b	$E1,$8F,$E8,$A7,$F3,$F2,$69,$FD,$8D,$A6,$9C,$11,$C1,$0F,$EB,$47
	.dc.b	$02,$BB,$69,$B3,$FA,$E1,$B2,$BB,$69,$B9,$FE,$E2,$ED,$B1,$B6,$C6
	.dc.b	$F5,$FE,$F3,$39,$6C,$64,$FF,$60,$46,$5A,$64,$ED,$1C,$1F,$49,$0F
	.dc.b	$29,$67,$C2,$D8,$E0,$E8,$91,$A6,$C6,$A3,$04,$70,$48,$F2,$9D,$4A
	.dc.b	$72,$2C,$71,$6C,$71,$7D,$66,$08,$60,$93,$F0,$DA,$7C,$37,$04,$70
	.dc.b	$43,$C8,$70,$2B,$B6,$9B,$3D,$06,$CA,$ED,$A6,$E7,$D8,$ED,$B1,$B6
	.dc.b	$C6,$E7,$E0,$4E,$5B,$19,$3E,$0C,$65,$A6,$4E,$D1,$C1,$F4,$90,$F2
	.dc.b	$96,$7C,$2D,$8E,$0E,$89,$1A,$6C,$6A,$30,$47,$04,$8F,$29,$D4,$A7
	.dc.b	$22,$C7,$16,$C7,$17,$D6,$60,$86,$09,$3F,$0D,$A7,$C3,$70,$47,$04
	.dc.b	$3C,$87,$02,$BB,$69,$B3,$D0,$6C,$AE,$DA,$6E,$7D,$8E,$DB,$1B,$6C
	.dc.b	$6E,$7E,$04,$E5,$B1,$93,$E0,$C6,$5A,$64,$ED,$1C,$1F,$49,$0F,$29
	.dc.b	$67,$C2,$D8,$E0,$E8,$FE,$05,$9F,$FA,$59,$FE,$91,$87,$FA,$6D,$FF
	.dc.b	$4D,$9F,$E8,$8E,$3F,$F3,$A9,$1F,$F3,$9B,$75,$FF,$40,$B7,$1F,$E7
	.dc.b	$CD,$9E,$47,$02,$4F,$06,$CF,$06,$9C,$23,$44,$8D,$36,$35,$18,$23
	.dc.b	$82,$1F,$D5,$A7,$C8,$D3,$B9,$67,$A3,$67,$9B,$4E,$67,$FA,$CE,$79
	.dc.b	$B4,$E6,$59,$E6,$D9,$E6,$D3,$9E,$2A,$E6,$D8,$E6,$DA,$E6,$DA,$E6
	.dc.b	$D9,$E7,$5C,$88,$F2,$6D,$72,$6D,$72,$6C,$F2,$38,$12,$78,$36,$78
	.dc.b	$34,$E1,$1A,$24,$69,$B1,$A8,$C1,$1C,$10,$FA,$04,$3E,$41,$6B,$E0
	.dc.b	$36,$BD,$2D,$3C,$85,$AF,$0B,$6B,$C0,$D3,$C0,$76,$25,$76,$6D,$76
	.dc.b	$69,$D8,$B5,$D9,$B5,$D9,$A7,$6A,$E8,$4A,$E8,$D8,$E8,$DA,$E8,$DA
	.dc.b	$E8,$D9,$E9,$5C,$88,$F2,$6D,$72,$6D,$72,$6C,$F2,$38,$12,$78,$36
	.dc.b	$78,$34,$E1,$1A,$24,$69,$B1,$A8,$C1,$1C,$10,$FA,$84,$3E,$49,$6B
	.dc.b	$E0,$36,$BD,$2D,$3C,$85,$AF,$0B,$6B,$C0,$D3,$C0,$76,$25,$76,$6D
	.dc.b	$76,$69,$D8,$B5,$D9,$B5,$D9,$A7,$6A,$E8,$4C,$FF,$CF,$70,$3F,$CD
	.dc.b	$AE,$1B,$9F,$FA,$0D,$D1,$FE,$83,$70,$EB,$8B,$FD,$12,$47,$FC,$E5
	.dc.b	$3F,$E9,$08,$FF,$45,$C7,$F9,$EC,$99,$FF,$A0,$33,$FF,$45,$36,$3F
	.dc.b	$D0,$E7,$FA,$2A,$33,$B7,$FA,$17,$3F,$F4,$43,$67,$CF,$C7,$FE,$9A
	.dc.b	$23,$E2,$C9,$1F,$E8,$68,$F5,$E7,$FE,$86,$C3,$11,$F0,$1F,$3C,$7F
	.dc.b	$25,$8F,$E9,$44,$D3,$FA,$53,$4F,$EA,$25,$3F,$A0,$14,$FE,$70,$58
	.dc.b	$E9,$1C,$89,$1C,$9A,$F2,$6B,$C9,$B1,$CA,$38,$10,$E0,$53,$81,$5E
	.dc.b	$0E,$88,$68,$AE,$9C,$11,$C1,$35,$FF,$3E,$12,$3F,$CE,$AD,$7E,$03
	.dc.b	$67,$D2,$D8,$F2,$35,$F0,$B5,$F0,$36,$7C,$11,$D8,$87,$62,$CF,$66
	.dc.b	$9D,$8A,$76,$29,$D8,$B1,$D9,$E8,$48,$E8,$D3,$A1,$4E,$85,$3A,$16
	.dc.b	$3A,$47,$22,$47,$26,$BC,$9A,$F2,$6C,$72,$8E,$04,$38,$14,$E0,$57
	.dc.b	$83,$A2,$1A,$2B,$A7,$04,$70,$49,$FA,$84,$8F,$92,$D7,$E0,$36,$7D
	.dc.b	$2D,$8F,$23,$5F,$0B,$5F,$03,$67,$C1,$1D,$88,$76,$2C,$F6,$69,$D8
	.dc.b	$A7,$62,$9D,$8B,$1D,$9E,$84,$8E,$8D,$3A,$14,$E8,$53,$A1,$63,$A4
	.dc.b	$72,$24,$72,$6B,$C9,$AF,$26,$C7,$28,$E0,$43,$81,$4E,$05,$78,$3A
	.dc.b	$21,$A2,$BA,$70,$47,$04,$9F,$A8,$7E,$DF,$3F,$F4,$04,$FF,$48,$6B
	.dc.b	$FD,$3E,$5D,$1F,$D4,$74,$61,$FD,$3A,$95,$A5,$3A,$7F,$9C,$D7,$1A
	.dc.b	$E1,$5B,$AD,$57,$7A,$D3,$C0,$E7,$5C,$23,$31,$B3,$26,$E7,$9B,$E4
	.dc.b	$8D,$B4,$D9,$C6,$37,$5E,$3A,$ED,$3A,$38,$4E,$AB,$9D,$6E,$7C,$15
	.dc.b	$C2,$B5,$1C,$67,$4F,$79,$D1,$1E,$24,$34,$53,$C2,$ED,$C9,$0C,$9B
	.dc.b	$8E,$EE,$0D,$38,$25,$7A,$89,$1E,$76,$D7,$A8,$95,$E1,$9E,$84,$AE
	.dc.b	$6F,$12,$4F,$07,$C9,$5D,$2B,$95,$70,$AD,$57,$9A,$B4,$F0,$39,$D7
	.dc.b	$08,$CC,$6C,$C9,$B9,$E6,$F9,$63,$6D,$36,$71,$8D,$D7,$8E,$BB,$4E
	.dc.b	$8E,$13,$AA,$E7,$5B,$9F,$05,$70,$AD,$47,$19,$D3,$DE,$74,$47,$89
	.dc.b	$0D,$14,$F0,$BB,$72,$43,$26,$E3,$BB,$83,$4E,$09,$5E,$A2,$47,$9D
	.dc.b	$B5,$EA,$25,$78,$67,$A1,$2B,$9B,$C4,$93,$C1,$F2,$57,$4A,$E5,$5C
	.dc.b	$2B,$55,$E6,$AD,$3C,$0E,$75,$C2,$33,$1B,$32,$6E,$79,$BE,$58,$DB
	.dc.b	$4D,$9C,$63,$75,$E3,$AE,$D3,$A3,$84,$EA,$B9,$D6,$E7,$C1,$5C,$2B
	.dc.b	$51,$C6,$74,$F7,$9D,$11,$E2,$43,$45,$3C,$2E,$DC,$90,$C9,$B8,$EE
	.dc.b	$E0,$D3,$82,$57,$A8,$91,$E7,$7F,$3B,$C9,$FC,$FB,$8B,$F9,$2B,$6F
	.dc.b	$F9,$CB,$6B,$FC,$EA,$D7,$FA,$47,$15,$7F,$A2,$B7,$03,$4C,$70,$FF
	.dc.b	$A6,$E7,$FD,$31,$87,$F9,$F2,$31,$46,$D7,$F5,$2C,$BF,$9C,$B7,$1F
	.dc.b	$E9,$AD,$C3,$FD,$12,$D7,$FD,$4F,$CB,$D7,$80,$D2,$47,$F5,$DD,$7F
	.dc.b	$D3,$D5,$FD,$72,$3F,$D3,$C7,$FA,$80,$FF,$52,$F0,$FF,$A7,$5B,$C1
	.dc.b	$FE,$A2,$23,$BD,$3D,$58,$B9,$BC,$44,$9F,$F5,$13,$FE,$9C,$C3,$FB
	.dc.b	$56,$7E,$C3,$FD,$84,$59,$F8,$2E,$1F,$F5,$3E,$AB,$63,$FE,$A8,$9F
	.dc.b	$F6,$23,$71,$FA,$77,$A3,$B4,$74,$23,$F8,$03,$53,$F8,$06,$BF,$E0
	.dc.b	$7D,$FF,$C0,$B3,$F8,$89,$CC,$FF,$71,$3F,$B2,$91,$FE,$DE,$D3,$FC
	.dc.b	$00,$53,$FB,$F1,$4E,$4F,$02,$3C,$1A,$F0,$6B,$C0,$D1,$0D,$14,$D1
	.dc.b	$82,$18,$25,$7F,$87,$DF,$23,$63,$BB,$63,$A3,$5E,$6D,$79,$F0,$7F
	.dc.b	$BE,$27,$8B,$63,$8B,$5E,$2D,$78,$9F,$81,$32,$53,$27,$99,$E1,$3E
	.dc.b	$12,$38,$24,$71,$8C,$B6,$32,$F3,$30,$43,$04,$9F,$5B,$96,$B9,$3D
	.dc.b	$A6,$4A,$64,$F5,$94,$F5,$BC,$88,$F0,$6B,$C1,$AF,$03,$44,$34,$53
	.dc.b	$46,$08,$60,$93,$F1,$1F,$23,$63,$BB,$63,$A3,$5E,$6D,$79,$CF,$CC
	.dc.b	$9E,$2D,$8E,$2D,$78,$B5,$E2,$7C,$93,$25,$32,$79,$9E,$13,$E1,$23
	.dc.b	$82,$47,$18,$CB,$63,$2F,$33,$04,$30,$49,$F5,$B9,$6B,$93,$DA,$64
	.dc.b	$A6,$4F,$59,$4F,$5B,$C8,$8F,$06,$BC,$1A,$F0,$34,$43,$45,$34,$60
	.dc.b	$86,$09,$3F,$11,$F2,$36,$3B,$B6,$3A,$35,$E6,$D7,$9C,$FC,$C9,$E2
	.dc.b	$D8,$F2,$6B,$FE,$B4,$4D,$A7,$5F,$FA,$50,$93,$FE,$91,$6F,$07,$FA
	.dc.b	$79,$B8,$7F,$AC,$9B,$23,$FD,$80,$E1,$A7,$32,$4F,$36,$CF,$36,$9C
	.dc.b	$CF,$F6,$2C,$F3,$69,$A3,$AC,$F3,$24,$F3,$6C,$73,$AF,$7B,$C4,$E7
	.dc.b	$1A,$79,$CF,$12,$3C,$6B,$C5,$5C,$AB,$84,$F2,$AD,$BC,$8E,$07,$79
	.dc.b	$E1,$1A,$8E,$06,$A7,$9C,$EA,$31,$3B,$8C,$47,$52,$1C,$8A,$E9,$C1
	.dc.b	$1C,$12,$BD,$64,$8F,$33,$6B,$C6,$DA,$F1,$36,$7C,$A4,$3C,$2E,$5A
	.dc.b	$74,$24,$F4,$6C,$F4,$69,$CC,$F4,$4F,$36,$9A,$3A,$CF,$32,$4F,$36
	.dc.b	$C7,$3A,$F8,$0F,$13,$9C,$69,$E7,$3C,$48,$F1,$AF,$15,$72,$AE,$13
	.dc.b	$CA,$B6,$F2,$38,$1D,$E7,$84,$6A,$38,$1A,$9E,$73,$A8,$C4,$EE,$31
	.dc.b	$1D,$48,$72,$2B,$A7,$04,$70,$4A,$F5,$92,$3C,$CD,$AF,$1B,$6B,$C4
	.dc.b	$D9,$F2,$90,$F0,$B9,$69,$D0,$93,$D1,$B3,$D1,$A7,$33,$D1,$3C,$DA
	.dc.b	$68,$EB,$3C,$C9,$3C,$DB,$1C,$EB,$E0,$3C,$4E,$71,$A7,$9C,$F1,$23
	.dc.b	$C6,$BC,$55,$CA,$B8,$4F,$2A,$DB,$FC,$8E,$51,$FE,$99,$20,$E1,$CC
	.dc.b	$FF,$15,$8F,$F5,$07,$63,$E8,$29,$23,$81,$FC,$96,$53,$B8,$CC,$7F
	.dc.b	$A6,$1A,$FF,$AB,$1A,$FF,$AB,$1B,$7F,$E7,$AB,$F1,$CF,$A7,$7F,$F5
	.dc.b	$49,$37,$F4,$47,$86,$BF,$D2,$75,$E3,$84,$39,$FF,$A2,$4F,$49,$CC
	.dc.b	$63,$0E,$4D,$8F,$E6,$CD,$7C,$57,$FE,$A9,$7F,$D2,$AD,$73,$9F,$FA
	.dc.b	$CC,$8F,$97,$1F,$F4,$D4,$62,$38,$BE,$7D,$7F,$D6,$AF,$F4,$F6,$C7
	.dc.b	$FA,$DB,$87,$C1,$18,$8F,$F5,$A9,$FE,$B9,$70,$67,$3F,$F5,$AC,$7F
	.dc.b	$B1,$48,$78,$F0,$F8,$CF,$C6,$C3,$FD,$5C,$E8,$F6,$94,$FC,$56,$3F
	.dc.b	$DB,$78,$3E,$55,$FC,$D9,$FF,$52,$B6,$33,$3B,$CF,$C1,$1F,$EB,$7B
	.dc.b	$FF,$5E,$3F,$EB,$9C,$BF,$D5,$F3,$FE,$B8,$26,$1F,$89,$9F,$B1,$AF
	.dc.b	$E3,$5B,$B7,$09,$FE,$ED,$39,$AF,$CA,$36,$3F,$39,$1E,$63,$83,$FA
	.dc.b	$56,$BF,$88,$6E,$5F,$DD,$A3,$FC,$0C,$48,$FF,$03,$34,$FE,$DA,$53
	.dc.b	$FC,$1E,$53,$FC,$0E,$58,$E9,$1C,$89,$1C,$9A,$F2,$6B,$C9,$B1,$CA
	.dc.b	$38,$10,$E0,$53,$81,$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$37,$FF,$59
	.dc.b	$92,$3F,$D6,$0D,$7E,$03,$67,$D2,$D8,$F2,$35,$F0,$B5,$F0,$36,$7C
	.dc.b	$11,$D8,$87,$62,$CF,$66,$9D,$8A,$76,$29,$D8,$B1,$D9,$E8,$48,$E8
	.dc.b	$D3,$A1,$4E,$85,$3A,$16,$3A,$47,$22,$47,$26,$BC,$9A,$F2,$6C,$72
	.dc.b	$8E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70,$49,$FA,$84
	.dc.b	$8F,$92,$D7,$E0,$36,$7D,$2D,$8F,$23,$5F,$0B,$5F,$03,$67,$C1,$1D
	.dc.b	$88,$76,$2C,$F6,$69,$D8,$A7,$62,$9D,$8B,$1D,$9E,$84,$8E,$8D,$3A
	.dc.b	$14,$E8,$53,$A1,$63,$A4,$72,$24,$72,$6B,$C9,$AF,$26,$C7,$28,$E0
	.dc.b	$43,$81,$4E,$05,$78,$3A,$21,$A2,$EB,$FE,$B7,$3F,$D6,$DF,$47,$FE
	.dc.b	$AD,$DF,$FA,$0B,$6F,$FB,$93,$6B,$FD,$3C,$D7,$FC,$24,$D7,$FB,$AB
	.dc.b	$6C,$D2,$42,$BB,$B6,$BB,$B4,$EE,$53,$B9,$4E,$E5,$AE,$ED,$AE,$F5
	.dc.b	$D8,$95,$D9,$B3,$D9,$AF,$66,$BD,$9B,$5D,$9B,$5D,$AB,$A1,$2B,$A3
	.dc.b	$5E,$8D,$7A,$36,$BA,$36,$7A,$4F,$22,$1C,$8A,$72,$2C,$F2,$6C,$72
	.dc.b	$78,$10,$E0,$58,$E0,$D8,$E0,$E8,$91,$A6,$C6,$A3,$04,$70,$48,$FA
	.dc.b	$84,$AF,$92,$DA,$F8,$0D,$3D,$25,$3C,$85,$3C,$25,$AF,$03,$6B,$C1
	.dc.b	$5D,$89,$5D,$9B,$3D,$9A,$F6,$6B,$D9,$B5,$D9,$B5,$DA,$BA,$12,$BA
	.dc.b	$35,$E8,$D7,$A3,$6B,$A3,$67,$A4,$F2,$21,$C8,$A7,$22,$CF,$26,$C7
	.dc.b	$27,$81,$0E,$05,$8E,$0D,$8E,$0E,$89,$1A,$6C,$6A,$30,$47,$04,$8F
	.dc.b	$A8,$4A,$F9,$2D,$AF,$80,$D3,$D2,$53,$C8,$53,$C2,$5A,$F0,$36,$BC
	.dc.b	$15,$D8,$95,$D9,$B3,$D9,$AF,$66,$BD,$9B,$5D,$9B,$5D,$AB,$A1,$2B
	.dc.b	$A3,$5E,$8D,$7A,$3F,$83,$EC,$76,$7F,$AF,$38,$BF,$D3,$91,$89,$FF
	.dc.b	$56,$B6,$33,$3F,$EC,$07,$FC,$F3,$34,$FF,$55,$1F,$EB,$D7,$FC,$DA
	.dc.b	$FF,$AD,$B4,$FF,$5E,$46,$0F,$F3,$9F,$CC,$7F,$B1,$89,$3B,$7C,$8F
	.dc.b	$FA,$3E,$3F,$D6,$99,$FF,$AD,$B1,$C1,$0F,$F5,$79,$FE,$C2,$E2,$FF
	.dc.b	$58,$3F,$EB,$63,$07,$FB,$1B,$A7,$32,$69,$23,$7A,$FF,$A4,$9F,$E8
	.dc.b	$15,$FD,$3C,$FE,$A1,$C7,$FE,$CB,$C3,$CA,$D9,$FF,$6E,$6F,$FE,$AC
	.dc.b	$E8,$71,$A7,$FD,$61,$18,$6C,$67,$93,$FD,$6F,$39,$6C,$67,$2F,$F5
	.dc.b	$F4,$E5,$B1,$9A,$FF,$67,$BF,$EA,$7E,$5E,$D1,$FE,$D0,$8F,$97,$9F
	.dc.b	$8E,$3D,$C4,$DB,$FD,$B6,$F9,$DF,$F7,$36,$BF,$EB,$38,$E1,$B7,$C9
	.dc.b	$7F,$D7,$31,$97,$DD,$19,$BF,$94,$4A,$EA,$D9,$F2,$1F,$EB,$E2,$18
	.dc.b	$3F,$D6,$BB,$FF,$AB,$B8,$0D,$25,$7C,$E3,$FD,$73,$CB,$B2,$39,$AF
	.dc.b	$F5,$FC,$7F,$AD,$A2,$1F,$EB,$D2,$5F,$E8,$1E,$13,$F6,$78,$FF,$4D
	.dc.b	$39,$24,$FF,$83,$27,$FC,$1E,$47,$04,$7F,$C1,$E6,$4A,$E5,$FF,$08
	.dc.b	$99,$2B,$97,$FC,$26,$64,$AE,$71,$FF,$0A,$C6,$5B,$19,$E4,$FC,$03
	.dc.b	$67,$EF,$1E,$22,$1E,$02,$9D,$8B,$3D,$1B,$3D,$1A,$74,$39,$10,$E4
	.dc.b	$58,$E4,$D8,$E4,$D3,$91,$C0,$91,$C1,$B1,$C1,$A7,$08,$D1,$23,$4D
	.dc.b	$8D,$46,$08,$E0,$87,$CB,$72,$D3,$33,$F2,$DC,$11,$C1,$1F,$8A,$64
	.dc.b	$AE,$5F,$80,$64,$AE,$5F,$69,$92,$B9,$9F,$64,$65,$B1,$99,$F4,$B6
	.dc.b	$7D,$27,$88,$87,$80,$A7,$62,$CF,$46,$CF,$46,$9D,$0E,$44,$39,$16
	.dc.b	$39,$36,$39,$34,$E4,$70,$24,$70,$6C,$70,$69,$C2,$34,$48,$D3,$63
	.dc.b	$51,$82,$38,$21,$F2,$DC,$B4,$CC,$FC,$B7,$04,$70,$47,$E2,$99,$2B
	.dc.b	$97,$E0,$19,$2B,$97,$DA,$64,$AE,$67,$D9,$19,$6C,$66,$7D,$2D,$9F
	.dc.b	$49,$E2,$21,$E0,$29,$D8,$B3,$D1,$B3,$D1,$A7,$43,$F8,$07,$0F,$FB
	.dc.b	$0F,$83,$FA,$9C,$7F,$58,$3F,$BE,$F0,$7F,$AE,$C8,$FF,$AB,$DA,$30
	.dc.b	$FE,$B1,$0C,$3F,$D7,$CE,$C9,$3B,$7F,$AB,$BC,$A7,$8C,$F0,$8D,$4F
	.dc.b	$F5,$87,$4F,$03,$9C,$F0,$8C,$C6,$CC,$9B,$79,$BE,$57,$6D,$36,$71
	.dc.b	$77,$3E,$39,$ED,$3A,$38,$4E,$8E,$73,$B8,$F0,$4F,$09,$D4,$71,$9D
	.dc.b	$3D,$E7,$44,$78,$90,$D1,$4F,$0B,$B7,$24,$32,$6E,$3B,$B8,$34,$E0
	.dc.b	$93,$EA,$24,$79,$DA,$7A,$89,$3E,$18,$E8,$49,$E6,$F1,$24,$F0,$7C
	.dc.b	$8F,$49,$E5,$3C,$23,$53,$E6,$74,$F0,$39,$CF,$08,$CC,$6C,$C9,$B7
	.dc.b	$9B,$E5,$76,$D3,$67,$17,$73,$E3,$9E,$D3,$A3,$84,$E8,$E7,$3B,$8F
	.dc.b	$04,$F0,$9D,$47,$19,$D3,$DE,$74,$47,$89,$0D,$14,$F0,$BB,$72,$43
	.dc.b	$26,$E3,$BB,$83,$4E,$09,$3E,$A2,$47,$9D,$A7,$A8,$93,$E1,$8E,$84
	.dc.b	$9E,$6F,$12,$4F,$07,$C8,$F4,$9E,$53,$C2,$35,$3E,$67,$4F,$03,$9C
	.dc.b	$F0,$8C,$C6,$CC,$9B,$79,$BE,$57,$6D,$36,$71,$77,$3E,$39,$ED,$3A
	.dc.b	$38,$4E,$8E,$73,$B8,$F0,$4F,$09,$D4,$71,$9D,$3F,$B7,$D7,$F9,$55
	.dc.b	$7F,$AD,$4F,$F5,$7C,$7F,$A0,$E3,$3C,$7F,$C6,$9B,$3F,$CA,$39,$F7
	.dc.b	$5F,$E7,$E6,$BF,$CC,$DB,$66,$F1,$CA,$61,$FC,$82,$30,$FF,$22,$E4
	.dc.b	$FF,$55,$61,$FE,$8E,$6B,$FE,$A9,$F0,$DE,$7F,$F4,$3E,$DF,$E5,$F1
	.dc.b	$8C,$51,$89,$D4,$7F,$48,$9D,$4F,$F3,$0C,$B3,$7F,$E9,$C6,$BE,$DA
	.dc.b	$E6,$7C,$12,$BE,$5E,$7E,$EF,$F4,$CC,$51,$88,$D5,$FF,$B1,$A3,$11
	.dc.b	$FE,$A6,$9D,$F6,$64,$91,$CE,$51,$C7,$C5,$93,$96,$BF,$DA,$F5,$FF
	.dc.b	$63,$BF,$EC,$C2,$38,$27,$06,$C9,$1E,$3B,$78,$DE,$72,$47,$8D,$F5
	.dc.b	$77,$48,$F3,$35,$F8,$9D,$FE,$F2,$3E,$B7,$FD,$F7,$B7,$F7,$B7,$E9
	.dc.b	$6C,$69,$1F,$B1,$C1,$B6,$9B,$2C,$7D,$F7,$15,$FE,$C9,$3F,$DC,$67
	.dc.b	$FB,$30,$A7,$FB,$6B,$3F,$F0,$67,$3F,$E4,$E7,$FD,$7F,$3F,$E0,$D9
	.dc.b	$CC,$7F,$82,$1B,$1F,$87,$8D,$F1,$FE,$7D,$B8,$EA,$7F,$15,$7F,$AB
	.dc.b	$9F,$F6,$D1,$0E,$FC,$9F,$B4,$8F,$EF,$33,$27,$FC,$06,$D3,$FC,$28
	.dc.b	$53,$FC,$24,$53,$FC,$1C,$5B,$FF,$12,$B7,$6F,$F5,$09,$1F,$F5,$33
	.dc.b	$5D,$18,$21,$82,$4F,$90,$8F,$91,$AF,$91,$AF,$91,$B3,$E4,$6C,$F8
	.dc.b	$5B,$3E,$06,$BE,$03,$B1,$0E,$C5,$3B,$16,$7B,$36,$7B,$34,$EC,$53
	.dc.b	$B1,$D0,$87,$42,$C7,$46,$C7,$46,$9D,$0A,$74,$39,$12,$39,$36,$39
	.dc.b	$34,$E4,$53,$94,$70,$24,$70,$6C,$70,$6B,$C2,$34,$47,$4D,$74,$60
	.dc.b	$86,$09,$3F,$50,$8F,$C9,$6B,$F0,$1A,$FA,$5B,$3E,$46,$CF,$85,$B3
	.dc.b	$E0,$6B,$E0,$3B,$10,$EC,$53,$B1,$67,$B3,$67,$B3,$4E,$C5,$3B,$1D
	.dc.b	$08,$74,$2C,$74,$6C,$74,$69,$D0,$A7,$43,$91,$23,$93,$63,$93,$4E
	.dc.b	$45,$39,$47,$02,$47,$06,$C7,$06,$BC,$23,$44,$74,$D7,$46,$08,$60
	.dc.b	$93,$F5,$08,$FC,$96,$BF,$01,$AF,$A5,$FF,$5C,$F3,$7F,$B0,$EF,$FD
	.dc.b	$77,$21,$0E,$6F,$F6,$3C,$67,$07,$93,$BC,$7F,$D5,$2D,$C7,$FD,$7A
	.dc.b	$DA,$E8,$7F,$B1,$DC,$3C,$DA,$E4,$4F,$46,$CF,$46,$9D,$2B,$C6,$E0
	.dc.b	$E8,$53,$A4,$F3,$24,$F3,$6C,$73,$AF,$1D,$74,$9C,$CF,$46,$BD,$2B
	.dc.b	$91,$1E,$47,$86,$79,$46,$63,$93,$5E,$47,$09,$EB,$3C,$23,$13,$B8
	.dc.b	$C4,$72,$32,$43,$04,$3A,$95,$E2,$E0,$8E,$09,$5F,$08,$91,$EB,$6D
	.dc.b	$79,$9B,$5E,$56,$CF,$8C,$EA,$E1,$F1,$35,$C9,$D4,$93,$D5,$B3,$D5
	.dc.b	$A7,$5A,$F3,$B8,$3A,$94,$EB,$3C,$C9,$3C,$DB,$1C,$EB,$C9,$5D,$27
	.dc.b	$33,$D1,$AF,$4A,$E4,$47,$91,$E1,$9E,$51,$98,$E4,$D7,$91,$C2,$7A
	.dc.b	$CF,$08,$C4,$EE,$31,$1C,$8C,$90,$C1,$0E,$A5,$78,$B8,$23,$82,$57
	.dc.b	$C2,$24,$7A,$DB,$5E,$66,$D7,$95,$B3,$E3,$3A,$B8,$7C,$4D,$72,$75
	.dc.b	$24,$F5,$6C,$F5,$69,$D6,$BC,$EE,$0E,$A5,$3A,$CF,$32,$4F,$36,$C7
	.dc.b	$3A,$F2,$57,$49,$CC,$F4,$6B,$D2,$B9,$11,$E4,$78,$67,$94,$66,$39
	.dc.b	$35,$E4,$70,$9E,$B3,$C2,$31,$3B,$8C,$47,$23,$27,$E8,$F3,$FF,$60
	.dc.b	$92,$7F,$D7,$E7,$FA,$A7,$9F,$FC,$DC,$47,$FC,$D8,$6A,$7F,$CE,$8D
	.dc.b	$7F,$D2,$91,$FE,$A9,$23,$FE,$A5,$7F,$A4,$D3,$B7,$FA,$54,$2B,$FA
	.dc.b	$0D,$2B,$4F,$53,$15,$C2,$77,$3A,$AF,$F3,$C4,$7F,$B0,$2B,$FA,$B6
	.dc.b	$BF,$EE,$AE,$7C,$92,$7D,$6F,$FA,$2C,$8E,$0E,$B8,$7F,$5F,$8F,$35
	.dc.b	$7F,$4B,$8C,$4E,$A3,$C0,$79,$8B,$83,$7F,$76,$7F,$E9,$92,$57,$CC
	.dc.b	$7F,$DD,$9C,$9F,$EC,$AB,$FF,$66,$73,$7F,$B3,$9F,$F6,$D3,$FD,$CB
	.dc.b	$83,$FB,$87,$0F,$F7,$3C,$7F,$D8,$F5,$FD,$D6,$72,$FF,$B3,$30,$FF
	.dc.b	$64,$BF,$DE,$38,$57,$C6,$73,$5C,$2F,$FD,$27,$9B,$9F,$FD,$CF,$CB
	.dc.b	$FE,$C9,$CB,$FD,$A4,$D7,$FD,$65,$AF,$F8,$14,$FC,$3B,$FE,$B7,$23
	.dc.b	$F1,$F9,$3E,$5C,$7F,$81,$F7,$FF,$06,$46,$23,$E5,$37,$0F,$C1,$BF
	.dc.b	$EC,$12,$57,$E5,$1B,$3F,$E0,$AE,$1F,$C6,$4F,$FB,$3A,$3F,$06,$FF
	.dc.b	$B5,$5D,$35,$CB,$B7,$ED,$6B,$E1,$C3,$FC,$40,$D8,$FF,$65,$C6,$27
	.dc.b	$FD,$B1,$1F,$99,$C9,$C1,$FE,$24,$CB,$E5,$36,$33,$87,$E9,$34,$FF
	.dc.b	$69,$5F,$CA,$8F,$F6,$F1,$30,$FC,$FF,$1E,$B1,$FF,$6D,$4F,$FB,$D0
	.dc.b	$93,$FE,$21,$AF,$F1,$21,$1C,$11,$FF,$08,$99,$2B,$97,$FC,$4A,$64
	.dc.b	$AE,$5F,$F1,$91,$92,$B9,$9F,$F1,$7C,$65,$B1,$9E,$4F,$F1,$FB,$67
	.dc.b	$F1,$07,$88,$87,$80,$A7,$62,$CF,$46,$CF,$46,$9D,$0E,$44,$39,$16
	.dc.b	$39,$36,$39,$34,$E4,$70,$24,$70,$6C,$70,$69,$C2,$34,$48,$D3,$63
	.dc.b	$51,$82,$38,$21,$F2,$DC,$B4,$CC,$FC,$B7,$04,$70,$47,$E2,$99,$2B
	.dc.b	$97,$E0,$19,$2B,$97,$DA,$64,$AE,$67,$D9,$19,$6C,$66,$7D,$2D,$9F
	.dc.b	$49,$E2,$21,$E0,$29,$D8,$B3,$D1,$B3,$D1,$A7,$43,$91,$0E,$45,$8E
	.dc.b	$4D,$8E,$4D,$39,$1C,$09,$1C,$1B,$1C,$1A,$70,$8D,$12,$34,$D8,$D4
	.dc.b	$60,$8E,$08,$7C,$B7,$2D,$33,$3F,$2D,$C1,$1C,$11,$F8,$A6,$4A,$E5
	.dc.b	$F8,$06,$4A,$E5,$F6,$99,$2B,$99,$F6,$46,$5B,$19,$9F,$4B,$67,$D2
	.dc.b	$78,$88,$78,$0A,$76,$2C,$F4,$6C,$F4,$69,$D0,$FE,$01,$BF,$FB,$6D
	.dc.b	$FE,$D5,$C3,$FE,$D8,$7F,$C2,$1E,$D7,$FB,$5D,$FF,$06,$B5,$FF,$65
	.dc.b	$35,$FF,$6F,$35,$E2,$7F,$B5,$0E,$25,$38,$94,$E2,$53,$8F,$17,$F8
	.dc.b	$9A,$3A,$60,$69,$27,$84,$65,$B1,$99,$E5,$18,$23,$82,$3E,$66,$BE
	.dc.b	$66,$BE,$66,$BE,$63,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1
	.dc.b	$0C,$12,$7E,$04,$79,$1B,$1D,$DA,$F4,$6B,$CD,$AF,$39,$F9,$0F,$16
	.dc.b	$BC,$5A,$F1,$6B,$C4,$F9,$A7,$12,$9C,$4A,$71,$29,$C6,$7D,$51,$D2
	.dc.b	$7C,$A4,$70,$49,$E3,$19,$6C,$66,$79,$C6,$08,$E0,$8F,$9D,$AF,$9D
	.dc.b	$AF,$9D,$AF,$9C,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$9F,$81,$1E,$46,$C7,$76,$BD,$1A,$F3,$6B,$CE,$7E,$43,$C5,$AF
	.dc.b	$16,$BC,$5A,$F1,$3E,$69,$C4,$A7,$12,$9C,$4A,$71,$9F,$54,$74,$9F
	.dc.b	$29,$1C,$12,$78,$C6,$5B,$19,$9E,$71,$82,$38,$23,$E7,$6B,$E7,$6B
	.dc.b	$E7,$6B,$E7,$38,$10,$E0,$53,$81,$4E,$06,$81,$3A,$DE,$4F,$FB,$0F
	.dc.b	$9F,$FD,$05,$18,$D3,$FD,$84,$D8,$CC,$EE,$7F,$DA,$B1,$FE,$D7,$AF
	.dc.b	$F6,$AB,$5F,$F6,$A9,$FE,$D3,$21,$E2,$3F,$CF,$1C,$28,$FF,$6B,$F1
	.dc.b	$7F,$B3,$E7,$FD,$80,$7F,$B4,$F3,$FF,$68,$5F,$FB,$06,$BF,$D2,$DD
	.dc.b	$FB,$AF,$6D,$23,$11,$FE,$DD,$6C,$66,$3A,$70,$FF,$B8,$FA,$3F,$D4
	.dc.b	$0D,$C3,$FD,$02,$7F,$B3,$1C,$4F,$A6,$BF,$D1,$B5,$FE,$EE,$E4,$FF
	.dc.b	$48,$B7,$33,$4F,$F7,$9E,$3F,$ED,$AB,$E4,$D7,$D2,$D7,$D0,$7F,$B7
	.dc.b	$2B,$FD,$3B,$B7,$52,$47,$FB,$D2,$32,$DB,$FE,$AF,$18,$8F,$43,$5F
	.dc.b	$8E,$D7,$FD,$AD,$C7,$FE,$D5,$7F,$DB,$9C,$3F,$ED,$87,$FD,$B8,$62
	.dc.b	$34,$F0,$32,$5E,$4F,$F7,$2B,$F7,$8F,$88,$43,$CC,$53,$CB,$7F,$85
	.dc.b	$69,$93,$E2,$9F,$EC,$7C,$FE,$D7,$37,$E0,$A3,$19,$FF,$B7,$36,$F8
	.dc.b	$53,$F4,$C9,$38,$3E,$EB,$F8,$08,$E9,$B7,$FB,$56,$56,$8C,$3F,$89
	.dc.b	$CB,$C0,$4A,$F4,$46,$5B,$C1,$FE,$D7,$BF,$CD,$D7,$FB,$6C,$93,$FE
	.dc.b	$DA,$AF,$F6,$BC,$7E,$24,$93,$D3,$3F,$F6,$91,$29,$3F,$ED,$79,$FD
	.dc.b	$67,$17,$EE,$63,$F6,$F3,$A3,$E6,$46,$5B,$C5,$FE,$DD,$E3,$FA,$17
	.dc.b	$FE,$28,$3F,$C3,$64,$3F,$C5,$A5,$34,$60,$86,$08,$7F,$8C,$0E,$05
	.dc.b	$36,$53,$67,$F8,$F0,$D9,$4D,$94,$DE,$3F,$E3,$37,$6D,$76,$D7,$71
	.dc.b	$EE,$7C,$04,$3A,$16,$38,$36,$38,$3A,$24,$69,$B1,$A8,$C1,$1C,$12
	.dc.b	$3C,$51,$96,$C6,$5F,$5B,$5F,$5B,$5F,$59,$C0,$87,$02,$9C,$0A,$70
	.dc.b	$34,$43,$45,$34,$60,$86,$08,$7C,$53,$A9,$4E,$45,$38,$94,$E2,$7C
	.dc.b	$83,$89,$4E,$25,$38,$94,$E3,$1F,$25,$DB,$5D,$B5,$DC,$7C,$97,$C4
	.dc.b	$43,$A1,$63,$83,$63,$83,$A2,$46,$9B,$1A,$8C,$11,$C1,$23,$DF,$19
	.dc.b	$6C,$65,$F5,$B5,$F5,$B5,$F5,$9C,$08,$70,$29,$C0,$A7,$03,$44,$34
	.dc.b	$53,$46,$08,$60,$87,$C5,$3A,$94,$E4,$53,$89,$4E,$27,$C8,$38,$94
	.dc.b	$E2,$53,$89,$4E,$31,$F2,$5D,$B5,$DB,$5D,$C7,$C9,$7C,$47,$F0,$1E
	.dc.b	$4F,$F6,$B3,$FD,$45,$70,$7F,$B5,$E9,$DD,$FE,$C6,$25,$7F,$B1,$1B
	.dc.b	$C7,$FC,$AA,$77,$39,$24,$66,$77,$1F,$D2,$1A,$FF,$83,$5A,$FF,$87
	.dc.b	$1A,$FF,$54,$6C,$F6,$9E,$93,$D8,$E8,$53,$A1,$4E,$85,$3A,$16,$3A
	.dc.b	$3C,$A3,$A9,$1C,$10,$EA,$E5,$A6,$4E,$EE,$5A,$64,$F0,$B9,$69,$98
	.dc.b	$F1,$BD,$08,$72,$29,$C0,$AF,$07,$44,$34,$57,$4E,$08,$E0,$93,$F4
	.dc.b	$09,$1F,$1D,$B3,$E9,$9D,$CE,$49,$19,$9D,$C7,$B5,$AF,$A1,$AF,$95
	.dc.b	$AF,$91,$B3,$DA,$7A,$4F,$63,$A1,$4E,$85,$3A,$14,$E8,$58,$E8,$F2
	.dc.b	$8E,$A4,$70,$43,$AB,$96,$99,$3B,$B9,$69,$93,$C2,$E5,$A6,$63,$C6
	.dc.b	$F4,$21,$C8,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$4F,$D0,$24
	.dc.b	$7C,$76,$CF,$A6,$77,$39,$24,$66,$77,$1E,$D6,$BE,$86,$BE,$56,$BE
	.dc.b	$46,$CF,$69,$E9,$3D,$8E,$85,$3A,$14,$E8,$53,$A1,$63,$A3,$CA,$3A
	.dc.b	$91,$C1,$F6,$73,$FF,$63,$67,$FE,$4C,$3F,$D9,$39,$FF,$B1,$E7,$FC
	.dc.b	$AA,$D7,$FD,$A5,$E4,$55,$48,$C3,$6B,$FD,$A7,$1B,$C1,$C8,$E4,$FF
	.dc.b	$65,$61,$FE,$CE,$3F,$DA,$27,$FA,$33,$9B,$FD,$9A,$FF,$B2,$F8,$BF
	.dc.b	$D8,$64,$7F,$D9,$DC,$BF,$EC,$62,$38,$3F,$DA,$D9,$7F,$A5,$7A,$1B
	.dc.b	$7F,$B3,$9B,$3A,$3F,$DB,$CE,$30,$57,$1A,$FF,$69,$F2,$7F,$B3,$2F
	.dc.b	$FD,$B2,$D7,$FD,$22,$DD,$0D,$CB,$FD,$8D,$06,$92,$7F,$DA,$37,$DD
	.dc.b	$A7,$92,$FF,$D9,$D8,$FA,$39,$FC,$B3,$93,$67,$FB,$B2,$CF,$C0,$BF
	.dc.b	$F4,$EB,$4F,$F7,$BD,$F2,$78,$19,$3E,$99,$0F,$F8,$03,$97,$FB,$9C
	.dc.b	$FF,$48,$F1,$FF,$70,$E5,$FF,$66,$B6,$7F,$D9,$AD,$D3,$FD,$9A,$DB
	.dc.b	$F2,$CF,$FB,$45,$B1,$98,$FF,$6F,$ED,$FE,$DB,$7F,$D9,$F7,$FE,$AE
	.dc.b	$69,$EC,$3D,$47,$FB,$0B,$2E,$B9,$64,$91,$F1,$F9,$F7,$AF,$E0,$9B
	.dc.b	$97,$FB,$4E,$75,$5F,$22,$7A,$ED,$F6,$67,$FD,$B9,$86,$DF,$93,$3A
	.dc.b	$D3,$FD,$A2,$4B,$FF,$71,$3F,$ED,$7C,$7F,$23,$97,$82,$FF,$D8,$07
	.dc.b	$DB,$70,$7F,$B5,$DF,$8F,$8F,$FB,$5D,$BB,$FF,$B5,$DF,$F7,$7C,$FF
	.dc.b	$B5,$9F,$F6,$CE,$1F,$80,$8C,$47,$CC,$7E,$53,$4F,$F7,$1F,$07,$ED
	.dc.b	$63,$FD,$9B,$C5,$FB,$B7,$E6,$B6,$BF,$DA,$BC,$5F,$A6,$9C,$B6,$3F
	.dc.b	$86,$47,$FB,$70,$8E,$0F,$F0,$FD,$FF,$88,$88,$60,$87,$F8,$88,$C9
	.dc.b	$4C,$9F,$E2,$43,$25,$32,$7F,$89,$8C,$94,$CE,$3F,$E4,$07,$2D,$73
	.dc.b	$C9,$FE,$C5,$23,$FE,$C2,$6B,$C1,$AF,$03,$44,$34,$53,$46,$08,$60
	.dc.b	$87,$BC,$87,$B4,$A7,$AC,$B1,$E9,$6C,$79,$1A,$78,$4A,$78,$0A,$78
	.dc.b	$0E,$C4,$3B,$16,$3B,$36,$3B,$34,$EC,$53,$B1,$4E,$C7,$42,$47,$46
	.dc.b	$C7,$46,$9D,$0A,$74,$29,$D2,$39,$12,$39,$36,$39,$35,$E4,$D7,$94
	.dc.b	$70,$23,$C1,$AF,$06,$BC,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$50,$87
	.dc.b	$C9,$29,$F0,$0B,$1E,$96,$C7,$91,$A7,$84,$A7,$80,$A7,$80,$EC,$43
	.dc.b	$B1,$63,$B3,$63,$B3,$4E,$C5,$3B,$14,$EC,$74,$24,$74,$6C,$74,$69
	.dc.b	$D0,$A7,$42,$9D,$23,$91,$23,$93,$63,$93,$5E,$4D,$79,$47,$02,$3C
	.dc.b	$1A,$F0,$7F,$AD,$BF,$ED,$3E,$5F,$EB,$0F,$F5,$C6,$BF,$DD,$38,$BF
	.dc.b	$B3,$3B,$C3,$8B,$FD,$E2,$DF,$43,$FD,$EB,$8B,$FD,$AC,$D9,$EF,$3A
	.dc.b	$7F,$C3,$67,$F4,$68,$D3,$4C,$9D,$5C,$B4,$C8,$8E,$93,$E0,$8C,$C6
	.dc.b	$1F,$0C,$F2,$78,$36,$35,$1E,$19,$E4,$47,$89,$D2,$78,$11,$C1,$23
	.dc.b	$83,$CC,$E2,$E1,$F5,$35,$F1,$9C,$4D,$90,$E4,$43,$89,$92,$99,$39
	.dc.b	$BD,$67,$DB,$1A,$21,$A3,$91,$4E,$4F,$07,$07,$07,$04,$9F,$85,$3D
	.dc.b	$DB,$3D,$27,$4F,$B4,$EB,$1A,$69,$93,$B3,$96,$99,$3C,$64,$8E,$B3
	.dc.b	$E1,$8D,$46,$1F,$14,$F3,$78,$36,$35,$1E,$29,$E4,$47,$89,$D2,$78
	.dc.b	$11,$C1,$23,$83,$CC,$E2,$E1,$F5,$B5,$F1,$9C,$4D,$90,$E4,$43,$89
	.dc.b	$92,$99,$39,$BD,$67,$DB,$1A,$21,$A3,$91,$4E,$4F,$07,$07,$07,$04
	.dc.b	$9F,$85,$3D,$DB,$3D,$27,$4F,$B4,$EB,$1A,$69,$93,$B3,$96,$99,$3C
	.dc.b	$64,$8E,$B3,$E1,$8D,$46,$1F,$14,$F3,$78,$36,$35,$1E,$29,$E4,$47
	.dc.b	$89,$D2,$78,$11,$C1,$23,$83,$CC,$E2,$E1,$F5,$B5,$F1,$9C,$4D,$90
	.dc.b	$E4,$7E,$F7,$7F,$F5,$FE,$7F,$EC,$38,$FF,$54,$C6,$39,$3F,$D0,$F3
	.dc.b	$FE,$E1,$8C,$3F,$CC,$B8,$7F,$DB,$C4,$79,$57,$FB,$11,$B8,$FF,$B8
	.dc.b	$B9,$3F,$DC,$33,$FE,$DF,$3F,$DC,$47,$FA,$B8,$FF,$6C,$4F,$FB,$36
	.dc.b	$FC,$BC,$BE,$28,$FF,$5F,$CA,$73,$1E,$33,$37,$FE,$D2,$6E,$1F,$ED
	.dc.b	$53,$FD,$A9,$5A,$E5,$FF,$69,$61,$FE,$9E,$7F,$D9,$F9,$C9,$EF,$8F
	.dc.b	$FB,$47,$0F,$F6,$B5,$FF,$B2,$7C,$E7,$A8,$F9,$8F,$F7,$4F,$A3,$98
	.dc.b	$C5,$A3,$13,$A8,$FF,$61,$D7,$28,$C6,$3F,$EA,$0D,$2B,$FE,$F0,$AE
	.dc.b	$47,$BE,$FF,$D6,$0E,$23,$FD,$C9,$1F,$EA,$D2,$47,$FB,$B5,$DE,$72
	.dc.b	$3E,$43,$67,$FD,$5F,$CE,$DD,$DF,$FD,$DB,$83,$ED,$36,$74,$D9,$DD
	.dc.b	$7F,$7D,$3F,$DC,$7C,$AE,$4F,$F7,$3B,$63,$31,$82,$2D,$3F,$BE,$E3
	.dc.b	$F8,$9E,$FE,$C4,$9F,$F5,$CE,$08,$F3,$53,$45,$7C,$62,$3F,$87,$7F
	.dc.b	$DA,$55,$23,$FC,$37,$C4,$D7,$8C,$E6,$BE,$A5,$7C,$53,$F2,$11,$F4
	.dc.b	$88,$FD,$CE,$1F,$C0,$C7,$E0,$23,$84,$F9,$38,$7F,$01,$18,$84,$62
	.dc.b	$7F,$DC,$DD,$72,$BC,$B8,$7F,$BC,$6F,$F4,$4D,$D3,$FD,$ED,$5F,$9B
	.dc.b	$9F,$F7,$B9,$1F,$C3,$92,$BF,$50,$D8,$F8,$E7,$EB,$63,$FD,$F0,$41
	.dc.b	$1F,$B3,$3E,$74,$7E,$22,$FF,$C7,$F5,$FB,$FD,$7F,$DD,$F3,$FE,$2D
	.dc.b	$E0,$FF,$77,$3F,$B7,$8F,$CC,$93,$0C,$1F,$C0,$8F,$C4,$15,$FE,$04
	.dc.b	$47,$FC,$57,$1F,$E3,$72,$18,$21,$FE,$43,$32,$53,$27,$F9,$0C,$C9
	.dc.b	$4C,$9F,$E2,$C3,$25,$33,$5F,$E5,$37,$2D,$73,$CD,$FE,$5E,$7A,$1E
	.dc.b	$12,$18,$21,$D4,$C9,$4C,$C7,$57,$2D,$73,$1D,$DE,$04,$8D,$36,$35
	.dc.b	$18,$23,$82,$47,$A6,$32,$D8,$CB,$F0,$1A,$FC,$03,$44,$34,$53,$46
	.dc.b	$08,$60,$87,$B9,$EA,$D3,$91,$4E,$25,$38,$9F,$10,$D9,$4D,$94,$D9
	.dc.b	$F1,$4D,$94,$D9,$4D,$C7,$C4,$72,$D7,$31,$E8,$7B,$1E,$52,$18,$21
	.dc.b	$D4,$C9,$4C,$C7,$57,$2D,$73,$1D,$DE,$04,$8D,$36,$35,$18,$23,$82
	.dc.b	$47,$C2,$8C,$B6,$32,$FC,$06,$BF,$00,$D1,$0D,$14,$D1,$82,$18,$21
	.dc.b	$EE,$7A,$B4,$E4,$53,$89,$4E,$27,$C4,$36,$53,$65,$36,$7C,$53,$65
	.dc.b	$36,$53,$71,$F1,$1C,$B5,$CC,$7A,$1E,$C7,$94,$86,$08,$75,$3F,$DD
	.dc.b	$55,$FC,$27,$9F,$FD,$D3,$18,$3F,$DD,$2E,$1F,$F7,$57,$CC,$FF,$B4
	.dc.b	$C8,$FF,$B6,$9B,$A7,$F7,$B6,$C7,$FB,$B1,$AF,$F4,$2C,$11,$88,$EA
	.dc.b	$73,$24,$F3,$69,$CC,$A7,$38,$E4,$FF,$BB,$DB,$3D,$A3,$67,$69,$E8
	.dc.b	$47,$A3,$5E,$87,$03,$A1,$63,$A4,$6E,$3A,$1C,$88,$72,$74,$F2,$6C
	.dc.b	$72,$8D,$9C,$8E,$06,$0E,$05,$78,$3A,$21,$C4,$87,$03,$25,$32,$F2
	.dc.b	$6C,$FC,$82,$3E,$F6,$CF,$A1,$B1,$E3,$6B,$DA,$7A,$3D,$C8,$E0,$91
	.dc.b	$C2,$3C,$27,$72,$4F,$76,$9D,$CA,$77,$8E,$4F,$66,$CF,$68,$D9,$DA
	.dc.b	$7A,$11,$E8,$D7,$A1,$C0,$E8,$58,$E9,$1B,$8E,$87,$22,$1C,$9D,$3C
	.dc.b	$9B,$1C,$A3,$67,$23,$81,$83,$81,$5E,$0E,$88,$71,$21,$C0,$C9,$4C
	.dc.b	$BC,$9B,$3F,$20,$8F,$BD,$B3,$E8,$6C,$78,$DA,$F6,$9E,$8F,$72,$38
	.dc.b	$24,$70,$8F,$09,$DC,$93,$DD,$A7,$72,$9D,$E3,$93,$D9,$B3,$DA,$36
	.dc.b	$76,$9E,$84,$7A,$35,$E8,$70,$3A,$16,$3F,$35,$DF,$FE,$83,$E9,$FF
	.dc.b	$2B,$CF,$FA,$12,$53,$9C,$FF,$CC,$32,$9C,$C7,$FB,$6A,$75,$3F,$ED
	.dc.b	$92,$1C,$F4,$FF,$54,$5F,$FA,$8C,$FF,$54,$74,$FF,$A4,$9F,$F3,$A1
	.dc.b	$2F,$BE,$09,$D4,$63,$0F,$F4,$D7,$4F,$A3,$1F,$24,$62,$74,$7F,$B8
	.dc.b	$1C,$47,$F9,$CE,$7F,$D6,$84,$7F,$D4,$EF,$FA,$EF,$B7,$FD,$4A,$F5
	.dc.b	$C1,$18,$BF,$2B,$6B,$FA,$23,$5F,$F5,$6F,$88,$F3,$F1,$08,$C3,$FE
	.dc.b	$F8,$6E,$1E,$B6,$DF,$D0,$5E,$1F,$FB,$82,$B9,$36,$33,$3F,$ED,$F9
	.dc.b	$F0,$CE,$F1,$6B,$FE,$D9,$EC,$FF,$6A,$12,$BF,$D8,$EF,$DE,$7C,$D9
	.dc.b	$7F,$B1,$A3,$C8,$DC,$3C,$8D,$9F,$F5,$2E,$3E,$CA,$F9,$17,$FE,$E7
	.dc.b	$DF,$FD,$C4,$4A,$FF,$6A,$B5,$FF,$53,$35,$FC,$47,$27,$FB,$3D,$FF
	.dc.b	$74,$74,$D5,$C6,$EB,$FF,$74,$4E,$63,$1C,$BF,$EE,$88,$C4,$EB,$0F
	.dc.b	$F0,$34,$62,$75,$1F,$EB,$7B,$FF,$7C,$10,$FF,$75,$94,$D1,$82,$1F
	.dc.b	$0B,$4F,$F7,$2E,$BF,$1A,$3F,$DB,$47,$CC,$2E,$BF,$82,$CF,$E3,$83
	.dc.b	$E9,$6D,$F6,$1C,$4F,$FB,$F1,$CE,$BB,$7E,$C5,$7E,$09,$FC,$56,$FF
	.dc.b	$EF,$F6,$E1,$E0,$8C,$3E,$5E,$7F,$BB,$87,$FB,$6A,$30,$47,$F3,$A6
	.dc.b	$4B,$1F,$A5,$E2,$FF,$77,$1F,$EE,$FC,$7D,$51,$E6,$C3,$FD,$CB,$5F
	.dc.b	$BC,$9C,$B6,$32,$FF,$8B,$78,$D8,$77,$8F,$E2,$91,$8B,$FD,$A3,$67
	.dc.b	$F7,$F1,$FE,$FD,$DA,$47,$F0,$AE,$45,$FF,$BA,$0F,$F1,$D3,$89,$FA
	.dc.b	$27,$FB,$D3,$5F,$F7,$B4,$F9,$63,$E9,$E1,$E4,$7C,$37,$CE,$7C,$AD
	.dc.b	$7F,$89,$B6,$BF,$C7,$51,$FE,$43,$21,$FE,$43,$29,$FE,$38,$29,$FE
	.dc.b	$4C,$29,$FE,$3A,$2B,$D1,$E4,$43,$91,$4E,$45,$39,$15,$E4,$F0,$21
	.dc.b	$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$73,$7F,$BC,$09,$1F
	.dc.b	$EE,$D6,$BF,$01,$AF,$A5,$AF,$91,$AF,$85,$AF,$81,$B1,$E0,$8E,$C4
	.dc.b	$3B,$14,$EC,$53,$B1,$4E,$C5,$3B,$15,$EC,$F4,$21,$D0,$A7,$42,$9D
	.dc.b	$0A,$74,$2B,$D1,$E4,$43,$91,$4E,$45,$39,$15,$E4,$F0,$21,$C0,$A7
	.dc.b	$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$47,$D4,$24,$7C,$96,$BF,$01
	.dc.b	$AF,$A5,$AF,$91,$AF,$85,$AF,$81,$B1,$E0,$8E,$C4,$3B,$14,$EC,$53
	.dc.b	$B1,$4E,$C5,$3B,$15,$EC,$F4,$21,$D0,$A7,$42,$9D,$0A,$74,$2B,$D1
	.dc.b	$E4,$43,$91,$4E,$45,$39,$15,$E4,$F0,$21,$C0,$A7,$EE,$78,$FF,$DD
	.dc.b	$8F,$F7,$2E,$36,$1F,$EF,$36,$C7,$FB,$C9,$AF,$F8,$F3,$E1,$7F,$99
	.dc.b	$F1,$FF,$5E,$6C,$64,$FE,$FC,$7F,$94,$DC,$BF,$EF,$16,$B9,$36,$7F
	.dc.b	$95,$5C,$B4,$CE,$08,$CB,$63,$31,$82,$38,$24,$F9,$DA,$79,$DC,$11
	.dc.b	$C1,$0F,$41,$C0,$AE,$DA,$6C,$F6,$39,$69,$93,$D8,$E5,$A6,$67,$D9
	.dc.b	$19,$6C,$66,$7D,$B3,$96,$C6,$4F,$31,$E0,$3C,$A4,$30,$43,$C2,$64
	.dc.b	$A6,$67,$C2,$E5,$AE,$63,$C2,$E5,$AE,$4F,$19,$D0,$91,$C9,$B1,$C1
	.dc.b	$A7,$08,$D1,$23,$4D,$8D,$46,$08,$E0,$93,$F1,$DA,$7C,$77,$04,$70
	.dc.b	$43,$D8,$70,$2B,$B6,$9B,$3E,$03,$96,$99,$3D,$AE,$5A,$66,$7D,$91
	.dc.b	$96,$C6,$67,$DB,$39,$6C,$64,$F3,$1E,$03,$CA,$43,$04,$3C,$26,$4A
	.dc.b	$66,$7C,$2E,$5A,$E6,$3C,$2E,$5A,$E4,$F1,$9D,$09,$1C,$9B,$1C,$1A
	.dc.b	$70,$8D,$12,$34,$D8,$D4,$60,$8E,$09,$3F,$1D,$A7,$C7,$70,$47,$04
	.dc.b	$3D,$87,$02,$BB,$69,$B3,$E0,$39,$69,$93,$DA,$FE,$F7,$7F,$F7,$6C
	.dc.b	$7F,$9A,$F9,$FF,$DD,$CD,$3F,$DA,$E1,$FF,$3C,$37,$4C,$C6,$0F,$F7
	.dc.b	$67,$5F,$FA,$C5,$B7,$FE,$F2,$AF,$F3,$58,$9D,$56,$23,$13,$B7,$FD
	.dc.b	$77,$D1,$FE,$EF,$7F,$DE,$47,$FB,$C0,$F2,$BE,$42,$4E,$0F,$F5,$53
	.dc.b	$FE,$96,$23,$83,$C7,$D4,$F5,$1E,$87,$FB,$CE,$30,$D8,$CD,$AF,$FD
	.dc.b	$1E,$FF,$A1,$AB,$6E,$6F,$D9,$AB,$8B,$FD,$D8,$48,$FF,$78,$4F,$FA
	.dc.b	$03,$AF,$E6,$92,$3E,$8B,$63,$B6,$3E,$28,$F0,$C3,$0F,$0C,$61,$B5
	.dc.b	$EE,$AE,$0F,$DD,$AF,$A7,$87,$FB,$1F,$5F,$C0,$C6,$36,$FC,$2B,$F4
	.dc.b	$08,$7D,$43,$E9,$1F,$EE,$4D,$63,$FE,$A8,$E2,$FF,$75,$AE,$BF,$F7
	.dc.b	$43,$67,$6F,$F8,$0F,$8F,$FD,$D6,$F1,$E3,$FF,$78,$3F,$EF,$CE,$6F
	.dc.b	$F6,$21,$1F,$F6,$13,$FE,$B1,$C7,$FD,$C5,$57,$9F,$FC,$13,$5F,$17
	.dc.b	$6F,$B3,$1F,$15,$BB,$7C,$5E,$3F,$F7,$8D,$7C,$E6,$BF,$EC,$2D,$DD
	.dc.b	$18,$84,$61,$FF,$7B,$B7,$0F,$D3,$9F,$EC,$02,$5F,$E1,$38,$FF,$DD
	.dc.b	$75,$FE,$D9,$8F,$F7,$D1,$E3,$7F,$DB,$5C,$5F,$8C,$D7,$D1,$97,$FB
	.dc.b	$51,$B3,$FE,$DF,$1D,$4E,$BD,$61,$FE,$F3,$9F,$F6,$DC,$E5,$E0,$DA
	.dc.b	$FD,$E9,$27,$F8,$0B,$67,$F0,$99,$7F,$BC,$1F,$A8,$7D,$22,$1F,$BE
	.dc.b	$E7,$71,$7F,$B9,$EF,$F7,$D1,$89,$FF,$70,$37,$83,$F8,$3C,$7F,$B8
	.dc.b	$77,$E0,$FE,$59,$AF,$D3,$6B,$F4,$9B,$BF,$FB,$CA,$3C,$F3,$F8,$33
	.dc.b	$FC,$64,$C3,$FC,$90,$53,$FC,$5E,$53,$FC,$8C,$53,$FC,$88,$58,$FF
	.dc.b	$21,$37,$A3,$F8,$D3,$C8,$EA,$43,$04,$3A,$99,$29,$93,$B9,$92,$99
	.dc.b	$8F,$0B,$96,$B9,$8F,$1B,$D0,$87,$22,$C7,$06,$C7,$07,$44,$8D,$36
	.dc.b	$35,$18,$23,$82,$47,$8A,$32,$D8,$CB,$F2,$4C,$10,$C1,$0F,$73,$C1
	.dc.b	$A6,$CA,$6C,$F8,$46,$4A,$64,$F7,$99,$29,$93,$DC,$64,$A6,$63,$D6
	.dc.b	$D8,$F2,$3D,$8F,$19,$0C,$10,$EE,$64,$A6,$4E,$E6,$4A,$66,$3C,$2E
	.dc.b	$5A,$E6,$3C,$6F,$42,$1C,$8B,$1C,$1B,$1C,$1D,$12,$34,$D8,$D4,$60
	.dc.b	$8E,$09,$1F,$26,$32,$D8,$CB,$F2,$4C,$10,$C1,$0F,$73,$C1,$A6,$CA
	.dc.b	$6C,$F8,$46,$4A,$64,$F7,$99,$29,$93,$DC,$64,$A6,$63,$D6,$D8,$F2
	.dc.b	$3D,$8F,$19,$0C,$10,$EE,$64,$A6,$4E,$E6,$4A,$66,$3C,$2E,$5F,$AB
	.dc.b	$CB,$FE,$EE,$27,$27,$F6,$29,$FE,$D9,$7F,$B3,$F2,$BF,$DB,$64,$8F
	.dc.b	$F6,$FB,$74,$FE,$37,$82,$31,$3F,$E1,$C6,$BF,$E0,$A6,$BF,$D9,$DA
	.dc.b	$FF,$66,$6C,$73,$8E,$E4,$9E,$F3,$A2,$46,$4F,$01,$4F,$01,$4F,$01
	.dc.b	$4F,$01,$67,$C1,$3D,$A3,$13,$DA,$30,$48,$EE,$D7,$BB,$5E,$ED,$7B
	.dc.b	$B6,$7B,$C7,$22,$1C,$8A,$72,$29,$C8,$AF,$27,$81,$0E,$05,$38,$15
	.dc.b	$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$7E,$34,$79,$1A,$F7,$6B,$D1,$AF
	.dc.b	$36,$C7,$38,$FA,$24,$9F,$91,$3A,$24,$64,$F0,$14,$F0,$14,$F0,$14
	.dc.b	$F0,$16,$7C,$13,$DA,$31,$3D,$A3,$04,$8E,$ED,$7B,$B5,$EE,$D7,$BB
	.dc.b	$67,$BC,$72,$21,$C8,$A7,$22,$9C,$8A,$F2,$78,$10,$E0,$53,$81,$5E
	.dc.b	$0E,$88,$68,$AE,$9C,$11,$C1,$27,$E3,$47,$91,$AF,$76,$BD,$1A,$F3
	.dc.b	$6C,$73,$8F,$A2,$49,$F9,$13,$A2,$46,$4F,$01,$4F,$01,$4F,$01,$4F
	.dc.b	$01,$67,$DD,$ED,$3B,$DB,$7F,$B7,$6B,$FD,$A5,$39,$8C,$47,$FB,$62
	.dc.b	$31,$D3,$FE,$3E,$6C,$7F,$AA,$CF,$F4,$37,$47,$FA,$A1,$BC,$3C,$9B
	.dc.b	$5F,$E4,$6F,$45,$F1,$A9,$78,$7F,$EB,$C5,$3A,$73,$97,$FB,$9E,$31
	.dc.b	$3B,$6C,$7F,$B4,$31,$63,$FE,$6F,$BF,$F3,$CE,$DF,$CF,$27,$2D,$9D
	.dc.b	$4F,$A2,$30,$47,$A4,$69,$FF,$50,$D7,$CA,$C3,$D1,$AF,$CC,$8F,$3C
	.dc.b	$FF,$B3,$9D,$1F,$EE,$42,$BE,$43,$AD,$F7,$70,$71,$86,$7D,$9F,$B2
	.dc.b	$D7,$27,$52,$C7,$CD,$5E,$F3,$F1,$2F,$A1,$7F,$EE,$A9,$CC,$61,$FA
	.dc.b	$95,$F8,$6A,$E0,$FF,$A7,$23,$27,$07,$07,$DB,$F7,$9E,$FB,$E3,$7F
	.dc.b	$1F,$8F,$FB,$E2,$73,$18,$BE,$F1,$87,$9D,$A3,$15,$F3,$B6,$FB,$33
	.dc.b	$F2,$08,$7C,$83,$D1,$97,$AD,$F8,$59,$7A,$AB,$E0,$37,$0F,$5B,$67
	.dc.b	$91,$FE,$DE,$BE,$EE,$34,$46,$23,$F1,$07,$88,$A7,$2E,$67,$37,$FB
	.dc.b	$F2,$3F,$E0,$00,$F0,$7F,$2D,$3A,$E7,$FD,$24,$62,$B8,$4E,$6B,$86
	.dc.b	$7F,$5A,$31,$5B,$BE,$33,$9B,$F0,$63,$E8,$E2,$FF,$4A,$65,$FA,$6C
	.dc.b	$BF,$44,$43,$F3,$E5,$3F,$DC,$25,$3E,$69,$F8,$D3,$E6,$EB,$F8,$46
	.dc.b	$99,$36,$7D,$99,$F0,$91,$E9,$B7,$E0,$5F,$C3,$70,$7E,$27,$7E,$53
	.dc.b	$97,$F0,$F1,$99,$FB,$77,$E1,$7B,$CF,$DC,$8C,$3F,$97,$8F,$D4,$69
	.dc.b	$23,$F5,$0D,$8F,$79,$F0,$88,$60,$FD,$A7,$91,$FB,$12,$4F,$EC,$5F
	.dc.b	$74,$7C,$73,$F1,$58,$FB,$DE,$67,$F2,$63,$59,$FD,$57,$F9,$37,$2F
	.dc.b	$F2,$5A,$FE,$3C,$DB,$FE,$44,$D9,$FD,$BF,$2F,$F2,$7A,$F9,$D8,$FF
	.dc.b	$C0,$19,$FE,$32,$FF,$39,$E7,$FE,$82,$DC,$3F,$CC,$DC,$FF,$D0,$63
	.dc.b	$07,$F9,$1E,$A1,$FE,$52,$29,$FE,$52,$29,$FE,$52,$29,$FE,$52,$2E
	.dc.b	$5F,$E6,$16,$F4,$7F,$BA,$88,$FF,$9D,$48,$68,$A7,$84,$87,$84,$A7
	.dc.b	$84,$A7,$84,$B1,$E1,$6C,$78,$1E,$C4,$32,$78,$0E,$C4,$3B,$14,$EC
	.dc.b	$58,$EC,$D8,$EC,$F4,$3B,$19,$3B,$1D,$08,$74,$2C,$74,$6C,$74,$79
	.dc.b	$1D,$0C,$9D,$0E,$44,$8E,$4D,$8E,$4F,$03,$91,$93,$94,$70,$24,$70
	.dc.b	$8D,$1C,$23,$2F,$08,$D4,$62,$35,$18,$23,$B2,$1A,$29,$F2,$48,$7C
	.dc.b	$02,$9E,$92,$9E,$42,$C7,$85,$B1,$E0,$7B,$10,$C9,$E0,$3B,$10,$EC
	.dc.b	$53,$B1,$63,$B3,$63,$B3,$D0,$EC,$64,$EC,$74,$21,$D0,$B1,$D1,$B1
	.dc.b	$D1,$E4,$74,$32,$74,$39,$12,$39,$36,$39,$3C,$0E,$46,$4E,$51,$C0
	.dc.b	$91,$C2,$34,$70,$8C,$BC,$23,$51,$88,$D4,$60,$8E,$C8,$68,$A7,$C9
	.dc.b	$21,$F0,$0F,$DF,$F2,$7F,$C0,$0A,$FF,$0C,$C7,$F6,$16,$BF,$E4,$F8
	.dc.b	$FE,$D4,$EA,$9F,$0D,$FE,$24,$8C,$35,$FF,$11,$C6,$9F,$F2,$13,$79
	.dc.b	$3A,$B6,$38,$9A,$9E,$4F,$F9,$7E,$34,$F4,$8C,$A8,$CB,$B8,$CC,$62
	.dc.b	$35,$18,$23,$DC,$93,$E4,$8C,$B6,$32,$79,$9C,$11,$C1,$0F,$31,$C0
	.dc.b	$AE,$DA,$6C,$F4,$1B,$2B,$B6,$9B,$3D,$66,$CA,$ED,$A6,$E7,$E1,$12
	.dc.b	$7D,$B1,$96,$BE,$42,$1E,$02,$9D,$0B,$1C,$DF,$13,$67,$C4,$D8,$EC
	.dc.b	$6A,$7B,$BD,$48,$75,$2C,$71,$7B,$C7,$07,$44,$8D,$3C,$23,$51,$88
	.dc.b	$D4,$60,$8F,$8C,$93,$E8,$8C,$B6,$32,$7A,$9C,$11,$C1,$0F,$39,$C0
	.dc.b	$AE,$DA,$6C,$F4,$1B,$2B,$B6,$9B,$3D,$66,$CA,$ED,$A6,$E7,$E1,$12
	.dc.b	$7D,$B1,$96,$BE,$42,$1E,$02,$9D,$0B,$1C,$DF,$13,$67,$C4,$D8,$EC
	.dc.b	$6A,$7B,$BD,$48,$75,$2C,$71,$7B,$C7,$07,$44,$8D,$3C,$23,$51,$88
	.dc.b	$D4,$60,$8F,$8C,$93,$E8,$8C,$B6,$32,$7A,$9C,$11,$C1,$0F,$39,$C0
	.dc.b	$AE,$DF,$93,$C7,$FF,$00,$5F,$F2,$FF,$47,$FB,$B4,$8F,$F3,$4C,$DF
	.dc.b	$03,$FF,$00,$67,$F9,$AC,$62,$7F,$DB,$E6,$E7,$4D,$32,$7F,$B5,$BC
	.dc.b	$E7,$B7,$FE,$CE,$8F,$E8,$31,$8D,$FF,$DF,$0D,$CB,$FA,$1B,$67,$FD
	.dc.b	$70,$7F,$9D,$A3,$FD,$E4,$FF,$AE,$87,$88,$F8,$B7,$C9,$7F,$B3,$67
	.dc.b	$31,$87,$FD,$B3,$3B,$9D,$36,$3F,$DB,$A7,$F3,$FF,$30,$BC,$06,$5F
	.dc.b	$D5,$67,$31,$89,$FE,$A7,$18,$9D,$47,$17,$D7,$C3,$FE,$BF,$CB,$D7
	.dc.b	$3E,$D3,$7B,$F3,$7F,$AA,$5F,$D3,$BC,$92,$3F,$DD,$3F,$0D,$EF,$87
	.dc.b	$80,$8C,$36,$7F,$DF,$33,$A9,$CF,$B9,$FE,$FF,$9F,$F7,$9B,$5F,$7D
	.dc.b	$23,$15,$D9,$D4,$FA,$4F,$F7,$97,$1F,$FB,$C5,$C1,$CA,$FC,$64,$7C
	.dc.b	$77,$E5,$CB,$FD,$A4,$7C,$4C,$FE,$34,$FC,$8B,$E9,$18,$8E,$FD,$8A
	.dc.b	$FF,$05,$F3,$7F,$BC,$1F,$F0,$74,$FF,$BD,$0F,$F0,$77,$81,$FE,$F8
	.dc.b	$27,$47,$FB,$D2,$32,$D5,$5C,$DC,$9C,$CF,$F6,$49,$FE,$C6,$C7,$FD
	.dc.b	$F0,$FF,$B2,$8C,$97,$1F,$F6,$53,$FE,$CE,$D7,$E9,$3F,$EF,$36,$BF
	.dc.b	$81,$69,$F9,$27,$E5,$1F,$EF,$93,$A9,$FE,$E1,$3F,$DB,$07,$5C,$FE
	.dc.b	$C4,$75,$6C,$75,$6C,$75,$3E,$E1,$BD,$3F,$DE,$B1,$F8,$99,$E1,$18
	.dc.b	$3F,$DF,$8F,$E9,$C6,$3F,$99,$70,$DB,$6D,$FA,$C8,$C3,$70,$F8,$D8
	.dc.b	$FC,$AB,$F6,$1E,$E3,$F4,$63,$AF,$8B,$67,$F5,$F1,$83,$F5,$5C,$5F
	.dc.b	$BD,$9B,$C9,$FC,$0D,$BC,$9F,$05,$FD,$FC,$EE,$FC,$7C,$3F,$E5,$69
	.dc.b	$F1,$13,$97,$F8,$D3,$FC,$63,$9B,$E0,$3F,$C2,$23,$2F,$F2,$08,$CA
	.dc.b	$E8,$FE,$59,$B7,$F2,$EE,$8F,$89,$3F,$CA,$E3,$0F,$CC,$3F,$CC,$9B
	.dc.b	$31,$C4,$7F,$98,$A3,$13,$FE,$60,$6C,$70,$9D,$BF,$E4,$23,$FC,$A0
	.dc.b	$43,$FC,$CA,$53,$FC,$9A,$53,$81,$A2,$1A,$29,$A3,$04,$30,$43,$FC
	.dc.b	$9A,$75,$29,$C8,$A7,$12,$9C,$78,$3F,$D0,$0F,$16,$BC,$5A,$F1,$6B
	.dc.b	$C7,$A7,$EE,$3D,$09,$1A,$6C,$6A,$30,$47,$04,$8F,$C3,$46,$5B,$19
	.dc.b	$7C,$CD,$7C,$CD,$7C,$CD,$7C,$C7,$22,$1C,$8A,$72,$29,$C8,$A7,$23
	.dc.b	$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$F9,$67,$90
	.dc.b	$A7,$72,$9D,$0A,$73,$29,$CE,$3E,$73,$C5,$AF,$16,$BC,$5A,$F1,$8F
	.dc.b	$9C,$F4,$24,$69,$B1,$A8,$C1,$1C,$12,$3D,$91,$96,$C6,$5F,$33,$5F
	.dc.b	$33,$5F,$33,$5F,$31,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81
	.dc.b	$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3E,$59,$E4,$29,$DC,$A7
	.dc.b	$42,$9C,$CA,$7E,$2F,$91,$CB,$FC,$72,$3F,$BA,$9F,$D9,$CF,$EF,$2E
	.dc.b	$5B,$8F,$F3,$BF,$9D,$FF,$22,$73,$EE,$51,$87,$93,$5F,$EE,$4D,$7F
	.dc.b	$C8,$AD,$78,$F1,$7F,$92,$27,$A4,$E4,$91,$99,$EA,$D7,$AB,$5E,$AD
	.dc.b	$7A,$9F,$E1,$F3,$89,$4E,$25,$38,$94,$E3,$3E,$48,$E9,$3E,$49,$D3
	.dc.b	$67,$51,$89,$D9,$1C,$12,$36,$47,$CA,$D7,$CA,$D7,$CA,$D7,$CA,$70
	.dc.b	$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$4F,$C1,$7A,$B5
	.dc.b	$E4,$D7,$8B,$5E,$33,$F1,$67,$D9,$39,$24,$66,$7A,$B5,$EA,$D7,$AB
	.dc.b	$5E,$A7,$CF,$38,$94,$E2,$53,$89,$4E,$33,$E4,$8E,$93,$E4,$9D,$36
	.dc.b	$75,$18,$9D,$91,$C1,$23,$64,$7C,$AD,$7C,$AD,$7C,$AD,$7C,$A7,$02
	.dc.b	$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18,$24,$FC,$17,$AB,$5E
	.dc.b	$4D,$78,$B5,$E3,$3F,$16,$7D,$93,$92,$46,$67,$AB,$5E,$AD,$7A,$B5
	.dc.b	$EA,$7C,$F3,$89,$4E,$25,$38,$94,$E3,$3E,$48,$E9,$3E,$49,$D3,$67
	.dc.b	$51,$89,$D9,$1F,$F7,$CF,$9B,$FE,$E4,$EA,$FF,$7C,$B6,$3F,$DF,$ED
	.dc.b	$7F,$D3,$5C,$48,$C4,$7F,$BB,$BE,$2D,$F0,$F8,$84,$63,$95,$87,$F3
	.dc.b	$78,$C3,$6B,$84,$3C,$6A,$4A,$E3,$7F,$D1,$67,$51,$88,$FF,$35,$71
	.dc.b	$57,$FE,$00,$F1,$38,$FF,$CE,$19,$7F,$9B,$6B,$54,$AD,$46,$1B,$DD
	.dc.b	$B2,$4F,$F9,$CD,$FE,$AD,$D0,$1D,$E9,$4E,$A7,$31,$83,$FD,$88,$F1
	.dc.b	$9E,$13,$B6,$CF,$FB,$27,$C5,$78,$AF,$91,$FB,$1A,$E6,$30,$FC,$E3
	.dc.b	$11,$AA,$F2,$57,$28,$C4,$7C,$6D,$7F,$DE,$A4,$D3,$FA,$86,$9F,$EF
	.dc.b	$79,$F7,$7E,$B5,$FE,$99,$C7,$B4,$7F,$AC,$23,$1B,$7F,$AE,$1B,$3E
	.dc.b	$E6,$CF,$C5,$87,$84,$F8,$7F,$91,$7F,$ED,$D6,$CF,$FB,$76,$75,$5E
	.dc.b	$01,$1C,$A7,$71,$F8,$2E,$7F,$BC,$47,$FD,$54,$FF,$83,$67,$84,$7D
	.dc.b	$67,$B4,$FD,$26,$C6,$4F,$45,$78,$B8,$3C,$31,$F3,$37,$7C,$F7,$93
	.dc.b	$1C,$CB,$83,$EC,$CE,$71,$FB,$0D,$9E,$4D,$7D,$B8,$BE,$1F,$FC,$29
	.dc.b	$5F,$0D,$B3,$E4,$6B,$C1,$AF,$E7,$78,$3E,$E6,$3F,$0E,$3E,$31,$E2
	.dc.b	$A4,$7A,$31,$F3,$C6,$0D,$46,$72,$F7,$C6,$6B,$99,$0F,$C6,$70,$7E
	.dc.b	$2B,$0F,$0C,$7D,$22,$1B,$3F,$C7,$A3,$CB,$FF,$12,$77,$7F,$91,$BB
	.dc.b	$7F,$59,$18,$EE,$FF,$76,$36,$33,$9F,$EC,$6F,$9C,$E5,$E9,$19,$BE
	.dc.b	$91,$83,$F7,$0E,$27,$FC,$69,$C9,$FC,$12,$7F,$84,$10,$FF,$7A,$14
	.dc.b	$FE,$22,$53,$E0,$9F,$E5,$BD,$7E,$5B,$F2,$78,$3F,$13,$AF,$D1,$8C
	.dc.b	$3F,$A5,$79,$9D,$88,$60,$FC,$01,$B3,$F0,$05,$7F,$5B,$BF,$C3,$7F
	.dc.b	$11,$1B,$AF,$1C,$78,$49,$1F,$92,$79,$70,$7E,$C3,$C4,$FD,$A4,$9A
	.dc.b	$49,$F2,$C7,$C0,$29,$F9,$1E,$0F,$F3,$B4,$7E,$C3,$6E,$64,$7F,$5E
	.dc.b	$FF,$09,$8F,$DE,$B6,$0D,$23,$DB,$3F,$AD,$57,$81,$7F,$E8,$5C,$BF
	.dc.b	$90,$5F,$23,$F9,$29,$6B,$F7,$5C,$DE,$08,$CE,$3F,$BA,$E8,$FE,$6D
	.dc.b	$5F,$04,$FE,$72,$FF,$3C,$CF,$F9,$C4,$64,$FE,$85,$1C,$23,$FA,$81
	.dc.b	$23,$FD,$3D,$9F,$F5,$08,$C4,$7F,$5A,$AF,$EA,$C4,$77,$9F,$B2,$3F
	.dc.b	$B5,$75,$7C,$69,$FE,$C8,$D7,$F4,$5D,$5F,$D9,$A7,$79,$7F,$AB,$E3
	.dc.b	$FB,$A4,$7F,$71,$8C,$75,$F1,$AF,$EF,$8D,$5D,$BF,$E0,$19,$FF,$5C
	.dc.b	$36,$BF,$C1,$31,$8E,$DE,$1D,$EE,$FF,$F5,$B6,$1F,$E9,$C3,$FD,$2A
	.dc.b	$43,$FD,$46,$53,$FD,$46,$53,$81,$A2,$1A,$29,$A3,$04,$30,$43,$FD
	.dc.b	$66,$75,$29,$C8,$A7,$12,$9C,$63,$D8,$FF,$B4,$E3,$DA,$E8,$91,$A7
	.dc.b	$B3,$5E,$CD,$7B,$35,$ED,$1D,$23,$11,$D2,$30,$47,$C4,$43,$BB,$96
	.dc.b	$99,$3C,$2E,$5A,$64,$F1,$B9,$69,$93,$CA,$53,$C8,$58,$F3,$92,$39
	.dc.b	$47,$03,$C6,$43,$C2,$53,$B9,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$0F
	.dc.b	$39,$D4,$A7,$22,$9C,$4A,$71,$8F,$63,$EF,$8F,$6B,$A2,$46,$9E,$CD
	.dc.b	$7B,$35,$EC,$D7,$B4,$74,$8C,$47,$48,$C1,$1F,$11,$0E,$EE,$5A,$64
	.dc.b	$F0,$B9,$69,$93,$C6,$E5,$A6,$4F,$29,$4F,$21,$63,$CE,$48,$E5,$1C
	.dc.b	$0F,$19,$0F,$09,$4E,$E5,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3C,$E7
	.dc.b	$53,$FC,$22,$7F,$82,$4A,$7F,$81,$FA,$7F,$A8,$B9,$32,$7F,$9F,$E5
	.dc.b	$F2,$1F,$E8,$1E,$AD,$46,$27,$FC,$5A,$75,$71,$1A,$9E,$13,$96,$BF
	.dc.b	$E1,$B9,$DB,$93,$99,$98,$DB,$E3,$6A,$8E,$2E,$CE,$26,$CA,$6E,$38
	.dc.b	$C7,$08,$D4,$F9,$DD,$C6,$5E,$06,$27,$89,$1C,$12,$79,$11,$F6,$35
	.dc.b	$F3,$35,$F0,$9A,$21,$A2,$9A,$30,$43,$04,$3C,$EF,$56,$9C,$8A,$71
	.dc.b	$29,$C6,$7E,$01,$27,$DD,$3C,$E3,$13,$EF,$3A,$38,$8D,$4F,$09,$CB
	.dc.b	$5F,$04,$ED,$C9,$CC,$CC,$6D,$F2,$35,$F3,$92,$39,$3C,$0E,$47,$02
	.dc.b	$9B,$8E,$51,$C2,$35,$3E,$87,$71,$97,$81,$89,$E2,$47,$04,$9E,$44
	.dc.b	$7D,$AD,$7C,$CD,$7C,$26,$88,$68,$A6,$8C,$10,$C1,$0F,$3B,$D5,$A7
	.dc.b	$22,$9C,$4A,$71,$9F,$80,$49,$F7,$4F,$38,$C4,$FB,$CE,$8E,$23,$53
	.dc.b	$C2,$72,$D7,$C1,$3B,$72,$73,$33,$1B,$7C,$8D,$7C,$E4,$8E,$4F,$03
	.dc.b	$91,$C0,$A6,$E3,$94,$70,$8D,$4F,$A1,$DC,$65,$E0,$62,$78,$91,$C1
	.dc.b	$27,$91,$1F,$D8,$6A,$D7,$FD,$FB,$DD,$FE,$55,$24,$FF,$35,$9D,$68
	.dc.b	$9F,$E6,$90,$F8,$AE,$35,$B6,$BF,$E6,$46,$A1,$1C,$6B,$A4,$7F,$9C
	.dc.b	$39,$EB,$FE,$CD,$E4,$71,$FF,$A3,$B1,$FE,$89,$18,$AD,$D7,$FB,$8A
	.dc.b	$7C,$BE,$87,$76,$C6,$8F,$0C,$65,$F7,$35,$FF,$3D,$BE,$7E,$FE,$24
	.dc.b	$8D,$87,$FA,$D6,$2E,$A7,$8D,$C6,$BF,$D8,$91,$87,$81,$88,$D3,$F1
	.dc.b	$67,$94,$62,$1D,$6F,$01,$F2,$9F,$62,$FF,$DE,$53,$98,$C5,$7C,$F8
	.dc.b	$C4,$FD,$93,$75,$C9,$C4,$FC,$FD,$BF,$D7,$7C,$5F,$39,$B1,$F5,$63
	.dc.b	$1B,$7F,$5C,$C3,$C9,$ED,$30,$7B,$37,$0F,$16,$4A,$D4,$E6,$7D,$B3
	.dc.b	$9A,$E2,$79,$AB,$6F,$CA,$F5,$FE,$E3,$6B,$D2,$78,$76,$4E,$A3,$E6
	.dc.b	$46,$1D,$9F,$EE,$7C,$FD,$64,$7D,$C7,$D8,$F8,$5F,$C1,$5D,$8F,$C0
	.dc.b	$37,$83,$D4,$FE,$06,$77,$3F,$03,$D8,$F8,$97,$EF,$D7,$ED,$46,$2F
	.dc.b	$E9,$B6,$B8,$42,$BA,$FE,$B2,$F6,$BF,$C4,$13,$FA,$BA,$FA,$13,$97
	.dc.b	$84,$64,$46,$6B,$9D,$72,$AF,$C0,$70,$7F,$85,$39,$BF,$59,$5F,$0E
	.dc.b	$72,$FF,$8D,$F2,$F4,$E3,$DA,$7C,$F3,$ED,$E0,$67,$EB,$9F,$BA,$D7
	.dc.b	$E8,$1F,$38,$EA,$53,$F8,$36,$1E,$27,$07,$03,$67,$F9,$1F,$BF,$F8
	.dc.b	$67,$8F,$FE,$36,$F2,$1D,$FF,$E4,$A8,$C3,$72,$E1,$69,$CD,$F2,$9C
	.dc.b	$C6,$27,$51,$83,$E3,$BF,$A2,$21,$FC,$64,$A6,$8C,$10,$FB,$A7,$C7
	.dc.b	$C7,$F4,$D9,$FE,$08,$FE,$05,$1A,$78,$3F,$13,$99,$B7,$FB,$FB,$4F
	.dc.b	$E4,$53,$E0,$3F,$07,$BF,$E0,$5F,$E1,$9C,$1F,$85,$7F,$02,$76,$36
	.dc.b	$5E,$0F,$BA,$F9,$71,$C7,$07,$DC,$6E,$7E,$89,$FD,$6F,$0F,$F3,$DF
	.dc.b	$13,$F4,$7C,$88,$C5,$FE,$A6,$73,$18,$AF,$8C,$43,$DD,$AF,$B7,$6F
	.dc.b	$7C,$7F,$2E,$C3,$F7,$5C,$3D,$B9,$7F,$A6,$36,$4D,$3F,$29,$CD,$FC
	.dc.b	$57,$D9,$BA,$1A,$47,$E7,$D3,$8B,$E6,$B8,$6B,$FC,$73,$A7,$F9,$06
	.dc.b	$1F,$A1,$6D,$7F,$2E,$6C,$EE,$3C,$A7,$F4,$15,$B7,$B7,$AB,$E5,$46
	.dc.b	$63,$0E,$8F,$EA,$2F,$F5,$C2,$38,$3F,$AA,$DF,$F6,$02,$38,$3F,$B7
	.dc.b	$3F,$CE,$88,$E0,$FE,$F5,$5F,$EB,$27,$FB,$D3,$5F,$EF,$5D,$BF,$DE
	.dc.b	$67,$FC,$07,$DD,$F0,$A7,$FC,$11,$DF,$5F,$E2,$D5,$FE,$0A,$AF,$F0
	.dc.b	$4C,$E6,$79,$B5,$FF,$4E,$9F,$EA,$12,$1F,$EB,$32,$9F,$EA,$D2,$9F
	.dc.b	$EA,$D2,$9C,$8E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30
	.dc.b	$4F,$07,$FD,$A8,$F9,$1A,$F7,$6B,$D1,$AF,$36,$BC,$FC,$1F,$F7,$23
	.dc.b	$96,$B9,$33,$1C,$09,$1A,$6C,$6A,$30,$47,$04,$3D,$0E,$5A,$64,$F5
	.dc.b	$39,$69,$93,$D8,$E5,$A6,$4F,$51,$4F,$51,$D8,$87,$42,$9C,$8A,$72
	.dc.b	$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$8F
	.dc.b	$9E,$F9,$1A,$F7,$6B,$D1,$AF,$36,$BC,$E3,$E5,$39,$6B,$93,$E5,$47
	.dc.b	$02,$46,$9B,$1A,$8C,$11,$C1,$0F,$43,$96,$99,$3D,$4E,$5A,$64,$F6
	.dc.b	$39,$69,$93,$D4,$53,$D4,$76,$21,$D0,$A7,$22,$9C,$8A,$72,$38,$10
	.dc.b	$E0,$53,$81,$4E,$06,$88,$68,$FF,$04,$9F,$E3,$11,$DB,$FE,$10,$BF
	.dc.b	$DA,$B6,$34,$6B,$B7,$F9,$53,$FE,$A3,$B7,$CA,$FF,$A9,$7B,$77,$18
	.dc.b	$9F,$F2,$11,$DD,$C4,$6A,$78,$4E,$5A,$FF,$91,$67,$6E,$4E,$66,$63
	.dc.b	$6F,$8D,$AA,$38,$BB,$38,$9B,$29,$B8,$E3,$1C,$23,$53,$E7,$77,$19
	.dc.b	$78,$18,$9E,$24,$70,$49,$E4,$47,$D8,$D7,$CC,$D7,$C2,$68,$86,$8A
	.dc.b	$68,$C1,$0C,$10,$F3,$BD,$5A,$72,$29,$C4,$A7,$19,$F8,$04,$9F,$74
	.dc.b	$F3,$8C,$4F,$BC,$E8,$E2,$35,$3C,$27,$2D,$7C,$13,$B7,$27,$33,$31
	.dc.b	$B7,$C8,$D7,$CE,$48,$E4,$F0,$39,$1C,$0A,$6E,$39,$47,$08,$D4,$FA
	.dc.b	$1D,$C6,$5E,$06,$27,$89,$1C,$12,$79,$11,$F6,$B5,$F3,$35,$F0,$9A
	.dc.b	$21,$A2,$9A,$30,$43,$04,$3C,$EF,$56,$9C,$8A,$71,$29,$C6,$7E,$01
	.dc.b	$27,$DD,$3C,$E3,$13,$EF,$3A,$38,$8D,$4F,$09,$CB,$5F,$04,$ED,$C9
	.dc.b	$CC,$CC,$6D,$F2,$35,$F3,$92,$39,$3C,$0E,$47,$02,$9B,$8E,$51,$C2
	.dc.b	$35,$3E,$87,$71,$97,$81,$89,$E2,$47,$04,$9E,$44,$7F,$61,$9B,$C6
	.dc.b	$6C,$F0,$FF,$98,$CE,$63,$15,$FC,$E6,$31,$3A,$AE,$2F,$F3,$8E,$37
	.dc.b	$47,$FA,$C2,$30,$DB,$FE,$85,$3A,$E9,$EB,$1F,$EE,$CA,$FE,$83,$E8
	.dc.b	$3D,$44,$E8,$B1,$97,$01,$C4,$F2,$9E,$2F,$4F,$92,$B5,$FD,$4E,$91
	.dc.b	$88,$EE,$73,$70,$F8,$1A,$B3,$F8,$DA,$FC,$3D,$7F,$9E,$C6,$27,$E4
	.dc.b	$6B,$F3,$AB,$E1,$E7,$FD,$0B,$4F,$F7,$C5,$7F,$BD,$8F,$F5,$DC,$FF
	.dc.b	$67,$C1,$E0,$BE,$2F,$FB,$04,$79,$8C,$B8,$AF,$F6,$CB,$5F,$BB,$3C
	.dc.b	$55,$7E,$01,$F0,$1F,$EB,$9D,$FF,$DC,$F1,$E6,$24,$7D,$88,$CB,$71
	.dc.b	$F5,$46,$23,$D5,$87,$91,$EE,$F9,$EF,$C5,$7E,$19,$F2,$E1,$E7,$AE
	.dc.b	$13,$99,$F2,$61,$D8,$91,$E4,$8F,$F7,$BE,$1C,$70,$FC,$8C,$78,$0C
	.dc.b	$94,$F8,$71,$BF,$56,$46,$70,$F9,$F1,$8B,$F3,$47,$D2,$F7,$3B,$12
	.dc.b	$B8,$60,$8F,$A1,$1E,$8A,$F3,$3F,$70,$F9,$0F,$DD,$95,$F7,$BF,$8C
	.dc.b	$DA,$FA,$B9,$FE,$19,$E4,$63,$1F,$A0,$E0,$87,$84,$FF,$18,$72,$FF
	.dc.b	$8B,$7C,$6F,$F0,$DB,$72,$FC,$FB,$67,$F5,$ED,$7F,$C6,$5C,$FF,$B2
	.dc.b	$AE,$3C,$E8,$C5,$F2,$8C,$52,$B7,$5A,$8F,$DD,$E3,$FC,$05,$FD,$19
	.dc.b	$0F,$C6,$14,$FD,$9B,$B7,$24,$3F,$82,$E7,$F6,$1A,$63,$4F,$9A,$E0
	.dc.b	$FD,$1E,$9F,$A9,$74,$D7,$CB,$A3,$6F,$F6,$4E,$1E,$19,$F1,$CE,$5F
	.dc.b	$31,$FA,$ED,$B9,$63,$F8,$78,$F0,$35,$FC,$5E,$3F,$87,$72,$F9,$9C
	.dc.b	$E3,$CA,$31,$3F,$84,$9D,$DE,$48,$E0,$F7,$1F,$A8,$3C,$A5,$3E,$41
	.dc.b	$FB,$03,$C3,$BF,$72,$3C,$B1,$F8,$0F,$C4,$CB,$E1,$D7,$C1,$8E,$67
	.dc.b	$53,$E2,$C6,$88,$FE,$7E,$BA,$BF,$09,$B7,$CA,$FB,$CF,$33,$C0,$7F
	.dc.b	$2D,$21,$FC,$B2,$3F,$DF,$BC,$3F,$D5,$E3,$F9,$F1,$FE,$88,$E3,$FE
	.dc.b	$7A,$DE,$3F,$E9,$7B,$68,$FE,$8D,$1B,$E5,$CB,$FD,$15,$AF,$F4,$4A
	.dc.b	$FE,$A1,$3F,$CE,$89,$1F,$C8,$DA,$FF,$4A,$BF,$ED,$51,$FD,$14,$91
	.dc.b	$E9,$7F,$CF,$CC,$3F,$D2,$05,$3F,$D1,$45,$3F,$D1,$45,$3F,$D0,$65
	.dc.b	$E7,$FF,$51,$37,$9F,$FD,$56,$F0,$3A,$90,$EA,$53,$46,$08,$60,$87
	.dc.b	$80,$E0,$53,$65,$37,$1E,$47,$6D,$76,$D7,$71,$E8,$79,$12,$34,$D8
	.dc.b	$D4,$60,$8E,$09,$1F,$0A,$32,$D8,$CB,$F1,$CC,$10,$C1,$0F,$4B,$C1
	.dc.b	$A6,$CA,$6C,$F6,$BB,$69,$B2,$9B,$3E,$09,$92,$99,$3E,$01,$92,$99
	.dc.b	$8F,$6B,$63,$C0,$F2,$3C,$A4,$3C,$25,$34,$60,$86,$08,$78,$0E,$05
	.dc.b	$36,$53,$71,$E4,$76,$D7,$6D,$77,$1E,$87,$91,$23,$4D,$8D,$46,$08
	.dc.b	$E0,$91,$F1,$E3,$2D,$8C,$BF,$1C,$C1,$0C,$10,$F4,$BC,$1A,$6C,$A6
	.dc.b	$CF,$6B,$B6,$9B,$29,$B3,$E0,$99,$29,$93,$E0,$19,$29,$98,$F6,$B6
	.dc.b	$3C,$0F,$23,$CA,$43,$C2,$53,$46,$0F,$EF,$93,$FC,$2B,$95,$C9,$FD
	.dc.b	$FA,$7F,$C1,$AD,$76,$7F,$9A,$8F,$F3,$CD,$BD,$AF,$F3,$DF,$36,$65
	.dc.b	$39,$7F,$CA,$A7,$17,$11,$B6,$C6,$67,$9B,$4E,$27,$F9,$EA,$36,$D8
	.dc.b	$DB,$4D,$9E,$67,$2D,$32,$79,$9C,$B4,$CC,$76,$78,$4F,$9E,$77,$39
	.dc.b	$9E,$13,$97,$8C,$E8,$8E,$09,$1B,$23,$EF,$6B,$EB,$6B,$E2,$34,$43
	.dc.b	$45,$34,$60,$86,$08,$7A,$0E,$05,$36,$53,$73,$EF,$24,$FB,$27,$CB
	.dc.b	$3B,$24,$E9,$F7,$9D,$9C,$47,$06,$C6,$67,$A3,$4E,$27,$92,$36,$D8
	.dc.b	$DB,$4D,$9E,$87,$2D,$32,$79,$9C,$B4,$CC,$76,$78,$4F,$9E,$77,$39
	.dc.b	$9E,$13,$97,$8C,$E8,$8E,$09,$1B,$23,$EF,$6B,$EB,$6B,$E2,$34,$43
	.dc.b	$45,$34,$60,$86,$08,$7A,$0E,$05,$36,$53,$73,$EF,$24,$FB,$27,$CB
	.dc.b	$3B,$24,$E9,$F7,$9D,$9C,$47,$06,$C6,$67,$A3,$4E,$27,$92,$36,$D8
	.dc.b	$DB,$4D,$9E,$87,$2D,$32,$79,$9C,$B4,$CC,$76,$78,$4F,$9E,$77,$39
	.dc.b	$9E,$13,$97,$8C,$E8,$8E,$09,$1B,$3F,$53,$BB,$8B,$FD,$2D,$D1,$FC
	.dc.b	$AE,$31,$C6,$8C,$4E,$AF,$FD,$F1,$1F,$CC,$76,$79,$49,$E0,$7F,$39
	.dc.b	$8C,$B8,$3F,$CB,$91,$B9,$E4,$D8,$FE,$81,$C1,$FE,$70,$9F,$F3,$99
	.dc.b	$23,$3C,$5B,$7F,$D7,$BE,$AD,$7F,$D8,$AD,$59,$A3,$0D,$8F,$21,$FE
	.dc.b	$E0,$7F,$DC,$CF,$FB,$84,$FF,$49,$E9,$E6,$7F,$9D,$E5,$EC,$8C,$12
	.dc.b	$3E,$10,$E4,$73,$3E,$2F,$FA,$CC,$68,$CB,$89,$FF,$68,$35,$E5,$3C
	.dc.b	$55,$75,$7F,$DF,$5B,$FF,$4E,$C7,$D2,$4E,$8F,$9D,$8F,$F6,$B8,$FF
	.dc.b	$6D,$EB,$C6,$FF,$DD,$75,$A7,$FD,$CE,$D7,$C3,$97,$8E,$FC,$F1,$E2
	.dc.b	$7D,$59,$79,$72,$FF,$6A,$D7,$A6,$BE,$51,$F3,$78,$3F,$DC,$93,$F8
	.dc.b	$CC,$FF,$B4,$E1,$D2,$38,$D7,$53,$E1,$94,$F1,$96,$3E,$6F,$C5,$F9
	.dc.b	$71,$34,$93,$B8,$CB,$63,$35,$FE,$EC,$7D,$1A,$7C,$AC,$7C,$F5,$F2
	.dc.b	$5A,$FE,$10,$F8,$33,$CB,$0B,$1D,$4F,$41,$F0,$89,$5D,$71,$FB,$17
	.dc.b	$D2,$7E,$34,$60,$FF,$04,$F2,$FF,$8A,$39,$BF,$5D,$D7,$FE,$16,$8C
	.dc.b	$52,$31,$CE,$AD,$E9,$FE,$2C,$9C,$C6,$2B,$F7,$17,$CA,$73,$1C,$EB
	.dc.b	$81,$FE,$F4,$7E,$C1,$0F,$F7,$86,$3F,$EF,$B7,$46,$31,$FF,$1B,$67
	.dc.b	$F5,$A3,$F8,$33,$F1,$9F,$A0,$60,$CE,$BF,$59,$FC,$B1,$FE,$FB,$26
	.dc.b	$9F,$EF,$E3,$F8,$BB,$FE,$47,$9E,$CF,$42,$3D,$C8,$F6,$D7,$F6,$4D
	.dc.b	$3C,$47,$EC,$1C,$BF,$90,$3F,$2A,$48,$FD,$0E,$DA,$8F,$2E,$3F,$0D
	.dc.b	$FE,$49,$7E,$78,$C5,$7E,$AA,$79,$1F,$9E,$29,$FB,$43,$ED,$3F,$B9
	.dc.b	$DF,$F9,$76,$DF,$12,$D1,$89,$D5,$70,$3E,$99,$0F,$1D,$F9,$5F,$E4
	.dc.b	$AF,$93,$2F,$E1,$D1,$E2,$9F,$31,$EE,$7E,$28,$E1,$FE,$7A,$FF,$25
	.dc.b	$21,$FD,$07,$97,$FD,$E5,$8F,$C1,$D7,$FC,$F1,$3F,$19,$B1,$FC,$CF
	.dc.b	$3F,$EA,$5C,$7E,$EA,$FE,$30,$D7,$F9,$F9,$FE,$87,$1C,$AE,$5D,$47
	.dc.b	$F5,$38,$C3,$B6,$D7,$FA,$25,$B1,$E7,$3A,$65,$FD,$62,$7F,$A0,$12
	.dc.b	$3F,$81,$63,$E1,$C7,$FB,$3C,$FF,$1B,$24,$FF,$A2,$CF,$F3,$B1,$0C
	.dc.b	$10,$FF,$45,$19,$29,$93,$FD,$14,$64,$A6,$4F,$F4,$51,$92,$99,$E7
	.dc.b	$FF,$52,$39,$6B,$9E,$7F,$A6,$F0,$3C,$24,$3B,$94,$D1,$82,$18,$24
	.dc.b	$76,$78,$35,$DB,$5D,$C7,$8A,$30,$47,$04,$8F,$C4,$C6,$5B,$19,$7D
	.dc.b	$CD,$7D,$C6,$88,$68,$A6,$8C,$10,$C1,$0F,$5B,$D5,$A7,$22,$9C,$4A
	.dc.b	$71,$3E,$0B,$C5,$A7,$12,$9C,$4A,$71,$3E,$39,$B2,$9B,$29,$B8,$F8
	.dc.b	$CE,$5A,$E6,$3C,$4F,$03,$CE,$43,$C2,$53,$46,$08,$60,$91,$D9,$E0
	.dc.b	$D7,$6D,$77,$1E,$28,$C1,$1C,$12,$3E,$04,$65,$B1,$97,$DC,$D7,$DC
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$10,$F5,$BD,$5A,$72,$29,$C4,$A7,$13,$E0
	.dc.b	$BC,$5A,$71,$29,$C4,$A7,$13,$E3,$9B,$29,$B2,$9B,$8F,$8C,$E5,$FD
	.dc.b	$87,$2F,$F8,$22,$3F,$CA,$E1,$C3,$FE,$5A,$6B,$FE,$06,$E6,$FE,$60
	.dc.b	$A5,$E9,$7F,$A0,$79,$B9,$CE,$63,$0F,$69,$DB,$FE,$5A,$31,$1A,$7F
	.dc.b	$CD,$31,$A9,$E6,$D8,$D9,$C4,$DC,$F1,$7F,$D2,$51,$B7,$26,$CC,$C6
	.dc.b	$DF,$31,$92,$CF,$82,$35,$3C,$E7,$89,$B7,$0F,$03,$CB,$3C,$5D,$4F
	.dc.b	$22,$38,$24,$6C,$8F,$22,$1E,$A7,$2D,$32,$79,$1C,$B4,$C9,$E5,$38
	.dc.b	$10,$D1,$4D,$18,$21,$82,$4F,$C0,$24,$FB,$27,$C3,$3A,$8C,$3E,$29
	.dc.b	$DB,$F0,$0C,$46,$9F,$14,$6A,$79,$B6,$36,$71,$37,$3C,$5F,$3C,$6D
	.dc.b	$C9,$B3,$31,$B7,$CC,$64,$B3,$E0,$8D,$4F,$39,$E2,$6D,$C3,$C0,$F2
	.dc.b	$CF,$17,$53,$C8,$8E,$09,$1B,$23,$C8,$87,$A9,$CB,$4C,$9E,$47,$2D
	.dc.b	$32,$79,$4E,$04,$34,$53,$46,$08,$60,$93,$F0,$09,$3E,$C9,$F0,$CE
	.dc.b	$A3,$0F,$8A,$76,$FC,$03,$11,$A7,$C5,$1A,$9E,$6D,$8D,$9C,$4D,$CF
	.dc.b	$17,$CF,$1B,$72,$6C,$CC,$6D,$F3,$19,$2C,$F8,$23,$53,$CE,$78,$9B
	.dc.b	$70,$F0,$3C,$B3,$C5,$D4,$F2,$3C,$1A,$3E,$61,$C4,$E9,$FE,$53,$39
	.dc.b	$8C,$5F,$F3,$68,$C4,$EA,$B8,$BF,$EE,$B2,$4F,$F9,$59,$AF,$FB,$5C
	.dc.b	$87,$FB,$3B,$0F,$F7,$91,$1F,$F2,$EB,$50,$F3,$1E,$63,$C3,$F4,$FC
	.dc.b	$DF,$F4,$A9,$CC,$62,$B8,$9C,$6B,$86,$6B,$E1,$18,$7F,$D7,$5D,$1F
	.dc.b	$D1,$71,$A7,$F3,$DA,$46,$6F,$AC,$62,$7F,$9E,$71,$FF,$BB,$5F,$54
	.dc.b	$FB,$CF,$69,$0F,$87,$3D,$3E,$2B,$FA,$D3,$D2,$D1,$89,$FF,$6A,$B5
	.dc.b	$FB,$31,$A7,$FD,$FF,$7E,$17,$B1,$E1,$3E,$89,$FE,$B0,$E5,$FF,$7F
	.dc.b	$61,$F0,$2F,$A6,$1E,$3A,$4F,$E1,$9F,$09,$88,$F3,$4F,$8E,$3A,$4F
	.dc.b	$C0,$69,$E2,$9F,$93,$F1,$5D,$B1,$34,$93,$D4,$FF,$79,$F8,$5B,$23
	.dc.b	$93,$E0,$BC,$FE,$22,$46,$6F,$EB,$C6,$30,$FF,$6C,$76,$B3,$F8,$37
	.dc.b	$F1,$8F,$39,$E0,$39,$64,$C7,$C3,$7F,$0E,$7C,$27,$D0,$BE,$91,$F0
	.dc.b	$74,$FA,$53,$F4,$1A,$FF,$B8,$B0,$FF,$00,$73,$B9,$DD,$3F,$A0,$F0
	.dc.b	$7F,$C1,$D3,$98,$C6,$3F,$E2,$5A,$DE,$0B,$DC,$61,$B5,$C2,$7F,$C5
	.dc.b	$4D,$7F,$C6,$13,$BA,$E9,$39,$C7,$FD,$D1,$5F,$E2,$C6,$BF,$ED,$23
	.dc.b	$19,$7F,$BE,$E3,$31,$F2,$4F,$8D,$B7,$EA,$34,$FD,$3C,$66,$3F,$3D
	.dc.b	$18,$8F,$88,$74,$D3,$FD,$EB,$7D,$4F,$B1,$1F,$B6,$DB,$85,$FD,$E8
	.dc.b	$FE,$0E,$FE,$A6,$33,$B1,$B8,$4A,$D9,$E3,$21,$FA,$CD,$F8,$91,$D9
	.dc.b	$92,$9F,$AB,$3D,$04,$3E,$31,$FA,$63,$8D,$FE,$B5,$F8,$B7,$FB,$96
	.dc.b	$F0,$7C,$37,$A1,$C8,$FE,$37,$9F,$CE,$7E,$39,$92,$9A,$CB,$E3,$BE
	.dc.b	$13,$A9,$E2,$AF,$05,$E4,$E5,$1E,$03,$96,$7F,$48,$8F,$B4,$C9,$4F
	.dc.b	$E2,$7A,$FF,$9C,$9A,$7F,$9D,$27,$F9,$33,$6B,$FC,$C0,$FF,$35,$E3
	.dc.b	$FE,$A7,$EF,$7F,$34,$6C,$7F,$3A,$E5,$FE,$7B,$6A,$D3,$67,$D3,$9F
	.dc.b	$C5,$7D,$47,$C4,$21,$82,$2C,$FF,$82,$70,$7E,$44,$93,$FE,$96,$9D
	.dc.b	$69,$FB,$08,$FE,$9B,$D1,$FD,$9E,$BF,$B2,$61,$99,$FE,$B3,$39,$7F
	.dc.b	$B5,$74,$7F,$70,$AF,$F3,$B4,$43,$FD,$1A,$53,$FD,$12,$53,$FD,$02
	.dc.b	$53,$FD,$12,$5E,$AF,$F4,$EB,$7A,$BE,$93,$A3,$A9,$0E,$A5,$3A,$94
	.dc.b	$E0,$68,$86,$8A,$68,$C1,$0C,$12,$3C,$AF,$56,$BC,$9A,$F1,$6B,$C6
	.dc.b	$3D,$71,$82,$38,$24,$7E,$46,$32,$D8,$CB,$F1,$4C,$10,$C1,$0F,$33
	.dc.b	$C1,$A6,$CA,$6C,$F5,$3B,$69,$B2,$9B,$3D,$CE,$DA,$6C,$A6,$CF,$84
	.dc.b	$64,$A6,$63,$DE,$D8,$E8,$E8,$F3,$90,$F1,$94,$EE,$53,$81,$A2,$1A
	.dc.b	$29,$A3,$04,$30,$48,$F2,$BD,$5A,$F2,$6B,$C5,$AF,$18,$F5,$C6,$08
	.dc.b	$E0,$91,$F1,$63,$2D,$8C,$BF,$14,$C1,$0C,$10,$F3,$3C,$1A,$6C,$A6
	.dc.b	$CF,$53,$B6,$9B,$29,$B3,$DC,$ED,$A6,$CA,$6C,$F8,$46,$4A,$66,$3D
	.dc.b	$ED,$8E,$8E,$8F,$39,$0F,$B3,$CF,$FE,$0D,$7F,$CC,$E7,$F8,$18,$A7
	.dc.b	$F8,$1A,$BF,$9E,$46,$E3,$74,$F6,$BF,$CE,$FC,$5C,$23,$13,$FE,$4A
	.dc.b	$37,$3A,$69,$93,$FC,$EF,$1C,$27,$94,$EE,$72,$D7,$FD,$0B,$3B,$79
	.dc.b	$CF,$07,$A1,$A9,$E0,$A7,$83,$D0,$D9,$98,$DC,$6A,$39,$1A,$70,$EC
	.dc.b	$C4,$70,$23,$82,$4F,$94,$8F,$A0,$87,$8D,$CB,$4C,$9E,$27,$2D,$32
	.dc.b	$78,$DC,$B4,$CC,$FA,$09,$3E,$69,$DC,$62,$7C,$E7,$09,$D3,$4C,$9E
	.dc.b	$52,$1E,$32,$9E,$12,$C7,$59,$EF,$3D,$27,$2D,$79,$10,$E4,$59,$E4
	.dc.b	$F8,$E7,$9B,$E4,$35,$3C,$DE,$24,$9E,$2F,$80,$E0,$66,$38,$46,$A3
	.dc.b	$99,$A7,$0E,$CC,$47,$02,$38,$24,$F9,$88,$FA,$48,$78,$DC,$B4,$C9
	.dc.b	$E2,$72,$D3,$27,$8D,$CB,$4C,$CF,$A0,$93,$E6,$9D,$C6,$27,$CE,$70
	.dc.b	$9D,$34,$C9,$E5,$21,$E3,$29,$E1,$2C,$75,$9E,$F3,$D2,$72,$D7,$91
	.dc.b	$0E,$45,$9E,$4F,$8E,$79,$BE,$43,$53,$CD,$E2,$49,$E2,$F8,$0E,$06
	.dc.b	$63,$84,$6A,$39,$9A,$70,$EC,$C4,$70,$23,$83,$AE,$BF,$E4,$AE,$BF
	.dc.b	$F4,$F1,$23,$FC,$9D,$BF,$FB,$56,$3F,$98,$EA,$F1,$14,$9C,$C6,$1B
	.dc.b	$19,$3F,$9D,$CF,$17,$07,$FB,$4E,$BF,$DA,$EF,$FB,$86,$FF,$CB,$B9
	.dc.b	$7F,$42,$F9,$E7,$E9,$AF,$77,$F9,$84,$61,$B1,$9B,$FE,$5F,$39,$7F
	.dc.b	$98,$F1,$39,$DF,$13,$EA,$F8,$AF,$E9,$F3,$98,$C5,$7B,$84,$65,$4F
	.dc.b	$0A,$E7,$18,$38,$4F,$7F,$8A,$FE,$8C,$D8,$F1,$CA,$72,$D8,$FE,$83
	.dc.b	$A7,$FB,$D9,$3F,$EF,$8A,$F5,$7C,$47,$02,$4F,$47,$9F,$C1,$FF,$62
	.dc.b	$9F,$1D,$A3,$15,$FE,$D6,$8C,$2D,$BF,$D8,$73,$A8,$FB,$F8,$78,$CF
	.dc.b	$AA,$FE,$1B,$A7,$FA,$EF,$C0,$BF,$7D,$7C,$17,$18,$C3,$63,$35,$DF
	.dc.b	$3F,$9D,$5A,$8C,$1E,$5C,$7C,$53,$F6,$5B,$9F,$A3,$3F,$3D,$70,$3C
	.dc.b	$FC,$7F,$8C,$7F,$1D,$9F,$CC,$8C,$47,$DC,$AE,$58,$7C,$6C,$3A,$92
	.dc.b	$7A,$47,$E2,$7C,$BF,$09,$23,$34,$8C,$63,$FE,$EC,$6D,$7D,$27,$E6
	.dc.b	$57,$DC,$C3,$C6,$DC,$7C,$C4,$BF,$C7,$57,$29,$FB,$0D,$7D,$F3,$B7
	.dc.b	$F5,$55,$D3,$5F,$B0,$D3,$BF,$57,$F8,$BF,$C6,$FF,$10,$37,$1F,$D0
	.dc.b	$36,$FF,$C6,$53,$AF,$19,$DB,$FC,$0A,$30,$DB,$E3,$3A,$A5,$69,$B3
	.dc.b	$FE,$F8,$7F,$DC,$F8,$32,$FF,$7E,$C6,$1A,$EC,$F8,$27,$F1,$2D,$3E
	.dc.b	$63,$4C,$9F,$44,$C1,$FE,$F0,$CF,$FD,$E8,$7F,$BA,$9F,$9A,$6A,$3E
	.dc.b	$F3,$75,$FB,$8D,$D7,$F7,$AF,$69,$F0,$BD,$C8,$FE,$2D,$E5,$B7,$07
	.dc.b	$F2,$0D,$8F,$CE,$47,$98,$FC,$11,$FE,$60,$BF,$3E,$3F,$16,$3F,$88
	.dc.b	$57,$1C,$3F,$73,$1F,$A2,$25,$FE,$B3,$83,$F5,$F1,$FC,$68,$FE,$73
	.dc.b	$C1,$FA,$F7,$01,$C4,$F1,$E1,$F4,$4F,$90,$FA,$87,$97,$1F,$23,$EF
	.dc.b	$3C,$27,$EE,$4F,$82,$7C,$BE,$2F,$F3,$DB,$FB,$16,$D7,$F1,$C8,$FF
	.dc.b	$3F,$B5,$ED,$AF,$E0,$9F,$B5,$5E,$07,$F8,$D7,$1F,$2C,$FF,$95,$47
	.dc.b	$F3,$58,$C1,$9E,$5F,$EB,$3E,$7F,$F3,$96,$CF,$F2,$66,$BF,$C9,$5A
	.dc.b	$B9,$FF,$A6,$3F,$D5,$7A,$3F,$A8,$37,$A7,$DB,$3A,$8F,$8E,$43,$FA
	.dc.b	$67,$57,$F6,$3E,$8F,$EC,$E4,$8F,$EC,$FD,$9F,$DB,$A3,$E0,$92,$3F
	.dc.b	$98,$35,$FF,$44,$1F,$E8,$F2,$1F,$E9,$C2,$9A,$30,$43,$04,$3F,$D1
	.dc.b	$A7,$02,$9B,$29,$B3,$FD,$20,$6C,$A6,$CA,$6F,$BB,$FD,$66,$ED,$AE
	.dc.b	$DA,$EF,$BB,$F0,$EF,$23,$C2,$43,$04,$8E,$2E,$5A,$E6,$39,$C6,$08
	.dc.b	$E0,$91,$F9,$68,$CB,$63,$2F,$A5,$AF,$A5,$AF,$A4,$E0,$43,$81,$4E
	.dc.b	$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3E,$13,$E4,$69,$DC,$A7,$42
	.dc.b	$9C,$CA,$73,$3E,$51,$C4,$A7,$12,$9C,$4A,$71,$8F,$96,$ED,$AE,$DA
	.dc.b	$EE,$3D,$0F,$23,$C6,$43,$04,$8E,$2E,$5A,$E6,$39,$C6,$08,$E0,$91
	.dc.b	$ED,$8C,$B6,$32,$FA,$5A,$FA,$5A,$FA,$4E,$04,$38,$14,$E0,$53,$81
	.dc.b	$A2,$1A,$29,$A3,$04,$30,$43,$E1,$3E,$46,$9D,$CA,$74,$29,$CC,$A7
	.dc.b	$33,$E5,$1C,$4F,$A3,$D8,$EC,$FF,$00,$47,$F9,$A4,$46,$6F,$FA,$0B
	.dc.b	$63,$46,$A7,$8B,$FE,$71,$BF,$F3,$67,$5E,$9B,$1D,$5A,$EF,$B7,$FD
	.dc.b	$2B,$3B,$6C,$6D,$AE,$CF,$F3,$01,$92,$99,$8E,$4E,$0F,$F5,$0C,$F8
	.dc.b	$1B,$3D,$1A,$6C,$46,$DB,$1B,$69,$B8,$D1,$23,$4D,$8D,$46,$08,$E0
	.dc.b	$93,$E8,$23,$ED,$6B,$ED,$30,$43,$04,$3D,$6F,$06,$9B,$29,$B3,$D6
	.dc.b	$ED,$A6,$CA,$6E,$7D,$91,$B6,$C6,$DA,$EE,$7D,$F3,$B6,$C6,$DA,$EC
	.dc.b	$F8,$66,$4A,$66,$39,$38,$3D,$64,$3C,$E5,$9F,$1B,$67,$BB,$4E,$47
	.dc.b	$02,$47,$06,$C7,$06,$9C,$23,$44,$8D,$36,$35,$18,$23,$82,$4F,$B0
	.dc.b	$8F,$C1,$6B,$F0,$4C,$10,$C1,$0F,$63,$C1,$A6,$CA,$6C,$F5,$BB,$69
	.dc.b	$B2,$9B,$9F,$64,$6D,$B1,$B6,$BB,$9F,$7C,$ED,$B1,$B6,$BB,$3E,$19
	.dc.b	$92,$99,$8E,$4E,$0F,$59,$0F,$39,$67,$C6,$D9,$EE,$D3,$91,$C0,$91
	.dc.b	$C1,$B1,$C1,$A7,$08,$FC,$37,$CF,$3D,$E7,$3B,$C0,$FE,$4F,$39,$8C
	.dc.b	$5F,$F2,$F8,$C4,$EA,$B8,$BF,$CB,$BE,$89,$CB,$8B,$FE,$6D,$18,$C3
	.dc.b	$9C,$09,$19,$AE,$55,$C6,$74,$25,$F2,$59,$8E,$2D,$74,$D8,$DD,$F8
	.dc.b	$27,$FD,$B7,$7F,$E8,$02,$3F,$D2,$FC,$07,$57,$FB,$1E,$7F,$D0,$BF
	.dc.b	$05,$2B,$53,$FE,$CB,$8C,$47,$AF,$99,$E2,$3C,$BF,$75,$7F,$57,$9C
	.dc.b	$C6,$2B,$DB,$18,$53,$BA,$E5,$18,$3F,$DE,$3D,$DF,$EF,$37,$E2,$E5
	.dc.b	$E8,$8F,$E9,$5E,$43,$CE,$7A,$5F,$1E,$30,$D8,$CE,$DF,$16,$72,$D8
	.dc.b	$72,$FE,$1B,$E8,$FF,$B8,$57,$58,$C6,$5F,$5A,$31,$6A,$DD,$F2,$8C
	.dc.b	$56,$E1,$F3,$9F,$92,$E8,$DB,$67,$4D,$7F,$11,$87,$B6,$31,$1F,$8B
	.dc.b	$D3,$E1,$57,$FA,$F0,$91,$F6,$34,$FE,$DD,$8B,$E6,$1E,$67,$F6,$D8
	.dc.b	$C3,$63,$35,$C1,$B7,$F0,$BF,$6D,$FC,$D5,$BE,$27,$FD,$BF,$A2,$73
	.dc.b	$18,$6C,$FC,$76,$E1,$F6,$B9,$59,$FD,$D8,$F6,$46,$31,$F5,$C7,$FB
	.dc.b	$BB,$F2,$77,$C5,$7E,$CB,$1F,$D3,$F9,$89,$CC,$7E,$F4,$CB,$8E,$9F
	.dc.b	$F7,$AC,$EE,$B9,$5F,$CE,$CF,$DB,$EF,$7A,$63,$F8,$3B,$71,$34,$FF
	.dc.b	$7F,$C3,$87,$E9,$BF,$98,$CF,$D2,$B7,$F8,$DE,$EF,$4C,$3E,$23,$67
	.dc.b	$C4,$D7,$86,$28,$C5,$7C,$FF,$6F,$F1,$57,$F3,$5B,$7E,$B6,$BC,$9B
	.dc.b	$3F,$5F,$CF,$B8,$1A,$4B,$63,$EE,$CB,$DB,$8F,$D9,$E1,$FB,$0F,$C1
	.dc.b	$E1,$FB,$0F,$8E,$F8,$96,$FF,$09,$9F,$80,$F2,$CF,$1B,$E7,$87,$68
	.dc.b	$FC,$1C,$63,$0E,$B1,$E1,$BF,$9A,$FF,$97,$FC,$6F,$E5,$DE,$9F,$F9
	.dc.b	$43,$C7,$FF,$32,$37,$C7,$4E,$AF,$F9,$7D,$69,$B3,$AB,$5F,$1B,$E1
	.dc.b	$1F,$CE,$EF,$83,$6B,$85,$7F,$98,$32,$FE,$2D,$3F,$E7,$46,$BF,$B5
	.dc.b	$31,$8F,$E8,$9B,$97,$F9,$E9,$F8,$86,$08,$7C,$D3,$E5,$E9,$FB,$47
	.dc.b	$F0,$0D,$7F,$00,$D7,$99,$E1,$21,$F6,$CF,$DF,$6B,$FD,$99,$FD,$D3
	.dc.b	$63,$F7,$4D,$8F,$D4,$35,$EE,$79,$09,$1E,$73,$F3,$E7,$F7,$22,$9F
	.dc.b	$C1,$F8,$30,$E0,$F2,$C7,$E0,$8F,$82,$79,$CF,$E1,$87,$03,$F7,$45
	.dc.b	$3F,$52,$5E,$0F,$E2,$CE,$78,$FF,$81,$4F,$E1,$0F,$E0,$FC,$7F,$E0
	.dc.b	$6D,$FF,$83,$F2,$FF,$0C,$6C,$66,$7C,$B7,$F3,$0F,$21,$0F,$0B,$FC
	.dc.b	$78,$F3,$95,$F3,$B5,$DE,$7F,$45,$FF,$0B,$F3,$FF,$34,$7F,$D8,$53
	.dc.b	$FC,$E1,$FF,$5F,$B5,$CC,$6D,$F2,$ED,$F9,$07,$C8,$79,$CF,$E2,$9D
	.dc.b	$5F,$D5,$2B,$FD,$93,$A7,$F4,$F6,$C6,$67,$FA,$BC,$63,$B3,$FC,$57
	.dc.b	$F4,$DF,$D8,$1B,$3F,$D7,$DA,$FF,$5A,$AF,$EC,$93,$FD,$9B,$BB,$E7
	.dc.b	$4F,$F6,$56,$BB,$6B,$E6,$6B,$E9,$26,$3E,$D7,$F0,$C7,$F8,$3B,$BF
	.dc.b	$FC,$17,$F0,$5F,$E0,$DF,$0F,$FC,$64,$D9,$34,$FD,$89,$1F,$ED,$4D
	.dc.b	$7F,$90,$70,$7E,$D8,$8F,$F6,$F7,$FC,$89,$E2,$F9,$BC,$5F,$F1,$9C
	.dc.b	$63,$6F,$F7,$9D,$7F,$8E,$1B,$3F,$E4,$79,$FF,$21,$C6,$3C,$6E,$47
	.dc.b	$C0,$F2,$3F,$CA,$D3,$FE,$53,$9E,$71,$8B,$FF,$2C,$79,$1F,$1B,$C9
	.dc.b	$FF,$0D,$C7,$F0,$B6,$BF,$EF,$22,$1F,$EF,$72,$9F,$EF,$B2,$9F,$EF
	.dc.b	$22,$9F,$EE,$E2,$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C
	.dc.b	$0A,$70,$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$27,$17,$F3,$18,$CB,$63
	.dc.b	$31,$F2,$E3,$04,$70,$43,$E2,$B9,$69,$93,$E0,$39,$69,$93,$DA,$E5
	.dc.b	$A6,$4F,$63,$96,$99,$3D,$25,$8F,$4B,$E2,$21,$E0,$29,$D8,$A7,$42
	.dc.b	$9D,$0A,$F4,$79,$10,$E4,$53,$91,$4E,$45,$79,$3C,$08,$70,$29,$C0
	.dc.b	$AF,$07,$44,$34,$57,$4E,$08,$E0,$91,$F2,$E3,$2D,$8C,$C7,$CB,$8C
	.dc.b	$11,$C1,$0F,$8A,$E5,$A6,$4F,$80,$E5,$A6,$4F,$6B,$96,$99,$3D,$8E
	.dc.b	$5A,$64,$F4,$96,$3D,$2F,$88,$87,$80,$A7,$62,$9D,$0A,$74,$3F,$81
	.dc.b	$F7,$3B,$BF,$8E,$C7,$F8,$CC,$FF,$6D,$9F,$E2,$5C,$BF,$C3,$93,$B7
	.dc.b	$F8,$A7,$8D,$FE,$E5,$F0,$37,$4C,$3F,$D4,$EF,$F9,$48,$E6,$E2,$36
	.dc.b	$D8,$CC,$F3,$69,$C4,$FF,$76,$C6,$DB,$1B,$69,$B3,$FD,$F1,$1B,$6C
	.dc.b	$6D,$A6,$CF,$43,$96,$99,$8E,$4E,$27,$D3,$3B,$9C,$91,$C1,$27,$93
	.dc.b	$67,$8C,$60,$8E,$08,$FC,$26,$BE,$E6,$BE,$A6,$BE,$53,$81,$0E,$05
	.dc.b	$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$7E,$13,$EC,$3B,$38,$8F
	.dc.b	$3B,$63,$33,$F1,$5A,$71,$3C,$31,$B6,$C6,$DA,$6C,$F2,$46,$DB,$1B
	.dc.b	$69,$B3,$D0,$E5,$A6,$63,$93,$89,$F4,$CE,$E7,$24,$70,$49,$E4,$D9
	.dc.b	$E3,$18,$23,$82,$3F,$09,$AF,$B9,$AF,$A9,$AF,$94,$E0,$43,$81,$4E
	.dc.b	$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$9F,$84,$FB,$0E,$CE,$23,$CE
	.dc.b	$D8,$CC,$FC,$56,$9C,$4F,$0C,$6D,$B1,$B6,$9B,$3C,$91,$B6,$C6,$DA
	.dc.b	$6C,$F4,$39,$69,$98,$E4,$E2,$7D,$33,$B9,$FE,$0B,$B3,$66,$CF,$13
	.dc.b	$FC,$8A,$4C,$7F,$96,$D6,$E7,$51,$8F,$3F,$8D,$7F,$2C,$A4,$62,$75
	.dc.b	$5C,$11,$27,$71,$98,$FF,$79,$D7,$FB,$58,$8F,$F3,$EE,$57,$82,$EC
	.dc.b	$E1,$2C,$1E,$87,$6C,$3C,$8D,$9F,$5B,$5F,$19,$C5,$C3,$E5,$7C,$FE
	.dc.b	$57,$28,$34,$8F,$2C,$9D,$DF,$D5,$E3,$66,$5C,$4F,$FA,$AB,$8F,$AC
	.dc.b	$E5,$AF,$93,$4F,$7E,$7F,$D0,$5B,$C3,$FE,$FA,$24,$FA,$BB,$9E,$33
	.dc.b	$E4,$3E,$EF,$C6,$7F,$66,$9C,$C6,$2B,$85,$6F,$55,$EE,$30,$DF,$9E
	.dc.b	$FC,$3F,$D4,$5A,$34,$97,$D3,$F1,$AF,$C1,$3F,$80,$CB,$CB,$18,$84
	.dc.b	$62,$BF,$04,$DE,$2F,$6C,$FC,$3E,$4F,$C0,$6B,$88,$FC,$5E,$DF,$11
	.dc.b	$F9,$DE,$8F,$F6,$9C,$7E,$9B,$78,$CD,$3F,$B7,$65,$F2,$3E,$57,$E3
	.dc.b	$D7,$C4,$6D,$FA,$78,$11,$88,$F4,$C6,$27,$E4,$9F,$97,$CF,$CF,$37
	.dc.b	$8B,$ED,$46,$23,$D9,$1F,$E0,$3C,$3E,$CF,$C9,$F5,$6D,$F4,$AF,$14
	.dc.b	$E7,$03,$71,$FD,$DF,$A1,$FE,$FD,$C3,$D9,$82,$31,$7E,$76,$D7,$09
	.dc.b	$F3,$6D,$F3,$71,$6D,$F3,$5F,$3C,$7E,$17,$E5,$3F,$84,$FC,$07,$46
	.dc.b	$DF,$E3,$1B,$5F,$25,$AF,$C8,$FB,$1F,$E2,$37,$E4,$6D,$7E,$51,$B3
	.dc.b	$F4,$5A,$FD,$0C,$AE,$9F,$A0,$6E,$9E,$FB,$61,$F0,$27,$8D,$FB,$37
	.dc.b	$FA,$4F,$E2,$9A,$E7,$7F,$AE,$7D,$5C,$19,$76,$26,$3D,$8F,$41,$6B
	.dc.b	$E3,$9F,$30,$86,$08,$78,$6B,$BE,$3F,$47,$C1,$FF,$2D,$4B,$D1,$FF
	.dc.b	$27,$46,$1B,$D4,$F0,$BF,$96,$C6,$1B,$7F,$C7,$AB,$83,$6B,$85,$F5
	.dc.b	$8C,$57,$08,$FF,$36,$19,$3C,$C7,$F9,$AB,$87,$F8,$AB,$75,$FE,$68
	.dc.b	$DC,$BF,$CD,$EF,$CB,$30,$43,$F4,$87,$F0,$FD,$3F,$A9,$BF,$83,$D3
	.dc.b	$FA,$39,$F3,$1C,$1A,$3C,$67,$22,$1F,$AC,$3F,$8D,$6D,$D2,$3F,$8F
	.dc.b	$B6,$3F,$78,$D7,$F1,$91,$E0,$7F,$20,$D7,$F3,$0D,$DF,$F9,$7D,$71
	.dc.b	$6C,$67,$77,$07,$F6,$EF,$43,$F4,$D2,$69,$1F,$4E,$1E,$C3,$D4,$62
	.dc.b	$FF,$77,$97,$EE,$27,$FB,$11,$FC,$5C,$B3,$FC,$5F,$8F,$FB,$EC,$FF
	.dc.b	$00,$E4,$FE,$FF,$F1,$D6,$CD,$23,$E2,$BF,$B0,$5D,$FE,$FB,$DE,$7C
	.dc.b	$44,$9C,$1F,$CC,$A3,$FD,$5E,$FF,$32,$6F,$3F,$F3,$26,$F3,$FE,$19
	.dc.b	$EE,$7C,$02,$1F,$0C,$FE,$5F,$D3,$FE,$18,$E3,$FE,$AC,$D8,$FE,$A1
	.dc.b	$7F,$9C,$7F,$9C,$46,$5F,$E7,$5D,$7F,$E2,$0F,$AC,$FE,$B4,$48,$FE
	.dc.b	$C7,$19,$7F,$B2,$F7,$7C,$86,$DF,$C8,$6E,$7F,$D9,$E7,$51,$E8,$3D
	.dc.b	$67,$A4,$B3,$FE,$06,$E8,$FF,$11,$F9,$5F,$E0,$39,$34,$8F,$F8,$60
	.dc.b	$C9,$5F,$F0,$A1,$1D,$67,$FE,$17,$23,$83,$FA,$93,$FB,$52,$38,$3F
	.dc.b	$C6,$EF,$F8,$EF,$C3,$F9,$11,$FE,$37,$6B,$FE,$2F,$6F,$87,$FE,$45
	.dc.b	$9F,$F2,$17,$89,$82,$63,$F4,$EF,$FD,$B6,$7F,$B3,$08,$7F,$BA,$8A
	.dc.b	$7F,$B9,$8A,$70,$34,$43,$45,$34,$60,$86,$08,$7F,$BA,$0E,$A5,$39
	.dc.b	$14,$E2,$53,$8F,$8A,$F1,$BF,$30,$D8,$FE,$3E,$D7,$AB,$5E,$AD,$7A
	.dc.b	$B5,$EB,$1D,$08,$F4,$6B,$D1,$AF,$46,$BD,$1A,$F4,$39,$10,$E4,$53
	.dc.b	$91,$4E,$45,$39,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08
	.dc.b	$60,$87,$D3,$24,$7C,$96,$C7,$C0,$69,$E9,$29,$E4,$29,$E1,$29,$E0
	.dc.b	$29,$E0,$8E,$C4,$8E,$CD,$8E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$D1,$D0
	.dc.b	$8F,$46,$BD,$1A,$F4,$6B,$D1,$AF,$43,$91,$0E,$45,$39,$14,$E4,$53
	.dc.b	$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$7D,$43
	.dc.b	$F0,$5D,$FF,$E2,$67,$FD,$76,$1C,$3F,$E1,$C3,$26,$74,$FE,$28,$FF
	.dc.b	$B5,$3F,$13,$7A,$BF,$ED,$6F,$07,$71,$89,$FF,$25,$9D,$9C,$46,$A7
	.dc.b	$84,$E5,$AF,$FB,$3A,$76,$E4,$E6,$66,$36,$F8,$DA,$A3,$8B,$B3,$89
	.dc.b	$B2,$9B,$8E,$31,$C2,$35,$3E,$77,$71,$97,$81,$89,$E2,$47,$04,$9E
	.dc.b	$44,$7D,$8D,$7C,$CD,$7C,$26,$88,$68,$A6,$8C,$10,$C1,$0F,$3B,$D5
	.dc.b	$A7,$22,$9C,$4A,$71,$9F,$80,$49,$F7,$4F,$38,$C4,$FB,$CE,$8E,$23
	.dc.b	$53,$C2,$72,$D7,$C1,$3B,$72,$73,$33,$1B,$7C,$8D,$7C,$E4,$8E,$4F
	.dc.b	$03,$91,$C0,$A6,$E3,$94,$70,$8D,$4F,$A1,$DC,$65,$E0,$62,$78,$91
	.dc.b	$C1,$27,$91,$1F,$6B,$5F,$33,$5F,$09,$A2,$1A,$29,$A3,$04,$30,$43
	.dc.b	$CE,$F5,$69,$C8,$A7,$12,$9C,$67,$E0,$12,$7D,$D3,$CE,$31,$3E,$F3
	.dc.b	$A3,$88,$D4,$F0,$9C,$B5,$F0,$4E,$DC,$9C,$CC,$C6,$DF,$23,$5F,$39
	.dc.b	$23,$93,$C0,$E4,$70,$29,$B8,$E5,$1C,$23,$53,$E8,$77,$19,$7F,$1F
	.dc.b	$AB,$CC,$68,$F1,$7F,$94,$CE,$63,$15,$FC,$C2,$31,$3A,$AE,$2F,$FB
	.dc.b	$0C,$93,$FC,$DB,$A5,$E1,$BB,$39,$C8,$FF,$72,$57,$83,$E8,$F0,$D8
	.dc.b	$D9,$BC,$53,$A8,$F2,$35,$E8,$21,$F4,$19,$8E,$4D,$74,$D8,$DC,$78
	.dc.b	$0E,$59,$7C,$6D,$7E,$06,$9F,$CC,$AB,$E3,$46,$0F,$F7,$21,$2B,$E1
	.dc.b	$67,$F0,$63,$11,$E6,$D3,$E2,$E9,$FC,$F1,$B3,$A9,$F8,$D1,$87,$9C
	.dc.b	$EE,$BA,$46,$2B,$87,$80,$F0,$DD,$FF,$7B,$C5,$FE,$D3,$39,$8C,$5F
	.dc.b	$C8,$8C,$68,$AD,$DF,$28,$C5,$7C,$E5,$F6,$5F,$ED,$F8,$79,$7F,$DD
	.dc.b	$A7,$3E,$2F,$DF,$8C,$56,$DB,$19,$AF,$51,$0F,$8D,$A7,$B9,$BA,$FC
	.dc.b	$12,$6B,$F1,$E1,$B7,$C7,$7F,$1D,$EB,$BB,$DF,$27,$FA,$0F,$AF,$FD
	.dc.b	$0D,$6A,$73,$B2,$73,$2B,$E5,$7C,$6B,$F1,$F1,$87,$81,$E9,$9F,$B3
	.dc.b	$9F,$A5,$B9,$FA,$1D,$1E,$88,$D3,$63,$4F,$96,$FF,$6F,$E6,$79,$30
	.dc.b	$F0,$37,$2F,$4B,$67,$78,$F6,$CF,$E2,$61,$E2,$D3,$E4,$7C,$EB,$D2
	.dc.b	$FD,$A7,$1F,$C0,$A4,$E5,$FC,$26,$1F,$E3,$28,$C4,$ED,$B1,$F8,$5E
	.dc.b	$2F,$A1,$4C,$3E,$5C,$FE,$23,$4F,$F1,$93,$B6,$CF,$F0,$58,$D3,$C5
	.dc.b	$FC,$B7,$1B,$E0,$BF,$69,$18,$6C,$FF,$0D,$E5,$FD,$14,$62,$11,$89
	.dc.b	$E9,$1F,$8C,$C3,$F1,$5A,$FC,$68,$FA,$51,$F9,$8C,$7C,$B1,$FC,$A3
	.dc.b	$2F,$6C,$FB,$1B,$7E,$B9,$EA,$7C,$F3,$E9,$71,$7D,$EC,$3B,$11,$C1
	.dc.b	$38,$BF,$C7,$FD,$EE,$F7,$87,$FC,$A7,$CF,$FF,$26,$CE,$63,$1A,$7F
	.dc.b	$98,$E7,$35,$C3,$24,$62,$B7,$1F,$CC,$EF,$93,$AB,$E4,$47,$F3,$27
	.dc.b	$F9,$BB,$2F,$E3,$2D,$CB,$F9,$F3,$72,$FF,$39,$BF,$9B,$30,$43,$F8
	.dc.b	$D9,$F3,$4F,$C0,$1F,$81,$D3,$F1,$2F,$CD,$D7,$F8,$83,$96,$BE,$43
	.dc.b	$99,$0F,$E0,$E7,$F1,$AD,$FF,$7F,$B7,$E1,$CE,$F1,$F8,$4E,$0F,$C6
	.dc.b	$46,$38,$3F,$0D,$3E,$13,$71,$F9,$47,$04,$7F,$49,$C7,$FD,$77,$2F
	.dc.b	$D2,$36,$33,$3F,$D8,$DB,$3F,$AF,$D3,$E5,$CF,$A4,$F6,$1E,$A3,$F8
	.dc.b	$7E,$5F,$C5,$BA,$3F,$7D,$66,$91,$E5,$3F,$44,$B5,$D9,$E8,$43,$B9
	.dc.b	$39,$FF,$8E,$7D,$8F,$62,$57,$8B,$9F,$FC,$13,$1E,$9E,$0F,$C0,$DF
	.dc.b	$66,$BD,$4F,$11,$70,$F1,$3E,$0B,$FE,$8E,$BC,$7F,$E8,$AD,$3F,$86
	.dc.b	$65,$FE,$09,$8C,$9F,$D5,$BB,$3F,$C3,$7A,$DB,$F0,$EF,$FD,$72,$30
	.dc.b	$7E,$47,$0F,$16,$5E,$67,$E0,$9E,$62,$E5,$E6,$70,$47,$04,$8E,$51
	.dc.b	$FE,$08,$BF,$D0,$F6,$FF,$8A,$E3,$FC,$12,$43,$FC,$59,$DF,$FE,$23
	.dc.b	$E7,$FF,$18,$B6,$32,$7F,$B6,$A7,$FC,$2B,$3F,$E2,$52,$47,$F8,$88
	.dc.b	$D7,$81,$C5,$FE,$CF,$1F,$B3,$23,$FE,$0A,$7F,$C3,$64,$7F,$C7,$DE
	.dc.b	$07,$F9,$12,$31,$1F,$ED,$33,$FD,$98,$43,$FD,$BC,$53,$46,$08,$60
	.dc.b	$87,$FB,$6C,$E0,$53,$65,$36,$7F,$B6,$0D,$94,$D9,$4D,$F8,$5F,$EF
	.dc.b	$B7,$6D,$76,$D7,$7E,$1F,$F1,$67,$91,$E1,$21,$82,$47,$17,$2D,$73
	.dc.b	$1C,$E3,$04,$70,$48,$FE,$4B,$19,$6C,$65,$F4,$B5,$F4,$B5,$F4,$9C
	.dc.b	$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$87,$C2,$7C,$8D
	.dc.b	$3B,$94,$E8,$53,$99,$4E,$67,$CA,$38,$94,$E2,$53,$89,$4E,$31,$F2
	.dc.b	$DD,$B5,$DB,$5D,$C7,$A1,$E4,$78,$C8,$60,$91,$C5,$CB,$5C,$C7,$38
	.dc.b	$C1,$1C,$12,$3D,$B1,$96,$C6,$5F,$4B,$5F,$4B,$5F,$49,$C0,$87,$02
	.dc.b	$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$7C,$27,$C8,$D3,$B9,$4E
	.dc.b	$85,$3F,$D6,$E3,$AF,$FC,$22,$E1,$AF,$FA,$F1,$AF,$F8,$5F,$B7,$FC
	.dc.b	$38,$F2,$A7,$BB,$FE,$CD,$ED,$EB,$29,$CB,$FE,$53,$38,$4E,$DA,$64
	.dc.b	$FF,$65,$C7,$09,$E6,$D8,$DC,$62,$79,$47,$06,$BF,$ED,$F9,$E0,$E4
	.dc.b	$EC,$66,$36,$F9,$DB,$1C,$23,$47,$03,$45,$74,$E2,$7C,$8E,$67,$99
	.dc.b	$1C,$12,$36,$47,$DA,$43,$D0,$E5,$A6,$4F,$2B,$96,$99,$3C,$85,$9F
	.dc.b	$39,$27,$C9,$3C,$67,$67,$62,$1D,$0A,$68,$C1,$0C,$12,$7B,$3E,$F3
	.dc.b	$C5,$3E,$06,$99,$3A,$C7,$09,$F2,$36,$37,$18,$9E,$51,$C1,$AF,$8E
	.dc.b	$78,$39,$3B,$19,$8D,$BE,$76,$C7,$08,$D1,$C0,$D1,$5D,$38,$9F,$23
	.dc.b	$99,$E6,$47,$04,$8D,$91,$F6,$90,$F4,$39,$69,$93,$CA,$E5,$A6,$4F
	.dc.b	$21,$67,$CE,$49,$F2,$4F,$19,$D9,$D8,$87,$42,$9A,$30,$43,$04,$9E
	.dc.b	$CF,$BC,$F1,$4F,$81,$A6,$4E,$B1,$C2,$7C,$8D,$8D,$C6,$27,$94,$70
	.dc.b	$6B,$E3,$9E,$0E,$4E,$C6,$63,$6F,$9D,$B1,$C2,$34,$70,$34,$57,$4F
	.dc.b	$E8,$38,$1C,$0E,$17,$89,$FE,$B5,$25,$7F,$32,$AD,$CE,$A3,$F9,$AF
	.dc.b	$03,$D4,$46,$4A,$23,$4E,$4D,$B8,$BF,$E6,$6D,$8F,$F2,$EC,$7F,$98
	.dc.b	$1E,$07,$3E,$2E,$F1,$93,$F9,$DF,$8E,$F3,$B3,$5F,$D2,$A3,$0F,$26
	.dc.b	$CF,$09,$DC,$FF,$BF,$CF,$F7,$9E,$3F,$CA,$F2,$FF,$80,$18,$FA,$B6
	.dc.b	$F3,$E7,$FC,$D2,$31,$7F,$12,$3F,$9B,$D7,$FB,$E6,$F9,$69,$FE,$F9
	.dc.b	$C3,$FA,$8F,$C1,$BD,$AF,$0E,$3F,$2A,$30,$FC,$06,$E9,$C2,$77,$D0
	.dc.b	$F0,$5F,$4A,$EA,$7D,$97,$F6,$AA,$D4,$E6,$31,$5C,$2B,$79,$7D,$8B
	.dc.b	$DC,$62,$3F,$DE,$1E,$A7,$F5,$88,$78,$C8,$C3,$67,$FA,$CC,$EA,$7E
	.dc.b	$1E,$DF,$06,$7E,$6E,$DE,$B7,$FA,$C6,$DF,$02,$30,$48,$FC,$5D,$BE
	.dc.b	$A1,$E2,$FB,$FC,$AF,$81,$5F,$92,$AD,$4E,$5E,$35,$C1,$57,$0A,$F9
	.dc.b	$13,$98,$FF,$6C,$71,$7D,$82,$61,$FE,$08,$E3,$FD,$1B,$7E,$53,$F2
	.dc.b	$78,$FE,$9D,$B8,$FE,$8A,$91,$88,$FD,$3C,$61,$FD,$8C,$F9,$B2,$FB
	.dc.b	$5F,$07,$E5,$8F,$58,$71,$9F,$AD,$B3,$C2,$7E,$6D,$43,$E2,$6E,$F6
	.dc.b	$3F,$6F,$7D,$63,$13,$D6,$BE,$7C,$62,$75,$5C,$27,$3E,$CF,$F8,$AD
	.dc.b	$B6,$D3,$A4,$E6,$3C,$2D,$7F,$2D,$F6,$3E,$3A,$E0,$7D,$E7,$15,$F0
	.dc.b	$5B,$19,$BF,$CF,$4F,$D6,$3E,$A7,$0F,$E3,$6F,$F7,$47,$80,$EA,$58
	.dc.b	$F2,$B8,$3F,$51,$93,$1F,$34,$35,$FD,$D5,$76,$DF,$F0,$73,$F3,$4F
	.dc.b	$B7,$87,$85,$FF,$29,$78,$7F,$C9,$3C,$EF,$F2,$47,$86,$F0,$D3,$9C
	.dc.b	$7F,$CB,$51,$8B,$5F,$08,$C3,$6B,$84,$F6,$37,$5C,$A3,$31,$FE,$69
	.dc.b	$AE,$33,$F9,$3C,$3F,$A0,$4F,$AC,$87,$F3,$F7,$2D,$C7,$FC,$CE,$FC
	.dc.b	$38,$CB,$CE,$33,$97,$F4,$68,$FF,$3C,$ED,$FC,$1B,$4F,$DC,$C7,$D3
	.dc.b	$3E,$89,$23,$B3,$E2,$3A,$14,$FC,$3E,$DF,$BE,$8F,$C1,$1F,$BF,$8F
	.dc.b	$E0,$1B,$FE,$2D,$FC,$44,$ED,$F0,$4F,$E5,$9F,$01,$83,$F4,$B1,$FC
	.dc.b	$26,$3F,$42,$D8,$C9,$FC,$B6,$32,$D3,$CE,$7A,$88,$60,$FE,$29,$C5
	.dc.b	$FC,$46,$3F,$6E,$71,$8F,$E2,$3C,$5D,$63,$F8,$11,$98,$FC,$97,$1F
	.dc.b	$F6,$ED,$FF,$8C,$46,$0F,$C5,$CF,$F1,$16,$9F,$C4,$67,$84,$79,$23
	.dc.b	$F2,$FA,$7D,$8D,$7E,$DC,$60,$FB,$38,$FF,$27,$9F,$39,$2B,$CA,$F8
	.dc.b	$B9,$7F,$AB,$D7,$F3,$B2,$1F,$EB,$A7,$2F,$84,$FC,$54,$78,$48,$FE
	.dc.b	$23,$8F,$FA,$53,$79,$BF,$23,$7F,$D2,$E3,$FD,$72,$D7,$FA,$DF,$57
	.dc.b	$F6,$1F,$B3,$B2,$69,$27,$C9,$D9,$F2,$DA,$7F,$B2,$0F,$EE,$4E,$5F
	.dc.b	$F0,$DE,$5F,$DD,$36,$FC,$A4,$7A,$8B,$C7,$FA,$17,$FC,$04,$7E,$A3
	.dc.b	$BF,$FC,$42,$FF,$82,$0C,$77,$FE,$B1,$E4,$7C,$59,$E4,$FE,$24,$FF
	.dc.b	$16,$F8,$3F,$E1,$AB,$FF,$0C,$47,$9C,$46,$5F,$F1,$07,$83,$E9,$7F
	.dc.b	$C5,$FE,$15,$7F,$C5,$47,$1C,$51,$96,$C6,$71,$C1,$1C,$1E,$1C,$7F
	.dc.b	$C8,$B8,$FF,$25,$24,$7F,$09,$7F,$D9,$EC,$3F,$DD,$65,$3F,$DC,$05
	.dc.b	$3F,$DB,$65,$3F,$DB,$65,$F0,$FF,$DE,$ED,$F1,$3E,$54,$62,$3A,$90
	.dc.b	$EA,$53,$A9,$4E,$A5,$79,$3C,$08,$70,$29,$C0,$AF,$07,$44,$34,$57
	.dc.b	$4E,$08,$E0,$91,$EC,$24,$7F,$32,$8C,$B6,$32,$FC,$33,$04,$30,$43
	.dc.b	$DC,$F0,$69,$B2,$9B,$3D,$AE,$DA,$6C,$A6,$CF,$73,$B6,$9B,$29,$B3
	.dc.b	$E0,$BB,$69,$B2,$9B,$8F,$82,$D8,$E1,$18,$8F,$49,$0F,$29,$4F,$09
	.dc.b	$4E,$A5,$79,$3C,$08,$70,$29,$C0,$AF,$07,$44,$34,$57,$4E,$08,$E0
	.dc.b	$91,$EC,$24,$7C,$38,$CB,$63,$2F,$C3,$30,$43,$04,$3D,$CF,$06,$9B
	.dc.b	$29,$B3,$DA,$ED,$A6,$CA,$6C,$F7,$3B,$69,$B2,$9B,$3E,$0B,$B6,$9B
	.dc.b	$29,$B8,$F9,$3D,$DF,$E2,$BA,$FF,$0A,$B6,$7F,$DA,$27,$F8,$94,$AE
	.dc.b	$5C,$FB,$2F,$6F,$FD,$B3,$DD,$C2,$53,$97,$FC,$B8,$70,$9D,$B4,$C9
	.dc.b	$FE,$E0,$8E,$13,$CD,$B1,$B8,$C4,$F2,$8E,$0D,$7F,$DD,$73,$C1,$C9
	.dc.b	$D8,$CC,$6D,$F3,$B6,$38,$46,$8E,$06,$8A,$E9,$C4,$F9,$1C,$CF,$32
	.dc.b	$38,$24,$6C,$8F,$B4,$87,$A1,$CB,$4C,$9E,$57,$2D,$32,$79,$0B,$3E
	.dc.b	$72,$4F,$92,$78,$CE,$CE,$C4,$3A,$14,$D1,$82,$18,$24,$F6,$7D,$E7
	.dc.b	$8A,$7C,$0D,$32,$75,$8E,$13,$E4,$6C,$6E,$31,$3C,$A3,$83,$5F,$1C
	.dc.b	$F0,$72,$76,$33,$1B,$7C,$ED,$8E,$11,$A3,$81,$A2,$BA,$71,$3E,$47
	.dc.b	$33,$CC,$8E,$09,$1B,$23,$ED,$21,$E8,$72,$D3,$27,$95,$CB,$4C,$9E
	.dc.b	$42,$CF,$9C,$93,$E4,$9E,$33,$B3,$B1,$0E,$85,$34,$60,$86,$09,$3D
	.dc.b	$9F,$79,$E2,$9F,$03,$4C,$9D,$63,$84,$F9,$1B,$1B,$8C,$4F,$28,$E0
	.dc.b	$D7,$C7,$3C,$1C,$9D,$8C,$C6,$DF,$3B,$63,$84,$68,$E0,$7E,$77,$E9
	.dc.b	$BF,$DE,$1E,$3F,$F1,$F6,$C7,$F2,$86,$A2,$38,$7F,$D1,$31,$FE,$E3
	.dc.b	$23,$FE,$C7,$7F,$DC,$D7,$23,$FD,$CD,$CA,$FC,$19,$EE,$21,$F6,$DF
	.dc.b	$EC,$89,$CC,$62,$BC,$46,$E3,$33,$BA,$E4,$FF,$A2,$7C,$EF,$E5,$56
	.dc.b	$EC,$46,$1B,$87,$F2,$B9,$D4,$E7,$D5,$F8,$1F,$6D,$FE,$93,$6D,$78
	.dc.b	$AB,$A4,$62,$3E,$34,$61,$4F,$C1,$7F,$DC,$D9,$D7,$FD,$62,$FF,$A7
	.dc.b	$FB,$79,$47,$C3,$C5,$18,$BF,$53,$63,$33,$D8,$FF,$7B,$57,$57,$FD
	.dc.b	$5A,$F7,$AF,$A1,$C1,$F1,$A7,$D5,$AF,$F6,$19,$F9,$D1,$EA,$8C,$69
	.dc.b	$F4,$EB,$84,$FB,$B0,$7D,$2F,$E1,$7E,$E3,$FD,$70,$D9,$F8,$D3,$A5
	.dc.b	$3A,$69,$F7,$3C,$D7,$9D,$D3,$3F,$1C,$A7,$38,$78,$A3,$15,$BC,$BF
	.dc.b	$BF,$D6,$AF,$8D,$7C,$23,$E2,$9F,$92,$CD,$E7,$73,$9F,$CA,$36,$33
	.dc.b	$1E,$23,$25,$FC,$05,$E2,$FE,$67,$D7,$FC,$D6,$1F,$82,$9C,$D2,$73
	.dc.b	$3F,$9F,$9C,$C6,$25,$18,$7C,$F3,$F8,$82,$9F,$88,$3F,$47,$4F,$31
	.dc.b	$E1,$7E,$8A,$B3,$18,$7F,$5A,$62,$35,$5F,$BD,$AE,$51,$88,$7B,$0F
	.dc.b	$87,$FF,$61,$FD,$DF,$AA,$31,$E2,$FF,$05,$8C,$4E,$B1,$FD,$BD,$6A
	.dc.b	$D5,$F0,$9F,$A8,$7B,$48,$7B,$4F,$26,$8F,$23,$F8,$97,$0F,$CA,$D3
	.dc.b	$E6,$46,$31,$F8,$75,$BB,$F7,$9C,$A7,$4F,$E6,$7E,$8F,$F7,$B5,$F9
	.dc.b	$66,$C6,$4F,$DC,$C6,$5A,$7F,$23,$E2,$FC,$BF,$D7,$FC,$76,$CF,$82
	.dc.b	$7F,$2B,$39,$E4,$3F,$36,$3E,$09,$EF,$21,$8A,$FD,$1D,$F1,$25,$F5
	.dc.b	$8C,$BD,$E7,$99,$3C,$67,$9D,$F9,$6E,$BF,$E0,$91,$8D,$7F,$C8,$6D
	.dc.b	$8C,$CE,$EB,$FC,$CF,$E9,$BC,$6F,$E8,$5E,$A7,$F9,$7B,$C7,$FF,$37
	.dc.b	$36,$75,$48,$C5,$70,$C3,$F9,$CD,$F0,$6D,$F0,$AD,$DF,$19,$CD,$F1
	.dc.b	$8F,$EA,$58,$FA,$5F,$F3,$B3,$5F,$F3,$FB,$5F,$E8,$ED,$78,$E3,$FB
	.dc.b	$87,$8B,$5E,$2D,$78,$B5,$FE,$1F,$AF,$F5,$7F,$BD,$FD,$DE,$DF,$49
	.dc.b	$BC,$09,$D6,$FF,$C7,$63,$0A,$7F,$45,$8F,$97,$03,$49,$8F,$87,$1F
	.dc.b	$C6,$D7,$6E,$0E,$B7,$F9,$3C,$7F,$1B,$8F,$53,$C1,$3A,$9F,$E9,$99
	.dc.b	$7F,$7B,$FC,$03,$F8,$87,$81,$E7,$94,$62,$75,$7F,$A1,$8F,$D8,$72
	.dc.b	$7E,$B7,$2B,$C9,$FE,$05,$6E,$1E,$13,$F6,$05,$3E,$8F,$27,$FA,$B5
	.dc.b	$F7,$ED,$F3,$5B,$19,$DF,$EC,$73,$FF,$09,$FB,$FF,$E1,$BD,$5F,$E1
	.dc.b	$36,$D9,$A7,$F0,$5E,$CF,$E4,$19,$5E,$CF,$85,$97,$F1,$ED,$FC,$8D
	.dc.b	$9F,$A7,$A7,$8F,$4F,$A3,$3E,$38,$FC,$15,$7F,$36,$8F,$F0,$C7,$67
	.dc.b	$F3,$9E,$DF,$E7,$ED,$8C,$CF,$F8,$1E,$72,$FF,$70,$EF,$FC,$56,$B6
	.dc.b	$3F,$08,$DC,$BF,$C4,$53,$A8,$FC,$9D,$F8,$C8,$FE,$69,$FF,$6C,$73
	.dc.b	$FF,$61,$AF,$EC,$7C,$FF,$DA,$DB,$E0,$73,$BF,$ED,$2D,$C7,$FC,$4C
	.dc.b	$FF,$8F,$FF,$00,$FE,$EC,$49,$FE,$F7,$19,$7F,$C1,$3E,$1F,$1F,$0B
	.dc.b	$FC,$15,$87,$F8,$1F,$0F,$5D,$7F,$82,$EB,$C0,$FD,$E3,$D0,$43,$F6
	.dc.b	$39,$FF,$BC,$67,$FC,$3F,$1F,$94,$3F,$C1,$7E,$27,$E6,$5A,$64,$FF
	.dc.b	$0A,$39,$7E,$39,$D4,$A7,$63,$FC,$78,$71,$3F,$C6,$1E,$37,$16,$99
	.dc.b	$9F,$F1,$E7,$57,$F9,$36,$3C,$DD,$9F,$E4,$A7,$04,$3F,$C5,$BE,$47
	.dc.b	$C5,$25,$7C,$48,$FF,$2D,$47,$C4,$E5,$FC,$7C,$79,$CF,$D3,$9F,$77
	.dc.b	$BE,$3F,$19,$AF,$F9,$93,$CA,$7A,$FF,$E6,$B9,$FF,$3E,$36,$7F,$CF
	.dc.b	$8D,$78,$35,$FF,$3F,$79,$8D,$FF,$6D,$08,$C3,$FB,$96,$BF,$E7,$43
	.dc.b	$8C,$70,$8C,$C6,$5A,$FF,$9E,$9A,$F0,$3A,$47,$12,$38,$21,$DC,$F2
	.dc.b	$3A,$71,$1D,$1C,$9C,$0F,$14,$6D,$A6,$4F,$23,$96,$99,$3C,$C7,$67
	.dc.b	$4E,$0E,$66,$4E,$07,$88,$C1,$0D,$90,$EE,$64,$A6,$4E,$C6,$08,$60
	.dc.b	$87,$73,$81,$4D,$94,$DC,$7C,$72,$47,$B2,$34,$48,$CB,$EC,$6B,$E7
	.dc.b	$39,$C7,$18,$CC,$78,$9A,$F8,$9A,$F0,$3A,$47,$12,$38,$21,$DC,$F2
	.dc.b	$3A,$71,$1D,$1C,$9C,$0F,$14,$6D,$A6,$4F,$23,$96,$99,$3C,$C7,$67
	.dc.b	$4E,$0E,$66,$4E,$07,$88,$C1,$0D,$90,$EE,$64,$A6,$4E,$C6,$08,$60
	.dc.b	$87,$73,$81,$4D,$94,$DC,$7C,$72,$47,$B2,$34,$48,$CB,$EC,$6B,$E7
	.dc.b	$39,$C7,$18,$CC,$78,$9A,$F8,$9A,$F0,$3A,$47,$12,$38,$21,$DC,$F2
	.dc.b	$3A,$7F,$09,$E3,$3C,$6F,$F2,$9C,$7F,$96,$B0,$FF,$08,$4E,$5B,$3A
	.dc.b	$9F,$E2,$98,$17,$8B,$FC,$32,$30,$D8,$C9,$FC,$76,$32,$D3,$27,$42
	.dc.b	$9F,$CB,$9D,$B9,$8E,$84,$79,$35,$D1,$82,$18,$21,$D4,$EE,$7F,$39
	.dc.b	$8E,$EE,$63,$A3,$4E,$65,$39,$9C,$4E,$78,$FF,$4E,$6C,$F4,$75,$3D
	.dc.b	$1A,$74,$29,$D2,$79,$46,$DB,$1B,$6B,$B9,$ED,$18,$8F,$31,$0E,$A5
	.dc.b	$74,$E0,$8E,$09,$1D,$09,$1E,$C2,$4F,$96,$32,$D8,$C9,$E7,$8C,$B4
	.dc.b	$C9,$E9,$29,$E5,$76,$E6,$3A,$11,$E4,$D7,$46,$08,$60,$87,$53,$B9
	.dc.b	$F0,$23,$BB,$98,$E8,$D3,$99,$4E,$67,$13,$9C,$FC,$86,$CF,$47,$53
	.dc.b	$D1,$A7,$42,$9D,$27,$94,$6D,$B1,$B6,$BB,$9E,$D1,$88,$F3,$10,$EA
	.dc.b	$57,$4E,$08,$E0,$91,$D0,$91,$EC,$24,$F9,$63,$2D,$8C,$9E,$78,$CB
	.dc.b	$4C,$9E,$92,$9E,$57,$6E,$63,$A1,$1E,$4D,$74,$60,$86,$0F,$F2,$16
	.dc.b	$2D,$5E,$67,$F8,$9F,$4F,$F1,$E4,$62,$BF,$C7,$11,$89,$D5,$71,$7F
	.dc.b	$D1,$84,$9F,$F3,$AB,$5F,$E4,$C3,$D0,$77,$7F,$29,$9F,$F1,$B4,$62
	.dc.b	$7C,$07,$07,$11,$A9,$E1,$3F,$E8,$2C,$7F,$D3,$84,$70,$7F,$BA,$F0
	.dc.b	$F7,$79,$3F,$E9,$E6,$C6,$72,$4E,$5B,$1E,$85,$C4,$FA,$AF,$E4,$B1
	.dc.b	$86,$C6,$67,$F9,$CC,$E5,$B0,$FA,$CF,$E9,$1E,$77,$F9,$8E,$BA,$C6
	.dc.b	$2B,$C1,$5B,$55,$B8,$C3,$EF,$8F,$F5,$FE,$55,$FF,$74,$BD,$7E,$FB
	.dc.b	$EF,$47,$23,$D8,$E2,$75,$87,$86,$7F,$AE,$29,$D4,$7C,$C3,$B9,$4E
	.dc.b	$E7,$DE,$F0,$1E,$03,$EC,$BE,$7C,$FF,$67,$9C,$C6,$21,$18,$BF,$A3
	.dc.b	$1B,$AE,$51,$88,$DC,$7D,$38,$7E,$05,$FE,$DA,$D7,$F3,$8D,$8C,$D7
	.dc.b	$96,$B8,$4F,$FB,$83,$5F,$16,$1E,$86,$CF,$D7,$BF,$04,$E5,$F1,$5B
	.dc.b	$A7,$F3,$7D,$DF,$DA,$9B,$B7,$D3,$9D,$71,$2B,$FB,$EE,$1F,$EB,$2C
	.dc.b	$3C,$3C,$DF,$B3,$7B,$B7,$8B,$FD,$51,$3B,$9F,$85,$4F,$31,$F2,$5F
	.dc.b	$E1,$3B,$FB,$F9,$77,$9C,$BF,$9E,$BF,$D9,$DF,$1B,$E1,$18,$AD,$DF
	.dc.b	$2B,$F4,$DF,$B7,$2F,$56,$7E,$DD,$3D,$F5,$F8,$1D,$1F,$61,$FB,$0C
	.dc.b	$F0,$D9,$FC,$14,$EA,$7C,$91,$89,$D3,$E9,$D3,$FC,$5D,$F8,$27,$B3
	.dc.b	$3F,$44,$A3,$1A,$FB,$9B,$5C,$30,$FC,$B5,$7A,$B3,$F4,$69,$EA,$8F
	.dc.b	$81,$F1,$DF,$E3,$BC,$79,$B6,$F3,$3F,$3D,$B3,$F3,$9F,$D3,$7D,$6F
	.dc.b	$EF,$1F,$D6,$B6,$7F,$57,$CC,$8C,$47,$F3,$18,$C3,$FB,$E9,$F9,$5E
	.dc.b	$A7,$C2,$E2,$FF,$7B,$B7,$80,$D3,$E3,$D7,$B4,$F7,$F0,$FD,$47,$F6
	.dc.b	$E7,$D0,$F3,$39,$13,$0F,$A3,$1F,$B2,$B4,$F4,$AE,$53,$FB,$E3,$99
	.dc.b	$48,$78,$4C,$94,$FF,$33,$78,$5F,$E6,$B7,$D3,$BF,$F4,$48,$C6,$3F
	.dc.b	$E6,$26,$C6,$7C,$17,$85,$FE,$6F,$AD,$46,$2F,$99,$D9,$C4,$6A,$78
	.dc.b	$57,$36,$D7,$F9,$E7,$0F,$DC,$BF,$D5,$8C,$70,$FE,$DD,$C1,$FA,$22
	.dc.b	$C7,$07,$83,$5F,$EB,$ED,$79,$F2,$72,$79,$1F,$A7,$CF,$FB,$97,$8F
	.dc.b	$FB,$8D,$DB,$7D,$2D,$D3,$98,$C3,$6B,$8C,$F8,$B2,$F4,$61,$71,$F0
	.dc.b	$65,$E3,$8F,$09,$FB,$2D,$78,$B6,$FB,$9F,$92,$2B,$9C,$7F,$41,$C1
	.dc.b	$F9,$8F,$16,$F1,$7F,$A8,$DB,$46,$9F,$AC,$7D,$87,$FA,$A7,$2F,$E3
	.dc.b	$5C,$9F,$EA,$BD,$74,$4C,$BD,$55,$F4,$35,$F6,$BF,$39,$AF,$CE,$3E
	.dc.b	$D5,$FF,$06,$FC,$17,$F8,$1C,$1A,$4A,$FE,$2A,$7B,$A3,$F4,$5D,$1F
	.dc.b	$11,$B1,$9E,$1F,$AB,$39,$9F,$96,$78,$4F,$39,$0F,$77,$37,$F3,$2D
	.dc.b	$FC,$64,$BF,$E6,$1C,$FF,$EC,$B8,$FB,$D1,$F8,$4A,$FE,$58,$D7,$F9
	.dc.b	$4B,$5F,$DC,$35,$F1,$93,$8B,$F0,$07,$E4,$F1,$FE,$B5,$5F,$D8,$6B
	.dc.b	$F1,$B5,$FD,$97,$AF,$A5,$6A,$3F,$B4,$76,$FF,$91,$FF,$06,$FE,$D8
	.dc.b	$48,$CD,$A3,$FA,$BF,$6F,$F7,$8C,$6F,$7F,$48,$C5,$FF,$73,$8F,$46
	.dc.b	$7E,$6E,$1F,$B7,$3E,$97,$EE,$E3,$FE,$13,$3F,$1D,$1F,$E1,$A6,$BF
	.dc.b	$CB,$9A,$FF,$2D,$6B,$CF,$87,$F4,$73,$DA,$BF,$6D,$5F,$E3,$2A,$FC
	.dc.b	$C1,$FE,$32,$D7,$9C,$67,$93,$FC,$65,$18,$F0,$7F,$C8,$15,$FE,$35
	.dc.b	$DB,$F7,$D7,$FE,$41,$8C,$1F,$C2,$B7,$FF,$12,$F8,$5F,$0C,$93,$F4
	.dc.b	$A7,$FC,$41,$1F,$13,$3F,$31,$34,$F4,$9F,$C2,$FC,$4F,$F3,$47,$67
	.dc.b	$F9,$96,$FF,$92,$1F,$E6,$9A,$38,$2C,$F0,$7F,$83,$4F,$EB,$6B,$51
	.dc.b	$FE,$00,$8C,$1F,$E8,$0C,$33,$87,$FA,$1A,$30,$49,$FE,$55,$E3,$7F
	.dc.b	$C0,$17,$F9,$4B,$5C,$F6,$FF,$23,$51,$97,$FD,$0E,$D7,$D2,$D7,$71
	.dc.b	$C6,$30,$47,$04,$8F,$E6,$11,$96,$C6,$5F,$2B,$5F,$29,$A2,$1A,$29
	.dc.b	$A3,$04,$30,$43,$D6,$F5,$69,$C8,$A7,$12,$9C,$4F,$82,$F1,$69,$C4
	.dc.b	$A7,$12,$9C,$4F,$8E,$6C,$A6,$CA,$6E,$3E,$33,$96,$B9,$8F,$13,$C0
	.dc.b	$F3,$90,$F0,$94,$D1,$82,$18,$24,$76,$78,$35,$DB,$5D,$C7,$8A,$30
	.dc.b	$47,$04,$8F,$81,$19,$6C,$65,$F7,$35,$F7,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$3D,$6F,$56,$9C,$8A,$71,$29,$C4,$F8,$2F,$16,$9C,$4A,$71,$29
	.dc.b	$C4,$F8,$E6,$CA,$6C,$A6,$E3,$E3,$39,$6B,$98,$F1,$3C,$0F,$39,$0F
	.dc.b	$09,$4D,$18,$21,$82,$47,$67,$83,$5D,$B5,$DC,$78,$A3,$04,$70,$48
	.dc.b	$F8,$11,$F8,$9F,$03,$FC,$A5,$1F,$E3,$F6,$BF,$E4,$53,$FC,$7E,$E1
	.dc.b	$FE,$23,$E4,$7F,$2A,$F0,$3F,$8D,$B6,$3F,$8F,$B5,$FE,$32,$D7,$88
	.dc.b	$8E,$2D,$38,$94,$E2,$53,$8B,$C0,$8F,$06,$BC,$1A,$F0,$34,$43,$45
	.dc.b	$34,$60,$86,$08,$7F,$36,$8F,$63,$4F,$31,$63,$C2,$D3,$B1,$4E,$A5
	.dc.b	$3A,$D2,$7A,$B6,$3A,$B6,$7A,$B6,$3A,$B5,$EA,$D7,$AC,$F4,$23,$D1
	.dc.b	$B3,$D1,$B1,$D1,$AF,$46,$BD,$0E,$44,$8E,$4D,$39,$14,$E4,$53,$93
	.dc.b	$C0,$8F,$06,$BC,$1A,$F0,$34,$43,$45,$34,$60,$86,$08,$7D,$32,$4F
	.dc.b	$C9,$6C,$FC,$06,$9E,$92,$CF,$91,$A7,$84,$A7,$80,$A7,$82,$7B,$12
	.dc.b	$7B,$36,$3B,$36,$7B,$36,$3B,$35,$EC,$D7,$B4,$F4,$23,$D1,$B3,$D1
	.dc.b	$B1,$D1,$AF,$46,$BD,$0E,$44,$8E,$4D,$39,$14,$E4,$53,$93,$C0,$8F
	.dc.b	$06,$BF,$A8,$FC,$45,$E7,$34,$79,$1F,$ED,$69,$CC,$62,$BF,$DB,$91
	.dc.b	$89,$D5,$71,$7F,$DB,$BF,$4E,$F8,$FC,$5F,$FB,$8E,$31,$87,$38,$FF
	.dc.b	$25,$88,$CD,$72,$AE,$33,$AF,$15,$C3,$88,$FF,$7D,$35,$46,$B0,$ED
	.dc.b	$1A,$6A,$F6,$9F,$79,$C5,$B3,$A6,$E3,$FE,$F8,$6D,$7F,$B8,$E3,$FC
	.dc.b	$BC,$47,$07,$FC,$01,$8E,$7F,$35,$C6,$FE,$73,$6C,$D3,$FC,$D0,$4A
	.dc.b	$FF,$78,$35,$FF,$6B,$63,$FC,$9B,$B1,$CF,$F1,$38,$3F,$CE,$04,$AF
	.dc.b	$9B,$83,$C8,$7C,$B7,$F9,$EB,$CE,$FA,$35,$F8,$19,$CA,$8C,$D7,$1A
	.dc.b	$E1,$3D,$2B,$75,$A5,$F1,$DF,$D5,$EF,$C0,$D8,$CD,$7F,$53,$9C,$BF
	.dc.b	$D5,$7B,$1D,$3E,$C9,$79,$DE,$69,$CE,$9F,$97,$8C,$56,$DB,$19,$BF
	.dc.b	$6E,$BF,$1B,$6F,$91,$3F,$1E,$72,$D7,$E6,$ED,$F0,$3E,$31,$DB,$CE
	.dc.b	$7F,$D4,$3A,$27,$2F,$9A,$3E,$7C,$63,$2E,$55,$C6,$7F,$D5,$39,$78
	.dc.b	$F0,$FA,$D1,$E0,$3C,$79,$7C,$5D,$5F,$29,$F4,$27,$C3,$18,$7E,$B1
	.dc.b	$F4,$E3,$4F,$DB,$6E,$7F,$81,$FA,$CF,$F5,$CF,$DA,$FD,$76,$D7,$DB
	.dc.b	$6B,$D9,$AF,$E1,$7E,$BB,$9E,$06,$92,$3C,$67,$AC,$87,$C0,$DF,$E7
	.dc.b	$37,$2F,$77,$CD,$74,$24,$72,$3F,$28,$F8,$3E,$27,$FD,$A3,$D3,$FC
	.dc.b	$5E,$7E,$1C,$63,$4F,$CA,$36,$32,$7E,$FE,$3E,$6E,$DF,$42,$3C,$4D
	.dc.b	$C3,$F9,$1E,$9F,$A5,$AF,$E2,$B1,$8A,$FD,$24,$63,$B9,$5B,$9F,$F7
	.dc.b	$2D,$6F,$9F,$CD,$0E,$6F,$CE,$FC,$7F,$F2,$FF,$C2,$33,$18,$C9,$18
	.dc.b	$E8,$56,$E7,$F9,$3D,$6E,$BA,$DB,$97,$F1,$8F,$B8,$87,$D2,$F8,$47
	.dc.b	$DC,$BD,$FF,$E8,$1C,$5E,$69,$CC,$63,$2F,$E6,$B1,$87,$E5,$58,$46
	.dc.b	$6B,$D1,$EC,$7E,$D6,$BC,$FA,$A3,$17,$C7,$4F,$DF,$C6,$23,$E3,$4E
	.dc.b	$EB,$F8,$09,$FC,$3F,$C4,$F9,$D3,$FC,$5D,$B1,$FC,$75,$BD,$1F,$D2
	.dc.b	$E7,$EB,$9F,$92,$3E,$37,$3F,$E4,$9B,$C5,$FC,$10,$FD,$87,$D7,$72
	.dc.b	$24,$72,$3F,$93,$3E,$27,$8F,$37,$E2,$DA,$7D,$03,$F1,$A5,$32,$78
	.dc.b	$FC,$97,$B4,$F2,$FE,$67,$8F,$FE,$79,$9C,$C6,$2F,$FD,$19,$39,$AE
	.dc.b	$17,$FD,$82,$72,$47,$EB,$5F,$27,$46,$27,$8D,$7F,$A1,$F9,$7F,$93
	.dc.b	$4F,$F4,$43,$E2,$95,$CB,$FC,$58,$C6,$7F,$DD,$1B,$9F,$17,$89,$B3
	.dc.b	$FA,$46,$7F,$89,$F4,$3F,$4B,$BF,$F5,$58,$C3,$A9,$4E,$B6,$56,$A7
	.dc.b	$FB,$19,$E0,$21,$DC,$FE,$4F,$BE,$DF,$C4,$1D,$2F,$F1,$8D,$36,$53
	.dc.b	$F2,$7C,$3F,$E2,$5F,$4F,$F3,$B2,$69,$1F,$D3,$F3,$FE,$B5,$7C,$2F
	.dc.b	$EB,$63,$0D,$E8,$FF,$18,$36,$76,$7A,$70,$FE,$4C,$7B,$0F,$79,$FA
	.dc.b	$CE,$AF,$F1,$C7,$E0,$9F,$C1,$DB,$39,$11,$96,$9F,$C6,$7B,$7E,$27
	.dc.b	$07,$F1,$DE,$0F,$E2,$ED,$9D,$61,$F0,$CF,$B6,$78,$88,$79,$8F,$E6
	.dc.b	$27,$A8,$FE,$63,$8F,$D9,$EF,$FE,$10,$D8,$D1,$FC,$F8,$FB,$E7,$F8
	.dc.b	$37,$C0,$FC,$33,$4C,$9F,$D8,$9C,$BE,$23,$CA,$78,$CF,$2D,$7F,$88
	.dc.b	$7C,$1D,$C7,$F5,$B3,$99,$71,$FF,$0D,$46,$A7,$FB,$44,$BF,$0A,$FE
	.dc.b	$D0,$D3,$FB,$41,$4F,$ED,$05,$3F,$B9,$1D,$A3,$FC,$65,$E2,$75,$E5
	.dc.b	$FE,$F5,$87,$F8,$A6,$3C,$B5,$EA,$27,$3F,$EC,$1F,$5F,$8F,$FE,$10
	.dc.b	$E4,$FF,$06,$E3,$F9,$39,$FF,$08,$F8,$DF,$97,$6C,$65,$F8,$D7,$CD
	.dc.b	$F8,$C4,$F0,$3F,$C5,$B3,$B8,$FE,$67,$D5,$CD,$FE,$71,$E0,$7F,$8D
	.dc.b	$71,$FF,$17,$D6,$CF,$F1,$DC,$FF,$90,$FC,$8F,$E0,$51,$88,$FF,$24
	.dc.b	$46,$36,$FF,$27,$ED,$FE,$49,$F2,$7E,$21,$2B,$EA,$9F,$0B,$2F,$E1
	.dc.b	$CF,$E9,$4C,$72,$BE,$83,$F8,$89,$23,$FC,$56,$FF,$9B,$C8,$FF,$9D
	.dc.b	$7C,$CF,$F5,$04,$62,$3F,$CD,$BE,$27,$F1,$A2,$3F,$E3,$47,$FC,$F0
	.dc.b	$47,$96,$9F,$CC,$9F,$F4,$DF,$8B,$FE,$8A,$9F,$E5,$5E,$17,$FA,$4D
	.dc.b	$B8,$FF,$AB,$DB,$7C,$6F,$FD,$2D,$1F,$EA,$18,$C4,$7F,$A9,$7C,$EE
	.dc.b	$0F,$FA,$AD,$AF,$FA,$AB,$C2,$FF,$56,$D7,$FA,$D2,$B8,$F9,$FF,$EB
	.dc.b	$A8,$C1,$2B,$FD,$03,$E8,$BD,$1F,$E7,$10,$8C,$3F,$CE,$DA,$FF,$45
	.dc.b	$6B,$CA,$38,$C6,$23,$FD,$7A,$D7,$FC,$10,$D7,$B4,$70,$23,$DD,$AF
	.dc.b	$76,$BD,$CD,$10,$D1,$4D,$18,$21,$82,$1E,$67,$AB,$4E,$45,$38,$94
	.dc.b	$E2,$7B,$1E,$2D,$38,$94,$E2,$53,$89,$F0,$8D,$94,$D9,$4D,$C7,$C9
	.dc.b	$24,$78,$23,$47,$94,$87,$72,$9A,$30,$43,$04,$8E,$AF,$91,$AF,$91
	.dc.b	$AF,$92,$3C,$51,$88,$EE,$D7,$B3,$5E,$D1,$C0,$8F,$76,$BD,$DA,$F7
	.dc.b	$34,$43,$45,$34,$60,$86,$08,$79,$9E,$AD,$39,$14,$E2,$53,$89,$EC
	.dc.b	$78,$B4,$E2,$53,$89,$4E,$27,$C2,$36,$53,$65,$37,$1F,$24,$91,$E0
	.dc.b	$8D,$1E,$52,$1D,$CA,$68,$C1,$0C,$12,$3A,$BE,$46,$BE,$46,$BE,$48
	.dc.b	$F1,$46,$23,$BB,$5E,$CD,$5E,$47,$F9,$55,$FF,$46,$1F,$E5,$5A,$FF
	.dc.b	$25,$46,$63,$2D,$7F,$85,$FC,$1F,$FA,$4F,$CA,$FE,$28,$D8,$FE,$3A
	.dc.b	$D7,$F8,$9B,$5E,$22,$38,$B4,$E2,$53,$89,$4E,$2F,$02,$3C,$1A,$F0
	.dc.b	$6B,$C0,$D1,$0D,$14,$D1,$82,$18,$21,$FC,$DE,$3D,$8D,$3C,$C5,$8F
	.dc.b	$0B,$4E,$C5,$3A,$94,$EB,$69,$EA,$D8,$EA,$D9,$EA,$D8,$EA,$D7,$AB
	.dc.b	$5E,$B3,$D0,$8F,$46,$CF,$46,$C7,$46,$BD,$1A,$F4,$39,$12,$39,$34
	.dc.b	$E4,$53,$91,$4E,$4F,$02,$3C,$1A,$F0,$6B,$C0,$D1,$0D,$14,$D1,$82
	.dc.b	$18,$21,$F4,$C9,$3F,$25,$B3,$F0,$1A,$7A,$4B,$3E,$46,$9E,$12,$9E
	.dc.b	$02,$9E,$09,$EC,$49,$EC,$D8,$EC,$D9,$EC,$D8,$EC,$D7,$B3,$5E,$D3
	.dc.b	$D0,$8F,$46,$CF,$46,$C7,$46,$BD,$1A,$F4,$39,$12,$39,$34,$E4,$53
	.dc.b	$91,$4E,$4F,$02,$3C,$1F,$F7,$36,$EF,$C0,$9A,$3D,$3C,$C6,$1B,$19
	.dc.b	$BF,$E3,$E2,$B9,$46,$63,$F9,$1F,$C0,$BC,$BE,$51,$C2,$31,$3B,$80
	.dc.b	$8C,$DF,$FB,$ED,$B1,$E0,$E3,$71,$F4,$8F,$E6,$31,$8B,$E9,$08,$C5
	.dc.b	$FF,$96,$0F,$4C,$38,$5F,$1F,$CE,$3C,$71,$8B,$E6,$DA,$E2,$D9,$F8
	.dc.b	$7D,$EE,$8F,$2D,$76,$8C,$65,$E4,$6C,$66,$77,$5F,$33,$57,$CB,$7C
	.dc.b	$4A,$FA,$ED,$9D,$35,$FF,$3C,$B5,$F9,$77,$E0,$F5,$BB,$62,$69,$27
	.dc.b	$99,$92,$BE,$07,$07,$62,$9D,$B1,$FA,$7C,$3D,$89,$1F,$12,$32,$FF
	.dc.b	$66,$D1,$DA,$F9,$BF,$C0,$FA,$9F,$8C,$9C,$C6,$2B,$85,$6D,$56,$E3
	.dc.b	$0D,$F8,$DF,$EC,$D7,$D9,$B2,$69,$FD,$AF,$C4,$78,$8F,$B7,$FD,$1F
	.dc.b	$B3,$F9,$E9,$CC,$63,$2F,$EF,$F1,$8C,$55,$BB,$E5,$18,$AF,$85,$BF
	.dc.b	$F5,$CE,$0F,$C1,$EF,$F1,$A7,$FC,$13,$18,$7A,$DB,$C3,$F0,$47,$FB
	.dc.b	$69,$AB,$2F,$4C,$61,$B8,$F8,$67,$AF,$07,$D2,$6E,$5E,$38,$C4,$7D
	.dc.b	$B8,$FF,$03,$4B,$F0,$6F,$E0,$D3,$F4,$0F,$B4,$E3,$83,$EC,$36,$33
	.dc.b	$D7,$F6,$5F,$F7,$8F,$8D,$FC,$3B,$F0,$BF,$8C,$D8,$F7,$B5,$DE,$09
	.dc.b	$EB,$F7,$5C,$E7,$F8,$BE,$08,$C5,$79,$DB,$19,$9F,$41,$F0,$38,$3E
	.dc.b	$83,$EF,$6D,$7B,$FE,$C7,$89,$27,$89,$E0,$7C,$5E,$D3,$F0,$D7,$D1
	.dc.b	$FF,$1A,$F0,$7F,$95,$DF,$F1,$CA,$D4,$E7,$0F,$F2,$3C,$62,$7F,$90
	.dc.b	$46,$2F,$9C,$60,$F4,$1F,$8A,$C7,$D2,$B9,$7F,$10,$FF,$2A,$E1,$7C
	.dc.b	$77,$F9,$0F,$8B,$C7,$3F,$CA,$23,$18,$FF,$96,$2B,$78,$7E,$CA,$31
	.dc.b	$3A,$AF,$E6,$B0,$F4,$1F,$09,$FD,$23,$F0,$EE,$B3,$98,$C3,$FB,$F9
	.dc.b	$DA,$9D,$C6,$1F,$2D,$7A,$F9,$BF,$26,$DC,$BF,$2C,$4C,$BF,$83,$7D
	.dc.b	$3B,$E3,$BF,$AC,$FC,$2F,$8A,$31,$9A,$31,$3A,$8F,$E2,$53,$A9,$F8
	.dc.b	$BD,$5F,$C0,$AF,$F8,$0D,$7E,$C9,$B8,$FF,$56,$9D,$4F,$F0,$03,$45
	.dc.b	$8F,$8F,$CD,$FC,$E9,$FE,$2F,$1F,$A8,$F0,$FE,$74,$69,$AF,$E7,$1A
	.dc.b	$FC,$1B,$FB,$FC,$5F,$7A,$3F,$30,$7E,$CB,$E0,$B8,$DF,$ED,$DB,$81
	.dc.b	$A7,$4A,$E5,$5D,$4F,$C6,$10,$FC,$79,$0E,$E5,$3F,$D4,$30,$F3,$BF
	.dc.b	$BC,$F9,$5F,$E8,$FF,$3B,$C1,$18,$CB,$FD,$43,$5B,$BF,$EE,$91,$89
	.dc.b	$D4,$7F,$AA,$4C,$96,$FA,$57,$16,$CF,$09,$FC,$31,$FD,$78,$87,$F3
	.dc.b	$02,$9F,$C8,$0A,$7F,$1F,$3F,$8C,$CF,$F8,$4B,$D4,$FC,$DE,$5F,$26
	.dc.b	$51,$89,$D6,$7F,$46,$72,$F7,$61,$DC,$A7,$DD,$36,$7E,$04,$A6,$67
	.dc.b	$C3,$A7,$68,$FC,$87,$85,$FE,$BF,$A7,$AB,$F9,$88,$C3,$74,$FF,$1D
	.dc.b	$B6,$76,$7E,$83,$8B,$FC,$7F,$F8,$7F,$E9,$20,$D2,$3E,$8C,$3E,$69
	.dc.b	$78,$3F,$C3,$6F,$F7,$7E,$7F,$DE,$FE,$21,$FB,$CE,$9F,$F2,$5B,$6C
	.dc.b	$D3,$F8,$9F,$3F,$F1,$0B,$F8,$2D,$DF,$3D,$5F,$ED,$B6,$ED,$F0,$EB
	.dc.b	$EC,$1E,$3E,$9F,$F2,$1B,$F7,$3B,$7F,$96,$7C,$87,$F2,$06,$E7,$FC
	.dc.b	$B7,$BF,$FC,$9F,$39,$EF,$FE,$5E,$FE,$03,$C1,$FC,$16,$1F,$D7,$9A
	.dc.b	$FE,$F1,$AF,$F4,$33,$F0,$A7,$8F,$5F,$F2,$6B,$F2,$0F,$C1,$1F,$C1
	.dc.b	$3C,$2E,$38,$FF,$57,$8E,$26,$4A,$7F,$5E,$8E,$FD,$BF,$E1,$6F,$13
	.dc.b	$FC,$EF,$F3,$FF,$DA,$AC,$D2,$5F,$F7,$23,$9C,$7F,$7B,$F1,$BC,$3D
	.dc.b	$BF,$DF,$38,$BF,$BC,$C7,$B2,$BD,$07,$CE,$3D,$25,$3F,$C2,$1D,$7F
	.dc.b	$E1,$5B,$FC,$9C,$7F,$86,$3C,$7F,$C8,$B6,$32,$FF,$86,$5A,$F3,$3B
	.dc.b	$93,$A7,$F1,$AF,$F8,$8B,$C9,$E6,$FE,$89,$AF,$89,$BA,$FF,$8C,$FB
	.dc.b	$7F,$C8,$53,$FE,$7E,$BF,$F2,$13,$7C,$BF,$F2,$13,$7C,$BF,$AF,$3F
	.dc.b	$E5,$08,$F8,$F3,$FE,$4F,$D7,$FC,$9D,$E6,$7D,$88,$C4,$FD,$F3,$F8
	.dc.b	$59,$EA,$F3,$FF,$A0,$7B,$1F,$E6,$92,$47,$F9,$B9,$FF,$39,$91,$FF
	.dc.b	$22,$B5,$FF,$37,$78,$BF,$E7,$12,$3F,$E7,$67,$FC,$F6,$47,$FC,$9A
	.dc.b	$D7,$FD,$56,$7F,$A7,$7D,$2F,$F5,$EC,$7F,$A6,$1B,$1F,$EB,$AF,$4B
	.dc.b	$FD,$3D,$39,$CF,$F9,$B4,$FF,$AB,$3A,$36,$D8,$CB,$FD,$02,$32,$D8
	.dc.b	$CB,$FE,$BB,$F4,$F2,$D8,$CF,$9F,$FE,$C2,$BF,$F5,$F5,$6B,$3F,$F6
	.dc.b	$5C,$7F,$B4,$09,$7F,$EA,$5E,$6E,$7E,$AB,$D5,$FD,$7C,$A3,$0F,$EC
	.dc.b	$9A,$FF,$B0,$1A,$FF,$B6,$9A,$FC,$86,$BF,$ED,$46,$BD,$63,$BC,$60
	.dc.b	$8E,$09,$1E,$16,$BE,$16,$BE,$16,$BE,$16,$BE,$16,$BE,$13,$A1,$0E
	.dc.b	$85,$3A,$14,$E8,$53,$A1,$4E,$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23
	.dc.b	$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$3E,$9B,$ED
	.dc.b	$6B,$E6,$6B,$E1,$6B,$D9,$AF,$56,$BD,$63,$E9,$C6,$08,$E0,$91,$E1
	.dc.b	$6B,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$3A,$10,$E8,$53,$A1,$4E
	.dc.b	$85,$3A,$14,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53
	.dc.b	$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$23,$E9,$BE,$D6,$BE,$66,$BE
	.dc.b	$16,$BD,$9A,$F5,$6B,$F8,$EF,$37,$FC,$BA,$FF,$2E,$3F,$CB,$05,$72
	.dc.b	$DF,$33,$F9,$1B,$FC,$57,$E6,$CB,$C6,$FE,$29,$18,$6C,$64,$FE,$2D
	.dc.b	$19,$69,$93,$F8,$C3,$96,$99,$3F,$8D,$39,$69,$93,$C8,$5B,$46,$E3
	.dc.b	$33,$E4,$23,$E2,$6B,$E0,$6B,$D9,$AF,$23,$81,$0E,$05,$38,$14,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$12,$7D,$AF,$BE,$7D,$D1,$89,$F3,$B6,$7C
	.dc.b	$4D,$7B,$B5,$EC,$D7,$B4,$F4,$23,$DC,$93,$DA,$32,$D8,$C9,$E0,$8C
	.dc.b	$B4,$C9,$E2,$72,$D3,$27,$91,$CB,$4C,$9E,$62,$CF,$A4,$91,$C2,$35
	.dc.b	$3E,$52,$3E,$36,$BE,$16,$BD,$DA,$F2,$38,$10,$E0,$53,$81,$4E,$06
	.dc.b	$88,$68,$A6,$8C,$10,$C1,$27,$DA,$FC,$09,$F7,$46,$27,$CE,$D9,$F1
	.dc.b	$35,$EE,$D7,$B3,$5E,$D3,$D0,$8F,$72,$4F,$68,$CB,$63,$27,$82,$32
	.dc.b	$D3,$27,$89,$CB,$4F,$BB,$E9,$3E,$91,$C0,$F5,$FF,$8C,$52,$B5,$39
	.dc.b	$8F,$E3,$91,$89,$DB,$63,$F9,$07,$E1,$AF,$83,$EB,$5D,$23,$17,$CE
	.dc.b	$BF,$CA,$71,$89,$D5,$70,$9F,$E5,$5C,$6E,$3C,$47,$F9,$6A,$31,$87
	.dc.b	$38,$FE,$53,$18,$BF,$F2,$4C,$E6,$5E,$9B,$97,$9C,$7F,$95,$6B,$E0
	.dc.b	$CE,$6B,$83,$63,$E2,$F2,$BE,$CB,$8D,$A3,$17,$DD,$B5,$C2,$B7,$0F
	.dc.b	$7A,$92,$B8,$3D,$3C,$27,$9D,$E7,$C3,$EA,$C6,$33,$F6,$36,$33,$3B
	.dc.b	$CB,$FA,$97,$D3,$BE,$D3,$FA,$6D,$7D,$E8,$C4,$FC,$56,$C6,$67,$73
	.dc.b	$FD,$6F,$5F,$89,$EE,$FC,$6C,$BF,$12,$DD,$0D,$3F,$B1,$2F,$B2,$7C
	.dc.b	$B7,$E4,$3F,$12,$CC,$62,$51,$89,$D6,$3F,$E9,$89,$E5,$18,$8F,$F4
	.dc.b	$DE,$DF,$0E,$BE,$0E,$3C,$A7,$FB,$94,$E5,$BB,$7E,$06,$B7,$D3,$27
	.dc.b	$37,$FD,$86,$FE,$74,$E6,$7B,$DF,$E2,$63,$11,$FD,$E2,$76,$F9,$FC
	.dc.b	$C7,$E0,$3F,$DB,$73,$46,$27,$FB,$94,$62,$75,$39,$BF,$F0,$57,$94
	.dc.b	$F3,$5F,$0D,$FB,$89,$CC,$63,$0F,$EF,$95,$BA,$D6,$3F,$EC,$08,$C4
	.dc.b	$FB,$73,$F5,$4F,$C0,$C3,$D6,$B3,$F5,$BF,$03,$4E,$04,$9D,$CE,$A7
	.dc.b	$E1,$57,$47,$84,$EA,$31,$1A,$AE,$4F,$84,$F9,$1F,$69,$FE,$22,$FB
	.dc.b	$FF,$C4,$EA,$8C,$4E,$84,$EA,$33,$1F,$E2,$AC,$BE,$7F,$D1,$FC,$C2
	.dc.b	$47,$D0,$B4,$7F,$86,$F3,$F5,$4F,$C1,$CB,$CE,$DC,$BC,$EF,$B0,$F4
	.dc.b	$14,$83,$D9,$7D,$6B,$EF,$FF,$41,$9F,$FB,$CA,$73,$18,$AF,$F2,$0C
	.dc.b	$62,$7F,$46,$6E,$B9,$38,$9E,$E7,$97,$6F,$9D,$87,$9A,$3E,$74,$60
	.dc.b	$FF,$2A,$5F,$5F,$84,$FE,$83,$3E,$0C,$3A,$C6,$23,$D5,$18,$C3,$C5
	.dc.b	$8F,$F4,$8A,$D5,$78,$4F,$86,$78,$0A,$7E,$2F,$C0,$FF,$2E,$F4,$70
	.dc.b	$D7,$F9,$E4,$FF,$08,$C3,$F6,$F5,$A6,$C6,$4F,$B9,$BB,$F0,$2F,$F3
	.dc.b	$24,$FE,$BD,$B3,$A6,$BF,$D7,$A7,$83,$FE,$6D,$9F,$D2,$72,$BE,$47
	.dc.b	$F5,$91,$86,$CF,$F9,$D3,$97,$F6,$51,$88,$46,$27,$A4,$7D,$4C,$BF
	.dc.b	$16,$4E,$2F,$D0,$CF,$F0,$2C,$BC,$AD,$AE,$27,$9E,$33,$97,$A3,$87
	.dc.b	$E4,$5F,$E5,$30,$E9,$C5,$F7,$F0,$FA,$47,$86,$74,$7E,$6B,$CA,$FF
	.dc.b	$4A,$F6,$7F,$A1,$5B,$87,$FA,$42,$91,$87,$FD,$26,$D7,$3E,$53,$CD
	.dc.b	$FF,$4B,$DF,$38,$C4,$FF,$6E,$BE,$97,$C4,$E5,$97,$F5,$99,$FE,$6E
	.dc.b	$43,$F9,$89,$4F,$E2,$05,$3F,$84,$1F,$C1,$34,$FE,$88,$FD,$F3,$F8
	.dc.b	$01,$4F,$CD,$6B,$F3,$5F,$93,$A7,$F8,$79,$FD,$E1,$D4,$A7,$E0,$63
	.dc.b	$9B,$F7,$B8,$3F,$0D,$18,$DF,$F0,$13,$FA,$76,$C7,$E4,$E3,$8B,$E1
	.dc.b	$30,$F9,$37,$FC,$9B,$FA,$3E,$1F,$F6,$0F,$A1,$60,$D2,$57,$C6,$88
	.dc.b	$7C,$62,$9F,$B8,$E5,$FF,$1E,$FD,$97,$EE,$FA,$3F,$86,$37,$A3,$F8
	.dc.b	$5B,$67,$77,$CA,$FE,$19,$D4,$87,$D2,$35,$BF,$DB,$7C,$38,$7F,$19
	.dc.b	$FB,$BE,$94,$69,$1E,$3C,$DF,$E3,$3A,$FA,$57,$CD,$EE,$78,$8F,$01
	.dc.b	$E2,$9F,$E7,$AF,$F9,$61,$AF,$F4,$8E,$AF,$BD,$18,$8F,$E9,$1F,$80
	.dc.b	$7F,$47,$4D,$23,$E4,$E2,$FC,$35,$F8,$E3,$C2,$43,$C8,$47,$4D,$C3
	.dc.b	$D8,$F3,$3D,$AE,$1F,$6C,$75,$E1,$FE,$EB,$1F,$DF,$3B,$BF,$96,$BF
	.dc.b	$91,$6C,$6A,$3F,$C1,$DD,$FE,$EE,$3F,$F0,$64,$7F,$68,$8C,$1F,$9F
	.dc.b	$7C,$98,$FB,$31,$F1,$47,$AF,$1F,$49,$23,$31,$C3,$5F,$F1,$69,$1F
	.dc.b	$F1,$3B,$FE,$37,$D3,$EF,$47,$F8,$DF,$C2,$FF,$1E,$4E,$49,$1F,$E4
	.dc.b	$BC,$3F,$D2,$3C,$7F,$E4,$92,$46,$6F,$F4,$2F,$F9,$6B,$C3,$3F,$C3
	.dc.b	$8F,$E3,$1E,$2F,$F9,$6B,$B7,$FC,$B1,$19,$37,$19,$7F,$C7,$B1,$FE
	.dc.b	$46,$23,$83,$F9,$11,$FE,$7B,$F1,$BF,$D4,$B1,$FE,$80,$6C,$6B,$C7
	.dc.b	$FF,$3D,$61,$DE,$38,$36,$3F,$D0,$75,$FE,$81,$8C,$79,$1C,$AB,$F9
	.dc.b	$85,$69,$F8,$15,$FE,$6B,$23,$83,$FD,$47,$D5,$C5,$FF,$4F,$B5,$79
	.dc.b	$3F,$EA,$49,$FF,$50,$9F,$EA,$CC,$3C,$DE,$5F,$FA,$CA,$30,$4A,$FF
	.dc.b	$5E,$F9,$BC,$FC,$DF,$F6,$0C,$FF,$AF,$A3,$07,$7F,$3D,$E7,$FE,$35
	.dc.b	$B7,$FC,$B5,$AF,$FB,$31,$AF,$CE,$6B,$FD,$45,$AF,$DE,$6A,$69,$23
	.dc.b	$C0,$48,$F0,$35,$F0,$35,$F0,$35,$F0,$35,$F0,$35,$F0,$36,$3C,$11
	.dc.b	$D8,$87,$62,$9D,$8A,$76,$29,$D8,$A7,$62,$BD,$9E,$84,$3A,$14,$E8
	.dc.b	$53,$A1,$4E,$85,$7A,$3C,$88,$72,$29,$C8,$A7,$22,$BC,$9E,$04,$38
	.dc.b	$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70,$48,$FA,$84,$8F,$92,$D7
	.dc.b	$E0,$35,$F4,$B5,$F2,$35,$F0,$B5,$F0,$36,$3C,$11,$D8,$87,$62,$9D
	.dc.b	$8A,$76,$29,$D8,$A7,$62,$BD,$9E,$84,$3A,$14,$E8,$53,$A1,$4E,$85
	.dc.b	$7A,$3C,$88,$72,$29,$C8,$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83
	.dc.b	$A3,$FC,$9B,$E2,$FF,$29,$7F,$C9,$B7,$FE,$35,$F1,$13,$96,$C7,$F0
	.dc.b	$DF,$90,$FE,$87,$E3,$7F,$0D,$6D,$FF,$0D,$6B,$FC,$35,$AF,$F0,$D6
	.dc.b	$BF,$C3,$5B,$81,$A4,$84,$F7,$69,$DC,$A7,$72,$9D,$CA,$77,$2C,$F7
	.dc.b	$6C,$F7,$9E,$C4,$8E,$CD,$7B,$35,$EC,$D7,$B3,$67,$B3,$67,$B4,$F4
	.dc.b	$21,$D0,$A7,$42,$9D,$0B,$3D,$1B,$1D,$1E,$44,$39,$14,$E4,$58,$E4
	.dc.b	$D8,$E4,$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D,$8D,$46,$08,$E0
	.dc.b	$91,$F5,$09,$3F,$25,$A7,$C0,$29,$E9,$29,$E4,$29,$E1,$2C,$F8,$1B
	.dc.b	$3E,$09,$EC,$48,$EC,$D7,$B3,$5E,$CD,$7B,$36,$7B,$36,$7B,$4F,$42
	.dc.b	$1D,$0A,$74,$29,$D0,$B3,$D1,$B1,$D1,$E4,$43,$91,$4E,$45,$8E,$4D
	.dc.b	$8E,$4F,$02,$1C,$0B,$1C,$1B,$1C,$1D,$12,$34,$D8,$D4,$60,$CF,$84
	.dc.b	$E0,$6E,$F3,$FF,$C8,$B8,$2B,$53,$98,$FF,$24,$C6,$27,$6D,$8F,$F2
	.dc.b	$6F,$A6,$F1,$FA,$D7,$48,$C5,$F3,$AF,$F7,$A4,$62,$75,$5C,$27,$FC
	.dc.b	$AF,$C4,$FB,$FC,$47,$FB,$EE,$31,$87,$38,$FF,$34,$C6,$2F,$D6,$D8
	.dc.b	$FF,$30,$FC,$83,$E4,$79,$C7,$F9,$C2,$31,$7C,$E1,$18,$BF,$8A,$D8
	.dc.b	$F7,$FA,$FC,$96,$68,$C3,$6F,$A3,$6F,$E7,$FD,$D3,$BF,$93,$6B,$C4
	.dc.b	$D9,$DC,$F7,$F0,$DE,$1F,$19,$FB,$31,$89,$F5,$36,$33,$3B,$CB,$FC
	.dc.b	$FD,$F1,$BF,$E9,$09,$FC,$0F,$6A,$31,$3A,$3E,$24,$65,$F8,$DC,$8F
	.dc.b	$88,$DC,$7F,$97,$DA,$FF,$97,$DA,$FF,$97,$DB,$AF,$C5,$F4,$BF,$0D
	.dc.b	$BF,$E5,$1B,$B9,$A7,$E5,$B9,$FF,$17,$EB,$5E,$6F,$CD,$B6,$8D,$3D
	.dc.b	$BA,$FF,$AB,$9B,$9E,$8F,$73,$83,$E3,$F0,$FF,$5C,$E1,$F9,$9D,$38
	.dc.b	$25,$7F,$AE,$0F,$D5,$F9,$3F,$A9,$7F,$B2,$E2,$8C,$4F,$FA,$CA,$31
	.dc.b	$D3,$FD,$E1,$B5,$C2,$9F,$5B,$96,$C6,$63,$06,$7D,$89,$1C,$5B,$87
	.dc.b	$9A,$3C,$D8,$FA,$9F,$F6,$1B,$5C,$F3,$7E,$4E,$7C,$D8,$3E,$9F,$C4
	.dc.b	$4A,$F3,$3F,$E0,$EB,$E5,$1F,$ED,$A6,$A6,$9A,$F2,$9F,$82,$BE,$BF
	.dc.b	$FC,$33,$18,$6C,$66,$BF,$C1,$F3,$97,$FC,$23,$E7,$FF,$0F,$F9,$8F
	.dc.b	$E3,$7F,$29,$FE,$DE,$8C,$70,$23,$1C,$AA,$DD,$FD,$9B,$DD,$7D,$5F
	.dc.b	$71,$F1,$DF,$55,$B5,$F7,$1B,$C1,$B9,$F8,$5C,$3F,$26,$BE,$6B,$5F
	.dc.b	$9B,$CB,$F8,$B8,$C4,$7F,$32,$B7,$DB,$77,$24,$77,$7F,$C7,$7F,$70
	.dc.b	$F7,$3F,$4D,$BF,$FB,$F2,$30,$FE,$9C,$C4,$6A,$BF,$2D,$3A,$EC,$FE
	.dc.b	$83,$1E,$4D,$3F,$0C,$FD,$9C,$BC,$AF,$D9,$7C,$F9,$BF,$3A,$FC,$9D
	.dc.b	$72,$AF,$E3,$51,$8C,$BF,$49,$5B,$E9,$FF,$2D,$C6,$27,$55,$E4,$CF
	.dc.b	$FA,$9F,$D2,$73,$86,$9E,$C8,$C3,$6F,$DA,$23,$87,$3F,$E3,$B4,$F4
	.dc.b	$B7,$3F,$F2,$AD,$FF,$41,$C3,$86,$1F,$27,$C7,$46,$0F,$8E,$E3,$1F
	.dc.b	$0B,$67,$F8,$00,$E0,$7D,$AF,$BA,$F9,$CE,$63,$11,$F7,$DB,$19,$8F
	.dc.b	$F3,$8B,$D3,$F1,$27,$C8,$7F,$76,$E1,$F9,$D2,$9C,$BC,$2B,$BC,$62
	.dc.b	$76,$D9,$FC,$9E,$9F,$91,$24,$FE,$02,$3C,$4F,$6C,$7D,$77,$D4,$F4
	.dc.b	$16,$76,$74,$9F,$C4,$C7,$E1,$EB,$AB,$92,$1F,$8F,$29,$FE,$A6,$F3
	.dc.b	$BE,$8F,$C1,$7F,$A3,$BD,$0F,$F5,$1C,$A3,$1E,$8A,$B7,$87,$F3,$1B
	.dc.b	$DD,$6A,$73,$7D,$4E,$55,$BB,$E5,$18,$35,$5F,$E9,$EE,$1F,$ED,$51
	.dc.b	$FD,$08,$87,$CA,$29,$FB,$C2,$9F,$A9,$C7,$FC,$57,$97,$DD,$8F,$99
	.dc.b	$B7,$E2,$1F,$C2,$E9,$F3,$9C,$9F,$88,$3A,$1D,$C8,$60,$FB,$FB,$7E
	.dc.b	$EA,$BF,$10,$D8,$CC,$73,$7F,$C7,$FB,$7E,$21,$EE,$48,$FC,$DD,$78
	.dc.b	$DB,$7E,$79,$FD,$2F,$0F,$EA,$71,$FD,$1C,$7F,$85,$CF,$D3,$94,$DF
	.dc.b	$07,$EB,$23,$F6,$BC,$5F,$E1,$F9,$F8,$E7,$E1,$CB,$5F,$C6,$A7,$F7
	.dc.b	$FC,$7F,$E5,$66,$E3,$D6,$BF,$82,$CE,$63,$F8,$67,$2F,$9E,$BF,$88
	.dc.b	$57,$C1,$6B,$FC,$5D,$B3,$F3,$23,$EC,$1E,$4D,$BF,$02,$FF,$0E,$F6
	.dc.b	$BC,$24,$9C,$1F,$23,$A3,$FC,$D3,$3F,$CE,$DB,$1F,$C1,$5B,$B7,$F3
	.dc.b	$E9,$DD,$F6,$7E,$E1,$E2,$3C,$95,$FC,$37,$A7,$89,$23,$F1,$B5,$FD
	.dc.b	$51,$FE,$BB,$D9,$FE,$7C,$F9,$5B,$66,$92,$7F,$C3,$1D,$FF,$DA,$DF
	.dc.b	$F0,$D0,$70,$EB,$C0,$FE,$E4,$EA,$32,$FA,$0F,$B4,$7C,$B2,$9F,$C8
	.dc.b	$BC,$3F,$F0,$53,$FA,$46,$BF,$E0,$EF,$0F,$E3,$C6,$23,$96,$9F,$15
	.dc.b	$F8,$87,$C6,$C7,$FC,$59,$E2,$70,$24,$75,$79,$F0,$C8,$F0,$35,$FF
	.dc.b	$0E,$F8,$BE,$02,$3E,$57,$ED,$BE,$A3,$C2,$43,$03,$CD,$FF,$32,$67
	.dc.b	$FE,$4C,$24,$7F,$9A,$5F,$F2,$A1,$1F,$9F,$E4,$FF,$A9,$23,$11,$FE
	.dc.b	$A4,$F2,$BF,$D6,$BC,$9F,$E6,$22,$47,$FA,$43,$34,$7F,$AE,$3C,$DF
	.dc.b	$F3,$94,$FF,$1C,$3F,$CE,$45,$3F,$CE,$45,$3F,$CE,$45,$C7,$FC,$F9
	.dc.b	$E0,$FF,$99,$09,$1F,$E8,$17,$FC,$C6,$47,$FA,$5B,$5F,$F3,$17,$9F
	.dc.b	$FE,$99,$8F,$F4,$BE,$1D,$E3,$13,$F8,$E6,$BE,$F8,$F2,$79,$BF,$EA
	.dc.b	$7A,$FF,$40,$92,$7F,$D7,$3D,$BF,$ED,$88,$FF,$59,$B6,$33,$5E,$DF
	.dc.b	$43,$FD,$77,$5F,$EC,$2F,$4B,$55,$FE,$C9,$6B,$FE,$C8,$6A,$F4,$BF
	.dc.b	$D9,$B3,$FE,$D4,$BF,$2F,$9B,$AC,$BF,$DB,$73,$93,$FD,$C7,$E9,$F2
	.dc.b	$E3,$FF,$74,$C6,$2F,$E8,$46,$31,$FD,$7E,$EE,$3F,$C7,$B6,$3F,$85
	.dc.b	$B5,$FB,$D1,$A2,$3A,$6B,$A3,$04,$30,$43,$C2,$F0,$69,$B2,$9B,$3F
	.dc.b	$D3,$D1,$E3,$72,$43,$31,$CD,$A7,$32,$9C,$CF,$F1,$7C,$73,$72,$73
	.dc.b	$33,$1C,$DA,$73,$29,$CC,$47,$37,$27,$33,$31,$CD,$A7,$32,$9C,$E3
	.dc.b	$93,$83,$93,$82,$47,$36,$C7,$36,$BC,$E3,$44,$74,$D7,$46,$08,$60
	.dc.b	$87,$89,$E0,$D3,$65,$36,$7C,$C2,$1F,$10,$A7,$B4,$B1,$E8,$79,$10
	.dc.b	$CC,$78,$1A,$78,$0A,$78,$0E,$C4,$3B,$16,$3B,$3C,$0E,$C6,$63,$B3
	.dc.b	$4E,$C5,$3B,$1D,$09,$1D,$1D,$1D,$0C,$C7,$46,$9D,$0A,$74,$8E,$4E
	.dc.b	$0E,$4E,$09,$1C,$DB,$1C,$DA,$F3,$8D,$11,$D3,$5D,$1F,$93,$F2,$7F
	.dc.b	$CC,$2F,$F7,$46,$BF,$E5,$9C,$3F,$C1,$F1,$88,$FD,$5F,$D5,$7F,$A6
	.dc.b	$BC,$AF,$E8,$8D,$8F,$E1,$ED,$7F,$76,$D7,$88,$8E,$2D,$38,$94,$E2
	.dc.b	$53,$8B,$C0,$8F,$06,$BC,$1A,$F0,$34,$43,$45,$34,$60,$86,$08,$7F
	.dc.b	$02,$8F,$63,$4F,$31,$63,$C2,$D3,$B1,$4E,$A5,$3A,$E0,$9E,$AD,$8E
	.dc.b	$AD,$9E,$AD,$8E,$AD,$7A,$B5,$EB,$3D,$08,$F4,$6C,$F4,$6C,$74,$6B
	.dc.b	$D1,$AF,$43,$91,$23,$93,$4E,$45,$39,$14,$E4,$F0,$23,$C1,$AF,$06
	.dc.b	$BC,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$4C,$93,$F2,$5B,$3F,$01,$A7
	.dc.b	$A4,$B3,$E4,$69,$E1,$29,$E0,$29,$E0,$9E,$C4,$9E,$CD,$8E,$CD,$9E
	.dc.b	$CD,$8E,$CD,$7B,$35,$ED,$3D,$08,$F4,$6C,$F4,$6C,$74,$6B,$D1,$AF
	.dc.b	$43,$91,$23,$93,$4F,$85,$8B,$36,$6F,$53,$FC,$34,$4A,$FF,$43,$D6
	.dc.b	$E7,$51,$FE,$8A,$D5,$EA,$72,$BF,$F4,$6C,$A7,$35,$C2,$04,$8C,$CF
	.dc.b	$06,$C7,$FA,$5B,$81,$C3,$D6,$30,$D7,$FC,$7D,$7D,$5F,$F1,$F3,$5F
	.dc.b	$17,$AD,$B7,$C0,$D7,$31,$BB,$4E,$57,$45,$24,$ED,$E5,$EB,$FB,$23
	.dc.b	$C0,$DC,$BE,$1B,$70,$FF,$58,$B5,$F8,$63,$B5,$E4,$BC,$7F,$54,$61
	.dc.b	$B1,$9B,$F9,$F3,$97,$E8,$FD,$B7,$FA,$19,$BB,$FD,$06,$E1,$F1,$67
	.dc.b	$4F,$FB,$13,$B1,$E7,$EE,$30,$D7,$CF,$1A,$7E,$73,$6B,$ED,$6B,$F8
	.dc.b	$CF,$4A,$EC,$69,$2F,$E1,$9B,$2C,$6D,$B1,$F2,$9D,$1F,$3F,$C8,$7D
	.dc.b	$8F,$E3,$3D,$7F,$F6,$ED,$A7,$34,$9C,$C6,$1B,$5C,$67,$DE,$6C,$B1
	.dc.b	$F9,$CC,$BE,$03,$73,$E1,$3F,$6B,$6F,$AD,$D1,$F5,$38,$3E,$D3,$78
	.dc.b	$7F,$AA,$CE,$AB,$F0,$3A,$79,$FE,$F3,$FB,$B3,$63,$FB,$83,$6F,$D3
	.dc.b	$3F,$DE,$FD,$77,$80,$FA,$3F,$DD,$D7,$FB,$FA,$73,$18,$C3,$F1,$F1
	.dc.b	$8C,$3F,$BE,$36,$BE,$2F,$2F,$8B,$87,$FD,$9F,$18,$BF,$1C,$61,$FC
	.dc.b	$C9,$F9,$9D,$7D,$EE,$34,$F7,$3F,$16,$32,$DD,$79,$E9,$DC,$91,$83
	.dc.b	$E3,$CF,$07,$F8,$4D,$74,$6C,$65,$F1,$57,$F8,$5F,$C6,$7E,$37,$E7
	.dc.b	$9F,$E2,$31,$8C,$FF,$C4,$8D,$8E,$13,$B5,$A7,$CE,$F8,$DF,$E4,$FE
	.dc.b	$4F,$CD,$D1,$18,$AF,$9B,$18,$DD,$5B,$AD,$57,$E6,$7D,$8F,$9C,$47
	.dc.b	$E6,$BF,$E1,$CD,$BD,$D1,$EC,$E0,$FC,$51,$1F,$C4,$F0,$7F,$16,$7E
	.dc.b	$AF,$47,$D9,$23,$F2,$B8,$25,$71,$8F,$A4,$D7,$E5,$C6,$9F,$B5,$C3
	.dc.b	$F4,$61,$FB,$0F,$F2,$69,$FF,$25,$6B,$FC,$6A,$73,$DD,$FE,$55,$8C
	.dc.b	$56,$F0,$FE,$9F,$7F,$D0,$AF,$6F,$23,$C6,$78,$1F,$13,$82,$1E,$68
	.dc.b	$F4,$3F,$14,$FD,$5F,$D7,$FE,$91,$7C,$FF,$81,$FE,$BF,$19,$8C,$60
	.dc.b	$8C,$3C,$B8,$7F,$AB,$CE,$B5,$7D,$2F,$B7,$0F,$E2,$4D,$7F,$B5,$C6
	.dc.b	$9F,$61,$88,$D5,$BE,$E3,$F7,$2E,$8C,$61,$FB,$A7,$F8,$2C,$6B,$2F
	.dc.b	$3B,$63,$E3,$47,$C2,$E0,$FC,$39,$FA,$58,$FE,$45,$7E,$0C,$7C,$95
	.dc.b	$EB,$8C,$1E,$9C,$7E,$51,$F3,$8A,$7C,$F3,$B9,$0D,$5F,$E5,$BC,$9F
	.dc.b	$E9,$3E,$57,$F7,$EF,$77,$FD,$15,$18,$A4,$63,$B5,$5B,$C7,$FD,$41
	.dc.b	$5A,$6D,$6E,$75,$3F,$E0,$AB,$E9,$39,$8E,$07,$E6,$8F,$E6,$04,$3F
	.dc.b	$18,$53,$F8,$91,$4E,$07,$EB,$31,$FE,$6A,$FD,$E3,$FB,$F1,$4F,$94
	.dc.b	$60,$E2,$74,$3B,$90,$FF,$03,$69,$F8,$07,$ED,$EB,$FA,$4D,$7E,$FB
	.dc.b	$63,$6D,$7F,$14,$77,$8F,$BA,$DD,$7F,$20,$F9,$0F,$CE,$EF,$FA,$39
	.dc.b	$B5,$FE,$17,$7F,$3D,$3F,$E4,$26,$9F,$A7,$7F,$95,$9F,$14,$A7,$AC
	.dc.b	$F7,$90,$C9,$F0,$4F,$DF,$70,$FF,$07,$7F,$65,$C3,$FC,$01,$B1,$99
	.dc.b	$FE,$0F,$C5,$F4,$27,$A7,$0F,$F8,$85,$F1,$35,$F1,$1E,$03,$7C,$7F
	.dc.b	$E2,$97,$FB,$FF,$27,$F9,$51,$FD,$AB,$6F,$FC,$0C,$DD,$3F,$95,$4F
	.dc.b	$6A,$E4,$F6,$BF,$03,$FC,$E4,$FF,$2F,$DF,$F8,$3F,$97,$91,$23,$FC
	.dc.b	$C7,$D1,$FD,$26,$5F,$35,$FD,$0D,$B3,$FD,$23,$A3,$FC,$E7,$39,$3F
	.dc.b	$B6,$F5,$FF,$55,$9B,$D7,$FE,$70,$7F,$CD,$F1,$97,$FC,$96,$D7,$E9
	.dc.b	$1F,$13,$83,$E1,$67,$EA,$8F,$E2,$FC,$3F,$04,$8F,$B3,$6F,$E5,$4F
	.dc.b	$C3,$AE,$3B,$7C,$48,$EB,$A7,$F8,$BF,$B3,$E5,$12,$39,$3F,$47,$6F
	.dc.b	$F0,$84,$7F,$69,$ED,$FC,$73,$FD,$BB,$2F,$F0,$D5,$FC,$86,$C7,$F8
	.dc.b	$80,$F2,$77,$7D,$72,$3E,$4C,$A3,$E7,$3F,$C6,$C3,$5A,$47,$FC,$7E
	.dc.b	$FF,$90,$9D,$1F,$DB,$3C,$1F,$F0,$3B,$4C,$9F,$E5,$17,$2F,$F9,$EB
	.dc.b	$C5,$FF,$41,$7A,$3F,$E5,$7C,$8D,$24,$FF,$AA,$79,$7F,$D6,$31,$FE
	.dc.b	$66,$24,$66,$78,$BF,$E5,$4E,$DF,$59,$1C,$1F,$E3,$2F,$1B,$F0,$91
	.dc.b	$FE,$7E,$24,$67,$83,$FD,$63,$1F,$E5,$32,$47,$F6,$F6,$BE,$B3,$FD
	.dc.b	$8A,$FA,$C8,$FF,$9D,$9F,$F3,$E9,$1F,$43,$5F,$F4,$EF,$8F,$FE,$AC
	.dc.b	$BF,$E7,$F1,$FE,$C9,$E9,$EA,$D8,$CB,$C2,$32,$FF,$AF,$AB,$FD,$9B
	.dc.b	$E4,$FF,$B1,$2B,$73,$FE,$C0,$9C,$F9,$3C,$23,$FD,$9D,$8D,$7F,$DA
	.dc.b	$8D,$C7,$FD,$AB,$5F,$ED,$18,$D7,$95,$FE,$DF,$8F,$F6,$DB,$6B,$FD
	.dc.b	$C7,$E5,$E8,$E9,$E6,$BC,$DF,$E1,$D2,$8C,$3F,$C4,$5A,$FF,$B9,$9A
	.dc.b	$FF,$9A,$9A,$FF,$B1,$1A,$FF,$BA,$1A,$F5,$8E,$F1,$82,$38,$24,$78
	.dc.b	$5A,$F8,$5A,$F8,$5A,$F8,$5A,$F8,$5A,$F8,$4E,$84,$3A,$14,$E8,$53
	.dc.b	$A1,$4E,$85,$3A,$1C,$88,$72,$29,$C8,$A7,$22,$9C,$8E,$04,$38,$14
	.dc.b	$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$48,$FA,$6F,$B5,$AF,$99,$AF
	.dc.b	$85,$AF,$66,$BD,$5A,$F5,$8F,$A7,$18,$23,$82,$47,$85,$AF,$85,$AF
	.dc.b	$85,$AF,$85,$AF,$85,$AF,$84,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$53
	.dc.b	$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A
	.dc.b	$21,$A2,$9A,$30,$43,$04,$8F,$A6,$FB,$5A,$FE,$5B,$C5,$FF,$32,$3F
	.dc.b	$EE,$A6,$BF,$E6,$23,$1E,$3F,$F1,$E4,$3E,$6B,$FD,$57,$E4,$7F,$BC
	.dc.b	$A3,$1B,$EA,$31,$3A,$9F,$F7,$BB,$5E,$67,$01,$1C,$27,$94,$70,$6B
	.dc.b	$C2,$79,$3E,$46,$BC,$E7,$93,$D0,$E4,$53,$94,$F4,$79,$35,$E4,$70
	.dc.b	$8E,$64,$70,$43,$99,$D4,$E6,$53,$99,$A7,$93,$4E,$47,$07,$2D,$32
	.dc.b	$72,$30,$71,$21,$82,$1B,$24,$FC,$62,$4F,$BD,$B1,$E7,$6C,$F8,$A3
	.dc.b	$8C,$EC,$8E,$DB,$1B,$9E,$33,$E8,$6B,$E6,$39,$9C,$49,$1C,$67,$9C
	.dc.b	$71,$6B,$C6,$79,$BE,$E6,$BD,$27,$93,$D0,$E4,$53,$94,$F4,$79,$35
	.dc.b	$E4,$70,$8E,$64,$70,$43,$99,$D4,$E6,$53,$99,$A7,$93,$4E,$47,$07
	.dc.b	$2D,$32,$72,$30,$71,$21,$82,$1B,$24,$FC,$62,$4F,$BD,$B1,$E7,$6C
	.dc.b	$F8,$A3,$8C,$EC,$8E,$DB,$1B,$9E,$33,$E8,$6B,$E6,$1F,$A1,$BE,$65
	.dc.b	$6F,$43,$F8,$64,$E6,$31,$5F,$C2,$0E,$35,$C2,$B6,$D8,$D7,$1B,$E3
	.dc.b	$B1,$1F,$C4,$23,$17,$D2,$31,$3A,$7F,$8F,$CE,$AF,$F8,$F7,$23,$B3
	.dc.b	$54,$8C,$61,$CE,$3F,$93,$C6,$2B,$8D,$70,$E5,$77,$52,$61,$C7,$04
	.dc.b	$EA,$31,$8F,$F9,$47,$0F,$F2,$7C,$63,$1F,$1C,$E6,$FE,$35,$BC,$7F
	.dc.b	$01,$2F,$B4,$65,$F0,$7A,$6F,$53,$DA,$D8,$CC,$60,$F8,$3E,$AF,$B7
	.dc.b	$0F,$A8,$D7,$F9,$EB,$5F,$E9,$2D,$5E,$3B,$C6,$F9,$9A,$FF,$4B,$8C
	.dc.b	$6F,$F0,$DB,$19,$9D,$D7,$E0,$FC,$1F,$73,$F2,$5B,$3F,$D5,$5B,$1B
	.dc.b	$D7,$FA,$DC,$FF,$59,$F8,$B7,$C6,$62,$BF,$1D,$18,$7E,$A3,$67,$84
	.dc.b	$EE,$7F,$CD,$1D,$F5,$FF,$34,$35,$FF,$3A,$37,$AF,$FD,$D5,$5F,$DA
	.dc.b	$75,$FC,$4F,$AD,$76,$FD,$1B,$78,$0D,$3E,$14,$63,$1F,$D2,$6F,$F2
	.dc.b	$27,$24,$8F,$AB,$3F,$AE,$D7,$A1,$23,$37,$F2,$9F,$EE,$B7,$F4,$A7
	.dc.b	$B9,$23,$BE,$FF,$A3,$8F,$1F,$54,$9C,$E1,$FD,$86,$FE,$DC,$FE,$E7
	.dc.b	$07,$DB,$BC,$6F,$E0,$3F,$51,$98,$C4,$A3,$13,$AC,$7F,$C1,$33,$CA
	.dc.b	$31,$1F,$E1,$68,$79,$8F,$AD,$FE,$FD,$18,$6C,$66,$7F,$BC,$4E,$5B
	.dc.b	$0F,$8A,$FF,$10,$F3,$7A,$63,$12,$8C,$4E,$A3,$CB,$3A,$9F,$89,$1F
	.dc.b	$33,$7F,$E1,$5A,$62,$3F,$C5,$BA,$FC,$67,$DB,$E9,$BE,$6B,$A4,$7B
	.dc.b	$63,$0F,$84,$DC,$69,$FE,$57,$1A,$C3,$B4,$7E,$C3,$0F,$0B,$F5,$B0
	.dc.b	$F1,$67,$F5,$39,$BE,$5C,$FF,$34,$E0,$46,$27,$47,$D5,$9F,$A7,$3A
	.dc.b	$8C,$47,$F9,$23,$D0,$FE,$3F,$DD,$FC,$FB,$F1,$CF,$E7,$51,$8A,$46
	.dc.b	$3C,$0F,$F7,$73,$6B,$85,$FF,$92,$2F,$F9,$4F,$77,$99,$B8,$78,$25
	.dc.b	$18,$E5,$E9,$5E,$C9,$F8,$5C,$DF,$95,$DB,$E1,$C7,$D5,$8C,$1F,$84
	.dc.b	$D3,$F9,$35,$FE,$57,$9F,$E1,$46,$2B,$E0,$C7,$C5,$EA,$7B,$DF,$D7
	.dc.b	$3E,$C7,$FA,$ED,$7E,$56,$73,$AA,$73,$3F,$D6,$EB,$84,$FE,$7A,$B7
	.dc.b	$5A,$AF,$4F,$17,$DC,$E6,$FC,$44,$7A,$63,$FB,$36,$8F,$61,$F2,$DF
	.dc.b	$AE,$CB,$FC,$E3,$39,$8C,$5F,$F9,$FA,$31,$3F,$B1,$37,$5C,$9C,$4F
	.dc.b	$EB,$FA,$5F,$67,$FE,$86,$FC,$2F,$D7,$5F,$B2,$8C,$3F,$A4,$6D,$71
	.dc.b	$6B,$FA,$2E,$2F,$A6,$DC,$7E,$9E,$7F,$96,$C5,$F5,$3F,$E9,$4F,$91
	.dc.b	$EF,$18,$B4,$62,$75,$97,$ED,$1B,$3F,$C1,$FE,$7F,$EA,$F6,$7A,$1B
	.dc.b	$5F,$C3,$9B,$3B,$8F,$E0,$67,$E4,$FA,$DF,$67,$F2,$A7,$F2,$31,$89
	.dc.b	$F8,$F1,$89,$D7,$57,$F8,$09,$E1,$1F,$70,$FD,$A1,$0F,$B8,$77,$F3
	.dc.b	$38,$DF,$F8,$97,$AB,$F1,$37,$FE,$0C,$9C,$9D,$39,$7F,$0B,$7F,$2F
	.dc.b	$7F,$C4,$70,$FC,$E8,$F0,$1D,$88,$6C,$FF,$61,$FA,$9F,$EC,$2F,$81
	.dc.b	$FF,$55,$7A,$4F,$49,$39,$C3,$FD,$7B,$39,$8C,$5F,$FB,$2A,$30,$AF
	.dc.b	$9D,$F2,$8C,$5F,$28,$C4,$EA,$F9,$D2,$F9,$CF,$FA,$F7,$1F,$F0,$4C
	.dc.b	$FF,$66,$21,$FD,$50,$A7,$EF,$CA,$7E,$3C,$FE,$25,$97,$F9,$53,$D2
	.dc.b	$F9,$BB,$7F,$94,$DB,$19,$E0,$FA,$87,$FB,$6E,$B9,$1D,$C8,$77,$3F
	.dc.b	$11,$C1,$C5,$FE,$4D,$7F,$87,$BE,$F3,$99,$FC,$45,$75,$1C,$3F,$C3
	.dc.b	$3E,$8F,$F3,$11,$FE,$F6,$6C,$9A,$7E,$8D,$7C,$5D,$24,$FE,$96,$32
	.dc.b	$FA,$1F,$59,$E9,$21,$FE,$78,$E7,$FE,$02,$BD,$FF,$D9,$F5,$23,$15
	.dc.b	$FC,$05,$A7,$F0,$12,$9F,$E6,$DE,$CF,$E1,$DA,$7F,$24,$EB,$FE,$26
	.dc.b	$D8,$CD,$7F,$14,$ED,$F8,$D7,$FC,$F0,$FB,$65,$3D,$51,$FE,$8D,$F0
	.dc.b	$3F,$D1,$FF,$63,$FC,$5B,$C1,$FE,$66,$D8,$CC,$FF,$31,$6D,$7F,$83
	.dc.b	$7C,$3F,$F3,$46,$5F,$D0,$E8,$D2,$4F,$C9,$3C,$47,$C8,$29,$FC,$1B
	.dc.b	$C4,$FC,$73,$FC,$2D,$AF,$E1,$1A,$EE,$7B,$BF,$C4,$7C,$5F,$F5,$4F
	.dc.b	$E3,$DF,$D9,$E4,$D2,$4F,$F7,$6F,$23,$9C,$FF,$7A,$C3,$FC,$61,$18
	.dc.b	$C7,$FB,$D4,$7C,$AE,$9F,$D5,$DF,$AD,$AF,$F8,$4F,$CA,$FC,$83,$FE
	.dc.b	$12,$3F,$20,$5E,$6F,$F2,$07,$95,$FE,$18,$7A,$78,$3F,$EA,$B9,$BC
	.dc.b	$9F,$EB,$D7,$F1,$9E,$5F,$47,$FA,$73,$5C,$9D,$8C,$94,$FC,$A7,$99
	.dc.b	$CD,$A7,$5D,$BF,$C7,$CF,$F8,$EE,$35,$1F,$E4,$B3,$E6,$C7,$DF,$F3
	.dc.b	$7E,$21,$23,$31,$FA,$42,$1F,$70,$9A,$7F,$A6,$B0,$FF,$3B,$92,$3F
	.dc.b	$CC,$AF,$F9,$F4,$8F,$F9,$C9,$AF,$F8,$4F,$DD,$FF,$3F,$91,$FF,$36
	.dc.b	$BF,$E8,$52,$3F,$E7,$A6,$BF,$E9,$43,$F9,$61,$7C,$3F,$9D,$E8,$FF
	.dc.b	$8F,$23,$27,$FA,$60,$A3,$D3,$FF,$51,$4F,$FA,$82,$33,$18,$7F,$D4
	.dc.b	$FE,$9F,$EF,$E7,$FD,$61,$E8,$59,$FF,$58,$1D,$3C,$CF,$F5,$BC,$EA
	.dc.b	$3B,$E7,$FE,$C0,$8F,$F4,$E9,$23,$FD,$16,$FF,$A8,$C8,$FF,$A6,$1A
	.dc.b	$FF,$0F,$1D,$3F,$ED,$98,$FC,$63,$63,$46,$BD,$46,$DF,$02,$75,$3C
	.dc.b	$FD,$5F,$F7,$35,$FE,$55,$B1,$FE,$EE,$CB,$8C,$E5,$B1,$DB,$D6,$76
	.dc.b	$62,$3C,$D1,$8C,$7E,$7B,$6F,$FD,$FA,$77,$F2,$38,$57,$F2,$BC,$51
	.dc.b	$88,$46,$2F,$EC,$B6,$BF,$59,$83,$8B,$83,$6B,$EF,$B6,$77,$3D,$71
	.dc.b	$7A,$F4,$97,$E8,$7C,$5E,$C3,$CB,$A4,$9F,$C5,$C6,$5F,$C6,$F6,$FF
	.dc.b	$96,$F4,$FB,$0D,$7F,$A4,$35,$FE,$D8,$7D,$87,$04,$8F,$DF,$46,$8E
	.dc.b	$6E,$23,$51,$FC,$0D,$AF,$F0,$F6,$BD,$63,$81,$1E,$0D,$78,$35,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$10,$F4,$9C,$0A,$6C,$A6,$CF,$8E,$F8,$DC
	.dc.b	$90,$C9,$CC,$A7,$32,$9C,$CF,$94,$60,$E4,$60,$91,$F2,$1C,$B5,$E0
	.dc.b	$43,$71,$E5,$71,$1B,$72,$79,$4A,$79,$4A,$79,$5E,$04,$8E,$91,$A3
	.dc.b	$C6,$E2,$35,$1D,$5A,$F5,$6B,$D6,$38,$11,$E0,$D7,$83,$5E,$06,$88
	.dc.b	$68,$A6,$8C,$10,$C1,$0F,$49,$C0,$A6,$CA,$6C,$F9,$6F,$8D,$C9,$0C
	.dc.b	$9C,$CA,$73,$29,$CC,$F9,$46,$0E,$46,$09,$1F,$21,$CB,$F6,$FC,$DF
	.dc.b	$F3,$EC,$7F,$A2,$0F,$EF,$8E,$1F,$F3,$E6,$7F,$B3,$FB,$9F,$F2,$37
	.dc.b	$A1,$FB,$36,$CF,$EF,$DA,$FF,$06,$6B,$FC,$19,$AF,$F8,$C9,$B6,$69
	.dc.b	$21,$3D,$DA,$77,$29,$DC,$A7,$72,$9D,$CB,$3D,$DB,$3D,$E7,$B1,$23
	.dc.b	$B3,$5E,$CD,$7B,$35,$EC,$D9,$EC,$D9,$ED,$3D,$08,$74,$29,$D0,$A7
	.dc.b	$42,$CF,$46,$C7,$47,$91,$0E,$45,$39,$16,$39,$36,$39,$3C,$08,$70
	.dc.b	$2C,$70,$6C,$70,$74,$48,$D3,$63,$51,$82,$38,$24,$7D,$42,$4F,$C9
	.dc.b	$69,$F0,$0A,$7A,$4A,$79,$0A,$78,$4B,$3E,$06,$CF,$82,$7B,$12,$3B
	.dc.b	$35,$EC,$D7,$B3,$5E,$CD,$9E,$CD,$9E,$D3,$D0,$87,$42,$9D,$0A,$74
	.dc.b	$2C,$F4,$6C,$74,$79,$10,$E4,$53,$91,$63,$93,$63,$93,$C0,$87,$02
	.dc.b	$C7,$07,$FD,$91,$EE,$3E,$D3,$FC,$6B,$E4,$7F,$A1,$63,$19,$A3,$13
	.dc.b	$AA,$FF,$1E,$D6,$A7,$FD,$19,$BF,$F0,$0B,$E1,$5F,$E3,$5C,$B8,$3F
	.dc.b	$E3,$46,$BF,$ED,$99,$E1,$5E,$1F,$77,$32,$C1,$18,$6E,$1D,$A7,$FC
	.dc.b	$A9,$18,$AF,$57,$D4,$6C,$95,$BA,$F8,$31,$DB,$0F,$F2,$35,$7A,$B8
	.dc.b	$F0,$4B,$E1,$7F,$E4,$2F,$7F,$CB,$CA,$69,$27,$7D,$0E,$9F,$5B,$63
	.dc.b	$31,$83,$9F,$C0,$C8,$EF,$68,$C6,$5F,$EB,$9E,$B7,$89,$FE,$4F,$8C
	.dc.b	$37,$3F,$F3,$8C,$EB,$AB,$D2,$DA,$FF,$61,$7C,$0F,$FA,$55,$BE,$52
	.dc.b	$FD,$F3,$97,$FC,$F6,$D7,$FD,$05,$E0,$BC,$1E,$17,$FE,$99,$24,$FF
	.dc.b	$A6,$0F,$93,$E0,$BE,$4F,$71,$FE,$86,$8C,$3A,$8C,$BF,$E8,$F8,$CC
	.dc.b	$FF,$B6,$7E,$35,$F4,$1A,$AF,$44,$61,$FA,$ED,$9E,$13,$B9,$69,$F1
	.dc.b	$7E,$0B,$E4,$EC,$69,$27,$E0,$ED,$F1,$63,$E3,$C6,$CC,$94,$FF,$53
	.dc.b	$6D,$FE,$C1,$8E,$BE,$2C,$9F,$AB,$C1,$FE,$87,$8C,$47,$0B,$FA,$51
	.dc.b	$D8,$91,$D9,$F3,$F6,$48,$E7,$8F,$F6,$8C,$BF,$01,$3F,$EB,$2F,$73
	.dc.b	$FD,$37,$D7,$FE,$B2,$F7,$71,$E3,$A3,$18,$FF,$AF,$6B,$7D,$0B,$DD
	.dc.b	$6A,$BF,$0D,$2F,$69,$F6,$1F,$E0,$88,$C3,$63,$33,$FD,$FA,$72,$D8
	.dc.b	$FC,$DA,$E5,$FF,$64,$CF,$5E,$0F,$CA,$46,$2B,$6D,$8C,$D7,$C2,$D3
	.dc.b	$E2,$3F,$ED,$7D,$7E,$1B,$93,$FD,$B5,$87,$A7,$C2,$E3,$1E,$CC,$7F
	.dc.b	$92,$3E,$33,$6E,$31,$F1,$4F,$09,$FD,$45,$F6,$7C,$18,$FF,$B0,$F1
	.dc.b	$F1,$C6,$A3,$0F,$CF,$FD,$9B,$FD,$8D,$AF,$D7,$6C,$67,$83,$E8,$2D
	.dc.b	$D3,$A8,$F4,$7C,$57,$EE,$7A,$FE,$33,$63,$D2,$D7,$62,$3C,$1D,$7D
	.dc.b	$AF,$AB,$6F,$F7,$DC,$1F,$83,$BF,$44,$F9,$1F,$D5,$E7,$EC,$7E,$0E
	.dc.b	$5F,$93,$8C,$47,$C1,$8F,$F1,$FE,$1E,$AF,$A9,$EC,$4C,$3B,$57,$F1
	.dc.b	$39,$D4,$62,$0B,$D5,$7C,$76,$3E,$13,$35,$F9,$69,$CD,$27,$33,$CA
	.dc.b	$73,$18,$9E,$95,$CD,$FC,$45,$7B,$38,$3E,$DB,$F0,$32,$FF,$2C,$8F
	.dc.b	$DD,$1F,$07,$FA,$FB,$FE,$59,$4A,$D4,$E6,$3F,$98,$46,$27,$F6,$07
	.dc.b	$07,$1C,$5E,$5E,$67,$A3,$FC,$DF,$EC,$F0,$DB,$FC,$BB,$70,$FC,$E3
	.dc.b	$5F,$A5,$9F,$AB,$9B,$F3,$11,$83,$F3,$FF,$55,$FD,$0F,$F2,$1F,$03
	.dc.b	$70,$CC,$A7,$2D,$8F,$D3,$FC,$D7,$EE,$1B,$1F,$BD,$6F,$5F,$AE,$75
	.dc.b	$3F,$7F,$A1,$F4,$BF,$16,$3F,$41,$18,$BF,$8B,$1B,$9D,$73,$7E,$91
	.dc.b	$E1,$1F,$57,$4F,$77,$47,$E4,$A3,$07,$EC,$7E,$C3,$8E,$26,$93,$1E
	.dc.b	$F3,$F1,$4F,$C1,$1F,$84,$30,$43,$A9,$A2,$1F,$DB,$7D,$57,$C5,$3D
	.dc.b	$6F,$EE,$1C,$FF,$EA,$49,$CC,$63,$3F,$F5,$DD,$FE,$31,$B5,$BB,$E5
	.dc.b	$39,$3B,$38,$8D,$57,$39,$CB,$5F,$CD,$1F,$D8,$88,$7F,$4C,$29,$FB
	.dc.b	$C2,$9F,$BB,$3F,$8C,$63,$FB,$4F,$57,$F5,$BB,$7F,$16,$8C,$67,$F3
	.dc.b	$C4,$EB,$4F,$A1,$18,$3C,$17,$0F,$01,$4F,$C1,$ED,$D1,$FC,$14,$71
	.dc.b	$6D,$75,$9D,$47,$F0,$8E,$67,$07,$F8,$7F,$DE,$FC,$94,$9A,$4C,$3F
	.dc.b	$3F,$F9,$17,$E7,$88,$E0,$F2,$4F,$CA,$2F,$37,$F8,$C5,$FE,$B1,$CF
	.dc.b	$FE,$2D,$FB,$5F,$DB,$C1,$A4,$AF,$E0,$1D,$9F,$C2,$9F,$64,$FF,$09
	.dc.b	$8C,$6B,$FC,$6B,$B3,$F8,$43,$63,$35,$E2,$E7,$FE,$21,$7F,$60,$F3
	.dc.b	$1F,$C9,$7B,$7F,$95,$FC,$5F,$F2,$1F,$03,$F9,$3B,$63,$37,$FD,$A2
	.dc.b	$72,$FF,$59,$F0,$7F,$01,$76,$3E,$EB,$72,$FE,$B9,$3A,$8F,$B6,$43
	.dc.b	$F1,$47,$DD,$E6,$FE,$A3,$5C,$A3,$F8,$27,$85,$C5,$FE,$06,$D9,$DD
	.dc.b	$FF,$60,$97,$E0,$74,$8F,$F6,$23,$27,$F6,$DB,$EC,$FF,$0E,$DB,$A7
	.dc.b	$3F,$F3,$79,$CE,$1F,$DC,$27,$C5,$5F,$28,$F4,$9E,$B3,$13,$FE,$0D
	.dc.b	$7F,$18,$7F,$82,$BC,$6F,$C8,$34,$C9,$FE,$11,$72,$F3,$E3,$FC,$53
	.dc.b	$F6,$EB,$FA,$5B,$FE,$17,$E0,$DB,$F9,$D6,$DF,$F8,$A6,$3A,$99,$2B
	.dc.b	$FE,$32,$7F,$C5,$9E,$3F,$47,$FC,$44,$D7,$F5,$0D,$C7,$FC,$8B,$A7
	.dc.b	$F9,$12,$BE,$2F,$93,$FE,$4B,$6F,$93,$F1,$B3,$F4,$E3,$EB,$CB,$F2
	.dc.b	$6F,$EA,$3C,$D7,$99,$FD,$8A,$3F,$CD,$7A,$FF,$9B,$88,$FF,$97,$9F
	.dc.b	$F3,$B9,$1C,$1F,$E7,$2F,$7F,$FC,$EA,$47,$07,$F9,$99,$FF,$3F,$91
	.dc.b	$C1,$F3,$3D,$1F,$F4,$57,$83,$FE,$88,$8F,$E5,$A7,$F7,$A8,$CB,$BF
	.dc.b	$43,$FD,$51,$3F,$27,$C7,$FF,$4D,$B7,$D2,$CD,$7F,$A7,$DA,$FF,$A7
	.dc.b	$5A,$FF,$8E,$79,$EB,$FE,$35,$6F,$9D,$FE,$AF,$8F,$19,$B2,$8D,$7E
	.dc.b	$5B,$86,$DF,$FA,$60,$91,$FE,$88,$7F,$D4,$24,$70,$7E,$3B,$D4,$FF
	.dc.b	$69,$71,$76,$6C,$FE,$1D,$B3,$BF,$53,$FC,$6B,$3F,$8B,$F5,$BF,$DC
	.dc.b	$0D,$F5,$B3,$5F,$9A,$9C,$B6,$32,$FF,$8A,$2F,$FD,$DD,$87,$87,$D7
	.dc.b	$CC,$FF,$B5,$E7,$38,$78,$E7,$32,$F6,$5E,$CE,$1B,$87,$C8,$6D,$FF
	.dc.b	$C0,$18,$F9,$7E,$D7,$E6,$E3,$F4,$8D,$BF,$D8,$B5,$FF,$1E,$B5,$FF
	.dc.b	$75,$2F,$6F,$C6,$DA,$EB,$18,$3C,$9E,$8C,$8F,$23,$6F,$C3,$E0,$72
	.dc.b	$24,$F9,$9F,$7F,$73,$DC,$FD,$C6,$5F,$DB,$9B,$19,$8F,$DE,$35,$FE
	.dc.b	$00,$D7,$F8,$6B,$5F,$F2,$CB,$51,$D4,$E7,$1D,$88,$E0,$87,$67,$2D
	.dc.b	$32,$78,$1C,$B4,$C9,$E2,$72,$D3,$27,$92,$3C,$CF,$94,$EC,$43,$A1
	.dc.b	$4E,$45,$39,$47,$66,$9D,$8E,$87,$22,$1C,$8A,$70,$3A,$1C,$0D,$10
	.dc.b	$D1,$C0,$D1,$83,$46,$09,$1E,$E2,$47,$9E,$32,$D8,$CC,$7A,$5A,$F9
	.dc.b	$1A,$F8,$5A,$F8,$1A,$F8,$C8,$76,$3A,$47,$72,$38,$21,$DD,$CB,$4C
	.dc.b	$9E,$17,$2D,$32,$78,$DC,$B4,$C9,$E4,$8F,$3B,$E5,$3B,$10,$E8,$53
	.dc.b	$91,$4E,$51,$D9,$A7,$63,$A1,$C8,$87,$22,$9C,$0E,$87,$03,$44,$34
	.dc.b	$70,$34,$60,$D1,$82,$47,$B8,$F7,$FA,$7F,$E8,$88,$FE,$F4,$D7,$FC
	.dc.b	$0E,$D7,$FD,$16,$7E,$FA,$BF,$77,$8F,$F0,$26,$C6,$44,$65,$A6,$5C
	.dc.b	$11,$C1,$0F,$E1,$31,$D1,$A7,$22,$C7,$16,$9C,$4F,$DE,$47,$16,$9C
	.dc.b	$4B,$1C,$5A,$71,$3F,$D1,$F1,$C5,$A7,$12,$C7,$16,$9C,$69,$3C,$5B
	.dc.b	$1C,$5B,$3C,$5B,$1C,$67,$81,$1E,$0D,$9E,$0D,$8E,$06,$89,$1A,$69
	.dc.b	$A7,$04,$70,$43,$E5,$90,$F8,$A5,$3D,$E5,$9F,$4B,$67,$C8,$D3,$C2
	.dc.b	$59,$F0,$34,$F0,$1D,$88,$76,$2C,$F6,$6C,$F6,$69,$D8,$B3,$D9,$A7
	.dc.b	$63,$A1,$27,$A3,$67,$A3,$4E,$85,$9E,$8D,$3A,$4F,$22,$4F,$26,$C7
	.dc.b	$26,$CF,$26,$C7,$29,$E0,$47,$83,$67,$83,$63,$81,$A2,$46,$9A,$69
	.dc.b	$FC,$C7,$D4,$3E,$2D,$93,$DE,$FF,$3E,$4E,$63,$17,$FE,$8A,$8C,$4E
	.dc.b	$AB,$8B,$FE,$83,$DF,$FC,$30,$DD,$B3,$7F,$E9,$29,$CB,$63,$2F,$F8
	.dc.b	$93,$2A,$FF,$89,$1A,$B7,$FD,$6E,$5E,$48,$E0,$D7,$FC,$1F,$7F,$C4
	.dc.b	$A3,$11,$FE,$3D,$CF,$9C,$E7,$1F,$44,$76,$BF,$2D,$7F,$8D,$B1,$C1
	.dc.b	$30,$E7,$3F,$E9,$9B,$FF,$1F,$F8,$5C,$49,$3E,$0C,$7F,$C9,$33,$FE
	.dc.b	$47,$26,$1F,$0F,$87,$DB,$F9,$3B,$E5,$69,$30,$F4,$C6,$5F,$5F,$3B
	.dc.b	$9F,$E2,$B6,$33,$18,$3E,$4F,$3F,$F9,$D3,$2F,$F3,$97,$46,$09,$5F
	.dc.b	$1B,$E5,$7F,$DB,$53,$FE,$BD,$6E,$9F,$EC,$A6,$BF,$E3,$C0,$E1,$FF
	.dc.b	$1F,$F8,$AF,$15,$E2,$7D,$28,$C3,$63,$37,$FE,$CB,$9C,$BF,$EC,$DE
	.dc.b	$EF,$F4,$D3,$7B,$FF,$C9,$6D,$C3,$EE,$CE,$9F,$F5,$41,$F8,$0F,$05
	.dc.b	$E0,$F0,$8C,$D7,$FA,$BD,$F9,$BC,$5F,$A2,$F8,$1F,$F6,$E7,$1F,$CF
	.dc.b	$6C,$66,$7F,$DD,$73,$97,$FD,$25,$BA,$BF,$D3,$0D,$8C,$C6,$0F,$A3
	.dc.b	$19,$8F,$F4,$C7,$07,$56,$C6,$5F,$F4,$C4,$65,$FF,$5D,$4F,$C1,$23
	.dc.b	$FE,$FE,$CF,$FD,$81,$1D,$BE,$76,$4F,$FA,$BE,$D1,$88,$FF,$3A,$C3
	.dc.b	$EB,$1F,$1B,$FC,$13,$D0,$FB,$D5,$FC,$1A,$73,$49,$CB,$C6,$B8,$57
	.dc.b	$E1,$63,$06,$DF,$F7,$47,$D5,$7F,$A1,$2B,$AC,$3C,$94,$E6,$36,$71
	.dc.b	$3F,$DD,$8E,$5F,$C9,$71,$FE,$33,$BF,$F1,$5C,$BF,$90,$6F,$37,$F7
	.dc.b	$99,$D5,$79,$7E,$C9,$F1,$1F,$E9,$6E,$7F,$5C,$62,$51,$89,$D4,$71
	.dc.b	$6B,$F2,$BD,$C7,$CE,$FF,$84,$77,$46,$27,$FC,$25,$18,$9D,$4E,$73
	.dc.b	$F9,$FE,$8B,$D3,$7D,$AF,$F3,$6F,$86,$CC,$63,$14,$62,$75,$97,$ED
	.dc.b	$6F,$94,$62,$3E,$3F,$07,$C5,$7E,$EE,$FF,$11,$CF,$07,$C7,$9D,$47
	.dc.b	$FA,$EA,$DE,$A3,$ED,$3A,$4A,$73,$18,$6E,$7E,$69,$F2,$57,$18,$C4
	.dc.b	$78,$35,$FA,$E4,$7C,$B8,$78,$9F,$EA,$FF,$48,$F6,$7E,$8C,$61,$B6
	.dc.b	$9D,$4F,$DA,$3F,$28,$E3,$8F,$F3,$3C,$2F,$1B,$FD,$AF,$F8,$FE,$1B
	.dc.b	$7F,$9B,$9D,$29,$D3,$4D,$7C,$D7,$8A,$4D,$24,$78,$CF,$80,$7B,$B8
	.dc.b	$3E,$81,$35,$FD,$6E,$5F,$82,$9E,$07,$C2,$E2,$F7,$C6,$73,$DC,$78
	.dc.b	$79,$BF,$DE,$3F,$15,$FD,$FA,$FF,$2D,$18,$DB,$D8,$D8,$C8,$9E,$0F
	.dc.b	$F3,$CE,$0F,$A4,$B9,$3F,$10,$E7,$4F,$43,$FD,$15,$7B,$5F,$E8,$2C
	.dc.b	$F9,$C6,$35,$FD,$86,$7B,$AD,$36,$33,$7F,$91,$E7,$7E,$21,$FE,$18
	.dc.b	$FC,$73,$CB,$CF,$FA,$78,$C6,$1F,$A1,$AD,$AA,$DC,$61,$F6,$DF,$AF
	.dc.b	$5F,$6E,$BE,$C8,$C6,$9E,$C7,$C5,$F6,$5F,$E2,$8F,$C4,$B9,$B7,$2F
	.dc.b	$03,$67,$70,$9F,$0F,$77,$EE,$E3,$CC,$D9,$F7,$35,$FD,$66,$7F,$C9
	.dc.b	$27,$F7,$E7,$F2,$DF,$C8,$DE,$67,$C6,$AF,$F1,$DC,$61,$FE,$33,$3B
	.dc.b	$9D,$29,$D7,$17,$C2,$D3,$E0,$13,$5F,$86,$7E,$C6,$B8,$75,$F2,$BF
	.dc.b	$D5,$47,$8E,$30,$78,$F1,$F9,$3C,$5F,$BB,$7F,$3E,$43,$F1,$0E,$5A
	.dc.b	$7E,$8D,$FF,$6A,$FC,$6B,$DC,$4B,$E3,$7F,$D8,$9E,$E7,$FB,$6D,$B5
	.dc.b	$B9,$D5,$FF,$85,$EB,$4D,$9D,$57,$0B,$EA,$DB,$E2,$78,$2F,$8B,$F9
	.dc.b	$F3,$FA,$69,$0F,$DD,$14,$FE,$56,$53,$81,$F1,$C7,$C5,$7E,$BB,$2F
	.dc.b	$9B,$18,$75,$19,$CF,$F8,$DB,$F5,$F4,$FF,$16,$D7,$53,$F2,$E7,$7A
	.dc.b	$FB,$DB,$FC,$8A,$E4,$D8,$FC,$24,$F7,$3F,$C0,$77,$F8,$9A,$FC,$BC
	.dc.b	$BD,$DA,$47,$F3,$26,$4F,$D0,$8F,$C9,$3F,$4A,$D3,$2E,$0F,$57,$37
	.dc.b	$F8,$0B,$0F,$49,$F2,$88,$7F,$01,$E8,$FF,$4B,$FE,$4D,$FB,$A6,$DF
	.dc.b	$EF,$DB,$3B,$84,$FF,$10,$EB,$71,$7C,$1A,$FC,$5E,$1F,$C4,$DF,$84
	.dc.b	$7F,$96,$27,$6E,$23,$59,$FA,$23,$E8,$72,$7F,$97,$1F,$A1,$E0,$FF
	.dc.b	$82,$7E,$2B,$EE,$37,$3F,$E5,$F2,$9C,$F7,$FE,$71,$FC,$17,$83,$FC
	.dc.b	$FF,$1F,$E7,$75,$F8,$63,$FA,$CF,$83,$F6,$DA,$64,$FC,$79,$F2,$0F
	.dc.b	$C7,$98,$3F,$C2,$DE,$17,$16,$9C,$4F,$F6,$1E,$5F,$D6,$1C,$65,$FE
	.dc.b	$1A,$9E,$11,$AF,$13,$F6,$7F,$93,$FF,$62,$24,$7F,$67,$8C,$BF,$DC
	.dc.b	$3B,$38,$DF,$F7,$56,$E7,$F9,$C9,$D1,$E3,$3F,$06,$7A,$4F,$59,$3C
	.dc.b	$3F,$F0,$6F,$37,$F8,$A3,$C6,$FF,$06,$37,$C7,$FF,$6E,$B7,$1F,$CD
	.dc.b	$5F,$C5,$3A,$9D,$0B,$7F,$E2,$88,$DC,$7F,$51,$DB,$9B,$FD,$35,$AE
	.dc.b	$63,$6F,$CA,$F2,$7F,$39,$1F,$E3,$F8,$FF,$2F,$B5,$FF,$08,$C6,$BC
	.dc.b	$AF,$F2,$87,$0F,$F9,$3E,$7E,$8C,$61,$FF,$28,$1F,$17,$93,$F2,$0F
	.dc.b	$AD,$AF,$C7,$3F,$CD,$9E,$27,$F2,$8F,$B9,$FF,$37,$F9,$9F,$E6,$E9
	.dc.b	$46,$27,$51,$FE,$70,$8F,$F3,$EF,$9F,$FE,$80,$6F,$9E,$D3,$8D,$7F
	.dc.b	$9F,$EB,$4D,$9D,$57,$FA,$83,$EA,$24,$FF,$A4,$DA,$FF,$B2,$FD,$3F
	.dc.b	$EB,$1F,$11,$FE,$BD,$24,$7F,$A7,$7C,$4F,$E8,$31,$FE,$AA,$24,$7F
	.dc.b	$A7,$5A,$FF,$B5,$7A,$3F,$DC,$91,$FE,$B4,$24,$7F,$AE,$1F,$F5,$EE
	.dc.b	$66,$92,$3F,$D7,$FE,$BF,$BE,$3F,$D9,$1E,$4C,$9F,$F6,$7B,$5F,$56
	.dc.b	$1F,$D5,$A3,$FD,$AC,$48,$F5,$1F,$D5,$4F,$F3,$7B,$FE,$96,$23,$FE
	.dc.b	$43,$7F,$DE,$5E,$07,$CF,$6E,$1C,$67,$FB,$1F,$B5,$FE,$EE,$8C,$47
	.dc.b	$FB,$D2,$B8,$3F,$EC,$1A,$FF,$68,$91,$C1,$FE,$DA,$7F,$DB,$24,$70
	.dc.b	$7C,$29,$EB,$EC,$32,$EB,$3F,$DD,$04,$66,$77,$3A,$9F,$EF,$3E,$F3
	.dc.b	$DE,$C4,$23,$15,$EA,$6E,$1F,$EC,$66,$CF,$FB,$53,$E0,$1F,$01,$C9
	.dc.b	$B7,$F0,$9B,$7F,$EC,$E6,$D7,$1F,$81,$E0,$4A,$E0,$F9,$30,$FE,$8F
	.dc.b	$1E,$3F,$27,$04,$7B,$7C,$17,$84,$8E,$0F,$55,$73,$23,$83,$EB,$FC
	.dc.b	$23,$E1,$3E,$1B,$63,$31,$83,$E3,$F8,$4B,$FE,$D1,$97,$F7,$56,$C6
	.dc.b	$63,$FB,$63,$5F,$F7,$7B,$5F,$F7,$EB,$5F,$B4,$D7,$FC,$5E,$D7,$B1
	.dc.b	$D2,$3B,$91,$C1,$0E,$EE,$5A,$64,$F0,$B9,$69,$93,$C6,$E5,$A6,$4F
	.dc.b	$2B,$96,$99,$3C,$E7,$80,$87,$62,$9D,$0A,$72,$29,$C8,$E0,$43,$81
	.dc.b	$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$8F,$A4,$48,$F8,$11,$96
	.dc.b	$C6,$63,$E1,$35,$F5,$B5,$F3,$35,$F1,$B5,$F1,$35,$EC,$74,$8E,$E4
	.dc.b	$70,$43,$BB,$96,$99,$3C,$2E,$5A,$64,$F1,$B9,$69,$93,$CA,$E5,$A6
	.dc.b	$4F,$39,$E0,$21,$D8,$A7,$42,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E
	.dc.b	$06,$8F,$F4,$86,$9F,$E5,$FF,$49,$EA,$7F,$38,$94,$62,$75,$39,$B6
	.dc.b	$7F,$EA,$2D,$3F,$76,$DC,$3F,$76,$D7,$F7,$6D,$7F,$76,$D7,$FC,$F8
	.dc.b	$D9,$EB,$1F,$E6,$03,$89,$4E,$25,$38,$95,$E3,$3E,$48,$C4,$F5,$23
	.dc.b	$82,$46,$C8,$FA,$8C,$95,$CB,$E4,$6B,$E4,$6B,$E4,$6C,$F9,$23,$91
	.dc.b	$0E,$45,$39,$14,$E4,$57,$93,$C0,$87,$02,$9C,$0A,$F0,$74,$43,$45
	.dc.b	$74,$E0,$8E,$09,$3F,$2A,$7D,$AD,$9F,$33,$5F,$0B,$5E,$CD,$7A,$B6
	.dc.b	$7A,$C7,$D0,$38,$94,$E2,$53,$89,$5E,$33,$E4,$8C,$4F,$52,$38,$24
	.dc.b	$6C,$8F,$A8,$C9,$5C,$BE,$46,$BE,$46,$BE,$46,$CF,$92,$39,$10,$E4
	.dc.b	$53,$91,$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF,$04,$3E,$0D,$F5,$3F
	.dc.b	$E6,$BF,$83,$FF,$15,$4E,$A7,$3B,$27,$33,$C6,$B8,$3C,$2B,$73,$A5
	.dc.b	$F8,$6E,$A7,$FC,$6D,$18,$BE,$91,$C6,$74,$F8,$5A,$FF,$8D,$7E,$62
	.dc.b	$F1,$1A,$4A,$E7,$3F,$E4,$3F,$CC,$DF,$1B,$C9,$B3,$A9,$F8,$71,$8C
	.dc.b	$BD,$AD,$8F,$87,$83,$E4,$BC,$24,$BE,$53,$FE,$49,$8F,$3E,$6F,$5A
	.dc.b	$92,$B7,$19,$6F,$9F,$C0,$8E,$0D,$FC,$67,$42,$38,$39,$FC,$D7,$80
	.dc.b	$8E,$0E,$FF,$15,$E4,$23,$83,$D3,$F0,$1F,$84,$6E,$7F,$42,$31,$B7
	.dc.b	$E0,$1B,$1F,$86,$F3,$5E,$FB,$C3,$F8,$71,$86,$C6,$6F,$F1,$33,$97
	.dc.b	$F1,$7F,$1D,$F1,$EF,$E5,$B6,$33,$BF,$E6,$1B,$3E,$8F,$8C,$C3,$F9
	.dc.b	$8C,$91,$88,$FF,$42,$46,$1E,$07,$D8,$F0,$1F,$61,$B8,$FD,$2C,$61
	.dc.b	$D4,$65,$46,$6F,$F3,$DB,$FF,$B4,$FE,$36,$EE,$69,$2F,$F4,$DC,$1B
	.dc.b	$6C,$ED,$B3,$FA,$99,$D4,$7A,$8A,$7E,$EB,$2F,$B1,$3F,$56,$32,$23
	.dc.b	$2F,$EF,$3E,$13,$FD,$6D,$F1,$BF,$BA,$6C,$67,$8B,$FD,$D9,$39,$6C
	.dc.b	$7C,$D2,$9E,$3C,$53,$E1,$6C,$6D,$AF,$FA,$D7,$E2,$BF,$0E,$BE,$37
	.dc.b	$F8,$5C,$61,$BC,$9F,$C5,$9B,$3B,$3F,$8A,$FA,$DF,$E1,$87,$07,$FB
	.dc.b	$05,$C4,$7F,$16,$6C,$70,$F0,$FA,$3F,$99,$AF,$CC,$78,$5F,$94,$6C
	.dc.b	$67,$3F,$F0,$77,$3F,$E6,$EB,$F9,$2F,$B6,$F8,$8F,$F1,$4E,$AF,$BA
	.dc.b	$CC,$61,$B9,$A7,$59,$7F,$B5,$EF,$94,$E6,$BF,$98,$D3,$E2,$5F,$33
	.dc.b	$FE,$29,$8C,$36,$33,$87,$F8,$96,$72,$FF,$89,$F0,$F3,$7C,$97,$87
	.dc.b	$E4,$3C,$AD,$9F,$16,$68,$C4,$7F,$46,$AE,$36,$BF,$8D,$B7,$C4,$DF
	.dc.b	$E3,$DF,$FB,$C7,$7F,$8C,$FC,$9D,$FE,$1F,$D8,$74,$86,$5F,$26,$30
	.dc.b	$DC,$BB,$4F,$92,$78,$2B,$EA,$F7,$C3,$C2,$FF,$5C,$8C,$B6,$33,$7F
	.dc.b	$5F,$EA,$BE,$74,$FC,$06,$BE,$FD,$51,$88,$F3,$C6,$31,$FB,$31,$BF
	.dc.b	$A8,$FB,$9F,$95,$7E,$01,$B1,$F3,$1A,$EE,$53,$F7,$B8,$3F,$35,$D3
	.dc.b	$79,$3F,$79,$7F,$DD,$8F,$C4,$4E,$9C,$5F,$E1,$67,$71,$F0,$F8,$3E
	.dc.b	$A1,$38,$3E,$B9,$DF,$8B,$73,$D1,$B7,$F8,$F6,$E1,$F8,$39,$FE,$14
	.dc.b	$78,$3E,$09,$F0,$9F,$E0,$AE,$7F,$CE,$4F,$F4,$99,$CA,$8C,$BF,$A1
	.dc.b	$8C,$C6,$30,$46,$1F,$C4,$4C,$3D,$98,$FA,$9F,$89,$93,$F2,$17,$AD
	.dc.b	$FA,$9B,$CC,$61,$FD,$A1,$88,$D5,$7F,$88,$AB,$94,$62,$3F,$2B,$E4
	.dc.b	$3D,$5F,$DC,$7D,$7F,$AB,$99,$39,$C5,$39,$8C,$36,$B8,$CF,$AE,$FF
	.dc.b	$29,$8F,$D5,$71,$87,$F0,$BF,$C2,$FF,$83,$7E,$59,$DF,$9D,$18,$C3
	.dc.b	$F6,$ED,$BF,$CC,$35,$FC,$BF,$DC,$FE,$FE,$7F,$7A,$D8,$FC,$F3,$5D
	.dc.b	$C7,$EF,$4F,$C6,$F1,$F8,$23,$FA,$54,$62,$BF,$A8,$36,$32,$7F,$82
	.dc.b	$67,$87,$37,$E0,$23,$F7,$1D,$5F,$B6,$6E,$9F,$05,$FB,$A7,$83,$1E
	.dc.b	$AF,$36,$BF,$A2,$6E,$1F,$C9,$5C,$C7,$4E,$2F,$C3,$3D,$08,$7D,$07
	.dc.b	$2D,$3E,$A1,$FE,$E0,$E7,$7B,$1F,$E6,$6F,$13,$FD,$A5,$EC,$FF,$B8
	.dc.b	$E9,$18,$AD,$C6,$2F,$FD,$DA,$D3,$FC,$DF,$5C,$6B,$85,$F5,$AD,$C6
	.dc.b	$2B,$81,$F1,$8F,$E5,$E4,$3F,$AA,$14,$FE,$00,$53,$F1,$A7,$F0,$DC
	.dc.b	$3F,$CF,$FF,$13,$FA,$FD,$91,$8C,$FE,$85,$2B,$53,$98,$EA,$78,$08
	.dc.b	$60,$FC,$06,$FB,$7F,$8B,$B6,$BF,$AD,$36,$FF,$19,$7F,$9D,$BF,$EC
	.dc.b	$9E,$CD,$B3,$48,$FE,$80,$7E,$5B,$FA,$5C,$3F,$B6,$36,$32,$6D,$FC
	.dc.b	$97,$83,$FE,$53,$CB,$E3,$1F,$34,$F5,$9F,$C0,$39,$BF,$BB,$7D,$7F
	.dc.b	$EE,$6C,$D2,$3F,$C2,$C7,$E0,$BF,$C7,$B8,$FF,$85,$C6,$0D,$46,$7B
	.dc.b	$3F,$8A,$3E,$4E,$5F,$F3,$46,$3E,$4C,$BC,$CF,$92,$FF,$92,$F7,$FF
	.dc.b	$AE,$3F,$09,$FE,$51,$DF,$FC,$B9,$B1,$99,$FE,$6C,$DA,$FD,$A7,$83
	.dc.b	$F8,$27,$F7,$AD,$72,$7E,$04,$C9,$4F,$BC,$7E,$30,$F8,$E7,$C9,$3F
	.dc.b	$80,$74,$F3,$7F,$AB,$B6,$BF,$AB,$47,$33,$25,$ED,$FF,$15,$78,$7F
	.dc.b	$E2,$5F,$C5,$BF,$B5,$51,$A4,$AF,$EF,$11,$F4,$1B,$B7,$F8,$C3,$C5
	.dc.b	$FF,$17,$36,$75,$3E,$A3,$D0,$7A,$CF,$6E,$1F,$E1,$1E,$6F,$F0,$BC
	.dc.b	$FE,$4A,$7F,$C2,$4D,$7F,$C0,$DE,$37,$E6,$1F,$87,$C1,$F6,$1E,$CD
	.dc.b	$7F,$C1,$3A,$ED,$FF,$07,$35,$CD,$75,$6D,$FF,$81,$7C,$8F,$F7,$7C
	.dc.b	$7F,$80,$9B,$3F,$E3,$E8,$D1,$FE,$F4,$C7,$FC,$97,$1F,$02,$3F,$C3
	.dc.b	$3B,$7C,$47,$FC,$7E,$D7,$FC,$40,$D7,$E2,$10,$F8,$E4,$F1,$BF,$CE
	.dc.b	$1F,$17,$FE,$6D,$94,$A3,$0D,$3F,$CE,$05,$3F,$D0,$5C,$9C,$23,$FC
	.dc.b	$FA,$48,$FF,$10,$0F,$33,$FD,$0F,$1F,$E8,$DF,$8B,$93,$FE,$8D,$6B
	.dc.b	$FE,$27,$F0,$3F,$DA,$59,$27,$FD,$34,$D9,$FF,$4F,$35,$E1,$D3,$FE
	.dc.b	$75,$9F,$F4,$D7,$C4,$C9,$FF,$4D,$35,$FF,$1A,$7B,$4F,$4B,$FC,$F9
	.dc.b	$3F,$EB,$7C,$0D,$25,$7B,$29,$E9,$F0,$8C,$35,$EA,$D7,$E6,$C7,$0A
	.dc.b	$EB,$1F,$1B,$D5,$FF,$6B,$78,$FF,$22,$3F,$DA,$BE,$B7,$D4,$8C,$4E
	.dc.b	$A7,$84,$E5,$FF,$74,$7A,$F5,$FF,$73,$B5,$FF,$67,$35,$FF,$66,$37
	.dc.b	$D9,$FF,$7C,$5F,$FB,$D0,$93,$F6,$A3,$2F,$FA,$AE,$3F,$D1,$24,$70
	.dc.b	$7F,$B1,$DF,$F6,$51,$1C,$1F,$EE,$67,$E7,$11,$C1,$F2,$B0,$FE,$E6
	.dc.b	$FC,$A6,$BF,$2B,$2F,$F5,$94,$FF,$BD,$8F,$91,$93,$DA,$CD,$7D,$76
	.dc.b	$BF,$80,$6B,$F5,$1A,$A3,$FC,$8B,$7E,$08,$C7,$A3,$A9,$CD,$F8,$23
	.dc.b	$1E,$D4,$BE,$14,$8C,$57,$DD,$F7,$1E,$E7,$81,$B7,$DD,$B3,$B9,$F2
	.dc.b	$FB,$AF,$7A,$92,$FC,$D1,$97,$97,$BF,$D4,$8E,$0F,$2F,$C0,$7A,$C8
	.dc.b	$E0,$F4,$FC,$0F,$BC,$8E,$0F,$89,$E0,$BE,$0B,$F8,$EF,$07,$F8,$01
	.dc.b	$B1,$98,$FE,$42,$D7,$F3,$CD,$7F,$83,$B5,$FF,$54,$35,$FE,$2A,$D7
	.dc.b	$B1,$D2,$3B,$91,$C1,$0E,$EE,$5A,$64,$F0,$B9,$69,$93,$C6,$E5,$A6
	.dc.b	$4F,$2B,$96,$99,$3C,$E7,$80,$87,$62,$9D,$0A,$72,$29,$C8,$E0,$43
	.dc.b	$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$8F,$A4,$48,$F8,$11
	.dc.b	$96,$C6,$63,$E1,$35,$F5,$B5,$F3,$35,$F1,$B5,$F1,$35,$EC,$74,$8E
	.dc.b	$E4,$70,$43,$BB,$96,$99,$3C,$2E,$5A,$64,$F1,$B9,$69,$93,$CA,$E5
	.dc.b	$A6,$4F,$39,$E0,$21,$D8,$A7,$42,$9C,$8A,$72,$38,$10,$E0,$53,$81
	.dc.b	$4F,$F5,$26,$9F,$E1,$2F,$5B,$FD,$45,$1F,$DD,$9A,$FF,$A8,$0F,$D9
	.dc.b	$FE,$03,$FE,$04,$F5,$FF,$6E,$D8,$C8,$8C,$B4,$CB,$82,$38,$21,$FB
	.dc.b	$E8,$E8,$D3,$91,$63,$8B,$4E,$27,$F0,$48,$E2,$D3,$89,$63,$8B,$4E
	.dc.b	$27,$FA,$5A,$38,$B4,$E2,$58,$E2,$D3,$8E,$09,$E2,$D8,$E2,$D9,$E2
	.dc.b	$D8,$E3,$3C,$08,$F0,$6C,$F0,$6C,$70,$34,$48,$D3,$4D,$38,$23,$82
	.dc.b	$1F,$2C,$87,$C5,$29,$EF,$2C,$FA,$5B,$3E,$46,$9E,$12,$CF,$81,$A7
	.dc.b	$80,$EC,$43,$B1,$67,$B3,$67,$B3,$4E,$C5,$9E,$CD,$3B,$1D,$09,$3D
	.dc.b	$1B,$3D,$1A,$74,$2C,$F4,$69,$D2,$79,$12,$79,$36,$39,$36,$79,$36
	.dc.b	$39,$4F,$02,$3C,$1B,$3C,$1B,$1F,$E6,$CD,$5F,$0A,$CD,$F0,$9F,$E4
	.dc.b	$99,$CC,$62,$B8,$56,$EB,$4D,$8F,$F3,$BF,$13,$C0,$C4,$7F,$A1,$04
	.dc.b	$66,$FA,$46,$63,$17,$FE,$1D,$AE,$33,$FE,$8C,$EC,$7C,$2E,$21,$18
	.dc.b	$C3,$9C,$6E,$75,$39,$BF,$F1,$47,$07,$1C,$7F,$C2,$18,$70,$C3,$CB
	.dc.b	$39,$3B,$7E,$4E,$F8,$EA,$4A,$E1,$19,$79,$5F,$F1,$B6,$EF,$DC,$8F
	.dc.b	$7F,$81,$F0,$E5,$FE,$77,$6E,$5F,$EA,$06,$BF,$25,$AF,$FA,$B9,$FA
	.dc.b	$B1,$C7,$6F,$89,$18,$9E,$51,$EF,$F8,$89,$7E,$66,$DF,$FA,$1F,$E2
	.dc.b	$9F,$15,$F3,$5B,$5F,$15,$B3,$B9,$FA,$5F,$1A,$FA,$1A,$4B,$F9,$91
	.dc.b	$96,$FD,$E7,$02,$38,$3E,$97,$C7,$7F,$AE,$30,$FF,$4D,$37,$4F,$F6
	.dc.b	$1B,$5F,$C6,$35,$FF,$68,$FA,$8F,$3D,$E6,$FD,$D8,$C3,$63,$37,$FE
	.dc.b	$A7,$9C,$BF,$EA,$9F,$85,$FF,$5A,$DF,$E1,$5B,$19,$F2,$3F,$DB,$ED
	.dc.b	$9F,$C5,$7C,$83,$E4,$33,$1F,$ED,$A8,$C3,$A8,$CB,$FA,$58,$CC,$FF
	.dc.b	$AE,$7F,$29,$FD,$A4,$FE,$55,$B1,$9D,$78,$CF,$FB,$C2,$72,$FE,$CB
	.dc.b	$8F,$F7,$9F,$23,$FE,$87,$6C,$66,$93,$96,$C7,$FB,$C3,$85,$C3,$C2
	.dc.b	$3F,$75,$18,$9D,$C6,$27,$48,$4E,$EB,$ED,$F0,$6A,$5C,$3F,$EF,$69
	.dc.b	$CE,$1D,$27,$94,$E5,$FE,$14,$38,$3E,$DD,$7D,$9E,$1F,$BE,$DC,$39
	.dc.b	$56,$E7,$2F,$2F,$59,$F5,$7F,$C8,$7E,$4B,$FD,$D1,$18,$CD,$18,$9D
	.dc.b	$57,$16,$BF,$C7,$72,$FC,$6C,$F8,$49,$1E,$88,$CB,$C6,$9F,$84,$75
	.dc.b	$6C,$66,$30,$7A,$C8,$F3,$C7,$F3,$B1,$88,$F1,$47,$E8,$3A,$FE,$11
	.dc.b	$23,$86,$7F,$E1,$4C,$3F,$47,$39,$9E,$33,$9A,$E5,$1F,$CE,$70,$7C
	.dc.b	$63,$B3,$FC,$4F,$18,$6C,$66,$7F,$C5,$33,$97,$FC,$57,$ED,$BE,$37
	.dc.b	$F8,$26,$5F,$72,$31,$D7,$FD,$4E,$31,$3A,$AF,$E9,$55,$AB,$F3,$67
	.dc.b	$2F,$D5,$5F,$C0,$B3,$F4,$BE,$C8,$CB,$F8,$39,$E0,$FC,$48,$CE,$9F
	.dc.b	$12,$30,$43,$91,$4E,$D3,$EC,$C7,$E2,$60,$F9,$1F,$4E,$BF,$DB,$5B
	.dc.b	$19,$BF,$5D,$70,$9F,$CE,$CF,$CE,$FB,$FF,$96,$4A,$F9,$7B,$23,$FC
	.dc.b	$9F,$97,$9E,$7C,$A4,$3C,$B9,$FB,$34,$F8,$71,$83,$D5,$EE,$3F,$08
	.dc.b	$7C,$77,$F8,$0F,$8B,$F8,$8D,$2B,$53,$9D,$3F,$BE,$46,$1F,$CD,$98
	.dc.b	$BE,$6E,$2B,$C8,$7D,$1C,$7C,$9C,$1F,$4A,$33,$9F,$AD,$F3,$1F,$C2
	.dc.b	$F0,$ED,$EA,$79,$1B,$87,$7C,$BF,$C3,$53,$9C,$BC,$2F,$AA,$FC,$07
	.dc.b	$D9,$29,$E4,$FC,$0F,$FA,$D2,$F0,$78,$46,$1B,$CB,$FC,$75,$B3,$BC
	.dc.b	$7F,$4D,$F0,$CF,$AB,$FF,$50,$46,$1B,$3F,$C9,$61,$3B,$8F,$F0,$41
	.dc.b	$97,$1B,$FD,$2E,$7F,$D4,$77,$7E,$96,$FF,$12,$DB,$FE,$61,$3A,$9E
	.dc.b	$70,$8C,$57,$C3,$E0,$F9,$D3,$F0,$8F,$9E,$4A,$F2,$E3,$E8,$8F,$F2
	.dc.b	$36,$3E,$49,$F6,$47,$67,$6F,$80,$E8,$4A,$EE,$4E,$3F,$C9,$C6,$5B
	.dc.b	$1E,$4F,$49,$F0,$6F,$53,$FA,$67,$A3,$FE,$BC,$9C,$C6,$34,$FF,$27
	.dc.b	$46,$31,$FF,$29,$46,$2F,$94,$60,$FE,$AD,$3B,$BE,$93,$A6,$DF,$5A
	.dc.b	$DC,$E6,$FF,$93,$C7,$F4,$F2,$1F,$CB,$8A,$68,$C1,$0F,$BD,$A7,$F2
	.dc.b	$D6,$9F,$85,$29,$FB,$2C,$FE,$6B,$F6,$73,$FF,$0E,$BD,$8F,$E2,$65
	.dc.b	$D7,$F0,$4F,$E2,$7A,$FF,$BC,$37,$7F,$F0,$6D,$FE,$28,$F0,$47,$DE
	.dc.b	$C3,$C8,$D3,$31,$E2,$3F,$02,$FE,$8B,$9E,$47,$F7,$6E,$3F,$D5,$C6
	.dc.b	$27,$E1,$10,$F3,$9F,$0C,$FE,$1D,$CB,$FC,$23,$F0,$3F,$DD,$73,$FF
	.dc.b	$A5,$5B,$81,$A7,$C3,$BF,$F0,$2D,$F4,$3B,$1F,$CF,$CF,$8C,$53,$C0
	.dc.b	$53,$F8,$75,$FF,$18,$FC,$23,$B9,$2F,$F8,$70,$7C,$13,$D3,$93,$FC
	.dc.b	$9A,$DD,$FE,$AB,$DA,$32,$FB,$EB,$F9,$F3,$E8,$6B,$FD,$0B,$B3,$F9
	.dc.b	$F3,$96,$BF,$D2,$BD,$68,$FF,$4A,$EF,$FF,$5D,$46,$23,$E0,$DF,$81
	.dc.b	$B8,$79,$1B,$9F,$C8,$7E,$11,$F2,$48,$60,$93,$CD,$CB,$5C,$CF,$57
	.dc.b	$F8,$87,$0F,$FA,$05,$FE,$F7,$D5,$FC,$CD,$CB,$78,$FF,$BE,$47,$C1
	.dc.b	$CB,$C9,$A7,$F8,$26,$30,$7C,$38,$F4,$3E,$0C,$BD,$59,$78,$C9,$97
	.dc.b	$8E,$74,$7D,$59,$FB,$4F,$F8,$93,$5F,$F2,$41,$1F,$F1,$B3,$FE,$35
	.dc.b	$23,$83,$FC,$57,$F0,$9F,$E4,$F2,$38,$3F,$C7,$CF,$F8,$F4,$8E,$0F
	.dc.b	$F1,$9F,$C1,$FF,$95,$C8,$E0,$FF,$4F,$D7,$FB,$D6,$3F,$CC,$A4,$8C
	.dc.b	$E4,$8F,$D0,$F8,$FF,$E7,$56,$F3,$71,$AF,$F3,$E3,$5F,$F3,$C5,$71
	.dc.b	$7F,$D0,$93,$FE,$83,$23,$83,$FC,$E3,$E4,$FF,$A5,$3A,$BF,$D2,$51
	.dc.b	$FC,$B0,$FF,$7F,$46,$5D,$D7,$FA,$0E,$7F,$6B,$E5,$FF,$A8,$9B,$E6
	.dc.b	$66,$BF,$D4,$CD,$7F,$D4,$8D,$7F,$D1,$9E,$6F,$FA,$D2,$7F,$D6,$04
	.dc.b	$8F,$F4,$2C,$65,$FF,$2F,$C7,$FA,$0C,$8E,$0F,$92,$FF,$A4,$08,$E0
	.dc.b	$F8,$F7,$ED,$7F,$D3,$99,$FF,$A8,$67,$E6,$CE,$67,$FD,$3F,$9A,$31
	.dc.b	$5F,$EE,$1A,$FA,$BC,$79,$6C,$EA,$31,$7F,$E9,$EF,$3F,$FD,$E3,$5F
	.dc.b	$EE,$FA,$ED,$18,$9D,$79,$FC,$1F,$F5,$80,$F4,$5E,$7E,$63,$FD,$69
	.dc.b	$18,$BF,$43,$70,$FF,$61,$1F,$CA,$B9,$9E,$8F,$1B,$FF,$40,$36,$33
	.dc.b	$7F,$05,$B5,$C7,$D4,$D1,$27,$4F,$3C,$1E,$15,$24,$F9,$1F,$07,$A6
	.dc.b	$F5,$29,$27,$E8,$46,$5F,$5D,$F5,$23,$83,$EB,$7A,$AF,$5B,$D8,$D8
	.dc.b	$CC,$60,$F3,$79,$12,$3E,$3B,$73,$FE,$9F,$EB,$FC,$0D,$FF,$DA,$AD
	.dc.b	$7E,$9B,$5F,$F7,$23,$57,$84,$F6,$3F,$8D,$B6,$3F,$9A,$35,$FE,$BA
	.dc.b	$D7,$FD,$C6,$D7,$F0,$2D,$7F,$DE,$4D,$7B,$C7,$62,$3D,$9A,$F6,$6B
	.dc.b	$D9,$AF,$66,$BD,$9A,$F6,$3A,$10,$E8,$53,$A1,$4E,$85,$3A,$14,$E8
	.dc.b	$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E,$06,$88
	.dc.b	$68,$A6,$8C,$10,$C1,$23,$EA,$12,$3E,$4B,$63,$E0,$35,$F4,$B5,$F2
	.dc.b	$35,$F0,$B5,$F0,$35,$F0,$47,$62,$3D,$9A,$F6,$6B,$D9,$AF,$66,$BD
	.dc.b	$9A,$F6,$3A,$10,$E8,$53,$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8,$A7
	.dc.b	$22,$9C,$8A,$3C,$BF,$F4,$33,$FC,$A0,$C7,$97,$FC,$ED,$B8,$7F,$89
	.dc.b	$67,$71,$FA,$EE,$03,$FC,$68,$FD,$9B,$83,$F6,$E5,$72,$6C,$E4,$53
	.dc.b	$F6,$EE,$DC,$BC,$9D,$1F,$C2,$1D,$3C,$0C,$E1,$E3,$8E,$04,$8F,$0B
	.dc.b	$5E,$A6,$08,$60,$87,$57,$83,$4D,$94,$D9,$E5,$3C,$6F,$99,$C3,$C9
	.dc.b	$A7,$22,$9C,$A3,$D4,$E5,$AE,$63,$A4,$62,$3C,$44,$39,$95,$D3,$82
	.dc.b	$38,$24,$74,$24,$7A,$88,$FA,$63,$71,$83,$CD,$1C,$0D,$15,$C9,$B3
	.dc.b	$D4,$53,$CC,$ED,$CB,$C9,$D1,$E8,$74,$F0,$33,$1E,$48,$E0,$48,$F0
	.dc.b	$B5,$EA,$60,$86,$08,$75,$78,$34,$D9,$4D,$9E,$53,$C6,$F9,$9C,$3C
	.dc.b	$9A,$72,$29,$CA,$3D,$4E,$5A,$FE,$3B,$C5,$7C,$8F,$F8,$BF,$D9,$FF
	.dc.b	$60,$C6,$36,$46,$27,$55,$FB,$FA,$D4,$FF,$B2,$39,$1F,$59,$AA,$FF
	.dc.b	$15,$C6,$30,$E7,$1F,$C1,$E3,$15,$C6,$B8,$4F,$F0,$4F,$80,$7B,$98
	.dc.b	$84,$63,$0E,$71,$FE,$38,$8C,$5F,$FB,$69,$B1,$FC,$55,$7B,$BC,$DB
	.dc.b	$5C,$A7,$C1,$39,$3C,$18,$BD,$EA,$4A,$E1,$5F,$E4,$78,$F3,$7D,$AB
	.dc.b	$E9,$A9,$2B,$71,$97,$8D,$BD,$7E,$E4,$8C,$BE,$7F,$82,$91,$E0,$CD
	.dc.b	$18,$D3,$E0,$7C,$1F,$B4,$95,$B7,$E0,$7C,$24,$8E,$4D,$9F,$91,$ED
	.dc.b	$7C,$52,$4F,$77,$E4,$7C,$44,$8F,$13,$67,$FA,$16,$EF,$C2,$BF,$CB
	.dc.b	$B1,$86,$C7,$F4,$76,$BF,$E7,$06,$BF,$D1,$3D,$47,$A8,$F4,$BE,$7C
	.dc.b	$61,$B1,$9B,$F0,$CE,$5F,$1F,$C3,$FC,$6B,$FA,$ED,$8C,$ED,$FD,$49
	.dc.b	$B3,$F7,$3D,$17,$A3,$98,$FF,$56,$99,$70,$FF,$AB,$8F,$C1,$FA,$2F
	.dc.b	$88,$CC,$7F,$AC,$1A,$A3,$4F,$08,$CF,$DB,$BD,$EA,$49,$FC,$7C,$65
	.dc.b	$FB,$3C,$2F,$87,$FD,$07,$27,$FA,$A5,$B1,$99,$FD,$14,$E5,$FA,$5E
	.dc.b	$7D,$7E,$93,$5F,$EE,$BC,$7F,$E9,$6B,$FB,$53,$99,$70,$FD,$39,$FA
	.dc.b	$3C,$5F,$5D,$BC,$7E,$1A,$FF,$56,$D6,$DF,$F0,$47,$1F,$D5,$AF,$A2
	.dc.b	$48,$CF,$1F,$F8,$1A,$3F,$1B,$F0,$EF,$AD,$FF,$63,$7C,$3E,$1B,$7F
	.dc.b	$E1,$39,$D6,$EA,$D3,$67,$A9,$4F,$F6,$0E,$3F,$87,$9F,$C1,$92,$3A
	.dc.b	$5F,$FB,$57,$E1,$79,$92,$33,$68,$FC,$59,$23,$BB,$5F,$87,$AF,$E7
	.dc.b	$7C,$7F,$8E,$48,$FF,$16,$6B,$F9,$F8,$FF,$17,$5B,$F2,$D7,$C1,$7F
	.dc.b	$86,$A3,$0D,$8C,$DF,$F8,$96,$72,$D8,$FC,$8D,$74,$6C,$7F,$B8,$A7
	.dc.b	$B7,$37,$33,$A7,$E0,$F2,$79,$36,$FD,$11,$EC,$E2,$FF,$17,$E5,$E9
	.dc.b	$8F,$5E,$7E,$F7,$B6,$4F,$83,$F4,$12,$3C,$CF,$F8,$86,$B9,$47,$5C
	.dc.b	$3B,$BF,$C1,$EF,$A3,$FE,$4C,$F8,$07,$9E,$FA,$9F,$F1,$FC,$61,$B1
	.dc.b	$99,$FF,$25,$4E,$5F,$F2,$67,$C1,$BC,$DF,$F7,$37,$E2,$DF,$E5,$F8
	.dc.b	$C7,$B4,$E2,$4E,$6B,$77,$FE,$5E,$17,$C2,$B7,$1F,$0B,$CE,$FB,$04
	.dc.b	$7E,$B3,$F0,$74,$F6,$C7,$BF,$87,$E6,$C6,$23,$F6,$6E,$AB,$F0,$DE
	.dc.b	$77,$E0,$C9,$3F,$81,$7F,$11,$F4,$BF,$46,$5E,$47,$AE,$BF,$CE,$B0
	.dc.b	$9D,$4E,$67,$E0,$46,$27,$F1,$0D,$8F,$AC,$7E,$03,$0F,$39,$38,$7F
	.dc.b	$9B,$FD,$3F,$F9,$FE,$FF,$3E,$DA,$FE,$0E,$D7,$FA,$23,$63,$C9,$C9
	.dc.b	$F4,$7E,$57,$C9,$18,$0E,$31,$F7,$36,$78,$4F,$97,$0F,$C5,$E1,$F2
	.dc.b	$B1,$F3,$3F,$A3,$FB,$E7,$99,$FE,$98,$C7,$BC,$62,$7B,$E1,$BA,$D3
	.dc.b	$63,$27,$CC,$E8,$FE,$A9,$8F,$A1,$B3,$FA,$4F,$0B,$DD,$18,$AF,$A7
	.dc.b	$F9,$4F,$F6,$4A,$5F,$C0,$8C,$37,$87,$F2,$8D,$9D,$9E,$FA,$87,$D6
	.dc.b	$21,$F9,$BB,$F0,$37,$0F,$19,$1E,$67,$03,$E4,$57,$CA,$EA,$FD,$BC
	.dc.b	$7E,$5E,$31,$87,$E5,$E7,$33,$FE,$C6,$1E,$DF,$FA,$F7,$D8,$FD,$CF
	.dc.b	$B3,$FE,$C6,$8C,$76,$23,$18,$FE,$B6,$B7,$7C,$A3,$15,$B8,$C1,$FE
	.dc.b	$D0,$9D,$DF,$4B,$E7,$39,$AE,$13,$FE,$CF,$C3,$FC,$21,$3F,$D7,$08
	.dc.b	$7E,$74,$A7,$F2,$12,$9F,$B7,$3F,$97,$65,$FC,$B5,$FC,$13,$5F,$9C
	.dc.b	$7E,$C0,$C9,$4E,$47,$72,$1D,$CF,$E7,$1A,$FE,$15,$FC,$01,$35,$CE
	.dc.b	$9F,$BD,$7F,$1D,$1D,$5F,$E9,$95,$E3,$6C,$67,$5F,$F1,$53,$FA,$0D
	.dc.b	$BF,$82,$CF,$E7,$4C,$47,$F2,$37,$07,$99,$F4,$9E,$72,$1F,$C2,$B8
	.dc.b	$B8,$47,$F0,$A6,$C7,$F1,$2E,$3F,$E2,$D3,$FB,$93,$B4,$7F,$06,$3E
	.dc.b	$31,$FE,$14,$E4,$F9,$8D,$33,$A2,$32,$FD,$C8,$F3,$1F,$C0,$4F,$B3
	.dc.b	$3F,$C2,$F9,$3C,$04,$7C,$C7,$F9,$54,$FE,$B7,$CD,$FE,$62,$9F,$D5
	.dc.b	$36,$3F,$99,$C7,$F9,$99,$AF,$E1,$8F,$01,$E2,$AF,$23,$FC,$1F,$9F
	.dc.b	$96,$9F,$EA,$48,$E4,$D7,$FB,$D4,$78,$7A,$3F,$A7,$7E,$63,$FD,$4F
	.dc.b	$B3,$FA,$B3,$71,$34,$FE,$AD,$6E,$DF,$EC,$4E,$1A,$FF,$5B,$6B,$FD
	.dc.b	$BF,$97,$FB,$53,$EB,$CF,$D2,$FE,$00,$97,$FE,$19,$E0,$F5,$92,$3D
	.dc.b	$6F,$C9,$7E,$19,$F8,$62,$9C,$B4,$77,$FC,$42,$47,$C4,$7E,$3F,$81
	.dc.b	$23,$E3,$B5,$FF,$10,$78,$3F,$30,$8F,$CC,$7C,$58,$FA,$A3,$C2,$47
	.dc.b	$1E,$2F,$FA,$37,$E2,$BF,$D1,$24,$9F,$F2,$6B,$FE,$50,$23,$FE,$46
	.dc.b	$6B,$FE,$50,$E7,$FF,$0C,$3F,$E5,$4D,$BF,$D2,$64,$7F,$C9,$AF,$F9
	.dc.b	$68,$8E,$0F,$F2,$DF,$C5,$7F,$A6,$88,$E0,$FE,$E4,$FF,$9D,$08,$E0
	.dc.b	$FF,$41,$EC,$8F,$F3,$FB,$63,$31,$83,$FD,$1A,$47,$FD,$1B,$1B,$7F
	.dc.b	$DE,$9E,$67,$F8,$96,$7F,$D8,$BE,$6F,$FA,$67,$BA,$F9,$CB,$FE,$6C
	.dc.b	$7F,$89,$E7,$4F,$E5,$1A,$FF,$8A,$FC,$FF,$F5,$4D,$7F,$B4,$1B,$1F
	.dc.b	$9A,$6B,$BE,$0E,$D3,$FE,$D6,$F4,$7F,$D7,$B3,$E0,$3C,$4E,$23,$5E
	.dc.b	$8F,$09,$FF,$60,$C7,$FB,$18,$8F,$FA,$29,$AF,$F9,$85,$FF,$50,$91
	.dc.b	$FF,$11,$3F,$30,$8F,$D0,$6B,$FE,$47,$F5,$1E,$A7,$F9,$36,$7F,$DD
	.dc.b	$B1,$8D,$BD,$14,$8C,$4E,$AB,$84,$FE,$F2,$7F,$C7,$B5,$C1,$B5,$FF
	.dc.b	$00,$23,$1E,$4F,$0A,$FB,$31,$BD,$F1,$1F,$5E,$D1,$88,$F1,$36,$3F
	.dc.b	$C4,$F3,$DF,$C2,$C1,$1F,$17,$16,$C8,$E0,$F6,$FA,$CF,$5B,$B3,$63
	.dc.b	$31,$83,$C5,$EB,$BD,$8A,$49,$F8,$F1,$97,$AF,$B3,$D0,$8E,$0F,$8D
	.dc.b	$E2,$7A,$C8,$E0,$F9,$5E,$27,$BC,$8E,$0F,$4F,$B5,$E8,$23,$83,$EC
	.dc.b	$FA,$0F,$6F,$E5,$36,$33,$18,$3E,$77,$B9,$FE,$40,$E2,$FE,$9C,$D7
	.dc.b	$F9,$C8,$70,$FF,$3C,$6B,$FE,$93,$6B,$FB,$16,$BC,$7D,$0F,$E6,$51
	.dc.b	$82,$38,$24,$78,$9A,$F8,$A3,$6E,$48,$66,$36,$F9,$5A,$F9,$5A,$F9
	.dc.b	$5A,$F9,$4E,$67,$12,$9C,$4A,$71,$29,$C4,$F1,$18,$8F,$3B,$82,$1D
	.dc.b	$4A,$75,$29,$D4,$A7,$53,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68
	.dc.b	$C1,$0C,$11,$F5,$BD,$5A,$F2,$6B,$C5,$AF,$18,$F9,$F1,$82,$38,$24
	.dc.b	$7C,$96,$BF,$0A,$36,$E4,$86,$63,$6F,$95,$AF,$95,$AF,$95,$AF,$94
	.dc.b	$E6,$71,$29,$C4,$A7,$12,$9C,$4F,$11,$88,$F3,$BF,$AA,$F4,$3F,$D1
	.dc.b	$6F,$F8,$91,$AF,$FA,$33,$9B,$FC,$7D,$18,$8F,$D5,$70,$7F,$5B,$BF
	.dc.b	$D1,$36,$3F,$64,$D7,$F6,$4D,$78,$9F,$B6,$BF,$E4,$0D,$8D,$3F,$BD
	.dc.b	$6C,$6A,$7B,$CE,$C8,$E0,$8F,$88,$C9,$5C,$BE,$43,$25,$72,$F9,$9A
	.dc.b	$F8,$CD,$99,$3A,$10,$E4,$53,$46,$08,$60,$87,$57,$BC,$7B,$E3,$91
	.dc.b	$92,$B9,$79,$99,$2B,$97,$AB,$5D,$18,$9F,$9A,$47,$DA,$68,$91,$C4
	.dc.b	$91,$B9,$F3,$B6,$7C,$AD,$7C,$6D,$7C,$07,$69,$E6,$D8,$D3,$E0,$6C
	.dc.b	$6A,$7B,$CE,$C8,$E0,$8F,$88,$C9,$5C,$BE,$43,$25,$72,$F9,$9A,$F8
	.dc.b	$CD,$99,$3A,$10,$E4,$53,$46,$08,$60,$FC,$0E,$5F,$D8,$FD,$CF,$EB
	.dc.b	$AD,$8F,$F4,$BB,$5F,$F2,$50,$70,$FF,$93,$7D,$E6,$AD,$57,$FD,$A2
	.dc.b	$73,$18,$AF,$F5,$1C,$62,$75,$5C,$5F,$F1,$9E,$D6,$BF,$C5,$CD,$7F
	.dc.b	$C7,$4D,$7F,$C5,$0D,$54,$F4,$B0,$DC,$3C,$53,$EC,$8C,$5F,$9E,$B9
	.dc.b	$72,$38,$A9,$30,$E3,$87,$F9,$B6,$75,$18,$8F,$0F,$CA,$78,$09,$19
	.dc.b	$C5,$1D,$B7,$93,$C6,$FF,$C6,$59,$7F,$84,$1B,$87,$F9,$3F,$9B,$DD
	.dc.b	$D0,$69,$2F,$B7,$A5,$F0,$08,$E0,$F3,$4F,$F9,$39,$BD,$5D,$08,$F9
	.dc.b	$F9,$BE,$49,$1C,$1C,$BB,$3C,$64,$70,$7B,$FE,$0B,$E8,$91,$C1,$F1
	.dc.b	$FC,$86,$2F,$47,$0D,$DB,$FC,$63,$3A,$9C,$9F,$3F,$C9,$77,$FF,$A7
	.dc.b	$EB,$FC,$6F,$18,$DF,$53,$97,$FD,$E7,$19,$9F,$F4,$F7,$95,$99,$FF
	.dc.b	$51,$35,$FF,$55,$78,$A8,$C4,$23,$0F,$E0,$7E,$09,$F2,$BA,$6C,$66
	.dc.b	$30,$72,$9F,$F7,$7C,$7E,$25,$B7,$FE,$B1,$F3,$74,$78,$38,$FE,$C7
	.dc.b	$BF,$FE,$98,$E4,$FD,$83,$78,$4D,$3F,$65,$C9,$AC,$BF,$62,$D7,$F4
	.dc.b	$4D,$7F,$78,$D5,$C3,$F4,$67,$E9,$F1,$7D,$76,$F1,$78,$2B,$F8,$39
	.dc.b	$C2,$74,$FD,$0E,$2F,$F6,$A5,$7D,$66,$C6,$4E,$47,$F1,$0E,$2F,$B5
	.dc.b	$3F,$5A,$32,$7D,$4E,$2E,$2F,$D8,$32,$7D,$57,$2F,$FB,$8F,$DF,$7A
	.dc.b	$5F,$EE,$5F,$7F,$FC,$FB,$18,$94,$62,$75,$1E,$C9,$D4,$F8,$C9,$1F
	.dc.b	$EE,$5A,$DB,$E3,$8E,$4B,$DE,$A4,$7A,$5F,$FB,$DE,$3D,$C4,$8E,$87
	.dc.b	$12,$BF,$43,$C3,$91,$F4,$36,$FD,$34,$62,$03,$D7,$FE,$75,$F2,$5F
	.dc.b	$CA,$5B,$C5,$FC,$71,$BC,$DF,$69,$B5,$FD,$09,$7E,$A5,$D2,$BF,$D4
	.dc.b	$71,$89,$FE,$85,$87,$1A,$D3,$63,$27,$F4,$CA,$7A,$CF,$3B,$FC,$71
	.dc.b	$18,$6C,$66,$7F,$D6,$33,$97,$FD,$63,$9F,$C0,$7E,$17,$27,$F0,$BC
	.dc.b	$3E,$14,$7C,$29,$D3,$E9,$B7,$77,$62,$47,$68,$CB,$DA,$FA,$3D,$B9
	.dc.b	$B0,$4C,$3E,$07,$17,$CD,$D3,$F8,$46,$7F,$35,$BB,$6A,$3F,$A5,$9F
	.dc.b	$0B,$D3,$FA,$FF,$99,$FE,$AA,$31,$92,$31,$A2,$B7,$3F,$2E,$30,$EC
	.dc.b	$C7,$E0,$7F,$28,$8F,$B5,$FF,$30,$EF,$F1,$23,$E3,$6D,$EE,$7F,$C0
	.dc.b	$31,$97,$F7,$9B,$7C,$5F,$C0,$BE,$A9,$23,$07,$C3,$3E,$36,$FF,$49
	.dc.b	$E2,$DA,$E8,$4B,$F0,$4F,$D7,$D3,$FC,$08,$7E,$74,$AF,$91,$AE,$DF
	.dc.b	$F0,$89,$E0,$61,$EA,$8E,$AF,$88,$C9,$F0,$30,$F4,$3B,$7F,$05,$F8
	.dc.b	$1F,$EA,$B8,$7F,$C3,$D3,$F8,$98,$C5,$FD,$16,$C6,$4F,$C6,$C7,$D5
	.dc.b	$F0,$BF,$D3,$4D,$86,$5D,$1F,$F1,$56,$AA,$DC,$7E,$5D,$AF,$F8,$BF
	.dc.b	$F0,$D7,$C4,$FE,$C3,$0F,$D6,$46,$1E,$A6,$23,$53,$FB,$99,$D6,$1E
	.dc.b	$18,$F8,$87,$C1,$7E,$B6,$9F,$52,$FF,$6F,$87,$89,$B5,$D8,$F3,$14
	.dc.b	$F5,$C7,$A1,$EA,$7C,$FA,$EB,$A7,$D0,$8C,$1F,$50,$D8,$F4,$7F,$9C
	.dc.b	$FA,$5F,$C7,$3C,$EF,$93,$3F,$C7,$A7,$2F,$F9,$06,$FF,$D9,$8D,$35
	.dc.b	$D8,$BE,$53,$AB,$E7,$5D,$EB,$4F,$13,$15,$C1,$E6,$D7,$F4,$A7,$F6
	.dc.b	$42,$1F,$C1,$CA,$7E,$34,$A7,$E4,$4F,$DE,$E3,$FC,$95,$F9,$67,$E1
	.dc.b	$72,$F9,$ED,$DB,$F0,$CF,$F3,$33,$A1,$DC,$86,$0F,$C0,$6D,$F9,$A9
	.dc.b	$FC,$21,$C6,$3F,$19,$19,$77,$19,$3C,$27,$E6,$5F,$15,$79,$DF,$BE
	.dc.b	$6E,$9B,$FF,$52,$F5,$BD,$0D,$BF,$D4,$52,$72,$D8,$F4,$BF,$14,$A7
	.dc.b	$CE,$2F,$17,$F9,$15,$F9,$66,$4F,$E0,$D1,$F7,$CF,$E1,$45,$73,$97
	.dc.b	$E8,$67,$FD,$35,$39,$9F,$01,$F2,$4A,$43,$B1,$63,$C4,$FA,$23,$FA
	.dc.b	$EF,$1F,$F8,$92,$3F,$A8,$35,$FD,$FB,$63,$71,$99,$E6,$F8,$4E,$E7
	.dc.b	$82,$3F,$B4,$F2,$71,$C8,$D2,$4F,$86,$7F,$A5,$FE,$69,$FD,$2B,$9B
	.dc.b	$FC,$0A,$D9,$34,$FE,$AD,$94,$8F,$F0,$17,$4F,$F5,$C8,$C4,$79,$72
	.dc.b	$F3,$BF,$0F,$97,$FC,$F8,$FF,$58,$E7,$F3,$11,$F2,$69,$FD,$B9,$FB
	.dc.b	$35,$C9,$F7,$BD,$33,$FF,$00,$47,$F6,$2E,$BF,$80,$47,$E7,$76,$7F
	.dc.b	$84,$E6,$E7,$F9,$D9,$FF,$11,$CE,$5F,$F0,$A4,$F0,$F4,$DD,$DF,$ED
	.dc.b	$BD,$7F,$C4,$51,$89,$FE,$07,$5C,$A3,$13,$AA,$E1,$3F,$94,$D7,$E0
	.dc.b	$B7,$AB,$F0,$1E,$4F,$FA,$23,$DD,$FF,$2E,$12,$7F,$C7,$EF,$F9,$20
	.dc.b	$8F,$F9,$18,$38,$7F,$CC,$1E,$EF,$F9,$94,$8F,$F9,$21,$FF,$28,$91
	.dc.b	$E4,$D7,$FC,$D1,$EE,$FF,$9B,$48,$FF,$94,$1F,$F3,$B2,$69,$1F,$F3
	.dc.b	$C3,$FE,$82,$23,$FE,$25,$6B,$FE,$8E,$D3,$93,$FE,$92,$6B,$98,$FF
	.dc.b	$4B,$C6,$23,$FC,$B3,$E5,$FF,$B5,$E7,$FD,$3C,$48,$FF,$47,$C6,$5F
	.dc.b	$F4,$8E,$1F,$EA,$CD,$3F,$D5,$51,$F6,$0F,$F5,$DF,$99,$FE,$64,$8C
	.dc.b	$9F,$B5,$EB,$FF,$5B,$B7,$CD,$D5,$7F,$A7,$27,$2D,$8C,$BF,$E9,$19
	.dc.b	$FF,$4B,$11,$FF,$3C,$BF,$28,$8F,$FA,$7D,$AF,$F8,$3D,$FF,$55,$91
	.dc.b	$FF,$27,$3F,$C2,$7D,$37,$9F,$F8,$E9,$FF,$3E,$88,$CE,$9E,$F8,$C4
	.dc.b	$ED,$FF,$41,$CE,$AB,$F8,$A6,$7F,$E1,$66,$F4,$6E,$93,$97,$E7,$C6
	.dc.b	$69,$E7,$52,$63,$E1,$8C,$BE,$3B,$79,$FD,$5B,$19,$8C,$1D,$78,$E4
	.dc.b	$73,$6E,$5E,$8F,$43,$91,$27,$B3,$E9,$F4,$5E,$95,$24,$FC,$58,$CB
	.dc.b	$E7,$F4,$FD,$84,$70,$6F,$D4,$E8,$47,$07,$BF,$C8,$F8,$84,$70,$7C
	.dc.b	$1F,$57,$E0,$11,$C1,$F1,$FD,$6F,$98,$47,$07,$DB,$ED,$F9,$7A,$7F
	.dc.b	$4B,$6F,$1F,$FB,$C1,$AF,$E1,$1A,$FF,$B0,$BD,$7F,$86,$FF,$51,$6B
	.dc.b	$FD,$55,$AF,$FB,$E9,$AF,$E3,$9A,$F3,$D7,$FA,$D2,$8C,$B6,$33,$18
	.dc.b	$23,$82,$47,$8D,$AF,$8D,$AF,$8D,$AF,$8D,$AF,$8C,$E4,$43,$91,$4E
	.dc.b	$45,$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21
	.dc.b	$82,$1F,$24,$F2,$14,$EE,$53,$A1,$4E,$65,$39,$C7,$D0,$79,$B5,$E6
	.dc.b	$D7,$9B,$5E,$6D,$79,$C7,$D0,$7B,$12,$34,$D8,$D4,$60,$8E,$09,$1E
	.dc.b	$46,$BE,$46,$BE,$46,$BE,$46,$BE,$43,$91,$0E,$45,$39,$14,$E4,$53
	.dc.b	$91,$C0,$87,$02,$9C,$0F,$F1,$E7,$9B,$FE,$90,$7F,$99,$9F,$E8,$63
	.dc.b	$3E,$67,$F3,$37,$07,$EB,$7F,$0C,$3F,$E9,$03,$F6,$CE,$3C,$EF,$DA
	.dc.b	$36,$38,$1B,$3F,$60,$E5,$FD,$B3,$5E,$66,$F0,$46,$63,$89,$97,$11
	.dc.b	$A8,$C1,$27,$C6,$D3,$C6,$E0,$8E,$08,$79,$0E,$05,$76,$D3,$67,$A1
	.dc.b	$CB,$4C,$9E,$87,$2D,$33,$1E,$98,$CB,$63,$33,$EE,$23,$E4,$3B,$1E
	.dc.b	$32,$18,$21,$DC,$C9,$4C,$C7,$77,$2D,$73,$1E,$09,$F2,$B6,$3C,$A7
	.dc.b	$8C,$E8,$49,$E4,$D8,$E0,$F6,$6B,$D8,$E5,$1C,$09,$1A,$8E,$46,$9C
	.dc.b	$46,$A3,$04,$9F,$3B,$4F,$3B,$82,$38,$21,$E5,$38,$15,$DB,$4D,$9E
	.dc.b	$97,$2D,$32,$7A,$1C,$B4,$CF,$B4,$FC,$09,$AB,$D3,$FF,$59,$4E,$63
	.dc.b	$15,$FB,$88,$C4,$EA,$B8,$BF,$EB,$9E,$47,$06,$2F,$F7,$F1,$8C,$39
	.dc.b	$C7,$F8,$D2,$31,$5C,$6B,$84,$FF,$8E,$3A,$1C,$58,$8F,$F2,$FB,$5F
	.dc.b	$F1,$8B,$63,$78,$76,$9F,$F3,$2F,$B7,$92,$46,$63,$FC,$73,$1C,$F1
	.dc.b	$72,$D2,$56,$E3,$2F,$1D,$1E,$E7,$8D,$B1,$98,$C1,$D7,$8E,$47,$36
	.dc.b	$EB,$FE,$E1,$F9,$07,$8F,$ED,$6D,$79,$DB,$3B,$9F,$81,$EE,$CA,$F6
	.dc.b	$6C,$8C,$57,$F9,$6B,$B7,$E3,$12,$B6,$FC,$9F,$32,$47,$26,$CF,$F9
	.dc.b	$82,$3A,$F8,$3F,$3C,$8F,$CB,$F8,$D7,$C6,$FD,$56,$C6,$63,$07,$D9
	.dc.b	$F3,$DE,$17,$C0,$C3,$FA,$2F,$2A,$73,$B7,$D7,$6C,$E9,$B3,$FD,$33
	.dc.b	$E0,$3F,$DD,$73,$C9,$B1,$9A,$FE,$97,$39,$6C,$71,$F5,$AB,$FE,$88
	.dc.b	$A4,$62,$3F,$A9,$C6,$1F,$C5,$36,$BD,$3E,$5B,$E0,$78,$52,$30,$FE
	.dc.b	$45,$B3,$C2,$77,$F2,$DF,$B4,$AF,$CB,$36,$7E,$2B,$63,$FB,$5B,$5F
	.dc.b	$F4,$FF,$C1,$BD,$7F,$0B,$63,$F4,$0D,$AF,$9E,$D8,$FF,$63,$F1,$7F
	.dc.b	$04,$F8,$2F,$F5,$0B,$63,$3C,$29,$CB,$63,$FD,$42,$D7,$F7,$5C,$7C
	.dc.b	$5B,$19,$7F,$DB,$31,$97,$E6,$EF,$FE,$00,$AF,$AD,$1B,$1B,$7E,$99
	.dc.b	$C3,$63,$F0,$1B,$F1,$9F,$B8,$69,$C3,$F7,$5A,$FF,$83,$B7,$FA,$B3
	.dc.b	$F4,$E3,$C1,$18,$7F,$DD,$3B,$F0,$C9,$EB,$BC,$FF,$F0,$BF,$C0,$FF
	.dc.b	$87,$A7,$31,$8B,$FF,$69,$C6,$14,$EE,$B9,$46,$0F,$97,$B7,$EC,$48
	.dc.b	$E0,$E7,$2F,$13,$F9,$DF,$E3,$7E,$18,$C6,$08,$C4,$EA,$B8,$B5,$FF
	.dc.b	$73,$FC,$1B,$E2,$7F,$C6,$35,$FB,$48,$C4,$F3,$3F,$C7,$6E,$23,$53
	.dc.b	$C2,$7F,$C7,$FC,$FF,$BA,$BF,$DC,$74,$7F,$1E,$6F,$1F,$F8,$DE,$75
	.dc.b	$5F,$EF,$8F,$69,$ED,$FE,$AF,$47,$CE,$FF,$92,$A7,$3C,$5F,$D6,$A3
	.dc.b	$15,$B6,$C6,$6F,$E8,$F0,$BE,$47,$FC,$85,$8F,$DA,$6C,$66,$7F,$CA
	.dc.b	$13,$96,$C3,$F0,$BF,$C4,$47,$46,$BF,$E5,$89,$FC,$2C,$62,$3F,$B7
	.dc.b	$CF,$17,$F8,$21,$F1,$B6,$F8,$8D,$D7,$E2,$46,$23,$FC,$C6,$7F,$25
	.dc.b	$70,$FF,$7A,$FC,$1B,$F8,$47,$37,$E4,$5B,$1F,$90,$6B,$E5,$6B,$FE
	.dc.b	$6E,$DF,$E8,$BD,$9A,$E6,$BB,$BE,$1B,$FF,$3A,$6F,$F5,$BD,$BF,$9D
	.dc.b	$3F,$E7,$88,$C6,$9F,$62,$3E,$E7,$02,$B7,$0A,$DF,$07,$A7,$D5,$FF
	.dc.b	$40,$F8,$B6,$8D,$23,$D3,$E2,$B9,$DF,$E2,$1B,$89,$A7,$82,$7D,$5C
	.dc.b	$9F,$92,$23,$F4,$5F,$C5,$FE,$11,$C8,$9A,$7A,$E7,$DD,$69,$F5,$CB
	.dc.b	$ED,$5F,$15,$FE,$2E,$FC,$1F,$F9,$E6,$1F,$E3,$09,$CD,$27,$2F,$1A
	.dc.b	$E0,$FF,$08,$9D,$CF,$2A,$F4,$9E,$72,$1F,$67,$87,$F0,$BC,$2F,$CF
	.dc.b	$9F,$21,$FA,$7C,$7F,$17,$39,$8C,$5F,$FA,$C2,$31,$3F,$B0,$37,$5C
	.dc.b	$9C,$4F,$E2,$FD,$2F,$F5,$BF,$E7,$1E,$AE,$7F,$F1,$5C,$A3,$13,$AA
	.dc.b	$FC,$A3,$67,$D1,$87,$E8,$09,$C1,$F5,$23,$F5,$9F,$7C,$C7,$FD,$8B
	.dc.b	$AF,$C8,$8C,$4F,$CB,$C7,$F9,$7C,$62,$75,$5C,$27,$F4,$9F,$27,$FD
	.dc.b	$05,$B7,$FD,$DD,$BD,$5E,$B9,$D4,$FE,$03,$D7,$F8,$EF,$E6,$79,$3F
	.dc.b	$CA,$3D,$3F,$7B,$9F,$F8,$CD,$69,$B8,$FC,$8B,$FD,$66,$DE,$E7,$EC
	.dc.b	$98,$DB,$F8,$DF,$AF,$C7,$23,$49,$7C,$EB,$F0,$67,$CA,$AF,$90,$F3
	.dc.b	$3A,$9A,$30,$6C,$FF,$7A,$FB,$AF,$75,$EF,$FF,$BC,$BB,$7F,$A7,$57
	.dc.b	$FB,$D2,$B5,$39,$BF,$E2,$B1,$89,$57,$07,$FE,$00,$36,$34,$DB,$EF
	.dc.b	$3C,$2B,$9C,$62,$75,$7D,$EB,$E4,$9F,$D2,$88,$7F,$29,$29,$F8,$C2
	.dc.b	$9C,$0F,$E2,$78,$FF,$A8,$BD,$DF,$B1,$9F,$CE,$CD,$18,$9D,$47,$03
	.dc.b	$27,$7A,$87,$72,$9F,$85,$DB,$8B,$F8,$98,$E2,$DA,$F0,$CE,$A3,$F9
	.dc.b	$E7,$01,$7D,$AF,$E4,$DB,$7F,$98,$8C,$1F,$9E,$87,$E1,$34,$93,$FA
	.dc.b	$4E,$4E,$91,$FC,$2F,$C2,$FF,$37,$D7,$C6,$3F,$9F,$1F,$22,$BF,$61
	.dc.b	$D1,$E3,$FC,$E7,$F7,$72,$69,$27,$F8,$3F,$67,$AF,$97,$F8,$5E,$DF
	.dc.b	$C8,$9B,$3A,$3E,$1D,$FD,$53,$E5,$10,$C1,$FC,$A7,$BF,$FD,$B3,$9F
	.dc.b	$F2,$8E,$FF,$E5,$6D,$8C,$D7,$F2,$96,$D7,$F3,$8F,$03,$F0,$0F,$F5
	.dc.b	$46,$B9,$9F,$C0,$B9,$6B,$E3,$3E,$29,$F7,$8A,$7F,$44,$F0,$79,$BF
	.dc.b	$C1,$5B,$D5,$FD,$56,$79,$CE,$5A,$7F,$1D,$E7,$FC,$07,$E7,$5F,$D8
	.dc.b	$49,$3F,$DE,$E3,$2F,$F6,$CF,$13,$AF,$47,$F6,$DB,$EB,$87,$F7,$2D
	.dc.b	$FF,$C4,$B5,$E4,$BF,$39,$F3,$0A,$62,$BF,$97,$1F,$91,$E5,$FF,$06
	.dc.b	$36,$3F,$C1,$8D,$7F,$C1,$8D,$7A,$72,$7E,$31,$F8,$CD,$7F,$C4,$DE
	.dc.b	$2F,$F8,$8E,$BF,$CC,$5E,$37,$1B,$FF,$14,$D7,$E8,$0F,$F1,$71,$4E
	.dc.b	$3E,$3F,$FB,$21,$FF,$16,$B6,$3F,$C7,$11,$FE,$6D,$8D,$67,$FE,$48
	.dc.b	$9F,$8B,$A7,$F9,$37,$4F,$F2,$3C,$63,$C8,$F8,$DA,$79,$F1,$F5,$F8
	.dc.b	$FF,$E9,$17,$F0,$1E,$5F,$F9,$93,$F1,$EF,$F3,$0C,$1A,$47,$FC,$DF
	.dc.b	$86,$5B,$1F,$E7,$48,$C4,$7F,$92,$7C,$DF,$F3,$CC,$FF,$BC,$7A,$64
	.dc.b	$FF,$AB,$9A,$FC,$AF,$3F,$F8,$37,$C6,$7F,$A2,$DB,$1F,$E6,$06,$BB
	.dc.b	$11,$FE,$9F,$F8,$C9,$1F,$E9,$E0,$E1,$FF,$54,$18,$3F,$D5,$AF,$FA
	.dc.b	$C1,$C1,$A7,$FD,$72,$47,$FD,$E4,$68,$FF,$5F,$94,$FF,$1E,$FA,$AC
	.dc.b	$B1,$E7,$A3,$15,$FE,$62,$AF,$F6,$4D,$6A,$73,$1F,$ED,$69,$FF,$6A
	.dc.b	$91,$C1,$C7,$D6,$FF,$72,$78,$FF,$EE,$2F,$63,$F9,$73,$67,$FA,$99
	.dc.b	$FE,$EC,$8C,$BF,$EF,$0C,$2B,$FE,$74,$F6,$7F,$DF,$15,$FE,$F6,$9C
	.dc.b	$CF,$F3,$99,$CB,$F3,$67,$FD,$9C,$47,$07,$FB,$61,$F9,$84,$70,$7C
	.dc.b	$A7,$FD,$CA,$47,$07,$FA,$25,$FF,$56,$91,$C1,$F2,$7D,$B7,$B7,$F3
	.dc.b	$E7,$E6,$4A,$73,$7F,$0A,$3F,$A9,$46,$27,$6D,$8F,$F8,$03,$1C,$B4
	.dc.b	$FB,$15,$C6,$7E,$EB,$63,$33,$BF,$5F,$C5,$5F,$76,$51,$88,$FF,$50
	.dc.b	$C6,$2B,$94,$E6,$7F,$D0,$FE,$E5,$7F,$D0,$ED,$A3,$4C,$7B,$B2,$FC
	.dc.b	$8D,$C3,$E1,$FB,$CE,$3E,$8D,$AF,$63,$67,$73,$E1,$F8,$27,$C0,$52
	.dc.b	$5F,$C3,$8C,$BF,$17,$E0,$9F,$05,$E9,$6C,$66,$30,$75,$F3,$E4,$73
	.dc.b	$6E,$3F,$47,$D3,$F8,$04,$9F,$A0,$FC,$3F,$4E,$47,$D3,$6C,$FE,$0B
	.dc.b	$17,$C1,$FA,$C9,$5E,$97,$CD,$F0,$92,$3D,$4D,$AF,$C5,$F8,$6F,$85
	.dc.b	$FE,$6D,$C9,$FE,$2C,$6C,$66,$3F,$9D,$35,$E3,$1A,$23,$A6,$BA,$30
	.dc.b	$43,$04,$3F,$84,$47,$56,$C7,$26,$9C,$4A,$71,$3F,$DE,$31,$C5,$B1
	.dc.b	$C5,$A7,$12,$9C,$4F,$F1,$D4,$71,$6C,$71,$69,$C4,$A7,$13,$E4,$19
	.dc.b	$29,$98,$E8,$E0,$F4,$90,$F1,$94,$EA,$53,$81,$A2,$1A,$29,$A3,$04
	.dc.b	$30,$47,$C6,$48,$F7,$46,$5B,$19,$8F,$63,$5F,$64,$68,$8E,$9A,$E8
	.dc.b	$C1,$0C,$10,$F7,$47,$56,$C7,$26,$9C,$4A,$71,$3E,$0C,$71,$6C,$71
	.dc.b	$69,$C4,$A7,$13,$E3,$47,$16,$C7,$16,$9C,$4A,$7F,$34,$F6,$7F,$D5
	.dc.b	$8F,$FA,$DB,$D8,$FE,$58,$A3,$2F,$FA,$A2,$FF,$57,$F6,$7F,$E8,$9F
	.dc.b	$6B,$F6,$6D,$9F,$D1,$B5,$FC,$EB,$5F,$CB,$B5,$FF,$51,$B7,$13,$49
	.dc.b	$1F,$EB,$F3,$89,$4E,$25,$8E,$2D,$8E,$33,$E6,$8C,$11,$C1,$1F,$49
	.dc.b	$92,$B9,$7D,$66,$4A,$E5,$F2,$B5,$F2,$B6,$7C,$AD,$9F,$2C,$F6,$21
	.dc.b	$D0,$A7,$42,$9D,$0B,$3D,$1B,$1D,$1E,$44,$39,$14,$E4,$58,$E4,$D8
	.dc.b	$E4,$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D,$8D,$46,$08,$E0,$91
	.dc.b	$F3,$CE,$A5,$39,$16,$38,$B6,$38,$CF,$CF,$8C,$11,$C1,$1F,$92,$64
	.dc.b	$AE,$5F,$88,$64,$AE,$5F,$2B,$5F,$2B,$67,$CA,$D9,$F2,$CF,$F9,$47
	.dc.b	$8B,$FA,$E6,$8F,$86,$FF,$14,$C6,$1B,$A7,$F9,$6D,$B3,$BA,$FF,$0F
	.dc.b	$70,$65,$B1,$97,$FC,$3F,$19,$6C,$65,$0F,$8A,$7A,$5B,$8C,$C6,$27
	.dc.b	$84,$6E,$74,$D8,$CF,$13,$93,$2D,$8C,$C6,$0E,$36,$F8,$9A,$49,$EF
	.dc.b	$19,$7A,$F3,$F4,$23,$83,$CB,$D8,$F8,$AF,$2B,$63,$31,$83,$3D,$92
	.dc.b	$38,$B7,$6F,$F2,$97,$6F,$22,$4F,$67,$D3,$F1,$72,$3C,$ED,$9F,$F2
	.dc.b	$D7,$8D,$E2,$24,$F9,$9F,$83,$F1,$92,$3D,$ED,$9F,$F3,$5F,$9C,$F3
	.dc.b	$BF,$C9,$51,$86,$F1,$7F,$A6,$E7,$5E,$0F,$D6,$6D,$7F,$9D,$3E,$61
	.dc.b	$AF,$E0,$23,$0D,$BF,$F5,$24,$E9,$FB,$ED,$9F,$F5,$B1,$23,$FD,$55
	.dc.b	$F1,$8F,$8C,$E5,$1F,$EB,$E9,$FF,$41,$FC,$73,$CF,$CC,$FE,$42,$30
	.dc.b	$FE,$31,$B3,$C2,$77,$2F,$86,$7B,$78,$6C,$7E,$51,$B5,$FE,$92,$6C
	.dc.b	$7F,$AF,$FE,$81,$F4,$3D,$1B,$1F,$9C,$6D,$7F,$A6,$9B,$1F,$EC,$7A
	.dc.b	$7B,$FE,$12,$47,$85,$F3,$F4,$7F,$A6,$3E,$3F,$F0,$3C,$A6,$92,$BF
	.dc.b	$DA,$5D,$6E,$6F,$F6,$C4,$FF,$AA,$A3,$15,$C6,$B7,$5A,$6C,$65,$FF
	.dc.b	$76,$D7,$FB,$AC,$8F,$D6,$E0,$FD,$FC,$7C,$D8,$E0,$7D,$3D,$F2,$FF
	.dc.b	$B9,$78,$3E,$AC,$EC,$F9,$9C,$1A,$9F,$E1,$27,$D5,$70,$FF,$0C,$D7
	.dc.b	$EA,$CF,$08,$EE,$7F,$AE,$BC,$EF,$8A,$47,$FD,$71,$D1,$FE,$F9,$8F
	.dc.b	$85,$A5,$9F,$7C,$63,$4F,$F5,$EC,$60,$F1,$AF,$AA,$7C,$17,$F1,$EF
	.dc.b	$90,$FF,$80,$13,$98,$C5,$70,$AD,$AA,$DC,$61,$E3,$F1,$2F,$27,$8D
	.dc.b	$7F,$2A,$C1,$39,$8E,$47,$91,$C5,$72,$AE,$33,$FC,$CB,$AF,$F7,$31
	.dc.b	$FB,$DE,$CF,$F1,$04,$62,$7F,$7D,$DB,$FE,$CB,$AF,$1F,$83,$FC,$EB
	.dc.b	$E2,$ED,$1A,$47,$EB,$FC,$03,$E0,$3F,$92,$5F,$DA,$94,$E7,$C4,$6D
	.dc.b	$FE,$33,$8C,$36,$F8,$D7,$B7,$D1,$FC,$03,$4F,$DC,$17,$9B,$E4,$C6
	.dc.b	$A3,$F1,$3A,$FB,$5B,$AF,$B5,$F7,$ED,$F0,$A3,$11,$F3,$F8,$3F,$D8
	.dc.b	$18,$7D,$4F,$2F,$A4,$FA,$27,$2F,$D1,$8C,$AC,$3C,$0F,$D1,$BE,$AD
	.dc.b	$C3,$C0,$7F,$94,$F3,$FB,$BE,$6F,$CF,$9F,$B8,$D7,$C8,$D7,$C4,$DD
	.dc.b	$7E,$87,$8B,$F3,$3F,$3B,$FA,$11,$8B,$46,$37,$FE,$6F,$18,$C3,$EC
	.dc.b	$C6,$27,$5E,$47,$D6,$E3,$B7,$F8,$1B,$DC,$E5,$B1,$97,$E7,$E7,$EE
	.dc.b	$E3,$FC,$81,$23,$BF,$47,$17,$FC,$0F,$9F,$E2,$63,$11,$F8,$8C,$D1
	.dc.b	$89,$D6,$1D,$EB,$F1,$7F,$12,$F6,$FF,$1B,$F8,$77,$E3,$A7,$F5,$53
	.dc.b	$9A,$4E,$67,$8D,$70,$9F,$DF,$D6,$EB,$55,$F7,$78,$7E,$DB,$74,$F7
	.dc.b	$3F,$8C,$D3,$F4,$10,$FC,$9D,$F1,$5F,$AB,$CB,$F8,$EC,$E6,$31,$7F
	.dc.b	$E2,$88,$C3,$FA,$F9,$DD,$72,$8C,$1F,$C1,$7E,$41,$D7,$FE,$30,$FB
	.dc.b	$1F,$65,$A7,$34,$9C,$C6,$1B,$5C,$6B,$D5,$50,$FB,$79,$7D,$97,$F4
	.dc.b	$DE,$E7,$F8,$FB,$F3,$CE,$FD,$68,$C5,$FF,$15,$6D,$70,$15,$C3,$E0
	.dc.b	$3F,$9E,$57,$F1,$B6,$CF,$EC,$DA,$FE,$C5,$BC,$7F,$E1,$0F,$A6,$FC
	.dc.b	$B5,$7F,$12,$EA,$46,$2F,$F1,$B1,$89,$F9,$31,$FC,$52,$31,$D3,$F7
	.dc.b	$BB,$3F,$73,$3F,$77,$5F,$80,$FE,$E5,$FD,$57,$8B,$D6,$3F,$2C,$DC
	.dc.b	$BF,$91,$37,$A8,$EE,$7A,$1F,$34,$E8,$53,$27,$D0,$31,$7E,$4F,$71
	.dc.b	$EE,$BD,$DF,$F3,$7F,$B5,$FD,$16,$FF,$DE,$13,$97,$FC,$E3,$9F,$FB
	.dc.b	$C5,$A7,$F9,$DA,$F9,$DF,$29,$D5,$F3,$9C,$D7,$03,$AD,$6D,$E5,$7D
	.dc.b	$EB,$C5,$87,$F9,$22,$7F,$A8,$10,$FE,$62,$53,$F8,$39,$4F,$E4,$C7
	.dc.b	$CE,$D9,$EE,$FE,$DF,$5F,$9B,$18,$3F,$8C,$E9,$FD,$32,$75,$AF,$D0
	.dc.b	$9D,$5F,$63,$C2,$43,$07,$E1,$B8,$39,$47,$E2,$DB,$19,$8E,$6F,$E2
	.dc.b	$63,$2F,$E5,$47,$C6,$7E,$75,$B3,$F9,$D6,$9F,$9D,$29,$FD,$37,$93
	.dc.b	$FD,$53,$F9,$EF,$EB,$1B,$3F,$AB,$94,$E5,$B1,$EB,$7E,$49,$4F,$E2
	.dc.b	$C7,$F8,$5B,$A7,$F8,$1F,$E1,$1F,$C0,$20,$D2,$4F,$F1,$0E,$DF,$F5
	.dc.b	$DE,$9F,$C9,$FB,$7F,$8A,$B7,$B7,$F8,$A3,$67,$75,$E1,$E1,$FC,$16
	.dc.b	$37,$B9,$8F,$F8,$16,$3F,$98,$D3,$F1,$0F,$E3,$3E,$0F,$F3,$76,$F8
	.dc.b	$5F,$E0,$86,$CE,$CF,$E9,$67,$E1,$AA,$E7,$FD,$7F,$C3,$FC,$24,$60
	.dc.b	$F1,$63,$E1,$7E,$29,$F2,$79,$BF,$C3,$5E,$27,$08,$FE,$AB,$CC,$8C
	.dc.b	$47,$F5,$56,$CF,$0C,$7C,$53,$C3,$BB,$FC,$3F,$ED,$FF,$71,$6C,$7F
	.dc.b	$70,$6B,$B9,$4F,$F9,$13,$C7,$FA,$33,$FE,$04,$3A,$9F,$E0,$22,$CF
	.dc.b	$F8,$29,$C9,$EC,$3F,$3A,$7E,$10,$AF,$F9,$33,$C8,$FF,$7B,$F4,$FF
	.dc.b	$87,$E7,$F3,$27,$F8,$74,$A7,$F8,$96,$36,$FC,$A3,$F7,$07,$73,$FC
	.dc.b	$4D,$E5,$69,$FF,$16,$35,$C9,$D0,$B3,$FE,$15,$F2,$FF,$B7,$C7,$F8
	.dc.b	$05,$B9,$FF,$91,$A3,$58,$7E,$24,$FF,$32,$F8,$FF,$1A,$3F,$C7,$4D
	.dc.b	$7F,$C7,$0D,$8D,$C7,$C6,$E6,$FC,$AB,$F1,$8F,$51,$EB,$8F,$F3,$8F
	.dc.b	$86,$87,$C5,$7F,$9B,$DA,$7F,$9C,$E3,$51,$93,$FC,$FF,$E7,$E2,$7F
	.dc.b	$D0,$8D,$7F,$CF,$DE,$7F,$FA,$1A,$77,$1D,$0F,$F5,$1F,$AD,$FE,$B8
	.dc.b	$E5,$FF,$4B,$12,$3F,$D6,$EF,$FB,$10,$8F,$FA,$83,$1F,$D7,$B6,$3F
	.dc.b	$D5,$1F,$13,$23,$FD,$A5,$EA,$7B,$63,$11,$FE,$B2,$F5,$7F,$DD,$71
	.dc.b	$FE,$BD,$24,$7F,$AE,$FD,$6E,$0F,$FB,$0C,$47,$EF,$7D,$DF,$F7,$D4
	.dc.b	$60,$D4,$67,$D8,$E3,$1E,$EF,$5F,$FD,$0B,$3F,$EB,$32,$47,$FB,$51
	.dc.b	$FF,$70,$91,$F7,$35,$FF,$64,$FA,$DF,$EE,$C7,$FB,$11,$8F,$6B,$ED
	.dc.b	$79,$FF,$EF,$B6,$C7,$FB,$CD,$AF,$FB,$4B,$DB,$FF,$7D,$CF,$FB,$E7
	.dc.b	$DD,$CC,$FE,$0E,$72,$FF,$A7,$DA,$F8,$B6,$DB,$FE,$DB,$3C,$67,$FB
	.dc.b	$6D,$CE,$07,$7B,$36,$A4,$AE,$22,$34,$E3,$6F,$9F,$97,$CD,$8C,$6D
	.dc.b	$E2,$9C,$C7,$FA,$D7,$B9,$DF,$E6,$6C,$76,$6D,$FF,$37,$6C,$7F,$AA
	.dc.b	$A3,$FD,$98,$47,$07,$FB,$07,$8D,$C3,$FE,$C0,$9F,$F7,$71,$27,$51
	.dc.b	$97,$F3,$11,$FD,$17,$DC,$FE,$F4,$DF,$43,$53,$FB,$31,$3A,$7C,$1E
	.dc.b	$F6,$9F,$DC,$60,$8C,$42,$31,$CD,$F3,$1B,$87,$87,$C3,$DE,$1F,$D0
	.dc.b	$6B,$E8,$B6,$75,$18,$D3,$64,$70,$70,$EF,$E6,$47,$07,$DC,$F8,$1F
	.dc.b	$F3,$DB,$7B,$FC,$F1,$8B,$FE,$D0,$DC,$FE,$77,$C0,$7B,$34,$34,$98
	.dc.b	$75,$F8,$1E,$64,$70,$7C,$1F,$57,$91,$1C,$1F,$03,$E0,$BE,$19,$1C
	.dc.b	$1F,$94,$F8,$37,$91,$F3,$9B,$19,$8C,$1F,$4B,$CC,$79,$5F,$DC,$F4
	.dc.b	$FF,$1F,$B6,$33,$1F,$DD,$DA,$FF,$82,$5A,$FE,$0C,$E2,$6E,$39,$11
	.dc.b	$C1,$0E,$4E,$5A,$64,$E9,$1D,$9F,$C8,$35,$FF,$12,$C7,$17,$6D,$76
	.dc.b	$D7,$67,$53,$25,$32,$78,$88,$74,$34,$7B,$8C,$9C,$8A,$72,$38,$18
	.dc.b	$38,$18,$21,$C4,$A7,$13,$04,$30,$43,$99,$92,$99,$8F,$8C,$E0,$8E
	.dc.b	$09,$1F,$16,$32,$D8,$CC,$7C,$06,$BE,$B6,$BE,$83,$89,$B8,$E4,$47
	.dc.b	$04,$39,$39,$69,$93,$A4,$76,$7D,$CD,$7D,$D1,$C5,$DB,$5D,$B5,$D9
	.dc.b	$D4,$C9,$4C,$9E,$22,$1D,$0D,$1E,$E3,$27,$FA,$F4,$FD,$17,$B3,$FE
	.dc.b	$B8,$6F,$B3,$FD,$21,$FF,$5C,$9F,$A9,$F6,$3F,$C3,$9E,$4D,$8F,$F6
	.dc.b	$51,$FE,$80,$8C,$BF,$A2,$6B,$FA,$26,$BB,$38,$9F,$EB,$87,$2D,$33
	.dc.b	$82,$32,$D8,$CC,$60,$8E,$09,$3E,$72,$3E,$53,$25,$72,$F9,$CC,$95
	.dc.b	$CB,$E9,$6B,$E7,$6C,$F9,$5B,$3E,$46,$CF,$43,$91,$E1,$21,$DC,$A6
	.dc.b	$8C,$10,$C1,$0F,$11,$C0,$A6,$CA,$6E,$7C,$CE,$DA,$ED,$AE,$E7,$D4
	.dc.b	$ED,$AE,$DA,$EC,$F7,$1E,$02,$47,$46,$C7,$06,$9C,$23,$44,$8D,$36
	.dc.b	$35,$18,$23,$82,$4F,$D0,$23,$E9,$32,$57,$2F,$AC,$C9,$5C,$BE,$D6
	.dc.b	$BE,$86,$CF,$95,$B3,$E4,$6C,$FE,$83,$3F,$EB,$79,$BE,$0F,$FD,$BD
	.dc.b	$18,$6E,$69,$D5,$7F,$59,$AD,$3F,$D6,$F9,$DF,$0B,$BC,$3F,$C4,$8D
	.dc.b	$9D,$35,$FF,$12,$4F,$0B,$EB,$1F,$E2,$7E,$47,$16,$E3,$FD,$39,$18
	.dc.b	$9D,$C6,$27,$4A,$75,$87,$B3,$4C,$E5,$FB,$BC,$38,$DA,$31,$5B,$9C
	.dc.b	$DF,$F8,$7A,$3F,$C3,$7C,$DD,$A3,$17,$FD,$ED,$AF,$F8,$FF,$E1,$9E
	.dc.b	$4F,$91,$B8,$78,$1B,$87,$F8,$0E,$91,$87,$E6,$75,$7B,$35,$FF,$30
	.dc.b	$DF,$A1,$B7,$FE,$51,$9E,$DF,$84,$4A,$EC,$D9,$FF,$2B,$7A,$7C,$C9
	.dc.b	$3C,$DE,$DD,$F2,$3B,$36,$7C,$DF,$10,$F8,$0F,$9E,$DA,$F9,$6D,$9D
	.dc.b	$CF,$D4,$F4,$5E,$CF,$C4,$9F,$81,$18,$DF,$E9,$B6,$33,$3B,$DB,$FC
	.dc.b	$51,$E3,$7F,$B1,$A7,$FD,$71,$7F,$ED,$F6,$CF,$DC,$9D,$3F,$EB,$F8
	.dc.b	$FF,$47,$11,$C1,$FE,$33,$F2,$1F,$21,$CE,$30,$D7,$FD,$9C,$D7,$FD
	.dc.b	$DD,$1C,$1F,$C5,$FC,$57,$E9,$DB,$AE,$D4,$66,$FF,$00,$D8,$FF,$22
	.dc.b	$7C,$5B,$DB,$E0,$D8,$FC,$BB,$6B,$F1,$0D,$8F,$F2,$67,$2F,$E3,$BE
	.dc.b	$2B,$F1,$8D,$8C,$EB,$FB,$B9,$CB,$63,$F1,$F9,$27,$F1,$8D,$8C,$C6
	.dc.b	$0F,$BB,$19,$8F,$F4,$F7,$0F,$E0,$E7,$24,$8F,$F3,$07,$07,$DF,$8F
	.dc.b	$BB,$7C,$64,$27,$4F,$0E,$0F,$D5,$D7,$F9,$A7,$D8,$7B,$7F,$1F,$27
	.dc.b	$C6,$E6,$30,$DD,$FF,$8D,$36,$77,$8F,$F1,$9E,$6C,$B6,$32,$FF,$16
	.dc.b	$8C,$B6,$32,$FA,$3C,$7F,$DE,$D7,$EE,$DB,$1F,$E7,$F6,$BF,$EC,$36
	.dc.b	$BF,$03,$6F,$D5,$5F,$CC,$24,$7F,$A1,$E3,$2F,$C3,$D9,$3F,$28,$91
	.dc.b	$97,$EA,$FD,$4C,$8E,$5C,$FF,$CE,$E3,$11,$FE,$D5,$E0,$FD,$94,$7D
	.dc.b	$42,$47,$77,$EB,$F8,$72,$3C,$4D,$7E,$CE,$8F,$BD,$7C,$C7,$DC,$F5
	.dc.b	$BF,$D4,$13,$98,$C5,$7F,$B9,$63,$0A,$77,$5C,$A3,$07,$FB,$AF,$8B
	.dc.b	$F0,$3F,$43,$FD,$15,$B7,$FD,$11,$BD,$1F,$83,$9D,$72,$3E,$8B,$FA
	.dc.b	$E7,$C5,$7E,$1D,$B8,$F5,$9D,$62,$AD,$36,$7E,$09,$0F,$95,$D1,$FD
	.dc.b	$03,$6F,$89,$1F,$3F,$D8,$EA,$FC,$6E,$04,$62,$35,$7D,$5D,$46,$5F
	.dc.b	$D2,$5F,$57,$C0,$7E,$A3,$87,$F9,$56,$1E,$0B,$FA,$5E,$2F,$CE,$AF
	.dc.b	$AB,$C0,$8C,$47,$F7,$D8,$C3,$F3,$8D,$C6,$A5,$F9,$E3,$EE,$7F,$6F
	.dc.b	$F1,$BF,$72,$31,$68,$C4,$EA,$3E,$F4,$EA,$78,$7E,$11,$D6,$0D,$24
	.dc.b	$F8,$CF,$76,$9F,$05,$F9,$FD,$3F,$E7,$AD,$7E,$0C,$7C,$4C,$BF,$25
	.dc.b	$E9,$7A,$75,$FF,$7F,$CF,$39,$CB,$63,$3C,$FE,$BF,$C3,$3F,$89,$FE
	.dc.b	$2D,$FC,$EE,$BF,$C4,$5A,$A7,$38,$7B,$1B,$3F,$A2,$9D,$2A,$E4,$FD
	.dc.b	$AC,$FD,$1A,$7A,$E3,$DB,$18,$3F,$D3,$A7,$EF,$BE,$47,$F5,$58,$FF
	.dc.b	$26,$9F,$F0,$6F,$6A,$73,$1F,$CA,$AB,$83,$FA,$BA,$E1,$18,$CF,$D9
	.dc.b	$D7,$FC,$AB,$EA,$3F,$C7,$DF,$88,$E3,$A9,$18,$C3,$F4,$8D,$C7,$F9
	.dc.b	$BD,$F1,$95,$FB,$B8,$BE,$9D,$7D,$6E,$7F,$D1,$BF,$05,$FE,$75,$F4
	.dc.b	$CF,$B2,$FF,$27,$7E,$21,$E4,$8C,$62,$8C,$4E,$A3,$F8,$14,$EA,$78
	.dc.b	$7E,$0F,$FB,$C9,$CB,$70,$F3,$D2,$31,$0C,$D3,$9A,$FE,$00,$7E,$63
	.dc.b	$0F,$9B,$5F,$82,$6C,$7E,$91,$BC,$9E,$B9,$FA,$07,$C1,$3F,$6E,$7C
	.dc.b	$13,$F2,$DF,$75,$E3,$76,$7F,$9B,$DC,$70,$7E,$46,$D1,$87,$F9,$43
	.dc.b	$67,$B1,$0F,$C5,$E9,$F4,$35,$FA,$71,$83,$C6,$E1,$FF,$77,$F8,$C9
	.dc.b	$FC,$77,$B8,$F7,$1F,$01,$FD,$7E,$73,$18,$B5,$F1,$7F,$BD,$DF,$1B
	.dc.b	$E1,$18,$AE,$15,$B8,$F0,$DF,$4A,$DC,$60,$FD,$29,$FE,$08,$21,$FD
	.dc.b	$8C,$A7,$F0,$42,$9C,$0F,$8F,$C1,$FE,$0E,$F8,$FF,$99,$9F,$D1,$6C
	.dc.b	$67,$4D,$D2,$7F,$39,$D3,$FE,$7C,$8E,$A7,$EE,$CE,$F3,$F8,$37,$83
	.dc.b	$F8,$96,$9F,$C1,$78,$3B,$34,$C9,$F9,$3E,$1F,$C9,$7B,$DF,$9C,$A3
	.dc.b	$48,$FE,$8A,$BF,$D6,$DF,$88,$7E,$9A,$0D,$23,$F1,$EE,$1F,$1C,$A7
	.dc.b	$E8,$FA,$3F,$80,$C3,$F0,$FF,$E0,$9D,$3F,$EC,$36,$C6,$4F,$E0,$A5
	.dc.b	$AF,$E2,$5D,$9F,$C5,$5F,$6F,$5F,$F1,$4D,$BF,$96,$36,$BF,$8A,$D6
	.dc.b	$E7,$E7,$CF,$FA,$96,$BE,$71,$E6,$29,$FC,$BF,$BB,$FC,$93,$F8,$B7
	.dc.b	$E9,$FA,$BF,$98,$B6,$33,$8F,$F3,$7A,$4F,$EB,$7C,$0F,$E8,$B8,$FF
	.dc.b	$84,$3C,$1F,$C3,$57,$F8,$56,$72,$D8,$CB,$F2,$48,$7E,$3C,$B7,$FD
	.dc.b	$A3,$C2,$DC,$7F,$4A,$6C,$66,$3F,$1C,$7F,$66,$8C,$E5,$FD,$B3,$E6
	.dc.b	$BF,$B5,$C9,$A4,$7C,$DE,$37,$4C,$7F,$BD,$E5,$FE,$2A,$8C,$1E,$29
	.dc.b	$F9,$27,$9C,$86,$31,$79,$1F,$E1,$67,$F2,$31,$FE,$37,$8C,$72,$FF
	.dc.b	$8F,$1B,$19,$F2,$15,$C2,$3E,$29,$E0,$27,$81,$FE,$C1,$7F,$C5,$B7
	.dc.b	$1F,$F1,$97,$95,$CA,$31,$1F,$D4,$39,$9E,$5F,$6E,$EF,$F1,$F6,$3F
	.dc.b	$C4,$EB,$53,$97,$FC,$78,$D7,$FC,$47,$E6,$FC,$72,$3F,$48,$C9,$4F
	.dc.b	$8B,$97,$9D,$FC,$93,$5F,$F0,$9D,$7F,$A7,$08,$FF,$9D,$DF,$F5,$39
	.dc.b	$1F,$F3,$9B,$FE,$04,$5E,$7F,$FB,$AE,$3F,$D4,$BC,$F2,$7F,$D0,$0D
	.dc.b	$7F,$D5,$FE,$7F,$CE,$8F,$F5,$99,$23,$AB,$FE,$93,$3E,$9F,$D0,$DA
	.dc.b	$34,$93,$FE,$A1,$7F,$C1,$CF,$FA,$D3,$DD,$FF,$47,$CF,$08,$D1,$0F
	.dc.b	$A9,$E9,$F6,$7F,$CD,$4D,$73,$1C,$5B,$5F,$EC,$2F,$91,$91,$FE,$9D
	.dc.b	$6B,$EE,$94,$7F,$B4,$09,$1F,$29,$FF,$6B,$91,$FF,$6D,$B5,$FF,$14
	.dc.b	$7A,$CC,$7F,$2B,$3F,$EE,$2A,$EB,$5F,$EE,$66,$BC,$27,$6F,$FB,$DB
	.dc.b	$0F,$F6,$BE,$FF,$EF,$26,$C6,$4F,$C9,$1F,$ED,$CF,$5F,$FD,$F5,$3F
	.dc.b	$F0,$02,$F7,$3F,$E3,$BF,$63,$F4,$EF,$00,$E2,$78,$46,$7D,$97,$B5
	.dc.b	$CA,$3F,$DD,$31,$87,$51,$95,$19,$9F,$BB,$0F,$6D,$F1,$BB,$6C,$78
	.dc.b	$DB,$3C,$1B,$1F,$86,$7F,$B2,$46,$74,$E4,$49,$F6,$71,$BC,$AF,$A5
	.dc.b	$5E,$28,$C5,$6A,$73,$5F,$ED,$FA,$F3,$F2,$39,$3F,$75,$3F,$EE,$96
	.dc.b	$BF,$AA,$6C,$6F,$6F,$99,$C8,$F7,$76,$D8,$CC,$60,$CF,$8D,$23,$8B
	.dc.b	$71,$F9,$3C,$AF,$7B,$C2,$DA,$FB,$4D,$9D,$CF,$83,$DF,$95,$E4,$6C
	.dc.b	$FE,$4C,$9A,$7E,$4D,$AF,$FA,$07,$E0,$3F,$D7,$55,$F8,$76,$CF,$DC
	.dc.b	$6C,$FE,$E9,$AF,$FA,$27,$E0,$5F,$05,$F2,$1B,$7F,$91,$6D,$FE,$19
	.dc.b	$B5,$CF,$E0,$FE,$7E,$06,$92,$FE,$7F,$C2,$7C,$E2,$38,$3B,$7C,$2F
	.dc.b	$84,$8E,$0F,$D1,$78,$5E,$6C,$3F,$C0,$4D,$7F,$32,$D7,$F8,$BB,$5F
	.dc.b	$F5,$33,$51,$FE,$2A,$CF,$FC,$18,$D3,$AB,$88,$E1,$19,$3B,$C6,$48
	.dc.b	$F0,$21,$A2,$9E,$47,$6E,$48,$64,$D9,$E8,$76,$E4,$D9,$93,$67,$B2
	.dc.b	$3C,$E4,$8E,$31,$B3,$89,$B7,$2D,$39,$9E,$F3,$C0,$43,$91,$4E,$07
	.dc.b	$43,$81,$A2,$1A,$38,$1A,$30,$68,$C1,$23,$D4,$F9,$1A,$F7,$6B,$D1
	.dc.b	$AF,$36,$BD,$88,$75,$8F,$82,$D3,$B3,$88,$E3,$19,$3C,$11,$92,$3C
	.dc.b	$08,$68,$A7,$95,$DB,$92,$19,$36,$7A,$5D,$B9,$36,$64,$D9,$EC,$8F
	.dc.b	$39,$23,$8C,$6C,$FE,$7F,$EC,$FF,$AC,$DB,$EC,$FF,$4F,$51,$97,$FC
	.dc.b	$02,$D7,$F5,$7C,$BF,$AA,$F6,$3F,$62,$D8,$FD,$8B,$5D,$9F,$EB,$D3
	.dc.b	$F2,$CE,$5A,$64,$FF,$4E,$B9,$69,$9C,$3F,$A0,$36,$35,$18,$9F,$1C
	.dc.b	$E4,$8E,$08,$F9,$8C,$95,$CB,$E8,$6B,$E6,$6B,$E4,$6C,$78,$63,$91
	.dc.b	$0E,$45,$39,$14,$E4,$57,$93,$C0,$87,$02,$9C,$0A,$F0,$74,$43,$45
	.dc.b	$74,$E0,$8E,$09,$3F,$30,$91,$F1,$8E,$EE,$27,$E0,$B6,$3D,$4D,$76
	.dc.b	$71,$3C,$4E,$5A,$64,$F1,$B9,$69,$99,$F2,$B6,$35,$18,$9F,$1C,$E4
	.dc.b	$8E,$08,$F9,$8C,$95,$CB,$E8,$6B,$E6,$6B,$E4,$6C,$78,$63,$91,$0E
	.dc.b	$45,$39,$14,$FA,$5B,$38,$BF,$D4,$FF,$07,$FD,$BD,$B3,$FE,$72,$9D
	.dc.b	$68,$AD,$36,$7F,$BA,$35,$FF,$03,$6F,$C9,$B1,$97,$FB,$1F,$07,$F7
	.dc.b	$98,$C4,$3E,$21,$F1,$1C,$63,$FB,$4C,$62,$76,$DC,$3F,$C9,$6D,$9F
	.dc.b	$F0,$87,$99,$C5,$61,$FE,$18,$AF,$F0,$B4,$E6,$31,$3F,$E9,$0E,$97
	.dc.b	$C4,$F1,$6E,$3E,$66,$E1,$FE,$5A,$6D,$71,$E7,$E0,$4A,$E0,$F3,$F8
	.dc.b	$A9,$1C,$DB,$7C,$FA,$BC,$04,$9F,$03,$E3,$EB,$91,$E3,$6C,$FC,$2F
	.dc.b	$8C,$F8,$1B,$1A,$4A,$F9,$FE,$07,$F9,$82,$BF,$CC,$5D,$FF,$E9,$56
	.dc.b	$F8,$1A,$8C,$6B,$FE,$B4,$F8,$D6,$DF,$E1,$CB,$FF,$3D,$C6,$27,$ED
	.dc.b	$B6,$33,$3B,$9F,$F3,$C7,$87,$5F,$F1,$29,$FE,$83,$70,$FF,$8A,$3C
	.dc.b	$3F,$F4,$1D,$7F,$A0,$67,$33,$FE,$8C,$F3,$EF,$8A,$69,$23,$F1,$BE
	.dc.b	$3F,$FA,$7A,$FF,$D3,$64,$8F,$F6,$7C,$65,$FF,$68,$FC,$7B,$E4,$3B
	.dc.b	$B6,$3F,$2E,$DA,$FF,$4D,$B6,$3F,$DB,$3F,$06,$F9,$1F,$3B,$63,$F3
	.dc.b	$ED,$AF,$F5,$1B,$63,$F1,$5C,$4F,$8E,$FC,$76,$E8,$C4,$EA,$73,$18
	.dc.b	$8F,$97,$D2,$E9,$E1,$5F,$97,$6C,$FE,$7D,$AF,$06,$BF,$E5,$8E,$5F
	.dc.b	$F5,$2C,$FE,$10,$91,$FE,$F6,$8C,$BF,$85,$C3,$FC,$BF,$3F,$80,$8F
	.dc.b	$21,$FE,$62,$E3,$FC,$34,$64,$FF,$61,$FD,$D3,$EE,$BD,$DF,$21,$86
	.dc.b	$DF,$E1,$84,$EF,$45,$6D,$B3,$FB,$BE,$BE,$2F,$C4,$6B,$FE,$74,$D7
	.dc.b	$FD,$65,$5F,$BB,$9C,$D7,$F1,$FF,$22,$F0,$1A,$48,$EF,$87,$E6,$B6
	.dc.b	$FF,$65,$F6,$FE,$A8,$91,$F1,$FA,$9D,$69,$F8,$C4,$9F,$94,$64,$AF
	.dc.b	$ED,$3D,$89,$1C,$E5,$18,$8F,$D5,$F0,$FE,$CE,$3E,$91,$23,$AB,$FB
	.dc.b	$EF,$1E,$47,$36,$BF,$E9,$8F,$89,$7B,$1F,$EF,$AD,$1F,$87,$7D,$39
	.dc.b	$CF,$77,$E5,$63,$15,$B6,$C6,$6F,$F7,$F6,$F7,$BE,$64,$7F,$8D,$E3
	.dc.b	$1C,$1F,$E3,$76,$C6,$67,$79,$76,$EF,$FD,$27,$C8,$DB,$34,$97,$ED
	.dc.b	$21,$F5,$F8,$7E,$9D,$FC,$08,$6B,$F0,$3A,$3A,$C7,$F1,$68,$C6,$1D
	.dc.b	$A3,$1B,$7E,$4F,$1F,$83,$7C,$27,$B7,$0F,$F9,$3F,$8D,$9F,$82,$7C
	.dc.b	$67,$EB,$DC,$3F,$47,$07,$D3,$FC,$98,$C3,$5F,$3C,$6B,$2F,$AB,$1E
	.dc.b	$49,$D6,$7F,$6B,$D6,$FE,$15,$F5,$1F,$59,$B5,$EA,$6B,$E1,$6B,$FC
	.dc.b	$47,$81,$F7,$FF,$ED,$CF,$03,$F6,$99,$7D,$B8,$C5,$7D,$EA,$DA,$AD
	.dc.b	$C6,$1F,$A1,$0D,$7E,$2E,$DF,$1F,$6F,$87,$3F,$16,$3F,$05,$F1,$9D
	.dc.b	$A3,$FD,$DB,$97,$E2,$E3,$11,$E7,$6C,$66,$3C,$47,$E4,$BE,$87,$FD
	.dc.b	$FD,$F4,$7F,$CB,$63,$9C,$62,$75,$EB,$7F,$BE,$AB,$55,$FC,$CE,$B4
	.dc.b	$FD,$EE,$5F,$5E,$7E,$A7,$F1,$99,$3F,$12,$F3,$F5,$72,$8C,$5F,$EA
	.dc.b	$73,$FE,$C7,$18,$7F,$5E,$62,$B8,$BF,$93,$F8,$F7,$C5,$FF,$45,$9F
	.dc.b	$2F,$2A,$73,$5E,$13,$85,$6D,$A6,$70,$F5,$1F,$94,$3E,$A7,$07,$D2
	.dc.b	$6E,$3F,$EA,$DC,$1F,$B6,$7F,$55,$B7,$E5,$5E,$28,$C3,$63,$FA,$D3
	.dc.b	$5D,$9F,$AD,$87,$D5,$7E,$F2,$30,$DC,$3F,$8A,$4E,$A7,$F7,$E6,$8B
	.dc.b	$1F,$84,$E5,$7E,$05,$F9,$79,$FC,$A3,$67,$4D,$7F,$3C,$D7,$DF,$C1
	.dc.b	$F7,$34,$F8,$5C,$3F,$76,$7F,$3D,$F6,$1C,$70,$FE,$38,$DC,$4D,$3A
	.dc.b	$57,$1A,$F9,$E4,$3F,$3D,$87,$7D,$BF,$44,$47,$FD,$D3,$F0,$29,$7C
	.dc.b	$B2,$7F,$C5,$7F,$01,$FE,$EF,$8C,$4E,$E3,$13,$AF,$0D,$7C,$EF,$77
	.dc.b	$FE,$F4,$9C,$BD,$AB,$85,$6D,$B1,$93,$B4,$F0,$AE,$78,$FE,$F2,$7F
	.dc.b	$C9,$A4,$3F,$AD,$14,$FE,$4E,$53,$F8,$01,$FC,$23,$2F,$81,$F1,$9F
	.dc.b	$B2,$D7,$E7,$CA,$31,$AF,$D0,$8C,$4F,$D6,$69,$D8,$F0,$90,$C1,$F8
	.dc.b	$43,$67,$E1,$0A,$FF,$59,$DF,$C6,$FE,$32,$37,$C7,$FA,$4E,$2F,$F1
	.dc.b	$C7,$C0,$7E,$8D,$B5,$F9,$78,$C2,$7F,$49,$F8,$94,$8F,$D2,$1F,$E4
	.dc.b	$17,$17,$E6,$3E,$29,$E9,$29,$FD,$E7,$9F,$F8,$0F,$E2,$9F,$C1,$1B
	.dc.b	$5F,$C0,$23,$1D,$1F,$E4,$D7,$F8,$A7,$4F,$F2,$CF,$C2,$AE,$1F,$E5
	.dc.b	$16,$D9,$A7,$F2,$48,$FE,$0A,$7D,$92,$9E,$58,$76,$3B,$BF,$35,$F5
	.dc.b	$7F,$6E,$31,$48,$C4,$EA,$BF,$99,$CE,$5F,$E9,$DE,$07,$E1,$22,$9F
	.dc.b	$83,$2F,$1F,$F5,$08,$D4,$7C,$93,$CA,$5E,$5F,$CB,$3F,$D6,$7C,$2D
	.dc.b	$BF,$D8,$5B,$7F,$D7,$E3,$B1,$FC,$3E,$37,$C7,$FD,$BE,$5F,$9F,$FF
	.dc.b	$6F,$6D,$7F,$6F,$6B,$FD,$BD,$AF,$F7,$CF,$1B,$E8,$3F,$DE,$E3,$E8
	.dc.b	$37,$6F,$F1,$7C,$EA,$3D,$5D,$FF,$E4,$E7,$E9,$65,$EC,$7F,$C1,$9E
	.dc.b	$47,$E5,$1F,$F0,$CB,$5C,$E1,$F9,$37,$2D,$7A,$F8,$1F,$E7,$37,$F1
	.dc.b	$E4,$CF,$FC,$8F,$E5,$73,$8F,$F0,$AB,$63,$35,$CE,$72,$D3,$FC,$7E
	.dc.b	$FF,$81,$BB,$BB,$3F,$E1,$16,$C6,$A3,$FC,$63,$E5,$FC,$17,$FC,$46
	.dc.b	$DA,$FF,$2A,$CF,$C8,$9F,$F1,$64,$ED,$FC,$09,$EA,$29,$3C,$DF,$F6
	.dc.b	$E4,$FF,$A6,$C9,$1F,$E7,$67,$FD,$CE,$47,$FC,$E8,$D7,$FC,$E7,$EA
	.dc.b	$FF,$37,$CB,$FD,$48,$48,$FF,$3F,$BF,$EF,$02,$3F,$E7,$D6,$BF,$E1
	.dc.b	$EF,$43,$FD,$25,$4F,$8E,$FF,$18,$C6,$1B,$9F,$FB,$1D,$B3,$BA,$FF
	.dc.b	$52,$35,$FF,$52,$35,$FF,$52,$7A,$5C,$5D,$CF,$FA,$B4,$8F,$FA,$CF
	.dc.b	$D5,$FF,$59,$46,$23,$DE,$3E,$3B,$FD,$84,$47,$FD,$3A,$FE,$27,$87
	.dc.b	$FC,$7E,$DE,$2F,$F5,$2B,$74,$E5,$5F,$E9,$F8,$FF,$6D,$65,$88,$FF
	.dc.b	$6C,$35,$FF,$5E,$7B,$1F,$EE,$79,$D9,$FE,$E6,$24,$7F,$BB,$9A,$FF
	.dc.b	$6B,$3F,$CA,$DE,$CF,$F6,$DF,$2B,$FD,$EA,$D8,$FF,$7E,$B5,$E1,$E8
	.dc.b	$7F,$BF,$67,$F4,$9E,$D3,$DB,$D5,$7F,$77,$8C,$3F,$DB,$0F,$03,$89
	.dc.b	$E3,$3C,$25,$EE,$7F,$95,$67,$F6,$0D,$8F,$F7,$23,$5F,$53,$5F,$F0
	.dc.b	$37,$33,$A3,$A3,$63,$C4,$6A,$32,$78,$7D,$E7,$BD,$49,$3C,$04,$69
	.dc.b	$C7,$4B,$A7,$D4,$D8,$CC,$60,$F7,$72,$7F,$A0,$33,$FC,$01,$27,$FD
	.dc.b	$BC,$FF,$0F,$F8,$07,$C0,$39,$3C,$51,$86,$C6,$67,$83,$6B,$F4,$31
	.dc.b	$FE,$FF,$9F,$57,$BE,$EA,$D5,$7E,$92,$BF,$8F,$4F,$FB,$DE,$B7,$EF
	.dc.b	$6A,$BF,$C6,$53,$9A,$DC,$E6,$52,$E7,$C3,$6F,$C3,$18,$33,$A4,$8F
	.dc.b	$2B,$70,$F1,$FC,$0F,$22,$4F,$A9,$F0,$7C,$13,$E0,$E9,$27,$EC,$46
	.dc.b	$5F,$93,$F0,$9E,$7D,$CD,$24,$70,$F8,$5E,$A4,$70,$7E,$0F,$E1,$9F
	.dc.b	$0F,$F0,$5B,$19,$8C,$1F,$17,$D5,$7C,$47,$F3,$EE,$1F,$F1,$13,$63
	.dc.b	$31,$FD,$09,$AF,$F5,$16,$BF,$B9,$38,$9B,$8E,$44,$70,$43,$93,$96
	.dc.b	$99,$3A,$47,$67,$F8,$33,$5F,$F3,$93,$5F,$09,$C8,$87,$28,$EC,$D3
	.dc.b	$B1,$D1,$CB,$4C,$9D,$9C,$B4,$C9,$E0,$3A,$1E,$03,$91,$4E,$05,$38
	.dc.b	$1A,$38,$90,$C1,$0E,$26,$4A,$64,$E6,$60,$86,$09,$1F,$1C,$91,$EC
	.dc.b	$8C,$B6,$33,$1E,$E6,$BE,$96,$BE,$33,$89,$B8,$E4,$47,$04,$39,$39
	.dc.b	$69,$93,$A4,$76,$7D,$CD,$7D,$AD,$7C,$27,$22,$1C,$A3,$B3,$4E,$C7
	.dc.b	$47,$2D,$32,$76,$72,$D3,$27,$80,$FD,$17,$BA,$F1,$BF,$9B,$47,$FA
	.dc.b	$EF,$83,$FC,$57,$39,$6C,$EB,$47,$C5,$FF,$B3,$7D,$EF,$F1,$F4,$63
	.dc.b	$0F,$F6,$29,$FE,$D1,$71,$1A,$9E,$13,$FB,$39,$CB,$63,$2F,$F6,$F8
	.dc.b	$CB,$63,$2F,$56,$BF,$C5,$1B,$1C,$23,$47,$90,$D1,$5D,$38,$9F,$33
	.dc.b	$99,$E6,$47,$04,$8D,$91,$F5,$19,$2B,$97,$D4,$64,$AE,$5F,$3B,$5F
	.dc.b	$21,$C8,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$09,$3F,$18
	.dc.b	$93,$F0,$E7,$CD,$18,$9F,$61,$C9,$C4,$6A,$78,$4F,$9A,$72,$D8,$CB
	.dc.b	$E5,$8C,$B6,$32,$F5,$6B,$E6,$6C,$70,$8D,$1E,$43,$45,$74,$E2,$7C
	.dc.b	$CE,$67,$99,$1C,$12,$36,$47,$D4,$64,$AE,$5F,$51,$92,$B9,$7C,$ED
	.dc.b	$7C,$87,$F5,$9D,$5F,$83,$7F,$87,$7C,$AF,$F5,$74,$63,$74,$62,$75
	.dc.b	$5C,$5A,$FF,$96,$EB,$FD,$DD,$5C,$23,$B7,$A7,$A4,$27,$77,$E0,$9D
	.dc.b	$4B,$8F,$FC,$11,$3F,$EC,$16,$C7,$F7,$16,$BF,$E6,$B6,$BD,$B8,$5E
	.dc.b	$47,$46,$E3,$E5,$9F,$F3,$94,$62,$FF,$D9,$CD,$8F,$EF,$DD,$3F,$E4
	.dc.b	$DA,$E1,$1F,$E3,$2E,$8F,$3C,$E6,$FF,$C7,$31,$DF,$69,$7C,$2F,$FC
	.dc.b	$2F,$18,$AF,$F1,$CC,$E6,$FF,$C8,$11,$82,$57,$F8,$3F,$E3,$7E,$3B
	.dc.b	$74,$F8,$F1,$88,$FF,$46,$35,$F9,$EF,$F8,$82,$FC,$4D,$AF,$F7,$63
	.dc.b	$67,$B7,$27,$D2,$EE,$79,$5F,$2A,$FF,$D4,$C4,$AF,$F1,$14,$EB,$6F
	.dc.b	$77,$83,$F4,$26,$C7,$E0,$2B,$FD,$F5,$39,$6C,$66,$7F,$C5,$93,$FE
	.dc.b	$7D,$9D,$F8,$59,$9F,$F5,$44,$E5,$DC,$65,$FF,$48,$FC,$2D,$5E,$2A
	.dc.b	$30,$D8,$FA,$8D,$BF,$F4,$FE,$BC,$7C,$6F,$F4,$05,$7D,$96,$CE,$A3
	.dc.b	$EE,$79,$0F,$BC,$F0,$36,$33,$18,$3C,$9F,$9C,$7F,$B0,$6B,$FC,$95
	.dc.b	$28,$C3,$FE,$4D,$6B,$FE,$C7,$6B,$FA,$AF,$2D,$F5,$5E,$F6,$C7,$E3
	.dc.b	$DB,$5D,$9B,$1F,$E5,$9E,$8F,$EE,$39,$BE,$3B,$F4,$3D,$1F,$E9,$F6
	.dc.b	$C6,$6B,$84,$E5,$FD,$2F,$91,$5F,$F6,$E7,$4F,$FA,$66,$FF,$3A,$48
	.dc.b	$FB,$BC,$9F,$C0,$A3,$F0,$31,$D4,$FB,$B1,$97,$66,$4F,$E1,$3C,$9F
	.dc.b	$EC,$19,$FC,$2C,$64,$FF,$62,$F2,$71,$7F,$0E,$D5,$E9,$3E,$19,$F0
	.dc.b	$DF,$EB,$DF,$8E,$FF,$3E,$4E,$63,$15,$FE,$F7,$8C,$6C,$AD,$DF,$28
	.dc.b	$C5,$7E,$A9,$AF,$ED,$3A,$F9,$B6,$32,$FE,$DA,$32,$FF,$26,$FB,$47
	.dc.b	$CF,$E2,$11,$8A,$E3,$18,$9D,$4E,$70,$F9,$BC,$12,$BE,$5B,$5F,$8A
	.dc.b	$1F,$96,$47,$F5,$A6,$4F,$85,$8F,$FA,$E3,$07,$E0,$1F,$0B,$C6,$FF
	.dc.b	$73,$36,$33,$5C,$27,$2F,$FA,$A7,$1F,$55,$7E,$89,$B1,$9C,$B6,$FF
	.dc.b	$48,$3D,$D7,$D1,$C7,$77,$F1,$1F,$1B,$F5,$6D,$8F,$E2,$BE,$07,$F4
	.dc.b	$29,$CB,$63,$A5,$3D,$B7,$D4,$FF,$BF,$EF,$FD,$7D,$39,$8C,$57,$0A
	.dc.b	$DC,$FF,$95,$8D,$1F,$E5,$8E,$1F,$C7,$61,$FD,$85,$B1,$FD,$29,$BD
	.dc.b	$7F,$8A,$9D,$6F,$D3,$2F,$C6,$E5,$E0,$CB,$FB,$4E,$1F,$99,$8C,$47
	.dc.b	$E7,$2B,$83,$F0,$39,$FF,$CA,$DA,$7C,$38,$F8,$F8,$FC,$16,$CF,$A7
	.dc.b	$47,$CB,$F4,$D7,$E2,$36,$33,$1B,$8F,$0D,$30,$EC,$FF,$7E,$6B,$9C
	.dc.b	$3C,$0E,$5B,$ED,$7F,$83,$BC,$4F,$5B,$70,$FF,$40,$B7,$6F,$F7,$83
	.dc.b	$6F,$EC,$FE,$20,$F6,$BF,$C1,$BF,$7B,$F6,$23,$12,$8C,$4E,$A3,$8B
	.dc.b	$5F,$F0,$AE,$0A,$FE,$57,$87,$3F,$0B,$FD,$0F,$39,$6C,$EA,$BE,$85
	.dc.b	$7E,$3B,$7F,$A3,$A7,$C0,$9F,$85,$1F,$3F,$D4,$F9,$84,$C7,$07,$E2
	.dc.b	$FA,$BF,$BC,$7E,$51,$FD,$4A,$FF,$8B,$76,$29,$39,$7F,$8A,$46,$63
	.dc.b	$15,$CA,$7C,$B5,$0F,$C0,$E5,$E7,$3F,$5C,$FF,$8A,$3E,$9D,$F2,$1F
	.dc.b	$A3,$AF,$D9,$66,$9C,$C2,$FC,$D3,$9A,$E1,$7F,$E4,$08,$C5,$6C,$FD
	.dc.b	$8E,$1F,$E1,$5F,$AF,$F3,$E7,$ED,$6C,$9A,$7C,$3B,$F9,$78,$FA,$5F
	.dc.b	$61,$F8,$DC,$3F,$D6,$DF,$85,$3F,$0A,$FF,$28,$F8,$3C,$63,$15,$FC
	.dc.b	$87,$1F,$83,$18,$9D,$57,$09,$FE,$0F,$F2,$EF,$73,$FD,$B1,$18,$6E
	.dc.b	$1F,$8A,$9D,$3E,$F3,$11,$A7,$ED,$7E,$11,$FE,$6F,$CF,$ED,$36,$FF
	.dc.b	$8A,$B6,$7F,$2A,$D7,$DD,$3F,$05,$FB,$67,$C2,$95,$F8,$A7,$F0,$CD
	.dc.b	$8F,$44,$F9,$4F,$C6,$46,$5E,$64,$3B,$9F,$38,$FA,$25,$EA,$FD,$F9
	.dc.b	$FC,$03,$DA,$FF,$80,$11,$FD,$BA,$FF,$9D,$46,$3C,$6F,$C4,$4F,$FA
	.dc.b	$12,$31,$3A,$F6,$5E,$D7,$FA,$22,$F7,$6B,$DB,$C6,$B6,$F7,$31,$5C
	.dc.b	$6F,$A3,$63,$F0,$C7,$ED,$C8,$7F,$29,$29,$FC,$98,$A7,$03,$E4,$9F
	.dc.b	$33,$4F,$E9,$6F,$DE,$3F,$85,$94,$FC,$C6,$3F,$E6,$77,$99,$FC,$10
	.dc.b	$EB,$9B,$4F,$F5,$94,$7E,$13,$DA,$FF,$29,$CF,$E1,$1B,$1F,$84,$6B
	.dc.b	$F6,$AF,$C6,$D3,$C6,$53,$D0,$60,$70,$7E,$93,$87,$F4,$31,$FE,$A8
	.dc.b	$3F,$B9,$46,$6F,$F4,$FC,$1F,$BD,$7E,$44,$C3,$E4,$14,$FE,$32,$BE
	.dc.b	$4F,$F7,$A4,$7F,$81,$F2,$7F,$1D,$8F,$E1,$7C,$FF,$C3,$AA,$F3,$FB
	.dc.b	$B2,$FE,$17,$18,$C3,$D5,$B7,$CD,$8F,$01,$F3,$4F,$14,$7F,$26,$1F
	.dc.b	$A0,$FF,$0E,$23,$E2,$32,$7F,$7F,$EA,$FE,$21,$3F,$CC,$7C,$34,$62
	.dc.b	$7F,$C1,$4D,$8C,$C7,$F9,$FA,$7C,$47,$C4,$21,$F8,$B3,$FA,$8F,$6F
	.dc.b	$F9,$E9,$FE,$A0,$7F,$A1,$A3,$FA,$83,$63,$4F,$F5,$A1,$FA,$17,$F5
	.dc.b	$E6,$99,$70,$7F,$6A,$F2,$3F,$D4,$95,$FE,$92,$F0,$FF,$B8,$56,$7C
	.dc.b	$4F,$7C,$62,$3E,$5E,$9F,$35,$F4,$9E,$C3,$E2,$78,$BF,$04,$97,$E9
	.dc.b	$CF,$FC,$13,$3F,$A0,$AE,$4F,$C1,$27,$91,$FE,$1D,$AF,$F5,$2B,$63
	.dc.b	$FC,$47,$18,$F1,$FF,$C6,$31,$FE,$29,$8F,$E0,$2F,$F8,$9A,$3F,$DB
	.dc.b	$31,$88,$FF,$13,$78,$9F,$DA,$7A,$F8,$1F,$E3,$EC,$79,$CF,$F8,$F7
	.dc.b	$C8,$FB,$44,$BF,$B5,$97,$A2,$7E,$01,$1C,$79,$5F,$E7,$5F,$D0,$FF
	.dc.b	$CB,$9E,$5F,$F7,$76,$D9,$A7,$F9,$9C,$97,$FE,$67,$8D,$36,$3F,$D3
	.dc.b	$FC,$D2,$3F,$CE,$AD,$7F,$D5,$27,$FB,$C4,$AF,$FA,$B8,$8F,$F9,$B6
	.dc.b	$5C,$1E,$42,$46,$5F,$F6,$AF,$B4,$EB,$FF,$08,$46,$1B,$3F,$E9,$F8
	.dc.b	$C5,$74,$8F,$F4,$3F,$89,$A7,$EA,$35,$FF,$2C,$37,$8B,$FD,$5B,$5F
	.dc.b	$EB,$CF,$88,$95,$FE,$92,$6B,$EE,$7F,$D9,$24,$7F,$D8,$EF,$FB,$B3
	.dc.b	$D3,$FE,$F8,$DF,$23,$FC,$50,$DC,$4D,$3F,$DB,$E4,$AF,$F6,$FC,$69
	.dc.b	$B1,$FE,$EB,$F3,$11,$F8,$D1,$EB,$7F,$BC,$B9,$75,$19,$8C,$3F,$EF
	.dc.b	$D3,$8C,$7F,$8C,$17,$AF,$FE,$9F,$8C,$36,$38,$1D,$FD,$97,$B3,$CA
	.dc.b	$3F,$C7,$31,$87,$51,$97,$FD,$55,$19,$AF,$07,$B5,$FE,$4B,$9F,$53
	.dc.b	$63,$39,$F1,$9F,$7C,$E5,$FF,$23,$FB,$6E,$ED,$42,$30,$FC,$56,$DF
	.dc.b	$BD,$B1,$FE,$CE,$8F,$F7,$D9,$1C,$1F,$E5,$4F,$65,$CD,$D9,$B1,$F4
	.dc.b	$5B,$3F,$2D,$B1,$FE,$A3,$8F,$F3,$A1,$1C,$1F,$EF,$77,$F3,$24,$70
	.dc.b	$7F,$99,$39,$9E,$E7,$F0,$98,$C3,$5F,$F6,$B4,$6B,$3F,$C2,$36,$BF
	.dc.b	$20,$49,$FF,$37,$F1,$FF,$94,$63,$11,$FE,$DF,$F7,$1E,$E7,$58,$FC
	.dc.b	$7C,$63,$0F,$C8,$36,$BF,$14,$D9,$EB,$EE,$F1,$58,$23,$0D,$BF,$CD
	.dc.b	$36,$F9,$E9,$C7,$0F,$A3,$5E,$08,$C1,$D7,$DE,$FE,$BF,$5F,$9B,$6E
	.dc.b	$3F,$E8,$C6,$BF,$EF,$B6,$BF,$AA,$F7,$FE,$33,$FE,$92,$6B,$FF,$00
	.dc.b	$5A,$EF,$5F,$64,$3D,$1A,$4C,$7F,$69,$19,$7F,$6D,$EE,$FC,$CB,$FE
	.dc.b	$E8,$D7,$F6,$4D,$7F,$D4,$6D,$7F,$7E,$1C,$3F,$C0,$9A,$E7,$6F,$F0
	.dc.b	$03,$4E,$EE,$23,$64,$70,$48,$D9,$0F,$2B,$96,$99,$3C,$EE,$5A,$64
	.dc.b	$F4,$C7,$77,$07,$A6,$38,$3A,$6B,$93,$67,$AC,$F0,$90,$EE,$53,$91
	.dc.b	$C2,$3C,$0E,$4E,$85,$3A,$1C,$88,$72,$34,$F2,$72,$72,$29,$C8,$E0
	.dc.b	$61,$E0,$E0,$87,$12,$9C,$4C,$10,$C1,$1E,$6E,$5A,$E6,$3E,$4B,$4F
	.dc.b	$84,$E2,$36,$47,$04,$8D,$90,$F8,$2E,$5A,$64,$F7,$39,$69,$93,$DB
	.dc.b	$1D,$DC,$1E,$98,$E0,$E9,$AE,$4D,$9E,$B3,$EC,$7A,$DF,$EB,$E7,$F4
	.dc.b	$A7,$FA,$D7,$D6,$FE,$C8,$DD,$3F,$CC,$53,$B8,$FC,$9F,$E1,$FF,$DA
	.dc.b	$BD,$7F,$D8,$B6,$3F,$62,$D7,$67,$FB,$08,$FD,$83,$96,$99,$3F,$6C
	.dc.b	$E5,$A6,$71,$46,$5B,$19,$8C,$11,$C1,$27,$CE,$D3,$CE,$53,$CE,$E8
	.dc.b	$8E,$9A,$E8,$C1,$0C,$10,$F6,$BC,$1A,$6C,$A6,$CF,$7B,$B6,$9B,$29
	.dc.b	$B9,$F8,$31,$B6,$C6,$DA,$EE,$7E,$24,$ED,$B1,$B6,$BB,$3E,$21,$E6
	.dc.b	$21,$E2,$2C,$F6,$6C,$72,$69,$C8,$E0,$48,$E0,$D8,$E0,$D3,$84,$68
	.dc.b	$91,$A6,$C6,$A3,$04,$70,$49,$F8,$0D,$3E,$01,$4F,$80,$E8,$8E,$9A
	.dc.b	$E8,$C1,$0C,$10,$F8,$2F,$06,$9B,$29,$FD,$9F,$EF,$19,$BC,$DF,$F2
	.dc.b	$76,$A9,$CC,$61,$B5,$FD,$C5,$AF,$F7,$7F,$42,$BF,$E6,$0E,$1F,$F6
	.dc.b	$ED,$7F,$7F,$9C,$CA,$72,$FF,$87,$BD,$7C,$F2,$9A,$4C,$BC,$D3,$FE
	.dc.b	$18,$6B,$FD,$C2,$7F,$C2,$D8,$7F,$9F,$79,$B9,$9D,$3E,$16,$5F,$0B
	.dc.b	$FF,$3A,$DA,$73,$7F,$E8,$12,$4F,$F8,$CB,$C3,$F2,$9E,$6F,$86,$98
	.dc.b	$73,$9F,$F1,$4C,$62,$BF,$D3,$E4,$8F,$F2,$37,$91,$F0,$BA,$FF,$D5
	.dc.b	$17,$C7,$CC,$79,$FF,$E5,$AC,$3F,$D0,$11,$8E,$CD,$4E,$71,$FE,$57
	.dc.b	$39,$AF,$F4,$37,$C4,$7F,$8E,$67,$C8,$D8,$CE,$FF,$E9,$46,$CF,$A7
	.dc.b	$BD,$F0,$79,$8F,$F3,$63,$5F,$F0,$EC,$69,$E1,$19,$7F,$D3,$FE,$0B
	.dc.b	$E2,$71,$1F,$E7,$F6,$C6,$63,$E4,$36,$FF,$D1,$71,$F2,$FC,$4F,$F4
	.dc.b	$84,$60,$91,$FE,$9A,$F8,$AE,$BE,$31,$A4,$8F,$A9,$E4,$7F,$8D,$2B
	.dc.b	$FD,$4F,$1E,$48,$C3,$FE,$A8,$8F,$CB,$70,$BE,$1A,$92,$7F,$CA,$DD
	.dc.b	$7F,$A8,$8F,$F6,$57,$55,$9F,$CD,$57,$16,$CE,$A3,$11,$F9,$7E,$8E
	.dc.b	$0D,$8C,$BF,$94,$8C,$B6,$32,$FE,$39,$BD,$4E,$9D,$CE,$6F,$FD,$BD
	.dc.b	$18,$E2,$FB,$11,$FE,$A1,$E2,$FF,$6A,$B6,$BA,$CE,$9F,$F5,$CF,$16
	.dc.b	$9A,$78,$B8,$BE,$D5,$7F,$AD,$E7,$53,$FE,$62,$F1,$5F,$33,$FC,$3B
	.dc.b	$E0,$FF,$DB,$51,$8D,$51,$89,$D5,$7F,$81,$AB,$53,$FA,$2E,$9F,$F5
	.dc.b	$EB,$7A,$B8,$D7,$E8,$A7,$2D,$8C,$BF,$E7,$4F,$51,$E3,$FF,$BD,$2F
	.dc.b	$FD,$D9,$18,$9D,$9F,$EF,$87,$11,$A9,$E1,$3F,$A4,$1C,$5F,$24,$8E
	.dc.b	$5F,$F7,$C7,$9A,$8F,$E5,$FF,$75,$86,$9F,$E2,$48,$D6,$A9,$D3,$67
	.dc.b	$FD,$1A,$BE,$69,$EF,$7F,$3A,$F8,$9F,$F1,$6C,$E6,$31,$5C,$2B,$72
	.dc.b	$AD,$C6,$1F,$74,$E4,$FF,$1B,$E9,$F1,$63,$24,$8F,$9D,$E4,$7D,$32
	.dc.b	$38,$3C,$FF,$0C,$FB,$6F,$F1,$96,$28,$C4,$FF,$8D,$E3,$13,$A9,$CD
	.dc.b	$7B,$3E,$03,$E7,$DE,$DB,$19,$DB,$AC,$62,$7E,$3B,$FA,$57,$87,$56
	.dc.b	$88,$FF,$90,$FC,$5F,$E5,$D7,$EA,$8F,$F5,$DC,$61,$F6,$1F,$5B,$0F
	.dc.b	$E5,$D3,$F4,$1B,$1F,$03,$C4,$F5,$1F,$E4,$DA,$E5,$1D,$39,$5C,$BE
	.dc.b	$16,$E1,$E1,$97,$AE,$FA,$DF,$EC,$91,$FE,$5D,$8C,$78,$1F,$E5,$D6
	.dc.b	$C6,$67,$75,$F2,$7E,$9F,$F2,$5F,$85,$DA,$34,$8F,$C8,$FA,$9F,$96
	.dc.b	$47,$E8,$3E,$BC,$BC,$F1,$F1,$0F,$CF,$69,$EE,$6C,$69,$F9,$1B,$FD
	.dc.b	$D2,$3F,$81,$DF,$EB,$65,$FE,$86,$FC,$3D,$F1,$CF,$98,$FF,$0B,$E9
	.dc.b	$FE,$18,$C3,$F5,$95,$A9,$CF,$37,$FB,$CA,$30,$FB,$8C,$5F,$37,$15
	.dc.b	$F4,$F8,$BE,$91,$31,$FA,$4B,$F2,$0F,$F0,$34,$A7,$C4,$FD,$A8,$CC
	.dc.b	$62,$78,$7A,$3F,$E9,$9B,$FE,$05,$7B,$8C,$3E,$AC,$3C,$9E,$4F,$4C
	.dc.b	$BD,$2D,$8C,$DF,$AB,$3F,$F1,$CC,$F8,$AD,$C1,$F4,$5B,$C1,$F4,$5F
	.dc.b	$A7,$9F,$B5,$F8,$DF,$8B,$7C,$BB,$F0,$37,$1C,$9F,$C6,$63,$2D,$3F
	.dc.b	$89,$FC,$83,$C8,$F5,$DF,$F9,$32,$30,$FC,$79,$DC,$E9,$4E,$B7,$F9
	.dc.b	$71,$F3,$5F,$F2,$A9,$83,$91,$4C,$BF,$C9,$1C,$BF,$93,$AF,$83,$8F
	.dc.b	$BD,$FB,$D3,$FC,$9A,$BC,$15,$E4,$9F,$31,$FC,$CA,$32,$DB,$F6,$BD
	.dc.b	$2B,$F6,$CD,$3C,$19,$FD,$12,$67,$F4,$63,$47,$E0,$BD,$9F,$EA,$7E
	.dc.b	$D7,$F4,$5F,$83,$FC,$64,$FF,$47,$9C,$A8,$CB,$FE,$E7,$31,$87,$3F
	.dc.b	$19,$7C,$E3,$0D,$9E,$93,$A7,$89,$88,$D3,$D6,$BB,$E3,$FE,$51,$8F
	.dc.b	$EE,$04,$3F,$B7,$14,$FD,$51,$4F,$C7,$9F,$3F,$2F,$EB,$2F,$CB,$3F
	.dc.b	$AF,$94,$FC,$0E,$7F,$88,$7F,$39,$3E,$17,$B9,$0E,$E7,$E0,$71,$FE
	.dc.b	$09,$A5,$3A,$C7,$E2,$A3,$2F,$F6,$E7,$EF,$C7,$E0,$63,$C3,$7E,$67
	.dc.b	$F3,$4D,$8F,$CB,$B9,$E2,$70,$7F,$13,$FB,$FF,$D2,$D9,$A4,$BF,$44
	.dc.b	$7C,$42,$9F,$96,$3F,$90,$71,$FF,$15,$9F,$E0,$1C,$7F,$EB,$66,$D7
	.dc.b	$53,$F8,$11,$6B,$FC,$1D,$C5,$FC,$41,$F8,$2D,$7F,$8A,$72,$FF,$1A
	.dc.b	$6F,$2F,$BD,$F1,$1E,$4D,$BE,$83,$F3,$8F,$E2,$D8,$FF,$B2,$3F,$0B
	.dc.b	$F1,$93,$1F,$1B,$FB,$AE,$8F,$EE,$15,$FD,$97,$A7,$F9,$0B,$FE,$04
	.dc.b	$6B,$FD,$10,$A7,$88,$F0,$1F,$1C,$B4,$F0,$1D,$7C,$23,$0D,$8C,$CF
	.dc.b	$69,$CB,$61,$D9,$FE,$35,$C7,$FB,$21,$27,$FC,$13,$19,$7F,$9C,$F7
	.dc.b	$7F,$6E,$7D,$6D,$73,$DD,$FD,$D9,$CB,$5F,$99,$D9,$FE,$00,$7F,$4C
	.dc.b	$7B,$CF,$F7,$9E,$9F,$EB,$97,$FC,$19,$E0,$7B,$63,$11,$FE,$0D,$3E
	.dc.b	$3E,$FF,$59,$F8,$C4,$3A,$69,$F1,$F6,$F8,$84,$8F,$84,$FC,$7F,$06
	.dc.b	$47,$C5,$6B,$FE,$1D,$F0,$BC,$44,$7E,$66,$1F,$E3,$A7,$D7,$87,$8D
	.dc.b	$F6,$13,$C4,$FF,$37,$7C,$37,$F9,$38,$91,$FE,$4E,$F0,$3F,$CC,$D1
	.dc.b	$FE,$54,$F1,$DA,$23,$0D,$AF,$F2,$E4,$EA,$72,$7F,$9A,$3B,$24,$FF
	.dc.b	$9B,$4E,$6E,$1F,$F6,$DF,$A1,$FA,$87,$FC,$F9,$D5,$FE,$A3,$23,$E5
	.dc.b	$3F,$C6,$65,$E3,$FF,$33,$91,$FF,$48,$3F,$E9,$52,$38,$3F,$D3,$9E
	.dc.b	$6F,$B2,$3F,$D4,$19,$74,$9F,$F5,$44,$E5,$DC,$65,$FF,$76,$F9,$DF
	.dc.b	$E8,$BF,$85,$FF,$58,$41,$A4,$7F,$D2,$CF,$FA,$F4,$8E,$0F,$D9,$FA
	.dc.b	$5F,$EC,$9A,$E5,$18,$E4,$FA,$35,$FE,$CA,$9C,$CF,$F8,$6A,$B8,$4F
	.dc.b	$EE,$9F,$F5,$1C,$EA,$7F,$DB,$13,$9E,$4E,$71,$E0,$6B,$CF,$CA,$AF
	.dc.b	$FA,$9E,$D1,$88,$46,$2F,$B4,$62,$75,$ED,$BC,$AA,$4A,$DC,$65,$E3
	.dc.b	$EA,$FF,$8D,$3C,$AF,$E2,$6D,$C3,$F8,$BB,$5F,$F1,$CB,$5F,$F5,$E7
	.dc.b	$AE,$F5,$FF,$D7,$D5,$FC,$72,$31,$3F,$DB,$CF,$F2,$0B,$88,$D4,$F0
	.dc.b	$9F,$E4,$BE,$C7,$F9,$76,$7F,$92,$B6,$33,$86,$DF,$F2,$5C,$65,$FF
	.dc.b	$26,$F9,$4E,$1D,$47,$E2,$0F,$11,$6F,$F1,$27,$FC,$01,$23,$FE,$DD
	.dc.b	$6B,$FE,$58,$F6,$5D,$FD,$21,$18,$7A,$98,$8D,$46,$67,$FD,$F0,$49
	.dc.b	$FF,$7B,$35,$FF,$32,$FB,$4F,$6F,$F8,$0C,$FF,$46,$8C,$67,$D9,$B1
	.dc.b	$99,$DD,$7E,$66,$BC,$7C,$5F,$9D,$AF,$CD,$12,$7F,$25,$3A,$E3,$E3
	.dc.b	$1F,$9D,$6B,$F9,$36,$BF,$92,$F2,$6D,$9A,$4A,$F0,$E7,$A2,$38,$3C
	.dc.b	$BC,$9C,$88,$E0,$F9,$7C,$7F,$97,$CB,$FB,$3B,$71,$FE,$D6,$D7,$FD
	.dc.b	$1A,$D4,$3D,$D7,$BB,$EA,$6D,$FD,$86,$DF,$F7,$46,$D7,$F7,$8F,$7D
	.dc.b	$EF,$FF,$76,$84,$61,$FE,$F2,$D7,$FB,$EB,$5E,$51,$C6,$31,$1F,$EA
	.dc.b	$46,$BE,$86,$BD,$A3,$81,$1E,$ED,$7B,$B5,$EE,$68,$86,$8A,$68,$C1
	.dc.b	$0C,$10,$F3,$3D,$5A,$72,$29,$C4,$A7,$13,$D8,$F1,$69,$C4,$A7,$12
	.dc.b	$9C,$4F,$84,$6C,$A6,$CA,$6E,$3E,$49,$23,$C1,$1A,$3C,$A4,$3B,$94
	.dc.b	$D1,$82,$18,$24,$75,$7C,$8D,$7C,$8D,$7C,$91,$E2,$8C,$47,$76,$BD
	.dc.b	$9A,$F6,$8E,$04,$7B,$B5,$EE,$D7,$B9,$A2,$1A,$29,$A3,$04,$30,$43
	.dc.b	$CC,$F5,$69,$C8,$A7,$12,$9C,$4F,$63,$FA,$9F,$57,$FD,$7B,$1F,$E6
	.dc.b	$46,$BF,$EB,$83,$FC,$CA,$E1,$FD,$27,$DE,$7E,$F3,$D4,$FD,$6B,$63
	.dc.b	$F6,$CD,$7F,$58,$D7,$89,$FE,$CA,$2C,$72,$70,$7E,$F2,$38,$34,$CD
	.dc.b	$A3,$2D,$8C,$C6,$08,$E0,$93,$DC,$8F,$A0,$C9,$5C,$BE,$A3,$25,$72
	.dc.b	$FA,$8E,$04,$34,$53,$46,$08,$60,$87,$BD,$EA,$D3,$91,$4E,$25,$38
	.dc.b	$CF,$C3,$8E,$2D,$8E,$2D,$78,$B5,$E3,$3F,$22,$78,$B6,$38,$B5,$E2
	.dc.b	$D7,$89,$F1,$8B,$1C,$9C,$1E,$72,$4F,$85,$B3,$CD,$A7,$08,$D1,$23
	.dc.b	$4D,$8D,$46,$08,$E0,$93,$E3,$23,$EC,$32,$57,$2F,$B8,$C9,$5C,$BE
	.dc.b	$E3,$81,$0D,$14,$D1,$82,$1F,$D9,$B0,$7C,$17,$FA,$43,$E0,$7F,$B1
	.dc.b	$36,$7F,$DB,$33,$AD,$15,$A6,$CF,$F9,$B3,$7E,$0F,$F9,$76,$32,$D9
	.dc.b	$FF,$34,$4E,$A3,$FB,$BF,$CC,$3C,$DC,$42,$31,$7C,$A3,$73,$A9,$CD
	.dc.b	$FF,$9C,$F8,$F8,$F2,$7F,$85,$70,$E1,$87,$F8,$65,$B1,$9A,$FF,$79
	.dc.b	$DB,$E2,$BA,$12,$B8,$C6,$5E,$3E,$17,$62,$38,$37,$F3,$3E,$52,$38
	.dc.b	$3C,$7E,$6F,$BF,$C7,$BA,$9A,$4C,$BF,$C3,$7A,$7F,$94,$FA,$7F,$C3
	.dc.b	$B1,$88,$FF,$0F,$F6,$FF,$94,$E3,$11,$FE,$53,$3E,$0C,$65,$E0,$FF
	.dc.b	$8E,$5C,$EF,$D1,$CB,$5F,$F3,$13,$5F,$F3,$64,$FF,$AC,$88,$F8,$FB
	.dc.b	$FF,$CE,$4F,$D0,$F8,$47,$C2,$F1,$84,$61,$D4,$66,$31,$5F,$E8,$69
	.dc.b	$FB,$3E,$27,$FA,$32,$30,$48,$FF,$44,$FC,$97,$5F,$10,$D2,$47,$E0
	.dc.b	$BE,$BF,$CB,$5F,$19,$B1,$F2,$DA,$FD,$06,$BF,$EB,$5E,$7F,$F7,$AF
	.dc.b	$C3,$FE,$6B,$8C,$D2,$4F,$D5,$E6,$FC,$CC,$7E,$3B,$8F,$8C,$E5,$AF
	.dc.b	$DE,$9D,$BF,$E7,$B9,$FF,$69,$11,$C1,$D7,$97,$8B,$FA,$A6,$BF,$A9
	.dc.b	$6C,$9C,$F0,$7D,$6A,$FF,$75,$92,$BE,$D7,$8D,$FE,$6A,$8F,$C5,$F0
	.dc.b	$7D,$76,$F0,$E6,$BF,$25,$39,$6C,$65,$FC,$CF,$E2,$5F,$9D,$D1,$F0
	.dc.b	$FF,$96,$8C,$37,$9D,$3A,$AF,$CF,$56,$9F,$CC,$F3,$FF,$9F,$AF,$F3
	.dc.b	$FC,$FF,$EB,$36,$C6,$67,$67,$33,$FD,$0B,$6F,$69,$F6,$FB,$9C,$C6
	.dc.b	$1E,$07,$91,$C5,$FE,$C6,$B9,$57,$1A,$F4,$FD,$8B,$D0,$F5,$FC,$46
	.dc.b	$E3,$16,$8C,$4E,$AB,$F7,$F2,$AD,$D7,$7E,$0F,$3C,$F1,$AD,$C6,$27
	.dc.b	$58,$73,$9F,$7C,$E5,$EB,$93,$8B,$FD,$4F,$F0,$FE,$B8,$C4,$A3,$13
	.dc.b	$A8,$E2,$D7,$E3,$DD,$7E,$33,$71,$F8,$B1,$CC,$F0,$C6,$D7,$77,$F2
	.dc.b	$3F,$13,$F8,$CB,$63,$F8,$F3,$5F,$E4,$6D,$5F,$1E,$F9,$2F,$CF,$61
	.dc.b	$FD,$1E,$31,$D1,$FD,$3D,$B1,$99,$DF,$6F,$F2,$B9,$E5,$F3,$BF,$92
	.dc.b	$CB,$14,$8C,$63,$FC,$D2,$B7,$87,$F9,$35,$B7,$F2,$FE,$83,$FC,$A5
	.dc.b	$3F,$8F,$6E,$E6,$9F,$E5,$0D,$3D,$98,$7C,$4E,$DF,$E8,$98,$7C,$46
	.dc.b	$C7,$C4,$6B,$F9,$8F,$90,$EA,$47,$FC,$C1,$8F,$E7,$E3,$B4,$F0,$9F
	.dc.b	$F7,$3D,$F8,$09,$19,$CB,$E6,$7C,$AF,$C9,$D4,$D2,$63,$F2,$FF,$0E
	.dc.b	$FD,$CF,$D0,$FC,$86,$E5,$F1,$ED,$18,$8F,$F7,$BE,$1F,$66,$31,$1F
	.dc.b	$C0,$BF,$43,$7C,$7F,$CA,$6C,$7B,$1B,$B7,$EC,$DB,$1F,$17,$BF,$FC
	.dc.b	$FB,$B7,$B1,$BB,$7C,$1D,$57,$FE,$FC,$BE,$9E,$87,$72,$3F,$E8,$6F
	.dc.b	$9D,$7D,$27,$F4,$2B,$FE,$E5,$18,$7F,$A1,$98,$8D,$4F,$F7,$A9,$D7
	.dc.b	$5F,$F1,$E8,$FC,$DE,$3F,$4D,$FC,$06,$5E,$62,$69,$FD,$FF,$E0,$59
	.dc.b	$BE,$5B,$FC,$71,$A7,$F8,$1A,$73,$18,$C0,$23,$33,$FE,$3A,$38,$57
	.dc.b	$37,$13,$E7,$D3,$F9,$9F,$B5,$E7,$8E,$7A,$FB,$E3,$10,$8C,$57,$C1
	.dc.b	$13,$AD,$FE,$96,$7F,$9B,$D3,$DB,$87,$F2,$0D,$FF,$C6,$4B,$E1,$FC
	.dc.b	$18,$FD,$6F,$4B,$D7,$18,$AD,$B6,$33,$3C,$FB,$7F,$D9,$0D,$B6,$3C
	.dc.b	$67,$31,$FD,$13,$0F,$EC,$CE,$AB,$F9,$9F,$CF,$FF,$94,$AF,$C6,$D8
	.dc.b	$CF,$17,$F0,$76,$9C,$6B,$EE,$E7,$EE,$6D,$FC,$13,$F3,$18,$F8,$A3
	.dc.b	$F4,$0D,$AF,$CA,$B7,$1F,$2D,$7C,$BC,$3F,$11,$A7,$CC,$C3,$F4,$13
	.dc.b	$D5,$A4,$3C,$27,$FB,$C7,$E1,$DE,$DF,$E1,$B9,$BE,$2C,$FE,$23,$DA
	.dc.b	$56,$A3,$F6,$77,$FE,$F7,$AD,$C2,$B7,$7C,$EB,$44,$9E,$97,$CE,$F8
	.dc.b	$DF,$09,$EA,$7F,$14,$C7,$FC,$C8,$FE,$1C,$87,$EF,$4A,$7C,$E2,$9C
	.dc.b	$75,$FF,$31,$EB,$F2,$A3,$FC,$02,$7D,$FD,$7F,$B7,$B4,$C9,$F4,$8A
	.dc.b	$7E,$D0,$EA,$53,$C4,$7E,$00,$FC,$2D,$7E,$F6,$70,$48,$EC,$7E,$44
	.dc.b	$EE,$F9,$5C,$1F,$9B,$3F,$CB,$6F,$E9,$77,$FD,$DF,$A5,$92,$47,$8A
	.dc.b	$7E,$11,$4F,$84,$7F,$A1,$38,$FF,$8B,$63,$FB,$86,$CF,$ED,$E3,$1C
	.dc.b	$9F,$E3,$27,$F8,$07,$37,$F0,$8E,$0B,$CD,$FC,$1D,$BC,$FF,$E3,$86
	.dc.b	$E1,$F8,$C3,$C5,$97,$C7,$3C,$A4,$3F,$85,$E3,$FE,$40,$F8,$3F,$0C
	.dc.b	$9A,$47,$F8,$36,$1F,$E9,$EA,$FE,$8C,$D8,$FE,$76,$DE,$5F,$E7,$93
	.dc.b	$F6,$39,$7F,$D6,$0D,$E0,$FC,$6B,$FC,$47,$8F,$81,$1F,$C4,$D7,$F5
	.dc.b	$07,$FA,$E6,$0F,$C3,$A9,$23,$FA,$F7,$5F,$F9,$6A,$3F,$C1,$3D,$FF
	.dc.b	$16,$7F,$B1,$35,$FE,$C3,$DD,$FD,$B2,$FF,$5D,$18,$9D,$3F,$68,$F9
	.dc.b	$79,$FA,$9F,$BD,$3F,$DF,$5F,$CF,$F8,$4F,$07,$E2,$36,$33,$1F,$25
	.dc.b	$FA,$87,$53,$EA,$18,$9F,$F0,$C3,$D3,$C2,$E6,$47,$FD,$9F,$E0,$FF
	.dc.b	$88,$9F,$F1,$3F,$87,$FD,$11,$FF,$1A,$E7,$FE,$29,$BD,$B6,$3F,$C5
	.dc.b	$BD,$7E,$EB,$FF,$1A,$F2,$E2,$FE,$1D,$72,$AF,$86,$FA,$09,$A7,$C3
	.dc.b	$7F,$19,$3F,$E5,$0F,$87,$FF,$2E,$78,$DF,$D8,$9B,$26,$9F,$E5,$E7
	.dc.b	$07,$F9,$78,$A7,$FA,$13,$C7,$E4,$E6,$3F,$CD,$F5,$FD,$B6,$3F,$CE
	.dc.b	$75,$FE,$9A,$EA,$C9,$23,$FC,$FB,$E5,$7F,$A0,$A3,$FD,$03,$5C,$67
	.dc.b	$FD,$2B,$39,$77,$E5,$F3,$7F,$CD,$23,$E1,$BD,$64,$72,$FF,$A8,$9D
	.dc.b	$1F,$EA,$27,$04,$7F,$D5,$BD,$DF,$EB,$08,$FF,$59,$5F,$09,$FF,$5C
	.dc.b	$CE,$5D,$C6,$5F,$F0,$6F,$A5,$FE,$FB,$ED,$FF,$62,$C7,$EE,$4F,$F6
	.dc.b	$19,$4F,$F0,$98,$8E,$0F,$FB,$40,$D1,$FE,$18,$D5,$E9,$FF,$86,$AB
	.dc.b	$8C,$E6,$30,$F1,$8C,$BF,$EE,$1B,$E6,$D8,$CF,$A9,$FE,$86,$6F,$0F
	.dc.b	$F1,$19,$EB,$ED,$BD,$BC,$47,$A2,$31,$87,$46,$DF,$81,$B3,$FE,$FD
	.dc.b	$9F,$0E,$0F,$03,$9C,$FF,$8D,$23,$13,$A9,$CC,$A9,$EB,$E1,$B5,$E1
	.dc.b	$8C,$1C,$3D,$8F,$8F,$5F,$A8,$0E,$39,$FE,$2B,$63,$E5,$B5,$FA,$BE
	.dc.b	$CB,$C8,$FF,$24,$D7,$CC,$8C,$4F,$56,$C6,$67,$73,$F5,$BD,$A7,$C2
	.dc.b	$62,$3F,$03,$18,$7C,$CD,$AF,$B2,$D8,$FC,$4C,$7E,$30,$8E,$0F,$C4
	.dc.b	$3F,$80,$E2,$F5,$47,$E1,$4F,$84,$E1,$7B,$2F,$73,$C9,$1F,$E6,$68
	.dc.b	$C3,$F2,$5B,$5F,$8C,$6C,$7E,$49,$7B,$5C,$A3,$AB,$63,$31,$CE,$35
	.dc.b	$DD,$F8,$3E,$8F,$AC,$DB,$FF,$32,$12,$BF,$3F,$C7,$FE,$DC,$8F,$CB
	.dc.b	$F0,$FD,$D8,$C4,$EB,$9B,$F9,$2F,$1F,$E5,$AF,$F2,$9C,$D8,$26,$1C
	.dc.b	$7C,$EF,$E5,$78,$9A,$48,$FC,$55,$FC,$76,$ED,$D8,$8F,$8F,$9B,$99
	.dc.b	$1C,$1C,$B9,$BC,$A4,$70,$7E,$5F,$DD,$7B,$DF,$CF,$9B,$19,$8C,$1F
	.dc.b	$D0,$BC,$D7,$BF,$FD,$6D,$BC,$1F,$E0,$56,$BF,$C1,$1A,$FF,$0A,$6B
	.dc.b	$FC,$51,$AF,$FA,$A1,$AF,$FA,$C2,$3B,$11,$EC,$D7,$B3,$5E,$CD,$7B
	.dc.b	$35,$EC,$D7,$B1,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0E
	.dc.b	$45,$39,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34
	.dc.b	$60,$86,$09,$1F,$50,$91,$F2,$5B,$1F,$01,$AF,$A5,$AF,$91,$AF,$85
	.dc.b	$AF,$81,$AF,$82,$3B,$11,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B1
	.dc.b	$D0,$87,$42,$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0F,$F6,$16,$FF,$E6
	.dc.b	$BF,$4B,$FD,$65,$18,$3F,$B5,$B8,$7F,$4D,$F3,$67,$FD,$82,$7E,$9D
	.dc.b	$C7,$AB,$FA,$F6,$C7,$EB,$DA,$F2,$38,$8C,$11,$A8,$E8,$6A,$32,$D8
	.dc.b	$D3,$C5,$B1,$A9,$F1,$4E,$C8,$E0,$8F,$94,$C9,$5C,$BE,$76,$BE,$46
	.dc.b	$BE,$13,$89,$B3,$91,$0C,$10,$E4,$64,$A6,$4E,$8F,$68,$F6,$46,$CC
	.dc.b	$95,$CB,$C5,$AF,$42,$1C,$0D,$4F,$C8,$23,$EA,$37,$1C,$9E,$06,$23
	.dc.b	$73,$E5,$6C,$78,$DA,$F6,$3A,$1C,$89,$3D,$49,$1C,$23,$B1,$C2,$32
	.dc.b	$D8,$D3,$C9,$B1,$A9,$F2,$4E,$C8,$E0,$8F,$9C,$C9,$5C,$BE,$96,$BE
	.dc.b	$46,$BE,$13,$89,$B3,$91,$0C,$10,$E4,$64,$A7,$F9,$9B,$37,$C2,$34
	.dc.b	$7B,$9F,$E5,$E9,$CC,$62,$BF,$CC,$B1,$89,$D5,$71,$7F,$CA,$DD,$16
	.dc.b	$BF,$CB,$ED,$7F,$CB,$CD,$7F,$7D,$C5,$FE,$1E,$C3,$B4,$7F,$98,$F8
	.dc.b	$B4,$FF,$95,$DB,$3F,$E0,$7B,$FF,$32,$46,$3D,$A9,$7C,$AF,$AC,$62
	.dc.b	$BF,$C2,$F3,$99,$FF,$31,$73,$74,$24,$67,$E1,$7D,$D3,$FE,$84,$6B
	.dc.b	$FE,$8E,$6B,$FC,$5D,$AB,$A7,$FC,$65,$7F,$E8,$86,$C7,$FA,$39,$AF
	.dc.b	$FA,$61,$AF,$FA,$83,$AB,$F9,$14,$FF,$A4,$3C,$44,$62,$7F,$D2,$F1
	.dc.b	$87,$D8,$D9,$E1,$F1,$1B,$9F,$F5,$2C,$A3,$11,$FE,$AE,$6B,$FE,$A9
	.dc.b	$6F,$77,$FA,$02,$BF,$CC,$5E,$2E,$09,$5F,$2F,$C3,$FC,$53,$74,$E7
	.dc.b	$18,$AF,$F3,$F4,$76,$F0,$7F,$CE,$D1,$FE,$74,$F0,$B0,$4B,$FA,$DF
	.dc.b	$0F,$FE,$BC,$9F,$F5,$8B,$63,$FD,$6E,$D7,$FD,$82,$D7,$FA,$57,$8B
	.dc.b	$E9,$7F,$D2,$79,$7B,$A3,$11,$FE,$C8,$6B,$FD,$47,$9F,$F4,$3E,$F7
	.dc.b	$E7,$B3,$34,$93,$F9,$E8,$CB,$63,$F4,$71,$88,$FD,$0B,$5F,$D0,$F2
	.dc.b	$FE,$C6,$7F,$37,$39,$9F,$D0,$37,$9B,$FC,$CD,$3F,$ED,$2E,$4F,$2D
	.dc.b	$7E,$57,$A3,$93,$6F,$94,$EA,$3F,$1D,$E0,$FE,$2A,$FF,$28,$D8,$FC
	.dc.b	$A3,$5D,$F2,$F8,$F8,$BE,$ED,$7D,$A9,$C6,$1F,$DD,$E3,$07,$AB,$E0
	.dc.b	$A4,$FF,$B8,$34,$FE,$1D,$18,$8F,$EF,$D4,$F6,$1E,$97,$F8,$06,$BF
	.dc.b	$DE,$73,$98,$C5,$7F,$BF,$A3,$0A,$77,$5C,$A3,$07,$E9,$B9,$F8,$BF
	.dc.b	$A2,$6B,$FA,$13,$66,$4A,$7F,$84,$FE,$29,$F8,$A7,$08,$46,$1D,$98
	.dc.b	$8D,$46,$6F,$FC,$39,$32,$1F,$0F,$FC,$C7,$E1,$73,$3D,$E3,$0E,$27
	.dc.b	$79,$2B,$73,$E2,$AD,$BD,$BC,$CF,$F7,$EB,$B3,$FD,$99,$E4,$FF,$8B
	.dc.b	$6B,$FD,$A0,$D8,$C9,$B3,$CD,$5F,$D1,$7E,$8B,$BE,$3E,$26,$C6,$67
	.dc.b	$CC,$D7,$C5,$5F,$E3,$B9,$EF,$06,$91,$FC,$8F,$17,$F8,$E3,$C0,$FF
	.dc.b	$6D,$78,$9F,$96,$8C,$72,$FF,$22,$8F,$8F,$EB,$7F,$B9,$BA,$3F,$AD
	.dc.b	$7D,$9E,$3C,$6F,$EB,$2D,$8C,$E5,$F4,$AB,$87,$0A,$B9,$FB,$0C,$7F
	.dc.b	$AB,$52,$31,$5F,$D3,$63,$13,$A9,$CF,$83,$F9,$2F,$7B,$FC,$B5,$F3
	.dc.b	$77,$23,$48,$FC,$0B,$F7,$11,$C1,$F9,$FF,$AA,$FF,$75,$3C,$8C,$7A
	.dc.b	$3D,$9A,$64,$FD,$6C,$70,$AE,$91,$FB,$CA,$AF,$F9,$BF,$87,$E7,$BF
	.dc.b	$17,$EB,$BF,$A7,$EB,$F1,$63,$1C,$5F,$29,$FF,$3B,$6C,$AD,$EF,$F5
	.dc.b	$1B,$3F,$B7,$EC,$F9,$8B,$F4,$4F,$DB,$E7,$FE,$7F,$6C,$64,$E1,$19
	.dc.b	$7E,$C7,$0B,$E4,$3C,$50,$CB,$EB,$46,$1B,$A7,$A6,$7C,$F5,$C6,$31
	.dc.b	$1F,$07,$97,$F1,$8F,$D0,$3E,$31,$4E,$D9,$BE,$3B,$B3,$6B,$9E,$1F
	.dc.b	$0E,$72,$79,$A3,$FC,$5B,$EA,$7F,$8B,$FE,$B7,$FB,$3E,$1F,$E2,$98
	.dc.b	$C4,$FF,$1D,$E6,$46,$27,$F9,$3C,$62,$55,$CA,$7F,$C2,$19,$7A,$9F
	.dc.b	$49,$0F,$77,$0F,$FA,$B0,$7E,$3E,$F8,$7F,$D4,$E1,$FC,$5A,$73,$18
	.dc.b	$AF,$F5,$A4,$62,$7F,$50,$6E,$B9,$38,$9F,$E7,$BC,$4F,$C1,$BF,$D7
	.dc.b	$9F,$A2,$FC,$0A,$FE,$81,$18,$7F,$0C,$DA,$E2,$D7,$D3,$C7,$F9,$87
	.dc.b	$DF,$8F,$F0,$43,$FD,$8F,$5F,$D2,$FE,$FB,$FD,$95,$F9,$2F,$7C,$91
	.dc.b	$89,$FC,$A3,$6F,$CE,$D7,$FC,$C9,$F8,$27,$F9,$99,$FD,$CC,$FF,$51
	.dc.b	$8C,$75,$7A,$DB,$19,$EE,$FE,$02,$FE,$62,$3E,$55,$7E,$5F,$9D,$18
	.dc.b	$E6,$FE,$1C,$D8,$CD,$7C,$BA,$F8,$27,$E4,$CF,$83,$AF,$EE,$AD,$F7
	.dc.b	$4F,$63,$A3,$6F,$94,$EA,$72,$79,$2F,$99,$D7,$B7,$F8,$1B,$F8,$93
	.dc.b	$25,$72,$F9,$0F,$EB,$BE,$E3,$E4,$5E,$F7,$FB,$EF,$9F,$F8,$0D,$2B
	.dc.b	$53,$9D,$38,$DF,$0A,$FC,$A4,$62,$75,$03,$94,$EE,$FA,$DF,$49,$E2
	.dc.b	$6D,$C4,$6A,$78,$5F,$F8,$7A,$BF,$A7,$10,$FE,$A6,$53,$46,$08,$7F
	.dc.b	$22,$D3,$E7,$7D,$87,$EF,$34,$F9,$DA,$A3,$13,$A8,$FE,$A4,$DA,$EA
	.dc.b	$43,$F6,$E7,$84,$FC,$46,$FF,$EB,$7A,$ED,$BF,$E1,$9B,$19,$BF,$C8
	.dc.b	$1E,$13,$F2,$94,$FB,$3F,$CC,$70,$FC,$B6,$C6,$67,$71,$FA,$9E,$3F
	.dc.b	$A1,$FA,$35,$A3,$49,$5F,$12,$3D,$25,$3F,$6E,$7F,$8D,$BA,$3E,$D7
	.dc.b	$CA,$FE,$D5,$B3,$FC,$0A,$53,$96,$C7,$F0,$FE,$1F,$74,$7F,$14,$EB
	.dc.b	$FE,$23,$AF,$F2,$CE,$AF,$E2,$CD,$8C,$DF,$8F,$99,$D9,$FE,$A9,$C3
	.dc.b	$0D,$7F,$97,$F7,$FF,$B9,$7E,$DB,$F8,$C7,$81,$FC,$C5,$B1,$9C,$3F
	.dc.b	$9C,$36,$BF,$B2,$F8,$5F,$D1,$73,$FE,$C9,$E1,$7D,$99,$FE,$DC,$D7
	.dc.b	$6D,$7E,$49,$F8,$33,$E4,$9F,$37,$A3,$FA,$EC,$FF,$86,$3C,$3E,$11
	.dc.b	$89,$FE,$CF,$1C,$F1,$FF,$37,$75,$7E,$A7,$F4,$7F,$ED,$E4,$AF,$ED
	.dc.b	$E1,$FE,$F5,$C3,$D3,$3F,$F0,$1E,$58,$C3,$F3,$97,$FE,$04,$6B,$E7
	.dc.b	$E0,$F9,$B7,$FE,$56,$8F,$49,$F9,$D9,$78,$FF,$E1,$CE,$8F,$F3,$FB
	.dc.b	$67,$FC,$37,$18,$F2,$3F,$2D,$1F,$5A,$FA,$61,$DA,$3F,$4D,$87,$FA
	.dc.b	$2F,$C9,$DC,$7F,$52,$6A,$8D,$3F,$A3,$3F,$C2,$D1,$AD,$FF,$C7,$B3
	.dc.b	$FC,$4E,$7F,$C1,$3B,$71,$F0,$7F,$C8,$ED,$8F,$F0,$AB,$5F,$F1,$2F
	.dc.b	$94,$ED,$FA,$AD,$BF,$8F,$3F,$E1,$98,$C6,$FF,$E5,$79,$CC,$FC,$83
	.dc.b	$F0,$64,$30,$3D,$0F,$F3,$FF,$D8,$FF,$9B,$89,$1F,$E7,$C7,$FC,$EF
	.dc.b	$26,$92,$3F,$CF,$31,$97,$FD,$07,$F1,$DF,$E7,$82,$38,$36,$F4,$23
	.dc.b	$83,$FD,$31,$EA,$7F,$37,$F8,$CF,$F4,$4B,$63,$22,$32,$D3,$FD,$70
	.dc.b	$53,$FD,$68,$FF,$A8,$3E,$11,$EB,$F3,$9C,$9F,$E4,$FF,$63,$0F,$DF
	.dc.b	$F3,$7F,$D8,$97,$FE,$C6,$9C,$E5,$FE,$CA,$8F,$F7,$89,$2B,$FD,$B1
	.dc.b	$1E,$17,$FD,$EE,$D5,$ED,$FE,$2D,$7C,$67,$FB,$94,$8F,$FA,$F2,$35
	.dc.b	$1F,$E9,$B2,$47,$FA,$81,$AF,$FB,$1B,$DD,$FC,$34,$7F,$BF,$F8,$BF
	.dc.b	$30,$DB,$EF,$5F,$EF,$F8,$FF,$61,$FB,$CF,$77,$73,$EA,$9C,$C6,$23
	.dc.b	$FD,$A7,$3B,$AE,$71,$8E,$4D,$3F,$DB,$E5,$18,$84,$62,$78,$C6,$E7
	.dc.b	$5E,$FF,$AA,$3F,$DB,$AD,$7F,$DC,$6D,$7F,$BE,$35,$7C,$03,$E0,$7F
	.dc.b	$C0,$1D,$3F,$EE,$08,$C4,$FE,$D4,$FF,$02,$38,$8D,$4F,$09,$FF,$76
	.dc.b	$FC,$17,$FB,$CA,$7F,$DD,$2D,$8C,$9B,$3F,$DE,$EE,$5F,$F7,$CF,$C2
	.dc.b	$3E,$13,$51,$FE,$13,$8C,$3E,$30,$E2,$BF,$DE,$ED,$8F,$C2,$67,$C9
	.dc.b	$FC,$5B,$78,$7F,$19,$5F,$EF,$E6,$CF,$F8,$84,$FF,$11,$7C,$2F,$BD
	.dc.b	$C0,$70,$ED,$AF,$0C,$BC,$1C,$FF,$AA,$8F,$43,$6F,$F5,$DC,$FA,$37
	.dc.b	$E0,$CA,$DB,$67,$FD,$EC,$48,$FF,$7D,$B5,$FC,$E7,$27,$F2,$57,$FC
	.dc.b	$75,$CB,$FC,$8E,$7F,$38,$48,$F5,$F1,$FF,$96,$63,$F3,$BE,$C7,$19
	.dc.b	$FC,$8B,$67,$51,$C3,$E1,$B0,$4D,$FE,$A4,$A3,$87,$34,$9C,$B7,$0F
	.dc.b	$D5,$F3,$75,$24,$F2,$7C,$5C,$DE,$9C,$7F,$3B,$A6,$09,$5F,$83,$F8
	.dc.b	$77,$C4,$7F,$63,$A4,$61,$FE,$CC,$D7,$ED,$9A,$8E,$11,$98,$FB,$ED
	.dc.b	$7F,$08,$D7,$FC,$CC,$D7,$91,$DA,$39,$91,$C1,$0F,$0C,$71,$24,$6E
	.dc.b	$32,$D7,$C6,$D7,$C6,$D7,$C6,$D7,$C6,$72,$3C,$07,$22,$9C,$8A,$72
	.dc.b	$29,$C8,$E0,$73,$21,$82,$1C,$CC,$94,$C9,$D4,$C9,$4C,$9D,$CE,$04
	.dc.b	$34,$53,$46,$08,$60,$91,$F3,$1F,$5B,$5F,$09,$A8,$F9,$11,$98,$F1
	.dc.b	$B5,$F0,$35,$EE,$D7,$91,$DA,$39,$91,$C1,$0F,$0C,$71,$24,$6E,$32
	.dc.b	$D7,$C6,$D7,$C6,$D7,$C6,$D7,$C6,$72,$3C,$07,$22,$9C,$8A,$72,$3F
	.dc.b	$03,$ED,$FF,$AF,$DF,$E4,$2D,$7F,$D9,$07,$EA,$9C,$3F,$A4,$F9,$5F
	.dc.b	$F6,$57,$89,$F9,$E6,$C7,$EB,$1A,$FE,$B1,$AF,$13,$FD,$B6,$58,$E4
	.dc.b	$E0,$FD,$EC,$70,$69,$9B,$46,$5B,$19,$8C,$11,$C1,$27,$B9,$1F,$41
	.dc.b	$92,$B9,$7D,$46,$4A,$E5,$F5,$1C,$08,$68,$A6,$8C,$10,$C1,$0F,$7B
	.dc.b	$D5,$A7,$22,$9C,$4A,$71,$9F,$87,$1C,$5B,$1C,$5A,$F1,$6B,$C6,$7E
	.dc.b	$44,$F1,$6C,$71,$6B,$C5,$AF,$13,$E3,$16,$39,$38,$3C,$E4,$9F,$0B
	.dc.b	$67,$9B,$4E,$11,$A2,$46,$9B,$1A,$8C,$11,$C1,$27,$C6,$47,$D8,$64
	.dc.b	$AE,$5F,$71,$92,$B9,$7D,$C7,$02,$1A,$29,$A3,$04,$03,$E5,$DC,$7F
	.dc.b	$EA,$FF,$4B,$FA,$DC,$63,$34,$62,$75,$5F,$E1,$AA,$D4,$FF,$96,$2B
	.dc.b	$73,$FE,$74,$8C,$11,$FF,$10,$FC,$63,$CB,$E7,$1F,$E2,$78,$C5,$72
	.dc.b	$8C,$4E,$9F,$4C,$EA,$BF,$C1,$F5,$B9,$FF,$31,$72,$ED,$FF,$08,$C6
	.dc.b	$5B,$3F,$DE,$29,$F1,$BB,$6D,$6D,$B3,$B9,$E9,$E6,$7F,$BA,$23,$FC
	.dc.b	$02,$DB,$FF,$1E,$B5,$FF,$21,$35,$FA,$3E,$4F,$F9,$D9,$F2,$36,$7F
	.dc.b	$C1,$6D,$8F,$F2,$3B,$5F,$F2,$67,$5F,$F9,$0E,$7F,$C8,$31,$ED,$8C
	.dc.b	$3F,$EB,$48,$FF,$0A,$F6,$FF,$9C,$9B,$DB,$F8,$26,$CF,$C5,$9D,$3F
	.dc.b	$E5,$6A,$7A,$5B,$8C,$35,$F1,$B5,$FC,$6C,$70,$7E,$87,$81,$FE,$70
	.dc.b	$BF,$F3,$6F,$83,$B9,$FF,$38,$B6,$75,$1F,$EB,$3F,$8E,$D7,$84,$69
	.dc.b	$23,$EE,$78,$9E,$87,$FC,$E0,$D7,$FD,$1F,$A7,$78,$C4,$3C,$47,$D3
	.dc.b	$F8,$DB,$1F,$87,$6E,$7F,$E9,$29,$F4,$73,$7E,$97,$E3,$BF,$49,$CA
	.dc.b	$69,$30,$FF,$6A,$BF,$A5,$23,$83,$F4,$7C,$E9,$E5,$18,$6B,$FA,$26
	.dc.b	$BF,$A2,$8E,$13,$F9,$2F,$63,$F7,$6D,$E9,$DD,$39,$95,$A8,$FC,$BB
	.dc.b	$5F,$F4,$C7,$37,$E5,$6F,$F2,$7E,$0F,$F0,$86,$CF,$49,$D3,$FE,$9C
	.dc.b	$F9,$FF,$E1,$99,$BE,$3F,$FC,$CB,$18,$6E,$BE,$F6,$CE,$EB,$E0,$47
	.dc.b	$ED,$75,$D4,$60,$87,$FA,$B7,$E1,$7F,$D0,$CF,$13,$1E,$3F,$F1,$EA
	.dc.b	$FF,$3A,$CE,$9B,$19,$9F,$F5,$9F,$5F,$FB,$5A,$7F,$46,$48,$CF,$2E
	.dc.b	$DF,$E5,$38,$BC,$EF,$E5,$51,$E1,$3E,$D3,$8A,$FF,$5F,$CE,$F3,$E5
	.dc.b	$5A,$6A,$F8,$EF,$F0,$6F,$C7,$FF,$A3,$1B,$1F,$EC,$86,$BF,$EC,$D6
	.dc.b	$BF,$CF,$34,$FC,$CC,$FC,$0E,$5F,$E7,$ED,$93,$4E,$DE,$1F,$F4,$4F
	.dc.b	$8D,$B2,$69,$1F,$F6,$CD,$FE,$9A,$3A,$12,$33,$7C,$DF,$C6,$F3,$FF
	.dc.b	$8B,$F9,$FF,$22,$DE,$7F,$EA,$3E,$27,$E5,$CF,$8F,$C0,$FA,$DF,$EA
	.dc.b	$FF,$1D,$86,$E3,$FE,$ED,$E1,$FA,$51,$88,$FA,$B5,$C5,$57,$4F,$A0
	.dc.b	$FF,$22,$DF,$D9,$6D,$1A,$63,$E4,$BE,$65,$F7,$94,$62,$3B,$B5,$FC
	.dc.b	$0B,$5F,$D8,$65,$EE,$7E,$1E,$5E,$F6,$9D,$7C,$CE,$58,$FE,$35,$B8
	.dc.b	$1A,$75,$D7,$E5,$DF,$3B,$FE,$9F,$87,$76,$EB,$FE,$62,$C5,$F4,$9F
	.dc.b	$E0,$09,$F8,$A7,$CB,$71,$AF,$E6,$1B,$19,$AF,$9C,$BD,$B7,$C5,$BE
	.dc.b	$23,$F8,$D7,$0F,$F8,$3A,$BF,$C1,$F5,$A9,$CC,$66,$31,$97,$D7,$8D
	.dc.b	$DF,$38,$C4,$77,$87,$C5,$7F,$27,$CF,$B4,$E6,$30,$F5,$9D,$D7,$BA
	.dc.b	$3A,$CE,$9F,$8B,$AF,$C9,$7F,$C4,$BB,$7C,$C7,$EA,$9E,$4F,$BE,$E0
	.dc.b	$F8,$DB,$97,$C2,$6E,$27,$78,$FF,$8A,$BE,$8F,$F9,$F7,$E4,$1F,$CF
	.dc.b	$F0,$FE,$17,$E3,$27,$35,$F9,$D8,$C2,$AE,$15,$FE,$37,$AD,$BE,$A3
	.dc.b	$CE,$7A,$B8,$BF,$04,$DE,$1F,$F5,$18,$F1,$9E,$EF,$EB,$B1,$FB,$13
	.dc.b	$98,$C5,$7F,$92,$23,$0F,$EB,$27,$75,$CA,$30,$7F,$26,$FA,$17,$CE
	.dc.b	$7F,$94,$3F,$08,$F4,$DA,$73,$49,$CC,$61,$B5,$C6,$BD,$B5,$FA,$CC
	.dc.b	$3F,$85,$13,$8B,$F3,$1F,$17,$FE,$5E,$FD,$23,$BF,$4A,$31,$87,$E7
	.dc.b	$DB,$87,$F3,$C9,$D2,$AE,$3F,$82,$FF,$2D,$BF,$D3,$36,$78,$35,$FE
	.dc.b	$F5,$48,$C7,$2B,$E5,$1F,$51,$F9,$9C,$3F,$45,$18,$7E,$4E,$1F,$AD
	.dc.b	$8C,$75,$7E,$02,$F9,$5F,$1A,$FC,$01,$0F,$BD,$A7,$E6,$E7,$F6,$BE
	.dc.b	$3E,$E7,$F5,$4D,$8C,$E3,$E0,$6E,$1F,$8E,$AF,$9D,$7F,$A5,$69,$F4
	.dc.b	$BA,$BF,$4C,$F8,$4C,$97,$DC,$7C,$B3,$E0,$3F,$D2,$3F,$26,$C3,$FE
	.dc.b	$00,$B6,$FF,$D2,$B7,$C2,$B7,$5A,$24,$EA,$FA,$2B,$E4,$AF,$93,$9A
	.dc.b	$E2,$FE,$24,$FF,$13,$10,$FE,$AE,$53,$F9,$49,$4F,$D5,$1F,$C4,$32
	.dc.b	$F9,$3F,$17,$F5,$7B,$FF,$C4,$2D,$DB,$FC,$78,$DD,$8F,$CD,$D7,$FB
	.dc.b	$1E,$1D,$4F,$21,$E2,$3F,$0B,$BF,$2A,$A7,$22,$D7,$E1,$78,$3F,$14
	.dc.b	$FE,$62,$9F,$1B,$49,$5F,$9A,$8C,$BF,$A3,$5F,$A4,$FE,$A3,$9B,$FD
	.dc.b	$8E,$D8,$CC,$EE,$3D,$5D,$BF,$D2,$E7,$D0,$7C,$82,$9F,$E5,$DE,$BF
	.dc.b	$DF,$FD,$57,$EF,$20,$D2,$5F,$F0,$CE,$EF,$F6,$BE,$7F,$C8,$3A,$FF
	.dc.b	$88,$B7,$BB,$F8,$83,$67,$75,$E0,$C3,$E5,$9E,$32,$1F,$CA,$57,$D8
	.dc.b	$7F,$14,$F0,$BF,$97,$B7,$C1,$FF,$3A,$36,$76,$7F,$4D,$F0,$FE,$ED
	.dc.b	$D8,$FB,$AD,$C7,$F0,$DC,$DF,$D0,$E7,$33,$E2,$3C,$A7,$E3,$0F,$8F
	.dc.b	$CF,$FD,$B3,$C4,$EB,$1F,$D6,$27,$8B,$FD,$7D,$B3,$B6,$D7,$F6,$9F
	.dc.b	$A1,$F0,$F8,$DF,$E8,$D6,$C9,$A7,$F7,$FE,$0F,$9F,$87,$F7,$F3,$BC
	.dc.b	$FF,$82,$23,$31,$FC,$9D,$AF,$A4,$F6,$74,$FE,$71,$FC,$0C,$FF,$86
	.dc.b	$DF,$C9,$9F,$E1,$F2,$9F,$E3,$8F,$20,$E0,$FF,$35,$B8,$E8,$FD,$35
	.dc.b	$75,$9E,$FC,$DF,$91,$7F,$C6,$67,$1B,$FF,$3B,$79,$5C,$63,$FC,$74
	.dc.b	$D7,$77,$FE,$45,$7F,$C1,$67,$F9,$EB,$CB,$AD,$33,$5F,$E4,$96,$F9
	.dc.b	$7F,$E4,$FE,$1F,$F2,$B4,$7C,$88,$C3,$FE,$49,$F2,$FE,$3B,$F8,$02
	.dc.b	$67,$EB,$7E,$51,$FE,$7B,$F3,$3E,$1F,$C8,$7F,$9E,$49,$1F,$E6,$E0
	.dc.b	$FF,$9E,$08,$FF,$A1,$FD,$0E,$2D,$36,$71,$F9,$0F,$F4,$21,$1C,$1B
	.dc.b	$7F,$D0,$64,$70,$7F,$98,$FC,$3F,$AF,$F1,$FF,$E7,$86,$C6,$44,$65
	.dc.b	$A7,$FA,$B0,$A7,$F9,$1F,$D7,$7A,$FC,$A3,$0D,$7F,$0B,$1A,$78,$9F
	.dc.b	$EA,$1F,$63,$4B,$D5,$FF,$62,$DF,$FB,$02,$36,$66,$77,$F0,$CF,$6B
	.dc.b	$FD,$E5,$5F,$ED,$1E,$3D,$CE,$5B,$3A,$BF,$F7,$06,$75,$FF,$30,$7B
	.dc.b	$7F,$EE,$2C,$3F,$DB,$A4,$8F,$F2,$BF,$B9,$FE,$31,$5F,$1F,$FE,$EC
	.dc.b	$23,$AF,$1F,$F4,$71,$FE,$FB,$24,$7F,$AB,$1A,$FF,$9B,$F3,$F3,$47
	.dc.b	$A4,$91,$FE,$66,$8C,$BF,$E6,$8F,$8D,$7C,$6B,$5F,$1C,$E6,$31,$7F
	.dc.b	$A4,$AD,$D6,$AB,$FC,$FD,$18,$9F,$D4,$E8,$EF,$E3,$5F,$E7,$68,$C5
	.dc.b	$F1,$8F,$DF,$C6,$2B,$8D,$70,$9F,$F4,$57,$C0,$B9,$FF,$82,$FA,$BF
	.dc.b	$B6,$8C,$4F,$F8,$08,$DC,$E9,$FF,$42,$C6,$A7,$F7,$BF,$05,$F8,$96
	.dc.b	$F3,$6E,$BF,$D3,$13,$97,$FD,$1A,$D7,$FD,$39,$F0,$8F,$84,$E1,$1F
	.dc.b	$C6,$23,$0F,$8C,$C4,$69,$46,$AF,$F8,$7F,$C2,$BE,$1B,$4D,$AE,$41
	.dc.b	$C4,$F0,$9E,$9F,$0E,$F8,$8A,$4B,$E3,$19,$79,$F6,$7A,$DF,$D5,$E9
	.dc.b	$E5,$8C,$41,$B2,$F8,$FE,$E4,$9E,$AF,$B7,$A3,$F3,$99,$7F,$2A,$6C
	.dc.b	$FE,$6C,$F5,$B8,$7F,$DA,$B1,$FC,$C3,$C1,$79,$1F,$89,$8C,$36,$7F
	.dc.b	$D7,$F3,$AE,$3F,$8B,$D2,$F8,$AF,$CD,$61,$FC,$EA,$31,$D1,$A9,$CE
	.dc.b	$1F,$A3,$6B,$F9,$AF,$8B,$AF,$F5,$2E,$A4,$62,$11,$8A,$FA,$AD,$C3
	.dc.b	$F8,$41,$27,$F6,$1D,$FC,$71,$FE,$15,$3F,$AB,$EC,$E9,$7F,$AC,$8C
	.dc.b	$61,$D2,$7C,$1D,$9F,$A8,$9F,$D2,$CF,$E1,$23,$18,$FE,$B7,$4F,$37
	.dc.b	$C6,$3C,$4F,$EB,$32,$8C,$3F,$D7,$1A,$FF,$60,$6B,$FE,$EC,$6B,$FE
	.dc.b	$FB,$6B,$FD,$D9,$AF,$58,$EF,$18,$23,$82,$47,$85,$AF,$85,$AF,$85
	.dc.b	$AF,$85,$AF,$85,$AF,$84,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$53,$A1
	.dc.b	$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21
	.dc.b	$A2,$9A,$30,$43,$04,$8F,$A6,$FB,$5A,$F9,$9A,$F8,$5A,$F6,$6B,$D5
	.dc.b	$AF,$58,$FA,$71,$82,$38,$24,$78,$5A,$F8,$5A,$F8,$5A,$F8,$5A,$F8
	.dc.b	$5A,$F8,$4E,$84,$3A,$14,$E8,$53,$A1,$4E,$85,$3E,$4F,$BD,$FE,$CA
	.dc.b	$7F,$CA,$A7,$FB,$47,$1F,$F2,$2C,$66,$32,$D7,$F4,$DE,$CF,$F9,$73
	.dc.b	$DE,$FD,$73,$63,$F5,$CD,$7F,$5C,$D7,$89,$FE,$DC,$2C,$72,$70,$7E
	.dc.b	$F6,$38,$34,$CE,$08,$CB,$63,$31,$82,$38,$24,$F7,$23,$E8,$32,$57
	.dc.b	$2F,$A8,$C9,$5C,$BE,$A3,$81,$0D,$14,$D1,$82,$18,$21,$EF,$7A,$B4
	.dc.b	$E4,$53,$89,$4E,$33,$F0,$E3,$8B,$63,$8B,$5E,$2D,$78,$CF,$C8,$9E
	.dc.b	$2D,$8E,$2D,$78,$B5,$E2,$7C,$62,$C7,$27,$07,$9C,$93,$E1,$6C,$F3
	.dc.b	$69,$C2,$34,$48,$D3,$63,$51,$82,$38,$24,$F8,$C8,$FB,$0C,$95,$CB
	.dc.b	$EE,$32,$57,$2F,$B8,$E0,$43,$45,$34,$60,$87,$F9,$23,$97,$FC,$B1
	.dc.b	$EC,$7F,$94,$1B,$1F,$E4,$F6,$BF,$B8,$D3,$FC,$35,$AA,$BF,$CB,$9E
	.dc.b	$36,$AB,$FC,$00,$D7,$FB,$FB,$78,$0D,$E3,$97,$CE,$BF,$CD,$91,$8B
	.dc.b	$FF,$65,$FC,$7B,$D1,$E8,$DB,$E1,$7F,$C3,$23,$15,$C2,$BC,$5C,$FF
	.dc.b	$E1,$F8,$FF,$11,$F1,$F0,$BF,$F7,$3C,$E5,$FF,$12,$F4,$7F,$9C,$6B
	.dc.b	$FD,$B7,$D5,$FE,$73,$9E,$01,$C4,$6A,$FF,$DE,$2D,$9D,$CF,$F9,$05
	.dc.b	$BB,$E9,$FE,$46,$73,$70,$FF,$26,$8E,$13,$FE,$EC,$EB,$7D,$26,$AB
	.dc.b	$FD,$76,$D7,$FC,$A0,$D8,$DC,$7A,$3C,$1F,$F4,$B3,$76,$E1,$18,$BF
	.dc.b	$F7,$CB,$67,$CB,$E0,$FF,$9A,$E7,$FC,$D1,$1F,$0A,$30,$EB,$C1,$F1
	.dc.b	$78,$9F,$CF,$AB,$FD,$1D,$38,$21,$F2,$BC,$4F,$F6,$94,$7F,$A2,$A3
	.dc.b	$1E,$17,$AE,$73,$1F,$1B,$E4,$1E,$8F,$FA,$BF,$E3,$FE,$4C,$A7,$3A
	.dc.b	$A7,$31,$86,$D7,$1A,$FA,$3E,$63,$E5,$B1,$1F,$ED,$26,$A1,$1B,$C3
	.dc.b	$B4,$72,$9D,$AE,$CF,$D3,$E1,$FA,$DA,$DC,$FE,$AF,$93,$75,$A8,$FF
	.dc.b	$66,$75,$F2,$EC,$FD,$3D,$F0,$9F,$D6,$B5,$FD,$6B,$63,$FC,$D9,$87
	.dc.b	$FB,$3F,$B3,$91,$27,$87,$3F,$E6,$E7,$F2,$C4,$BF,$C9,$C6,$5B,$97
	.dc.b	$C1,$9F,$D1,$B6,$32,$74,$70,$FF,$7E,$FB,$B7,$D5,$7F,$B9,$7E,$4B
	.dc.b	$EB,$C6,$31,$46,$27,$55,$FA,$4A,$D4,$FA,$EB,$73,$FC,$19,$AF,$FA
	.dc.b	$45,$AF,$FA,$47,$2E,$6F,$CF,$8F,$E2,$2F,$F1,$D6,$B9,$D3,$EB,$1F
	.dc.b	$C7,$23,$31,$F4,$23,$04,$7C,$4F,$94,$8E,$0F,$F0,$FC,$3E,$2D,$F2
	.dc.b	$7F,$8B,$F9,$2F,$E0,$33,$98,$C5,$70,$AD,$AA,$DC,$61,$F8,$9F,$67
	.dc.b	$27,$E2,$35,$FC,$C7,$C0,$3E,$01,$ED,$32,$7E,$4F,$FC,$4A,$B5,$39
	.dc.b	$8C,$57,$0A,$DD,$FF,$8E,$AF,$71,$88,$FF,$1E,$FC,$AF,$FA,$9A,$FF
	.dc.b	$6A,$D8,$FE,$3E,$D7,$7E,$77,$EB,$67,$F9,$AF,$95,$27,$F9,$9F,$8D
	.dc.b	$FE,$36,$F2,$3F,$DB,$13,$F2,$3F,$06,$7C,$0F,$ED,$7D,$CC,$37,$0F
	.dc.b	$A3,$3A,$53,$A6,$9E,$BD,$BF,$84,$F8,$9F,$03,$C1,$FE,$B9,$C1,$F2
	.dc.b	$5B,$3A,$AE,$18,$FC,$BF,$90,$F8,$ED,$8C,$F0,$A7,$2D,$8F,$6E,$9E
	.dc.b	$F7,$FC,$CB,$AF,$B9,$C9,$F3,$FD,$C7,$D5,$F5,$8F,$BB,$18,$BE,$91
	.dc.b	$89,$D2,$D1,$5B,$C7,$C1,$3F,$D8,$BC,$E7,$0F,$CF,$6E,$7A,$C9,$F9
	.dc.b	$2B,$DB,$F9,$D8,$7F,$25,$9C,$C6,$23,$F1,$D3,$B7,$E7,$9B,$71,$C5
	.dc.b	$F9,$CF,$99,$FE,$5D,$FA,$55,$A3,$48,$FE,$97,$F0,$7E,$D8,$FE,$B5
	.dc.b	$BD,$A6,$9E,$39,$71,$7C,$D7,$E7,$EB,$EE,$7E,$79,$8D,$BF,$D1,$18
	.dc.b	$7C,$0F,$96,$EF,$A9,$A4,$AF,$69,$FA,$DC,$BF,$D0,$DF,$92,$FF,$70
	.dc.b	$8C,$7B,$1F,$D2,$E3,$18,$7F,$05,$8C,$3F,$C0,$5B,$3E,$3E,$1F,$AE
	.dc.b	$FA,$B3,$F4,$39,$3F,$D5,$B9,$BE,$A9,$5F,$A5,$C3,$31,$8A,$FD,$1C
	.dc.b	$62,$75,$1F,$D7,$EB,$94,$62,$3F,$15,$D1,$FE,$31,$FD,$2F,$D9,$CF
	.dc.b	$FC,$37,$D9,$FE,$19,$18,$9D,$63,$FA,$D9,$CB,$F3,$AF,$F2,$18,$FE
	.dc.b	$3C,$98,$FF,$13,$FA,$17,$DB,$FF,$17,$DF,$BC,$63,$04,$62,$75,$1F
	.dc.b	$C5,$67,$53,$E3,$F8,$5F,$DB,$4F,$EF,$DB,$8F,$F2,$6C,$D1,$88,$F7
	.dc.b	$B6,$7F,$9D,$7E,$49,$F0,$DB,$3F,$28,$FD,$17,$12,$74,$D8,$D1,$F0
	.dc.b	$33,$FB,$7C,$BF,$93,$8C,$1F,$D2,$3F,$07,$7C,$AF,$76,$C7,$2C,$51
	.dc.b	$8E,$BF,$C5,$B6,$3E,$74,$FE,$1F,$83,$E6,$B8,$C3,$C0,$E7,$1F,$1F
	.dc.b	$2F,$E3,$23,$FD,$EC,$3E,$73,$F1,$0B,$E0,$DF,$3A,$F8,$5F,$F1,$35
	.dc.b	$6A,$73,$18,$C3,$FD,$FB,$5B,$BE,$91,$89,$D4,$78,$2F,$95,$6D,$B3
	.dc.b	$A9,$F6,$1F,$E7,$C2,$1F,$D2,$0A,$7E,$EC,$A7,$E4,$0F,$EE,$3A,$7C
	.dc.b	$CF,$92,$F9,$3B,$FF,$8F,$DB,$26,$9F,$BB,$7C,$27,$72,$1D,$CF,$EF
	.dc.b	$5C,$3C,$F7,$B1,$CD,$B3,$FC,$0B,$87,$F0,$CF,$E5,$E1,$F2,$54,$8F
	.dc.b	$E6,$4C,$9F,$9F,$E7,$FE,$F9,$F9,$37,$E8,$E0,$D2,$3E,$6C,$3E,$11
	.dc.b	$4F,$59,$FA,$3E,$9F,$F0,$37,$E9,$9F,$BB,$6C,$FE,$EC,$46,$5A,$7F
	.dc.b	$0B,$9F,$71,$FC,$2F,$B7,$F8,$86,$5F,$C8,$F9,$BF,$86,$CE,$8F,$EA
	.dc.b	$33,$AA,$F2,$73,$39,$BF,$D6,$55,$86,$BF,$B0,$EE,$FF,$3B,$69,$F2
	.dc.b	$DB,$7F,$CA,$E7,$E1,$B6,$7F,$98,$F8,$1F,$CE,$E7,$EF,$1F,$E3,$87
	.dc.b	$19,$FF,$8D,$9B,$19,$F0,$3F,$0C,$F8,$F4,$FF,$49,$BF,$28,$F9,$A7
	.dc.b	$F0,$4F,$07,$6D,$3B,$96,$FF,$C7,$B1,$A8,$FE,$95,$E1,$FF,$A4,$7F
	.dc.b	$47,$FE,$C8,$48,$FF,$23,$46,$5F,$47,$8B,$C7,$4F,$EE,$71,$D8,$FE
	.dc.b	$E7,$19,$CF,$FB,$A6,$3F,$A4,$9F,$41,$37,$FA,$19,$FF,$AB,$BC,$4F
	.dc.b	$F0,$4F,$91,$FE,$49,$E6,$FF,$24,$79,$1F,$E1,$16,$C6,$70,$FF,$0C
	.dc.b	$35,$FD,$21,$0F,$8A,$4A,$FF,$2B,$79,$3B,$8F,$E9,$ED,$8C,$EB,$FA
	.dc.b	$49,$FF,$16,$DF,$F8,$DE,$3F,$86,$47,$F3,$BF,$2B,$FC,$BD,$DD,$92
	.dc.b	$47,$F9,$13,$0F,$8D,$E6,$7F,$91,$76,$FF,$22,$12,$7E,$E1,$F1,$76
	.dc.b	$F8,$6F,$F9,$73,$CE,$FF,$62,$7E,$9B,$60,$D2,$3F,$E4,$DF,$47,$FC
	.dc.b	$E1,$1F,$E6,$C6,$C6,$74,$E2,$D9,$FF,$50,$7D,$E4,$8F,$F5,$03,$5F
	.dc.b	$E9,$FE,$AF,$FB,$57,$E5,$7F,$D0,$AD,$8F,$F4,$53,$5D,$88,$FF,$56
	.dc.b	$72,$48,$FF,$56,$35,$FF,$70,$1F,$EC,$E7,$FD,$48,$47,$FD,$AF,$87
	.dc.b	$EC,$9F,$F7,$07,$B9,$FE,$8D,$9F,$F7,$01,$23,$FD,$60,$FF,$AE,$FE
	.dc.b	$56,$47,$FA,$ED,$AF,$FB,$27,$DD,$FF,$4B,$47,$FB,$16,$CD,$24,$FF
	.dc.b	$B4,$BE,$57,$FD,$AA,$9A,$47,$FD,$BB,$7F,$EF,$B8,$F6,$C7,$FB,$F7
	.dc.b	$E0,$3F,$CC,$2F,$13,$65,$3E,$1B,$A8,$FF,$6A,$12,$3F,$DA,$8D,$7E
	.dc.b	$4F,$C0,$FF,$BE,$A3,$FD,$98,$48,$F9,$2F,$89,$F9,$38,$77,$46,$23
	.dc.b	$5F,$04,$CF,$FD,$75,$18,$6C,$FB,$3E,$11,$F0,$98,$8F,$B7,$18,$C3
	.dc.b	$B4,$23,$17,$E3,$6C,$7C,$6F,$85,$75,$73,$8F,$2C,$62,$F9,$B6,$FD
	.dc.b	$8D,$9F,$9F,$F0,$CF,$87,$F6,$F6,$7C,$18,$C4,$FE,$E8,$F8,$2E,$23
	.dc.b	$53,$C2,$7E,$CF,$C4,$7E,$5A,$7E,$53,$63,$26,$CF,$CD,$B9,$7F,$3B
	.dc.b	$F1,$4F,$8A,$D4,$7D,$08,$C3,$E3,$0E,$2B,$EC,$36,$3F,$17,$F1,$7F
	.dc.b	$EE,$D7,$83,$67,$EF,$36,$37,$A7,$81,$7C,$63,$E3,$70,$D8,$F1,$36
	.dc.b	$FF,$10,$D8,$FC,$DF,$2F,$E1,$5F,$43,$67,$F4,$A4,$8F,$CE,$BF,$91
	.dc.b	$8F,$D0,$B7,$9B,$C9,$18,$9F,$F7,$9C,$FE,$5F,$BF,$FA,$2B,$7C,$FF
	.dc.b	$F6,$4B,$6F,$E4,$7C,$73,$D3,$FE,$67,$87,$EC,$23,$1A,$6A,$73,$3F
	.dc.b	$C4,$27,$37,$FA,$5E,$A7,$C5,$EA,$BF,$4B,$18,$C3,$EB,$36,$B8,$56
	.dc.b	$E9,$2F,$8B,$C3,$6F,$8C,$60,$E1,$D9,$FA,$F8,$FC,$13,$70,$FD,$84
	.dc.b	$63,$B3,$B7,$C7,$BD,$4F,$F0,$1E,$48,$C3,$FE,$06,$6B,$FC,$05,$AF
	.dc.b	$F1,$46,$BF,$CC,$9A,$FF,$04,$6B,$D6,$3B,$C6,$08,$E0,$91,$E1,$6B
	.dc.b	$E1,$6B,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$3A,$10,$E8,$53,$A1,$4E,$85
	.dc.b	$3A,$14,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81
	.dc.b	$4E,$06,$88,$68,$A6,$8C,$10,$C1,$23,$E9,$BE,$D6,$BE,$66,$BE,$16
	.dc.b	$BD,$9A,$F5,$6B,$D6,$3E,$9C,$60,$8E,$09,$1E,$16,$BE,$16,$BE,$16
	.dc.b	$BE,$16,$BE,$16,$BE,$13,$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$4F,$93
	.dc.b	$EF,$FF,$B2,$9F,$EC,$CD,$7F,$BC,$B5,$FF,$65,$1F,$AB,$F6,$BF,$B4
	.dc.b	$FB,$F6,$3F,$DB,$27,$EA,$63,$2F,$EB,$DA,$FE,$BD,$AE,$CE,$27,$EE
	.dc.b	$DC,$B4,$CD,$FF,$BF,$E7,$C0,$48,$EC,$6A,$78,$CE,$E3,$33,$B7,$C4
	.dc.b	$6C,$8E,$5F,$23,$5F,$13,$63,$A4,$72,$8C,$4F,$80,$91,$D0,$F0,$90
	.dc.b	$E0,$53,$B9,$4E,$E5,$9E,$EE,$8E,$24,$30,$43,$89,$92,$99,$79,$B8
	.dc.b	$23,$82,$47,$AA,$7D,$B3,$EB,$3A,$4E,$9C,$1D,$67,$83,$E9,$6B,$C1
	.dc.b	$AE,$CE,$27,$89,$CB,$4C,$C7,$79,$F1,$92,$3B,$1A,$9E,$33,$B8,$CC
	.dc.b	$ED,$F1,$1B,$23,$97,$C8,$D7,$C4,$D8,$E9,$1C,$A3,$13,$E0,$24,$74
	.dc.b	$3C,$24,$3E,$A6,$6F,$49,$9B,$E2,$3F,$CA,$71,$86,$C6,$6B,$FC,$DB
	.dc.b	$5C,$63,$11,$FE,$64,$E0,$E0,$FF,$9C,$B8,$3F,$BD,$5F,$F7,$8A,$DC
	.dc.b	$FF,$7E,$AD,$BF,$E9,$4F,$91,$7C,$9E,$23,$FC,$CF,$18,$AE,$30,$8C
	.dc.b	$5F,$F9,$BD,$B1,$FE,$82,$E7,$FF,$6F,$CF,$F8,$64,$91,$9E,$77,$3F
	.dc.b	$4F,$94,$7A,$BF,$E7,$6B,$FF,$3D,$46,$2F,$53,$97,$FD,$1F,$19,$9F
	.dc.b	$F4,$B9,$23,$FC,$73,$CB,$B7,$FD,$1E,$D7,$FD,$07,$D7,$FE,$66,$9F
	.dc.b	$F2,$0F,$83,$99,$FF,$24,$B6,$75,$1F,$E4,$E3,$E1,$F7,$7F,$95,$6B
	.dc.b	$FC,$A7,$B7,$18,$C4,$FF,$A9,$3B,$FF,$CC,$35,$FE,$5E,$F0,$3B,$46
	.dc.b	$27,$51,$C7,$E1,$64,$76,$6D,$7F,$9E,$23,$E9,$F8,$7E,$47,$FD,$9A
	.dc.b	$D7,$FC,$FF,$9F,$AF,$C3,$7C,$AD,$24,$FD,$C8,$CB,$F7,$BC,$57,$A9
	.dc.b	$F9,$6F,$93,$FF,$48,$B6,$7F,$D3,$4D,$7F,$DB,$B3,$C3,$75,$7F,$93
	.dc.b	$FD,$FD,$57,$F9,$39,$AF,$F9,$29,$BD,$67,$3D,$9F,$9F,$AF,$D1,$92
	.dc.b	$BF,$47,$DB,$C3,$B6,$63,$E0,$9F,$D1,$46,$2B,$F2,$DD,$BE,$0E,$DF
	.dc.b	$CE,$46,$2F,$85,$FE,$8E,$BC,$BE,$3F,$FA,$06,$BF,$DC,$64,$9F,$C6
	.dc.b	$9C,$0F,$C5,$F0,$FF,$B3,$A3,$F1,$71,$93,$F1,$65,$E8,$FF,$5A,$C7
	.dc.b	$FB,$5F,$D3,$7E,$0F,$FC,$13,$E4,$F0,$DB,$FF,$76,$CE,$B3,$56,$9B
	.dc.b	$3F,$EF,$2D,$3F,$71,$3E,$D2,$46,$74,$E8,$FF,$B9,$06,$FF,$C3,$7C
	.dc.b	$8F,$E2,$24,$8F,$9F,$87,$F9,$FD,$F9,$AD,$7F,$89,$35,$FE,$1C,$D7
	.dc.b	$FD,$0F,$BF,$F2,$09,$FA,$04,$8F,$24,$EF,$1F,$E2,$13,$FE,$DB,$E1
	.dc.b	$FF,$7C,$FC,$9D,$B3,$48,$F2,$F9,$3F,$B8,$47,$07,$BC,$FF,$74,$78
	.dc.b	$3F,$CC,$BE,$4B,$11,$8C,$91,$8E,$3F,$E2,$B1,$89,$E5,$18,$9E,$92
	.dc.b	$FB,$07,$90,$BE,$05,$9D,$46,$6B,$8D,$70,$AD,$9F,$D1,$9C,$47,$FB
	.dc.b	$0E,$64,$FF,$2B,$F1,$7F,$C5,$7E,$3B,$C7,$FE,$5B,$5E,$FF,$79,$F1
	.dc.b	$9F,$CC,$33,$7E,$1B,$98,$C3,$71,$FE,$AA,$D9,$DE,$5F,$87,$5F,$15
	.dc.b	$FE,$FC,$C7,$E0,$B6,$74,$D7,$E0,$35,$63,$E4,$F2,$BF,$B1,$FC,$A7
	.dc.b	$8D,$B3,$F9,$59,$D6,$0A,$D3,$67,$E1,$61,$F1,$1F,$CE,$6B,$F0,$F3
	.dc.b	$F8,$BA,$FC,$49,$E9,$AF,$F6,$BC,$7E,$17,$81,$E0,$9F,$B1,$9F,$C8
	.dc.b	$8C,$47,$A2,$B8,$3E,$02,$1E,$0B,$EC,$DC,$3F,$CC,$F7,$F4,$BE,$DB
	.dc.b	$E6,$CF,$89,$AF,$82,$D1,$88,$F0,$37,$47,$E5,$BF,$B9,$FD,$3B,$F2
	.dc.b	$2D,$AF,$AC,$D7,$F2,$0D,$7F,$C2,$2F,$F4,$0F,$92,$E3,$E4,$FE,$5D
	.dc.b	$B2,$69,$F3,$5F,$CE,$6F,$F3,$78,$31,$08,$C3,$E2,$FC,$7F,$A3,$4E
	.dc.b	$85,$AF,$4C,$6A,$3C,$FF,$8C,$FE,$A7,$F1,$8F,$F1,$3C,$60,$38,$7D
	.dc.b	$71,$BD,$3E,$0B,$63,$E9,$63,$E6,$C7,$CB,$18,$3F,$91,$70,$3C,$CF
	.dc.b	$D5,$61,$FA,$63,$2E,$39,$FC,$13,$BE,$3F,$D3,$D6,$9B,$7F,$6F,$EA
	.dc.b	$7F,$C7,$FF,$93,$79,$F9,$7F,$67,$18,$BF,$C9,$B6,$B8,$42,$BD,$D1
	.dc.b	$F3,$6F,$F1,$AF,$D2,$6B,$9C,$3F,$3E,$7E,$43,$95,$F2,$3F,$E5,$0F
	.dc.b	$D3,$FC,$51,$8C,$51,$89,$D4,$7E,$99,$B7,$F9,$5F,$C4,$FF,$6F,$3F
	.dc.b	$AA,$6C,$66,$3F,$6C,$64,$AF,$F2,$1F,$98,$7C,$4F,$E3,$E7,$F8,$64
	.dc.b	$62,$7F,$21,$18,$9D,$47,$16,$BF,$07,$2F,$7F,$17,$DD,$D3,$E2,$DF
	.dc.b	$E9,$BE,$97,$6D,$BE,$A7,$E9,$23,$3D,$27,$AC,$7C,$BB,$E9,$87,$68
	.dc.b	$FD,$F9,$E1,$8C,$9F,$EE,$CD,$9F,$01,$FD,$07,$E6,$BF,$9E,$53,$E0
	.dc.b	$3F,$DE,$51,$86,$DF,$FB,$C2,$31,$7C,$A9,$18,$BE,$95,$B2,$4F,$C3
	.dc.b	$3F,$AD,$F3,$FF,$21,$6E,$BF,$E7,$06,$E5,$FD,$81,$FC,$41,$82,$1F
	.dc.b	$D3,$4F,$E1,$B9,$FF,$5B,$F8,$FF,$D8,$6D,$F2,$DB,$B7,$C9,$A4,$62
	.dc.b	$B8,$57,$39,$87,$F2,$53,$B9,$F7,$4E,$2F,$E0,$5A,$66,$7C,$3C,$1F
	.dc.b	$87,$7F,$37,$C3,$FE,$AC,$F9,$1B,$66,$91,$FC,$E0,$FD,$27,$F4,$0D
	.dc.b	$32,$E0,$FD,$8E,$BF,$2F,$1F,$49,$F3,$CB,$D1,$FB,$CA,$7E,$51,$FA
	.dc.b	$EE,$9F,$F5,$FB,$63,$27,$F0,$02,$D7,$F8,$1F,$7F,$E1,$90,$FC,$43
	.dc.b	$F8,$7F,$67,$F1,$46,$F1,$7F,$B2,$9B,$3B,$3C,$B7,$F7,$0B,$9F,$A5
	.dc.b	$FE,$55,$F8,$DF,$D8,$EE,$FF,$0B,$36,$0D,$3E,$EF,$81,$F8,$18,$FE
	.dc.b	$75,$C5,$FC,$E1,$FB,$E6,$0E,$F9,$7E,$3D,$BA,$7F,$8B,$B2,$FE,$B3
	.dc.b	$E1,$73,$9F,$EA,$AD,$7F,$AA,$B6,$37,$1D,$8F,$EB,$67,$F8,$B7,$F5
	.dc.b	$0F,$EC,$2D,$3F,$B0,$94,$D8,$7F,$C4,$1E,$2F,$C9,$8F,$EE,$53,$F3
	.dc.b	$DB,$B6,$A3,$11,$F9,$73,$D6,$7E,$60,$FF,$02,$78,$DF,$91,$7F,$C1
	.dc.b	$AD,$7F,$BE,$35,$DC,$7E,$55,$E6,$43,$FC,$05,$E4,$6F,$93,$FC,$43
	.dc.b	$CD,$FE,$20,$24,$FF,$8B,$BE,$0B,$F7,$57,$FE,$34,$6F,$94,$69,$FB
	.dc.b	$FB,$FF,$21,$ED,$F0,$6F,$FC,$20,$D7,$FC,$3C,$D8,$DD,$FC,$29,$FD
	.dc.b	$01,$F0,$C8,$60,$FF,$2C,$76,$FF,$98,$7E,$53,$FC,$BA,$D8,$FF,$25
	.dc.b	$79,$BF,$E5,$F9,$CC,$1F,$F3,$87,$9E,$BC,$4F,$F9,$C2,$31,$3A,$9C
	.dc.b	$CF,$FA,$6F,$CF,$FF,$3B,$CF,$FA,$F7,$E5,$24,$FF,$A3,$9A,$FF,$BA
	.dc.b	$3D,$3F,$F3,$B7,$CA,$7F,$A3,$49,$1F,$E8,$7F,$83,$FF,$6A,$28,$FF
	.dc.b	$64,$6D,$23,$FD,$32,$D7,$FD,$46,$FF,$A3,$3E,$05,$EC,$FA,$E3,$0D
	.dc.b	$9F,$F5,$5C,$7F,$A4,$BD,$A7,$36,$67,$FD,$6D,$18,$DB,$F5,$91,$FE
	.dc.b	$8D,$8C,$57,$1A,$E1,$3F,$E9,$1F,$69,$ED,$FC,$BF,$93,$FF,$70,$12
	.dc.b	$7F,$D6,$2F,$FB,$2C,$8F,$FB,$43,$DF,$E0,$DC,$93,$FE,$DE,$F9,$09
	.dc.b	$3F,$EA,$DF,$82,$72,$7F,$A8,$67,$33,$FE,$CF,$F8,$37,$6E,$63,$FD
	.dc.b	$D7,$3F,$EB,$29,$CC,$23,$13,$B8,$FF,$7B,$63,$97,$FD,$93,$3F,$EF
	.dc.b	$B2,$3F,$ED,$67,$E7,$C6,$E3,$FD,$AF,$18,$3E,$44,$66,$3E,$66,$BF
	.dc.b	$91,$9C,$92,$3F,$D9,$5C,$2E,$4D,$46,$1A,$FF,$BB,$63,$57,$D6,$11
	.dc.b	$8A,$FF,$68,$FC,$23,$E1,$78,$D7,$FB,$26,$31,$7C,$5B,$5C,$2B,$75
	.dc.b	$FE,$59,$F0,$FF,$DA,$1E,$97,$FB,$73,$CB,$46,$27,$FD,$C5,$18,$7F
	.dc.b	$C1,$C7,$FB,$5C,$E4,$60,$FF,$6D,$74,$FD,$C7,$01,$C5,$7F,$BA,$9B
	.dc.b	$1F,$A9,$E0,$F0,$3D,$7E,$25,$F1,$18,$8F,$F7,$34,$61,$EA,$DB,$FF
	.dc.b	$7A,$B6,$3F,$DF,$5F,$13,$FC,$75,$F1,$B6,$7F,$DF,$6D,$8D,$F8,$7D
	.dc.b	$D5,$7F,$A3,$2B,$AE,$F9,$8C,$4F,$C3,$37,$1C,$79,$7D,$B1,$F1,$DA
	.dc.b	$FF,$BF,$DA,$9B,$CB,$F9,$CB,$FC,$DC,$FC,$A8,$C3,$DD,$AF,$D3,$F8
	.dc.b	$A7,$83,$E9,$6C,$7D,$26,$D7,$9E,$11,$87,$D5,$E7,$FF,$B8,$E7,$FC
	.dc.b	$82,$48,$CF,$96,$EE,$FB,$BD,$DF,$C2,$34,$BD,$DF,$CC,$F0,$ED,$39
	.dc.b	$6C,$67,$0F,$D2,$F8,$15,$FD,$13,$51,$FA,$27,$0D,$BF,$C6,$F8,$1F
	.dc.b	$C7,$EF,$F6,$A4,$9F,$A9,$19,$7E,$B7,$C6,$3E,$33,$FC,$B9,$08,$C3
	.dc.b	$FE,$5F,$6B,$F1,$DA,$F2,$8E,$31,$88,$FD,$23,$5F,$4B,$5E,$D1,$C0
	.dc.b	$8F,$76,$BD,$DA,$F7,$34,$43,$45,$34,$60,$86,$08,$79,$9E,$AD,$39
	.dc.b	$14,$E2,$53,$89,$EC,$78,$B4,$E2,$53,$89,$4E,$27,$C2,$36,$53,$65
	.dc.b	$37,$1F,$24,$91,$E0,$8D,$1E,$52,$1D,$CA,$68,$C1,$0C,$12,$3A,$BE
	.dc.b	$46,$BE,$46,$BE,$48,$F1,$46,$23,$BB,$5E,$CD,$7B,$47,$02,$3D,$DA
	.dc.b	$F7,$6B,$DC,$D1,$0D,$14,$D1,$82,$18,$21,$E6,$7A,$B4,$E4,$53,$89
	.dc.b	$4E,$27,$B1,$FD,$4F,$B6,$F6,$FF,$9F,$CF,$FB,$65,$B1,$9C,$3F,$D9
	.dc.b	$4D,$9F,$D6,$78,$9F,$BE,$F2,$6C,$7F,$B8,$8F,$D4,$C6,$5F,$DC,$B5
	.dc.b	$FD,$7B,$5D,$9C,$4F,$DE,$39,$69,$9C,$51,$96,$C6,$63,$04,$70,$49
	.dc.b	$F3,$91,$F2,$99,$2B,$97,$CE,$64,$AE,$5F,$4B,$5F,$3B,$67,$CA,$D9
	.dc.b	$F2,$36,$7A,$1C,$8F,$09,$0E,$E5,$34,$60,$86,$08,$78,$8E,$05,$36
	.dc.b	$53,$73,$E6,$76,$D7,$6D,$77,$3E,$A7,$6D,$76,$D7,$67,$B8,$F0,$12
	.dc.b	$3A,$36,$38,$34,$E1,$1A,$24,$69,$B1,$A8,$C1,$1C,$12,$7E,$81,$1F
	.dc.b	$49,$92,$B9,$7D,$66,$4A,$E5,$F6,$B5,$F4,$36,$7C,$AD,$9F,$23,$67
	.dc.b	$F4,$19,$3C,$EF,$F5,$27,$C6,$7F,$A1,$DB,$3F,$D9,$1A,$FF,$A2,$1A
	.dc.b	$FF,$70,$F8,$E7,$C7,$E2,$11,$8D,$D7,$FE,$0B,$8C,$36,$B8,$72,$B9
	.dc.b	$7F,$C1,$13,$FE,$11,$8C,$72,$F2,$C7,$FC,$1A,$D7,$FC,$9D,$1A,$9F
	.dc.b	$F4,$04,$FF,$81,$1B,$3B,$3F,$C0,$FD,$1F,$E1,$57,$01,$C4,$7F,$8B
	.dc.b	$09,$1F,$E2,$BF,$23,$A3,$4E,$3D,$4F,$90,$CC,$7F,$8E,$DA,$FF,$8E
	.dc.b	$DA,$FF,$8E,$DB,$D9,$FE,$D5,$8F,$1B,$6F,$FC,$97,$18,$DB,$8F,$73
	.dc.b	$F1,$1A,$49,$F5,$C6,$5F,$6F,$87,$FE,$73,$BF,$F3,$D4,$78,$A3,$0F
	.dc.b	$F9,$E8,$F2,$F8,$7F,$E7,$59,$FF,$39,$78,$78,$24,$FC,$8F,$D3,$3E
	.dc.b	$3B,$FE,$5D,$6B,$FE,$81,$6B,$FE,$36,$6A,$F9,$47,$D8,$FF,$BD,$7E
	.dc.b	$4F,$FC,$A1,$18,$D5,$18,$9D,$57,$F9,$AE,$B5,$3F,$EA,$0E,$6C,$CF
	.dc.b	$FA,$47,$CB,$FC,$94,$E7,$0E,$73,$FE,$73,$9C,$BF,$EB,$2F,$1B,$8B
	.dc.b	$FE,$4C,$F3,$3F,$CC,$F5,$C2,$3F,$D6,$53,$A7,$FD,$62,$DE,$CF,$D6
	.dc.b	$4E,$E7,$FD,$D9,$D9,$92,$46,$63,$F4,$F9,$73,$EC,$F1,$C7,$FA,$A5
	.dc.b	$AF,$EB,$E3,$C1,$D7,$FA,$4D,$3F,$4D,$D9,$FE,$6F,$6F,$67,$59,$D5
	.dc.b	$FF,$BA,$FA,$7F,$CC,$F5,$FE,$7D,$EA,$FC,$C4,$F8,$DB,$19,$9D,$CF
	.dc.b	$FA,$67,$AB,$F2,$D3,$F9,$42,$46,$74,$DB,$FE,$C1,$F9,$37,$E0,$C9
	.dc.b	$7C,$A7,$FA,$36,$73,$18,$AF,$F5,$34,$62,$75,$5C,$5F,$69,$27,$F8
	.dc.b	$1B,$5F,$E2,$D4,$F9,$4F,$E3,$8D,$8F,$E3,$8D,$77,$1F,$C4,$E3,$11
	.dc.b	$F3,$DA,$FD,$66,$F0,$7F,$1D,$9F,$9E,$49,$FE,$2A,$D7,$E1,$5F,$C1
	.dc.b	$F9,$3F,$5D,$9A,$49,$FE,$3A,$FD,$A2,$38,$3E,$2C,$FC,$BF,$97,$E6
	.dc.b	$48,$CD,$23,$F0,$64,$8F,$E5,$2D,$7F,$94,$F0,$FF,$AD,$FC,$3F,$C4
	.dc.b	$92,$3F,$97,$B5,$DF,$17,$F8,$A7,$8B,$F9,$AC,$E6,$5E,$D7,$F8,$FF
	.dc.b	$E2,$FF,$A3,$B7,$2F,$E6,$4D,$EF,$FF,$72,$36,$BF,$D7,$67,$2F,$29
	.dc.b	$F3,$3F,$89,$C7,$71,$8C,$3F,$C8,$57,$FE,$4D,$AD,$5F,$28,$C4,$F9
	.dc.b	$F4,$F7,$4F,$B0,$87,$B3,$0F,$52,$CF,$DA,$E8,$FF,$67,$B8,$78,$19
	.dc.b	$38,$4F,$67,$24,$38,$B9,$6E,$9E,$53,$D7,$97,$F9,$9F,$07,$C8,$7C
	.dc.b	$08,$FD,$53,$63,$35,$EF,$9E,$13,$F2,$A3,$F6,$1F,$18,$FC,$3B,$FC
	.dc.b	$C5,$68,$C4,$FF,$9A,$E3,$13,$A9,$CE,$7F,$13,$3F,$81,$7E,$88,$C6
	.dc.b	$7E,$B8,$F8,$39,$FA,$E7,$F0,$98,$FD,$AF,$87,$FC,$C3,$F7,$8C,$1F
	.dc.b	$97,$31,$B2,$37,$7F,$EF,$E9,$DF,$0F,$F6,$78,$F1,$1F,$47,$2F,$2B
	.dc.b	$E2,$C3,$CB,$8F,$8E,$3E,$D5,$F8,$3E,$03,$93,$4F,$0E,$3F,$C5,$E3
	.dc.b	$39,$78,$E5,$1E,$1A,$EC,$78,$73,$F8,$59,$7C,$6A,$7E,$58,$F0,$56
	.dc.b	$FA,$1F,$0C,$E6,$31,$5F,$16,$31,$3A,$AE,$2F,$A3,$F1,$27,$81,$FE
	.dc.b	$A5,$BF,$99,$18,$C1,$BE,$EB,$51,$F2,$E7,$53,$F2,$F9,$7F,$D6,$DF
	.dc.b	$43,$FA,$FC,$3F,$18,$DB,$F4,$CE,$A7,$A9,$A2,$C7,$E7,$CF,$8A,$5D
	.dc.b	$3E,$26,$1E,$09,$F3,$E3,$E4,$AF,$E8,$D3,$97,$CD,$19,$79,$9F,$80
	.dc.b	$3A,$6B,$F4,$09,$C3,$F7,$E3,$C5,$E9,$FF,$B2,$FE,$4B,$E1,$C7,$D2
	.dc.b	$8C,$6B,$FC,$41,$B1,$9E,$87,$AB,$F6,$2F,$8C,$FF,$B6,$63,$17,$C6
	.dc.b	$B7,$2A,$DD,$F4,$8E,$06,$6B,$F9,$5B,$FD,$E0,$87,$F9,$78,$A6,$8C
	.dc.b	$10,$F8,$A7,$E9,$4A,$7D,$F3,$07,$CC,$CB,$F6,$0F,$32,$1F,$C4,$8E
	.dc.b	$E7,$F3,$1D,$FF,$CA,$9A,$FD,$D8,$FC,$36,$29,$CC,$7E,$12,$31,$3E
	.dc.b	$01,$AF,$E4,$E7,$C8,$48,$FC,$0B,$BE,$2F,$D1,$47,$E7,$E3,$FC,$10
	.dc.b	$FE,$8F,$8B,$F8,$13,$63,$27,$A8,$FE,$22,$7A,$8A,$7E,$F1,$7C,$5F
	.dc.b	$EF,$39,$BF,$C2,$8D,$8C,$9B,$7F,$7D,$C5,$FE,$92,$AE,$8D,$7E,$91
	.dc.b	$F2,$4A,$78,$0F,$9B,$3F,$C7,$BD,$EE,$C4,$7C,$8F,$6E,$3F,$F4,$A4
	.dc.b	$F2,$AE,$CF,$B5,$AF,$73,$11,$C1,$FE,$4D,$D3,$FC,$F1,$FE,$85,$7F
	.dc.b	$CF,$1F,$E8,$87,$F4,$4E,$AA,$D3,$3C,$CD,$FF,$A9,$7C,$4F,$F4,$F2
	.dc.b	$4F,$F4,$F8,$FF,$1E,$47,$C4,$C7,$C5,$3E,$52,$1F,$0C,$B9,$7A,$9C
	.dc.b	$11,$C1,$23,$9C,$65,$B1,$9E,$7F,$3D,$FF,$80,$EB,$FC,$0B,$D5,$FC
	.dc.b	$D5,$FF,$05,$37,$1F,$F0,$1C,$7F,$82,$BB,$BF,$CA,$9D,$37,$5F,$DC
	.dc.b	$CF,$E6,$E7,$2F,$F6,$D6,$BF,$43,$1F,$93,$97,$8E,$3E,$14,$63,$3F
	.dc.b	$39,$16,$FF,$E7,$62,$4F,$FB,$71,$FF,$22,$91,$FF,$1D,$B5,$FF,$18
	.dc.b	$FC,$37,$F9,$F8,$8F,$FB,$91,$FF,$26,$91,$FF,$21,$B5,$FF,$1C,$CF
	.dc.b	$FA,$18,$8F,$FB,$B1,$FF,$31,$11,$FF,$05,$78,$7F,$EF,$B8,$C4,$7F
	.dc.b	$9B,$9F,$F3,$69,$1F,$F3,$6B,$FE,$92,$F8,$59,$1F,$EF,$86,$BF,$E6
	.dc.b	$4C,$D1,$FE,$64,$24,$7F,$96,$9F,$F7,$0F,$C1,$C8,$FF,$47,$07,$0F
	.dc.b	$FB,$BC,$C1,$FE,$9F,$29,$97,$FD,$38,$E0,$FF,$18,$BF,$EA,$82,$3F
	.dc.b	$E2,$9C,$D9,$B0,$E1,$3F,$E8,$3A,$FF,$44,$11,$F8,$FE,$5F,$FA,$7E
	.dc.b	$3F,$D2,$24,$8F,$F4,$8B,$5F,$F6,$5E,$3D,$3C,$EF,$F6,$7C,$FE,$D9
	.dc.b	$B3,$FE,$D5,$9F,$0F,$9E,$BE,$D3,$FE,$DB,$6B,$FE,$E1,$6B,$FE,$10
	.dc.b	$8E,$13,$EA,$F4,$3F,$DD,$D8,$7F,$BB,$63,$1E,$8F,$0B,$FF,$7B,$37
	.dc.b	$D2,$F8,$51,$FA,$63,$FD,$F9,$9F,$F8,$73,$8F,$F8,$6C,$E9,$B1,$99
	.dc.b	$FF,$2A,$7A,$8F,$53,$11,$FE,$80,$8C,$3D,$43,$8A,$FE,$32,$D8,$FF
	.dc.b	$2F,$7A,$BF,$E8,$89,$FE,$3A,$D8,$C9,$D8,$FF,$18,$B9,$7F,$D1,$FE
	.dc.b	$B3,$D7,$D4,$23,$0F,$8D,$B5,$FC,$A9,$B1,$F9,$48,$FE,$5B,$EA,$7F
	.dc.b	$B7,$E3,$D4,$D8,$CE,$BD,$BC,$6F,$F7,$8C,$7E,$50,$95,$F9,$97,$F9
	.dc.b	$BF,$B2,$F6,$7C,$AD,$9F,$8C,$D9,$DC,$EB,$69,$5A,$6C,$FF,$42,$E7
	.dc.b	$FC,$EC,$7C,$F6,$C6,$79,$FF,$40,$D9,$FA,$9D,$3F,$8A,$8F,$F7,$41
	.dc.b	$2F,$F4,$39,$FD,$9F,$65,$EC,$FF,$28,$9F,$F3,$0C,$62,$75,$39,$94
	.dc.b	$E6,$BF,$AB,$FB,$6F,$6F,$FA,$BD,$23,$0F,$F5,$A6,$BC,$23,$71,$88
	.dc.b	$FF,$52,$35,$FF,$67,$B5,$FF,$3A,$B5,$ED,$1C,$88,$F7,$6B,$DC,$C1
	.dc.b	$0C,$10,$F0,$BC,$1A,$6C,$A6,$CF,$2B,$B6,$9B,$29,$B3,$D2,$ED,$A6
	.dc.b	$CA,$6C,$F6,$99,$29,$98,$F8,$44,$8E,$91,$A3,$CE,$43,$C6,$53,$B9
	.dc.b	$4E,$06,$88,$68,$A6,$8C,$10,$C1,$23,$C8,$FA,$5A,$FA,$63,$CF,$18
	.dc.b	$8F,$13,$5E,$ED,$7B,$35,$ED,$1C,$88,$F7,$6B,$DC,$C1,$0C,$10,$F0
	.dc.b	$BC,$1A,$6C,$A6,$CF,$2B,$B6,$9B,$29,$B3,$D2,$ED,$A6,$CF,$F4,$E4
	.dc.b	$FF,$90,$3C,$EF,$F1,$F3,$7C,$4F,$E6,$13,$FE,$96,$3F,$4F,$F0,$3F
	.dc.b	$ED,$8F,$3F,$F6,$6D,$8C,$9F,$EA,$43,$F5,$2E,$5F,$D6,$B5,$C9,$B3
	.dc.b	$F7,$4E,$5A,$67,$14,$65,$B1,$98,$C1,$1C,$12,$7C,$ED,$3C,$EE,$08
	.dc.b	$E0,$87,$A0,$E0,$57,$6D,$36,$7B,$1C,$B4,$C9,$EC,$72,$D3,$33,$EC
	.dc.b	$8C,$B6,$33,$3E,$D9,$CB,$63,$27,$98,$F0,$1E,$52,$18,$21,$E1,$32
	.dc.b	$53,$33,$E1,$72,$D7,$31,$E1,$72,$D7,$27,$8C,$E8,$48,$E4,$D8,$E0
	.dc.b	$D3,$84,$68,$91,$A6,$C6,$A3,$04,$70,$49,$F8,$ED,$3E,$3B,$82,$38
	.dc.b	$21,$EC,$38,$15,$DB,$4D,$9F,$01,$CB,$4C,$9E,$D7,$2B,$57,$C1,$B6
	.dc.b	$7A,$DF,$EB,$69,$CC,$62,$BF,$CE,$75,$BA,$D3,$63,$FC,$B1,$C3,$FE
	.dc.b	$57,$9F,$EF,$04,$8C,$F0,$F3,$44,$F8,$27,$9F,$8A,$FF,$63,$C6,$27
	.dc.b	$81,$B9,$D3,$C8,$C1,$A1,$CD,$FE,$17,$AF,$F0,$AC,$66,$30,$F1,$37
	.dc.b	$CD,$D1,$E9,$CF,$FE,$44,$AF,$F6,$CB,$63,$83,$5D,$F6,$F1,$7D,$BD
	.dc.b	$3F,$E3,$3A,$FF,$18,$C7,$96,$30,$FF,$A3,$23,$FC,$79,$C3,$C6,$3F
	.dc.b	$C8,$4D,$EA,$FE,$23,$3A,$9F,$77,$67,$C1,$8F,$F2,$8C,$71,$8C,$3F
	.dc.b	$E5,$23,$E2,$76,$F9,$DF,$A2,$D7,$FC,$BF,$8F,$86,$31,$3F,$EC,$89
	.dc.b	$EF,$E3,$F5,$23,$F3,$FE,$27,$D6,$FD,$36,$BF,$79,$A9,$BF,$07,$FE
	.dc.b	$2B,$E5,$7B,$FF,$4E,$93,$98,$C3,$63,$EC,$35,$FC,$17,$77,$FA,$9E
	.dc.b	$B9,$C7,$DD,$8C,$57,$56,$C6,$79,$7F,$45,$3F,$EA,$1F,$AE,$B3,$F8
	.dc.b	$5F,$25,$18,$8F,$C8,$46,$30,$EF,$18,$9D,$3F,$9E,$EA,$FD,$5D,$FE
	.dc.b	$A2,$B7,$39,$EA,$DB,$FB,$56,$F8,$5F,$E6,$6A,$FF,$31,$12,$7F,$56
	.dc.b	$FE,$9B,$A7,$F5,$6D,$8E,$A3,$AB,$F5,$51,$93,$F2,$64,$8F,$F6,$EB
	.dc.b	$5F,$F4,$AF,$37,$47,$F4,$DC,$DF,$D8,$A7,$F2,$44,$8F,$D5,$7E,$21
	.dc.b	$F4,$7E,$03,$ED,$36,$33,$C1,$FC,$3E,$72,$D8,$FD,$96,$3F,$BF,$9F
	.dc.b	$51,$23,$39,$7F,$08,$8F,$95,$F0,$12,$7F,$8A,$EC,$8C,$47,$F0,$5D
	.dc.b	$7F,$D1,$F1,$FE,$8B,$24,$7F,$14,$7E,$41,$1E,$E7,$32,$BE,$89,$7C
	.dc.b	$05,$24,$FB,$A3,$2F,$28,$FA,$84,$70,$6F,$DE,$E8,$47,$07,$89,$FB
	.dc.b	$84,$70,$7F,$2C,$F0,$7F,$D4,$C4,$70,$7F,$25,$D1,$F5,$CF,$C7,$3F
	.dc.b	$9C,$4F,$F3,$29,$CC,$62,$11,$8C,$7F,$98,$B7,$BB,$F9,$44,$3D,$8F
	.dc.b	$F7,$4E,$38,$6D,$FF,$35,$9D,$77,$FF,$8C,$DB,$1F,$E3,$97,$97,$D1
	.dc.b	$66,$8D,$24,$FF,$8E,$77,$FF,$64,$E1,$EF,$8F,$BD,$E2,$FA,$23,$11
	.dc.b	$FD,$76,$DF,$7B,$D4,$91,$E1,$8C,$BD,$6B,$93,$F2,$8E,$BA,$64,$FF
	.dc.b	$28,$0F,$DC,$9F,$0B,$F2,$63,$0D,$8C,$E9,$F2,$9B,$B7,$C7,$F8,$7F
	.dc.b	$EC,$7F,$85,$DC,$4D,$23,$F2,$FF,$11,$FF,$7F,$6D,$F4,$9B,$19,$C7
	.dc.b	$E7,$46,$23,$D3,$3F,$11,$6F,$F2,$9F,$71,$FE,$03,$97,$C6,$F2,$24
	.dc.b	$7D,$78,$CB,$F6,$7F,$15,$7C,$D7,$F8,$2F,$E2,$7F,$C1,$93,$FD,$FB
	.dc.b	$25,$69,$4E,$A7,$FC,$25,$39,$8C,$4F,$0A,$E8,$F9,$AB,$D3,$C3,$F5
	.dc.b	$9F,$69,$9F,$85,$73,$7F,$85,$EB,$FB,$F4,$61,$FF,$0F,$98,$8D,$4F
	.dc.b	$F8,$2E,$75,$C1,$E2,$CB,$D1,$F1,$D7,$2F,$F1,$43,$71,$34,$F6,$61
	.dc.b	$E6,$E4,$FC,$47,$07,$D1,$E0,$FA,$13,$F1,$7E,$21,$F8,$7F,$F8,$93
	.dc.b	$0F,$83,$18,$9E,$B8,$FE,$AE,$31,$3A,$AE,$13,$FA,$FF,$B9,$7C,$8F
	.dc.b	$FA,$A6,$30,$DC,$3E,$A4,$E9,$65,$F4,$23,$0F,$D2,$FC,$2D,$F2,$3F
	.dc.b	$E4,$78,$C3,$70,$FD,$F4,$E9,$E6,$62,$35,$7F,$58,$87,$F9,$0B,$4F
	.dc.b	$86,$DE,$3F,$B8,$FF,$95,$71,$F1,$DF,$AA,$3A,$9E,$AC,$BD,$8E,$DA
	.dc.b	$FD,$1C,$BE,$85,$F4,$8C,$1F,$8B,$C3,$C2,$E0,$F2,$8F,$6F,$FD,$AE
	.dc.b	$BE,$3F,$E3,$BF,$ED,$C3,$17,$FE,$53,$9D,$CE,$A7,$97,$B6,$F7,$3F
	.dc.b	$CC,$33,$9A,$DD,$F1,$AD,$5F,$43,$B5,$6E,$32,$F5,$C1,$8F,$F3,$8A
	.dc.b	$FF,$15,$46,$38,$BF,$C5,$91,$F9,$33,$33,$B7,$FC,$EA,$62,$37,$3C
	.dc.b	$63,$07,$F4,$0D,$7F,$10,$FE,$1F,$4F,$E7,$CF,$D3,$3F,$80,$14,$FD
	.dc.b	$1C,$F3,$7B,$90,$FE,$5A,$7F,$4C,$DF,$F1,$4F,$E3,$B6,$E6,$FF,$A5
	.dc.b	$E3,$31,$F8,$D3,$22,$FC,$AF,$E1,$E7,$CC,$6A,$32,$FE,$C7,$83,$FD
	.dc.b	$27,$ED,$7E,$AE,$0D,$25,$79,$E3,$F2,$D9,$7C,$6C,$FE,$54,$7A,$8F
	.dc.b	$EC,$1C,$BF,$C7,$EF,$F8,$31,$2F,$3C,$BF,$EB,$18,$FE,$0F,$CF,$FC
	.dc.b	$57,$8E,$F4,$7F,$AD,$5B,$7F,$89,$3F,$8C,$17,$1F,$01,$0F,$29,$F6
	.dc.b	$0E,$EF,$FA,$FC,$A7,$F1,$92,$1F,$D3,$7A,$FF,$0E,$BA,$FF,$9F,$CF
	.dc.b	$F4,$F9,$CC,$62,$3E,$21,$F8,$D2,$F0,$FD,$F7,$FA,$14,$7F,$24,$6D
	.dc.b	$72,$23,$FE,$81,$7F,$C1,$9D,$BF,$EC,$6F,$C2,$BF,$B2,$12,$7F,$C1
	.dc.b	$9E,$0A,$75,$1F,$E1,$FF,$03,$FB,$7E,$7F,$EB,$DE,$2F,$EE,$78,$FE
	.dc.b	$B2,$BF,$C4,$15,$A7,$87,$83,$F3,$DF,$4F,$1F,$DC,$7E,$59,$F8,$1A
	.dc.b	$FF,$10,$78,$5F,$1B,$3F,$F0,$4C,$FF,$88,$63,$07,$EC,$1F,$88,$7D
	.dc.b	$52,$9D,$0F,$EC,$D1,$FE,$20,$7E,$13,$5F,$F1,$67,$87,$DC,$F0,$70
	.dc.b	$C8,$DB,$5F,$81,$E2,$FF,$8F,$F0,$F8,$11,$FE,$3C,$8C,$6B,$AF,$17
	.dc.b	$E0,$4F,$F9,$0E,$72,$FC,$2D,$BF,$3C,$47,$FD,$1D,$EB,$C7,$FC,$CE
	.dc.b	$D7,$FD,$E9,$E6,$FF,$AC,$E3,$FC,$D0,$48,$FF,$33,$3F,$E9,$7F,$88
	.dc.b	$91,$FE,$70,$6B,$CA,$99,$27,$FC,$E4,$49,$FF,$38,$3F,$EA,$19,$91
	.dc.b	$FE,$7A,$6B,$FE,$3A,$7F,$8F,$11,$19,$3F,$D1,$6F,$FA,$DC,$8E,$0F
	.dc.b	$F4,$2F,$C4,$7F,$AB,$C8,$E0,$FF,$41,$3F,$E8,$72,$38,$3F,$D5,$7E
	.dc.b	$97,$83,$E1,$7F,$DF,$84,$8C,$87,$FD,$C9,$EE,$FF,$8B,$1B,$8F,$19
	.dc.b	$F4,$9E,$A9,$D3,$80,$E1,$DF,$A4,$F4,$B9,$46,$1A,$FC,$48,$D3,$E3
	.dc.b	$3F,$D4,$B7,$FD,$1C,$8F,$7E,$BF,$EA,$11,$88,$F2,$FA,$CF,$5B,$ED
	.dc.b	$CF,$FB,$8C,$93,$FE,$BA,$7F,$DE,$59,$77,$8F,$F7,$63,$5F,$AF,$EC
	.dc.b	$7F,$BC,$67,$67,$FB,$F7,$6E,$B3,$FE,$FD,$AE,$0B,$D9,$FF,$4E,$C6
	.dc.b	$1B,$5F,$1E,$5E,$D3,$E0,$39,$C7,$CE,$8C,$5F,$46,$D7,$16,$CF,$D1
	.dc.b	$F7,$3F,$CC,$7D,$1D,$5B,$19,$BF,$C8,$CE,$5B,$1F,$67,$DD,$73,$62
	.dc.b	$3E,$FC,$61,$FD,$E8,$71,$5E,$B6,$C7,$E0,$FD,$EF,$F3,$2B,$C1,$B3
	.dc.b	$F0,$5B,$1B,$EB,$F0,$3F,$A6,$AF,$C7,$D7,$58,$C9,$E1,$2C,$69,$FC
	.dc.b	$DC,$7C,$DF,$75,$CF,$C1,$B3,$EA,$6C,$FE,$7A,$3D,$BC,$FF,$21,$FC
	.dc.b	$FB,$5F,$D0,$67,$E2,$F7,$FD,$31,$FA,$46,$BF,$BE,$6A,$6F,$C0,$7F
	.dc.b	$80,$23,$E5,$B6,$FF,$08,$D8,$FD,$63,$5F,$D4,$1F,$4B,$A3,$F4,$F3
	.dc.b	$F8,$B6,$C7,$EC,$DA,$FF,$B3,$23,$FC,$60,$47,$F5,$5E,$17,$FB,$CA
	.dc.b	$3F,$29,$D1,$FA,$9C,$7E,$FB,$63,$3D,$1F,$CE,$AF,$F8,$DC,$E5,$FE
	.dc.b	$05,$CB,$F3,$9F,$CF,$35,$CE,$DF,$D5,$E3,$11,$FC,$31,$AF,$F2,$66
	.dc.b	$BF,$A5,$8E,$04,$7F,$C9,$F1,$A2,$46,$63,$C2,$D7,$C2,$70,$70,$F8
	.dc.b	$5A,$F8,$5A,$F8,$4E,$04,$3A,$94,$EA,$60,$86,$08,$77,$38,$14,$D9
	.dc.b	$4D,$9E,$33,$65,$36,$53,$67,$9C,$C9,$4C,$C7,$C6,$79,$99,$3C,$04
	.dc.b	$3A,$14,$D1,$82,$18,$23,$D5,$F0,$B5,$F0,$C7,$CC,$8C,$47,$56,$BD
	.dc.b	$1A,$F4,$8E,$04,$7D,$31,$A2,$46,$63,$C2,$D7,$C2,$70,$70,$F8,$5A
	.dc.b	$F8,$5A,$F8,$4E,$04,$3A,$94,$EA,$60,$86,$08,$77,$3E,$8F,$A9,$FE
	.dc.b	$C0,$7F,$B5,$35,$FF,$5B,$61,$FE,$AC,$8C,$47,$E9,$7C,$4F,$F0,$5F
	.dc.b	$A9,$FA,$E6,$C7,$EC,$9A,$FE,$B5,$AF,$13,$FD,$90,$58,$E4,$E0,$FE
	.dc.b	$A9,$1C,$1A,$67,$04,$65,$B1,$98,$C1,$1C,$12,$7B,$91,$F4,$19,$2B
	.dc.b	$97,$D4,$64,$AE,$5F,$51,$C0,$86,$8A,$68,$C1,$0C,$10,$F7,$BD,$5A
	.dc.b	$72,$29,$C4,$A7,$19,$F8,$71,$C5,$B1,$C5,$AF,$16,$BC,$67,$E4,$4F
	.dc.b	$16,$C7,$16,$BC,$5A,$F1,$3E,$31,$63,$93,$83,$CE,$49,$F0,$B6,$79
	.dc.b	$B4,$E1,$1A,$24,$69,$B1,$A8,$C1,$1C,$12,$7C,$64,$7D,$86,$4A,$E5
	.dc.b	$F7,$19,$2B,$97,$DC,$70,$21,$A2,$9A,$30,$43,$FD,$DF,$EC,$7F,$8D
	.dc.b	$7B,$5B,$3B,$13,$98,$C3,$63,$35,$FD,$C7,$E0,$3F,$CF,$F3,$FD,$D5
	.dc.b	$B1,$93,$67,$F9,$99,$CB,$FE,$41,$F8,$5A,$FF,$A2,$3B,$11,$88,$FF
	.dc.b	$24,$B6,$33,$1D,$0F,$F4,$7F,$23,$E5,$78,$36,$34,$1C,$61,$FE,$12
	.dc.b	$9E,$DC,$DF,$EC,$78,$FF,$2C,$36,$BF,$D0,$0D,$7F,$C4,$5E,$3F,$97
	.dc.b	$A1,$E7,$52,$4F,$82,$32,$F8,$BA,$BF,$84,$E1,$FE,$39,$E2,$F7,$46
	.dc.b	$27,$FD,$1C,$D7,$FC,$AB,$DD,$F0,$FB,$3F,$CA,$D7,$FE,$53,$8C,$1F
	.dc.b	$1F,$A1,$EB,$D2,$5F,$C2,$8C,$B7,$87,$81,$1C,$1F,$23,$87,$FC,$EB
	.dc.b	$F0,$5F,$E7,$3E,$8F,$F6,$4B,$7B,$4D,$3F,$D4,$DF,$16,$F6,$9E,$F7
	.dc.b	$0C,$7F,$D5,$B3,$98,$C4,$7F,$B2,$23,$17,$D2,$37,$5C,$A3,$11,$FE
	.dc.b	$9A,$BF,$F4,$E9,$1F,$F5,$B7,$8E,$E9,$EB,$0C,$3C,$9F,$55,$86,$E1
	.dc.b	$D2,$75,$8A,$B4,$D9,$D4,$BB,$7F,$53,$D9,$86,$E7,$FA,$72,$4F,$E8
	.dc.b	$DF,$D3,$3C,$BC,$3F,$33,$73,$FF,$74,$CF,$FA,$87,$82,$CF,$FB,$47
	.dc.b	$83,$F4,$33,$98,$F1,$C6,$1F,$F6,$B7,$47,$E4,$27,$FD,$BF,$D3,$CA
	.dc.b	$7F,$CF,$ED,$7F,$DB,$CD,$5D,$1F,$D7,$A7,$F2,$D1,$F0,$23,$0F,$E5
	.dc.b	$BA,$78,$7D,$37,$FA,$A7,$E1,$7F,$D8,$0D,$9F,$F7,$43,$5F,$F7,$6B
	.dc.b	$5F,$E0,$19,$94,$FB,$48,$FC,$18,$D4,$7F,$13,$F8,$59,$3F,$E7,$EE
	.dc.b	$04,$62,$3F,$88,$18,$3F,$8A,$3F,$48,$8F,$53,$47,$F1,$52,$9E,$62
	.dc.b	$9F,$C5,$0B,$7F,$EC,$9F,$85,$F8,$3D,$7E,$B6,$C9,$A7,$D9,$23,$E1
	.dc.b	$6B,$F1,$A3,$E2,$7C,$2F,$22,$47,$CC,$8C,$BE,$98,$FC,$21,$1C,$1F
	.dc.b	$17,$C4,$FE,$74,$47,$07,$EB,$BE,$5B,$F3,$E4,$E6,$FD,$27,$4F,$EE
	.dc.b	$E3,$F5,$70,$F6,$BF,$A4,$D3,$F3,$FF,$B4,$FD,$B9,$D7,$8F,$F7,$67
	.dc.b	$32,$9C,$BC,$4E,$2E,$23,$6F,$7F,$BB,$FF,$1E,$B6,$7F,$C7,$AD,$CB
	.dc.b	$78,$2B,$FC,$79,$D7,$FB,$69,$F8,$17,$E6,$8C,$3F,$08,$F4,$53,$E1
	.dc.b	$3B,$B6,$3B,$B6,$BB,$CF,$6C,$BF,$B5,$47,$6A,$FC,$E5,$F6,$7C,$13
	.dc.b	$A7,$F8,$EF,$DA,$7F,$0E,$26,$5F,$43,$B9,$1F,$43,$2F,$EC,$3F,$AA
	.dc.b	$7B,$B9,$FF,$50,$DC,$0D,$3E,$1F,$D8,$FA,$89,$1E,$A7,$D1,$CF,$FE
	.dc.b	$63,$D3,$CD,$3E,$8D,$3C,$CF,$A2,$7E,$0F,$E1,$5F,$10,$91,$83,$94
	.dc.b	$F0,$79,$4E,$09,$B7,$C5,$8F,$E5,$53,$B6,$CE,$9F,$C3,$FC,$13,$CB
	.dc.b	$FE,$A3,$C1,$86,$CF,$E4,$27,$53,$FD,$50,$D1,$63,$D5,$87,$87,$2F
	.dc.b	$7C,$78,$23,$18,$F8,$E7,$27,$C4,$AF,$D0,$4F,$F1,$7E,$4F,$CD,$5F
	.dc.b	$E2,$23,$13,$F9,$AD,$D1,$89,$D0,$AE,$2F,$E8,$E5,$CD,$FE,$A0,$BE
	.dc.b	$4D,$9F,$CA,$B5,$E0,$D8,$7C,$47,$F8,$F1,$F0,$9F,$AC,$71,$87,$82
	.dc.b	$37,$97,$EC,$E3,$13,$A9,$FD,$25,$FD,$A9,$F8,$BA,$FE,$8B,$0F,$06
	.dc.b	$3E,$8A,$EE,$79,$8A,$79,$8A,$7C,$F3,$A9,$4F,$AA,$63,$3F,$A3,$97
	.dc.b	$D3,$3F,$0D,$E8,$7F,$83,$7D,$1F,$E9,$1E,$67,$CE,$BF,$E9,$53,$97
	.dc.b	$FC,$47,$9F,$FB,$39,$A6,$B9,$97,$CA,$75,$7C,$E9,$7C,$E3,$0D,$9D
	.dc.b	$9F,$A5,$9F,$F3,$31,$0F,$EA,$85,$34,$60,$87,$C7,$3E,$A6,$5F,$2D
	.dc.b	$B9,$FF,$24,$75,$AF,$07,$F5,$A7,$52,$9F,$7F,$3F,$CA,$C5,$3F,$9A
	.dc.b	$47,$F3,$5D,$BF,$0C,$FE,$02,$76,$F6,$21,$F6,$4F,$19,$C8,$FC,$C6
	.dc.b	$1F,$4A,$BF,$40,$D8,$FC,$FE,$5E,$C8,$C4,$79,$8F,$86,$43,$04,$3F
	.dc.b	$81,$9E,$D2,$9F,$BC,$F5,$FF,$52,$B8,$7F,$80,$C6,$1B,$7B,$3E,$44
	.dc.b	$F2,$3C,$05,$3C,$1B,$FF,$8F,$1F,$03,$5F,$DE,$71,$FF,$5A,$7F,$A3
	.dc.b	$B5,$FD,$D3,$67,$FB,$33,$98,$EC,$49,$E8,$FC,$19,$FE,$D9,$C9,$C7
	.dc.b	$63,$49,$3E,$19,$FE,$E5,$CD,$FE,$29,$FC,$17,$FA,$11,$27,$F8,$94
	.dc.b	$65,$FE,$9F,$2C,$FF,$CE,$FC,$7F,$D3,$A3,$13,$A9,$CF,$47,$F7,$77
	.dc.b	$E1,$F0,$7F,$B1,$E6,$F2,$F8,$F4,$FE,$CD,$1F,$D9,$0C,$74,$FA,$CF
	.dc.b	$85,$B7,$C0,$8F,$9F,$7C,$DF,$CE,$65,$FD,$F5,$FE,$F7,$D5,$F0,$C8
	.dc.b	$F7,$9F,$F2,$47,$E8,$1F,$E0,$06,$C7,$F3,$16,$BF,$DF,$1A,$FF,$9F
	.dc.b	$BA,$7D,$8F,$F8,$86,$FE,$0B,$FE,$93,$DB,$F9,$B4,$62,$38,$35,$FB
	.dc.b	$55,$E9,$7F,$17,$8F,$B5,$FF,$23,$77,$7F,$9D,$C8,$FF,$93,$DF,$F2
	.dc.b	$29,$1F,$F1,$E0,$70,$FF,$94,$FD,$DF,$F4,$01,$1F,$F2,$C3,$FE,$4E
	.dc.b	$23,$C9,$AF,$F9,$6F,$6F,$F4,$41,$1F,$F3,$03,$FE,$94,$23,$FE,$F9
	.dc.b	$6B,$FE,$01,$7F,$41,$5F,$EA,$06,$D9,$F8,$53,$FD,$E9,$FF,$4B,$7B
	.dc.b	$B2,$3F,$CD,$CD,$7F,$CB,$C1,$FF,$33,$91,$F8,$6F,$FA,$4B,$82,$47
	.dc.b	$FA,$44,$FA,$5E,$33,$C6,$FF,$45,$4F,$FA,$82,$7E,$B7,$93,$FE,$F2
	.dc.b	$8C,$12,$3F,$D0,$0F,$FA,$AC,$8E,$0F,$F2,$BF,$99,$FE,$B6,$8F,$F5
	.dc.b	$C9,$23,$3E,$67,$81,$FF,$49,$47,$C4,$CB,$F0,$DA,$7F,$B1,$A3,$07
	.dc.b	$EC,$3C,$DF,$C8,$D7,$E3,$DA,$FF,$B4,$9B,$E3,$7F,$B5,$A7,$F8,$87
	.dc.b	$9D,$FE,$DD,$E2,$FC,$93,$63,$39,$72,$C1,$39,$7F,$DD,$7E,$7F,$4E
	.dc.b	$4F,$F7,$65,$FF,$02,$8C,$1C,$7C,$F7,$A1,$49,$97,$E8,$63,$2F,$E8
	.dc.b	$FD,$17,$77,$FA,$EF,$C8,$FE,$4D,$18,$CF,$FB,$79,$FE,$82,$71,$1A
	.dc.b	$9E,$13,$9F,$16,$BF,$E8,$CA,$46,$23,$FD,$91,$18,$7A,$9B,$8D,$2F
	.dc.b	$4F,$F5,$8E,$8F,$F6,$83,$8B,$FE,$78,$D8,$FF,$4B,$B5,$FF,$4E,$F3
	.dc.b	$FE,$3D,$E4,$7F,$BA,$39,$FB,$C6,$6B,$FD,$D2,$7E,$4C,$93,$FE,$F6
	.dc.b	$F4,$BC,$BD,$07,$CF,$EA,$7F,$97,$E7,$AB,$6B,$FA,$83,$63,$FD,$56
	.dc.b	$D7,$FD,$63,$EB,$3D,$6F,$82,$D8,$F3,$B6,$BF,$AD,$36,$3F,$DC,$51
	.dc.b	$FD,$77,$83,$EA,$C6,$08,$7F,$BB,$23,$FB,$17,$5B,$9F,$E7,$B6,$7E
	.dc.b	$2B,$67,$F4,$84,$8F,$F7,$C3,$5F,$F6,$4F,$1F,$CD,$7F,$B7,$35,$FF
	.dc.b	$64,$B5,$DE,$DF,$DB,$63,$10,$3F,$33,$19,$23,$FD,$DE,$36,$E8,$C4
	.dc.b	$69,$F0,$B5,$F0,$B5,$F0,$9C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53
	.dc.b	$46,$08,$60,$87,$BE,$3D,$64,$8F,$34,$65,$AF,$BC,$EC,$64,$F2,$94
	.dc.b	$C9,$B3,$C2,$64,$A6,$4F,$19,$4D,$18,$8F,$23,$D8,$D1,$1C,$BE,$06
	.dc.b	$BD,$9A,$F3,$8F,$5C,$62,$3B,$90,$E9,$1A,$23,$E1,$8E,$0E,$8C,$46
	.dc.b	$9F,$13,$5F,$13,$5F,$11,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34
	.dc.b	$60,$86,$08,$7B,$FC,$DF,$F5,$03,$FA,$13,$FD,$3D,$5F,$E5,$58,$CC
	.dc.b	$7F,$A8,$CD,$0F,$C1,$9E,$2F,$EA,$7C,$CF,$F2,$F4,$62,$7F,$D4,$87
	.dc.b	$EB,$5C,$46,$A7,$84,$F4,$9C,$B6,$32,$F9,$63,$2D,$8C,$BD,$5A,$F9
	.dc.b	$5B,$1C,$23,$47,$90,$D1,$5D,$38,$9F,$33,$99,$E6,$47,$04,$8D,$91
	.dc.b	$F5,$19,$2B,$97,$D4,$64,$AE,$5F,$3B,$5F,$21,$C8,$87,$02,$9C,$0A
	.dc.b	$70,$34,$43,$45,$34,$60,$86,$09,$3F,$18,$93,$F0,$E7,$CD,$18,$9F
	.dc.b	$61,$C9,$C4,$6A,$78,$4F,$9A,$72,$D8,$CB,$E5,$8C,$B6,$32,$F5,$6B
	.dc.b	$E6,$6C,$70,$8D,$1E,$43,$45,$74,$E2,$7C,$CE,$67,$99,$1C,$12,$36
	.dc.b	$47,$D4,$64,$AE,$5F,$51,$F9,$3F,$CD,$9D,$CE,$97,$B1,$FE,$C9,$9C
	.dc.b	$C6,$2B,$F6,$71,$89,$D5,$71,$7F,$DA,$1F,$A1,$3D,$2C,$5F,$EE,$A3
	.dc.b	$18,$73,$8F,$E0,$51,$8A,$E3,$5C,$27,$F8,$37,$84,$F1,$7F,$CF,$55
	.dc.b	$FE,$59,$24,$FF,$06,$9D,$78,$9D,$63,$D9,$E5,$BE,$83,$FB,$87,$AB
	.dc.b	$FC,$33,$44,$E6,$93,$98,$C3,$6B,$8C,$FA,$F3,$F9,$1B,$FA,$30,$FD
	.dc.b	$F6,$BE,$E8,$C1,$DF,$9F,$FD,$25,$3F,$C5,$8F,$F0,$A6,$7C,$B0,$FE
	.dc.b	$49,$18,$35,$3C,$2B,$FD,$C7,$F8,$BF,$D7,$F9,$6B,$46,$92,$F9,$FE
	.dc.b	$B3,$FE,$F6,$9D,$46,$3D,$3E,$4D,$8C,$BF,$E1,$55,$3C,$6B,$E2,$7C
	.dc.b	$1F,$F8,$67,$C0,$B6,$69,$1F,$5F,$0B,$E4,$B8,$46,$1A,$F7,$8D,$3C
	.dc.b	$CD,$14,$F2,$7E,$B5,$FE,$6E,$C9,$EE,$FD,$B8,$C3,$7C,$8F,$CE,$36
	.dc.b	$77,$5F,$D3,$3E,$69,$EE,$6A,$30,$D7,$FA,$B4,$6A,$78,$C6,$27,$4F
	.dc.b	$EA,$70,$FF,$33,$4E,$A3,$1E,$3E,$A1,$3A,$C3,$D1,$39,$C9,$F3,$9F
	.dc.b	$8E,$C3,$E2,$35,$FC,$DB,$70,$FC,$CB,$7F,$5A,$FF,$6F,$4F,$E6,$8F
	.dc.b	$C5,$B8,$EA,$FD,$03,$63,$3C,$FD,$7D,$37,$A4,$F9,$AF,$E0,$9F,$6B
	.dc.b	$FD,$CA,$73,$18,$BF,$EF,$51,$8C,$D5,$BB,$E5,$18,$AD,$C3,$D1,$E5
	.dc.b	$1F,$E6,$3B,$F0,$C6,$21,$18,$AE,$51,$BA,$E5,$18,$8F,$F7,$B7,$E7
	.dc.b	$4F,$93,$FE,$39,$AF,$F8,$2B,$D8,$4E,$6F,$E3,$5F,$F9,$CE,$73,$7C
	.dc.b	$23,$FC,$2D,$18,$9D,$C7,$F9,$E7,$C3,$FE,$1F,$87,$FB,$AC,$91,$FC
	.dc.b	$4F,$C5,$6F,$C3,$07,$EB,$1F,$90,$AC,$36,$FC,$CD,$8F,$0B,$5F,$D3
	.dc.b	$76,$7F,$A2,$7C,$5F,$D8,$92,$3F,$5F,$19,$7F,$53,$1F,$C6,$FB,$31
	.dc.b	$1F,$AF,$6B,$FC,$99,$BA,$7E,$32,$1F,$A5,$7F,$23,$E6,$FF,$4C,$36
	.dc.b	$32,$7E,$FA,$32,$FF,$10,$FC,$33,$E5,$3F,$C2,$FC,$14,$62,$3F,$A8
	.dc.b	$46,$2F,$E6,$9F,$C4,$27,$5A,$FD,$1C,$3E,$D7,$17,$D9,$6F,$17,$F5
	.dc.b	$96,$DF,$93,$1F,$E9,$3C,$3F,$4A,$7F,$C9,$70,$E2,$FE,$8F,$18,$6C
	.dc.b	$7D,$9E,$2E,$4B,$2F,$B3,$DB,$EE,$8F,$C4,$71,$7D,$BC,$3E,$CC,$E6
	.dc.b	$78,$4E,$5F,$D0,$F9,$FF,$DB,$FF,$59,$FD,$33,$63,$3A,$27,$2D,$8F
	.dc.b	$C1,$DF,$FB,$26,$7B,$C6,$4F,$1C,$65,$D9,$93,$F5,$7C,$CB,$FF,$00
	.dc.b	$7E,$B5,$D8,$93,$E2,$7A,$91,$E6,$D7,$E6,$E9,$FE,$5D,$D7,$FB,$8F
	.dc.b	$8F,$F2,$F7,$FD,$0B,$67,$FC,$EB,$D3,$FC,$3B,$F0,$6F,$EC,$18,$7E
	.dc.b	$ED,$AF,$99,$B1,$BD,$BF,$B4,$77,$3E,$73,$FC,$E7,$08,$C5,$EA,$73
	.dc.b	$18,$C7,$C4,$49,$FB,$CD,$7D,$6D,$7F,$7D,$97,$A9,$D9,$FC,$B3,$E9
	.dc.b	$BA,$91,$F2,$BF,$E8,$4D,$3E,$1C,$70,$3A,$94,$FA,$7E,$47,$BB,$83
	.dc.b	$FB,$D6,$7F,$19,$B8,$FF,$2B,$9F,$8B,$3E,$F9,$57,$DB,$3F,$9F,$B8
	.dc.b	$E0,$FE,$7E,$D8,$C9,$B3,$F9,$87,$EA,$BF,$DF,$FE,$13,$F9,$F3,$63
	.dc.b	$F9,$EB,$5D,$9F,$E4,$78,$F8,$CF,$EA,$F4,$F8,$CF,$B8,$87,$F3,$A9
	.dc.b	$7C,$27,$DC,$6E,$3F,$84,$6E,$7E,$29,$F7,$FC,$2B,$F1,$3F,$F2,$B7
	.dc.b	$5F,$F9,$5F,$1F,$F0,$B4,$E7,$1F,$E7,$91,$8A,$DC,$7F,$8B,$27,$55
	.dc.b	$C6,$BE,$EF,$2F,$E6,$73,$F4,$C6,$32,$FC,$17,$23,$47,$D6,$7F,$90
	.dc.b	$33,$F8,$AD,$BF,$96,$D7,$E0,$36,$19,$F9,$FC,$0F,$14,$63,$5F,$F3
	.dc.b	$94,$62,$75,$3E,$99,$CB,$F8,$9B,$FA,$1A,$7A,$49,$9F,$F7,$EF,$43
	.dc.b	$F0,$B8,$F6,$E6,$46,$30,$FF,$02,$B7,$0F,$F2,$04,$E9,$FF,$09,$CF
	.dc.b	$FB,$DF,$EC,$5E,$EB,$E2,$BF,$DE,$53,$98,$C5,$32,$F9,$D7,$C2,$F7
	.dc.b	$3F,$98,$8C,$3B,$BE,$DF,$2F,$E6,$71,$1B,$8C,$5F,$46,$CF,$0C,$BE
	.dc.b	$85,$7C,$4A,$FC,$AF,$4F,$EC,$A3,$F2,$78,$78,$27,$CB,$97,$A1,$B9
	.dc.b	$79,$A9,$5B,$C7,$E8,$61,$CE,$7F,$0C,$63,$8F,$F1,$2F,$84,$C6,$DF
	.dc.b	$F0,$07,$DD,$FF,$16,$FB,$DF,$B8,$F5,$3E,$35,$7E,$EA,$72,$A3,$2A
	.dc.b	$33,$18,$6E,$9F,$C0,$A7,$5D,$4B,$ED,$39,$AD,$D7,$7A,$D3,$DE,$75
	.dc.b	$18,$BF,$0E,$5F,$EA,$CA,$FE,$74,$43,$FB,$31,$4F,$E8,$65,$3F,$5A
	.dc.b	$79,$4F,$97,$B7,$E1,$9F,$9E,$7F,$A3,$CF,$C2,$39,$D3,$F1,$9B,$7F
	.dc.b	$A9,$27,$A9,$E0,$21,$83,$F0,$DB,$FF,$05,$9F,$C2,$1D,$63,$ED,$F0
	.dc.b	$7E,$35,$FC,$54,$ED,$F0,$D7,$77,$CE,$DE,$0F,$CB,$C7,$A0,$FD,$64
	.dc.b	$FF,$32,$F5,$EC,$1A,$4C,$3C,$F1,$F1,$0A,$7E,$E8,$FF,$6A,$72,$FF
	.dc.b	$1B,$9F,$E0,$5C,$BF,$EF,$56,$D7,$53,$F8,$21,$6B,$F8,$8B,$EF,$E8
	.dc.b	$FE,$31,$CF,$ED,$9F,$E2,$73,$9E,$8F,$E3,$6F,$C5,$8F,$11,$E4,$DF
	.dc.b	$F0,$0F,$A2,$3F,$8B,$0F,$AF,$F1,$91,$F1,$BF,$CB,$70,$FE,$51,$D7
	.dc.b	$FC,$BE,$BF,$A7,$B5,$FE,$88,$D8,$DD,$7F,$8D,$A7,$C6,$7D,$D3,$E4
	.dc.b	$16,$3F,$9A,$28,$FD,$43,$FD,$60,$C7,$17,$F5,$C6,$99,$8D,$BF,$C8
	.dc.b	$38,$FF,$B5,$FC,$3F,$F6,$88,$34,$8F,$FA,$6B,$BF,$D1,$3F,$DE,$23
	.dc.b	$FB,$B5,$7F,$93,$B0,$FF,$55,$D7,$BE,$7D,$67,$A8,$86,$27,$FC,$35
	.dc.b	$E0,$7C,$16,$C7,$F9,$61,$AF,$FA,$EA,$3E,$33,$D0,$9A,$7D,$87,$E4
	.dc.b	$18,$EB,$FF,$62,$BF,$75,$AF,$F4,$0F,$07,$C1,$DB,$FE,$2A,$8F,$F6
	.dc.b	$3C,$60,$FA,$99,$7F,$8C,$E3,$FD,$7D,$E1,$7C,$22,$65,$F6,$71,$F1
	.dc.b	$4F,$B9,$AF,$C2,$F2,$3F,$CD,$6B,$CB,$FF,$2D,$91,$FF,$2C,$BF,$E6
	.dc.b	$22,$3F,$E6,$66,$BF,$E6,$03,$FD,$DA,$7F,$98,$3D,$6F,$F3,$29,$1F
	.dc.b	$F3,$6B,$FE,$6C,$23,$83,$FC,$CF,$FA,$07,$F9,$B0,$8E,$0F,$F4,$23
	.dc.b	$FE,$5B,$23,$83,$FD,$8F,$E6,$7F,$B3,$63,$FD,$26,$48,$CF,$8B,$F9
	.dc.b	$78,$FE,$49,$97,$FA,$7C,$FF,$02,$CF,$FA,$A5,$AF,$FA,$A7,$CD,$FC
	.dc.b	$F7,$67,$0C,$BF,$D5,$0F,$FA,$CE,$E5,$7F,$AC,$C3,$87,$FD,$80,$7F
	.dc.b	$86,$8F,$E9,$16,$F5,$5E,$7F,$F9,$33,$B3,$FD,$91,$5B,$38,$BE,$89
	.dc.b	$7A,$3B,$9F,$F6,$89,$B3,$F6,$91,$A7,$17,$DE,$78,$47,$D9,$C7,$FD
	.dc.b	$36,$47,$FD,$32,$FF,$A2,$48,$FF,$AD,$5A,$FF,$BE,$75,$FF,$7E,$47
	.dc.b	$FB,$08,$91,$F4,$2B,$E5,$BF,$EA,$2F,$4F,$F8,$47,$DB,$FF,$55,$E4
	.dc.b	$75,$27,$35,$FD,$B6,$73,$18,$8F,$F3,$8C,$62,$D7,$CE,$DE,$B6,$EF
	.dc.b	$FD,$62,$DB,$EE,$D9,$DD,$F2,$AF,$AB,$D0,$F2,$FF,$CF,$F9,$FE,$22
	.dc.b	$31,$3E,$CA,$FF,$51,$56,$A3,$8B,$5F,$F4,$3C,$7F,$BF,$6B,$FD,$F4
	.dc.b	$48,$FC,$43,$F9,$4E,$57,$AF,$FC,$B7,$ED,$F0,$DA,$FC,$4C,$EA,$D5
	.dc.b	$A6,$CF,$F8,$07,$DA,$7A,$7C,$63,$FC,$31,$18,$9D,$C6,$27,$4F,$FA
	.dc.b	$7A,$75,$5E,$5F,$0B,$36,$C3,$FC,$50,$F2,$BF,$E8,$91,$8A,$E1,$39
	.dc.b	$3F,$41,$CD,$FE,$AA,$EE,$FC,$F9,$2F,$F4,$BC,$DF,$A4,$7F,$92,$1F
	.dc.b	$9C,$E6,$E0,$FE,$82,$7F,$49,$18,$8F,$9F,$3F,$A1,$9F,$DC,$7B,$9F
	.dc.b	$D5,$FE,$46,$E2,$69,$2F,$F2,$1C,$7F,$E6,$38,$FC,$67,$95,$C6,$72
	.dc.b	$D7,$F2,$73,$B7,$F7,$BE,$4F,$E2,$6F,$FA,$53,$63,$F8,$8B,$5D,$F2
	.dc.b	$7F,$57,$F7,$9E,$FD,$25,$FF,$75,$8C,$BF,$D5,$75,$7B,$7F,$8F,$C3
	.dc.b	$F8,$0B,$67,$F8,$73,$67,$73,$FB,$43,$FD,$37,$EA,$7F,$91,$DA,$7F
	.dc.b	$92,$7C,$57,$8F,$FA,$39,$D4,$63,$3F,$F2,$DD,$7F,$98,$F2,$F2,$D7
	.dc.b	$36,$C7,$F9,$4D,$AE,$F3,$FF,$05,$4F,$87,$E0,$9F,$05,$8A,$FF,$2D
	.dc.b	$46,$35,$EF,$1F,$EF,$28,$C5,$71,$AE,$13,$FE,$75,$F8,$47,$C2,$7F
	.dc.b	$78,$6C,$7F,$87,$DA,$FC,$16,$BF,$E6,$96,$BC,$E3,$91,$1E,$4D,$79
	.dc.b	$35,$E4,$D7,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86
	.dc.b	$08,$04,$7B,$9C,$35,$CC,$7A,$5C,$B5,$C9,$E6,$32,$53,$27,$94,$C9
	.dc.b	$4C,$9E,$63,$25,$32,$7A,$0A,$68,$C4,$79,$C9,$1E,$56,$C7,$8D,$AF
	.dc.b	$85,$AF,$76,$BD,$23,$91,$1E,$4D,$79,$35,$E4,$D7,$91,$C0,$87,$02
	.dc.b	$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$0F,$91,$EC,$7F,$B7,$9F,$F6
	.dc.b	$23,$5F,$F6,$8D,$FF,$A0,$63,$11,$F8,$2F,$C1,$3F,$0D,$EC,$FE,$95
	.dc.b	$B1,$FA,$56,$BF,$AF,$6B,$C4,$FF,$75,$96,$39,$38,$3F,$23,$1C,$1A
	.dc.b	$67,$04,$65,$B1,$98,$C1,$1C,$12,$7B,$91,$F4,$19,$2B,$97,$D4,$64
	.dc.b	$AE,$5F,$51,$C0,$86,$8A,$68,$C1,$0C,$10,$F7,$BD,$5A,$72,$29,$C4
	.dc.b	$A7,$19,$F8,$71,$C5,$B1,$C5,$AF,$16,$BC,$67,$E4,$4F,$16,$C7,$16
	.dc.b	$BC,$5A,$F1,$3E,$31,$63,$93,$83,$CE,$49,$F0,$B6,$79,$B4,$E1,$1A
	.dc.b	$24,$69,$B1,$A8,$C1,$1C,$12,$7C,$64,$7D,$86,$4A,$E5,$F7,$19,$2B
	.dc.b	$97,$FC,$A3,$E0,$7F,$B8,$7E,$17,$FD,$A0,$D8,$FF,$2E,$35,$FF,$2D
	.dc.b	$35,$FE,$61,$BF,$F7,$A9,$FE,$CD,$BE,$65,$39,$7F,$B4,$46,$78,$7F
	.dc.b	$B8,$57,$F7,$2E,$2F,$F2,$B3,$78,$8D,$3F,$BE,$7C,$EB,$3F,$E8,$11
	.dc.b	$86,$E3,$FD,$0C,$FE,$89,$3B,$79,$99,$0E,$67,$FC,$11,$FA,$DF,$F8
	.dc.b	$52,$AE,$4E,$4F,$DF,$4E,$70,$E4,$D7,$C1,$9F,$F8,$2E,$FD,$44,$8F
	.dc.b	$F0,$96,$7F,$E7,$38,$FF,$0B,$7D,$77,$F8,$E6,$7F,$85,$36,$33,$D7
	.dc.b	$FC,$2E,$31,$3F,$E7,$CF,$80,$FE,$C1,$EA,$FF,$9E,$DB,$19,$11,$96
	.dc.b	$9F,$E8,$4F,$53,$6F,$F9,$F3,$C9,$46,$23,$E0,$46,$1E,$26,$E3,$5A
	.dc.b	$BF,$24,$FE,$3A,$BE,$3B,$FD,$AB,$18,$6F,$37,$CB,$6C,$EE,$FF,$D4
	.dc.b	$DE,$DB,$E5,$F5,$18,$6B,$FE,$A8,$8D,$4F,$18,$C4,$E9,$FB,$1F,$06
	.dc.b	$F7,$3F,$C0,$DF,$15,$FE,$E8,$8C,$4A,$31,$3A,$8F,$F6,$EC,$EA,$7F
	.dc.b	$C9,$9F,$04,$FA,$0D,$57,$E3,$E3,$0F,$33,$F1,$CE,$27,$8C,$F0,$9F
	.dc.b	$2F,$B1,$72,$61,$FE,$ED,$BF,$34,$E6,$31,$5F,$90,$F7,$3F,$29,$CD
	.dc.b	$F9,$86,$C6,$70,$FC,$BC,$62,$3F,$49,$CB,$F8,$A6,$F3,$7E,$34,$8F
	.dc.b	$F5,$EA,$F1,$FE,$4B,$FE,$2D,$F8,$CF,$90,$D9,$F8,$AD,$7F,$D8,$B9
	.dc.b	$23,$11,$F4,$BF,$26,$B7,$EE,$6D,$1A,$70,$97,$67,$EB,$5F,$F1,$F3
	.dc.b	$5C,$F6,$FB,$5C,$B5,$76,$FE,$CD,$FD,$8F,$81,$FD,$DE,$31,$1F,$B8
	.dc.b	$EE,$FD,$A3,$FC,$17,$6F,$8C,$FF,$93,$5A,$9A,$6B,$3F,$D1,$74,$FF
	.dc.b	$07,$FC,$97,$F3,$A4,$AF,$E2,$93,$A1,$1F,$C4,$FD,$A7,$57,$E7,$23
	.dc.b	$0D,$AF,$D1,$57,$06,$CF,$E0,$63,$EF,$93,$93,$07,$F0,$A9,$FE,$1B
	.dc.b	$C7,$F7,$A3,$07,$F1,$E9,$FE,$3B,$B7,$C9,$DF,$E7,$1F,$3F,$6F,$8F
	.dc.b	$87,$F9,$BD,$AF,$CE,$DB,$E4,$BF,$2F,$7E,$EC,$C7,$BB,$F3,$72,$FF
	.dc.b	$13,$60,$E4,$F0,$13,$3F,$55,$6C,$E4,$5B,$E6,$FF,$3F,$AE,$53,$C2
	.dc.b	$31,$5C,$EB,$71,$F6,$EF,$E1,$3C,$CF,$C2,$C7,$F3,$77,$D4,$D7,$F8
	.dc.b	$5F,$DE,$7F,$91,$2F,$F6,$ED,$E3,$34,$FF,$24,$F0,$7D,$CE,$4F,$ED
	.dc.b	$0D,$AF,$EA,$77,$FC,$6E,$73,$C3,$FA,$99,$CC,$FF,$51,$FB,$2F,$E4
	.dc.b	$8F,$ED,$5B,$97,$F5,$B9,$46,$27,$F1,$ED,$7D,$67,$B8,$86,$0F,$DC
	.dc.b	$E3,$E9,$7F,$81,$65,$EC,$6C,$7F,$97,$78,$3F,$02,$4E,$0F,$C0,$CF
	.dc.b	$E1,$5B,$87,$F1,$CF,$A9,$F8,$C4,$9E,$AF,$C6,$CF,$F8,$14,$BF,$5C
	.dc.b	$FD,$0F,$3F,$F7,$16,$C6,$6F,$FA,$3C,$E5,$FE,$6F,$F2,$7F,$EC,$3D
	.dc.b	$3F,$9F,$36,$32,$7F,$2E,$2C,$7D,$D7,$F3,$5C,$9F,$C1,$73,$F5,$C6
	.dc.b	$37,$FE,$91,$D9,$F7,$C9,$1F,$19,$FC,$27,$C1,$BE,$BF,$F9,$AE,$9F
	.dc.b	$C0,$23,$0F,$F8,$64,$C4,$6A,$BF,$DB,$50,$9D,$DF,$F8,$46,$7E,$B1
	.dc.b	$F3,$71,$F3,$12,$3F,$AE,$7D,$6F,$FA,$7A,$FF,$AC,$B6,$BF,$9A,$B5
	.dc.b	$FE,$CA,$D8,$F0,$61,$E3,$9F,$F1,$06,$1E,$29,$F3,$9E,$98,$F3,$C6
	.dc.b	$5F,$29,$F1,$CF,$C6,$9E,$7B,$FC,$B7,$EB,$BE,$B7,$D9,$1F,$E2,$D8
	.dc.b	$C5,$7F,$26,$6C,$7C,$46,$AF,$8E,$FF,$26,$FC,$27,$AD,$B1,$E5,$35
	.dc.b	$E7,$27,$4D,$8D,$4F,$F4,$EF,$67,$E0,$37,$0C,$F6,$A7,$2D,$8E,$D5
	.dc.b	$EF,$3E,$0F,$3F,$E8,$F7,$FF,$63,$E1,$E4,$D3,$DE,$DB,$FF,$2B,$36
	.dc.b	$C7,$A3,$3F,$AC,$F5,$3E,$51,$F3,$0C,$13,$0F,$05,$FE,$6F,$D2,$F4
	.dc.b	$7A,$7D,$3E,$07,$FC,$C5,$69,$CA,$8C,$C6,$1B,$AA,$B9,$5F,$FB,$BA
	.dc.b	$F8,$36,$B7,$3A,$9F,$5D,$F3,$AF,$E8,$91,$FE,$7D,$21,$FD,$D0,$A6
	.dc.b	$8C,$10,$F8,$B9,$7D,$E7,$E5,$ED,$F6,$DB,$9F,$F6,$E7,$71,$9D,$39
	.dc.b	$47,$EA,$4E,$E5,$3E,$9E,$DF,$75,$FA,$ED,$8C,$E9,$FB,$36,$9F,$B8
	.dc.b	$9E,$F7,$E1,$7E,$D1,$92,$9E,$63,$10,$DF,$F4,$6F,$E8,$4E,$46,$4A
	.dc.b	$EA,$3E,$31,$0F,$51,$63,$F7,$67,$6E,$0F,$E2,$31,$D9,$B1,$FB,$C8
	.dc.b	$F0,$F2,$B8,$FC,$2D,$8F,$3B,$5D,$C7,$BD,$FE,$52,$7C,$C2,$9F,$42
	.dc.b	$7C,$6F,$87,$8F,$E5,$3F,$C2,$9A,$FF,$10,$6C,$7F,$99,$39,$3F,$D7
	.dc.b	$13,$FE,$4E,$6B,$FE,$7E,$6C,$6E,$33,$C6,$E5,$72,$E1,$B3,$E7,$9D
	.dc.b	$4F,$6A,$FF,$39,$F3,$7E,$A6,$BF,$A4,$F3,$7F,$4C,$7F,$50,$60,$FE
	.dc.b	$A7,$2F,$D7,$BF,$AB,$92,$32,$FF,$A4,$7A,$FC,$FB,$1A,$49,$F3,$63
	.dc.b	$1F,$36,$7E,$87,$FD,$2B,$D9,$C8,$8F,$A6,$FF,$BC,$3C,$2B,$93,$F4
	.dc.b	$FB,$7F,$C1,$6F,$F8,$25,$AF,$F8,$23,$CB,$F9,$71,$88,$FF,$0B,$67
	.dc.b	$FC,$0D,$FF,$5B,$F7,$FF,$B9,$7B,$3F,$CA,$ED,$8C,$CE,$DF,$F0,$69
	.dc.b	$E5,$CB,$CE,$FF,$AC,$DF,$B6,$FE,$20,$FC,$E1,$4F,$CE,$9F,$E3,$93
	.dc.b	$F0,$3E,$37,$FB,$A5,$FF,$2F,$10,$FF,$51,$1F,$E6,$52,$18,$3F,$C9
	.dc.b	$86,$CF,$F5,$19,$D3,$C2,$F1,$F5,$7F,$93,$23,$89,$FE,$4C,$BF,$C8
	.dc.b	$C6,$63,$99,$0E,$67,$53,$D2,$7A,$C9,$C7,$F9,$07,$FC,$E4,$43,$FA
	.dc.b	$19,$FE,$9C,$9E,$FE,$27,$F6,$88,$EA,$77,$3E,$21,$DC,$FF,$54,$F8
	.dc.b	$9B,$7F,$D5,$2D,$73,$3E,$A8,$D7,$89,$FE,$92,$DD,$2F,$D4,$BF,$D2
	.dc.b	$71,$86,$C6,$6F,$FD,$2B,$1F,$EF,$C2,$3F,$EA,$0F,$39,$E4,$71,$8C
	.dc.b	$35,$F5,$90,$F5,$9E,$D3,$E3,$F2,$FE,$85,$FC,$2E,$9F,$25,$F8,$E7
	.dc.b	$FB,$13,$0F,$8E,$47,$E6,$1F,$ED,$6F,$17,$FB,$7F,$C3,$FF,$9A,$79
	.dc.b	$FF,$DA,$52,$8C,$4E,$A3,$FD,$73,$3A,$9F,$F7,$1C,$6C,$FF,$71,$94
	.dc.b	$FF,$71,$F2,$71,$69,$93,$DA,$7B,$CF,$F6,$67,$17,$E8,$B8,$FF,$39
	.dc.b	$1F,$97,$3F,$20,$7E,$48,$FB,$78,$FE,$FF,$CB,$FF,$04,$FC,$1F,$AD
	.dc.b	$B3,$E8,$6B,$FE,$F6,$6B,$FE,$10,$F3,$9D,$F8,$84,$61,$57,$F3,$28
	.dc.b	$C3,$67,$9D,$B4,$79,$F8,$6E,$1C,$8F,$97,$19,$3C,$7C,$BF,$91,$A7
	.dc.b	$EB,$FF,$8E,$25,$7F,$B8,$9F,$F5,$07,$27,$E3,$27,$73,$F8,$C6,$BF
	.dc.b	$E6,$C8,$D4,$F0,$E4,$FC,$EB,$FE,$A9,$F4,$DE,$9F,$F8,$CB,$87,$0D
	.dc.b	$C3,$FD,$F3,$3A,$7E,$AE,$28,$C5,$73,$F5,$78,$57,$F9,$F2,$51,$88
	.dc.b	$FD,$6C,$61,$E6,$D9,$FD,$D7,$17,$E0,$A7,$F0,$CD,$7F,$85,$F3,$7E
	.dc.b	$85,$FD,$1F,$3F,$E9,$5F,$E1,$07,$E8,$0A,$7E,$9B,$B1,$D9,$FC,$0F
	.dc.b	$D7,$79,$DF,$E5,$2F,$03,$FB,$7C,$62,$FF,$9D,$1F,$E5,$67,$11,$A9
	.dc.b	$E1,$3F,$B2,$C7,$FB,$5C,$FF,$78,$6C,$67,$1F,$EE,$73,$96,$C7,$EE
	.dc.b	$BC,$DF,$2E,$FE,$08,$46,$2B,$FC,$06,$D8,$E1,$97,$F5,$17,$F9,$4F
	.dc.b	$23,$D8,$C4,$7F,$9B,$A3,$18,$75,$6E,$1F,$E0,$E6,$CF,$F0,$CF,$67
	.dc.b	$F8,$CB,$FE,$10,$6B,$FE,$14,$6B,$BE,$5F,$EB,$71,$88,$FE,$76,$D7
	.dc.b	$FC,$FE,$D4,$D2,$47,$22,$47,$85,$AF,$85,$AF,$84,$D1,$0D,$14,$D1
	.dc.b	$82,$18,$21,$E8,$7A,$B4,$E4,$53,$89,$4E,$27,$B9,$E2,$D3,$89,$4E
	.dc.b	$25,$38,$9F,$11,$E2,$D3,$89,$4E,$25,$38,$C7,$CA,$76,$D7,$6D,$77
	.dc.b	$1D,$A3,$11,$EB,$21,$E5,$29,$DC,$AF,$07,$44,$34,$57,$4E,$08,$E0
	.dc.b	$91,$E3,$24,$7A,$9A,$FA,$9A,$FA,$8D,$10,$D1,$4D,$18,$21,$82,$1F
	.dc.b	$EB,$F3,$FB,$6F,$A3,$FE,$C8,$6F,$A1,$FB,$57,$FD,$98,$DF,$DF,$7F
	.dc.b	$DD,$11,$FE,$D0,$3F,$DB,$AE,$38,$3F,$CF,$6D,$8F,$EF,$2D,$7F,$C4
	.dc.b	$2D,$78,$CF,$FB,$9A,$78,$B6,$38,$B5,$E2,$D7,$89,$E3,$13,$D5,$B1
	.dc.b	$B6,$9B,$8D,$12,$34,$D8,$D4,$60,$8E,$09,$3E,$86,$9E,$82,$9E,$82
	.dc.b	$9E,$87,$81,$1E,$0D,$78,$35,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$F8
	.dc.b	$6F,$56,$9C,$8A,$71,$29,$C6,$7E,$34,$71,$6C,$71,$6B,$C5,$AF,$19
	.dc.b	$F9,$73,$C5,$B1,$C5,$AF,$16,$BC,$4F,$96,$79,$09,$3D,$9B,$1C,$1A
	.dc.b	$70,$8D,$12,$34,$D8,$D4,$60,$8E,$0F,$0F,$53,$EA,$BF,$CD,$BE,$DF
	.dc.b	$FB,$72,$31,$B2,$31,$3A,$AF,$F2,$ED,$6A,$7F,$AB,$6F,$99,$FE,$BE
	.dc.b	$D7,$FA,$F3,$56,$DF,$D8,$27,$FD,$D7,$FB,$AF,$F8,$03,$1F,$05,$7F
	.dc.b	$9A,$A3,$13,$AD,$39,$46,$27,$FB,$D7,$97,$C1,$FF,$4B,$F2,$FF,$BA
	.dc.b	$AF,$FB,$E4,$E6,$42,$74,$FF,$9E,$39,$BF,$C1,$D5,$FE,$0B,$8E,$02
	.dc.b	$32,$EF,$0E,$73,$FE,$80,$87,$CE,$BE,$37,$31,$FE,$84,$8C,$3D,$1B
	.dc.b	$3C,$5B,$1F,$EB,$0F,$5D,$EA,$7F,$11,$6F,$17,$F0,$B6,$CF,$F1,$D6
	.dc.b	$E3,$E9,$24,$79,$CF,$E2,$19,$7F,$8A,$E3,$FD,$2F,$BB,$D9,$7C,$AF
	.dc.b	$D9,$8C,$36,$33,$C3,$F2,$E7,$2D,$87,$E1,$3F,$EB,$8F,$D8,$BA,$C6
	.dc.b	$25,$18,$9D,$65,$FE,$26,$AD,$4F,$0F,$C1,$33,$3F,$E2,$79,$46,$23
	.dc.b	$F1,$71,$8B,$EB,$1C,$E7,$50,$F6,$DE,$07,$E0,$BE,$07,$FD,$6F,$39
	.dc.b	$8C,$5F,$FB,$82,$30,$A7,$75,$CA,$30,$6D,$7D,$57,$28,$FC,$7D,$78
	.dc.b	$63,$11,$FE,$DC,$9D,$BC,$8D,$B8,$AF,$2F,$EB,$DF,$E8,$4B,$7E,$0B
	.dc.b	$8F,$2B,$FD,$CA,$D8,$CE,$1E,$9A,$E1,$3F,$A2,$24,$7E,$A1,$AF,$83
	.dc.b	$E6,$BF,$4B,$1E,$E6,$C7,$B5,$AE,$CE,$E7,$D5,$FB,$5F,$C3,$12,$FF
	.dc.b	$08,$FE,$27,$E3,$64,$7E,$1D,$BD,$5F,$33,$E3,$FF,$6B,$E7,$5C,$3F
	.dc.b	$D2,$87,$C4,$8C,$BF,$3D,$AF,$C8,$F3,$6B,$F3,$5B,$87,$E0,$5B,$B9
	.dc.b	$DC,$BE,$91,$EC,$FF,$17,$6D,$FF,$02,$6E,$5F,$BF,$6D,$7F,$12,$8F
	.dc.b	$E1,$E4,$70,$7F,$10,$1E,$5F,$F2,$2F,$8A,$FE,$0B,$1F,$E9,$FE,$A4
	.dc.b	$62,$37,$18,$7F,$8D,$4F,$17,$F6,$1E,$43,$17,$DE,$FE,$C2,$30,$D8
	.dc.b	$CC,$FE,$BE,$72,$FE,$C7,$EF,$AB,$FA,$F6,$BF,$AE,$6E,$1F,$C1,$DC
	.dc.b	$D7,$EB,$7B,$FD,$0D,$AF,$E6,$0D,$CF,$F5,$33,$ED,$BF,$E2,$1C,$5E
	.dc.b	$9F,$B5,$FE,$0E,$DC,$3F,$82,$35,$FE,$04,$D7,$F9,$DE,$DF,$5A,$7E
	.dc.b	$A6,$F9,$9F,$B8,$D9,$D4,$7F,$48,$9F,$E8,$A4,$70,$7D,$7D,$BE,$74
	.dc.b	$7D,$98,$FA,$7B,$FF,$4A,$8C,$4E,$A7,$86,$3F,$5E,$7F,$0D,$B7,$D5
	.dc.b	$FB,$5B,$B1,$A4,$BE,$38,$FF,$56,$8E,$B8,$7E,$05,$B1,$99,$D9,$CC
	.dc.b	$F3,$65,$FD,$EB,$ED,$7B,$92,$33,$8F,$E1,$E3,$9C,$1A,$47,$AB,$FD
	.dc.b	$AB,$69,$1F,$DC,$1A,$FF,$18,$F9,$FF,$F2,$F6,$7F,$CB,$DB,$C8,$69
	.dc.b	$FE,$63,$D3,$FA,$87,$C2,$7F,$4D,$25,$7F,$4B,$7F,$C1,$4B,$EE,$9E
	.dc.b	$AF,$66,$E3,$FE,$7D,$9D,$4E,$4F,$F3,$57,$9B,$FD,$03,$CE,$FF,$38
	.dc.b	$37,$4F,$4D,$FF,$81,$23,$0F,$F3,$BF,$99,$F4,$91,$ED,$AA,$3A,$F8
	.dc.b	$8C,$3A,$B7,$0E,$B7,$FE,$0F,$8C,$3F,$D1,$FE,$6B,$E2,$91,$EA,$FC
	.dc.b	$8F,$5B,$FC,$2F,$F9,$EF,$F5,$06,$CF,$F4,$F6,$BF,$D2,$1A,$FF,$4D
	.dc.b	$FB,$9F,$F0,$FF,$C0,$7F,$58,$6C,$7F,$57,$6B,$FD,$41,$AF,$C3,$C3
	.dc.b	$CE,$FF,$8B,$1A,$E7,$2F,$79,$F1,$3E,$A3,$EF,$92,$30,$7D,$5F,$A6
	.dc.b	$7D,$FF,$F1,$7E,$8F,$E3,$13,$FD,$9A,$73,$C7,$FE,$52,$8C,$56,$E3
	.dc.b	$FB,$A4,$EA,$B8,$D7,$8F,$7F,$9F,$C5,$F6,$A3,$D1,$97,$E1,$7F,$55
	.dc.b	$7C,$9F,$FA,$E7,$2F,$EF,$D1,$8A,$FF,$29,$1F,$DE,$1C,$61,$DA,$51
	.dc.b	$87,$C7,$9F,$F3,$1F,$B4,$D6,$9E,$A6,$C6,$67,$62,$3D,$38,$7E,$17
	.dc.b	$2F,$97,$8F,$AB,$6F,$9F,$7E,$0F,$83,$FF,$2C,$E1,$E3,$CD,$18,$9F
	.dc.b	$5B,$71,$F5,$35,$7A,$EF,$AB,$7C,$BF,$C3,$9F,$F3,$2C,$E6,$31,$1C
	.dc.b	$27,$6F,$C6,$31,$1A,$97,$C5,$B9,$7F,$AA,$6B,$F5,$A7,$31,$8C,$BF
	.dc.b	$CB,$D5,$B7,$A4,$E9,$ED,$87,$C5,$BF,$89,$B7,$C6,$8F,$F3,$36,$5E
	.dc.b	$49,$EC,$DA,$CE,$5E,$58,$C5,$7E,$12,$B9,$91,$FE,$38,$43,$47,$F4
	.dc.b	$1F,$83,$7B,$9F,$DE,$A1,$F2,$AA,$FE,$FD,$5A,$F2,$3F,$D1,$F3,$9A
	.dc.b	$FF,$48,$CE,$5E,$35,$C2,$FA,$56,$E5,$5B,$8C,$36,$B8,$5F,$66,$D7
	.dc.b	$43,$F6,$C7,$FA,$88,$87,$FA,$A0,$A7,$E9,$8A,$70,$3E,$49,$FE,$5A
	.dc.b	$3F,$0B,$87,$E0,$9F,$9F,$8F,$F9,$9D,$CB,$5E,$47,$53,$C0,$43,$F1
	.dc.b	$FA,$FE,$05,$FC,$33,$73,$FC,$2E,$DF,$C1,$9A,$78,$E7,$F2,$25,$DB
	.dc.b	$F3,$0D,$C3,$CD,$3E,$A8,$FD,$46,$FF,$E7,$48,$FD,$3C,$7F,$8A,$5A
	.dc.b	$69,$C1,$E9,$21,$EC,$22,$E4,$FE,$32,$FF,$00,$38,$C7,$F8,$D6,$33
	.dc.b	$1F,$C0,$8F,$DD,$72,$FC,$77,$F8,$6B,$78,$BF,$8A,$56,$8F,$80,$7F
	.dc.b	$07,$3E,$B9,$4F,$98,$77,$E5,$FF,$22,$71,$F8,$C9,$7F,$C4,$9F,$E3
	.dc.b	$9D,$CE,$9F,$F6,$D5,$FF,$A5,$5B,$5F,$EA,$96,$CE,$E7,$7D,$1F,$DB
	.dc.b	$5F,$1B,$5F,$1D,$7F,$AC,$7A,$BF,$47,$1F,$D4,$4F,$D4,$97,$9B,$FA
	.dc.b	$B3,$83,$FA,$E0,$FD,$8F,$FB,$13,$4F,$EC,$45,$36,$FF,$AD,$BB,$BC
	.dc.b	$E4,$8F,$F7,$B5,$FF,$B0,$E3,$DF,$1F,$24,$FB,$A5,$3E,$EE,$3F,$DF
	.dc.b	$DE,$2D,$7F,$D8,$7D,$FF,$10,$E7,$97,$C1,$8E,$6D,$7F,$0F,$E0,$FF
	.dc.b	$86,$1F,$F6,$8B,$5F,$F6,$83,$5F,$F6,$4B,$5F,$F6,$FF,$85,$FE,$7F
	.dc.b	$EA,$FF,$76,$B6,$3F,$DD,$4D,$7F,$DD,$0D,$7F,$DA,$1E,$27,$DA,$23
	.dc.b	$EC,$7D,$3D,$FF,$E1,$B8,$FB,$6D,$7F,$1B,$5F,$E9,$1F,$B8,$FF,$24
	.dc.b	$92,$3F,$CA,$6F,$F9,$78,$8F,$F9,$43,$54,$62,$3F,$D4,$1E,$41,$7D
	.dc.b	$5F,$F9,$50,$8F,$F9,$79,$FF,$34,$91,$E8,$D7,$98,$FB,$7F,$F2,$E9
	.dc.b	$1F,$F3,$3B,$FE,$7B,$7F,$B8,$37,$5F,$F2,$33,$72,$FF,$1E,$37,$97
	.dc.b	$A5,$7F,$9B,$BC,$DF,$F5,$84,$FF,$A4,$33,$CC,$FF,$AF,$FC,$BF,$F6
	.dc.b	$4C,$E4,$FF,$4E,$CE,$A5,$E5,$7F,$A8,$A3,$0D,$9D,$9F,$EA,$89,$FE
	.dc.b	$A0,$7C,$79,$FF,$57,$6D,$92,$47,$FA,$EE,$1F,$AC,$7F,$A6,$6F,$FD
	.dc.b	$71,$18,$9D,$4E,$7D,$E7,$A5,$FE,$CB,$8F,$F6,$41,$27,$FD,$91,$5C
	.dc.b	$5F,$B1,$6B,$79,$5C,$67,$FD,$2B,$8F,$FA,$38,$8E,$0F,$F5,$66,$FF
	.dc.b	$E7,$58,$F9,$C4,$8C,$EF,$FE,$6D,$8F,$9C,$48,$FA,$58,$FC,$47,$FD
	.dc.b	$87,$B7,$E7,$E3,$FE,$00,$43,$D8,$FF,$5D,$46,$1B,$D0,$F5,$7F,$25
	.dc.b	$18,$6D,$7F,$BE,$69,$4F,$5B,$33,$F4,$E3,$0F,$73,$83,$8B,$EF,$3C
	.dc.b	$27,$F0,$3A,$BD,$4F,$DB,$6F,$F8,$3D,$3F,$0E,$4A,$FD,$31,$C3,$D8
	.dc.b	$DD,$9A,$4C,$FE,$21,$E8,$CB,$F5,$DF,$83,$FC,$06,$DF,$99,$AF,$AD
	.dc.b	$AF,$F9,$FE,$3F,$44,$47,$EA,$F3,$23,$E6,$FB,$6F,$71,$86,$6F,$E5
	.dc.b	$4E,$63,$11,$F9,$B8,$C5,$72,$AE,$2F,$D6,$F7,$BF,$83,$7C,$7D,$B3
	.dc.b	$48,$FD,$0E,$4D,$B4,$D9,$63,$F0,$31,$A8,$FC,$DE,$AF,$80,$F4,$FE
	.dc.b	$C9,$FA,$56,$CF,$D2,$E1,$FD,$04,$62,$38,$B5,$FB,$BF,$07,$67,$F4
	.dc.b	$CD,$B3,$4C,$7C,$27,$DC,$26,$FF,$9D,$AF,$D7,$47,$F3,$BB,$7C,$2F
	.dc.b	$F9,$1D,$FC,$29,$8F,$5B,$FA,$EC,$66,$3F,$12,$D7,$FD,$07,$F0,$DF
	.dc.b	$E5,$C7,$F9,$27,$A8,$8C,$47,$48,$C7,$B3,$FE,$06,$6C,$7F,$82,$3E
	.dc.b	$1D,$F1,$1F,$CE,$FA,$BF,$97,$46,$27,$F6,$C1,$C4,$6A,$78,$4F,$AB
	.dc.b	$1F,$F3,$44,$FF,$37,$6C,$67,$2D,$BE,$C6,$BF,$ED,$59,$FE,$14,$FF
	.dc.b	$B0,$9A,$FF,$4F,$F8,$9F,$F3,$C6,$DE,$46,$CF,$F4,$56,$CF,$CC,$6B
	.dc.b	$F8,$1F,$8A,$FA,$5B,$9A,$4B,$F7,$FC,$5F,$F2,$46,$F8,$9F,$66,$31
	.dc.b	$8F,$F5,$36,$E1,$F6,$BE,$33,$E8,$B7,$83,$FC,$2F,$18,$8F,$E0,$6D
	.dc.b	$7F,$B5,$35,$34,$91,$C8,$91,$FE,$40,$8C,$B6,$32,$FF,$67,$8C,$B6
	.dc.b	$32,$F9,$4E,$04,$34,$53,$46,$08,$60,$87,$B1,$EA,$D3,$91,$4E,$25
	.dc.b	$38,$9F,$09,$E2,$D3,$89,$4E,$25,$38,$9F,$21,$E2,$D3,$89,$4E,$25
	.dc.b	$38,$C7,$C6,$6C,$72,$8C,$47,$9C,$87,$84,$A7,$32,$BC,$1D,$10,$D1
	.dc.b	$5D,$38,$23,$82,$47,$8C,$91,$EC,$8C,$B6,$32,$FB,$A3,$2D,$8C,$BE
	.dc.b	$E3,$81,$0D,$1F,$E8,$CF,$7F,$FD,$C0,$FF,$BA,$E3,$F8,$1B,$FE,$DE
	.dc.b	$6C,$6A,$3F,$3B,$F4,$9F,$E4,$5D,$BF,$24,$D8,$FD,$33,$5D,$9F,$C7
	.dc.b	$ED,$19,$8F,$D2,$B5,$FD,$2B,$63,$71,$C6,$30,$47,$04,$9F,$1B,$4F
	.dc.b	$19,$4F,$1B,$A2,$3A,$6B,$A3,$04,$30,$43,$D2,$75,$2B,$C9,$A7,$12
	.dc.b	$9C,$4F,$79,$C4,$AF,$16,$9C,$4A,$71,$9F,$8B,$1B,$6C,$6D,$AE,$E7
	.dc.b	$E4,$4E,$DB,$1B,$6B,$B3,$C9,$3C,$9F,$49,$0F,$19,$5D,$38,$23,$82
	.dc.b	$47,$68,$E0,$D7,$6D,$8D,$C7,$8A,$30,$47,$04,$9F,$63,$4F,$61,$4F
	.dc.b	$63,$A2,$3A,$6B,$FE,$3C,$FA,$5F,$CB,$E8,$F8,$9F,$F4,$B4,$61,$B7
	.dc.b	$FE,$2C,$6C,$EE,$BF,$C0,$3F,$20,$F8,$7D,$46,$1A,$FF,$92,$A3,$53
	.dc.b	$C6,$31,$3A,$7F,$D6,$9E,$5B,$B9,$E2,$6E,$73,$18,$C7,$FD,$4C,$48
	.dc.b	$FF,$09,$FD,$C3,$C2,$FF,$21,$FC,$5F,$FA,$C6,$31,$9A,$31,$3A,$AE
	.dc.b	$2D,$7C,$B3,$B7,$D8,$D7,$33,$FE,$BD,$D3,$DA,$FF,$82,$77,$FF,$67
	.dc.b	$36,$78,$1F,$ED,$1E,$94,$EA,$3F,$C1,$C6,$AD,$EF,$BF,$11,$FF,$6C
	.dc.b	$7A,$7F,$E8,$C8,$C5,$23,$13,$A8,$FF,$45,$4E,$A7,$CD,$C3,$FE,$66
	.dc.b	$C2,$E1,$FE,$37,$CF,$EA,$C6,$0F,$F3,$6F,$16,$F1,$FF,$35,$B5,$FF
	.dc.b	$35,$35,$FF,$38,$F9,$3F,$8C,$9F,$37,$D9,$7F,$BB,$FC,$AB,$5F,$E7
	.dc.b	$73,$CB,$19,$69,$93,$FD,$A1,$A3,$F0,$9D,$46,$1A,$FF,$A0,$23,$4F
	.dc.b	$13,$11,$AE,$FF,$F6,$47,$E1,$5F,$E8,$6F,$C4,$7F,$DA,$4D,$AF,$F3
	.dc.b	$77,$92,$8C,$47,$F9,$D9,$AF,$7F,$86,$2A,$FA,$91,$86,$C6,$4F,$A7
	.dc.b	$19,$7F,$D7,$DF,$6A,$FC,$93,$FD,$33,$FA,$C7,$FB,$22,$31,$C6,$8C
	.dc.b	$4E,$AB,$8B,$5F,$F5,$BF,$77,$EA,$7B,$B8,$47,$FA,$A2,$BF,$CE,$EA
	.dc.b	$75,$18,$6C,$7E,$23,$93,$DB,$3F,$8F,$6C,$64,$EA,$7D,$67,$2F,$D8
	.dc.b	$F8,$CF,$C9,$34,$FC,$B1,$F9,$47,$2F,$D9,$6E,$3F,$A0,$FB,$8B,$C4
	.dc.b	$69,$2B,$99,$FE,$AA,$FA,$7F,$F6,$0F,$D5,$FF,$A5,$38,$DC,$69,$CD
	.dc.b	$7D,$E6,$BC,$71,$5F,$EB,$7E,$47,$EE,$B6,$72,$7E,$B2,$32,$D3,$27
	.dc.b	$FA,$4F,$F6,$07,$D2,$7F,$B5,$3B,$BF,$D4,$FE,$5A,$73,$8F,$AA,$D3
	.dc.b	$9B,$E1,$1F,$EF,$28,$C4,$EE,$3E,$D7,$8D,$FC,$9B,$0F,$B4,$D8,$FF
	.dc.b	$73,$C6,$39,$3B,$5B,$F0,$37,$C1,$7D,$28,$C3,$63,$37,$C1,$B7,$FA
	.dc.b	$8C,$5F,$B0,$52,$57,$EA,$BB,$3F,$DE,$71,$FA,$22,$47,$E8,$8E,$25
	.dc.b	$7F,$8E,$72,$FE,$B3,$EF,$EC,$9A,$47,$F6,$FF,$25,$F1,$9F,$F7,$5B
	.dc.b	$57,$2F,$E0,$A3,$0D,$C7,$E1,$F0,$FD,$2B,$FA,$AD,$BF,$A0,$D7,$EE
	.dc.b	$70,$AB,$C9,$1F,$EE,$2E,$AF,$D8,$5F,$CF,$24,$67,$8B,$FA,$DC,$7E
	.dc.b	$0B,$7F,$9A,$DE,$0F,$81,$5F,$7E,$72,$D8,$CB,$F8,$DF,$80,$F1,$7D
	.dc.b	$F7,$A1,$B1,$E8,$6B,$BC,$93,$DE,$29,$D4,$A7,$FB,$AE,$FF,$03,$1B
	.dc.b	$3B,$4B,$EF,$A9,$2B,$C1,$19,$7A,$47,$72,$38,$3E,$27,$2F,$F6,$26
	.dc.b	$F8,$BF,$97,$23,$FC,$D3,$EC,$3F,$A7,$70,$FE,$F1,$BD,$66,$9F,$D4
	.dc.b	$F1,$7E,$C5,$FC,$DA,$FF,$C1,$51,$87,$F9,$CF,$A1,$97,$EE,$4E,$E3
	.dc.b	$13,$B9,$E3,$F6,$0F,$AF,$FF,$3F,$4A,$31,$3E,$B3,$FA,$2C,$66,$77
	.dc.b	$3A,$CF,$DB,$5F,$CE,$38,$FF,$C3,$F1,$88,$FF,$0F,$EB,$CD,$F2,$FA
	.dc.b	$FE,$0C,$CD,$24,$6C,$FE,$0B,$37,$97,$FA,$25,$F5,$9C,$BF,$C9,$F6
	.dc.b	$7D,$5F,$CC,$26,$BE,$1E,$0F,$E5,$71,$F3,$73,$F7,$FE,$3B,$F1,$9B
	.dc.b	$8F,$CF,$BF,$A3,$BF,$F6,$6A,$D3,$67,$FB,$7F,$A1,$FA,$BE,$BF,$13
	.dc.b	$63,$36,$9C,$B6,$3F,$0E,$43,$E5,$6D,$EF,$E0,$F9,$71,$FD,$F3,$E0
	.dc.b	$3F,$0E,$4A,$F6,$6C,$8F,$C4,$7A,$0F,$33,$F7,$DF,$63,$FB,$FC,$3E
	.dc.b	$DC,$E7,$04,$E6,$7F,$C4,$35,$C2,$7E,$15,$6E,$B5,$5F,$40,$F4,$1F
	.dc.b	$7B,$8B,$F0,$1C,$5F,$D6,$27,$FC,$6D,$7F,$E1,$46,$D7,$F8,$42,$75
	.dc.b	$3F,$E4,$A3,$45,$8F,$3E,$1F,$C5,$FE,$6F,$CD,$2C,$7D,$51,$86,$C7
	.dc.b	$A2,$78,$3E,$6A,$F4,$1F,$3B,$1F,$CB,$47,$EB,$BE,$2B,$FC,$47,$87
	.dc.b	$6E,$34,$62,$7F,$09,$18,$B5,$6E,$B5,$1F,$E2,$CF,$83,$7E,$A6,$F2
	.dc.b	$7F,$DE,$98,$FE,$EE,$73,$18,$AE,$15,$B7,$E1,$09,$DB,$85,$F6,$4F
	.dc.b	$8C,$F9,$B3,$B8,$C6,$1B,$6E,$9F,$7E,$F8,$DF,$0B,$F8,$A7,$DB,$3E
	.dc.b	$47,$57,$EC,$1F,$46,$3E,$5B,$E4,$DC,$73,$97,$AE,$BD,$13,$97,$E8
	.dc.b	$1D,$39,$3F,$3A,$E0,$87,$84,$A7,$90,$7C,$03,$E0,$7F,$C1,$F0,$F9
	.dc.b	$CF,$E3,$13,$9F,$3B,$FC,$BB,$18,$AD,$B6,$33,$9F,$F2,$DA,$E1,$7C
	.dc.b	$E3,$17,$C1,$B5,$C2,$7F,$CD,$35,$C2,$FF,$94,$69,$FE,$59,$9F,$F0
	.dc.b	$C1,$0F,$E8,$E5,$3F,$7E,$53,$F9,$39,$FC,$EF,$6F,$BE,$FC,$ED,$3F
	.dc.b	$8A,$3F,$86,$6B,$F3,$9A,$FF,$28,$3A,$14,$F0,$9F,$91,$DF,$F0,$AF
	.dc.b	$E0,$DB,$B6,$77,$FE,$10,$D3,$FA,$1C,$F8,$4A,$78,$4B,$7E,$88,$FC
	.dc.b	$1C,$69,$FD,$72,$F1,$BF,$50,$D8,$FD,$6F,$0F,$FA,$52,$72,$7F,$10
	.dc.b	$CB,$E6,$57,$A8,$F9,$25,$3F,$8C,$D7,$37,$F8,$4B,$5C,$CF,$F0,$7E
	.dc.b	$4F,$F1,$13,$FC,$87,$9B,$F8,$9C,$DE,$9F,$E4,$B7,$EE,$9C,$BF,$C5
	.dc.b	$A7,$C9,$C7,$FE,$FE,$7E,$7E,$3E,$67,$CA,$78,$B1,$F0,$57,$F3,$18
	.dc.b	$C3,$FC,$5C,$C3,$FE,$40,$C3,$F9,$E4,$7F,$A8,$30,$FF,$41,$F4,$FF
	.dc.b	$32,$AD,$36,$33,$3F,$EA,$69,$FC,$01,$E2,$3F,$1A,$7C,$78,$FF,$46
	.dc.b	$F6,$7F,$90,$FD,$CF,$EA,$A4,$9F,$EB,$2F,$F6,$05,$FB,$2D,$23,$FC
	.dc.b	$93,$BF,$FD,$97,$1F,$DC,$FC,$1F,$26,$7F,$DD,$32,$F5,$36,$F5,$1E
	.dc.b	$7D,$7E,$44,$79,$8F,$41,$EA,$3F,$C1,$27,$FB,$0A,$3F,$C4,$BE,$17
	.dc.b	$BD,$B1,$9E,$DF,$F0,$6B,$D7,$83,$EB,$3F,$87,$3A,$1D,$67,$FC,$3C
	.dc.b	$F5,$F0,$FF,$C4,$6F,$F8,$A9,$AF,$F8,$15,$AF,$F8,$9A,$BF,$D3,$CF
	.dc.b	$F4,$63,$FD,$DB,$D5,$FE,$0D,$69,$93,$F8,$C9,$FD,$8F,$2F,$F2,$15
	.dc.b	$FC,$6F,$17,$E4,$91,$F5,$78,$52,$BF,$1F,$97,$B1,$FF,$30,$79,$3F
	.dc.b	$EF,$CF,$63,$FC,$CC,$48,$FF,$33,$3F,$E6,$22,$3F,$E6,$06,$BC,$87
	.dc.b	$DF,$7F,$9B,$C8,$FF,$9B,$DF,$F3,$59,$1F,$F3,$4B,$5E,$5F,$6D,$FE
	.dc.b	$77,$23,$FE,$77,$7F,$D7,$84,$7F,$CD,$6D,$7F,$D2,$F8,$A3,$FD,$3B
	.dc.b	$E7,$7F,$A6,$5B,$19,$9F,$F4,$FC,$E5,$FF,$56,$C3,$F0,$1F,$F2,$37
	.dc.b	$93,$FE,$A1,$6C,$66,$77,$1F,$EB,$8B,$91,$FE,$FF,$3B,$1F,$E6,$4F
	.dc.b	$61,$E9,$23,$27,$FA,$B6,$3F,$C6,$4B,$F6,$6F,$F3,$9C,$61,$BD,$DC
	.dc.b	$5B,$3B,$3F,$DA,$58,$D7,$F0,$DE,$AF,$FA,$E6,$7F,$DB,$44,$8F,$F5
	.dc.b	$16,$5F,$EA,$52,$38,$3F,$D2,$EF,$FB,$AC,$8E,$0F,$93,$5F,$0D,$FA
	.dc.b	$31,$F3,$88,$FF,$99,$70,$FF,$0B,$7D,$C7,$B3,$C6,$F8,$4D,$93,$4F
	.dc.b	$F3,$B7,$B3,$FE,$D1,$6E,$1A,$9F,$F4,$F7,$82,$E7,$5E,$E7,$31,$8A
	.dc.b	$E1,$1F,$4F,$DA,$FA,$59,$FC,$E6,$C7,$9D,$AE,$CF,$71,$F5,$F7,$CB
	.dc.b	$4C,$D3,$4F,$BF,$18,$6D,$FD,$CF,$6F,$F0,$5F,$4F,$68,$D2,$3F,$EC
	.dc.b	$87,$FD,$70,$FF,$A7,$9B,$DE,$E9,$75,$72,$AE,$07,$FA,$57,$A5,$EF
	.dc.b	$7F,$BE,$F0,$FF,$5F,$C6,$34,$E1,$5F,$EB,$A8,$C4,$EA,$B8,$4F,$DB
	.dc.b	$E6,$FC,$D7,$EC,$7F,$92,$24,$64,$3F,$EA,$FE,$76,$38,$9F,$CB,$CE
	.dc.b	$63,$11,$F9,$86,$BF,$A5,$F8,$1F,$D4,$7C,$B7,$E9,$1B,$1F,$8D,$6B
	.dc.b	$FA,$F6,$BF,$AD,$F8,$3B,$5F,$55,$BC,$46,$9A,$F5,$9C,$5F,$DA,$6B
	.dc.b	$FD,$14,$D8,$CC,$7F,$4D,$6D,$FD,$6F,$85,$79,$7F,$DC,$61,$18,$AF
	.dc.b	$EE,$0D,$8E,$13,$DF,$17,$97,$F8,$EB,$FC,$63,$63,$35,$F8,$D8,$FC
	.dc.b	$9A,$F8,$67,$AB,$FC,$3F,$E6,$FF,$13,$18,$94,$62,$75,$1F,$C5,$4F
	.dc.b	$E2,$D3,$B9,$F2,$E1,$C2,$74,$7B,$31,$FB,$B3,$A2,$46,$BE,$29,$F1
	.dc.b	$5A,$8C,$35,$F0,$46,$AF,$B4,$62,$74,$FF,$29,$EA,$FD,$07,$4F,$F6
	.dc.b	$48,$C7,$57,$E8,$BC,$AF,$D2,$B6,$3C,$F5,$C2,$7D,$64,$8F,$5F,$07
	.dc.b	$D5,$8C,$47,$C3,$F8,$BF,$BF,$1F,$B2,$6C,$7C,$26,$BB,$E3,$FC,$FC
	.dc.b	$FE,$E3,$E3,$3F,$17,$3F,$E2,$F6,$B9,$9F,$F1,$92,$8C,$BF,$39,$AF
	.dc.b	$D5,$6B,$FD,$2D,$AF,$18,$E7,$18,$23,$82,$47,$8D,$AF,$8C,$C1,$0C
	.dc.b	$10,$F2,$3C,$1A,$6C,$A6,$CF,$43,$B6,$9B,$29,$B3,$D8,$ED,$A6,$CA
	.dc.b	$6C,$F8,$06,$4A,$66,$3E,$0B,$96,$B9,$8E,$CF,$03,$D2,$43,$CA,$53
	.dc.b	$C2,$53,$81,$A2,$1A,$29,$A3,$04,$30,$48,$F2,$BD,$5A,$F2,$6B,$C5
	.dc.b	$AF,$18,$F5,$C6,$08,$E0,$91,$F0,$DA,$FC,$33,$04,$30,$43,$C8,$F0
	.dc.b	$7F,$81,$FB,$5F,$EE,$46,$FB,$7F,$B5,$7F,$DC,$2D,$7F,$DE,$AD,$FC
	.dc.b	$95,$EA,$FF,$89,$79,$7F,$AF,$C6,$2B,$F8,$C1,$FC,$55,$C4,$6A,$78
	.dc.b	$4F,$F1,$49,$CB,$63,$2F,$16,$BF,$E3,$38,$DB,$93,$B9,$98,$DA,$8D
	.dc.b	$B9,$36,$66,$37,$1A,$8C,$1A,$70,$48,$D9,$23,$4D,$9F,$74,$60,$8E
	.dc.b	$08,$FB,$CC,$95,$CB,$ED,$6B,$E9,$6B,$E6,$6C,$F9,$63,$99,$C4,$E8
	.dc.b	$43,$04,$3A,$19,$29,$93,$B1,$92,$99,$7C,$11,$E2,$9F,$85,$3B,$9C
	.dc.b	$B6,$32,$F1,$6B,$E0,$21,$D8,$B1,$C9,$E0,$7B,$8C,$C7,$27,$81,$23
	.dc.b	$83,$A3,$81,$98,$E1,$1A,$8C,$1A,$70,$48,$D9,$23,$4D,$9F,$AB,$A7
	.dc.b	$F8,$27,$E0,$7F,$B9,$92,$BF,$AC,$F2,$FF,$8B,$A3,$FB,$0F,$09,$A4
	.dc.b	$9F,$EC,$7C,$3F,$E4,$88,$FF,$01,$46,$CF,$EF,$BC,$DF,$E0,$78,$C9
	.dc.b	$23,$FB,$06,$DF,$E3,$CF,$8D,$FE,$C8,$D8,$CF,$2A,$72,$D8,$F0,$E3
	.dc.b	$5F,$F0,$87,$27,$F7,$CA,$F4,$92,$3F,$C1,$FF,$24,$F9,$2F,$F5,$54
	.dc.b	$FF,$99,$23,$13,$CC,$FF,$57,$38,$8D,$4F,$09,$FF,$0C,$FD,$AB,$D9
	.dc.b	$FF,$2A,$F0,$FF,$9B,$A3,$19,$FE,$F1,$B1,$99,$DD,$7A,$A3,$E7,$91
	.dc.b	$C1,$FE,$26,$FA,$9F,$F1,$84,$7F,$9E,$9B,$19,$8F,$E1,$B1,$88,$FF
	.dc.b	$63,$79,$CF,$D8,$BF,$CD,$38,$BF,$67,$FF,$18,$CE,$63,$FC,$DB,$18
	.dc.b	$9D,$B6,$33,$5F,$EA,$89,$7C,$AE,$23,$BC,$62,$7B,$46,$E7,$4A,$FC
	.dc.b	$91,$87,$CD,$F0,$BF,$E7,$EF,$BD,$F1,$36,$BC,$0D,$7B,$B5,$FF,$42
	.dc.b	$F8,$AF,$57,$18,$A3,$0A,$FF,$CA,$2E,$67,$73,$A9,$CF,$DD,$3E,$27
	.dc.b	$FD,$23,$F9,$16,$1B,$8F,$F9,$7F,$15,$6D,$56,$DB,$3F,$E6,$0F,$C1
	.dc.b	$9F,$83,$F2,$8F,$F3,$44,$62,$77,$18,$9D,$3F,$F0,$06,$75,$5E,$6F
	.dc.b	$AF,$CD,$F8,$9B,$81,$A7,$2B,$72,$BE,$1F,$F2,$71,$FE,$79,$8C,$5F
	.dc.b	$E6,$67,$32,$38,$B8,$5C,$9F,$91,$74,$D7,$3C,$BF,$96,$3F,$B4,$F1
	.dc.b	$7E,$57,$A7,$FD,$7B,$F9,$0F,$E4,$5B,$5F,$8F,$6B,$F8,$E6,$BF,$8B
	.dc.b	$E6,$FF,$6D,$64,$AF,$CD,$78,$1A,$C3,$F4,$4D,$AD,$CF,$EC,$EB,$F6
	.dc.b	$DD,$3E,$BD,$7F,$6F,$8F,$EF,$A3,$DD,$D9,$FE,$88,$D3,$F6,$D3,$FC
	.dc.b	$43,$2F,$F7,$4C,$62,$3E,$0B,$EF,$7F,$BE,$EB,$F0,$CF,$8D,$8F,$FB
	.dc.b	$A6,$1F,$8C,$7E,$53,$D6,$FF,$6F,$C6,$0F,$C7,$38,$8D,$4F,$D3,$78
	.dc.b	$75,$DA,$DB,$5F,$B1,$3F,$EF,$A8,$C4,$7F,$BB,$2B,$F7,$6F,$73,$F0
	.dc.b	$5C,$3F,$0C,$95,$F8,$33,$D7,$8F,$B4,$FE,$28,$59,$F9,$11,$F3,$09
	.dc.b	$1F,$3C,$F8,$D0,$CF,$E6,$3F,$C8,$0F,$99,$DF,$D9,$93,$D9,$A7,$F3
	.dc.b	$2B,$F2,$C7,$F3,$16,$C7,$F2,$53,$67,$FB,$57,$3F,$A5,$3F,$E3,$4D
	.dc.b	$BF,$3E,$F1,$C7,$DC,$D3,$C0,$4A,$E8,$D7,$CE,$FB,$31,$F5,$C6,$0F
	.dc.b	$B3,$F8,$2F,$F9,$22,$99,$A3,$0D,$9F,$F2,$4C,$EA,$73,$F1,$DF,$B1
	.dc.b	$F0,$FF,$0C,$DA,$FC,$2B,$5F,$BF,$68,$C4,$2D,$F5,$EF,$1B,$C8,$D9
	.dc.b	$F0,$B7,$3F,$E4,$ED,$8F,$AC,$7F,$BA,$37,$FF,$2C,$E9,$F0,$23,$15
	.dc.b	$F9,$BD,$3D,$A4,$8F,$6E,$7E,$F9,$F9,$1C,$5F,$AA,$9E,$A7,$E1,$5C
	.dc.b	$3E,$CF,$5B,$E3,$91,$E6,$FF,$0C,$E6,$7A,$1E,$FE,$AF,$EB,$8D,$AF
	.dc.b	$EB,$4D,$7F,$AC,$B5,$FD,$26,$1F,$C7,$FE,$13,$F6,$8D,$8F,$D9,$B5
	.dc.b	$F9,$ED,$7E,$09,$EA,$D3,$E2,$3E,$BD,$7E,$0F,$7F,$E0,$09,$18,$3F
	.dc.b	$03,$EE,$7F,$1C,$BF,$E0,$71,$F6,$63,$1C,$5F,$D4,$5B,$1F,$BE,$6B
	.dc.b	$E2,$3F,$24,$43,$07,$F0,$1D,$FF,$D5,$13,$FC,$39,$B1,$F8,$A6,$BB
	.dc.b	$15,$E4,$95,$78,$1B,$BF,$F8,$91,$F3,$36,$BC,$CD,$7C,$67,$C5,$AF
	.dc.b	$C1,$E3,$F8,$6F,$CB,$FC,$37,$FA,$86,$E2,$69,$E6,$F9,$07,$C9,$3E
	.dc.b	$F3,$0D,$CB,$D3,$87,$F5,$B5,$5C,$2B,$67,$3F,$93,$FC,$5D,$70,$AF
	.dc.b	$EC,$54,$9C,$E7,$F5,$A3,$1C,$3E,$D6,$DF,$1B,$F7,$F0,$7E,$10,$9B
	.dc.b	$7E,$7A,$3C,$79,$F9,$F1,$E8,$DB,$F4,$61,$E7,$AE,$8D,$AF,$C3,$C7
	.dc.b	$CB,$3F,$25,$7D,$5C,$1F,$83,$29,$E4,$75,$EA,$3D,$47,$AD,$FE,$01
	.dc.b	$F3,$7F,$DD,$73,$98,$C6,$1F,$2E,$31,$82,$B7,$7C,$A3,$15,$B8,$C4
	.dc.b	$EA,$F9,$D7,$FC,$00,$6B,$FF,$00,$1B,$A3,$17,$17,$DB,$AF,$E9,$27
	.dc.b	$FA,$68,$87,$FA,$20,$A6,$8C,$1F,$33,$5F,$C3,$3F,$43,$83,$FC,$7C
	.dc.b	$FE,$02,$33,$B6,$E7,$33,$FC,$8C,$EA,$53,$A9,$F8,$8E,$0F,$C5,$BF
	.dc.b	$87,$6C,$FD,$C6,$C6,$F8,$3F,$C0,$73,$DE,$3F,$BD,$9E,$3A,$F0,$F0
	.dc.b	$7E,$0E,$3D,$06,$8F,$EF,$DC,$3E,$6A,$FD,$54,$65,$AF,$EE,$F8,$BF
	.dc.b	$58,$F1,$3F,$8C,$1E,$C2,$9E,$C3,$F8,$A6,$9E,$67,$F7,$B3,$D7,$8F
	.dc.b	$FC,$6B,$1F,$C1,$DB,$1F,$C3,$E3,$C5,$5F,$D9,$1A,$F9,$CC,$3F,$9C
	.dc.b	$DB,$F0,$1B,$7D,$F8,$F9,$C7,$DB,$21,$E4,$AF,$16,$1E,$58,$C1,$E3
	.dc.b	$71,$1F,$DD,$39,$3A,$4F,$F8,$95,$B1,$98,$F0,$35,$F8,$97,$E4,$6D
	.dc.b	$7E,$51,$FE,$92,$3D,$AA,$47,$FA,$59,$93,$FA,$AD,$3E,$4F,$FA,$CB
	.dc.b	$63,$31,$83,$FC,$91,$D5,$E6,$CC,$D2,$4F,$B3,$1F,$33,$F6,$4F,$B6
	.dc.b	$7C,$8B,$FF,$08,$3E,$EA,$FF,$0A,$3F,$DE,$CC,$75,$FC,$67,$89,$F4
	.dc.b	$88,$62,$7F,$C1,$FD,$BF,$2B,$B7,$FC,$1D,$3F,$E0,$C8,$C1,$E9,$EE
	.dc.b	$FF,$47,$73,$FF,$87,$A7,$F8,$41,$FE,$1C,$8C,$BF,$E2,$03,$27,$F9
	.dc.b	$6F,$0F,$84,$47,$CE,$7C,$1D,$BE,$A3,$F0,$73,$F8,$6F,$E2,$0F,$F4
	.dc.b	$B9,$F8,$2F,$07,$FA,$FB,$7A,$FE,$04,$62,$3F,$C9,$04,$3F,$CA,$9A
	.dc.b	$7E,$2B,$C1,$FF,$55,$C7,$F9,$64,$87,$F9,$54,$A7,$F9,$C2,$38,$78
	.dc.b	$5F,$CE,$A3,$FD,$61,$3C,$9E,$A7,$03,$FD,$1E,$79,$48,$7E,$4B,$93
	.dc.b	$E0,$11,$FF,$50,$10,$FF,$3B,$94,$FF,$33,$61,$EE,$23,$D0,$F0,$90
	.dc.b	$F4,$1F,$04,$FF,$28,$F8,$7F,$EC,$7F,$C6,$3F,$D2,$B0,$69,$1F,$F4
	.dc.b	$EE,$3B,$69,$B2,$C7,$FA,$81,$C1,$E6,$3C,$85,$3E,$21,$F9,$42,$67
	.dc.b	$F7,$9F,$CA,$E2,$F1,$BE,$74,$7F,$B9,$1B,$19,$BF,$B5,$1F,$4F,$C8
	.dc.b	$FF,$65,$C3,$F1,$FF,$F6,$7B,$6F,$FC,$19,$3A,$9C,$9F,$ED,$D9,$D4
	.dc.b	$FF,$B6,$DA,$FF,$B6,$FC,$AF,$F3,$B4,$EC,$F5,$EF,$F8,$F9,$FC,$41
	.dc.b	$0F,$A2,$7D,$33,$F0,$FB,$FA,$8F,$C5,$5B,$CB,$78,$FF,$E8,$3F,$5B
	.dc.b	$FB,$44,$E6,$31,$7F,$E2,$58,$C0,$9D,$D7,$28,$C4,$FF,$71,$F3,$9D
	.dc.b	$DC,$A3,$FC,$5D,$18,$9D,$C6,$27,$48,$4E,$EF,$BF,$9F,$AB,$F5,$37
	.dc.b	$03,$4D,$7A,$2F,$4B,$FD,$41,$E3,$FB,$A3,$17,$FE,$39,$CB,$DD,$18
	.dc.b	$9D,$57,$09,$FB,$7E,$6F,$F8,$6E,$F8,$47,$9F,$89,$39,$BF,$F0,$94
	.dc.b	$63,$CA,$E5,$5B,$AD,$3D,$3A,$3F,$D6,$9F,$B1,$F0,$94,$62,$75,$8A
	.dc.b	$B5,$39,$9E,$93,$56,$1E,$48,$C3,$5E,$51,$EC,$78,$3F,$C8,$0F,$C7
	.dc.b	$F5,$7F,$00,$7F,$85,$10,$FE,$0E,$43,$FC,$D5,$D1,$FA,$77,$F8,$51
	.dc.b	$F6,$F9,$DE,$AF,$DD,$F7,$BF,$BC,$46,$25,$18,$9D,$47,$F8,$EE,$75
	.dc.b	$3F,$DF,$FA,$FF,$C0,$2D,$EE,$FF,$20,$B7,$BB,$9D,$FF,$90,$EB,$6F
	.dc.b	$F9,$17,$DD,$7B,$3F,$EE,$3B,$FF,$05,$46,$35,$E3,$5F,$E8,$58,$C4
	.dc.b	$EA,$B8,$4F,$F9,$3F,$A3,$FA,$2F,$87,$FE,$13,$6C,$67,$7F,$EA,$EF
	.dc.b	$FA,$3E,$32,$FF,$A4,$72,$FF,$0D,$D7,$A4,$91,$E9,$D3,$FC,$BB,$1F
	.dc.b	$E1,$FF,$6B,$FD,$F1,$A7,$F6,$0A,$46,$27,$5E,$2F,$FB,$FD,$B5,$BF
	.dc.b	$71,$EE,$7F,$79,$6E,$1F,$E2,$06,$BF,$E2,$D6,$BF,$E6,$D6,$BF,$EA
	.dc.b	$96,$BF,$B9,$6A,$69,$23,$C0,$48,$F0,$35,$F0,$35,$F0,$35,$F0,$35
	.dc.b	$F0,$35,$F0,$36,$3C,$11,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$A7,$62
	.dc.b	$BD,$9E,$84,$3A,$14,$E8,$53,$A1,$4E,$85,$7A,$3C,$88,$72,$29,$C8
	.dc.b	$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70
	.dc.b	$48,$FA,$84,$8F,$92,$D7,$E0,$35,$F4,$B5,$F2,$35,$F0,$B5,$F0,$FA
	.dc.b	$9F,$ED,$27,$FD,$88,$D7,$FD,$A1,$7F,$E7,$E8,$C4,$7E,$67,$E2,$BF
	.dc.b	$A9,$6D,$FA,$16,$C7,$E8,$5A,$FE,$85,$AF,$13,$FB,$B1,$FA,$17,$2D
	.dc.b	$33,$82,$32,$D8,$CC,$60,$8E,$09,$3E,$56,$9E,$52,$9E,$52,$9E,$57
	.dc.b	$81,$1E,$0D,$78,$35,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$F8,$4F,$56
	.dc.b	$9C,$8A,$71,$29,$C6,$7E,$34,$71,$6C,$71,$6B,$C5,$AF,$19,$F9,$73
	.dc.b	$C5,$B1,$C5,$AF,$16,$BC,$4F,$96,$79,$09,$3D,$9B,$1C,$1A,$70,$8D
	.dc.b	$12,$34,$D8,$D4,$60,$8E,$09,$3E,$96,$9E,$92,$9E,$92,$9E,$97,$81
	.dc.b	$FD,$9F,$D7,$91,$FD,$3F,$8D,$18,$8F,$EB,$13,$FE,$35,$8F,$EB,$CD
	.dc.b	$8F,$EB,$C7,$02,$C7,$F7,$B7,$07,$F8,$F4,$72,$7F,$80,$63,$44,$8D
	.dc.b	$1F,$D7,$35,$6A,$F7,$7F,$AD,$92,$77,$BF,$F6,$48,$F2,$C6,$CF,$2B
	.dc.b	$82,$3D,$BF,$5D,$76,$FF,$94,$67,$C3,$29,$CB,$CC,$F5,$38,$8D,$B6
	.dc.b	$3D,$DC,$CD,$1E,$5F,$F0,$68,$C3,$63,$33,$FE,$7D,$9C,$BF,$C2,$B9
	.dc.b	$BF,$DA,$33,$EC,$24,$7D,$07,$FC,$91,$5F,$1B,$5B,$A7,$D0,$AF,$F2
	.dc.b	$1F,$77,$8A,$31,$D9,$FE,$58,$BE,$15,$FE,$7F,$F3,$9F,$6F,$FE,$73
	.dc.b	$F7,$FF,$D0,$B1,$89,$46,$27,$51,$C5,$AF,$F9,$EF,$05,$3E,$F3,$33
	.dc.b	$F7,$23,$0F,$43,$C5,$1A,$9E,$2D,$9F,$F4,$CC,$BE,$03,$CB,$E5,$FF
	.dc.b	$A2,$5B,$19,$3D,$91,$97,$DD,$E4,$57,$CD,$5F,$E8,$A7,$81,$88,$D3
	.dc.b	$E7,$6E,$0F,$D9,$1D,$3F,$E9,$9F,$81,$FF,$55,$C6,$2D,$18,$9D,$47
	.dc.b	$FA,$B6,$75,$3F,$90,$F8,$AF,$F6,$A4,$F0,$3F,$46,$E2,$35,$E9,$72
	.dc.b	$9F,$D6,$4E,$5E,$9F,$77,$97,$F5,$36,$8C,$63,$E9,$6C,$66,$7C,$07
	.dc.b	$D9,$F8,$BF,$F2,$3D,$7E,$2C,$38,$E5,$FF,$64,$B6,$7F,$D9,$0D,$7F
	.dc.b	$D4,$1E,$B7,$E3,$9B,$5F,$92,$E6,$4E,$67,$F2,$ED,$E6,$FF,$56,$FB
	.dc.b	$1F,$C0,$BF,$0E,$FF,$53,$36,$BF,$D6,$AD,$7F,$D5,$CD,$7F,$84,$E3
	.dc.b	$F9,$39,$F3,$C6,$71,$F5,$B6,$74,$FF,$17,$FB,$67,$EA,$7F,$C4,$AB
	.dc.b	$F5,$91,$87,$D0,$D7,$31,$BD,$3F,$95,$57,$FB,$77,$CA,$79,$5F,$88
	.dc.b	$9C,$CA,$73,$C9,$DB,$B9,$F6,$BF,$CB,$EB,$FD,$CD,$9F,$FB,$CE,$31
	.dc.b	$08,$C3,$C6,$33,$9F,$EC,$BA,$7F,$DD,$DE,$3F,$E9,$89,$5F,$A6,$3F
	.dc.b	$C2,$47,$E8,$4A,$7E,$84,$B3,$FB,$CE,$9E,$71,$FA,$C3,$3C,$AF,$C9
	.dc.b	$3C,$2D,$8F,$C5,$37,$6F,$D3,$CF,$AB,$93,$FD,$D7,$CB,$FD,$03,$C4
	.dc.b	$FF,$72,$37,$0F,$F7,$9B,$5F,$C5,$C7,$F4,$8D,$BE,$93,$78,$FE,$C3
	.dc.b	$74,$E1,$5F,$1E,$72,$FF,$61,$9F,$F6,$51,$1C,$1F,$83,$E0,$FA,$6F
	.dc.b	$F4,$33,$E8,$38,$8F,$13,$63,$87,$07,$F4,$D7,$FB,$16,$4E,$7F,$BB
	.dc.b	$8B,$F0,$3F,$EC,$33,$9B,$FE,$C3,$18,$AD,$B6,$33,$7D,$8E,$45,$8F
	.dc.b	$1C,$23,$27,$E2,$B0,$F5,$CF,$3C,$1F,$81,$D2,$57,$A3,$0F,$3C,$74
	.dc.b	$8D,$9E,$17,$04,$7F,$C1,$9E,$27,$FB,$D4,$8E,$0F,$E4,$5F,$21,$FE
	.dc.b	$6A,$BF,$E5,$8D,$E2,$34,$FF,$37,$E9,$F2,$6F,$ED,$B6,$BE,$A3,$75
	.dc.b	$F9,$13,$F8,$09,$FE,$71,$D3,$D4,$98,$FF,$77,$3F,$9A,$3F,$E7,$8E
	.dc.b	$8F,$D9,$D7,$A2,$7F,$DF,$1D,$3F,$D0,$23,$11,$FC,$BB,$F0,$2F,$41
	.dc.b	$1E,$87,$F6,$3D,$11,$D6,$B9,$4F,$C3,$3A,$F5,$7F,$A2,$B0,$F0,$C7
	.dc.b	$F4,$7F,$47,$E2,$11,$EA,$7F,$3E,$7E,$47,$CA,$FD,$FF,$DA,$3F,$A9
	.dc.b	$B6,$7F,$A8,$35,$F2,$35,$FE,$8F,$F2,$FF,$87,$F8,$EF,$EB,$6D,$8F
	.dc.b	$EB,$0D,$7D,$8D,$7F,$8F,$65,$EC,$7D,$7A,$FC,$67,$F8,$F4,$65,$FE
	.dc.b	$AB,$E0,$7D,$F2,$3E,$CE,$0F,$ED,$D1,$F8,$5F,$A4,$7C,$67,$F9,$13
	.dc.b	$EC,$BF,$C8,$B7,$FE,$0F,$9C,$DA,$73,$3F,$E4,$A9,$CC,$62,$51,$87
	.dc.b	$E9,$4F,$AB,$83,$E9,$63,$E7,$8F,$4C,$79,$78,$9E,$D7,$FB,$22,$FF
	.dc.b	$BE,$36,$BF,$00,$D7,$FB,$9B,$63,$C3,$97,$F9,$7B,$8F,$C9,$97,$A9
	.dc.b	$B1,$99,$FF,$35,$1C,$0F,$C5,$9E,$8B,$F2,$EB,$F3,$F8,$FF,$13,$1F
	.dc.b	$91,$FB,$2F,$E0,$58,$7C,$1C,$D1,$8C,$3F,$C2,$8D,$AE,$07,$63,$F8
	.dc.b	$57,$C7,$3E,$41,$F7,$BE,$D8,$C3,$63,$33,$FE,$63,$BF,$F4,$15,$6A
	.dc.b	$30,$74,$F8,$2F,$F4,$2C,$E2,$12,$9C,$E9,$F7,$23,$1E,$D7,$FA,$1A
	.dc.b	$31,$3A,$BF,$89,$AF,$C2,$C3,$EE,$57,$EE,$30,$F1,$3C,$F1,$F3,$B7
	.dc.b	$2F,$51,$F6,$6B,$E6,$9D,$38,$BF,$09,$87,$58,$C9,$FA,$03,$5C,$5F
	.dc.b	$B6,$5E,$E3,$E4,$9E,$F7,$F3,$5F,$6F,$F8,$3C,$E6,$31,$9F,$FA,$A6
	.dc.b	$31,$7F,$E9,$88,$C5,$F2,$8C,$1F,$CF,$E7,$77,$D2,$74,$DB,$EB,$5B
	.dc.b	$9C,$DF,$F1,$28,$FF,$51,$E3,$FE,$14,$6E,$5F,$E5,$D7,$FD,$90,$63
	.dc.b	$2F,$F5,$4B,$FA,$63,$04,$3F,$03,$A7,$E1,$1F,$9C,$7F,$A1,$CF,$C6
	.dc.b	$B9,$DB,$7A,$FE,$82,$7C,$87,$80,$86,$0F,$C5,$70,$7F,$04,$9F,$C7
	.dc.b	$F0,$7E,$32,$7F,$0B,$39,$77,$3E,$17,$F1,$F7,$E1,$7C,$77,$E1,$72
	.dc.b	$DE,$0F,$CD,$47,$EA,$F8,$7F,$D4,$17,$FA,$A8,$FF,$24,$1F,$AA,$74
	.dc.b	$E0,$FE,$1C,$7C,$53,$D4,$60,$FD,$FF,$2F,$F1,$C8,$FE,$06,$71,$3F
	.dc.b	$81,$18,$3F,$79,$CD,$E7,$7F,$D2,$CD,$E4,$FE,$2F,$5E,$F9,$CB,$F0
	.dc.b	$4F,$B0,$7C,$C3,$C4,$79,$1F,$E2,$1A,$F8,$88,$F9,$B9,$BF,$DE,$F1
	.dc.b	$FC,$37,$A3,$F0,$73,$FC,$EA,$BF,$DD,$AD,$7F,$D4,$F1,$F8,$87,$C2
	.dc.b	$7C,$22,$9E,$23,$FA,$A8,$F8,$FF,$EA,$AD,$3F,$AA,$94,$FE,$AA,$53
	.dc.b	$FD,$47,$D9,$FE,$23,$FD,$A7,$FB,$2C,$1A,$47,$FB,$81,$E8,$C9,$DF
	.dc.b	$EC,$6C,$66,$3D,$5C,$BF,$A7,$8F,$94,$7A,$8F,$5E,$5F,$EC,$9E,$BE
	.dc.b	$44,$8E,$CF,$2E,$6F,$D7,$47,$C3,$AE,$8F,$D3,$C3,$FD,$32,$FF,$B4
	.dc.b	$3C,$0F,$21,$1F,$F2,$01,$FE,$39,$E8,$62,$F7,$3F,$C4,$AD,$9D,$35
	.dc.b	$FF,$24,$78,$7F,$E2,$D9,$FF,$6E,$F8,$9F,$04,$93,$F1,$5F,$47,$17
	.dc.b	$F0,$2C,$FC,$F3,$F6,$4F,$B7,$E5,$FF,$A5,$BE,$FF,$FC,$94,$48,$FF
	.dc.b	$2D,$3F,$E5,$32,$3F,$E5,$1F,$2F,$D3,$18,$8F,$F2,$D4,$7F,$A9,$1F
	.dc.b	$F3,$14,$FF,$95,$48,$F1,$CF,$FD,$4C,$D8,$FF,$30,$11,$FF,$35,$65
	.dc.b	$FE,$5D,$23,$83,$E5,$9F,$E0,$27,$F9,$01,$1C,$E1,$FE,$BA,$8F,$F4
	.dc.b	$85,$7C,$B9,$FF,$48,$3F,$4A,$32,$48,$FF,$4E,$C6,$F3,$FF,$51,$B7
	.dc.b	$C9,$FF,$73,$B7,$CD,$E5,$5F,$E0,$4E,$0F,$F5,$77,$85,$FE,$AC,$8F
	.dc.b	$E8,$87,$F9,$47,$CE,$FF,$41,$46,$4F,$5D,$FF,$AE,$E7,$FD,$6C,$7F
	.dc.b	$B0,$67,$3E,$C3,$D1,$FF,$45,$46,$1B,$3F,$EC,$98,$DF,$A7,$88,$FF
	.dc.b	$48,$5F,$FB,$45,$EA,$7E,$4A,$D3,$B8,$FF,$55,$F0,$79,$9F,$F4,$E6
	.dc.b	$3F,$E8,$F2,$3F,$3B,$99,$C7,$EE,$8F,$F1,$C4,$FF,$53,$23,$FE,$FB
	.dc.b	$F4,$7F,$DF,$51,$F5,$09,$19,$CB,$F4,$B1,$F1,$39,$FD,$38,$5A,$FD
	.dc.b	$4B,$6B,$FD,$F7,$1F,$E3,$49,$DB,$FE,$A7,$F5,$36,$D9,$DB,$FA,$CE
	.dc.b	$14,$EA,$30,$D5,$C5,$F8,$2C,$7F,$DC,$77,$EF,$8C,$4F,$E2,$27,$D1
	.dc.b	$F0,$5E,$A8,$F2,$35,$FA,$78,$A3,$11,$FC,$12,$31,$7F,$2C,$71,$3D
	.dc.b	$6F,$F6,$57,$A9,$F8,$88,$C5,$A3,$13,$A8,$FC,$A3,$5F,$E2,$7E,$BB
	.dc.b	$D8,$C4,$7F,$B5,$63,$0F,$33,$F5,$EE,$2B,$F6,$30,$8C,$3F,$A1,$E1
	.dc.b	$E7,$8F,$A1,$B8,$FE,$1A,$7A,$1F,$EB,$38,$DF,$27,$E7,$63,$FC,$9C
	.dc.b	$7F,$9B,$9C,$79,$9F,$9F,$8C,$9F,$9C,$8C,$BB,$E5,$FF,$77,$4F,$E8
	.dc.b	$A3,$67,$E7,$E3,$2E,$CC,$E0,$A7,$B7,$FE,$F7,$F8,$6F,$D8,$B6,$BC
	.dc.b	$8D,$7C,$4D,$7F,$59,$EE,$DB,$E8,$DB,$34,$E1,$C6,$E5,$7B,$FF,$84
	.dc.b	$8C,$36,$33,$7F,$91,$6F,$2F,$D4,$E1,$FF,$2F,$65,$F8,$6F,$0B,$FB
	.dc.b	$E4,$62,$3F,$0D,$1F,$E4,$6E,$E3,$E0,$8F,$BB,$18,$C7,$53,$99,$FE
	.dc.b	$63,$39,$7F,$C9,$FF,$00,$F4,$D1,$FC,$BC,$46,$5C,$36,$7F,$16,$D8
	.dc.b	$FE,$7B,$87,$91,$E4,$D9,$F7,$5F,$66,$CF,$7F,$81,$7C,$17,$F3,$48
	.dc.b	$FE,$E3,$18,$D7,$53,$99,$F3,$CE,$6B,$F3,$BE,$A7,$F8,$DB,$5B,$E7
	.dc.b	$FF,$23,$BF,$EF,$73,$97,$FA,$9B,$5F,$EF,$F9,$FA,$6B,$CA,$48,$F6
	.dc.b	$E7,$CD,$FE,$4E,$7F,$2B,$E7,$FE,$54,$FF,$B7,$A7,$F9,$4B,$FE,$0A
	.dc.b	$F8,$37,$C0,$7F,$8D,$DB,$1F,$E8,$06,$BF,$B6,$6B,$FD,$A1,$AF,$38
	.dc.b	$E4,$47,$93,$5E,$4D,$79,$35,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10
	.dc.b	$D1,$4D,$18,$21,$82,$1F,$90,$8F,$6B,$63,$CC,$D3,$C2,$53,$B1,$4E
	.dc.b	$A5,$3A,$88,$EA,$D8,$EA,$D3,$A9,$4E,$A5,$3A,$94,$EB,$1D,$09,$1D
	.dc.b	$1B,$1D,$1A,$F4,$6B,$D1,$AF,$48,$E4,$47,$93,$5E,$4D,$79,$35,$E4
	.dc.b	$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$FD,$2B,$D8,$FF,$75,$BF,$E3
	.dc.b	$C6,$BF,$EE,$CA,$FF,$6D,$C6,$23,$F8,$6F,$A1,$FC,$2F,$1F,$EA,$6D
	.dc.b	$8F,$F2,$2B,$5F,$F3,$8B,$5E,$3C,$5F,$C5,$A7,$8B,$63,$8B,$5E,$2D
	.dc.b	$78,$9F,$C5,$4C,$94,$CC,$74,$70,$7F,$A9,$E3,$93,$4C,$CF,$95,$C4
	.dc.b	$F1,$23,$E4,$37,$3C,$A7,$81,$1C,$12,$36,$49,$E6,$FA,$1A,$FA,$0E
	.dc.b	$CE,$1F,$19,$82,$1C,$4A,$71,$30,$43,$04,$39,$99,$29,$99,$F5,$3E
	.dc.b	$73,$A9,$A2,$3A,$8F,$33,$63,$6D,$7B,$35,$ED,$3F,$16,$78,$B6,$38
	.dc.b	$B5,$E2,$D7,$89,$E5,$32,$53,$31,$D1,$C1,$E8,$8E,$4D,$33,$3E,$57
	.dc.b	$DD,$F0,$9F,$ED,$EE,$27,$C0,$FF,$8C,$23,$0D,$F4,$3F,$D9,$6D,$9D
	.dc.b	$D7,$FA,$9B,$8F,$FB,$04,$FF,$8E,$1B,$19,$E4,$E2,$DA,$E2,$FF,$B6
	.dc.b	$3C,$C7,$8F,$FE,$7E,$C3,$FC,$83,$18,$E2,$DD,$7F,$80,$23,$13,$AA
	.dc.b	$E1,$3F,$EB,$8F,$DA,$9E,$EF,$FB,$8B,$83,$FC,$97,$18,$CB,$F4,$ED
	.dc.b	$8C,$CE,$EB,$CD,$1E,$BC,$F1,$1F,$DF,$9A,$FF,$7C,$6A,$F8,$F7,$E2
	.dc.b	$7F,$B2,$6C,$7E,$D9,$B5,$FE,$5E,$6C,$7F,$B4,$3C,$8F,$F4,$3E,$5F
	.dc.b	$BC,$6D,$7F,$99,$1B,$3B,$F2,$FC,$3C,$1F,$E1,$1A,$F4,$5F,$F0,$48
	.dc.b	$C5,$7B,$0F,$F6,$EE,$AF,$63,$F8,$3B,$63,$F8,$5B,$6F,$FC,$E6,$D8
	.dc.b	$FF,$18,$8B,$7E,$C0,$FC,$43,$FC,$6B,$F6,$DF,$ED,$08,$C5,$23,$13
	.dc.b	$A8,$E7,$5C,$A3,$11,$F3,$E3,$D5,$BF,$B8,$8F,$F2,$31,$F9,$8F,$AD
	.dc.b	$C3,$72,$FF,$70,$B6,$BB,$B5,$FF,$5B,$7C,$97,$AF,$F3,$BF,$F8,$03
	.dc.b	$E3,$FF,$99,$E3,$13,$F0,$23,$18,$2B,$7E,$3A,$F7,$7F,$EC,$1E,$4A
	.dc.b	$FF,$AF,$A7,$F0,$B7,$C6,$3A,$46,$2F,$9D,$70,$AF,$9B,$F8,$27,$F9
	.dc.b	$83,$EB,$7E,$63,$70,$F9,$3C,$08,$C4,$7F,$BD,$23,$0A,$7E,$A7,$DF
	.dc.b	$6A,$3E,$D4,$A3,$17,$F7,$1B,$19,$9E,$47,$D8,$FC,$05,$EC,$FE,$0F
	.dc.b	$F1,$5F,$C2,$CF,$E1,$67,$38,$27,$2A,$33,$18,$6C,$F2,$F9,$8F,$99
	.dc.b	$5C,$63,$1B,$F7,$6C,$65,$FC,$AC,$F0,$BE,$9F,$35,$B9,$FF,$44,$D7
	.dc.b	$86,$31,$1E,$76,$C6,$4E,$91,$F1,$BE,$77,$F2,$AD,$3F,$38,$7D,$27
	.dc.b	$31,$FA,$46,$E6,$F8,$7A,$4A,$E2,$7D,$77,$AF,$1F,$FA,$DF,$D3,$FD
	.dc.b	$77,$ED,$7F,$8E,$6D,$7E,$21,$AF,$FA,$92,$78,$62,$AF,$CA,$FE,$11
	.dc.b	$AA,$FC,$93,$6B,$FD,$88,$DE,$63,$BF,$83,$FF,$79,$72,$7E,$FF,$B3
	.dc.b	$EE,$C6,$32,$F4,$E2,$8C,$4E,$BD,$CF,$C8,$56,$A9,$E3,$FF,$3D,$AF
	.dc.b	$E7,$5E,$46,$63,$13,$FE,$97,$9E,$F0,$F0,$FE,$8E,$3C,$9B,$19,$AF
	.dc.b	$1C,$E5,$F2,$E1,$FA,$A7,$F9,$99,$FA,$5E,$AF,$F6,$5F,$5F,$EB,$23
	.dc.b	$27,$E9,$23,$2F,$EB,$BA,$EB,$FA,$D6,$BF,$AA,$6D,$7E,$A6,$39,$BF
	.dc.b	$83,$5F,$79,$FC,$97,$9B,$FD,$B6,$D8,$C9,$F8,$B2,$9F,$98,$E1,$7C
	.dc.b	$3F,$D3,$8F,$E5,$31,$8B,$F7,$B7,$6F,$D7,$B6,$7F,$9C,$E9,$F6,$70
	.dc.b	$FA,$DC,$3F,$CE,$DB,$87,$82,$BE,$E7,$0F,$DA,$9F,$AE,$48,$FF,$66
	.dc.b	$35,$FF,$7A,$DF,$D1,$8F,$B3,$C1,$EE,$9F,$0F,$07,$D7,$9C,$C7,$7E
	.dc.b	$36,$6F,$A2,$FE,$9D,$F7,$7F,$CB,$9B,$5F,$CA,$1A,$FE,$86,$78,$68
	.dc.b	$AE,$A4,$9F,$0E,$47,$46,$0A,$73,$BF,$EB,$DF,$77,$E1,$B3,$49,$3C
	.dc.b	$5F,$59,$1C,$1F,$EF,$1F,$0F,$F3,$E4,$70,$7F,$42,$FC,$2B,$FC,$1B
	.dc.b	$9F,$C2,$6E,$66,$98,$F7,$7F,$B9,$7E,$01,$FD,$49,$B7,$FD,$31,$AF
	.dc.b	$C5,$6B,$FD,$03,$8B,$FD,$21,$9F,$F7,$A6,$F7,$7F,$3E,$CF,$FB,$D4
	.dc.b	$7F,$A3,$38,$7F,$C6,$91,$E9,$BF,$F1,$D7,$92,$EB,$FF,$49,$D6,$A7
	.dc.b	$FA,$AF,$C2,$74,$23,$FD,$E7,$14,$75,$BE,$58,$75,$AF,$F2,$04,$61
	.dc.b	$F8,$E7,$DF,$F8,$0F,$8A,$4C,$FF,$D4,$DC,$1F,$8A,$8F,$F1,$BC,$7F
	.dc.b	$85,$BE,$8F,$F9,$99,$2B,$F8,$2E,$48,$FE,$63,$C2,$F8,$3F,$A2,$4B
	.dc.b	$EB,$19,$7D,$F9,$79,$70,$F8,$7B,$60,$9A,$7F,$81,$BE,$2B,$ED,$92
	.dc.b	$F9,$BE,$6F,$A3,$4F,$F8,$B7,$EF,$3F,$C5,$D8,$7F,$58,$9C,$DA,$73
	.dc.b	$3F,$E3,$39,$CC,$62,$51,$87,$E4,$CF,$CC,$C3,$CA,$E3,$77,$ED,$3F
	.dc.b	$6E,$BB,$CF,$F4,$D8,$C6,$3E,$0A,$DD,$FF,$9B,$EF,$FC,$4D,$5A,$7F
	.dc.b	$C7,$B8,$7A,$32,$F2,$36,$33,$87,$F9,$2E,$72,$D8,$F3,$E3,$FD,$23
	.dc.b	$3F,$9C,$FA,$67,$F0,$D2,$FD,$5F,$FD,$17,$2B,$F8,$51,$86,$DF,$F0
	.dc.b	$16,$CE,$EB,$FC,$A7,$4F,$AA,$FF,$7D,$46,$1B,$3F,$E0,$A9,$D3,$F1
	.dc.b	$8F,$F3,$23,$8C,$3D,$FE,$6F,$C5,$9F,$B4,$D7,$EC,$36,$FE,$BB,$98
	.dc.b	$F8,$1A,$7D,$67,$F1,$B5,$F0,$2F,$B5,$F9,$67,$C4,$FC,$3C,$3C,$F1
	.dc.b	$E4,$6B,$D2,$BF,$5C,$D3,$E7,$F2,$FE,$E5,$EE,$63,$3F,$A2,$BD,$F7
	.dc.b	$CD,$3E,$07,$F9,$67,$C0,$FE,$6A,$BF,$97,$56,$A7,$38,$FF,$A2,$A3
	.dc.b	$12,$AE,$0F,$5B,$E8,$E6,$B8,$46,$27,$75,$C9,$B1,$99,$DD,$F8,$2F
	.dc.b	$F4,$07,$F8,$40,$87,$F9,$63,$1F,$F6,$83,$83,$F7,$65,$38,$9F,$3C
	.dc.b	$FC,$6E,$5F,$E9,$A7,$F0,$8D,$7E,$81,$FE,$50,$39,$9D,$F2,$FE,$50
	.dc.b	$F5,$3E,$F9,$FE,$66,$E0,$FF,$30,$BF,$8A,$D3,$93,$F8,$86,$9F,$88
	.dc.b	$2E,$BF,$94,$8F,$29,$8B,$F2,$4F,$E1,$63,$77,$FD,$8F,$CE,$FD,$19
	.dc.b	$27,$F4,$71,$97,$D1,$BF,$F9,$FA,$BE,$19,$EC,$21,$FC,$4F,$4F,$F6
	.dc.b	$35,$FE,$F2,$BF,$CC,$47,$F0,$02,$CF,$F1,$D9,$DC,$7F,$0A,$AF,$E4
	.dc.b	$2D,$3E,$0F,$1F,$F1,$06,$F1,$FF,$17,$BF,$E4,$67,$CA,$3C,$E7,$CD
	.dc.b	$D7,$EE,$BE,$1F,$8D,$F2,$12,$B0,$7F,$AD,$39,$39,$CF,$FA,$E5,$B8
	.dc.b	$7F,$85,$9B,$CD,$FC,$26,$7B,$6B,$F7,$1F,$09,$0F,$F0,$B7,$3F,$02
	.dc.b	$61,$F2,$F0,$FF,$10,$C7,$82,$3F,$AC,$4B,$EC,$FF,$DD,$9D,$3F,$E1
	.dc.b	$96,$C6,$4F,$F6,$1C,$65,$FF,$66,$F5,$F9,$89,$5D,$F2,$47,$BA,$7E
	.dc.b	$41,$F6,$8A,$7D,$E3,$FD,$8F,$D9,$F8,$16,$9F,$83,$A7,$67,$F7,$C8
	.dc.b	$C3,$5F,$8B,$C3,$C7,$83,$E9,$C7,$E2,$2B,$FC,$24,$F0,$ED,$ED,$18
	.dc.b	$8F,$F0,$99,$E9,$EE,$43,$ED,$7F,$C4,$4D,$9F,$F1,$13,$5F,$F1,$13
	.dc.b	$5F,$F7,$A7,$83,$F6,$89,$1F,$15,$F8,$98,$79,$A7,$E1,$1F,$10,$F8
	.dc.b	$BE,$3F,$FA,$3F,$EB,$3F,$CB,$64,$8F,$F2,$63,$FE,$49,$23,$FE,$59
	.dc.b	$B4,$62,$3F,$CD,$35,$FE,$9B,$FB,$9F,$F3,$29,$23,$FC,$AC,$FF,$94
	.dc.b	$C8,$F3,$6B,$CB,$C7,$FF,$2F,$7D,$C7,$F9,$B8,$91,$FE,$5F,$7F,$D6
	.dc.b	$24,$7F,$7A,$D7,$FC,$F7,$E5,$FF,$AE,$A3,$FD,$0A,$48,$FF,$45,$C6
	.dc.b	$5F,$F4,$9F,$9B,$F5,$39,$BF,$D2,$2D,$8C,$E7,$F9,$77,$FD,$39,$19
	.dc.b	$7F,$DE,$BC,$4F,$3B,$D3,$1F,$EA,$3C,$7F,$D4,$C4,$7F,$D6,$83,$DD
	.dc.b	$FF,$3E,$3F,$EB,$53,$06,$9C,$BF,$EB,$73,$27,$FB,$0C,$A7,$FB,$04
	.dc.b	$B8,$7F,$B2,$63,$FD,$8C,$7F,$A6,$31,$93,$FE,$89,$6B,$FE,$42,$7F
	.dc.b	$DB,$A4,$7F,$94,$BF,$EA,$E2,$3F,$2B,$8F,$E1,$3F,$EB,$06,$F5,$7F
	.dc.b	$BD,$AD,$F9,$BF,$FB,$F1,$B5,$F0,$27,$53,$93,$DE,$FF,$AB,$3D,$6F
	.dc.b	$D3,$C6,$3A,$91,$8A,$E5,$18,$BE,$B1,$FE,$AD,$9D,$3F,$5B,$C2,$FE
	.dc.b	$1C,$D3,$EE,$1F,$B4,$D9,$1A,$8C,$36,$BF,$DC,$99,$FF,$B8,$49,$3F
	.dc.b	$EE,$17,$EB,$C3,$DA,$7B,$1F,$C0,$3F,$17,$CC,$62,$51,$89,$D5,$FE
	.dc.b	$12,$79,$46,$23,$F0,$FE,$DB,$CE,$C4,$7F,$B1,$63,$15,$D2,$3F,$13
	.dc.b	$18,$BF,$EC,$4D,$8E,$D8,$3D,$CE,$77,$E9,$6C,$66,$3A,$4F,$FB,$4D
	.dc.b	$D1,$87,$F2,$DF,$8D,$FC,$2D,$5E,$C7,$FB,$66,$30,$D9,$FD,$1C,$6F
	.dc.b	$D9,$C4,$7E,$76,$30,$9A,$7E,$0F,$DA,$FE,$57,$F9,$ED,$C4,$D2,$57
	.dc.b	$E1,$B6,$DB,$67,$6D,$9F,$C7,$4E,$A3,$DD,$F0,$0F,$82,$FE,$BC,$DB
	.dc.b	$FE,$94,$D9,$FC,$0B,$5F,$DC,$FC,$1B,$E0,$FF,$AC,$36,$BF,$AA,$36
	.dc.b	$7F,$C2,$CD,$7F,$05,$B7,$F0,$4E,$4F,$99,$D7,$F9,$38,$C4,$EA,$3F
	.dc.b	$30,$BC,$DF,$43,$FD,$85,$AF,$F8,$8D,$AF,$F0,$E6,$BF,$C5,$2F,$C4
	.dc.b	$DC,$7D,$24,$C7,$FB,$67,$C2,$3C,$2F,$EE,$8D,$8C,$C6,$0F,$EF,$1F
	.dc.b	$0A,$E1,$FE,$E9,$82,$30,$FF,$77,$6B,$FE,$39,$6B,$F3,$30,$F1,$3F
	.dc.b	$21,$AF,$F2,$8F,$87,$7C,$3F,$F8,$1D,$B8,$FF,$80,$DB,$3F,$E4,$56
	.dc.b	$BF,$E4,$8F,$67,$FC,$91,$E4,$7C,$BF,$3B,$FC,$FB,$3F,$EE,$89,$CE
	.dc.b	$E9,$CC,$86,$FF,$7E,$BC,$D8,$EA,$73,$9F,$F8,$12,$72,$7F,$95,$FE
	.dc.b	$29,$F1,$59,$AF,$E8,$71,$8B,$E4,$DA,$E1,$5B,$AF,$E9,$BF,$17,$FE
	.dc.b	$97,$7F,$C4,$4D,$7F,$CC,$8D,$7F,$A7,$B5,$E3,$EF,$7D,$68,$C4,$7B
	.dc.b	$5A,$9A,$48,$E0,$48,$F0,$B5,$F0,$B5,$F0,$B5,$F0,$9C,$08,$70,$29
	.dc.b	$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$87,$BD,$F2,$34,$EE,$53,$A1
	.dc.b	$4E,$65,$39,$9F,$1D,$E6,$D3,$99,$4E,$65,$39,$94,$E7,$1F,$3D,$E2
	.dc.b	$D7,$8B,$5E,$2D,$78,$C7,$8A,$31,$1E,$72,$1D,$CA,$E9,$C1,$1C,$12
	.dc.b	$3A,$12,$3C,$AD,$7C,$AD,$7F,$C3,$FE,$FF,$FB,$B9,$BE,$FF,$E9,$DF
	.dc.b	$F7,$76,$1F,$E7,$B9,$D4,$7E,$7B,$E1,$FF,$61,$EE,$FE,$71,$B1,$F9
	.dc.b	$C6,$BB,$3F,$DF,$27,$E6,$DC,$B4,$C9,$F9,$D7,$2D,$33,$8A,$32,$D8
	.dc.b	$CC,$60,$8E,$09,$3E,$76,$9E,$72,$9E,$77,$44,$74,$D7,$46,$08,$60
	.dc.b	$87,$B5,$E0,$D3,$65,$36,$7B,$DD,$B4,$D9,$4D,$CF,$C1,$8D,$B6,$36
	.dc.b	$D7,$73,$F1,$27,$6D,$8D,$B5,$D9,$F1,$0F,$31,$0F,$11,$67,$B3,$63
	.dc.b	$93,$4E,$47,$02,$47,$06,$C7,$06,$9C,$23,$44,$8D,$36,$35,$18,$23
	.dc.b	$83,$F1,$BF,$9E,$3D,$FF,$F6,$2F,$C6,$7F,$A0,$25,$39,$DD,$39,$8C
	.dc.b	$36,$B8,$D7,$F6,$5A,$DB,$67,$FA,$89,$1F,$F4,$6F,$27,$F5,$D8,$FE
	.dc.b	$B7,$C7,$C2,$7F,$D2,$B3,$97,$7C,$9F,$D7,$EB,$CD,$F1,$ED,$7F,$B0
	.dc.b	$46,$1B,$5F,$E9,$A9,$D3,$D4,$FE,$CC,$E2,$7E,$07,$E0,$6F,$17,$FD
	.dc.b	$CF,$78,$6D,$FF,$A9,$27,$59,$FE,$D9,$B5,$EA,$24,$FA,$33,$FD,$DE
	.dc.b	$7F,$DE,$E7,$FC,$45,$E5,$FE,$F3,$4F,$F3,$BD,$FF,$00,$6C,$EA,$3F
	.dc.b	$81,$F7,$3E,$03,$C7,$1F,$EB,$A8,$C6,$7A,$9C,$CF,$F8,$0E,$73,$5E
	.dc.b	$48,$F9,$B7,$FE,$09,$8C,$47,$F8,$27,$1F,$33,$FE,$C2,$FD,$47,$FD
	.dc.b	$9C,$FF,$14,$6C,$FE,$1D,$B1,$BF,$57,$F8,$BF,$D8,$3E,$A3,$C7,$28
	.dc.b	$C6,$9F,$C7,$9B,$87,$E4,$5B,$3F,$26,$3C,$AB,$1F,$2B,$EA,$DB,$F9
	.dc.b	$3E,$CF,$7F,$F9,$3B,$67,$31,$83,$F9,$7E,$2F,$86,$FC,$2F,$C8,$FF
	.dc.b	$AA,$9B,$5F,$EA,$36,$BF,$E9,$16,$A9,$E5,$E6,$7F,$8C,$6B,$93,$67
	.dc.b	$93,$5E,$4D,$79,$79,$6F,$A4,$E9,$F8,$C7,$F9,$F1,$B3,$FE,$B5,$9D
	.dc.b	$60,$AD,$36,$7F,$DB,$DF,$12,$E5,$E3,$1F,$E4,$5F,$6D,$39,$AE,$13
	.dc.b	$98,$C5,$FF,$B9,$1B,$1F,$0F,$BB,$FC,$9D,$F2,$7F,$1D,$B1,$91,$19
	.dc.b	$69,$FE,$55,$3F,$CA,$DF,$00,$F8,$6E,$31,$98,$C3,$C0,$C4,$69,$E2
	.dc.b	$DC,$7E,$D7,$C7,$DE,$13,$49,$5C,$CF,$F3,$37,$3B,$F1,$3F,$CA,$B6
	.dc.b	$FF,$29,$18,$3F,$33,$CB,$F8,$BE,$8F,$E4,$CD,$BF,$61,$FE,$F0,$E6
	.dc.b	$FC,$73,$FD,$FA,$DC,$6F,$A9,$FC,$47,$E4,$FF,$EB,$86,$D7,$DD,$6B
	.dc.b	$FE,$C2,$6A,$E4,$FB,$3F,$79,$86,$E0,$BF,$B1,$39,$E4,$3C,$AF,$F6
	.dc.b	$5D,$9F,$00,$97,$FB,$37,$F6,$FD,$9F,$AE,$9F,$E0,$9D,$FF,$C1,$DF
	.dc.b	$DA,$10,$FB,$9B,$7C,$36,$ED,$F0,$B7,$F8,$EF,$EC,$37,$FC,$CF,$D9
	.dc.b	$FE,$60,$95,$F9,$DE,$BF,$D6,$47,$E4,$7A,$EC,$FE,$71,$AF,$E5,$DA
	.dc.b	$FC,$5D,$FF,$39,$3F,$C2,$F8,$FE,$D6,$FF,$70,$99,$F9,$2F,$F8,$29
	.dc.b	$1C,$1F,$C4,$F7,$F8,$AD,$9F,$8C,$7F,$3C,$DB,$E2,$57,$F1,$39,$FE
	.dc.b	$2F,$BF,$C9,$8F,$C2,$1D,$4F,$C3,$EF,$F1,$DF,$F5,$C3,$5F,$E4,$D8
	.dc.b	$FA,$2B,$D5,$87,$A2,$78,$EF,$F3,$5A,$75,$3F,$5D,$C9,$F9,$BA,$E2
	.dc.b	$48,$FC,$BC,$65,$E3,$1F,$01,$A4,$3D,$87,$DF,$3F,$07,$DD,$F0,$4F
	.dc.b	$CF,$7E,$01,$FE,$B3,$8B,$E4,$FF,$2E,$8C,$57,$E9,$B1,$46,$2B,$F8
	.dc.b	$24,$63,$DC,$FF,$6D,$7C,$A6,$E9,$18,$AD,$C6,$14,$EE,$75,$1F,$EE
	.dc.b	$3A,$EF,$F7,$FE,$32,$47,$A1,$FF,$36,$EF,$F5,$C9,$C1,$F7,$27,$DF
	.dc.b	$97,$8F,$3F,$61,$23,$DE,$7C,$38,$FC,$0C,$F1,$8F,$EB,$0F,$3C,$3E
	.dc.b	$0F,$D5,$7C,$52,$4F,$CA,$7F,$D0,$7D,$5F,$CD,$B8,$BF,$A7,$7E,$A3
	.dc.b	$FC,$CD,$B5,$FC,$F9,$AF,$F3,$96,$BF,$D5,$77,$7C,$AF,$D3,$24,$7D
	.dc.b	$87,$E0,$E5,$E7,$D7,$E0,$F0,$FA,$EB,$E7,$FC,$6F,$DB,$23,$F8,$07
	.dc.b	$FB,$5F,$E8,$8F,$A8,$FF,$14,$EF,$F7,$23,$0F,$F8,$DC,$C4,$6A,$BE
	.dc.b	$B4,$EB,$83,$FB,$64,$7E,$40,$87,$D2,$3F,$25,$5F,$8A,$FA,$EF,$F5
	.dc.b	$8B,$F8,$46,$C7,$E0,$75,$46,$23,$E2,$36,$57,$E1,$7C,$A7,$9E,$7C
	.dc.b	$B8,$79,$DF,$F0,$04,$ED,$F2,$9F,$7B,$2F,$90,$FA,$2B,$F3,$FD,$BD
	.dc.b	$F1,$FE,$FD,$3D,$4F,$AD,$19,$35,$F7,$52,$35,$5F,$A8,$8C,$67,$F1
	.dc.b	$A7,$C7,$F8,$2F,$96,$BD,$6D,$C8,$D3,$D5,$3E,$7A,$FC,$BE,$5E,$88
	.dc.b	$FC,$46,$1D,$B0,$FD,$7B,$63,$DF,$B7,$E1,$AF,$89,$23,$F8,$31,$FE
	.dc.b	$E8,$1E,$CF,$F9,$C9,$C3,$7D,$77,$B2,$F6,$55,$C2,$FF,$C1,$15,$A9
	.dc.b	$CD,$F1,$BE,$17,$B3,$8D,$6D,$C1,$E0,$BE,$73,$C2,$BC,$98,$38,$DA
	.dc.b	$7F,$A8,$F0,$FB,$87,$FA,$58,$87,$F9,$F4,$A7,$F9,$4C,$A7,$CA,$3F
	.dc.b	$89,$1F,$93,$CF,$F8,$FB,$4C,$9B,$3E,$E1,$E0,$3B,$10,$EC,$7E,$B7
	.dc.b	$5F,$D3,$EB,$4E,$67,$E0,$8B,$B7,$E0,$9C,$F0,$2F,$C7,$7E,$19,$FB
	.dc.b	$EE,$A3,$31,$F9,$A8,$E3,$C1,$8A,$FC,$DC,$7E,$4E,$74,$FA,$B3,$F9
	.dc.b	$99,$7C,$B6,$C7,$C6,$3D,$C7,$EB,$A7,$CF,$C3,$EC,$8C,$1D,$72,$FE
	.dc.b	$16,$F7,$C3,$F7,$AF,$B3,$8B,$C8,$47,$E6,$9D,$8F,$EB,$27,$87,$0F
	.dc.b	$1B,$FC,$00,$ED,$7E,$17,$F7,$ED,$C3,$F8,$3B,$6D,$C7,$F8,$79,$FE
	.dc.b	$BA,$D8,$CD,$7F,$BB,$67,$2F,$C3,$DB,$F1,$EF,$8D,$AF,$DF,$3F,$9D
	.dc.b	$72,$73,$23,$CE,$BF,$9E,$BF,$EF,$9E,$7F,$E9,$16,$FD,$43,$FA,$43
	.dc.b	$67,$FA,$43,$5F,$E9,$0D,$7F,$A4,$74,$FA,$48,$F7,$7F,$48,$FC,$B2
	.dc.b	$69,$F2,$9F,$CA,$9F,$CC,$7A,$BE,$11,$1F,$43,$CA,$38,$47,$D4,$3E
	.dc.b	$01,$F5,$EB,$FB,$CB,$EF,$EB,$F9,$84,$7E,$17,$67,$F0,$8F,$98,$FF
	.dc.b	$01,$41,$A4,$7F,$C1,$38,$78,$DF,$EF,$87,$90,$FA,$9C,$71,$FA,$8F
	.dc.b	$F0,$3E,$CF,$F6,$64,$7F,$92,$48,$7F,$8C,$8F,$F1,$B9,$0F,$F1,$7F
	.dc.b	$83,$FB,$C7,$FD,$22,$D7,$FC,$8A,$72,$3F,$CD,$27,$7F,$07,$E7,$F8
	.dc.b	$3F,$E9,$58,$E8,$7F,$8F,$A7,$94,$66,$3C,$44,$39,$9D,$4F,$9C,$7D
	.dc.b	$62,$90,$FF,$42,$E9,$FC,$86,$3F,$DE,$F1,$FE,$61,$21,$D4,$FF,$2C
	.dc.b	$D7,$E1,$2F,$E2,$C6,$4F,$F2,$DD,$73,$BF,$39,$1F,$CC,$78,$9F,$ED
	.dc.b	$3F,$AB,$FF,$40,$41,$A4,$AF,$F4,$34,$65,$B1,$FE,$E2,$23,$E1,$BF
	.dc.b	$1B,$F0,$8F,$26,$FF,$90,$7F,$33,$CB,$FB,$07,$E2,$1F,$08,$F7,$9F
	.dc.b	$E6,$CF,$13,$E7,$9F,$6F,$93,$FD,$63,$F8,$67,$FA,$C3,$23,$49,$5F
	.dc.b	$EB,$D8,$CB,$63,$FD,$85,$18,$8F,$F6,$03,$5F,$9A,$7E,$0C,$87,$DB
	.dc.b	$3E,$69,$F8,$43,$F1,$AF,$CF,$3F,$88,$79,$BF,$EE,$48,$7C,$87,$F9
	.dc.b	$CE,$30,$DF,$3B,$FB,$8B,$67,$67,$FB,$7B,$D0,$FF,$7A,$C7,$FB,$82
	.dc.b	$11,$87,$5E,$87,$26,$9C,$8F,$F3,$A5,$BD,$2D,$46,$1A,$B1,$4E,$E3
	.dc.b	$91,$97,$18,$7E,$23,$F1,$49,$3F,$EC,$7A,$46,$23,$FD,$09,$E7,$BC
	.dc.b	$4E,$D3,$FE,$95,$8C,$57,$0A,$FC,$7C,$E6,$B7,$1F,$E9,$8C,$1E,$AF
	.dc.b	$F8,$63,$E3,$7E,$63,$67,$4D,$7E,$0B,$5F,$57,$97,$71,$F7,$B6,$CD
	.dc.b	$38,$4F,$F0,$83,$F3,$DD,$3F,$B1,$3F,$96,$72,$7E,$6E,$3F,$79,$CB
	.dc.b	$F9,$F3,$F4,$1D,$5F,$BC,$8F,$F6,$1F,$37,$E7,$32,$FE,$00,$FF,$15
	.dc.b	$F6,$5E,$CF,$F4,$06,$D7,$F8,$95,$AF,$E0,$DA,$FF,$3A,$C3,$C1,$8F
	.dc.b	$8E,$3F,$A4,$7A,$6F,$6B,$F9,$13,$72,$FF,$18,$36,$7F,$1C,$D7,$FC
	.dc.b	$8A,$7F,$91,$A9,$D7,$FD,$51,$B1,$FE,$39,$6B,$F9,$76,$BE,$57,$C9
	.dc.b	$87,$9B,$BF,$FC,$81,$ED,$BD,$CF,$E6,$CD,$CB,$31,$83,$F9,$D7,$BD
	.dc.b	$F5,$B1,$FF,$29,$9E,$67,$1E,$A7,$E9,$DB,$1E,$A0,$E1,$F6,$E3,$FE
	.dc.b	$28,$9F,$D7,$36,$38,$35,$DE,$5F,$E5,$89,$FD,$A7,$C2,$3E,$13,$F5
	.dc.b	$51,$86,$D7,$86,$75,$97,$A2,$3F,$CE,$31,$8A,$FD,$F7,$C1,$3E,$0B
	.dc.b	$8D,$7D,$28,$C5,$F1,$84,$62,$B8,$D7,$09,$FF,$3E,$F9,$1F,$BB,$7F
	.dc.b	$85,$35,$F8,$8D,$7E,$FB,$5E,$3E,$07,$F8,$4A,$31,$06,$92,$34,$48
	.dc.b	$FF,$3E,$46,$5B,$19,$7C,$6D,$7C,$6D,$7C,$6D,$7C,$67,$22,$1C,$8A
	.dc.b	$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1
	.dc.b	$0C,$10,$F9,$4F,$B5,$A7,$98,$A7,$84,$A7,$62,$9D,$4A,$75,$8F,$A0
	.dc.b	$F1,$6B,$C5,$AF,$16,$BC,$63,$C9,$18,$8E,$A4,$70,$48,$D9,$23,$D5
	.dc.b	$19,$6C,$7F,$97,$3D,$8F,$F6,$44,$7E,$D5,$AF,$FB,$22,$FF,$D8,$91
	.dc.b	$88,$FF,$6B,$FE,$ED,$F9,$17,$FD,$BC,$62,$FF,$DB,$7D,$7F,$D9,$E7
	.dc.b	$4D,$8C,$CF,$F0,$39,$CB,$63,$2F,$F2,$08,$CB,$63,$2F,$73,$C2,$23
	.dc.b	$6E,$67,$C0,$EA,$32,$EE,$79,$B4,$D3,$88,$D9,$1C,$12,$36,$49,$EE
	.dc.b	$47,$D8,$64,$AE,$5F,$61,$92,$B9,$7D,$47,$02,$1A,$29,$A3,$04,$30
	.dc.b	$43,$DC,$F5,$69,$C8,$A7,$12,$9C,$67,$E1,$3E,$73,$13,$F0,$E7,$A4
	.dc.b	$E9,$B1,$99,$ED,$39,$6C,$65,$F0,$46,$5B,$19,$7B,$9E,$13,$CA,$48
	.dc.b	$E0,$EA,$7F,$C7,$1F,$A6,$3E,$A3,$FC,$5D,$F0,$B8,$6D,$7F,$55,$9D
	.dc.b	$6C,$AD,$36,$7F,$AD,$6D,$FD,$0A,$7F,$9F,$CE,$67,$FA,$36,$DD,$1B
	.dc.b	$19,$43,$F0,$57,$E2,$BB,$8C,$C6,$27,$84,$6E,$74,$D8,$CF,$23,$91
	.dc.b	$D1,$86,$E5,$FD,$B1,$B7,$FD,$B1,$AF,$F7,$9D,$BC,$B3,$FD,$E5,$B1
	.dc.b	$FD,$E7,$9B,$FC,$01,$39,$8F,$2F,$D6,$3E,$B7,$71,$FE,$18,$8C,$3A
	.dc.b	$8C,$BF,$0A,$33,$3F,$E6,$DE,$A7,$CE,$FF,$98,$F1,$C3,$6F,$FC,$15
	.dc.b	$3A,$D7,$F8,$0B,$6B,$C8,$49,$F4,$3F,$BE,$CB,$FC,$30,$FF,$89,$7C
	.dc.b	$7F,$E0,$3A,$7F,$99,$6B,$F8,$5B,$67,$51,$FC,$3B,$E3,$5E,$DF,$8E
	.dc.b	$3F,$C6,$B1,$8C,$B5,$39,$9F,$BF,$39,$AF,$24,$7C,$F3,$F8,$CE,$3E
	.dc.b	$47,$CE,$7F,$1C,$F0,$9F,$4F,$FC,$59,$B1,$98,$C1,$FC,$93,$E8,$9E
	.dc.b	$1B,$C8,$C3,$6B,$F9,$BC,$EA,$72,$7C,$5C,$64,$F9,$1A,$7F,$8E,$7D
	.dc.b	$37,$AA,$F9,$AF,$CC,$46,$1B,$19,$DF,$F4,$13,$96,$C7,$F8,$65,$7E
	.dc.b	$DF,$D6,$30,$DF,$07,$FD,$48,$D9,$DE,$3F,$E6,$4B,$F1,$FB,$5A,$8C
	.dc.b	$60,$8C,$47,$EC,$27,$75,$CA,$31,$5B,$8C,$47,$FA,$CF,$F5,$17,$E2
	.dc.b	$3F,$B8,$F9,$0F,$F2,$D5,$27,$34,$9C,$C6,$1B,$5C,$67,$FC,$F7,$5C
	.dc.b	$7E,$63,$11,$FC,$12,$30,$F3,$37,$69,$DC,$79,$5B,$C0,$FC,$05,$F2
	.dc.b	$3F,$EC,$BF,$C8,$3F,$CE,$13,$98,$C5,$FF,$A3,$23,$18,$AB,$77,$CA
	.dc.b	$31,$5F,$C1,$FC,$1A,$FF,$9C,$9A,$FF,$1C,$F0,$BF,$89,$46,$21,$18
	.dc.b	$BF,$1F,$A7,$FC,$82,$78,$E2,$8C,$4E,$BC,$4F,$E4,$ED,$AE,$FC,$3F
	.dc.b	$92,$F8,$2F,$C7,$93,$3F,$C6,$BF,$92,$F8,$39,$1F,$90,$6F,$89,$F8
	.dc.b	$FF,$8D,$FF,$78,$7E,$E1,$68,$D2,$63,$F8,$CF,$88,$DB,$67,$6D,$9F
	.dc.b	$CD,$4E,$A3,$F9,$FE,$CF,$67,$FD,$59,$8F,$D4,$6C,$FD,$36,$BC,$1A
	.dc.b	$FC,$DF,$17,$F9,$BB,$78,$BF,$9F,$12,$BE,$FF,$E3,$AF,$89,$FE,$5F
	.dc.b	$8A,$31,$7F,$D4,$E3,$13,$A9,$CD,$7E,$CF,$C7,$6A,$FC,$87,$0D,$C7
	.dc.b	$F5,$F3,$A9,$C9,$FB,$BF,$98,$C4,$FE,$D1,$AF,$EC,$1A,$B8,$FF,$86
	.dc.b	$C6,$2F,$F5,$DF,$43,$FD,$21,$B5,$FD,$AD,$B8,$FE,$C2,$7F,$A6,$DF
	.dc.b	$F1,$6E,$3F,$4F,$E0,$BF,$C3,$5B,$87,$F0,$66,$BF,$C0,$9A,$B6,$FB
	.dc.b	$1E,$E6,$1B,$92,$FE,$D4,$E7,$8B,$FD,$D1,$1F,$DB,$F8,$75,$87,$EA
	.dc.b	$63,$04,$7F,$65,$C1,$F5,$23,$EC,$F0,$FF,$81,$9B,$3C,$8F,$C5,$CE
	.dc.b	$DF,$E3,$FF,$1F,$F5,$BF,$06,$FE,$54,$D8,$CE,$69,$CB,$63,$C3,$87
	.dc.b	$F8,$52,$7B,$92,$32,$6C,$FE,$4F,$5F,$84,$ED,$E6,$48,$EE,$FA,$08
	.dc.b	$F2,$6B,$FC,$27,$C8,$F8,$C4,$7F,$87,$3F,$CA,$BE,$9B,$FC,$FF,$8F
	.dc.b	$F2,$F6,$F8,$46,$98,$EF,$FF,$16,$FC,$BF,$F3,$26,$E1,$FC,$A9,$AF
	.dc.b	$F2,$FA,$FE,$20,$A7,$F9,$CF,$E2,$7D,$89,$1F,$E9,$EC,$3F,$A0,$47
	.dc.b	$FA,$23,$C2,$FE,$3F,$97,$A6,$BD,$87,$A8,$FE,$73,$F2,$1E,$82,$3C
	.dc.b	$F1,$47,$3E,$DF,$F1,$CD,$73,$AF,$DF,$37,$1F,$EA,$1C,$1F,$08,$93
	.dc.b	$CD,$F8,$BE,$E7,$F9,$3B,$F1,$8F,$EA,$6D,$9F,$E9,$6D,$7E,$13,$5F
	.dc.b	$EC,$1F,$83,$FC,$E2,$3E,$77,$CD,$CD,$FE,$B0,$D7,$E0,$CF,$BC,$F8
	.dc.b	$27,$CD,$F6,$3E,$B9,$1F,$80,$FD,$CF,$C0,$5F,$07,$FC,$3B,$8F,$F8
	.dc.b	$7D,$7F,$63,$9C,$EB,$FD,$14,$FE,$D2,$E2,$77,$3F,$C5,$A3,$13,$E3
	.dc.b	$8F,$92,$53,$FA,$9E,$FF,$39,$FE,$DB,$F0,$BF,$ED,$89,$B0,$E0,$EB
	.dc.b	$1F,$DF,$E7,$53,$F8,$96,$BE,$1C,$7F,$93,$F8,$9C,$31,$F4,$B6,$33
	.dc.b	$7E,$A6,$D7,$C8,$3C,$E5,$3C,$FA,$FE,$1F,$D6,$FF,$44,$DF,$93,$93
	.dc.b	$FC,$86,$D8,$CE,$1F,$E0,$86,$DF,$F8,$37,$F1,$9F,$9B,$A7,$B1,$B5
	.dc.b	$FE,$14,$6C,$EE,$3D,$A3,$37,$C9,$7F,$9D,$63,$E9,$B6,$7E,$C3,$5F
	.dc.b	$A8,$D7,$D7,$7F,$9D,$69,$F5,$E5,$7E,$28,$FE,$A2,$D8,$C9,$E4,$75
	.dc.b	$1C,$8F,$91,$C3,$F9,$46,$EF,$F7,$DB,$AF,$E6,$63,$11,$FE,$B4,$F8
	.dc.b	$57,$96,$F8,$6F,$F5,$BF,$CD,$FE,$0E,$B5,$39,$51,$95,$5C,$AB,$8E
	.dc.b	$9F,$EC,$09,$CD,$F1,$BE,$17,$B9,$E7,$5B,$8C,$4E,$AB,$84,$E5,$B5
	.dc.b	$FE,$03,$3F,$A8,$10,$FF,$1C,$14,$FF,$1C,$14,$E0,$7E,$07,$4F,$F2
	.dc.b	$AB,$F8,$13,$FC,$A8,$53,$F1,$0E,$9E,$27,$43,$B9,$0F,$C1,$BC,$9F
	.dc.b	$C1,$B4,$FC,$66,$BF,$C0,$1F,$C5,$99,$D7,$F6,$33,$E1,$8F,$33,$97
	.dc.b	$C4,$7F,$9C,$F7,$FF,$35,$3F,$A1,$8F,$F0,$CB,$4F,$D3,$38,$3D,$04
	.dc.b	$3D,$64,$7F,$41,$C5,$C9,$FD,$C1,$C6,$3F,$80,$34,$D9,$78,$9C,$7E
	.dc.b	$66,$CF,$C1,$6B,$FB,$B8,$F9,$8F,$F0,$03,$EB,$14,$F9,$87,$83,$83
	.dc.b	$F8,$65,$7F,$C0,$03,$CA,$E0,$FE,$0D,$CB,$F9,$19,$FF,$45,$B5,$FE
	.dc.b	$6C,$59,$FF,$4E,$46,$63,$E1,$9E,$13,$E3,$14,$FE,$8E,$71,$B7,$37
	.dc.b	$1A,$FF,$53,$36,$33,$3B,$E7,$2F,$C5,$7F,$AB,$B4,$FE,$AE,$53,$FA
	.dc.b	$B9,$4F,$F5,$B7,$57,$52,$3E,$37,$D4,$F9,$C9,$CD,$FE,$96,$7F,$45
	.dc.b	$8F,$FA,$EF,$AF,$CC,$48,$F6,$3C,$A7,$84,$7D,$23,$9F,$3B,$5F,$85
	.dc.b	$D9,$EF,$24,$E0,$F8,$DD,$7F,$03,$E5,$FF,$C2,$D0,$69,$1F,$F6,$0F
	.dc.b	$7F,$C9,$23,$F3,$DF,$26,$1E,$18,$FC,$3E,$3E,$67,$F3,$9E,$3F,$F9
	.dc.b	$F9,$7E,$0B,$FE,$4C,$23,$FE,$3E,$7F,$C8,$A4,$7F,$C9,$2D,$7F,$CA
	.dc.b	$3B,$7F,$A7,$1F,$F2,$2C,$FF,$95,$88,$FF,$94,$1F,$F2,$79,$1C,$1F
	.dc.b	$E4,$CF,$BF,$FF,$2E,$91,$C1,$FE,$CF,$7F,$CD,$44,$70,$7F,$A9,$FC
	.dc.b	$9F,$F4,$AC,$7F,$9C,$B8,$0D,$24,$FF,$9F,$E3,$2D,$8F,$F4,$24,$62
	.dc.b	$3F,$D0,$4D,$7F,$D4,$3C,$6F,$33,$F2,$FE,$07,$FA,$3E,$7F,$A1,$1F
	.dc.b	$E4,$B8,$CB,$FE,$F8,$6B,$FE,$9F,$AF,$F5,$93,$FE,$A3,$6B,$9F,$23
	.dc.b	$FD,$51,$18,$8F,$F5,$27,$B2,$F1,$3F,$D6,$75,$FE,$B2,$24,$FF,$AC
	.dc.b	$AB,$86,$9F,$7F,$1F,$F4,$49,$27,$FC,$F6,$FF,$A5,$48,$FF,$BF,$5A
	.dc.b	$FF,$B4,$74,$FF,$60,$C7,$CD,$24,$7D,$1A,$F8,$2F,$9B,$D1,$FF,$63
	.dc.b	$FC,$A7,$FB,$8A,$4D,$24,$FF,$63,$F3,$DA,$7F,$70,$AD,$CE,$63,$15
	.dc.b	$FE,$F5,$24,$7D,$BC,$3E,$C9,$1C,$1C,$7D,$4F,$4B,$6F,$CF,$87,$82
	.dc.b	$73,$3F,$C4,$A3,$13,$EC,$CD,$CE,$F5,$BF,$8B,$FE,$3B,$FC,$AE,$73
	.dc.b	$18,$C3,$FB,$F4,$61,$4E,$EB,$94,$60,$FE,$69,$CC,$E7,$FE,$15,$5F
	.dc.b	$EC,$92,$4E,$A3,$2F,$6F,$35,$EC,$7F,$BD,$AB,$C9,$18,$CF,$B5,$7F
	.dc.b	$2E,$C9,$39,$AD,$DF,$1A,$FF,$0B,$5B,$99,$EC,$FF,$36,$F9,$5F,$E9
	.dc.b	$53,$98,$C5,$70,$AD,$AA,$DC,$61,$BE,$DF,$09,$FF,$59,$35,$FD,$7B
	.dc.b	$6F,$C4,$E7,$83,$F0,$FE,$EA,$5F,$B7,$FF,$5D,$9C,$C6,$2F,$F1,$55
	.dc.b	$BA,$D7,$32,$31,$3F,$8F,$F8,$0A,$FE,$0E,$51,$88,$FD,$9B,$70,$FC
	.dc.b	$CB,$65,$F0,$3F,$8C,$AF,$DC,$B6,$BF,$3E,$D9,$DE,$3F,$AA,$A7,$C2
	.dc.b	$FF,$92,$DF,$EF,$0D,$7F,$C9,$4D,$7F,$B4,$B7,$89,$F0,$8F,$84,$FE
	.dc.b	$F5,$18,$6C,$66,$BF,$C0,$53,$97,$FC,$01,$F0,$AF,$43,$FB,$4B,$63
	.dc.b	$31,$83,$FB,$6E,$1E,$38,$F4,$FA,$98,$23,$FE,$0B,$F8,$6F,$E3,$59
	.dc.b	$DF,$0F,$FD,$73,$3F,$E6,$59,$CB,$FE,$15,$6B,$FE,$19,$EA,$7C,$3F
	.dc.b	$F7,$C6,$C7,$F8,$5D,$AF,$F8,$75,$AF,$CA,$C3,$C6,$FF,$88,$71,$F4
	.dc.b	$1E,$9F,$88,$7A,$5E,$49,$FF,$0A,$C6,$33,$FF,$12,$36,$3F,$C5,$4D
	.dc.b	$7F,$C5,$DF,$12,$F8,$AF,$E8,$31,$FE,$1E,$8C,$3F,$7D,$BE,$27,$F8
	.dc.b	$CD,$B1,$E7,$3F,$09,$8F,$FA,$46,$7D,$19,$74,$9F,$9E,$7E,$27,$C1
	.dc.b	$FE,$15,$E8,$5D,$7F,$C2,$97,$FC,$F2,$BF,$C5,$F5,$A8,$FF,$21,$7C
	.dc.b	$5F,$F1,$DD,$7F,$24,$D8,$CF,$8B,$F9,$BB,$4E,$5F,$F2,$67,$C6,$BE
	.dc.b	$37,$4D,$C3,$F3,$8D,$C3,$FC,$A0,$DA,$FF,$2A,$7C,$73,$D7,$EC,$DB
	.dc.b	$FD,$23,$6F,$FC,$B0,$DA,$FF,$2E,$7C,$7F,$ED,$9F,$F2,$C3,$5F,$F2
	.dc.b	$F3,$5D,$F8,$1F,$E4,$48,$C4,$7F,$AD,$5A,$FF,$95,$1A,$9A,$48,$E4
	.dc.b	$48,$F0,$B5,$F0,$B5,$F0,$9A,$21,$A2,$9A,$30,$43,$04,$3D,$0F,$56
	.dc.b	$9C,$8A,$71,$29,$C4,$F7,$3C,$5A,$71,$29,$C4,$A7,$13,$E2,$3C,$5A
	.dc.b	$71,$29,$C4,$A7,$18,$F9,$4E,$DA,$ED,$AE,$E3,$B4,$62,$3D,$64,$3C
	.dc.b	$A5,$3B,$95,$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$3F,$D1,$BF,$01,$FE
	.dc.b	$F9,$9F,$F5,$73,$5F,$F7,$B5,$FF,$AA,$E3,$11,$F8,$DF,$88,$FF,$50
	.dc.b	$F1,$7E,$59,$B1,$F9,$C6,$BF,$95,$6B,$C4,$FE,$89,$82,$33,$1F,$97
	.dc.b	$6C,$66,$37,$18,$23,$82,$4F,$8D,$A7,$8C,$A7,$8C,$A7,$8D,$E0,$47
	.dc.b	$83,$5E,$0D,$78,$1A,$21,$A2,$9A,$30,$43,$04,$3E,$01,$E4,$2B,$DD
	.dc.b	$A7,$42,$9C,$CA,$73,$9F,$91,$1C,$5B,$1C,$5A,$F1,$6B,$C6,$7E,$6C
	.dc.b	$F1,$6C,$71,$6B,$C5,$AF,$13,$D5,$3D,$1F,$29,$1C,$12,$38,$C6,$5B
	.dc.b	$19,$8E,$71,$83,$F0,$FF,$4E,$F8,$BF,$F1,$67,$C5,$FF,$BD,$2A,$4E
	.dc.b	$75,$4E,$67,$FD,$F3,$39,$8C,$42,$30,$EF,$E4,$5F,$01,$DA,$30,$D7
	.dc.b	$C1,$1A,$BE,$F1,$A6,$BE,$5E,$5F,$EB,$F8,$78,$1B,$1F,$E6,$7E,$7D
	.dc.b	$BF,$D8,$39,$FF,$DE,$D5,$E6,$FA,$4F,$EA,$DF,$19,$79,$FF,$C9,$8D
	.dc.b	$C8,$D3,$DB,$9F,$95,$BA,$70,$CB,$CD,$18,$AE,$31,$FD,$EE,$3F,$BC
	.dc.b	$B8,$3F,$C7,$7C,$27,$07,$F8,$F1,$C4,$7C,$CF,$15,$F5,$7F,$E6,$19
	.dc.b	$C3,$6B,$EA,$4E,$AF,$FC,$90,$DA,$F2,$12,$7E,$63,$5F,$C0,$6A,$FA
	.dc.b	$EF,$E1,$6D,$8F,$F2,$93,$67,$FC,$5D,$3F,$C5,$2F,$FD,$7F,$E2,$78
	.dc.b	$EF,$D9,$18,$3F,$CB,$85,$AF,$F4,$5F,$AF,$FE,$44,$7C,$53,$F0,$F5
	.dc.b	$FC,$14,$62,$11,$87,$FD,$2F,$F0,$DF,$EA,$97,$AB,$67,$FD,$2E,$D8
	.dc.b	$DE,$5F,$CA,$B8,$9C,$7F,$E4,$AF,$BF,$FC,$E4,$63,$C6,$FA,$91,$89
	.dc.b	$D5,$FE,$96,$B4,$A7,$CD,$97,$96,$3F,$37,$C5,$F0,$DF,$63,$7C,$A7
	.dc.b	$E1,$FF,$3F,$7F,$F4,$04,$63,$87,$53,$9D,$FF,$DA,$0D,$C3,$FD,$CD
	.dc.b	$F5,$9F,$EA,$3F,$DC,$3F,$D3,$AD,$8C,$E2,$9C,$B6,$3F,$DD,$DB,$ED
	.dc.b	$FF,$50,$35,$7A,$3F,$EF,$68,$C3,$67,$94,$FF,$A9,$FE,$9F,$FD,$69
	.dc.b	$F8,$EF,$FA,$B9,$B1,$FE,$FB,$6B,$B9,$4F,$D6,$E8,$FD,$13,$67,$27
	.dc.b	$DE,$8C,$B4,$C9,$FE,$DD,$FC,$1D,$F1,$3F,$EC,$5E,$7F,$F7,$B4,$63
	.dc.b	$04,$62,$75,$5F,$EF,$8A,$D4,$FF,$0B,$E7,$73,$70,$AF,$E1,$93,$98
	.dc.b	$C4,$7F,$13,$9D,$CF,$F2,$5E,$37,$E6,$3F,$CB,$6B,$62,$33,$E0,$FF
	.dc.b	$01,$AE,$15,$B6,$C6,$6B,$F3,$FC,$9F,$94,$69,$F9,$12,$F4,$7E,$82
	.dc.b	$35,$1F,$B2,$A7,$D0,$E2,$3E,$33,$76,$F8,$AD,$9D,$CF,$53,$FD,$C9
	.dc.b	$E7,$FF,$BC,$7F,$29,$FF,$75,$B6,$3F,$A4,$72,$3D,$7F,$F5,$16,$2B
	.dc.b	$DB,$6B,$73,$FA,$19,$FF,$7A,$F5,$73,$9F,$F7,$93,$5F,$C9,$CE,$DF
	.dc.b	$EA,$DB,$BE,$19,$EE,$31,$F4,$57,$CF,$9C,$C6,$2B,$85,$6F,$7F,$B1
	.dc.b	$5A,$7E,$F1,$2F,$FA,$63,$5F,$E7,$14,$FC,$41,$F1,$5F,$36,$30,$D8
	.dc.b	$CC,$F8,$5B,$5F,$A7,$C5,$F2,$D4,$95,$FA,$98,$CB,$FA,$18,$FD,$31
	.dc.b	$1C,$1F,$84,$FC,$1B,$F0,$49,$A4,$7E,$8F,$E1,$FF,$05,$B5,$F2,$38
	.dc.b	$3E,$D4,$E6,$7F,$AB,$46,$27,$E5,$69,$F3,$2F,$E4,$EB,$F4,$5B,$3E
	.dc.b	$1B,$FA,$F5,$B7,$FB,$B4,$FE,$AC,$8E,$0F,$BB,$9F,$C8,$8F,$D6,$36
	.dc.b	$33,$A7,$F6,$99,$CB,$63,$F0,$BE,$5F,$FB,$2F,$F0,$7F,$DA,$36,$33
	.dc.b	$EA,$7E,$1E,$72,$D8,$EA,$23,$AB,$4C,$B8,$38,$E4,$FC,$1E,$92,$7B
	.dc.b	$C6,$5E,$B1,$D0,$8E,$0F,$F0,$D7,$47,$F0,$F2,$38,$3F,$53,$F0,$FF
	.dc.b	$D3,$77,$FD,$83,$79,$0D,$35,$F3,$5F,$DD,$3F,$06,$FF,$1D,$52,$31
	.dc.b	$7F,$BA,$6B,$F8,$06,$BF,$CB,$67,$FC,$01,$D4,$69,$21,$8F,$A1,$FE
	.dc.b	$4D,$DF,$87,$D4,$62,$3F,$BB,$FD,$37,$42,$47,$F7,$C7,$FD,$3B,$E6
	.dc.b	$3C,$1F,$E6,$B5,$D6,$BF,$C7,$D7,$D5,$E8,$7A,$F9,$FE,$09,$38,$35
	.dc.b	$1F,$D7,$B4,$FE,$0B,$7F,$11,$7C,$57,$F0,$BF,$AB,$C3,$6B,$E3,$35
	.dc.b	$F8,$CD,$7A,$7D,$87,$D0,$26,$1E,$0E,$0F,$A1,$1F,$E0,$59,$FF,$09
	.dc.b	$E9,$EC,$7E,$29,$EC,$CB,$DD,$5F,$5F,$EC,$BE,$D9,$30,$E9,$9A,$3F
	.dc.b	$C2,$11,$FC,$CF,$F2,$1F,$E4,$9F,$0F,$FC,$97,$1F,$F1,$AF,$2A,$73
	.dc.b	$5F,$E8,$13,$80,$AE,$15,$FC,$B6,$B8,$4E,$4F,$3C,$7A,$B8,$3E,$6B
	.dc.b	$73,$F4,$E4,$F9,$4F,$F2,$D5,$FE,$3D,$B5,$F8,$F6,$BF,$55,$B1,$E4
	.dc.b	$E3,$F1,$B4,$F4,$63,$E4,$8C,$B6,$3C,$BC,$7F,$89,$CB,$CA,$DC,$BF
	.dc.b	$15,$2F,$C4,$DF,$29,$FE,$75,$C7,$BC,$63,$34,$62,$75,$5F,$26,$B5
	.dc.b	$3C,$3D,$DF,$F7,$FC,$E5,$B7,$FE,$66,$AF,$F6,$7C,$E6,$3F,$52,$FB
	.dc.b	$3E,$B3,$E0,$E8,$69,$23,$73,$FA,$02,$1F,$89,$3F,$16,$FB,$6F,$C3
	.dc.b	$1F,$D9,$EF,$C2,$F8,$E3,$33,$B8,$EA,$73,$3A,$9F,$74,$F9,$BC,$7F
	.dc.b	$99,$73,$A7,$CD,$7F,$0B,$F0,$0F,$A1,$7C,$17,$FA,$87,$E8,$BF,$71
	.dc.b	$39,$8C,$70,$7F,$B4,$E3,$16,$AD,$DF,$28,$C5,$6E,$31,$3A,$BE,$75
	.dc.b	$FE,$AF,$BE,$73,$98,$E2,$7F,$0A,$3F,$C7,$C4,$3F,$D9,$85,$3F,$8D
	.dc.b	$94,$E0,$7D,$93,$F0,$07,$FA,$88,$F9,$85,$3E,$86,$7F,$AF,$79,$9D
	.dc.b	$88,$60,$FD,$BE,$BF,$B9,$D3,$F0,$87,$33,$FC,$60,$E6,$37,$19,$D5
	.dc.b	$5D,$E7,$C2,$65,$C4,$FD,$F8,$FC,$E0,$F9,$CA,$48,$FC,$F7,$0F,$FB
	.dc.b	$66,$3F,$82,$CF,$F0,$13,$E2,$14,$F5,$9F,$C2,$79,$BF,$C6,$D1,$FB
	.dc.b	$66,$D7,$EE,$9B,$3F,$C5,$9A,$FF,$03,$8F,$95,$D1,$FD,$67,$97,$E0
	.dc.b	$4E,$4F,$E2,$F1,$99,$FA,$11,$E3,$3E,$61,$F6,$0A,$FF,$20,$F9,$9F
	.dc.b	$22,$69,$1F,$EC,$DD,$3F,$02,$7F,$A5,$36,$BF,$9B,$4F,$FB,$F6,$76
	.dc.b	$D7,$C2,$78,$CF,$11,$23,$FA,$2C,$6E,$3F,$A5,$B6,$33,$1F,$D5,$FA
	.dc.b	$BC,$1D,$3F,$D6,$2B,$F8,$B4,$FF,$54,$6B,$FD,$7F,$AF,$FC,$CD,$AF
	.dc.b	$F5,$C9,$CE,$1F,$E0,$3E,$DE,$64,$8F,$06,$1F,$E6,$AC,$3E,$EF,$3F
	.dc.b	$F2,$B6,$F4,$7F,$AA,$1F,$EF,$67,$A6,$BF,$D6,$3D,$5E,$B8,$FF,$5B
	.dc.b	$36,$33,$19,$AE,$8F,$E2,$0F,$8C,$7C,$78,$FF,$5D,$F7,$71,$EF,$FF
	.dc.b	$09,$4F,$76,$BF,$3E,$76,$FA,$8F,$F4,$37,$37,$F7,$F6,$9F,$E2,$12
	.dc.b	$9F,$E2,$12,$9F,$E9,$2F,$03,$EF,$76,$FF,$8B,$AB,$EA,$36,$7F,$C5
	.dc.b	$AD,$7D,$58,$FB,$1F,$C0,$B5,$FF,$62,$78,$FF,$E8,$6F,$C1,$7F,$C9
	.dc.b	$E4,$8F,$F2,$9B,$FE,$4E,$23,$FE,$4C,$6B,$FE,$94,$8F,$F4,$7F,$E0
	.dc.b	$BF,$E5,$82,$47,$F9,$71,$FF,$2B,$91,$FF,$2A,$B5,$E5,$E4,$17,$E0
	.dc.b	$BF,$E6,$22,$3F,$E6,$67,$FD,$C8,$47,$F3,$D8,$7F,$6F,$8C,$47,$FA
	.dc.b	$A3,$CB,$FF,$59,$C7,$FA,$18,$91,$FE,$86,$E3,$FF,$5D,$47,$CA,$F3
	.dc.b	$3F,$D7,$9F,$80,$5D,$8D,$25,$FF,$B9,$2B,$FD,$42,$47,$07,$F8,$77
	.dc.b	$85,$EC,$FB,$7E,$8F,$F0,$D1,$89,$46,$27,$51,$C5,$AF,$FA,$93,$85
	.dc.b	$EE,$7F,$B4,$6B,$FD,$78,$49,$FF,$5D,$3F,$E8,$AC,$64,$7F,$A3,$9A
	.dc.b	$FF,$99,$9F,$E0,$1B,$7F,$85,$DB,$CE,$F2,$7E,$0D,$7C,$E8,$C4,$FF
	.dc.b	$AC,$09,$1F,$2F,$5D,$BF,$47,$7F,$96,$FF,$BC,$BD,$4F,$F7,$47,$E5
	.dc.b	$3F,$EF,$2A,$34,$93,$FE,$F8,$F5,$F6,$D9,$DB,$67,$FD,$FE,$7C,$CF
	.dc.b	$63,$F1,$58,$D9,$F1,$9F,$EF,$7E,$14,$EA,$3E,$B3,$5F,$C9,$5A,$7F
	.dc.b	$D1,$55,$FE,$DE,$24,$FD,$DF,$37,$82,$F5,$DE,$D3,$DA,$FC,$E7,$DA
	.dc.b	$FF,$33,$9C,$C6,$2B,$F9,$C4,$61,$4E,$EB,$94,$60,$FA,$5C,$35,$FF
	.dc.b	$4F,$CF,$D8,$8C,$47,$FA,$5A,$3C,$1E,$E3,$DD,$6B,$AC,$53,$98,$C3
	.dc.b	$6B,$8D,$70,$7F,$2D,$6F,$C7,$29,$1F,$CC,$19,$3F,$AA,$78,$CE,$9F
	.dc.b	$CC,$E3,$99,$FC,$E7,$4F,$92,$FF,$8D,$B6,$38,$D7,$09,$FE,$3F,$F0
	.dc.b	$1E,$AF,$DC,$6E,$06,$92,$F8,$FC,$1E,$DB,$3B,$6C,$FF,$2A,$9D,$47
	.dc.b	$F1,$3D,$9F,$08,$E9,$FE,$E5,$18,$6C,$67,$0F,$ED,$53,$97,$FA,$97
	.dc.b	$C3,$3E,$1B,$FB,$D6,$08,$C4,$EA,$73,$18,$9F,$EF,$5C,$6E,$E6,$BF
	.dc.b	$90,$8C,$36,$33,$3F,$A0,$9C,$BF,$A2,$C3,$F3,$9D,$DF,$D6,$C9,$1F
	.dc.b	$D6,$DF,$ED,$0F,$97,$1F,$19,$27,$CC,$7F,$6F,$F8,$77,$C4,$7F,$85
	.dc.b	$61,$18,$75,$19,$8C,$57,$F8,$3E,$FE,$63,$FD,$DD,$AE,$76,$F1,$C6
	.dc.b	$23,$FC,$39,$A7,$C5,$6E,$3E,$57,$FC,$35,$19,$7F,$C4,$9F,$13,$FC
	.dc.b	$CB,$2F,$F0,$23,$63,$FC,$40,$D7,$FC,$50,$D7,$FC,$59,$F2,$8E,$CF
	.dc.b	$EA,$DF,$49,$FE,$27,$8C,$4A,$31,$3A,$8F,$F1,$94,$EA,$7C,$B8,$FE
	.dc.b	$C9,$B8,$71,$AF,$5B,$5F,$F1,$E6,$3F,$86,$9F,$F1,$B7,$C7,$3E,$3B
	.dc.b	$95,$7F,$8F,$E3,$17,$C6,$3F,$B6,$46,$2B,$8D,$70,$9F,$F2,$47,$0F
	.dc.b	$FB,$5F,$E0,$3F,$85,$12,$3F,$87,$BE,$62,$3E,$66,$BF,$E5,$1F,$8F
	.dc.b	$FE,$41,$8F,$FA,$9E,$31,$E0,$FF,$10,$AF,$F2,$75,$69,$B1,$99,$FF
	.dc.b	$2C,$7C,$87,$FA,$29,$FF,$1D,$35,$FF,$2B,$35,$DF,$BF,$F9,$88,$C4
	.dc.b	$7F,$98,$DA,$FF,$83,$5A,$9A,$48,$E4,$48,$F0,$B5,$F0,$B5,$F0,$9A
	.dc.b	$21,$A2,$9A,$30,$43,$04,$3D,$0F,$56,$9C,$8A,$71,$29,$C4,$F7,$3C
	.dc.b	$5A,$71,$29,$C4,$A7,$13,$E2,$3C,$5A,$71,$29,$C4,$A7,$18,$F9,$4E
	.dc.b	$DA,$ED,$AE,$E3,$B4,$62,$3D,$64,$3C,$A5,$3B,$95,$E0,$E8,$86,$8A
	.dc.b	$E9,$C1,$1C,$12,$3F,$90,$FC,$13,$E0,$FF,$35,$5F,$EF,$FC,$3F,$D1
	.dc.b	$93,$96,$CE,$A7,$F1,$DE,$AF,$F9,$6F,$8F,$F2,$ED,$8F,$CE,$B5,$FC
	.dc.b	$B3,$5E,$27,$F0,$C3,$25,$33,$1D,$1C,$1F,$9B,$8E,$4D,$33,$8A,$32
	.dc.b	$D8,$CC,$60,$8E,$09,$3D,$C8,$FA,$8C,$95,$CB,$E8,$6B,$E8,$34,$43
	.dc.b	$45,$34,$60,$86,$08,$7B,$5E,$AD,$39,$14,$E2,$53,$8C,$FC,$18,$E2
	.dc.b	$D8,$E2,$D7,$8B,$5E,$33,$F1,$A7,$8B,$63,$8B,$5E,$2D,$78,$9F,$20
	.dc.b	$C9,$4C,$C7,$47,$07,$A4,$93,$E3,$6C,$F5,$69,$C2,$34,$6F,$87,$FC
	.dc.b	$1F,$F1,$3F,$D4,$09,$3F,$D4,$1F,$F2,$51,$1F,$F2,$7B,$5F,$EC,$77
	.dc.b	$FE,$B7,$8F,$EB,$F0,$69,$1F,$F5,$E1,$FE,$BD,$D7,$FA,$2F,$C8,$7F
	.dc.b	$40,$DC,$D2,$4F,$F9,$79,$FF,$2E,$11,$C1,$D0,$7C,$B3,$CE,$FF,$65
	.dc.b	$4E,$67,$31,$87,$84,$ED,$F0,$98,$8D,$3F,$E0,$8F,$6F,$FC,$CF,$97
	.dc.b	$F8,$19,$B1,$FE,$06,$6B,$BC,$3F,$67,$C6,$EA,$FF,$55,$5F,$F9,$D4
	.dc.b	$95,$FE,$7A,$7C,$DF,$34,$F6,$FF,$72,$D9,$FD,$EB,$67,$73,$FB,$EE
	.dc.b	$C7,$E2,$1E,$2B,$FF,$75,$46,$32,$FE,$0A,$DB,$FF,$0F,$D6,$EB,$FC
	.dc.b	$99,$8F,$99,$F2,$10,$FF,$02,$B9,$7E,$DF,$8D,$FE,$64,$7F,$88,$B6
	.dc.b	$BE,$83,$63,$EB,$35,$FC,$1F,$D7,$BE,$1F,$C4,$DC,$BF,$8D,$B6,$FE
	.dc.b	$D3,$6B,$C8,$0F,$4F,$89,$F4,$31,$F3,$46,$76,$FA,$53,$FC,$9B,$ED
	.dc.b	$5F,$85,$FF,$26,$6D,$66,$30,$7F,$33,$C9,$F1,$5F,$EB,$2F,$96,$FC
	.dc.b	$84,$63,$C0,$F1,$46,$27,$57,$FA,$0A,$D2,$9D,$7D,$87,$08,$D5,$23
	.dc.b	$11,$FE,$D2,$8C,$3C,$4D,$C6,$97,$A2,$F8,$57,$D4,$FF,$AF,$BE,$FF
	.dc.b	$F4,$D3,$98,$C5,$7F,$AD,$2B,$79,$2F,$71,$87,$DF,$E9,$BE,$53,$8D
	.dc.b	$FF,$B9,$A3,$17,$C6,$3E,$04,$62,$B8,$D7,$09,$FF,$68,$F9,$DF,$B7
	.dc.b	$BF,$9B,$1F,$1E,$31,$A7,$CE,$6C,$67,$B7,$FD,$B3,$3F,$BD,$FC,$1F
	.dc.b	$FD,$F7,$7C,$AB,$F8,$7C,$62,$75,$EE,$F0,$9F,$E1,$13,$97,$C1,$EB
	.dc.b	$E5,$FF,$80,$38,$FF,$B5,$AF,$C5,$39,$9F,$4B,$57,$47,$FB,$A7,$A7
	.dc.b	$F2,$FE,$CF,$E7,$09,$9F,$E3,$6A,$4F,$F1,$5E,$7F,$C8,$B7,$9F,$F2
	.dc.b	$19,$3E,$0F,$F3,$DF,$33,$F8,$D6,$E1,$F8,$86,$BF,$87,$6B,$F1,$BC
	.dc.b	$AB,$97,$E6,$9B,$7F,$E0,$1B,$FC,$C4,$70,$AF,$5F,$77,$ED,$09,$7F
	.dc.b	$B4,$7F,$84,$76,$FE,$B3,$3F,$E3,$75,$FC,$1B,$B7,$F6,$47,$B9,$FE
	.dc.b	$6D,$AF,$B9,$BA,$FE,$B9,$FF,$0E,$62,$F6,$7F,$A3,$FE,$37,$FA,$E6
	.dc.b	$CF,$EA,$DA,$FF,$01,$9E,$0F,$F0,$39,$E1,$F0,$56,$38,$35,$E1,$48
	.dc.b	$C4,$2B,$C5,$18,$8F,$BB,$C3,$F6,$B3,$FB,$B5,$FC,$5A,$3F,$95,$65
	.dc.b	$ED,$AF,$EB,$ED,$7F,$8B,$35,$37,$8B,$F8,$E4,$7F,$24,$DB,$E7,$E1
	.dc.b	$FC,$81,$F9,$FE,$0F,$F1,$C9,$FE,$3F,$3F,$41,$FE,$3D,$19,$BE,$C4
	.dc.b	$D7,$E7,$57,$A0,$95,$FD,$A7,$0F,$CC,$37,$1F,$5D,$70,$8F,$AE,$7D
	.dc.b	$B3,$F9,$CC,$7C,$37,$86,$5F,$0C,$D3,$F0,$27,$93,$F1,$1F,$F2,$AE
	.dc.b	$FF,$89,$8F,$E7,$67,$B0,$FE,$7B,$3F,$83,$ED,$FE,$65,$B7,$F1,$C9
	.dc.b	$FC,$3B,$72,$FC,$34,$EB,$E8,$3F,$A2,$7C,$67,$E2,$5B,$7F,$89,$6B
	.dc.b	$F8,$26,$BF,$81,$FD,$0B,$FC,$09,$7F,$CF,$9B,$1F,$CE,$DA,$FE,$45
	.dc.b	$AF,$D2,$E3,$FD,$0C,$FD,$58,$F6,$1F,$90,$C3,$D6,$47,$D6,$FC,$4E
	.dc.b	$0F,$BD,$A7,$C5,$9F,$D6,$57,$DF,$D3,$F4,$DF,$29,$F1,$09,$3F,$11
	.dc.b	$F8,$FE,$4F,$F6,$5F,$D8,$7F,$54,$D9,$FD,$53,$5F,$D1,$B5,$FD,$16
	.dc.b	$FF,$DD,$5F,$9F,$C9,$F4,$88,$FB,$8F,$B3,$97,$9D,$F8,$59,$7E,$E7
	.dc.b	$EB,$BE,$C9,$23,$EC,$BF,$DB,$7F,$0C,$6F,$FE,$2E,$D7,$FB,$44,$61
	.dc.b	$5F,$ED,$6F,$F8,$04,$E6,$31,$1F,$B9,$8C,$4F,$93,$6F,$96,$4B,$FF
	.dc.b	$55,$7E,$30,$FC,$A1,$F5,$7F,$EB,$7B,$FF,$2B,$CE,$63,$18,$FF,$83
	.dc.b	$E3,$0F,$92,$77,$5C,$A3,$07,$9A,$FD,$8D,$9C,$BE,$78,$CB,$63,$2F
	.dc.b	$E2,$0F,$96,$7A,$E3,$F4,$3F,$23,$E4,$8F,$AE,$D9,$FE,$54,$DD,$CE
	.dc.b	$7F,$29,$7C,$85,$25,$F9,$2D,$3A,$8C,$6A,$F4,$BC,$F4,$8C,$3A,$8C
	.dc.b	$C6,$37,$FA,$B5,$EB,$27,$0F,$E1,$DF,$75,$F8,$31,$F1,$D7,$DC,$8C
	.dc.b	$4F,$97,$1F,$24,$F4,$BE,$44,$E0,$FB,$D1,$FB,$18,$C6,$BF,$97,$BF
	.dc.b	$F7,$D0,$F9,$CF,$F7,$77,$BD,$FE,$FB,$8C,$5E,$A7,$3E,$EB,$DE,$FF
	.dc.b	$28,$46,$1B,$3B,$25,$7F,$95,$6F,$A5,$7F,$BE,$F1,$FF,$29,$CF,$FA
	.dc.b	$55,$AF,$FA,$6D,$AF,$F6,$B3,$59,$7E,$99,$B9,$7F,$A7,$5F,$F3,$09
	.dc.b	$83,$FC,$AB,$9F,$D9,$7E,$53,$74,$FF,$2A,$36,$3F,$B5,$35,$FE,$34
	.dc.b	$73,$3C,$95,$E1,$6D,$7F,$AB,$6B,$F0,$5B,$FE,$D5,$FD,$CB,$67,$F6
	.dc.b	$ED,$7F,$6C,$D8,$FB,$B1,$E6,$DB,$FC,$D0,$F9,$1A,$FF,$75,$C3,$CD
	.dc.b	$3F,$9E,$8F,$F0,$11,$FD,$F3,$83,$71,$98,$FC,$57,$0F,$E8,$9D,$9E
	.dc.b	$E3,$D6,$7B,$CF,$A0,$7F,$12,$C7,$CD,$C3,$FC,$4E,$3F,$28,$DA,$D1
	.dc.b	$D3,$8F,$F8,$2C,$FF,$14,$6C,$7F,$06,$E4,$F5,$C6,$27,$B9,$0F,$21
	.dc.b	$67,$F5,$D7,$E2,$23,$EA,$32,$5C,$1C,$BF,$85,$8F,$F1,$63,$63,$35
	.dc.b	$F6,$5A,$FC,$13,$C2,$43,$F1,$A7,$F8,$C7,$9B,$91,$1E,$4F,$5E,$7F
	.dc.b	$F5,$A7,$E3,$1F,$D4,$24,$D2,$3F,$E3,$BE,$AF,$21,$1F,$23,$F9,$D7
	.dc.b	$C2,$7C,$52,$9E,$6C,$7F,$C0,$CF,$AF,$AF,$DC,$47,$F4,$DC,$9F,$AC
	.dc.b	$7D,$8D,$79,$CB,$B1,$E6,$FC,$96,$CF,$E7,$DA,$FF,$81,$63,$E6,$77
	.dc.b	$7F,$B4,$F3,$FF,$08,$4F,$FB,$E4,$FF,$07,$14,$FE,$DB,$1B,$7F,$CB
	.dc.b	$98,$F9,$DF,$F0,$D9,$EB,$3D,$3C,$FF,$B1,$7E,$16,$5E,$93,$D5,$E1
	.dc.b	$FF,$B6,$A3,$FD,$14,$7F,$05,$70,$FF,$8E,$8F,$F2,$C9,$0F,$F1,$C9
	.dc.b	$4F,$F1,$B1,$C7,$C1,$FA,$51,$CA,$32,$6C,$FF,$79,$78,$6F,$0F,$F9
	.dc.b	$FC,$61,$AF,$F9,$AB,$6E,$8C,$3C,$E4,$3F,$66,$43,$DF,$8F,$C1,$7B
	.dc.b	$F0,$F4,$B3,$49,$5D,$67,$E0,$10,$F4,$9F,$00,$87,$F9,$71,$FE,$59
	.dc.b	$3E,$B7,$F4,$DE,$22,$5F,$8F,$7F,$A3,$9A,$7F,$A3,$8A,$7F,$A3,$A3
	.dc.b	$84,$7F,$A6,$E2,$9F,$E9,$B2,$9F,$E9,$42,$9F,$E3,$22,$9F,$E8,$8C
	.dc.b	$FE,$99,$1F,$71,$F1,$B9,$BF,$0E,$47,$FD,$15,$7E,$CC,$7E,$2C,$60
	.dc.b	$FF,$60,$8F,$92,$A4,$7E,$01,$FE,$CB,$34,$7F,$B4,$4A,$7F,$B4,$4A
	.dc.b	$7F,$BD,$BC,$BF,$91,$1F,$ED,$98,$C9,$EB,$3D,$27,$FA,$1C,$A7,$E1
	.dc.b	$CF,$C7,$14,$FA,$E7,$FB,$CA,$9F,$0F,$FE,$F6,$6C,$7F,$BF,$1A,$FF
	.dc.b	$BF,$3C,$DF,$F5,$F4,$FB,$23,$EE,$79,$EF,$43,$31,$98,$C5,$F2,$8D
	.dc.b	$CE,$9F,$8B,$3A,$AE,$94,$F4,$3F,$85,$BC,$4C,$74,$71,$BF,$F6,$84
	.dc.b	$EA,$B8,$B6,$3E,$8D,$3D,$37,$A5,$FE,$D7,$FC,$1B,$BC,$62,$51,$89
	.dc.b	$D5,$72,$AD,$4F,$C1,$F5,$9E,$B3,$AF,$77,$F9,$F9,$CC,$62,$11,$8C
	.dc.b	$7B,$C7,$E1,$A3,$17,$F0,$BD,$57,$AF,$FA,$9F,$6B,$0D,$BF,$6C,$E9
	.dc.b	$4E,$9A,$6B,$DA,$E0,$FF,$C0,$19,$46,$23,$FD,$F5,$18,$78,$9B,$8D
	.dc.b	$3F,$91,$3F,$88,$1F,$C1,$08,$7F,$46,$EC,$FE,$20,$FF,$50,$27,$67
	.dc.b	$F1,$0C,$FF,$58,$48,$FD,$61,$F8,$DC,$1F,$1B,$FD,$6D,$B1,$FE,$39
	.dc.b	$6B,$F9,$33,$E0,$C7,$F6,$1F,$03,$F9,$F6,$1E,$26,$F5,$FF,$90,$7C
	.dc.b	$E7,$B9,$FC,$B9,$B9,$66,$30,$7F,$32,$F7,$7F,$A9,$65,$FB,$C6,$BF
	.dc.b	$C7,$5A,$9B,$87,$82,$3F,$C0,$C7,$C3,$B7,$BD,$E1,$26,$1E,$57,$FA
	.dc.b	$0F,$BF,$FD,$AB,$E1,$FD,$CD,$BF,$D9,$35,$FE,$1E,$D7,$FD,$D9,$1F
	.dc.b	$E6,$02,$3F,$E5,$27,$E4,$7C,$17,$E0,$AD,$18,$E1,$46,$2B,$F8,$05
	.dc.b	$6E,$B5,$39,$9F,$E3,$FD,$7F,$ED,$07,$FD,$00,$7F,$07,$71,$EC,$7D
	.dc.b	$76,$C6,$7D,$8F,$E8,$D3,$FE,$76,$24,$7F,$9A,$1A,$FF,$2C,$F8,$47
	.dc.b	$C4,$F9,$A7,$F0,$51,$8C,$3C,$35,$BA,$D3,$FC,$66,$75,$5F,$89,$F8
	.dc.b	$67,$C0,$FF,$8A,$5B,$5F,$E8,$76,$BB,$8D,$11,$D3,$5D,$18,$21,$82
	.dc.b	$1F,$CA,$E3,$AB,$63,$93,$4E,$25,$38,$9F,$D3,$A3,$8B,$63,$8B,$4E
	.dc.b	$25,$38,$9F,$DD,$63,$8B,$63,$8B,$4E,$25,$38,$88,$E2,$D8,$E2,$D3
	.dc.b	$89,$4E,$31,$C0,$91,$C1,$B1,$C1,$AF,$08,$D1,$1D,$35,$D1,$82,$18
	.dc.b	$21,$F2,$C8,$7C,$52,$9E,$F2,$9E,$92,$C7,$91,$B1,$E1,$7F,$B3,$7B
	.dc.b	$9F,$EE,$56,$FB,$BF,$94,$7F,$DD,$66,$2F,$FD,$A9,$FB,$DF,$FA,$3B
	.dc.b	$4F,$E3,$8D,$9F,$EE,$0D,$7F,$C1,$AD,$7F,$C5,$CD,$4D,$27,$AF,$FE
	.dc.b	$33,$9E,$AD,$9E,$AD,$7A,$B5,$EA,$D7,$AB,$67,$AC,$7F,$01,$38,$94
	.dc.b	$E2,$53,$89,$5E,$33,$E4,$8C,$4F,$52,$38,$24,$6C,$8F,$A8,$C9,$5C
	.dc.b	$BE,$46,$BE,$46,$BE,$46,$CF,$92,$39,$10,$E4,$53,$91,$4E,$45,$79
	.dc.b	$3C,$08,$70,$29,$C0,$AF,$07,$44,$34,$57,$4E,$08,$E0,$93,$F2,$A7
	.dc.b	$DA,$D9,$F3,$35,$F0,$B5,$EC,$D5,$CA,$E9,$74,$BE,$1F,$FB,$A4,$E6
	.dc.b	$31,$5F,$E4,$98,$C4,$EA,$B8,$BF,$D1,$70,$FF,$44,$36,$B8,$9F,$D3
	.dc.b	$A3,$2F,$F4,$93,$27,$FA,$3F,$9B,$FD,$A3,$3F,$DA,$3C,$7D,$CF,$F6
	.dc.b	$66,$BB,$6A,$E1,$FE,$95,$F0,$FF,$D3,$38,$FF,$B1,$B7,$90,$D3,$CC
	.dc.b	$4A,$F2,$35,$FF,$41,$FE,$32,$F8,$1F,$F5,$24,$FF,$86,$63,$13,$D0
	.dc.b	$FF,$37,$38,$8D,$4F,$09,$FF,$39,$7D,$83,$E3,$BF,$BB,$ED,$FE,$23
	.dc.b	$8C,$63,$FA,$F6,$C6,$67,$75,$F2,$63,$E7,$11,$C1,$FE,$7F,$EF,$7D
	.dc.b	$1F,$F0,$16,$C7,$EE,$DB,$3F,$E3,$36,$C7,$F9,$CB,$E0,$5F,$57,$E0
	.dc.b	$8F,$F6,$26,$49,$CB,$FC,$1C,$C4,$EE,$7F,$C2,$73,$AB,$F3,$47,$E0
	.dc.b	$CF,$E1,$9D,$BF,$E1,$18,$CC,$7F,$86,$BE,$01,$F4,$BF,$C6,$F1,$46
	.dc.b	$27,$C0,$DA,$FF,$49,$B6,$3F,$D7,$7E,$03,$E6,$1F,$75,$FC,$96,$30
	.dc.b	$D8,$CC,$FF,$21,$9C,$BE,$6B,$FE,$55,$97,$9E,$3E,$79,$EB,$8F,$F4
	.dc.b	$77,$A0,$F7,$FF,$C6,$7D,$38,$6D,$7F,$A2,$67,$5A,$7E,$8B,$D0,$FF
	.dc.b	$4A,$D6,$AB,$FD,$29,$4F,$82,$FF,$1E,$FE,$AF,$D9,$B3,$CD,$AF,$06
	.dc.b	$BF,$E9,$A9,$7C,$16,$30,$FF,$51,$B6,$33,$8B,$0F,$5C,$F2,$8F,$A3
	.dc.b	$CA,$C1,$F1,$FF,$42,$30,$D8,$C9,$FE,$BC,$E2,$FC,$8C,$7F,$B2,$3F
	.dc.b	$5F,$FF,$2B,$CE,$8F,$C6,$B8,$8D,$5F,$49,$FF,$62,$CE,$5B,$E8,$78
	.dc.b	$FD,$3E,$8D,$8C,$88,$CB,$4F,$F7,$27,$CC,$FE,$2E,$3F,$D2,$8D,$8C
	.dc.b	$F3,$F3,$7E,$00,$8D,$3E,$8E,$DD,$3F,$E7,$2C,$3F,$82,$DF,$BE,$73
	.dc.b	$8A,$73,$5F,$EE,$0D,$FF,$CF,$33,$FC,$90,$FF,$6B,$9F,$83,$8F,$F3
	.dc.b	$EC,$6B,$8F,$FD,$E2,$D8,$FC,$67,$C0,$6F,$8B,$FD,$BE,$D9,$FF,$5D
	.dc.b	$57,$E5,$A3,$A5,$F8,$FC,$4F,$B9,$F2,$DF,$55,$B8,$FD,$06,$BF,$59
	.dc.b	$AF,$BF,$C6,$B5,$ED,$6E,$7F,$D3,$E5,$7F,$81,$9E,$3F,$30,$FC,$B3
	.dc.b	$FD,$05,$9F,$CE,$B4,$E7,$C6,$F8,$4D,$CB,$FA,$4B,$6B,$FA,$A7,$95
	.dc.b	$FD,$32,$FF,$AF,$E3,$FD,$59,$FE,$B4,$D7,$ED,$F9,$5F,$D7,$1F,$BB
	.dc.b	$6F,$B1,$FF,$52,$61,$E3,$8C,$77,$7F,$52,$6D,$78,$AB,$6F,$E7,$FA
	.dc.b	$BF,$4D,$EF,$7E,$8C,$91,$90,$FE,$79,$A7,$E7,$B4,$E2,$D3,$27,$E9
	.dc.b	$9C,$BF,$A8,$F9,$8F,$EC,$B1,$F8,$D6,$F0,$9A,$7F,$6A,$E5,$FC,$2C
	.dc.b	$FF,$59,$AF,$E1,$FC,$9F,$DA,$1F,$C5,$C6,$5F,$E1,$51,$97,$F8,$16
	.dc.b	$7F,$4F,$CA,$F8,$DC,$7F,$DC,$A7,$E5,$B5,$DB,$5F,$E1,$5C,$1F,$E0
	.dc.b	$F9,$FB,$24,$8F,$AF,$C5,$E1,$BF,$9F,$3F,$4F,$7F,$F6,$C3,$78,$3F
	.dc.b	$B7,$5F,$92,$72,$FD,$CE,$1F,$E8,$3A,$3F,$02,$FB,$91,$86,$F8,$DF
	.dc.b	$EE,$A6,$CE,$EB,$F0,$B7,$FE,$E7,$9F,$04,$64,$F0,$61,$FC,$96,$32
	.dc.b	$72,$FB,$C9,$3E,$3C,$3F,$DE,$F1,$88,$E2,$1E,$A4,$7C,$35,$FC,$C6
	.dc.b	$3F,$3E,$C8,$F8,$ED,$7F,$95,$F4,$FF,$69,$E3,$34,$91,$FD,$E6,$13
	.dc.b	$FC,$F7,$E5,$BF,$9B,$36,$7F,$98,$B5,$FE,$6E,$D7,$F9,$07,$C2,$F7
	.dc.b	$23,$D5,$FF,$49,$70,$FF,$93,$67,$CD,$3F,$82,$E8,$FF,$29,$C6,$36
	.dc.b	$FE,$4D,$F8,$9F,$A4,$95,$E7,$79,$D7,$08,$F8,$24,$3F,$7A,$E5,$FE
	.dc.b	$9B,$F8,$B7,$C3,$23,$F0,$5F,$F5,$47,$3F,$C6,$F5,$18,$3F,$45,$FE
	.dc.b	$FB,$18,$6C,$66,$FF,$BE,$4E,$5F,$EA,$FF,$82,$7D,$12,$3F,$3D,$F4
	.dc.b	$F4,$7F,$AB,$75,$F8,$93,$E9,$3E,$24,$7D,$5F,$C6,$7E,$D1,$23,$07
	.dc.b	$F8,$C7,$DA,$FE,$17,$8B,$1F,$ED,$B3,$EA,$9C,$C6,$25,$18,$7F,$AE
	.dc.b	$98,$EE,$F8,$33,$CA,$BF,$95,$E1,$E7,$DB,$E6,$47,$9B,$83,$EC,$71
	.dc.b	$BE,$3B,$FD,$B1,$97,$F7,$86,$D7,$F7,$E6,$BF,$D9,$DB,$1E,$13,$FC
	.dc.b	$C7,$1F,$E5,$0C,$FD,$BE,$AF,$A5,$B3,$E8,$6C,$FE,$03,$2F,$31,$31
	.dc.b	$F2,$FC,$0F,$F8,$67,$0F,$85,$92,$31,$8F,$F8,$39,$B8,$7F,$85,$1A
	.dc.b	$FE,$13,$F7,$3F,$F5,$4B,$FA,$89,$F9,$31,$8C,$3E,$73,$63,$38,$FB
	.dc.b	$B2,$7E,$0B,$FE,$82,$9F,$53,$70,$F8,$0D,$7E,$D3,$5F,$A7,$87,$EA
	.dc.b	$9A,$7E,$26,$58,$78,$63,$FC,$F2,$D7,$D5,$87,$D0,$8C,$47,$23,$E5
	.dc.b	$57,$5E,$1F,$CA,$C6,$0F,$CB,$1B,$3F,$04,$7F,$79,$F7,$FF,$1B,$EF
	.dc.b	$FF,$91,$FC,$AF,$E1,$B3,$FE,$49,$9C,$BF,$EA,$DB,$FF,$16,$C6,$27
	.dc.b	$72,$8C,$4E,$B4,$5F,$58,$C3,$6B,$B4,$E9,$EC,$62,$34,$FE,$70,$FF
	.dc.b	$04,$90,$FE,$A2,$53,$F1,$85,$3F,$6A,$7F,$85,$F3,$F9,$F9,$FE,$AA
	.dc.b	$3E,$7B,$5F,$9E,$7F,$AA,$8C,$94,$E6,$78,$08,$78,$0F,$C2,$C6,$CF
	.dc.b	$DC,$E9,$FE,$6A,$73,$1F,$A5,$8C,$3D,$8F,$C9,$CF,$8E,$7C,$F1,$CF
	.dc.b	$4F,$CD,$BF,$A4,$D7,$FD,$7F,$3F,$A3,$8F,$F2,$A9,$FA,$32,$9F,$E5
	.dc.b	$F8,$DB,$E7,$21,$EC,$2C,$7E,$F4,$FD,$D7,$1F,$F9,$81,$FD,$D3,$63
	.dc.b	$F8,$03,$5F,$E0,$C7,$C1,$DF,$F8,$73,$78,$7D,$91,$88,$FE,$31,$C0
	.dc.b	$E2,$F6,$C7,$90,$FC,$01,$F3,$8F,$E4,$D7,$FD,$03,$3F,$2E,$26,$91
	.dc.b	$FF,$49,$F2,$7F,$95,$EB,$FD,$32,$D7,$FD,$04,$D8,$FF,$40,$39,$8F
	.dc.b	$09,$E3,$C3,$CD,$5C,$5B,$1F,$EA,$7E,$5E,$59,$1A,$48,$F1,$73,$FF
	.dc.b	$59,$C5,$FA,$17,$F5,$62,$4F,$F5,$68,$CB,$FE,$AC,$E9,$EA,$47,$D9
	.dc.b	$1F,$E9,$78,$F4,$63,$E4,$9F,$90,$7A,$8F,$BC,$7F,$A8,$7A,$BC,$E4
	.dc.b	$79,$E4,$8E,$7B,$7D,$49,$E2,$73,$3A,$61,$FE,$A7,$EB,$F8,$24,$8E
	.dc.b	$6F,$C4,$E6,$FF,$6A,$7A,$5F,$E1,$89,$FF,$6F,$9F,$E1,$82,$9F,$E1
	.dc.b	$82,$9F,$EB,$7E,$EF,$9A,$47,$CE,$F9,$B8,$3E,$A6,$3E,$79,$F3,$69
	.dc.b	$F0,$DF,$F1,$CF,$E0,$BF,$E4,$42,$3F,$E3,$80,$FF,$93,$08,$FF,$90
	.dc.b	$F3,$CB,$63,$FC,$9B,$E3,$48,$E4,$D7,$FC,$E3,$E2,$FF,$97,$23,$FC
	.dc.b	$AE,$48,$E8,$FF,$95,$BD,$89,$1E,$36,$BF,$D9,$DF,$F5,$49,$1F,$F3
	.dc.b	$8B,$FE,$7C,$F3,$3F,$C4,$6D,$C0,$D2,$4F,$FA,$21,$FF,$67,$46,$A3
	.dc.b	$04,$3F,$D3,$07,$F3,$52,$F6,$7F,$2B,$E2,$FF,$4C,$46,$4F,$F7,$67
	.dc.b	$96,$71,$5F,$EE,$B3,$1F,$35,$FE,$AB,$24,$7F,$AA,$5F,$E6,$DE,$AB
	.dc.b	$CE,$DC,$FF,$AC,$E7,$31,$88,$ED,$4A,$E1,$0A,$FF,$40,$E3,$23,$EF
	.dc.b	$35,$FF,$09,$3F,$ED,$42,$3F,$E9,$C7,$E5,$11,$FF,$53,$E7,$FE,$50
	.dc.b	$8C,$47,$8B,$2F,$F7,$57,$E4,$3F,$EE,$98,$34,$8F,$FB,$DC,$C9,$5F
	.dc.b	$F7,$E3,$83,$FD,$F4,$53,$FD,$7B,$8F,$CF,$EA,$FF,$62,$DF,$DA,$24
	.dc.b	$FF,$91,$B9,$3F,$CC,$B5,$F4,$9B,$3F,$D8,$DB,$3B,$9F,$63,$FE,$4E
	.dc.b	$C1,$E2,$7D,$BF,$6B,$F0,$0D,$9D,$35,$F8,$4D,$7F,$05,$EA,$3D,$5C
	.dc.b	$42,$30,$AB,$F9,$B4,$61,$B3,$CF,$06,$0F,$5B,$0D,$C3,$91,$E9,$8C
	.dc.b	$9A,$9F,$CA,$7C,$6F,$FB,$EC,$93,$FE,$F9,$7F,$1A,$9A,$47,$FC,$E3
	.dc.b	$C3,$F8,$E8,$FC,$F4,$7E,$7F,$6F,$F5,$CF,$5F,$FA,$53,$F5,$CC,$CA
	.dc.b	$73,$82,$73,$18,$6E,$EB,$E5,$7D,$AB,$FD,$8B,$ED,$E2,$3C,$14,$8C
	.dc.b	$47,$78,$C6,$1D,$E3,$94,$63,$0E,$FE,$E7,$E0,$27,$9B,$5E,$4D,$C3
	.dc.b	$FD,$AA,$E6,$3E,$C7,$7F,$F6,$7E,$0F,$F2,$1C,$7C,$06,$BF,$19,$AF
	.dc.b	$F1,$4F,$75,$E9,$FF,$73,$EB,$46,$27,$FB,$C4,$62,$75,$39,$9F,$E0
	.dc.b	$FC,$7F,$DE,$78,$7F,$17,$18,$3F,$B2,$B8,$7C,$98,$F9,$C8,$FF,$80
	.dc.b	$F1,$F2,$EC,$F7,$9E,$FF,$FA,$B6,$BF,$D9,$F1,$89,$FE,$E2,$D8,$FF
	.dc.b	$27,$35,$FC,$0F,$C0,$7F,$14,$7F,$C2,$4D,$73,$DB,$FE,$0F,$8C,$47
	.dc.b	$C7,$C7,$CA,$FF,$86,$71,$F9,$2F,$D0,$3F,$C3,$5F,$03,$FE,$B2,$7F
	.dc.b	$1E,$D7,$F2,$AD,$7F,$9A,$35,$7C,$17,$EF,$9F,$E8,$47,$E7,$1C,$70
	.dc.b	$7F,$8A,$5B,$1F,$9C,$6B,$E7,$38,$9F,$C1,$EF,$FA,$14,$FA,$5B,$1F
	.dc.b	$48,$D8,$F8,$37,$7F,$F5,$A8,$FD,$4C,$63,$5D,$B6,$FF,$C7,$8D,$9F
	.dc.b	$D7,$FC,$03,$E1,$3F,$A9,$78,$1F,$E4,$18,$C4,$FF,$57,$0E,$23,$53
	.dc.b	$C2,$7C,$D1,$E9,$78,$6B,$FE,$A6,$8C,$1F,$D2,$23,$FA,$59,$FD,$24
	.dc.b	$87,$F0,$6F,$85,$FF,$80,$0F,$F9,$45,$AF,$F9,$55,$AF,$F0,$86,$BC
	.dc.b	$7D,$9F,$F2,$A4,$62,$3F,$B7,$35,$34,$91,$C0,$91,$E1,$6B,$E1,$6B
	.dc.b	$E1,$6B,$E1,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1
	.dc.b	$0F,$7B,$E4,$69,$DC,$A7,$42,$9C,$CA,$73,$3E,$3B,$CD,$A7,$32,$9C
	.dc.b	$CA,$73,$29,$CE,$3E,$7B,$C5,$AF,$16,$BC,$5A,$F1,$8F,$14,$62,$3C
	.dc.b	$E4,$3B,$95,$D3,$82,$38,$24,$7F,$A6,$FD,$8F,$F7,$DC,$FF,$C0,$1F
	.dc.b	$63,$F4,$4F,$FB,$E5,$B1,$A8,$FE,$BB,$FA,$23,$D1,$FF,$1E,$70,$FF
	.dc.b	$AA,$E3,$13,$FC,$F9,$B1,$99,$DC,$FF,$62,$9C,$B6,$32,$F6,$6B,$FE
	.dc.b	$49,$8D,$B9,$3B,$46,$94,$69,$E1,$19,$8C,$46,$A3,$04,$79,$92,$7D
	.dc.b	$0D,$3C,$C5,$3C,$4E,$88,$E9,$AE,$8C,$10,$C1,$0F,$39,$D4,$AF,$26
	.dc.b	$9C,$4A,$71,$3D,$87,$12,$BC,$5A,$71,$29,$C6,$7E,$29,$27,$E2,$4F
	.dc.b	$18,$C4,$F5,$6C,$66,$77,$3D,$67,$2D,$8C,$BD,$DA,$FA,$88,$3E,$25
	.dc.b	$F1,$4E,$27,$9B,$FE,$F9,$9C,$C6,$2B,$FC,$5B,$18,$9D,$57,$17,$FA
	.dc.b	$2F,$25,$AF,$E9,$6D,$7F,$C7,$3C,$7F,$D3,$67,$78,$75,$9F,$F1,$EF
	.dc.b	$9F,$FF,$00,$27,$FC,$77,$08,$C3,$FE,$3E,$6B,$C3,$D0,$7A,$3C,$E3
	.dc.b	$D9,$FB,$6F,$F9,$B3,$0F,$F1,$38,$71,$BF,$F9,$22,$31,$E6,$FC,$38
	.dc.b	$C4,$EA,$BD,$75,$FE,$2E,$CF,$FB,$35,$7A,$89,$3F,$E5,$69,$7B,$6F
	.dc.b	$9B,$FD,$7C,$61,$B1,$99,$F2,$B6,$BC,$9F,$A5,$3D,$FF,$25,$75,$8C
	.dc.b	$61,$EC,$6D,$70,$AD,$D7,$D0,$E0,$AF,$FB,$1F,$2F,$4B,$EC,$AF,$4B
	.dc.b	$FE,$BE,$C7,$FC,$66,$F9,$3C,$3F,$39,$1F,$F2,$8F,$AC,$FC,$DF,$FD
	.dc.b	$89,$0F,$C8,$BF,$CE,$73,$9F,$1F,$FD,$AD,$18,$AD,$B6,$33,$5E,$32
	.dc.b	$1E,$9D,$BE,$86,$9E,$AC,$3A,$4B,$E7,$B9,$34,$FF,$43,$C6,$A3,$27
	.dc.b	$93,$CB,$FF,$02,$FC,$15,$EF,$FF,$75,$36,$CD,$3E,$2B,$F2,$70,$F9
	.dc.b	$6F,$F8,$AD,$BC,$1F,$32,$3F,$DE,$98,$FF,$9C,$2F,$AC,$79,$30,$F0
	.dc.b	$B6,$32,$F0,$8C,$BF,$4F,$F0,$8F,$F7,$F7,$E0,$5F,$97,$E1,$46,$27
	.dc.b	$F2,$11,$85,$3B,$9D,$47,$CB,$E2,$F5,$BE,$9D,$BE,$69,$FA,$6F,$D1
	.dc.b	$DF,$43,$F8,$88,$46,$1D,$46,$63,$15,$FE,$AC,$E0,$FE,$19,$F5,$96
	.dc.b	$CD,$26,$1F,$A7,$FA,$87,$BF,$C2,$30,$D4,$7F,$17,$8D,$D7,$48,$DB
	.dc.b	$5F,$F5,$BE,$DF,$C7,$7F,$04,$FF,$5F,$B6,$3F,$D6,$AD,$77,$29,$F9
	.dc.b	$DD,$3E,$06,$CE,$4F,$9F,$19,$69,$93,$CF,$2F,$C6,$5E,$CF,$F2,$FF
	.dc.b	$B5,$FE,$21,$18,$94,$62,$75,$5F,$C2,$EB,$53,$FC,$93,$C3,$78,$7F
	.dc.b	$C8,$23,$0D,$AF,$E4,$CD,$7F,$93,$78,$7D,$9F,$C3,$FC,$F7,$CC,$AF
	.dc.b	$C6,$36,$32,$6C,$FE,$08,$E5,$FE,$0D,$DB,$F8,$C7,$FA,$57,$47,$F0
	.dc.b	$2E,$8F,$CB,$4E,$75,$FE,$46,$D9,$7D,$95,$25,$FE,$60,$FC,$DB,$D7
	.dc.b	$DB,$F9,$3F,$57,$FD,$5F,$F2,$5F,$F7,$CB,$6B,$FD,$DE,$D7,$FD,$EE
	.dc.b	$D7,$FD,$ED,$F1,$B6,$BE,$9B,$78,$8D,$38,$7C,$43,$F0,$77,$B9,$FD
	.dc.b	$9F,$7F,$C9,$E3,$F9,$4A,$D4,$E6,$33,$18,$DF,$D9,$1B,$BE,$71,$88
	.dc.b	$FC,$EF,$27,$DC,$E3,$C4,$7F,$61,$E4,$FE,$BB,$97,$DE,$9C,$D6,$E3
	.dc.b	$C7,$4C,$5E,$13,$F0,$3F,$AD,$18,$6C,$66,$7C,$53,$97,$C9,$BF,$E6
	.dc.b	$3A,$BF,$C0,$DF,$83,$FE,$AB,$53,$49,$5F,$A2,$7F,$4E,$47,$07,$F1
	.dc.b	$6E,$6F,$DD,$FE,$0F,$F8,$68,$34,$8F,$F6,$BF,$DA,$3F,$88,$6B,$FD
	.dc.b	$DD,$B1,$9F,$13,$E8,$3F,$B0,$8C,$BF,$3F,$8F,$FC,$3D,$7F,$1B,$5F
	.dc.b	$A8,$DE,$3F,$A6,$D9,$DD,$7E,$AB,$8B,$CA,$D9,$FA,$44,$7E,$FF,$07
	.dc.b	$CB,$8E,$CD,$8C,$F0,$FC,$06,$9D,$0F,$E3,$BC,$3F,$C1,$BC,$9F,$F1
	.dc.b	$A7,$E1,$1F,$72,$31,$C4,$8C,$73,$FF,$56,$8C,$5F,$E5,$2B,$75,$E2
	.dc.b	$C3,$F9,$4B,$67,$67,$E6,$71,$FC,$14,$64,$EB,$8F,$F9,$17,$F0,$8B
	.dc.b	$89,$A4,$BF,$5B,$D4,$8E,$0F,$95,$E3,$7E,$64,$8E,$0F,$46,$6F,$C2
	.dc.b	$FF,$C2,$FA,$7F,$28,$6C,$67,$05,$FC,$29,$CF,$C2,$FF,$96,$FF,$21
	.dc.b	$F8,$AD,$C7,$AB,$5F,$9F,$82,$31,$1F,$E6,$3F,$D4,$DF,$84,$7F,$8E
	.dc.b	$1B,$1F,$E4,$A6,$E3,$FE,$49,$6C,$7A,$CF,$E5,$F1,$EE,$69,$EE,$F4
	.dc.b	$3D,$64,$CF,$D7,$19,$6B,$C0,$87,$F0,$6E,$EF,$F2,$FD,$75,$86,$BF
	.dc.b	$D3,$BD,$1F,$8E,$4B,$E7,$3A,$D3,$FC,$6D,$18,$AF,$EA,$BC,$DF,$D4
	.dc.b	$2B,$E2,$B5,$E0,$D8,$DC,$FF,$45,$9F,$27,$37,$67,$F8,$C3,$7B,$FF
	.dc.b	$8E,$65,$E7,$9C,$B6,$33,$1F,$24,$F5,$FE,$4B,$FE,$95,$BF,$33,$6B
	.dc.b	$F4,$8D,$9F,$D0,$35,$FE,$E3,$F6,$1F,$E2,$6A,$FE,$5F,$18,$F4,$7F
	.dc.b	$D4,$19,$FF,$A8,$63,$13,$A7,$FD,$49,$5C,$67,$51,$EE,$C3,$D1,$B7
	.dc.b	$D2,$8F,$46,$8C,$9F,$23,$FE,$8B,$CB,$CA,$DA,$FD,$FC,$EA,$7F,$36
	.dc.b	$68,$B1,$E4,$F9,$CF,$0C,$33,$F6,$46,$1B,$1F,$E9,$16,$CF,$E2,$0F
	.dc.b	$9E,$7C,$D2,$9F,$E1,$6E,$27,$D2,$7F,$8B,$AF,$BC,$62,$51,$89,$D5
	.dc.b	$FF,$A7,$EB,$53,$F1,$B2,$F8,$5E,$E7,$FB,$37,$CC,$C6,$7F,$EB,$36
	.dc.b	$C6,$70,$FB,$2D,$CB,$C9,$EA,$7F,$9D,$27,$B3,$63,$27,$F2,$12,$F6
	.dc.b	$7D,$67,$DA,$7D,$73,$E0,$71,$FE,$39,$FA,$37,$FE,$E1,$9F,$1D,$F3
	.dc.b	$9F,$C8,$CE,$63,$CA,$66,$3A,$92,$BA,$3D,$CF,$A0,$E1,$D7,$07,$E1
	.dc.b	$DF,$F0,$4F,$C1,$3E,$0D,$F0,$BF,$E5,$EC,$7F,$80,$CE,$63,$1C,$9C
	.dc.b	$2B,$76,$BD,$C6,$1B,$5C,$27,$FC,$D3,$7C,$DE,$17,$CE,$75,$09,$D5
	.dc.b	$7F,$12,$C7,$FC,$29,$3F,$E1,$12,$1F,$D9,$4A,$7F,$93,$8A,$7F,$04
	.dc.b	$3F,$6F,$0C,$9E,$E7,$D0,$23,$F3,$9F,$B5,$5E,$0A,$EC,$43,$B1,$F8
	.dc.b	$43,$9C,$D9,$FC,$CE,$BF,$E8,$86,$CE,$8F,$C0,$CF,$C9,$3E,$ED,$FE
	.dc.b	$08,$8F,$E7,$78,$FF,$D2,$27,$E8,$67,$F4,$F1,$96,$9F,$AB,$8D,$C7
	.dc.b	$9E,$3E,$11,$4F,$E4,$A7,$F8,$E1,$FD,$E7,$3F,$C5,$9F,$DE,$36,$3F
	.dc.b	$78,$D9,$FE,$2F,$3F,$AC,$E6,$FE,$35,$5F,$C2,$67,$0F,$FA,$49,$FE
	.dc.b	$1F,$1A,$7E,$49,$0F,$29,$63,$C6,$3F,$0C,$F1,$B4,$F1,$96,$3F,$94
	.dc.b	$C7,$EF,$A3,$F6,$51,$FC,$E9,$BC,$9F,$C9,$5B,$C9,$FC,$F6,$7F,$0E
	.dc.b	$7C,$43,$E3,$6B,$F8,$17,$CC,$64,$AE,$5F,$EA,$B2,$F9,$6F,$F1,$2F
	.dc.b	$4F,$E3,$DB,$19,$9F,$F0,$7C,$E5,$FF,$08,$F6,$79,$BB,$3F,$B2,$C7
	.dc.b	$F8,$BA,$30,$7B,$1F,$11,$F2,$08,$63,$2F,$F5,$FF,$6F,$32,$61,$FE
	.dc.b	$76,$9F,$EE,$51,$CF,$2F,$54,$74,$23,$ED,$97,$5F,$F8,$4A,$7F,$C2
	.dc.b	$1D,$DC,$DF,$F0,$C3,$67,$6E,$70,$FF,$68,$F7,$FE,$C7,$F1,$DF,$F1
	.dc.b	$2D,$1A,$4B,$FF,$19,$F8,$5F,$0C,$93,$E9,$BF,$F1,$A3,$E1,$CB,$D2
	.dc.b	$4E,$C7,$83,$96,$E1,$FE,$51,$FC,$2A,$3F,$E6,$26,$BF,$E3,$4E,$FF
	.dc.b	$E0,$D1,$FE,$50,$24,$7F,$91,$A3,$2F,$F9,$AB,$F0,$9F,$F2,$E1,$1C
	.dc.b	$1B,$C5,$1F,$E5,$A2,$46,$5F,$F3,$27,$E1,$12,$3F,$CE,$AD,$7F,$DD
	.dc.b	$2F,$FA,$C4,$8F,$FA,$15,$FF,$44,$49,$A4,$8F,$F6,$1F,$97,$FE,$C0
	.dc.b	$8F,$F4,$D7,$9D,$FC,$8F,$E5,$2D,$9A,$4B,$FF,$54,$1D,$4B,$5F,$EA
	.dc.b	$92,$3F,$ED,$49,$7B,$0F,$6D,$2F,$C2,$3F,$88,$4E,$63,$07,$17,$11
	.dc.b	$A9,$E2,$FF,$B0,$49,$3F,$E7,$8F,$5B,$F9,$E4,$62,$3F,$D1,$9E,$87
	.dc.b	$83,$2F,$F4,$81,$23,$FD,$42,$FF,$B6,$C8,$FF,$AB,$DA,$FF,$AE,$71
	.dc.b	$F8,$8F,$CA,$EC,$F9,$0D,$8C,$BB,$9F,$F5,$2F,$AF,$FD,$3F,$F2,$6B
	.dc.b	$46,$92,$BF,$D2,$3E,$CE,$DB,$3B,$6C,$FF,$AC,$27,$51,$F3,$27,$E7
	.dc.b	$FB,$7E,$F8,$F9,$4D,$75,$B2,$31,$1F,$D8,$E3,$0F,$A8,$C7,$B3,$AA
	.dc.b	$FB,$6D,$93,$4C,$79,$FF,$7B,$C3,$FB,$AD,$BF,$B8,$D7,$C6,$D7,$FC
	.dc.b	$F3,$1F,$C5,$7A,$9D,$3F,$BA,$8C,$36,$7F,$DC,$91,$FD,$67,$DD,$78
	.dc.b	$9F,$CA,$E7,$E2,$46,$33,$FE,$61,$18,$9D,$3F,$E0,$39,$D5,$7E,$3E
	.dc.b	$1F,$83,$D2,$3F,$95,$E6,$FC,$DC,$7F,$B8,$B9,$BF,$C1,$8D,$C7,$8D
	.dc.b	$7E,$55,$AF,$FB,$02,$76,$FE,$9F,$E0,$1F,$03,$FE,$15,$FB,$5F,$EE
	.dc.b	$91,$8B,$46,$27,$55,$C5,$AF,$FB,$33,$D0,$AF,$FA,$BD,$BE,$17,$C6
	.dc.b	$6E,$C7,$78,$3E,$11,$DD,$FD,$AA,$30,$D8,$CC,$7F,$4B,$32,$7F,$43
	.dc.b	$F8,$67,$95,$FD,$5D,$B1,$98,$C1,$D7,$9D,$E8,$7E,$34,$9A,$7E,$4E
	.dc.b	$BF,$85,$C7,$E6,$3D,$C7,$0B,$E1,$BF,$C5,$D7,$FC,$38,$4E,$A3,$31
	.dc.b	$FC,$4A,$30,$FF,$90,$8E,$0E,$3D,$4F,$EC,$D9,$FA,$26,$E7,$E8,$CB
	.dc.b	$D4,$7F,$6B,$BE,$13,$CB,$1F,$23,$D4,$F0,$38,$C3,$CD,$1E,$68,$C3
	.dc.b	$E2,$F8,$7F,$D6,$AF,$B4,$DD,$A7,$31,$86,$CF,$1E,$DF,$E5,$F3,$E8
	.dc.b	$C7,$CF,$34,$86,$CF,$F4,$11,$7D,$AE,$95,$FC,$E1,$B1,$A3,$F8,$5E
	.dc.b	$5F,$DF,$F7,$F7,$47,$F3,$98,$C1,$FE,$05,$5A,$FC,$B7,$E0,$B5,$FB
	.dc.b	$0D,$7F,$A3,$B7,$07,$C4,$FF,$4B,$7F,$20,$D9,$FF,$2D,$35,$FE,$A4
	.dc.b	$D7,$F9,$0F,$C5,$79,$C8,$FE,$67,$04,$79,$EF,$C5,$3E,$8D,$FF,$00
	.dc.b	$FC,$83,$F9,$87,$C5,$FF,$9E,$5F,$F3,$53,$5F,$EC,$4D,$77,$ED,$7F
	.dc.b	$8B,$FC,$0F,$F0,$AF,$B9,$FC,$73,$D1,$FF,$55,$D7,$E6,$7C,$3F,$E4
	.dc.b	$31,$88,$FE,$D3,$18,$9F,$B8,$48,$F9,$78,$FF,$A7,$63,$11,$FA,$3F
	.dc.b	$8E,$7C,$76,$63,$F2,$31,$8E,$0E,$D0,$8C,$57,$F9,$1D,$B1,$F9,$8F
	.dc.b	$90,$7C,$87,$DD,$6C,$7F,$90,$DA,$FF,$93,$DA,$F1,$8E,$04,$78,$35
	.dc.b	$E0,$D7,$81,$A2,$1A,$29,$A3,$04,$30,$43,$F8,$34,$79,$1B,$1D,$DA
	.dc.b	$74,$29,$CC,$A7,$33,$FC,$3B,$1C,$DB,$1C,$DA,$73,$29,$CC,$A7,$31
	.dc.b	$1C,$DB,$1C,$DA,$73,$29,$CC,$A7,$38,$E4,$48,$E4,$D8,$E4,$D7,$93
	.dc.b	$5E,$51,$C0,$8F,$06,$BC,$1F,$E0,$1F,$01,$FE,$FF,$8F,$E3,$ED,$7F
	.dc.b	$E0,$01,$FE,$93,$70,$FF,$91,$BC,$27,$D0,$FF,$92,$3E,$0B,$FD,$31
	.dc.b	$18,$94,$62,$75,$1F,$CB,$27,$53,$98,$D9,$99,$DB,$FD,$1E,$76,$E4
	.dc.b	$D9,$98,$DB,$E5,$6B,$FD,$E1,$B1,$CA,$38,$1C,$8E,$05,$38,$15,$E0
	.dc.b	$EA,$7D,$04,$7C,$8D,$7A,$B5,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$F2
	.dc.b	$BD,$5A,$72,$29,$C4,$A7,$18,$F5,$B8,$23,$82,$4F,$C8,$24,$7C,$36
	.dc.b	$CF,$86,$77,$39,$24,$66,$77,$1E,$59,$DC,$E6,$36,$66,$76,$FF,$80
	.dc.b	$3D,$76,$8F,$91,$FE,$F9,$18,$6C,$66,$7F,$B9,$CE,$5F,$E6,$31,$97
	.dc.b	$FA,$DE,$F9,$6C,$65,$FE,$BB,$19,$7F,$9E,$F1,$7F,$B6,$2F,$FA,$21
	.dc.b	$23,$FC,$0F,$C3,$FD,$12,$3F,$DF,$5F,$27,$F9,$C6,$E7,$FE,$AE,$6D
	.dc.b	$77,$13,$B7,$FC,$23,$E3,$FF,$93,$6B,$FD,$D9,$AF,$C2,$6E,$7F,$97
	.dc.b	$9D,$57,$8B,$2F,$F7,$6D,$7F,$59,$24,$67,$93,$F3,$DF,$9C,$B9,$BF
	.dc.b	$4E,$D9,$CC,$60,$FD,$5E,$1F,$E8,$0C,$7C,$18,$F8,$EB,$CE,$D7,$FC
	.dc.b	$7D,$3B,$C7,$C3,$1F,$DA,$67,$CA,$F8,$E3,$2F,$FB,$4B,$07,$CB,$79
	.dc.b	$09,$1E,$87,$F7,$BF,$93,$3C,$CF,$F6,$CF,$C1,$FF,$8E,$29,$39,$51
	.dc.b	$98,$C3,$67,$8C,$F8,$C8,$79,$F5,$FE,$09,$C3,$FE,$00,$CB,$84,$FF
	.dc.b	$9D,$4F,$F2,$3E,$1D,$FC,$BC,$92,$33,$3F,$EE,$EE,$4F,$13,$63,$22
	.dc.b	$32,$D3,$C6,$7E,$4B,$0F,$1E,$BF,$2A,$3C,$D5,$C2,$A4,$74,$6B,$FE
	.dc.b	$71,$E0,$FF,$1B,$4F,$C7,$6C,$7F,$98,$1A,$EF,$CF,$FB,$8F,$86,$7E
	.dc.b	$BA,$CF,$E2,$3E,$61,$FA,$E3,$E6,$3F,$99,$B6,$3F,$97,$36,$BE,$C3
	.dc.b	$63,$FD,$0D,$C6,$F8,$57,$C1,$25,$F2,$FF,$E7,$6A,$D4,$E6,$31,$5C
	.dc.b	$2B,$67,$47,$11,$AF,$85,$E1,$39,$6B,$FE,$F9,$94,$62,$3F,$89,$C6
	.dc.b	$1E,$A7,$97,$9D,$F0,$FF,$9A,$C5,$F6,$9F,$BC,$8C,$37,$A3,$F8,$EB
	.dc.b	$67,$77,$FA,$AA,$FD,$84,$F1,$E3,$DC,$E5,$AF,$86,$76,$FF,$97,$BE
	.dc.b	$9F,$F9,$4F,$DA,$7F,$98,$9B,$19,$C1,$39,$6C,$7F,$A8,$BE,$F1,$F5
	.dc.b	$3C,$23,$FD,$75,$18,$9D,$C6,$27,$4F,$D7,$9D,$57,$F9,$CA,$73,$4F
	.dc.b	$27,$37,$FE,$74,$9C,$C6,$23,$F3,$1E,$F7,$E4,$9F,$E3,$0D,$E8,$34
	.dc.b	$D7,$27,$E2,$E7,$FD,$A0,$D7,$F1,$4D,$7F,$2F,$CD,$FE,$0A,$C7,$FD
	.dc.b	$DD,$F8,$87,$F5,$0F,$C9,$BF,$1D,$ED,$23,$17,$F9,$68,$C5,$2B,$75
	.dc.b	$AA,$FA,$9E,$1D,$5C,$BF,$86,$8C,$36,$F9,$4F,$E0,$67,$83,$FC,$DF
	.dc.b	$B3,$E0,$12,$7F,$C3,$1A,$FE,$E2,$7F,$6C,$7E,$CF,$B3,$F6,$AF,$F0
	.dc.b	$63,$FD,$BB,$AF,$C1,$26,$BF,$B6,$3E,$17,$E3,$6C,$BF,$61,$F8,$57
	.dc.b	$EB,$E3,$16,$8C,$4E,$A3,$F7,$33,$A9,$E1,$F7,$B9,$9F,$DD,$66,$A4
	.dc.b	$E6,$7F,$BF,$C6,$1E,$C7,$48,$D3,$F8,$29,$F8,$3C,$3F,$81,$69,$FC
	.dc.b	$3B,$1F,$64,$D2,$77,$FF,$21,$DB,$E5,$36,$BE,$5C,$7E,$3F,$5F,$93
	.dc.b	$5F,$30,$F9,$44,$3F,$17,$7D,$AF,$F8,$DB,$63,$F9,$66,$1E,$59,$FE
	.dc.b	$60,$D8,$FA,$73,$B7,$F2,$ED,$7F,$90,$6D,$F9,$EC,$7D,$B7,$C9,$B3
	.dc.b	$C8,$F6,$96,$3C,$6F,$F4,$8F,$06,$47,$F3,$AC,$7D,$B5,$F4,$F9,$7F
	.dc.b	$02,$4B,$FF,$33,$46,$5B,$AB,$A3,$F9,$A7,$CB,$FE,$15,$B3,$F8,$26
	.dc.b	$BF,$80,$9E,$0A,$7F,$19,$E6,$FE,$40,$91,$D9,$FF,$3D,$E3,$F4,$A3
	.dc.b	$FA,$87,$67,$F6,$17,$EA,$13,$1F,$C6,$61,$F9,$A2,$47,$A1,$E7,$3C
	.dc.b	$63,$9E,$9F,$11,$FE,$DF,$5F,$95,$FB,$0F,$D1,$92,$3E,$0B,$FA,$6F
	.dc.b	$B1,$FE,$A9,$F6,$DF,$A5,$6C,$FE,$89,$AF,$C1,$6B,$FE,$5C,$FC,$D5
	.dc.b	$F8,$07,$F1,$16,$C7,$D0,$6F,$0F,$F5,$F6,$C7,$EF,$73,$F5,$6B,$F1
	.dc.b	$4F,$B7,$AB,$F1,$7F,$DD,$92,$BE,$CB,$FC,$7F,$E3,$7F,$C4,$BB,$7F
	.dc.b	$AE,$0F,$EE,$2E,$32,$FF,$38,$D7,$EE,$23,$13,$AF,$3B,$FC,$A8,$F8
	.dc.b	$F8,$3E,$55,$7E,$96,$3F,$7D,$F7,$1F,$EC,$77,$F8,$2B,$63,$F0,$8D
	.dc.b	$7F,$80,$B6,$17,$E0,$9F,$15,$F8,$1B,$3E,$96,$E1,$E8,$7F,$0A,$7D
	.dc.b	$E7,$F4,$1A,$7E,$26,$5F,$86,$32,$F0,$36,$FA,$1F,$E9,$27,$17,$D6
	.dc.b	$B8,$D7,$0F,$82,$FF,$0B,$75,$66,$D1,$8A,$FF,$0E,$C6,$30,$FD,$E1
	.dc.b	$FC,$91,$C5,$71,$AF,$F2,$45,$70,$6D,$6D,$AF,$95,$AF,$BC,$FA,$98
	.dc.b	$7F,$9B,$A7,$DF,$3C,$1B,$3F,$E5,$17,$B3,$4E,$C5,$3E,$5E,$FF,$8F
	.dc.b	$CB,$F2,$24,$CB,$F0,$33,$A8,$FD,$77,$AE,$F8,$17,$B7,$FC,$73,$3F
	.dc.b	$F1,$E5,$AB,$53,$98,$FF,$20,$46,$1F,$F4,$C4,$F0,$8C,$76,$23,$17
	.dc.b	$DA,$B6,$72,$71,$3B,$8C,$5F,$77,$C5,$3A,$AF,$F4,$14,$7F,$A3,$72
	.dc.b	$63,$F8,$58,$FF,$56,$19,$70,$F0,$32,$43,$07,$CC,$F8,$1F,$BC,$9A
	.dc.b	$4D,$38,$47,$FA,$04,$ED,$A7,$FA,$7D,$FE,$32,$7F,$98,$F7,$FC,$65
	.dc.b	$71,$8E,$D5,$F7,$23,$BE,$92,$3E,$FB,$67,$F2,$57,$FE,$72,$69,$FD
	.dc.b	$E0,$A7,$E7,$5D,$3E,$53,$CE,$79,$0F,$4C,$FE,$EB,$83,$F8,$5B,$FC
	.dc.b	$24,$E9,$1F,$EF,$EE,$1F,$E2,$6F,$EB,$5B,$1F,$C2,$23,$6F,$F0,$C6
	.dc.b	$F2,$38,$FC,$73,$99,$FE,$04,$78,$CF,$A2,$79,$CA,$F9,$4E,$B5,$E0
	.dc.b	$8F,$E1,$A7,$8E,$FD,$2F,$F5,$BE,$4F,$A5,$5F,$E1,$B6,$E1,$FC,$C6
	.dc.b	$7F,$DA,$93,$97,$EC,$6D,$F8,$AC,$AE,$DF,$88,$9F,$21,$29,$CB,$FD
	.dc.b	$12,$3F,$A1,$73,$78,$23,$11,$FE,$A8,$6B,$FD,$3C,$7D,$1F,$F5,$56
	.dc.b	$9F,$D5,$4A,$6D,$FF,$1F,$75,$7A,$89,$1E,$AC,$D1,$F9,$18,$F9,$87
	.dc.b	$A0,$A6,$30,$FE,$E6,$F1,$9F,$8A,$47,$E0,$E5,$1F,$CE,$90,$FF,$25
	.dc.b	$F5,$FC,$C2,$3E,$37,$E7,$F6,$FF,$B1,$7E,$6B,$FC,$25,$26,$91,$FF
	.dc.b	$2A,$F7,$FD,$62,$3E,$D7,$D9,$BF,$D5,$8F,$11,$F8,$83,$DB,$E2,$FF
	.dc.b	$B6,$DF,$F1,$69,$0F,$F1,$91,$FE,$35,$21,$FE,$42,$CF,$F7,$1C,$BF
	.dc.b	$D0,$23,$64,$3F,$C8,$27,$EF,$67,$B3,$C4,$E6,$7F,$A9,$27,$9F,$6F
	.dc.b	$E7,$A3,$FC,$A6,$71,$F0,$7E,$B5,$72,$8E,$A7,$B0,$FB,$18,$7B,$DF
	.dc.b	$DC,$E0,$F0,$BA,$47,$FA,$47,$1F,$4C,$62,$75,$1D,$E6,$1F,$0C,$EE
	.dc.b	$7A,$48,$7A,$4F,$59,$FE,$80,$F0,$FF,$DF,$DF,$83,$FF,$A2,$20,$D2
	.dc.b	$3F,$E8,$EF,$1B,$6D,$36,$58,$FF,$4A,$B8,$3C,$64,$3E,$39,$E7,$3E
	.dc.b	$1F,$3F,$EC,$DF,$CD,$E5,$F1,$5F,$EA,$F8,$FC,$07,$FD,$10,$43,$FA
	.dc.b	$D9,$F1,$CF,$F8,$01,$6C,$5F,$21,$86,$CF,$FA,$FD,$AF,$FB,$21,$AF
	.dc.b	$1F,$2F,$FD,$9F,$3F,$EC,$E9,$C9,$FE,$D8,$2C,$6A,$7F,$18,$7B,$8A
	.dc.b	$7F,$B8,$0F,$C4,$C4,$7E,$99,$FE,$AD,$3F,$DF,$0B,$DB,$A4,$8F,$F7
	.dc.b	$C4,$65,$F8,$3E,$7B,$8B,$E0,$57,$C4,$94,$E6,$B8,$47,$F8,$0A,$31
	.dc.b	$3B,$6C,$7F,$81,$A5,$E7,$7F,$A4,$5E,$31,$F3,$63,$15,$F4,$A7,$7C
	.dc.b	$BE,$2A,$E5,$5A,$7F,$C2,$3E,$63,$D0,$FE,$FB,$6F,$B5,$FB,$73,$9E
	.dc.b	$1F,$B5,$18,$AD,$B6,$33,$7D,$25,$E9,$E2,$02,$FC,$11,$98,$C4,$F0
	.dc.b	$C3,$C5,$1C,$E7,$50,$F5,$1E,$AF,$4F,$8F,$F8,$11,$8B,$46,$27,$51
	.dc.b	$F2,$E7,$53,$C3,$D8,$CC,$FF,$BD,$A5,$18,$8F,$CD,$46,$1E,$A7,$38
	.dc.b	$D3,$F8,$D3,$F5,$85,$3F,$23,$CD,$FA,$B7,$F5,$FD,$FF,$D0,$5F,$D4
	.dc.b	$F3,$FE,$B5,$BD,$5F,$C3,$C9,$9F,$F8,$87,$DB,$FF,$1D,$79,$5F,$8D
	.dc.b	$6D,$7E,$C9,$AF,$F9,$39,$AF,$80,$FE,$44,$7F,$58,$72,$9A,$4F,$33
	.dc.b	$EC,$CF,$7F,$73,$ED,$FA,$36,$FF,$CF,$87,$78,$CB,$E3,$6B,$FD,$2F
	.dc.b	$4F,$84,$FF,$0E,$E4,$FE,$21,$8A,$BF,$0F,$E7,$F5,$25,$73,$3F,$CD
	.dc.b	$87,$F1,$0F,$77,$F1,$EF,$F9,$15,$AF,$EC,$5A,$EF,$93,$FA,$FE,$1F
	.dc.b	$C7,$F1,$F1,$91,$FE,$45,$CD,$FA,$4A,$FD,$D3,$5F,$D2,$C7,$F4,$68
	.dc.b	$C4,$7F,$9F,$2D,$C1,$FE,$07,$6C,$79,$63,$07,$8B,$87,$F0,$51,$F1
	.dc.b	$6F,$C3,$DF,$D4,$9C,$3F,$E6,$1C,$51,$FD,$5B,$AF,$FC,$F9,$E8,$FF
	.dc.b	$0F,$6E,$1F,$CD,$1A,$FF,$A4,$1A,$FE,$D3,$4F,$F1,$A4,$FF,$9B,$89
	.dc.b	$1F,$CF,$23,$2F,$F3,$EF,$84,$7C,$07,$F3,$DA,$FE,$3D,$18,$CB,$FD
	.dc.b	$25,$5F,$E9,$E8,$C4,$EA,$B8,$4F,$F4,$8F,$8A,$FF,$59,$4F,$F2,$76
	.dc.b	$C6,$7D,$FD,$BF,$EA,$88,$CB,$FE,$AA,$CD,$F0,$5F,$E3,$56,$C7,$FA
	.dc.b	$1D,$AF,$F3,$36,$BF,$D6,$1A,$F3,$8E,$44,$79,$35,$E4,$D7,$93,$5E
	.dc.b	$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18,$21,$FC,$C2
	.dc.b	$3D,$AD,$8F,$33,$4F,$09,$4E,$C5,$3A,$94,$EA,$7E,$49,$E6,$D3,$99
	.dc.b	$4E,$65,$39,$94,$E7,$1E,$67,$07,$72,$18,$23,$B2,$47,$89,$B1,$E2
	.dc.b	$6B,$E2,$F5,$FF,$DD,$8F,$F8,$61,$AF,$FB,$AB,$0F,$F6,$74,$62,$3E
	.dc.b	$57,$E0,$BF,$D6,$BA,$BF,$14,$D8,$FC,$53,$5F,$CA,$35,$E2,$7F,$BC
	.dc.b	$4C,$94,$CC,$74,$70,$7E,$FA,$39,$34,$CD,$A3,$2D,$8C,$C6,$08,$E0
	.dc.b	$93,$DC,$8F,$A8,$C9,$5C,$BE,$86,$BE,$83,$44,$34,$53,$46,$08,$60
	.dc.b	$87,$B5,$EA,$D3,$91,$4E,$25,$38,$CF,$C1,$8E,$2D,$8E,$2D,$78,$B5
	.dc.b	$E3,$3F,$1A,$78,$B6,$38,$B5,$F0,$E2,$FC,$57,$FC,$BD,$F0,$9F,$C6
	.dc.b	$5B,$3F,$C7,$5A,$FF,$19,$6B,$FE,$24,$96,$98,$84,$63,$75,$FF,$25
	.dc.b	$AD,$4E,$63,$FC,$9D,$F1,$7F,$D1,$2B,$77,$FE,$74,$8C,$4E,$B3,$ED
	.dc.b	$18,$9F,$E7,$DC,$8E,$5E,$31,$FE,$38,$1C,$7F,$EF,$9C,$3F,$CB,$8D
	.dc.b	$8E,$33,$C1,$FF,$1F,$F2,$37,$FF,$20,$DF,$F5,$56,$CF,$F5,$56,$BF
	.dc.b	$D5,$5A,$FF,$92,$3A,$1E,$76,$23,$FC,$9A,$D8,$CC,$78,$4C,$9E,$3D
	.dc.b	$7F,$D2,$27,$FA,$4E,$BF,$CC,$DE,$5F,$FA,$56,$3F,$64,$68,$AE,$63
	.dc.b	$D1,$3F,$B4,$C7,$D8,$D8,$FD,$AC,$61,$FF,$39,$7B,$6F,$89,$FE,$05
	.dc.b	$82,$31,$7E,$06,$D7,$0A,$DF,$E2,$3F,$EC,$E8,$FF,$3E,$36,$BF,$D0
	.dc.b	$8D,$7F,$CD,$B6,$8C,$47,$D2,$9F,$29,$1F,$41,$E1,$F0,$BC,$B8,$1A
	.dc.b	$48,$FE,$2D,$C6,$F8,$D7,$A7,$FE,$7C,$F7,$3F,$BD,$46,$29,$18,$9D
	.dc.b	$47,$F7,$99,$D4,$FC,$CC,$64,$FE,$0C,$F2,$BC,$57,$C4,$52,$3C,$EF
	.dc.b	$FD,$1B,$1F,$43,$E4,$3F,$D1,$FD,$36,$BF,$5A,$7F,$AA,$F4,$4E,$A3
	.dc.b	$0D,$7F,$24,$78,$8F,$26,$1E,$67,$E7,$CF,$27,$A9,$89,$E4,$FF,$B8
	.dc.b	$63,$55,$E3,$3F,$DC,$FF,$60,$F9,$F7,$4F,$F4,$78,$C3,$63,$37,$FD
	.dc.b	$1E,$72,$F9,$4F,$3E,$3E,$57,$F0,$07,$F4,$DF,$0F,$FD,$7D,$EB,$FF
	.dc.b	$10,$6F,$1F,$ED,$DB,$D3,$FE,$BD,$9D,$7A,$0F,$BA,$FF,$7D,$7E,$07
	.dc.b	$FE,$C1,$6E,$3F,$EB,$F6,$BF,$EB,$EC,$3F,$D9,$91,$88,$FF,$5E,$F9
	.dc.b	$76,$7F,$DA,$0D,$BF,$F4,$06,$4C,$3F,$D9,$D3,$C6,$3F,$51,$E5,$FF
	.dc.b	$B2,$BE,$33,$F5,$AD,$C7,$EE,$35,$FA,$AD,$7E,$97,$D2,$5A,$FD,$93
	.dc.b	$78,$CD,$31,$E8,$7F,$B6,$3E,$37,$F7,$6D,$BF,$C5,$35,$FC,$1B,$5F
	.dc.b	$F6,$FF,$CC,$FF,$B4,$E7,$F9,$54,$63,$D6,$D2,$9D,$3A,$6B,$F8,$EF
	.dc.b	$92,$E9,$97,$F0,$96,$E4,$69,$E1,$E0,$FF,$7E,$7C,$C7,$DA,$25,$7D
	.dc.b	$FE,$14,$7D,$EE,$59,$39,$6F,$2F,$FB,$FF,$DD,$4B,$F1,$FC,$37,$4F
	.dc.b	$F8,$01,$3A,$9C,$9F,$03,$26,$4F,$77,$57,$FD,$36,$73,$18,$8F,$C5
	.dc.b	$4E,$F0,$ED,$18,$9D,$2E,$97,$EE,$9F,$D6,$E0,$65,$C3,$7C,$EF,$EE
	.dc.b	$33,$C1,$FE,$E1,$E6,$FF,$5B,$9F,$EC,$0D,$7F,$D4,$DD,$DF,$D9,$76
	.dc.b	$FE,$C2,$D8,$FE,$CD,$6F,$AF,$FE,$55,$1F,$D8,$E3,$1E,$3F,$F5,$C6
	.dc.b	$C6,$67,$77,$F9,$C8,$7C,$FD,$23,$F9,$E3,$27,$F1,$BE,$AF,$7C,$7F
	.dc.b	$50,$6C,$64,$D9,$FC,$61,$CB,$FC,$6B,$E7,$BF,$B7,$EF,$FD,$55,$BC
	.dc.b	$E6,$9F,$DD,$38,$BE,$DF,$2F,$E1,$3F,$03,$FB,$DC,$9F,$DF,$F0,$46
	.dc.b	$27,$58,$FD,$FA,$D4,$FF,$22,$E3,$FF,$0C,$37,$8F,$75,$F6,$B8,$3E
	.dc.b	$B6,$3D,$23,$E7,$DF,$07,$FC,$23,$7F,$53,$D1,$D4,$7D,$AE,$3F,$F0
	.dc.b	$E4,$E4,$FE,$63,$C5,$F5,$9B,$C7,$F8,$86,$D7,$23,$F1,$D3,$B7,$D5
	.dc.b	$F1,$1F,$A4,$F9,$2F,$82,$D8,$CE,$69,$CB,$63,$F0,$F5,$F8,$69,$EE
	.dc.b	$48,$CE,$1F,$9A,$8E,$3F,$81,$49,$EA,$D7,$E3,$57,$E4,$23,$D2,$48
	.dc.b	$E6,$1F,$8D,$C1,$23,$FC,$72,$D7,$FA,$EF,$E1,$7F,$E0,$42,$3F,$E1
	.dc.b	$77,$FC,$0F,$F8,$3F,$EB,$BE,$5B,$FC,$3E,$D9,$FE,$DC,$D7,$FB,$3B
	.dc.b	$5F,$E1,$7F,$22,$FA,$0F,$F1,$2B,$63,$FC,$66,$DC,$3F,$94,$EA,$8C
	.dc.b	$4F,$A0,$87,$DF,$E1,$FF,$54,$F6,$7F,$6C,$7F,$C7,$BF,$69,$FE,$5B
	.dc.b	$C3,$9B,$6B,$FC,$BD,$82,$31,$1F,$E2,$C6,$BF,$C0,$AB,$AF,$1F,$F9
	.dc.b	$9E,$FE,$44,$E5,$EB,$F4,$5F,$E4,$0B,$34,$93,$FD,$D7,$A1,$F2,$5F
	.dc.b	$CB,$B7,$FD,$5B,$67,$F5,$0D,$7F,$92,$B5,$FE,$6B,$EA,$FF,$36,$7F
	.dc.b	$CB,$6D,$C7,$E4,$B6,$3D,$ED,$7E,$44,$7A,$9F,$F3,$FE,$5E,$D7,$E6
	.dc.b	$9F,$D0,$B2,$7C,$DF,$B4,$91,$F3,$9F,$F2,$FF,$E0,$0F,$33,$FA,$4F
	.dc.b	$17,$96,$BF,$A5,$CE,$78,$3F,$04,$7D,$27,$13,$B8,$46,$1F,$1C,$F9
	.dc.b	$B8,$3E,$99,$33,$FB,$BF,$51,$FE,$F7,$AF,$BB,$9A,$31,$7F,$75,$AF
	.dc.b	$B5,$B3,$E0,$CB,$FB,$0F,$9B,$E3,$CB,$D0,$D8,$CC,$FA,$5B,$7F,$30
	.dc.b	$F3,$F1,$FE,$4F,$2F,$C3,$D7,$E3,$FE,$FB,$FA,$BE,$1E,$FC,$D1,$89
	.dc.b	$F4,$B7,$0F,$CA,$35,$FE,$EB,$F1,$DF,$E6,$67,$FC,$EB,$3F,$9C,$8C
	.dc.b	$7B,$3F,$2D,$B1,$9F,$67,$FB,$3C,$FD,$5F,$13,$FB,$BF,$C5,$7F,$80
	.dc.b	$3D,$9C,$74,$A3,$18,$7F,$3B,$6E,$5F,$DE,$1A,$FE,$8C,$F8,$3C,$DF
	.dc.b	$8B,$7E,$21,$FE,$98,$CF,$D1,$87,$97,$3F,$F3,$CE,$1E,$5A,$FF,$71
	.dc.b	$1B,$8C,$BC,$8E,$C5,$3F,$09,$BF,$E6,$F8,$3E,$84,$78,$B9,$3F,$D7
	.dc.b	$2F,$FB,$77,$E1,$BF,$2D,$F0,$FF,$E0,$8D,$73,$18,$9F,$81,$86,$F9
	.dc.b	$17,$B8,$C3,$67,$FC,$23,$5C,$6F,$9D,$F0,$9E,$66,$9C,$46,$AB,$94
	.dc.b	$FE,$74,$FE,$B8,$43,$FD,$DE,$53,$F9,$01,$4F,$E2,$C7,$F2,$8D,$7F
	.dc.b	$09,$E1,$FC,$D8,$FF,$5F,$6B,$F3,$63,$07,$ED,$63,$C0,$76,$21,$D8
	.dc.b	$FC,$04,$72,$7F,$00,$D5,$C3,$F8,$A8,$C3,$67,$C5,$84,$9F,$BC,$DA
	.dc.b	$FC,$E9,$FE,$F5,$2C,$7E,$97,$8F,$FD,$BE,$FE,$92,$34,$7C,$03,$CE
	.dc.b	$53,$D0,$7F,$B9,$B9,$BF,$88,$47,$EE,$4E,$27,$EF,$CA,$E9,$FE,$13
	.dc.b	$CF,$FD,$1A,$7F,$84,$1F,$EF,$02,$9F,$B9,$74,$F8,$48,$7C,$C3,$C6
	.dc.b	$7F,$22,$D3,$FC,$DB,$BF,$8C,$95,$FC,$2C,$3F,$C1,$7A,$7F,$D1,$75
	.dc.b	$FD,$55,$AF,$F3,$56,$CF,$F5,$B7,$31,$E1,$3C,$7C,$3F,$D9,$5F,$C4
	.dc.b	$4F,$FB,$F4,$E3,$D5,$F8,$38,$FF,$35,$C6,$5F,$0B,$5F,$EA,$EB,$E5
	.dc.b	$A9,$23,$FA,$CC,$65,$FE,$67,$DD,$E6,$EA,$FE,$CB,$1D,$1A,$FF,$6A
	.dc.b	$3D,$B8,$79,$DF,$BA,$7A,$CF,$65,$7F,$79,$7A,$10,$E8,$7C,$28,$E4
	.dc.b	$FD,$23,$93,$87,$FC,$2B,$D9,$F2,$BB,$FF,$C2,$91,$FE,$12,$6B,$FE
	.dc.b	$08,$6B,$FE,$1E,$3D,$C5,$C7,$FC,$4A,$E0,$FF,$55,$F9,$6F,$07,$FD
	.dc.b	$2F,$3F,$E1,$E8,$78,$5F,$69,$B5,$F4,$1A,$EE,$3D,$9D,$BF,$86,$8F
	.dc.b	$E0,$47,$B1,$C3,$FE,$42,$F7,$FF,$CB,$A4,$70,$7F,$96,$1F,$F2,$B9
	.dc.b	$1C,$1F,$E4,$AF,$98,$FF,$32,$91,$C1,$FE,$60,$7F,$CB,$E4,$70,$7F
	.dc.b	$94,$BC,$BF,$F3,$69,$1C,$1F,$EE,$6D,$7F,$C7,$11,$FE,$7B,$24,$66
	.dc.b	$FF,$47,$1F,$E8,$7D,$BF,$D8,$53,$FE,$84,$F3,$3F,$DA,$F3,$92,$47
	.dc.b	$FA,$47,$BF,$2D,$AE,$2F,$FB,$42,$32,$D8,$CB,$F6,$3C,$DF,$F6,$6F
	.dc.b	$7F,$D7,$6C,$7F,$92,$B6,$FF,$54,$3F,$CF,$CF,$F5,$71,$4F,$C0,$F9
	.dc.b	$DF,$ED,$6F,$8A,$FF,$5C,$41,$A4,$7F,$D8,$0F,$FB,$14,$8E,$0F,$F3
	.dc.b	$ED,$FF,$A6,$08,$E0,$F9,$1D,$3E,$17,$FC,$51,$1F,$EA,$72,$3F,$27
	.dc.b	$1F,$36,$7F,$EB,$09,$F9,$C4,$9F,$F5,$D7,$A4,$A7,$D9,$7F,$B0,$A3
	.dc.b	$0D,$8C,$CF,$06,$BF,$EF,$EF,$57,$6F,$FB,$15,$AE,$63,$6F,$F4,$CF
	.dc.b	$21,$3A,$8C,$6F,$E3,$C7,$FD,$A1,$7F,$3F,$C4,$FF,$6B,$B7,$0F,$F6
	.dc.b	$B4,$FF,$B6,$3D,$6E,$18,$7F,$B5,$5A,$8F,$03,$86,$D7,$DB,$1C,$0F
	.dc.b	$5F,$87,$C2,$FE,$ED,$B5,$F8,$56,$BF,$CB,$27,$82,$9F,$E7,$BE,$CB
	.dc.b	$DA,$DC,$F8,$E3,$0E,$CF,$31,$98,$DA,$8D,$CF,$F0,$DF,$3F,$9E,$3E
	.dc.b	$86,$E0,$69,$8E,$5F,$85,$F2,$7F,$96,$26,$3F,$9F,$8C,$BF,$A1,$F1
	.dc.b	$AB,$ED,$6B,$FC,$81,$B8,$7C,$87,$3C,$0E,$07,$BD,$F8,$DF,$BF,$F6
	.dc.b	$E1,$F8,$18,$C3,$F0,$2B,$84,$62,$3E,$C4,$63,$05,$72,$8E,$DF,$03
	.dc.b	$BA,$FC,$63,$63,$35,$C2,$7B,$60,$AD,$2F,$82,$FC,$25,$FA,$5B,$19
	.dc.b	$BF,$C8,$47,$E2,$6F,$EF,$72,$FF,$74,$E0,$FF,$29,$C7,$B9,$AF,$DC
	.dc.b	$6A,$E9,$7C,$1B,$BF,$FC,$03,$18,$6C,$66,$7F,$C0,$F3,$97,$FA,$57
	.dc.b	$C2,$7E,$77,$D1,$FE,$A0,$D8,$FE,$34,$D7,$F2,$CD,$7F,$3B,$8F,$A4
	.dc.b	$8F,$F8,$4B,$1F,$3E,$7F,$A4,$AE,$0D,$95,$F8,$1A,$78,$0A,$7E,$C7
	.dc.b	$C4,$79,$8F,$5B,$F6,$9F,$25,$F8,$49,$CC,$62,$BF,$47,$18,$53,$BA
	.dc.b	$E5,$18,$3E,$86,$7E,$A7,$FC,$5B,$9F,$D0,$7E,$B9,$FE,$30,$BE,$9C
	.dc.b	$BC,$F1,$FE,$0C,$D9,$FF,$70,$1E,$2C,$3B,$C7,$F7,$DC,$BF,$BE,$12
	.dc.b	$7F,$BE,$3E,$17,$C6,$7C,$BC,$70,$FF,$8C,$BE,$1F,$F6,$DA,$7F,$8C
	.dc.b	$9B,$1F,$A6,$84,$62,$3F,$82,$C6,$3A,$7F,$C7,$8F,$F1,$8F,$88,$FF
	.dc.b	$07,$12,$7F,$C1,$CF,$81,$FB,$E7,$8C,$87,$7F,$89,$FF,$38,$BF,$E4
	.dc.b	$66,$BF,$D1,$5B,$27,$30,$F8,$A7,$C5,$FF,$6A,$8F,$E3,$71,$87,$FA
	.dc.b	$C3,$7C,$3F,$F3,$0B,$63,$E3,$C7,$D8,$CF,$94,$7A,$23,$0F,$F6,$1F
	.dc.b	$2F,$F5,$1E,$45,$E8,$FF,$36,$D7,$F6,$2A,$FE,$57,$5A,$8F,$F3,$7F
	.dc.b	$C6,$3C,$CF,$DD,$37,$5F,$ED,$8D,$9D,$CF,$EF,$F9,$5E,$ED,$25,$FF
	.dc.b	$73,$8C,$B7,$E3,$78,$11,$C1,$FD,$E6,$9F,$1D,$FE,$7C,$6E,$7F,$E8
	.dc.b	$26,$BF,$E8,$66,$BF,$CA,$9A,$FF,$48,$6B,$FD,$7D,$AF,$BE,$3B,$11
	.dc.b	$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B1,$D0,$87,$42,$9D,$0A,$74
	.dc.b	$29,$D0,$A7,$43,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0,$87,$02,$9C
	.dc.b	$0A,$70,$34,$43,$45,$34,$60,$86,$09,$1F,$50,$91,$FE,$BC,$F7,$9C
	.dc.b	$FF,$E3,$DA,$46,$33,$FF,$62,$B6,$3F,$D9,$2D,$7F,$09,$E9,$7F,$87
	.dc.b	$F8,$7F,$10,$D8,$C9,$FD,$53,$1F,$F4,$9C,$EA,$31,$3B,$7F,$18,$D7
	.dc.b	$F1,$8D,$7F,$54,$D9,$E7,$1D,$A7,$A1,$1C,$10,$EA,$78,$08,$ED,$CB
	.dc.b	$5F,$09,$A3,$91,$87,$4F,$90,$D1,$C8,$D9,$97,$71,$E8,$8D,$10,$E8
	.dc.b	$70,$74,$E0,$D1,$CC,$D4,$FC,$32,$3E,$23,$81,$0E,$07,$28,$EA,$E0
	.dc.b	$D3,$82,$47,$49,$F4,$B6,$7F,$C5,$BF,$49,$FE,$59,$CD,$F1,$5F,$E3
	.dc.b	$09,$4E,$63,$0D,$8E,$3E,$6F,$FA,$0E,$78,$7C,$8B,$E4,$71,$5F,$E3
	.dc.b	$A8,$C6,$1D,$A3,$13,$A7,$FC,$B7,$3A,$AF,$17,$27,$F3,$86,$F2,$B9
	.dc.b	$B9,$5F,$F3,$88,$C4,$EA,$73,$3F,$E7,$BF,$C4,$FF,$9C,$7C,$5F,$B5
	.dc.b	$B1,$FE,$4E,$6B,$FE,$6C,$6B,$FE,$83,$C7,$D3,$3E,$EC,$BF,$D4,$4D
	.dc.b	$CC,$D3,$FA,$89,$2B,$FB,$EB,$5F,$F4,$74,$3E,$9F,$FB,$EF,$07,$F7
	.dc.b	$B6,$CF,$F7,$A6,$BC,$1A,$F9,$6F,$FD,$69,$3E,$32,$47,$CE,$AF,$F2
	.dc.b	$78,$FA,$7F,$F1,$8C,$FF,$82,$5B,$3F,$E0,$86,$BE,$66,$BF,$EC,$1B
	.dc.b	$FF,$1A,$57,$46,$D7,$A7,$C3,$F1,$D7,$A5,$AF,$DE,$C2,$BE,$9B,$FF
	.dc.b	$30,$E5,$F8,$6B,$F4,$CF,$C6,$F7,$DE,$FF,$F0,$DB,$46,$27,$DA,$DA
	.dc.b	$FF,$12,$36,$3F,$C8,$BF,$99,$3E,$D7,$FD,$7D,$F8,$2F,$F9,$2E,$31
	.dc.b	$28,$C4,$EA,$38,$B5,$F5,$61,$E8,$CF,$D7,$0C,$FD,$4F,$D1,$CB,$9B
	.dc.b	$FE,$52,$3F,$CB,$D8,$F8,$34,$C9,$23,$FD,$43,$F3,$2F,$77,$F9,$53
	.dc.b	$63,$FD,$0E,$DB,$FD,$C3,$63,$C0,$43,$CF,$1F,$31,$FF,$6F,$46,$DF
	.dc.b	$F6,$5B,$5C,$CF,$80,$FA,$1A,$BE,$2F,$FD,$CD,$F1,$FF,$39,$B5,$FE
	.dc.b	$50,$9D,$29,$D3,$4F,$21,$F8,$1C,$3C,$8F,$C9,$D3,$F0,$BA,$BE,$73
	.dc.b	$FC,$61,$5F,$E5,$F6,$D7,$F9,$79,$AF,$EB,$1A,$A5,$F1,$5F,$E2,$A5
	.dc.b	$F0,$DF,$E6,$98,$C3,$7C,$3E,$0D,$9D,$D7,$FB,$13,$E8,$72,$D8,$CB
	.dc.b	$FE,$D2,$F5,$D3,$A8,$C3,$57,$B1,$FE,$E1,$FA,$6F,$37,$D3,$FF,$B5
	.dc.b	$ED,$18,$AF,$F6,$9B,$5E,$CD,$79,$C3,$E9,$71,$7F,$EE,$D6,$C6,$71
	.dc.b	$61,$DE,$79,$47,$FB,$CB,$6F,$91,$F9,$57,$FA,$37,$DF,$46,$32,$FF
	.dc.b	$46,$C6,$29,$5B,$AD,$50,$9F,$44,$F8,$7C,$C6,$1A,$A3,$57,$D2,$39
	.dc.b	$1B,$70,$FF,$2A,$C1,$F0,$FD,$25,$18,$77,$3F,$83,$8C,$63,$ED,$AE
	.dc.b	$1C,$CF,$9F,$FC,$DD,$FF,$11,$6B,$F8,$B8,$D3,$F9,$B3,$45,$35,$F6
	.dc.b	$1C,$31,$FE,$2E,$D9,$34,$EB,$C0,$E0,$FE,$25,$4F,$C5,$3F,$09,$E0
	.dc.b	$FE,$6D,$B1,$9C,$7E,$2D,$F1,$AD,$46,$2D,$F5,$58,$8F,$CB,$35,$FC
	.dc.b	$4B,$6B,$C4,$E7,$E0,$BC,$FE,$6F,$EA,$FB,$7E,$9D,$A7,$39,$7C,$16
	.dc.b	$CE,$9B,$3C,$3C,$DF,$F5,$9C,$FF,$77,$24,$67,$CD,$FE,$FD,$1C,$76
	.dc.b	$7C,$77,$D2,$C3,$E4,$B6,$33,$7E,$16,$D7,$FB,$97,$A5,$F2,$DF,$93
	.dc.b	$FD,$6F,$0D,$BF,$F7,$3C,$E9,$4E,$9A,$6B,$DB,$E0,$FF,$7A,$9F,$F6
	.dc.b	$E3,$A3,$11,$A7,$F7,$AD,$D7,$FA,$CF,$2F,$E9,$7E,$F3,$FA,$BB,$6B
	.dc.b	$F9,$53,$5F,$C6,$C7,$F0,$5F,$2F,$E9,$47,$DD,$6B,$FC,$45,$AF,$F0
	.dc.b	$46,$ED,$F5,$A7,$E6,$70,$FF,$25,$6E,$06,$9F,$E2,$B2,$5F,$D5,$F4
	.dc.b	$74,$D8,$FA,$DC,$1E,$78,$FC,$27,$07,$D4,$9C,$C7,$03,$F8,$26,$3F
	.dc.b	$69,$C0,$F9,$27,$E1,$5F,$63,$27,$DE,$FE,$42,$73,$C5,$FC,$F6,$31
	.dc.b	$5B,$6C,$66,$FF,$84,$E1,$DE,$7C,$64,$8C,$E5,$F8,$A8,$E3,$8B,$EF
	.dc.b	$69,$2B,$F8,$9E,$5E,$29,$D4,$72,$D3,$F9,$05,$7C,$02,$46,$4D,$9F
	.dc.b	$E5,$9D,$7F,$96,$BF,$37,$C2,$FF,$2D,$11,$FD,$EF,$E5,$5F,$E6,$1C
	.dc.b	$37,$1F,$CB,$7A,$93,$9A,$FD,$AC,$62,$FF,$C5,$ED,$BE,$37,$FD,$4A
	.dc.b	$DF,$15,$FD,$9F,$EC,$B8,$C6,$29,$18,$9D,$57,$91,$AF,$E8,$FC,$3F
	.dc.b	$F1,$B7,$47,$F3,$28,$FF,$28,$B7,$3F,$DA,$4F,$FB,$17,$B1,$D9,$FE
	.dc.b	$7D,$9F,$81,$7F,$15,$AF,$9F,$D8,$F5,$E8,$69,$34,$F6,$CB,$0E,$BB
	.dc.b	$61,$B8,$74,$9D,$63,$E1,$7E,$7F,$C3,$76,$26,$1F,$23,$7F,$9F,$1F
	.dc.b	$E1,$6D,$7F,$AE,$53,$F0,$9F,$E2,$1E,$47,$F6,$16,$C6,$4F,$E1,$F1
	.dc.b	$97,$FA,$EE,$0F,$9A,$7D,$8F,$F9,$7D,$B3,$FE,$7A,$6E,$9F,$CC,$AB
	.dc.b	$85,$7C,$BD,$3D,$57,$F2,$4F,$C6,$10,$FC,$47,$03,$E6,$BE,$59,$2F
	.dc.b	$DD,$19,$7F,$D1,$3F,$80,$7F,$65,$7F,$AC,$9E,$87,$1D,$5F,$E2,$FB
	.dc.b	$FF,$4E,$C6,$27,$5E,$0F,$F8,$CE,$B5,$5F,$E0,$5E,$2F,$A6,$F9,$4C
	.dc.b	$65,$EA,$6F,$1F,$93,$E3,$BF,$D3,$F8,$FE,$91,$B3,$EC,$6B,$F9,$D6
	.dc.b	$C3,$2F,$57,$85,$E2,$8C,$67,$FE,$47,$8C,$4E,$A7,$D2,$7F,$A2,$4F
	.dc.b	$35,$FE,$23,$0F,$C6,$63,$E7,$8C,$6F,$FD,$47,$CF,$FF,$11,$63,$DF
	.dc.b	$93,$FC,$7B,$5C,$CF,$E9,$53,$AC,$7F,$A9,$B6,$3E,$BF,$C8,$FF,$96
	.dc.b	$FE,$2B,$E4,$F4,$7F,$00,$8C,$5F,$F0,$66,$D7,$08,$57,$87,$CB,$FF
	.dc.b	$1A,$67,$6B,$FC,$74,$7F,$5B,$8C,$D7,$F6,$46,$C7,$E2,$C8,$7D,$49
	.dc.b	$FD,$AE,$1E,$3A,$E7,$3F,$C3,$70,$FA,$73,$9A,$F4,$F1,$A7,$EB,$D7
	.dc.b	$C8,$3E,$6F,$17,$E1,$1E,$E6,$4A,$E5,$F1,$1F,$E1,$4F,$86,$7C,$43
	.dc.b	$E2,$3F,$CA,$5E,$37,$F2,$89,$CC,$63,$2F,$F2,$6C,$62,$FF,$C0,$F1
	.dc.b	$8B,$E5,$18,$3F,$CB,$93,$BB,$E9,$3A,$6D,$F5,$AD,$CE,$6F,$FA,$04
	.dc.b	$7F,$8A,$8F,$C5,$17,$0F,$E2,$ED,$E2,$71,$EE,$36,$7F,$73,$34,$E6
	.dc.b	$78,$C7,$E0,$F6,$FF,$39,$15,$FA,$3A,$7E,$DE,$30,$73,$88,$79,$8A
	.dc.b	$7F,$A5,$B7,$FE,$2E,$FE,$1C,$EA,$53,$F1,$87,$68,$73,$52,$3E,$57
	.dc.b	$F4,$9C,$5F,$E8,$7F,$C0,$2D,$1A,$4A,$F1,$C7,$C2,$3C,$C4,$3F,$D8
	.dc.b	$1C,$DF,$C5,$E7,$F7,$7C,$FF,$EC,$36,$D7,$53,$F7,$AF,$E5,$30,$F8
	.dc.b	$97,$FC,$2E,$7D,$AF,$F0,$B6,$F2,$7F,$70,$9D,$CF,$F1,$63,$E9,$90
	.dc.b	$F9,$33,$FC,$38,$7C,$BF,$88,$8F,$F0,$E7,$F4,$BD,$5F,$C9,$A7,$F9
	.dc.b	$BB,$5F,$EF,$0D,$BF,$F0,$0B,$98,$F8,$87,$8C,$F8,$FB,$7E,$11,$FC
	.dc.b	$1F,$67,$F4,$C7,$FA,$53,$5F,$EA,$6D,$77,$D9,$F9,$1C,$7F,$D9,$3D
	.dc.b	$BF,$EB,$DF,$D8,$FF,$AE,$92,$BF,$AE,$87,$FC,$17,$E0,$79,$FA,$CD
	.dc.b	$25,$FE,$67,$9B,$F3,$F7,$F2,$4F,$B8,$7B,$35,$FF,$0F,$F8,$3C,$C9
	.dc.b	$5F,$AB,$8C,$BC,$F9,$E5,$73,$3E,$21,$FE,$17,$3F,$65,$1F,$EE,$4F
	.dc.b	$0B,$DD,$DF,$FE,$19,$9F,$F0,$B9,$D7,$3F,$F1,$27,$D8,$FF,$88,$EC
	.dc.b	$D2,$4F,$F8,$E3,$0F,$80,$49,$E4,$FD,$7E,$1F,$A9,$3E,$42,$1F,$6B
	.dc.b	$CA,$FF,$33,$7D,$FF,$F9,$A8,$91,$FE,$4B,$7F,$CA,$E4,$7F,$CA,$BE
	.dc.b	$67,$F9,$A6,$31,$1F,$E7,$0F,$17,$FD,$4F,$F2,$FF,$E7,$42,$47,$F9
	.dc.b	$55,$FF,$30,$91,$E7,$E1,$B3,$4E,$67,$A7,$97,$FE,$FF,$FB,$EF,$F3
	.dc.b	$F9,$23,$FC,$C0,$FF,$16,$23,$FE,$8B,$6B,$E5,$F4,$3B,$47,$FA,$50
	.dc.b	$91,$FE,$76,$8C,$BF,$EA,$1B,$7D,$EF,$FA,$95,$B1,$98,$C1,$FE,$AE
	.dc.b	$55,$48,$FF,$AC,$0C,$9F,$EB,$F1,$F8,$CF,$FB,$2F,$D3,$46,$0D,$46
	.dc.b	$5C,$1F,$EC,$5C,$F7,$1F,$EC,$B2,$47,$FB,$42,$32,$FF,$A1,$6F,$FD
	.dc.b	$4E,$47,$07,$FA,$C5,$FE,$7E,$47,$07,$C5,$C3,$E0,$BF,$2D,$AF,$FA
	.dc.b	$D3,$1F,$B5,$3F,$EC,$03,$FD,$0F,$EB,$BD,$7F,$BD,$F8,$9E,$1B,$7F
	.dc.b	$00,$74,$AB,$86,$2B,$E0,$DA,$E1,$5B,$F2,$39,$C7,$FA,$66,$D1,$88
	.dc.b	$EB,$18,$AE,$51,$B9,$D4,$3D,$A7,$17,$D4,$CF,$FC,$C9,$18,$BF,$83
	.dc.b	$5B,$AD,$47,$F9,$C2,$75,$5F,$85,$24,$7E,$0D,$AF,$F6,$AE,$B7,$B7
	.dc.b	$FA,$3F,$CA,$BF,$CE,$51,$8C,$51,$89,$D5,$7F,$9F,$AB,$53,$FE,$01
	.dc.b	$F7,$9E,$1E,$AB,$E7,$46,$30,$E7,$1F,$44,$46,$6B,$95,$71,$9F,$DB
	.dc.b	$5B,$CE,$C6,$08,$C3,$D2,$7D,$93,$A9,$CD,$FE,$13,$DF,$FE,$01,$F7
	.dc.b	$37,$03,$49,$87,$DE,$E0,$79,$FC,$23,$0D,$7F,$00,$D7,$EF,$B7,$8B
	.dc.b	$F1,$9F,$03,$FC,$3F,$7F,$CA,$B6,$BB,$35,$FC,$53,$5F,$E2,$DF,$08
	.dc.b	$F8,$4C,$66,$8C,$3F,$6A,$7F,$D6,$31,$8B,$FF,$56,$36,$3F,$D6,$BF
	.dc.b	$0A,$F9,$4F,$EE,$56,$8C,$3F,$D2,$9B,$3F,$EB,$56,$BF,$67,$E3,$5F
	.dc.b	$11,$FD,$DD,$B5,$FE,$3D,$6C,$FF,$91,$5A,$FE,$5F,$E2,$7F,$32,$FF
	.dc.b	$8F,$3A,$11,$88,$FE,$6D,$18,$7F,$42,$DE,$8F,$F2,$87,$C5,$BE,$2F
	.dc.b	$FC,$99,$48,$C3,$FE,$4F,$6B,$C2,$37,$87,$92,$3F,$82,$35,$FE,$4A
	.dc.b	$D4,$D9,$E7,$1F,$9B,$6B,$FC,$35,$AF,$F2,$E6,$FC,$B3,$E3,$9F,$03
	.dc.b	$FE,$7C,$AF,$E2,$93,$98,$C4,$7F,$36,$8C,$2F,$0F,$F5,$51,$87,$E7
	.dc.b	$6D,$F4,$C9,$B7,$D3,$7F,$C6,$13,$CB,$F0,$DE,$38,$F8,$B0,$46,$31
	.dc.b	$FE,$8A,$D8,$C9,$FC,$3F,$E4,$3F,$C5,$5B,$9A,$48,$F0,$65,$1F,$2B
	.dc.b	$5F,$E6,$4B,$E4,$5F,$23,$F9,$46,$C7,$F8,$71,$B3,$FC,$26,$38,$4F
	.dc.b	$E6,$BE,$4B,$FC,$EB,$3F,$E2,$26,$BB,$6B,$E0,$6B,$F8,$0D,$7E,$3B
	.dc.b	$E6,$21,$E6,$B7,$CD,$F9,$49,$87,$95,$FF,$2C,$7C,$F7,$F1,$7A,$7A
	.dc.b	$5F,$EA,$58,$C3,$7C,$4F,$EE,$2D,$9D,$9F,$AD,$C7,$FC,$2D,$1F,$20
	.dc.b	$91,$9C,$36,$FF,$5B,$F6,$3F,$73,$3F,$EB,$36,$C7,$F2,$F6,$BB,$EA
	.dc.b	$EA,$FF,$33,$F5,$1E,$CF,$FA,$DB,$E1,$BF,$D6,$64,$9D,$46,$5F,$14
	.dc.b	$7D,$C2,$38,$3F,$9F,$7C,$BB,$D0,$FF,$18,$4F,$FB,$1E,$31,$9F,$FB
	.dc.b	$52,$BF,$BA,$46,$27,$55,$C2,$7F,$A4,$62,$F9,$8F,$DF,$B6,$3F,$D9
	.dc.b	$8D,$7F,$DA,$8D,$78,$C7,$02,$3C,$1A,$F0,$6B,$C0,$D1,$0D,$14,$D1
	.dc.b	$82,$18,$21,$FE,$11,$8F,$23,$63,$BB,$4E,$85,$39,$94,$E6,$7F,$8E
	.dc.b	$A3,$9B,$63,$9B,$4E,$65,$39,$94,$E6,$7F,$45,$78,$B4,$E2,$53,$89
	.dc.b	$4E,$31,$E2,$70,$79,$C8,$77,$29,$A3,$04,$30,$47,$A1,$D7,$E2,$5F
	.dc.b	$13,$F8,$99,$FF,$1E,$46,$30,$FF,$07,$47,$F9,$7A,$30,$E1,$AF,$F8
	.dc.b	$4F,$F3,$1F,$F7,$B3,$FB,$93,$FC,$8C,$E3,$87,$9B,$63,$87,$C3,$FF
	.dc.b	$9A,$A7,$2D,$8C,$9C,$CF,$F3,$7C,$F1,$7A,$C6,$DA,$6C,$46,$DE,$31
	.dc.b	$B6,$9B,$8D,$3C,$23,$4D,$34,$E2,$36,$47,$04,$8D,$92,$7D,$AD,$3D
	.dc.b	$AE,$08,$E0,$87,$B1,$CB,$4C,$9E,$A7,$2D,$32,$7A,$5C,$B4,$CC,$FA
	.dc.b	$5F,$21,$D4,$87,$32,$9C,$4A,$F1,$8E,$AD,$8E,$B3,$F0,$E7,$C3,$E0
	.dc.b	$7F,$87,$BE,$67,$FA,$D9,$2B,$F9,$A3,$F9,$0E,$1F,$E5,$7E,$F2,$31
	.dc.b	$C4,$8C,$57,$F2,$C6,$BF,$E9,$5A,$E3,$5F,$D1,$49,$1F,$E6,$9E,$6F
	.dc.b	$E9,$D1,$88,$FE,$53,$C3,$FE,$DD,$F9,$9F,$E5,$3C,$86,$92,$7C,$0F
	.dc.b	$93,$1D,$46,$08,$7F,$A9,$FD,$A7,$CE,$7F,$B2,$2B,$F5,$B1,$89,$E2
	.dc.b	$6E,$74,$FF,$9F,$E3,$53,$FB,$4F,$3B,$FA,$7E,$1F,$9F,$6E,$F9,$DB
	.dc.b	$FD,$EC,$D9,$FD,$26,$DF,$32,$3D,$04,$C7,$E5,$3F,$C0,$3E,$E5,$97
	.dc.b	$E8,$9B,$3F,$AD,$6C,$EE,$7F,$4F,$8F,$0C,$3C,$91,$E2,$BF,$F4,$C4
	.dc.b	$E6,$BD,$53,$99,$FB,$18,$FD,$D9,$F2,$46,$34,$F4,$CE,$63,$FD,$45
	.dc.b	$8F,$F7,$07,$C9,$1E,$72,$3F,$C0,$FC,$7F,$F3,$F7,$B3,$FE,$60,$F3
	.dc.b	$BF,$DF,$2D,$9F,$F6,$89,$FE,$53,$8C,$BF,$10,$87,$8F,$1F,$2B,$FE
	.dc.b	$55,$83,$FE,$32,$39,$14,$FF,$2E,$3A,$F0,$BC,$07,$F2,$EC,$1F,$0B
	.dc.b	$FC,$41,$B5,$FC,$6D,$B7,$B9,$FE,$2B,$BF,$DB,$AF,$29,$E2,$70,$F9
	.dc.b	$CF,$F6,$45,$72,$6C,$F7,$23,$FC,$A7,$CA,$7C,$57,$D5,$8F,$6C,$63
	.dc.b	$87,$53,$99,$FF,$36,$37,$0F,$0C,$78,$F0,$FB,$0F,$8C,$87,$C8,$FB
	.dc.b	$C7,$E2,$5F,$CC,$1B,$1F,$CE,$DB,$9F,$F5,$46,$C7,$F9,$8B,$ED,$DE
	.dc.b	$1F,$D2,$F9,$EF,$F2,$F4,$62,$51,$89,$D4,$71,$6B,$F6,$B4,$57,$D2
	.dc.b	$AE,$7B,$6A,$95,$A8,$C3,$63,$B9,$E5,$F8,$EF,$C2,$FC,$5F,$91,$B1
	.dc.b	$91,$19,$69,$ED,$F8,$FD,$BE,$5C,$7F,$CE,$F3,$EE,$9C,$CF,$92,$72
	.dc.b	$FC,$5F,$92,$FC,$AF,$CF,$7C,$46,$C6,$44,$65,$A7,$E7,$0F,$CE,$FD
	.dc.b	$4B,$EA,$B8,$C6,$63,$0F,$03,$11,$A7,$8B,$73,$F9,$BF,$8F,$5C,$4D
	.dc.b	$25,$73,$3F,$23,$E0,$FE,$3F,$9F,$FC,$C3,$C7,$FE,$16,$6C,$F9,$CF
	.dc.b	$F5,$F4,$FE,$57,$8F,$F0,$F1,$EB,$6B,$F9,$1E,$4F,$C8,$1F,$E0,$5C
	.dc.b	$BF,$03,$E4,$FE,$F3,$E8,$7F,$16,$DA,$FC,$83,$5F,$A8,$D7,$E9,$74
	.dc.b	$DA,$FC,$5B,$75,$FD,$6C,$AF,$F2,$F3,$C6,$3F,$57,$5F,$AE,$26,$9E
	.dc.b	$F9,$FD,$77,$47,$EB,$2B,$F5,$DD,$FF,$C9,$89,$5F,$A7,$CF,$DB,$3F
	.dc.b	$AB,$6C,$7E,$3F,$F5,$49,$3F,$93,$6B,$F8,$FE,$5F,$E3,$B1,$F8,$F8
	.dc.b	$34,$8F,$D1,$E0,$FA,$6F,$E2,$5A,$FD,$AC,$FF,$87,$70,$7D,$5C,$7C
	.dc.b	$73,$F6,$78,$3E,$AA,$6D,$FF,$11,$CB,$E1,$61,$F1,$2B,$C1,$AF,$C4
	.dc.b	$D3,$E0,$C7,$F2,$41,$D7,$FC,$51,$E9,$97,$81,$FC,$23,$5F,$36,$1F
	.dc.b	$8D,$CF,$F2,$04,$C7,$E5,$4C,$9F,$9C,$D3,$E1,$4E,$E3,$94,$FC,$03
	.dc.b	$53,$E0,$29,$F9,$33,$D2,$7F,$26,$3D,$3B,$7D,$F6,$E3,$FE,$65,$97
	.dc.b	$E3,$1F,$EB,$BB,$FF,$7F,$36,$BF,$75,$3A,$F3,$BF,$3D,$39,$9F,$E1
	.dc.b	$9F,$01,$FD,$6B,$E7,$7F,$94,$B6,$3F,$97,$35,$D8,$8F,$E5,$FE,$47
	.dc.b	$EA,$79,$7F,$C3,$99,$62,$BC,$D1,$87,$EA,$6F,$F4,$C9,$AF,$E6,$F4
	.dc.b	$FE,$63,$BF,$32,$57,$27,$FB,$0C,$FC,$3A,$E3,$B7,$C8,$AF,$E4,$5B
	.dc.b	$78,$C8,$F8,$9F,$F4,$17,$8E,$F9,$5F,$E2,$39,$3F,$64,$FF,$0B,$C6
	.dc.b	$1B,$D5,$FD,$01,$B3,$BC,$3E,$6F,$D4,$FF,$44,$7E,$A3,$77,$FE,$29
	.dc.b	$28,$C4,$F0,$8F,$5E,$7E,$98,$FB,$B5,$FE,$0E,$97,$EB,$BF,$CC,$E3
	.dc.b	$9B,$67,$EE,$B5,$FE,$B4,$D7,$FA,$BF,$D5,$3F,$1F,$FF,$22,$DF,$E1
	.dc.b	$23,$0A,$FD,$55,$C2,$B6,$FF,$65,$AD,$CE,$A3,$CE,$7D,$CE,$0F,$A8
	.dc.b	$7D,$BF,$C7,$BF,$C6,$D3,$D3,$04,$61,$FC,$AB,$6B,$FB,$93,$5F,$14
	.dc.b	$79,$4F,$46,$1E,$48,$CD,$75,$C3,$CB,$5E,$3C,$BE,$67,$0F,$F8,$52
	.dc.b	$3E,$5F,$DA,$FF,$B4,$5F,$92,$DC,$3E,$4B,$5F,$EF,$FA,$23,$13,$F5
	.dc.b	$3E,$AF,$FC,$83,$F4,$7E,$DA,$46,$27,$EA,$B6,$FE,$AB,$5E,$AF,$49
	.dc.b	$E3,$1E,$F3,$65,$36,$53,$E8,$9F,$C1,$F0,$FD,$1E,$7F,$86,$AE,$84
	.dc.b	$9E,$57,$E9,$79,$57,$07,$E4,$57,$F9,$79,$EA,$7F,$4E,$71,$EC,$3C
	.dc.b	$37,$B3,$FE,$46,$9F,$F0,$44,$E6,$31,$5F,$E6,$68,$C2,$F0,$3F,$20
	.dc.b	$E6,$F9,$B9,$BE,$71,$89,$DC,$AB,$9C,$61,$B5,$C2,$FC,$35,$FC,$A0
	.dc.b	$FF,$5C,$10,$FF,$65,$14,$FF,$48,$14,$FE,$DA,$7C,$B3,$FC,$8C,$7D
	.dc.b	$43,$F0,$19,$7F,$2E,$71,$97,$46,$1D,$0A,$7D,$A3,$89,$D0,$C3,$F8
	.dc.b	$87,$0F,$D7,$F0,$F6,$47,$4F,$E4,$76,$F1,$4F,$E6,$1B,$3F,$A0,$8E
	.dc.b	$7A,$FE,$6D,$F0,$1F,$14,$F2,$94,$F4,$1F,$C0,$78,$7F,$8F,$4F,$EC
	.dc.b	$8F,$2F,$0F,$F0,$A9,$FD,$94,$E9,$FD,$DF,$1F,$8B,$8F,$FD,$71,$1F
	.dc.b	$C0,$DB,$7F,$C0,$A7,$D3,$3E,$13,$E8,$9E,$4E,$14,$38,$FF,$C8,$13
	.dc.b	$FC,$18,$F0,$D7,$93,$0F,$14,$66,$3F,$93,$BC,$0F,$DE,$95,$FE,$5E
	.dc.b	$DC,$B4,$F8,$8F,$86,$52,$47,$F3,$D3,$97,$27,$F2,$58,$FF,$46,$35
	.dc.b	$FF,$44,$B7,$A3,$FA,$97,$3F,$FA,$33,$F6,$4F,$E9,$A4,$AF,$F3,$94
	.dc.b	$EA,$D3,$FE,$9E,$EA,$F4,$76,$1A,$4B,$F7,$6D,$F2,$35,$BB,$7C,$99
	.dc.b	$FC,$F1,$F3,$F1,$FE,$E2,$FC,$C6,$BF,$EA,$CE,$CE,$E7,$C4,$BE,$91
	.dc.b	$C9,$AF,$E7,$A7,$FD,$57,$DB,$E6,$24,$75,$C9,$1E,$AE,$FF,$F7,$5F
	.dc.b	$CF,$FF,$84,$A8,$D2,$3F,$E1,$CF,$07,$E0,$3E,$76,$BF,$E2,$2D,$7E
	.dc.b	$E4,$62,$3F,$15,$A7,$C1,$6E,$BF,$83,$F1,$3F,$DA,$31,$FE,$3C,$21
	.dc.b	$FE,$55,$DB,$F0,$7D,$AF,$0F,$EE,$37,$7F,$C6,$36,$BF,$C8,$27,$23
	.dc.b	$FC,$80,$53,$FC,$80,$68,$FF,$24,$C7,$29,$EA,$7E,$26,$FC,$B3,$E3
	.dc.b	$3D,$17,$F9,$68,$CC,$7E,$58,$FD,$51,$38,$3E,$AB,$FE,$75,$F1,$3D
	.dc.b	$24,$C3,$D2,$F7,$BE,$84,$9E,$E7,$96,$63,$E9,$3F,$D0,$18,$7E,$86
	.dc.b	$3F,$DD,$7D,$69,$7C,$0F,$FA,$2C,$8F,$FB,$7C,$C9,$FE,$E5,$7F,$DC
	.dc.b	$84,$70,$7C,$13,$E5,$1F,$E7,$C3,$CA,$7E,$4F,$5F,$BC,$FC,$3C,$BE
	.dc.b	$0B,$E3,$D1,$9F,$C1,$8F,$F4,$69,$0F,$F3,$4F,$1F,$FA,$E3,$E6,$3F
	.dc.b	$D6,$F0,$69,$1F,$F6,$29,$92,$BF,$EC,$A7,$07,$FB,$20,$A7,$E0,$8F
	.dc.b	$61,$0F,$C3,$9F,$84,$3E,$71,$FE,$AB,$2E,$7F,$EE,$C5,$F2,$9F,$E5
	.dc.b	$0A,$FF,$76,$C6,$0F,$5F,$47,$D3,$9D,$36,$3F,$DD,$5E,$5F,$FB,$F2
	.dc.b	$7F,$96,$B6,$33,$E5,$ED,$FF,$7F,$9D,$4F,$BD,$83,$07,$9B,$B9,$4E
	.dc.b	$63,$0D,$8E,$46,$A7,$8B,$88,$FC,$5A,$F9,$3F,$F6,$A1,$1F,$C5,$3F
	.dc.b	$07,$D0,$FC,$83,$63,$81,$F7,$23,$2D,$32,$7A,$A9,$E8,$FE,$67,$D5
	.dc.b	$F9,$8D,$9F,$C2,$B5,$E0,$D7,$D5,$EA,$5C,$7E,$23,$6C,$D3,$84,$FE
	.dc.b	$6C,$FC,$EF,$47,$EF,$1F,$C3,$1F,$9B,$E4,$FC,$EB,$FD,$4C,$FD,$FF
	.dc.b	$1F,$E7,$30,$FD,$D9,$23,$F3,$87,$EC,$CF,$ED,$78,$FF,$5E,$AF,$7C
	.dc.b	$7F,$65,$6B,$FD,$B9,$AF,$F2,$83,$F5,$DD,$3B,$9E,$44,$8C,$1F,$DB
	.dc.b	$FD,$76,$9F,$B1,$8F,$10,$8C,$BF,$E4,$53,$2E,$31,$FE,$FE,$D8,$F1
	.dc.b	$10,$F6,$3C,$BE,$23,$75,$FC,$45,$B2,$69,$FE,$0A,$F6,$7F,$C2,$B1
	.dc.b	$E4,$6B,$F1,$DA,$FF,$1B,$6B,$FC,$B7,$2F,$A6,$DC,$BF,$C1,$DE,$1B
	.dc.b	$DA,$FF,$00,$13,$5C,$BF,$E0,$5C,$35,$7F,$E1,$56,$BE,$58,$FF,$09
	.dc.b	$C6,$23,$E0,$1F,$81,$EA,$FE,$3A,$AF,$FC,$4F,$7E,$2C,$7F,$C2,$04
	.dc.b	$7F,$C4,$58,$7F,$72,$A7,$C1,$FF,$8B,$23,$0D,$F1,$3F,$9D,$36,$76
	.dc.b	$7D,$AC,$7F,$C1,$D1,$FE,$34,$6C,$67,$0F,$F1,$C4,$E5,$B1,$F8,$2F
	.dc.b	$83,$FE,$D5,$3F,$E3,$C6,$C6,$7C,$0F,$E1,$33,$96,$C7,$E2,$BD,$FF
	.dc.b	$F4,$5F,$9F,$FE,$44,$6C,$67,$D0,$FE,$A5,$39,$6C,$79,$78,$3F,$D2
	.dc.b	$2F,$E9,$73,$FF,$14,$57,$9E,$73,$3F,$D6,$E7,$2F,$E6,$BE,$01,$F0
	.dc.b	$9F,$C7,$AF,$FC,$AB,$18,$E0,$E3,$48,$C4,$EA,$B8,$4F,$E9,$3E,$0D
	.dc.b	$F0,$7F,$E5,$26,$C7,$F9,$65,$AF,$F9,$89,$AF,$18,$E0,$47,$83,$5E
	.dc.b	$0D,$78,$1A,$21,$A2,$9A,$30,$43,$04,$3F,$87,$47,$91,$B1,$DD,$A7
	.dc.b	$42,$9C,$CA,$73,$3F,$C1,$51,$CD,$B1,$CD,$A7,$32,$9C,$CA,$73,$11
	.dc.b	$CD,$B1,$CD,$A7,$32,$9C,$CA,$73,$8E,$44,$8E,$4D,$8E,$4D,$79,$35
	.dc.b	$FE,$47,$EA,$FF,$B9,$9B,$EB,$7E,$21,$FF,$70,$36,$35,$1F,$CE,$FF
	.dc.b	$44,$FE,$DA,$FF,$B9,$8F,$F1,$83,$8D,$BF,$D4,$ED,$8F,$E6,$2D,$79
	.dc.b	$69,$FE,$37,$9D,$B6,$36,$D7,$67,$71,$3C,$DF,$0C,$6D,$A6,$E3,$4F
	.dc.b	$08,$D3,$4D,$38,$8D,$91,$C1,$23,$64,$9F,$4B,$4F,$49,$4F,$4B,$A2
	.dc.b	$3A,$6B,$A3,$04,$30,$43,$DA,$F0,$69,$B2,$9B,$3D,$CE,$DA,$6C,$A6
	.dc.b	$E7,$E0,$3E,$93,$99,$0D,$95,$DC,$75,$6C,$7F,$1F,$F8,$4F,$E3,$84
	.dc.b	$70,$7F,$27,$0F,$F2,$76,$99,$70,$7F,$43,$23,$FD,$F3,$9B,$FA,$64
	.dc.b	$62,$3F,$C2,$1B,$7F,$23,$F8,$4F,$E4,$18,$9A,$49,$F1,$3E,$12,$38
	.dc.b	$3F,$C1,$3F,$8A,$FF,$81,$E3,$FC,$17,$08,$C3,$AE,$EF,$F4,$94,$EA
	.dc.b	$73,$3F,$E5,$8F,$61,$F0,$FF,$E5,$6D,$7F,$C2,$B8,$A7,$2F,$EB,$4C
	.dc.b	$4E,$EB,$FC,$59,$18,$3F,$49,$CC,$BF,$CE,$B6,$73,$18,$3F,$4D,$2F
	.dc.b	$B5,$7A,$18,$6D,$78,$5B,$5E,$66,$C7,$8B,$17,$C3,$79,$63,$DC,$D7
	.dc.b	$FC,$E7,$1A,$C3,$D8,$DB,$FC,$01,$27,$F0,$8D,$7F,$CF,$FE,$CB,$EE
	.dc.b	$BD,$2D,$8F,$8C,$D9,$F9,$6D,$8F,$F2,$97,$C7,$BE,$3D,$F3,$BF,$EC
	.dc.b	$7F,$8B,$FF,$22,$CE,$63,$15,$C2,$B6,$AB,$71,$87,$CD,$1F,$ED,$1D
	.dc.b	$FE,$86,$38,$84,$61,$E9,$7F,$ED,$7D,$79,$B6,$32,$7C,$C2,$C7,$F9
	.dc.b	$97,$E8,$3F,$8C,$F8,$86,$92,$7E,$F9,$30,$F1,$61,$E1,$8F,$F3,$A5
	.dc.b	$71,$7F,$96,$36,$7B,$46,$2B,$E7,$F9,$12,$3E,$7B,$67,$A9,$F8,$D8
	.dc.b	$F0,$1F,$E8,$1F,$1D,$F0,$CF,$81,$FF,$5A,$F5,$FF,$9D,$27,$31,$8A
	.dc.b	$E1,$5B,$E4,$FE,$07,$1F,$EF,$DA,$DC,$62,$3F,$CA,$52,$EF,$FF,$60
	.dc.b	$FE,$85,$D6,$31,$48,$C4,$EA,$BB,$B5,$ED,$5F,$E9,$D9,$E5,$ED,$66
	.dc.b	$7B,$B7,$0F,$6B,$6B,$F7,$3E,$97,$87,$AF,$FD,$9D,$7E,$C6,$BD,$9A
	.dc.b	$FF,$B6,$3E,$5B,$FD,$33,$9F,$FB,$76,$3F,$7F,$18,$AF,$E0,$ED,$8C
	.dc.b	$F9,$FC,$A9,$F8,$47,$FB,$9B,$F2,$EF,$E1,$AD,$AF,$E1,$2D,$7F,$DD
	.dc.b	$6D,$7F,$CA,$1C,$8F,$31,$5F,$E5,$52,$4F,$F9,$6D,$FD,$7F,$E1,$E8
	.dc.b	$D9,$FC,$5B,$B5,$19,$9D,$CE,$8F,$DE,$C6,$5F,$7F,$CF,$7E,$0D,$FF
	.dc.b	$7F,$E8,$8C,$47,$C7,$6E,$69,$D3,$F8,$5E,$1F,$C8,$F9,$FC,$1B,$19
	.dc.b	$E1,$4E,$5B,$1F,$8D,$19,$FD,$3F,$81,$FE,$1B,$C3,$EF,$6E,$66,$9F
	.dc.b	$C4,$BE,$2F,$51,$F0,$DA,$FF,$1C,$6A,$BE,$F1,$8F,$B0,$7E,$72,$FB
	.dc.b	$6F,$83,$18,$6C,$66,$BF,$49,$39,$7F,$4D,$6F,$51,$C9,$F7,$7B,$3F
	.dc.b	$BA,$4F,$FA,$4E,$30,$FF,$76,$3E,$E5,$BF,$7C,$FE,$47,$5F,$EE,$C8
	.dc.b	$C7,$87,$F8,$18,$C4,$EA,$BF,$08,$2B,$6F,$D1,$EA,$F8,$1F,$11,$FA
	.dc.b	$7C,$CD,$24,$FE,$7A,$32,$D8,$FD,$1F,$5F,$28,$C1,$F9,$1F,$57,$C8
	.dc.b	$D3,$EE,$71,$FF,$AA,$9B,$1A,$7F,$80,$F2,$FD,$3F,$3B,$F8,$53,$63
	.dc.b	$F8,$AB,$5D,$E3,$FE,$0D,$7D,$F9,$FD,$1A,$F8,$DC,$3F,$55,$BC,$5C
	.dc.b	$6B,$E2,$4E,$5F,$F6,$34,$FF,$B0,$C8,$E0,$EB,$B7,$F5,$48,$F9,$FC
	.dc.b	$1F,$E1,$09,$CB,$5F,$B3,$3B,$7F,$96,$F8,$DF,$8F,$FC,$17,$F9,$33
	.dc.b	$63,$39,$A7,$2D,$8E,$A5,$3A,$DF,$F0,$19,$F1,$C6,$4F,$C0,$5F,$F6
	.dc.b	$A8,$E7,$F8,$2C,$9E,$ED,$79,$87,$99,$1F,$0B,$FE,$E7,$F1,$64,$7F
	.dc.b	$BB,$5A,$FF,$13,$C1,$F9,$7B,$F1,$6F,$F1,$0B,$67,$FC,$5C,$DA,$FF
	.dc.b	$19,$B6,$3F,$C5,$EB,$D3,$FE,$3D,$F2,$BA,$6C,$E9,$AF,$86,$78,$29
	.dc.b	$EB,$EB,$78,$89,$1E,$27,$FD,$2B,$D7,$FB,$6F,$17,$FD,$40,$DC,$7D
	.dc.b	$B5,$FC,$DA,$7D,$3D,$9E,$C2,$46,$0E,$2F,$F3,$B9,$E3,$1F,$C0,$E3
	.dc.b	$0F,$C9,$3D,$FF,$37,$F1,$88,$FC,$67,$E4,$FD,$EF,$F8,$9A,$3D,$CD
	.dc.b	$7F,$8D,$66,$8C,$6D,$F4,$9B,$0F,$BE,$FA,$64,$AF,$A6,$FC,$2C,$3C
	.dc.b	$B1,$FC,$57,$2F,$59,$32,$FA,$FF,$53,$F7,$09,$3F,$71,$FF,$2F,$F9
	.dc.b	$4F,$AC,$FE,$5D,$F4,$3F,$ED,$AA,$FE,$61,$39,$D5,$39,$7F,$B5,$D7
	.dc.b	$07,$FB,$4B,$6B,$C9,$87,$CC,$DF,$EB,$47,$CB,$3F,$05,$1F,$ED,$FA
	.dc.b	$FC,$5B,$63,$27,$F9,$E6,$32,$D3,$FB,$AD,$7F,$45,$F0,$BC,$98,$F9
	.dc.b	$DB,$19,$AF,$29,$FE,$8A,$AF,$29,$E3,$3F,$11,$7F,$8A,$8F,$D2,$FD
	.dc.b	$27,$F7,$9B,$F8,$1B,$A3,$13,$FA,$26,$DF,$E7,$5A,$FF,$59,$FA,$0F
	.dc.b	$F3,$32,$F1,$BE,$54,$61,$BE,$B7,$F8,$1D,$B3,$BB,$FF,$20,$FC,$77
	.dc.b	$FA,$95,$A7,$C3,$8F,$F1,$FC,$65,$FF,$1B,$B5,$F6,$9E,$F3,$F3,$DC
	.dc.b	$3F,$87,$7F,$D5,$38,$79,$B4,$F0,$92,$F3,$8F,$D5,$AE,$67,$CA,$21
	.dc.b	$82,$1D,$A7,$F0,$5C,$5F,$98,$8F,$F1,$27,$C0,$FF,$95,$7E,$13,$FC
	.dc.b	$4D,$D5,$FE,$B5,$94,$E5,$FE,$1D,$7F,$DA,$23,$13,$B6,$C6,$7D,$55
	.dc.b	$F4,$8C,$36,$FC,$33,$A7,$A9,$88,$D3,$F1,$8F,$F6,$EF,$27,$F1,$16
	.dc.b	$E3,$FE,$43,$6E,$BF,$EE,$27,$81,$FD,$F8,$A7,$ED,$CD,$9F,$ED,$ED
	.dc.b	$7F,$0A,$FE,$00,$9A,$FD,$07,$F9,$C1,$C4,$E9,$7E,$36,$DF,$8D,$B7
	.dc.b	$FE,$F4,$F3,$FF,$D1,$FB,$FF,$8A,$E7,$F7,$D1,$8C,$3A,$C7,$E2,$9B
	.dc.b	$19,$8F,$C9,$FB,$F2,$7C,$8D,$8F,$D0,$1F,$E3,$78,$FD,$23,$4C,$BF
	.dc.b	$AA,$E2,$FE,$A4,$F9,$A3,$E1,$94,$87,$EF,$9E,$4F,$F1,$4E,$5E,$D1
	.dc.b	$FB,$C3,$F7,$E6,$8F,$EB,$BB,$FF,$8B,$9F,$E1,$33,$FB,$A8,$C3,$FC
	.dc.b	$38,$FE,$40,$7D,$63,$FA,$51,$E4,$3E,$81,$E3,$CB,$B9,$27,$F8,$6B
	.dc.b	$FC,$1B,$A3,$F0,$51,$FD,$B9,$BC,$A6,$9F,$87,$AF,$01,$36,$FB,$8F
	.dc.b	$80,$FE,$93,$D3,$C8,$8F,$E3,$4C,$96,$FF,$A6,$2F,$63,$FB,$CF,$5F
	.dc.b	$FA,$4D,$B1,$93,$FB,$DC,$65,$FE,$BE,$73,$C5,$D9,$EA,$24,$7E,$7F
	.dc.b	$1F,$3B,$F2,$0F,$B2,$53,$FC,$01,$C5,$F0,$88,$F5,$7E,$14,$72,$8F
	.dc.b	$85,$1F,$9B,$7E,$16,$3F,$E1,$8E,$DF,$96,$48,$F3,$3F,$3B,$A3,$FD
	.dc.b	$61,$F2,$DF,$E1,$39,$34,$8F,$F8,$6B,$C0,$FA,$A4,$7E,$03,$E4,$C3
	.dc.b	$C5,$1E,$B2,$1F,$87,$F2,$7F,$CF,$3E,$EF,$F8,$F8,$91,$FE,$44,$7F
	.dc.b	$C8,$E4,$7F,$C9,$6D,$7F,$C7,$B1,$FE,$46,$23,$FE,$4C,$7F,$C9,$E4
	.dc.b	$7F,$CA,$6D,$7F,$C8,$BF,$0D,$FE,$4E,$23,$FE,$54,$7F,$8A,$91,$FE
	.dc.b	$0A,$D7,$FD,$37,$E5,$7F,$A9,$E3,$FC,$E7,$A1,$A4,$9F,$F3,$FC,$65
	.dc.b	$B1,$FE,$84,$8C,$47,$FA,$0A,$3F,$91,$47,$FB,$23,$CD,$FF,$72,$76
	.dc.b	$E3,$CE,$FF,$48,$36,$33,$7F,$E9,$58,$FF,$4D,$D7,$EA,$63,$FD,$41
	.dc.b	$CD,$82,$3F,$EA,$CE,$C7,$A8,$F1,$3F,$D9,$55,$FE,$AB,$25,$6E,$75
	.dc.b	$7D,$63,$FC,$DF,$87,$F9,$E0,$8F,$F9,$F9,$FE,$7E,$47,$F9,$7B,$5F
	.dc.b	$97,$E7,$7F,$B6,$63,$E5,$92,$3F,$D4,$7A,$FF,$B5,$63,$FD,$C9,$1F
	.dc.b	$E4,$13,$FD,$C8,$53,$FD,$C8,$53,$FD,$C9,$A7,$F8,$2D,$C5,$FF,$BD
	.dc.b	$3D,$44,$62,$51,$8E,$BF,$F7,$A9,$27,$ED,$DF,$D9,$23,$83,$F9,$57
	.dc.b	$AA,$F5,$BD,$91,$86,$BF,$E8,$D8,$D6,$3F,$DB,$63,$13,$A7,$D3,$EB
	.dc.b	$F9,$BF,$2E,$FC,$13,$63,$F9,$C6,$68,$C4,$7F,$A6,$63,$0A,$7F,$B8
	.dc.b	$91,$FE,$C2,$D7,$FD,$41,$E5,$FF,$A8,$63,$F2,$6D,$8F,$E8,$CD,$76
	.dc.b	$30,$EC,$FF,$BB,$FE,$3F,$F2,$64,$70,$7E,$2F,$8B,$84,$7F,$BE,$49
	.dc.b	$19,$94,$7F,$81,$78,$FF,$2F,$ED,$F7,$FA,$BC,$72,$23,$18,$7E,$5D
	.dc.b	$AF,$DE,$AE,$38,$2F,$8F,$97,$AA,$57,$E9,$23,$0D,$AE,$13,$FA,$39
	.dc.b	$CE,$0F,$7B,$EC,$5F,$EB,$1B,$19,$BF,$BD,$7D,$A3,$1C,$5F,$D5,$7E
	.dc.b	$03,$FC,$79,$3F,$E2,$36,$BF,$96,$0E,$1F,$CC,$B6,$9C,$1F,$D5,$E7
	.dc.b	$F8,$47,$8D,$FD,$BA,$31,$1F,$88,$8F,$ED,$9D,$CF,$80,$A4,$9F,$F1
	.dc.b	$B4,$65,$FF,$1C,$61,$E1,$7C,$7E,$4F,$F7,$58,$C1,$99,$E1,$3F,$DC
	.dc.b	$5A,$FF,$8F,$DA,$FB,$5A,$BC,$07,$C0,$BD,$8F,$F3,$0C,$FC,$39,$CC
	.dc.b	$62,$3F,$77,$18,$7F,$DD,$26,$27,$93,$83,$E4,$E1,$E9,$23,$EE,$7F
	.dc.b	$CA,$D3,$CF,$2F,$F2,$B3,$67,$C5,$5D,$27,$2D,$8F,$DA,$FC,$17,$D1
	.dc.b	$23,$FE,$11,$D1,$1E,$08,$FC,$16,$15,$FC,$13,$FE,$26,$F8,$37,$B7
	.dc.b	$FE,$32,$9F,$D8,$46,$37,$FF,$0E,$36,$3F,$A5,$B5,$FF,$1F,$7C,$23
	.dc.b	$D2,$FF,$18,$B6,$3F,$C4,$8D,$7F,$A9,$B5,$F1,$9F,$14,$87,$A0,$F0
	.dc.b	$F5,$79,$6D,$7F,$EC,$B9,$FF,$27,$35,$FE,$63,$1A,$9F,$64,$3E,$15
	.dc.b	$E5,$FF,$36,$F2,$3F,$C7,$51,$89,$FE,$D2,$7F,$3A,$71,$1A,$9E,$13
	.dc.b	$E8,$8F,$31,$1C,$1F,$E9,$1F,$86,$79,$3F,$C6,$E7,$FC,$91,$18,$C7
	.dc.b	$BD,$23,$13,$AA,$E1,$3F,$E9,$BF,$35,$F1,$7F,$DD,$A3,$F7,$51,$87
	.dc.b	$F8,$18,$71,$D5,$FE,$57,$6C,$7D,$08,$F9,$C4,$70,$7F,$6C,$21,$FC
	.dc.b	$1F,$CF,$74,$BD,$3E,$F1,$87,$C3,$F1,$1F,$42,$8D,$26,$BF,$E7,$16
	.dc.b	$C7,$FA,$E5,$AF,$F9,$B5,$AF,$FA,$ED,$AF,$F1,$96,$BF,$DC,$1A,$F8
	.dc.b	$23,$B1,$1E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$1D,$08,$74,$29
	.dc.b	$D0,$A7,$42,$9D,$0A,$74,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08
	.dc.b	$70,$29,$C0,$A7,$03,$44,$34,$7F,$BF,$0F,$F5,$17,$B2,$F0,$BF,$23
	.dc.b	$08,$C6,$5F,$EF,$56,$C7,$DD,$F5,$FF,$C4,$FD,$16,$3F,$AA,$E2,$9C
	.dc.b	$C7,$E0,$5A,$FE,$05,$AF,$E0,$5A,$FE,$41,$B1,$CE,$3A,$C6,$08,$E0
	.dc.b	$93,$E3,$23,$E1,$32,$57,$2F,$8C,$C9,$5C,$BE,$53,$25,$72,$F9,$CC
	.dc.b	$95,$CC,$FA,$5B,$3E,$86,$CF,$09,$D3,$E5,$21,$E3,$29,$E1,$29,$DC
	.dc.b	$AF,$27,$FC,$0D,$F1,$6F,$83,$FF,$1A,$7C,$37,$FB,$86,$31,$A2,$31
	.dc.b	$3A,$AF,$F0,$70,$AD,$CF,$F1,$DA,$E1,$3F,$C9,$F8,$B8,$3F,$E1,$5E
	.dc.b	$4F,$E4,$D5,$B3,$FD,$E1,$C1,$FC,$3D,$BC,$3F,$C4,$DB,$C3,$D2,$FF
	.dc.b	$8C,$4E,$5F,$F1,$06,$3F,$E2,$DE,$FF,$F2,$76,$DF,$EA,$6F,$17,$FD
	.dc.b	$F1,$3F,$E2,$83,$45,$73,$1E,$92,$57,$A6,$7F,$C5,$B9,$FF,$3D,$9F
	.dc.b	$EF,$BF,$89,$BF,$17,$FC,$BF,$86,$8C,$61,$FE,$36,$8C,$4E,$A7,$35
	.dc.b	$E5,$F8,$DF,$3D,$FA,$B2,$46,$23,$FB,$0C,$63,$2F,$44,$E6,$7F,$D2
	.dc.b	$B9,$79,$5F,$45,$FA,$9F,$A1,$C1,$BB,$7B,$6F,$AD,$FD,$7C,$61,$B1
	.dc.b	$9A,$F9,$0D,$BF,$C0,$C7,$F8,$17,$F0,$9F,$F5,$2F,$E0,$38,$F1,$11
	.dc.b	$89,$FC,$3C,$61,$4E,$E7,$51,$E6,$C7,$E7,$E5,$ED,$8F,$ED,$4D,$CF
	.dc.b	$DD,$1F,$E5,$FC,$3F,$BF,$63,$E1,$F0,$3F,$83,$C6,$27,$C3,$1F,$C2
	.dc.b	$BC,$A5,$7F,$AF,$3F,$00,$FF,$0A,$12,$BF,$C3,$2F,$C9,$7E,$6E,$3E
	.dc.b	$12,$65,$FE,$70,$BE,$72,$F8,$BF,$F0,$F1,$1F,$F1,$1B,$F2,$3D,$FF
	.dc.b	$F7,$86,$37,$DD,$FE,$3B,$18,$AF,$D6,$B5,$F8,$93,$E3,$7E,$1E,$1F
	.dc.b	$18,$FF,$40,$CE,$E3,$FD,$AE,$D4,$D3,$51,$F8,$2F,$63,$EF,$92,$7E
	.dc.b	$FB,$FC,$C3,$6F,$9F,$1F,$0A,$7C,$17,$F1,$5E,$51,$C0,$91,$FC,$1A
	.dc.b	$32,$FE,$23,$F2,$7F,$CE,$CF,$E0,$8F,$E1,$6E,$38,$75,$C9,$F8,$49
	.dc.b	$FF,$78,$CE,$5F,$CE,$90,$F8,$A5,$3F,$DE,$FF,$66,$FB,$3F,$C5,$37
	.dc.b	$1F,$F5,$3B,$71,$FC,$DB,$6B,$FD,$55,$EA,$BD,$57,$C0,$7F,$B6,$7E
	.dc.b	$B3,$FD,$96,$DA,$DC,$EB,$15,$69,$B3,$FC,$5B,$A7,$FD,$1B,$3F,$15
	.dc.b	$B1,$9F,$58,$F0,$3F,$CC,$31,$A7,$FD,$AD,$7F,$EB,$3B,$7E,$25,$FE
	.dc.b	$DE,$F6,$3F,$DD,$CD,$8C,$D7,$9A,$72,$F8,$FE,$C3,$BD,$78,$DB,$19
	.dc.b	$38,$9E,$27,$2F,$FB,$3F,$F3,$17,$C3,$7A,$39,$BF,$82,$C6,$25,$18
	.dc.b	$9D,$47,$F4,$A9,$D4,$FD,$2E,$26,$4F,$AF,$C3,$6F,$FA,$8F,$17,$DB
	.dc.b	$9C,$9F,$C5,$3F,$47,$7E,$05,$F2,$A3,$0D,$AF,$EB,$33,$A7,$A1,$FE
	.dc.b	$E9,$71,$3F,$99,$FC,$B3,$FC,$C9,$3F,$97,$3F,$D2,$6E,$3A,$75,$E5
	.dc.b	$FE,$5F,$24,$E6,$BC,$F9,$FE,$48,$93,$F9,$27,$F9,$D5,$7F,$28,$FC
	.dc.b	$07,$E0,$F2,$7F,$6E,$8C,$57,$F3,$8B,$FF,$7F,$46,$27,$5E,$2A,$F8
	.dc.b	$D6,$BF,$7C,$7D,$87,$E3,$2F,$A4,$62,$7A,$1F,$90,$C1,$19,$9D,$D7
	.dc.b	$1A,$FD,$BF,$77,$F2,$96,$8F,$17,$F7,$2D,$32,$FF,$7A,$DA,$47,$F8
	.dc.b	$2B,$D0,$FF,$47,$46,$23,$F4,$27,$E0,$7F,$0E,$FE,$04,$D3,$E9,$1F
	.dc.b	$75,$CF,$09,$FE,$11,$DF,$FB,$9E,$BF,$C5,$FA,$BF,$A3,$25,$FE,$8F
	.dc.b	$8B,$FC,$94,$FC,$1F,$97,$FC,$DC,$3F,$1F,$FA,$51,$86,$F6,$7E,$59
	.dc.b	$B3,$BA,$FF,$09,$F3,$FE,$23,$7F,$C2,$C7,$C3,$3F,$13,$C9,$F2,$5A
	.dc.b	$64,$FE,$03,$D1,$F9,$5E,$0F,$CC,$C7,$C0,$11,$97,$F3,$7D,$3F,$1B
	.dc.b	$C4,$FE,$5B,$F2,$7F,$48,$93,$F5,$9F,$F1,$5E,$32,$3E,$97,$A9,$FE
	.dc.b	$C1,$8C,$47,$D2,$FD,$39,$F5,$9F,$B4,$CF,$FC,$03,$18,$DB,$C8,$D8
	.dc.b	$CC,$EE,$BE,$1E,$BF,$C7,$32,$F8,$64,$8C,$87,$DF,$8C,$8F,$69,$F1
	.dc.b	$4F,$83,$A7,$C9,$5C,$9F,$CD,$FE,$C3,$F8,$4F,$EE,$FF,$E5,$68,$C6
	.dc.b	$A8,$C7,$8C,$AD,$DF,$F3,$A6,$BF,$07,$4F,$E2,$7F,$85,$7B,$30,$34
	.dc.b	$93,$F2,$B5,$FF,$01,$BF,$E5,$66,$BF,$CF,$3A,$FF,$DD,$9E,$5F,$F9
	.dc.b	$70,$FE,$CD,$F2,$0F,$97,$FF,$1B,$37,$1F,$F2,$C3,$67,$73,$FE,$41
	.dc.b	$DB,$FA,$B7,$C2,$7F,$28,$F8,$AF,$6B,$72,$FE,$AA,$D7,$FA,$4B,$5F
	.dc.b	$93,$BB,$E8,$F1,$7F,$D8,$5B,$19,$CD,$87,$CA,$9E,$56,$D3,$DA,$FC
	.dc.b	$3C,$FD,$CD,$3F,$CE,$FA,$3D,$3F,$79,$31,$F7,$BF,$E4,$1B,$E7,$5F
	.dc.b	$2E,$F9,$B4,$FE,$0F,$F3,$1F,$E9,$7E,$6F,$F3,$23,$70,$FA,$2D,$AF
	.dc.b	$F4,$CB,$6B,$F6,$FE,$CF,$F7,$0F,$B5,$FE,$80,$D8,$C8,$8C,$B4,$FD
	.dc.b	$FF,$CD,$FF,$9A,$BB,$CD,$24,$7E,$08,$F8,$DD,$AE,$5C,$6B,$EE,$9F
	.dc.b	$C3,$4E,$73,$F7,$C7,$F6,$4D,$FF,$C2,$CF,$F8,$1B,$54,$62,$35,$E6
	.dc.b	$FE,$28,$FF,$40,$7E,$0E,$F9,$FF,$F0,$6F,$A7,$FD,$BA,$B9,$4E,$54
	.dc.b	$65,$FE,$E7,$19,$8C,$60,$8C,$3F,$76,$7E,$A1,$0F,$5E,$9F,$A2,$1F
	.dc.b	$8C,$BE,$9B,$FD,$6F,$5F,$D7,$23,$0F,$F9,$D8,$C4,$6A,$BF,$AB,$4E
	.dc.b	$B4,$F1,$E1,$E9,$F4,$6E,$7F,$E2,$A6,$DF,$F6,$5C,$7F,$D8,$F5,$B7
	.dc.b	$D1,$1E,$6B,$F9,$EF,$D6,$CB,$EA,$ED,$F4,$E3,$66,$67,$EA,$D7,$C3
	.dc.b	$F8,$5F,$F1,$CC,$79,$1B,$1E,$06,$CF,$D0,$6B,$FE,$20,$FB,$7F,$F2
	.dc.b	$C3,$F0,$4F,$BC,$E2,$BE,$8B,$6B,$FD,$C2,$D9,$FA,$07,$94,$F3,$F0
	.dc.b	$FE,$36,$7C,$D1,$D4,$FB,$B7,$E2,$8C,$DF,$A6,$74,$73,$3A,$D7,$E5
	.dc.b	$DC,$1F,$24,$EF,$87,$C7,$7C,$7F,$16,$F8,$D7,$C6,$FF,$97,$BE,$13
	.dc.b	$F9,$3C,$E6,$31,$8F,$F8,$0A,$31,$6A,$DD,$F2,$8C,$56,$E3,$13,$AB
	.dc.b	$E7,$5F,$E6,$CB,$E7,$39,$8E,$27,$E5,$8F,$F6,$EE,$2C,$7F,$C8,$31
	.dc.b	$FC,$F4,$CB,$87,$F9,$F1,$8D,$7F,$DC,$AF,$E2,$4C,$1F,$EA,$3D,$9A
	.dc.b	$FF,$87,$B5,$FC,$24,$FE,$EC,$FC,$21,$4F,$AA,$53,$A1,$E3,$21,$E2
	.dc.b	$3F,$07,$C9,$D5,$FC,$4B,$5F,$C7,$36,$1B,$FE,$42,$7F,$00,$48,$F2
	.dc.b	$B4,$EC,$43,$E4,$1E,$22,$99,$3C,$E6,$DF,$E1,$1C,$3F,$E9,$ED,$FF
	.dc.b	$6A,$48,$FD,$AC,$65,$FE,$02,$7C,$3C,$BF,$84,$F1,$FF,$19,$8C,$DF
	.dc.b	$F1,$79,$CC,$7C,$7D,$7E,$D4,$7D,$46,$CF,$D4,$3D,$13,$E4,$5E,$A7
	.dc.b	$F0,$82,$47,$92,$32,$FF,$5A,$CF,$F9,$34,$7F,$33,$E6,$FF,$67,$3F
	.dc.b	$DF,$23,$2F,$F3,$49,$F1,$6F,$F8,$98,$F8,$E4,$3E,$39,$FE,$A2,$E8
	.dc.b	$FE,$96,$FF,$99,$0F,$F4,$8F,$47,$E3,$AB,$FD,$1E,$D9,$DC,$FF,$56
	.dc.b	$FC,$A2,$3F,$D6,$9A,$FF,$B3,$BB,$3F,$B3,$3F,$D9,$3E,$2F,$FC,$65
	.dc.b	$87,$78,$4E,$9F,$ED,$F3,$A7,$E5,$1F,$7F,$5F,$9A,$FC,$CD,$3F,$BF
	.dc.b	$5F,$33,$FA,$BF,$7F,$C4,$69,$98,$DC,$F4,$3E,$25,$74,$7F,$21,$E1
	.dc.b	$FF,$85,$6B,$F8,$24,$7F,$BC,$BC,$1F,$F6,$4F,$47,$F8,$56,$73,$3F
	.dc.b	$EE,$89,$CB,$FE,$24,$8C,$BF,$E2,$2F,$13,$2F,$F8,$9A,$33,$7F,$E2
	.dc.b	$E7,$67,$F7,$0F,$17,$D5,$97,$F8,$DE,$BF,$C6,$B3,$EB,$9F,$C0,$C7
	.dc.b	$DC,$26,$5E,$B7,$F0,$1E,$6F,$F9,$AF,$F0,$5F,$F2,$71,$23,$F0,$11
	.dc.b	$97,$FC,$AD,$6F,$DD,$BF,$CB,$7E,$5F,$FB,$A1,$B1,$99,$DC,$7F,$A2
	.dc.b	$BF,$06,$91,$FE,$6C,$6B,$FE,$92,$F9,$CF,$F4,$B9,$1F,$F3,$AB,$FE
	.dc.b	$61,$FC,$0A,$47,$93,$CE,$FF,$7B,$C6,$23,$FC,$C1,$5F,$C8,$C8,$FF
	.dc.b	$99,$5F,$F6,$2D,$48,$FF,$64,$B5,$FF,$79,$7A,$05,$5F,$EA,$22,$3F
	.dc.b	$E8,$47,$FD,$5B,$52,$3F,$D1,$6D,$7F,$D7,$9E,$B3,$27,$E3,$FF,$EB
	.dc.b	$82,$4F,$FB,$B3,$D8,$FF,$60,$C7,$FA,$1B,$DA,$E1,$F2,$16,$CD,$25
	.dc.b	$FE,$D7,$DD,$FF,$69,$3E,$83,$1E,$EF,$AF,$4F,$D6,$4E,$9B,$19,$9F
	.dc.b	$F4,$54,$7F,$4B,$23,$83,$FD,$DF,$A7,$FB,$B8,$8E,$0F,$F8,$01,$EF
	.dc.b	$7C,$9F,$CD,$7F,$DF,$30,$69,$1F,$F7,$7F,$8C,$F8,$2F,$F1,$34,$FF
	.dc.b	$0F,$8C,$7A,$3F,$26,$31,$3A,$AE,$2D,$7F,$DF,$3F,$07,$FC,$89,$FE
	.dc.b	$30,$D7,$3E,$47,$E9,$23,$11,$FC,$7F,$0F,$84,$FF,$B9,$33,$FF,$74
	.dc.b	$91,$FF,$61,$FD,$D7,$FB,$90,$8E,$0E,$F7,$F7,$E3,$F8,$D5,$A3,$13
	.dc.b	$AE,$05,$6A,$73,$3F,$91,$1F,$0C,$E5,$FC,$A7,$CE,$FF,$25,$8C,$4A
	.dc.b	$31,$3A,$8F,$E7,$13,$A9,$FE,$75,$83,$E1,$7F,$CD,$F5,$FC,$E0,$FC
	.dc.b	$B3,$8F,$2F,$6D,$8F,$F0,$43,$5F,$D3,$78,$6F,$75,$D1,$F9,$FB,$FC
	.dc.b	$C1,$2B,$D9,$5B,$9D,$47,$4F,$89,$FD,$87,$D7,$F8,$1B,$19,$11,$96
	.dc.b	$9D,$30,$7C,$67,$EB,$E7,$FD,$DE,$D9,$D3,$5F,$4C,$F0,$CF,$9C,$7E
	.dc.b	$19,$FF,$15,$F0,$FE,$1A,$0D,$31,$F1,$AF,$91,$FE,$5E,$DA,$FF,$19
	.dc.b	$B6,$BF,$AE,$35,$FE,$D9,$F0,$FF,$E0,$07,$F7,$07,$E0,$5C,$72,$FF
	.dc.b	$63,$6C,$70,$F5,$3F,$2E,$FD,$8F,$8C,$FE,$E9,$C4,$69,$23,$C7,$97
	.dc.b	$98,$99,$60,$FE,$CD,$5C,$66,$D7,$17,$F0,$6F,$F9,$39,$7C,$7F,$FB
	.dc.b	$AD,$FF,$28,$B5,$FE,$F4,$D7,$FC,$14,$D7,$F3,$D9,$FC,$36,$E1,$E4
	.dc.b	$7E,$3B,$FD,$6A,$78,$47,$F3,$56,$BE,$2F,$47,$C9,$18,$F9,$17,$C0
	.dc.b	$7F,$3B,$F9,$AF,$F0,$64,$62,$D1,$89,$D4,$7F,$87,$E7,$53,$E2,$D3
	.dc.b	$E3,$4D,$21,$BA,$FE,$91,$E1,$73,$83,$48,$FF,$8A,$FE,$17,$FC,$02
	.dc.b	$FE,$FD,$AF,$F5,$16,$A6,$E9,$F1,$E3,$FA,$B3,$53,$4C,$47,$F2,$D8
	.dc.b	$FC,$69,$D7,$C0,$FC,$4B,$D5,$B1,$BE,$2F,$E0,$3F,$27,$F8,$C2,$57
	.dc.b	$E1,$9F,$C1,$5F,$7D,$FE,$C4,$FF,$91,$E3,$CD,$9B,$E5,$3F,$CB,$33
	.dc.b	$FC,$49,$AF,$FA,$4D,$AF,$F7,$18,$E1,$C7,$FE,$5A,$F9,$67,$CB,$7F
	.dc.b	$62,$AF,$F2,$34,$63,$CA,$FF,$50,$B6,$33,$3B,$7E,$B9,$F8,$EB,$FB
	.dc.b	$6F,$9F,$2F,$F2,$26,$3F,$EA,$69,$FF,$24,$36,$7D,$1F,$07,$A3,$FC
	.dc.b	$5B,$EA,$FF,$9F,$BE,$1B,$FD,$64,$49,$FF,$34,$46,$5E,$D7,$FE,$C9
	.dc.b	$7C,$53,$F8,$D2,$3F,$E6,$FF,$97,$7C,$C7,$F8,$26,$7F,$CC,$91,$8A
	.dc.b	$E5,$5F,$E1,$78,$C4,$EA,$B8,$4F,$F9,$E7,$E6,$5F,$2B,$FB,$96,$C7
	.dc.b	$FB,$21,$AF,$F9,$C5,$AF,$18,$E0,$47,$83,$5E,$0D,$78,$1A,$21,$A2
	.dc.b	$9A,$30,$43,$04,$3F,$B2,$47,$91,$B1,$DD,$A7,$42,$9C,$CA,$73,$3F
	.dc.b	$C7,$D1,$CD,$B1,$CD,$A7,$32,$9C,$CA,$73,$11,$CD,$B1,$CD,$A7,$32
	.dc.b	$9C,$CA,$73,$F8,$A7,$C4,$7D,$E8,$FF,$2A,$46,$30,$FF,$7F,$B6,$3F
	.dc.b	$D5,$0D,$7F,$01,$CF,$FE,$5B,$E9,$FB,$8D,$8C,$9F,$C8,$CF,$9A,$E5
	.dc.b	$A6,$4F,$8E,$E5,$A6,$4F,$82,$E5,$A6,$71,$FE,$25,$3E,$28,$C4,$EE
	.dc.b	$73,$1E,$49,$D1,$1C,$10,$F2,$B9,$69,$93,$CA,$E5,$A6,$4F,$2B,$96
	.dc.b	$99,$8F,$1B,$D0,$87,$22,$9C,$0A,$F0,$74,$43,$45,$74,$E0,$FF,$6F
	.dc.b	$FC,$BF,$FA,$1F,$E0,$FF,$DC,$5F,$08,$8C,$4F,$FB,$CE,$31,$BA,$B7
	.dc.b	$5A,$AF,$E3,$D1,$FD,$64,$8F,$F1,$27,$FD,$DD,$C0,$F8,$77,$A5,$CE
	.dc.b	$BF,$8C,$C6,$2B,$6D,$8C,$CE,$DF,$F2,$C7,$81,$FD,$B6,$BF,$B1,$12
	.dc.b	$3F,$B3,$1B,$3F,$D5,$DE,$D3,$E0,$7F,$D4,$DB,$7F,$C0,$18,$C6,$9F
	.dc.b	$8D,$6C,$66,$77,$5E,$28,$F2,$E3,$F9,$3C,$FD,$8D,$CF,$DB,$3F,$E1
	.dc.b	$4F,$85,$FC,$B7,$89,$FE,$FA,$8C,$61,$F9,$B8,$C3,$FE,$73,$19,$7C
	.dc.b	$7C,$FC,$98,$78,$A7,$FC,$F1,$39,$BF,$54,$E6,$7F,$D9,$99,$79,$27
	.dc.b	$CE,$48,$F5,$1F,$E2,$6F,$9A,$F2,$91,$F2,$3F,$AD,$FC,$89,$F5,$1D
	.dc.b	$65,$18,$CB,$F6,$8D,$BF,$54,$EA,$B8,$61,$E3,$8F,$09,$FE,$91,$70
	.dc.b	$DC,$7E,$4B,$F0,$FF,$0B,$FA,$BC,$1D,$1B,$19,$E0,$FC,$93,$FE,$EB
	.dc.b	$8C,$BF,$E9,$DD,$1F,$11,$FE,$EF,$F9,$7F,$2C,$62,$51,$89,$D4,$7D
	.dc.b	$39,$D4,$FC,$98,$F3,$71,$7D,$F2,$6B,$C6,$1F,$3B,$48,$F3,$C3,$FE
	.dc.b	$00,$47,$F9,$63,$E4,$FF,$0A,$49,$FC,$23,$FC,$77,$6F,$8B,$1F,$73
	.dc.b	$6F,$8A,$F8,$B0,$F2,$3F,$E5,$7A,$E5,$1D,$63,$99,$FC,$A7,$E0,$7F
	.dc.b	$D1,$4D,$F0,$FF,$95,$46,$27,$EC,$B6,$7E,$B9,$E1,$CF,$DE,$E0,$46
	.dc.b	$DF,$F6,$63,$51,$AA,$EE,$7F,$39,$F2,$3F,$D3,$57,$FE,$61,$24,$FF
	.dc.b	$86,$9F,$6A,$CB,$E1,$B7,$2F,$86,$F9,$67,$F2,$3F,$78,$FB,$CF,$E8
	.dc.b	$CD,$9F,$E9,$4D,$9D,$CF,$F4,$FE,$67,$C8,$7F,$64,$B7,$CF,$FF,$B7
	.dc.b	$23,$0D,$F0,$FD,$AD,$9D,$9F,$EC,$7F,$92,$5B,$ED,$7F,$7D,$EF,$7F
	.dc.b	$B6,$E7,$31,$8F,$69,$18,$BE,$57,$C5,$FF,$69,$61,$D3,$E0,$78,$36
	.dc.b	$33,$3F,$EF,$C6,$CF,$83,$F3,$CF,$F3,$03,$63,$81,$FE,$EE,$8C,$BF
	.dc.b	$EE,$A6,$BF,$68,$7E,$9C,$FC,$27,$DB,$F1,$7F,$C0,$23,$18,$23,$13
	.dc.b	$AA,$FE,$3D,$5A,$9F,$A3,$1B,$3F,$91,$71,$7D,$78,$C9,$23,$8F,$AB
	.dc.b	$F3,$3E,$E3,$C0,$D8,$CE,$29,$CB,$63,$F9,$77,$C0,$3E,$D3,$F8,$E3
	.dc.b	$67,$FC,$1C,$DE,$6F,$E3,$6D,$AF,$2F,$37,$E3,$08,$FE,$31,$FF,$0C
	.dc.b	$76,$3E,$17,$F9,$DC,$BE,$07,$F9,$F4,$E7,$C9,$FD,$2C,$62,$B6,$D8
	.dc.b	$CD,$7D,$1C,$BF,$2B,$F3,$FF,$AC,$EB,$FC,$A3,$67,$50,$8C,$47,$FA
	.dc.b	$7E,$30,$FE,$78,$FE,$A1,$F0,$6F,$C2,$BF,$6F,$F4,$7F,$CE,$A3,$12
	.dc.b	$8C,$4E,$A3,$8B,$5F,$E2,$87,$F0,$5F,$03,$F6,$E7,$EE,$3C,$67,$DA
	.dc.b	$7F,$38,$9F,$E6,$AD,$8C,$E7,$FD,$E6,$72,$FF,$7B,$3F,$6B,$F0,$FF
	.dc.b	$C3,$3E,$9B,$F6,$8D,$9F,$82,$D7,$E0,$35,$78,$4E,$DF,$F6,$17,$E4
	.dc.b	$1F,$A3,$C0,$D2,$57,$E2,$3A,$3F,$0E,$49,$FC,$3B,$FD,$97,$B7,$F4
	.dc.b	$3E,$7F,$E8,$27,$D5,$18,$AF,$D0,$F7,$FC,$1E,$2F,$9B,$E6,$7C,$C9
	.dc.b	$FE,$A6,$7C,$C2,$9F,$B1,$8D,$BF,$37,$C7,$FE,$42,$FF,$23,$EA,$FE
	.dc.b	$6A,$FE,$DF,$E6,$BF,$DC,$8F,$CB,$6E,$7F,$DD,$9B,$1F,$E1,$16,$BF
	.dc.b	$CE,$25,$FA,$67,$FB,$1A,$3F,$04,$7F,$03,$8C,$FB,$DF,$81,$8F,$F1
	.dc.b	$EC,$EF,$A3,$F3,$ED,$8F,$83,$B7,$F6,$66,$F9,$FF,$1C,$8F,$F8,$8F
	.dc.b	$D6,$7D,$FF,$F2,$AC,$3F,$C1,$B1,$8B,$F1,$D7,$F1,$58,$C4,$EA,$B8
	.dc.b	$4F,$B7,$E0,$BF,$A9,$FD,$DF,$C1,$6C,$64,$46,$5A,$7C,$FD,$3E,$6B
	.dc.b	$FD,$3C,$FD,$A7,$E0,$BF,$E0,$BF,$1B,$FC,$75,$9A,$31,$3A,$EC,$FE
	.dc.b	$9D,$18,$8F,$E6,$1F,$8D,$FF,$46,$F0,$BF,$9B,$36,$3F,$99,$35,$DD
	.dc.b	$FE,$E9,$FD,$E7,$CF,$FF,$6E,$AF,$EA,$CD,$8F,$E0,$6D,$76,$71,$3D
	.dc.b	$47,$EE,$08,$60,$FE,$B3,$7F,$D7,$31,$F6,$B6,$3E,$0F,$9D,$EC,$3F
	.dc.b	$C3,$1C,$DF,$94,$F0,$BF,$80,$B4,$FE,$EF,$EF,$BE,$35,$F4,$9F,$EB
	.dc.b	$F8,$C3,$63,$38,$7F,$B1,$67,$2F,$FB,$1F,$5F,$EE,$FB,$FF,$83,$BE
	.dc.b	$EB,$FB,$B3,$67,$F9,$1B,$5F,$E4,$4D,$7F,$BA,$F9,$0F,$A4,$FF,$01
	.dc.b	$36,$3B,$37,$7F,$F0,$14,$FB,$B8,$FF,$07,$5F,$B1,$9F,$F1,$9E,$2E
	.dc.b	$B7,$C6,$FF,$83,$5B,$5E,$96,$D7,$E2,$1B,$1F,$CF,$FF,$55,$FE,$63
	.dc.b	$F5,$7F,$E0,$A8,$FF,$15,$E0,$9C,$BB,$F3,$3F,$C5,$F5,$BA,$FF,$08
	.dc.b	$D6,$9F,$BA,$79,$CF,$CE,$61,$FD,$53,$ED,$3D,$D8,$7E,$F5,$B3,$FC
	.dc.b	$0D,$AF,$FA,$E9,$B1,$E3,$C3,$FC,$7F,$EF,$F8,$63,$14,$8C,$61,$FD
	.dc.b	$31,$B5,$C2,$BC,$75,$0F,$7D,$7F,$52,$FB,$2D,$C1,$A4,$9F,$EC,$DF
	.dc.b	$9D,$7F,$4E,$96,$5E,$38,$C3,$7D,$0F,$F2,$AB,$67,$75,$CA,$72,$D8
	.dc.b	$C9,$F4,$AB,$9C,$66,$BE,$84,$FC,$D3,$E7,$F4,$7E,$8B,$83,$F3,$17
	.dc.b	$C2,$79,$61,$D9,$A7,$63,$FC,$93,$1B,$7F,$68,$75,$2D,$75,$BF,$F3
	.dc.b	$3B,$63,$FD,$99,$F1,$8F,$7D,$F1,$BF,$E7,$4F,$17,$E2,$D7,$F9,$DA
	.dc.b	$B5,$39,$C3,$FC,$81,$18,$95,$70,$51,$98,$C3,$67,$8D,$F8,$2B,$8C
	.dc.b	$62,$FB,$47,$86,$75,$7D,$70,$FF,$76,$CF,$FB,$C0,$87,$FB,$9D,$CB
	.dc.b	$7B,$1A,$7F,$B4,$A3,$FB,$21,$C0,$FF,$5C,$9F,$CD,$B3,$FC,$33,$F8
	.dc.b	$12,$6B,$F7,$DF,$B7,$7D,$9E,$27,$7A,$ED,$19,$8F,$AF,$AF,$F8,$DD
	.dc.b	$FB,$67,$42,$D7,$52,$38,$21,$FB,$8E,$0F,$DF,$C7,$EF,$63,$FD,$16
	.dc.b	$D3,$F4,$67,$9B,$1F,$83,$1E,$F3,$FD,$70,$58,$FD,$9F,$B5,$FA,$A2
	.dc.b	$3D,$5F,$E1,$18,$B8,$FF,$91,$56,$1B,$5F,$C1,$79,$FD,$D3,$FC,$0D
	.dc.b	$AF,$D0,$3E,$31,$E3,$30,$7E,$F3,$1F,$F6,$2F,$D2,$FE,$F0,$95,$EF
	.dc.b	$7F,$85,$73,$FF,$28,$AF,$E4,$DD,$1F,$76,$BF,$97,$36,$75,$3E,$02
	.dc.b	$1F,$60,$87,$F4,$0C,$39,$74,$FB,$A3,$FC,$D3,$18,$33,$D5,$FE,$DA
	.dc.b	$FB,$4F,$E9,$B2,$69,$1F,$F5,$37,$6F,$F5,$67,$FD,$50,$1C,$3A,$EE
	.dc.b	$FE,$BA,$EA,$32,$F9,$8F,$8E,$7D,$A3,$D6,$7F,$74,$F0,$3E,$17,$83
	.dc.b	$FE,$D0,$8F,$EE,$B1,$83,$F4,$4F,$22,$72,$FE,$5D,$EB,$B7,$FB,$83
	.dc.b	$C2,$FE,$09,$1F,$E0,$83,$F7,$E5,$E8,$FF,$06,$B8,$3F,$C2,$6F,$F8
	.dc.b	$62,$71,$1F,$E6,$EF,$4D,$E2,$FF,$B8,$AB,$47,$F8,$2B,$6F,$49,$2B
	.dc.b	$E3,$9F,$E1,$07,$D3,$8F,$9C,$9D,$1F,$9E,$9F,$86,$3C,$CF,$F2,$FE
	.dc.b	$DF,$E3,$92,$47,$F8,$F5,$FF,$22,$66,$F9,$6A,$49,$FF,$28,$79,$5F
	.dc.b	$EF,$88,$FF,$3D,$7C,$C4,$9F,$F2,$CB,$5F,$F3,$37,$9D,$FE,$65,$9F
	.dc.b	$F2,$FF,$3A,$7B,$1F,$E6,$79,$D3,$63,$53,$FE,$4F,$F0,$64,$FF,$94
	.dc.b	$DA,$FE,$0F,$5F,$83,$DB,$FC,$43,$6F,$85,$3F,$E8,$5E,$CF,$07,$3F
	.dc.b	$FA,$2C,$91,$FE,$8B,$7F,$D2,$98,$48,$FF,$39,$B5,$FF,$7F,$7B,$BF
	.dc.b	$EE,$37,$FD,$C3,$EA,$FF,$5C,$7D,$91,$97,$FD,$49,$C1,$F6,$3D,$7F
	.dc.b	$F5,$64,$21,$F8,$DF,$FA,$B4,$91,$F1,$4F,$C3,$1B,$7F,$D7,$6B,$E9
	.dc.b	$A9,$1F,$F5,$E9,$93,$F5,$BC,$DF,$EC,$8C,$BF,$12,$D8,$CF,$89,$E3
	.dc.b	$9C,$B6,$3F,$13,$C1,$F3,$1F,$E2,$DE,$AF,$FB,$83,$8B,$FD,$68,$48
	.dc.b	$FF,$5A,$3F,$EF,$05,$F8,$25,$23,$FE,$F2,$F4,$7F,$05,$1F,$EF,$B8
	.dc.b	$7D,$2D,$23,$FE,$AB,$F8,$07,$BF,$F8,$99,$FD,$1F,$C2,$3E,$0B,$F7
	.dc.b	$6D,$BF,$EB,$CD,$77,$1F,$C0,$39,$3E,$B4,$7D,$CE,$5F,$B4,$A3,$2F
	.dc.b	$DC,$8C,$FB,$CF,$37,$ED,$7D,$67,$ED,$23,$1C,$E8,$C4,$EA,$FF,$8F
	.dc.b	$D6,$A7,$C7,$0F,$A6,$A4,$7F,$92,$E7,$EB,$8F,$F7,$41,$23,$FD,$8E
	.dc.b	$D7,$F5,$9D,$3F,$8E,$C5,$F9,$CF,$FB,$E4,$93,$F8,$D9,$D4,$7F,$32
	.dc.b	$5F,$14,$F2,$BF,$2D,$39,$8C,$5F,$4B,$E5,$18,$9D,$57,$09,$CF,$C5
	.dc.b	$7E,$7F,$EE,$B9,$36,$33,$82,$72,$D8,$FD,$12,$F4,$3F,$C6,$D0,$FA
	.dc.b	$3F,$F6,$5C,$61,$BE,$B7,$F0,$D6,$CE,$CF,$97,$C9,$F8,$62,$3F,$86
	.dc.b	$7F,$95,$F8,$9F,$C2,$23,$FA,$61,$FC,$35,$C2,$F8,$E7,$C7,$FE,$21
	.dc.b	$B9,$7F,$8B,$5B,$3F,$CE,$1A,$FF,$60,$F9,$0E,$2F,$F8,$CD,$AF,$F3
	.dc.b	$E6,$BB,$AF,$2F,$91,$FC,$74,$91,$FE,$09,$9D,$BF,$C2,$5A,$FF,$2F
	.dc.b	$6A,$6F,$CB,$7F,$A4,$63,$FC,$88,$D7,$FA,$7B,$5F,$EE,$ED,$7E,$24
	.dc.b	$F8,$9F,$E9,$F1,$97,$F9,$F3,$53,$4C,$7C,$93,$C6,$FE,$89,$E6,$FF
	.dc.b	$5E,$8C,$52,$31,$3A,$8F,$F0,$4C,$EA,$7F,$C5,$5E,$2F,$F1,$79,$FE
	.dc.b	$C3,$3C,$BC,$AE,$31,$FE,$ED,$38,$DF,$F4,$46,$99,$3F,$99,$7C,$9B
	.dc.b	$E5,$3F,$92,$46,$1A,$FF,$5B,$8D,$6F,$FE,$7B,$8D,$35,$FC,$59,$0F
	.dc.b	$F0,$7F,$07,$E3,$3E,$59,$F2,$DF,$C4,$9B,$5F,$E7,$76,$D7,$F8,$09
	.dc.b	$AF,$D4,$CF,$FC,$37,$B1,$A4,$8F,$AC,$7D,$83,$FC,$A1,$E2,$B1,$FE
	.dc.b	$42,$DC,$3F,$D0,$CD,$9F,$F0,$8B,$5F,$C1,$FA,$BE,$02,$3F,$8A,$7F
	.dc.b	$2E,$FD,$4A,$FD,$4E,$FF,$E2,$DA,$F8,$F3,$D3,$5F,$09,$30,$FF,$4B
	.dc.b	$7C,$BF,$FA,$CB,$CE,$FF,$10,$B6,$BF,$C9,$2D,$7F,$B4,$35,$FF,$07
	.dc.b	$61,$ED,$E3,$FA,$50,$8C,$57,$F8,$B1,$B1,$FE,$99,$8F,$DC,$72,$FD
	.dc.b	$E8,$FE,$00,$DE,$6F,$DF,$CE,$B1,$EB,$EA,$BE,$63,$83,$63,$31,$83
	.dc.b	$A7,$CC,$BE,$6B,$FC,$A6,$DC,$BF,$D6,$CD,$7F,$C8,$2D,$7F,$CD,$2D
	.dc.b	$7F,$C1,$0D,$7F,$C5,$CD,$4D,$24,$78,$09,$1E,$06,$BE,$06,$BE,$06
	.dc.b	$BE,$06,$BE,$06,$BE,$06,$C7,$82,$3B,$10,$EC,$53,$B1,$4E,$C5,$3B
	.dc.b	$14,$EC,$57,$B3,$D0,$87,$42,$9F,$EE,$F3,$FC,$DB,$F0,$BF,$EF,$B7
	.dc.b	$FD,$2A,$D7,$47,$D0,$FC,$4B,$FC,$B5,$F0,$DF,$51,$B1,$91,$19,$69
	.dc.b	$97,$04,$70,$43,$ED,$C7,$46,$9C,$8B,$1C,$5A,$71,$3F,$03,$1C,$5A
	.dc.b	$71,$2C,$71,$69,$C4,$FC,$3C,$71,$69,$C4,$B1,$C5,$A7,$1A,$4F,$16
	.dc.b	$C7,$16,$CF,$16,$C7,$19,$E0,$47,$83,$67,$83,$E5,$CB,$F5,$DF,$23
	.dc.b	$FE,$A6,$6D,$7F,$B2,$72,$FF,$38,$4E,$5B,$1F,$ED,$2F,$D1,$FF,$9C
	.dc.b	$CF,$FA,$B9,$B1,$9E,$3D,$BF,$E7,$58,$CB,$FE,$77,$E1,$FE,$1D,$5F
	.dc.b	$C2,$F9,$5C,$68,$C3,$6B,$FD,$C3,$5C,$1F,$E3,$7A,$7F,$A9,$6B,$F8
	.dc.b	$FB,$63,$27,$13,$F8,$FB,$97,$FD,$11,$CF,$FD,$A2,$7F,$AF,$92,$3F
	.dc.b	$AE,$BF,$EF,$0F,$A1,$7E,$96,$F5,$3F,$D2,$98,$66,$31,$C9,$F8,$7A
	.dc.b	$DD,$6B,$0F,$F4,$B4,$62,$7F,$09,$D1,$F9,$EC,$3D,$35,$87,$F1,$51
	.dc.b	$98,$FF,$1E,$F8,$DF,$E4,$F7,$F2,$95,$FE,$3F,$F2,$7F,$30,$D8,$CC
	.dc.b	$EE,$FF,$D4,$D1,$FE,$AB,$FD,$4D,$E4,$79,$63,$FD,$9D,$69,$CB,$D9
	.dc.b	$B5,$CA,$35,$5C,$6B,$FD,$AD,$8F,$9D,$F9,$E7,$AC,$95,$C7,$3F,$F7
	.dc.b	$DC,$7A,$9B,$19,$F8,$2F,$5C,$62,$7D,$3E,$1F,$F8,$22,$BF,$DF,$AD
	.dc.b	$8C,$F2,$F8,$D4,$65,$F1,$CF,$CE,$23,$F8,$67,$E2,$BF,$E6,$9F,$AA
	.dc.b	$FC,$CB,$FE,$F1,$D5,$18,$8D,$69,$F8,$D7,$51,$9F,$BC,$78,$5F,$EF
	.dc.b	$9F,$98,$EB,$18,$A4,$62,$75,$1F,$5A,$75,$3E,$68,$F4,$9E,$72,$1F
	.dc.b	$EE,$18,$E9,$F3,$1C,$E0,$D2,$3F,$82,$FC,$7D,$F2,$3F,$C5,$DB,$1F
	.dc.b	$C5,$DB,$A7,$C4,$6C,$78,$0F,$BF,$1F,$96,$23,$C6,$7A,$11,$C1,$F7
	.dc.b	$FE,$3B,$EE,$91,$C1,$F8,$63,$B0,$E2,$FC,$63,$E0,$37,$5F,$8B,$9E
	.dc.b	$4D,$8F,$E6,$1D,$5F,$E1,$B8,$FE,$16,$48,$FE,$17,$D9,$FC,$42,$3E
	.dc.b	$74,$75,$6B,$E1,$7F,$9E,$7B,$2F,$A5,$FC,$13,$78,$7F,$05,$48,$C7
	.dc.b	$67,$FB,$7E,$BF,$17,$EB,$BE,$2D,$A3,$17,$D1,$FF,$B3,$67,$31,$8C
	.dc.b	$BF,$DA,$51,$89,$D5,$71,$7F,$59,$F8,$8A,$B8,$1F,$BC,$EA,$46,$67
	.dc.b	$73,$A3,$C5,$3A,$71,$87,$F1,$FF,$53,$A7,$D2,$7E,$E5,$B3,$FE,$D0
	.dc.b	$AF,$F5,$32,$9D,$34,$F2,$1F,$CF,$BF,$64,$FF,$80,$3F,$A8,$7F,$14
	.dc.b	$F1,$91,$89,$DC,$63,$E0,$7E,$35,$6E,$B5,$5D,$38,$BF,$6B,$9B,$E9
	.dc.b	$7F,$77,$18,$6F,$77,$CF,$6C,$EE,$BF,$07,$95,$7F,$AB,$73,$7F,$54
	.dc.b	$AF,$EA,$64,$8F,$E5,$7F,$AE,$FF,$96,$5E,$07,$D1,$71,$1F,$C3,$1B
	.dc.b	$1C,$3C,$BF,$AD,$3F,$C4,$3E,$15,$F5,$BF,$C3,$5B,$97,$E3,$1B,$D3
	.dc.b	$B9,$FE,$25,$6F,$77,$FA,$E6,$3F,$81,$8C,$3F,$94,$6D,$7F,$1E,$6C
	.dc.b	$7D,$9D,$7F,$8F,$34,$78,$9F,$CD,$E3,$31,$F8,$B6,$AE,$AF,$CE,$7D
	.dc.b	$A7,$E2,$A3,$F2,$ED,$7F,$29,$1A,$D3,$FB,$54,$69,$AF,$D0,$E0,$FD
	.dc.b	$CF,$07,$ED,$63,$EA,$F6,$FE,$E5,$D6,$9F,$E0,$62,$3F,$69,$FD,$85
	.dc.b	$FF,$22,$8F,$EC,$3C,$5F,$B7,$F0,$B2,$48,$FC,$D7,$4F,$EA,$7E,$97
	.dc.b	$F3,$24,$8C,$87,$FA,$D7,$C8,$FE,$BF,$E7,$2C,$9A,$47,$F3,$7E,$FB
	.dc.b	$F1,$0F,$AF,$BF,$FB,$ED,$B1,$99,$FE,$99,$39,$7E,$D7,$2F,$EF,$73
	.dc.b	$FB,$4D,$8F,$EF,$BC,$BF,$80,$9C,$C7,$C1,$3E,$67,$07,$C9,$7F,$D7
	.dc.b	$FC,$3F,$35,$C9,$FA,$29,$FF,$18,$91,$C1,$FE,$0A,$F9,$7F,$F0,$4A
	.dc.b	$FA,$3F,$ED,$31,$86,$FA,$1F,$C4,$DB,$3B,$3E,$1E,$BE,$97,$F8,$C6
	.dc.b	$9E,$FC,$3F,$82,$6B,$F0,$DB,$5B,$AF,$E3,$DF,$55,$FE,$18,$FD,$6B
	.dc.b	$FB,$CB,$63,$FC,$3E,$D7,$F8,$8B,$5F,$F2,$7D,$FC,$FF,$C3,$3A,$C1
	.dc.b	$A4,$7E,$06,$7F,$CB,$7C,$0F,$59,$23,$27,$F1,$A3,$F2,$E4,$3E,$69
	.dc.b	$0F,$E8,$1F,$32,$FC,$7B,$FC,$72,$D8,$FF,$1E,$B7,$87,$FC,$60,$D8
	.dc.b	$FE,$61,$E2,$7F,$3F,$F9,$5F,$F1,$63,$63,$22,$32,$D3,$F9,$C6,$5B
	.dc.b	$7F,$A3,$35,$FE,$86,$DE,$BF,$F3,$03,$98,$79,$BF,$DB,$5F,$71,$F8
	.dc.b	$CE,$2F,$DC,$9E,$AE,$9F,$69,$23,$06,$B3,$FC,$A3,$D3,$9B,$F2,$8F
	.dc.b	$38,$FF,$2C,$FB,$AF,$C9,$7F,$B9,$4F,$FA,$D6,$31,$E7,$FF,$AE,$5B
	.dc.b	$19,$9D,$D7,$F6,$DF,$37,$FB,$87,$EB,$57,$63,$49,$7F,$D7,$7E,$CB
	.dc.b	$FC,$A6,$47,$FC,$B0,$FF,$AF,$F6,$F8,$51,$F7,$4F,$56,$DE,$D7,$FB
	.dc.b	$0D,$B8,$9F,$5B,$F8,$02,$4F,$E0,$1E,$9B,$FC,$28,$E7,$A7,$C6,$C7
	.dc.b	$BC,$7C,$53,$E8,$9A,$3E,$9E,$3F,$E0,$FE,$3F,$F7,$8E,$5F,$DB,$3D
	.dc.b	$24,$62,$BF,$63,$18,$7F,$DE,$73,$B9,$D4,$77,$3D,$D7,$E2,$7C,$06
	.dc.b	$31,$F4,$46,$4F,$3D,$3F,$01,$7D,$67,$F8,$5F,$4F,$F1,$16,$1F,$E9
	.dc.b	$29,$CB,$F1,$AB,$E4,$C6,$27,$73,$FE,$29,$8C,$4F,$CE,$FC,$05,$F7
	.dc.b	$7E,$88,$46,$2B,$DA,$7F,$A8,$1C,$6D,$D2,$77,$5C,$6F,$F1,$31,$E2
	.dc.b	$E5,$C1,$30,$E6,$7C,$52,$1F,$10,$C4,$74,$76,$DB,$F0,$61,$E2,$8F
	.dc.b	$56,$1F,$42,$B9,$13,$3F,$99,$A7,$CE,$26,$FF,$8F,$9F,$C9,$7C,$47
	.dc.b	$FA,$FB,$E2,$BF,$C8,$9E,$F3,$4F,$F1,$9C,$E6,$BF,$D2,$F1,$8E,$E5
	.dc.b	$F0,$8C,$36,$7F,$C9,$D5,$C6,$FA,$46,$2F,$8B,$D0,$C5,$71,$55,$C6
	.dc.b	$FF,$C0,$11,$FF,$00,$08,$7F,$8D,$0A,$68,$C1,$0F,$B5,$9F,$EB,$9A
	.dc.b	$7F,$B7,$0A,$7C,$E7,$4F,$F1,$73,$A9,$4E,$87,$E3,$76,$DE,$BF,$99
	.dc.b	$6E,$FF,$C4,$A7,$F1,$0D,$9F,$B3,$1E,$13,$FC,$FD,$BF,$E4,$5F,$11
	.dc.b	$98,$FC,$23,$CB,$87,$FC,$F5,$F1,$B6,$0D,$24,$F9,$A3,$F8,$79,$EA
	.dc.b	$29,$9E,$17,$0B,$87,$FD,$BF,$3A,$3E,$61,$FC,$0B,$8F,$E2,$C7,$F0
	.dc.b	$98,$F5,$35,$FD,$8B,$6F,$B5,$75,$9F,$15,$F8,$2B,$BB,$63,$33,$B7
	.dc.b	$E7,$D7,$F2,$5E,$3F,$2B,$63,$FD,$EA,$D7,$FB,$1C,$7A,$9F,$51,$F6
	.dc.b	$4F,$59,$4F,$F4,$EF,$37,$4F,$60,$D2,$47,$FA,$58,$FF,$2B,$F3,$7F
	.dc.b	$AA,$2D,$F9,$7F,$F4,$72,$4F,$FA,$D3,$AB,$FD,$75,$1F,$D5,$BA,$FF
	.dc.b	$09,$3F,$EA,$0E,$CF,$EA,$6E,$79,$3E,$8C,$62,$3E,$4E,$FF,$1B,$6F
	.dc.b	$89,$1E,$AC,$BF,$A4,$76,$FC,$4E,$C3,$49,$3C,$6E,$3F,$39,$F9,$3B
	.dc.b	$FF,$87,$61,$D9,$EF,$C3,$FB,$EC,$60,$FF,$02,$94,$FF,$02,$94,$FF
	.dc.b	$63,$F1,$FF,$0C,$7F,$C1,$F3,$FE,$0E,$23,$FD,$83,$1F,$F0,$E3,$FE
	.dc.b	$1A,$F0,$3D,$44,$7C,$E7,$C2,$3F,$35,$E4,$7F,$93,$BE,$43,$FC,$92
	.dc.b	$48,$FF,$17,$BF,$E3,$8C,$E4,$7F,$93,$0F,$92,$FF,$9A,$4F,$F2,$2B
	.dc.b	$FE,$EF,$FC,$F6,$E6,$69,$23,$97,$CA,$7F,$91,$08,$E0,$FF,$1F,$F0
	.dc.b	$7F,$97,$08,$E0,$FF,$65,$C7,$E2,$DF,$F0,$1B,$7A,$3F,$48,$DE,$1F
	.dc.b	$DA,$3F,$5E,$61,$FE,$7C,$F3,$7F,$DF,$91,$FE,$7C,$24,$64,$3F,$E5
	.dc.b	$AF,$06,$47,$C3,$6A,$3F,$DF,$FC,$77,$F3,$EF,$F4,$C9,$23,$FD,$F9
	.dc.b	$19,$7F,$D4,$3F,$4B,$FE,$FC,$23,$83,$FD,$5B,$CC,$DB,$C7,$18,$6B
	.dc.b	$FE,$F7,$8D,$63,$FE,$B6,$3E,$16,$FF,$D2,$27,$FD,$85,$96,$08,$FE
	.dc.b	$14,$87,$ED,$BD,$6F,$06,$BF,$B9,$F5,$BF,$D9,$CD,$EC,$4E,$AB,$FD
	.dc.b	$A1,$3F,$ED,$D5,$F4,$34,$8F,$FB,$8D,$FF,$55,$E1,$23,$E8,$35,$F9
	.dc.b	$83,$DB,$FF,$7C,$C7,$FB,$DF,$5F,$64,$FD,$19,$CB,$FC,$3D,$AF,$2F
	.dc.b	$75,$EA,$7D,$7F,$9F,$FC,$54,$62,$51,$89,$D4,$7F,$19,$9D,$4F,$CC
	.dc.b	$24,$7F,$1D,$E5,$FE,$69,$18,$8F,$4F,$B2,$F6,$FE,$D4,$FE,$66,$31
	.dc.b	$97,$82,$BF,$DD,$75,$A6,$C6,$5F,$F5,$DF,$A3,$FE,$4E,$87,$EF,$1F
	.dc.b	$CB,$78,$D3,$98,$C3,$6B,$E5,$CE,$BC,$1F,$C6,$57,$F1,$7F,$09,$ED
	.dc.b	$FE,$6F,$E7,$70,$DC,$3F,$DE,$F1,$84,$F1,$F1,$FF,$45,$F3,$96,$4D
	.dc.b	$23,$F7,$3A,$5D,$2F,$85,$FC,$F5,$7E,$35,$B5,$B9,$D6,$FD,$23,$13
	.dc.b	$A7,$F4,$51,$F8,$45,$C2,$EC,$FD,$57,$47,$D1,$6C,$FE,$AE,$3F,$86
	.dc.b	$7C,$37,$F8,$C2,$2F,$A5,$FD,$8E,$31,$1F,$D6,$DA,$FF,$64,$3F,$B3
	.dc.b	$93,$9F,$FB,$A3,$FE,$96,$E1,$F7,$3F,$D2,$33,$FE,$D0,$47,$FC,$13
	.dc.b	$5D,$1F,$E1,$8D,$4D,$35,$9A,$BA,$CF,$DB,$AE,$8D,$3F,$C2,$BF,$10
	.dc.b	$F8,$8F,$EF,$CD,$8C,$C6,$0F,$81,$94,$8F,$2B,$5F,$E6,$53,$C5,$B3
	.dc.b	$C5,$E6,$7F,$89,$3E,$27,$FB,$17,$83,$FC,$E5,$B1,$FE,$05,$6B,$FC
	.dc.b	$01,$AF,$C2,$BF,$03,$F4,$5A,$EF,$CC,$FE,$65,$F3,$7D,$5B,$1D,$5A
	.dc.b	$EC,$47,$F4,$BF,$85,$FD,$9F,$05,$ED,$FC,$2D,$7F,$12,$9C,$BF,$E1
	.dc.b	$B6,$BF,$C5,$B8,$3E,$A6,$3E,$48,$F4,$1F,$E4,$4F,$8C,$7C,$6F,$FA
	.dc.b	$BA,$3F,$8E,$C6,$38,$B5,$39,$9F,$E9,$93,$97,$F6,$DF,$1D,$FE,$2F
	.dc.b	$23,$FE,$12,$7C,$0F,$E1,$AE,$E1,$F8,$6F,$8F,$7B,$BD,$72,$46,$1F
	.dc.b	$F2,$1B,$6F,$FC,$34,$D7,$FD,$DF,$9B,$E4,$3F,$C9,$6D,$8F,$F2,$5B
	.dc.b	$5F,$F1,$13,$5F,$B0,$7E,$28,$87,$EB,$8F,$EF,$9F,$22,$F9,$2F,$CE
	.dc.b	$61,$86,$CF,$F9,$2E,$75,$D1,$F1,$DB,$5F,$E7,$2F,$93,$7C,$2F,$F7
	.dc.b	$5E,$3F,$F0,$44,$63,$0F,$E0,$CD,$8C,$CE,$E7,$CF,$1E,$52,$38,$3F
	.dc.b	$A0,$79,$BF,$E8,$18,$FE,$D7,$1F,$DB,$8C,$F8,$FF,$C4,$FE,$53,$EF
	.dc.b	$4F,$F8,$65,$B3,$FE,$65,$6B,$FD,$81,$AF,$F8,$37,$D4,$FE,$B3,$97
	.dc.b	$F1,$D6,$D7,$F8,$8D,$B3,$FE,$6F,$6B,$FD,$AB,$E5,$9F,$2D,$FE,$7D
	.dc.b	$6C,$7F,$A6,$1A,$FF,$8B,$5A,$FF,$9E,$5A,$F3,$8E,$44,$79,$35,$E4
	.dc.b	$D7,$93,$5E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18
	.dc.b	$21,$FE,$39,$8F,$6B,$63,$CC,$D3,$C2,$53,$B1,$4E,$A7,$FB,$D0,$FF
	.dc.b	$2B,$7C,$2F,$EC,$2F,$FD,$8A,$D8,$CF,$27,$F5,$75,$6F,$A1,$FF,$26
	.dc.b	$73,$FD,$36,$CF,$D3,$6B,$F7,$5A,$FA,$DA,$F4,$6C,$F5,$8F,$EB,$2D
	.dc.b	$9E,$D1,$89,$D9,$1C,$12,$36,$47,$C8,$64,$AE,$5F,$31,$92,$B9,$7D
	.dc.b	$06,$4A,$E5,$F5,$19,$2B,$99,$F5,$47,$80,$87,$62,$9D,$0A,$72,$3F
	.dc.b	$DA,$DF,$32,$F9,$AF,$F3,$87,$C9,$FF,$A3,$E3,$1B,$23,$13,$AA,$E2
	.dc.b	$D7,$FC,$B3,$E3,$7F,$BD,$E7,$FC,$E6,$48,$FE,$0A,$FF,$A6,$FE,$3F
	.dc.b	$FA,$17,$C8,$5F,$03,$FD,$00,$DC,$4D,$3C,$2E,$2F,$C2,$D7,$FC,$D3
	.dc.b	$97,$A2,$76,$7F,$2F,$24,$7F,$34,$6B,$FE,$9B,$97,$D9,$FF,$93,$F5
	.dc.b	$FF,$6E,$36,$7F,$CE,$B3,$AF,$4F,$FD,$2D,$5A,$6C,$F9,$23,$CE,$47
	.dc.b	$07,$2F,$B7,$7D,$D7,$E6,$AD,$18,$9F,$F1,$8C,$62,$75,$39,$AF,$F4
	.dc.b	$A7,$A9,$E0,$BF,$F5,$A7,$A5,$FE,$4E,$8C,$42,$31,$87,$F8,$F1,$B5
	.dc.b	$E5,$9F,$93,$C3,$C1,$6B,$F8,$29,$D4,$7F,$AD,$7B,$DF,$17,$E9,$6C
	.dc.b	$7A,$1B,$3F,$E6,$F9,$FD,$AF,$1B,$B5,$DA,$FB,$6F,$F5,$C9,$31,$FE
	.dc.b	$96,$F8,$1F,$1E,$7E,$5C,$FC,$91,$F3,$0F,$C8,$DF,$4E,$6F,$C8,$46
	.dc.b	$30,$FC,$D3,$F9,$23,$FC,$4C,$3E,$D7,$FC,$54,$D3,$2E,$0F,$15,$F8
	.dc.b	$23,$E3,$EF,$FE,$64,$8C,$47,$FB,$97,$CD,$79,$BF,$EE,$7F,$92,$EA
	.dc.b	$49,$F0,$BF,$E4,$CF,$B7,$FE,$33,$E9,$1A,$49,$ED,$B7,$C1,$26,$58
	.dc.b	$3F,$CE,$B5,$C6,$6D,$71,$E2,$ED,$1F,$57,$DE,$FB,$04,$70,$73,$8F
	.dc.b	$7E,$1D,$9F,$F5,$44,$ED,$B3,$C1,$FF,$6C,$A1,$F5,$AF,$BF,$FF,$7E
	.dc.b	$71,$EA,$73,$18,$78,$CE,$F3,$FC,$24,$62,$74,$FD,$0B,$F2,$CF,$9E
	.dc.b	$3F,$C2,$31,$D7,$93,$F1,$DC,$5F,$E9,$86,$C6,$79,$3F,$22,$FA,$2D
	.dc.b	$F8,$DB,$E4,$90,$FA,$7D,$4E,$63,$19,$7F,$04,$8C,$4E,$AB,$8B,$F1
	.dc.b	$3A,$BF,$D8,$3F,$1D,$D3,$EE,$71,$EF,$23,$18,$FF,$B0,$5A,$FE,$F2
	.dc.b	$35,$92,$BF,$DB,$FE,$0F,$9B,$EE,$AC,$9A,$4B,$FF,$71,$FE,$40,$FC
	.dc.b	$B7,$EB,$4F,$FB,$AE,$31,$3C,$4D,$CE,$9A,$64,$FB,$C3,$F0,$D7,$93
	.dc.b	$F8,$1F,$D0,$FE,$F4,$62,$51,$89,$D4,$7E,$22,$75,$3F,$57,$95,$F9
	.dc.b	$B5,$25,$7F,$93,$9F,$C5,$7C,$23,$E6,$BE,$BF,$E6,$58,$6D,$7E,$52
	.dc.b	$75,$B2,$B4,$D9,$D5,$7E,$39,$FC,$D5,$70,$EC,$D9,$23,$F1,$E4,$3F
	.dc.b	$87,$74,$7A,$4F,$BB,$F0,$3F,$8A,$FC,$16,$E2,$69,$2B,$FA,$6F,$ED
	.dc.b	$3F,$90,$5F,$9B,$FE,$56,$30,$DA,$FD,$73,$67,$67,$EA,$BF,$1A,$FD
	.dc.b	$97,$E2,$5F,$AF,$F0,$91,$89,$D0,$9D,$46,$63,$F9,$46,$9F,$CC,$DA
	.dc.b	$7F,$34,$3E,$01,$FB,$FF,$07,$F5,$AE,$03,$87,$FB,$76,$5F,$C0,$D7
	.dc.b	$E2,$BE,$DF,$0F,$F8,$43,$63,$27,$C4,$8C,$BF,$92,$FC,$9F,$F3,$1D
	.dc.b	$3F,$C3,$DB,$06,$9F,$7B,$EC,$A4,$7D,$ED,$D1,$88,$FB,$27,$A7,$2F
	.dc.b	$43,$ED,$F1,$7F,$83,$BF,$23,$F2,$DF,$8E,$9A,$47,$FD,$47,$5F,$9B
	.dc.b	$23,$83,$B4,$F4,$7C,$24,$3F,$23,$CD,$FC,$4B,$CF,$FC,$B9,$23,$D4
	.dc.b	$FE,$3D,$E7,$D5,$FC,$7E,$BA,$47,$F1,$AF,$A1,$7A,$9F,$CD,$7F,$59
	.dc.b	$FE,$41,$18,$94,$62,$75,$1F,$E2,$09,$D4,$F7,$8F,$31,$F7,$AB,$BB
	.dc.b	$6F,$C4,$4E,$AE,$11,$F2,$E3,$0F,$A2,$33,$E3,$BC,$67,$C8,$FF,$32
	.dc.b	$F5,$7F,$93,$B6,$FF,$88,$D6,$E3,$11,$FC,$9D,$AF,$F8,$86,$9F,$3B
	.dc.b	$8A,$EC,$D8,$CF,$1F,$F8,$6A,$BF,$98,$4F,$27,$D9,$8F,$A9,$A7,$A4
	.dc.b	$87,$C4,$E3,$FC,$4B,$A3,$1A,$7C,$68,$C9,$1C,$1F,$62,$DD,$1F,$E2
	.dc.b	$7A,$7E,$13,$FC,$0E,$BF,$B7,$36,$32,$7F,$0A,$8C,$BF,$C3,$3D,$DF
	.dc.b	$F2,$75,$1A,$48,$FA,$67,$9B,$7A,$FD,$7C,$BE,$06,$5E,$79,$FF,$2A
	.dc.b	$DB,$F0,$0F,$F0,$07,$C6,$7F,$19,$C3,$FB,$4C,$E5,$46,$74,$F6,$D7
	.dc.b	$07,$F9,$0B,$67,$C5,$8F,$83,$3F,$8B,$8F,$92,$7E,$61,$E5,$38,$9F
	.dc.b	$DD,$EE,$F9,$FD,$F2,$F1,$37,$1F,$F0,$B3,$E7,$AE,$64,$CF,$DD,$97
	.dc.b	$F8,$17,$F0,$9F,$C4,$CB,$E5,$3D,$51,$86,$F0,$7F,$88,$5B,$3B,$CF
	.dc.b	$FD,$13,$F5,$2F,$AB,$FF,$6F,$C6,$1B,$3F,$E9,$39,$D2,$C7,$E4,$46
	.dc.b	$1F,$67,$87,$C2,$7E,$7B,$6F,$F1,$2D,$DB,$FA,$0C,$F3,$9F,$CB,$BF
	.dc.b	$24,$FE,$20,$AF,$B3,$E5,$8E,$07,$97,$0E,$D3,$A8,$FA,$E7,$43,$E5
	.dc.b	$70,$7D,$8C,$7F,$51,$F0,$EF,$87,$7C,$47,$F9,$1F,$DE,$F7,$61,$FE
	.dc.b	$49,$AD,$4E,$6F,$FC,$1F,$18,$9F,$F3,$6C,$62,$F9,$C6,$0E,$4E,$2F
	.dc.b	$AC,$2B,$94,$61,$B5,$CE,$FC,$15,$FC,$5C,$FF,$0A,$10,$FF,$2F,$F0
	.dc.b	$FF,$93,$9C,$1F,$E8,$F2,$9C,$4F,$94,$7D,$AC,$FF,$D9,$6F,$CB,$3F
	.dc.b	$DA,$45,$3F,$29,$1D,$5E,$F9,$7F,$AA,$1E,$A7,$F2,$F3,$FC,$0D,$B7
	.dc.b	$47,$F0,$67,$42,$9F,$88,$31,$AF,$DF,$BE,$0D,$8C,$EB,$F9,$27,$D0
	.dc.b	$70,$E1,$FD,$1C,$FE,$78,$87,$FB,$99,$CB,$E8,$E4,$FF,$40,$4F,$9C
	.dc.b	$C1,$F1,$A3,$F8,$1A,$FB,$3F,$DB,$36,$3F,$7B,$18,$3E,$2E,$5F,$0E
	.dc.b	$79,$90,$FA,$45,$3E,$91,$82,$1F,$C5,$B9,$7F,$D3,$1F,$77,$E0,$25
	.dc.b	$FF,$0C,$7B,$E1,$EA,$D7,$EA,$5F,$23,$D8,$57,$2F,$80,$95,$FC,$F3
	.dc.b	$A3,$E1,$B6,$3F,$D4,$8D,$7F,$C8,$11,$AE,$4F,$E8,$5E,$BF,$F3,$F2
	.dc.b	$4F,$F3,$F0,$FE,$12,$FC,$31,$E4,$3C,$45,$23,$A2,$67,$F2,$E3,$D4
	.dc.b	$D7,$8C,$6B,$1F,$85,$39,$9E,$DD,$9F,$DB,$32,$7E,$19,$FE,$9E,$E9
	.dc.b	$FF,$55,$B6,$33,$5F,$DE,$BB,$7F,$B8,$CF,$F7,$E9,$CC,$FF,$7B,$6B
	.dc.b	$FD,$EF,$BB,$9B,$C5,$AF,$A3,$2F,$29,$1F,$89,$97,$9F,$3F,$54,$62
	.dc.b	$74,$FE,$1E,$DE,$0F,$F8,$97,$2F,$F1,$24,$FF,$85,$4F,$F1,$27,$85
	.dc.b	$FE,$DF,$8C,$9C,$E2,$1F,$E4,$FF,$C2,$7F,$D1,$E4,$70,$7E,$54,$FA
	.dc.b	$79,$79,$49,$8B,$E6,$7F,$D4,$84,$9F,$D6,$3F,$E5,$2F,$CA,$E4,$7F
	.dc.b	$9F,$1A,$FC,$33,$C6,$7F,$AE,$3D,$4F,$F5,$1C,$FF,$A8,$89,$1F,$E6
	.dc.b	$57,$FC,$D7,$F9,$C4,$8F,$F6,$0B,$5F,$F3,$DF,$9D,$FE,$FF,$FD,$0B
	.dc.b	$FC,$F0,$48,$FF,$3C,$07,$C2,$FE,$2B,$2F,$6F,$E0,$BF,$EB,$F2,$47
	.dc.b	$F0,$77,$FC,$DB,$F8,$B4,$8F,$F4,$BB,$5F,$F4,$5F,$09,$7E,$71,$FE
	.dc.b	$A3,$23,$FE,$AC,$8D,$47,$FA,$CB,$9A,$4F,$E4,$7D,$9F,$F7,$FC,$62
	.dc.b	$3E,$41,$F5,$79,$7F,$57,$BF,$FB,$2B,$2F,$F6,$49,$27,$FD,$18,$FF
	.dc.b	$B3,$3E,$76,$47,$FA,$3D,$AF,$FB,$A2,$9A,$38,$9F,$83,$7F,$B8,$49
	.dc.b	$5F,$CA,$5F,$F7,$42,$FC,$9E,$91,$FF,$75,$7C,$0F,$C3,$8F,$F5,$1F
	.dc.b	$E1,$52,$7F,$DF,$2D,$7E,$45,$BE,$A7,$F8,$4F,$0A,$BF,$DF,$B5,$E5
	.dc.b	$9C,$C6,$2B,$87,$7A,$F8,$5F,$D3,$38,$46,$6B,$9F,$7F,$F9,$F7,$E1
	.dc.b	$F8,$7E,$1D,$8E,$D4,$63,$0F,$E0,$8D,$7E,$F5,$71,$C5,$7E,$65,$F1
	.dc.b	$4F,$8A,$F2,$79,$19,$8C,$4F,$F4,$DC,$3C,$CF,$0A,$DC,$62,$77,$5F
	.dc.b	$98,$8F,$DA,$EB,$FD,$6B,$E4,$5F,$1B,$FE,$49,$9F,$C9,$46,$31,$D4
	.dc.b	$E6,$53,$97,$FD,$E9,$DD,$F2,$7A,$2D,$9A,$49,$F3,$FC,$9B,$E3,$BF
	.dc.b	$A4,$B6,$3F,$A4,$B5,$F8,$8D,$7E,$DE,$1E,$16,$E3,$E5,$8C,$47,$D9
	.dc.b	$6B,$F7,$67,$93,$FA,$FC,$11,$88,$D5,$F9,$0F,$EF,$EB,$E4,$1F,$21
	.dc.b	$86,$C7,$F8,$11,$AF,$E4,$0E,$11,$CB,$0F,$1C,$7C,$AC,$7D,$04,$C7
	.dc.b	$95,$75,$24,$E0,$FF,$08,$FC,$93,$D7,$FD,$0C,$7E,$7A,$30,$FF,$5F
	.dc.b	$6E,$DE,$36,$C7,$C8,$BF,$BA,$47,$D1,$8F,$9A,$79,$F9,$1C,$70,$FE
	.dc.b	$2C,$DA,$34,$F8,$1D,$DF,$C1,$31,$FF,$2D,$CF,$F0,$7E,$EF,$E0,$DE
	.dc.b	$AE,$9B,$1F,$E2,$D8,$F0,$EB,$F0,$E2,$E9,$E5,$7F,$96,$10,$E9,$7F
	.dc.b	$8F,$70,$7D,$FF,$77,$F1,$A4,$7F,$1A,$FD,$3A,$ED,$1F,$0C,$EE,$7E
	.dc.b	$EF,$CB,$FD,$27,$4F,$FB,$ED,$BD,$66,$9F,$A9,$C5,$F2,$8F,$95,$FF
	.dc.b	$12,$B7,$0F,$F1,$2B,$5F,$E0,$0D,$7C,$44,$3F,$35,$8F,$99,$FC,$C7
	.dc.b	$CB,$7F,$1E,$7F,$84,$35,$FE,$06,$D7,$7E,$07,$EE,$7C,$5F,$F4,$AF
	.dc.b	$D2,$FE,$55,$B3,$FC,$99,$AF,$F2,$16,$BF,$C1,$2F,$F9,$34,$FD,$42
	.dc.b	$47,$E2,$DF,$D0,$78,$BF,$EF,$5C,$D1,$8F,$53,$8D,$7F,$34,$AD,$4E
	.dc.b	$6B,$73,$F8,$32,$1F,$7B,$8B,$F1,$1C,$5F,$8C,$8F,$24,$D7,$F5,$8D
	.dc.b	$E8,$FE,$35,$5F,$8A,$3F,$69,$F3,$2F,$9A,$FD,$EB,$6F,$FC,$FA,$D7
	.dc.b	$71,$A2,$3A,$6B,$A3,$04,$30,$43,$FA,$9C,$75,$6C,$72,$69,$C4,$A7
	.dc.b	$13,$FA,$A4,$71,$6C,$71,$69,$C4,$A7,$13,$FC,$1D,$1C,$5B,$1C,$5A
	.dc.b	$71,$29,$C4,$47,$16,$C7,$16,$9C,$4A,$7F,$23,$F7,$BF,$DC,$0F,$F7
	.dc.b	$46,$BF,$ED,$73,$5E,$FF,$DB,$7B,$FC,$87,$F8,$97,$E0,$3E,$CB,$63
	.dc.b	$E5,$B5,$F8,$AD,$7D,$ED,$79,$9F,$EE,$A3,$25,$33,$1D,$9C,$60,$9E
	.dc.b	$4D,$8C,$C6,$08,$E0,$93,$CC,$8F,$A0,$C9,$5C,$BE,$A3,$25,$72,$FA
	.dc.b	$1A,$FA,$0E,$44,$38,$14,$E0,$53,$81,$A2,$1A,$28,$EB,$FE,$79,$F3
	.dc.b	$3F,$B2,$24,$7F,$B9,$5F,$E1,$93,$FE,$5B,$6C,$1A,$47,$FC,$6B,$F8
	.dc.b	$97,$FA,$5B,$67,$CD,$FF,$69,$8C,$37,$DA,$FF,$1B,$36,$77,$5F,$EF
	.dc.b	$58,$FE,$0A,$47,$07,$F9,$0B,$E8,$1F,$41,$FE,$6F,$9F,$EE,$D1,$89
	.dc.b	$E6,$7F,$90,$5C,$46,$A7,$84,$FF,$93,$3C,$DF,$F3,$0D,$FE,$25,$B9
	.dc.b	$FF,$80,$5B,$3B,$F0,$BF,$C9,$D3,$FC,$68,$91,$FE,$56,$D7,$FB,$44
	.dc.b	$62,$3F,$CA,$5F,$83,$B3,$FC,$D3,$63,$F3,$0D,$9F,$F0,$83,$63,$CD
	.dc.b	$3E,$06,$E3,$E7,$8C,$47,$F9,$7F,$1E,$D3,$FA,$3C,$3C,$73,$E5,$C7
	.dc.b	$FC,$FD,$8F,$96,$3F,$D3,$13,$A7,$CF,$3E,$62,$3E,$67,$F6,$5D,$6A
	.dc.b	$FF,$50,$7C,$9F,$E1,$89,$5F,$D4,$E3,$2F,$87,$A3,$C9,$7E,$28,$C4
	.dc.b	$F8,$D5,$F5,$7F,$20,$73,$29,$FD,$76,$3F,$2C,$E7,$87,$FC,$71,$FB
	.dc.b	$03,$97,$FD,$69,$F1,$D8,$6D,$7F,$A3,$27,$4A,$74,$D3,$CA,$43,$FC
	.dc.b	$2F,$7E,$79,$E3,$F3,$FC,$09,$3D,$6F,$FD,$89,$1F,$E4,$BF,$13,$FC
	.dc.b	$21,$3F,$C5,$DB,$79,$7F,$8F,$18,$8F,$01,$F5,$B3,$F9,$0F,$88,$C9
	.dc.b	$FE,$1B,$8E,$24,$7B,$3D,$3F,$16,$91,$E1,$6D,$7D,$1D,$FF,$05,$3D
	.dc.b	$F6,$FA,$E6,$D9,$3E,$BA,$E3,$1F,$EB,$6F,$C3,$9E,$1F,$E0,$23,$FD
	.dc.b	$65,$18,$5C,$3F,$E6,$D8,$C3,$67,$84,$F8,$6F,$C9,$9F,$C0,$8C,$1F
	.dc.b	$EC,$4F,$7B,$FC,$2D,$C3,$FE,$77,$6C,$67,$C3,$FF,$62,$4E,$5B,$1F
	.dc.b	$EC,$5F,$25,$E4,$FF,$B4,$3E,$87,$FD,$85,$9F,$FB,$2D,$BC,$E6,$9F
	.dc.b	$EB,$FF,$87,$7C,$67,$FC,$01,$8C,$37,$1F,$F5,$EC,$E9,$E4,$62,$34
	.dc.b	$FF,$A6,$E1,$F8,$A2,$1F,$45,$DE,$30,$D8,$C9,$DA,$32,$FF,$92,$BF
	.dc.b	$22,$FF,$72,$7E,$0F,$FE,$4E,$6C,$67,$24,$E5,$B1,$FE,$F0,$F8,$3F
	.dc.b	$F7,$9F,$D0,$FF,$A8,$DB,$19,$D7,$FD,$E1,$39,$6C,$7F,$BE,$3E,$D7
	.dc.b	$E2,$DB,$E8,$BF,$D5,$91,$86,$ED,$FE,$F9,$6C,$EE,$BF,$E0,$0F,$D0
	.dc.b	$FF,$1C,$F6,$BF,$D6,$8D,$8C,$F8,$BF,$F0,$06,$72,$D8,$F1,$61,$FC
	.dc.b	$99,$F9,$98,$F9,$3A,$7B,$92,$3D,$98,$FF,$0E,$7E,$96,$BF,$9C,$8C
	.dc.b	$42,$31,$CF,$F0,$67,$87,$37,$E3,$3C,$04,$3E,$8B,$EE,$61,$F7,$6B
	.dc.b	$53,$98,$FC,$A7,$42,$F8,$DF,$09,$5F,$08,$FC,$97,$D0,$7E,$1F,$F3
	.dc.b	$AB,$46,$91,$FD,$B7,$1F,$E3,$7E,$7F,$F6,$10,$69,$1F,$E1,$BE,$1F
	.dc.b	$F0,$47,$FC,$15,$DF,$FB,$36,$F8,$7F,$E3,$1F,$09,$F2,$7F,$CD,$27
	.dc.b	$F8,$0B,$5F,$CE,$46,$9F,$96,$DE,$1F,$CE,$7D,$0C,$9F,$CE,$07,$0F
	.dc.b	$FA,$AB,$B7,$E6,$FE,$1D,$F9,$92,$47,$E7,$5F,$F7,$D7,$E0,$CF,$B2
	.dc.b	$FE,$1B,$28,$C3,$FC,$50,$F9,$CE,$3C,$7E,$33,$C3,$A3,$F0,$5E,$8F
	.dc.b	$DF,$25,$FE,$21,$F9,$7E,$6C,$8F,$9E,$7F,$4B,$3F,$91,$F1,$FF,$52
	.dc.b	$7F,$9D,$10,$FE,$3A,$51,$B7,$C7,$AF,$E5,$1B,$FF,$86,$A7,$E6,$9F
	.dc.b	$E2,$03,$E6,$B9,$7F,$85,$7C,$13,$87,$F2,$7A,$7E,$2E,$53,$9E,$BF
	.dc.b	$85,$1F,$96,$8C,$4E,$DB,$1F,$CF,$EF,$F8,$B7,$E1,$1E,$48,$34,$8F
	.dc.b	$E6,$7F,$4A,$72,$FF,$1D,$FA,$8F,$14,$62,$51,$89,$D4,$71,$6B,$F0
	.dc.b	$3F,$27,$F0,$11,$F2,$07,$FA,$4E,$9F,$D6,$BC,$0F,$F1,$06,$AF,$C1
	.dc.b	$DF,$6B,$FE,$8F,$8C,$36,$33,$87,$FA,$3E,$72,$FF,$A4,$7C,$0F,$E9
	.dc.b	$3F,$53,$FB,$EF,$88,$FD,$B4,$62,$91,$8F,$11,$5B,$BF,$DF,$B5,$FE
	.dc.b	$85,$F1,$1F,$D0,$6B,$FA,$73,$63,$FA,$5B,$5D,$9D,$8F,$DB,$ED,$FB
	.dc.b	$5F,$03,$F7,$07,$F7,$CC,$7E,$07,$DC,$7B,$49,$5E,$DD,$11,$FE,$AF
	.dc.b	$AE,$37,$CC,$9B,$FF,$AC,$7E,$C7,$FC,$5D,$F0,$7F,$EB,$76,$E3,$9F
	.dc.b	$67,$FD,$E9,$C5,$FC,$66,$72,$FF,$69,$FC,$67,$FA,$F7,$DE,$DB,$34
	.dc.b	$8F,$F6,$AC,$3E,$B7,$AD,$D2,$30,$7D,$57,$0F,$EA,$74,$F4,$E7,$F8
	.dc.b	$18,$FE,$63,$BF,$E0,$7E,$77,$EF,$12,$3E,$13,$F8,$1E,$2F,$EE,$5F
	.dc.b	$59,$FC,$EA,$7F,$BA,$57,$E2,$67,$2F,$4E,$2F,$A5,$5B,$AF,$E8,$31
	.dc.b	$89,$FA,$B8,$78,$E7,$CD,$F2,$DF,$EE,$59,$FF,$45,$36,$3F,$A3,$E4
	.dc.b	$8C,$47,$F9,$E1,$B3,$E3,$87,$BB,$E1,$8C,$37,$0F,$C3,$37,$2F,$F0
	.dc.b	$73,$E7,$AF,$E7,$7C,$1F,$A8,$3F,$4F,$5F,$8D,$E6,$E1,$3F,$9A,$C3
	.dc.b	$E0,$61,$FE,$17,$9C,$9E,$1D,$BF,$DD,$D8,$5F,$86,$F3,$C6,$2F,$FC
	.dc.b	$4B,$18,$26,$1F,$E8,$D8,$CB,$4F,$F1,$34,$7C,$27,$CE,$7C,$CA,$FD
	.dc.b	$ED,$75,$8D,$9C,$4F,$EE,$6E,$5B,$3C,$F4,$FC,$3D,$7C,$19,$E8,$43
	.dc.b	$FB,$CF,$C4,$7F,$A3,$7E,$27,$F9,$6C,$E3,$99,$18,$AF,$F1,$D5,$6D
	.dc.b	$FE,$5F,$5B,$9D,$42,$75,$E1,$7F,$A8,$63,$17,$D1,$E2,$74,$71,$1B
	.dc.b	$9E,$37,$F8,$8E,$8F,$E8,$17,$FC,$FF,$1F,$E6,$3C,$5F,$D0,$63,$FB
	.dc.b	$C9,$99,$DB,$FE,$B7,$31,$1B,$9E,$31,$83,$FC,$C5,$9F,$EB,$1F,$9E
	.dc.b	$7D,$F3,$F4,$E6,$35,$E8,$C3,$CD,$5D,$6F,$C3,$1B,$21,$F5,$8E,$65
	.dc.b	$3B,$12,$7F,$0E,$D3,$F6,$FC,$1F,$C9,$7E,$13,$F3,$24,$9F,$CC,$CA
	.dc.b	$3F,$45,$8F,$BA,$3C,$C7,$FA,$8C,$F8,$07,$0D,$7F,$46,$4A,$C1,$FB
	.dc.b	$A8,$F3,$89,$FD,$D3,$86,$BE,$0E,$3F,$94,$F7,$3C,$27,$D6,$3C,$73
	.dc.b	$FB,$FF,$1F,$F6,$64,$7D,$6F,$EF,$F9,$BE,$F4,$7F,$1F,$9F,$F0,$03
	.dc.b	$5F,$F2,$63,$5E,$F7,$F0,$5F,$AA,$7C,$03,$FD,$2D,$CF,$D3,$8C,$D2
	.dc.b	$47,$68,$74,$FF,$A6,$FF,$79,$FE,$7C,$49,$FF,$52,$C6,$5F,$F3,$2F
	.dc.b	$5F,$F4,$A7,$FA,$46,$BD,$9F,$F0,$C4,$65,$DC,$65,$F3,$1F,$24,$F4
	.dc.b	$9F,$23,$7F,$EB,$AF,$C2,$6B,$F0,$9A,$FF,$3B,$EC,$FD,$33,$F4,$4E
	.dc.b	$64,$3B,$1F,$C3,$3B,$7F,$D7,$1F,$4B,$FD,$E2,$0D,$23,$FE,$00,$DB
	.dc.b	$2D,$8F,$F0,$37,$57,$FA,$21,$B3,$A9,$FC,$DF,$83,$F7,$31,$FF,$0A
	.dc.b	$C7,$A6,$30,$7A,$38,$3E,$D4,$79,$8F,$51,$7C,$9F,$F3,$97,$DB,$7F
	.dc.b	$8E,$49,$1F,$E2,$A7,$FC,$64,$BE,$8A,$91,$FF,$1A,$19,$3F,$C6,$3B
	.dc.b	$FF,$90,$C8,$E0,$FF,$4D,$76,$FF,$93,$FE,$67,$FC,$99,$D7,$FE,$A6
	.dc.b	$6E,$A6,$9F,$E4,$DF,$90,$97,$FE,$3E,$6B,$F4,$CF,$69,$34,$F8,$2F
	.dc.b	$F9,$AB,$A7,$FD,$D9,$F2,$BF,$EA,$02,$47,$F9,$D9,$FF,$5D,$54,$8F
	.dc.b	$F6,$1B,$5F,$F5,$CF,$A5,$FE,$88,$C3,$FD,$0B,$1E,$73,$FD,$13,$E9
	.dc.b	$7F,$BE,$63,$27,$FB,$B7,$F2,$59,$3F,$EF,$16,$BF,$F0,$07,$E1,$7F
	.dc.b	$E0,$07,$D2,$FF,$BF,$49,$1F,$EA,$27,$FC,$DD,$E3,$48,$FF,$2F,$B5
	.dc.b	$F8,$5A,$FC,$02,$78,$7F,$CB,$23,$FD,$63,$B1,$7D,$37,$FB,$08,$8F
	.dc.b	$FA,$61,$FF,$65,$FE,$6F,$23,$FC,$53,$E0,$BD,$EE,$53,$99,$FF,$6E
	.dc.b	$79,$DF,$ED,$9E,$47,$C6,$7F,$98,$7E,$07,$FD,$C6,$D8,$CD,$7F,$B8
	.dc.b	$27,$2F,$FB,$CB,$6F,$F7,$61,$1C,$1F,$EF,$C9,$FE,$B5,$F0,$4F,$83
	.dc.b	$FF,$0D,$4F,$FB,$F9,$AA,$35,$3F,$D8,$7E,$19,$F0,$BF,$EF,$63,$FD
	.dc.b	$48,$E2,$33,$5F,$EF,$18,$C4,$E8,$E1,$1E,$3F,$8F,$FF,$56,$5F,$66
	.dc.b	$E0,$C1,$18,$6D,$7F,$BF,$7D,$F7,$16,$FE,$97,$FD,$6D,$18,$A4,$62
	.dc.b	$75,$5C,$5A,$FF,$B9,$65,$F8,$1F,$F3,$EF,$9A,$7C,$27,$8E,$30,$DA
	.dc.b	$FF,$32,$CE,$AB,$8C,$7F,$BC,$BE,$93,$FA,$3F,$33,$AF,$F1,$51,$86
	.dc.b	$CE,$CF,$C3,$ED,$F8,$FF,$A3,$E2,$D8,$C8,$8C,$B4,$FC,$B0,$F7,$5C
	.dc.b	$FF,$0E,$7F,$D5,$9B,$7C,$D6,$E7,$C6,$FF,$27,$39,$79,$9F,$94,$F9
	.dc.b	$2F,$A9,$CD,$FE,$E3,$E8,$46,$27,$71,$8B,$FF,$19,$47,$92,$31,$5F
	.dc.b	$D4,$23,$F8,$87,$67,$EB,$DF,$E2,$7E,$57,$DC,$7F,$83,$1E,$C7,$1E
	.dc.b	$0F,$FB,$CD,$B1,$E6,$6B,$F0,$BE,$4F,$FC,$FA,$FF,$8E,$5A,$FF,$BE
	.dc.b	$DA,$FB,$9A,$FF,$04,$AF,$27,$81,$FC,$13,$C7,$E5,$EE,$FF,$5E,$BD
	.dc.b	$DF,$F0,$56,$A6,$95,$F1,$3C,$48,$FF,$8E,$E6,$3F,$16,$BF,$99,$CF
	.dc.b	$18,$F4,$3C,$1A,$7F,$95,$7E,$51,$EA,$7F,$17,$BF,$B3,$29,$CB,$FC
	.dc.b	$75,$B1,$A6,$C7,$F8,$73,$1F,$82,$49,$ED,$5F,$D1,$27,$F0,$B5,$D3
	.dc.b	$CA,$E5,$39,$9F,$C9,$CE,$5F,$C5,$F7,$3D,$CF,$E5,$9B,$F9,$A3,$13
	.dc.b	$F8,$83,$F9,$83,$88,$D4,$F0,$9F,$36,$3F,$21,$FF,$20,$D7,$2D,$F7
	.dc.b	$1F,$8A,$39,$D7,$F4,$F7,$C6,$D7,$F5,$3F,$2A,$D3,$FC,$36,$D8,$FF
	.dc.b	$41,$35,$F9,$AD,$7C,$67,$D9,$8F,$CE,$7B,$9F,$C9,$63,$E9,$B5,$FD
	.dc.b	$6B,$5F,$DC,$B5,$FE,$A1,$3F,$CB,$1E,$FE,$1F,$90,$8F,$C1,$3F,$31
	.dc.b	$95,$7F,$AC,$57,$AF,$93,$FD,$34,$D8,$F1,$E1,$FD,$02,$31,$1F,$C0
	.dc.b	$2D,$F2,$DF,$D8,$7C,$EF,$C5,$C6,$25,$18,$9D,$47,$16,$BF,$E9,$88
	.dc.b	$FF,$52,$91,$FE,$1B,$C7,$FE,$AE,$8F,$CB,$FC,$C3,$E6,$3F,$A1,$46
	.dc.b	$1B,$5F,$C5,$27,$59,$F8,$A3,$13,$A7,$F9,$1F,$07,$E2,$5F,$BE,$7E
	.dc.b	$34,$FE,$9B,$BF,$E2,$DD,$B5,$FF,$00,$63,$FE,$C2,$E1,$FF,$5E,$92
	.dc.b	$3D,$3B,$37,$E9,$F3,$2F,$81,$FF,$31,$B6,$BF,$D9,$4D,$7F,$6A,$D7
	.dc.b	$F9,$7B,$5F,$E7,$6D,$7F,$C2,$4D,$4D,$24,$78,$09,$1E,$06,$BE,$06
	.dc.b	$BE,$06,$BE,$06,$BE,$06,$BE,$06,$C7,$82,$3B,$10,$EC,$53,$B1,$4E
	.dc.b	$C5,$3B,$14,$EC,$57,$B3,$FC,$73,$E0,$BF,$8F,$E1,$FE,$F5,$6C,$67
	.dc.b	$9B,$EF,$28,$CB,$FD,$73,$F7,$E7,$E3,$DF,$E3,$DD,$7F,$93,$C6,$27
	.dc.b	$F8,$99,$FD,$B1,$C4,$6A,$78,$4F,$36,$BB,$6B,$FE,$49,$9E,$2E,$CE
	.dc.b	$A6,$63,$6A,$36,$F3,$36,$66,$37,$1A,$8E,$46,$9C,$3B,$31,$1C,$08
	.dc.b	$E0,$93,$ED,$69,$E7,$70,$47,$04,$3C,$CE,$5A,$64,$FF,$35,$7A,$FF
	.dc.b	$E3,$5F,$92,$FF,$6A,$F8,$3F,$E3,$F6,$F1,$9A,$7F,$5A,$E1,$FE,$69
	.dc.b	$C1,$9C,$3F,$62,$D7,$FC,$43,$1A,$9E,$1C,$3F,$B5,$6D,$7C,$A9,$E7
	.dc.b	$96,$09,$7F,$E2,$AF,$13,$FC,$53,$F8,$6F,$96,$31,$E2,$A3,$11,$FE
	.dc.b	$7A,$6B,$FE,$31,$8D,$67,$E5,$8F,$F2,$E7,$92,$F8,$8F,$F4,$67,$C8
	.dc.b	$FF,$91,$9B,$3F,$E8,$86,$BF,$E9,$1A,$57,$18,$57,$A1,$F8,$3B,$7F
	.dc.b	$13,$CF,$11,$FC,$6A,$30,$FA,$BC,$DF,$F3,$77,$A7,$FE,$3E,$D7,$FC
	.dc.b	$8F,$18,$9D,$64,$AD,$69,$FE,$5F,$7F,$C8,$B0,$FC,$7D,$E0,$FE,$82
	.dc.b	$30,$D8,$CD,$FE,$82,$72,$FD,$3C,$BC,$AF,$AB,$87,$04,$D3,$C3,$F2
	.dc.b	$7F,$E8,$56,$DF,$EA,$B5,$FF,$2F,$CE,$6F,$EB,$B5,$FF,$32,$FE,$A5
	.dc.b	$FD,$35,$FA,$81,$C7,$27,$F8,$A1,$B1,$FE,$62,$6B,$E5,$C3,$E6,$AE
	.dc.b	$4F,$C5,$BF,$94,$8E,$5D,$FF,$EC,$9F,$93,$B2,$69,$1F,$F3,$7F,$E8
	.dc.b	$3F,$71,$D1,$8E,$4E,$4D,$32,$7F,$A0,$23,$84,$F9,$30,$F1,$CF,$9C
	.dc.b	$F2,$1F,$E7,$CF,$D0,$5F,$6D,$F9,$88,$46,$37,$F8,$AD,$D7,$FC,$96
	.dc.b	$D9,$FF,$47,$CF,$E0,$7F,$1E,$FF,$73,$79,$DF,$84,$E4,$46,$2B,$71
	.dc.b	$87,$FC,$E3,$5C,$5F,$F3,$8E,$5E,$87,$F3,$1C,$7F,$81,$7D,$27,$FB
	.dc.b	$A3,$E4,$7F,$D8,$DF,$15,$F9,$46,$C7,$ED,$9A,$EE,$D3,$FE,$32,$FB
	.dc.b	$4F,$C2,$4F,$F8,$99,$AF,$E6,$9B,$D0,$73,$8F,$96,$7F,$01,$D1,$F9
	.dc.b	$87,$CF,$CB,$F4,$63,$11,$AC,$FD,$4F,$D3,$3F,$E0,$0F,$D5,$78,$BA
	.dc.b	$7F,$D6,$ED,$EB,$34,$F0,$DF,$4C,$7C,$44,$7B,$F9,$DF,$E5,$7C,$BF
	.dc.b	$CA,$B1,$FD,$32,$31,$8F,$FA,$F1,$A8,$FE,$31,$F8,$6F,$FB,$C7,$E6
	.dc.b	$FF,$7E,$D8,$FE,$42,$D7,$86,$7F,$EC,$E8,$C4,$7C,$23,$EB,$94,$F8
	.dc.b	$07,$7F,$CB,$1F,$51,$FD,$81,$B1,$FD,$79,$B9,$7F,$A6,$9B,$1F,$8A
	.dc.b	$F7,$BD,$1F,$81,$7E,$3D,$B1,$91,$19,$69,$F9,$2F,$33,$F2,$8F,$F2
	.dc.b	$DE,$7F,$F5,$2F,$4F,$E6,$67,$32,$E7,$FE,$3D,$18,$7F,$90,$7A,$FF
	.dc.b	$27,$F1,$AF,$D0,$36,$BF,$3E,$D7,$B3,$5F,$E8,$DF,$57,$D7,$EA,$3F
	.dc.b	$96,$B6,$3F,$A7,$B5,$F9,$FC,$E8,$C4,$79,$73,$FE,$2A,$DF,$13,$EE
	.dc.b	$78,$1F,$B3,$6C,$7E,$C9,$AF,$D4,$3F,$2D,$C3,$F9,$9E,$AF,$E6,$D6
	.dc.b	$FC,$5B,$75,$DD,$BB,$7F,$41,$6C,$EE,$7F,$96,$FD,$B7,$E6,$9F,$F0
	.dc.b	$CB,$70,$FF,$6A,$B6,$3F,$24,$7E,$4D,$C3,$F0,$38,$7F,$B6,$F3,$B6
	.dc.b	$46,$1B,$97,$F6,$F9,$D4,$FF,$74,$8F,$EA,$DF,$51,$F9,$FF,$C8,$7F
	.dc.b	$8B,$37,$0F,$E2,$AD,$7F,$08,$D7,$E7,$7D,$97,$E9,$FF,$08,$FE,$3C
	.dc.b	$D8,$FE,$3A,$D7,$F1,$8D,$7F,$5B,$7F,$DA,$DF,$D5,$B5,$FB,$BD,$7F
	.dc.b	$DC,$5B,$19,$7F,$54,$43,$EB,$9F,$AE,$CB,$E2,$79,$5F,$0C,$93,$F0
	.dc.b	$F4,$FF,$03,$BF,$E1,$33,$F3,$34,$E5,$FC,$CB,$4F,$CC,$94,$FC,$C9
	.dc.b	$4F,$C8,$7B,$BF,$C2,$7F,$2D,$FE,$80,$D8,$FE,$7E,$D7,$F5,$CD,$7E
	.dc.b	$F7,$87,$FE,$60,$E1,$FB,$F1,$83,$F0,$F1,$FE,$09,$FC,$8F,$F1,$A4
	.dc.b	$8F,$57,$22,$3F,$2B,$AF,$F1,$3C,$FF,$AE,$D7,$CE,$F2,$FF,$8D,$C6
	.dc.b	$23,$E6,$C7,$E7,$F9,$64,$7F,$8F,$B7,$FE,$37,$7F,$32,$DE,$17,$F8
	.dc.b	$8A,$73,$C1,$F3,$23,$FC,$55,$EA,$FF,$07,$AF,$C9,$37,$AF,$31,$F9
	.dc.b	$A8,$C6,$3D,$E1,$F5,$1D,$5B,$19,$8C,$1F,$E3,$0F,$CA,$3F,$4F,$76
	.dc.b	$7F,$67,$18,$7F,$BA,$B5,$FE,$13,$97,$F1,$BF,$A7,$EC,$48,$F1,$3F
	.dc.b	$09,$FE,$09,$E4,$FF,$73,$C5,$6F,$98,$FF,$1F,$36,$BF,$C7,$8D,$DF
	.dc.b	$CD,$3A,$FC,$4F,$FA,$77,$EA,$FE,$96,$DF,$99,$AF,$F9,$7E,$D1,$88
	.dc.b	$FE,$AD,$FB,$9F,$F0,$DD,$7F,$83,$C6,$3C,$1F,$E2,$AA,$74,$E9,$AF
	.dc.b	$B3,$CD,$FE,$D4,$4F,$0F,$FC,$95,$C4,$FB,$AF,$EE,$8D,$DF,$31,$83
	.dc.b	$FB,$9C,$FF,$7C,$BF,$DA,$70,$7C,$7D,$5F,$17,$F1,$89,$3F,$19,$FF
	.dc.b	$62,$F3,$B2,$7C,$F7,$F4,$FF,$D9,$FF,$16,$DB,$FF,$4A,$4E,$B4,$56
	.dc.b	$9B,$3F,$E9,$9F,$D5,$FF,$A3,$43,$D6,$FF,$3F,$46,$1B,$E5,$FF,$5D
	.dc.b	$6C,$EE,$BF,$BC,$9F,$C3,$8F,$91,$B7,$DE,$FA,$BF,$C0,$92,$30,$7C
	.dc.b	$AF,$8D,$FF,$1A,$D7,$A1,$B1,$FE,$08,$CD,$18,$86,$BF,$0E,$31,$1E
	.dc.b	$13,$E6,$F1,$7D,$7E,$9F,$D1,$78,$7F,$EF,$9C,$7F,$3E,$DA,$FC,$53
	.dc.b	$5F,$EF,$5C,$48,$C6,$3D,$EB,$FA,$F1,$E7,$2D,$79,$F1,$FF,$01,$73
	.dc.b	$7F,$80,$E3,$13,$A7,$CB,$AF,$E0,$DF,$45,$FD,$DC,$F8,$D7,$DA,$6C
	.dc.b	$FC,$36,$BF,$0B,$44,$62,$BF,$0D,$F2,$5F,$EA,$1B,$FC,$1B,$6F,$E6
	.dc.b	$35,$E4,$D9,$FC,$77,$AF,$70,$FF,$2E,$5B,$6F,$66,$1E,$1A,$D1,$CD
	.dc.b	$FA,$44,$E6,$FF,$28,$67,$FA,$8C,$BA,$13,$0E,$87,$98,$B5,$F8,$AC
	.dc.b	$3F,$87,$3F,$21,$AB,$E2,$1E,$9B,$E2,$7F,$CA,$DD,$1E,$7A,$FF,$2C
	.dc.b	$56,$A7,$37,$FE,$0A,$8C,$4F,$EB,$23,$17,$CE,$30,$72,$71,$7D,$63
	.dc.b	$C5,$18,$9E,$37,$DA,$78,$35,$FF,$3F,$1F,$C5,$0A,$7F,$16,$CF,$ED
	.dc.b	$E7,$98,$C9,$FE,$DF,$38,$46,$9B,$BA,$78,$4E,$67,$E6,$1F,$63,$7F
	.dc.b	$C2,$E9,$FE,$7B,$32,$48,$EC,$78,$EF,$A3,$DA,$FA,$C6,$23,$F8,$7E
	.dc.b	$DF,$9F,$AA,$75,$8F,$DD,$46,$5B,$19,$7F,$22,$DB,$F2,$C7,$9F,$76
	.dc.b	$FC,$63,$66,$70,$FF,$38,$6D,$70,$FE,$41,$3F,$A7,$6B,$FA,$79,$DC
	.dc.b	$7B,$72,$F5,$92,$7D,$C7,$F7,$6E,$1D,$BF,$BA,$BF,$24,$62,$3F,$64
	.dc.b	$DA,$FE,$C3,$C5,$FE,$E0,$9F,$E1,$A7,$BE,$32,$DE,$3F,$E1,$CF,$C3
	.dc.b	$3E,$F9,$FE,$DC,$3C,$67,$9C,$F2,$43,$DA,$F1,$B4,$FE,$28,$E0,$FD
	.dc.b	$BB,$FD,$D3,$97,$F1,$8D,$9F,$C9,$35,$FF,$04,$47,$E4,$5F,$19,$38
	.dc.b	$3F,$18,$FE,$18,$FF,$5E,$F4,$7F,$4A,$FA,$6F,$E8,$A4,$8F,$E9,$DD
	.dc.b	$3F,$EA,$58,$FE,$A5,$F1,$72,$7F,$AB,$ED,$FE,$A9,$8C,$47,$F7,$DE
	.dc.b	$CF,$3E,$BF,$D9,$A3,$CE,$D7,$FC,$3D,$1E,$7C,$BC,$B1,$F2,$4F,$90
	.dc.b	$43,$FA,$67,$6F,$F1,$A6,$EB,$FB,$3E,$5F,$ED,$D3,$99,$FE,$5C,$D7
	.dc.b	$E1,$71,$FE,$85,$F8,$47,$73,$FD,$2B,$DD,$FE,$9D,$E8,$FD,$F3,$63
	.dc.b	$FC,$1D,$D9,$FE,$EB,$9C,$B6,$3F,$C3,$C3,$F4,$9F,$F1,$13,$4C,$B8
	.dc.b	$3F,$C5,$2F,$C4,$F0,$FE,$39,$1F,$CE,$1F,$04,$F3,$13,$2F,$F2,$C7
	.dc.b	$D2,$FF,$92,$C9,$1F,$E4,$F7,$FC,$90,$FF,$23,$3F,$C6,$07,$9C,$A6
	.dc.b	$4F,$ED,$26,$4F,$EE,$FE,$4F,$FA,$33,$E2,$FF,$CB,$04,$8F,$F2,$E3
	.dc.b	$FE,$60,$DB,$95,$FF,$87,$DF,$F7,$66,$79,$25,$7F,$98,$FE,$47,$FC
	.dc.b	$C4,$47,$07,$BC,$FA,$FD,$9F,$09,$FF,$53,$D7,$F9,$F8,$8F,$FA,$B9
	.dc.b	$FF,$23,$8F,$3B,$FD,$11,$1E,$D8,$C7,$9D,$ED,$6C,$66,$77,$87,$FB
	.dc.b	$63,$CF,$FD,$57,$77,$FB,$76,$B0,$7F,$3D,$70,$FF,$A8,$87,$E6,$DF
	.dc.b	$EA,$66,$99,$70,$7F,$AC,$BE,$6E,$47,$FA,$C9,$AF,$F9,$EB,$1F,$97
	.dc.b	$BF,$FA,$F3,$1F,$F4,$41,$27,$E3,$73,$FE,$A1,$F8,$E7,$FA,$8B,$58
	.dc.b	$FF,$B4,$9A,$FF,$B6,$07,$E3,$3F,$ED,$E6,$9F,$ED,$E2,$9B,$7F,$DD
	.dc.b	$6B,$F7,$BA,$47,$FD,$DA,$64,$FF,$7F,$F0,$FF,$BD,$B2,$F5,$FB,$9F
	.dc.b	$84,$9C,$E7,$FE,$FC,$CF,$FC,$FF,$7B,$7E,$7C,$C3,$FD,$07,$B7,$D0
	.dc.b	$47,$FB,$2C,$FC,$0F,$07,$ED,$DF,$EC,$CD,$58,$FE,$17,$9F,$F0,$53
	.dc.b	$99,$F9,$27,$CC,$E9,$FE,$0C,$F0,$87,$BB,$C0,$91,$F8,$97,$F5,$19
	.dc.b	$33,$FF,$3A,$F0,$FB,$1B,$5F,$03,$C1,$F1,$CE,$5B,$1F,$EF,$4F,$C6
	.dc.b	$3F,$DE,$E4,$70,$7C,$03,$C9,$E5,$FC,$23,$E0,$95,$D6,$BF,$A3,$8C
	.dc.b	$47,$C4,$6B,$ED,$F8,$2F,$97,$E0,$7C,$A6,$C7,$2D,$51,$88,$F9,$AD
	.dc.b	$7E,$9F,$C1,$BE,$13,$FA,$93,$63,$FC,$48,$D7,$FC,$54,$D7,$E8,$73
	.dc.b	$FE,$FD,$FE,$3D,$B3,$DD,$FB,$3D,$BF,$EF,$E2,$4F,$E2,$A3,$2F,$FB
	.dc.b	$23,$E1,$7F,$CD,$AF,$F8,$CD,$AF,$F8,$E5,$AE,$F9,$3F,$A9,$1F,$80
	.dc.b	$7F,$06,$7F,$46,$E3,$FB,$9D,$BF,$D3,$5B,$C3,$FD,$47,$24,$E5,$B1
	.dc.b	$EB,$AE,$67,$F3,$A2,$93,$CB,$FF,$3B,$BF,$E4,$D6,$BF,$EE,$26,$BF
	.dc.b	$81,$6B,$FC,$DB,$D2,$FC,$EC,$7E,$3D,$B1,$FA,$E6,$BF,$A8,$C3,$F2
	.dc.b	$D1,$88,$F2,$77,$7F,$0B,$26,$5F,$A2,$C3,$F9,$67,$C3,$FE,$21,$B5
	.dc.b	$FE,$7A,$3F,$0D,$19,$E3,$3F,$8E,$87,$C4,$3D,$DF,$E0,$51,$F9,$B8
	.dc.b	$C5,$FF,$53,$8D,$CE,$9F,$DF,$CE,$BB,$BF,$C5,$B1,$FD,$E8,$F9,$7F
	.dc.b	$14,$F7,$3F,$82,$FC,$BF,$F9,$DE,$31,$28,$C4,$EA,$3F,$58,$27,$73
	.dc.b	$FE,$7F,$F8,$CF,$E0,$33,$C4,$DB,$88,$D7,$81,$CE,$7F,$6D,$39,$7F
	.dc.b	$3A,$7A,$0F,$C1,$3D,$30,$E3,$1F,$06,$30,$F6,$9F,$E1,$BC,$6F,$8D
	.dc.b	$FC,$71,$35,$FC,$93,$FE,$1F,$E1,$FB,$3D,$3F,$91,$7F,$C5,$B3,$FD
	.dc.b	$97,$97,$FC,$97,$C5,$FE,$53,$CB,$FC,$96,$D7,$F9,$5C,$7F,$A6,$F9
	.dc.b	$7F,$8E,$7C,$7F,$F2,$AC,$7F,$8E,$36,$BF,$9D,$B5,$FC,$76,$68,$C4
	.dc.b	$7E,$3B,$97,$F5,$D9,$F9,$A3,$FB,$CF,$C8,$3E,$47,$F9,$EF,$C6,$61
	.dc.b	$B3,$FC,$72,$74,$A7,$4D,$3F,$A6,$FC,$97,$E2,$74,$FE,$68,$DC,$CD
	.dc.b	$3F,$1D,$3F,$B4,$C7,$C8,$FF,$A3,$F8,$FF,$D5,$44,$70,$7E,$23,$CE
	.dc.b	$FF,$10,$47,$FB,$15,$B1,$9F,$73,$AC,$62,$3F,$9D,$7C,$A3,$E5,$3F
	.dc.b	$CF,$0D,$8F,$F6,$53,$5F,$F6,$83,$5F,$CB,$B5,$E7,$1C,$88,$F2,$6B
	.dc.b	$C9,$AF,$26,$BC,$8E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04
	.dc.b	$30,$43,$FB,$EC,$7B,$5F,$F7,$99,$FE,$F1,$F8,$1F,$F8,$02,$FE,$E1
	.dc.b	$AF,$FB,$E8,$F8,$FF,$96,$FF,$81,$BE,$0A,$C7,$FC,$01,$3A,$C6,$5F
	.dc.b	$96,$D7,$EA,$B5,$D9,$C4,$52,$35,$18,$6C,$66,$38,$36,$7C,$64,$8F
	.dc.b	$09,$A9,$E2,$47,$04,$8E,$0F,$98,$C1,$C9,$CB,$E8,$6B,$E5,$36,$71
	.dc.b	$73,$3E,$58,$E4,$43,$FC,$21,$9B,$89,$C8,$F8,$4F,$E5,$34,$AD,$4E
	.dc.b	$63,$FD,$F1,$3A,$9C,$BC,$3D,$C7,$CB,$7F,$8F,$23,$0D,$AF,$E6,$B3
	.dc.b	$A7,$A1,$FE,$23,$71,$3D,$70,$F2,$B7,$0D,$D7,$F9,$23,$0F,$0B,$E4
	.dc.b	$3F,$CA,$FB,$BF,$0C,$FF,$23,$FC,$5F,$F9,$52,$31,$D8,$8C,$4E,$AB
	.dc.b	$8B,$5F,$F2,$AC,$BD,$FF,$E7,$10,$FB,$9F,$F4,$BC,$61,$BD,$FF,$E5
	.dc.b	$C6,$CE,$CF,$EA,$B9,$7C,$CC,$7D,$84,$8C,$EB,$F5,$23,$D1,$F6,$9F
	.dc.b	$E6,$0F,$8C,$B9,$9A,$4B,$FF,$23,$E7,$FE,$1A,$7F,$4D,$3E,$9F,$2F
	.dc.b	$F4,$4D,$8C,$CE,$F9,$3E,$71,$23,$C7,$C1,$FA,$7D,$3F,$07,$3F,$2F
	.dc.b	$F0,$AF,$F3,$9B,$70,$FD,$A7,$6B,$89,$5A,$C7,$EA,$B5,$F9,$B9,$3D
	.dc.b	$6F,$C3,$46,$1A,$FF,$A0,$23,$4F,$E0,$B0,$FA,$93,$A8,$FC,$C7,$07
	.dc.b	$E7,$32,$F5,$47,$E8,$23,$0F,$B2,$3E,$CF,$B5,$FE,$92,$F9,$9B,$26
	.dc.b	$91,$FF,$3C,$7C,$03,$F2,$C7,$D5,$7D,$F9,$CC,$61,$FC,$01,$FE,$9E
	.dc.b	$9E,$13,$B8,$C4,$EF,$9B,$CF,$1F,$9D,$CB,$CE,$FA,$78,$BE,$94,$62
	.dc.b	$18,$78,$F7,$FC,$C4,$63,$0F,$A3,$B7,$E7,$E3,$CB,$3A,$AE,$13,$99
	.dc.b	$FC,$FF,$CD,$7F,$AB,$9F,$DB,$18,$8F,$F5,$BF,$67,$FB,$DA,$74,$D8
	.dc.b	$CC,$FC,$D2,$1E,$9D,$FD,$BC,$7F,$85,$7F,$D7,$7E,$5F,$ED,$5B,$3F
	.dc.b	$BA,$6B,$FE,$5F,$E4,$46,$23,$FD,$87,$8B,$E2,$3F,$3A,$D8,$FD,$0B
	.dc.b	$67,$D1,$3F,$CF,$B9,$BF,$37,$5E,$63,$E8,$70,$7C,$E3,$FA,$37,$B6
	.dc.b	$3F,$A2,$96,$32,$E0,$F0,$FC,$64,$8F,$0B,$70,$FB,$B1,$F9,$6A,$ED
	.dc.b	$1F,$80,$D5,$97,$FA,$82,$3E,$77,$17,$B0,$97,$FA,$ED,$1F,$47,$FE
	.dc.b	$E4,$F0,$3F,$8F,$36,$7F,$91,$B5,$F9,$4D,$7D,$F7,$E4,$D7,$E1,$C7
	.dc.b	$D7,$3F,$95,$FD,$26,$F8,$7F,$78,$DC,$4D,$31,$CF,$FE,$F6,$FD,$57
	.dc.b	$F9,$83,$6F,$F9,$AB,$5F,$B8,$D5,$7F,$C2,$7F,$2A,$FF,$7F,$FE,$E5
	.dc.b	$FC,$E5,$B3,$FC,$F9,$AF,$E1,$5A,$FE,$13,$C4,$2F,$CF,$7F,$31,$18
	.dc.b	$6C,$64,$FF,$77,$F0,$FF,$49,$70,$7E,$45,$71,$7F,$42,$8C,$35,$F4
	.dc.b	$EB,$F6,$A3,$19,$7F,$23,$AE,$3E,$2F,$F5,$3F,$B6,$E7,$F8,$15,$B3
	.dc.b	$49,$8F,$FA,$73,$EB,$BE,$FC,$EA,$31,$C9,$C9,$B1,$97,$E0,$89,$E3
	.dc.b	$5F,$D8,$FE,$E7,$D7,$FA,$15,$B3,$49,$7F,$DB,$3E,$F3,$F9,$C3,$EE
	.dc.b	$6E,$BF,$CF,$DB,$B1,$DD,$FE,$73,$97,$F9,$DB,$6F,$F9,$D6,$FF,$8A
	.dc.b	$6C,$EA,$3F,$95,$73,$FF,$0F,$E5,$FD,$E9,$2B,$F7,$BC,$9F,$7B,$9B
	.dc.b	$F8,$A3,$63,$ED,$B5,$FE,$3A,$D7,$EB,$F0,$FF,$28,$D7,$FA,$C3,$75
	.dc.b	$F9,$FE,$5F,$EB,$35,$FE,$4D,$7F,$A8,$3F,$84,$1F,$BB,$CF,$F8,$ED
	.dc.b	$7E,$92,$77,$87,$EA,$1B,$19,$BF,$F0,$A6,$BD,$FC,$0F,$C6,$FD,$37
	.dc.b	$F8,$5D,$B9,$7F,$87,$1A,$FF,$23,$6B,$FC,$8B,$8B,$F3,$FF,$B1,$FF
	.dc.b	$35,$6C,$7E,$71,$AF,$F3,$E6,$BF,$6F,$2F,$E1,$9C,$5F,$68,$8F,$F0
	.dc.b	$FC,$7F,$43,$F4,$1E,$88,$34,$95,$FC,$42,$1F,$53,$F2,$49,$19,$7F
	.dc.b	$88,$77,$FF,$5B,$8F,$94,$7F,$33,$DF,$E4,$4F,$72,$47,$F1,$4A,$EA
	.dc.b	$FE,$79,$7E,$CF,$FE,$5C,$8F,$EF,$46,$5C,$79,$5F,$A4,$8F,$DD,$C6
	.dc.b	$1D,$78,$5F,$E0,$37,$B7,$9F,$EF,$23,$83,$67,$F7,$E3,$9C,$7F,$7D
	.dc.b	$69,$9B,$F2,$C6,$71,$FC,$27,$2F,$CB,$25,$7B,$5F,$8B,$3F,$D1,$A3
	.dc.b	$FA,$B5,$7B,$BE,$1B,$FC,$77,$89,$A4,$9F,$E8,$DE,$EF,$F3,$72,$57
	.dc.b	$F9,$AB,$1F,$F1,$B4,$7F,$43,$8F,$93,$F3,$7F,$CA,$3E,$8B,$E3,$46
	.dc.b	$2D,$18,$C5,$5B,$9F,$9C,$D7,$FA,$51,$0F,$D4,$17,$07,$E6,$9F,$E9
	.dc.b	$2C,$3F,$C9,$2D,$EA,$FA,$D3,$AD,$3F,$87,$CE,$6F,$F6,$3E,$3F,$F5
	.dc.b	$87,$A9,$30,$FC,$1F,$B1,$F1,$89,$1F,$E6,$4C,$FF,$D4,$CF,$F4,$6C
	.dc.b	$7C,$7F,$05,$FC,$FB,$9B,$C3,$92,$31,$5F,$CB,$2B,$78,$FF,$19,$8E
	.dc.b	$55,$B8,$C4,$7E,$3B,$E4,$7F,$A9,$FC,$5F,$C9,$6C,$7E,$55,$AF,$FA
	.dc.b	$43,$24,$62,$3D,$A7,$E9,$8F,$98,$53,$F3,$7F,$05,$F8,$42,$3E,$F7
	.dc.b	$FD,$C5,$8F,$D0,$F9,$0F,$EB,$7F,$19,$FE,$3F,$AF,$EE,$39,$A7,$39
	.dc.b	$FB,$E3,$0A,$B8,$5F,$38,$C1,$F5,$3A,$BF,$79,$C1,$F5,$71,$F3,$4F
	.dc.b	$ED,$BD,$3F,$F7,$E3,$FB,$56,$C7,$C1,$6B,$FB,$66,$DF,$7F,$AB,$F0
	.dc.b	$E5,$E4,$6E,$7F,$DD,$9B,$81,$F7,$61,$F1,$C9,$8F,$94,$FC,$56,$1F
	.dc.b	$94,$E8,$FE,$FF,$E2,$F6,$6E,$1F,$0D,$B7,$F8,$06,$BF,$2B,$EE,$7F
	.dc.b	$D4,$B3,$F2,$5B,$1F,$86,$6B,$B8,$FF,$2E,$9F,$3F,$A3,$38,$7E,$99
	.dc.b	$BA,$7E,$C2,$77,$5F,$90,$9E,$6F,$B3,$6F,$A2,$4E,$7F,$D6,$E3,$C8
	.dc.b	$97,$CA,$BC,$5D,$1F,$1B,$5F,$89,$87,$18,$FE,$26,$49,$FE,$C3,$F0
	.dc.b	$3F,$96,$F8,$4F,$F1,$7F,$AF,$FE,$76,$9F,$F1,$8C,$E5,$FF,$3B,$57
	.dc.b	$F8,$52,$31,$3B,$94,$62,$75,$CC,$BE,$B1,$86,$D7,$69,$D3,$D8,$C4
	.dc.b	$69,$FF,$38,$1F,$D0,$4A,$7F,$B7,$8C,$62,$C7,$FA,$54,$64,$87,$FA
	.dc.b	$CC,$DB,$A7,$EE,$65,$FE,$50,$3F,$00,$57,$27,$E6,$33,$FF,$57,$47
	.dc.b	$48,$FF,$48,$9D,$34,$F3,$3E,$63,$EF,$47,$37,$F6,$FA,$F2,$72,$7E
	.dc.b	$91,$C3,$F8,$D3,$C1,$1C,$77,$FE,$63,$7A,$6C,$67,$7F,$E9,$8B,$EF
	.dc.b	$BF,$44,$D3,$F4,$45,$BF,$D1,$35,$F5,$9E,$D3,$D4,$60,$FE,$4F,$C9
	.dc.b	$FE,$6C,$EA,$FD,$B9,$27,$F8,$84,$65,$FE,$0D,$3F,$05,$FE,$14,$DE
	.dc.b	$97,$3F,$F8,$0E,$3F,$85,$1F,$1C,$F0,$93,$7F,$BD,$B7,$CE,$6C,$7F
	.dc.b	$40,$C3,$FC,$FF,$87,$F0,$A2,$1E,$43,$31,$FC,$DF,$A7,$EB,$C7,$F8
	.dc.b	$49,$B8,$9A,$7D,$C9,$F1,$70,$7E,$25,$F8,$84,$AF,$F0,$DF,$17,$FA
	.dc.b	$6B,$87,$FA,$44,$FF,$70,$3F,$A4,$14,$FE,$91,$1B,$7F,$9C,$6B,$FE
	.dc.b	$98,$F9,$DF,$EB,$6D,$8C,$DA,$72,$D8,$FF,$14,$76,$FA,$88,$FE,$A6
	.dc.b	$3F,$B6,$BF,$27,$1F,$53,$EC,$DB,$07,$EB,$F8,$BF,$BD,$55,$E2,$FF
	.dc.b	$00,$57,$F6,$EA,$FE,$6A,$FF,$52,$6B,$CC,$FD,$21,$0F,$F0,$78,$F9
	.dc.b	$EF,$F0,$A6,$BF,$EC,$06,$C6,$4D,$BF,$E0,$CF,$31,$E0,$FF,$AF,$39
	.dc.b	$F0,$DA,$FE,$FD,$3A,$CB,$F8,$6C,$7F,$81,$E3,$15,$FE,$23,$EB,$FB
	.dc.b	$E4,$9F,$8D,$C7,$FE,$35,$7E,$19,$E9,$3F,$85,$1F,$C0,$75,$FF,$43
	.dc.b	$7D,$EF,$F9,$B4,$91,$FE,$5F,$7F,$C6,$1A,$FF,$B7,$3E,$EA,$C9,$A4
	.dc.b	$7F,$CC,$DE,$EF,$F9,$DC,$8E,$0F,$F2,$CF,$C8,$72,$23,$83,$FC,$DD
	.dc.b	$5F,$E7,$F2,$38,$3F,$79,$D3,$FA,$86,$F1,$FE,$6F,$2F,$F5,$5E,$FF
	.dc.b	$C9,$5B,$7F,$EA,$B2,$57,$F9,$EF,$E2,$BF,$D1,$84,$70,$7F,$B4,$6B
	.dc.b	$F1,$5C,$1F,$EC,$D6,$F9,$9F,$88,$C3,$FD,$A5,$1F,$95,$F5,$1E,$73
	.dc.b	$07,$E7,$5F,$EA,$56,$9F,$EA,$52,$9F,$EA,$52,$9F,$EF,$CF,$47,$FE
	.dc.b	$00,$FC,$D7,$FA,$D7,$03,$49,$3F,$E9,$5F,$98,$FF,$53,$11,$C1,$F1
	.dc.b	$B6,$FC,$1B,$F8,$1D,$7F,$DF,$B3,$FE,$CF,$24,$7F,$AA,$DF,$F6,$EA
	.dc.b	$E0,$A4,$7F,$DB,$E6,$4F,$F5,$26,$8F,$5F,$F8,$AB,$63,$E3,$46,$0C
	.dc.b	$F9,$BF,$1F,$A7,$FD,$03,$5F,$EF,$56,$C6,$6B,$AB,$5F,$D2,$70,$FD
	.dc.b	$3C,$BE,$B4,$7E,$BE,$30,$67,$81,$B3,$F4,$3F,$DA,$46,$1B,$19,$AF
	.dc.b	$D9,$CE,$5F,$F5,$D7,$B6,$F4,$79,$FD,$97,$F7,$D8,$C5,$A3,$13,$AA
	.dc.b	$E2,$D7,$F7,$9E,$ED,$5C,$CD,$13,$98,$C3,$63,$38,$77,$AF,$E1,$3B
	.dc.b	$FB,$BC,$BF,$E2,$0D,$8C,$CA,$72,$D8,$FB,$3F,$01,$F9,$9F,$C4,$3E
	.dc.b	$9B,$63,$22,$32,$D3,$F8,$CF,$A3,$F3,$BF,$72,$FE,$40,$D8,$CD,$A7
	.dc.b	$2D,$8F,$E4,$3C,$BF,$5B,$9B,$F9,$43,$63,$33,$F8,$F7,$F9,$24,$65
	.dc.b	$FE,$4D,$CB,$FC,$AD,$B3,$FC,$A9,$F9,$BA,$A7,$51,$86,$BF,$D0,$4F
	.dc.b	$D4,$1F,$CE,$78,$37,$C1,$FC,$3B,$9B,$18,$75,$8C,$57,$E8,$BB,$7F
	.dc.b	$42,$47,$F8,$33,$FD,$CF,$3F,$E3,$51,$FC,$D0,$87,$F7,$3E,$17,$C2
	.dc.b	$FE,$A8,$91,$FC,$3D,$FE,$F9,$F0,$C2,$92,$3F,$C7,$6E,$5F,$E0,$27
	.dc.b	$63,$DC,$7F,$92,$65,$EA,$7E,$E0,$91,$FC,$75,$FF,$22,$E3,$F9,$5F
	.dc.b	$4F,$F5,$7F,$4F,$FD,$75,$B5,$FD,$69,$AF,$F5,$96,$BF,$E1,$03,$C2
	.dc.b	$7B,$F4,$FE,$51,$F4,$5F,$C0,$9B,$1F,$C3,$1A,$F9,$DA,$BC,$57,$C5
	.dc.b	$3E,$2B,$F7,$FE,$0F,$F3,$3A,$FE,$69,$5A,$9C,$C6,$63,$16,$AE,$13
	.dc.b	$D6,$FA,$4E,$63,$F0,$A4,$3A,$C7,$F1,$FF,$23,$8D,$F9,$39,$FF,$B5
	.dc.b	$57,$7A,$D3,$63,$2F,$F9,$EF,$E3,$9F,$1B,$FD,$2F,$CA,$F1,$E4,$9C
	.dc.b	$D2,$73,$18,$6D,$71,$9F,$31,$F4,$79,$7F,$22,$FF,$24,$A7,$C7,$7F
	.dc.b	$2D,$8C,$37,$0F,$03,$67,$73,$FA,$1E,$BF,$8A,$FF,$87,$F7,$FE,$68
	.dc.b	$4B,$FC,$8C,$78,$7A,$3F,$C7,$38,$F8,$A7,$D9,$F1,$FF,$E5,$CC,$4D
	.dc.b	$25,$FF,$74,$F9,$0F,$E7,$7C,$1F,$E2,$D6,$C7,$F8,$A9,$AF,$F8,$A5
	.dc.b	$AF,$F7,$5C,$3F,$CE,$7C,$5F,$EA,$12,$47,$E0,$23,$2F,$E2,$71,$FA
	.dc.b	$1F,$19,$F5,$A3,$E4,$36,$33,$1F,$E2,$53,$86,$49,$E1,$CF,$FC,$F7
	.dc.b	$C2,$FC,$8D,$7F,$8F,$5A,$FF,$16,$6C,$6E,$FF,$21,$3F,$9B,$C3,$FD
	.dc.b	$17,$DB,$FE,$B9,$24,$7A,$76,$79,$5F,$C8,$61,$F2,$2F,$37,$EF,$46
	.dc.b	$1B,$19,$8E,$0D,$BE,$B1,$E1,$BE,$B1,$88,$F9,$91,$87,$F5,$3F,$24
	.dc.b	$F6,$3F,$D2,$4D,$8F,$F6,$7B,$5F,$F7,$33,$5E,$31,$C0,$8F,$06,$BC
	.dc.b	$1A,$F0,$34,$43,$45,$34,$60,$86,$08,$7F,$09,$8F,$23,$63,$BB,$4E
	.dc.b	$85,$39,$94,$E6,$7F,$91,$63,$9B,$63,$FD,$6D,$F0,$0F,$73,$EC,$4A
	.dc.b	$31,$87,$FB,$D9,$B1,$FC,$58,$D4,$F0,$FC,$4B,$FD,$59,$E0,$65,$B3
	.dc.b	$FC,$71,$AF,$F2,$56,$BF,$E5,$96,$A6,$93,$DC,$FF,$2F,$4F,$56,$CF
	.dc.b	$56,$BD,$5A,$F5,$6B,$D5,$B3,$D6,$13,$D5,$A7,$52,$9D,$4A,$75,$29
	.dc.b	$D4,$B1,$D6,$3A,$12,$3F,$9F,$FD,$1F,$F9,$67,$85,$F2,$7F,$36,$30
	.dc.b	$DF,$03,$FC,$96,$D9,$DD,$7E,$C7,$8A,$BF,$A7,$6B,$FA,$66,$BF,$E6
	.dc.b	$DE,$0F,$DB,$D7,$F5,$0E,$47,$BC,$F9,$5F,$F2,$CF,$DA,$66,$31,$92
	.dc.b	$31,$3A,$C7,$FB,$05,$F2,$8C,$47,$FA,$56,$9E,$E3,$F1,$9F,$F4,$C7
	.dc.b	$BF,$FD,$3A,$31,$5F,$ED,$83,$F9,$13,$88,$D4,$F0,$9F,$81,$C1,$F9
	.dc.b	$3E,$A6,$DF,$C4,$A3,$0D,$9D,$9F,$CB,$7F,$23,$FD,$06,$5F,$18,$38
	.dc.b	$DF,$FB,$0B,$67,$87,$2F,$D2,$7C,$BF,$8D,$FE,$3A,$FD,$AD,$D4,$D3
	.dc.b	$F2,$39,$F9,$B8,$BF,$0E,$DD,$3C,$F1,$8D,$3D,$18,$FF,$AF,$78,$3E
	.dc.b	$AF,$4F,$FA,$32,$FE,$A3,$67,$E8,$F1,$FF,$21,$D3,$F2,$B7,$C1,$FC
	.dc.b	$AB,$FE,$0C,$BF,$DD,$36,$BF,$05,$E2,$73,$7E,$DB,$5F,$C5,$4F,$98
	.dc.b	$8F,$CC,$3F,$20,$E1,$A6,$4F,$F4,$7D,$BE,$9B,$A5,$FE,$41,$B1,$9C
	.dc.b	$3D,$2D,$BE,$97,$E2,$3A,$1F,$9B,$CF,$F6,$3F,$A6,$7F,$61,$8B,$E0
	.dc.b	$FF,$00,$8C,$57,$F1,$96,$CF,$DC,$C1,$F2,$1F,$80,$E2,$FF,$5C,$B6
	.dc.b	$33,$3B,$8F,$1B,$E1,$BB,$C1,$F6,$E3,$A7,$27,$12,$38,$3F,$C8,$5F
	.dc.b	$38,$FB,$3F,$F7,$D7,$CB,$FF,$8E,$A3,$12,$8C,$4E,$A3,$8B,$5F,$9E
	.dc.b	$43,$E3,$17,$C9,$7E,$49,$FB,$9D,$3A,$4F,$F9,$4E,$31,$7F,$E4,$A6
	.dc.b	$C6,$7B,$7F,$DB,$73,$FE,$44,$FD,$FB,$F9,$B4,$FF,$22,$B4,$63,$5D
	.dc.b	$78,$3F,$A5,$8C,$57,$D8,$DB,$E3,$3E,$73,$E5,$1F,$E2,$18,$DB,$FC
	.dc.b	$25,$AF,$FB,$B1,$B7,$E0,$AE,$DF,$64,$A7,$85,$E7,$8C,$36,$32,$77
	.dc.b	$8C,$BE,$5E,$BF,$D8,$BF,$28,$97,$EE,$9E,$DF,$0B,$CC,$93,$CC,$3F
	.dc.b	$CA,$79,$3F,$3A,$4A,$FC,$EB,$FA,$D9,$F8,$27,$C3,$CF,$E2,$BE,$11
	.dc.b	$F5,$6F,$8B,$FE,$B6,$D8,$FE,$A2,$DD,$BF,$BB,$36,$3F,$82,$FC,$7B
	.dc.b	$17,$DB,$7F,$3D,$8C,$36,$33,$3C,$1A,$FF,$13,$EC,$7C,$AF,$3F,$81
	.dc.b	$FE,$8C,$D9,$FE,$26,$D7,$83,$5F,$B5,$F3,$3F,$5E,$7F,$18,$D8,$FC
	.dc.b	$13,$5F,$C1,$35,$FE,$17,$F2,$9E,$BF,$D5,$BF,$AB,$36,$3F,$91,$B5
	.dc.b	$F4,$35,$FE,$27,$C9,$F0,$FA,$7F,$AF,$36,$3F,$95,$B5,$F8,$0D,$7F
	.dc.b	$1D,$F8,$1B,$EB,$BF,$21,$F8,$47,$F8,$96,$31,$28,$C4,$EA,$3F,$D7
	.dc.b	$F3,$A9,$FD,$67,$D4,$FE,$67,$F1,$7B,$46,$91,$FD,$37,$8B,$FA,$D2
	.dc.b	$3F,$B3,$ED,$FE,$3A,$FD,$7E,$FF,$EB,$0D,$E3,$FF,$02,$F9,$1F,$E0
	.dc.b	$09,$CC,$FF,$36,$6C,$7F,$81,$3E,$F7,$F1,$1F,$91,$7F,$80,$DB,$5F
	.dc.b	$D4,$1A,$FE,$15,$AF,$F4,$0F,$83,$FC,$87,$ED,$5F,$E0,$B6,$C7,$F5
	.dc.b	$86,$BF,$8D,$6B,$FD,$23,$BF,$F5,$1D,$BF,$AB,$CF,$F5,$27,$F6,$7F
	.dc.b	$2E,$47,$F3,$2F,$0F,$F7,$3A,$7F,$87,$3B,$7D,$7D,$BF,$A0,$C3,$E0
	.dc.b	$B7,$C9,$FF,$0F,$4E,$BE,$E3,$D3,$7F,$E2,$46,$D1,$A6,$3A,$FF,$59
	.dc.b	$F1,$5F,$92,$C4,$D2,$6D,$F6,$0F,$AD,$CB,$F6,$35,$FE,$58,$FF,$27
	.dc.b	$F8,$5F,$E3,$5A,$FF,$2F,$6D,$7F,$25,$6B,$FC,$95,$AF,$C7,$C3,$E2
	.dc.b	$DF,$C7,$24,$7F,$1D,$DF,$E3,$BF,$DD,$75,$F8,$D5,$FD,$12,$3F,$7D
	.dc.b	$B7,$F4,$07,$FC,$1A,$D7,$31,$B7,$F9,$A6,$5E,$B7,$DD,$87,$B1,$FE
	.dc.b	$F7,$5F,$C4,$7E,$93,$E7,$D9,$A4,$9E,$24,$3F,$A1,$5F,$57,$C3,$F4
	.dc.b	$9C,$88,$F9,$A9,$1F,$83,$CB,$DF,$3F,$D9,$65,$FB,$53,$E3,$7F,$C7
	.dc.b	$4D,$9F,$EB,$8D,$CF,$FA,$64,$7F,$96,$E3,$0F,$F8,$C3,$D9,$FE,$59
	.dc.b	$F2,$1F,$E6,$76,$C6,$70,$4E,$5B,$1F,$E3,$8F,$63,$F2,$A4,$8F,$CA
	.dc.b	$BF,$B4,$8F,$E8,$A7,$F4,$3F,$33,$FA,$2B,$8E,$07,$E4,$1F,$E6,$FC
	.dc.b	$FF,$CF,$8D,$9F,$F3,$AB,$5F,$F2,$63,$5E,$9D,$1F,$D8,$1E,$E7,$83
	.dc.b	$1F,$8F,$F7,$DF,$18,$91,$F1,$9F,$F5,$7E,$1E,$2F,$9F,$FE,$7C,$DA
	.dc.b	$79,$7E,$49,$CE,$27,$FA,$24,$BC,$4F,$F2,$36,$3F,$E0,$D6,$E7,$F4
	.dc.b	$A3,$FA,$D4,$61,$FA,$B3,$C7,$D9,$FF,$37,$E9,$EB,$D7,$E3,$91,$FE
	.dc.b	$0B,$93,$E0,$7F,$CA,$6D,$BF,$F3,$4B,$72,$F9,$73,$AF,$5F,$FC,$E5
	.dc.b	$5F,$31,$B9,$7C,$C6,$AD,$BE,$AC,$62,$3E,$6E,$5E,$3E,$5F,$CA,$C7
	.dc.b	$F7,$5E,$4F,$F6,$8E,$5F,$DD,$E3,$F2,$0F,$FB,$8B,$1F,$7C,$E9,$B1
	.dc.b	$99,$F1,$47,$FB,$70,$F3,$15,$CC,$75,$7C,$D9,$FF,$94,$E7,$51,$F8
	.dc.b	$4B,$F2,$BF,$8F,$97,$D2,$7B,$B3,$FC,$8B,$63,$37,$CA,$72,$D8,$7C
	.dc.b	$9E,$C9,$19,$79,$7A,$F2,$38,$37,$2F,$46,$3F,$2B,$7F,$BC,$4C,$FE
	.dc.b	$CE,$1D,$2B,$DD,$97,$E9,$A3,$18,$FF,$23,$9F,$E0,$C4,$C3,$07,$E3
	.dc.b	$FE,$11,$E2,$3E,$17,$FC,$9D,$D6,$BF,$E0,$B5,$A9,$CC,$62,$B8,$5F
	.dc.b	$F0,$F8,$C4,$EA,$3F,$CB,$15,$C5,$D1,$8A,$E3,$7D,$E3,$15,$C5,$57
	.dc.b	$19,$D4,$7F,$A9,$4F,$F6,$69,$0F,$D5,$70,$FF,$B4,$DC,$1F,$C7,$9C
	.dc.b	$3A,$D5,$AF,$EC,$63,$FC,$DD,$3A,$7F,$85,$E9,$F9,$7D,$3F,$87,$B6
	.dc.b	$33,$1F,$E7,$53,$25,$7C,$17,$D9,$F1,$9D,$9C,$3E,$3B,$FF,$6C,$EE
	.dc.b	$DF,$B4,$FE,$69,$BB,$FF,$24,$6C,$EF,$0F,$05,$7E,$83,$1C,$92,$33
	.dc.b	$7F,$87,$6D,$7E,$72,$BF,$16,$D8,$FE,$5B,$C1,$D8,$F1,$CF,$D1,$3D
	.dc.b	$07,$B0,$C1,$FB,$63,$9F,$17,$98,$93,$E6,$3F,$AB,$F0,$FF,$1E,$9F
	.dc.b	$E0,$66,$78,$BE,$FC,$FF,$02,$8F,$69,$D8,$F9,$5A,$7D,$97,$06,$72
	.dc.b	$FE,$2B,$3E,$EC,$3C,$33,$E4,$8F,$83,$C7,$FB,$08,$FE,$0E,$DA,$FE
	.dc.b	$A2,$DC,$FF,$95,$4F,$BF,$7F,$C3,$3E,$23,$EE,$DF,$F3,$77,$F1,$8D
	.dc.b	$7F,$9F,$35,$FE,$7B,$CB,$FD,$11,$FE,$95,$1F,$CE,$5F,$EB,$BC,$DF
	.dc.b	$E9,$DC,$32,$48,$FE,$4F,$D1,$F0,$6B,$FA,$DC,$78,$35,$4E,$5B,$1F
	.dc.b	$1F,$1F,$4B,$F2,$1A,$7B,$33,$FE,$E3,$1F,$04,$97,$F9,$67,$E4,$1C
	.dc.b	$78,$BF,$36,$FE,$83,$6E,$D1,$88,$FE,$91,$E2,$BA,$BF,$D7,$1C,$FF
	.dc.b	$D8,$71,$4E,$71,$FE,$1B,$7F,$EB,$5B,$DB,$FE,$0D,$8C,$BD,$3B,$3F
	.dc.b	$D8,$5B,$FF,$84,$C9,$1F,$E1,$77,$E8,$EB,$FE,$23,$8B,$CF,$F1,$48
	.dc.b	$FA,$35,$8F,$DF,$D7,$D5,$A7,$F9,$33,$DB,$FF,$1F,$92,$7F,$C6,$CF
	.dc.b	$F9,$15,$87,$F9,$94,$A7,$F9,$3B,$E8,$3F,$C8,$E4,$7F,$C7,$CE,$9F
	.dc.b	$F2,$89,$F8,$AF,$35,$E0,$FD,$F6,$C7,$F1,$9B,$FF,$2F,$FC,$67,$F9
	.dc.b	$48,$91,$FE,$49,$7D,$59,$7B,$23,$F7,$27,$F9,$0A,$7E,$8F,$77,$E8
	.dc.b	$78,$BF,$A2,$F0,$F5,$6D,$7F,$A5,$CF,$F3,$D1,$F4,$A1,$8B,$C5,$4E
	.dc.b	$63,$0D,$8C,$F2,$7B,$EB,$8E,$9E,$D2,$57,$87,$C8,$FF,$7F,$77,$7F
	.dc.b	$A4,$C9,$1F,$EE,$4A,$FE,$E2,$FA,$F2,$FF,$4E,$43,$F2,$0F,$C7,$46
	.dc.b	$1B,$E5,$27,$55,$C6,$31,$3F,$E7,$CC,$A4,$FF,$9C,$DA,$FE,$7C,$F9
	.dc.b	$A7,$F3,$ED,$BE,$5D,$FC,$9D,$BE,$61,$1F,$99,$E5,$F8,$7F,$29,$FF
	.dc.b	$65,$41,$A4,$BF,$F6,$A4,$79,$9F,$F6,$A3,$5F,$D6,$F9,$DE,$87,$79
	.dc.b	$7C,$0F,$43,$F5,$31,$FE,$E7,$CF,$17,$FA,$AC,$FF,$DD,$0F,$D7,$E0
	.dc.b	$FA,$35,$F5,$78,$BF,$1E,$FF,$BE,$BF,$3B,$8F,$F9,$5A,$3F,$CB,$CA
	.dc.b	$32,$FF,$A3,$BD,$47,$AB,$F8,$4F,$CE,$FF,$B8,$46,$25,$18,$9D,$47
	.dc.b	$F9,$96,$75,$3F,$6B,$D7,$FB,$1F,$AA,$5C,$CD,$25,$FE,$17,$F5,$8F
	.dc.b	$F6,$D1,$1C,$1E,$9F,$73,$F2,$FE,$3F,$F3,$6D,$51,$89,$D1,$FC,$D4
	.dc.b	$FF,$00,$46,$9C,$1F,$C7,$BD,$DF,$E7,$0F,$E4,$DA,$FE,$93,$83,$F8
	.dc.b	$FC,$62,$38,$65,$CA,$7F,$91,$FB,$DF,$95,$F2,$7F,$32,$DA,$FC,$F3
	.dc.b	$5E,$9D,$68,$C4,$7E,$CB,$DF,$7C,$07,$F6,$78,$46,$1F,$ED,$4D,$7F
	.dc.b	$C4,$8D,$7F,$83,$9F,$9F,$29,$FD,$84,$FD,$87,$67,$E7,$5F,$E0,$C6
	.dc.b	$32,$FE,$35,$F0,$5F,$C1,$3F,$1A,$F9,$CD,$9F,$C5,$35,$F9,$AD,$7B
	.dc.b	$1E,$12,$18,$3F,$75,$B2,$3F,$CB,$71,$D0,$FE,$80,$69,$CB,$D8,$CD
	.dc.b	$7F,$1A,$F4,$FF,$97,$12,$3F,$99,$3F,$27,$CE,$FE,$CF,$E4,$FF,$5D
	.dc.b	$9F,$EC,$5E,$36,$98,$7E,$2E,$B9,$3F,$EB,$FF,$25,$F0,$2F,$03,$F8
	.dc.b	$97,$93,$98,$C6,$68,$C4,$EA,$FF,$DF,$37,$CA,$31,$1F,$10,$F1,$9F
	.dc.b	$E5,$08,$7C,$2B,$D3,$FE,$6B,$7E,$38,$C5,$77,$3A,$B8,$8D,$4F,$09
	.dc.b	$FF,$6A,$FC,$5F,$F4,$6F,$13,$FC,$3A,$D8,$CF,$1F,$F1,$07,$F9,$74
	.dc.b	$65,$F3,$70,$7C,$AD,$FF,$AA,$47,$F9,$AE,$7F,$A8,$FC,$C7,$4B,$FE
	.dc.b	$6C,$DD,$8D,$3F,$CF,$5F,$13,$FC,$5A,$FF,$DD,$8D,$7F,$A9,$87,$0F
	.dc.b	$F5,$66,$BF,$20,$FC,$E9,$FE,$4B,$E9,$7B,$5F,$D9,$49,$7F,$D9,$5F
	.dc.b	$F1,$1E,$7E,$BA,$FE,$86,$72,$70,$F4,$EC,$7C,$57,$F8,$A9,$B7,$FE
	.dc.b	$5B,$6C,$FF,$7D,$9F,$EE,$9F,$17,$2B,$FB,$A3,$67,$B9,$F3,$39,$FF
	.dc.b	$C6,$12,$F8,$CF,$F1,$84,$FF,$95,$1A,$FF,$95,$1A,$FC,$68,$E1,$EC
	.dc.b	$7F,$93,$78,$3E,$C4,$FD,$AE,$0F,$B9,$49,$CB,$63,$F0,$13,$F2,$CF
	.dc.b	$27,$0F,$87,$D8,$FF,$02,$FC,$BF,$85,$B3,$E1,$C9,$18,$8F,$F5,$74
	.dc.b	$61,$4F,$F7,$CF,$8F,$FE,$F5,$F3,$97,$23,$49,$5F,$EA,$FE,$6F,$F6
	.dc.b	$07,$13,$7F,$F6,$1C,$61,$B3,$B3,$F1,$23,$CA,$7C,$8F,$B9,$B8,$FF
	.dc.b	$95,$27,$53,$93,$E1,$F7,$FF,$46,$8F,$C9,$B5,$FF,$13,$B5,$FF,$61
	.dc.b	$79,$3F,$CD,$E3,$04,$70,$48,$FF,$51,$46,$5B,$19,$7F,$DB,$D1,$96
	.dc.b	$C6,$5F,$6C,$65,$B1,$97,$CE,$D7,$CE,$FF,$4B,$F8,$1F,$F7,$F4,$7C
	.dc.b	$66,$BF,$B0,$6B,$FE,$FE,$3C,$7D,$9F,$E4,$6F,$80,$B1,$FE,$FE,$3E
	.dc.b	$1C,$65,$F2,$B5,$F9,$2D,$76,$71,$3F,$DD,$4E,$5A,$66,$FF,$8F,$CF
	.dc.b	$80,$91,$D8,$D4,$F1,$9D,$C6,$67,$6F,$88,$D9,$1C,$BE,$46,$BE,$26
	.dc.b	$EB,$F9,$1E,$37,$C3,$FE,$45,$B5,$F9,$09,$D4,$E4,$FC,$9C,$7E,$4F
	.dc.b	$E8,$7F,$CD,$F1,$8D,$91,$89,$FC,$CC,$61,$EA,$7F,$24,$71,$3F,$EB
	.dc.b	$3F,$82,$FE,$D5,$BB,$EB,$F1,$E3,$FF,$99,$DB,$19,$C3,$FD,$15,$5C
	.dc.b	$27,$FD,$5F,$F4,$BF,$E8,$2F,$94,$B4,$69,$1F,$77,$23,$6F,$B5,$B7
	.dc.b	$C5,$DF,$E2,$12,$BB,$F9,$7F,$E7,$9F,$93,$F4,$36,$33,$AA,$72,$D8
	.dc.b	$FE,$3D,$F0,$8F,$97,$FC,$54,$A3,$13,$F9,$38,$C4,$EA,$73,$87,$8E
	.dc.b	$BF,$DC,$0D,$7F,$D4,$38,$FA,$5F,$8B,$F3,$DF,$E8,$F9,$F9,$CD,$8C
	.dc.b	$E1,$FA,$1F,$15,$C2,$AD,$57,$F9,$26,$3F,$90,$D3,$EB,$BE,$C7,$7F
	.dc.b	$FA,$6D,$B1,$9A,$FA,$B3,$97,$E7,$EF,$F3,$B8,$3E,$86,$7E,$99,$69
	.dc.b	$E9,$7F,$B4,$FE,$02,$9F,$F5,$EE,$5F,$92,$8C,$4F,$73,$73,$A7,$F9
	.dc.b	$4C,$6A,$7F,$1D,$EB,$FB,$25,$B7,$C0,$8C,$37,$4F,$89,$AF,$F9,$D1
	.dc.b	$F4,$69,$32,$F4,$67,$E7,$C7,$E1,$7D,$77,$DC,$CC,$D2,$47,$4E,$3F
	.dc.b	$F6,$17,$DB,$FE,$18,$91,$F8,$A7,$AE,$FF,$1A,$3E,$4E,$3F,$7F,$5F
	.dc.b	$AF,$7F,$91,$9F,$AC,$D7,$DB,$AA,$31,$3C,$35,$FC,$A5,$7D,$DF,$B0
	.dc.b	$FF,$6A,$7C,$97,$EC,$DB,$1F,$AE,$6B,$B9,$4F,$9B,$7F,$89,$3E,$63
	.dc.b	$FB,$E6,$CF,$17,$FD,$EB,$CF,$FD,$73,$0F,$C5,$46,$2B,$6D,$7F,$DF
	.dc.b	$71,$A7,$CE,$7E,$3F,$CB,$FF,$7F,$FC,$DF,$3C,$11,$89,$F0,$B5,$E7
	.dc.b	$1A,$53,$E9,$DE,$47,$03,$F8,$B9,$DC,$B7,$FA,$1B,$FE,$3D,$1D,$1A
	.dc.b	$FD,$0F,$4F,$FD,$FD,$E0,$7F,$1D,$6C,$7F,$18,$6B,$FA,$D6,$BF,$BA
	.dc.b	$E2,$FC,$03,$FB,$E3,$E6,$14,$F5,$F9,$65,$F3,$7F,$0E,$30,$D8,$C9
	.dc.b	$F5,$23,$2F,$F2,$AF,$90,$FD,$0F,$29,$A4,$CF,$EE,$CF,$84,$F3,$70
	.dc.b	$E1,$FC,$7F,$C2,$A7,$E4,$F2,$FF,$0C,$8C,$78,$9F,$CF,$DB,$5F,$CE
	.dc.b	$9A,$FF,$0F,$E2,$E5,$FB,$3F,$F4,$56,$C7,$F4,$06,$BB,$C1,$3F,$D2
	.dc.b	$FE,$8B,$C5,$F2,$D6,$4D,$25,$7F,$4A,$F8,$7F,$47,$E5,$9F,$D4,$DB
	.dc.b	$19,$A4,$E5,$B1,$FD,$4B,$E9,$3E,$07,$E5,$7F,$D6,$DB,$19,$94,$E5
	.dc.b	$B1,$FA,$6F,$8D,$FB,$5F,$B2,$FE,$6D,$B1,$91,$19,$69,$F8,$43,$E9
	.dc.b	$F5,$7E,$69,$FE,$71,$F2,$9F,$4F,$B3,$FB,$73,$63,$FB,$4B,$5D,$E1
	.dc.b	$FA,$47,$F7,$3F,$37,$FC,$6C,$8F,$F5,$5E,$14,$7F,$1F,$F9,$3F,$D0
	.dc.b	$D7,$F4,$56,$C9,$A7,$E9,$B9,$7F,$A7,$FE,$03,$FA,$06,$D7,$F9,$7F
	.dc.b	$0F,$EF,$B3,$9E,$93,$FA,$69,$FE,$A5,$EE,$FE,$87,$8B,$F8,$13,$6F
	.dc.b	$F8,$8B,$5F,$E0,$2D,$7F,$92,$F5,$7E,$77,$7F,$F0,$93,$63,$E6,$1F
	.dc.b	$D5,$3B,$7F,$66,$FC,$E3,$E8,$67,$F9,$DC,$7F,$8A,$B7,$C2,$FC,$F4
	.dc.b	$63,$47,$57,$F3,$58,$7D,$B3,$E6,$BF,$75,$FB,$5F,$F2,$58,$C6,$48
	.dc.b	$C4,$EA,$3F,$D5,$13,$A9,$FC,$66,$D2,$7F,$20,$D7,$F9,$0F,$1F,$DA
	.dc.b	$DF,$F9,$43,$63,$F6,$58,$FF,$3C,$6F,$8D,$EA,$23,$FC,$1E,$3F,$B2
	.dc.b	$F0,$7C,$78,$FF,$03,$3F,$DB,$5F,$91,$E0,$7F,$05,$6E,$1F,$D6,$76
	.dc.b	$F8,$76,$F3,$3E,$39,$2B,$C6,$F9,$9F,$C8,$65,$E7,$7D,$67,$F7,$6D
	.dc.b	$FF,$7E,$FA,$BB,$BE,$89,$1F,$CD,$9C,$89,$8F,$53,$EB,$FC,$F4,$8F
	.dc.b	$86,$D7,$F9,$01,$EA,$3F,$0B,$E6,$FF,$80,$72,$FB,$7E,$B7,$F8,$F3
	.dc.b	$88,$D2,$57,$F8,$F6,$3F,$BA,$2F,$D7,$3F,$01,$E6,$FF,$3B,$6C,$64
	.dc.b	$FB,$F1,$97,$FA,$01,$FE,$1A,$FD,$87,$F9,$86,$0F,$8C,$FF,$64,$F4
	.dc.b	$A7,$31,$86,$D7,$29,$D5,$7F,$39,$AF,$E1,$53,$F9,$F3,$FA,$04,$F3
	.dc.b	$FA,$DF,$F1,$3D,$9A,$4B,$FC,$06,$9F,$E2,$5B,$F9,$07,$F4,$98,$FD
	.dc.b	$0F,$17,$C7,$24,$7C,$57,$E5,$7A,$1F,$D7,$7F,$1F,$FD,$23,$67,$F4
	.dc.b	$2D,$7F,$C0,$FB,$23,$11,$FD,$4E,$1E,$77,$F9,$16,$36,$D7,$4D,$C3
	.dc.b	$F6,$AE,$63,$D2,$7C,$32,$18,$D3,$E7,$7D,$07,$C4,$25,$7F,$9B,$75
	.dc.b	$47,$C8,$F8,$4F,$F3,$A4,$FD,$19,$46,$2F,$FB,$E5,$FD,$98,$C4,$7D
	.dc.b	$F6,$BE,$42,$6B,$F3,$CF,$CA,$65,$FE,$73,$E4,$FE,$E9,$3F,$1B,$C4
	.dc.b	$FF,$71,$65,$FC,$29,$B1,$93,$FC,$E1,$1E,$5C,$3C,$38,$F8,$E3,$33
	.dc.b	$E6,$CB,$D3,$18,$8F,$33,$5F,$2F,$83,$FC,$A9,$BD,$7F,$16,$3A,$D1
	.dc.b	$A4,$AE,$97,$F9,$BF,$C3,$76,$48,$CD,$23,$9F,$CA,$C9,$CB,$74,$F8
	.dc.b	$79,$7D,$AD,$BE,$46,$5E,$57,$FC,$F9,$97,$7C,$3A,$E3,$DF,$0E,$33
	.dc.b	$97,$9E,$1E,$C7,$F5,$2D,$3F,$83,$E1,$F0,$7E,$0D,$E9,$3E,$1B,$FC
	.dc.b	$75,$C5,$FE,$9C,$B5,$6A,$73,$87,$F0,$08,$C4,$FE,$32,$31,$7C,$E3
	.dc.b	$07,$F9,$16,$77,$7D,$5B,$5C,$2F,$AD,$6E,$F9,$5F,$8F,$4F,$A9,$BF
	.dc.b	$F1,$0A,$FE,$23,$18,$70,$7F,$40,$2B,$97,$FD,$82,$63,$2F,$F7,$83
	.dc.b	$67,$8B,$C4,$FD,$26,$9F,$98,$D6,$9F,$E8,$48,$FC,$84,$65,$B1,$97
	.dc.b	$FC,$AA,$75,$21,$E2,$3F,$61,$B7,$17,$F9,$53,$76,$FE,$50,$DD,$CD
	.dc.b	$B9,$7D,$EB,$F0,$D7,$F8,$4B,$FC,$94,$E5,$FC,$E3,$71,$E8,$47,$D1
	.dc.b	$3F,$04,$F9,$B1,$F1,$9F,$41,$8C,$7E,$7D,$FE,$CE,$3C,$1C,$1E,$72
	.dc.b	$5F,$C3,$3F,$C1,$FC,$3F,$D2,$1F,$E0,$13,$9E,$2F,$F7,$54,$62,$3F
	.dc.b	$37,$AF,$D9,$7C,$07,$CA,$29,$FC,$16,$3D,$D7,$E0,$8F,$1C,$7C,$1D
	.dc.b	$BF,$4B,$1F,$D5,$DB,$8F,$F2,$89,$F5,$E9,$FC,$99,$FC,$1E,$1E,$1B
	.dc.b	$FB,$67,$C2,$29,$FC,$E8,$7D,$D7,$F8,$A7,$93,$FD,$48,$D8,$C9,$FE
	.dc.b	$28,$8C,$BF,$D2,$87,$E3,$FF,$D3,$9A,$65,$C1,$FE,$35,$E8,$FC,$BE
	.dc.b	$1F,$D6,$A7,$F4,$2D,$74,$7A,$B8,$3E,$D3,$F2,$08,$7A,$8F,$EF,$5D
	.dc.b	$3F,$13,$3F,$ED,$95,$F1,$23,$07,$CA,$9F,$C3,$1C,$4A,$63,$4F,$EF
	.dc.b	$72,$F9,$AF,$F0,$03,$FE,$B5,$6B,$99,$FE,$A4,$E5,$FF,$07,$8F,$C5
	.dc.b	$BF,$C2,$64,$72,$FF,$95,$FB,$FE,$1F,$17,$F8,$82,$7E,$1B,$5D,$1F
	.dc.b	$0F,$92,$3F,$8B,$3E,$16,$3F,$E6,$BF,$27,$FC,$D2,$48,$FF,$2A,$BF
	.dc.b	$E4,$36,$1F,$E3,$C2,$9F,$DA,$3C,$5F,$F2,$6F,$CA,$7F,$9C,$89,$1F
	.dc.b	$E5,$E7,$6F,$FB,$5F,$9F,$FD,$36,$FD,$A6,$BA,$BF,$F3,$F6,$1F,$E7
	.dc.b	$C2,$47,$F9,$A1,$F5,$3F,$4B,$97,$F3,$0F,$D3,$C3,$FD,$D9,$AF,$EE
	.dc.b	$1B,$1F,$E9,$08,$C7,$8F,$FA,$49,$FF,$32,$FD,$27,$FA,$28,$95,$FE
	.dc.b	$76,$7D,$34,$CB,$CF,$3F,$0C,$87,$C3,$F5,$7F,$DE,$9F,$21,$FE,$D1
	.dc.b	$24,$7F,$B5,$5F,$F4,$6A,$FD,$72,$91,$FF,$47,$EE,$74,$7F,$A2,$3E
	.dc.b	$5B,$FD,$3E,$47,$07,$ED,$CF,$89,$CF,$FA,$27,$EF,$E9,$F0,$9F,$F4
	.dc.b	$61,$F0,$C8,$7F,$AE,$87,$EC,$DF,$EB,$E6,$99,$70,$7F,$98,$23,$FD
	.dc.b	$8F,$A7,$F8,$C2,$7F,$D9,$6D,$7F,$0D,$E9,$A3,$10,$E2,$FF,$69,$F7
	.dc.b	$7F,$5F,$9F,$F6,$E4,$FF,$6D,$3E,$13,$8F,$0F,$5C,$1F,$ED,$F9,$D1
	.dc.b	$F8,$DF,$53,$E4,$4F,$D8,$9F,$AF,$3F,$E9,$A3,$F1,$D9,$3D,$5F,$27
	.dc.b	$CB,$FF,$BE,$31,$34,$95,$F3,$79,$BF,$1E,$4D,$FE,$75,$7E,$17,$D9
	.dc.b	$FB,$1F,$BE,$F9,$9B,$7F,$D0,$F3,$46,$23,$DB,$18,$53,$F3,$79,$BE
	.dc.b	$46,$97,$03,$49,$5F,$E8,$D5,$EE,$7C,$FE,$3F,$AC,$D9,$D3,$5F,$DD
	.dc.b	$35,$FA,$92,$F7,$7E,$C7,$67,$DF,$6C,$F7,$6B,$FC,$0D,$AF,$2C,$FF
	.dc.b	$01,$FB,$F7,$E1,$DB,$1E,$76,$BF,$C2,$DA,$F8,$BD,$5F,$C4,$71,$FF
	.dc.b	$BF,$89,$1F,$EE,$A7,$F8,$1F,$0F,$D3,$EE,$FE,$3F,$3F,$9E,$21,$FE
	.dc.b	$E8,$DB,$F8,$09,$FC,$47,$CC,$FD,$77,$A5,$F8,$C6,$CF,$D7,$6B,$F5
	.dc.b	$9A,$FF,$43,$BF,$03,$FC,$B7,$A7,$EA,$F9,$3F,$9F,$EA,$FD,$33,$67
	.dc.b	$EB,$35,$FE,$54,$D7,$F9,$37,$C1,$FE,$69,$FC,$43,$5F,$C3,$B5,$D9
	.dc.b	$FA,$23,$F8,$0E,$7F,$09,$FE,$C3,$9F,$F1,$99,$FE,$19,$D3,$FA,$A6
	.dc.b	$CF,$63,$F4,$25,$8F,$D7,$FC,$27,$F1,$3E,$EB,$26,$91,$FD,$37,$5F
	.dc.b	$F1,$17,$C6,$7F,$07,$F4,$BF,$81,$78,$77,$23,$49,$3F,$9B,$E1,$FE
	.dc.b	$11,$87,$F0,$F6,$C6,$6B,$C4,$D9,$FC,$1E,$1E,$1E,$0F,$AB,$3B,$FB
	.dc.b	$6E,$56,$69,$23,$F6,$DF,$0F,$FE,$10,$E6,$34,$91,$F3,$4F,$C1,$61
	.dc.b	$DC,$FE,$61,$C7,$FB,$6F,$37,$FA,$71,$31,$FE,$A8,$FF,$83,$7E,$21
	.dc.b	$EA,$7F,$89,$B0,$FE,$80,$23,$2F,$F8,$B0,$CB,$8F,$47,$FD,$06,$D8
	.dc.b	$FA,$3C,$BE,$57,$FB,$27,$C4,$FF,$31,$AF,$E7,$0D,$8F,$E1,$0D,$7F
	.dc.b	$BE,$35,$FE,$07,$F1,$5F,$D9,$C9,$97,$F6,$D7,$FC,$4D,$C1,$F3,$A7
	.dc.b	$FC,$5E,$7F,$8A,$25,$F1,$6F,$8C,$FF,$1D,$B6,$7F,$C7,$6D,$7F,$A1
	.dc.b	$35,$FB,$84,$3F,$C2,$1C,$9F,$82,$6D,$7E,$09,$F0,$DF,$8D,$FF,$20
	.dc.b	$47,$8B,$E3,$BF,$B6,$65,$D9,$B3,$FE,$77,$C1,$18,$8F,$EA,$31,$85
	.dc.b	$3F,$D8,$7D,$1F,$EB,$9F,$7F,$70,$34,$95,$FE,$BB,$F8,$D7,$C8,$FE
	.dc.b	$E6,$11,$87,$F7,$4D,$F0,$BF,$D7,$ED,$8F,$F6,$26,$9F,$EC,$69,$FF
	.dc.b	$5E,$12,$33,$CC,$E2,$FE,$09,$F2,$4F,$93,$FF,$60,$D2,$30,$FF,$B1
	.dc.b	$1A,$FF,$B3,$9A,$FF,$B4,$DA,$FE,$59,$AF,$F8,$5D,$AF,$58,$EF,$18
	.dc.b	$23,$82,$47,$85,$AF,$85,$AF,$85,$AF,$85,$AF,$85,$AF,$84,$FF,$6F
	.dc.b	$FB,$FF,$EF,$A7,$FD,$D8,$D7,$FD,$E8,$7F,$BB,$1C,$3F,$EF,$BF,$9F
	.dc.b	$FF,$19,$74,$65,$B1,$C9,$AF,$EE,$25,$18,$8E,$47,$EE,$8D,$39,$51
	.dc.b	$C1,$D1,$C2,$7C,$0E,$0D,$CF,$19,$E1,$3A,$3F,$0F,$E2,$FF,$8B,$FE
	.dc.b	$42,$F0,$9A,$4C,$3F,$29,$F6,$FB,$6C,$ED,$B3,$F9,$29,$D4,$7F,$B9
	.dc.b	$63,$F8,$BB,$FE,$57,$DD,$B7,$E4,$67,$32,$9C,$ED,$D6,$39,$1F,$C8
	.dc.b	$3B,$5E,$F3,$ED,$BF,$C9,$1F,$B4,$76,$9C,$C6,$2B,$85,$6F,$45,$EE
	.dc.b	$30,$FF,$93,$7F,$4C,$7C,$0F,$D5,$A4,$62,$FE,$8C,$62,$75,$39,$E0
	.dc.b	$E9,$BF,$F9,$6E,$2D,$72,$BF,$1C,$E5,$B1,$97,$FC,$F9,$9F,$BB,$DD
	.dc.b	$FF,$32,$61,$E7,$6C,$FB,$1A,$F0,$CD,$18,$8F,$A1,$C1,$FB,$FA,$F9
	.dc.b	$FB,$FC,$E6,$C6,$67,$F8,$2E,$FF,$8B,$FB,$87,$D8,$FF,$A3,$FE,$4F
	.dc.b	$EB,$C6,$29,$18,$9D,$47,$E0,$27,$53,$FD,$4B,$6F,$F4,$C7,$B9,$6B
	.dc.b	$FC,$DC,$7E,$1B,$C6,$4E,$A3,$0D,$7F,$B0,$E9,$E9,$7C,$FA,$7A,$B3
	.dc.b	$FD,$14,$E6,$77,$A7,$B5,$B3,$A8,$C4,$7C,$68,$FB,$BC,$3F,$96,$8F
	.dc.b	$F5,$E1,$FD,$C3,$87,$F0,$75,$FE,$B1,$F2,$7F,$10,$D9,$D5,$71,$9F
	.dc.b	$26,$32,$7C,$E4,$3F,$D3,$3E,$87,$FB,$27,$0E,$F1,$8A,$FF,$72,$3E
	.dc.b	$19,$D4,$7F,$A8,$DB,$1E,$0F,$88,$FF,$62,$7D,$3F,$E6,$DB,$19,$11
	.dc.b	$96,$9E,$33,$E1,$61,$E7,$7E,$3B,$DF,$DA,$FF,$5B,$4D,$C7,$8D,$7A
	.dc.b	$A7,$2F,$9D,$AF,$F8,$63,$CB,$7D,$DF,$F1,$D6,$C7,$F1,$96,$E1,$FC
	.dc.b	$11,$B1,$E5,$3F,$83,$6B,$F6,$1B,$B7,$DB,$39,$76,$FE,$4A,$3C,$0D
	.dc.b	$7F,$DD,$15,$FE,$D3,$E4,$ED,$39,$97,$E2,$0D,$BF,$99,$36,$3F,$97
	.dc.b	$37,$2F,$E3,$2D,$8F,$37,$17,$E4,$5F,$96,$7A,$63,$4F,$F8,$E9,$AF
	.dc.b	$FB,$ED,$AF,$8A,$BC,$3C,$7F,$EF,$FE,$EE,$ED,$9F,$E5,$B5,$FA,$87
	.dc.b	$F9,$41,$CE,$34,$FD,$A3,$CB,$AF,$C9,$C3,$CD,$1F,$02,$7E,$15,$73
	.dc.b	$CF,$BC,$FE,$F8,$EF,$3F,$2F,$83,$F9,$2B,$FC,$88,$91,$FA,$A7,$FA
	.dc.b	$D5,$F6,$26,$1E,$13,$F6,$3D,$7F,$CF,$34,$FE,$30,$DC,$3F,$AF,$B6
	.dc.b	$3F,$9C,$1F,$C7,$70,$7E,$D3,$FE,$28,$AF,$C9,$C0,$CF,$8D,$F9,$27
	.dc.b	$13,$BE,$4F,$EA,$2D,$8F,$8B,$C5,$F8,$BC,$3F,$45,$1C,$FF,$01,$7C
	.dc.b	$8F,$85,$B9,$7F,$71,$6D,$7F,$5C,$6D,$7F,$04,$AF,$27,$DE,$FF,$4F
	.dc.b	$6C,$64,$46,$5A,$7F,$0C,$EE,$F6,$7E,$2F,$FD,$5D,$B1,$91,$19,$69
	.dc.b	$FD,$8A,$98,$BE,$F3,$FB,$14,$61,$B1,$99,$E0,$D7,$FA,$75,$F2,$FC
	.dc.b	$23,$FB,$53,$63,$FB,$7B,$5D,$CA,$7F,$B8,$7D,$2F,$79,$FE,$EA,$D7
	.dc.b	$FA,$D3,$57,$0F,$F8,$02,$31,$0D,$3E,$8F,$E1,$5F,$3F,$3F,$6B,$71
	.dc.b	$34,$FE,$FD,$C5,$F8,$5F,$7E,$E0,$69,$1F,$F0,$4F,$6F,$C8,$FC,$D7
	.dc.b	$FC,$10,$D8,$CE,$09,$CB,$63,$FC,$43,$97,$E0,$9F,$F5,$CB,$5F,$C6
	.dc.b	$65,$F8,$B3,$F4,$3E,$3B,$EF,$3E,$BF,$C6,$FF,$85,$DB,$5F,$E1,$C6
	.dc.b	$BF,$E0,$16,$BF,$E1,$1F,$67,$F4,$5C,$7F,$E1,$D6,$C7,$F5,$66,$BF
	.dc.b	$A7,$CB,$FC,$3F,$18,$8F,$D3,$ED,$F8,$8F,$80,$FD,$3B,$63,$FC,$7B
	.dc.b	$D3,$FB,$19,$CD,$9F,$D4,$CF,$F3,$6E,$8B,$D9,$FC,$43,$27,$DD,$7F
	.dc.b	$31,$6D,$7F,$0F,$3F,$DB,$EE,$3B,$FF,$C8,$ED,$8F,$CA,$FE,$5B,$F9
	.dc.b	$82,$38,$3F,$27,$93,$F2,$4F,$CB,$B6,$33,$18,$3E,$CE,$F2,$3E,$EB
	.dc.b	$5F,$F7,$8F,$D1,$BF,$0B,$FE,$90,$D8,$F4,$1F,$C4,$5C,$65,$FE,$5E
	.dc.b	$6C,$7C,$8E,$0F,$D9,$FB,$3F,$20,$91,$9C,$BF,$D8,$2F,$F1,$4D,$FE
	.dc.b	$4F,$87,$FD,$87,$F4,$07,$99,$FB,$CC,$FF,$CD,$71,$8C,$3C,$78,$23
	.dc.b	$13,$AA,$E1,$3F,$9A,$FA,$BF,$CC,$91,$C1,$FE,$7A,$D3,$DD,$F5,$7E
	.dc.b	$BD,$BF,$97,$36,$4D,$3F,$47,$E9,$48,$FD,$3B,$5F,$E6,$C7,$97,$0F
	.dc.b	$E2,$BF,$95,$79,$74,$34,$93,$F0,$C8,$7E,$67,$F3,$D7,$D0,$7F,$98
	.dc.b	$69,$18,$7F,$CF,$2D,$E0,$FF,$4A,$C6,$A7,$F6,$3C,$5F,$C9,$BE,$3F
	.dc.b	$FD,$0C,$DC,$3F,$D1,$6D,$7F,$CE,$0D,$7F,$96,$FE,$87,$F5,$1F,$86
	.dc.b	$7F,$92,$5C,$6B,$E0,$6C,$67,$4F,$96,$FB,$0F,$F0,$49,$FD,$E4,$F7
	.dc.b	$79,$DE,$DD,$3F,$CC,$F9,$FB,$23,$07,$F6,$5E,$BF,$ED,$FE,$6F,$F5
	.dc.b	$BB,$FF,$1F,$7C,$6F,$FA,$FF,$C1,$FF,$43,$37,$0F,$F6,$6B,$78,$3F
	.dc.b	$D7,$2D,$AF,$F5,$5F,$C3,$FF,$4A,$F8,$7F,$F5,$53,$63,$36,$9C,$B6
	.dc.b	$3E,$17,$C8,$7F,$AB,$77,$34,$91,$E7,$3E,$01,$F0,$89,$3F,$73,$C6
	.dc.b	$F8,$44,$7E,$EB,$F8,$1F,$C9,$BA,$C7,$F9,$B6,$7F,$8F,$46,$3D,$AF
	.dc.b	$EC,$71,$8B,$FE,$CB,$18,$8F,$EC,$F3,$BA,$FA,$91,$FC,$6C,$FF,$34
	.dc.b	$6D,$FD,$E3,$EC,$FF,$D6,$F3,$F9,$C8,$C2,$DB,$A4,$61,$BE,$A2,$B8
	.dc.b	$DF,$8E,$F8,$1F,$E9,$1C,$3D,$77,$E5,$84,$61,$D6,$5C,$AB,$E4,$1E
	.dc.b	$73,$E5,$94,$CF,$0F,$E1,$AF,$FC,$3F,$FA,$03,$E8,$5F,$86,$FF,$8A
	.dc.b	$E7,$E2,$4E,$63,$18,$FE,$EA,$31,$87,$A0,$FF,$19,$38,$AF,$47,$EF
	.dc.b	$5E,$C6,$9E,$7A,$46,$5F,$F5,$83,$6B,$D3,$8F,$9D,$FF,$4C,$5F,$5A
	.dc.b	$F1,$63,$E8,$9F,$10,$9D,$3E,$48,$D6,$1F,$7E,$3E,$21,$2F,$B3,$E0
	.dc.b	$F1,$3F,$CB,$1B,$FE,$56,$3F,$C9,$DF,$0F,$F9,$AF,$8B,$FF,$28,$7B
	.dc.b	$FF,$7E,$73,$BA,$73,$97,$F2,$0B,$E1,$4B,$E0,$A3,$31,$86,$CF,$1B
	.dc.b	$E9,$5C,$63,$17,$DA,$3C,$2D,$7F,$D4,$E7,$FB,$9C,$87,$FC,$00,$29
	.dc.b	$FE,$E6,$29,$FE,$E5,$3F,$9B,$9F,$2F,$3F,$9C,$FE,$11,$BA,$FD,$2D
	.dc.b	$FF,$DD,$13,$F3,$DA,$F3,$3C,$04,$3C,$07,$E0,$2F,$F6,$B3,$4F,$F0
	.dc.b	$B4,$7F,$85,$F7,$FC,$74,$66,$3C,$55,$DB,$7F,$C9,$C6,$0E,$27,$A6
	.dc.b	$3C,$6F,$C4,$3C,$47,$94,$F1,$90,$C1,$0E,$07,$ED,$CF,$DA,$70,$7F
	.dc.b	$10,$6E,$7E,$A7,$57,$FC,$17,$8B,$F8,$03,$E4,$DF,$F8,$3C,$FF,$04
	.dc.b	$E2,$E0,$A3,$33,$E6,$8F,$09,$E3,$E2,$71,$6E,$3F,$AE,$E5,$FC,$8B
	.dc.b	$E0,$3C,$76,$69,$1F,$EE,$67,$F2,$D3,$EF,$CF,$FA,$23,$B3,$FD,$3E
	.dc.b	$FF,$68,$6B,$C1,$AF,$C7,$3C,$A4,$C3,$CB,$8F,$4A,$73,$FF,$42,$AC
	.dc.b	$35,$FE,$CD,$AF,$89,$FE,$DA,$DF,$65,$E9,$BA,$FF,$D3,$BF,$B3,$7F
	.dc.b	$50,$25,$FF,$54,$7F,$AD,$FC,$03,$BF,$F8,$04,$BB,$FF,$B1,$3A,$8C
	.dc.b	$C6,$23,$E8,$6D,$EA,$7F,$4A,$7D,$2D,$7F,$C1,$FE,$07,$46,$9E,$FE
	.dc.b	$0F,$EE,$F3,$F2,$67,$C2,$FC,$5E,$3F,$B0,$F6,$6B,$FE,$75,$E0,$FF
	.dc.b	$06,$7C,$A7,$F8,$27,$C2,$FF,$6F,$36,$4D,$3F,$C3,$38,$FE,$DA,$3F
	.dc.b	$D5,$7C,$1F,$EE,$7E,$8C,$92,$3E,$D7,$8D,$E9,$D3,$FC,$51,$1F,$7E
	.dc.b	$30,$7A,$35,$F8,$51,$E7,$E5,$FA,$D1,$88,$FF,$1E,$F9,$BF,$E5,$3F
	.dc.b	$CB,$BF,$C7,$84,$8F,$E1,$71,$97,$FD,$57,$CF,$FE,$AA,$23,$83,$FC
	.dc.b	$A5,$E5,$FC,$06,$9F,$00,$BB,$7F,$94,$CF,$F5,$9F,$C8,$C8,$FF,$5B
	.dc.b	$B5,$F2,$F0,$7F,$B9,$3E,$63,$FC,$A0,$48,$F2,$C6,$5F,$71,$F8,$0C
	.dc.b	$FD,$CF,$FA,$9B,$87,$F9,$D3,$FE,$79,$F1,$7F,$D4,$64,$7C,$F7,$FE
	.dc.b	$67,$23,$83,$FD,$97,$EB,$7F,$B2,$88,$E0,$FF,$4D,$F8,$DF,$E9,$A2
	.dc.b	$38,$3F,$DE,$FF,$81,$FF,$BE,$88,$E0,$FE,$73,$EB,$FF,$AB,$AF,$FD
	.dc.b	$57,$69,$46,$1A,$70,$8D,$47,$C1,$E8,$FC,$AC,$FF,$11,$6A,$F6,$5E
	.dc.b	$73,$4F,$F5,$34,$FF,$A0,$27,$31,$88,$F9,$5C,$3F,$EC,$A3,$FD,$B0
	.dc.b	$BF,$03,$A4,$7F,$DB,$46,$4F,$F7,$24,$BF,$02,$FF,$74,$36,$33,$18
	.dc.b	$3F,$8E,$FB,$BF,$C6,$23,$FD,$E7,$3F,$C3,$E3,$10,$75,$E2,$7F,$10
	.dc.b	$F7,$FD,$9F,$9B,$7F,$C0,$02,$57,$FA,$72,$75,$69,$FF,$47,$FC,$5F
	.dc.b	$C4,$A7,$CD,$63,$6F,$A4,$D8,$C9,$FC,$D2,$78,$47,$A3,$D3,$F9,$9F
	.dc.b	$89,$5C,$4D,$25,$F8,$FC,$37,$C3,$7D,$3E,$0F,$EF,$7A,$FF,$2A,$8C
	.dc.b	$56,$DA,$FE,$16,$35,$87,$CF,$8C,$5B,$E2,$3E,$F7,$5F,$FB,$B8,$93
	.dc.b	$FE,$F2,$7B,$CB,$3E,$B3,$FC,$51,$B8,$7F,$3C,$6C,$EF,$0F,$05,$7E
	.dc.b	$2F,$5F,$02,$F8,$FF,$9B,$97,$E6,$DB,$19,$3E,$54,$65,$F9,$98,$7E
	.dc.b	$87,$F3,$8F,$AA,$D8,$C8,$8C,$B4,$FF,$11,$F1,$FE,$18,$93,$FC,$F1
	.dc.b	$FC,$4E,$9F,$C1,$A7,$F9,$F7,$89,$FC,$9B,$A3,$F4,$FD,$1F,$A2,$D7
	.dc.b	$FB,$7B,$6B,$FC,$5E,$D7,$F5,$1C,$9F,$AB,$9F,$97,$EA,$FF,$5A,$E1
	.dc.b	$FE,$C6,$DE,$1F,$F3,$0D,$FD,$68,$FE,$71,$8F,$ED,$B3,$F2,$11,$FD
	.dc.b	$DE,$0E,$17,$95,$FD,$C6,$30,$D8,$CE,$1F,$DE,$DA,$FF,$82,$3E,$37
	.dc.b	$3D,$96,$BF,$84,$CF,$F9,$2A,$72,$D8,$CB,$FE,$09,$C7,$C9,$BF,$C7
	.dc.b	$3F,$C0,$55,$E0,$F6,$3A,$57,$F9,$88,$FE,$34,$23,$4F,$53,$47,$F1
	.dc.b	$F5,$F1,$DF,$E0,$D8,$E0,$D8,$CC,$78,$23,$57,$E4,$8F,$E4,$57,$96
	.dc.b	$1F,$06,$79,$1D,$7D,$3F,$E8,$B9,$7F,$8B,$1B,$3F,$E6,$C6,$BE,$A6
	.dc.b	$BF,$CB,$6B,$E1,$3F,$E1,$08,$76,$F5,$24,$7F,$89,$E3,$2F,$F8,$BF
	.dc.b	$C5,$7C,$7F,$E2,$63,$FC,$7F,$18,$DF,$FD,$44,$1C,$3F,$E8,$03,$13
	.dc.b	$C6,$7E,$B7,$07,$E2,$C9,$8F,$E0,$F3,$FF,$37,$CF,$ED,$DB,$7F,$8D
	.dc.b	$6F,$07,$CE,$8E,$D8,$BD,$BF,$E9,$19,$E1,$B8,$7A,$1A,$FF,$59,$AF
	.dc.b	$44,$62,$3E,$41,$F5,$8F,$E6,$39,$7F,$5F,$AB,$1F,$E6,$33,$F9,$74
	.dc.b	$65,$A6,$4F,$7E,$5F,$49,$FF,$01,$90,$FF,$30,$7C,$87,$E9,$3C,$BF
	.dc.b	$F3,$4B,$63,$FD,$52,$D7,$F2,$ED,$7F,$8C,$76,$7E,$89,$A7,$F7,$23
	.dc.b	$F3,$BD,$BF,$C8,$38,$7F,$CE,$AD,$8F,$11,$F0,$0F,$E8,$1F,$23,$FC
	.dc.b	$0C,$9A,$FF,$9F,$3A,$7F,$CD,$31,$FC,$1F,$9F,$FB,$B7,$E0,$5F,$CD
	.dc.b	$9B,$7F,$D2,$5A,$FF,$40,$6B,$FC,$CF,$E4,$FF,$82,$11,$FE,$08,$FF
	.dc.b	$9D,$FA,$DF,$2B,$FE,$03,$EF,$FE,$86,$DA,$FE,$AC,$D7,$FA,$6B,$5F
	.dc.b	$F0,$B7,$57,$FA,$2B,$B3,$FD,$2A,$D8,$FF,$6A,$B5,$FE,$06,$D7,$FC
	.dc.b	$3D,$F1,$3F,$C7,$36,$FE,$07,$08,$C5,$7F,$A8,$1B,$1F,$EE,$06,$BF
	.dc.b	$C3,$9F,$F7,$C4,$F2,$6C,$F1,$F3,$BF,$C4,$0D,$9E,$FE,$73,$E5,$BF
	.dc.b	$88,$37,$0F,$F4,$F3,$5F,$F5,$8B,$5F,$F7,$8B,$5F,$E3,$ED,$7F,$C7
	.dc.b	$4D,$4D,$24,$78,$09,$1E,$0E,$6F,$86,$F0,$6C,$9A,$63,$D6,$FF,$24
	.dc.b	$78,$5E,$66,$D7,$99,$AF,$99,$AF,$99,$AF,$99,$B4,$69,$AC,$FF,$DB
	.dc.b	$B8,$BD,$AF,$AD,$D7,$FE,$4D,$6C,$67,$0F,$B1,$39,$7E,$D7,$1D,$7E
	.dc.b	$D3,$5F,$C2,$B5,$FC,$23,$77,$FF,$6C,$DF,$E4,$2B,$FD,$E8,$DE,$BF
	.dc.b	$05,$FE,$42,$72,$FE,$27,$4F,$F3,$75,$7E,$35,$B1,$93,$89,$F9,$67
	.dc.b	$2F,$F8,$BB,$EF,$3F,$CD,$5F,$8C,$7F,$8A,$9B,$1F,$E3,$D6,$BB,$CD
	.dc.b	$3F,$CE,$7D,$A7,$E0,$9F,$4A,$7F,$DE,$2D,$8C,$D3,$0E,$13,$F6,$3A
	.dc.b	$FF,$DD,$FF,$25,$FE,$7B,$FA,$FF,$F1,$DB,$6F,$FC,$9C,$D7,$FC,$8A
	.dc.b	$D7,$FD,$A3,$F1,$DF,$E8,$3C,$3F,$C8,$6D,$8F,$F2,$B3,$5F,$F2,$6B
	.dc.b	$5F,$E0,$3B,$37,$F3,$47,$F0,$08,$C5,$7D,$32,$46,$67,$6F,$D5,$6B
	.dc.b	$D1,$AB,$81,$5D,$23,$FC,$1B,$EF,$BF,$16,$7E,$1B,$FE,$A9,$FB,$BF
	.dc.b	$F0,$8D,$2B,$53,$98,$E3,$5C,$15,$70,$AD,$C6,$0F,$46,$5E,$A7,$D0
	.dc.b	$D7,$FD,$71,$F9,$5F,$ED,$57,$D2,$F8,$94,$66,$30,$D8,$F3,$B5,$FE
	.dc.b	$D5,$C4,$FB,$BF,$F6,$27,$97,$F8,$96,$CF,$F8,$BB,$95,$5B,$8F,$EF
	.dc.b	$06,$5C,$57,$9E,$3E,$66,$DF,$2D,$F5,$9E,$E3,$A7,$C2,$B8,$FF,$D9
	.dc.b	$15,$FE,$11,$8C,$61,$DA,$BF,$C0,$11,$89,$D5,$70,$9F,$17,$C5,$7F
	.dc.b	$B7,$7E,$7F,$F4,$4D,$8C,$88,$CB,$4F,$21,$E6,$21,$F8,$7D,$BF,$C4
	.dc.b	$9F,$BB,$FE,$4F,$2E,$13,$FE,$13,$8C,$56,$B9,$7F,$DC,$4D,$AD,$D7
	.dc.b	$33,$B9,$4E,$E7,$C8,$9F,$CA,$F2,$FE,$51,$BA,$FC,$B9,$E0,$43,$86
	.dc.b	$3F,$46,$B8,$3F,$E3,$BF,$C3,$FF,$CC,$ED,$9F,$E0,$91,$87,$EF,$35
	.dc.b	$F1,$CF,$27,$CE,$76,$36,$72,$21,$FA,$EC,$13,$CE,$1D,$3F,$EF,$7C
	.dc.b	$7F,$99,$36,$BF,$91,$B7,$B7,$FD,$AF,$F0,$1F,$3D,$F6,$67,$F0,$70
	.dc.b	$F1,$1F,$5A,$94,$F8,$1F,$E1,$51,$86,$C6,$70,$FE,$07,$39,$7D,$19
	.dc.b	$3E,$8F,$E2,$93,$0C,$BF,$D6,$B4,$F8,$91,$F1,$B2,$F2,$46,$23,$DF
	.dc.b	$8F,$FC,$00,$C3,$BB,$75,$F6,$FC,$0F,$90,$F2,$F8,$EA,$7D,$EF,$C2
	.dc.b	$8C,$36,$33,$87,$FA,$8B,$5F,$83,$5A,$8F,$AF,$97,$8D,$F8,$E7,$E1
	.dc.b	$DC,$13,$3F,$49,$F8,$DA,$F9,$7E,$37,$8F,$F6,$8F,$A6,$DB,$F0,$6B
	.dc.b	$FD,$E7,$25,$69,$B3,$F8,$CF,$A8,$2F,$DD,$FF,$29,$18,$6C,$64,$E0
	.dc.b	$53,$8E,$3F,$7B,$D8,$F8,$DF,$7B,$F8,$66,$CF,$91,$AF,$0C,$91,$88
	.dc.b	$FA,$BF,$5D,$F3,$FF,$4C,$FC,$7B,$63,$DA,$D7,$93,$5F,$BB,$EE,$7D
	.dc.b	$7F,$D5,$3F,$2E,$D8,$F8,$AD,$7C,$4D,$7E,$97,$07,$D0,$EA,$FD,$E9
	.dc.b	$F8,$6F,$C3,$3F,$05,$F9,$C7,$E8,$9B,$5F,$41,$AF,$B5,$AF,$E4,$3D
	.dc.b	$3F,$C5,$7E,$D7,$FA,$A6,$C7,$D8,$6B,$F1,$5A,$F9,$2F,$C7,$D5,$FD
	.dc.b	$47,$87,$F5,$BC,$1F,$95,$FC,$D7,$F6,$AD,$BF,$C1,$B5,$FA,$0D,$7F
	.dc.b	$39,$DA,$FC,$03,$F0,$7F,$BF,$7F,$15,$6C,$FF,$12,$6B,$FE,$BF,$9E
	.dc.b	$0A,$7F,$45,$F1,$72,$3F,$27,$DB,$FA,$98,$C4,$7E,$D7,$4F,$5B,$FA
	.dc.b	$4E,$1F,$A1,$AF,$F0,$C9,$FD,$AF,$37,$EC,$7E,$37,$F1,$9D,$66,$93
	.dc.b	$0F,$C2,$C3,$EB,$7F,$12,$D8,$CC,$60,$FA,$51,$CE,$3E,$93,$5F,$A4
	.dc.b	$D7,$ED,$35,$FD,$C9,$3B,$FF,$71,$87,$F0,$CD,$3F,$84,$7C,$2F,$F0
	.dc.b	$52,$57,$C4,$B4,$7B,$F3,$93,$F0,$EF,$E2,$BF,$CA,$F8,$7F,$D0,$56
	.dc.b	$F8,$CF,$49,$27,$D2,$F2,$7E,$75,$FB,$BE,$B2,$47,$B9,$AE,$8F,$04
	.dc.b	$71,$7F,$20,$D7,$DD,$BF,$F1,$3C,$3F,$BC,$60,$FC,$4B,$FA,$2F,$93
	.dc.b	$FC,$9D,$B1,$9B,$FE,$2B,$39,$7F,$C3,$5E,$1B,$E1,$3F,$89,$FD,$57
	.dc.b	$F8,$EA,$FF,$8C,$C6,$2B,$6D,$7F,$8C,$C6,$94,$F9,$F8,$7E,$9E,$97
	.dc.b	$4F,$C6,$D7,$B8,$87,$C1,$21,$FC,$B7,$E0,$BF,$92,$FE,$11,$E3,$CD
	.dc.b	$18,$9F,$4B,$5F,$1C,$69,$4F,$1F,$C1,$BF,$D0,$12,$69,$27,$EF,$9F
	.dc.b	$40,$F2,$94,$FC,$17,$B7,$FD,$4F,$E4,$3F,$AA,$47,$F2,$B8,$C6,$BF
	.dc.b	$D0,$9B,$19,$84,$F0,$8E,$84,$3E,$EF,$27,$89,$FE,$AB,$DF,$FD,$EA
	.dc.b	$FF,$B1,$7C,$0F,$18,$C6,$08,$C7,$27,$F9,$D2,$30,$78,$CB,$7F,$38
	.dc.b	$EF,$BF,$DF,$FB,$0F,$F4,$5F,$C2,$79,$69,$18,$7F,$CC,$0D,$AF,$C0
	.dc.b	$B5,$FE,$C9,$F2,$1F,$EE,$49,$FC,$6B,$63,$EA,$B5,$D9,$FE,$AF,$8F
	.dc.b	$ED,$35,$FE,$A5,$69,$98,$FE,$B1,$19,$6C,$65,$F9,$5C,$7F,$90,$7C
	.dc.b	$26,$63,$FB,$9D,$78,$5B,$19,$E0,$4F,$96,$3E,$29,$F8,$22,$9F,$27
	.dc.b	$E1,$9E,$2B,$E1,$FF,$BD,$F2,$FF,$AB,$A7,$31,$8D,$7F,$BF,$DA,$72
	.dc.b	$FE,$72,$33,$18,$AE,$57,$D6,$B4,$7F,$82,$27,$55,$C2,$75,$5C,$2F
	.dc.b	$C1,$7F,$EB,$D3,$FD,$EC,$43,$FD,$88,$53,$FD,$88,$53,$81,$F2,$8F
	.dc.b	$C2,$63,$F1,$9F,$E0,$27,$DC,$29,$F7,$0F,$CF,$1F,$E7,$93,$A1,$4F
	.dc.b	$09,$F7,$8F,$CC,$E9,$FA,$CC,$F0,$48,$EC,$78,$B3,$FF,$59,$38,$8F
	.dc.b	$31,$F9,$6C,$BE,$9B,$63,$F8,$4B,$71,$FC,$F6,$BF,$EE,$A7,$CF,$AF
	.dc.b	$FA,$36,$BC,$ED,$7C,$E7,$EE,$FE,$DF,$F5,$FC,$5F,$E5,$56,$E0,$69
	.dc.b	$D3,$9F,$FD,$F1,$5D,$1A,$F4,$3E,$79,$F4,$8C,$1F,$1C,$FE,$15,$F8
	.dc.b	$C7,$62,$57,$EF,$5E,$DC,$BF,$EE,$19,$E0,$74,$21,$82,$63,$ED,$7C
	.dc.b	$07,$F2,$AE,$7E,$57,$FE,$5A,$6F,$07,$F2,$99,$FF,$77,$4E,$5B,$1F
	.dc.b	$CC,$7E,$0B,$F9,$C6,$BF,$E5,$E6,$C9,$A7,$C3,$C7,$D1,$8F,$8E,$7C
	.dc.b	$ED,$7C,$F9,$FC,$87,$CE,$69,$C3,$67,$9C,$62,$38,$C6,$0F,$3F,$5F
	.dc.b	$F9,$CB,$E4,$3F,$A7,$36,$7F,$AF,$88,$CB,$4F,$ED,$79,$BF,$28,$FE
	.dc.b	$DE,$48,$CB,$ED,$61,$F7,$31,$F6,$13,$1F,$33,$71,$F3,$4F,$EC,$FA
	.dc.b	$BF,$D7,$7C,$FF,$EA,$3F,$6B,$FC,$52,$49,$FF,$14,$BD,$9F,$8A,$7E
	.dc.b	$0A,$78,$E4,$FB,$8F,$F1,$89,$27,$FC,$62,$F1,$7E,$51,$4F,$92,$7F
	.dc.b	$AB,$FF,$11,$8F,$F8,$AD,$AF,$83,$2F,$53,$E9,$EC,$FF,$14,$FB,$9F
	.dc.b	$E4,$02,$47,$F9,$01,$FF,$23,$FB,$92,$3F,$C9,$6D,$7F,$D3,$1E,$5F
	.dc.b	$FB,$0E,$7F,$CA,$64,$8F,$F4,$F6,$5F,$ED,$77,$E7,$6F,$F3,$09,$BF
	.dc.b	$CC,$8C,$B5,$FF,$52,$79,$3F,$EC,$FF,$C1,$BF,$D9,$84,$8F,$F2,$43
	.dc.b	$FE,$75,$5F,$B4,$52,$3F,$E7,$73,$27,$FA,$0B,$07,$EE,$FF,$E8,$66
	.dc.b	$C6,$63,$07,$CF,$DB,$E4,$C7,$F9,$93,$F6,$AF,$F3,$39,$1C,$1F,$EA
	.dc.b	$1A,$7E,$77,$FE,$A5,$6C,$66,$30,$75,$D3,$E1,$C7,$FA,$DA,$5F,$B7
	.dc.b	$52,$47,$FA,$E2,$32,$FF,$9F,$BF,$4C,$FF,$3E,$91,$C1,$F4,$BF,$0F
	.dc.b	$FF,$43,$91,$C1,$FE,$DA,$FD,$E3,$FD,$BA,$47,$07,$FB,$AB,$E1,$3F
	.dc.b	$E0,$07,$EB,$9F,$EE,$92,$46,$70,$47,$AF,$8B,$ED,$4F,$E6,$0F,$BA
	.dc.b	$7F,$70,$1D,$2E,$77,$23,$07,$E1,$3B,$AD,$4E,$63,$15,$C2,$B7,$7F
	.dc.b	$E8,$B9,$CF,$47,$F6,$08,$EF,$48,$C6,$1E,$16,$C6,$6F,$D3,$5F,$AE
	.dc.b	$3F,$A2,$2F,$6B,$F3,$91,$FE,$75,$6B,$FE,$0C,$8D,$74,$7F,$38,$9C
	.dc.b	$CF,$FB,$FF,$B3,$B9,$2B,$E2,$47,$9A,$3D,$0F,$84,$FD,$D7,$1F,$F2
	.dc.b	$CC,$BE,$46,$BF,$1F,$CC,$F8,$6D,$AF,$F0,$EC,$EB,$0F,$F0,$C4,$0F
	.dc.b	$87,$3B,$7F,$DB,$47,$9C,$87,$F4,$CC,$7F,$AD,$3F,$6C,$F1,$C6,$5D
	.dc.b	$9F,$0C,$A7,$E0,$08,$70,$3F,$0C,$43,$E4,$7C,$4A,$3F,$05,$5F,$EA
	.dc.b	$18,$C3,$FE,$1A,$0E,$1D,$47,$09,$EE,$79,$CF,$C6,$7C,$73,$E2,$FF
	.dc.b	$C4,$B1,$86,$D2,$75,$3D,$DB,$1F,$EB,$3F,$8E,$7C,$6F,$E5,$FC,$0C
	.dc.b	$36,$FF,$05,$3A,$A5,$69,$B3,$E6,$3E,$F1,$E6,$F8,$FF,$AD,$3F,$EC
	.dc.b	$1C,$11,$89,$FC,$74,$61,$E6,$7F,$B0,$9C,$4F,$E3,$3E,$43,$FC,$CB
	.dc.b	$3F,$DD,$1B,$9E,$7C,$5F,$C2,$36,$7F,$BC,$F8,$1F,$C8,$1F,$B0,$7F
	.dc.b	$9A,$3C,$97,$C9,$BD,$0F,$C6,$7C,$EF,$EF,$67,$31,$8A,$E1,$5B,$55
	.dc.b	$B8,$C3,$F8,$38,$EE,$43,$FC,$51,$C8,$F9,$C7,$CB,$7E,$E9,$BC,$3F
	.dc.b	$AA,$9D,$4E,$4D,$71,$FF,$0D,$FC,$AB,$F8,$65,$FD,$56,$C6,$70,$FD
	.dc.b	$4C,$E5,$F9,$FF,$33,$FE,$32,$E3,$34,$91,$E1,$9F,$F5,$FF,$E5,$5D
	.dc.b	$BD,$CF,$E3,$8D,$D0,$D3,$F8,$0F,$B6,$F6,$BF,$6D,$DB,$86,$E3,$FC
	.dc.b	$A6,$74,$FE,$F0,$FD,$E3,$89,$FE,$17,$C8,$EE,$FD,$64,$61,$B5,$FC
	.dc.b	$CE,$75,$CD,$CE,$11,$8A,$F2,$FC,$E7,$F0,$79,$EE,$D5,$82,$30,$DC
	.dc.b	$FA,$3E,$36,$E3,$A2,$3E,$1F,$75,$E4,$7F,$10,$6E,$1F,$E6,$B6,$BF
	.dc.b	$CA,$1A,$FF,$4D,$6B,$FE,$BA,$8E,$44,$79,$75,$7B,$DF,$F6,$13,$67
	.dc.b	$FD,$87,$1C,$E3,$1D,$9F,$CA,$76,$F0,$36,$7C,$0D,$7F,$D9,$2D,$7F
	.dc.b	$D9,$FF,$55,$FD,$77,$E6,$3F,$D0,$8D,$8F,$E8,$2D,$7F,$B3,$35,$FB
	.dc.b	$BF,$63,$5F,$9E,$DD,$4D,$35,$C1,$F7,$2F,$F0,$1C,$26,$92,$FF,$19
	.dc.b	$97,$F4,$58,$FC,$84,$6C,$C9,$4F,$F7,$4E,$1F,$92,$8F,$CC,$FC,$07
	.dc.b	$19,$FC,$B3,$5D,$B5,$FC,$BB,$7A,$FF,$C0,$5C,$FF,$E0,$1F,$B6,$FE
	.dc.b	$FA,$DA,$FE,$B7,$E4,$FE,$C3,$8D,$5A,$6C,$FF,$71,$F7,$7F,$86,$BF
	.dc.b	$A6,$3F,$B0,$42,$33,$A7,$F8,$4E,$31,$DF,$F5,$DB,$5C,$4F,$F0,$A5
	.dc.b	$7E,$C3,$EE,$B8,$36,$3C,$17,$FD,$B6,$72,$D8,$F2,$7C,$6F,$5A,$34
	.dc.b	$91,$DC,$FD,$8F,$47,$F6,$3A,$FB,$24,$8F,$DC,$46,$5F,$E6,$1F,$85
	.dc.b	$FE,$17,$C5,$34,$99,$7D,$99,$FF,$01,$75,$3E,$17,$FC,$09,$E2,$7F
	.dc.b	$84,$A3,$13,$FC,$B8,$FF,$1C,$38,$8D,$4F,$09,$F1,$F3,$7A,$79,$FF
	.dc.b	$85,$E9,$E7,$9F,$F0,$81,$EC,$9F,$76,$1D,$BE,$23,$9C,$9A,$47,$FC
	.dc.b	$91,$EC,$BF,$0E,$FF,$23,$7C,$F7,$F8,$A2,$31,$AA,$31,$3A,$AE,$2D
	.dc.b	$7C,$79,$F9,$9F,$4E,$7E,$77,$FC,$4D,$19,$7D,$3F,$4F,$FC,$C2,$FF
	.dc.b	$C6,$ED,$8F,$F2,$E3,$5D,$D7,$F8,$BE,$7F,$C7,$BF,$89,$3E,$FF,$FC
	.dc.b	$BD,$A6,$1B,$7F,$E6,$59,$D6,$7F,$0E,$31,$3A,$7D,$3C,$7F,$81,$6E
	.dc.b	$FF,$46,$31,$1E,$A6,$BF,$CB,$38,$3F,$C8,$58,$FC,$D8,$FA,$BC,$1F
	.dc.b	$4D,$FB,$F3,$A7,$FC,$F1,$FA,$87,$93,$5F,$F2,$BB,$63,$39,$7F,$9E
	.dc.b	$67,$2D,$8F,$9A,$43,$CF,$1F,$3D,$7E,$49,$FC,$6B,$4F,$CB,$46,$3C
	.dc.b	$1F,$E4,$13,$9F,$03,$FC,$CE,$D9,$FF,$46,$7D,$77,$E5,$F0,$FF,$35
	.dc.b	$B6,$33,$CD,$FC,$CD,$46,$5F,$19,$E5,$3E,$47,$07,$CB,$3F,$69,$1B
	.dc.b	$8F,$F3,$53,$5E,$73,$D6,$30,$47,$04,$8E,$95,$F0,$79,$3F,$1D,$1E
	.dc.b	$08,$F4,$CE,$DB,$1E,$28,$F5,$70,$FF,$AA,$23,$F8,$43,$67,$FD,$16
	.dc.b	$D8,$FF,$58,$35,$F7,$90,$F0,$1F,$04,$FF,$04,$7B,$7D,$88,$F6,$7E
	.dc.b	$57,$D7,$BE,$F3,$FD,$71,$F4,$FF,$E9,$A8,$C7,$02,$31,$3A,$BF,$F6
	.dc.b	$15,$6A,$7E,$16,$1F,$96,$CF,$E1,$D7,$F0,$7C,$BF,$D4,$BF,$6C,$FB
	.dc.b	$1F,$B4,$D9,$FE,$50,$D9,$DC,$FF,$2F,$E5,$FF,$50,$7D,$2F,$F3,$86
	.dc.b	$DF,$F4,$76,$BF,$D1,$BC,$1F,$E4,$33,$FE,$CE,$FC,$4F,$FD,$A9,$F8
	.dc.b	$1D,$E0,$34,$97,$FD,$47,$E6,$FF,$D5,$DF,$8D,$7F,$46,$6C,$64,$46
	.dc.b	$5A,$7F,$5E,$F8,$1F,$F7,$17,$E9,$7F,$ED,$E6,$C6,$70,$4E,$5B,$1F
	.dc.b	$D9,$BD,$FF,$F7,$57,$ED,$3F,$EE,$66,$C6,$65,$39,$6C,$7F,$BD,$78
	.dc.b	$7E,$67,$DD,$FC,$A8,$34,$8F,$F6,$FF,$77,$FD,$99,$FB,$8F,$F6,$46
	.dc.b	$C6,$44,$65,$A7,$E0,$A3,$FE,$00,$7D,$87,$E2,$5B,$19,$C1,$39,$6C
	.dc.b	$79,$F9,$3F,$06,$47,$F0,$6F,$F8,$C7,$D3,$7A,$4F,$C6,$16,$FD,$E3
	.dc.b	$FA,$AE,$0B,$DD,$6A,$73,$18,$AE,$15,$FE,$14,$AD,$3F,$F0,$07,$F1
	.dc.b	$1F,$F6,$D2,$FD,$3B,$FC,$1D,$18,$6F,$89,$FE,$F3,$6C,$EC,$FD,$2F
	.dc.b	$83,$FC,$FD,$FF,$7C,$78,$3F,$AC,$F2,$FF,$57,$97,$F4,$59,$FE,$DB
	.dc.b	$39,$7F,$75,$E0,$7E,$A8,$9E,$0F,$F4,$2E,$77,$B1,$FE,$4B,$6E,$1F
	.dc.b	$CC,$1B,$E4,$7C,$4C,$3F,$DA,$35,$FE,$53,$AF,$F1,$57,$6A,$5F,$43
	.dc.b	$F9,$B2,$3F,$9F,$7F,$2D,$FA,$44,$8F,$CD,$35,$F4,$F4,$7D,$4F,$5B
	.dc.b	$EF,$12,$3F,$04,$F8,$FE,$B1,$F4,$FF,$E6,$36,$E5,$FD,$21,$BA,$6E
	.dc.b	$7F,$CD,$1C,$1F,$C5,$1F,$92,$4C,$3E,$7D,$F4,$7E,$4E,$FF,$C4,$C9
	.dc.b	$C9,$FE,$35,$FC,$11,$F5,$FF,$E5,$BA,$46,$1F,$EB,$2D,$CB,$E5,$36
	.dc.b	$3F,$39,$F5,$5F,$9B,$23,$83,$F2,$7A,$7B,$3F,$20,$F5,$66,$69,$27
	.dc.b	$F4,$DE,$5F,$E9,$48,$E0,$FC,$EE,$1E,$5F,$A9,$F1,$E7,$F6,$5B,$26
	.dc.b	$9F,$CD,$34,$FE,$AF,$1F,$D9,$71,$7E,$84,$E7,$FF,$34,$36,$7F,$CD
	.dc.b	$13,$A9,$C9,$FE,$6D,$EC,$7D,$9F,$E2,$7A,$3F,$4A,$DB,$FD,$53,$5F
	.dc.b	$F3,$CC,$F0,$53,$FE,$79,$F5,$BC,$1E,$85,$C3,$FC,$F5,$18,$AF,$F2
	.dc.b	$8E,$3F,$E3,$19,$FF,$01,$76,$7E,$93,$4F,$6D,$7F,$63,$AF,$6F,$DC
	.dc.b	$FB,$09,$1E,$CD,$51,$CB,$9B,$F0,$B9,$FC,$73,$F8,$5F,$C0,$3F,$06
	.dc.b	$FF,$2A,$61,$FE,$8B,$8C,$7B,$BF,$EC,$56,$C6,$67,$75,$FC,$43,$C7
	.dc.b	$7C,$B7,$F1,$8E,$1C,$36,$FF,$D4,$AD,$7F,$69,$E3,$FF,$20,$8C,$47
	.dc.b	$F2,$0E,$37,$C5,$7D,$32,$47,$D8,$8C,$BF,$10,$FF,$02,$9F,$E0,$73
	.dc.b	$FB,$A1,$F3,$3E,$51,$F4,$BF,$EA,$96,$C7,$E0,$9B,$9F,$FA,$BD,$B1
	.dc.b	$F5,$3F,$0C,$7C,$47,$F3,$4F,$8B,$FE,$6B,$3F,$EB,$89,$CF,$2A,$73
	.dc.b	$3F,$CE,$A7,$31,$89,$46,$1F,$DD,$F3,$7A,$EB,$D2,$7F,$19,$5F,$AD
	.dc.b	$FF,$45,$86,$5D,$A7,$3E,$77,$FC,$01,$BF,$F4,$DC,$62,$B7,$18,$BE
	.dc.b	$57,$E3,$C3,$D3,$CB,$73,$FF,$4C,$DF,$C1,$6E,$1E,$F6,$BF,$40,$87
	.dc.b	$A0,$C4,$3F,$2E,$7C,$E7,$FA,$3E,$3F,$3F,$18,$9F,$F5,$24,$62,$75
	.dc.b	$1F,$EE,$49,$D5,$FC,$CF,$C3,$9F,$84,$FE,$0B,$3C,$37,$1F,$F5,$DC
	.dc.b	$EB,$8F,$E2,$63,$FD,$0E,$30,$FC,$43,$1F,$90,$3D,$CF,$F4,$35,$F2
	.dc.b	$8C,$3B,$95,$27,$38,$F9,$5B,$5F,$72,$3F,$61,$5E,$4C,$7C,$CF,$8D
	.dc.b	$AF,$46,$DF,$9A,$7F,$17,$3F,$34,$9D,$FF,$C1,$63,$AC,$7E,$86,$3F
	.dc.b	$CA,$BF,$29,$FA,$AF,$96,$FE,$C5,$F6,$1F,$88,$AF,$EC,$93,$95,$19
	.dc.b	$51,$98,$C3,$6F,$FD,$CE,$D9,$E7,$AA,$FB,$D6,$A3,$13,$DA,$31,$3A
	.dc.b	$8F,$F4,$D9,$FD,$EF,$07,$81,$FE,$16,$8F,$F7,$C9,$97,$0F,$FB,$30
	.dc.b	$C7,$2F,$F8,$0D,$FF,$7D,$18,$3F,$C5,$39,$FE,$01,$FC,$3E,$3F,$CC
	.dc.b	$1F,$A4,$D7,$E9,$35,$E6,$78,$88,$78,$4F,$C3,$6F,$FC,$13,$6A,$7F
	.dc.b	$84,$4B,$1F,$85,$8D,$3E,$03,$45,$3C,$A6,$0F,$19,$63,$F5,$9C,$5F
	.dc.b	$E1,$DF,$96,$FE,$19,$C7,$FA,$B6,$C6,$67,$F5,$94,$9F,$C5,$E9,$F3
	.dc.b	$B3,$F9,$0D,$9F,$90,$7B,$8F,$E3,$5C,$9F,$E5,$4A,$FE,$0D,$7C,$CF
	.dc.b	$E0,$C5,$9F,$E0,$CD,$57,$FC,$4D,$FE,$31,$CD,$FC,$6D,$BC,$FF,$46
	.dc.b	$75,$5E,$19,$FA,$27,$CF,$21,$E4,$7F,$32,$78,$B8,$FC,$A4,$7F,$A1
	.dc.b	$75,$7E,$15,$75,$7F,$42,$AF,$E8,$13,$98,$C4,$78,$74,$F8,$8D,$E2
	.dc.b	$FC,$8B,$FD,$5F,$87,$A7,$5F,$F5,$48,$FF,$35,$C6,$0F,$CA,$76,$FF
	.dc.b	$5F,$C9,$F5,$BF,$D9,$9B,$5F,$D9,$9A,$FF,$66,$F0,$3F,$B6,$C6,$21
	.dc.b	$E1,$7F,$74,$7F,$BD,$F8,$5F,$DE,$9B,$DD,$FA,$39,$D4,$7A,$78,$7E
	.dc.b	$FE,$BE,$A8,$F9,$A7,$FA,$43,$C4,$FF,$04,$BF,$E6,$DC,$7F,$C1,$ED
	.dc.b	$8C,$BF,$E9,$48,$CB,$F8,$83,$EA,$9D,$0E,$AF,$F8,$75,$E9,$E3,$7F
	.dc.b	$8C,$DF,$F1,$21,$8F,$1B,$E5,$F8,$FF,$E7,$9D,$BF,$C6,$33,$F0,$6D
	.dc.b	$C3,$FE,$2F,$9C,$9E,$3C,$7A,$57,$F8,$B1,$AF,$EC,$DA,$FB,$F0,$FE
	.dc.b	$10,$F9,$F8,$BE,$FB,$EA,$6B,$FE,$8B,$F2,$BF,$CD,$5B,$7F,$95,$09
	.dc.b	$1F,$E5,$AF,$2B,$FD,$21,$1F,$EC,$3F,$B2,$93,$FE,$C7,$6B,$FE,$95
	.dc.b	$1E,$77,$FA,$1B,$F1,$CF,$F3,$59,$23,$FC,$DF,$E9,$7F,$B1,$23,$FD
	.dc.b	$9D,$D5,$27,$FD,$A4,$D7,$A6,$5F,$E7,$FF,$13,$FC,$D2,$48,$F3,$1F
	.dc.b	$CD,$0F,$77,$27,$E0,$F5,$F7,$C7,$D9,$F4,$DE,$D7,$FB,$53,$C3,$FF
	.dc.b	$5E,$92,$7F,$D2,$AF,$FA,$0B,$09,$1E,$8E,$5E,$CF,$FA,$CF,$6F,$F5
	.dc.b	$1C,$FF,$B9,$FE,$BE,$4F,$FB,$AD,$AF,$E8,$7C,$6F,$F7,$1F,$E8,$FF
	.dc.b	$E4,$F6,$C7,$C9,$EF,$FF,$74,$C6,$23,$86,$2A,$FC,$B7,$C5,$7F,$9D
	.dc.b	$27,$FC,$12,$7F,$9C,$1C,$79,$9A,$F4,$7F,$D9,$0D,$9D,$9F,$5F,$A7
	.dc.b	$FD,$A7,$A7,$FB,$45,$B1,$FE,$F6,$6B,$BF,$73,$FD,$F7,$3F,$A5,$8F
	.dc.b	$EB,$E7,$EC,$CF,$F7,$3E,$51,$FF,$5F,$B5,$FF,$46,$F8,$1F,$ED,$BF
	.dc.b	$CF,$3F,$DE,$64,$8F,$F7,$E4,$65,$FF,$80,$12,$FD,$F7,$FC,$ED,$E4
	.dc.b	$A3,$13,$A9,$CC,$62,$3F,$29,$F0,$CF,$83,$FD,$2B,$63,$FB,$73,$5F
	.dc.b	$F7,$93,$5F,$CE,$0E,$77,$BD,$E1,$FA,$AF,$F7,$C4,$63,$04,$62,$75
	.dc.b	$5C,$AB,$53,$F6,$A7,$FD,$F9,$F9,$A5,$A3,$48,$F8,$B0,$F4,$7E,$1D
	.dc.b	$E6,$C9,$28,$C3,$4E,$11,$A8,$FF,$4F,$70,$FC,$9F,$DA,$6E,$86,$92
	.dc.b	$FF,$29,$E6,$FF,$92,$3F,$3B,$FC,$C5,$1A,$49,$FC,$97,$37,$93,$F2
	.dc.b	$6F,$C8,$12,$32,$1F,$F5,$87,$C8,$FD,$BF,$C5,$2E,$06,$92,$BF,$3D
	.dc.b	$F0,$5F,$03,$ED,$3F,$46,$D8,$C8,$8C,$B4,$F8,$67,$E0,$37,$FC,$19
	.dc.b	$1F,$E9,$7C,$F1,$FC,$F1,$FC,$27,$C3,$FC,$84,$7E,$01,$B1,$F7,$5A
	.dc.b	$FF,$23,$E6,$FE,$1F,$97,$F6,$EF,$15,$F2,$DF,$CD,$9B,$5F,$E3,$F6
	.dc.b	$BF,$90,$3F,$C9,$31,$FC,$F3,$C6,$FE,$36,$49,$FE,$3C,$7F,$61,$38
	.dc.b	$CF,$E6,$1A,$FF,$35,$6B,$F8,$86,$AF,$98,$7C,$C7,$F4,$A6,$C7,$F9
	.dc.b	$4D,$AF,$95,$AF,$F0,$12,$1F,$D0,$DC,$B6,$3A,$11,$E8,$D7,$A3,$5E
	.dc.b	$67,$F0,$E3,$FB,$9C,$74,$7F,$42,$D7,$D5,$1C,$5F,$EA,$9F,$32,$F9
	.dc.b	$AF,$E3,$31,$FB,$C8,$C6,$5F,$CF,$21,$18,$F0,$7E,$93,$63,$C6,$7F
	.dc.b	$6D,$C7,$FC,$21,$F3,$6F,$5F,$FD,$3B,$3F,$B4,$8C,$61,$FE,$7B,$6C
	.dc.b	$7D,$D6,$BF,$C3,$BE,$71,$C9,$FE,$67,$8F,$DE,$C6,$1E,$AD,$AF,$C3
	.dc.b	$36,$3C,$BB,$7E,$17,$8B,$FD,$15,$F3,$BE,$2B,$FC,$63,$6B,$F8,$EB
	.dc.b	$5F,$EF,$4D,$7F,$86,$FC,$F7,$FA,$BF,$E0,$7F,$28,$D8,$FE,$4A,$D7
	.dc.b	$FC,$04,$D7,$EC,$10,$F9,$17,$FE,$9D,$FA,$07,$D0,$7F,$9C,$E7,$F9
	.dc.b	$7C,$63,$1F,$F3,$F4,$62,$74,$FF,$83,$E7,$58,$7C,$2D,$3E,$44,$B4
	.dc.b	$F8,$8F,$F8,$55,$AF,$CC,$8D,$E9,$F1,$E3,$F7,$EF,$F8,$22,$7E,$87
	.dc.b	$E2,$9D,$64,$D2,$3F,$D1,$BD,$97,$D1,$7E,$3A,$3F,$C4,$91,$8D,$3F
	.dc.b	$1C,$DE,$4F,$DF,$36,$7E,$CF,$1F,$EF,$37,$FF,$2C,$79,$EF,$94,$FF
	.dc.b	$40,$36,$B3,$18,$33,$F1,$5F,$D1,$3F,$16,$FF,$03,$B6,$7F,$C0,$CD
	.dc.b	$7F,$D7,$95,$FE,$EB,$9F,$F0,$6F,$C4,$F6,$24,$F8,$9F,$F4,$0F,$8E
	.dc.b	$FA,$6F,$EE,$5F,$8A,$7F,$66,$F6,$13,$9E,$04,$E6,$30,$DA,$E3,$87
	.dc.b	$F6,$EF,$91,$7A,$DF,$D7,$38,$BF,$91,$46,$27,$F6,$CD,$8C,$CE,$E7
	.dc.b	$96,$D5,$FF,$2A,$52,$31,$1F,$DA,$23,$0F,$48,$CA,$E6,$FF,$66,$5F
	.dc.b	$FB,$D0,$91,$FE,$04,$8C,$BE,$9F,$03,$F0,$0F,$56,$AB,$FE,$14,$F9
	.dc.b	$23,$2B,$EA,$29,$35,$FE,$29,$19,$7F,$8B,$7C,$87,$E9,$23,$FA,$33
	.dc.b	$5F,$F0,$BB,$5F,$F3,$A3,$5F,$F0,$3A,$F8,$BE,$AD,$CB,$FA,$84,$A3
	.dc.b	$0F,$69,$DF,$E1,$5F,$E1,$0D,$3C,$2D,$9F,$0B,$5F,$0B,$5F,$A1,$F1
	.dc.b	$DF,$E3,$9F,$99,$FB,$AD,$8F,$C1,$35,$FB,$2D,$7F,$C3,$FB,$7F,$BF
	.dc.b	$E7,$FB,$33,$63,$EA,$74,$FF,$BD,$A7,$2D,$8F,$D9,$FE,$1E,$F7,$3F
	.dc.b	$C9,$5F,$33,$FB,$98,$C6,$E8,$C4,$EA,$B8,$B5,$F2,$CE,$DF,$4B,$5C
	.dc.b	$CF,$BA,$31,$1E,$AB,$76,$62,$3F,$21,$18,$AF,$4C,$E7,$24,$E6,$BF
	.dc.b	$C6,$BF,$31,$FE,$63,$FC,$13,$F5,$CD,$8C,$E2,$9C,$B6,$3F,$D7,$0F
	.dc.b	$FB,$4B,$9F,$EB,$36,$3F,$4A,$7F,$93,$9C,$70,$7F,$B9,$FE,$3F,$F8
	.dc.b	$14,$E8,$75,$FF,$8E,$5F,$F3,$B7,$D4,$FF,$05,$F5,$79,$36,$32,$7F
	.dc.b	$A8,$63,$2D,$F8,$6F,$F3,$F6,$1C,$5B,$19,$C7,$C1,$87,$F9,$1E,$73
	.dc.b	$3F,$ED,$3D,$BE,$3E,$FE,$58,$F9,$AD,$8C,$E1,$F3,$8F,$F1,$87,$C2
	.dc.b	$7F,$51,$FC,$12,$F9,$9F,$E8,$F6,$E6,$69,$FD,$93,$C2,$FF,$39,$FD
	.dc.b	$7D,$CC,$D2,$3F,$96,$CF,$C5,$C9,$F8,$2C,$FC,$53,$96,$BE,$B8,$D3
	.dc.b	$92,$1F,$57,$85,$EF,$FD,$A6,$C7,$D5,$CD,$18,$BF,$B5,$5F,$B5,$AF
	.dc.b	$8B,$1E,$07,$F2,$D8,$79,$8E,$5F,$36,$F7,$BE,$7C,$FF,$AC,$E3,$18
	.dc.b	$EA,$72,$A3,$37,$F8,$9F,$9C,$FF,$4C,$FC,$85,$C0,$D2,$3E,$E3,$C7
	.dc.b	$3E,$57,$CE,$7E,$6A,$BE,$7D,$F2,$8C,$6B,$FE,$C7,$BF,$F5,$DC,$F8
	.dc.b	$E3,$13,$A8,$E0,$FF,$AB,$07,$0F,$F8,$4F,$B7,$F0,$24,$8F,$F6,$2C
	.dc.b	$65,$F3,$1E,$43,$EC,$69,$F2,$72,$FC,$3C,$70,$E2,$FC,$63,$70,$FC
	.dc.b	$63,$78,$BF,$CC,$6F,$E6,$7D,$9F,$F3,$9D,$E0,$91,$E0,$E4,$FC,$DB
	.dc.b	$78,$FF,$0E,$FF,$9C,$4E,$5E,$CF,$DF,$BE,$71,$FE,$DE,$6B,$FE,$CC
	.dc.b	$F8,$77,$C5,$FD,$F8,$FE,$33,$18,$E4,$FF,$68,$F7,$FF,$B9,$A3,$13
	.dc.b	$AA,$E1,$3F,$5F,$2F,$2C,$FC,$4C,$FF,$1D,$7C,$3F,$19,$E5,$87,$76
	.dc.b	$E3,$FE,$DE,$6D,$9F,$B7,$F4,$DF,$5F,$0B,$3F,$C6,$A3,$0F,$F1,$E6
	.dc.b	$BE,$13,$E3,$70,$FD,$57,$F0,$1D,$6F,$B2,$7D,$C7,$E7,$A3,$0D,$8C
	.dc.b	$F1,$FF,$21,$9C,$BF,$C3,$FE,$05,$F8,$E7,$E3,$B0,$FF,$62,$35,$FF
	.dc.b	$7C,$C6,$A7,$F2,$31,$89,$D3,$F8,$B3,$F4,$27,$A3,$3F,$F6,$37,$E8
	.dc.b	$7F,$BE,$E6,$BD,$CA,$FC,$59,$F9,$B9,$D3,$F9,$D0,$E1,$FE,$7D,$B7
	.dc.b	$C0,$5F,$BB,$FC,$ED,$BA,$36,$32,$7C,$D8,$CB,$F3,$FE,$63,$C5,$EC
	.dc.b	$7D,$66,$C6,$4F,$F2,$D1,$E0,$72,$FA,$7C,$47,$8A,$F9,$77,$03,$F2
	.dc.b	$0F,$E9,$91,$86,$C6,$6F,$FA,$64,$E5,$FE,$9D,$97,$2F,$C8,$3F,$AA
	.dc.b	$36,$33,$49,$CB,$63,$F1,$5F,$8A,$7C,$DF,$D4,$FF,$22,$D8,$C8,$8C
	.dc.b	$B4,$FA,$1F,$26,$F8,$DF,$AD,$F6,$1F,$EF,$F8,$C4,$A3,$13,$A8,$E2
	.dc.b	$D7,$FB,$A7,$CA,$73,$C5,$F9,$FF,$F1,$E8,$C3,$6B,$E4,$B6,$77,$5F
	.dc.b	$DC,$78,$7E,$6F,$4F,$F7,$76,$C6,$67,$F3,$6D,$9F,$C5,$2F,$CF,$7E
	.dc.b	$AB,$F9,$03,$F2,$4E,$3A,$76,$D8,$FC,$93,$5F,$C8,$FC,$37,$E6,$7E
	.dc.b	$D3,$F7,$AD,$8F,$DD,$35,$D8,$8F,$F0,$77,$BD,$F8,$6F,$C9,$2E,$46
	.dc.b	$92,$BF,$45,$DF,$FB,$1F,$81,$F5,$B6,$33,$A7,$E4,$27,$2D,$8F,$D9
	.dc.b	$10,$FE,$91,$FB,$A7,$F8,$F3,$0F,$85,$08,$C6,$BF,$21,$B1,$FB,$36
	.dc.b	$BF,$B2,$FC,$47,$F7,$5F,$BE,$FF,$1E,$6C,$7F,$1B,$6B,$B1,$1F,$98
	.dc.b	$FA,$B9,$1F,$9C,$6B,$FA,$DE,$D7,$17,$EB,$3C,$FF,$C0,$66,$69,$2B
	.dc.b	$EB,$C7,$F3,$36,$9F,$CC,$CB,$E4,$FD,$E8,$D4,$7F,$20,$5F,$55,$F2
	.dc.b	$9B,$1F,$39,$AE,$E3,$FA,$47,$A3,$FD,$6A,$3F,$90,$9F,$CE,$4F,$9F
	.dc.b	$E5,$FF,$04,$BF,$98,$D8,$F9,$8D,$77,$7E,$77,$E1,$63,$FC,$33,$F2
	.dc.b	$5F,$82,$48,$FD,$03,$D2,$5F,$77,$49,$3D,$A3,$2F,$E9,$FD,$7F,$96
	.dc.b	$47,$06,$FF,$31,$E8,$47,$07,$E2,$CF,$79,$25,$F8,$67,$D5,$7F,$C7
	.dc.b	$8D,$F3,$BF,$C3,$AD,$9F,$F3,$EB,$5F,$E7,$BE,$BF,$F4,$5F,$CC,$FF
	.dc.b	$B4,$B6,$3F,$B3,$35,$D8,$8F,$ED,$F9,$FF,$4D,$FC,$96,$C9,$A4,$7D
	.dc.b	$47,$F5,$93,$FA,$D6,$2F,$CD,$5F,$86,$7F,$91,$9B,$1F,$E5,$76,$E7
	.dc.b	$FE,$38,$6D,$74,$F6,$7F,$C1,$6F,$C6,$F9,$47,$E2,$4F,$98,$FF,$2A
	.dc.b	$C6,$1B,$19,$C7,$FD,$31,$39,$7F,$CC,$DE,$D7,$F1,$AE,$3F,$F2,$3B
	.dc.b	$70,$FF,$53,$B5,$FF,$52,$35,$FE,$AA,$BF,$15,$FF,$55,$B6,$3A,$C6
	.dc.b	$0F,$F3,$86,$5E,$A8,$F8,$CD,$3D,$B7,$F1,$FE,$6B,$E3,$11,$F8,$CF
	.dc.b	$F9,$CB,$D5,$FE,$CF,$F6,$7F,$EA,$89,$FF,$56,$36,$FF,$80,$35,$FF
	.dc.b	$5B,$F7,$23,$11,$E2,$EC,$FD,$83,$7A,$FF,$D5,$5E,$A7,$DF,$C1,$C7
	.dc.b	$FD,$02,$BF,$C2,$13,$98,$C4,$7F,$8A,$B9,$7A,$CE,$6B,$CA,$72,$2D
	.dc.b	$73,$CF,$D7,$19,$BF,$4B,$5F,$D2,$9E,$59,$FE,$0F,$87,$F8,$EE,$3F
	.dc.b	$83,$1F,$AA,$FB,$8F,$F3,$DF,$C9,$7F,$89,$6B,$D1,$18,$9E,$8D,$8C
	.dc.b	$FC,$22,$B8,$61,$FE,$D7,$F5,$BE,$56,$5F,$E1,$66,$EE,$69,$F3,$38
	.dc.b	$7E,$E6,$FF,$31,$BA,$FF,$BA,$74,$F0,$63,$FE,$91,$6E,$BF,$E3,$B6
	.dc.b	$A6,$BF,$04,$FD,$25,$F5,$25,$7C,$38,$FF,$22,$7C,$9F,$AF,$E5,$3F
	.dc.b	$C8,$DF,$27,$FE,$FE,$8C,$4A,$31,$B2,$B7,$C1,$FE,$A5,$BD,$D6,$A7
	.dc.b	$37,$DA,$F8,$3C,$EF,$84,$62,$B8,$4E,$4F,$F6,$79,$FE,$0C,$D3,$FD
	.dc.b	$F2,$DC,$99,$7F,$84,$23,$FD,$F6,$65,$C4,$F0,$8F,$F7,$71,$93,$FC
	.dc.b	$C1,$9F,$DF,$7E,$EE,$1F,$EB,$FD,$3E,$87,$07,$DF,$9F,$F3,$54,$62
	.dc.b	$75,$1F,$EB,$C3,$A1,$D8,$87,$F6,$DD,$BF,$C4,$73,$F8,$73,$9C,$65
	.dc.b	$AF,$EC,$A3,$6F,$DF,$AF,$13,$E6,$37,$7E,$9A,$F2,$47,$23,$E1,$9E
	.dc.b	$23,$E5,$94,$C9,$E9,$31,$87,$C7,$7F,$5A,$7E,$C7,$7F,$E0,$93,$E6
	.dc.b	$E0,$FD,$9C,$E7,$8F,$CE,$FF,$00,$E3,$AF,$EF,$EB,$C9,$C7,$FC,$1E
	.dc.b	$3F,$81,$B6,$3C,$3B,$FD,$C9,$F9,$67,$94,$F1,$C7,$F0,$67,$C0,$F8
	.dc.b	$F1,$F2,$C7,$F1,$18,$FE,$4C,$7F,$95,$0A,$FF,$49,$E4,$FE,$26,$FF
	.dc.b	$A2,$BA,$D1,$88,$F8,$A7,$C8,$3C,$45,$3F,$9E,$F3,$7E,$46,$9C,$FF
	.dc.b	$CF,$67,$FA,$3C,$E6,$31,$1F,$D2,$27,$FD,$69,$87,$F4,$C6,$C6,$4F
	.dc.b	$F4,$F9,$63,$FD,$6D,$D3,$FD,$56,$2F,$57,$F5,$96,$F5,$FB,$27,$51
	.dc.b	$F3,$39,$BF,$47,$AF,$CD,$8F,$C2,$63,$FD,$C7,$B3,$E1,$B4,$F9,$1B
	.dc.b	$7F,$72,$9F,$EE,$13,$F1,$78,$FF,$0D,$1F,$0C,$F8,$E7,$C8,$A7,$BE
	.dc.b	$EE,$FF,$69,$7C,$C7,$F8,$18,$95,$FE,$F3,$8C,$BF,$E1,$01,$F9,$67
	.dc.b	$F8,$51,$A6,$5C,$1F,$E1,$EF,$27,$D3,$E2,$7F,$88,$E7,$18,$FE,$06
	.dc.b	$31,$1F,$0C,$FC,$69,$E9,$F3,$7F,$CE,$DF,$97,$FF,$8F,$49,$1F,$E3
	.dc.b	$B7,$FC,$C7,$F8,$CC,$8F,$F5,$13,$5F,$F2,$5F,$2F,$F9,$14,$8F,$F9
	.dc.b	$0D,$FF,$35,$F8,$12,$3F,$D5,$4D,$7F,$CA,$7F,$73,$FE,$4D,$23,$FE
	.dc.b	$4B,$7C,$EF,$CE,$E0,$FA,$2F,$D0,$F5,$1E,$9F,$FB,$93,$B3,$FC,$D8
	.dc.b	$49,$FF,$36,$3F,$E7,$35,$74,$8F,$F9,$E9,$FF,$40,$FB,$12,$3F,$D8
	.dc.b	$0D,$7F,$D1,$1D,$DF,$EF,$2E,$EF,$F7,$29,$23,$FD,$D2,$FF,$A5,$D7
	.dc.b	$C7,$D2,$3F,$E9,$93,$27,$F9,$A7,$F1,$8F,$F3,$71,$1C,$1F,$64,$87
	.dc.b	$FA,$2F,$C7,$FF,$45,$11,$FF,$45,$3F,$EC,$0F,$CA,$24,$0F,$F5,$07
	.dc.b	$B4,$E6,$4E,$A7,$FD,$97,$F5,$DF,$ED,$04,$D2,$3F,$EA,$0C,$3E,$61
	.dc.b	$1C,$1F,$EE,$7B,$FF,$55,$91,$C1,$FD,$3F,$E0,$3F,$DF,$1F,$A1,$FF
	.dc.b	$BC,$C9,$19,$F8,$1F,$34,$7F,$BB,$BE,$0D,$1E,$BA,$C3,$6B,$DC,$7B
	.dc.b	$E7,$6A,$76,$D3,$FB,$07,$81,$FE,$17,$5F,$AA,$F2,$7E,$34,$66,$31
	.dc.b	$AF,$EF,$BC,$FF,$99,$18,$9D,$57,$8B,$CD,$75,$B9,$3E,$87,$A5,$D6
	.dc.b	$73,$18,$AE,$15,$BC,$97,$B8,$C3,$7E,$1F,$ED,$6B,$68,$D2,$5F,$BB
	.dc.b	$D9,$FC,$17,$C6,$F9,$DB,$19,$11,$96,$9F,$1B,$2F,$C5,$7E,$C1,$EF
	.dc.b	$6C,$64,$46,$5A,$7C,$BF,$8D,$7C,$77,$E4,$F9,$BE,$74,$A7,$2A,$33
	.dc.b	$18,$6C,$F1,$9F,$D1,$9F,$CC,$FC,$1F,$E0,$BE,$0F,$F0,$4D,$5F,$73
	.dc.b	$FE,$EF,$25,$7F,$BB,$DF,$BD,$A3,$E4,$3E,$84,$E1,$AF,$E3,$23,$5A
	.dc.b	$FF,$6A,$6D,$FF,$4E,$F4,$25,$7F,$18,$CB,$FA,$3C,$FF,$44,$8F,$ED
	.dc.b	$1C,$BF,$E3,$97,$F1,$A6,$0F,$F1,$E7,$C9,$3E,$4B,$FB,$6C,$7F,$18
	.dc.b	$3F,$3B,$19,$F5,$BF,$79,$E2,$FE,$5E,$31,$3B,$3F,$80,$38,$8F,$27
	.dc.b	$A7,$FD,$8E,$7F,$BE,$9F,$E0,$09,$E3,$F7,$7C,$F0,$F0,$B7,$0F,$E1
	.dc.b	$4D,$A3,$F6,$BE,$55,$E8,$7B,$5B,$3F,$E5,$76,$BB,$8F,$85,$AF,$F3
	.dc.b	$C2,$65,$E5,$9F,$ED,$D5,$C1,$FD,$A3,$5F,$DF,$B5,$FE,$3C,$D5,$EE
	.dc.b	$3E,$07,$F9,$07,$3F,$E4,$23,$1C,$1F,$8F,$AF,$E0,$11,$89,$D5,$70
	.dc.b	$9F,$C7,$9F,$D4,$F0,$F3,$3D,$A3,$F9,$67,$BF,$C3,$0F,$F4,$0C,$FF
	.dc.b	$14,$C3,$BC,$E4,$FE,$17,$E5,$3E,$0F,$F9,$A7,$46,$1B,$7F,$E7,$59
	.dc.b	$D6,$1F,$9C,$8C,$4E,$9F,$B7,$AF,$CA,$E5,$FC,$9C,$7F,$9F,$7E,$65
	.dc.b	$F3,$3F,$D0,$F2,$FE,$37,$18,$9E,$87,$FA,$01,$C4,$6A,$78,$4F,$8F
	.dc.b	$8F,$FC,$53,$9F,$F9,$0B,$3F,$F2,$83,$67,$DB,$3F,$E2,$9D,$FE,$3E
	.dc.b	$7E,$59,$FF,$09,$DB,$2F,$E0,$31,$E4,$EC,$FF,$27,$46,$23,$D5,$1F
	.dc.b	$88,$BE,$64,$D3,$DF,$3F,$0A,$30,$FF,$82,$BE,$A9,$F3,$7F,$D8,$FD
	.dc.b	$5F,$E2,$31,$89,$46,$27,$51,$FD,$0E,$75,$3F,$77,$92,$4F,$E2,$39
	.dc.b	$3F,$0C,$FD,$F6,$BF,$C5,$70,$F1,$BF,$E9,$5F,$1F,$FA,$2F,$6F,$F9
	.dc.b	$D1,$BB,$7F,$24,$6C,$FF,$41,$9D,$78,$FF,$EB,$4D,$7F,$A2,$4F,$F7
	.dc.b	$B6,$DF,$EF,$6F,$FC,$D6,$FE,$FC,$F9,$DF,$3B,$F3,$09,$5F,$E8,$7E
	.dc.b	$04,$7F,$B3,$69,$EE,$3E,$7F,$FC,$11,$F8,$E7,$F5,$98,$C6,$28,$C4
	.dc.b	$EB,$1F,$EC,$55,$A9,$FF,$6B,$77,$49,$FE,$DD,$CA,$DF,$FD,$BB,$39
	.dc.b	$9F,$EF,$7F,$23,$FE,$F3,$CB,$F8,$7C,$7F,$75,$8C,$4F,$F4,$A6,$C6
	.dc.b	$7D,$5F,$F0,$F4,$FF,$A2,$BE,$89,$F4,$58,$AF,$F0,$3F,$C2,$27,$38
	.dc.b	$F6,$9F,$F1,$44,$E6,$B8,$46,$2B,$77,$FE,$F1,$F0,$24,$FB,$33,$FF
	.dc.b	$7B,$C6,$23,$D7,$97,$F0,$67,$CD,$97,$F0,$67,$FC,$6D,$9F,$BE,$BD
	.dc.b	$95,$B9,$FF,$0E,$FA,$0F,$A3,$FE,$94,$D8,$FE,$D6,$7F,$82,$23,$26
	.dc.b	$BD,$EE,$D1,$FE,$43,$94,$62,$3F,$C1,$ED,$8C,$DF,$45,$F7,$FF,$E4
	.dc.b	$BF,$8C,$FE,$F3,$18,$B4,$62,$75,$1F,$E2,$B9,$D4,$FF,$80,$38,$7F
	.dc.b	$01,$D0,$FA,$1C,$6E,$8C,$61,$F2,$DA,$FF,$8F,$AB,$8D,$7E,$07,$E9
	.dc.b	$BF,$CB,$CD,$9D,$9F,$32,$32,$FE,$25,$AF,$D4,$E0,$7D,$17,$D0,$6C
	.dc.b	$7D,$68,$C1,$E6,$C3,$31,$E3,$C7,$D1,$39,$24,$77,$36,$78,$CA,$38
	.dc.b	$58,$7F,$9D,$E7,$D4,$7F,$90,$7C,$0F,$F3,$C7,$E8,$97,$23,$49,$87
	.dc.b	$F9,$8A,$3F,$DE,$BF,$94,$7D,$28,$FF,$41,$46,$1F,$AE,$6F,$C6,$4E
	.dc.b	$E3,$FD,$0D,$18,$76,$72,$5F,$89,$7C,$A8,$DB,$69,$3A,$9F,$A7,$18
	.dc.b	$9E,$1D,$DF,$E6,$6F,$CC,$BF,$D2,$9C,$5F,$E5,$F6,$DF,$F9,$91,$AF
	.dc.b	$FA,$6E,$78,$5F,$92,$3F,$D3,$DC,$5F,$8C,$BF,$AC,$48,$FC,$53,$5F
	.dc.b	$B1,$C3,$C5,$4F,$17,$67,$F8,$C3,$F0,$AF,$D7,$79,$7F,$E2,$F6,$C6
	.dc.b	$7A,$7F,$D5,$D3,$96,$C7,$23,$BE,$1E,$17,$CF,$C9,$F9,$F8,$FF,$31
	.dc.b	$FD,$83,$E4,$7F,$D7,$1F,$85,$E1,$B7,$FE,$95,$9D,$29,$D3,$4F,$B9
	.dc.b	$5F,$6F,$5F,$A6,$DE,$1F,$C0,$BF,$E9,$C6,$CF,$9C,$F2,$ED,$F1,$76
	.dc.b	$F8,$91,$EA,$3F,$1D,$F8,$EF,$FA,$23,$0F,$A4,$1C,$61,$FE,$A8,$6C
	.dc.b	$F0,$F2,$7E,$DB,$F8,$EF,$C2,$FF,$DA,$DF,$59,$E3,$6C,$7E,$55,$AF
	.dc.b	$E4,$5A,$F9,$AF,$E6,$3F,$6B,$7F,$98,$FA,$F1,$F2,$7B,$7F,$EE,$1C
	.dc.b	$3C,$8D,$9F,$5B,$5F,$D3,$E5,$FE,$B3,$9E,$B9,$FC,$29,$F0,$1F,$1C
	.dc.b	$F0,$69,$F1,$F1,$F0,$C7,$C7,$21,$C0,$A7,$0C,$BF,$09,$97,$EA,$38
	.dc.b	$3E,$B1,$27,$E6,$ED,$F8,$0A,$F1,$DF,$69,$E8,$4B,$E4,$F0,$BE,$55
	.dc.b	$F7,$DF,$DF,$43,$EC,$3F,$C0,$2F,$F2,$83,$1E,$5F,$EF,$5A,$66,$37
	.dc.b	$DF,$E2,$CF,$E0,$13,$0F,$F5,$EC,$F4,$FA,$3B,$83,$04,$61,$A7,$FB
	.dc.b	$C6,$78,$3F,$C5,$7E,$43,$F3,$04,$7E,$1B,$EA,$DB,$EA,$47,$CB,$3E
	.dc.b	$A3,$F5,$FA,$7F,$DE,$7F,$1F,$F9,$06,$ED,$FC,$25,$B9,$FF,$BB,$E7
	.dc.b	$5E,$83,$EB,$96,$FA,$DF,$F7,$A4,$61,$B1,$9C,$BF,$DF,$73,$97,$C7
	.dc.b	$C5,$F5,$DF,$31,$F8,$2C,$FD,$24,$8D,$3F,$8F,$AF,$E5,$7E,$D7,$C9
	.dc.b	$FD,$5B,$F7,$0D,$BF,$E7,$4D,$7F,$9B,$35,$6A,$CD,$CD,$EC,$F8,$3F
	.dc.b	$A9,$5F,$CB,$67,$38,$FF,$B9,$63,$37,$C6,$F8,$46,$2B,$77,$CA,$3B
	.dc.b	$C3,$E9,$BE,$4F,$E4,$59,$C3,$C5,$18,$AE,$4D,$7C,$4D,$79,$FE,$49
	.dc.b	$F4,$A3,$95,$A3,$13,$E3,$6B,$EB,$6B,$F6,$BF,$07,$FF,$55,$63,$E7
	.dc.b	$CD,$18,$9F,$ED,$4D,$7F,$B4,$B5,$FD,$79,$F6,$8B,$B7,$D7,$FD,$CF
	.dc.b	$FB,$DD,$7F,$B4,$67,$FB,$A4,$63,$DB,$F0,$36,$33,$E6,$FE,$3A,$7F
	.dc.b	$CD,$DF,$7D,$F0,$76,$FF,$39,$B7,$DA,$34,$F8,$D7,$E3,$26,$5F,$8B
	.dc.b	$ED,$FE,$6B,$A3,$C6,$FD,$97,$E2,$3F,$06,$31,$68,$C4,$EA,$3E,$1B
	.dc.b	$5F,$87,$F3,$5F,$BB,$FD,$1B,$93,$63,$E4,$B5,$D8,$8F,$D3,$F8,$F2
	.dc.b	$3F,$4F,$C2,$8C,$47,$E9,$7A,$BF,$46,$4C,$FF,$87,$47,$EB,$1A,$FE
	.dc.b	$22,$BF,$C2,$2F,$E3,$1A,$E7,$C5,$FC,$4F,$3F,$F8,$7A,$3E,$FF,$E1
	.dc.b	$9F,$C5,$7F,$4E,$F8,$8D,$8C,$88,$CB,$4F,$F4,$FF,$0F,$37,$FC,$2D
	.dc.b	$C5,$F5,$B1,$F0,$C7,$EC,$A3,$18,$FD,$E9,$CC,$FD,$8C,$FD,$DE,$0F
	.dc.b	$F8,$1F,$93,$F7,$1D,$FF,$C4,$BC,$7F,$E1,$E4,$BF,$92,$FF,$33,$A9
	.dc.b	$1F,$15,$AF,$93,$3F,$EB,$CF,$8D,$B7,$F8,$EC,$7F,$82,$CF,$B1,$AF
	.dc.b	$D3,$DF,$E7,$BF,$D4,$27,$87,$AB,$EB,$24,$71,$7C,$2F,$E9,$B3,$F6
	.dc.b	$BE,$52,$1D,$71,$F8,$1D,$AF,$C0,$7F,$98,$FD,$3F,$F0,$06,$D7,$F1
	.dc.b	$16,$BF,$C6,$5A,$FE,$FF,$C9,$FE,$A1,$F5,$BF,$9F,$6C,$7E,$F9,$AF
	.dc.b	$EF,$9A,$F9,$F4,$F7,$BF,$55,$BD,$9F,$AD,$CB,$DB,$39,$6C,$66,$3F
	.dc.b	$85,$FE,$0B,$FD,$87,$F1,$5F,$D8,$36,$32,$23,$2D,$3F,$89,$FC,$87
	.dc.b	$F8,$EF,$BC,$D2,$47,$93,$7F,$9A,$FC,$B3,$E6,$9F,$5F,$77,$D8,$7F
	.dc.b	$5C,$F2,$3F,$AE,$CF,$F4,$18,$C7,$27,$F3,$D6,$C6,$61,$3C,$23,$E3
	.dc.b	$6B,$F0,$30,$F0,$C6,$0E,$55,$E0,$2D,$F1,$E4,$FE,$A8,$FF,$6E,$31
	.dc.b	$8E,$DE,$D5,$F3,$4F,$5D,$3F,$11,$7C,$7F,$F8,$36,$BC,$93,$98,$C3
	.dc.b	$FE,$A5,$9D,$D7,$8C,$FF,$52,$B8,$AE,$3A,$BD,$7F,$F7,$05,$FF,$44
	.dc.b	$6D,$FF,$44,$6B,$FD,$35,$B9,$7F,$6F,$FC,$07,$FD,$3B,$5F,$D9,$DB
	.dc.b	$5F,$DB,$DA,$FF,$68,$6B,$EF,$3E,$D1,$DC,$FF,$77,$C7,$8D,$A6,$75
	.dc.b	$F4,$DF,$92,$39,$1F,$4C,$E6,$53,$FC,$69,$F0,$1F,$F0,$01,$7C,$27
	.dc.b	$FB,$02,$30,$DF,$84,$7C,$27,$F8,$EE,$30,$DC,$3F,$C1,$75,$C1,$B5
	.dc.b	$C2,$FA,$C6,$2B,$84,$78,$A7,$57,$FE,$D0,$E6,$71,$39,$FF,$C1,$D7
	.dc.b	$FE,$96,$3F,$DE,$84,$3F,$C2,$53,$C2,$37,$1F,$E7,$63,$47,$FB,$08
	.dc.b	$FE,$C5,$9F,$FB,$A5,$F9,$C7,$CA,$3F,$DD,$26,$34,$EB,$7E,$38,$E6
	.dc.b	$78,$CC,$D7,$9B,$6F,$C1,$57,$29,$FB,$CD,$8C,$C7,$E8,$36,$FD,$9C
	.dc.b	$FD,$D8,$F1,$11,$C1,$F9,$BE,$2F,$F3,$67,$8F,$FA,$0E,$3F,$CE,$B6
	.dc.b	$33,$3B,$A4,$FE,$B3,$4F,$8D,$1E,$A3,$E3,$1E,$A3,$FC,$5F,$C7,$FE
	.dc.b	$62,$9F,$DA,$B6,$BF,$7B,$18,$7A,$9C,$66,$1F,$3C,$A7,$C7,$30,$43
	.dc.b	$07,$F9,$2A,$7F,$CE,$3F,$05,$D4,$93,$FC,$31,$F5,$6B,$F4,$EB,$A9
	.dc.b	$EA,$3A,$14,$EA,$53,$13,$FE,$42,$E7,$FC,$3B,$63,$F1,$ED,$7F,$C5
	.dc.b	$11,$AE,$57,$4F,$FA,$77,$E5,$3F,$9C,$92,$BF,$9E,$BF,$37,$4F,$9F
	.dc.b	$1E,$66,$BE,$66,$BE,$66,$BE,$76,$93,$1F,$81,$1C,$5A,$F1,$6E,$3F
	.dc.b	$E5,$EE,$BF,$F5,$97,$CA,$FF,$63,$A3,$48,$FF,$6D,$97,$8B,$FD,$C5
	.dc.b	$B1,$98,$C1,$ED,$BF,$0C,$79,$DB,$87,$9F,$93,$F6,$6D,$9E,$11,$89
	.dc.b	$D9,$EA,$E4,$FF,$0A,$73,$7F,$89,$C9,$5F,$E1,$57,$8B,$0F,$F0,$C7
	.dc.b	$1F,$F8,$B4,$8E,$0F,$A1,$5D,$AF,$C0,$48,$E4,$7E,$4E,$FF,$C6,$39
	.dc.b	$3E,$6B,$FD,$7E,$49,$FF,$2C,$BC,$1F,$A1,$5F,$C0,$73,$FF,$21,$7D
	.dc.b	$77,$F8,$FC,$93,$FE,$2D,$7F,$C9,$7F,$23,$23,$FC,$9E,$D7,$FC,$B9
	.dc.b	$1F,$E5,$8C,$0D,$24,$7B,$2F,$D4,$D3,$C8,$7E,$87,$1F,$5F,$CE,$FF
	.dc.b	$B6,$09,$1F,$E7,$27,$FC,$F6,$BF,$47,$A4,$7F,$CF,$A6,$4F,$F4,$28
	.dc.b	$FC,$3B,$FD,$12,$D3,$2E,$0F,$1F,$07,$E0,$63,$FC,$9D,$F0,$5F,$E6
	.dc.b	$32,$38,$3F,$D4,$B9,$BC,$EF,$F5,$4F,$AB,$FE,$F4,$6C,$66,$77,$1D
	.dc.b	$B8,$7E,$F4,$7F,$A7,$F8,$8B,$F5,$DC,$34,$FF,$60,$C6,$A3,$27,$FB
	.dc.b	$43,$D9,$60,$FA,$FC,$36,$FF,$DA,$4D,$7F,$DA,$4D,$7F,$DC,$BE,$9F
	.dc.b	$2D,$3E,$E1,$23,$FD,$30,$FF,$B7,$BD,$CE,$F8,$DB,$F3,$57,$FB,$DE
	.dc.b	$72,$FF,$BC,$DA,$FB,$A1,$E9,$FF,$8E,$21,$F8,$3F,$FB,$86,$30,$DF
	.dc.b	$33,$E2,$B6,$76,$7E,$2C,$87,$F0,$FE,$6F,$C7,$BF,$A7,$F3,$1E,$A7
	.dc.b	$C6,$F8,$6F,$F7,$BB,$67,$FD,$7F,$18,$C8,$FF,$BF,$3E,$07,$FC,$7F
	.dc.b	$2E,$A6,$09,$CC,$61,$B1,$9B,$F1,$D7,$EC,$8F,$EE,$AF,$C2,$F8,$2F
	.dc.b	$E7,$BA,$9A,$49,$FF,$03,$FC,$1B,$C3,$FC,$3F,$CB,$FF,$84,$23,$12
	.dc.b	$8C,$4E,$A3,$F1,$A7,$E3,$A7,$73,$F9,$3D,$BC,$B2,$9E,$7E,$87,$28
	.dc.b	$FF,$13,$1C,$A3,$2F,$43,$85,$FE,$6A,$7E,$D7,$B3,$EC,$24,$FF,$58
	.dc.b	$C9,$1F,$E2,$28,$FE,$69,$87,$8B,$B7,$FA,$7E,$9F,$86,$AE,$35,$D0
	.dc.b	$E4,$F1,$9F,$DC,$47,$E0,$4F,$61,$0F,$C2,$9F,$A6,$F8,$77,$3F,$EA
	.dc.b	$3A,$FD,$58,$27,$2F,$F9,$D2,$BF,$31,$18,$9D,$B6,$3A,$9E,$A3,$F3
	.dc.b	$5E,$E7,$EF,$26,$C7,$33,$F6,$F9,$A7,$51,$F5,$AF,$F4,$B1,$88,$F0
	.dc.b	$F2,$7E,$4A,$7C,$CD,$C3,$F2,$F1,$FE,$04,$6C,$FC,$13,$FC,$35,$1F
	.dc.b	$B0,$F5,$3F,$1E,$48,$FA,$8F,$D3,$F7,$BE,$AC,$7E,$81,$B3,$FE,$5E
	.dc.b	$6C,$7E,$05,$AF,$D1,$3E,$7F,$96,$F8,$A7,$C5,$7F,$1E,$F8,$CF,$C4
	.dc.b	$4E,$63,$15,$F9,$68,$C2,$9D,$D7,$28,$C1,$F4,$38,$3F,$02,$FE,$B4
	.dc.b	$FB,$FD,$5F,$B6,$F8,$CF,$D9,$93,$3F,$F3,$17,$52,$3E,$67,$0F,$4E
	.dc.b	$6F,$E2,$59,$FE,$6F,$1F,$A9,$39,$6C,$65,$FE,$39,$0F,$8D,$FD,$D9
	.dc.b	$23,$F4,$0F,$FA,$03,$2F,$E1,$3E,$A7,$F3,$8F,$D2,$3F,$55,$18,$D1
	.dc.b	$18,$D1,$5B,$C3,$F7,$77,$BA,$FF,$48,$FC,$0B,$CA,$FE,$97,$F2,$58
	.dc.b	$6D,$FE,$DE,$74,$A7,$4D,$3F,$D0,$EF,$FA,$73,$83,$FD,$30,$DF,$2B
	.dc.b	$FD,$40,$DE,$DE,$15,$FC,$3A,$FF,$D6,$FA,$E2,$3F,$AE,$C6,$2B,$CF
	.dc.b	$E9,$FF,$58,$6D,$6E,$95,$E6,$E2,$F2,$92,$3D,$59,$FD,$17,$FB,$4E
	.dc.b	$7F,$EA,$D9,$F2,$12,$3F,$D4,$2F,$FA,$DB,$DE,$7C,$97,$F1,$16,$CF
	.dc.b	$FA,$D1,$AE,$E3,$5E,$FF,$F2,$28,$FF,$63,$1C,$DC,$3F,$EC,$7A,$E5
	.dc.b	$1F,$EC,$8E,$BD,$EB,$FE,$D4,$AF,$F6,$79,$B2,$8C,$3F,$DA,$CE,$63
	.dc.b	$81,$27,$99,$23,$4D,$79,$57,$85,$C9,$0C,$DF,$FB,$79,$A7,$8A,$FC
	.dc.b	$2E,$2F,$C0,$F1,$8C,$92,$BB,$46,$EF,$C5,$39,$BE,$CD,$9E,$D5,$C2
	.dc.b	$31,$7D,$A7,$91,$DE,$74,$4A,$E6,$D8,$E7,$18,$23,$82,$4F,$5A,$EF
	.dc.b	$7E,$6B,$E5,$39,$6C,$66,$F9,$CE,$5B,$19,$BE,$AD,$9D,$46,$0F,$54
	.dc.b	$F0,$69,$93,$DE,$4B,$F2,$BA,$9E,$64,$AE,$C4,$8D,$35,$E5,$7E,$A7
	.dc.b	$24,$33,$7E,$36,$9E,$3B,$F0,$B8,$BF,$03,$C6,$32,$4A,$ED,$1B,$BF
	.dc.b	$14,$E6,$FB,$36,$7B,$57,$08,$C5,$F6,$9E,$47,$79,$D1,$2B,$9B,$63
	.dc.b	$9C,$60,$8E,$09,$3D,$6B,$BD,$F9,$AF,$94,$E5,$B1,$9B,$E7,$39,$6C
	.dc.b	$66,$FA,$B6,$75,$18,$3D,$53,$C1,$A6,$4F,$79,$2F,$CA,$EA,$79,$92
	.dc.b	$BB,$12,$34,$D7,$95,$FA,$9C,$90,$CD,$F8,$DA,$78,$EF,$C2,$E2,$FC
	.dc.b	$0F,$18,$C9,$2B,$B4,$6E,$FC,$53,$9B,$EC,$D9,$ED,$5C,$23,$17,$DA
	.dc.b	$79,$1D,$E7,$44,$AE,$6D,$8E,$71,$82,$38,$24,$F5,$AE,$F7,$E6,$BE
	.dc.b	$53,$96,$C6,$6F,$9C,$E5,$B1,$9B,$EA,$D9,$D4,$60,$F5,$4F,$06,$99
	.dc.b	$3D,$E4,$BF,$2B,$A9,$E6,$4A,$EC,$48,$D3,$5E,$57,$EA,$72,$43,$37
	.dc.b	$E3,$69,$E3,$BF,$0B,$8B,$F0,$3C,$63,$24,$AE,$D1,$BB,$F1,$4E,$6F
	.dc.b	$B3,$67,$B5,$70,$8C,$5F,$69,$E4,$77,$9D,$12,$B9,$B6,$39,$C6,$08
	.dc.b	$E0,$93,$D6,$BB,$DF,$9A,$F9,$4E,$5B,$19,$BE,$73,$96,$C6,$6F,$AB
	.dc.b	$67,$51,$83,$D5,$3C,$1A,$64,$F7,$92,$FC,$AF,$E1,$8F,$F5,$72,$BF
	.dc.b	$F0,$7C,$7F,$78,$32,$E2,$35,$3F,$E1,$22,$4F,$F8,$5D,$AF,$06,$BF
	.dc.b	$E1,$C3,$06,$C8,$60,$86,$C8,$7F,$86,$CC,$94,$C9,$E7,$2C,$79,$9E
	.dc.b	$A6,$27,$D4,$49,$F2,$C7,$32,$1C,$CB,$3C,$DF,$09,$92,$39,$7B,$47
	.dc.b	$47,$33,$DD,$F1,$9D,$1C,$3D,$27,$93,$99,$E5,$18,$8E,$71,$89,$E3
	.dc.b	$1E,$03,$A9,$A2,$CF,$12,$47,$08,$EC,$69,$C3,$C4,$F2,$38,$8E,$33
	.dc.b	$E1,$24,$F1,$6B,$C1,$AF,$94,$C1,$B2,$18,$21,$B2,$1E,$63,$25,$32
	.dc.b	$79,$CB,$1E,$67,$A9,$89,$F5,$12,$7C,$B1,$CC,$87,$32,$CF,$37,$C2
	.dc.b	$64,$8E,$5E,$D1,$D1,$CC,$F7,$7C,$67,$47,$0F,$49,$E4,$E6,$79,$46
	.dc.b	$23,$9C,$62,$78,$C7,$80,$EA,$68,$B3,$C4,$91,$C2,$3B,$1A,$70,$F1
	.dc.b	$3C,$8E,$23,$8C,$F8,$49,$3C,$5A,$F0,$6B,$E5,$30,$6C,$86,$08,$6C
	.dc.b	$87,$98,$C9,$4C,$9E,$72,$C7,$99,$EA,$62,$7D,$44,$9F,$2C,$73,$21
	.dc.b	$CC,$B3,$CD,$F0,$99,$23,$97,$B4,$74,$73,$3D,$DF,$19,$D1,$C3,$D2
	.dc.b	$79,$39,$9E,$51,$88,$E7,$18,$9E,$31,$E0,$3A,$9A,$2C,$F1,$24,$70
	.dc.b	$8E,$C6,$9C,$3C,$4F,$23,$88,$E3,$3E,$12,$4F,$16,$BC,$1A,$F9,$4C
	.dc.b	$1B,$21,$82,$1B,$21,$E6,$32,$53,$27,$9C,$B1,$E6,$7A,$98,$9F,$51
	.dc.b	$27,$CB,$1C,$C8,$73,$2C,$F3,$7C,$26,$48,$E5,$ED,$1D,$1C,$CF,$77
	.dc.b	$C6,$74,$70,$F4,$9E,$4E,$67,$94,$62,$39,$C6,$27,$8C,$78,$0E,$A6
	.dc.b	$8B,$3C,$49,$1C,$23,$B1,$A7,$0F,$13,$C8,$A5,$10,$7C,$74,$00

;--------------------------------------------------------------------------------
;	FSABS.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsabs_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FSABS.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSABS.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsabs_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsabs.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSABS.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsabs.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSABS.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsabs_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$E6,$3A,$29,$F7,$3F,$80,$9C,$C6,$1C,$4E,$E7,$51
	.dc.b	$E6,$9D,$3F,$88,$9D,$3C,$23,$2F,$96,$78,$CE,$A3,$84,$E9,$F0,$CE
	.dc.b	$A3,$11,$DE,$30,$7F,$4A,$FA,$EF,$F3,$4F,$C1,$7E,$A1,$B9,$FF,$3A
	.dc.b	$7F,$CD,$B8,$79,$1B,$87,$B5,$DE,$1C,$FD,$2F,$DD,$5F,$17,$D1,$09
	.dc.b	$E3,$8E,$E3,$F7,$95,$B6,$99,$3D,$07,$1A,$E7,$5E,$08,$C1,$C0,$A7
	.dc.b	$07,$BB,$8C,$7D,$D3,$E6,$8E,$D8,$FB,$63,$A9,$98,$E1,$7E,$57,$A4
	.dc.b	$7B,$3E,$1B,$C4,$47,$04,$8F,$24,$65,$FA,$A7,$F6,$38,$D3,$FC,$7E
	.dc.b	$7F,$B4,$CE,$E3,$E0,$BE,$AC,$3E,$01,$2B,$CF,$3D,$23,$0F,$C3,$78
	.dc.b	$46,$A3,$B1,$F3,$70,$F8,$6F,$18,$D4,$62,$36,$7C,$CC,$3E,$33,$B3
	.dc.b	$B4,$FC,$7A,$EF,$19,$38,$BE,$47,$E4,$99,$AE,$C7,$7A,$EB,$1A,$38
	.dc.b	$D7,$53,$D5,$1A,$9E,$38,$7D,$58,$E2,$65,$FB,$11,$D7,$2F,$CC,$57
	.dc.b	$BA,$72,$71,$8C,$DF,$96,$32,$FE,$1A,$32,$F9,$6B,$C2,$E7,$28,$FE
	.dc.b	$20,$F9,$D5,$D6,$33,$8F,$C3,$AF,$24,$63,$0F,$75,$FC,$C3,$E5,$63
	.dc.b	$EB,$79,$9B,$3C,$C6,$0F,$AD,$8F,$9A,$38,$47,$8C,$CD,$71,$7E,$69
	.dc.b	$DC,$FA,$11,$C3,$0F,$B0,$FC,$CA,$F8,$38,$7C,$0A,$E6,$4A,$F6,$63
	.dc.b	$EF,$8C,$63,$EC,$7B,$9D,$AF,$9C,$FC,$E8,$D9,$EC,$3D,$15,$F8,$C7
	.dc.b	$E8,$46,$E3,$18,$79,$EF,$B3,$E7,$C7,$E2,$4F,$AB,$0F,$88,$78,$30
	.dc.b	$CD,$FE,$32,$31,$1E,$08,$C5,$6A,$3E,$35,$F5,$74,$62,$BA,$CE,$E3
	.dc.b	$DF,$3B,$BF,$8B,$1F,$46,$3D,$0F,$19,$FA,$98,$FC,$DA,$F6,$DF,$91
	.dc.b	$F5,$4E,$B1,$F8,$58,$F9,$B0,$F1,$9E,$0B,$F1,$B8,$C3,$F1,$EF,$CE
	.dc.b	$3E,$9B,$8C,$3E,$73,$E1,$AE,$91,$E8,$9E,$18,$FA,$1E,$D8,$F8,$DF
	.dc.b	$81,$1E,$FA,$FC,$05,$F5,$7F,$61,$3A,$C3,$CB,$87,$53,$DF,$87,$C9
	.dc.b	$BE,$15,$F8,$39,$F8,$2F,$A6,$3C,$CF,$E9,$AB,$E3,$5F,$E0,$9E,$AF
	.dc.b	$8B,$1F,$CA,$5F,$B9,$F9,$35,$C7,$0F,$04,$6A,$BC,$F3,$F3,$31,$F6
	.dc.b	$3F,$A7,$C7,$CA,$FE,$50,$F6,$5F,$C3,$78,$63,$E2,$78,$61,$D5,$F6
	.dc.b	$61,$E7,$AE,$18,$FB,$8F,$34,$F4,$3E,$05,$7A,$1F,$D5,$4F,$A1,$FC
	.dc.b	$65,$FC,$27,$F1,$87,$03,$95,$74,$7C,$B3,$E0,$BF,$14,$6C,$C9,$67
	.dc.b	$C9,$1F,$11,$B1,$EA,$3B,$BD,$4C,$BF,$14,$E6,$E6,$3B,$B6,$38,$C7
	.dc.b	$53,$25,$72,$77,$3F,$22,$79,$5D,$B9,$C3,$EA,$E3,$F9,$76,$E1,$F1
	.dc.b	$67,$8D,$F8,$5B,$7D,$DA,$73,$C7,$EC,$3E,$B9,$F6,$E1,$D6,$72,$D9
	.dc.b	$FA,$78,$FD,$57,$F5,$38,$78,$71,$F1,$CE,$70,$FD,$0D,$7E,$DF,$0E
	.dc.b	$77,$B6,$D6,$F1,$FD,$A9,$F9,$DA,$EE,$71,$2D,$FE,$B6,$37,$1F,$92
	.dc.b	$32,$F1,$7C,$B3,$D5,$FD,$21,$CC,$C9,$67,$CE,$75,$76,$D7,$F2,$A6
	.dc.b	$E7,$C2,$FF,$10,$AF,$5B,$93,$8C,$78,$E7,$AB,$4E,$38,$7F,$00,$9E
	.dc.b	$B3,$DF,$1F,$DA,$C7,$53,$DA,$7B,$CF,$4C,$6F,$1F,$DA,$46,$CC,$DF
	.dc.b	$77,$C6,$71,$9F,$1B,$63,$C7,$87,$EC,$EF,$C3,$87,$95,$B1,$9C,$3B
	.dc.b	$CE,$5B,$19,$C7,$F6,$58,$7E,$78,$FC,$76,$1F,$9C,$69,$9C,$3F,$1C
	.dc.b	$FA,$0F,$21,$E8,$3D,$98,$FA,$1B,$3E,$87,$F8,$1B,$F9,$FC,$7F,$27
	.dc.b	$1F,$4C,$F9,$B1,$F0,$DA,$76,$29,$D0,$F0,$47,$B6,$FD,$71,$FA,$A3
	.dc.b	$FC,$B7,$3F,$E6,$DC,$BF,$CB,$D8,$7F,$A8,$1F,$F3,$B9,$FE,$61,$BF
	.dc.b	$F7,$DB,$89,$FF,$33,$C6,$1E,$47,$F9,$AA,$7E,$1E,$1E,$F9,$F1,$5F
	.dc.b	$F0,$53,$E4,$65,$FA,$B7,$F0,$A7,$A4,$F5,$C7,$09,$FE,$08,$FC,$47
	.dc.b	$F1,$18,$FB,$23,$18,$7C,$5C,$3E,$FD,$FC,$C3,$F8,$41,$DA,$7A,$4F
	.dc.b	$CF,$CB,$F4,$38,$74,$AF,$31,$F1,$EF,$D7,$5F,$5E,$3F,$94,$BC,$5E
	.dc.b	$95,$F8,$97,$8D,$F9,$1F,$AD,$5E,$47,$89,$E4,$3F,$10,$7C,$17,$57
	.dc.b	$FC,$73,$1F,$8D,$3A,$AF,$4B,$C6,$7E,$75,$F8,$5F,$8E,$FB,$30,$F8
	.dc.b	$97,$F9,$EA,$F8,$87,$E6,$8F,$73,$DA,$B9,$C6,$AD,$68,$98,$BE,$3B
	.dc.b	$FD,$6C,$D8,$FF,$61,$35,$FE,$A0,$DA,$E7,$1C,$A7,$89,$FD,$FA,$75
	.dc.b	$5D,$E7,$FD,$FB,$3A,$7F,$CA,$93,$A7,$91,$E0,$8E,$46,$DF,$F5,$4C
	.dc.b	$78,$F2,$FF,$39,$BC,$63,$A9,$C4,$FF,$30,$1E,$13,$FD,$62,$5B,$F4
	.dc.b	$B9,$32,$7F,$B5,$CD,$5E,$5F,$F7,$21,$83,$11,$C9,$C1,$CD,$C3,$CC
	.dc.b	$F8,$4B,$7E,$90,$8C,$3D,$23,$31,$88,$E9,$18,$23,$B2,$1A,$29,$EF
	.dc.b	$79,$F0,$48,$F4,$B5,$F4,$35,$F8,$CD,$7D,$2D,$7F,$15,$28,$C4,$7A
	.dc.b	$F8,$7D,$24,$7E,$3B,$5F,$8C,$D7,$EA,$35,$F8,$ED,$7B,$BF,$13,$8A
	.dc.b	$47,$D2,$6B,$F4,$5A,$FD,$F6,$BF,$49,$AF,$A9,$AB,$27,$B5,$F9,$9C
	.dc.b	$FF,$45,$F7,$7F,$8B,$6D,$67,$5F,$CF,$5F,$08,$FC,$D9,$CE,$F4,$FE
	.dc.b	$C6,$F9,$57,$E9,$5B,$1F,$B5,$AE,$4F,$E7,$0E,$47,$48,$DA,$6C,$70
	.dc.b	$BF,$D3,$5F,$82,$B4,$5A,$DC,$66,$B8,$5F,$8E,$B4,$66,$63,$E1,$32
	.dc.b	$7B,$A7,$8B,$FC,$37,$E1,$DC,$FE,$16,$C7,$C4,$6B,$ED,$C1,$ED,$6E
	.dc.b	$A4,$ED,$F4,$46,$23,$E2,$57,$2C,$BF,$8B,$37,$2F,$99,$5B,$AF,$E0
	.dc.b	$53,$E0,$6D,$FE,$CA,$31,$1F,$C5,$DA,$FA,$E3,$44,$7F,$82,$46,$93
	.dc.b	$49,$1B,$8F,$77,$B3,$F4,$4F,$56,$1E,$89,$C4,$FD,$D9,$CE,$1E,$39
	.dc.b	$CC,$F2,$3E,$21,$0C,$12,$BA,$39,$7F,$33,$5C,$9C,$C7,$CA,$9C,$57
	.dc.b	$E4,$EB,$C3,$3F,$6F,$BB,$ED,$11,$FC,$97,$B3,$5F,$A7,$17,$0F,$D3
	.dc.b	$CD,$BF,$99,$36,$BF,$45,$35,$F9,$31,$73,$4E,$72,$4E,$63,$0D,$AF
	.dc.b	$C6,$3F,$6C,$8F,$D6,$6B,$FC,$FE,$79,$B9,$21,$98,$DB,$F8,$86,$BE
	.dc.b	$A3,$F6,$5E,$C7,$EA,$CF,$1B,$82,$1F,$10,$F2,$98,$C7,$F8,$5C,$70
	.dc.b	$8F,$D1,$3C,$DD,$9C,$4C,$9C,$71,$F1,$39,$9E,$6E,$4F,$B0,$72,$72
	.dc.b	$F2,$38,$47,$28,$C9,$1C,$12,$3F,$8C,$4F,$95,$E2,$7E,$FF,$C4,$FD
	.dc.b	$51,$1F,$D2,$B5,$FA,$0D,$7F,$66,$D7,$F6,$0D,$7A,$1F,$AE,$F1,$7F
	.dc.b	$6C,$47,$04,$3E,$E3,$96,$99,$3F,$80,$B9,$69,$FB,$AF,$1B,$F8,$11
	.dc.b	$1C,$10,$FC,$43,$96,$99,$3F,$86,$B9,$69,$FC,$1B,$C7,$FE,$1C,$47
	.dc.b	$04,$3F,$28,$E5,$A6,$4F,$E3,$2E,$5A,$7F,$12,$F2,$3F,$8D,$11,$C1
	.dc.b	$0F,$D0,$39,$69,$93,$F9,$2B,$96,$9F,$C7,$BC,$9F,$E4,$C4,$70,$43
	.dc.b	$F5,$8E,$5A,$64,$FE,$62,$E5,$A7,$F2,$AF,$2B,$F9,$91,$1C,$10,$FD
	.dc.b	$C3,$96,$99,$3F,$9E,$B9,$69,$FC,$DB,$CB,$FE,$7C,$47,$04,$3F,$82
	.dc.b	$39,$69,$93,$FA,$4B,$96,$9F,$D0,$BE,$1F,$FA,$51,$1C,$10,$FE,$20
	.dc.b	$E5,$A6,$4F,$EA,$AE,$5A,$7F,$94,$7E,$21,$ED,$7F,$7B,$6C,$7F,$83
	.dc.b	$9A,$FF,$6B,$F6,$B5,$32,$34,$FF,$80,$63,$11,$FE,$0D,$9E,$3C,$3F
	.dc.b	$E7,$F6,$F0,$FF,$85,$6B,$73,$9B,$B3,$9B,$EA,$43,$A4,$77,$71,$3C
	.dc.b	$5B,$3C,$5D,$98,$3F,$D2,$D7,$E2,$7F,$C5,$17,$E0,$73,$75,$FF,$11
	.dc.b	$45,$9E,$10,$6F,$2A,$75,$5C,$63,$15,$E1,$38,$46,$67,$71,$99,$DB
	.dc.b	$E8,$26,$1E,$D7,$E9,$1C,$EF,$C6,$78,$AB,$9C,$70,$C3,$E1,$57,$12
	.dc.b	$3F,$EA,$9A,$E1,$3C,$F0,$F2,$34,$F2,$5F,$4B,$E3,$1F,$20,$D1,$DC
	.dc.b	$C1,$4F,$79,$C8,$DB,$C9,$F9,$A6,$2F,$C9,$5F,$76,$39,$5F,$BA,$31
	.dc.b	$5C,$63,$13,$FE,$59,$BF,$35,$F2,$8C,$1D,$4F,$BA,$60,$EB,$7F,$01
	.dc.b	$A7,$9C,$A7,$AF,$0F,$AB,$87,$86,$78,$36,$78,$35,$E1,$7E,$47,$C2
	.dc.b	$7D,$A3,$AC,$7C,$3A,$EB,$7E,$99,$CB,$F6,$8C,$4F,$53,$FD,$13,$5D
	.dc.b	$67,$E7,$9D,$63,$53,$C0,$E5,$1E,$23,$FD,$B5,$87,$EF,$9B,$87,$FB
	.dc.b	$2A,$31,$1E,$33,$27,$F0,$98,$D1,$1C,$10,$FF,$71,$47,$27,$E0,$D7
	.dc.b	$FA,$83,$D9,$FF,$44,$1E,$9C,$4E,$27,$EC,$1F,$E9,$37,$0F,$C4,$3F
	.dc.b	$D4,$0E,$27,$DF,$18,$3F,$D3,$91,$D4,$D6,$1F,$0E,$3C,$6F,$FA,$9E
	.dc.b	$33,$18,$21,$E1,$39,$9E,$4A,$F3,$B8,$3C,$86,$CF,$F4,$94,$78,$5D
	.dc.b	$57,$99,$D3,$F0,$DE,$91,$E2,$7E,$DC,$7E,$16,$31,$7F,$A0,$9E,$A7
	.dc.b	$4B,$F6,$3C,$1F,$63,$93,$D7,$1F,$C6,$27,$9E,$1F,$C5,$DB,$3F,$C4
	.dc.b	$DA,$FE,$85,$B7,$E3,$8E,$87,$73,$E3,$E1,$F1,$1F,$29,$F1,$1C,$47
	.dc.b	$BE,$31,$3D,$30,$F4,$D6,$9F,$4C,$E8,$EF,$87,$C1,$BE,$44,$9F,$83
	.dc.b	$1E,$D9,$DE,$1E,$D6,$9E,$D2,$9E,$D2,$D7,$B5,$F4,$BF,$5E,$FC,$91
	.dc.b	$F8,$D3,$C8,$6A,$FC,$35,$F6,$67,$73,$F1,$48,$E0,$CD,$FA,$EB,$DF
	.dc.b	$7E,$BA,$FA,$F7,$CF,$0F,$C5,$47,$73,$F2,$BA,$BB,$BF,$DE,$8D,$81
	.dc.b	$FE,$FF,$70,$FF,$B9,$7D,$8D,$CC,$8D,$E0,$C7,$8C,$2F,$8C,$E6,$3A
	.dc.b	$E6,$C3,$94,$77,$C2,$D6,$E1,$5B,$7B,$E1,$DB,$0E,$AD,$BE,$2F,$8A
	.dc.b	$B8,$57,$53,$89,$EE,$AE,$2D,$AF,$4D,$F8,$6B,$8B,$E3,$AD,$D7,$B6
	.dc.b	$BB,$BE,$CC,$2B,$93,$B9,$4F,$94,$62,$79,$92,$7C,$38,$7A,$6B,$9C
	.dc.b	$E6,$B9,$E1,$E4,$AF,$81,$1C,$EF,$C3,$39,$8F,$61,$92,$1E,$23,$E2
	.dc.b	$13,$0F,$74,$F8,$A3,$D8,$70,$8C,$B6,$BC,$CD,$3B,$9C,$DC,$B4,$CC
	.dc.b	$72,$7C,$4D,$24,$7B,$23,$2D,$8C,$C7,$28,$EB,$18,$8F,$9C,$DA,$F8
	.dc.b	$51,$89,$E4,$47,$DA,$79,$4D,$1D,$4C,$CF,$5B,$FA,$51,$E4,$34,$70
	.dc.b	$25,$78,$23,$73,$E3,$C3,$ED,$E1,$F5,$9B,$7F,$24,$FC,$FB,$87,$C9
	.dc.b	$7C,$E7,$15,$F9,$5A,$B8,$7A,$EF,$89,$CF,$0F,$0C,$7C,$63,$4E,$23
	.dc.b	$DB,$18,$BF,$24,$71,$6B,$F1,$67,$83,$D8,$C4,$6A,$FE,$33,$6B,$A4
	.dc.b	$FB,$70,$F5,$3D,$4F,$CD,$E1,$EA,$7E,$5C,$69,$EA,$68,$FC,$C9,$FC
	.dc.b	$38,$FD,$61,$70,$FE,$2E,$DB,$D3,$F8,$B3,$06,$0E,$26,$48,$7E,$68
	.dc.b	$D9,$CA,$3A,$BF,$3E,$5E,$BF,$E9,$E3,$F8,$BB,$EB,$69,$FB,$87,$11
	.dc.b	$C2,$30,$7C,$31,$8F,$F1,$19,$E9,$3B,$9F,$86,$6B,$1F,$E3,$D3,$A3
	.dc.b	$A9,$63,$A5,$73,$7B,$9C,$9C,$3C,$CF,$E6,$B1,$C9,$DC,$7D,$C6,$CF
	.dc.b	$81,$EA,$D7,$91,$F0,$0E,$CE,$63,$B9,$CE,$37,$5E,$D8,$CC,$FE,$22
	.dc.b	$73,$5E,$B7,$99,$E6,$3C,$67,$EC,$AF,$F9,$AD,$F3,$AF,$01,$CE,$FA
	.dc.b	$D7,$13,$99,$E7,$72,$FE,$EE,$B9,$DF,$86,$7D,$06,$AF,$B5,$78,$70
	.dc.b	$D6,$1E,$08,$C1,$E6,$3B,$3E,$33,$4E,$0D,$9F,$00,$E1,$87,$EC,$2B
	.dc.b	$BC,$6C,$F3,$57,$81,$F6,$5F,$9A,$FA,$D7,$0B,$E5,$86,$EF,$89,$E5
	.dc.b	$9D,$61,$FD,$2A,$75,$5C,$2B,$C7,$39,$69,$FB,$D9,$DC,$66,$3D,$47
	.dc.b	$CE,$AF,$49,$D9,$E0,$72,$3F,$8F,$CF,$27,$27,$2C,$3D,$8F,$BB,$0F
	.dc.b	$93,$87,$E0,$3D,$6F,$D3,$92,$7F,$B3,$FA,$D5,$FC,$7C,$5B,$FE,$FD
	.dc.b	$36,$FF,$C1,$13,$6B,$FC,$2F,$35,$FF,$09,$C5,$C5,$36,$30,$D9,$FE
	.dc.b	$EF,$1F,$BA,$8F,$E0,$DF,$14,$E7,$FE,$25,$48,$C3,$FC,$55,$AF,$F1
	.dc.b	$A6,$BF,$C8,$5A,$FF,$8C,$1A,$FF,$8E,$9A,$F5,$8E,$F1,$82,$38,$33
	.dc.b	$CF,$27,$C4,$D7,$C4,$D7,$C4,$D7,$C4,$D7,$C4,$D7,$DD,$CF,$E2,$23
	.dc.b	$EF,$6B,$EF,$6B,$EF,$6B,$EF,$6B,$EF,$7D,$BC,$F2,$3E,$53,$5F,$94
	.dc.b	$D7,$E5,$35,$F9,$4D,$7E,$53,$5F,$AD,$AF,$CA,$23,$F5,$DA,$FD,$76
	.dc.b	$BF,$5D,$AF,$D7,$6B,$F5,$DF,$AB,$CF,$23,$F0,$4D,$7F,$04,$D7,$F0
	.dc.b	$4D,$7F,$04,$D7,$F0,$4D,$7F,$1D,$CF,$F8,$22,$3F,$8F,$6B,$F8,$F6
	.dc.b	$BF,$8F,$6B,$F8,$F6,$BF,$8F,$7F,$1B,$CF,$23,$F3,$4D,$7F,$34,$D7
	.dc.b	$F3,$4D,$7F,$34,$D7,$F3,$4D,$7F,$4D,$F1,$BF,$9A,$23,$FA,$76,$BF
	.dc.b	$A7,$6B,$FA,$76,$BF,$A7,$6B,$FA,$77,$F4,$BC,$12,$3F,$64,$D7,$F6
	.dc.b	$4D,$7F,$64,$D7,$F6,$4D,$7F,$64,$D5,$F1,$8F,$67,$F8,$0B,$63,$FC
	.dc.b	$28,$D7,$FC,$32,$D7,$FC,$40,$D7,$FD,$5E,$D7,$FD,$72,$D7,$BC,$76
	.dc.b	$23,$D9,$AF,$6E,$0F,$19,$1F,$1B,$5D,$18,$21,$82,$1E,$67,$83,$E8
	.dc.b	$E1,$F7,$91,$F7,$B5,$F7,$B5,$E0,$68,$86,$8A,$7C,$4F,$6B,$E5,$91
	.dc.b	$F9,$6D,$7E,$5B,$5E,$06,$88,$68,$A7,$D0,$E5,$FA,$E4,$7E,$BB,$5F
	.dc.b	$AE,$D7,$81,$A2,$1A,$29,$F7,$3D,$BF,$C1,$91,$FC,$1B,$5F,$C1,$B5
	.dc.b	$E0,$68,$86,$8A,$7E,$23,$A3,$F1,$E4,$7F,$1E,$D7,$F1,$ED,$78,$1A
	.dc.b	$21,$A2,$9F,$94,$F7,$3F,$36,$47,$F3,$6D,$7F,$36,$D7,$81,$A2,$1A
	.dc.b	$29,$FA,$0E,$AF,$D3,$91,$FD,$3B,$5F,$D3,$B5,$E0,$68,$86,$8A,$7F
	.dc.b	$A2,$FE,$35,$E4,$7F,$9E,$DB,$1F,$EF,$A6,$BF,$E5,$EF,$35,$C3,$BA
	.dc.b	$93,$B7,$FD,$09,$3A,$8C,$10,$FF,$33,$D7,$4E,$C7,$12,$65,$71,$77
	.dc.b	$7D,$9A,$6E,$B8,$BC,$30,$D4,$62,$18,$77,$C3,$B4,$62,$FC,$F7,$C2
	.dc.b	$73,$5B,$AF,$23,$56,$1E,$5C,$3B,$4E,$49,$87,$4A,$F8,$13,$A8,$ED
	.dc.b	$1E,$73,$37,$E0,$8C,$C7,$6C,$2D,$F4,$8F,$81,$5E,$AC,$3C,$4F,$18
	.dc.b	$E6,$7A,$2F,$C4,$FC,$82,$61,$E3,$79,$9E,$33,$33,$CD,$A6,$63,$83
	.dc.b	$83,$E5,$1C,$8F,$41,$E2,$32,$6E,$BC,$EF,$94,$D5,$FC,$D8,$CC,$70
	.dc.b	$9E,$E4,$AF,$81,$87,$C4,$C3,$E3,$DF,$0C,$3E,$36,$1D,$EB,$4F,$8E
	.dc.b	$75,$58,$8F,$CE,$4D,$C3,$B4,$7B,$F0,$F8,$51,$88,$F2,$47,$89,$F4
	.dc.b	$D7,$8A,$B9,$4F,$03,$53,$F1,$F0,$F2,$DF,$E5,$70,$F1,$CF,$23,$E5
	.dc.b	$38,$78,$61,$F6,$A7,$C5,$5C,$AF,$AC,$62,$7C,$55,$CC,$F8,$45,$8F
	.dc.b	$55,$F5,$78,$47,$83,$0F,$4B,$F5,$4F,$8A,$71,$BF,$63,$F1,$4D,$4F
	.dc.b	$27,$67,$A0,$FA,$13,$D1,$F9,$27,$12,$93,$0F,$C1,$47,$EF,$E3,$F8
	.dc.b	$B6,$1F,$E8,$89,$FE,$27,$18,$3F,$76,$7E,$5C,$C1,$83,$F2,$26,$08
	.dc.b	$6C,$87,$BC,$E3,$5F,$57,$DC,$FE,$36,$7A,$71,$38,$9F,$88,$7F,$18
	.dc.b	$70,$FC,$13,$F9,$43,$89,$F8,$C7,$58,$FB,$F1,$D4,$D6,$3F,$3A,$3E
	.dc.b	$2D,$78,$1D,$9E,$03,$F8,$9C,$F0,$7E,$9C,$F1,$9E,$11,$E2,$38,$38
	.dc.b	$AF,$14,$79,$DA,$49,$F1,$5F,$ED,$6F,$B5,$7D,$7C,$2E,$1C,$EF,$71
	.dc.b	$8A,$F6,$57,$A0,$ED,$3F,$C6,$63,$9D,$FE,$6E,$7E,$51,$BC,$3E,$3D
	.dc.b	$6E,$30,$F6,$69,$C0,$EA,$58,$F5,$3F,$10,$87,$C8,$38,$9E,$B2,$E1
	.dc.b	$EB,$77,$5F,$05,$DB,$E7,$79,$1F,$34,$F0,$98,$9F,$0B,$D4,$D3,$88
	.dc.b	$E5,$3D,$E7,$85,$7B,$5D,$BF,$05,$EB,$1B,$38,$4F,$2B,$F9,$78,$7F
	.dc.b	$24,$6D,$7F,$21,$6B,$FC,$CD,$B8,$7F,$04,$6E,$1F,$89,$6C,$FC,$8C
	.dc.b	$3C,$11,$F7,$23,$CA,$7C,$6B,$FC,$03,$F8,$E8,$E2,$BC,$E7,$8B,$FD
	.dc.b	$CB,$1F,$EE,$D1,$89,$FE,$94,$D5,$1F,$DE,$FE,$F1,$F1,$DC,$73,$FE
	.dc.b	$F7,$7B,$AE,$78,$67,$0F,$F0,$64,$E6,$BF,$C0,$38,$72,$9E,$F5,$B7
	.dc.b	$CC,$DA,$E1,$1F,$E1,$59,$E0,$F9,$2B,$C3,$5B,$24,$7F,$63,$9D,$DF
	.dc.b	$92,$3B,$90,$E4,$75,$C3,$8B,$83,$C6,$6E,$3E,$14,$69,$E9,$1B,$C3
	.dc.b	$CE,$AB,$C3,$87,$9A,$39,$35,$ED,$1A,$7E,$3E,$1D,$67,$57,$E0,$9E
	.dc.b	$26,$8A,$77,$C3,$83,$93,$C9,$5F,$36,$FB,$47,$B0,$C9,$C9,$CE,$1E
	.dc.b	$A2,$4F,$1A,$EB,$5C,$6F,$E0,$9E,$57,$99,$1C,$E1,$F8,$09,$E3,$7D
	.dc.b	$2B,$EB,$46,$2F,$AE,$1E,$27,$13,$A3,$DF,$5C,$5F,$93,$5A,$7C,$17
	.dc.b	$DD,$DD,$74,$9D,$61,$E4,$9F,$C0,$1A,$3E,$14,$6A,$BE,$89,$CE,$3A
	.dc.b	$9F,$36,$79,$3F,$44,$E4,$68,$C1,$6F,$D1,$1F,$1C,$E2,$E2,$7B,$46
	.dc.b	$2F,$D1,$5D,$23,$BE,$1F,$16,$31,$5C,$0F,$3B,$E5,$6B,$E2,$9E,$8F
	.dc.b	$B8,$E2,$6C,$F8,$67,$9E,$BC,$2D,$38,$14,$F7,$57,$8D,$C4,$FB,$6B
	.dc.b	$C3,$7E,$D9,$EB,$3F,$A8,$7F,$C1,$31,$FC,$3D,$B7,$FE,$09,$7F,$34
	.dc.b	$60,$C4,$6D,$FD,$51,$93,$FC,$20,$E5,$E2,$68,$87,$BC,$FD,$B7,$C7
	.dc.b	$7E,$94,$FC,$03,$FC,$3D,$C3,$F1,$8D,$9F,$C3,$5D,$B8,$3F,$76,$E3
	.dc.b	$03,$CA,$3E,$2E,$1F,$C8,$A3,$18,$7D,$49,$E2,$75,$3C,$15,$C9,$C1
	.dc.b	$C8,$F3,$18,$8E,$47,$F3,$79,$EB,$5F,$4B,$0B,$5C,$0E,$47,$A6,$B8
	.dc.b	$3E,$0A,$DB,$FB,$FA,$FC,$5D,$F8,$E3,$CE,$72,$71,$5C,$A3,$15,$CB
	.dc.b	$0F,$C1,$47,$7B,$F8,$4F,$73,$D4,$F9,$DA,$79,$EB,$DB,$1C,$23,$C3
	.dc.b	$18,$75,$19,$8C,$47,$73,$27,$E0,$67,$F9,$BD,$FA,$2B,$81,$E1,$9E
	.dc.b	$AE,$1E,$18,$7C,$49,$E1,$18,$3C,$E7,$73,$E4,$1F,$8F,$76,$E6,$BE
	.dc.b	$1C,$60,$CB,$8C,$3E,$45,$7F,$24,$6D,$7F,$37,$6B,$FB,$26,$E1,$FC
	.dc.b	$19,$B8,$7D,$86,$CF,$8D,$AF,$98,$F4,$61,$E2,$BF,$04,$7B,$30,$F2
	.dc.b	$36,$32,$FC,$63,$E6,$5F,$EE,$5F,$F0,$83,$64,$D2,$47,$EF,$1A,$FE
	.dc.b	$E9,$AF,$F0,$66,$BF,$E1,$07,$B7,$C7,$BA,$29,$27,$F8,$54,$65,$B1
	.dc.b	$97,$FD,$2D,$19,$6C,$65,$FF,$4E,$C6,$5B,$19,$44,$87,$47,$90,$91
	.dc.b	$E2,$6B,$E0,$6B,$D9,$AF,$41,$1D,$1E,$44,$3D,$3D,$1F,$08,$8F,$C0
	.dc.b	$6B,$A3,$04,$30,$43,$DE,$F0,$7E,$0F,$47,$CE,$23,$F3,$1A,$FC,$A6
	.dc.b	$BC,$0D,$10,$D1,$4F,$9B,$D1,$F6,$88,$FD,$86,$BF,$59,$AF,$03,$44
	.dc.b	$34,$53,$FD,$AB,$A3,$87,$FC,$9D,$83,$CE,$A7,$F8,$96,$25,$7D,$CE
	.dc.b	$0F,$CB,$BF,$76,$76,$FF,$93,$67,$6F,$F9,$1E,$BA,$3C,$75,$FB,$51
	.dc.b	$88,$FF,$1F,$CF,$27,$85,$5B,$FF,$37,$4D,$BE,$33,$6B,$FC,$C5,$35
	.dc.b	$FC,$2C,$5C,$3D,$D1,$E1,$3D,$05,$36,$5C,$3C,$98,$78,$23,$15,$D6
	.dc.b	$73,$87,$F9,$E3,$0C,$4F,$76,$D7,$03,$E2,$14,$F6,$96,$FB,$38,$23
	.dc.b	$F3,$9F,$29,$F1,$8E,$15,$F0,$9A,$7C,$22,$9C,$23,$95,$FA,$23,$47
	.dc.b	$69,$E4,$D3,$33,$C2,$FD,$71,$E1,$34,$75,$9E,$33,$EA,$38,$47,$8A
	.dc.b	$BC,$0F,$FA,$06,$F9,$CE,$A3,$A3,$49,$5D,$0F,$9D,$1C,$EF,$E7,$34
	.dc.b	$F7,$E1,$F7,$9F,$2E,$1E,$03,$C3,$19,$BE,$B3,$F8,$23,$D7,$19,$BF
	.dc.b	$04,$66,$76,$FA,$4F,$01,$C3,$0F,$5C,$F0,$8D,$1E,$2C,$3C,$ED,$3C
	.dc.b	$F3,$C1,$ED,$3E,$E7,$47,$53,$BD,$78,$AF,$CD,$1C,$08,$7D,$33,$EC
	.dc.b	$CF,$8F,$0F,$85,$87,$ED,$2F,$D7,$18,$AF,$CA,$9E,$08,$ED,$87,$96
	.dc.b	$3A,$1E,$C8,$E3,$5E,$C6,$C7,$B0,$FC,$F1,$FD,$1A,$3F,$A9,$91,$FE
	.dc.b	$A2,$D7,$FA,$7B,$5E,$06,$88,$68,$B1,$FA,$5F,$90,$7C,$7F,$FB,$9D
	.dc.b	$B3,$F2,$31,$FE,$91,$18,$8F,$F6,$D3,$6B,$D3,$18,$AE,$47,$C1,$9D
	.dc.b	$57,$28,$C1,$A9,$EC,$7F,$60,$9D,$B6,$36,$FF,$0F,$8C,$BF,$33,$1F
	.dc.b	$D2,$C6,$0C,$BF,$56,$3C,$87,$19,$E6,$E0,$FB,$15,$E5,$74,$79,$05
	.dc.b	$75,$9E,$37,$F2,$5B,$5C,$6B,$99,$E9,$39,$98,$39,$C7,$C1,$71,$5D
	.dc.b	$5F,$03,$5F,$E0,$ED,$AF,$04,$7F,$16,$8C,$57,$86,$7F,$5F,$18,$9F
	.dc.b	$11,$CA,$BC,$13,$CA,$30,$79,$9C,$1C,$A3,$75,$D4,$FD,$B5,$75,$AF
	.dc.b	$59,$E3,$9D,$BE,$0A,$E4,$F5,$AE,$4F,$EC,$2B,$91,$F6,$30,$E3,$1D
	.dc.b	$0F,$AD,$5C,$9B,$5C,$9F,$D6,$D7,$2B,$FE,$B4,$D8,$F2,$1E,$13,$C9
	.dc.b	$3E,$A9,$FF,$26,$4F,$E7,$67,$FC,$4C,$78,$4A,$E5,$FB,$C6,$23,$89
	.dc.b	$1F,$F6,$74,$6D,$FA,$26,$CC,$C6,$D5,$7F,$57,$FB,$D6,$9F,$E3,$78
	.dc.b	$C3,$F1,$2F,$F4,$33,$AA,$FE,$E4,$D7,$E2,$9E,$E3,$FC,$96,$E6,$B9
	.dc.b	$C7,$C3,$3F,$3B,$ED,$FF,$01,$23,$D3,$7A,$FF,$00,$8B,$3E,$79,$B9
	.dc.b	$7F,$06,$9B,$5E,$09,$E2,$FC,$A8,$F9,$E4,$7E,$3B,$5F,$A2,$79,$0C
	.dc.b	$90,$CE,$08,$E1,$1D,$4D,$5F,$F8,$07,$E4,$9E,$FF,$CB,$C7,$FC,$C5
	.dc.b	$18,$AF,$E1,$97,$E8,$51,$97,$EA,$CF,$0A,$D6,$1E,$19,$FF,$11,$56
	.dc.b	$F0,$F0,$56,$B0,$F0,$57,$F9,$F2,$B5,$87,$38,$F0,$90,$EE,$6E,$B8
	.dc.b	$BE,$47,$A3,$E0,$73,$3D,$23,$F9,$55,$7A,$E3,$11,$C4,$ED,$7F,$D0
	.dc.b	$67,$8D,$49,$CD,$FD,$B7,$B1,$0F,$C1,$1A,$3C,$C6,$6B,$A3,$93,$C0
	.dc.b	$7B,$0A,$E9,$E6,$66,$FC,$E7,$A3,$0F,$9D,$1A,$9F,$9A,$F7,$AF,$33
	.dc.b	$97,$E6,$3C,$EF,$E3,$B4,$E0,$7C,$62,$9E,$48,$E3,$3E,$53,$BC,$71
	.dc.b	$23,$98,$D9,$23,$E1,$9E,$32,$C7,$22,$47,$A1,$B1,$A8,$E4,$FE,$84
	.dc.b	$EE,$E2,$FE,$26,$1F,$5E,$72,$71,$3D,$86,$0F,$61,$93,$A5,$78,$1C
	.dc.b	$93,$0F,$9F,$19,$AF,$A7,$87,$E2,$CF,$04,$71,$9D,$E1,$E5,$9C,$BE
	.dc.b	$D3,$17,$F3,$A7,$6F,$8E,$FA,$BF,$87,$6B,$E9,$8D,$BE,$68,$CD,$FD
	.dc.b	$18,$E5,$7E,$07,$E0,$C7,$A6,$31,$1F,$02,$FF,$0A,$FF,$3E,$AF,$DF
	.dc.b	$36,$7F,$88,$BF,$68,$C1,$88,$DB,$C8,$86,$CA,$7E,$80,$E2,$7F,$1C
	.dc.b	$7F,$3B,$F0,$9F,$B8,$3C,$AE,$23,$F8,$5C,$62,$75,$18,$3F,$2F,$3B
	.dc.b	$7F,$51,$3B,$74,$60,$DC,$F2,$70,$7E,$16,$79,$5F,$F2,$7A,$E5,$3A
	.dc.b	$8C,$3C,$8E,$AE,$1E,$51,$D6,$30,$7E,$25,$C5,$FF,$30,$7D,$E6,$27
	.dc.b	$F9,$23,$87,$AC,$6D,$F7,$C6,$DF,$E6,$AE,$DF,$34,$6D,$F1,$9F,$5E
	.dc.b	$3E,$DB,$C0,$C1,$0F,$A8,$EA,$8D,$24,$6C,$FC,$43,$87,$89,$FD,$0B
	.dc.b	$C6,$FE,$66,$79,$E7,$9C,$64,$87,$A0,$F8,$26,$2F,$FA,$C4,$F1,$AE
	.dc.b	$73,$9B,$F7,$57,$CE,$7F,$AD,$D6,$A7,$A1,$EF,$3F,$96,$39,$9E,$11
	.dc.b	$FD,$4E,$31,$3C,$A3,$E8,$99,$3C,$05,$3A,$DA,$78,$C6,$6B,$AB,$FD
	.dc.b	$AA,$32,$D3,$A9,$4F,$29,$D8,$F4,$47,$C2,$71,$3D,$DE,$2D,$78,$92
	.dc.b	$3C,$31,$97,$94,$66,$BE,$49,$EF,$70,$7A,$5C,$BE,$F3,$F8,$35,$7E
	.dc.b	$14,$8F,$F0,$36,$BF,$BD,$6B,$C0,$D1,$0D,$16,$52,$89,$F7,$CF,$84
	.dc.b	$FF,$00,$B6,$3E,$1D,$E2,$64,$60,$E5,$7D,$1E,$71,$A7,$DE,$7F,$87
	.dc.b	$BC,$7F,$D0,$11,$FE,$3C,$D7,$F8,$C3,$5F,$E5,$0D,$7F,$C5,$6D,$7F
	.dc.b	$C3,$CF,$EA,$FC,$79,$1F,$CA,$5A,$FF,$23,$6B,$FC,$CD,$AF,$F8,$D9
	.dc.b	$AF,$F8,$A5,$AF,$F8,$DB,$C7,$FD,$C1,$1F,$E6,$CD,$7F,$98,$35,$FE
	.dc.b	$80,$D7,$FC,$86,$D7,$FC,$6C,$FF,$03,$F0,$A4,$7F,$41,$6B,$FC,$ED
	.dc.b	$AF,$F4,$B6,$BF,$E4,$E6,$BF,$E4,$16,$AC,$9F,$07,$F7,$B7,$93,$F6
	.dc.b	$B5,$FC,$23,$B9,$FC,$19,$A7,$FC,$59,$5F,$E3,$E3,$9B,$C4,$FC,$5C
	.dc.b	$6F,$4F,$C2,$47,$86,$F9,$BB,$3B,$16,$BA,$46,$A3,$BD,$73,$BF,$2D
	.dc.b	$78,$63,$57,$DA,$38,$D7,$39,$FF,$2D,$FA,$EE,$CF,$E6,$AD,$9F,$EE
	.dc.b	$2D,$7F,$D0,$6D,$7F,$D6,$AD,$7E,$2B,$5F,$D5,$35,$34,$91,$E0,$24
	.dc.b	$78,$39,$3C,$64,$70,$43,$C6,$E5,$A6,$4F,$2B,$96,$9E,$CE,$6F,$82
	.dc.b	$47,$E8,$70,$BC,$1C,$47,$F9,$9F,$0F,$F5,$C6,$9E,$EB,$FF,$41,$DE
	.dc.b	$EB,$EC,$56,$9F,$83,$3C,$B4,$FB,$F5,$B9,$E9,$3F,$40,$8F,$94,$87
	.dc.b	$D0,$72,$D3,$27,$D2,$72,$D3,$8C,$F8,$3D,$AF,$39,$E3,$C3,$D5,$3F
	.dc.b	$02,$32,$FA,$AB,$4F,$F2,$3B,$4E,$A3,$D2,$7C,$0F,$1F,$F1,$64,$7E
	.dc.b	$5F,$A3,$5F,$95,$17,$1F,$87,$37,$1F,$C4,$CD,$AF,$04,$F1,$7F,$1F
	.dc.b	$1F,$96,$23,$F9,$66,$BF,$96,$6B,$C0,$D1,$0D,$16,$BE,$AF,$8D,$F7
	.dc.b	$0F,$14,$7E,$62,$70,$FE,$1A,$33,$9A,$73,$3F,$EA,$DC,$95,$AF,$53
	.dc.b	$F5,$E4,$9F,$C7,$FB,$D5,$FC,$74,$5C,$7F,$0F,$37,$3F,$D6,$CD,$AF
	.dc.b	$A9,$3C,$5F,$D9,$C7,$EE,$88,$FE,$E9,$AF,$EE,$9A,$F0,$34,$43,$45
	.dc.b	$AF,$CC,$F8,$FF,$A0,$3C,$51,$FB,$C9,$C3,$FA,$A8,$CE,$69,$CC,$FC
	.dc.b	$59,$CF,$C4,$7F,$12,$24,$7E,$CB,$C0,$AF,$EC,$62,$E1,$FA,$A9,$B9
	.dc.b	$7F,$10,$9B,$5F,$97,$9E,$2F,$F1,$68,$FE,$3A,$47,$F8,$EB,$5F,$E3
	.dc.b	$AD,$78,$1A,$21,$A2,$D7,$EF,$3E,$47,$F8,$19,$E2,$8F,$E3,$F3,$87
	.dc.b	$F8,$6C,$67,$24,$E6,$7F,$11,$39,$9F,$E6,$A4,$8F,$E2,$BF,$21,$5F
	.dc.b	$E2,$91,$6F,$F8,$6C,$DC,$7F,$99,$CD,$AF,$DD,$4F,$17,$F9,$CC,$7F
	.dc.b	$40,$23,$FD,$01,$AF,$F4,$06,$BC,$0D,$10,$D1,$6B,$F8,$FF,$C7,$FF
	.dc.b	$26,$3C,$51,$FD,$0A,$70,$FF,$30,$8C,$C7,$F4,$68,$C1,$0F,$79,$FC
	.dc.b	$BB,$E3,$7F,$AC,$91,$FE,$73,$E3,$D7,$F9,$C4,$5B,$FE,$63,$37,$1F
	.dc.b	$EA,$F3,$6B,$F8,$F4,$F1,$7F,$AE,$C7,$F6,$42,$3F,$D9,$1A,$FF,$64
	.dc.b	$6B,$C0,$D1,$0D,$16,$BF,$C0,$3F,$7E,$F8,$FF,$F0,$6C,$61,$F2,$4F
	.dc.b	$F6,$8A,$CE,$1F,$E1,$2B,$E1,$3F,$E0,$76,$BF,$00,$F8,$C6,$6B,$8B
	.dc.b	$FE,$5D,$8D,$DF,$F8,$4B,$16,$1A,$A7,$CA,$B2,$C3,$63,$FC,$2C,$D5
	.dc.b	$7E,$CF,$27,$75,$27,$6F,$5C,$53,$99,$E3,$9F,$F8,$BB,$0E,$98,$7F
	.dc.b	$89,$9B,$1F,$06,$B8,$57,$F8,$D0,$F3,$E1,$FE,$CE,$AF,$24,$65,$AF
	.dc.b	$58,$D1,$1E,$47,$FA,$06,$36,$D8,$D3,$A8,$CB,$E0,$38,$94,$F5,$9F
	.dc.b	$3F,$2F,$80,$7B,$F0,$F6,$56,$27,$ED,$4E,$70,$F3,$46,$08,$74,$3B
	.dc.b	$57,$C5,$F9,$2F,$AA,$48,$FC,$27,$C1,$57,$86,$29,$C5,$7E,$4A,$AD
	.dc.b	$FD,$19,$B7,$F8,$79,$E4,$FC,$B8,$F9,$C4,$7E,$0D,$F5,$6C,$65,$F9
	.dc.b	$0D,$79,$1A,$21,$A3,$F3,$9F,$03,$F8,$73,$C6,$FE,$52,$30,$E8,$FC
	.dc.b	$9B,$AD,$13,$A9,$F8,$33,$F9,$AE,$19,$1F,$9A,$7F,$29,$CB,$88,$F8
	.dc.b	$F1,$73,$FD,$BC,$DC,$FF,$2D,$36,$BF,$59,$3C,$5F,$C8,$47,$E4,$C8
	.dc.b	$FC,$66,$BF,$8B,$6B,$C0,$D1,$0D,$16,$BF,$84,$FA,$BF,$AB,$3C,$51
	.dc.b	$FB,$89,$C3,$FC,$16,$33,$9A,$73,$3F,$8B,$CD,$5A,$F9,$5F,$DF,$92
	.dc.b	$7F,$87,$7A,$F5,$FC,$5C,$5C,$7F,$8D,$CD,$CF,$F7,$53,$6B,$F8,$84
	.dc.b	$F1,$7F,$69,$1F,$B7,$23,$F8,$A6,$BF,$AF,$6B,$C0,$D1,$0D,$16,$BF
	.dc.b	$96,$F7,$FF,$0F,$3C,$51,$FC,$72,$70,$FF,$29,$8C,$E6,$9C,$CF,$C5
	.dc.b	$9C,$F5,$7F,$22,$24,$7F,$31,$F0,$6B,$FA,$E8,$B8,$7F,$3E,$9B,$97
	.dc.b	$F1,$D9,$B5,$FC,$CE,$78,$BF,$C5,$E3,$F8,$D1,$1F,$D6,$B5,$FE,$24
	.dc.b	$D7,$81,$A2,$1A,$2D,$7F,$50,$F4,$FF,$99,$1E,$28,$FE,$7F,$38,$7F
	.dc.b	$A6,$46,$72,$4E,$67,$F1,$13,$9F,$53,$FA,$29,$23,$FD,$B5,$F8,$03
	.dc.b	$C9,$C4,$7F,$12,$9B,$8F,$FB,$8A,$B5,$3F,$D3,$27,$39,$FF,$43,$7F
	.dc.b	$AA,$4E,$A7,$91,$FC,$42,$3F,$A0,$11,$C1,$0F,$E7,$0E,$5A,$64,$FE
	.dc.b	$8C,$E5,$A7,$AB,$CB,$F0,$9E,$47,$FA,$34,$61,$D1,$E9,$21,$FC,$DE
	.dc.b	$53,$B8,$F4,$9F,$CD,$3C,$5F,$ED,$A4,$7E,$4F,$B7,$5D,$1F,$DC,$23
	.dc.b	$E1,$C6,$71,$FE,$CD,$17,$15,$4C,$17,$CE,$3F,$B2,$57,$F6,$72,$3F
	.dc.b	$CE,$9A,$FF,$5F,$6B,$C0,$D1,$0D,$16,$FE,$E7,$B9,$F6,$0F,$24,$7F
	.dc.b	$6D,$9C,$3F,$86,$8C,$BF,$E1,$08,$CB,$F2,$63,$3F,$2E,$F8,$FF,$F0
	.dc.b	$B3,$67,$DF,$6C,$BE,$0C,$27,$4F,$D3,$F8,$0E,$37,$23,$8D,$70,$AD
	.dc.b	$DF,$E5,$6F,$55,$FD,$C6,$73,$5B,$BF,$CA,$D6,$AB,$FC,$29,$5E,$06
	.dc.b	$CF,$F8,$19,$F2,$36,$35,$1C,$CF,$F0,$5B,$87,$ED,$B5,$E2,$6A,$39
	.dc.b	$11,$C1,$FB,$2F,$73,$F5,$27,$AB,$0F,$8D,$1D,$27,$2E,$CF,$86,$64
	.dc.b	$A7,$80,$FC,$E4,$F0,$7E,$24,$7D,$4F,$3A,$47,$0B,$47,$EE,$BE,$47
	.dc.b	$33,$F2,$E2,$E1,$FC,$0E,$6E,$5F,$4E,$6D,$7E,$6E,$78,$BF,$3A,$3E
	.dc.b	$51,$1F,$92,$D7,$FC,$7B,$7E,$27,$24,$33,$1B,$7A,$19,$BF,$E2,$DE
	.dc.b	$A7,$F1,$23,$C7,$1F,$8D,$9C,$3F,$C8,$23,$35,$DE,$3F,$81,$46,$9F
	.dc.b	$80,$7C,$5F,$77,$F3,$64,$7F,$94,$D5,$7F,$21,$16,$FF,$94,$4D,$C7
	.dc.b	$F3,$13,$6B,$F8,$04,$F1,$7F,$27,$1F,$8F,$23,$F8,$E6,$BF,$19,$AF
	.dc.b	$03,$44,$34,$5A,$FE,$75,$F1,$1F,$CD,$8F,$14,$7E,$C6,$70,$FF,$44
	.dc.b	$8C,$E2,$9C,$CF,$F2,$79,$CE,$9F,$BE,$24,$7E,$4B,$F0,$17,$A5,$88
	.dc.b	$FD,$A4,$DC,$7F,$3B,$0A,$DB,$F9,$39,$D6,$7F,$BF,$7F,$D4,$13,$A9
	.dc.b	$E6,$7E,$CA,$3F,$6C,$47,$04,$3F,$1A,$E5,$A6,$4F,$E0,$6E,$5A,$7A
	.dc.b	$F8,$7C,$27,$91,$FD,$FC,$61,$D1,$E9,$21,$FE,$83,$94,$EE,$3D,$27
	.dc.b	$F1,$7F,$83,$FE,$4E,$47,$E5,$7A,$75,$D1,$FB,$78,$F8,$71,$9C,$7E
	.dc.b	$5E,$A9,$C6,$5F,$CB,$B1,$5F,$3C,$3F,$90,$D7,$F1,$D2,$3F,$C7,$1A
	.dc.b	$FE,$D1,$AF,$03,$44,$34,$5C,$3E,$ED,$3D,$1F,$B2,$79,$27,$F9,$3D
	.dc.b	$61,$FC,$44,$65,$FE,$89,$19,$7E,$54,$67,$17,$A5,$FC,$D4,$93,$FC
	.dc.b	$D1,$AF,$F0,$C6,$BF,$D0,$1A,$FF,$4B,$6B,$FC,$E9,$FE,$81,$CD,$23
	.dc.b	$F9,$FB,$5F,$E2,$ED,$7F,$A5,$B5,$FE,$AC,$D7,$FA,$2B,$5F,$EC,$1E
	.dc.b	$C7,$F4,$D2,$3F,$D3,$1A,$FF,$24,$6B,$FD,$61,$AF,$F6,$36,$BF,$D4
	.dc.b	$9F,$EB,$1D,$32,$3F,$AB,$B5,$FE,$5E,$D7,$FB,$1B,$5F,$ED,$CD,$7F
	.dc.b	$AE,$B5,$FE,$F1,$EC,$7F,$65,$23,$FD,$91,$AF,$F3,$C6,$BF,$DC,$1A
	.dc.b	$FF,$7B,$6B,$FD,$A9,$FE,$E1,$D9,$23,$FB,$7B,$5F,$E8,$ED,$7F,$BD
	.dc.b	$B5,$FF,$03,$35,$FE,$EA,$D5,$F3,$2E,$EF,$F0,$9E,$FF,$C5,$23,$13
	.dc.b	$FD,$B1,$AF,$EE,$FB,$F7,$C5,$FE,$1A,$9C,$D7,$1B,$CD,$FF,$8F,$A7
	.dc.b	$35,$FB,$3B,$CC,$F8,$AA,$D6,$A3,$84,$7F,$20,$9C,$BF,$E1,$59,$F1
	.dc.b	$CF,$F8,$A8,$E4,$53,$89,$4C,$99,$78,$BE,$98,$E8,$47,$84,$69,$E6
	.dc.b	$71,$8D,$3F,$C7,$7D,$5F,$E6,$A7,$99,$F9,$11,$87,$FA,$24,$65,$61
	.dc.b	$F1,$DC,$90,$CE,$1E,$57,$E9,$7C,$0F,$C6,$23,$FD,$3F,$D6,$AF,$03
	.dc.b	$E8,$46,$2B,$F8,$9C,$DA,$FB,$13,$6B,$FC,$E7,$C8,$AE,$71,$F3,$72
	.dc.b	$57,$D1,$24,$7C,$66,$BF,$35,$AF,$16,$BE,$03,$81,$0F,$E8,$DE,$57
	.dc.b	$F5,$E3,$C8,$FE,$5A,$30,$FF,$6D,$8C,$CF,$92,$30,$43,$E1,$1C,$4F
	.dc.b	$CD,$79,$1F,$8E,$23,$FD,$DF,$D5,$AF,$09,$4E,$33,$FE,$77,$57,$3F
	.dc.b	$D0,$4D,$BF,$93,$3C,$9F,$91,$1F,$97,$23,$F8,$C6,$BF,$93,$6B,$C0
	.dc.b	$D1,$0D,$16,$BF,$A6,$FA,$BF,$E0,$43,$C5,$1F,$BC,$9C,$3F,$E1,$38
	.dc.b	$CE,$1E,$C8,$C1,$0F,$81,$3F,$B2,$F9,$59,$1F,$C0,$5F,$F0,$8F,$A7
	.dc.b	$88,$F8,$D1,$70,$FE,$BF,$37,$2F,$E0,$93,$6B,$F1,$B3,$C5,$FC,$64
	.dc.b	$7E,$EC,$8F,$EC,$1A,$FE,$DD,$AF,$03,$44,$34,$5A,$FF,$16,$FA,$4F
	.dc.b	$5B,$F8,$7B,$67,$D8,$43,$F0,$EE,$5B,$1E,$97,$F9,$3D,$72,$89,$1C
	.dc.b	$6B,$F9,$47,$A7,$FC,$68,$91,$FE,$94,$B7,$AB,$88,$FC,$74,$DA,$F3
	.dc.b	$CD,$CB,$F9,$64,$DA,$FF,$73,$DF,$27,$F9,$A4,$7F,$23,$23,$FC,$61
	.dc.b	$AF,$F1,$F6,$BF,$BB,$6B,$D8,$E0,$43,$81,$E9,$F5,$7E,$49,$E3,$7F
	.dc.b	$A4,$C6,$1D,$1E,$42,$1F,$32,$34,$FC,$03,$FA,$67,$C8,$7F,$40,$23
	.dc.b	$F1,$EA,$BF,$B3,$8B,$8F,$DC,$9B,$87,$F5,$19,$B5,$F1,$A7,$8B,$FC
	.dc.b	$66,$3F,$A4,$11,$FE,$7A,$D7,$FA,$23,$5E,$06,$88,$68,$B5,$F8,$EC
	.dc.b	$3F,$1A,$78,$A3,$FB,$6C,$E1,$FC,$14,$67,$64,$E6,$7F,$23,$39,$CB
	.dc.b	$FB,$19,$23,$F1,$5E,$AD,$7F,$8B,$45,$C3,$F3,$F3,$72,$FE,$E9,$36
	.dc.b	$BF,$13,$3C,$5F,$E7,$91,$FD,$AC,$8F,$F6,$06,$BF,$D9,$DA,$F0,$34
	.dc.b	$43,$45,$AF,$D8,$FA,$9F,$B0,$3C,$51,$FE,$10,$9C,$3F,$A5,$8C,$E5
	.dc.b	$F9,$D8,$FD,$9C,$69,$F7,$9F,$E1,$33,$FC,$0E,$43,$F5,$DC,$15,$FE
	.dc.b	$74,$5B,$FE,$07,$17,$1F,$F0,$C4,$D3,$F5,$B1,$C5,$FE,$C0,$7F,$83
	.dc.b	$C8,$7F,$81,$0A,$7F,$82,$CA,$70,$34,$43,$45,$9F,$E2,$BE,$9F,$F1
	.dc.b	$43,$C5,$1F,$E3,$43,$07,$F0,$97,$38,$7F,$87,$E7,$8C,$E4,$F8,$80
	.dc.b	$FE,$C3,$DC,$E3,$FF,$17,$B6,$35,$3F,$ED,$28,$C4,$7F,$22,$F4,$F7
	.dc.b	$52,$37,$97,$55,$5C,$2F,$F9,$7D,$EE,$7C,$37,$67,$31,$C8,$F3,$57
	.dc.b	$81,$FF,$1D,$35,$EC,$D7,$73,$FE,$3B,$7C,$2D,$8D,$46,$0F,$F2,$03
	.dc.b	$82,$46,$C9,$1A,$7F,$A2,$F5,$C8,$F3,$1C,$CF,$59,$FC,$ED,$D3,$82
	.dc.b	$57,$F5,$08,$CB,$F3,$67,$E8,$FC,$C4,$AE,$B6,$8F,$F3,$1F,$E0,$4E
	.dc.b	$FD,$57,$DA,$9B,$83,$3F,$A5,$7B,$AE,$75,$A9,$F9,$2F,$4B,$E7,$3F
	.dc.b	$3A,$3E,$41,$1F,$8E,$D7,$FC,$D5,$87,$95,$C9,$0C,$C6,$DE,$86,$70
	.dc.b	$F3,$3F,$93,$C1,$E0,$FB,$8F,$34,$7E,$0A,$B0,$FD,$09,$DB,$EC,$21
	.dc.b	$FE,$3C,$72,$D9,$E6,$FE,$53,$B7,$F3,$E4,$7A,$78,$55,$FD,$2C,$5C
	.dc.b	$3C,$F0,$B5,$67,$3F,$C9,$DD,$AF,$64,$F3,$7F,$33,$1F,$93,$23,$F9
	.dc.b	$26,$BF,$31,$AF,$03,$44,$34,$5B,$F2,$DB,$D2,$FC,$39,$E3,$9F,$80
	.dc.b	$7C,$97,$2F,$E3,$AB,$53,$C6,$34,$64,$FC,$9F,$C1,$7F,$06,$23,$F4
	.dc.b	$FD,$3A,$FF,$0C,$8B,$87,$92,$6E,$5F,$B8,$9B,$5F,$81,$9E,$2F,$EF
	.dc.b	$A3,$F7,$04,$7F,$6E,$D7,$F2,$2D,$78,$1A,$21,$A2,$D7,$C6,$F3,$3F
	.dc.b	$56,$78,$A3,$F8,$BC,$E1,$FC,$74,$67,$64,$E6,$7F,$1D,$39,$F2,$7F
	.dc.b	$94,$92,$3F,$2D,$E9,$57,$F9,$7C,$5C,$3E,$BC,$DC,$BF,$8D,$CD,$AF
	.dc.b	$D2,$4F,$17,$F9,$0C,$7F,$1B,$23,$FC,$69,$AF,$ED,$1A,$F0,$34,$43
	.dc.b	$45,$AF,$C4,$FA,$5F,$C3,$8F,$14,$7F,$3A,$9C,$3F,$B1,$8C,$E4,$9C
	.dc.b	$CF,$EC,$67,$3F,$21,$FD,$30,$91,$FB,$9F,$4E,$BF,$D4,$A2,$E1,$F9
	.dc.b	$A9,$B9,$7F,$3E,$9B,$5F,$C1,$E7,$8B,$FD,$12,$3F,$9F,$11,$FE,$7A
	.dc.b	$D7,$F8,$BB,$5E,$06,$88,$68,$B5,$FA,$CF,$4B,$F9,$89,$E2,$8F,$EB
	.dc.b	$B3,$87,$F8,$A4,$67,$24,$E6,$7F,$8A,$4E,$7D,$2F,$ED,$E4,$8F,$E9
	.dc.b	$7F,$81,$A3,$11,$FD,$D9,$B8,$7F,$12,$9D,$4F,$F4,$69,$CE,$5F,$D9
	.dc.b	$1F,$EE,$55,$B7,$9C,$FF,$6A,$7F,$B2,$11,$FE,$C6,$D7,$F9,$EB,$5E
	.dc.b	$06,$88,$68,$B1,$E5,$F4,$FC,$35,$FE,$03,$B4,$79,$AB,$FB,$ED,$E9
	.dc.b	$FF,$0A,$4E,$9F,$87,$3B,$7F,$C1,$73,$B5,$3B,$7F,$DF,$6D,$BF,$F0
	.dc.b	$C4,$EF,$E7,$1F,$05,$49,$7F,$43,$D5,$AA,$FF,$9F,$D7,$08,$FE,$FE
	.dc.b	$D7,$BC,$F0,$9D,$D5,$CB,$E0,$4F,$4A,$FF,$12,$0B,$E4,$E1,$AF,$F8
	.dc.b	$A5,$AF,$F8,$DD,$AF,$1A,$FF,$0F,$46,$08,$E0,$91,$E0,$39,$38,$7E
	.dc.b	$3F,$07,$CB,$CB,$F7,$51,$EE,$9F,$53,$F5,$27,$46,$1F,$11,$F9,$87
	.dc.b	$2F,$13,$E3,$F0,$7D,$12,$3F,$9E,$F4,$6B,$F3,$F2,$F8,$13,$8C,$FF
	.dc.b	$3D,$56,$73,$5D,$0E,$45,$3D,$D9,$2F,$E2,$92,$3E,$11,$F4,$5C,$3C
	.dc.b	$1A,$FD,$36,$BE,$03,$47,$28,$EC,$FE,$E3,$D1,$F9,$47,$9D,$FB,$13
	.dc.b	$87,$47,$9F,$0F,$C6,$C6,$63,$F3,$33,$C5,$FC,$37,$C2,$7E,$64,$8F
	.dc.b	$F0,$6E,$9A,$FC,$93,$F0,$D1,$8B,$FE,$0D,$37,$1F,$C6,$4D,$AF,$D7
	.dc.b	$CF,$27,$F2,$51,$F8,$C2,$3E,$F6,$BF,$39,$AF,$03,$44,$34,$5A,$FE
	.dc.b	$31,$E8,$7E,$3C,$F1,$47,$B8,$FE,$42,$E5,$FD,$7C,$EB,$64,$E6,$7E
	.dc.b	$01,$C4,$B9,$BE,$43,$F4,$A4,$8F,$B0,$D7,$F0,$ED,$7F,$64,$D7,$F6
	.dc.b	$4D,$7F,$64,$FE,$C3,$AA,$47,$E0,$DA,$FE,$49,$AF,$EF,$5A,$FE,$F5
	.dc.b	$AF,$EF,$5A,$FF,$05,$F5,$FF,$7A,$47,$F2,$0D,$7F,$3E,$D7,$F8,$53
	.dc.b	$5F,$E1,$4D,$7F,$85,$3F,$C2,$3D,$6F,$E4,$84,$8F,$EA,$3E,$9D,$7E
	.dc.b	$5C,$5D,$3F,$A8,$4D,$DB,$F8,$B4,$DA,$FE,$73,$3C,$5F,$E3,$91,$DD
	.dc.b	$FD,$43,$5F,$DB,$B5,$D9,$A2,$1A,$29,$A3,$F9,$FF,$89,$FC,$64,$F1
	.dc.b	$3F,$CB,$63,$0F,$F6,$C8,$CE,$69,$CC,$F2,$3F,$9E,$BF,$C8,$FD,$3F
	.dc.b	$E9,$DF,$82,$7F,$BA,$F1,$FF,$6E,$36,$BF,$AB,$CE,$DF,$F6,$1C,$65
	.dc.b	$37,$74,$EE,$7F,$33,$3B,$35,$19,$CF,$FD,$1F,$19,$9C,$5F,$F6,$BA
	.dc.b	$F2,$C4,$3F,$C1,$2E,$9F,$F7,$F4,$F1,$72,$79,$BC,$DF,$69,$1F,$4F
	.dc.b	$CD,$D7,$CB,$17,$0F,$1C,$DC,$3E,$1C,$DA,$EF,$35,$F8,$31,$6D,$36
	.dc.b	$30,$D9,$F2,$1E,$23,$FD,$DB,$F3,$CF,$2B,$FC,$00,$D8,$CC,$7D,$6F
	.dc.b	$4F,$2C,$8C,$9F,$52,$30,$FF,$08,$9E,$4D,$9E,$4F,$03,$FB,$44,$69
	.dc.b	$FE,$F2,$D4,$47,$0D,$8F,$E9,$EA,$32,$D8,$CC,$60,$8E,$09,$1F,$D6
	.dc.b	$23,$CF,$97,$E1,$3D,$2F,$CA,$9E,$C7,$FD,$AB,$1A,$CF,$F2,$F3,$AA
	.dc.b	$F6,$56,$AF,$11,$F9,$29,$E5,$5F,$9E,$F4,$BF,$43,$EB,$4A,$FC,$EC
	.dc.b	$7E,$90,$C3,$F9,$A8,$C3,$A3,$F4,$6E,$B0,$FC,$DC,$EA,$78,$1A,$38
	.dc.b	$1F,$A4,$78,$57,$EC,$2B,$85,$72,$AE,$07,$23,$81,$FB,$2F,$4B,$F7
	.dc.b	$24,$7F,$6B,$F1,$D5,$FD,$8C,$59,$FD,$7C,$DB,$FD,$FC,$DA,$FD,$5C
	.dc.b	$D7,$F7,$B1,$77,$4D,$8C,$36,$7C,$87,$88,$FC,$EF,$C3,$7E,$70,$8F
	.dc.b	$F0,$AF,$4A,$BF,$97,$8B,$8F,$F0,$49,$B8,$FE,$4A,$BF,$1D,$58,$7F
	.dc.b	$80,$CF,$27,$F3,$31,$F9,$02,$3F,$8D,$6B,$FA,$06,$BC,$0D,$10,$D1
	.dc.b	$6B,$F8,$BE,$BF,$CA,$0F,$14,$79,$0F,$E4,$AE,$5F,$82,$43,$F9,$1B
	.dc.b	$97,$E0,$1F,$AE,$85,$FB,$FE,$69,$E9,$FF,$A0,$D7,$B9,$4C,$44,$BE
	.dc.b	$55,$C6,$75,$18,$3F,$77,$5D,$1C,$07,$0F,$F8,$41,$EE,$E9,$AE,$6B
	.dc.b	$C0,$EE,$7B,$B9,$38,$46,$2F,$DB,$1E,$29,$E3,$18,$8F,$14,$EE,$FD
	.dc.b	$51,$C8,$F1,$7A,$5F,$0C,$8F,$97,$A2,$BE,$98,$B7,$D7,$1F,$EC,$93
	.dc.b	$C8,$86,$CE,$46,$23,$84,$F2,$BF,$A9,$76,$BE,$5C,$D7,$E5,$45,$B4
	.dc.b	$D8,$FE,$7B,$C9,$2B,$FA,$0B,$F7,$7B,$71,$1F,$CE,$62,$E3,$E9,$9B
	.dc.b	$8F,$E8,$F9,$55,$8B,$FB,$B5,$C9,$FE,$77,$1F,$CC,$88,$FF,$2F,$6B
	.dc.b	$FD,$09,$AF,$03,$44,$34,$5B,$F9,$9E,$7F,$E5,$4F,$14,$7F,$43,$9C
	.dc.b	$3F,$89,$8C,$DF,$A2,$30,$43,$E0,$1F,$8F,$F3,$FF,$41,$EF,$48,$C1
	.dc.b	$0F,$CB,$39,$7F,$31,$1F,$A7,$31,$87,$E0,$67,$13,$C6,$30,$FE,$76
	.dc.b	$32,$F2,$30,$43,$04,$8E,$2F,$88,$E9,$97,$EE,$7D,$0F,$DC,$12,$3F
	.dc.b	$55,$E5,$57,$F5,$D1,$6B,$F2,$D3,$6B,$95,$7E,$C6,$BF,$81,$4E,$63
	.dc.b	$F7,$71,$5F,$DD,$45,$DD,$36,$30,$D9,$FF,$08,$9F,$E1,$02,$1F,$C1
	.dc.b	$3F,$05,$79,$B8,$8F,$F0,$61,$71,$FD,$C4,$EA,$39,$90,$D9,$CE,$79
	.dc.b	$3F,$E0,$E3,$FC,$0A,$43,$FC,$04,$53,$FC,$2A,$53,$81,$A2,$1A,$2C
	.dc.b	$F9,$3C,$EF,$E9,$0D,$8F,$23,$FE,$0A,$73,$97,$F4,$DB,$56,$27,$FC
	.dc.b	$43,$18,$78,$CF,$EC,$5D,$08,$D3,$FE,$B8,$6D,$77,$8F,$99,$E7,$70
	.dc.b	$97,$C1,$53,$11,$27,$E9,$46,$5E,$24,$3B,$39,$7C,$11,$C5,$C9,$C8
	.dc.b	$E1,$5F,$62,$38,$3F,$6A,$78,$39,$9E,$07,$23,$A9,$F8,$5F,$3B,$D4
	.dc.b	$47,$F1,$1E,$9D,$7E,$14,$59,$FC,$3C,$DC,$7E,$84,$DA,$F9,$F3,$5F
	.dc.b	$99,$17,$14,$D8,$C3,$67,$C8,$78,$8F,$EB,$BF,$41,$FE,$EC,$6F,$99
	.dc.b	$F6,$23,$11,$F5,$FC,$EC,$4C,$8D,$9F,$A3,$8C,$3F,$EF,$49,$E3,$97
	.dc.b	$F5,$96,$E5,$F8,$BA,$DD,$7F,$86,$44,$F8,$1C,$35,$FF,$0F,$35,$FF
	.dc.b	$38,$B5,$E3,$7F,$E7,$08,$C1,$1C,$12,$3C,$07,$27,$0F,$E5,$5F,$C9
	.dc.b	$1E,$CB,$F5,$38,$8F,$E0,$4E,$6F,$CA,$E6,$39,$1F,$18,$FD,$07,$97
	.dc.b	$F9,$DF,$87,$91,$FC,$2E,$3F,$40,$61,$FD,$34,$61,$D1,$FC,$45,$D5
	.dc.b	$FF,$0E,$9D,$4F,$03,$47,$03,$C2,$60,$D1,$82,$47,$F1,$89,$EB,$7F
	.dc.b	$C0,$6B,$F6,$9E,$4C,$BF,$E4,$4F,$EF,$BF,$06,$FF,$70,$6D,$FE,$DE
	.dc.b	$9C,$90,$63,$FE,$1B,$9D,$D7,$39,$FE,$08,$F3,$71,$3F,$CA,$A2,$E1
	.dc.b	$FE,$5F,$AE,$35,$D6,$B4,$FF,$8E,$23,$2F,$43,$53,$C5,$FC,$F7,$2C
	.dc.b	$7F,$46,$FA,$FC,$AC,$47,$E6,$E2,$DF,$A6,$6D,$7E,$52,$6D,$7F,$B1
	.dc.b	$2D,$5C,$A3,$F3,$73,$F9,$52,$3F,$92,$3F,$42,$E1,$FC,$AE,$BF,$E5
	.dc.b	$17,$4D,$73,$1B,$7A,$9A,$79,$61,$F3,$FD,$FF,$AC,$79,$A7,$B9,$F1
	.dc.b	$DC,$BF,$08,$87,$D8,$C7,$E8,$46,$A3,$E2,$1F,$3F,$B3,$F0,$5F,$15
	.dc.b	$23,$04,$39,$1F,$15,$D3,$81,$9F,$F8,$85,$C3,$70,$FC,$44,$F0,$7F
	.dc.b	$15,$5F,$91,$31,$87,$E1,$67,$17,$E1,$AE,$2F,$E1,$63,$2F,$6A,$E2
	.dc.b	$F9,$4E,$15,$F8,$FF,$2B,$F3,$C4,$8F,$C7,$79,$55,$F3,$C5,$C3,$F3
	.dc.b	$D3,$6B,$95,$7E,$56,$BF,$51,$39,$8F,$C4,$45,$7E,$1C,$5C,$53,$63
	.dc.b	$0D,$9F,$E7,$7D,$3F,$CE,$48,$FE,$8F,$C9,$AF,$F3,$58,$B8,$FE,$CE
	.dc.b	$6E,$3F,$CB,$A6,$D7,$E6,$E7,$8B,$FC,$D2,$3F,$97,$11,$FC,$AB,$5F
	.dc.b	$CC,$B5,$E0,$68,$86,$8B,$5F,$C1,$7D,$EF,$E1,$87,$8A,$3F,$A2,$CE
	.dc.b	$1F,$DF,$46,$77,$FF,$17,$47,$E2,$23,$4F,$BC,$FE,$09,$E4,$FF,$17
	.dc.b	$F8,$B9,$18,$21,$F9,$07,$2F,$E4,$A3,$F8,$F1,$8C,$3F,$8D,$4E,$27
	.dc.b	$8C,$61,$FE,$25,$19,$79,$18,$21,$82,$47,$17,$C4,$74,$CB,$FD,$FD
	.dc.b	$F8,$3F,$FB,$4A,$7F,$D8,$8D,$8C,$F9,$1F,$E3,$08,$FF,$69,$CE,$A3
	.dc.b	$04,$AF,$F6,$EC,$EA,$B8,$C6,$27,$4F,$ED,$67,$48,$3E,$17,$0D,$73
	.dc.b	$1E,$3F,$27,$9D,$F8,$E7,$B1,$E2,$8C,$DF,$8A,$3C,$15,$C5,$FF,$79
	.dc.b	$F7,$BC,$3F,$F2,$6B,$63,$31,$F0,$7C,$9C,$B2,$32,$7C,$08,$C3,$FC
	.dc.b	$46,$79,$36,$79,$3C,$0F,$F0,$3C,$69,$FF,$0C,$35,$F8,$81,$C3,$F1
	.dc.b	$DA,$E6,$3F,$A0,$34,$FE,$85,$1A,$24,$69,$B1,$A8,$FB,$DE,$EC,$8F
	.dc.b	$36,$4C,$7C,$77,$EC,$8F,$C3,$D6,$EB,$DC,$70,$AF,$C6,$C6,$67,$94
	.dc.b	$FA,$6F,$F1,$9E,$4C,$8F,$C4,$BF,$8F,$F5,$71,$1F,$87,$8B,$5F,$96
	.dc.b	$9B,$87,$E6,$A6,$D7,$E2,$26,$BF,$8C,$8B,$AA,$6C,$61,$B3,$E2,$3F
	.dc.b	$36,$FE,$87,$C8,$FD,$16,$32,$7F,$3F,$19,$7F,$43,$19,$7F,$3D,$19
	.dc.b	$7F,$41,$19,$C7,$F5,$F3,$97,$F6,$31,$9C,$7F,$6B,$39,$7F,$6F,$19
	.dc.b	$AF,$D2,$CE,$5F,$D3,$C6,$5F,$D5,$C7,$E7,$32,$91,$FA,$07,$F7,$9E
	.dc.b	$3E,$23,$F3,$11,$70,$FE,$07,$37,$0F,$C0,$CD,$AF,$DA,$CF,$17,$F3
	.dc.b	$31,$F8,$D2,$3F,$92,$6B,$FA,$06,$BC,$0D,$10,$D1,$6B,$F8,$AF,$B5
	.dc.b	$FC,$4C,$F1,$47,$E9,$67,$0F,$F1,$28,$CE,$A9,$CC,$FF,$11,$9E,$31
	.dc.b	$FC,$FA,$BF,$2B,$3B,$E5,$FF,$67,$CE,$71,$FD,$9A,$F9,$DC,$CC,$AF
	.dc.b	$E3,$57,$CA,$B8,$47,$F1,$D7,$83,$AB,$E7,$5C,$63,$FC,$31,$3C,$1F
	.dc.b	$45,$E6,$FA,$55,$BF,$1C,$DA,$ED,$35,$E2,$77,$32,$53,$DC,$5B,$F0
	.dc.b	$1E,$67,$C7,$8F,$C2,$F2,$24,$6E,$3E,$29,$87,$D3,$18,$74,$75,$75
	.dc.b	$8F,$C5,$9D,$4F,$03,$47,$03,$C2,$60,$D1,$82,$47,$CF,$9E,$B8,$FD
	.dc.b	$6A,$FE,$76,$CA,$FE,$7E,$FD,$CF,$23,$11,$FC,$DE,$2D,$FD,$99,$B1
	.dc.b	$FC,$80,$B3,$F6,$E3,$8B,$FC,$E2,$3F,$96,$11,$FE,$62,$D7,$F9,$FB
	.dc.b	$5E,$06,$88,$68,$B3,$F8,$BF,$67,$F2,$67,$8A,$3F,$A2,$4E,$1F,$B5
	.dc.b	$19,$BF,$CD,$D7,$19,$C9,$F0,$8F,$C0,$79,$1F,$9E,$23,$F8,$3F,$A2
	.dc.b	$AF,$E5,$22,$CF,$E7,$66,$E1,$FA,$49,$B5,$F9,$F9,$AF,$EC,$62,$E6
	.dc.b	$9B,$18,$6C,$F9,$0F,$11,$F9,$5F,$1F,$F6,$7E,$44,$8C,$10,$FD,$4B
	.dc.b	$97,$F5,$71,$FB,$A3,$18,$FE,$D6,$71,$3C,$63,$0F,$E8,$E3,$2F,$23
	.dc.b	$04,$30,$48,$E2,$F8,$8E,$98,$FF,$7A,$FA,$37,$7F,$F7,$CA,$FE,$1F
	.dc.b	$18,$AF,$DA,$35,$FE,$07,$E4,$6E,$F9,$5D,$9F,$F1,$45,$6E,$F9,$CF
	.dc.b	$F1,$8B,$D5,$2B,$57,$CA,$3B,$D6,$DA,$7F,$B2,$67,$C4,$FF,$84,$8E
	.dc.b	$51,$FE,$0E,$8C,$C7,$33,$46,$8E,$51,$B2,$38,$21,$C2,$39,$BE,$C8
	.dc.b	$F4,$FE,$10,$F8,$7A,$7B,$2B,$D7,$5B,$3F,$8E,$4E,$DB,$3B,$A9,$18
	.dc.b	$21,$FE,$C7,$AE,$6A,$BF,$DD,$EF,$4F,$2B,$C6,$79,$3D,$8B,$19,$21
	.dc.b	$FE,$D1,$94,$6A,$31,$37,$3F,$E9,$F5,$A9,$E3,$18,$9D,$65,$FE,$42
	.dc.b	$AD,$57,$23,$71,$99,$F2,$46,$0F,$F3,$64,$FA,$9E,$07,$96,$FE,$6F
	.dc.b	$91,$F3,$BC,$99,$3F,$2A,$32,$FC,$C8,$CB,$ED,$8C,$BE,$F8,$CE,$3E
	.dc.b	$C9,$CB,$EE,$8C,$E3,$EB,$9C,$BE,$D8,$CD,$7A,$27,$2F,$AA,$32,$FD
	.dc.b	$58,$FC,$14,$48,$FD,$4B,$F8,$8F,$1F,$11,$F7,$62,$DF,$D0,$9B,$7F
	.dc.b	$B5,$9B,$5F,$22,$78,$BF,$A1,$CD,$3F,$A1,$24,$7E,$9D,$AF,$E6,$DA
	.dc.b	$F1,$6B,$DC,$E0,$43,$F1,$1F,$0D,$FA,$03,$C6,$FE,$E6,$30,$FE,$7A
	.dc.b	$33,$3E,$48,$C1,$0F,$80,$7E,$93,$5F,$D5,$91,$FD,$3F,$AF,$5F,$CC
	.dc.b	$45,$CB,$F2,$B3,$72,$FC,$94,$DA,$FC,$6C,$D7,$F5,$11,$75,$4D,$8C
	.dc.b	$36,$7C,$87,$88,$FD,$CF,$8F,$FC,$03,$C7,$91,$82,$1F,$AD,$72,$FE
	.dc.b	$BE,$3F,$83,$98,$C7,$F5,$B3,$89,$E3,$18,$7F,$83,$C6,$5E,$46,$08
	.dc.b	$60,$91,$C5,$F1,$1D,$31,$FE,$7D,$C3,$FC,$54,$91,$FE,$2F,$FC,$27
	.dc.b	$FD,$B5,$A7,$F1,$58,$FE,$21,$3A,$57,$FE,$52,$AD,$CE,$63,$FA,$3B
	.dc.b	$FC,$F2,$33,$3F,$CD,$E3,$F9,$A1,$1C,$10,$FE,$5C,$E5,$A6,$4F,$CF
	.dc.b	$39,$69,$E9,$F8,$5F,$01,$E3,$C3,$DF,$19,$21,$EC,$3F,$AD,$CF,$07
	.dc.b	$37,$EA,$7E,$27,$8F,$EB,$23,$F1,$FD,$7A,$FC,$4B,$FF,$71,$E2,$AC
	.dc.b	$DF,$C7,$BB,$5A,$9B,$61,$6A,$CC,$62,$6C,$E6,$3C,$A4,$3C,$47,$37
	.dc.b	$C4,$F9,$CD,$5F,$D9,$9E,$75,$F5,$FC,$7F,$B1,$E3,$CB,$FC,$24,$65
	.dc.b	$FC,$34,$65,$EE,$7A,$23,$4E,$30,$FF,$6D,$3C,$0E,$64,$35,$3C,$F0
	.dc.b	$FC,$65,$72,$7F,$1D,$19,$C7,$E6,$CE,$5F,$9F,$19,$AF,$81,$3F,$E0
	.dc.b	$CF,$A6,$7C,$3D,$24,$FE,$7F,$C8,$AF,$AA,$76,$FF,$A4,$5A,$FA,$E3
	.dc.b	$78,$FF,$AC,$A7,$2D,$AF,$D0,$4F,$4A,$FF,$0A,$57,$26,$CF,$F8,$49
	.dc.b	$FC,$04,$65,$FC,$10,$70,$EA,$38,$39,$6C,$6A,$30,$75,$70,$48,$D9
	.dc.b	$FB,$5F,$83,$93,$EF,$8F,$6E,$5F,$9A,$9C,$CF,$C5,$9C,$BF,$8D,$8C
	.dc.b	$E1,$F0,$63,$F8,$57,$93,$FC,$04,$8F,$F0,$DF,$62,$BF,$85,$8B,$3F
	.dc.b	$C2,$E6,$E1,$FA,$A9,$B5,$F9,$B9,$AF,$E9,$A2,$EA,$9B,$18,$6C,$F9
	.dc.b	$0F,$11,$FE,$CD,$FC,$2B,$FD,$B4,$7F,$B2,$F0,$FF,$76,$B7,$0F,$A6
	.dc.b	$EB,$C9,$FF,$2B,$B9,$AF,$DC,$35,$3B,$7B,$79,$5C,$0E,$33,$23,$24
	.dc.b	$3F,$57,$3C,$5B,$3C,$5F,$44,$F1,$72,$71,$BF,$4C,$7E,$87,$C9,$93
	.dc.b	$F9,$B7,$D3,$E4,$62,$3F,$41,$16,$BE,$0C,$DC,$7F,$27,$BA,$B1,$7F
	.dc.b	$02,$B9,$3F,$9E,$8F,$C9,$91,$FC,$93,$5F,$F3,$8C,$F3,$72,$43,$31
	.dc.b	$B7,$A1,$9B,$FA,$BF,$0B,$F4,$8F,$1C,$7E,$9E,$70,$FD,$D8,$CD,$77
	.dc.b	$8F,$9B,$1A,$7E,$09,$F8,$39,$FB,$A4,$7F,$0D,$EC,$D7,$EA,$C5,$BF
	.dc.b	$C3,$CD,$C7,$F0,$F3,$6B,$F0,$B3,$5F,$C2,$C5,$D1,$36,$30,$D9,$F2
	.dc.b	$1E,$23,$F3,$1E,$4F,$E4,$BC,$99,$18,$21,$F8,$A7,$2F,$E3,$23,$F3
	.dc.b	$26,$31,$FC,$FC,$E2,$78,$C6,$1F,$D3,$46,$5E,$46,$08,$60,$91,$C5
	.dc.b	$F1,$1D,$31,$FC,$5E,$DF,$CD,$49,$1F,$B3,$F2,$6B,$FC,$CE,$2D,$7E
	.dc.b	$D2,$6D,$72,$AF,$E6,$D5,$FB,$49,$E3,$1F,$CC,$23,$F9,$29,$1F,$E4
	.dc.b	$8D,$7F,$26,$D7,$81,$A2,$1A,$2D,$7F,$06,$F8,$3F,$E0,$67,$92,$3F
	.dc.b	$9D,$CE,$1F,$E1,$D1,$97,$F9,$E4,$65,$FD,$FC,$67,$47,$93,$FC,$38
	.dc.b	$93,$FE,$05,$B6,$3F,$D1,$BC,$5F,$F7,$04,$7F,$18,$AD,$D7,$F0,$A7
	.dc.b	$FD,$DB,$3A,$AF,$E3,$71,$AA,$E1,$5D,$67,$A3,$97,$6F,$F2,$79,$E6
	.dc.b	$E4,$FE,$42,$D7,$FC,$FC,$6C,$87,$8C,$EC,$7F,$87,$EB,$C8,$DF,$27
	.dc.b	$CF,$E4,$C8,$D9,$FE,$9C,$8F,$59,$97,$FC,$23,$3F,$ED,$0C,$3E,$2C
	.dc.b	$73,$38,$C7,$FB,$36,$33,$1F,$EA,$6A,$E5,$18,$9D,$61,$F3,$6F,$2F
	.dc.b	$CF,$AF,$0D,$F4,$9C,$BE,$B8,$CB,$E4,$8C,$BD,$CE,$07,$B3,$EA,$1D
	.dc.b	$3F,$E0,$0C,$7C,$B1,$89,$D4,$FD,$AF,$2B,$58,$FF,$81,$AB,$55,$FE
	.dc.b	$32,$6D,$F2,$BE,$13,$C8,$86,$DC,$B4,$FF,$0A,$4F,$47,$FC,$26,$1C
	.dc.b	$3E,$18,$FF,$09,$C6,$23,$FC,$B4,$D7,$84,$68,$8F,$66,$BD,$8D,$99
	.dc.b	$3E,$CF,$C1,$7E,$68,$F8,$2F,$68,$AF,$03,$F0,$71,$B7,$DF,$5F,$0D
	.dc.b	$E6,$43,$E0,$1F,$8A,$F2,$FF,$4E,$47,$F1,$3E,$BD,$74,$7E,$32,$3F
	.dc.b	$5B,$19,$9F,$C5,$45,$C7,$F6,$D3,$6B,$F4,$13,$5F,$F7,$BC,$5D,$53
	.dc.b	$63,$0D,$9F,$CD,$F9,$5F,$BC,$F2,$A4,$60,$87,$E9,$1C,$BF,$A6,$8F
	.dc.b	$E0,$66,$31,$FD,$04,$E2,$78,$C6,$1F,$D4,$C6,$5E,$46,$08,$60,$91
	.dc.b	$C5,$F1,$1D,$31,$FC,$C7,$C2,$FD,$8D,$4D,$3F,$8B,$F9,$76,$BF,$4D
	.dc.b	$15,$FD,$FC,$5B,$FD,$3C,$DB,$FE,$49,$3B,$AF,$CC,$DF,$67,$F3,$B5
	.dc.b	$F9,$82,$38,$21,$F9,$38,$E0,$D8,$DB,$4D,$9D,$9F,$E5,$FF,$03,$23
	.dc.b	$D2,$43,$F8,$BB,$97,$F7,$11,$E9,$7F,$92,$46,$70,$F6,$C7,$F2,$2F
	.dc.b	$C2,$DE,$BF,$E8,$1B,$19,$8C,$7B,$19,$64,$64,$E5,$18,$7F,$D7,$33
	.dc.b	$C9,$B3,$C9,$E0,$60,$B5,$FC,$AA,$55,$CE,$31,$57,$3F,$F4,$DD,$6A
	.dc.b	$F8,$C6,$27,$51,$FE,$E1,$9D,$3E,$4F,$2F,$DD,$E5,$C8,$F8,$51,$F0
	.dc.b	$4C,$3E,$A8,$C3,$A3,$E4,$3A,$C3,$D1,$3A,$9E,$06,$8E,$07,$84,$C1
	.dc.b	$A3,$04,$8F,$85,$3D,$70,$F7,$D7,$E7,$70,$97,$FC,$FD,$FB,$1E,$56
	.dc.b	$23,$F9,$C4,$5C,$3E,$64,$DB,$FE,$7D,$36,$BE,$34,$F1,$7F,$3B,$1F
	.dc.b	$CC,$88,$FF,$3A,$6B,$F9,$B6,$BC,$0D,$10,$D1,$6B,$EF,$7C,$17,$E3
	.dc.b	$CF,$14,$7F,$43,$9C,$3F,$95,$8C,$ED,$F8,$BA,$E3,$39,$3E,$11,$F9
	.dc.b	$BF,$2F,$F3,$44,$7F,$3D,$EC,$57,$F2,$B1,$67,$F2,$13,$72,$FC,$6C
	.dc.b	$DA,$FC,$54,$D5,$AA,$70,$A6,$C6,$1B,$3E,$53,$C6,$7E,$C7,$CA,$FD
	.dc.b	$87,$95,$23,$04,$3F,$56,$E5,$FD,$6C,$7E,$DC,$C6,$3F,$A9,$9C,$4F
	.dc.b	$18,$C3,$FB,$F8,$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$26,$76,$B9
	.dc.b	$7F,$B4,$B6,$3F,$C5,$ED,$55,$FE,$28,$FC,$31,$E5,$F0,$B9,$5C,$1F
	.dc.b	$EE,$73,$A8,$C1,$0E,$78,$78,$D6,$1D,$F0,$ED,$87,$49,$FF,$02,$4F
	.dc.b	$8D,$FF,$09,$B5,$E4,$D7,$DB,$87,$F8,$2D,$A5,$8D,$3E,$08,$D3,$4D
	.dc.b	$38,$8D,$91,$F6,$FC,$14,$9F,$65,$FF,$93,$2B,$D5,$87,$B7,$15,$EE
	.dc.b	$FE,$09,$E4,$29,$E1,$2C,$70,$7F,$CC,$2D,$77,$5D,$9F,$B9,$0F,$33
	.dc.b	$E8,$92,$38,$FB,$35,$EB,$17,$1F,$B9,$37,$1F,$44,$DA,$EB,$35,$5F
	.dc.b	$77,$CB,$3F,$80,$9D,$4E,$67,$B9,$E2,$36,$43,$27,$C0,$8F,$C4,$F9
	.dc.b	$7F,$8A,$F2,$E4,$FC,$28,$CB,$F1,$23,$2F,$C8,$8C,$BF,$2A,$33,$8F
	.dc.b	$E6,$67,$2F,$E6,$E3,$38,$FD,$79,$CB,$F6,$63,$35,$F3,$A7,$2F,$D0
	.dc.b	$8C,$BF,$12,$3F,$57,$E0,$C8,$FD,$5B,$FA,$4F,$2F,$11,$F6,$A2,$E1
	.dc.b	$FA,$D9,$B8,$7E,$72,$6D,$7D,$E9,$E2,$FE,$8A,$3F,$2C,$47,$F3,$CD
	.dc.b	$7F,$52,$D7,$81,$A2,$1A,$2D,$7E,$FF,$E0,$BF,$76,$78,$A3,$F6,$F3
	.dc.b	$87,$F2,$B1,$9D,$53,$99,$FC,$DC,$F1,$D1,$E5,$FF,$06,$24,$FE,$7B
	.dc.b	$D8,$AF,$E8,$A2,$E3,$FC,$46,$6E,$7F,$AD,$9B,$5F,$A2,$9A,$FE,$5A
	.dc.b	$2D,$A6,$C6,$1B,$3F,$AE,$8F,$E3,$31,$FC,$6B,$C9,$FF,$46,$FE,$1B
	.dc.b	$FE,$D6,$A3,$6F,$FD,$EF,$1F,$B6,$9D,$B0,$C1,$0F,$DF,$CF,$33,$94
	.dc.b	$77,$8F,$F6,$FC,$65,$C1,$FC,$12,$3A,$3C,$8C,$E3,$FC,$D6,$32,$FF
	.dc.b	$B7,$A7,$AC,$FF,$3C,$F9,$7F,$E7,$E4,$7B,$F9,$75,$FE,$1B,$16,$B9
	.dc.b	$4D,$CF,$F9,$7C,$DA,$FE,$99,$7F,$EE,$87,$FA,$6C,$7F,$29,$23,$FC
	.dc.b	$C5,$AF,$F3,$C6,$BF,$CF,$1A,$F6,$38,$10,$E0,$7C,$6F,$82,$FA,$A7
	.dc.b	$8D,$FE,$97,$18,$74,$7C,$12,$1F,$4A,$34,$FC,$03,$E8,$F9,$9F,$80
	.dc.b	$23,$F5,$3C,$BA,$FD,$A8,$B9,$7C,$F9,$B8,$FF,$5D,$C5,$58,$BF,$8F
	.dc.b	$55,$FC,$5C,$59,$4D,$8C,$36,$7C,$A7,$8C,$FC,$3F,$95,$F9,$6F,$2E
	.dc.b	$46,$08,$7E,$41,$CB,$F9,$28,$FC,$E9,$8C,$7F,$11,$38,$9E,$31,$87
	.dc.b	$F2,$B1,$97,$91,$82,$18,$24,$71,$7C,$47,$4C,$7F,$C0,$5F,$59,$FC
	.dc.b	$12,$3F,$C2,$D2,$9C,$F0,$FF,$BF,$AB,$55,$FA,$AF,$33,$13,$23,$8D
	.dc.b	$70,$7F,$57,$39,$7F,$B7,$D7,$2B,$ED,$1F,$E2,$58,$C5,$7D,$9A,$E1
	.dc.b	$3F,$E1,$71,$3E,$27,$0D,$7C,$06,$CD,$DF,$F8,$3A,$35,$1D,$E3,$04
	.dc.b	$7F,$C3,$11,$A7,$89,$88,$D3,$FC,$2B,$E0,$7F,$8A,$9E,$DC,$3D,$71
	.dc.b	$C0,$D1,$FC,$32,$35,$87,$9A,$7F,$83,$47,$29,$F9,$27,$94,$FE,$27
	.dc.b	$E6,$FF,$22,$23,$FC,$5B,$DB,$AF,$F1,$C8,$B5,$FC,$46,$6D,$FE,$8E
	.dc.b	$6D,$7F,$08,$9A,$FF,$2C,$8B,$BA,$6C,$61,$B3,$E4,$3C,$47,$A0,$75
	.dc.b	$7F,$A4,$6F,$FC,$9C,$FF,$B3,$30,$F8,$AD,$8C,$F9,$1F,$E5,$39,$CB
	.dc.b	$63,$8F,$99,$6F,$FC,$D5,$7C,$5B,$59,$7C,$2D,$8D,$47,$92,$71,$32
	.dc.b	$3D,$13,$A2,$46,$63,$FC,$8B,$19,$9C,$5F,$9A,$78,$C7,$73,$37,$FA
	.dc.b	$CE,$DF,$D4,$12,$7B,$F9,$75,$FD,$1C,$5B,$F8,$B3,$6B,$95,$7C,$DD
	.dc.b	$D7,$ED,$AE,$53,$FA,$28,$FC,$F1,$1F,$CD,$B5,$FD,$24,$F2,$72,$43
	.dc.b	$31,$B7,$A1,$9B,$FB,$3E,$F7,$CD,$3C,$D1,$FB,$99,$C3,$F3,$23,$2F
	.dc.b	$EE,$23,$2F,$CC,$8C,$E4,$F3,$3E,$C7,$B3,$2B,$EA,$C7,$DB,$30,$FD
	.dc.b	$48,$C3,$A3,$ED,$BA,$C3,$F1,$F3,$A9,$E0,$68,$E0,$7D,$E7,$86,$7F
	.dc.b	$98,$AE,$15,$CA,$B8,$1C,$8E,$07,$E2,$BC,$AF,$C8,$11,$FC,$6F,$95
	.dc.b	$5F,$C5,$C5,$9F,$D1,$CD,$BF,$D3,$CD,$AF,$CF,$CD,$7F,$5B,$17,$1F
	.dc.b	$F7,$54,$78,$48,$76,$3C,$26,$2F,$F6,$33,$C6,$BF,$9E,$78,$D2,$BF
	.dc.b	$1C,$FE,$A7,$CA,$C4,$7F,$38,$8B,$7F,$BA,$9B,$8F,$F4,$A9,$B5,$FB
	.dc.b	$69,$E2,$FF,$36,$8F,$E6,$44,$7F,$97,$B5,$FC,$AB,$5E,$06,$88,$68
	.dc.b	$B5,$FC,$33,$E0,$3F,$80,$1E,$28,$FE,$8F,$38,$7F,$7F,$19,$DF,$FD
	.dc.b	$79,$5C,$67,$27,$C2,$3F,$8E,$7E,$1D,$FE,$EB,$7F,$D8,$EB,$D6,$FE
	.dc.b	$20,$D8,$DF,$99,$83,$33,$23,$24,$39,$B9,$7A,$C7,$17,$C3,$19,$AF
	.dc.b	$E9,$B5,$E3,$6D,$67,$3F,$F7,$E4,$F8,$EF,$99,$DE,$B8,$DE,$E0,$27
	.dc.b	$A5,$72,$F2,$FD,$A4,$9F,$87,$E5,$D7,$E1,$C5,$C3,$CB,$36,$B9,$57
	.dc.b	$CA,$AF,$8F,$39,$8F,$0D,$FF,$B5,$EF,$DB,$5A,$9E,$04,$3B,$9E,$23
	.dc.b	$11,$C1,$B7,$F0,$AF,$FA,$27,$D7,$3E,$B2,$92,$BE,$EF,$15,$58,$7F
	.dc.b	$7A,$AE,$11,$FD,$75,$AF,$E0,$A7,$84,$EE,$AE,$1F,$4A,$7A,$57,$F8
	.dc.b	$66,$F8,$B6,$BF,$C2,$CF,$D2,$8C,$BF,$50,$38,$75,$1C,$1C,$B6,$35
	.dc.b	$18,$3A,$B8,$24,$6C,$FC,$CF,$C1,$C9,$F7,$47,$A9,$FA,$B1,$97,$E2
	.dc.b	$DF,$96,$BE,$1E,$1F,$06,$7F,$4F,$5B,$8C,$47,$EA,$3C,$EF,$D3,$FB
	.dc.b	$52,$3F,$5B,$1F,$AA,$30,$FE,$BA,$30,$E8,$FD,$9B,$A9,$FD,$1C,$EA
	.dc.b	$78,$1A,$38,$1E,$13,$06,$8C,$12,$3F,$5B,$3D,$70,$FD,$5D,$7E,$FB
	.dc.b	$C1,$97,$FC,$11,$FE,$03,$E6,$62,$3F,$82,$C5,$C3,$F6,$F3,$6F,$F6
	.dc.b	$B3,$6B,$F4,$33,$5F,$D9,$45,$DF,$FD,$D1,$1D,$C8,$76,$3F,$8A,$CE
	.dc.b	$E7,$83,$FA,$0E,$99,$3F,$99,$7F,$8B,$F9,$78,$8F,$B6,$7E,$46,$31
	.dc.b	$8F,$F0,$E9,$B8,$FD,$F9,$B5,$FB,$89,$E4,$FE,$76,$3F,$24,$47,$F2
	.dc.b	$2D,$7F,$CE,$9A,$FF,$7F,$72,$43,$31,$B7,$A1,$9B,$FF,$36,$FE,$1E
	.dc.b	$F8,$0F,$F2,$03,$67,$DC,$43,$DA,$1D,$38,$6C,$FE,$FE,$BD,$B3,$BA
	.dc.b	$EA,$62,$BF,$DD,$D1,$CF,$D1,$F0,$9E,$2F,$3A,$47,$F9,$8E,$53,$A8
	.dc.b	$C4,$E9,$F1,$46,$5F,$24,$67,$0F,$7C,$E5,$F8,$31,$9C,$FF,$83,$5F
	.dc.b	$6B,$F0,$5E,$5F,$E7,$B1,$E8,$34,$F8,$0F,$19,$E7,$F3,$3E,$59,$1F
	.dc.b	$57,$99,$5F,$91,$17,$1F,$A7,$37,$0F,$1C,$DA,$F0,$CD,$7E,$94,$5B
	.dc.b	$FF,$77,$C7,$84,$87,$63,$C2,$62,$FF,$03,$3C,$6B,$F9,$EF,$89,$2B
	.dc.b	$F9,$B3,$F8,$4F,$33,$11,$F8,$2D,$BF,$D0,$67,$FA,$13,$14,$EB,$1F
	.dc.b	$E8,$17,$72,$F9,$B3,$CB,$0F,$E7,$53,$FC,$BC,$8F,$F2,$E6,$BF,$9E
	.dc.b	$6B,$C0,$D1,$0D,$16,$BF,$3B,$F0,$9F,$9C,$3C,$51,$FD,$16,$70,$FE
	.dc.b	$36,$32,$FF,$46,$8C,$BF,$9D,$8C,$E0,$F3,$3F,$4F,$E9,$CA,$FC,$B4
	.dc.b	$7E,$A8,$C3,$FA,$58,$C3,$A3,$F3,$8E,$B0,$FD,$9C,$EA,$78,$1A,$38
	.dc.b	$1F,$AF,$78,$65,$F9,$8A,$E1,$5C,$AB,$81,$C8,$E0,$7E,$AF,$CB,$FD
	.dc.b	$F9,$1F,$D7,$79,$75,$FD,$E4,$59,$FE,$0F,$36,$FF,$4F,$36,$BF,$4B
	.dc.b	$35,$FE,$0D,$17,$1F,$F7,$14,$78,$48,$76,$3C,$26,$2F,$F8,$AC,$F1
	.dc.b	$AF,$F6,$A7,$D8,$3C,$EA,$4B,$FF,$74,$7E,$21,$FE,$E7,$7F,$C6,$73
	.dc.b	$B7,$FD,$D8,$D7,$F8,$DB,$70,$65,$FE,$07,$AE,$54,$AD,$4F,$F8,$2D
	.dc.b	$B1,$DA,$FF,$DD,$15,$C2,$7B,$4F,$F8,$44,$E5,$3F,$E0,$F8,$CC,$73
	.dc.b	$8C,$11,$FF,$0D,$C6,$9E,$C6,$23,$4F,$53,$A9,$92,$1F,$CE,$25,$F0
	.dc.b	$9E,$63,$E4,$47,$C2,$9C,$3F,$2A,$33,$87,$C4,$8F,$09,$CE,$36,$F1
	.dc.b	$3E,$8F,$9D,$F0,$BD,$B9,$18,$21,$E3,$72,$F9,$63,$E4,$98,$AF,$04
	.dc.b	$30,$56,$A3,$9C,$EA,$73,$1F,$82,$9D,$3D,$4C,$10,$C1,$27,$94,$78
	.dc.b	$72,$FC,$15,$FD,$CE,$99,$7F,$8E,$7E,$FF,$9B,$88,$F9,$D1,$70,$F2
	.dc.b	$4D,$BF,$C6,$4D,$AF,$F4,$3E,$2A,$C4,$7B,$E6,$CE,$63,$C0,$43,$B9
	.dc.b	$FA,$09,$DC,$F0,$7F,$4F,$F5,$32,$7F,$42,$FE,$5B,$CE,$C4,$7E,$8A
	.dc.b	$2E,$3F,$2E,$6E,$3F,$9F,$9B,$5E,$89,$E2,$FE,$77,$64,$FE,$74,$91
	.dc.b	$F9,$A6,$BF,$97,$8D,$B9,$21,$98,$DB,$D5,$E3,$7F,$7F,$DF,$FD,$79
	.dc.b	$E4,$9F,$DB,$4E,$1F,$E0,$71,$9A,$EF,$1F,$A3,$8D,$3F,$04,$FE,$13
	.dc.b	$E7,$7E,$EB,$DA,$91,$82,$1F,$AB,$72,$FE,$B6,$3F,$80,$98,$C3,$F2
	.dc.b	$93,$89,$E3,$18,$7F,$8B,$C6,$5E,$46,$08,$60,$91,$C5,$F1,$1D,$32
	.dc.b	$FF,$4F,$7B,$FF,$ED,$08,$FF,$69,$B5,$34,$FF,$6E,$4F,$EB,$9C,$79
	.dc.b	$BF,$EA,$78,$C3,$6B,$AF,$9B,$6F,$AC,$D7,$AB,$6F,$B4,$D8,$93,$FE
	.dc.b	$E6,$BE,$AD,$BE,$AD,$3A,$94,$EA,$5B,$EA,$DA,$EA,$FF,$95,$A3,$2F
	.dc.b	$F3,$EF,$07,$FA,$49,$1E,$5E,$6D,$74,$7F,$46,$8F,$54,$67,$0F,$E2
	.dc.b	$7A,$FF,$96,$0F,$F2,$CD,$FF,$BB,$F1,$FE,$89,$7F,$CF,$49,$87,$F3
	.dc.b	$A6,$BF,$A1,$6B,$FA,$96,$BC,$8E,$04,$38,$1F,$07,$E0,$BE,$71,$E5
	.dc.b	$9E,$11,$F0,$63,$2F,$AC,$86,$4E,$46,$30,$FC,$17,$9B,$F6,$7D,$49
	.dc.b	$3F,$3E,$3E,$E1,$87,$E7,$46,$1D,$1F,$55,$D4,$F4,$8C,$9C,$0D,$18
	.dc.b	$74,$E4,$8E,$09,$7F,$9E,$9E,$AF,$E8,$2B,$F2,$1E,$5C,$BF,$C3,$3F
	.dc.b	$93,$F3,$31,$1F,$85,$8B,$87,$E8,$66,$DF,$EA,$A6,$D7,$EA,$A6,$BF
	.dc.b	$AA,$8B,$9A,$6C,$61,$B3,$E2,$3F,$30,$FE,$9F,$AB,$FD,$A7,$1F,$EA
	.dc.b	$C6,$E1,$FD,$81,$B8,$7E,$F3,$CB,$E1,$E5,$CB,$FD,$25,$69,$FD,$FC
	.dc.b	$E5,$E9,$36,$FF,$7D,$39,$9F,$F0,$A5,$78,$9B,$5F,$E0,$B4,$7F,$82
	.dc.b	$1C,$34,$D4,$71,$76,$64,$FF,$0A,$39,$79,$B5,$D1,$E0,$39,$C7,$68
	.dc.b	$F6,$47,$F1,$8F,$82,$FE,$18,$7A,$F0,$EF,$3A,$CD,$3A,$AF,$8B,$7B
	.dc.b	$7F,$91,$C6,$5F,$8A,$7F,$0B,$F3,$3F,$8C,$11,$FE,$21,$F8,$8F,$F8
	.dc.b	$E7,$3F,$E4,$D3,$99,$FE,$1C,$FF,$21,$9D,$65,$FC,$CA,$35,$5C,$2B
	.dc.b	$99,$89,$D3,$E1,$32,$53,$A9,$63,$FC,$47,$5D,$A3,$F9,$ED,$73,$69
	.dc.b	$C8,$FF,$7B,$C7,$FB,$B8,$FF,$6B,$CF,$2F,$33,$D1,$C5,$2B,$DF,$1E
	.dc.b	$B3,$0F,$28,$C3,$A3,$E2,$BA,$9E,$91,$93,$81,$A3,$0E,$9C,$91,$C1
	.dc.b	$30,$FC,$5D,$AB,$F5,$1E,$7C,$BF,$D0,$BF,$47,$CD,$C4,$7E,$AE,$2E
	.dc.b	$1F,$22,$6E,$1F,$88,$9B,$5F,$9E,$C1,$5C,$A3,$F4,$B3,$F9,$32,$3F
	.dc.b	$A0,$6B,$F9,$E6,$BC,$0D,$10,$D1,$6F,$ED,$FB,$DF,$89,$3C,$71,$FE
	.dc.b	$76,$9C,$3F,$92,$8C,$EE,$9C,$CF,$DE,$9E,$38,$7F,$BC,$A3,$F3,$3E
	.dc.b	$67,$E5,$4F,$CB,$7B,$56,$32,$43,$F0,$EE,$5F,$C4,$C5,$C7,$F2,$33
	.dc.b	$73,$FA,$B3,$6B,$E5,$4D,$57,$E3,$79,$9F,$AC,$8D,$46,$6F,$F3,$B5
	.dc.b	$E6,$AF,$D8,$F9,$7F,$AD,$F2,$E4,$60,$87,$E7,$9C,$BF,$A0,$8F,$DA
	.dc.b	$18,$AF,$D4,$CE,$27,$8C,$61,$FD,$FC,$65,$E4,$60,$86,$09,$1C,$5F
	.dc.b	$11,$D3,$1F,$E2,$BF,$29,$FC,$FC,$91,$FC,$33,$CB,$AF,$F4,$18,$B5
	.dc.b	$FB,$E9,$B5,$CA,$BF,$9F,$D7,$F0,$F9,$E3,$1F,$CE,$A3,$F9,$41,$1F
	.dc.b	$E6,$AD,$7F,$99,$B5,$E0,$68,$86,$8B,$5F,$C4,$3E,$0F,$F8,$F1,$E4
	.dc.b	$8F,$E7,$73,$87,$F9,$3C,$65,$FF,$67,$C6,$5F,$E2,$71,$9F,$C4,$BF
	.dc.b	$DD,$A4,$F7,$3F,$C9,$4B,$D1,$B1,$2B,$FC,$85,$3C,$27,$10,$C3,$FD
	.dc.b	$8F,$3C,$E3,$73,$A9,$CD,$5C,$BF,$AB,$DF,$7B,$EB,$39,$7F,$B0,$C7
	.dc.b	$77,$24,$3C,$13,$CA,$7D,$37,$EA,$F3,$3D,$7E,$6C,$BF,$4C,$65,$F5
	.dc.b	$C6,$5E,$67,$82,$34,$E0,$61,$FE,$DE,$70,$DC,$7E,$0C,$F0,$7E,$1D
	.dc.b	$75,$73,$5E,$23,$C2,$43,$9C,$F1,$C7,$E2,$D7,$17,$E3,$C7,$F9,$FF
	.dc.b	$ED,$1E,$45,$25,$FD,$6F,$36,$BE,$99,$DB,$FE,$7B,$6B,$EA,$8D,$E3
	.dc.b	$FF,$00,$27,$2D,$AF,$C2,$CF,$4A,$FF,$07,$8A,$E6,$E1,$AF,$F8,$51
	.dc.b	$AF,$F8,$71,$AF,$19,$FF,$0E,$C6,$08,$E0,$91,$FE,$25,$8C,$BF,$94
	.dc.b	$F8,$5F,$CC,$9E,$AC,$7C,$15,$89,$D1,$F0,$5D,$3F,$8E,$9D,$3F,$1E
	.dc.b	$3F,$47,$E6,$C8,$FD,$0B,$FA,$6F,$77,$11,$F8,$28,$B9,$7E,$8A,$6E
	.dc.b	$1F,$9E,$9B,$5F,$9C,$9A,$FE,$32,$2E,$A9,$B1,$86,$CF,$88,$FD,$9B
	.dc.b	$FB,$8C,$3F,$73,$E6,$49,$FD,$BC,$65,$FD,$CC,$65,$FC,$CC,$65,$FC
	.dc.b	$DC,$67,$1F,$DD,$CE,$5F,$DE,$C6,$71,$FD,$CC,$E5,$FD,$DC,$66,$BF
	.dc.b	$73,$39,$7F,$77,$19,$7F,$57,$1F,$9B,$F3,$E4,$7E,$55,$FF,$6E,$7E
	.dc.b	$27,$FE,$E9,$D5,$1A,$9F,$CE,$3A,$F3,$71,$97,$E6,$62,$E5,$FE,$64
	.dc.b	$C3,$94,$7E,$38,$97,$F8,$B6,$BF,$9C,$3F,$38,$E1,$FC,$E3,$5F,$CE
	.dc.b	$35,$FC,$E3,$CA,$95,$E1,$F7,$BD,$67,$92,$B6,$7F,$BB,$6D,$1A,$8F
	.dc.b	$84,$43,$F8,$94,$69,$E1,$1C,$D4,$7F,$B8,$DF,$F4,$4A,$F7,$3E,$59
	.dc.b	$F3,$3C,$F9,$3E,$18,$CB,$E3,$8C,$BE,$08,$CB,$E2,$8C,$E1,$EE,$8E
	.dc.b	$F3,$A8,$C4,$79,$27,$78,$7D,$3A,$D3,$F5,$63,$33,$C8,$F2,$1E,$33
	.dc.b	$E3,$79,$9F,$84,$23,$F2,$7C,$DA,$FC,$68,$B8,$78,$B1,$56,$30,$FC
	.dc.b	$25,$5B,$F8,$F3,$5F,$97,$16,$B8,$C7,$88,$87,$73,$C4,$62,$FC,$13
	.dc.b	$C6,$BF,$9C,$78,$72,$BF,$99,$3F,$9E,$F3,$71,$1F,$CD,$E2,$DF,$A6
	.dc.b	$6E,$3F,$CA,$E6,$D7,$DE,$9E,$2F,$F3,$18,$FE,$56,$47,$F9,$43,$5F
	.dc.b	$CF,$6E,$8C,$47,$13,$64,$36,$F1,$BF,$99,$F0,$9F,$B6,$3C,$73,$FD
	.dc.b	$0E,$70,$FE,$76,$33,$5D,$E3,$04,$3E,$01,$FA,$3F,$37,$F8,$0F,$99
	.dc.b	$23,$04,$3F,$46,$E5,$FD,$2C,$7F,$08,$31,$87,$E0,$67,$13,$C6,$30
	.dc.b	$FE,$C6,$32,$F2,$30,$43,$04,$8E,$2F,$88,$E9,$97,$F0,$FF,$2F,$F8
	.dc.b	$C9,$23,$FD,$F7,$3F,$A4,$F1,$FF,$D9,$51,$FB,$89,$D3,$FB,$E7,$FD
	.dc.b	$F7,$3B,$8C,$BF,$9B,$8B,$5F,$F0,$02,$B9,$EE,$9C,$DF,$F0,$27,$9D
	.dc.b	$6E,$32,$78,$4F,$E2,$51,$A4,$2B,$AC,$70,$AE,$D3,$EC,$AE,$8D,$38
	.dc.b	$FD,$B3,$E7,$7A,$42,$30,$EC,$C4,$6B,$0F,$F6,$84,$66,$76,$FF,$8C
	.dc.b	$DB,$8F,$F0,$E6,$DF,$C3,$F3,$3B,$C6,$2B,$C7,$86,$0E,$D3,$63,$FA
	.dc.b	$01,$E2,$2C,$F8,$1F,$F0,$FB,$5F,$8A,$6C,$FE,$90,$E5,$EE,$72,$9E
	.dc.b	$AE,$CE,$6E,$23,$84,$64,$EA,$FA,$5F,$A5,$F0,$9F,$14,$F4,$3F,$0E
	.dc.b	$70,$FC,$BC,$53,$AC,$7E,$35,$6A,$BF,$CB,$95,$A7,$E7,$47,$D1,$F3
	.dc.b	$64,$73,$7E,$A7,$BB,$88,$E3,$16,$FF,$39,$37,$3F,$CA,$4D,$AF,$C5
	.dc.b	$E0,$AC,$47,$E3,$66,$CE,$63,$C0,$43,$B9,$ED,$9D,$CF,$07,$F1,$1E
	.dc.b	$5F,$E2,$7C,$D9,$5F,$0E,$32,$FC,$58,$CB,$F0,$A3,$2F,$C4,$8C,$E3
	.dc.b	$FB,$79,$CB,$FB,$98,$CE,$3F,$B1,$9C,$BF,$B3,$8C,$D7,$C0,$9C,$BF
	.dc.b	$0A,$32,$FE,$B6,$3F,$2E,$47,$F4,$0F,$E8,$FC,$CC,$47,$E8,$62,$DF
	.dc.b	$F0,$B9,$B7,$F8,$89,$B5,$F4,$27,$8B,$F9,$A8,$FC,$81,$1F,$CF,$B5
	.dc.b	$FC,$DB,$5E,$06,$88,$68,$B5,$FC,$77,$D9,$FD,$D1,$E2,$8F,$E0,$13
	.dc.b	$87,$F2,$D1,$9D,$93,$99,$FC,$AC,$F1,$D1,$E6,$7F,$05,$F6,$A5,$7E
	.dc.b	$7E,$3F,$84,$18,$7F,$3B,$18,$74,$7E,$95,$D6,$3F,$CC,$E7,$53,$C0
	.dc.b	$D1,$C0,$FD,$3B,$C3,$3F,$E6,$15,$C2,$B9,$57,$03,$91,$C0,$FE,$35
	.dc.b	$B7,$EF,$C8,$FF,$0E,$6B,$FC,$29,$AF,$F1,$06,$BF,$C1,$1A,$FF,$2E
	.dc.b	$7F,$86,$FE,$28,$F3,$3F,$97,$BF,$EE,$D9,$FF,$6D,$36,$BF,$C6,$AE
	.dc.b	$E3,$F9,$93,$5D,$F9,$9C,$8E,$75,$23,$71,$97,$8C,$65,$E2,$68,$C9
	.dc.b	$0C,$92,$BC,$71,$C5,$F2,$CF,$59,$F4,$CF,$F3,$CF,$4B,$F9,$41,$23
	.dc.b	$F9,$BB,$5F,$E6,$6D,$7F,$9C,$B5,$FE,$5A,$D7,$FA,$BB,$FC,$CF,$96
	.dc.b	$47,$F4,$26,$BF,$CF,$9A,$FF,$44,$6B,$FC,$E1,$AF,$F6,$26,$BF,$D8
	.dc.b	$7E,$1B,$FA,$01,$1F,$E9,$ED,$7F,$A5,$B5,$FE,$A2,$D7,$FA,$2B,$5F
	.dc.b	$ED,$EF,$F4,$BF,$80,$91,$FD,$69,$AF,$F5,$66,$BF,$D7,$1A,$FF,$50
	.dc.b	$6B,$FD,$E9,$AF,$F7,$93,$FA,$C1,$0F,$EC,$E5,$3F,$B1,$94,$FE,$D2
	.dc.b	$53,$FA,$E9,$4F,$F0,$39,$FD,$9B,$EE,$3F,$C1,$AF,$FA,$CD,$AF,$F7
	.dc.b	$00,$E1,$FE,$E6,$D7,$FC,$32,$D7,$7D,$5D,$48,$FF,$80,$23,$44,$8C
	.dc.b	$C7,$87,$EE,$3C,$84,$78,$35,$D1,$82,$18,$21,$E9,$94,$71,$8C,$7A
	.dc.b	$FF,$04,$8F,$95,$AF,$C3,$6B,$C0,$D1,$0D,$14,$F0,$76,$FC,$D2,$3F
	.dc.b	$05,$AF,$CF,$6B,$C0,$D1,$0D,$14,$F6,$77,$7D,$92,$3F,$35,$AF,$DB
	.dc.b	$6B,$C0,$D1,$0D,$14,$F9,$1F,$1F,$F8,$52,$3F,$65,$AF,$E1,$DA,$F0
	.dc.b	$34,$43,$45,$3E,$A7,$83,$F9,$12,$3F,$85,$6B,$F9,$36,$BC,$0D,$10
	.dc.b	$D1,$4F,$C0,$7B,$5F,$9D,$23,$F9,$16,$BF,$9F,$6B,$C0,$D1,$0D,$14
	.dc.b	$FC,$67,$89,$FA,$92,$3F,$9D,$6B,$FA,$B6,$BC,$0D,$10,$D1,$4F,$CC
	.dc.b	$7D,$37,$ED,$48,$FE,$A5,$AF,$ED,$DA,$F0,$34,$43,$45,$3F,$49,$ED
	.dc.b	$7F,$01,$23,$FB,$56,$BF,$C0,$DA,$F0,$34,$43,$45,$3F,$61,$E3,$FF
	.dc.b	$0D,$23,$FC,$05,$AF,$F0,$F6,$BC,$0D,$10,$D1,$4F,$DE,$79,$1F,$C6
	.dc.b	$48,$FF,$0D,$6B,$FC,$6D,$AF,$03,$44,$34,$53,$F8,$47,$CB,$7F,$25
	.dc.b	$23,$FC,$65,$AF,$F2,$76,$BC,$0D,$10,$D1,$4F,$E2,$9E,$EF,$F3,$12
	.dc.b	$3F,$C9,$5A,$FF,$33,$6B,$C0,$D1,$0D,$14,$FE,$41,$E6,$7F,$3D,$23
	.dc.b	$FC,$C5,$AF,$F3,$F6,$BC,$0D,$10,$D1,$4F,$E5,$9E,$67,$F4,$92,$3F
	.dc.b	$CF,$5A,$FF,$4B,$6B,$C0,$D1,$0D,$14,$FE,$71,$ED,$7F,$55,$23,$FD
	.dc.b	$25,$AF,$F5,$76,$BC,$0D,$10,$D1,$4F,$E8,$9E,$77,$F6,$12,$3F,$D5
	.dc.b	$5A,$FF,$63,$6B,$C0,$D1,$0D,$14,$FE,$A1,$EF,$FF,$6D,$23,$FD,$85
	.dc.b	$AF,$F6,$F6,$BC,$0D,$10,$D1,$4F,$EB,$9E,$8F,$F7,$92,$3F,$DB,$5A
	.dc.b	$FF,$7B,$6B,$C0,$D1,$0D,$14,$FF,$00,$53,$E9,$3F,$DD,$6D,$8F,$F7
	.dc.b	$CB,$5F,$F0,$30,$70,$F2,$8E,$24,$78,$B5,$E3,$1C,$1F,$F7,$FB,$51
	.dc.b	$D6,$3C,$7F,$15,$23,$C6,$D7,$9B,$5F,$5B,$5E,$AD,$7A,$CA,$31,$1E
	.dc.b	$BF,$57,$E0,$91,$F8,$2D,$7D,$2D,$7E,$4B,$5F,$5B,$5E,$EF,$CC,$F5
	.dc.b	$64,$7C,$C6,$BF,$11,$AF,$D4,$6B,$F1,$9A,$FA,$9A,$FD,$3F,$89,$FB
	.dc.b	$24,$7E,$CB,$5F,$A2,$D7,$F0,$2D,$7E,$9B,$5F,$8E,$FE,$13,$E0,$E4
	.dc.b	$7E,$11,$AF,$DC,$6B,$F8,$C6,$BF,$79,$AF,$D2,$6B,$F8,$BF,$63,$F2
	.dc.b	$24,$7F,$22,$D7,$F1,$2D,$7F,$32,$D7,$F1,$6D,$7E,$FB,$F9,$CF,$66
	.dc.b	$47,$E7,$1A,$FE,$51,$AF,$E9,$1A,$FE,$59,$AF,$E2,$9A,$FE,$8F,$E7
	.dc.b	$BF,$52,$47,$F5,$2D,$7F,$42,$D7,$F6,$2D,$7F,$46,$D7,$F2,$EF,$ED
	.dc.b	$3D,$C9,$1F,$B4,$6B,$FA,$C6,$BF,$BC,$6B,$FA,$E6,$BF,$A2,$6B,$FB
	.dc.b	$BF,$6F,$F8,$09,$1F,$E0,$2D,$7F,$72,$D7,$F8,$4B,$5F,$DD,$B5,$FD
	.dc.b	$7B,$FC,$33,$DD,$91,$FC,$31,$AF,$F0,$46,$BF,$C5,$1A,$FF,$06,$6B
	.dc.b	$FB,$A6,$BF,$C4,$FE,$0F,$F8,$C9,$1F,$E3,$2D,$7F,$88,$B5,$FE,$42
	.dc.b	$D7,$F8,$9B,$5F,$E0,$EF,$F2,$4F,$8C,$91,$FC,$91,$AF,$F1,$C6,$BF
	.dc.b	$CB,$1A,$FF,$1E,$6B,$FC,$49,$AF,$F2,$BF,$81,$FE,$62,$47,$F9,$8B
	.dc.b	$5F,$E5,$2D,$7F,$9C,$B5,$FE,$56,$D7,$F8,$FB,$FC,$F3,$E0,$64,$7F
	.dc.b	$3C,$6B,$FC,$D1,$AF,$F4,$46,$BF,$CD,$9A,$FF,$2A,$6B,$FD,$0F,$E7
	.dc.b	$FF,$A4,$91,$FE,$92,$D7,$FA,$0B,$5F,$EA,$2D,$7F,$A1,$B5,$FE,$6E
	.dc.b	$FF,$54,$F8,$19,$1F,$D5,$1A,$FF,$4C,$6B,$FD,$71,$AF,$F4,$E6,$BF
	.dc.b	$D0,$9A,$FF,$5B,$F4,$3F,$B0,$91,$FE,$C2,$D7,$FA,$CB,$5F,$ED,$2D
	.dc.b	$7F,$AD,$B5,$FE,$9E,$FF,$6C,$F4,$24,$7F,$6C,$6B,$FD,$91,$AF,$F7
	.dc.b	$46,$BF,$D9,$9A,$FF,$5A,$6B,$FD,$CF,$D5,$FE,$F2,$47,$FB,$CB,$5F
	.dc.b	$EE,$2D,$7F,$C0,$2D,$7F,$B9,$B5,$FE,$CE,$FF,$80,$1F,$EE,$BF,$72
	.dc.b	$F8,$4C,$42,$30,$ED,$AF,$F8,$21,$AF,$F7,$E3,$9C,$72,$8C,$C6,$5A
	.dc.b	$E8,$E9,$1B,$23,$83,$C1,$EB,$F3,$23,$82,$1E,$37,$2D,$32,$79,$4F
	.dc.b	$63,$A7,$12,$8F,$47,$B1,$27,$E0,$35,$F8,$0D,$78,$B5,$F7,$35,$F9
	.dc.b	$AD,$7E,$67,$C0,$7C,$52,$3F,$35,$AF,$CD,$6B,$E6,$6B,$F2,$DA,$FD
	.dc.b	$A7,$E8,$7B,$72,$3E,$C3,$5F,$B0,$D7,$E0,$B5,$FA,$CD,$7F,$0A,$D7
	.dc.b	$F0,$9F,$0B,$F7,$48,$FE,$15,$AF,$E1,$5A,$FC,$E6,$BF,$83,$6B,$F9
	.dc.b	$27,$F1,$1E,$FC,$8F,$C8,$35,$FC,$83,$5F,$B2,$D7,$F1,$CD,$7F,$3A
	.dc.b	$D7,$F3,$9F,$3B,$F9,$52,$3F,$9D,$6B,$F9,$D6,$BF,$86,$6B,$F9,$B6
	.dc.b	$BF,$AA,$7F,$41,$F1,$92,$3F,$50,$D7,$F5,$0D,$7F,$22,$D7,$F4,$CD
	.dc.b	$7F,$6A,$D7,$F6,$9F,$05,$FA,$D2,$3F,$B5,$6B,$FB,$56,$BF,$9E,$6B
	.dc.b	$FB,$36,$BF,$C0,$9F,$DC,$7C,$0C,$8F,$E0,$0D,$7F,$80,$35,$FD,$4B
	.dc.b	$5F,$DF,$35,$FE,$1A,$D7,$F8,$67,$C5,$7F,$05,$23,$FC,$35,$AF,$F0
	.dc.b	$D6,$BF,$B6,$6B,$FC,$2D,$AF,$F1,$A7,$F8,$87,$C1,$48,$FE,$30,$D7
	.dc.b	$F8,$C3,$5F,$E0,$2D,$7F,$8B,$35,$FE,$4A,$D7,$F9,$27,$C2,$7F,$1D
	.dc.b	$23,$FC,$95,$AF,$F2,$56,$BF,$C3,$9A,$FF,$23,$6B,$FC,$C9,$FE,$51
	.dc.b	$F0,$72,$3F,$98,$35,$FE,$60,$D7,$F8,$CB,$5F,$E5,$CD,$7F,$9E,$B5
	.dc.b	$FE,$79,$F4,$7F,$CD,$48,$FF,$3D,$6B,$FC,$F5,$AF,$F2,$66,$BF,$CE
	.dc.b	$DA,$FF,$4A,$7F,$A0,$7C,$34,$8F,$E9,$0D,$7F,$A4,$35,$FE,$62,$D7
	.dc.b	$FA,$33,$5F,$EA,$AD,$7F,$AA,$7C,$1F,$F4,$D2,$3F,$D5,$5A,$FF,$55
	.dc.b	$6B,$FC,$F9,$AF,$F5,$36,$BF,$D8,$9F,$EB,$1F,$07,$23,$FB,$03,$5F
	.dc.b	$EC,$0D,$7F,$A4,$B5,$FE,$BC,$D7,$FB,$6B,$5F,$ED,$8F,$F6,$7F,$C3
	.dc.b	$BF,$DE,$11,$FE,$E6,$FB,$B7,$C0,$E7,$CE,$FE,$D9,$18,$9F,$F7,$5D
	.dc.b	$70,$79,$98,$9F,$0B,$4E,$95,$CD,$A6,$4E,$22,$7A,$74,$B0,$E1,$2C
	.dc.b	$3C,$4E,$F0,$F2,$E1,$AB,$E1,$3D,$5A,$EC,$F0,$94,$E9,$1C,$6E,$47
	.dc.b	$96,$FB,$E1,$EF,$6E,$1E,$E9,$E5,$87,$6A,$D6,$1E,$8B,$DD,$7D,$23
	.dc.b	$A4,$64,$8F,$9F,$0A,$F9,$67,$8B,$E6,$34,$72,$21,$B9,$F4,$B4,$CD
	.dc.b	$79,$1E,$26,$48,$E5,$F5,$D7,$8A,$30,$4C,$3E,$03,$BE,$BF,$F0,$E6
	.dc.b	$1F,$51,$B8,$7F,$8C,$DB,$80,$38,$69,$F6,$8D,$39,$23,$93,$67,$DA
	.dc.b	$72,$D3,$F2,$FF,$03,$E1,$69,$D0,$A7,$23,$F1,$8E,$DF,$D1,$35,$C9
	.dc.b	$C4,$C9,$4C,$9C,$E9,$1C,$5B,$1B,$6B,$B8,$D7,$C0,$C8,$D3,$5E,$8D
	.dc.b	$AF,$9E,$D8,$FD,$AB,$5F,$DD,$35,$FD,$47,$C0,$F8,$C8,$F8,$DA,$FA
	.dc.b	$5B,$3F,$71,$B1,$FC,$09,$AF,$F0,$77,$DD,$F0,$32,$3D,$CD,$7E,$23
	.dc.b	$67,$F0,$ED,$8F,$E1,$AD,$7F,$89,$35,$FE,$01,$F0,$3F,$2C,$8F,$CB
	.dc.b	$6B,$F4,$5B,$3F,$94,$6C,$7F,$1A,$6B,$FC,$7D,$FA,$DF,$03,$23,$EB
	.dc.b	$35,$FB,$8D,$9F,$CF,$B6,$3F,$92,$B5,$FE,$54,$D7,$F8,$C3,$FC,$6B
	.dc.b	$EF,$1E,$33,$D7,$FE,$45,$18,$7F,$88,$4F,$F0,$93,$84,$EE,$BF,$8C
	.dc.b	$56,$DF,$E0,$D3,$B7,$9C,$EF,$7F,$E6,$D5,$B7,$F8,$D5,$6E,$B9,$D6
	.dc.b	$F8,$3F,$D5,$57,$B7,$DB,$3E,$43,$AE,$11,$F0,$99,$39,$B9,$C3,$BC
	.dc.b	$59,$EB,$7D,$E3,$F8,$25,$F7,$7E,$49,$B3,$31,$B5,$1B,$7E,$69,$B3
	.dc.b	$31,$B8,$D6,$EC,$7E,$5E,$3B,$9C,$C6,$2F,$D2,$4C,$7D,$2D,$C7,$D3
	.dc.b	$7D,$F0,$E9,$3C,$F1,$F0,$1E,$78,$CC,$71,$3D,$73,$C7,$0F,$06,$1F
	.dc.b	$5A,$F8,$CF,$CB,$8C,$3D,$4C,$61,$E2,$BE,$87,$C0,$8C,$BD,$8C,$E3
	.dc.b	$EE,$79,$1C,$31,$F2,$DF,$68,$DE,$1E,$9C,$7C,$B8,$F7,$AD,$1F,$10
	.dc.b	$AE,$49,$5F,$68,$E7,$8F,$BD,$F0,$CF,$07,$4D,$7D,$47,$96,$BC,$2F
	.dc.b	$AE,$35,$87,$A2,$FB,$CF,$09,$CD,$70,$8C,$4E,$AB,$84,$73,$8E,$38
	.dc.b	$FC,$58,$F3,$9E,$13,$CF,$1C,$6B,$A4,$FE,$44,$E1,$7E,$97,$18,$FC
	.dc.b	$9B,$E2,$77,$3D,$F7,$E0,$8F,$BF,$1F,$AA,$3C,$35,$EA,$78,$9E,$A3
	.dc.b	$89,$CC,$C1,$30,$F6,$B8,$BF,$1E,$3F,$84,$8C,$5F,$82,$FE,$D6,$1E
	.dc.b	$3A,$FA,$D5,$E2,$8F,$80,$F8,$AF,$DF,$1B,$8F,$8C,$6C,$F2,$61,$E2
	.dc.b	$9F,$0C,$6A,$FE,$36,$1E,$1B,$FE,$1B,$3C,$4F,$56,$3F,$3A,$75,$1F
	.dc.b	$00,$F4,$E3,$E6,$BE,$6D,$BF,$24,$7C,$99,$E9,$87,$A1,$F9,$0F,$92
	.dc.b	$3E,$0C,$62,$FB,$BF,$C3,$0E,$2F,$DB,$6D,$FC,$BA,$E5,$7F,$1A,$72
	.dc.b	$7D,$BB,$F2,$BF,$06,$3C,$A6,$E3,$27,$9C,$AE,$4D,$C7,$CC,$C3,$E2
	.dc.b	$5F,$BE,$3D,$0D,$7E,$29,$DC,$C9,$4C,$D7,$7A,$EB,$1F,$40,$FB,$D8
	.dc.b	$7B,$E7,$C7,$1E,$13,$35,$F0,$67,$E3,$E1,$F3,$67,$AD,$FA,$5F,$0D
	.dc.b	$F7,$6C,$65,$D9,$D4,$F9,$31,$F5,$1C,$9E,$53,$31,$F8,$88,$F2,$C6
	.dc.b	$5B,$3E,$33,$EA,$C7,$58,$CD,$7B,$6B,$E0,$D6,$E3,$D6,$66,$BF,$95
	.dc.b	$1F,$E5,$92,$9F,$CC,$4C,$1F,$C6,$0B,$1F,$DD,$5B,$1F,$DE,$5B,$7C
	.dc.b	$1E,$64,$39,$1D,$48,$66,$7A,$38,$36,$66,$7B,$39,$8D,$B9,$9F,$03
	.dc.b	$C0,$91,$F1,$9F,$95,$5F,$12,$BE,$14,$7D,$29,$F8,$D5,$FC,$53,$1F
	.dc.b	$DF,$57,$C8,$8F,$C2,$19,$C7,$F2,$F7,$F9,$1B,$EC,$72,$7F,$80,$5F
	.dc.b	$2A,$E1,$1F,$8E,$3F,$55,$87,$8E,$3A,$C7,$0B,$EB,$8F,$E3,$63,$CD
	.dc.b	$8F,$A3,$1E,$ED,$3C,$D1,$D0,$FC,$99,$F8,$5C,$3B,$39,$39,$E1,$E0
	.dc.b	$BF,$C7,$E3,$E7,$9F,$11,$E7,$8F,$CC,$61,$E0,$AE,$17,$FA,$EB,$E1
	.dc.b	$5C,$B1,$F4,$CF,$E8,$EF,$86,$1E,$16,$D7,$4C,$7D,$2D,$C7,$E3,$D7
	.dc.b	$E9,$0F,$E4,$D5,$C2,$FF,$4C,$7B,$2B,$84,$FA,$0E,$07,$F1,$F7,$F4
	.dc.b	$B5,$E8,$8F,$8A,$66,$79,$1D,$1C,$D7,$DB,$AF,$64,$7C,$39,$D3,$F2
	.dc.b	$8F,$65,$FA,$A3,$35,$E7,$9F,$4C,$E5,$B7,$EE,$BF,$2C,$72,$8C,$5F
	.dc.b	$92,$36,$7B,$67,$67,$C0,$3C,$AF,$C0,$C7,$E9,$CF,$D5,$C7,$E9,$CF
	.dc.b	$E0,$CF,$6E,$1F,$89,$C3,$BC,$61,$F8,$CF,$F1,$6B,$E6,$FC,$86,$B9
	.dc.b	$AE,$AF,$EB,$8E,$47,$AC,$F4,$5F,$AD,$FA,$06,$B0,$F5,$47,$9E,$3E
	.dc.b	$71,$F0,$23,$DA,$7C,$08,$F9,$B8,$7C,$B9,$E6,$7D,$39,$E6,$F9,$EB
	.dc.b	$CB,$8F,$DB,$BE,$D3,$F1,$E7,$37,$F0,$E3,$CE,$79,$67,$BE,$3F,$1A
	.dc.b	$3B,$DF,$F0,$DC,$7F,$C2,$A7,$C4,$7F,$A4,$36,$FF,$87,$B6,$BF,$C3
	.dc.b	$0D,$AF,$DA,$37,$1F,$E8,$CF,$F9,$BB,$2F,$F5,$AC,$FF,$AE,$88,$7F
	.dc.b	$AE,$8A,$E9,$C1,$1C,$11,$FF,$0A,$BF,$EC,$B7,$FD,$A0,$43,$FD,$A0
	.dc.b	$57,$4E,$08,$E0,$8F,$FB,$4C,$E0,$7F,$B8,$8F,$F7,$49,$0F,$F7,$49
	.dc.b	$5F,$F7,$4B,$4E,$06,$88,$E9,$A7,$FB,$E4,$FF,$80,$04,$3F,$E0,$01
	.dc.b	$5F,$F8,$00,$D3,$81,$A2,$3A,$69,$FD,$96,$1B,$7F,$67,$9F,$EC,$B1
	.dc.b	$89,$08,$CC,$EE,$39,$35,$DC,$F1,$78,$1A,$39,$CF,$26,$99,$36,$72
	.dc.b	$9E,$8A,$7A,$4F,$18,$D9,$99,$DC,$B7,$A4,$9E,$E7,$99,$C3,$CA,$38
	.dc.b	$3A,$3C,$11,$AA,$F8,$AD,$8F,$94,$D7,$E0,$F0,$78,$88,$F3,$A4,$62
	.dc.b	$3C,$0D,$AF,$A2,$D8,$FA,$CD,$7A,$9C,$DF,$F0,$F6,$0D,$37,$3F,$62
	.dc.b	$30,$F0,$0E,$17,$13,$8B,$9D,$72,$9D,$B6,$B8,$63,$EA,$9E,$D3,$FE
	.dc.b	$2F,$BE,$51,$83,$C1,$5C,$30,$EB,$18,$AE,$18,$62,$3E,$F4,$DC,$3B
	.dc.b	$47,$88,$E0,$7E,$53,$0E,$EE,$4E,$27,$84,$A6,$63,$9E,$1F,$06,$78
	.dc.b	$BE,$59,$E1,$3A,$C3,$E0,$1E,$68,$E9,$5C,$23,$31,$E4,$3E,$5E,$1E
	.dc.b	$B9,$D4,$66,$BA,$3E,$99,$D1,$EF,$C2,$47,$68,$CB,$E4,$8C,$BE,$FA
	.dc.b	$EA,$F8,$0D,$10,$D4,$F2,$7E,$14,$F2,$C3,$E1,$4E,$67,$FC,$CB,$5F
	.dc.b	$31,$B1,$FE,$52,$7B,$1F,$E6,$38,$D3,$F1,$4C,$46,$E3,$31,$EF,$3F
	.dc.b	$CB,$6E,$1E,$67,$63,$3C,$7C,$48,$E0,$87,$85,$CB,$4C,$9E,$A7,$2D
	.dc.b	$3C,$3C,$9E,$52,$38,$21,$ED,$72,$D3,$27,$C6,$72,$D3,$DB,$CB,$F0
	.dc.b	$48,$E0,$87,$C9,$72,$D3,$27,$D2,$72,$D3,$E4,$F3,$7C,$D2,$38,$21
	.dc.b	$F5,$5C,$B4,$C9,$F7,$9C,$B4,$FA,$BC,$FF,$64,$8E,$08,$7E,$05,$CB
	.dc.b	$4C,$9F,$8A,$72,$D3,$F0,$3D,$1F,$85,$23,$82,$1F,$8D,$72,$D3,$27
	.dc.b	$E5,$9C,$B4,$FC,$6F,$4F,$E4,$48,$E0,$87,$E6,$5C,$B4,$C9,$FA,$27
	.dc.b	$2D,$3F,$33,$D5,$F9,$D2,$38,$21,$FA,$57,$2D,$32,$7E,$B9,$CB,$4F
	.dc.b	$D2,$F5,$FE,$A4,$8E,$08,$7E,$C5,$CB,$4C,$9F,$BA,$72,$D3,$F6,$3D
	.dc.b	$9F,$B5,$23,$82,$1F,$BD,$72,$D3,$27,$F0,$67,$2D,$3F,$7B,$DB,$FC
	.dc.b	$04,$8E,$08,$7F,$09,$72,$D3,$27,$F1,$27,$2D,$3F,$84,$F7,$7F,$0D
	.dc.b	$23,$82,$1F,$C5,$5C,$B4,$C9,$FC,$79,$CB,$4F,$E2,$BD,$FF,$C6,$48
	.dc.b	$E0,$87,$F2,$17,$2D,$32,$7F,$2A,$72,$D3,$F9,$0F,$81,$FC,$94,$8E
	.dc.b	$08,$7F,$2D,$72,$D3,$27,$F3,$67,$2D,$3F,$96,$F8,$3F,$CC,$48,$E0
	.dc.b	$87,$F3,$97,$2D,$32,$7F,$42,$72,$D3,$F9,$CF,$85,$FC,$F4,$8E,$08
	.dc.b	$7F,$45,$72,$D3,$27,$F4,$E7,$2D,$3F,$A2,$F8,$7F,$D2,$48,$E0,$87
	.dc.b	$F5,$17,$2D,$32,$7F,$5A,$72,$D3,$FA,$8F,$89,$FD,$54,$8E,$08,$7F
	.dc.b	$5D,$72,$D3,$27,$F6,$67,$2D,$3F,$CC,$E4,$3F,$C6,$9D,$7F,$DA,$7A
	.dc.b	$FF,$CA,$F1,$A6,$BF,$E6,$E6,$BC,$0E,$46,$63,$91,$1C,$9B,$37,$1C
	.dc.b	$89,$1E,$1F,$16,$4F,$09,$46,$23,$8B,$5F,$54,$69,$E1,$19,$7D,$4D
	.dc.b	$7D,$4F,$BF,$C6,$91,$DD,$AF,$99,$AF,$66,$BF,$15,$AF,$C5,$6B,$FE
	.dc.b	$5F,$F1,$FE,$61,$1F,$63,$5F,$86,$D7,$D6,$D7,$E9,$35,$FA,$4F,$D7
	.dc.b	$F0,$A4,$7C,$76,$BF,$39,$AF,$C6,$6B,$F7,$5A,$FD,$D6,$BF,$E7,$7F
	.dc.b	$23,$F0,$84,$7E,$A3,$5F,$B6,$D7,$E9,$B5,$FC,$53,$5F,$C5,$3F,$8F
	.dc.b	$F2,$64,$7D,$F6,$BF,$86,$6B,$F7,$9A,$FE,$55,$AF,$E5,$5A,$FF,$A3
	.dc.b	$FC,$AF,$CE,$11,$FC,$63,$5F,$C9,$B5,$FC,$5B,$5F,$D1,$35,$FD,$13
	.dc.b	$FA,$7F,$2E,$47,$E5,$DA,$FE,$79,$AF,$E5,$9A,$FE,$B5,$AF,$EB,$5A
	.dc.b	$FF,$A9,$FC,$CF,$DA,$11,$FD,$23,$5F,$D5,$B5,$FD,$1B,$5F,$DD,$35
	.dc.b	$FD,$D3,$FB,$FF,$36,$47,$EB,$DA,$FE,$D9,$AF,$EB,$9A,$FF,$05,$6B
	.dc.b	$FC,$15,$AF,$FA,$FF,$CE,$FE,$18,$47,$F7,$8D,$7F,$81,$B5,$FD,$DB
	.dc.b	$5F,$E2,$4D,$7F,$89,$3F,$C5,$FC,$F9,$1F,$C1,$DA,$FF,$0E,$6B,$FC
	.dc.b	$19,$AF,$F1,$D6,$BF,$C7,$5A,$FF,$B5,$FD,$0F,$E4,$84,$7F,$8A,$35
	.dc.b	$FE,$36,$D7,$F8,$9B,$5F,$E5,$4D,$7F,$95,$3F,$CB,$FD,$19,$1F,$C7
	.dc.b	$DA,$FF,$26,$6B,$FC,$79,$AF,$F3,$56,$BF,$CD,$5A,$FF,$BB,$FD,$2F
	.dc.b	$E7,$84,$7F,$96,$35,$FE,$66,$D7,$F9,$5B,$5F,$E8,$4D,$7F,$A1,$3F
	.dc.b	$D1,$FD,$39,$1F,$CD,$DA,$FF,$3E,$6B,$FC,$D9,$AF,$F4,$D6,$BF,$D3
	.dc.b	$5A,$B8,$DE,$A7,$F5,$52,$47,$F4,$56,$BF,$D3,$1A,$FF,$44,$6B,$FD
	.dc.b	$6D,$AF,$F5,$B7,$FB,$07,$AB,$23,$FA,$83,$5F,$EA,$ED,$7F,$A7,$B5
	.dc.b	$FE,$CC,$D7,$FB,$33,$5F,$57,$AD,$FD,$B4,$8F,$F5,$D6,$BF,$D9,$1A
	.dc.b	$FF,$5C,$6B,$FD,$CD,$AF,$F7,$37,$FB,$C7,$AF,$23,$FB,$43,$5F,$ED
	.dc.b	$ED,$7F,$B3,$B5,$FE,$FC,$D7,$FB,$F3,$57,$C1,$BD,$3F,$F7,$73,$63
	.dc.b	$FB,$EB,$5F,$F0,$1B,$5F,$F0,$73,$5F,$F0,$8B,$51,$DA,$3A,$91,$EA
	.dc.b	$D7,$AB,$E2,$F6,$64,$78,$9A,$F8,$9A,$F8,$9A,$F8,$9A,$F8,$A5,$18
	.dc.b	$8F,$83,$ED,$7C,$02,$3F,$01,$AF,$C0,$6B,$F0,$1A,$FC,$06,$BD,$DF
	.dc.b	$97,$9C,$8F,$96,$D7,$E5,$B5,$F9,$6D,$7E,$5B,$5F,$53,$5F,$B3,$ED
	.dc.b	$FD,$82,$3F,$61,$AF,$D8,$6B,$F6,$1A,$FD,$86,$BF,$1D,$FC,$1F,$B9
	.dc.b	$23,$F0,$6D,$7F,$06,$D7,$F0,$6D,$7F,$06,$D7,$E9,$35,$FC,$8F,$BB
	.dc.b	$F9,$02,$3F,$90,$6B,$F9,$06,$BF,$90,$6B,$F9,$06,$BF,$7D,$FC,$DF
	.dc.b	$BD,$23,$F3,$6D,$7F,$36,$D7,$F3,$6D,$7F,$36,$D7,$F1,$4D,$7F,$53
	.dc.b	$EF,$FE,$A0,$8F,$EA,$1A,$FE,$A1,$AF,$EA,$1A,$FE,$A1,$AF,$E5,$DF
	.dc.b	$D9,$FC,$04,$8F,$D9,$B5,$FD,$9B,$5F,$D9,$B5,$FD,$9B,$5F,$D1,$35
	.dc.b	$FE,$03,$F0,$3F,$C0,$08,$FF,$00,$6B,$FC,$01,$AF,$F0,$06,$BF,$C0
	.dc.b	$1A,$FE,$BD,$FE,$17,$F0,$52,$3F,$85,$B5,$FE,$16,$D7,$F8,$5B,$5F
	.dc.b	$E1,$6D,$7F,$74,$D7,$F8,$CF,$C1,$FF,$18,$23,$FC,$61,$AF,$F1,$86
	.dc.b	$BF,$C6,$1A,$FF,$18,$6B,$FC,$1D,$FE,$47,$F0,$12,$3F,$91,$B5,$FE
	.dc.b	$46,$D7,$F9,$1B,$5F,$E4,$6D,$7F,$89,$35,$FE,$63,$F0,$3F,$CC,$08
	.dc.b	$FF,$30,$6B,$FC,$C1,$AF,$F3,$06,$BF,$CC,$1A,$FF,$1F,$7F,$9D,$FC
	.dc.b	$14,$8F,$E7,$6D,$7F,$9D,$B5,$FE,$76,$D7,$F9,$DB,$5F,$E5,$4D,$7F
	.dc.b	$A4,$FC,$1F,$F4,$82,$3F,$D2,$1A,$FF,$48,$6B,$FD,$21,$AF,$F4,$86
	.dc.b	$BF,$CD,$DF,$EA,$7F,$05,$23,$FA,$9B,$5F,$EA,$6D,$7F,$A9,$B5,$FE
	.dc.b	$A6,$D7,$FA,$13,$5F,$EC,$3F,$05,$FD,$80,$8F,$F6,$06,$BF,$D8,$1A
	.dc.b	$FF,$60,$6B,$FD,$81,$AF,$F4,$F7,$FB,$5F,$C0,$C8,$FE,$D6,$D7,$FB
	.dc.b	$5B,$5F,$ED,$6D,$7F,$B5,$B5,$FE,$B4,$D7,$FB,$CF,$C2,$7F,$78,$23
	.dc.b	$FD,$E1,$AF,$F7,$86,$BF,$DE,$1A,$FF,$78,$6B,$FD,$9D,$FF,$7E,$AF
	.dc.b	$89,$AD,$32,$E0,$8E,$08,$7F,$80,$88,$71,$29,$FE,$06,$78,$B8,$34
	.dc.b	$69,$ED,$F0,$DE,$02,$46,$E1,$18,$8E,$0D,$7C,$31,$A7,$84,$65,$F3
	.dc.b	$46,$5E,$27,$5F,$87,$F7,$91,$C1,$0F,$33,$96,$99,$3C,$2E,$5A,$7A
	.dc.b	$7E,$13,$E5,$91,$C1,$0F,$84,$E5,$A6,$4F,$6B,$96,$9F,$17,$E0,$FE
	.dc.b	$B9,$1C,$10,$F9,$CE,$5A,$64,$F9,$2E,$5A,$7D,$7F,$89,$7A,$5F,$7F
	.dc.b	$E2,$5E,$A7,$E0,$65,$EA,$66,$30,$DB,$FF,$23,$CF,$27,$FC,$93,$5C
	.dc.b	$9C,$DE,$D4,$F4,$77,$7D,$1A,$70,$29,$C2,$79,$4F,$18,$57,$77,$AD
	.dc.b	$EA,$A4,$72,$AE,$15,$B6,$DF,$BE,$B7,$5D,$23,$C8,$D3,$DF,$7A,$BE
	.dc.b	$55,$E4,$89,$3D,$E3,$E0,$66,$F8,$3F,$D8,$92,$7F,$44,$D7,$F5,$0D
	.dc.b	$7F,$42,$D7,$F3,$ED,$7F,$64,$FF,$00,$F8,$27,$B3,$F3,$D7,$B3,$4C
	.dc.b	$44,$AF,$9B,$3A,$6C,$E5,$F8,$02,$3A,$BA,$21,$A7,$2D,$9F,$03,$AC
	.dc.b	$BE,$B4,$E9,$B3,$A9,$E5,$5E,$19,$CC,$71,$CF,$F8,$A5,$F9,$E3,$B1
	.dc.b	$F8,$6C,$5F,$03,$F0,$7D,$C9,$39,$8F,$A6,$61,$F4,$46,$1D,$1F,$11
	.dc.b	$D6,$1F,$8D,$AE,$B3,$A3,$F8,$F5,$7F,$2C,$23,$FC,$75,$AE,$8C,$10
	.dc.b	$C1,$0F,$E3,$EF,$09,$F0,$3E,$58,$F9,$DE,$F5,$9C,$C7,$8D,$F9,$B1
	.dc.b	$9C,$7F,$92,$D5,$AF,$F5,$EE,$2A,$FC,$34,$FE,$1F,$E0,$65,$7E,$12
	.dc.b	$32,$FE,$1A,$32,$FD,$F8,$CB,$F8,$18,$CE,$1F,$7A,$72,$FE,$02,$33
	.dc.b	$87,$DA,$9C,$BF,$72,$33,$5E,$69,$CB,$E8,$8C,$BF,$93,$8F,$CF,$6B
	.dc.b	$23,$F3,$6F,$E8,$7E,$0F,$11,$F9,$58,$B8,$7E,$4A,$6E,$1F,$8C,$9B
	.dc.b	$5F,$1A,$6B,$FA,$18,$B9,$26,$C6,$1B,$3E,$23,$F5,$AF,$F0,$2F,$7F
	.dc.b	$F6,$9F,$0B,$27,$F3,$11,$97,$F6,$71,$97,$F6,$3A,$A7,$51,$FA,$F9
	.dc.b	$D6,$5F,$95,$9C,$CF,$F6,$87,$FB,$F9,$1F,$EE,$0D,$7F,$BB,$35,$E0
	.dc.b	$68,$86,$8B,$1D,$DF,$49,$FC,$2B,$97,$CB,$19,$8F,$4C,$FF,$13,$9C
	.dc.b	$E7,$FD,$E2,$2C,$FE,$52,$7F,$A3,$7C,$53,$A6,$92,$7F,$8B,$94,$FE
	.dc.b	$2B,$1B,$7F,$A3,$B5,$FE,$28,$6F,$0F,$E8,$D1,$96,$CF,$F2,$79,$D6
	.dc.b	$1F,$E2,$1A,$D3,$67,$53,$D9,$DA,$13,$C1,$FD,$79,$A3,$31,$75,$FF
	.dc.b	$16,$DF,$93,$B7,$E0,$F6,$CB,$F5,$C7,$C4,$30,$FC,$08,$C3,$A3,$E0
	.dc.b	$BA,$9E,$91,$93,$81,$A3,$0E,$9C,$91,$C1,$31,$F8,$93,$D5,$F8,$D5
	.dc.b	$FD,$F7,$E2,$DF,$03,$F6,$BC,$19,$8F,$F7,$A9,$CB,$67,$2F,$DA,$9E
	.dc.b	$51,$23,$ED,$C6,$B1,$FE,$FD,$5C,$A7,$FA,$AB,$5F,$F2,$BB,$5F,$F2
	.dc.b	$D3,$5F,$AC,$D4,$D2,$57,$F9,$A9,$B1,$FE,$5D,$6B,$FE,$62,$6C,$F9
	.dc.b	$27,$D1,$3F,$94,$F0,$EC,$E6,$3E,$01,$F9,$68,$CE,$3E,$6A,$B3,$F8
	.dc.b	$BD,$55,$FA,$4F,$1E,$4F,$E7,$9F,$D3,$F8,$F8,$8F,$D1,$C5,$C7,$F3
	.dc.b	$F3,$71,$FC,$E4,$DA,$ED,$35,$FC,$CC,$5C,$13,$63,$0D,$9F,$11,$FB
	.dc.b	$07,$F7,$3E,$5F,$EE,$BC,$B9,$3F,$B4,$8C,$BF,$B6,$8C,$BF,$BA,$8C
	.dc.b	$BF,$BC,$8C,$E3,$FB,$89,$CB,$FB,$A8,$CE,$3F,$B7,$9C,$BF,$B9,$8C
	.dc.b	$D7,$CB,$9C,$BF,$36,$32,$FE,$E2,$3F,$35,$ED,$FF,$12,$F3,$A4,$FE
	.dc.b	$66,$32,$FF,$0E,$8C,$BF,$92,$8C,$BF,$C5,$23,$38,$7E,$62,$73,$3F
	.dc.b	$96,$7F,$1C,$47,$F1,$CD,$7F,$1C,$D7,$81,$A2,$1A,$2C,$77,$7D,$07
	.dc.b	$F2,$4F,$47,$CB,$19,$8F,$41,$93,$66,$08,$FF,$28,$8F,$F3,$D7,$C6
	.dc.b	$3D,$6B,$EB,$4B,$FF,$3B,$CE,$5B,$39,$7F,$C9,$70,$9E,$73,$27,$8C
	.dc.b	$6E,$FF,$CF,$95,$CE,$78,$D6,$74,$FF,$6B,$4F,$39,$EE,$43,$BD,$ED
	.dc.b	$FE,$6D,$1C,$EF,$AD,$E6,$53,$CD,$EE,$7B,$3D,$AF,$82,$47,$DF,$ED
	.dc.b	$57,$DB,$16,$FD,$33,$71,$F3,$CD,$AF,$1C,$D7,$C3,$16,$D3,$63,$0D
	.dc.b	$9F,$21,$E2,$3F,$97,$61,$F5,$BD,$C9,$18,$21,$F4,$9C,$BF,$CA,$A3
	.dc.b	$F9,$D1,$8C,$7E,$8C,$E2,$78,$9F,$CB,$A7,$F9,$31,$1F,$E4,$CD,$7F
	.dc.b	$93,$35,$E0,$68,$86,$8B,$3F,$CF,$A3,$C1,$1F,$88,$F7,$6C,$E6,$3D
	.dc.b	$6F,$E1,$A3,$39,$7F,$40,$AB,$5F,$80,$9F,$CC,$FC,$0F,$E6,$BE,$06
	.dc.b	$4F,$E6,$23,$2F,$E6,$A3,$2F,$E6,$23,$2F,$E6,$A3,$38,$7E,$5A,$72
	.dc.b	$FE,$62,$33,$87,$E5,$A7,$2F,$E6,$23,$35,$F9,$49,$CB,$F9,$68,$CB
	.dc.b	$F9,$48,$FD,$8F,$95,$23,$F5,$EF,$ED,$3C,$BC,$47,$EB,$A2,$E1,$FA
	.dc.b	$A9,$B8,$7E,$A2,$6D,$7E,$92,$6B,$FA,$28,$BB,$7F,$B9,$A3,$F5,$31
	.dc.b	$83,$C0,$F8,$88,$78,$4F,$E0,$73,$FF,$00,$7A,$5E,$C7,$F8,$9E,$7F
	.dc.b	$D9,$BF,$1A,$F8,$4C,$67,$FE,$FB,$8C,$4F,$F8,$B6,$3F,$DF,$D7,$CA
	.dc.b	$31,$9F,$38,$C4,$74,$BE,$73,$DE,$F9,$CF,$FB,$72,$B7,$7D,$30,$DC
	.dc.b	$FF,$8E,$EB,$73,$E3,$C3,$D5,$52,$BD,$47,$85,$AD,$29,$FE,$24,$6B
	.dc.b	$4C,$DF,$A1,$E6,$D7,$91,$D1,$C3,$FE,$4F,$6B,$CC,$D4,$76,$8F,$6E
	.dc.b	$1C,$AF,$DD,$5F,$20,$F7,$79,$DE,$B7,$E2,$5D,$9F,$A5,$36,$B9,$12
	.dc.b	$B8,$9C,$C6,$3F,$EC,$D7,$0B,$E1,$7E,$6C,$69,$B1,$B8,$FC,$1F,$9B
	.dc.b	$89,$91,$CA,$36,$61,$FC,$4C,$61,$D1,$E0,$75,$87,$E2,$E7,$53,$C0
	.dc.b	$D1,$C0,$F0,$98,$34,$60,$FC,$BF,$99,$E8,$23,$EE,$F3,$6B,$A3,$E0
	.dc.b	$C7,$BA,$33,$5F,$9D,$8B,$8F,$E5,$66,$D7,$E4,$66,$BF,$32,$2E,$29
	.dc.b	$B1,$86,$CF,$D3,$D1,$EC,$FC,$EF,$1E,$4E,$63,$F6,$86,$1F,$D9,$C6
	.dc.b	$1D,$1F,$69,$D6,$5F,$B2,$9D,$4F,$E5,$5F,$C8,$11,$FC,$EB,$5D,$18
	.dc.b	$21,$82,$1F,$A2,$78,$47,$78,$F1,$9F,$C2,$3C,$EF,$0C,$66,$3F,$79
	.dc.b	$3F,$8C,$9C,$E7,$FC,$16,$2C,$FE,$F6,$7F,$23,$E5,$7E,$4B,$D2,$93
	.dc.b	$F9,$88,$CB,$F9,$A8,$CB,$FC,$62,$32,$FF,$1A,$8C,$E1,$FC,$66,$72
	.dc.b	$FF,$1B,$8C,$E1,$FC,$4E,$72,$FF,$15,$8C,$D7,$F1,$19,$CB,$FC,$4E
	.dc.b	$32,$FE,$BA,$3F,$57,$F0,$7F,$EC,$F8,$34,$C6,$DF,$EC,$EB,$FE,$61
	.dc.b	$19,$F3,$38,$9F,$B2,$AA,$7E,$EE,$34,$FE,$E8,$C8,$C5,$5D,$2B,$31
	.dc.b	$FC,$E6,$6C,$E5,$E9,$3D,$DA,$43,$A4,$6D,$FE,$8A,$7F,$49,$F3,$3D
	.dc.b	$FE,$BC,$8F,$CF,$46,$5F,$44,$65,$FE,$7D,$19,$7A,$9F,$EF,$7C,$13
	.dc.b	$BC,$3F,$A6,$D6,$EB,$F9,$9B,$FC,$AC,$8F,$F3,$86,$BF,$CD,$5A,$F0
	.dc.b	$34,$43,$45,$8E,$EF,$A8,$FA,$1E,$6F,$9E,$33,$1E,$A3,$26,$CC,$11
	.dc.b	$FA,$6F,$B6,$3F,$0B,$E5,$C8,$FB,$8F,$E2,$3D,$9C,$47,$C4,$3F,$13
	.dc.b	$18,$AE,$71,$F8,$C9,$C6,$5F,$8A,$AB,$3F,$2E,$6B,$C8,$EE,$7E,$2D
	.dc.b	$CE,$5F,$ED,$C8,$C9,$E1,$51,$A9,$DB,$F9,$EF,$2F,$F3,$FE,$5C,$BF
	.dc.b	$CB,$46,$5F,$CC,$46,$5E,$91,$97,$B4,$67,$1F,$6E,$0A,$D4,$62,$75
	.dc.b	$8F,$EA,$27,$2F,$EA,$A3,$37,$F8,$B9,$CB,$F8,$D8,$CB,$FB,$48,$FE
	.dc.b	$C3,$F1,$CE,$0F,$DC,$FB,$B2,$BF,$B5,$4E,$5B,$39,$7E,$34,$F2,$89
	.dc.b	$1F,$B7,$8D,$63,$F6,$2B,$94,$FD,$D6,$BF,$E1,$86,$BF,$E3,$36,$A0
	.dc.b	$E1,$B5,$FE,$1A,$8C,$11,$C1,$23,$FC,$43,$19,$6C,$7A,$1F,$23,$FC
	.dc.b	$6B,$CC,$C5,$7A,$62,$E5,$FC,$62,$6D,$79,$6F,$DB,$7B,$8F,$E3,$13
	.dc.b	$FC,$93,$CA,$91,$FC,$6D,$FE,$4F,$EF,$62,$3F,$21,$16,$BF,$07,$37
	.dc.b	$0F,$E4,$F3,$6B,$F8,$44,$D7,$F9,$7C,$5D,$53,$63,$0D,$9F,$11,$FC
	.dc.b	$B5,$FF,$11,$FC,$7F,$FB,$4E,$FF,$CC,$8F,$FB,$47,$0F,$D8,$B6,$33
	.dc.b	$E4,$70,$38,$F9,$93,$0F,$F1,$95,$73,$7E,$D3,$53,$CD,$CC,$73,$78
	.dc.b	$CF,$37,$F8,$24,$ED,$E3,$1B,$73,$3B,$BF,$3C,$E5,$F4,$C6,$6F,$D3
	.dc.b	$5F,$9E,$F8,$AF,$83,$F0,$B2,$30,$43,$C2,$E5,$FC,$D4,$7E,$A8,$C5
	.dc.b	$FC,$69,$C4,$F1,$3F,$3D,$3F,$97,$23,$F9,$46,$BF,$E7,$8D,$91,$88
	.dc.b	$E2,$6C,$86,$DE,$35,$FA,$F7,$C3,$1F,$23,$CE,$B3,$98,$F6,$3F,$72
	.dc.b	$33,$87,$EC,$6A,$D7,$D8,$9F,$C2,$F9,$92,$3E,$EB,$F8,$8F,$3F,$11
	.dc.b	$F5,$A2,$E1,$F8,$89,$B8,$7E,$32,$6D,$7E,$12,$6B,$F8,$48,$BA,$A6
	.dc.b	$C6,$1B,$3E,$23,$F1,$EF,$E6,$BC,$CF,$CD,$F9,$B2,$7F,$27,$19,$7F
	.dc.b	$2B,$19,$7F,$1B,$19,$7F,$1F,$19,$C7,$F4,$33,$97,$F4,$71,$9C,$7F
	.dc.b	$4F,$39,$7F,$53,$19,$AF,$D0,$CE,$5F,$D1,$C6,$5F,$D1,$C7,$F3,$4F
	.dc.b	$67,$F6,$7E,$24,$9F,$E6,$71,$97,$F5,$B1,$97,$F9,$64,$65,$FD,$3C
	.dc.b	$67,$0F,$E6,$53,$99,$FE,$60,$FF,$29,$23,$FC,$99,$AF,$E4,$DA,$F0
	.dc.b	$34,$43,$45,$8E,$EF,$A0,$FE,$07,$E6,$F9,$63,$31,$E8,$32,$6C,$C1
	.dc.b	$1F,$E2,$31,$FE,$B8,$F9,$07,$97,$49,$5F,$EC,$18,$7B,$78,$8F,$F0
	.dc.b	$64,$70,$75,$3C,$E7,$57,$FE,$0E,$AD,$57,$0A,$D3,$C0,$86,$4F,$05
	.dc.b	$72,$D3,$FA,$C4,$2B,$9D,$F5,$BE,$73,$C4,$91,$97,$99,$EB,$F2,$FE
	.dc.b	$07,$99,$23,$04,$3D,$AE,$5F,$7C,$7C,$73,$13,$ED,$9C,$4F,$18,$C3
	.dc.b	$F2,$63,$2F,$23,$04,$30,$48,$E2,$F8,$8E,$98,$7C,$2F,$91,$7A,$1F
	.dc.b	$67,$E3,$25,$7F,$BD,$E7,$2D,$9C,$BF,$62,$7F,$DE,$F3,$A8,$FF,$23
	.dc.b	$36,$FE,$C4,$F0,$7F,$C8,$F1,$9B,$F1,$5F,$D4,$B9,$3C,$DB,$19,$76
	.dc.b	$51,$1B,$72,$6D,$C3,$FE,$17,$6B,$C2,$39,$34,$E9,$1A,$24,$7B,$89
	.dc.b	$1F,$9B,$F3,$AB,$F0,$22,$DF,$E5,$E6,$DF,$92,$3E,$0C,$FE,$9E,$7F
	.dc.b	$51,$E6,$4B,$FD,$33,$FA,$BF,$77,$11,$FA,$68,$B7,$FA,$39,$B8,$7E
	.dc.b	$73,$65,$62,$FF,$C6,$D0,$AC,$47,$E8,$66,$CE,$63,$C2,$43,$C0,$7E
	.dc.b	$E6,$77,$3C,$1F,$E0,$1E,$17,$F0,$1F,$36,$57,$EF,$E3,$2F,$F0,$18
	.dc.b	$CB,$FC,$02,$32,$FF,$02,$8C,$E3,$FB,$F9,$CB,$FC,$06,$33,$8F,$BE
	.dc.b	$72,$FC,$18,$CD,$7C,$09,$CB,$F0,$A3,$2F,$F0,$18,$FC,$CF,$47,$F1
	.dc.b	$9F,$80,$93,$FA,$18,$CB,$FC,$66,$32,$FE,$7A,$32,$FF,$1A,$8C,$E1
	.dc.b	$F9,$E9,$CC,$FE,$71,$FC,$7E,$C6,$92,$3F,$22,$D7,$F3,$4D,$79,$1A
	.dc.b	$21,$A3,$D0,$43,$FD,$71,$F2,$5F,$F0,$06,$7C,$11,$9A,$F3,$3F,$E7
	.dc.b	$B7,$FD,$77,$3B,$AF,$6C,$6A,$79,$F9,$DF,$CA,$A3,$7E,$65,$D3,$F9
	.dc.b	$6E,$08,$C5,$FF,$A1,$E3,$0F,$79,$DF,$9B,$A8,$4E,$9F,$4D,$66,$31
	.dc.b	$7E,$6A,$C4,$CA,$E8,$77,$78,$46,$63,$99,$A3,$BC,$69,$C5,$76,$BF
	.dc.b	$5C,$F1,$BF,$05,$6A,$31,$3A,$8E,$37,$E7,$F3,$7D,$1E,$6C,$9E,$B1
	.dc.b	$97,$BC,$65,$FA,$F1,$97,$EC,$C6,$70,$FA,$B3,$97,$EB,$C6,$6F,$E8
	.dc.b	$CE,$5F,$A7,$19,$AF,$A5,$39,$7E,$A4,$65,$FA,$31,$FC,$F3,$E2,$7E
	.dc.b	$87,$A3,$27,$FA,$24,$65,$F8,$F1,$97,$FA,$0C,$65,$FC,$8C,$67,$0F
	.dc.b	$E8,$33,$99,$FE,$7E,$FE,$90,$8F,$F3,$26,$BF,$CE,$DA,$F0,$34,$43
	.dc.b	$45,$8E,$EF,$A0,$FB,$BE,$77,$96,$33,$1E,$83,$26,$CC,$11,$FD,$2C
	.dc.b	$7E,$4F,$CD,$93,$F8,$87,$F2,$DE,$D6,$23,$F6,$31,$6B,$F5,$B3,$73
	.dc.b	$FD,$3C,$DA,$FD,$34,$D7,$F4,$71,$76,$4D,$8C,$36,$7C,$47,$E6,$9F
	.dc.b	$D4,$79,$BF,$A9,$F3,$64,$FE,$7A,$32,$FE,$82,$32,$FF,$08,$8C,$BF
	.dc.b	$C2,$A3,$38,$FF,$06,$9C,$BF,$C2,$23,$38,$FF,$02,$9C,$BF,$C1,$23
	.dc.b	$35,$FC,$0E,$72,$FF,$05,$8C,$BF,$C0,$A3,$F8,$27,$C9,$BD,$BF,$E0
	.dc.b	$DA,$3E,$2A,$9A,$89,$7F,$BD,$9D,$36,$72,$B2,$65,$FE,$44,$C3,$B4
	.dc.b	$EC,$86,$E3,$58,$72,$8F,$F0,$C4,$F0,$7F,$C2,$CD,$73,$1F,$E0,$E8
	.dc.b	$C4,$72,$3F,$C3,$4E,$10,$70,$D8,$E4,$47,$8B,$5F,$31,$CA,$3C,$95
	.dc.b	$EE,$8C,$4F,$CA,$F4,$3E,$1C,$E6,$7B,$1F,$C9,$6B,$8D,$FF,$9A,$31
	.dc.b	$5F,$08,$F5,$57,$80,$E2,$59,$DB,$F4,$FC,$CF,$B2,$47,$EB,$7B,$95
	.dc.b	$E1,$17,$1F,$25,$2B,$17,$E3,$AB,$7F,$22,$B4,$F7,$21,$D0,$EE,$62
	.dc.b	$FA,$CF,$28,$4F,$27,$81,$27,$6F,$23,$DD,$D1,$F9,$4F,$32,$47,$E4
	.dc.b	$63,$F2,$E6,$1F,$89,$18,$74,$7E,$5D,$D4,$FC,$09,$D4,$F0,$34,$70
	.dc.b	$3C,$26,$0D,$18,$24,$7C,$A9,$EB,$87,$DB,$AF,$D6,$FA,$9F,$AC,$F8
	.dc.b	$C9,$87,$E0,$63,$2F,$EA,$E3,$2F,$DF,$D5,$3A,$8F,$C0,$4E,$B1,$FC
	.dc.b	$65,$6A,$7A,$1F,$A0,$AF,$D2,$11,$C1,$0F,$D5,$B9,$69,$93,$F6,$2E
	.dc.b	$5A,$76,$3C,$CF,$F0,$8F,$37,$CD,$5E,$48,$B8,$FE,$76,$6D,$7E,$A6
	.dc.b	$6B,$F8,$48,$FE,$23,$E5,$FF,$12,$F1,$A4,$FF,$0F,$8C,$BF,$C4,$63
	.dc.b	$2F,$E8,$E3,$2F,$E9,$63,$38,$7E,$7A,$72,$FE,$82,$33,$87,$EC,$67
	.dc.b	$2F,$EC,$E3,$35,$FA,$19,$CB,$FA,$38,$CB,$F9,$C8,$FF,$17,$7C,$A7
	.dc.b	$FB,$82,$BF,$DC,$6D,$9F,$E2,$2D,$4F,$FB,$E1,$FF,$14,$39,$27,$97
	.dc.b	$D3,$CC,$EA,$C9,$DB,$96,$90,$FE,$11,$3C,$E7,$84,$62,$6E,$3F,$C5
	.dc.b	$6B,$C5,$39,$39,$46,$67,$B4,$7E,$57,$C0,$F7,$F9,$B2,$3F,$8E,$47
	.dc.b	$F4,$C3,$0F,$8A,$30,$E8,$F7,$BA,$AF,$CD,$CE,$A7,$F9,$FB,$FC,$F4
	.dc.b	$8F,$F3,$C6,$BA,$30,$43,$04,$3F,$A5,$BC,$23,$BC,$78,$CF,$8F,$E6
	.dc.b	$F8,$63,$31,$EF,$3E,$AC,$67,$2F,$EB,$95,$67,$EE,$CF,$E0,$BC,$BF
	.dc.b	$C1,$FB,$52,$7E,$EC,$65,$FB,$F1,$97,$EB,$C6,$5F,$B3,$19,$C3,$EB
	.dc.b	$4E,$5F,$B1,$19,$C3,$F1,$F3,$97,$F2,$31,$9A,$FC,$8C,$E5,$FC,$9C
	.dc.b	$65,$FC,$9C,$7E,$6B,$CB,$91,$F9,$47,$F3,$BE,$5E,$23,$F1,$B1,$70
	.dc.b	$FC,$54,$DC,$3F,$43,$36,$BF,$43,$35,$FD,$0C,$5D,$7F,$DC,$D1,$D0
	.dc.b	$F0,$10,$FC,$DC,$EE,$78,$3F,$E9,$DF,$2D,$F0,$DF,$B2,$F8,$49,$7F
	.dc.b	$E9,$39,$CB,$67,$2F,$E9,$A7,$94,$48,$FD,$9C,$6B,$2F,$EF,$55,$CA
	.dc.b	$7F,$BA,$35,$FF,$06,$9F,$E0,$E7,$0E,$1A,$E6,$B6,$C8,$CA,$69,$23
	.dc.b	$71,$CD,$AF,$32,$C6,$DB,$1E,$B7,$CE,$FF,$16,$F3,$71,$5E,$C8,$B9
	.dc.b	$FF,$18,$9B,$5D,$4E,$27,$F1,$67,$FD,$99,$F2,$AF,$33,$F9,$0C,$C3
	.dc.b	$FD,$85,$19,$6C,$65,$FE,$2A,$72,$63,$FC,$88,$D6,$1F,$E2,$99,$E5
	.dc.b	$1F,$E2,$36,$C7,$F1,$98,$DD,$FD,$29,$CB,$6B,$33,$FC,$C8,$EE,$73
	.dc.b	$72,$FF,$34,$33,$5E,$1E,$1F,$79,$23,$DB,$E6,$57,$C9,$16,$7D,$13
	.dc.b	$6F,$C9,$36,$BF,$CA,$78,$2B,$10,$8F,$F7,$33,$A9,$F9,$F3,$A9,$CC
	.dc.b	$EE,$72,$FD,$07,$46,$1F,$CB,$FB,$BF,$63,$E1,$24,$7E,$96,$3F,$54
	.dc.b	$61,$F9,$D1,$87,$47,$DC,$75,$8F,$E9,$27,$53,$F9,$6D,$51,$F9,$68
	.dc.b	$FC,$E3,$63,$31,$F9,$88,$C1,$1C,$10,$E8,$E5,$A6,$67,$C4,$F9,$8F
	.dc.b	$C3,$79,$BE,$48,$CC,$7C,$23,$F2,$B1,$9C,$FF,$6D,$56,$7E,$AC,$FE
	.dc.b	$7B,$C1,$FC,$FF,$6C,$9F,$CE,$C6,$5F,$CF,$C6,$5F,$CB,$46,$5F,$CC
	.dc.b	$46,$70,$FC,$E4,$E5,$FC,$F4,$67,$0F,$CD,$4E,$5F,$CE,$46,$6B,$F1
	.dc.b	$33,$97,$F1,$71,$97,$F1,$51,$FB,$6F,$2A,$47,$ED,$1F,$DC,$F9,$58
	.dc.b	$8F,$D6,$45,$C3,$F6,$13,$70,$FD,$64,$DA,$FC,$CC,$D7,$F3,$11,$76
	.dc.b	$FF,$75,$C7,$72,$1D,$8F,$E0,$B3,$B9,$E0,$FF,$3B,$F4,$FF,$86,$F9
	.dc.b	$12,$BF,$9D,$46,$5F,$E1,$B1,$97,$F9,$9C,$65,$FE,$0D,$19,$C7,$F9
	.dc.b	$CC,$E6,$7F,$98,$3F,$93,$23,$FC,$A9,$AF,$E6,$DA,$F0,$34,$43,$45
	.dc.b	$8E,$EF,$A0,$FE,$47,$E5,$79,$63,$31,$E8,$32,$6C,$C1,$1F,$E4,$51
	.dc.b	$FD,$39,$FF,$66,$7C,$B7,$FB,$0A,$7F,$D8,$CB,$DB,$D6,$9F,$D5,$23
	.dc.b	$13,$FE,$E2,$AE,$3E,$5E,$A3,$FD,$CB,$3C,$E7,$AD,$E6,$53,$C9,$EF
	.dc.b	$59,$8E,$17,$DE,$B0,$66,$65,$78,$67,$4D,$3C,$33,$A7,$A9,$93,$C2
	.dc.b	$F4,$71,$5E,$07,$D3,$59,$7E,$1C,$F3,$BF,$8F,$E5,$79,$C9,$3C,$BC
	.dc.b	$BA,$FC,$A8,$B3,$F3,$A6,$E3,$F2,$E6,$D7,$C8,$9A,$FC,$68,$B8,$FF
	.dc.b	$B7,$EF,$C3,$3E,$89,$CD,$69,$B1,$EF,$F9,$77,$8D,$F3,$7E,$12,$4E
	.dc.b	$63,$EA,$4E,$A3,$E1,$CE,$96,$0C,$BF,$C2,$B5,$B9,$CC,$62,$F8,$57
	.dc.b	$E3,$2F,$AC,$4A,$FF,$07,$DF,$17,$FC,$32,$D7,$91,$A8,$FE,$81,$19
	.dc.b	$80,$70,$D8,$D9,$1E,$44,$3F,$C4,$31,$E9,$AF,$64,$74,$8C,$5F,$8A
	.dc.b	$3F,$47,$E6,$7A,$5F,$79,$BB,$C5,$FE,$AA,$6D,$71,$8E,$B5,$E5,$3F
	.dc.b	$1F,$3F,$9F,$F2,$BF,$41,$EC,$C9,$FC,$9C,$65,$FC,$AC,$65,$FD,$AC
	.dc.b	$65,$FD,$BC,$66,$BF,$77,$39,$7F,$7B,$19,$BF,$DD,$4E,$5F,$DE,$46
	.dc.b	$6B,$F7,$33,$97,$F7,$71,$97,$F7,$31,$FB,$7C,$24,$7E,$A9,$FD,$D7
	.dc.b	$97,$88,$FE,$17,$17,$0F,$E1,$F3,$70,$FE,$19,$36,$BF,$84,$CD,$7F
	.dc.b	$83,$C5,$D9,$36,$30,$D9,$F1,$1F,$BC,$7F,$3B,$EF,$7F,$11,$F1,$1F
	.dc.b	$07,$4C,$D7,$E9,$E7,$0F,$F0,$79,$DB,$92,$1F,$82,$0E,$9C,$37,$2F
	.dc.b	$C0,$EC,$AE,$F5,$F9,$27,$F3,$A4,$7F,$4A,$D7,$46,$08,$60,$87,$EA
	.dc.b	$9E,$13,$E0,$7B,$BF,$92,$A7,$99,$8D,$3F,$77,$56,$FF,$5B,$5A,$7D
	.dc.b	$F7,$CD,$B5,$FE,$96,$9D,$47,$C0,$F2,$FE,$77,$3C,$8F,$95,$1F,$44
	.dc.b	$C3,$F1,$A3,$0E,$8F,$A0,$EA,$FD,$13,$A9,$E0,$68,$E0,$78,$4C,$1A
	.dc.b	$30,$48,$F8,$53,$D6,$FD,$55,$F8,$5F,$2A,$5F,$E0,$5F,$C4,$79,$78
	.dc.b	$8F,$BB,$17,$0F,$99,$36,$FE,$24,$DA,$F9,$73,$5F,$C3,$C5,$C5,$36
	.dc.b	$30,$D9,$F1,$1F,$93,$7F,$9B,$FB,$FF,$9B,$F0,$A4,$FF,$3F,$8C,$BF
	.dc.b	$99,$8C,$BF,$97,$D9,$3A,$8F,$CB,$CE,$B2,$FC,$BC,$E6,$7F,$9A,$BF
	.dc.b	$CC,$88,$FF,$3A,$6B,$FC,$B1,$AF,$03,$44,$34,$58,$EE,$FA,$4F,$D8
	.dc.b	$79,$7E,$58,$CC,$7A,$67,$F7,$53,$9C,$FF,$A9,$C5,$9F,$CC,$CF,$EF
	.dc.b	$FC,$9F,$E0,$1E,$2C,$9F,$DE,$46,$5F,$DF,$46,$5F,$DC,$C6,$5F,$DD
	.dc.b	$C6,$70,$FD,$54,$E5,$FD,$64,$67,$0F,$D2,$4E,$5F,$D3,$46,$6B,$F5
	.dc.b	$B3,$97,$F5,$F1,$97,$F8,$5C,$7F,$89,$FE,$63,$FD,$A3,$1F,$D8,$1A
	.dc.b	$9B,$B7,$F6,$48,$FF,$0F,$46,$89,$E5,$7F,$21,$8E,$8C,$76,$E5,$A4
	.dc.b	$3F,$8F,$47,$67,$81,$88,$B8,$FF,$0D,$8B,$8F,$8E,$BC,$31,$FC,$AC
	.dc.b	$FF,$21,$7C,$CB,$D5,$F7,$7C,$1C,$9F,$F1,$B4,$E5,$B3,$97,$C7,$3C
	.dc.b	$A2,$47,$C0,$8D,$61,$FD,$C2,$B9,$4F,$F3,$76,$BF,$1C,$38,$7E,$53
	.dc.b	$5F,$F0,$7B,$5D,$D7,$FB,$4E,$30,$47,$04,$8E,$C7,$27,$0E,$23,$CB
	.dc.b	$3E,$79,$F5,$CF,$D0,$F3,$2C,$E6,$3E,$19,$F7,$A3,$38,$FA,$2A,$CF
	.dc.b	$E1,$E7,$F1,$5E,$5F,$E2,$FD,$A9,$3F,$86,$8C,$BF,$88,$8C,$BF,$80
	.dc.b	$8C,$BF,$82,$8C,$E1,$FE,$1C,$CD,$5A,$8C,$4E,$B1,$FC,$14,$E5,$FC
	.dc.b	$24,$66,$FF,$39,$39,$7F,$3D,$19,$7F,$3D,$1F,$A3,$F2,$E4,$7E,$79
	.dc.b	$FD,$37,$99,$88,$FC,$BC,$5C,$3C,$D3,$70,$FC,$A4,$DA,$FD,$64,$D7
	.dc.b	$F5,$91,$71,$4D,$8C,$36,$7C,$47,$EC,$1F,$CE,$C7,$EE,$7E,$0E,$4F
	.dc.b	$E7,$23,$2F,$ED,$A3,$2F,$E5,$63,$2F,$EC,$23,$38,$7E,$6E,$73,$3F
	.dc.b	$99,$D5,$1F,$93,$24,$7E,$40,$FC,$EB,$87,$F2,$0D,$79,$35,$FD,$0B
	.dc.b	$5F,$D0,$BE,$49,$F1,$47,$F0,$EF,$33,$19,$70,$8F,$71,$92,$9E,$D2
	.dc.b	$CF,$E2,$67,$F8,$F7,$95,$FC,$7F,$79,$5F,$C6,$A3,$2F,$F1,$C8,$CB
	.dc.b	$FC,$52,$32,$FF,$16,$8C,$D7,$E5,$67,$2F,$E5,$E3,$37,$FC,$56,$72
	.dc.b	$FF,$17,$8C,$D7,$F2,$F9,$CB,$FC,$C6,$32,$FF,$31,$8F,$F6,$E7,$CD
	.dc.b	$7F,$B2,$60,$D2,$79,$5A,$F2,$E4,$FF,$B4,$E7,$2D,$8C,$BF,$E4,$D8
	.dc.b	$CB,$63,$38,$FF,$A3,$27,$2D,$8C,$E3,$FD,$0E,$72,$D8,$CD,$7F,$AA
	.dc.b	$A7,$2F,$F1,$EF,$8E,$F5,$F5,$48,$FE,$67,$1F,$CE,$0C,$3E,$58,$C3
	.dc.b	$A3,$D8,$EB,$0F,$E6,$53,$A9,$FE,$40,$FF,$29,$23,$F9,$7D,$91,$88
	.dc.b	$D9,$97,$0F,$E8,$5A,$FF,$38,$78,$4F,$82,$3C,$87,$CB,$F2,$FC,$51
	.dc.b	$98,$FE,$89,$3F,$6A,$73,$97,$39,$E3,$5F,$3E,$7E,$FF,$95,$F8,$0F
	.dc.b	$66,$57,$DB,$8C,$BF,$76,$32,$FD,$B8,$CB,$F7,$63,$35,$F6,$E7,$2F
	.dc.b	$DD,$8C,$DF,$DB,$9C,$BF,$76,$33,$5F,$6E,$72,$FD,$D8,$CB,$F9,$48
	.dc.b	$FC,$BF,$95,$23,$F2,$4F,$E6,$BC,$BC,$47,$E4,$62,$E1,$F8,$F9,$B8
	.dc.b	$7E,$36,$6D,$7E,$2E,$6B,$FA,$28,$BA,$A6,$C6,$1B,$3E,$23,$F4,$8F
	.dc.b	$F8,$6E,$FF,$DD,$3F,$36,$F8,$0F,$D7,$FC,$1C,$B1,$FE,$B3,$9D,$38
	.dc.b	$AE,$35,$B8,$FD,$65,$75,$CB,$FC,$51,$7C,$27,$95,$6E,$FF,$5D,$7B
	.dc.b	$9E,$F5,$B9,$EB,$1F,$CB,$E3,$93,$FE,$0D,$6B,$DA,$BF,$9A,$28,$CB
	.dc.b	$D2,$33,$18,$8D,$46,$08,$F7,$6B,$FE,$0E,$8F,$25,$F7,$8F,$84,$D7
	.dc.b	$F8,$67,$9B,$EB,$7A,$9A,$3C,$45,$32,$70,$31,$3D,$9D,$9F,$20,$FE
	.dc.b	$39,$1D,$7E,$73,$FD,$8B,$82,$FF,$CA,$D0,$F6,$F4,$6F,$CD,$98,$78
	.dc.b	$67,$4D,$9C,$BF,$EF,$18,$EB,$5D,$27,$9C,$F1,$AD,$DF,$86,$3F,$0F
	.dc.b	$7D,$1E,$55,$AB,$F2,$DF,$29,$C9,$CC,$86,$09,$7D,$55,$F9,$1E,$FE
	.dc.b	$5F,$C3,$23,$ED,$F3,$2B,$EB,$8B,$5C,$E9,$58,$C3,$E4,$D5,$BF,$5C
	.dc.b	$D7,$C3,$16,$B8,$C7,$CB,$8C,$1F,$56,$6C,$64,$F2,$1F,$7B,$C5,$FB
	.dc.b	$5F,$09,$23,$04,$3E,$93,$97,$F6,$51,$F8,$DD,$4E,$72,$F6,$56,$62
	.dc.b	$57,$EA,$5F,$D0,$91,$C1,$0F,$CF,$39,$69,$93,$F6,$0E,$5A,$79,$E7
	.dc.b	$F8,$5C,$7E,$57,$CD,$F0,$CF,$A0,$D6,$7F,$96,$8C,$D7,$09,$C9,$F8
	.dc.b	$A8,$FC,$F7,$95,$F9,$FB,$93,$F9,$98,$CB,$F9,$B8,$CB,$F9,$98,$CB
	.dc.b	$F9,$B8,$CD,$7E,$06,$72,$FE,$0E,$33,$7F,$AA,$9C,$BF,$AC,$8C,$D7
	.dc.b	$E7,$A7,$2F,$E8,$23,$2F,$E5,$A3,$F6,$DE,$4C,$8F,$D8,$3F,$B9,$F2
	.dc.b	$B1,$1F,$AF,$8B,$87,$E5,$66,$E1,$FB,$C9,$B5,$FA,$E9,$AF,$E9,$22
	.dc.b	$EC,$9B,$18,$6C,$F8,$8F,$E0,$4F,$E2,$F6,$FE,$19,$C5,$27,$F8,$9C
	.dc.b	$65,$FE,$11,$19,$7F,$A2,$46,$5F,$E1,$11,$9C,$7F,$A6,$CE,$67,$F9
	.dc.b	$C3,$FC,$C8,$8F,$F3,$D6,$BF,$CB,$DA,$F0,$34,$43,$45,$8E,$EF,$A0
	.dc.b	$FE,$5D,$EF,$7F,$BF,$37,$57,$9E,$B7,$47,$F8,$75,$6E,$75,$9F,$F5
	.dc.b	$A7,$F9,$E4,$3C,$BB,$E5,$63,$D9,$98,$F8,$23,$2F,$8A,$32,$F7,$C3
	.dc.b	$FB,$6A,$B5,$70,$8D,$1F,$DC,$A1,$5C,$B1,$F3,$D6,$DF,$4C,$66,$FA
	.dc.b	$D7,$4B,$F2,$1E,$32,$1D,$8F,$15,$F3,$9E,$35,$CE,$7C,$53,$F0,$BC
	.dc.b	$A9,$1F,$39,$F8,$BE,$5E,$23,$A6,$1F,$ED,$A7,$47,$D4,$8D,$63,$F4
	.dc.b	$2A,$E3,$E8,$9B,$5F,$06,$6B,$F1,$E2,$DF,$58,$EF,$3E,$27,$B9,$82
	.dc.b	$DF,$DE,$AF,$EE,$DE,$3F,$F4,$D8,$FC,$0F,$C1,$7D,$E3,$38,$CB,$FC
	.dc.b	$5C,$65,$FE,$F5,$5C,$67,$ED,$B6,$3E,$C4,$ED,$F9,$D1,$94,$FF,$5D
	.dc.b	$7F,$06,$62,$7F,$C3,$71,$98,$EB,$18,$23,$FE,$D3,$8D,$3F,$E3,$86
	.dc.b	$BC,$CD,$47,$07,$B9,$92,$C7,$94,$91,$FA,$0F,$27,$E1,$4F,$AA,$2D
	.dc.b	$FE,$A2,$6C,$79,$A7,$E2,$BC,$5F,$D9,$D7,$EA,$BD,$89,$18,$21,$F8
	.dc.b	$87,$2F,$E2,$A3,$F6,$26,$2F,$F6,$F3,$89,$E3,$18,$7F,$80,$C6,$5E
	.dc.b	$46,$08,$60,$91,$C5,$F1,$1D,$2B,$F5,$7E,$4F,$F0,$22,$47,$F0,$EF
	.dc.b	$2A,$BF,$A0,$8B,$5F,$C3,$26,$D7,$2A,$FE,$2B,$5F,$C3,$27,$31,$FB
	.dc.b	$F8,$AF,$F0,$58,$BA,$A6,$C6,$1B,$3F,$94,$9F,$F2,$97,$CE,$BE,$06
	.dc.b	$9F,$A0,$89,$3F,$E5,$39,$D3,$67,$2F,$E6,$27,$8B,$93,$B1,$0F,$C4
	.dc.b	$BF,$98,$23,$F9,$56,$BA,$30,$43,$04,$3F,$47,$9F,$F2,$57,$59,$7F
	.dc.b	$90,$E3,$11,$E5,$3D,$5E,$5F,$A2,$75,$57,$3F,$E5,$F8,$AB,$17,$FA
	.dc.b	$7A,$AA,$74,$FC,$5F,$23,$E5,$FB,$32,$30,$43,$F9,$54,$A3,$51,$89
	.dc.b	$FA,$66,$31,$FF,$2E,$56,$E7,$94,$62,$75,$8F,$A2,$F8,$D6,$A3,$0F
	.dc.b	$D9,$8C,$BD,$CC,$10,$C1,$E6,$F2,$3F,$0A,$47,$F0,$5E,$4D,$74,$7E
	.dc.b	$16,$3F,$17,$19,$C3,$D3,$16,$FE,$6C,$DA,$FF,$64,$E2,$AC,$47,$CD
	.dc.b	$9B,$39,$8E,$C7,$88,$A7,$88,$97,$F2,$27,$F9,$CE,$1F,$9E,$CA,$57
	.dc.b	$F3,$48,$CB,$F9,$D8,$CB,$FC,$CA,$32,$FD,$88,$CD,$7E,$36,$73,$3F
	.dc.b	$92,$7F,$99,$91,$FE,$60,$D7,$F9,$4B,$5E,$06,$88,$68,$B1,$DD,$F4
	.dc.b	$1F,$B3,$F2,$3C,$B1,$98,$F4,$19,$36,$60,$8F,$E3,$23,$F8,$0F,$8D
	.dc.b	$27,$F7,$CF,$F0,$4F,$5F,$11,$F9,$88,$B5,$F9,$49,$BA,$7E,$A2,$6D
	.dc.b	$7D,$C9,$AF,$E9,$E2,$E6,$9B,$18,$6C,$F8,$8F,$E2,$0F,$FB,$9F,$E7
	.dc.b	$FF,$D7,$90,$7F,$D9,$58,$7F,$B0,$5B,$87,$F8,$A1,$D9,$3C,$8E,$47
	.dc.b	$3F,$22,$5F,$FB,$BE,$78,$A3,$B4,$E9,$FF,$11,$47,$37,$84,$66,$FC
	.dc.b	$55,$DA,$72,$75,$75,$7E,$5A,$D4,$F4,$9F,$F1,$FD,$7F,$B6,$FE,$7F
	.dc.b	$F9,$D4,$F9,$7E,$0B,$E1,$1A,$C3,$FC,$8B,$3A,$9F,$F1,$B3,$6B,$E2
	.dc.b	$57,$09,$F6,$47,$5E,$6E,$51,$EF,$AE,$17,$F2,$6F,$84,$F4,$AD,$AA
	.dc.b	$F0,$3F,$E1,$46,$BE,$48,$FF,$04,$BD,$5A,$F1,$6C,$70,$8D,$1C,$1C
	.dc.b	$3B,$31,$1C,$08,$E0,$F6,$D7,$85,$A7,$43,$EC,$F8,$1E,$AA,$F7,$C6
	.dc.b	$70,$F8,$8F,$E0,$E7,$57,$23,$99,$E3,$39,$9F,$89,$F2,$3F,$2C,$47
	.dc.b	$F1,$9E,$C5,$7F,$13,$16,$FF,$31,$37,$1F,$CC,$4D,$AF,$C9,$6A,$AC
	.dc.b	$47,$E8,$66,$CE,$63,$C4,$43,$B9,$E2,$31,$7F,$A7,$9E,$35,$FA,$DF
	.dc.b	$23,$F5,$DE,$4C,$BF,$D2,$46,$5F,$D3,$46,$5F,$D1,$C6,$5F,$D2,$C6
	.dc.b	$6F,$F6,$93,$97,$F6,$D1,$9C,$7F,$6D,$39,$7F,$71,$19,$AF,$83,$39
	.dc.b	$7E,$1C,$65,$FE,$05,$1F,$6F,$E5,$7F,$83,$7C,$14,$9F,$AF,$B2,$75
	.dc.b	$1F,$BF,$9D,$3F,$A8,$8C,$BF,$BE,$8C,$E3,$FB,$39,$CC,$FE,$7D,$FC
	.dc.b	$C9,$1F,$CC,$35,$FF,$34,$D7,$85,$C9,$0C,$C6,$DE,$86,$67,$C0,$FA
	.dc.b	$A3,$F9,$17,$95,$E6,$9C,$CF,$A4,$FE,$35,$19,$CB,$F8,$2D,$59,$FC
	.dc.b	$5C,$FE,$EB,$E8,$3F,$DA,$D1,$F9,$C6,$BF,$EB,$F6,$DF,$F2,$D8,$E1
	.dc.b	$E4,$47,$83,$FC,$D5,$C1,$C5,$B1,$C6,$36,$43,$47,$F3,$13,$78,$7F
	.dc.b	$B8,$2B,$C2,$DA,$ED,$3C,$69,$39,$8C,$3F,$D0,$5F,$77,$8F,$EF,$F2
	.dc.b	$24,$7B,$E3,$2F,$C1,$8C,$BF,$06,$32,$FC,$38,$CD,$78,$23,$FA,$36
	.dc.b	$1F,$EE,$97,$81,$CC,$86,$09,$1C,$54,$7F,$BE,$1D,$C6,$63,$A3,$6B
	.dc.b	$C3,$7F,$16,$FC,$B3,$FD,$1F,$E0,$BE,$E1,$30,$FD,$0B,$FC,$7C,$C3
	.dc.b	$F7,$5C,$3A,$3F,$04,$EA,$3A,$46,$4F,$CD,$C7,$F3,$82,$3F,$CD,$DA
	.dc.b	$E8,$C1,$0C,$10,$FE,$90,$F0,$8E,$F1,$FD,$6A,$3F,$2D,$E4,$F9,$A3
	.dc.b	$07,$8E,$7E,$34,$65,$E2,$4B,$FC,$54,$7E,$7F,$C7,$95,$F9,$F7,$F4
	.dc.b	$5E,$C6,$23,$F4,$11,$6F,$E9,$4D,$BF,$CA,$4D,$AF,$C0,$CD,$7E,$DC
	.dc.b	$5C,$D3,$63,$0D,$9F,$11,$FB,$17,$F6,$9E,$2F,$ED,$7C,$89,$3F,$B5
	.dc.b	$8C,$BF,$B7,$8C,$BF,$B7,$8C,$BF,$B9,$8C,$E3,$F8,$B9,$CB,$F8,$D8
	.dc.b	$CE,$3F,$AF,$9C,$BF,$B1,$8C,$D7,$E7,$A7,$2F,$E8,$23,$2F,$E6,$A3
	.dc.b	$FC,$0B,$F4,$3F,$E2,$98,$FE,$19,$F0,$5F,$C3,$CC,$B0,$C9,$06,$1F
	.dc.b	$A5,$9E,$51,$8C,$3A,$46,$27,$F4,$D5,$D6,$BF,$BA,$CE,$6B,$71,$FE
	.dc.b	$10,$8F,$F0,$8A,$69,$1D,$10,$FF,$0C,$B9,$69,$93,$CA,$73,$31,$5E
	.dc.b	$97,$DC,$7F,$2B,$EA,$B1,$98,$F3,$9F,$CC,$E3,$35,$E8,$8C,$9F,$CD
	.dc.b	$63,$FC,$63,$C7,$FE,$D0,$8F,$88,$7F,$95,$DC,$57,$F0,$D6,$D7,$D3
	.dc.b	$5E,$CF,$2F,$26,$5F,$F8,$72,$76,$F3,$9C,$BD,$8F,$ED,$53,$8B,$ED
	.dc.b	$36,$FB,$4D,$AE,$47,$57,$0F,$23,$C1,$38,$BF,$0C,$64,$F0,$3E,$87
	.dc.b	$CD,$E4,$FC,$2F,$1A,$46,$08,$70,$72,$F2,$8F,$92,$63,$1F,$8D,$38
	.dc.b	$9E,$31,$87,$E2,$46,$5E,$46,$08,$60,$91,$C5,$F1,$1D,$31,$FC,$CF
	.dc.b	$57,$DA,$F8,$29,$3F,$98,$8F,$D4,$98,$7E,$0C,$61,$D1,$F5,$DD,$57
	.dc.b	$E9,$E7,$53,$F9,$4D,$51,$FA,$12,$47,$E7,$1A,$ED,$AF,$EC,$23,$8B
	.dc.b	$B2,$1B,$2C,$F8,$5F,$29,$F6,$3C,$AF,$1C,$66,$3F,$15,$3D,$1D,$1C
	.dc.b	$23,$44,$C3,$F3,$15,$FA,$0F,$22,$57,$E5,$DF,$D1,$FA,$D8,$8F,$C3
	.dc.b	$C5,$BF,$D1,$4D,$BF,$D1,$4D,$AF,$D4,$4D,$7F,$C8,$D1,$75,$4D,$8C
	.dc.b	$36,$7C,$47,$EA,$5F,$DA,$F9,$1F,$B6,$F2,$24,$FE,$B6,$32,$FE,$BE
	.dc.b	$32,$FE,$7A,$32,$FE,$82,$33,$8F,$EE,$E7,$2F,$EF,$63,$38,$FE,$F6
	.dc.b	$72,$FE,$FE,$33,$5F,$C1,$67,$2F,$F0,$78,$CB,$FE,$65,$8F,$E7,$3F
	.dc.b	$39,$FE,$BE,$FA,$27,$C1,$D3,$35,$FC,$BA,$70,$FF,$A9,$27,$6E,$48
	.dc.b	$7F,$09,$72,$D9,$E6,$FE,$12,$7F,$9B,$11,$FE,$62,$D7,$F3,$8D,$7F
	.dc.b	$2C,$D7,$FA,$2B,$5F,$F7,$D6,$C8,$E5,$97,$F4,$D7,$C2,$FF,$8C,$ED
	.dc.b	$E5,$63,$3E,$78,$7A,$AB,$53,$98,$EB,$75,$5F,$27,$E2,$79,$3F,$2C
	.dc.b	$8F,$C7,$F6,$2B,$F1,$62,$E1,$E3,$9B,$8F,$F6,$BB,$56,$2C,$78,$2B
	.dc.b	$0F,$8A,$2C,$A6,$C6,$1B,$3E,$63,$C8,$7D,$BF,$27,$F0,$5E,$4C,$8C
	.dc.b	$10,$FB,$8E,$5F,$BD,$1F,$89,$31,$8F,$CD,$9C,$4F,$18,$C3,$F8,$98
	.dc.b	$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$3F,$E0,$6D,$7F,$35,$F0,$32
	.dc.b	$7F,$C0,$B1,$FE,$16,$30,$FE,$56,$30,$E8,$FC,$E3,$AA,$FF,$0A,$4E
	.dc.b	$A7,$F9,$73,$FE,$0D,$23,$FE,$08,$6B,$A3,$04,$30,$43,$F9,$83,$C2
	.dc.b	$3B,$C7,$8C,$FD,$57,$2F,$86,$33,$1F,$CB,$27,$F7,$13,$9B,$F8,$53
	.dc.b	$C6,$BF,$3D,$3F,$BD,$F1,$E4,$FE,$E5,$FE,$01,$EB,$E2,$3F,$71,$16
	.dc.b	$BF,$4D,$36,$FF,$29,$36,$BF,$27,$B2,$B1,$1F,$A4,$9B,$39,$8F,$01
	.dc.b	$0E,$E7,$F0,$B9,$DC,$F0,$7F,$C9,$97,$FE,$A9,$FA,$3F,$F6,$2D,$78
	.dc.b	$0F,$E3,$06,$B0,$FF,$29,$46,$A0,$DC,$3F,$C8,$D1,$C4,$FD,$5D,$75
	.dc.b	$F2,$79,$47,$EB,$AB,$83,$D8,$DD,$F6,$9D,$C7,$67,$4F,$F2,$13,$33
	.dc.b	$E3,$39,$BF,$CA,$4C,$C4,$70,$43,$FC,$65,$1E,$BF,$A4,$FE,$AB,$1E
	.dc.b	$23,$CD,$F0,$7D,$8F,$1C,$6A,$3F,$C7,$4D,$C3,$E5,$4E,$E7,$90,$9D
	.dc.b	$61,$FE,$7B,$6C,$76,$37,$6A,$EF,$D1,$27,$FC,$CA,$73,$6B,$4C,$D7
	.dc.b	$95,$FF,$3F,$B5,$F0,$47,$08,$C4,$7F,$20,$6B,$CC,$FF,$39,$B8,$24
	.dc.b	$71,$3C,$91,$F0,$9F,$A7,$E5,$E3,$1F,$F7,$A5,$7B,$AB,$A4,$7E,$3E
	.dc.b	$75,$8F,$92,$7E,$3D,$6F,$1F,$F7,$BC,$FC,$7F,$85,$F4,$12,$BD,$ED
	.dc.b	$7C,$ED,$7E,$0B,$5E,$47,$02,$1E,$EE,$DF,$88,$47,$E5,$B5,$D1,$82
	.dc.b	$18,$21,$F4,$1E,$0F,$D2,$F5,$7E,$81,$1F,$AE,$D7,$E7,$B5,$E0,$68
	.dc.b	$86,$8A,$7D,$EF,$9E,$FD,$1F,$9B,$23,$E5,$C7,$E6,$0C,$3F,$A0,$8C
	.dc.b	$3A,$3F,$22,$EB,$4F,$B3,$3A,$9F,$C1,$BF,$89,$23,$F9,$06,$BA,$30
	.dc.b	$43,$04,$3D,$06,$63,$BB,$E3,$3F,$6B,$E6,$F8,$63,$31,$FB,$09,$FE
	.dc.b	$03,$39,$CB,$F5,$91,$67,$F8,$74,$FE,$E3,$E0,$FF,$83,$FB,$52,$7F
	.dc.b	$1F,$19,$7F,$63,$19,$7F,$7B,$19,$7F,$83,$C6,$70,$FC,$EC,$E6,$7F
	.dc.b	$5E,$FE,$B4,$8F,$ED,$1A,$FE,$B1,$AF,$03,$44,$34,$58,$EE,$FA,$0F
	.dc.b	$E0,$FE,$77,$96,$33,$1E,$83,$26,$CC,$11,$FE,$81,$1F,$C7,$3C,$8F
	.dc.b	$E5,$7F,$09,$2B,$F6,$71,$97,$F8,$AC,$65,$FE,$43,$19,$7F,$95,$C6
	.dc.b	$6B,$F8,$0C,$E6,$7F,$89,$BF,$C4,$48,$FF,$18,$6B,$FC,$41,$AF,$03
	.dc.b	$44,$34,$58,$EE,$FA,$0F,$E5,$7E,$6F,$96,$33,$1E,$83,$26,$CC,$11
	.dc.b	$FE,$C9,$1F,$D4,$BE,$7D,$E6,$FF,$50,$F3,$A5,$FF,$3E,$9C,$B6,$72
	.dc.b	$FF,$4E,$9E,$51,$23,$F9,$FC,$6A,$DB,$FF,$94,$A5,$7D,$14,$E9,$FF
	.dc.b	$2B,$C6,$5E,$07,$0B,$FE,$8B,$19,$8E,$8D,$78,$C6,$88,$FF,$4E,$8D
	.dc.b	$3C,$CC,$46,$9E,$E6,$4B,$1E,$79,$F3,$46,$5B,$19,$AF,$F0,$5D,$7C
	.dc.b	$3A,$F6,$C7,$9A,$31,$5F,$E1,$56,$C6,$67,$E7,$1F,$DA,$23,$EB,$7C
	.dc.b	$AF,$F8,$23,$E9,$5E,$ED,$33,$5F,$5E,$70,$FF,$76,$9D,$B9,$21,$E4
	.dc.b	$3F,$D0,$AE,$9C,$07,$18,$7B,$A7,$BC,$FC,$47,$E9,$91,$F9,$CD,$74
	.dc.b	$60,$86,$08,$7D,$EC,$51,$C6,$7C,$13,$E8,$7E,$8D,$3D,$1F,$1C,$6A
	.dc.b	$7D,$37,$C9,$F8,$53,$AC,$BF,$0F,$85,$9F,$83,$3F,$A3,$DB,$E7,$70
	.dc.b	$49,$FD,$14,$65,$F9,$91,$97,$BC,$65,$F9,$39,$27,$59,$7E,$1A,$B5
	.dc.b	$5E,$F7,$F3,$64,$7F,$2C,$D7,$F2,$2D,$78,$1A,$21,$A2,$C7,$77,$D2
	.dc.b	$7E,$0F,$D1,$F3,$46,$63,$D2,$64,$D9,$82,$3F,$42,$3F,$84,$7A,$7F
	.dc.b	$93,$F4,$65,$7F,$07,$8C,$BF,$91,$8C,$BF,$52,$32,$FC,$68,$CD,$7E
	.dc.b	$AA,$73,$3F,$83,$7F,$7E,$47,$F7,$4D,$7F,$6A,$D7,$81,$A2,$1A,$2C
	.dc.b	$77,$7D,$07,$E9,$FA,$7C,$B1,$98,$F4,$19,$36,$60,$8F,$E5,$23,$F9
	.dc.b	$67,$37,$ED,$FD,$D9,$5F,$CA,$E3,$2F,$ED,$63,$2F,$E6,$23,$2F,$E2
	.dc.b	$A3,$35,$FC,$3A,$73,$3F,$A7,$7F,$91,$91,$FE,$3C,$D7,$F8,$CB,$5E
	.dc.b	$06,$88,$68,$B1,$DD,$F4,$1F,$C2,$FD,$1F,$2C,$66,$3D,$06,$4D,$98
	.dc.b	$23,$FB,$88,$FE,$A1,$F0,$DF,$C6,$FE,$16,$57,$F4,$F8,$CB,$FC,$66
	.dc.b	$32,$FE,$F2,$32,$FE,$BA,$33,$5F,$CC,$A7,$33,$FC,$2D,$FE,$8E,$47
	.dc.b	$FA,$13,$5F,$E7,$AD,$78,$1A,$21,$A2,$C7,$77,$D0,$7F,$2F,$F4,$BC
	.dc.b	$B1,$98,$F4,$19,$36,$60,$8F,$F1,$C8,$FE,$E9,$E2,$7F,$B1,$BC,$97
	.dc.b	$95,$4C,$DF,$F7,$59,$C3,$FE,$B4,$9D,$B9,$21,$FC,$F1,$CB,$67,$9B
	.dc.b	$FC,$A2,$7F,$B6,$91,$FE,$D0,$D7,$FB,$1B,$5F,$E7,$0D,$7F,$B9,$B5
	.dc.b	$FE,$F6,$F1,$BF,$EC,$EF,$81,$FE,$D7,$4F,$47,$1C,$FC,$AF,$D3,$5A
	.dc.b	$9C,$C7,$43,$33,$FE,$0D,$AE,$51,$FE,$D6,$FA,$6F,$F7,$CA,$87,$BD
	.dc.b	$F3,$4F,$F7,$9C,$E8,$DC,$6A,$7E,$64,$E9,$86,$08,$7F,$68,$9E,$4F
	.dc.b	$FB,$69,$FF,$6D,$A6,$91,$D1,$0F,$F7,$1B,$96,$99,$3F,$C2,$AE,$5F
	.dc.b	$01,$E4,$3E,$F7,$A5,$71,$FF,$0B,$CF,$92,$7F,$0D,$39,$C3,$FC,$69
	.dc.b	$39,$9E,$27,$C9,$F3,$BF,$25,$F0,$D2,$30,$43,$F1,$8E,$5F,$85,$1F
	.dc.b	$30,$C6,$1F,$7A,$71,$3C,$4F,$F1,$ED,$2B,$E4,$12,$3E,$1B,$5F,$82
	.dc.b	$D7,$8B,$5F,$5B,$51,$F4,$5F,$0B,$E6,$3F,$51,$E9,$62,$FE,$3C,$DC
	.dc.b	$3F,$19,$36,$B9,$DF,$9D,$F8,$57,$F9,$A9,$FC,$5F,$D1,$7E,$E7,$D2
	.dc.b	$95,$F9,$38,$CB,$FB,$78,$CB,$F8,$98,$CB,$FB,$78,$CD,$7D,$A9,$CC
	.dc.b	$FC,$87,$F1,$C4,$7F,$12,$D7,$F0,$ED,$78,$1A,$21,$A2,$C7,$77,$D0
	.dc.b	$7F,$0D,$D7,$CB,$19,$8F,$41,$93,$66,$08,$FE,$FA,$3F,$5F,$5F,$C7
	.dc.b	$7D,$E9,$5F,$B7,$8C,$BF,$C6,$E3,$2F,$EB,$63,$2F,$F1,$B8,$CD,$7E
	.dc.b	$7A,$73,$3F,$8C,$7F,$64,$47,$F5,$AD,$7F,$56,$D7,$81,$A2,$1A,$2C
	.dc.b	$77,$7D,$07,$F3,$1C,$FC,$B1,$98,$F4,$19,$36,$60,$8F,$F2,$28,$FE
	.dc.b	$27,$EA,$7F,$41,$F8,$69,$5F,$C6,$E3,$2F,$F3,$F8,$CB,$FC,$46,$32
	.dc.b	$FF,$3F,$8C,$D7,$F0,$29,$CC,$FE,$C1,$FE,$2C,$47,$F8,$8B,$5F,$E1
	.dc.b	$ED,$78,$1A,$21,$A2,$C7,$77,$D0,$7E,$C9,$FF,$54,$FD,$3B,$D4,$F3
	.dc.b	$C6,$E7,$D4,$6F,$7F,$F6,$45,$6E,$39,$9A,$24,$F4,$AF,$E7,$7C,$3F
	.dc.b	$EC,$F8,$EF,$C9,$70,$E6,$EA,$FF,$A3,$56,$5F,$E0,$F8,$27,$10,$8C
	.dc.b	$3C,$08,$7F,$AA,$1C,$B6,$3B,$3F,$D1,$67,$FA,$21,$1F,$E7,$ED,$7F
	.dc.b	$9E,$B5,$FE,$4A,$D7,$F9,$BB,$5F,$E2,$6F,$1B,$FE,$7E,$F8,$1F,$9F
	.dc.b	$E9,$DC,$38,$D7,$A2,$72,$D8,$CB,$E7,$3E,$91,$F1,$27,$FB,$3F,$CE
	.dc.b	$FC,$4C,$5E,$85,$33,$5F,$DC,$67,$0F,$23,$E5,$46,$9F,$ED,$11,$A7
	.dc.b	$E5,$46,$70,$FE,$B3,$3F,$DA,$C9,$3F,$D9,$9A,$FF,$64,$6B,$FD,$31
	.dc.b	$AF,$F5,$E6,$BF,$CE,$9E,$31,$DA,$3D,$07,$9E,$9D,$BE,$3C,$79,$4E
	.dc.b	$8F,$C0,$46,$5E,$86,$6B,$F3,$F1,$FE,$0E,$E9,$F9,$1E,$7C,$AF,$F0
	.dc.b	$B4,$65,$F0,$46,$5F,$F0,$64,$65,$FC,$74,$66,$BF,$BE,$CE,$67,$FB
	.dc.b	$5B,$FE,$11,$23,$FE,$0C,$6B,$FE,$0A,$6B,$C0,$D1,$0D,$16,$3B,$BE
	.dc.b	$83,$E3,$79,$FE,$58,$CC,$7A,$0C,$9B,$30,$47,$F8,$1C,$7E,$BB,$EA
	.dc.b	$1D,$BF,$DE,$2A,$1F,$DE,$A3,$2E,$0C,$45,$69,$63,$6E,$DA,$F4,$6B
	.dc.b	$DD,$AF,$23,$84,$75,$51,$A6,$C6,$61,$EE,$52,$4F,$8D,$AF,$8D,$AF
	.dc.b	$8D,$AF,$8D,$AF,$C1,$7B,$78,$12,$3D,$CD,$7D,$CD,$7D,$CD,$7D,$CD
	.dc.b	$7E,$63,$5F,$6F,$99,$EB,$23,$F2,$DA,$FC,$B6,$BF,$2D,$AF,$CB,$6B
	.dc.b	$F6,$5F,$8D,$E1,$48,$FA,$CD,$7E,$B3,$5F,$AC,$D7,$EB,$35,$FC,$23
	.dc.b	$5F,$AB,$EB,$FD,$32,$3F,$83,$6B,$F8,$36,$BF,$83,$6B,$F8,$36,$BF
	.dc.b	$91,$7E,$F7,$AB,$FC,$87,$CA,$93,$FC,$77,$24,$FE,$6C,$CC,$7F,$45
	.dc.b	$9C,$3B,$8C,$EB,$FD,$92,$73,$28,$70,$FE,$56,$BC,$33,$F9,$D6,$BF
	.dc.b	$9D,$6B,$F9,$D6,$BC,$4E,$04,$38,$16,$7B,$BF,$AF,$8F,$1C,$7F,$78
	.dc.b	$F4,$AD,$73,$AF,$33,$67,$51,$ED,$21,$A3,$A1,$FC,$B6,$7F,$80,$79
	.dc.b	$5F,$D2,$FD,$29,$3F,$2E,$32,$FF,$33,$8C,$BC,$CF,$D9,$C6,$9F,$EE
	.dc.b	$51,$AA,$F5,$C6,$8E,$87,$F0,$59,$FD,$C1,$1C,$10,$FE,$10,$E5,$A6
	.dc.b	$4F,$E1,$4E,$5A,$76,$3C,$D9,$23,$FB,$47,$A5,$E7,$AF,$2C,$E6,$B8
	.dc.b	$47,$F8,$9A,$75,$8F,$F1,$5B,$B5,$FE,$16,$9F,$E5,$1E,$F7,$F8,$5B
	.dc.b	$EA,$5E,$A5,$33,$5F,$95,$9C,$3F,$DC,$E7,$6E,$48,$7C,$D7,$2D,$9E
	.dc.b	$6F,$B2,$7F,$90,$11,$FE,$60,$D7,$F9,$83,$5F,$E6,$0D,$7F,$98,$35
	.dc.b	$FE,$7C,$F1,$C7,$E2,$BE,$07,$ED,$D3,$D3,$C6,$5C,$AF,$D3,$5A,$9C
	.dc.b	$C7,$43,$33,$FD,$36,$B9,$47,$EE,$67,$E5,$7A,$52,$30,$43,$E2,$39
	.dc.b	$7E,$EC,$7F,$59,$31,$7E,$69,$C4,$F1,$3F,$A8,$4F,$F4,$72,$3F,$D5
	.dc.b	$DA,$FF,$57,$6B,$C0,$D1,$0D,$16,$7F,$17,$1E,$08,$FB,$5E,$95,$9C
	.dc.b	$C7,$AD,$F9,$71,$9C,$3F,$B1,$55,$AF,$F7,$4E,$0A,$FE,$ED,$F4,$BF
	.dc.b	$8F,$D6,$57,$F2,$18,$CB,$F8,$D8,$CB,$FA,$28,$CB,$F8,$D8,$CE,$3F
	.dc.b	$A2,$9C,$CF,$F7,$57,$FB,$61,$1F,$EF,$8D,$7F,$BE,$35,$E0,$68,$86
	.dc.b	$8B,$1D,$DF,$41,$FA,$3F,$4B,$CB,$19,$8F,$41,$93,$66,$08,$FE,$52
	.dc.b	$3F,$D0,$FF,$54,$E1,$FD,$A7,$A7,$2F,$FC,$C5,$39,$6C,$E5,$FD,$84
	.dc.b	$F2,$89,$1F,$B7,$8D,$5F,$F1,$CA,$E5,$3F,$C8,$DA,$FF,$8A,$1A,$FF
	.dc.b	$8A,$DA,$83,$86,$D7,$F8,$7E,$30,$47,$04,$8F,$F1,$84,$65,$B1,$E8
	.dc.b	$7C,$8F,$F1,$5E,$7C,$57,$A6,$2E,$DF,$B3,$9B,$5E,$5B,$F6,$DE,$E3
	.dc.b	$F8,$C4,$FC,$EF,$82,$FE,$41,$D5,$27,$E6,$C6,$5F,$E3,$D1,$97,$E3
	.dc.b	$46,$5F,$E3,$D1,$9A,$F9,$93,$99,$F9,$4F,$C3,$23,$F0,$1A,$FF,$94
	.dc.b	$ED,$18,$8E,$26,$C8,$6D,$E3,$3E,$08,$F4,$9F,$CD,$7B,$3C,$D1,$98
	.dc.b	$F4,$99,$36,$60,$8F,$F2,$68,$FC,$9F,$6F,$FB,$6B,$EA,$DD,$F4,$CD
	.dc.b	$FE,$52,$70,$FF,$AD,$E7,$6E,$48,$7F,$71,$72,$D9,$E6,$FE,$4E,$7F
	.dc.b	$14,$47,$F0,$ED,$7E,$43,$5F,$CC,$35,$FC,$C3,$5F,$C2,$3C,$6F,$F4
	.dc.b	$4F,$81,$FE,$FF,$4F,$53,$1A,$F2,$BF,$4D,$6A,$73,$1D,$0C,$CF,$FA
	.dc.b	$46,$B9,$47,$EE,$3E,$9B,$E5,$7A,$B2,$30,$43,$E2,$39,$7F,$69,$1F
	.dc.b	$C1,$0C,$61,$E6,$9C,$4F,$13,$F7,$13,$FB,$12,$3F,$AE,$6B,$F8,$F6
	.dc.b	$BC,$0D,$10,$D1,$67,$F8,$44,$78,$23,$ED,$79,$36,$73,$1E,$B7,$E5
	.dc.b	$C6,$70,$FE,$15,$56,$BF,$85,$60,$AF,$E4,$3E,$47,$E3,$FD,$59,$5F
	.dc.b	$C8,$23,$2F,$E3,$63,$2F,$F1,$98,$CB,$F8,$D8,$CE,$3F,$C7,$E7,$33
	.dc.b	$FC,$75,$FE,$2C,$47,$F8,$9B,$5F,$DA,$35,$E0,$68,$86,$8B,$1D,$DF
	.dc.b	$41,$FA,$3F,$53,$CB,$19,$8F,$41,$93,$66,$08,$FE,$52,$3F,$A2,$FC
	.dc.b	$37,$EC,$FD,$29,$5F,$D1,$23,$2F,$EC,$63,$2F,$F3,$D8,$CB,$FB,$18
	.dc.b	$CD,$7F,$43,$9C,$CF,$F4,$17,$F9,$D1,$1F,$E6,$ED,$7F,$8C,$35,$E0
	.dc.b	$68,$86,$8B,$1D,$DF,$41,$FC,$1F,$D4,$F2,$C6,$63,$D0,$64,$D9,$82
	.dc.b	$3F,$B8,$8F,$ED,$3E,$87,$F1,$7F,$52,$57,$F6,$88,$CB,$FC,$56,$32
	.dc.b	$FF,$61,$8C,$BF,$C5,$63,$35,$FD,$9E,$73,$3F,$D9,$5F,$EB,$C4,$7F
	.dc.b	$AD,$B5,$FE,$78,$D7,$81,$A2,$1A,$2C,$77,$7D,$07,$F2,$BF,$53,$CB
	.dc.b	$19,$8F,$41,$93,$66,$08,$FF,$1C,$8F,$EF,$11,$FE,$A1,$A7,$CB,$7F
	.dc.b	$3D,$F5,$64,$FF,$7D,$BF,$F5,$0D,$6A,$31,$C5,$CA,$31,$1F,$CE,$EB
	.dc.b	$8B,$D6,$FB,$DF,$27,$2F,$F2,$E9,$CD,$F3,$9B,$5F,$E0,$DB,$EA,$FA
	.dc.b	$03,$87,$81,$93,$84,$56,$96,$38,$38,$38,$38,$7A,$C7,$16,$9C,$5C
	.dc.b	$46,$CF,$83,$7E,$D9,$F8,$A7,$57,$FD,$93,$E6,$32,$F6,$CF,$43,$E3
	.dc.b	$1C,$0C,$4F,$08,$DD,$76,$C3,$DF,$5C,$0F,$F6,$D5,$F5,$79,$4F,$C6
	.dc.b	$7A,$43,$D4,$F0,$B6,$37,$3D,$8C,$CC,$8C,$90,$EE,$E5,$FC,$54,$71
	.dc.b	$7E,$73,$F1,$88,$E0,$87,$D0,$72,$D3,$27,$DF,$7D,$4E,$49,$87,$E5
	.dc.b	$23,$C1,$1F,$0B,$D4,$B3,$98,$F4,$3F,$4A,$33,$87,$E6,$6A,$D7,$F9
	.dc.b	$6F,$05,$7D,$0F,$89,$F8,$FC,$52,$BE,$6C,$65,$F9,$B1,$97,$F5,$91
	.dc.b	$97,$EA,$C6,$71,$FD,$5C,$E6,$7D,$CF,$E3,$08,$FE,$5D,$AF,$E5,$9A
	.dc.b	$F0,$34,$43,$45,$8E,$EF,$A0,$FC,$1F,$A7,$E5,$8C,$C7,$A0,$C9,$B3
	.dc.b	$04,$7F,$37,$1F,$94,$E3,$FC,$5F,$A7,$2B,$F2,$31,$97,$F2,$31,$97
	.dc.b	$F8,$84,$65,$FC,$CC,$66,$BF,$87,$CE,$67,$F0,$4F,$EC,$08,$FE,$ED
	.dc.b	$AF,$EE,$9A,$F0,$34,$43,$45,$8E,$EF,$A0,$FD,$3F,$5F,$96,$33,$1E
	.dc.b	$83,$26,$CC,$11,$FD,$FC,$7E,$E3,$E1,$7F,$5F,$E9,$4A,$FD,$AC,$65
	.dc.b	$FD,$AC,$65,$FE,$69,$19,$7F,$7B,$19,$AF,$E6,$73,$99,$FD,$33,$FC
	.dc.b	$50,$8F,$F1,$F6,$BF,$C7,$9A,$F0,$34,$43,$45,$8E,$EF,$A0,$FE,$17
	.dc.b	$E8,$F9,$63,$31,$E8,$32,$6C,$C1,$1F,$E4,$71,$FC,$73,$B7,$F8,$9F
	.dc.b	$8B,$2B,$F8,$CC,$65,$FE,$33,$19,$7F,$AC,$46,$5F,$E4,$31,$9A,$FE
	.dc.b	$AF,$39,$9F,$E1,$4F,$F3,$82,$3F,$D0,$DA,$FF,$42,$6B,$C0,$D1,$0D
	.dc.b	$16,$3B,$BE,$83,$F9,$7F,$A1,$E5,$8C,$C7,$A0,$C9,$B3,$04,$7F,$A3
	.dc.b	$C7,$F4,$0F,$03,$FA,$79,$27,$FA,$D3,$5F,$EB,$2D,$7F,$94,$35,$FE
	.dc.b	$EC,$D7,$FB,$63,$FD,$6B,$E2,$E4,$7F,$65,$6B,$FD,$91,$AF,$F3,$36
	.dc.b	$BF,$E0,$16,$BF,$DD,$DA,$BE,$B1,$E7,$FF,$60,$5F,$01,$4C,$44,$9F
	.dc.b	$EC,$53,$A6,$CE,$56,$2E,$8F,$F0,$DD,$76,$9D,$90,$DC,$6A,$F9,$57
	.dc.b	$19,$1E,$7B,$F2,$3E,$29,$DD,$76,$9C,$CF,$43,$8D,$F8,$5C,$1F,$06
	.dc.b	$B7,$7D,$E3,$B1,$D4,$FF,$00,$7D,$6B,$C7,$FF,$08,$36,$3F,$C3,$4D
	.dc.b	$7F,$C3,$CD,$7F,$C4,$8D,$7F,$C7,$AD,$7D,$AD,$4D,$24,$78,$09,$1E
	.dc.b	$0F,$63,$C6,$47,$04,$3C,$6E,$5A,$64,$F2,$B9,$69,$EC,$F2,$FD,$E4
	.dc.b	$70,$43,$DE,$E5,$A6,$4F,$82,$E5,$A7,$C8,$F6,$3E,$59,$1C,$10,$F9
	.dc.b	$6E,$5A,$64,$F9,$AE,$5A,$7D,$4E,$CF,$CE,$7A,$92,$3E,$DC,$7F,$8F
	.dc.b	$B1,$39,$8F,$C2,$E8,$AD,$4C,$9F,$D0,$C6,$F3,$FC,$45,$F2,$9E,$A7
	.dc.b	$84,$87,$E1,$2B,$F1,$24,$72,$7E,$19,$C3,$B3,$27,$E2,$1C,$CF,$81
	.dc.b	$F2,$47,$E7,$25,$F0,$58,$AD,$4F,$90,$D4,$67,$7E,$91,$99,$F5,$CE
	.dc.b	$63,$F2,$B7,$5F,$CA,$C5,$BF,$81,$1E,$19,$F9,$93,$97,$4D,$8E,$D7
	.dc.b	$FA,$28,$C4,$71,$3F,$5E,$E2,$FD,$31,$E8,$32,$72,$3C,$A6,$0D,$1C
	.dc.b	$C8,$FB,$1D,$3D,$09,$3E,$38,$CB,$F1,$1E,$2E,$5E,$B1,$C1,$F0,$98
	.dc.b	$8D,$46,$63,$F7,$75,$F1,$1F,$31,$FC,$12,$BF,$85,$91,$FE,$16,$D7
	.dc.b	$F8,$5B,$5E,$06,$88,$68,$B3,$EF,$7C,$87,$F0,$ED,$FC,$4E,$0F,$B6
	.dc.b	$7F,$1E,$73,$1C,$49,$7F,$C6,$23,$2F,$B8,$93,$D4,$F9,$27,$F1,$CA
	.dc.b	$DB,$93,$C1,$87,$E5,$9D,$9F,$C8,$2B,$F6,$87,$07,$CA,$7E,$75,$CB
	.dc.b	$C2,$39,$39,$3D,$46,$67,$6E,$4F,$0B,$8B,$FD,$CD,$77,$AE,$71,$8A
	.dc.b	$E6,$70,$2B,$97,$07,$57,$07,$C8,$7E,$11,$C2,$FE,$53,$E5,$30,$7C
	.dc.b	$53,$FA,$3D,$7F,$4E,$23,$FD,$39,$AF,$F4,$E6,$BC,$0D,$10,$D1,$67
	.dc.b	$CB,$3E,$28,$FE,$C7,$F5,$CF,$81,$F8,$93,$9A,$DD,$7F,$72,$AD,$3E
	.dc.b	$83,$EC,$15,$CB,$F0,$E7,$91,$1E,$55,$D9,$FC,$25,$76,$55,$CA,$31
	.dc.b	$87,$F8,$3E,$7A,$1C,$8F,$13,$98,$E5,$3C,$6F,$C1,$5E,$63,$89,$DA
	.dc.b	$FC,$4E,$6B,$8C,$62,$FA,$57,$2B,$E9,$7E,$88,$EB,$3A,$8F,$21,$D9
	.dc.b	$C3,$D6,$3F,$C3,$6F,$84,$F4,$E1,$E2,$AE,$B1,$B3,$B1,$0E,$07,$58
	.dc.b	$E1,$7E,$76,$C7,$43,$EA,$46,$5A,$7F,$8B,$30,$FF,$10,$57,$83,$0F
	.dc.b	$39,$E8,$7C,$66,$6F,$AD,$7E,$22,$FA,$34,$E9,$87,$BA,$B8,$47,$80
	.dc.b	$D1,$E9,$36,$7C,$8C,$3C,$75,$D1,$B1,$A3,$E0,$BF,$84,$BF,$96,$EA
	.dc.b	$3A,$B9,$78,$57,$AA,$BA,$5F,$95,$E2,$D9,$F9,$07,$27,$0F,$9C,$C1
	.dc.b	$E7,$7F,$03,$7C,$A3,$04,$3E,$D1,$2B,$C4,$F6,$BF,$41,$2B,$9C,$7A
	.dc.b	$2F,$A3,$F8,$03,$F1,$25,$3C,$A6,$2B,$F6,$2D,$8D,$BA,$21,$F9,$63
	.dc.b	$46,$8E,$86,$4D,$9F,$42,$7E,$1D,$79,$E3,$FC,$EB,$F5,$EE,$4F,$3E
	.dc.b	$1E,$E9,$F8,$E2,$74,$FB,$48,$69,$CB,$FC,$53,$BB,$E0,$C3,$9E,$9D
	.dc.b	$A2,$4F,$E1,$67,$4D,$9C,$BC,$CC,$10,$C1,$0F,$95,$3F,$0D,$FE,$26
	.dc.b	$DA,$34,$91,$F2,$63,$2D,$8C,$BF,$4A,$32,$D8,$CB,$E0,$3F,$2E,$FA
	.dc.b	$30,$FF,$4A,$7D,$83,$C8,$C6,$7F,$AD,$AB,$7F,$EA,$6A,$D3,$C8,$87
	.dc.b	$BA,$F9,$DF,$08,$F6,$9A,$8C,$B6,$FB,$34,$ED,$3C,$70,$F5,$D7,$F0
	.dc.b	$AC,$3F,$2F,$87,$8A,$FC,$18,$72,$69,$C8,$E8,$78,$8C,$88,$E5,$5D
	.dc.b	$63,$77,$E0,$AF,$E5,$75,$A7,$B5,$6E,$B9,$CE,$5F,$EB,$57,$E9,$3C
	.dc.b	$F8,$58,$F1,$CE,$DF,$25,$73,$25,$F8,$2B,$84,$F6,$6C,$EA,$3D,$C7
	.dc.b	$FB,$42,$BA,$61,$E9,$8F,$19,$FC,$76,$3C,$06,$B0,$FA,$CE,$2F,$AE
	.dc.b	$1E,$1A,$E1,$18,$7B,$E1,$F6,$9F,$97,$87,$77,$E6,$4F,$C0,$7E,$5E
	.dc.b	$15,$F3,$47,$37,$67,$63,$65,$36,$60,$A4,$AF,$84,$FC,$AB,$F4,$39
	.dc.b	$7E,$B9,$1F,$4B,$A8,$E3,$7E,$9C,$3E,$15,$77,$9F,$71,$2B,$94,$77
	.dc.b	$39,$90,$C1,$27,$9B,$F1,$B0,$F8,$55,$CE,$3E,$A9,$F7,$4F,$80,$F9
	.dc.b	$23,$75,$EC,$9F,$E6,$B5,$FA,$16,$E1,$FC,$2E,$31,$1B,$21,$F8,$63
	.dc.b	$67,$F3,$37,$6E,$4D,$9F,$50,$FE,$7A,$FA,$AF,$D1,$0F,$B1,$78,$FE
	.dc.b	$68,$F8,$55,$F9,$D9,$CC,$7B,$30,$FA,$04,$47,$3A,$E4,$F8,$6F,$BD
	.dc.b	$72,$8F,$B8,$76,$C3,$FA,$CC,$66,$7C,$B8,$79,$F0,$FF,$11,$CE,$5F
	.dc.b	$05,$F5,$6C,$66,$3A,$1E,$73,$CA,$E5,$E0,$68,$52,$7B,$E1,$EA,$BF
	.dc.b	$15,$72,$8C,$61,$EE,$C3,$A5,$6A,$BA,$9F,$0C,$F1,$18,$7E,$65,$72
	.dc.b	$8C,$CF,$2A,$C4,$79,$2F,$C6,$F2,$BE,$D5,$B8,$C5,$7C,$AB,$E3,$87
	.dc.b	$D0,$6E,$1F,$19,$B5,$F9,$FB,$E4,$FD,$43,$D2,$E2,$BE,$9D,$F5,$9E
	.dc.b	$E7,$A2,$B9,$BC,$1B,$87,$BA,$36,$75,$BE,$98,$7D,$A9,$D4,$F8,$5F
	.dc.b	$6D,$72,$BF,$04,$E5,$FA,$11,$98,$FF,$20,$5F,$47,$EE,$61,$F6,$DF
	.dc.b	$B0,$78,$9F,$88,$DC,$3E,$23,$4F,$89,$5D,$5F,$A4,$72,$29,$F7,$4F
	.dc.b	$21,$E2,$7C,$F3,$F2,$23,$04,$36,$72,$AF,$C4,$E7,$FC,$04,$91,$FB
	.dc.b	$D6,$BF,$BA,$6B,$FC,$29,$AF,$F5,$E6,$BF,$DF,$DF,$E1,$9F,$01,$23
	.dc.b	$F8,$43,$5F,$E0,$AD,$7F,$8A,$B5,$FE,$D2,$D7,$FC,$14,$D7,$FB,$96
	.dc.b	$DF,$C6,$48,$FF,$15,$6B,$FC,$49,$AF,$F2,$26,$BF,$DD,$9A,$FF,$85
	.dc.b	$DF,$E4,$9D,$72,$3F,$90,$35,$FE,$3A,$D7,$F9,$6B,$5F,$F0,$0B,$5F
	.dc.b	$F1,$13,$5F,$F0,$57,$A5,$FC,$C4,$8F,$F2,$D6,$BF,$CA,$9A,$FF,$3A
	.dc.b	$6B,$FE,$0E,$6B,$FE,$2F,$7F,$C8,$DE,$9F,$E4,$E3,$FC,$DE,$D7,$FC
	.dc.b	$EE,$D7,$FD,$82,$D7,$F9,$6B,$53,$49,$A7,$62,$47,$F9,$DE,$32,$DF
	.dc.b	$73,$F9,$7F,$8D,$23,$FD,$51,$29,$FF,$45,$B6,$3F,$5B,$3B,$7F,$43
	.dc.b	$19,$6C,$E7,$5F,$76,$AB,$E7,$32,$B9,$90,$E2,$78,$9E,$58,$7F,$40
	.dc.b	$9E,$13,$A3,$F5,$0E,$9F,$8B,$3A,$BF,$0C,$E6,$7E,$3C,$E6,$BA,$47
	.dc.b	$86,$BE,$A6,$1E,$B6,$D7,$1C,$3D,$55,$E5,$21,$E2,$39,$5F,$D8,$9E
	.dc.b	$D1,$FD,$43,$0F,$16,$1D,$6B,$53,$C0,$D1,$D3,$0F,$0B,$93,$FB,$B5
	.dc.b	$74,$7D,$66,$70,$F2,$BF,$16,$78,$BE,$52,$38,$25,$7C,$28,$CB,$E4
	.dc.b	$79,$39,$8F,$97,$1E,$58,$D4,$62,$38,$9F,$48,$D1,$EC,$32,$7A,$0E
	.dc.b	$A6,$5E,$47,$91,$E4,$70,$9F,$6C,$66,$39,$3C,$27,$CC,$79,$09,$1D
	.dc.b	$5C,$BD,$23,$94,$F4,$9C,$CF,$47,$8B,$F4,$0F,$A5,$7F,$65,$E5,$1C
	.dc.b	$27,$94,$60,$87,$13,$E0,$CF,$17,$B1,$E5,$21,$82,$47,$47,$95,$7C
	.dc.b	$19,$E5,$39,$34,$F5,$33,$7F,$90,$C3,$F1,$13,$B8,$FC,$54,$EA,$FD
	.dc.b	$B5,$A9,$E1,$1F,$68,$F0,$E1,$F1,$AB,$B5,$7E,$96,$B9,$B6,$B8,$4F
	.dc.b	$57,$E0,$C7,$57,$B9,$FA,$A3,$F9,$0B,$FC,$98,$8F,$F1,$72,$1F,$CA
	.dc.b	$1C,$B4,$C9,$C8,$A7,$DA,$8F,$3B,$FE,$06,$F0,$3C,$EF,$B7,$04,$62
	.dc.b	$3F,$2A,$7F,$99,$5C,$63,$F8,$86,$E3,$F8,$6A,$E3,$5F,$03,$0F,$9F
	.dc.b	$5A,$8F,$C1,$CE,$9E,$B8,$79,$1F,$F0,$14,$ED,$FC,$0B,$63,$EF,$B6
	.dc.b	$BF,$CC,$ED,$9F,$61,$E1,$3D,$8E,$5E,$A6,$4F,$6D,$78,$F0,$FC,$65
	.dc.b	$7A,$27,$51,$EF,$6E,$1E,$FB,$E9,$3F,$CF,$9B,$1F,$E7,$66,$CF,$F9
	.dc.b	$C1,$AF,$CB,$39,$9F,$11,$CD,$F8,$E7,$EC,$5F,$8E,$BE,$3B,$6F,$CF
	.dc.b	$1E,$E3,$99,$F2,$0B,$87,$C8,$6C,$FC,$86,$9F,$20,$EA,$73,$29,$F1
	.dc.b	$B0,$F2,$91,$E2,$7C,$C3,$75,$F1,$DB,$7F,$1D,$DC,$7C,$77,$FD,$4E
	.dc.b	$FF,$80,$FE,$CB,$FD,$29,$E2,$BD,$BF,$F5,$5D,$7F,$93,$A3,$0F,$0B
	.dc.b	$CD,$7F,$A6,$CE,$35,$C2,$FF,$C1,$D7,$C2,$FF,$CA,$F5,$BE,$2D,$61
	.dc.b	$DA,$B5,$7C,$30,$CE,$1F,$EB,$CC,$38,$3E,$13,$FD,$7F,$5C,$2F,$95
	.dc.b	$70,$95,$70,$7D,$51,$9A,$E1,$1D,$70,$E1,$5E,$A8,$EC,$27,$6F,$49
	.dc.b	$D9,$CE,$A4,$70,$AF,$06,$1E,$B8,$E7,$7C,$A7,$38,$7A,$27,$37,$D1
	.dc.b	$E5,$87,$47,$D2,$60,$86,$09,$87,$DB,$8E,$51,$8A,$EC,$FD,$28,$CC
	.dc.b	$F9,$63,$47,$C3,$3A,$B9,$24,$FE,$02,$78,$BD,$23,$33,$C4,$EC,$7B
	.dc.b	$08,$60,$8F,$07,$E7,$57,$47,$94,$67,$0F,$91,$79,$9D,$9D,$27,$E0
	.dc.b	$3E,$8B,$F2,$11,$E6,$FA,$5E,$0E,$5F,$B8,$71,$3F,$1B,$7F,$46,$7A
	.dc.b	$C7,$A2,$78,$5F,$92,$38,$1E,$43,$E5,$9D,$0E,$07,$43,$C2,$79,$08
	.dc.b	$71,$9F,$89,$87,$E1,$A3,$81,$D1,$C3,$BA,$F5,$46,$2B,$F2,$E7,$96
	.dc.b	$7C,$26,$CE,$75,$EC,$25,$7B,$2F,$E7,$5F,$E8,$EB,$C1,$87,$D1,$C3
	.dc.b	$AD,$6A,$3E,$35,$F8,$9D,$46,$2B,$DB,$7C,$F0,$F6,$D6,$A7,$E1,$9F
	.dc.b	$1E,$3C,$6F,$9A,$FA,$3F,$3B,$0F,$0B,$F4,$2D,$7F,$C6,$22,$46,$1C
	.dc.b	$1F,$95,$33,$18,$70,$7E,$14,$DC,$71,$6C,$6D,$FB,$87,$23,$94,$6D
	.dc.b	$C1,$C0,$F3,$47,$F1,$1F,$0B,$E4,$47,$AD,$AF,$D6,$A4,$63,$0F,$87
	.dc.b	$3F,$C4,$E3,$18,$79,$E7,$F9,$7C,$62,$7F,$17,$5C,$9F,$09,$DC,$A7
	.dc.b	$78,$E0,$FD,$1A,$FE,$69,$18,$C3,$FA,$2B,$70,$F5,$57,$2C,$3C,$B8
	.dc.b	$77,$8F,$C7,$DF,$E4,$EF,$84,$7F,$22,$6B,$FD,$35,$B5,$EA,$8E,$98
	.dc.b	$7B,$4F,$1B,$83,$DB,$3C,$23,$31,$E0,$C3,$F2,$D3,$98,$FE,$8E,$D7
	.dc.b	$E5,$1C,$CF,$80,$68,$C5,$FC,$06,$CF,$C0,$7B,$1F,$2C,$A7,$CB,$36
	.dc.b	$7B,$CA,$7B,$CB,$7E,$F7,$8B,$F2,$AB,$CD,$87,$CB,$8F,$C0,$DF,$CB
	.dc.b	$AE,$38,$7C,$09,$CE,$1F,$02,$72,$7A,$CE,$B3,$F2,$DC,$4F,$07,$D4
	.dc.b	$D7,$D4,$14,$7F,$81,$3E,$CD,$EA,$7E,$8D,$B5,$FA,$B6,$BF,$B0,$6B
	.dc.b	$FB,$56,$BC,$E3,$91,$1E,$4D,$79,$35,$E5,$E4,$78,$48,$F8,$5A,$F8
	.dc.b	$5A,$F0,$34,$43,$45,$3C,$DE,$AF,$B4,$8F,$B5,$AF,$B5,$AF,$03,$44
	.dc.b	$34,$53,$E1,$7C,$77,$C9,$23,$F2,$5A,$FC,$96,$BC,$0D,$10,$D1,$4F
	.dc.b	$9D,$D1,$F5,$48,$FD,$56,$BF,$55,$AF,$03,$44,$34,$53,$ED,$74,$7E
	.dc.b	$04,$8F,$E0,$5A,$FE,$05,$AF,$03,$44,$34,$53,$F0,$DD,$1F,$8D,$23
	.dc.b	$F8,$D6,$BF,$8D,$6B,$C0,$D1,$0D,$14,$FC,$97,$47,$E6,$48,$FE,$65
	.dc.b	$AF,$E6,$5A,$F0,$34,$43,$45,$3F,$3D,$D1,$FA,$52,$3F,$A5,$6B,$FA
	.dc.b	$56,$BC,$0D,$10,$D1,$4F,$EB,$DF,$69,$FB,$47,$FA,$D3,$5F,$EB,$ED
	.dc.b	$77,$B7,$F5,$98,$C4,$7F,$9F,$9A,$83,$86,$C7,$12,$3E,$06,$BE,$0E
	.dc.b	$3E,$C4,$7C,$AD,$74,$60,$86,$08,$7B,$9E,$0F,$C0,$F2,$3D,$44,$7E
	.dc.b	$0B,$5F,$82,$D7,$81,$A2,$1A,$29,$F3,$3C,$8F,$8C,$47,$E6,$B5,$F9
	.dc.b	$AD,$78,$1A,$21,$A2,$9F,$63,$C8,$FA,$44,$7E,$CB,$5F,$B2,$D7,$81
	.dc.b	$A2,$1A,$29,$F8,$4F,$6F,$EF,$11,$FC,$2B,$5F,$C2,$B5,$E0,$68,$86
	.dc.b	$8A,$7E,$43,$C7,$FC,$51,$1F,$C8,$B5,$FC,$8B,$5E,$06,$88,$68,$A7
	.dc.b	$E7,$3D,$CF,$CB,$11,$FC,$EB,$5F,$CE,$B5,$E0,$68,$86,$8A,$7E,$A3
	.dc.b	$E0,$FF,$44,$47,$F5,$2D,$7F,$52,$D7,$81,$A2,$1A,$29,$FB,$4F,$7B
	.dc.b	$F5,$C4,$7F,$6A,$D7,$F6,$AD,$78,$1A,$21,$A2,$9F,$E5,$DF,$B5,$70
	.dc.b	$FF,$B7,$DB,$1F,$ED,$86,$BF,$DE,$76,$FF,$7D,$C3,$87,$85,$4A,$E0
	.dc.b	$F3,$8C,$47,$13,$F8,$24,$F4,$E9,$FE,$1C,$DB,$EB,$5C,$27,$37,$67
	.dc.b	$37,$D8,$87,$51,$7C,$DE,$F7,$C5,$F0,$D5,$77,$1F,$EF,$FD,$1E,$87
	.dc.b	$C3,$E5,$95,$FE,$FA,$9C,$B6,$72,$F8,$70,$57,$39,$93,$F2,$E3,$79
	.dc.b	$7F,$2D,$BE,$73,$E9,$6C,$7C,$49,$DE,$5F,$CA,$E9,$5A,$8C,$36,$3E
	.dc.b	$6C,$F0,$7F,$99,$46,$5E,$37,$97,$E3,$CF,$99,$F1,$46,$70,$F5,$C7
	.dc.b	$BC,$87,$BE,$79,$57,$53,$B4,$F0,$9D,$1C,$30,$F8,$17,$D6,$FC,$11
	.dc.b	$89,$F8,$C4,$3C,$67,$1C,$3F,$23,$3E,$0A,$E3,$18,$9E,$15,$CD,$E2
	.dc.b	$78,$88,$E4,$91,$C6,$32,$FC,$C9,$EB,$1F,$30,$EB,$19,$7C,$86,$4E
	.dc.b	$27,$D0,$8E,$2F,$CE,$38,$99,$38,$99,$38,$9F,$58,$CD,$FC,$BA,$FC
	.dc.b	$24,$70,$C3,$E7,$E1,$F0,$E3,$6D,$7E,$0C,$F0,$7A,$CF,$D7,$BF,$A5
	.dc.b	$1C,$DA,$F0,$C3,$DC,$F9,$30,$F1,$46,$2F,$E4,$CF,$C1,$7B,$18,$7D
	.dc.b	$47,$1C,$3E,$1E,$1E,$88,$F8,$66,$A7,$85,$F6,$8E,$97,$DC,$8E,$08
	.dc.b	$71,$72,$FC,$B9,$F1,$C6,$2B,$C1,$3F,$0B,$0F,$AB,$7D,$E3,$85,$6E
	.dc.b	$7E,$1D,$FA,$EF,$8B,$D2,$BE,$D1,$F1,$CF,$1D,$F3,$9F,$9D,$1E,$E8
	.dc.b	$C5,$78,$30,$FA,$D7,$F9,$7A,$FD,$F1,$F2,$0B,$7F,$C1,$9B,$5A,$7E
	.dc.b	$A9,$83,$07,$13,$24,$3C,$E6,$CE,$51,$D5,$F8,$07,$E4,$39,$FE,$BC
	.dc.b	$FC,$66,$C6,$4F,$C4,$14,$F8,$47,$D3,$3E,$2D,$7F,$3E,$8E,$31,$AC
	.dc.b	$11,$89,$FC,$0B,$71,$FB,$73,$AA,$EE,$71,$9D,$3D,$A7,$FA,$AC,$62
	.dc.b	$3C,$47,$F4,$47,$18,$7F,$16,$6E,$1F,$D8,$5B,$3F,$D8,$5A,$FF,$3D
	.dc.b	$6E,$1E,$B8,$F1,$BE,$86,$DF,$A2,$76,$F9,$23,$27,$9C,$FE,$C7,$1A
	.dc.b	$69,$A3,$D3,$87,$C1,$7E,$6E,$1F,$8C,$7C,$86,$DF,$63,$70,$F6,$34
	.dc.b	$F6,$14,$E2,$74,$7E,$F3,$9C,$3D,$CE,$4F,$BC,$FE,$8E,$37,$19,$3E
	.dc.b	$A1,$C7,$0F,$80,$E6,$7E,$33,$63,$E3,$3B,$3E,$09,$93,$E3,$1F,$64
	.dc.b	$D1,$4D,$1F,$1B,$0F,$EC,$D3,$FE,$99,$FB,$67,$D4,$FF,$87,$7A,$E5
	.dc.b	$FF,$A5,$6B,$53,$C3,$0F,$ED,$D7,$C2,$BF,$C4,$57,$C1,$FF,$07,$46
	.dc.b	$5B,$59,$BF,$F0,$DD,$73,$9F,$F5,$15,$6E,$7B,$C7,$42,$1C,$8F,$25
	.dc.b	$F8,$AF,$A3,$FE,$30,$BE,$92,$BE,$93,$B8,$E7,$7E,$27,$99,$E2,$BE
	.dc.b	$35,$C2,$3F,$C6,$77,$D9,$D1,$C0,$ED,$39,$7F,$C5,$B1,$CA,$24,$78
	.dc.b	$2F,$CC,$F1,$AE,$CF,$23,$47,$80,$D4,$F2,$7F,$C5,$55,$DE,$78,$9C
	.dc.b	$B3,$46,$6B,$86,$1E,$FB,$E2,$74,$21,$82,$3F,$49,$FA,$06,$30,$F4
	.dc.b	$B8,$8D,$1D,$49,$1F,$3E,$34,$F0,$8C,$C7,$CA,$32,$7B,$CE,$C6,$5E
	.dc.b	$B1,$C4,$91,$97,$B4,$74,$8F,$A2,$7B,$1E,$2F,$57,$58,$7C,$8A,$FB
	.dc.b	$18,$7C,$8A,$E1,$7D,$0E,$EF,$43,$CE,$7D,$73,$CE,$66,$FB,$B9,$7C
	.dc.b	$93,$D6,$38,$DF,$D7,$C3,$EB,$47,$80,$F3,$5F,$82,$7C,$D1,$F8,$43
	.dc.b	$B4,$79,$5F,$84,$70,$21,$E1,$74,$F0,$33,$1D,$1F,$4D,$79,$6F,$D3
	.dc.b	$3E,$59,$F0,$3B,$8F,$5D,$FE,$32,$FA,$DF,$17,$D1,$86,$70,$F1,$CE
	.dc.b	$5F,$39,$E5,$33,$5B,$70,$79,$C9,$87,$86,$FE,$9D,$73,$9D,$CF,$9D
	.dc.b	$FA,$6F,$B5,$FF,$03,$BF,$C4,$DB,$7F,$E0,$27,$ED,$18,$3F,$C0,$B1
	.dc.b	$B7,$F3,$46,$4F,$F0,$63,$97,$F3,$46,$48,$7D,$07,$D0,$FF,$AF,$3D
	.dc.b	$4F,$74,$62,$3F,$D6,$71,$89,$F8,$54,$8C,$4F,$E1,$4D,$96,$B9,$4E
	.dc.b	$E3,$15,$CA,$B8,$C7,$E6,$27,$4F,$E4,$A7,$4F,$03,$B4,$6B,$1F,$C1
	.dc.b	$CE,$A7,$81,$AA,$EB,$3F,$A0,$AE,$18,$7C,$1B,$E1,$3C,$AF,$81,$DC
	.dc.b	$F2,$5F,$2A,$FC,$5B,$63,$F0,$F7,$D1,$FC,$4E,$1F,$02,$75,$1F,$88
	.dc.b	$9D,$3F,$84,$C3,$C3,$7E,$38,$F0,$9E,$A3,$C2,$78,$F0,$F5,$5F,$4C
	.dc.b	$3F,$96,$36,$FF,$A6,$E1,$DA,$7E,$D1,$AB,$E5,$87,$A9,$A7,$AA,$F9
	.dc.b	$3F,$D2,$6B,$95,$FA,$9B,$1C,$0F,$BE,$79,$F0,$F2,$B9,$C3,$C0,$E7
	.dc.b	$0F,$03,$4F,$01,$BB,$F4,$E1,$E2,$8F,$F7,$A3,$6D,$5F,$EE,$BC,$3C
	.dc.b	$75,$CC,$F1,$CE,$90,$FB,$77,$3F,$EB,$5B,$1F,$E7,$06,$BF,$E7,$56
	.dc.b	$BF,$E7,$C6,$BF,$EF,$B6,$BD,$9A,$F7,$8E,$C4,$7B,$35,$ED,$CF,$E3
	.dc.b	$23,$E3,$6B,$A3,$04,$30,$43,$CC,$F0,$7D,$1C,$FE,$F2,$3E,$F6,$BE
	.dc.b	$F6,$BC,$0D,$10,$D1,$4F,$89,$C1,$F2,$C8,$FC,$B6,$BF,$2D,$AF,$03
	.dc.b	$44,$34,$53,$E8,$73,$FD,$72,$3F,$5D,$AF,$D7,$6B,$C0,$D1,$0D,$14
	.dc.b	$FB,$9C,$FF,$83,$23,$F8,$36,$BF,$83,$6B,$C0,$D1,$0D,$14,$FC,$47
	.dc.b	$3F,$E3,$C8,$FE,$3D,$AF,$E3,$DA,$F0,$34,$43,$45,$3F,$29,$CF,$F9
	.dc.b	$B2,$3F,$9B,$6B,$F9,$B6,$BC,$0D,$10,$D1,$4F,$D0,$7C,$7F,$E9,$C8
	.dc.b	$FE,$9D,$AF,$E9,$DA,$F0,$34,$43,$45,$3F,$59,$E2,$FF,$18,$F8,$09
	.dc.b	$1F,$E5,$9D,$93,$FE,$68,$6C,$7F,$15,$9D,$AD,$FF,$80,$3F,$BF,$AE
	.dc.b	$06,$A3,$3C,$0A,$CD,$77,$AD,$70,$FF,$0D,$9C,$CF,$58,$FE,$01,$3A
	.dc.b	$8F,$E1,$B5,$CD,$E4,$43,$81,$AA,$E0,$F9,$4D,$5F,$92,$33,$87,$BA
	.dc.b	$78,$9B,$2D,$77,$6C,$F9,$AB,$E0,$D7,$17,$A4,$FC,$73,$BE,$1E,$A7
	.dc.b	$F9,$06,$1F,$C8,$C8,$E0,$87,$F2,$37,$2D,$32,$7D,$A2,$9D,$4F,$4D
	.dc.b	$78,$E7,$15,$F4,$BD,$FF,$44,$60,$95,$F3,$23,$2F,$F2,$DB,$AB,$14
	.dc.b	$E3,$0E,$D3,$B3,$FA,$D4,$66,$B8,$D7,$09,$F8,$66,$CF,$3C,$7F,$6A
	.dc.b	$9F,$C8,$7C,$3F,$F6,$CF,$76,$57,$E5,$63,$2F,$F6,$78,$CB,$F0,$23
	.dc.b	$2F,$D8,$8C,$E5,$F5,$A7,$2F,$E5,$23,$39,$7F,$4D,$9C,$BF,$D7,$E3
	.dc.b	$35,$FD,$3E,$72,$FE,$4A,$32,$FF,$5F,$8F,$0B,$FD,$65,$FE,$C0,$47
	.dc.b	$FB,$03,$5F,$EC,$0D,$78,$1A,$21,$A2,$D7,$AD,$F8,$87,$F8,$27,$CD
	.dc.b	$F2,$BE,$33,$27,$E6,$9C,$E6,$AD,$E6,$9F,$F7,$3F,$DC,$3C,$7F,$F0
	.dc.b	$EE,$10,$FF,$6F,$46,$5B,$19,$7F,$4E,$72,$63,$FE,$1D,$35,$8F,$F0
	.dc.b	$89,$E5,$1F,$B0,$6C,$7F,$08,$8D,$E7,$FE,$0C,$53,$A8,$C3,$63,$F8
	.dc.b	$29,$C0,$FF,$09,$B9,$78,$D6,$5F,$CE,$1E,$63,$F8,$0F,$DC,$BD,$BF
	.dc.b	$F4,$23,$63,$C4,$78,$DC,$38,$0E,$1D,$B5,$E1,$1D,$08,$F1,$38,$B8
	.dc.b	$7A,$47,$17,$FC,$5A,$D7,$BF,$C8,$F9,$88,$F3,$84,$62,$39,$B5,$F2
	.dc.b	$B5,$F6,$35,$EE,$FC,$2F,$02,$47,$56,$BE,$66,$BF,$01,$AF,$C7,$6B
	.dc.b	$EA,$6B,$F2,$3C,$FF,$9E,$47,$D4,$D7,$E1,$B5,$F9,$AD,$7E,$A3,$5F
	.dc.b	$8E,$FD,$AF,$36,$47,$C5,$6B,$F3,$9A,$FD,$86,$BF,$7D,$AF,$D2,$6B
	.dc.b	$F8,$07,$F2,$DA,$7F,$64,$C1,$18,$D3,$95,$7E,$AB,$DD,$FE,$D4,$BE
	.dc.b	$E9,$F1,$FA,$8F,$C9,$DF,$17,$C0,$71,$C7,$FC,$DF,$5C,$23,$27,$03
	.dc.b	$FC,$E7,$1E,$39,$FD,$ED,$F8,$2B,$85,$7A,$4D,$E1,$E1,$BE,$11,$DC
	.dc.b	$F3,$61,$99,$91,$E0,$9F,$DE,$5F,$5A,$CC,$F5,$3C,$87,$EF,$E3,$4F
	.dc.b	$ED,$9A,$EE,$FC,$B8,$FC,$A8,$F7,$D7,$F0,$0A,$DE,$E7,$8D,$6E,$BC
	.dc.b	$E7,$C4,$BF,$7D,$6F,$0E,$73,$E1,$6B,$E1,$C3,$E9,$57,$0C,$3B,$9D
	.dc.b	$B2,$FE,$29,$1E,$66,$CF,$96,$F9,$E1,$DF,$0F,$19,$D4,$F2,$4E,$DC
	.dc.b	$C6,$EF,$A4,$79,$31,$F6,$47,$8E,$FA,$E3,$E3,$AD,$63,$E2,$7D,$87
	.dc.b	$48,$D3,$E2,$BE,$78,$F8,$70,$E9,$39,$BE,$4F,$9C,$FE,$63,$7F,$0B
	.dc.b	$0E,$F1,$0E,$CF,$90,$C5,$7A,$A3,$24,$70,$4B,$F1,$12,$3E,$29,$0F
	.dc.b	$0C,$F5,$7C,$C7,$68,$E0,$E8,$C9,$27,$DA,$48,$FB,$A4,$8F,$8F,$19
	.dc.b	$7C,$18,$7C,$67,$33,$DA,$31,$1F,$17,$1F,$B1,$39,$C7,$E4,$D6,$A7
	.dc.b	$35,$C8,$EB,$3F,$3A,$33,$1E,$9B,$F8,$8F,$C1,$8C,$9A,$AF,$09,$D7
	.dc.b	$1F,$A9,$3E,$FC,$3D,$8F,$37,$C8,$FC,$73,$C4,$64,$F1,$9F,$36,$FE
	.dc.b	$13,$F8,$1A,$ED,$8E,$6B,$AC,$62,$7E,$E4,$63,$1F,$66,$1E,$4C,$3D
	.dc.b	$75,$F1,$E3,$15,$B8,$E8,$79,$2F,$DB,$7C,$B0,$CC,$F8,$E3,$0F,$80
	.dc.b	$F4,$94,$F6,$1B,$38,$E3,$F2,$63,$EA,$9C,$B1,$F7,$DF,$82,$B7,$8F
	.dc.b	$F3,$6F,$9D,$FC,$D1,$32,$FD,$C3,$FA,$83,$13,$FC,$C9,$C4,$68,$FE
	.dc.b	$78,$EA,$3F,$2C,$F8,$30,$F1,$C7,$83,$2F,$AC,$FC,$77,$E0,$E1,$E5
	.dc.b	$8F,$76,$5E,$B7,$6F,$E1,$B1,$FD,$FE,$5F,$42,$31,$7D,$2F,$EC,$4F
	.dc.b	$86,$F8,$DF,$D4,$7E,$5C,$7E,$0A,$FB,$47,$E3,$4F,$C4,$B8,$8E,$D3
	.dc.b	$C6,$7E,$81,$E6,$7C,$79,$7E,$52,$BC,$73,$CC,$F7,$1F,$66,$FD,$D7
	.dc.b	$E1,$BE,$33,$FA,$37,$B4,$EA,$FF,$80,$DF,$8D,$ED,$3A,$C7,$E6,$DF
	.dc.b	$8F,$1F,$85,$1F,$81,$AF,$D2,$CF,$8E,$FC,$B8,$FE,$63,$1F,$8F,$5A
	.dc.b	$8F,$7E,$1F,$53,$0E,$B1,$E7,$8E,$B8,$7D,$B3,$DD,$7C,$9F,$CF,$57
	.dc.b	$0B,$F4,$E3,$EE,$9F,$87,$87,$D5,$7C,$91,$EB,$7E,$BE,$1E,$16,$9F
	.dc.b	$26,$FA,$3F,$5A,$7A,$61,$E1,$8F,$D2,$BF,$25,$D5,$7A,$1F,$8C,$7A
	.dc.b	$2B,$E3,$3E,$83,$BD,$F9,$4F,$86,$F8,$4C,$E1,$F1,$CF,$A7,$5E,$78
	.dc.b	$F5,$B6,$78,$9C,$E7,$73,$CC,$D9,$63,$71,$D6,$3F,$09,$3C,$E3,$8C
	.dc.b	$F3,$9C,$9C,$63,$27,$17,$E7,$B6,$FE,$36,$3F,$8B,$9E,$35,$CD,$D9
	.dc.b	$F0,$CF,$2B,$99,$E3,$8F,$D0,$C7,$E4,$D7,$D0,$7E,$AB,$FC,$5A,$FB
	.dc.b	$3D,$0D,$1F,$28,$ED,$7F,$01,$B8,$7A,$9A,$79,$9F,$55,$78,$2B,$AE
	.dc.b	$1E,$4C,$7E,$F6,$1F,$97,$C3,$EF,$1F,$97,$7F,$41,$7C,$71,$F5,$4E
	.dc.b	$71,$FE,$70,$FE,$1A,$B6,$E4,$FE,$13,$1F,$9C,$BE,$ED,$8E,$E7,$17
	.dc.b	$9C,$F8,$CE,$71,$93,$99,$F9,$63,$8B,$B8,$E2,$75,$78,$99,$8E,$27
	.dc.b	$CF,$AF,$5B,$D4,$FE,$2F,$1B,$6C,$F3,$3A,$BE,$18,$EB,$19,$8E,$B8
	.dc.b	$FF,$08,$8D,$C7,$7B,$F8,$2F,$18,$CC,$EE,$38,$DF,$EF,$5F,$C5,$CF
	.dc.b	$DA,$7E,$61,$E4,$39,$1C,$2B,$F4,$0F,$E0,$4F,$C1,$9F,$8B,$AF,$C0
	.dc.b	$E1,$F6,$E7,$C7,$3E,$53,$E5,$D7,$AD,$EC,$67,$0F,$7B,$E0,$33,$7E
	.dc.b	$96,$E1,$E9,$9F,$CB,$47,$E6,$8F,$59,$E4,$C7,$F2,$4F,$80,$E8,$FF
	.dc.b	$13,$7C,$17,$F1,$AB,$A3,$F9,$F8,$E9,$5D,$AF,$DB,$1F,$1A,$F9,$47
	.dc.b	$95,$B7,$CA,$7E,$31,$CA,$FA,$57,$28,$FC,$25,$70,$C3,$E6,$C7,$CB
	.dc.b	$BF,$C4,$E1,$F8,$7A,$D4,$FF,$0D,$3A,$DF,$39,$D6,$3F,$C0,$23,$C8
	.dc.b	$DB,$F2,$57,$5C,$3E,$AD,$6A,$3F,$81,$1E,$73,$AC,$F8,$B1,$FB,$18
	.dc.b	$F8,$B1,$FE,$25,$8F,$DF,$BF,$9C,$7C,$C3,$F1,$07,$9C,$F0,$1A,$C3
	.dc.b	$C0,$FD,$83,$C0,$6A,$7F,$06,$F3,$3F,$03,$8F,$C6,$BF,$BB,$3A,$AF
	.dc.b	$EF,$F9,$7F,$88,$67,$FD,$6C,$FF,$89,$AF,$FD,$2E,$E7,$04,$48,$D3
	.dc.b	$FE,$28,$C7,$FC,$56,$47,$8B,$FA,$77,$EC,$CF,$CD,$C7,$C8,$FC,$3C
	.dc.b	$7C,$99,$7D,$0C,$7C,$F8,$7E,$3E,$FF,$D1,$D8,$78,$8F,$F5,$59,$4F
	.dc.b	$F4,$71,$72,$FF,$5F,$36,$FF,$DF,$AD,$9E,$0F,$23,$FD,$80,$64,$CC
	.dc.b	$F2,$72,$43,$E1,$AF,$69,$F8,$A7,$49,$4E,$63,$0D,$8E,$B3,$C1,$B3
	.dc.b	$A8,$C4,$7A,$DA,$FC,$19,$E4,$E6,$78,$11,$E6,$7C,$7F,$BC,$10

;--------------------------------------------------------------------------------
;	FSADD.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsadd_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FSADD.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSADD.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsadd_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsadd.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#64,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSADD.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsadd.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#64,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSADD.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsadd_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3F,$D9,$BE,$53,$FB,$3B,$4A,$7F,$6C,$6B,$4A,$FF
	.dc.b	$70,$FA,$0F,$4C,$7F,$72,$F9,$AF,$E3,$1F,$05,$9D,$3A,$DD,$AD,$7C
	.dc.b	$67,$E8,$F0,$E3,$1F,$D5,$83,$8C,$39,$C6,$2F,$9B,$4E,$78,$75,$79
	.dc.b	$7C,$FF,$1C,$3F,$60,$F6,$8F,$53,$99,$FF,$04,$B9,$86,$3C,$F0,$F0
	.dc.b	$E3,$98,$C6,$1E,$9C,$7C,$73,$CE,$FC,$87,$AD,$C1,$C5,$CD,$F4,$9E
	.dc.b	$27,$F8,$37,$1F,$54,$FC,$FA,$E3,$3A,$8F,$79,$D8,$D3,$C5,$C1,$EA
	.dc.b	$3B,$DF,$9D,$FB,$D8,$F9,$CF,$21,$67,$FC,$55,$F9,$B3,$E0,$7F,$45
	.dc.b	$58,$6B,$FE,$30,$8D,$63,$F8,$86,$C7,$E3,$1B,$F3,$BF,$C8,$47,$EE
	.dc.b	$6F,$8C,$C5,$F5,$3F,$C7,$73,$B9,$D4,$60,$FF,$18,$96,$FF,$21,$D9
	.dc.b	$C9,$FF,$1B,$FE,$61,$FE,$1A,$9D,$43,$7F,$DF,$46,$1B,$1F,$A2,$6B
	.dc.b	$FE,$4A,$6B,$E0,$35,$5D,$F0,$F5,$F9,$1E,$19,$ED,$39,$C7,$F8,$F6
	.dc.b	$1E,$2F,$94,$FF,$19,$37,$C4,$F4,$7C,$FB,$F1,$BF,$F1,$54,$E4,$98
	.dc.b	$FE,$DA,$F8,$B6,$FE,$9B,$FE,$11,$6C,$6B,$3F,$EA,$1F,$36,$FC,$85
	.dc.b	$EF,$F9,$EB,$58,$FF,$9A,$AB,$53,$9F,$02,$4F,$CF,$1F,$BF,$7F,$99
	.dc.b	$1F,$64,$6B,$3F,$33,$E2,$24,$7F,$0C,$C1,$1F,$5E,$39,$F4,$71,$CF
	.dc.b	$FB,$34,$60,$FB,$3F,$42,$F8,$CF,$9C,$D8,$FC,$16,$9E,$DF,$82,$CD
	.dc.b	$49,$FE,$37,$5B,$CF,$F8,$FB,$5F,$F4,$3B,$5F,$F3,$7C,$FE,$2E,$FE
	.dc.b	$47,$C9,$F3,$96,$3F,$33,$74,$FE,$07,$1F,$F3,$A1,$F9,$02,$D7,$07
	.dc.b	$FD,$2E,$7C,$AF,$C1,$78,$78,$32,$3F,$98,$9C,$27,$FD,$36,$D7,$FD
	.dc.b	$03,$1F,$80,$8E,$B8,$7C,$8F,$8A,$7A,$9C,$EF,$C0,$7F,$8E,$E3,$35
	.dc.b	$FA,$4F,$0B,$FA,$E3,$7C,$CF,$81,$A3,$E5,$3F,$39,$5A,$C3,$99,$FC
	.dc.b	$9A,$BF,$23,$19,$8F,$E9,$91,$83,$F2,$BE,$07,$E0,$30,$FE,$0D,$C9
	.dc.b	$82,$3F,$C2,$7E,$0B,$64,$70,$7F,$54,$FA,$1F,$F5,$B3,$7C,$AF,$D0
	.dc.b	$C6,$31,$FE,$B6,$D7,$FD,$87,$F9,$CF,$F9,$D6,$BF,$AE,$B6,$3F,$B0
	.dc.b	$B5,$DF,$4F,$EB,$3E,$BB,$E3,$C7,$C0,$6F,$17,$F8,$EF,$E8,$DF,$AE
	.dc.b	$7F,$A8,$F1,$FE,$1F,$5A,$7F,$DA,$8F,$2B,$FF,$6A,$B5,$CF,$27,$72
	.dc.b	$4F,$F1,$DF,$05,$EA,$BD,$6F,$E1,$D8,$75,$C1,$5A,$9F,$D6,$79,$DF
	.dc.b	$0F,$4F,$D5,$BF,$E9,$1C,$7F,$74,$E5,$FC,$9B,$9B,$FC,$3C,$7F,$82
	.dc.b	$38,$39,$FC,$97,$46,$C7,$2F,$A3,$EE,$78,$3E,$02,$C6,$49,$7D,$30
	.dc.b	$FF,$22,$46,$31,$FF,$17,$7C,$47,$F1,$2A,$FF,$7D,$B7,$87,$E6,$F9
	.dc.b	$FF,$CB,$63,$53,$FE,$5D,$F2,$3F,$90,$C7,$F2,$48,$FF,$80,$2D,$47
	.dc.b	$48,$D9,$FD,$1E,$BE,$CF,$C6,$75,$8C,$72,$23,$FA,$67,$D2,$FF,$BC
	.dc.b	$2F,$F5,$6F,$4B,$CC,$7E,$F0,$C9,$FC,$94,$A6,$9D,$B7,$3F,$EB,$53
	.dc.b	$E6,$F4,$B9,$9D,$3C,$3B,$3A,$AF,$F2,$3B,$67,$4D,$3F,$AF,$FD,$2F
	.dc.b	$CE,$9F,$DB,$B7,$6F,$F4,$97,$90,$FC,$15,$F3,$3F,$D6,$6B,$58,$7F
	.dc.b	$BE,$8F,$F2,$1B,$CF,$0F,$96,$7C,$C7,$0E,$CF,$F2,$5B,$C9,$FE,$E1
	.dc.b	$F8,$07,$F9,$C6,$7C,$BC,$9F,$DD,$FD,$44,$E6,$BE,$4F,$CF,$FC,$C3
	.dc.b	$E6,$FC,$A5,$C3,$E1,$45,$9F,$9F,$37,$5F,$F5,$77,$03,$EA,$7F,$DB
	.dc.b	$58,$7C,$E7,$F6,$D7,$F6,$9C,$B7,$E7,$FE,$FC,$62,$F9,$BC,$1B,$5A
	.dc.b	$9F,$F0,$07,$D4,$3D,$2A,$7B,$7E,$2F,$FD,$65,$5F,$E1,$29,$D3,$67
	.dc.b	$FB,$CC,$63,$1F,$F3,$9B,$5F,$9E,$1C,$3F,$49,$E1,$87,$F8,$B7,$F1
	.dc.b	$47,$C2,$78,$B0,$FF,$13,$4A,$73,$1F,$0F,$E5,$FD,$FF,$2F,$F0,$1B
	.dc.b	$19,$3E,$44,$65,$A7,$CB,$F5,$FE,$3F,$CF,$FC,$86,$C6,$4F,$A1,$19
	.dc.b	$69,$FE,$34,$FA,$8F,$8B,$A7,$FA,$49,$AF,$C0,$6B,$F9,$06,$BF,$BC
	.dc.b	$78,$E5,$FE,$B7,$5F,$31,$FE,$33,$73,$5C,$63,$FC,$6D,$E7,$6A,$3F
	.dc.b	$C8,$AD,$BF,$EB,$B5,$6F,$FD,$85,$36,$BF,$42,$EE,$BF,$A6,$DF,$F8
	.dc.b	$CA,$31,$1A,$8F,$F1,$AD,$BF,$4F,$FF,$0E,$CF,$AD,$AF,$AD,$AA,$38
	.dc.b	$5F,$F0,$D8,$D3,$FE,$B5,$AF,$89,$F1,$C8,$C4,$E8,$FF,$4D,$CE,$A3
	.dc.b	$30,$FA,$47,$CD,$BC,$3C,$C6,$1D,$E2,$C7,$B5,$F4,$3F,$96,$D6,$DC
	.dc.b	$1D,$6F,$A3,$88,$D7,$B1,$FE,$DE,$8F,$75,$74,$8C,$4E,$B2,$FE,$1E
	.dc.b	$DA,$EF,$7D,$2B,$96,$1E,$5C,$BD,$33,$9C,$38,$B6,$F8,$E5,$EC,$CB
	.dc.b	$C1,$5A,$8E,$24,$9F,$5E,$1D,$5F,$55,$F3,$7D,$D3,$BC,$7C,$78,$74
	.dc.b	$9E,$E7,$09,$D6,$1F,$4E,$30,$4A,$F8,$37,$D7,$0F,$04,$62,$F8,$D7
	.dc.b	$48,$EF,$8E,$23,$C6,$73,$C3,$B4,$F1,$8C,$1C,$63,$27,$C1,$72,$72
	.dc.b	$8E,$AE,$32,$FA,$4E,$67,$8B,$CC,$86,$4F,$DB,$65,$F0,$DC,$12,$79
	.dc.b	$E3,$EE,$AF,$15,$F9,$C8,$F6,$AE,$19,$7D,$1B,$E3,$87,$57,$CA,$48
	.dc.b	$F9,$51,$97,$81,$F9,$49,$FE,$5B,$3F,$99,$6C,$65,$F8,$86,$BE,$8D
	.dc.b	$EA,$7F,$34,$8F,$D0,$C6,$1C,$52,$31,$1C,$67,$83,$FB,$18,$E1,$8F
	.dc.b	$E8,$1B,$8F,$E8,$1B,$3F,$E1,$AF,$A3,$7C,$EF,$F7,$2B,$E1,$87,$EE
	.dc.b	$2F,$77,$D6,$3F,$7F,$7C,$23,$51,$1F,$DA,$4F,$09,$D4,$62,$75,$5D
	.dc.b	$A3,$F7,$B3,$A8,$C1,$30,$F4,$57,$83,$0F,$34,$62,$FA,$12,$BF,$81
	.dc.b	$5F,$39,$D4,$73,$9F,$9B,$7C,$DF,$1C,$E8,$8F,$C5,$BE,$EF,$03,$B4
	.dc.b	$69,$5F,$77,$57,$DD,$F9,$07,$BA,$B9,$B8,$8D,$63,$F7,$70,$F2,$4F
	.dc.b	$63,$9E,$1D,$F0,$ED,$1C,$63,$C5,$5C,$63,$B1,$EC,$38,$90,$F3,$E1
	.dc.b	$E2,$8E,$18,$78,$B1,$FB,$13,$97,$F0,$33,$AB,$F1,$4E,$71,$ED,$87
	.dc.b	$8C,$F9,$06,$2B,$83,$E1,$C7,$D2,$3F,$46,$F8,$6E,$2E,$C4,$9F,$EA
	.dc.b	$C7,$CE,$7F,$CD,$0F,$E6,$5A,$E4,$DF,$E7,$AF,$2B,$F4,$B0,$8C,$3F
	.dc.b	$A8,$6B,$FB,$16,$BF,$E2,$67,$3C,$9F,$D5,$A3,$94,$FF,$96,$23,$13
	.dc.b	$AE,$1F,$F2,$DF,$D2,$BE,$A3,$51,$8F,$32,$E3,$FE,$8A,$AD,$3F,$E8
	.dc.b	$CC,$3B,$46,$23,$FD,$CA,$A7,$8C,$23,$0F,$F0,$26,$BF,$C2,$DA,$FF
	.dc.b	$90,$1A,$FF,$9C,$3E,$95,$F5,$1D,$17,$C8,$53,$11,$27,$CD,$8F,$19
	.dc.b	$D4,$60,$FF,$52,$D7,$47,$03,$F1,$CF,$F0,$03,$85,$F4,$9E,$C6,$C6
	.dc.b	$7A,$30,$66,$64,$64,$EB,$F4,$AF,$07,$97,$AD,$27,$31,$F2,$27,$51
	.dc.b	$FC,$A1,$AF,$F3,$16,$BF,$E6,$66,$BF,$E9,$67,$7B,$FE,$0D,$BA,$66
	.dc.b	$BD,$5C,$FF,$5B,$F1,$C7,$AF,$7E,$92,$65,$F1,$27,$2D,$9C,$AC,$BF
	.dc.b	$21,$2A,$E9,$5F,$82,$FC,$73,$E7,$5E,$F2,$FF,$55,$C6,$30,$FC,$C3
	.dc.b	$63,$3C,$3C,$78,$A5,$FD,$6A,$D3,$FE,$DE,$7F,$06,$FD,$AE,$FF,$AA
	.dc.b	$F7,$3F,$DC,$7E,$D3,$E9,$FF,$2A,$D8,$E1,$C5,$F9,$8A,$E3,$1F,$E9
	.dc.b	$26,$BF,$EC,$26,$BF,$EE,$E6,$BF,$E4,$C8,$E4,$47,$F1,$B3,$F8,$C7
	.dc.b	$C8,$1F,$EC,$F9,$FE,$E5,$B9,$FF,$B4,$1B,$3E,$AF,$98,$F4,$F0,$48
	.dc.b	$EA,$FB,$23,$F6,$3F,$8E,$7D,$89,$EE,$FD,$57,$F3,$ED,$73,$BF,$03
	.dc.b	$8F,$04,$8D,$C7,$EF,$3E,$FD,$F5,$BF,$BE,$FC,$71,$F5,$B4,$D4,$4B
	.dc.b	$FF,$00,$D7,$06,$C7,$F2,$08,$C4,$7F,$83,$5B,$B7,$F1,$B6,$ED,$FE
	.dc.b	$F2,$6C,$86,$1F,$84,$7F,$D8,$1D,$3F,$C4,$BC,$FF,$A1,$87,$F2,$56
	.dc.b	$E1,$FE,$FC,$7F,$0F,$19,$9E,$04,$AF,$E0,$B3,$AA,$E1,$BF,$5D,$E5
	.dc.b	$6A,$3C,$06,$1F,$BF,$49,$CC,$7F,$10,$FC,$05,$F2,$9F,$CA,$3E,$81
	.dc.b	$F0,$14,$FE,$57,$3A,$AE,$15,$FD,$22,$31,$1F,$E3,$56,$BF,$EE,$A6
	.dc.b	$BF,$1C,$3C,$34,$FE,$90,$DC,$F4,$D9,$F9,$D8,$7E,$2E,$BF,$A7,$7E
	.dc.b	$1A,$F7,$3F,$AC,$B6,$BF,$C7,$ED,$7F,$C9,$0D,$7F,$DF,$BF,$58,$FB
	.dc.b	$2F,$EC,$2D,$8F,$F2,$3B,$5F,$F2,$83,$5F,$50,$FC,$35,$F2,$3F,$DA
	.dc.b	$9B,$1F,$E5,$06,$BF,$E5,$96,$BF,$DA,$7E,$87,$FB,$8F,$96,$69,$23
	.dc.b	$C7,$97,$98,$8E,$08,$7E,$F3,$0F,$81,$9F,$F7,$36,$ED,$F9,$7F,$9C
	.dc.b	$7E,$51,$FD,$DE,$BF,$BC,$7C,$07,$99,$FE,$F5,$F2,$0F,$7B,$FB,$5B
	.dc.b	$72,$FE,$AF,$5B,$C7,$EE,$3E,$C8,$BB,$FF,$83,$23,$11,$FA,$76,$A7
	.dc.b	$93,$98,$FE,$F3,$18,$85,$3E,$0B,$FB,$EF,$D7,$FF,$00,$9F,$ED,$9A
	.dc.b	$7B,$27,$FC,$12,$D8,$FE,$DB,$3B,$7F,$90,$B5,$FB,$6D,$7F,$2C,$D7
	.dc.b	$F7,$CD,$4F,$3C,$FF,$C4,$0D,$CB,$8F,$7F,$8B,$5F,$E0,$F6,$F9,$5F
	.dc.b	$BC,$FF,$80,$3F,$0E,$E0,$FF,$0B,$6B,$BE,$2F,$F0,$DC,$F0,$8F,$F5
	.dc.b	$13,$5F,$C9,$B5,$FD,$73,$5F,$E2,$CF,$1C,$BF,$C6,$CD,$C7,$75,$FE
	.dc.b	$28,$BF,$E3,$5F,$5B,$FC,$7B,$4F,$C3,$F9,$1E,$B3,$FC,$5B,$1E,$C3
	.dc.b	$F9,$4E,$5F,$A4,$6B,$FB,$F6,$BF,$C9,$DA,$9E,$5E,$1F,$43,$FC,$83
	.dc.b	$F2,$CF,$C8,$DF,$4F,$E0,$F9,$CF,$37,$B3,$31,$FF,$14,$E2,$9D,$47
	.dc.b	$FB,$02,$30,$D8,$FF,$2F,$C6,$23,$FD,$98,$DE,$0F,$F2,$D3,$77,$F5
	.dc.b	$57,$C5,$ED,$FF,$26,$F6,$3E,$2B,$FC,$AF,$83,$F7,$0F,$F0,$2B,$7D
	.dc.b	$2D,$CC,$CF,$FC,$D8,$FF,$00,$23,$FE,$D6,$FA,$B7,$87,$FE,$1D,$8F
	.dc.b	$E7,$1E,$E7,$F2,$9F,$AD,$FD,$1C,$7F,$51,$F2,$7F,$CB,$71,$88,$D9
	.dc.b	$FD,$56,$F9,$E5,$BB,$FE,$65,$18,$97,$D6,$72,$F0,$3F,$9B,$36,$3F
	.dc.b	$AF,$3F,$E6,$89,$E0,$DC,$BC,$CE,$E7,$9D,$F9,$EF,$9B,$E0,$AD,$B7
	.dc.b	$0E,$AE,$EA,$9F,$DB,$71,$F6,$61,$DE,$B3,$2C,$7C,$77,$C6,$3C,$27
	.dc.b	$1C,$7C,$2F,$C3,$C7,$C2,$DB,$E7,$8F,$6B,$E5,$8F,$7A,$E6,$77,$BF
	.dc.b	$A9,$97,$BC,$95,$EB,$9F,$A2,$6F,$0E,$B8,$F8,$A1,$8F,$77,$C4,$76
	.dc.b	$CB,$D6,$FA,$4E,$27,$C0,$72,$F5,$3F,$A9,$94,$FE,$AA,$FD,$D6,$E3
	.dc.b	$F8,$B7,$3E,$4B,$E3,$FF,$20,$D8,$E7,$5F,$E0,$49,$C9,$FE,$C6,$87
	.dc.b	$C3,$D2,$3F,$EC,$A6,$BE,$43,$66,$48,$65,$DB,$EA,$6B,$FD,$B9,$AF
	.dc.b	$F8,$25,$AF,$F8,$C5,$AF,$63,$A1,$D8,$E8,$53,$A1,$4E,$85,$3A,$14
	.dc.b	$E8,$72,$3A,$90,$C1,$0E,$A6,$4A,$64,$EE,$64,$A6,$4F,$09,$92,$99
	.dc.b	$3C,$67,$42,$1C,$8A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18,$23,$F4
	.dc.b	$08,$FC,$76,$BE,$93,$66,$48,$65,$DB,$ED,$6B,$E8,$6B,$E5,$6B,$E4
	.dc.b	$6B,$D8,$E8,$76,$3A,$14,$E8,$53,$A1,$4E,$85,$3A,$1C,$8E,$A4,$30
	.dc.b	$43,$A9,$92,$99,$3B,$99,$29,$93,$C2,$64,$A6,$4F,$19,$D0,$87,$22
	.dc.b	$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$FD,$02,$3F,$1D,$AF,$A4
	.dc.b	$D9,$92,$19,$76,$FB,$5A,$FA,$1A,$F9,$5A,$F9,$1A,$F6,$3A,$1D,$8E
	.dc.b	$85,$3A,$14,$E8,$53,$A1,$4E,$87,$23,$A9,$0C,$10,$EA,$64,$A6,$4E
	.dc.b	$E6,$4A,$64,$F0,$99,$29,$93,$C6,$74,$21,$C8,$A7,$02,$9C,$0D,$10
	.dc.b	$D1,$4D,$18,$21,$82,$3F,$40,$8F,$C7,$6B,$E9,$36,$64,$86,$5D,$BE
	.dc.b	$D6,$BE,$86,$BE,$56,$BE,$46,$BD,$8E,$87,$63,$A1,$4E,$85,$3A,$14
	.dc.b	$E8,$53,$A1,$C8,$EA,$43,$04,$3A,$99,$29,$93,$B9,$92,$99,$3C,$26
	.dc.b	$4A,$64,$F1,$9D,$08,$72,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60
	.dc.b	$8F,$D0,$23,$F1,$DA,$FA,$4D,$99,$21,$F0,$67,$FC,$1C,$BA,$9E,$DF
	.dc.b	$F9,$5B,$0F,$F0,$A4,$62,$7F,$CB,$91,$89,$D6,$09,$E5,$18,$7F,$C3
	.dc.b	$73,$FE,$1D,$7F,$C4,$33,$FE,$1C,$7C,$9B,$B6,$E5,$0F,$93,$7D,$3B
	.dc.b	$A3,$FC,$A3,$3F,$E0,$6B,$E0,$F8,$6B,$73,$A8,$FF,$10,$CE,$9E,$95
	.dc.b	$CD,$FF,$26,$9C,$EF,$C3,$7C,$A3,$27,$28,$CC,$F1,$AE,$93,$C4,$F1
	.dc.b	$46,$67,$FC,$43,$3E,$1A,$E8,$F0,$AE,$87,$0B,$F4,$6B,$FE,$6E,$AD
	.dc.b	$DF,$23,$C3,$3A,$AF,$F3,$2C,$E9,$FF,$16,$DF,$7C,$3C,$51,$8B,$E2
	.dc.b	$FB,$30,$ED,$3A,$8F,$F1,$9D,$71,$C3,$B3,$63,$81,$E0,$9E,$0F,$C1
	.dc.b	$3E,$6C,$F0,$BE,$F5,$EF,$BE,$F3,$9A,$F8,$07,$49,$EC,$6D,$ED,$1B
	.dc.b	$7A,$9B,$AE,$A7,$C5,$79,$CF,$13,$9C,$66,$38,$C6,$63,$8C,$7C,$53
	.dc.b	$8B,$BC,$3E,$64,$7C,$48,$DB,$4D,$9C,$4E,$AF,$1C,$3E,$2C,$7D,$E3
	.dc.b	$26,$E7,$AB,$F2,$5A,$ED,$B8,$7C,$88,$F8,$07,$80,$F1,$1D,$0B,$1C
	.dc.b	$1F,$92,$70,$34,$77,$9F,$2C,$F5,$BF,$89,$3D,$4D,$61,$F4,$6F,$A4
	.dc.b	$FC,$C9,$F5,$5F,$27,$BC,$72,$7B,$9E,$B8,$D4,$71,$8F,$3C,$63,$0F
	.dc.b	$9B,$1E,$03,$D1,$5E,$08,$CC,$73,$BF,$25,$71,$8F,$41,$93,$8B,$97
	.dc.b	$C7,$87,$BE,$BE,$B5,$72,$24,$F0,$6E,$1E,$BC,$3C,$53,$DA,$F9,$3F
	.dc.b	$0A,$B7,$7D,$23,$91,$2F,$96,$1E,$3C,$39,$CF,$90,$F9,$35,$C6,$B7
	.dc.b	$7D,$27,$B1,$CE,$74,$47,$33,$C1,$D1,$88,$D3,$F4,$27,$A3,$4D,$96
	.dc.b	$79,$12,$78,$C7,$D5,$8C,$3D,$48,$71,$8D,$BF,$4E,$36,$E4,$D9,$CE
	.dc.b	$37,$87,$CC,$8D,$10,$D1,$C0,$E9,$1C,$1F,$92,$66,$39,$3C,$0E,$47
	.dc.b	$03,$47,$03,$2F,$07,$46,$0D,$1C,$C8,$68,$AF,$12,$3C,$0D,$11,$CC
	.dc.b	$FE,$8E,$38,$9B,$8C,$CF,$E8,$1B,$1F,$91,$34,$49,$E4,$F5,$31,$3C
	.dc.b	$5A,$71,$8D,$CF,$48,$FC,$94,$61,$D9,$89,$E5,$19,$23,$82,$47,$56
	.dc.b	$BC,$4C,$10,$C1,$2F,$F4,$64,$C3,$F4,$C4,$C3,$F2,$F3,$F1,$AF,$C2
	.dc.b	$F3,$AE,$13,$A8,$F8,$53,$A7,$B5,$73,$7F,$3E,$73,$BF,$3D,$F2,$8C
	.dc.b	$9C,$A3,$33,$C6,$BF,$37,$3C,$4F,$14,$66,$7F,$29,$3E,$1A,$E8,$F0
	.dc.b	$AE,$87,$0B,$F4,$61,$F4,$EB,$77,$C8,$F0,$CE,$AB,$EA,$CE,$9F,$CD
	.dc.b	$DF,$7C,$3C,$51,$8B,$E2,$FB,$30,$ED,$3A,$8F,$CF,$57,$1C,$3B,$36
	.dc.b	$38,$1E,$09,$E0,$FC,$13,$E6,$CF,$0B,$EF,$5E,$FB,$EF,$39,$AF,$80
	.dc.b	$74,$9E,$C6,$DE,$D1,$B7,$A9,$BA,$EA,$7C,$57,$9C,$F1,$39,$C6,$63
	.dc.b	$8C,$66,$38,$C7,$C5,$38,$BB,$C3,$E6,$47,$C4,$8D,$B4,$D9,$C4,$EA
	.dc.b	$F1,$C3,$E2,$C7,$DE,$32,$6E,$7A,$BF,$25,$AE,$DB,$87,$C8,$8F,$80
	.dc.b	$78,$0F,$11,$D0,$B1,$C1,$FE,$0B,$C5,$FE,$39,$3F,$C5,$85,$BF,$36
	.dc.b	$3F,$36,$BF,$C1,$13,$9A,$FF,$39,$8B,$FF,$3F,$46,$61,$F0,$EF,$89
	.dc.b	$7A,$7C,$27,$FD,$09,$39,$8C,$47,$0F,$05,$F1,$78,$AD,$4F,$1E,$37
	.dc.b	$23,$9F,$8C,$61,$B1,$99,$E0,$D7,$FD,$1E,$7F,$A4,$B8,$BF,$D2,$0E
	.dc.b	$0F,$F5,$36,$BF,$37,$8B,$F0,$91,$F3,$4F,$A9,$C5,$F6,$A7,$5C,$5F
	.dc.b	$22,$7E,$99,$F4,$4D,$F1,$7C,$BA,$E1,$1F,$9A,$37,$C5,$F5,$70,$FC
	.dc.b	$AC,$7E,$0E,$77,$18,$8F,$B7,$18,$BF,$C8,$C7,$63,$9C,$65,$F4,$1F
	.dc.b	$89,$70,$F9,$4C,$5F,$94,$8F,$26,$D7,$C2,$8E,$18,$7C,$07,$CD,$3A
	.dc.b	$BF,$85,$5B,$C3,$D9,$1E,$6A,$EE,$75,$24,$EF,$0F,$2D,$70,$9F,$4E
	.dc.b	$1D,$DF,$83,$87,$77,$C2,$72,$C3,$C2,$47,$38,$75,$9E,$11,$C8,$E0
	.dc.b	$6A,$78,$39,$9E,$0E,$8C,$46,$9F,$A1,$1A,$24,$66,$3D,$F1,$87,$B1
	.dc.b	$0E,$65,$39,$C7,$1A,$F8,$11,$83,$47,$48,$D3,$F8,$B9,$EC,$F4,$21
	.dc.b	$C4,$A7,$13,$9C,$F6,$70,$69,$C1,$1E,$D3,$E0,$9F,$CA,$47,$52,$1C
	.dc.b	$8B,$3C,$63,$71,$92,$46,$67,$F0,$CD,$8F,$C0,$9E,$03,$CE,$53,$CC
	.dc.b	$53,$B9,$4E,$A5,$7A,$3F,$9A,$30,$49,$F2,$C6,$AF,$F2,$53,$B9,$CB
	.dc.b	$BA,$F3,$B4,$F3,$C6,$89,$1A,$6C,$6A,$30,$47,$04,$3C,$0E,$5A,$66
	.dc.b	$7C,$A4,$C3,$F5,$24,$C3,$F3,$F3,$F1,$2F,$C2,$F2,$25,$71,$9D,$C7
	.dc.b	$C8,$9D,$CE,$63,$F4,$47,$3B,$F2,$46,$67,$F2,$33,$93,$CD,$7E,$37
	.dc.b	$FD,$17,$AF,$F3,$C7,$FA,$5B,$51,$FD,$1B,$7F,$E8,$6F,$F4,$98,$D3
	.dc.b	$AD,$BF,$A0,$2E,$26,$5C,$A1,$CC,$E7,$C5,$6A,$7B,$CE,$5F,$EB,$75
	.dc.b	$C6,$FA,$D6,$E5,$1E,$07,$FB,$1B,$5E,$1C,$1F,$D3,$DB,$C3,$FD,$39
	.dc.b	$E9,$CA,$E2,$C4,$7A,$E3,$0F,$F6,$43,$FB,$71,$78,$B0,$E0,$FE,$D2
	.dc.b	$E0,$FC,$C6,$FF,$D3,$B2,$FE,$E9,$3F,$32,$7D,$AF,$CE,$E1,$53,$8B
	.dc.b	$33,$FA,$CE,$2F,$C0,$CF,$D9,$7E,$EC,$7E,$17,$6F,$C0,$CE,$5F,$E0
	.dc.b	$86,$2B,$F1,$78,$79,$1B,$5F,$A7,$3F,$23,$5C,$5F,$DE,$DF,$82,$76
	.dc.b	$D8,$FC,$89,$E1,$3F,$1B,$1A,$BF,$04,$6A,$FC,$0D,$3C,$18,$7E,$5E
	.dc.b	$7C,$F1,$C8,$D5,$F7,$3D,$F5,$C9,$B1,$CB,$0F,$43,$F1,$EB,$C3,$3A
	.dc.b	$39,$3D,$A3,$DF,$87,$9D,$F0,$4E,$E7,$89,$BC,$3C,$D3,$F1,$23,$2D
	.dc.b	$38,$BB,$BF,$1C,$7E,$C4,$EE,$7B,$23,$8B,$E7,$38,$9C,$CD,$D7,$8D
	.dc.b	$C9,$FE,$DF,$C7,$F1,$6F,$E5,$1B,$8F,$E2,$30,$FC,$5E,$3F,$8A,$AF
	.dc.b	$C8,$1F,$B5,$C3,$F6,$93,$A8,$C1,$E1,$7D,$EF,$94,$F1,$C7,$B6,$7E
	.dc.b	$44,$60,$E1,$8F,$C0,$7F,$08,$64,$E5,$5D,$1F,$88,$D7,$6D,$BF,$BD
	.dc.b	$7D,$E7,$E4,$1F,$08,$E3,$5D,$E7,$71,$E2,$BF,$2D,$F3,$8F,$C3,$46
	.dc.b	$2B,$84,$60,$EA,$79,$CE,$B1,$B2,$3A,$9E,$98,$FE,$32,$3C,$06,$6B
	.dc.b	$AC,$66,$3E,$0C,$63,$0F,$C0,$C7,$63,$85,$F6,$AF,$C0,$47,$8C,$F8
	.dc.b	$B5,$C5,$F8,$46,$27,$85,$7C,$7A,$E3,$7F,$07,$1F,$C2,$DF,$DE,$11
	.dc.b	$F8,$18,$FF,$48,$D7,$E1,$30,$FE,$25,$5B,$3F,$13,$3F,$E8,$DE,$1F
	.dc.b	$E8,$31,$FE,$8E,$E5,$FF,$46,$BF,$D1,$78,$FF,$A3,$BF,$D3,$23,$2F
	.dc.b	$F3,$6E,$0F,$F4,$26,$7F,$CE,$F8,$BF,$A0,$4A,$31,$C7,$FD,$25,$FE
	.dc.b	$81,$C7,$FD,$39,$B7,$CB,$5F,$E9,$D8,$72,$9E,$38,$7F,$5A,$77,$C7
	.dc.b	$6B,$FA,$1C,$FF,$A5,$23,$FD,$2E,$47,$FD,$28,$43,$FD,$28,$7F,$48
	.dc.b	$E3,$FE,$7C,$FF,$43,$21,$FD,$08,$C5,$7E,$05,$FB,$CF,$D4,$DF,$ED
	.dc.b	$47,$DD,$E2,$FC,$CB,$F8,$6E,$2F,$B6,$FE,$32,$B8,$46,$23,$47,$E2
	.dc.b	$09,$1A,$6E,$3F,$83,$23,$CD,$AF,$28,$E0,$48,$CC,$7A,$1A,$FE,$3C
	.dc.b	$EA,$73,$21,$CD,$DD,$7B,$23,$91,$A2,$9B,$30,$75,$9F,$13,$88,$E4
	.dc.b	$E6,$7A,$BF,$B9,$9D,$BC,$C8,$76,$21,$C4,$E8,$71,$8E,$71,$F5,$E3
	.dc.b	$47,$22,$9C,$A7,$C6,$47,$91,$D2,$39,$46,$5B,$1A,$3A,$46,$67,$F8
	.dc.b	$1C,$76,$24,$65,$FD,$FB,$5F,$C4,$1C,$CF,$21,$4E,$25,$78,$71,$7F
	.dc.b	$18,$9E,$F1,$8C,$3F,$27,$39,$AE,$E7,$12,$5F,$A8,$97,$CA,$32,$D8
	.dc.b	$CC,$72,$6B,$E8,$6C,$F7,$8C,$4F,$12,$3C,$5A,$E8,$C1,$0C,$1F,$C7
	.dc.b	$2B,$F8,$CD,$7F,$1C,$8F,$F1,$74,$7F,$1C,$E2,$C4,$6F,$1F,$F1,$7F
	.dc.b	$17,$F1,$C2,$71,$7F,$1A,$9F,$F2,$D4,$7F,$20,$E0,$FF,$29,$E5,$FC
	.dc.b	$43,$87,$FC,$B6,$DA,$FF,$1C,$CE,$9F,$F3,$37,$0F,$F1,$77,$FC,$BF
	.dc.b	$5F,$C7,$6B,$84,$7F,$21,$7F,$8A,$E9,$FC,$62,$30,$70,$3F,$8D,$14
	.dc.b	$C9,$FE,$64,$AE,$CD,$E0,$FE,$38,$FF,$31,$6B,$FC,$6C,$FE,$3C,$53
	.dc.b	$F8,$F9,$FE,$79,$3F,$CB,$C5,$29,$9D,$FF,$91,$BF,$C8,$6B,$EE,$ED
	.dc.b	$F8,$C2,$6B,$FC,$66,$7F,$2F,$3A,$8C,$47,$E6,$23,$0F,$F0,$E3,$F4
	.dc.b	$E7,$E3,$8F,$C0,$6F,$F8,$92,$57,$68,$E0,$76,$CB,$F1,$33,$AB,$EF
	.dc.b	$87,$E8,$5B,$5D,$4F,$C5,$CE,$DE,$67,$96,$B9,$DF,$7C,$3F,$88,$CF
	.dc.b	$73,$CB,$19,$9F,$E1,$A7,$0B,$EF,$5E,$E3,$8D,$75,$38,$C6,$A7,$84
	.dc.b	$78,$B0,$F4,$34,$F4,$61,$E7,$BE,$73,$A6,$CE,$B0,$F2,$3F,$C5,$EF
	.dc.b	$AC,$E8,$F2,$3E,$88,$F1,$9D,$E3,$2F,$13,$2F,$18,$E6,$71,$8F,$97
	.dc.b	$3E,$07,$ED,$CF,$26,$CF,$07,$A1,$C3,$0F,$84,$F6,$8E,$98,$7B,$DB
	.dc.b	$7D,$67,$E9,$1C,$4F,$44,$75,$7B,$9C,$CE,$F3,$CD,$B3,$CD,$FA,$87
	.dc.b	$3B,$F7,$3C,$4E,$66,$6B,$C7,$7F,$B6,$8F,$F1,$37,$2F,$EA,$5A,$7F
	.dc.b	$89,$F9,$71,$BF,$F0,$DD,$5A,$FF,$B0,$B8,$FF,$C4,$BC,$5F,$E1,$CC
	.dc.b	$3F,$C7,$5C,$3F,$C3,$5F,$DF,$1F,$E3,$D9,$DE,$9F,$C3,$2B,$F8,$51
	.dc.b	$FC,$2F,$5E,$44,$8F,$E1,$E7,$F8,$92,$3B,$11,$C1,$FE,$28,$AF,$E1
	.dc.b	$47,$F8,$A4,$87,$F0,$83,$FD,$EB,$CB,$FE,$EA,$71,$82,$31,$1F,$C2
	.dc.b	$F8,$7F,$84,$B4,$FF,$15,$9F,$28,$FB,$27,$CF,$3E,$EF,$0F,$D7,$7E
	.dc.b	$8B,$5C,$F0,$FD,$99,$FD,$16,$1C,$F2,$FB,$35,$F5,$CF,$BF,$C3,$F8
	.dc.b	$E7,$07,$4B,$E4,$FE,$A6,$35,$87,$78,$FC,$2B,$70,$EF,$1D,$0F,$49
	.dc.b	$D0,$F2,$CF,$2A,$EB,$7F,$A3,$9C,$9C,$2F,$A3,$E1,$31,$3C,$30,$FC
	.dc.b	$E4,$70,$AF,$6D,$70,$7A,$47,$9A,$B8,$3F,$2C,$C5,$F8,$C8,$F6,$6C
	.dc.b	$F4,$9E,$33,$EB,$38,$9B,$24,$6E,$78,$DF,$D1,$24,$FC,$59,$D3,$C2
	.dc.b	$32,$FD,$19,$E2,$EE,$32,$FB,$DA,$F8,$A7,$99,$23,$38,$7D,$49,$E8
	.dc.b	$77,$3A,$10,$CC,$F1,$70,$77,$29,$DE,$7F,$31,$07,$FD,$55,$5F,$C0
	.dc.b	$67,$FD,$E3,$39,$7F,$C2,$71,$97,$FD,$F5,$3C,$1F,$F5,$81,$83,$67
	.dc.b	$FA,$E7,$7F,$F0,$B7,$2F,$FA,$B2,$3F,$C2,$BC,$9F,$E1,$74,$7F,$0E
	.dc.b	$3F,$6F,$87,$EF,$DF,$F0,$43,$FC,$0E,$FF,$80,$C6,$1F,$F0,$9C,$66
	.dc.b	$3F,$C3,$3C,$1F,$C6,$F8,$3F,$D5,$D5,$A3,$8C,$FF,$12,$AE,$CA,$76
	.dc.b	$F2,$3F,$C2,$FB,$FF,$AE,$EB,$94,$78,$4D,$57,$F8,$85,$FE,$1E,$43
	.dc.b	$F8,$91,$FE,$17,$21,$F0,$27,$F8,$77,$2F,$F2,$F6,$C7,$F0,$C3,$FC
	.dc.b	$30,$53,$E3,$9F,$E1,$C3,$F0,$35,$F8,$87,$F6,$15,$F8,$37,$27,$E0
	.dc.b	$B5,$FC,$63,$F8,$53,$24,$30,$43,$F1,$07,$13,$AC,$F7,$9F,$DB,$CF
	.dc.b	$49,$D9,$CC,$A7,$37,$44,$79,$9D,$63,$9B,$83,$94,$6E,$32,$49,$FE
	.dc.b	$17,$3D,$89,$3B,$8C,$B6,$3F,$85,$35,$FE,$08,$DD,$FF,$89,$4F,$09
	.dc.b	$F1,$1E,$92,$9D,$8B,$3C,$77,$FE,$1F,$5C,$A7,$B9,$A2,$5F,$A5,$CB
	.dc.b	$5F,$19,$AB,$F1,$46,$67,$A4,$EA,$31,$5E,$02,$6F,$FC,$4A,$7E,$49
	.dc.b	$CC,$EF,$7D,$08,$E0,$87,$57,$2D,$33,$7E,$22,$57,$FA,$6A,$7F,$D2
	.dc.b	$3B,$7E,$FE,$BF,$7B,$BF,$FA,$5B,$8F,$FC,$13,$39,$9F,$F0,$1F,$17
	.dc.b	$F7,$97,$FD,$0C,$FF,$80,$AF,$F6,$FB,$FE,$D2,$7F,$6D,$CB,$FE,$93
	.dc.b	$7F,$63,$3A,$8C,$46,$8F,$F0,$E6,$FF,$B9,$76,$7E,$E4,$D1,$FB,$EE
	.dc.b	$4F,$F4,$9C,$6E,$33,$BF,$F0,$68,$C1,$9C,$3C,$87,$95,$CD,$7F,$89
	.dc.b	$2B,$FC,$1E,$7F,$02,$3F,$86,$1F,$E0,$EE,$4F,$F0,$6B,$FC,$0D,$AF
	.dc.b	$F8,$53,$8B,$FC,$1E,$FF,$02,$3F,$C6,$9C,$58,$7E,$EE,$3F,$82,$E2
	.dc.b	$FC,$04,$7E,$5E,$3F,$9D,$3F,$EC,$93,$FD,$63,$C4,$E2,$57,$F3,$2E
	.dc.b	$3B,$5F,$CE,$9A,$FF,$10,$C3,$FC,$8D,$7C,$6B,$81,$FD,$0E,$B7,$19
	.dc.b	$AF,$F2,$2C,$FF,$50,$DF,$4C,$8E,$CE,$5A,$43,$D1,$5E,$27,$FA,$AC
	.dc.b	$F1,$73,$35,$C9,$C6,$BC,$55,$E4,$C3,$DA,$F3,$96,$1E,$C9,$CD,$F5
	.dc.b	$BE,$2F,$96,$BB,$9D,$67,$6F,$F3,$6C,$7E,$7D,$70,$8D,$57,$C6,$75
	.dc.b	$8F,$5B,$E3,$1D,$0F,$9C,$7D,$67,$35,$CE,$34,$7B,$8E,$0E,$4D,$39
	.dc.b	$71,$1C,$8E,$F1,$97,$15,$E5,$8E,$A7,$C4,$3C,$4E,$67,$9B,$B8,$E4
	.dc.b	$E6,$79,$C7,$E0,$0F,$79,$F0,$8C,$1D,$B0,$FA,$13,$C4,$91,$A3,$A9
	.dc.b	$C2,$BC,$2E,$4F,$36,$1F,$42,$FC,$38,$FD,$FB,$F3,$56,$B1,$FA,$11
	.dc.b	$DA,$7F,$05,$8F,$8A,$BB,$5F,$2A,$FC,$06,$3E,$0B,$E9,$5E,$1B,$E1
	.dc.b	$18,$25,$70,$BE,$55,$F9,$B3,$D7,$3A,$70,$79,$4F,$8C,$E5,$FC,$1E
	.dc.b	$3E,$DA,$F4,$1F,$9F,$8E,$07,$E0,$AB,$BB,$C3,$1F,$7B,$EC,$9F,$87
	.dc.b	$87,$A1,$CE,$1F,$0A,$7E,$F9,$B2,$39,$C3,$C5,$18,$9F,$15,$73,$8F
	.dc.b	$4C,$7C,$13,$CE,$FB,$A3,$4F,$9E,$3C,$6F,$82,$BE,$06,$3F,$9A,$8F
	.dc.b	$6B,$F4,$67,$BC,$6D,$F4,$9D,$DF,$21,$DC,$D1,$E4,$3F,$26,$78,$4D
	.dc.b	$9F,$1C,$D1,$F9,$53,$89,$B9,$F8,$58,$78,$23,$B6,$1F,$12,$7D,$38
	.dc.b	$FE,$BA,$7D,$2F,$D4,$C3,$D9,$8F,$C2,$7D,$93,$E4,$6B,$E1,$BE,$98
	.dc.b	$F7,$8F,$31,$E3,$7E,$1C,$6A,$3D,$27,$E3,$A3,$83,$4C,$9E,$47,$EA
	.dc.b	$BD,$F1,$ED,$5E,$68,$F8,$2F,$91,$E4,$78,$4F,$46,$1E,$D3,$E0,$61
	.dc.b	$ED,$7F,$27,$1C,$27,$93,$E1,$3E,$8B,$DF,$0F,$54,$75,$3D,$F1,$F1
	.dc.b	$0D,$5F,$9A,$BB,$CF,$99,$EE,$F5,$C7,$E9,$E3,$F8,$EC,$3A,$E1,$E2
	.dc.b	$C7,$CD,$3F,$0A,$FC,$D8,$72,$9C,$C7,$E1,$8C,$CE,$DE,$CF,$CE,$69
	.dc.b	$F3,$A3,$8E,$1F,$BB,$8E,$27,$33,$67,$BC,$CB,$B7,$37,$E7,$7D,$D5
	.dc.b	$E1,$AE,$AF,$E0,$B1,$F8,$8F,$BD,$B3,$FC,$5C,$E2,$78,$67,$8B,$F9
	.dc.b	$A8,$DD,$FE,$7A,$3F,$47,$7F,$41,$B8,$7D,$0C,$7E,$5B,$F1,$E3,$AD
	.dc.b	$7A,$70,$EB,$1F,$8D,$38,$E3,$E5,$69,$E5,$BE,$71,$ED,$8E,$6F,$B5
	.dc.b	$DD,$79,$5F,$C3,$61,$E5,$6E,$1E,$5C,$7D,$77,$FC,$16,$FF,$07,$5D
	.dc.b	$E7,$2F,$E6,$A7,$F4,$77,$CC,$FC,$85,$F3,$9F,$0C,$F3,$3F,$01,$8F
	.dc.b	$BA,$B8,$57,$E8,$E7,$DD,$5E,$46,$CF,$46,$CE,$E7,$C8,$76,$3C,$87
	.dc.b	$E9,$9F,$13,$CC,$D9,$70,$F8,$95,$D6,$32,$D8,$CC,$F7,$6D,$7C,$08
	.dc.b	$EB,$3B,$8F,$75,$7E,$6C,$FA,$67,$E1,$CF,$31,$FB,$69,$E9,$1C,$23
	.dc.b	$F3,$87,$E6,$A3,$F0,$D8,$FD,$E8,$FC,$91,$D0,$FC,$31,$EB,$9E,$55
	.dc.b	$FA,$19,$E1,$7F,$8D,$AE,$13,$F4,$2B,$D4,$7C,$C2,$C7,$C2,$BF,$2B
	.dc.b	$F3,$2B,$F0,$67,$C2,$7E,$74,$F1,$69,$C5,$E7,$1B,$38,$BE,$69,$F6
	.dc.b	$E3,$F8,$CA,$F8,$93,$C0,$D1,$F9,$DC,$3D,$6F,$E7,$63,$E0,$1E,$57
	.dc.b	$F1,$AF,$BA,$BE,$A9,$EA,$7A,$9E,$6B,$F2,$4F,$0A,$EB,$3F,$5C,$FE
	.dc.b	$C0,$08,$7F,$A1,$4A,$53,$38,$7F,$9E,$DC,$AF,$17,$FC,$D5,$08,$C5
	.dc.b	$72,$8E,$D1,$87,$0D,$7B,$F2,$38,$FF,$D8,$8F,$FB,$1E,$39,$39,$E6
	.dc.b	$E9,$0E,$1F,$F5,$97,$1F,$FB,$03,$8F,$FD,$65,$C7,$FE,$BA,$3F,$D6
	.dc.b	$5A,$FF,$A8,$A3,$FD,$6F,$3F,$EC,$6A,$FF,$5B,$ED,$F0,$EB,$53,$CC
	.dc.b	$D5,$F3,$9D,$70,$7C,$8C,$39,$E3,$FE,$CE,$AD,$2C,$39,$BC,$0F,$F6
	.dc.b	$86,$3E,$07,$57,$CF,$1F,$04,$73,$C6,$3A,$3F,$DB,$18,$FB,$A3,$31
	.dc.b	$F6,$2D,$8F,$FB,$46,$30,$D7,$FD,$A4,$7F,$B3,$4F,$AD,$0E,$3F,$F6
	.dc.b	$8B,$78,$FE,$7B,$67,$E8,$9F,$E8,$53,$FD,$A6,$63,$8F,$FD,$A8,$FF
	.dc.b	$B4,$73,$FC,$4F,$1F,$FB,$33,$5F,$F6,$9C,$E4,$FC,$2C,$66,$7F,$C7
	.dc.b	$53,$A9,$FF,$03,$36,$BF,$CE,$CD,$8F,$F2,$BB,$4F,$F2,$A9,$67,$C3
	.dc.b	$C5,$F4,$67,$FD,$5B,$C9,$FE,$5A,$39,$16,$7F,$50,$7E,$E6,$3F,$B9
	.dc.b	$CF,$ED,$78,$FF,$92,$47,$ED,$4F,$E4,$97,$FA,$E7,$F8,$CC,$69,$FD
	.dc.b	$AE,$DF,$C1,$9E,$07,$F2,$4C,$7F,$A4,$65,$FB,$CC,$3F,$83,$4E,$AF
	.dc.b	$F9,$06,$3E,$38,$FE,$55,$3F,$C2,$63,$17,$CA,$3F,$8C,$61,$DE,$BC
	.dc.b	$31,$EC,$7F,$9B,$9E,$C3,$F8,$25,$F9,$DC,$1E,$AC,$3F,$87,$38,$8F
	.dc.b	$E5,$CF,$F0,$F9,$FE,$36,$FF,$3A,$C7,$E1,$4E,$AF,$C5,$3C,$6F,$DC
	.dc.b	$F1,$3D,$CF,$F2,$87,$CA,$7B,$8E,$98,$FB,$9F,$E4,$58,$7B,$8F,$63
	.dc.b	$FC,$A0,$F7,$9F,$52,$38,$61,$E4,$AF,$0D,$F9,$2B,$B3,$FD,$2E,$75
	.dc.b	$7D,$67,$91,$E4,$C7,$E7,$3F,$D0,$CF,$51,$E0,$3E,$E1,$E6,$35,$5D
	.dc.b	$E3,$8B,$FC,$DA,$3A,$9E,$93,$E3,$4F,$85,$E7,$5E,$47,$9C,$7C,$0C
	.dc.b	$7E,$C4,$7A,$A3,$F2,$D8,$7D,$88,$C1,$F7,$9C,$61,$E2,$9F,$B3,$87
	.dc.b	$C1,$C7,$E3,$92,$FE,$01,$EB,$8F,$43,$E4,$8D,$3E,$93,$B9,$F8,$99
	.dc.b	$FC,$DB,$F1,$70,$FC,$DB,$D3,$1F,$F0,$13,$FD,$33,$1F,$E9,$73,$FD
	.dc.b	$3E,$3F,$72,$7F,$48,$8C,$92,$BF,$A1,$CE,$E7,$F8,$5C,$7F,$5D,$C0
	.dc.b	$FF,$5A,$ED,$FF,$59,$11,$FF,$1F,$B6,$7F,$D2,$6D,$9F,$F6,$9B,$67
	.dc.b	$C7,$1D,$27,$FA,$EC,$E5,$F0,$88,$D3,$FD,$96,$34,$FF,$67,$BE,$53
	.dc.b	$AB,$56,$94,$EA,$FD,$47,$A0,$97,$E8,$6D,$FA,$1A,$7A,$0E,$31,$B2
	.dc.b	$46,$DB,$1B,$7D,$AF,$27,$DB,$1E,$7B,$F3,$57,$96,$76,$72,$37,$5E
	.dc.b	$0A,$EC,$47,$B4,$7C,$57,$B3,$F1,$5F,$B4,$71,$37,$7F,$19,$DB,$F1
	.dc.b	$9D,$C7,$8A,$7B,$3E,$23,$11,$CA,$FD,$2F,$08,$F1,$1E,$93,$C4,$7D
	.dc.b	$B7,$C4,$F3,$37,$3D,$9D,$BF,$16,$3C,$B7,$F1,$67,$71,$E1,$32,$78
	.dc.b	$6F,$D9,$1E,$17,$D8,$F8,$4F,$49,$E1,$77,$5E,$17,$6F,$86,$39,$57
	.dc.b	$86,$77,$5E,$19,$CB,$E1,$BF,$87,$1D,$CF,$39,$DD,$CD,$F5,$3C,$17
	.dc.b	$D6,$33,$5F,$1A,$BE,$1B,$6F,$CA,$DA,$E0,$FC,$29,$F0,$4F,$99,$F0
	.dc.b	$46,$6F,$C0,$FA,$5F,$0B,$E9,$34,$75,$31,$1E,$E3,$AD,$F9,$9F,$96
	.dc.b	$79,$8B,$5E,$66,$DF,$99,$E5,$7E,$27,$CA,$62,$FC,$0F,$67,$C6,$F2
	.dc.b	$7C,$71,$E8,$8D,$1C,$4D,$18,$74,$77,$7C,$C7,$D4,$2D,$7C,$76,$DF
	.dc.b	$B5,$D4,$7A,$DB,$7E,$BA,$EB,$7E,$87,$B5,$6C,$E8,$5B,$E8,$E2,$B9
	.dc.b	$57,$D3,$25,$FC,$4B,$EA,$FB,$1E,$B3,$C2,$35,$3C,$1F,$4B,$F2,$1F
	.dc.b	$3B,$F2,$0D,$57,$8D,$D3,$E3,$75,$1D,$08,$74,$3F,$79,$C7,$FE,$8B
	.dc.b	$7F,$D1,$BC,$9F,$E9,$0E,$77,$3F,$FA,$3A,$B5,$5F,$EA,$4E,$F7,$7F
	.dc.b	$F9,$89,$B1,$FE,$94,$6B,$FE,$A1,$6B,$FE,$72,$D3,$FC,$A5,$CD,$FE
	.dc.b	$88,$8F,$F3,$E1,$FE,$9B,$E6,$FF,$44,$F0,$7F,$A2,$23,$33,$FE,$7F
	.dc.b	$77,$C9,$FE,$A0,$79,$F3,$FF,$A5,$B0,$FF,$43,$DF,$12,$71,$FF,$AF
	.dc.b	$AF,$FD,$21,$18,$BF,$F4,$9E,$3E,$0A,$FF,$49,$47,$43,$9B,$C6,$3F
	.dc.b	$D5,$6E,$A3,$CB,$87,$85,$F2,$CF,$FA,$3D,$FF,$67,$5F,$FA,$53,$0F
	.dc.b	$F4,$5C,$E6,$7E,$04,$7F,$A3,$B9,$FF,$D2,$24,$E7,$C1,$FE,$8C,$9F
	.dc.b	$F4,$81,$FE,$CD,$E6,$C1,$FE,$84,$CB,$FD,$12,$48,$FF,$6D,$C6,$5A
	.dc.b	$7F,$A2,$23,$4F,$FB,$AB,$3F,$F4,$43,$FD,$E9,$AF,$F4,$26,$F2,$7F
	.dc.b	$34,$6F,$2F,$FA,$66,$7A,$C7,$FA,$4A,$BA,$C7,$FB,$BE,$FC,$4F,$F8
	.dc.b	$A4,$FF,$11,$47,$F6,$2F,$A2,$7B,$9F,$EE,$C6,$F3,$FF,$B8,$1B,$5F
	.dc.b	$EE,$D6,$BF,$EF,$87,$3A,$FF,$83,$3E,$AD,$F0,$3F,$D4,$67,$37,$FE
	.dc.b	$25,$9C,$BF,$D8,$7E,$29,$F4,$7F,$D5,$DB,$1F,$DB,$3C,$A7,$9B,$FE
	.dc.b	$F7,$8E,$15,$B3,$FC,$3F,$98,$97,$C9,$C1,$1C,$38,$0E,$1D,$47,$F8
	.dc.b	$36,$FC,$3E,$F7,$7B,$F2,$60,$8F,$ED,$DF,$9F,$BE,$1E,$9C,$65,$F0
	.dc.b	$38,$BF,$F0,$05,$6D,$F1,$65,$E9,$8C,$38,$6B,$E7,$6B,$EC,$77,$BF
	.dc.b	$F9,$17,$2F,$8F,$EE,$F7,$AF,$EF,$51,$8B,$F0,$FD,$0F,$A0,$F3,$71
	.dc.b	$58,$C9,$E4,$A6,$2A,$F9,$F1,$E7,$6B,$A6,$C6,$E7,$E3,$FC,$F7,$5A
	.dc.b	$F1,$E8,$8F,$F0,$37,$91,$2B,$FC,$57,$E1,$BE,$B5,$F0,$D9,$C7,$FB
	.dc.b	$BD,$DD,$FF,$CC,$97,$BA,$C6,$1F,$80,$EC,$5E,$30,$FE,$5F,$8F,$4A
	.dc.b	$FF,$11,$4E,$65,$5C,$27,$C7,$5B,$9F,$0C,$70,$23,$FE,$27,$C3,$C5
	.dc.b	$18,$C7,$FC,$DD,$87,$86,$73,$5C,$89,$3A,$AE,$F8,$F8,$F1,$F1,$56
	.dc.b	$A7,$F3,$D7,$CA,$F8,$4E,$4F,$D1,$57,$08,$F8,$B8,$FA,$5D,$92,$FE
	.dc.b	$36,$3F,$E6,$BC,$3A,$46,$0F,$51,$70,$F7,$5F,$38,$C5,$F3,$9F,$0D
	.dc.b	$F3,$AD,$46,$2B,$9B,$4E,$78,$FB,$EB,$E3,$57,$09,$FA,$71,$9A,$F8
	.dc.b	$A7,$28,$CB,$D3,$0F,$4C,$75,$C3,$C8,$E8,$EF,$3C,$71,$F8,$CF,$A8
	.dc.b	$F9,$B8,$78,$63,$33,$CE,$74,$FE,$2B,$1F,$23,$F4,$70,$F0,$BD,$F0
	.dc.b	$F0,$DF,$29,$CB,$D2,$75,$7D,$27,$38,$F4,$BF,$B3,$85,$AD,$63,$E5
	.dc.b	$AC,$61,$E3,$9F,$64,$7C,$AA,$CC,$F6,$34,$7E,$1C,$C1,$9C,$3C,$35
	.dc.b	$C8,$F9,$F1,$BC,$7D,$07,$AF,$0F,$14,$EE,$7E,$69,$BC,$3C,$51,$9C
	.dc.b	$7F,$0D,$87,$6A,$D4,$F6,$9F,$8F,$5C,$27,$27,$CD,$BE,$EE,$27,$E2
	.dc.b	$D7,$53,$E2,$E3,$F0,$31,$F3,$37,$0F,$CA,$36,$7C,$07,$DC,$BF,$87
	.dc.b	$5C,$23,$A4,$7E,$04,$F9,$37,$E6,$8F,$C8,$61,$E4,$8C,$4F,$C2,$69
	.dc.b	$F0,$23,$53,$DE,$BB,$57,$AE,$3D,$27,$81,$FA,$93,$C1,$F1,$18,$BF
	.dc.b	$53,$DE,$7A,$61,$F2,$70,$F0,$C7,$49,$EC,$7A,$F1,$FA,$53,$A8,$F6
	.dc.b	$CF,$23,$DB,$7D,$A7,$E7,$CF,$23,$C5,$3E,$7C,$7D,$73,$CE,$72,$6E
	.dc.b	$33,$3D,$63,$EE,$57,$58,$FD,$8B,$63,$C6,$7C,$09,$FE,$46,$FF,$79
	.dc.b	$6B,$FC,$54,$C6,$1F,$D5,$23,$31,$FC,$3C,$CD,$7F,$5D,$8C,$C7,$E9
	.dc.b	$6D,$CC,$F4,$BF,$A4,$46,$1B,$19,$BF,$E3,$73,$96,$C6,$6B,$F9,$A4
	.dc.b	$E5,$B1,$98,$FE,$50,$64,$AE,$67,$F8,$7C,$65,$B1,$99,$FC,$A4,$E5
	.dc.b	$B1,$97,$D8,$DB,$FE,$37,$3B,$9C,$DF,$B0,$97,$EA,$6C,$7A,$1B,$3E
	.dc.b	$66,$CF,$91,$B3,$E2,$6B,$DA,$3A,$11,$E8,$D9,$E8,$D3,$A1,$63,$A3
	.dc.b	$5E,$87,$22,$47,$26,$C7,$26,$9C,$8A,$72,$8E,$04,$8E,$0D,$8E,$0D
	.dc.b	$78,$46,$88,$E9,$AE,$8C,$10,$C1,$2B,$EA,$DF,$D7,$C1,$5F,$E5,$DB
	.dc.b	$FF,$37,$61,$F3,$2F,$6D,$AD,$E0,$AD,$DF,$C7,$BD,$B6,$B7,$3A,$AF
	.dc.b	$87,$5A,$6C,$EA,$BF,$CF,$57,$DD,$F2,$18,$3A,$18,$9F,$16,$1E,$B8
	.dc.b	$C6,$1F,$91,$C3,$B3,$F2,$DB,$1F,$21,$B5,$F1,$DA,$FC,$7C,$3C,$8F
	.dc.b	$6C,$3C,$8F,$C1,$AF,$23,$EA,$7C,$38,$78,$E3,$47,$18,$F4,$B8,$C3
	.dc.b	$CD,$7D,$9B,$3C,$AF,$BC,$7C,$93,$D5,$3B,$BE,$D1,$9A,$F6,$4E,$6B
	.dc.b	$8C,$79,$A7,$8B,$D4,$F1,$BC,$6F,$E1,$57,$19,$C9,$C4,$EA,$71,$8F
	.dc.b	$80,$7A,$EF,$C7,$1B,$8E,$6E,$EB,$C3,$1B,$BF,$9A,$BF,$40,$7D,$BE
	.dc.b	$1B,$7F,$EA,$4F,$3B,$13,$23,$FD,$15,$3A,$AF,$F4,$FB,$5F,$F5,$93
	.dc.b	$5F,$F5,$BB,$BC,$3F,$D3,$9C,$0F,$7B,$A5,$7F,$A2,$9F,$17,$C0,$4C
	.dc.b	$3F,$D6,$BF,$51,$6B,$EE,$9E,$99,$FF,$4B,$E2,$69,$2B,$CE,$FF,$AE
	.dc.b	$7E,$D3,$96,$5E,$27,$E1,$F0,$7F,$2B,$DB,$FC,$D1,$F0,$FE,$5C,$3F
	.dc.b	$41,$97,$FA,$2A,$76,$7F,$53,$24,$7F,$51,$6C,$7F,$9B,$3E,$9F,$73
	.dc.b	$FE,$90,$7E,$8F,$CE,$BD,$7F,$F5,$7B,$67,$FD,$6C,$DA,$FF,$65,$B5
	.dc.b	$FF,$6F,$7A,$2F,$4D,$EB,$7F,$AF,$63,$0D,$F5,$1D,$59,$AD,$DF,$FB
	.dc.b	$5B,$D2,$7A,$5F,$EE,$CE,$2F,$E1,$F2,$9F,$F6,$33,$F0,$9A,$FD,$27
	.dc.b	$8B,$99,$FF,$67,$3C,$5B,$1A,$8F,$04,$6B,$D2,$7A,$5F,$C7,$65,$E7
	.dc.b	$BC,$EF,$F7,$FE,$1C,$63,$12,$3B,$15,$CA,$BE,$B8,$E6,$F8,$4E,$70
	.dc.b	$F0,$5F,$0C,$3B,$57,$83,$1D,$CF,$4C,$79,$E1,$E6,$AE,$71,$89,$E5
	.dc.b	$5C,$27,$95,$78,$4F,$45,$6E,$38,$E1,$E2,$77,$3D,$4F,$76,$1E,$47
	.dc.b	$8C,$ED,$B8,$F5,$7E,$19,$C4,$B8,$FA,$67,$84,$7C,$19,$E1,$5D,$63
	.dc.b	$77,$E4,$9E,$98,$79,$E7,$C7,$8F,$B5,$A6,$B0,$EB,$3F,$1A,$3E,$79
	.dc.b	$DC,$A5,$32,$ED,$F9,$7E,$93,$D2,$DC,$7D,$B8,$C3,$CA,$51,$88,$E6
	.dc.b	$D7,$BF,$A3,$F9,$44,$D2,$3F,$98,$7E,$57,$43,$D2,$FE,$E3,$B7,$E3
	.dc.b	$2B,$F2,$15,$A9,$CE,$BD,$A7,$E8,$5F,$1E,$5F,$A9,$1C,$0C,$D7,$07
	.dc.b	$E2,$DF,$39,$DC,$79,$8D,$B8,$8D,$47,$DD,$3A,$61,$FA,$0B,$F0,$46
	.dc.b	$4E,$51,$9B,$F0,$47,$E8,$AB,$9C,$FB,$63,$78,$7E,$9A,$F9,$CE,$6B
	.dc.b	$73,$EF,$8E,$98,$79,$6F,$CE,$4D,$BF,$85,$E1,$E4,$7F,$86,$C7,$A2
	.dc.b	$FB,$63,$FC,$C8,$97,$FC,$C2,$74,$7E,$FE,$F9,$47,$EE,$BD,$17,$95
	.dc.b	$FD,$D7,$D2,$FE,$BB,$18,$9F,$F0,$2C,$62,$75,$1F,$CF,$27,$57,$FD
	.dc.b	$AE,$BF,$B2,$FA,$4F,$4B,$CF,$3F,$C3,$E3,$17,$F9,$46,$C7,$E7,$1A
	.dc.b	$FE,$81,$CE,$9F,$D8,$3D,$2E,$B7,$FE,$1E,$F3,$DE,$32,$73,$8F,$F8
	.dc.b	$BE,$73,$18,$C7,$FC,$3B,$18,$9F,$EA,$BE,$73,$D2,$F3,$D2,$31,$3F
	.dc.b	$AA,$6C,$7E,$C9,$AF,$ED,$9C,$E5,$FD,$DB,$D1,$FF,$03,$C6,$2B,$F9
	.dc.b	$86,$1E,$89,$CE,$1E,$BC,$91,$8A,$F5,$46,$1F,$3F,$A3,$E6,$C0,$D2
	.dc.b	$47,$9D,$E9,$E8,$F8,$9B,$1C,$8E,$B1,$96,$9F,$EF,$EE,$EF,$F7,$1B
	.dc.b	$FE,$D5,$CF,$FD,$BE,$7F,$B6,$F8,$FF,$DB,$3A,$FF,$B4,$FC,$FF,$F7
	.dc.b	$75,$6C,$FF,$77,$56,$E3,$FC,$AB,$7D,$2B,$85,$7F,$99,$63,$10,$B5
	.dc.b	$74,$8D,$9F,$EE,$6A,$FF,$77,$47,$4F,$45,$E8,$FF,$B9,$F0,$E7,$87
	.dc.b	$2B,$FF,$7C,$9C,$0F,$05,$70,$71,$87,$F9,$8A,$B8,$E1,$E1,$8F,$23
	.dc.b	$FE,$20,$31,$7D,$63,$FC,$5A,$DC,$7F,$73,$3E,$5C,$7C,$98,$F7,$BD
	.dc.b	$CF,$3C,$7C,$B1,$87,$E9,$4E,$EF,$E5,$47,$A2,$F9,$91,$F4,$C7,$D6
	.dc.b	$7E,$95,$7F,$01,$BF,$14,$68,$FF,$01,$63,$EE,$7C,$35,$E8,$7F,$C7
	.dc.b	$38,$FB,$6F,$FA,$9B,$5F,$E3,$D7,$FA,$88,$C4,$7F,$9A,$8F,$F3,$44
	.dc.b	$FF,$9D,$1F,$F4,$61,$FE,$72,$AF,$F3,$D3,$FE,$99,$8F,$F3,$81,$FE
	.dc.b	$78,$78,$9B,$7F,$CE,$CE,$CC,$C7,$F9,$D8,$E7,$3F,$EA,$42,$57,$FA
	.dc.b	$16,$79,$1E,$29,$E1,$5D,$2B,$6F,$8A,$7D,$53,$D2,$B9,$3C,$23,$94
	.dc.b	$F9,$E3,$4F,$18,$F3,$46,$1F,$3C,$6C,$F0,$4F,$47,$11,$D0,$F0,$BA
	.dc.b	$3A,$57,$9A,$39,$9A,$38,$1E,$08,$E0,$E8,$C1,$A8,$E5,$5E,$52,$4F
	.dc.b	$06,$CE,$EB,$CC,$F7,$8F,$F6,$0B,$D6,$B9,$D7,$12,$47,$17,$D3,$5C
	.dc.b	$A3,$C7,$5C,$8E,$E5,$9E,$2D,$8E,$47,$79,$E5,$5E,$08,$CC,$F2,$F4
	.dc.b	$7F,$D7,$B1,$8B,$E7,$18,$9E,$84,$70,$4F,$27,$FD,$A7,$5C,$49,$3B
	.dc.b	$8C,$B6,$BF,$3C,$D9,$FC,$7B,$4F,$C4,$16,$3E,$FB,$63,$E6,$57,$86
	.dc.b	$79,$B6,$39,$B6,$79,$B6,$79,$B6,$79,$D7,$9A,$31,$1F,$41,$B7,$F2
	.dc.b	$6F,$C5,$39,$AE,$04,$BF,$1B,$63,$C6,$D9,$F1,$B6,$7C,$6D,$9F,$1C
	.dc.b	$72,$23,$C9,$B3,$C9,$A7,$22,$C7,$23,$81,$23,$83,$63,$83,$4E,$11
	.dc.b	$A2,$46,$9B,$1A,$8C,$1F,$EF,$2F,$43,$FD,$CE,$47,$FD,$F6,$D8,$FF
	.dc.b	$79,$47,$FB,$83,$CE,$FF,$7C,$3F,$E0,$0E,$1F,$F7,$24,$FF,$BC,$A7
	.dc.b	$2B,$93,$FC,$37,$E8,$7F,$BF,$31,$FF,$7F,$BF,$EF,$99,$FF,$41,$D7
	.dc.b	$FB,$DF,$0F,$F7,$A9,$23,$FD,$0B,$19,$7F,$C3,$F8,$3C,$EF,$F7,$CD
	.dc.b	$78,$DB,$3E,$19,$D4,$FA,$7C,$FF,$F8,$00,$FF,$42,$D3,$FD,$EB,$5F
	.dc.b	$D0,$9A,$FF,$97,$DF,$EE,$5E,$77,$FB,$E2,$3F,$DF,$53,$C4,$FF,$7D
	.dc.b	$15,$C8,$CF,$FD,$D3,$E6,$7F,$BC,$A3,$FD,$3E,$D7,$FD,$F3,$E5,$FF
	.dc.b	$BD,$1C,$F9,$7F,$EF,$47,$FD,$4F,$F2,$0F,$4B,$FD,$D6,$D8,$FF,$51
	.dc.b	$B5,$FF,$55,$B5,$FF,$57,$F9,$BF,$EE,$FC,$7F,$DF,$6F,$FB,$E1,$BD
	.dc.b	$7F,$EF,$FC,$FF,$D9,$F1,$AF,$37,$17,$FE,$BC,$3F,$DF,$CE,$1F,$F5
	.dc.b	$D7,$91,$FE,$CB,$9F,$F6,$4C,$FF,$B0,$E7,$FD,$92,$7F,$B5,$0F,$F6
	.dc.b	$97,$97,$FE,$C9,$72,$47,$06,$71,$FF,$69,$1F,$ED,$37,$3E,$6F,$FB
	.dc.b	$62,$33,$5D,$E7,$37,$E2,$B8,$78,$2F,$FD,$A6,$71,$AF,$35,$F5,$8C
	.dc.b	$D7,$0C,$3F,$DF,$37,$FE,$DB,$88,$FF,$B3,$E7,$EF,$36,$3F,$DA,$B8
	.dc.b	$7F,$B6,$63,$FD,$FB,$E6,$7F,$B4,$5B,$E6,$EA,$31,$1F,$ED,$B8,$FF
	.dc.b	$68,$BF,$ED,$EF,$36,$BF,$EC,$B3,$07,$FA,$B3,$1F,$F5,$93,$7C,$DF
	.dc.b	$F5,$91,$FE,$AA,$D7,$FD,$65,$3F,$EB,$49,$FF,$61,$57,$FA,$E1,$FF
	.dc.b	$67,$9C,$4B,$1B,$3A,$96,$BA,$C7,$FA,$C7,$CD,$FF,$64,$BF,$EB,$43
	.dc.b	$FC,$BC,$E1,$FF,$64,$9B,$BF,$1C,$F9,$63,$C6,$78,$4F,$F6,$91,$7B
	.dc.b	$BF,$D7,$AF,$13,$F5,$93,$E2,$7B,$BF,$E3,$CA,$FF,$63,$C6,$8F,$5D
	.dc.b	$77,$D1,$49,$FF,$5D,$79,$3F,$EA,$B9,$CC,$7F,$09,$E4,$C5,$6E,$1E
	.dc.b	$57,$FA,$F1,$C3,$5E,$71,$FE,$C3,$BF,$F2,$57,$85,$FE,$6C,$3F,$CC
	.dc.b	$93,$FE,$7F,$3F,$CC,$84,$3F,$CA,$6F,$09,$F0,$47,$F9,$51,$AF,$A7
	.dc.b	$F4,$1F,$F6,$BE,$3E,$88,$46,$27,$F9,$B7,$DA,$7B,$2B,$53,$9A,$E5
	.dc.b	$E7,$E6,$FD,$D3,$9F,$0F,$FD,$A7,$86,$6F,$75,$F2,$70,$E7,$87,$FB
	.dc.b	$3B,$0F,$F7,$4D,$7A,$1A,$F4,$8D,$78,$7C,$E3,$CE,$FF,$B1,$FE,$D3
	.dc.b	$15,$FD,$32,$FC,$F3,$83,$58,$78,$4F,$3B,$97,$E2,$DF,$46,$D7,$F5
	.dc.b	$68,$F2,$4F,$F9,$F7,$C0,$FE,$79,$7F,$80,$F1,$DE,$A7,$87,$1F,$E8
	.dc.b	$47,$F3,$D9,$F9,$DC,$8B,$FF,$60,$D6,$F1,$FC,$64,$E6,$F8,$1C,$A3
	.dc.b	$FD,$65,$3B,$7F,$D8,$95,$E2,$9D,$5F,$8E,$B5,$7C,$EF,$FD,$67,$7B
	.dc.b	$58,$7A,$31,$C4,$74,$C7,$C2,$EB,$0E,$77,$AC,$3A,$63,$DF,$0C,$4F
	.dc.b	$5C,$7A,$63,$CB,$1E,$87,$89,$FF,$6F,$63,$D1,$D6,$3D,$1F,$F5,$9C
	.dc.b	$66,$78,$13,$1F,$04,$F0,$C3,$C1,$5B,$9F,$9F,$18,$C7,$CB,$5E,$CC
	.dc.b	$7C,$B8,$73,$8F,$92,$47,$E0,$4F,$07,$46,$23,$C0,$73,$C3,$C0,$F1
	.dc.b	$8F,$7E,$1E,$67,$89,$88,$ED,$7E,$29,$F3,$C7,$49,$FB,$F7,$E1,$9C
	.dc.b	$C6,$27,$51,$83,$B1,$5F,$16,$3F,$2E,$7D,$07,$89,$C3,$E3,$33,$3C
	.dc.b	$5F,$C9,$E3,$F2,$6B,$B9,$E8,$C7,$DF,$3A,$8F,$85,$87,$9D,$D3,$63
	.dc.b	$47,$A4,$F1,$57,$56,$CF,$23,$BB,$C2,$FD,$B1,$EC,$9E,$11,$E7,$9F
	.dc.b	$51,$EB,$C7,$E6,$56,$CE,$91,$D8,$FB,$71,$A7,$B4,$6D,$F3,$4F,$5C
	.dc.b	$7E,$39,$F2,$EB,$84,$F8,$23,$37,$FB,$CC,$3E,$04,$F8,$F1,$F8,$EE
	.dc.b	$E7,$58,$FB,$A3,$55,$C2,$FB,$5F,$07,$AD,$7E,$A2,$75,$7C,$AF,$D9
	.dc.b	$8F,$D9,$C7,$CB,$87,$09,$F7,$E3,$CE,$FE,$4E,$3E,$FC,$38,$4E,$63
	.dc.b	$E3,$C7,$C1,$9D,$D7,$C6,$21,$E8,$9F,$55,$FA,$31,$F2,$C7,$9C,$CD
	.dc.b	$79,$EF,$A4,$76,$3C,$47,$E2,$0D,$9C,$6F,$DA,$E5,$A7,$8F,$1F,$7D
	.dc.b	$7E,$3D,$B3,$F8,$A6,$CF,$06,$D7,$8A,$3E,$FC,$F8,$9E,$98,$FC,$37
	.dc.b	$E4,$E1,$EC,$76,$7C,$2C,$7E,$75,$FE,$5A,$31,$8F,$C8,$7F,$01,$87
	.dc.b	$BD,$CB,$5C,$9D,$9F,$82,$F6,$33,$3F,$5A,$FC,$4F,$78,$CC,$FC,$58
	.dc.b	$F6,$1B,$79,$BB,$BF,$3C,$6C,$F8,$13,$F6,$CF,$9A,$59,$E7,$1B,$8F
	.dc.b	$49,$E7,$37,$3F,$81,$8E,$E7,$F3,$AA,$F3,$B4,$CC,$FF,$10,$7C,$33
	.dc.b	$C6,$39,$DF,$E1,$E7,$2D,$8F,$0D,$FB,$5D,$CF,$E1,$E3,$CF,$1F,$6A
	.dc.b	$FC,$2F,$73,$8C,$FA,$DF,$92,$6D,$F2,$E3,$F8,$67,$F0,$8D,$8E,$27
	.dc.b	$33,$89,$F1,$E3,$CE,$78,$CA,$FD,$BC,$3F,$68,$FF,$B2,$63,$FC,$23
	.dc.b	$1F,$DD,$E3,$07,$F8,$E1,$FE,$21,$E5,$3C,$3F,$ED,$CD,$AF,$EB,$8D
	.dc.b	$AF,$E3,$AD,$C7,$F8,$23,$67,$FA,$8B,$67,$F5,$4D,$7F,$95,$36,$FC
	.dc.b	$37,$E0,$25,$78,$1B,$5E,$06,$DF,$81,$B3,$E0,$6C,$F8,$1A,$F8,$1B
	.dc.b	$5E,$09,$EC,$49,$EC,$DA,$EC,$D3,$B1,$63,$B3,$5E,$CD,$9E,$D1,$D0
	.dc.b	$93,$D1,$B3,$D1,$A7,$42,$9D,$0B,$3D,$27,$91,$27,$93,$67,$93,$57
	.dc.b	$8D,$F4,$CF,$AD,$C2,$F1,$9F,$3F,$FE,$63,$9D,$65,$4F,$F3,$85,$6E
	.dc.b	$71,$8F,$F9,$87,$0F,$F4,$D5,$FE,$1D,$FF,$43,$B5,$FF,$4C,$35,$DD
	.dc.b	$FF,$A1,$DF,$F4,$E7,$E8,$5F,$E4,$AA,$E9,$08,$C3,$FE,$AB,$C3,$D1
	.dc.b	$F4,$1F,$E7,$99,$CE,$7F,$E9,$16,$DF,$4D,$BD,$3A,$2F,$84,$49,$FF
	.dc.b	$3B,$D7,$6C,$7F,$D3,$ED,$79,$47,$F9,$F3,$E9,$9F,$4D,$E9,$9F,$F5
	.dc.b	$D4,$62,$75,$39,$94,$E5,$FF,$54,$FB,$0F,$67,$4D,$BF,$F4,$4B,$67
	.dc.b	$73,$CF,$CC,$77,$3D,$9F,$F4,$7C,$61,$BE,$6F,$F9,$D2,$FF,$D6,$F3
	.dc.b	$97,$FD,$73,$EE,$3B,$BF,$D9,$9C,$5F,$EB,$9F,$87,$7C,$D6,$36,$FF
	.dc.b	$68,$46,$30,$FA,$B8,$71,$6E,$1B,$7F,$DB,$93,$B7,$FD,$C4,$1C,$61
	.dc.b	$FE,$C7,$AE,$53,$AF,$31,$E6,$F3,$84,$3C,$EC,$36,$3F,$0E,$F0,$6C
	.dc.b	$E8,$8F,$E3,$3C,$D7,$9C,$C7,$FD,$E5,$E5,$BC,$AF,$F7,$75,$6A,$73
	.dc.b	$1F,$EF,$1C,$7B,$60,$BE,$0F,$FB,$EE,$33,$18,$85,$F3,$78,$57,$6C
	.dc.b	$3C,$51,$83,$96,$3F,$EF,$FB,$E5,$87,$4B,$E1,$8F,$92,$C7,$6C,$39
	.dc.b	$DF,$0C,$39,$5F,$A5,$B5,$C3,$1F,$25,$79,$DB,$5B,$9E,$C7,$D2,$AE
	.dc.b	$2F,$7C,$33,$5C,$23,$D2,$70,$C7,$D0,$F7,$C7,$CB,$87,$AA,$3E,$01
	.dc.b	$D4,$E1,$87,$6B,$E9,$1D,$CD,$DF,$47,$81,$CA,$36,$7C,$F7,$E2,$9E
	.dc.b	$1C,$3F,$DE,$B1,$98,$F1,$1A,$3E,$BF,$9C,$F3,$B6,$D8,$D1,$A8,$C9
	.dc.b	$FF,$00,$1F,$A7,$E7,$54,$D2,$3B,$7E,$6F,$86,$F3,$9C,$7F,$42,$B9
	.dc.b	$9F,$1A,$75,$19,$8E,$A6,$67,$8B,$F3,$0E,$3B,$FE,$4E,$B8,$C6,$4E
	.dc.b	$31,$99,$E2,$FE,$66,$38,$ED,$F2,$AB,$8C,$E6,$78,$CE,$6B,$8B,$F1
	.dc.b	$B0,$F6,$C6,$27,$E7,$46,$30,$F5,$36,$3C,$8D,$BE,$CD,$BE,$93,$EF
	.dc.b	$AE,$2D,$26,$9F,$09,$63,$F0,$23,$F9,$C5,$F7,$75,$19,$AF,$51,$FD
	.dc.b	$4A,$BE,$09,$FA,$E8,$D3,$FB,$0F,$39,$E7,$7F,$3F,$6D,$7E,$1A,$51
	.dc.b	$87,$F1,$2D,$7F,$20,$E6,$FE,$CF,$87,$FD,$D7,$DA,$FB,$71,$EB,$9D
	.dc.b	$5F,$F0,$0A,$D3,$FD,$A3,$17,$9F,$C2,$3E,$C3,$71,$78,$6F,$25,$EB
	.dc.b	$66,$7E,$3D,$5C,$39,$E1,$FE,$4C,$8F,$EB,$91,$89,$FF,$26,$46,$0E
	.dc.b	$D9,$BD,$6F,$F6,$FC,$61,$79,$BD,$FC,$F7,$A3,$4F,$EE,$FD,$F7,$3F
	.dc.b	$F0,$1C,$E9,$F9,$8D,$CB,$F6,$6D,$7F,$74,$D7,$F8,$0B,$BC,$BE,$6F
	.dc.b	$96,$FA,$3E,$D5,$FE,$0D,$8C,$5F,$83,$D0,$F3,$1F,$E0,$AF,$4A,$C6
	.dc.b	$4F,$1C,$7F,$7F,$C4,$D2,$47,$F8,$9F,$8F,$FC,$61,$F0,$1C,$B1,$F0
	.dc.b	$3F,$EC,$FF,$4F,$FD,$8D,$A7,$FA,$4C,$FF,$48,$CF,$FA,$0C,$FF,$43
	.dc.b	$94,$FF,$43,$F1,$7F,$A7,$E3,$6D,$C3,$FD,$0F,$1F,$EB,$7E,$0E,$D9
	.dc.b	$7F,$A6,$B9,$7F,$D2,$58,$7F,$A6,$1B,$87,$FA,$F2,$3F,$D2,$8D,$7C
	.dc.b	$0D,$3B,$3E,$46,$BE,$4A,$FF,$57,$37,$0F,$F4,$D3,$B8,$CC,$F3,$69
	.dc.b	$E6,$9F,$F5,$4E,$3F,$11,$FF,$57,$4E,$DE,$75,$D4,$FF,$52,$57,$17
	.dc.b	$FD,$76,$62,$3F,$DF,$55,$CE,$3F,$D6,$C6,$B1,$F8,$D3,$AC,$7E,$35
	.dc.b	$6E,$BE,$0D,$F5,$BF,$9E,$FA,$2F,$81,$E8,$BE,$24,$70,$4A,$F0,$3F
	.dc.b	$ED,$B6,$9F,$42,$7B,$47,$DA,$35,$5E,$07,$E0,$E1,$F0,$EB,$DC,$7F
	.dc.b	$8C,$1F,$B5,$8F,$DC,$BF,$DE,$1E,$79,$D3,$8A,$E6,$F8,$23,$EB,$46
	.dc.b	$0F,$6B,$E6,$9E,$91,$FC,$46,$7A,$5F,$FB,$A2,$73,$5F,$EF,$D8,$F4
	.dc.b	$3E,$67,$58,$FC,$DC,$3F,$12,$7A,$E3,$E9,$CF,$9B,$0E,$C7,$8D,$C1
	.dc.b	$D5,$F8,$CF,$C9,$7C,$F1,$E0,$39,$9F,$D0,$30,$FE,$36,$48,$FF,$74
	.dc.b	$B6,$7F,$DD,$0D,$7F,$DD,$67,$FB,$B2,$7F,$B7,$E1,$FC,$AF,$D1,$FF
	.dc.b	$67,$D7,$F2,$F8,$C5,$FF,$B1,$63,$1C,$DF,$EE,$36,$D7,$09,$FF,$56
	.dc.b	$57,$0C,$3F,$D8,$35,$A9,$E4,$6A,$F9,$36,$F9,$57,$F9,$76,$3F,$C8
	.dc.b	$B7,$CF,$0F,$F6,$26,$3F,$EC,$B9,$FF,$64,$92,$BF,$D9,$2D,$7F,$DA
	.dc.b	$53,$C3,$1F,$14,$79,$27,$C0,$79,$1F,$01,$FD,$3B,$1F,$2B,$8C,$3C
	.dc.b	$53,$D7,$0F,$14,$F8,$08,$78,$23,$57,$E0,$C3,$C5,$1E,$09,$F1,$1D
	.dc.b	$70,$EF,$3F,$51,$B3,$F0,$9B,$3E,$5C,$7E,$3D,$6D,$B5,$B6,$CE,$E3
	.dc.b	$DD,$3B,$69,$B2,$C6,$CF,$2C,$6D,$B1,$B6,$9B,$8F,$6B,$EB,$9F,$0E
	.dc.b	$3F,$22,$3E,$2E,$3F,$21,$F8,$C6,$31,$F8,$D8,$7C,$83,$E3,$61,$D0
	.dc.b	$F1,$15,$F1,$36,$7C,$51,$C5,$F1,$3E,$CB,$F1,$47,$80,$97,$E3,$25
	.dc.b	$F8,$A3,$C9,$5E,$23,$A5,$F8,$9A,$78,$8B,$1E,$23,$89,$E2,$3E,$3D
	.dc.b	$F8,$AF,$BC,$F5,$9D,$D7,$56,$C7,$56,$9D,$4F,$4D,$F9,$DB,$1C,$DA
	.dc.b	$78,$67,$E3,$5F,$86,$B8,$5F,$8A,$B5,$7E,$AA,$D3,$CE,$75,$7E,$5A
	.dc.b	$EA,$D8,$CC,$F9,$E7,$E7,$7A,$4C,$7F,$C6,$CD,$9F,$F0,$67,$0B,$D3
	.dc.b	$CD,$49,$FF,$29,$56,$EF,$FD,$32,$D7,$FD,$40,$D7,$FD,$4E,$EF,$1F
	.dc.b	$F1,$3F,$5F,$39,$FF,$13,$3F,$E4,$7F,$4B,$FC,$97,$7F,$ED,$2A,$FF
	.dc.b	$38,$46,$2B,$FC,$59,$E9,$EF,$59,$7F,$E6,$5A,$D5,$7F,$AC,$5A,$FF
	.dc.b	$AE,$1A,$FF,$B7,$9E,$97,$FE,$66,$FC,$B9,$E7,$F4,$AF,$E2,$3C,$A9
	.dc.b	$CE,$1E,$53,$FD,$87,$E9,$3D,$2F,$F4,$35,$E1,$AF,$FB,$B2,$35,$3F
	.dc.b	$E7,$B6,$C7,$FB,$0B,$D0,$7A,$5E,$28,$C3,$5F,$F6,$64,$3D,$2F,$F3
	.dc.b	$8C,$6B,$3F,$F4,$37,$AA,$F5,$BF,$D9,$4D,$9C,$C6,$0F,$F6,$77,$A4
	.dc.b	$F0,$7F,$DB,$BC,$1F,$E9,$FF,$4D,$EA,$63,$7F,$F7,$0C,$63,$1F,$84
	.dc.b	$06,$1C,$DC,$98,$C3,$B9,$1F,$F4,$E4,$6D,$FF,$71,$FA,$2F,$4B,$C1
	.dc.b	$1D,$61,$EA,$62,$11,$87,$BB,$C5,$B3,$B8,$C4,$7F,$4C,$3F,$DF,$F9
	.dc.b	$FF,$AF,$7D,$3F,$F7,$FF,$A8,$8F,$67,$A2,$F4,$7F,$E0,$0E,$1C,$A3
	.dc.b	$18,$79,$71,$F8,$4A,$32,$B1,$EB,$87,$4C,$39,$DF,$0B,$DE,$1D,$67
	.dc.b	$9D,$F0,$C3,$A4,$70,$79,$1E,$12,$9D,$F1,$F5,$3C,$0F,$AD,$28,$CD
	.dc.b	$F8,$30,$F1,$5F,$5B,$E0,$73,$C7,$AE,$1C,$5E,$F8,$F9,$E7,$C4,$70
	.dc.b	$9F,$46,$1D,$63,$78,$76,$7C,$78,$F8,$B1,$F7,$B6,$39,$D7,$17,$A4
	.dc.b	$65,$F2,$1F,$40,$E7,$8F,$FB,$9F,$0F,$34,$64,$D9,$B5,$EA,$3D,$4E
	.dc.b	$0D,$9E,$13,$A9,$C9,$FE,$F3,$8F,$D0,$7A,$75,$34,$8E,$DF,$AB,$E2
	.dc.b	$BD,$47,$1F,$D8,$AE,$67,$D0,$9D,$46,$63,$A9,$99,$E2,$FD,$63,$8E
	.dc.b	$9F,$90,$AE,$31,$93,$8C,$66,$78,$BF,$96,$8E,$3A,$FE,$6A,$B8,$CE
	.dc.b	$67,$8C,$E6,$B8,$BF,$37,$0F,$6C,$62,$7E,$A4,$63,$0F,$53,$63,$C8
	.dc.b	$DB,$EC,$DB,$E9,$3E,$FA,$E2,$D2,$69,$F0,$96,$3F,$02,$3F,$98,$DF
	.dc.b	$77,$51,$9A,$F5,$1F,$C5,$AB,$E0,$9F,$C6,$E3,$4F,$F1,$DF,$4D,$E1
	.dc.b	$FD,$0B,$FF,$80,$3E,$A5,$C3,$F8,$0D,$69,$FE,$D7,$19,$BE,$57,$7C
	.dc.b	$17,$90,$AD,$53,$0E,$4F,$F5,$29,$D6,$1F,$CB,$32,$7A,$7F,$D3,$AB
	.dc.b	$E0,$B6,$33,$3F,$A1,$9C,$B6,$33,$1F,$E6,$EC,$DE,$9B,$CB,$F1,$D6
	.dc.b	$A7,$E0,$4E,$6F,$C5,$87,$B6,$78,$DF,$08,$C5,$6F,$0F,$E6,$1E,$83
	.dc.b	$D3,$F4,$37,$1F,$33,$67,$F7,$0D,$7F,$7C,$E7,$1F,$D0,$ED,$F2,$BD
	.dc.b	$3F,$93,$39,$BF,$EA,$93,$97,$FA,$9C,$66,$D3,$97,$CB,$87,$D5,$D5
	.dc.b	$87,$4F,$47,$CE,$4C,$7C,$ED,$BF,$AD,$E9,$6E,$7F,$C5,$6D,$8C,$C7
	.dc.b	$8A,$7B,$4F,$FB,$96,$7F,$33,$B5,$DF,$FD,$B1,$7F,$EF,$63,$FD,$A7
	.dc.b	$E6,$7F,$B8,$B6,$FF,$68,$47,$FB,$65,$FF,$70,$E3,$FE,$F1,$C3,$FD
	.dc.b	$B7,$1A,$7F,$DC,$D3,$FB,$DC,$35,$3F,$ED,$A8,$C1,$FE,$E4,$9D,$9F
	.dc.b	$A3,$9D,$C7,$8E,$3F,$DF,$66,$8F,$4B,$C6,$7A,$47,$18,$C1,$C5,$D9
	.dc.b	$E6,$7C,$2F,$3C,$3F,$11,$3E,$93,$9C,$70,$71,$3C,$E7,$E3,$CF,$FB
	.dc.b	$DC,$E8,$78,$4E,$95,$EA,$75,$5F,$22,$BC,$B1,$B8,$EA,$EA,$3C,$B1
	.dc.b	$8C,$3E,$94,$60,$F1,$57,$D9,$23,$D2,$BD,$0E,$8C,$5F,$B9,$C4,$7F
	.dc.b	$85,$63,$F5,$0F,$F5,$D6,$BF,$18,$C4,$FE,$FC,$FF,$44,$92,$7F,$D0
	.dc.b	$CD,$8F,$F4,$59,$FE,$98,$AF,$F4,$5B,$83,$FD,$13,$3F,$E8,$7A,$FF
	.dc.b	$44,$C7,$FA,$75,$B3,$C2,$FF,$D1,$D7,$D5,$B7,$E0,$25,$F5,$BE,$3E
	.dc.b	$8F,$FB,$A3,$0E,$55,$DF,$0E,$4D,$38,$E1,$D3,$0E,$53,$A7,$C8,$DA
	.dc.b	$E7,$18,$AE,$27,$37,$8B,$87,$46,$09,$5D,$C9,$1C,$5B,$1C,$1B,$3E
	.dc.b	$02,$5F,$B4,$9E,$87,$FB,$C1,$B7,$FE,$E1,$6D,$FF,$B4,$9B,$7F,$EC
	.dc.b	$46,$CF,$FB,$05,$B3,$F6,$5A,$FD,$83,$C9,$1E,$72,$38,$25,$F9,$AF
	.dc.b	$C8,$49,$F2,$36,$7C,$8D,$BF,$23,$67,$C8,$D9,$F2,$35,$F2,$57,$79
	.dc.b	$EA,$48,$EA,$DA,$EA,$D3,$A9,$63,$AB,$5E,$B7,$F0,$B1,$FC,$6E,$1E
	.dc.b	$16,$D6,$DB,$5E,$02,$5F,$81,$B3,$E0,$69,$E0,$29,$E0,$AE,$D8,$FE
	.dc.b	$47,$0F,$13,$63,$35,$F8,$FC,$7E,$15,$F2,$C7,$C7,$8F,$84,$95,$E1
	.dc.b	$6C,$F8,$5A,$F8,$63,$D9,$87,$76,$9D,$F1,$F1,$B7,$1F,$1D,$7E,$7A
	.dc.b	$3F,$0B,$8F,$93,$0F,$E3,$97,$FE,$F7,$9F,$F7,$CD,$FF,$83,$7D,$0F
	.dc.b	$F7,$E6,$1F,$EB,$F6,$A6,$9C,$5F,$ED,$BB,$7F,$86,$B0,$FF,$1B,$4E
	.dc.b	$78,$3F,$C4,$3C,$DF,$E2,$08,$C5,$6F,$1F,$09,$FE,$CE,$AD,$BF,$EC
	.dc.b	$CF,$3F,$FD,$F7,$89,$A4,$D3,$FC,$48,$FF,$BF,$A7,$FC,$07,$E7,$FF
	.dc.b	$BE,$AB,$FD,$85,$39,$7F,$C0,$71,$9C,$FF,$C9,$93,$9C,$13,$9B,$FF
	.dc.b	$5F,$CE,$63,$17,$FE,$E3,$F3,$3F,$DF,$7E,$5F,$F8,$C5,$B1,$FE,$FF
	.dc.b	$F3,$3F,$DF,$2F,$FB,$F3,$CD,$FF,$7F,$3F,$EC,$AF,$37,$FD,$F0,$FF
	.dc.b	$A6,$5A,$FF,$A7,$DA,$EF,$CD,$FF,$4A,$53,$D4,$A4,$9F,$F2,$14,$65
	.dc.b	$FF,$78,$F9,$9F,$EF,$37,$FD,$4D,$E6,$7F,$BC,$DF,$F5,$63,$5F,$F6
	.dc.b	$EB,$FE,$F7,$E0,$FF,$38,$37,$7C,$D7,$FB,$D7,$CD,$7C,$F7,$FB,$C7
	.dc.b	$CC,$FF,$79,$79,$D3,$2F,$F7,$A9,$0F,$F7,$A7,$9D,$FE,$F6,$CF,$FC
	.dc.b	$F8,$DC,$BF,$D8,$B1,$FE,$D2,$F3,$BF,$D8,$53,$FE,$D1,$A3,$7C,$DF
	.dc.b	$F6,$8F,$95,$FE,$DB,$F2,$FF,$D9,$47,$FB,$D2,$33,$5F,$EC,$E9,$D1
	.dc.b	$D9,$FF,$67,$E1,$FE,$D8,$61,$82,$57,$49,$E3,$1F,$ED,$23,$2E,$1F
	.dc.b	$2E,$7F,$EE,$17,$FD,$B3,$3E,$3C,$7D,$73,$9C,$7D,$51,$FE,$DF,$C7
	.dc.b	$FD,$B7,$7F,$ED,$98,$FF,$7E,$B7,$1F,$F7,$04,$62,$07,$FC,$00,$F3
	.dc.b	$BF,$DB,$8F,$FB,$85,$AE,$85,$7F,$B7,$DF,$F7,$0C,$69,$FF,$5D,$9F
	.dc.b	$ED,$B2,$1F,$E8,$DF,$33,$FD,$BF,$97,$FB,$37,$4F,$F5,$BB,$67,$FD
	.dc.b	$7E,$7F,$AE,$5F,$F5,$FB,$74,$FF,$5F,$C6,$CC,$95,$C9,$FE,$B8,$8E
	.dc.b	$71,$97,$CA,$D7,$D9,$5F,$EC,$B7,$33,$C6,$7A,$B4,$EB,$1F,$EC,$1F
	.dc.b	$37,$FD,$B4,$F2,$3C,$C6,$6F,$FD,$65,$1F,$EB,$88,$FF,$64,$9F,$ED
	.dc.b	$27,$FD,$52,$48,$FF,$67,$47,$FB,$A8,$E4,$33,$FF,$6D,$7B,$8F,$7B
	.dc.b	$FD,$86,$DA,$FE,$45,$E0,$3E,$0B,$35,$27,$FD,$7D,$5B,$C3,$EE,$1F
	.dc.b	$82,$70,$E1,$AF,$E0,$7C,$CF,$F5,$22,$7F,$D4,$DE,$0F,$FB,$5B,$9F
	.dc.b	$FD,$5B,$E6,$4C,$3F,$D5,$6F,$FA,$9E,$31,$7F,$EA,$A9,$FF,$5D,$CF
	.dc.b	$FA,$F8,$8F,$FB,$4F,$C0,$FF,$6D,$CF,$C1,$8F,$C9,$46,$35,$47,$E6
	.dc.b	$BE,$2D,$F6,$AF,$AD,$FF,$73,$DE,$EB,$8F,$A7,$83,$FD,$E1,$7C,$2B
	.dc.b	$13,$FA,$4C,$3A,$E3,$FA,$6C,$3C,$6F,$C8,$6B,$D2,$7D,$18,$7F,$BB
	.dc.b	$FF,$43,$79,$FF,$EF,$86,$CF,$F5,$7F,$1D,$F1,$19,$A9,$3E,$13,$F6
	.dc.b	$15,$C3,$0F,$84,$7F,$52,$AD,$B8,$11,$B7,$EE,$B5,$E7,$5E,$53,$CD
	.dc.b	$BF,$94,$F7,$E7,$FE,$E8,$C7,$F0,$CF,$C2,$AE,$B3,$BC,$7C,$D1,$FC
	.dc.b	$2B,$04,$F2,$9F,$ED,$7C,$7B,$AF,$35,$6A,$7F,$23,$1C,$6B,$C7,$18
	.dc.b	$8E,$ED,$97,$B5,$F2,$CE,$D3,$FD,$D3,$E9,$31,$5E,$48,$F8,$73,$87
	.dc.b	$E4,$F1,$FF,$76,$AC,$65,$ED,$C3,$AE,$1D,$F0,$FB,$44,$7F,$29,$E9
	.dc.b	$FE,$30,$F0,$AF,$B5,$FF,$03,$64,$E1,$56,$EB,$C9,$18,$7E,$49,$C7
	.dc.b	$CE,$D5,$7D,$B9,$D3,$6B,$BD,$F6,$9C,$E5,$F5,$5B,$87,$E9,$AB,$78
	.dc.b	$74,$AD,$45,$9F,$E8,$27,$84,$F2,$B9,$C3,$F4,$0F,$28,$C9,$23,$C8
	.dc.b	$FA,$B0,$F2,$4F,$19,$F8,$4F,$17,$64,$AF,$1D,$71,$BE,$18,$7A,$AB
	.dc.b	$C3,$39,$3C,$15,$C2,$39,$47,$AB,$0F,$0B,$A3,$E1,$1C,$63,$67,$42
	.dc.b	$BF,$05,$C9,$E8,$79,$3F,$C3,$A3,$AB,$AA,$F3,$4F,$1A,$F2,$11,$F0
	.dc.b	$CF,$EE,$1D,$18,$8D,$47,$8D,$FE,$1F,$1A,$8C,$3B,$23,$E3,$BF,$8F
	.dc.b	$87,$D7,$9E,$13,$E9,$3C,$D1,$F4,$6B,$BC,$65,$F0,$CF,$0B,$FA,$13
	.dc.b	$98,$EE,$66,$BB,$BF,$2E,$B9,$D7,$03,$E1,$E1,$F1,$A3,$35,$E1,$AF
	.dc.b	$4D,$F5,$3D,$A7,$AD,$F9,$EE,$4D,$3E,$13,$37,$DB,$0F,$1C,$FD,$AC
	.dc.b	$3C,$15,$DA,$B4,$F0,$BF,$8D,$3A,$AF,$09,$D9,$C3,$E1,$AE,$2F,$CA
	.dc.b	$AF,$C1,$9E,$37,$07,$DD,$9E,$33,$F0,$EB,$E0,$4F,$68,$C5,$7C,$93
	.dc.b	$CD,$3E,$4B,$F5,$E1,$F5,$30,$E7,$39,$AF,$51,$E2,$8C,$B6,$B8,$BF
	.dc.b	$62,$33,$87,$C5,$9C,$BF,$42,$7D,$57,$F5,$9F,$24,$79,$EF,$C0,$FB
	.dc.b	$A7,$C9,$5C,$A3,$E4,$37,$0F,$8F,$3C,$E7,$C1,$7E,$86,$9C,$DF,$35
	.dc.b	$71,$BF,$8F,$3F,$45,$F0,$D7,$33,$C3,$5C,$DB,$3C,$CF,$66,$1F,$02
	.dc.b	$32,$71,$9E,$E7,$12,$CF,$13,$D6,$E5,$E7,$7E,$58,$F8,$B1,$E5,$7E
	.dc.b	$81,$B7,$89,$CF,$0F,$87,$1F,$78,$DC,$67,$0F,$7C,$FE,$12,$3E,$2D
	.dc.b	$F9,$DF,$0D,$75,$9E,$71,$F5,$63,$9C,$EE,$3C,$2D,$9D,$C7,$9C,$EB
	.dc.b	$3E,$36,$9C,$CA,$6E,$7E,$4C,$6D,$B1,$B6,$BB,$8F,$97,$1E,$1A,$F6
	.dc.b	$9F,$90,$8F,$C7,$9C,$CF,$E7,$17,$F3,$6B,$F5,$4D,$AF,$E4,$78,$7D
	.dc.b	$BA,$EA,$71,$AF,$8A,$7C,$B7,$EA,$E1,$F2,$EB,$CA,$D8,$E3,$3E,$96
	.dc.b	$BE,$93,$9C,$FA,$63,$9E,$1E,$9B,$FD,$5D,$7F,$90,$63,$FC,$48,$FF
	.dc.b	$3E,$31,$3F,$E0,$63,$FB,$27,$5F,$F1,$02,$3F,$B9,$6C,$7F,$19,$6D
	.dc.b	$FF,$01,$6C,$FE,$ED,$B3,$F9,$E6,$BF,$C6,$5B,$87,$F3,$26,$DF,$F8
	.dc.b	$21,$B5,$E4,$3C,$44,$AF,$13,$6F,$C4,$D3,$C4,$58,$F1,$35,$F1,$36
	.dc.b	$FC,$4D,$BF,$13,$4F,$14,$F8,$09,$5E,$06,$C7,$81,$FF,$2F,$7E,$8B
	.dc.b	$FE,$63,$9F,$F2,$A3,$67,$FD,$B7,$B3,$85,$F7,$3F,$E4,$6C,$17,$B7
	.dc.b	$FD,$C2,$D7,$3C,$7F,$E4,$CB,$E5,$5C,$6B,$47,$F9,$AE,$B8,$D7,$E0
	.dc.b	$9B,$87,$F9,$66,$75,$1E,$0E,$2E,$47,$5C,$5E,$36,$2F,$FC,$B1,$7F
	.dc.b	$E6,$5B,$F7,$DE,$EB,$51,$8A,$ED,$3E,$18,$ED,$18,$3F,$CD,$5F,$80
	.dc.b	$BE,$FF,$E1,$36,$3F,$C7,$1F,$1D,$89,$91,$FE,$69,$9D,$57,$F5,$DA
	.dc.b	$CB,$FE,$8F,$8C,$BF,$EC,$A6,$A2,$38,$6D,$79,$FE,$15,$F0,$D4,$98
	.dc.b	$7F,$A0,$63,$2F,$FA,$0F,$E0,$9F,$07,$EF,$6C,$66,$30,$7C,$3F,$59
	.dc.b	$EB,$7F,$9C,$F2,$FF,$47,$C6,$39,$BF,$D1,$CB,$D3,$75,$62,$F8,$5F
	.dc.b	$FB,$9B,$89,$EF,$7B,$63,$FD,$17,$C8,$F4,$71,$D3,$FE,$A9,$6E,$DB
	.dc.b	$C3,$8D,$FD,$EC,$32,$BA,$7F,$31,$78,$6C,$6C,$87,$FA,$A2,$B9,$B6
	.dc.b	$BF,$D5,$0F,$E7,$9B,$FA,$3B,$9B,$A6,$6F,$BC,$A7,$8A,$D5,$B7,$FD
	.dc.b	$54,$D7,$DF,$1C,$1C,$D7,$1F,$4F,$E1,$47,$FA,$C5,$FC,$EB,$7A,$B5
	.dc.b	$A4,$8D,$35,$FF,$6D,$7A,$4F,$4D,$C1,$F8,$BF,$45,$E8,$FD,$1A,$D4
	.dc.b	$E6,$3E,$A6,$5E,$2C,$17,$C1,$FC,$A4,$66,$31,$0B,$E6,$F0,$AE,$D8
	.dc.b	$78,$A3,$07,$2C,$7F,$05,$7C,$B0,$E9,$7C,$31,$F2,$5F,$E8,$F1,$EB
	.dc.b	$7B,$C3,$8D,$FA,$1B,$5B,$C7,$C1,$5E,$6B,$E3,$7B,$AE,$A3,$1F,$0B
	.dc.b	$C7,$1C,$D7,$08,$F4,$1C,$27,$6F,$C1,$C3,$BD,$FA,$63,$DE,$75,$C7
	.dc.b	$CA,$D3,$AC,$F0,$7C,$07,$0A,$EA,$76,$BF,$14,$77,$C3,$FD,$FD,$1F
	.dc.b	$42,$3F,$58,$64,$A5,$32,$BD,$37,$C2,$72,$6C,$F0,$9D,$4E,$4F,$D2
	.dc.b	$C3,$D4,$A4,$7F,$52,$DD,$3F,$0B,$F0,$2F,$19,$CF,$F4,$2F,$F8,$F5
	.dc.b	$FD,$2A,$D4,$E6,$7B,$1F,$2A,$32,$FC,$C3,$89,$9A,$EA,$E4,$E3,$7E
	.dc.b	$28,$CB,$F9,$88,$E2,$E6,$BC,$17,$E3,$AD,$EF,$F2,$A3,$13,$AB,$F0
	.dc.b	$36,$F9,$47,$86,$76,$D3,$65,$8D,$9E,$37,$CF,$5F,$C9,$70,$F7,$CF
	.dc.b	$4D,$3D,$B4,$9F,$87,$1F,$DA,$6F,$BB,$A8,$CE,$1C,$F1,$FE,$F5,$7F
	.dc.b	$03,$1C,$4F,$F4,$FB,$EB,$E9,$BC,$CF,$EC,$8D,$9F,$C5,$DA,$30,$FE
	.dc.b	$39,$AF,$E4,$9C,$D7,$D4,$E9,$7A,$AF,$8D,$FF,$06,$56,$B0,$FF,$08
	.dc.b	$CE,$5F,$21,$FD,$1F,$D5,$7A,$BF,$E1,$36,$C6,$B2,$FE,$85,$5F,$E0
	.dc.b	$1C,$BE,$3D,$6A,$7F,$91,$D7,$3B,$FF,$00,$4E,$67,$F5,$2D,$75,$3F
	.dc.b	$02,$39,$57,$AE,$FF,$AD,$CF,$86,$95,$A8,$F5,$F6,$7F,$8A,$7C,$27
	.dc.b	$A5,$4F,$CB,$E3,$F9,$8F,$51,$EA,$BD,$5D,$63,$D7,$0E,$2F,$D2,$6E
	.dc.b	$7F,$BE,$6B,$FC,$09,$AF,$F6,$F7,$78,$7D,$3F,$B9,$7D,$E7,$6A,$F8
	.dc.b	$A7,$F6,$B8,$CD,$FB,$7D,$5F,$F1,$F1,$FE,$41,$E3,$FF,$23,$7A,$FF
	.dc.b	$E4,$58,$CF,$A5,$FE,$2B,$7F,$C7,$24,$70,$7F,$8E,$7C,$0E,$0D,$BE
	.dc.b	$DE,$97,$33,$FC,$7D,$D1,$63,$27,$E4,$AF,$FC,$7D,$9F,$F2,$06,$BF
	.dc.b	$E0,$D6,$BF,$E2,$B7,$F0,$91,$FE,$5E,$F1,$DE,$BF,$5A,$FF,$09,$BE
	.dc.b	$BF,$5E,$57,$E3,$75,$FF,$32,$4E,$2B,$E3,$8C,$17,$F0,$49,$7E,$76
	.dc.b	$CF,$FA,$DA,$3E,$64,$7F,$AF,$9F,$F6,$99,$FE,$B5,$8F,$F6,$9B,$78
	.dc.b	$7F,$DA,$67,$FA,$CE,$3F,$D5,$F1,$9F,$5B,$94,$7F,$BE,$E3,$FD,$8E
	.dc.b	$FF,$AD,$CF,$E6,$3E,$BF,$FB,$0D,$E2,$64,$EE,$5A,$EF,$5F,$EB,$9F
	.dc.b	$3B,$FD,$99,$7F,$ED,$03,$FD,$D9,$1E,$0E,$4F,$F7,$84,$78,$2F,$FD
	.dc.b	$88,$DB,$F4,$36,$BF,$48,$71,$7F,$DA,$86,$4F,$87,$3E,$38,$D3,$EB
	.dc.b	$8D,$3F,$EF,$09,$F8,$8F,$A8,$FF,$6A,$15,$C9,$EC,$8F,$D6,$4F,$89
	.dc.b	$FF,$6B,$98,$3C,$33,$CA,$78,$3F,$37,$0F,$94,$F5,$AF,$8D,$5F,$10
	.dc.b	$E7,$1B,$C3,$E4,$BF,$49,$B1,$EA,$C3,$FD,$DF,$5C,$AB,$FD,$E8,$7E
	.dc.b	$1C,$B1,$FE,$FC,$AE,$37,$E2,$C3,$CB,$3F,$4A,$31,$7F,$22,$FE,$9C
	.dc.b	$63,$0F,$2D,$6F,$0E,$64,$8F,$55,$78,$67,$FD,$F8,$D8,$FC,$81,$C8
	.dc.b	$FB,$35,$D6,$34,$7D,$98,$D3,$F6,$63,$53,$EB,$9E,$ED,$9F,$7C,$7E
	.dc.b	$3D,$FD,$7E,$0C,$3E,$C4,$7F,$21,$7F,$D7,$51,$99,$FE,$4D,$18,$23
	.dc.b	$82,$4F,$17,$FD,$D7,$5F,$EE,$92,$3F,$EE,$96,$C7,$FB,$E0,$FF,$76
	.dc.b	$95,$CB,$FE,$FA,$32,$57,$35,$FE,$FC,$8C,$B6,$33,$5F,$EF,$E9,$CB
	.dc.b	$63,$35,$E6,$3C,$84,$9F,$13,$5F,$03,$6B,$B3,$6B,$A3,$4E,$93,$C8
	.dc.b	$87,$22,$CF,$26,$CF,$26,$CF,$27,$81,$27,$83,$67,$83,$67,$84,$E8
	.dc.b	$93,$A6,$C6,$A3,$04,$70,$49,$FE,$B6,$AF,$FC,$47,$1E,$63,$2E,$1F
	.dc.b	$11,$8F,$03,$F1,$B8,$7F,$93,$2B,$94,$61,$FF,$7B,$18,$3F,$84,$4F
	.dc.b	$16,$CF,$1A,$ED,$87,$AA,$B4,$7D,$EA,$F4,$56,$EF,$B4,$FF,$38,$6C
	.dc.b	$F7,$8E,$07,$29,$FC,$2B,$63,$EE,$34,$F8,$11,$D6,$3C,$A7,$AA,$FC
	.dc.b	$B8,$7C,$BA,$E0,$7C,$4A,$E0,$F8,$1A,$78,$0B,$5E,$06,$C7,$82,$3B
	.dc.b	$5F,$AA,$FB,$C7,$C0,$32,$79,$27,$E0,$34,$F8,$15,$DD,$DC,$78,$DB
	.dc.b	$19,$8E,$26,$4A,$E6,$39,$9E,$73,$89,$63,$6D,$36,$87,$9A,$E2,$FF
	.dc.b	$1D,$CE,$67,$31,$85,$BB,$81,$EB,$E6,$77,$7F,$E4,$BC,$3A,$3F,$E7
	.dc.b	$06,$C7,$FA,$F1,$AF,$FB,$01,$AF,$85,$FF,$1B,$57,$FA,$FE,$7A,$61
	.dc.b	$E3,$C3,$C5,$E4,$78,$6F,$FC,$4B,$87,$B5,$B3,$E7,$F2,$F5,$CB,$31
	.dc.b	$F2,$3F,$E9,$4B,$F0,$35,$E4,$D7,$E3,$BE,$1C,$9F,$80,$F3,$9F,$F2
	.dc.b	$8E,$1F,$16,$3E,$27,$B0,$F6,$29,$F2,$F1,$FD,$E7,$9F,$87,$FC,$E5
	.dc.b	$3B,$7F,$D2,$ED,$C3,$FD,$C6,$D7,$FD,$CE,$D7,$EC,$3B,$E0,$FD,$C5
	.dc.b	$F5,$9F,$E0,$EF,$D7,$F5,$9E,$B7,$B5,$B8,$7F,$A9,$9B,$5F,$EF,$26
	.dc.b	$BF,$EF,$6F,$59,$EA,$BA,$3F,$D5,$B1,$86,$FA,$CE,$AC,$D6,$EF,$FD
	.dc.b	$FB,$EF,$BD,$FF,$F5,$AC,$23,$0F,$FA,$E1,$AF,$08,$DF,$33,$C9,$FC
	.dc.b	$C7,$07,$FA,$DF,$95,$EB,$63,$83,$FC,$D0,$DC,$7F,$0D,$75,$FD,$1B
	.dc.b	$D0,$8F,$E6,$5A,$FF,$9A,$30,$F0,$3F,$9F,$8E,$9C,$DF,$A5,$F1,$FA
	.dc.b	$7B,$0E,$7C,$E6,$F6,$29,$A8,$98,$FF,$B5,$AB,$83,$63,$E2,$C6,$23
	.dc.b	$F6,$C7,$EE,$FD,$8F,$7B,$FE,$BB,$5D,$9E,$D6,$C6,$75,$C1,$1C,$1F
	.dc.b	$EC,$3F,$61,$E4,$38,$7F,$84,$7A,$EF,$57,$F8,$35,$6A,$73,$1F,$C2
	.dc.b	$71,$F1,$E0,$BE,$0F,$9F,$1F,$E1,$D1,$88,$61,$CD,$E1,$7D,$B1,$F1
	.dc.b	$46,$0E,$47,$80,$A6,$4E,$6F,$0A,$ED,$5D,$5F,$06,$1E,$66,$E1,$E4
	.dc.b	$C3,$C1,$8F,$5C,$79,$B6,$F8,$4F,$BE,$B8,$57,$F2,$DC,$3B,$4F,$28
	.dc.b	$F2,$57,$07,$CF,$87,$93,$1F,$35,$6E,$3C,$73,$DF,$0E,$AF,$88,$F7
	.dc.b	$61,$D9,$E7,$5E,$37,$D1,$19,$78,$1F,$88,$7E,$73,$6F,$E6,$B7,$1F
	.dc.b	$C6,$BE,$BF,$05,$EC,$62,$3E,$BC,$61,$DD,$A3,$11,$DD,$AF,$4F,$63
	.dc.b	$F7,$E9,$A4,$7F,$81,$BF,$47,$D8,$79,$0E,$8F,$A9,$5F,$D6,$EB,$53
	.dc.b	$98,$FC,$11,$93,$F0,$95,$CA,$B8,$C7,$84,$CC,$F1,$DB,$F2,$91,$C5
	.dc.b	$FC,$B5,$71,$8E,$A7,$18,$CC,$F1,$D7,$DD,$3D,$2F,$BE,$1E,$66,$C6
	.dc.b	$4E,$71,$96,$99,$BF,$1B,$DE,$BF,$BB,$61,$E3,$9C,$D7,$16,$D7,$1A
	.dc.b	$DE,$1F,$DD,$EF,$FC,$91,$3E,$D9,$CE,$1F,$E0,$5D,$3E,$54,$AB,$FC
	.dc.b	$9C,$70,$7C,$47,$F8,$2D,$FF,$29,$7A,$EF,$3B,$FB,$B7,$AC,$F6,$3F
	.dc.b	$C0,$D3,$9C,$7B,$1F,$E1,$3E,$A5,$6E,$33,$8F,$F8,$5E,$B8,$3F,$DB
	.dc.b	$A7,$4F,$F1,$59,$FF,$0C,$4E,$B2,$FC,$15,$F4,$C3,$FC,$2D,$EB,$BD
	.dc.b	$8F,$EF,$8D,$C3,$F3,$8D,$C7,$F4,$2D,$7F,$97,$B9,$C3,$AE,$5F,$D4
	.dc.b	$6B,$CF,$4A,$D4,$7B,$67,$53,$E7,$3F,$C6,$53,$A7,$FC,$5B,$1A,$AF
	.dc.b	$A3,$39,$C7,$E2,$F6,$FF,$8B,$FC,$47,$B1,$FE,$C8,$26,$5F,$B5,$8C
	.dc.b	$B6,$3F,$C5,$1E,$AB,$D6,$E5,$18,$6C,$FD,$0C,$15,$B8,$F6,$DF,$4D
	.dc.b	$3F,$C3,$F3,$97,$D8,$3D,$4F,$F1,$DB,$FE,$4F,$6B,$9F,$43,$FC,$59
	.dc.b	$18,$3F,$C7,$FD,$7F,$E4,$B6,$E1,$F3,$1B,$5D,$9B,$B7,$E5,$97,$A1
	.dc.b	$D0,$95,$FC,$5E,$32,$D8,$FF,$1D,$7A,$BD,$DB,$1D,$DB,$7D,$DB,$7D
	.dc.b	$FD,$5F,$49,$2B,$B3,$6B,$FD,$6B,$E5,$FA,$E3,$FC,$81,$EA,$7F,$88
	.dc.b	$EB,$D9,$18,$9D,$47,$E8,$0F,$F6,$03,$FE,$AA,$AF,$F1,$77,$AB,$B9
	.dc.b	$FF,$47,$46,$1F,$F4,$75,$F5,$8F,$A7,$BA,$B9,$36,$BB,$56,$F1,$FD
	.dc.b	$FC,$E6,$F8,$DF,$FA,$7F,$0E,$2D,$C3,$C3,$5B,$BE,$33,$E5,$BE,$35
	.dc.b	$A8,$C4,$EA,$30,$78,$AF,$AB,$EA,$6D,$72,$3E,$5C,$70,$71,$3F,$ED
	.dc.b	$59,$DB,$D6,$36,$6A,$7F,$D3,$97,$E7,$6D,$F9,$1A,$75,$AE,$8D,$33
	.dc.b	$5D,$23,$B1,$23,$2F,$08,$F0,$1F,$EA,$FA,$F0,$36,$BC,$0D,$AE,$CF
	.dc.b	$F5,$DC,$3E,$A4,$74,$24,$6C,$F1,$38,$7F,$04,$78,$1F,$C4,$9D,$1C
	.dc.b	$DF,$94,$F4,$3C,$CF,$F2,$5D,$FE,$05,$B7,$FE,$A1,$8C,$47,$FB,$2E
	.dc.b	$FF,$DC,$2F,$FB,$BA,$3D,$B7,$FE,$CE,$8F,$F7,$83,$FE,$D4,$7F,$DD
	.dc.b	$47,$FB,$50,$AE,$E3,$8C,$F4,$BF,$F7,$8D,$72,$38,$16,$B8,$5F,$FB
	.dc.b	$FE,$77,$7C,$BB,$3F,$DE,$C4,$AF,$D2,$B6,$FF,$DD,$91,$C8,$FF,$7D
	.dc.b	$3D,$A7,$94,$70,$24,$70,$C3,$CD,$1E,$D8,$F2,$46,$1E,$58,$79,$67
	.dc.b	$55,$D2,$79,$5F,$8F,$0F,$F7,$E6,$1C,$6F,$C9,$5F,$37,$0E,$4F,$48
	.dc.b	$F3,$9A,$AF,$83,$3A,$8C,$1E,$07,$15,$D4,$97,$E1,$9F,$B8,$DB,$FA
	.dc.b	$B3,$A9,$F8,$37,$E6,$AD,$57,$C0,$6E,$1E,$D9,$D5,$70,$3A,$96,$BA
	.dc.b	$D7,$0C,$3C,$AD,$9F,$6E,$1E,$1A,$F8,$F1,$C5,$F0,$37,$0F,$03,$DA
	.dc.b	$36,$4C,$3C,$6D,$8D,$46,$08,$E0,$91,$E6,$7B,$61,$E6,$69,$E6,$C3
	.dc.b	$D1,$3F,$5E,$30,$69,$C1,$2B,$EB,$C6,$A3,$C8,$77,$29,$EE,$2D,$FB
	.dc.b	$9D,$4F,$A1,$B3,$E8,$78,$46,$5B,$19,$9E,$51,$83,$89,$0C,$11,$D9
	.dc.b	$2F,$E1,$47,$91,$B5,$DE,$75,$5E,$D8,$C4,$EA,$3C,$4D,$A0,$F2,$BF
	.dc.b	$C0,$E4,$E0,$FF,$29,$C6,$8D,$38,$78,$C2,$3F,$D0,$57,$FE,$42,$F2
	.dc.b	$7F,$D2,$98,$7F,$8D,$AB,$4F,$F8,$7E,$33,$7F,$EF,$DB,$E3,$3F,$E5
	.dc.b	$59,$CD,$FF,$99,$2F,$8D,$FF,$8E,$F8,$D5,$AB,$F3,$57,$F8,$42,$BF
	.dc.b	$D5,$6D,$7F,$DF,$2D,$77,$BF,$AD,$F8,$4F,$89,$5F,$6F,$53,$11,$AB
	.dc.b	$F5,$46,$2F,$C1,$8F,$C1,$8C,$61,$CF,$0E,$57,$FE,$58,$F5,$1E,$A7
	.dc.b	$85,$B1,$FE,$BF,$6E,$5F,$EC,$46,$BF,$4C,$3A,$C7,$FD,$65,$2E,$0C
	.dc.b	$3F,$66,$6D,$7F,$5F,$BF,$5C,$62,$78,$46,$0F,$F1,$49,$F0,$3D,$27
	.dc.b	$A2,$E1,$FF,$6A,$46,$1B,$E9,$39,$33,$5B,$CB,$FD,$BB,$E9,$3D,$2F
	.dc.b	$F6,$E4,$23,$0F,$FB,$7D,$AF,$08,$DD,$BD,$FA,$49,$FF,$73,$46,$5F
	.dc.b	$F7,$47,$1B,$BF,$BB,$63,$31,$82,$3F,$97,$7F,$41,$C3,$F9,$16,$EF
	.dc.b	$F9,$5F,$4E,$D7,$E8,$A6,$BF,$AB,$6D,$F7,$8F,$CC,$73,$2B,$7E,$8B
	.dc.b	$D2,$F8,$6D,$AF,$F8,$00,$D7,$F0,$01,$C3,$F8,$37,$59,$7F,$B2,$9B
	.dc.b	$97,$EE,$5B,$3F,$9C,$C1,$ED,$FA,$3D,$27,$A5,$7D,$39,$7F,$EA,$E9
	.dc.b	$CB,$67,$81,$1F,$F5,$9F,$A4,$F1,$1A,$7F,$0E,$F4,$1E,$87,$F0,$BA
	.dc.b	$D4,$E6,$3F,$87,$63,$DB,$05,$F0,$7F,$8B,$C6,$63,$10,$BE,$6F,$0A
	.dc.b	$ED,$87,$8A,$30,$72,$C7,$F8,$E5,$F2,$C3,$A5,$F0,$C7,$C9,$7F,$C9
	.dc.b	$B1,$EB,$7B,$C3,$8D,$FA,$1B,$5B,$C7,$C1,$5E,$6B,$E3,$7B,$AE,$A3
	.dc.b	$1F,$0B,$C7,$1C,$D7,$08,$F4,$1C,$27,$6F,$C1,$C3,$BD,$FA,$63,$DE
	.dc.b	$75,$C7,$CA,$D3,$AC,$F0,$7C,$07,$0A,$EA,$76,$BF,$14,$77,$C3,$F2
	.dc.b	$6D,$3F,$2B,$1F,$49,$FE,$2C,$D7,$F0,$1E,$93,$D3,$E3,$28,$C4,$F2
	.dc.b	$6B,$C9,$AF,$E8,$BD,$3F,$E1,$29,$A4,$7F,$86,$BF,$3B,$D3,$78,$EE
	.dc.b	$3F,$A3,$5F,$D8,$AB,$53,$98,$F9,$26,$4F,$97,$5C,$AB,$8C,$78,$4C
	.dc.b	$CF,$1D,$FF,$2B,$1C,$5F,$CB,$D7,$18,$EA,$71,$8C,$CF,$1D,$FD,$D3
	.dc.b	$D2,$FB,$E1,$E6,$6C,$64,$E7,$19,$69,$9B,$F1,$BD,$EB,$FB,$E6,$1E
	.dc.b	$39,$CD,$71,$6D,$71,$AD,$E1,$FD,$8A,$FD,$5A,$7F,$80,$F0,$F0,$3F
	.dc.b	$26,$D3,$AC,$78,$9F,$E1,$09,$E1,$3E,$97,$87,$A4,$E1,$FE,$DE,$DA
	.dc.b	$E0,$7F,$86,$3C,$47,$A0,$F4,$77,$72,$BF,$2B,$7C,$71,$FE,$52,$D7
	.dc.b	$FB,$53,$5F,$F0,$F7,$A3,$FE,$0E,$CB,$D4,$DA,$F0,$62,$F8,$CF,$F0
	.dc.b	$1C,$E6,$FF,$C5,$6D,$4E,$73,$F3,$CF,$0F,$2E,$61,$FA,$AA,$FF,$0E
	.dc.b	$DF,$F7,$D6,$BE,$49,$FF,$0B,$79,$FD,$4E,$9D,$4F,$67,$FC,$2B,$19
	.dc.b	$9A,$7F,$84,$67,$58,$7C,$AB,$FF,$20,$D6,$E3,$F7,$6F,$82,$33,$5E
	.dc.b	$03,$FD,$A9,$E7,$7F,$88,$BC,$8F,$F6,$4F,$91,$62,$61,$D6,$FF,$DC
	.dc.b	$2F,$F9,$38,$FF,$72,$BF,$EF,$63,$FD,$D2,$7F,$B8,$1F,$E1,$75,$FE
	.dc.b	$E0,$8F,$F7,$31,$FE,$51,$FA,$DF,$F0,$44,$F5,$6C,$7F,$6C,$F8,$AF
	.dc.b	$F1,$AC,$63,$87,$F2,$4D,$AE,$5E,$47,$C3,$AD,$B2,$3F,$8D,$CF,$4B
	.dc.b	$FF,$17,$35,$E3,$1A,$F8,$4E,$E7,$CD,$9B,$1E,$58,$B3,$FE,$00,$9B
	.dc.b	$87,$92,$7F,$CA,$10,$69,$2B,$FC,$85,$E6,$FF,$8C,$DF,$B0,$7F,$8F
	.dc.b	$7C,$5C,$61,$FD,$BA,$3F,$C7,$D3,$8C,$3F,$C8,$7E,$7F,$F8,$36,$7F
	.dc.b	$CA,$13,$97,$67,$48,$D5,$73,$3F,$D0,$BE,$7F,$F9,$21,$FF,$27,$D7
	.dc.b	$17,$F9,$54,$F0,$BF,$E0,$15,$B2,$57,$F9,$AE,$FD,$07,$88,$97,$DC
	.dc.b	$EA,$F8,$F0,$ED,$18,$35,$5F,$E5,$7C,$3C,$15,$EA,$9C,$DF,$96,$3B
	.dc.b	$C6,$1E,$F8,$79,$A7,$C9,$5C,$AB,$DF,$1E,$0B,$ED,$85,$8E,$95,$BA
	.dc.b	$F2,$5F,$3C,$3A,$9F,$9F,$9D,$61,$F0,$A3,$35,$F2,$30,$E9,$18,$3C
	.dc.b	$F7,$CE,$7D,$04,$7E,$DC,$F0,$8C,$4F,$78,$C4,$F1,$37,$1E,$09,$F3
	.dc.b	$BF,$1E,$33,$1C,$8F,$83,$1E,$3B,$F4,$E1,$D2,$3B,$1C,$F0,$F6,$57
	.dc.b	$E1,$E3,$D2,$75,$23,$D0,$D5,$74,$8E,$55,$F4,$B0,$ED,$5B,$C3,$CA
	.dc.b	$7A,$EB,$9C,$7A,$8F,$94,$F2,$3D,$13,$CA,$35,$1E,$47,$A3,$CB,$0F
	.dc.b	$BA,$F1,$8F,$84,$ED,$EC,$48,$DB,$F1,$CC,$5F,$A6,$79,$5F,$9E,$7C
	.dc.b	$A6,$EB,$E3,$C7,$B2,$7C,$55,$F8,$CB,$F4,$57,$6A,$E5,$7D,$AF,$F4
	.dc.b	$87,$63,$E6,$9E,$79,$EA,$E3,$0F,$9D,$3E,$1C,$3C,$CD,$9F,$8E,$73
	.dc.b	$7E,$29,$E8,$BF,$0D,$F4,$8C,$4F,$A0,$E2,$F3,$3E,$4D,$F8,$2B,$8D
	.dc.b	$F8,$1B,$5C,$1B,$5C,$30,$F8,$35,$BA,$ED,$39,$7C,$CF,$DC,$BE,$D1
	.dc.b	$AB,$F9,$93,$EA,$3E,$41,$F9,$0B,$EB,$3F,$2A,$3D,$13,$F5,$E3,$E5
	.dc.b	$5F,$E4,$9A,$7C,$52,$CF,$07,$F4,$6D,$3E,$4B,$A9,$F6,$9E,$48,$F1
	.dc.b	$BD,$EF,$D3,$1E,$7C,$3F,$01,$5E,$B3,$ED,$61,$E8,$9F,$54,$7C,$29
	.dc.b	$DC,$71,$C3,$F1,$53,$97,$E5,$46,$63,$E8,$1D,$5F,$3D,$F8,$9F,$2C
	.dc.b	$7B,$30,$F4,$BE,$23,$A1,$F2,$5C,$C7,$2C,$3C,$D1,$EF,$C3,$F1,$13
	.dc.b	$F5,$E7,$35,$D4,$F1,$46,$5E,$AF,$B8,$F9,$86,$4B,$3F,$01,$F3,$9E
	.dc.b	$18,$E3,$5E,$1C,$3E,$D6,$1E,$DB,$FC,$D5,$7D,$53,$E1,$16,$FE,$31
	.dc.b	$E1,$BF,$6B,$E1,$6C,$F8,$4F,$69,$6F,$BE,$1E,$EA,$FD,$39,$D2,$BF
	.dc.b	$01,$19,$BF,$04,$FA,$9B,$7E,$6C,$3D,$18,$7D,$CA,$F8,$2F,$DC,$7F
	.dc.b	$3A,$0F,$EC,$45,$3F,$98,$14,$FE,$D4,$59,$FE,$D5,$1F,$D7,$09,$5F
	.dc.b	$D0,$5B,$3F,$BC,$6E,$1F,$99,$6C,$FF,$0F,$6C,$FE,$31,$AF,$ED,$DB
	.dc.b	$5E,$09,$EC,$49,$EC,$DA,$EC,$D3,$B1,$63,$B3,$5E,$CD,$9E,$D1,$D0
	.dc.b	$93,$D1,$B3,$D1,$A7,$42,$9D,$0B,$3D,$27,$91,$27,$E8,$7E,$94,$F6
	.dc.b	$3F,$D9,$53,$FE,$A3,$8C,$4F,$F9,$C7,$A1,$CA,$FB,$1D,$CF,$2C,$2E
	.dc.b	$5F,$77,$0E,$2F,$F8,$9D,$B8,$FF,$B5,$F1,$E7,$87,$0A,$56,$B0,$E8
	.dc.b	$7F,$BE,$1E,$71,$F8,$56,$BA,$3F,$10,$72,$E0,$E9,$1F,$E6,$B8,$5F
	.dc.b	$AB,$E3,$F3,$52,$FF,$D0,$57,$BA,$D4,$62,$75,$7D,$AB,$FC,$DC,$FB
	.dc.b	$E3,$FD,$94,$FC,$C6,$BF,$9F,$F8,$27,$97,$88,$46,$2F,$D3,$18,$9D
	.dc.b	$4E,$63,$DB,$EF,$3D,$FE,$4D,$BE,$2D,$9D,$CF,$7F,$59,$EB,$3A,$7B
	.dc.b	$C6,$1B,$EA,$3B,$33,$5B,$C3,$E6,$FB,$AF,$80,$FD,$C7,$27,$FB,$A7
	.dc.b	$E1,$5F,$21,$8D,$FF,$79,$18,$C3,$FD,$5D,$87,$16,$E1,$B7,$F8,$24
	.dc.b	$ED,$FF,$37,$37,$0F,$F7,$B5,$71,$9F,$F3,$93,$5F,$E1,$0D,$7F,$86
	.dc.b	$35,$FE,$20,$D7,$F1,$FE,$DF,$87,$97,$F3,$BE,$D5,$EA,$3E,$5A,$D9
	.dc.b	$F8,$6A,$D4,$48,$C1,$0F,$C3,$47,$DC,$78,$98,$3E,$FA,$FA,$BF,$01
	.dc.b	$23,$FD,$A9,$19,$6C,$65,$FF,$50,$5F,$ED,$23,$CD,$3B,$9F,$F4,$B7
	.dc.b	$AA,$F5,$FD,$4B,$CD,$A6,$22,$4F,$FB,$8E,$78,$36,$3E,$17,$AA,$F5
	.dc.b	$DB,$7F,$3B,$F5,$5E,$AF,$F3,$BA,$D4,$E6,$3F,$9F,$63,$D3,$15,$F0
	.dc.b	$7F,$A2,$46,$63,$10,$BE,$6F,$0A,$ED,$87,$8A,$30,$72,$C7,$FA,$65
	.dc.b	$F2,$C3,$A5,$F0,$C7,$C9,$7F,$D4,$F1,$EB,$7B,$C3,$8D,$FA,$1B,$5B
	.dc.b	$C7,$C1,$5E,$6B,$E3,$7B,$AE,$A3,$1F,$0B,$C7,$1C,$D7,$08,$F4,$1C
	.dc.b	$27,$6F,$C1,$C3,$BD,$FA,$63,$DE,$75,$C7,$CA,$D3,$AC,$F0,$7C,$07
	.dc.b	$0A,$EA,$76,$BF,$14,$77,$C3,$F8,$9C,$7D,$08,$FE,$E8,$D7,$C4,$63
	.dc.b	$D6,$7A,$FC,$1B,$1F,$CD,$25,$18,$8E,$13,$FA,$55,$EB,$52,$3F,$A9
	.dc.b	$6E,$9F,$84,$F5,$9E,$BB,$A3,$E7,$DF,$F8,$36,$FE,$8D,$6A,$73,$3D
	.dc.b	$8F,$93,$19,$7E,$59,$C4,$CD,$75,$72,$71,$BF,$14,$65,$FC,$BC,$71
	.dc.b	$73,$5E,$0B,$F1,$D6,$F7,$FE,$AF,$18,$9D,$5F,$81,$B7,$CA,$3C,$33
	.dc.b	$B6,$9B,$2C,$6C,$F1,$BE,$7A,$FF,$17,$61,$EF,$9E,$9A,$7B,$69,$3F
	.dc.b	$0E,$3F,$AC,$DF,$77,$51,$9C,$39,$E3,$FD,$E6,$FE,$06,$38,$9F,$F1
	.dc.b	$F5,$F5,$F5,$BF,$88,$47,$F5,$E6,$D7,$F7,$86,$BF,$E4,$16,$BF,$E4
	.dc.b	$57,$73,$FE,$09,$F5,$9F,$29,$FE,$09,$9C,$DF,$F9,$37,$99,$5A,$8F
	.dc.b	$19,$FE,$4F,$F5,$9E,$A7,$F8,$45,$B1,$FE,$1F,$97,$AF,$9A,$93,$FD
	.dc.b	$DA,$B7,$9F,$F8,$41,$AF,$F9,$65,$AF,$79,$FF,$12,$E1,$E4,$F5,$79
	.dc.b	$CF,$F3,$AC,$51,$FE,$23,$F5,$BA,$EB,$F5,$5B,$97,$B7,$B5,$EB,$EF
	.dc.b	$39,$3F,$E0,$AA,$F0,$E5,$FE,$6F,$6A,$69,$D1,$FF,$12,$DF,$F3,$D9
	.dc.b	$F8,$F1,$C6,$BD,$33,$98,$FE,$83,$3F,$E7,$9F,$1B,$D2,$F0,$8F,$F0
	.dc.b	$76,$5F,$E4,$48,$7A,$7F,$E0,$D7,$55,$C2,$B9,$9F,$E8,$57,$0F,$FA
	.dc.b	$0D,$AF,$7A,$FF,$15,$46,$23,$84,$77,$AF,$CB,$46,$2B,$D6,$78,$B8
	.dc.b	$FF,$CA,$1E,$B7,$F9,$16,$4F,$F9,$07,$17,$AF,$AF,$3E,$61,$FE,$36
	.dc.b	$AF,$F2,$0D,$F0,$9F,$F1,$E4,$60,$E5,$F1,$9C,$27,$FB,$BE,$08,$F0
	.dc.b	$FB,$13,$0F,$F3,$74,$F8,$9F,$95,$69,$D5,$FF,$97,$A3,$13,$E3,$9F
	.dc.b	$09,$1E,$7F,$47,$C3,$8B,$FC,$E1,$82,$3F,$D8,$33,$FE,$D2,$3F,$DB
	.dc.b	$D8,$7F,$B0,$5F,$F7,$53,$FE,$BA,$7F,$DD,$47,$FB,$8B,$D7,$FF,$65
	.dc.b	$3F,$7D,$FF,$61,$7B,$3F,$EC,$7D,$FF,$D8,$73,$FE,$FA,$AE,$87,$FB
	.dc.b	$84,$E8,$7F,$B6,$5E,$9E,$CF,$FB,$1E,$FF,$DA,$ED,$3C,$F3,$C3,$2F
	.dc.b	$F7,$45,$78,$67,$FD,$A1,$1F,$EE,$07,$84,$64,$F1,$4F,$FB,$C8,$FF
	.dc.b	$73,$3E,$03,$DA,$F9,$8C,$96,$BD,$EF,$E5,$8F,$F7,$15,$FB,$9C,$46
	.dc.b	$AF,$F9,$63,$F1,$2B,$E7,$4F,$FB,$B8,$F1,$46,$E3,$FD,$FB,$5C,$EB
	.dc.b	$FD,$E3,$1C,$0F,$F7,$7D,$FB,$9F,$45,$FC,$93,$CD,$1F,$01,$F1,$BB
	.dc.b	$73,$1E,$E7,$C3,$1B,$9F,$3B,$B8,$CC,$EE,$FE,$9B,$67,$E9,$D7,$CF
	.dc.b	$8C,$1F,$3E,$BC,$2E,$1A,$48,$FB,$2D,$8F,$C7,$C7,$38,$F6,$1E,$22
	.dc.b	$D7,$89,$B5,$E2,$AF,$87,$5D,$0F,$5B,$F8,$0A,$E1,$19,$8F,$9F,$87
	.dc.b	$E5,$C8,$FC,$FC,$3E,$93,$E4,$3A,$9F,$A6,$C3,$C9,$5C,$DF,$CE,$C7
	.dc.b	$D6,$3F,$D7,$76,$C3,$FD,$80,$D8,$CC,$60,$8E,$09,$3F,$ED,$C9,$FF
	.dc.b	$7D,$47,$FC,$01,$AF,$F7,$D4,$FF,$C0,$16,$BF,$F0,$05,$B3,$C2,$35
	.dc.b	$3C,$48,$E0,$87,$17,$2D,$33,$1C,$DC,$11,$C1,$2B,$F9,$91,$FE,$FD
	.dc.b	$71,$3F,$CC,$AF,$E1,$57,$E0,$4F,$F1,$D3,$8B,$F4,$DF,$2A,$FC,$64
	.dc.b	$7F,$8C,$DF,$E9,$71,$D4,$D5,$73,$8C,$9C,$03,$DA,$38,$47,$8D,$B1
	.dc.b	$FD,$4A,$38,$BC,$EB,$C9,$1F,$12,$FD,$73,$B9,$CB,$F1,$AB,$C7,$7E
	.dc.b	$AB,$EC,$78,$E7,$C4,$D3,$BC,$6E,$3A,$4F,$92,$3E,$05,$76,$9F,$A1
	.dc.b	$3D,$AB,$C8,$D8,$F1,$47,$81,$F5,$47,$38,$F8,$67,$80,$E2,$6E,$BC
	.dc.b	$47,$82,$38,$4F,$A0,$E7,$3E,$9A,$F2,$5F,$C4,$BE,$13,$9B,$E2,$F7
	.dc.b	$BF,$29,$EE,$74,$47,$A9,$27,$C7,$1E,$8B,$EF,$5E,$FA,$F8,$73,$E9
	.dc.b	$3D,$6F,$53,$93,$8A,$F2,$C6,$0F,$1B,$83,$BB,$EF,$6C,$7A,$63,$B9
	.dc.b	$D6,$33,$5E,$49,$F1,$4F,$47,$D1,$5F,$0A,$75,$3C,$67,$C8,$68,$F8
	.dc.b	$55,$D2,$B8,$12,$32,$70,$9F,$33,$4F,$35,$FD,$37,$A4,$7B,$CF,$4D
	.dc.b	$7B,$E3,$F7,$FE,$B3,$D9,$FF,$50,$4F,$FA,$16,$31,$3F,$E1,$6F,$67
	.dc.b	$2F,$EB,$2D,$3A,$7F,$B3,$35,$D1,$AB,$FF,$48,$3C,$63,$4D,$8C,$BF
	.dc.b	$B8,$8E,$97,$FE,$B3,$EA,$71,$7E,$6A,$BB,$DF,$F9,$BA,$31,$5E,$97
	.dc.b	$CF,$3C,$DF,$F2,$DF,$B3,$88,$F4,$C5,$C3,$FC,$7A,$47,$C7,$1A,$C7
	.dc.b	$D6,$F3,$7F,$CC,$F8,$F9,$3D,$3E,$55,$E3,$7B,$7A,$EF,$5F,$9B,$6F
	.dc.b	$9E,$08,$C3,$D9,$AF,$CA,$F5,$DE,$B3,$B3,$8C,$61,$BE,$BB,$2C,$D6
	.dc.b	$EF,$D3,$EF,$FC,$A7,$C2,$D7,$3A,$79,$14,$65,$FB,$7E,$AB,$D7,$FC
	.dc.b	$17,$17,$FB,$2F,$D7,$7B,$18,$E0,$FC,$34,$63,$0F,$F4,$36,$1C,$5B
	.dc.b	$87,$78,$C4,$7C,$E6,$BE,$08,$D1,$D5,$FE,$7B,$1C,$30,$FF,$6E,$36
	.dc.b	$FC,$3F,$1F,$BA,$FC,$BF,$AF,$C7,$91,$DC,$F6,$70,$DA,$FF,$4E,$73
	.dc.b	$AF,$93,$63,$EC,$B5,$FC,$2B,$5F,$D0,$B5,$FD,$04,$72,$23,$F8,$AE
	.dc.b	$07,$89,$E7,$97,$B3,$E4,$25,$7F,$AB,$E3,$2D,$8F,$B1,$18,$8F,$CB
	.dc.b	$B5,$FD,$9B,$5F,$DD,$7A,$CF,$63,$E3,$2F,$66,$98,$89,$3F,$EB,$F9
	.dc.b	$E0,$D8,$F9,$DE,$C3,$D9,$71,$FF,$72,$F5,$DE,$AF,$F0,$9A,$D4,$E6
	.dc.b	$3C,$77,$FC,$36,$B7,$87,$F7,$AC,$7B,$A8,$CA,$C3,$9B,$9B,$ED,$8F
	.dc.b	$8A,$30,$73,$C3,$B4,$F8,$23,$C4,$7F,$82,$2B,$9D,$F6,$6C,$EE,$79
	.dc.b	$8C,$3C,$B5,$BC,$3C,$0D,$C7,$AD,$6F,$1F,$0B,$E3,$33,$8F,$67,$C5
	.dc.b	$7C,$B1,$F0,$61,$F0,$31,$EF,$1E,$EC,$78,$D6,$EF,$A1,$E6,$AD,$BD
	.dc.b	$27,$67,$92,$33,$1E,$53,$27,$BB,$0F,$9A,$FF,$3C,$6B,$E1,$3F,$92
	.dc.b	$B8,$7E,$57,$B2,$F6,$7F,$8F,$CA,$31,$3C,$1A,$F5,$6B,$FC,$C3,$D9
	.dc.b	$FE,$28,$9A,$47,$F8,$C3,$F3,$BD,$9F,$D5,$F4,$7D,$09,$FF,$21,$CE
	.dc.b	$78,$93,$99,$E8,$7D,$CA,$E2,$FD,$E3,$89,$99,$E2,$E4,$E3,$B7,$E5
	.dc.b	$27,$8B,$F9,$68,$E2,$E6,$78,$B9,$8E,$3B,$7F,$6D,$AE,$2F,$C1,$8F
	.dc.b	$05,$F9,$F0,$F5,$36,$32,$78,$E3,$2D,$33,$7D,$E3,$2D,$8C,$D7,$87
	.dc.b	$0F,$EC,$D8,$7C,$2D,$7F,$BA,$E1,$D1,$F8,$36,$9D,$63,$C4,$FF,$38
	.dc.b	$CF,$09,$F2,$BC,$3D,$8F,$E1,$D3,$FD,$AB,$D6,$7B,$3F,$DD,$67,$37
	.dc.b	$FC,$2A,$BF,$CF,$7D,$4B,$DC,$76,$3F,$C2,$95,$C3,$0F,$F0,$44,$EA
	.dc.b	$FF,$DF,$53,$9A,$E7,$39,$AF,$5F,$B0,$F5,$BF,$C0,$2D,$AF,$EE,$ED
	.dc.b	$C3,$FC,$06,$D7,$FC,$3E,$E6,$BC,$3E,$B7,$F8,$3E,$31,$5E,$08,$C5
	.dc.b	$FF,$85,$B2,$F6,$46,$27,$E2,$3F,$E1,$7F,$5F,$B6,$3F,$E3,$16,$BF
	.dc.b	$E9,$E6,$A6,$9E,$05,$5F,$D8,$39,$BA,$61,$E1,$ED,$FE,$71,$8F,$CB
	.dc.b	$BE,$93,$9A,$DC,$FF,$88,$2B,$75,$F5,$4F,$F5,$7F,$90,$CB,$FC,$52
	.dc.b	$DA,$FF,$0F,$B6,$7F,$C5,$6D,$7F,$C8,$EE,$74,$FF,$50,$F1,$7F,$8D
	.dc.b	$E3,$15,$FE,$83,$8C,$5F,$F8,$FA,$1E,$8F,$E4,$B0,$F1,$CE,$63,$10
	.dc.b	$F5,$7F,$C9,$75,$FE,$32,$6C,$7F,$8F,$9A,$FF,$96,$9A,$FF,$98,$5D
	.dc.b	$E3,$FE,$55,$C9,$EB,$7F,$93,$67,$37,$F9,$89,$CB,$F3,$6F,$C5,$E0
	.dc.b	$7D,$1B,$E0,$D9,$D5,$BD,$6F,$F2,$9B,$63,$D1,$18,$23,$F5,$7D,$6F
	.dc.b	$14,$7D,$D6,$C6,$4E,$B1,$97,$FD,$AD,$83,$D4,$FF,$6E,$C7,$CC,$F5
	.dc.b	$BF,$D8,$73,$9F,$5B,$FD,$39,$18,$AF,$F4,$EE,$BF,$ED,$88,$C4,$EA
	.dc.b	$FA,$D6,$9F,$F5,$E4,$E9,$FF,$66,$36,$FF,$DE,$4D,$BF,$E1,$F5,$E4
	.dc.b	$D9,$39,$BF,$F4,$A4,$E6,$31,$63,$D1,$3F,$9D,$3D,$87,$93,$0F,$34
	.dc.b	$7E,$6C,$D9,$6B,$D0,$FF,$48,$BF,$0D,$71,$3D,$25,$8F,$4B,$70,$F4
	.dc.b	$DF,$E4,$F0,$F2,$D7,$69,$F9,$15,$B3,$D7,$5B,$71,$3E,$BC,$3D,$D1
	.dc.b	$88,$E4,$FC,$C3,$11,$A8,$F5,$4F,$C4,$9F,$83,$18,$7F,$29,$29,$FF
	.dc.b	$50,$B6,$3F,$D2,$ED,$9F,$F5,$1B,$5F,$F5,$FB,$FE,$D1,$7F,$DA,$A7
	.dc.b	$FB,$8C,$A7,$FB,$4A,$3F,$DA,$B5,$FE,$DF,$7F,$DB,$B1,$B7,$FD,$B7
	.dc.b	$5F,$EF,$FA,$E9,$18,$3F,$DD,$F3,$CC,$EB,$1A,$C3,$EB,$CE,$E7,$FD
	.dc.b	$E0,$75,$8C,$92,$BF,$DD,$D1,$FE,$E7,$AE,$51,$8B,$F3,$4E,$63,$C8
	.dc.b	$4B,$F1,$D7,$07,$D7,$3E,$57,$C5,$5F,$EF,$18,$F1,$57,$59,$FF,$7D
	.dc.b	$5F,$8D,$E3,$7C,$EB,$C1,$1D,$AB,$C1,$1E,$C3,$B5,$71,$76,$66,$B9
	.dc.b	$5F,$A8,$97,$E0,$BF,$15,$79,$23,$C3,$1D,$A7,$D1,$3D,$A3,$6D,$9D
	.dc.b	$9E,$52,$3E,$18,$E5,$1F,$80,$3C,$07,$08,$D3,$C2,$7B,$D7,$A2,$37
	.dc.b	$3E,$C9,$CD,$74,$8F,$C0,$B6,$7E,$79,$DE,$3D,$87,$17,$0F,$A4,$C1
	.dc.b	$C9,$F1,$1A,$AF,$14,$72,$AF,$92,$D9,$F2,$C6,$23,$A3,$67,$8C,$62
	.dc.b	$76,$47,$91,$0D,$16,$38,$92,$BD,$D7,$F7,$EF,$B4,$60,$FA,$B3,$BB
	.dc.b	$F5,$DF,$B4,$F9,$E4,$AF,$97,$1A,$3E,$4C,$6E,$32,$EE,$3E,$25,$77
	.dc.b	$7E,$D4,$E8,$8E,$8E,$97,$E7,$7E,$05,$7B,$89,$7E,$77,$B1,$E1,$8E
	.dc.b	$4E,$1F,$29,$FE,$00,$BF,$F3,$9D,$FF,$95,$3C,$67,$AB,$FE,$56,$BF
	.dc.b	$E8,$42,$33,$7F,$EF,$A9,$D4,$FF,$47,$9C,$BF,$EF,$C8,$FF,$1B,$6D
	.dc.b	$FD,$2E,$BF,$BD,$56,$AF,$FD,$07,$CC,$AD,$61,$FD,$2E,$B5,$7F,$E5
	.dc.b	$A9,$CD,$F9,$F0,$EB,$7E,$AA,$F3,$DF,$F8,$42,$3C,$97,$E6,$7D,$4D
	.dc.b	$8D,$46,$27,$FC,$75,$87,$8F,$D0,$E9,$3C,$A7,$36,$9C,$CF,$13,$FC
	.dc.b	$79,$5C,$6B,$DF,$5F,$0A,$B7,$E9,$3D,$17,$3F,$CE,$C7,$E7,$CE,$63
	.dc.b	$11,$FD,$CE,$1E,$97,$86,$33,$1B,$77,$E9,$7C,$B7,$88,$70,$EB,$1C
	.dc.b	$B9,$6D,$3C,$DA,$49,$E7,$19,$7A,$F9,$EF,$43,$9B,$63,$31,$82,$3F
	.dc.b	$E3,$B8,$D3,$FE,$09,$6F,$0F,$FA,$9B,$D1,$7A,$3F,$92,$8F,$F5,$4C
	.dc.b	$EB,$83,$F2,$AD,$BF,$F0,$C3,$55,$F2,$7F,$23,$E8,$7C,$0E,$33,$49
	.dc.b	$1F,$22,$32,$F0,$21,$F7,$B5,$77,$F9,$71,$FC,$8D,$DA,$FD,$23,$5F
	.dc.b	$D5,$39,$E4,$FD,$05,$66,$1E,$8F,$D4,$6D,$7D,$A6,$BF,$84,$6B,$F8
	.dc.b	$97,$3A,$7F,$B3,$9B,$87,$EE,$9B,$3F,$E9,$CC,$7F,$C8,$6F,$F9,$B7
	.dc.b	$CF,$7A,$1E,$75,$E8,$D3,$11,$27,$83,$5F,$F2,$17,$A0,$F1,$59,$FF
	.dc.b	$92,$FC,$F7,$9F,$FC,$32,$B5,$39,$8F,$E1,$F9,$78,$B0,$C1,$FE,$53
	.dc.b	$70,$AF,$9B,$FE,$57,$BE,$6E,$6B,$B6,$1E,$28,$C1,$CF,$1F,$E3,$B7
	.dc.b	$CA,$18,$76,$8C,$61,$D1,$B7,$CD,$E1,$3B,$C3,$D3,$87,$47,$C3,$8F
	.dc.b	$AB,$16,$3C,$E7,$71,$9B,$E4,$FA,$F1,$ED,$3C,$A3,$D2,$79,$B0,$EC
	.dc.b	$FA,$CE,$D7,$D5,$E5,$87,$59,$DC,$71,$37,$7D,$9D,$9D,$B0,$F1,$BB
	.dc.b	$3C,$37,$D5,$F5,$CF,$CD,$7F,$8D,$47,$D2,$6C,$65,$F5,$1A,$F4,$1E
	.dc.b	$8F,$F1,$6A,$46,$27,$F9,$0C,$62,$75,$39,$9F,$E3,$BE,$8F,$F1,$A4
	.dc.b	$D2,$3F,$C7,$9F,$A5,$E8,$FE,$E3,$87,$EA,$4F,$FA,$7E,$73,$C6,$9C
	.dc.b	$CF,$43,$E8,$D7,$17,$E9,$9C,$4C,$CF,$17,$27,$1D,$FF,$1D,$3C,$5F
	.dc.b	$C8,$47,$17,$33,$C5,$CC,$71,$DF,$FB,$9D,$71,$7F,$A2,$E1,$E8,$C3
	.dc.b	$CE,$DA,$F0,$B6,$F9,$B6,$B8,$CF,$A2,$B8,$B4,$E2,$5A,$F2,$C7,$FB
	.dc.b	$CE,$FE,$16,$FF,$E0,$2C,$3C,$0F,$C1,$B4,$EA,$BD,$38,$FF,$68,$9F
	.dc.b	$81,$8E,$27,$A7,$C8,$FF,$6C,$B3,$FD,$CB,$C1,$7C,$87,$F5,$5A,$D9
	.dc.b	$30,$FF,$69,$5F,$08,$FF,$6D,$65,$F4,$EF,$8F,$77,$5C,$7B,$76,$4B
	.dc.b	$D9,$FE,$E1,$A7,$AE,$F3,$F5,$9F,$6B,$B9,$75,$AD,$61,$F1,$67,$2F
	.dc.b	$D8,$7D,$51,$AA,$E5,$3E,$07,$FD,$E1,$E8,$7F,$54,$8C,$42,$30,$FF
	.dc.b	$83,$BC,$F7,$C4,$7F,$84,$A7,$38,$FD,$16,$DF,$F8,$E9,$B6,$EE,$61
	.dc.b	$CA,$BB,$D6,$AB,$C1,$DB,$FE,$29,$8F,$0C,$7F,$85,$DA,$F5,$8F,$F1
	.dc.b	$77,$9F,$FE,$3B,$AF,$4B,$67,$D1,$ED,$7F,$8A,$23,$19,$7E,$1D,$B5
	.dc.b	$FE,$50,$6D,$79,$F1,$5F,$1E,$F9,$7F,$E1,$EB,$FF,$1D,$92,$3F,$C8
	.dc.b	$1F,$6E,$F9,$4F,$F1,$BB,$6B,$FC,$7F,$F1,$2F,$73,$35,$27,$E8,$D6
	.dc.b	$F3,$FF,$1E,$DE,$E7,$35,$B9,$FF,$1B,$D6,$DF,$F1,$E3,$5F,$F3,$5B
	.dc.b	$50,$FD,$33,$FC,$7B,$5F,$80,$6C,$7F,$83,$FD,$0F,$F2,$7C,$63,$6F
	.dc.b	$D7,$36,$BF,$CE,$ED,$AF,$EF,$D5,$C3,$CB,$95,$FE,$44,$AE,$B5,$FE
	.dc.b	$51,$6B,$CA,$7F,$CD,$10,$F8,$FF,$09,$FE,$7C,$F9,$FB,$7E,$78,$B7
	.dc.b	$FE,$25,$9B,$8F,$9A,$79,$92,$BC,$4F,$6F,$77,$DA,$7F,$A2,$3B,$EC
	.dc.b	$FC,$38,$B8,$7F,$8C,$66,$DF,$C2,$9E,$55,$ED,$6D,$7F,$9A,$A3,$FA
	.dc.b	$69,$FA,$C2,$BF,$E1,$67,$FD,$2D,$EA,$BE,$47,$FC,$FB,$1A,$BF,$E3
	.dc.b	$8D,$3F,$D4,$A5,$AF,$E9,$15,$C6,$3B,$78,$BC,$88,$FE,$FC,$46,$E7
	.dc.b	$C4,$7F,$A9,$E7,$6E,$0F,$EA,$8F,$FA,$52,$78,$DE,$AF,$CB,$36,$7D
	.dc.b	$73,$6B,$A4,$79,$0F,$15,$F8,$6F,$8C,$66,$7C,$0F,$B4,$EE,$66,$37
	.dc.b	$1A,$BF,$25,$73,$AF,$8B,$1E,$19,$F4,$C7,$82,$F0,$FC,$98,$C9,$F0
	.dc.b	$9E,$EF,$57,$C0,$74,$39,$18,$7D,$A6,$CE,$55,$E3,$73,$7E,$82,$4F
	.dc.b	$C6,$39,$47,$8A,$32,$FD,$CA,$F3,$DF,$CB,$AF,$2D,$6A,$B8,$47,$95
	.dc.b	$F0,$D7,$08,$F3,$1C,$0D,$4F,$17,$E7,$5F,$9E,$B9,$9E,$88,$D9,$E7
	.dc.b	$7D,$D3,$C6,$FE,$8D,$F7,$9C,$D6,$EB,$4F,$8A,$FB,$3F,$44,$97,$C9
	.dc.b	$F8,$66,$0F,$9F,$5D,$0F,$21,$1F,$63,$83,$D2,$7A,$9C,$B8,$7A,$3F
	.dc.b	$19,$EF,$1F,$36,$7C,$AF,$AE,$BF,$C5,$93,$C8,$E2,$FD,$67,$11,$B7
	.dc.b	$E0,$1F,$22,$BC,$8F,$CE,$BD,$47,$A6,$FB,$CF,$88,$F7,$DF,$C8,$9F
	.dc.b	$81,$1D,$8D,$3E,$93,$E2,$1E,$DB,$F9,$4E,$23,$CE,$79,$1F,$1C,$7B
	.dc.b	$EF,$C7,$1F,$1E,$FC,$6F,$19,$E7,$7E,$33,$E0,$4E,$9C,$5F,$5A,$F1
	.dc.b	$3F,$7E,$FA,$BF,$04,$F4,$CF,$28,$CD,$F6,$9F,$91,$5C,$8F,$75,$76
	.dc.b	$7E,$45,$70,$9F,$8E,$70,$3D,$F1,$FA,$97,$B9,$F4,$CF,$94,$FB,$5E
	.dc.b	$2F,$C4,$7D,$07,$07,$BB,$5F,$8D,$1E,$17,$CB,$1E,$1A,$F9,$8E,$CF
	.dc.b	$81,$1D,$CF,$B7,$7F,$26,$38,$1E,$28,$F8,$17,$F1,$63,$95,$7C,$59
	.dc.b	$F3,$1C,$AF,$C5,$5F,$96,$BE,$93,$A8,$C4,$EA,$FE,$B1,$F1,$A7,$D6
	.dc.b	$F6,$3C,$46,$4F,$4C,$FD,$18,$D9,$EB,$7C,$CD,$8F,$9D,$5E,$28,$F5
	.dc.b	$C7,$82,$FF,$42,$FD,$2A,$FB,$8D,$3E,$1C,$FC,$C3,$D4,$F4,$BF,$54
	.dc.b	$7D,$D3,$C0,$7D,$62,$9E,$69,$F4,$4F,$A8,$EA,$FA,$0C,$D7,$CD,$9F
	.dc.b	$39,$CE,$33,$3F,$12,$7D,$37,$ED,$8F,$C7,$9E,$9A,$F3,$9E,$5A,$E2
	.dc.b	$FA,$CF,$CE,$1F,$E3,$23,$FA,$49,$4F,$EA,$65,$8F,$F0,$3B,$6B,$FB
	.dc.b	$97,$DF,$3D,$AF,$E8,$8D,$AF,$E7,$ED,$AF,$E1,$2D,$BF,$E3,$0D,$9F
	.dc.b	$D2,$36,$7F,$96,$35,$FE,$54,$D8,$FE,$E8,$DB,$F1,$DF,$88,$93,$E2
	.dc.b	$6D,$78,$9B,$7E,$26,$CF,$89,$B3,$E2,$6B,$E2,$6C,$78,$9B,$5E,$28
	.dc.b	$F0,$12,$7C,$0F,$FA,$73,$B9,$F4,$7F,$EE,$7C,$3F,$D9,$11,$89,$FF
	.dc.b	$52,$76,$BA,$3F,$B2,$4E,$70,$FF,$12,$36,$F8,$E1,$CF,$0F,$F3,$6E
	.dc.b	$1C,$75,$C6,$1F,$E7,$0C,$38,$57,$E0,$5E,$91,$F8,$16,$B9,$BF,$F6
	.dc.b	$E3,$67,$F0,$CF,$3F,$B7,$EA,$A9,$1E,$0C,$3D,$5A,$EA,$73,$7F,$0E
	.dc.b	$B0,$6F,$35,$62,$74,$78,$30,$ED,$5E,$47,$C5,$19,$BE,$B1,$E6,$FB
	.dc.b	$7F,$24,$FF,$B3,$1B,$1F,$E2,$6B,$F4,$C6,$34,$FF,$7D,$36,$BF,$4C
	.dc.b	$DA,$FF,$2F,$D7,$0B,$95,$E5,$AE,$B5,$FA,$B6,$BC,$A7,$D0,$BE,$21
	.dc.b	$F1,$38,$6D,$FB,$DB,$3F,$41,$AF,$ED,$FE,$1B,$FD,$68,$FD,$46,$B9
	.dc.b	$ED,$EF,$18,$8F,$E0,$1E,$D3,$DC,$7B,$9F,$36,$30,$DE,$C7,$93,$9A
	.dc.b	$DD,$FE,$0B,$57,$DF,$FF,$0B,$E1,$6D,$F9,$2E,$37,$CD,$E6,$30,$DC
	.dc.b	$3F,$29,$75,$FC,$53,$9E,$5F,$D1,$4D,$BE,$75,$FA,$BA,$E1,$8F,$8A
	.dc.b	$3F,$05,$1F,$87,$79,$98,$3F,$15,$FA,$73,$EA,$7A,$79,$D2,$7F,$DA
	.dc.b	$BA,$7A,$9F,$F6,$BD,$6D,$CE,$9F,$CA,$63,$11,$FE,$9A,$D3,$F7,$17
	.dc.b	$CE,$3F,$D3,$53,$AB,$E9,$08,$E2,$FF,$31,$6B,$D7,$0F,$54,$62,$3F
	.dc.b	$DD,$3F,$34,$F2,$FE,$47,$BD,$F1,$27,$FD,$50,$DC,$3B,$D6,$E7,$D2
	.dc.b	$72,$AE,$3B,$C8,$C8,$9E,$0F,$FA,$CD,$B5,$EA,$BE,$DA,$7F,$4B,$6D
	.dc.b	$75,$39,$D3,$E1,$FC,$FE,$2C,$9C,$C7,$E0,$67,$51,$E2,$A5,$6E,$3C
	.dc.b	$55,$F8,$7F,$6D,$EE,$5F,$76,$57,$AA,$72,$D9,$E0,$47,$DF,$EE,$3D
	.dc.b	$D6,$A9,$7B,$8F,$2B,$FB,$45,$EE,$B5,$39,$AE,$18,$F8,$1E,$37,$FD
	.dc.b	$BE,$31,$3B,$95,$F3,$76,$78,$5C,$46,$E7,$8D,$73,$C7,$FB,$D4,$F3
	.dc.b	$AD,$CF,$9B,$1F,$5E,$1D,$06,$1C,$9F,$3E,$1C,$AB,$86,$3E,$C6,$CF
	.dc.b	$1A,$E0,$F2,$AE,$0F,$9E,$FB,$E3,$E3,$6D,$F9,$E7,$A1,$E0,$29,$F0
	.dc.b	$A3,$6F,$3C,$33,$87,$80,$F0,$E3,$EE,$9D,$1E,$5C,$7C,$6F,$94,$EB
	.dc.b	$1B,$7B,$57,$F4,$B8,$FA,$11,$FE,$21,$6B,$F0,$0C,$78,$EF,$77,$83
	.dc.b	$63,$FB,$1C,$A3,$11,$C2,$7F,$89,$2F,$76,$91,$FE,$2C,$DD,$3F,$09
	.dc.b	$ED,$BD,$DF,$E2,$51,$F8,$7A,$FC,$A5,$7C,$8A,$D7,$27,$CA,$9C,$9F
	.dc.b	$7E,$32,$FE,$04,$CD,$76,$BE,$BB,$FF,$7C,$C3,$A4,$62,$BF,$C8,$91
	.dc.b	$8C,$3B,$47,$AC,$ED,$3B,$7F,$C9,$93,$B7,$C0,$6E,$FC,$15,$CA,$75
	.dc.b	$1E,$D9,$D5,$FF,$95,$70,$F4,$CE,$6B,$73,$E7,$AF,$0D,$F3,$69,$34
	.dc.b	$FE,$5F,$87,$C2,$7F,$C1,$17,$E5,$73,$8F,$F8,$6A,$75,$7D,$71,$F3
	.dc.b	$E1,$FC,$12,$3C,$9D,$3B,$3F,$C5,$BD,$EF,$75,$EE,$E6,$BF,$95,$E2
	.dc.b	$BC,$61,$CC,$FE,$DB,$7C,$2B,$73,$E3,$EF,$91,$FC,$E3,$17,$B8,$F7
	.dc.b	$73,$8F,$31,$97,$F8,$5D,$B8,$79,$AB,$86,$1E,$48,$E3,$B7,$C6,$75
	.dc.b	$86,$27,$FD,$0D,$EE,$7F,$4F,$7F,$A2,$B7,$0F,$EC,$1E,$E3,$DD,$FF
	.dc.b	$21,$CE,$71,$FE,$BE,$EE,$BF,$D1,$CD,$7A,$F8,$1F,$E2,$48,$C4,$7F
	.dc.b	$31,$79,$61,$FE,$5C,$C1,$F2,$DD,$AF,$FC,$9D,$18,$9F,$EB,$3E,$EF
	.dc.b	$F9,$A5,$BE,$3E,$BD,$EF,$F2,$BC,$E7,$2F,$90,$D7,$FD,$44,$D7,$FD
	.dc.b	$4C,$D7,$FA,$23,$53,$CF,$3F,$F0,$47,$8E,$C5,$F1,$1E,$0C,$7C,$EF
	.dc.b	$F9,$A3,$DD,$7B,$D4,$FF,$34,$7B,$FF,$E3,$CC,$7F,$C0,$F3,$AD,$7F
	.dc.b	$BA,$B7,$4F,$F3,$9B,$5F,$F5,$D3,$5F,$F5,$F3,$5F,$EB,$8F,$19,$F3
	.dc.b	$D7,$F8,$EA,$BA,$DF,$8B,$CC,$FF,$3E,$FB,$CF,$3E,$9F,$E7,$7F,$7E
	.dc.b	$DF,$F8,$22,$74,$FF,$9D,$DB,$9F,$FA,$21,$AF,$FB,$3D,$AF,$FB,$4D
	.dc.b	$DE,$5F,$2B,$57,$D2,$76,$AF,$F1,$74,$62,$FC,$1F,$01,$E6,$3F,$D1
	.dc.b	$5E,$65,$8C,$9E,$38,$FF,$38,$E2,$69,$23,$FD,$63,$4F,$A1,$E5,$87
	.dc.b	$81,$F8,$7F,$03,$2B,$7E,$35,$AF,$D9,$62,$AE,$E4,$AF,$88,$D9,$F5
	.dc.b	$FB,$DC,$75,$F8,$0F,$FB,$FB,$DF,$FF,$56,$DF,$F4,$D3,$FC,$6A,$7F
	.dc.b	$C0,$0F,$81,$46,$BC,$BE,$13,$A6,$E1,$FC,$16,$39,$36,$38,$47,$F9
	.dc.b	$10,$57,$47,$91,$27,$83,$B8,$D4,$60,$98,$7F,$97,$A7,$BB,$FE,$6B
	.dc.b	$7A,$DF,$82,$77,$1F,$E2,$F7,$FD,$13,$7F,$ED,$88,$D9,$E1,$AE,$91
	.dc.b	$C5,$E9,$7E,$27,$33,$B3,$A1,$E2,$3F,$C7,$87,$4A,$F0,$98,$AE,$91
	.dc.b	$AA,$E8,$F2,$8F,$97,$5D,$AB,$89,$E8,$9F,$89,$7C,$A3,$A4,$78,$DF
	.dc.b	$F5,$2D,$F4,$72,$72,$8F,$33,$A9,$E4,$F5,$3C,$F1,$E9,$AF,$11,$1F
	.dc.b	$31,$E2,$3C,$11,$C6,$34,$F5,$8C,$9F,$8B,$9E,$ED,$33,$7E,$B7,$C0
	.dc.b	$76,$9F,$44,$7D,$7B,$F6,$47,$79,$D5,$F8,$1E,$31,$F0,$23,$C6,$EC
	.dc.b	$F2,$1D,$23,$E4,$1C,$1D,$11,$F8,$E6,$4F,$92,$E6,$BC,$97,$EC,$7D
	.dc.b	$57,$E6,$AD,$9E,$FA,$F0,$1F,$3A,$35,$3F,$80,$9F,$59,$27,$D4,$FE
	.dc.b	$1E,$BB,$3E,$8A,$E2,$FC,$A3,$EA,$4F,$6B,$F8,$93,$A8,$C4,$F3,$38
	.dc.b	$3E,$3A,$EA,$D8,$CD,$72,$9C,$B6,$32,$79,$CF,$8D,$7E,$BA,$F4,$90
	.dc.b	$FA,$51,$A3,$8B,$F5,$27,$EB,$C7,$C7,$6C,$FC,$CA,$F0,$CF,$7B,$F6
	.dc.b	$56,$A3,$E0,$B6,$FD,$47,$C0,$3C,$71,$E1,$70,$D8,$E0,$7F,$00,$AF
	.dc.b	$87,$3F,$C2,$4F,$D3,$16,$7F,$8F,$B6,$FF,$DE,$2F,$E6,$4C,$47,$17
	.dc.b	$F6,$66,$0C,$3F,$8E,$F7,$DF,$03,$FA,$36,$C7,$E5,$9B,$3F,$B6,$6E
	.dc.b	$1F,$A1,$6C,$FD,$36,$CF,$C0,$6B,$F4,$9B,$1E,$1A,$F0,$11,$F0,$36
	.dc.b	$7C,$0D,$AF,$03,$67,$C0,$D9,$F0,$35,$F0,$36,$3C,$07,$62,$47,$66
	.dc.b	$CF,$66,$9D,$8B,$1D,$9A,$F6,$69,$D9,$E8,$48,$E8,$D8,$E8,$D3,$A1
	.dc.b	$4E,$85,$7A,$47,$22,$3C,$9B,$1C,$9A,$F2,$6C,$72,$38,$12,$38,$35
	.dc.b	$E0,$D3,$83,$A2,$1A,$2B,$A7,$04,$70,$4B,$FE,$C9,$7F,$84,$6D,$FD
	.dc.b	$76,$CF,$CC,$6D,$7C,$36,$DF,$B1,$B3,$E7,$6C,$F9,$1A,$F8,$DB,$1E
	.dc.b	$33,$FC,$11,$7E,$36,$DF,$8D,$A7,$8C,$B5,$E3,$6D,$F8,$DA,$78,$CB
	.dc.b	$1E,$36,$AF,$0D,$9A,$53,$FE,$A3,$9C,$C7,$F8,$F3,$2D,$4E,$6A,$DE
	.dc.b	$23,$A4,$7F,$06,$79,$DF,$27,$F8,$33,$5F,$E2,$0D,$7F,$C3,$4E,$E7
	.dc.b	$FC,$B5,$F4,$3D,$B0,$CD,$DB,$76,$B8,$97,$99,$F7,$7C,$15,$C8,$F1
	.dc.b	$BE,$1A,$E6,$26,$5F,$C6,$EF,$8B,$6F,$8B,$67,$DD,$3A,$3F,$CD,$B3
	.dc.b	$C9,$B3,$F3,$33,$F8,$DF,$05,$FE,$CD,$3E,$94,$7F,$90,$F0,$F4,$46
	.dc.b	$6F,$AB,$AB,$FE,$5C,$D7,$F9,$C3,$5F,$F2,$13,$BB,$FF,$1E,$FB,$3D
	.dc.b	$2B,$FC,$1C,$FC,$5F,$7D,$F0,$1E,$A6,$E5,$F0,$5B,$8F,$F3,$D6,$BF
	.dc.b	$D1,$FE,$01,$EF,$BE,$03,$E4,$46,$1B,$DE,$F3,$F3,$5B,$BF,$E9,$1F
	.dc.b	$0F,$F6,$5F,$A0,$D7,$3C,$9F,$4D,$46,$5F,$C4,$F8,$5F,$E6,$1E,$0F
	.dc.b	$F5,$06,$F7,$6F,$CD,$F2,$7E,$16,$31,$7F,$94,$BA,$FE,$8D,$ED,$C2
	.dc.b	$69,$23,$F0,$4F,$08,$D4,$7D,$DB,$FC,$CC,$F6,$BF,$CA,$72,$76,$F8
	.dc.b	$07,$C0,$E7,$57,$A7,$4D,$44,$CB,$F2,$55,$C1,$B1,$F0,$E3,$11,$FD
	.dc.b	$C5,$BA,$FE,$CD,$B9,$FF,$A1,$5B,$3F,$A8,$C1,$F1,$FF,$0B,$E0,$9E
	.dc.b	$7F,$C4,$F8,$27,$0D,$35,$13,$0F,$80,$7F,$08,$9D,$B8,$0E,$16,$7F
	.dc.b	$C1,$EF,$BC,$6D,$AE,$B3,$FF,$4E,$57,$1A,$FE,$1C,$D8,$FF,$50,$57
	.dc.b	$09,$E9,$F0,$5E,$9F,$22,$61,$D6,$3D,$C6,$1E,$C7,$A6,$32,$FA,$7E
	.dc.b	$01,$F0,$1F,$86,$5F,$05,$4C,$44,$9E,$0D,$7A,$8F,$81,$7C,$17,$C2
	.dc.b	$87,$BE,$F3,$3F,$D7,$55,$AD,$3F,$99,$CE,$6B,$F9,$AE,$5E,$47,$37
	.dc.b	$FE,$C1,$8C,$47,$48,$E7,$87,$7A,$E5,$3E,$38,$C4,$F1,$C7,$FD,$95
	.dc.b	$5A,$BE,$6F,$0C,$7C,$98,$75,$3F,$A3,$B8,$8E,$15,$CA,$FA,$57,$98
	.dc.b	$DE,$1E,$06,$E3,$95,$8F,$82,$B8,$C7,$73,$D1,$3B,$C7,$C1,$3A,$C7
	.dc.b	$C3,$59,$C3,$B3,$C0,$E5,$87,$57,$E5,$E3,$E7,$7C,$67,$7C,$3A,$BA
	.dc.b	$3C,$D3,$C5,$E9,$7F,$16,$FE,$E4,$7F,$9A,$EB,$E3,$B6,$32,$FD,$4C
	.dc.b	$7F,$A7,$4B,$C9,$7C,$16,$1B,$5F,$E6,$D9,$D4,$E4,$FE,$49,$0F,$82
	.dc.b	$A4,$7F,$94,$37,$1F,$C0,$FB,$EF,$82,$FE,$F7,$1F,$84,$AF,$F0,$0D
	.dc.b	$7D,$3A,$D7,$1F,$D5,$9C,$9F,$4E,$32,$FD,$53,$35,$DA,$FA,$EF,$FE
	.dc.b	$85,$C3,$A4,$62,$BF,$D2,$11,$8C,$3B,$47,$AC,$ED,$3B,$7F,$C2,$A6
	.dc.b	$6F,$B4,$F0,$AE,$53,$E3,$35,$7C,$30,$F3,$DF,$BE,$72,$7F,$88,$23
	.dc.b	$33,$E7,$8F,$0D,$F5,$69,$D6,$FB,$DF,$8A,$3F,$7F,$7F,$13,$4F,$F1
	.dc.b	$15,$79,$1F,$F1,$34,$66,$FC,$98,$FF,$85,$EF,$E4,$E3,$89,$FE,$55
	.dc.b	$5F,$D1,$9B,$D7,$FE,$21,$F0,$9F,$02,$F8,$2F,$F1,$05,$EF,$1F,$1D
	.dc.b	$E0,$E1,$5B,$C3,$99,$DB,$C2,$E2,$75,$F8,$27,$C1,$62,$74,$4C,$3F
	.dc.b	$94,$E6,$AD,$CE,$76,$F7,$E1,$C6,$F7,$5A,$7E,$3C,$7F,$8D,$FE,$07
	.dc.b	$FA,$8D,$FF,$8C,$5B,$3F,$EC,$D6,$BF,$ED,$76,$BF,$84,$77,$7F,$E5
	.dc.b	$6F,$11,$EF,$3E,$07,$B7,$81,$FE,$41,$8C,$63,$FD,$5F,$0E,$2F,$F8
	.dc.b	$22,$33,$8F,$F3,$A9,$CB,$FE,$50,$8C,$E3,$FE,$61,$F8,$0F,$F2,$95
	.dc.b	$FF,$92,$5B,$5F,$EE,$56,$BF,$EE,$F6,$BF,$9C,$77,$3E,$9A,$7C,$07
	.dc.b	$F9,$6E,$73,$7E,$58,$C7,$8B,$EB,$6C,$7F,$74,$D3,$E0,$C7,$F9,$8F
	.dc.b	$E0,$1E,$FF,$F9,$61,$B5,$FE,$5F,$6E,$7F,$EF,$B6,$BF,$B0,$0E,$B3
	.dc.b	$F9,$DF,$01,$FE,$27,$AF,$F0,$04,$E5,$FF,$18,$C6,$6B,$FB,$1C,$E5
	.dc.b	$FF,$3C,$C6,$6B,$FD,$15,$39,$C5,$39,$BF,$E0,$3E,$E3,$D0,$FF,$38
	.dc.b	$B6,$3F,$CF,$0D,$C7,$D0,$1C,$3E,$B7,$58,$FE,$1F,$47,$BF,$FE,$89
	.dc.b	$9C,$DF,$F9,$2A,$72,$FF,$87,$63,$37,$FE,$2B,$AE,$18,$7F,$A3,$6F
	.dc.b	$73,$FE,$02,$A7,$BF,$E8,$26,$1E,$86,$CF,$F8,$53,$3F,$F4,$E7,$BF
	.dc.b	$E3,$9E,$8D,$AF,$BD,$82,$31,$7C,$63,$13,$E7,$F7,$FE,$39,$1F,$8E
	.dc.b	$D9,$E1,$EE,$7C,$18,$F5,$36,$7B,$B6,$77,$1F,$E3,$F3,$FA,$D1,$7C
	.dc.b	$5F,$F5,$B2,$F7,$DE,$F7,$FA,$D6,$37,$5F,$E2,$66,$F0,$7F,$B2,$DA
	.dc.b	$7F,$B5,$0E,$53,$C6,$E4,$77,$3F,$CA,$66,$8C,$57,$58,$DB,$FE,$40
	.dc.b	$8D,$39,$77,$5E,$19,$F4,$DF,$A2,$72,$7F,$52,$3C,$B5,$E5,$9D,$DF
	.dc.b	$6B,$E5,$5C,$2B,$FD,$DD,$7C,$DF,$01,$27,$6F,$96,$35,$18,$24,$F8
	.dc.b	$6F,$BB,$D6,$BF,$DF,$51,$E1,$8F,$91,$5C,$9E,$06,$CF,$49,$F0,$AB
	.dc.b	$AD,$7A,$E3,$D0,$64,$E2,$F9,$DB,$07,$BD,$F8,$67,$91,$B3,$E0,$B9
	.dc.b	$7C,$F1,$DA,$BC,$D1,$C2,$BE,$1B,$FC,$46,$7F,$0E,$DB,$FF,$78,$C6
	.dc.b	$23,$C6,$FE,$1F,$DD,$7B,$FE,$38,$46,$1F,$6B,$67,$CD,$1A,$9E,$13
	.dc.b	$9B,$F8,$B3,$93,$8C,$E8,$8E,$09,$5D,$2B,$94,$79,$A3,$E4,$46,$1F
	.dc.b	$81,$1E,$99,$E7,$1C,$5B,$5E,$2A,$E1,$3F,$28,$FC,$E3,$8A,$FB,$AD
	.dc.b	$BF,$D9,$B6,$3F,$4E,$D9,$E3,$7D,$A7,$CC,$7B,$E3,$C4,$FB,$2F,$C5
	.dc.b	$3B,$9F,$49,$EC,$70,$F9,$63,$F0,$CE,$67,$91,$DA,$BC,$53,$A8,$F0
	.dc.b	$3D,$9D,$CF,$67,$27,$1A,$EE,$FC,$06,$DF,$C0,$6C,$7C,$06,$CF,$C0
	.dc.b	$8E,$AF,$9D,$FA,$46,$0E,$85,$9E,$8D,$3A,$16,$3A,$1F,$20,$95,$EB
	.dc.b	$6C,$7A,$4E,$11,$F0,$6F,$E1,$C7,$8A,$76,$7C,$18,$CB,$F2,$CC,$9E
	.dc.b	$93,$13,$DC,$93,$DE,$36,$FB,$1B,$5E,$C6,$C7,$B1,$A7,$B2,$79,$BE
	.dc.b	$3B,$F9,$71,$E9,$6C,$FA,$63,$73,$E5,$8F,$B2,$F1,$7E,$09,$B3,$CD
	.dc.b	$1C,$1F,$33,$6B,$CC,$D8,$F3,$36,$7C,$D1,$C8,$8F,$73,$D7,$1E,$78
	.dc.b	$F0,$C7,$C8,$76,$F9,$EF,$E7,$C7,$A2,$31,$7E,$E7,$81,$D6,$BC,$47
	.dc.b	$57,$BC,$FB,$5D,$3D,$8F,$69,$6F,$DA,$D3,$DA,$58,$F6,$9C,$88,$79
	.dc.b	$1F,$59,$E7,$3A,$BA,$6B,$B3,$CA,$7A,$8C,$4F,$AA,$3D,$11,$87,$C5
	.dc.b	$1E,$D3,$51,$C6,$3A,$92,$7D,$71,$E1,$6B,$CC,$C4,$7B,$9A,$7B,$8D
	.dc.b	$11,$E8,$6D,$C4,$74,$8C,$C7,$16,$CF,$9D,$F4,$9E,$C3,$25,$9E,$44
	.dc.b	$7B,$1A,$9F,$74,$65,$EC,$D9,$ED,$01,$D8,$CF,$F9,$1C,$7F,$9B,$63
	.dc.b	$17,$FE,$58,$8D,$CE,$9F,$F7,$0C,$E9,$FF,$2E,$B6,$BF,$CB,$5E,$FF
	.dc.b	$F2,$E7,$AC,$66,$FF,$CB,$53,$97,$9D,$FF,$95,$E3,$17,$E7,$AE,$0B
	.dc.b	$85,$BA,$B7,$3E,$CF,$5F,$CC,$F1,$6B,$9C,$7D,$98,$27,$2D,$8F,$F2
	.dc.b	$DB,$D0,$C4,$FF,$96,$DC,$4F,$F9,$B5,$B3,$C0,$F9,$D9,$7C,$CF,$7F
	.dc.b	$E7,$C7,$CB,$6D,$FF,$A1,$1A,$FF,$A7,$5A,$FD,$70,$F0,$9F,$EA,$9B
	.dc.b	$BD,$5F,$83,$39,$BF,$F1,$7C,$63,$0F,$F2,$4B,$63,$FC,$AB,$5C,$5F
	.dc.b	$39,$DF,$DD,$7B,$8F,$77,$F0,$51,$86,$F4,$BC,$5C,$D6,$F5,$FF,$4F
	.dc.b	$FB,$BF,$59,$FC,$53,$5C,$F0,$FE,$35,$46,$5F,$F5,$F5,$BC,$BD,$36
	.dc.b	$3F,$28,$D7,$FD,$5A,$D7,$F3,$5E,$EF,$5D,$CD,$24,$7E,$09,$F9,$DC
	.dc.b	$3F,$96,$6E,$DF,$9E,$6C,$FE,$47,$2F,$C4,$3F,$20,$FC,$B4,$69,$FD
	.dc.b	$6B,$6F,$C3,$1F,$EA,$6E,$2D,$7B,$8F,$77,$E1,$36,$7F,$60,$D7,$FD
	.dc.b	$92,$D7,$FD,$C4,$E7,$2F,$DC,$B7,$2F,$C4,$36,$7D,$1E,$EF,$82,$E4
	.dc.b	$7A,$44,$EF,$9F,$F5,$B5,$EB,$F7,$1E,$57,$B7,$DD,$FC,$86,$1C,$5A
	.dc.b	$B3,$FE,$17,$5C,$23,$B3,$5D,$67,$F9,$DA,$E3,$5F,$C4,$9B,$1F,$A1
	.dc.b	$AE,$13,$D2,$FC,$5D,$13,$0E,$B1,$E7,$F7,$7C,$DE,$EC,$9F,$0C,$65
	.dc.b	$F4,$46,$5F,$19,$FC,$97,$04,$FD,$BF,$6D,$EE,$5F,$76,$5F,$C9,$9C
	.dc.b	$B6,$78,$11,$F9,$DE,$E3,$DD,$F8,$D0,$F7,$1E,$D7,$EF,$EB,$5B,$FF
	.dc.b	$37,$9C,$D7,$F3,$9C,$7B,$BC,$6F,$F8,$34,$62,$77,$2B,$E6,$EC,$FE
	.dc.b	$84,$E2,$37,$3C,$6B,$9E,$3F,$C4,$67,$9D,$6E,$7C,$D8,$FA,$A7,$8D
	.dc.b	$6E,$78,$E1,$E9,$77,$8D,$8F,$0D,$F1,$9D,$C7,$94,$63,$E3,$77,$5C
	.dc.b	$5F,$65,$71,$7E,$25,$F8,$5A,$6F,$1F,$43,$C8,$ED,$7D,$5E,$F3,$C5
	.dc.b	$EB,$7C,$6F,$B4,$68,$F1,$57,$1B,$EB,$1C,$0F,$E7,$07,$72,$BF,$49
	.dc.b	$B3,$F8,$27,$3E,$2B,$DD,$FE,$E1,$48,$C4,$F4,$3E,$E3,$87,$8C,$FF
	.dc.b	$2B,$5E,$ED,$23,$FC,$BD,$BA,$7E,$1B,$DC,$7B,$BF,$EA,$48,$FC,$4D
	.dc.b	$7F,$81,$EB,$E8,$56,$B9,$7E,$94,$E4,$FB,$11,$97,$ED,$19,$AE,$D7
	.dc.b	$D7,$7F,$56,$1D,$23,$15,$F2,$23,$18,$76,$8F,$59,$DA,$76,$FF,$65
	.dc.b	$9D,$BE,$03,$77,$E0,$AE,$53,$A8,$F6,$CE,$AF,$FB,$76,$1E,$99,$CD
	.dc.b	$6E,$7C,$F5,$E1,$BE,$6D,$26,$DF,$E2,$AC,$3E,$13,$FE,$12,$BF,$2B
	.dc.b	$9C,$7F,$C2,$F3,$AB,$EB,$8F,$9F,$0F,$E2,$62,$5E,$EF,$03,$FC,$4B
	.dc.b	$E2,$3D,$DF,$EC,$D8,$7F,$3B,$AD,$B6,$B8,$3F,$D5,$EB,$87,$87,$8C
	.dc.b	$3B,$DF,$F8,$25,$FF,$23,$36,$5E,$E3,$DD,$F4,$F8,$53,$1F,$03,$E4
	.dc.b	$9D,$46,$09,$87,$69,$D5,$71,$8F,$7F,$B8,$F9,$9F,$F1,$DB,$6F,$FB
	.dc.b	$77,$82,$F7,$5E,$F6,$AE,$57,$F5,$BB,$E1,$87,$F3,$CC,$38,$5E,$A3
	.dc.b	$BD,$FF,$91,$2B,$71,$FB,$86,$BF,$E3,$16,$BF,$E6,$9F,$79,$F1,$BF
	.dc.b	$E5,$06,$E3,$E6,$C9,$19,$99,$3F,$DB,$E3,$75,$E5,$AD,$CE,$6B,$73
	.dc.b	$FB,$FA,$DB,$FC,$31,$AF,$F9,$15,$AF,$F9,$F7,$DA,$79,$9F,$E6,$46
	.dc.b	$DF,$CA,$F3,$1F,$05,$9A,$93,$FE,$03,$AD,$DF,$C9,$BD,$CE,$6B,$73
	.dc.b	$FC,$4E,$B6,$FF,$20,$6B,$FE,$59,$6B,$FE,$91,$9F,$E3,$87,$F8,$44
	.dc.b	$AF,$F7,$25,$E9,$3D,$8F,$F3,$EC,$6B,$AF,$FD,$2E,$D3,$FC,$CC,$5A
	.dc.b	$3C,$A7,$8F,$7C,$8F,$F4,$E1,$E0,$9F,$E6,$ED,$79,$C7,$FA,$23,$F5
	.dc.b	$0F,$F4,$74,$FF,$89,$DB,$3F,$17,$E6,$FF,$D2,$31,$8C,$3F,$D5,$6D
	.dc.b	$AF,$F3,$D3,$6B,$FC,$95,$5C,$3D,$79,$5F,$EA,$FA,$EB,$5F,$D2,$9A
	.dc.b	$F2,$9F,$27,$9D,$E0,$3F,$D4,$FC,$DF,$EA,$4F,$84,$C6,$1E,$79,$B7
	.dc.b	$F6,$E6,$E5,$E6,$9E,$75,$E1,$6D,$7F,$A5,$7A,$7F,$D1,$0F,$D5,$D9
	.dc.b	$ED,$E3,$0F,$F4,$F5,$5C,$3E,$F5,$69,$FF,$27,$7B,$3F,$C3,$70,$FF
	.dc.b	$4D,$D6,$DF,$AF,$5B,$7F,$90,$CE,$C9,$3F,$EC,$4A,$FC,$2E,$1C,$6B
	.dc.b	$9C,$76,$9D,$5F,$3C,$3C,$B3,$92,$57,$AE,$F9,$61,$DA,$7F,$D4,$B1
	.dc.b	$8C,$3C,$D1,$E0,$38,$DF,$4B,$F3,$CF,$82,$B8,$56,$8F,$86,$E2,$77
	.dc.b	$1D,$8D,$3F,$EF,$5B,$F1,$3E,$58,$CB,$EE,$38,$99,$2D,$F8,$C9,$1E
	.dc.b	$03,$8C,$F4,$8F,$F6,$4D,$78,$DD,$D7,$D2,$8C,$57,$FB,$46,$BF,$A2
	.dc.b	$3F,$06,$7B,$B9,$BF,$8B,$87,$CF,$8C,$1E,$08,$F8,$8D,$3D,$E7,$3C
	.dc.b	$3C,$35,$ED,$7C,$71,$93,$9C,$78,$0F,$8D,$3D,$1C,$1C,$63,$73,$EE
	.dc.b	$7C,$11,$C5,$E8,$F2,$73,$1B,$8E,$77,$F1,$E7,$E5,$46,$2B,$DF,$87
	.dc.b	$C6,$8E,$47,$D3,$BF,$5D,$F1,$8F,$99,$7D,$8F,$29,$D2,$77,$7D,$23
	.dc.b	$53,$D2,$7C,$86,$A3,$E4,$E1,$F3,$E7,$51,$D1,$A7,$80,$F4,$3C,$30
	.dc.b	$F4,$47,$C7,$35,$5D,$5F,$8D,$5D,$5D,$13,$0F,$C1,$C6,$5F,$74,$73
	.dc.b	$BF,$9A,$EF,$0F,$63,$E2,$3A,$E1,$E2,$AE,$6D,$9F,$B9,$87,$8C,$FC
	.dc.b	$1E,$1D,$E3,$E8,$61,$DA,$FC,$27,$27,$E4,$E1,$E3,$AE,$11,$F3,$0E
	.dc.b	$15,$CA,$33,$3E,$5B,$F9,$F5,$F2,$E3,$D4,$FE,$DD,$D3,$F3,$CE,$67
	.dc.b	$59,$E7,$3E,$98,$FB,$C7,$3B,$FC,$23,$C2,$B9,$9F,$49,$D5,$FC,$8A
	.dc.b	$E5,$3F,$72,$78,$46,$1A,$70,$3C,$CF,$08,$F8,$58,$7C,$D9,$C9,$B8
	.dc.b	$FA,$6F,$C8,$BF,$95,$1E,$43,$8C,$66,$FE,$AB,$C6,$FE,$AD,$F5,$AF
	.dc.b	$D1,$4E,$4F,$8D,$87,$C4,$71,$1E,$F3,$F2,$E7,$CE,$38,$98,$C3,$E3
	.dc.b	$CF,$18,$C6,$1E,$26,$D7,$E7,$4F,$31,$EE,$72,$7B,$8F,$65,$7B,$B0
	.dc.b	$F8,$CD,$9F,$2C,$FC,$03,$75,$DD,$F3,$57,$C0,$9E,$F1,$E8,$32,$FC
	.dc.b	$0C,$3E,$24,$FE,$32,$7C,$6F,$17,$AC,$F9,$DE,$27,$DF,$C3,$E0,$38
	.dc.b	$8F,$BE,$F2,$7F,$1A,$D3,$85,$7F,$4F,$7F,$DA,$78,$30,$FF,$12,$C7
	.dc.b	$F7,$DA,$FF,$66,$3F,$D3,$4D,$C7,$F6,$AF,$43,$F9,$E9,$27,$F8,$8B
	.dc.b	$67,$F8,$3B,$71,$FD,$BB,$67,$F0,$2D,$9F,$63,$5F,$DB,$36,$FF,$86
	.dc.b	$B6,$FC,$55,$E0,$25,$78,$1B,$7E,$06,$9E,$02,$C7,$81,$AF,$81,$BF
	.dc.b	$A8,$BD,$CF,$8B,$5F,$1A,$93,$9A,$FF,$35,$7A,$FA,$53,$27,$4F,$FA
	.dc.b	$E9,$B5,$CD,$B3,$C3,$77,$3B,$0E,$98,$73,$C6,$BE,$27,$C8,$F8,$0C
	.dc.b	$5F,$DA,$C3,$C3,$19,$32,$51,$FA,$97,$FB,$45,$F4,$B7,$50,$F8,$1F
	.dc.b	$0C,$6E,$78,$F9,$38,$3C,$F5,$C2,$B1,$87,$C4,$D7,$E5,$60,$BF,$4E
	.dc.b	$1E,$06,$BD,$2B,$F2,$73,$B7,$F2,$BF,$12,$F8,$AD,$B6,$33,$18,$39
	.dc.b	$FC,$43,$9A,$92,$7F,$3F,$19,$7F,$43,$F0,$0F,$7D,$F0,$18,$6E,$7F
	.dc.b	$26,$7F,$D5,$BE,$56,$6B,$F5,$3E,$BB,$DB,$FA,$9C,$1F,$89,$FA,$F7
	.dc.b	$CF,$63,$4F,$B1,$18,$CF,$E3,$47,$DB,$C3,$94,$60,$FC,$6B,$85,$7D
	.dc.b	$6F,$A4,$7E,$4A,$FA,$3A,$BE,$98,$6A,$18,$73,$7F,$0C,$D7,$67,$B2
	.dc.b	$3F,$82,$BF,$88,$F5,$7B,$F5,$7E,$63,$F0,$3B,$D6,$7B,$72,$AA,$9A
	.dc.b	$89,$7E,$D3,$A1,$67,$83,$F8,$29,$F0,$C6,$23,$F3,$4D,$7F,$3C,$BB
	.dc.b	$7C,$BF,$9C,$7E,$E7,$7F,$E3,$F9,$23,$1A,$FE,$A6,$F8,$61,$E1,$C3
	.dc.b	$8C,$C8,$C1,$0F,$0C,$79,$08,$FF,$2A,$F7,$BE,$B6,$3F,$C2,$7E,$03
	.dc.b	$AC,$FD,$33,$F8,$3B,$88,$FE,$1D,$3C,$2F,$F7,$8D,$AD,$DF,$D9,$BE
	.dc.b	$15,$AD,$BE,$E5,$EE,$D5,$A7,$89,$F3,$E3,$43,$DF,$E9,$F0,$12,$75
	.dc.b	$1D,$0C,$3E,$93,$AC,$65,$FB,$5E,$FB,$E0,$3F,$1C,$FE,$3A,$33,$C7
	.dc.b	$89,$91,$C1,$AF,$5F,$7D,$F0,$1F,$9E,$87,$BD,$F8,$19,$CF,$07,$F3
	.dc.b	$08,$FC,$26,$3E,$1A,$C7,$27,$F2,$58,$C4,$72,$8F,$E5,$17,$CA,$30
	.dc.b	$7F,$3A,$71,$5D,$31,$FE,$5C,$D9,$EA,$6F,$1F,$34,$79,$5B,$7E,$1A
	.dc.b	$E2,$DA,$E2,$2F,$BC,$79,$2B,$8C,$FF,$4C,$C3,$D3,$87,$86,$3C,$E7
	.dc.b	$19,$DB,$71,$EB,$5C,$2F,$C1,$3C,$30,$EE,$FA,$6B,$78,$F9,$A3,$C0
	.dc.b	$D7,$A1,$C7,$0E,$F8,$75,$8D,$9E,$CB,$EA,$F0,$3F,$A9,$3F,$3D,$B3
	.dc.b	$F7,$5B,$7F,$E1,$29,$79,$CF,$80,$C3,$6B,$FA,$D4,$EA,$72,$7F,$16
	.dc.b	$87,$C0,$52,$3F,$C6,$9B,$A7,$F0,$BF,$77,$FA,$74,$FC,$C3,$FB,$DC
	.dc.b	$FD,$68,$CC,$FF,$0F,$9E,$A7,$D2,$8C,$EB,$FD,$B6,$B8,$78,$BF,$E2
	.dc.b	$EB,$E1,$18,$E5,$FC,$0E,$1D,$8E,$55,$BB,$E5,$18,$9D,$BE,$0B,$F4
	.dc.b	$56,$E3,$FB,$F5,$F5,$7D,$46,$EF,$FC,$0D,$87,$9A,$36,$75,$37,$3C
	.dc.b	$6F,$CF,$19,$69,$E0,$C3,$B6,$9F,$E2,$5B,$FF,$02,$57,$AE,$B4,$FF
	.dc.b	$07,$8F,$F0,$41,$A7,$C8,$7F,$8A,$1C,$FB,$FF,$E1,$E4,$FF,$88,$BC
	.dc.b	$07,$C0,$3E,$03,$53,$32,$FE,$95,$7C,$1F,$ED,$17,$C2,$B5,$87,$8F
	.dc.b	$CE,$EC,$75,$F8,$07,$C0,$63,$0E,$73,$6B,$F9,$C6,$0B,$DD,$68,$93
	.dc.b	$D5,$F0,$C6,$A3,$07,$F2,$9F,$79,$EF,$FF,$8F,$1B,$5F,$DF,$FC,$27
	.dc.b	$BE,$F8,$0D,$5C,$AF,$F3,$BD,$F0,$C3,$FC,$78,$D7,$FC,$80,$D7,$FC
	.dc.b	$8C,$EF,$8B,$FD,$03,$2F,$59,$A7,$5B,$FE,$AD,$39,$8F,$F2,$8F,$BE
	.dc.b	$F8,$0A,$7F,$95,$B5,$F3,$C6,$2B,$FB,$E3,$B7,$FD,$1A,$DB,$FF,$29
	.dc.b	$35,$FF,$2B,$35,$FF,$2E,$3B,$C7,$FB,$CE,$BF,$E6,$89,$FF,$37,$CE
	.dc.b	$6B,$FC,$D5,$39,$97,$BC,$F7,$FF,$CD,$2D,$AF,$9D,$E5,$BE,$03,$35
	.dc.b	$27,$FD,$47,$5B,$BF,$F3,$43,$5F,$F3,$63,$5F,$F3,$93,$BC,$BF,$C1
	.dc.b	$F8,$7F,$9A,$6B,$FC,$3D,$39,$7F,$D0,$11,$9A,$FE,$D9,$1F,$C8,$0F
	.dc.b	$F1,$21,$5F,$A7,$B3,$E0,$3F,$D0,$73,$9C,$3F,$D2,$CE,$EB,$FD,$00
	.dc.b	$D7,$FD,$0C,$D7,$FD,$2E,$D4,$F3,$33,$EE,$7C,$B7,$11,$A8,$F9,$DE
	.dc.b	$FB,$DF,$FF,$4A,$B6,$BE,$6F,$C0,$62,$64,$7F,$B3,$A7,$57,$FE,$93
	.dc.b	$6B,$FE,$97,$6B,$FE,$9D,$77,$87,$F8,$EB,$D0,$7C,$17,$4A,$FF,$16
	.dc.b	$71,$FF,$8F,$63,$C7,$F0,$33,$1F,$BF,$ED,$5B,$FF,$0D,$64,$AF,$F4
	.dc.b	$8B,$6B,$CD,$18,$23,$FB,$5C,$5F,$55,$CF,$3F,$23,$FE,$B6,$F7,$FF
	.dc.b	$DD,$D5,$FE,$5D,$6E,$3F,$6B,$7D,$77,$4A,$D3,$FE,$C0,$7F,$DF,$B3
	.dc.b	$FE,$FF,$E2,$FF,$80,$37,$CB,$1F,$F6,$A3,$6B,$FE,$00,$D7,$FA,$E2
	.dc.b	$FA,$3F,$E3,$79,$D3,$F7,$AF,$9C,$EA,$FF,$81,$C6,$2F,$C6,$7F,$A4
	.dc.b	$4F,$1D,$F2,$BE,$D1,$99,$E1,$1D,$EF,$D5,$18,$BF,$3C,$63,$0F,$05
	.dc.b	$FF,$B6,$70,$E8,$79,$9E,$F3,$AC,$3D,$15,$5F,$88,$DA,$E3,$3D,$6B
	.dc.b	$D3,$5E,$36,$DF,$BA,$F8,$57,$59,$EE,$68,$F9,$77,$D1,$C5,$F8,$9A
	.dc.b	$F8,$9B,$1B,$8F,$BD,$18,$9E,$D8,$7C,$38,$F8,$11,$E7,$9F,$31,$30
	.dc.b	$F8,$F3,$F3,$70,$F3,$BE,$23,$CF,$3B,$7C,$F5,$A8,$F0,$99,$07,$9C
	.dc.b	$C1,$6B,$BB,$8C,$3E,$96,$1E,$78,$E3,$18,$C3,$B5,$70,$3E,$66,$1E
	.dc.b	$F3,$CD,$7C,$6F,$F0,$E7,$92,$78,$4F,$C2,$37,$5D,$1C,$4E,$DB,$3C
	.dc.b	$23,$C7,$87,$C9,$8C,$47,$AE,$76,$47,$6F,$CB,$3C,$CF,$C5,$7E,$5C
	.dc.b	$75,$69,$C4,$B1,$C8,$F2,$16,$3C,$38,$7C,$59,$EB,$1C,$F0,$F8,$53
	.dc.b	$9A,$F9,$D7,$F8,$4C,$3E,$04,$73,$C3,$ED,$9C,$8A,$F0,$7D,$4D,$3D
	.dc.b	$57,$DA,$BE,$24,$FC,$19,$F8,$87,$53,$D4,$7E,$10,$A7,$D0,$30,$7D
	.dc.b	$82,$DF,$F1,$A6,$CF,$F4,$87,$66,$CE,$47,$C9,$AF,$F7,$A5,$7E,$13
	.dc.b	$D4,$FD,$19,$1C,$12,$7D,$71,$96,$C6,$71,$F9,$93,$96,$C6,$4F,$04
	.dc.b	$65,$A6,$67,$E1,$3F,$E2,$18,$D4,$62,$BC,$EF,$80,$95,$D9,$B5,$D1
	.dc.b	$A7,$1A,$E6,$FE,$82,$7F,$96,$C6,$30,$FD,$6B,$70,$FE,$4E,$DB,$FE
	.dc.b	$58,$DB,$FF,$00,$B6,$F9,$CF,$AE,$7B,$3F,$DF,$30,$F4,$CF,$9C,$91
	.dc.b	$E7,$6C,$F9,$E3,$96,$1E,$46,$E1,$E4,$69,$E4,$2C,$F9,$1B,$5E,$48
	.dc.b	$F8,$67,$90,$FE,$A1,$87,$A6,$7C,$CD,$8C,$C7,$23,$25,$72,$74,$BF
	.dc.b	$33,$6F,$CC,$D9,$F3,$34,$F3,$16,$FC,$C7,$C2,$7C,$D1,$F8,$33,$C0
	.dc.b	$57,$83,$E4,$69,$E4,$8E,$55,$E4,$6D,$79,$1B,$5E,$46,$D7,$91,$A7
	.dc.b	$92,$3E,$B1,$23,$EB,$1E,$07,$C9,$1C,$1F,$23,$63,$C9,$1F,$BD,$F8
	.dc.b	$0F,$F3,$21,$FE,$69,$E1,$72,$BE,$0B,$F7,$58,$7E,$EF,$05,$F0,$F8
	.dc.b	$2C,$47,$36,$BF,$D3,$21,$18,$8E,$2D,$8F,$07,$07,$F9,$86,$78,$4F
	.dc.b	$FA,$26,$31,$0F,$63,$FD,$E7,$7E,$36,$E7,$FD,$4B,$5F,$F3,$44,$E7
	.dc.b	$CC,$F4,$CF,$F1,$BA,$FF,$7D,$DF,$F1,$A6,$B9,$9F,$34,$62,$3F,$B2
	.dc.b	$BD,$96,$AC,$5E,$F7,$69,$DD,$FA,$63,$10,$F7,$DC,$FF,$0D,$B9,$7F
	.dc.b	$26,$9D,$4E,$4F,$91,$EF,$F6,$F7,$F4,$D9,$FE,$5A,$FF,$A6,$FC,$4C
	.dc.b	$E2,$AE,$9F,$0B,$F7,$2E,$F2,$77,$51,$97,$F9,$BB,$5F,$E7,$FE,$C3
	.dc.b	$DD,$FE,$89,$C5,$F8,$3F,$7D,$F0,$18,$DF,$FA,$44,$63,$0F,$F0,$6D
	.dc.b	$7E,$27,$0E,$51,$8A,$FE,$9D,$18,$8F,$E8,$ED,$7C,$38,$79,$E3,$11
	.dc.b	$C2,$3C,$0F,$F8,$6A,$39,$57,$79,$CC,$AF,$E2,$3E,$13,$EE,$FA,$BC
	.dc.b	$B9,$3F,$43,$F0,$17,$53,$CB,$99,$54,$F9,$4E,$2F,$ED,$35,$FE,$C2
	.dc.b	$D7,$F2,$11,$C5,$F1,$1F,$9D,$7C,$5C,$CF,$92,$F0,$76,$E3,$79,$7C
	.dc.b	$30,$4E,$A3,$C2,$F9,$08,$E0,$E0,$ED,$FF,$52,$DB,$E2,$BE,$3F,$C0
	.dc.b	$FC,$AC,$3F,$C0,$5F,$05,$DA,$FC,$ED,$5C,$3F,$C1,$6B,$84,$7F,$93
	.dc.b	$1A,$A7,$84,$FF,$0B,$6D,$7F,$83,$EB,$67,$08,$CD,$7F,$95,$9A,$A3
	.dc.b	$7F,$05,$F7,$A5,$F0,$54,$F5,$C4,$AF,$4C,$65,$FA,$34,$9F,$C3,$FC
	.dc.b	$0B,$E0,$AF,$34,$BF,$AF,$39,$6C,$F0,$23,$F7,$3E,$01,$F0,$5F,$1A
	.dc.b	$1E,$FF,$E1,$67,$3B,$7F,$9C,$A3,$F1,$18,$75,$AD,$F2,$7F,$8C,$E3
	.dc.b	$13,$C7,$0E,$8D,$BE,$55,$BC,$7F,$C7,$53,$D1,$B1,$FE,$3D,$C7,$C3
	.dc.b	$8F,$63,$FD,$06,$E2,$B8,$47,$43,$FD,$0F,$8F,$8A,$B9,$63,$E2,$6E
	.dc.b	$3B,$A5,$6D,$E3,$8F,$89,$EF,$7C,$5B,$5C,$6B,$84,$72,$AE,$0B,$1F
	.dc.b	$36,$1E,$08,$F5,$1B,$39,$4F,$09,$DC,$76,$37,$5D,$8E,$F1,$A8,$ED
	.dc.b	$87,$D9,$7F,$CC,$31,$F2,$1B,$19,$7C,$46,$BE,$09,$F0,$5F,$CF,$69
	.dc.b	$18,$9F,$F3,$2C,$62,$75,$39,$9F,$E8,$9F,$05,$FC,$DD,$34,$8F,$F3
	.dc.b	$B7,$E8,$7C,$03,$C6,$FF,$39,$6D,$FC,$46,$B8,$9F,$1A,$75,$C9,$F5
	.dc.b	$E3,$FD,$87,$7C,$AF,$8E,$DF,$E7,$A8,$C5,$6E,$B9,$E1,$FC,$7F,$0E
	.dc.b	$71,$F4,$8E,$67,$7C,$39,$BF,$E8,$AB,$E6,$EC,$E7,$87,$FA,$46,$B9
	.dc.b	$DF,$A6,$B6,$DA,$E9,$87,$8E,$73,$7E,$79,$C9,$D6,$32,$F8,$4C,$D7
	.dc.b	$A7,$0F,$EE,$F8,$7F,$7F,$9F,$F5,$15,$79,$B7,$FF,$11,$4E,$4D,$4F
	.dc.b	$BA,$F9,$63,$FC,$F2,$5F,$03,$D0,$FF,$03,$78,$8F,$82,$7C,$16,$6F
	.dc.b	$85,$5C,$7F,$96,$C7,$FB,$DA,$F8,$56,$72,$F1,$F9,$DE,$83,$AF,$C0
	.dc.b	$BE,$0B,$18,$78,$E6,$E3,$D0,$62,$BE,$15,$B2,$4E,$C6,$8B,$EF,$87
	.dc.b	$4C,$7F,$C8,$DF,$00,$F8,$1F,$F2,$03,$71,$FF,$0F,$37,$1F,$F6,$2B
	.dc.b	$5F,$F6,$6B,$9A,$FF,$65,$4B,$D3,$F3,$BF,$DA,$27,$4F,$FA,$77,$4F
	.dc.b	$F1,$7C,$62,$D3,$C6,$73,$87,$F8,$DE,$30,$7F,$72,$F8,$07,$C0,$FF
	.dc.b	$95,$DB,$1F,$E3,$56,$DF,$FB,$71,$AF,$FB,$91,$CE,$5F,$EE,$1F,$21
	.dc.b	$F0,$3F,$E6,$2B,$ED,$87,$F9,$1E,$73,$EA,$7F,$91,$B0,$E3,$4C,$7F
	.dc.b	$C9,$F8,$71,$7E,$94,$E9,$CD,$7F,$93,$BD,$F7,$C0,$7F,$9B,$DB,$87
	.dc.b	$F9,$35,$B5,$FE,$F4,$6B,$FE,$FC,$7F,$CE,$58,$FF,$A0,$BE,$17,$A5
	.dc.b	$F9,$FC,$DF,$F3,$6E,$1E,$5C,$18,$70,$9C,$C7,$A6,$31,$7F,$E6,$2A
	.dc.b	$FE,$42,$7F,$97,$CB,$3C,$CF,$80,$E1,$C3,$5F,$03,$C3,$5F,$C2,$FB
	.dc.b	$5F,$15,$FF,$4B,$46,$4E,$71,$97,$8F,$77,$E0,$63,$15,$E0,$8C,$69
	.dc.b	$FE,$78,$F7,$5E,$FF,$FA,$65,$B8,$7F,$9D,$DB,$7E,$90,$E1,$F6,$3A
	.dc.b	$C7,$F0,$9E,$63,$DF,$ED,$87,$FA,$7D,$B8,$7F,$95,$78,$3F,$D1,$D1
	.dc.b	$8C,$7F,$C6,$EF,$97,$DF,$F2,$12,$BC,$8D,$9F,$F1,$EF,$BF,$B9,$E9
	.dc.b	$8A,$31,$3A,$9F,$F2,$9E,$6B,$FF,$59,$57,$FA,$6B,$DD,$FE,$01,$5F
	.dc.b	$E5,$86,$DF,$F0,$B8,$FD,$B6,$FB,$EC,$91,$B7,$F6,$1D,$69,$CD,$7F
	.dc.b	$0D,$BE,$7B,$FF,$B3,$DB,$87,$E1,$2F,$83,$CC,$95,$C7,$0E,$B8,$74
	.dc.b	$C3,$94,$5A,$E5,$36,$BF,$5D,$87,$66,$D7,$2A,$EB,$7C,$9A,$72,$2C
	.dc.b	$F0,$69,$F0,$8C,$AA,$EC,$DA,$EC,$FA,$67,$73,$E3,$79,$46,$49,$5E
	.dc.b	$37,$E8,$61,$F1,$1C,$CF,$ED,$6B,$B9,$CE,$75,$5C,$49,$3B,$BF,$1C
	.dc.b	$62,$FD,$87,$43,$F9,$5D,$72,$AE,$07,$A2,$B8,$B8,$9F,$41,$FB,$DB
	.dc.b	$FC,$3B,$67,$F9,$AB,$FA,$98,$CB,$C6,$FF,$A0,$DF,$FA,$7A,$FE,$6B
	.dc.b	$67,$C9,$1F,$D1,$7D,$1F,$B8,$D3,$6E,$27,$85,$7B,$67,$55,$CA,$73
	.dc.b	$1E,$13,$F2,$15,$C4,$8F,$42,$4F,$18,$CB,$63,$35,$DE,$BA,$CF,$CC
	.dc.b	$8F,$6D,$70,$74,$DA,$F1,$4F,$D0,$8C,$3C,$63,$CA,$76,$36,$7B,$27
	.dc.b	$9B,$98,$E0,$73,$70,$FE,$4D,$FF,$72,$1C,$4E,$75,$E8,$8D,$1F,$90
	.dc.b	$AF,$1B,$8A,$F5,$D7,$4B,$FC,$85,$69,$B1,$9C,$3F,$81,$CE,$5B,$19
	.dc.b	$3C,$78,$78,$6B,$8E,$1F,$34,$98,$7B,$9B,$5E,$8C,$3A,$3C,$9B,$1E
	.dc.b	$98,$EC,$D3,$89,$F1,$9F,$84,$DB,$F1,$61,$E6,$8D,$9F,$06,$7E,$9B
	.dc.b	$8C,$3C,$F8,$79,$49,$3E,$53,$C2,$FC,$89,$E1,$5D,$63,$71,$E1,$C3
	.dc.b	$D3,$1B,$33,$1E,$13,$EE,$1C,$AB,$E0,$1D,$8F,$2C,$70,$BF,$2C,$78
	.dc.b	$E7,$81,$F0,$E7,$57,$E3,$9D,$46,$1E,$B3,$C2,$3C,$91,$A2,$47,$A5
	.dc.b	$F1,$3E,$68,$E4,$F7,$8D,$BE,$B7,$E8,$B7,$0F,$A3,$1E,$09,$E2,$78
	.dc.b	$27,$B1,$E3,$8F,$79,$D2,$FD,$F1,$A9,$E3,$3A,$8C,$3B,$BF,$81,$18
	.dc.b	$21,$E3,$C3,$E9,$CF,$18,$C3,$EE,$3D,$37,$E0,$8C,$9E,$53,$E1,$47
	.dc.b	$73,$D8,$57,$E2,$1E,$08,$E4,$4A,$F2,$C7,$4B,$F5,$BD,$0E,$11,$A9
	.dc.b	$E1,$5D,$A7,$2D,$79,$3E,$1A,$F9,$91,$83,$67,$90,$C1,$27,$AB,$96
	.dc.b	$C7,$26,$CF,$C7,$6C,$F9,$CE,$55,$E7,$24,$F9,$67,$6F,$CC,$8C,$CF
	.dc.b	$76,$D7,$76,$CF,$76,$CF,$7A,$E7,$87,$F1,$DF,$7B,$FC,$DD,$3F,$E5
	.dc.b	$36,$BF,$E5,$F6,$BF,$E7,$56,$A7,$FC,$97,$C0,$DF,$FC,$DF,$97,$F5
	.dc.b	$EA,$D7,$17,$FA,$06,$73,$9F,$58,$C7,$3F,$F6,$4B,$E1,$8F,$4B,$DE
	.dc.b	$1F,$E8,$5F,$4F,$FD,$DE,$FF,$9A,$5A,$FF,$9D,$5A,$EF,$C7,$F1,$39
	.dc.b	$7C,$5E,$DF,$7C,$F0,$7F,$91,$9C,$4F,$18,$C1,$FE,$95,$C7,$D9,$7F
	.dc.b	$F0,$01,$FF,$2E,$D7,$0A,$7B,$88,$7B,$9F,$E6,$79,$CC,$61,$FF,$35
	.dc.b	$4F,$FA,$6B,$C1,$D3,$C6,$B8,$7B,$8E,$EE,$56,$8C,$3A,$8C,$C6,$23
	.dc.b	$FC,$ED,$BF,$FB,$C1,$F1,$35,$CF,$8F,$D2,$31,$1F,$E9,$5F,$49,$ED
	.dc.b	$F8,$1B,$1E,$B6,$BF,$E8,$36,$BC,$63,$FD,$51,$AF,$E0,$E7,$6D,$8C
	.dc.b	$EF,$FE,$AD,$F6,$FE,$25,$FF,$BC,$6F,$73,$FE,$B1,$6C,$7F,$B1,$DA
	.dc.b	$A7,$8D,$FD,$6F,$3F,$E3,$F0,$9A,$4A,$FA,$91,$97,$EF,$61,$F8,$B9
	.dc.b	$F2,$C7,$E4,$5B,$CD,$F9,$FF,$6B,$32,$BE,$4B,$0F,$13,$D0,$E4,$7F
	.dc.b	$A5,$27,$8C,$BA,$FC,$E4,$9F,$2B,$5E,$98,$FE,$7E,$38,$9E,$4E,$67
	.dc.b	$B5,$8A,$F0,$1E,$59,$C3,$C8,$65,$FB,$99,$F5,$FB,$1F,$33,$DA,$F2
	.dc.b	$57,$26,$BC,$44,$7D,$23,$85,$FC,$49,$DE,$9F,$C0,$E7,$5C,$19,$A9
	.dc.b	$7E,$69,$E5,$EC,$F1,$F6,$A4,$E6,$3C,$A6,$1F,$16,$29,$CC,$7C,$0F
	.dc.b	$65,$ED,$7D,$65,$C9,$4C,$44,$9E,$0D,$7D,$DE,$CB,$DA,$FB,$90,$F6
	.dc.b	$3E,$FC,$E7,$74,$E7,$9A,$AC,$BC,$6E,$AF,$F0,$F1,$88,$E0,$72,$BE
	.dc.b	$8E,$03,$89,$E3,$8F,$66,$CF,$5B,$F0,$D6,$DB,$8F,$AA,$B7,$5D,$70
	.dc.b	$F1,$47,$B2,$75,$5C,$31,$E9,$7E,$EB,$DC,$FB,$9B,$1E,$9A,$E2,$DA
	.dc.b	$E2,$FC,$3C,$7D,$17,$C5,$B1,$B3,$E0,$63,$E9,$7C,$11,$A7,$91,$D2
	.dc.b	$34,$FC,$6B,$F0,$BA,$3D,$D3,$C6,$76,$7C,$E8,$FA,$E7,$1B,$FE,$71
	.dc.b	$19,$8F,$84,$68,$FA,$7E,$63,$DA,$DB,$63,$52,$8C,$47,$68,$C4,$FF
	.dc.b	$3C,$87,$B5,$48,$FF,$40,$6E,$9F,$C3,$BD,$77,$89,$FE,$CC,$8F,$E2
	.dc.b	$35,$F9,$2A,$F9,$F5,$AE,$9F,$A3,$39,$3F,$19,$19,$7F,$1C,$66,$BB
	.dc.b	$5F,$5D,$BE,$06,$1D,$23,$15,$F2,$23,$18,$76,$8F,$59,$DA,$76,$FE
	.dc.b	$92,$76,$F8,$0D,$DF,$82,$B9,$4E,$A3,$DB,$3A,$BF,$D6,$E1,$E9,$9C
	.dc.b	$D6,$E7,$CF,$5E,$1B,$E6,$D2,$69,$FE,$17,$C3,$FB,$E4,$F0,$C3,$C5
	.dc.b	$5F,$C7,$E7,$2F,$EF,$9A,$FA,$CE,$73,$E4,$F6,$7F,$C0,$92,$7F,$C1
	.dc.b	$BE,$23,$DA,$78,$3A,$99,$7F,$CE,$2F,$83,$FD,$26,$F8,$56,$B0,$F2
	.dc.b	$79,$7D,$8E,$BE,$D3,$DA,$C6,$1C,$E6,$D7,$F4,$1C,$17,$BA,$D1,$27
	.dc.b	$AB,$E1,$8D,$46,$0F,$E8,$5E,$3B,$AF,$FC,$44,$DA,$FF,$6D,$DE,$26
	.dc.b	$47,$F5,$49,$D5,$7F,$53,$87,$98,$BD,$CF,$6B,$BC,$0B,$FE,$5E,$70
	.dc.b	$AC,$BF,$CC,$3E,$55,$F0,$7F,$E3,$56,$C7,$FB,$BB,$E5,$B1,$32,$3F
	.dc.b	$B3,$CE,$B1,$F2,$64,$BD,$47,$F6,$19,$D6,$3F,$E5,$F6,$DF,$92,$F8
	.dc.b	$3D,$30,$FF,$23,$7B,$1F,$DD,$A3,$FB,$AB,$70,$FE,$E9,$E2,$FF,$8F
	.dc.b	$FC,$0D,$F8,$B2,$F6,$FF,$95,$A7,$51,$82,$57,$F1,$5B,$E8,$FF,$1C
	.dc.b	$6B,$FD,$51,$AF,$F5,$8F,$55,$EB,$FF,$96,$DB,$5F,$E1,$9B,$FF,$29
	.dc.b	$7C,$1E,$6A,$4F,$F8,$3A,$B7,$7E,$AF,$3D,$7A,$8F,$F0,$44,$E9,$FE
	.dc.b	$5E,$D7,$FB,$3B,$5F,$25,$FF,$9E,$9F,$E3,$9F,$55,$FE,$6F,$EE,$3F
	.dc.b	$0F,$E3,$9F,$03,$98,$99,$FF,$88,$AB,$6F,$AE,$B6,$E6,$77,$7F,$C7
	.dc.b	$9F,$03,$FC,$D1,$B5,$CA,$3F,$A1,$7C,$EF,$EE,$AB,$FC,$FC,$D8,$CF
	.dc.b	$67,$28,$C4,$7F,$4E,$5F,$25,$FE,$89,$6D,$7F,$8E,$1E,$5C,$DF,$EA
	.dc.b	$B7,$35,$FE,$31,$6B,$CE,$BB,$12,$BF,$AB,$70,$7F,$AD,$7E,$63,$FD
	.dc.b	$2D,$F5,$4F,$A8,$FF,$5A,$46,$1B,$5F,$E4,$89,$D3,$F8,$1F,$49,$5B
	.dc.b	$8D,$55,$C7,$F3,$35,$A9,$C9,$23,$FC,$91,$7D,$1D,$DF,$1C,$FF,$D9
	.dc.b	$13,$96,$E1,$E4,$AE,$2E,$EF,$FC,$5B,$5B,$7F,$C6,$38,$FB,$DF,$05
	.dc.b	$F2,$69,$AC,$7C,$D8,$74,$8C,$BF,$E3,$AB,$E8,$FF,$99,$A7,$75,$CE
	.dc.b	$38,$11,$FE,$28,$EF,$1F,$33,$CF,$1B,$8F,$96,$38,$63,$EA,$9C,$E1
	.dc.b	$DA,$30,$79,$25,$8F,$92,$F9,$63,$E3,$25,$70,$AF,$9A,$F1,$AF,$5C
	.dc.b	$72,$C3,$E6,$57,$08,$F3,$57,$C2,$BE,$AF,$39,$F8,$13,$FE,$26,$BF
	.dc.b	$B4,$DB,$FD,$3C,$62,$38,$57,$F9,$8B,$F0,$4F,$F3,$8B,$EB,$6C,$7C
	.dc.b	$23,$FC,$D2,$E2,$FC,$90,$8C,$4F,$48,$DF,$9F,$F3,$9D,$36,$77,$1E
	.dc.b	$C3,$FD,$30,$E1,$F8,$53,$FE,$A6,$8C,$5F,$C6,$6E,$1F,$EF,$D6,$CE
	.dc.b	$EF,$B1,$1F,$63,$6F,$D8,$79,$23,$24,$73,$3D,$1F,$2C,$F1,$C3,$E0
	.dc.b	$CE,$5F,$C6,$CF,$33,$8C,$EE,$BE,$24,$66,$78,$CE,$63,$E1,$47,$2A
	.dc.b	$E9,$5F,$56,$B5,$5F,$10,$F1,$C7,$9E,$B9,$B4,$EF,$5C,$A7,$E9,$C6
	.dc.b	$5E,$EF,$53,$0F,$59,$F1,$57,$83,$0F,$A0,$DC,$3E,$83,$6F,$E8,$1C
	.dc.b	$A7,$6D,$9D,$B4,$DD,$FC,$67,$0F,$B0,$C1,$2F,$DA,$D8,$F6,$DF,$86
	.dc.b	$38,$4E,$89,$3A,$3A,$BE,$C6,$9F,$06,$3A,$1F,$18,$E6,$F1,$69,$98
	.dc.b	$EA,$D8,$E5,$18,$8F,$BD,$18,$39,$47,$86,$39,$9F,$3E,$34,$F3,$21
	.dc.b	$92,$E1,$FA,$53,$EE,$3D,$8C,$D7,$E1,$A3,$4D,$3E,$8C,$7B,$8E,$04
	.dc.b	$C3,$EB,$BE,$49,$EA,$72,$AF,$AF,$7E,$D8,$C4,$F3,$6C,$66,$FB,$B6
	.dc.b	$B8,$9F,$08,$F7,$D7,$87,$0D,$C6,$2F,$C3,$78,$AF,$87,$87,$9B,$0F
	.dc.b	$94,$4B,$E9,$3F,$09,$FC,$DD,$F9,$1C,$C7,$C0,$8E,$B1,$E4,$8E,$B8
	.dc.b	$77,$75,$87,$C9,$9D,$47,$D3,$32,$E3,$0F,$23,$E9,$9E,$13,$F7,$A7
	.dc.b	$2F,$94,$F9,$11,$DE,$33,$1D,$0C,$CF,$5A,$F2,$5F,$AE,$7F,$72,$73
	.dc.b	$C3,$E1,$38,$39,$38,$7D,$11,$C5,$FA,$15,$EC,$8F,$E7,$F1,$FE,$EE
	.dc.b	$6B,$FA,$53,$18,$7F,$8C,$23,$F2,$5B,$79,$09,$5C,$DB,$5F,$8B,$6D
	.dc.b	$F8,$EB,$FC,$53,$5F,$7E,$72,$D3,$91,$2B,$93,$FD,$0E,$3F,$A0,$1C
	.dc.b	$23,$FA,$AB,$88,$E1,$5F,$CA,$E3,$13,$C8,$FE,$9E,$E2,$3C,$8D,$9F
	.dc.b	$EC,$8F,$59,$FE,$BC,$E5,$FD,$9B,$D1,$F9,$0F,$C5,$24,$22,$1D,$1C
	.dc.b	$4F,$F4,$8F,$BF,$78,$3D,$69,$18,$AD,$9C,$BF,$55,$FE,$F5,$3D,$5A
	.dc.b	$9A,$63,$17,$BD,$F6,$49,$5E,$4A,$FF,$01,$47,$F9,$B3,$C2,$7C,$25
	.dc.b	$25,$6E,$BD,$93,$A8,$C4,$3E,$45,$F2,$3F,$E7,$66,$CE,$9B,$3C,$DA
	.dc.b	$F8,$FE,$39,$F1,$DF,$E7,$E6,$D7,$66,$CF,$8D,$AF,$AF,$E0,$5F,$05
	.dc.b	$F3,$23,$D1,$18,$E0,$F3,$AF,$81,$FF,$6E,$4E,$AB,$83,$F1,$7F,$56
	.dc.b	$7C,$DF,$DA,$E4,$FF,$25,$F4,$BA,$9F,$49,$9D,$BF,$CB,$4D,$C7,$F0
	.dc.b	$58,$F2,$7E,$D9,$8C,$75,$87,$3A,$B8,$74,$AD,$8B,$E4,$B1,$FF,$08
	.dc.b	$C6,$23,$D1,$87,$D4,$78,$36,$35,$1F,$94,$5D,$3E,$0F,$98,$93,$E0
	.dc.b	$9C,$B6,$72,$FF,$A3,$27,$DC,$F9,$62,$CF,$08,$FC,$43,$5F,$99,$1E
	.dc.b	$33,$CD,$F0,$8F,$9F,$E7,$F0,$2F,$8C,$A7,$69,$D5,$FC,$46,$BF,$0F
	.dc.b	$7F,$D7,$D7,$18,$F9,$2D,$74,$7E,$D6,$78,$E9,$FE,$3A,$6D,$72,$BF
	.dc.b	$0D,$77,$C3,$DF,$F6,$4F,$67,$CF,$F3,$52,$BE,$BC,$E5,$B3,$97,$C2
	.dc.b	$33,$57,$49,$93,$E8,$8E,$19,$FF,$93,$AF,$A4,$F2,$F8,$57,$C3,$FC
	.dc.b	$8F,$87,$93,$F8,$89,$CB,$67,$2F,$8B,$25,$7D,$AA,$DC,$F5,$AD,$E1
	.dc.b	$FE,$5E,$BD,$CF,$33,$B7,$B3,$F6,$3D,$59,$18,$21,$DD,$CB,$E9,$8F
	.dc.b	$C4,$FC,$3B,$E1,$AF,$1C,$AF,$D0,$CE,$5B,$3C,$08,$FE,$9B,$E1,$5F
	.dc.b	$0F,$FA,$58,$7C,$1B,$E1,$3F,$99,$D6,$B5,$FE,$6D,$39,$AF,$E7,$18
	.dc.b	$F7,$78,$DF,$F3,$A8,$C4,$EE,$57,$CD,$D9,$FD,$05,$C4,$6E,$78,$D7
	.dc.b	$3C,$7F,$A2,$CF,$3A,$DC,$F9,$B1,$F5,$4F,$1A,$DC,$F1,$C3,$D2,$EF
	.dc.b	$1B,$1E,$1B,$E3,$3B,$8F,$28,$C7,$C6,$EE,$B8,$BE,$CA,$E2,$FC,$4B
	.dc.b	$F0,$B4,$DE,$3E,$87,$91,$DA,$FA,$BD,$E7,$8B,$D6,$F8,$DF,$68,$D1
	.dc.b	$E2,$AE,$37,$D6,$38,$1F,$D4,$DF,$A0,$FF,$5B,$6B,$F0,$CD,$7C,$3B
	.dc.b	$E1,$FF,$AB,$CA,$31,$3A,$9C,$C6,$23,$FA,$EF,$A9,$FC,$AD,$34,$8F
	.dc.b	$F2,$F7,$E4,$FC,$2B,$E1,$DD,$7F,$32,$BF,$C0,$95,$A9,$CC,$7D,$43
	.dc.b	$27,$D6,$AE,$55,$C6,$3C,$26,$67,$8E,$FF,$93,$8E,$2F,$E5,$6B,$8C
	.dc.b	$75,$38,$C6,$67,$8E,$FF,$DF,$70,$F2,$61,$DD,$B7,$C5,$B7,$DE,$39
	.dc.b	$D7,$F8,$7B,$0E,$F3,$9A,$DC,$F5,$AF,$45,$F5,$69,$D4,$B5,$E5,$8F
	.dc.b	$EC,$37,$EA,$DB,$FD,$DF,$5E,$EB,$F0,$63,$F2,$A7,$FD,$E7,$5B,$BC
	.dc.b	$9F,$DA,$AF,$81,$FE,$35,$8D,$37,$A7,$94,$FA,$DB,$0F,$2D,$F0,$FF
	.dc.b	$E5,$A8,$E1,$5E,$6A,$DC,$2F,$FA,$34,$FF,$80,$2F,$8F,$A5,$A9,$CD
	.dc.b	$7F,$44,$5F,$0D,$93,$C3,$4F,$86,$7C,$3E,$71,$F8,$B5,$70,$FE,$BB
	.dc.b	$7B,$AC,$69,$F1,$2A,$DF,$F8,$EB,$0E,$6F,$23,$FC,$1F,$E6,$3A,$7D
	.dc.b	$6D,$8F,$F0,$C8,$F8,$5F,$88,$F0,$9D,$91,$DC,$6B,$3F,$EC,$35,$CD
	.dc.b	$B5,$B7,$FB,$AD,$75,$CF,$19,$7E,$9A,$5F,$0C,$F8,$6F,$9E,$DB,$EF
	.dc.b	$F0,$F8,$99,$1F,$E2,$A9,$D5,$76,$CD,$7A,$8F,$EE,$B3,$AC,$3C,$0D
	.dc.b	$BF,$F3,$E1,$F0,$6B,$84,$F9,$7E,$19,$F0,$FF,$82,$6D,$7F,$90,$3C
	.dc.b	$F7,$C3,$BC,$2D,$5C,$AF,$F3,$5D,$F0,$CF,$FC,$E0,$D7,$FC,$F0,$D7
	.dc.b	$FD,$1A,$EF,$1F,$F2,$AE,$0F,$9B,$E9,$5F,$E2,$0C,$BC,$31,$F9,$6F
	.dc.b	$87,$7C,$3D,$3F,$3F,$3E,$6F,$88,$C6,$7F,$E5,$2A,$DB,$FE,$80,$6E
	.dc.b	$1F,$E8,$56,$BF,$E8,$D6,$BF,$EA,$27,$79,$7F,$9A,$BD,$37,$C8,$76
	.dc.b	$AF,$34,$62,$FF,$CC,$3E,$0B,$E1,$FF,$5A,$D8,$F9,$9F,$10,$F8,$9C
	.dc.b	$D4,$9F,$E7,$2F,$F0,$8C,$3F,$98,$B5,$FE,$76,$DB,$E9,$1F,$E3,$6F
	.dc.b	$86,$FC,$B6,$7F,$91,$9C,$BF,$93,$8C,$DF,$FA,$FB,$E2,$5F,$13,$FC
	.dc.b	$09,$B1,$FE,$7A,$7F,$84,$60,$72,$FF,$91,$1A,$FF,$AD,$5A,$F5,$9F
	.dc.b	$44,$62,$3F,$D9,$CF,$2B,$FD,$37,$CD,$FF,$9A,$6F,$C1,$7F,$C5,$23
	.dc.b	$17,$ED,$F8,$9F,$E2,$F8,$9F,$C0,$EC,$F8,$AC,$C4,$C7,$FD,$7F,$5F
	.dc.b	$ED,$02,$47,$CB,$79,$47,$FA,$63,$E2,$3F,$D1,$AF,$FA,$6A,$33,$87
	.dc.b	$F9,$8A,$72,$F8,$E7,$87,$A5,$DE,$34,$DD,$3F,$81,$57,$07,$E9,$61
	.dc.b	$FC,$C2,$31,$7F,$EF,$0C,$3B,$D7,$0B,$E5,$3F,$EF,$4A,$DE,$1F,$EF
	.dc.b	$7B,$DC,$E6,$B7,$3F,$EF,$BA,$DA,$AD,$B9,$9D,$A9,$DC,$EA,$31,$3A
	.dc.b	$9C,$E1,$E9,$8C,$12,$FE,$44,$65,$A7,$C6,$2D,$7C,$66,$E1,$EF,$69
	.dc.b	$EB,$2C,$79,$DA,$F9,$1B,$7E,$47,$B4,$75,$24,$FD,$08,$CB,$DE,$33
	.dc.b	$1F,$46,$BD,$11,$83,$C0,$7D,$F3,$07,$84,$B8,$7C,$7B,$F9,$E4,$C3
	.dc.b	$E2,$4F,$08,$F3,$36,$3B,$E1,$E5,$8C,$61,$E6,$3C,$85,$7C,$71,$E9
	.dc.b	$38,$16,$3E,$44,$79,$30,$EB,$18,$9D,$4F,$E2,$67,$2F,$49,$D1,$E0
	.dc.b	$7C,$67,$37,$C4,$FC,$19,$E6,$FE,$36,$79,$B9,$9E,$6F,$77,$8D,$FC
	.dc.b	$57,$EC,$DF,$08,$F8,$B7,$D1,$D5,$7D,$18,$C4,$F6,$C3,$E8,$3C,$4F
	.dc.b	$9F,$1A,$BE,$A7,$82,$32,$F9,$2B,$CC,$79,$4F,$7C,$FB,$2B,$AC,$F1
	.dc.b	$39,$CE,$9F,$8D,$3D,$E3,$E3,$19,$3E,$61,$D5,$D1,$D7,$0F,$44,$FC
	.dc.b	$CA,$EF,$5C,$4F,$99,$1E,$17,$AD,$72,$8F,$54,$F0,$8C,$4F,$79,$F7
	.dc.b	$61,$F2,$A7,$F0,$F1,$8C,$3E,$56,$1E,$2B,$EB,$39,$0C,$3F,$1A,$53
	.dc.b	$F4,$45,$BF,$E4,$AD,$AF,$EB,$11,$88,$E2,$43,$E2,$9F,$C4,$7C,$7F
	.dc.b	$C1,$12,$BF,$0E,$D8,$FD,$53,$70,$FC,$33,$67,$DE,$D9,$EB,$5F,$D6
	.dc.b	$8F,$DA,$C6,$63,$F0,$27,$86,$7B,$91,$EE,$DA,$EE,$D3,$B9,$63,$BC
	.dc.b	$F5,$AE,$F3,$99,$EF,$1E,$3B,$FE,$85,$5B,$6D,$6E,$78,$D7,$88,$97
	.dc.b	$E2,$6C,$F8,$9A,$78,$9E,$55,$D2,$32,$78,$AF,$C1,$3F,$D8,$5B,$1F
	.dc.b	$D7,$DB,$5F,$D6,$9B,$3F,$D4,$DB,$3F,$D3,$DA,$FC,$B3,$BD,$FA,$AB
	.dc.b	$D0,$4A,$E4,$78,$4F,$81,$5E,$78,$D5,$75,$9D,$3D,$27,$51,$F0,$67
	.dc.b	$CC,$E9,$E3,$3E,$37,$2F,$18,$FC,$D3,$D4,$B4,$AF,$CC,$92,$7F,$47
	.dc.b	$5F,$E7,$DF,$8A,$78,$9E,$2C,$3F,$4F,$18,$AF,$11,$FA,$97,$11,$F3
	.dc.b	$63,$99,$99,$E5,$09,$DD,$7A,$0F,$04,$EA,$33,$5F,$2A,$95,$A8,$F0
	.dc.b	$C7,$92,$B8,$12,$33,$1D,$E3,$FD,$87,$5D,$23,$84,$EA,$31,$1B,$8F
	.dc.b	$4B,$A3,$D8,$78,$9F,$61,$A7,$81,$93,$85,$7B,$63,$71,$93,$71,$EB
	.dc.b	$7A,$B4,$91,$C1,$E5,$1D,$23,$04,$7E,$F4,$70,$74,$43,$F0,$D1,$B7
	.dc.b	$CE,$7F,$1F,$9F,$B2,$DA,$FF,$3A,$46,$23,$87,$EA,$FF,$E5,$3A,$F8
	.dc.b	$13,$F2,$23,$13,$FE,$4B,$F8,$AF,$E4,$A4,$AF,$C9,$C7,$F8,$CE,$3F
	.dc.b	$D0,$1E,$2B,$E2,$A9,$2B,$71,$97,$8F,$67,$FA,$4E,$FF,$3A,$D7,$F9
	.dc.b	$C3,$5F,$E7,$8D,$7F,$9F,$FC,$3B,$E2,$3F,$D2,$0D,$AF,$2B,$67,$F5
	.dc.b	$2D,$7F,$A2,$7C,$43,$E1,$FF,$4B,$1F,$AE,$8C,$6D,$EE,$5F,$10,$CF
	.dc.b	$17,$C2,$FF,$A7,$7C,$77,$FB,$49,$F8,$AD,$7F,$72,$D7,$7E,$2F,$CA
	.dc.b	$5F,$0C,$F8,$8F,$EB,$5C,$9F,$EB,$BF,$89,$BB,$FF,$5E,$E1,$76,$27
	.dc.b	$38,$FF,$62,$C7,$93,$F9,$B6,$BF,$C2,$9A,$FF,$65,$6B,$FD,$A5,$AF
	.dc.b	$3C,$7A,$BE,$66,$C7,$4C,$58,$FF,$72,$F9,$07,$9C,$EF,$DF,$67,$9B
	.dc.b	$5F,$4F,$C5,$5D,$8E,$B5,$F4,$57,$C5,$D0,$D2,$47,$E4,$A3,$2F,$02
	.dc.b	$1F,$EE,$0D,$FD,$3F,$14,$F8,$9F,$5F,$C2,$3E,$2A,$9A,$89,$7F,$C0
	.dc.b	$E7,$4D,$9C,$BF,$C0,$6B,$A3,$92,$19,$AF,$93,$F1,$2F,$73,$C3,$F1
	.dc.b	$52,$BF,$88,$4E,$5B,$39,$7F,$87,$62,$AE,$73,$27,$CD,$1B,$E4,$FA
	.dc.b	$71,$FE,$24,$BE,$8F,$F1,$7F,$89,$7C,$4F,$C2,$F8,$A9,$3F,$C8,$67
	.dc.b	$2D,$9C,$BE,$29,$FC,$1C,$EA,$3A,$4E,$AF,$C3,$0C,$97,$C2,$B8,$FB
	.dc.b	$9F,$4B,$D9,$97,$E1,$8C,$BF,$46,$32,$FA,$63,$F2,$7F,$10,$F8,$AB
	.dc.b	$C9,$2B,$F9,$D4,$E5,$B3,$C0,$8F,$F3,$FF,$89,$7C,$57,$F5,$48,$7C
	.dc.b	$43,$E2,$3F,$C9,$95,$AD,$FF,$C9,$F3,$9A,$FF,$28,$E3,$DD,$E3,$7F
	.dc.b	$E5,$38,$C4,$EE,$57,$CD,$D9,$FE,$5A,$71,$1B,$9E,$35,$CF,$1F,$F2
	.dc.b	$F4,$F3,$AD,$CF,$9B,$1F,$54,$F1,$AD,$CF,$1C,$3D,$2E,$F1,$B1,$E1
	.dc.b	$BE,$33,$B8,$F2,$8C,$7C,$6E,$EB,$8B,$EC,$AE,$2F,$C4,$BF,$0B,$4D
	.dc.b	$E3,$E8,$79,$1D,$AF,$AB,$DE,$78,$BD,$6F,$8D,$F6,$8D,$1E,$2A,$E3
	.dc.b	$7D,$63,$81,$FE,$2A,$7E,$83,$FE,$85,$6B,$F0,$CD,$7C,$4B,$E2,$BF
	.dc.b	$AF,$CA,$31,$3A,$9C,$C6,$23,$FB,$2F,$AF,$FC,$F5,$34,$8F,$F4,$17
	.dc.b	$E4,$FC,$43,$E2,$BF,$73,$B7,$E1,$EB,$FD,$3B,$5A,$9C,$ED,$FE,$44
	.dc.b	$8C,$78,$9F,$4A,$B6,$7D,$8C,$3A,$46,$63,$C6,$66,$F9,$BF,$EA,$7A
	.dc.b	$E6,$FE,$00,$E7,$87,$E6,$6F,$9C,$64,$E7,$19,$AE,$71,$FE,$B3,$AE
	.dc.b	$73,$EB,$8D,$E1,$FE,$B7,$BE,$73,$9A,$DC,$FB,$6B,$CD,$87,$A6,$73
	.dc.b	$1E,$73,$24,$DB,$F8,$8E,$1E,$67,$F8,$9C,$7A,$AF,$C1,$8F,$F2,$69
	.dc.b	$DB,$70,$C9,$FE,$43,$AE,$07,$FB,$3A,$FA,$47,$A7,$A7,$84,$E5,$BE
	.dc.b	$4B,$E2,$BF,$CE,$D5,$FE,$4D,$BF,$E8,$4D,$AE,$36,$9C,$BF,$E5,$1B
	.dc.b	$E5,$EA,$6A,$73,$97,$BB,$E2,$70,$78,$3E,$25,$F1,$58,$CF,$C1,$37
	.dc.b	$2E,$95,$A9,$C9,$23,$FC,$B9,$82,$BA,$5F,$1C,$3F,$9E,$FA,$2E,$9F
	.dc.b	$4B,$71,$FF,$37,$8F,$89,$FF,$14,$3F,$E6,$D8,$D6,$5F,$E6,$96,$DF
	.dc.b	$4B,$F9,$75,$C7,$5E,$B1,$23,$93,$AC,$3A,$E1,$CE,$3E,$5F,$96,$F8
	.dc.b	$AA,$6F,$E2,$AE,$1F,$E7,$F9,$D3,$67,$4D,$25,$FF,$9F,$E7,$6F,$FA
	.dc.b	$16,$3D,$D5,$D2,$3F,$D0,$9F,$12,$F8,$AF,$C8,$C0,$46,$6B,$FC,$D6
	.dc.b	$D8,$FF,$5F,$B5,$E0,$71,$75,$A7,$F7,$EF,$39,$E5,$BE,$2B,$F1,$75
	.dc.b	$AC,$3F,$C9,$13,$97,$FD,$3D,$19,$C1,$7C,$AF,$8E,$3F,$9C,$8C,$5F
	.dc.b	$FA,$97,$E2,$1F,$15,$EC,$C1,$18,$9F,$F4,$2B,$63,$FD,$B2,$D7,$D6
	.dc.b	$70,$9E,$DE,$8F,$FA,$0F,$4F,$F0,$4D,$71,$7D,$D1,$9B,$F8,$73,$97
	.dc.b	$FC,$C9,$19,$BF,$9D,$39,$7D,$B1,$97,$F9,$9F,$C3,$7A,$7B,$8D,$25
	.dc.b	$FF,$3A,$7F,$6B,$F0,$7F,$96,$8F,$E3,$91,$8E,$39,$1F,$EA,$49,$FC
	.dc.b	$ED,$FB,$9A,$FD,$A6,$BF,$83,$F8,$67,$C3,$7C,$B6,$D7,$FA,$B9,$B3
	.dc.b	$F2,$03,$87,$E5,$BA,$D3,$FC,$63,$EB,$7F,$97,$6B,$FD,$B9,$39,$7F
	.dc.b	$17,$19,$AF,$F6,$D3,$5F,$AA,$D7,$7E,$A7,$DF,$8C,$47,$E6,$F1,$7C
	.dc.b	$2F,$F1,$AF,$85,$FF,$3B,$E1,$FE,$EB,$9C,$BF,$EE,$D8,$CD,$7F,$BA
	.dc.b	$5B,$97,$F9,$21,$B3,$F1,$EF,$A7,$9B,$23,$93,$F5,$23,$17,$CE,$2D
	.dc.b	$FF,$94,$B0,$F0,$CE,$6F,$89,$1F,$F7,$D4,$6A,$7C,$55,$CA,$BF,$63
	.dc.b	$87,$92,$FE,$06,$19,$7F,$79,$19,$4D,$C3,$A5,$72,$BE,$15,$A7,$D5
	.dc.b	$1E,$3C,$3C,$2D,$3E,$11,$67,$BC,$F2,$AE,$B1,$89,$D3,$D5,$FA,$58
	.dc.b	$78,$30,$F3,$1E,$57,$17,$CD,$B3,$CA,$FA,$DF,$33,$E5,$13,$0F,$9D
	.dc.b	$5C,$23,$F2,$D8,$7D,$E7,$F7,$ED,$7E,$15,$FE,$6D,$E9,$5F,$81,$F0
	.dc.b	$BC,$C4,$9F,$5D,$7F,$6A,$AF,$04,$7D,$49,$D4,$60,$FA,$6E,$2B,$E5
	.dc.b	$C7,$E3,$E3,$13,$E2,$8C,$3D,$A3,$31,$F2,$CC,$92,$7C,$D3,$B3,$F3
	.dc.b	$93,$FA,$13,$8C,$FE,$09,$C4,$6E,$39,$3E,$93,$13,$D2,$32,$78,$5C
	.dc.b	$12,$38,$47,$BC,$EC,$F8,$88,$E6,$BE,$44,$F1,$8F,$8B,$3E,$03,$9C
	.dc.b	$66,$3E,$9C,$60,$8F,$03,$A5,$7B,$1F,$39,$C0,$D4,$F3,$AF,$0D,$7C
	.dc.b	$F8,$C5,$71,$7E,$A9,$88,$F4,$B8,$7C,$35,$EE,$3B,$D7,$E6,$AB,$94
	.dc.b	$EA,$30,$F2,$31,$8E,$AB,$CB,$18,$34,$E0,$87,$AC,$F0,$5F,$95,$F8
	.dc.b	$31,$C1,$CB,$ED,$3C,$57,$DE,$39,$5F,$C2,$7A,$B5,$CC,$6E,$30,$7A
	.dc.b	$A3,$DF,$1F,$8C,$8C,$B4,$E1,$7C,$5B,$5E,$59,$DB,$F7,$EB,$ED,$1F
	.dc.b	$B4,$9F,$EF,$53,$FA,$D8,$C4,$FE,$D6,$31,$1F,$90,$32,$7E,$FA,$78
	.dc.b	$9F,$E5,$AF,$85,$77,$FC,$16,$DF,$C7,$6D,$7E,$21,$B8,$FE,$45,$B3
	.dc.b	$F8,$06,$CF,$D4,$6B,$F4,$0F,$F1,$63,$8C,$3E,$63,$70,$F2,$61,$E3
	.dc.b	$25,$F8,$DB,$7E,$36,$E1,$E3,$6C,$F8,$DB,$3E,$36,$BE,$31,$7E,$19
	.dc.b	$FF,$20,$51,$FF,$1D,$C6,$65,$39,$7D,$75,$D3,$0F,$5B,$70,$F5,$E1
	.dc.b	$E9,$25,$7A,$5B,$7E,$96,$9E,$92,$C7,$A5,$AF,$A7,$0F,$24,$F8,$E3
	.dc.b	$D7,$7E,$1A,$F5,$9E,$A7,$11,$EA,$8C,$1D,$30,$F4,$B7,$0F,$4C,$F9
	.dc.b	$C9,$7E,$76,$CF,$9D,$A7,$9C,$A7,$9E,$F9,$CF,$9F,$0F,$4D,$78,$AB
	.dc.b	$C8,$7C,$C7,$D2,$D3,$9C,$7E,$30,$E2,$E6,$7A,$57,$6A,$F2,$46,$1E
	.dc.b	$73,$F1,$9F,$97,$5C,$EB,$F0,$AD,$8C,$C7,$F9,$A6,$3F,$47,$3F,$E6
	.dc.b	$46,$BE,$93,$11,$FE,$C6,$F8,$57,$C3,$78,$E7,$FD,$19,$18,$BF,$1D
	.dc.b	$FE,$9A,$31,$3F,$08,$F3,$15,$CE,$1E,$B8,$C5,$F5,$53,$AB,$F9,$95
	.dc.b	$F8,$58,$C4,$FC,$79,$FF,$4B,$C6,$23,$BC,$EE,$BA,$4E,$6F,$CD,$18
	.dc.b	$25,$FA,$B0,$F8,$11,$23,$E7,$D7,$19,$C6,$1E,$89,$CD,$72,$8C,$13
	.dc.b	$0F,$71,$E3,$9D,$92,$3E,$36,$1E,$48,$C4,$EE,$73,$7E,$08,$C1,$27
	.dc.b	$E2,$9F,$0A,$FC,$B5,$FE,$A0,$C3,$E9,$B6,$3F,$D5,$51,$88,$E1,$3F
	.dc.b	$96,$E8,$73,$7F,$BD,$E7,$D5,$3F,$EA,$0F,$0B,$F8,$46,$3F,$B7,$F0
	.dc.b	$9F,$0B,$FE,$B5,$6D,$7F,$8B,$DB,$7F,$E8,$B6,$BF,$EB,$4C,$5F,$13
	.dc.b	$FE,$BD,$6D,$7F,$8D,$DB,$3F,$E9,$36,$BE,$DF,$03,$D6,$49,$FF,$1F
	.dc.b	$69,$FE,$D6,$7F,$C8,$1F,$08,$F8,$3F,$E8,$F7,$FE,$9D,$8C,$70,$6B
	.dc.b	$E1,$1A,$70,$BD,$DF,$FB,$2B,$E1,$3F,$DC,$2F,$F9,$2D,$AF,$FA,$9D
	.dc.b	$AE,$F3,$FF,$28,$2B,$72,$61,$B1,$FE,$55,$6B,$FE,$B1,$6B,$FE,$DA
	.dc.b	$F8,$07,$C1,$F4,$6C,$7F,$97,$5A,$FF,$AE,$5A,$F1,$8F,$F7,$EF,$17
	.dc.b	$F2,$F9,$DB,$63,$3B,$29,$F0,$6F,$25,$DB,$A9,$D6,$3C,$AF,$84,$23
	.dc.b	$0E,$B0,$E1,$87,$3B,$E5,$3B,$EB,$F6,$11,$F6,$35,$FE,$19,$68,$E2
	.dc.b	$D8,$FC,$D7,$47,$D4,$CF,$DD,$5A,$24,$EB,$6F,$0F,$C1,$3E,$0B,$C7
	.dc.b	$EF,$BC,$1A,$6A,$25,$FF,$32,$9D,$36,$72,$FF,$21,$AE,$8E,$48,$66
	.dc.b	$BE,$0F,$C0,$BD,$3F,$0F,$C1,$4A,$FE,$7F,$39,$6C,$E5,$FE,$5D,$8A
	.dc.b	$B9,$CC,$9F,$34,$6F,$4F,$ED,$D7,$CE,$7F,$9B,$7C,$13,$E0,$7E,$0F
	.dc.b	$C1,$49,$FE,$9F,$39,$6C,$E5,$F0,$CF,$DA,$9D,$47,$49,$D5,$FF,$7D
	.dc.b	$C9,$7B,$9E,$1E,$57,$D0,$F8,$29,$87,$82,$32,$FC,$F8,$CB,$E8,$8F
	.dc.b	$C4,$FC,$03,$E0,$AF,$1C,$AF,$EC,$F3,$96,$CF,$02,$3F,$DB,$3E,$05
	.dc.b	$F0,$5F,$E2,$68,$7C,$03,$E0,$3F,$3B,$5A,$DF,$F4,$D3,$9A,$FD,$46
	.dc.b	$3D,$DE,$37,$FA,$48,$C4,$EE,$57,$CD,$D9,$FB,$17,$11,$B9,$E3,$5C
	.dc.b	$F1,$FD,$7C,$F3,$AD,$CF,$9B,$1F,$54,$F1,$AD,$CF,$1C,$3D,$2E,$F1
	.dc.b	$B1,$E1,$BE,$33,$B8,$F2,$8C,$7C,$6E,$EB,$8B,$EC,$AE,$2F,$C4,$BF
	.dc.b	$0B,$4D,$E3,$E8,$79,$1D,$AF,$AB,$DE,$78,$BD,$6F,$8D,$F6,$8D,$1E
	.dc.b	$2A,$E3,$7D,$63,$81,$FC,$58,$EE,$57,$E9,$36,$7F,$04,$E7,$E0,$5F
	.dc.b	$05,$C6,$3E,$B4,$61,$C5,$23,$11,$C6,$7F,$A1,$AF,$56,$91,$FE,$8E
	.dc.b	$DD,$3F,$0D,$EF,$BE,$0B,$FC,$F1,$1F,$89,$AF,$E6,$D5,$F4,$2B,$5E
	.dc.b	$17,$D2,$9C,$9F,$62,$32,$FD,$A3,$35,$DA,$FA,$EF,$ED,$C3,$A4,$62
	.dc.b	$BE,$44,$63,$0E,$D1,$EB,$3B,$4E,$DF,$E8,$73,$B7,$C0,$6E,$FC,$15
	.dc.b	$CA,$75,$1E,$D9,$D5,$FF,$4C,$C3,$D3,$39,$AD,$CF,$9E,$BC,$37,$CD
	.dc.b	$A4,$DB,$FD,$1D,$87,$C2,$7F,$D2,$11,$F0,$AB,$B6,$3F,$E1,$EB,$A7
	.dc.b	$F9,$C2,$F8,$3D,$F1,$F4,$DF,$F9,$7E,$7F,$9D,$4E,$FA,$79,$4F,$F8
	.dc.b	$33,$E0,$BF,$DE,$55,$A8,$E1,$3A,$51,$9C,$BF,$D6,$DE,$4A,$F8,$DE
	.dc.b	$E9,$F0,$34,$EF,$4F,$82,$FF,$1D,$61,$F0,$27,$59,$7A,$DD,$CE,$A3
	.dc.b	$E7,$46,$B2,$F6,$3A,$F4,$B1,$7F,$EB,$7F,$35,$D1,$E9,$6E,$1F,$EA
	.dc.b	$01,$F0,$5F,$E2,$C7,$FD,$33,$1A,$C7,$FD,$0C,$DB,$E9,$7F,$22,$B8
	.dc.b	$E7,$D6,$24,$72,$75,$87,$5C,$39,$C7,$CB,$F2,$9F,$05,$4D,$FC,$15
	.dc.b	$C3,$FD,$57,$3A,$6C,$E9,$A4,$BF,$F4,$E4,$ED,$FF,$5F,$C7,$BA,$BA
	.dc.b	$47,$FA,$87,$E0,$9F,$07,$F7,$5B,$5F,$E6,$FF,$27,$FA,$63,$FE,$BC
	.dc.b	$8C,$D7,$1A,$E0,$7E,$2B,$D5,$5F,$09,$D5,$F1,$AF,$CD,$C7,$F8,$E6
	.dc.b	$F9,$BF,$C7,$DA,$FF,$8D,$9A,$FF,$8F,$FE,$03,$FC,$41,$5F,$EC,$E6
	.dc.b	$C7,$FB,$73,$0F,$0F,$C3,$7E,$5E,$73,$C1,$FE,$F9,$6D,$FF,$90,$9B
	.dc.b	$6B,$E3,$3D,$7C,$CE,$47,$A2,$55,$E0,$7F,$9B,$35,$E9,$1B,$F7,$FF
	.dc.b	$CD,$75,$FC,$CE,$51,$87,$F9,$B3,$5F,$E8,$0D,$7F,$CA,$EB,$E0,$3F
	.dc.b	$64,$D9,$FF,$77,$BC,$F1,$47,$09,$D9,$E3,$23,$E3,$9E,$55,$FD,$1F
	.dc.b	$E0,$1F,$01,$EF,$6D,$7A,$5B,$3B,$9F,$89,$F0,$1F,$E5,$FA,$FF,$80
	.dc.b	$2D,$9F,$F7,$75,$70,$C7,$CB,$1F,$E7,$99,$D4,$62,$11,$C5,$FE,$C4
	.dc.b	$D7,$FC,$F2,$D7,$97,$91,$FE,$C8,$F9,$6F,$E2,$DE,$E9,$FC,$A7,$A8
	.dc.b	$F8,$2C,$C4,$CF,$FD,$66,$F1,$69,$30,$65,$E5,$F8,$EF,$54,$FF,$BF
	.dc.b	$2B,$F9,$24,$62,$BF,$D2,$98,$F8,$6B,$FD,$2D,$7C,$75,$F0,$57,$49
	.dc.b	$FF,$49,$DF,$18,$5F,$17,$47,$FA,$7E,$35,$19,$BE,$95,$D8,$F2,$16
	.dc.b	$B8,$5F,$82,$77,$1D,$0E,$F8,$7A,$A7,$55,$E4,$6C,$F2,$8E,$8F,$06
	.dc.b	$E1,$E9,$24,$F0,$94,$62,$77,$39,$8C,$4F,$8D,$F6,$D7,$8E,$B6,$49
	.dc.b	$F4,$61,$F0,$63,$CC,$67,$0F,$69,$F8,$B8,$D4,$FA,$4E,$D8,$75,$79
	.dc.b	$DD,$AE,$37,$D2,$72,$F5,$30,$FF,$8D,$70,$FD,$1B,$6F,$F8,$34,$62
	.dc.b	$39,$57,$E7,$74,$7C,$4F,$7C,$3F,$36,$DA,$FD,$23,$5F,$D1,$B6,$78
	.dc.b	$CF,$8E,$76,$D7,$6D,$8D,$C7,$C0,$8F,$DA,$CF,$32,$3B,$6C,$6C,$FC
	.dc.b	$34,$F2,$AF,$54,$FF,$80,$E3,$15,$EE,$3D,$67,$62,$57,$3A,$E3,$18
	.dc.b	$6C,$66,$38,$BE,$0A,$EC,$79,$27,$64,$9F,$04,$FC,$23,$83,$D4,$CB
	.dc.b	$F1,$0C,$92,$3E,$FB,$C5,$E8,$D8,$E3,$5E,$E8,$C1,$0F,$CE,$E1,$E6
	.dc.b	$9F,$04,$67,$0F,$DE,$57,$18,$F1,$5F,$67,$8D,$70,$72,$FC,$28,$F2
	.dc.b	$C6,$0F,$04,$F3,$AE,$B8,$78,$67,$D9,$7E,$88,$C5,$F7,$AD,$CF,$7B
	.dc.b	$E5,$86,$A3,$C0,$4C,$3F,$51,$7E,$7A,$F5,$1C,$27,$96,$1C,$6B,$71
	.dc.b	$C5,$F4,$13,$0F,$0B,$E6,$3E,$36,$1C,$5D,$DF,$57,$46,$2F,$F0,$57
	.dc.b	$D6,$75,$87,$96,$79,$61,$DA,$31,$5F,$9D,$3C,$E5,$BE,$B7,$CA,$3A
	.dc.b	$5F,$2C,$3C,$93,$FA,$AA,$FE,$A7,$3F,$D8,$09,$5F,$DF,$67,$53,$F9
	.dc.b	$A3,$71,$99,$DC,$7F,$31,$ED,$7C,$17,$C1,$6D,$7E,$0D,$B3,$F0,$9B
	.dc.b	$8F,$E2,$5B,$3F,$3D,$B3,$F9,$56,$BF,$49,$B8,$7E,$4D,$B8,$7F,$9F
	.dc.b	$5B,$5F,$CF,$9B,$5E,$7C,$3C,$C4,$9F,$33,$63,$CC,$DC,$3C,$CD,$9F
	.dc.b	$33,$67,$CC,$D7,$CC,$DC,$3C,$CD,$C3,$CC,$DA,$F3,$36,$BC,$D1,$E4
	.dc.b	$23,$E4,$6D,$F9,$1A,$79,$0B,$1E,$46,$BE,$46,$DF,$91,$B7,$E4,$69
	.dc.b	$E4,$2C,$F9,$0F,$11,$2B,$C4,$D8,$F1,$34,$0C,$3E,$D9,$F5,$27,$FC
	.dc.b	$50,$E3,$0F,$14,$F1,$38,$38,$9E,$E7,$07,$15,$E0,$8C,$3E,$99,$E5
	.dc.b	$5D,$63,$E1,$B6,$33,$5F,$E6,$58,$FC,$BC,$7A,$0F,$4B,$87,$0D,$73
	.dc.b	$1F,$91,$E3,$F1,$C7,$06,$BF,$05,$B3,$FE,$78,$8F,$6D,$7E,$5E,$72
	.dc.b	$73,$70,$7A,$EF,$E3,$C6,$67,$F6,$4F,$09,$DC,$61,$E0,$7A,$23,$84
	.dc.b	$7B,$4F,$24,$7F,$9F,$6B,$FD,$05,$1E,$98,$C4,$FF,$BA,$79,$5F,$05
	.dc.b	$E1,$C3,$FC,$C1,$18,$AF,$0C,$62,$75,$1F,$2E,$75,$7E,$53,$FC,$74
	.dc.b	$E2,$BD,$B3,$FE,$36,$8C,$3C,$6B,$C4,$74,$AF,$C6,$CE,$9B,$1A,$9F
	.dc.b	$19,$C4,$E8,$E6,$3C,$55,$E7,$9C,$9F,$E9,$FA,$F3,$CF,$C7,$6B,$CC
	.dc.b	$F9,$47,$FA,$C6,$BF,$D0,$B5,$FE,$83,$F0,$DF,$55,$DE,$30,$EB,$0F
	.dc.b	$C7,$7C,$73,$E3,$FF,$D6,$CD,$8F,$F5,$83,$71,$EE,$D7,$ED,$7C,$63
	.dc.b	$B3,$FD,$7C,$DA,$FF,$5C,$36,$7C,$ED,$7F,$05,$F0,$CF,$86,$78,$FF
	.dc.b	$EB,$F8,$C3,$7E,$13,$F7,$55,$F0,$27,$2F,$C2,$FD,$65,$F2,$1F,$EC
	.dc.b	$78,$46,$1F,$F6,$53,$5E,$11,$BF,$C2,$DF,$1B,$49,$3F,$ED,$18,$CB
	.dc.b	$78,$38,$11,$C1,$DB,$E4,$3F,$75,$C5,$FE,$B8,$E5,$78,$0F,$99,$CF
	.dc.b	$17,$E2,$A3,$18,$7F,$A1,$B5,$63,$A8,$FE,$37,$18,$7F,$D2,$D8,$74
	.dc.b	$C3,$4D,$8E,$55,$C5,$F8,$6F,$C5,$AE,$51,$F2,$63,$07,$C9,$E6,$7C
	.dc.b	$BF,$4C,$5F,$01,$4E,$93,$AB,$FF,$4E,$B5,$FF,$58,$E3,$F0,$27,$84
	.dc.b	$6A,$F8,$9D,$C8,$FF,$BF,$EB,$93,$FC,$1F,$E1,$DF,$0F,$EB,$F8,$47
	.dc.b	$6D,$3D,$D3,$AA,$FF,$57,$B5,$FF,$5C,$4F,$07,$FD,$66,$D7,$FD,$79
	.dc.b	$1B,$E0,$FF,$76,$37,$83,$F9,$A5,$6E,$7F,$DE,$0D,$8F,$E7,$13,$BA
	.dc.b	$FE,$A3,$F0,$CF,$63,$D1,$F0,$F2,$7F,$D8,$73,$96,$CE,$5F,$3C,$FD
	.dc.b	$99,$D4,$7F,$B4,$5B,$8F,$8A,$78,$2C,$3F,$A5,$4E,$9F,$E9,$9F,$0F
	.dc.b	$F6,$AF,$E0,$FC,$3F,$CB,$3F,$0D,$19,$8F,$F6,$F3,$6F,$DF,$3B,$9F
	.dc.b	$F6,$DC,$E6,$BA,$F0,$66,$7E,$7D,$74,$8E,$BE,$BF,$D9,$F8,$79,$5C
	.dc.b	$A3,$2F,$D4,$8C,$BE,$28,$FC,$D7,$C3,$3E,$1E,$FA,$12,$BF,$DD,$B3
	.dc.b	$96,$CF,$02,$3F,$EF,$2F,$86,$7C,$3B,$6F,$F0,$3F,$C1,$3E,$13,$FB
	.dc.b	$95,$6A,$73,$1F,$DD,$71,$ED,$82,$F8,$3F,$E1,$08,$CC,$62,$17,$CD
	.dc.b	$E1,$5D,$B0,$F1,$46,$0E,$58,$FF,$80,$EF,$96,$1D,$2F,$86,$3E,$4B
	.dc.b	$FF,$0F,$E3,$D6,$F7,$87,$1B,$F4,$36,$B7,$8F,$82,$BC,$D7,$C6,$F7
	.dc.b	$5D,$46,$3E,$17,$8E,$39,$AE,$11,$E8,$38,$4E,$DF,$83,$87,$7B,$F4
	.dc.b	$C7,$BC,$EB,$8F,$95,$A7,$59,$E0,$F8,$0E,$15,$D4,$ED,$7E,$28,$EF
	.dc.b	$87,$F6,$66,$9F,$DA,$23,$E9,$3F,$E1,$36,$BF,$80,$F8,$77,$C3,$F1
	.dc.b	$94,$62,$79,$35,$E4,$D7,$FB,$AF,$C3,$FF,$53,$4D,$23,$FD,$5D,$F9
	.dc.b	$DF,$0A,$F8,$7F,$DD,$4F,$C5,$3F,$CB,$73,$A8,$CE,$DF,$8D,$9C,$9F
	.dc.b	$2E,$B9,$72,$7D,$98,$CC,$FE,$4E,$73,$1D,$4F,$29,$C6,$BA,$EF,$EA
	.dc.b	$8C,$5F,$37,$77,$CE,$B7,$1F,$E3,$DC,$3C,$D5,$B6,$C7,$76,$D7,$A6
	.dc.b	$39,$DF,$F9,$87,$0F,$2C,$E6,$B7,$3E,$38,$F0,$5F,$5B,$F2,$93,$4F
	.dc.b	$92,$B0,$F7,$C6,$23,$E1,$DF,$7C,$7F,$C4,$91,$FE,$88,$AD,$DE,$4F
	.dc.b	$F4,$5D,$F1,$9F,$39,$BA,$E2,$F7,$E8,$B7,$FD,$06,$F8,$9F,$E7,$CF
	.dc.b	$87,$7C,$66,$AA,$4E,$AF,$C5,$1E,$4F,$4D,$7C,$2B,$55,$DD,$B8,$3E
	.dc.b	$23,$C2,$71,$F0,$5F,$11,$8A,$DC,$DB,$FE,$45,$1F,$E9,$0A,$DB,$A2
	.dc.b	$46,$8F,$8D,$3D,$6B,$96,$BF,$43,$CF,$74,$FC,$46,$E1,$ED,$1F,$11
	.dc.b	$FE,$B5,$7D,$71,$AC,$7B,$1F,$E4,$4A,$E4,$7D,$69,$E3,$AE,$1E,$31
	.dc.b	$67,$CD,$1A,$32,$7C,$8F,$87,$7C,$47,$D8,$6C,$79,$71,$7C,$46,$6A
	.dc.b	$4F,$17,$C5,$3A,$8C,$12,$FD,$37,$D1,$F1,$CE,$DC,$D7,$E2,$3E,$19
	.dc.b	$F0,$F4,$FC,$9F,$96,$F8,$8F,$EA,$99,$7E,$02,$B6,$DA,$FC,$A6,$09
	.dc.b	$CB,$FE,$86,$6B,$FE,$B1,$6B,$FE,$CF,$77,$BF,$F9,$5E,$DE,$3F,$8F
	.dc.b	$0F,$F2,$7C,$61,$FC,$8F,$C4,$7F,$B2,$CF,$F1,$9B,$AF,$6B,$F4,$71
	.dc.b	$9F,$33,$CA,$FF,$AF,$B2,$34,$91,$F0,$3C,$FF,$F2,$8D,$F6,$8C,$79
	.dc.b	$FF,$AE,$F8,$77,$C4,$53,$1D,$D6,$FF,$4B,$3A,$7F,$A1,$37,$2F,$F5
	.dc.b	$4B,$5F,$F6,$A3,$5F,$F7,$93,$BD,$BF,$85,$FC,$57,$FB,$A6,$31,$5F
	.dc.b	$EF,$78,$FF,$76,$7C,$1B,$CB,$FE,$06,$D9,$FD,$BF,$C4,$BE,$17,$35
	.dc.b	$27,$FA,$B5,$6F,$2F,$F5,$EB,$5F,$F7,$43,$51,$1F,$E0,$8F,$0C,$FF
	.dc.b	$07,$18,$8E,$31,$F8,$6F,$86,$F1,$57,$F0,$A6,$CF,$EC,$3E,$21,$F1
	.dc.b	$1F,$C7,$27,$3B,$7F,$2C,$77,$5F,$EF,$B6,$BE,$40,$E1,$F3,$B5,$CE
	.dc.b	$7F,$1B,$E2,$3B,$37,$1F,$23,$67,$F8,$1D,$78,$A7,$B6,$6B,$E3,$1F
	.dc.b	$C9,$23,$0F,$23,$11,$AA,$FE,$1B,$3A,$7A,$DF,$7C,$3D,$CD,$BF,$66
	.dc.b	$1E,$1A,$FA,$18,$73,$BF,$E9,$F8,$73,$78,$DF,$80,$F1,$57,$82,$B8
	.dc.b	$47,$A9,$ED,$1D,$23,$73,$D2,$73,$2A,$ED,$5E,$4C,$33,$18,$9E,$47
	.dc.b	$B2,$7C,$4F,$85,$B7,$E6,$8C,$11,$F0,$47,$88,$F3,$1C,$0B,$1D,$4E
	.dc.b	$0E,$1F,$CB,$DF,$E4,$89,$1F,$9E,$7F,$B3,$F8,$0F,$88,$F6,$B6,$7B
	.dc.b	$57,$E7,$A3,$0F,$63,$13,$C0,$93,$C2,$13,$C2,$FD,$71,$FD,$32,$31
	.dc.b	$3E,$A9,$FF,$03,$C6,$27,$E4,$C6,$2B,$AC,$78,$4E,$AF,$A8,$95,$E1
	.dc.b	$AE,$53,$A8,$FE,$EF,$18,$C3,$EB,$4E,$4F,$25,$72,$7B,$4F,$22,$47
	.dc.b	$8A,$FC,$97,$D2,$30,$FE,$C6,$76,$F8,$E3,$A4,$F0,$AE,$B1,$89,$E0
	.dc.b	$F2,$8F,$9B,$1D,$4E,$45,$7C,$EE,$1A,$49,$F9,$E7,$72,$47,$63,$83
	.dc.b	$88,$EE,$79,$23,$B4,$66,$3F,$41,$87,$9D,$DC,$67,$0F,$97,$87,$C0
	.dc.b	$9F,$15,$FD,$D3,$C0,$E0,$E2,$E5,$A4,$3E,$A5,$7A,$A3,$8B,$E3,$9F
	.dc.b	$B5,$7E,$3A,$DD,$68,$E8,$E0,$E9,$3C,$23,$DA,$6E,$3C,$17,$EA,$C3
	.dc.b	$A5,$6E,$7D,$07,$32,$4E,$EB,$9C,$72,$BE,$37,$E5,$AD,$4F,$2B,$E9
	.dc.b	$87,$AE,$B5,$7C,$30,$E7,$7F,$41,$B7,$F3,$6F,$E2,$4F,$B2,$FF,$C2
	.dc.b	$33,$FB,$68,$FE,$84,$7E,$74,$AE,$5F,$A6,$62,$38,$92,$BF,$D7,$13
	.dc.b	$FA,$1E,$4F,$86,$48,$F1,$B6,$BD,$8D,$BF,$AC,$D9,$F9,$D3,$FC,$12
	.dc.b	$30,$FD,$46,$BF,$36,$BF,$C3,$6F,$49,$EE,$47,$04,$AE,$F1,$96,$C6
	.dc.b	$63,$C3,$3D,$27,$24,$8C,$D7,$8A,$7E,$04,$73,$32,$5A,$F4,$35,$F4
	.dc.b	$4F,$1B,$FE,$5A,$D9,$FE,$22,$D9,$FE,$44,$DB,$FE,$42,$D9,$FE,$14
	.dc.b	$D9,$FE,$64,$D7,$E5,$47,$86,$FC,$F1,$EF,$37,$1E,$03,$53,$E4,$69
	.dc.b	$98,$E0,$74,$3F,$12,$53,$F1,$31,$F8,$E2,$47,$FA,$8B,$B1,$F1,$1D
	.dc.b	$6B,$FD,$13,$18,$9E,$D8,$7F,$B1,$23,$13,$F0,$A7,$95,$71,$AD,$36
	.dc.b	$3F,$C3,$95,$C2,$73,$1C,$0E,$A6,$A7,$FD,$87,$3B,$8C,$1A,$AF,$F1
	.dc.b	$CC,$65,$B3,$C0,$F0,$38,$9E,$CF,$C4,$9F,$4B,$FA,$E6,$BD,$8D,$4F
	.dc.b	$FB,$53,$C6,$7B,$FD,$6F,$FD,$93,$18,$7A,$E1,$FE,$E8,$8C,$5F,$CA
	.dc.b	$BE,$57,$C6,$3F,$D6,$B3,$AA,$F8,$71,$89,$DD,$F6,$78,$4F,$57,$33
	.dc.b	$B9,$F0,$9F,$E4,$9A,$E8,$D9,$DD,$76,$39,$47,$53,$F0,$F4,$AF,$4B
	.dc.b	$FC,$49,$AF,$38,$DC,$76,$8F,$F3,$5F,$0F,$FA,$BE,$7A,$47,$F9,$8F
	.dc.b	$B3,$C3,$1F,$E5,$D9,$D4,$FA,$4D,$47,$0B,$3B,$AE,$33,$D9,$CB,$C0
	.dc.b	$93,$EC,$6C,$F1,$3F,$DF,$F5,$E0,$9D,$1C,$A5,$19,$9E,$11,$C8,$FF
	.dc.b	$43,$DF,$BD,$DD,$F8,$AF,$95,$FB,$27,$F9,$CB,$5E,$66,$27,$FD,$9F
	.dc.b	$1D,$E3,$FD,$27,$FA,$D7,$D1,$C3,$CF,$5D,$5F,$F4,$24,$66,$7F,$DF
	.dc.b	$BF,$0E,$EE,$FF,$6C,$B6,$BE,$DB,$67,$F8,$1B,$5F,$E9,$BF,$0A,$F8
	.dc.b	$77,$93,$F8,$28,$C3,$7E,$1F,$FC,$87,$5F,$C3,$27,$2F,$F0,$EF,$8E
	.dc.b	$FF,$77,$3F,$8A,$6B,$9D,$3F,$1A,$A3,$2F,$F5,$EF,$86,$7C,$3E,$9B
	.dc.b	$1F,$94,$6B,$FC,$71,$AF,$1E,$2E,$04,$78,$3E,$0F,$C3,$1E,$C5,$24
	.dc.b	$FE,$82,$32,$FE,$8B,$E1,$DF,$0F,$FE,$B3,$E5,$FC,$7F,$43,$C8,$7C
	.dc.b	$46,$79,$7F,$89,$C6,$31,$FF,$66,$DF,$E3,$71,$E7,$18,$AF,$F5,$F4
	.dc.b	$62,$3F,$9B,$B5,$FF,$00,$B5,$DE,$3F,$2A,$38,$57,$58,$F1,$4F,$FB
	.dc.b	$72,$73,$5E,$4F,$1F,$33,$FB,$DF,$27,$79,$FA,$FE,$0A,$E6,$77,$D8
	.dc.b	$AA,$7D,$A7,$17,$FC,$0A,$30,$47,$F9,$2B,$C1,$F8,$02,$FE,$57,$D1
	.dc.b	$F9,$3E,$1D,$F1,$3E,$6F,$89,$78,$54,$D4,$4C,$39,$89,$DB,$86,$AE
	.dc.b	$1F,$DF,$5F,$68,$E2,$D7,$5B,$FF,$BC,$6B,$8D,$7F,$05,$6C,$7F,$BD
	.dc.b	$6B,$84,$FF,$09,$F8,$97,$BB,$E9,$F4,$A6,$3E,$69,$CB,$67,$2F,$80
	.dc.b	$1E,$B9,$FA,$F3,$C2,$34,$1C,$57,$96,$79,$BE,$3A,$FB,$FF,$13,$F6
	.dc.b	$EF,$E2,$7C,$4F,$CD,$33,$C5,$23,$24,$3B,$CF,$16,$C7,$17,$57,$FC
	.dc.b	$A6,$74,$DF,$6F,$ED,$7C,$4C,$AF,$AF,$49,$F2,$99,$8F,$6E,$CA,$CC
	.dc.b	$7D,$EF,$87,$7C,$4F,$E6,$97,$A9,$4C,$44,$9E,$0D,$7E,$4F,$C3,$BE
	.dc.b	$27,$F0,$90,$F8,$77,$C3,$FD,$28,$56,$F6,$FA,$B5,$A9,$FA,$F9,$79
	.dc.b	$1C,$DF,$18,$C4,$75,$8E,$97,$CE,$30,$7E,$01,$C5,$75,$85,$72,$8C
	.dc.b	$36,$B9,$E1,$EA,$9E,$17,$CA,$7C,$58,$FA,$E7,$8B,$63,$8E,$3E,$8C
	.dc.b	$7C,$58,$73,$9E,$31,$93,$C7,$5C,$5D,$D6,$70,$F0,$47,$C4,$9D,$3C
	.dc.b	$0D,$5F,$57,$89,$F0,$8D,$17,$1F,$73,$AC,$7D,$B3,$C6,$37,$1E,$C7
	.dc.b	$AE,$11,$FF,$30,$E1,$F3,$DF,$F5,$5B,$5F,$82,$6B,$C2,$7C,$4F,$F8
	.dc.b	$8A,$7F,$A3,$B6,$33,$3B,$8F,$EB,$6B,$E2,$69,$1F,$EB,$ED,$C3,$EE
	.dc.b	$7C,$3B,$E2,$7F,$C3,$31,$F7,$EB,$F8,$0D,$7C,$FA,$D7,$57,$D1,$9C
	.dc.b	$9F,$1E,$32,$FC,$93,$35,$DA,$FA,$EF,$FE,$82,$C3,$A4,$62,$BF,$D9
	.dc.b	$B1,$8C,$3B,$47,$AC,$ED,$3B,$7F,$49,$3B,$7C,$06,$EF,$C1,$5C,$A7
	.dc.b	$51,$ED,$9D,$5F,$EB,$70,$F4,$CE,$6B,$73,$E7,$AF,$0D,$F3,$69,$34
	.dc.b	$FF,$1C,$61,$F0,$9F,$F1,$D4,$7C,$2A,$ED,$8F,$F8,$A6,$3F,$96,$D6
	.dc.b	$EF,$27,$F4,$A8,$E0,$6A,$78,$29,$FE,$7B,$3F,$D4,$1B,$1F,$CE,$7D
	.dc.b	$57,$C4,$BD,$DE,$35,$AC,$BC,$FD,$58,$3C,$97,$C2,$B1,$8F,$5F,$89
	.dc.b	$E2,$6B,$C4,$7C,$4E,$27,$44,$D3,$C5,$3A,$8C,$65,$F1,$AF,$85,$6A
	.dc.b	$EC,$FF,$98,$7D,$3F,$E6,$97,$FD,$B1,$B5,$FD,$56,$1F,$0E,$CF,$E2
	.dc.b	$D6,$AF,$97,$66,$0F,$A1,$7C,$2B,$0F,$58,$B8,$75,$C3,$A4,$E4,$FE
	.dc.b	$FB,$F1,$0F,$88,$FA,$6D,$8F,$F0,$59,$E4,$F8,$9C,$CC,$8E,$2F,$8A
	.dc.b	$93,$9B,$FE,$D7,$7C,$DB,$5B,$9D,$65,$FE,$85,$F8,$87,$C4,$FE,$3A
	.dc.b	$FF,$BB,$C6,$32,$FF,$54,$B6,$3F,$DE,$2D,$7F,$A6,$07,$55,$FE,$4C
	.dc.b	$F5,$DE,$37,$E4,$27,$37,$FE,$DE,$9C,$BF,$E2,$08,$CD,$FF,$8A,$E7
	.dc.b	$2F,$F8,$1A,$33,$E6,$27,$38,$7F,$A7,$FE,$27,$FC,$5B,$87,$FB,$01
	.dc.b	$B8,$79,$83,$87,$D2,$D7,$FB,$33,$C2,$7F,$CC,$BE,$7F,$F9,$1E,$BF
	.dc.b	$C6,$B3,$97,$FC,$CB,$E7,$FE,$B2,$31,$7F,$E4,$3A,$E0,$FF,$87,$A3
	.dc.b	$38,$FF,$9B,$EB,$F6,$7E,$83,$E2,$7E,$95,$A3,$18,$FF,$B6,$9B,$1E
	.dc.b	$B6,$BF,$49,$CE,$5F,$EC,$1F,$89,$F1,$CF,$3B,$FE,$03,$18,$3E,$04
	.dc.b	$7F,$8D,$9A,$9A,$63,$E1,$7F,$C2,$B8,$7F,$BB,$1B,$3F,$35,$AF,$E1
	.dc.b	$1A,$FF,$84,$5F,$E1,$51,$C7,$AF,$DF,$18,$9E,$51,$F0,$BE,$4B,$FD
	.dc.b	$73,$5F,$EF,$A6,$CF,$DB,$6B,$F9,$26,$BD,$FE,$1B,$F8,$FD,$9F,$E4
	.dc.b	$3F,$10,$F8,$8C,$C4,$C3,$D8,$F9,$08,$E0,$FF,$3A,$F5,$BE,$23,$86
	.dc.b	$3F,$E5,$A8,$C4,$FF,$9E,$B1,$F0,$3C,$2B,$7B,$7F,$BD,$6B,$77,$FE
	.dc.b	$62,$BD,$BC,$EB,$77,$FE,$39,$AD,$DF,$FB,$F6,$F6,$F3,$AD,$DF,$F9
	.dc.b	$C2,$B7,$7F,$0C,$F1,$1E,$5C,$3F,$A1,$C6,$A7,$FC,$EF,$87,$4A,$F0
	.dc.b	$5F,$49,$FE,$4C,$F0,$24,$73,$3A,$C7,$A5,$B8,$79,$F0,$DC,$E4,$91
	.dc.b	$99,$D9,$E0,$2C,$79,$27,$8C,$FB,$E7,$32,$BF,$43,$F0,$6F,$73,$CE
	.dc.b	$3E,$75,$E1,$D4,$72,$AE,$31,$EE,$8F,$5B,$FC,$7B,$0F,$AE,$DB,$FD
	.dc.b	$BC,$62,$3B,$61,$F7,$7D,$7E,$D1,$FD,$35,$B3,$FB,$43,$E8,$46,$6F
	.dc.b	$D7,$19,$BF,$AB,$52,$7F,$22,$F2,$31,$5F,$91,$71,$1A,$24,$7C,$E7
	.dc.b	$F1,$51,$C2,$3D,$8F,$A8,$91,$FC,$9E,$3C,$87,$FA,$DB,$89,$F1,$1E
	.dc.b	$0C,$55,$FE,$8D,$9C,$BE,$08,$CD,$7F,$B0,$A7,$2F,$9E,$7F,$DB,$F1
	.dc.b	$89,$F9,$11,$8A,$EC,$F5,$7F,$C3,$33,$B8,$C1,$DC,$B5,$E7,$9D,$9C
	.dc.b	$9F,$13,$88,$EC,$71,$8C,$B8,$79,$39,$7F,$DF,$D1,$C5,$C9,$EB,$2D
	.dc.b	$7C,$07,$73,$CE,$3D,$B1,$A8,$F3,$C6,$EB,$D9,$3F,$B9,$BE,$D3,$9B
	.dc.b	$B7,$F6,$8E,$87,$22,$57,$17,$2F,$B0,$F0,$19,$29,$E6,$2E,$3F,$D2
	.dc.b	$30,$F1,$C6,$6F,$C5,$5F,$0E,$72,$4B,$F9,$D5,$B9,$F1,$B4,$F0,$57
	.dc.b	$27,$C3,$5D,$AF,$C1,$7E,$E9,$F4,$57,$C8,$38,$12,$FA,$D7,$07,$DF
	.dc.b	$3D,$CF,$CC,$CE,$DE,$71,$C3,$0E,$D1,$BC,$3C,$51,$87,$D9,$3E,$FC
	.dc.b	$30,$7B,$A3,$35,$F2,$2B,$35,$E2,$C3,$C9,$3E,$EB,$E0,$69,$FC,$46
	.dc.b	$1E,$57,$E9,$1F,$E2,$CC,$3F,$4E,$D9,$FF,$02,$C6,$23,$F3,$E6,$6F
	.dc.b	$F7,$11,$98,$FC,$01,$91,$5F,$95,$F0,$9E,$F7,$C2,$6D,$FC,$A6,$D7
	.dc.b	$E1,$9B,$5E,$76,$CF,$DE,$6C,$FC,$E0,$1C,$BF,$55,$B7,$F8,$E6,$E1
	.dc.b	$FE,$96,$6D,$72,$BF,$46,$1E,$72,$47,$9D,$B8,$79,$DB,$3E,$76,$CF
	.dc.b	$9D,$B3,$E7,$3C,$CE,$1F,$33,$70,$F3,$37,$0F,$33,$6B,$CC,$7C,$0A
	.dc.b	$E6,$F9,$48,$F9,$5B,$7E,$56,$9E,$52,$C7,$97,$0F,$86,$FD,$A8,$ED
	.dc.b	$18,$9F,$8B,$1C,$27,$D2,$DC,$3D,$2D,$C3,$D2,$D3,$D3,$3F,$62,$FA
	.dc.b	$47,$6C,$3D,$D3,$E1,$34,$F2,$70,$7C,$03,$DC,$62,$7F,$02,$74,$78
	.dc.b	$CF,$81,$E1,$87,$A9,$E2,$D7,$FD,$1F,$3F,$94,$6D,$7F,$9A,$23,$11
	.dc.b	$E7,$AE,$9F,$11,$FE,$87,$7B,$4F,$E4,$A3,$13,$E5,$AF,$F7,$2C,$62
	.dc.b	$FC,$47,$F3,$B7,$17,$D6,$BD,$F1,$87,$D6,$63,$A7,$99,$27,$C3,$76
	.dc.b	$75,$1E,$69,$F9,$57,$DD,$D3,$63,$89,$B2,$4F,$FA,$76,$FD,$6F,$FA
	.dc.b	$D1,$AF,$E4,$DA,$FE,$D1,$F4,$2F,$88,$FD,$B4,$F4,$AF,$EB,$51,$89
	.dc.b	$F3,$9F,$79,$C4,$79,$63,$18,$FA,$2B,$F4,$55,$A7,$C5,$87,$F8,$12
	.dc.b	$FA,$D5,$9D,$B6,$FF,$84,$DF,$17,$D1,$79,$7D,$13,$67,$51,$8A,$F5
	.dc.b	$1D,$0F,$C7,$46,$8D,$E1,$F0,$9C,$57,$FA,$C9,$F8,$15,$FE,$DE,$6B
	.dc.b	$CC,$FF,$66,$3E,$0F,$D6,$DE,$57,$8E,$BC,$D1,$8C,$3C,$31,$89,$D4
	.dc.b	$7E,$86,$75,$7F,$2A,$31,$3A,$8F,$D6,$CE,$B2,$FF,$07,$36,$BF,$88
	.dc.b	$36,$7E,$4D,$F7,$9E,$64,$8E,$C7,$4B,$D3,$63,$CE,$6E,$FB,$CF,$FB
	.dc.b	$5A,$FC,$F5,$FC,$7C,$38,$7C,$71,$B3,$EE,$6A,$EC,$7C,$57,$83,$0F
	.dc.b	$F7,$7B,$F7,$7E,$19,$F1,$0F,$23,$FC,$A9,$18,$6F,$C4,$39,$33,$5B
	.dc.b	$D3,$FE,$00,$7C,$43,$8F,$FC,$B9,$08,$C3,$FE,$5F,$6B,$C2,$37,$6E
	.dc.b	$DA,$49,$FF,$33,$46,$5F,$F3,$47,$C3,$FF,$6C,$6F,$37,$82,$31,$1F
	.dc.b	$1C,$38,$7E,$57,$C3,$BD,$6F,$3B,$63,$FC,$E8,$D7,$93,$5F,$A5,$F7
	.dc.b	$8F,$8B,$F7,$B6,$3F,$CF,$8D,$7C,$4D,$7E,$D7,$C3,$BE,$23,$E3,$B6
	.dc.b	$3F,$D0,$8D,$7D,$4D,$78,$C7,$F6,$1E,$4F,$F3,$14,$ED,$B1,$9E,$9F
	.dc.b	$F4,$27,$33,$C7,$C6,$1F,$E0,$8C,$38,$57,$F6,$B6,$C6,$6F,$7D,$1F
	.dc.b	$8B,$BE,$33,$FE,$0F,$F2,$3E,$FF,$61,$A4,$8F,$F4,$F4,$65,$FE,$8B
	.dc.b	$1C,$A3,$11,$FD,$7E,$BF,$31,$F1,$19,$A9,$3C,$0F,$ED,$33,$A7,$FA
	.dc.b	$E7,$C3,$BE,$23,$CB,$F0,$8F,$12,$9E,$89,$D5,$70,$AC,$BF,$D9,$6B
	.dc.b	$83,$92,$1F,$DC,$1C,$B6,$BA,$4F,$C4,$F8,$87,$B3,$E2,$F4,$65,$FF
	.dc.b	$71,$9C,$B6,$72,$F5,$9E,$51,$23,$C9,$1A,$E3,$FF,$50,$57,$29,$FF
	.dc.b	$01,$FC,$33,$E1,$FE,$0F,$C4,$49,$FF,$00,$4E,$5B,$39,$7D,$93,$F7
	.dc.b	$27,$51,$D2,$75,$7F,$EA,$FD,$17,$B9,$E1,$ED,$7D,$0F,$88,$98,$78
	.dc.b	$23,$2F,$CD,$8C,$BE,$88,$FC,$5F,$C3,$3E,$1E,$FC,$44,$AF,$F0,$C4
	.dc.b	$E5,$B3,$C0,$8F,$F8,$77,$E1,$DF,$11,$FE,$72,$87,$C3,$BE,$1B,$FC
	.dc.b	$91,$5A,$DF,$FC,$D3,$39,$AF,$F3,$5E,$3D,$DE,$37,$FE,$4F,$8C,$4E
	.dc.b	$E5,$7C,$DD,$9F,$E7,$47,$11,$B9,$E3,$5C,$F1,$FF,$2D,$4F,$3A,$DC
	.dc.b	$F9,$B1,$F5,$4F,$1A,$DC,$F1,$C3,$D2,$EF,$1B,$1E,$1B,$E3,$3B,$8F
	.dc.b	$28,$C7,$C6,$EE,$B8,$BE,$CA,$E2,$FC,$4B,$F0,$B4,$DE,$3E,$87,$91
	.dc.b	$DA,$FA,$BD,$E7,$8B,$D6,$F8,$DF,$68,$D1,$E2,$AE,$37,$D6,$38,$1F
	.dc.b	$CF,$5F,$A0,$FF,$65,$6B,$F0,$CD,$78,$6F,$88,$FE,$89,$1F,$62,$30
	.dc.b	$EA,$33,$3F,$D1,$5F,$EB,$70,$F8,$8A,$47,$FA,$FB,$74,$FC,$27,$C3
	.dc.b	$FF,$83,$9F,$C3,$4F,$FA,$DA,$7E,$6C,$E7,$5F,$EA,$D1,$8F,$49,$E1
	.dc.b	$FD,$2B,$E0,$7D,$8B,$E0,$FF,$80,$EB,$84,$EE,$3C,$C6,$EB,$9B,$F8
	.dc.b	$03,$9E,$1E,$68,$CB,$E6,$8E,$2E,$67,$A5,$76,$9E,$46,$B0,$FF,$59
	.dc.b	$61,$E4,$8F,$F5,$A4,$EE,$FA,$36,$BD,$F3,$E5,$BE,$AD,$26,$DE,$85
	.dc.b	$87,$F8,$BD,$EA,$6A,$7C,$38,$FF,$95,$E7,$53,$EE,$BC,$D7,$FA,$86
	.dc.b	$7F,$9A,$B7,$A7,$FC,$C5,$E7,$BE,$23,$FC,$25,$39,$26,$1F,$EB,$0B
	.dc.b	$E2,$FF,$AD,$6B,$87,$9B,$E1,$C3,$C5,$F1,$13,$0F,$F5,$FD,$7F,$BD
	.dc.b	$7E,$23,$18,$73,$AB,$AF,$D6,$AD,$5F,$9E,$72,$FC,$E7,$CF,$E2,$B0
	.dc.b	$CE,$03,$FC,$75,$D1,$FE,$79,$F4,$CF,$F9,$FF,$07,$C4,$7F,$8E,$DF
	.dc.b	$F6,$24,$E9,$B3,$C4,$E1,$3B,$AF,$F6,$8D,$6D,$B3,$BA,$FF,$6E,$9D
	.dc.b	$B2,$C4,$FF,$B6,$3E,$21,$F1,$1F,$E8,$D6,$D7,$0C,$5F,$11,$9A,$93
	.dc.b	$FE,$D9,$AD,$DF,$FB,$96,$3D,$75,$A8,$ED,$57,$6F,$D0,$D6,$8E,$51
	.dc.b	$FE,$9D,$7F,$C8,$47,$FB,$94,$B3,$D7,$07,$D1,$6F,$AA,$56,$DF,$8B
	.dc.b	$18,$AF,$F7,$AC,$F2,$C3,$F3,$4D,$3E,$B1,$F4,$A7,$84,$7E,$7F,$E4
	.dc.b	$DC,$3F,$EB,$16,$D7,$C5,$F9,$DC,$4C,$8F,$F7,$E4,$EB,$0F,$45,$65
	.dc.b	$5F,$5B,$C3,$F1,$32,$56,$30,$FF,$23,$FC,$B7,$FB,$07,$E4,$1F,$0D
	.dc.b	$FE,$8C,$F5,$13,$98,$95,$F9,$1C,$7F,$71,$7C,$23,$FC,$6F,$7C,$1D
	.dc.b	$1F,$97,$CD,$3B,$7C,$F1,$98,$C5,$7F,$03,$F8,$57,$8B,$E0,$BD,$9F
	.dc.b	$A1,$8C,$DF,$68,$FF,$79,$7C,$07,$FB,$6F,$E0,$FF,$D4,$73,$E4,$F9
	.dc.b	$4C,$4F,$FB,$7F,$E2,$3B,$CF,$95,$B7,$CA,$3A,$28,$CA,$C1,$F1,$1F
	.dc.b	$EE,$96,$D7,$0A,$DC,$62,$BD,$23,$F0,$CC,$69,$FC,$85,$B7,$FB,$DE
	.dc.b	$A7,$E1,$6F,$89,$FF,$7B,$56,$B0,$FE,$04,$D7,$A6,$7F,$C7,$AB,$86
	.dc.b	$1E,$AA,$D1,$27,$57,$D1,$FE,$11,$F8,$63,$E9,$A9,$8F,$5E,$E1,$FC
	.dc.b	$02,$74,$D9,$D3,$4D,$71,$74,$F4,$3A,$B6,$32,$73,$8C,$B4,$FA,$3F
	.dc.b	$8C,$3D,$8F,$EB,$BC,$1F,$D1,$E3,$13,$FC,$C3,$6F,$F6,$EB,$C2,$B6
	.dc.b	$FE,$46,$77,$BF,$D2,$6D,$FF,$33,$BE,$33,$9B,$E2,$FF,$2F,$AD,$DF
	.dc.b	$39,$F9,$DF,$13,$AA,$56,$9F,$EF,$BA,$3E,$CF,$E6,$9F,$3B,$E6,$AE
	.dc.b	$3F,$2E,$2D,$7F,$1F,$A5,$63,$3F,$E3,$B0,$FA,$BF,$AF,$F4,$BF,$61
	.dc.b	$B8,$FD,$56,$E1,$CD,$B7,$FD,$BF,$E7,$9F,$01,$E9,$6D,$FC,$BF,$A1
	.dc.b	$C4,$C8,$FD,$1C,$EB,$0F,$EC,$9C,$0B,$D4,$7E,$7A,$75,$87,$7C,$3F
	.dc.b	$29,$5A,$8F,$5F,$CB,$7E,$40,$FC,$8F,$35,$CB,$F1,$D1,$71,$F8,$93
	.dc.b	$6B,$FC,$3F,$E3,$7F,$0D,$BF,$E0,$4F,$EA,$5B,$5F,$CC,$8F,$E1,$31
	.dc.b	$96,$99,$9E,$6F,$F6,$16,$BF,$82,$8E,$11,$F2,$7C,$CF,$90,$7E,$BF
	.dc.b	$BE,$DF,$C6,$8B,$7F,$1A,$6D,$71,$AD,$1F,$99,$8C,$BF,$DF,$6F,$FB
	.dc.b	$F7,$A9,$E1,$BF,$F4,$E4,$FB,$3C,$9C,$4C,$8F,$7B,$B8,$C4,$FB,$BC
	.dc.b	$7E,$5E,$F7,$36,$DF,$C6,$6C,$EE,$3E,$4C,$FE,$77,$E7,$BF,$27,$1F
	.dc.b	$C9,$1B,$3F,$CD,$3E,$AD,$F8,$6F,$F4,$18,$FE,$3F,$3A,$D7,$FC,$7C
	.dc.b	$DB,$FE,$88,$D9,$5F,$2F,$2F,$C1,$D3,$2F,$C0,$FE,$D2,$75,$18,$25
	.dc.b	$78,$27,$58,$F8,$F1,$F0,$C7,$39,$D5,$7F,$49,$BE,$71,$82,$57,$09
	.dc.b	$D5,$F3,$7D,$37,$CE,$75,$5F,$0E,$B4,$FF,$8D,$F1,$F7,$26,$9F,$07
	.dc.b	$0F,$0D,$72,$26,$1D,$8F,$0E,$3D,$DC,$5F,$99,$FF,$22,$57,$27,$35
	.dc.b	$CA,$7C,$58,$FC,$4B,$E5,$18,$AF,$93,$5F,$E8,$AB,$E5,$39,$3B,$16
	.dc.b	$35,$8F,$BA,$3D,$98,$74,$6D,$FA,$0E,$A7,$90,$8F,$E0,$27,$96,$3F
	.dc.b	$12,$BC,$B8,$78,$AB,$99,$E2,$2E,$1E,$2C,$7C,$17,$C7,$1F,$05,$6E
	.dc.b	$30,$FD,$59,$DE,$1D,$B1,$F0,$B8,$BE,$38,$7C,$9B,$EF,$1E,$8C,$7C
	.dc.b	$98,$F4,$AF,$9D,$18,$BE,$57,$DE,$FE,$0B,$F2,$5E,$58,$78,$1C,$C7
	.dc.b	$63,$E2,$63,$CB,$0E,$B5,$B9,$F6,$5F,$DB,$BE,$38,$F8,$DF,$C1,$D6
	.dc.b	$EF,$D5,$3F,$8F,$8C,$4F,$D2,$BF,$06,$3E,$98,$E7,$87,$59,$EF,$1E
	.dc.b	$A9,$D4,$60,$F7,$17,$0F,$3E,$3E,$9A,$F7,$B6,$7E,$11,$EE,$8E,$E7
	.dc.b	$5C,$3C,$B1,$A3,$CA,$E1,$F8,$C6,$27,$A3,$67,$9B,$F9,$8B,$F3,$DF
	.dc.b	$C2,$C3,$D8,$DA,$F0,$5F,$2C,$7E,$33,$6B,$73,$F1,$E7,$E7,$61,$E9
	.dc.b	$3E,$04,$7B,$8F,$0D,$FC,$26,$99,$8F,$D5,$ED,$FE,$BB,$FA,$2F,$E7
	.dc.b	$DE,$77,$F3,$F6,$D7,$F5,$A6,$D7,$F5,$86,$DF,$F5,$8F,$9A,$FE,$8C
	.dc.b	$7F,$47,$F8,$7B,$7F,$D8,$A2,$DF,$F6,$09,$B5,$FE,$F1,$FA,$07,$C2
	.dc.b	$7F,$86,$DB,$87,$F8,$B7,$E4,$F1,$32,$3F,$CC,$B3,$AC,$3F,$C0,$75
	.dc.b	$97,$FC,$BF,$19,$BE,$D7,$FE,$55,$AD,$47,$F7,$AF,$82,$61,$FD,$BD
	.dc.b	$FF,$2C,$E1,$FD,$E1,$CC,$7F,$7E,$73,$3C,$63,$87,$95,$C5,$A6,$FE
	.dc.b	$2F,$A1,$D7,$E5,$2C,$64,$98,$77,$8C,$BE,$AF,$92,$F8,$87,$F9,$13
	.dc.b	$E0,$5F,$80,$3C,$3C,$E1,$F1,$2A,$DF,$FA,$62,$F7,$58,$CF,$E8,$55
	.dc.b	$BF,$F7,$E5,$69,$E4,$7B,$FC,$2F,$88,$7C,$5F,$9F,$B1,$92,$63,$F1
	.dc.b	$E3,$2F,$C9,$F9,$7F,$99,$F0,$3F,$35,$B5,$F2,$DB,$3B,$8F,$9D,$3F
	.dc.b	$E5,$DF,$C4,$0D,$3E,$16,$BF,$EB,$2A,$FF,$4E,$CE,$AB,$FD,$52,$D8
	.dc.b	$CD,$70,$F4,$79,$56,$A2,$CF,$C7,$BE,$9A,$FE,$12,$03,$0E,$91,$F8
	.dc.b	$9F,$92,$FC,$47,$B3,$F8,$96,$E3,$F8,$76,$DF,$E2,$DB,$5F,$91,$E5
	.dc.b	$FC,$87,$8F,$F9,$16,$DF,$E3,$DB,$7F,$93,$6D,$79,$61,$7F,$EA,$47
	.dc.b	$FD,$C0,$7A,$1C,$38,$6B,$FE,$E5,$6B,$EE,$FC,$69,$87,$58,$FF,$75
	.dc.b	$46,$30,$EB,$1B,$9D,$36,$32,$BE,$6D,$E9,$BE,$17,$31,$86,$F3,$BD
	.dc.b	$AC,$D6,$EF,$BE,$6F,$67,$FD,$F9,$AF,$D5,$F9,$47,$C0,$38,$33,$B7
	.dc.b	$D8,$9C,$C6,$31,$F8,$98,$55,$F2,$EF,$A4,$C7,$EC,$2F,$8E,$A6,$E7
	.dc.b	$55,$FE,$F6,$F9,$C7,$D2,$F3,$FC,$9D,$F3,$74,$ED,$3A,$AF,$F7,$EF
	.dc.b	$0B,$EE,$FC,$9E,$73,$8D,$F6,$78,$6D,$F0,$AF,$47,$CF,$BC,$DF,$56
	.dc.b	$3C,$EB,$47,$E6,$2F,$F0,$D5,$B8,$C7,$2F,$E7,$17,$D0,$7F,$B0,$1C
	.dc.b	$9F,$A0,$9E,$0F,$96,$78,$46,$2B,$F2,$55,$C6,$B4,$7E,$A3,$C3,$7C
	.dc.b	$E7,$CF,$FA,$D7,$D8,$D3,$E9,$4E,$AB,$C2,$1C,$3D,$04,$F2,$76,$D7
	.dc.b	$51,$C3,$0F,$D3,$37,$2F,$81,$5B,$9F,$D5,$36,$3E,$24,$EE,$BF,$80
	.dc.b	$F8,$AF,$7F,$F1,$1F,$13,$FC,$52,$7C,$ED,$7C,$D3,$B7,$D6,$D7,$D5
	.dc.b	$1B,$C7,$F7,$4D,$C3,$EA,$56,$E7,$F7,$CD,$8F,$B1,$3B,$AF,$E2,$AD
	.dc.b	$7F,$A2,$DF,$E7,$7E,$0B,$F5,$11,$F3,$3E,$7F,$EC,$5F,$E0,$5B,$1F
	.dc.b	$C7,$A7,$6F,$CF,$6E,$3F,$C3,$E7,$72,$6C,$EA,$BF,$18,$D9,$FE,$53
	.dc.b	$3C,$0D,$F6,$FE,$C7,$E8,$5F,$33,$FB,$39,$7C,$65,$35,$12,$FD,$51
	.dc.b	$97,$F0,$F5,$C9,$B3,$97,$F1,$15,$D1,$C9,$F8,$EF,$A1,$FB,$CF,$AF
	.dc.b	$E0,$BC,$47,$F0,$58,$C9,$A9,$46,$A7,$97,$1E,$CC,$3E,$08,$C3,$DF
	.dc.b	$CB,$F7,$7B,$32,$30,$43,$CE,$E5,$FD,$06,$1F,$CD,$1F,$D2,$FD,$03
	.dc.b	$8B,$F8,$D3,$FA,$78,$CF,$26,$26,$47,$06,$BF,$AB,$FA,$07,$D0,$FF
	.dc.b	$11,$87,$D0,$BE,$27,$F8,$35,$6B,$9B,$FB,$1C,$E6,$BF,$B2,$E5,$E3
	.dc.b	$73,$7F,$C3,$63,$11,$D2,$39,$E1,$DE,$B9,$4B,$0E,$CE,$CE,$B8,$FF
	.dc.b	$17,$9D,$E1,$D1,$E3,$5C,$1B,$7D,$70,$F1,$5F,$16,$E1,$E6,$3F,$BE
	.dc.b	$DF,$18,$F2,$E3,$99,$63,$DE,$78,$C7,$94,$F4,$61,$DD,$E6,$D7,$8E
	.dc.b	$1E,$17,$D6,$75,$C3,$C0,$F0,$C3,$C1,$87,$78,$D9,$DE,$76,$F8,$27
	.dc.b	$75,$D6,$3B,$9D,$6F,$C5,$5F,$32,$3F,$C0,$AD,$7D,$E7,$F7,$C7,$0F
	.dc.b	$C4,$F9,$A7,$BF,$FD,$F2,$51,$89,$E0,$D7,$AB,$5F,$F0,$3F,$D0,$FF
	.dc.b	$81,$1F,$F1,$5B,$53,$48,$FC,$DF,$73,$FB,$97,$57,$D3,$8E,$26,$7C
	.dc.b	$BF,$F1,$B6,$FF,$8A,$8C,$BF,$8C,$AE,$31,$F5,$CC,$EF,$FD,$86,$33
	.dc.b	$3C,$67,$C6,$75,$BF,$03,$88,$E9,$7E,$0B,$F6,$CE,$CE,$C6,$DE,$D1
	.dc.b	$B8,$FF,$1A,$E1,$F8,$DC,$3C,$35,$CA,$FA,$E1,$E0,$AD,$61,$C7,$0F
	.dc.b	$EF,$11,$89,$D4,$78,$6B,$D5,$39,$8F,$26,$DF,$E4,$7C,$7E,$1B,$FE
	.dc.b	$38,$9D,$39,$C3,$FC,$C0,$DC,$3F,$B9,$DF,$C8,$3F,$D0,$F3,$A8,$F9
	.dc.b	$5C,$F6,$7F,$D1,$6D,$F4,$9F,$42,$FA,$1D,$53,$3F,$E8,$38,$7F,$80
	.dc.b	$1B,$1F,$DF,$DB,$97,$F9,$CF,$0E,$77,$CA,$73,$7F,$6F,$E1,$B0,$78
	.dc.b	$7E,$87,$FC,$EF,$19,$3B,$46,$9F,$F0,$FD,$A7,$78,$FF,$A9,$2B,$4F
	.dc.b	$8D,$E9,$EE,$E3,$2F,$F3,$97,$B3,$FE,$65,$AF,$F0,$B3,$63,$FD,$15
	.dc.b	$8B,$E7,$BF,$CB,$CF,$F8,$7A,$74,$FF,$8C,$DB,$5F,$E9,$06,$C7,$1C
	.dc.b	$BE,$4C,$6A,$7F,$CD,$9D,$98,$3E,$95,$6C,$EC,$FD,$0F,$A1,$7D,$05
	.dc.b	$37,$8B,$E8,$71,$7F,$E3,$2A,$DB,$6B,$73,$AC,$3F,$C6,$95,$A6,$CE
	.dc.b	$AB,$81,$FE,$3C,$3F,$84,$97,$9F,$C3,$F2,$5F,$84,$8C,$D7,$F8,$FD
	.dc.b	$BA,$FF,$A7,$DA,$7F,$AD,$0F,$99,$3C,$6A,$47,$53,$FC,$B6,$68,$C5
	.dc.b	$7F,$92,$FD,$27,$07,$E5,$1B,$3C,$34,$7C,$96,$6A,$4F,$F9,$5A,$B7
	.dc.b	$7F,$E6,$08,$FB,$D5,$A8,$ED,$57,$2F,$C1,$E1,$D2,$F9,$4F,$EA,$7D
	.dc.b	$CF,$F4,$DF,$C8,$7E,$9B,$E8,$1F,$43,$9F,$80,$4E,$62,$61,$FE,$B0
	.dc.b	$AD,$B6,$3F,$D6,$B9,$A7,$2F,$FA,$EB,$E7,$FF,$DB,$F5,$FE,$83,$6C
	.dc.b	$7F,$B8,$27,$7A,$7F,$B3,$E7,$32,$FA,$17,$C3,$F6,$6E,$1F,$EC,$26
	.dc.b	$CF,$FB,$21,$AF,$FB,$5B,$E8,$7C,$6D,$CB,$CF,$18,$8F,$F4,$B3,$5F
	.dc.b	$F7,$9E,$1F,$E7,$3B,$FF,$67,$B6,$3D,$9E,$9F,$C0,$8C,$47,$FB,$93
	.dc.b	$E8,$75,$9F,$FB,$A5,$B7,$FE,$FE,$EB,$7C,$DB,$E4,$BF,$8B,$D6,$B0
	.dc.b	$07,$AB,$FE,$AB,$6B,$A6,$BB,$CB,$E9,$C7,$3F,$BD,$7A,$8F,$9C,$EF
	.dc.b	$5F,$CA,$E7,$31,$FB,$CF,$9F,$FF,$7F,$6B,$FC,$31,$BA,$FE,$8D,$FF
	.dc.b	$73,$3F,$A0,$73,$92,$11,$86,$FD,$17,$5F,$67,$B3,$DF,$D3,$F0,$7A
	.dc.b	$56,$BA,$37,$0F,$D0,$61,$FE,$53,$9F,$05,$70,$7F,$D7,$7F,$45,$FE
	.dc.b	$7C,$6F,$B5,$FD,$17,$89,$F1,$BF,$D1,$2B,$58,$FF,$AF,$DA,$FF,$B5
	.dc.b	$C3,$87,$FD,$B4,$D7,$EA,$B5,$DD,$77,$F7,$3B,$37,$0F,$8D,$F4,$5F
	.dc.b	$34,$F9,$DF,$3B,$63,$24,$BF,$7C,$65,$F8,$3F,$43,$F5,$7E,$8B,$EB
	.dc.b	$36,$BE,$93,$67,$71,$F5,$67,$FB,$1F,$CD,$32,$A7,$C5,$F9,$9B,$5F
	.dc.b	$EE,$49,$D3,$F0,$EF,$F2,$71,$87,$0D,$7E,$43,$5F,$F8,$00,$EF,$93
	.dc.b	$FB,$66,$EF,$99,$EF,$5F,$EF,$AE,$CF,$D1,$C7,$E4,$79,$E6,$3F,$8D
	.dc.b	$E9,$B7,$F8,$29,$B8,$FE,$56,$6D,$7E,$A3,$C9,$FD,$4D,$FF,$81,$DF
	.dc.b	$D8,$47,$EB,$2B,$F2,$D2,$9D,$CE,$4E,$51,$FD,$A8,$4F,$09,$DC,$62
	.dc.b	$77,$8F,$78,$FE,$90,$64,$EC,$73,$AF,$0A,$CF,$FB,$71,$FD,$D2,$7F
	.dc.b	$82,$FD,$17,$D8,$F4,$26,$1F,$4E,$F8,$47,$F3,$0B,$FF,$02,$47,$D2
	.dc.b	$8C,$1F,$A1,$FA,$BE,$31,$F3,$F6,$47,$9F,$DC,$98,$7C,$1A,$F3,$F9
	.dc.b	$F8,$7D,$5A,$FF,$14,$AC,$6B,$F0,$E5,$58,$BF,$93,$E7,$7C,$3F,$83
	.dc.b	$F8,$8D,$8E,$D8,$FF,$1E,$8C,$5F,$63,$FA,$C1,$6B,$FC,$77,$F4,$2E
	.dc.b	$8A,$7F,$90,$FE,$89,$F4,$7F,$D6,$32,$F5,$60,$BE,$11,$86,$C7,$D5
	.dc.b	$35,$7C,$EB,$94,$66,$77,$A7,$D8,$6D,$FE,$39,$B2,$C3,$C7,$E7,$79
	.dc.b	$AB,$C1,$8D,$AF,$C2,$56,$A7,$24,$8F,$04,$F8,$EB,$86,$1E,$18,$B5
	.dc.b	$FD,$1A,$F9,$5F,$08,$E5,$5B,$7E,$E4,$EA,$31,$3B,$23,$FA,$48,$D4
	.dc.b	$FB,$6B,$BD,$7F,$54,$C3,$DF,$5A,$8C,$4E,$9F,$97,$5E,$0C,$7E,$5E
	.dc.b	$1E,$02,$47,$53,$58,$F1,$C7,$E5,$DF,$23,$E3,$DF,$2A,$E1,$1E,$86
	.dc.b	$DF,$A6,$75,$3E,$49,$F8,$66,$8F,$91,$1A,$8E,$D3,$E4,$78,$1E,$0B
	.dc.b	$F2,$61,$F0,$B1,$F8,$13,$C3,$0E,$8F,$0B,$F1,$5F,$09,$F0,$CF,$E1
	.dc.b	$70,$F0,$63,$EE,$26,$3F,$7E,$B7,$5D,$4E,$8F,$A2,$FA,$47,$C1,$3A
	.dc.b	$38,$C3,$D5,$8F,$9B,$0E,$B7,$E2,$8C,$61,$B8,$E8,$78,$27,$F4,$B8
	.dc.b	$7F,$63,$6B,$F4,$0C,$4F,$F9,$02,$33,$1F,$20,$CC,$FF,$17,$8C,$C7
	.dc.b	$AF,$DA,$7E,$23,$FE,$0A,$8F,$54,$63,$C6,$FC,$7B,$6F,$E4,$B6,$77
	.dc.b	$5C,$72,$FB,$13,$96,$C6,$6B,$EA,$CE,$5B,$19,$AF,$9A,$7F,$B1,$DC
	.dc.b	$3E,$76,$E1,$FB,$EA,$E5,$3C,$1A,$EF,$0F,$4D,$70,$C7,$FD,$B1,$5A
	.dc.b	$BE,$53,$9C,$3F,$98,$4E,$6B,$73,$9C,$3F,$A5,$4E,$6B,$73,$9C,$7F
	.dc.b	$AD,$CE,$6B,$73,$9A,$FF,$25,$CE,$6B,$73,$9A,$F8,$18,$FA,$09,$7F
	.dc.b	$42,$75,$7C,$63,$9C,$60,$93,$C2,$BE,$AB,$6F,$E9,$B6,$BE,$8B,$67
	.dc.b	$E5,$57,$29,$F0,$4E,$67,$E6,$46,$08,$F6,$7F,$A5,$AF,$A4,$FE,$E5
	.dc.b	$EA,$7F,$73,$6C,$E9,$B3,$FD,$3D,$B8,$7F,$5B,$FA,$4F,$EF,$47,$F7
	.dc.b	$BF,$92,$B7,$DA,$2D,$FF,$57,$9B,$5F,$DF,$3E,$91,$F1,$7F,$E0,$E8
	.dc.b	$C3,$FE,$53,$F9,$DC,$63,$FE,$16,$AD,$BF,$D6,$1B,$87,$F8,$C5,$AF
	.dc.b	$F9,$55,$AF,$1A,$FF,$58,$C7,$0B,$FF,$17,$F5,$F6,$AF,$F4,$5C,$E6
	.dc.b	$FF,$D2,$5D,$6A,$FF,$13,$7C,$4F,$F9,$E6,$7F,$BA,$BF,$E9,$56,$9A
	.dc.b	$F4,$B6,$D3,$27,$07,$9F,$D2,$48,$E7,$F3,$F7,$2F,$F0,$5D,$AB,$19
	.dc.b	$FF,$8B,$61,$58,$BF,$8D,$F4,$9E,$BC,$1A,$23,$F5,$BF,$25,$FE,$2B
	.dc.b	$AF,$95,$1B,$9D,$47,$F8,$A5,$B5,$F3,$67,$E4,$92,$7F,$C4,$58,$BF
	.dc.b	$08,$70,$73,$AF,$99,$39,$8F,$93,$E4,$CB,$F9,$3E,$4D,$C7,$DF,$36
	.dc.b	$FD,$D3,$6B,$FC,$B1,$AB,$E9,$7E,$9F,$C3,$FD,$46,$CF,$D4,$6E,$3F
	.dc.b	$29,$B8,$7F,$93,$3E,$95,$7F,$1D,$B7,$FC,$3B,$7F,$F7,$6C,$63,$0F
	.dc.b	$F1,$13,$5F,$F5,$13,$5F,$F3,$5B,$5F,$F6,$AB,$53,$CE,$FF,$CD,$BE
	.dc.b	$53,$F5,$FF,$F6,$2D,$6B,$1F,$CA,$4E,$7D,$8F,$C8,$FC,$DF,$E4,$BE
	.dc.b	$06,$E9,$F9,$26,$E1,$F8,$86,$E5,$F8,$77,$F3,$3C,$3F,$9A,$E8,$B7
	.dc.b	$F9,$A6,$DF,$E4,$1B,$7F,$8F,$7F,$19,$F4,$DF,$F0,$06,$4D,$27,$C0
	.dc.b	$FF,$9E,$23,$FD,$1B,$F4,$DF,$EC,$CF,$8C,$FF,$3A,$36,$3F,$D5,$0F
	.dc.b	$E6,$B6,$CF,$B6,$AF,$F3,$D7,$E2,$0F,$91,$FF,$3E,$60,$8C,$3A,$8C
	.dc.b	$C6,$23,$FD,$69,$C0,$F1,$B8,$B6,$3F,$D0,$CD,$77,$1D,$BE,$95,$F4
	.dc.b	$F4,$93,$FE,$8D,$8C,$BF,$E8,$FF,$A3,$7D,$37,$99,$B1,$98,$C1,$E2
	.dc.b	$FA,$39,$1E,$26,$BF,$EC,$BF,$A3,$79,$3F,$01,$B1,$FE,$9D,$6B,$B8
	.dc.b	$F8,$BF,$47,$FE,$81,$8F,$F5,$13,$5F,$F6,$B3,$5F,$F7,$D3,$5F,$F6
	.dc.b	$1F,$C2,$BE,$83,$FD,$8B,$D1,$FE,$B2,$FA,$27,$D1,$E3,$AF,$FD,$93
	.dc.b	$18,$C3,$FD,$17,$87,$16,$E1,$B4,$27,$83,$FE,$8D,$6E,$1F,$EC,$0A
	.dc.b	$E5,$3F,$E9,$16,$C7,$8F,$0E,$93,$CD,$B3,$AC,$3F,$DB,$BE,$13,$E0
	.dc.b	$77,$F1,$5F,$C0,$30,$FF,$50,$B5,$FF,$4F,$4E,$DF,$F5,$2B,$5F,$F5
	.dc.b	$14,$6F,$1F,$F6,$AB,$6F,$BD,$F9,$2F,$85,$6B,$C2,$F0,$99,$9F,$4F
	.dc.b	$C2,$7E,$46,$7F,$DB,$FF,$47,$F7,$1B,$C1,$AE,$79,$A7,$FA,$E2,$B5
	.dc.b	$9F,$FB,$E9,$B1,$F3,$B0,$E9,$3F,$EF,$E6,$C7,$D2,$9D,$A6,$C6,$A1
	.dc.b	$F1,$BF,$AB,$C7,$CB,$F1,$5F,$42,$CF,$D2,$9D,$61,$F7,$AF,$8F,$66
	.dc.b	$EA,$47,$FB,$42,$B6,$FE,$11,$AF,$8E,$36,$FF,$B4,$63,$3F,$47,$F7
	.dc.b	$63,$D5,$F0,$1E,$33,$F0,$71,$98,$37,$14,$EE,$7F,$DC,$33,$B3,$51
	.dc.b	$98,$FF,$70,$46,$7C,$5F,$81,$EB,$49,$E3,$1E,$03,$0F,$FB,$9E,$FF
	.dc.b	$C3,$AE,$8F,$F7,$7F,$D1,$B9,$3F,$32,$B9,$69,$88,$93,$C1,$AF,$FB
	.dc.b	$CF,$E8,$9F,$46,$E1,$FD,$6F,$D1,$BE,$23,$E9,$56,$A7,$31,$F5,$72
	.dc.b	$F1,$61,$83,$F6,$AE,$15,$F3,$7F,$73,$7C,$DC,$D7,$6C,$3C,$51,$83
	.dc.b	$9E,$3F,$83,$BE,$50,$C3,$B4,$63,$0E,$8D,$BE,$6F,$09,$DE,$1E,$9C
	.dc.b	$3A,$3E,$1C,$7D,$58,$B1,$E7,$3B,$8C,$DF,$27,$D7,$8F,$69,$E5,$1E
	.dc.b	$93,$CD,$87,$67,$D6,$76,$BE,$AF,$2C,$3A,$CE,$E3,$89,$BB,$EC,$EC
	.dc.b	$ED,$87,$8D,$D9,$E1,$BE,$AF,$AE,$7E,$6B,$FA,$D3,$17,$5E,$86,$67
	.dc.b	$EB,$37,$E6,$DE,$9E,$E3,$E3,$46,$1E,$52,$8C,$47,$19,$FE,$E2,$BE
	.dc.b	$8E,$91,$FE,$EA,$DD,$3E,$FF,$C1,$3E,$8D,$D1,$F5,$AF,$FA,$1D,$FD
	.dc.b	$8A,$D4,$E6,$7B,$1F,$4E,$32,$FD,$53,$89,$9A,$EA,$E4,$E3,$7E,$28
	.dc.b	$CB,$F9,$28,$E2,$E6,$BC,$17,$E3,$AD,$EF,$F3,$63,$13,$AB,$F0,$36
	.dc.b	$F9,$47,$86,$76,$D3,$65,$8D,$9E,$37,$CF,$5F,$CA,$70,$F7,$CF,$4D
	.dc.b	$3D,$B4,$9F,$87,$1F,$E2,$F8,$E2,$7F,$B6,$B1,$FF,$41,$4E,$B1,$E9
	.dc.b	$1F,$E8,$59,$F8,$38,$62,$7F,$BA,$DF,$6B,$FF,$18,$46,$0F,$27,$47
	.dc.b	$F9,$53,$E8,$DF,$47,$DA,$24,$EF,$1F,$F7,$5C,$6E,$73,$1F,$C5,$E7
	.dc.b	$DD,$F0,$9E,$63,$FC,$BF,$F4,$7F,$E6,$58,$C9,$CE,$7E,$04,$EE,$3B
	.dc.b	$7A,$2A,$E1,$5F,$EF,$AA,$D3,$DC,$C5,$70,$9D,$C7,$48,$ED,$E8,$7F
	.dc.b	$AF,$DB,$D3,$FE,$C5,$5F,$40,$C7,$FD,$5D,$5B,$BE,$C7,$F9,$66,$F2
	.dc.b	$71,$AE,$17,$D0,$93,$C2,$B9,$6D,$F9,$A9,$EF,$87,$27,$E1,$7D,$13
	.dc.b	$E8,$BF,$D7,$CD,$8F,$A0,$79,$BE,$8F,$33,$23,$8B,$E3,$8C,$63,$E8
	.dc.b	$AE,$4D,$9D,$57,$03,$E9,$1F,$C1,$4B,$CF,$ED,$F9,$2F,$F6,$84,$66
	.dc.b	$BE,$1B,$72,$FC,$CB,$4F,$F2,$19,$F3,$27,$8D,$48,$EA,$7B,$34,$4E
	.dc.b	$6F,$E7,$FC,$93,$83,$FD,$C0,$DB,$E1,$F2,$58,$99,$1F,$8B,$9D,$57
	.dc.b	$E1,$63,$EE,$D6,$A3,$AD,$5C,$BE,$A6,$2A,$C5,$FF,$87,$FE,$85,$ED
	.dc.b	$FF,$BC,$5B,$1F,$A3,$FA,$37,$D1,$BE,$53,$57,$2B,$F3,$DF,$43,$FE
	.dc.b	$03,$7F,$41,$87,$2D,$7F,$1B,$39,$6F,$8B,$FE,$FF,$8C,$6B,$FC,$1A
	.dc.b	$B8,$3F,$C2,$23,$37,$FE,$35,$FA,$37,$C1,$FF,$28,$AF,$E1,$BF,$47
	.dc.b	$C2,$B5,$39,$7F,$C7,$91,$9C,$BF,$CF,$6D,$58,$7F,$9B,$E7,$F7,$13
	.dc.b	$98,$C4,$FA,$BD,$2E,$8F,$FA,$13,$A5,$F5,$9E,$B9,$CC,$FE,$BE,$2F
	.dc.b	$D2,$72,$CF,$F9,$1B,$72,$E5,$F2,$6F,$93,$DE,$32,$7F,$8A,$7D,$27
	.dc.b	$F9,$73,$E7,$F1,$9F,$F0,$58,$C4,$FF,$0F,$C7,$A4,$A7,$2F,$EF,$A3
	.dc.b	$31,$8C,$BF,$71,$F4,$8F,$8C,$FE,$10,$DC,$3F,$6A,$FF,$B4,$DB,$1A
	.dc.b	$23,$AF,$66,$E3,$F9,$5A,$FB,$2F,$F6,$A7,$FC,$95,$3D,$39,$7B,$FA
	.dc.b	$5E,$06,$C6,$4F,$0C,$7C,$CF,$A4,$FF,$3C,$C7,$F8,$0D,$B7,$FE,$0B
	.dc.b	$6B,$FE,$0F,$6B,$FE,$7D,$77,$97,$F7,$0E,$27,$0B,$DA,$ED,$ED,$7D
	.dc.b	$E9,$CC,$63,$1F,$E7,$35,$B7,$F9,$F6,$1D,$1F,$B7,$3F,$2B,$E9,$24
	.dc.b	$7C,$AF,$9C,$B9,$FC,$59,$B9,$FD,$29,$B5,$FC,$3B,$67,$D2,$7D,$4F
	.dc.b	$A4,$FA,$AD,$9F,$AA,$DC,$7E,$83,$70,$FE,$E3,$F4,$7E,$79,$FF,$18
	.dc.b	$35,$FF,$1A,$35,$FF,$51,$35,$FF,$70,$BC,$63,$E4,$6A,$F9,$6F,$CF
	.dc.b	$CE,$70,$FC,$DC,$63,$5F,$95,$5B,$9F,$EC,$95,$BC,$3F,$AA,$4E,$5F
	.dc.b	$C4,$73,$FE,$48,$FC,$9F,$45,$BF,$C9,$C5,$9F,$C7,$4D,$CB,$FC,$57
	.dc.b	$E3,$FF,$AE,$EF,$FB,$E3,$FB,$76,$D7,$F3,$A3,$FB,$74,$65,$A6,$67
	.dc.b	$9B,$FD,$88,$C9,$FC,$0A,$38,$47,$CE,$F3,$BD,$FF,$47,$F1,$1B,$7F
	.dc.b	$01,$B7,$FD,$C3,$3F,$ED,$33,$FC,$0A,$B8,$3F,$8E,$6C,$FF,$80,$F9
	.dc.b	$3F,$2F,$1E,$18,$FF,$54,$4F,$B3,$D5,$C7,$97,$96,$46,$4F,$83,$18
	.dc.b	$9F,$7F,$93,$CF,$E0,$3A,$36,$BE,$3B,$67,$71,$F2,$A7,$F8,$FF,$D0
	.dc.b	$B9,$FF,$92,$FA,$EF,$A1,$FE,$11,$87,$F8,$3A,$B5,$87,$F8,$BE,$B5
	.dc.b	$DF,$CB,$0E,$38,$70,$E1,$FF,$24,$E1,$C3,$1F,$F1,$26,$1C,$31,$FF
	.dc.b	$17,$4E,$6B,$FC,$D7,$39,$AF,$06,$1D,$31,$ED,$87,$87,$1E,$CF,$FC
	.dc.b	$01,$C7,$92,$BF,$F2,$C4,$E9,$C4,$E9,$CD,$74,$7C,$B3,$B5,$3B,$C3
	.dc.b	$C9,$8F,$B3,$1F,$26,$3E,$A6,$DF,$A2,$3C,$D8,$F5,$AF,$6E,$3D,$71
	.dc.b	$E8,$DC,$38,$CF,$5B,$E2,$75,$2E,$1D,$4F,$3E,$1D,$63,$E0,$63,$E6
	.dc.b	$7E,$25,$F4,$69,$CA,$FB,$5F,$28,$EC,$D3,$B4,$7A,$2F,$B4,$7C,$89
	.dc.b	$DC,$E5,$A7,$8B,$0F,$63,$DC,$F0,$CF,$59,$F0,$9C,$5E,$F1,$C6,$BB
	.dc.b	$E1,$EB,$9F,$B3,$87,$AD,$B5,$D5,$A7,$57,$BD,$F5,$3B,$BD,$4F,$3C
	.dc.b	$7A,$CF,$09,$63,$6D,$8E,$67,$17,$75,$E5,$7F,$0F,$5E,$56,$CF,$95
	.dc.b	$A7,$94,$F3,$D7,$96,$3A,$C7,$73,$2F,$78,$F1,$D7,$7C,$7F,$13,$87
	.dc.b	$C6,$C7,$E7,$E1,$D4,$EF,$7D,$5E,$67,$53,$38,$75,$3C,$38,$75,$C7
	.dc.b	$D5,$8F,$69,$F8,$F3,$97,$D9,$19,$AF,$87,$87,$8B,$0F,$04,$76,$3C
	.dc.b	$13,$B8,$F5,$CE,$EB,$C5,$5E,$6B,$EC,$FA,$EF,$84,$F9,$AB,$84,$F4
	.dc.b	$C3,$CE,$49,$F2,$B6,$39,$37,$0F,$2D,$F4,$9F,$05,$F4,$C3,$C3,$5A
	.dc.b	$BE,$18,$79,$AF,$D9,$5A,$8C,$61,$CE,$73,$87,$32,$63,$F5,$C9,$5F
	.dc.b	$32,$32,$D8,$CC,$FC,$C7,$FA,$47,$D0,$FF,$49,$F4,$2D,$7F,$64,$6D
	.dc.b	$7F,$62,$6E,$7F,$D8,$9F,$EA,$1F,$43,$2B,$FB,$5F,$C3,$5B,$FE,$D3
	.dc.b	$36,$FF,$B4,$CD,$AF,$E5,$94,$F9,$F7,$C2,$FF,$81,$B9,$FF,$CD,$31
	.dc.b	$8A,$FF,$17,$C6,$27,$51,$FE,$ED,$9D,$57,$F6,$5E,$BF,$ED,$6F,$F3
	.dc.b	$98,$CE,$3F,$E1,$8C,$39,$5F,$43,$FC,$0B,$F0,$27,$FB,$E3,$FE,$73
	.dc.b	$72,$63,$CA,$CB,$4C,$9C,$9E,$3F,$43,$23,$AF,$D0,$DC,$BF,$C1,$F6
	.dc.b	$AC,$63,$FE,$18,$85,$62,$FE,$3E,$4F,$A0,$FF,$48,$DB,$E8,$5F,$3B
	.dc.b	$F2,$27,$31,$8B,$F8,$D1,$89,$D4,$7F,$5D,$AF,$F5,$55,$6F,$34,$7C
	.dc.b	$37,$E5,$5F,$1A,$E6,$71,$39,$57,$1A,$F5,$7D,$0F,$CA,$3E,$5F,$8B
	.dc.b	$6B,$E7,$C5,$C3,$E1,$CD,$AF,$85,$F4,$3F,$4B,$E3,$FE,$9B,$70,$FA
	.dc.b	$ED,$BF,$96,$DA,$F4,$43,$E8,$57,$88,$FC,$8C,$62,$FF,$27,$18,$9D
	.dc.b	$3F,$E0,$19,$D3,$A3,$A4,$7F,$A1,$7E,$87,$FD,$4D,$1F,$22,$B9,$1F
	.dc.b	$15,$78,$89,$DE,$1E,$6C,$3A,$BF,$E9,$38,$CD,$F2,$BF,$F2,$04,$61
	.dc.b	$78,$CE,$27,$CE,$FE,$63,$D9,$FC,$CB,$70,$FC,$EB,$6F,$F2,$2D,$C7
	.dc.b	$F2,$1C,$5F,$A0,$E7,$FD,$0B,$6F,$F4,$AD,$BF,$CC,$B6,$BE,$7F,$D0
	.dc.b	$3E,$81,$F0,$FF,$1E,$D3,$98,$C3,$F9,$EE,$0B,$5C,$2B,$E7,$7D,$03
	.dc.b	$C7,$E5,$08,$C3,$D5,$AF,$08,$DF,$6B,$E0,$69,$27,$C9,$19,$7C,$DB
	.dc.b	$BC,$4F,$E2,$BA,$FE,$2A,$31,$3F,$6C,$F9,$F3,$AD,$FF,$8D,$FD,$43
	.dc.b	$E3,$B1,$5F,$E4,$9F,$A0,$7D,$0E,$21,$18,$7A,$35,$FC,$51,$FE,$51
	.dc.b	$FA,$87,$D4,$71,$FA,$37,$D1,$D2,$57,$96,$1F,$4F,$DF,$67,$D2,$52
	.dc.b	$57,$AF,$E9,$1E,$67,$95,$B1,$98,$C1,$E9,$FA,$4F,$F3,$F4,$7C,$36
	.dc.b	$BE,$80,$E3,$8F,$F6,$6D,$E2,$FE,$83,$F8,$4B,$E5,$7E,$27,$C3,$3E
	.dc.b	$97,$E3,$7D,$2B,$E9,$69,$A8,$99,$FD,$39,$D3,$67,$2F,$8E,$BA,$39
	.dc.b	$21,$9A,$FA,$9F,$4A,$F8,$1F,$AD,$F1,$5F,$C2,$2B,$F0,$0D,$7D,$D3
	.dc.b	$B7,$F0,$8D,$7E,$14,$6F,$6F,$E2,$6D,$D7,$FB,$0D,$6E,$7F,$8B,$B6
	.dc.b	$3F,$B2,$CE,$EB,$FD,$24,$D5,$3B,$7F,$D2,$DF,$09,$F9,$67,$E9,$7D
	.dc.b	$1F,$DA,$6F,$0F,$E4,$67,$58,$FF,$26,$6E,$3F,$DD,$2F,$84,$FF,$2A
	.dc.b	$6C,$7F,$78,$9D,$DF,$83,$BB,$F3,$95,$A8,$FF,$54,$35,$4F,$05,$F4
	.dc.b	$AF,$9E,$E3,$F1,$B2,$7F,$5D,$39,$6C,$E5,$F6,$CF,$EA,$E7,$51,$F9
	.dc.b	$16,$E5,$E9,$9E,$0F,$F4,$48,$CE,$5F,$E1,$2F,$A5,$FC,$2D,$FC,$8F
	.dc.b	$17,$D6,$7E,$DE,$33,$7F,$D3,$1B,$5F,$1E,$77,$5F,$C1,$67,$2F,$FB
	.dc.b	$25,$AB,$3F,$E2,$F7,$F3,$6B,$B7,$0F,$7F,$33,$E5,$FB,$93,$0E,$91
	.dc.b	$97,$F8,$96,$1F,$E2,$86,$CE,$5F,$D9,$7D,$2B,$8F,$F8,$82,$E5,$A6
	.dc.b	$22,$4F,$F1,$D9,$E0,$D8,$FE,$31,$F4,$8F,$A5,$FF,$02,$CB,$E9,$1F
	.dc.b	$15,$FE,$3C,$AD,$6F,$FE,$E9,$9C,$D7,$FB,$AF,$1E,$AF,$1B,$FF,$21
	.dc.b	$C6,$27,$72,$BE,$6E,$CF,$F7,$A3,$88,$DC,$F1,$AE,$78,$FF,$93,$A7
	.dc.b	$9D,$6E,$7C,$D8,$FA,$A7,$8D,$6E,$78,$E1,$E9,$77,$8D,$8F,$0D,$F1
	.dc.b	$9D,$C7,$94,$63,$E3,$77,$5C,$5F,$65,$71,$7E,$25,$F8,$5A,$6F,$1F
	.dc.b	$43,$C8,$ED,$7D,$5E,$F3,$C5,$EB,$7C,$6F,$B4,$68,$F1,$57,$1B,$EB
	.dc.b	$1C,$0F,$F0,$F3,$F4,$1F,$F1,$2B,$5F,$86,$6B,$E7,$FF,$C2,$4D,$F5
	.dc.b	$7F,$C0,$12,$9C,$CF,$26,$BF,$E2,$BF,$A5,$FE,$FA,$9A,$47,$FC,$02
	.dc.b	$FC,$9F,$89,$FE,$DD,$CB,$F2,$E4,$71,$3E,$94,$69,$FA,$86,$A7,$97
	.dc.b	$9B,$FE,$5D,$DF,$F1,$F3,$93,$8C,$67,$7F,$34,$63,$0F,$03,$DE,$BF
	.dc.b	$2F,$87,$77,$9C,$66,$BF,$D3,$F3,$97,$9C,$F0,$BE,$D5,$FE,$52,$C3
	.dc.b	$CB,$87,$66,$C7,$A8,$E8,$7C,$0C,$3A,$3C,$4F,$31,$B8,$E3,$7E,$46
	.dc.b	$E1,$E4,$C3,$FC,$FD,$B7,$F9,$A3,$0F,$90,$FF,$98,$AB,$C2,$E6,$77
	.dc.b	$8F,$13,$FC,$49,$8F,$C9,$C7,$D9,$1F,$E2,$7C,$3F,$9F,$37,$A7,$8D
	.dc.b	$7F,$99,$3E,$95,$F4,$B8,$85,$F0,$7F,$C1,$8D,$CB,$CD,$7C,$BD,$56
	.dc.b	$1C,$B1,$EC,$78,$7E,$23,$A6,$07,$87,$D2,$FF,$9C,$27,$EA,$C6,$88
	.dc.b	$FF,$9E,$6D,$3B,$CB,$E8,$56,$9F,$2B,$D3,$DF,$C6,$7F,$E9,$2F,$6B
	.dc.b	$FD,$A7,$5F,$E6,$4E,$BF,$83,$93,$E8,$FF,$D0,$AF,$F9,$36,$B6,$49
	.dc.b	$DC,$EA,$BF,$C5,$D3,$9A,$E1,$5A,$CF,$E5,$4F,$83,$E9,$5F,$4B,$F3
	.dc.b	$5B,$7F,$E9,$53,$C5,$E5,$E6,$64,$7F,$97,$E7,$78,$FF,$95,$23,$66
	.dc.b	$67,$71,$FE,$41,$9D,$E7,$FE,$C9,$AD,$CE,$6B,$EF,$47,$F0,$53,$FC
	.dc.b	$D8,$58,$ED,$9B,$E4,$B7,$D1,$2B,$6F,$C4,$8C,$57,$F9,$46,$79,$61
	.dc.b	$FE,$C2,$69,$FC,$18,$FA,$13,$C2,$3F,$D8,$FF,$03,$F8,$59,$FF,$3F
	.dc.b	$36,$BC,$7F,$25,$F9,$58,$C5,$70,$E1,$C1,$F5,$A6,$DF,$28,$B7,$CA
	.dc.b	$30,$7F,$A2,$FE,$89,$F4,$7F,$A1,$6C,$7F,$B2,$7E,$95,$F4,$BF,$E9
	.dc.b	$B8,$E1,$5B,$3F,$51,$F4,$7F,$DF,$63,$27,$FA,$5D,$C6,$1F,$E7,$56
	.dc.b	$E7,$FE,$E6,$6C,$FF,$2A,$96,$3E,$27,$87,$8B,$FB,$5F,$A5,$7C,$47
	.dc.b	$ED,$FE,$97,$59,$7F,$BC,$27,$38,$FF,$A9,$E7,$2F,$56,$E1,$FC,$E3
	.dc.b	$2F,$F3,$84,$EF,$3F,$5F,$A9,$23,$FD,$61,$D2,$FA,$5C,$61,$FE,$8F
	.dc.b	$8C,$4F,$E6,$FE,$83,$FD,$CC,$9A,$7F,$AF,$5D,$7D,$1F,$EC,$A3,$2D
	.dc.b	$F5,$70,$C9,$D9,$D6,$31,$7F,$A3,$AF,$92,$A7,$C1,$E6,$F5,$E1,$EC
	.dc.b	$D8,$F3,$36,$79,$3F,$C8,$7E,$8D,$EA,$D3,$F8,$F7,$C4,$7F,$BB,$6F
	.dc.b	$FD,$A5,$3A,$6C,$F8,$6F,$ED,$1F,$D9,$27,$4E,$03,$87,$6D,$7F,$4D
	.dc.b	$8B,$E8,$FE,$31,$F1,$FE,$7A,$E1,$F3,$E2,$D7,$C3,$9B,$CB,$FC,$06
	.dc.b	$F9,$5E,$CF,$AD,$3A,$7B,$FD,$17,$D3,$FA,$3F,$A8,$D8,$FB,$8D,$C3
	.dc.b	$E8,$37,$0F,$81,$0F,$A2,$FF,$47,$D7,$FB,$DD,$B1,$FE,$EC,$E2,$6C
	.dc.b	$F9,$7F,$EA,$15,$AC,$39,$BC,$2F,$E5,$B5,$FD,$33,$5F,$F0,$8B,$70
	.dc.b	$FC,$87,$D1,$FE,$34,$FC,$77,$35,$8F,$CB,$45,$AF,$C5,$4D,$BF,$91
	.dc.b	$36,$BE,$9C,$DA,$78,$3F,$B7,$9F,$F1,$0D,$FE,$06,$3F,$C4,$79,$FF
	.dc.b	$6E,$75,$49,$D3,$FD,$2D,$AE,$CC,$57,$F8,$B2,$78,$CF,$0A,$F8,$FE
	.dc.b	$4C,$AF,$EC,$DA,$3E,$23,$18,$F7,$9F,$8F,$5F,$5A,$2C,$7E,$4D,$FB
	.dc.b	$11,$97,$E5,$B7,$1F,$F1,$EF,$A9,$E5,$3C,$DE,$2D,$BF,$82,$79,$EF
	.dc.b	$FD,$26,$E1,$F8,$E6,$0F,$21,$88,$E4,$E2,$BB,$CF,$16,$C6,$4F,$EA
	.dc.b	$91,$97,$A6,$7F,$33,$C5,$F9,$5E,$E7,$CB,$6D,$6D,$B5,$CD,$B5,$F9
	.dc.b	$AF,$A2,$72,$FF,$88,$DB,$7E,$BF,$A7,$7D,$47,$F4,$3C,$7D,$62,$B6
	.dc.b	$E1,$B8,$75,$C3,$FC,$63,$7C,$BE,$4F,$B4,$7E,$07,$E8,$FF,$05,$E5
	.dc.b	$DC,$FD,$CD,$CB,$E1,$37,$1F,$C5,$B6,$7E,$93,$67,$E8,$B7,$6F,$F3
	.dc.b	$BB,$63,$FB,$5C,$FD,$9B,$F0,$BF,$35,$B1,$F4,$5B,$7F,$9A,$6C,$F1
	.dc.b	$3C,$E7,$12,$CF,$16,$CF,$16,$9C,$63,$DA,$F1,$69,$C4,$B1,$C5,$B1
	.dc.b	$C6,$3E,$1B,$C5,$B1,$C5,$A7,$12,$C7,$13,$E4,$9C,$4B,$1C,$5B,$1C
	.dc.b	$5A,$71,$8F,$9E,$F1,$69,$C4,$B1,$C5,$B1,$C6,$FE,$AB,$C5,$B1,$C5
	.dc.b	$B3,$C5,$B3,$C6,$7E,$D9,$FB,$0A,$FA,$AD,$8F,$9E,$D3,$E4,$96,$BE
	.dc.b	$1B,$63,$DA,$D3,$CE,$5A,$F0,$B4,$F0,$BF,$EC,$4A,$F0,$B6,$7C,$2D
	.dc.b	$9F,$0B,$6B,$C2,$D9,$F0,$B6,$7C,$2D,$3C,$25,$7C,$22,$BC,$2D,$9F
	.dc.b	$0B,$4F,$09,$6B,$C2,$D8,$F0,$B4,$F0,$96,$BC,$2D,$3C,$35,$E0,$24
	.dc.b	$F8,$1B,$1E,$06,$D7,$81,$B3,$E0,$6C,$F8,$1B,$5E,$06,$C7,$82,$7B
	.dc.b	$11,$EC,$DA,$EC,$D3,$B1,$63,$B3,$6B,$B3,$63,$B1,$D0,$93,$D1,$B1
	.dc.b	$D1,$A7,$42,$CF,$46,$9F,$63,$E8,$FF,$B1,$1F,$D8,$FE,$1A,$D7,$F3
	.dc.b	$F8,$B3,$FD,$0E,$6D,$FF,$84,$BE,$81,$F0,$5F,$ED,$B6,$E1,$FD,$9B
	.dc.b	$E6,$B1,$32,$3F,$AB,$60,$AD,$BF,$E1,$7E,$85,$EE,$3F,$AB,$7C,$0F
	.dc.b	$F9,$3E,$FF,$AB,$2F,$A3,$FF,$00,$39,$9D,$C6,$27,$2D,$78,$B6,$37
	.dc.b	$1D,$0D,$9D,$BE,$8F,$A9,$DB,$E8,$EC,$64,$93,$E6,$8C,$BF,$D5,$B2
	.dc.b	$7D,$1F,$C5,$3F,$CB,$1A,$FF,$96,$7D,$E7,$C6,$FF,$8E,$F1,$F8,$B7
	.dc.b	$C1,$B7,$FE,$07,$C3,$E4,$36,$75,$1F,$E5,$F6,$BF,$E9,$DE,$26,$BF
	.dc.b	$E4,$1B,$F9,$3E,$27,$C8,$FA,$2F,$92,$C8,$C9,$F3,$A3,$17,$F4,$A6
	.dc.b	$E9,$FE,$31,$B5,$62,$FF,$C5,$50,$FA,$2F,$AB,$F0,$1F,$59,$B3,$F6
	.dc.b	$9B,$8F,$DE,$6E,$1F,$E5,$EF,$A2,$5F,$C8,$BF,$F5,$C4,$61,$FF,$10
	.dc.b	$6F,$F2,$BD,$0D,$62,$15,$B7,$FC,$12,$D7,$FD,$12,$D7,$AD,$FC,$F8
	.dc.b	$C4,$7F,$94,$B2,$7C,$BF,$E4,$3D,$9F,$C8,$B7,$2F,$CA,$B6,$BF,$32
	.dc.b	$DC,$FE,$43,$67,$E3,$B6,$7B,$73,$7E,$73,$C6,$FC,$EB,$6B,$F4,$2D
	.dc.b	$9F,$D2,$B6,$FF,$D0,$BF,$51,$FE,$DB,$9F,$F5,$2B,$5F,$F7,$03,$5F
	.dc.b	$8C,$D5,$F4,$4F,$A2,$7C,$2F,$F9,$66,$30,$DE,$97,$99,$9A,$DD,$FF
	.dc.b	$AC,$7C,$9F,$D0,$BF,$C0,$17,$DC,$7F,$BC,$73,$FF,$4D,$C6,$27,$5D
	.dc.b	$79,$DB,$FD,$ED,$3C,$A7,$FC,$82,$ED,$B7,$B7,$F0,$11,$BB,$EB,$79
	.dc.b	$94,$F1,$9F,$0D,$66,$38,$5F,$E2,$6B,$EF,$FB,$1C,$78,$4D,$3A,$7D
	.dc.b	$13,$E3,$7B,$7D,$0B,$E8,$A9,$30,$FE,$43,$F4,$4F,$A2,$F2,$7D,$0B
	.dc.b	$E8,$A9,$2B,$F9,$3C,$7F,$A3,$DF,$E5,$3A,$3E,$8B,$C8,$48,$FE,$5B
	.dc.b	$F4,$4F,$33,$E0,$36,$33,$18,$3E,$27,$D1,$7F,$9C,$A3,$F9,$AB,$5F
	.dc.b	$F3,$7B,$79,$3F,$DA,$2D,$E2,$F8,$DF,$44,$F9,$2F,$9D,$E2,$3E,$77
	.dc.b	$E8,$7D,$13,$E8,$A9,$A8,$99,$FF,$44,$9D,$36,$72,$FF,$A0,$2B,$A3
	.dc.b	$92,$19,$AF,$BD,$E7,$3D,$FF,$C0,$7C,$47,$F0,$AA,$FE,$A2,$D7,$FD
	.dc.b	$17,$3B,$7F,$AA,$B5,$FF,$47,$C6,$F6,$FF,$78,$B7,$5F,$C3,$56,$E7
	.dc.b	$FD,$EA,$D8,$FC,$54,$EE,$BF,$8A,$35,$FE,$83,$7F,$95,$FA,$2F,$C6
	.dc.b	$C7,$F6,$6F,$A0,$FB,$0F,$FA,$81,$B8,$B0,$FC,$B5,$70,$AF,$F5,$2B
	.dc.b	$67,$51,$FC,$8E,$B8,$D7,$E7,$5B,$1F,$C9,$EB,$84,$F4,$ED,$FD,$45
	.dc.b	$6A,$8D,$AD,$C7,$F2,$FF,$A2,$7C,$DF,$0F,$8C,$93,$FE,$00,$9C,$B6
	.dc.b	$72,$FB,$67,$F6,$D3,$A8,$FF,$5F,$B7,$2F,$5C,$78,$A7,$8C,$62,$75
	.dc.b	$97,$EE,$7E,$8B,$F0,$F7,$F2,$7C,$2F,$59,$FC,$0E,$33,$7D,$1B,$5F
	.dc.b	$16,$77,$5F,$E1,$89,$CB,$FD,$35,$AF,$C4,$8D,$BF,$ED,$68,$C8,$F3
	.dc.b	$3E,$5F,$B5,$27,$A4,$78,$CC,$3F,$7E,$30,$FF,$8A,$3E,$85,$C3,$FC
	.dc.b	$6D,$72,$53,$11,$27,$83,$5F,$F7,$4F,$D0,$BE,$89,$AF,$F6,$7F,$A1
	.dc.b	$7C,$4F,$F2,$1A,$D4,$E6,$3F,$92,$65,$E2,$C3,$07,$F6,$F7,$0A,$F9
	.dc.b	$BF,$DC,$EF,$9B,$9A,$ED,$87,$8A,$30,$73,$C7,$F9,$95,$F2,$86,$1D
	.dc.b	$A3,$18,$74,$6D,$F3,$78,$4E,$F0,$F4,$E1,$D1,$F0,$E3,$EA,$C5,$8F
	.dc.b	$39,$DC,$66,$F9,$3E,$BC,$7B,$4F,$28,$F4,$9E,$6C,$3B,$3E,$B3,$B5
	.dc.b	$F5,$79,$61,$D6,$77,$1C,$4D,$DF,$67,$67,$6C,$3C,$6E,$CF,$0D,$F5
	.dc.b	$7D,$73,$F3,$5F,$EB,$86,$2E,$BD,$0C,$CF,$D6,$6F,$CD,$FF,$92,$9D
	.dc.b	$C2,$31,$1A,$F5,$79,$BF,$2A,$32,$F6,$FA,$2C,$BF,$E3,$56,$A6,$91
	.dc.b	$F9,$7E,$CF,$F6,$DE,$3F,$AD,$1C,$4C,$F9,$9F,$C7,$B7,$FC,$34,$65
	.dc.b	$FC,$45,$71,$8F,$AA,$67,$7F,$EF,$D1,$99,$E3,$3E,$33,$AD,$F8,$1C
	.dc.b	$47,$4B,$F0,$5F,$B6,$76,$76,$36,$F6,$8D,$C7,$D2,$C3,$F2,$18,$78
	.dc.b	$6B,$95,$F5,$C3,$C1,$5A,$C3,$8E,$1F,$E1,$58,$C4,$EA,$3C,$35,$EA
	.dc.b	$9C,$C7,$93,$4F,$E5,$38,$FC,$37,$F9,$5C,$7C,$33,$FC,$D1,$19,$BF
	.dc.b	$41,$FE,$B8,$71,$1E,$93,$A4,$7C,$AE,$7F,$F3,$8C,$27,$FA,$1C,$FF
	.dc.b	$9C,$EB,$FA,$9F,$A2,$FA,$17,$D1,$70,$BD,$E1,$FC,$53,$0E,$87,$87
	.dc.b	$1F,$F3,$EE,$1C,$6F,$84,$71,$F8,$69,$5D,$CF,$F4,$15,$3E,$8B,$FD
	.dc.b	$05,$9F,$C2,$AD,$E3,$DC,$F8,$2F,$69,$DD,$FE,$1A,$37,$5C,$63,$B5
	.dc.b	$7F,$98,$AB,$6F,$28,$FF,$6F,$FA,$FF,$D0,$5B,$D3,$FE,$96,$5F,$41
	.dc.b	$FE,$6A,$72,$4A,$F2,$1F,$E7,$29,$E1,$3B,$8C,$D7,$56,$CF,$D2,$8C
	.dc.b	$61,$FE,$02,$BE,$F9,$66,$BE,$BF,$D1,$3E,$8A,$9D,$BC,$9B,$7F,$EF
	.dc.b	$89,$D3,$67,$4D,$26,$5E,$89,$DB,$F3,$63,$DF,$5D,$23,$DB,$1F,$C2
	.dc.b	$CF,$96,$FF,$87,$79,$BD,$FF,$23,$F8,$C9,$D4,$F0,$9C,$1A,$8E,$26
	.dc.b	$B3,$FF,$03,$B4,$FF,$3E,$16,$7F,$C5,$5A,$AF,$9C,$F8,$BE,$15,$C1
	.dc.b	$F9,$C6,$C7,$E0,$0F,$B3,$F2,$39,$99,$1C,$5F,$AF,$18,$C3,$ED,$D7
	.dc.b	$26,$CE,$AB,$81,$F0,$BE,$81,$F3,$BF,$AC,$6C,$7E,$43,$E8,$9F,$44
	.dc.b	$F9,$2D,$5C,$AF,$97,$F4,$1F,$E0,$67,$A4,$7E,$DA,$31,$97,$E3,$9B
	.dc.b	$BF,$F8,$B9,$B5,$FE,$A3,$CB,$C3,$5B,$F1,$7F,$7F,$F4,$4F,$88,$FE
	.dc.b	$03,$F4,$5A,$CB,$FC,$6F,$39,$C7,$EE,$CE,$5F,$F5,$73,$70,$FF,$7C
	.dc.b	$65,$FE,$79,$9D,$E7,$E7,$F4,$64,$7E,$1B,$A9,$F4,$58,$C3,$F4,$F1
	.dc.b	$89,$FA,$DF,$3E,$FA,$2F,$D8,$36,$33,$EA,$62,$64,$70,$77,$18,$BF
	.dc.b	$CD,$57,$16,$CE,$AB,$81,$E0,$E0,$EF,$E6,$F8,$1B,$19,$3C,$B1,$96
	.dc.b	$9F,$A2,$FA,$27,$A7,$FC,$DB,$00,$8C,$CF,$F7,$06,$C7,$F7,$C6,$BF
	.dc.b	$E0,$07,$3C,$5F,$CC,$BE,$59,$EA,$FF,$38,$9C,$DF,$F2,$89,$CB,$F2
	.dc.b	$BE,$8B,$E4,$9F,$2B,$E6,$AC,$64,$98,$FC,$68,$CB,$F2,$26,$E3,$FC
	.dc.b	$4F,$65,$62,$FD,$A2,$1F,$43,$F6,$3E,$8B,$EC,$B6,$7E,$EB,$71,$FA
	.dc.b	$AD,$C3,$FD,$0F,$F4,$5E,$A9,$FF,$0F,$B5,$FF,$11,$B5,$FF,$80,$0D
	.dc.b	$7E,$7B,$C6,$3E,$47,$CB,$FF,$5E,$8C,$5F,$C7,$8F,$C6,$FD,$14,$8F
	.dc.b	$C8,$F3,$DC,$3F,$13,$37,$0F,$85,$36,$BE,$0E,$6A,$C5,$FF,$AF,$E3
	.dc.b	$F4,$8F,$29,$FE,$A3,$E1,$7F,$1C,$8F,$EB,$4F,$F6,$08,$CC,$7C,$C3
	.dc.b	$27,$F4,$9C,$3F,$D8,$B3,$C5,$B1,$99,$FC,$7C,$E5,$B1,$93,$E2,$F9
	.dc.b	$5F,$DD,$38,$1F,$4F,$F1,$63,$0D,$AF,$EC,$FD,$2A,$DC,$7F,$40,$DB
	.dc.b	$FA,$BB,$F9,$1A,$FF,$6C,$E5,$E4,$F5,$6C,$F8,$9B,$93,$3F,$D2,$F8
	.dc.b	$F8,$7F,$4F,$36,$7F,$D3,$B5,$F1,$23,$15,$96,$46,$4F,$93,$18,$9F
	.dc.b	$93,$E3,$F3,$F7,$3A,$36,$BE,$7B,$67,$71,$F4,$A7,$FC,$4B,$F4,$3F
	.dc.b	$A0,$8F,$F0,$4F,$A6,$FA,$77,$81,$AE,$9F,$6C,$E6,$31,$5C,$2B,$6F
	.dc.b	$F6,$8B,$E9,$8F,$F8,$6E,$F8,$4F,$E0,$BE,$87,$E1,$79,$9F,$0D,$B1
	.dc.b	$93,$F0,$F1,$96,$99,$AF,$C4,$C6,$5B,$19,$E0,$FC,$4C,$E5,$B1,$93
	.dc.b	$F1,$31,$96,$99,$9F,$B4,$7F,$65,$9F,$C9,$D7,$9A,$3F,$32,$D8,$CC
	.dc.b	$F9,$A7,$2D,$8C,$C7,$98,$C9,$5C,$9E,$67,$2D,$33,$1E,$68,$CB,$63
	.dc.b	$31,$E8,$AF,$63,$67,$D9,$C1,$FC,$E2,$72,$D8,$CC,$FB,$DF,$D6,$DF
	.dc.b	$B5,$A7,$AC,$B1,$E9,$6D,$F9,$DA,$79,$4B,$1E,$33,$D2,$58,$F4,$D7
	.dc.b	$8E,$32,$D8,$CC,$79,$4F,$AB,$5E,$56,$D7,$95,$A7,$94,$B3,$E5,$6C
	.dc.b	$79,$5A,$78,$DF,$3D,$78,$5B,$1E,$18,$FE,$C3,$5E,$16,$CF,$85,$B3
	.dc.b	$E1,$6D,$78,$5B,$3E,$16,$CF,$84,$F1,$BE,$16,$9E,$11,$5E,$16,$CF
	.dc.b	$85,$A7,$84,$B5,$E1,$6C,$78,$5A,$78,$6B,$C6,$78,$4A,$F8,$6B,$C0
	.dc.b	$49,$F0,$36,$3C,$0D,$AF,$03,$67,$C0,$D9,$F0,$1E,$2A,$F0,$34,$F0
	.dc.b	$4F,$62,$3D,$9B,$5D,$9A,$76,$2C,$76,$AF,$05,$76,$6C,$76,$3A,$12
	.dc.b	$BA,$36,$BA,$34,$E8,$F6,$AE,$8D,$8E,$93,$C8,$91,$F4,$7E,$87,$FA
	.dc.b	$21,$FD,$17,$E2,$6C,$7F,$68,$8B,$5F,$DA,$26,$D2,$BF,$C6,$72,$F9
	.dc.b	$F7,$C0,$FF,$63,$E9,$FF,$71,$C6,$2B,$FB,$0F,$47,$F5,$EB,$FF,$1D
	.dc.b	$CE,$5F,$F6,$97,$4F,$F6,$5B,$E3,$87,$3C,$3F,$BF,$FB,$DE,$16,$FD
	.dc.b	$0F,$F5,$F8,$C6,$1F,$E0,$25,$19,$9D,$35,$E2,$74,$32,$53,$33,$CF
	.dc.b	$E8,$7B,$1D,$FE,$6A,$CF,$92,$2C,$F8,$66,$EB,$F1,$E3,$DF,$F3,$3F
	.dc.b	$DE,$21,$F4,$EF,$9E,$F8,$F3,$98,$C6,$3F,$EC,$4D,$D3,$9A,$DC,$FF
	.dc.b	$BC,$AF,$9C,$7C,$3F,$A2,$93,$F2,$3C,$5B,$97,$C5,$9B,$7F,$E2,$2D
	.dc.b	$95,$8B,$FF,$10,$C2,$B1,$7F,$E0,$F8,$DC,$FD,$49,$FA,$5F,$45,$F4
	.dc.b	$FE,$8A,$CF,$D8,$6D,$FD,$56,$E3,$E1,$7F,$C3,$70,$FA,$24,$7F,$A0
	.dc.b	$63,$0D,$8C,$D7,$E4,$A7,$F2,$71,$87,$07,$08,$FF,$18,$FD,$3F,$FA
	.dc.b	$66,$FE,$A4,$E4,$FC,$8C,$66,$7E,$0F,$CC,$7E,$37,$D9,$FC,$73,$63
	.dc.b	$F2,$6D,$AF,$C8,$36,$FE,$63,$67,$99,$F9,$AA,$FC,$C6,$DF,$99,$E6
	.dc.b	$B3,$F9,$E6,$E1,$F9,$B6,$E1,$F5,$DF,$8F,$F4,$4F,$A2,$7C,$2F,$C0
	.dc.b	$B4,$E6,$30,$B9,$9D,$58,$BE,$37,$F2,$3E,$9D,$E3,$F3,$84,$61,$EC
	.dc.b	$D7,$84,$6C,$FC,$BB,$FE,$B8,$CF,$E7,$C3,$E8,$FF,$84,$C6,$1B,$AF
	.dc.b	$F0,$C9,$E3,$4A,$E3,$D1,$A8,$FF,$61,$CF,$37,$B5,$71,$BD,$47,$AA
	.dc.b	$AD,$74,$BF,$04,$F0,$39,$D6,$AB,$8F,$0F,$F9,$52,$74,$1C,$3E,$5C
	.dc.b	$3F,$91,$C6,$2F,$F1,$BF,$46,$F8,$FF,$1F,$D1,$BE,$8E,$93,$1E,$BF
	.dc.b	$46,$FA,$3F,$4F,$D1,$3E,$8E,$92,$BC,$70,$FA,$7F,$3F,$D2,$49,$F3
	.dc.b	$EC,$8F,$87,$F4,$8F,$2A,$92,$BD,$B1,$97,$DF,$F4,$9F,$ED,$66,$FA
	.dc.b	$9F,$32,$31,$1F,$04,$FF,$7A,$B8,$E2,$FD,$6F,$D2,$3E,$4B,$E9,$FC
	.dc.b	$33,$E8,$BE,$AF,$D2,$3E,$92,$9A,$89,$97,$D0,$9D,$36,$72,$F8,$45
	.dc.b	$75,$74,$43,$4E,$5E,$1F,$48,$F8,$3F,$C3,$FC,$57,$F1,$09,$FB,$ED
	.dc.b	$77,$1F,$8C,$9E,$11,$AF,$2F,$89,$99,$91,$92,$1C,$8F,$C5,$FC,$4F
	.dc.b	$E1,$FB,$BF,$B5,$FD,$27,$D1,$9F,$63,$78,$3F,$8B,$D6,$E4,$1C,$29
	.dc.b	$E1,$5D,$39,$3F,$47,$5A,$37,$3A,$AE,$6D,$7A,$7B,$FF,$AA,$FA,$47
	.dc.b	$CE,$7E,$B2,$9F,$1B,$4D,$44,$BF,$D3,$CE,$9B,$39,$7E,$C5,$74,$72
	.dc.b	$43,$35,$FB,$AF,$A4,$FB,$1A,$7F,$38,$C3,$F8,$83,$7C,$7D,$72,$CC
	.dc.b	$3C,$31,$E7,$30,$FE,$1E,$F9,$ED,$F4,$27,$53,$FB,$E9,$CB,$F3,$97
	.dc.b	$9D,$F0,$7D,$B9,$3B,$8F,$31,$87,$EC,$C6,$1F,$E1,$3F,$48,$E5,$FE
	.dc.b	$2A,$FE,$62,$33,$CD,$89,$91,$C1,$AF,$E7,$FE,$8D,$F4,$9F,$C9,$61
	.dc.b	$F4,$8F,$8A,$FF,$1A,$D6,$B8,$3F,$0F,$39,$AF,$C4,$E5,$E3,$73,$7F
	.dc.b	$E3,$C8,$C4,$74,$8E,$78,$77,$AE,$52,$C3,$B3,$B3,$AE,$3F,$E4,$69
	.dc.b	$DE,$1D,$1E,$35,$C1,$B7,$D7,$0F,$15,$F1,$6E,$1E,$63,$F4,$37,$C6
	.dc.b	$3C,$B8,$E6,$58,$F7,$9E,$31,$E5,$3D,$18,$77,$79,$B5,$E3,$87,$85
	.dc.b	$F5,$9D,$70,$F0,$3C,$30,$F0,$61,$DE,$36,$77,$9D,$BE,$09,$DD,$75
	.dc.b	$8E,$E7,$5B,$F1,$57,$CC,$8F,$F0,$F3,$5F,$79,$FE,$1E,$70,$FC,$4F
	.dc.b	$9E,$7B,$9F,$E0,$39,$46,$27,$83,$5E,$AD,$7F,$C5,$3F,$49,$FE,$01
	.dc.b	$4D,$23,$FE,$05,$7E,$87,$C4,$7F,$5F,$E3,$FA,$53,$FC,$5A,$73,$E7
	.dc.b	$FF,$98,$F7,$FC,$4C,$E4,$FA,$97,$C5,$FA,$C6,$63,$F1,$F1,$97,$8C
	.dc.b	$79,$0E,$33,$D7,$7F,$4C,$63,$0F,$1B,$BA,$E7,$5B,$8F,$F2,$56,$1E
	.dc.b	$6A,$DB,$63,$C6,$71,$BF,$F5,$0E,$1E,$37,$BE,$1D,$2F,$8D,$78,$4E
	.dc.b	$37,$CE,$7D,$2D,$AE,$78,$7F,$98,$B0,$F8,$5B,$7F,$95,$B0,$F8,$55
	.dc.b	$AC,$7F,$C2,$75,$A7,$13,$FD,$8C,$D9,$FD,$1A,$34,$DE,$7E,$C7,$FA
	.dc.b	$56,$FF,$D2,$DE,$AB,$E9,$1F,$49,$AC,$7B,$DD,$C3,$FD,$6F,$5A,$9C
	.dc.b	$9E,$4F,$8A,$F5,$1D,$5F,$EA,$F8,$3E,$93,$FC,$ED,$97,$F9,$CB,$35
	.dc.b	$F0,$BE,$47,$F8,$D6,$F8,$78,$79,$CB,$FC,$01,$7C,$2B,$71,$FE,$CA
	.dc.b	$C7,$AC,$7D,$6C,$3F,$D1,$BE,$E7,$FB,$46,$BF,$CC,$FD,$5E,$57,$FD
	.dc.b	$B9,$F4,$5F,$E6,$4B,$FF,$1F,$D6,$C9,$5B,$84,$EE,$BF,$C2,$55,$A9
	.dc.b	$E3,$5B,$AD,$69,$FE,$90,$AF,$17,$D1,$BE,$93,$FD,$42,$DC,$3F,$CA
	.dc.b	$87,$6F,$2F,$33,$23,$FC,$A7,$3B,$CB,$FC,$65,$1B,$33,$3B,$8F,$F1
	.dc.b	$34,$EF,$0F,$F6,$A5,$6E,$73,$5F,$EB,$28,$FE,$0E,$7F,$98,$0B,$1D
	.dc.b	$B1,$7C,$96,$FA,$25,$6D,$F8,$11,$8A,$FF,$1B,$4F,$2C,$3F,$D6,$2D
	.dc.b	$3F,$C3,$67,$CC,$9E,$11,$FE,$B6,$F8,$EF,$BD,$5F,$E3,$F6,$C7,$AB
	.dc.b	$07,$C9,$7F,$B2,$A7,$38,$7F,$89,$AD,$18,$7F,$C5,$2D,$8F,$B7,$3C
	.dc.b	$77,$91,$FE,$80,$9D,$5F,$68,$C4,$FF,$B6,$3E,$85,$F3,$D4,$FC,$67
	.dc.b	$D2,$3E,$91,$F4,$99,$C3,$FD,$17,$7C,$1F,$F4,$6F,$D1,$7F,$96,$9C
	.dc.b	$E1,$FE,$E4,$8C,$6B,$FE,$5F,$6E,$DF,$ED,$A6,$D7,$BE,$FA,$56,$FC
	.dc.b	$6F,$F7,$6F,$D2,$3E,$2F,$FD,$DF,$F4,$9A,$CB,$FD,$C7,$39,$C7,$FD
	.dc.b	$3F,$39,$7E,$53,$70,$FF,$28,$E5,$FE,$69,$9D,$E7,$E7,$F5,$24,$7F
	.dc.b	$AA,$3B,$9F,$49,$8C,$3F,$CF,$11,$89,$FC,$FF,$CF,$7A,$A5,$08,$FF
	.dc.b	$14,$3A,$DB,$F6,$51,$89,$FD,$0B,$C6,$78,$79,$7A,$AD,$B6,$3B,$B6
	.dc.b	$B8,$BF,$C0,$D7,$D1,$FF,$97,$A7,$FD,$82,$D8,$FF,$50,$60,$F9,$9F
	.dc.b	$F7,$95,$6B,$D5,$E5,$7B,$BF,$C9,$1D,$23,$F9,$E3,$FE,$7A,$8D,$3F
	.dc.b	$97,$6B,$F0,$B2,$7D,$1F,$C0,$3E,$0F,$BF,$70,$F9,$31,$6B,$E1,$CD
	.dc.b	$E3,$FD,$FD,$BE,$8B,$E4,$FD,$1F,$CA,$6E,$3F,$49,$B8,$7C,$C6,$DF
	.dc.b	$26,$CF,$EE,$2F,$A3,$F6,$AF,$E5,$7D,$17,$FA,$56,$3F,$DD,$2D,$C3
	.dc.b	$E4,$EA,$F9,$8D,$FB,$B2,$FE,$3D,$7F,$A7,$AF,$83,$F2,$9A,$FE,$F5
	.dc.b	$AF,$C7,$FA,$3F,$C2,$9F,$86,$E5,$B5,$F8,$E8,$B3,$F8,$89,$BA,$FE
	.dc.b	$42,$3F,$83,$9F,$C5,$0F,$F6,$F7,$83,$FD,$6A,$FF,$D6,$33,$B8,$FE
	.dc.b	$1D,$18,$3F,$86,$CF,$F4,$06,$BF,$D3,$E3,$93,$D0,$ED,$3C,$DE,$07
	.dc.b	$88,$E0,$62,$3B,$BC,$78,$FE,$7F,$93,$2F,$FB,$46,$AF,$88,$C5,$FF
	.dc.b	$B9,$F4,$FE,$C4,$7F,$A7,$2B,$73,$AC,$BC,$5E,$8F,$80,$F0,$F8,$D6
	.dc.b	$BD,$D8,$7C,$69,$C3,$EB,$9B,$5E,$BF,$1B,$D3,$ED,$FA,$9B,$87,$2D
	.dc.b	$BF,$DB,$71,$8C,$3E,$23,$6F,$95,$6E,$57,$CE,$30,$FE,$E7,$E8,$5C
	.dc.b	$F4,$FE,$FD,$F4,$CF,$A7,$FE,$67,$7E,$AC,$17,$C2,$30,$D8,$63,$F8
	.dc.b	$57,$78,$74,$BE,$71,$AB,$FE,$A1,$7C,$6F,$EE,$7D,$1F,$DA,$F2,$BE
	.dc.b	$DB,$6B,$D8,$DB,$FC,$23,$6B,$F0,$9F,$47,$F8,$23,$F0,$7F,$0B,$6F
	.dc.b	$E2,$45,$BF,$C6,$4D,$AE,$35,$A3,$E5,$46,$5F,$CB,$D5,$AF,$D0,$CD
	.dc.b	$DD,$09,$C2,$9B,$18,$6D,$77,$9F,$1B,$6F,$C7,$1F,$A1,$AF,$5C,$E4
	.dc.b	$F1,$96,$3C,$6D,$BF,$1C,$F7,$8C,$4F,$95,$B7,$E2,$8E,$D1,$F0,$0D
	.dc.b	$CF,$18,$C9,$2F,$CA,$7E,$CE,$B9,$38,$AF,$DE,$C6,$1F,$3C,$FB,$8E
	.dc.b	$4E,$27,$84,$7A,$8C,$9F,$C3,$AB,$C5,$1A,$BF,$6D,$7A,$8F,$39,$6B
	.dc.b	$8C,$69,$A7,$D3,$8D,$B8,$BF,$33,$DE,$3E,$13,$C2,$FC,$31,$E7,$24
	.dc.b	$6C,$F6,$5F,$95,$F9,$C7,$58,$EF,$19,$AE,$F7,$E5,$8F,$95,$3A,$3A
	.dc.b	$C7,$E3,$27,$75,$D2,$B6,$F9,$6B,$E0,$47,$86,$FC,$F3,$E5,$8C,$4F
	.dc.b	$B1,$B1,$E5,$39,$15,$C9,$E7,$8F,$24,$73,$7E,$E9,$8B,$F5,$38,$8E
	.dc.b	$73,$C5,$B1,$EC,$7E,$91,$83,$A1,$F8,$18,$F0,$C6,$EF,$E0,$DF,$E8
	.dc.b	$AB,$85,$F8,$1E,$15,$D6,$7E,$0C,$E4,$93,$F2,$67,$CD,$7D,$A3,$C3
	.dc.b	$3F,$3E,$7D,$91,$B3,$A9,$A3,$11,$D2,$FC,$B3,$F3,$4F,$C2,$CB,$E8
	.dc.b	$1F,$03,$FE,$29,$6C,$FF,$52,$F9,$CC,$4C,$8F,$E6,$13,$AA,$FF,$01
	.dc.b	$D6,$5F,$ED,$FF,$01,$FE,$42,$AF,$E8,$CB,$E8,$FF,$B5,$39,$9D,$C6
	.dc.b	$27,$2D,$78,$BC,$3E,$8E,$47,$4F,$9D,$B5,$E0,$9B,$AF,$F7,$DC,$15
	.dc.b	$8B,$FF,$44,$C2,$B1,$7F,$E8,$E8,$DC,$FA,$27,$E2,$7C,$D4,$8F,$ED
	.dc.b	$3E,$E3,$E3,$1D,$B9,$C7,$FA,$51,$FC,$C3,$66,$19,$D3,$FC,$2D,$87
	.dc.b	$1B,$C6,$3E,$4F,$13,$E2,$9F,$1B,$E8,$AC,$7C,$A8,$B9,$7C,$E9,$B8
	.dc.b	$7B,$FE,$8B,$E6,$FC,$07,$CE,$6E,$1F,$49,$B7,$F5,$9B,$5C,$67,$EB
	.dc.b	$D7,$16,$9F,$DB,$FE,$8B,$FD,$EB,$7F,$D3,$1B,$9F,$C3,$F4,$3F,$CE
	.dc.b	$51,$A9,$2B,$FF,$34,$C6,$1B,$19,$BF,$F2,$4C,$E5,$FF,$25,$F3,$B8
	.dc.b	$9F,$2D,$F8,$AF,$5F,$F1,$6D,$AF,$C7,$B6,$FF,$26,$DD,$3F,$09,$CF
	.dc.b	$F9,$2F,$1F,$F2,$6D,$BF,$CB,$B6,$FF,$36,$DA,$E3,$3F,$9C,$AE,$2D
	.dc.b	$3F,$C2,$9F,$4F,$FE,$BA,$BF,$F4,$4B,$5F,$F6,$E3,$5F,$EE,$AD,$5F
	.dc.b	$44,$FA,$27,$C2,$7C,$48,$C3,$7A,$DE,$66,$6B,$77,$FE,$72,$F1,$FF
	.dc.b	$22,$FF,$89,$D7,$D0,$FF,$BB,$75,$FF,$12,$46,$27,$5D,$79,$D7,$FD
	.dc.b	$E5,$3C,$A7,$F3,$A7,$F8,$EA,$F4,$FF,$83,$A3,$93,$A7,$85,$F6,$F6
	.dc.b	$37,$3F,$E4,$2F,$B3,$B4,$6C,$C9,$F4,$46,$8C,$3D,$63,$0E,$8F,$73
	.dc.b	$DB,$E8,$9F,$19,$E0,$E4,$7C,$B5,$35,$1E,$4F,$A0,$7D,$07,$9B,$E8
	.dc.b	$9F,$45,$4D,$47,$AB,$E8,$7F,$CE,$DA,$7F,$94,$D7,$D1,$64,$D7,$AF
	.dc.b	$E1,$FA,$29,$5F,$E5,$9D,$51,$F1,$3E,$8B,$FC,$69,$3F,$E5,$D6,$BF
	.dc.b	$41,$BC,$7F,$E5,$B6,$F2,$FF,$8F,$A5,$F4,$5F,$E6,$A7,$FC,$CC,$D7
	.dc.b	$EC,$B6,$BF,$CC,$8D,$9E,$5F,$44,$F9,$3F,$AF,$E3,$BE,$7B,$EC,$FD
	.dc.b	$13,$E8,$A9,$A8,$99,$FF,$9D,$67,$4D,$9C,$BF,$8A,$AE,$8E,$48,$66
	.dc.b	$BF,$0F,$E8,$BE,$07,$F1,$3F,$15,$FC,$2E,$BF,$D0,$4D,$7F,$2B,$3B
	.dc.b	$7F,$D0,$CD,$7F,$33,$1B,$DB,$FD,$08,$DD,$7E,$55,$6E,$7F,$D1,$0D
	.dc.b	$8F,$9D,$3B,$AF,$F3,$6B,$5F,$E8,$97,$F9,$EF,$84,$FC,$7C,$7F,$A5
	.dc.b	$FE,$87,$EB,$BF,$E7,$46,$A8,$DB,$FA,$E6,$E3,$FE,$99,$9D,$9C,$23
	.dc.b	$38,$7E,$05,$B5,$D7,$D9,$FD,$7D,$6F,$B7,$F6,$3F,$44,$F9,$BC,$FC
	.dc.b	$6C,$C3,$FD,$59,$39,$6C,$E5,$FE,$03,$6A,$E7,$32,$7D,$D1,$BC,$FF
	.dc.b	$D6,$17,$CE,$7F,$84,$FD,$17,$E0,$1E,$7E,$2F,$98,$FE,$0B,$19,$AF
	.dc.b	$F5,$C3,$75,$FC,$F6,$5F,$EC,$AB,$F9,$97,$C3,$86,$46,$08,$78,$1E
	.dc.b	$BE,$77,$C4,$F7,$64,$FF,$B2,$F1,$4E,$A3,$EC,$4E,$9F,$4C,$7F,$1A
	.dc.b	$FA,$27,$1D,$E5,$95,$FE,$D5,$9C,$B6,$78,$11,$FF,$6D,$7D,$0B,$E8
	.dc.b	$BF,$DA,$F0,$FA,$27,$C5,$FE,$F6,$B5,$C3,$FE,$BC,$9C,$D7,$FA,$FB
	.dc.b	$1E,$EF,$1B,$FE,$09,$18,$9D,$CA,$F9,$BB,$3F,$D8,$EE,$23,$73,$C6
	.dc.b	$B9,$E3,$FC,$3E,$79,$D6,$E7,$CD,$8F,$AA,$78,$D6,$E7,$8E,$1E,$97
	.dc.b	$78,$D8,$F0,$DF,$19,$DC,$79,$46,$3E,$37,$75,$C5,$F6,$57,$17,$E2
	.dc.b	$5F,$85,$A6,$F1,$F4,$3C,$8E,$D7,$D5,$EF,$3C,$5E,$B7,$C6,$FB,$46
	.dc.b	$8F,$15,$71,$BE,$B1,$C0,$FE,$6C,$77,$2B,$F4,$9B,$3F,$82,$73,$F3
	.dc.b	$9F,$E2,$17,$89,$F6,$5C,$3A,$F6,$3F,$C5,$34,$9C,$CF,$0F,$A2,$FF
	.dc.b	$16,$A0,$E1,$A6,$9C,$1F,$37,$DD,$FE,$D7,$CB,$F3,$E7,$94,$E7,$CD
	.dc.b	$4E,$67,$9E,$FF,$90,$BE,$31,$F6,$CE,$27,$DD,$73,$1F,$96,$8C,$BC
	.dc.b	$6B,$AC,$71,$3A,$EF,$FD,$86,$B8,$BE,$F8,$E9,$3C,$2B,$A4,$70,$3E
	.dc.b	$15,$70,$7E,$AE,$1E,$BC,$3D,$4D,$8F,$03,$6B,$9B,$70,$F0,$47,$9E
	.dc.b	$B8,$D7,$89,$E5,$B7,$AA,$18,$7C,$48,$C4,$7C,$0B,$F6,$63,$F0,$27
	.dc.b	$85,$6E,$F2,$7D,$B8,$E3,$7E,$D3,$AD,$71,$9F,$8D,$D3,$67,$E0,$B7
	.dc.b	$D4,$7D,$0B,$E8,$B5,$5E,$6A,$B9,$7F,$49,$C3,$C0,$74,$2C,$FF,$95
	.dc.b	$2F,$94,$62,$39,$36,$5F,$13,$D0,$F1,$7D,$17,$FA,$26,$33,$3B,$9D
	.dc.b	$3D,$B2,$56,$E7,$37,$FE,$56,$BE,$2F,$FA,$D2,$B7,$F0,$3A,$D7,$E9
	.dc.b	$7B,$7F,$D0,$32,$F3,$36,$FF,$D7,$70,$FA,$0F,$F3,$7B,$AE,$AF,$8B
	.dc.b	$5B,$BE,$71,$F3,$67,$86,$3A,$89,$1C,$4F,$8E,$68,$C6,$1F,$E6,$EF
	.dc.b	$A2,$7D,$17,$D2,$6C,$F0,$CB,$FC,$6C,$FC,$D8,$CD,$71,$AE,$07,$8E
	.dc.b	$B7,$19,$AE,$13,$A7,$A7,$97,$67,$FC,$FF,$1F,$C2,$CF,$B4,$57,$E2
	.dc.b	$D3,$E4,$FF,$0F,$39,$CF,$F3,$CD,$3F,$CA,$C5,$AF,$E0,$F5,$C6,$3A
	.dc.b	$F4,$72,$23,$EF,$BB,$5F,$E9,$2F,$84,$FB,$B8,$7E,$AD,$B5,$FC,$74
	.dc.b	$FB,$1F,$27,$F9,$A8,$CD,$7F,$21,$6A,$73,$3C,$0F,$CC,$3A,$27,$17
	.dc.b	$CB,$AE,$8F,$E9,$BE,$7D,$F3,$94,$FD,$77,$D1,$3E,$89,$F4,$59,$CB
	.dc.b	$F4,$97,$C1,$FD,$37,$D0,$3C,$6D,$6B,$FE,$B0,$8C,$4E,$A7,$F6,$B1
	.dc.b	$89,$FF,$5A,$D7,$37,$84,$FE,$EB,$E8,$9F,$19,$FB,$CF,$A2,$D7,$0F
	.dc.b	$F1,$39,$CE,$5F,$B7,$9C,$BF,$E8,$D6,$E1,$FD,$27,$2F,$F3,$A4,$EF
	.dc.b	$6F,$1F,$A9,$23,$F8,$07,$53,$E1,$71,$87,$FB,$2A,$31,$3F,$A2,$F9
	.dc.b	$FF,$56,$9F,$CF,$A1,$1F,$E2,$E7,$5B,$7E,$B6,$31,$7F,$B6,$78,$CF
	.dc.b	$E0,$5C,$D7,$37,$33,$FD,$59,$CC,$EE,$BA,$79,$9D,$5B,$1F,$C6,$7E
	.dc.b	$8B,$FC,$C0,$FF,$72,$B4,$61,$FE,$EA,$D7,$FB,$FB,$5F,$F5,$B3,$BC
	.dc.b	$3F,$AD,$E2,$F9,$EF,$F4,$15,$F7,$F5,$7F,$9A,$46,$30,$F7,$7D,$17
	.dc.b	$C1,$3E,$17,$CF,$5B,$F9,$51,$67,$E2,$4D,$C7,$E1,$7D,$17,$C9,$FA
	.dc.b	$2F,$94,$DC,$3E,$93,$6F,$E6,$36,$B8,$CF,$D7,$AE,$2D,$3F,$9F,$7C
	.dc.b	$CF,$9E,$FF,$C3,$ED,$7F,$DA,$CD,$7F,$C2,$4D,$4F,$BE,$BE,$2E,$4F
	.dc.b	$98,$FE,$A7,$39,$BF,$F0,$D6,$BF,$26,$7F,$09,$F4,$52,$7F,$19,$C9
	.dc.b	$72,$FC,$34,$DC,$BF,$09,$36,$BF,$82,$57,$F0,$91,$3F,$E1,$0F,$03
	.dc.b	$BC,$FF,$5E,$7F,$8D,$CE,$A3,$8E,$3F,$DB,$67,$35,$C6,$BF,$A4,$9B
	.dc.b	$3F,$88,$1D,$DD,$4F,$18,$C5,$74,$3C,$0E,$0E,$31,$F1,$3C,$7F,$F0
	.dc.b	$0C,$7C,$A6,$E7,$FD,$6E,$3E,$9C,$60,$FC,$6F,$D3,$F1,$BF,$ED,$9A
	.dc.b	$A3,$C5,$E8,$C9,$F0,$F8,$97,$0F,$56,$9F,$E4,$BA,$C6,$9E,$FA,$B7
	.dc.b	$F3,$3C,$4F,$67,$B7,$ED,$6C,$7B,$1B,$8F,$40,$E2,$FE,$53,$67,$A1
	.dc.b	$F1,$2B,$FC,$2F,$F4,$2E,$5F,$F0,$F7,$9C,$FA,$77,$C0,$6A,$73,$18
	.dc.b	$C3,$FC,$2D,$7C,$5F,$F0,$DD,$F1,$BE,$11,$8A,$E1,$9F,$9F,$E8,$BE
	.dc.b	$C7,$95,$F6,$5B,$5F,$61,$B3,$B8,$FB,$73,$F7,$FE,$8A,$47,$DE,$F8
	.dc.b	$3B,$5F,$0E,$6E,$FF,$8B,$9B,$5F,$98,$9B,$5F,$8C,$9B,$C5,$FE,$7F
	.dc.b	$9B,$29,$B1,$86,$D7,$E3,$A3,$F2,$51,$89,$F1,$4F,$94,$D1,$1C,$9F
	.dc.b	$66,$38,$5F,$E8,$6B,$9D,$6E,$30,$D3,$F6,$33,$C1,$F6,$1D,$EB,$AC
	.dc.b	$F0,$6C,$70,$39,$3F,$AC,$AF,$64,$62,$FE,$13,$67,$E1,$36,$77,$1D
	.dc.b	$0E,$51,$BB,$F2,$46,$0F,$74,$71,$34,$57,$95,$7F,$AB,$E7,$8B,$E2
	.dc.b	$8E,$26,$85,$78,$AF,$DC,$4B,$F8,$73,$C2,$30,$F8,$DB,$5D,$E7,$C5
	.dc.b	$18,$9F,$79,$DE,$BB,$4F,$49,$D4,$7D,$69,$D4,$FB,$1A,$79,$EF,$D6
	.dc.b	$FA,$23,$EB,$57,$27,$83,$63,$84,$79,$0F,$01,$F6,$5E,$B1,$E7,$9E
	.dc.b	$B5,$E3,$8E,$46,$6B,$95,$7C,$17,$94,$69,$EB,$1F,$A3,$77,$7F,$46
	.dc.b	$7F,$55,$18,$9F,$51,$F3,$5E,$33,$B6,$C7,$CA,$9E,$CF,$84,$EC,$7A
	.dc.b	$A3,$77,$EA,$7E,$29,$CA,$FE,$8C,$F1,$BE,$CD,$3B,$4E,$CF,$96,$57
	.dc.b	$E5,$9D,$CE,$15,$DD,$F5,$1C,$EB,$BD,$76,$87,$D0,$3E,$03,$FC,$1F
	.dc.b	$CF,$FE,$C0,$8C,$63,$FD,$12,$31,$3A,$8F,$F5,$E4,$EA,$BF,$D5,$DC
	.dc.b	$FF,$EA,$39,$FF,$37,$7B,$C7,$FB,$22,$FA,$2B,$18,$6B,$FD,$B6,$34
	.dc.b	$D3,$46,$7E,$8B,$99,$D3,$E7,$2C,$64,$98,$7F,$56,$A4,$EA,$31,$57
	.dc.b	$7F,$EF,$D5,$AB,$E3,$1F,$EF,$7C,$9F,$37,$FD,$B9,$B3,$FE,$43,$6E
	.dc.b	$02,$15,$CB,$83,$E8,$CF,$FB,$CE,$9F,$4E,$FA,$0F,$A5,$39,$8C,$61
	.dc.b	$F1,$7E,$8F,$E5,$9F,$33,$C4,$B1,$92,$63,$F0,$63,$2F,$C3,$FA,$3F
	.dc.b	$A7,$EF,$7D,$46,$D7,$D1,$6C,$EE,$3E,$A4,$ED,$A6,$CB,$1F,$E1,$D1
	.dc.b	$F4,$69,$FF,$0A,$B6,$7F,$CD,$4D,$9F,$88,$70,$9D,$9F,$89,$7F,$DF
	.dc.b	$5F,$4E,$FA,$8C,$6D,$99,$E1,$87,$D1,$8F,$F3,$6D,$6E,$3F,$19,$F3
	.dc.b	$1F,$92,$F6,$3F,$26,$D8,$C9,$F9,$18,$CB,$4F,$CA,$70,$7E,$6B,$A3
	.dc.b	$F3,$6D,$8C,$9F,$99,$8C,$B4,$CE,$3F,$93,$8C,$B6,$3D,$07,$E2,$3E
	.dc.b	$91,$F4,$8F,$84,$FF,$48,$46,$1B,$D0,$F3,$33,$5B,$C3,$E5,$FD,$43
	.dc.b	$C8,$FF,$4C,$42,$30,$FF,$A6,$DA,$F0,$8D,$9F,$A0,$7C,$A7,$E7,$57
	.dc.b	$D2,$FE,$7B,$87,$E9,$46,$27,$67,$87,$A7,$5A,$FE,$92,$7A,$4F,$E1
	.dc.b	$DE,$0D,$BD,$BD,$EA,$BE,$BF,$66,$D3,$E9,$69,$88,$93,$E2,$9D,$36
	.dc.b	$72,$F8,$67,$8B,$93,$83,$E0,$FA,$57,$C6,$F8,$B9,$7F,$C9,$75,$EE
	.dc.b	$6B,$FA,$19,$DB,$F0,$BE,$95,$F4,$7E,$AF,$A5,$7D,$2D,$3D,$B3,$AA
	.dc.b	$F9,$3F,$4A,$FA,$4F,$82,$BE,$96,$98,$8F,$27,$D2,$CA,$F2,$3F,$23
	.dc.b	$E9,$7F,$C8,$11,$F5,$1A,$FE,$DD,$BC,$BF,$AB,$6F,$2F,$F1,$AF,$A5
	.dc.b	$FF,$36,$57,$DC,$6B,$FB,$F6,$D7,$EC,$DB,$3F,$C8,$BE,$95,$F2,$5F
	.dc.b	$6B,$E1,$9F,$3F,$F7,$3E,$95,$F4,$B4,$D4,$4C,$FF,$17,$3A,$6C,$E5
	.dc.b	$FE,$21,$5D,$1C,$90,$CD,$7E,$2B,$E9,$5F,$09,$F8,$CF,$8B,$FE,$2D
	.dc.b	$5F,$98,$6B,$FC,$6E,$76,$FE,$71,$AF,$F1,$F8,$DE,$DF,$C4,$DB,$AF
	.dc.b	$F4,$2A,$DC,$FF,$17,$6C,$7F,$46,$9D,$D7,$F4,$96,$A9,$DB,$FD,$37
	.dc.b	$E9,$7F,$1F,$7F,$C8,$7E,$93,$EB,$CF,$F5,$16,$A9,$DE,$1F,$D5,$1B
	.dc.b	$1B,$F6,$FF,$5B,$5C,$27,$33,$C7,$DF,$CD,$48,$FD,$EC,$EB,$E9,$5F
	.dc.b	$3D,$CB,$E3,$A5,$7F,$02,$9C,$B6,$72,$FF,$47,$C5,$5F,$BE,$AD,$CF
	.dc.b	$F4,$B6,$E7,$EF,$9E,$2F,$F4,$18,$CE,$7F,$DC,$3E,$97,$F0,$57,$D7
	.dc.b	$C5,$F6,$9F,$C2,$A3,$37,$FD,$25,$B5,$FD,$E3,$1F,$E4,$73,$F3,$6B
	.dc.b	$87,$0C,$C3,$F8,$ED,$6A,$F8,$56,$67,$FB,$16,$0F,$3B,$E6,$FB,$D2
	.dc.b	$7F,$92,$CE,$5B,$39,$7E,$24,$F9,$E3,$F9,$0F,$D2,$B8,$EF,$2C,$BF
	.dc.b	$E5,$F3,$96,$CF,$02,$3F,$CD,$3E,$91,$F4,$AD,$FF,$C3,$1F,$48,$F8
	.dc.b	$DF,$F0,$CD,$6A,$73,$1F,$E1,$CC,$7B,$60,$BE,$0F,$F8,$86,$33,$18
	.dc.b	$85,$F3,$78,$57,$6C,$3C,$51,$83,$96,$3F,$E2,$AB,$E5,$87,$4B,$E1
	.dc.b	$8F,$92,$FF,$C6,$18,$F5,$BD,$E1,$C6,$FD,$0D,$AD,$E3,$E0,$AF,$35
	.dc.b	$F1,$BD,$D7,$51,$8F,$85,$E3,$8E,$6B,$84,$7A,$0E,$13,$B7,$E0,$E1
	.dc.b	$DE,$FD,$31,$EF,$3A,$E3,$E5,$69,$D6,$78,$3E,$03,$85,$75,$3B,$5F
	.dc.b	$8A,$3B,$E1,$FE,$2C,$69,$FE,$2E,$8F,$A4,$FF,$94,$DA,$FE,$03,$E7
	.dc.b	$9F,$03,$C6,$51,$89,$E4,$D7,$93,$5F,$F1,$DF,$D2,$FF,$84,$13,$48
	.dc.b	$FF,$85,$1F,$9D,$F1,$7F,$DB,$B8,$FE,$84,$FF,$9A,$E7,$3E,$77,$F9
	.dc.b	$0F,$7F,$C6,$4E,$4F,$95,$7C,$5F,$98,$66,$3F,$25,$19,$78,$C7,$90
	.dc.b	$E3,$3D,$77,$F4,$C6,$30,$F1,$BB,$AE,$75,$B8,$FF,$3E,$61,$E6,$AD
	.dc.b	$B6,$3C,$67,$1B,$FF,$42,$E1,$E3,$7B,$E1,$D2,$F8,$D7,$84,$E3,$7C
	.dc.b	$E7,$D2,$DA,$E7,$87,$F9,$92,$E9,$FE,$32,$BF,$85,$38,$7F,$CE,$D8
	.dc.b	$7C,$7D,$3F,$CD,$4D,$AF,$1D,$6E,$73,$D3,$D6,$FF,$D3,$D7,$FD,$02
	.dc.b	$FF,$C8,$BE,$AB,$E9,$1F,$4B,$BB,$5E,$DF,$0E,$19,$8C,$63,$FE,$7F
	.dc.b	$C3,$95,$E6,$B5,$F1,$73,$2E,$FF,$4B,$FE,$89,$C3,$D3,$39,$CB,$DA
	.dc.b	$F9,$F1,$56,$EF,$E8,$CE,$EF,$8C,$74,$F8,$1C,$3F,$EB,$AF,$73,$FD
	.dc.b	$B9,$8F,$FA,$F5,$B0,$FA,$3F,$F3,$93,$FE,$60,$8D,$CF,$F9,$E7,$AB
	.dc.b	$E1,$47,$0A,$DC,$FF,$A9,$E3,$64,$9F,$9D,$5D,$73,$E3,$1F,$2F,$E9
	.dc.b	$5F,$4B,$4D,$E3,$FE,$2E,$BF,$F3,$8C,$E9,$B3,$B7,$FC,$F6,$7A,$7C
	.dc.b	$BC,$65,$FE,$B9,$AE,$0D,$AE,$B3,$AD,$3F,$D6,$17,$FC,$10,$FF,$3F
	.dc.b	$96,$BF,$A6,$3F,$EE,$33,$E3,$7C,$AF,$E0,$DD,$4F,$2E,$8C,$1E,$A9
	.dc.b	$B3,$D2,$2D,$74,$11,$D0,$D1,$FE,$90,$F8,$F7,$07,$E4,$5B,$1F,$03
	.dc.b	$47,$CA,$E6,$A4,$F1,$7E,$5C,$62,$FF,$DB,$97,$C9,$B3,$AA,$E0,$7F
	.dc.b	$A8,$3E,$89,$F3,$FF,$A1,$6C,$7F,$AA,$BE,$95,$F4,$AF,$A5,$D5,$CA
	.dc.b	$FF,$54,$FD,$1B,$C6,$C6,$7F,$EE,$D8,$C4,$EA,$7F,$5D,$18,$9F,$F7
	.dc.b	$96,$5E,$07,$84,$FE,$CF,$E9,$5F,$19,$FB,$5F,$A5,$D7,$27,$FB,$AE
	.dc.b	$73,$97,$FA,$EA,$72,$B0,$91,$87,$19,$7F,$9B,$27,$86,$7E,$4F,$56
	.dc.b	$47,$FB,$13,$A9,$F0,$B8,$C3,$EE,$1F,$ED,$B7,$F8,$47,$D2,$7F,$91
	.dc.b	$2B,$FD,$96,$D9,$FC,$15,$3E,$67,$FD,$CF,$5A,$F5,$38,$91,$FE,$2B
	.dc.b	$1A,$AE,$F7,$CF,$6F,$C5,$C7,$B6,$3B,$3E,$F3,$17,$A7,$02,$36,$ED
	.dc.b	$AF,$E5,$BE,$95,$F4,$BF,$BC,$6E,$3F,$B2,$5F,$41,$99,$93,$97,$84
	.dc.b	$62,$BF,$45,$2A,$7D,$2F,$5F,$A5,$EC,$D9,$EC,$DA,$E3,$5C,$27,$C5
	.dc.b	$5F,$CD,$FE,$93,$FD,$11,$1F,$EF,$A6,$D7,$D0,$D5,$F3,$7B,$F7,$25
	.dc.b	$ED,$FE,$61,$3A,$8C,$12,$BE,$48,$BE,$AF,$CC,$6B,$F4,$5A,$FC,$BF
	.dc.b	$A5,$F9,$47,$CB,$E3,$B5,$F2,$E2,$CF,$BA,$6F,$0F,$DE,$EF,$FB,$97
	.dc.b	$F9,$B8,$C5,$7F,$34,$6B,$FD,$7E,$7F,$41,$C7,$63,$F8,$E1,$FC,$4E
	.dc.b	$33,$5C,$CF,$D3,$3D,$A3,$47,$58,$F6,$78,$FF,$3C,$F8,$FF,$0F,$63
	.dc.b	$E8,$45,$9F,$A1,$37,$0F,$C9,$67,$FD,$38,$FE,$AB,$8F,$E8,$E7,$58
	.dc.b	$78,$AF,$FC,$F8,$F9,$56,$3F,$BB,$F4,$3D,$9E,$37,$B5,$95,$B9,$D3
	.dc.b	$69,$5A,$7D,$BE,$27,$C5,$F6,$FE,$33,$6B,$83,$67,$71,$CE,$7F,$C0
	.dc.b	$DF,$49,$FA,$28,$F3,$4A,$30,$FF,$41,$FA,$87,$D4,$7F,$29,$AD,$D6
	.dc.b	$B7,$FD,$B3,$63,$F7,$4D,$B5,$F2,$F1,$7C,$3D,$52,$F9,$BF,$E1,$F9
	.dc.b	$D4,$60,$F8,$5F,$4B,$F8,$3F,$27,$F0,$8D,$8C,$9F,$1E,$32,$D3,$E5
	.dc.b	$7D,$2F,$E2,$CF,$C6,$7C,$25,$8C,$93,$2F,$C4,$C6,$5F,$E4,$FF,$46
	.dc.b	$F8,$1F,$F1,$73,$6B,$F8,$FF,$CE,$E2,$64,$7C,$C9,$D5,$7E,$EE,$B2
	.dc.b	$F4,$AB,$BF,$F8,$CF,$B1,$58,$C3,$F9,$26,$1C,$AB,$F9,$23,$63,$F4
	.dc.b	$F5,$C1,$F0,$5F,$1C,$3F,$A5,$56,$D7,$47,$F3,$BC,$75,$09,$D3,$6B
	.dc.b	$F8,$6C,$EA,$72,$73,$AE,$57,$E0,$25,$FB,$5F,$8A,$63,$1F,$4E,$1E
	.dc.b	$5A,$E3,$5B,$24,$EA,$B8,$C6,$1F,$87,$5C,$71,$F8,$13,$93,$DF,$19
	.dc.b	$C7,$D0,$DA,$F1,$1C,$8B,$3F,$12,$B9,$DF,$86,$3A,$1F,$0E,$7C,$71
	.dc.b	$92,$3C,$E3,$EF,$3D,$1E,$07,$97,$0F,$7C,$F0,$8F,$3C,$F8,$A7,$2F
	.dc.b	$39,$D4,$F2,$7F,$0D,$87,$B2,$78,$5F,$A1,$E0,$4A,$EA,$79,$0F,$5E
	.dc.b	$3F,$6B,$1F,$BB,$3E,$78,$C3,$EE,$39,$4F,$85,$E9,$8F,$9A,$3A,$E3
	.dc.b	$E6,$AF,$C2,$61,$E3,$9C,$DF,$29,$CE,$3E,$99,$F8,$95,$B7,$B1,$C0
	.dc.b	$98,$77,$72,$4A,$E7,$7E,$9C,$7E,$F4,$FB,$9B,$8F,$95,$B3,$E3,$6C
	.dc.b	$F7,$3A,$3F,$11,$E5,$1F,$3B,$1F,$45,$7D,$9C,$3C,$98,$7C,$8B,$F5
	.dc.b	$D6,$AF,$E3,$E3,$ED,$AD,$61,$C6,$31,$8E,$F0,$FC,$96,$15,$F8,$78
	.dc.b	$F6,$9D,$37,$1F,$80,$7D,$79,$F5,$63,$DA,$3F,$82,$DF,$38,$F8,$F3
	.dc.b	$EC,$3C,$F3,$C9,$A6,$71,$F3,$57,$1C,$3E,$24,$7C,$BA,$F8,$87,$EA
	.dc.b	$B9,$9E,$FF,$FA,$22,$3F,$AE,$2F,$A5,$FE,$FC,$E7,$1F,$EF,$8E,$67
	.dc.b	$66,$A7,$6D,$39,$FD,$2F,$23,$9F,$CF,$58,$C9,$2F,$AC,$65,$FE,$77
	.dc.b	$F3,$7F,$CE,$8F,$F7,$77,$BC,$F8,$CF,$F1,$35,$FF,$39,$9D,$34,$FE
	.dc.b	$F9,$3B,$F0,$33,$7D,$27,$4F,$FA,$A6,$75,$18,$D7,$FC,$49,$A7,$F8
	.dc.b	$DE,$BE,$0F,$8B,$F1,$3E,$93,$E2,$B2,$FE,$4B,$96,$9F,$33,$E9,$3E
	.dc.b	$5F,$BD,$F3,$1B,$19,$3E,$8C,$65,$A7,$F8,$03,$E9,$3F,$DE,$D8,$7F
	.dc.b	$A3,$DB,$9F,$C3,$F5,$BF,$CE,$FA,$7F,$9C,$E7,$74,$69,$C3,$7F,$F5
	.dc.b	$24,$EF,$1E,$38,$FF,$9F,$69,$1F,$E7,$DB,$C3,$5F,$13,$5E,$CE,$78
	.dc.b	$FF,$1B,$F3,$BF,$8A,$F5,$FF,$16,$DC,$3F,$1E,$DB,$FC,$6B,$71,$FA
	.dc.b	$3C,$FF,$92,$F2,$7F,$26,$DB,$FC,$BB,$6F,$F2,$AD,$AF,$F4,$A7,$D4
	.dc.b	$2F,$CC,$DB,$FF,$37,$DA,$30,$FF,$9C,$9A,$FF,$A7,$1F,$F6,$43,$7D
	.dc.b	$5E,$64,$9E,$7F,$46,$FA,$37,$C1,$E1,$B7,$FE,$7E,$97,$57,$F7,$E8
	.dc.b	$C4,$E9,$FF,$55,$F9,$4E,$BF,$F5,$86,$BF,$EA,$F8,$7D,$16,$33,$FF
	.dc.b	$5A,$46,$2F,$FC,$8E,$F1,$6D,$ED,$FF,$6C,$46,$DF,$F6,$7B,$6F,$B4
	.dc.b	$EE,$53,$FE,$D6,$7A,$FD,$16,$60,$DF,$6A,$4E,$88,$76,$7A,$7E,$28
	.dc.b	$FA,$0E,$5E,$1C,$AF,$F2,$FC,$E5,$B3,$97,$CF,$3C,$A2,$46,$88,$7B
	.dc.b	$3E,$17,$D5,$F4,$6F,$8D,$F6,$74,$3E,$62,$9A,$8F,$81,$F4,$2F,$A3
	.dc.b	$F8,$5F,$47,$E3,$CB,$FD,$EA,$DE,$2F,$F7,$15,$6E,$7F,$CE,$FF,$46
	.dc.b	$F6,$FE,$52,$FA,$3A,$62,$3C,$7F,$47,$2B,$FC,$FF,$AA,$3E,$97,$D1
	.dc.b	$BD,$3A,$4A,$FF,$42,$C6,$5F,$F4,$3F,$D1,$FF,$BA,$9B,$EC,$FD,$B8
	.dc.b	$C4,$35,$FF,$7D,$46,$1F,$F7,$E7,$D1,$B9,$3F,$05,$E5,$BE,$83,$F0
	.dc.b	$9F,$46,$FA,$3A,$6A,$26,$5F,$E9,$89,$D3,$67,$2F,$87,$C0,$32,$FB
	.dc.b	$3B,$21,$B3,$57,$F9,$5F,$39,$F0,$9F,$97,$FA,$0F,$E3,$35,$FE,$A8
	.dc.b	$6B,$C4,$FE,$95,$1C,$4E,$1E,$7C,$8F,$F5,$7D,$74,$6C,$E5,$E5,$5D
	.dc.b	$36,$65,$EA,$9E,$A7,$FA,$E3,$E1,$FF,$1C,$FF,$AE,$BE,$8B,$EB,$3E
	.dc.b	$36,$D7,$28,$F8,$73,$C1,$F4,$36,$35,$09,$E3,$5E,$16,$C7,$2E,$EF
	.dc.b	$DD,$5F,$1A,$D7,$2F,$EF,$6F,$76,$AD,$3F,$53,$E8,$DF,$37,$BF,$90
	.dc.b	$93,$FE,$D4,$9C,$B6,$72,$FB,$67,$F8,$54,$EA,$3E,$93,$73,$F5,$E0
	.dc.b	$AE,$31,$EE,$9D,$65,$F3,$7E,$8F,$F0,$B8,$7C,$9F,$27,$D4,$7F,$17
	.dc.b	$8C,$DF,$C6,$6D,$7C,$59,$DD,$7F,$BB,$27,$2F,$E4,$9A,$FC,$D8,$DB
	.dc.b	$F8,$98,$C9,$FD,$35,$FB,$9E,$37,$CB,$3C,$5E,$94,$8C,$90,$E4,$FF
	.dc.b	$2D,$FA,$27,$0D,$E6,$95,$FF,$00,$27,$2D,$9E,$04,$7F,$53,$9F,$E1
	.dc.b	$3E,$89,$F4,$7F,$BD,$97,$D1,$BE,$3F,$F2,$D5,$AB,$FD,$8C,$E6,$BF
	.dc.b	$67,$8F,$81,$E3,$7F,$9D,$8C,$4E,$E5,$7C,$DD,$9F,$BC,$71,$1B,$9E
	.dc.b	$35,$CF,$1F,$D4,$4F,$3A,$DC,$F9,$B1,$F5,$4F,$1A,$DC,$F1,$C3,$D2
	.dc.b	$EF,$1B,$1E,$1B,$E3,$3B,$8F,$28,$C7,$C6,$EE,$B8,$BE,$CA,$E2,$FC
	.dc.b	$4B,$F0,$B4,$DE,$3E,$87,$91,$DA,$FA,$BD,$E7,$8B,$D6,$F8,$DF,$68
	.dc.b	$D1,$E2,$AE,$37,$D6,$38,$1F,$C3,$CE,$E5,$7E,$93,$67,$F0,$4E,$7E
	.dc.b	$77,$FD,$02,$F1,$3E,$CB,$87,$5F,$03,$FE,$26,$A4,$E6,$78,$7D,$1F
	.dc.b	$F8,$A9,$FE,$72,$D4,$D2,$3F,$43,$DF,$FE,$E5,$C7,$F3,$A3,$89,$9F
	.dc.b	$3B,$F9,$16,$FF,$8D,$8C,$BF,$8F,$AE,$31,$F6,$8C,$EF,$FC,$C6,$33
	.dc.b	$3C,$67,$C6,$75,$BF,$03,$88,$E9,$7E,$0B,$F6,$CE,$CE,$C6,$DE,$D1
	.dc.b	$B8,$FA,$58,$7E,$23,$0F,$0D,$72,$BE,$B8,$78,$2B,$58,$71,$C3,$FA
	.dc.b	$44,$62,$75,$1E,$1A,$F5,$4E,$63,$C9,$B7,$F4,$2C,$7E,$1B,$FE,$75
	.dc.b	$9D,$39,$C3,$FC,$FB,$5F,$DB,$AB,$78,$79,$71,$F5,$BF,$C2,$8F,$95
	.dc.b	$D1,$C2,$95,$FD,$0A,$75,$0F,$87,$7D,$1B,$E8,$F8,$5E,$EB,$FB,$83
	.dc.b	$73,$F4,$63,$E4,$C3,$8D,$F0,$8E,$3F,$19,$32,$EF,$4F,$A3,$FF,$4E
	.dc.b	$E2,$AD,$65,$D9,$E9,$3A,$73,$97,$C7,$73,$E1,$E3,$1F,$EF,$1E,$F7
	.dc.b	$FA,$0A,$BF,$B6,$B6,$3F,$C2,$A3,$E8,$7E,$03,$F0,$B1,$FF,$3F,$F5
	.dc.b	$7C,$7A,$E1,$1F,$E1,$C6,$CF,$1D,$35,$52,$39,$4E,$AF,$AD,$FF,$85
	.dc.b	$E3,$9A,$FA,$37,$D1,$FD,$56,$DF,$D4,$3C,$DE,$5E,$66,$47,$9E,$77
	.dc.b	$5F,$E2,$B8,$D9,$99,$DC,$7F,$89,$E7,$79,$FF,$85,$6B,$73,$9A,$FC
	.dc.b	$34,$7F,$06,$3E,$21,$63,$B6,$2F,$97,$DF,$44,$AD,$BF,$96,$7E,$74
	.dc.b	$66,$BF,$C7,$11,$CF,$0F,$F1,$0B,$4F,$AB,$7D,$5E,$67,$D5,$F8,$7F
	.dc.b	$C3,$37,$83,$B4,$BE,$5F,$F3,$D5,$AC,$3F,$C8,$CD,$C3,$FC,$5E,$DC
	.dc.b	$3B,$D7,$0C,$24,$75,$3E,$96,$1D,$5C,$61,$FE,$51,$FA,$17,$C3,$7E
	.dc.b	$B5,$B3,$FA,$CF,$A3,$7D,$1B,$E6,$35,$72,$BF,$21,$F4,$3F,$EF,$A7
	.dc.b	$FC,$A1,$87,$2C,$7F,$CB,$D3,$97,$FC,$DA,$D4,$D8,$E1,$B7,$F9,$32
	.dc.b	$35,$F4,$58,$BF,$F2,$E3,$6F,$FC,$EF,$F4,$6F,$8F,$FE,$0D,$39,$93
	.dc.b	$6F,$FC,$EB,$7F,$A7,$8C,$7A,$7F,$C3,$7A,$9F,$0F,$4F,$E2,$17,$C3
	.dc.b	$0F,$F3,$C4,$EB,$0F,$D8,$7D,$1F,$F9,$1D,$BE,$8F,$F0,$BF,$99,$FF
	.dc.b	$79,$46,$09,$7F,$D7,$EB,$83,$6B,$FC,$D6,$FF,$9E,$63,$2F,$F6,$D7
	.dc.b	$7B,$7F,$9D,$70,$61,$FE,$67,$C3,$AC,$E6,$31,$7F,$AD,$FA,$17,$D1
	.dc.b	$FE,$75,$B7,$FB,$75,$EF,$E6,$64,$E5,$E1,$18,$AE,$9F,$47,$CF,$E8
	.dc.b	$FA,$36,$FA,$36,$77,$1D,$E7,$F9,$E7,$CD,$79,$B3,$FF,$45,$B6,$7F
	.dc.b	$D6,$7A,$7B,$BE,$67,$7E,$DC,$9F,$F0,$46,$2B,$F7,$BF,$E1,$36,$BF
	.dc.b	$EA,$36,$BF,$EB,$E7,$B3,$FD,$27,$67,$CD,$75,$8F,$F5,$F6,$3F,$3A
	.dc.b	$3E,$47,$D1,$CC,$3E,$47,$2D,$BF,$9D,$37,$4F,$E2,$DC,$3F,$89,$6C
	.dc.b	$7F,$62,$F0,$7F,$D6,$D8,$7F,$5A,$7D,$ED,$BF,$E7,$E7,$FB,$3E,$32
	.dc.b	$D3,$35,$FE,$B7,$8C,$B6,$33,$3F,$ED,$C9,$CB,$F0,$FC,$8F,$A2,$7E
	.dc.b	$5F,$E2,$6C,$7F,$4B,$D9,$3F,$94,$BF,$90,$D7,$E8,$36,$37,$3F,$51
	.dc.b	$FE,$A3,$AF,$E5,$BE,$A7,$AD,$FC,$CD,$D1,$F0,$E3,$FD,$09,$7F,$07
	.dc.b	$D1,$C7,$8D,$96,$46,$4F,$8D,$18,$9F,$89,$E3,$73,$F7,$7A,$36,$BE
	.dc.b	$63,$67,$71,$F3,$E7,$FC,$1D,$F4,$7F,$A2,$8F,$ED,$8D,$9F,$F8,$01
	.dc.b	$F4,$6F,$A4,$FE,$49,$5A,$EB,$E2,$7C,$5B,$E1,$7F,$E6,$56,$BF,$EF
	.dc.b	$56,$A0,$F7,$E3,$FF,$79,$7A,$4F,$A9,$7C,$2F,$6C,$7F,$C4,$33,$98
	.dc.b	$F9,$5F,$4B,$F2,$FC,$EB,$5F,$8B,$6E,$5F,$89,$6E,$3F,$87,$7E,$97
	.dc.b	$D2,$CA,$FC,$97,$C3,$DB,$FC,$84,$DB,$FC,$74,$DA,$F5,$53,$E9,$5F
	.dc.b	$0B,$FE,$39,$6C,$FF,$8A,$B9,$DF,$43,$9A,$93,$FE,$1B,$AD,$DF,$E1
	.dc.b	$EB,$2B,$15,$78,$2B,$71,$89,$F9,$91,$89,$FE,$17,$87,$81,$ED,$87
	.dc.b	$4B,$E1,$7E,$98,$FE,$75,$86,$6B,$B4,$59,$FE,$11,$35,$E1,$3D,$0F
	.dc.b	$74,$67,$0F,$14,$F9,$2B,$AD,$6E,$3D,$8F,$63,$17,$FC,$EE,$F9,$57
	.dc.b	$87,$A7,$FA,$0C,$62,$B6,$7B,$9C,$5F,$43,$C1,$1E,$43,$6E,$0D,$1E
	.dc.b	$A3,$E8,$3A,$71,$7F,$1B,$0F,$5D,$F8,$AB,$51,$EE,$AD,$C7,$8C,$CE
	.dc.b	$1E,$9C,$7E,$94,$F8,$2B,$81,$27,$E2,$C6,$5F,$5E,$1F,$0E,$F9,$9E
	.dc.b	$AC,$3C,$55,$B8,$CB,$EA,$30,$49,$F5,$57,$66,$CF,$07,$B4,$EE,$3E
	.dc.b	$A3,$4F,$C0,$BB,$33,$1C,$0E,$71,$B3,$E1,$9C,$0C,$5F,$E7,$9F,$F7
	.dc.b	$53,$5E,$E6,$2B,$F9,$84,$66,$3E,$C1,$9A,$FE,$C7,$19,$8E,$F7,$FE
	.dc.b	$06,$EB,$FA,$04,$AF,$91,$8F,$F9,$F2,$FE,$DB,$70,$FF,$0D,$56,$F0
	.dc.b	$F8,$18,$73,$C3,$8C,$E7,$1F,$79,$FE,$11,$71,$1F,$10,$FD,$8B,$8A
	.dc.b	$FC,$84,$EE,$79,$92,$3B,$E1,$E3,$8F,$E1,$15,$C6,$7D,$B3,$97,$FA
	.dc.b	$64,$67,$1F,$5B,$71,$F5,$DF,$5C,$3C,$EF,$90,$C1,$E7,$BF,$F0,$25
	.dc.b	$71,$8F,$09,$A8,$F4,$BC,$5F,$3C,$6D,$F1,$C6,$DF,$3C,$66,$7E,$04
	.dc.b	$70,$9D,$5F,$A1,$B7,$E8,$9D,$12,$BF,$C0,$1F,$05,$FC,$8A,$BF,$C1
	.dc.b	$4B,$E9,$7F,$C1,$CE,$6B,$FC,$14,$E6,$76,$6A,$76,$D3,$9F,$D2,$F2
	.dc.b	$39,$FD,$0D,$8C,$92,$FA,$C6,$5F,$DE,$7C,$FF,$F4,$13,$FE,$00,$F2
	.dc.b	$BE,$3B,$FC,$71,$7F,$CF,$E7,$4F,$E6,$9F,$CF,$91,$C1,$FF,$00,$33
	.dc.b	$7D,$4F,$F0,$A7,$F8,$A4,$FF,$41,$64,$66,$7F,$C4,$5A,$3C,$AF,$8A
	.dc.b	$7C,$6F,$26,$CF,$CA,$8B,$9F,$C8,$9B,$87,$D0,$FA,$6F,$9B,$F0,$3F
	.dc.b	$39,$B8,$7D,$26,$DF,$D0,$6D,$7F,$4E,$FA,$65,$F9,$EF,$FB,$4C,$61
	.dc.b	$FE,$FB,$AF,$FA,$9E,$31,$87,$F3,$66,$BF,$DC,$03,$87,$9D,$7F,$8D
	.dc.b	$E3,$11,$FE,$0F,$79,$DF,$FA,$2F,$EA,$1F,$53,$E3,$AD,$5D,$E8,$56
	.dc.b	$B2,$FC,$7F,$CD,$FE,$3B,$D9,$FC,$7B,$63,$27,$E5,$23,$2D,$3F,$2F
	.dc.b	$D7,$F9,$6E,$AF,$CB,$B6,$32,$7E,$72,$32,$D3,$FC,$51,$A7,$F7,$8F
	.dc.b	$A8,$FF,$68,$69,$F0,$DA,$FF,$85,$1A,$FF,$8B,$DA,$BE,$A1,$F5,$0F
	.dc.b	$88,$FF,$46,$C6,$1B,$D4,$DF,$35,$BB,$FA,$43,$EA,$5E,$5F,$FA,$5A
	.dc.b	$30,$D8,$CC,$7F,$A6,$9A,$F2,$87,$5F,$F8,$B3,$4F,$F0,$AC,$3E,$A7
	.dc.b	$1B,$7F,$8C,$23,$17,$F1,$23,$17,$A8,$FF,$27,$4F,$47,$46,$2B,$4F
	.dc.b	$A1,$B1,$B8,$FF,$02,$BF,$EE,$67,$89,$EE,$FA,$9D,$9C,$3D,$F9,$1A
	.dc.b	$94,$EA,$31,$3E,$0F,$A9,$7D,$4F,$83,$C4,$95,$DA,$72,$D9,$F9,$91
	.dc.b	$88,$FA,$2D,$7F,$48,$D7,$FC,$AA,$EF,$4F,$F1,$FB,$74,$CD,$7C,$0F
	.dc.b	$8D,$F8,$7F,$52,$F9,$3F,$B4,$DC,$7F,$DA,$2D,$7F,$02,$D7,$F6,$9F
	.dc.b	$4E,$FA,$9F,$C1,$36,$3F,$DB,$2D,$7F,$12,$D7,$F7,$9F,$50,$F7,$3F
	.dc.b	$14,$D8,$FF,$70,$B5,$FC,$8B,$5F,$E0,$9F,$4C,$F3,$FE,$EF,$80,$FA
	.dc.b	$7A,$7E,$02,$75,$5F,$71,$AF,$F9,$11,$BC,$3F,$E6,$66,$F0,$7F,$A0
	.dc.b	$3E,$9D,$EC,$7E,$2B,$E9,$DC,$5F,$8C,$9C,$E3,$F8,$86,$BF,$E5,$09
	.dc.b	$DB,$F8,$CF,$A1,$7D,$2F,$E5,$3E,$9D,$F4,$F4,$FC,$BC,$EA,$B8,$56
	.dc.b	$5F,$CA,$D7,$07,$24,$3F,$CB,$EE,$5B,$5D,$27,$F4,$7E,$6B,$E2,$3F
	.dc.b	$4B,$F4,$9F,$CE,$2F,$F4,$2D,$7F,$CD,$33,$B7,$F4,$AD,$7F,$CD,$B1
	.dc.b	$BE,$0F,$F4,$7B,$77,$FF,$54,$D6,$E7,$FD,$26,$D8,$FF,$56,$CE,$EB
	.dc.b	$FD,$90,$D7,$FB,$65,$FE,$EB,$E2,$BF,$D8,$F1,$FB,$6F,$A7,$FA,$EF
	.dc.b	$FB,$3D,$AA,$36,$FF,$A0,$9B,$8F,$FA,$92,$76,$70,$8C,$E1,$FE,$C0
	.dc.b	$6D,$75,$E5,$FE,$17,$5B,$F7,$FF,$86,$FD,$3B,$E7,$F3,$F2,$53,$0F
	.dc.b	$E1,$73,$96,$CE,$5F,$F4,$AD,$AB,$9C,$C9,$F7,$46,$F3,$FF,$5D,$DF
	.dc.b	$39,$FF,$4E,$FD,$3F,$E1,$DE,$7E,$37,$98,$FE,$43,$19,$AF,$F6,$1B
	.dc.b	$75,$FF,$6E,$E5,$FC,$DA,$7E,$25,$70,$E1,$98,$7C,$F8,$E8,$FF,$26
	.dc.b	$C3,$93,$FE,$FF,$5E,$87,$C6,$F1,$64,$EE,$3C,$66,$1F,$83,$18,$7F
	.dc.b	$98,$FD,$3B,$8F,$F9,$CA,$E7,$A6,$22,$4F,$06,$BF,$EC,$0F,$A6,$7D
	.dc.b	$3F,$FB,$B6,$1F,$4E,$F8,$EF,$F7,$ED,$6B,$75,$6A,$7F,$E0,$0E,$5E
	.dc.b	$38,$CD,$F1,$77,$19,$78,$1C,$AF,$A3,$80,$E2,$79,$42,$79,$46,$1B
	.dc.b	$5E,$6C,$3D,$53,$C2,$F9,$4F,$9B,$1F,$4E,$1D,$F0,$E2,$F8,$F0,$E3
	.dc.b	$5C,$F1,$F5,$BC,$66,$9E,$CB,$E7,$8F,$7B,$E3,$3C,$EF,$6F,$A6,$B7
	.dc.b	$7C,$E3,$E0,$63,$E6,$78,$1E,$59,$DB,$DC,$E7,$1B,$7E,$3D,$F8,$1E
	.dc.b	$C7,$83,$1F,$6B,$AC,$38,$CF,$DF,$8F,$F2,$55,$7D,$26,$C6,$5F,$84
	.dc.b	$7F,$87,$A1,$F4,$2F,$80,$FF,$10,$CF,$D8,$8C,$3B,$6B,$D9,$AF,$F8
	.dc.b	$A3,$E9,$FF,$C4,$C9,$A4,$7F,$C5,$4F,$CA,$F8,$3F,$EE,$9C,$BF,$32
	.dc.b	$7F,$27,$39,$F3,$13,$99,$E8,$7E,$36,$B8,$BF,$8F,$38,$99,$9E,$2E
	.dc.b	$4E,$3B,$FE,$4A,$78,$BF,$94,$8E,$2E,$67,$8B,$98,$E3,$BF,$E4,$2B
	.dc.b	$8B,$F0,$63,$C1,$7E,$7C,$3D,$4D,$8C,$9E,$38,$CB,$4C,$DF,$78,$CB
	.dc.b	$63,$35,$E1,$C3,$FC,$FD,$0C,$3E,$1E,$BF,$02,$3F,$4B,$87,$68,$C4
	.dc.b	$EB,$1E,$47,$F9,$0A,$BE,$0E,$18,$55,$C3,$A7,$BD,$FF,$41,$BF,$F2
	.dc.b	$8F,$AA,$FA,$77,$D3,$EA,$E5,$7F,$93,$AF,$86,$3D,$23,$F8,$8E,$1C
	.dc.b	$6F,$73,$93,$F8,$AD,$3E,$2F,$C2,$7F,$A6,$FE,$9F,$FD,$43,$87,$D3
	.dc.b	$8C,$BD,$DD,$7C,$0E,$2F,$A1,$F1,$E3,$1A,$7E,$1B,$26,$1D,$2B,$C3
	.dc.b	$5F,$3B,$DC,$FE,$7F,$8F,$F9,$A1,$B3,$E7,$87,$D2,$FF,$A3,$5D,$75
	.dc.b	$F1,$23,$FE,$78,$37,$8E,$2B,$CB,$7D,$32,$FB,$4D,$AF,$E6,$E7,$03
	.dc.b	$A9,$FE,$7A,$FA,$77,$D3,$FE,$11,$B1,$DF,$2F,$F1,$DB,$B9,$D1,$1D
	.dc.b	$79,$B7,$1F,$8F,$7C,$9B,$5F,$EC,$DA,$F7,$D6,$F1,$FC,$84,$7F,$0C
	.dc.b	$3F,$D1,$A5,$AF,$8D,$83,$E5,$BF,$33,$39,$AE,$3D,$18,$3B,$56,$E7
	.dc.b	$9C,$65,$A7,$B6,$B1,$7F,$7B,$E6,$7F,$11,$3F,$E9,$E6,$C7,$8B,$E5
	.dc.b	$FF,$53,$18,$C7,$F5,$0D,$AF,$E9,$AD,$AF,$E2,$78,$AF,$8F,$04,$BE
	.dc.b	$CF,$A2,$31,$5F,$90,$FA,$37,$D2,$7E,$E1,$B7,$FE,$BA,$FA,$77,$D3
	.dc.b	$BE,$9F,$57,$2B,$FD,$6F,$F4,$BF,$E0,$57,$FD,$73,$87,$2C,$7F,$3B
	.dc.b	$39,$6F,$8D,$FC,$1A,$31,$BF,$F0,$5A,$E0,$FF,$07,$8C,$DF,$F6,$CF
	.dc.b	$A6,$CB,$71,$FE,$1F,$F4,$EF,$8E,$FE,$23,$5A,$CF,$FC,$17,$87,$F9
	.dc.b	$BF,$0F,$3F,$A9,$AC,$7C,$4F,$F1,$8E,$A7,$C2,$D3,$35,$FE,$D4,$8C
	.dc.b	$BF,$ED,$6F,$A7,$FF,$24,$37,$D3,$FE,$7B,$B7,$EE,$7E,$6B,$FE,$00
	.dc.b	$CE,$49,$A7,$F0,$BC,$13,$FE,$67,$9C,$D7,$1F,$5F,$F9,$5E,$3D,$B0
	.dc.b	$E1,$7B,$7F,$84,$C6,$6B,$85,$FE,$EF,$E9,$1F,$4F,$FA,$56,$C6,$7D
	.dc.b	$FC,$4C,$8E,$0E,$E3,$18,$7F,$0B,$A7,$D3,$F3,$FA,$7E,$8D,$C3,$C0
	.dc.b	$DA,$E0,$DC,$7F,$A1,$7D,$37,$A6,$7F,$C4,$E1,$C3,$FE,$2A,$6B,$FE
	.dc.b	$44,$6B,$BD,$7E,$7D,$FF,$38,$D5,$F3,$5F,$D4,$AF,$E3,$61,$F4,$E3
	.dc.b	$17,$C3,$06,$FF,$4A,$73,$8F,$F4,$AC,$79,$BF,$2F,$E9,$FE,$51,$F2
	.dc.b	$F8,$EC,$64,$98,$7C,$78,$CB,$FD,$27,$C1,$FF,$22,$CF,$F9,$7B,$3F
	.dc.b	$F0,$1C,$7F,$50,$8C,$1F,$E6,$27,$73,$C4,$FE,$74,$53,$D4,$F1,$71
	.dc.b	$1C,$94,$66,$7E,$8F,$91,$FD,$CE,$7E,$73,$5F,$9C,$D7,$EA,$37,$5F
	.dc.b	$E8,$11,$F2,$32,$7D,$3F,$CF,$AF,$99,$AF,$E7,$23,$19,$7C,$B6,$DF
	.dc.b	$E6,$9F,$4B,$FE,$83,$8F,$46,$5F,$E8,$2F,$43,$3E,$2E,$99,$3A,$AE
	.dc.b	$07,$C5,$9C,$BF,$1F,$C3,$EB,$ED,$F6,$6D,$71,$6C,$EE,$3A,$4F,$F8
	.dc.b	$6B,$E9,$5C,$FF,$87,$F3,$DF,$4A,$F8,$1D,$4E,$63,$18,$7D,$1C,$58
	.dc.b	$72,$8F,$A9,$87,$2C,$38,$C6,$2F,$8D,$FF,$81,$2F,$83,$FE,$11,$BE
	.dc.b	$15,$D6,$73,$C7,$FE,$4C,$9C,$D7,$E1,$FE,$97,$E6,$F9,$5F,$39,$B8
	.dc.b	$7C,$96,$CE,$E3,$E6,$4F,$D4,$FA,$59,$1F,$3F,$E1,$2D,$7E,$4E,$6E
	.dc.b	$3F,$9A,$84,$E2,$BF,$D4,$7F,$4A,$F8,$1F,$C2,$F4,$7E,$2A,$D3,$9A
	.dc.b	$FD,$87,$47,$F9,$2A,$31,$3F,$C8,$23,$17,$C9,$B5,$C0,$FA,$15,$C3
	.dc.b	$1F,$E6,$53,$AB,$E5,$5A,$BE,$4D,$39,$47,$F8,$D2,$F9,$47,$F0,$DA
	.dc.b	$E5,$7D,$23,$85,$74,$38,$5F,$46,$C7,$49,$FF,$31,$5F,$49,$FD,$55
	.dc.b	$F4,$AE,$C7,$4A,$ED,$3C,$8E,$CF,$2B,$EC,$D9,$EC,$7D,$08,$CB,$4C
	.dc.b	$CF,$FB,$BA,$FC,$5D,$1F,$D4,$B0,$F1,$4F,$92,$BC,$47,$91,$F1,$61
	.dc.b	$E4,$3C,$0F,$92,$3B,$34,$ED,$87,$89,$B1,$E2,$7F,$03,$19,$6C,$67
	.dc.b	$0F,$2E,$3F,$87,$6E,$1E,$6C,$7D,$15,$E6,$BF,$41,$E6,$7D,$18,$F9
	.dc.b	$8F,$43,$E6,$6C,$79,$0F,$41,$E2,$2D,$F8,$27,$B5,$78,$27,$B1,$E0
	.dc.b	$AE,$D3,$E0,$3B,$16,$3B,$47,$81,$EC,$DA,$ED,$1D,$2B,$B4,$F4,$8E
	.dc.b	$D5,$D1,$A7,$47,$B3,$D1,$B5,$D2,$79,$57,$49,$E4,$DA,$E4,$74,$39
	.dc.b	$16,$B9,$4F,$0C,$3E,$A6,$3F,$66,$73,$5C,$AF,$A3,$C9,$B1,$A8,$C5
	.dc.b	$72,$BF,$C0,$5F,$32,$38,$24,$FD,$48,$CB,$63,$38,$7C,$B9,$CB,$63
	.dc.b	$27,$C4,$8C,$B4,$CC,$FB,$E3,$2D,$8C,$E1,$EA,$9C,$B6,$33,$7E,$BA
	.dc.b	$F2,$12,$FC,$4D,$9F,$03,$4F,$B7,$CE,$F7,$53,$FD,$F9,$7D,$2D,$8C
	.dc.b	$35,$FE,$D3,$1A,$69,$A3,$3F,$4B,$CC,$E9,$F3,$96,$32,$4C,$3F,$C3
	.dc.b	$14,$9D,$47,$F2,$5F,$9A,$FF,$15,$64,$FA,$57,$D0,$7F,$39,$9C,$C6
	.dc.b	$31,$FE,$6B,$18,$9D,$47,$F8,$5E,$B9,$46,$2F,$FD,$E1,$18,$9E,$26
	.dc.b	$EF,$A4,$F9,$BE,$97,$E3,$1F,$1F,$C4,$B5,$F1,$22,$CF,$B2,$6D,$FF
	.dc.b	$8C,$B3,$7D,$2F,$CF,$F7,$FE,$83,$71,$F9,$AD,$C3,$E2,$B6,$FE,$BC
	.dc.b	$07,$D2,$FF,$BE,$78,$DF,$48,$FA,$5F,$F2,$95,$6B,$2F,$F1,$4C,$E7
	.dc.b	$2F,$F0,$56,$1C,$63,$14,$C7,$BB,$C6,$BF,$C8,$18,$F6,$C7,$AE,$1C
	.dc.b	$5B,$8F,$67,$F1,$9F,$31,$F8,$DF,$5A,$E1,$F8,$A6,$E3,$F8,$26,$E5
	.dc.b	$F1,$DF,$CA,$70,$FE,$57,$AA,$DF,$E4,$9B,$7F,$8A,$6D,$FC,$F7,$E0
	.dc.b	$FD,$23,$E9,$1F,$09,$FE,$84,$BF,$95,$39,$8C,$42,$3F,$29,$BE,$1D
	.dc.b	$47,$1F,$A4,$79,$1C,$21,$18,$79,$B5,$E1,$1B,$87,$C4,$52,$4F,$07
	.dc.b	$A4,$BA,$7F,$01,$3F,$57,$5B,$2F,$A4,$CE,$DF,$62,$31,$5F,$E4,$77
	.dc.b	$93,$6F,$77,$CA,$EC,$FE,$36,$B8,$BD,$CC,$56,$9F,$83,$EC,$D9,$7D
	.dc.b	$25,$31,$12,$7C,$51,$97,$C9,$1C,$FE,$91,$F4,$9F,$2D,$B3,$F5,$1A
	.dc.b	$FD,$96,$BF,$EB,$E7,$39,$FE,$61,$B9,$FC,$86,$CF,$B7,$E2,$3C,$17
	.dc.b	$23,$FC,$D9,$3A,$7E,$EC,$FB,$FE,$91,$F2,$14,$93,$FE,$73,$FA,$9E
	.dc.b	$1F,$48,$FA,$5F,$8C,$D9,$FF,$3C,$35,$F9,$83,$E9,$1F,$53,$F3,$57
	.dc.b	$D3,$52,$4F,$FA,$06,$3F,$91,$BF,$E8,$3F,$A5,$FF,$32,$BF,$E8,$5E
	.dc.b	$F7,$D3,$7D,$76,$C6,$FD,$1F,$B5,$F4,$DF,$E9,$28,$C1,$0F,$2B,$97
	.dc.b	$CF,$F4,$CE,$1E,$3E,$DC,$9F,$F4,$A4,$E5,$FF,$4B,$7C,$7B,$E8,$7F
	.dc.b	$11,$F4,$CF,$A6,$A7,$E2,$E7,$55,$C2,$B2,$FF,$A7,$EB,$83,$92,$1F
	.dc.b	$29,$CB,$6B,$A4,$FE,$5F,$E9,$9F,$0D,$F9,$9F,$A3,$79,$B4,$D4,$4C
	.dc.b	$3F,$D5,$D3,$A6,$CE,$5F,$B3,$5D,$1C,$90,$CD,$7E,$9F,$E2,$7F,$D7
	.dc.b	$91,$FE,$B9,$FA,$5F,$AE,$FE,$09,$BE,$07,$F3,$09,$DC,$FE,$19,$B1
	.dc.b	$FC,$D2,$77,$E0,$7F,$04,$6D,$7F,$45,$AD,$CF,$F0,$86,$C7,$F4,$99
	.dc.b	$DA,$6C,$6A,$1F,$4D,$FC,$59,$F4,$7C,$8F,$F0,$1A,$3F,$C0,$E7,$53
	.dc.b	$F9,$C6,$E1,$E7,$9E,$06,$FC,$89,$5D,$08,$71,$7F,$87,$7D,$37,$E1
	.dc.b	$23,$E3,$78,$FE,$13,$3C,$12,$3E,$64,$65,$FF,$73,$4F,$1C,$3F,$AD
	.dc.b	$B7,$4F,$91,$5B,$AF,$D9,$CE,$5F,$ED,$CB,$CF,$F8,$1E,$FC,$9E,$B1
	.dc.b	$E6,$30,$FD,$B8,$C3,$FE,$F7,$FA,$67,$1F,$F2,$95,$CD,$4C,$44,$9E
	.dc.b	$0D,$7F,$83,$FD,$2B,$E9,$BF,$BB,$43,$E9,$9F,$1D,$FD,$1A,$B5,$BF
	.dc.b	$F8,$1E,$73,$5F,$E0,$9C,$BC,$6E,$6F,$FA,$6C,$62,$3A,$47,$3C,$3B
	.dc.b	$D7,$29,$61,$D9,$D9,$D7,$1F,$EA,$F3,$BC,$3A,$3C,$6B,$83,$6F,$AE
	.dc.b	$1E,$2B,$E2,$DC,$3C,$C7,$F8,$86,$F8,$C7,$97,$1C,$CB,$1E,$F3,$C6
	.dc.b	$3C,$A7,$A3,$0E,$EF,$36,$BC,$70,$F0,$BE,$B3,$AE,$1E,$07,$86,$1E
	.dc.b	$0C,$3B,$C6,$CE,$F3,$B7,$C1,$3B,$AE,$B1,$DC,$EB,$7E,$2A,$F9,$91
	.dc.b	$FE,$00,$38,$9F,$E1,$97,$2F,$C1,$31,$F4,$1F,$E7,$D6,$FB,$BF,$E1
	.dc.b	$C9,$4E,$67,$A4,$62,$7F,$C3,$50,$FA,$6A,$47,$FC,$3C,$DD,$3F,$05
	.dc.b	$F1,$7F,$D9,$38,$FE,$8D,$7F,$95,$6B,$E9,$CE,$63,$E8,$99,$3E,$9B
	.dc.b	$9A,$FC,$64,$65,$EB,$E6,$FF,$98,$1C,$C7,$94,$E3,$19,$DF,$FC,$79
	.dc.b	$7E,$08,$C4,$FC,$98,$C6,$1E,$68,$E5,$7D,$A3,$13,$A9,$F3,$9A,$BF
	.dc.b	$F1,$8E,$1E,$77,$BC,$78,$AB,$8B,$4E,$35,$CE,$FD,$2D,$AE,$78,$7F
	.dc.b	$9D,$B0,$F8,$DB,$7F,$9C,$30,$F8,$55,$AC,$7F,$D0,$13,$FE,$74,$AD
	.dc.b	$C6,$6B,$FC,$16,$70,$3F,$D0,$D1,$A6,$F3,$77,$3F,$D1,$75,$FE,$8C
	.dc.b	$F5,$DF,$4C,$FA,$6D,$63,$E0,$3E,$0D,$E3,$0F,$F2,$25,$F0,$AC,$61
	.dc.b	$E5,$F8,$CF,$61,$DB,$0F,$F4,$C5,$3E,$9B,$FB,$8C,$75,$C3,$CF,$5C
	.dc.b	$23,$10,$CD,$3C,$E1,$7C,$FC,$2E,$11,$FE,$2C,$AD,$DF,$26,$C7,$A3
	.dc.b	$0F,$0B,$BC,$3F,$D4,$FE,$E7,$F3,$FB,$F0,$B6,$7F,$CC,$50,$FA,$4F
	.dc.b	$0B,$AE,$BE,$24,$7D,$86,$F3,$C5,$7F,$90,$2F,$A6,$9F,$EA,$E6,$D7
	.dc.b	$FA,$F4,$E0,$75,$3E,$1F,$D3,$3E,$9B,$FD,$7A,$D8,$EF,$97,$F8,$D5
	.dc.b	$DC,$E8,$8E,$BC,$BB,$8F,$F9,$4A,$F9,$36,$BE,$75,$7B,$EB,$78,$FF
	.dc.b	$B1,$23,$F8,$59,$F5,$8B,$5F,$1B,$07,$CB,$7F,$B4,$A7,$35,$C7,$AB
	.dc.b	$07,$6A,$DC,$F3,$8C,$B4,$F6,$D6,$2F,$FC,$D3,$F1,$DF,$73,$0F,$F6
	.dc.b	$BB,$6B,$E5,$60,$F9,$6F,$F7,$1C,$E6,$FF,$DD,$4D,$4E,$6B,$EF,$4F
	.dc.b	$0E,$09,$1F,$94,$9D,$5F,$58,$C4,$FF,$BC,$3E,$8D,$F4,$14,$FC,$C7
	.dc.b	$D3,$3E,$99,$F4,$D9,$C3,$F3,$F7,$C1,$FD,$0F,$D2,$7F,$81,$9C,$E1
	.dc.b	$FE,$FC,$8C,$6B,$FE,$8A,$6E,$DF,$EF,$86,$D7,$FB,$22,$FA,$56,$FC
	.dc.b	$6F,$D5,$2F,$A5,$CE,$1F,$F0,$01,$B7,$FE,$D0,$FA,$67,$C6,$F0,$9C
	.dc.b	$D8,$65,$FE,$6E,$BF,$4F,$A5,$BD,$3D,$0F,$83,$A9,$F0,$D4,$CD,$7F
	.dc.b	$00,$8C,$BF,$AA,$FA,$47,$D2,$FE,$51,$B1,$9F,$6B,$13,$23,$83,$B8
	.dc.b	$C6,$7F,$BA,$AE,$2D,$9D,$57,$03,$C1,$F4,$9D,$FE,$97,$C0,$D8,$C9
	.dc.b	$E5,$8C,$B4,$FE,$33,$F4,$8F,$67,$E7,$B6,$3F,$90,$66,$F9,$D7,$CC
	.dc.b	$EA,$E5,$7F,$80,$73,$F9,$D8,$71,$70,$D7,$FC,$0C,$D7,$F8,$56,$8F
	.dc.b	$A5,$F9,$07,$C9,$E0,$B7,$F3,$62,$D7,$C5,$9B,$CC,$AE,$75,$87,$AC
	.dc.b	$DA,$FD,$49,$FC,$64,$57,$17,$3A,$7F,$39,$EE,$FC,$44,$53,$F9,$59
	.dc.b	$C2,$FF,$6A,$F3,$AE,$53,$FD,$2A,$7F,$A4,$3C,$9B,$3C,$A7,$C0,$B2
	.dc.b	$FE,$5A,$FC,$AF,$1B,$FA,$56,$1F,$47,$E1,$FE,$AC,$4C,$3E,$85,$71
	.dc.b	$8F,$EE,$8F,$18,$CD,$FF,$45,$8F,$5E,$3F,$A3,$F3,$BD,$87,$B7,$C3
	.dc.b	$B7,$E6,$8B,$5E,$F3,$E4,$CE,$33,$F7,$C7,$F9,$F2,$B8,$1E,$E9,$D3
	.dc.b	$F1,$BC,$3F,$95,$EC,$FC,$B6,$C7,$C4,$6D,$77,$C7,$F7,$11,$8C,$3E
	.dc.b	$3F,$D1,$3A,$3F,$82,$B7,$0F,$EE,$DF,$52,$FA,$9F,$E3,$31,$FD,$DA
	.dc.b	$75,$97,$AC,$38,$7F,$CB,$4D,$7F,$02,$D7,$F8,$53,$72,$FE,$19,$F4
	.dc.b	$9F,$85,$F1,$FF,$0C,$DB,$FB,$ED,$BF,$60,$71,$9F,$CA,$6C,$FE,$31
	.dc.b	$B3,$F9,$8F,$A4,$FC,$91,$F9,$3F,$88,$B5,$F8,$C8,$B3,$DA,$6D,$FE
	.dc.b	$3B,$E9,$1E,$FF,$F2,$C6,$E1,$FB,$EE,$D7,$CD,$E6,$A4,$FF,$87,$AB
	.dc.b	$78,$FF,$8B,$2F,$73,$9A,$FD,$47,$B9,$FB,$38,$FD,$0A,$FA,$4C,$C6
	.dc.b	$9A,$E9,$B7,$F8,$19,$E1,$3C,$8E,$05,$38,$16,$38,$4F,$2C,$BF,$67
	.dc.b	$5C,$A7,$A5,$72,$6B,$C9,$A7,$27,$A4,$F2,$AE,$91,$FD,$D2,$BA,$4F
	.dc.b	$68,$E8,$D7,$A3,$6B,$A1,$D9,$E9,$1D,$8E,$95,$DA,$7F,$BF,$D7,$69
	.dc.b	$F0,$1D,$8A,$76,$2C,$76,$AF,$01,$D9,$F0,$47,$63,$C0,$F6,$AF,$06
	.dc.b	$EB,$F0,$4F,$8A,$BC,$0D,$7C,$0D,$3C,$0F,$8A,$FC,$07,$89,$F0,$47
	.dc.b	$88,$F0,$3E,$2A,$F0,$5F,$8A,$FC,$31,$E3,$9F,$0B,$5F,$0B,$6B,$C2
	.dc.b	$78,$EF,$C3,$3E,$33,$C2,$F8,$E3,$C2,$78,$E7,$C3,$3E,$3A,$F0,$9F
	.dc.b	$9B,$9E,$B3,$DE,$BA,$C7,$7A,$EA,$77,$7A,$C7,$79,$EB,$1D,$EB,$AD
	.dc.b	$7B,$1E,$D5,$D2,$BB,$4F,$4A,$EC,$74,$7B,$47,$49,$ED,$5D,$27,$E2
	.dc.b	$BD,$1E,$53,$D2,$B9,$1D,$1E,$51,$D2,$79,$4F,$C0,$79,$3C,$27,$91
	.dc.b	$C1,$E5,$1C,$2B,$D4,$F0,$9D,$1C,$1D,$57,$91,$D1,$8A,$EC,$E2,$FF
	.dc.b	$08,$49,$FA,$CD,$9F,$98,$DB,$F8,$6D,$9F,$63,$67,$CE,$DB,$F2,$36
	.dc.b	$FC,$15,$D8,$95,$D9,$B7,$D9,$A7,$62,$C7,$66,$DF,$66,$D7,$69,$E8
	.dc.b	$4A,$FE,$33,$F4,$9F,$CB,$0F,$E5,$BF,$49,$6F,$F9,$6C,$5A,$FE,$51
	.dc.b	$36,$BF,$A0,$7C,$D7,$F3,$F3,$FD,$A9,$EC,$BE,$2B,$FC,$21,$8F,$F3
	.dc.b	$D9,$D3,$4F,$F7,$04,$EF,$B7,$38,$FF,$35,$F1,$7F,$9F,$9F,$D0,$3E
	.dc.b	$8E,$DF,$F4,$08,$B3,$FC,$EE,$6E,$5F,$CD,$26,$D7,$F2,$B9,$B5,$FC
	.dc.b	$83,$E8,$FF,$A5,$FB,$BF,$D3,$1B,$5F,$D3,$1B,$3F,$D1,$DB,$87,$F8
	.dc.b	$B3,$E8,$BF,$DD,$2D,$BF,$83,$AF,$FA,$42,$FF,$D0,$95,$AC,$3F,$C7
	.dc.b	$EF,$FB,$C2,$3F,$C8,$21,$C3,$AA,$FF,$17,$B9,$7F,$C6,$3C,$3F,$CA
	.dc.b	$7E,$4B,$FB,$1F,$A5,$FD,$91,$B8,$7F,$64,$6D,$FF,$5F,$6E,$5F,$D6
	.dc.b	$1B,$3F,$D3,$DB,$3F,$D1,$3A,$3F,$B7,$F8,$BF,$DC,$1B,$5F,$DC,$1B
	.dc.b	$3F,$DA,$DB,$7F,$E5,$0F,$A9,$FF,$57,$CF,$B9,$AF,$F9,$81,$AF,$FA
	.dc.b	$A5,$AB,$E8,$5F,$44,$F8,$1F,$F2,$9C,$61,$BC,$EF,$2F,$35,$BB,$FF
	.dc.b	$44,$79,$0E,$9F,$F3,$7E,$7F,$E6,$48,$7D,$0E,$36,$FF,$39,$C6,$2F
	.dc.b	$FD,$79,$5B,$BD,$47,$FA,$FE,$7A,$3A,$AE,$37,$A8,$4F,$37,$FC,$88
	.dc.b	$F0,$6D,$6E,$35,$7E,$5A,$C4,$E7,$D4,$D7,$D0,$CC,$3C,$07,$8E,$34
	.dc.b	$E0,$F0,$C6,$DF,$0C,$79,$7E,$81,$F4,$3E,$1F,$0A,$57,$19,$CB,$67
	.dc.b	$FD,$06,$A3,$2F,$FA,$D5,$AF,$E7,$1A,$FF,$A7,$BF,$11,$7C,$3F,$C5
	.dc.b	$5F,$11,$4C,$44,$9F,$54,$E9,$B3,$97,$FC,$DD,$1D,$1C,$90,$CC,$FD
	.dc.b	$3F,$A2,$7C,$65,$24,$FC,$7F,$A2,$F5,$F0,$7F,$AA,$3E,$85,$F4,$5F
	.dc.b	$6D,$B5,$F3,$5A,$FF,$9F,$FE,$85,$F4,$5F,$81,$5E,$D5,$24,$FD,$58
	.dc.b	$FF,$31,$BF,$5F,$E8,$9E,$8F,$E2,$7C,$27,$D1,$53,$F1,$93,$AA,$FB
	.dc.b	$DF,$45,$FE,$97,$79,$5E,$8F,$C9,$46,$63,$FD,$29,$F4,$4F,$91,$CF
	.dc.b	$B9,$27,$F1,$13,$97,$F1,$5F,$16,$F9,$DF,$CE,$7D,$13,$E8,$A9,$F9
	.dc.b	$F9,$D5,$70,$AC,$BF,$93,$AE,$0E,$48,$7F,$A9,$DC,$B6,$BA,$4F,$EA
	.dc.b	$FD,$07,$C2,$7E,$B7,$E7,$5E,$8D,$35,$13,$0F,$D1,$4E,$9B,$39,$7F
	.dc.b	$D6,$D5,$D1,$C9,$0C,$D7,$EE,$FE,$21,$F5,$1F,$BD,$FA,$3F,$A5,$5F
	.dc.b	$EB,$F6,$F5,$FF,$BC,$2B,$73,$FB,$36,$BB,$E3,$E4,$75,$B9,$1F,$B9
	.dc.b	$BE,$7E,$52,$24,$61,$C5,$F4,$C3,$FD,$FF,$F4,$6F,$9B,$F3,$7D,$1C
	.dc.b	$BF,$E0,$93,$96,$CE,$5F,$54,$FF,$19,$9D,$47,$FB,$6D,$B4,$C7,$C9
	.dc.b	$2C,$57,$CE,$BC,$78,$7F,$23,$FA,$3F,$C2,$DF,$C7,$F3,$BE,$51,$9E
	.dc.b	$29,$1F,$06,$32,$FF,$1B,$9E,$37,$E1,$8F,$89,$18,$CB,$C7,$97,$F3
	.dc.b	$09,$E4,$71,$C3,$C5,$EB,$7D,$9F,$2E,$4F,$58,$CB,$F4,$63,$2F,$FB
	.dc.b	$EF,$04,$FF,$3D,$FA,$37,$2D,$E8,$97,$FC,$C6,$72,$D9,$E0,$47,$F9
	.dc.b	$B7,$D1,$3E,$8D,$B7,$DF,$FA,$37,$C8,$7C,$EA,$D4,$E6,$3E,$8E,$3D
	.dc.b	$B0,$5F,$07,$F1,$11,$98,$C4,$2F,$9B,$C2,$BB,$61,$E2,$8C,$1C,$B1
	.dc.b	$FB,$F7,$CB,$0E,$97,$C3,$1F,$25,$FE,$5F,$1E,$B7,$BC,$38,$DF,$A1
	.dc.b	$B5,$BC,$7C,$15,$E6,$BE,$37,$BA,$EA,$31,$F0,$BC,$71,$CD,$70,$8F
	.dc.b	$41,$C2,$76,$FC,$1C,$3B,$DF,$A6,$3D,$E7,$5C,$7C,$AD,$3A,$CF,$07
	.dc.b	$C0,$70,$AE,$A7,$6B,$F1,$47,$7C,$3F,$C7,$EF,$03,$FC,$83,$1F,$49
	.dc.b	$B5,$F4,$5C,$FC,$FF,$FB,$9D,$E3,$28,$C4,$6B,$E1,$BA,$BF,$76,$32
	.dc.b	$F8,$3E,$8F,$FB,$CA,$69,$1F,$EF,$AF,$CE,$F8,$6F,$EC,$FC,$9F,$42
	.dc.b	$7F,$89,$4E,$7C,$EF,$D9,$6F,$F8,$C9,$C9,$F2,$AF,$8B,$F3,$0C,$C7
	.dc.b	$E4,$A3,$2F,$18,$F2,$1C,$67,$AE,$FE,$98,$C6,$1E,$37,$75,$CE,$B7
	.dc.b	$1F,$E6,$CC,$3C,$D5,$B6,$C7,$8C,$E3,$7F,$C6,$B0,$F1,$BD,$F0,$E9
	.dc.b	$7C,$6B,$C2,$71,$BE,$73,$E9,$6D,$73,$C3,$F8,$B6,$BF,$C5,$ED,$8F
	.dc.b	$C4,$8C,$47,$C4,$AE,$18,$FF,$8E,$6B,$4E,$27,$F8,$2B,$E5,$9C,$57
	.dc.b	$5A,$FE,$81,$3C,$79,$FC,$2F,$F4,$9F,$69,$F4,$7F,$E3,$08,$D1,$31
	.dc.b	$EA,$7F,$4F,$BE,$6E,$2B,$8F,$A9,$D2,$35,$85,$7D,$BF,$1D,$6C,$EE
	.dc.b	$FF,$C8,$3E,$1E,$62,$63,$EF,$75,$EA,$7F,$A9,$F0,$F1,$46,$9F,$29
	.dc.b	$DA,$7F,$CC,$37,$D1,$E4,$7F,$98,$7D,$EF,$E6,$ED,$E9,$F4,$2F,$A1
	.dc.b	$FF,$44,$39,$25,$F6,$AE,$2F,$C4,$8E,$17,$EF,$9C,$EF,$F7,$5B,$3F
	.dc.b	$D9,$1B,$5F,$D9,$EF,$BB,$D7,$0F,$91,$F4,$6F,$A3,$A7,$9A,$BF,$C6
	.dc.b	$D8,$7F,$9E,$E7,$4D,$9D,$BF,$E8,$23,$D9,$E6,$E3,$0F,$8F,$5C,$1B
	.dc.b	$5D,$67,$59,$7E,$56,$FF,$86,$1F,$E8,$92,$D7,$C1,$C1,$F2,$DF,$7E
	.dc.b	$73,$5C,$7A,$30,$7A,$66,$CF,$38,$B7,$CE,$6D,$7E,$97,$E1,$FF,$0F
	.dc.b	$7F,$E9,$96,$D7,$87,$E5,$BF,$1D,$18,$AE,$1C,$18,$3E,$44,$DB,$E5
	.dc.b	$16,$F9,$46,$0F,$F5,$37,$D0,$3E,$77,$F3,$4D,$8F,$F5,$6F,$D1,$BE
	.dc.b	$8D,$F4,$7A,$B9,$5F,$EA,$FF,$A1,$FF,$80,$18,$B1,$E4,$FE,$62,$32
	.dc.b	$4F,$1B,$8C,$65,$FD,$57,$D1,$70,$DB,$F8,$43,$70,$FF,$17,$FD,$1B
	.dc.b	$E4,$37,$9F,$EC,$72,$FF,$32,$D7,$8F,$D4,$CB,$4F,$F6,$27,$5B,$E2
	.dc.b	$3F,$70,$D8,$D6,$BF,$AE,$8C,$3F,$A1,$FA,$07,$D1,$FE,$A9,$B5,$F9
	.dc.b	$15,$EE,$66,$64,$E5,$E1,$18,$AE,$91,$77,$FF,$1E,$52,$F9,$BC,$4E
	.dc.b	$DF,$47,$E1,$FA,$3F,$13,$63,$27,$8E,$32,$D3,$FD,$C9,$F3,$2F,$67
	.dc.b	$F9,$8E,$5F,$E6,$D8,$C4,$FF,$6A,$6C,$7F,$78,$6B,$FE,$60,$0E,$B1
	.dc.b	$FF,$2A,$68,$F9,$C7,$CC,$FF,$25,$AD,$61,$FC,$CE,$72,$FC,$9F,$A3
	.dc.b	$F9,$87,$CD,$E1,$B1,$92,$65,$F4,$E3,$2F,$D5,$9B,$97,$CA,$DB,$F0
	.dc.b	$73,$9B,$FF,$45,$4F,$F9,$A7,$BB,$FC,$F3,$18,$3F,$D1,$CF,$F4,$77
	.dc.b	$12,$8D,$BF,$D1,$9D,$34,$FF,$4B,$47,$08,$CC,$F7,$9C,$9B,$D3,$E2
	.dc.b	$78,$BF,$4A,$3E,$5B,$74,$FE,$63,$1F,$3D,$AF,$FA,$41,$FE,$5B,$3F
	.dc.b	$3F,$EA,$79,$E3,$FD,$13,$54,$7F,$57,$CB,$FD,$55,$E7,$5C,$7F,$CD
	.dc.b	$B7,$E7,$7F,$8E,$B6,$34,$47,$5E,$15,$9F,$81,$37,$4F,$74,$DA,$F9
	.dc.b	$74,$F0,$BC,$1E,$DF,$85,$B3,$E1,$6C,$FC,$B6,$DF,$F8,$27,$E8,$BF
	.dc.b	$47,$3F,$DD,$BC,$D7,$D4,$FF,$B5,$30,$FA,$F2,$AD,$DF,$FA,$9E,$F7
	.dc.b	$F0,$1F,$DD,$E7,$37,$FD,$FF,$0E,$58,$7F,$6E,$9C,$BF,$E1,$88,$CE
	.dc.b	$1F,$6F,$E8,$FE,$57,$91,$F2,$DB,$87,$CB,$6D,$FE,$15,$B5,$C6,$7E
	.dc.b	$85,$71,$69,$F4,$FE,$8F,$EB,$1F,$5F,$E0,$EC,$64,$98,$7E,$46,$32
	.dc.b	$FF,$81,$BE,$8D,$F0,$1F,$E1,$1E,$5F,$EE,$72,$9C,$DF,$E1,$FA,$3F
	.dc.b	$C3,$51,$89,$FF,$10,$57,$08,$C6,$7F,$BE,$8F,$EF,$BC,$AF,$6F,$FD
	.dc.b	$C9,$8F,$E8,$57,$D1,$FE,$45,$CE,$3F,$A9,$73,$3B,$35,$3B,$69,$93
	.dc.b	$A3,$96,$99,$9E,$91,$96,$C6,$70,$FE,$01,$5E,$09,$F1,$4F,$66,$BD
	.dc.b	$1A,$72,$2C,$F2,$AE,$CD,$9E,$C7,$F0,$AA,$ED,$1E,$08,$EC,$D7,$B3
	.dc.b	$6B,$B3,$4E,$8F,$82,$3A,$57,$81,$A7,$82,$3F,$8B,$57,$82,$3C,$47
	.dc.b	$80,$A7,$80,$B1,$E0,$6D,$78,$0F,$13,$DA,$3C,$47,$6A,$F1,$36,$3C
	.dc.b	$57,$F8,$39,$CB,$63,$37,$E5,$AF,$3C,$79,$5A,$F9,$5A,$79,$4B,$7E
	.dc.b	$58,$F3,$9E,$57,$CF,$1E,$33,$CF,$5E,$16,$D7,$86,$3F,$31,$5E,$16
	.dc.b	$BE,$16,$CF,$85,$B5,$E1,$8F,$1C,$F8,$4F,$1B,$E1,$8F,$1C,$F8,$5B
	.dc.b	$5E,$13,$F4,$6F,$38,$EB,$5C,$EB,$A9,$CD,$EB,$1C,$DB,$5C,$DF,$53
	.dc.b	$D1,$E5,$1D,$27,$91,$D1,$E4,$D9,$E5,$3F,$05,$E5,$1C,$27,$91,$C0
	.dc.b	$B1,$C2,$3E,$33,$C2,$34,$D3,$55,$EF,$71,$7F,$86,$73,$7E,$87,$2F
	.dc.b	$E2,$23,$44,$70,$4B,$FB,$B1,$96,$C6,$6B,$E8,$CE,$5B,$19,$3E,$5C
	.dc.b	$65,$A6,$67,$E3,$46,$5B,$19,$BF,$87,$39,$6C,$66,$FD,$75,$E4,$25
	.dc.b	$F8,$9B,$3F,$C2,$7E,$8F,$F9,$41,$FC,$A7,$E7,$6C,$7F,$2D,$8B,$3F
	.dc.b	$CA,$E6,$D7,$FB,$BF,$37,$CD,$FF,$B5,$E1,$F5,$2F,$A1,$FE,$7F,$39
	.dc.b	$8C,$65,$FD,$AA,$F8,$BF,$E2,$2B,$E3,$7C,$23,$15,$C3,$1F,$E7,$9F
	.dc.b	$49,$FC,$FC,$FE,$81,$E2,$5A,$FE,$89,$16,$7F,$A1,$4D,$C3,$F9,$AC
	.dc.b	$DA,$FE,$55,$36,$BF,$8E,$FD,$27,$F4,$BF,$77,$FA,$63,$6B,$FA,$83
	.dc.b	$67,$FA,$73,$70,$4A,$67,$D2,$7F,$BB,$FC,$37,$D4,$BE,$A7,$FD,$F9
	.dc.b	$39,$8C,$5F,$F8,$3A,$31,$3A,$CB,$FD,$FD,$5A,$C3,$FA,$F7,$CB,$FF
	.dc.b	$61,$F4,$FF,$B1,$37,$1F,$EC,$CD,$9D,$C7,$F6,$99,$DB,$4D,$96,$3F
	.dc.b	$B5,$70,$7F,$6D,$E9,$FE,$DC,$D9,$FE,$E4,$D9,$DC,$7F,$75,$9F,$87
	.dc.b	$0D,$6A,$FA,$57,$D2,$BE,$0B,$FD,$01,$3F,$26,$73,$18,$8E,$1C,$EE
	.dc.b	$1C,$56,$A7,$8F,$D4,$BC,$8E,$10,$8C,$3C,$DA,$F0,$8D,$9F,$9C,$7F
	.dc.b	$C7,$EB,$E9,$BE,$E4,$7A,$F5,$B3,$98,$DF,$56,$B5,$F7,$CF,$39,$E0
	.dc.b	$7E,$15,$E0,$E0,$F4,$5F,$1B,$EB,$76,$78,$35,$D9,$E6,$9C,$57,$7B
	.dc.b	$F0,$FA,$D8,$FA,$69,$7E,$C8,$D9,$87,$C5,$49,$CC,$7F,$98,$3E,$95
	.dc.b	$F4,$DF,$31,$B1,$F5,$5A,$FD,$A6,$BF,$EB,$B7,$38,$7E,$09,$B9,$FC
	.dc.b	$C6,$CF,$C5,$F8,$5F,$77,$D3,$3E,$26,$FC,$5C,$AF,$F3,$9C,$E5,$B3
	.dc.b	$97,$E6,$42,$79,$CC,$9E,$31,$BC,$7F,$2B,$5C,$E7,$EA,$7D,$33,$E3
	.dc.b	$BE,$83,$63,$FD,$06,$D7,$ED,$7D,$37,$A9,$7D,$2B,$E9,$BE,$C3,$63
	.dc.b	$FD,$16,$D7,$F0,$5F,$48,$FA,$6F,$BC,$BE,$9A,$92,$7F,$D2,$71,$FC
	.dc.b	$BD,$FF,$4A,$FD,$2B,$D2,$FC,$37,$86,$FA,$6A,$7E,$26,$75,$5F,$E9
	.dc.b	$CF,$A6,$7C,$46,$70,$93,$98,$FC,$87,$D3,$39,$BF,$25,$7C,$2B,$47
	.dc.b	$FA,$9A,$74,$FF,$AA,$3E,$39,$F4,$1F,$99,$FA,$67,$D3,$53,$F3,$93
	.dc.b	$AA,$E1,$59,$7F,$D6,$35,$C1,$C9,$0F,$D5,$39,$6D,$74,$9F,$D4,$7D
	.dc.b	$33,$E2,$BF,$55,$F4,$4F,$42,$9A,$89,$87,$FB,$02,$74,$D9,$CB,$FB
	.dc.b	$9A,$E8,$E4,$86,$6B,$F7,$1F,$18,$F8,$4F,$DD,$7D,$27,$D2,$AF,$E0
	.dc.b	$4D,$EF,$FE,$4F,$5B,$9F,$F6,$8B,$5D,$F1,$F2,$3A,$DC,$8F,$F6,$BD
	.dc.b	$F3,$F0,$FF,$8C,$B6,$3F,$AB,$5F,$3C,$3F,$9A,$FD,$33,$E7,$BC,$BF
	.dc.b	$4D,$2F,$FD,$C3,$39,$6C,$E5,$F4,$CF,$F1,$59,$D4,$7F,$1D,$6D,$7F
	.dc.b	$42,$C1,$5C,$63,$C3,$7F,$C7,$7E,$9B,$F0,$58,$7C,$4F,$2B,$E3,$99
	.dc.b	$E2,$91,$EE,$8C,$BF,$EF,$69,$E3,$7F,$D4,$1B,$A7,$C2,$AD,$D7,$F3
	.dc.b	$19,$CB,$FD,$E5,$7A,$9F,$5B,$E1,$24,$F5,$8F,$31,$87,$EC,$C6,$1F
	.dc.b	$F0,$16,$1F,$D8,$3E,$99,$CB,$FC,$E5,$74,$53,$11,$2B,$8E,$7F,$D4
	.dc.b	$2B,$8C,$60,$FF,$07,$38,$7F,$A5,$7D,$2B,$E9,$BF,$B8,$43,$E9,$9F
	.dc.b	$23,$FD,$6A,$B5,$9F,$F8,$5E,$73,$5D,$32,$FE,$BF,$18,$86,$3D,$E3
	.dc.b	$17,$CA,$7B,$61,$D7,$1F,$0D,$71,$BE,$13,$EB,$C3,$AB,$BC,$7C,$4D
	.dc.b	$C3,$99,$E3,$32,$53,$23,$1F,$16,$1E,$4C,$7C,$35,$DE,$B8,$9E,$AC
	.dc.b	$3C,$2F,$F8,$DE,$F8,$CF,$86,$FD,$15,$AA,$E0,$72,$AD,$BC,$6F,$BB
	.dc.b	$E5,$AE,$2E,$E3,$2F,$9D,$B8,$FC,$48,$E1,$87,$84,$F8,$4F,$F8,$40
	.dc.b	$E1,$8F,$F8,$8F,$0F,$A1,$19,$6B,$F7,$7E,$85,$F0,$FB,$6C,$7F,$8B
	.dc.b	$A5,$18,$8E,$4D,$7F,$C6,$7F,$4D,$FE,$2A,$4D,$23,$FE,$2E,$7E,$6F
	.dc.b	$C8,$FF,$67,$DB,$E7,$CF,$F9,$8E,$73,$E7,$FF,$91,$36,$FC,$64,$E4
	.dc.b	$FB,$77,$C5,$FB,$A6,$63,$F2,$51,$97,$8C,$79,$0E,$33,$D7,$4F,$4C
	.dc.b	$63,$0F,$1B,$BA,$E7,$5B,$8F,$9D,$87,$9A,$B6,$D8,$F1,$9C,$6F,$FC
	.dc.b	$87,$87,$8D,$EF,$87,$4B,$E3,$5E,$13,$8D,$F3,$9F,$4B,$6B,$9E,$1F
	.dc.b	$E6,$8D,$3F,$CC,$F8,$FC,$37,$FC,$D3,$87,$C3,$AD,$E3,$FC,$0E,$15
	.dc.b	$B8,$CC,$7F,$93,$4E,$07,$FA,$56,$34,$DE,$6F,$04,$FB,$1B,$1F,$E6
	.dc.b	$FF,$A6,$FF,$15,$4F,$7A,$FE,$8C,$DA,$DD,$78,$9E,$3E,$E2,$F8,$CF
	.dc.b	$F9,$7E,$F8,$B7,$E3,$F8,$C1,$E3,$EA,$FF,$75,$C7,$C1,$1A,$23,$FC
	.dc.b	$8F,$0F,$F2,$37,$87,$8C,$3A,$4F,$03,$57,$CE,$32,$7F,$9C,$7D,$FF
	.dc.b	$E7,$F3,$F2,$DB,$1F,$E7,$88,$7D,$27,$FA,$45,$D7,$57,$C0,$AD,$EF
	.dc.b	$FE,$9E,$6C,$FF,$B0,$CE,$39,$6E,$A4,$72,$3F,$9E,$1A,$31,$87,$FA
	.dc.b	$17,$E9,$9F,$4D,$F3,$9B,$3C,$32,$FF,$1A,$3F,$83,$8C,$D7,$1A,$E0
	.dc.b	$7D,$C9,$F5,$F9,$98,$78,$45,$9F,$81,$7D,$1E,$53,$F8,$F3,$F8,$49
	.dc.b	$79,$BA,$4B,$E5,$FF,$07,$3A,$AF,$F4,$F3,$75,$FF,$61,$B4,$F8,$D5
	.dc.b	$C5,$E8,$49,$E3,$7D,$AF,$AB,$F9,$0F,$90,$FC,$45,$7E,$8D,$B5,$EA
	.dc.b	$F9,$7D,$71,$4A,$D3,$E8,$8C,$4F,$FA,$CE,$B8,$E5,$FE,$D8,$6D,$F9
	.dc.b	$AB,$73,$FE,$DC,$FA,$37,$D0,$FE,$89,$B5,$FD,$DF,$E9,$9F,$4C,$FA
	.dc.b	$6D,$5C,$AF,$DD,$7D,$27,$F8,$11,$FD,$E6,$1C,$B4,$FF,$63,$4E,$5B
	.dc.b	$E2,$FE,$C2,$31,$9F,$8F,$0C,$4F,$ED,$3E,$9B,$94,$FF,$BC,$9B,$8F
	.dc.b	$FB,$73,$E9,$9F,$21,$BC,$3A,$7A,$72,$BF,$87,$5F,$0C,$C3,$3F,$F3
	.dc.b	$64,$FF,$13,$EA,$7C,$4F,$F0,$46,$C7,$BB,$0C,$4F,$E4,$FE,$89,$F4
	.dc.b	$B7,$E9,$70,$DB,$FD,$2C,$E9,$B3,$A6,$9A,$9E,$19,$7E,$E6,$B8,$36
	.dc.b	$BF,$8B,$FD,$23,$D6,$A7,$F1,$8F,$9F,$7C,$DE,$32,$FE,$5D,$5B,$7F
	.dc.b	$BA,$61,$F0,$A3,$0E,$1A,$FF,$79,$6B,$FC,$0F,$17,$D2,$FC,$13,$E1
	.dc.b	$6B,$6F,$DF,$16,$BD,$53,$79,$15,$F4,$BA,$FE,$9D,$7F,$38,$7E,$AF
	.dc.b	$05,$8F,$E2,$42,$B8,$A9,$DD,$7F,$93,$5E,$6D,$8F,$E4,$07,$17,$C0
	.dc.b	$F4,$9E,$73,$CA,$34,$75,$35,$3C,$2B,$C9,$1E,$57,$E9,$F8,$72,$7F
	.dc.b	$9E,$6B,$FE,$E1,$7E,$8C,$66,$BE,$73,$FD,$91,$FA,$4D,$5A,$7F,$3A
	.dc.b	$7E,$2E,$1F,$98,$F2,$BE,$21,$F1,$7C,$0B,$87,$BE,$2D,$7A,$AB,$F3
	.dc.b	$F5,$8C,$FD,$F2,$EE,$F9,$DE,$AF,$CF,$6E,$3F,$25,$B3,$CC,$E1,$1F
	.dc.b	$36,$78,$34,$E0,$58,$F9,$63,$E9,$3F,$DE,$F1,$FE,$01,$6C,$7D,$0F
	.dc.b	$A9,$7D,$4F,$F0,$FA,$D7,$C2,$7F,$76,$BD,$E5,$F8,$2C,$3A,$4F,$F8
	.dc.b	$03,$0E,$3E,$0E,$E3,$F1,$5F,$4B,$F8,$BF,$16,$E5,$F8,$56,$E1,$ED
	.dc.b	$6E,$3F,$19,$FC,$97,$D2,$CA,$FC,$67,$C0,$DB,$F8,$93,$6F,$E6,$CD
	.dc.b	$AF,$C7,$CD,$AF,$C6,$4D,$5F,$46,$F7,$3F,$C2,$6D,$BF,$C5,$FC,$EE
	.dc.b	$26,$47,$F8,$A6,$75,$5F,$98,$E9,$74,$B0,$DC,$FE,$0F,$DA,$FD,$8E
	.dc.b	$1F,$A7,$5F,$4B,$FB,$17,$33,$B8,$C4,$E5,$AF,$17,$87,$D2,$C8,$E9
	.dc.b	$F3,$D6,$BC,$13,$77,$FD,$7E,$0A,$C5,$FF,$03,$85,$62,$FF,$7D,$1B
	.dc.b	$9F,$44,$E7,$AB,$FD,$0E,$FA,$6F,$15,$CA,$B6,$79,$0B,$7E,$46,$E1
	.dc.b	$E4,$AE,$37,$C9,$B3,$A3,$CC,$E2,$B9,$1F,$35,$C5,$7B,$AB,$C9,$5C
	.dc.b	$6B,$D7,$1D,$63,$18,$7A,$9B,$87,$BE,$BA,$DF,$39,$CE,$1D,$9F,$91
	.dc.b	$87,$6A,$DC,$60,$D9,$D2,$3C,$11,$A3,$C7,$1A,$71,$87,$8D,$B8,$78
	.dc.b	$F0,$F1,$46,$2F,$C1,$18,$25,$F8,$67,$E2,$3E,$13,$B4,$6A,$3C,$27
	.dc.b	$6C,$3D,$8E,$23,$EA,$4F,$18,$D3,$63,$96,$1F,$EC,$8A,$E2,$FA,$63
	.dc.b	$8D,$79,$EB,$C2,$79,$63,$8E,$1F,$87,$9F,$39,$B3,$B9,$63,$77,$DA
	.dc.b	$78,$47,$63,$9E,$1F,$39,$B7,$E0,$BF,$63,$E1,$23,$F9,$EB,$E9,$5F
	.dc.b	$32,$73,$87,$AD,$B8,$7D,$36,$CF,$76,$9D,$E7,$85,$F8,$08,$FC,$78
	.dc.b	$F9,$47,$4B,$F0,$BF,$A7,$9F,$80,$E4,$DB,$CA,$FC,$4D,$BF,$13,$67
	.dc.b	$C5,$3F,$19,$F2,$C7,$CE,$C3,$E2,$BF,$39,$A7,$CE,$8E,$F8,$7A,$8F
	.dc.b	$44,$F9,$0F,$16,$1F,$C1,$DF,$E5,$7F,$39,$FC,$A8,$FF,$61,$FB,$0F
	.dc.b	$88,$FE,$DB,$87,$F2,$99,$D3,$4F,$F2,$2E,$3F,$E0,$EA,$E1,$DF,$AC
	.dc.b	$3F,$95,$78,$9F,$C9,$CF,$E5,$1F,$49,$70,$FE,$59,$16,$7F,$98,$4D
	.dc.b	$C3,$F9,$0F,$D2,$7F,$2F,$F6,$FF,$98,$37,$0F,$E6,$8D,$BF,$E7,$0D
	.dc.b	$AE,$33,$FC,$E6,$B8,$B4,$FF,$76,$7D,$27,$FB,$1E,$7F,$B3,$36,$3F
	.dc.b	$DD,$5A,$FF,$9C,$1F,$F1,$8C,$67,$0F,$F3,$2B,$53,$9D,$3E,$57,$A5
	.dc.b	$FE,$E8,$8E,$1C,$0E,$17,$C9,$7F,$4B,$F4,$BF,$A6,$37,$2F,$EA,$0D
	.dc.b	$CB,$FA,$A3,$74,$FE,$8B,$CD,$FD,$4F,$C8,$FE,$A8,$DB,$FE,$B0,$DB
	.dc.b	$FE,$B8,$DA,$E3,$3F,$D7,$6B,$8B,$4F,$BA,$7F,$94,$3E,$A7,$FD,$2D
	.dc.b	$7F,$E5,$20,$E1,$D7,$A7,$FE,$35,$8C,$47,$FA,$2B,$E8,$9F,$46,$F7
	.dc.b	$FF,$CD,$B2,$9C,$C6,$1F,$EC,$1D,$D6,$B8,$57,$F9,$D3,$CA,$75,$FF
	.dc.b	$99,$35,$FF,$0F,$43,$E8,$B1,$AF,$F9,$A2,$31,$7E,$C3,$FD,$3A,$F2
	.dc.b	$70,$1C,$3F,$E3,$47,$8B,$67,$AB,$CA,$EB,$C7,$D5,$B4,$FA,$2A,$62
	.dc.b	$24,$F8,$27,$4D,$9C,$BE,$19,$E2,$E4,$E0,$F8,$3E,$85,$F4,$5E,$0F
	.dc.b	$1A,$57,$B2,$72,$D9,$FA,$2A,$32,$FF,$A8,$1A,$FD,$F6,$BF,$E9,$2F
	.dc.b	$A1,$7C,$2F,$C0,$5F,$45,$4C,$44,$9F,$9B,$3A,$6C,$E5,$FF,$30,$C3
	.dc.b	$E3,$BA,$12,$38,$FC,$1D,$77,$17,$87,$FC,$BD,$4A,$C7,$0F,$9E,$AD
	.dc.b	$FF,$B7,$2B,$4F,$E0,$3E,$87,$E0,$E1,$FE,$67,$FA,$27,$C6,$7D,$C6
	.dc.b	$D6,$F6,$FB,$FF,$3C,$FA,$0F,$C0,$FD,$0B,$E8,$A9,$A8,$DF,$81,$C3
	.dc.b	$E8,$A6,$3F,$90,$C3,$FC,$A4,$FE,$4B,$E8,$5E,$AF,$E3,$3C,$47,$D1
	.dc.b	$53,$F1,$F3,$AA,$FC,$CF,$D1,$3D,$BC,$E1,$27,$31,$F9,$6F,$A2,$7C
	.dc.b	$97,$E6,$2F,$85,$68,$FD,$1C,$E9,$FD,$2F,$9C,$F9,$CF,$CF,$FD,$13
	.dc.b	$E8,$A9,$FA,$29,$D5,$70,$AC,$BF,$AE,$AE,$0E,$48,$7F,$A8,$9C,$B6
	.dc.b	$BA,$4F,$EB,$BD,$37,$C4,$7E,$C3,$E7,$5E,$B5,$35,$13,$0F,$DE,$CE
	.dc.b	$9B,$39,$7F,$D6,$35,$D1,$C9,$0C,$D7,$EF,$BE,$31,$F4,$5F,$C0,$3E
	.dc.b	$87,$E9,$57,$FA,$E9,$BD,$FF,$EB,$1A,$DC,$FF,$0E,$6B,$BE,$3E,$47
	.dc.b	$5B,$91,$FC,$52,$F9,$F8,$68,$91,$87,$17,$D3,$0F,$F6,$07,$D1,$3E
	.dc.b	$73,$CD,$F4,$52,$FF,$8F,$CE,$5B,$39,$7D,$53,$FC,$72,$75,$1F,$ED
	.dc.b	$66,$D7,$FB,$37,$05,$71,$8F,$0D,$FF,$24,$FA,$2F,$C1,$E1,$F1,$7C
	.dc.b	$FF,$90,$67,$8A,$47,$BE,$32,$FF,$34,$9E,$37,$FE,$DB,$6E,$9F,$0E
	.dc.b	$59,$7F,$39,$79,$1C,$70,$F1,$7B,$1F,$5F,$CE,$93,$D6,$32,$FC,$F8
	.dc.b	$CB,$FE,$F4,$C1,$3F,$CF,$BE,$89,$CB,$7A,$25,$FF,$4A,$9C,$B6,$78
	.dc.b	$11,$FE,$9F,$F4,$5F,$C7,$F5,$FB,$9F,$44,$F9,$3F,$64,$E6,$93,$97
	.dc.b	$F0,$58,$76,$73,$5F,$0E,$31,$3B,$95,$73,$8C,$36,$B8,$61,$F2,$E7
	.dc.b	$85,$F2,$9F,$1F,$9C,$C7,$CF,$18,$6E,$3E,$4C,$39,$4F,$48,$C9,$E2
	.dc.b	$AE,$4F,$A0,$FC,$E6,$3E,$37,$91,$C3,$0F,$13,$A9,$E3,$8F,$A9,$B7
	.dc.b	$DE,$7A,$0A,$E2,$F8,$71,$F1,$E1,$DE,$35,$8F,$A1,$F0,$9C,$A3,$46
	.dc.b	$EB,$99,$D1,$CE,$1E,$C9,$FF,$13,$47,$D9,$6C,$65,$F9,$07,$F8,$AC
	.dc.b	$FA,$BE,$7F,$F9,$15,$C4,$FF,$5B,$F7,$BF,$C4,$ED,$8C,$CE,$DF,$F1
	.dc.b	$5F,$D1,$7F,$7C,$4D,$23,$FE,$00,$7E,$9F,$C1,$7F,$5E,$E4,$FA,$B3
	.dc.b	$FC,$36,$73,$E8,$27,$33,$D0,$F9,$B5,$C5,$F9,$E7,$13,$33,$C5,$C9
	.dc.b	$C7,$7F,$CA,$4F,$17,$F2,$D1,$C5,$CC,$F1,$73,$1C,$77,$FE,$01,$5C
	.dc.b	$5F,$95,$87,$A3,$0F,$3B,$6B,$C2,$DB,$E6,$DA,$E3,$3E,$8A,$E2,$D3
	.dc.b	$89,$6B,$CB,$1F,$E4,$8B,$F8,$5B,$FB,$EB,$F8,$A6,$1E,$18,$C4,$EA
	.dc.b	$BD,$38,$FC,$59,$E1,$7F,$07,$1C,$D7,$5E,$7D,$3F,$01,$B2,$F6,$DF
	.dc.b	$45,$FE,$32,$AE,$04,$9F,$DC,$3C,$4F,$F3,$14,$F0,$F5,$79,$CB,$3F
	.dc.b	$E9,$15,$ED,$F8,$EC,$C1,$F0,$DF,$F9,$0B,$C3,$CC,$4A,$E0,$76,$F5
	.dc.b	$FF,$D4,$D7,$C6,$33,$87,$90,$FE,$B9,$19,$71,$87,$59,$DE,$BF,$37
	.dc.b	$DF,$FE,$67,$1F,$E7,$86,$D7,$38,$7D,$07,$C5,$75,$D5,$EE,$AD,$DF
	.dc.b	$38,$F9,$73,$C3,$0D,$44,$8E,$27,$FA,$24,$D1,$8C,$3D,$43,$E8,$9F
	.dc.b	$45,$F3,$1B,$3C,$72,$FF,$18,$BF,$E9,$08,$CD,$72,$6E,$3F,$8B,$6C
	.dc.b	$FF,$80,$4F,$6F,$9B,$D7,$09,$1C,$DF,$5E,$1D,$B0,$E9,$1F,$7A,$7F
	.dc.b	$CC,$C7,$FA,$80,$B1,$F2,$65,$F3,$1B,$E8,$95,$B7,$AC,$62,$BE,$0C
	.dc.b	$F2,$CB,$F4,$4D,$3D,$17,$D1,$E4,$7F,$AC,$3E,$1F,$F0,$13,$FA,$96
	.dc.b	$C7,$0F,$98,$FC,$AC,$62,$FF,$C4,$4D,$4E,$67,$F0,$33,$C3,$8A,$47
	.dc.b	$77,$E2,$C6,$2B,$EC,$FD,$03,$E2,$3F,$46,$DB,$FF,$64,$7D,$13,$E8
	.dc.b	$9F,$31,$AB,$95,$FE,$CA,$FA,$07,$8D,$8C,$BF,$0B,$18,$9D,$4F,$EC
	.dc.b	$23,$13,$E2,$AC,$4F,$ED,$3E,$87,$C5,$3F,$8E,$6E,$FF,$C3,$FE,$89
	.dc.b	$F1,$FB,$C3,$A7,$A9,$33,$FF,$6E,$E1,$C1,$FF,$14,$67,$FE,$6A,$7F
	.dc.b	$DC,$3D,$4F,$89,$FE,$06,$D8,$F6,$5E,$27,$F4,$FF,$40,$FA,$2B,$F4
	.dc.b	$58,$6D,$FE,$9E,$74,$D9,$D3,$4D,$4F,$0D,$FF,$75,$5C,$1B,$5F,$C5
	.dc.b	$BE,$5D,$EB,$FF,$28,$9F,$F3,$24,$63,$0F,$ED,$90,$8C,$3F,$DB,$DA
	.dc.b	$FF,$92,$4D,$4F,$F1,$CF,$9B,$7C,$BF,$78,$C3,$F0,$3E,$8A,$61,$ED
	.dc.b	$D6,$E1,$E8,$9B,$C9,$FD,$86,$9A,$7C,$FA,$D4,$7F,$9E,$E7,$FB,$37
	.dc.b	$67,$F9,$6E,$30,$7E,$7A,$7F,$8D,$1F,$C9,$E3,$4D,$33,$3C,$9F,$E0
	.dc.b	$C6,$78,$3F,$A3,$CF,$09,$D4,$62,$77,$18,$AF,$0C,$7E,$B0,$CB,$E3
	.dc.b	$9F,$A5,$E1,$FD,$19,$F9,$CD,$DB,$F9,$9C,$7C,$D6,$BF,$E7,$77,$F9
	.dc.b	$8D,$7F,$A6,$E9,$B3,$EA,$3A,$CF,$F5,$68,$C3,$E4,$CB,$FD,$45,$E6
	.dc.b	$7A,$4F,$57,$83,$67,$59,$7F,$9A,$DB,$3A,$69,$AE,$FF,$85,$EC,$FC
	.dc.b	$36,$C6,$4E,$71,$96,$99,$AF,$A5,$19,$6C,$7C,$A5,$F4,$5F,$EF,$58
	.dc.b	$FC,$AB,$63,$FC,$84,$D7,$F2,$8D,$7F,$D3,$CE,$F7,$FF,$16,$79,$2F
	.dc.b	$A8,$FF,$69,$DF,$7A,$D3,$FD,$B2,$33,$7F,$87,$FA,$2F,$C0,$78,$BF
	.dc.b	$81,$6D,$7C,$86,$CE,$E3,$E5,$CF,$E1,$FE,$8A,$47,$CE,$F8,$1B,$5F
	.dc.b	$93,$9B,$9F,$E4,$E6,$D7,$E6,$E6,$D7,$E3,$66,$AF,$A2,$7B,$BF,$B3
	.dc.b	$E7,$FF,$0A,$C6,$30,$FC,$24,$62,$75,$1F,$3E,$75,$5F,$E0,$AE,$4F
	.dc.b	$C7,$BF,$D1,$BD,$9F,$F2,$E5,$7E,$2D,$7D,$17,$E9,$9C,$CF,$EB,$DC
	.dc.b	$CE,$CD,$4E,$DA,$73,$FA,$2E,$47,$3F,$A2,$B1,$92,$5F,$58,$CB,$DE
	.dc.b	$6E,$DF,$C1,$30,$56,$2F,$FB,$0C,$2B,$17,$ED,$AB,$D0,$AD,$5F,$83
	.dc.b	$0E,$95,$A9,$CD,$F8,$6B,$E2,$4E,$70,$D3,$63,$91,$E2,$71,$3C,$4F
	.dc.b	$64,$69,$FA,$13,$CA,$31,$87,$96,$31,$3C,$70,$F8,$11,$86,$D7,$0C
	.dc.b	$3C,$F1,$EA,$8C,$4E,$DF,$8A,$F9,$23,$55,$D6,$72,$73,$C3,$4E,$23
	.dc.b	$D2,$6C,$F4,$BB,$3B,$95,$D6,$1E,$B8,$C5,$7A,$27,$38,$7A,$27,$E4
	.dc.b	$1C,$A3,$37,$E2,$7B,$9E,$2A,$E4,$49,$F3,$3E,$C3,$11,$C1,$B7,$F8
	.dc.b	$1A,$E8,$F9,$67,$A5,$72,$23,$C5,$F2,$BE,$8C,$3E,$63,$92,$4F,$3B
	.dc.b	$F1,$57,$28,$C3,$70,$FC,$CB,$EF,$9F,$2C,$62,$BB,$4F,$C8,$38,$5F
	.dc.b	$82,$F9,$CF,$82,$BE,$53,$F9,$FA,$E8,$77,$C3,$DF,$3A,$AF,$85,$3A
	.dc.b	$AE,$55,$AB,$FD,$B4,$E6,$BB,$9C,$2B,$CA,$7D,$68,$D5,$F5,$3D,$58
	.dc.b	$7B,$2F,$D1,$19,$69,$E4,$C3,$C1,$5F,$AE,$39,$C7,$8A,$FB,$DF,$C6
	.dc.b	$AD,$C7,$D8,$3E,$14,$F4,$AF,$1C,$FF,$02,$9F,$E5,$5F,$37,$FE,$C6
	.dc.b	$B7,$D4,$3E,$4F,$F9,$5C,$E6,$31,$87,$F2,$98,$C4,$EA,$3F,$C0,$75
	.dc.b	$CA,$3F,$93,$7D,$1C,$9F,$E5,$5E,$1D,$C7,$F9,$44,$DB,$FE,$35,$36
	.dc.b	$BF,$8C,$FD,$1F,$F2,$FF,$6B,$F9,$83,$63,$F9,$A3,$70,$FE,$64,$DB
	.dc.b	$FE,$48,$D9,$E6,$7F,$3B,$AF,$F6,$DF,$D1,$9F,$F0,$B3,$63,$FC,$8A
	.dc.b	$DB,$FF,$21,$07,$0F,$F9,$15,$D5,$A7,$FC,$31,$F4,$EF,$A8,$EF,$87
	.dc.b	$FB,$92,$31,$3F,$D3,$7E,$5B,$FA,$67,$A3,$FD,$35,$B1,$93,$FA,$94
	.dc.b	$65,$A7,$F5,$5D,$BF,$AA,$74,$FF,$55,$6C,$64,$FE,$B5,$19,$69,$9D
	.dc.b	$3F,$AD,$46,$5B,$1F,$E1,$5F,$A4,$7D,$23,$DF,$FF,$34,$C6,$1B,$CA
	.dc.b	$F2,$73,$5B,$C3,$E4,$FD,$43,$C6,$FF,$38,$C2,$30,$FF,$9D,$1A,$F0
	.dc.b	$8D,$8E,$74,$7F,$8A,$61,$F4,$B8,$8F,$F7,$7C,$61,$D4,$E7,$93,$EB
	.dc.b	$47,$29,$E8,$F3,$AD,$6B,$CE,$79,$4F,$4B,$CB,$C7,$D3,$AF,$B7,$E9
	.dc.b	$6D,$1B,$32,$7D,$11,$A3,$0F,$A2,$30,$E8,$F7,$3D,$BE,$95,$F4,$5F
	.dc.b	$29,$B3,$FE,$9C,$6B,$F9,$86,$BF,$E9,$B7,$39,$FC,$28,$46,$32,$F9
	.dc.b	$4D,$8F,$8D,$F0,$7E,$EF,$A5,$7C,$35,$F8,$79,$5F,$E5,$B9,$CB,$67
	.dc.b	$2B,$1F,$04,$AB,$A5,$75,$F8,$E9,$7C,$DF,$07,$C3,$62,$39,$67,$F7
	.dc.b	$F0,$56,$71,$EB,$76,$B5,$36,$FF,$13,$6A,$FB,$3F,$4B,$EF,$97,$D2
	.dc.b	$BE,$37,$EE,$B6,$3F,$CF,$13,$BF,$A1,$7D,$1F,$E0,$BE,$95,$F4,$B4
	.dc.b	$FC,$2C,$EA,$FF,$D0,$0F,$F2,$F6,$FD,$2E,$8D,$F8,$3F,$8C,$FA,$47
	.dc.b	$A9,$F8,$EF,$15,$F4,$B4,$D4,$7E,$4B,$E9,$1F,$4B,$8C,$BF,$D1,$D8
	.dc.b	$72,$7F,$D2,$13,$A7,$FD,$25,$F4,$AE,$AE,$57,$27,$31,$F9,$EF,$8E
	.dc.b	$7C,$F7,$E8,$3E,$95,$F4,$B4,$D4,$4B,$FF,$4F,$4E,$9B,$39,$7E,$D9
	.dc.b	$FB,$CA,$EA,$E8,$86,$83,$A7,$1F,$4A,$F8,$BF,$D9,$FD,$07,$F3,$D9
	.dc.b	$FF,$57,$35,$DC,$7E,$E2,$78,$46,$BD,$2E,$26,$66,$46,$48,$72,$3F
	.dc.b	$D7,$5F,$1A,$FA,$5F,$E0,$7F,$47,$F3,$E7,$D0,$DF,$0F,$F6,$15,$B9
	.dc.b	$E3,$5C,$39,$24,$60,$87,$27,$2F,$C3,$9E,$4F,$C5,$FA,$57,$CF,$F8
	.dc.b	$BE,$96,$4F,$FB,$3E,$72,$D9,$CB,$E6,$9F,$E3,$73,$A8,$E9,$3A,$C7
	.dc.b	$F8,$46,$4B,$DC,$FF,$22,$FA,$5F,$A9,$5A,$F3,$BE,$19,$9E,$49,$8F
	.dc.b	$FB,$87,$1F,$E7,$F5,$B8,$FF,$72,$5F,$2F,$23,$F9,$E3,$63,$FA,$0D
	.dc.b	$70,$9F,$E8,$0D,$8F,$E8,$B3,$B4,$BD,$7F,$AB,$F0,$D2,$7B,$C7,$A4
	.dc.b	$C3,$F6,$23,$0F,$FB,$DF,$E9,$5C,$BF,$D0,$1F,$C9,$C6,$79,$B1,$32
	.dc.b	$38,$35,$FC,$D7,$D2,$FF,$18,$D3,$FA,$DF,$D2,$BE,$47,$FA,$D4,$E6
	.dc.b	$93,$97,$FB,$0E,$3E,$27,$8D,$7F,$61,$8C,$47,$08,$E7,$E7,$7F,$31
	.dc.b	$8C,$1F,$DA,$AF,$B5,$F5,$85,$F2,$8C,$36,$F9,$E1,$E0,$AE,$58,$75
	.dc.b	$BD,$E3,$FD,$CD,$B3,$B1,$8F,$83,$0E,$71,$E4,$3C,$35,$C6,$3D,$8D
	.dc.b	$8E,$F8,$73,$AE,$51,$F1,$27,$4F,$23,$58,$79,$6F,$BE,$1E,$49,$F0
	.dc.b	$63,$ED,$7A,$9E,$3C,$3A,$BC,$CD,$E1,$DA,$32,$7D,$78,$FF,$09,$1E
	.dc.b	$EC,$3E,$83,$88,$D3,$F0,$0C,$79,$BF,$EF,$37,$FC,$9F,$1F,$D5,$7E
	.dc.b	$1B,$FC,$A3,$1F,$22,$74,$F0,$8F,$F1,$5A,$FA,$5A,$47,$FC,$5E,$DC
	.dc.b	$BE,$F7,$C8,$FF,$5D,$E3,$FA,$95,$FE,$2C,$AF,$AD,$39,$8F,$A8,$64
	.dc.b	$FA,$CE,$6B,$F1,$11,$97,$AF,$A1,$FE,$2E,$73,$1E,$53,$8C,$67,$6F
	.dc.b	$F3,$6D,$F8,$23,$13,$F2,$A3,$18,$79,$A3,$95,$F6,$8C,$4E,$A7,$CE
	.dc.b	$6A,$FF,$C8,$98,$79,$DE,$F1,$E2,$AE,$2D,$38,$D7,$3B,$F4,$B6,$B9
	.dc.b	$E1,$FE,$4A,$DF,$FC,$55,$8F,$C7,$7F,$C5,$98,$7C,$3A,$DE,$3F,$E5
	.dc.b	$89,$DD,$F1,$8C,$D7,$F5,$D3,$81,$FE,$5E,$BE,$91,$EA,$E7,$E1,$39
	.dc.b	$6F,$B8,$FA,$5F,$F1,$65,$77,$38,$16,$B8,$35,$5F,$F4,$88,$D5,$73
	.dc.b	$F5,$7B,$D6,$B0,$B6,$F8,$FA,$78,$BD,$6F,$F5,$1E,$3F,$42,$32,$F7
	.dc.b	$9E,$0F,$69,$DD,$FB,$EB,$6F,$F7,$FA,$DB,$FE,$06,$C7,$FC,$A3,$E1
	.dc.b	$EB,$5F,$F2,$D7,$C0,$7F,$3D,$CB,$D8,$DB,$FF,$06,$C3,$E8,$FF,$D2
	.dc.b	$CE,$BA,$FD,$B5,$BB,$E7,$1F,$3A,$78,$63,$A8,$91,$C4,$F9,$86,$8C
	.dc.b	$61,$FE,$17,$FA,$57,$D2,$FC,$A6,$CF,$0C,$BF,$C6,$0F,$D3,$8C,$D7
	.dc.b	$1A,$E0,$78,$EB,$71,$9A,$E1,$3A,$7A,$79,$76,$7F,$C5,$11,$FE,$5F
	.dc.b	$3F,$00,$57,$E2,$D3,$E5,$FE,$FC,$E7,$3F,$F4,$33,$4F,$F5,$71,$6B
	.dc.b	$FD,$C1,$5C,$63,$AF,$4F,$22,$3E,$FB,$B5,$FE,$3B,$F8,$FF,$BB,$87
	.dc.b	$FA,$4D,$B5,$FE,$BE,$3E,$C7,$CB,$FE,$4A,$33,$5F,$EC,$26,$A7,$33
	.dc.b	$C0,$FC,$F3,$A2,$71,$7C,$BA,$E8,$FE,$7B,$E9,$5F,$43,$4F,$D7,$FD
	.dc.b	$2B,$E9,$5F,$4B,$9C,$BF,$55,$7C,$1F,$D6,$7D,$1F,$F9,$59,$CE,$1F
	.dc.b	$AA,$8C,$69,$FE,$5B,$6E,$BF,$EB,$36,$D7,$FB,$6E,$FA,$56,$FC,$6F
	.dc.b	$D9,$FD,$26,$30,$FF,$5D,$36,$FF,$DC,$5F,$4A,$F9,$0D,$E7,$E1,$F5
	.dc.b	$25,$7F,$01,$BE,$19,$7F,$BB,$34,$FF,$35,$4F,$F0,$5E,$A7,$C4,$FF
	.dc.b	$02,$6C,$7B,$B0,$C4,$FF,$06,$FA,$17,$A9,$4F,$E2,$BF,$36,$F9,$8C
	.dc.b	$61,$FC,$46,$B6,$DA,$DC,$EB,$3F,$F4,$05,$69,$FE,$DA,$1C,$3F,$DC
	.dc.b	$1A,$88,$E1,$53,$E8,$9F,$47,$FB,$F6,$C7,$EF,$1F,$87,$3F,$C2,$9C
	.dc.b	$B6,$78,$1C,$73,$B5,$FC,$BF,$AB,$F4,$F5,$FE,$9E,$7C,$9A,$FF,$3A
	.dc.b	$73,$C7,$E4,$9F,$CD,$1F,$C5,$8E,$07,$E8,$CB,$3C,$5C,$47,$19,$EB
	.dc.b	$5F,$37,$C1,$FA,$87,$D2,$F8,$5B,$19,$25,$FF,$2D,$DD,$3A,$8F,$A3
	.dc.b	$5F,$D7,$1F,$96,$D7,$E9,$37,$5F,$E6,$11,$F4,$4F,$5F,$97,$E7,$3D
	.dc.b	$1D,$D7,$1F,$8D,$3F,$E6,$16,$CE,$9A,$6B,$BB,$E0,$7A,$FF,$05,$B1
	.dc.b	$93,$9C,$65,$A7,$F5,$FF,$A2,$FF,$72,$B7,$E0,$3F,$B5,$FD,$3B,$EA
	.dc.b	$3F,$7B,$39,$26,$5F,$ED,$1B,$E2,$FF,$B6,$2B,$87,$81,$FD,$9A,$73
	.dc.b	$C5,$F6,$E9,$18,$BF,$88,$D8,$5F,$29,$EF,$5F,$6E,$5F,$47,$F7,$FC
	.dc.b	$4F,$C0,$36,$BE,$4B,$67,$F0,$ED,$AE,$5F,$47,$F8,$73,$F1,$1F,$01
	.dc.b	$6F,$E8,$C5,$BF,$C7,$CD,$AF,$57,$D1,$BD,$CF,$F0,$1B,$70,$FF,$03
	.dc.b	$F3,$BE,$2F,$35,$27,$FC,$1D,$5B,$C7,$F1,$D5,$97,$FD,$F9,$19,$C3
	.dc.b	$BD,$FF,$08,$AD,$47,$E7,$7D,$9F,$DC,$E1,$FA,$D5,$F4,$7F,$AB,$73
	.dc.b	$8F,$EB,$DC,$CE,$CD,$4E,$DA,$73,$FA,$3E,$47,$3F,$9C,$B1,$92,$5F
	.dc.b	$58,$CB,$EA,$F9,$9F,$8A,$7F,$8D,$BD,$67,$C4,$7F,$17,$BF,$87,$3A
	.dc.b	$69,$FE,$35,$9D,$F6,$E6,$FA,$4E,$96,$9F,$C1,$EB,$73,$98,$FA,$F7
	.dc.b	$C5,$D6,$1E,$1B,$E6,$48,$DD,$78,$67,$51,$89,$D4,$FE,$EF,$75,$FB
	.dc.b	$6F,$73,$98,$F0,$34,$E9,$8E,$2F,$F9,$A5,$F2,$9F,$41,$FE,$5F,$BE
	.dc.b	$73,$AB,$E0,$AB,$83,$F0,$AB,$85,$EA,$73,$87,$A4,$F8,$31,$9B,$F1
	.dc.b	$63,$F2,$30,$F6,$9D,$F0,$F1,$BF,$E6,$4C,$7C,$6F,$86,$B9,$4F,$B0
	.dc.b	$DC,$65,$E3,$5D,$70,$F1,$47,$C3,$9E,$F3,$98,$C4,$EA,$7C,$47,$B4
	.dc.b	$F1,$61,$E0,$78,$E1,$E4,$C7,$DF,$7C,$89,$3C,$1F,$2C,$6A,$3A,$DF
	.dc.b	$C4,$9C,$E3,$F1,$2F,$A3,$DE,$75,$5F,$18,$F7,$1E,$2C,$35,$7E,$0C
	.dc.b	$7D,$93,$DA,$3C,$6F,$0A,$E6,$7B,$DC,$47,$1C,$3C,$2F,$90,$F9,$D1
	.dc.b	$E1,$C7,$D4,$F0,$3D,$C6,$71,$F5,$63,$E4,$8F,$11,$AC,$3C,$15,$C8
	.dc.b	$F8,$98,$FA,$27,$A1,$23,$38,$F9,$23,$B6,$3E,$4B,$E7,$8F,$82,$3D
	.dc.b	$13,$F1,$5B,$3F,$77,$0E,$D8,$78,$6F,$85,$F4,$38,$16,$B8,$61,$E1
	.dc.b	$7B,$E1,$E1,$3E,$6D,$F3,$71,$5F,$00,$F5,$5F,$ED,$F1,$F8,$91,$F2
	.dc.b	$71,$F1,$CF,$C7,$6D,$7A,$0F,$D6,$96,$7E,$5C,$69,$F3,$46,$9A,$6B
	.dc.b	$1F,$44,$79,$AB,$C4,$F3,$9F,$14,$E9,$F4,$61,$F4,$30,$FE,$11,$E0
	.dc.b	$36,$FF,$25,$61,$FD,$07,$C5,$FE,$81,$F4,$5F,$D0,$59,$19,$3F,$A3
	.dc.b	$46,$31,$FE,$89,$F4,$5F,$D1,$FD,$9F,$E9,$0D,$AF,$E9,$8D,$9D,$C7
	.dc.b	$F4,$F9,$FF,$71,$FD,$17,$FB,$62,$4F,$FB,$6B,$7F,$E7,$FE,$8E,$AA
	.dc.b	$5F,$FB,$BE,$B5,$BF,$F9,$FB,$0F,$F2,$B4,$73,$25,$7F,$2D,$BE,$33
	.dc.b	$FD,$6F,$E4,$BF,$AF,$7A,$3F,$D7,$DB,$8F,$F6,$36,$CE,$E3,$FB,$34
	.dc.b	$FF,$65,$E7,$FE,$CD,$E3,$D9,$FE,$D4,$DA,$FE,$D0,$DE,$5F,$EE,$0F
	.dc.b	$F9,$55,$FF,$0E,$5B,$EB,$7F,$C8,$4D,$BF,$F3,$AD,$FF,$A4,$67,$27
	.dc.b	$F9,$1F,$EA,$7F,$D3,$71,$FE,$45,$0E,$1F,$F2,$43,$5F,$F3,$D3,$57
	.dc.b	$D1,$3E,$89,$EF,$F8,$23,$0D,$EB,$79,$D9,$AD,$DF,$FA,$2B,$CA,$FE
	.dc.b	$B6,$FF,$A1,$F5,$FF,$1C,$43,$E8,$7F,$CB,$51,$86,$EB,$FE,$5D,$9E
	.dc.b	$35,$FC,$DC,$38,$AE,$11,$D6,$78,$57,$28,$FF,$4B,$56,$E7,$A7,$A9
	.dc.b	$99,$E7,$F4,$36,$8D,$99,$3D,$4E,$71,$B3,$2F,$97,$E8,$1F,$43,$CF
	.dc.b	$C5,$93,$C4,$4E,$9C,$35,$F7,$61,$FE,$43,$78,$35,$ED,$5E,$99,$E1
	.dc.b	$1F,$7D,$AF,$FA,$15,$AF,$FA,$DD,$AB,$E8,$5F,$0B,$F0,$D7,$C3,$D3
	.dc.b	$11,$27,$D5,$3A,$6C,$E5,$FA,$D0,$F8,$DE,$84,$8E,$3F,$07,$5D,$C5
	.dc.b	$E0,$FF,$3E,$52,$BE,$B7,$D7,$BC,$2F,$37,$8F,$31,$F9,$93,$96,$CE
	.dc.b	$5F,$1C,$F2,$89,$1A,$21,$E8,$E7,$FC,$27,$D0,$FD,$3C,$BF,$D1,$DF
	.dc.b	$44,$F8,$DF,$C4,$36,$FE,$E5,$6F,$9B,$FD,$69,$F3,$EF,$A2,$FC,$77
	.dc.b	$D1,$3E,$8A,$9F,$91,$9D,$61,$F8,$57,$FC,$FC,$DF,$80,$D1,$BF,$17
	.dc.b	$F3,$1F,$42,$F5,$BF,$35,$E4,$3E,$8A,$9A,$8F,$CF,$7D,$13,$E0,$31
	.dc.b	$97,$E4,$F0,$E4,$FE,$56,$74,$FE,$5F,$E8,$9F,$23,$CA,$E4,$E6,$3F
	.dc.b	$55,$E7,$3E,$77,$F5,$9F,$44,$FA,$2A,$6A,$25,$FE,$8E,$74,$D9,$CB
	.dc.b	$FB,$FA,$E8,$E4,$86,$6B,$F7,$5E,$A3,$E2,$7F,$79,$F3,$8F,$56,$9A
	.dc.b	$89,$7F,$B3,$9D,$36,$72,$FF,$0F,$AE,$8E,$48,$66,$BF,$85,$7C,$6B
	.dc.b	$E8,$BF,$86,$7D,$07,$CC,$AF,$E3,$0D,$F0,$3F,$DB,$15,$B9,$FE,$08
	.dc.b	$D7,$7C,$9C,$8E,$B7,$23,$F8,$55,$F3,$ED,$44,$8C,$38,$BE,$98,$7F
	.dc.b	$B9,$FE,$89,$F3,$7E,$6F,$A2,$97,$FC,$56,$72,$D9,$CB,$EA,$9F,$E5
	.dc.b	$13,$A8,$FE,$68,$DA,$FF,$7A,$60,$AE,$31,$E1,$BF,$E6,$1F,$45,$F7
	.dc.b	$30,$F8,$BE,$87,$C8,$33,$C5,$23,$DF,$19,$7F,$95,$4F,$1B,$69,$F0
	.dc.b	$A5,$39,$8C,$65,$FD,$22,$79,$9C,$B0,$F1,$FB,$9F,$63,$D2,$93,$DA
	.dc.b	$32,$FD,$08,$CB,$FD,$4F,$04,$FF,$4B,$FA,$27,$2D,$E6,$97,$FD,$06
	.dc.b	$72,$D9,$E0,$47,$FA,$37,$D1,$7F,$29,$D7,$EE,$FD,$13,$E4,$FD,$D3
	.dc.b	$9A,$4E,$5F,$C1,$E1,$D9,$CD,$7C,$58,$C4,$EE,$55,$CE,$30,$DA,$E1
	.dc.b	$87,$CD,$9E,$17,$CA,$7C,$7E,$97,$FB,$A3,$0E,$E7,$E5,$30,$E4,$F9
	.dc.b	$30,$E5,$5D,$23,$DA,$F2,$C6,$C7,$1B,$E9,$8F,$82,$F8,$CF,$33,$67
	.dc.b	$AA,$77,$7C,$E3,$C6,$31,$F4,$3C,$4C,$E3,$E6,$79,$63,$D2,$77,$1C
	.dc.b	$70,$F0,$BA,$38,$E1,$E0,$7A,$63,$E7,$7C,$B8,$75,$3E,$C3,$FE,$35
	.dc.b	$6B,$E7,$31,$3F,$2D,$D7,$A3,$FE,$75,$7F,$C7,$87,$D8,$70,$EB,$E1
	.dc.b	$7F,$C8,$B5,$FD,$A5,$FF,$23,$C7,$F0,$6F,$A2,$A9,$A4,$76,$FC,$CF
	.dc.b	$84,$FE,$C1,$C9,$F3,$A7,$91,$F3,$23,$2F,$CE,$33,$97,$E3,$63,$2F
	.dc.b	$E3,$E3,$3B,$7E,$DE,$72,$FB,$2F,$C1,$E8,$7E,$F2,$BA,$1E,$73,$C0
	.dc.b	$E6,$FA,$46,$5F,$04,$66,$BA,$4F,$0C,$3D,$4E,$E3,$C5,$3B,$8F,$F3
	.dc.b	$45,$F8,$AB,$6D,$8F,$1B,$6F,$C2,$DB,$EF,$1E,$2A,$E1,$BF,$AE,$D8
	.dc.b	$FC,$88,$C4,$7C,$43,$F9,$DC,$66,$7C,$A7,$F9,$25,$C4,$79,$8F,$61
	.dc.b	$FE,$65,$7F,$9E,$C6,$F9,$FF,$D1,$8F,$F4,$6F,$79,$F4,$5F,$E3,$E8
	.dc.b	$D1,$2F,$F1,$B7,$C9,$B7,$E6,$7C,$AD,$7D,$7F,$21,$D9,$F4,$B7,$0F
	.dc.b	$F2,$6F,$85,$99,$DE,$1E,$BA,$C6,$07,$87,$B5,$FE,$C1,$8C,$4E,$CF
	.dc.b	$F5,$0B,$8A,$F3,$D7,$8E,$31,$1F,$EA,$1F,$57,$F9,$AD,$7F,$9C,$5B
	.dc.b	$0D,$3F,$1D,$2F,$A0,$FF,$4D,$BB,$E9,$F4,$5F,$0A,$FB,$4D,$9F,$EE
	.dc.b	$27,$1D,$37,$52,$3F,$D6,$F3,$E0,$C3,$C9,$3F,$26,$31,$1F,$D6,$7E
	.dc.b	$89,$F4,$5F,$1D,$B7,$FE,$91,$3C,$DE,$56,$66,$47,$FA,$36,$77,$5E
	.dc.b	$D8,$F2,$E7,$FE,$3A,$75,$5C,$27,$F8,$5D,$70,$7F,$D9,$0D,$7C,$91
	.dc.b	$F7,$E7,$FC,$E8,$7F,$A6,$CB,$1D,$CF,$EE,$74,$F9,$9F,$C0,$4E,$B4
	.dc.b	$FD,$13,$4F,$EF,$05,$AE,$33,$C3,$A6,$4F,$FB,$80,$EF,$3F,$ED,$D6
	.dc.b	$BC,$A3,$F2,$9F,$1B,$F8,$5B,$FD,$83,$6B,$E1,$FC,$CF,$E4,$E3,$17
	.dc.b	$FE,$2A,$6A,$73,$3F,$85,$9E,$1C,$32,$3F,$DE,$73,$F9,$0A,$FF,$09
	.dc.b	$35,$E9,$3F,$A7,$FA,$0F,$F7,$71,$3E,$7F,$07,$EA,$FE,$79,$E1,$7E
	.dc.b	$9D,$B8,$FF,$B2,$69,$59,$A9,$3F,$D1,$30,$5F,$07,$FA,$3B,$5F,$F1
	.dc.b	$31,$A9,$FD,$D7,$CF,$FF,$C0,$12,$5E,$0F,$27,$9D,$30,$ED,$87,$F9
	.dc.b	$BD,$B1,$FC,$0F,$99,$F1,$14,$93,$E5,$8C,$BE,$7F,$82,$79,$DF,$C1
	.dc.b	$5B,$1F,$EE,$F9,$7C,$BB,$E4,$F5,$72,$B9,$3F,$C4,$EF,$8E,$DF,$D9
	.dc.b	$9A,$FC,$07,$3C,$7F,$D3,$71,$EB,$F2,$7C,$EB,$F3,$51,$8C,$3F,$61
	.dc.b	$F3,$EF,$9F,$FE,$3D,$08,$C3,$AD,$B3,$18,$BF,$D2,$4E,$EB,$26,$AA
	.dc.b	$C7,$F9,$EB,$1F,$EC,$1C,$BF,$A4,$BF,$D6,$3F,$CE,$67,$47,$E5,$E3
	.dc.b	$F8,$89,$A3,$F8,$C9,$4F,$E6,$91,$C5,$C4,$72,$7F,$3C,$70,$9F,$EA
	.dc.b	$D9,$FF,$1D,$8F,$F4,$6B,$72,$FA,$7F,$01,$F5,$FB,$E6,$1F,$6D,$FA
	.dc.b	$44,$70,$7D,$BE,$07,$D4,$70,$C7,$E7,$C6,$27,$C9,$9F,$FA,$73,$C7
	.dc.b	$F4,$D7,$F9,$86,$72,$FF,$0D,$6C,$68,$8E,$BB,$2C,$FC,$5E,$BE,$7E
	.dc.b	$9F,$46,$D7,$46,$CE,$E3,$BC,$FF,$59,$F9,$EF,$F7,$54,$7C,$99,$46
	.dc.b	$1F,$E4,$5F,$3C,$FA,$07,$D0,$71,$F7,$E6,$1F,$B7,$C3,$86,$1F,$B4
	.dc.b	$6B,$FE,$9B,$6B,$FE,$BE,$7B,$72,$7F,$89,$7C,$87,$D4,$7F,$B4,$6F
	.dc.b	$C9,$5A,$7F,$C1,$71,$9B,$FC,$37,$D0,$FC,$CF,$13,$E6,$B6,$BE,$6B
	.dc.b	$67,$71,$F4,$27,$E9,$FD,$0C,$8F,$A7,$EF,$DA,$FC,$94,$DC,$FF,$37
	.dc.b	$09,$C5,$7F,$AD,$FE,$79,$ED,$FF,$81,$5B,$1F,$87,$F8,$9C,$4C,$8F
	.dc.b	$E3,$D3,$AB,$FF,$31,$B5,$FE,$3A,$D7,$B4,$FF,$84,$FA,$3F,$39,$09
	.dc.b	$E5,$8F,$F0,$39,$FD,$07,$43,$D7,$FF,$66,$E3,$F9,$45,$F4,$3F,$AC
	.dc.b	$73,$8F,$ED,$9C,$CE,$CD,$4E,$DA,$73,$FA,$1E,$47,$3F,$9B,$B1,$92
	.dc.b	$5F,$58,$CB,$FD,$BC,$F5,$7D,$0F,$C9,$3F,$C6,$5C,$AF,$86,$78,$79
	.dc.b	$C3,$F3,$1B,$AF,$18,$7F,$79,$C3,$85,$E3,$7F,$F2,$C5,$EF,$0E,$38
	.dc.b	$76,$C7,$AC,$60,$98,$F9,$6B,$83,$D6,$76,$E6,$BB,$57,$E0,$DE,$98
	.dc.b	$62,$47,$BA,$7A,$E1,$FA,$1B,$E6,$F9,$AF,$85,$6C,$F0,$5E,$30,$F1
	.dc.b	$D6,$F0,$FE,$45,$8F,$C0,$7A,$9E,$32,$DF,$F9,$E3,$1E,$F1,$98,$F5
	.dc.b	$3A,$9E,$51,$82,$57,$77,$C7,$3D,$1F,$34,$64,$F6,$61,$E7,$C7,$E1
	.dc.b	$5F,$49,$D5,$F8,$5B,$7E,$5B,$E1,$3E,$3C,$3C,$F3,$F2,$B1,$F8,$33
	.dc.b	$C4,$91,$C5,$F2,$46,$A3,$86,$3F,$52,$72,$F9,$2B,$AC,$FA,$A7,$9B
	.dc.b	$4E,$87,$C2,$73,$5D,$71,$F6,$46,$2F,$A6,$1E,$AB,$E0,$72,$C7,$DE
	.dc.b	$F5,$3B,$5F,$27,$CF,$7E,$63,$EE,$E3,$E6,$C7,$D9,$3F,$85,$3F,$83
	.dc.b	$E3,$FE,$9B,$8F,$CB,$C6,$2B,$F8,$7C,$E4,$8E,$09,$1F,$E6,$1C,$1F
	.dc.b	$50,$E2,$FC,$D4,$61,$B1,$9A,$FC,$6C,$E5,$B1,$9A,$F8,$33,$96,$C6
	.dc.b	$63,$F0,$26,$4A,$E6,$7E,$EC,$65,$B1,$99,$F8,$E7,$A5,$C3,$F6,$67
	.dc.b	$83,$A7,$FD,$E8,$61,$D4,$70,$C3,$F1,$13,$96,$C6,$70,$FF,$63,$4E
	.dc.b	$5B,$19,$BF,$F0,$A4,$E5,$B1,$9B,$F3,$61,$F2,$89,$3F,$21,$B1,$F1
	.dc.b	$9B,$5F,$11,$B3,$F0,$9B,$3E,$D8,$FF,$19,$57,$7A,$F7,$B6,$33,$5C
	.dc.b	$2B,$E0,$B7,$0F,$7B,$70,$F6,$B6,$BD,$6D,$AF,$5C,$7A,$48,$FA,$5F
	.dc.b	$EA,$1E,$57,$F4,$F3,$FA,$87,$93,$63,$24,$9F,$EA,$51,$97,$FA,$A7
	.dc.b	$CA,$FF,$57,$F6,$7F,$AC,$36,$BF,$AE,$36,$77,$1F,$D7,$E7,$FD,$F1
	.dc.b	$F4,$46,$9F,$D0,$7D,$2B,$3F,$E2,$39,$D3,$FE,$93,$6D,$7F,$A4,$03
	.dc.b	$87,$FD,$26,$D7,$F9,$F9,$AA,$FC,$7C,$FF,$42,$F2,$9F,$4E,$FA,$8F
	.dc.b	$0D,$FF,$9E,$23,$15,$FD,$DF,$E4,$BF,$BB,$7D,$27,$F7,$76,$C6,$4F
	.dc.b	$EF,$91,$96,$9F,$DF,$F7,$FE,$FD,$E6,$FF,$7F,$6C,$64,$FF,$02,$46
	.dc.b	$5A,$7F,$91,$7F,$13,$7B,$C8,$FF,$5B,$46,$1B,$19,$CB,$E3,$4E,$5F
	.dc.b	$91,$F4,$6F,$A3,$E7,$18,$6B,$FE,$BC,$87,$63,$7F,$F6,$24,$71,$C3
	.dc.b	$FD,$8D,$F5,$2F,$3B,$FD,$91,$82,$30,$EA,$33,$18,$8F,$B5,$1F,$E0
	.dc.b	$97,$FC,$C4,$BE,$93,$EC,$F6,$7A,$E7,$B4,$63,$5F,$BB,$5C,$67,$FC
	.dc.b	$34,$DA,$E0,$F3,$8E,$B5,$C5,$FF,$01,$C6,$EB,$9F,$B3,$99,$F6,$7D
	.dc.b	$25,$A3,$66,$4F,$A2,$34,$61,$FF,$16,$7D,$1B,$E9,$39,$79,$32,$7F
	.dc.b	$D0,$13,$96,$CF,$CA,$8C,$47,$E7,$DA,$FF,$BC,$1A,$FE,$79,$DE,$7F
	.dc.b	$1D,$B9,$66,$BD,$5F,$0B,$ED,$FA,$37,$C3,$DF,$A4,$99,$7F,$A4,$E7
	.dc.b	$2D,$9C,$BF,$E4,$78,$4F,$39,$E9,$F1,$F2,$B9,$3D,$FE,$1B,$11,$C6
	.dc.b	$2E,$9F,$77,$05,$7D,$1F,$A4,$78,$7E,$7F,$2E,$63,$FE,$A7,$9C,$B6
	.dc.b	$7F,$23,$18,$8F,$E0,$CD,$7F,$81,$87,$0F,$F0,$67,$86,$1F,$8F,$6E
	.dc.b	$D9,$AF,$91,$CD,$F8,$5F,$A3,$FA,$38,$7F,$93,$1A,$9A,$63,$E8,$1F
	.dc.b	$47,$FA,$66,$CF,$F1,$B0,$E1,$FE,$3C,$D7,$D2,$FF,$80,$DE,$0D,$77
	.dc.b	$E0,$75,$3F,$67,$F4,$1F,$EA,$78,$EC,$D7,$F9,$53,$5F,$8A,$D7,$F9
	.dc.b	$4F,$D0,$3D,$CF,$DE,$B6,$3D,$0D,$7F,$9A,$35,$F9,$FF,$40,$E6,$FE
	.dc.b	$0A,$D8,$F8,$0D,$7F,$9E,$35,$FA,$FE,$5B,$E6,$BF,$69,$F4,$0F,$A0
	.dc.b	$A7,$ED,$E7,$55,$FE,$EB,$6B,$FE,$9D,$9E,$0F,$FB,$BD,$AF,$FA,$86
	.dc.b	$37,$C9,$FC,$3D,$E3,$1E,$2B,$FE,$3B,$18,$8F,$E9,$AD,$7F,$D1,$BE
	.dc.b	$A3,$E2,$7F,$85,$7C,$D7,$F5,$F9,$FF,$7F,$35,$FF,$58,$4E,$DF,$F4
	.dc.b	$AE,$1F,$C7,$E3,$11,$FE,$B5,$9E,$13,$FC,$8D,$BC,$DF,$D7,$AB,$73
	.dc.b	$C6,$FF,$99,$C7,$2B,$FD,$0B,$5E,$11,$FC,$E3,$E3,$5F,$41,$FC,$87
	.dc.b	$E7,$BF,$0D,$87,$F3,$16,$E3,$FD,$B2,$B7,$87,$A2,$31,$7C,$7E,$02
	.dc.b	$4E,$4F,$F5,$8D,$70,$7F,$DA,$0D,$7F,$6A,$D7,$99,$FA,$57,$0F,$FA
	.dc.b	$EB,$E8,$1F,$31,$E7,$FA,$09,$3D,$C4,$E9,$C3,$5F,$F7,$06,$49,$E3
	.dc.b	$1F,$EE,$39,$D6,$7F,$D4,$5E,$11,$FC,$29,$AF,$F8,$39,$AF,$96,$FF
	.dc.b	$B1,$FD,$07,$E3,$34,$FF,$09,$E3,$FD,$87,$0F,$99,$E5,$FF,$4F,$AD
	.dc.b	$61,$FE,$D6,$6A,$AE,$17,$CA,$B8,$71,$4A,$FE,$15,$7F,$93,$4D,$27
	.dc.b	$A7,$F7,$7C,$A9,$3F,$1E,$53,$E3,$33,$1F,$1A,$70,$FC,$DF,$A0,$72
	.dc.b	$7F,$67,$5C,$B4,$C4,$49,$E0,$D5,$AF,$F7,$DF,$A0,$79,$18,$97,$CF
	.dc.b	$3E,$4B,$FC,$77,$5A,$BF,$F7,$DC,$E6,$BF,$DF,$B9,$79,$1C,$DF,$F9
	.dc.b	$0A,$31,$1D,$23,$9E,$1D,$EB,$94,$B0,$EC,$EC,$EB,$8F,$F9,$32,$77
	.dc.b	$87,$47,$8D,$70,$6D,$F5,$C3,$C5,$7C,$5B,$87,$98,$1E,$19,$E0,$E2
	.dc.b	$FA,$4F,$08,$F2,$0C,$3C,$51,$EC,$C7,$A4,$F3,$6C,$EA,$3D,$66,$AB
	.dc.b	$AB,$DE,$FC,$2F,$5C,$3D,$0F,$03,$C3,$3C,$1E,$C7,$92,$BB,$3C,$4E
	.dc.b	$D5,$D2,$7E,$6C,$7F,$74,$6B,$E9,$3F,$CF,$CE,$1F,$85,$E4,$BE,$1B
	.dc.b	$FB,$44,$FF,$82,$1B,$19,$9E,$2D,$7F,$B7,$7C,$FF,$F9,$59,$34,$8F
	.dc.b	$F9,$69,$F9,$3F,$0A,$FA,$0F,$F3,$66,$DF,$86,$AF,$AF,$5A,$9C,$ED
	.dc.b	$FE,$95,$8C,$72,$FD,$2A,$D9,$F6,$30,$E9,$19,$8F,$19,$9B,$E6,$FF
	.dc.b	$86,$2B,$9B,$F8,$03,$9E,$1F,$98,$BE,$71,$93,$9C,$66,$B9,$C7,$FA
	.dc.b	$92,$B9,$CF,$AE,$37,$87,$FA,$A2,$F9,$CE,$6B,$73,$ED,$AF,$36,$1E
	.dc.b	$99,$CC,$79,$CC,$93,$7F,$F3,$D6,$1E,$67,$FC,$F9,$1E,$AB,$F0,$63
	.dc.b	$FE,$B6,$8F,$D1,$D6,$F0,$C9,$FC,$F6,$38,$1A,$9E,$0A,$7F,$A2,$CF
	.dc.b	$F4,$36,$C7,$F1,$08,$FF,$61,$7D,$07,$F9,$69,$D7,$47,$2A,$DE,$3F
	.dc.b	$EA,$B1,$F8,$3B,$E3,$B5,$5E,$79,$CC,$71,$25,$FF,$61,$F6,$FF,$CD
	.dc.b	$FF,$07,$8C,$FC,$47,$BA,$B1,$7F,$EB,$9A,$DD,$FA,$67,$2F,$A3,$E6
	.dc.b	$6D,$FF,$AD,$74,$47,$C2,$F4,$3F,$9D,$D7,$F6,$36,$D7,$82,$1F,$3F
	.dc.b	$FE,$BD,$75,$D7,$C4,$8F,$F7,$93,$79,$62,$BD,$27,$FB,$52,$7E,$C4
	.dc.b	$F4,$E0,$D6,$9F,$ED,$38,$C1,$FD,$D7,$E8,$5F,$43,$F6,$1B,$1D,$F2
	.dc.b	$FF,$24,$3B,$9D,$11,$D7,$99,$72,$F1,$DF,$27,$DD,$77,$0F,$F6,$54
	.dc.b	$DB,$FE,$5E,$7C,$2C,$3F,$12,$FF,$A4,$8F,$F0,$51,$6F,$9F,$CD,$6B
	.dc.b	$A2,$4E,$9F,$83,$8A,$73,$5F,$1A,$B9,$63,$F0,$FE,$6B,$FC,$8D,$1C
	.dc.b	$4F,$9C,$E2,$7E,$6F,$C9,$7E,$3E,$7F,$C3,$8D,$8F,$6C,$FE,$82,$31
	.dc.b	$5F,$EE,$76,$D7,$F4,$CA,$46,$14,$7E,$0A,$B9,$71,$C8,$FA,$5E,$25
	.dc.b	$AF,$96,$FE,$BB,$E8,$5F,$41,$FB,$0F,$A0,$FF,$00,$DF,$F8,$C6,$72
	.dc.b	$FF,$8C,$E3,$37,$F8,$86,$E7,$FE,$FD,$6C,$F9,$EB,$8D,$6F,$C4,$EC
	.dc.b	$77,$F7,$ED,$75,$7F,$80,$FD,$0F,$0C,$7A,$7A,$18,$CF,$CC,$CB,$FF
	.dc.b	$23,$74,$3E,$13,$19,$FE,$72,$31,$3A,$9F,$E1,$B0,$F4,$2C,$61,$FE
	.dc.b	$23,$EF,$BD,$1A,$7F,$1B,$A7,$CB,$BE,$4B,$38,$FF,$95,$EF,$83,$FC
	.dc.b	$1E,$5F,$25,$AC,$FF,$83,$E1,$CA,$31,$B7,$F0,$7A,$DD,$F3,$CB,$C1
	.dc.b	$18,$C3,$F8,$07,$D0,$FF,$06,$AF,$DE,$36,$33,$B7,$EE,$A3,$18,$7E
	.dc.b	$E5,$B5,$FA,$86,$D7,$E2,$9F,$CC,$4E,$FC,$2F,$E7,$5D,$3F,$91,$BF
	.dc.b	$F4,$13,$FE,$1E,$D3,$38,$7E,$C8,$FE,$47,$19,$9D,$C6,$6F,$F2,$EF
	.dc.b	$E6,$8F,$E3,$07,$03,$51,$FE,$2A,$9E,$51,$FC,$5B,$5F,$E4,$35,$CA
	.dc.b	$11,$8A,$F2,$66,$FA,$1F,$B7,$E1,$49,$FA,$9F,$43,$70,$FA,$76,$8C
	.dc.b	$63,$F4,$DB,$5D,$1F,$7F,$95,$23,$DB,$DD,$70,$F5,$4D,$BF,$7E,$9F
	.dc.b	$E4,$4A,$C5,$FB,$6D,$DD,$F2,$BB,$7E,$5B,$67,$E4,$B6,$7E,$33,$70
	.dc.b	$F8,$5F,$3D,$FE,$FD,$AF,$F5,$43,$6B,$FA,$AF,$D0,$BE,$85,$F4,$3F
	.dc.b	$DB,$AB,$58,$72,$F8,$2F,$EF,$1E,$7B,$E8,$7F,$D9,$79,$F4,$9C,$C6
	.dc.b	$33,$FE,$93,$87,$2A,$FE,$FF,$39,$C3,$A4,$7F,$B2,$70,$E9,$EF,$7E
	.dc.b	$23,$E8,$7F,$13,$E3,$DA,$FC,$3B,$72,$FC,$23,$70,$F9,$71,$FC,$87
	.dc.b	$8B,$FC,$DF,$F4,$3F,$92,$3F,$27,$E0,$DC,$3F,$23,$16,$BF,$1D,$36
	.dc.b	$FF,$15,$F3,$CF,$73,$FC,$37,$5F,$E7,$A8,$C6,$1F,$CC,$63,$F9,$94
	.dc.b	$61,$C0,$70,$ED,$D5,$7E,$F7,$AB,$FC,$3B,$18,$AF,$CB,$C6,$2F,$F7
	.dc.b	$ED,$8F,$E1,$4F,$EC,$BD,$8F,$E7,$F1,$FB,$25,$F4,$39,$8D,$35,$D3
	.dc.b	$71,$FD,$AC,$F0,$9E,$5F,$43,$D4,$ED,$F3,$16,$32,$4A,$E3,$19,$7F
	.dc.b	$C1,$39,$3E,$83,$FC,$55,$F4,$2F,$8D,$F7,$C6,$1B,$87,$FA,$9B,$6F
	.dc.b	$F1,$A5,$6E,$3F,$E0,$05,$F3,$74,$6E,$35,$87,$83,$0E,$93,$C1,$B1
	.dc.b	$98,$E2,$FF,$B5,$A7,$8D,$FE,$02,$B6,$DA,$DA,$9D,$C7,$E0,$27,$6D
	.dc.b	$37,$1A,$AF,$C4,$4E,$9B,$3A,$C3,$E0,$C7,$F9,$22,$FB,$46,$1F,$86
	.dc.b	$D9,$F8,$0D,$9F,$53,$67,$8C,$FC,$13,$B1,$67,$B5,$F9,$5F,$CB,$5F
	.dc.b	$95,$B3,$C1,$F9,$07,$0B,$ED,$5C,$70,$FA,$93,$9A,$ED,$87,$A1,$A7
	.dc.b	$A0,$B1,$E8,$6C,$7A,$23,$BB,$C5,$B1,$C5,$A7,$12,$C7,$13,$EC,$3E
	.dc.b	$33,$35,$E7,$C3,$DD,$7E,$8A,$EC,$70,$2D,$FD,$BB,$E5,$19,$8F,$13
	.dc.b	$63,$C4,$D3,$C5,$1C,$5E,$71,$F8,$03,$9C,$64,$E6,$57,$9B,$63,$9C
	.dc.b	$71,$79,$BF,$58,$E6,$66,$39,$B4,$E6,$58,$E6,$F1,$79,$CF,$C3,$9E
	.dc.b	$33,$96,$C6,$63,$99,$92,$B9,$9E,$AD,$BF,$A1,$5F,$8B,$9F,$8D,$18
	.dc.b	$AF,$49,$CE,$38,$DF,$D6,$25,$71,$3E,$25,$71,$9F,$4C,$EA,$B9,$DF
	.dc.b	$87,$0F,$CB,$93,$0F,$C7,$E1,$C5,$F6,$DF,$1C,$39,$C7,$7C,$3E,$45
	.dc.b	$6D,$F7,$D6,$E7,$89,$B8,$F5,$4E,$EB,$D7,$39,$C3,$C8,$F8,$B0,$F2
	.dc.b	$5F,$E5,$9B,$5F,$19,$B3,$F0,$1B,$3E,$46,$BE,$06,$D7,$68,$FE,$7B
	.dc.b	$E5,$7F,$47,$3F,$A4,$79,$56,$FF,$A4,$45,$9F,$E8,$73,$70,$FF,$0B
	.dc.b	$EA,$F9,$1F,$EA,$1E,$DF,$F5,$16,$E3,$FD,$45,$B8,$7F,$4C,$6D,$FF
	.dc.b	$8D,$7E,$85,$7F,$E6,$9F,$A0,$7D,$0F,$FB,$FA,$73,$8F,$FB,$B6,$72
	.dc.b	$FF,$38,$C5,$3A,$7F,$CD,$33,$A7,$FD,$B3,$19,$C3,$FD,$E1,$39,$7C
	.dc.b	$58,$78,$70,$F0,$37,$0F,$ED,$5F,$43,$FD,$B7,$E8,$7F,$B7,$36,$32
	.dc.b	$7F,$6F,$8C,$B4,$FE,$E5,$C1,$FD,$D7,$9B,$FB,$B3,$63,$27,$F7,$78
	.dc.b	$CB,$4F,$77,$CF,$BE,$81,$F0,$3F,$E0,$CD,$D3,$98,$C2,$E7,$7A,$18
	.dc.b	$BE,$37,$FE,$14,$FA,$0F,$F2,$6B,$CD,$AE,$7C,$AE,$CA,$32,$FF,$87
	.dc.b	$FF,$65,$EA,$F8,$DA,$FF,$88,$1A,$EE,$F9,$F9,$9F,$E0,$38,$FB,$4B
	.dc.b	$E8,$7F,$DB,$5C,$9F,$E7,$98,$C4,$EB,$B7,$3B,$FF,$B7,$E7,$94,$AB
	.dc.b	$85,$EA,$02,$79,$BB,$AE,$57,$B8,$D7,$B1,$69,$F4,$34,$C4,$49,$F4
	.dc.b	$C6,$5F,$5C,$73,$FA,$17,$D0,$F3,$F1,$25,$7E,$66,$72,$D9,$E0,$47
	.dc.b	$E2,$FC,$35,$7D,$07,$B2,$78,$B8,$0E,$35,$F9,$53,$ED,$FA,$07,$D0
	.dc.b	$DF,$81,$98,$7E,$AE,$72,$D9,$CB,$D6,$13,$CE,$7A,$7C,$84,$AE,$4F
	.dc.b	$7F,$87,$C4,$71,$8B,$8F,$FA,$D7,$05,$7D,$1F,$A1,$FE,$19,$7E,$3F
	.dc.b	$7B,$D0,$67,$C5,$91,$FC,$06,$76,$FC,$16,$E1,$D7,$B7,$73,$9A,$B3
	.dc.b	$98,$FE,$15,$F4,$3F,$1F,$9B,$F0,$F0,$8C,$3F,$8A,$6A,$FA,$57,$D2
	.dc.b	$FE,$27,$E8,$9F,$47,$49,$5F,$C5,$E3,$FD,$A0,$DF,$13,$81,$1F,$E3
	.dc.b	$9F,$45,$F1,$1F,$E3,$DF,$46,$F5,$3F,$2C,$D8,$CC,$7E,$66,$5E,$ED
	.dc.b	$24,$F4,$7F,$3D,$F4,$6E,$9A,$49,$FE,$59,$19,$7F,$97,$7D,$13,$E2
	.dc.b	$BF,$4D,$F1,$CF,$9E,$FD,$44,$BE,$8E,$9A,$89,$7C,$A3,$2F,$F3,$9A
	.dc.b	$E4,$D9,$DB,$F9,$B8,$CB,$F9,$DA,$FE,$32,$FE,$7F,$E8,$FC,$8F,$F4
	.dc.b	$3F,$9F,$FE,$96,$DF,$4F,$F7,$93,$A8,$E1,$97,$EF,$EB,$71,$FA,$96
	.dc.b	$F4,$3A,$3F,$07,$7C,$63,$15,$FC,$32,$31,$3A,$BE,$E7,$5F,$77,$F8
	.dc.b	$67,$C8,$3E,$0F,$F8,$77,$D0,$FD,$9C,$7F,$70,$DC,$BA,$C2,$B8,$3F
	.dc.b	$D7,$DA,$F0,$E5,$E6,$76,$B9,$1B,$6B,$DB,$E0,$FF,$8F,$7D,$13,$E7
	.dc.b	$3C,$7F,$47,$2B,$FB,$64,$E5,$B3,$97,$CF,$3C,$A2,$46,$88,$7B,$3C
	.dc.b	$DF,$E5,$9F,$47,$F6,$B4,$F6,$EB,$FC,$CE,$BF,$BD,$35,$FE,$35,$3B
	.dc.b	$E0,$FD,$9B,$63,$F9,$DC,$EE,$B9,$FA,$1D,$B9,$E5,$7F,$81,$6F,$82
	.dc.b	$6D,$BE,$3F,$CB,$ED,$FD,$4F,$46,$5F,$18,$CB,$E6,$8C,$BE,$68,$FE
	.dc.b	$91,$F4,$6E,$8B,$D7,$2B,$FC,$2D,$39,$6C,$F0,$23,$FE,$1B,$FA,$37
	.dc.b	$9F,$FD,$6E,$1F,$46,$F9,$3F,$E2,$B5,$AE,$1F,$EB,$B3,$9A,$FE,$BF
	.dc.b	$8F,$77,$8D,$FF,$1C,$8C,$4E,$E5,$7C,$DD,$9F,$DA,$1C,$46,$E7,$8D
	.dc.b	$73,$C7,$F9,$3C,$F3,$AD,$CF,$9B,$1F,$54,$F1,$AD,$CF,$1C,$3D,$2E
	.dc.b	$F1,$B1,$E1,$BE,$33,$B8,$F2,$8C,$7C,$6E,$EB,$8B,$EC,$AE,$2F,$C4
	.dc.b	$BF,$0B,$4D,$E3,$E8,$79,$1D,$AF,$AB,$DE,$78,$BD,$6F,$8D,$F6,$8D
	.dc.b	$1E,$2A,$E3,$7D,$63,$81,$FD,$38,$EE,$57,$E9,$36,$7F,$04,$E7,$D0
	.dc.b	$7C,$2F,$F9,$06,$BF,$B4,$B6,$33,$3D,$4F,$BA,$E1,$FF,$24,$7D,$1F
	.dc.b	$F8,$F1,$34,$8F,$F9,$01,$F9,$7F,$1E,$FA,$1F,$E4,$3B,$7E,$26,$BF
	.dc.b	$C5,$B5,$A9,$CE,$DE,$18,$C7,$3F,$D4,$AD,$9F,$6B,$0E,$91,$98,$F1
	.dc.b	$99,$BE,$6F,$C7,$AE,$6F,$E0,$8E,$78,$7E,$6E,$F9,$C6,$4E,$71,$9A
	.dc.b	$E7,$1F,$DF,$AB,$9C,$FA,$E3,$78,$7F,$80,$AF,$9C,$E6,$B7,$3E,$DA
	.dc.b	$F3,$61,$E9,$9C,$C7,$9C,$C9,$36,$FE,$7B,$87,$99,$FE,$7F,$1E,$AB
	.dc.b	$F0,$63,$FD,$3A,$76,$DC,$32,$7F,$A8,$6B,$81,$FE,$5E,$BE,$91,$E9
	.dc.b	$E9,$E1,$39,$6F,$BA,$FA,$2F,$F1,$DD,$7F,$AA,$EF,$FA,$23,$6B,$8E
	.dc.b	$2F,$A3,$77,$F4,$C9,$87,$07,$FC,$AB,$5B,$9D,$4F,$0C,$3B,$47,$8B
	.dc.b	$D9,$FF,$5A,$47,$CF,$8D,$91,$FF,$3C,$D7,$F7,$0C,$FF,$CB,$3E,$3E
	.dc.b	$71,$FF,$30,$63,$D5,$FF,$32,$CF,$B7,$E6,$F5,$7C,$75,$47,$A7,$D7
	.dc.b	$75,$53,$E3,$E6,$FA,$1F,$F5,$4B,$FE,$CF,$9D,$B6,$78,$3E,$63,$FD
	.dc.b	$1E,$71,$C7,$39,$7F,$9D,$6B,$8B,$FE,$7A,$9E,$DC,$3A,$CF,$FC,$1B
	.dc.b	$18,$3F,$DB,$DF,$46,$FA,$3F,$A0,$D8,$F0,$5F,$F8,$5F,$3F,$F1,$E4
	.dc.b	$70,$AD,$91,$D9,$ED,$F4,$31,$5F,$E8,$9B,$E8,$F9,$6E,$E1,$FE,$14
	.dc.b	$9B,$7E,$E9,$B5,$F0,$3A,$3F,$09,$3F,$E7,$C3,$BC,$FF,$6F,$7F,$DF
	.dc.b	$06,$67,$FD,$30,$DB,$FF,$10,$B4,$ED,$E0,$74,$B9,$1E,$AF,$86,$FF
	.dc.b	$2A,$3D,$23,$EC,$C6,$2F,$FD,$49,$F2,$BF,$85,$C3,$FC,$5A,$DB,$F5
	.dc.b	$57,$E6,$E3,$15,$FE,$C8,$6A,$73,$3F,$85,$9E,$1C,$B2,$3D,$FE,$35
	.dc.b	$BF,$94,$FE,$9F,$E8,$9F,$43,$FA,$9F,$A3,$78,$B9,$8C,$33,$0F,$BB
	.dc.b	$82,$B7,$18,$BE,$6E,$A3,$33,$FE,$C3,$6E,$9F,$E4,$C7,$F7,$F1,$9B
	.dc.b	$F1,$7C,$1F,$EE,$FE,$8F,$9C,$F1,$F4,$F1,$9F,$9D,$99,$79,$3A,$1E
	.dc.b	$6E,$2F,$FD,$A5,$18,$9D,$4F,$F0,$78,$FF,$4D,$46,$0E,$5E,$8F,$EC
	.dc.b	$3E,$8F,$F6,$B1,$F9,$66,$DF,$E9,$DF,$B9,$B7,$E9,$9C,$B7,$B7,$91
	.dc.b	$CF,$2B,$7C,$E2,$DF,$39,$B5,$FE,$7C,$E4,$FF,$60,$DF,$EB,$15,$7F
	.dc.b	$02,$3F,$DE,$2E,$9A,$66,$79,$3F,$C0,$8C,$E5,$FC,$CA,$78,$4E,$A3
	.dc.b	$F9,$57,$C2,$7F,$04,$8F,$F4,$0B,$6F,$C3,$2F,$A2,$7D,$1F,$0F,$02
	.dc.b	$57,$F7,$FC,$98,$7F,$0E,$7F,$C1,$CD,$7F,$DB,$AD,$7F,$38,$7F,$93
	.dc.b	$63,$E0,$34,$F6,$F5,$7A,$5C,$65,$EE,$F1,$F0,$66,$6D,$72,$8B,$5C
	.dc.b	$A3,$07,$6E,$AE,$9D,$3D,$5B,$19,$3C,$71,$96,$9F,$A4,$F9,$FF,$F6
	.dc.b	$FB,$7D,$DF,$EB,$FF,$44,$FA,$27,$D1,$7F,$57,$BD,$E4,$D7,$FD,$5F
	.dc.b	$18,$8F,$F1,$A3,$5F,$C0,$87,$0F,$E1,$1B,$C7,$F8,$6F,$A1,$F8,$DE
	.dc.b	$07,$C7,$6E,$9F,$31,$B7,$F8,$86,$E1,$F8,$86,$CF,$F1,$19,$46,$1F
	.dc.b	$AF,$F4,$3F,$4C,$FA,$9D,$56,$BE,$CC,$59,$FC,$AC,$DB,$FF,$66,$FC
	.dc.b	$FB,$D7,$FE,$F4,$DC,$3F,$A5,$F4,$FE,$3F,$E1,$B3,$52,$7F,$1D,$5B
	.dc.b	$C7,$FC,$9C,$D7,$FC,$B8,$D7,$F9,$57,$A9,$FE,$EB,$9F,$D6,$2F,$80
	.dc.b	$FD,$23,$99,$FD,$6B,$99,$D9,$A9,$DB,$4C,$9D,$1C,$B4,$EF,$F4,$3D
	.dc.b	$8E,$FF,$31,$63,$24,$BF,$0C,$65,$FC,$34,$7E,$0B,$07,$CA,$7C,$83
	.dc.b	$FC,$5B,$E9,$3E,$13,$F6,$35,$FC,$E2,$B6,$DA,$E0,$FF,$78,$AE,$1D
	.dc.b	$D8,$C3,$E0,$FD,$07,$C4,$3E,$2F,$D0,$5B,$F9,$31,$67,$E3,$CD,$E2
	.dc.b	$FE,$35,$BA,$B1,$7F,$4A,$AD,$FF,$5B,$AD,$4F,$43,$EA,$CE,$B9,$FF
	.dc.b	$D1,$F7,$69,$36,$30,$DA,$E5,$7E,$46,$E1,$E4,$AD,$C6,$0F,$C2,$B8
	.dc.b	$9F,$83,$19,$7C,$D1,$F0,$A3,$18,$79,$1B,$87,$91,$B3,$CE,$FC,$11
	.dc.b	$83,$DA,$5B,$EE,$79,$CB,$1C,$1B,$1F,$24,$E1,$19,$3E,$7E,$1F,$06
	.dc.b	$76,$7C,$D7,$11,$B9,$E5,$3A,$C3,$D4,$DC,$3D,$4D,$C3,$D5,$5F,$18
	.dc.b	$F0,$9C,$AF,$A4,$FC,$68,$F8,$78,$78,$1E,$11,$ED,$9D,$E1,$E0,$6E
	.dc.b	$1E,$06,$E1,$E0,$AE,$87,$8E,$FE,$4B,$4F,$46,$1D,$DD,$D7,$C4,$9F
	.dc.b	$3C,$61,$F4,$1E,$48,$EE,$F4,$6C,$74,$8E,$38,$79,$DE,$31,$CC,$91
	.dc.b	$97,$E8,$47,$0C,$3F,$3B,$1F,$06,$3C,$04,$78,$C7,$C1,$8C,$4F,$32
	.dc.b	$46,$67,$78,$7C,$59,$CB,$67,$BD,$FA,$8F,$85,$5F,$1D,$F9,$F8,$7C
	.dc.b	$7B,$EA,$DA,$E3,$87,$BA,$7A,$BF,$70,$F6,$61,$FA,$3C,$3A,$47,$A2
	.dc.b	$BC,$D5,$C2,$7E,$51,$A3,$DF,$7D,$EF,$96,$1E,$7C,$39,$4E,$5F,$A8
	.dc.b	$75,$F9,$1F,$E6,$FE,$BF,$F3,$86,$C7,$F3,$C6,$E1,$FC,$E9,$B6,$9F
	.dc.b	$ED,$7F,$41,$FE,$C0,$AF,$F1,$6B,$75,$FE,$55,$E7,$7F,$B3,$63,$52
	.dc.b	$69,$BE,$1F,$F2,$ED,$F4,$CB,$FA,$5F,$CF,$7F,$4C,$F9,$EB,$3F,$D4
	.dc.b	$1B,$87,$F4,$E6,$E7,$FC,$A1,$B3,$FD,$61,$B3,$FC,$9D,$FE,$AF,$D1
	.dc.b	$FD,$63,$C7,$B5,$FD,$71,$B5,$FD,$69,$B7,$FC,$D1,$FF,$74,$7D,$23
	.dc.b	$DA,$FF,$80,$35,$FE,$4E,$8C,$5F,$F8,$ED,$B1,$FE,$12,$35,$3C,$17
	.dc.b	$CF,$BE,$7F,$A4,$6C,$CB,$87,$94,$3A,$9D,$DF,$E4,$98,$E5,$7F,$E4
	.dc.b	$BF,$23,$FA,$FD,$FF,$A4,$97,$CE,$FF,$89,$35,$FB,$93,$FE,$66,$9C
	.dc.b	$BB,$EC,$D6,$BF,$E2,$A9,$E7,$3F,$DB,$0E,$17,$A6,$9C,$48,$6C,$47
	.dc.b	$67,$0D,$BF,$1C,$6A,$FA,$3C,$3D,$2B,$4F,$9D,$A6,$22,$4F,$BE,$32
	.dc.b	$FC,$18,$E7,$F3,$8F,$9D,$E7,$E2,$4A,$FF,$5B,$CE,$5B,$3F,$E7,$55
	.dc.b	$19,$7E,$B3,$5F,$BC,$D7,$FC,$A3,$F3,$AF,$83,$F5,$AF,$6A,$98,$89
	.dc.b	$3F,$EC,$89,$D3,$67,$2F,$CB,$8E,$8E,$48,$66,$7B,$FC,$4C,$8E,$8F
	.dc.b	$8B,$DE,$C4,$73,$8B,$BF,$D2,$B5,$7D,$2F,$9D,$78,$5E,$AF,$1E,$63
	.dc.b	$FE,$DE,$9C,$B6,$7E,$44,$62,$3F,$38,$D7,$F4,$8D,$7F,$CE,$FF,$89
	.dc.b	$BE,$97,$F0,$2D,$8F,$F7,$5B,$5F,$C6,$2F,$91,$DC,$C9,$DB,$CA,$31
	.dc.b	$1C,$22,$E7,$F9,$0A,$E5,$3F,$EF,$8F,$9F,$FB,$D3,$F9,$46,$F4,$7F
	.dc.b	$A3,$1B,$3F,$EC,$EF,$9F,$7C,$FF,$E6,$3E,$79,$F3,$F4,$95,$FE,$D4
	.dc.b	$C7,$F3,$F3,$FE,$C4,$8C,$63,$FE,$98,$87,$89,$CC,$93,$A3,$FD,$BB
	.dc.b	$F3,$DF,$59,$C0,$F9,$F7,$AB,$FA,$A6,$C6,$9A,$FE,$D6,$5E,$EF,$EC
	.dc.b	$1B,$1D,$1A,$EE,$3F,$6B,$F3,$EF,$8D,$A4,$9F,$1C,$65,$F2,$FC,$F3
	.dc.b	$E1,$BF,$7B,$E9,$7D,$09,$FE,$0D,$2F,$9F,$FE,$02,$D8,$F6,$CE,$F8
	.dc.b	$3F,$83,$B7,$83,$FD,$7F,$5B,$AE,$77,$DB,$9E,$47,$26,$BE,$0F,$23
	.dc.b	$F8,$7F,$A5,$E5,$8F,$97,$F3,$2F,$4F,$F8,$9B,$63,$E7,$CE,$DF,$E3
	.dc.b	$CD,$D3,$F8,$DC,$EE,$7F,$91,$36,$3F,$8F,$CE,$F3,$FF,$6B,$B6,$87
	.dc.b	$FA,$57,$5F,$29,$E2,$F7,$3F,$92,$FC,$FB,$E7,$FF,$93,$FC,$E7,$DC
	.dc.b	$CB,$F9,$5B,$72,$FF,$71,$D6,$F1,$FC,$3B,$5F,$E6,$D3,$B7,$C2,$D5
	.dc.b	$1B,$AF,$E6,$ED,$AE,$B7,$E4,$3C,$7F,$05,$FC,$EB,$E7,$DF,$33,$E6
	.dc.b	$F9,$F9,$97,$E6,$27,$2D,$9C,$BE,$49,$E5,$12,$34,$43,$D3,$E7,$FF
	.dc.b	$48,$F9,$FF,$C2,$47,$AB,$D2,$F8,$A6,$79,$64,$7E,$A3,$7F,$EA,$AD
	.dc.b	$9C,$BF,$D6,$6B,$9A,$69,$89,$7C,$6F,$87,$DB,$FA,$7E,$9C,$BF,$DA
	.dc.b	$DA,$75,$1E,$19,$D3,$DE,$3F,$B0,$FC,$FB,$9E,$F6,$4A,$FD,$FC,$E5
	.dc.b	$B3,$C0,$8F,$F0,$4F,$9F,$7A,$2E,$7F,$C6,$FC,$FB,$E4,$3F,$1F,$6B
	.dc.b	$DD,$6A,$73,$8F,$7A,$5F,$17,$F2,$F1,$98,$C4,$73,$C3,$C0,$F0,$AE
	.dc.b	$F8,$76,$8C,$1C,$B1,$F1,$DF,$2C,$3A,$5F,$0C,$39,$CF,$6C,$39,$3B
	.dc.b	$BE,$35,$E8,$6E,$1F,$AC,$AD,$CF,$3C,$7D,$D3,$E1,$28,$C7,$C2,$F1
	.dc.b	$C7,$C3,$5C,$63,$D2,$6F,$0E,$ED,$39,$E1,$E0,$79,$61,$DB,$0E,$B1
	.dc.b	$B3,$9C,$ED,$F9,$53,$BA,$EB,$5C,$2B,$47,$CB,$AF,$F2,$83,$5F,$31
	.dc.b	$8C,$3E,$9B,$AF,$49,$F0,$FF,$E5,$18,$FC,$14,$61,$FF,$2A,$57,$F8
	.dc.b	$12,$74,$D8,$FF,$13,$7C,$FF,$F8,$9D,$34,$8E,$8F,$8F,$F0,$AF,$9D
	.dc.b	$FF,$29,$ED,$F8,$89,$FE,$43,$5F,$4A,$B5,$DB,$F5,$27,$27,$2B,$E3
	.dc.b	$BF,$F9,$A6,$F8,$46,$27,$FC,$E5,$18,$AE,$51,$E6,$39,$38,$8D,$47
	.dc.b	$DC,$3B,$E1,$F9,$CC,$3C,$71,$93,$94,$66,$FC,$11,$FA,$CA,$E7,$3E
	.dc.b	$D8,$DE,$1F,$B0,$BE,$73,$9A,$DC,$FA,$EB,$CA,$77,$8C,$BF,$D1,$AB
	.dc.b	$C7,$BF,$F9,$A9,$B8,$F1,$C3,$C8,$FF,$81,$A3,$35,$C7,$1F,$F4,$1D
	.dc.b	$FF,$A7,$30,$F8,$2F,$3E,$9C,$E1,$E7,$6D,$FF,$A4,$BE,$7B,$FC,$B9
	.dc.b	$3D,$67,$32,$C7,$FA,$0C,$FF,$86,$6D,$F3,$EE,$ED,$D6,$AF,$FD,$1D
	.dc.b	$5A,$9C,$92,$FF,$C4,$5F,$05,$FE,$6D,$F1,$71,$7F,$E8,$FC,$3A,$57
	.dc.b	$BE,$FC,$F7,$B8,$C1,$2B,$FD,$39,$0C,$DF,$2F,$D3,$2F,$F5,$3C,$E6
	.dc.b	$3D,$BE,$3F,$F3,$BC,$3F,$C6,$2D,$C3,$FD,$4D,$0F,$9D,$FF,$5B,$BA
	.dc.b	$EA,$E2,$47,$FC,$78,$6F,$0C,$57,$FA,$A6,$FA,$63,$FE,$B8,$AE,$9C
	.dc.b	$3A,$8C,$63,$F7,$2D,$F3,$CF,$9F,$FA,$6D,$C3,$BD,$7F,$8E,$5F,$F2
	.dc.b	$4C,$66,$BF,$D7,$0D,$BF,$F5,$AB,$67,$FC,$2A,$7B,$7C,$EE,$B8,$48
	.dc.b	$EA,$DB,$F8,$78,$71,$78,$F4,$FE,$1A,$BF,$D1,$E7,$93,$C0,$C4,$48
	.dc.b	$FF,$2E,$9A,$9E,$73,$97,$FD,$9D,$19,$CB,$E3,$7C,$C7,$F9,$3E,$38
	.dc.b	$DF,$E6,$14,$66,$BF,$DA,$DF,$21,$F8,$86,$F1,$EF,$2D,$DF,$F9,$BA
	.dc.b	$B5,$8F,$FB,$61,$B1,$FD,$A9,$B6,$8F,$C1,$61,$FA,$49,$CC,$9D,$CF
	.dc.b	$E1,$FC,$4D,$63,$E1,$7F,$57,$F3,$EF,$9D,$FD,$6F,$CF,$BC,$4C,$C6
	.dc.b	$19,$87,$FA,$12,$74,$FF,$A1,$AB,$93,$99,$DE,$5F,$EF,$7A,$FD,$DB
	.dc.b	$63,$BF,$C2,$FE,$F3,$E7,$F8,$CE,$FD,$1C,$65,$E4,$66,$3F,$E9,$4E
	.dc.b	$87,$99,$8B,$FF,$09,$4F,$F8,$42,$72,$EE,$7F,$84,$C7,$FA,$6A,$30
	.dc.b	$73,$F6,$7E,$D7,$CF,$FE,$DA,$3F,$52,$DB,$FD,$8B,$F7,$78,$FF,$50
	.dc.b	$E5,$BD,$BC,$8E,$79,$5B,$E7,$16,$F9,$CD,$AF,$B9,$1F,$64,$FD,$19
	.dc.b	$EC,$31,$8F,$F2,$6E,$3F,$F1,$6E,$1F,$EB,$97,$FC,$FB,$19,$9E,$35
	.dc.b	$D2,$73,$5B,$9C,$BF,$E8,$48,$CC,$77,$1F,$0B,$FC,$1A,$7F,$C7,$6D
	.dc.b	$85,$BE,$7D,$F3,$FF,$CD,$AB,$55,$CB,$C0,$FE,$4F,$5F,$EF,$3C,$3F
	.dc.b	$D9,$B5,$CF,$0E,$5F,$0B,$E9,$F2,$65,$79,$FB,$6E,$5F,$0F,$4F,$F4
	.dc.b	$DD,$63,$5F,$74,$FF,$9A,$2F,$DD,$39,$7D,$D1,$FE,$55,$7A,$1A,$3E
	.dc.b	$27,$5F,$C9,$EA,$F9,$4D,$8C,$9F,$26,$32,$D3,$FD,$CF,$F3,$9F,$EE
	.dc.b	$06,$FB,$BF,$D7,$3E,$7D,$F3,$EF,$9F,$FE,$B7,$7B,$CB,$E8,$37,$4F
	.dc.b	$C1,$F9,$2F,$A5,$FF,$5A,$67,$D0,$FF,$7C,$46,$7D,$BF,$C0,$7D,$07
	.dc.b	$E0,$7C,$4B,$5F,$7D,$B9,$7C,$93,$F8,$5F,$17,$FB,$12,$B8,$57,$E1
	.dc.b	$5B,$3B,$6E,$5F,$3B,$E8,$3F,$1C,$7E,$3F,$BA,$DF,$E3,$62,$DF,$79
	.dc.b	$B5,$F9,$E8,$7C,$FB,$D9,$C4,$7F,$A2,$23,$18,$7F,$92,$5B,$1F,$E8
	.dc.b	$C3,$FD,$B6,$E1,$C3,$AA,$FE,$DB,$D7,$F9,$28,$C5,$7E,$9A,$3F,$35
	.dc.b	$9F,$F9,$5A,$3F,$59,$EB,$3E,$02,$C6,$1A,$FE,$BE,$34,$D3,$45,$35
	.dc.b	$1C,$31,$FC,$C4,$F0,$6F,$D0,$78,$0F,$0F,$CC,$DB,$ED,$16,$78,$CD
	.dc.b	$CF,$E5,$66,$F9,$7F,$F1,$4C,$3E,$95,$F1,$7F,$E2,$8A,$D6,$5F,$E2
	.dc.b	$79,$CA,$C7,$F8,$04,$E9,$FF,$13,$CE,$9C,$E1,$DA,$31,$87,$F0,$DC
	.dc.b	$DF,$43,$F2,$8F,$97,$F4,$37,$2F,$91,$17,$1F,$87,$36,$F9,$4D,$AF
	.dc.b	$E2,$96,$AC,$5F,$F5,$D8,$EF,$3D,$0F,$A9,$3A,$DF,$FD,$23,$76,$B9
	.dc.b	$56,$EF,$E9,$CE,$6F,$4D,$9D,$61,$E4,$69,$E4,$9D,$B8,$AF,$24,$61
	.dc.b	$EB,$1E,$43,$58,$79,$1B,$87,$92,$76,$4B,$F0,$3F,$1A,$77,$87,$96
	.dc.b	$35,$1F,$1A,$31,$3C,$09,$19,$3C,$53,$C0,$F9,$53,$C2,$35,$19,$BF
	.dc.b	$33,$E1,$31,$7F,$41,$B8,$7B,$F0,$EF,$1F,$0A,$FC,$E7,$BA,$74,$F6
	.dc.b	$9F,$C4,$46,$27,$C7,$3E,$69,$E7,$7E,$26,$DF,$C7,$6E,$1E,$9B,$E2
	.dc.b	$79,$9F,$C8,$DF,$8A,$3C,$ED,$9E,$78,$78,$70,$ED,$3F,$74,$D1,$E7
	.dc.b	$31,$5D,$5F,$61,$A7,$C3,$5E,$C8,$C4,$F4,$9F,$88,$71,$3C,$65,$9F
	.dc.b	$1C,$F0,$C3,$E7,$47,$39,$D4,$F8,$DE,$71,$F0,$A3,$C7,$1A,$3C,$6F
	.dc.b	$A2,$35,$1E,$06,$9E,$07,$85,$79,$CF,$F7,$CD,$FC,$5B,$F1,$DF,$33
	.dc.b	$C5,$3B,$8C,$D7,$92,$73,$18,$9D,$57,$3C,$3E,$AD,$F8,$2B,$47,$DD
	.dc.b	$BE,$33,$F7,$6F,$B3,$67,$6D,$3C,$D7,$FB,$FF,$95,$FE,$77,$ED,$7F
	.dc.b	$3C,$6E,$1F,$CE,$5B,$8F,$F3,$46,$E5,$FD,$57,$E8,$57,$FD,$52,$BF
	.dc.b	$93,$46,$2F,$FA,$94,$62,$74,$A7,$4E,$9B,$5F,$EC,$0F,$A5,$7D,$2F
	.dc.b	$FB,$72,$73,$87,$F8,$EA,$72,$FF,$4F,$FA,$3F,$EA,$BF,$47,$FD,$59
	.dc.b	$B1,$93,$FA,$A4,$65,$A6,$71,$FE,$A3,$19,$6C,$7F,$5B,$E3,$FE,$C3
	.dc.b	$E6,$7F,$62,$6C,$64,$FE,$C1,$19,$69,$FD,$27,$3F,$EF,$9F,$42,$FA
	.dc.b	$27,$C0,$7C,$6C,$53,$98,$C2,$E8,$FF,$25,$CF,$CC,$8D,$3F,$3B,$E9
	.dc.b	$7F,$D5,$AF,$36,$B9,$F2,$FB,$28,$CB,$FE,$07,$F2,$FF,$B3,$3E,$7C
	.dc.b	$FF,$B0,$28,$7D,$1E,$34,$FF,$0B,$46,$27,$81,$FD,$9D,$E6,$FF,$99
	.dc.b	$DB,$1C,$27,$95,$76,$9E,$55,$BE,$BC,$BC,$DA,$F8,$EA,$CF,$0B,$F1
	.dc.b	$D6,$3D,$6C,$FD,$1C,$C3,$CA,$7A,$E3,$4E,$2D,$3B,$8F,$44,$F8,$3E
	.dc.b	$89,$F4,$7E,$0F,$12,$57,$49,$CB,$67,$81,$1E,$DF,$07,$EA,$FA,$27
	.dc.b	$D1,$DF,$7E,$57,$9A,$72,$D9,$CB,$FE,$91,$84,$F3,$99,$3C,$63,$7B
	.dc.b	$7F,$92,$EB,$9C,$F1,$F8,$EF,$11,$1F,$07,$C2,$D7,$47,$8E,$3D,$11
	.dc.b	$F4,$BE,$8D,$E1,$7A,$FD,$C9,$7F,$3A,$72,$D9,$E0,$47,$DD,$D5,$F7
	.dc.b	$3E,$8D,$EE,$52,$4F,$D8,$87,$CA,$E2,$3F,$D6,$51,$C1,$D4,$F3,$9D
	.dc.b	$6B,$FE,$6B,$AD,$57,$0A,$D3,$F8,$4F,$A3,$FB,$2F,$FA,$F5,$B5,$FE
	.dc.b	$73,$6F,$57,$E9,$3E,$8D,$F4,$7F,$94,$FA,$27,$D1,$D2,$63,$F9,$08
	.dc.b	$FF,$60,$B7,$C5,$E0,$47,$F2,$BF,$44,$E3,$FC,$FF,$D1,$BD,$7A,$4A
	.dc.b	$DF,$C1,$7E,$92,$CD,$24,$7E,$83,$E8,$DD,$9F,$AA,$6C,$66,$30,$7E
	.dc.b	$BB,$E8,$DF,$13,$FB,$0F,$8E,$FA,$35,$FA,$A9,$7D,$1E,$CD,$74,$49
	.dc.b	$D4,$65,$FD,$8D,$72,$7F,$DD,$4D,$E6,$FF,$50,$37,$97,$F8,$8B,$67
	.dc.b	$FC,$E5,$87,$F2,$87,$F7,$7F,$47,$E6,$7F,$7B,$F3,$DF,$D5,$1B,$EB
	.dc.b	$7F,$08,$9D,$47,$FB,$E5,$B8,$FF,$AB,$EB,$84,$FF,$BF,$5B,$1F,$EB
	.dc.b	$39,$DE,$3F,$C9,$1B,$5F,$E8,$56,$A6,$D7,$FA,$9E,$3F,$88,$FD,$1B
	.dc.b	$E1,$3F,$8D,$FD,$0F,$DD,$9F,$E2,$AD,$58,$FF,$B0,$EB,$84,$72,$AE
	.dc.b	$3C,$92,$30,$43,$8E,$5F,$CE,$AB,$8C,$60,$FF,$4E,$38,$BF,$F6,$86
	.dc.b	$0F,$85,$FE,$5F,$F4,$6F,$9D,$F3,$FD,$1C,$C7,$F9,$5C,$E5,$B3,$97
	.dc.b	$B0,$B5,$74,$99,$3C,$08,$70,$97,$9D,$FD,$03,$E8,$FF,$05,$3C,$FD
	.dc.b	$1F,$92,$67,$9A,$47,$CA,$8C,$BF,$D0,$27,$89,$27,$71,$DF,$E4,$7B
	.dc.b	$7B,$5F,$53,$D2,$93,$C6,$32,$FD,$68,$CB,$EE,$B4,$FF,$56,$FA,$37
	.dc.b	$2D,$E9,$97,$FD,$4A,$72,$D9,$E0,$47,$FA,$BF,$D1,$BC,$FF,$C1,$C3
	.dc.b	$E8,$DF,$29,$FD,$DA,$B5,$CD,$FE,$D7,$9C,$D7,$FB,$67,$1E,$EF,$1B
	.dc.b	$FE,$FB,$18,$9D,$CA,$F9,$BB,$3F,$DC,$4E,$23,$73,$C6,$B9,$E3,$FE
	.dc.b	$08,$9E,$75,$B9,$F3,$63,$EA,$9E,$35,$B9,$E3,$87,$A5,$DE,$36,$3C
	.dc.b	$37,$C6,$77,$1E,$51,$8F,$8D,$DD,$71,$7D,$95,$C5,$F8,$97,$E1,$69
	.dc.b	$BC,$7D,$0F,$23,$B5,$F5,$7B,$CF,$17,$AD,$F1,$BE,$D1,$A3,$C5,$5C
	.dc.b	$6F,$AC,$70,$3F,$CC,$EF,$D0,$7F,$C6,$26,$4A,$E5,$FC,$07,$8A,$F8
	.dc.b	$6F,$F2,$54,$FF,$49,$6C,$66,$7A,$B5,$FF,$28,$7D,$1F,$F9,$11,$34
	.dc.b	$8F,$F9,$21,$F9,$5F,$20,$FA,$17,$6F,$CD,$A1,$C8,$FA,$93,$B8,$D4
	.dc.b	$67,$7F,$C7,$CE,$63,$ED,$1C,$AF,$BD,$F6,$8C,$9C,$77,$FF,$2A,$D7
	.dc.b	$1B,$FC,$CC,$F1,$3C,$35,$C6,$33,$1C,$4F,$3C,$F1,$AF,$F4,$BD,$F8
	.dc.b	$30,$F3,$B6,$BC,$2D,$BE,$6D,$AE,$33,$E0,$AE,$2D,$38,$96,$BE,$0E
	.dc.b	$1E,$DD,$BD,$D2,$C7,$E2,$47,$FA,$B7,$1F,$13,$A8,$CE,$1F,$0A,$F9
	.dc.b	$3E,$2C,$3A,$BF,$E6,$09,$FE,$74,$DE,$9D,$E2,$F9,$77,$D1,$7F,$93
	.dc.b	$2B,$44,$9F,$E9,$19,$3E,$8F,$FA,$34,$E6,$3D,$51,$89,$FE,$83,$8D
	.dc.b	$9F,$47,$7F,$83,$1F,$F3,$0F,$89,$E2,$BF,$81,$36,$F8,$63,$F7,$A3
	.dc.b	$87,$B7,$FE,$C2,$AF,$B3,$1B,$3A,$66,$F9,$BC,$5F,$FA,$53,$2F,$6C
	.dc.b	$77,$F5,$FF,$9B,$CF,$FA,$11,$B8,$7F,$65,$87,$D0,$FC,$57,$5D,$7C
	.dc.b	$48,$FF,$A8,$0D,$E1,$8A,$FE,$C3,$7D,$36,$FC,$D5,$74,$E2,$D4,$63
	.dc.b	$1F,$B7,$7F,$AA,$FA,$37,$D1,$FC,$56,$E1,$DE,$BF,$C7,$6F,$FA,$72
	.dc.b	$33,$5F,$DC,$1B,$7F,$8C,$6C,$FE,$D8,$F6,$F9,$DD,$70,$91,$D5,$B7
	.dc.b	$F0,$F0,$E2,$F1,$E8,$FB,$15,$FE,$57,$3C,$9E,$0E,$22,$47,$FA,$C0
	.dc.b	$D4,$F3,$9C,$BF,$E0,$68,$CE,$5F,$1B,$E1,$BF,$CA,$31,$C6,$FF,$2E
	.dc.b	$A3,$35,$FE,$0E,$F9,$1F,$C3,$BF,$C4,$9B,$3F,$6A,$7F,$1D,$1A,$93
	.dc.b	$4D,$E3,$D6,$BF,$D8,$F1,$9E,$3F,$0C,$FF,$B8,$E7,$31,$8F,$17,$F3
	.dc.b	$5F,$46,$FA,$1F,$CE,$7D,$1B,$C5,$CC,$61,$98,$7F,$B5,$27,$4F,$FB
	.dc.b	$5A,$B9,$39,$9D,$E7,$FE,$2C,$AF,$DB,$36,$3B,$FC,$27,$EA,$FE,$8F
	.dc.b	$8C,$EF,$D5,$C6,$5E,$46,$63,$FE,$E4,$E8,$FF,$17,$BE,$38,$FD,$A4
	.dc.b	$62,$BF,$C7,$6F,$FA,$4E,$78,$7A,$3A,$F3,$76,$FF,$0C,$E1,$FE,$87
	.dc.b	$74,$FD,$51,$FE,$51,$73,$5F,$D2,$39,$3F,$9F,$47,$F4,$AC,$38,$57
	.dc.b	$16,$D6,$CF,$BD,$1D,$2B,$8B,$F9,$D3,$B1,$98,$DB,$CD,$E4,$FF,$58
	.dc.b	$F8,$4F,$DB,$47,$A1,$B1,$FC,$97,$E6,$9F,$45,$FC,$5E,$73,$27,$78
	.dc.b	$2B,$FD,$B7,$7D,$AB,$5D,$DF,$C8,$67,$31,$E6,$F2,$3E,$31,$F1,$FA
	.dc.b	$AD,$F9,$B4,$FE,$DB,$38,$AF,$6D,$5D,$3E,$0F,$4F,$CF,$E7,$FA,$0D
	.dc.b	$C7,$95,$FE,$74,$F8,$B3,$A8,$CB,$63,$31,$C2,$79,$47,$F9,$F3,$E7
	.dc.b	$FF,$D9,$F3,$ED,$6C,$7F,$A3,$7E,$89,$F4,$4F,$A2,$FE,$7D,$5A,$CB
	.dc.b	$FA,$FB,$70,$FE,$E7,$F4,$5F,$EA,$57,$A7,$BF,$FD,$5B,$C9,$4E,$73
	.dc.b	$FB,$BF,$45,$F8,$DF,$03,$F1,$CD,$C3,$DE,$D9,$DC,$7C,$29,$FC,$A7
	.dc.b	$D1,$48,$F8,$BD,$36,$BF,$13,$37,$4F,$C6,$CD,$AF,$C8,$CD,$AF,$E1
	.dc.b	$67,$E9,$4E,$31,$F8,$1F,$A0,$7B,$1F,$D7,$DB,$1F,$D5,$79,$DF,$0B
	.dc.b	$9A,$93,$FE,$42,$BF,$F1,$6D,$F0,$70,$D7,$FC,$8E,$D7,$F9,$17,$A9
	.dc.b	$FE,$36,$9F,$C5,$2F,$A2,$FD,$B3,$99,$FD,$A3,$99,$D9,$A9,$DB,$4E
	.dc.b	$7F,$45,$C8,$E7,$F3,$36,$32,$4B,$EB,$19,$7B,$CD,$DB,$F8,$0E,$0A
	.dc.b	$C5,$FF,$0A,$C5,$F2,$DF,$24,$FC,$27,$C2,$7E,$C6,$7E,$74,$65,$F9
	.dc.b	$19,$2B,$8C,$61,$B1,$F2,$A7,$FC,$62,$FC,$8F,$A1,$F8,$A7,$C6,$FA
	.dc.b	$1B,$7F,$2A,$2C,$FC,$89,$B9,$FA,$BE,$4F,$E6,$FB,$1F,$39,$B8,$7D
	.dc.b	$26,$DF,$D0,$6D,$71,$9F,$AF,$5C,$5A,$71,$2D,$71,$6C,$F1,$8F,$B7
	.dc.b	$3C,$5B,$3C,$5A,$71,$2C,$F1,$3E,$9C,$F1,$6C,$F1,$6C,$F1,$69,$C6
	.dc.b	$3F,$23,$3C,$5A,$71,$2C,$F1,$6C,$F1,$8F,$EE,$BA,$FF,$8F,$FC,$2F
	.dc.b	$F4,$F6,$1F,$1A,$FF,$3A,$DA,$DB,$67,$73,$F1,$0D,$96,$76,$D8,$DC
	.dc.b	$7C,$28,$DB,$4D,$96,$37,$7F,$02,$76,$D9,$DB,$67,$73,$EE,$36,$59
	.dc.b	$DB,$63,$71,$F0,$B0,$F9,$4D,$9F,$94,$D3,$E5,$16,$BE,$55,$7F,$B1
	.dc.b	$6B,$8B,$6B,$8B,$6B,$8B,$6B,$8E,$1F,$44,$61,$F3,$5B,$7F,$25,$B7
	.dc.b	$F1,$5B,$5F,$05,$B3,$ED,$6C,$F9,$E3,$E1,$B4,$F8,$65,$BF,$86,$79
	.dc.b	$63,$6D,$8D,$B4,$DC,$FA,$6F,$D0,$4A,$F4,$36,$BD,$0D,$BF,$43,$67
	.dc.b	$D0,$D9,$F2,$57,$AA,$FC,$0D,$3C,$15,$D8,$93,$D9,$B7,$D9,$A7,$62
	.dc.b	$C7,$6B,$F0,$5F,$66,$DF,$69,$E8,$4B,$E8,$DB,$E8,$D3,$A3,$DA,$FA
	.dc.b	$36,$BA,$57,$22,$47,$26,$D7,$23,$F4,$C3,$E8,$3F,$D6,$CD,$C7,$F9
	.dc.b	$BF,$9B,$FE,$BF,$9C,$92,$33,$7F,$E2,$78,$FC,$87,$CF,$FF,$56,$F9
	.dc.b	$FF,$EA,$ED,$9F,$EB,$6D,$C3,$FA,$CB,$71,$FE,$97,$CD,$FD,$7B,$C7
	.dc.b	$FE,$BE,$DB,$FE,$C6,$DB,$FE,$C2,$DA,$E3,$3F,$DA,$2B,$8B,$4F,$F1
	.dc.b	$E7,$D2,$3D,$9F,$E9,$38,$04,$66,$7F,$A9,$B6,$3F,$BE,$35,$FF,$33
	.dc.b	$8F,$9F,$7D,$07,$58,$FF,$01,$46,$2F,$FA,$EB,$FD,$9F,$CC,$B5,$BA
	.dc.b	$E7,$E5,$7F,$68,$7F,$CA,$4B,$E7,$FF,$CE,$BA,$33,$FF,$3B,$CE,$6B
	.dc.b	$7D,$1A,$9C,$B6,$3D,$D3,$B5,$5C,$DC,$DE,$D4,$F8,$DD,$DF,$83,$D0
	.dc.b	$D4,$BE,$7E,$98,$89,$5E,$A8,$CB,$EC,$8C,$BD,$0F,$2C,$69,$C1,$E2
	.dc.b	$8F,$27,$CF,$BE,$7F,$C3,$E1,$4A,$FF,$3C,$4E,$5B,$3C,$08,$FC,$CF
	.dc.b	$77,$DD,$F3,$EF,$83,$BE,$D4,$AF,$F4,$14,$E5,$B3,$97,$CB,$09,$E7
	.dc.b	$3D,$3E,$2A,$57,$27,$BF,$BF,$88,$E3,$17,$87,$E9,$43,$05,$67,$83
	.dc.b	$FD,$6B,$0B,$C5,$7E,$66,$36,$FF,$A6,$7E,$7F,$FA,$0B,$EA,$F6,$7D
	.dc.b	$E7,$DD,$8C,$C7,$CE,$6E,$3E,$26,$E1,$FE,$C4,$6C,$F7,$E8,$FC,$37
	.dc.b	$CF,$BE,$7E,$92,$7F,$D5,$50,$F9,$0C,$47,$DC,$8E,$0E,$A7,$9C,$F4
	.dc.b	$F9,$4E,$BF,$39,$2B,$9C,$65,$EB,$19,$7A,$46,$5E,$D1,$9C,$FE,$6E
	.dc.b	$0A,$FC,$F7,$CF,$3E,$7B,$F4,$1F,$3F,$F8,$9C,$7F,$22,$DC,$3E,$AD
	.dc.b	$6E,$7F,$D8,$DF,$3F,$FE,$CE,$6E,$5F,$A9,$83,$7E,$0F,$1E,$2C,$AE
	.dc.b	$2F,$EB,$FE,$7B,$EC,$C7,$E8,$3E,$7D,$EC,$FE,$CD,$B1,$FE,$D5,$6B
	.dc.b	$FA,$69,$7B,$DC,$C9,$1F,$ED,$B8,$CB,$FE,$DD,$F9,$F7,$6F,$EF,$DB
	.dc.b	$19,$8C,$1F,$C0,$FE,$79,$F1,$1F,$C1,$7D,$1F,$B9,$5F,$EE,$79,$7C
	.dc.b	$FE,$CD,$73,$C9,$D4,$65,$FF,$77,$57,$27,$F7,$ED,$E3,$FD,$1B,$7A
	.dc.b	$FF,$9B,$4A,$30,$FF,$39,$C3,$F9,$55,$7E,$AB,$D9,$F3,$CF,$FB,$EF
	.dc.b	$E6,$9E,$A7,$F1,$D6,$C7,$FB,$FE,$76,$FF,$10,$6D,$7E,$D2,$77,$3F
	.dc.b	$C5,$1B,$1D,$F4,$FE,$51,$5C,$32,$F3,$B6,$FF,$A7,$35,$5F,$FA,$F5
	.dc.b	$FE,$A5,$7F,$C0,$BA,$9F,$3F,$FC,$CF,$E7,$7E,$FC,$FF,$06,$6E,$3F
	.dc.b	$12,$B7,$7C,$A3,$17,$C7,$96,$4E,$4F,$EC,$15,$C1,$FE,$5E,$DC,$3C
	.dc.b	$55,$FD,$0E,$31,$7F,$4B,$07,$C1,$7F,$47,$F9,$F7,$CC,$F9,$FE,$7E
	.dc.b	$67,$DC,$4E,$9C,$35,$FE,$7F,$69,$E3,$1F,$D0,$A7,$59,$FF,$26,$6E
	.dc.b	$3F,$84,$6C,$FF,$7B,$6A,$B7,$9D,$FD,$77,$E7,$FF,$13,$39,$F4,$7E
	.dc.b	$71,$9E,$79,$1E,$9B,$FE,$C6,$D9,$CB,$FD,$5A,$B9,$92,$BE,$C7,$C7
	.dc.b	$F7,$F6,$7E,$D7,$A5,$2F,$E2,$5A,$75,$1D,$E7,$4F,$58,$FE,$E9,$F3
	.dc.b	$FF,$88,$8F,$98,$FE,$8D,$BD,$5A,$E6,$91,$A6,$BF,$DB,$3E,$7F,$FA
	.dc.b	$57,$0F,$F8,$87,$E7,$9F,$27,$FA,$E9,$CD,$27,$2F,$F8,$9F,$1F,$13
	.dc.b	$9A,$FD,$AC,$62,$77,$2A,$E7,$18,$6D,$70,$C3,$F7,$B3,$C2,$F9,$4F
	.dc.b	$8F,$D1,$FF,$0B,$E1,$DC,$FF,$1C,$61,$C9,$F2,$61,$CA,$BA,$47,$B5
	.dc.b	$E5,$8D,$8E,$37,$D3,$1F,$05,$F1,$9E,$66,$CF,$54,$EE,$F9,$C7,$8C
	.dc.b	$63,$E8,$78,$99,$C7,$CC,$F2,$C7,$A4,$EE,$38,$E1,$E1,$74,$71,$C3
	.dc.b	$C0,$F4,$C7,$CE,$F9,$70,$EA,$7D,$87,$FC,$AA,$D7,$CE,$62,$7E,$5B
	.dc.b	$AF,$49,$F0,$3F,$E5,$C8,$F9,$B1,$87,$15,$FE,$14,$9D,$47,$DD,$9D
	.dc.b	$38,$F9,$FC,$BF,$E6,$D6,$A6,$91,$F9,$5F,$1E,$F9,$CF,$F2,$A6,$5F
	.dc.b	$8A,$9E,$46,$A3,$3B,$79,$EF,$EB,$5F,$1E,$9F,$B1,$7C,$30,$FC,$9C
	.dc.b	$E6,$3E,$C1,$9B,$FE,$9B,$19,$9F,$04,$F9,$CF,$05,$74,$71,$1D,$30
	.dc.b	$F1,$C7,$FA,$13,$0F,$1C,$62,$3F,$AC,$C6,$2B,$93,$63,$C6,$74,$C3
	.dc.b	$C7,$5C,$63,$CC,$7F,$A5,$F0,$F3,$4F,$C1,$8F,$29,$CA,$BA,$61,$E5
	.dc.b	$C3,$F9,$95,$FC,$1D,$FF,$D4,$35,$E1,$7F,$CA,$31,$99,$F0,$9F,$CE
	.dc.b	$6F,$C2,$23,$4F,$8F,$AF,$FD,$15,$5C,$9B,$5F,$EA,$DF,$7D,$F3,$DF
	.dc.b	$E6,$78,$F1,$CE,$AF,$94,$7F,$55,$B7,$CF,$FF,$4E,$8D,$99,$71,$3F
	.dc.b	$E8,$6C,$6D,$7A,$BB,$FA,$65,$FE,$7D,$F1,$7B,$5F,$A6,$6D,$F0,$C7
	.dc.b	$F0,$31,$C3,$DA,$FF,$71,$CE,$6B,$E8,$4F,$4D,$1F,$3F,$CA,$30,$F6
	.dc.b	$F6,$3F,$9E,$63,$FD,$85,$B7,$FE,$9F,$87,$CE,$B4,$E1,$D7,$27,$83
	.dc.b	$F1,$5D,$C6,$6B,$FD,$4B,$7D,$36,$FF,$5D,$57,$4E,$4D,$46,$31,$FA
	.dc.b	$F6,$F9,$F7,$CF,$FC,$46,$E1,$FE,$05,$3C,$BE,$76,$66,$47,$F7,$B9
	.dc.b	$DD,$7A,$EF,$FC,$97,$39,$AD,$CE,$5B,$87,$8D,$FA,$73,$FE,$9E,$3F
	.dc.b	$C0,$C5,$9E,$9E,$0E,$BA,$64,$E9,$F8,$98,$27,$35,$FE,$C8,$AE,$58
	.dc.b	$7C,$4F,$9B,$FF,$2C,$C7,$13,$DC,$E2,$7F,$D9,$FF,$19,$F7,$D3,$F8
	.dc.b	$19,$FA,$F3,$F8,$A8,$D1,$38,$BC,$71,$96,$4F,$CB,$C7,$FD,$D6,$FF
	.dc.b	$88,$EB,$A4,$62,$FF,$DB,$FF,$3E,$F9,$DF,$CB,$FC,$FF,$F9,$7F,$0F
	.dc.b	$F1,$54,$E5,$FF,$16,$46,$6B,$FD,$CE,$DC,$FF,$D9,$6D,$9F,$F3,$4D
	.dc.b	$71,$AD,$F8,$9D,$8E,$FF,$09,$6B,$AB,$FA,$8F,$9F,$E1,$8F,$4F,$4F
	.dc.b	$19,$F9,$99,$7F,$E3,$FE,$7F,$F1,$7B,$E6,$8F,$D8,$C6,$32,$FF,$7E
	.dc.b	$BF,$E9,$E9,$E1,$E9,$6B,$CE,$DB,$FC,$17,$83,$FD,$03,$3F,$E5,$97
	.dc.b	$F8,$63,$6C,$6C,$FD,$63,$97,$17,$96,$C6,$4E,$51,$96,$9F,$E5,$9F
	.dc.b	$84,$FD,$CC,$FF,$A4,$83,$87,$FD,$2A,$D7,$C6,$D7,$7D,$7F,$C5,$3E
	.dc.b	$63,$FD,$D1,$5D,$63,$1F,$09,$E9,$F1,$E4,$FB,$7A,$2E,$0E,$04,$FC
	.dc.b	$0D,$3E,$0D,$EA,$73,$5F,$9A,$3F,$4D,$5F,$1E,$7E,$37,$47,$C7,$F4
	.dc.b	$2C,$FC,$B6,$D7,$78,$EC,$FE,$5E,$FA,$E1,$F4,$2F,$95,$7D,$36,$D7
	.dc.b	$13,$C4,$E2,$BC,$B0,$F9,$CF,$F6,$CD,$7F,$83,$1B,$1F,$E1,$36,$BF
	.dc.b	$45,$AF,$FB,$11,$DC,$FC,$0F,$0D,$F3,$9F,$EA,$AB,$FF,$49,$56,$9F
	.dc.b	$ED,$51,$9B,$F3,$63,$F8,$AF,$9C,$FC,$57,$6F,$E2,$DB,$7F,$8F,$6D
	.dc.b	$70,$0E,$32,$F8,$B0,$F9,$CF,$CA,$1F,$95,$E5,$B7,$F9,$98,$B3,$CE
	.dc.b	$6D,$72,$9B,$5F,$4A,$6D,$7E,$83,$E6,$5E,$9F,$CC,$8F,$F7,$3C,$62
	.dc.b	$BF,$C6,$ED,$8F,$F1,$F3,$5F,$CE,$39,$C3,$F1,$3C,$EE,$6F,$E9,$D7
	.dc.b	$D3,$0F,$CD,$73,$7E,$D7,$CF,$FC,$9C,$FE,$D5,$7C,$E6,$63,$4D,$74
	.dc.b	$DC,$FE,$8C,$F0,$9E,$5F,$39,$D4,$ED,$F3,$96,$32,$4A,$E3,$19,$79
	.dc.b	$CD,$D3,$F2,$F6,$AC,$5F,$C6,$F9,$3F,$F0,$95,$3E,$8D,$F0,$FF,$E0
	.dc.b	$8E,$05,$6A,$73,$7F,$ED,$D9,$CB,$FE,$19,$8C,$E7,$FA,$11,$3A,$7F
	.dc.b	$49,$87,$7C,$3E,$57,$CE,$CC,$3E,$57,$CE,$DB,$F8,$B3,$6F,$D7,$36
	.dc.b	$BF,$81,$68,$F9,$DF,$A9,$EA,$7D,$56,$CF,$D5,$6E,$3F,$41,$B8,$7C
	.dc.b	$66,$CF,$46,$CF,$F6,$4B,$ED,$3F,$7E,$FB,$34,$EC,$59,$EC,$DB,$EC
	.dc.b	$D3,$B1,$63,$B5,$F8,$3A,$7F,$D2,$AD,$C3,$C3,$3F,$88,$C3,$C2,$DB
	.dc.b	$F0,$B4,$F0,$97,$0F,$0B,$63,$C2,$D3,$C2,$F8,$F0,$F0,$B6,$FC,$35
	.dc.b	$F8,$CB,$F0,$B6,$7C,$2D,$BF,$0B,$4F,$09,$63,$C2,$D8,$F0,$9E,$3B
	.dc.b	$F0,$B6,$FC,$35,$F8,$0B,$F0,$B4,$F0,$96,$7C,$2D,$BF,$0B,$4F,$09
	.dc.b	$63,$C3,$1E,$3B,$F0,$B6,$FC,$33,$F8,$3B,$F0,$B6,$FC,$2D,$3C,$25
	.dc.b	$AF,$0B,$63,$C2,$D3,$C2,$F8,$EF,$C2,$DB,$F0,$C7,$FB,$62,$FC,$2D
	.dc.b	$9F,$0B,$6B,$C2,$DB,$F0,$B6,$7C,$2D,$9F,$09,$E3,$7C,$2D,$BF,$08
	.dc.b	$BF,$0B,$6B,$C2,$D9,$F0,$B6,$FC,$2D,$9F,$0B,$67,$C3,$7E,$33,$C2
	.dc.b	$58,$F0,$DF,$80,$95,$E0,$6D,$78,$1B,$7E,$06,$CF,$81,$B3,$E0,$AF
	.dc.b	$15,$F8,$1A,$78,$2B,$B1,$27,$B3,$6F,$B3,$4E,$C5,$8E,$D7,$E0,$BE
	.dc.b	$CD,$BE,$D3,$D0,$97,$D1,$B7,$D1,$A7,$47,$B5,$F4,$6D,$74,$AF,$F0
	.dc.b	$8F,$CE,$FC,$FE,$67,$D1,$3E,$8F,$FD,$8F,$39,$8C,$61,$FE,$4E,$BE
	.dc.b	$39,$FF,$B3,$A7,$39,$FF,$3F,$BE,$18,$7F,$65,$F9,$EF,$EC,$DF,$3F
	.dc.b	$72,$FE,$CC,$DB,$FE,$C2,$DC,$BF,$AB,$BF,$DB,$78,$7F,$B7,$74,$5B
	.dc.b	$FE,$DC,$DB,$FE,$D2,$DB,$FE,$BE,$D9,$FE,$8A,$D9,$FE,$74,$FF,$B9
	.dc.b	$3E,$79,$F3,$CF,$63,$FC,$BD,$5F,$E6,$C9,$CC,$62,$11,$FD,$AF,$83
	.dc.b	$0E,$A3,$8F,$D1,$BC,$6E,$10,$8C,$3C,$DA,$F0,$8D,$F8,$FF,$DC,$23
	.dc.b	$E8,$67,$FE,$7E,$50,$F9,$FC,$6B,$FE,$4C,$8C,$4F,$03,$FD,$0A,$F3
	.dc.b	$7F,$BA,$36,$38,$4F,$2A,$ED,$3C,$AB,$7D,$99,$79,$F9,$B5,$F1,$FC
	.dc.b	$FD,$A3,$66,$4F,$8C,$D9,$67,$6D,$8D,$BE,$69,$E8,$63,$1F,$F3,$4F
	.dc.b	$CF,$BE,$7F,$E6,$B6,$BE,$B3,$5F,$B6,$D7,$FD,$34,$E6,$79,$CF,$AE
	.dc.b	$31,$97,$C2,$F7,$3D,$DF,$3E,$F8,$1B,$EB,$CC,$BD,$93,$96,$CE,$5F
	.dc.b	$F1,$14,$27,$9C,$F4,$F8,$79,$5C,$9E,$FE,$EE,$23,$8C,$5D,$3F,$D1
	.dc.b	$38,$2B,$19,$7E,$16,$15,$8B,$FF,$62,$C6,$DF,$A5,$F3,$FF,$E2,$77
	.dc.b	$F3,$8D,$7F,$42,$D7,$FD,$80,$D7,$F2,$4E,$F0,$FF,$4D,$B6,$F3,$E1
	.dc.b	$FC,$8E,$8F,$C3,$7C,$FB,$D7,$A4,$C7,$F0,$90,$F8,$DC,$47,$F9,$26
	.dc.b	$38,$3A,$9E,$73,$D3,$E4,$7A,$FC,$DC,$AE,$71,$97,$AC,$65,$E9,$19
	.dc.b	$7B,$46,$74,$FF,$5A,$E0,$AF,$CF,$7C,$EB,$E7,$FF,$41,$F3,$FF,$86
	.dc.b	$C7,$FD,$E2,$D5,$3B,$7F,$41,$F3,$EF,$9F,$FD,$42,$CE,$98,$8C,$F8
	.dc.b	$B2,$B8,$BF,$AF,$F9,$DF,$B5,$1F,$E6,$DF,$9F,$7B,$3F,$B3,$6C,$7E
	.dc.b	$BD,$AF,$F9,$CE,$5F,$3F,$CC,$91,$FB,$58,$CB,$FB,$7F,$9F,$76,$FE
	.dc.b	$FD,$B1,$98,$C1,$FC,$0F,$E7,$9F,$0D,$FC,$17,$E2,$7E,$E5,$7F,$01
	.dc.b	$97,$CF,$EC,$D7,$3C,$9D,$46,$5F,$E1,$15,$C9,$FF,$46,$37,$9F,$FD
	.dc.b	$D6,$DE,$BF,$E4,$CD,$90,$C3,$F9,$F3,$FC,$4F,$E7,$FC,$EF,$F1,$5F
	.dc.b	$99,$FF,$0E,$37,$D6,$FE,$3D,$3A,$8F,$F4,$E3,$71,$FF,$7D,$D7,$09
	.dc.b	$FF,$50,$36,$3F,$DF,$F3,$BC,$7F,$9D,$36,$BC,$AD,$4D,$AF,$F6,$5C
	.dc.b	$7F,$29,$EA,$7C,$FF,$F3,$1F,$9D,$FB,$B8,$B1,$FE,$8D,$18,$7F,$A4
	.dc.b	$57,$0B,$E5,$7C,$79,$64,$60,$87,$FA,$E6,$78,$36,$7C,$1F,$0D,$FD
	.dc.b	$07,$E7,$9F,$33,$E3,$F9,$F9,$7F,$CF,$E7,$2D,$9C,$BF,$EC,$5C,$D5
	.dc.b	$CE,$64,$EC,$86,$FC,$EF,$EA,$1F,$3F,$F7,$61,$7F,$D9,$1F,$E9,$CD
	.dc.b	$7F,$DA,$51,$B3,$84,$67,$6F,$91,$3F,$86,$8C,$6D,$FD,$BE,$B8,$72
	.dc.b	$F8,$09,$5D,$BE,$3F,$C7,$EC,$FD,$3F,$88,$93,$FD,$7E,$D3,$A8,$F2
	.dc.b	$CE,$9F,$2C,$7F,$6B,$F9,$F7,$25,$E8,$95,$FD,$A6,$72,$D9,$E0,$47
	.dc.b	$FB,$77,$CF,$FF,$42,$DB,$F4,$3F,$3E,$F9,$2F,$F0,$54,$E6,$93,$97
	.dc.b	$F4,$F8,$76,$73,$5F,$E1,$18,$C4,$EE,$55,$CE,$30,$DA,$E1,$87,$F8
	.dc.b	$66,$78,$5F,$29,$F1,$FA,$0C,$7C,$F1,$86,$E3,$E4,$C3,$94,$F4,$8C
	.dc.b	$9E,$2A,$E4,$FA,$0F,$E0,$38,$F8,$DE,$47,$0C,$3C,$4E,$A7,$8E,$3E
	.dc.b	$A6,$DF,$79,$E8,$2B,$8B,$E1,$C7,$C7,$87,$78,$D6,$3E,$87,$C2,$72
	.dc.b	$8D,$1B,$AE,$67,$47,$38,$7B,$27,$FC,$84,$77,$3F,$CA,$F1,$F6,$DC
	.dc.b	$B5,$FB,$FE,$8B,$E1,$3F,$CB,$53,$FE,$09,$6C,$66,$79,$B5,$FF,$30
	.dc.b	$7C,$FF,$F9,$69,$34,$8F,$F9,$79,$FA,$7F,$26,$F9,$CF,$D6,$ED,$F8
	.dc.b	$7A,$FE,$67,$5A,$9C,$ED,$FE,$70,$8C,$72,$FD,$CA,$D9,$F4,$30,$E9
	.dc.b	$19,$8F,$19,$9B,$E6,$FC,$6A,$E6,$FD,$83,$9E,$1F,$99,$BE,$71,$93
	.dc.b	$9C,$66,$B9,$C7,$F9,$C6,$B9,$CF,$AE,$37,$87,$F9,$D6,$F9,$CE,$6B
	.dc.b	$73,$ED,$AF,$36,$1E,$99,$CC,$79,$CC,$93,$7F,$E2,$AB,$0F,$3C,$62
	.dc.b	$3D,$77,$E1,$C7,$F8,$DC,$F0,$AD,$E1,$93,$FC,$E5,$7C,$0F,$EC,$B1
	.dc.b	$A6,$F4,$72,$9F,$63,$61,$F0,$2F,$9E,$FF,$2E,$57,$F9,$F6,$FF,$A1
	.dc.b	$B6,$F8,$60,$F9,$F7,$77,$3C,$BD,$95,$A9,$CD,$F3,$AD,$61,$8B,$FE
	.dc.b	$0D,$3F,$DF,$32,$FF,$3A,$78,$7A,$B9,$3C,$4F,$F5,$14,$E9,$C5,$72
	.dc.b	$3E,$0C,$FB,$67,$D9,$F3,$FC,$A7,$51,$8A,$FF,$4D,$7A,$DF,$CF,$1B
	.dc.b	$D3,$F1,$57,$CE,$FC,$67,$24,$BF,$01,$FE,$CF,$9E,$13,$B8,$CD,$75
	.dc.b	$6C,$EB,$83,$E4,$4E,$77,$FE,$B3,$87,$82,$7E,$67,$CF,$BE,$7E,$9D
	.dc.b	$BC,$CB,$5F,$EB,$39,$D3,$67,$4D,$26,$5F,$1D,$F6,$C7,$28,$C6,$7C
	.dc.b	$9F,$6C,$FF,$A5,$4F,$F6,$01,$5D,$F8,$1F,$81,$8C,$65,$F9,$E6,$9F
	.dc.b	$ED,$F2,$CF,$F8,$76,$95,$C7,$A2,$57,$FB,$7C,$FF,$78,$79,$98,$BF
	.dc.b	$47,$C2,$FF,$95,$27,$FD,$A5,$F1,$DF,$8B,$CB,$F5,$AD,$8F,$F2,$39
	.dc.b	$F6,$A7,$F2,$D1,$9A,$FF,$25,$B5,$39,$9E,$07,$6E,$3E,$27,$DE,$F8
	.dc.b	$4B,$1E,$7D,$3F,$DF,$6F,$FB,$97,$E7,$9F,$3B,$FA,$5F,$9F,$78,$BF
	.dc.b	$A7,$9C,$D7,$FB,$AE,$72,$D9,$E0,$6E,$AC,$64,$E1,$F0,$98,$D3,$F2
	.dc.b	$77,$FB,$A7,$F2,$BF,$3F,$D9,$CF,$07,$ED,$E3,$1C,$1F,$C5,$2B,$E6
	.dc.b	$56,$B0,$FF,$7F,$74,$BC,$CF,$56,$3F,$BF,$8C,$4E,$BD,$1C,$19,$AB
	.dc.b	$1F,$E7,$8C,$7F,$90,$FC,$FF,$E7,$E3,$FC,$F7,$8A,$30,$FF,$14,$F9
	.dc.b	$BF,$F5,$74,$F1,$37,$3A,$9E,$1D,$BF,$C4,$63,$8E,$4B,$E3,$5F,$B6
	.dc.b	$9E,$06,$A3,$FD,$B5,$7D,$67,$FD,$1B,$39,$F8,$3F,$E4,$DC,$DF,$E9
	.dc.b	$0F,$16,$D7,$F8,$C6,$F8,$4F,$FA,$53,$AB,$35,$27,$35,$B3,$F8,$B5
	.dc.b	$6E,$33,$3D,$3A,$7A,$F4,$59,$E2,$DB,$E2,$7E,$38,$B3,$C2,$7D,$50
	.dc.b	$F9,$FF,$F6,$2C,$61,$B1,$FC,$9F,$E8,$DF,$46,$FA,$3F,$E7,$55,$AE
	.dc.b	$2F,$EA,$ED,$C3,$EB,$B7,$06,$1C,$AF,$8F,$BD,$2F,$E6,$78,$EF,$A2
	.dc.b	$FF,$4F,$C7,$8E,$7F,$A9,$4E,$7D,$9F,$9B,$F4,$5F,$3B,$C2,$B7,$F2
	.dc.b	$9B,$9F,$C2,$6E,$BF,$E8,$0E,$2F,$F6,$2C,$7D,$6F,$A2,$97,$F4,$BA
	.dc.b	$EE,$1F,$2A,$6E,$1C,$63,$FD,$E3,$5F,$62,$7F,$AE,$7C,$F3,$D7,$A7
	.dc.b	$F4,$DE,$A7,$C2,$62,$FE,$8D,$6D,$B5,$B9,$D6,$1F,$D2,$2B,$4F,$F9
	.dc.b	$08,$38,$7F,$C8,$AD,$7F,$4E,$D7,$F8,$BF,$A9,$FE,$C7,$BF,$C9,$AF
	.dc.b	$A2,$FC,$CB,$99,$FC,$8B,$99,$D9,$A9,$DB,$4E,$7F,$45,$C8,$E7,$F4
	.dc.b	$56,$32,$4B,$EB,$19,$7D,$DF,$2D,$F2,$8F,$BF,$F0,$7F,$BF,$AF,$8F
	.dc.b	$19,$69,$F8,$88,$D4,$F2,$8C,$B4,$FF,$00,$72,$FF,$BB,$BC,$1E,$51
	.dc.b	$EE,$FA,$1F,$82,$7C,$2F,$A1,$B7,$F1,$A2,$CF,$C4,$9B,$C7,$F8,$BD
	.dc.b	$9F,$25,$F2,$BD,$7F,$96,$DC,$7E,$7B,$70,$F9,$AD,$BF,$F1,$FF,$D0
	.dc.b	$2F,$FC,$4C,$DB,$F7,$79,$F8,$99,$1F,$BA,$9D,$61,$EE,$DB,$FA,$2B
	.dc.b	$FE,$10,$8C,$E1,$DA,$F8,$CE,$F0,$B3,$9C,$3A,$57,$D2,$9C,$C7,$F3
	.dc.b	$2B,$E6,$D9,$D3,$C0,$97,$E1,$8C,$B6,$33,$7D,$E3,$C5,$18,$3A,$15
	.dc.b	$FF,$31,$61,$EC,$9E,$04,$9D,$CF,$23,$B1,$23,$4F,$2C,$3D,$68,$61
	.dc.b	$EC,$70,$DD,$98,$78,$70,$EF,$87,$4B,$E3,$8F,$BA,$31,$87,$39,$CD
	.dc.b	$7D,$A3,$D8,$E0,$F3,$3F,$E7,$4B,$ED,$1E,$2B,$F3,$BB,$B5,$6F,$1C
	.dc.b	$E1,$DC,$F4,$61,$CA,$F8,$57,$A0,$8F,$C3,$7C,$73,$C7,$0F,$54,$6A
	.dc.b	$78,$9E,$37,$D4,$FA,$0C,$17,$1F,$8F,$87,$92,$BE,$7C,$62,$BB,$C6
	.dc.b	$30,$F3,$61,$E1,$9F,$C0,$56,$F0,$EC,$78,$8F,$1B,$9C,$7E,$7E,$1C
	.dc.b	$F1,$F0,$5F,$2C,$7A,$CE,$30,$FC,$35,$F0,$AD,$5E,$63,$F0,$F8,$75
	.dc.b	$9F,$85,$8F,$96,$F8,$C6,$30,$F0,$3E,$C3,$99,$63,$E2,$63,$E2,$8C
	.dc.b	$61,$F6,$A7,$E9,$1E,$EB,$ED,$19,$C3,$C5,$3F,$3A,$3D,$0F,$88,$C5
	.dc.b	$78,$CF,$AB,$1A,$3D,$77,$E1,$9E,$D8,$7D,$E7,$89,$0F,$9B,$1B,$9E
	.dc.b	$55,$F8,$F3,$D2,$7E,$2A,$FC,$B5,$F5,$70,$F4,$CE,$E3,$BB,$F1,$1A
	.dc.b	$7C,$4C,$7E,$84,$73,$C3,$C0,$FA,$E7,$DC,$75,$C7,$E7,$BC,$B1,$F6
	.dc.b	$DF,$28,$F5,$E3,$E4,$C7,$F7,$9D,$4D,$1F,$3F,$FD,$83,$E7,$FF,$B0
	.dc.b	$B7,$4F,$EC,$AD,$C7,$FB,$1B,$73,$FE,$81,$CF,$FD,$A3,$CB,$FE,$D2
	.dc.b	$DB,$FE,$DA,$DB,$FE,$D6,$DA,$FF,$7E,$FE,$CE,$FA,$1F,$F7,$E5,$7F
	.dc.b	$13,$8C,$5F,$F7,$F8,$46,$1F,$F0,$13,$5F,$F3,$28,$FA,$47,$B9,$D2
	.dc.b	$30,$D8,$CD,$75,$6C,$FF,$83,$7E,$85,$F4,$5C,$A3,$0D,$7B,$C3,$AD
	.dc.b	$E8,$F9,$E3,$8E,$3E,$9F,$39,$DB,$FE,$7A,$DB,$F4,$18,$A5,$F4,$39
	.dc.b	$D7,$FC,$FF,$18,$9F,$F2,$BB,$C9,$B7,$B7,$F9,$D4,$6D,$FF,$77,$B6
	.dc.b	$FB,$BC,$63,$3E,$9F,$17,$C5,$F4,$36,$4D,$99,$3E,$28,$D1,$87,$FD
	.dc.b	$FB,$F4,$0F,$A1,$E5,$E3,$49,$FF,$37,$CE,$5B,$39,$7C,$F3,$E3,$9D
	.dc.b	$47,$49,$C1,$1F,$87,$F0,$3E,$9F,$A0,$7D,$0D,$F7,$65,$7F,$9F,$A7
	.dc.b	$2D,$9C,$AD,$7E,$9C,$AB,$A5,$75,$F8,$C9,$7C,$DF,$07,$C1,$62,$39
	.dc.b	$70,$FF,$AD,$B0,$56,$71,$EB,$7F,$3F,$E8,$7F,$C4,$F8,$79,$7C,$AF
	.dc.b	$51,$9F,$22,$46,$48,$76,$3F,$BC,$4F,$27,$01,$C5,$FE,$39,$B9,$7F
	.dc.b	$B2,$9B,$3F,$E0,$1C,$BF,$39,$5F,$94,$FA,$17,$BD,$F8,$56,$D7,$FA
	.dc.b	$A1,$AF,$81,$7C,$96,$E6,$4E,$DE,$51,$CF,$E5,$BA,$7C,$EC,$AE,$91
	.dc.b	$97,$B4,$65,$ED,$1F,$97,$EE,$FE,$11,$1F,$EB,$AF,$C5,$1F,$6B,$F9
	.dc.b	$B6,$C7,$FA,$FA,$76,$FC,$66,$E7,$FB,$29,$DC,$FC,$A6,$C7,$ED,$A7
	.dc.b	$7B,$7F,$BC,$9B,$5F,$E2,$96,$BF,$C0,$FC,$4F,$17,$6F,$EB,$7E,$85
	.dc.b	$F4,$7F,$AF,$FA,$3F,$CE,$E5,$FC,$09,$B9,$7F,$BF,$AB,$78,$FF,$B5
	.dc.b	$BE,$8D,$F4,$7F,$B7,$5C,$14,$C4,$67,$CA,$95,$C5,$FD,$F7,$D1,$3A
	.dc.b	$BF,$80,$7D,$1B,$DE,$A4,$BF,$F7,$2D,$3E,$0B,$89,$23,$2F,$F0,$9F
	.dc.b	$A3,$78,$54,$93,$FE,$EE,$8C,$BF,$EF,$0F,$A2,$7C,$6F,$F1,$1F,$67
	.dc.b	$F1,$33,$9F,$A3,$FE,$29,$47,$5D,$B2,$74,$43,$F4,$35,$C9,$B1,$B6
	.dc.b	$9E,$1F,$3B,$F9,$07,$B9,$E2,$8F,$F3,$17,$47,$F3,$4F,$9F,$FF,$49
	.dc.b	$46,$3A,$3F,$CC,$CA,$32,$FE,$C2,$B9,$37,$4F,$E5,$93,$AB,$E9,$EE
	.dc.b	$F8,$26,$57,$53,$FC,$E5,$7C,$AB,$FA,$2B,$70,$F0,$E1,$FE,$D7,$AE
	.dc.b	$03,$C1,$7C,$3F,$F3,$BF,$A1,$FB,$93,$A6,$BF,$C1,$A7,$6F,$26,$BF
	.dc.b	$C2,$A3,$78,$FF,$56,$6E,$1F,$0A,$B7,$3F,$D6,$9B,$19,$BF,$E9,$91
	.dc.b	$8B,$FF,$47,$5B,$E1,$FF,$A8,$7D,$1B,$E7,$3D,$1F,$47,$2F,$DB,$39
	.dc.b	$6C,$E5,$F3,$4F,$28,$91,$A2,$1E,$BF,$57,$FA,$F7,$D1,$FE,$17,$5F
	.dc.b	$67,$AF,$F1,$CC,$F2,$4B,$F9,$11,$97,$E7,$D7,$1D,$FF,$06,$D8,$FF
	.dc.b	$55,$35,$37,$57,$CA,$78,$BD,$FF,$AD,$EC,$4B,$EB,$19,$7E,$C4,$65
	.dc.b	$FE,$7B,$69,$FE,$F1,$F4,$7F,$93,$9F,$C1,$AE,$9C,$9A,$E5,$93,$A6
	.dc.b	$BF,$D2,$3E,$8F,$FA,$BE,$FF,$EC,$3F,$A3,$7C,$BF,$E8,$27,$34,$9C
	.dc.b	$BF,$EC,$9C,$7C,$4E,$6B,$F4,$F1,$89,$DC,$AB,$9C,$61,$B5,$C3,$0F
	.dc.b	$D7,$CF,$0B,$E5,$3E,$3F,$63,$FB,$C6,$1D,$CF,$F6,$D6,$1C,$9F,$26
	.dc.b	$1C,$AB,$A4,$7B,$5E,$58,$D8,$E3,$7D,$31,$F0,$5F,$19,$E6,$6C,$F5
	.dc.b	$4E,$EF,$9C,$78,$C6,$3E,$87,$89,$9C,$7C,$CF,$2C,$7A,$4E,$E3,$8E
	.dc.b	$1E,$17,$47,$1C,$3C,$0F,$4C,$7C,$EF,$97,$0E,$A7,$D8,$7F,$CC,$AD
	.dc.b	$7C,$E6,$27,$E5,$BA,$F1,$BF,$D0,$CF,$F9,$B0,$FB,$0E,$1D,$7C,$37
	.dc.b	$F9,$D6,$BF,$A6,$3F,$E7,$78,$0F,$A3,$C3,$4D,$38,$23,$F2,$7E,$29
	.dc.b	$F4,$1F,$EC,$5C,$BF,$13,$3D,$2B,$E9,$56,$B9,$7E,$A4,$E4,$E5,$7C
	.dc.b	$77,$F5,$5F,$08,$C4,$FF,$A2,$23,$15,$CA,$3C,$D8,$7E,$63,$0E,$D1
	.dc.b	$F7,$4E,$D1,$9B,$E7,$1F,$D1,$AF,$9B,$B3,$9E,$1F,$D2,$AF,$9D,$FB
	.dc.b	$27,$27,$C8,$8C,$CF,$A2,$3D,$78,$79,$6F,$AD,$F1,$6D,$6E,$3C,$93
	.dc.b	$E2,$DF,$FD,$1F,$87,$A9,$FF,$57,$DE,$67,$C9,$8F,$FB,$CE,$76,$D8
	.dc.b	$E9,$86,$9F,$EA,$37,$FC,$E5,$BD,$3B,$CB,$FD,$47,$F4,$7F,$E6,$B9
	.dc.b	$C9,$23,$FC,$B5,$8B,$E8,$FF,$A2,$4E,$63,$C5,$18,$9F,$F2,$EE,$16
	.dc.b	$7D,$7D,$DD,$FE,$1B,$FD,$03,$E2,$78,$6D,$7B,$C7,$E8,$5E,$63,$13
	.dc.b	$FE,$B8,$F8,$05,$E6,$A5,$FF,$75,$C1,$9B,$E8,$F3,$8F,$F9,$CE,$73
	.dc.b	$1E,$1F,$1F,$F9,$B5,$FF,$76,$6D,$7F,$9E,$61,$F4,$3F,$EC,$87,$5D
	.dc.b	$7C,$48,$FF,$81,$4D,$E1,$8A,$FF,$3F,$DF,$4D,$95,$FB,$03,$AF,$16
	.dc.b	$E3,$35,$AC,$7F,$05,$5F,$E1,$CF,$A3,$7D,$1F,$C9,$6C,$78,$67,$FC
	.dc.b	$7A,$FF,$84,$23,$35,$FE,$90,$6D,$FF,$B7,$1B,$3E,$3A,$E3,$5B,$BE
	.dc.b	$C7,$E1,$E7,$E1,$B8,$7C,$0F,$DB,$7F,$CB,$07,$F8,$80,$B1,$F2,$E7
	.dc.b	$FB,$6C,$67,$A7,$81,$1F,$57,$81,$6B,$FD,$4F,$1C,$B1,$F8,$AB,$E7
	.dc.b	$FF,$CA,$AF,$33,$E9,$38,$9F,$F5,$7F,$C8,$FE,$29,$3F,$8C,$9F,$BB
	.dc.b	$3F,$93,$8D,$13,$8B,$E0,$65,$DE,$BF,$C7,$73,$97,$E8,$78,$9C,$6B
	.dc.b	$6F,$E7,$7E,$8D,$F4,$7F,$9F,$FA,$2F,$F0,$05,$FF,$91,$27,$2F,$F9
	.dc.b	$1A,$33,$9F,$FB,$19,$BA,$7E,$12,$11,$89,$FF,$2F,$D7,$29,$E1,$E2
	.dc.b	$77,$3C,$1F,$15,$6B,$B3,$FB,$1F,$A3,$E1,$8F,$5C,$7E,$43,$DE,$33
	.dc.b	$5F,$E6,$C9,$CB,$FE,$6E,$E6,$79,$7F,$BB,$6C,$7A,$FD,$1C,$4E,$7D
	.dc.b	$8F,$56,$9C,$5A,$FF,$52,$BF,$F4,$F5,$FF,$5B,$84,$61,$FE,$BD,$F1
	.dc.b	$7F,$AB,$9F,$F3,$FB,$63,$7F,$39,$FE,$B2,$9F,$F3,$8C,$E6,$4F,$EF
	.dc.b	$9F,$F4,$39,$8E,$CF,$1B,$67,$C5,$3B,$8E,$F3,$FC,$6E,$BF,$DE,$35
	.dc.b	$DA,$31,$F1,$1F,$CB,$79,$7F,$B6,$F8,$B7,$0F,$F1,$BD,$FF,$A2,$E7
	.dc.b	$FB,$8F,$4E,$6A,$4E,$6B,$63,$87,$F8,$2D,$F0,$9D,$57,$5E,$9E,$DC
	.dc.b	$F6,$79,$37,$0E,$51,$FE,$00,$BF,$F0,$0C,$E7,$0E,$B4,$8F,$E6,$DF
	.dc.b	$45,$FE,$CD,$BF,$F4,$DB,$67,$98,$FA,$27,$D1,$BE,$8F,$8F,$BD,$30
	.dc.b	$FB,$98,$71,$C3,$E8,$B5,$FF,$5C,$B5,$FF,$0A,$BD,$F2,$FE,$4B,$E4
	.dc.b	$3E,$8F,$FD,$57,$7E,$4C,$3F,$AE,$4E,$7D,$AF,$A1,$F4,$7F,$47,$C1
	.dc.b	$B7,$F3,$5B,$9F,$AB,$83,$FC,$47,$18,$CF,$E1,$BF,$67,$E8,$E5,$FD
	.dc.b	$5E,$AB,$87,$11,$38,$BF,$97,$36,$BF,$DA,$BF,$46,$F6,$3F,$B1,$36
	.dc.b	$3F,$AD,$75,$3E,$17,$35,$27,$F4,$95,$BC,$BF,$C8,$0D,$7F,$36,$D7
	.dc.b	$BC,$FE,$5F,$AF,$94,$7F,$63,$C3,$F5,$1A,$BD,$4F,$EA,$B8,$FE,$55
	.dc.b	$7D,$1F,$E6,$DC,$DF,$C7,$73,$3B,$35,$3B,$69,$CF,$E8,$F9,$1C,$FE
	.dc.b	$42,$C6,$49,$7D,$63,$2F,$E3,$86,$8F,$8F,$F9,$27,$E0,$3E,$0F,$F7
	.dc.b	$B7,$FE,$F0,$9D,$36,$7F,$7B,$18,$AF,$93,$18,$24,$FF,$BF,$F8,$5F
	.dc.b	$4D,$F4,$E3,$91,$97,$15,$F0,$FE,$93,$E3,$1F,$1F,$E9,$2D,$7C,$B8
	.dc.b	$B3,$F2,$66,$DF,$EF,$36,$7C,$CF,$CF,$F6,$3E,$83,$71,$FA,$8D,$C3
	.dc.b	$E9,$36,$FF,$A3,$7D,$27,$F8,$9A,$BF,$CF,$B1,$8B,$F8,$DE,$7F,$F8
	.dc.b	$C6,$B5,$87,$CD,$6D,$6E,$F8,$9F,$E5,$FD,$FF,$A8,$57,$FA,$22,$B7
	.dc.b	$3F,$E8,$A8,$46,$1D,$61,$E0,$9E,$71,$86,$BC,$23,$C8,$6D,$F2,$38
	.dc.b	$8D,$CF,$5E,$D5,$F9,$2F,$95,$6A,$F8,$47,$F1,$6A,$DE,$38,$C7,$DD
	.dc.b	$3D,$AF,$CB,$5A,$79,$12,$34,$73,$C3,$C1,$8F,$BE,$3C,$26,$AF,$F8
	.dc.b	$BC,$EB,$1F,$05,$FC,$7B,$F5,$4E,$63,$B5,$79,$0E,$38,$F0,$C3,$E4
	.dc.b	$1D,$67,$EA,$46,$63,$C3,$3C,$67,$27,$43,$C1,$87,$3C,$7D,$D5,$CC
	.dc.b	$EB,$5C,$5F,$F3,$AE,$1E,$57,$47,$6B,$EA,$E3,$1F,$84,$7B,$31,$F2
	.dc.b	$38,$C7,$EE,$61,$D6,$B9,$1E,$0B,$E7,$19,$AE,$11,$F5,$30,$F1,$4E
	.dc.b	$6F,$9C,$FB,$0F,$75,$F3,$3D,$38,$78,$A7,$E0,$5F,$2A,$E1,$5F,$07
	.dc.b	$1F,$87,$87,$59,$FF,$09,$63,$FA,$6A,$FF,$03,$B5,$E2,$63,$0F,$F0
	.dc.b	$4C,$66,$3E,$B9,$99,$E7,$3F,$EA,$3F,$A6,$7C,$B7,$DD,$F3,$11,$8A
	.dc.b	$F5,$C6,$27,$53,$9B,$FB,$13,$9A,$DC,$E7,$1F,$B5,$39,$AD,$CE,$6B
	.dc.b	$E0,$D7,$FA,$42,$B4,$F0,$9D,$57,$CB,$9C,$D7,$03,$FC,$71,$3A,$9F
	.dc.b	$B7,$3A,$78,$46,$71,$FE,$09,$39,$9D,$CE,$71,$F8,$98,$FB,$89,$87
	.dc.b	$B1,$B5,$EA,$6E,$3E,$76,$CF,$96,$7F,$C9,$37,$D6,$3F,$D8,$CD,$8F
	.dc.b	$F6,$23,$6B,$E1,$36,$7D,$B5,$D2,$BD,$6D,$7D,$2D,$C7,$DB,$5C,$67
	.dc.b	$C3,$39,$C3,$DB,$7E,$B2,$4F,$AD,$B8,$7A,$DA,$7A,$F1,$F9,$18,$F6
	.dc.b	$C3,$C9,$CF,$FD,$E7,$E9,$3F,$BB,$7D,$2F,$F7,$76,$E1,$FD,$ED,$B9
	.dc.b	$7F,$79,$6E,$5F,$D4,$7B,$7F,$BF,$76,$7F,$7F,$6D,$FF,$80,$DB,$7F
	.dc.b	$E0,$16,$D7,$FA,$57,$E9,$5F,$53,$FE,$6A,$6D,$FF,$9E,$DB,$7F,$E7
	.dc.b	$00,$E1,$FF,$39,$FD,$23,$E9,$5E,$FF,$FA,$0E,$30,$DE,$EF,$F4,$2D
	.dc.b	$73,$9C,$BD,$7E,$9F,$FD,$8C,$FF,$A2,$DA,$E7,$D0,$FF,$47,$28,$CB
	.dc.b	$FE,$84,$F4,$3F,$C1,$CF,$F5,$65,$F4,$DF,$D7,$F6,$FF,$5A,$46,$27
	.dc.b	$5E,$0E,$76,$FE,$C9,$3C,$A7,$FC,$34,$ED,$B7,$B7,$FC,$E7,$55,$FF
	.dc.b	$39,$7A,$D6,$9F,$4D,$4C,$44,$9F,$34,$65,$F4,$47,$3F,$A6,$7D,$37
	.dc.b	$3F,$2A,$57,$F7,$29,$CB,$67,$E2,$C6,$23,$EC,$35,$FF,$54,$B5,$FF
	.dc.b	$01,$3B,$D3,$FD,$B0,$DD,$33,$5E,$BF,$99,$F7,$7D,$2B,$E2,$6F,$C0
	.dc.b	$4C,$BF,$C0,$B3,$96,$CE,$5F,$F2,$24,$27,$9C,$F4,$F8,$F9,$5C,$9E
	.dc.b	$FF,$0B,$88,$E3,$17,$4F,$F7,$76,$0A,$FA,$5F,$4D,$FE,$34,$BF,$1F
	.dc.b	$C0,$7A,$0C,$F9,$52,$3F,$DB,$13,$B7,$F3,$CD,$7F,$D9,$ED,$7F,$C5
	.dc.b	$0E,$F0,$FF,$7F,$37,$6C,$D7,$C6,$EA,$FC,$27,$D3,$3E,$06,$93,$1F
	.dc.b	$F1,$5C,$3E,$5B,$11,$FE,$69,$8E,$0E,$A7,$9C,$F4,$F9,$9E,$BF,$41
	.dc.b	$2B,$9C,$65,$EB,$19,$7A,$47,$E5,$FB,$FF,$AD,$C7,$F8,$F3,$E9,$5F
	.dc.b	$4D,$F9,$B6,$C7,$F9,$02,$76,$FF,$9F,$5A,$FF,$BD,$1A,$FF,$93,$5A
	.dc.b	$83,$86,$F9,$5E,$32,$3E,$3E,$EF,$D6,$7D,$23,$E9,$BF,$80,$B6,$7F
	.dc.b	$8A,$07,$0F,$F1,$76,$BF,$E5,$F1,$F4,$AF,$A5,$FE,$16,$D8,$E6,$D7
	.dc.b	$F9,$03,$5E,$4F,$FA,$FD,$FF,$31,$B5,$FF,$50,$37,$95,$C9,$FE,$6F
	.dc.b	$FA,$4F,$CB,$57,$F9,$AB,$E9,$5F,$01,$FC,$25,$B1,$98,$FE,$19,$2F
	.dc.b	$85,$FE,$1D,$F4,$AF,$07,$31,$2F,$FC,$EB,$39,$6C,$F0,$23,$FE,$7A
	.dc.b	$FA,$57,$C7,$FF,$19,$F7,$3F,$25,$39,$FA,$5F,$E3,$94,$75,$D5,$27
	.dc.b	$44,$3F,$D8,$15,$C9,$B1,$B6,$9E,$1E,$DF,$E5,$1E,$EF,$8A,$3F,$D1
	.dc.b	$BF,$40,$F7,$3F,$96,$36,$3F,$D2,$33,$B7,$FD,$9C,$DE,$3F,$9F,$1F
	.dc.b	$8F,$9E,$0F,$FB,$4D,$B1,$A8,$FF,$49,$E9,$E2,$23,$D9,$F0,$FC,$2F
	.dc.b	$F3,$DE,$E7,$C5,$7F,$3F,$FA,$3F,$AF,$7F,$ED,$F6,$E1,$D6,$B7,$3F
	.dc.b	$EA,$86,$BB,$E5,$E4,$75,$B9,$1B,$6B,$D7,$E2,$BF,$A7,$7D,$2B,$E7
	.dc.b	$FC,$5F,$4B,$2B,$FD,$6B,$39,$6C,$E5,$F3,$4F,$28,$91,$A2,$1E,$BF
	.dc.b	$4F,$FA,$E7,$D2,$FD,$BD,$3E,$46,$AF,$5B,$E1,$1A,$E5,$98,$7F,$B1
	.dc.b	$A7,$4D,$9C,$AA,$FA,$57,$D7,$BB,$FD,$A8,$D9,$FE,$10,$D5,$6F,$96
	.dc.b	$F2,$FC,$07,$D3,$F5,$E5,$7F,$B5,$AD,$3A,$8F,$24,$E9,$F1,$61,$E0
	.dc.b	$9F,$EF,$5F,$4A,$E7,$BD,$72,$FF,$DC,$33,$96,$CF,$02,$3F,$EE,$6F
	.dc.b	$A5,$7A,$BE,$F8,$7D,$2B,$E6,$7F,$DE,$55,$AE,$0F,$E4,$D3,$9A,$FE
	.dc.b	$51,$8F,$77,$8D,$FF,$BE,$63,$13,$B9,$5F,$37,$67,$F3,$17,$11,$B9
	.dc.b	$E3,$5C,$F1,$FE,$6D,$7C,$0E,$78,$FA,$6F,$B6,$1C,$AF,$85,$7A,$1A
	.dc.b	$F1,$8D,$3E,$1A,$E5,$3A,$AF,$15,$F8,$1E,$B3,$AA,$E1,$8F,$57,$E1
	.dc.b	$D6,$EF,$91,$F2,$AB,$84,$FA,$2B,$82,$C7,$DE,$F2,$33,$8F,$99,$E0
	.dc.b	$7C,$18,$DB,$C7,$0F,$03,$E4,$3D,$D7,$F7,$5F,$E9,$D5,$F4,$5B,$19
	.dc.b	$7C,$D7,$FD,$62,$7F,$C3,$1F,$19,$7D,$73,$49,$5F,$E6,$0F,$A5,$FF
	.dc.b	$30,$A6,$91,$D1,$F6,$7E,$35,$F4,$5F,$C9,$36,$FC,$34,$FF,$73,$AF
	.dc.b	$8B,$5A,$E1,$F8,$F3,$93,$95,$F1,$DF,$FD,$05,$7C,$23,$13,$FD,$B6
	.dc.b	$31,$5C,$A3,$CC,$72,$71,$1A,$8F,$C7,$1D,$F0,$FC,$D6,$1E,$38,$C9
	.dc.b	$CA,$33,$7E,$08,$FC,$BD,$73,$9F,$6C,$6F,$0F,$CD,$DF,$39,$CD,$6E
	.dc.b	$7D,$75,$E5,$3B,$C6,$5F,$F0,$C5,$78,$F4,$FE,$90,$DC,$78,$E1,$E4
	.dc.b	$7F,$C3,$D3,$AC,$7F,$A9,$B6,$32,$70,$C7,$CB,$3A,$8F,$3F,$4D,$C7
	.dc.b	$C2,$DF,$8D,$7D,$27,$F9,$A6,$B9,$D6,$9A,$B3,$FE,$8D,$1A,$C9,$F4
	.dc.b	$AE,$FE,$D7,$BC,$BC,$15,$A9,$CE,$1B,$C3,$AB,$6F,$F9,$48,$CB,$FD
	.dc.b	$09,$E3,$70,$B9,$7C,$1F,$57,$BF,$FD,$F6,$30,$4B,$E7,$7E,$EF,$A5
	.dc.b	$E3,$18,$D5,$1E,$8F,$27,$F9,$ED,$FC,$06,$D7,$92,$1F,$47,$E5,$75
	.dc.b	$D7,$C4,$8F,$CA,$37,$8E,$2B,$DA,$7F,$B7,$27,$EA,$4F,$4E,$1D,$6F
	.dc.b	$FE,$10,$8C,$1F,$4B,$E9,$5F,$4B,$F3,$9B,$1D,$F2,$FF,$1D,$3B,$9D
	.dc.b	$11,$D7,$9F,$72,$F1,$DF,$27,$DD,$77,$0F,$F0,$E4,$DB,$FF,$38,$1F
	.dc.b	$0B,$0F,$BE,$FF,$A9,$CF,$C3,$96,$F9,$F8,$3A,$EA,$93,$A7,$E0,$E3
	.dc.b	$FD,$C6,$30,$4A,$FA,$1F,$15,$FE,$4C,$8C,$65,$F1,$63,$13,$F3,$3E
	.dc.b	$4B,$F1,$53,$FE,$33,$6D,$7A,$2D,$7F,$96,$9C,$DF,$FA,$2D,$A9,$CD
	.dc.b	$7E,$02,$78,$72,$48,$FA,$1E,$2D,$BF,$92,$FE,$8B,$E9,$1F,$4B,$FA
	.dc.b	$4F,$A4,$78,$B9,$8C,$33,$0F,$D7,$CE,$9F,$D8,$D7,$27,$33,$CF,$E2
	.dc.b	$64,$60,$E2,$F7,$F4,$A4,$7E,$EB,$E9,$79,$67,$F9,$2C,$7E,$4F,$07
	.dc.b	$F9,$67,$A1,$E6,$FE,$E9,$B8,$79,$31,$C4,$E7,$D1,$FD,$FF,$D7,$BD
	.dc.b	$4C,$5F,$F0,$88,$C3,$FC,$2B,$E2,$BF,$57,$97,$F9,$9D,$B1,$FE,$BB
	.dc.b	$C9,$F3,$FF,$EA,$B8,$E1,$DD,$FC,$32,$B7,$B7,$F0,$5A,$FF,$71,$56
	.dc.b	$8F,$E3,$37,$CB,$E2,$38,$61,$FC,$7E,$B7,$7F,$B2,$C3,$9D,$EF,$0F
	.dc.b	$E4,$3D,$1F,$EC,$DF,$23,$F9,$43,$63,$37,$F9,$73,$F8,$BC,$65,$A7
	.dc.b	$F8,$CE,$37,$3C,$A3,$86,$7F,$E8,$CA,$E0,$D8,$FC,$AD,$76,$E7,$E1
	.dc.b	$D3,$C5,$B1,$93,$C7,$19,$7F,$9B,$FD,$1F,$FA,$F5,$BE,$EF,$F4,$1F
	.dc.b	$A6,$7D,$33,$E9,$BF,$9B,$DE,$F2,$FE,$2A,$DC,$FF,$D4,$3E,$5B,$E9
	.dc.b	$BF,$D2,$F9,$F4,$3F,$A3,$C6,$7D,$CE,$7A,$7F,$4F,$8C,$4E,$F1,$FE
	.dc.b	$7B,$18,$7E,$3F,$D3,$7C,$5F,$0B,$E3,$36,$BE,$7B,$67,$71,$F4,$A7
	.dc.b	$E6,$7D,$34,$8F,$AB,$D7,$6B,$E9,$CD,$D3,$E7,$4D,$AF,$F2,$AF,$D2
	.dc.b	$BD,$AF,$EB,$2D,$8F,$C5,$F4,$BE,$7B,$35,$27,$3D,$77,$3F,$E9,$F7
	.dc.b	$F9,$65,$87,$37,$33,$BC,$BF,$D9,$BD,$5F,$A0,$8F,$C7,$FA,$FF,$EF
	.dc.b	$EC,$7F,$26,$FE,$AA,$FF,$1E,$E5,$BD,$FC,$27,$8B,$5E,$2F,$4F,$A6
	.dc.b	$91,$D3,$E3,$ED,$74,$9B,$97,$E5,$70,$56,$2F,$E3,$FC,$77,$C6,$3E
	.dc.b	$F7,$C2,$BC,$1C,$4F,$F8,$23,$65,$7F,$80,$BE,$9B,$E5,$4E,$63,$1C
	.dc.b	$5C,$35,$FF,$00,$E8,$BD,$C7,$F6,$8C,$39,$E1,$FE,$11,$9C,$D7,$9F
	.dc.b	$E9,$BE,$49,$F2,$BE,$9A,$D7,$CA,$8B,$87,$CA,$9B,$5E,$FF,$97,$FA
	.dc.b	$3E,$DF,$D2,$6E,$1F,$49,$B7,$F4,$9B,$5F,$F0,$07,$E9,$7F,$C2,$53
	.dc.b	$FE,$10,$FA,$67,$CA,$FF,$89,$27,$35,$F8,$09,$CE,$1F,$E2,$E9,$CB
	.dc.b	$FC,$8E,$33,$7F,$E0,$89,$CB,$FE,$25,$8C,$D7,$87,$1F,$F1,$E4,$62
	.dc.b	$FF,$D6,$37,$C3,$9D,$39,$C0,$27,$51,$98,$C5,$F4,$C7,$D4,$DC,$3D
	.dc.b	$0D,$9F,$33,$71,$F2,$4F,$F3,$4B,$E9,$87,$6C,$39,$CE,$5B,$87,$C9
	.dc.b	$6D,$FC,$76,$CF,$C5,$6C,$FC,$36,$E3,$EE,$6E,$1E,$0C,$7C,$97,$F1
	.dc.b	$1B,$7F,$0B,$1F,$24,$FB,$0F,$21,$63,$C8,$DC,$7C,$8D,$3C,$85,$8F
	.dc.b	$23,$71,$F2,$37,$0F,$23,$DA,$7E,$4B,$6B,$E4,$C7,$80,$C9,$5C,$9E
	.dc.b	$27,$2D,$33,$1E,$48,$CB,$63,$31,$E6,$32,$57,$35,$E8,$8C,$B6,$33
	.dc.b	$5E,$A9,$F2,$57,$CD,$3D,$45,$72,$7C,$FC,$3E,$8B,$F8,$E8,$DB,$96
	.dc.b	$B9,$38,$96,$BD,$D1,$D5,$B1,$C4,$DE,$1F,$71,$F4,$DF,$9E,$77,$1C
	.dc.b	$EB,$E2,$1C,$CA,$F3,$8D,$D7,$37,$E0,$E3,$F8,$E2,$57,$DD,$6D,$78
	.dc.b	$1B,$3C,$27,$44,$9D,$36,$35,$18,$23,$82,$47,$9E,$3E,$C3,$71,$FA
	.dc.b	$CD,$C3,$E0,$3E,$C9,$F8,$07,$03,$DC,$E7,$0F,$73,$C7,$0F,$4B,$4F
	.dc.b	$31,$67,$CA,$DC,$3C,$B1,$D0,$EC,$78,$30,$F2,$B9,$6B,$9C,$3C,$EF
	.dc.b	$0C,$3C,$6E,$E3,$33,$CB,$C8,$72,$BE,$97,$FB,$5F,$D3,$7F,$6C,$6E
	.dc.b	$7F,$DB,$1B,$8F,$F6,$C6,$E7,$FD,$87,$8F,$FB,$9F,$57,$F7,$46,$DF
	.dc.b	$F7,$46,$DF,$F7,$46,$D2,$7F,$96,$CB,$E8,$DF,$48,$F7,$F3,$3A,$9C
	.dc.b	$C6,$23,$F9,$AC,$7F,$72,$DB,$11,$CE,$B9,$3B,$FA,$57,$C4,$EA,$11
	.dc.b	$87,$8B,$5E,$11,$B8,$7B,$94,$93,$C1,$E9,$E8,$7F,$7C,$8F,$E9,$3A
	.dc.b	$7B,$DF,$E9,$70,$FA,$4F,$83,$39,$94,$E5,$E3,$AF,$C6,$8E,$91,$FE
	.dc.b	$E6,$6C,$F1,$78,$46,$6B,$AD,$71,$ED,$B3,$E2,$F5,$6B,$E0,$FA,$4B
	.dc.b	$81,$B3,$27,$C6,$7A,$A3,$66,$5F,$F7,$CF,$D2,$3E,$93,$9F,$91,$27
	.dc.b	$88,$9D,$38,$6B,$CA,$7C,$B1,$8D,$3D,$DF,$09,$89,$95,$C5,$F4,$C7
	.dc.b	$B3,$E9,$1F,$13,$7D,$F9,$7E,$29,$CB,$67,$2B,$0F,$A9,$2A,$E9,$5D
	.dc.b	$78,$E5,$F3,$7C,$1C,$58,$8E,$5C,$3F,$9E,$C3,$F1,$98,$F4,$BD,$D7
	.dc.b	$C4,$FA,$47,$87,$E9,$F2,$24,$E6,$3E,$BC,$EA,$30,$43,$B9,$FE,$18
	.dc.b	$C3,$F4,$2E,$04,$71,$70,$DC,$3F,$24,$DC,$7D,$1D,$5F,$86,$FA,$47
	.dc.b	$BB,$49,$87,$DC,$87,$CA,$E2,$3C,$11,$C1,$D4,$F3,$9E,$9F,$31,$D7
	.dc.b	$E9,$25,$73,$8C,$BD,$63,$2F,$48,$FC,$CF,$9F,$FE,$4A,$8F,$C6,$FD
	.dc.b	$1B,$E9,$3F,$3A,$D8,$FC,$8C,$ED,$F8,$8D,$DB,$F6,$4D,$D7,$F8,$AB
	.dc.b	$67,$FC,$87,$5F,$B0,$9C,$F9,$7E,$22,$4F,$8B,$BB,$F5,$BF,$46,$FA
	.dc.b	$4A,$49,$FD,$1F,$D1,$B6,$FD,$9A,$FA,$4A,$49,$DF,$91,$FB,$8F,$A3
	.dc.b	$FE,$27,$18,$21,$F7,$DC,$B7,$DF,$FD,$FF,$D2,$3D,$CA,$73,$8F,$E0
	.dc.b	$7F,$48,$F0,$B1,$7F,$B8,$C1,$5A,$99,$3F,$BC,$AE,$0D,$8F,$E1,$BF
	.dc.b	$48,$F8,$CF,$E1,$FF,$1F,$F8,$6A,$FE,$07,$2F,$A4,$D9,$AE,$B9,$3A
	.dc.b	$8C,$BF,$C2,$AB,$93,$F9,$D6,$F4,$FF,$2B,$6F,$3F,$F4,$E6,$CF,$F9
	.dc.b	$F3,$0F,$D2,$D7,$F3,$1F,$7B,$CD,$3F,$C5,$BE,$7D,$ED,$FF,$26,$6C
	.dc.b	$7F,$1A,$9D,$BF,$AD,$6D,$7F,$3B,$9D,$CF,$EC,$5B,$1D,$F4,$FE,$63
	.dc.b	$5C,$32,$FE,$C6,$DB,$FF,$48,$B5,$5B,$E0,$7F,$9C,$F7,$BE,$1F,$F9
	.dc.b	$DF,$D1,$7D,$DB,$FE,$5E,$D7,$F8,$14,$ED,$FE,$66,$D7,$F8,$34,$6F
	.dc.b	$3F,$EA,$8D,$CB,$FB,$BD,$6E,$7F,$AC,$36,$33,$7E,$43,$C7,$F0,$DF
	.dc.b	$D3,$BE,$91,$F3,$BE,$6F,$A4,$99,$FF,$44,$9C,$B6,$72,$F9,$27,$94
	.dc.b	$48,$D1,$0F,$4F,$AB,$FD,$73,$E9,$3E,$FC,$35,$FE,$DA,$FC,$4F,$5F
	.dc.b	$5C,$F2,$30,$43,$F9,$45,$72,$24,$6E,$3E,$3F,$CD,$77,$F7,$BE,$8F
	.dc.b	$B1,$27,$FA,$FD,$A7,$51,$F5,$27,$4F,$68,$FE,$E7,$F4,$9F,$7A,$3F
	.dc.b	$B3,$AE,$AC,$9A,$E6,$93,$A6,$BF,$CF,$FE,$93,$FC,$3F,$CD,$FE,$DA
	.dc.b	$FA,$37,$CB,$FF,$8C,$27,$34,$9C,$BF,$EE,$0C,$7C,$4E,$6B,$FC,$6F
	.dc.b	$18,$9D,$CA,$B9,$C6,$1B,$5C,$30,$FF,$1F,$CF,$0B,$E5,$3E,$3F,$5F
	.dc.b	$FB,$86,$1D,$CF,$F7,$8E,$1C,$9F,$26,$1C,$AB,$A4,$7B,$5E,$58,$D8
	.dc.b	$E3,$7D,$31,$F0,$5F,$19,$E6,$6C,$F5,$4E,$EF,$9C,$78,$C6,$3E,$87
	.dc.b	$89,$9C,$7C,$CF,$2C,$7A,$4E,$E3,$8E,$1E,$17,$47,$1C,$3C,$0F,$4C
	.dc.b	$7C,$EF,$97,$0E,$A7,$D8,$7F,$CC,$AD,$7C,$E6,$27,$E5,$BA,$F6,$3F
	.dc.b	$D0,$2F,$F9,$A6,$7F,$CB,$7F,$0B,$FE,$68,$6C,$66,$76,$FF,$9B,$3E
	.dc.b	$93,$FC,$D0,$9A,$47,$FC,$D8,$FC,$BF,$92,$7D,$0F,$F9,$87,$2F,$C3
	.dc.b	$D0,$E4,$7C,$D9,$DF,$6F,$D3,$8D,$5F,$3B,$E5,$BF,$FA,$06,$31,$5B
	.dc.b	$AE,$78,$7E,$53,$0E,$71,$F6,$4E,$67,$7C,$39,$BF,$E8,$EB,$E6,$EC
	.dc.b	$E7,$87,$FA,$5A,$B9,$DF,$AA,$72,$7A,$E3,$33,$E5,$8E,$D8,$79,$6F
	.dc.b	$AD,$F1,$6D,$6E,$3D,$18,$79,$2F,$FD,$4B,$5A,$7B,$47,$FA,$2F,$7F
	.dc.b	$F4,$46,$3E,$E7,$FD,$15,$3E,$EB,$E7,$8F,$F9,$C2,$B4,$FA,$EF,$93
	.dc.b	$E1,$C7,$0E,$FA,$73,$87,$B5,$BF,$0A,$FA,$3F,$F3,$25,$6C,$95,$FE
	.dc.b	$52,$CD,$F4,$8E,$FE,$72,$D3,$FA,$1D,$FF,$A3,$2F,$75,$A9,$E1,$3E
	.dc.b	$98,$F0,$FC,$07,$F7,$48,$C1,$2F,$FC,$BD,$5C,$1F,$F3,$1E,$5F,$E8
	.dc.b	$2F,$1B,$A6,$12,$FF,$D3,$73,$EB,$FA,$4C,$5F,$0D,$11,$E8,$F7,$7F
	.dc.b	$9D,$5F,$FA,$69,$B5,$FE,$70,$87,$D1,$30,$E1,$D7,$27,$83,$F3,$1D
	.dc.b	$C6,$6B,$FC,$ED,$7D,$36,$FF,$64,$57,$4E,$2D,$46,$31,$FB,$77,$FA
	.dc.b	$BF,$A4,$7D,$27,$CE,$6E,$1F,$EB,$83,$CB,$E7,$66,$64,$7F,$AD,$27
	.dc.b	$75,$EB,$BF,$F2,$14,$E6,$B7,$39,$6E,$1E,$37,$EE,$CF,$FB,$98,$FF
	.dc.b	$5F,$96,$7A,$78,$3A,$E9,$93,$A7,$E2,$60,$9C,$D7,$FA,$66,$B9,$61
	.dc.b	$F1,$3E,$7F,$FC,$99,$1C,$4F,$73,$89,$FF,$50,$7C,$87,$E0,$EB,$FD
	.dc.b	$DC,$DA,$FE,$24,$7D,$8A,$FC,$AC,$66,$BF,$8B,$35,$39,$9E,$07,$FB
	.dc.b	$71,$D1,$9C,$3F,$DF,$71,$9E,$3F,$6F,$8B,$FA,$2F,$A3,$7D,$27,$E9
	.dc.b	$3E,$8F,$FB,$F5,$FF,$B9,$A7,$2F,$FB,$A2,$33,$97,$FA,$ED,$BA,$27
	.dc.b	$F9,$64,$A7,$31,$8A,$E7,$3C,$7C,$5F,$07,$C5,$4A,$ED,$16,$7B,$7A
	.dc.b	$5E,$53,$F7,$3F,$49,$CB,$2F,$04,$FF,$3B,$CB,$E8,$46,$23,$F4,$5E
	.dc.b	$6F,$F0,$0E,$93,$CF,$2F,$E8,$55,$B9,$CC,$60,$F5,$BF,$E9,$77,$15
	.dc.b	$FD,$27,$E9,$3F,$CC,$51,$88,$FD,$87,$B5,$FC,$3A,$73,$1F,$EE,$4E
	.dc.b	$3F,$EA,$1E,$27,$F1,$C6,$DF,$47,$FA,$A7,$36,$6B,$FC,$4F,$39,$69
	.dc.b	$CC,$B9,$7C,$2A,$E3,$3B,$BE,$44,$79,$5D,$BE,$5C,$DD,$F9,$3C,$0D
	.dc.b	$AF,$03,$67,$71,$E3,$9F,$E5,$7F,$43,$FE,$B4,$8F,$84,$D9,$65,$F5
	.dc.b	$BE,$91,$F4,$8F,$A4,$E3,$EE,$CC,$7E,$6F,$8C,$FA,$4F,$F4,$64,$74
	.dc.b	$9F,$E8,$73,$9F,$63,$E3,$7D,$27,$C7,$EE,$B7,$F1,$DB,$AF,$C7,$6E
	.dc.b	$BF,$1E,$3F,$5D,$C7,$FE,$03,$BE,$57,$F3,$DB,$3F,$3F,$E9,$3E,$A1
	.dc.b	$F5,$79,$AD,$7D,$58,$B3,$F5,$66,$E1,$F9,$3F,$A2,$7A,$DF,$D4,$9B
	.dc.b	$87,$F4,$2F,$A4,$7C,$2E,$6A,$4E,$7C,$BB,$97,$F4,$CB,$FE,$6D,$3F
	.dc.b	$C1,$BD,$5F,$D9,$3F,$8B,$7F,$23,$97,$E1,$5C,$D7,$E4,$DC,$CF,$13
	.dc.b	$F2,$8E,$23,$F8,$7B,$C2,$79,$7B,$FA,$AD,$B5,$F0,$3C,$BE,$92,$47
	.dc.b	$8F,$E3,$EE,$1E,$19,$B7,$E5,$9B,$5F,$C2,$B5,$7C,$6F,$C7,$3F,$03
	.dc.b	$F0,$6F,$96,$C5,$7E,$C6,$B6,$FF,$7B,$FA,$4F,$96,$E6,$FF,$C1,$78
	.dc.b	$A7,$39,$FC,$3F,$A4,$F8,$27,$C2,$FA,$4B,$87,$CB,$8B,$3F,$26,$6D
	.dc.b	$FC,$D9,$B5,$F4,$E6,$D7,$CA,$F9,$9F,$9B,$EC,$FC,$E6,$D7,$D7,$6C
	.dc.b	$FD,$56,$E1,$FC,$6F,$E9,$3F,$C2,$13,$FA,$CA,$FE,$5F,$18,$9F,$E4
	.dc.b	$7D,$AD,$FF,$98,$DF,$F9,$7E,$B5,$7C,$31,$FF,$1B,$CE,$5A,$A7,$BB
	.dc.b	$FC,$F0,$C4,$70,$EB,$72,$3E,$8B,$F1,$3F,$47,$F8,$A6,$E3,$F9,$36
	.dc.b	$E1,$F9,$16,$E9,$F9,$56,$CF,$E6,$DB,$3F,$92,$6E,$1E,$06,$E1,$E1
	.dc.b	$AF,$D0,$61,$E1,$69,$E1,$2D,$F8,$5B,$87,$85,$A7,$84,$B1,$E1,$6E
	.dc.b	$1E,$16,$DF,$86,$7F,$59,$7E,$16,$CF,$85,$A7,$84,$B7,$E1,$6C,$78
	.dc.b	$5A,$78,$4B,$7E,$16,$DF,$86,$BF,$7D,$7E,$16,$DF,$85,$B5,$E1,$69
	.dc.b	$E1,$2C,$78,$5B,$1E,$16,$DF,$85,$B7,$E1,$AF,$E2,$57,$E1,$69,$E1
	.dc.b	$2D,$F8,$5B,$5E,$16,$9E,$12,$C7,$85,$B7,$E1,$6D,$F8,$67,$F8,$BD
	.dc.b	$F8,$5B,$3E,$16,$9E,$12,$DF,$85,$B1,$E1,$69,$E1,$2D,$F8,$5B,$7E
	.dc.b	$18,$FD,$1D,$F8,$5B,$5E,$16,$D7,$85,$B7,$E1,$6C,$F8,$5B,$3E,$16
	.dc.b	$9E,$12,$DF,$84,$5F,$85,$B5,$E1,$6D,$78,$5B,$7E,$16,$CF,$85,$B3
	.dc.b	$E1,$6D,$F8,$5A,$78,$6F,$C0,$4A,$F0,$36,$BC,$0D,$BF,$03,$67,$C0
	.dc.b	$D9,$F0,$36,$FC,$0D,$BF,$05,$76,$25,$76,$6D,$F6,$69,$D8,$B1,$D9
	.dc.b	$B7,$D9,$B5,$DA,$7A,$12,$BA,$36,$7A,$34,$E8,$7F,$6F,$EA,$FE,$DB
	.dc.b	$D3,$FD,$B9,$B3,$FD,$DD,$B3,$B8,$FE,$F5,$3F,$ED,$8F,$A1,$7D,$37
	.dc.b	$F3,$66,$CF,$F3,$5A,$08,$CB,$96,$BF,$CE,$FE,$91,$F4,$8F,$7B,$94
	.dc.b	$A7,$31,$85,$D7,$FE,$CD,$9F,$04,$69,$F1,$79,$7F,$DE,$1F,$F3,$22
	.dc.b	$FA,$4F,$F4,$2E,$DF,$E7,$58,$C4,$EB,$AF,$3A,$7F,$A2,$A7,$94,$FF
	.dc.b	$5F,$3D,$AF,$07,$01,$C5,$75,$AE,$73,$C1,$AE,$EE,$D7,$0B,$F0,$56
	.dc.b	$3D,$2C,$FD,$24,$C3,$D7,$19,$7D,$B1,$97,$C7,$7F,$E8,$E7,$FD,$09
	.dc.b	$F4,$8F,$A4,$EF,$E2,$C9,$CC,$79,$27,$51,$82,$1D,$3E,$07,$D3,$F4
	.dc.b	$8F,$86,$BE,$EC,$AF,$F4,$94,$E5,$B3,$97,$E3,$C2,$79,$CC,$9E,$31
	.dc.b	$BE,$0F,$F5,$55,$73,$9E,$3F,$19,$E2,$23,$E0,$F8,$2A,$E8,$F1,$C7
	.dc.b	$A2,$3E,$8F,$D2,$3C,$4F,$5F,$93,$2F,$FD,$55,$39,$6C,$F0,$23,$EE
	.dc.b	$E8,$FB,$7F,$48,$FA,$4A,$49,$FF,$5A,$43,$E4,$F1,$1F,$88,$8E,$0E
	.dc.b	$A7,$9C,$EB,$5F,$F6,$55,$6A,$B8,$56,$9E,$BF,$2F,$E2,$F9,$F9,$18
	.dc.b	$21,$DD,$CB,$E1,$8F,$C9,$F6,$BE,$8B,$F2,$BF,$46,$FA,$4A,$6A,$25
	.dc.b	$FF,$B3,$EB,$83,$63,$8F,$6F,$2E,$D9,$3F,$ED,$5C,$3F,$28,$FE,$A1
	.dc.b	$76,$EA,$31,$37,$6F,$F7,$55,$6A,$79,$4E,$7B,$BF,$DF,$F3,$D2,$B7
	.dc.b	$7F,$AE,$FA,$47,$D2,$52,$63,$FE,$E7,$FA,$47,$07,$ED,$17,$D2,$52
	.dc.b	$4E,$FC,$9F,$DC,$FD,$1B,$AF,$84,$49,$FF,$79,$4E,$5B,$EF,$FF,$00
	.dc.b	$FA,$47,$B9,$4E,$71,$FC,$13,$E9,$1E,$26,$2F,$89,$FE,$FD,$8C,$B6
	.dc.b	$38,$11,$FE,$63,$C2,$E3,$FE,$67,$0F,$A3,$7C,$7F,$F1,$2F,$90,$FC
	.dc.b	$A4,$F0,$AD,$D3,$E9,$3F,$8B,$B6,$74,$7F,$3E,$AE,$2F,$F2,$26,$E1
	.dc.b	$E1,$8F,$E8,$73,$C5,$FE,$4E,$D8,$D4,$27,$8D,$FA,$1B,$1D,$FC,$0F
	.dc.b	$41,$E4,$F0,$7F,$95,$FB,$3E,$C9,$F2,$8F,$9F,$7B,$3F,$CC,$1B,$1A
	.dc.b	$9E,$0F,$F3,$A6,$E5,$E4,$9D,$CF,$F3,$E6,$C7,$A2,$77,$97,$CE,$6D
	.dc.b	$7F,$62,$6B,$FE,$A3,$D7,$C5,$F0,$3F,$D0,$BC,$37,$C3,$7F,$44,$FA
	.dc.b	$2F,$C2,$DF,$A9,$AF,$F4,$E9,$DB,$EE,$6B,$FD,$4A,$37,$97,$CC,$6E
	.dc.b	$5F,$82,$AD,$CF,$D0,$6C,$66,$FC,$87,$8F,$E1,$FF,$AB,$FD,$23,$E7
	.dc.b	$BC,$DF,$49,$33,$F9,$F3,$96,$CE,$5F,$24,$F2,$89,$1A,$21,$E9,$F1
	.dc.b	$7F,$B2,$FD,$27,$E0,$21,$AF,$F6,$D7,$E2,$7A,$FA,$E5,$91,$82,$1F
	.dc.b	$DD,$2B,$91,$23,$71,$F1,$FE,$4B,$BF,$BD,$F4,$7D,$99,$3F,$8A,$B4
	.dc.b	$EA,$3E,$A4,$E9,$ED,$1F,$DF,$3E,$93,$F2,$11,$F9,$27,$F3,$8D,$EA
	.dc.b	$D7,$44,$8D,$35,$FF,$05,$FD,$23,$D6,$73,$FF,$8B,$3E,$8D,$F2,$DF
	.dc.b	$BC,$AD,$4E,$63,$F7,$F9,$78,$B0,$5F,$07,$FC,$6B,$19,$8C,$42,$F9
	.dc.b	$BC,$2B,$B6,$1E,$28,$C1,$CB,$1F,$E1,$B7,$CB,$0E,$97,$C3,$1F,$25
	.dc.b	$FF,$91,$31,$EB,$7B,$C3,$8D,$FA,$1B,$5B,$C7,$C1,$5E,$6B,$E3,$7B
	.dc.b	$AE,$A3,$1F,$0B,$C7,$1C,$D7,$08,$F4,$1C,$27,$6F,$C1,$C3,$BD,$FA
	.dc.b	$63,$DE,$75,$C7,$CA,$D3,$AC,$F0,$7C,$07,$0A,$EA,$76,$BF,$14,$77
	.dc.b	$C3,$FC,$BF,$1F,$42,$3F,$9B,$99,$3F,$CD,$8E,$5F,$C0,$79,$0F,$83
	.dc.b	$E2,$D8,$DB,$5E,$73,$FD,$5E,$7F,$CD,$B1,$FE,$76,$FA,$4A,$9A,$47
	.dc.b	$6F,$CE,$F8,$97,$D0,$FF,$3D,$DB,$F1,$35,$CC,$F8,$F3,$AE,$DF,$AB
	.dc.b	$19,$9F,$C7,$CE,$63,$E6,$99,$DF,$FA,$BC,$66,$7A,$CF,$94,$EB,$7E
	.dc.b	$1A,$FF,$48,$61,$E5,$8C,$4F,$CA,$8C,$61,$E2,$8E,$57,$E2,$AD,$C7
	.dc.b	$F6,$6A,$DB,$E4,$37,$7E,$49,$DC,$79,$CF,$F4,$FE,$1E,$79,$F1,$C7
	.dc.b	$9B,$0F,$05,$F4,$C3,$D6,$DA,$E9,$3F,$D6,$2F,$E2,$69,$FD,$93,$0E
	.dc.b	$C2,$33,$3C,$23,$FD,$23,$18,$7E,$01,$E1,$33,$D3,$FD,$CE,$BF,$B7
	.dc.b	$CF,$F3,$BA,$FE,$F7,$F3,$4F,$A3,$FF,$3A,$47,$8E,$F8,$3F,$51,$B8
	.dc.b	$7F,$14,$FA,$47,$7E,$23,$CD,$3A,$C9,$5A,$BF,$75,$EF,$2C,$61,$FE
	.dc.b	$AD,$CB,$FD,$19,$E2,$E6,$A5,$F0,$3C,$FF,$05,$FD,$F1,$C5,$FC,$29
	.dc.b	$F2,$DF,$AF,$E9,$38,$CF,$5D,$51,$E7,$F2,$3F,$9E,$E1,$FE,$08,$6D
	.dc.b	$7F,$AF,$E1,$F4,$4C,$38,$75,$C9,$E0,$FC,$47,$71,$9A,$FF,$62,$DF
	.dc.b	$4D,$FF,$C4,$95,$D3,$93,$51,$8C,$7E,$D5,$BE,$91,$F4,$9F,$35,$B8
	.dc.b	$7F,$8C,$8F,$2F,$9D,$99,$91,$FE,$22,$9D,$D7,$AE,$FF,$C9,$B3,$9A
	.dc.b	$DC,$E5,$B8,$78,$DF,$B9,$3F,$E5,$F3,$FC,$5A,$59,$E9,$E0,$EB,$AE
	.dc.b	$4E,$9F,$89,$82,$73,$5F,$EE,$8A,$E5,$87,$C4,$F9,$FF,$F2,$AC,$71
	.dc.b	$3D,$CE,$27,$FD,$DF,$F1,$FF,$88,$AF,$F2,$4B,$69,$3F,$E6,$1C,$17
	.dc.b	$F9,$5A,$D4,$E9,$A9,$CD,$7E,$02,$78,$71,$C8,$FA,$17,$FE,$FF,$7F
	.dc.b	$C9,$F5,$C2,$31,$7F,$DE,$27,$FC,$E3,$F4,$6F,$A4,$FD,$37,$D1,$FF
	.dc.b	$97,$E7,$FC,$AF,$39,$7F,$CB,$31,$9A,$E4,$7B,$1C,$67,$FE,$68,$6C
	.dc.b	$F6,$BE,$BE,$27,$6F,$8B,$93,$D2,$2E,$1D,$05,$FC,$DC,$3A,$7A,$3F
	.dc.b	$BB,$FA,$4F,$25,$6E,$B5,$87,$FA,$33,$9D,$E6,$FF,$01,$6C,$7A,$B0
	.dc.b	$C4,$E7,$D8,$FE,$0D,$F4,$8F,$53,$17,$FE,$78,$8C,$3F,$E8,$EE,$2F
	.dc.b	$B7,$E2,$7F,$10,$6C,$67,$4F,$C6,$36,$75,$1F,$E2,$68,$C1,$BE,$6C
	.dc.b	$CC,$8C,$CF,$CE,$AF,$C4,$D6,$63,$07,$E2,$A7,$C1,$E7,$F6,$E7,$EE
	.dc.b	$D8,$C9,$E5,$8C,$BF,$E9,$4F,$A4,$7B,$5F,$C9,$DB,$1F,$EA,$CF,$A4
	.dc.b	$7D,$23,$E9,$35,$72,$BF,$D6,$3E,$43,$E9,$3F,$D1,$31,$D6,$7F,$9A
	.dc.b	$CE,$7D,$9F,$87,$F4,$9F,$13,$C0,$B7,$F2,$5B,$BF,$B9,$BD,$7F,$EB
	.dc.b	$9E,$2F,$D5,$C6,$2B,$F5,$15,$FA,$A9,$FD,$7D,$F5,$9F,$A9,$F4,$9F
	.dc.b	$48,$FA,$7C,$F6,$32,$4B,$F9,$B1,$97,$F2,$9F,$44,$F5,$FF,$A6,$36
	.dc.b	$BF,$AB,$FD,$23,$E1,$73,$52,$73,$E6,$5C,$3F,$A6,$5F,$F3,$19,$FE
	.dc.b	$01,$EA,$FF,$B9,$5F,$C7,$3F,$90,$9F,$C1,$39,$6F,$6F,$E4,$63,$1C
	.dc.b	$1F,$15,$E0,$DA,$D4,$ED,$A7,$62,$D7,$6F,$A4,$E8,$75,$F9,$0B,$3E
	.dc.b	$48,$B3,$E4,$9B,$7F,$EE,$8D,$5F,$1F,$FE,$07,$AF,$E4,$31,$8C,$3F
	.dc.b	$C9,$71,$89,$D2,$9D,$3A,$6D,$7F,$81,$FE,$93,$E7,$5F,$FB,$B2,$72
	.dc.b	$FF,$2E,$94,$EA,$BE,$4F,$D2,$7C,$73,$E4,$7D,$25,$BF,$9F,$16,$7E
	.dc.b	$7C,$DB,$F9,$93,$6B,$EB,$CD,$AF,$E6,$DA,$BE,$5F,$EA,$7B,$1F,$55
	.dc.b	$B7,$F7,$1B,$5F,$71,$B8,$A7,$FA,$3C,$BE,$93,$FC,$18,$FF,$84,$FE
	.dc.b	$95,$F2,$DF,$E1,$D9,$CD,$F1,$AF,$F0,$25,$6A,$FE,$21,$D2,$B6,$FF
	.dc.b	$54,$F2,$D3,$BA,$FE,$65,$5A,$79,$E1,$F8,$FF,$A5,$FC,$87,$D2,$DC
	.dc.b	$7F,$2E,$DA,$FC,$BB,$72,$FC,$A3,$67,$F3,$ED,$9F,$92,$DE,$DF,$F4
	.dc.b	$53,$76,$34,$95,$FA,$6C,$3C,$6D,$3C,$65,$BF,$1B,$70,$F1,$B4,$F1
	.dc.b	$96,$3C,$6D,$C3,$C6,$DC,$3C,$6D,$C3,$C7,$5F,$B6,$C3,$C6,$D9,$F1
	.dc.b	$B4,$F1,$97,$0F,$1B,$63,$C6,$D3,$C6,$5C,$3C,$6D,$C3,$C6,$DC,$3C
	.dc.b	$77,$FC,$0F,$0F,$1B,$70,$F1,$B6,$BC,$6D,$3C,$65,$8F,$1B,$63,$C6
	.dc.b	$DC,$3C,$6D,$C3,$C6,$DC,$3C,$77,$FC,$7B,$0F,$1B,$4F,$19,$70,$F1
	.dc.b	$B6,$BC,$6D,$3C,$65,$8F,$1B,$70,$F1,$B7,$0F,$1B,$70,$F1,$D7,$EE
	.dc.b	$30,$F1,$B6,$7C,$6D,$3C,$65,$C3,$C6,$D8,$F1,$B4,$F1,$97,$0F,$1B
	.dc.b	$70,$F1,$B7,$0F,$1C,$FF,$BA,$70,$F1,$B6,$FC,$6D,$AF,$1B,$70,$F1
	.dc.b	$B6,$7C,$6D,$9F,$1B,$4F,$19,$70,$F1,$B7,$0F,$1C,$30,$F1,$B6,$FC
	.dc.b	$6D,$BF,$1B,$70,$F1,$B6,$7C,$6D,$9F,$1B,$70,$F1,$B4,$F1,$97,$0F
	.dc.b	$1E,$1E,$22,$5F,$89,$B7,$E2,$6E,$1E,$26,$CF,$89,$B3,$E2,$6E,$1E
	.dc.b	$26,$E1,$E2,$6E,$1E,$2B,$F0,$1F,$0B,$AB,$FB,$97,$4F,$F7,$36,$C7
	.dc.b	$F7,$A6,$E3,$FD,$E9,$B9,$7F,$37,$87,$D1,$3E,$8D,$EE,$7F,$66,$9C
	.dc.b	$C6,$1F,$ED,$32,$E9,$FF,$61,$46,$27,$4F,$F4,$A1,$F4,$8F,$87,$E9
	.dc.b	$18,$6C,$66,$3C,$0D,$7F,$BB,$F9,$BF,$DF,$1F,$E9,$67,$FA,$71,$7D
	.dc.b	$27,$F5,$8C,$FD,$91,$89,$D8,$79,$3F,$E7,$F6,$C7,$09,$E5,$AF,$F6
	.dc.b	$29,$E5,$5B,$ED,$CB,$E1,$6B,$D6,$AC,$F0,$BF,$1D,$63,$D2,$CF,$CD
	.dc.b	$CC,$3C,$B7,$FE,$88,$6C,$E5,$F3,$4F,$17,$D7,$F4,$6F,$A4,$F9,$6D
	.dc.b	$8F,$B0,$D7,$FC,$3E,$D7,$EA,$39,$C7,$F0,$4D,$CF,$FC,$02,$D9,$F9
	.dc.b	$DF,$03,$EE,$FA,$47,$C3,$5F,$76,$57,$CE,$9C,$B6,$72,$FF,$AB,$61
	.dc.b	$3C,$E6,$4F,$18,$DE,$3F,$95,$AE,$73,$C7,$87,$C4,$47,$C1,$F0,$35
	.dc.b	$D1,$E3,$8F,$44,$7D,$6F,$A3,$FF,$13,$4F,$9F,$E9,$3D,$86,$7C,$39
	.dc.b	$1F,$88,$9D,$BF,$EC,$56,$E3,$E5,$E9,$FC,$0F,$D2,$3E,$92,$92,$FF
	.dc.b	$23,$0F,$93,$C4,$7F,$B3,$E3,$83,$A9,$E7,$3A,$BF,$DB,$56,$AB,$85
	.dc.b	$69,$EB,$F2,$FE,$2F,$A0,$91,$82,$1D,$DC,$BE,$18,$FC,$CF,$6F,$F8
	.dc.b	$F2,$3F,$4F,$F4,$6F,$A4,$FC,$EB,$63,$F5,$73,$B7,$FD,$D0,$DC,$BF
	.dc.b	$86,$37,$9F,$FC,$96,$D9,$0A,$FD,$8C,$FF,$94,$1F,$D1,$47,$FB,$CD
	.dc.b	$B5,$FC,$5D,$B3,$BE,$9F,$2F,$54,$BF,$4B,$E8,$8C,$47,$38,$B7,$CE
	.dc.b	$7F,$73,$F4,$8F,$A4,$A4,$BF,$E1,$3F,$48,$DB,$F7,$EB,$E9,$29,$27
	.dc.b	$7E,$2F,$F0,$4F,$A4,$74,$F0,$89,$3F,$C5,$27,$2D,$F7,$FF,$85,$FD
	.dc.b	$23,$DB,$A7,$38,$FE,$1F,$F4,$8F,$0B,$17,$C4,$FE,$3F,$19,$6C,$70
	.dc.b	$23,$FC,$93,$E8,$DF,$19,$FC,$6B,$E3,$FF,$2D,$39,$FA,$4F,$E3,$B4
	.dc.b	$75,$DD,$27,$44,$1C,$DF,$D3,$6F,$9C,$63,$A3,$FD,$1B,$18,$9D,$57
	.dc.b	$3A,$D1,$F8,$6C,$3F,$A5,$CF,$F5,$7F,$A4,$F3,$CF,$F3,$CF,$A0,$7B
	.dc.b	$5F,$CC,$9B,$1F,$D0,$27,$6F,$40,$E2,$BF,$B1,$4F,$09,$D3,$63,$C1
	.dc.b	$A7,$F3,$EA,$E1,$9F,$FA,$99,$B7,$F9,$C6,$AB,$7C,$07,$F4,$6F,$01
	.dc.b	$F0,$DF,$D2,$3E,$8B,$EF,$5F,$F5,$66,$BE,$79,$DB,$FD,$69,$AF,$AE
	.dc.b	$37,$9F,$F7,$B6,$E5,$FE,$BA,$AD,$CF,$F7,$F6,$C6,$6F,$C8,$78,$FE
	.dc.b	$1F,$FA,$DF,$D2,$3E,$7B,$CD,$F4,$93,$3F,$ED,$73,$96,$CE,$5F,$24
	.dc.b	$F2,$89,$1A,$21,$E9,$F0,$FF,$B4,$FD,$27,$E0,$F2,$FF,$68,$EB,$FD
	.dc.b	$DE,$BE,$27,$A9,$AE,$59,$18,$21,$F7,$6F,$93,$FC,$39,$A9,$A6,$29
	.dc.b	$F2,$7E,$2F,$6F,$EA,$7C,$54,$BF,$F0,$4D,$A7,$51,$F5,$E7,$4F,$86
	.dc.b	$3F,$C0,$7F,$49,$F8,$18,$FF,$08,$AE,$AC,$9A,$E6,$93,$A6,$BF,$97
	.dc.b	$FA,$4F,$F1,$47,$37,$F2,$8F,$A4,$7C,$BF,$FB,$DA,$73,$49,$CB,$FC
	.dc.b	$BB,$1F,$13,$9A,$FF,$7E,$C6,$27,$72,$AE,$71,$86,$D7,$0C,$3F,$9C
	.dc.b	$D7,$33,$B3,$8A,$E1,$EB,$7F,$02,$9D,$1D,$5C,$47,$83,$0F,$15,$F5
	.dc.b	$AF,$73,$ED,$AD,$B7,$1C,$CF,$80,$A2,$B9,$BF,$D3,$B1,$F4,$B6,$F9
	.dc.b	$BC,$4F,$2C,$F0,$7C,$D5,$D2,$B8,$C7,$5C,$7D,$D8,$79,$E3,$DB,$3A
	.dc.b	$37,$7C,$CE,$8E,$70,$F8,$93,$FE,$65,$8F,$B2,$D8,$CB,$EF,$3F,$CD
	.dc.b	$47,$D5,$F1,$DF,$07,$FE,$75,$AF,$E9,$AD,$8C,$CF,$09,$CB,$FE,$70
	.dc.b	$FA,$4F,$F3,$8A,$69,$1D,$1F,$63,$E4,$9F,$42,$E7,$F9,$F3,$FD,$EA
	.dc.b	$BE,$8D,$6A,$73,$3D,$4E,$55,$C5,$FC,$19,$C4,$CC,$70,$AE,$73,$BD
	.dc.b	$FF,$2B,$5B,$7F,$2F,$7D,$E3,$99,$DE,$33,$3C,$75,$F9,$37,$D3,$0E
	.dc.b	$D7,$E0,$38,$C6,$5A,$66,$FC,$2F,$5A,$FD,$0E,$1E,$19,$CD,$71,$6D
	.dc.b	$71,$AD,$E1,$F0,$77,$FF,$58,$61,$E9,$7F,$D6,$55,$ED,$AD,$E3,$FE
	.dc.b	$98,$AD,$38,$9F,$EB,$46,$CF,$F1,$5C,$69,$BD,$3D,$E7,$CE,$D8,$7C
	.dc.b	$5B,$E8,$FF,$CE,$75,$DE,$B5,$4B,$FE,$85,$3F,$EB,$0C,$5F,$48,$EF
	.dc.b	$DD,$6A,$FF,$BC,$D6,$A7,$24,$BF,$F5,$CE,$5F,$E8,$AF,$17,$17,$FD
	.dc.b	$D7,$0E,$95,$F0,$6F,$D1,$7B,$8C,$12,$BF,$C0,$F1,$EB,$FA,$4E,$78
	.dc.b	$FE,$3E,$31,$5F,$EC,$2F,$6F,$F9,$D3,$7A,$7E,$22,$FA,$2F,$8A,$E4
	.dc.b	$98,$FF,$82,$6B,$8B,$FE,$16,$8E,$17,$F2,$27,$38,$7F,$BB,$5B,$3F
	.dc.b	$E6,$03,$B7,$0E,$F1,$F0,$E3,$F6,$71,$7D,$23,$E9,$3E,$9B,$70,$FF
	.dc.b	$70,$9E,$7F,$37,$33,$23,$FD,$BF,$3B,$AF,$35,$FF,$91,$E7,$35,$B9
	.dc.b	$CB,$73,$FD,$B4,$FC,$38,$FC,$1B,$FE,$5A,$3F,$DD,$65,$9E,$BE,$06
	.dc.b	$BA,$64,$E9,$F6,$C6,$27,$FC,$65,$3C,$6F,$BF,$CF,$FF,$94,$A3,$81
	.dc.b	$EE,$71,$3F,$E3,$9F,$8F,$FC,$2B,$78,$BE,$4C,$FE,$62,$73,$7F,$C3
	.dc.b	$E9,$18,$BF,$EB,$2D,$8F,$E3,$99,$AF,$96,$12,$FE,$7C,$BC,$5C,$5F
	.dc.b	$CC,$7F,$4F,$F4,$8F,$A4,$FD,$4F,$D1,$FF,$7F,$C3,$FD,$3F,$A7,$8E
	.dc.b	$B5,$18,$3F,$B5,$CE,$F0,$FF,$2B,$36,$BD,$4D,$9F,$EE,$35,$CA,$B7
	.dc.b	$E2,$78,$3E,$22,$57,$59,$FF,$57,$DF,$16,$CF,$9B,$D1,$95,$B9,$7D
	.dc.b	$27,$5C,$BC,$35,$FC,$13,$A5,$E6,$53,$8D,$FF,$9B,$63,$38,$73,$FA
	.dc.b	$4F,$F3,$0B,$7D,$2E,$5E,$C7,$F0,$E9,$CD,$FF,$A0,$39,$7F,$D8,$BC
	.dc.b	$BF,$C5,$DF,$F6,$3F,$1D,$9F,$AD,$8F,$B6,$72,$D8,$C9,$CC,$FB,$0E
	.dc.b	$6B,$95,$73,$6C,$64,$E5,$1F,$C8,$BE,$91,$EC,$D3,$F9,$57,$D2,$3E
	.dc.b	$91,$F4,$99,$C3,$E5,$5F,$07,$E9,$78,$8F,$A4,$FF,$40,$C6,$B1,$FE
	.dc.b	$69,$39,$F6,$3D,$DF,$49,$EF,$EE,$B7,$ED,$6E,$FE,$D6,$F3,$FF,$6D
	.dc.b	$E1,$FF,$76,$C7,$C7,$FA,$49,$7F,$17,$92,$E1,$F1,$66,$E1,$C6,$3F
	.dc.b	$DA,$15,$F2,$E7,$2F,$FB,$E3,$0F,$F7,$CC,$7D,$3B,$CD,$7F,$BF,$3E
	.dc.b	$89,$EB,$7F,$4A,$6E,$1F,$D0,$3E,$91,$F0,$99,$A9,$39,$F2,$6E,$1F
	.dc.b	$CF,$2F,$F9,$4B,$FB,$AF,$57,$FC,$25,$5F,$81,$7F,$13,$AF,$C1,$73
	.dc.b	$3F,$8B,$73,$3E,$97,$33,$BE,$BF,$C7,$46,$27,$51,$8F,$A4,$91,$DF
	.dc.b	$E3,$6E,$1D,$66,$E1,$E1,$9B,$5D,$E6,$D7,$A6,$6A,$D5,$F3,$1F,$1C
	.dc.b	$FC,$1F,$C1,$3E,$53,$18,$FE,$D6,$B6,$FF,$73,$FA,$4F,$96,$E6,$FF
	.dc.b	$BF,$66,$9C,$E3,$F0,$FE,$93,$DE,$7C,$0F,$A4,$B8,$7C,$B8,$B3,$F2
	.dc.b	$66,$DF,$CD,$F9,$7F,$8F,$EB,$FC,$86,$E1,$F4,$DB,$7F,$45,$B5,$C6
	.dc.b	$7E,$C5,$71,$69,$FC,$47,$E8,$FF,$DF,$6D,$8F,$87,$E6,$FF,$85,$27
	.dc.b	$38,$7F,$BE,$DB,$7F,$DF,$FE,$91,$F4,$9B,$C3,$FC,$13,$18,$C7,$F1
	.dc.b	$1F,$49,$F8,$0F,$A4,$FC,$0B,$63,$27,$E3,$63,$2D,$3F,$21,$D7,$F8
	.dc.b	$8E,$AF,$C4,$B6,$32,$7E,$56,$32,$D3,$3A,$7E,$56,$32,$D8,$CE,$7F
	.dc.b	$9B,$9C,$B6,$32,$7E,$6E,$32,$D3,$33,$FA,$78,$CB,$63,$3D,$DF,$E8
	.dc.b	$A9,$CB,$63,$3D,$5F,$B0,$8F,$F4,$7E,$1E,$DB,$FD,$63,$70,$FD,$83
	.dc.b	$6F,$F5,$CD,$9F,$DC,$36,$B9,$CF,$C3,$9E,$6D,$AE,$6D,$39,$96,$39
	.dc.b	$B6,$79,$C7,$C9,$39,$96,$79,$B6,$79,$B4,$E6,$59,$E7,$3F,$3E,$39
	.dc.b	$B4,$E6,$59,$E6,$D8,$E6,$D9,$E7,$3F,$56,$79,$B6,$79,$B4,$E6,$58
	.dc.b	$E6,$D9,$E7,$1F,$6C,$E6,$59,$E6,$D9,$E6,$D3,$99,$67,$99,$F8,$17
	.dc.b	$9B,$67,$9B,$67,$9B,$67,$9B,$4E,$77,$F8,$93,$F8,$8D,$7C,$36,$D7
	.dc.b	$AD,$B5,$E3,$6D,$73,$69,$CE,$FF,$14,$E3,$0F,$CC,$B9,$BF,$C2,$39
	.dc.b	$7F,$37,$1A,$23,$82,$5F,$C0,$9E,$4F,$E0,$DE,$4E,$67,$8B,$67,$8E
	.dc.b	$1F,$1A,$B8,$BE,$FB,$F3,$92,$7C,$AD,$AE,$CD,$9E,$13,$A2,$4E,$9B
	.dc.b	$1A,$8C,$11,$C1,$23,$E7,$4F,$56,$9C,$8B,$3C,$5B,$3C,$63,$F1,$03
	.dc.b	$E8,$3F,$C5,$68,$85,$72,$E0,$E1,$E5,$7F,$B7,$9D,$35,$7D,$23,$E9
	.dc.b	$1E,$F7,$19,$ED,$39,$8C,$47,$0E,$DF,$F6,$DB,$B9,$D3,$C6,$FF,$B6
	.dc.b	$BF,$E5,$D5,$F4,$7E,$8C,$BF,$D0,$D1,$89,$D7,$5E,$74,$F6,$CF,$29
	.dc.b	$57,$0B,$D4,$7F,$95,$9A,$FF,$A1,$EE,$D7,$F9,$A2,$3C,$5F,$37,$83
	.dc.b	$3E,$9C,$8C,$90,$EF,$8F,$8A,$35,$18,$3F,$CC,$B3,$CD,$F7,$47,$8B
	.dc.b	$E8,$FE,$C4,$7C,$1F,$17,$C6,$67,$82,$47,$FA,$06,$76,$F5,$0E,$34
	.dc.b	$F8,$FF,$01,$EE,$FA,$27,$D1,$DF,$6A,$61,$FE,$8B,$9C,$B6,$72,$F5
	.dc.b	$84,$F3,$9E,$9F,$19,$2B,$93,$DF,$E0,$71,$1C,$62,$E3,$FE,$BF,$C1
	.dc.b	$58,$DF,$F2,$B0,$AC,$5F,$E6,$63,$6F,$FA,$87,$E8,$DE,$1F,$B7,$C8
	.dc.b	$93,$98,$FB,$B3,$A8,$C1,$0F,$67,$47,$E0,$BE,$8D,$F4,$74,$93,$FE
	.dc.b	$AF,$87,$CA,$62,$3E,$84,$70,$75,$3C,$E7,$A7,$CB,$F5,$F9,$F9,$5C
	.dc.b	$E3,$2F,$58,$CB,$D2,$32,$F6,$8C,$F0,$7F,$B9,$F0,$57,$E6,$BB,$7F
	.dc.b	$C5,$77,$FE,$C5,$FA,$27,$D1,$FE,$79,$B1,$FE,$C9,$9D,$BF,$89,$6E
	.dc.b	$1F,$EF,$36,$F1,$7F,$08,$6C,$FF,$0B,$AF,$D7,$57,$FB,$E5,$FD,$44
	.dc.b	$7E,$4D,$B1,$9E,$AF,$17,$5F,$8D,$95,$E7,$72,$FA,$7A,$3D,$1F,$15
	.dc.b	$23,$04,$3B,$39,$7C,$11,$EF,$31,$7E,$59,$C4,$FF,$BA,$7E,$85,$BF
	.dc.b	$F0,$2F,$A3,$7D,$1D,$3F,$82,$CE,$AB,$87,$8B,$8F,$A3,$91,$83,$FD
	.dc.b	$E9,$F4,$7F,$96,$7F,$69,$5F,$EA,$A7,$47,$ED,$E3,$59,$B8,$7F,$9A
	.dc.b	$C7,$FB,$6A,$7F,$DF,$DF,$46,$F6,$FF,$8B,$B6,$3B,$FC,$0F,$F1,$AF
	.dc.b	$A3,$70,$FF,$1C,$C3,$95,$77,$AD,$4C,$8F,$E8,$99,$78,$5F,$E8,$D7
	.dc.b	$C9,$CE,$5F,$D2,$23,$11,$FC,$3F,$E8,$9F,$19,$FC,$AB,$D7,$FD,$2C
	.dc.b	$F0,$3F,$A8,$4B,$E8,$FF,$97,$B6,$34,$27,$8B,$FC,$6D,$B8,$F9,$61
	.dc.b	$3C,$5F,$15,$78,$7B,$64,$74,$BE,$39,$7F,$60,$6C,$79,$3B,$FF,$A0
	.dc.b	$FB,$BE,$A8,$F1,$FC,$FB,$D8,$FE,$8A,$D8,$F3,$CE,$DF,$E6,$0D,$D3
	.dc.b	$CD,$3B,$9F,$E6,$8D,$8F,$54,$EF,$1F,$EE,$6D,$AF,$EE,$CD,$4D,$A7
	.dc.b	$C0,$7F,$58,$F0,$5F,$0B,$FD,$6B,$E8,$7F,$09,$87,$F4,$36,$E5,$F2
	.dc.b	$6B,$73,$F4,$1A,$EF,$9F,$91,$D6,$E4,$6D,$AF,$5F,$85,$FE,$D3,$F4
	.dc.b	$6F,$9B,$F1,$7D,$1C,$AF,$BB,$39,$6C,$E5,$F3,$4F,$28,$91,$A2,$1E
	.dc.b	$BF,$0F,$FB,$BF,$D1,$FD,$A8,$6A,$F5,$BE,$11,$AE,$69,$3F,$8E,$9D
	.dc.b	$36,$72,$FF,$69,$AE,$84,$AF,$99,$F3,$3E,$0F,$77,$E7,$7B,$12,$FF
	.dc.b	$33,$69,$D4,$77,$9D,$3D,$63,$FC,$1F,$F4,$6E,$5B,$D5,$2B,$F4,$73
	.dc.b	$96,$CF,$02,$3F,$A8,$FA,$3F,$EC,$DC,$DF,$E3,$DF,$A2,$7D,$17,$F8
	.dc.b	$FA,$73,$49,$CB,$FE,$43,$C3,$B3,$9A,$FF,$22,$C6,$27,$72,$AE,$71
	.dc.b	$86,$D7,$0C,$3F,$C9,$B3,$C2,$F9,$4F,$8F,$D7,$FF,$19,$E1,$DC,$FF
	.dc.b	$2A,$61,$C9,$F2,$61,$CA,$BA,$47,$B5,$E5,$8D,$8E,$37,$D3,$1F,$05
	.dc.b	$F1,$9E,$66,$CF,$54,$EE,$F9,$C7,$8C,$63,$E8,$78,$99,$C7,$CC,$F2
	.dc.b	$C7,$A4,$EE,$38,$E1,$E1,$74,$71,$C3,$C0,$F4,$C7,$CE,$F9,$70,$EA
	.dc.b	$7D,$87,$FC,$E8,$D7,$CE,$62,$7E,$5B,$AF,$61,$F0,$3F,$E7,$0A,$FF
	.dc.b	$16,$B6,$33,$3F,$E7,$79,$CB,$FE,$76,$1F,$47,$5A,$65,$C1,$1F,$91
	.dc.b	$F2,$2F,$A0,$FE,$95,$97,$E2,$27,$FD,$21,$5F,$46,$B5,$E0,$FD,$39
	.dc.b	$C9,$CA,$F8,$EF,$FE,$8B,$BE,$11,$89,$FF,$4B,$46,$2B,$94,$79,$B0
	.dc.b	$FC,$BE,$1D,$A3,$EE,$1D,$A3,$37,$CE,$3F,$D4,$77,$CD,$D9,$CF,$0F
	.dc.b	$F5,$3D,$F3,$BF,$64,$E4,$FF,$56,$46,$67,$D1,$1E,$BC,$3C,$B7,$D6
	.dc.b	$F8,$B6,$B7,$1E,$49,$F1,$6F,$FD,$7F,$0F,$53,$FE,$B3,$BC,$CF,$93
	.dc.b	$1F,$EF,$13,$B6,$C7,$4C,$34,$FF,$6E,$BF,$E6,$CD,$E9,$DE,$4F,$9A
	.dc.b	$7D,$1F,$F9,$FA,$B4,$49,$FF,$4D,$64,$FA,$37,$7F,$29,$65,$D9,$DC
	.dc.b	$EA,$FC,$55,$BA,$C5,$7F,$A9,$71,$FF,$4A,$F8,$B9,$A9,$7C,$0F,$F0
	.dc.b	$64,$E9,$C5,$7F,$A8,$6F,$9E,$3F,$E1,$BA,$F5,$FD,$1F,$2A,$ED,$9A
	.dc.b	$3D,$1E,$3F,$F3,$9C,$7F,$C2,$6D,$AF,$F5,$C4,$3E,$8B,$E3,$3A,$EA
	.dc.b	$E2,$47,$FC,$40,$6F,$1C,$57,$FA,$D6,$FA,$65,$FE,$2F,$AE,$9C,$3A
	.dc.b	$8C,$63,$F8,$09,$FF,$7B,$FD,$1B,$D8,$F9,$CD,$8E,$F3,$FE,$40,$7F
	.dc.b	$C5,$71,$9C,$7F,$D8,$8D,$BF,$F1,$53,$65,$3F,$E5,$1A,$E4,$F0,$BE
	.dc.b	$E7,$E1,$A7,$E1,$38,$7C,$2F,$E0,$5F,$F5,$D9,$FE,$3C,$2C,$7C,$A9
	.dc.b	$77,$EF,$A2,$56,$DE,$D3,$A8,$C1,$2B,$E9,$FC,$F7,$F9,$42,$31,$97
	.dc.b	$CB,$8C,$4F,$FB,$8F,$E3,$3F,$07,$5E,$E8,$46,$1F,$81,$3F,$99,$8C
	.dc.b	$46,$B8,$F0,$79,$26,$D7,$F9,$62,$B9,$3F,$85,$F0,$F3,$7E,$F7,$F4
	.dc.b	$9F,$44,$FA,$3F,$D3,$7D,$1F,$F9,$7E,$FF,$CC,$33,$97,$FC,$C7,$19
	.dc.b	$BF,$F7,$E3,$73,$FF,$31,$36,$7D,$95,$C6,$B7,$E1,$F6,$F8,$99,$5D
	.dc.b	$22,$CF,$4F,$4F,$C6,$7E,$E7,$E8,$F9,$65,$DA,$3F,$79,$CE,$F2,$E9
	.dc.b	$C3,$1F,$F3,$CC,$65,$FF,$3D,$FD,$1F,$F9,$59,$BE,$97,$2F,$47,$F8
	.dc.b	$34,$E7,$36,$5F,$E9,$1E,$2F,$C9,$F1,$FF,$14,$7F,$2B,$C7,$69,$36
	.dc.b	$30,$DC,$BF,$D2,$15,$C9,$B5,$FE,$45,$6C,$EE,$3C,$31,$FA,$29,$E1
	.dc.b	$18,$3F,$0E,$71,$30,$3E,$81,$EC,$FF,$21,$6D,$7F,$A8,$7E,$8D,$F4
	.dc.b	$6F,$A3,$D5,$CA,$FF,$53,$78,$8F,$A3,$FF,$3D,$C7,$FA,$93,$1E,$79
	.dc.b	$78,$EB,$5E,$BF,$C3,$FA,$3F,$89,$DB,$70,$F7,$B6,$FD,$5B,$7E,$FA
	.dc.b	$31,$C3,$F3,$E7,$F4,$53,$FC,$0B,$E8,$FE,$71,$F3,$F9,$2E,$1F,$2A
	.dc.b	$2D,$74,$3F,$DA,$91,$92,$39,$8E,$51,$8B,$FF,$5D,$7D,$0B,$D5,$FE
	.dc.b	$90,$D9,$FE,$7B,$F4,$6F,$86,$CD,$49,$CF,$95,$72,$FE,$91,$7F,$CB
	.dc.b	$27,$ED,$7A,$9F,$EE,$B7,$F0,$6F,$E3,$F2,$FC,$73,$9A,$F8,$AE,$67
	.dc.b	$89,$F9,$17,$11,$FC,$21,$E1,$3C,$BA,$F5,$F4,$72,$FB,$FC,$6D,$C3
	.dc.b	$AC,$DB,$F0,$CD,$AF,$3C,$DA,$ED,$35,$6A,$F9,$8F,$8E,$7F,$78,$E5
	.dc.b	$FF,$62,$FC,$A6,$31,$FF,$75,$D6,$DF,$F7,$96,$4F,$A3,$F9,$91,$AC
	.dc.b	$3F,$C0,$33,$9E,$1F,$81,$F4,$72,$BE,$4F,$D1,$DD,$3E,$3C,$DB,$F9
	.dc.b	$73,$6B,$E8,$FC,$BF,$C9,$F5,$BE,$53,$63,$E9,$B7,$0F,$A2,$DB,$FA
	.dc.b	$AD,$9E,$67,$DA,$AF,$F0,$77,$D1,$FF,$BD,$63,$F9,$E5,$FF,$3D,$8C
	.dc.b	$38,$0E,$1D,$B5,$E1,$3F,$E2,$2F,$19,$F4,$4F,$A3,$EF,$39,$8C,$61
	.dc.b	$F8,$CF,$A3,$FC,$37,$D1,$FE,$1D,$B1,$93,$F2,$31,$96,$9F,$94,$DF
	.dc.b	$F1,$DD,$5F,$8F,$6C,$64,$FC,$CC,$65,$A6,$73,$FC,$CC,$65,$B1,$9D
	.dc.b	$BF,$3F,$39,$6C,$64,$FD,$3C,$65,$A6,$67,$F4,$11,$96,$C6,$7B,$BF
	.dc.b	$D1,$F3,$96,$C6,$7A,$BF,$69,$5F,$A9,$C3,$DA,$FE,$C1,$B8,$7E,$D1
	.dc.b	$B5,$FB,$A6,$CF,$18,$F8,$2F,$16,$CF,$16,$9C,$4B,$1C,$4F,$8A,$71
	.dc.b	$2C,$71,$6C,$71,$69,$C6,$3E,$4B,$C5,$A7,$12,$C7,$16,$C7,$18,$F9
	.dc.b	$AF,$16,$C7,$16,$9C,$4B,$1C,$4F,$A2,$71,$2C,$71,$6C,$71,$69,$C6
	.dc.b	$FE,$AB,$C5,$B3,$C5,$B3,$C5,$B3,$C6,$FE,$DC,$7F,$B3,$AF,$EA,$B6
	.dc.b	$BE,$7B,$6B,$E4,$B6,$FE,$1B,$67,$DA,$D9,$F3,$B4,$F0,$96,$FC,$27
	.dc.b	$F1,$A8,$EA,$DA,$EA,$DA,$EA,$D9,$EA,$DA,$EA,$D3,$AD,$FB,$1C,$61
	.dc.b	$F9,$D7,$37,$EE,$72,$FE,$7E,$34,$47,$04,$BF,$A1,$3C,$9F,$4D,$77
	.dc.b	$24,$F5,$6C,$6A,$30,$47,$04,$C3,$E0,$CF,$06,$CE,$DB,$3B,$8F,$8B
	.dc.b	$3B,$69,$B2,$C6,$CF,$93,$1B,$6C,$6D,$A6,$F0,$F9,$B3,$B7,$F1,$70
	.dc.b	$FA,$27,$D1,$BD,$BF,$E7,$13,$98,$C3,$FC,$EE,$5D,$7F,$ED,$78,$C4
	.dc.b	$E9,$FE,$87,$FB,$47,$47,$9B,$5F,$E8,$8D,$77,$1E,$05,$E6,$FF,$6F
	.dc.b	$7F,$D4,$BA,$34,$FF,$45,$4B,$E9,$3F,$D1,$93,$9A,$FF,$2B,$C6,$2B
	.dc.b	$7D,$56,$7F,$CB,$6D,$7B,$CD,$BE,$91,$E0,$AE,$2D,$3F,$CC,$7E,$86
	.dc.b	$FE,$6E,$46,$4F,$4C,$73,$32,$FF,$96,$29,$39,$8E,$11,$FE,$7D,$6C
	.dc.b	$70,$C7,$D9,$1D,$0E,$DF,$46,$FA,$4F,$47,$8D,$27,$31,$ED,$9D,$46
	.dc.b	$08,$78,$FE,$5B,$E1,$7D,$1B,$E9,$2F,$B5,$2B,$D9,$39,$6C,$E5,$FF
	.dc.b	$39,$C2,$79,$CF,$4F,$8C,$95,$C9,$EF,$F0,$18,$8E,$31,$74,$F7,$E0
	.dc.b	$AC,$70,$7F,$B2,$61,$58,$BF,$F7,$4C,$6D,$FA,$5F,$48,$F1,$3D,$BE
	.dc.b	$4C,$9C,$C7,$E0,$67,$51,$82,$1E,$CE,$AF,$C3,$7D,$1B,$E9,$29,$27
	.dc.b	$F0,$10,$F9,$3C,$47,$FA,$5A,$38,$3A,$9E,$73,$D3,$E5,$FA,$FD,$14
	.dc.b	$AE,$71,$97,$AC,$65,$E9,$19,$7B,$46,$78,$3F,$0F,$82,$BF,$3D,$DD
	.dc.b	$FE,$2B,$BF,$CB,$7D,$1B,$E9,$3F,$44,$D8,$FC,$D4,$ED,$FF,$57,$B7
	.dc.b	$0F,$C9,$B7,$8B,$FD,$F8,$D9,$0A,$FD,$D3,$FE,$B8,$7F,$3C,$DE,$CF
	.dc.b	$09,$E2,$EB,$B1,$E6,$8A,$F1,$8F,$3F,$47,$A3,$E2,$A4,$FB,$23,$2F
	.dc.b	$BA,$32,$F8,$63,$2F,$8E,$33,$97,$EB,$B0,$57,$F0,$5F,$A1,$6F,$FC
	.dc.b	$1F,$E9,$1F,$49,$4D,$46,$FC,$7E,$1F,$49,$34,$FE,$09,$87,$FA,$AD
	.dc.b	$FE,$0D,$F4,$7F,$97,$7F,$DB,$71,$96,$D7,$F1,$59,$D6,$9F,$C0,$9F
	.dc.b	$F6,$FC,$FF,$0F,$FA,$47,$BB,$FC,$6D,$B1,$DB,$E0,$FF,$8F,$7D,$27
	.dc.b	$B2,$30,$43,$FD,$D6,$E5,$BB,$F0,$23,$FC,$77,$E9,$1F,$17,$FC,$A7
	.dc.b	$E3,$7F,$41,$39,$FA,$4F,$E5,$94,$75,$DB,$27,$44,$3F,$DF,$75,$C9
	.dc.b	$B1,$B6,$9E,$1E,$FF,$E7,$1F,$49,$E2,$8F,$E6,$1F,$3E,$F6,$7F,$9E
	.dc.b	$36,$3F,$9A,$4E,$D7,$17,$F2,$C8,$C3,$FC,$BA,$78,$4E,$BB,$BF,$36
	.dc.b	$A3,$33,$C7,$5F,$19,$27,$BB,$E2,$F8,$3F,$E9,$DE,$0B,$E2,$3F,$A8
	.dc.b	$7D,$17,$D6,$BE,$C7,$EA,$1C,$61,$FD,$02,$B8,$4F,$F4,$D6,$BC,$04
	.dc.b	$F0,$CB,$CE,$D8,$FD,$9C,$F0,$AE,$97,$E4,$24,$FF,$59,$A7,$C4,$7F
	.dc.b	$65,$FA,$47,$CE,$FA,$3E,$92,$5F,$F5,$F9,$CB,$67,$2F,$96,$D5,$CE
	.dc.b	$64,$EC,$86,$FC,$5F,$EE,$9F,$49,$F8,$3C,$FF,$B0,$E1,$FE,$04,$AF
	.dc.b	$37,$AD,$AE,$59,$18,$21,$F1,$6F,$96,$FF,$69,$B1,$FC,$59,$A9,$B6
	.dc.b	$F9,$AF,$1F,$BF,$F6,$3D,$89,$7F,$DF,$ED,$3A,$8F,$95,$3A,$7C,$51
	.dc.b	$FE,$0F,$FA,$4F,$C5,$47,$F8,$1D,$75,$64,$D7,$3C,$9D,$35,$FC,$07
	.dc.b	$D2,$7F,$62,$DF,$F9,$7F,$D2,$3E,$8F,$F2,$F3,$9A,$4E,$5F,$E6,$B8
	.dc.b	$F8,$9C,$D7,$E7,$A3,$13,$B9,$57,$38,$C3,$6B,$86,$1F,$A6,$9E,$17
	.dc.b	$CA,$7C,$7E,$BF,$EC,$30,$EE,$7F,$46,$C3,$93,$E4,$C3,$95,$74,$8F
	.dc.b	$6B,$CB,$1B,$1C,$6F,$A6,$3E,$0B,$E3,$3C,$CD,$9E,$A9,$DD,$F3,$8F
	.dc.b	$18,$C7,$D0,$F1,$33,$8F,$99,$E5,$8F,$49,$DC,$71,$C3,$C2,$E8,$E3
	.dc.b	$87,$81,$E9,$8F,$9D,$F2,$E1,$D4,$FB,$0F,$F1,$03,$91,$5C,$BE,$93
	.dc.b	$1E,$C3,$E0,$FF,$CF,$33,$FE,$29,$6C,$66,$7A,$1F,$71,$C3,$FE,$80
	.dc.b	$FA,$4F,$F3,$92,$69,$1F,$F3,$B3,$F3,$3E,$3D,$F4,$3F,$E7,$1C,$7F
	.dc.b	$11,$5F,$E0,$2A,$D4,$E7,$6F,$0C,$63,$97,$EA,$D6,$CF,$A7,$87,$48
	.dc.b	$CC,$78,$CC,$DF,$37,$E3,$57,$37,$ED,$9C,$F0,$FC,$D5,$F3,$8C,$9C
	.dc.b	$E3,$35,$CE,$3F,$9F,$D7,$39,$F5,$C6,$F0,$FE,$87,$7C,$E7,$35,$B9
	.dc.b	$F6,$D7,$9B,$0F,$4C,$E6,$3C,$E6,$49,$BF,$F9,$FB,$0F,$33,$FE,$80
	.dc.b	$8F,$55,$F8,$31,$FF,$4F,$4E,$DB,$86,$4F,$F4,$95,$70,$3F,$C6,$77
	.dc.b	$D2,$3D,$3D,$3C,$27,$2D,$F8,$B7,$D1,$FF,$9E,$EB,$FD,$39,$7F,$D0
	.dc.b	$9B,$5C,$71,$7D,$27,$F4,$89,$C9,$FE,$A7,$8C,$F7,$F7,$C9,$86,$DF
	.dc.b	$EB,$4D,$9F,$F5,$D6,$5F,$E9,$3F,$17,$84,$E7,$0F,$A3,$86,$0F,$47
	.dc.b	$C0,$7F,$80,$63,$CF,$19,$7F,$D7,$17,$EC,$FA,$4C,$65,$FD,$C2,$31
	.dc.b	$3F,$EB,$9F,$23,$F9,$DB,$7A,$BE,$32,$FA,$35,$7A,$EB,$77,$FD,$CD
	.dc.b	$BA,$FF,$AE,$1B,$7F,$E6,$43,$95,$71,$C2,$47,$FB,$34,$ED,$C3,$8C
	.dc.b	$3F,$C0,$B1,$83,$FD,$9D,$F4,$8F,$6B,$E6,$36,$39,$67,$FE,$5A,$77
	.dc.b	$3A,$23,$AF,$3A,$DF,$F8,$2E,$F9,$3E,$0B,$BA,$7F,$B4,$A6,$DF,$AE
	.dc.b	$6D,$7B,$7A,$3E,$F4,$FF,$AE,$4F,$1F,$83,$88,$91,$FE,$E7,$35,$3D
	.dc.b	$E7,$2F,$F8,$76,$33,$87,$FB,$F5,$7C,$FF,$F9,$49,$E5,$7F,$5A,$31
	.dc.b	$5F,$E2,$5F,$8D,$FC,$43,$FE,$94,$6D,$7D,$B9,$FC,$A4,$6A,$4D,$37
	.dc.b	$8F,$82,$BF,$DF,$D1,$9E,$2F,$C2,$CB,$C6,$E7,$18,$7F,$41,$F4,$4F
	.dc.b	$A4,$FD,$17,$D2,$7F,$80,$31,$FF,$53,$EB,$FE,$FE,$AD,$46,$0F,$F2
	.dc.b	$8C,$EF,$3F,$F2,$03,$69,$3A,$F1,$7B,$E1,$D9,$EF,$F1,$32,$7A,$61
	.dc.b	$FE,$B6,$C3,$94,$76,$9F,$2F,$A3,$23,$74,$FA,$4F,$2B,$E2,$9F,$DE
	.dc.b	$74,$3C,$DA,$71,$9F,$F2,$B4,$66,$F9,$FD,$23,$D3,$FE,$08,$D8,$EF
	.dc.b	$3F,$E9,$C7,$73,$FC,$27,$E8,$1E,$CD,$3F,$8D,$7D,$23,$E9,$1F,$49
	.dc.b	$9C,$3D,$37,$C1,$B7,$C2,$B7,$18,$26,$7F,$E7,$4F,$11,$F4,$9F,$E7
	.dc.b	$99,$CF,$2F,$D0,$AF,$E3,$73,$97,$FC,$F7,$D5,$FE,$E2,$FA,$4F,$E5
	.dc.b	$CF,$FB,$93,$AA,$CA,$6C,$61,$B8,$7D,$6A,$E4,$DA,$FC,$DC,$7F,$BB
	.dc.b	$BE,$93,$99,$D3,$E0,$6D,$6E,$2D,$79,$0F,$98,$57,$F9,$A7,$D1,$3D
	.dc.b	$6A,$7F,$43,$FA,$47,$C2,$E3,$0F,$C1,$D6,$DB,$5B,$9D,$61,$FE,$95
	.dc.b	$AD,$3F,$CB,$EF,$F9,$1A,$11,$AD,$BF,$A0,$CE,$B3,$FF,$0E,$63,$F2
	.dc.b	$BD,$4F,$C3,$61,$F3,$9F,$D0,$C7,$C7,$73,$7F,$56,$B8,$CE,$79,$B9
	.dc.b	$3C,$DE,$9F,$49,$23,$97,$C7,$5C,$7B,$4D,$C3,$A6,$2A,$C5,$FF,$B1
	.dc.b	$B3,$7C,$CF,$CC,$3F,$B7,$72,$3E,$0D,$F2,$B9,$BF,$97,$57,$4F,$C8
	.dc.b	$56,$A7,$24,$8F,$F6,$3E,$6F,$A4,$FA,$B3,$98,$FF,$02,$C6,$27,$E3
	.dc.b	$7D,$27,$C9,$3E,$57,$D2,$5B,$F9,$11,$67,$E6,$4D,$BF,$7F,$CC,$7D
	.dc.b	$1F,$5F,$E9,$37,$0F,$A0,$DB,$FA,$8D,$AF,$D3,$43,$E9,$3F,$BB,$3F
	.dc.b	$ED,$96,$C7,$EA,$F6,$F8,$DE,$4F,$F8,$1E,$73,$87,$FB,$C9,$B7,$F8
	.dc.b	$96,$D7,$09,$FE,$2F,$1A,$9F,$06,$3F,$E1,$DF,$A3,$7D,$27,$58,$C3
	.dc.b	$F8,$DF,$A4,$FC,$77,$D2,$5C,$7F,$18,$DC,$3F,$20,$DD,$3F,$06,$FE
	.dc.b	$57,$AF,$F2,$DE,$CD,$BF,$CA,$36,$FF,$30,$DB,$FC,$5B,$FE,$25,$FA
	.dc.b	$9F,$C0,$4F,$F9,$D9,$AF,$FA,$0D,$AF,$EA,$43,$87,$F5,$8F,$29,$FA
	.dc.b	$2F,$C0,$AF,$E7,$6E,$5E,$4E,$65,$3C,$32,$F2,$1E,$5E,$6B,$7C,$A3
	.dc.b	$FC,$AD,$5E,$06,$BE,$46,$DF,$58,$F5,$7C,$6E,$09,$5F,$E9,$48,$F5
	.dc.b	$C6,$1F,$1C,$79,$B8,$7E,$34,$FC,$76,$47,$B1,$CB,$49,$3C,$8F,$8B
	.dc.b	$1A,$70,$7C,$A7,$9B,$8A,$FE,$CA,$FC,$CA,$CE,$3F,$32,$AD,$FC,$A9
	.dc.b	$B5,$DA,$3C,$35,$F4,$A7,$33,$DE,$72,$71,$2B,$C4,$F1,$57,$9E,$BC
	.dc.b	$13,$A9,$E4,$79,$EB,$AD,$FA,$23,$35,$E1,$AE,$13,$92,$57,$0B,$E2
	.dc.b	$D9,$DB,$6F,$C7,$1F,$8A,$8C,$5F,$92,$7C,$53,$DA,$B9,$B4,$E6,$58
	.dc.b	$E6,$70,$2C,$F3,$24,$7C,$58,$CB,$F8,$33,$F1,$17,$FE,$25,$8F,$D3
	.dc.b	$57,$E4,$09,$3A,$6C,$7E,$B5,$BF,$07,$FC,$12,$3F,$53,$F1,$FF,$DF
	.dc.b	$63,$11,$F3,$9B,$69,$D6,$5E,$3A,$FF,$74,$3F,$AC,$8C,$C7,$DB,$AE
	.dc.b	$31,$88,$E9,$3B,$C3,$FD,$D0,$F3,$AD,$C6,$27,$94,$ED,$EE,$60,$F4
	.dc.b	$C6,$DE,$A6,$0D,$98,$3F,$84,$D7,$8E,$FF,$C9,$D7,$E0,$9D,$C6,$2F
	.dc.b	$C1,$3F,$0B,$0E,$17,$C5,$B8,$78,$6F,$86,$1E,$1B,$E1,$39,$BE,$0D
	.dc.b	$38,$5F,$2A,$F0,$B6,$B8,$C7,$92,$B8,$B6,$B8,$9C,$DE,$2D,$9E,$27
	.dc.b	$B4,$B8,$7B,$0F,$A0,$59,$F9,$87,$C6,$9E,$AF,$73,$A9,$67,$9B,$EF
	.dc.b	$9E,$6D,$39,$E1,$F8,$9A,$E7,$5E,$3A,$DD,$77,$3A,$C6,$5E,$37,$F0
	.dc.b	$E7,$CD,$1F,$6E,$FA,$3E,$5B,$FE,$91,$5F,$EA,$98,$FC,$44,$60,$FF
	.dc.b	$6A,$3F,$D0,$BE,$03,$4F,$F4,$A6,$BF,$D3,$9B,$46,$9F,$9B,$FA,$5F
	.dc.b	$F1,$27,$9D,$F8,$18,$C7,$77,$FB,$A2,$FF,$23,$5A,$6C,$67,$3F,$D6
	.dc.b	$D7,$0B,$FC,$D5,$BE,$91,$F0,$3F,$D1,$E3,$0D,$C3,$B6,$3F,$D9,$44
	.dc.b	$E9,$C1,$FE,$B7,$EF,$FE,$D5,$C1,$FE,$16,$F6,$5F,$49,$8D,$FF,$B7
	.dc.b	$46,$2F,$FD,$46,$FF,$8F,$9B,$7B,$7F,$CB,$91,$B7,$FA,$6B,$6F,$B4
	.dc.b	$EE,$7F,$D9,$1E,$A5,$A7,$CE,$D3,$11,$27,$C3,$19,$7C,$71,$CF,$E8
	.dc.b	$DF,$49,$CF,$CC,$95,$FE,$BA,$9C,$B6,$72,$F9,$A7,$94,$48,$D1,$0F
	.dc.b	$5F,$C1,$FA,$7E,$93,$F0,$51,$FE,$C6,$5E,$E6,$4D,$72,$C9,$D4,$65
	.dc.b	$E1,$1C,$FE,$3A,$47,$37,$B7,$C1,$E2,$3F,$BB,$DA,$71,$C1,$FE,$4E
	.dc.b	$85,$7C,$DF,$A4,$79,$1E,$7F,$02,$63,$FE,$DC,$9C,$B6,$72,$F6,$9E
	.dc.b	$51,$23,$44,$3C,$7D,$1F,$6F,$E8,$DF,$49,$49,$3F,$EE,$A8,$7C,$B6
	.dc.b	$23,$FC,$23,$1C,$1D,$4F,$39,$E9,$F3,$3D,$7E,$82,$57,$38,$CB,$D6
	.dc.b	$32,$F4,$8F,$C7,$76,$FF,$5A,$DF,$FC,$E7,$F4,$6F,$A4,$FC,$93,$6F
	.dc.b	$FD,$FF,$5B,$ED,$FF,$27,$36,$3F,$D7,$13,$B7,$FC,$52,$DB,$FF,$3F
	.dc.b	$CE,$E7,$FD,$4D,$E2,$F8,$88,$FF,$B0,$67,$C7,$DD,$E2,$EC,$93,$9C
	.dc.b	$7F,$CB,$D8,$A7,$CF,$DB,$AA,$F0,$C5,$C3,$DB,$3E,$CE,$AF,$6F,$C7
	.dc.b	$4B,$E8,$7A,$A3,$4E,$06,$09,$E1,$1E,$E9,$F9,$7D,$7F,$B9,$F4,$9F
	.dc.b	$5D,$4C,$D7,$CA,$9C,$3D,$67,$6E,$48,$7F,$94,$B7,$FD,$4B,$8C,$FF
	.dc.b	$D5,$33,$FA,$E8,$FF,$2B,$FD,$13,$8F,$F8,$57,$D2,$BE,$96,$9F,$C3
	.dc.b	$67,$55,$ED,$7F,$DF,$6D,$FA,$5D,$1B,$F4,$3F,$8A,$7D,$2F,$EA,$A7
	.dc.b	$84,$65,$F8,$D1,$97,$FC,$D2,$BE,$1F,$F8,$E7,$D2,$BE,$12,$9D,$23
	.dc.b	$F9,$07,$D2,$B8,$F1,$87,$49,$CC,$48,$FA,$93,$B6,$C7,$F2,$9F,$A5
	.dc.b	$7C,$7F,$F2,$BF,$57,$F5,$B5,$F6,$E7,$FC,$6C,$DF,$57,$5E,$0C,$8C
	.dc.b	$10,$FF,$41,$D7,$26,$C6,$DA,$75,$F1,$3F,$9D,$7C,$07,$8A,$3F,$15
	.dc.b	$F4,$0F,$73,$F9,$F3,$63,$F1,$D3,$B7,$FD,$22,$DE,$9F,$F7,$3C,$EE
	.dc.b	$7F,$D2,$AD,$8F,$F7,$5C,$EF,$C1,$FF,$61,$36,$81,$C3,$75,$F1,$FC
	.dc.b	$2F,$F5,$1F,$19,$ED,$7F,$53,$FA,$3F,$B1,$7F,$A3,$6B,$FE,$A6,$9D
	.dc.b	$BF,$A7,$6B,$FE,$AA,$8D,$E5,$FF,$00,$1B,$97,$FB,$5E,$B7,$28,$C5
	.dc.b	$F8,$C9,$7E,$1F,$8B,$FE,$CD,$F4,$AF,$9E,$F3,$7D,$2C,$AF,$DC,$4E
	.dc.b	$5B,$39,$7C,$93,$CA,$24,$68,$87,$A7,$C9,$FE,$E9,$F4,$BF,$81,$CF
	.dc.b	$D5,$ED,$7C,$53,$3C,$B2,$FE,$34,$65,$FE,$15,$5C,$75,$FF,$7B,$36
	.dc.b	$3E,$AB,$53,$77,$7C,$E7,$8B,$E0,$BE,$A7,$B7,$2F,$AC,$65,$FA,$D1
	.dc.b	$97,$FD,$B7,$69,$FF,$07,$7D,$2B,$9E,$F5,$4B,$FE,$41,$39,$6C,$F0
	.dc.b	$23,$FC,$97,$E9,$7F,$B8,$6F,$F8,$EF,$A5,$7D,$2F,$E4,$29,$5A,$53
	.dc.b	$A7,$F2,$B8,$77,$73,$5C,$63,$13,$B9,$E7,$EE,$B0,$F0,$5F,$7C,$3B
	.dc.b	$5E,$F0,$E3,$3C,$F1,$F2,$4F,$3A,$DC,$F1,$C3,$A6,$3E,$23,$CB,$7C
	.dc.b	$27,$94,$78,$0F,$D6,$D7,$26,$D7,$2B,$F1,$5F,$19,$E6,$71,$3D,$78
	.dc.b	$DB,$E5,$0C,$7D,$2F,$03,$C8,$70,$2C,$70,$8D,$C7,$1A,$F0,$3B,$BF
	.dc.b	$81,$87,$97,$1F,$5C,$F4,$BE,$38,$78,$8F,$B1,$1F,$E6,$96,$BF,$1F
	.dc.b	$1F,$F3,$4B,$E3,$35,$EE,$3E,$3F,$35,$FE,$32,$6C,$66,$7F,$CE,$33
	.dc.b	$97,$98,$F8,$6A,$D3,$2E,$08,$FC,$4F,$8C,$7D,$17,$F0,$B8,$F8,$E7
	.dc.b	$F2,$29,$FA,$13,$AE,$5F,$A5,$39,$39,$57,$1D,$BF,$21,$19,$7F,$25
	.dc.b	$5C,$63,$A9,$E5,$8C,$CF,$1D,$7F,$D1,$B8,$78,$23,$15,$FE,$93,$8C
	.dc.b	$61,$E2,$8E,$55,$FA,$9C,$3B,$61,$E2,$9C,$9F,$CF,$23,$33,$DA,$3D
	.dc.b	$35,$CE,$FA,$C7,$99,$B7,$CD,$B5,$C6,$3D,$35,$C3,$4F,$95,$6C,$3E
	.dc.b	$04,$62,$3E,$25,$F9,$F1,$FA,$13,$C2,$B7,$79,$3F,$C5,$97,$C0,$F5
	.dc.b	$57,$08,$F8,$1D,$36,$7D,$CD,$F8,$87,$D2,$FF,$A3,$2B,$FC,$6D,$87
	.dc.b	$F4,$36,$D7,$1C,$5F,$4A,$EF,$E9,$97,$AE,$B5,$39,$BE,$75,$AC,$31
	.dc.b	$7F,$C3,$E7,$FB,$5E,$5F,$ED,$FF,$17,$57,$27,$89,$FE,$4C,$9D,$38
	.dc.b	$AE,$47,$D9,$9F,$66,$6F,$A5,$E5,$3A,$BF,$7C,$7A,$3C,$7F,$E7,$58
	.dc.b	$7F,$95,$5B,$5D,$21,$F4,$6B,$E1,$D5,$27,$83,$F3,$5D,$C6,$6B,$CA
	.dc.b	$7F,$BB,$27,$55,$F6,$A3,$FC,$17,$7E,$0D,$FF,$5D,$87,$83,$83,$E6
	.dc.b	$FD,$2F,$F9,$9E,$3F,$CD,$4D,$BF,$36,$1F,$E3,$F9,$CF,$B9,$C0,$8F
	.dc.b	$F9,$D0,$F5,$F9,$B8,$AF,$2D,$F3,$C7,$C3,$85,$BF,$EA,$13,$6B,$FC
	.dc.b	$19,$35,$FF,$07,$74,$7E,$0A,$BF,$D8,$A7,$79,$FE,$DE,$FF,$A0,$4C
	.dc.b	$CF,$C4,$6D,$FE,$91,$A7,$F8,$B8,$FA,$3E,$07,$5B,$91,$EB,$FA,$5F
	.dc.b	$F2,$AB,$D6,$3E,$94,$62,$FE,$8F,$CC,$7E,$43,$0F,$D8,$B6,$FB,$62
	.dc.b	$BF,$CD,$CE,$6F,$FC,$7E,$D4,$E6,$BF,$0B,$3C,$38,$E4,$7C,$1F,$12
	.dc.b	$DF,$D6,$7F,$4F,$F4,$AF,$A5,$FD,$4F,$D2,$BC,$4C,$C6,$19,$87,$FA
	.dc.b	$A6,$74,$FF,$AA,$EB,$93,$99,$E7,$F0,$12,$30,$71,$7B,$FA,$12,$3F
	.dc.b	$D6,$9F,$4B,$CB,$3F,$CB,$E3,$F2,$78,$3F,$8B,$F4,$FF,$B1,$6F,$FD
	.dc.b	$74,$DA,$DE,$1F,$C0,$A3,$15,$FC,$7B,$E9,$5E,$97,$F0,$66,$E1,$CB
	.dc.b	$D8,$C4,$FF,$0A,$FA,$3F,$F4,$F4,$FF,$B1,$DB,$97,$EF,$BE,$95,$F4
	.dc.b	$8F,$A5,$E1,$EE,$4C,$38,$3F,$C7,$EB,$73,$9B,$FD,$9E,$1D,$34,$FE
	.dc.b	$B9,$E2,$BE,$97,$FC,$EF,$97,$23,$FD,$B7,$19,$F6,$3F,$95,$75,$FF
	.dc.b	$A0,$BE,$96,$E5,$F8,$78,$C5,$FF,$B8,$9F,$F4,$3F,$5E,$AB,$6D,$9D
	.dc.b	$57,$11,$F4,$BD,$0E,$BC,$56,$77,$16,$7F,$90,$E1,$FE,$94,$AC,$69
	.dc.b	$FD,$3B,$E8,$DE,$B7,$F3,$66,$E3,$FC,$AF,$E9,$5F,$09,$9A,$93,$FE
	.dc.b	$F7,$AD,$E5,$FC,$A6,$F7,$39,$AD,$CE,$7C,$AB,$8F,$F8,$1F,$A5,$87
	.dc.b	$F1,$DC,$3F,$57,$EA,$7F,$A5,$5F,$8C,$FE,$13,$0F,$5B,$96,$F4,$FD
	.dc.b	$08,$C6,$1F,$E1,$3D,$BF,$C4,$93,$9B,$D3,$BF,$A5,$91,$DB,$E3,$2E
	.dc.b	$3D,$A6,$E1,$CF,$15,$62,$FF,$CE,$1B,$3E,$43,$E5,$1F,$DA,$F9,$BF
	.dc.b	$D8,$5F,$2D,$8A,$F9,$13,$73,$FE,$AB,$5A,$9C,$92,$3F,$AC,$E4,$FA
	.dc.b	$5F,$9B,$1F,$D2,$31,$F0,$57,$FB,$42,$BF,$C0,$75,$A7,$D1,$F1,$FF
	.dc.b	$14,$F8,$DF,$4B,$6B,$E6,$C5,$C3,$E6,$CD,$AF,$81,$F3,$1F,$37,$D6
	.dc.b	$F9,$CD,$C3,$EA,$B6,$FE,$AB,$6B,$FC,$37,$F4,$BF,$EF,$3A,$F3,$1F
	.dc.b	$E9,$77,$15,$F0,$BE,$23,$F8,$CB,$FD,$92,$33,$5C,$B0,$FF,$05,$C6
	.dc.b	$2F,$FC,$69,$7C,$2B,$51,$DA,$FF,$DE,$75,$B7,$FB,$B7,$D2,$BE,$5F
	.dc.b	$9E,$3F,$E1,$F8,$C5,$FE,$43,$E9,$7F,$19,$F4,$BF,$8D,$6C,$64,$FC
	.dc.b	$AC,$65,$A7,$E6,$3B,$7F,$29,$D9,$F9,$56,$C6,$4F,$CE,$C6,$5A,$7A
	.dc.b	$72,$7D,$2B,$9B,$FC,$56,$F9,$A3,$FC,$12,$E6,$78,$39,$DB,$FC,$96
	.dc.b	$E6,$77,$AF,$2B,$E6,$CB,$FF,$46,$D7,$5A,$FF,$66,$1E,$29,$FF,$29
	.dc.b	$23,$13,$C2,$3F,$C7,$FF,$73,$E1,$8F,$F6,$A5,$5A,$E1,$35,$F6,$F8
	.dc.b	$9E,$CF,$43,$DA,$DB,$F5,$B6,$7D,$ED,$CB,$DB,$5F,$E9,$1F,$83,$FB
	.dc.b	$95,$E5,$C5,$F1,$F9,$CF,$C7,$76,$B5,$35,$F9,$F1,$6B,$8E,$5F,$E9
	.dc.b	$BB,$E2,$DA,$E2,$D7,$8B,$67,$8C,$7F,$AB,$67,$8B,$4E,$25,$38,$96
	.dc.b	$38,$E1,$EC,$8C,$B6,$33,$3D,$23,$18,$7A,$70,$F5,$37,$0F,$23,$70
	.dc.b	$EC,$DA,$E3,$3E,$9A,$E2,$D3,$89,$6B,$8B,$67,$8C,$7A,$67,$8B,$67
	.dc.b	$8B,$4E,$25,$9E,$27,$B6,$78,$B6,$78,$B6,$78,$B4,$E3,$7F,$0C,$FC
	.dc.b	$8D,$FB,$5B,$5E,$76,$D7,$85,$B5,$CD,$A7,$38,$FA,$F1,$96,$C6,$5F
	.dc.b	$DB,$D7,$77,$27,$E4,$EF,$D7,$3B,$C3,$EC,$CE,$70,$F6,$CE,$6F,$94
	.dc.b	$F4,$9F,$43,$D0,$FC,$1C,$F9,$9E,$47,$C4,$79,$47,$08,$E4,$7B,$CE
	.dc.b	$55,$FC,$C3,$0F,$93,$1C,$9A,$F6,$37,$87,$A9,$D5,$F8,$5D,$1D,$A7
	.dc.b	$E3,$36,$3C,$6F,$84,$C1,$C8,$EB,$7F,$09,$C1,$A3,$AD,$FC,$86,$E1
	.dc.b	$F1,$A3,$07,$AE,$BC,$8D,$9E,$A7,$C5,$3A,$BC,$30,$F6,$47,$4B,$F9
	.dc.b	$13,$97,$C9,$19,$3D,$51,$97,$A1,$9C,$3E,$0C,$65,$F9,$6F,$8E,$3D
	.dc.b	$8D,$3D,$38,$7F,$11,$FB,$4F,$1C,$1A,$4B,$FE,$1D,$F4,$5F,$C6,$7E
	.dc.b	$9B,$FC,$2B,$E2,$E2,$7F,$8A,$46,$31,$FE,$19,$5B,$F0,$3F,$DB,$11
	.dc.b	$8A,$DC,$63,$3F,$E2,$16,$FA,$77,$C1,$7F,$44,$C3,$FA,$14,$E6,$31
	.dc.b	$08,$F0,$3F,$D2,$9D,$3C,$5B,$5D,$AB,$AC,$FF,$23,$9C,$EF,$FD,$3E
	.dc.b	$1F,$8B,$3E,$D7,$18,$7F,$25,$F7,$3F,$C6,$55,$FD,$A5,$7D,$4F,$1E
	.dc.b	$46,$DC,$E7,$35,$BF,$1B,$53,$96,$C7,$F5,$F9,$DB,$FE,$F8,$6D,$74
	.dc.b	$8D,$4F,$05,$2F,$93,$C2,$FA,$2A,$62,$24,$FA,$63,$2F,$AE,$3A,$7D
	.dc.b	$4B,$EA,$79,$FA,$12,$BF,$C1,$13,$96,$CE,$5F,$44,$F2,$89,$1A,$21
	.dc.b	$ED,$F8,$7F,$57,$D4,$B9,$EF,$C0,$CA,$FF,$0B,$4E,$5B,$39,$5B,$7C
	.dc.b	$55,$2A,$EB,$5D,$BE,$4E,$57,$47,$C3,$F0,$F8,$8E,$7C,$3F,$83,$C1
	.dc.b	$59,$C7,$B4,$2F,$E9,$7D,$4B,$CC,$F5,$F8,$93,$2F,$F1,$74,$E5,$B3
	.dc.b	$97,$B9,$FE,$1F,$BF,$F1,$93,$A2,$1A,$13,$CD,$C3,$70,$F2,$F6,$7E
	.dc.b	$17,$EA,$5F,$53,$49,$7F,$E4,$18,$7C,$D6,$23,$B4,$70,$75,$3C,$E7
	.dc.b	$A7,$CE,$F5,$FA,$69,$5C,$E3,$2F,$58,$CB,$D2,$3F,$31,$E1,$7F,$83
	.dc.b	$F2,$FB,$1F,$50,$FA,$9F,$CE,$36,$FF,$CA,$D5,$BE,$3F,$DB,$36,$3F
	.dc.b	$C8,$D3,$B7,$E3,$36,$FF,$09,$3B,$9F,$F5,$6F,$85,$E2,$23,$FE,$63
	.dc.b	$AF,$DE,$BC,$BA,$FC,$64,$75,$D9,$5D,$1C,$63,$D3,$1E,$AE,$9F,$5F
	.dc.b	$C9,$4A,$ED,$19,$7C,$11,$97,$D3,$1F,$1B,$AB,$F7,$DE,$B3,$EA,$69
	.dc.b	$9A,$F8,$B3,$87,$FC,$F9,$3B,$72,$43,$F1,$5A,$7E,$95,$C6,$BF,$A9
	.dc.b	$9F,$D7,$47,$E3,$FE,$9D,$C5,$FC,$3F,$EA,$5F,$53,$4F,$E2,$53,$AA
	.dc.b	$FF,$46,$3F,$EF,$C6,$FD,$4E,$8D,$FA,$9F,$C6,$7E,$A7,$F5,$53,$C2
	.dc.b	$32,$FF,$A5,$63,$2F,$E8,$57,$C6,$FF,$21,$FA,$97,$C3,$D3,$A4,$7F
	.dc.b	$25,$FA,$97,$1E,$30,$E9,$39,$89,$1F,$EA,$39,$DB,$63,$F9,$77,$D4
	.dc.b	$BE,$4F,$F9,$87,$B1,$FA,$8A,$FF,$55,$4F,$FA,$09,$BE,$C6,$BC,$09
	.dc.b	$18,$21,$FB,$BA,$E4,$D8,$DB,$4E,$BE,$1F,$F3,$FF,$81,$F1,$47,$FA
	.dc.b	$E7,$E9,$5F,$03,$FD,$0D,$B1,$FE,$BD,$9D,$BF,$C2,$5B,$D3,$FC,$9E
	.dc.b	$77,$3F,$C3,$5B,$1F,$CA,$E7,$7E,$17,$F6,$56,$D7,$FA,$D5,$A9,$B4
	.dc.b	$F8,$9F,$EA,$DE,$2B,$DD,$FE,$B1,$F4,$FF,$63,$0F,$F6,$93,$5F,$E3
	.dc.b	$D3,$B7,$FD,$AC,$D7,$F9,$14,$6F,$2F,$E8,$0D,$CB,$FB,$D5,$6E,$7F
	.dc.b	$A2,$36,$33,$7E,$43,$C7,$F1,$DF,$DA,$FE,$A5,$F4,$7E,$6F,$A9,$99
	.dc.b	$FF,$B9,$E7,$2D,$9C,$BE,$49,$E5,$12,$34,$43,$D3,$E4,$7F,$79,$FA
	.dc.b	$9F,$C0,$E9,$EA,$F6,$7E,$29,$9E,$49,$7F,$1A,$32,$FF,$BE,$6B,$8E
	.dc.b	$BF,$E1,$96,$C7,$FB,$A5,$A9,$BB,$BE,$83,$C5,$F0,$DF,$53,$DA,$97
	.dc.b	$D6,$32,$FD,$68,$CB,$FD,$4E,$D3,$FE,$13,$FA,$97,$2D,$E9,$97,$FE
	.dc.b	$FB,$DB,$FB,$B5,$6A,$31,$A7,$F8,$C6,$31,$1F,$D7,$DB,$3C,$2F,$FC
	.dc.b	$41,$F5,$3F,$DD,$21,$4F,$A9,$7D,$4F,$F8,$EE,$B5,$39,$8E,$D7,$FE
	.dc.b	$3F,$AD,$CF,$2C,$3C,$31,$88,$5F,$28,$C3,$6F,$9C,$F8,$A7,$85,$72
	.dc.b	$9F,$17,$B9,$FE,$16,$AF,$11,$E8,$BE,$4F,$87,$0F,$25,$74,$8D,$9E
	.dc.b	$2A,$E5,$87,$89,$B3,$B8,$F0,$D7,$2A,$ED,$5A,$6E,$3E,$7B,$E0,$48
	.dc.b	$C9,$E0,$C3,$CA,$B1,$F6,$3E,$73,$91,$C3,$1F,$25,$E2,$38,$63,$E3
	.dc.b	$69,$9B,$EA,$7D,$67,$FC,$E4,$7B,$E7,$FC,$FF,$19,$8F,$97,$18,$F7
	.dc.b	$1F,$27,$49,$3F,$E8,$9A,$FF,$1B,$47,$FA,$22,$3F,$0F,$F0,$75,$34
	.dc.b	$8E,$DF,$57,$C9,$BE,$9B,$FC,$E5,$9F,$E1,$6B,$99,$F5,$A7,$5B,$FC
	.dc.b	$18,$CC,$FE,$2E,$73,$1F,$70,$CE,$DF,$E9,$18,$CC,$F5,$9F,$29,$D6
	.dc.b	$FC,$35,$FE,$A5,$C3,$CB,$18,$9F,$93,$18,$C3,$C5,$1C,$AF,$C5,$5B
	.dc.b	$8F,$F4,$E9,$D6,$77,$5C,$63,$CC,$71,$BF,$35,$71,$8F,$49,$FB,$2C
	.dc.b	$3D,$33,$E5,$8E,$87,$2A,$E9,$87,$8B,$0F,$E0,$33,$C7,$0F,$F4,$F6
	.dc.b	$9F,$E7,$B8,$EA,$31,$F9,$31,$FE,$7F,$6D,$79,$0E,$73,$EC,$BE,$CF
	.dc.b	$F9,$A6,$7F,$9E,$CE,$9B,$1F,$E7,$2F,$8A,$7D,$47,$FA,$26,$B5,$D7
	.dc.b	$E4,$7E,$B3,$6B,$FD,$23,$F5,$2E,$FC,$47,$9E,$75,$9A,$B5,$7E,$1B
	.dc.b	$DE,$58,$C3,$C5,$F0,$9F,$DF,$2B,$44,$8F,$55,$78,$73,$FF,$54,$78
	.dc.b	$BC,$A2,$57,$CC,$9F,$5F,$D4,$E2,$FD,$DA,$A3,$CF,$E3,$FF,$3A,$BF
	.dc.b	$92,$DA,$FF,$4A,$C3,$EA,$18,$70,$EB,$93,$C1,$F8,$8E,$E3,$35,$FE
	.dc.b	$99,$BE,$9B,$FF,$B6,$EB,$A7,$16,$A3,$18,$FD,$CB,$FE,$7B,$F5,$2F
	.dc.b	$77,$E6,$B7,$0F,$C4,$9E,$5F,$3B,$33,$23,$F0,$73,$BA,$F5,$DF,$F9
	.dc.b	$12,$73,$5B,$9C,$B7,$0F,$1B,$F7,$27,$FD,$80,$7E,$38,$B3,$D3,$C1
	.dc.b	$D7,$5C,$9D,$3F,$13,$0F,$EE,$11,$82,$57,$CE,$FA,$9F,$F2,$74,$63
	.dc.b	$1F,$8D,$18,$9F,$F6,$17,$CE,$7E,$1E,$7F,$DC,$0D,$AF,$EC,$67,$D7
	.dc.b	$AF,$C9,$C6,$6B,$FB,$3B,$53,$99,$E0,$7E,$A9,$D1,$9B,$78,$BD,$39
	.dc.b	$3D,$8F,$E8,$3E,$A1,$F5,$3F,$A2,$FA,$87,$8B,$98,$C3,$32,$FD,$BC
	.dc.b	$E9,$FD,$CD,$72,$73,$3C,$FE,$06,$46,$0E,$2F,$7F,$52,$47,$F0,$2F
	.dc.b	$A9,$E5,$B7,$FB,$A7,$1F,$95,$C0,$AF,$F0,$4C,$FE,$73,$CD,$FD,$CF
	.dc.b	$49,$E7,$1F,$EF,$18,$C4,$FF,$0E,$FA,$9F,$F2,$F3,$7D,$3E,$3E,$CF
	.dc.b	$F0,$29,$CD,$FF,$BE,$7E,$9B,$FD,$33,$7F,$C5,$9B,$0C,$BC,$8B,$EA
	.dc.b	$1F,$52,$FA,$9F,$E1,$B7,$C2,$F7,$68,$C6,$1E,$C6,$C7,$3B,$E5,$3E
	.dc.b	$1F,$77,$99,$FC,$6F,$C7,$7D,$4F,$F9,$DE,$74,$7F,$36,$8C,$FB,$3F
	.dc.b	$C8,$7B,$3F,$C6,$3F,$53,$72,$FC,$74,$62,$FF,$9E,$BF,$E3,$6E,$CD
	.dc.b	$56,$DB,$3A,$AE,$23,$EA,$7A,$1D,$78,$AC,$EE,$2C,$FC,$98,$FC,$EB
	.dc.b	$FE,$C4,$74,$7F,$44,$FA,$77,$B1,$FC,$D9,$B1,$FC,$9F,$EA,$5F,$0B
	.dc.b	$9A,$93,$C5,$FE,$61,$3A,$8C,$12,$FF,$A5,$47,$F1,$E8,$FC,$6D,$F6
	.dc.b	$9C,$F9,$7F,$27,$D5,$FC,$95,$7C,$67,$F0,$FC,$BF,$82,$DF,$FC,$3B
	.dc.b	$3B,$E9,$E0,$D3,$81,$70,$F8,$47,$EF,$2D,$F5,$3D,$CF,$07,$C6,$DA
	.dc.b	$F0,$45,$AE,$38,$7F,$8B,$2B,$18,$FE,$3B,$37,$CD,$FC,$93,$F1,$1F
	.dc.b	$06,$F9,$9C,$67,$F3,$E6,$E5,$E4,$E7,$FF,$63,$56,$31,$E9,$5F,$D9
	.dc.b	$BE,$A7,$E6,$5F,$A8,$65,$E0,$C7,$FC,$0F,$6A,$DB,$F1,$F6,$7C,$87
	.dc.b	$CD,$3E,$77,$D4,$DB,$F9,$D1,$71,$F7,$D2,$B1,$7C,$FE,$6F,$EB,$7B
	.dc.b	$1F,$5D,$B8,$FD,$76,$E1,$C9,$B5,$FB,$4F,$A9,$FE,$F5,$3E,$B6,$C7
	.dc.b	$EF,$B7,$FF,$74,$79,$7F,$E0,$89,$CE,$1F,$C5,$1B,$7F,$E9,$E6,$D7
	.dc.b	$09,$FF,$0F,$C6,$A4,$F0,$F1,$9F,$52,$F9,$8E,$DA,$7F,$82,$63,$18
	.dc.b	$7E,$47,$EA,$7F,$29,$F5,$3F,$95,$6C,$64,$FC,$B4,$65,$A7,$E6,$7B
	.dc.b	$BF,$39,$D9,$F9,$D6,$C6,$4F,$CF,$46,$5A,$7F,$8E,$3E,$A3,$C7,$AF
	.dc.b	$E0,$5B,$C1,$FC,$79,$FC,$74,$EF,$4F,$F1,$7C,$70,$AF,$F4,$4B,$5F
	.dc.b	$E7,$AD,$7F,$D8,$ED,$5F,$53,$DA,$BF,$A1,$B6,$37,$5F,$E3,$C8,$C4
	.dc.b	$70,$8C,$1E,$22,$CF,$AF,$87,$CA,$79,$BC,$5B,$19,$25,$7B,$E3,$2F
	.dc.b	$C1,$F9,$0F,$E8,$7F,$0B,$F6,$63,$D3,$18,$C3,$DB,$85,$9C,$E1,$FE
	.dc.b	$9E,$9F,$83,$59,$96,$48,$C3,$FD,$C6,$3F,$D4,$73,$C2,$B9,$47,$43
	.dc.b	$25,$C3,$C8,$F4,$C3,$9E,$12,$B8,$CF,$82,$3B,$CE,$63,$99,$99,$DE
	.dc.b	$1E,$56,$CE,$EF,$BC,$E6,$B8,$B6,$33,$3B,$C3,$D5,$5A,$C3,$C3,$87
	.dc.b	$67,$51,$EB,$9F,$97,$39,$24,$71,$7C,$66,$2F,$C7,$5C,$A7,$E1,$4E
	.dc.b	$6B,$C3,$5A,$8E,$33,$CC,$91,$97,$D7,$7C,$F0,$F6,$5F,$23,$CB,$3C
	.dc.b	$1C,$3E,$93,$F3,$B3,$C7,$0E,$4D,$C3,$C9,$3A,$6B,$F4,$4E,$86,$4A
	.dc.b	$67,$0F,$31,$F3,$A3,$8E,$1E,$69,$F1,$1F,$08,$E2,$E5,$E5,$87,$AA
	.dc.b	$B9,$BE,$A8,$C9,$E1,$8C,$BD,$27,$91,$F9,$78,$CC,$70,$24,$66,$7A
	.dc.b	$3C,$30,$FC,$0C,$F9,$1F,$8B,$7E,$A6,$C6,$63,$DB,$87,$C9,$8F,$5B
	.dc.b	$70,$F0,$DF,$79,$F3,$5F,$29,$CB,$70,$F1,$B6,$77,$87,$63,$C4,$57
	.dc.b	$C4,$78,$1C,$B4,$CD,$78,$B0,$F2,$5F,$C0,$7F,$15,$F5,$3F,$BE,$2D
	.dc.b	$FE,$B9,$C3,$F6,$9F,$E0,$DF,$50,$F0,$3F,$D7,$F8,$FE,$DA,$31,$CB
	.dc.b	$FB,$6F,$84,$DD,$6B,$3F,$E0,$74,$FA,$97,$C1,$FF,$9B,$B1,$FF,$37
	.dc.b	$CE,$63,$10,$11,$D9,$CB,$B7,$7F,$52,$FA,$9F,$F3,$CE,$1E,$0C,$3B
	.dc.b	$D7,$9D,$B1,$FC,$5A,$B8,$4F,$F1,$46,$BB,$C3,$F8,$BE,$1D,$EB,$86
	.dc.b	$1D,$B7,$FF,$43,$FB,$BF,$E2,$2C,$3D,$1C,$BF,$E7,$28,$7D,$4F,$FA
	.dc.b	$32,$30,$DE,$1E,$3E,$2D,$BF,$81,$7C,$DF,$F1,$03,$C9,$EB,$3E,$0A
	.dc.b	$E1,$1F,$EB,$2A,$B5,$FE,$97,$8D,$9D,$3E,$87,$26,$BD,$99,$1A,$3F
	.dc.b	$D7,$71,$A7,$13,$CD,$FF,$19,$7D,$4B,$EA,$7B,$FA,$12,$79,$4E,$5B
	.dc.b	$39,$7D,$73,$E8,$9D,$47,$49,$C1,$1F,$0F,$C4,$7B,$BE,$A5,$CF,$7E
	.dc.b	$06,$57,$A6,$72,$D9,$CB,$FE,$49,$84,$F3,$9E,$9F,$29,$2B,$93,$DF
	.dc.b	$E1,$F1,$1C,$62,$F1,$7F,$AB,$B0,$57,$D2,$FA,$97,$97,$E7,$F4,$66
	.dc.b	$3F,$36,$72,$D9,$CB,$E5,$9E,$51,$23,$44,$3D,$5D,$5F,$80,$FA,$97
	.dc.b	$D4,$D2,$4F,$DB,$87,$CD,$62,$3F,$CD,$51,$C1,$D4,$F3,$9E,$9F,$3B
	.dc.b	$D7,$E9,$A5,$73,$8C,$BD,$63,$2F,$48,$FC,$9F,$85,$FE,$02,$DF,$FD
	.dc.b	$A7,$F5,$0F,$A9,$FC,$BB,$6F,$F2,$15,$BE,$EF,$F7,$8B,$63,$F8,$14
	.dc.b	$ED,$FF,$41,$36,$FF,$DB,$F3,$B9,$FE,$2D,$DF,$E2,$23,$F9,$DA,$FD
	.dc.b	$65,$79,$3B,$7C,$64,$9D,$76,$D7,$BC,$59,$F5,$CF,$AB,$A7,$D7,$F2
	.dc.b	$32,$FB,$46,$5F,$04,$65,$F4,$C7,$C6,$EA,$FD,$CF,$AE,$FA,$9A,$66
	.dc.b	$BE,$2C,$E1,$FD,$B4,$ED,$C9,$0F,$F5,$36,$BF,$9E,$71,$93,$6F,$E5
	.dc.b	$91,$FE,$A4,$EC,$FE,$13,$F4,$EE,$3F,$E1,$7F,$52,$FA,$9A,$6A,$37
	.dc.b	$E9,$F0,$FA,$99,$8F,$F0,$9C,$3F,$DC,$EF,$F0,$BF,$A9,$FD,$2B,$9A
	.dc.b	$E2,$E7,$C0,$FE,$33,$2F,$8A,$FE,$37,$F5,$2F,$88,$A6,$A3,$F8,$FF
	.dc.b	$D4,$B9,$B1,$8F,$79,$CC,$48,$FE,$39,$3B,$6C,$7F,$28,$FA,$97,$CA
	.dc.b	$FF,$2A,$F5,$BF,$4F,$5F,$C8,$E7,$FC,$C8,$DF,$5F,$5E,$14,$8C,$10
	.dc.b	$FF,$6B,$57,$26,$C6,$DA,$75,$F1,$7F,$9C,$FC,$37,$8A,$3F,$99,$7D
	.dc.b	$23,$E0,$7F,$9E,$B6,$3F,$9B,$4E,$DF,$F7,$23,$7A,$FE,$F4,$7F,$69
	.dc.b	$9E,$0F,$FB,$A5,$B1,$A8,$4F,$1E,$BF,$C2,$B6,$3F,$B9,$35,$DE,$BE
	.dc.b	$4F,$8A,$FE,$A3,$E3,$BD,$DF,$EA,$7F,$4F,$F6,$6F,$FA,$63,$5F,$F7
	.dc.b	$D4,$ED,$FE,$A0,$D7,$FD,$FD,$1B,$CB,$CC,$DC,$BF,$33,$5B,$9F,$53
	.dc.b	$63,$37,$E4,$3C,$7F,$1D,$FD,$9B,$EA,$5F,$47,$E6,$FA,$99,$9F,$F6
	.dc.b	$19,$CB,$67,$2F,$92,$79,$44,$8D,$10,$F4,$F9,$5F,$DD,$3E,$A7,$F0
	.dc.b	$5A,$7A,$BD,$9F,$8A,$67,$96,$5F,$C6,$8C,$BF,$DC,$EB,$8E,$BF,$B8
	.dc.b	$6C,$7F,$89,$5A,$9B,$BB,$E8,$3C,$5F,$0D,$F5,$3D,$B9,$7D,$63,$2F
	.dc.b	$D6,$8C,$AC,$FF,$0D,$82,$BF,$C1,$FF,$52,$E7,$BD,$53,$0F,$F0,$54
	.dc.b	$E5,$B3,$C0,$8F,$F8,$3F,$EA,$5E,$DF,$78,$7D,$43,$EA,$7F,$8B,$D6
	.dc.b	$B4,$FF,$22,$4E,$6B,$FC,$8D,$8F,$77,$8D,$FF,$1E,$8C,$4E,$E5,$7C
	.dc.b	$DD,$9F,$E4,$F7,$11,$B9,$E3,$5C,$F1,$FE,$53,$3C,$EB,$73,$E6,$C7
	.dc.b	$D5,$3C,$6B,$73,$C7,$0F,$4B,$BC,$6C,$78,$6F,$8C,$EE,$3C,$A3,$1F
	.dc.b	$1B,$BA,$E2,$FB,$2B,$8B,$F1,$2F,$C2,$D3,$78,$FA,$1E,$47,$6B,$EA
	.dc.b	$F7,$9E,$2F,$5B,$E3,$7D,$A3,$47,$8A,$B8,$DF,$58,$E0,$7F,$9F,$1F
	.dc.b	$A0,$FF,$A0,$5A,$FC,$33,$5E,$3B,$E4,$7F,$D0,$71,$F6,$23,$0E,$A3
	.dc.b	$33,$FE,$32,$77,$1F,$E8,$9F,$84,$A9,$A4,$76,$FC,$AF,$90,$7D,$37
	.dc.b	$F2,$AD,$BF,$11,$5C,$CF,$A7,$3A,$E5,$FA,$31,$99,$FC,$74,$E6,$3E
	.dc.b	$C9,$9D,$FF,$B8,$C6,$67,$AC,$F9,$4E,$B7,$E1,$AF,$F5,$1E,$1E,$58
	.dc.b	$C4,$FF,$AA,$63,$18,$78,$A3,$95,$F8,$AB,$71,$FE,$A4,$3A,$CE,$EB
	.dc.b	$8C,$79,$8E,$37,$E6,$AE,$31,$E9,$3F,$D6,$98,$7A,$67,$CB,$1D,$0E
	.dc.b	$55,$D3,$0F,$16,$1F,$E8,$DB,$F8,$9B,$7F,$51,$BE,$82,$33,$3E,$13
	.dc.b	$86,$3F,$D7,$2B,$85,$79,$8F,$1C,$66,$7E,$6F,$4D,$9F,$3B,$7E,$29
	.dc.b	$F5,$1F,$E8,$2A,$F2,$4E,$EB,$C4,$F1,$C1,$F5,$2E,$FD,$67,$FD,$2B
	.dc.b	$0F,$F5,$6D,$6A,$72,$48,$FF,$58,$E5,$FE,$98,$F1,$71,$97,$F2,$D9
	.dc.b	$FC,$1D,$62,$FF,$BC,$DE,$E7,$E4,$4E,$5F,$F7,$24,$7A,$BE,$A7,$16
	.dc.b	$8C,$4F,$FA,$EF,$C7,$FE,$70,$DE,$9F,$88,$BE,$9F,$83,$92,$5F,$94
	.dc.b	$FF,$79,$57,$27,$11,$C6,$FE,$44,$E7,$2F,$F0,$63,$67,$FD,$F4,$77
	.dc.b	$E0,$DE,$3E,$2C,$7E,$FE,$3F,$EF,$FF,$A9,$7B,$BF,$4D,$B8,$7F,$B6
	.dc.b	$8F,$47,$9B,$99,$91,$FE,$D9,$9D,$D7,$9A,$FF,$C8,$D3,$9A,$DC,$E5
	.dc.b	$B9,$FF,$8A,$27,$E2,$47,$E0,$DF,$F5,$E9,$FE,$E5,$2C,$F5,$F0,$35
	.dc.b	$D5,$27,$4F,$B6,$31,$3F,$2C,$FF,$39,$47,$D2,$FA,$9F,$F2,$8D,$70
	.dc.b	$3D,$CE,$27,$87,$CD,$FE,$46,$7F,$DE,$CD,$8F,$74,$FE,$66,$31,$8F
	.dc.b	$F8,$F5,$B5,$FE,$80,$6D,$7F,$A0,$F2,$5F,$1E,$29,$7F,$3E,$5E,$26
	.dc.b	$2F,$E6,$3F,$A7,$FA,$87,$D4,$FE,$A7,$EA,$1E,$26,$63,$0C,$C5,$87
	.dc.b	$C3,$AD,$C6,$35,$FF,$2A,$61,$D2,$30,$7F,$A6,$27,$75,$C7,$E0,$7C
	.dc.b	$07,$57,$13,$C1,$EF,$3F,$2E,$FF,$CC,$0D,$9F,$F5,$1F,$D4,$F7,$AF
	.dc.b	$27,$A3,$FC,$0F,$A5,$E6,$53,$DD,$16,$FB,$53,$EA,$5E,$97,$F0,$B6
	.dc.b	$CE,$B0,$FF,$4E,$3C,$23,$35,$F8,$5F,$A7,$FF,$51,$4F,$46,$C7,$EC
	.dc.b	$FE,$A5,$F5,$0F,$A9,$FE,$31,$5A,$DF,$FC,$F8,$DC,$3F,$D7,$ED,$C1
	.dc.b	$87,$2B,$E3,$ED,$CB,$F9,$1E,$23,$EA,$7F,$CF,$91,$E3,$9F,$E5,$93
	.dc.b	$9F,$5F,$F9,$77,$67,$DC,$FA,$9B,$9F,$E3,$A3,$18,$7D,$37,$F0,$3D
	.dc.b	$9A,$AD,$B6,$75,$5C,$47,$D4,$F4,$3A,$FB,$F6,$77,$16,$7F,$8C,$61
	.dc.b	$F8,$EA,$C7,$07,$FA,$7F,$E9,$DE,$BF,$F3,$F6,$E3,$FC,$DB,$EA,$5F
	.dc.b	$09,$9A,$93,$F7,$AB,$79,$7F,$25,$E7,$61,$C2,$B5,$7C,$2B,$53,$FC
	.dc.b	$8D,$FD,$0D,$F2,$C7,$F9,$FC,$E7,$3F,$F1,$06,$1F,$33,$D4,$FC,$2D
	.dc.b	$7F,$AF,$27,$33,$F2,$E7,$2F,$EB,$23,$38,$7C,$29,$CA,$8C,$D7,$39
	.dc.b	$CC,$7F,$8B,$FE,$A7,$C0,$78,$7E,$32,$E1,$DE,$2E,$1D,$66,$E1,$F3
	.dc.b	$7E,$6F,$E6,$1F,$86,$F8,$27,$CB,$E3,$1F,$95,$37,$0F,$E9,$DC,$DF
	.dc.b	$ED,$2A,$C6,$1C,$EF,$FB,$6F,$D4,$FC,$DB,$FE,$55,$BB,$1F,$EE,$D3
	.dc.b	$9B,$F6,$7C,$87,$C9,$3E,$57,$D4,$DC,$7E,$44,$5C,$7E,$34,$DA,$F8
	.dc.b	$9F,$33,$F4,$7D,$7F,$A4,$DC,$3E,$83,$6F,$E7,$36,$BF,$DD,$9F,$53
	.dc.b	$FE,$F9,$AF,$E0,$AD,$AF,$7E,$FF,$C6,$AB,$78,$7F,$84,$6B,$51,$8B
	.dc.b	$FD,$9D,$6D,$FE,$53,$6A,$DC,$61,$B1,$FE,$03,$FA,$97,$CC,$75,$C7
	.dc.b	$FC,$45,$18,$BF,$C5,$7D,$4F,$E3,$FE,$A7,$F2,$0D,$8C,$9F,$8F,$8C
	.dc.b	$B4,$FC,$97,$5F,$E5,$FD,$9F,$CC,$36,$32,$7E,$5E,$32,$D3,$F9,$4F
	.dc.b	$D4,$78,$F3,$F6,$9F,$D1,$9C,$38,$6B,$F0,$9A,$EF,$3F,$F1,$D7,$D4
	.dc.b	$F3,$DD,$7B,$9C,$91,$FE,$73,$3B,$74,$62,$35,$5E,$41,$3E,$06,$C7
	.dc.b	$F9,$51,$B3,$FE,$98,$BF,$EA,$F1,$E9,$F4,$B5,$18,$8F,$54,$60,$F3
	.dc.b	$BF,$E9,$7A,$F7,$7C,$7E,$3E,$13,$ED,$CF,$C5,$AC,$56,$B0,$FF,$50
	.dc.b	$36,$79,$1F,$2E,$F6,$F1,$9D,$B9,$6A,$C1,$3C,$23,$0D,$8F,$EE,$78
	.dc.b	$79,$27,$35,$CA,$78,$B6,$B8,$61,$C5,$91,$93,$84,$62,$BD,$27,$42
	.dc.b	$D7,$46,$C6,$EF,$C5,$39,$6D,$F8,$2F,$B5,$F2,$8E,$18,$7D,$29,$CD
	.dc.b	$70,$6C,$66,$B8,$46,$1F,$90,$78,$9C,$3E,$79,$E4,$E4,$E2,$47,$27
	.dc.b	$A3,$0F,$1C,$74,$8C,$D7,$85,$B1,$98,$FB,$D1,$C6,$BC,$38,$7C,$19
	.dc.b	$C9,$C4,$B5,$C4,$87,$DF,$BF,$0B,$E9,$8D,$B9,$AE,$B8,$6A,$FF,$03
	.dc.b	$5A,$7C,$15,$D1,$CB,$67,$FC,$DC,$FE,$56,$BF,$79,$18,$8F,$97,$18
	.dc.b	$3F,$D1,$2E,$C8,$68,$F0,$7D,$4F,$8C,$91,$F0,$9B,$1F,$3E,$FF,$84
	.dc.b	$46,$31,$F9,$73,$F9,$F8,$C4,$F4,$AE,$44,$9E,$47,$F1,$59,$DB,$EF
	.dc.b	$6B,$F2,$4D,$47,$1A,$EF,$82,$B6,$FC,$8A,$F1,$56,$EF,$C5,$1E,$03
	.dc.b	$25,$9E,$53,$B8,$F1,$35,$F1,$1C,$2F,$C3,$7E,$0C,$3F,$74,$DA,$FD
	.dc.b	$C3,$6B,$F5,$EE,$A3,$27,$5B,$F0,$CF,$FB,$ED,$B1,$FC,$DE,$7E,$E4
	.dc.b	$FA,$23,$F9,$EF,$D3,$BE,$A5,$F0,$5F,$CB,$69,$39,$8C,$2E,$FF,$F6
	.dc.b	$F4,$FF,$33,$8D,$36,$B9,$BF,$C2,$23,$39,$FF,$15,$6E,$9F,$DA,$BE
	.dc.b	$A5,$F5,$3F,$DB,$2B,$F8,$D4,$62,$FF,$91,$DF,$07,$FD,$32,$D5,$1B
	.dc.b	$7F,$8B,$B7,$3F,$EB,$6D,$8F,$05,$FF,$7A,$9C,$B6,$3F,$BE,$FB,$9F
	.dc.b	$E1,$D7,$FD,$1A,$BE,$A7,$FC,$03,$C7,$FD,$1A,$7F,$D6,$B1,$8B,$E3
	.dc.b	$E0,$59,$FF,$57,$DF,$37,$FD,$A4,$F3,$7A,$D6,$DF,$F5,$FC,$D9,$FF
	.dc.b	$50,$C7,$8B,$E8,$70,$67,$D9,$91,$92,$1D,$9E,$9F,$52,$FA,$9E,$9E
	.dc.b	$64,$AF,$F5,$4C,$E5,$B3,$97,$CF,$3C,$A2,$46,$88,$7B,$3E,$1F,$D5
	.dc.b	$F5,$2F,$A3,$BF,$03,$2B,$FD,$6F,$39,$6C,$E5,$FF,$7B,$42,$79,$CF
	.dc.b	$4F,$93,$95,$C9,$EF,$F0,$F8,$8E,$31,$78,$BE,$B6,$0A,$F9,$DF,$52
	.dc.b	$F2,$7C,$FE,$74,$C7,$FD,$97,$39,$6C,$E5,$F2,$CF,$28,$91,$A2,$1E
	.dc.b	$AE,$DF,$B9,$F5,$2F,$A9,$A4,$9F,$F6,$CC,$3E,$6B,$10,$D3,$F1,$D3
	.dc.b	$C6,$37,$5D,$2B,$AF,$CE,$F6,$FA,$49,$5D,$23,$2F,$68,$CB,$D7,$AF
	.dc.b	$FC,$B7,$82,$BF,$25,$E1,$7F,$82,$F2,$FC,$F7,$D4,$3E,$A7,$F2,$CD
	.dc.b	$C7,$FD,$E3,$5B,$C7,$FC,$CA,$D8,$09,$E0,$F8,$4C,$B8,$BF,$D6,$4F
	.dc.b	$19,$FE,$21,$E1,$F9,$08,$FF,$C0,$0A,$FD,$65,$79,$BA,$FC,$A4,$9D
	.dc.b	$76,$57,$C0,$78,$E3,$13,$EE,$9F,$6F,$4F,$BB,$E4,$65,$F7,$8C,$BE
	.dc.b	$18,$CB,$D3,$04,$FC,$AE,$AF,$DE,$7A,$8F,$A9,$A6,$6F,$E4,$CE,$1F
	.dc.b	$8F,$97,$F0,$DA,$E1,$1A,$21,$EA,$D7,$F4,$CE,$27,$87,$67,$F0,$BF
	.dc.b	$A6,$72,$7F,$0D,$FA,$97,$D4,$D3,$51,$BF,$47,$87,$D4,$CC,$BC,$7C
	.dc.b	$BF,$EA,$0C,$7F,$DE,$B1,$83,$EE,$FD,$4F,$EB,$5C,$08,$DB,$F4,$1B
	.dc.b	$A7,$F2,$99,$E1,$3F,$51,$7C,$47,$F2,$3F,$A9,$7C,$2D,$3F,$93,$4E
	.dc.b	$AB,$BF,$D4,$FD,$07,$C3,$19,$7A,$5F,$F2,$D9,$D4,$7E,$09,$B8,$7F
	.dc.b	$3B,$6E,$5F,$EB,$E6,$CF,$E6,$BE,$A5,$F2,$7F,$CD,$FD,$7F,$DC,$4F
	.dc.b	$C0,$9F,$F3,$BB,$7D,$7D,$78,$52,$30,$43,$F2,$75,$C9,$B1,$B6,$9D
	.dc.b	$7C,$4F,$E8,$DE,$FF,$8A,$3E,$7F,$D2,$BD,$FF,$E9,$4D,$8F,$A7,$3B
	.dc.b	$7F,$42,$DD,$BF,$AD,$4E,$E7,$F4,$AD,$8F,$EB,$D3,$BD,$BF,$DC,$2D
	.dc.b	$AF,$DE,$B5,$36,$9F,$29,$FD,$77,$C6,$7C,$77,$F5,$FF,$A7,$FB,$B8
	.dc.b	$7E,$21,$AF,$ED,$A7,$6F,$E3,$1A,$FE,$EA,$37,$97,$F7,$36,$E5,$FE
	.dc.b	$F6,$AD,$CF,$F7,$76,$C6,$6F,$C8,$78,$FE,$3B,$FB,$97,$D4,$BE,$8B
	.dc.b	$CD,$F5,$33,$3F,$CE,$CE,$5B,$39,$7C,$93,$CA,$24,$68,$87,$A7,$CA
	.dc.b	$FF,$00,$7D,$4F,$E0,$74,$F5,$7B,$3F,$14,$CF,$24,$BF,$8D,$19,$7F
	.dc.b	$5F,$5C,$75,$FE,$5F,$08,$C3,$FC,$C9,$AE,$6B,$E9,$7D,$07,$8B,$E1
	.dc.b	$7E,$A7,$B5,$2B,$AC,$65,$FA,$D1,$97,$F9,$5D,$A7,$FC,$35,$F5,$2E
	.dc.b	$5B,$D3,$2F,$F8,$2C,$E5,$B3,$C0,$8F,$F0,$AF,$A9,$7B,$7F,$D8,$E1
	.dc.b	$F5,$0F,$A9,$F7,$D6,$B8,$7F,$A7,$CE,$6B,$FA,$8E,$3D,$DE,$37,$F1
	.dc.b	$A3,$13,$B9,$5F,$37,$67,$F5,$A7,$11,$B9,$E3,$5C,$F1,$FA,$33,$CE
	.dc.b	$B7,$3E,$6C,$7D,$53,$C6,$B7,$3C,$70,$F4,$BB,$C6,$C7,$86,$F8,$CE
	.dc.b	$E3,$CA,$31,$F1,$BB,$AE,$2F,$B2,$B8,$BF,$12,$FC,$2D,$37,$8F,$A1
	.dc.b	$E4,$76,$BE,$AF,$79,$E2,$F5,$BE,$37,$DA,$34,$78,$AB,$8D,$F5,$8E
	.dc.b	$07,$E6,$0E,$E5,$7E,$93,$67,$F0,$4E,$7D,$C7,$CA,$7F,$A6,$6B,$FC
	.dc.b	$7C,$D8,$CC,$F5,$3E,$EB,$87,$FD,$41,$F0,$DF,$EA,$17,$FC,$2E,$D4
	.dc.b	$D2,$3F,$43,$E4,$9F,$4C,$E5,$F9,$F3,$C8,$D4,$65,$FA,$E6,$77,$FC
	.dc.b	$85,$72,$AE,$33,$E0,$38,$B8,$BF,$B7,$3B,$9E,$66,$EB,$9D,$F7,$7F
	.dc.b	$39,$1D,$F7,$FE,$0B,$5C,$67,$33,$C6,$73,$5C,$5F,$91,$7E,$08,$C4
	.dc.b	$FD,$78,$C6,$1E,$58,$F3,$9E,$48,$CB,$4C,$DF,$82,$32,$FA,$27,$FC
	.dc.b	$13,$87,$B7,$6F,$F5,$75,$FA,$4F,$F2,$5C,$66,$7C,$E7,$0C,$3E,$75
	.dc.b	$70,$AF,$61,$C2,$33,$3F,$0B,$A2,$CF,$46,$FC,$BB,$EA,$7F,$D2,$55
	.dc.b	$E4,$9D,$D3,$1F,$E8,$D3,$FE,$23,$C5,$F5,$2E,$ED,$D6,$AF,$FC,$C1
	.dc.b	$5A,$9C,$92,$FF,$C5,$99,$7F,$AA,$7C,$4C,$5F,$F9,$63,$0E,$95,$F1
	.dc.b	$AF,$CF,$EA,$7F,$80,$63,$04,$AF,$F3,$76,$1E,$BF,$A9,$E7,$97,$E4
	.dc.b	$E3,$15,$FE,$38,$F1,$FF,$9D,$37,$AB,$E2,$AF,$A7,$57,$C2,$AD,$E3
	.dc.b	$FE,$6C,$6E,$1F,$EE,$56,$DF,$FA,$08,$E5,$5C,$70,$91,$FE,$4B,$3B
	.dc.b	$70,$E3,$0F,$F4,$34,$60,$FF,$26,$FD,$4B,$DE,$FA,$6D,$8E,$59,$FF
	.dc.b	$9C,$9D,$CE,$88,$EB,$DD,$B7,$FE,$84,$BE,$4F,$82,$EE,$9F,$EF,$F9
	.dc.b	$B7,$EB,$9B,$5E,$DE,$8F,$C2,$CF,$FB,$3C,$F1,$F8,$38,$89,$1F,$E6
	.dc.b	$63,$53,$DE,$72,$FF,$A5,$A3,$38,$7F,$09,$5F,$53,$FE,$52,$79,$5F
	.dc.b	$D6,$8C,$57,$FA,$7B,$E7,$3F,$0E,$DE,$3F,$B7,$97,$2C,$3F,$CE,$F5
	.dc.b	$AC,$7F,$8B,$5A,$31,$3F,$E0,$C6,$C7,$F1,$FA,$E4,$F5,$9E,$77,$DB
	.dc.b	$C5,$FD,$3D,$F0,$9F,$C5,$47,$EA,$BE,$A1,$F5,$3F,$AC,$FA,$97,$8B
	.dc.b	$98,$C3,$2F,$FD,$1B,$3A,$7F,$D1,$F5,$C9,$CC,$F3,$F8,$19,$18,$38
	.dc.b	$BD,$FD,$49,$1F,$E9,$8F,$A9,$E5,$AF,$FB,$1B,$1F,$97,$BF,$C4,$E8
	.dc.b	$79,$9F,$C0,$9B,$87,$93,$1C,$4F,$F0,$6F,$A9,$FF,$30,$CF,$FA,$91
	.dc.b	$B8,$FF,$B4,$AF,$FD,$3B,$39,$F4,$FF,$86,$FD,$47,$FA,$8A,$3F,$D5
	.dc.b	$6D,$AF,$F7,$47,$D4,$3E,$A5,$F5,$3C,$3D,$C9,$87,$07,$F8,$ED,$6E
	.dc.b	$73,$7F,$EE,$2C,$3A,$6D,$FD,$57,$C4,$7D,$4F,$F9,$DB,$2E,$47,$FA
	.dc.b	$FE,$33,$EB,$FF,$28,$E9,$FF,$7E,$FD,$4D,$CB,$F0,$B1,$8B,$FF,$63
	.dc.b	$28,$C8,$9E,$1D,$7C,$5B,$5C,$25,$F5,$3D,$4E,$DF,$03,$67,$51,$65
	.dc.b	$A7,$F6,$FC,$7C,$17,$98,$FC,$EF,$D3,$BD,$8F,$E7,$6D,$C7,$F9,$8F
	.dc.b	$D4,$BE,$1F,$35,$27,$FD,$BB,$5B,$CB,$F9,$6D,$EE,$73,$5B,$9C,$F9
	.dc.b	$57,$1F,$E6,$17,$FC,$92,$7F,$5D,$EA,$FF,$82,$70,$F8,$EF,$E1,$AB
	.dc.b	$D4,$E6,$7D,$B3,$F2,$F4,$FF,$10,$F4,$7D,$3A,$D5,$F0,$8C,$56,$5D
	.dc.b	$7D,$4C,$8E,$FF,$19,$71,$EA,$79,$27,$17,$E2,$9B,$5F,$E8,$ED,$5F
	.dc.b	$37,$F2,$8F,$ED,$7C,$BF,$EC,$2F,$97,$C5,$7C,$89,$B9,$76,$AD,$4E
	.dc.b	$49,$1F,$D9,$BE,$A7,$E6,$3E,$41,$87,$82,$7F,$B8,$4E,$B2,$F8,$FF
	.dc.b	$21,$F0,$CF,$89,$F5,$36,$BE,$6C,$5C,$3D,$FC,$1F,$E4,$AA,$C5,$FD
	.dc.b	$0F,$99,$F9,$9E,$BF,$CD,$6E,$3F,$59,$B8,$72,$E2,$46,$2F,$F9,$5F
	.dc.b	$D4,$FF,$75,$AF,$53,$63,$F6,$1F,$13,$FC,$6E,$73,$8F,$F8,$3A,$B5
	.dc.b	$8F,$FA,$F1,$B3,$FE,$11,$AE,$17,$DA,$F8,$1F,$E1,$AF,$A9,$7C,$C6
	.dc.b	$31,$FF,$06,$C6,$1F,$C7,$FD,$4F,$E2,$7E,$A7,$F1,$4D,$8C,$9F,$94
	.dc.b	$8C,$B4,$FC,$BF,$67,$E4,$7A,$FF,$24,$D8,$C9,$F9,$C8,$CB,$4F,$E0
	.dc.b	$99,$7F,$50,$FA,$8F,$E2,$1A,$FF,$9D,$DA,$FF,$A1,$5A,$FF,$B7,$9A
	.dc.b	$FF,$0D,$78,$C7,$3F,$43,$FC,$6F,$18,$AE,$11,$82,$3E,$23,$A7,$87
	.dc.b	$8B,$FC,$85,$7D,$57,$C5,$5F,$E4,$D5,$19,$7A,$D7,$FA,$3D,$CF,$E1
	.dc.b	$1F,$DE,$5F,$43,$5F,$43,$5D,$C7,$F8,$9D,$DB,$FE,$94,$DB,$E0,$7B
	.dc.b	$FF,$E8,$BF,$84,$FB,$B5,$CA,$B4,$FC,$39,$D5,$70,$9C,$CF,$A2,$72
	.dc.b	$F1,$3F,$D3,$F3,$B6,$9C,$E7,$8E,$0A,$E3,$3E,$83,$67,$59,$DC,$6A
	.dc.b	$78,$47,$2C,$31,$32,$36,$64,$B5,$D2,$B8,$B6,$7A,$57,$18,$C5,$70
	.dc.b	$C3,$E2,$C7,$95,$B7,$C9,$B7,$C5,$CD,$DC,$3E,$14,$6B,$0F,$26,$1D
	.dc.b	$89,$19,$7C,$D5,$C5,$F9,$06,$AB,$ED,$B4,$CC,$F1,$3C,$35,$E9,$9F
	.dc.b	$09,$27,$EB,$4E,$E3,$11,$F3,$CF,$0C,$F1,$7A,$C6,$B0,$5F,$96,$78
	.dc.b	$5F,$37,$AD,$6F,$0D,$5F,$0C,$3A,$57,$0A,$E9,$86,$AF,$6D,$AF,$CD
	.dc.b	$57,$C2,$7F,$75,$3F,$9F,$BF,$B7,$18,$37,$3E,$B7,$0F,$FB,$0A,$36
	.dc.b	$FC,$0F,$A7,$77,$FF,$A5,$23,$E6,$08,$CF,$9D,$E8,$6C,$69,$B9,$FC
	.dc.b	$18,$E1,$5C,$E3,$13,$AC,$FC,$13,$96,$DF,$58,$C5,$FD,$03,$FC,$41
	.dc.b	$5C,$5C,$1D,$A7,$81,$1F,$24,$F2,$7E,$71,$C9,$C6,$1F,$99,$BF,$F1
	.dc.b	$64,$F3,$38,$D7,$3C,$3D,$37,$C3,$2F,$DB,$E3,$E3,$BE,$A7,$1A,$DD
	.dc.b	$F1,$C3,$DB,$3D,$09,$3A,$6C,$6A,$30,$47,$04,$8F,$83,$5E,$1C,$7D
	.dc.b	$51,$ED,$BF,$03,$4F,$E1,$73,$BC,$3D,$38,$FC,$58,$7D,$43,$E0,$7F
	.dc.b	$98,$46,$1B,$3F,$85,$BF,$8E,$7E,$16,$B8,$38,$1F,$50,$F9,$DF,$9B
	.dc.b	$B7,$08,$C4,$FE,$59,$AF,$FA,$A0,$E3,$3C,$1F,$AB,$1A,$DB,$F8,$9B
	.dc.b	$76,$CD,$7D,$A9,$CF,$C2,$59,$E2,$DB,$FF,$35,$5F,$16,$DE,$5F,$4C
	.dc.b	$D9,$3E,$38,$E2,$D7,$67,$AD,$C7,$B9,$FE,$3C,$8F,$EC,$87,$F9,$E5
	.dc.b	$7D,$3F,$E4,$39,$7F,$D8,$F1,$89,$D8,$79,$3F,$E3,$96,$C7,$09,$E5
	.dc.b	$9F,$E6,$A7,$95,$6F,$C6,$CB,$E1,$F5,$EB,$D7,$E8,$2D,$1B,$32,$7C
	.dc.b	$73,$96,$CE,$5F,$F2,$67,$D3,$BE,$9F,$9F,$9F,$27,$D7,$39,$6C,$E5
	.dc.b	$F6,$4F,$92,$75,$1D,$27,$04,$7C,$7F,$0D,$EA,$FA,$7F,$ED,$51,$F2
	.dc.b	$17,$C0,$64,$D7,$4C,$9D,$46,$5E,$11,$CF,$E4,$24,$73,$7B,$7C,$36
	.dc.b	$23,$FC,$D5,$69,$C7,$17,$FC,$00,$85,$7C,$EF,$A7,$FF,$2D,$DF,$97
	.dc.b	$D2,$F4,$19,$F2,$E4,$64,$87,$F9,$DE,$78,$B6,$36,$D3,$E0,$76,$FD
	.dc.b	$CF,$A6,$7D,$3D,$24,$FE,$32,$1F,$2F,$88,$FF,$42,$47,$07,$53,$CE
	.dc.b	$7A,$7C,$D7,$5F,$A3,$95,$CE,$32,$F5,$8C,$BD,$23,$F1,$FE,$17,$F8
	.dc.b	$2B,$77,$5F,$F3,$0F,$A6,$7D,$3F,$E5,$1B,$87,$E8,$AF,$85,$7F,$36
	.dc.b	$6C,$7F,$B8,$E7,$6F,$FA,$75,$B7,$DA,$3F,$DD,$13,$C1,$FE,$BF,$E1
	.dc.b	$79,$08,$FE,$C2,$BF,$5B,$7E,$5E,$DF,$21,$27,$5D,$95,$F0,$45,$C7
	.dc.b	$C1,$0B,$FD,$43,$E5,$E9,$F7,$7C,$8C,$8F,$04,$7C,$13,$0F,$AE,$30
	.dc.b	$FC,$3E,$AF,$DC,$FA,$FF,$CA,$30,$EE,$DC,$FE,$34,$FD,$03,$7E,$24
	.dc.b	$BE,$44,$3F,$D8,$9B,$7E,$A1,$C6,$1D,$3B,$3F,$83,$FD,$23,$8F,$F8
	.dc.b	$4F,$D3,$BE,$9E,$9A,$8D,$FA,$3C,$3E,$9E,$65,$FC,$73,$04,$FF,$12
	.dc.b	$FA,$6F,$D3,$CE,$63,$8B,$FC,$8A,$32,$FF,$B6,$57,$C5,$FF,$1A,$FA
	.dc.b	$77,$C2,$D3,$A4,$7F,$1E,$FA,$77,$36,$30,$E9,$39,$89,$1F,$CC,$27
	.dc.b	$6D,$8F,$E4,$FF,$4E,$F9,$1F,$E5,$3E,$BF,$EA,$AB,$F9,$BC,$BE,$9F
	.dc.b	$66,$BC,$19,$3A,$8C,$BF,$CF,$AB,$93,$FE,$F5,$6F,$2F,$E2,$9B,$D3
	.dc.b	$FE,$16,$6C,$86,$1F,$DC,$9F,$F7,$F7,$BF,$E7,$7F,$A5,$FD,$1B,$E0
	.dc.b	$3F,$A0,$B6,$3F,$A7,$CE,$DE,$3A,$6A,$A4,$6A,$3F,$A4,$CE,$F2,$FC
	.dc.b	$FB,$72,$FF,$15,$B6,$BC,$D8,$3E,$27,$FA,$8F,$8A,$F7,$7F,$A9,$FD
	.dc.b	$3F,$DA,$BF,$EC,$2D,$59,$7E,$B2,$B8,$47,$F6,$66,$BA,$CB,$FC,$77
	.dc.b	$5C,$6B,$F6,$6D,$8F,$F1,$FD,$70,$9E,$97,$E4,$25,$F8,$BE,$3B,$FB
	.dc.b	$3F,$D3,$BE,$83,$CF,$F4,$F2,$BF,$BB,$CE,$5B,$39,$7C,$87,$D7,$CD
	.dc.b	$5D,$26,$4F,$02,$1C,$04,$BC,$8F,$EF,$3F,$4F,$F8,$29,$CF,$B3,$F2
	.dc.b	$8C,$F3,$C8,$F9,$71,$97,$FC,$1B,$3C,$49,$3B,$8F,$05,$BE,$7F,$BF
	.dc.b	$C2,$FD,$6F,$6E,$57,$28,$CB,$F6,$23,$2F,$A7,$04,$FF,$84,$3E,$9D
	.dc.b	$CB,$7A,$A5,$FF,$88,$27,$2D,$9E,$04,$7F,$C4,$BF,$4F,$FD,$C7,$93
	.dc.b	$F4,$5F,$4C,$FA,$7F,$F3,$DC,$E6,$93,$97,$F5,$18,$76,$73,$5F,$E8
	.dc.b	$28,$C4,$EE,$55,$CE,$30,$DA,$E1,$87,$FA,$2A,$78,$5F,$29,$F1,$FB
	.dc.b	$BF,$9A,$C3,$B9,$FB,$9C,$39,$3E,$4C,$39,$57,$48,$F6,$BC,$B1,$B1
	.dc.b	$C6,$FA,$63,$E0,$BE,$33,$CC,$D9,$EA,$9D,$DF,$38,$F1,$8C,$7D,$0F
	.dc.b	$13,$38,$F9,$9E,$58,$F4,$9D,$C7,$1C,$3C,$2E,$8E,$38,$78,$1E,$98
	.dc.b	$F9,$DF,$2E,$1D,$4F,$B0,$FF,$A3,$9A,$F9,$CC,$4F,$CB,$75,$EE,$3E
	.dc.b	$3B,$FD,$25,$1F,$62,$30,$FF,$A5,$6B,$FC,$6B,$3A,$6C,$70,$1F,$0B
	.dc.b	$5A,$65,$C1,$1F,$93,$F2,$6F,$A5,$FF,$08,$65,$F8,$99,$FE,$6D,$5F
	.dc.b	$4A,$B5,$DD,$F5,$27,$27,$2B,$E3,$BF,$FA,$AE,$F8,$46,$27,$FD,$63
	.dc.b	$18,$AE,$51,$E6,$C3,$F3,$18,$76,$8F,$BA,$76,$8C,$DF,$38,$FF,$54
	.dc.b	$1E,$1A,$E7,$7D,$63,$89,$D7,$0F,$F7,$3E,$1D,$70,$F7,$4E,$4F,$EA
	.dc.b	$91,$99,$F5,$47,$53,$9D,$75,$AE,$2D,$9E,$B8,$7F,$8E,$77,$FF,$15
	.dc.b	$E1,$EB,$7F,$D8,$D8,$77,$11,$9C,$78,$9F,$EA,$A7,$11,$F2,$0F,$14
	.dc.b	$F4,$E9,$ED,$5E,$C6,$CB,$E2,$9F,$4F,$FE,$8C,$AE,$95,$AA,$F0,$BC
	.dc.b	$2D,$F4,$EE,$FD,$65,$FD,$27,$0F,$F6,$1D,$6A,$72,$48,$FF,$63,$E4
	.dc.b	$F5,$3F,$BC,$47,$AE,$75,$97,$F4,$5C,$3B,$4F,$0A,$DC,$E6,$B7,$87
	.dc.b	$F9,$8F,$0F,$5F,$C4,$E3,$0F,$F0,$54,$61,$FF,$68,$FC,$0F,$F3,$86
	.dc.b	$F4,$FC,$45,$F4,$BF,$15,$C9,$32,$FE,$B1,$5C,$5F,$EC,$31,$C2,$FE
	.dc.b	$44,$E7,$1F,$F3,$3B,$67,$FC,$38,$76,$E0,$DE,$3E,$1C,$7E,$DE,$2F
	.dc.b	$A7,$7B,$BF,$4D,$B8,$7F,$BC,$0F,$3F,$9D,$99,$91,$FE,$EE,$9D,$D7
	.dc.b	$9A,$FF,$C8,$F3,$9A,$DC,$E5,$B9,$FF,$A3,$67,$E1,$C7,$E0,$DF,$F6
	.dc.b	$09,$FE,$FA,$2C,$F5,$F0,$35,$D5,$27,$4F,$B6,$31,$3F,$E0,$19,$E3
	.dc.b	$7D,$FE,$9F,$FC,$A5,$1C,$0F,$73,$89,$FF,$04,$7C,$BF,$E1,$1B,$C5
	.dc.b	$F2,$67,$F3,$13,$9B,$FF,$4B,$B6,$FF,$C9,$96,$8C,$3F,$EB,$8C,$D7
	.dc.b	$CB,$09,$7F,$3E,$5E,$2E,$2F,$E6,$3F,$A7,$FA,$67,$D3,$FE,$A7,$E9
	.dc.b	$9E,$26,$63,$0C,$C7,$FC,$BD,$AF,$FA,$B2,$B7,$18,$3C,$F7,$D1,$C6
	.dc.b	$3F,$E6,$59,$DB,$FE,$2B,$7F,$74,$DF,$81,$EC,$D9,$E2,$A7,$E6,$5F
	.dc.b	$87,$D1,$FE,$01,$F4,$FD,$EB,$A4,$7F,$03,$E9,$FF,$1B,$61,$C9,$B3
	.dc.b	$A8,$C7,$99,$FC,$27,$E9,$FF,$CC,$91,$E0,$6C,$FF,$90,$6B,$FD,$3B
	.dc.b	$39,$F4,$BF,$88,$7D,$27,$FA,$8A,$3C,$ED,$AF,$F2,$7F,$D3,$3E,$9D
	.dc.b	$F4,$FC,$3D,$C9,$87,$07,$F8,$FD,$6E,$73,$7F,$E5,$1C,$3A,$6F,$FE
	.dc.b	$EB,$F1,$1F,$4F,$FE,$79,$CB,$91,$F4,$23,$3E,$BF,$F2,$AE,$CF,$F4
	.dc.b	$B7,$D3,$DC,$BF,$17,$18,$BF,$AE,$FF,$A6,$FB,$35,$5B,$6C,$EA,$B8
	.dc.b	$8F,$A7,$E8,$75,$F8,$0B,$3B,$8B,$3F,$E7,$2C,$3F,$D5,$15,$8D,$3F
	.dc.b	$DE,$FF,$4A,$F5,$FF,$9E,$37,$1F,$E6,$7F,$4F,$FE,$D8,$7F,$0B,$19
	.dc.b	$C7,$FC,$EC,$DB,$E7,$7F,$D2,$B9,$98,$71,$BD,$CF,$63,$51,$FC,$86
	.dc.b	$3F,$D7,$97,$CF,$0F,$E6,$B3,$9C,$BF,$C3,$D1,$F3,$7D,$4F,$C3,$E4
	.dc.b	$8F,$F6,$24,$E9,$F9,$B3,$A7,$2E,$DC,$F3,$F3,$7A,$3D,$7E,$9E,$47
	.dc.b	$2F,$8C,$B9,$FC,$BB,$56,$31,$F1,$4F,$F0,$99,$56,$6F,$F8,$6E,$6F
	.dc.b	$9B,$F9,$E7,$F7,$0E,$47,$C0,$BE,$63,$37,$F4,$2A,$E5,$FE,$AC,$BD
	.dc.b	$D6,$33,$FB,$75,$FD,$CF,$E9,$FE,$84,$63,$0F,$EE,$31,$87,$E1,$FC
	.dc.b	$8F,$CB,$3E,$67,$D3,$D8,$F9,$31,$72,$F8,$13,$6F,$DF,$C3,$FE,$51
	.dc.b	$F9,$9F,$A9,$F0,$FF,$55,$B8,$FD,$26,$E1,$F2,$5B,$7F,$22,$1F,$4F
	.dc.b	$FD,$EA,$30,$DA,$FF,$16,$EF,$FC,$72,$73,$87,$F8,$4E,$B5,$63,$9C
	.dc.b	$7C,$DA,$E1,$7F,$CF,$AB,$75,$A8,$FF,$08,$FD,$3B,$E6,$3A,$61,$F9
	.dc.b	$B8,$C4,$FE,$33,$E9,$FF,$25,$F4,$FF,$93,$6C,$64,$FC,$8C,$65,$A7
	.dc.b	$E5,$3E,$4B,$F3,$5E,$CF,$E6,$DB,$19,$3F,$33,$19,$69,$FC,$DB,$E9
	.dc.b	$BF,$89,$67,$FE,$E7,$7D,$AF,$C1,$73,$B7,$CD,$73,$07,$33,$FD,$81
	.dc.b	$C9,$88,$E5,$E8,$FF,$8F,$E3,$2D,$F1,$30,$CA,$FF,$3F,$ED,$CF,$E4
	.dc.b	$3F,$47,$F0,$9F,$2F,$1E,$75,$A7,$CD,$3A,$AF,$17,$D3,$D9,$F5,$7C
	.dc.b	$0D,$7D,$11,$73,$FF,$23,$61,$CE,$72,$D8,$CC,$7F,$A7,$2F,$AC,$E6
	.dc.b	$F9,$1D,$7E,$9F,$F6,$FE,$2E,$67,$A4,$62,$F9,$1E,$48,$CF,$81,$FE
	.dc.b	$67,$8C,$9F,$36,$FA,$46,$74,$F8,$53,$9A,$E3,$87,$78,$E3,$87,$7A
	.dc.b	$DE,$5E,$A7,$8C,$67,$2F,$B1,$97,$6C,$39,$56,$B2,$F0,$56,$B2,$EF
	.dc.b	$97,$9B,$2E,$93,$FE,$20,$CB,$A4,$FB,$4F,$1D,$F2,$71,$87,$5C,$7B
	.dc.b	$DF,$1A,$D5,$7B,$4E,$33,$A7,$F6,$97,$D9,$D5,$F3,$7C,$99,$7A,$DB
	.dc.b	$97,$A2,$7E,$1C,$E5,$B1,$9C,$7D,$D5,$E0,$38,$96,$76,$D8,$DC,$78
	.dc.b	$4C,$95,$C9,$ED,$C3,$D5,$2C,$3D,$59,$7C,$A8,$F6,$E5,$F0,$A3,$11
	.dc.b	$B8,$C6,$5E,$9C,$3A,$CF,$0C,$3A,$E1,$D2,$B5,$3F,$42,$BC,$4F,$B3
	.dc.b	$1F,$33,$93,$91,$98,$DC,$7B,$1A,$78,$23,$C9,$7E,$0C,$3C,$B9,$7C
	.dc.b	$B9,$E8,$68,$E8,$76,$BE,$8D,$BE,$98,$FA,$6B,$94,$FD,$E9,$F0,$BD
	.dc.b	$27,$4F,$D3,$31,$7E,$27,$07,$13,$B1,$AC,$3D,$04,$7C,$A7,$9F,$0F
	.dc.b	$2B,$6F,$CB,$5D,$67,$E4,$CF,$53,$47,$53,$9D,$F9,$A3,$79,$7D,$C8
	.dc.b	$C5,$FB,$2B,$94,$61,$F3,$98,$BF,$04,$EE,$B9,$46,$27,$71,$8C,$3D
	.dc.b	$98,$7A,$B0,$E9,$5E,$63,$F2,$5F,$4E,$F8,$2F,$F4,$CC,$7E,$EE,$30
	.dc.b	$E0,$FB,$46,$A5,$1A,$8C,$3C,$2B,$F8,$7B,$72,$E9,$7F,$CB,$FE,$9D
	.dc.b	$F5,$1F,$EA,$8B,$FE,$3B,$0F,$9E,$C5,$FF,$AB,$23,$18,$F1,$9C,$BF
	.dc.b	$DD,$E1,$F0,$D8,$84,$70,$7F,$96,$B6,$FB,$D6,$E3,$8C,$F9,$6B,$13
	.dc.b	$23,$04,$39,$C7,$13,$C1,$F5,$CF,$86,$F1,$7C,$07,$F8,$DE,$BF,$D4
	.dc.b	$AB,$EA,$7F,$D8,$FD,$3F,$E1,$98,$C4,$EB,$C7,$CF,$2F,$FB,$32,$79
	.dc.b	$4A,$B8,$5E,$A3,$FC,$31,$3C,$9D,$57,$1B,$D4,$7F,$B3,$BD,$BB,$4F
	.dc.b	$A2,$A6,$22,$4F,$A2,$32,$FA,$A3,$9F,$D4,$BE,$A7,$9F,$A7,$2B,$FD
	.dc.b	$87,$39,$6C,$E5,$F2,$9E,$89,$E7,$12,$36,$43,$67,$FB,$9B,$5F,$5F
	.dc.b	$D4,$BB,$2F,$C1,$CA,$FF,$69,$CE,$5B,$39,$79,$C2,$79,$CF,$4F,$93
	.dc.b	$95,$C9,$EF,$F1,$58,$8E,$31,$78,$BF,$CA,$58,$2B,$E7,$FD,$4B,$CD
	.dc.b	$F3,$FA,$53,$1F,$F7,$3C,$E5,$B3,$97,$CB,$3C,$A2,$46,$88,$7A,$BB
	.dc.b	$7E,$EF,$D4,$BE,$A6,$92,$7F,$DE,$B0,$F9,$BC,$47,$CC,$8E,$0E,$A7
	.dc.b	$9C,$F4,$F9,$EE,$BF,$4B,$2B,$9C,$65,$EB,$19,$7A,$47,$E4,$3C,$7F
	.dc.b	$F0,$CE,$FF,$E7,$7F,$A8,$7D,$4F,$E5,$1B,$7F,$E7,$19,$FF,$3E,$57
	.dc.b	$0E,$DF,$F3,$A4,$FF,$A9,$3C,$4E,$18,$48,$E1,$3A,$AE,$11,$FE,$7B
	.dc.b	$6B,$C4,$D9,$9A,$F1,$77,$78,$FB,$24,$E8,$64,$8F,$47,$6E,$A7,$F2
	.dc.b	$B3,$97,$83,$5C,$C7,$BF,$AB,$E0,$7C,$9C,$9D,$C6,$5E,$31,$97,$AC
	.dc.b	$7C,$BE,$BF,$DD,$FB,$3F,$C7,$77,$FF,$51,$37,$6F,$85,$3B,$BF,$54
	.dc.b	$E5,$FF,$4F,$35,$FF,$67,$C6,$DF,$D7,$ED,$FA,$57,$E1,$35,$FD,$9C
	.dc.b	$FE,$B5,$FD,$AF,$D3,$38,$BF,$87,$FD,$4B,$EA,$69,$FC,$4A,$75,$5F
	.dc.b	$35,$7A,$99,$FA,$99,$39,$8F,$E3,$3F,$53,$FA,$D8,$C1,$0F,$E0,$AE
	.dc.b	$5B,$F1,$DF,$C8,$3E,$A5,$F1,$14,$E7,$1F,$C9,$3E,$A7,$F9,$C5,$7D
	.dc.b	$FB,$4E,$A3,$F0,$53,$A7,$F8,$9B,$78,$BF,$DA,$4D,$E2,$FC,$BB,$67
	.dc.b	$FD,$99,$F5,$2F,$94,$FE,$67,$ED,$7E,$CE,$7F,$1D,$2F,$6B,$F9,$BB
	.dc.b	$63,$F2,$53,$B7,$F9,$1B,$71,$FF,$71,$4E,$E7,$F9,$3B,$63,$FD,$CD
	.dc.b	$3B,$C7,$F5,$6D,$AF,$F7,$03,$53,$6B,$FB,$F5,$7F,$BB,$FE,$0B,$D5
	.dc.b	$5F,$A5,$FA,$47,$C1,$7F,$4E,$6C,$7E,$A6,$76,$FF,$3A,$6D,$7F,$BE
	.dc.b	$67,$73,$FC,$F9,$B1,$DF,$3F,$EB,$35,$C3,$2F,$E0,$6D,$BF,$F7,$C3
	.dc.b	$55,$BE,$2F,$FB,$0F,$86,$F7,$FF,$B1,$FD,$3F,$E1,$2F,$F7,$CD,$7F
	.dc.b	$A7,$CE,$DF,$E0,$4D,$7F,$A9,$C6,$F3,$65,$FC,$5A,$30,$FF,$18,$AE
	.dc.b	$13,$AA,$08,$CC,$EE,$FC,$E7,$9B,$E3,$FF,$BC,$FD,$4B,$E8,$BD,$3F
	.dc.b	$53,$31,$FE,$2D,$39,$6C,$E5,$F3,$CF,$28,$91,$A2,$1E,$CF,$1B,$FC
	.dc.b	$0F,$F5,$3F,$83,$D5,$EC,$FC,$83,$3E,$64,$BF,$93,$19,$7F,$94,$57
	.dc.b	$E2,$C9,$3B,$8F,$93,$F4,$3D,$BE,$1F,$E9,$FB,$52,$78,$C6,$5F,$AB
	.dc.b	$19,$7F,$C0,$76,$9F,$F0,$EF,$D4,$FE,$02,$7F,$9C,$3F,$9A,$6F,$56
	.dc.b	$B9,$A4,$69,$AF,$F8,$3B,$EA,$5E,$D7,$F8,$BE,$1F,$52,$FA,$9F,$EA
	.dc.b	$75,$AE,$9F,$B7,$39,$AF,$BB,$97,$8D,$E3,$7F,$D6,$A3,$13,$B9,$5F
	.dc.b	$37,$67,$E1,$9C,$46,$E7,$8D,$73,$C7,$FB,$2C,$F3,$AD,$CF,$9B,$1F
	.dc.b	$54,$F1,$AD,$CF,$1C,$3D,$2E,$F1,$B1,$E1,$BE,$33,$B8,$F2,$8C,$7C
	.dc.b	$6E,$EB,$8B,$EC,$AE,$2F,$C4,$BF,$0B,$4D,$E3,$E8,$79,$1D,$AF,$AB
	.dc.b	$DE,$78,$BD,$6F,$8D,$F6,$8D,$1E,$2A,$E3,$7D,$63,$81,$FE,$9D,$7E
	.dc.b	$83,$FE,$A2,$6B,$F0,$CD,$7B,$6F,$95,$FF,$52,$47,$D8,$8C,$3A,$8C
	.dc.b	$CF,$F9,$0D,$DC,$7F,$00,$F8,$6A,$9A,$47,$6F,$CA,$F9,$47,$D3,$FF
	.dc.b	$09,$DB,$F1,$15,$CC,$FA,$73,$AE,$DF,$A3,$19,$9F,$C7,$4E,$63,$EC
	.dc.b	$99,$DF,$FC,$5B,$19,$9E,$B3,$E5,$3A,$DF,$86,$BF,$D8,$98,$79,$63
	.dc.b	$13,$FE,$CA,$8C,$61,$E2,$8E,$57,$E2,$AD,$C7,$FB,$04,$EB,$3B,$AE
	.dc.b	$31,$E6,$38,$DF,$9A,$B8,$C7,$A4,$FE,$57,$87,$A6,$7C,$B1,$D0,$E5
	.dc.b	$5D,$30,$F1,$61,$FE,$3D,$BF,$89,$B7,$F3,$5B,$F2,$56,$9F,$F4,$8C
	.dc.b	$EA,$BC,$47,$F9,$16,$3C,$42,$34,$F9,$3A,$7F,$A0,$5F,$59,$FE,$75
	.dc.b	$5F,$E5,$2F,$8C,$7D,$4F,$FA,$4E,$3C,$97,$C1,$FA,$AD,$C3,$FD,$85
	.dc.b	$F5,$2E,$FC,$47,$86,$75,$92,$B5,$7E,$FB,$DE,$58,$C3,$FD,$67,$97
	.dc.b	$FA,$AF,$C5,$CD,$4B,$E0,$79,$FE,$13,$FB,$FB,$8B,$F8,$B3,$E5,$BF
	.dc.b	$5F,$C4,$F1,$9E,$BA,$FF,$BD,$AF,$FC,$D9,$F0,$3F,$CE,$5B,$D5,$ED
	.dc.b	$5F,$4E,$C7,$DB,$5B,$C3,$FD,$86,$DD,$FE,$C3,$6F,$FB,$69,$CA,$B8
	.dc.b	$E1,$23,$FD,$02,$76,$E4,$C6,$1F,$ED,$48,$C1,$FE,$81,$FA,$97,$B9
	.dc.b	$F3,$9B,$1C,$B3,$FF,$80,$0E,$E7,$44,$75,$EE,$5B,$FF,$6D,$DF,$27
	.dc.b	$C1,$77,$4F,$C9,$4D,$BF,$5C,$DA,$F6,$F5,$7E,$02,$7F,$DA,$07,$8F
	.dc.b	$C1,$C4,$48,$FF,$4D,$1A,$9E,$F3,$97,$FD,$DB,$19,$C3,$FC,$2C,$BE
	.dc.b	$A7,$FC,$B0,$F2,$BF,$AD,$18,$AF,$F7,$AF,$CC,$7E,$2E,$FF,$54,$DB
	.dc.b	$F4,$D7,$E5,$A3,$15,$FE,$22,$6A,$73,$3F,$82,$9E,$1C,$72,$3E,$87
	.dc.b	$8B,$6F,$E4,$BF,$A2,$FA,$87,$D4,$FE,$93,$EA,$7F,$CB,$77,$FE,$BB
	.dc.b	$9C,$BF,$EB,$D8,$CD,$B3,$FE,$05,$18,$6F,$8B,$D7,$0E,$8F,$5F,$81
	.dc.b	$98,$73,$8B,$87,$38,$FF,$6C,$E1,$F1,$F0,$EB,$E9,$7E,$E3,$EA,$7C
	.dc.b	$77,$B3,$FC,$8B,$5B,$BF,$F7,$0F,$47,$FB,$45,$FF,$6A,$36,$78,$09
	.dc.b	$E0,$FF,$93,$A3,$58,$7F,$1D,$FA,$9F,$F3,$25,$FF,$B7,$5B,$5C,$AB
	.dc.b	$FD,$4F,$39,$F4,$FF,$86,$FD,$37,$FA,$92,$3F,$DC,$8D,$AF,$47,$D4
	.dc.b	$3E,$A5,$F5,$3C,$3D,$C9,$87,$07,$F8,$D5,$6E,$73,$7E,$AC,$3A,$69
	.dc.b	$FE,$67,$F1,$5F,$53,$FE,$79,$CB,$91,$FE,$FA,$8C,$FB,$1F,$C9,$7A
	.dc.b	$BF,$CD,$BF,$53,$72,$FC,$7C,$62,$FF,$E0,$0B,$FE,$72,$EA,$D5,$6D
	.dc.b	$B3,$AA,$E2,$3E,$A7,$A1,$D7,$E0,$6C,$EE,$2C,$FD,$BC,$3F,$CF,$F5
	.dc.b	$8D,$3F,$A7,$7D,$3B,$D7,$FE,$72,$DC,$7F,$A2,$7D,$4F,$FB,$91,$FF
	.dc.b	$43,$D7,$0A,$D9,$23,$7F,$0B,$72,$FC,$1E,$1C,$DF,$E7,$B8,$7F,$25
	.dc.b	$7F,$D2,$31,$9C,$3F,$D2,$9C,$EA,$D6,$7F,$E1,$FC,$7E,$67,$A9,$F8
	.dc.b	$6B,$FF,$4D,$4E,$67,$E5,$CE,$5F,$DE,$3F,$EA,$47,$3D,$1C,$9E,$6F
	.dc.b	$4F,$A9,$91,$CB,$E3,$6E,$9F,$26,$D5,$8C,$7C,$30,$AC,$5F,$FA,$B2
	.dc.b	$DA,$3E,$43,$E6,$9F,$DA,$79,$5F,$06,$F9,$6C,$E1,$F3,$8E,$57,$C5
	.dc.b	$C6,$1F,$23,$0E,$57,$C6,$73,$3E,$09,$E6,$FE,$EF,$37,$D4,$FD,$B8
	.dc.b	$FA,$B3,$FD,$DA,$31,$3E,$CF,$90,$F9,$E7,$D0,$FA,$9B,$87,$CE,$8B
	.dc.b	$87,$C5,$9B,$5F,$13,$E6,$3E,$BF,$B1,$F6,$1B,$87,$D6,$6D,$FC,$D6
	.dc.b	$D7,$F8,$87,$EA,$7F,$BE,$CF,$F0,$C6,$C7,$EC,$F4,$FE,$49,$7F,$3E
	.dc.b	$72,$DC,$3F,$C2,$71,$8B,$FF,$12,$57,$07,$FD,$C3,$19,$4E,$F9,$9F
	.dc.b	$52,$F9,$7E,$78,$7F,$83,$63,$18,$FE,$43,$EA,$7F,$2B,$F5,$3F,$96
	.dc.b	$6C,$64,$FC,$AC,$65,$A7,$E6,$3A,$FF,$3B,$EC,$FE,$79,$B1,$93,$F3
	.dc.b	$B1,$96,$9F,$80,$FA,$8F,$66,$A6,$9F,$C5,$DF,$E6,$AE,$3C,$E4,$FF
	.dc.b	$A0,$F4,$FC,$C4,$BE,$43,$37,$FE,$85,$9C,$DE,$3E,$0F,$E2,$48,$F0
	.dc.b	$56,$6B,$9B,$63,$57,$D7,$0F,$04,$63,$0D,$3C,$A3,$2D,$3C,$47,$2A
	.dc.b	$DD,$7F,$68,$C3,$C7,$3B,$6C,$3E,$03,$67,$0F,$A9,$B8,$76,$9D,$3E
	.dc.b	$0F,$A9,$FD,$E7,$16,$6F,$CB,$18,$C7,$C6,$1C,$63,$E2,$6E,$9F,$32
	.dc.b	$B8,$F7,$7F,$92,$EF,$8D,$69,$F9,$4D,$7E,$D4,$F0,$C3,$C9,$5C,$70
	.dc.b	$EC,$DC,$3B,$63,$AA,$98,$76,$AD,$3F,$4B,$0E,$8E,$67,$78,$7C,$89
	.dc.b	$CB,$DA,$33,$87,$6A,$E0,$49,$D3,$F1,$32,$F9,$6E,$67,$F9,$26,$5E
	.dc.b	$D3,$DD,$3A,$AE,$24,$9D,$E3,$E8,$BF,$87,$97,$89,$F8,$F1,$9C,$39
	.dc.b	$E5,$E8,$BE,$39,$6E,$F8,$9E,$33,$8B,$9C,$BC,$53,$C2,$74,$7D,$DC
	.dc.b	$3C,$53,$A8,$E2,$6B,$0F,$0E,$3D,$DF,$75,$70,$C7,$C9,$5C,$32,$F7
	.dc.b	$3F,$4E,$F8,$57,$AC,$E0,$E3,$2F,$35,$F2,$AF,$24,$62,$75,$1E,$8C
	.dc.b	$7C,$CE,$8F,$35,$F4,$6D,$71,$69,$C6,$FB,$63,$EA,$9F,$D1,$61,$DB
	.dc.b	$1F,$4D,$FC,$C6,$D6,$E3,$F0,$57,$CB,$0F,$66,$1C,$A7,$BD,$F2,$C7
	.dc.b	$CF,$87,$18,$C6,$1C,$6F,$CD,$1D,$B0,$E7,$87,$6C,$7C,$99,$7D,$9A
	.dc.b	$FC,$D4,$62,$7E,$84,$FE,$66,$79,$5F,$48,$E0,$49,$D9,$F5,$4F,$3B
	.dc.b	$97,$E0,$37,$1F,$43,$D4,$C7,$D4,$BE,$07,$FC,$09,$5F,$A2,$8C,$38
	.dc.b	$3E,$B1,$AA,$46,$A3,$8F,$D4,$3E,$A7,$FC,$1D,$97,$F0,$FA,$E6,$DC
	.dc.b	$3F,$94,$D6,$EE,$DF,$0C,$7C,$55,$FC,$7F,$0E,$31,$FE,$19,$BE,$17
	.dc.b	$FC,$79,$7C,$2F,$18,$FE,$81,$0F,$A9,$C4,$23,$0F,$8A,$73,$1C,$A3
	.dc.b	$04,$3F,$95,$FD,$4B,$E1,$39,$B6,$35,$A7,$F3,$EA,$E5,$1C,$2B,$67
	.dc.b	$6B,$91,$FD,$2A,$F9,$3F,$CE,$3D,$FF,$F1,$A3,$FE,$28,$5F,$53,$FE
	.dc.b	$39,$F0,$3F,$B0,$CA,$72,$EF,$C5,$D7,$1F,$F8,$FE,$79,$CF,$FA,$89
	.dc.b	$DB,$6F,$6F,$F6,$78,$DD,$F6,$5F,$29,$A9,$F0,$7D,$0D,$93,$66,$4F
	.dc.b	$A6,$34,$61,$FF,$59,$FD,$4B,$EA,$79,$7A,$52,$7F,$C8,$F3,$96,$CE
	.dc.b	$5F,$54,$F8,$E7,$51,$D2,$70,$47,$E5,$6F,$E9,$FA,$9F,$EE,$11,$FE
	.dc.b	$54,$5F,$07,$93,$5D,$92,$75,$19,$78,$47,$3F,$94,$91,$CD,$ED,$F1
	.dc.b	$18,$8F,$F6,$8D,$A7,$1B,$FD,$A8,$57,$CD,$FA,$97,$99,$E7,$F4,$66
	.dc.b	$3F,$E6,$A9,$CB,$67,$2F,$69,$E5,$12,$34,$43,$C7,$DB,$F6,$FE,$A5
	.dc.b	$F5,$34,$93,$FE,$78,$87,$CD,$62,$3F,$DD,$71,$C1,$D4,$F3,$9E,$9F
	.dc.b	$3D,$D7,$E9,$E5,$73,$8C,$BD,$63,$2F,$48,$FC,$77,$8D,$FE,$28,$DF
	.dc.b	$F3,$FF,$50,$FA,$9F,$C9,$36,$FF,$D1,$B5,$BE,$DF,$F4,$F3,$63,$FC
	.dc.b	$85,$3B,$57,$D2,$FB,$F8,$92,$79,$B5,$DE,$1F,$EA,$9B,$FD,$2B,$7B
	.dc.b	$3C,$64,$9F,$F4,$FC,$FE,$91,$B1,$97,$99,$E4,$7D,$7D,$9E,$CE,$9F
	.dc.b	$6F,$CA,$4A,$E7,$19,$7A,$C6,$5E,$62,$FF,$85,$E7,$FC,$1B,$4F,$CE
	.dc.b	$CF,$FA,$CB,$D7,$FE,$42,$F0,$6C,$EF,$AF,$E8,$9F,$BF,$8C,$E9,$FE
	.dc.b	$C9,$6D,$7F,$9B,$6F,$AB,$E2,$6D,$F3,$8F,$AB,$7C,$FA,$FF,$85,$7D
	.dc.b	$3B,$8B,$F8,$67,$D4,$BE,$A6,$9A,$8D,$FA,$7C,$3E,$A6,$65,$FE,$CC
	.dc.b	$C1,$3F,$C5,$7E,$A3,$F5,$73,$98,$E2,$FF,$B4,$A3,$2F,$C8,$5F,$1D
	.dc.b	$FC,$77,$EA,$5F,$0F,$4E,$91,$FC,$87,$EA,$7E,$6F,$1F,$FB,$AE,$F8
	.dc.b	$E1,$D6,$B5,$CD,$23,$FD,$C3,$7C,$1F,$F4,$A3,$57,$D4,$BE,$5B,$F9
	.dc.b	$77,$B5,$FB,$09,$FF,$74,$CF,$FA,$85,$BE,$D6,$BC,$09,$18,$21,$F8
	.dc.b	$4A,$E4,$D8,$DB,$4E,$BE,$1F,$F3,$DF,$81,$F1,$47,$FB,$E3,$E9,$5F
	.dc.b	$03,$FD,$05,$B1,$FE,$FC,$9D,$BF,$92,$6F,$5F,$F5,$69,$DC,$FE,$59
	.dc.b	$B1,$FD,$6A,$77,$C3,$FE,$B7,$A4,$61,$FF,$5D,$35,$CE,$BE,$3F,$89
	.dc.b	$FE,$A9,$E2,$BE,$3F,$FA,$B7,$D3,$FD,$AB,$FF,$61,$65,$FD,$A2,$31
	.dc.b	$1F,$A8,$AE,$0E,$B2,$F1,$46,$23,$F5,$B3,$C2,$B9,$73,$78,$30,$95
	.dc.b	$C4,$FF,$68,$5F,$1A,$FE,$EE,$DC,$3D,$D8,$3E,$3F,$FB,$67,$D4,$BE
	.dc.b	$8F,$D3,$F5,$33,$2E,$82,$74,$E1,$AF,$9A,$38,$3F,$C1,$23,$39,$7F
	.dc.b	$82,$1B,$8F,$CF,$6C,$FF,$B9,$DA,$B1,$FF,$00,$BE,$2F,$A9,$FC,$5D
	.dc.b	$FC,$BF,$63,$E7,$9F,$E0,$58,$CE,$1F,$6D,$B5,$FE,$F3,$BF,$F0,$C3
	.dc.b	$EC,$6B,$FC,$6A,$36,$9A,$4E,$5F,$BB,$F4,$1D,$FE,$1F,$EF,$FB,$52
	.dc.b	$FC,$D1,$97,$B4,$65,$ED,$1F,$E1,$BF,$A9,$73,$DE,$A9,$5F,$42,$72
	.dc.b	$D9,$E0,$47,$EA,$FD,$4F,$F7,$DD,$D0,$FA,$87,$D4,$FE,$76,$B5,$39
	.dc.b	$8F,$D0,$61,$DA,$30,$72,$71,$5C,$A3,$AD,$72,$8C,$42,$31,$5D,$30
	.dc.b	$FD,$65,$6A,$F9,$CF,$8F,$DC,$FE,$C7,$87,$80,$E4,$E2,$B8,$5F,$2A
	.dc.b	$E9,$5E,$E9,$F0,$E1,$E0,$8F,$2D,$F1,$AD,$C7,$5C,$3C,$2E,$B1,$E5
	.dc.b	$3E,$36,$C7,$C1,$C7,$C0,$D3,$27,$9A,$2E,$1E,$18,$5F,$66,$D7,$27
	.dc.b	$81,$EF,$9E,$13,$B3,$83,$CC,$FA,$F1,$FE,$A6,$6B,$E7,$3F,$D4,$CE
	.dc.b	$1F,$AB,$ED,$BE,$53,$FD,$55,$1F,$56,$30,$F0,$6B,$D5,$AF,$FA,$C7
	.dc.b	$E1,$78,$BF,$8C,$6A,$69,$1F,$B5,$F2,$8F,$A7,$73,$7C,$89,$E4,$6A
	.dc.b	$32,$FE,$04,$CE,$FF,$8C,$AE,$55,$C6,$7C,$07,$17,$17,$F5,$27,$73
	.dc.b	$CC,$DD,$73,$BE,$EF,$E6,$23,$BE,$DF,$D0,$6B,$8C,$E6,$78,$CE,$6B
	.dc.b	$8B,$F1,$63,$C4,$77,$BF,$14,$EA,$3F,$DB,$B1,$8C,$3C,$F1,$C2,$FB
	.dc.b	$4E,$63,$91,$9B,$F0,$57,$F3,$2C,$3D,$BB,$7F,$AC,$AB,$D2,$7F,$00
	.dc.b	$8C,$CF,$08,$FF,$1A,$C6,$1F,$19,$E8,$9F,$F5,$DC,$7F,$41,$9E,$DD
	.dc.b	$5F,$52,$7E,$AF,$C6,$3E,$A3,$FD,$2F,$5A,$26,$1F,$E4,$CA,$7D,$4F
	.dc.b	$F4,$D9,$CC,$73,$8C,$4F,$93,$BB,$B5,$7D,$16,$C3,$07,$A9,$FE,$02
	.dc.b	$AF,$85,$5A,$A6,$7D,$31,$F5,$CF,$2C,$38,$F8,$DB,$AD,$61,$DC,$FE
	.dc.b	$D1,$3E,$DF,$89,$D5,$F1,$D3,$FD,$FB,$5F,$D8,$BC,$87,$57,$85,$B1
	.dc.b	$FD,$C4,$7D,$47,$A5,$E1,$3B,$23,$B8,$D6,$BF,$E6,$AA,$E6,$DA,$F8
	.dc.b	$75,$D3,$93,$8C,$FF,$9E,$A3,$07,$F7,$6F,$A9,$7B,$9F,$15,$B1,$DF
	.dc.b	$CF,$C4,$C8,$E0,$F6,$C9,$39,$BF,$F3,$ED,$F2,$7C,$13,$78,$7F,$73
	.dc.b	$09,$C6,$1F,$D4,$8F,$87,$87,$D7,$7F,$D9,$E7,$F8,$30,$B3,$D3,$C0
	.dc.b	$D7,$4C,$9D,$3E,$58,$C4,$FF,$A4,$27,$8E,$1F,$C2,$97,$D4,$FF,$96
	.dc.b	$DE,$27,$D5,$71,$3F,$E9,$8F,$9C,$FC,$1B,$FD,$A1,$B3,$F6,$A7,$F1
	.dc.b	$D1,$A9,$34,$DE,$7E,$0A,$FF,$13,$C6,$78,$7E,$67,$8D,$C9,$B3,$FE
	.dc.b	$2C,$FA,$97,$D4,$FE,$6F,$EA,$1E,$2F,$E7,$67,$35,$FE,$32,$9C,$B6
	.dc.b	$78,$1B,$AB,$1B,$F7,$BA,$18,$D3,$FD,$6F,$7F,$BA,$6F,$A3,$E1,$3F
	.dc.b	$59,$F5,$3E,$08,$CD,$FF,$90,$BA,$DF,$09,$FB,$16,$C7,$96,$F1,$3F
	.dc.b	$B5,$FA,$97,$A9,$4E,$1E,$A5,$BF,$F2,$6F,$D3,$7F,$A8,$9B,$ED,$7F
	.dc.b	$02,$FA,$87,$D4,$BE,$A7,$F8,$05,$EF,$6F,$F6,$8B,$7A,$BF,$26,$DC
	.dc.b	$7F,$C4,$78,$72,$BE,$38,$C9,$FF,$34,$78,$AF,$A9,$FF,$3D,$47,$88
	.dc.b	$CB,$8C,$BF,$DD,$3D,$5F,$E2,$EF,$A9,$FE,$2E,$FF,$8C,$3A,$AC,$A6
	.dc.b	$C6,$1B,$7F,$E7,$7A,$E4,$DA,$FC,$9C,$7F,$8E,$BE,$A7,$99,$D3,$82
	.dc.b	$D6,$E2,$D7,$90,$FF,$64,$C6,$4F,$F3,$CF,$D3,$BD,$6F,$E5,$0D,$8F
	.dc.b	$E4,$3F,$53,$FE,$DA,$77,$3A,$23,$AF,$85,$B7,$FF,$00,$2F,$92,$E3
	.dc.b	$FE,$8F,$87,$F1,$A7,$F8,$2C,$E9,$CF,$93,$F2,$3D,$4F,$F0,$2E,$3F
	.dc.b	$0D,$FC,$3B,$4D,$73,$ED,$A6,$4E,$0F,$3F,$A9,$91,$CF,$E3,$2E,$7F
	.dc.b	$62,$D5,$8C,$7F,$C2,$3C,$1F,$E2,$7B,$C6,$1F,$EC,$3C,$DF,$37,$F2
	.dc.b	$0F,$EA,$BC,$BF,$EC,$6F,$96,$C5,$7C,$E9,$BA,$7C,$38,$FE,$E9,$F0
	.dc.b	$8A,$CE,$7F,$12,$FF,$AF,$AF,$A9,$FA,$58,$7C,$D9,$CC,$62,$7C,$FB
	.dc.b	$3E,$43,$E5,$1F,$2F,$EA,$6E,$7F,$2E,$2E,$3F,$06,$6D,$F2,$F9,$8F
	.dc.b	$A5,$EB,$FD,$36,$E1,$F4,$DB,$7F,$25,$B5,$FD,$03,$EA,$7F,$BA,$4F
	.dc.b	$BC,$FF,$48,$B8,$8F,$F1,$EE,$FF,$EE,$CF,$2F,$FB,$C4,$E7,$0F,$F7
	.dc.b	$83,$6F,$8E,$3F,$3B,$83,$F8,$FC,$75,$13,$B6,$CF,$F7,$FF,$A9,$7C
	.dc.b	$BF,$68,$C3,$F8,$FF,$A9,$FC,$87,$D4,$DC,$BF,$20,$DC,$3F,$0C,$DD
	.dc.b	$7E,$7B,$F9,$7E,$CF,$CC,$75,$DB,$FC,$C3,$6F,$F1,$CD,$BF,$AE,$FF
	.dc.b	$85,$FE,$A3,$D7,$1F,$DA,$1B,$5F,$EB,$8A,$47,$F0,$F9,$D6,$FC,$4F
	.dc.b	$F3,$3F,$C1,$7B,$AB,$FD,$7F,$4F,$8F,$CE,$BC,$E3,$E4,$D6,$EB,$AD
	.dc.b	$69,$E2,$D7,$38,$FF,$9F,$EF,$96,$19,$C3,$B4,$E6,$B8,$B5,$E2,$7F
	.dc.b	$A1,$6B,$8C,$66,$78,$CE,$4F,$4F,$C0,$75,$3B,$7D,$4D,$AE,$11,$E5
	.dc.b	$FA,$97,$77,$F9,$16,$F9,$1F,$22,$75,$19,$D3,$E0,$57,$0F,$47,$CC
	.dc.b	$DB,$E3,$97,$F9,$3B,$1C,$E1,$C7,$3F,$A9,$7B,$C7,$9C,$E7,$0F,$4D
	.dc.b	$F0,$C6,$47,$92,$F8,$61,$F4,$F2,$F2,$D6,$B1,$E6,$FB,$B0,$C5,$78
	.dc.b	$31,$ED,$8B,$1E,$6F,$80,$ED,$1A,$8C,$C6,$E3,$8E,$5E,$D8,$DE,$1D
	.dc.b	$91,$B7,$C4,$E8,$98,$F9,$9F,$8F,$97,$A9,$CE,$3E,$7A,$E3,$5B,$24
	.dc.b	$66,$78,$E5,$E1,$7C,$07,$23,$D9,$3C,$1F,$2E,$5E,$C3,$DB,$8F,$AE
	.dc.b	$75,$3C,$9B,$19,$BF,$04,$6A,$B8,$E3,$E9,$8C,$65,$F1,$A3,$15,$C0
	.dc.b	$98,$79,$1A,$F7,$6E,$1D,$E3,$CF,$5F,$4C,$99,$7C,$4A,$E3,$8F,$AE
	.dc.b	$7C,$19,$7C,$38,$C5,$6E,$F9,$B9,$AD,$C7,$C7,$3C,$46,$8F,$9A,$66
	.dc.b	$79,$B8,$CB,$DF,$5F,$47,$1F,$44,$F2,$25,$7C,$6C,$3C,$15,$BC,$BE
	.dc.b	$2D,$77,$C3,$E7,$C6,$23,$8E,$3E,$69,$EF,$1E,$7C,$7B,$E3,$E4,$9E
	.dc.b	$38,$F6,$AD,$61,$CE,$B5,$97,$EA,$6B,$57,$C2,$FD,$D8,$FC,$E7,$DD
	.dc.b	$5F,$3A,$31,$3F,$C8,$71,$F0,$C6,$27,$EC,$E3,$DF,$EA,$1F,$03,$FC
	.dc.b	$CF,$1F,$E0,$31,$87,$07,$E6,$0D,$66,$8D,$46,$1E,$19,$7F,$12,$6E
	.dc.b	$1D,$2F,$F9,$47,$D3,$BE,$A7,$FA,$15,$FF,$37,$8C,$61,$FC,$63,$0E
	.dc.b	$2F,$E9,$1A,$A3,$6F,$F2,$46,$CF,$86,$F9,$1E,$0B,$EF,$87,$F5,$1A
	.dc.b	$7C,$2E,$30,$FE,$0B,$0F,$A9,$C4,$77,$6C,$70,$8F,$5C,$E0,$8F,$F4
	.dc.b	$5F,$A9,$FF,$66,$B7,$E1,$72,$F9,$E7,$AC,$EA,$3F,$9E,$B7,$3F,$89
	.dc.b	$3C,$1F,$EB,$91,$9C,$7F,$B5,$7B,$FF,$E3,$8B,$FE,$3B,$CF,$FE,$D6
	.dc.b	$87,$D4,$FF,$6F,$8C,$37,$93,$8F,$8F,$6F,$F9,$2D,$F3,$7F,$C2,$4F
	.dc.b	$27,$AC,$F8,$2B,$84,$27,$85,$F8,$7D,$AC,$C3,$E8,$A9,$88,$95,$D0
	.dc.b	$F0,$C6,$9C,$1D,$63,$BF,$D4,$BE,$A7,$AF,$A5,$2B,$73,$96,$CE,$5F
	.dc.b	$5C,$F2,$89,$1A,$21,$F0,$38,$7D,$BF,$52,$E9,$BF,$05,$2B,$CB,$39
	.dc.b	$6C,$E5,$FF,$1D,$42,$79,$CF,$4F,$94,$95,$C9,$EF,$F1,$38,$8E,$31
	.dc.b	$77,$70,$FF,$8A,$71,$5F,$D2,$FA,$97,$9B,$E8,$F4,$A6,$5F,$2A,$72
	.dc.b	$D9,$CB,$E6,$9E,$51,$23,$44,$3D,$7D,$BF,$80,$FA,$97,$D4,$D2,$4F
	.dc.b	$D8,$87,$CC,$62,$3F,$CB,$91,$C1,$D4,$F3,$9E,$9F,$3B,$D7,$E9,$65
	.dc.b	$73,$8C,$BD,$63,$2F,$48,$FC,$9F,$8B,$FE,$41,$E0,$F8,$F3,$FE,$00
	.dc.b	$FA,$87,$D4,$FE,$61,$B1,$F8,$EA,$E1,$C5,$FE,$55,$6C,$71,$AE,$07
	.dc.b	$6E,$F9,$1C,$5A,$FF,$9E,$6F,$F4,$77,$FE,$5F,$5D,$7E,$12,$4F,$E7
	.dc.b	$A3,$F4,$CD,$8C,$BC,$DF,$47,$67,$A7,$A7,$D5,$F2,$52,$B9,$46,$5E
	.dc.b	$91,$97,$C9,$1F,$1B,$AB,$F7,$3E,$C7,$F1,$F9,$FD,$7B,$5F,$F4,$96
	.dc.b	$BF,$99,$7E,$34,$F0,$F1,$A4,$60,$87,$08,$EC,$F4,$EC,$FE,$0F,$F4
	.dc.b	$CE,$3F,$E1,$3F,$52,$FA,$9A,$6A,$37,$E9,$F0,$FA,$99,$8F,$F0,$4C
	.dc.b	$13,$FC,$4B,$EA,$3F,$47,$39,$8E,$2F,$F0,$A8,$CB,$FE,$AD,$5F,$15
	.dc.b	$FC,$6B,$EA,$5F,$11,$4E,$91,$FC,$7B,$EA,$7F,$9B,$75,$FE,$33,$06
	.dc.b	$1C,$9E,$D5,$BE,$49,$1F,$C6,$AF,$8F,$4F,$FA,$85,$B1,$FE,$54,$FA
	.dc.b	$97,$C9,$FF,$2C,$F6,$7F,$53,$3F,$C8,$A7,$FD,$40,$DF,$6B,$5E,$0C
	.dc.b	$8C,$10,$FF,$65,$57,$26,$C6,$DA,$75,$F1,$3F,$9D,$FC,$0F,$8A,$3F
	.dc.b	$98,$FD,$23,$E0,$3F,$9F,$B6,$3F,$9A,$CE,$DF,$F6,$C3,$78,$7E,$34
	.dc.b	$EE,$7F,$DB,$8D,$8F,$95,$3B,$DB,$FD,$88,$DA,$FF,$3B,$B5,$36,$9F
	.dc.b	$15,$FD,$4B,$C6,$7D,$4F,$F5,$4F,$A7,$FB,$78,$7F,$4A,$6B,$FE,$ED
	.dc.b	$9D,$BF,$D3,$9A,$FF,$BC,$63,$79,$7D,$F6,$E5,$FE,$D9,$AD,$CF,$E0
	.dc.b	$DB,$19,$BF,$21,$E3,$F9,$0F,$EC,$FF,$50,$FA,$3F,$37,$D4,$CC,$FF
	.dc.b	$AF,$CE,$5B,$39,$7C,$93,$CA,$24,$68,$87,$A7,$C9,$FE,$EB,$F5,$3F
	.dc.b	$81,$D3,$D5,$EC,$FC,$53,$3C,$92,$FE,$34,$65,$FE,$E3,$5C,$75,$FF
	.dc.b	$7C,$36,$3F,$D5,$6D,$4D,$DD,$F4,$3E,$2F,$87,$FA,$9E,$D4,$BE,$B1
	.dc.b	$97,$EB,$46,$56,$7F,$AF,$C1,$5F,$E1,$0F,$A9,$72,$DE,$99,$87,$F8
	.dc.b	$22,$72,$D9,$E0,$47,$FC,$1B,$F5,$2F,$6F,$BC,$3E,$A5,$F5,$3F,$EC
	.dc.b	$88,$56,$F4,$FF,$65,$D6,$A7,$FD,$9B,$8F,$81,$E5,$7C,$63,$13,$B9
	.dc.b	$E7,$87,$82,$31,$87,$59,$E2,$74,$C7,$C5,$3C,$6F,$9E,$1E,$5B,$E1
	.dc.b	$3C,$B0,$E9,$3E,$1B,$E2,$DC,$3D,$2F,$19,$DC,$7F,$B9,$AF,$9C,$EE
	.dc.b	$3C,$63,$0F,$0B,$BA,$E2,$FB,$31,$EA,$FA,$4A,$71,$C3,$C6,$F1,$8C
	.dc.b	$E1,$E0,$8E,$07,$82,$FB,$BD,$4F,$1C,$6D,$E8,$76,$8D,$C6,$4F,$99
	.dc.b	$1F,$E9,$F6,$BF,$06,$7F,$D3,$EF,$53,$5E,$E3,$E4,$B3,$5F,$E3,$06
	.dc.b	$C6,$67,$FD,$4F,$39,$79,$8F,$84,$AD,$32,$E0,$8F,$CE,$F9,$17,$D3
	.dc.b	$FF,$86,$B6,$FC,$44,$8E,$4F,$E2,$E7,$7C,$BF,$3A,$75,$7F,$42,$73
	.dc.b	$1F,$68,$C9,$DE,$FB,$6F,$FE,$C0,$AE,$11,$89,$FF,$63,$46,$2F,$B4
	.dc.b	$78,$AB,$F4,$D7,$DB,$0F,$3C,$E4,$F4,$46,$67,$B4,$78,$6B,$9D,$F5
	.dc.b	$8F,$31,$C4,$57,$17,$C6,$71,$37,$1C,$6F,$C2,$DB,$F2,$E1,$FE,$44
	.dc.b	$D3,$FB,$D6,$1F,$11,$FE,$F9,$87,$C4,$C3,$E6,$56,$F1,$E2,$7F,$A5
	.dc.b	$2F,$83,$EA,$9D,$B9,$E9,$E1,$7F,$01,$BF,$1A,$FA,$8F,$F4,$25,$6C
	.dc.b	$95,$FC,$81,$FF,$5D,$FD,$4B,$BF,$9C,$FC,$2C,$3A,$E8,$AD,$5F,$96
	.dc.b	$F7,$9E,$30,$E8,$32,$FF,$4E,$78,$9A,$A9,$7C,$4F,$F5,$A4,$E9,$C5
	.dc.b	$72,$3E,$9C,$FB,$33,$FF,$75,$CE,$B2,$FF,$38,$DF,$BF,$E2,$BD,$3E
	.dc.b	$37,$F3,$AA,$FF,$5E,$B7,$1F,$2C,$3E,$9F,$CC,$EB,$AB,$89,$1F,$F6
	.dc.b	$41,$BC,$31,$5E,$5B,$E9,$8F,$FA,$0A,$BA,$70,$EA,$31,$8F,$DA,$BF
	.dc.b	$D6,$7D,$4B,$DD,$F9,$ED,$C3,$BD,$3C,$EC,$D4,$9F,$F6,$BD,$6E,$FD
	.dc.b	$77,$B9,$CD,$6E,$72,$DC,$3C,$6F,$DE,$9F,$F5,$E9,$FE,$E2,$2C,$F4
	.dc.b	$F0,$35,$D3,$27,$4F,$E4,$1F,$95,$19,$9F,$B1,$1C,$AF,$E2,$7D,$4F
	.dc.b	$F9,$42,$38,$9F,$31,$C4,$FD,$FF,$9B,$FC,$55,$7F,$98,$5B,$5F,$C4
	.dc.b	$CF,$B1,$5F,$96,$8C,$D7,$F1,$76,$A7,$33,$C0,$FF,$7F,$BA,$33,$93
	.dc.b	$C4,$E9,$C9,$ED,$8C,$57,$F2,$3C,$7F,$CE,$9F,$52,$FA,$9F,$D3,$FD
	.dc.b	$4B,$C4,$FD,$4C,$E6,$F8,$1F,$DF,$67,$4E,$34,$FE,$5D,$18,$7F,$40
	.dc.b	$DC,$3B,$E1,$DB,$0E,$FE,$FC,$C3,$96,$0C,$38,$47,$79,$F2,$FA,$72
	.dc.b	$37,$F5,$3E,$5B,$EF,$3F,$C0,$3A,$3F,$D9,$35,$D1,$B3,$A8,$C7,$99
	.dc.b	$FC,$1B,$EA,$7F,$CC,$B1,$E2,$6C,$FE,$E6,$BF,$D3,$F3,$9F,$4B,$F8
	.dc.b	$6F,$D4,$7F,$A9,$23,$D2,$DA,$FE,$13,$F5,$0F,$A9,$7D,$4F,$0F,$6E
	.dc.b	$61,$C1,$FE,$43,$5B,$9C,$DF,$F0,$9C,$3A,$71,$FF,$BD,$BC,$57,$D4
	.dc.b	$FF,$9E,$B2,$E4,$7D,$38,$CF,$AF,$FC,$B3,$AF,$FB,$17,$D4,$DC,$BF
	.dc.b	$1F,$18,$BF,$B4,$FF,$67,$EB,$D5,$6D,$B3,$AA,$E2,$3E,$A7,$A1,$D7
	.dc.b	$E0,$2C,$EE,$2C,$FF,$2C,$C3,$FB,$A5,$63,$4F,$F6,$37,$D3,$BD,$7F
	.dc.b	$E7,$2D,$C7,$F9,$6F,$D4,$FF,$B6,$9F,$C3,$C6,$71,$FE,$5C,$DB,$E7
	.dc.b	$7F,$D2,$AF,$85,$6A,$3A,$99,$F2,$AE,$1D,$9F,$91,$EA,$22,$F3,$E1
	.dc.b	$A7,$C8,$8D,$36,$34,$D3,$5F,$53,$D0,$EB,$F1,$B6,$32,$4C,$7E,$35
	.dc.b	$A7,$51,$EE,$F9,$CF,$8A,$7F,$68,$E4,$FF,$5F,$FC,$BE,$30,$F8,$B3
	.dc.b	$6B,$FA,$DD,$6A,$72,$48,$FE,$BF,$B3,$EA,$7E,$3B,$F5,$2F,$FB,$4C
	.dc.b	$62,$BF,$8B,$64,$F9,$1F,$8E,$7C,$8F,$A9,$B3,$F1,$62,$E3,$EC,$9B
	.dc.b	$8F,$CD,$8F,$E3,$D6,$F9,$9F,$A3,$EB,$7D,$26,$E3,$F3,$DB,$5F,$19
	.dc.b	$B7,$CE,$1F,$53,$FE,$F3,$AF,$E1,$AD,$8F,$87,$F0,$FF,$C3,$9F,$F2
	.dc.b	$34,$66,$B9,$36,$BF,$D3,$8D,$9F,$F1,$6D,$71,$7B,$E1,$FE,$10,$9D
	.dc.b	$78,$6F,$A9,$7C,$C6,$6B,$FC,$3B,$18,$C7,$F1,$5F,$53,$F9,$0F,$A9
	.dc.b	$FC,$8B,$63,$27,$E3,$E3,$2D,$3F,$25,$F2,$3F,$98,$E9,$FC,$CB,$63
	.dc.b	$27,$E5,$E3,$2D,$3F,$98,$FD,$47,$AC,$9E,$1F,$F8,$A9,$C7,$A1,$F5
	.dc.b	$FE,$0F,$D7,$C1,$FE,$6D,$97,$C7,$67,$7F,$F6,$14,$7F,$43,$AD,$D2
	.dc.b	$37,$8F,$F9,$C2,$B7,$7C,$6B,$A3,$5E,$18,$5B,$E3,$1F,$E8,$7A,$E3
	.dc.b	$7C,$E3,$8B,$5E,$2D,$9E,$38,$79,$17,$CC,$75,$3B,$7D,$4D,$AE,$11
	.dc.b	$E3,$FA,$9F,$DA,$71,$E2,$7B,$46,$31,$F2,$1F,$E9,$17,$18,$FC,$26
	.dc.b	$E5,$F1,$B0,$E9,$DF,$FE,$4F,$BE,$22,$B6,$F3,$8D,$5F,$A6,$B8,$E3
	.dc.b	$E2,$9C,$E1,$FE,$F3,$C3,$8E,$1F,$46,$F6,$F6,$8C,$9A,$CF,$E1,$E5
	.dc.b	$CF,$19,$5B,$C7,$AB,$F8,$1C,$39,$E5,$ED,$9C,$CF,$A5,$B8,$74,$8E
	.dc.b	$57,$D2,$B8,$3F,$2E,$32,$48,$CB,$F1,$27,$94,$60,$99,$7B,$EF,$BE
	.dc.b	$1E,$48,$C6,$1D,$DF,$A1,$8D,$C7,$E6,$4D,$7E,$34,$5C,$BE,$2E,$1D
	.dc.b	$23,$0F,$BA,$3E,$46,$1D,$EB,$99,$E9,$C3,$BB,$4C,$E3,$E8,$7E,$56
	.dc.b	$1D,$31,$F1,$65,$EE,$BF,$4D,$71,$3E,$64,$FC,$7B,$E5,$18,$35,$5C
	.dc.b	$5E,$F5,$CA,$77,$39,$9E,$11,$CB,$1F,$5E,$1E,$39,$E7,$8F,$B5,$DE
	.dc.b	$1C,$CF,$6D,$F1,$CB,$E2,$1F,$1E,$7D,$18,$7C,$AC,$BD,$17,$C6,$7C
	.dc.b	$A4,$3E,$89,$C2,$B9,$CF,$08,$D5,$73,$7F,$09,$87,$96,$31,$87,$E9
	.dc.b	$71,$F5,$3F,$02,$75,$87,$4B,$E3,$5A,$C7,$C9,$1E,$2B,$F6,$63,$F5
	.dc.b	$30,$E2,$DB,$E1,$5E,$9C,$BE,$66,$5E,$2C,$79,$C6,$32,$FC,$47,$D4
	.dc.b	$BE,$09,$84,$3F,$71,$19,$70,$7E,$30,$B8,$23,$51,$87,$86,$9F,$C3
	.dc.b	$DB,$8F,$4B,$FE,$49,$F5,$0F,$A9,$F0,$5E,$E7,$F8,$DD,$F0,$C3,$F8
	.dc.b	$D3,$55,$F1,$79,$DF,$2C,$3B,$47,$F0,$DA,$E1,$5B,$8F,$E3,$AD,$C3
	.dc.b	$C1,$1B,$C3,$9B,$E6,$F8,$5E,$8F,$C6,$A7,$CF,$62,$3C,$6D,$84,$D2
	.dc.b	$4F,$33,$FA,$27,$D4,$FF,$B5,$5B,$F0,$DA,$7C,$B3,$D6,$75,$1F,$CD
	.dc.b	$1B,$8F,$F5,$CB,$EF,$3D,$4D,$5F,$97,$E0,$5E,$37,$DD,$F0,$7F,$C2
	.dc.b	$D0,$FA,$9C,$72,$7E,$06,$31,$87,$FA,$91,$E2,$DB,$DB,$F8,$88,$DA
	.dc.b	$BE,$B7,$A8,$FF,$62,$55,$A7,$B7,$7E,$86,$57,$78,$D1,$87,$A6,$3E
	.dc.b	$59,$CC,$7F,$AE,$3E,$A5,$F5,$3D,$7D,$09,$39,$8F,$14,$EA,$3D,$93
	.dc.b	$A7,$A9,$FA,$28,$D3,$80,$E3,$4F,$F1,$F3,$70,$FD,$43,$67,$E2,$6B
	.dc.b	$F0,$7E,$A7,$FB,$EC,$7F,$B3,$97,$C1,$64,$D7,$6C,$9D,$46,$5E,$11
	.dc.b	$CF,$E5,$64,$73,$7B,$7C,$56,$23,$D1,$69,$C6,$7F,$E5,$58,$57,$D5
	.dc.b	$FA,$9F,$F3,$05,$F9,$7D,$3F,$41,$9F,$3A,$46,$48,$7C,$59,$E2,$D8
	.dc.b	$DB,$4F,$81,$D7,$F8,$3F,$A9,$7D,$4D,$24,$FF,$BC,$A1,$F3,$78,$8F
	.dc.b	$A5,$1C,$1D,$4F,$39,$E9,$F3,$DD,$7E,$96,$57,$38,$CB,$D6,$32,$F4
	.dc.b	$8F,$CB,$78,$BF,$E4,$6D,$FF,$CF,$5F,$50,$FA,$9F,$CD,$36,$FF,$93
	.dc.b	$CF,$F9,$FE,$B8,$73,$FF,$29,$6C,$FF,$2C,$AE,$0F,$E2,$DB,$5C,$7C
	.dc.b	$5F,$09,$27,$AD,$F3,$8F,$0F,$77,$8B,$B6,$56,$8E,$CF,$9B,$B7,$31
	.dc.b	$DE,$2E,$1E,$C9,$F5,$F5,$7B,$3E,$4E,$5F,$31,$9F,$F1,$0B,$F8,$75
	.dc.b	$A9,$F5,$57,$27,$86,$9F,$9C,$79,$7B,$3F,$C9,$9B,$E3,$7C,$EE,$D9
	.dc.b	$18,$21,$FA,$FA,$E4,$D8,$DB,$4E,$BD,$BF,$C3,$7E,$99,$C9,$FC,$3F
	.dc.b	$EA,$5F,$53,$4D,$46,$FD,$4E,$1F,$53,$31,$F8,$18,$27,$F8,$C7,$D4
	.dc.b	$7E,$9A,$73,$1C,$5F,$8D,$19,$7F,$83,$2F,$8E,$FE,$41,$F5,$2F,$87
	.dc.b	$A7,$48,$FE,$49,$F5,$3F,$37,$AF,$FB,$85,$F1,$C3,$AD,$6B,$96,$47
	.dc.b	$D5,$BE,$0F,$F7,$86,$AF,$A9,$7C,$A7,$F3,$1F,$67,$F5,$73,$F7,$67
	.dc.b	$FD,$4E,$DF,$6B,$5E,$14,$8C,$10,$FE,$45,$5C,$9B,$1B,$69,$D7,$C5
	.dc.b	$FE,$81,$F0,$5E,$28,$FC,$67,$D2,$BE,$0B,$FA,$23,$63,$F2,$13,$B7
	.dc.b	$F9,$83,$7B,$FF,$DD,$93,$B9,$FE,$68,$D8,$FF,$78,$4E,$F8,$7F,$C3
	.dc.b	$4D,$AF,$F0,$EB,$53,$69,$F1,$3F,$D5,$FC,$77,$BF,$FD,$67,$E9,$FE
	.dc.b	$D6,$1F,$A5,$6B,$FD,$16,$76,$FE,$A5,$AF,$F4,$98,$DE,$4C,$BF,$C5
	.dc.b	$D1,$87,$FC,$61,$5C,$27,$55,$FE,$34,$8C,$4E,$AF,$CC,$79,$7E,$3F
	.dc.b	$FB,$77,$D4,$BE,$8F,$D1,$F5,$33,$1F,$DE,$4E,$5B,$39,$7C,$D3,$CA
	.dc.b	$24,$68,$87,$AF,$CA,$FE,$F9,$F5,$3F,$83,$D3,$CF,$2D,$5E,$D7,$C9
	.dc.b	$37,$C9,$27,$E5,$46,$5F,$E2,$17,$CF,$6F,$F2,$93,$63,$FC,$AA,$D4
	.dc.b	$DB,$7D,$0F,$93,$E1,$FE,$C7,$B7,$2F,$AC,$65,$FB,$51,$97,$FB,$F5
	.dc.b	$A7,$FC,$31,$F5,$2E,$5B,$D3,$2F,$F9,$34,$E5,$B3,$C0,$8F,$F2,$BF
	.dc.b	$A9,$FE,$E5,$BF,$F9,$BB,$EA,$1F,$53,$FE,$6F,$9C,$D2,$72,$FF,$9D
	.dc.b	$30,$EC,$E6,$BF,$CE,$D1,$89,$DC,$AB,$9C,$61,$B5,$C3,$0F,$F3,$F4
	.dc.b	$F0,$BE,$53,$E3,$F7,$BF,$C7,$58,$77,$3F,$D0,$B8,$72,$7C,$98,$72
	.dc.b	$AE,$91,$ED,$79,$63,$63,$8D,$F4,$C7,$C1,$7C,$67,$99,$B3,$D5,$3B
	.dc.b	$BE,$71,$E3,$18,$FA,$1E,$26,$71,$F3,$3C,$B1,$E9,$3B,$8E,$38,$78
	.dc.b	$5D,$1C,$70,$F0,$3D,$31,$F3,$BE,$5C,$3A,$9F,$61,$FF,$52,$B5,$F3
	.dc.b	$98,$9F,$96,$EB,$DD,$7C,$87,$FA,$B2,$3E,$6C,$61,$C5,$7F,$8D,$A7
	.dc.b	$51,$F7,$67,$4E,$3E,$13,$2F,$FA,$E9,$A9,$A4,$7E,$57,$C9,$BE,$9F
	.dc.b	$F8,$7E,$5F,$8A,$9E,$46,$A3,$3B,$79,$EF,$EB,$5F,$1E,$5F,$B1,$7C
	.dc.b	$30,$FC,$9C,$E6,$3E,$C1,$9B,$FF,$67,$46,$67,$C1,$3E,$73,$C1,$5D
	.dc.b	$1C,$47,$4C,$3C,$71,$FE,$D4,$C3,$C7,$18,$8F,$F6,$59,$C8,$F0,$9C
	.dc.b	$B0,$F0,$C6,$E3,$C8,$7F,$B8,$B0,$F2,$4F,$BE,$39,$5F,$9A,$FB,$E1
	.dc.b	$E5,$6C,$F7,$DF,$F9,$1E,$1F,$05,$FE,$4B,$1D,$84,$66,$BC,$67,$F8
	.dc.b	$C5,$C4,$7B,$4E,$91,$F2,$3A,$7F,$A2,$D7,$66,$C7,$FB,$EB,$E3,$5F
	.dc.b	$51,$FE,$8E,$8E,$93,$AB,$F7,$36,$DF,$52,$EF,$D5,$79,$AB,$53,$9B
	.dc.b	$F3,$56,$B2,$C5,$FF,$AA,$B2,$7A,$BF,$DF,$EF,$64,$9F,$F5,$65,$70
	.dc.b	$CB,$FA,$FD,$F0,$F1,$7A,$63,$30,$FF,$2F,$4F,$B3,$E2,$F1,$8F,$7D
	.dc.b	$7F,$DD,$F5,$FE,$59,$F1,$FF,$9D,$37,$AB,$E0,$2F,$A7,$63,$EE,$AD
	.dc.b	$E1,$FE,$BF,$6E,$9F,$D7,$1B,$69,$FE,$F5,$7C,$A3,$B1,$23,$FD,$96
	.dc.b	$F7,$E4,$ED,$8F,$8A,$BE,$ED,$3E,$A5,$EE,$FD,$06,$E1,$FE,$7D,$3D
	.dc.b	$1E,$7E,$66,$47,$F9,$DE,$77,$5E,$0B,$5E,$A3,$FD,$B5,$3A,$CF,$FC
	.dc.b	$1B,$7F,$0F,$0D,$46,$30,$D4,$FB,$EF,$FC,$31,$D1,$F8,$48,$FF,$64
	.dc.b	$9E,$59,$FE,$E2,$FF,$A2,$4C,$CF,$FB,$A5,$B5,$FE,$06,$69,$F2,$CF
	.dc.b	$A9,$E0,$F5,$B9,$1F,$07,$EA,$7F,$CB,$6F,$58,$FB,$71,$8B,$FF,$7B
	.dc.b	$7C,$EF,$E2,$EB,$D0,$D8,$FA,$F3,$F9,$C8,$C4,$1A,$6E,$7F,$0D,$3B
	.dc.b	$E3,$91,$EF,$F1,$6D,$FC,$F7,$F4,$FF,$52,$FA,$9F,$D4,$FD,$4B,$C5
	.dc.b	$CC,$61,$98,$7F,$AB,$E7,$4F,$FA,$C6,$B9,$39,$9E,$7E,$FC,$8C,$1C
	.dc.b	$5E,$FE,$A4,$8F,$F5,$D7,$D4,$F2,$CD,$C3,$FE,$36,$CB,$E4,$3F,$E3
	.dc.b	$8E,$97,$9B,$FC,$01,$B8,$79,$71,$C4,$FF,$04,$FA,$9F,$F3,$3C,$FF
	.dc.b	$B1,$DB,$8F,$68,$FC,$45,$FF,$A8,$5D,$7A,$7F,$C3,$3E,$9B,$FD,$4D
	.dc.b	$1F,$EC,$F6,$D7,$11,$F5,$0F,$A9,$7D,$4F,$1F,$72,$61,$C5,$FE,$39
	.dc.b	$5B,$9C,$DF,$31,$87,$6D,$BF,$CD,$BE,$2B,$EA,$7F,$CF,$F9,$73,$3F
	.dc.b	$DD,$91,$9F,$63,$F9,$47,$5F,$F0,$1F,$A9,$B9,$7E,$36,$31,$7F,$EF
	.dc.b	$37,$F8,$37,$5E,$AB,$6D,$9D,$57,$11,$F5,$3D,$0E,$BF,$03,$67,$71
	.dc.b	$67,$E3,$61,$FC,$4E,$B1,$A7,$F9,$CB,$EA,$5E,$C7,$F3,$A6,$E3,$FC
	.dc.b	$AB,$EA,$7F,$DB,$CF,$FB,$FE,$33,$8F,$C7,$6D,$F3,$BF,$E9,$1C,$CC
	.dc.b	$38,$DE,$E7,$B1,$A8,$FE,$4F,$1F,$C9,$EF,$9E,$1F,$D5,$67,$39,$7F
	.dc.b	$88,$A3,$E6,$FA,$9F,$8A,$CB,$FD,$43,$39,$9F,$99,$39,$7F,$99,$46
	.dc.b	$70,$F8,$73,$97,$F7,$8F,$FA,$BE,$78,$F3,$F6,$FA,$99,$1D,$3E,$32
	.dc.b	$DF,$29,$B8,$F8,$B0,$56,$2F,$F9,$26,$6F,$91,$F9,$C7,$F6,$DE,$4F
	.dc.b	$F6,$6F,$CC,$62,$BE,$74,$DC,$FF,$3D,$5A,$9C,$92,$3F,$45,$93,$EA
	.dc.b	$7E,$7B,$F7,$6F,$FB,$C4,$62,$BE,$1E,$6F,$91,$F9,$67,$CC,$FA,$9B
	.dc.b	$3F,$2A,$2E,$3F,$22,$6E,$3E,$FF,$99,$FA,$7E,$B7,$D4,$6E,$1F,$49
	.dc.b	$B7,$F4,$1B,$5F,$EC,$2F,$A9,$FF,$00,$4F,$EC,$5B,$1F,$B0,$F8,$8F
	.dc.b	$E3,$F3,$9C,$3F,$C1,$F5,$AC,$3F,$A8,$36,$7F,$C4,$F5,$C2,$FB,$5F
	.dc.b	$03,$F2,$BF,$52,$F9,$8C,$63,$FE,$08,$8C,$3F,$8A,$FA,$9F,$C7,$FD
	.dc.b	$4F,$E4,$1B,$19,$3F,$1F,$19,$69,$F9,$2F,$92,$FC,$BF,$B3,$F9,$86
	.dc.b	$C6,$4F,$CB,$C6,$5A,$7F,$82,$72,$FF,$73,$FD,$47,$AE,$3F,$B7,$37
	.dc.b	$5F,$F0,$73,$FC,$51,$FE,$34,$E5,$3F,$2B,$E0,$BE,$04,$70,$F8,$F5
	.dc.b	$7F,$A0,$F5,$EB,$E9,$62,$BF,$D0,$91,$E0,$AC,$5F,$29,$EB,$18,$9D
	.dc.b	$5F,$39,$F2,$DE,$5F,$04,$65,$E2,$DC,$17,$B9,$F4,$D5,$9C,$D6,$FE
	.dc.b	$07,$06,$7E,$A6,$C6,$4E,$DF,$53,$FB,$4F,$13,$18,$FB,$0F,$F1,$B4
	.dc.b	$65,$C0,$71,$A7,$BD,$BA,$7D,$2C,$7C,$D1,$8A,$E7,$DF,$FE,$48,$79
	.dc.b	$C6,$5F,$A7,$3C,$27,$57,$D5,$B7,$FE,$2B,$BE,$19,$FF,$99,$B2,$F5
	.dc.b	$D6,$B1,$EF,$96,$0F,$6E,$3C,$DC,$47,$8A,$78,$3E,$2A,$CD,$6E,$35
	.dc.b	$5C,$B1,$F1,$5F,$0C,$7D,$47,$93,$2F,$55,$72,$AD,$BC,$6F,$CA,$74
	.dc.b	$26,$1F,$89,$CB,$D9,$18,$AE,$2D,$3F,$3B,$7D,$9C,$1F,$57,$1F,$5D
	.dc.b	$72,$BF,$51,$2F,$9B,$DE,$F9,$39,$BF,$2B,$9C,$BD,$51,$8A,$E2,$7D
	.dc.b	$F8,$CC,$78,$71,$B8,$78,$6B,$73,$C4,$87,$CF,$3D,$39,$7C,$AC,$BD
	.dc.b	$39,$79,$2B,$8D,$75,$3F,$27,$7D,$5E,$77,$C7,$0F,$9D,$1B,$CB,$D8
	.dc.b	$DB,$F4,$3E,$0C,$7E,$D1,$E0,$CB,$D7,$87,$39,$EF,$97,$8D,$EF,$5B
	.dc.b	$79,$63,$F0,$27,$51,$DC,$F7,$E5,$E7,$70,$4B,$FC,$46,$1E,$0C,$3C
	.dc.b	$78,$7A,$A7,$C5,$97,$AB,$0E,$51,$8C,$BE,$2C,$62,$F8,$E3,$F8,$48
	.dc.b	$C6,$1D,$4E,$93,$AC,$3A,$65,$E8,$C7,$A6,$1F,$8E,$CB,$83,$CE,$75
	.dc.b	$F5,$0F,$82,$FF,$4F,$4F,$ED,$63,$0E,$0F,$B0,$6A,$D1,$A8,$C3,$C3
	.dc.b	$0F,$E1,$8D,$D3,$A5,$FF,$25,$FA,$77,$D4,$FF,$AB,$6F,$73,$FC,$72
	.dc.b	$F8,$61,$FC,$69,$AA,$F8,$BB,$98,$FF,$14,$9D,$CE,$A3,$F9,$0B,$6B
	.dc.b	$BC,$6E,$BC,$97,$E3,$C3,$FD,$81,$7B,$7D,$35,$B3,$9C,$ED,$FF,$72
	.dc.b	$4B,$E1,$B1,$1E,$7F,$9E,$B2,$6C,$C9,$F8,$8D,$7F,$9B,$FD,$4F,$FB
	.dc.b	$69,$FE,$98,$D7,$D7,$1B,$CB,$FA,$C4,$E5,$BF,$0D,$9C,$BD,$75,$E4
	.dc.b	$AD,$47,$FB,$77,$E0,$3F,$C7,$58,$7F,$75,$E0,$C7,$D4,$FF,$B8,$7A
	.dc.b	$3F,$B4,$4A,$93,$9A,$E5,$E3,$5A,$1B,$70,$F1,$9B,$5C,$A3,$C0,$70
	.dc.b	$7C,$6F,$FA,$65,$FF,$5A,$BA,$3C,$5F,$45,$C8,$E7,$EE,$48,$D4,$A7
	.dc.b	$51,$89,$F1,$FD,$43,$EA,$7C,$7E,$94,$AE,$D3,$96,$CE,$5F,$24,$F2
	.dc.b	$89,$1A,$21,$E9,$E0,$F8,$3F,$52,$E9,$BF,$07,$2B,$D9,$39,$6C,$E5
	.dc.b	$FF,$5D,$42,$79,$CF,$4F,$93,$95,$C9,$EF,$F1,$18,$8E,$31,$78,$BE
	.dc.b	$DE,$0A,$FA,$BF,$52,$F3,$3C,$FE,$94,$C7,$E8,$4E,$5B,$39,$7C,$B3
	.dc.b	$CA,$24,$68,$87,$AB,$B7,$F0,$7F,$52,$FA,$9A,$49,$FB,$D0,$F9,$BC
	.dc.b	$47,$FB,$6E,$38,$3A,$9E,$73,$D3,$E7,$BA,$FD,$2C,$AE,$71,$97,$AC
	.dc.b	$65,$E9,$1F,$96,$F1,$7F,$C8,$5B,$FE,$83,$EA,$1F,$53,$F9,$A6,$DF
	.dc.b	$E4,$EB,$7D,$DF,$A0,$8F,$F5,$3F,$7E,$EE,$46,$E3,$B1,$1C,$1F,$9D
	.dc.b	$BF,$D0,$E1,$E1,$EB,$F0,$12,$75,$D9,$5F,$F7,$F5,$7E,$99,$B3,$E7
	.dc.b	$9F,$47,$4F,$A7,$E4,$E5,$77,$8C,$BE,$18,$CB,$D2,$3E,$27,$57,$EE
	.dc.b	$BD,$7F,$E3,$53,$FB,$36,$AC,$BF,$83,$ED,$F9,$78,$F8,$D5,$C7,$C5
	.dc.b	$91,$82,$1C,$77,$F9,$55,$C6,$30,$7F,$0C,$71,$7F,$C5,$30,$FD,$44
	.dc.b	$E8,$7D,$33,$8B,$F8,$8F,$D4,$BE,$A6,$9F,$C5,$27,$55,$FC,$35,$7A
	.dc.b	$79,$FA,$99,$39,$8F,$E3,$7F,$51,$FA,$98,$C1,$0F,$13,$96,$FC,$77
	.dc.b	$F2,$2F,$A9,$7C,$3D,$39,$C7,$F2,$6F,$A9,$FE,$7D,$BF,$F3,$0C,$13
	.dc.b	$AC,$3F,$91,$56,$B8,$7F,$20,$D8,$FE,$62,$D4,$DF,$A9,$7C,$A7,$F3
	.dc.b	$4F,$67,$F5,$D5,$FC,$B6,$7F,$D2,$ED,$F6,$B5,$E0,$48,$C1,$0F,$A1
	.dc.b	$5C,$9B,$1B,$69,$D7,$C3,$FE,$87,$F0,$1E,$28,$FE,$79,$F4,$8F,$81
	.dc.b	$FE,$8E,$D8,$FE,$81,$3B,$7E,$F3,$78,$3F,$AA,$CE,$E7,$F0,$4D,$8F
	.dc.b	$EB,$33,$BD,$BF,$6C,$DA,$FE,$AE,$D4,$DA,$7C,$4F,$F5,$AF,$15,$F1
	.dc.b	$FF,$D7,$3E,$9F,$ED,$E1,$FD,$55,$AF,$E4,$E7,$6F,$F5,$96,$BF,$97
	.dc.b	$8D,$E5,$FD,$B9,$B9,$7F,$0B,$AD,$CF,$F7,$26,$C6,$6F,$C8,$78,$FE
	.dc.b	$3F,$FB,$7F,$D4,$BE,$8F,$CD,$F5,$33,$3F,$ED,$53,$96,$CE,$5F,$24
	.dc.b	$F2,$89,$1A,$21,$E9,$F2,$3F,$BE,$FD,$4F,$E0,$74,$F5,$7B,$3F,$14
	.dc.b	$CF,$24,$BF,$8D,$19,$7F,$BE,$57,$1D,$7F,$95,$36,$3F,$C1,$6D,$4D
	.dc.b	$DD,$F4,$1E,$2F,$87,$FA,$9E,$D4,$BE,$B1,$97,$EB,$46,$5F,$E1,$16
	.dc.b	$9F,$F0,$BF,$D4,$B9,$6F,$4C,$BF,$F0,$94,$E5,$B3,$C0,$8F,$F8,$5F
	.dc.b	$EA,$5E,$D7,$F2,$58,$7D,$43,$EA,$7F,$A4,$56,$B8,$BF,$C5,$33,$9A
	.dc.b	$FF,$15,$E3,$DD,$E3,$7F,$D3,$E3,$13,$B9,$5F,$37,$67,$F8,$D1,$C4
	.dc.b	$6E,$78,$D7,$3C,$7F,$AD,$4F,$3A,$DC,$F9,$B1,$F5,$4F,$1A,$DC,$F1
	.dc.b	$C3,$D2,$EF,$1B,$1E,$1B,$E3,$3B,$8F,$28,$C7,$C6,$EE,$B8,$BE,$CA
	.dc.b	$E2,$FC,$4B,$F0,$B4,$DE,$3E,$87,$91,$DA,$FA,$BD,$E7,$8B,$D6,$F8
	.dc.b	$DF,$68,$D1,$E2,$AE,$37,$D6,$38,$1F,$EA,$77,$E8,$3F,$EA,$C6,$BF
	.dc.b	$0C,$D7,$B6,$F9,$1F,$F5,$74,$7D,$88,$C3,$A8,$CC,$FF,$8D,$DD,$C7
	.dc.b	$F9,$BB,$E1,$6A,$69,$1D,$BF,$2B,$E4,$DF,$4E,$E5,$F9,$B5,$CC,$FA
	.dc.b	$73,$A8,$CC,$75,$33,$3C,$5F,$B2,$71,$DF,$F2,$55,$C6,$32,$71,$8C
	.dc.b	$CF,$17,$F3,$11,$C7,$7F,$F1,$55,$71,$9C,$CF,$19,$CD,$71,$7F,$DA
	.dc.b	$D8,$7B,$63,$13,$FE,$DC,$8C,$61,$EA,$6C,$79,$1B,$7D,$9B,$7D,$27
	.dc.b	$DF,$5C,$5A,$4D,$BF,$D4,$B8,$FB,$DF,$F5,$3C,$7A,$CF,$F5,$1C,$66
	.dc.b	$7D,$27,$FA,$6D,$C4,$79,$CE,$91,$F0,$BA,$7E,$75,$76,$6C,$7F,$AB
	.dc.b	$BE,$31,$F5,$1F,$E8,$58,$E9,$3A,$BF,$0B,$6D,$F5,$2E,$FD,$53,$2F
	.dc.b	$E9,$18,$7F,$AD,$2F,$75,$A9,$E1,$3E,$98,$ED,$97,$FA,$77,$C4,$D4
	.dc.b	$49,$D1,$FE,$BE,$9D,$38,$C3,$FC,$71,$7D,$1F,$F1,$FC,$FB,$3E,$2B
	.dc.b	$94,$63,$4F,$F7,$75,$FF,$B0,$FC,$6F,$E7,$0D,$E9,$F8,$0B,$EA,$18
	.dc.b	$FB,$AB,$78,$7F,$91,$DB,$AF,$FB,$79,$B7,$FE,$DA,$39,$57,$1C,$24
	.dc.b	$7F,$B6,$0E,$DC,$78,$C3,$FC,$AF,$18,$3F,$DB,$1F,$52,$F7,$3E,$7B
	.dc.b	$63,$96,$6F,$73,$35,$27,$8B,$EC,$9D,$46,$09,$7E,$08,$E8,$FF,$98
	.dc.b	$EE,$E9,$FB,$7A,$F8,$97,$F8,$07,$FD,$84,$7F,$BB,$4B,$5C,$FC,$0D
	.dc.b	$75,$49,$D3,$E3,$8C,$4F,$F9,$C6,$78,$DF,$7F,$A9,$FF,$27,$C7,$03
	.dc.b	$E4,$38,$9F,$F3,$C7,$CE,$7E,$06,$BF,$86,$42,$31,$5F,$4F,$35,$FE
	.dc.b	$62,$73,$43,$87,$26,$4F,$91,$36,$F5,$87,$6B,$E3,$1F,$87,$F1,$35
	.dc.b	$87,$CB,$7F,$4B,$F5,$0F,$A9,$FD,$3F,$D4,$3C,$4C,$C6,$19,$87,$63
	.dc.b	$FD,$89,$3B,$71,$87,$9E,$FA,$46,$21,$3B,$7F,$D3,$8F,$EF,$5B,$F0
	.dc.b	$3D,$9B,$3C,$67,$E6,$67,$F2,$9B,$1F,$ED,$5F,$A9,$EF,$5D,$3D,$4F
	.dc.b	$E0,$3D,$0F,$32,$9E,$E8,$B7,$E5,$FA,$9F,$F3,$1B,$7D,$2E,$3E,$CF
	.dc.b	$F0,$A9,$CE,$3F,$EB,$3F,$A6,$FF,$4C,$DF,$B5,$B1,$FE,$BC,$FA,$87
	.dc.b	$D4,$BE,$A7,$F8,$9D,$6B,$83,$F0,$6D,$C3,$FD,$DC,$DC,$18,$72,$BE
	.dc.b	$3E,$E4,$BF,$9F,$E2,$BE,$A7,$FC,$F3,$1E,$39,$FE,$4D,$39,$F6,$3F
	.dc.b	$94,$75,$7E,$2F,$EA,$6E,$7F,$8E,$8C,$61,$F6,$9F,$C8,$76,$EA,$B6
	.dc.b	$D9,$D5,$71,$1F,$53,$D0,$EB,$C5,$67,$71,$67,$FD,$C3,$87,$E7,$6B
	.dc.b	$1C,$1F,$A6,$FA,$77,$AF,$FC,$E9,$B8,$FF,$2F,$FA,$9F,$F6,$DB,$F8
	.dc.b	$D8,$CE,$3F,$EE,$76,$DF,$3B,$FE,$77,$7C,$2B,$51,$D4,$CF,$93,$70
	.dc.b	$EC,$FC,$8F,$4F,$FC,$19,$1F,$15,$FC,$1E,$1F,$E0,$7C,$FF,$C4,$13
	.dc.b	$BE,$8E,$0D,$38,$17,$1F,$83,$6F,$A9,$EC,$77,$F8,$CB,$5D,$E2,$D7
	.dc.b	$09,$B8,$FC,$7F,$9C,$F8,$C7,$E0,$3E,$05,$F2,$D8,$C7,$E3,$4D,$C1
	.dc.b	$7F,$E1,$FE,$6F,$F6,$2D,$E6,$3A,$5F,$F6,$7F,$A9,$F9,$11,$8B,$E3
	.dc.b	$3F,$C4,$27,$2F,$F1,$7D,$BF,$C8,$5F,$1F,$F2,$4F,$95,$F5,$36,$7E
	.dc.b	$54,$5B,$F8,$73,$70,$F9,$9B,$BE,$63,$E9,$7A,$FF,$4D,$B8,$FD,$36
	.dc.b	$E1,$F3,$1B,$7F,$CC,$3E,$A7,$FD,$E5,$5F,$C4,$DB,$5E,$A3,$FC,$0D
	.dc.b	$C1,$FC,$56,$B8,$61,$FE,$0A,$AD,$C6,$2F,$F6,$35,$B7,$F9,$B5,$AB
	.dc.b	$71,$86,$C7,$F7,$8F,$A9,$7C,$BF,$5C,$7F,$C1,$F1,$8B,$FC,$77,$D4
	.dc.b	$FE,$47,$EA,$7F,$24,$D8,$C9,$F9,$38,$CB,$4F,$CB,$7C,$8F,$E6,$7A
	.dc.b	$7F,$34,$D8,$C9,$F9,$B8,$CB,$4F,$E7,$1F,$51,$ED,$CF,$FA,$93,$53
	.dc.b	$FC,$41,$F9,$46,$3D,$1F,$F1,$8C,$7F,$3F,$F8,$2F,$7C,$F2,$F8,$EF
	.dc.b	$F3,$74,$62,$38,$5F,$FA,$07,$6F,$97,$6B,$DC,$61,$B5,$CE,$B8,$3E
	.dc.b	$6C,$32,$F9,$A6,$DF,$3A,$DC,$E5,$B1,$E7,$BE,$D1,$8F,$A0,$E2,$72
	.dc.b	$FA,$9B,$19,$39,$FD,$4B,$BB,$FC,$6B,$8F,$FA,$67,$0F,$5E,$1E,$4A
	.dc.b	$D7,$A5,$C2,$F7,$AF,$C3,$9C,$E1,$C4,$FA,$37,$C6,$B7,$87,$E4,$2B
	.dc.b	$77,$C6,$73,$7C,$31,$ED,$8C,$8E,$82,$F8,$E1,$E6,$AD,$61,$CE,$EC
	.dc.b	$E6,$F8,$DF,$07,$FC,$2C,$D7,$A6,$1E,$27,$F0,$F8,$FA,$6F,$9C,$70
	.dc.b	$9D,$66,$AD,$4F,$69,$E5,$7D,$6B,$59,$72,$AE,$F8,$59,$CC,$71,$8D
	.dc.b	$57,$1C,$3A,$E5,$ED,$AD,$1F,$12,$F8,$C6,$49,$97,$CF,$CB,$CF,$7C
	.dc.b	$A3,$18,$F8,$63,$18,$7B,$63,$19,$78,$AF,$D6,$72,$71,$1A,$9F,$1E
	.dc.b	$1E,$07,$91,$E5,$BE,$AF,$C0,$37,$87,$D5,$9E,$11,$96,$E3,$EC,$9F
	.dc.b	$3C,$F0,$3A,$CF,$EC,$E7,$51,$89,$DC,$F2,$AF,$0E,$5F,$7C,$F9,$98
	.dc.b	$FA,$AF,$A1,$F8,$07,$15,$C5,$B7,$D0,$F4,$16,$7C,$04,$39,$1D,$2B
	.dc.b	$C1,$97,$C9,$8C,$3F,$33,$2F,$DA,$46,$1F,$99,$87,$C4,$AE,$31,$A8
	.dc.b	$C5,$71,$AF,$9D,$87,$99,$F1,$65,$F1,$1F,$99,$8F,$9D,$A7,$EB,$CF
	.dc.b	$01,$F0,$A3,$86,$3F,$1F,$EA,$5F,$05,$FD,$22,$BF,$75,$18,$70,$7D
	.dc.b	$D3,$52,$8D,$46,$1E,$19,$7F,$09,$6E,$1D,$2F,$F8,$FF,$D4,$3E,$A7
	.dc.b	$FA,$BD,$EE,$7F,$8C,$DF,$0C,$3F,$76,$D5,$7C,$5D,$CC,$7F,$81,$CE
	.dc.b	$E7,$51,$89,$D5,$77,$8F,$E2,$33,$A8,$FE,$84,$DC,$3D,$13,$C4,$DE
	.dc.b	$1E,$88,$FE,$DF,$3B,$9F,$3D,$F6,$9C,$D6,$E7,$FD,$55,$4F,$86,$C4
	.dc.b	$7B,$3E,$7A,$C9,$B3,$27,$E4,$B5,$FE,$53,$F5,$3F,$ED,$D7,$FA,$5B
	.dc.b	$5F,$6C,$6F,$2F,$EB,$D3,$96,$FC,$36,$72,$F5,$D7,$92,$B5,$1F,$E0
	.dc.b	$CF,$80,$79,$1F,$E0,$EF,$03,$FD,$97,$0F,$A9,$C7,$1E,$DB,$A7,$FB
	.dc.b	$26,$78,$BF,$EB,$36,$BF,$DE,$1E,$4E,$63,$8A,$8D,$D7,$83,$DC,$CC
	.dc.b	$3E,$86,$98,$89,$5D,$E3,$2F,$86,$3A,$FD,$4B,$EA,$79,$FA,$12,$BF
	.dc.b	$D9,$13,$96,$CE,$5F,$44,$F2,$89,$1A,$21,$ED,$E0,$F5,$FD,$4F,$F4
	.dc.b	$D8,$FF,$6A,$2F,$83,$C9,$AE,$D9,$3A,$8C,$BC,$23,$9F,$CA,$48,$E6
	.dc.b	$F6,$F8,$AC,$47,$F8,$92,$D3,$8D,$DC,$3F,$E4,$89,$5F,$D0,$FA,$9F
	.dc.b	$F3,$06,$1E,$6F,$4B,$D2,$67,$CD,$91,$92,$1F,$E3,$29,$E2,$D8,$DB
	.dc.b	$4F,$83,$DB,$F7,$BE,$A5,$F5,$34,$93,$FE,$F9,$87,$CD,$E2,$3F,$C7
	.dc.b	$F1,$C1,$D4,$F3,$9E,$9F,$3D,$D7,$E9,$E5,$73,$8C,$BD,$63,$2F,$48
	.dc.b	$FC,$8F,$8D,$FE,$45,$E0,$F8,$F2,$1F,$50,$FA,$9F,$CB,$B6,$35,$3C
	.dc.b	$1F,$F1,$A7,$7F,$43,$75,$23,$7C,$7F,$E7,$AC,$3C,$0D,$AE,$13,$A3
	.dc.b	$C5,$3A,$7C,$98,$7E,$9A,$BF,$D0,$6B,$AF,$CA,$49,$E4,$7A,$BB,$30
	.dc.b	$FF,$9A,$A7,$83,$A6,$B9,$8F,$7F,$4F,$C0,$F9,$39,$3B,$8C,$BC,$63
	.dc.b	$2F,$58,$F9,$BD,$5F,$BE,$F6,$3F,$90,$4F,$91,$AF,$F9,$F3,$3F,$D3
	.dc.b	$3E,$86,$BF,$E8,$08,$DE,$DF,$3D,$BA,$7C,$BB,$EF,$5C,$5A,$F1,$EB
	.dc.b	$FE,$1F,$F4,$EE,$2F,$E2,$3F,$52,$FA,$9A,$6A,$37,$E9,$F0,$FA,$99
	.dc.b	$8F,$CC,$C1,$3F,$C6,$BE,$A3,$F5,$B3,$98,$E2,$FD,$28,$CB,$FE,$9A
	.dc.b	$5F,$15,$FC,$8B,$EA,$5F,$0F,$4E,$91,$FC,$9B,$EA,$7F,$A0,$71,$7E
	.dc.b	$3E,$F8,$E1,$D6,$B5,$C9,$23,$F0,$37,$C3,$8B,$FD,$92,$D8,$FC,$F7
	.dc.b	$D4,$BE,$53,$F9,$9F,$B5,$FB,$29,$FC,$4C,$FF,$A9,$5B,$ED,$6B,$C0
	.dc.b	$91,$82,$1F,$EB,$6A,$E4,$D8,$DB,$4E,$BE,$1F,$F4,$2F,$81,$F1,$47
	.dc.b	$E6,$3E,$95,$F0,$3F,$D1,$9B,$1F,$9C,$9D,$BF,$EC,$46,$F0,$FE,$CE
	.dc.b	$77,$3F,$EC,$86,$C7,$ED,$E7,$7B,$7F,$BB,$1B,$5F,$C0,$5A,$9B,$4F
	.dc.b	$89,$FE,$B3,$E2,$BE,$3F,$FA,$DF,$D3,$FD,$AC,$3F,$62,$D7,$FD,$B3
	.dc.b	$3B,$7F,$6A,$D7,$FD,$BB,$1B,$CB,$F8,$63,$72,$FF,$80,$35,$B9,$FE
	.dc.b	$20,$D8,$CD,$F9,$0F,$1F,$C7,$FF,$6E,$FA,$97,$D1,$F9,$BE,$A6,$67
	.dc.b	$FC,$1A,$72,$D9,$CB,$E4,$9E,$51,$23,$44,$3D,$3E,$4F,$F7,$CF,$A9
	.dc.b	$FC,$0E,$9E,$AF,$63,$E2,$99,$E4,$97,$F1,$A3,$2F,$F1,$8A,$E3,$AF
	.dc.b	$F3,$38,$46,$1F,$E6,$CD,$73,$5F,$4B,$E8,$7C,$5F,$0F,$F5,$3D,$99
	.dc.b	$5D,$63,$2F,$D6,$8C,$AC,$BF,$9E,$60,$AF,$F0,$BF,$D4,$B9,$6F,$4C
	.dc.b	$C3,$F9,$7C,$E5,$B3,$C0,$8F,$F3,$4F,$A9,$FE,$EF,$8F,$F4,$FF,$A9
	.dc.b	$7D,$4F,$BE,$73,$49,$CB,$FD,$57,$0E,$CE,$6B,$E3,$46,$27,$72,$AE
	.dc.b	$71,$86,$D7,$0C,$3E,$74,$F0,$BE,$53,$E3,$F7,$3E,$26,$1D,$CF,$EC
	.dc.b	$D8,$72,$7C,$98,$72,$AE,$91,$ED,$79,$63,$63,$8D,$F4,$C7,$C1,$7C
	.dc.b	$67,$99,$B3,$D5,$3B,$BE,$71,$E3,$18,$FA,$1E,$26,$71,$F3,$3C,$B1
	.dc.b	$E9,$3B,$8E,$38,$78,$5D,$1C,$70,$F0,$3D,$31,$F3,$BE,$5C,$3A,$9F
	.dc.b	$61,$FF,$58,$35,$F3,$98,$9F,$96,$EB,$DB,$7C,$87,$FA,$D2,$3E,$C4
	.dc.b	$61,$FF,$5B,$D7,$F8,$EA,$74,$D8,$E0,$3E,$12,$B4,$CB,$82,$3F,$27
	.dc.b	$E5,$1F,$4F,$FE,$22,$CB,$F1,$33,$FE,$23,$AF,$A5,$5A,$E5,$FA,$93
	.dc.b	$93,$95,$F1,$DF,$FD,$9B,$7C,$23,$13,$FE,$D3,$8C,$57,$28,$F3,$61
	.dc.b	$F9,$8C,$3B,$47,$DD,$3B,$46,$6F,$9C,$7F,$B3,$8F,$0D,$73,$BE,$B1
	.dc.b	$C4,$EB,$87,$F0,$BC,$3A,$E1,$EE,$9C,$9F,$E4,$18,$CC,$FA,$A3,$A9
	.dc.b	$CE,$BA,$D7,$16,$CF,$5C,$3F,$C9,$7B,$FF,$91,$70,$F5,$BF,$E4,$7C
	.dc.b	$3D,$75,$BC,$7F,$C6,$F5,$A7,$13,$FC,$4D,$F1,$63,$F0,$5D,$F4,$76
	.dc.b	$7D,$8D,$97,$C6,$BE,$A3,$FD,$25,$5C,$09,$3D,$4F,$E4,$1F,$52,$EE
	.dc.b	$E5,$0C,$BF,$A4,$D7,$F9,$57,$45,$EE,$3D,$37,$C2,$B1,$7D,$71,$7A
	.dc.b	$BF,$DF,$2F,$64,$9F,$11,$DE,$FD,$55,$C3,$C5,$E9,$12,$BF,$CC,$D3
	.dc.b	$ED,$F8,$9C,$5F,$7D,$3F,$DE,$55,$FE,$62,$F1,$FF,$9D,$37,$AB,$E3
	.dc.b	$2F,$A8,$F8,$EE,$49,$7E,$4A,$E2,$FC,$78,$E1,$8F,$C6,$9C,$EF,$FE
	.dc.b	$83,$6C,$FF,$A2,$CE,$DC,$7B,$C7,$C3,$8F,$DF,$C7,$FD,$21,$F5,$2F
	.dc.b	$73,$E8,$B7,$0F,$F4,$21,$E7,$F3,$F3,$32,$3F,$CF,$F3,$BA,$F3,$5A
	.dc.b	$F5,$1F,$3A,$75,$9F,$FA,$82,$FE,$16,$1A,$8C,$61,$A9,$FE,$E4,$7F
	.dc.b	$B2,$8F,$F4,$69,$6F,$BF,$83,$F8,$88,$C5,$7F,$A9,$5A,$72,$EB,$F8
	.dc.b	$33,$B6,$47,$BB,$EA,$7F,$CA,$8F,$68,$FA,$F1,$8B,$FC,$3F,$CC,$FE
	.dc.b	$3B,$0F,$F5,$93,$6F,$E8,$62,$BF,$CD,$CE,$6F,$FD,$7C,$D4,$E6,$BF
	.dc.b	$09,$3C,$38,$E4,$7D,$2B,$78,$B8,$C3,$E7,$BF,$A8,$FA,$87,$D4,$FE
	.dc.b	$AB,$EA,$5E,$2E,$63,$0C,$C7,$FD,$6F,$3A,$7F,$D7,$35,$C9,$CC,$F3
	.dc.b	$F7,$E4,$60,$E2,$F7,$F5,$24,$7F,$B0,$BE,$A7,$96,$7F,$AE,$C7,$E5
	.dc.b	$70,$7F,$B6,$7A,$7F,$D9,$57,$FE,$C8,$6D,$6F,$0F,$E0,$71,$8A,$FF
	.dc.b	$70,$7D,$4F,$F9,$8E,$FF,$D9,$ED,$BF,$DD,$D7,$FA,$72,$73,$E9,$FF
	.dc.b	$0C,$FA,$8F,$F5,$0C,$7F,$B5,$9B,$5F,$C4,$FE,$A1,$F5,$2F,$A9,$E1
	.dc.b	$EE,$4C,$38,$3F,$C7,$EB,$73,$9B,$FE,$17,$87,$4D,$3F,$DF,$BE,$2B
	.dc.b	$EA,$7F,$CF,$19,$72,$3F,$DE,$11,$9F,$5F,$F9,$57,$5B,$17,$D4,$E1
	.dc.b	$FC,$54,$66,$31,$D7,$96,$E1,$FE,$FB,$9D,$36,$B8,$8F,$A9,$E8,$75
	.dc.b	$F8,$1B,$3B,$8B,$3F,$CD,$EB,$C9,$58,$D1,$7F,$E6,$7F,$A9,$7A,$FF
	.dc.b	$CE,$9B,$1F,$CA,$3E,$A7,$FD,$B2,$FB,$6B,$8E,$3E,$FA,$DB,$FC,$D1
	.dc.b	$B1,$A8,$46,$27,$E3,$73,$2F,$AD,$71,$C3,$C4,$7F,$3C,$8F,$E4,$B1
	.dc.b	$DE,$31,$E3,$FD,$0F,$53,$D7,$87,$C9,$7F,$19,$97,$9B,$4F,$F1,$3C
	.dc.b	$EF,$A3,$83,$4E,$05,$C7,$E4,$5B,$EA,$7B,$1D,$FE,$36,$D7,$78,$B5
	.dc.b	$C2,$6E,$3F,$07,$67,$CD,$FC,$D3,$FB,$87,$2F,$FB,$1F,$E5,$F1,$97
	.dc.b	$CD,$3E,$A4,$E3,$1F,$EE,$15,$B9,$D4,$7D,$19,$D6,$1F,$E0,$AF,$A9
	.dc.b	$F9,$6F,$86,$3F,$BF,$28,$CD,$7F,$1C,$D1,$F2,$1F,$2C,$F9,$9F,$53
	.dc.b	$67,$E6,$45,$C7,$E2,$CD,$C3,$E7,$5B,$E6,$3E,$A7,$AB,$F5,$5B,$8F
	.dc.b	$D5,$6E,$1F,$39,$B7,$F8,$7F,$A9,$FF,$00,$4F,$F7,$E6,$C7,$ED,$3E
	.dc.b	$23,$F9,$04,$E7,$0F,$F0,$8D,$6B,$0F,$F6,$13,$67,$FC,$29,$5C,$2F
	.dc.b	$B5,$F0,$3F,$C1,$3F,$52,$F9,$7C,$63,$FE,$08,$8C,$3F,$8D,$FA,$9F
	.dc.b	$C8,$7D,$4F,$E4,$5B,$19,$3F,$25,$19,$69,$F9,$5F,$92,$FC,$C7,$57
	.dc.b	$E6,$5B,$19,$3F,$35,$19,$69,$F8,$0D,$1F,$51,$EB,$8F,$F7,$03,$63
	.dc.b	$F8,$93,$FC,$61,$FF,$13,$B9,$78,$F9,$FF,$BA,$F8,$3F,$81,$1C,$3E
	.dc.b	$3F,$1A,$F4,$AB,$39,$C3,$FC,$FD,$3E,$0A,$CC,$F9,$1B,$87,$D6,$C3
	.dc.b	$AC,$E6,$F8,$BC,$CF,$F4,$95,$F3,$8C,$C7,$99,$B4,$BE,$BF,$01,$27
	.dc.b	$3F,$53,$6F,$99,$F0,$A7,$CB,$F5,$2E,$EF,$F2,$FC,$E8,$F9,$73,$A8
	.dc.b	$CE,$7E,$8A,$E1,$E2,$79,$1B,$7C,$73,$FF,$2B,$E5,$9C,$38,$E9,$FD
	.dc.b	$0E,$97,$C3,$1E,$95,$A9,$F3,$5F,$1C,$64,$79,$6F,$86,$1F,$7B,$2F
	.dc.b	$35,$6B,$1E,$6F,$C6,$8C,$DF,$1C,$7B,$3E,$26,$E5,$E8,$8D,$57,$3C
	.dc.b	$3D,$18,$F8,$27,$CA,$68,$98,$F8,$1F,$85,$8F,$82,$30,$4A,$F4,$65
	.dc.b	$D6,$BF,$B3,$5F,$57,$C9,$87,$A6,$75,$18,$3B,$D7,$26,$9C,$B2,$F6
	.dc.b	$DF,$3C,$BD,$98,$F8,$6B,$58,$F4,$9C,$E1,$EE,$C3,$89,$27,$C5,$5D
	.dc.b	$8F,$15,$F3,$7F,$0F,$8F,$9B,$1F,$51,$E1,$CB,$E1,$63,$D7,$0F,$04
	.dc.b	$62,$FC,$67,$5C,$BC,$6E,$31,$ED,$87,$3C,$7B,$65,$E5,$AF,$9F,$18
	.dc.b	$C7,$AB,$63,$A6,$3D,$DF,$4E,$5E,$37,$89,$ED,$C3,$D3,$87,$E1,$2B
	.dc.b	$84,$63,$0F,$87,$8F,$92,$BC,$4F,$EF,$32,$F4,$E1,$D8,$F5,$63,$DE
	.dc.b	$75,$8F,$4B,$E3,$3F,$6E,$7A,$61,$EA,$BF,$25,$69,$FC,$8C,$E9,$F6
	.dc.b	$E5,$EF,$8F,$A9,$8F,$8B,$2F,$45,$68,$F7,$7D,$4B,$E0,$7F,$69,$1F
	.dc.b	$9E,$8C,$38,$3F,$20,$6A,$51,$A8,$C3,$C2,$FF,$83,$37,$3E,$97,$FC
	.dc.b	$6B,$EA,$1F,$53,$FC,$0E,$F7,$3F,$C2,$EF,$86,$1F,$A4,$6A,$BE,$2E
	.dc.b	$E6,$3F,$C1,$67,$73,$A8,$C4,$EA,$BB,$C7,$F1,$F9,$D4,$60,$98,$7A
	.dc.b	$27,$58,$79,$E3,$F8,$D5,$70,$75,$5C,$2B,$FC,$C1,$39,$6D,$71,$7D
	.dc.b	$97,$E3,$BE,$73,$99,$F7,$7C,$2D,$B7,$D4,$D3,$11,$27,$E6,$35,$FE
	.dc.b	$75,$F5,$3F,$ED,$97,$FA,$4B,$5F,$81,$1B,$CB,$FA,$9C,$E5,$BF,$0B
	.dc.b	$9C,$BD,$F5,$E5,$AD,$47,$F3,$BF,$7F,$FC,$77,$87,$F7,$35,$F5,$3F
	.dc.b	$CF,$FB,$BF,$CC,$11,$89,$D7,$8D,$9E,$3E,$4D,$8F,$F4,$44,$ED,$57
	.dc.b	$2B,$D4,$7F,$B2,$AA,$D3,$E4,$EF,$D0,$CA,$F2,$46,$8C,$3D,$31,$F5
	.dc.b	$4E,$63,$FC,$19,$F5,$2F,$A9,$EB,$E8,$C9,$CC,$78,$A7,$51,$F0,$67
	.dc.b	$4F,$53,$FD,$45,$1A,$70,$1C,$69,$FE,$85,$6E,$1F,$D9,$DB,$3F,$13
	.dc.b	$83,$E0,$FD,$4B,$B2,$FC,$1C,$AF,$F1,$24,$E5,$B3,$97,$C3,$09,$E7
	.dc.b	$3D,$3E,$4E,$57,$27,$BF,$C4,$E2,$38,$C5,$CF,$FD,$39,$82,$BE,$AF
	.dc.b	$D4,$FF,$98,$6F,$C7,$E9,$7A,$0C,$F9,$B2,$32,$43,$E2,$CF,$16,$C6
	.dc.b	$DA,$7B,$BB,$7F,$07,$F5,$2F,$A9,$A4,$9F,$F2,$34,$3E,$6B,$11,$F4
	.dc.b	$A3,$83,$A9,$E7,$3D,$3E,$77,$AF,$D2,$CA,$E7,$19,$7A,$C6,$5E,$91
	.dc.b	$F9,$6F,$1B,$FC,$7F,$BF,$FB,$07,$EA,$1F,$53,$F9,$A6,$DF,$F9,$6E
	.dc.b	$B7,$CF,$FE,$27,$6C,$7F,$BA,$E7,$6F,$E2,$9B,$7F,$EC,$D9,$DC,$FF
	.dc.b	$8F,$7B,$FC,$44,$7F,$CD,$15,$FB,$87,$97,$67,$8C,$8E,$BB,$2B,$A3
	.dc.b	$8C,$7A,$63,$D5,$D3,$EB,$F9,$39,$5D,$A3,$2F,$82,$32,$FA,$63,$E3
	.dc.b	$75,$7E,$F3,$D8,$FE,$3F,$3F,$E7,$D6,$BF,$A8,$CF,$F3,$0F,$C7,$9E
	.dc.b	$1E,$2C,$8C,$10,$E1,$1D,$9E,$9D,$9F,$C2,$BE,$99,$C7,$FC,$33,$EA
	.dc.b	$5F,$53,$4D,$46,$FD,$3E,$1F,$53,$31,$FF,$49,$60,$9F,$E2,$BF,$51
	.dc.b	$FA,$39,$CC,$71,$7F,$D3,$11,$97,$F8,$1A,$F8,$EF,$E3,$BF,$52,$F8
	.dc.b	$7A,$74,$8F,$E4,$3F,$53,$FC,$EF,$A5,$D5,$FE,$6D,$C3,$96,$3D,$AF
	.dc.b	$7C,$B2,$3F,$D5,$18,$71,$BF,$F3,$A3,$60,$3E,$A5,$F2,$9F,$CC,$3D
	.dc.b	$AF,$D5,$4F,$FA,$C6,$7F,$D2,$4D,$F6,$B5,$E0,$C8,$C1,$0F,$E4,$35
	.dc.b	$C9,$B1,$B6,$9D,$7C,$4F,$E7,$FF,$01,$E2,$8F,$F6,$07,$D2,$3E,$07
	.dc.b	$FA,$1B,$63,$FD,$89,$3B,$7F,$97,$B7,$87,$E3,$C7,$BA,$78,$3F,$CD
	.dc.b	$1B,$1A,$84,$F1,$DB,$FD,$34,$D8,$F8,$ED,$77,$AF,$93,$E2,$7F,$AB
	.dc.b	$F8,$CF,$8F,$FE,$B3,$F5,$3F,$76,$FF,$DB,$4D,$7F,$A2,$CE,$DF,$F6
	.dc.b	$F3,$5F,$E9,$31,$BC,$BC,$CD,$CB,$FD,$5D,$5B,$9F,$53,$63,$37,$E4
	.dc.b	$3C,$7F,$21,$FD,$B3,$EA,$5F,$47,$E6,$FA,$99,$9F,$FB,$BE,$72,$D9
	.dc.b	$CB,$E4,$9E,$51,$23,$44,$3D,$3E,$4F,$F7,$AF,$A9,$FC,$16,$9E,$AF
	.dc.b	$67,$E2,$99,$E5,$97,$F1,$A3,$2F,$FC,$01,$AE,$3A,$FF,$B3,$5B,$1F
	.dc.b	$9A,$6A,$6E,$EF,$A1,$F1,$7C,$3F,$D4,$F6,$A5,$F5,$8C,$BF,$5A,$32
	.dc.b	$FF,$7A,$B4,$FF,$85,$3E,$A5,$CF,$7A,$A5,$FE,$B3,$6F,$C6,$56,$A3
	.dc.b	$1A,$7F,$B9,$63,$11,$FE,$06,$6C,$F0,$BF,$F1,$0F,$D4,$BD,$BF,$F1
	.dc.b	$2D,$BE,$A1,$F5,$3F,$EE,$FA,$D6,$1F,$BC,$9C,$D7,$EF,$B1,$EF,$18
	.dc.b	$84,$62,$F9,$61,$DA,$B5,$87,$4B,$F0,$57,$A7,$1F,$F7,$F5,$73,$C3
	.dc.b	$95,$7F,$0C,$BE,$13,$CB,$0E,$98,$71,$AF,$33,$5E,$A7,$18,$DB,$71
	.dc.b	$DB,$EB,$C7,$C3,$5C,$6F,$B5,$6A,$98,$78,$AB,$44,$8E,$87,$80,$DE
	.dc.b	$1E,$07,$E2,$E1,$D9,$F0,$1D,$6B,$A3,$CB,$0F,$85,$3B,$8E,$26,$EB
	.dc.b	$A5,$7C,$B8,$FF,$59,$B5,$FA,$38,$7F,$AC,$DF,$1B,$63,$FC,$61,$F2
	.dc.b	$17,$DD,$34,$95,$C4,$7C,$2D,$69,$97,$04,$7E,$17,$C9,$BE,$9F,$F9
	.dc.b	$4E,$9F,$87,$9F,$E8,$F5,$F3,$6B,$5B,$FC,$F9,$C9,$CA,$F8,$ED,$FE
	.dc.b	$CD,$BE,$11,$89,$FF,$69,$C6,$2B,$94,$79,$B0,$FC,$B6,$1D,$A3,$EB
	.dc.b	$1D,$A3,$37,$CE,$3F,$D9,$C7,$86,$B9,$DF,$58,$E2,$75,$C3,$FB,$0A
	.dc.b	$C3,$B6,$1E,$F8,$D1,$FD,$9A,$33,$3E,$B8,$EC,$74,$AE,$D5,$C5,$B3
	.dc.b	$DA,$BD,$5A,$7F,$4F,$6E,$1C,$EF,$D6,$FF,$06,$C3,$C1,$8F,$F5,$AB
	.dc.b	$E0,$D9,$E1,$87,$58,$C5,$79,$FA,$6B,$ED,$73,$F1,$AF,$A8,$FF,$42
	.dc.b	$61,$DE,$B4,$DC,$FF,$A2,$BE,$3C,$9F,$52,$EF,$EB,$5A,$BF,$F5,$15
	.dc.b	$6A,$72,$4B,$F6,$63,$FE,$9B,$F1,$71,$7F,$EA,$6C,$3A,$57,$86,$FC
	.dc.b	$FE,$AF,$F8,$06,$30,$4A,$FF,$58,$61,$EB,$F8,$AE,$79,$7F,$80,$23
	.dc.b	$15,$ED,$F2,$3F,$9C,$B7,$AB,$E3,$AF,$A7,$57,$C8,$AD,$E3,$FE,$BA
	.dc.b	$6E,$1F,$E0,$76,$DF,$F8,$A8,$E5,$5C,$70,$91,$F6,$CE,$DC,$38,$C3
	.dc.b	$FD,$91,$18,$3E,$97,$D4,$BD,$DF,$B2,$D8,$E5,$9B,$CF,$CD,$49,$E2
	.dc.b	$FB,$27,$51,$82,$5F,$82,$3A,$3F,$ED,$6B,$BA,$7F,$8A,$2B,$E2,$5F
	.dc.b	$E2,$9F,$F6,$01,$F8,$72,$D7,$3F,$07,$5D,$52,$74,$F8,$E3,$13,$FE
	.dc.b	$E4,$9E,$37,$DF,$EA,$7F,$CA,$31,$C0,$F9,$0E,$27,$FD,$D7,$F3,$9F
	.dc.b	$86,$6F,$1F,$BE,$7F,$3F,$39,$BF,$F1,$E3,$6F,$F7,$ED,$BF,$F3,$5E
	.dc.b	$8B,$E3,$84,$BF,$9F,$E2,$DB,$F9,$0F,$EB,$3E,$A1,$F5,$3F,$AE,$FA
	.dc.b	$87,$8B,$98,$C3,$30,$FD,$74,$E9,$FD,$85,$72,$73,$3C,$FD,$F9,$18
	.dc.b	$38,$BD,$FD,$19,$1F,$B9,$FA,$9E,$59,$B8,$3F,$CB,$99,$7C,$97,$FC
	.dc.b	$BD,$D0,$F3,$7F,$82,$B7,$0F,$2E,$38,$9F,$E1,$3F,$52,$F4,$E9,$C3
	.dc.b	$DA,$B7,$FC,$27,$E9,$BF,$D3,$8D,$F6,$BF,$89,$FD,$43,$EA,$5F,$53
	.dc.b	$FC,$52,$F7,$C1,$E7,$C3,$F9,$9C,$63,$7F,$F3,$A3,$63,$FD,$2F,$87
	.dc.b	$3B,$E5,$8C,$9F,$E5,$1E,$2B,$EA,$7F,$CE,$51,$E3,$32,$E3,$2F,$00
	.dc.b	$EC,$FF,$51,$FD,$4F,$F2,$A7,$FD,$4B,$D7,$65,$36,$30,$DB,$FE,$71
	.dc.b	$5C,$9B,$5F,$8F,$8F,$F5,$7F,$D4,$F3,$3A,$70,$5A,$DC,$5A,$F2,$D7
	.dc.b	$FA,$DA,$B1,$BF,$FA,$5F,$E9,$DE,$BF,$F3,$E6,$E3,$FC,$BB,$EA,$7F
	.dc.b	$DA,$CF,$F3,$D8,$CE,$3F,$00,$38,$BE,$97,$FD,$1A,$F8,$D6,$EB,$57
	.dc.b	$FD,$A1,$B1,$FC,$A2,$7F,$91,$BF,$EC,$C8,$CF,$42,$73,$97,$F8,$62
	.dc.b	$FE,$77,$A7,$F8,$D7,$E4,$BF,$ED,$E6,$F3,$69,$DB,$5C,$9C,$3E,$A7
	.dc.b	$99,$D3,$E3,$6C,$64,$99,$FC,$CB,$4E,$A3,$DD,$F3,$7F,$2C,$FE,$D7
	.dc.b	$C9,$FE,$C4,$F9,$8C,$61,$F2,$E6,$D7,$E2,$AB,$53,$92,$47,$F6,$4F
	.dc.b	$A9,$F9,$7C,$3F,$66,$FF,$BB,$A8,$CD,$7F,$BD,$35,$FF,$1E,$FC,$8F
	.dc.b	$C6,$3E,$3F,$D4,$D9,$F8,$F1,$71,$F6,$4D,$C3,$FD,$C3,$7F,$E4,$6F
	.dc.b	$99,$F9,$FE,$AF,$D0,$6E,$3F,$41,$B8,$7C,$56,$DF,$FB,$97,$EA,$7F
	.dc.b	$B9,$4F,$E2,$5B,$1F,$EE,$ED,$7F,$DD,$3E,$3F,$F8,$42,$73,$87,$F8
	.dc.b	$B9,$B6,$0E,$2B,$8C,$FF,$84,$E3,$73,$AC,$7F,$C3,$3F,$52,$F9,$8E
	.dc.b	$D1,$87,$F1,$7F,$53,$F8,$CF,$A9,$B8,$FE,$31,$B8,$7E,$05,$BA,$7C
	.dc.b	$6E,$54,$7E,$53,$E4,$BF,$2B,$EC,$5C,$3F,$2A,$DC,$3F,$14,$DC,$39
	.dc.b	$3F,$E3,$0F,$A8,$FE,$7D,$3F,$AF,$6C,$AB,$FC,$A3,$1F,$C4,$E3,$13
	.dc.b	$F1,$63,$73,$AE,$0C,$FC,$17,$C1,$AD,$7C,$7E,$36,$D5,$FF,$9D,$9E
	.dc.b	$B1,$99,$DC,$E5,$A7,$FA,$1A,$78,$B4,$CD,$77,$58,$2F,$A4,$61,$B1
	.dc.b	$F0,$27,$83,$BB,$F0,$57,$2F,$7B,$67,$0F,$A9,$B3,$AA,$4F,$6F,$A9
	.dc.b	$FD,$BF,$83,$98,$C7,$77,$F8,$E3,$1F,$40,$AD,$B8,$6E,$5E,$4A,$E3
	.dc.b	$7E,$49,$CE,$9F,$E8,$46,$DF,$D3,$AD,$BC,$F0,$F0,$61,$F1,$6B,$73
	.dc.b	$ED,$CB,$A5,$6B,$0E,$8D,$9F,$D1,$57,$0C,$BD,$F7,$C3,$29,$5D,$DF
	.dc.b	$81,$18,$C7,$C1,$87,$3A,$F8,$0D,$BE,$71,$C4,$93,$C2,$F9,$C6,$31
	.dc.b	$F2,$C7,$57,$F2,$B9,$7C,$0A,$E1,$5B,$3E,$26,$3B,$7D,$78,$F8,$F2
	.dc.b	$F6,$5F,$1C,$3A,$D7,$13,$E3,$16,$7E,$49,$C2,$FA,$47,$B3,$1B,$97
	.dc.b	$B3,$2F,$19,$30,$EA,$FD,$DB,$E3,$87,$4C,$BE,$26,$5E,$3C,$BC,$F9
	.dc.b	$74,$8C,$47,$73,$CD,$97,$7C,$BA,$92,$38,$9F,$90,$8D,$38,$CB,$C8
	.dc.b	$D9,$F9,$79,$79,$31,$F8,$C7,$91,$C6,$3E,$36,$D7,$8B,$1F,$2E,$5E
	.dc.b	$BA,$FB,$38,$F6,$AF,$43,$72,$F1,$C6,$32,$F4,$34,$FB,$27,$D1,$CB
	.dc.b	$C7,$7D,$27,$78,$77,$AF,$76,$3E,$2B,$DC,$63,$2F,$3E,$5E,$1A,$D6
	.dc.b	$3E,$BA,$EC,$D9,$F4,$60,$FA,$97,$C0,$7F,$BD,$2F,$F6,$51,$87,$07
	.dc.b	$D5,$35,$3C,$67,$31,$FC,$06,$78,$57,$ED,$5B,$9F,$4A,$FE,$27,$F5
	.dc.b	$0F,$A9,$FE,$A9,$87,$F0,$7B,$DB,$FB,$66,$BD,$6B,$8C,$F0,$C3,$B4
	.dc.b	$7F,$0B,$AD,$E3,$E5,$BF,$E2,$B8,$77,$75,$51,$F0,$3D,$1C,$5F,$47
	.dc.b	$27,$2C,$3C,$18,$FA,$E3,$33,$B3,$F9,$C5,$F7,$8E,$07,$49,$E1,$3A
	.dc.b	$AE,$73,$FD,$D6,$B8,$4F,$3A,$D9,$F1,$A3,$2D,$F8,$3E,$34,$97,$CF
	.dc.b	$61,$B1,$F4,$27,$04,$70,$7F,$46,$FA,$9F,$F6,$F3,$7E,$13,$4F,$CA
	.dc.b	$9E,$F3,$A8,$FE,$82,$DC,$7E,$0C,$F0,$7F,$AD,$46,$71,$FB,$FE,$F3
	.dc.b	$C6,$FC,$0F,$47,$F7,$2C,$52,$FA,$9C,$F2,$70,$6C,$FF,$86,$E7,$93
	.dc.b	$FE,$1D,$AE,$51,$88,$FF,$38,$3C,$DD,$57,$36,$9C,$EF,$AB,$E2,$F6
	.dc.b	$6D,$3E,$86,$98,$89,$3C,$8F,$4C,$69,$C4,$FF,$B0,$5D,$7D,$43,$EA
	.dc.b	$7A,$FA,$32,$73,$1E,$29,$D4,$72,$9D,$3F,$E8,$B8,$CB,$67,$81,$1F
	.dc.b	$0F,$0F,$B3,$EA,$7F,$BE,$C7,$9D,$7C,$0E,$4D,$75,$C9,$D4,$65,$E1
	.dc.b	$1C,$FE,$46,$47,$37,$B7,$C4,$62,$3F,$D4,$16,$9C,$70,$FD,$F8,$57
	.dc.b	$D0,$FA,$97,$9B,$E7,$F4,$A6,$3F,$36,$72,$D9,$CB,$DA,$79,$44,$8D
	.dc.b	$10,$F1,$F6,$FD,$EF,$A9,$7D,$4D,$24,$FD,$B8,$7C,$CE,$23,$FD,$75
	.dc.b	$1C,$1D,$4F,$39,$E9,$F3,$BD,$7E,$96,$57,$38,$CB,$D6,$32,$F4,$8F
	.dc.b	$C8,$F8,$DF,$E4,$1D,$FF,$49,$F5,$0F,$A9,$FC,$AB,$6F,$F2,$15,$BE
	.dc.b	$DF,$E5,$6D,$8F,$F4,$04,$ED,$FF,$69,$36,$FF,$61,$3B,$9F,$F5,$AF
	.dc.b	$7F,$88,$8F,$E7,$6B,$F5,$95,$E4,$ED,$F1,$92,$75,$DB,$5E,$F1,$67
	.dc.b	$D7,$3E,$AE,$9F,$5F,$C8,$CB,$ED,$19,$7C,$11,$97,$D3,$1F,$1B,$AB
	.dc.b	$F6,$FE,$CF,$F1,$E9,$FF,$77,$37,$0F,$6C,$EE,$7F,$6D,$39,$D7,$FA
	.dc.b	$8B,$63,$FD,$4D,$AF,$E9,$1F,$9B,$5C,$5F,$F7,$D4,$65,$EF,$D7,$FC
	.dc.b	$23,$E9,$9C,$5F,$C2,$BE,$A5,$F5,$34,$D4,$6F,$D3,$E1,$F5,$33,$1F
	.dc.b	$E1,$58,$27,$F8,$9F,$D4,$7E,$9A,$73,$1C,$5F,$E2,$11,$95,$BF,$F2
	.dc.b	$B8,$7C,$4F,$F1,$CF,$A9,$7C,$35,$3A,$C7,$F2,$0F,$A9,$FE,$79,$7C
	.dc.b	$63,$38,$7F,$1D,$9C,$F1,$FF,$78,$6C,$7F,$B3,$5A,$9B,$F5,$2F,$92
	.dc.b	$FE,$5B,$ED,$7E,$B2,$BF,$94,$4F,$FA,$75,$BE,$D6,$BC,$09,$18,$21
	.dc.b	$EE,$C7,$FC,$13,$7C,$E3,$07,$FB,$7D,$C6,$7F,$D3,$5B,$3C,$2B,$F9
	.dc.b	$FE,$1F,$E1,$FB,$E2,$3E,$A7,$D0,$FF,$3D,$FA,$37,$C0,$FF,$46,$6C
	.dc.b	$7F,$41,$9D,$BC,$9B,$97,$F5,$D9,$DC,$F6,$6C,$7F,$61,$9D,$E3,$FE
	.dc.b	$24,$6D,$7F,$BE,$5A,$F9,$74,$F1,$7C,$4F,$F5,$8F,$11,$F5,$3F,$D6
	.dc.b	$BE,$9F,$EF,$DF,$F5,$56,$BE,$E9,$DB,$FD,$65,$AF,$C2,$8D,$E5,$FD
	.dc.b	$D5,$B9,$7F,$8E,$2B,$73,$FD,$E5,$B1,$9B,$F2,$1E,$3F,$90,$FE,$DB
	.dc.b	$F5,$2F,$A3,$F3,$7D,$4C,$CF,$FB,$54,$E5,$B3,$97,$C9,$3C,$A2,$46
	.dc.b	$88,$7A,$7C,$7F,$EF,$7F,$53,$F8,$08,$0D,$7F,$C1,$EF,$C5,$F6,$37
	.dc.b	$CB,$23,$04,$3F,$0B,$5C,$D3,$49,$1F,$47,$E8,$7C,$3F,$0F,$F4,$FD
	.dc.b	$A9,$7F,$E0,$7B,$4E,$A3,$EB,$4E,$9E,$F1,$FE,$14,$FA,$97,$3D,$EA
	.dc.b	$95,$FE,$11,$9C,$B6,$78,$11,$FF,$0B,$7D,$4F,$F7,$3E,$6F,$87,$F5
	.dc.b	$2F,$A9,$FF,$39,$CE,$69,$39,$7E,$4E,$1D,$9C,$D7,$F9,$EA,$31,$3B
	.dc.b	$95,$73,$8C,$36,$B8,$61,$FE,$82,$9E,$17,$CA,$7C,$7E,$E7,$F1,$CC
	.dc.b	$3B,$9F,$6B,$0E,$4F,$93,$0E,$55,$D2,$3D,$AF,$2C,$6C,$71,$BE,$98
	.dc.b	$F8,$2F,$8C,$F3,$36,$7A,$A7,$77,$CE,$3C,$63,$1F,$43,$C4,$CE,$3E
	.dc.b	$67,$96,$3D,$27,$71,$C7,$0F,$0B,$A3,$8E,$1E,$07,$A6,$3E,$77,$CB
	.dc.b	$87,$53,$EC,$3F,$EB,$B6,$BE,$73,$13,$F2,$DD,$7B,$6F,$8D,$FF,$5D
	.dc.b	$D7,$F8,$D1,$B1,$99,$FF,$61,$CE,$5F,$F6,$08,$F8,$3A,$D3,$2E,$08
	.dc.b	$FC,$8F,$91,$7D,$3B,$B7,$E7,$CF,$ED,$AB,$E8,$D6,$A7,$33,$D7,$3F
	.dc.b	$C7,$4E,$63,$E9,$9C,$4E,$D7,$D5,$C9,$C7,$7F,$F6,$8C,$F1,$AF,$CC
	.dc.b	$4F,$17,$AD,$71,$73,$1C,$4F,$E1,$73,$C6,$BF,$DC,$F8,$7A,$2F,$84
	.dc.b	$62,$BC,$78,$7A,$AB,$4D,$8C,$9E,$68,$CB,$4C,$E1,$E2,$8C,$BE,$69
	.dc.b	$FF,$49,$61,$F0,$F7,$FF,$56,$5F,$AF,$0F,$7E,$3F,$0E,$7F,$D6,$15
	.dc.b	$BB,$C9,$FE,$9B,$BE,$07,$BA,$B8,$47,$C8,$E9,$B3,$F0,$9B,$F1,$8F
	.dc.b	$A9,$FF,$46,$D7,$92,$77,$4C,$3F,$A1,$4F,$FA,$CB,$17,$D4,$BB,$F7
	.dc.b	$5A,$BF,$EF,$B5,$A9,$C9,$2F,$FD,$77,$93,$D3,$FE,$EF,$1F,$12,$75
	.dc.b	$7F,$DF,$B0,$ED,$3C,$2B,$73,$9A,$DE,$1F,$89,$C3,$D9,$F1,$38,$C3
	.dc.b	$FA,$5C,$61,$FF,$61,$7C,$07,$F3,$C6,$F4,$FC,$65,$F5,$0A,$F8,$75
	.dc.b	$BC,$BF,$C2,$2D,$C7,$FD,$C8,$DB,$FE,$B2,$72,$AE,$38,$48,$FF,$6D
	.dc.b	$9D,$B8,$31,$87,$F8,$92,$30,$7F,$B5,$FE,$A5,$EE,$7D,$46,$C7,$2C
	.dc.b	$DE,$E6,$6A,$4F,$17,$D9,$3A,$8C,$12,$FC,$11,$D1,$FF,$16,$DD,$D3
	.dc.b	$F6,$95,$F1,$2F,$F0,$AF,$FB,$10,$FF,$76,$16,$B9,$F8,$1A,$E9,$93
	.dc.b	$A7,$C7,$18,$9F,$F1,$DC,$F1,$BE,$FF,$53,$FE,$53,$8E,$07,$C8,$71
	.dc.b	$3F,$E4,$2F,$98,$FC,$2B,$78,$BD,$F3,$F9,$A9,$CD,$FF,$0D,$A4,$62
	.dc.b	$FF,$C1,$F1,$8B,$E3,$1F,$E1,$2A,$E1,$5B,$AF,$F0,$AB,$67,$F0,$FE
	.dc.b	$27,$7B,$F9,$EF,$EA,$BE,$A1,$F5,$3F,$AC,$FA,$87,$89,$98,$C3,$30
	.dc.b	$F0,$1E,$49,$DB,$8C,$BF,$C4,$B7,$D2,$31,$09,$DB,$FE,$66,$7F,$7A
	.dc.b	$DF,$80,$EC,$D9,$E3,$3F,$33,$4F,$8C,$D8,$F8,$BF,$53,$DE,$BA,$7A
	.dc.b	$3F,$C1,$BA,$1E,$65,$3D,$D1,$6F,$CB,$F5,$2F,$4B,$F8,$6B,$63,$D3
	.dc.b	$7F,$EA,$07,$73,$FC,$4B,$EA,$3F,$D4,$B1,$EE,$6E,$3F,$E8,$AF,$A8
	.dc.b	$7D,$4B,$EA,$78,$7B,$93,$0E,$0F,$F2,$2A,$DC,$E6,$FF,$D1,$78,$74
	.dc.b	$E0,$FE,$A5,$E2,$3E,$A7,$FC,$EF,$97,$23,$E9,$46,$7D,$7F,$E5,$BD
	.dc.b	$5F,$E5,$8F,$A9,$B9,$7E,$1E,$31,$7F,$65,$FF,$2E,$F5,$EA,$B6,$D9
	.dc.b	$D5,$71,$1F,$53,$D0,$EB,$F0,$16,$77,$16,$7F,$D5,$B8,$7F,$9A,$AB
	.dc.b	$1A,$7E,$87,$EA,$5E,$B7,$F3,$E6,$E3,$FC,$D3,$EA,$7F,$DB,$8F,$E3
	.dc.b	$23,$38,$FF,$AC,$1B,$7C,$EF,$FA,$5D,$F0,$AD,$47,$53,$3E,$4D,$C3
	.dc.b	$B3,$F2,$3D,$4F,$F0,$64,$7C,$57,$F0,$95,$EA,$72,$DE,$7E,$13,$C5
	.dc.b	$AF,$17,$A7,$D4,$C8,$E9,$F1,$96,$BA,$4D,$CF,$FC,$21,$B7,$F8,$B6
	.dc.b	$B1,$7F,$C1,$35,$7C,$D7,$C7,$3F,$B5,$F2,$7F,$B4,$7E,$5B,$15,$F1
	.dc.b	$E6,$E7,$FE,$DE,$AD,$4E,$49,$1F,$D9,$BE,$A7,$E4,$67,$F5,$6F,$FB
	.dc.b	$BA,$8C,$D7,$F1,$0D,$9F,$21,$F2,$4F,$95,$F5,$36,$7E,$54,$5C,$7E
	.dc.b	$54,$DC,$3D,$F6,$F9,$8F,$A5,$F0,$FF,$4D,$B8,$FD,$36,$E1,$F4,$DB
	.dc.b	$7F,$CA,$FE,$A7,$FB,$9B,$6F,$E2,$FC,$47,$F1,$28,$C6,$1F,$C5,$AB
	.dc.b	$6F,$F3,$06,$E1,$FE,$0F,$AE,$13,$9B,$FF,$09,$D6,$BC,$37,$D4,$BE
	.dc.b	$5F,$35,$FE,$1C,$8C,$5F,$E2,$FE,$A7,$F1,$DF,$53,$F8,$F6,$C6,$4F
	.dc.b	$C8,$46,$5A,$7E,$4F,$E4,$BF,$2D,$D7,$F9,$76,$C6,$4F,$CC,$46,$5A
	.dc.b	$7F,$34,$FA,$8F,$5E,$DF,$ED,$C6,$A7,$F8,$6A,$DF,$FA,$B4,$68,$FE
	.dc.b	$77,$F0,$9E,$F9,$E5,$F2,$1F,$E6,$78,$C4,$73,$F4,$F1,$BF,$F5,$CB
	.dc.b	$5E,$2B,$8E,$3F,$E7,$9A,$D4,$E6,$79,$4E,$5D,$9F,$E8,$99,$DC,$66
	.dc.b	$3D,$53,$C6,$BC,$F5,$A9,$E0,$7C,$1F,$80,$D1,$BF,$A9,$B8,$73,$BF
	.dc.b	$F2,$23,$D7,$EA,$1D,$FF,$E3,$3C,$39,$9F,$27,$0F,$44,$EB,$93,$85
	.dc.b	$6F,$5F,$8B,$39,$BE,$38,$F6,$79,$C6,$67,$75,$C7,$3F,$F0,$2C,$E5
	.dc.b	$F9,$98,$F8,$27,$56,$15,$B7,$0D,$7D,$19,$79,$B2,$91,$E2,$AE,$6E
	.dc.b	$8C,$46,$B1,$F2,$BF,$1F,$0F,$13,$9C,$7C,$F7,$CB,$2F,$75,$F0,$C3
	.dc.b	$99,$1F,$C4,$C6,$AB,$8D,$F3,$AF,$4D,$F3,$C3,$BC,$71,$AD,$BF,$32
	.dc.b	$33,$87,$72,$4F,$C8,$C7,$A5,$7C,$6B,$F7,$E5,$F1,$72,$F1,$4F,$83
	.dc.b	$2D,$CE,$72,$EB,$8F,$83,$2E,$B7,$F8,$DC,$B7,$97,$43,$D0,$E3,$1F
	.dc.b	$0E,$5D,$AB,$C5,$19,$7D,$53,$6F,$AC,$7D,$36,$E5,$F6,$CF,$24,$EF
	.dc.b	$1F,$21,$EF,$AF,$8C,$7A,$CA,$73,$76,$E6,$BB,$E1,$E5,$BE,$D8,$FA
	.dc.b	$F1,$FC,$06,$3D,$F0,$E1,$8F,$7C,$7C,$B7,$B9,$F5,$9E,$0C,$39,$E1
	.dc.b	$CA,$31,$3B,$AF,$C1,$E5,$F3,$A7,$A9,$0F,$41,$6F,$9B,$DA,$7C,$37
	.dc.b	$F3,$AB,$94,$FC,$E3,$E0,$E5,$ED,$C3,$F0,$3F,$52,$F8,$2F,$F3,$9D
	.dc.b	$7E,$C2,$30,$E0,$E4,$6B,$24,$6A,$30,$F0,$CB,$F8,$0B,$73,$E9,$7F
	.dc.b	$C6,$7E,$9D,$F5,$3F,$E8,$2B,$DC,$FF,$11,$BE,$18,$7E,$E9,$AA,$F8
	.dc.b	$BB,$98,$FF,$0C,$9D,$CE,$A3,$13,$AA,$EF,$1F,$C7,$27,$51,$82,$61
	.dc.b	$E8,$9D,$61,$E7,$8C,$4E,$A7,$CD,$7E,$DA,$D4,$60,$F0,$92,$7F,$A4
	.dc.b	$DF,$82,$FF,$D4,$6D,$9E,$13,$D5,$F8,$15,$D1,$A7,$AE,$BC,$67,$28
	.dc.b	$D5,$3E,$16,$FD,$4C,$BF,$A5,$1A,$23,$83,$FA,$77,$D4,$FF,$B7,$9B
	.dc.b	$F0,$DA,$7E,$44,$F8,$E7,$51,$FD,$19,$B9,$7B,$A7,$83,$FD,$A2,$33
	.dc.b	$97,$FB,$0B,$E0,$3F,$C7,$F7,$FE,$AC,$5F,$53,$FE,$C6,$E8,$FF,$65
	.dc.b	$46,$27,$5E,$36,$79,$B9,$36,$3F,$A2,$4E,$DF,$EF,$D3,$DE,$F5,$09
	.dc.b	$F0,$BA,$38,$FB,$9A,$87,$D0,$D3,$11,$27,$D1,$19,$7D,$51,$DB,$EA
	.dc.b	$5F,$53,$CF,$D1,$95,$FE,$CA,$9C,$B6,$72,$F8,$CF,$34,$F3,$89,$1B
	.dc.b	$21,$BC,$3E,$9E,$9F,$EF,$1F,$A9,$75,$FC,$05,$F0,$74,$C4,$4B,$FF
	.dc.b	$6E,$CE,$9B,$39,$7C,$90,$F9,$3E,$84,$8E,$3F,$13,$5D,$C5,$C3,$C9
	.dc.b	$3F,$4B,$EA,$7F,$CC,$95,$E0,$F4,$BC,$C6,$7C,$D9,$19,$21,$F1,$67
	.dc.b	$8B,$63,$6D,$3D,$7D,$BF,$80,$FA,$97,$D4,$D2,$4F,$FB,$F2,$1F,$35
	.dc.b	$88,$FA,$51,$C1,$D4,$F3,$9E,$9F,$3D,$D7,$E9,$65,$73,$8C,$BD,$63
	.dc.b	$2F,$48,$FC,$9F,$8B,$FE,$4C,$D9,$D5,$FE,$1D,$A7,$D4,$3E,$A7,$F3
	.dc.b	$2D,$8F,$F1,$15,$70,$7F,$C4,$97,$CE,$F9,$36,$3F,$C5,$15,$C5,$FC
	.dc.b	$73,$6B,$94,$7F,$90,$BB,$BC,$A4,$7B,$DF,$48,$F1,$F7,$79,$3B,$25
	.dc.b	$68,$EE,$FA,$3B,$73,$1E,$08,$B8,$78,$31,$FD,$7C,$F1,$3E,$0F,$85
	.dc.b	$F0,$3E,$4A,$46,$67,$F3,$F9,$FC,$76,$C7,$BA,$76,$FB,$AB,$E7,$F5
	.dc.b	$7F,$02,$F6,$1E,$2D,$33,$5E,$1D,$FF,$44,$FE,$CE,$76,$F9,$5A,$FE
	.dc.b	$DE,$37,$9F,$CA,$6E,$7F,$4B,$0E,$B5,$CF,$B7,$F8,$9F,$D3,$39,$3F
	.dc.b	$8A,$FD,$4B,$EA,$69,$A8,$DF,$A7,$C3,$EA,$66,$3F,$1F,$04,$FF,$1E
	.dc.b	$FA,$97,$83,$C2,$B5,$39,$8D,$7C,$77,$F2,$4F,$A9,$7C,$3D,$33,$7F
	.dc.b	$4F,$EA,$7E,$6A,$C5,$7A,$8F,$E3,$13,$A2,$73,$7F,$30,$78,$FD,$4B
	.dc.b	$E5,$3F,$99,$FB,$5F,$B0,$BF,$C0,$4F,$FA,$A9,$BE,$D6,$BC,$39,$18
	.dc.b	$21,$FC,$9E,$B9,$36,$36,$D3,$AF,$8D,$FD,$0B,$E0,$7C,$51,$F8,$FF
	.dc.b	$A4,$7C,$0F,$F4,$66,$C7,$E4,$E7,$6F,$F3,$56,$F6,$FE,$A2,$77,$3F
	.dc.b	$CE,$5B,$1F,$AC,$9D,$F6,$7E,$F5,$B5,$FE,$9B,$6A,$6D,$3E,$2B,$FA
	.dc.b	$CF,$90,$F8,$FF,$EB,$7F,$4F,$F6,$30,$FD,$43,$5F,$E9,$53,$B7,$F5
	.dc.b	$8D,$7F,$A7,$46,$F2,$FE,$02,$DC,$BF,$89,$56,$E7,$F8,$2B,$63,$37
	.dc.b	$E4,$3C,$7F,$1F,$FD,$BB,$EA,$5F,$45,$E6,$FA,$99,$9F,$EF,$67,$2D
	.dc.b	$9C,$BE,$49,$E5,$12,$34,$43,$D3,$E5,$7F,$7C,$FA,$9F,$C0,$E9,$EA
	.dc.b	$F5,$FE,$29,$9E,$49,$7F,$1A,$32,$FF,$0F,$AE,$3A,$FF,$31,$6C,$7F
	.dc.b	$B3,$DA,$9B,$BB,$E8,$3C,$5F,$0F,$F5,$3D,$B9,$7D,$63,$2F,$D6,$8C
	.dc.b	$BF,$DF,$ED,$3F,$E1,$6F,$A9,$72,$DE,$99,$7F,$C9,$67,$2D,$9E,$04
	.dc.b	$7F,$95,$7D,$4F,$F7,$7D,$FF,$DD,$1F,$52,$FA,$9F,$E9,$D3,$9A,$4E
	.dc.b	$5F,$F7,$76,$1D,$9C,$D7,$F5,$58,$C4,$EE,$55,$CE,$30,$DA,$E1,$87
	.dc.b	$F5,$D9,$E1,$7C,$A7,$C7,$EE,$B1,$F3,$C6,$1B,$8F,$93,$0E,$53,$D2
	.dc.b	$32,$78,$AB,$93,$E8,$06,$6B,$77,$E1,$8E,$87,$2C,$7D,$0F,$06,$A8
	.dc.b	$E0,$DC,$3C,$6E,$EF,$C3,$8F,$C2,$6E,$3E,$79,$F2,$1C,$4F,$04,$69
	.dc.b	$E5,$8F,$9A,$FA,$C6,$CE,$38,$F9,$0F,$B4,$FF,$AF,$1A,$FC,$93,$FD
	.dc.b	$78,$E1,$F6,$78,$EF,$90,$FF,$60,$47,$DA,$8C,$3C,$1A,$F5,$6B,$FE
	.dc.b	$C7,$F8,$4E,$2F,$CE,$6A,$69,$1F,$83,$F2,$6F,$A7,$FF,$0D,$6B,$F8
	.dc.b	$B9,$E4,$6A,$33,$AF,$78,$C7,$2F,$E0,$1D,$D7,$E4,$EF,$A4,$E6,$7C
	.dc.b	$47,$03,$E8,$9C,$2B,$FC,$61,$3C,$23,$27,$0A,$E9,$7E,$28,$FF,$74
	.dc.b	$E1,$F9,$0A,$E1,$5E,$99,$E1,$7D,$2F,$84,$7F,$B8,$4F,$14,$EE,$F9
	.dc.b	$4E,$A7,$C8,$7E,$62,$FA,$DF,$C1,$BF,$1E,$1E,$DC,$38,$B2,$FF,$D1
	.dc.b	$57,$EB,$D7,$FC,$7B,$1D,$04,$66,$7C,$C7,$F8,$C5,$C4,$7B,$8F,$91
	.dc.b	$3F,$E4,$68,$FE,$7B,$3B,$E9,$FF,$29,$4F,$F9,$4F,$E3,$5F,$53,$FE
	.dc.b	$90,$AD,$13,$0F,$F5,$85,$3E,$A5,$DF,$CA,$BE,$9E,$36,$FF,$82,$56
	.dc.b	$A7,$CD,$59,$9F,$5B,$BC,$7F,$D4,$3E,$26,$73,$E1,$5F,$12,$B0,$6F
	.dc.b	$D5,$FF,$00,$E1,$D0,$93,$FE,$C5,$C3,$D7,$F1,$5C,$27,$FC,$BD,$18
	.dc.b	$7F,$CB,$1E,$3F,$F3,$A6,$F5,$7A,$57,$D4,$AB,$D1,$5B,$C3,$FD,$80
	.dc.b	$DD,$FF,$CE,$6D,$BF,$E3,$27,$2A,$E3,$84,$8F,$F3,$81,$DB,$93,$18
	.dc.b	$7F,$B7,$63,$07,$F9,$B7,$EA,$5E,$EF,$C9,$6C,$72,$CD,$E7,$E6,$A4
	.dc.b	$F1,$7D,$93,$A8,$C1,$2F,$C1,$1D,$1F,$F6,$E5,$DD,$3F,$D3,$15,$F1
	.dc.b	$2F,$ED,$3F,$E5,$83,$FD,$04,$5A,$E7,$E0,$EB,$A6,$4E,$9F,$1C,$62
	.dc.b	$7F,$DD,$53,$C6,$FB,$FD,$4F,$F9,$46,$38,$1F,$21,$C4,$FF,$BC,$7E
	.dc.b	$6F,$F0,$F5,$FE,$A4,$6D,$7D,$2C,$D7,$F9,$19,$CD,$FF,$4E,$6A,$73
	.dc.b	$5F,$82,$9E,$1C,$72,$3E,$87,$8B,$6F,$E4,$BF,$9D,$FA,$87,$D4,$FE
	.dc.b	$7F,$EA,$5E,$2E,$63,$0C,$C3,$FD,$55,$3A,$7F,$D5,$95,$C9,$CC,$F3
	.dc.b	$F7,$E4,$60,$E2,$F7,$F5,$24,$7F,$AD,$7E,$A7,$96,$7F,$DD,$E3,$FD
	.dc.b	$B1,$8F,$CA,$E0,$FF,$65,$F4,$3C,$DF,$DB,$36,$BC,$B8,$E2,$7F,$75
	.dc.b	$F5,$3F,$E5,$F9,$FF,$61,$37,$1E,$C7,$F7,$AB,$FF,$4B,$CE,$BD,$3F
	.dc.b	$E0,$5F,$4D,$FE,$94,$8F,$F6,$53,$6B,$8F,$D4,$3E,$A5,$F5,$3C,$3D
	.dc.b	$C9,$87,$07,$F8,$85,$6E,$73,$7E,$21,$87,$5D,$BF,$DE,$5E,$23,$EA
	.dc.b	$7F,$CE,$D9,$73,$3F,$DC,$91,$9F,$63,$F8,$E7,$5F,$F8,$53,$EA,$6E
	.dc.b	$5F,$8A,$8C,$5F,$FB,$AD,$FF,$0C,$F5,$EA,$B6,$D9,$D5,$71,$1F,$53
	.dc.b	$D0,$EB,$F0,$36,$77,$16,$7E,$6E,$1F,$E2,$6A,$C6,$8B,$FF,$14,$D3
	.dc.b	$E9,$DE,$C7,$F3,$16,$CF,$F2,$4F,$A9,$FF,$69,$BF,$EF,$B8,$CD,$7C
	.dc.b	$96,$DF,$58,$FF,$1A,$C6,$27,$EF,$57,$29,$E1,$7D,$CF,$E7,$D1,$FC
	.dc.b	$96,$38,$C6,$3C,$9F,$9B,$EA,$7A,$6F,$E5,$3F,$8B,$C7,$C1,$9F,$F8
	.dc.b	$86,$77,$CF,$C1,$A7,$02,$E3,$F1,$AD,$F5,$3D,$8E,$FF,$19,$6B,$BC
	.dc.b	$5A,$E1,$37,$1F,$81,$1F,$2F,$E4,$7E,$51,$FD,$7F,$93,$FD,$85,$F3
	.dc.b	$18,$C7,$E5,$4D,$AF,$D1,$D6,$A7,$24,$8F,$ED,$BF,$53,$F2,$70,$EF
	.dc.b	$C2,$C3,$FB,$64,$27,$57,$FC,$67,$47,$C8,$7C,$A3,$E5,$FD,$4D,$9F
	.dc.b	$97,$17,$2F,$89,$37,$0F,$9B,$6F,$99,$FA,$7E,$B7,$D4,$6E,$3F,$51
	.dc.b	$B8,$7C,$D6,$DF,$C6,$FA,$9F,$EE,$B3,$FB,$16,$C7,$EC,$3E,$23,$F8
	.dc.b	$F4,$E7,$0F,$F0,$25,$6B,$0F,$CC,$B6,$7F,$C4,$35,$C2,$FB,$5F,$03
	.dc.b	$FC,$15,$F5,$2F,$98,$C6,$3F,$E0,$08,$C3,$F8,$CF,$A9,$FC,$7F,$D4
	.dc.b	$FE,$41,$B1,$93,$F2,$31,$96,$9F,$94,$F9,$2F,$CB,$F5,$7E,$61,$B1
	.dc.b	$93,$F3,$31,$96,$9F,$E9,$CC,$BF,$17,$F5,$1E,$B8,$FD,$BB,$75,$FF
	.dc.b	$50,$BF,$C5,$1F,$E3,$4E,$53,$F2,$BE,$0B,$E0,$47,$0F,$8F,$FD,$B4
	.dc.b	$62,$37,$E9,$63,$6E,$F5,$67,$86,$1F,$E8,$39,$CB,$63,$C7,$5B,$7C
	.dc.b	$35,$70,$34,$C4,$DA,$E5,$5E,$8F,$80,$E0,$71,$FA,$9B,$3A,$C1,$3D
	.dc.b	$FE,$A5,$DD,$FE,$33,$C3,$FD,$2D,$9F,$C8,$AD,$4E,$74,$F8,$17,$C3
	.dc.b	$CF,$F0,$B7,$0E,$27,$CE,$BE,$33,$AB,$E5,$87,$F2,$AB,$E1,$87,$CB
	.dc.b	$C7,$CB,$96,$F0,$99,$74,$17,$C3,$0F,$35,$6B,$0F,$2E,$3D,$AF,$FD
	.dc.b	$BF,$7B,$7C,$95,$BB,$F1,$56,$E7,$AB,$71,$F1,$C7,$86,$F9,$3A,$C7
	.dc.b	$93,$D0,$D5,$73,$9D,$9C,$2B,$9E,$3E,$6B,$E3,$7C,$CF,$74,$EB,$0F
	.dc.b	$2C,$EA,$F9,$65,$F3,$2F,$BC,$F2,$BE,$11,$F0,$31,$F1,$12,$7F,$05
	.dc.b	$5C,$71,$F1,$57,$23,$C3,$19,$C7,$C3,$19,$9F,$AD,$39,$8E,$D5,$E1
	.dc.b	$8C,$47,$AE,$7A,$9C,$32,$F5,$61,$E3,$9E,$F1,$8B,$E4,$D8,$F0,$61
	.dc.b	$F7,$9A,$78,$0A,$7E,$86,$38,$5F,$82,$78,$36,$78,$35,$E0,$72,$CB
	.dc.b	$EB,$E1,$F8,$1C,$7D,$71,$8C,$7E,$4E,$3D,$B1,$F8,$B8,$FB,$71,$E5
	.dc.b	$8F,$B6,$30,$77,$C3,$94,$FC,$03,$78,$7C,$06,$9E,$CC,$BE,$DB,$F2
	.dc.b	$F0,$EE,$47,$F2,$17,$D7,$EA,$5F,$05,$FE,$16,$9F,$D7,$46,$1C,$1F
	.dc.b	$8F,$35,$8A,$35,$18,$78,$57,$F0,$06,$DF,$4B,$FE,$23,$F5,$0F,$A9
	.dc.b	$FF,$12,$DE,$E7,$F8,$2D,$F0,$C3,$F6,$6D,$57,$C5,$DC,$C7,$F6,$13
	.dc.b	$B9,$D4,$62,$75,$5D,$E3,$F8,$34,$EA,$30,$4C,$3D,$13,$AC,$3C,$F1
	.dc.b	$89,$D4,$F9,$AF,$DB,$5A,$8C,$1E,$12,$4F,$F4,$3B,$F0,$5F,$32,$4F
	.dc.b	$09,$D5,$73,$8F,$82,$DC,$7E,$96,$1E,$CB,$FF,$27,$E1,$D2,$C7,$03
	.dc.b	$47,$58,$DD,$F5,$97,$C2,$DF,$9E,$97,$F6,$23,$44,$70,$7F,$4C,$FA
	.dc.b	$9F,$F7,$0B,$7E,$1B,$4F,$CE,$9F,$54,$EA,$3F,$9B,$37,$4F,$89,$3C
	.dc.b	$1F,$ED,$11,$9C,$BF,$CD,$7F,$01,$FE,$33,$BF,$EF,$5C,$58,$FA,$9F
	.dc.b	$F3,$7F,$47,$F9,$1A,$91,$89,$E3,$E3,$5B,$F8,$B5,$CD,$FF,$4B,$3C
	.dc.b	$1E,$A6,$AB,$82,$8E,$17,$DF,$DB,$CC,$3E,$8A,$98,$89,$5D,$0E,$51
	.dc.b	$A7,$13,$FE,$C1,$75,$F5,$2F,$A9,$EB,$E8,$C9,$CC,$79,$27,$51,$CA
	.dc.b	$74,$FF,$AC,$63,$2D,$9E,$04,$7C,$3C,$3E,$EF,$A9,$FF,$00,$C7,$9D
	.dc.b	$7C,$1E,$4D,$76,$C9,$D4,$65,$E1,$1C,$FE,$52,$47,$37,$B7,$C5,$62
	.dc.b	$3F,$D8,$D6,$9C,$6F,$FE,$80,$85,$7D,$2F,$A9,$FF,$33,$5F,$97,$D3
	.dc.b	$F4,$19,$F3,$64,$64,$87,$FB,$4E,$78,$B6,$36,$D3,$E0,$76,$FE,$03
	.dc.b	$EA,$5F,$53,$49,$3F,$6E,$1F,$37,$88,$FF,$70,$47,$07,$53,$CE,$7A
	.dc.b	$7C,$F7,$5F,$A5,$95,$CE,$32,$F5,$8C,$BD,$23,$F2,$7E,$3F,$F9,$37
	.dc.b	$7F,$F5,$2F,$D4,$3E,$A7,$F2,$ED,$BF,$C8,$56,$FB,$BF,$DA,$AD,$8F
	.dc.b	$E0,$F3,$B7,$FD,$F0,$DB,$FF,$58,$CE,$E7,$FC,$87,$DF,$E2,$23,$F9
	.dc.b	$DA,$FD,$2D,$FF,$B8,$9F,$CE,$47,$4E,$DF,$29,$E6,$ED,$B1,$C6,$2B
	.dc.b	$E2,$8F,$5F,$4F,$B3,$E4,$E4,$F7,$8C,$BE,$18,$CB,$EA,$8F,$8F,$D5
	.dc.b	$FB,$AF,$61,$E2,$D3,$35,$F1,$A7,$0F,$ED,$E7,$6E,$48,$36,$F8,$B3
	.dc.b	$CA,$31,$A7,$F3,$27,$FD,$8B,$D9,$FC,$2B,$E9,$9C,$7F,$C3,$3E,$A5
	.dc.b	$F5,$34,$D4,$6F,$D3,$E1,$F5,$33,$1F,$E1,$58,$27,$F8,$AF,$D4,$7E
	.dc.b	$9A,$73,$1C,$5F,$E2,$11,$97,$CA,$7F,$46,$5F,$1D,$FC,$7B,$EA,$5F
	.dc.b	$0F,$4E,$B1,$FC,$8B,$EA,$7F,$9D,$70,$FF,$BD,$2F,$8E,$1D,$AB,$5C
	.dc.b	$92,$3F,$90,$5F,$0E,$1F,$88,$D8,$FE,$AF,$F5,$2F,$94,$FE,$5F,$EC
	.dc.b	$FE,$B2,$7F,$94,$4F,$FA,$79,$BE,$D6,$BC,$19,$18,$21,$EE,$15,$CD
	.dc.b	$C3,$75,$65,$F5,$E3,$0D,$AF,$E8,$38,$7F,$88,$2B,$77,$FD,$BB,$E0
	.dc.b	$3D,$0F,$F3,$DF,$A4,$7C,$0F,$F4,$86,$C7,$F4,$19,$DB,$E5,$6D,$74
	.dc.b	$84,$F0,$7D,$4D,$8D,$4F,$0C,$7F,$1C,$DA,$F2,$69,$FD,$62,$31,$1D
	.dc.b	$3E,$27,$FA,$DF,$8A,$F8,$FF,$EB,$BF,$4F,$F8,$0A,$FE,$AC,$D7,$E5
	.dc.b	$4E,$DF,$EB,$4D,$7E,$74,$6F,$2F,$53,$72,$FD,$15,$6E,$7D,$CD,$8C
	.dc.b	$DF,$90,$F1,$FC,$7F,$F7,$0F,$A9,$7D,$17,$9B,$EA,$66,$7F,$DA,$E7
	.dc.b	$2D,$9C,$BE,$49,$E5,$12,$34,$43,$D3,$E3,$FF,$7E,$FA,$9F,$C0,$C7
	.dc.b	$F8,$AB,$5F,$F0,$03,$F1,$3D,$9D,$72,$C8,$C1,$0F,$C8,$57,$24,$D2
	.dc.b	$47,$D0,$FA,$1F,$07,$C3,$FD,$2F,$6E,$5F,$F8,$1E,$D3,$A8,$FA,$B3
	.dc.b	$A7,$BC,$7F,$85,$BE,$A7,$F1,$31,$FE,$0F,$7F,$42,$DE,$AD,$74,$48
	.dc.b	$D3,$5F,$D4,$7D,$4F,$F7,$3E,$6F,$F2,$5F,$D4,$3E,$A7,$F8,$E4,$E6
	.dc.b	$93,$97,$FC,$A3,$8F,$89,$CD,$7F,$23,$8C,$4E,$E5,$5C,$E3,$0D,$AE
	.dc.b	$18,$7F,$2B,$9E,$17,$CA,$7C,$7E,$EF,$F3,$3C,$3B,$9F,$E6,$2C,$39
	.dc.b	$3E,$4C,$39,$57,$48,$F6,$BC,$B1,$B1,$C6,$FA,$63,$E0,$BE,$33,$CC
	.dc.b	$D9,$EA,$9D,$DF,$38,$F1,$8C,$7D,$0F,$13,$38,$F9,$9E,$58,$F4,$9D
	.dc.b	$C7,$1C,$3C,$2E,$8E,$38,$78,$1E,$98,$F9,$DF,$2E,$1D,$4F,$B0,$FF
	.dc.b	$B0,$5A,$F9,$CC,$4F,$CB,$75,$EE,$3E,$43,$FD,$8B,$1F,$62,$30,$FF
	.dc.b	$B2,$2B,$FC,$6F,$3A,$6C,$70,$1F,$09,$5A,$65,$C1,$1F,$93,$F2,$8F
	.dc.b	$A7,$FE,$63,$97,$E2,$67,$FD,$3F,$5F,$4A,$B5,$DB,$F5,$27,$27,$2B
	.dc.b	$E3,$BF,$FB,$7A,$F8,$46,$27,$FD,$C9,$18,$AE,$51,$E6,$C3,$F3,$18
	.dc.b	$76,$8F,$BA,$76,$8C,$DF,$38,$FF,$6F,$9E,$1A,$E7,$7D,$63,$89,$D7
	.dc.b	$0F,$F0,$46,$1D,$70,$F7,$4E,$4F,$F5,$EC,$66,$7D,$51,$D4,$E7,$5D
	.dc.b	$6B,$8B,$67,$AE,$1F,$D6,$B7,$FE,$A1,$87,$AD,$FE,$A5,$87,$AE,$B7
	.dc.b	$8F,$FA,$66,$B4,$E2,$7F,$B7,$1B,$3C,$B7,$D6,$3D,$7D,$36,$7D,$2D
	.dc.b	$F8,$D7,$D4,$FF,$A3,$AB,$B4,$EE,$98,$7F,$42,$9F,$F5,$AE,$4F,$A9
	.dc.b	$77,$EE,$B5,$7F,$E1,$EA,$D4,$E4,$97,$FE,$BE,$C9,$EA,$7F,$78,$8F
	.dc.b	$74,$EA,$FF,$C1,$58,$76,$9E,$15,$B9,$CD,$6F,$0F,$F0,$2E,$1E,$CF
	.dc.b	$8A,$C6,$1F,$EF,$48,$C3,$FE,$C3,$F8,$1F,$E7,$AD,$E9,$F8,$CB,$EA
	.dc.b	$15,$F0,$EB,$79,$7F,$87,$5B,$8F,$F8,$25,$B6,$9F,$F0,$ED,$F2,$8E
	.dc.b	$C4,$8F,$F2,$1B,$DF,$83,$B6,$3E,$2A,$FB,$F5,$CF,$EA,$5E,$E7,$D5
	.dc.b	$6E,$1F,$EE,$83,$D1,$E7,$E6,$64,$7F,$B8,$67,$75,$E0,$B5,$EA,$3F
	.dc.b	$C7,$13,$AC,$FF,$C5,$F7,$F0,$F0,$D4,$63,$0D,$4F,$F9,$BC,$FF,$63
	.dc.b	$9F,$EE,$F2,$DF,$7F,$03,$F1,$91,$8A,$FF,$18,$34,$E5,$D5,$F0,$67
	.dc.b	$6C,$8F,$77,$D4,$FF,$95,$DE,$D1,$F6,$23,$17,$FE,$51,$F9,$CF,$C4
	.dc.b	$37,$8B,$E6,$4F,$E7,$67,$38,$FF,$90,$5B,$87,$FA,$25,$B7,$F6,$AB
	.dc.b	$85,$CA,$FA,$19,$3C,$4C,$5F,$CD,$7F,$55,$F5,$0F,$A9,$FD,$67,$D4
	.dc.b	$7F,$99,$30,$FC,$36,$9F,$E5,$1A,$D4,$60,$FF,$4D,$4E,$F0,$FF,$36
	.dc.b	$36,$BF,$CB,$0D,$9F,$F5,$05,$72,$AD,$F8,$9E,$0F,$7E,$57,$59,$FC
	.dc.b	$B5,$F1,$6C,$F9,$BD,$19,$5B,$97,$D4,$F5,$CB,$C3,$5F,$C1,$FA,$5E
	.dc.b	$65,$38,$DF,$FA,$0E,$33,$87,$3F,$A9,$7A,$5F,$C3,$DB,$1D,$E7,$FD
	.dc.b	$46,$EE,$7F,$8A,$7D,$37,$FA,$A2,$3C,$8D,$BF,$F5,$6F,$D4,$3E,$A5
	.dc.b	$F5,$3C,$3D,$C9,$87,$07,$F9,$25,$6E,$73,$7F,$E9,$9C,$3A,$6D,$FE
	.dc.b	$90,$F1,$1F,$53,$FE,$79,$CB,$91,$F2,$E3,$3E,$BF,$F2,$FE,$CF,$E0
	.dc.b	$5F,$53,$72,$FC,$5C,$62,$FE,$93,$FC,$1F,$B3,$55,$B6,$CE,$AB,$88
	.dc.b	$FA,$9E,$87,$5F,$81,$B3,$B8,$B3,$FE,$D0,$C3,$F8,$A5,$63,$4F,$F5
	.dc.b	$07,$D3,$BD,$6F,$E8,$0D,$C7,$F9,$B7,$D4,$FF,$B7,$DF,$C1,$46,$71
	.dc.b	$FF,$5F,$36,$F9,$DF,$F4,$DB,$E1,$5A,$8E,$A6,$7C,$8B,$87,$67,$E4
	.dc.b	$7A,$9F,$EC,$48,$F8,$AF,$E1,$2B,$D4,$E5,$BC,$FC,$27,$8B,$5E,$2F
	.dc.b	$4F,$A9,$91,$D3,$E4,$6D,$74,$9B,$9F,$FB,$3B,$6F,$F1,$85,$62,$FF
	.dc.b	$9E,$5B,$57,$CE,$7C,$83,$FB,$77,$23,$E0,$5F,$31,$9C,$3E,$4D,$5C
	.dc.b	$19,$FE,$36,$F7,$5A,$24,$FF,$6B,$FA,$9F,$97,$5F,$62,$FF,$BE,$28
	.dc.b	$CD,$7F,$51,$D3,$FC,$91,$F2,$1F,$30,$F9,$BF,$53,$67,$E6,$C5,$C7
	.dc.b	$E6,$CD,$C3,$E1,$70,$BE,$67,$EA,$FC,$3F,$D6,$6E,$3F,$59,$B8,$7D
	.dc.b	$66,$DF,$FB,$B7,$EA,$7F,$BC,$36,$FE,$3E,$DF,$C4,$6D,$5A,$C3,$E7
	.dc.b	$B6,$3F,$B6,$37,$83,$E7,$79,$5F,$E1,$78,$D4,$F2,$C3,$FC,$3D,$F5
	.dc.b	$2F,$98,$D4,$61,$FC,$7F,$D4,$FE,$43,$EA,$6E,$7F,$90,$6E,$1F,$90
	.dc.b	$6E,$BF,$11,$FC,$BF,$C9,$7E,$63,$AE,$DF,$E6,$1B,$7F,$98,$6D,$FC
	.dc.b	$C7,$FC,$5D,$F5,$1F,$04,$9E,$1E,$37,$98,$7F,$9B,$B4,$FC,$8D,$3E
	.dc.b	$41,$39,$CB,$FC,$21,$3F,$D1,$FE,$13,$E0,$46,$2B,$6F,$FA,$0A,$F9
	.dc.b	$C6,$27,$72,$AE,$D7,$D6,$31,$86,$F0,$E3,$39,$AD,$D7,$53,$C1,$19
	.dc.b	$AE,$11,$83,$47,$7F,$80,$CB,$67,$A7,$D4,$E2,$FF,$C8,$F1,$DE,$B7
	.dc.b	$F5,$0F,$03,$FC,$6F,$87,$4B,$DB,$DA,$75,$E7,$FA,$DB,$19,$AE,$3A
	.dc.b	$7C,$2B,$E2,$FC,$4A,$DB,$67,$7A,$7F,$04,$9C,$BF,$5B,$1F,$2C,$EB
	.dc.b	$0F,$F2,$25,$69,$B1,$97,$E0,$1E,$3C,$7C,$D1,$99,$E1,$3A,$C3,$C5
	.dc.b	$5A,$24,$7C,$1B,$F0,$C6,$27,$73,$9C,$B1,$8F,$A7,$1E,$C7,$8F,$2F
	.dc.b	$83,$19,$C3,$77,$E8,$7E,$95,$E2,$19,$79,$8F,$3D,$F2,$8E,$07,$B4
	.dc.b	$98,$FC,$FC,$BC,$F1,$8C,$7C,$0F,$9C,$C1,$CA,$7D,$67,$1C,$3C,$58
	.dc.b	$78,$72,$F5,$61,$D2,$B7,$87,$47,$86,$1D,$0F,$4D,$7C,$2C,$BD,$B1
	.dc.b	$9C,$3A,$46,$2B,$9C,$F8,$EF,$A1,$F8,$1C,$7C,$78,$73,$9F,$55,$70
	.dc.b	$7B,$DF,$C9,$CB,$E3,$46,$0F,$05,$7A,$AF,$9D,$6F,$2F,$66,$3E,$0B
	.dc.b	$F1,$1F,$C0,$A3,$2E,$0E,$98,$FC,$03,$C5,$97,$B2,$3C,$B5,$CD,$DB
	.dc.b	$72,$F2,$CE,$EB,$D5,$18,$CB,$DD,$18,$C3,$C3,$87,$23,$C5,$5B,$C7
	.dc.b	$F6,$5F,$52,$F8,$2F,$ED,$B5,$FA,$B8,$C3,$83,$E2,$1A,$C9,$1A,$8C
	.dc.b	$3C,$31,$FD,$C3,$73,$E9,$7F,$C6,$FE,$A1,$F5,$3F,$DF,$2F,$73,$E4
	.dc.b	$26,$1F,$C5,$AB,$6E,$27,$6F,$F0,$DC,$3B,$3F,$EC,$19,$DB,$9A,$E1
	.dc.b	$87,$8A,$3A,$61,$E1,$8C,$61,$D0,$8F,$F8,$42,$F9,$C6,$2B,$F8,$5E
	.dc.b	$1E,$07,$F6,$2D,$9D,$D7,$42,$4F,$0C,$3E,$16,$1E,$18,$C1,$E7,$26
	.dc.b	$1F,$D1,$2F,$9C,$6A,$3C,$4F,$82,$B9,$B0,$E3,$7E,$05,$5C,$9B,$7E
	.dc.b	$17,$91,$C0,$F9,$4E,$61,$F0,$B7,$E7,$A5,$7D,$B8,$D1,$1C,$1F,$D6
	.dc.b	$7E,$A7,$FD,$BE,$DF,$86,$D3,$F7,$67,$C5,$3A,$8F,$EA,$AD,$CF,$E1
	.dc.b	$4F,$07,$FB,$3C,$67,$2F,$F2,$2F,$C0,$7F,$91,$2F,$FD,$50,$BE,$A7
	.dc.b	$FC,$93,$D5,$FD,$CA,$7F,$D9,$11,$8E,$5E,$3E,$4D,$91,$FE,$CD,$70
	.dc.b	$F0,$3A,$CE,$2F,$C1,$1F,$E4,$CA,$F1,$FB,$79,$FA,$19,$3E,$48,$E0
	.dc.b	$61,$E0,$78,$29,$3A,$8F,$F6,$17,$D4,$BE,$A7,$C1,$E9,$49,$CC,$79
	.dc.b	$A7,$51,$F0,$A7,$4F,$61,$1A,$70,$DD,$3E,$33,$72,$FF,$38,$B6,$7C
	.dc.b	$DC,$3F,$13,$EA,$5D,$77,$E0,$A5,$7F,$B6,$67,$2D,$9C,$BE,$38,$4F
	.dc.b	$39,$E9,$F2,$72,$B9,$3D,$FE,$27,$11,$C6,$2E,$7F,$77,$05,$7D,$8F
	.dc.b	$A9,$FF,$35,$5F,$8F,$D4,$F4,$19,$F3,$E4,$64,$87,$C7,$9E,$2D,$8D
	.dc.b	$B4,$F7,$76,$FE,$1B,$EA,$5F,$53,$49,$3F,$EF,$B8,$7C,$D6,$23,$EA
	.dc.b	$47,$07,$53,$CE,$7A,$7C,$F7,$5F,$A5,$95,$CE,$32,$F5,$8C,$BD,$23
	.dc.b	$F3,$3E,$3F,$F9,$3B,$7F,$D1,$7D,$43,$EA,$7F,$3A,$DB,$FD,$0C,$FE
	.dc.b	$9E,$B8,$73,$FF,$AC,$9B,$3F,$A6,$AE,$0F,$E3,$5B,$5C,$7B,$FC,$24
	.dc.b	$9E,$B7,$CE,$3C,$3D,$FE,$2E,$E9,$5A,$3B,$3E,$6E,$DC,$C7,$78,$B8
	.dc.b	$7B,$27,$D7,$D5,$EC,$F9,$39,$7C,$C6,$7F,$C3,$EF,$E1,$D6,$A7,$D5
	.dc.b	$5C,$9E,$1A,$7E,$59,$E5,$ED,$3C,$5F,$E0,$AD,$8D,$F6,$FD,$03,$33
	.dc.b	$23,$24,$3F,$69,$5D,$5B,$3D,$3B,$7F,$88,$FD,$33,$93,$F8,$9F,$D4
	.dc.b	$BE,$A6,$9A,$8D,$FA,$9C,$3E,$A6,$65,$F0,$70,$4F,$F1,$CF,$A8,$FD
	.dc.b	$3C,$E6,$38,$BF,$1E,$32,$FF,$09,$5F,$1D,$FC,$93,$EA,$5F,$11,$4E
	.dc.b	$91,$FC,$A3,$EA,$7E,$77,$5F,$FB,$EA,$F8,$E1,$D6,$B5,$CB,$23,$EB
	.dc.b	$5F,$07,$F9,$9B,$57,$D4,$BE,$53,$F9,$AF,$B3,$FA,$C9,$FB,$D3,$FE
	.dc.b	$B7,$6F,$B5,$AF,$0A,$46,$08,$7F,$25,$AE,$4D,$8D,$B4,$EB,$E2,$FF
	.dc.b	$44,$F8,$3F,$14,$7E,$37,$E9,$1F,$03,$FD,$21,$B1,$F9,$19,$DB,$FC
	.dc.b	$CD,$BD,$FF,$D5,$67,$73,$FC,$DD,$B1,$FD,$66,$77,$C3,$FD,$5A,$91
	.dc.b	$87,$FA,$CB,$5C,$EB,$E3,$F8,$BF,$EB,$7E,$3B,$DF,$FE,$BB,$F4,$FF
	.dc.b	$6A,$FF,$4C,$D7,$FA,$44,$ED,$FD,$53,$5F,$E9,$91,$BC,$BF,$B7,$37
	.dc.b	$2F,$35,$6E,$7F,$B9,$36,$33,$7E,$43,$C7,$F1,$FF,$DC,$3E,$A5,$F4
	.dc.b	$5E,$6F,$A9,$99,$FE,$EE,$72,$D9,$CB,$E4,$9E,$51,$23,$44,$3D,$3E
	.dc.b	$57,$F7,$EF,$A9,$FC,$0E,$9E,$AF,$6B,$E2,$99,$E5,$97,$F1,$A3,$2F
	.dc.b	$F0,$DA,$E3,$AF,$D5,$6C,$7F,$83,$1A,$9B,$BB,$E8,$7C,$5F,$0F,$F5
	.dc.b	$3D,$B9,$7D,$63,$2F,$D6,$8C,$BF,$DF,$6D,$3F,$E1,$8F,$A9,$73,$DE
	.dc.b	$A9,$7F,$C8,$E7,$2D,$9E,$04,$7F,$94,$7D,$4B,$DC,$70,$7F,$89,$BE
	.dc.b	$A5,$F5,$3F,$93,$AD,$4E,$63,$F2,$D8,$F6,$C1,$7C,$1F,$F1,$7C,$66
	.dc.b	$31,$0B,$E6,$F0,$AE,$D8,$78,$A3,$07,$2C,$7F,$49,$7C,$B0,$E9,$7C
	.dc.b	$31,$F2,$5F,$F8,$FB,$1E,$B7,$BC,$38,$DF,$A1,$B5,$BC,$7C,$15,$E6
	.dc.b	$BE,$37,$BA,$EA,$31,$F0,$BC,$71,$CD,$70,$8F,$41,$C2,$76,$FC,$1C
	.dc.b	$3B,$DF,$A6,$3D,$E7,$5C,$7C,$AD,$3A,$CF,$07,$C0,$70,$AE,$A7,$6B
	.dc.b	$F1,$47,$7C,$3F,$D8,$0D,$3F,$D8,$51,$F4,$9F,$F6,$3B,$5F,$C0,$7B
	.dc.b	$AF,$91,$E3,$3F,$E3,$26,$C6,$67,$9B,$5E,$FF,$09,$E0,$7F,$CE,$2D
	.dc.b	$4D,$23,$F0,$BE,$45,$F4,$EE,$5F,$A3,$3C,$8D,$46,$5F,$92,$67,$7F
	.dc.b	$C7,$D7,$2A,$E3,$3E,$03,$8B,$8B,$FB,$B3,$B9,$E6,$6E,$B9,$DF,$77
	.dc.b	$F3,$71,$DF,$7F,$E7,$15,$C6,$73,$3C,$67,$35,$C5,$FF,$79,$5F,$82
	.dc.b	$31,$3F,$EF,$88,$C6,$1E,$58,$F3,$9E,$48,$CB,$4C,$DF,$82,$32,$FA
	.dc.b	$27,$FC,$4D,$87,$B7,$4F,$F5,$45,$FA,$4F,$F5,$2C,$66,$7C,$E7,$0C
	.dc.b	$3E,$75,$70,$AF,$61,$C2,$33,$3F,$0B,$A6,$CF,$46,$FC,$63,$EA,$7F
	.dc.b	$D0,$B5,$E4,$9D,$D3,$1F,$E8,$B3,$FE,$47,$C5,$F5,$2E,$FD,$D6,$AF
	.dc.b	$FD,$5F,$5A,$9C,$92,$FF,$CA,$19,$7F,$A7,$3C,$5C,$5F,$FA,$A7,$0E
	.dc.b	$95,$F1,$AF,$CF,$EA,$FF,$80,$23,$04,$AF,$F5,$EE,$1E,$BF,$8A,$E7
	.dc.b	$97,$E4,$E3,$15,$FE,$55,$F2,$3F,$9C,$B7,$AB,$E2,$AF,$A7,$57,$C2
	.dc.b	$AD,$E3,$FE,$BD,$6E,$1F,$ED,$F6,$DF,$FB,$70,$E5,$5C,$70,$91,$FE
	.dc.b	$6F,$3B,$70,$E3,$0F,$F6,$AC,$60,$FF,$35,$7D,$4B,$DD,$FA,$6D,$8E
	.dc.b	$59,$BC,$FC,$D4,$9E,$2F,$B2,$75,$18,$25,$F8,$23,$A3,$FE,$D9,$BB
	.dc.b	$A7,$EE,$EB,$E2,$5F,$E1,$1F,$F6,$09,$FE,$7F,$2D,$73,$F0,$75,$D3
	.dc.b	$27,$4F,$8E,$31,$3F,$EE,$79,$E3,$7D,$FE,$A7,$FC,$A1,$1C,$0F,$90
	.dc.b	$E2,$7F,$DD,$DF,$33,$F8,$66,$F1,$FB,$E7,$F3,$33,$9B,$FE,$29,$48
	.dc.b	$C5,$FF,$8F,$5B,$1F,$C7,$74,$5F,$2C,$25,$FD,$0F,$16,$DF,$C9,$7F
	.dc.b	$4F,$F5,$0F,$A9,$FD,$4F,$D4,$BC,$5C,$C6,$19,$87,$FA,$9E,$74,$FF
	.dc.b	$AA,$6B,$93,$99,$E7,$EF,$C8,$C1,$C5,$EF,$E8,$C8,$FF,$59,$7D,$4F
	.dc.b	$2C,$FF,$C1,$71,$FE,$5B,$C7,$E5,$70,$7C,$4E,$87,$9B,$FC,$01,$B5
	.dc.b	$E5,$C7,$13,$FC,$13,$EA,$5E,$9D,$38,$7B,$56,$FF,$D8,$5F,$4D,$FE
	.dc.b	$9A,$6F,$B5,$FC,$4B,$EA,$1F,$52,$FA,$9F,$E1,$F7,$BE,$0F,$38,$0E
	.dc.b	$77,$FB,$6D,$C7,$58,$74,$BE,$78,$C9,$FF,$6F,$78,$AF,$A9,$FF,$3A
	.dc.b	$47,$90,$CB,$8C,$BC,$3D,$9E,$1F,$A9,$FE,$52,$F8,$FA,$EC,$A6,$C6
	.dc.b	$1B,$7F,$EE,$DA,$E4,$DA,$FC,$7C,$7B,$7E,$A7,$99,$D3,$82,$D6,$E2
	.dc.b	$D7,$90,$FF,$61,$C6,$4F,$F7,$87,$D3,$BD,$7F,$E7,$2D,$8F,$E5,$FF
	.dc.b	$53,$FE,$D4,$77,$3A,$23,$AF,$85,$B7,$F4,$85,$F3,$7F,$97,$DF,$F2
	.dc.b	$07,$EB,$46,$78,$FF,$98,$4E,$73,$FF,$0C,$61,$F2,$BD,$4F,$C4,$BF
	.dc.b	$15,$FF,$06,$37,$9B,$4E,$DA,$E4,$E1,$F5,$3C,$CE,$9F,$19,$63,$24
	.dc.b	$CF,$ED,$DA,$75,$1E,$DC,$DF,$23,$F2,$0F,$EC,$FC,$9F,$EC,$0F,$98
	.dc.b	$C6,$3F,$3A,$6D,$FE,$5A,$B5,$39,$24,$7F,$62,$FA,$9F,$A0,$F7,$AF
	.dc.b	$EE,$7B,$A7,$2F,$F8,$A7,$6F,$F1,$F7,$C8,$FC,$73,$E4,$7D,$4D,$C7
	.dc.b	$E4,$45,$C7,$DB,$36,$F9,$6C,$F9,$9F,$A1,$EB,$7D,$16,$E3,$F4,$5B
	.dc.b	$87,$C6,$6D,$FF,$28,$FA,$9F,$EE,$13,$F8,$36,$C7,$F8,$FF,$6F,$F7
	.dc.b	$2F,$91,$FE,$08,$9C,$E1,$FE,$E9,$6D,$FE,$ED,$B5,$C2,$7F,$7D,$7D
	.dc.b	$F1,$E4,$7F,$85,$BE,$A5,$F3,$18,$AF,$F0,$4C,$61,$FC,$57,$D4,$FE
	.dc.b	$37,$EA,$7F,$1C,$D8,$C9,$F8,$F8,$CB,$4F,$C9,$75,$7E,$57,$D8,$FC
	.dc.b	$B3,$63,$27,$E5,$E3,$2D,$3D,$27,$F1,$7D,$5F,$51,$FC,$F1,$FD,$53
	.dc.b	$74,$FE,$A6,$FF,$0F,$6E,$FF,$E3,$08,$E1,$F0,$3F,$02,$BF,$91,$43
	.dc.b	$E3,$FF,$CD,$D3,$99,$EF,$58,$DB,$C5,$7B,$C7,$FC,$FF,$18,$C3,$57
	.dc.b	$D1,$AF,$29,$DB,$96,$BE,$C8,$DD,$75,$8F,$F4,$4D,$78,$2F,$35,$AF
	.dc.b	$7A,$61,$CB,$07,$D4,$E3,$1E,$27,$BA,$BC,$7F,$53,$FB,$7F,$43,$84
	.dc.b	$63,$BF,$FC,$6B,$5C,$05,$6D,$C3,$75,$F8,$D8,$AF,$96,$1E,$CA,$D6
	.dc.b	$BF,$65,$B8,$78,$F3,$D6,$5E,$0C,$BE,$85,$6A,$BC,$33,$9B,$FA,$38
	.dc.b	$F6,$9F,$1C,$E5,$FE,$B9,$5C,$72,$F2,$63,$DB,$29,$5C,$1F,$83,$18
	.dc.b	$CB,$C3,$87,$3A,$F0,$D6,$9E,$04,$9D,$C6,$5B,$19,$CB,$E0,$57,$3C
	.dc.b	$3C,$73,$93,$CF,$1D,$F1,$F4,$65,$ED,$C3,$99,$30,$F8,$39,$79,$DE
	.dc.b	$B9,$5C,$7E,$61,$D4,$F4,$97,$2F,$0D,$F3,$8F,$0C,$F1,$7F,$5B,$87
	.dc.b	$86,$FA,$C7,$18,$C3,$C0,$C6,$1C,$71,$FB,$07,$A9,$C6,$5E,$D3,$E1
	.dc.b	$57,$2A,$F2,$E3,$E3,$7C,$35,$C3,$1F,$26,$5E,$8C,$BC,$77,$BC,$BB
	.dc.b	$61,$EF,$7C,$39,$76,$C7,$94,$F9,$8D,$47,$8F,$1F,$B1,$87,$42,$4F
	.dc.b	$B7,$0F,$CE,$E5,$E7,$C7,$AD,$F0,$AD,$5F,$07,$D6,$FD,$67,$BC,$7A
	.dc.b	$AB,$97,$D4,$BE,$03,$F8,$E5,$7E,$96,$30,$E0,$FB,$26,$B0,$46,$A3
	.dc.b	$0F,$0C,$7F,$52,$DC,$7A,$5F,$F0,$EF,$A8,$7D,$4F,$F2,$BB,$DC,$FE
	.dc.b	$DE,$F8,$61,$FA,$A6,$AB,$E2,$EE,$63,$FC,$12,$77,$3A,$8C,$4E,$AB
	.dc.b	$BC,$7F,$0D,$9D,$46,$09,$87,$A2,$75,$87,$9E,$31,$3A,$9F,$35,$FB
	.dc.b	$6B,$51,$83,$C2,$49,$FE,$79,$7E,$0B,$E6,$49,$E1,$3A,$AE,$71,$FC
	.dc.b	$8B,$0F,$87,$8F,$D2,$C7,$D9,$87,$F5,$5C,$7A,$5F,$28,$E8,$F1,$9F
	.dc.b	$33,$0F,$24,$6A,$78,$C7,$CF,$C3,$C3,$7D,$E3,$67,$F9,$E6,$76,$3E
	.dc.b	$0F,$CE,$BE,$7A,$98,$89,$3F,$83,$6B,$FC,$FF,$EA,$7F,$DC,$0F,$F5
	.dc.b	$46,$BF,$5E,$37,$A7,$F6,$59,$CB,$7E,$13,$39,$FC,$3A,$F9,$35,$A8
	.dc.b	$FE,$F5,$EF,$7F,$91,$30,$FE,$FA,$7F,$A4,$D7,$D4,$FF,$7F,$E8,$FF
	.dc.b	$66,$C6,$27,$61,$E4,$FF,$9A,$DB,$1C,$27,$97,$27,$81,$B5,$DF,$C7
	.dc.b	$AF,$3F,$66,$E0,$FA,$2A,$62,$24,$F8,$A7,$4D,$9C,$BF,$E7,$DF,$A8
	.dc.b	$7D,$4F,$4F,$46,$4F,$AA,$72,$D9,$CB,$EB,$9F,$2C,$EA,$3A,$4E,$08
	.dc.b	$F8,$F8,$3D,$7F,$53,$FD,$CE,$3E,$3A,$F8,$1C,$9A,$EA,$93,$A8,$CB
	.dc.b	$C2,$39,$FC,$94,$8E,$6F,$6F,$88,$C4,$7F,$A6,$ED,$38,$E2,$71,$7F
	.dc.b	$8E,$65,$7F,$43,$EA,$7F,$CD,$58,$79,$BD,$2F,$49,$9F,$36,$46,$48
	.dc.b	$7F,$AA,$67,$8B,$63,$6D,$3E,$0F,$77,$DE,$FA,$97,$D4,$D2,$4F,$E3
	.dc.b	$21,$F3,$38,$8F,$F5,$B4,$70,$75,$3C,$E7,$A7,$CE,$F5,$FA,$59,$5C
	.dc.b	$E3,$2F,$58,$CB,$D2,$3F,$23,$E3,$7F,$93,$F8,$3E,$3C,$FF,$BD,$7E
	.dc.b	$A1,$F5,$3F,$96,$6C,$7E,$8A,$B8,$71,$7F,$99,$5B,$1C,$6B,$81,$DB
	.dc.b	$BE,$47,$16,$BF,$ED,$3B,$FD,$4D,$FF,$9B,$D7,$6F,$84,$93,$FB,$28
	.dc.b	$FD,$0B,$C6,$32,$D3,$5D,$9E,$9E,$9F,$57,$C8,$C9,$E5,$19,$7A,$46
	.dc.b	$5F,$24,$7C,$6E,$AF,$DB,$FB,$1F,$C7,$67,$F8,$1B,$5F,$F7,$66,$BF
	.dc.b	$9F,$7E,$34,$F0,$F1,$64,$60,$87,$08,$EC,$F4,$EC,$FE,$0B,$F4,$CE
	.dc.b	$3F,$E0,$FF,$52,$FA,$9A,$6A,$37,$E9,$F0,$FA,$99,$8F,$F1,$8C,$13
	.dc.b	$FC,$43,$EA,$3F,$47,$39,$8E,$2F,$F1,$E8,$CA,$E7,$FC,$2A,$87,$C4
	.dc.b	$FF,$1A,$FA,$97,$C3,$D3,$B4,$7F,$1E,$FA,$9F,$E7,$78,$72,$BE,$38
	.dc.b	$77,$AD,$72,$48,$FE,$5F,$7C,$3A,$7F,$D6,$4D,$80,$FA,$97,$C9,$7F
	.dc.b	$2C,$F6,$BF,$57,$3F,$CE,$27,$FD,$5C,$DF,$6B,$5E,$0C,$8C,$10,$F8
	.dc.b	$18,$FF,$B0,$6F,$9C,$60,$F0,$B8,$D7,$E2,$B6,$78,$57,$F3,$EC,$3F
	.dc.b	$C5,$15,$F2,$BE,$A7,$CF,$87,$F4,$CF,$A4,$7C,$0F,$F4,$46,$C7,$F5
	.dc.b	$09,$DB,$E1,$0E,$2B,$EA,$CF,$09,$D3,$63,$C1,$A7,$F5,$0A,$E1,$97
	.dc.b	$FB,$75,$B7,$F3,$1A,$AD,$F1,$5F,$D6,$3C,$57,$D4,$FF,$5A,$FA,$7F
	.dc.b	$C0,$DF,$F6,$66,$BE,$79,$DB,$FD,$A9,$AF,$AE,$37,$9F,$E3,$9B,$97
	.dc.b	$FB,$C6,$B7,$3F,$92,$6C,$66,$FC,$87,$8F,$E4,$3F,$B6,$FD,$4B,$E8
	.dc.b	$BC,$DF,$53,$33,$FE,$F7,$39,$6C,$E5,$F2,$4F,$28,$91,$A2,$1E,$9F
	.dc.b	$23,$FB,$DF,$D4,$FE,$06,$3F,$2B,$AF,$F8,$1D,$F8,$9E,$CE,$B9,$64
	.dc.b	$60,$87,$DD,$AE,$49,$A4,$8F,$A1,$F4,$3E,$0F,$87,$FA,$5E,$DC,$BF
	.dc.b	$F0,$BD,$A7,$51,$F5,$67,$4F,$78,$FF,$09,$FD,$4F,$DF,$8F,$F0,$FA
	.dc.b	$EA,$C9,$AE,$69,$3A,$6B,$F9,$5F,$A9,$FE,$E7,$CD,$FB,$6F,$A8,$7D
	.dc.b	$4F,$EE,$29,$5A,$53,$A7,$F7,$B8,$F8,$DC,$D7,$18,$C4,$EE,$79,$FB
	.dc.b	$BF,$C5,$DD,$9F,$C1,$AF,$B4,$EE,$B8,$DF,$3C,$7C,$93,$CE,$B7,$3E
	.dc.b	$76,$DF,$92,$B7,$3C,$63,$DC,$EF,$0F,$15,$F3,$6C,$6E,$7C,$A3,$0F
	.dc.b	$1B,$E7,$9D,$B7,$1E,$AF,$84,$A7,$1C,$3C,$6E,$CE,$B8,$FA,$1D,$1C
	.dc.b	$2F,$BB,$F2,$2B,$9C,$EE,$72,$69,$E4,$7D,$79,$FF,$64,$B5,$F8,$13
	.dc.b	$FE,$C9,$7A,$1F,$EC,$B8,$FF,$18,$FC,$8E,$5B,$19,$F7,$53,$96,$C7
	.dc.b	$31,$F0,$B5,$A6,$5C,$11,$F9,$BF,$20,$FA,$7F,$F0,$EE,$DF,$87,$9F
	.dc.b	$E9,$35,$F2,$AB,$5D,$BF,$32,$72,$72,$BE,$3B,$FF,$B9,$6F,$84,$62
	.dc.b	$7F,$DD,$71,$8A,$E5,$1E,$6C,$3F,$2D,$87,$68,$FC,$09,$DA,$33,$7C
	.dc.b	$E3,$FD,$CC,$78,$6B,$9D,$F5,$8E,$27,$5C,$3F,$59,$87,$5C,$3D,$D3
	.dc.b	$93,$FB,$3C,$66,$7D,$51,$D4,$E7,$5D,$6B,$8B,$67,$AE,$1F,$E4,$6D
	.dc.b	$3F,$C0,$78,$7A,$DF,$F0,$2E,$1E,$BA,$DE,$3F,$EA,$8A,$D3,$89,$FE
	.dc.b	$B4,$6C,$F2,$DF,$58,$F5,$F4,$D9,$F4,$B7,$E3,$1F,$53,$FE,$8D,$AE
	.dc.b	$D3,$BA,$61,$FD,$06,$7F,$D7,$F9,$3E,$A5,$DF,$BA,$D5,$FF,$7D,$AD
	.dc.b	$4E,$49,$7F,$EC,$8C,$9E,$A7,$F7,$88,$F7,$4E,$AF,$FB,$B6,$1D,$A7
	.dc.b	$85,$6E,$73,$5B,$C3,$FC,$E3,$87,$B3,$E2,$B1,$87,$F8,$7A,$30,$FF
	.dc.b	$B4,$BE,$07,$F9,$E3,$7A,$7E,$32,$FA,$85,$7C,$3A,$DE,$5F,$E0,$96
	.dc.b	$E3,$FE,$3C,$6D,$FF,$8B,$0E,$55,$C7,$09,$1F,$EE,$83,$B7,$06,$30
	.dc.b	$FF,$12,$46,$0F,$F7,$3F,$D4,$BD,$DF,$A8,$D8,$E5,$9B,$CF,$CD,$49
	.dc.b	$E2,$FB,$27,$51,$82,$5F,$82,$3A,$3F,$E2,$5B,$BA,$7F,$A7,$2B,$E2
	.dc.b	$5F,$E1,$5F,$F6,$19,$FE,$FC,$2D,$73,$F0,$35,$D1,$27,$4F,$8E,$31
	.dc.b	$3F,$E3,$39,$E3,$7D,$FE,$A7,$FC,$A7,$1C,$0F,$90,$E2,$7F,$C7,$5F
	.dc.b	$33,$F8,$56,$F1,$7B,$E7,$F3,$53,$9B,$FF,$38,$B6,$FF,$B5,$36,$FF
	.dc.b	$DB,$FA,$2F,$8E,$12,$FE,$7F,$89,$6F,$E4,$3F,$A7,$FA,$87,$D4,$FE
	.dc.b	$A7,$EA,$3F,$C0,$17,$FE,$6E,$D3,$FD,$03,$5A,$8C,$1F,$DE,$E7,$78
	.dc.b	$7F,$95,$5B,$5F,$E8,$86,$CF,$F8,$06,$B9,$56,$FC,$4F,$07,$BF,$2B
	.dc.b	$AC,$FF,$9F,$AF,$8B,$67,$CD,$E8,$CA,$DC,$BE,$A7,$AE,$5E,$1A,$FE
	.dc.b	$09,$D1,$FE,$CC,$BD,$B6,$FD,$13,$AF,$33,$F8,$57,$D4,$FF,$98,$2F
	.dc.b	$AB,$6B,$FC,$E5,$7F,$EA,$09,$CF,$A5,$FC,$47,$EA,$3F,$D4,$B1,$E4
	.dc.b	$6D,$7F,$A2,$7E,$A1,$F5,$2F,$A9,$E1,$EE,$4C,$38,$3F,$C8,$6B,$73
	.dc.b	$9B,$FF,$41,$61,$D3,$6F,$F7,$CF,$88,$FA,$9F,$F3,$AE,$5C,$8F,$97
	.dc.b	$19,$F5,$FF,$96,$76,$7F,$AA,$BE,$A6,$E5,$F8,$B8,$C5,$FD,$27,$FD
	.dc.b	$5F,$D7,$AA,$DB,$67,$55,$C4,$7D,$4F,$43,$AF,$C0,$D9,$DC,$59,$FF
	.dc.b	$56,$61,$FE,$BC,$AC,$69,$FE,$D0,$FA,$77,$AD,$FC,$F5,$B8,$FF,$33
	.dc.b	$FA,$9F,$F6,$EB,$F8,$28,$CE,$3F,$EA,$76,$DF,$3B,$FE,$95,$7C,$2B
	.dc.b	$51,$D4,$CF,$93,$70,$EC,$FC,$8F,$4D,$17,$9F,$0D,$3E,$44,$69,$B1
	.dc.b	$A6,$9A,$FA,$9E,$87,$5F,$8D,$B1,$92,$63,$F1,$AD,$3A,$8F,$77,$CD
	.dc.b	$FC,$53,$FB,$47,$27,$FB,$1B,$E6,$31,$87,$C5,$9B,$5F,$ED,$BA,$D4
	.dc.b	$E4,$91,$FD,$8F,$EA,$7E,$36,$FF,$4E,$FF,$BA,$28,$CD,$7F,$BD,$38
	.dc.b	$BF,$C7,$7F,$23,$F1,$CF,$91,$F5,$36,$7E,$2C,$5C,$7D,$93,$70,$F9
	.dc.b	$B0,$BF,$F2,$2F,$CC,$FD,$1F,$87,$FA,$4D,$C7,$E7,$B6,$BE,$33,$6F
	.dc.b	$9C,$3E,$A7,$FB,$94,$61,$B5,$FE,$26,$F8,$8F,$E1,$F5,$BC,$3F,$C1
	.dc.b	$95,$A8,$19,$AF,$D6,$D7,$09,$D4,$E4,$5F,$F7,$EF,$A9,$7C,$C7,$4C
	.dc.b	$7F,$C3,$91,$8C,$7F,$13,$F5,$3F,$8F,$FA,$9F,$C8,$36,$32,$7E,$3A
	.dc.b	$32,$D3,$F2,$3F,$25,$F9,$7E,$9F,$CC,$36,$32,$7E,$5A,$32,$D3,$F9
	.dc.b	$5F,$D4,$7C,$22,$78,$5F,$E2,$D7,$1E,$87,$D6,$F8,$3F,$5E,$FF,$12
	.dc.b	$5F,$21,$9D,$FF,$D8,$11,$FD,$1A,$B7,$48,$DE,$3F,$E7,$4A,$DD,$F1
	.dc.b	$AE,$8D,$78,$61,$6F,$8C,$7F,$A1,$AB,$8D,$F3,$8E,$2D,$78,$B6,$78
	.dc.b	$E1,$E4,$7F,$C7,$B1,$9F,$80,$ED,$F5,$3D,$D9,$1C,$FE,$9F,$F6,$7E
	.dc.b	$7E,$E3,$1D,$FF,$E3,$2C,$BC,$87,$E2,$2B,$6E,$03,$8C,$BC,$58,$7B
	.dc.b	$E3,$19,$FF,$BA,$5B,$5C,$EF,$B5,$F0,$C3,$C3,$8C,$AE,$D7,$C1,$FA
	.dc.b	$18,$F5,$C3,$8E,$3D,$5B,$3D,$6F,$D8,$F2,$8C,$BE,$86,$DF,$5C,$FE
	.dc.b	$4C,$EE,$BC,$59,$7E,$02,$B5,$8F,$86,$73,$8F,$85,$EF,$5B,$C7,$A9
	.dc.b	$23,$33,$BC,$3B,$12,$BE,$13,$63,$51,$CE,$78,$CF,$B4,$D1,$63,$89
	.dc.b	$B7,$27,$BE,$38,$B4,$CE,$1E,$CC,$BD,$71,$CE,$FC,$27,$E8,$B1,$F8
	.dc.b	$11,$C3,$1F,$97,$97,$C8,$CB,$C0,$4B,$E3,$5F,$AD,$BE,$33,$98,$EC
	.dc.b	$67,$0F,$0B,$E5,$8D,$3F,$80,$8C,$E3,$F8,$33,$D1,$1E,$3C,$B8,$E3
	.dc.b	$E3,$C7,$C3,$3C,$DB,$19,$8F,$09,$92,$B9,$70,$4A,$EF,$87,$91,$F1
	.dc.b	$65,$F0,$B2,$F6,$E1,$E0,$9D,$B6,$7F,$05,$7F,$A9,$C3,$B5,$7A,$FE
	.dc.b	$A5,$F0,$5F,$CD,$AA,$1F,$AE,$8C,$B8,$3E,$D1,$A9,$46,$A3,$0F,$0C
	.dc.b	$BF,$34,$DC,$7F,$68,$D9,$FE,$8B,$F5,$0F,$A9,$F1,$5F,$0C,$7F,$84
	.dc.b	$5E,$EF,$C9,$1F,$C5,$6F,$84,$6A,$23,$FB,$A9,$E1,$3A,$8C,$4E,$AB
	.dc.b	$B4,$7F,$0D,$9D,$46,$09,$87,$A2,$BC,$18,$79,$A3,$17,$D0,$95,$FC
	.dc.b	$E2,$F9,$CE,$A3,$9C,$FF,$1D,$C3,$CB,$7C,$E3,$18,$79,$6F,$81,$F5
	.dc.b	$70,$E7,$3A,$3B,$C6,$9F,$97,$5C,$DF,$8E,$68,$E1,$87,$96,$35,$8F
	.dc.b	$DA,$C3,$C1,$5A,$BF,$05,$73,$C3,$C3,$3D,$30,$F5,$A8,$C9,$AC,$3A
	.dc.b	$47,$84,$D5,$73,$F8,$5E,$9F,$3D,$27,$EE,$46,$5A,$43,$FA,$17,$D4
	.dc.b	$FF,$B8,$DF,$EA,$CD,$7E,$D4,$71,$77,$F0,$DF,$0E,$75,$97,$D5,$AD
	.dc.b	$3F,$DB,$A3,$39,$FE,$8F,$E0,$1E,$37,$E9,$7B,$FF,$C4,$B0,$FA,$9C
	.dc.b	$73,$6D,$BA,$7F,$88,$27,$8B,$FE,$B0,$6B,$FE,$1E,$9E,$0E,$67,$8A
	.dc.b	$9D,$D7,$83,$DB,$CC,$3E,$86,$98,$89,$5D,$E3,$2F,$86,$3A,$FD,$4B
	.dc.b	$EA,$79,$FA,$32,$BF,$D8,$B3,$96,$CE,$5F,$41,$ED,$9E,$71,$23,$64
	.dc.b	$36,$7A,$F8,$7D,$BF,$52,$EB,$BF,$05,$2B,$FD,$AB,$39,$6C,$E5,$E7
	.dc.b	$09,$E7,$3D,$3E,$56,$57,$27,$BF,$C5,$62,$38,$C5,$DF,$FC,$BF,$82
	.dc.b	$BE,$8F,$D4,$BC,$DF,$3F,$A7,$31,$FF,$74,$CE,$5B,$39,$7C,$B3,$CA
	.dc.b	$24,$68,$87,$AB,$BB,$EF,$FD,$4B,$EA,$69,$27,$FD,$EF,$0F,$9A,$C4
	.dc.b	$7C,$C8,$E0,$EA,$79,$CF,$4F,$9E,$EB,$F4,$B2,$B9,$C6,$5E,$B1,$97
	.dc.b	$A4,$7E,$4B,$C6,$FF,$28,$EF,$FE,$84,$FA,$87,$D4,$FE,$59,$B7,$FE
	.dc.b	$7E,$9F,$F4,$4D,$70,$ED,$FE,$98,$D9,$FF,$41,$D7,$07,$F0,$AD,$AE
	.dc.b	$3D,$FE,$12,$4F,$5B,$E7,$1E,$1E,$FF,$17,$64,$AD,$1D,$9F,$37,$6E
	.dc.b	$63,$BC,$5C,$3D,$93,$EB,$EA,$F6,$7C,$9C,$BE,$63,$3F,$E1,$D7,$F0
	.dc.b	$EB,$53,$EA,$AE,$4F,$0D,$3F,$2C,$F2,$F6,$5E,$2F,$EF,$9B,$1B,$ED
	.dc.b	$FA,$06,$66,$46,$48,$7E,$B2,$BA,$B6,$7A,$76,$FF,$0A,$FA,$67,$27
	.dc.b	$F0,$CF,$A9,$7D,$4D,$35,$1B,$F5,$38,$7D,$4C,$CB,$E0,$E0,$9F,$E2
	.dc.b	$BF,$51,$FA,$79,$CC,$71,$7E,$3C,$65,$FE,$02,$BE,$3B,$F8,$EF,$D4
	.dc.b	$BE,$22,$9D,$23,$F9,$0F,$D4,$FC,$EE,$BF,$F0,$A5,$F1,$C3,$AD,$6B
	.dc.b	$96,$47,$D6,$BE,$0F,$FB,$41,$AB,$EA,$5F,$29,$FC,$BB,$D9,$FD,$64
	.dc.b	$FD,$E9,$FF,$59,$B7,$DA,$D7,$85,$23,$04,$3F,$8E,$D7,$26,$C6,$DA
	.dc.b	$75,$F1,$7F,$9E,$FC,$17,$8A,$3F,$1B,$F4,$AF,$82,$FE,$82,$D8,$FC
	.dc.b	$8C,$ED,$FE,$56,$DE,$FF,$F7,$BC,$EE,$7F,$97,$B6,$3F,$DF,$73,$BE
	.dc.b	$1F,$F1,$F3,$6B,$FD,$EA,$D4,$DA,$7C,$57,$F5,$4F,$1D,$EF,$FF,$56
	.dc.b	$FA,$7F,$B5,$87,$E9,$9A,$FF,$40,$9D,$BF,$AA,$6B,$FD,$12,$37,$93
	.dc.b	$2F,$F2,$94,$61,$FF,$2A,$57,$09,$D5,$04,$66,$77,$7E,$73,$CD,$F1
	.dc.b	$FF,$DB,$3E,$A5,$F4,$5E,$9F,$A9,$98,$FE,$FA,$72,$D9,$CB,$E7,$9E
	.dc.b	$51,$23,$44,$3D,$9E,$57,$F7,$AF,$A9,$FC,$26,$9E,$89,$6B,$FD,$F5
	.dc.b	$F9,$1E,$D6,$F9,$24,$60,$87,$F6,$AB,$E7,$B7,$F9,$DD,$B1,$F0,$DA
	.dc.b	$9B,$6F,$A1,$F2,$7C,$3F,$D9,$F6,$E5,$FF,$1E,$B4,$EA,$3E,$E4,$E9
	.dc.b	$F1,$47,$F8,$5B,$EA,$5C,$B7,$A6,$57,$F2,$89,$CB,$67,$81,$1F,$E5
	.dc.b	$BF,$52,$F7,$1C,$1F,$77,$EA,$5F,$53,$FE,$95,$AD,$4E,$63,$FD,$31
	.dc.b	$8F,$6C,$17,$C1,$FC,$34,$66,$31,$0B,$E6,$F0,$AE,$D8,$78,$A3,$07
	.dc.b	$2C,$7F,$D4,$D7,$CB,$0E,$97,$C3,$1F,$25,$FE,$57,$1E,$B7,$BC,$38
	.dc.b	$DF,$A1,$B5,$BC,$7C,$15,$E6,$BE,$37,$BA,$EA,$31,$F0,$BC,$71,$CD
	.dc.b	$70,$8F,$41,$C2,$76,$FC,$1C,$3B,$DF,$A6,$3D,$E7,$5C,$7C,$AD,$3A
	.dc.b	$CF,$07,$C0,$70,$AE,$A7,$6B,$F1,$47,$7C,$3F,$D9,$4D,$3F,$D9,$91
	.dc.b	$F4,$9F,$F6,$8B,$5F,$C0,$7B,$AF,$94,$E3,$3F,$E3,$76,$C6,$67,$9B
	.dc.b	$5E,$FF,$09,$E0,$7F,$84,$35,$34,$8F,$C2,$F9,$47,$D3,$FF,$10,$DB
	.dc.b	$F1,$33,$C8,$D4,$67,$6F,$F7,$35,$FD,$6B,$E3,$CB,$F6,$2F,$86,$1F
	.dc.b	$92,$9C,$C7,$D0,$33,$7F,$EF,$38,$CC,$F8,$27,$CE,$78,$2B,$A3,$88
	.dc.b	$E9,$87,$8E,$3F,$DF,$78,$78,$E3,$11,$FE,$EE,$39,$1E,$13,$96,$1E
	.dc.b	$18,$DC,$79,$0F,$E6,$18,$79,$27,$DF,$1C,$AF,$CD,$7D,$F0,$F2,$B6
	.dc.b	$7B,$E9,$FC,$8F,$0F,$82,$FF,$25,$8E,$C2,$33,$5E,$33,$FC,$66,$E2
	.dc.b	$3D,$A7,$48,$F9,$1D,$3F,$D1,$AB,$B3,$63,$FA,$47,$C6,$BE,$A7,$FD
	.dc.b	$15,$1D,$27,$57,$EE,$6D,$BE,$A5,$DF,$AA,$F3,$56,$A7,$37,$E6,$AD
	.dc.b	$65,$8B,$FF,$54,$E5,$FE,$A1,$F1,$33,$52,$F8,$1F,$E5,$A9,$D3,$8A
	.dc.b	$FF,$55,$DF,$3C,$BF,$B1,$D7,$AF,$E2,$B9,$57,$6D,$11,$E7,$F1,$BF
	.dc.b	$9E,$E3,$FE,$5C,$6D,$7F,$AF,$21,$F5,$0C,$38,$75,$49,$E0,$FC,$67
	.dc.b	$71,$9A,$FF,$5F,$DF,$4C,$BF,$BB,$D7,$4E,$3D,$46,$31,$FB,$56,$FA
	.dc.b	$97,$BB,$F3,$5B,$87,$F9,$F0,$F2,$F9,$D9,$99,$1F,$E7,$69,$DD,$7A
	.dc.b	$ED,$7A,$8F,$F6,$9C,$EB,$0F,$16,$19,$6E,$1F,$E8,$03,$FD,$88,$5E
	.dc.b	$7F,$81,$8F,$F6,$B8,$C1,$30,$FF,$6F,$BC,$1F,$17,$81,$F8,$78,$C5
	.dc.b	$FF,$80,$DA,$7F,$86,$C7,$D4,$FF,$95,$9E,$91,$F2,$A3,$15,$FE,$EC
	.dc.b	$F9,$8F,$C1,$57,$F8,$8E,$11,$87,$FC,$59,$87,$49,$FF,$4D,$CE,$67
	.dc.b	$7C,$9F,$9B,$8E,$13,$F8,$2B,$C4,$C8,$FA,$56,$F1,$71,$87,$CF,$7F
	.dc.b	$49,$F5,$0F,$A9,$FD,$37,$D4,$BC,$4C,$C6,$19,$8F,$FA,$BA,$74,$FF
	.dc.b	$AC,$2B,$93,$99,$E7,$EF,$C8,$C1,$C5,$EF,$E9,$C8,$FF,$5C,$FD,$4F
	.dc.b	$2C,$FF,$C1,$B1,$F1,$F1,$F9,$7C,$1F,$E3,$5E,$87,$99,$FB,$D6,$D7
	.dc.b	$97,$1C,$4F,$F0,$1F,$A9,$FF,$32,$CF,$FB,$19,$B8,$F6,$3F,$87,$FB
	.dc.b	$3B,$F4,$FF,$84,$7D,$37,$FA,$8E,$FF,$D9,$AD,$AE,$1F,$50,$FA,$97
	.dc.b	$D4,$F0,$F7,$26,$1C,$1F,$E2,$F5,$B9,$CD,$F8,$46,$1D,$77,$FF,$34
	.dc.b	$F8,$AF,$A9,$FF,$3D,$E5,$CC,$FF,$74,$C6,$7D,$8F,$E4,$7D,$BF,$E6
	.dc.b	$FF,$A9,$B9,$7E,$36,$31,$7F,$EF,$07,$FC,$E9,$DB,$AA,$DB,$67,$55
	.dc.b	$C4,$7D,$4F,$43,$AF,$15,$9D,$C5,$9F,$99,$87,$FA,$06,$B1,$A7,$F9
	.dc.b	$B7,$EA,$5E,$B7,$F3,$56,$E3,$FC,$AF,$EA,$7F,$DB,$4F,$FB,$F2,$33
	.dc.b	$8F,$C7,$6D,$F3,$BF,$E8,$7D,$2C,$38,$DE,$E7,$B1,$A8,$FE,$41,$87
	.dc.b	$EC,$F0,$E6,$FF,$50,$8C,$D7,$ED,$BD,$3F,$F5,$04,$FC,$97,$F1,$38
	.dc.b	$FF,$A7,$B3,$FF,$10,$4F,$D2,$9C,$BF,$EA,$18,$CB,$C7,$9F,$4E,$DE
	.dc.b	$1F,$53,$23,$B7,$C6,$5C,$7E,$5E,$0A,$C6,$3E,$18,$56,$2F,$D9,$F2
	.dc.b	$3F,$34,$FE,$CF,$C9,$FE,$C8,$F9,$7C,$4F,$CD,$9B,$97,$E8,$EB,$53
	.dc.b	$92,$47,$E9,$F1,$7D,$4F,$CE,$7E,$E5,$FF,$72,$8C,$57,$C1,$C9,$F2
	.dc.b	$3F,$14,$F8,$DF,$53,$67,$E6,$45,$C7,$DF,$37,$1F,$77,$CC,$FC,$DF
	.dc.b	$5B,$E7,$37,$0F,$A8,$DB,$F8,$ED,$AF,$F5,$E7,$D4,$FF,$BE,$6B,$FD
	.dc.b	$A9,$18,$BF,$8B,$F0,$FF,$C7,$30,$FD,$A5,$ED,$B3,$FE,$0B,$8C,$57
	.dc.b	$ED,$AB,$83,$CE,$FA,$9F,$CF,$E7,$6F,$F8,$73,$EA,$5F,$31,$CF,$0F
	.dc.b	$CC,$C6,$27,$F1,$DF,$53,$F8,$BF,$A9,$FC,$63,$63,$27,$E4,$E3,$2D
	.dc.b	$3F,$2D,$F2,$3F,$93,$F6,$FF,$28,$D8,$C9,$F9,$B8,$CB,$4F,$4E,$BF
	.dc.b	$EF,$9F,$A8,$F6,$C7,$F5,$96,$C6,$7C,$3F,$F1,$74,$62,$0F,$CB,$F8
	.dc.b	$1F,$7C,$73,$F8,$E5,$FF,$9F,$B6,$E9,$E9,$62,$BF,$CF,$F1,$DE,$B1
	.dc.b	$7C,$A7,$AC,$62,$75,$7C,$E7,$D1,$79,$7C,$11,$97,$8B,$70,$5E,$E7
	.dc.b	$D3,$56,$73,$5B,$F9,$8C,$19,$FA,$9B,$19,$3B,$7D,$3F,$ED,$3C,$5C
	.dc.b	$63,$EC,$02,$34,$E4,$C7,$7F,$F8,$F5,$C4,$6B,$5F,$85,$93,$2F,$44
	.dc.b	$E7,$5F,$F0,$73,$6F,$BB,$5F,$AB,$3C,$30,$ED,$87,$CD,$AD,$1B,$C3
	.dc.b	$B6,$5E,$CA,$DE,$3D,$9B,$8F,$66,$BD,$A7,$86,$52,$3D,$73,$AB,$EB
	.dc.b	$18,$9D,$63,$E3,$AD,$65,$E7,$9C,$E3,$E1,$7E,$6D,$6D,$CD,$F8,$DC
	.dc.b	$CE,$E3,$B1,$27,$E5,$D7,$58,$C4,$EE,$73,$8F,$49,$FA,$B7,$71,$F4
	.dc.b	$D7,$86,$78,$9E,$13,$EE,$46,$AF,$A1,$23,$53,$D9,$E2,$76,$2C,$F1
	.dc.b	$77,$3C,$5D,$4F,$13,$C4,$48,$D4,$F1,$9D,$C7,$1C,$BE,$F6,$3F,$06
	.dc.b	$F8,$57,$63,$F8,$14,$66,$B8,$65,$F5,$32,$F8,$98,$7E,$22,$31,$7E
	.dc.b	$A8,$C6,$5F,$4A,$3A,$1E,$22,$E5,$E5,$9D,$C7,$C0,$C7,$C0,$F0,$8F
	.dc.b	$3E,$1E,$0C,$3F,$1F,$18,$9F,$07,$D4,$BE,$0B,$FC,$E1,$1F,$A4,$8C
	.dc.b	$38,$3E,$71,$A9,$46,$A3,$0F,$0D,$3F,$4C,$DD,$3A,$5F,$F0,$DF,$A9
	.dc.b	$7D,$4F,$F9,$FE,$F7,$3F,$C1,$2F,$86,$1F,$B4,$6A,$BE,$2E,$E6,$3F
	.dc.b	$C0,$A7,$73,$A8,$C4,$EA,$BB,$C7,$F0,$B9,$D4,$60,$98,$7A,$27,$58
	.dc.b	$79,$E3,$13,$A9,$F3,$5F,$B6,$B5,$18,$3C,$24,$9F,$E6,$D7,$E0,$BE
	.dc.b	$64,$9E,$13,$AA,$E7,$1F,$C8,$F0,$F8,$78,$FD,$2C,$7D,$98,$7F,$AA
	.dc.b	$71,$E9,$7C,$A3,$A3,$C6,$7C,$CF,$03,$44,$3E,$76,$1E,$0B,$ED,$1E
	.dc.b	$9A,$E7,$7D,$A3,$67,$33,$B1,$F2,$0E,$E7,$13,$B3,$C4,$FF,$7E,$BF
	.dc.b	$85,$F9,$EF,$7C,$1B,$F0,$D2,$7F,$14,$D7,$FA,$27,$D4,$FF,$B9,$9F
	.dc.b	$EB,$4D,$7E,$F4,$6F,$4F,$EC,$D3,$96,$FC,$36,$71,$F9,$55,$F4,$2B
	.dc.b	$51,$FE,$D5,$F8,$07,$8D,$FE,$D7,$F0,$3F,$BA,$65,$FE,$CB,$FA,$9C
	.dc.b	$71,$ED,$B1,$FD,$72,$F8,$BF,$E0,$A3,$FD,$A5,$7A,$79,$19,$84,$E5
	.dc.b	$E3,$1F,$EA,$57,$FD,$7E,$EF,$E8,$B4,$6F,$DC,$91,$92,$1C,$5F,$07
	.dc.b	$D4,$BE,$A7,$B7,$A7,$2B,$CD,$39,$6C,$E5,$F1,$4F,$28,$91,$A2,$1E
	.dc.b	$7E,$0F,$77,$D4,$FF,$75,$8F,$84,$BE,$0F,$26,$BA,$E4,$EA,$32,$F0
	.dc.b	$8E,$7F,$29,$23,$9B,$DB,$E2,$B1,$1F,$EC,$BB,$4E,$38,$BE,$CC,$2B
	.dc.b	$E9,$7D,$4B,$CD,$F3,$FA,$73,$1F,$B1,$39,$6C,$E5,$ED,$3C,$A2,$46
	.dc.b	$88,$78,$FB,$BF,$01,$F5,$2F,$A9,$A4,$9F,$C3,$43,$E6,$F1,$1F,$EE
	.dc.b	$58,$E0,$EA,$79,$CF,$4F,$9E,$EB,$F4,$B2,$B9,$C6,$5E,$B1,$97,$A4
	.dc.b	$7E,$4F,$C5,$FF,$26,$6F,$F9,$EF,$A8,$7D,$4F,$E5,$DB,$7F,$9B,$AD
	.dc.b	$F7,$7F,$8E,$A3,$FD,$63,$DF,$BB,$91,$B8,$EC,$47,$07,$E9,$6F,$F3
	.dc.b	$F8,$78,$7A,$FC,$04,$9D,$76,$55,$3B,$79,$4F,$9E,$7D,$1D,$3E,$9F
	.dc.b	$93,$95,$DE,$32,$F8,$63,$2F,$4C,$FD,$98,$FF,$0D,$C7,$F8,$0E,$9F
	.dc.b	$99,$AF,$DE,$7B,$1F,$C6,$5B,$E2,$FC,$BE,$C9,$18,$21,$D4,$FF,$30
	.dc.b	$D7,$37,$07,$FB,$6A,$38,$B8,$8D,$5F,$F1,$2B,$FD,$64,$F0,$FA,$67
	.dc.b	$17,$F0,$DF,$A9,$7D,$4D,$3F,$88,$4E,$AB,$F8,$A2,$F5,$33,$F5,$32
	.dc.b	$73,$1F,$C5,$FE,$A3,$F5,$11,$82,$1E,$87,$2D,$F8,$EF,$E3,$DF,$52
	.dc.b	$F8,$8A,$73,$8F,$E4,$5F,$53,$F3,$79,$3F,$15,$82,$75,$87,$F2,$9A
	.dc.b	$D3,$FE,$88,$6A,$69,$8F,$A9,$7C,$A7,$F2,$FF,$67,$F5,$95,$FC,$CE
	.dc.b	$7F,$D3,$ED,$F6,$B5,$E0,$48,$C1,$0F,$AB,$5C,$9B,$1B,$69,$D7,$C3
	.dc.b	$FE,$7D,$F0,$5E,$28,$FE,$85,$F4,$AF,$82,$FE,$84,$D8,$FE,$8D,$3B
	.dc.b	$7F,$06,$DE,$9F,$EA,$33,$B9,$FC,$3B,$63,$FA,$AC,$EF,$83,$F6,$8D
	.dc.b	$AF,$F5,$7B,$53,$69,$F1,$3F,$D5,$7C,$57,$BF,$FD,$5F,$E9,$FE,$D6
	.dc.b	$1F,$D6,$DA,$FE,$62,$76,$FF,$5F,$6B,$F9,$C8,$DE,$5F,$DA,$9B,$97
	.dc.b	$F0,$9A,$DC,$FF,$6E,$6C,$66,$FC,$87,$8F,$E3,$FF,$B5,$7D,$4B,$E8
	.dc.b	$FC,$DF,$53,$33,$FE,$E1,$39,$6C,$E5,$F2,$4F,$28,$91,$A2,$1E,$9F
	.dc.b	$27,$FB,$C7,$D4,$FE,$07,$4F,$57,$B3,$F1,$4C,$F2,$4B,$F8,$D1,$97
	.dc.b	$FC,$05,$5C,$75,$FE,$50,$D8,$FF,$71,$B5,$37,$77,$D0,$F8,$BE,$1F
	.dc.b	$EA,$7B,$52,$FA,$C6,$5F,$AD,$19,$7F,$86,$DA,$7F,$C2,$5F,$52,$E5
	.dc.b	$BD,$32,$FF,$C3,$13,$96,$CF,$02,$3F,$E1,$DF,$A9,$FE,$DF,$BF,$FB
	.dc.b	$F3,$EA,$1F,$53,$FD,$16,$73,$49,$CA,$18,$77,$75,$5F,$D3,$63,$13
	.dc.b	$C2,$55,$D2,$30,$DA,$E1,$87,$F5,$59,$E1,$7C,$A7,$C9,$EE,$7F,$4A
	.dc.b	$C3,$B9,$E3,$C3,$93,$E5,$AE,$0D,$C3,$A5,$F2,$C7,$E0,$36,$79,$9E
	.dc.b	$C9,$DB,$C4,$D9,$DF,$0F,$16,$3E,$A6,$D7,$AA,$76,$73,$2E,$3E,$67
	.dc.b	$E1,$CF,$07,$B1,$E2,$C7,$D2,$EB,$1E,$33,$B8,$E8,$6F,$1F,$3B,$B3
	.dc.b	$D7,$7F,$69,$FF,$65,$B5,$F8,$06,$27,$E6,$3A,$F6,$DF,$23,$FE,$CF
	.dc.b	$8F,$B3,$18,$7F,$DA,$55,$FE,$35,$9D,$36,$38,$0F,$85,$AD,$32,$E0
	.dc.b	$8F,$D2,$F9,$37,$D3,$B9,$7E,$5C,$FE,$46,$BE,$6D,$6A,$73,$3D,$77
	.dc.b	$FC,$84,$E6,$3E,$E9,$C4,$ED,$7D,$5C,$9C,$77,$FF,$13,$4F,$1A,$FC
	.dc.b	$D4,$F1,$7A,$D7,$17,$31,$C4,$FD,$34,$F1,$AF,$F7,$C6,$1E,$8B,$E1
	.dc.b	$18,$AF,$1E,$1E,$AA,$D3,$63,$27,$9A,$32,$D3,$38,$78,$A3,$2F,$9A
	.dc.b	$7F,$D3,$98,$7C,$3D,$BF,$D4,$B7,$EB,$C3,$DF,$8F,$C3,$9F,$F5,$4D
	.dc.b	$6E,$F2,$7F,$A9,$6F,$81,$EE,$AE,$11,$F2,$3A,$6C,$FC,$26,$FC,$6B
	.dc.b	$EA,$3F,$D1,$95,$E4,$9D,$D3,$0F,$E8,$D3,$FC,$3F,$1F,$27,$D4,$BB
	.dc.b	$F7,$5A,$BF,$F1,$B5,$2B,$57,$F0,$2F,$67,$8B,$E1,$3F,$BE,$46,$49
	.dc.b	$7F,$E3,$CC,$38,$BF,$E4,$1C,$D5,$C2,$35,$7D,$2F,$CC,$7F,$25,$AF
	.dc.b	$77,$C5,$66,$B9,$68,$8F,$7F,$8F,$FC,$F6,$FF,$DA,$0D,$AF,$F2,$84
	.dc.b	$3E,$9D,$87,$0E,$A9,$3C,$1F,$9E,$EE,$33,$5F,$E5,$6B,$E9,$B7,$E7
	.dc.b	$EB,$A7,$06,$A3,$18,$FD,$FB,$FE,$7D,$F5,$2F,$77,$EA,$B7,$0F,$7C
	.dc.b	$FE,$D6,$9E,$F6,$AE,$57,$FB,$AE,$F8,$5F,$B2,$F7,$39,$AD,$CE,$5B
	.dc.b	$87,$91,$FC,$2C,$FF,$B0,$CF,$F7,$B1,$67,$A7,$81,$AE,$A9,$3A,$7E
	.dc.b	$2E,$1F,$DD,$23,$04,$AF,$9F,$F5,$3F,$E5,$28,$C6,$3F,$1E,$31,$3F
	.dc.b	$E8,$0F,$9B,$FC,$2C,$FF,$09,$84,$62,$BF,$B3,$1F,$66,$BF,$3B,$19
	.dc.b	$9F,$ED,$4D,$78,$F1,$F9,$A7,$6C,$AF,$9E,$7E,$0F,$C4,$C5,$7C,$B7
	.dc.b	$F5,$3F,$50,$FA,$9F,$D5,$FD,$43,$C4,$CC,$61,$98,$7D,$AC,$15,$B8
	.dc.b	$C5,$F3,$75,$19,$9F,$F5,$0B,$FB,$A6,$FC,$0F,$46,$CF,$0F,$53,$C6
	.dc.b	$7E,$FB,$EA,$7A,$4F,$28,$FE,$03,$D0,$F3,$29,$C2,$FC,$D1,$97,$D1
	.dc.b	$F5,$3F,$E6,$26,$FA,$5C,$BD,$AF,$E1,$73,$9C,$7F,$D6,$FF,$4D,$FE
	.dc.b	$97,$BF,$80,$D8,$FF,$5F,$7D,$43,$EA,$5F,$53,$FC,$52,$B5,$D1,$F7
	.dc.b	$DB,$87,$FB,$AD,$B8,$30,$E5,$7C,$7D,$D9,$7F,$A8,$F1,$9F,$53,$FE
	.dc.b	$78,$8F,$1C,$FF,$28,$9C,$FB,$1F,$CA,$BB,$3F,$C6,$1F,$53,$73,$FC
	.dc.b	$54,$63,$0F,$D9,$BF,$E3,$5E,$CD,$56,$DB,$3A,$AE,$23,$EA,$7A,$1D
	.dc.b	$78,$AC,$EE,$2C,$FF,$B8,$F0,$FF,$21,$56,$38,$3F,$45,$F5,$2F,$5F
	.dc.b	$F9,$DB,$71,$FE,$0F,$CC,$FA,$9F,$F6,$D4,$7E,$46,$75,$97,$FB,$B1
	.dc.b	$B8,$74,$C3,$FA,$4D,$F0,$AD,$47,$63,$3E,$55,$C3,$BD,$7F,$20,$7F
	.dc.b	$5F,$EA,$7F,$82,$DF,$8E,$FE,$17,$0F,$F0,$16,$9F,$E2,$09,$FA,$13
	.dc.b	$97,$89,$8E,$9C,$B4,$E7,$F5,$3C,$0E,$3F,$1B,$63,$24,$D3,$E5,$5A
	.dc.b	$75,$1F,$07,$E7,$3E,$49,$F8,$3F,$81,$7C,$BE,$30,$F9,$13,$69,$A7
	.dc.b	$F8,$6F,$9F,$FD,$8F,$79,$8F,$AB,$5F,$DD,$17,$D4,$FC,$B7,$38,$7F
	.dc.b	$3C,$73,$F1,$5F,$0B,$E4,$24,$FC,$8F,$A9,$B9,$FA,$A6,$DF,$CA,$D9
	.dc.b	$58,$BF,$A3,$F3,$3F,$33,$E2,$3E,$6B,$63,$E9,$B7,$1F,$8A,$DC,$3F
	.dc.b	$97,$7D,$4F,$F7,$16,$DF,$C5,$F3,$3F,$C1,$33,$9C,$3F,$D2,$D8,$23
	.dc.b	$0F,$FA,$65,$BC,$1F,$EE,$CA,$E3,$72,$FF,$2F,$F5,$2F,$98,$C6,$3F
	.dc.b	$E0,$38,$C4,$FE,$2B,$EA,$7F,$0F,$F5,$3F,$88,$6C,$64,$FC,$7C,$65
	.dc.b	$A7,$E4,$B8,$7F,$1F,$D7,$F9,$06,$C6,$4F,$CB,$C6,$5A,$7F,$39,$FA
	.dc.b	$8F,$E6,$9B,$FF,$53,$6C,$7C,$27,$F8,$76,$5F,$86,$8C,$6B,$FE,$30
	.dc.b	$9F,$E7,$9F,$09,$EF,$AF,$F5,$F4,$3E,$3F,$FC,$CD,$39,$94,$F0,$DB
	.dc.b	$B5,$69,$B8,$7F,$9D,$DB,$5F,$E8,$37,$C3,$19,$8E,$F3,$C5,$B1,$97
	.dc.b	$BC,$65,$F0,$E1,$6F,$CD,$5B,$7F,$C6,$ED,$F8,$1D,$7D,$4E,$D9,$1D
	.dc.b	$7E,$9F,$F6,$5E,$26,$E3,$1D,$DF,$E3,$0C,$BD,$07,$FB,$86,$B6,$E0
	.dc.b	$38,$D7,$E3,$65,$E2,$9D,$4E,$75,$FC,$AB,$63,$A5,$F7,$BE,$39,$7B
	.dc.b	$EB,$55,$DA,$B5,$7C,$9B,$1F,$62,$B8,$46,$23,$86,$5E,$C3,$DC,$F9
	.dc.b	$23,$57,$D8,$EF,$8F,$A7,$1F,$C0,$46,$67,$67,$59,$D6,$3E,$3C,$3B
	.dc.b	$61,$D5,$EF,$87,$2B,$E3,$1A,$8F,$7E,$56,$F9,$5F,$13,$E7,$61,$D9
	.dc.b	$F8,$F7,$B6,$9A,$CB,$E0,$C6,$63,$E0,$99,$BC,$65,$EB,$9F,$2E,$1E
	.dc.b	$0C,$7C,$D1,$C7,$1F,$06,$1C,$A7,$2F,$95,$F4,$B6,$B9,$11,$FA,$98
	.dc.b	$78,$32,$F8,$31,$E3,$38,$1E,$AA,$EA,$74,$2D,$73,$8F,$13,$FB,$DC
	.dc.b	$BD,$F1,$C8,$C9,$E0,$31,$97,$BD,$F0,$57,$47,$C1,$5D,$71,$F7,$D6
	.dc.b	$EB,$55,$FC,$02,$FB,$C6,$27,$F0,$F1,$8B,$F3,$46,$31,$F9,$FF,$52
	.dc.b	$F8,$2F,$E6,$B8,$FE,$8A,$30,$E0,$FA,$E6,$A9,$1A,$8C,$3C,$31,$FC
	.dc.b	$AB,$70,$E9,$7F,$C2,$7E,$A1,$F5,$3F,$D1,$2F,$73,$FC,$06,$F8,$61
	.dc.b	$FB,$D6,$AB,$E2,$EE,$63,$FB,$C9,$DC,$EA,$31,$3A,$AE,$F1,$FC,$1E
	.dc.b	$75,$18,$26,$1E,$89,$D6,$1E,$78,$C4,$EA,$7C,$D7,$ED,$AD,$46,$0F
	.dc.b	$09,$27,$F9,$C5,$F8,$2F,$99,$27,$84,$EA,$B9,$C7,$F1,$CC,$3E,$1E
	.dc.b	$3F,$4B,$1F,$66,$1F,$DB,$B1,$E9,$7C,$A3,$A3,$C6,$7C,$CF,$03,$44
	.dc.b	$3E,$76,$1E,$0B,$ED,$1E,$9A,$E7,$7D,$A3,$A1,$93,$93,$93,$53,$CB
	.dc.b	$1F,$91,$3D,$8F,$19,$C4,$E6,$FC,$17,$F1,$1F,$0B,$F0,$A9,$F3,$D4
	.dc.b	$C4,$49,$D9,$0F,$E8,$9F,$53,$FE,$E4,$7F,$AD,$35,$FC,$14,$71,$77
	.dc.b	$F0,$DF,$42,$75,$87,$D8,$AD,$3F,$DC,$63,$39,$7F,$87,$BE,$01,$E2
	.dc.b	$FF,$88,$3A,$3F,$DA,$90,$FA,$9C,$71,$6D,$BA,$04,$F2,$7F,$D8,$8D
	.dc.b	$7F,$C1,$06,$DE,$8E,$A3,$9B,$CA,$EC,$FF,$AE,$E3,$9F,$D1,$60,$CF
	.dc.b	$B9,$23,$24,$3B,$9E,$47,$AF,$D4,$BE,$A7,$AF,$A3,$2B,$FD,$B5,$39
	.dc.b	$6C,$E5,$E7,$87,$FA,$55,$C9,$0E,$F3,$C5,$B3,$EA,$E1,$F6,$FD,$4B
	.dc.b	$A6,$FC,$14,$BF,$F7,$54,$E5,$B3,$97,$D7,$09,$E7,$3D,$3E,$4E,$57
	.dc.b	$27,$BF,$C4,$E2,$38,$C5,$E2,$71,$7F,$97,$B1,$5F,$D2,$FA,$97,$99
	.dc.b	$E8,$F4,$66,$5F,$EF,$F9,$CB,$67,$2F,$9A,$79,$44,$8D,$10,$F5,$F7
	.dc.b	$7E,$03,$EA,$5F,$53,$49,$3F,$8D,$D3,$CE,$FE,$3E,$5F,$37,$A8,$FB
	.dc.b	$F1,$C9,$E1,$AF,$F9,$EA,$31,$1A,$9F,$07,$CF,$78,$7E,$9E,$4F,$03
	.dc.b	$B4,$69,$C0,$A4,$F0,$8F,$0C,$FE,$6B,$C6,$FF,$2F,$63,$E8,$97,$D4
	.dc.b	$3E,$A7,$F3,$ED,$8E,$B5,$C3,$3F,$F4,$8B,$63,$F5,$D3,$B7,$F3,$AD
	.dc.b	$BE,$D3,$B9,$FF,$62,$77,$F8,$C8,$FA,$AB,$F6,$95,$E5,$7F,$5D,$1D
	.dc.b	$3B,$3C,$A7,$9B,$B6,$C7,$18,$AF,$8A,$3D,$7D,$3E,$CF,$93,$93,$DE
	.dc.b	$32,$F8,$63,$2F,$AA,$3E,$47,$57,$F0,$2F,$63,$F9,$3C,$FD,$06,$BF
	.dc.b	$BC,$D3,$F4,$4F,$C8,$9E,$1E,$2C,$8C,$10,$E1,$1D,$9E,$9D,$9F,$C4
	.dc.b	$3E,$9F,$ED,$C7,$DE,$FA,$97,$D4,$FF,$14,$6C,$7E,$0A,$76,$FF,$0E
	.dc.b	$5E,$9F,$3F,$A9,$93,$B8,$FE,$3B,$F5,$1F,$AE,$8C,$10,$FE,$2E,$E5
	.dc.b	$BF,$15,$FC,$93,$EA,$5F,$0F,$4E,$71,$FC,$A3,$EA,$7F,$A2,$73,$FE
	.dc.b	$3F,$04,$EB,$0F,$CC,$D6,$BA,$7F,$DB,$AD,$8F,$E7,$4D,$4D,$FA,$97
	.dc.b	$CA,$7F,$36,$F6,$7F,$63,$5F,$A4,$9F,$F5,$F3,$7D,$AD,$78,$52,$30
	.dc.b	$43,$F9,$AD,$72,$6C,$6D,$A7,$5F,$13,$FA,$2F,$C2,$F8,$A3,$F6,$7F
	.dc.b	$48,$F8,$1F,$E9,$2D,$8F,$DB,$CE,$DF,$E8,$6D,$E0,$FD,$A4,$EE,$7F
	.dc.b	$A3,$B6,$3F,$71,$3B,$DB,$FA,$F5,$23,$0F,$F6,$16,$B9,$D7,$C7,$F1
	.dc.b	$1F,$D7,$3C,$57,$C7,$FF,$5E,$FA,$7F,$B7,$7F,$C3,$1A,$FF,$58,$9D
	.dc.b	$BF,$C4,$1A,$FF,$5C,$8D,$E5,$FC,$35,$B9,$79,$AB,$73,$FC,$45,$B1
	.dc.b	$9B,$F2,$1E,$3F,$8F,$FE,$E3,$F5,$2F,$A3,$F3,$7D,$4C,$CF,$F9,$0C
	.dc.b	$E5,$B3,$97,$C9,$3C,$A2,$46,$88,$7A,$7C,$8F,$EF,$FF,$53,$F8,$1D
	.dc.b	$3D,$5E,$C7,$C5,$33,$CB,$2F,$E3,$46,$5F,$E6,$75,$C7,$5F,$AA,$D8
	.dc.b	$FF,$0A,$35,$37,$77,$D0,$F8,$BE,$1F,$EA,$7B,$72,$FA,$C6,$5F,$AD
	.dc.b	$19,$7F,$C1,$B6,$9F,$F0,$CF,$D4,$B9,$EF,$54,$BF,$E9,$33,$96,$CF
	.dc.b	$02,$3F,$D3,$BE,$A5,$ED,$B8,$3F,$C5,$DF,$52,$FA,$9F,$C9,$D6,$A7
	.dc.b	$31,$F9,$6C,$7B,$60,$BE,$0F,$F8,$DE,$33,$18,$85,$F3,$78,$57,$6C
	.dc.b	$3C,$51,$83,$96,$3F,$A4,$BE,$58,$74,$BE,$18,$F9,$2F,$FC,$8D,$8F
	.dc.b	$5B,$DE,$1C,$6F,$D0,$DA,$DE,$3E,$0A,$F3,$5F,$1B,$DD,$75,$18,$F8
	.dc.b	$5E,$38,$E6,$B8,$47,$A0,$E1,$3B,$7E,$0E,$1D,$EF,$D3,$1E,$F3,$AE
	.dc.b	$3E,$56,$9D,$67,$83,$E0,$38,$57,$53,$B5,$F8,$A3,$BE,$1F,$ED,$76
	.dc.b	$9F,$ED,$98,$FA,$4F,$FB,$79,$AF,$E0,$3D,$C7,$C9,$71,$9F,$F1,$CB
	.dc.b	$63,$33,$CD,$AF,$7F,$84,$F0,$3F,$E7,$96,$A6,$91,$F8,$5F,$20,$FA
	.dc.b	$7F,$F1,$9E,$DF,$89,$9E,$46,$A3,$3B,$7F,$BC,$EF,$EB,$5F,$1E,$5F
	.dc.b	$B1,$7C,$30,$FC,$94,$E6,$3E,$81,$9B,$FE,$63,$19,$9F,$04,$F9,$CF
	.dc.b	$05,$74,$71,$1D,$30,$F1,$C7,$F3,$A3,$37,$D2,$33,$1F,$EF,$D3,$99
	.dc.b	$E2,$36,$7F,$A7,$EF,$C5,$3E,$D8,$DD,$79,$30,$F7,$DF,$49,$F5,$1C
	.dc.b	$5C,$3D,$05,$F4,$75,$87,$F9,$42,$3E,$26,$9F,$E4,$B9,$E6,$70,$AF
	.dc.b	$F3,$2C,$7F,$93,$E7,$75,$ED,$3F,$CB,$35,$C2,$32,$FA,$FA,$78,$CF
	.dc.b	$CE,$6C,$3E,$35,$F5,$1F,$E9,$1A,$F2,$56,$AB,$D2,$F0,$C1,$F5,$2E
	.dc.b	$FD,$65,$FD,$37,$0F,$F3,$75,$6A,$72,$48,$FF,$39,$65,$FE,$A7,$F1
	.dc.b	$71,$97,$FA,$F3,$0E,$95,$F1,$6F,$CF,$7B,$8C,$12,$BF,$DA,$31,$EA
	.dc.b	$F8,$AE,$78,$FF,$A3,$23,$15,$FE,$7E,$F1,$FF,$9E,$37,$A7,$E1,$AF
	.dc.b	$A8,$F8,$8E,$49,$8F,$FB,$42,$B8,$BF,$EE,$08,$E1,$7F,$1E,$73,$87
	.dc.b	$FA,$15,$B3,$FE,$EB,$3B,$70,$6F,$1F,$0E,$3F,$63,$17,$D4,$BD,$CF
	.dc.b	$A4,$DC,$3F,$D3,$47,$9F,$DC,$CC,$C8,$FF,$4D,$CE,$EB,$CD,$6B,$D4
	.dc.b	$7F,$BA,$A7,$59,$FF,$AC,$AF,$E1,$61,$A8,$C6,$1A,$9F,$F1,$39,$FE
	.dc.b	$CC,$3F,$D5,$25,$BE,$FE,$07,$E2,$63,$15,$FE,$A7,$69,$CB,$A3,$E0
	.dc.b	$CE,$D9,$1E,$EF,$A9,$FF,$2A,$BD,$A3,$EB,$C6,$2F,$FB,$54,$FF,$8E
	.dc.b	$BE,$6F,$F1,$31,$80,$1C,$E1,$EC,$AF,$CF,$46,$2B,$97,$16,$0F,$AB
	.dc.b	$36,$3F,$D8,$F3,$BB,$FF,$77,$E4,$F1,$35,$1F,$41,$FD,$57,$D4,$3E
	.dc.b	$A7,$F5,$9F,$52,$F1,$33,$18,$66,$5F,$ED,$49,$D3,$FE,$D6,$AE,$4E
	.dc.b	$67,$9F,$BD,$23,$07,$17,$BF,$A1,$23,$FD,$BF,$F5,$3C,$B3,$F8,$C3
	.dc.b	$1F,$99,$C1,$FE,$DD,$E9,$79,$9F,$C0,$DB,$87,$97,$1C,$4F,$F0,$7F
	.dc.b	$A9,$7A,$54,$E1,$EC,$DB,$FF,$77,$FD,$47,$FA,$69,$BE,$CF,$F1,$4F
	.dc.b	$A8,$7D,$4B,$EA,$7F,$89,$5E,$F8,$3C,$ED,$DF,$FD,$E4,$DC,$7E,$FE
	.dc.b	$1C,$AF,$8E,$32,$7F,$07,$67,$FA,$07,$D4,$FF,$9E,$9F,$19,$A3,$3E
	.dc.b	$C7,$F2,$8E,$AF,$C5,$FD,$4D,$D3,$F0,$51,$8A,$E1,$1F,$8F,$32,$27
	.dc.b	$87,$6F,$16,$CF,$0A,$FC,$BF,$D4,$F5,$3B,$70,$59,$D6,$09,$C3,$F9
	.dc.b	$FA,$FC,$FD,$63,$3F,$74,$FE,$8B,$EA,$5E,$BF,$F3,$D6,$E3,$FC,$D3
	.dc.b	$EA,$7F,$DB,$EF,$AA,$93,$A7,$F0,$ED,$BE,$B1,$AB,$FE,$6F,$7C,$A3
	.dc.b	$71,$DC,$CF,$8F,$70,$F0,$3F,$37,$D3,$FA,$91,$F0,$DF,$C5,$B4,$D7
	.dc.b	$36,$DA,$64,$E0,$F3,$FA,$99,$1C,$FE,$32,$E3,$F1,$ED,$58,$CF,$E3
	.dc.b	$ED,$FE,$32,$BC,$61,$FE,$C6,$D5,$F3,$7F,$24,$FE,$D7,$E7,$FF,$B2
	.dc.b	$7E,$5F,$15,$F2,$6B,$6D,$AD,$FC,$12,$B7,$AF,$CC,$9F,$ED,$DF,$53
	.dc.b	$F2,$74,$F9,$B8,$F2,$AF,$E3,$90,$9D,$63,$FC,$7B,$4F,$F2,$4F,$C8
	.dc.b	$7C,$B3,$E6,$7D,$4D,$9F,$99,$17,$0F,$85,$37,$0F,$83,$C0,$F9,$9F
	.dc.b	$A9,$EB,$7D,$56,$E3,$F5,$5B,$87,$CB,$6D,$FE,$13,$EA,$7F,$DF,$55
	.dc.b	$F8,$B6,$D7,$F3,$5E,$1F,$F7,$87,$95,$FE,$0F,$9C,$DF,$1C,$70,$7C
	.dc.b	$89,$B8,$73,$8F,$F0,$2F,$D4,$BE,$5F,$39,$7F,$81,$23,$15,$F8,$BF
	.dc.b	$A9,$FC,$77,$D4,$FE,$3D,$B1,$93,$F2,$11,$96,$9F,$93,$F9,$1F,$CB
	.dc.b	$75,$FE,$5D,$B1,$93,$F3,$11,$96,$9F,$AB,$FA,$8F,$E7,$78,$FF,$30
	.dc.b	$6E,$BF,$AA,$7F,$89,$46,$75,$FF,$19,$47,$0F,$81,$F7,$4F,$F5,$C8
	.dc.b	$7C,$77,$EF,$A7,$32,$9E,$1B,$7F,$62,$8C,$56,$E3,$13,$CB,$1F,$F4
	.dc.b	$44,$62,$75,$5D,$F0,$AF,$2A,$DB,$E3,$9C,$DF,$38,$FF,$48,$5F,$6B
	.dc.b	$CD,$6B,$E5,$E6,$1C,$AF,$D1,$F5,$38,$78,$9E,$F9,$F0,$FD,$3B,$BF
	.dc.b	$FC,$B5,$3A,$3E,$04,$EA,$33,$A7,$C4,$AE,$1E,$97,$76,$DF,$1D,$58
	.dc.b	$71,$AD,$61,$CB,$5F,$95,$87,$0C,$7E,$7E,$1C,$33,$DE,$13,$2F,$25
	.dc.b	$EF,$0F,$9F,$97,$59,$CE,$5D,$5F,$F0,$BD,$6D,$F8,$18,$73,$CB,$D5
	.dc.b	$5B,$95,$6D,$F8,$35,$BC,$3C,$F5,$BA,$D6,$5E,$7B,$F0,$E5,$B7,$C3
	.dc.b	$5D,$67,$58,$F8,$2B,$55,$E0,$AD,$65,$EB,$AD,$4F,$43,$E6,$46,$71
	.dc.b	$F1,$DF,$2C,$3A,$C6,$2B,$85,$F2,$3E,$6D,$F1,$8F,$2D,$F1,$8C,$47
	.dc.b	$8E,$31,$97,$9B,$2F,$88,$7C,$0A,$DB,$8C,$BD,$33,$B7,$E3,$18,$BF
	.dc.b	$1E,$3E,$5C,$7C,$8F,$DA,$AE,$18,$F8,$2B,$67,$98,$A7,$D4,$2E,$1E
	.dc.b	$18,$FD,$06,$5F,$0A,$78,$61,$D9,$B8,$77,$78,$E3,$E6,$BE,$39,$7B
	.dc.b	$23,$13,$F4,$32,$F8,$5F,$52,$F8,$2F,$F7,$34,$FE,$3A,$30,$E3,$4F
	.dc.b	$D1,$BF,$EE,$B5,$1A,$8C,$3C,$33,$FD,$63,$73,$FD,$2B,$67,$FC,$E3
	.dc.b	$F5,$0F,$A9,$FE,$25,$7C,$31,$FD,$95,$EE,$FF,$36,$DB,$E1,$1A,$98
	.dc.b	$FE,$B2,$75,$7D,$30,$F0,$D6,$B0,$ED,$7F,$C1,$70,$E6,$F1,$9D,$57
	.dc.b	$08,$D5,$F3,$7E,$34,$E9,$A6,$8C,$1E,$86,$D7,$6B,$F0,$92,$FB,$3F
	.dc.b	$CD,$A3,$51,$82,$57,$69,$D5,$F8,$E3,$B1,$E9,$24,$69,$F1,$E3,$F6
	.dc.b	$30,$F1,$D7,$18,$C4,$F0,$BE,$EB,$0F,$1B,$E1,$8C,$CF,$0C,$3B,$4F
	.dc.b	$13,$44,$39,$5F,$9E,$F9,$C7,$A2,$7E,$13,$F8,$49,$D1,$E5,$C7,$F0
	.dc.b	$57,$DF,$0F,$EB,$18,$72,$1F,$0B,$D7,$E7,$66,$3F,$D2,$67,$2D,$9F
	.dc.b	$11,$1F,$E9,$DF,$53,$FE,$E3,$6F,$C3,$65,$FC,$34,$FC,$F9,$D4,$7F
	.dc.b	$4D,$6E,$9F,$66,$78,$3F,$DC,$23,$38,$FF,$6F,$F8,$0F,$F2,$3D,$FF
	.dc.b	$9A,$D7,$D4,$FF,$73,$E8,$FF,$34,$C6,$27,$5E,$46,$78,$B9,$36,$3F
	.dc.b	$DB,$73,$B7,$FD,$72,$DA,$E7,$3B,$9F,$F6,$A7,$B7,$69,$F4,$14,$C4
	.dc.b	$49,$F3,$46,$5F,$44,$73,$FA,$97,$D4,$F3,$F4,$65,$7F,$A1,$E7,$2D
	.dc.b	$9C,$BE,$69,$E5,$12,$34,$43,$D7,$C3,$E9,$FA,$9F,$EE,$71,$FE,$93
	.dc.b	$5F,$07,$93,$5D,$72,$75,$19,$78,$47,$3F,$94,$91,$CD,$ED,$F1,$38
	.dc.b	$8F,$F7,$1D,$A7,$1C,$1F,$E9,$E8,$57,$CD,$FA,$9F,$F3,$35,$F9,$7D
	.dc.b	$3F,$41,$9F,$36,$46,$48,$7F,$BC,$27,$8B,$63,$6D,$3E,$07,$7F,$DB
	.dc.b	$FA,$87,$D4,$D2,$4F,$FA,$DE,$1F,$37,$88,$FF,$7E,$C7,$07,$53,$CE
	.dc.b	$7A,$7C,$F7,$5F,$A9,$95,$CE,$32,$F5,$8C,$BD,$23,$F1,$DE,$2F,$F9
	.dc.b	$43,$7F,$F6,$1F,$D4,$3E,$A7,$F2,$4D,$BF,$F6,$65,$6F,$B7,$FC,$A8
	.dc.b	$D8,$09,$E0,$AF,$AD,$F8,$3B,$E4,$F4,$6B,$BC,$3F,$CC,$17,$FB,$28
	.dc.b	$C1,$D9,$75,$F9,$89,$1F,$EE,$18,$FD,$3B,$63,$2F,$23,$CC,$FB,$BB
	.dc.b	$3D,$FD,$3F,$03,$E4,$E5,$73,$8C,$BD,$63,$2F,$3B,$FE,$1F,$9F,$FB
	.dc.b	$B3,$4F,$CF,$CF,$FB,$CF,$D8,$78,$DF,$BD,$6C,$7C,$CE,$CF,$A6,$66
	.dc.b	$64,$64,$87,$A6,$BA,$B6,$7A,$76,$7F,$09,$FA,$77,$1F,$F0,$BF,$A8
	.dc.b	$7D,$4D,$35,$1B,$F4,$F8,$7D,$4C,$CB,$ED,$EC,$E4,$FF,$49,$64,$BF
	.dc.b	$E2,$DF,$51,$FA,$CB,$DD,$6A,$7A,$46,$0F,$C2,$CE,$DF,$A4,$BE,$2F
	.dc.b	$F9,$07,$D4,$BE,$1E,$9D,$A3,$F9,$27,$D4,$FC,$EC,$7F,$D4,$97,$C7
	.dc.b	$0E,$C1,$C3,$F7,$E7,$83,$F9,$66,$A6,$98,$FA,$97,$CA,$7F,$33,$F6
	.dc.b	$BF,$6B,$5D,$A7,$FD,$72,$DF,$6B,$5E,$0C,$8C,$10,$FC,$85,$72,$6C
	.dc.b	$6D,$A7,$5F,$13,$FA,$17,$C3,$F8,$A3,$DB,$F4,$AF,$81,$FE,$8C,$D8
	.dc.b	$F8,$33,$B7,$F3,$CD,$E4,$FB,$51,$FB,$69,$E0,$FE,$8D,$B1,$A8,$4F
	.dc.b	$1D,$FF,$DA,$4D,$8F,$DF,$B5,$DE,$BE,$4F,$8A,$FE,$B5,$E3,$3E,$3F
	.dc.b	$FA,$E7,$D3,$FD,$DB,$FA,$ED,$7F,$69,$3B,$7E,$DB,$5F,$DC,$46,$F2
	.dc.b	$F3,$37,$2F,$F7,$35,$6E,$7D,$4D,$8C,$DF,$90,$F1,$FC,$87,$F6,$FF
	.dc.b	$A9,$7D,$17,$9B,$EA,$66,$7F,$8C,$9C,$B6,$72,$F9,$27,$94,$48,$D1
	.dc.b	$0F,$4F,$95,$FD,$F7,$EA,$7F,$05,$A7,$AB,$D9,$F8,$A6,$79,$65,$FC
	.dc.b	$68,$CB,$F9,$EA,$E3,$AF,$F3,$08,$46,$1F,$E6,$6D,$73,$5F,$4B,$E8
	.dc.b	$7C,$5F,$0F,$F5,$3D,$A9,$5D,$63,$2F,$D6,$8C,$BF,$CA,$6D,$3F,$E1
	.dc.b	$7F,$A9,$73,$DE,$A9,$7F,$B4,$9C,$B6,$78,$11,$FD,$CF,$D4,$FF,$79
	.dc.b	$DB,$FA,$5F,$D4,$3E,$A7,$DD,$39,$A4,$E5,$FE,$A3,$87,$67,$35,$F1
	.dc.b	$63,$13,$B9,$57,$38,$C3,$6B,$86,$1F,$36,$78,$5F,$29,$F1,$FB,$BF
	.dc.b	$D7,$70,$EE,$7F,$62,$C3,$93,$E4,$C3,$95,$74,$8F,$6B,$CB,$1B,$1C
	.dc.b	$6F,$A6,$3E,$0B,$E3,$3C,$CD,$9E,$A9,$DD,$F3,$8F,$18,$C7,$D0,$F1
	.dc.b	$33,$8F,$99,$E5,$8F,$49,$DC,$71,$C3,$C2,$E8,$E3,$87,$81,$E9,$8F
	.dc.b	$9D,$F2,$E1,$D4,$FB,$0F,$E5,$CE,$45,$72,$FA,$4C,$7B,$8F,$91,$FF
	.dc.b	$70,$C7,$D8,$8C,$3F,$EE,$49,$FF,$1E,$CE,$9B,$1C,$07,$C2,$56,$99
	.dc.b	$70,$47,$E4,$FC,$8B,$E9,$FF,$94,$E3,$F8,$99,$FF,$0F,$D7,$D2,$AD
	.dc.b	$72,$FD,$49,$C9,$CA,$F8,$EF,$FE,$FA,$BE,$11,$89,$F8,$91,$8A,$E5
	.dc.b	$1E,$6C,$3F,$31,$87,$68,$FB,$A7,$68,$CD,$F3,$8F,$F7,$B9,$E1,$AE
	.dc.b	$77,$D6,$38,$9D,$70,$FE,$15,$87,$5C,$3D,$D3,$93,$FC,$7B,$19,$9F
	.dc.b	$54,$75,$39,$D7,$5A,$E2,$D9,$EB,$87,$F5,$2D,$FF,$A7,$61,$EB,$7F
	.dc.b	$A8,$61,$EB,$AD,$E3,$FD,$F2,$B4,$E2,$7F,$C7,$46,$CF,$2D,$F5,$8F
	.dc.b	$5F,$45,$9F,$4B,$7E,$31,$F5,$3F,$E9,$4A,$ED,$3B,$A6,$1F,$D1,$27
	.dc.b	$FC,$35,$93,$EA,$5D,$DB,$AD,$5F,$F9,$5A,$B5,$39,$25,$FF,$88,$F2
	.dc.b	$7A,$9F,$DE,$E3,$DD,$3A,$BF,$F2,$B6,$1D,$A7,$85,$6E,$73,$5B,$C3
	.dc.b	$FB,$EE,$1E,$CF,$8A,$C6,$1F,$E7,$48,$C3,$FE,$26,$F2,$3F,$9C,$B7
	.dc.b	$A7,$E3,$2F,$A7,$57,$C3,$AD,$E5,$FE,$69,$6E,$3F,$E0,$76,$DF,$FA
	.dc.b	$08,$E5,$5C,$70,$91,$FE,$43,$3B,$70,$63,$0F,$F3,$EC,$60,$FF,$1D
	.dc.b	$FD,$4B,$DD,$FA,$8D,$8E,$59,$BD,$EC,$D4,$9E,$2F,$B2,$75,$18,$25
	.dc.b	$F8,$23,$A3,$FE,$85,$BB,$A7,$F8,$A2,$BE,$25,$FE,$15,$FF,$65,$9F
	.dc.b	$E5,$02,$D7,$3F,$03,$5D,$52,$74,$F8,$E3,$13,$FE,$93,$9E,$37,$DF
	.dc.b	$EA,$7F,$CA,$31,$C0,$F9,$0E,$27,$FD,$35,$F3,$3F,$85,$6F,$17,$BE
	.dc.b	$7F,$35,$39,$BF,$F1,$E3,$6F,$FC,$10,$DB,$FF,$3F,$68,$BE,$38,$4B
	.dc.b	$F9,$FE,$25,$BF,$90,$FE,$9F,$EA,$1F,$53,$FA,$9F,$A8,$78,$99,$8C
	.dc.b	$33,$0F,$F3,$BC,$E9,$FF,$3C,$D7,$27,$33,$CF,$DF,$91,$83,$8B,$DF
	.dc.b	$D0,$91,$FE,$82,$FA,$9E,$59,$FF,$AF,$F1,$F9,$1C,$1F,$E5,$DE,$8F
	.dc.b	$F6,$4D,$FF,$A2,$1B,$5B,$C3,$F8,$14,$62,$BF,$CC,$9F,$52,$F4,$BF
	.dc.b	$83,$37,$0E,$5E,$D6,$27,$F8,$57,$D4,$7F,$A7,$67,$FD,$2A,$DC,$BF
	.dc.b	$DB,$7F,$50,$FA,$97,$D4,$F0,$F7,$A6,$1C,$1F,$E3,$55,$B9,$CD,$FF
	.dc.b	$B6,$F0,$E9,$A7,$F9,$FB,$C4,$7D,$4F,$F9,$C7,$2E,$47,$FA,$CE,$33
	.dc.b	$EB,$FF,$25,$EC,$FF,$7C,$FD,$4D,$CB,$F0,$91,$8B,$FF,$5D,$3F,$EF
	.dc.b	$EE,$BD,$56,$DB,$3A,$AE,$23,$EA,$7A,$1D,$78,$AC,$EE,$2C,$FF,$BF
	.dc.b	$B0,$3F,$EC,$39,$D1,$FE,$BE,$FA,$77,$AD,$FC,$DD,$B1,$FC,$B7,$EA
	.dc.b	$7F,$DA,$CE,$E7,$44,$75,$F0,$B6,$DC,$1F,$EA,$1C,$39,$C6,$2B,$F9
	.dc.b	$56,$1F,$C7,$A3,$CB,$3A,$AF,$E6,$33,$9C,$7F,$C3,$17,$F3,$7D,$4F
	.dc.b	$C3,$61,$ED,$7F,$DD,$91,$F3,$1C,$DF,$FA,$DA,$B8,$CE,$79,$F9,$3C
	.dc.b	$DE,$9F,$53,$23,$97,$C6,$5C,$7B,$4D,$C3,$A6,$2A,$C5,$FF,$A2,$73
	.dc.b	$7C,$DF,$CB,$3F,$B4,$72,$7F,$B0,$BE,$63,$15,$F4,$66,$E7,$F1,$4F
	.dc.b	$EC,$FF,$06,$AC,$E7,$F2,$EB,$FB,$92,$FA,$9F,$AB,$7C,$46,$5E,$2C
	.dc.b	$7F,$B7,$CE,$AB,$D3,$F2,$3F,$2C,$F9,$9F,$53,$6B,$E5,$45,$C3,$E7
	.dc.b	$4D,$AF,$83,$F3,$3F,$4F,$E2,$3E,$A3,$70,$FA,$4D,$BF,$AC,$DA,$FF
	.dc.b	$72,$7D,$4F,$F7,$16,$DF,$C6,$DB,$F8,$D5,$FA,$EB,$58,$7F,$AD,$5B
	.dc.b	$1F,$9C,$6E,$FF,$EE,$EF,$27,$FC,$21,$1A,$93,$C3,$C5,$7D,$4B,$E6
	.dc.b	$37,$9F,$F8,$6E,$31,$97,$E2,$FE,$A7,$F2,$1F,$53,$F9,$16,$C6,$4F
	.dc.b	$C8,$46,$5A,$7E,$4F,$B3,$F3,$1E,$CF,$E6,$5B,$19,$3F,$31,$19,$69
	.dc.b	$FB,$8E,$03,$FE,$D5,$FA,$8F,$8B,$18,$DB,$F7,$B1,$89,$D7,$89,$FE
	.dc.b	$31,$8C,$4E,$B8,$38,$7C,$17,$C1,$8E,$3F,$1F,$8E,$0E,$D5,$67,$38
	.dc.b	$FF,$9D,$A7,$C5,$59,$96,$08,$C3,$F6,$B0,$ED,$3A,$9C,$92,$FA,$47
	.dc.b	$95,$AE,$DB,$5F,$E3,$E6,$FB,$FD,$BE,$A7,$BB,$23,$8F,$D3,$FE,$D7
	.dc.b	$93,$71,$8E,$FF,$F1,$9E,$3F,$E9,$AD,$3F,$E0,$05,$EE,$30,$7F,$28
	.dc.b	$71,$A7,$AB,$3F,$14,$E8,$4E,$AB,$9D,$F6,$BF,$83,$39,$AF,$AF,$97
	.dc.b	$C9,$8C,$47,$2A,$E1,$7C,$B2,$F4,$CF,$7A,$D9,$2B,$71,$D3,$2F,$75
	.dc.b	$48,$E1,$3A,$7E,$BE,$1E,$18,$C5,$F3,$AE,$F8,$FA,$27,$24,$8C,$E3
	.dc.b	$E5,$C3,$AC,$77,$AF,$8D,$87,$58,$C5,$6E,$73,$8E,$2F,$E7,$61,$D5
	.dc.b	$F9,$99,$7C,$EA,$F1,$61,$B9,$F8,$CF,$89,$E2,$E7,$2F,$97,$5C,$71
	.dc.b	$F0,$12,$B6,$76,$CB,$CC,$FA,$89,$87,$99,$B3,$C5,$F0,$9C,$72,$F0
	.dc.b	$65,$E9,$AE,$07,$A2,$78,$9E,$07,$35,$CF,$1F,$A1,$3C,$CD,$9D,$A7
	.dc.b	$91,$C1,$E5,$3C,$6B,$EE,$63,$EF,$9C,$90,$E5,$3F,$C2,$71,$F7,$DF
	.dc.b	$48,$F6,$57,$1A,$D9,$FA,$1F,$A8,$7C,$0F,$F8,$2A,$21,$FA,$08,$CB
	.dc.b	$83,$E2,$9A,$C9,$1A,$8C,$3C,$30,$FD,$2B,$70,$FD,$9B,$67,$FC,$33
	.dc.b	$F5,$0F,$A9,$F1,$5F,$0C,$7F,$7F,$7B,$BF,$24,$7F,$0A,$BE,$11,$A8
	.dc.b	$8F,$EA,$E7,$84,$EA,$31,$3A,$AE,$D1,$FB,$F9,$D4,$60,$98,$7A,$2B
	.dc.b	$C1,$87,$9A,$31,$7D,$09,$5F,$C9,$2F,$9C,$EA,$39,$CF,$D1,$BE,$6F
	.dc.b	$8E,$74,$47,$E2,$DF,$77,$81,$DA,$34,$AF,$BB,$AB,$EE,$FC,$83,$DD
	.dc.b	$5C,$DC,$46,$B1,$FB,$78,$79,$27,$B1,$CF,$0E,$F8,$76,$8E,$31,$E2
	.dc.b	$AE,$31,$D8,$F6,$1C,$48,$79,$F0,$F1,$47,$0C,$3C,$58,$FD,$89,$F3
	.dc.b	$61,$DE,$B5,$87,$8A,$72,$72,$AF,$C3,$4E,$B0,$F9,$27,$E1,$FE,$13
	.dc.b	$9F,$CF,$48,$C1,$0F,$C1,$47,$80,$F8,$DF,$53,$FE,$E9,$8F,$EB,$2D
	.dc.b	$76,$71,$CF,$FB,$1C,$65,$BF,$0B,$CE,$75,$9F,$EB,$EB,$4F,$EC,$7D
	.dc.b	$FF,$F2,$6C,$FF,$81,$78,$B1,$F5,$3F,$B5,$E9,$FF,$02,$4A,$93,$9A
	.dc.b	$E5,$E4,$5A,$FF,$42,$D7,$47,$FC,$94,$F0,$7B,$1A,$AE,$0F,$59,$B2
	.dc.b	$F9,$3B,$F4,$52,$B8,$9E,$38,$D9,$97,$BC,$61,$DF,$D4,$BE,$A7,$AF
	.dc.b	$A7,$27,$31,$E2,$9D,$47,$6C,$3F,$D3,$6D,$AD,$54,$8E,$24,$3D,$FB
	.dc.b	$FB,$3E,$A5,$D5,$7E,$0E,$57,$AA,$72,$D9,$CB,$FE,$71,$84,$F3,$9E
	.dc.b	$9F,$27,$2B,$93,$DF,$E2,$B1,$1C,$62,$F1,$7D,$AC,$15,$F4,$3E,$A5
	.dc.b	$E7,$79,$FD,$39,$8F,$CE,$9C,$B6,$72,$F9,$67,$94,$48,$D1,$0F,$57
	.dc.b	$7F,$DE,$FA,$97,$D4,$D2,$4F,$DC,$87,$CD,$62,$3F,$D2,$51,$C1,$D4
	.dc.b	$F3,$9E,$9F,$3B,$D7,$E9,$65,$73,$8C,$BD,$63,$2F,$48,$FC,$8F,$8D
	.dc.b	$FE,$4C,$DF,$F3,$FF,$50,$FA,$9F,$CA,$B6,$FF,$23,$5B,$EE,$FE,$B4
	.dc.b	$D8,$FF,$72,$CE,$DF,$F5,$53,$6F,$F5,$73,$B9,$FF,$60,$77,$F8,$88
	.dc.b	$FE,$7A,$BF,$55,$5E,$4E,$DF,$19,$27,$5D,$B5,$EF,$16,$7D,$73,$EA
	.dc.b	$E9,$F5,$FC,$9C,$BE,$D1,$97,$C1,$19,$7D,$31,$F1,$BA,$BF,$6F,$EC
	.dc.b	$BC,$5A,$66,$BE,$2C,$E1,$FD,$BC,$ED,$C9,$0F,$F6,$76,$BF,$9E,$71
	.dc.b	$97,$F1,$09,$FD,$2C,$7F,$B4,$FE,$99,$C5,$FC,$23,$EA,$5F,$53,$4F
	.dc.b	$E1,$73,$AA,$FE,$0E,$BD,$4C,$FD,$4C,$9C,$C7,$F1,$2F,$A8,$FD,$2C
	.dc.b	$60,$87,$FB,$8D,$CB,$7E,$3B,$F8,$CF,$D4,$BE,$26,$9C,$E3,$F8,$EF
	.dc.b	$D4,$FF,$39,$DF,$F9,$5E,$09,$D6,$1F,$C7,$2B,$5D,$3F,$E2,$56,$C0
	.dc.b	$38,$6F,$27,$F2,$AF,$A9,$7C,$A7,$F2,$CF,$6B,$F5,$75,$FC,$A2,$7F
	.dc.b	$D5,$8D,$F6,$B5,$E0,$48,$C1,$06,$BF,$D0,$AF,$9C,$63,$3F,$F1,$D4
	.dc.b	$62,$75,$5C,$EB,$47,$A7,$0F,$F1,$C4,$FF,$4C,$F8,$3F,$3C,$FF,$3C
	.dc.b	$FA,$47,$C1,$7F,$44,$6C,$7F,$40,$9D,$BD,$03,$8A,$FE,$AF,$3C,$27
	.dc.b	$4D,$8F,$06,$9F,$D4,$2B,$86,$7F,$E5,$36,$DF,$CF,$6A,$B7,$C5,$7F
	.dc.b	$58,$F1,$1E,$FF,$F5,$AF,$A9,$FC,$15,$FF,$56,$6B,$E7,$9D,$BF,$D6
	.dc.b	$9A,$FA,$E3,$79,$FF,$70,$6E,$5F,$E6,$7A,$DC,$FF,$74,$6C,$66,$FC
	.dc.b	$87,$8F,$E3,$FF,$B6,$FD,$4B,$E8,$BC,$DF,$53,$33,$FE,$D7,$39,$6C
	.dc.b	$E5,$F2,$4F,$28,$91,$A2,$1E,$9F,$1F,$FB,$DF,$D4,$FE,$0F,$2F,$F3
	.dc.b	$F6,$BF,$E0,$9A,$F8,$9E,$C6,$B9,$64,$60,$87,$DD,$BE,$4F,$E8,$5A
	.dc.b	$9A,$62,$9F,$43,$E2,$F8,$7F,$A9,$ED,$4B,$FF,$04,$DA,$75,$1F,$5E
	.dc.b	$74,$F8,$63,$FC,$2B,$F5,$3F,$81,$8F,$F0,$8A,$EA,$C9,$AE,$69,$3A
	.dc.b	$6B,$F9,$7F,$A9,$7B,$5F,$AC,$87,$D4,$BE,$A7,$FD,$49,$5A,$E9,$FE
	.dc.b	$01,$39,$AF,$E0,$59,$78,$DE,$37,$FE,$AB,$8C,$4E,$E5,$7C,$DD,$9F
	.dc.b	$C2,$DC,$46,$E7,$8D,$73,$C7,$FD,$6D,$3C,$EB,$73,$E6,$C7,$D5,$3C
	.dc.b	$6B,$73,$C7,$0F,$4B,$BC,$6C,$78,$6F,$8C,$EE,$3C,$A3,$1F,$1B,$BA
	.dc.b	$E2,$FB,$2B,$8B,$F1,$2F,$C2,$D3,$78,$FA,$1E,$47,$6B,$EA,$F7,$9E
	.dc.b	$2F,$5B,$E3,$7D,$A3,$47,$8A,$B8,$DF,$58,$E0,$7F,$B6,$5F,$A0,$FF
	.dc.b	$B7,$9A,$FC,$33,$5E,$DB,$E5,$3F,$DC,$11,$F6,$23,$0E,$A3,$33,$FE
	.dc.b	$3A,$77,$1F,$D1,$7E,$12,$A6,$91,$DB,$F2,$BE,$45,$F4,$EE,$7F,$9B
	.dc.b	$5C,$CF,$A7,$3A,$8C,$C7,$53,$33,$C5,$FB,$27,$1D,$FF,$25,$5C,$63
	.dc.b	$27,$18,$CC,$F1,$7F,$31,$1C,$77,$FE,$C0,$AB,$94,$EA,$39,$4E,$A7
	.dc.b	$2F,$08,$E4,$7F,$69,$8C,$B9,$C3,$E1,$5F,$86,$B8,$4F,$43,$84,$65
	.dc.b	$E0,$7C,$0C,$3E,$1D,$72,$6C,$64,$F7,$C6,$5C,$6D,$FE,$AC,$C3,$E2
	.dc.b	$3F,$EA,$E8,$F3,$9E,$FB,$F9,$91,$B6,$DF,$BD,$B3,$E4,$C7,$E5,$CF
	.dc.b	$FA,$9A,$FF,$A0,$37,$A7,$7F,$1A,$FA,$9F,$F4,$6C,$E4,$98,$7F,$78
	.dc.b	$C9,$F5,$2E,$FE,$F3,$F4,$F1,$F0,$D6,$A7,$39,$6F,$0F,$14,$78,$3E
	.dc.b	$13,$FB,$EC,$60,$98,$7F,$7B,$AE,$0F,$F8,$23,$3F,$F5,$4F,$8D,$D3
	.dc.b	$09,$7F,$EC,$A9,$F5,$7C,$56,$2F,$86,$A8,$F3,$FC,$0F,$F3,$FB,$FF
	.dc.b	$64,$B6,$BF,$C2,$B0,$FA,$86,$1C,$3A,$E4,$F0,$7E,$7B,$B8,$CD,$7F
	.dc.b	$84,$EF,$A6,$DF,$A3,$AE,$9C,$5A,$8C,$63,$F8,$09,$FC,$EF,$D4,$BD
	.dc.b	$DF,$9A,$D8,$FF,$73,$9E,$5F,$3B,$33,$23,$FD,$C7,$3B,$CB,$D7,$6B
	.dc.b	$D4,$7F,$89,$E7,$58,$78,$B0,$CB,$70,$FF,$76,$9F,$EC,$62,$F5,$7C
	.dc.b	$0C,$7F,$B6,$C6,$09,$87,$F8,$CD,$E0,$F8,$BC,$1F,$C3,$C6,$2F,$F8
	.dc.b	$34,$FF,$0F,$BF,$97,$F5,$3F,$E5,$8A,$E7,$5F,$26,$31,$5F,$E3,$FF
	.dc.b	$9C,$FC,$2C,$F2,$6D,$7F,$99,$4F,$B5,$5F,$96,$8C,$D7,$F9,$A5,$A9
	.dc.b	$CC,$F0,$3F,$CD,$D5,$D7,$8F,$91,$F0,$7C,$6B,$3F,$53,$34,$7E,$93
	.dc.b	$EA,$1F,$53,$FA,$6F,$A8,$FF,$03,$5F,$43,$F8,$DC,$E9,$C6,$7F,$E7
	.dc.b	$D9,$DB,$FE,$5D,$6E,$5F,$21,$B3,$CA,$FB,$78,$BD,$FE,$06,$4F,$4C
	.dc.b	$17,$C1,$ED,$3E,$5F,$4A,$56,$FE,$A7,$C9,$5E,$19,$FD,$FF,$43,$CD
	.dc.b	$A7,$09,$FF,$39,$46,$6F,$C1,$F5,$3F,$E6,$66,$FA,$7C,$BD,$AF,$E1
	.dc.b	$33,$9B,$FF,$3D,$7D,$37,$FA,$8E,$FD,$0D,$8F,$F4,$4F,$D4,$3E,$A5
	.dc.b	$F5,$3F,$C4,$AB,$5C,$7F,$8C,$6E,$1F,$D2,$1B,$83,$0E,$57,$C7,$DD
	.dc.b	$97,$F4,$BC,$67,$D4,$FF,$9F,$23,$C7,$3F,$CB,$A7,$3E,$CF,$F3,$0E
	.dc.b	$CF,$F5,$A7,$D4,$DC,$FF,$1F,$18,$C3,$EE,$BF,$EB,$9E,$DD,$56,$DB
	.dc.b	$3A,$AE,$23,$EA,$7A,$1D,$78,$AC,$EE,$2C,$FF,$AB,$B0,$FF,$63,$56
	.dc.b	$38,$3F,$5F,$F4,$EF,$5F,$F9,$CB,$71,$FE,$55,$F5,$3F,$EE,$57,$F1
	.dc.b	$11,$9C,$7F,$D5,$CD,$BE,$77,$FD,$2A,$F8,$56,$A3,$A9,$9F,$2A,$E1
	.dc.b	$D9,$F9,$1E,$A7,$F8,$3E,$3E,$2B,$F8,$3C,$3F,$C1,$59,$FF,$88,$E7
	.dc.b	$7D,$1C,$1A,$70,$2E,$3F,$06,$DF,$53,$D8,$EF,$F1,$B6,$BB,$C5,$AE
	.dc.b	$13,$71,$F8,$FF,$37,$F1,$8F,$C0,$7C,$13,$E5,$F1,$8F,$C6,$9B,$87
	.dc.b	$FB,$7B,$9B,$FD,$A7,$58,$C3,$9D,$FE,$F9,$7D,$4F,$C8,$73,$7F,$DB
	.dc.b	$E3,$13,$FC,$5F,$47,$C8,$7C,$83,$E4,$FD,$4D,$9F,$93,$16,$FE,$14
	.dc.b	$DC,$3E,$5D,$BE,$67,$E8,$FA,$FF,$49,$B8,$FD,$26,$E1,$F2,$DB,$7F
	.dc.b	$CE,$3E,$A7,$FB,$CC,$FF,$BC,$5B,$1F,$E4,$1D,$7F,$DE,$7E,$57,$F8
	.dc.b	$1A,$73,$87,$F8,$E9,$B6,$0E,$2B,$8C,$FF,$19,$8D,$CE,$B1,$FF,$0C
	.dc.b	$FD,$4B,$E6,$3B,$46,$1F,$C6,$FD,$4F,$E3,$BE,$A6,$E3,$F8,$E6,$E1
	.dc.b	$F8,$76,$E9,$F1,$9F,$CA,$FC,$97,$E5,$BA,$AD,$FE,$59,$B7,$F8,$F6
	.dc.b	$DF,$CE,$7F,$C4,$BF,$51,$FC,$BE,$4D,$3F,$86,$29,$FE,$C7,$19,$DF
	.dc.b	$FC,$D5,$F0,$7E,$DA,$F8,$D2,$F9,$0C,$EB,$FE,$64,$8F,$E8,$F5,$BA
	.dc.b	$EF,$5A,$78,$B5,$CE,$5F,$E8,$7B,$E5,$86,$70,$ED,$39,$AE,$2D,$78
	.dc.b	$9F,$E8,$DA,$E3,$19,$9E,$33,$93,$D3,$F0,$3D,$4E,$DF,$53,$6B,$84
	.dc.b	$79,$7E,$99,$DB,$FE,$32,$BE,$47,$CA,$9D,$46,$74,$F8,$15,$C3,$D0
	.dc.b	$F3,$36,$F8,$E4,$C3,$8D,$6B,$0E,$59,$FF,$A3,$B0,$E1,$8F,$CB,$C3
	.dc.b	$86,$5B,$C2,$65,$E4,$BD,$E1,$F4,$F2,$EB,$39,$CB,$AB,$F9,$FA,$DB
	.dc.b	$F1,$30,$E7,$97,$AA,$B7,$2A,$DB,$F0,$6B,$78,$79,$EB,$75,$AC,$BC
	.dc.b	$F7,$E1,$CB,$6F,$86,$BA,$CE,$B1,$F0,$56,$AB,$C1,$5A,$CB,$D7,$5A
	.dc.b	$9E,$87,$CC,$8C,$E3,$E3,$BE,$58,$75,$8C,$57,$0B,$E4,$7C,$DB,$E3
	.dc.b	$1E,$5B,$E3,$18,$8F,$1C,$63,$2F,$36,$5F,$10,$F8,$15,$B7,$19,$7A
	.dc.b	$67,$6F,$C6,$31,$7E,$3C,$7C,$B8,$F9,$1F,$B5,$5C,$31,$F0,$56,$CF
	.dc.b	$31,$4F,$A8,$5C,$3C,$31,$F9,$C9,$E5,$3B,$CB,$E2,$3C,$B0,$FC,$CF
	.dc.b	$D3,$BE,$0B,$FE,$00,$E2,$8C,$57,$FB,$52,$3F,$38,$FF,$B7,$5D,$4F
	.dc.b	$09,$DD,$74,$8F,$D1,$C6,$32,$FF,$75,$FD,$4B,$EA,$7F,$DD,$B7,$BC
	.dc.b	$7F,$39,$7B,$7F,$30,$D5,$5C,$5E,$B8,$47,$8B,$FC,$1E,$35,$18,$25
	.dc.b	$77,$3F,$8C,$46,$9C,$61,$FB,$3B,$F0,$63,$EE,$BF,$E0,$77,$BA,$D6
	.dc.b	$1E,$1C,$38,$D7,$F4,$DA,$D3,$EA,$9D,$57,$08,$E7,$8F,$C2,$8C,$5F
	.dc.b	$29,$FD,$96,$3F,$07,$0E,$71,$8C,$39,$BD,$4E,$66,$0D,$1F,$02,$FE
	.dc.b	$8E,$1D,$E3,$86,$1D,$F1,$F7,$3F,$12,$75,$3C,$A3,$58,$F8,$E7,$35
	.dc.b	$C2,$FD,$24,$BF,$85,$3A,$AF,$96,$4C,$3E,$B4,$65,$F4,$5F,$3A,$E2
	.dc.b	$62,$B8,$3E,$E2,$61,$E4,$AE,$71,$0F,$80,$7B,$B0,$F8,$2F,$E1,$7E
	.dc.b	$17,$DB,$93,$E7,$A9,$88,$93,$B2,$1F,$D1,$3E,$A7,$FD,$BC,$FF,$55
	.dc.b	$6B,$F8,$68,$E2,$EF,$E1,$BE,$5C,$EB,$2F,$C8,$D6,$9F,$ED,$51,$9C
	.dc.b	$7F,$C2,$D7,$FE,$BE,$F1,$6F,$3F,$FB,$6B,$E0,$1F,$53,$8E,$3D,$C6
	.dc.b	$2F,$FD,$64,$AF,$51,$FE,$21,$8C,$3A,$AC,$BF,$EC,$C6,$C6,$A3,$FC
	.dc.b	$11,$5D,$BD,$BC,$FD,$0C,$9E,$F1,$C0,$C3,$C0,$F0,$46,$5F,$F6,$37
	.dc.b	$D4,$BE,$A7,$BF,$A1,$27,$31,$E4,$9D,$47,$B2,$74,$F5,$11,$A7,$0D
	.dc.b	$D3,$FE,$00,$B7,$1F,$9A,$D9,$F8,$9C,$3F,$0B,$EA,$7F,$BC,$47,$FB
	.dc.b	$6D,$7C,$16,$4D,$76,$C9,$D4,$65,$E1,$1C,$FE,$52,$47,$37,$B7,$C4
	.dc.b	$E2,$3D,$56,$9C,$66,$CF,$F0,$72,$BF,$AF,$F5,$3F,$E6,$DC,$3C,$DE
	.dc.b	$8F,$A4,$CF,$9B,$23,$24,$3E,$44,$F1,$6C,$6D,$A7,$C1,$EC,$FC,$2F
	.dc.b	$D4,$BE,$A6,$92,$7F,$E0,$04,$3E,$6F,$11,$F5,$63,$83,$A9,$E7,$3D
	.dc.b	$3E,$7B,$AF,$D3,$CA,$E7,$19,$7A,$C6,$5E,$91,$F9,$8F,$1F,$FC,$A5
	.dc.b	$C1,$F1,$E7,$FC,$FD,$F5,$0F,$A9,$FC,$EB,$63,$FD,$05,$3B,$7F,$C8
	.dc.b	$3D,$FC,$CD,$D4,$8D,$F1,$FE,$D3,$0E,$ED,$AE,$13,$A3,$FD,$1D,$3A
	.dc.b	$7C,$7D,$FE,$2E,$D9,$1B,$19,$23,$CF,$DB,$A9,$FC,$F4,$E5,$E0,$D7
	.dc.b	$31,$EE,$EA,$F7,$FC,$9C,$9D,$C6,$5E,$31,$97,$AC,$7C,$BE,$BF,$E0
	.dc.b	$7E,$C7,$F2,$59,$F3,$35,$FD,$86,$7F,$A4,$7D,$4D,$7F,$69,$1B,$DF
	.dc.b	$E7,$37,$6F,$95,$7D,$EB,$8B,$5E,$3D,$9F,$C5,$3E,$9D,$C7,$FC,$5B
	.dc.b	$EA,$5F,$53,$4D,$46,$FD,$3E,$1F,$53,$31,$F9,$D8,$27,$F8,$FF,$D4
	.dc.b	$7E,$AE,$73,$1C,$5F,$A9,$19,$7F,$87,$2F,$8A,$FE,$4F,$F5,$2F,$87
	.dc.b	$A7,$48,$FE,$57,$F5,$3F,$D1,$38,$BF,$1D,$7C,$70,$EB,$5A,$E4,$91
	.dc.b	$F8,$3B,$E1,$C5,$FD,$09,$B1,$FE,$D9,$FA,$97,$CA,$7F,$38,$F6,$7F
	.dc.b	$61,$3F,$8B,$9F,$F5,$C3,$7D,$AD,$78,$32,$30,$43,$F9,$55,$72,$6C
	.dc.b	$6D,$A7,$5F,$13,$FA,$3F,$C2,$F8,$A3,$F3,$5F,$4A,$F8,$0F,$E9,$6D
	.dc.b	$8F,$CF,$4E,$DF,$E7,$0D,$E1,$FD,$94,$EE,$7F,$9E,$36,$3F,$6D,$3B
	.dc.b	$DB,$FB,$2B,$6B,$FD,$F8,$D4,$DA,$7C,$47,$F5,$EF,$19,$F1,$FF,$D8
	.dc.b	$3E,$9F,$ED,$61,$FB,$36,$BF,$D3,$67,$6F,$ED,$DA,$FF,$51,$8D,$E5
	.dc.b	$FC,$2D,$B9,$7F,$7A,$AD,$CF,$F0,$F6,$C6,$6F,$C8,$78,$FE,$3F,$FB
	.dc.b	$9F,$D4,$BE,$8B,$CD,$F5,$33,$3F,$E1,$13,$96,$CE,$5F,$24,$F2,$89
	.dc.b	$1A,$21,$E9,$F2,$7F,$C0,$3F,$53,$F8,$1D,$3D,$5E,$CF,$C5,$33,$C9
	.dc.b	$2F,$E3,$46,$5F,$E3,$55,$C7,$5F,$F0,$CB,$60,$1C,$36,$BE,$97,$D0
	.dc.b	$F8,$BE,$1F,$EA,$7B,$52,$BA,$C6,$5F,$AD,$19,$7F,$C0,$56,$9F,$F0
	.dc.b	$DF,$D4,$B9,$6F,$4C,$BF,$E6,$13,$96,$CF,$02,$3F,$CD,$7E,$A7,$FB
	.dc.b	$A6,$DE,$AF,$A8,$7D,$4F,$F8,$DA,$73,$49,$CB,$F0,$30,$EC,$E6,$BF
	.dc.b	$C7,$B1,$89,$DC,$AB,$9C,$61,$B5,$C3,$0F,$F2,$2C,$F0,$BE,$53,$E3
	.dc.b	$F7,$58,$F9,$E3,$0D,$C7,$C9,$87,$29,$E9,$19,$3C,$55,$C9,$F4,$1F
	.dc.b	$6B,$1F,$1B,$C8,$E1,$87,$89,$D4,$F1,$C7,$D4,$DB,$EF,$3D,$05,$71
	.dc.b	$7C,$38,$F8,$F0,$EF,$1A,$C7,$D0,$F8,$4E,$51,$A3,$75,$CC,$E8,$E7
	.dc.b	$0F,$64,$FF,$9B,$8E,$E5,$29,$98,$FB,$AD,$F7,$1F,$23,$B8,$FB,$11
	.dc.b	$87,$94,$FF,$8D,$E7,$4D,$8E,$03,$E0,$EB,$4C,$B8,$23,$F1,$FE,$45
	.dc.b	$F4,$FF,$E1,$FD,$BF,$13,$3F,$AE,$AF,$B5,$5A,$E5,$FB,$93,$93,$95
	.dc.b	$F1,$DF,$E0,$5F,$08,$C4,$FC,$58,$C5,$72,$8F,$36,$1F,$98,$C3,$B4
	.dc.b	$7D,$B3,$B4,$66,$F9,$C7,$FC,$00,$3C,$35,$CE,$FA,$C7,$13,$AE,$1F
	.dc.b	$EA,$2C,$3A,$E1,$EE,$9C,$9F,$C3,$23,$33,$EA,$8E,$A7,$3A,$EB,$5C
	.dc.b	$5B,$3D,$70,$FF,$24,$6F,$FE,$43,$C3,$D6,$FF,$91,$70,$F5,$D6,$F1
	.dc.b	$FF,$1C,$56,$9C,$4F,$FA,$4C,$D9,$E5,$BE,$B1,$EB,$E8,$B3,$E9,$6F
	.dc.b	$C6,$BE,$A3,$FD,$1F,$5D,$A7,$74,$C3,$FA,$24,$FF,$93,$F2,$7D,$4B
	.dc.b	$BB,$75,$AB,$FF,$59,$56,$A7,$24,$BF,$F2,$C6,$4F,$57,$FB,$D4,$7B
	.dc.b	$A7,$57,$FE,$AF,$C3,$B4,$F0,$AD,$CE,$6B,$78,$7F,$9C,$F0,$F6,$7C
	.dc.b	$4E,$30,$FE,$87,$18,$7F,$CC,$5E,$3F,$F3,$96,$F4,$FC,$65,$F4,$EA
	.dc.b	$F8,$75,$BC,$BF,$D8,$4D,$C7,$FC,$F8,$DB,$FE,$A2,$72,$AE,$38,$48
	.dc.b	$FF,$3C,$1D,$B8,$31,$87,$FB,$66,$30,$7F,$9D,$FE,$A5,$EE,$7D,$46
	.dc.b	$C7,$2C,$DE,$7E,$6A,$4F,$17,$D9,$3A,$8C,$12,$FC,$11,$D1,$FF,$6F
	.dc.b	$5D,$D3,$FD,$3D,$5F,$12,$FF,$0A,$FF,$B1,$4F,$F4,$61,$6B,$9F,$81
	.dc.b	$AE,$A9,$3A,$7C,$71,$89,$FF,$76,$4F,$1B,$EF,$F5,$3F,$E5,$08,$E0
	.dc.b	$7C,$87,$13,$FE,$F3,$F9,$BF,$C2,$B7,$8B,$DF,$3F,$9A,$9C,$DF,$FA
	.dc.b	$C5,$B7,$FE,$06,$B4,$61,$FF,$71,$68,$BE,$58,$4B,$FA,$1E,$25,$BF
	.dc.b	$92,$FE,$A3,$EA,$5F,$53,$FA,$AF,$A9,$78,$99,$8C,$33,$0F,$F5,$A4
	.dc.b	$E9,$FF,$5B,$57,$27,$33,$CF,$DF,$91,$83,$8B,$DF,$D4,$91,$FE,$BF
	.dc.b	$FA,$9E,$59,$FF,$76,$8F,$81,$8F,$CA,$E0,$FF,$6A,$74,$7F,$B0,$E7
	.dc.b	$FD,$8C,$DA,$E0,$7F,$8B,$2B,$84,$FC,$99,$D5,$7F,$B7,$7E,$A7,$FC
	.dc.b	$C3,$3F,$EC,$F6,$DF,$23,$F5,$9E,$D7,$F0,$E8,$CD,$7F,$B9,$7E,$A3
	.dc.b	$FD,$41,$7F,$ED,$76,$DF,$8B,$EA,$1F,$52,$FA,$9F,$E2,$F5,$AC,$39
	.dc.b	$7B,$B8,$3F,$90,$DF,$0A,$C6,$3D,$67,$F9,$2F,$86,$FA,$9F,$F3,$BC
	.dc.b	$E4,$FF,$77,$46,$7D,$6F,$E5,$7D,$7F,$E4,$EF,$A9,$BB,$7E,$26,$31
	.dc.b	$7F,$EF,$57,$FC,$A7,$D5,$AA,$DB,$67,$55,$C4,$7D,$4F,$43,$AF,$15
	.dc.b	$9D,$C5,$9F,$AB,$87,$F9,$82,$B1,$BA,$FF,$3D,$F5,$2F,$57,$F9,$EB
	.dc.b	$63,$F9,$9F,$D4,$FF,$B5,$DF,$F3,$4D,$71,$C7,$FC,$D7,$5B,$7E,$D3
	.dc.b	$63,$50,$8C,$4F,$F3,$0A,$E9,$3C,$2F,$C2,$7F,$52,$8F,$E5,$11,$DA
	.dc.b	$31,$E3,$FC,$EF,$4B,$D5,$7F,$2D,$FC,$3B,$FE,$7E,$CF,$FC,$4A,$EF
	.dc.b	$DD,$E0,$D3,$81,$70,$F8,$F4,$FA,$9E,$C7,$7F,$8D,$B5,$DE,$2D,$70
	.dc.b	$9B,$AF,$C0,$D5,$F2,$1F,$30,$FC,$5F,$C0,$BE,$5F,$19,$7C,$C3,$E9
	.dc.b	$CE,$31,$FD,$6F,$37,$FB,$2A,$B1,$8F,$5A,$FE,$F2,$BE,$A7,$E5,$B9
	.dc.b	$C3,$FB,$A4,$61,$F8,$9A,$BE,$43,$E5,$1F,$2F,$EA,$6C,$FC,$B8,$B9
	.dc.b	$FC,$49,$B8,$7E,$07,$87,$FC,$A3,$F3,$1F,$4F,$E1,$FE,$A3,$71,$FA
	.dc.b	$8D,$C3,$E6,$B6,$FF,$0B,$F5,$3F,$DE,$1B,$7F,$1B,$C9,$FF,$07,$CE
	.dc.b	$70,$FE,$68,$DC,$3F,$D7,$6D,$E0,$FF,$77,$D7,$0B,$97,$FE,$0C,$FA
	.dc.b	$97,$CB,$E3,$1F,$E3,$3E,$3A,$73,$5F,$8B,$FA,$9F,$C7,$7D,$4F,$E3
	.dc.b	$DB,$19,$3F,$21,$19,$69,$F9,$3F,$92,$FC,$B7,$67,$E5,$DB,$19,$3F
	.dc.b	$31,$19,$69,$F8,$3C,$7F,$DA,$FF,$51,$FD,$05,$FE,$4C,$DE,$0F,$D6
	.dc.b	$BF,$C5,$23,$3B,$7F,$8D,$A3,$87,$C1,$7C,$09,$E7,$F1,$FF,$B9,$8C
	.dc.b	$42,$77,$BF,$F6,$15,$5B,$8C,$36,$79,$E3,$FE,$7F,$9E,$18,$66,$7A
	.dc.b	$55,$BE,$77,$B9,$E2,$7F,$A4,$6F,$8C,$66,$38,$B6,$7F,$C8,$8D,$F7
	.dc.b	$FA,$7D,$4F,$56,$47,$1F,$A8,$78,$1F,$E3,$7C,$3B,$1F,$32,$BC,$D3
	.dc.b	$AE,$3E,$15,$BD,$7E,$2C,$E6,$F8,$9F,$1E,$F8,$D6,$F3,$FF,$04,$56
	.dc.b	$EF,$8C,$E6,$FC,$18,$F6,$C6,$47,$4B,$E1,$87,$D1,$C7,$AD,$6B,$1E
	.dc.b	$6F,$C1,$C3,$15,$E0,$C7,$B6,$2C,$79,$BE,$03,$B4,$6A,$33,$1B,$8E
	.dc.b	$39,$FC,$E8,$DE,$1D,$91,$B7,$C4,$E8,$98,$F9,$9F,$8F,$3C,$E3,$04
	.dc.b	$C3,$D1,$5C,$A3,$15,$DB,$0F,$29,$E6,$BE,$57,$C2,$7B,$B6,$BF,$37
	.dc.b	$3A,$9E,$58,$7B,$C9,$87,$7B,$E0,$49,$DC,$65,$A7,$62,$DF,$6C,$BE
	.dc.b	$AC,$64,$F8,$87,$C0,$73,$1E,$C8,$CB,$C7,$2F,$99,$18,$BF,$13,$F8
	.dc.b	$D8,$CE,$5F,$0B,$1F,$AB,$87,$4B,$DC,$FC,$58,$C5,$F9,$A7,$9E,$5F
	.dc.b	$0E,$3B,$1F,$37,$EA,$1F,$03,$FE,$5F,$8F,$CC,$C6,$1C,$1E,$C3,$58
	.dc.b	$23,$51,$87,$86,$7F,$A8,$6E,$1D,$2F,$F7,$9F,$50,$FA,$9F,$F3,$75
	.dc.b	$EE,$7F,$71,$7C,$30,$F3,$47,$0A,$DE,$1C,$E7,$F5,$15,$B7,$F4,$0D
	.dc.b	$C3,$B4,$74,$B8,$F0,$7A,$38,$C3,$C4,$E4,$FD,$FC,$F2,$71,$1A,$25
	.dc.b	$FF,$A1,$E7,$51,$89,$F4,$9C,$27,$58,$7A,$A3,$33,$B7,$D3,$7E,$47
	.dc.b	$81,$AA,$F8,$13,$FC,$E2,$FB,$C6,$09,$3B,$3F,$6F,$1A,$70,$76,$3A
	.dc.b	$87,$4E,$67,$C9,$8F,$DF,$BF,$24,$63,$1F,$A1,$39,$AF,$8F,$8F,$B5
	.dc.b	$E5,$8F,$91,$A7,$92,$35,$3C,$6F,$A6,$1D,$AF,$84,$E5,$E5,$8F,$9A
	.dc.b	$31,$87,$B1,$AF,$AF,$0F,$14,$EF,$1F,$39,$F8,$58,$C9,$E2,$73,$87
	.dc.b	$86,$BE,$4F,$C2,$72,$FA,$99,$1F,$9A,$8E,$A4,$70,$7F,$59,$FA,$9F
	.dc.b	$F7,$8B,$7E,$17,$4F,$E3,$27,$E3,$4E,$A3,$FA,$6B,$75,$FB,$D3,$C1
	.dc.b	$FE,$EF,$19,$D7,$FD,$AD,$EF,$FF,$93,$AF,$FC,$10,$17,$D4,$FF,$B6
	.dc.b	$FA,$3F,$CE,$71,$89,$DB,$C5,$FF,$59,$36,$37,$3C,$79,$3B,$B6,$BF
	.dc.b	$A1,$D6,$E7,$C7,$F2,$79,$9F,$17,$D0,$DA,$36,$64,$F8,$A3,$46,$1F
	.dc.b	$F6,$27,$D4,$3E,$A7,$97,$A3,$27,$DF,$39,$6C,$E5,$F5,$CF,$8E,$75
	.dc.b	$1D,$27,$04,$7E,$1F,$07,$A7,$EA,$5D,$37,$E0,$E5,$7C,$E9,$CB,$67
	.dc.b	$2F,$FB,$6E,$13,$CE,$7A,$7C,$9C,$AE,$4F,$7F,$8A,$C4,$71,$8B,$C3
	.dc.b	$FE,$AA,$C1,$5F,$37,$EA,$5E,$5F,$9F,$D2,$98,$FE,$0A,$72,$D9,$CB
	.dc.b	$E5,$9E,$51,$23,$44,$3D,$5D,$5F,$6F,$EA,$5F,$53,$49,$3F,$90,$87
	.dc.b	$CD,$62,$3F,$DF,$91,$C1,$D4,$F3,$9E,$9F,$3B,$D7,$E9,$65,$73,$8C
	.dc.b	$BD,$63,$2F,$48,$FC,$77,$8B,$FE,$5A,$DF,$FD,$9F,$F5,$0F,$A9,$FC
	.dc.b	$93,$6F,$F4,$B5,$BE,$CF,$F0,$DC,$7F,$91,$BB,$B7,$72,$37,$1D,$88
	.dc.b	$E0,$FD,$7D,$FE,$85,$FF,$12,$AE,$CF,$09,$23,$7D,$95,$65,$E8,$9C
	.dc.b	$3E,$A9,$F4,$F4,$FA,$BE,$4E,$5F,$82,$32,$F8,$A3,$2F,$4C,$FD,$B8
	.dc.b	$FF,$0D,$C3,$FD,$E7,$A7,$E5,$2B,$F8,$3F,$AE,$F1,$3F,$70,$D8,$F8
	.dc.b	$FD,$9F,$38,$CC,$C8,$C9,$0F,$01,$FE,$46,$AE,$CE,$0F,$F3,$9C,$F6
	.dc.b	$EC,$FE,$0D,$F4,$CE,$3F,$E1,$1F,$52,$FA,$9A,$6A,$37,$E9,$70,$FA
	.dc.b	$99,$97,$F2,$0C,$13,$FC,$47,$EA,$3F,$4F,$39,$8E,$2F,$F2,$68,$CB
	.dc.b	$E5,$5F,$17,$FC,$67,$EA,$5F,$0F,$4E,$91,$FC,$77,$EA,$7E,$67,$2F
	.dc.b	$E2,$AF,$8E,$1D,$6B,$5C,$F2,$3F,$9A,$DF,$07,$FC,$D6,$D5,$F5,$2F
	.dc.b	$96,$FE,$55,$EC,$FE,$B2,$7F,$9E,$CF,$FA,$D1,$BE,$CE,$BC,$19,$18
	.dc.b	$21,$F4,$6B,$93,$63,$6D,$3A,$F8,$9F,$CE,$7E,$1B,$C5,$1F,$D3,$3E
	.dc.b	$95,$F0,$1F,$CF,$5B,$1F,$D4,$27,$6F,$DF,$6F,$73,$87,$F5,$F5,$C2
	.dc.b	$BF,$08,$D9,$D4,$7F,$A3,$AB,$8D,$7E,$D9,$B1,$9D,$7C,$C7,$8F,$E2
	.dc.b	$BF,$A8,$78,$CF,$8F,$FE,$A5,$F5,$3F,$6A,$FF,$B2,$B5,$FC,$A4,$ED
	.dc.b	$FE,$D2,$D7,$F3,$11,$BC,$BC,$B0,$8C,$65,$FC,$2A,$B8,$56,$9B,$19
	.dc.b	$BF,$29,$E4,$F8,$FF,$EC,$DF,$52,$FA,$2F,$3F,$D4,$CC,$FF,$BD,$CE
	.dc.b	$5B,$39,$7C,$B3,$CA,$24,$68,$87,$AB,$CA,$FE,$E9,$F5,$3F,$82,$D3
	.dc.b	$D7,$AF,$F8,$1E,$BE,$2F,$B3,$AE,$59,$18,$21,$FB,$BB,$E5,$B7,$F2
	.dc.b	$A6,$C7,$FB,$19,$A9,$B6,$FA,$1F,$1F,$C3,$7D,$6F,$6E,$5F,$F8,$6A
	.dc.b	$D3,$A8,$FB,$33,$A7,$C5,$1F,$E0,$FF,$A9,$73,$DE,$A9,$5F,$E2,$49
	.dc.b	$CB,$67,$81,$1F,$F1,$4F,$D4,$BD,$BF,$E6,$B0,$FA,$97,$D4,$FF,$48
	.dc.b	$AD,$71,$7F,$B9,$27,$35,$FE,$E6,$C7,$BB,$C6,$FF,$A7,$C6,$27,$72
	.dc.b	$BE,$6E,$CF,$F7,$7B,$88,$DC,$F1,$AE,$78,$FF,$5A,$9E,$75,$B9,$F3
	.dc.b	$63,$EA,$9E,$35,$B9,$E3,$87,$A5,$DE,$36,$3C,$37,$C6,$77,$1E,$51
	.dc.b	$8F,$8D,$DD,$71,$7D,$95,$C5,$F8,$97,$E1,$69,$BC,$7D,$0F,$23,$B5
	.dc.b	$F5,$7B,$CF,$17,$AD,$F1,$BE,$D1,$A3,$C5,$5C,$6F,$AC,$70,$3F,$DB
	.dc.b	$AF,$D0,$7F,$DC,$4D,$7E,$19,$AF,$71,$F2,$5F,$EE,$48,$FB,$11,$87
	.dc.b	$51,$99,$FF,$1B,$BB,$80,$F8,$3C,$34,$D3,$82,$3F,$3F,$E3,$DF,$4F
	.dc.b	$FC,$9B,$6F,$C4,$CF,$4A,$FA,$15,$AE,$5F,$A5,$39,$39,$5F,$1D,$FF
	.dc.b	$DF,$B7,$C2,$31,$3F,$E2,$43,$C9,$F8,$4F,$E5,$B0,$E9,$08,$F0,$BA
	.dc.b	$BF,$F1,$5C,$EA,$73,$7D,$70,$F1,$DF,$BF,$1E,$98,$F8,$23,$FC,$67
	.dc.b	$5B,$7F,$1C,$7A,$67,$75,$CA,$75,$1D,$A7,$DB,$39,$69,$C8,$F0,$63
	.dc.b	$E6,$BF,$06,$3E,$CD,$FF,$9E,$B6,$FE,$2D,$F2,$9C,$D7,$F4,$28,$E5
	.dc.b	$3F,$83,$AF,$8B,$1F,$25,$AF,$C4,$EA,$E0,$FE,$0A,$3F,$A0,$4F,$19
	.dc.b	$CC,$A7,$2F,$D3,$31,$1F,$D9,$AE,$7D,$4B,$BB,$18,$77,$3F,$1F,$3A
	.dc.b	$8C,$DF,$85,$AF,$77,$3F,$1D,$FE,$A0,$F1,$73,$97,$5C,$3F,$D9,$AD
	.dc.b	$BF,$23,$5F,$81,$7E,$A8,$C4,$7E,$12,$3A,$FD,$53,$E2,$BA,$C6,$1F
	.dc.b	$1F,$91,$FD,$03,$0F,$F6,$DB,$6F,$FC,$9B,$0F,$A9,$67,$C3,$AE,$4F
	.dc.b	$ED,$6F,$85,$FE,$6D,$AF,$FA,$01,$AF,$ED,$1E,$DA,$FF,$84,$A7,$B7
	.dc.b	$07,$4A,$F0,$E3,$F8,$57,$F7,$7F,$54,$F7,$FE,$5B,$6F,$E8,$9F,$C0
	.dc.b	$0F,$47,$9F,$A9,$91,$FC,$2E,$78,$57,$DF,$6B,$FB,$36,$BF,$E9,$87
	.dc.b	$78,$F9,$F0,$B5,$9B,$FB,$EF,$C7,$7E,$FB,$FE,$CA,$3C,$B7,$E5,$EA
	.dc.b	$DD,$7F,$75,$9C,$CF,$F8,$E6,$78,$9F,$96,$8E,$51,$F9,$56,$BF,$C2
	.dc.b	$5A,$FF,$AB,$9D,$DF,$CD,$FA,$AF,$F2,$F4,$76,$32,$7C,$77,$33,$FE
	.dc.b	$89,$F9,$8F,$C8,$57,$F8,$C9,$B5,$F3,$70,$5F,$E7,$27,$37,$FC,$C1
	.dc.b	$FC,$9B,$78,$7F,$5F,$18,$83,$4E,$37,$F8,$FF,$1B,$A5,$7C,$67,$F5
	.dc.b	$3F,$4E,$F0,$A9,$FA,$CF,$A9,$7D,$56,$30,$F9,$75,$B7,$FA,$53,$72
	.dc.b	$FD,$F3,$5F,$E6,$4D,$7F,$DA,$EB,$E0,$3F,$83,$B6,$79,$C6,$0C,$FA
	.dc.b	$32,$3B,$B6,$BF,$86,$73,$FF,$B5,$E7,$FA,$E3,$63,$3E,$67,$F1,$48
	.dc.b	$C4,$7F,$45,$FA,$97,$A5,$FC,$61,$B7,$FD,$95,$B3,$FC,$65,$AF,$F4
	.dc.b	$DF,$A8,$FF,$57,$CF,$E3,$9B,$1F,$EC,$AF,$A8,$7D,$4B,$EA,$7F,$8B
	.dc.b	$D6,$B9,$FF,$CE,$6F,$0B,$E9,$ED,$FF,$2A,$8C,$47,$FB,$F9,$AF,$F7
	.dc.b	$27,$96,$3F,$CC,$FE,$A7,$FD,$10,$F8,$EB,$F9,$3F,$86,$9C,$DF,$FB
	.dc.b	$7B,$A3,$FB,$DF,$D4,$FF,$2C,$7F,$BE,$F4,$D9,$4D,$8C,$37,$1F,$D8
	.dc.b	$D7,$26,$D7,$E6,$23,$FC,$0F,$F5,$3C,$CE,$9F,$01,$6B,$71,$6B,$C8
	.dc.b	$7F,$B6,$23,$27,$EE,$BE,$9D,$EA,$7F,$40,$6C,$7F,$02,$E5,$7D,$4F
	.dc.b	$FB,$AA,$38,$56,$CF,$EA,$B5,$C2,$FF,$AA,$35,$FF,$04,$35,$FF,$0F
	.dc.b	$87,$87,$0F,$FA,$FB,$0E,$FE,$3F,$5C,$3C,$58,$FF,$32,$7F,$71,$E8
	.dc.b	$FF,$B2,$5F,$8E,$FE,$32,$FF,$D8,$FA,$FF,$8D,$27,$E9,$4E,$5E,$26
	.dc.b	$3A,$32,$D3,$9F,$D4,$F0,$38,$FC,$65,$8C,$93,$5F,$C0,$5A,$75,$0C
	.dc.b	$7F,$D8,$BA,$BE,$A7,$E7,$1F,$8A,$F8,$17,$CB,$E3,$2F,$A9,$3F,$30
	.dc.b	$8F,$F9,$36,$B9,$36,$3F,$BC,$DB,$EA,$7E,$B6,$3E,$1A,$F9,$F7,$C2
	.dc.b	$30,$79,$67,$75,$F2,$33,$7C,$8F,$C8,$3E,$4F,$D4,$DC,$7E,$7C,$5C
	.dc.b	$3E,$0C,$DB,$E5,$F3,$3F,$43,$D4,$FA,$2D,$C3,$EB,$B6,$FE,$4B,$6B
	.dc.b	$FD,$D7,$F5,$2B,$FF,$09,$B6,$FF,$AA,$F6,$B8,$3F,$93,$5F,$0B,$D9
	.dc.b	$FE,$28,$BE,$18,$FF,$89,$DA,$FF,$96,$DA,$FE,$51,$DC,$FE,$4F,$EA
	.dc.b	$5F,$31,$D3,$2F,$D2,$46,$1F,$C8,$FD,$4F,$E3,$BE,$A7,$F1,$ED,$8C
	.dc.b	$9F,$96,$8C,$B4,$FC,$CF,$C8,$FE,$5B,$AF,$F2,$ED,$8C,$9F,$9E,$8C
	.dc.b	$B4,$FE,$E7,$F5,$1F,$D0,$F4,$FF,$45,$B5,$34,$9E,$82,$7F,$D0,$D9
	.dc.b	$FE,$D2,$5F,$1D,$9B,$FF,$4A,$4E,$6F,$1F,$03,$F1,$27,$BD,$62,$B9
	.dc.b	$36,$33,$87,$67,$CB,$79,$7C,$B3,$70,$F0,$D7,$F8,$96,$79,$B5,$E6
	.dc.b	$7F,$9B,$E3,$6D,$F9,$8F,$F1,$B3,$EB,$6B,$F1,$1A,$FC,$46,$C1,$F2
	.dc.b	$7D,$47,$EE,$FC,$2F,$45,$6E,$EC,$E6,$3E,$29,$FE,$BA,$71,$9F,$C5
	.dc.b	$84,$63,$3F,$25,$72,$EE,$FF,$2B,$D7,$23,$6E,$2F,$95,$F1,$9F,$55
	.dc.b	$71,$C2,$47,$19,$D3,$E7,$31,$7C,$F0,$FE,$63,$3A,$37,$97,$C1,$C3
	.dc.b	$9E,$3E,$0B,$E1,$87,$B5,$AF,$58,$D6,$3C,$31,$EB,$5A,$C3,$A5,$F8
	.dc.b	$AB,$4F,$37,$E6,$C6,$AF,$9C,$EC,$8F,$CB,$8D,$5F,$6A,$E7,$5F,$21
	.dc.b	$E7,$97,$DA,$C3,$CF,$97,$BD,$F2,$46,$67,$8C,$FC,$28,$C5,$78,$2F
	.dc.b	$B3,$F5,$70,$F0,$BE,$93,$45,$C7,$CD,$5C,$9F,$59,$2B,$F0,$33,$C0
	.dc.b	$F4,$16,$BA,$B8,$9E,$0F,$59,$E3,$18,$8D,$47,$9A,$3C,$39,$7B,$32
	.dc.b	$F8,$27,$C0,$71,$97,$FB,$26,$FC,$27,$FB,$F7,$2F,$CE,$36,$BF,$2A
	.dc.b	$E7,$E9,$9E,$F7,$1A,$46,$27,$FD,$A9,$1C,$9F,$C7,$C6,$5E,$53,$FA
	.dc.b	$A8,$E1,$8F,$E7,$9B,$97,$F8,$1B,$EA,$1F,$51,$FE,$08,$AD,$5F,$12
	.dc.b	$61,$FA,$9A,$DB,$89,$DB,$FA,$5C,$3A,$A1,$3C,$1D,$57,$1C,$3D,$2F
	.dc.b	$F0,$29,$D5,$F4,$24,$EA,$BC,$58,$79,$E3,$13,$A9,$FE,$0D,$87,$89
	.dc.b	$F0,$D6,$DC,$E1,$E1,$69,$D0,$C1,$FC,$32,$79,$63,$F4,$A7,$CD,$5C
	.dc.b	$EB,$51,$C2,$75,$1C,$CF,$79,$2F,$CC,$F9,$0E,$17,$D8,$F4,$63,$F0
	.dc.b	$B1,$F6,$92,$7C,$F5,$FC,$86,$75,$18,$9D,$3E,$43,$C5,$3E,$0B,$F2
	.dc.b	$BE,$2A,$E3,$3B,$8D,$5F,$8B,$1F,$75,$F0,$C3,$AB,$6F,$EF,$61,$CA
	.dc.b	$31,$1C,$E7,$C7,$39,$BF,$76,$3F,$3A,$3C,$26,$5C,$61,$F3,$63,$F1
	.dc.b	$1A,$7F,$97,$30,$F9,$07,$F6,$93,$C3,$3F,$ED,$F7,$F3,$8D,$7C,$BF
	.dc.b	$03,$5F,$C9,$B6,$8D,$24,$7E,$4D,$F1,$1F,$7F,$EA,$1F,$03,$4C,$CE
	.dc.b	$CF,$EB,$71,$A6,$C7,$F0,$35,$19,$7D,$86,$23,$51,$98,$F8,$1E,$E7
	.dc.b	$F9,$75,$FD,$67,$57,$C1,$87,$D4,$7F,$A0,$23,$0D,$CF,$FD,$07,$3C
	.dc.b	$6B,$FC,$AA,$1C,$57,$08,$EB,$3C,$2B,$94,$7E,$EE,$B7,$3D,$3D,$8C
	.dc.b	$CF,$3F,$A1,$B4,$6C,$C9,$EA,$73,$8D,$99,$7C,$BF,$50,$FA,$8E,$7E
	.dc.b	$84,$9E,$22,$74,$E1,$AF,$CC,$6B,$F8,$96,$BF,$EA,$66,$BF,$C5,$DE
	.dc.b	$39,$FE,$35,$B9,$6E,$FD,$BA,$FC,$0F,$A8,$FF,$0E,$61,$E4,$5E,$F6
	.dc.b	$4D,$75,$C9,$D4,$65,$E1,$1C,$FE,$42,$47,$37,$B7,$C2,$E2,$3E,$CD
	.dc.b	$A7,$1A,$7E,$8E,$15,$F5,$3E,$A1,$E5,$79,$FD,$19,$8F,$CA,$9C,$B6
	.dc.b	$7F,$2D,$18,$8F,$CA,$B5,$FD,$93,$5F,$F6,$7B,$BC,$3F,$68,$DD,$B3
	.dc.b	$5F,$1F,$A7,$F0,$FF,$50,$FA,$8A,$4C,$7E,$EC,$3E,$63,$11,$F9,$88
	.dc.b	$E0,$EA,$79,$CF,$4F,$9C,$EB,$F4,$72,$B9,$C6,$5E,$B1,$97,$A4,$7E
	.dc.b	$6B,$C3,$FF,$26,$6B,$FC,$37,$E9,$DF,$51,$F9,$E6,$DF,$E4,$AB,$7B
	.dc.b	$7F,$BC,$1F,$F2,$A4,$7F,$17,$6B,$9F,$0F,$F8,$14,$62,$3F,$B3,$3C
	.dc.b	$9B,$1E,$33,$C1,$EC,$78,$7B,$64,$FE,$8B,$04,$79,$7B,$73,$3F,$BE
	.dc.b	$9C,$BB,$9D,$3C,$7A,$BD,$9F,$1D,$23,$04,$39,$39,$7A,$47,$C7,$EB
	.dc.b	$FE,$03,$EB,$3E,$A2,$99,$AF,$8D,$3F,$00,$8F,$F8,$97,$5F,$D3,$B6
	.dc.b	$38,$3F,$C5,$A3,$2F,$F1,$8A,$FD,$7B,$FC,$6B,$E9,$5C,$7F,$CB,$5B
	.dc.b	$1F,$E3,$16,$BF,$CB,$9A,$FF,$4B,$5E,$87,$12,$47,$17,$F9,$C7,$D2
	.dc.b	$7E,$B2,$3F,$C7,$8D,$7F,$9C,$B5,$FE,$AC,$D7,$FC,$44,$17,$C3,$74
	.dc.b	$24,$74,$7F,$A3,$7D,$27,$F4,$18,$E2,$D7,$FC,$58,$D4,$D3,$FA,$67
	.dc.b	$37,$F6,$56,$F2,$F2,$DF,$F9,$BF,$D2,$3E,$3B,$F9,$CF,$A9,$FA,$FC
	.dc.b	$7F,$96,$CF,$FB,$59,$BE,$A6,$BB,$24,$7F,$97,$6B,$83,$FD,$69,$AF
	.dc.b	$F7,$76,$BF,$09,$DE,$7F,$DE,$5B,$96,$6B,$CB,$DF,$FD,$33,$DB,$F3
	.dc.b	$61,$FD,$03,$E7,$9E,$E7,$F5,$06,$C7,$F4,$49,$DB,$FD,$79,$AF,$F8
	.dc.b	$25,$AF,$D4,$6B,$FE,$57,$6A,$79,$E5,$FE,$0F,$6E,$3C,$75,$F1,$7C
	.dc.b	$1F,$F6,$2F,$05,$F4,$9F,$D9,$3E,$89,$F1,$34,$D4,$4C,$FF,$D0,$D5
	.dc.b	$C1,$B1,$FE,$07,$8C,$47,$F8,$19,$B9,$FF,$88,$5B,$97,$E3,$9B,$3F
	.dc.b	$E7,$1C,$1F,$17,$FD,$DB,$E9,$1F,$39,$E5,$FA,$49,$7F,$DB,$27,$2D
	.dc.b	$9F,$F0,$C4,$62,$3F,$C2,$ED,$7F,$C6,$8D,$7F,$3E,$EF,$2F,$F1,$B3
	.dc.b	$72,$CD,$7C,$5F,$0F,$FC,$13,$F4,$7F,$82,$C3,$D1,$EB,$7C,$93,$3C
	.dc.b	$D2,$3F,$D5,$93,$F8,$C7,$FD,$1C,$D4,$D3,$A3,$88,$E4,$FF,$88,$B5
	.dc.b	$7D,$27,$8B,$E0,$BE,$C7,$B1,$2B,$FC,$1F,$82,$75,$1E,$27,$FC,$79
	.dc.b	$19,$8C,$1F,$EC,$1F,$A4,$74,$7F,$90,$5B,$19,$8C,$11,$FF,$0E,$C6
	.dc.b	$9F,$F1,$BB,$78,$3F,$CC,$1F,$49,$FE,$01,$7F,$D5,$DF,$46,$FA,$3F
	.dc.b	$E1,$F3,$9B,$4E,$5F,$F5,$A5,$F5,$8C,$42,$31,$5C,$BD,$96,$1D,$A3
	.dc.b	$0D,$AE,$78,$FF,$1E,$C3,$9E,$1C,$A7,$FD,$85,$8F,$8E,$7B,$B7,$0F
	.dc.b	$E4,$B5,$BA,$F1,$5F,$39,$EF,$1E,$A6,$DF,$69,$D4,$F8,$8F,$20,$BE
	.dc.b	$CF,$8B,$1F,$5A,$FA,$4E,$4F,$FB,$5D,$AE,$70,$F0,$C7,$A2,$78,$E1
	.dc.b	$A2,$4F,$90,$95,$E3,$3D,$05,$8E,$59,$79,$B0,$FA,$D3,$FE,$9D,$6B
	.dc.b	$F0,$0C,$5F,$A1,$D7,$B2,$F8,$EF,$F4,$4C,$7B,$E3,$0F,$FA,$32,$BF
	.dc.b	$C8,$73,$A6,$C7,$FA,$4F,$E1,$BF,$D2,$89,$A4,$74,$7A,$87,$C6,$BE
	.dc.b	$89,$D3,$F4,$27,$FE,$00,$57,$D2,$AD,$4E,$67,$B1,$CE,$32,$F5,$38
	.dc.b	$99,$AE,$AE,$4E,$3B,$7E,$62,$78,$BF,$9A,$8E,$2E,$67,$8B,$98,$F0
	.dc.b	$EB,$9B,$F2,$61,$E7,$BE,$F3,$B8,$C4,$EA,$BC,$8D,$9E,$51,$E5,$9D
	.dc.b	$B4,$D9,$63,$67,$0D,$7F,$DB,$58,$FC,$27,$FD,$B9,$3F,$0A,$75,$7F
	.dc.b	$E0,$AA,$F8,$F3,$FE,$E3,$6B,$9C,$38,$A3,$EB,$B8,$7E,$CF,$4D,$9F
	.dc.b	$E8,$75,$F8,$A6,$CF,$FB,$51,$AF,$79,$FA,$78,$4F,$A4,$FE,$9B,$7F
	.dc.b	$EB,$4B,$E3,$39,$ED,$DC,$FA,$5B,$87,$96,$B8,$BF,$C3,$A7,$D1,$F1
	.dc.b	$4F,$81,$FE,$FF,$1D,$6F,$8D,$7E,$49,$F5,$12,$3D,$4D,$8E,$82,$9F
	.dc.b	$49,$FF,$00,$2B,$A1,$FE,$D4,$8C,$FC,$3F,$6F,$27,$FA,$04,$FF,$B4
	.dc.b	$1B,$7D,$21,$F4,$7D,$5D,$75,$F1,$3E,$85,$70,$AF,$F2,$18,$70,$FF
	.dc.b	$91,$9A,$FD,$97,$BE,$BF,$A2,$AE,$FC,$1D,$5F,$F7,$59,$83,$FD,$CB
	.dc.b	$F4,$8F,$6B,$E3,$B6,$37,$93,$D0,$CD,$49,$FE,$17,$5B,$CB,$CE,$1C
	.dc.b	$3E,$A6,$BF,$E5,$E7,$86,$3E,$7C,$2D,$66,$FF,$02,$FC,$87,$ED,$BF
	.dc.b	$ED,$83,$FD,$FA,$5B,$F9,$D3,$FD,$DA,$33,$D3,$C0,$FC,$E3,$B9,$F9
	.dc.b	$AD,$7F,$CE,$2D,$7F,$48,$F5,$BF,$99,$F4,$9F,$E6,$38,$EA,$7C,$C7
	.dc.b	$13,$F8,$A1,$F3,$1F,$98,$AF,$E2,$B1,$8E,$2F,$6E,$2B,$FC,$BD,$EF
	.dc.b	$0F,$F4,$1B,$FA,$0B,$FC,$1B,$5C,$DF,$ED,$E3,$11,$FB,$B7,$93,$FC
	.dc.b	$C2,$DE,$2F,$6A,$F0,$BF,$A6,$FA,$27,$A3,$4F,$D4,$FD,$23,$E9,$31
	.dc.b	$9F,$F0,$7D,$7D,$77,$C2,$3F,$A6,$B7,$3F,$C1,$B5,$FC,$DB,$5F,$F4
	.dc.b	$FA,$F7,$BF,$88,$36,$79,$C6,$0C,$FA,$52,$3B,$B6,$BF,$23,$D1,$FE
	.dc.b	$DE,$9F,$EB,$ED,$8C,$F9,$9F,$C7,$63,$11,$FA,$EF,$A4,$7A,$7F,$C8
	.dc.b	$5B,$7F,$DA,$1B,$3F,$9E,$6B,$FB,$6F,$A1,$FF,$5B,$CF,$C5,$3F,$D7
	.dc.b	$6E,$23,$F8,$6F,$D2,$3E,$91,$F4,$9F,$C5,$AB,$5B,$FE,$19,$E1,$7D
	.dc.b	$7D,$DF,$E6,$91,$88,$EB,$18,$3F,$93,$F8,$6F,$A4,$FF,$46,$E7,$E3
	.dc.b	$AF,$E7,$D7,$DA,$3F,$96,$4E,$A7,$F9,$0F,$4F,$F4,$3F,$A4,$FE,$62
	.dc.b	$FF,$45,$E9,$B2,$9B,$18,$6D,$F7,$AE,$4D,$AF,$C9,$C7,$F4,$FF,$A4
	.dc.b	$E6,$74,$F6,$ED,$6E,$2D,$79,$0F,$F5,$2C,$64,$F6,$2F,$A2,$7A,$BF
	.dc.b	$D0,$1B,$3F,$CD,$FE,$93,$FD,$F2,$F0,$8C,$D7,$ED,$DA,$FF,$30,$6B
	.dc.b	$FE,$F4,$6B,$FD,$9D,$AF,$86,$BF,$AB,$F3,$2F,$F9,$9D,$F1,$9C,$C6
	.dc.b	$2B,$F8,$2F,$A3,$FE,$24,$AF,$90,$FE,$3B,$8B,$FC,$37,$A7,$F8,$E6
	.dc.b	$7E,$9C,$E5,$FD,$F3,$C6,$BE,$CB,$97,$AC,$70,$E7,$E3,$F4,$92,$3B
	.dc.b	$FC,$4D,$AE,$33,$71,$F0,$E0,$AC,$5F,$C3,$F8,$DF,$9C,$7E,$43,$E0
	.dc.b	$5F,$29,$89,$F9,$D3,$FE,$0B,$E0,$34,$91,$FE,$54,$8C,$BF,$97,$C5
	.dc.b	$F4,$9F,$63,$4F,$AD,$7F,$E0,$49,$CF,$C4,$FC,$4F,$8C,$98,$7C,$AF
	.dc.b	$A4,$B8,$7C,$49,$B7,$EC,$9B,$5F,$2B,$E5,$7E,$6F,$C4,$7C,$E6,$C7
	.dc.b	$D4,$6E,$1F,$39,$B7,$FC,$F3,$E9,$3F,$C2,$33,$FD,$05,$B1,$FE,$54
	.dc.b	$EE,$70,$7F,$25,$8E,$18,$FF,$85,$AB,$8D,$FF,$80,$1A,$FF,$8D,$43
	.dc.b	$87,$FC,$70,$D7,$31,$F9,$2F,$A4,$7C,$A7,$58,$C3,$F8,$AF,$A4,$FC
	.dc.b	$5F,$D2,$5D,$7F,$22,$DC,$3F,$16,$DC,$BF,$06,$FE,$4B,$83,$F2,$7E
	.dc.b	$C5,$BF,$CC,$B6,$FF,$26,$DB,$FC,$5B,$FE,$34,$FA,$4F,$E8,$53,$FE
	.dc.b	$51,$6A,$69,$3D,$0F,$F3,$EF,$C0,$7B,$27,$87,$C7,$E3,$5F,$8D,$7F
	.dc.b	$D4,$6B,$55,$CE,$B4,$F0,$CB,$FD,$23,$7C,$27,$33,$E1,$C2,$E0,$9A
	.dc.b	$7F,$8B,$67,$85,$FF,$81,$9A,$7F,$9D,$27,$84,$66,$3D,$DE,$F7,$16
	.dc.b	$C6,$5F,$81,$19,$6C,$66,$7F,$CB,$CF,$A2,$3F,$CF,$1F,$49,$FB,$5F
	.dc.b	$0F,$11,$C0,$FF,$5B,$65,$F8,$99,$D4,$60,$E3,$DB,$FE,$5F,$70,$23
	.dc.b	$7A,$7C,$AC,$7C,$33,$A9,$CD,$F5,$BE,$57,$F4,$27,$35,$EE,$38,$4E
	.dc.b	$AB,$86,$1E,$8A,$DD,$7A,$6F,$95,$EE,$3C,$F5,$B8,$C4,$6E,$F9,$BD
	.dc.b	$A7,$47,$C4,$C3,$D3,$7D,$23,$33,$B3,$E0,$CE,$AB,$9D,$6E,$74,$FC
	.dc.b	$3C,$BF,$0F,$8F,$A0,$98,$F9,$30,$E5,$7E,$83,$8E,$1D,$0F,$CB,$DF
	.dc.b	$2A,$F5,$DF,$2C,$7D,$13,$97,$CF,$8F,$8C,$F1,$65,$8C,$3A,$63,$DE
	.dc.b	$72,$F8,$0C,$46,$9F,$90,$60,$87,$E0,$71,$F5,$61,$E1,$8E,$46,$70
	.dc.b	$F2,$E1,$E1,$8E,$38,$FD,$AC,$BE,$0C,$6C,$E3,$19,$BE,$99,$7A,$F2
	.dc.b	$F2,$5F,$0A,$F9,$9F,$49,$FE,$A1,$C3,$FD,$1B,$08,$C4,$7F,$A9,$A3
	.dc.b	$F3,$AD,$8C,$BF,$A1,$6B,$FA,$86,$BF,$9F,$6C,$75,$F7,$BF,$D3,$73
	.dc.b	$97,$F2,$D3,$E3,$C3,$F3,$CD,$C3,$F9,$B7,$D1,$BE,$93,$F9,$C5,$6A
	.dc.b	$F8,$93,$0F,$D7,$D6,$DC,$4E,$DF,$D7,$61,$D5,$FF,$7B,$4E,$DC,$D7
	.dc.b	$0C,$3D,$CF,$F0,$29,$D5,$F3,$24,$EA,$BC,$38,$79,$A3,$13,$A9,$F8
	.dc.b	$D7,$CE,$B5,$18,$3C,$24,$9F,$E5,$F7,$E0,$BE,$64,$9E,$13,$AA,$E7
	.dc.b	$1F,$34,$E5,$3B,$55,$CD,$E3,$19,$C7,$EB,$47,$A8,$F8,$46,$88,$72
	.dc.b	$AE,$F1,$C2,$FD,$D3,$E1,$7C,$8F,$23,$15,$E3,$BF,$14,$60,$EB,$5C
	.dc.b	$E7,$F0,$37,$DB,$1F,$B1,$5B,$25,$F2,$BE,$11,$CA,$77,$39,$78,$CE
	.dc.b	$A7,$8D,$75,$C7,$CD,$39,$BF,$29,$C6,$3D,$4E,$27,$FD,$BB,$5E,$63
	.dc.b	$FA,$84,$7E,$2F,$4F,$F0,$74,$7C,$73,$3F,$03,$C1,$FC,$8B,$5F,$AA
	.dc.b	$D9,$34,$FB,$AE,$23,$F2,$91,$F8,$0F,$A4,$7C,$0D,$37,$12,$7F,$AD
	.dc.b	$CE,$9B,$3F,$B4,$51,$97,$D2,$62,$35,$19,$8F,$6F,$B9,$FE,$52,$7F
	.dc.b	$D1,$2B,$E9,$3F,$C5,$DC,$FF,$EF,$68,$C4,$EB,$C3,$CE,$7F,$E3,$39
	.dc.b	$E5,$3F,$EF,$F7,$6D,$BD,$A3,$FD,$E3,$58,$7F,$DE,$7E,$C5,$A7,$CE
	.dc.b	$D3,$11,$27,$CF,$19,$7D,$31,$CF,$E9,$1F,$49,$CF,$D0,$95,$FE,$A6
	.dc.b	$9C,$B6,$7E,$54,$62,$3E,$4B,$5F,$C1,$B5,$FF,$29,$3B,$D1,$A6,$2B
	.dc.b	$C4,$7A,$F6,$F7,$7D,$23,$A6,$FB,$D3,$2F,$F5,$CC,$E5,$B3,$95,$8F
	.dc.b	$67,$E5,$4A,$BA,$D7,$6F,$8E,$95,$D1,$F0,$FC,$2E,$23,$9E,$7F,$E6
	.dc.b	$BC,$15,$9C,$7B,$52,$57,$F5,$FE,$91,$E5,$FB,$3D,$09,$7F,$ED,$29
	.dc.b	$CB,$67,$F2,$D1,$88,$FC,$AB,$5F,$D7,$B5,$FF,$3F,$3B,$C3,$C6,$DC
	.dc.b	$B3,$5F,$33,$AB,$F1,$5F,$48,$FA,$4A,$4C,$7F,$DC,$B0,$F9,$8C,$47
	.dc.b	$C2,$3F,$29,$1C,$5D,$CF,$49,$EB,$F3,$7D,$BE,$86,$57,$48,$CB,$DA
	.dc.b	$32,$F5,$16,$9F,$D0,$78,$DF,$E7,$4D,$7E,$7F,$D1,$BE,$93,$F4,$8D
	.dc.b	$C3,$FD,$FB,$5B,$DB,$FD,$44,$FF,$9F,$23,$F8,$BB,$5C,$F8,$BF,$C1
	.dc.b	$63,$11,$FB,$17,$93,$63,$C6,$78,$3B,$BC,$3D,$92,$7F,$71,$97,$9A
	.dc.b	$3C,$BD,$B9,$9F,$76,$49,$C5,$7B,$2B,$D7,$D5,$EC,$F8,$E9,$87,$4D
	.dc.b	$3F,$D7,$98,$FF,$19,$C3,$E1,$DE,$AB,$AD,$F2,$7E,$1F,$5F,$F0,$9F
	.dc.b	$59,$F4,$94,$F9,$B1,$27,$FC,$33,$BF,$E9,$DB,$1F,$14,$8F,$42,$1F
	.dc.b	$66,$BF,$68,$FD,$BF,$A4,$72,$7F,$31,$6C,$7F,$89,$9A,$FF,$32,$6B
	.dc.b	$FD,$31,$7A,$1C,$49,$1C,$5F,$E7,$9F,$49,$FB,$08,$FF,$18,$B5,$FE
	.dc.b	$7A,$D7,$FA,$BB,$5F,$F7,$32,$F8,$7E,$64,$8E,$6F,$F4,$9F,$A4,$FE
	.dc.b	$89,$1F,$EE,$D6,$BF,$CD,$DA,$9A,$7D,$7E,$1E,$1B,$FF,$3B,$8C,$4F
	.dc.b	$F3,$FF,$A4,$7C,$77,$F3,$FF,$53,$F6,$53,$F8,$13,$5E,$A7,$F4,$3A
	.dc.b	$FF,$67,$46,$44,$F0,$7F,$58,$D7,$FB,$BB,$5F,$E9,$C7,$0E,$DF,$EC
	.dc.b	$D1,$98,$FE,$A4,$F3,$E5,$FE,$3E,$DC,$F8,$B6,$7D,$5D,$FF,$D5,$FD
	.dc.b	$BF,$5C,$79,$7E,$75,$ED,$7F,$5B,$6C,$7A,$67,$6F,$F0,$36,$BF,$E0
	.dc.b	$E6,$BE,$86,$BF,$D9,$DA,$9E,$78,$FF,$36,$6E,$3C,$75,$F1,$7C,$1F
	.dc.b	$F6,$DF,$05,$F4,$9F,$DB,$FE,$89,$F1,$34,$D4,$4C,$FF,$CF,$F5,$C1
	.dc.b	$B1,$FE,$0F,$8C,$47,$F8,$39,$B9,$FF,$48,$6E,$5F,$49,$B3,$FE,$01
	.dc.b	$C1,$F1,$7F,$E0,$1F,$A3,$7C,$E7,$97,$E9,$25,$FE,$06,$72,$D9,$FF
	.dc.b	$10,$46,$23,$FC,$3E,$D7,$FC,$72,$D7,$F0,$6E,$F2,$FE,$BA,$DC,$B3
	.dc.b	$5F,$17,$C5,$FF,$0A,$7D,$1F,$E0,$70,$66,$F5,$BE,$51,$AE,$79,$3F
	.dc.b	$EA,$6A,$FC,$72,$69,$23,$71,$FD,$16,$3B,$53,$E9,$3C,$3F,$05,$F5
	.dc.b	$FD,$89,$7F,$A3,$B4,$EA,$3C,$0F,$F9,$0E,$33,$18,$3F,$D7,$3F,$48
	.dc.b	$E8,$FF,$24,$36,$33,$18,$23,$FB,$18,$D3,$FD,$6D,$BC,$1F,$E0,$6F
	.dc.b	$A3,$FF,$04,$BF,$E3,$DF,$A2,$7D,$27,$EC,$A7,$36,$9C,$BF,$E4,$3B
	.dc.b	$EB,$18,$84,$62,$B9,$7B,$3F,$E2,$29,$E5,$18,$C3,$C0,$FF,$00,$AE
	.dc.b	$13,$D7,$1F,$E0,$78,$75,$C7,$C6,$DC,$7A,$3E,$1B,$E3,$7E,$2B,$E3
	.dc.b	$87,$78,$F0,$9F,$E5,$BC,$39,$C7,$86,$77,$3E,$13,$CA,$6F,$0E,$F3
	.dc.b	$CF,$E9,$3C,$15,$FC,$61,$B5,$FE,$67,$7C,$D9,$5A,$F1,$65,$AA,$F1
	.dc.b	$9E,$73,$B8,$C7,$CA,$F8,$6B,$39,$7B,$8F,$7C,$EA,$78,$65,$F3,$DF
	.dc.b	$F5,$1C,$7D,$A6,$C6,$5F,$88,$6B,$D9,$7C,$87,$FA,$4A,$3C,$51,$87
	.dc.b	$FD,$2B,$5F,$E4,$C9,$D3,$63,$FD,$39,$F0,$DF,$E9,$D4,$D2,$3A,$3E
	.dc.b	$0F,$C6,$BE,$8B,$FC,$CD,$A7,$E2,$E7,$FD,$21,$5F,$12,$B5,$D3,$F1
	.dc.b	$A7,$27,$2B,$E3,$B7,$FA,$AA,$F8,$46,$27,$FD,$8B,$18,$AE,$51,$E6
	.dc.b	$39,$38,$8D,$47,$D6,$3B,$E1,$F9,$FC,$3C,$71,$93,$94,$66,$FC,$11
	.dc.b	$FD,$72,$B9,$CF,$B6,$37,$87,$F6,$0B,$E7,$39,$AD,$CF,$AE,$BC,$A7
	.dc.b	$78,$CB,$FE,$B5,$AF,$1E,$DF,$E8,$26,$E3,$C7,$0F,$23,$FD,$D6,$33
	.dc.b	$5C,$7C,$9F,$F4,$B5,$F1,$CA,$DF,$96,$6D,$77,$EA,$B6,$FA,$4F,$E8
	.dc.b	$B7,$D6,$73,$18,$AF,$F3,$66,$3F,$D2,$2B,$FC,$15,$7C,$7B,$BA,$D2
	.dc.b	$B4,$FF,$9F,$E7,$55,$E9,$AD,$D7,$59,$FF,$57,$7C,$5F,$FA,$DF,$C6
	.dc.b	$CD,$4B,$FF,$0C,$4B,$E9,$3F,$DF,$F1,$CC,$CB,$8C,$7C,$F5,$F0,$2B
	.dc.b	$87,$C3,$F3,$9F,$F4,$94,$62,$FF,$C4,$1E,$4F,$F3,$E6,$F5,$FC,$85
	.dc.b	$F4,$4A,$F7,$D6,$EF,$FD,$20,$DD,$3F,$D9,$17,$C3,$87,$78,$F3,$C7
	.dc.b	$EC,$E0,$C3,$85,$FF,$BD,$CE,$07,$88,$FF,$1C,$18,$8F,$F4,$F7,$D2
	.dc.b	$3D,$9F,$A2,$D9,$DD,$B8,$F3,$52,$72,$DB,$FF,$5A,$47,$BA,$31,$9F
	.dc.b	$AA,$FA,$57,$82,$7D,$47,$E0,$2B,$FD,$AE,$7F,$92,$4B,$1D,$27,$FB
	.dc.b	$74,$67,$A7,$81,$F3,$7E,$93,$FC,$AF,$19,$3D,$EE,$2B,$FD,$79,$1D
	.dc.b	$30,$F9,$B3,$97,$FD,$D3,$19,$3E,$54,$65,$FF,$2A,$FC,$C7,$E3,$DB
	.dc.b	$C3,$F4,$E7,$F3,$33,$9B,$07,$15,$F8,$2F,$17,$77,$F6,$DB,$1F,$41
	.dc.b	$B5,$F0,$5F,$D2,$FD,$17,$F9,$B2,$7F,$CD,$CD,$9F,$F6,$C7,$D2,$3E
	.dc.b	$93,$7E,$94,$BD,$FB,$96,$FF,$DB,$D8,$23,$A9,$FB,$2F,$AC,$FC,$65
	.dc.b	$70,$3F,$CF,$8E,$2B,$8F,$A7,$FB,$8E,$8F,$F6,$55,$79,$23,$1E,$6C
	.dc.b	$8F,$F4,$24,$FF,$96,$9B,$1C,$7D,$2F,$E0,$3F,$4A,$F5,$3F,$81,$FC
	.dc.b	$67,$FA,$C2,$FF,$D1,$93,$97,$FD,$1D,$19,$CB,$FD,$F1,$F4,$5F,$EA
	.dc.b	$99,$FF,$49,$36,$1C,$C9,$FA,$BF,$4A,$FA,$57,$D2,$FF,$12,$C3,$85
	.dc.b	$EF,$0F,$B7,$18,$C3,$F9,$87,$D2,$FF,$9D,$9E,$FE,$FF,$F2,$6F,$1D
	.dc.b	$39,$BF,$2C,$E6,$31,$87,$59,$61,$D5,$FE,$45,$3A,$C3,$F1,$7D,$7F
	.dc.b	$8C,$FA,$5F,$E6,$6F,$E3,$BB,$2D,$A6,$C6,$1B,$5F,$EB,$9A,$E4,$DA
	.dc.b	$FC,$C4,$7E,$63,$E9,$79,$9D,$38,$ED,$6E,$2D,$7C,$C8,$F2,$BF,$74
	.dc.b	$C1,$FA,$11,$1F,$CF,$BE,$8D,$EB,$D3,$F9,$77,$D2,$FD,$BD,$BE,$24
	.dc.b	$EB,$0F,$F6,$63,$70,$FE,$63,$5F,$C8,$1F,$D7,$C6,$6B,$FA,$1C,$E6
	.dc.b	$53,$9A,$FD,$AF,$42,$B5,$39,$AF,$E6,$D1,$D7,$3F,$F1,$2D,$FD,$0F
	.dc.b	$4F,$F2,$55,$EE,$9C,$CF,$CD,$9C,$E3,$F8,$99,$CD,$6E,$73,$5F,$71
	.dc.b	$FC,$A3,$C7,$A3,$B7,$D2,$C8,$E9,$F1,$56,$F9,$4D,$C7,$B6,$0A,$C5
	.dc.b	$FF,$BB,$33,$7C,$77,$D1,$3F,$2D,$F0,$6F,$95,$C5,$7C,$E9,$FE,$FB
	.dc.b	$F4,$BF,$4E,$39,$46,$3E,$2B,$7A,$FF,$75,$AD,$3F,$4A,$75,$5F,$0F
	.dc.b	$E3,$7E,$41,$F2,$7E,$96,$C6,$49,$5F,$16,$32,$FC,$7F,$96,$FA,$1E
	.dc.b	$BF,$D1,$6D,$7C,$F6,$CE,$E3,$E9,$4F,$F8,$33,$E9,$7F,$DF,$F1,$FC
	.dc.b	$D3,$87,$F6,$11,$8A,$FC,$BF,$7F,$FB,$DF,$CA,$E1,$84,$C3,$FC,$13
	.dc.b	$F4,$AF,$95,$C6,$1F,$E0,$28,$C4,$F5,$3F,$A0,$C3,$1E,$D9,$7F,$85
	.dc.b	$AB,$73,$FA,$63,$58,$7F,$83,$E3,$33,$F7,$3E,$97,$F2,$1F,$4B,$F9
	.dc.b	$16,$E3,$F8,$F6,$E1,$F8,$D6,$D7,$E1,$38,$BF,$31,$EC,$FE,$65,$B7
	.dc.b	$F9,$76,$DF,$E5,$5B,$5F,$C5,$7E,$97,$FA,$77,$C0,$FC,$1A,$FE,$BB
	.dc.b	$0F,$91,$FE,$4D,$39,$94,$F0,$DF,$E1,$BB,$C7,$FC,$FB,$18,$E0,$E2
	.dc.b	$EB,$0B,$7D,$A6,$BD,$22,$DF,$38,$FF,$47,$D7,$1B,$EB,$1C,$5A,$F1
	.dc.b	$6C,$F1,$C3,$D0,$BE,$B3,$AE,$1D,$9E,$FE,$6F,$8B,$E9,$7F,$5D,$E1
	.dc.b	$F2,$8C,$77,$7F,$8F,$73,$F1,$9F,$1A,$33,$3B,$CF,$C7,$7C,$AB,$61
	.dc.b	$C6,$7F,$2D,$BA,$7C,$AC,$39,$D7,$09,$CB,$FE,$10,$6B,$F4,$A3,$75
	.dc.b	$CF,$1F,$33,$71,$F2,$57,$39,$CE,$1F,$9A,$C3,$99,$D3,$46,$3D,$27
	.dc.b	$55,$C2,$FC,$4D,$7C,$59,$7C,$08,$C6,$1D,$1D,$61,$D2,$75,$8C,$BF
	.dc.b	$23,$E6,$CB,$D1,$18,$23,$F6,$B2,$F2,$47,$8A,$79,$4E,$6A,$E5,$E8
	.dc.b	$9B,$5F,$5E,$6B,$F4,$F1,$F3,$9E,$6C,$3A,$C6,$6B,$78,$FA,$63,$C7
	.dc.b	$87,$56,$9D,$4A,$71,$2C,$71,$C7,$DA,$49,$E4,$FA,$67,$8E,$5F,$26
	.dc.b	$FE,$25,$68,$93,$F0,$4F,$C6,$DF,$58,$CC,$7E,$3F,$E9,$BF,$D1,$4F
	.dc.b	$F1,$83,$F1,$4E,$1C,$5F,$E5,$DB,$19,$7F,$0C,$F2,$C3,$F3,$0D,$AC
	.dc.b	$FB,$FF,$B2,$FA,$57,$D3,$7F,$B9,$67,$38,$7E,$6E,$F8,$3E,$2B,$E1
	.dc.b	$5B,$8C,$1B,$8F,$D5,$CF,$08,$5F,$67,$33,$1E,$07,$EC,$23,$4E,$30
	.dc.b	$F4,$CE,$DF,$E3,$93,$A7,$33,$BC,$3D,$31,$AC,$3C,$D3,$F3,$6F,$95
	.dc.b	$68,$FE,$8B,$7C,$DC,$4F,$B2,$FA,$9C,$DE,$26,$63,$6F,$BA,$FB,$E2
	.dc.b	$C7,$E1,$37,$1E,$35,$B2,$C6,$5F,$51,$FE,$98,$C7,$C8,$F7,$21,$EB
	.dc.b	$3E,$45,$F8,$2F,$B9,$2F,$EB,$CF,$68,$C3,$C0,$C4,$79,$2B,$BC,$75
	.dc.b	$C7,$E0,$E3,$EE,$9C,$E3,$E7,$9C,$D7,$92,$73,$87,$69,$CC,$FB,$CF
	.dc.b	$69,$D6,$FA,$63,$E9,$C7,$C7,$3E,$52,$57,$2C,$7C,$17,$F6,$2F,$EE
	.dc.b	$E3,$E8,$23,$8F,$B7,$CF,$C2,$4C,$7F,$40,$FE,$13,$E9,$FF,$DD,$91
	.dc.b	$82,$1F,$93,$8D,$7C,$1C,$8C,$E7,$F9,$29,$C4,$7F,$48,$AE,$6F,$F8
	.dc.b	$2D,$BC,$DF,$E8,$D6,$F1,$FF,$85,$1F,$F8,$01,$3F,$E1,$73,$FD,$AF
	.dc.b	$EF,$FF,$91,$9C,$05,$F4,$FF,$E1,$CB,$FF,$4C,$46,$27,$6F,$17,$FC
	.dc.b	$38,$D8,$DC,$F1,$BF,$F1,$2C,$F1,$AD,$F8,$99,$7C,$7E,$D5,$7C,$5F
	.dc.b	$41,$68,$D9,$93,$E2,$8D,$18,$7F,$C5,$DF,$4E,$FA,$7E,$5E,$94,$9F
	.dc.b	$7C,$E5,$B3,$E1,$8C,$47,$28,$F1,$70,$65,$91,$C0,$F8,$71,$E9,$FA
	.dc.b	$7F,$EC,$F1,$F3,$1F,$F7,$83,$7E,$07,$5D,$92,$30,$43,$FC,$8D,$3C
	.dc.b	$9B,$F2,$3D,$48,$F5,$F8,$7A,$E8,$F0,$47,$8E,$3E,$67,$D3,$BC,$CF
	.dc.b	$37,$A1,$2F,$F0,$33,$96,$CF,$02,$3E,$BE,$EA,$FF,$96,$E7,$83,$67
	.dc.b	$E1,$4F,$DB,$FA,$67,$D3,$D2,$57,$E4,$21,$F3,$78,$8F,$F3,$34,$70
	.dc.b	$75,$3C,$E7,$A7,$CF,$75,$FA,$49,$5C,$E3,$2F,$58,$CB,$D2,$3F,$1D
	.dc.b	$E3,$FF,$96,$7B,$3F,$DE,$9F,$4C,$FA,$7F,$C9,$36,$FF,$4B,$5B,$ED
	.dc.b	$FF,$3F,$B6,$07,$FA,$CF,$C5,$EA,$72,$EE,$B1,$CA,$2B,$FE,$83,$BF
	.dc.b	$D4,$5F,$FA,$29,$76,$F8,$89,$3F,$B6,$8F,$D1,$B6,$32,$F3,$7D,$3D
	.dc.b	$BE,$AE,$AF,$5F,$C8,$CA,$E5,$19,$7A,$46,$5F,$24,$7C,$6E,$BF,$DA
	.dc.b	$7A,$EF,$A7,$A6,$6B,$E2,$CD,$7F,$D3,$DB,$7E,$81,$FF,$50,$57,$E6
	.dc.b	$F8,$1A,$FC,$AB,$E1,$18,$AF,$97,$7C,$63,$10,$BE,$57,$FE,$B0,$FA
	.dc.b	$77,$1F,$F0,$96,$D7,$F1,$C6,$BF,$EB,$15,$F5,$3B,$F4,$A4,$ED,$FE
	.dc.b	$25,$F5,$0E,$EA,$49,$FE,$4F,$19,$7F,$94,$FD,$4F,$DA,$7F,$07,$0F
	.dc.b	$8B,$E6,$48,$FE,$5F,$F5,$3F,$CD,$9F,$C4,$B5,$34,$C7,$D4,$BE,$4B
	.dc.b	$F9,$33,$67,$9F,$27,$F2,$9A,$DD,$7F,$2B,$F6,$3F,$5B,$5F,$CF,$AF
	.dc.b	$FD,$74,$DF,$63,$5D,$F2,$3F,$A1,$D7,$07,$FD,$B0,$DE,$4F,$A7,$7F
	.dc.b	$E4,$77,$EA,$C6,$78,$FF,$DC,$6D,$AF,$D2,$B5,$57,$19,$FF,$74,$FC
	.dc.b	$17,$AE,$7F,$DC,$DF,$46,$F7,$FF,$A2,$36,$3F,$AB,$CE,$F1,$FC,$2B
	.dc.b	$6F,$9E,$9F,$D2,$E3,$15,$FB,$6B,$7C,$3E,$6A,$4F,$47,$BC,$7F,$53
	.dc.b	$F0,$DF,$53,$FD,$57,$E9,$DF,$1B,$4D,$44,$C3,$FB,$5D,$70,$6C,$71
	.dc.b	$F8,$FD,$43,$5F,$CE,$D6,$E3,$1A,$FF,$0C,$51,$99,$FD,$1E,$2B,$FE
	.dc.b	$D1,$F5,$0F,$A1,$F3,$7D,$4C,$C7,$FB,$EC,$E5,$B3,$C0,$8F,$F8,$0A
	.dc.b	$FF,$88,$BA,$21,$E4,$3F,$8F,$CF,$37,$02,$38,$B8,$FA,$8F,$C0,$47
	.dc.b	$F8,$36,$3B,$7B,$1F,$DF,$E7,$F0,$84,$E6,$C1,$F3,$FE,$A6,$D7,$F8
	.dc.b	$5A,$74,$F9,$E3,$97,$C3,$7D,$5F,$6A,$4E,$E3,$2D,$21,$E8,$C7,$F6
	.dc.b	$1F,$50,$E6,$FF,$09,$36,$7F,$C4,$AD,$7D,$FF,$50,$F6,$3E,$D4,$3E
	.dc.b	$9D,$F5,$1F,$66,$F8,$72,$7F,$1A,$6F,$36,$72,$FF,$10,$65,$CE,$BF
	.dc.b	$A5,$E1,$CB,$1F,$E9,$B3,$97,$B6,$3C,$5D,$B5,$C9,$C3,$2F,$2D,$73
	.dc.b	$85,$73,$8C,$65,$ED,$6E,$3F,$91,$78,$4E,$F0,$FC,$A3,$6F,$FB,$0D
	.dc.b	$F1,$9F,$15,$70,$C3,$C3,$1C,$1B,$5C,$30,$ED,$3C,$0D,$63,$EC,$9F
	.dc.b	$53,$56,$1D,$DE,$46,$71,$F2,$3E,$EC,$3B,$BC,$30,$EE,$F0,$3E,$39
	.dc.b	$BB,$E5,$3C,$2F,$B4,$F1,$33,$31,$FF,$75,$1C,$4B,$7F,$5D,$FF,$77
	.dc.b	$35,$F9,$3E,$2B,$E4,$F8,$C7,$D4,$8C,$3C,$9A,$F2,$6B,$E0,$F8,$3E
	.dc.b	$2F,$F8,$49,$A9,$A4,$7D,$FF,$20,$FA,$6F,$EE,$1B,$7E,$16,$79,$1A
	.dc.b	$8C,$ED,$DE,$31,$BF,$D3,$77,$5F,$8D,$BE,$93,$99,$F1,$1C,$0F,$9E
	.dc.b	$70,$AF,$E3,$D3,$C2,$32,$70,$AE,$97,$E2,$8F,$E4,$86,$6B,$8C,$66
	.dc.b	$3F,$94,$61,$F8,$7B,$E9,$7E,$D9,$E0,$7F,$8E,$23,$37,$ED,$9E,$F8
	.dc.b	$7A,$6F,$96,$1D,$70,$F4,$1C,$5C,$3D,$06,$1D,$1D,$57,$FA,$BA,$3D
	.dc.b	$DA,$7F,$87,$27,$99,$C2,$BF,$C4,$51,$FE,$20,$9D,$D7,$C6,$F1,$B2
	.dc.b	$FC,$03,$A3,$8B,$F6,$74,$DB,$7D,$47,$F4,$7B,$F1,$4E,$63,$15,$FC
	.dc.b	$EB,$0F,$E9,$55,$FE,$C7,$BE,$3D,$9D,$6B,$DB,$87,$18,$C5,$F2,$AF
	.dc.b	$44,$78,$FE,$35,$F0,$5F,$DE,$F1,$F8,$53,$C0,$D7,$D4,$3E,$23,$13
	.dc.b	$FE,$D7,$B4,$E6,$38,$99,$71,$9F,$A6,$F9,$DF,$9F,$E0,$1D,$54,$F8
	.dc.b	$D8,$3E,$9D,$96,$63,$FD,$C7,$5C,$1F,$F7,$41,$BE,$1D,$65,$FD,$8E
	.dc.b	$30,$70,$C3,$AE,$1F,$62,$7F,$D0,$B7,$CD,$E1,$19,$5F,$50,$F6,$BE
	.dc.b	$73,$70,$F0,$D3,$8F,$35,$27,$2D,$BE,$AD,$E0,$FF,$36,$4F,$BE,$30
	.dc.b	$FF,$A6,$0C,$1F,$E9,$A8,$FF,$37,$BF,$E6,$13,$07,$F8,$4C,$C4,$FA
	.dc.b	$BB,$FF,$09,$18,$9E,$3D,$1F,$3B,$EA,$3F,$CB,$13,$CA,$FC,$13,$9F
	.dc.b	$2B,$8B,$67,$53,$FE,$86,$8C,$3F,$E0,$EF,$9B,$FC,$74,$FF,$89,$5B
	.dc.b	$1F,$E8,$CD,$18,$7E,$5E,$73,$27,$75,$F8,$1F,$17,$5C,$9F,$61,$B1
	.dc.b	$F4,$1B,$5F,$61,$FD,$17,$D3,$3D,$0A,$7E,$97,$EA,$1F,$51,$8C,$3D
	.dc.b	$75,$B6,$FB,$D9,$C7,$FC,$53,$28,$EA,$7E,$BF,$EA,$1C,$D8,$BF,$8B
	.dc.b	$18,$7A,$F0,$FE,$D7,$A7,$FD,$93,$18,$3E,$5B,$8C,$7F,$C6,$ED,$E3
	.dc.b	$FF,$56,$BF,$E5,$C9,$DF,$C1,$73,$F4,$7F,$80,$7D,$47,$F9,$A2,$7E
	.dc.b	$AF,$7B,$DC,$CF,$A9,$FC,$16,$31,$3F,$6B,$E9,$7F,$D4,$4D,$F6,$FF
	.dc.b	$8B,$7D,$43,$EA,$1F,$51,$FC,$36,$F7,$C1,$FE,$4C,$6F,$07,$FB,$17
	.dc.b	$C7,$7D,$47,$F9,$D3,$3E,$87,$E3,$63,$3E,$D7,$3D,$3F,$92,$46,$27
	.dc.b	$78,$FF,$96,$E3,$0F,$F9,$77,$BB,$FD,$F7,$F5,$1F,$CA,$DF,$F7,$EF
	.dc.b	$7D,$94,$D8,$C3,$6F,$F3,$D5,$C9,$B5,$F9,$68,$2F,$A8,$E8,$75,$E5
	.dc.b	$B5,$C2,$2C,$F9,$4F,$AC,$57,$F9,$A7,$D3,$3D,$8A,$7F,$31,$FA,$8F
	.dc.b	$A7,$5F,$E7,$69,$D5,$FE,$B9,$B1,$FE,$7D,$EA,$57,$F1,$7E,$1F,$F7
	.dc.b	$8D,$69,$FF,$7B,$4E,$BC,$BE,$F8,$7F,$37,$BE,$0F,$FA,$66,$F8,$3F
	.dc.b	$02,$B8,$3F,$1B,$D4,$3F,$3D,$FC,$76,$3F,$4D,$CC,$6F,$A7,$4E,$DA
	.dc.b	$E4,$E9,$F5,$1C,$CE,$9F,$17,$63,$24,$BE,$D1,$97,$E3,$7C,$87,$CC
	.dc.b	$3F,$13,$F0,$4F,$96,$C5,$FC,$99,$FE,$DD,$F5,$1F,$3B,$83,$E3,$C6
	.dc.b	$3C,$0D,$E3,$FE,$A9,$95,$6E,$FE,$35,$ED,$F8,$9F,$1D,$F0,$CF,$89
	.dc.b	$F5,$16,$32,$4C,$3E,$34,$65,$F9,$1F,$2F,$F2,$FE,$23,$E6,$36,$BE
	.dc.b	$83,$67,$71,$F4,$E7,$FB,$F7,$D4,$7F,$77,$9F,$F5,$EB,$73,$FD,$96
	.dc.b	$DF,$C6,$E7,$35,$C0,$FF,$00,$7D,$43,$E5,$B1,$8F,$F7,$B8,$C3,$FC
	.dc.b	$FB,$1E,$8F,$E1,$63,$37,$D2,$38,$57,$E1,$7E,$9F,$F0,$DF,$51,$67
	.dc.b	$F1,$4D,$AF,$C4,$37,$3F,$C6,$C7,$F4,$1E,$64,$FE,$3F,$8B,$F2,$1D
	.dc.b	$57,$0F,$CA,$36,$FF,$24,$DC,$39,$F7,$7F,$9E,$63,$FC,$49,$F4,$FF
	.dc.b	$0F,$E0,$BE,$05,$FE,$FA,$1F,$21,$E2,$9C,$CA,$78,$6D,$FE,$FF,$77
	.dc.b	$97,$F9,$EE,$31,$EB,$71,$75,$85,$BE,$D3,$5E,$91,$6F,$9C,$7F,$A0
	.dc.b	$EB,$8D,$F5,$8E,$2D,$78,$B6,$78,$E1,$E8,$5F,$51,$D7,$0E,$CF,$7F
	.dc.b	$3B,$C5,$F4,$EE,$DF,$F1,$6E,$5E,$13,$E2,$4E,$A3,$38,$F5,$8C,$78
	.dc.b	$1E,$16,$D7,$11,$7D,$67,$D0,$D8,$D6,$5F,$E5,$0C,$F3,$8F,$6D,$3F
	.dc.b	$98,$D7,$28,$C3,$FC,$D2,$F9,$65,$E3,$BE,$38,$CA,$F1,$56,$B0,$FB
	.dc.b	$19,$79,$B0,$FB,$B3,$9A,$E9,$8F,$67,$F0,$15,$BB,$F1,$D6,$E7,$C9
	.dc.b	$87,$C1,$C7,$C1,$8F,$4C,$79,$5E,$F2,$F7,$61,$E4,$9C,$93,$0E,$51
	.dc.b	$97,$9E,$3E,$09,$F2,$57,$2C,$3B,$56,$AF,$D2,$F2,$8C,$C7,$32,$3F
	.dc.b	$1D,$B7,$D6,$79,$47,$C0,$C6,$E3,$EF,$3D,$59,$7C,$39,$EE,$72,$3F
	.dc.b	$0F,$3C,$32,$F6,$47,$DE,$9E,$0F,$C8,$38,$1D,$CB,$5D,$E7,$DF,$3C
	.dc.b	$21,$F5,$1F,$E2,$EA,$FF,$7D,$C6,$3D,$5F,$F7,$AB,$F8,$D8,$FF,$68
	.dc.b	$B9,$77,$3C,$73,$FC,$0B,$73,$FB,$AD,$9F,$F0,$0F,$D4,$3E,$A3,$FD
	.dc.b	$AF,$5A,$C7,$F3,$77,$BB,$FC,$FB,$6F,$84,$6A,$63,$F9,$49,$D5,$F4
	.dc.b	$8F,$CE,$47,$3C,$7D,$18,$F9,$AF,$F8,$06,$1D,$2F,$94,$74,$9D,$47
	.dc.b	$EC,$67,$51,$82,$63,$E9,$9D,$63,$E1,$7D,$37,$CE,$75,$5F,$C8,$AF
	.dc.b	$9C,$60,$95,$E8,$AF,$2D,$F3,$C3,$D3,$5A,$AE,$A7,$80,$EC,$71,$3D
	.dc.b	$66,$4B,$87,$A1,$D3,$E3,$AE,$D8,$FC,$72,$61,$E2,$AE,$A7,$82,$B9
	.dc.b	$2C,$7D,$6F,$4C,$3C,$D8,$FA,$23,$F0,$38,$F6,$C3,$9C,$E6,$79,$4E
	.dc.b	$5F,$3D,$7B,$EF,$AC,$78,$63,$AE,$3F,$80,$8E,$6F,$A1,$F8,$18,$F9
	.dc.b	$B0,$EC,$77,$BE,$6E,$2B,$84,$FC,$3A,$E7,$8F,$C0,$BF,$09,$C2,$B7
	.dc.b	$87,$F2,$C8,$CE,$3F,$81,$6D,$FE,$3B,$EA,$31,$F0,$D2,$3F,$22,$FE
	.dc.b	$1F,$EA,$3F,$DB,$D1,$82,$1F,$91,$8D,$7C,$2C,$8C,$E9,$F8,$B9,$C4
	.dc.b	$7F,$47,$AE,$6B,$A3,$FC,$06,$A3,$31,$86,$F0,$FF,$46,$7F,$DE,$93
	.dc.b	$FE,$D4,$F8,$07,$8B,$FD,$2A,$F9,$C3,$EA,$31,$87,$F4,$E8,$C5,$F1
	.dc.b	$BC,$BF,$E8,$28,$E6,$F5,$AC,$A8,$E2,$FA,$6B,$A5,$6A,$5E,$DD,$FA
	.dc.b	$09,$7E,$78,$D1,$87,$91,$E1,$8C,$BF,$EF,$0F,$A8,$7D,$47,$5F,$46
	.dc.b	$4E,$63,$CB,$3A,$8C,$10,$EA,$7A,$78,$37,$32,$35,$AF,$C8,$9F,$7F
	.dc.b	$D4,$3A,$6F,$C1,$4B,$FF,$53,$DF,$D8,$AD,$46,$37,$FE,$F7,$18,$8F
	.dc.b	$0B,$FE,$EB,$74,$7F,$AC,$1C,$3A,$BF,$0F,$CA,$48,$D8,$94,$79,$BE
	.dc.b	$27,$53,$DA,$2E,$1E,$59,$FB,$3F,$50,$F3,$BD,$BE,$94,$C3,$AC,$E5
	.dc.b	$B3,$C0,$8F,$8B,$1F,$EA,$8F,$CC,$9E,$0D,$9D,$BF,$03,$BB,$F1,$3F
	.dc.b	$50,$FA,$8A,$49,$F6,$C3,$E6,$F1,$1F,$62,$38,$3A,$9E,$73,$D3,$E7
	.dc.b	$BA,$FD,$2C,$AE,$71,$97,$AC,$65,$E9,$1F,$9C,$F1,$7F,$CC,$3B,$FE
	.dc.b	$CB,$E9,$DF,$51,$FA,$06,$DF,$D2,$AD,$F2,$7F,$8E,$1B,$1F,$EE,$A6
	.dc.b	$BF,$E4,$FF,$17,$97,$77,$76,$47,$37,$2F,$DE,$BF,$D4,$B7,$B7,$C4
	.dc.b	$78,$FB,$6C,$64,$87,$E7,$A5,$1E,$9E,$AF,$57,$C9,$CA,$ED,$19,$7C
	.dc.b	$11,$97,$94,$7C,$5E,$BF,$E0,$3E,$CB,$EA,$29,$9A,$F8,$93,$5F,$D7
	.dc.b	$EB,$F9,$CE,$0F,$E3,$B7,$FA,$37,$F8,$FC,$67,$7F,$95,$7C,$67,$7D
	.dc.b	$BF,$C4,$3E,$A1,$C9,$49,$97,$E9,$21,$F5,$1B,$F5,$24,$E5,$FE,$33
	.dc.b	$F4,$EF,$02,$92,$7F,$5B,$19,$7F,$5F,$F5,$1F,$C8,$9F,$E1,$30,$F8
	.dc.b	$9E,$64,$8F,$DA,$FD,$47,$F3,$64,$1C,$37,$8F,$F9,$5C,$63,$EA,$1F
	.dc.b	$27,$FC,$B9,$B5,$D2,$B6,$66,$7F,$99,$FB,$1F,$A7,$9F,$E0,$B5,$FE
	.dc.b	$BC,$6F,$B1,$AF,$02,$47,$F0,$AA,$E0,$FF,$24,$6F,$1F,$F5,$2B,$FD
	.dc.b	$EB,$FD,$52,$33,$CB,$FE,$96,$6D,$7B,$9A,$AB,$8C,$FF,$A7,$3E,$07
	.dc.b	$D7,$3F,$CC,$BE,$91,$F0,$5F,$D3,$1B,$1F,$C7,$A7,$78,$FF,$64,$6D
	.dc.b	$F3,$D3,$FA,$9C,$62,$BE,$7D,$BE,$1B,$35,$27,$A3,$DE,$3F,$AD,$F8
	.dc.b	$8F,$8E,$FE,$BB,$F4,$AF,$8C,$A6,$A2,$61,$FC,$CA,$B8,$36,$38,$FC
	.dc.b	$76,$A3,$FA,$7C,$E9,$B3,$A7,$F9,$DF,$D3,$BE,$87,$C3,$F5,$12,$73
	.dc.b	$1F,$DC,$27,$51,$82,$1F,$D7,$32,$FE,$0E,$FF,$5E,$8D,$B6,$79,$1E
	.dc.b	$EF,$23,$FB,$DF,$D4,$7D,$8B,$FE,$F9,$5F,$D3,$9A,$FF,$68,$8F,$BA
	.dc.b	$4E,$6E,$27,$C8,$FA,$0B,$3F,$D5,$A7,$4F,$38,$EF,$F0,$DF,$3B,$D9
	.dc.b	$93,$B8,$CB,$48,$7B,$7A,$FF,$C2,$3F,$50,$E5,$A4,$9F,$EC,$BF,$51
	.dc.b	$FD,$93,$A7,$F6,$3F,$4E,$FA,$8F,$F7,$6E,$1C,$6F,$FB,$5B,$5F,$F0
	.dc.b	$2B,$6B,$FC,$67,$7C,$E5,$7C,$DF,$DE,$C6,$63,$10,$AE,$91,$86,$DF
	.dc.b	$FB,$FA,$B8,$37,$DB,$FF,$21,$5F,$68,$F3,$E3,$FC,$2A,$30,$7F,$0C
	.dc.b	$BE,$6F,$8A,$B8,$E1,$E2,$C3,$AC,$EE,$F9,$4F,$8D,$B8,$78,$71,$F8
	.dc.b	$ED,$AE,$E7,$AE,$76,$F3,$36,$78,$66,$AB,$F0,$BC,$4C,$E3,$E8,$75
	.dc.b	$8F,$A3,$0F,$06,$3E,$59,$DE,$3E,$07,$78,$F6,$C7,$AE,$3C,$CE,$8F
	.dc.b	$80,$FB,$13,$F2,$4E,$25,$29,$97,$6F,$D5,$F1,$5F,$1B,$B8,$F6,$46
	.dc.b	$1E,$53,$FD,$2E,$74,$D8,$E0,$3E,$0E,$B4,$CB,$82,$3F,$43,$E6,$1F
	.dc.b	$4C,$F0,$7E,$6C,$FF,$4C,$AF,$9F,$5A,$9C,$CF,$5D,$7F,$27,$39,$8F
	.dc.b	$B6,$71,$3B,$5F,$57,$27,$1D,$BF,$21,$3C,$6B,$F3,$B3,$C5,$EB,$5C
	.dc.b	$5C,$C7,$13,$FB,$0C,$F1,$AF,$A1,$87,$A2,$F8,$46,$2B,$C7,$87,$AA
	.dc.b	$B4,$D8,$C9,$E6,$8C,$B4,$CE,$1E,$28,$CB,$E6,$9F,$F3,$4E,$1F,$0F
	.dc.b	$5F,$F5,$7D,$FA,$F0,$F7,$E3,$F0,$E7,$FD,$67,$5B,$BF,$87,$C1,$97
	.dc.b	$67,$47,$18,$7C,$0E,$8B,$6F,$A8,$77,$6A,$98,$FF,$41,$C7,$FD,$07
	.dc.b	$7B,$AD,$4E,$6B,$FD,$4B,$39,$AF,$5E,$1D,$B0,$EB,$1E,$EC,$38,$9E
	.dc.b	$4F,$8C,$FF,$4D,$78,$98,$9D,$1F,$E9,$28,$7D,$43,$E2,$35,$9F,$FA
	.dc.b	$D2,$73,$18,$CF,$F1,$11,$8B,$E7,$5B,$BE,$51,$FE,$98,$F7,$DD,$3F
	.dc.b	$05,$B1,$FE,$A2,$1F,$4E,$C3,$55,$27,$53,$F4,$F0,$FF,$16,$DF,$27
	.dc.b	$FD,$93,$3C,$B8,$78,$46,$2B,$82,$AE,$19,$7E,$86,$F8,$1F,$EA,$EF
	.dc.b	$A8,$7B,$5F,$5D,$B1,$E2,$E3,$C4,$C8,$F2,$B7,$0F,$03,$6F,$B6,$3C
	.dc.b	$A7,$FD,$A7,$85,$9C,$D7,$E1,$6F,$FC,$AC,$7F,$B0,$0B,$5E,$BE,$ED
	.dc.b	$75,$49,$F9,$5F,$51,$FE,$4A,$78,$47,$E5,$14,$66,$BF,$DC,$71,$D3
	.dc.b	$0F,$85,$39,$7F,$82,$4F,$F1,$1A,$E5,$39,$AF,$EA,$9F,$27,$F9,$0A
	.dc.b	$FF,$76,$B6,$FD,$F5,$F9,$F8,$C5,$70,$E4,$F2,$68,$F8,$AD,$46,$1B
	.dc.b	$3F,$80,$6E,$3F,$6D,$FD,$57,$D3,$3D,$0A,$7E,$B7,$EA,$1F,$51,$8C
	.dc.b	$3F,$DD,$35,$B6,$FB,$B9,$C5,$B7,$C0,$A4,$F6,$3F,$71,$F5,$1F,$89
	.dc.b	$BE,$27,$FB,$D1,$C5,$F2,$8F,$EF,$15,$F1,$E3,$FD,$07,$D0,$F2,$FD
	.dc.b	$3E,$97,$F0,$38,$C5,$FF,$BF,$9B,$5C,$CE,$3E,$8E,$EF,$83,$FC,$2F
	.dc.b	$EA,$3F,$CC,$17,$FE,$90,$D3,$EC,$FB,$EF,$73,$5E,$AF,$F1,$18,$C5
	.dc.b	$6C,$FF,$07,$7D,$2F,$FA,$75,$C6,$7F,$E9,$A8,$C3,$E3,$FA,$87,$D4
	.dc.b	$3E,$A3,$F8,$E5,$6B,$3F,$C4,$B7,$0F,$E8,$7F,$51,$FE,$6F,$7A,$E2
	.dc.b	$9D,$3F,$C9,$27,$4F,$C1,$35,$EF,$FF,$2E,$F2,$D3,$AC,$3B,$CE,$63
	.dc.b	$18,$7F,$AC,$BB,$7F,$D6,$7F,$51,$FC,$D9,$FF,$5A,$F6,$DC,$93,$63
	.dc.b	$0D,$AF,$87,$5C,$9B,$5F,$9C,$11,$FE,$C0,$FA,$8E,$87,$5E,$4B,$5C
	.dc.b	$22,$D7,$DD,$AF,$F6,$3D,$63,$3F,$D8,$FD,$3B,$D7,$FE,$90,$DC,$7F
	.dc.b	$9C,$7D,$47,$DA,$C7,$E3,$4E,$5B,$5F,$D0,$67,$F8,$FD,$7F,$92,$EB
	.dc.b	$57,$FC,$EA,$73,$7F,$ED,$89,$CD,$7E,$62,$73,$29,$CB,$FC,$DE,$38
	.dc.b	$E3,$FE,$28,$AF,$9D,$EA,$FE,$52,$BD,$CF,$F8,$6E,$3E,$5B,$9C,$3F
	.dc.b	$C2,$55,$C6,$73,$CD,$C9,$E6,$F4,$FA,$89,$1C,$BE,$2A,$E3,$DA,$6E
	.dc.b	$1D,$31,$56,$2F,$E1,$7C,$87,$CC,$3F,$19,$F0,$6F,$97,$C4,$FE,$5B
	.dc.b	$65,$7F,$7C,$FA,$8F,$A1,$97,$D6,$8C,$76,$F0,$D7,$FB,$BD,$6B,$0E
	.dc.b	$E7,$F7,$E7,$11,$F8,$8D,$1F,$1D,$F2,$0F,$93,$F5,$16,$FE,$3C,$5A
	.dc.b	$F9,$73,$72,$F7,$7C,$BF,$D0,$F6,$3E,$8B,$70,$F9,$ED,$BF,$A6,$DA
	.dc.b	$FE,$7F,$F5,$1F,$EF,$4A,$FD,$AB,$6B,$F8,$97,$6B,$87,$F9,$54,$7E
	.dc.b	$D6,$75,$7C,$B1,$FF,$08,$7D,$43,$E5,$BF,$C4,$15,$AC,$BF,$91,$CE
	.dc.b	$5E,$F5,$C2,$B6,$D8,$FC,$3F,$D4,$7E,$33,$EA,$3F,$1A,$D8,$C9,$F8
	.dc.b	$C8,$CB,$4F,$C7,$F1,$7E,$53,$DB,$FC,$AB,$63,$27,$E5,$23,$2D,$3F
	.dc.b	$8B,$FD,$3F,$C0,$F8,$3F,$66,$C1,$0F,$90,$FF,$36,$4E,$A7,$3E,$A6
	.dc.b	$35,$FF,$5E,$5E,$F0,$FF,$3E,$46,$27,$55,$D6,$B4,$FF,$A2,$E7,$55
	.dc.b	$C2,$35,$87,$07,$C6,$DA,$E3,$87,$8D,$F5,$5F,$8A,$74,$6F,$EA,$2E
	.dc.b	$1D,$70,$F6,$C7,$3F,$33,$B7,$D4,$7E,$B3,$D0,$D6,$49,$CF,$7F,$F8
	.dc.b	$EF,$0F,$41,$EB,$8C,$D7,$06,$C6,$85,$F5,$70,$DD,$3D,$53,$C1,$F9
	.dc.b	$D1,$9D,$3F,$DD,$0D,$AF,$06,$7E,$B8,$C4,$76,$C7,$CD,$7C,$9B,$7F
	.dc.b	$5B,$0E,$CF,$48,$CD,$7E,$06,$F9,$E5,$F0,$71,$ED,$5A,$AE,$D9,$60
	.dc.b	$F8,$13,$63,$DB,$3B,$BE,$78,$F8,$27,$B4,$E4,$E9,$87,$78,$E3,$87
	.dc.b	$3C,$BD,$4F,$68,$C9,$23,$2F,$CF,$9E,$51,$82,$65,$EF,$AE,$51,$8A
	.dc.b	$F1,$C7,$79,$F8,$D9,$59,$C9,$DE,$32,$D3,$35,$CD,$FD,$4E,$3E,$F7
	.dc.b	$E3,$9B,$33,$5D,$5C,$CF,$1C,$3D,$37,$E7,$9F,$C3,$FD,$47,$F9,$FD
	.dc.b	$FF,$80,$19,$A3,$11,$FB,$93,$CE,$53,$F2,$2E,$9F,$C5,$BC,$EF,$F1
	.dc.b	$ED,$AC,$FC,$07,$E9,$FE,$9D,$F5,$1D,$6F,$73,$F9,$1B,$E1,$87,$E2
	.dc.b	$5A,$AF,$8B,$B9,$8F,$B1,$ED,$7D,$70,$F0,$CE,$49,$87,$F1,$1B,$E7
	.dc.b	$18,$9D,$CF,$92,$3F,$61,$3A,$8C,$12,$76,$7C,$78,$D3,$89,$F5,$5F
	.dc.b	$9A,$BC,$17,$E5,$9C,$9E,$12,$C6,$9E,$15,$CD,$FA,$35,$EE,$72,$70
	.dc.b	$8D,$5F,$8A,$30,$43,$89,$EC,$BF,$13,$8B,$F6,$3B,$9E,$4E,$6B,$C5
	.dc.b	$8F,$E1,$F0,$F1,$4F,$A4,$D1,$C7,$0E,$B3,$B8,$D3,$70,$F7,$46,$31
	.dc.b	$FB,$F3,$9C,$3E,$65,$F2,$9C,$93,$0F,$0E,$1E,$8C,$3C,$24,$C7,$D3
	.dc.b	$8F,$38,$F8,$4D,$7F,$19,$87,$5B,$F9,$71,$C6,$30,$F9,$63,$BE,$3F
	.dc.b	$0B,$1F,$14,$FC,$B3,$27,$67,$F3,$38,$7C,$A2,$5F,$D0,$71,$F5,$17
	.dc.b	$E1,$A5,$FE,$9D,$FA,$BF,$50,$F8,$5C,$CF,$E5,$26,$EB,$FD,$AE,$B5
	.dc.b	$3F,$D5,$5A,$EC,$E8,$66,$BA,$9F,$BF,$ED,$FF,$6E,$CF,$06,$C6,$78
	.dc.b	$3F,$82,$4B,$E1,$31,$7F,$DF,$FE,$03,$FC,$86,$FF,$A6,$97,$D4,$6A
	.dc.b	$3F,$DB,$8D,$8D,$C7,$FA,$6D,$B8,$FF,$BB,$9E,$31,$CE,$78,$79,$1C
	.dc.b	$7D,$BE,$5F,$41,$2B,$C1,$19,$7C,$51,$97,$B4,$78,$FE,$A1,$F5,$1E
	.dc.b	$3F,$4E,$57,$FA,$CA,$72,$D9,$E0,$47,$DD,$BD,$7F,$D6,$93,$C1,$B3
	.dc.b	$EF,$9F,$85,$F5,$1F,$E0,$39,$FF,$5F,$BF,$EE,$C6,$FC,$1E,$BB,$24
	.dc.b	$60,$87,$FB,$06,$79,$37,$E5,$3A,$91,$EB,$F1,$35,$D1,$E0,$8F,$1C
	.dc.b	$7D,$5F,$A8,$79,$DE,$6F,$52,$5F,$FB,$52,$72,$D9,$E0,$47,$D7,$DF
	.dc.b	$5F,$F6,$B4,$F0,$6C,$FC,$29,$FC,$27,$D4,$3E,$A2,$92,$BF,$DC,$70
	.dc.b	$F9,$9C,$47,$FB,$86,$38,$3A,$9E,$73,$D3,$E7,$BA,$FD,$2C,$AE,$71
	.dc.b	$97,$AC,$65,$E9,$1F,$97,$F2,$3F,$CA,$DD,$6E,$EF,$EC,$9F,$4E,$FA
	.dc.b	$8F,$CE,$37,$0F,$F7,$CD,$F0,$AF,$ED,$4D,$8F,$F3,$23,$5F,$F5,$8F
	.dc.b	$8F,$CB,$B7,$C0,$C8,$E6,$E5,$FF,$36,$63,$E3,$7F,$CD,$D8,$3B,$3C
	.dc.b	$64,$8D,$E3,$E8,$ED,$C4,$2F,$92,$BE,$4F,$03,$AD,$FB,$AB,$DF,$D5
	.dc.b	$F0,$3C,$09,$5B,$39,$46,$9C,$5F,$F0,$76,$CE,$5F,$16,$3F,$1B,$AF
	.dc.b	$F8,$37,$B0,$FA,$8A,$7D,$38,$95,$F3,$9A,$F2,$3F,$D2,$13,$FA,$4D
	.dc.b	$3D,$BB,$FE,$B6,$74,$7F,$B8,$2B,$8C,$FC,$2A,$E1,$87,$F8,$87,$EA
	.dc.b	$1C,$7F,$C6,$5B,$87,$BD,$AF,$45,$F5,$1B,$F4,$E4,$ED,$FE,$45,$F4
	.dc.b	$EE,$EA,$49,$F9,$51,$97,$E6,$7D,$47,$F2,$4C,$BF,$C6,$92,$F8,$AE
	.dc.b	$64,$AF,$A5,$F5,$1F,$CE,$6B,$FC,$74,$D8,$FF,$5A,$35,$37,$EA,$1F
	.dc.b	$25,$FC,$DD,$B3,$D3,$9B,$F9,$D5,$6E,$BF,$9E,$7B,$1F,$B5,$AF,$C1
	.dc.b	$5F,$FA,$FD,$BE,$C6,$BB,$E4,$7E,$1E,$B8,$3F,$51,$BC,$9F,$94,$BF
	.dc.b	$F2,$9B,$F9,$68,$CE,$7F,$E5,$56,$D7,$FB,$3D,$AA,$B8,$CF,$F9,$6F
	.dc.b	$DF,$F5,$CF,$E0,$BE,$8D,$EF,$7F,$55,$6C,$7E,$6E,$77,$8F,$E9,$9B
	.dc.b	$7C,$F4,$FE,$B9,$18,$AF,$F7,$05,$BE,$1F,$35,$27,$A3,$DE,$3F,$B2
	.dc.b	$78,$6F,$A8,$FE,$CD,$F4,$CF,$8C,$A6,$A2,$61,$FA,$FA,$E0,$D8,$E3
	.dc.b	$F1,$FA,$8F,$CF,$CE,$9B,$3A,$7F,$71,$F4,$EF,$A2,$F0,$FD,$44,$9C
	.dc.b	$C7,$F7,$99,$D4,$60,$87,$EB,$32,$FE,$12,$FE,$BA,$36,$D9,$E4,$7B
	.dc.b	$BC,$7F,$F0,$27,$D4,$7D,$8B,$7B,$3F,$08,$CF,$3C,$C3,$F8,$75,$7D
	.dc.b	$E3,$87,$D0,$66,$3F,$7B,$19,$9F,$97,$3D,$BE,$1B,$E6,$FB,$52,$F9
	.dc.b	$C6,$5A,$43,$D9,$DB,$FE,$19,$FA,$87,$35,$24,$FF,$21,$FA,$87,$B1
	.dc.b	$FC,$B6,$1F,$50,$FA,$8F,$F5,$66,$1C,$A7,$F9,$43,$5F,$E2,$AD,$ED
	.dc.b	$FE,$6B,$87,$45,$3D,$2B,$FD,$6D,$39,$8C,$1A,$AE,$97,$E1,$8C,$63
	.dc.b	$FE,$BD,$8C,$42,$B9,$46,$1B,$7E,$8A,$E0,$DA,$F1,$D6,$F1,$F8,$0D
	.dc.b	$C7,$35,$BA,$F2,$8C,$7C,$B1,$EA,$BE,$2D,$AE,$35,$E8,$29,$C7,$0F
	.dc.b	$1B,$F1,$31,$F3,$B6,$F9,$BC,$4F,$0E,$1E,$07,$57,$D2,$B8,$C7,$53
	.dc.b	$BC,$ED,$F8,$D5,$D1,$B5,$AB,$F1,$C7,$13,$DF,$7D,$0F,$AC,$FF,$BA
	.dc.b	$DA,$FB,$CF,$F7,$71,$D0,$AF,$F4,$7F,$91,$CC,$7D,$7F,$1D,$18,$9D
	.dc.b	$4F,$09,$CB,$C0,$7C,$15,$69,$97,$04,$7E,$BF,$C8,$3E,$9B,$FC,$21
	.dc.b	$B7,$E2,$27,$F2,$75,$F5,$2B,$5E,$17,$D6,$9C,$9C,$AF,$8E,$FE,$EB
	.dc.b	$E1,$18,$9F,$CE,$9C,$30,$FC,$AE,$1D,$23,$F1,$26,$6F,$71,$99,$F2
	.dc.b	$4F,$B1,$E3,$5D,$63,$AD,$F8,$30,$F3,$46,$F0,$E7,$5B,$8E,$C7,$EC
	.dc.b	$30,$ED,$3E,$98,$F6,$57,$2B,$E9,$1E,$63,$A1,$C6,$FA,$38,$17,$D1
	.dc.b	$D5,$7F,$8A,$E3,$E0,$6B,$FE,$36,$9F,$19,$C2,$BF,$C8,$51,$FE,$3A
	.dc.b	$9D,$D7,$C4,$E0,$CB,$EC,$3A,$38,$BF,$67,$4D,$B7,$D4,$3B,$F5,$5E
	.dc.b	$6A,$D4,$E7,$0F,$34,$62,$B7,$8F,$F4,$B9,$FF,$25,$61,$D2,$72,$48
	.dc.b	$FF,$27,$7C,$6B,$E0,$BF,$B9,$C7,$C1,$9D,$65,$FE,$3D,$A7,$D4,$3E
	.dc.b	$23,$38,$FF,$98,$E7,$31,$8C,$3F,$93,$C6,$1E,$87,$82,$B9,$3F,$E5
	.dc.b	$9F,$80,$74,$FC,$06,$C7,$F9,$8C,$7D,$32,$F5,$52,$75,$3F,$4B,$2F
	.dc.b	$E6,$F7,$C9,$FF,$2B,$4F,$2E,$1E,$11,$8A,$E0,$AB,$86,$7F,$E7,$FB
	.dc.b	$E0,$7F,$9B,$FE,$A3,$FD,$E4,$DF,$6F,$EE,$3E,$6E,$3F,$B3,$3A,$C3
	.dc.b	$FC,$C2,$DB,$E9,$8D,$BF,$D8,$CD,$AF,$F5,$04,$D5,$19,$9F,$87,$87
	.dc.b	$63,$C3,$5B,$8F,$F2,$F1,$FE,$85,$2D,$7B,$FB,$BF,$1B,$18,$C3,$FD
	.dc.b	$26,$D3,$FA,$F8,$FA,$8F,$F2,$AB,$CF,$A7,$DF,$18,$BF,$F3,$EC,$F1
	.dc.b	$7F,$D7,$11,$9A,$EB,$5C,$3C,$BF,$CE,$FC,$DF,$E6,$EF,$FD,$0A,$DA
	.dc.b	$F8,$B7,$AE,$49,$8F,$DE,$F1,$2E,$1F,$65,$B7,$F3,$1B,$3E,$AD,$3F
	.dc.b	$C3,$AF,$FA,$9F,$E9,$5E,$8F,$EB,$9B,$1F,$EA,$BF,$A8,$7D,$46,$6A
	.dc.b	$4F,$1F,$7A,$E7,$FE,$9C,$C1,$1C,$CF,$DC,$FD,$47,$E2,$EB,$81,$FE
	.dc.b	$B7,$71,$5C,$78,$3F,$7F,$D3,$FE,$CB,$AF,$34,$63,$CC,$91,$FE,$BF
	.dc.b	$9F,$F2,$B3,$63,$8F,$A3,$FC,$23,$EA,$3F,$CC,$53,$FE,$C5,$EF,$7B
	.dc.b	$99,$F5,$3F,$86,$C6,$27,$FD,$95,$F4,$BF,$E9,$D6,$FB,$7F,$C5,$FE
	.dc.b	$A1,$F5,$0F,$A8,$FE,$2B,$7B,$E0,$FF,$5E,$B7,$A7,$FD,$CD,$E3,$BE
	.dc.b	$A3,$FC,$DD,$9F,$43,$FD,$BD,$19,$F6,$79,$E9,$FC,$96,$31,$3B,$C7
	.dc.b	$FD,$9D,$18,$7F,$DA,$1D,$FF,$E3,$4F,$A8,$FE,$58,$FF,$8D,$BB,$EC
	.dc.b	$A6,$C6,$1B,$7F,$EE,$FA,$E4,$DA,$FC,$93,$FE,$40,$8F,$F2,$0F,$D4
	.dc.b	$74,$3A,$F2,$D9,$E1,$16,$BC,$A7,$D6,$2B,$FC,$F3,$E9,$9E,$BD,$3F
	.dc.b	$A2,$74,$BE,$A3,$EA,$5F,$FB,$A2,$B7,$87,$FB,$FD,$B1,$FC,$D6,$7F
	.dc.b	$8B,$CF,$F7,$A9,$CF,$0F,$F3,$59,$CD,$7F,$96,$A7,$35,$FC,$DA,$73
	.dc.b	$29,$CB,$FC,$BE,$38,$E7,$FE,$20,$AF,$9F,$EA,$FE,$3A,$BE,$9A,$8C
	.dc.b	$F4,$69,$A7,$CF,$8D,$36,$39,$FD,$44,$8D,$7C,$5D,$BE,$B3,$73,$ED
	.dc.b	$0C,$7F,$C6,$75,$9B,$F8,$5F,$21,$F3,$CF,$C6,$7C,$13,$E5,$B1,$3F
	.dc.b	$2E,$7F,$B9,$7D,$47,$D0,$CB,$AE,$5F,$0E,$73,$DF,$C2,$13,$FE,$02
	.dc.b	$AE,$17,$E0,$3F,$B8,$B8,$9F,$36,$CF,$8E,$F9,$27,$CA,$FA,$8B,$7F
	.dc.b	$16,$2D,$7C,$C9,$B8,$7C,$0D,$BF,$CA,$3F,$2F,$F4,$BE,$23,$E9,$B7
	.dc.b	$1F,$9C,$DC,$3E,$AB,$6F,$DB,$F5,$1F,$E0,$09,$F5,$D7,$F9,$0A,$31
	.dc.b	$3F,$BA,$E0,$FE,$4B,$5A,$3F,$D4,$73,$AA,$E1,$87,$E6,$BE,$A1,$F2
	.dc.b	$DF,$E1,$AA,$D5,$FF,$79,$9C,$DF,$FA,$B3,$1F,$1E,$1C,$70,$E0,$D9
	.dc.b	$FC,$47,$D4,$7E,$43,$EA,$3F,$22,$D8,$C9,$F8,$D8,$CB,$4F,$C8,$71
	.dc.b	$7E,$63,$DC,$FC,$CB,$63,$27,$E5,$63,$2D,$3C,$C7,$F9,$9B,$E9,$FE
	.dc.b	$1F,$C1,$7B,$B5,$FF,$62,$C3,$E4,$3F,$2B,$39,$94,$F0,$D3,$95,$69
	.dc.b	$E9,$E9,$DA,$E7,$35,$E3,$16,$BB,$61,$FE,$82,$BE,$6D,$AE,$2D,$78
	.dc.b	$B6,$78,$C7,$FA,$52,$78,$B4,$F5,$FD,$47,$7C,$3C,$0F,$87,$CE,$F2
	.dc.b	$7D,$43,$B7,$FC,$69,$7C,$4F,$7C,$EA,$33,$A7,$C3,$CB,$FC,$A1,$9F
	.dc.b	$97,$0E,$5E,$17,$97,$0E,$0E,$70,$E3,$A7,$FA,$5A,$B7,$97,$7C,$79
	.dc.b	$BE,$18,$CE,$5F,$45,$B5,$C6,$F9,$D7,$D8,$C7,$BB,$C6,$32,$AB,$8B
	.dc.b	$BC,$BC,$99,$48,$E0,$FA,$63,$18,$F8,$6F,$9D,$F2,$BF,$8B,$5A,$C7
	.dc.b	$C9,$85,$9E,$C7,$42,$61,$CE,$3B,$46,$27,$8B,$4E,$33,$B7,$07,$AF
	.dc.b	$0F,$6E,$5F,$22,$74,$4A,$E6,$FD,$7C,$BE,$2B,$9B,$1C,$F0,$F1,$C6
	.dc.b	$63,$07,$82,$FB,$5F,$42,$47,$2C,$BE,$24,$EA,$31,$8F,$C8,$26,$1D
	.dc.b	$AF,$8C,$FD,$5F,$A8,$FF,$3C,$3F,$EF,$AC,$11,$88,$FF,$6E,$9E,$42
	.dc.b	$C6,$BE,$07,$FD,$FA,$D8,$FC,$54,$F4,$CB,$F1,$4D,$BF,$F7,$3F,$D3
	.dc.b	$BE,$A3,$FD,$BB,$1F,$95,$9D,$5F,$B5,$AF,$9F,$1F,$DA,$5F,$18,$FD
	.dc.b	$14,$EB,$1F,$16,$3E,$0B,$DE,$3D,$AF,$A4,$70,$23,$FB,$1C,$3A,$3E
	.dc.b	$A8,$CD,$73,$8D,$57,$37,$C8,$6C,$91,$FC,$66,$B9,$BC,$49,$8F,$BE
	.dc.b	$FC,$B5,$DA,$30,$F2,$7E,$3D,$F7,$C3,$DB,$1A,$BE,$75,$F8,$CC,$3B
	.dc.b	$CE,$49,$1F,$00,$F4,$9D,$CC,$4F,$25,$3C,$9C,$CF,$27,$DF,$3E,$DB
	.dc.b	$F2,$63,$F4,$49,$8F,$B2,$BD,$47,$22,$1D,$A7,$8C,$78,$E3,$AD,$FC
	.dc.b	$07,$18,$FC,$07,$8E,$3E,$57,$67,$48,$D3,$D4,$E3,$87,$91,$F8,$27
	.dc.b	$90,$E2,$49,$E9,$3C,$89,$3F,$92,$8C,$BC,$F1,$F9,$11,$8C,$3B,$9F
	.dc.b	$06,$3E,$65,$72,$8C,$61,$ED,$C3,$C5,$3E,$8B,$F6,$13,$1F,$C7,$38
	.dc.b	$FA,$8B,$F0,$F2,$FF,$46,$FE,$4F,$EA,$1F,$0D,$99,$FC,$E4,$DD,$3F
	.dc.b	$B6,$56,$A7,$FA,$F3,$5D,$9D,$0C,$D7,$53,$F8,$CB,$FE,$1B,$11,$C7
	.dc.b	$AF,$FC,$82,$DE,$3D,$AF,$85,$E3,$3F,$E4,$6F,$82,$78,$F8,$BE,$B0
	.dc.b	$FA,$8C,$63,$FC,$A2,$30,$FF,$83,$0C,$56,$94,$74,$74,$62,$75,$1C
	.dc.b	$A3,$D4,$ED,$BE,$E6,$BE,$86,$47,$29,$4F,$23,$31,$CA,$70,$F8,$3E
	.dc.b	$A1,$F5,$1E,$0F,$4A,$4E,$63,$D1,$3A,$8C,$10,$F2,$EF,$A9,$91,$9D
	.dc.b	$7E,$54,$FC,$1F,$A8,$76,$DF,$83,$97,$F0,$A7,$2D,$9C,$BF,$6E,$3F
	.dc.b	$DE,$6E,$4E,$1F,$29,$D4,$8F,$3F,$8A,$AF,$18,$B8,$FF,$A0,$A9,$5F
	.dc.b	$57,$EA,$1E,$6F,$9F,$D3,$98,$FD,$69,$CB,$67,$81,$1F,$57,$7D,$7F
	.dc.b	$25,$3C,$1B,$3F,$0E,$7F,$09,$F5,$0F,$A8,$A4,$AF,$C2,$C3,$E6,$71
	.dc.b	$1F,$9D,$8E,$0E,$A7,$9C,$F4,$F9,$EE,$BF,$4B,$2B,$9C,$65,$EB,$19
	.dc.b	$7A,$47,$E5,$FC,$7F,$F2,$F6,$FF,$EA,$8F,$A7,$7D,$47,$E6,$DB,$7F
	.dc.b	$9A,$AD,$F7,$7E,$C9,$B1,$FE,$0C,$6B,$FE,$4D,$F1,$F9,$77,$F7,$64
	.dc.b	$73,$72,$FE,$9A,$FF,$3A,$DE,$DF,$11,$E3,$ED,$B1,$92,$1F,$C0,$E5
	.dc.b	$1E,$9E,$AF,$57,$81,$2B,$B4,$65,$F0,$46,$5E,$51,$F1,$7A,$FF,$79
	.dc.b	$EC,$3E,$A2,$99,$AF,$89,$35,$FE,$29,$AF,$E6,$9F,$E2,$D5,$FA,$07
	.dc.b	$33,$BE,$0F,$F6,$BD,$6D,$FF,$6C,$D6,$DF,$F6,$DF,$D4,$38,$FF,$87
	.dc.b	$37,$0F,$E1,$AD,$7F,$92,$2F,$A8,$DF,$A7,$27,$6F,$F1,$7F,$A7,$FD
	.dc.b	$2C,$7F,$16,$6B,$FC,$B5,$B8,$7F,$BB,$1B,$C9,$FC,$CB,$EA,$3F,$8C
	.dc.b	$4E,$7E,$27,$A1,$2F,$F9,$0F,$D4,$7F,$33,$7F,$CA,$2D,$4D,$31,$F5
	.dc.b	$0F,$93,$FE,$58,$D9,$E7,$C7,$FC,$BE,$B7,$5F,$CC,$7D,$8F,$D5,$57
	.dc.b	$F3,$0B,$FF,$5E,$37,$D8,$D7,$74,$8F,$E6,$B5,$C1,$FE,$98,$DC,$DA
	.dc.b	$FF,$50,$C3,$FC,$A9,$18,$AF,$F3,$5D,$6E,$BF,$AA,$B6,$33,$7E,$AF
	.dc.b	$07,$FA,$37,$BF,$EB,$9F,$EB,$5F,$46,$F7,$BF,$A5,$36,$3F,$A4,$CE
	.dc.b	$F2,$F0,$42,$31,$97,$F6,$46,$C7,$F9,$F9,$AA,$9F,$0D,$E2,$DA,$4F
	.dc.b	$67,$C3,$1F,$D6,$FC,$37,$C7,$7F,$5D,$FA,$67,$C6,$53,$51,$30,$FE
	.dc.b	$BF,$5C,$1B,$1C,$7E,$3B,$51,$FD,$E6,$74,$D9,$D3,$FD,$A3,$E9,$DF
	.dc.b	$43,$E1,$FA,$89,$39,$8F,$EE,$13,$A8,$C1,$0F,$F0,$36,$5F,$C3,$1F
	.dc.b	$F0,$44,$6D,$B3,$C8,$F7,$78,$FF,$DE,$FE,$A3,$ED,$4A,$FF,$BE,$BF
	.dc.b	$DF,$1A,$FF,$84,$E3,$64,$E5,$E4,$7C,$9F,$A0,$B3,$FE,$04,$9D,$3C
	.dc.b	$E3,$C1,$F0,$DF,$3F,$D9,$93,$B8,$CB,$48,$7B,$BB,$7F,$C2,$5F,$51
	.dc.b	$F7,$E3,$FC,$22,$D7,$FC,$50,$DE,$8F,$C5,$FD,$43,$D8,$47,$FB,$1B
	.dc.b	$EA,$1F,$51,$FE,$30,$BE,$1D,$97,$AF,$B9,$FE,$21,$C7,$96,$1C,$6F
	.dc.b	$FC,$6F,$18,$C9,$8F,$47,$FD,$A7,$8F,$96,$31,$08,$C5,$72,$CB,$CB
	.dc.b	$3C,$A3,$18,$77,$8E,$78,$FF,$91,$B0,$E7,$87,$29,$FF,$70,$E3,$E3
	.dc.b	$9E,$ED,$C3,$D9,$3C,$CF,$15,$F3,$8E,$F1,$E1,$39,$E3,$E4,$BE,$F5
	.dc.b	$B9,$F1,$9C,$EF,$B3,$C3,$1E,$D8,$F9,$63,$B9,$BC,$3A,$63,$E0,$C7
	.dc.b	$AC,$F0,$C7,$B2,$C7,$AB,$6F,$8B,$E2,$3D,$07,$62,$C7,$03,$EB,$3F
	.dc.b	$EF,$06,$BF,$00,$C4,$7D,$27,$5E,$CB,$E4,$7F,$DE,$F1,$F3,$A3,$0E
	.dc.b	$2B,$FC,$67,$3A,$8F,$44,$E9,$C7,$C1,$65,$07,$0D,$34,$E0,$FB,$1F
	.dc.b	$30,$FA,$6F,$E6,$D9,$FE,$32,$B9,$D6,$A7,$3A,$7A,$23,$1B,$FE,$06
	.dc.b	$B6,$7C,$8C,$3A,$46,$63,$C5,$87,$E6,$27,$31,$F3,$8E,$31,$9C,$3B
	.dc.b	$C7,$FA,$68,$E3,$87,$FA,$72,$B8,$D7,$9E,$72,$7C,$28,$CC,$F9,$E3
	.dc.b	$C9,$5D,$1F,$F5,$2E,$1E,$A7,$C4,$72,$15,$C9,$F1,$46,$EF,$C0,$DB
	.dc.b	$ED,$1E,$9C,$3A,$6F,$FD,$4B,$0F,$7B,$FD,$53,$0F,$7D,$FA,$F1,$FE
	.dc.b	$CD,$3B,$6D,$FB,$F8,$2D,$FB,$66,$BE,$68,$B8,$7F,$42,$AF,$F5,$85
	.dc.b	$BE,$A1,$DD,$AE,$AE,$73,$98,$C5,$70,$AD,$BF,$EB,$4B,$E9,$8F,$F2
	.dc.b	$08,$FC,$7D,$F1,$75,$7E,$23,$CF,$F1,$9F,$E9,$EF,$13,$13,$A3,$FD
	.dc.b	$89,$0F,$A8,$7C,$46,$B1,$EC,$27,$51,$9C,$7F,$C1,$91,$8B,$E9,$5C
	.dc.b	$23,$0F,$FB,$87,$C7,$74,$FC,$86,$D7,$FB,$54,$7D,$37,$77,$FD,$A7
	.dc.b	$1A,$AE,$55,$F4,$A7,$7C,$3C,$E3,$17,$F7,$A7,$FD,$E3,$7C,$71,$FF
	.dc.b	$11,$47,$03,$AD,$70,$BF,$B7,$F5,$1F,$EF,$19,$FF,$71,$B7,$1F,$87
	.dc.b	$1E,$6E,$2D,$FB,$32,$3D,$59,$EF,$2F,$F1,$0E,$16,$7F,$82,$9E,$EA
	.dc.b	$E1,$87,$4B,$D3,$FC,$2A,$AD,$6C,$FF,$7A,$9F,$EC,$62,$F3,$FC,$0E
	.dc.b	$EF,$C9,$C6,$6B,$EC,$9F,$E4,$49,$F9,$1F,$51,$FE,$57,$9E,$27,$D1
	.dc.b	$71,$3C,$23,$96,$5F,$22,$72,$FF,$8F,$23,$27,$B2,$32,$FF,$22,$97
	.dc.b	$CB,$FE,$69,$F1,$B6,$BE,$9E,$4B,$FD,$5C,$E6,$F9,$36,$FF,$CB,$B4
	.dc.b	$F8,$AD,$F2,$FE,$01,$B7,$F4,$9B,$5F,$35,$FD,$BF,$D3,$3C,$FA,$7E
	.dc.b	$EB,$EA,$1F,$51,$8C,$3F,$5A,$7A,$EB,$83,$8C,$3F,$9F,$C3,$DE,$E1
	.dc.b	$1F,$42,$31,$3B,$84,$61,$FA,$BF,$50,$E7,$F2,$DF,$F0,$C8,$C5,$F1
	.dc.b	$BF,$9B,$7F,$D3,$BA,$1E,$5F,$B3,$D2,$FE,$27,$18,$BF,$03,$6B,$F0
	.dc.b	$6F,$F9,$92,$33,$5C,$3D,$1F,$E3,$7F,$51,$FE,$70,$9F,$47,$C6,$BD
	.dc.b	$CC,$FA,$BF,$C8,$23,$13,$ED,$FA,$57,$B7,$FC,$91,$B1,$F1,$FE,$A1
	.dc.b	$F5,$0F,$A8,$D5,$CA,$F9,$9E,$43,$EA,$3F,$CF,$71,$D6,$7F,$98,$CE
	.dc.b	$7D,$AE,$1C,$5F,$C5,$A3,$15,$B3,$9D,$F2,$9F,$E3,$7D,$DF,$DD,$BE
	.dc.b	$A3,$F9,$DB,$FD,$E3,$BA,$CA,$6C,$61,$B7,$F8,$1A,$E4,$DA,$FC,$B4
	.dc.b	$7F,$80,$BE,$A3,$99,$D3,$92,$D6,$E2,$D7,$90,$FA,$85,$7F,$A6,$FD
	.dc.b	$43,$D8,$A7,$F3,$CF,$A8,$FA,$35,$FB,$E9,$D5,$FE,$3D,$B1,$FC,$CE
	.dc.b	$7F,$91,$4F,$ED,$67,$3E,$2F,$F3,$A9,$CD,$7F,$86,$A7,$35,$F8,$F9
	.dc.b	$CC,$A7,$2F,$F4,$58,$E3,$97,$F8,$B6,$BE,$6F,$AB,$F8,$BA,$FA,$2F
	.dc.b	$FB,$5A,$3E,$53,$9C,$3F,$DA,$D5,$C6,$73,$CF,$C9,$E6,$F4,$FA,$89
	.dc.b	$1C,$BE,$2E,$E3,$DA,$6E,$1D,$31,$56,$2F,$E1,$7C,$87,$CE,$3F,$1B
	.dc.b	$F0,$6F,$96,$C4,$FC,$B9,$FE,$EF,$F5,$1F,$43,$1F,$F6,$EE,$E9,$CF
	.dc.b	$77,$0D,$7F,$B4,$D2,$F8,$61,$E0,$3F,$C0,$AE,$27,$F9,$3E,$8F,$8E
	.dc.b	$F9,$27,$CA,$FA,$8B,$7F,$22,$2D,$7C,$C9,$B8,$7B,$FE,$5F,$E8,$FB
	.dc.b	$3F,$49,$B8,$7D,$06,$DF,$D4,$6D,$7F,$BE,$3E,$A3,$FD,$FD,$5F,$C6
	.dc.b	$5B,$5F,$AF,$E0,$FE,$46,$FF,$4E,$8C,$D7,$1C,$3F,$C2,$3F,$50,$F9
	.dc.b	$6F,$F1,$3D,$6B,$1F,$E3,$53,$97,$B5,$70,$AD,$B6,$3F,$0F,$F5,$1F
	.dc.b	$8C,$FA,$8F,$C6,$B6,$32,$7E,$32,$32,$D3,$F1,$FC,$3F,$94,$F7,$3F
	.dc.b	$2A,$D8,$C9,$F9,$48,$CB,$4F,$E7,$5F,$4F,$F0,$3E,$13,$D9,$B7,$FA
	.dc.b	$02,$1F,$20,$AF,$F3,$CE,$9F,$D8,$E7,$83,$67,$55,$FE,$87,$8D,$7A
	.dc.b	$D8,$C3,$AC,$AF,$AE,$1D,$23,$C1,$5B,$BE,$93,$99,$E1,$86,$5F,$33
	.dc.b	$5E,$95,$6F,$AC,$78,$CD,$D7,$38,$7D,$46,$EF,$83,$C7,$CD,$E9,$F4
	.dc.b	$FF,$AE,$F4,$78,$46,$3B,$BF,$C9,$59,$75,$3D,$91,$99,$DE,$5E,$9B
	.dc.b	$E5,$5B,$6E,$BF,$11,$BA,$FC,$DC,$39,$56,$F1,$EF,$7B,$6D,$6F,$0E
	.dc.b	$D1,$8A,$FF,$5C,$E1,$D1,$B1,$97,$CF,$3C,$31,$91,$DD,$F4,$B4,$98
	.dc.b	$78,$63,$2D,$8C,$E1,$FE,$27,$C3,$C1,$1D,$8C,$9E,$2A,$EF,$5F,$84
	.dc.b	$D1,$97,$C0,$8E,$84,$C3,$DD,$7C,$86,$1D,$8E,$F7,$CB,$1E,$B5,$E3
	.dc.b	$CB,$E0,$56,$B1,$ED,$39,$CB,$78,$FE,$86,$72,$F9,$CF,$06,$3D,$5F
	.dc.b	$15,$F8,$70,$F9,$93,$9C,$BD,$15,$F4,$BE,$A3,$FC,$48,$FF,$BD,$AD
	.dc.b	$18,$8F,$F6,$A4,$7E,$19,$B1,$97,$E6,$3C,$A7,$F1,$0D,$CB,$3E,$BF
	.dc.b	$E9,$FE,$9D,$F5,$1F,$ED,$49,$CE,$1F,$91,$BE,$0F,$96,$F8,$56,$E3
	.dc.b	$06,$E3,$F3,$53,$C2,$17,$D9,$CC,$C7,$81,$FA,$58,$D3,$8C,$3D,$33
	.dc.b	$B7,$F8,$2C,$E9,$CC,$EF,$0F,$4C,$6B,$0F,$34,$FE,$0B,$0F,$26,$3F
	.dc.b	$33,$1E,$98,$7F,$9E,$71,$E8,$F2,$25,$FB,$B1,$E8,$F0,$31,$1A,$7D
	.dc.b	$F8,$F8,$30,$ED,$1E,$53,$60,$D9,$CC,$F3,$61,$70,$F1,$C7,$03,$27
	.dc.b	$33,$89,$67,$BE,$1F,$10,$F3,$C6,$71,$F8,$27,$D5,$AE,$B1,$D2,$78
	.dc.b	$63,$EB,$9C,$DF,$48,$F7,$36,$BD,$53,$AB,$EB,$87,$A7,$1F,$85,$18
	.dc.b	$AF,$45,$6E,$BE,$05,$F3,$9F,$5E,$3E,$27,$E0,$46,$6B,$84,$7A,$EB
	.dc.b	$F2,$6F,$2B,$ED,$7F,$00,$E5,$1E,$DC,$7D,$98,$F9,$2F,$81,$CA,$7F
	.dc.b	$25,$19,$7C,$EF,$E2,$B0,$F8,$E4,$71,$F5,$19,$F8,$69,$5F,$AD,$7F
	.dc.b	$23,$F5,$0F,$86,$CC,$FE,$7E,$6E,$7F,$D8,$2B,$53,$FD,$6D,$AE,$CE
	.dc.b	$86,$6B,$A9,$FE,$D1,$ED,$FE,$DB,$09,$E3,$18,$6F,$1F,$FB,$5D,$7C
	.dc.b	$37,$18,$FF,$6D,$7C,$0F,$F8,$F6,$FF,$D7,$4B,$EA,$35,$1F,$99,$8C
	.dc.b	$3A,$F1,$F2,$FF,$B5,$DB,$1A,$8F,$F7,$AB,$71,$F3,$BD,$E3,$3E,$E7
	.dc.b	$0F,$A0,$91,$E2,$8E,$66,$1E,$71,$87,$FD,$A1,$F5,$0F,$A8,$F0,$FA
	.dc.b	$52,$73,$1E,$69,$D4,$60,$87,$9B,$7D,$4C,$8C,$E5,$E8,$9F,$7F,$D4
	.dc.b	$7F,$54,$9F,$F7,$03,$FE,$E5,$6F,$C2,$6B,$B6,$46,$08,$2F,$EC,$D7
	.dc.b	$38,$C7,$07,$F0,$75,$0F,$93,$F0,$92,$3C,$3F,$15,$5E,$27,$F0,$F9
	.dc.b	$E0,$F9,$A7,$D3,$3F,$63,$EA,$1E,$6F,$B3,$D3,$97,$FE,$FB,$9C,$B6
	.dc.b	$78,$11,$F8,$3D,$B5,$F0,$89,$E2,$E1,$B9,$FC,$09,$FC,$4F,$D4,$3E
	.dc.b	$A2,$92,$FF,$94,$57,$E7,$A5,$F3,$59,$9E,$F1,$C5,$DE,$DF,$26,$31
	.dc.b	$1A,$9E,$FF,$3D,$E0,$FA,$59,$3C,$0F,$E6,$D4,$9D,$C6,$44,$27,$8C
	.dc.b	$78,$A7,$F4,$7E,$3F,$F9,$73,$1F,$DA,$FD,$3B,$EA,$3F,$4E,$DB,$E9
	.dc.b	$5B,$CF,$F0,$0D,$8F,$E9,$0D,$7F,$D7,$BE,$37,$2E,$EE,$EC,$8E,$6E
	.dc.b	$5F,$45,$BB,$7C,$24,$8C,$F6,$D7,$EC,$CE,$DE,$33,$E8,$9F,$4F,$57
	.dc.b	$AB,$C0,$95,$DA,$32,$F8,$23,$2F,$48,$F8,$BD,$7F,$C2,$3D,$77,$D4
	.dc.b	$53,$35,$F1,$26,$BF,$8C,$D7,$F5,$CF,$E3,$AB,$F5,$1C,$1F,$C8,$A7
	.dc.b	$2D,$AE,$2F,$F2,$69,$DE,$FF,$AD,$FA,$8F,$B5,$87,$DD,$6B,$F9,$A6
	.dc.b	$D7,$F2,$E5,$F5,$1F,$C8,$17,$A7,$49,$5F,$87,$FA,$87,$77,$F2,$66
	.dc.b	$C6,$63,$07,$F2,$AF,$A8,$FE,$49,$8F,$F3,$CA,$7C,$5F,$12,$5F,$E4
	.dc.b	$FE,$A3,$F9,$EB,$FE,$0D,$6A,$69,$8F,$A8,$7C,$9F,$F3,$86,$CF,$3E
	.dc.b	$3F,$E7,$75,$BA,$FE,$7B,$EC,$7E,$C2,$BF,$45,$7F,$EC,$26,$FB,$1A
	.dc.b	$F0,$24,$7E,$9E,$B8,$3F,$C1,$1B,$A7,$F5,$8B,$FF,$2F,$3F,$D6,$A3
	.dc.b	$3C,$5F,$C8,$5B,$5F,$E2,$96,$AA,$E3,$3F,$C9,$FD,$FF,$5C,$FF,$12
	.dc.b	$FA,$37,$BD,$FD,$59,$B1,$FB,$B9,$DE,$3F,$DA,$9B,$7C,$F4,$FE,$BB
	.dc.b	$18,$AF,$F1,$D5,$BE,$1B,$35,$27,$A3,$DE,$3F,$B2,$F8,$4F,$8F,$FE
	.dc.b	$CF,$F4,$CF,$8C,$A6,$A2,$61,$FC,$3E,$B8,$36,$38,$FC,$76,$A3,$F9
	.dc.b	$7C,$E9,$B3,$A7,$F8,$C7,$D3,$BE,$8B,$C3,$F5,$12,$73,$1F,$DE,$A7
	.dc.b	$51,$82,$1F,$CF,$32,$FE,$08,$FF,$3E,$8D,$B6,$79,$1E,$EF,$17,$FC
	.dc.b	$0B,$F5,$1F,$62,$DE,$C7,$C2,$33,$CF,$30,$FE,$5D,$5F,$78,$E1,$F4
	.dc.b	$19,$8F,$E9,$B1,$99,$F9,$73,$DB,$E1,$BE,$6F,$B5,$2F,$9C,$65,$A4
	.dc.b	$3D,$9D,$BF,$E1,$AF,$A8,$73,$52,$4F,$F4,$1F,$A8,$FE,$E5,$D5,$FE
	.dc.b	$7B,$FA,$87,$D4,$7F,$77,$C3,$8D,$FF,$47,$6B,$FD,$91,$B5,$FE,$38
	.dc.b	$BE,$72,$BE,$6F,$FA,$1E,$33,$18,$85,$74,$8C,$36,$FF,$C1,$15,$C1
	.dc.b	$BE,$D7,$F8,$3E,$FB,$47,$9F,$1F,$F0,$76,$1C,$9B,$1C,$31,$F7,$36
	.dc.b	$B8,$9F,$E9,$BC,$7C,$0D,$AE,$2F,$A7,$0F,$06,$1E,$9C,$7C,$93,$B3
	.dc.b	$A9,$4E,$A1,$DB,$67,$9E,$1E,$98,$E2,$7C,$4B,$F0,$BB,$C7,$37,$CB
	.dc.b	$1F,$3C,$EC,$CE,$1D,$1F,$45,$74,$9D,$9C,$2B,$C6,$7D,$89,$FF,$1B
	.dc.b	$1D,$CA,$53,$2E,$DF,$AB,$E3,$3E,$43,$71,$EF,$8C,$3C,$A7,$FA,$4C
	.dc.b	$E9,$B1,$C0,$7C,$15,$69,$97,$04,$7E,$6F,$CC,$BE,$99,$E1,$7C,$A9
	.dc.b	$FF,$6B,$57,$CC,$AD,$4E,$67,$AE,$FF,$93,$9C,$C7,$D8,$38,$9D,$AF
	.dc.b	$AB,$93,$8E,$FF,$E5,$D9,$E3,$5F,$9D,$9E,$2F,$5A,$E2,$E6,$38,$9F
	.dc.b	$EE,$B9,$E3,$5F,$43,$0F,$45,$F0,$8C,$57,$8F,$0F,$55,$69,$B1,$93
	.dc.b	$CD,$19,$69,$9C,$3C,$51,$97,$CD,$3F,$E4,$5C,$3E,$1E,$BF,$EB,$4B
	.dc.b	$F5,$E1,$EF,$C7,$E1,$CF,$FA,$E2,$B7,$7F,$0F,$83,$2E,$CE,$8E,$30
	.dc.b	$F8,$1D,$36,$DF,$50,$EF,$D5,$31,$FE,$81,$8F,$F9,$76,$F7,$5A,$9C
	.dc.b	$D7,$FA,$22,$73,$5E,$BC,$3B,$61,$D6,$3D,$D8,$71,$3C,$9F,$1A,$F8
	.dc.b	$2F,$ED,$F3,$A2,$63,$FE,$94,$97,$D4,$3E,$23,$95,$7F,$9C,$67,$31
	.dc.b	$8A,$E1,$5B,$AE,$79,$FE,$32,$F7,$1F,$E7,$5F,$80,$74,$FC,$26,$C7
	.dc.b	$F9,$F4,$7D,$32,$B5,$52,$75,$3F,$53,$10,$BE,$6F,$FA,$BA,$79,$F0
	.dc.b	$F1,$8C,$57,$17,$85,$FD,$9A,$D9,$FE,$8B,$FA,$8F,$F7,$9B,$7D,$AD
	.dc.b	$5B,$8F,$ED,$D6,$B0,$FF,$56,$B7,$4F,$D6,$E1,$EC,$C3,$67,$22,$BE
	.dc.b	$C6,$D7,$99,$FC,$3C,$7F,$95,$8F,$F4,$F1,$63,$D7,$DD,$AE,$A9,$3F
	.dc.b	$2F,$EA,$3F,$C9,$6F,$08,$F7,$46,$2B,$FD,$87,$3C,$F0,$F3,$CE,$5F
	.dc.b	$E0,$B3,$FC,$4A,$B8,$F9,$7F,$9A,$F9,$BF,$C8,$4F,$FB,$31,$B5,$F0
	.dc.b	$35,$61,$BE,$59,$97,$DA,$F1,$2E,$3F,$51,$B8,$71,$6C,$FD,$E3,$F0
	.dc.b	$AF,$EA,$BE,$95,$E8,$D3,$F5,$BF,$50,$FA,$8C,$5F,$FB,$16,$B6,$DF
	.dc.b	$77,$39,$7F,$B9,$25,$1D,$4F,$DB,$FD,$47,$E3,$6B,$81,$FE,$CF,$71
	.dc.b	$5C,$78,$3F,$7B,$D3,$FE,$BB,$AF,$04,$63,$CC,$91,$FE,$D7,$9F,$F2
	.dc.b	$73,$63,$8F,$A3,$FC,$1B,$EA,$3F,$CB,$53,$FE,$DD,$EF,$FF,$4D,$BF
	.dc.b	$ED,$F8,$CE,$3F,$EF,$C9,$CB,$FE,$FE,$FA,$8F,$F4,$AD,$7F,$B9,$5B
	.dc.b	$0E,$2F,$E8,$3F,$50,$FA,$87,$D4,$7F,$16,$BD,$D7,$E0,$9B,$87,$F3
	.dc.b	$DF,$A8,$FF,$34,$3D,$7D,$EF,$E4,$5E,$42,$73,$9F,$8E,$73,$18,$C3
	.dc.b	$A4,$63,$D7,$FE,$55,$DB,$FB,$CF,$A8,$B8,$7E,$33,$2F,$DF,$CE,$5F
	.dc.b	$E0,$38,$7F,$5A,$8F,$E0,$7D,$9C,$2F,$8B,$6B,$67,$2B,$E8,$2B,$F8
	.dc.b	$5F,$D4,$78,$0F,$0F,$1D,$AE,$35,$B1,$39,$7E,$64,$75,$9C,$63,$FA
	.dc.b	$AF,$A6,$7A,$BF,$D0,$9B,$5F,$CB,$7E,$A3,$EC,$D6,$EB,$47,$F3,$A8
	.dc.b	$FE,$3B,$19,$9E,$1E,$46,$9A,$79,$B1,$61,$D9,$FE,$7D,$7C,$A7,$E5
	.dc.b	$FA,$7F,$E0,$E7,$DB,$A7,$F8,$87,$DD,$FA,$73,$9B,$F6,$9F,$1D,$FC
	.dc.b	$8E,$5F,$45,$D3,$67,$8D,$F2,$79,$FD,$44,$8E,$F4,$F8,$AC,$61,$DB
	.dc.b	$1F,$F1,$95,$E3,$1E,$D5,$70,$FE,$83,$7E,$EF,$8F,$F9,$87,$E4,$BE
	.dc.b	$05,$F2,$D8,$AF,$89,$B2,$BF,$BB,$7D,$47,$D0,$9E,$73,$97,$FB,$A4
	.dc.b	$67,$0F,$E9,$F3,$97,$FB,$A4,$65,$F1,$4F,$0F,$8A,$F9,$3F,$1B,$23
	.dc.b	$DF,$F5,$16,$BD,$DC,$0A,$C6,$DF,$2E,$AD,$FD,$6F,$98,$FA,$7F,$11
	.dc.b	$F5,$1B,$1F,$29,$B8,$F4,$6D,$FF,$3A,$FA,$8F,$F0,$0C,$FD,$A0,$E2
	.dc.b	$3F,$C8,$7D,$EE,$1F,$E4,$F5,$BA,$D6,$1F,$41,$B1,$FD,$DF,$EA,$1F
	.dc.b	$2B,$FE,$1E,$9C,$F9,$29,$CD,$FE,$56,$73,$18,$BE,$DA,$7F,$86,$23
	.dc.b	$19,$FE,$37,$EA,$3F,$27,$F5,$1F,$94,$6E,$3F,$8A,$6E,$1F,$31,$B5
	.dc.b	$F9,$5E,$2F,$CD,$F6,$7E,$71,$B7,$F9,$26,$DF,$D4,$6D,$7F,$62,$FA
	.dc.b	$7F,$87,$F0,$9E,$EA,$FF,$08,$C3,$E4,$3F,$CD,$73,$99,$4F,$0E,$0F
	.dc.b	$CB,$BB,$C7,$FC,$F7,$18,$F5,$78,$BA,$C2,$DF,$69,$AF,$48,$B7,$CE
	.dc.b	$3F,$D1,$F5,$C6,$FA,$C7,$16,$BC,$5B,$3C,$70,$F4,$2F,$A8,$EB,$87
	.dc.b	$67,$BF,$9D,$E2,$FA,$8F,$DC,$72,$72,$8C,$77,$FF,$8E,$73,$F1,$9F
	.dc.b	$1A,$33,$3B,$CF,$C7,$7C,$AB,$61,$C6,$7F,$2D,$BA,$7C,$AC,$39,$D7
	.dc.b	$09,$CB,$FE,$44,$6B,$F4,$A3,$75,$CF,$1F,$33,$71,$F2,$57,$39,$CE
	.dc.b	$1F,$C3,$B0,$E6,$74,$D3,$FC,$D9,$97,$AA,$B7,$7E,$16,$BE,$B6,$CF
	.dc.b	$08,$ED,$87,$12,$3F,$0A,$35,$8F,$96,$36,$47,$EB,$E5,$EC,$CB,$CD
	.dc.b	$87,$83,$2F,$34,$E6,$FC,$98,$F6,$C3,$8E,$57,$1E,$33,$6B,$EC,$4D
	.dc.b	$7D,$51,$72,$F2,$D7,$09,$F2,$36,$FA,$C6,$0F,$A9,$F5,$1F,$E7,$77
	.dc.b	$FD,$E7,$92,$31,$1F,$EF,$88,$FC,$23,$63,$2F,$D3,$79,$4F,$E1,$5B
	.dc.b	$59,$F8,$2F,$CF,$FD,$43,$EA,$3F,$A3,$DE,$E7,$F2,$37,$C3,$0F,$2C
	.dc.b	$70,$AD,$E1,$CE,$7F,$1D,$5B,$7F,$2C,$DC,$3B,$47,$4B,$8F,$07,$A3
	.dc.b	$8C,$3C,$4E,$4F,$D0,$CF,$27,$11,$A2,$5F,$F6,$09,$D4,$62,$7D,$27
	.dc.b	$09,$D6,$1E,$A8,$CC,$ED,$F4,$DF,$91,$E0,$6A,$BE,$04,$FF,$1B,$BE
	.dc.b	$F1,$82,$4E,$CF,$F5,$9C,$69,$C1,$D8,$EA,$1D,$39,$9F,$26,$3F,$83
	.dc.b	$BF,$24,$63,$1F,$A1,$39,$AF,$8F,$8F,$B5,$E5,$8F,$91,$A7,$92,$35
	.dc.b	$3C,$6F,$A6,$1D,$AF,$84,$E5,$E5,$8F,$9A,$31,$87,$B1,$AF,$AF,$0E
	.dc.b	$B8,$F9,$89,$87,$59,$E1,$18,$7D,$27,$E0,$F0,$F0,$C7,$E2,$EB,$9D
	.dc.b	$66,$BE,$5E,$1E,$3A,$E5,$8F,$DB,$8C,$56,$EB,$E0,$9E,$53,$18,$FA
	.dc.b	$0F,$4C,$F0,$8F,$AB,$18,$C3,$E6,$4E,$6F,$B4,$60,$95,$ED,$3E,$E7
	.dc.b	$D4,$70,$F8,$79,$1F,$B2,$7E,$B7,$D4,$7F,$BA,$63,$04,$3F,$43,$1A
	.dc.b	$F8,$69,$19,$C3,$F2,$53,$88,$FE,$8F,$5C,$DF,$F0,$E3,$79,$BF,$DB
	.dc.b	$2D,$E1,$FF,$23,$AF,$85,$FF,$24,$CF,$FB,$7B,$E0,$7F,$C8,$97,$FE
	.dc.b	$69,$5F,$51,$AC,$3F,$C9,$F1,$89,$D7,$93,$94,$23,$9B,$FE,$2F,$78
	.dc.b	$36,$B8,$3B,$9B,$2F,$72,$FC,$FC,$AF,$24,$68,$C3,$D6,$30,$FF,$A0
	.dc.b	$7E,$A1,$F5,$1D,$3D,$39,$39,$8F,$0C,$EA,$30,$43,$D1,$C3,$A9,$91
	.dc.b	$9D,$7E,$44,$FA,$BE,$A1,$D9,$7E,$0E,$5F,$FA,$4E,$72,$D9,$CB,$FE
	.dc.b	$53,$8F,$F7,$8B,$93,$87,$CA,$75,$23,$CF,$E2,$AB,$C6,$2E,$2E,$2F
	.dc.b	$F4,$2D,$AF,$E7,$FD,$43,$CE,$F4,$7A,$93,$2F,$F5,$44,$E5,$B3,$C0
	.dc.b	$8F,$AF,$BA,$BF,$E6,$B9,$E0,$D9,$F8,$93,$F7,$BE,$9D,$F5,$14,$95
	.dc.b	$FE,$B7,$87,$CC,$E2,$3F,$CE,$D1,$C1,$D4,$F3,$9E,$9F,$3D,$D7,$EA
	.dc.b	$25,$73,$8C,$BD,$63,$2F,$48,$FC,$8F,$91,$FE,$57,$E0,$F9,$13,$F9
	.dc.b	$AF,$A7,$7D,$47,$E5,$9B,$1F,$EC,$DA,$E1,$C5,$FE,$B2,$6C,$7E,$89
	.dc.b	$AF,$FA,$F7,$C8,$E5,$DD,$E0,$64,$73,$72,$FF,$B6,$6F,$F4,$4D,$ED
	.dc.b	$F1,$1E,$3E,$DB,$19,$21,$FE,$9D,$94,$7A,$7A,$BD,$5E,$04,$AE,$D1
	.dc.b	$97,$C1,$19,$79,$47,$C6,$EB,$FD,$BF,$B0,$FA,$8A,$66,$BE,$24,$D7
	.dc.b	$FD,$5D,$AF,$E7,$1F,$F5,$85,$7E,$89,$CC,$EF,$83,$E8,$C7,$FB,$8E
	.dc.b	$3B,$F6,$FF,$06,$FA,$77,$25,$25,$7F,$BF,$E1,$F5,$1B,$F4,$A4,$E5
	.dc.b	$FE,$1F,$F4,$FF,$A3,$8F,$E2,$F9,$78,$23,$11,$FE,$C5,$6C,$67,$BF
	.dc.b	$F8,$BF,$D4,$7F,$1C,$8E,$1F,$15,$D0,$97,$D3,$A3,$FD,$EB,$F5,$1F
	.dc.b	$CC,$63,$F9,$2B,$53,$4C,$7D,$43,$E4,$FF,$92,$B6,$7A,$1C,$B9,$7F
	.dc.b	$94,$D7,$0A,$FE,$57,$EC,$7E,$AA,$B6,$2F,$FD,$6C,$E1,$AF,$FB,$8E
	.dc.b	$38,$37,$BF,$99,$1E,$1E,$15,$74,$76,$9E,$3A,$7D,$AB,$FF,$2D,$BF
	.dc.b	$73,$DF,$F5,$57,$FB,$C7,$E9,$1E,$FF,$F4,$26,$C7,$A2,$77,$5C,$73
	.dc.b	$FE,$8F,$18,$DB,$FA,$7D,$7F,$BD,$E7,$97,$C4,$69,$91,$CD,$CB,$F1
	.dc.b	$3C,$47,$C7,$7F,$52,$FA,$67,$C6,$53,$FA,$AC,$EA,$B8,$56,$08,$FC
	.dc.b	$EA,$7C,$76,$21,$A7,$E7,$2F,$94,$63,$3F,$4C,$62,$7F,$43,$7C,$AB
	.dc.b	$EC,$FD,$3B,$E8,$BC,$DF,$51,$27,$31,$FD,$AA,$75,$18,$21,$DC,$FE
	.dc.b	$EB,$5F,$C3,$1C,$08,$E2,$E1,$B7,$FB,$46,$E3,$F4,$2B,$94,$FE,$E3
	.dc.b	$EA,$3F,$03,$39,$F5,$FE,$61,$FD,$FA,$33,$87,$F8,$19,$A9,$A6,$2D
	.dc.b	$F4,$1C,$7A,$25,$7E,$72,$B5,$3D,$7E,$1B,$ED,$7B,$32,$B5,$19,$69
	.dc.b	$0F,$83,$87,$EC,$FE,$A1,$CF,$FE,$14,$6C,$FE,$A9,$AF,$C8,$FA,$87
	.dc.b	$B3,$F7,$A1,$F5,$0F,$A8,$FC,$AD,$F0,$E1,$FE,$3A,$DE,$1C,$E5,$FE
	.dc.b	$22,$CB,$9D,$7F,$8A,$B0,$E5,$8F,$F8,$B2,$72,$F6,$C7,$8B,$B6,$B9
	.dc.b	$38,$65,$E5,$AE,$70,$AE,$71,$8C,$BD,$AD,$C7,$F5,$AF,$09,$DE,$1F
	.dc.b	$B0,$6D,$FF,$90,$2F,$8C,$F8,$AB,$86,$1E,$18,$E0,$DA,$E1,$87,$69
	.dc.b	$E0,$6B,$1F,$64,$FA,$9A,$B0,$EE,$F2,$33,$8F,$91,$F7,$61,$DD,$E1
	.dc.b	$87,$77,$81,$F1,$CD,$DF,$29,$E1,$7D,$A7,$89,$99,$8F,$F1,$23,$CC
	.dc.b	$53,$99,$6F,$EC,$BB,$7F,$C6,$9F,$23,$B8,$FA,$9E,$2A,$31,$3A,$9E
	.dc.b	$53,$97,$80,$F8,$2A,$D3,$2E,$08,$FB,$FE,$41,$F4,$DF,$E2,$0D,$BF
	.dc.b	$0B,$3F,$E7,$AA,$F9,$F5,$AD,$FE,$8C,$E4,$E5,$7C,$77,$F8,$17,$C2
	.dc.b	$31,$3F,$16,$31,$5C,$A3,$CD,$87,$E5,$30,$ED,$1F,$4C,$ED,$19,$BE
	.dc.b	$71,$FD,$02,$F9,$BB,$39,$E1,$FD,$12,$F9,$DF,$B2,$72,$7F,$A5,$A3
	.dc.b	$33,$E8,$8F,$5E,$1E,$5B,$EB,$7C,$5B,$5B,$8F,$24,$F8,$B4,$FF,$0C
	.dc.b	$E1,$EA,$7F,$C3,$78,$7A,$AF,$C9,$8F,$F9,$3A,$76,$DB,$F4,$78,$D6
	.dc.b	$FC,$D3,$56,$09,$C3,$FD,$16,$BF,$C0,$B6,$FA,$8F,$EA,$71,$FE,$02
	.dc.b	$9D,$5F,$F4,$3A,$D7,$6F,$2E,$BF,$04,$62,$F8,$46,$30,$E7,$7F,$D2
	.dc.b	$30,$B7,$F0,$30,$F4,$7C,$67,$FA,$8B,$C4,$F2,$C4,$CB,$FC,$33,$2F
	.dc.b	$A8,$7C,$3E,$72,$FE,$AD,$39,$8C,$63,$FE,$74,$8C,$5F,$AF,$0E,$F8
	.dc.b	$72,$8F,$F0,$FF,$92,$EA,$F9,$0D,$8F,$F1,$48,$FA,$76,$1A,$A9,$3A
	.dc.b	$9F,$B4,$FF,$83,$AB,$96,$7F,$DA,$E3,$97,$0F,$08,$C5,$70,$7F,$DC
	.dc.b	$75,$C3,$3F,$F4,$6D,$F0,$3F,$C6,$BF,$51,$FE,$F3,$6F,$B3,$F8,$17
	.dc.b	$CD,$C7,$F8,$19,$D6,$1F,$DE,$1B,$7D,$31,$B7,$FE,$24,$9B,$5F,$EC
	.dc.b	$39,$AA,$33,$3F,$0F,$0E,$C7,$86,$B7,$1F,$EC,$93,$FC,$96,$5A,$F7
	.dc.b	$F7,$FE,$56,$31,$87,$F9,$E9,$A2,$7F,$DA,$9F,$51,$FE,$5D,$8E,$7D
	.dc.b	$3E,$F8,$C4,$FF,$85,$27,$8B,$C3,$C9,$ED,$5B,$7F,$49,$F3,$7F,$92
	.dc.b	$AF,$F2,$1B,$6B,$E0,$5E,$B8,$E5,$7D,$DF,$16,$D7,$D8,$6E,$5F,$2D
	.dc.b	$B3,$E9,$3F,$10,$FE,$CF,$E9,$9E,$7D,$3F,$6D,$F5,$0F,$A8,$C5,$FF
	.dc.b	$9C,$AB,$6D,$F6,$F3,$97,$F8,$CA,$51,$D4,$FE,$01,$F5,$0E,$6C,$5F
	.dc.b	$F9,$F2,$30,$F5,$E0,$FE,$0D,$D2,$F8,$2F,$34,$61,$93,$FE,$84,$9F
	.dc.b	$F2,$FB,$63,$8F,$A3,$FC,$3F,$EA,$1E,$9F,$F1,$1F,$79,$ED,$E6,$30
	.dc.b	$FF,$15,$FA,$57,$B5,$4F,$E3,$5F,$50,$FA,$87,$D4,$67,$5F,$F4,$AD
	.dc.b	$F0,$7F,$D3,$BE,$43,$EA,$3F,$CE,$B1,$AC,$7F,$92,$4E,$7D,$BE,$5C
	.dc.b	$FF,$E5,$D8,$C5,$6C,$E7,$7C,$A7,$FC,$C5,$DF,$F9,$3F,$A8,$FE,$5C
	.dc.b	$FE,$57,$BE,$CA,$6C,$61,$B7,$FE,$B2,$AE,$4D,$AF,$C8,$47,$E7,$FE
	.dc.b	$A3,$99,$D3,$92,$D6,$E2,$D7,$90,$FA,$65,$7F,$A0,$7D,$33,$D9,$A7
	.dc.b	$F2,$EF,$A8,$FA,$77,$FE,$BE,$9D,$36,$75,$E6,$70,$8F,$E6,$35,$FC
	.dc.b	$4D,$FF,$54,$D7,$0C,$3F,$98,$CE,$5F,$DC,$46,$7C,$4F,$E6,$53,$9C
	.dc.b	$FF,$C2,$D7,$F2,$BD,$6F,$C3,$BF,$15,$39,$E8,$D3,$4D,$17,$1F,$AB
	.dc.b	$2F,$A8,$E8,$75,$F8,$AB,$5D,$62,$D7,$09,$B9,$78,$23,$D3,$F1,$FF
	.dc.b	$28,$FC,$4F,$C1,$BE,$57,$18,$FC,$99,$FE,$C3,$F5,$1F,$32,$FF,$D4
	.dc.b	$BB,$A7,$3D,$DC,$27,$FB,$55,$6F,$2F,$04,$2F,$86,$1F,$DD,$2B,$4F
	.dc.b	$F2,$3D,$FF,$C7,$DF,$1B,$F2,$4F,$95,$F5,$16,$7E,$54,$5C,$3E,$2C
	.dc.b	$DC,$7D,$D7,$FE,$47,$F9,$6F,$A5,$EC,$FD,$36,$E3,$F4,$DB,$87,$CE
	.dc.b	$6D,$FA,$3E,$A3,$FD,$E7,$5F,$CD,$E3,$17,$F2,$3C,$FF,$F0,$95,$6B
	.dc.b	$0F,$F6,$4B,$6B,$FB,$B7,$D4,$3E,$57,$78,$7F,$81,$23,$18,$F5,$3F
	.dc.b	$DF,$37,$C5,$FE,$8D,$E5,$27,$78,$F6,$BE,$58,$7E,$3F,$E9,$FF,$21
	.dc.b	$F5,$17,$1F,$C8,$36,$FF,$16,$DC,$FE,$3F,$4A,$3F,$31,$C9,$F9,$9E
	.dc.b	$CB,$87,$E6,$5B,$87,$E5,$1B,$87,$27,$FC,$69,$F4,$DF,$CF,$27,$FC
	.dc.b	$E5,$9F,$F5,$F5,$0F,$A8,$FC,$A7,$C2,$FC,$59,$FE,$C9,$3C,$63,$11
	.dc.b	$C2,$77,$5C,$EB,$4D,$8C,$CF,$2E,$2A,$F1,$CB,$FC,$F5,$5B,$6C,$78
	.dc.b	$2B,$77,$D2,$31,$3A,$79,$1E,$C3,$CF,$F5,$1C,$2B,$8B,$CB,$CE,$F0
	.dc.b	$FD,$3B,$BF,$FC,$5B,$7C,$4F,$87,$3A,$8C,$E9,$F0,$32,$4E,$79,$7C
	.dc.b	$51,$8C,$FC,$98,$F5,$9B,$87,$2D,$3E,$84,$E7,$2F,$06,$1C,$5F,$14
	.dc.b	$67,$2F,$9E,$DA,$E3,$7C,$EB,$EA,$E5,$E3,$78,$C6,$55,$71,$77,$97
	.dc.b	$93,$29,$1C,$1F,$5D,$F3,$C7,$C3,$87,$3C,$39,$5F,$C6,$C3,$8D,$6B
	.dc.b	$0B,$8F,$E0,$F1,$E4,$FA,$23,$37,$E2,$3C,$04,$AE,$4F,$C4,$AE,$2F
	.dc.b	$F8,$D7,$0F,$09,$E2,$CB,$E4,$57,$02,$61,$DE,$B9,$63,$E0,$7E,$1E
	.dc.b	$5E,$B3,$E3,$5F,$2A,$E1,$7F,$43,$EA,$3F,$C3,$97,$FE,$E7,$C5,$18
	.dc.b	$8F,$F7,$64,$7D,$B6,$C6,$5F,$BE,$F2,$9E,$9E,$AF,$E1,$A3,$17,$FB
	.dc.b	$3F,$A8,$7D,$47,$FB,$26,$38,$D6,$C9,$87,$E5,$EB,$6E,$30,$EA,$FE
	.dc.b	$2B,$0E,$AF,$F9,$76,$76,$E6,$B8,$5F,$68,$E9,$87,$85,$FA,$17,$C9
	.dc.b	$FC,$5B,$71,$F8,$77,$FB,$3C,$3A,$E3,$E2,$26,$3D,$5F,$E1,$31,$A8
	.dc.b	$C1,$30,$EF,$5C,$F1,$F1,$47,$82,$3C,$EE,$8F,$39,$83,$47,$B4,$C9
	.dc.b	$6F,$C6,$F1,$7C,$73,$DB,$0F,$5C,$EE,$FC,$8B,$0F,$43,$C6,$33,$3C
	.dc.b	$2F,$9C,$F8,$E3,$13,$CA,$B8,$CF,$9A,$BC,$6D,$9F,$C1,$5F,$8F,$1F
	.dc.b	$9C,$4C,$7E,$25,$75,$8C,$3E,$EB,$E7,$8F,$82,$3C,$C7,$E0,$6F,$A5
	.dc.b	$F1,$39,$C7,$BC,$D9,$E3,$38,$1B,$34,$7C,$38,$E6,$E2,$7C,$0E,$4E
	.dc.b	$15,$EE,$C7,$E0,$CE,$89,$87,$E3,$63,$2F,$B2,$78,$5F,$67,$DB,$1B
	.dc.b	$8C,$63,$F4,$A3,$18,$7B,$2B,$78,$DA,$F9,$B3,$4F,$69,$FA,$BF,$A8
	.dc.b	$FC,$7E,$3F,$F0,$02,$3F,$39,$F0,$F8,$3F,$27,$F5,$0F,$86,$C4,$E8
	.dc.b	$87,$F4,$B8,$D3,$FD,$5D,$AF,$EE,$67,$8B,$B9,$EA,$7F,$75,$E7,$FF
	.dc.b	$1B,$D7,$06,$C6,$78,$7F,$BD,$4B,$E1,$71,$7F,$ED,$BF,$81,$FF,$22
	.dc.b	$3F,$E1,$B3,$1F,$51,$A5,$08,$C3,$C2,$18,$FF,$92,$23,$0D,$9D,$9C
	.dc.b	$3C,$8E,$3E,$E7,$2F,$A0,$98,$78,$4E,$F1,$A7,$18,$27,$71,$E0,$CF
	.dc.b	$FC,$2D,$3F,$EC,$78,$E5,$F5,$0F,$A8,$F3,$FA,$52,$73,$1E,$D9,$D4
	.dc.b	$60,$87,$53,$E3,$6F,$B9,$91,$AD,$7E,$04,$FC,$7F,$A8,$FF,$00,$CF
	.dc.b	$AD,$FF,$71,$B7,$E0,$F5,$DB,$23,$04,$3C,$02,$79,$B8,$5F,$29,$DC
	.dc.b	$91,$DF,$E2,$6B,$B8,$B8,$FF,$9F,$29,$5F,$77,$EA,$1E,$6F,$A7,$D3
	.dc.b	$98,$FD,$39,$CB,$67,$81,$1F,$77,$65,$7D,$33,$C1,$B3,$F1,$67,$F1
	.dc.b	$9F,$50,$FA,$8A,$4A,$FC,$14,$3E,$6F,$11,$F1,$23,$83,$A9,$E7,$3D
	.dc.b	$3E,$7B,$AF,$D2,$CA,$E7,$19,$7A,$C6,$5E,$91,$F9,$FF,$1F,$FC,$B3
	.dc.b	$BF,$FA,$8B,$E9,$DF,$51,$FA,$36,$DF,$E5,$EB,$7D,$5F,$E6,$76,$C0
	.dc.b	$FF,$25,$78,$9D,$4E,$5D,$D6,$39,$45,$7F,$03,$7F,$AA,$BF,$F3,$93
	.dc.b	$F9,$D8,$CF,$77,$8C,$F2,$77,$58,$FD,$54,$EF,$3F,$55,$6A,$73,$D5
	.dc.b	$EB,$F0,$24,$60,$87,$89,$CB,$E4,$8F,$8D,$D7,$FC,$17,$D8,$7D,$45
	.dc.b	$33,$5F,$16,$6B,$F9,$CD,$FF,$46,$FE,$7A,$BF,$3B,$97,$FB,$42,$72
	.dc.b	$DA,$E2,$FF,$B5,$27,$7A,$FF,$A5,$7E,$A1,$C7,$FC,$5D,$B8,$FF,$0B
	.dc.b	$6B,$FA,$C5,$F5,$1B,$F4,$E4,$ED,$FE,$41,$F5,$1F,$A6,$8F,$E2,$8D
	.dc.b	$7F,$6A,$DC,$BF,$DC,$ED,$C7,$FD,$51,$F5,$1F,$66,$BF,$05,$2F,$8B
	.dc.b	$EA,$48,$FE,$41,$F5,$1F,$CE,$6F,$FD,$60,$D8,$FC,$53,$53,$7E,$A1
	.dc.b	$F2,$7F,$CD,$DB,$3D,$38,$FF,$9D,$56,$EB,$F9,$E7,$B1,$FA,$DA,$FE
	.dc.b	$61,$7F,$EB,$D6,$FB,$1A,$EE,$91,$FC,$D6,$B8,$3F,$C5,$9B,$B3,$3F
	.dc.b	$F6,$46,$1F,$E6,$18,$C5,$7E,$8E,$B7,$5F,$EC,$D6,$C6,$6F,$D5,$E0
	.dc.b	$FF,$51,$F8,$4F,$5C,$FF,$24,$FA,$47,$BF,$FD,$55,$B1,$FD,$26,$77
	.dc.b	$97,$82,$11,$8C,$BF,$DB,$8D,$8F,$DB,$35,$53,$E1,$FC,$5B,$49,$EC
	.dc.b	$F8,$63,$FB,$37,$86,$F8,$FF,$ED,$1F,$4A,$F8,$CA,$6A,$26,$1F,$D7
	.dc.b	$EB,$83,$63,$8F,$C7,$6A,$3F,$A1,$CE,$9B,$3A,$7F,$B4,$7D,$3B,$E8
	.dc.b	$BC,$3F,$51,$27,$31,$FD,$EE,$75,$18,$21,$FD,$43,$2F,$E1,$1B,$FF
	.dc.b	$C0,$19,$DC,$FF,$54,$6F,$07,$CD,$6C,$FF,$1F,$95,$72,$7F,$91,$7D
	.dc.b	$47,$DE,$73,$EC,$7C,$63,$FC,$1D,$19,$27,$2E,$0F,$97,$F3,$F6,$BF
	.dc.b	$C0,$D3,$A7,$FC,$11,$69,$E7,$F0,$DF,$4B,$DA,$95,$C2,$32,$D2,$1F
	.dc.b	$07,$B7,$FC,$41,$F5,$1F,$82,$8F,$F0,$A3,$5F,$EE,$AD,$DB,$F1,$DF
	.dc.b	$51,$FD,$D1,$FE,$81,$F5,$0F,$A8,$F8,$95,$BE,$0B,$D5,$D8,$FF,$17
	.dc.b	$61,$C7,$16,$1C,$5F,$E9,$51,$98,$C4,$2F,$C9,$18,$6E,$1F,$4E,$B8
	.dc.b	$37,$DB,$FF,$1B,$DF,$68,$F1,$BE,$86,$B9,$3C,$31,$C2,$FC,$35,$D2
	.dc.b	$7F,$AF,$5F,$81,$B8,$7E,$0E,$B7,$7C,$70,$F1,$65,$DE,$F6,$53,$D9
	.dc.b	$7C,$67,$71,$E3,$18,$F8,$5D,$9C,$71,$F1,$3A,$C3,$C0,$FA,$B0,$D6
	.dc.b	$1D,$63,$C2,$75,$BE,$87,$5B,$E6,$74,$73,$7F,$2A,$7F,$40,$77,$BA
	.dc.b	$F7,$2C,$FD,$D7,$6F,$F8,$CB,$E3,$37,$1E,$CF,$6D,$18,$9D,$4F,$29
	.dc.b	$CB,$C0,$7C,$2D,$69,$97,$04,$7E,$57,$CC,$BE,$9B,$F9,$AE,$DF,$8C
	.dc.b	$9F,$F0,$C5,$7D,$3A,$D6,$FF,$56,$72,$72,$BE,$3B,$FC,$0B,$E1,$18
	.dc.b	$9F,$8B,$18,$AE,$51,$E6,$C3,$F3,$78,$76,$8F,$BC,$76,$8C,$DF,$38
	.dc.b	$FE,$1D,$7C,$DD,$9C,$F0,$FE,$25,$7C,$EF,$D9,$39,$3F,$C6,$91,$99
	.dc.b	$F4,$47,$AF,$0F,$2D,$F5,$BE,$2D,$AD,$C7,$92,$7C,$5A,$7F,$4F,$C3
	.dc.b	$D4,$FF,$51,$C3,$D5,$7E,$4C,$7F,$B1,$CE,$DB,$7E,$8E,$0B,$7E,$69
	.dc.b	$AB,$0F,$F6,$65,$7F,$A9,$DF,$E8,$0F,$FA,$E6,$DF,$50,$EE,$E3,$5A
	.dc.b	$BF,$F4,$E5,$6A,$73,$D7,$E0,$8C,$61,$D6,$31,$87,$38,$F0,$57,$43
	.dc.b	$CF,$F1,$AF,$82,$FE,$EB,$3A,$26,$3F,$EA,$B9,$7D,$43,$E2,$39,$57
	.dc.b	$FB,$46,$73,$18,$AE,$15,$BA,$E7,$9F,$F8,$32,$F7,$1F,$EC,$9F,$21
	.dc.b	$D3,$F2,$1B,$1F,$ED,$D1,$F4,$EA,$D5,$49,$D4,$FD,$A7,$FA,$E5,$72
	.dc.b	$C7,$FD,$89,$1C,$B8,$38,$46,$2B,$83,$FE,$81,$AE,$18,$FF,$88,$AF
	.dc.b	$81,$FE,$DF,$FA,$8F,$F7,$A3,$7D,$AF,$C0,$BE,$6E,$3F,$C0,$CE,$B0
	.dc.b	$FF,$62,$B6,$FA,$63,$6F,$FB,$94,$DA,$FF,$1A,$4D,$51,$99,$F8,$78
	.dc.b	$76,$3C,$35,$B8,$FF,$63,$9F,$EF,$52,$D7,$BF,$BB,$F2,$B1,$8C,$3F
	.dc.b	$C6,$0D,$3F,$D4,$43,$EA,$3F,$CA,$EF,$3E,$9F,$7C,$62,$FF,$DC,$33
	.dc.b	$C5,$FF,$26,$46,$6B,$AD,$70,$F2,$BF,$4B,$F2,$FF,$95,$BF,$F7,$4B
	.dc.b	$6B,$E2,$DE,$B9,$66,$3F,$7B,$C5,$B8,$7D,$96,$DF,$CC,$6C,$FA,$B4
	.dc.b	$47,$ED,$3E,$99,$E7,$D3,$F6,$FF,$50,$FA,$8C,$61,$FE,$C3,$C7,$FC
	.dc.b	$A9,$7C,$23,$07,$F8,$A5,$7B,$BC,$33,$57,$08,$F0,$1F,$C1,$3E,$A1
	.dc.b	$CD,$8C,$39,$DF,$FB,$46,$73,$18,$86,$1F,$2E,$3F,$C7,$1D,$2F,$81
	.dc.b	$F6,$71,$FF,$10,$8C,$4F,$16,$DF,$92,$38,$FA,$3B,$BE,$0F,$F1,$8F
	.dc.b	$A8,$FF,$35,$4F,$87,$BD,$EE,$67,$D4,$FE,$3B,$18,$9F,$37,$D2,$FF
	.dc.b	$A9,$DB,$ED,$FF,$24,$FA,$87,$D4,$3E,$A3,$F9,$2D,$EF,$6F,$21,$FE
	.dc.b	$F1,$71,$CB,$FE,$92,$F2,$5F,$51,$FE,$77,$CF,$A9,$F2,$23,$3E,$CF
	.dc.b	$4D,$3F,$99,$C6,$27,$78,$F6,$8C,$3E,$0F,$03,$FE,$00,$7D,$47,$F3
	.dc.b	$A5,$1F,$93,$AB,$1A,$7E,$BB,$AF,$03,$84,$AA,$C6,$1B,$5C,$25,$F5
	.dc.b	$1C,$0E,$3C,$B6,$75,$16,$7C,$C7,$FA,$56,$32,$7E,$17,$E9,$9E,$BF
	.dc.b	$F5,$06,$C7,$F3,$5F,$A8,$FA,$D5,$F3,$84,$E9,$C3,$78,$3F,$9B,$4F
	.dc.b	$F2,$2A,$FF,$43,$4E,$5F,$8B,$D4,$9D,$5F,$C8,$AD,$3F,$CD,$E3,$2F
	.dc.b	$83,$0F,$EA,$31,$C3,$5F,$F1,$4E,$1F,$47,$D5,$FC,$8D,$FD,$57,$FD
	.dc.b	$AD,$1F,$41,$CE,$5F,$ED,$6A,$E3,$39,$E8,$E4,$F3,$7A,$7D,$44,$8E
	.dc.b	$5F,$17,$71,$ED,$37,$0E,$98,$AB,$17,$F0,$BE,$43,$E7,$1F,$8E,$F8
	.dc.b	$27,$CB,$62,$7E,$A4,$FF,$75,$FA,$8F,$A3,$8F,$FB,$77,$81,$39,$EF
	.dc.b	$E1,$AF,$F7,$EB,$DE,$1D,$CF,$F0,$33,$88,$6F,$FE,$50,$F8,$EF,$90
	.dc.b	$7C,$9F,$A8,$B7,$F1,$E2,$D7,$CB,$9B,$9F,$BB,$E5,$FE,$87,$C4,$7D
	.dc.b	$16,$E1,$F3,$DB,$7F,$4D,$B5,$FE,$F7,$FA,$8F,$F0,$4C,$FE,$35,$B1
	.dc.b	$FB,$9E,$0F,$E5,$13,$9C,$3F,$C4,$B5,$AC,$3F,$17,$F5,$0F,$96,$CD
	.dc.b	$7F,$85,$A3,$13,$FC,$8E,$B8,$5F,$F8,$0E,$72,$F5,$9F,$F0,$A4,$63
	.dc.b	$1F,$B7,$F5,$1F,$8D,$FA,$8F,$C7,$36,$BF,$18,$DB,$FC,$83,$6B,$F0
	.dc.b	$7C,$3F,$95,$F7,$3F,$2C,$DB,$FC,$A3,$6F,$F3,$0D,$A4,$FF,$BB,$FE
	.dc.b	$9F,$E1,$FC,$27,$BE,$FF,$87,$C3,$E4,$37,$1F,$E8,$6D,$F8,$7A,$D8
	.dc.b	$BE,$12,$BE,$18,$7F,$A0,$AF,$6D,$9E,$B8,$71,$9C,$C7,$53,$33,$B7
	.dc.b	$A4,$ED,$C9,$B3,$C1,$86,$5D,$CF,$1B,$F1,$C6,$D7,$D4,$6D,$CE,$3E
	.dc.b	$A7,$35,$C7,$CE,$E9,$F5,$1F,$B1,$F0,$F1,$3E,$58,$C6,$3D,$6B,$7D
	.dc.b	$DF,$E4,$58,$C5,$6E,$31,$8F,$C8,$BE,$2D,$8C,$EB,$F5,$70,$E7,$8F
	.dc.b	$12,$63,$D0,$F8,$58,$72,$AF,$E8,$F0,$8C,$3F,$D2,$9A,$FD,$7A,$EA
	.dc.b	$F2,$BF,$0B,$B6,$BB,$C7,$C3,$8F,$9F,$1B,$8F,$9E,$6D,$71,$9A,$F1
	.dc.b	$3B,$38,$C7,$C7,$97,$FA,$06,$B1,$8F,$82,$73,$97,$9B,$2F,$06,$3C
	.dc.b	$8F,$84,$E0,$99,$7C,$2A,$E3,$87,$7F,$A7,$FF,$0F,$DF,$FB,$BA,$51
	.dc.b	$88,$FF,$79,$C7,$DC,$6C,$65,$E8,$F2,$BF,$B4,$DB,$CF,$AD,$F9,$9F
	.dc.b	$A6,$7D,$47,$FB,$42,$73,$87,$D8,$BE,$0F,$D5,$6A,$AE,$2F,$E3,$E3
	.dc.b	$38,$FA,$30,$C1,$DF,$0E,$D5,$CE,$32,$F7,$74,$E3,$1F,$4C,$5C,$7F
	.dc.b	$13,$8F,$82,$B8,$12,$B6,$FE,$C6,$35,$18,$24,$FB,$9F,$E0,$B1,$A8
	.dc.b	$C1,$27,$CC,$73,$BF,$4B,$F9,$9C,$7D,$E4,$C3,$D3,$1E,$3A,$E4,$EE
	.dc.b	$33,$87,$8A,$3A,$B5,$F9,$97,$DB,$1F,$95,$1C,$CD,$DF,$C2,$9D,$61
	.dc.b	$E1,$9D,$1C,$8F,$90,$70,$21,$93,$75,$DA,$FC,$58,$FC,$39,$DC,$E7
	.dc.b	$0F,$5C,$E6,$7D,$87,$2B,$FC,$2C,$66,$B8,$4F,$29,$CE,$1E,$8C,$7C
	.dc.b	$B3,$CB,$1F,$A7,$39,$8E,$93,$F2,$31,$F9,$78,$74,$9E,$11,$F0,$DE
	.dc.b	$71,$93,$58,$F8,$F1,$F8,$D8,$FB,$B1,$EC,$FD,$1A,$DC,$63,$0E,$F5
	.dc.b	$E8,$C3,$AB,$CA,$3C,$11,$E6,$3E,$8E,$1F,$14,$F0,$63,$EE,$71,$7F
	.dc.b	$8A,$6C,$FE,$CF,$EA,$31,$F0,$F2,$3F,$6C,$FE,$43,$EA,$1F,$0D,$99
	.dc.b	$FC,$F4,$DD,$7F,$AF,$D6,$A7,$FA,$B3,$5D,$9D,$0C,$D7,$53,$F9,$AF
	.dc.b	$37,$F6,$C8,$4F,$18,$C3,$77,$FE,$76,$BE,$17,$8C,$7F,$3E,$F8,$1F
	.dc.b	$F1,$CD,$FF,$AD,$57,$D4,$6A,$3F,$D1,$31,$87,$5E,$3E,$5F,$F6,$9B
	.dc.b	$63,$51,$FE,$75,$6E,$3E,$77,$BC,$67,$DC,$E1,$F4,$32,$3C,$51,$CC
	.dc.b	$C3,$CE,$30,$FF,$B2,$FE,$A1,$F5,$1E,$1F,$4A,$4E,$63,$CD,$3A,$8C
	.dc.b	$10,$F3,$6F,$A9,$91,$9C,$BD,$13,$EF,$FA,$87,$55,$F8,$49,$7F,$ED
	.dc.b	$D9,$CB,$67,$2F,$FA,$5E,$3F,$DD,$AE,$4E,$1F,$27,$D4,$8F,$3F,$8A
	.dc.b	$AF,$18,$B8,$FE,$06,$95,$F4,$FE,$A1,$E6,$F9,$FD,$39,$8F,$FB,$C6
	.dc.b	$72,$D9,$E0,$47,$D5,$DB,$5F,$F5,$74,$F0,$6C,$FC,$39,$FC,$17,$D4
	.dc.b	$3E,$A2,$92,$BF,$E0,$04,$3E,$6F,$11,$FE,$B9,$8E,$0E,$A7,$9C,$F4
	.dc.b	$F9,$EE,$BF,$4B,$2B,$9C,$65,$EB,$19,$7A,$47,$E5,$7C,$6F,$F2,$6E
	.dc.b	$FF,$A7,$FA,$77,$D4,$7E,$65,$B7,$FE,$E6,$9F,$D6,$57,$0E,$EF,$F1
	.dc.b	$1B,$67,$FD,$D6,$D7,$FD,$59,$E3,$73,$EE,$F0,$32,$3A,$57,$28,$D4
	.dc.b	$60,$FF,$7A,$E0,$ED,$F2,$12,$34,$21,$DB,$98,$FF,$71,$47,$17,$6D
	.dc.b	$73,$1E,$FE,$AF,$81,$E0,$49,$DC,$65,$E3,$19,$7A,$C7,$CB,$EB,$FE
	.dc.b	$01,$EB,$BE,$A2,$99,$AF,$3E,$DF,$9E,$7F,$DE,$CD,$DF,$E6,$DF,$2E
	.dc.b	$DE,$15,$A8,$B3,$FC,$77,$0F,$D5,$CF,$FB,$FF,$EA,$1C,$9F,$C4,$5B
	.dc.b	$1F,$19,$AE,$FE,$A3,$5E,$94,$8D,$3F,$C6,$7E,$9F,$F5,$31,$F3,$9A
	.dc.b	$B1,$FE,$5D,$18,$6F,$7F,$F2,$1F,$A8,$FE,$37,$1A,$F8,$9E,$84,$BF
	.dc.b	$B3,$F5,$1F,$CC,$D0,$70,$DE,$2F,$E5,$91,$8F,$A8,$7C,$9F,$F2,$F6
	.dc.b	$D7,$4A,$D9,$99,$FE,$69,$EB,$FE,$BE,$7F,$13,$5F,$EB,$36,$FB,$1A
	.dc.b	$F0,$24,$7E,$3A,$B8,$3F,$07,$B7,$FD,$07,$18,$D3,$FA,$86,$1F,$E5
	.dc.b	$F7,$FA,$94,$66,$77,$7D,$2F,$91,$F0,$2F,$84,$FF,$A3,$BD,$FF,$5C
	.dc.b	$E8,$7D,$23,$E0,$3F,$A7,$36,$3F,$47,$3C,$32,$FE,$C4,$DB,$E9,$A7
	.dc.b	$F5,$58,$C5,$7D,$0B,$7C,$3E,$6A,$4F,$47,$BC,$7F,$5D,$F1,$1F,$1D
	.dc.b	$FD,$7F,$E9,$5F,$19,$4D,$44,$C3,$F6,$F5,$C1,$B1,$C7,$E3,$B5,$1E
	.dc.b	$F9,$D3,$67,$4F,$F0,$0F,$A7,$7D,$17,$87,$EA,$24,$E6,$3F,$B9,$4E
	.dc.b	$A3,$04,$3E,$66,$5F,$C0,$F8,$3F,$C1,$13,$B9,$FA,$0D,$8C,$D7,$BF
	.dc.b	$C8,$FE,$FD,$F5,$1F,$76,$FF,$33,$83,$D8,$F8,$86,$B9,$A4,$FF,$18
	.dc.b	$AE,$09,$B4,$FA,$0D,$CF,$E0,$67,$66,$A3,$C3,$F0,$DF,$47,$DA,$95
	.dc.b	$D6,$32,$D2,$1F,$03,$B3,$FC,$2D,$F5,$1F,$8C,$8F,$E5,$6D,$7F,$20
	.dc.b	$DD,$BF,$C5,$9F,$50,$F6,$BF,$81,$53,$EA,$1F,$51,$FE,$E6,$C3,$87
	.dc.b	$6D,$EC,$EE,$7F,$89,$71,$E5,$97,$F0,$0C,$38,$E3,$FC,$0A,$72,$F5
	.dc.b	$C7,$93,$B6,$B9,$38,$65,$E4,$AE,$70,$AE,$71,$8C,$BD,$CD,$C7,$FD
	.dc.b	$FA,$F0,$9D,$E1,$FF,$00,$1B,$7F,$C5,$AF,$8C,$F8,$AB,$86,$1E,$18
	.dc.b	$E0,$DA,$E1,$87,$69,$E0,$6B,$1F,$64,$FA,$9A,$B0,$EE,$F2,$33,$8F
	.dc.b	$91,$F7,$61,$DD,$E1,$87,$77,$81,$F1,$CD,$DF,$29,$E1,$7D,$A7,$89
	.dc.b	$99,$8F,$F3,$63,$D0,$53,$D0,$53,$A9,$6F,$FC,$67,$F2,$79,$8F,$A9
	.dc.b	$E3,$23,$13,$A9,$E1,$39,$78,$0F,$82,$AD,$32,$E0,$8F,$BF,$E6,$1F
	.dc.b	$4C,$DF,$E6,$CF,$F5,$0A,$F9,$F5,$A9,$CC,$F5,$DF,$F1,$73,$98,$F9
	.dc.b	$47,$13,$B5,$F5,$72,$71,$DF,$E4,$4F,$1A,$FC,$AC,$F1,$7A,$D7,$17
	.dc.b	$31,$C4,$FE,$CB,$3C,$6B,$E7,$61,$E8,$BE,$11,$8A,$F1,$E1,$EA,$AD
	.dc.b	$36,$32,$79,$A3,$2D,$33,$87,$8A,$32,$F9,$A7,$FC,$63,$87,$C3,$D3
	.dc.b	$FD,$61,$7E,$BC,$3D,$F8,$FC,$39,$FF,$5A,$56,$EF,$E1,$F0,$65,$D9
	.dc.b	$D1,$C6,$1F,$03,$A6,$DB,$EA,$1D,$FA,$A6,$3F,$D1,$31,$FF,$25,$5E
	.dc.b	$EB,$53,$9A,$FE,$E3,$39,$AF,$5E,$1D,$B0,$EB,$1E,$EC,$38,$9E,$4F
	.dc.b	$8D,$FF,$4F,$F8,$B8,$9D,$1F,$E5,$98,$7D,$43,$E2,$35,$9F,$F8,$02
	.dc.b	$73,$18,$CF,$F1,$11,$8B,$E7,$5B,$BE,$51,$FE,$60,$F8,$07,$4F,$C1
	.dc.b	$6C,$7F,$9A,$47,$D3,$30,$D5,$49,$D4,$FD,$3C,$3F,$C6,$37,$C9,$FF
	.dc.b	$0A,$CF,$2E,$1E,$11,$8A,$E0,$AB,$86,$5F,$A1,$BE,$07,$F9,$DB,$EA
	.dc.b	$3F,$DD,$CD,$F6,$FE,$EB,$E6,$E3,$FB,$53,$AC,$3F,$C4,$0D,$BE,$98
	.dc.b	$DB,$FE,$25,$36,$BF,$79,$35,$46,$67,$E1,$E1,$D8,$F0,$D6,$E3,$FD
	.dc.b	$8E,$7F,$A3,$0B,$5E,$FE,$EF,$C7,$46,$30,$FD,$F4,$FF,$0B,$E8,$F5
	.dc.b	$FD,$47,$F9,$52,$B9,$46,$3C,$BE,$2D,$8C,$9F,$5E,$78,$57,$F8,$FF
	.dc.b	$E6,$3F,$22,$DE,$5F,$3C,$FE,$7E,$73,$8F,$89,$B7,$FE,$6A,$3F,$03
	.dc.b	$E2,$EE,$FD,$ED,$8F,$C0,$B7,$0F,$B8,$FE,$AF,$E9,$5E,$85,$3F,$5D
	.dc.b	$F5,$0F,$A8,$C6,$1F,$EB,$2A,$DB,$7D,$DC,$E3,$FE,$59,$94,$75,$3F
	.dc.b	$71,$F5,$1F,$89,$AE,$07,$FA,$F1,$C5,$71,$E0,$FD,$F7,$4F,$FB,$0E
	.dc.b	$BC,$11,$8F,$32,$47,$FB,$12,$7F,$CA,$6D,$8E,$3E,$8F,$F0,$7F,$A8
	.dc.b	$FF,$2F,$CF,$FB,$2B,$E0,$1E,$E6,$7D,$5F,$E1,$91,$89,$FF,$67,$FD
	.dc.b	$2B,$DB,$FE,$20,$D8,$FF,$6C,$7D,$43,$EA,$1F,$51,$AB,$95,$FE,$DD
	.dc.b	$F2,$9F,$51,$FE,$6A,$8E,$B3,$FC,$82,$73,$ED,$70,$E3,$FF,$43,$C6
	.dc.b	$2B,$67,$3B,$E5,$3F,$E8,$CE,$CF,$F5,$5F,$D4,$7F,$2A,$7F,$D5,$BD
	.dc.b	$B6,$53,$63,$0D,$BF,$F7,$95,$72,$6D,$7E,$38,$47,$FA,$E3,$EA,$3A
	.dc.b	$1D,$79,$2D,$70,$8B,$5E,$53,$EA,$95,$FE,$75,$F4,$CF,$62,$9F,$CB
	.dc.b	$FE,$A3,$E9,$57,$FA,$8E,$75,$7F,$EC,$8F,$0B,$F6,$71,$89,$FE,$5D
	.dc.b	$5F,$C5,$2B,$FC,$31,$39,$AF,$E5,$F3,$9A,$FF,$68,$CE,$6B,$FD,$A7
	.dc.b	$D4,$AD,$52,$B4,$FF,$2D,$8E,$59,$FF,$86,$EF,$E8,$7A,$DF,$8E,$AF
	.dc.b	$A8,$FF,$84,$63,$E6,$B9,$C7,$FC,$11,$5C,$67,$3D,$1C,$9E,$6F,$4F
	.dc.b	$A8,$91,$CB,$E2,$EE,$3D,$A6,$E1,$D3,$15,$62,$FE,$1F,$C8,$7D,$03
	.dc.b	$F2,$1F,$06,$F9,$6C,$4F,$CD,$36,$7F,$6C,$FA,$8F,$A7,$8F,$D0,$BF
	.dc.b	$EE,$B1,$89,$DE,$7F,$DB,$A3,$0F,$CE,$9D,$F7,$FC,$6F,$8E,$97,$F1
	.dc.b	$3E,$A2,$DF,$C8,$9B,$8F,$AE,$6D,$2B,$F2,$7A,$BE,$5F,$E8,$FB,$3F
	.dc.b	$49,$B5,$F4,$1B,$5F,$51,$B8,$7F,$3C,$FA,$8F,$F7,$75,$7C,$1A,$ED
	.dc.b	$18,$7F,$67,$BF,$F2,$07,$FD,$ED,$19,$AE,$58,$7F,$81,$FE,$A1,$F2
	.dc.b	$DF,$E1,$7A,$D6,$3F,$8E,$9C,$BD,$F0,$4E,$9F,$EF,$53,$A7,$BD,$F5
	.dc.b	$BF,$C7,$FD,$3F,$E4,$3E,$A2,$E3,$F8,$E6,$DF,$E4,$9B,$97,$E1,$5F
	.dc.b	$CB,$F0,$7E,$63,$DC,$B7,$F9,$66,$DF,$E6,$9B,$7F,$8D,$7F,$C6,$5F
	.dc.b	$4F,$F0,$FE,$13,$DF,$5F,$56,$1F,$21,$FE,$65,$9C,$CA,$78,$69,$FE
	.dc.b	$8D,$77,$8F,$F9,$D6,$31,$EA,$F1,$75,$85,$BE,$D3,$5E,$91,$6F,$9C
	.dc.b	$7F,$A2,$6B,$8D,$F5,$8E,$2D,$78,$B6,$78,$E1,$E8,$5F,$51,$D7,$0E
	.dc.b	$CF,$7F,$3B,$C5,$F5,$1F,$B0,$F0,$B9,$46,$3B,$7F,$C6,$99,$F8,$CF
	.dc.b	$8D,$19,$9D,$E7,$E3,$BE,$55,$B0,$E3,$3F,$96,$DD,$3E,$56,$1C,$EB
	.dc.b	$84,$E5,$FD,$43,$5F,$A5,$1B,$AE,$78,$F9,$9B,$8F,$92,$B9,$CE,$70
	.dc.b	$FF,$62,$61,$CC,$E9,$A3,$1E,$93,$AA,$E1,$7E,$26,$BE,$2C,$BE,$04
	.dc.b	$63,$0E,$8E,$B0,$E9,$3A,$C6,$5F,$91,$F3,$65,$E8,$8C,$11,$FB,$59
	.dc.b	$79,$23,$C5,$3C,$A7,$35,$72,$F6,$FD,$47,$F9,$7E,$FF,$DB,$D9,$A3
	.dc.b	$11,$FE,$E3,$8F,$AE,$D8,$CB,$CB,$E0,$7E,$E4,$73,$C3,$EC,$B6,$FF
	.dc.b	$DD,$DF,$4E,$FA,$8F,$F7,$85,$6A,$F8,$93,$0F,$C2,$56,$DC,$4E,$DF
	.dc.b	$C0,$E1,$D5,$FE,$51,$3B,$73,$5C,$30,$F3,$47,$4C,$3C,$31,$8C,$3A
	.dc.b	$11,$F3,$C6,$A7,$F1,$B8,$77,$7F,$2E,$DC,$7E,$46,$1F,$AF,$C7,$C5
	.dc.b	$8F,$3B,$E3,$3E,$37,$CF,$87,$57,$9C,$65,$A7,$9C,$C5,$F6,$7D,$A6
	.dc.b	$4B,$5D,$DE,$2F,$F1,$4B,$F1,$E3,$EF,$24,$F9,$0E,$53,$C6,$BB,$3C
	.dc.b	$EB,$8C,$EE,$35,$8F,$C3,$9C,$CF,$82,$FA,$B4,$F7,$9E,$A9,$F9,$07
	.dc.b	$1C,$7E,$16,$1E,$0A,$E2,$DA,$E2,$7B,$A3,$35,$C6,$3A,$63,$F2,$70
	.dc.b	$F1,$DF,$19,$F1,$12,$33,$5E,$2A,$E1,$87,$8A,$31,$5D,$CF,$0E,$3F
	.dc.b	$02,$B7,$1E,$03,$42,$FA,$C6,$67,$E9,$5F,$58,$C6,$3F,$8E,$BA,$7C
	.dc.b	$D7,$0F,$4C,$7D,$47,$AF,$0E,$B1,$9C,$7E,$13,$F6,$AB,$E2,$DF,$E4
	.dc.b	$F0,$FC,$75,$FD,$62,$38,$FA,$8C,$FC,$3C,$C3,$F3,$4F,$E5,$7E,$A1
	.dc.b	$F0,$D9,$9F,$DA,$CD,$D7,$FB,$0D,$6A,$7F,$A8,$35,$D9,$D0,$CD,$75
	.dc.b	$3F,$C4,$0B,$9F,$FD,$A7,$3C,$5B,$3A,$73,$F0,$B7,$83,$8F,$C0,$BC
	.dc.b	$5C,$63,$FE,$2B,$87,$D4,$62,$11,$87,$FC,$2C,$F1,$6C,$EE,$35,$1E
	.dc.b	$5A,$EF,$59,$AD,$7B,$9B,$FA,$09,$7D,$23,$2F,$68,$CB,$CC,$F3,$47
	.dc.b	$97,$EA,$1F,$51,$E1,$F4,$A5,$7F,$8A,$E7,$2D,$9E,$04,$7D,$7B,$57
	.dc.b	$B0,$9E,$2E,$1B,$B3,$4F,$97,$5F,$1F,$EA,$3F,$B0,$5F,$F8,$ED,$FF
	.dc.b	$70,$B7,$E0,$F5,$D9,$23,$04,$3C,$33,$C9,$BF,$29,$D4,$8F,$5F,$89
	.dc.b	$AE,$8F,$04,$78,$E3,$ED,$7D,$47,$F9,$96,$7C,$9E,$97,$9C,$CF,$9D
	.dc.b	$23,$FC,$A5,$3B,$7E,$1B,$75,$F8,$11,$FE,$71,$C7,$FB,$0C,$62,$13
	.dc.b	$BD,$7F,$2E,$D8,$FF,$3C,$35,$57,$29,$FC,$EF,$D4,$3E,$A3,$F2,$4D
	.dc.b	$AF,$F3,$33,$5F,$AE,$BE,$6F,$73,$27,$6F,$28,$E7,$F3,$DD,$3E,$96
	.dc.b	$57,$48,$CB,$DA,$32,$F6,$8F,$D2,$79,$1F,$E5,$AD,$3E,$07,$D3,$BE
	.dc.b	$A3,$F5,$0D,$BF,$F3,$E5,$6F,$4F,$DC,$B6,$3F,$D3,$4D,$7F,$D6,$FE
	.dc.b	$37,$2E,$CE,$EC,$8E,$6E,$5F,$F4,$45,$BA,$FC,$24,$8C,$F6,$57,$F3
	.dc.b	$B3,$B7,$8C,$FA,$27,$D3,$D3,$EA,$EF,$95,$DA,$32,$F8,$23,$2F,$48
	.dc.b	$F8,$BD,$5F,$C2,$7D,$87,$D4,$53,$35,$F1,$26,$BF,$B0,$D7,$F4,$EF
	.dc.b	$EC,$AB,$F4,$5C,$1F,$8D,$9C,$B6,$B8,$BF,$91,$9D,$EF,$FC,$9B,$EA
	.dc.b	$1C,$7F,$C6,$DB,$8F,$FA,$BD,$AF,$F0,$05,$F5,$1B,$F4,$E4,$ED,$FE
	.dc.b	$49,$F5,$0E,$EA,$49,$FF,$5D,$46,$5F,$F5,$E7,$D4,$7F,$27,$7F,$86
	.dc.b	$C3,$E2,$F9,$92,$3F,$D8,$5F,$51,$FC,$EF,$4F,$E7,$ED,$8F,$F7,$03
	.dc.b	$53,$7E,$A1,$F2,$7F,$CE,$5B,$3D,$39,$3F,$9E,$56,$EB,$F9,$F7,$B1
	.dc.b	$FA,$9A,$FF,$69,$5F,$FA,$E1,$BE,$C6,$BB,$E4,$7F,$B5,$EB,$83,$FC
	.dc.b	$A1,$BC,$9F,$C0,$EF,$F7,$93,$FE,$F7,$9C,$E7,$FD,$69,$B1,$9B,$F4
	.dc.b	$7A,$7F,$D4,$7D,$FF,$4C,$FF,$34,$FA,$47,$BD,$FD,$55,$B1,$FE,$E9
	.dc.b	$9D,$E7,$FC,$4D,$B9,$7F,$68,$6C,$83,$85,$4F,$85,$F1,$6D,$27,$B3
	.dc.b	$E1,$8F,$EC,$DE,$23,$E3,$FF,$B4,$7D,$33,$E3,$29,$A8,$98,$7F,$BF
	.dc.b	$EB,$83,$63,$8F,$C7,$6A,$3F,$A9,$CE,$9B,$3A,$7D,$5A,$FF,$31,$FA
	.dc.b	$77,$D1,$78,$BE,$A2,$56,$B8,$3F,$C1,$35,$A8,$C1,$F0,$DC,$3F,$D8
	.dc.b	$9B,$7D,$6F,$DB,$E3,$FF,$81,$67,$37,$D2,$70,$6A,$FA,$8F,$A8,$FB
	.dc.b	$EB,$6F,$F0,$6B,$A6,$BF,$DC,$A3,$64,$E5,$FF,$0A,$3C,$A5,$F3,$F8
	.dc.b	$99,$5E,$0A,$D4,$F0,$F8,$6F,$AB,$EC,$CA,$D4,$65,$A4,$3F,$C0,$78
	.dc.b	$7E,$BF,$EA,$3F,$0B,$1E,$C6,$BF,$E0,$86,$ED,$FD,$8F,$EA,$1E,$C7
	.dc.b	$F8,$DE,$9F,$50,$FA,$8F,$F1,$DE,$1C,$38,$2F,$57,$F8,$C6,$31,$7F
	.dc.b	$97,$C7,$9E,$5F,$99,$C7,$94,$E7,$2E,$D1,$88,$5F,$9A,$30,$DC,$BF
	.dc.b	$C9,$78,$74,$AF,$2C,$62,$F9,$BE,$48,$F0,$35,$D9,$E1,$C3,$2D,$32
	.dc.b	$78,$B1,$FF,$2C,$E3,$E2,$6D,$77,$AE,$34,$C3,$C9,$7B,$8C,$E3,$E9
	.dc.b	$7B,$47,$BE,$77,$7E,$58,$E1,$3E,$29,$CB,$C2,$74,$47,$04,$3D,$07
	.dc.b	$83,$0F,$3B,$99,$F9,$0F,$F9,$D0,$F4,$63,$F7,$DC,$46,$9F,$71,$B3
	.dc.b	$DD,$E3,$3E,$3F,$11,$86,$CF,$F4,$F6,$CE,$E7,$63,$E0,$AB,$4C,$B8
	.dc.b	$23,$F5,$3E,$65,$F4,$DF,$E3,$CD,$3F,$11,$3F,$C9,$6B,$E9,$D6,$B7
	.dc.b	$FA,$B3,$93,$95,$F1,$D7,$D9,$7C,$23,$13,$F1,$63,$15,$CA,$3C,$D8
	.dc.b	$7E,$5F,$0E,$D1,$F8,$83,$B4,$66,$F9,$C7,$FA,$8A,$F9,$BB,$39,$E1
	.dc.b	$FE,$A6,$BE,$77,$EC,$9C,$9F,$D0,$63,$33,$E8,$8F,$5E,$1E,$5B,$EB
	.dc.b	$7C,$5B,$5B,$8F,$24,$F8,$B7,$FF,$31,$61,$EA,$7F,$CC,$98,$7A,$AF
	.dc.b	$C9,$8F,$F9,$AA,$76,$DB,$F4,$78,$F6,$FC,$D3,$56,$1F,$EC,$EA,$FE
	.dc.b	$5A,$FF,$40,$7F,$9D,$DB,$EA,$1D,$DC,$6B,$57,$FE,$AB,$AD,$4E,$7A
	.dc.b	$7C,$11,$8C,$3A,$C6,$30,$E7,$1E,$0A,$E8,$79,$FE,$31,$F0,$3F,$DD
	.dc.b	$A7,$44,$C7,$FD,$75,$2F,$A8,$7C,$47,$2A,$FE,$A7,$39,$8C,$57,$0A
	.dc.b	$DD,$73,$CF,$FD,$27,$7B,$8F,$E9,$FE,$43,$AB,$E4,$36,$3F,$AF,$8F
	.dc.b	$A7,$56,$AA,$4E,$A7,$ED,$3F,$EF,$0A,$E5,$8F,$FB,$4E,$39,$70,$70
	.dc.b	$8C,$57,$07,$FC,$1D,$5C,$31,$FF,$54,$DF,$03,$FB,$2F,$D4,$3D,$AF
	.dc.b	$BC,$D8,$F1,$71,$E2,$64,$79,$5B,$87,$81,$B7,$DB,$1E,$53,$FE,$DC
	.dc.b	$C2,$CE,$6B,$F1,$97,$FE,$5B,$3F,$BB,$96,$BD,$7D,$FA,$E8,$93,$F2
	.dc.b	$BE,$A3,$FC,$9E,$F0,$8F,$CC,$28,$CD,$7F,$BB,$23,$A6,$1F,$0A,$72
	.dc.b	$FF,$B2,$63,$27,$F8,$D4,$E5,$E5,$7E,$87,$E6,$BF,$23,$7F,$EF,$66
	.dc.b	$D7,$BE,$F5,$CB,$30,$FB,$5F,$17,$70,$FA,$AD,$BF,$8C,$D9,$F0,$6C
	.dc.b	$8F,$D8,$7D,$33,$D0,$A7,$EC,$FE,$A1,$F5,$18,$C3,$FC,$3F,$5B,$6F
	.dc.b	$BD,$9C,$D9,$FF,$B9,$E9,$3D,$8F,$DF,$7D,$43,$9B,$18,$7F,$8A,$A3
	.dc.b	$0F,$6E,$1F,$E0,$9D,$3F,$EC,$18,$C1,$FE,$30,$71,$97,$82,$3F,$CC
	.dc.b	$0F,$E8,$3C,$DE,$3E,$8F,$26,$CF,$F8,$E3,$EA,$3F,$CC,$4F,$F8,$EB
	.dc.b	$BD,$EE,$7F,$13,$9C,$E1,$D4,$FF,$36,$4E,$AB,$FC,$83,$F5,$1F,$E9
	.dc.b	$F6,$FB,$7F,$C7,$7E,$A1,$F5,$0F,$A8,$FE,$3D,$7B,$C7,$A0,$71,$C8
	.dc.b	$C7,$F9,$D7,$D4,$7F,$9C,$5E,$A7,$F9,$73,$C9,$46,$6F,$9D,$F2,$AE
	.dc.b	$3E,$C7,$83,$4F,$E6,$18,$70,$8F,$E6,$5D,$FF,$E8,$7F,$A8,$B8,$FE
	.dc.b	$43,$0F,$F4,$54,$E5,$FF,$35,$BF,$E8,$DE,$FD,$D7,$06,$CE,$AB,$90
	.dc.b	$AF,$F4,$A7,$D4,$76,$3B,$F2,$DA,$E1,$16,$BE,$85,$7F,$A6,$EB,$1B
	.dc.b	$7C,$59,$FF,$4F,$7D,$43,$D7,$FE,$94,$DC,$7F,$97,$FD,$47,$D7,$AF
	.dc.b	$9F,$39,$6D,$7F,$3D,$9F,$E3,$55,$F8,$AA,$D5,$FF,$32,$9C,$DF,$82
	.dc.b	$F8,$79,$7E,$06,$E1,$F1,$BD,$3F,$B4,$DF,$7B,$E7,$C6,$2F,$DA,$FE
	.dc.b	$3B,$1F,$66,$9F,$E2,$7B,$DD,$F0,$69,$D7,$EA,$39,$9D,$3E,$2E,$C6
	.dc.b	$49,$8F,$68,$CB,$F0,$BE,$43,$E3,$9F,$84,$F8,$27,$CB,$E2,$FF,$D3
	.dc.b	$BC,$0A,$FE,$D3,$F5,$1F,$2F,$5F,$A5,$8A,$B5,$7F,$DD,$AB,$4F,$CB
	.dc.b	$BE,$3E,$07,$5B,$FE,$ED,$5A,$7F,$0D,$BB,$E3,$BE,$39,$F2,$3E,$A2
	.dc.b	$E3,$F3,$62,$D7,$CA,$9B,$9F,$CA,$B7,$CC,$FD,$0F,$88,$FA,$2D,$C7
	.dc.b	$EB,$37,$0F,$A6,$DB,$F6,$FD,$47,$F7,$F9,$FD,$BB,$63,$F2,$5D,$BF
	.dc.b	$EE,$EF,$3F,$FC,$35,$39,$C3,$F8,$9B,$6F,$FB,$9F,$D4,$3E,$5B,$79
	.dc.b	$FF,$83,$63,$17,$FE,$CC,$BE,$1E,$52,$73,$87,$8A,$F8,$61,$F8,$CF
	.dc.b	$A8,$FC,$6F,$D4,$5C,$BF,$24,$DB,$FC,$7B,$75,$F8,$CF,$E5,$38,$FF
	.dc.b	$2B,$DB,$6F,$F3,$4D,$BF,$CB,$B6,$FE,$73,$FE,$39,$FA,$7F,$E8,$93
	.dc.b	$FE,$77,$CB,$F9,$9C,$3E,$43,$F7,$1F,$0B,$F0,$68,$27,$8C,$66,$31
	.dc.b	$3C,$2B,$9F,$0D,$9E,$37,$D7,$2F,$F3,$E5,$69,$B3,$AB,$E0,$F4,$8C
	.dc.b	$B4,$F4,$1D,$2F,$77,$D6,$32,$6F,$0F,$34,$3E,$A3,$78,$70,$78,$F9
	.dc.b	$DE,$0F,$A8,$77,$FF,$8D,$31,$E7,$87,$F9,$33,$3F,$1D,$6B,$93,$CB
	.dc.b	$39,$AF,$0C,$E5,$E3,$87,$5C,$3A,$67,$F2,$6B,$4D,$8C,$EB,$F5,$2B
	.dc.b	$86,$3E,$09,$CE,$1F,$53,$0E,$37,$BC,$B5,$72,$BC,$77,$C1,$FB,$58
	.dc.b	$74,$73,$3B,$C7,$E3,$CE,$5F,$24,$67,$0E,$D5,$C0,$93,$A7,$DF,$86
	.dc.b	$DF,$36,$1E,$1C,$7C,$F1,$F8,$4C,$7B,$61,$FE,$04,$9C,$C9,$B9,$7C
	.dc.b	$DC,$7B,$1F,$27,$EA,$3F,$C2,$D5,$FE,$DB,$A4,$62,$3F,$DC,$91,$F4
	.dc.b	$DB,$19,$7C,$0F,$29,$E9,$EA,$FD,$E8,$C6,$1F,$E1,$8F,$A8,$7D,$47
	.dc.b	$FB,$06,$38,$D6,$C9,$87,$E4,$AB,$6E,$30,$EA,$FE,$1B,$0E,$AF,$EB
	.dc.b	$A7,$6E,$6B,$85,$F6,$8E,$98,$78,$5F,$C1,$5F,$27,$E4,$37,$1F,$87
	.dc.b	$7F,$AA,$C3,$AE,$3E,$22,$63,$D5,$FD,$E4,$6A,$30,$4C,$3B,$D7,$3C
	.dc.b	$7C,$51,$E0,$8F,$3B,$A3,$CE,$60,$D1,$ED,$32,$5B,$F1,$BC,$5F,$1C
	.dc.b	$F6,$C3,$D7,$3B,$BF,$22,$C3,$D0,$F1,$8C,$CF,$0B,$E7,$3E,$38,$C4
	.dc.b	$F2,$AE,$33,$E6,$AF,$1B,$67,$F0,$57,$E3,$C7,$E7,$13,$1F,$89,$5D
	.dc.b	$63,$0F,$BA,$F9,$E3,$E0,$8F,$31,$F8,$1B,$E9,$7C,$4E,$71,$EF,$36
	.dc.b	$78,$CE,$06,$CD,$1F,$0E,$39,$B8,$9F,$03,$93,$85,$7B,$B1,$F8,$33
	.dc.b	$A2,$61,$F8,$D8,$CB,$EC,$9E,$17,$D9,$F6,$C6,$E3,$18,$FD,$28,$C6
	.dc.b	$1E,$08,$FA,$13,$A8,$F6,$E1,$E6,$C3,$BE,$3B,$9F,$79,$F2,$EF,$A4
	.dc.b	$7C,$6C,$7E,$EC,$E4,$DB,$F9,$CF,$A8,$FC,$BD,$FF,$C0,$09,$FD,$27
	.dc.b	$C3,$E0,$FC,$DF,$D4,$3E,$1B,$13,$A2,$1F,$D3,$A3,$4F,$F5,$66,$BF
	.dc.b	$C0,$67,$8B,$B9,$EA,$7F,$A8,$B9,$FF,$C6,$B5,$C1,$B1,$9E,$2F,$F5
	.dc.b	$3C,$FF,$BD,$6B,$FB,$9D,$7F,$8F,$3E,$03,$FC,$81,$18,$AF,$F5,$6C
	.dc.b	$3E,$A3,$94,$FF,$7A,$8C,$3B,$31,$E3,$69,$FF,$68,$36,$37,$1F,$EA
	.dc.b	$26,$DF,$86,$75,$ED,$EF,$E8,$25,$F5,$3F,$DB,$51,$A7,$16,$9D,$C7
	.dc.b	$59,$F1,$7D,$3B,$EA,$3C,$5E,$8C,$AE,$93,$96,$CF,$02,$3D,$B6,$AF
	.dc.b	$FA,$EA,$78,$36,$7D,$D3,$F0,$7E,$9D,$D5,$7E,$0A,$5F,$AE,$72,$D9
	.dc.b	$CB,$FE,$C6,$8F,$F7,$5B,$93,$87,$C9,$F5,$23,$CF,$E2,$6B,$C6,$2F
	.dc.b	$07,$F9,$E2,$95,$F5,$7E,$A3,$FC,$CF,$7E,$3F,$4B,$D0,$67,$CD,$91
	.dc.b	$F4,$67,$6F,$FB,$69,$B8,$75,$EB,$DC,$E6,$AC,$E6,$3E,$D7,$D4,$3E
	.dc.b	$A3,$F0,$CD,$8C,$C6,$3E,$6F,$2C,$8C,$9C,$A3,$8F,$CF,$72,$FA,$59
	.dc.b	$3D,$A3,$2F,$82,$32,$FF,$BC,$AD,$3F,$97,$F1,$BF,$CA,$58,$FF,$A7
	.dc.b	$BE,$9D,$F5,$1F,$9B,$6E,$1F,$93,$AD,$F7,$7F,$97,$23,$FC,$7F,$E2
	.dc.b	$73,$23,$F9,$9A,$FD,$6A,$C3,$FD,$57,$7C,$A3,$FC,$CA,$D7,$99,$C7
	.dc.b	$BB,$C5,$D9,$23,$A3,$E5,$ED,$C4,$75,$CB,$C9,$8A,$B3,$8F,$B6,$FD
	.dc.b	$9D,$5E,$DF,$93,$98,$F9,$E3,$2F,$A6,$32,$F5,$3F,$CF,$78,$7F,$0F
	.dc.b	$C7,$E1,$67,$F9,$89,$D1,$FE,$BD,$FA,$8F,$E0,$9E,$C9,$FA,$1D,$B2
	.dc.b	$7E,$63,$5E,$33,$FA,$67,$9D,$75,$D3,$E4,$D6,$DF,$F6,$AD,$6D,$FF
	.dc.b	$6B,$FD,$43,$93,$F8,$A3,$70,$FE,$18,$D7,$CE,$BE,$A3,$7E,$9C,$9D
	.dc.b	$BF,$C7,$7E,$9F,$F5,$71,$FC,$55,$AF,$C0,$6E,$1F,$EE,$86,$EB,$F8
	.dc.b	$7F,$A8,$FC,$1C,$E7,$E2,$BA,$12,$FF,$90,$7D,$47,$CD,$7F,$D5,$2D
	.dc.b	$4D,$31,$F5,$0F,$93,$FE,$64,$D9,$E7,$CB,$FC,$D6,$B7,$5F,$CD,$FD
	.dc.b	$87,$77,$0F,$62,$61,$FC,$C6,$72,$D9,$E0,$47,$F9,$B6,$3F,$E6,$7C
	.dc.b	$DA,$FE,$96,$F8,$D7,$E1,$1B,$3A,$8F,$F6,$2C,$62,$7F,$59,$5C,$DF
	.dc.b	$D7,$7C,$27,$AE,$BF,$1B,$F4,$6F,$7B,$FA,$83,$63,$FA,$44,$EF,$1E
	.dc.b	$D0,$8C,$5F,$4D,$3F,$AC,$46,$27,$FD,$AF,$6F,$87,$CD,$49,$EA,$F8
	.dc.b	$23,$FB,$07,$84,$F8,$EF,$EC,$5F,$4A,$F8,$CA,$6A,$26,$1F,$D7,$6B
	.dc.b	$83,$63,$8F,$C7,$6A,$3F,$57,$3A,$6C,$E9,$FE,$CD,$F4,$EF,$A2,$F0
	.dc.b	$FD,$44,$9C,$C7,$F7,$59,$D4,$60,$87,$EE,$32,$FE,$10,$E3,$83,$E4
	.dc.b	$CF,$08,$FD,$EB,$63,$E6,$CF,$09,$FE,$03,$F5,$1F,$6D,$FF,$80,$30
	.dc.b	$F6,$3E,$29,$FE,$07,$9D,$49,$A4,$E6,$F9,$7F,$41,$B6,$4F,$F8,$16
	.dc.b	$78,$DA,$BA,$7C,$37,$D2,$F6,$A5,$EE,$32,$D2,$1F,$07,$B3,$FC,$35
	.dc.b	$F5,$0E,$6A,$49,$FF,$0A,$7D,$43,$D9,$FE,$79,$0F,$A8,$7D,$47,$F4
	.dc.b	$5C,$39,$4F,$F8,$6D,$AF,$F2,$66,$F0,$FE,$5B,$0E,$8A,$7A,$57,$F4
	.dc.b	$F9,$CC,$60,$D5,$74,$BF,$0C,$63,$1F,$EA,$D1,$88,$57,$28,$C3,$6F
	.dc.b	$D1,$5C,$1B,$5E,$3A,$DE,$3F,$01,$B8,$E6,$B7,$5E,$51,$8F,$96,$3D
	.dc.b	$57,$C5,$B5,$C6,$BD,$05,$38,$E1,$E3,$7E,$26,$3E,$76,$DF,$37,$89
	.dc.b	$E1,$C3,$C0,$EA,$FA,$57,$18,$EA,$77,$9D,$BF,$1A,$BA,$36,$B5,$7E
	.dc.b	$38,$E2,$7B,$EF,$A1,$F5,$9F,$F0,$31,$C4,$A5,$32,$ED,$FB,$DE,$D3
	.dc.b	$E4,$37,$1F,$62,$30,$F2,$9F,$F1,$EC,$E9,$B1,$C0,$7C,$2D,$69,$97
	.dc.b	$04,$7E,$C7,$C8,$3E,$9B,$F9,$B6,$DF,$89,$9F,$CA,$57,$D5,$AD,$78
	.dc.b	$3F,$5E,$72,$72,$BE,$3B,$FC,$0B,$E1,$18,$9F,$8B,$18,$AE,$51,$E6
	.dc.b	$C3,$F3,$18,$76,$8F,$C6,$1D,$A3,$37,$CE,$3F,$C6,$97,$CD,$D9,$CF
	.dc.b	$0F,$F1,$C5,$F3,$BF,$64,$E4,$FD,$84,$66,$7D,$11,$EB,$C3,$CB,$7D
	.dc.b	$6F,$8B,$6B,$71,$E4,$9F,$16,$BF,$D4,$B0,$F5,$3F,$D5,$30,$F5,$5F
	.dc.b	$93,$1F,$EC,$93,$B6,$DF,$A3,$82,$DF,$9A,$6A,$C3,$FD,$A1,$5F,$EA
	.dc.b	$97,$FA,$23,$FE,$B7,$B7,$D4,$3B,$B8,$D6,$AF,$FC,$7D,$5A,$9C,$F5
	.dc.b	$F8,$23,$18,$75,$8C,$61,$CE,$3C,$15,$D0,$F3,$FC,$6F,$FA,$93,$C4
	.dc.b	$C4,$E8,$FF,$65,$43,$EA,$1F,$11,$AC,$FF,$CA,$13,$98,$C6,$7F,$E0
	.dc.b	$98,$C5,$F3,$AD,$DF,$28,$FF,$65,$7C,$03,$A7,$E3,$B6,$3F,$DB,$43
	.dc.b	$E9,$98,$6A,$A4,$EA,$7E,$CB,$FE,$81,$AE,$58,$7F,$99,$63,$97,$07
	.dc.b	$08,$C5,$70,$7F,$9E,$57,$0C,$BF,$C4,$57,$C0,$FF,$70,$7D,$43,$DA
	.dc.b	$FB,$AD,$8F,$17,$1E,$26,$47,$95,$B8,$78,$1B,$7D,$B1,$E5,$3F,$E7
	.dc.b	$6C,$2C,$E6,$BF,$17,$7F,$EC,$83,$FD,$E4,$5A,$F5,$F7,$6B,$AA,$4F
	.dc.b	$CA,$FA,$8F,$F2,$93,$C2,$3F,$2E,$A3,$35,$FE,$86,$8E,$98,$7C,$29
	.dc.b	$CB,$FE,$38,$8C,$9F,$D8,$CE,$5E,$57,$E8,$3E,$6F,$F2,$17,$FE,$90
	.dc.b	$6D,$7B,$EF,$5C,$B3,$0F,$B5,$E2,$DC,$3E,$AB,$6F,$E3,$36,$7C,$1B
	.dc.b	$7F,$86,$5F,$EE,$93,$FE,$4C,$FA,$67,$9F,$FB,$26,$CE,$8F,$F6,$37
	.dc.b	$D4,$3E,$A3,$55,$27,$59,$7F,$7F,$B7,$B9,$98,$FF,$55,$C6,$27,$70
	.dc.b	$8C,$3F,$EB,$0F,$A8,$73,$78,$EF,$F8,$2C,$62,$F8,$DF,$CE,$7F,$DB
	.dc.b	$5D,$3F,$EC,$77,$3C,$5F,$C3,$23,$18,$FF,$AF,$1B,$AF,$F9,$B1,$FF
	.dc.b	$2F,$57,$91,$CB,$D3,$D0,$FE,$2B,$F5,$1F,$E6,$B8,$F4,$FC,$6B,$DB
	.dc.b	$CF,$A9,$FC,$6A,$31,$3E,$EF,$A4,$7B,$5F,$C7,$9B,$1F,$23,$EA,$1F
	.dc.b	$50,$FA,$8D,$5C,$AF,$9B,$E4,$BE,$A3,$FC,$ED,$1D,$67,$F9,$5C,$E7
	.dc.b	$D9,$E1,$C5,$FE,$DF,$8C,$56,$CE,$77,$CA,$7F,$DC,$9D,$FF,$E3,$7F
	.dc.b	$A8,$FE,$6A,$FF,$8E,$7B,$EC,$A6,$C6,$1B,$7F,$82,$AE,$4D,$AF,$CD
	.dc.b	$BF,$E4,$38,$FF,$22,$7D,$47,$43,$AF,$25,$9E,$11,$6B,$CA,$7D,$62
	.dc.b	$BF,$D2,$3E,$99,$EB,$D3,$F9,$BF,$D4,$7D,$2A,$FF,$7D,$CE,$AF,$F2
	.dc.b	$2D,$8F,$E6,$B3,$FC,$6E,$7F,$CB,$10,$9D,$74,$FF,$37,$AD,$4F,$F9
	.dc.b	$76,$73,$5F,$E5,$FE,$95,$6A,$B9,$61,$E4,$F2,$FE,$6F,$AB,$FE,$CE
	.dc.b	$C3,$E4,$3F,$8D,$AF,$A6,$E5,$BD,$3C,$27,$8B,$5E,$2F,$4F,$A8,$91
	.dc.b	$D3,$E2,$AD,$74,$9B,$AF,$FB,$67,$83,$FC,$6D,$58,$BF,$83,$B3,$E3
	.dc.b	$FE,$71,$F8,$EF,$82,$7C,$B6,$2B,$E5,$CF,$F6,$DF,$A8,$FA,$19,$7F
	.dc.b	$B6,$30,$E5,$5F,$DE,$67,$2F,$6A,$E1,$E0,$73,$D7,$FB,$CD,$69,$69
	.dc.b	$FE,$4B,$F8,$DF,$92,$7C,$AF,$A8,$B8,$FC,$A8,$B5,$F2,$A6,$E7,$EF
	.dc.b	$E1,$7C,$BF,$D2,$F6,$3E,$9B,$71,$FA,$6D,$C3,$E9,$B6,$FF,$DE,$7F
	.dc.b	$51,$FE,$F9,$AF,$E3,$6D,$AF,$D8,$70,$FF,$2A,$55,$C3,$0F,$F0,$FD
	.dc.b	$6E,$3F,$C0,$5F,$50,$F9,$6C,$D7,$F8,$42,$31,$8F,$68,$C4,$EA,$3F
	.dc.b	$C2,$33,$AC,$3F,$13,$F5,$1F,$8E,$FA,$8F,$C7,$B6,$BF,$1E,$D9,$DC
	.dc.b	$7E,$4A,$7F,$2B,$C5,$F9,$6E,$9B,$3F,$96,$6D,$7E,$59,$B9,$FC,$D7
	.dc.b	$FC,$77,$F4,$FF,$07,$E0,$FD,$95,$FE,$C3,$87,$C7,$B0,$FF,$39,$EB
	.dc.b	$FB,$09,$E0,$D9,$D5,$7F,$A0,$A3,$5E,$9E,$30,$EB,$2B,$EB,$87,$48
	.dc.b	$F0,$56,$EF,$A4,$E6,$78,$61,$97,$CC,$D7,$A5,$5B,$EB,$1E,$33,$75
	.dc.b	$CE,$1F,$51,$BB,$E0,$F1,$F3,$BA,$7D,$33,$BB,$FC,$6F,$8F,$33,$D5
	.dc.b	$3A,$8C,$E9,$EE,$C5,$39,$F0,$FC,$0D,$BE,$58,$75,$9F,$73,$63,$B5
	.dc.b	$72,$CF,$38,$F6,$D7,$F9,$0D,$EF,$2F,$9A,$D8,$FA,$D9,$77,$9C,$D7
	.dc.b	$1C,$7F,$E0,$05,$ED,$FC,$05,$6E,$FA,$D6,$E5,$5B,$7E,$0E,$1E,$4C
	.dc.b	$A4,$7B,$2B,$86,$5E,$98,$C4,$EB,$2F,$1B,$F1,$2F,$18,$F8,$22,$D7
	.dc.b	$92,$6B,$E2,$FA,$8F,$F0,$8B,$FE,$DE,$C9,$18,$8F,$F7,$1C,$7C,$B6
	.dc.b	$C6,$5F,$92,$F2,$C3,$E4,$B7,$0C,$FA,$DF,$91,$FA,$67,$D4,$7F,$B0
	.dc.b	$27,$38,$7D,$AB,$E0,$FC,$C6,$AA,$E2,$FD,$E8,$CE,$3E,$8C,$30,$77
	.dc.b	$C3,$B5,$73,$8C,$BD,$DD,$38,$C7,$D3,$17,$1F,$EC,$B8,$F8,$2B,$81
	.dc.b	$2B,$6F,$E6,$E3,$51,$82,$4F,$B9,$FD,$CC,$6A,$30,$49,$F3,$1C,$EF
	.dc.b	$D2,$FD,$7B,$F2,$3A,$BE,$73,$AC,$3E,$14,$F9,$6F,$A3,$F2,$67,$58
	.dc.b	$FC,$82,$61,$F0,$A7,$81,$D6,$B8,$E3,$E7,$8C,$10,$F2,$1E,$73,$CA
	.dc.b	$63,$1F,$82,$8F,$11,$AC,$7C,$F1,$A8,$C9,$E2,$39,$57,$8B,$1F,$4E
	.dc.b	$1C,$71,$FC,$2E,$3E,$5B,$D9,$E5,$C7,$AE,$1F,$6B,$0E,$33,$D6,$72
	.dc.b	$F9,$6B,$94,$62,$39,$92,$33,$7F,$22,$3C,$D1,$8C,$3C,$87,$CB,$C7
	.dc.b	$D9,$7C,$CE,$93,$BA,$F1,$5F,$AB,$0F,$2D,$F0,$8C,$4F,$43,$BC,$EB
	.dc.b	$1F,$BB,$87,$48,$C4,$FC,$1A,$E2,$F6,$C7,$EA,$E3,$E1,$C7,$B4,$79
	.dc.b	$F1,$F8,$D5,$DE,$7F,$1A,$E4,$F8,$D8,$7E,$52,$FE,$80,$FA,$8C,$7C
	.dc.b	$34,$AF,$E0,$0F,$E6,$BE,$A1,$F0,$B9,$9F,$D4,$4D,$C7,$FA,$ED,$6A
	.dc.b	$7F,$AC,$35,$D9,$D0,$CD,$75,$3F,$CB,$7C,$FF,$DA,$A1,$3C,$63,$0D
	.dc.b	$DF,$FC,$C2,$FF,$BE,$67,$FD,$6F,$F0,$1F,$E4,$36,$FD,$47,$4A,$FF
	.dc.b	$33,$CA,$31,$3F,$9D,$8E,$0F,$F8,$8A,$77,$E3,$EA,$13,$C5,$EA,$78
	.dc.b	$23,$50,$FA,$1B,$F4,$32,$BC,$B1,$A3,$0F,$73,$C5,$19,$7F,$DA,$3F
	.dc.b	$50,$FA,$8E,$BE,$84,$9C,$C7,$96,$75,$18,$21,$D4,$F4,$ED,$B9,$91
	.dc.b	$AD,$BE,$86,$5F,$36,$3F,$80,$7D,$47,$F7,$A7,$FD,$C6,$FF,$BA,$1B
	.dc.b	$F0,$7F,$16,$35,$1D,$DB,$5F,$6D,$E1,$19,$9E,$1D,$BC,$7E,$52,$4F
	.dc.b	$67,$A7,$C4,$E2,$3A,$45,$C3,$A4,$FD,$5F,$A8,$FF,$33,$D7,$9B,$D3
	.dc.b	$F2,$19,$F3,$64,$7F,$BF,$27,$6F,$C5,$6D,$F9,$FB,$77,$39,$AB,$39
	.dc.b	$8F,$E4,$8F,$E5,$FE,$9D,$F5,$1F,$87,$6C,$EB,$9F,$FD,$6F,$18,$8F
	.dc.b	$A8,$BE,$6F,$85,$49,$E0,$7F,$31,$9E,$17,$DE,$BB,$7C,$F7,$7F,$A5
	.dc.b	$97,$C0,$5A,$77,$19,$84,$F0,$8F,$1C,$FE,$7F,$C7,$FF,$2B,$E1,$FB
	.dc.b	$3F,$A7,$7D,$47,$E8,$DB,$7D,$EB,$7A,$7F,$B5,$1B,$1F,$D2,$5A,$FF
	.dc.b	$92,$FC,$5E,$5D,$DD,$D9,$1C,$DC,$BE,$BB,$FD,$63,$7B,$BC,$47,$8F
	.dc.b	$BA,$C6,$48,$7E,$86,$51,$E9,$EC,$F5,$7C,$9C,$AE,$D1,$97,$C1,$19
	.dc.b	$79,$47,$C5,$EA,$FE,$09,$EC,$3E,$A2,$99,$AF,$89,$35,$FD,$96,$BF
	.dc.b	$A2,$E0,$FE,$3D,$7F,$A6,$DF,$E4,$5E,$E7,$F9,$15,$6E,$7F,$DF,$F5
	.dc.b	$BB,$FE,$E9,$F5,$1F,$69,$FC,$03,$5F,$E0,$0D,$BF,$E5,$8B,$EA,$3F
	.dc.b	$8E,$2F,$4E,$92,$BF,$15,$F5,$0E,$EF,$E4,$4D,$8C,$C6,$0F,$E4,$DF
	.dc.b	$51,$FC,$AB,$1F,$E7,$54,$F8,$BE,$24,$BF,$CB,$7D,$47,$F0,$F4,$D2
	.dc.b	$70,$FF,$31,$FA,$87,$C9,$D3,$F9,$A4,$EA,$BF,$41,$3A,$7F,$45,$EC
	.dc.b	$7E,$B5,$CF,$B1,$CC,$FE,$79,$19,$8F,$E4,$CD,$CF,$FA,$93,$78,$BF
	.dc.b	$C4,$52,$8C,$3F,$E2,$5B,$FF,$38,$57,$0F,$03,$A9,$27,$CB,$E1,$FF
	.dc.b	$4B,$F7,$FC,$D1,$FC,$D7,$E8,$DE,$FF,$F4,$F6,$C7,$EE,$E7,$78,$7F
	.dc.b	$65,$6E,$7E,$46,$CF,$F8,$DD,$AA,$9F,$0F,$E1,$DE,$4F,$57,$C1,$1F
	.dc.b	$D7,$FC,$37,$C7,$FF,$61,$FA,$57,$C6,$53,$51,$30,$FE,$21,$5C,$1B
	.dc.b	$1C,$7E,$3B,$51,$FD,$4E,$74,$D9,$D3,$FC,$67,$E9,$DF,$45,$E1,$FA
	.dc.b	$89,$39,$8F,$EE,$93,$A8,$C1,$0F,$EC,$19,$7F,$08,$7F,$B1,$46,$DB
	.dc.b	$3C,$8F,$77,$8F,$FD,$FF,$EA,$3E,$EC,$AF,$FC,$02,$FF,$2D,$6B,$FD
	.dc.b	$B6,$36,$4E,$5E,$47,$C9,$FA,$0B,$3F,$CD,$67,$4F,$38,$F0,$7C,$37
	.dc.b	$CF,$F6,$A4,$EE,$32,$D2,$1E,$EE,$DF,$F0,$B7,$D4,$7D,$F8,$FE,$84
	.dc.b	$D7,$FC,$04,$DE,$8F,$F1,$77,$D4,$3D,$74,$7F,$9F,$FE,$A1,$F5,$1F
	.dc.b	$A0,$BE,$1D,$97,$AF,$B9,$FE,$25,$C7,$96,$1C,$6F,$F4,$F1,$8C,$98
	.dc.b	$F4,$7F,$D1,$98,$F9,$63,$10,$8C,$57,$2C,$BC,$B3,$CA,$31,$87,$78
	.dc.b	$E7,$8F,$ED,$B0,$E7,$87,$29,$FF,$4D,$63,$E3,$9E,$ED,$C3,$D9,$3C
	.dc.b	$CF,$15,$F3,$8E,$F1,$E1,$39,$E3,$E4,$BE,$F5,$B9,$F1,$9C,$EF,$B3
	.dc.b	$C3,$1E,$D8,$F9,$63,$B9,$BC,$3A,$63,$E0,$C7,$AC,$F0,$C7,$B2,$C7
	.dc.b	$AB,$6F,$8B,$E2,$3D,$07,$62,$C7,$03,$EB,$3F,$C7,$8E,$58,$D7,$D0
	.dc.b	$58,$DB,$F7,$BC,$67,$C7,$EE,$3C,$91,$87,$83,$5E,$D3,$FD,$36,$78
	.dc.b	$47,$FB,$3F,$E0,$AA,$69,$1D,$BF,$5F,$E6,$5F,$4C,$DF,$ED,$57,$33
	.dc.b	$E0,$CE,$A3,$31,$D4,$CC,$F1,$7E,$39,$C7,$5F,$CA,$D7,$18,$C9,$C6
	.dc.b	$33,$3C,$5F,$CD,$C7,$1D,$BF,$A4,$D7,$19,$CC,$F1,$9C,$D7,$17,$E8
	.dc.b	$61,$ED,$8C,$4F,$D5,$8C,$61,$EA,$6C,$79,$1B,$7D,$9B,$7D,$27,$DF
	.dc.b	$5C,$5A,$4D,$3F,$D6,$58,$FB,$DF,$F5,$A4,$7A,$C0,$8D,$4F,$A8,$FF
	.dc.b	$29,$3F,$EC,$86,$E3,$F3,$6B,$AF,$05,$9F,$8B,$D1,$70,$FE,$27,$F5
	.dc.b	$0E,$EC,$47,$79,$D6,$0A,$D5,$F8,$27,$31,$8A,$FF,$51,$65,$FD,$32
	.dc.b	$BF,$CD,$B7,$C7,$0E,$B3,$B8,$F8,$31,$E3,$F8,$C7,$C0,$7F,$76,$BF
	.dc.b	$6C,$F0,$35,$F5,$0F,$88,$C4,$FF,$9E,$AD,$39,$8E,$26,$5C,$65,$E6
	.dc.b	$BE,$77,$F0,$3C,$87,$4D,$3E,$36,$0F,$A6,$65,$98,$FF,$41,$57,$07
	.dc.b	$FD,$16,$6F,$87,$59,$7F,$B4,$63,$07,$0C,$3A,$E1,$F7,$27,$EF,$DF
	.dc.b	$37,$84,$65,$7D,$43,$D9,$FA,$ED,$C3,$C3,$5F,$83,$AF,$F6,$A4,$E5
	.dc.b	$B5,$E0,$38,$96,$BF,$D3,$93,$BD,$FF,$6D,$1F,$06,$76,$F6,$7F,$0E
	.dc.b	$FF,$96,$4F,$F5,$21,$6B,$D3,$DD,$AE,$A9,$3F,$1B,$EA,$3F,$C9,$EF
	.dc.b	$08,$F9,$F1,$8A,$FF,$76,$4F,$3C,$3C,$73,$97,$F8,$54,$FF,$13,$AE
	.dc.b	$3E,$57,$E6,$BE,$63,$F1,$D8,$7F,$93,$1B,$5F,$C6,$74,$63,$BE,$29
	.dc.b	$7F,$2B,$E2,$2D,$FD,$26,$E7,$F1,$30,$46,$2B,$F6,$E7,$E1,$DF,$D5
	.dc.b	$FD,$2F,$F9,$6E,$3F,$D9,$0D,$BF,$F0,$1C,$7F,$98,$3E,$A1,$F5,$1C
	.dc.b	$3D,$39,$3C,$3D,$EB,$7C,$0F,$E0,$B8,$23,$B9,$FB,$AF,$A8,$72,$E2
	.dc.b	$FF,$DB,$11,$87,$C7,$C3,$FC,$03,$A1,$E6,$7A,$23,$0C,$9F,$F7,$04
	.dc.b	$FF,$95,$9B,$1C,$7D,$1F,$E1,$3F,$51,$FE,$62,$9F,$F7,$37,$77,$FA
	.dc.b	$85,$FF,$74,$46,$75,$F5,$1F,$C0,$3D,$CF,$E2,$3F,$51,$FE,$9B,$AF
	.dc.b	$F7,$73,$67,$E0,$7D,$43,$EA,$1F,$51,$C3,$DE,$98,$7F,$BF,$3C,$57
	.dc.b	$D4,$7F,$9B,$A3,$9C,$FF,$21,$9C,$FB,$1C,$32,$F9,$91,$8A,$D9,$CE
	.dc.b	$F9,$4F,$D1,$EF,$FE,$6B,$F5,$1F,$CA,$DF,$E6,$FD,$F6,$53,$63,$0D
	.dc.b	$BF,$E7,$BD,$FF,$D8,$EF,$9C,$60,$FE,$82,$F4,$71,$68,$FE,$8B,$F5
	.dc.b	$1D,$CF,$07,$25,$AD,$CE,$9B,$4A,$DC,$61,$FE,$83,$F4,$CF,$5A,$9F
	.dc.b	$CD,$7E,$A3,$ED,$5F,$D8,$9D,$5F,$92,$11,$87,$F2,$DD,$2A,$FE,$39
	.dc.b	$9F,$EE,$2B,$4F,$F8,$0E,$75,$E4,$F8,$30,$FE,$79,$7C,$1F,$F6,$2D
	.dc.b	$F0,$7F,$B2,$D7,$07,$DF,$E9,$FF,$85,$5B,$EF,$7D,$58,$C6,$5F,$65
	.dc.b	$FC,$96,$5F,$45,$CB,$67,$85,$71,$79,$7D,$44,$8E,$FF,$15,$70,$E9
	.dc.b	$36,$FA,$4D,$AF,$4F,$C7,$FC,$E3,$F2,$5F,$04,$F9,$6C,$4F,$CB,$D5
	.dc.b	$5F,$DC,$BE,$A3,$E8,$E3,$F2,$63,$1D,$FC,$34,$FE,$09,$2A,$DE,$1E
	.dc.b	$03,$FB,$9B,$89,$FC,$56,$CF,$8D,$F8,$C7,$C7,$FA,$8B,$7F,$32,$2D
	.dc.b	$7C,$99,$B8,$7C,$9F,$97,$F9,$DF,$11,$F3,$DB,$87,$D4,$6D,$FD,$16
	.dc.b	$D7,$F3,$AF,$A8,$FF,$00,$4F,$F0,$46,$C7,$F9,$1F,$B9,$C3,$FC,$A6
	.dc.b	$38,$63,$FE,$18,$AD,$E3,$F8,$4F,$A8,$7C,$AE,$6B,$FC,$1D,$18,$9F
	.dc.b	$CC,$57,$0C,$BF,$BE,$CE,$5F,$15,$70,$BF,$C5,$7D,$47,$E2,$FE,$A2
	.dc.b	$E3,$F9,$06,$DF,$E3,$5B,$97,$E3,$5F,$C9,$71,$7E,$4F,$AE,$DF,$E6
	.dc.b	$1B,$7F,$95,$6D,$FE,$55,$FF,$1A,$FD,$3F,$F4,$9F,$83,$F6,$D7,$F9
	.dc.b	$1A,$1F,$21,$FE,$6D,$9C,$CA,$78,$69,$FE,$4C,$8C,$56,$E3,$13,$A9
	.dc.b	$E9,$39,$7B,$7A,$95,$E3,$97,$FA,$2E,$F8,$D6,$6F,$C7,$39,$AE,$2D
	.dc.b	$78,$9F,$E9,$1A,$E3,$19,$97,$D4,$73,$BE,$8F,$5F,$3F,$C5,$F5,$0E
	.dc.b	$EF,$F1,$AE,$5C,$CF,$89,$3A,$8C,$E9,$F1,$31,$FF,$27,$E5,$E3,$C3
	.dc.b	$97,$83,$E1,$C3,$83,$9C,$38,$E9,$F2,$6B,$79,$77,$3E,$75,$F0,$71
	.dc.b	$1A,$CB,$E9,$B6,$BE,$7D,$6E,$73,$7D,$AF,$FD,$03,$7B,$7D,$78,$F8
	.dc.b	$F1,$91,$DA,$B8,$3E,$29,$CD,$F4,$AB,$8F,$82,$6D,$76,$BE,$38,$77
	.dc.b	$8E,$47,$5B,$F8,$DF,$51,$FE,$0C,$8F,$F6,$C4,$A3,$11,$FE,$DD,$8F
	.dc.b	$8A,$D8,$CB,$F2,$5E,$58,$FC,$46,$E3,$9F,$57,$F0,$DF,$4E,$FA,$8F
	.dc.b	$CC,$5E,$E7,$C6,$4C,$3E,$95,$6D,$C4,$ED,$FA,$78,$76,$7F,$D5,$B3
	.dc.b	$B7,$35,$C3,$0F,$14,$74,$C3,$C3,$18,$C3,$A1,$1F,$D6,$5F,$38,$C5
	.dc.b	$7D,$AC,$3C,$0B,$1F,$91,$8F,$BB,$1E,$B3,$B8,$F1,$34,$F1,$18,$39
	.dc.b	$57,$4A,$E7,$1A,$3D,$64,$BF,$1A,$C3,$CD,$8F,$AC,$91,$DE,$7F,$84
	.dc.b	$CE,$A3,$13,$A7,$C3,$87,$A5,$F2,$5F,$8A,$FB,$DF,$19,$D9,$E3,$BE
	.dc.b	$51,$9A,$E1,$1D,$5B,$3F,$5B,$1F,$8A,$F2,$BF,$5C,$EA,$3D,$93,$A8
	.dc.b	$C1,$31,$FA,$38,$77,$7E,$41,$31,$F5,$5F,$94,$F7,$90,$EA,$7B,$A3
	.dc.b	$83,$F3,$F0,$F1,$C6,$1F,$93,$87,$8E,$31,$8F,$C0,$8C,$5F,$33,$C3
	.dc.b	$5C,$9F,$BD,$87,$9D,$D6,$1D,$EF,$E3,$61,$D2,$3E,$31,$AB,$FA,$38
	.dc.b	$76,$8F,$61,$F2,$6B,$F2,$F1,$99,$F0,$4F,$53,$E0,$63,$F2,$70,$F2
	.dc.b	$DF,$81,$F2,$DF,$18,$C5,$71,$3E,$39,$E0,$AF,$AD,$8F,$E0,$C9,$5F
	.dc.b	$71,$C7,$D4,$5F,$85,$97,$FC,$09,$FC,$0F,$D4,$3E,$2B,$33,$FC,$02
	.dc.b	$6E,$9F,$D8,$2B,$53,$FD,$3D,$AE,$CE,$86,$6B,$A9,$FD,$97,$9F,$FD
	.dc.b	$9B,$3C,$1B,$19,$E1,$FE,$D5,$3F,$EF,$3A,$FF,$31,$7C,$07,$F8,$E1
	.dc.b	$BF,$51,$B4,$7F,$B2,$1C,$3F,$EC,$AF,$1E,$CF,$18,$FF,$52,$A9,$C9
	.dc.b	$A9,$DF,$B7,$C3,$E7,$E5,$F8,$A3,$2F,$92,$32,$F4,$8F,$0F,$D4,$3E
	.dc.b	$A3,$C3,$E8,$CA,$FF,$3E,$4E,$5B,$3C,$08,$FC,$0D,$6B,$FE,$BC,$9E
	.dc.b	$0D,$9F,$6C,$FC,$0F,$A8,$75,$DF,$82,$97,$FE,$8B,$9C,$B6,$72,$FF
	.dc.b	$B2,$23,$FD,$C0,$E4,$E1,$F2,$7D,$48,$F3,$F8,$9A,$F1,$8B,$C6,$E3
	.dc.b	$FF,$12,$5A,$FE,$AF,$D4,$3C,$DF,$47,$A7,$32,$FF,$50,$4E,$5B,$3C
	.dc.b	$08,$FA,$FB,$AB,$FE,$E0,$9E,$0D,$9F,$89,$3F,$84,$FA,$77,$D4,$52
	.dc.b	$57,$FA,$BE,$1F,$35,$88,$FF,$75,$C7,$07,$53,$CE,$7A,$7C,$F7,$5F
	.dc.b	$A8,$95,$CE,$32,$F5,$8C,$BD,$23,$F2,$FE,$37,$F9,$33,$83,$E4,$4F
	.dc.b	$E7,$3E,$9D,$F5,$1F,$9C,$6C,$7F,$B1,$6B,$87,$17,$F9,$45,$B1,$FA
	.dc.b	$46,$BF,$E3,$BF,$1B,$97,$6F,$81,$91,$CD,$CB,$FE,$D1,$B7,$67,$84
	.dc.b	$91,$9E,$DA,$B3,$F4,$AA,$E2,$F3,$AF,$55,$7A,$FA,$BD,$9F,$27,$2B
	.dc.b	$C1,$19,$7C,$51,$97,$B6,$9F,$E6,$FC,$3F,$85,$3F,$07,$AF,$F8,$0F
	.dc.b	$B0,$FA,$8A,$7C,$E8,$97,$F2,$5A,$F4,$3F,$85,$CF,$E9,$9C,$1F,$E6
	.dc.b	$38,$DE,$9E,$ED,$FF,$4F,$3A,$AE,$13,$F0,$BE,$9D,$C7,$FC,$51,$B5
	.dc.b	$FE,$FD,$6B,$D9,$7D,$46,$FD,$39,$3B,$7F,$8E,$FD,$3F,$EA,$E3,$F9
	.dc.b	$0E,$1F,$42,$31,$1E,$86,$C6,$7B,$BF,$92,$FD,$47,$F2,$38,$E1,$F1
	.dc.b	$7D,$09,$7D,$38,$3F,$D3,$9F,$51,$F3,$A3,$F9,$7B,$53,$4C,$7D,$43
	.dc.b	$E4,$FF,$99,$B6,$7A,$1C,$B9,$BF,$9B,$D7,$0A,$FE,$73,$EC,$7E,$CA
	.dc.b	$B6,$2F,$FD,$66,$E1,$AF,$D6,$8E,$0D,$EE,$E6,$47,$87,$83,$5D,$1D
	.dc.b	$A7,$8E,$9F,$9A,$BF,$F3,$8B,$F9,$CF,$7F,$D5,$5F,$84,$FA,$37,$BD
	.dc.b	$FD,$41,$B1,$E8,$9D,$D7,$1C,$FF,$AA,$C6,$36,$FF,$64,$36,$BF,$AD
	.dc.b	$DB,$E1,$F7,$72,$7A,$3D,$E3,$FA,$FF,$86,$F8,$FF,$EC,$3F,$4C,$F8
	.dc.b	$DA,$6A,$26,$1F,$32,$B8,$36,$38,$FC,$7E,$A3,$F4,$33,$A6,$CE,$9F
	.dc.b	$AB,$F4,$EF,$A2,$F0,$FD,$44,$9C,$C7,$F7,$49,$D4,$60,$87,$EB,$B2
	.dc.b	$FE,$0E,$FE,$C2,$36,$D9,$E4,$7B,$BC,$7F,$EF,$FF,$51,$F6,$38,$7F
	.dc.b	$DE,$58,$3D,$8F,$86,$6B,$A2,$63,$F8,$CB,$E0,$FF,$82,$9A,$A9,$F4
	.dc.b	$1C,$23,$F8,$1C,$E8,$DC,$EA,$3A,$7C,$37,$D2,$F6,$64,$60,$87,$28
	.dc.b	$F4,$9F,$63,$B7,$FC,$33,$F5,$0E,$7A,$4A,$FD,$1F,$D4,$3D,$8F,$E6
	.dc.b	$10,$FA,$87,$D4,$7F,$89,$21,$87,$37,$F5,$6D,$7F,$8E,$B7,$B7,$F9
	.dc.b	$CE,$1D,$5F,$F1,$6D,$F4,$AF,$F1,$7C,$E5,$F0,$5F,$7B,$E9,$5B,$9F
	.dc.b	$0E,$3E,$38,$C5,$F1,$6E,$3E,$A7,$85,$F2,$C3,$A5,$F1,$7D,$97,$C5
	.dc.b	$B7,$C5,$DE,$1E,$86,$BE,$63,$D3,$1B,$6E,$3B,$7F,$C9,$78,$78,$5B
	.dc.b	$87,$85,$D9,$CC,$61,$E3,$76,$73,$3E,$09,$5D,$3C,$2F,$C4,$E8,$E9
	.dc.b	$5E,$17,$57,$C6,$77,$39,$35,$5D,$0F,$A9,$5F,$81,$38,$94,$A6,$5D
	.dc.b	$BF,$33,$C5,$7C,$8E,$E3,$DB,$18,$79,$4F,$F8,$EE,$74,$D8,$E0,$3E
	.dc.b	$06,$B4,$CB,$82,$3F,$2F,$E4,$1F,$4D,$FE,$22,$DB,$F1,$53,$FE,$7E
	.dc.b	$AF,$99,$5A,$E7,$F9,$D3,$93,$95,$F1,$DF,$E0,$5F,$08,$C4,$FC,$58
	.dc.b	$C5,$72,$8F,$36,$1F,$99,$C3,$B4,$7D,$F3,$B4,$66,$F9,$C7,$EB,$EF
	.dc.b	$9B,$B3,$9E,$1F,$B3,$BE,$77,$EC,$9C,$9F,$E9,$A8,$CC,$FA,$23,$D7
	.dc.b	$87,$96,$FA,$DF,$16,$D6,$E3,$C9,$3E,$2D,$BF,$C6,$78,$7A,$9F,$F1
	.dc.b	$AE,$1E,$AB,$F2,$63,$FE,$50,$9D,$B6,$FD,$1C,$16,$FC,$D3,$56,$1F
	.dc.b	$EC,$9A,$FE,$EE,$FF,$43,$7F,$BF,$DB,$EA,$1D,$FC,$6B,$57,$FE,$31
	.dc.b	$AD,$4E,$7A,$FC,$11,$8C,$3A,$C6,$30,$E7,$1E,$0A,$E8,$79,$FE,$2F
	.dc.b	$FD,$47,$E2,$E2,$74,$7F,$84,$A1,$F5,$0F,$87,$D6,$7F,$E4,$39,$CC
	.dc.b	$63,$3F,$F3,$AC,$62,$F9,$D6,$EF,$94,$7F,$86,$3C,$A7,$4F,$C7,$6C
	.dc.b	$7F,$88,$87,$D3,$30,$D5,$49,$D4,$FD,$97,$F9,$E5,$72,$C3,$FC,$B1
	.dc.b	$1C,$B8,$38,$46,$2B,$83,$FE,$E6,$AE,$19,$7F,$A3,$AF,$81,$FE,$2E
	.dc.b	$FA,$87,$B5,$F7,$5B,$1E,$2E,$3C,$4C,$8F,$2B,$70,$F0,$36,$FB,$63
	.dc.b	$CA,$7F,$CD,$78,$59,$CD,$7E,$2E,$FF,$D8,$C7,$F9,$00,$B5,$EB,$F0
	.dc.b	$35,$D5,$27,$E5,$7D,$47,$F9,$3D,$E1,$1F,$97,$51,$9A,$FF,$3D,$C7
	.dc.b	$4C,$3E,$14,$E5,$FF,$4C,$C6,$44,$FF,$B6,$E7,$51,$FD,$B7,$E7,$7F
	.dc.b	$23,$3F,$E8,$66,$D7,$C0,$AF,$D3,$46,$2B,$87,$27,$97,$4F,$F0,$B4
	.dc.b	$7F,$97,$A7,$58,$7F,$BA,$67,$2F,$5A,$E1,$E2,$74,$6D,$FF,$99,$BE
	.dc.b	$99,$E8,$7E,$D5,$B1,$FE,$6A,$FA,$87,$D4,$66,$A4,$F1,$F7,$6E,$5F
	.dc.b	$EA,$0C,$11,$CC,$FE,$01,$F5,$0E,$5C,$5F,$F9,$DA,$30,$F8,$FD,$2F
	.dc.b	$E0,$DD,$3F,$EC,$A8,$C1,$FE,$7C,$71,$8F,$FA,$B9,$BC,$5F,$EC,$A7
	.dc.b	$FC,$B5,$3B,$F3,$39,$FA,$3F,$C4,$7E,$A3,$FC,$CD,$3F,$E8,$6F,$8C
	.dc.b	$7B,$79,$F5,$7F,$8B,$46,$27,$FD,$17,$F5,$0F,$6B,$F8,$D3,$63,$FD
	.dc.b	$2B,$F5,$0F,$A8,$7D,$46,$AE,$57,$FA,$6B,$CA,$7D,$47,$F9,$C6,$3A
	.dc.b	$CF,$F2,$79,$CF,$B5,$C3,$8F,$FD,$9D,$18,$AD,$9C,$EF,$94,$FF,$B4
	.dc.b	$FB,$FF,$23,$F5,$1F,$CC,$5F,$C9,$F7,$59,$4D,$8C,$36,$FF,$D5,$F5
	.dc.b	$C9,$B5,$F8,$E8,$FC,$EF,$D4,$73,$3A,$72,$5A,$DC,$5A,$F2,$1F,$50
	.dc.b	$AF,$F4,$3F,$A6,$7B,$14,$FE,$65,$F5,$1F,$46,$BF,$DD,$73,$AB,$FF
	.dc.b	$5E,$B6,$3F,$98,$4F,$F1,$69,$FF,$2D,$4E,$7A,$BF,$98,$CE,$6B,$F6
	.dc.b	$B3,$9A,$FD,$BF,$52,$B5,$4A,$D3,$FC,$BE,$39,$67,$FE,$20,$BF,$9D
	.dc.b	$EB,$7E,$32,$BE,$92,$8C,$F4,$69,$A7,$CE,$8D,$36,$39,$FD,$44,$8D
	.dc.b	$7C,$55,$BE,$B3,$74,$ED,$0C,$7F,$C6,$75,$9B,$F8,$7F,$1F,$F3,$CF
	.dc.b	$C7,$7C,$1B,$E5,$71,$3F,$32,$7F,$B4,$FD,$47,$D1,$CB,$AE,$5F,$0E
	.dc.b	$73,$DF,$C2,$3F,$B9,$56,$EF,$B9,$FD,$E9,$C4,$79,$76,$7C,$6F,$C8
	.dc.b	$3E,$4F,$D4,$5B,$F8,$91,$6B,$E5,$CD,$CB,$DF,$B7,$F9,$43,$E5,$BE
	.dc.b	$8F,$B3,$F4,$9B,$8F,$CD,$6E,$1F,$51,$B7,$EC,$FA,$8F,$F7,$9D,$7F
	.dc.b	$38,$8C,$5F,$C6,$F3,$3F,$C3,$15,$AC,$15,$AB,$FF,$02,$7D,$43,$E5
	.dc.b	$77,$8F,$F8,$2E,$31,$3D,$4F,$F5,$76,$1C,$EF,$FC,$0B,$3A,$C3,$AD
	.dc.b	$6A,$F9,$61,$F8,$EF,$A8,$FC,$7F,$D4,$5A,$FC,$53,$6F,$F2,$2D,$C7
	.dc.b	$E3,$F0,$23,$F2,$FC,$5F,$98,$F7,$2E,$1F,$93,$6E,$1F,$9A,$6E,$1C
	.dc.b	$9F,$F1,$D7,$D3,$7F,$46,$9F,$F3,$C6,$7F,$EC,$98,$7D,$47,$E5,$3E
	.dc.b	$17,$E2,$57,$F6,$49,$E1,$18,$84,$70,$C7,$FC,$F1,$7C,$F8,$79,$3D
	.dc.b	$70,$B7,$CE,$6B,$CA,$2D,$F3,$8F,$29,$E1,$8C,$CE,$99,$7E,$09,$F3
	.dc.b	$7D,$46,$1E,$27,$99,$EB,$E7,$78,$3E,$9D,$DB,$FE,$33,$AD,$E4,$AD
	.dc.b	$4E,$74,$F8,$13,$9C,$7C,$16,$C3,$9F,$85,$8A,$D4,$F8,$A7,$2D,$C3
	.dc.b	$A6,$BF,$EF,$99,$CE,$3D,$B2,$F0,$E3,$DA,$73,$5B,$BE,$38,$7D,$5B
	.dc.b	$DB,$D6,$F8,$4E,$A3,$EB,$CE,$B0,$FB,$58,$74,$9F,$F1,$0B,$5D,$E5
	.dc.b	$CB,$09,$1D,$4F,$8F,$7C,$F1,$F5,$7D,$47,$F8,$1E,$7F,$DA,$58,$23
	.dc.b	$11,$FE,$D7,$8F,$86,$D8,$CB,$CB,$D5,$F8,$11,$CF,$2F,$3B,$6F,$FC
	.dc.b	$6F,$4F,$A7,$7D,$47,$FA,$D6,$79,$4F,$D1,$AE,$0F,$B9,$AF,$19,$E0
	.dc.b	$E7,$0F,$09,$1F,$C5,$DF,$67,$18,$79,$26,$E1,$E4,$AD,$57,$86,$72
	.dc.b	$F5,$39,$DF,$81,$FC,$56,$1E,$4C,$7E,$45,$70,$9F,$31,$63,$2F,$53
	.dc.b	$99,$99,$E4,$7B,$AF,$C7,$8F,$9E,$78,$34,$FD,$E4,$6B,$1F,$93,$19
	.dc.b	$69,$ED,$C7,$D1,$87,$8A,$75,$7E,$D7,$9C,$EB,$0F,$1E,$3E,$BC,$7B
	.dc.b	$5E,$F1,$EA,$D3,$FC,$E7,$1A,$7E,$2C,$FB,$4F,$D8,$C6,$71,$F1,$57
	.dc.b	$2A,$D9,$F1,$0F,$7B,$98,$DC,$76,$C3,$CD,$7D,$63,$86,$3E,$C7,$9B
	.dc.b	$70,$F0,$63,$EC,$9E,$A7,$33,$94,$EE,$FA,$C6,$8E,$78,$FC,$B7,$6F
	.dc.b	$A6,$FA,$E1,$89,$E8,$71,$C7,$D2,$F2,$33,$87,$9E,$32,$6A,$B9,$CF
	.dc.b	$1C,$7F,$09,$8F,$92,$F8,$46,$27,$C0,$75,$3F,$22,$71,$C3,$ED,$57
	.dc.b	$1C,$7E,$65,$71,$AD,$1E,$FC,$3C,$38,$76,$C3,$EC,$E1,$FA,$19,$CE
	.dc.b	$1F,$06,$30,$4B,$D9,$F8,$0F,$A8,$E1,$F0,$F2,$3F,$06,$FE,$AF,$EA
	.dc.b	$1F,$0D,$99,$FD,$CC,$DC,$FF,$AD,$56,$A7,$F9,$DB,$5D,$9D,$0C,$D7
	.dc.b	$53,$F8,$8F,$3F,$F8,$BA,$78,$36,$33,$C3,$FC,$5A,$7F,$DD,$F5,$FE
	.dc.b	$0B,$3F,$D9,$FF,$00,$F1,$77,$18,$FA,$8A,$A5,$18,$78,$47,$33,$53
	.dc.b	$99,$91,$99,$DB,$FE,$16,$3C,$3E,$DF,$3F,$A1,$91,$B8,$EC,$61,$E0
	.dc.b	$78,$E3,$2F,$87,$EA,$1F,$51,$E5,$F4,$64,$E6,$3D,$73,$A8,$C1,$0E
	.dc.b	$B8,$7C,$AD,$3F,$9E,$CE,$E7,$C2,$DD,$BF,$CB,$AD,$9F,$E8,$75,$C6
	.dc.b	$BF,$DF,$7F,$51,$FD,$ED,$F8,$CF,$FB,$79,$BF,$05,$F3,$A3,$51,$EE
	.dc.b	$6E,$1F,$E6,$A7,$84,$66,$78,$75,$71,$F9,$49,$3D,$9E,$9F,$13,$88
	.dc.b	$E9,$17,$0E,$93,$F7,$BE,$A1,$E6,$78,$FD,$19,$87,$DD,$9C,$B6,$78
	.dc.b	$11,$F6,$F6,$D7,$EB,$CF,$06,$CF,$BE,$7F,$1B,$F5,$0F,$A8,$A4,$AF
	.dc.b	$C6,$43,$E6,$F1,$1F,$82,$8E,$0E,$A7,$9C,$F4,$F9,$EE,$BF,$4B,$2B
	.dc.b	$9C,$65,$EB,$19,$7A,$47,$E8,$3C,$6F,$F2,$76,$DF,$E9,$AF,$A7,$7D
	.dc.b	$47,$E9,$1B,$7F,$A1,$AD,$F3,$7F,$81,$1B,$03,$FC,$7F,$E2,$75,$39
	.dc.b	$77,$58,$E5,$15,$FC,$FD,$FE,$A3,$1F,$17,$6F,$84,$93,$FB,$19,$76
	.dc.b	$E2,$38,$CD,$AF,$54,$FA,$7C,$0F,$57,$C9,$CB,$E5,$19,$7A,$46,$5F
	.dc.b	$1C,$7C,$5E,$AF,$E0,$BE,$C3,$EA,$29,$9A,$F8,$93,$5F,$DC,$ED,$F9
	.dc.b	$AC,$BF,$D8,$B7,$F9,$ED,$FE,$45,$EE,$7F,$D9,$35,$B9,$FF,$17,$D6
	.dc.b	$EF,$E6,$7D,$43,$8F,$F8,$BB,$63,$F8,$A3,$5F,$E1,$2B,$EA,$37,$E9
	.dc.b	$C9,$DB,$FC,$83,$EA,$1D,$D4,$93,$FC,$7E,$32,$FF,$21,$FA,$8F,$AD
	.dc.b	$9F,$F9,$12,$5F,$17,$CC,$95,$FC,$9F,$EA,$3F,$83,$26,$93,$8B,$F9
	.dc.b	$8F,$D4,$3E,$4E,$9F,$CD,$27,$55,$FC,$C6,$74,$FF,$33,$F6,$3F,$50
	.dc.b	$E7,$D8,$E6,$7F,$3C,$8C,$C7,$F2,$C6,$F2,$7F,$BE,$1B,$A7,$F9,$79
	.dc.b	$B3,$F9,$BB,$FF,$35,$D6,$FB,$FA,$12,$7C,$9E,$17,$F4,$AF,$83,$F2
	.dc.b	$C7,$F3,$AF,$A4,$7B,$FF,$D3,$9B,$1F,$D2,$E7,$78,$33,$FF,$37,$C6
	.dc.b	$1B,$B7,$53,$97,$C3,$5B,$E5,$15,$F0,$47,$F5,$CF,$11,$F1,$FF,$D7
	.dc.b	$BE,$95,$F1,$94,$D4,$4B,$FE,$BD,$5C,$1B,$1C,$7E,$3B,$51,$FD,$5A
	.dc.b	$74,$D9,$D3,$FD,$9F,$E9,$DF,$45,$E1,$FA,$89,$39,$8F,$EE,$33,$A8
	.dc.b	$C1,$0F,$EC,$79,$3C,$7D,$D4,$9D,$70,$7C,$97,$F8,$95,$6D,$FE,$F1
	.dc.b	$F5,$1F,$71,$47,$F7,$F7,$FB,$E3,$5F,$EE,$31,$B2,$72,$FF,$81,$9E
	.dc.b	$52,$FA,$0C,$4C,$AE,$EF,$CF,$8E,$1F,$0D,$F4,$7D,$A9,$5E,$48,$CB
	.dc.b	$48,$7C,$0E,$DF,$F0,$AF,$D4,$3A,$29,$27,$FC,$29,$F5,$0F,$5D,$B7
	.dc.b	$F3,$3F,$A8,$7D,$47,$FA,$EB,$1E,$58,$7F,$86,$DA,$FF,$84,$1B,$5F
	.dc.b	$2E,$18,$75,$8E,$95,$FE,$C4,$8C,$4A,$FB,$BF,$D0,$EF,$AC,$62,$11
	.dc.b	$8A,$E5,$87,$A2,$79,$46,$30,$EF,$1C,$F1,$FF,$69,$E1,$CF,$0E,$53
	.dc.b	$FD,$47,$1F,$1C,$F7,$6E,$1F,$06,$F8,$B6,$BA,$9E,$EC,$7C,$57,$C5
	.dc.b	$B3,$B9,$F2,$9E,$33,$8D,$F6,$7C,$53,$C1,$D1,$C0,$F8,$37,$E2,$77
	.dc.b	$8F,$29,$E1,$1A,$3C,$B8,$FB,$9F,$34,$EC,$CD,$73,$8E,$07,$7B,$F2
	.dc.b	$3C,$0F,$F7,$D9,$E0,$29,$93,$2F,$DC,$6F,$8C,$F8,$CD,$C7,$E1,$23
	.dc.b	$0F,$29,$FE,$97,$3A,$6C,$70,$1F,$05,$5A,$65,$C1,$1F,$91,$F3,$2F
	.dc.b	$A6,$FD,$F6,$DF,$8B,$9F,$F0,$85,$7D,$AA,$D7,$81,$F7,$27,$27,$2B
	.dc.b	$E3,$BF,$C0,$BE,$11,$89,$F8,$B1,$8A,$E5,$1E,$6C,$3F,$35,$87,$68
	.dc.b	$FB,$87,$68,$CD,$F3,$8F,$F1,$21,$C7,$0B,$5C,$6B,$D5,$39,$3F,$C5
	.dc.b	$51,$99,$F3,$47,$A6,$FC,$78,$79,$63,$C2,$71,$15,$C5,$F5,$9C,$4D
	.dc.b	$C7,$1B,$F3,$D7,$FA,$76,$BD,$9A,$FF,$01,$BF,$66,$1E,$8C,$7F,$97
	.dc.b	$4F,$F0,$5A,$DD,$FB,$38,$32,$EC,$E8,$E3,$0F,$6F,$3D,$B7,$D4,$3B
	.dc.b	$B5,$5E,$AA,$D4,$E7,$0F,$34,$62,$B7,$8F,$F4,$E9,$FF,$60,$61,$D2
	.dc.b	$72,$48,$FF,$62,$7C,$63,$E0,$3F,$BA,$C7,$C1,$9D,$65,$FE,$B9,$A7
	.dc.b	$D4,$3E,$23,$38,$FF,$4F,$9C,$C6,$30,$FF,$31,$46,$1E,$87,$82,$B9
	.dc.b	$3F,$EC,$BF,$21,$D3,$F0,$1B,$1F,$ED,$91,$F4,$CB,$D5,$49,$D4,$FD
	.dc.b	$2C,$BF,$CE,$57,$C9,$FF,$69,$4F,$2E,$1E,$11,$8A,$E0,$AB,$86,$7F
	.dc.b	$DA,$EF,$81,$FE,$E1,$FA,$87,$B5,$F5,$9B,$1E,$2E,$3C,$4C,$8F,$2B
	.dc.b	$70,$F0,$36,$FB,$63,$CA,$7F,$DC,$38,$59,$CD,$7E,$12,$FF,$CA,$E7
	.dc.b	$FB,$CC,$B5,$EB,$EE,$D7,$54,$9F,$95,$F5,$1F,$E4,$D7,$84,$7E,$4D
	.dc.b	$46,$6B,$FD,$E3,$1D,$30,$F8,$53,$97,$FB,$24,$64,$FF,$50,$1C,$BC
	.dc.b	$AF,$CC,$FC,$BF,$E5,$EF,$FD,$F6,$DA,$F7,$DE,$B9,$26,$1F,$6B,$E2
	.dc.b	$2E,$1F,$55,$B7,$F1,$9B,$3E,$0D,$91,$FA,$8F,$A6,$7A,$34,$FD,$5F
	.dc.b	$D4,$3E,$A3,$18,$7F,$AD,$F1,$FE,$FD,$7C,$23,$07,$F1,$85,$EF,$F0
	.dc.b	$CF,$FC,$1B,$39,$8F,$01,$FB,$8F,$A8,$FC,$54,$72,$9F,$F6,$25,$72
	.dc.b	$8C,$42,$76,$F4,$37,$C1,$FC,$0B,$A7,$FC,$81,$5C,$2F,$E9,$3F,$CB
	.dc.b	$23,$35,$C4,$FD,$37,$99,$C7,$D1,$E4,$DA,$F0,$FD,$43,$D3,$FE,$1F
	.dc.b	$E0,$3D,$CF,$E2,$33,$9A,$F3,$CE,$5F,$4F,$D4,$7F,$A8,$1B,$ED,$FF
	.dc.b	$1D,$FA,$87,$D4,$3E,$A3,$F8,$DD,$EF,$6F,$33,$79,$3F,$CB,$1E,$3B
	.dc.b	$EA,$3F,$CE,$F9,$F4,$3E,$54,$67,$D9,$E7,$A7,$F2,$B8,$C4,$EF,$1F
	.dc.b	$95,$18,$7E,$67,$7F,$FB,$D7,$EA,$3F,$99,$3F,$EF,$7E,$FB,$29,$B1
	.dc.b	$86,$DF,$DB,$AE,$4D,$AF,$CE,$3F,$F0,$06,$10,$FA,$8E,$A7,$6E,$6B
	.dc.b	$5C,$62,$CF,$98,$FF,$4A,$C6,$4F,$C3,$FD,$43,$D7,$FE,$8A,$D8,$FE
	.dc.b	$77,$F5,$1F,$56,$BF,$05,$39,$6D,$7F,$51,$EA,$57,$F1,$EB,$FE,$EB
	.dc.b	$7B,$E1,$FE,$79,$39,$C3,$D7,$39,$AF,$E7,$B3,$99,$4E,$5F,$E6,$D1
	.dc.b	$C7,$3F,$F1,$45,$7D,$0F,$57,$F1,$95,$F5,$1F,$F6,$04,$7C,$D7,$38
	.dc.b	$7F,$A6,$EB,$8C,$E7,$AB,$93,$CD,$E9,$F5,$12,$39,$7C,$5D,$C7,$B4
	.dc.b	$DC,$3A,$62,$AC,$5F,$C2,$F9,$0F,$9A,$7E,$37,$E0,$9F,$2D,$89,$FA
	.dc.b	$73,$FD,$E7,$EA,$3E,$86,$3F,$E8,$7D,$D3,$9F,$03,$86,$BF,$BD,$A5
	.dc.b	$F0,$C3,$C0,$7F,$78,$71,$2D,$FF,$CA,$1F,$1D,$F2,$4F,$95,$F5,$16
	.dc.b	$FE,$44,$5A,$F9,$93,$70,$F7,$FC,$BF,$D1,$F8,$8F,$A4,$DC,$3E,$83
	.dc.b	$6F,$EA,$36,$BF,$D4,$5F,$51,$FE,$06,$9F,$E1,$CD,$8F,$DC,$F0,$7F
	.dc.b	$29,$9C,$E1,$FE,$1A,$AD,$61,$FE,$04,$FA,$87,$CB,$66,$BF,$29,$18
	.dc.b	$9F,$E4,$B5,$C2,$FF,$C0,$13,$97,$AC,$FF,$86,$A3,$18,$FD,$CF,$A8
	.dc.b	$FC,$77,$D4,$7E,$3D,$B5,$F8,$D6,$DF,$E4,$5B,$5F,$84,$E1,$FC,$B7
	.dc.b	$B9,$F9,$76,$DF,$E5,$5B,$7F,$99,$6D,$7F,$B1,$FE,$9F,$E1,$7C,$27
	.dc.b	$BA,$BF,$81,$C3,$E4,$3F,$DE,$D3,$99,$4F,$0D,$FF,$CC,$EE,$F1,$FF
	.dc.b	$3E,$C6,$3D,$5E,$2E,$B0,$B7,$DA,$6B,$D2,$2D,$F3,$8F,$F4,$8D,$71
	.dc.b	$BE,$B1,$C5,$AF,$16,$CF,$1C,$3D,$0B,$EA,$3A,$E1,$D9,$EF,$E7,$78
	.dc.b	$BE,$A3,$F5,$DE,$17,$28,$C7,$77,$F8,$EF,$3F,$19,$F1,$A3,$33,$BC
	.dc.b	$FC,$77,$CA,$B6,$1C,$67,$F2,$DB,$A7,$CA,$C3,$9D,$70,$9C,$BF,$CD
	.dc.b	$9A,$FD,$28,$DD,$73,$C7,$CC,$DC,$7C,$95,$CE,$73,$87,$FA,$7B,$0E
	.dc.b	$67,$4D,$3F,$CD,$D9,$7A,$AB,$77,$E1,$FA,$8F,$F2,$83,$FE,$D2,$B4
	.dc.b	$62,$3F,$DA,$F1,$F0,$5B,$19,$7C,$0F,$2C,$BE,$13,$71,$CF,$C0,$FD
	.dc.b	$CF,$A8,$7D,$47,$FA,$D2,$F7,$3F,$3A,$F8,$61,$E5,$8E,$15,$BC,$39
	.dc.b	$CF,$CB,$AD,$BF,$4D,$B8,$76,$8E,$97,$1E,$0F,$47,$18,$78,$9C,$9F
	.dc.b	$7A,$79,$38,$8D,$12,$FF,$D9,$B3,$A8,$C4,$FA,$4E,$13,$AC,$3D,$51
	.dc.b	$99,$DB,$E9,$BF,$23,$C0,$D5,$7C,$09,$FD,$C5,$F7,$8C,$12,$76,$7F
	.dc.b	$80,$A3,$4E,$0E,$C7,$50,$E9,$CC,$F9,$31,$FC,$1D,$F9,$23,$18,$FD
	.dc.b	$09,$CD,$7C,$7C,$7D,$AF,$2C,$7C,$8D,$3C,$91,$A9,$E3,$7D,$30,$ED
	.dc.b	$7C,$27,$2F,$2C,$7C,$D1,$8C,$3D,$8D,$7D,$78,$75,$C7,$CC,$4C,$3A
	.dc.b	$CF,$08,$C3,$E9,$3F,$07,$87,$86,$3F,$17,$5C,$EB,$35,$F2,$F0,$F1
	.dc.b	$D7,$2C,$7E,$DC,$62,$B7,$5F,$04,$F2,$98,$C7,$D0,$7A,$67,$84,$7D
	.dc.b	$58,$C6,$1E,$18,$C6,$1E,$47,$93,$E1,$C7,$F1,$D1,$8C,$3C,$38,$74
	.dc.b	$C7,$E7,$4F,$BC,$E7,$3A,$C3,$B4,$E8,$FC,$7B,$8B,$F8,$93,$F2,$AB
	.dc.b	$D9,$8F,$AB,$1F,$24,$75,$C3,$95,$F9,$23,$0D,$8E,$B1,$FC,$17,$EA
	.dc.b	$3F,$13,$97,$FB,$FE,$3F,$84,$7C,$3E,$0F,$C6,$FD,$43,$E1,$B1,$3A
	.dc.b	$21,$FC,$DA,$34,$FF,$4C,$6B,$FB,$D9,$E2,$EE,$7A,$9F,$E6,$A7,$FD
	.dc.b	$39,$3C,$39,$9C,$19,$8C,$37,$E1,$78,$4E,$AB,$FC,$E7,$F0,$3F,$E3
	.dc.b	$47,$FD,$62,$BE,$A3,$77,$DE,$31,$3A,$F1,$F2,$F2,$6B,$FE,$D0,$AB
	.dc.b	$3C,$23,$FD,$64,$F7,$6F,$B9,$AF,$A0,$91,$E2,$8E,$26,$1E,$31,$87
	.dc.b	$FD,$8F,$F4,$EF,$A8,$EF,$E9,$49,$CC,$79,$27,$51,$82,$1E,$4D,$F5
	.dc.b	$32,$33,$AF,$C7,$7F,$DD,$73,$EE,$FA,$87,$6D,$F8,$49,$5F,$ED,$A9
	.dc.b	$CB,$67,$2F,$FB,$4E,$3F,$DD,$2E,$4E,$1F,$2B,$D4,$8F,$3F,$8A,$AF
	.dc.b	$18,$B8,$F9,$69,$5F,$4B,$EA,$1E,$6F,$9F,$D3,$98,$FF,$BB,$A7,$2D
	.dc.b	$9E,$04,$7D,$5D,$F5,$FF,$75,$4F,$06,$CF,$C3,$9F,$C0,$FD,$43,$EA
	.dc.b	$29,$2B,$FD,$FB,$0F,$99,$C4,$7F,$BD,$63,$83,$A9,$E7,$3D,$3E,$7B
	.dc.b	$AF,$D2,$CA,$E7,$19,$7A,$C6,$5E,$91,$F9,$4F,$1B,$FC,$99,$BF,$DB
	.dc.b	$FA,$77,$D4,$7E,$61,$B7,$FE,$D6,$9F,$C0,$57,$0E,$EF,$D8,$36,$7F
	.dc.b	$DB,$6D,$7F,$D5,$1E,$2F,$3E,$FF,$03,$23,$A5,$72,$8D,$46,$0F,$F7
	.dc.b	$2E,$0E,$DF,$21,$23,$42,$1D,$B9,$86,$5E,$DA,$C3,$EF,$9F,$77,$57
	.dc.b	$BF,$C0,$97,$D3,$0F,$E0,$EF,$8A,$72,$F4,$CB,$F8,$84,$AB,$9C,$F9
	.dc.b	$34,$FC,$DB,$B3,$FD,$FF,$EB,$BE,$A3,$F8,$2B,$63,$87,$6F,$D5,$23
	.dc.b	$83,$DF,$1F,$AA,$7A,$1F,$ED,$58,$DB,$8A,$F0,$E9,$FB,$7A,$E0,$FE
	.dc.b	$E7,$EA,$1C,$7F,$C5,$9B,$7F,$31,$AF,$65,$F5,$1B,$F4,$A4,$ED,$FE
	.dc.b	$3F,$F4,$FF,$B0,$8F,$AA,$D7,$CE,$DC,$3F,$85,$37,$7F,$E7,$5F,$51
	.dc.b	$FC,$8E,$73,$F1,$3D,$09,$7F,$80,$FA,$8F,$AA,$9A,$4E,$3F,$E6,$3F
	.dc.b	$50,$F9,$4A,$7F,$34,$9D,$57,$E2,$A7,$4F,$E3,$3D,$8F,$DA,$39,$F6
	.dc.b	$39,$9F,$CF,$23,$31,$F4,$9B,$97,$F2,$56,$E9,$FD,$55,$B2,$17,$FE
	.dc.b	$79,$AE,$1D,$DD,$49,$3E,$5F,$07,$FA,$5F,$BF,$E6,$8F,$C0,$FD,$1B
	.dc.b	$E0,$3F,$A7,$B6,$3F,$4B,$3B,$C3,$F9,$C3,$73,$FE,$D0,$D9,$F2,$35
	.dc.b	$53,$E1,$FC,$3B,$C9,$EA,$F8,$23,$FA,$FF,$86,$F8,$EF,$EC,$3F,$4A
	.dc.b	$F8,$CA,$6A,$26,$1F,$BA,$AE,$0D,$8E,$3F,$1D,$A8,$FC,$FC,$E9,$B3
	.dc.b	$A7,$F8,$1F,$D3,$BE,$87,$C3,$F5,$12,$73,$1F,$DD,$27,$51,$82,$1F
	.dc.b	$AC,$CB,$F8,$83,$8E,$0F,$EC,$13,$C2,$3F,$62,$D8,$FE,$C9,$3C,$27
	.dc.b	$F6,$BF,$51,$F8,$05,$0F,$63,$E2,$9F,$E0,$69,$D1,$39,$70,$7C,$AF
	.dc.b	$A0,$B7,$FC,$82,$74,$FF,$22,$B4,$F3,$F8,$6F,$A3,$ED,$4A,$E1,$19
	.dc.b	$69,$0F,$81,$DB,$FE,$18,$FA,$8F,$C0,$C7,$F2,$E6,$BF,$C4,$1B,$A7
	.dc.b	$F8,$0F,$EA,$1E,$CF,$F3,$0A,$7D,$43,$EA,$3F,$C6,$D8,$70,$EC,$BD
	.dc.b	$7D,$CF,$F1,$3E,$3C,$B2,$FD,$36,$1C,$71,$FD,$44,$E5,$EB,$8F,$27
	.dc.b	$6D,$72,$70,$CB,$C9,$5C,$E1,$5C,$E3,$19,$7B,$9B,$8F,$F9,$35,$E1
	.dc.b	$3B,$C3,$FC,$A0,$DB,$FD,$E5,$F1,$9F,$15,$70,$C3,$C3,$1C,$1B,$5C
	.dc.b	$30,$ED,$3C,$0D,$63,$EC,$9F,$53,$56,$1D,$DE,$46,$71,$F2,$3E,$EC
	.dc.b	$3B,$BC,$30,$EE,$F0,$3E,$39,$BB,$E5,$3C,$2F,$B4,$F1,$33,$31,$FF
	.dc.b	$39,$1E,$62,$9C,$CB,$7F,$65,$DB,$FE,$36,$F9,$3D,$C7,$D4,$F6,$D1
	.dc.b	$89,$D4,$F2,$9C,$BC,$07,$C2,$D6,$99,$70,$47,$DF,$F3,$0F,$A6,$FF
	.dc.b	$41,$6D,$F8,$59,$FE,$6D,$5F,$3E,$B5,$BF,$D1,$9C,$9C,$AF,$8E,$FF
	.dc.b	$02,$F8,$46,$27,$E2,$C6,$2B,$94,$79,$B0,$FC,$A6,$1D,$A3,$E9,$9D
	.dc.b	$A3,$37,$CE,$3F,$D4,$17,$CD,$D9,$CF,$0F,$F5,$25,$F3,$BF,$64,$E4
	.dc.b	$FE,$A5,$19,$9F,$44,$7A,$F0,$F2,$DF,$5B,$E2,$DA,$DC,$79,$27,$C5
	.dc.b	$A7,$FA,$17,$0F,$53,$FE,$87,$C3,$D5,$7E,$4C,$7F,$D6,$13,$B6,$DF
	.dc.b	$A3,$82,$DF,$9A,$6A,$C3,$FD,$99,$5F,$E4,$27,$FA,$03,$FE,$49,$B7
	.dc.b	$D4,$3B,$B8,$D6,$AF,$FC,$69,$5A,$9C,$F5,$F8,$23,$18,$75,$8C,$61
	.dc.b	$CE,$3C,$15,$D0,$F3,$FC,$6F,$FA,$8B,$C4,$C4,$E8,$FF,$2E,$43,$EA
	.dc.b	$1F,$11,$AC,$FF,$C8,$D3,$98,$C6,$7F,$ED,$58,$C5,$F3,$AD,$DF,$28
	.dc.b	$FF,$31,$F8,$EE,$9F,$8E,$D8,$FF,$36,$0F,$A6,$61,$AA,$93,$A9,$FB
	.dc.b	$2F,$FB,$B6,$B9,$61,$FE,$5B,$8E,$5C,$1C,$23,$15,$C1,$FF,$0F,$57
	.dc.b	$0C,$BF,$DD,$D7,$C0,$FF,$3C,$FD,$47,$F9,$A5,$BE,$CF,$E0,$1F,$37
	.dc.b	$1F,$E0,$27,$58,$7F,$98,$1B,$7D,$31,$B7,$FE,$2D,$A5,$61,$47,$C1
	.dc.b	$C3,$A1,$DE,$B5,$3C,$23,$07,$E4,$27,$FD,$96,$7F,$A4,$0B,$1F,$03
	.dc.b	$BB,$5D,$12,$7E,$7F,$D4,$7F,$94,$5E,$11,$ED,$8C,$57,$F9,$EA,$79
	.dc.b	$E5,$F1,$A7,$2F,$FB,$BE,$32,$7F,$92,$44,$69,$F8,$1F,$31,$F9,$5A
	.dc.b	$FF,$42,$B6,$BE,$35,$7E,$AA,$31,$5C,$39,$3E,$17,$89,$96,$C7,$FA
	.dc.b	$BA,$77,$87,$CA,$D3,$FC,$35,$39,$AD,$CF,$FA,$53,$E9,$5E,$7F,$EE
	.dc.b	$1B,$5F,$EB,$8F,$A8,$7D,$46,$6A,$4F,$FA,$F2,$9E,$EE,$30,$E8,$F3
	.dc.b	$3F,$81,$FD,$43,$97,$17,$FE,$C4,$8C,$3C,$FD,$1F,$E1,$3D,$3F,$EC
	.dc.b	$B8,$C1,$FE,$CA,$71,$87,$FA,$B1,$BC,$5F,$88,$8F,$F3,$C4,$7F,$96
	.dc.b	$E7,$87,$C1,$74,$F4,$7F,$8A,$FD,$47,$F9,$9E,$7F,$DA,$BF,$1A,$F7
	.dc.b	$33,$EA,$FF,$1A,$8C,$4F,$FB,$6B,$E9,$5E,$DF,$F1,$E6,$C7,$FB,$8F
	.dc.b	$EA,$1F,$50,$FA,$8D,$5C,$AF,$F7,$47,$94,$FA,$8F,$F3,$94,$75,$9F
	.dc.b	$E5,$73,$9F,$6B,$87,$17,$FB,$56,$31,$5B,$39,$DF,$29,$FF,$6D,$77
	.dc.b	$7F,$A6,$7E,$A3,$F9,$AB,$FE,$9B,$EE,$B2,$9B,$18,$6D,$FF,$BF,$6B
	.dc.b	$93,$6B,$F1,$B1,$FE,$A5,$FA,$8E,$67,$4E,$4B,$5B,$8B,$5E,$43,$EA
	.dc.b	$15,$FE,$8F,$F4,$CF,$62,$9F,$CD,$BE,$A3,$E8,$D7,$FB,$AA,$75,$7F
	.dc.b	$EB,$5E,$AF,$C0,$4A,$72,$FF,$36,$AF,$E3,$95,$FC,$3E,$73,$5F,$CD
	.dc.b	$E7,$35,$FE,$BF,$9C,$D7,$E2,$A7,$33,$D6,$FC,$9E,$67,$CC,$F5,$BF
	.dc.b	$8B,$DF,$C7,$7F,$19,$5F,$49,$CB,$7A,$38,$4F,$16,$BC,$5E,$9F,$51
	.dc.b	$23,$A7,$C5,$DA,$E9,$37,$4F,$D0,$6F,$FE,$32,$AC,$5F,$BF,$E4,$3E
	.dc.b	$69,$F8,$CF,$83,$7C,$B6,$27,$E5,$1B,$3F,$BA,$7D,$47,$D0,$C7,$E6
	.dc.b	$F0,$B0,$FE,$E1,$39,$AE,$1A,$7F,$76,$8C,$3F,$2E,$77,$DF,$F1,$FE
	.dc.b	$3A,$5F,$C7,$FA,$8B,$7F,$1E,$6E,$5E,$CE,$15,$62,$FF,$DE,$9B,$FF
	.dc.b	$94,$3E,$5F,$E9,$7B,$3F,$4D,$B3,$F4,$DB,$97,$D3,$6E,$3F,$88,$FA
	.dc.b	$8F,$F7,$A5,$7C,$2A,$FF,$80,$11,$87,$F6,$1C,$3F,$C9,$15,$71,$C3
	.dc.b	$FC,$3F,$5B,$8F,$F0,$17,$D4,$3E,$5B,$35,$FE,$10,$8C,$63,$DF,$06
	.dc.b	$1C,$63,$F8,$CC,$EA,$7B,$C7,$F6,$79,$D4,$7F,$81,$27,$58,$7C,$7F
	.dc.b	$A7,$FC,$A7,$D4,$7E,$55,$B8,$7E,$55,$B7,$F9,$56,$D7,$CF,$E1,$FC
	.dc.b	$E7,$4F,$E7,$5B,$7F,$9D,$6D,$FE,$75,$B5,$FB,$2F,$A7,$F8,$DF,$09
	.dc.b	$F0,$68,$21,$F2,$1F,$CD,$E7,$53,$9F,$53,$1C,$1F,$E0,$5B,$DE,$5F
	.dc.b	$E7,$E8,$C4,$EA,$BA,$D6,$9F,$F4,$6C,$EA,$B8,$46,$B0,$E0,$F8,$DB
	.dc.b	$5C,$70,$F1,$BE,$AB,$F1,$4E,$8D,$FD,$45,$C3,$AE,$1E,$D8,$E7,$E6
	.dc.b	$F6,$FA,$77,$77,$F8,$EA,$BC,$A7,$A2,$75,$19,$CF,$C3,$18,$F4,$7A
	.dc.b	$E8,$8C,$5F,$21,$87,$6A,$F1,$B6,$35,$5C,$72,$CE,$3D,$F4,$FF,$19
	.dc.b	$5E,$F2,$FA,$6D,$8F,$A5,$97,$82,$73,$5C,$71,$FB,$B7,$B7,$D9,$87
	.dc.b	$6C,$BC,$7F,$4F,$FE,$01,$9F,$F6,$7D,$23,$11,$FE,$D4,$8F,$43,$63
	.dc.b	$2F,$99,$E5,$87,$99,$B9,$67,$D5,$FA,$7F,$4C,$FA,$8F,$F5,$8C,$E7
	.dc.b	$0F,$91,$8F,$86,$FC,$A7,$0A,$DC,$FB,$30,$EA,$D9,$DB,$F4,$70,$EC
	.dc.b	$E3,$1F,$24,$DC,$7F,$CB,$D5,$AA,$F0,$DF,$47,$EC,$B7,$1F,$24,$75
	.dc.b	$C7,$C6,$EB,$0E,$8F,$8C,$D1,$83,$47,$67,$D8,$E6,$7A,$E3,$E6,$8E
	.dc.b	$26,$C8,$E4,$FD,$AE,$1E,$97,$18,$FC,$B6,$CF,$EF,$A7,$53,$92,$47
	.dc.b	$9E,$79,$E1,$EC,$BF,$1D,$6A,$BD,$35,$A7,$D1,$8F,$C5,$3E,$46,$1D
	.dc.b	$A3,$A5,$71,$BE,$71,$E3,$C3,$B5,$F1,$C7,$E3,$61,$C6,$BB,$E1,$C6
	.dc.b	$BA,$DF,$0B,$F0,$9F,$2A,$BE,$2D,$6E,$31,$8F,$C2,$9E,$67,$90,$F1
	.dc.b	$1C,$8E,$26,$0A,$70,$26,$1E,$5C,$7D,$71,$96,$C6,$71,$F6,$4F,$28
	.dc.b	$C5,$F7,$BF,$95,$3D,$AF,$99,$CB,$1F,$74,$67,$0F,$82,$F1,$C7,$EA
	.dc.b	$63,$D2,$35,$8F,$C0,$9F,$85,$1E,$1B,$E7,$8F,$C3,$C7,$B4,$F3,$3E
	.dc.b	$20,$C3,$A4,$EF,$1F,$0C,$F6,$3A,$1C,$4F,$8B,$7F,$A4,$AE,$38,$79
	.dc.b	$A3,$04,$9F,$31,$F9,$5F,$A8,$E1,$F0,$F2,$3F,$83,$BF,$AE,$FA,$87
	.dc.b	$C3,$66,$7F,$7D,$37,$3F,$EA,$F5,$A9,$FE,$82,$D7,$67,$43,$35,$D4
	.dc.b	$FF,$70,$73,$7F,$96,$27,$83,$63,$3C,$1F,$EE,$59,$7C,$2E,$2F,$FB
	.dc.b	$5F,$C1,$7F,$8C,$DF,$F0,$52,$FA,$8D,$42,$30,$EB,$C7,$CC,$71,$8F
	.dc.b	$F0,$69,$D2,$72,$6A,$77,$EE,$70,$FA,$19,$5E,$08,$CB,$E2,$8C,$BD
	.dc.b	$0F,$14,$79,$3E,$A1,$F5,$1E,$2F,$46,$57,$F8,$8E,$72,$D9,$E0,$47
	.dc.b	$DF,$B5,$7B,$4F,$06,$CF,$BA,$7E,$1F,$D4,$7F,$69,$9F,$F1,$83,$FE
	.dc.b	$D8,$6F,$C1,$EB,$B6,$46,$08,$7A,$E7,$93,$7E,$53,$A9,$1E,$BF,$15
	.dc.b	$5D,$1E,$08,$F1,$C7,$D6,$FA,$87,$9B,$E6,$F5,$25,$FF,$91,$67,$2D
	.dc.b	$9E,$04,$7D,$7E,$05,$7E,$84,$F0,$6C,$FC,$29,$FC,$2F,$D4,$3E,$A2
	.dc.b	$92,$BF,$CA,$90,$F9,$BC,$47,$DB,$8E,$0E,$A7,$9C,$F4,$F9,$EE,$BF
	.dc.b	$4B,$2B,$9C,$65,$EB,$19,$7A,$47,$E6,$3C,$7F,$F2,$1F,$77,$FA,$F3
	.dc.b	$E9,$DF,$51,$F9,$C6,$DF,$F9,$A6,$B7,$E0,$7F,$93,$A3,$FC,$5F,$E2
	.dc.b	$F3,$23,$FE,$6F,$AF,$D0,$61,$FE,$C8,$BE,$33,$FE,$56,$6B,$C8,$E1
	.dc.b	$DB,$E1,$EC,$91,$CD,$F2,$76,$E2,$3F,$41,$82,$71,$87,$AE,$BD,$5D
	.dc.b	$5E,$BF,$02,$61,$E5,$8C,$BE,$78,$CB,$D2,$3E,$37,$5F,$EF,$BD,$97
	.dc.b	$D4,$53,$35,$F1,$66,$BF,$B3,$D7,$F4,$0F,$ED,$6B,$F3,$79,$7C,$3A
	.dc.b	$DB,$67,$FD,$2D,$3A,$DF,$E7,$7D,$43,$8F,$F8,$7B,$6B,$FD,$38,$D7
	.dc.b	$F8,$12,$FA,$8D,$FA,$52,$76,$FF,$18,$FA,$87,$75,$24,$FF,$A9,$E3
	.dc.b	$2F,$FA,$A7,$EA,$3E,$DE,$5F,$81,$97,$C5,$F3,$25,$7F,$AB,$FE,$A3
	.dc.b	$F9,$7D,$7E,$1D,$B1,$FE,$96,$6A,$6F,$D4,$3E,$4F,$F9,$6B,$67,$A7
	.dc.b	$27,$F3,$0A,$DD,$7F,$32,$F6,$3F,$55,$5F,$EB,$FB,$FF,$50,$B7,$D8
	.dc.b	$D7,$7C,$8F,$F6,$25,$70,$7F,$95,$37,$91,$9F,$E7,$F0,$FF,$34,$C6
	.dc.b	$2B,$FD,$5F,$5B,$AF,$D2,$36,$33,$7E,$AF,$0B,$FA,$3F,$BF,$EB,$9F
	.dc.b	$E7,$1F,$48,$F8,$0F,$E9,$6D,$8F,$F6,$C4,$EF,$2F,$04,$23,$19,$7E
	.dc.b	$CD,$B1,$FE,$C2,$6A,$A7,$C3,$F8,$B6,$93,$D9,$F0,$C7,$F5,$CF,$11
	.dc.b	$F1,$FF,$D7,$BE,$95,$F1,$94,$D4,$4C,$3F,$DD,$D5,$C1,$B1,$C7,$E3
	.dc.b	$B5,$1F,$D5,$E7,$4D,$9D,$3F,$EF,$7F,$A7,$7D,$17,$87,$EA,$24,$E6
	.dc.b	$3F,$B8,$CE,$A3,$04,$3F,$B2,$65,$FC,$15,$C7,$07,$CA,$9E,$11,$FD
	.dc.b	$A5,$B1,$F3,$A7,$84,$FF,$6D,$FA,$8F,$B8,$A3,$FC,$00,$FF,$BC,$27
	.dc.b	$87,$B3,$FE,$04,$AD,$93,$97,$07,$CB,$FA,$0B,$7C,$79,$3F,$96,$56
	.dc.b	$E3,$07,$FB,$F3,$15,$75,$F8,$6F,$A9,$ED,$4A,$D8,$B4,$EE,$33,$18
	.dc.b	$6C,$7D,$DE,$DF,$F0,$DF,$D4,$7E,$1E,$79,$B5,$FF,$07,$37,$5F,$C9
	.dc.b	$FD,$43,$D8,$42,$9F,$50,$FA,$8F,$E9,$18,$71,$DE,$F5,$F8,$0F,$F1
	.dc.b	$5E,$5C,$F1,$E5,$87,$1C,$7A,$4E,$5E,$B8,$F2,$76,$D7,$27,$0C,$BC
	.dc.b	$AE,$04,$F3,$AE,$D0,$CB,$E0,$46,$1B,$7F,$D8,$6B,$83,$6F,$AD,$6F
	.dc.b	$1F,$EC,$CD,$C3,$CA,$78,$F1,$AF,$8A,$B8,$DF,$A6,$B8,$4F,$9F,$0E
	.dc.b	$EF,$13,$D5,$8F,$91,$FA,$0D,$7A,$61,$DA,$78,$C7,$B7,$1F,$2E,$1D
	.dc.b	$CF,$55,$6F,$1F,$1D,$74,$AD,$B4,$D4,$C7,$FC,$08,$79,$4B,$7F,$5D
	.dc.b	$B5,$F4,$9C,$F8,$CF,$91,$E3,$1F,$7A,$30,$E2,$BF,$AB,$CE,$A3,$EC
	.dc.b	$CE,$9C,$7C,$16,$5F,$C9,$35,$34,$8F,$E0,$FE,$65,$F4,$DF,$CF,$76
	.dc.b	$FC,$1C,$F2,$35,$19,$D7,$CF,$7F,$6A,$F8,$EF,$F7,$2F,$86,$1F,$8C
	.dc.b	$9C,$C7,$E4,$4C,$DF,$F8,$AE,$33,$3E,$09,$F3,$9E,$0A,$E8,$E2,$3A
	.dc.b	$61,$E3,$8F,$A3,$87,$8E,$31,$1F,$E3,$78,$C5,$72,$6C,$78,$CE,$98
	.dc.b	$78,$EB,$8C,$79,$8F,$DE,$61,$E6,$9F,$83,$1E,$53,$95,$74,$C3,$CB
	.dc.b	$87,$F1,$4B,$F8,$3A,$FF,$5A,$BE,$82,$33,$3E,$13,$87,$07,$F6,$FA
	.dc.b	$FE,$C3,$3A,$AF,$1C,$E6,$BE,$57,$4D,$9F,$E2,$7F,$50,$EE,$C4,$77
	.dc.b	$9D,$62,$AD,$5F,$9E,$73,$18,$AF,$F5,$76,$3F,$D5,$2B,$F9,$AD,$F1
	.dc.b	$C3,$AC,$EE,$3C,$F1,$E3,$F8,$C7,$C1,$7F,$7F,$BF,$89,$3C,$0D,$7D
	.dc.b	$43,$E2,$31,$3F,$D0,$AD,$39,$8E,$26,$5C,$65,$E6,$BE,$77,$F0,$3C
	.dc.b	$8F,$E8,$77,$FD,$19,$B5,$FE,$D4,$87,$D3,$31,$E1,$D5,$27,$FA,$99
	.dc.b	$AE,$4C,$5F,$1B,$FA,$8B,$0E,$BA,$FF,$89,$EF,$81,$E1,$C3,$EB,$38
	.dc.b	$AF,$F6,$D7,$D4,$7F,$BC,$A3,$FA,$DB,$63,$B7,$17,$DA,$8C,$5F,$F8
	.dc.b	$91,$B5,$FD,$44,$F5,$FB,$5B,$3B,$E5,$AC,$3F,$AB,$55,$AD,$9E,$0A
	.dc.b	$D9,$F8,$87,$FC,$CE,$7F,$6E,$2C,$FC,$1E,$FD,$74,$49,$F8,$FF,$51
	.dc.b	$FE,$5B,$78,$47,$C0,$8C,$57,$FB,$E6,$79,$E5,$F1,$E7,$2F,$F9,$02
	.dc.b	$32,$7B,$63,$2F,$FA,$CF,$E5,$FF,$20,$DE,$2F,$7C,$FE,$82,$73,$7F
	.dc.b	$DD,$5B,$5F,$80,$F1,$35,$7F,$19,$B1,$F8,$0D,$51,$8B,$FC,$03,$FA
	.dc.b	$CF,$A6,$79,$F4,$FD,$7F,$D4,$3E,$A3,$18,$FF,$85,$AB,$6D,$F7,$73
	.dc.b	$8F,$F8,$2A,$3F,$D9,$D2,$9E,$C7,$EE,$BE,$A3,$F1,$D3,$C4,$FF,$10
	.dc.b	$B8,$AE,$47,$F8,$4E,$BE,$3B,$FE,$15,$E8,$FF,$64,$57,$8F,$D1,$FE
	.dc.b	$0B,$18,$9E,$27,$E8,$FE,$0B,$3E,$8E,$9B,$7F,$E3,$0F,$A8,$FF,$36
	.dc.b	$BF,$E3,$2E,$F7,$B9,$FC,$42,$73,$87,$49,$CD,$7F,$8E,$3E,$A3,$FD
	.dc.b	$56,$DF,$6F,$F8,$D7,$D4,$3E,$A1,$F5,$1F,$C6,$AF,$78,$F9,$9B,$D7
	.dc.b	$FE,$95,$F1,$9F,$51,$FE,$7A,$CF,$A1,$FE,$4C,$8C,$FB,$1C,$F4,$FE
	.dc.b	$51,$18,$9D,$E3,$F2,$A3,$0F,$CC,$ED,$FF,$1F,$7D,$47,$F2,$F7,$FC
	.dc.b	$81,$DD,$65,$36,$30,$DB,$FF,$2E,$D7,$26,$D7,$E5,$A3,$FC,$95,$F5
	.dc.b	$1C,$CE,$9C,$B6,$B7,$16,$BC,$87,$FA,$5A,$32,$7F,$99,$3E,$99,$EB
	.dc.b	$FF,$42,$6C,$7F,$2D,$FA,$8F,$A7,$5F,$80,$9C,$B6,$BF,$9D,$4F,$F1
	.dc.b	$CA,$FF,$2F,$4A,$B7,$C1,$FE,$C5,$E9,$5E,$EB,$FC,$C9,$5A,$BF,$C9
	.dc.b	$4E,$67,$96,$1D,$BC,$9F,$99,$E9,$FF,$B4,$AF,$E3,$BF,$8A,$CB,$FD
	.dc.b	$A5,$AF,$F8,$A6,$77,$EF,$70,$69,$C0,$B8,$FD,$9B,$7D,$47,$63,$BF
	.dc.b	$C5,$DA,$EF,$16,$B8,$4D,$C7,$FD,$A7,$AB,$E3,$FE,$61,$F8,$AF,$82
	.dc.b	$7C,$B6,$32,$FA,$53,$FD,$C3,$EA,$3E,$7E,$1C,$A3,$1E,$0E,$F3,$FD
	.dc.b	$CC,$AB,$77,$F4,$32,$F0,$3F,$DC,$A3,$34,$DF,$FC,$99,$F1,$DF,$28
	.dc.b	$F9,$7F,$51,$6F,$E5,$C5,$AF,$89,$37,$1F,$9B,$C0,$F9,$7F,$A7,$F1
	.dc.b	$1F,$51,$B8,$FD,$46,$E1,$F3,$5B,$7F,$EF,$8F,$A8,$FF,$00,$4F,$F0
	.dc.b	$C6,$C7,$ED,$F8,$7F,$91,$CE,$70,$FF,$0B,$D6,$B0,$FE,$31,$F5,$0F
	.dc.b	$96,$CD,$7F,$83,$63,$13,$FE,$C2,$AE,$17,$FD,$EE,$72,$F5,$9F,$F0
	.dc.b	$64,$63,$1F,$7F,$D3,$FE,$43,$EA,$3F,$22,$DA,$FC,$8B,$6F,$F1,$4D
	.dc.b	$AF,$8F,$C7,$F9,$8E,$9F,$CC,$B6,$FF,$32,$DB,$FC,$93,$69,$3F,$AC
	.dc.b	$FA,$7F,$87,$F0,$7E,$FB,$FF,$66,$43,$E4,$37,$1F,$E7,$CD,$F8,$7A
	.dc.b	$B8,$BE,$12,$BE,$18,$7F,$A1,$2F,$6D,$9E,$B8,$71,$9C,$C7,$53,$33
	.dc.b	$B7,$A4,$ED,$C9,$B3,$C1,$86,$5D,$CF,$1B,$F1,$C6,$D7,$D4,$6D,$CE
	.dc.b	$3E,$A7,$35,$C7,$CE,$E9,$F5,$0E,$EF,$F1,$C6,$3C,$E7,$31,$8C,$39
	.dc.b	$46,$3C,$3E,$58,$7A,$F0,$E5,$16,$F8,$6B,$FC,$7A,$73,$8F,$5B,$D5
	.dc.b	$C8,$F8,$13,$AC,$7B,$D6,$5F,$A7,$0F,$A8,$FF,$23,$57,$FB,$02,$31
	.dc.b	$F0,$3F,$EB,$D7,$BC,$7F,$B3,$1C,$BB,$9E,$35,$D1,$B8,$F9,$DB,$21
	.dc.b	$F5,$0F,$A8,$FF,$52,$56,$F1,$F8,$57,$C2,$D8,$F8,$31,$FC,$5E,$3C
	.dc.b	$67,$77,$1F,$83,$7C,$67,$51,$89,$D5,$79,$DF,$C5,$4E,$A3,$13,$B2
	.dc.b	$3E,$0C,$7C,$F1,$EB,$C3,$C9,$8F,$A6,$36,$FF,$54,$BE,$8E,$6F,$A9
	.dc.b	$1D,$5F,$4A,$F7,$E3,$F0,$70,$EB,$1D,$E3,$D9,$3B,$9C,$CF,$0C,$7D
	.dc.b	$B7,$D2,$29,$C6,$2B,$F4,$AB,$95,$F7,$24,$FC,$48,$CB,$C2,$FB,$E3
	.dc.b	$EC,$BE,$87,$C0,$2E,$3F,$1D,$E1,$87,$47,$F6,$93,$C5,$F4,$1C,$4B
	.dc.b	$3C,$5C,$57,$B8,$F0,$E1,$E4,$8C,$DF,$C4,$7B,$E1,$EE,$C3,$B5,$F1
	.dc.b	$AF,$16,$1D,$4F,$A5,$7E,$8C,$7E,$B1,$F2,$5E,$B8,$73,$C7,$C1,$5C
	.dc.b	$27,$CF,$5B,$8C,$61,$DE,$FE,$55,$79,$0F,$76,$3E,$4C,$7C,$78,$78
	.dc.b	$27,$F4,$D3,$9A,$E1,$87,$9E,$B5,$5E,$03,$D0,$70,$C7,$CB,$1A,$C7
	.dc.b	$E3,$CF,$28,$E3,$87,$58,$F6,$9F,$52,$B8,$BD,$0E,$C7,$E1,$CE,$46
	.dc.b	$63,$95,$FC,$0C,$7E,$94,$F4,$8F,$8D,$3A,$BF,$25,$6A,$7D,$95,$DF
	.dc.b	$0F,$3B,$E3,$3E,$49,$4F,$C4,$FD,$47,$E9,$F0,$FF,$7B,$47,$EE,$BE
	.dc.b	$1B,$07,$EA,$FE,$A1,$F0,$D8,$9D,$10,$FE,$8D,$1A,$7F,$A6,$35,$FC
	.dc.b	$BC,$F1,$77,$3D,$4F,$F2,$8F,$37,$F6,$28,$57,$18,$C3,$77,$FF,$2B
	.dc.b	$AF,$85,$E3,$1F,$E5,$AF,$81,$FF,$1D,$5F,$F2,$7B,$E9,$0F,$A8,$DC
	.dc.b	$FF,$2B,$8C,$3B,$AE,$5E,$4E,$A3,$FD,$8E,$D9,$DC,$7F,$AB,$1B,$7D
	.dc.b	$A7,$5E,$E6,$FE,$82,$5F,$53,$C1,$1A,$70,$7F,$B5,$23,$6F,$58,$F1
	.dc.b	$7D,$43,$EA,$3C,$5E,$8C,$AE,$33,$96,$CF,$02,$3D,$B2,$F9,$4F,$F4
	.dc.b	$B9,$E0,$FF,$B1,$9B,$97,$D3,$6E,$7F,$37,$7F,$89,$F5,$0E,$CB,$F0
	.dc.b	$93,$1F,$5C,$E5,$B3,$97,$FD,$A7,$1F,$EE,$37,$27,$0F,$93,$EA,$47
	.dc.b	$9F,$C5,$57,$8C,$5C,$FF,$11,$4A,$FB,$1F,$51,$FE,$5F,$BF,$1F,$A9
	.dc.b	$E8,$33,$E7,$48,$FA,$33,$B7,$FD,$CE,$DC,$3A,$F8,$1B,$9C,$D5,$9C
	.dc.b	$C7,$DA,$FA,$87,$D4,$7E,$25,$B1,$98,$C7,$CD,$E5,$91,$93,$94,$71
	.dc.b	$F9,$EE,$5F,$4B,$27,$B4,$65,$F0,$46,$5F,$F7,$FD,$A7,$F3,$5E,$3F
	.dc.b	$F9,$2B,$1F,$D7,$7D,$3B,$EA,$3F,$3C,$DC,$3F,$27,$5B,$E6,$FF,$62
	.dc.b	$B6,$3F,$C2,$4D,$7F,$C7,$1E,$3F,$2E,$EE,$EC,$8E,$6E,$5F,$CF,$DB
	.dc.b	$B7,$C2,$48,$CF,$75,$59,$FA,$27,$0F,$AA,$7D,$3D,$5E,$AF,$02,$5F
	.dc.b	$68,$CB,$E0,$8C,$BD,$34,$FF,$70,$E3,$FA,$48,$FF,$29,$75,$FF,$01
	.dc.b	$F6,$1F,$51,$4F,$97,$12,$7E,$3B,$5E,$87,$F8,$DE,$7F,$44,$E0,$EA
	.dc.b	$70,$21,$97,$79,$FF,$23,$FA,$87,$27,$F1,$36,$D7,$F0,$86,$BD,$57
	.dc.b	$D4,$6F,$D2,$93,$B7,$F8,$E7,$D3,$FE,$9E,$3F,$88,$B5,$F3,$37,$0F
	.dc.b	$E6,$8D,$D7,$FC,$A5,$2F,$A8,$FE,$5E,$F0,$F8,$9E,$A4,$7F,$8E,$FD
	.dc.b	$47,$CD,$27,$17,$F2,$D7,$E8,$7D,$43,$E4,$FF,$98,$36,$B8,$D6,$CC
	.dc.b	$CF,$F3,$5F,$5D,$DD,$C3,$D8,$97,$FC,$B2,$72,$D9,$E0,$47,$F9,$8E
	.dc.b	$1F,$E7,$A7,$EB,$4F,$06,$CE,$DF,$B5,$3B,$DB,$FA,$C7,$BF,$E7,$9F
	.dc.b	$BD,$F4,$6F,$80,$FE,$94,$D8,$FE,$7F,$3B,$AE,$39,$FF,$4F,$8C,$70
	.dc.b	$7C,$E6,$D7,$FA,$06,$DF,$0F,$BB,$93,$D1,$EF,$1F,$D6,$7C,$27,$C7
	.dc.b	$7F,$5B,$FA,$67,$C6,$53,$51,$30,$FE,$A7,$5C,$1B,$1C,$7E,$3B,$51
	.dc.b	$F9,$B9,$D3,$67,$4F,$F5,$CF,$A7,$7D,$17,$87,$EA,$24,$E6,$3F,$B7
	.dc.b	$4E,$A3,$04,$3F,$51,$97,$F0,$77,$1C,$1F,$E0,$79,$E1,$1F,$AD,$6C
	.dc.b	$7F,$82,$E7,$84,$FE,$C7,$EA,$3E,$E4,$E7,$D6,$F8,$87,$F7,$F8,$CF
	.dc.b	$17,$E7,$9B,$5F,$EB,$16,$A6,$DB,$E8,$39,$73,$4A,$FE,$FB,$5A,$9E
	.dc.b	$DF,$0D,$F4,$FD,$A9,$5A,$8C,$B4,$87,$C2,$EB,$FF,$0A,$7D,$43,$9E
	.dc.b	$92,$7F,$C1,$7F,$50,$F6,$7F,$94,$C3,$E9,$DF,$51,$FB,$AC,$39,$4F
	.dc.b	$F8,$49,$AF,$F1,$76,$F1,$7F,$8D,$70,$E8,$FF,$B4,$EF,$9D,$7F,$B5
	.dc.b	$67,$2F,$7B,$ED,$8F,$F0,$8A,$DC,$B0,$EF,$18,$BE,$2D,$C7,$D2,$F0
	.dc.b	$9D,$E1,$D2,$FC,$4F,$F1,$4B,$E2,$DB,$E2,$EE,$7C,$87,$FB,$AE,$B9
	.dc.b	$BE,$99,$DB,$71,$EA,$EF,$1F,$55,$F3,$C7,$CD,$5A,$A6,$3E,$27,$81
	.dc.b	$9C,$7B,$BA,$C7,$BB,$6B,$58,$F6,$8E,$27,$A0,$D1,$6F,$B3,$F2,$E3
	.dc.b	$83,$4F,$24,$68,$FA,$AF,$F3,$A3,$99,$5C,$BF,$26,$74,$FD,$DF,$69
	.dc.b	$F1,$F8,$8F,$7C,$61,$DD,$7F,$8C,$E7,$4D,$8E,$03,$E1,$AB,$4C,$B8
	.dc.b	$23,$F0,$7E,$61,$F4,$CE,$6F,$A1,$23,$96,$DF,$8E,$AD,$CE,$A7,$3B
	.dc.b	$FF,$5A,$9C,$C7,$CD,$39,$61,$F6,$E3,$37,$F9,$89,$C9,$E3,$C3,$C4
	.dc.b	$F1,$3A,$9E,$39,$E3,$7F,$2A,$FA,$57,$0B,$E9,$7C,$05,$F0,$7E,$B5
	.dc.b	$70,$C3,$D4,$D8,$F0,$36,$B9,$B7,$0F,$04,$F9,$EB,$8B,$49,$A7,$FA
	.dc.b	$C7,$0F,$86,$FF,$AC,$E3,$DF,$87,$CB,$C7,$DF,$3B,$6D,$FC,$BE,$0B
	.dc.b	$7F,$06,$6B,$EC,$8B,$87,$F4,$4A,$FF,$52,$5B,$EA,$1D,$DA,$EB,$E7
	.dc.b	$39,$8C,$57,$0A,$DB,$FE,$AA,$BE,$98,$FF,$8C,$AF,$84,$E6,$FC,$27
	.dc.b	$9B,$E3,$3F,$D3,$5E,$26,$27,$47,$FA,$DE,$1F,$50,$F8,$8D,$63,$FE
	.dc.b	$3E,$9C,$C6,$31,$FC,$1C,$62,$F9,$D6,$EF,$94,$7F,$B0,$3E,$01,$D3
	.dc.b	$F1,$9B,$1F,$EC,$71,$F4,$CC,$35,$52,$75,$3F,$61,$FF,$1A,$D7,$2C
	.dc.b	$3F,$CA,$71,$CB,$87,$84,$62,$B8,$3F,$A9,$AE,$19,$7E,$72,$F8,$1F
	.dc.b	$ED,$6F,$A8,$FF,$78,$37,$D9,$FB,$EF,$9B,$8F,$EF,$CE,$B0,$FF,$30
	.dc.b	$36,$FA,$63,$6F,$FC,$97,$36,$BF,$79,$35,$46,$67,$E1,$E1,$D8,$F0
	.dc.b	$D6,$E3,$FD,$84,$7F,$BB,$0B,$5E,$FE,$EF,$C9,$C6,$30,$FD,$CC,$FF
	.dc.b	$0B,$EA,$F5,$FD,$47,$F9,$4E,$B9,$46,$3C,$AE,$2D,$8C,$9F,$5E,$78
	.dc.b	$57,$F9,$FF,$E6,$3F,$2A,$DE,$5F,$3C,$FE,$92,$73,$8F,$89,$B7,$FC
	.dc.b	$6C,$FC,$0F,$8B,$BB,$F7,$B6,$3F,$02,$DC,$3E,$E3,$FB,$0F,$A5,$79
	.dc.b	$F4,$FD,$9F,$D4,$3E,$A3,$18,$7F,$29,$CF,$E1,$DF,$08,$C1,$FE,$80
	.dc.b	$5E,$E7,$0C,$BF,$D3,$91,$89,$DC,$7F,$33,$8C,$3F,$EA,$1F,$A8,$73
	.dc.b	$79,$6F,$F8,$2C,$62,$F8,$D2,$FE,$5E,$1F,$5B,$A7,$FD,$91,$5A,$9C
	.dc.b	$BF,$E9,$48,$CD,$7F,$AB,$DB,$5C,$7D,$0F,$E2,$71,$8A,$E1,$1F,$C5
	.dc.b	$BE,$A3,$FC,$DB,$1E,$2F,$8D,$7B,$79,$F5,$3F,$8D,$C6,$27,$CF,$F4
	.dc.b	$AF,$6B,$F8,$FB,$63,$E1,$FD,$43,$EA,$1F,$51,$AB,$95,$F2,$3C,$77
	.dc.b	$D4,$7F,$9E,$23,$AC,$FF,$2C,$9C,$FB,$3C,$38,$BF,$D9,$F1,$8A,$D9
	.dc.b	$CE,$F9,$4F,$FB,$53,$B3,$FB,$37,$D4,$7F,$36,$7F,$B4,$76,$D9,$4D
	.dc.b	$8C,$36,$FE,$ED,$72,$6D,$7E,$60,$47,$F7,$3F,$A8,$E8,$75,$E4,$B5
	.dc.b	$C2,$2D,$79,$4F,$F4,$AC,$64,$FC,$1F,$D3,$3D,$7F,$E9,$6D,$8F,$E6
	.dc.b	$DF,$51,$F4,$EB,$FD,$DD,$39,$6D,$7F,$3F,$9F,$E3,$B5,$FE,$DA,$AD
	.dc.b	$74,$FF,$38,$9C,$DF,$F8,$2E,$73,$5F,$E0,$DE,$95,$6A,$95,$A7,$F9
	.dc.b	$B4,$72,$CF,$FC,$4F,$7F,$3F,$D5,$FC,$6D,$7D,$37,$FC,$35,$1F,$31
	.dc.b	$CE,$3F,$E0,$CA,$E3,$39,$E8,$E4,$F3,$7A,$7D,$44,$8E,$5F,$21,$71
	.dc.b	$ED,$37,$0E,$98,$AB,$17,$F0,$FE,$43,$E6,$9F,$8F,$F8,$27,$CB,$62
	.dc.b	$7E,$9C,$FF,$74,$FA,$8F,$A1,$8F,$F2,$AD,$D3,$9E,$FE,$1A,$FF,$73
	.dc.b	$BD,$E1,$DC,$FF,$00,$B8,$8F,$C4,$E8,$F8,$DF,$90,$7C,$9F,$A8,$B7
	.dc.b	$F1,$E2,$D7,$CB,$9B,$9F,$BB,$E5,$FE,$87,$B1,$F4,$5B,$87,$CF,$6D
	.dc.b	$FD,$36,$D7,$F4,$4F,$A8,$FF,$7D,$57,$F1,$46,$D7,$EC,$38,$3F,$93
	.dc.b	$2A,$E1,$87,$F8,$82,$B7,$1F,$E0,$4F,$A8,$7C,$B6,$6B,$FC,$25,$18
	.dc.b	$C7,$B4,$62,$75,$1F,$E1,$39,$D6,$1F,$89,$FA,$7F,$C6,$7D,$47,$E3
	.dc.b	$5B,$5F,$8B,$6C,$EE,$3F,$1D,$3F,$93,$E1,$FC,$A7,$B7,$67,$F2,$4D
	.dc.b	$AF,$CB,$37,$3F,$C4,$BF,$E3,$CF,$A7,$F8,$3F,$07,$EC,$AF,$F3,$94
	.dc.b	$3E,$3F,$FC,$CB,$39,$94,$F0,$D7,$FC,$F1,$18,$AD,$C6,$27,$53,$D2
	.dc.b	$72,$F6,$F5,$2B,$C7,$2F,$F3,$CD,$F1,$AC,$DF,$8E,$73,$5C,$5A,$F1
	.dc.b	$3F,$D2,$95,$C6,$33,$2F,$A8,$E7,$7D,$1E,$BE,$77,$8B,$E9,$BF,$59
	.dc.b	$E1,$70,$8C,$77,$7F,$8E,$73,$EA,$7C,$68,$CC,$EF,$2F,$1D,$F2,$AD
	.dc.b	$87,$1A,$FC,$B6,$E9,$F1,$F0,$E7,$5C,$25,$F4,$FF,$DC,$23,$FD,$77
	.dc.b	$18,$F5,$7F,$D7,$4F,$78,$FF,$62,$39,$77,$3C,$70,$F0,$37,$0F,$B3
	.dc.b	$5B,$9F,$BD,$F4,$CF,$A8,$FB,$B1,$C3,$1F,$6E,$1C,$71,$FF,$5C,$61
	.dc.b	$C7,$1F,$C1,$E3,$C6,$77,$71,$F6,$5F,$15,$3B,$73,$3B,$C3,$C1,$1F
	.dc.b	$2A,$75,$18,$25,$FA,$AF,$A4,$EA,$31,$7C,$DB,$7C,$F1,$F6,$C6,$30
	.dc.b	$E7,$3E,$CC,$39,$D6,$A3,$17,$CD,$A7,$33,$07,$DF,$9F,$76,$3F,$16
	.dc.b	$3F,$DB,$D8,$F9,$5E,$38,$78,$1E,$17,$F9,$C9,$D4,$60,$F8,$A4,$9F
	.dc.b	$AF,$87,$96,$35,$1C,$EF,$C1,$1F,$3A,$75,$18,$26,$1E,$E9,$F4,$C6
	.dc.b	$30,$F1,$57,$0A,$F5,$92,$33,$5C,$23,$0F,$C3,$3C,$D8,$FC,$D2,$3F
	.dc.b	$0A,$BA,$57,$13,$9C,$70,$3E,$2C,$F0,$7D,$46,$CD,$18,$2D,$FA,$5F
	.dc.b	$16,$1F,$0F,$0F,$2B,$F2,$23,$93,$F4,$71,$F9,$F5,$C6,$31,$8F,$9A
	.dc.b	$BA,$DF,$C7,$6E,$3F,$80,$C3,$8C,$F6,$C7,$D1,$18,$9E,$F1,$FA,$6B
	.dc.b	$E9,$87,$96,$7E,$0B,$70,$F6,$DF,$17,$B9,$EC,$BF,$89,$8F,$9D,$E9
	.dc.b	$1D,$8F,$92,$7C,$F3,$07,$E1,$9D,$BD,$2B,$DE,$EA,$7F,$13,$5F,$92
	.dc.b	$3C,$53,$A9,$F1,$1D,$4A,$7E,$F3,$EA,$3F,$29,$3F,$EF,$18,$FC,$E7
	.dc.b	$C3,$60,$FC,$C7,$D4,$3E,$17,$13,$A2,$1F,$CA,$A3,$4F,$F4,$06,$BF
	.dc.b	$C1,$67,$8B,$B9,$EA,$7F,$56,$E7,$FF,$63,$D7,$06,$C6,$78,$3F,$AD
	.dc.b	$CB,$E1,$31,$61,$EF,$FF,$89,$9F,$F4,$FA,$FA,$8D,$C6,$9B,$1B,$8F
	.dc.b	$F0,$75,$FF,$68,$8C,$63,$FE,$DA,$79,$4F,$4A,$E3,$E3,$F2,$FA,$1E
	.dc.b	$7F,$43,$2B,$C3,$19,$7C,$71,$97,$B9,$E9,$B4,$F9,$BE,$9D,$F5,$1E
	.dc.b	$6F,$4A,$5F,$FA,$EA,$72,$D9,$E0,$47,$E0,$EF,$5E,$E2,$78,$B8,$6E
	.dc.b	$9E,$F9,$F8,$FF,$50,$EA,$BF,$07,$30,$FF,$65,$CE,$5B,$39,$7B,$C7
	.dc.b	$FB,$71,$C9,$C3,$E5,$3A,$91,$E7,$F1,$15,$E3,$17,$17,$17,$F8,$72
	.dc.b	$D7,$F7,$3E,$A1,$E6,$FA,$3D,$39,$97,$FB,$82,$72,$D9,$E0,$47,$D7
	.dc.b	$DD,$5F,$8F,$3C,$1B,$3F,$12,$7F,$17,$F4,$EF,$A8,$A4,$AF,$F7,$7C
	.dc.b	$3E,$6B,$11,$F4,$A3,$83,$A9,$E7,$3D,$3E,$7B,$AF,$D2,$CA,$E7,$19
	.dc.b	$7A,$C6,$5E,$91,$F9,$EF,$1F,$FC,$89,$C1,$F2,$27,$F4,$DF,$4E,$FA
	.dc.b	$8F,$D1,$B6,$3F,$55,$3B,$7F,$D3,$5E,$37,$32,$3A,$E3,$FF,$26,$E1
	.dc.b	$D9,$FF,$27,$DF,$EA,$1C,$D7,$0A,$FC,$8D,$6D,$F1,$F7,$78,$BB,$24
	.dc.b	$75,$3C,$EF,$9B,$BB,$31,$F9,$AC,$7F,$4E,$EC,$38,$75,$1E,$EE,$AF
	.dc.b	$7F,$81,$27,$85,$7F,$0F,$CF,$D3,$59,$9E,$D5,$C5,$ED,$97,$E6,$DF
	.dc.b	$D5,$FA,$EF,$A8,$FE,$1A,$D8,$DF,$6F,$D3,$23,$83,$D1,$1F,$A9,$7E
	.dc.b	$2C,$69,$A6,$BB,$7F,$8C,$FD,$43,$92,$92,$7E,$14,$3E,$A3,$7E,$9C
	.dc.b	$9C,$BF,$C8,$BE,$A3,$F5,$31,$F2,$5A,$FF,$06,$6F,$2F,$E0,$DB,$C1
	.dc.b	$FE,$8C,$FA,$8F,$BD,$5F,$CC,$65,$F1,$7D,$49,$1F,$57,$EA,$3E,$73
	.dc.b	$FC,$DD,$A9,$A6,$3E,$A1,$F3,$1F,$CE,$1B,$3C,$F9,$7F,$9D,$D6,$EB
	.dc.b	$F9,$EF,$B1,$FA,$FA,$FC,$25,$FF,$A9,$DB,$EB,$EB,$BE,$47,$E2,$6B
	.dc.b	$83,$FC,$99,$BB,$7E,$8E,$FF,$CF,$D3,$9A,$E2,$D3,$BF,$85,$FD,$3B
	.dc.b	$DF,$F3,$C7,$F3,$0F,$A3,$7B,$FF,$D4,$9B,$1F,$99,$9D,$E3,$FB,$16
	.dc.b	$F1,$7F,$AE,$9B,$3F,$D7,$DA,$A9,$F0,$DE,$1D,$A4,$F5,$7C,$11,$FD
	.dc.b	$8B,$C4,$7C,$7F,$F6,$4F,$A6,$7C,$65,$35,$13,$0F,$D7,$57,$06,$C7
	.dc.b	$1F,$8E,$D4,$7F,$4E,$9D,$36,$74,$FE,$DF,$E9,$DF,$45,$E1,$FA,$89
	.dc.b	$39,$8F,$EE,$F3,$A8,$C1,$0F,$EB,$79,$7E,$FF,$83,$F8,$C4,$EE,$7F
	.dc.b	$AF,$B6,$33,$5E,$FF,$23,$FC,$07,$F5,$1F,$7B,$8B,$FD,$CF,$8F,$F8
	.dc.b	$2B,$1F,$E1,$8D,$7F,$B4,$CE,$DF,$F0,$83,$53,$4C,$43,$E7,$FB,$F3
	.dc.b	$4A,$FE,$31,$5A,$9F,$1F,$C3,$7D,$3F,$66,$56,$A3,$2D,$21,$F0,$BB
	.dc.b	$7F,$C3,$BF,$51,$F8,$38,$FE,$48,$D7,$FC,$02,$DD,$BF,$9F,$7D,$43
	.dc.b	$D9,$47,$F8,$AF,$E9,$DF,$51,$FE,$2D,$C1,$87,$1E,$0C,$36,$3F,$C6
	.dc.b	$4D,$58,$F5,$C3,$A5,$F3,$8C,$4B,$0E,$EF,$2E,$BF,$04,$62,$3C,$31
	.dc.b	$8A,$E5,$97,$A2,$79,$46,$30,$EF,$1C,$CA,$72,$2E,$1E,$2C,$3B,$36
	.dc.b	$BD,$B8,$77,$AE,$47,$63,$FC,$A1,$3B,$6E,$1E,$3B,$F4,$63,$F4,$5B
	.dc.b	$7C,$EB,$E2,$D6,$F0,$F4,$47,$53,$CC,$5B,$E5,$1A,$BF,$3D,$71,$AF
	.dc.b	$2C,$76,$7D,$B1,$97,$BE,$3F,$0A,$F9,$C7,$0B,$E8,$78,$9F,$C0,$1C
	.dc.b	$0A,$53,$38,$FD,$A6,$FB,$4F,$90,$DC,$7E,$0A,$30,$F2,$9F,$F1,$CC
	.dc.b	$E9,$B1,$C0,$7C,$2D,$69,$97,$04,$7E,$47,$CC,$BE,$9B,$FC,$41,$AF
	.dc.b	$E2,$67,$FD,$0D,$5F,$82,$AD,$6F,$F8,$49,$C9,$CA,$F8,$E7,$F0,$2F
	.dc.b	$84,$62,$7E,$2C,$62,$B9,$47,$9B,$0F,$CC,$61,$DA,$3E,$E1,$DA,$33
	.dc.b	$7C,$E3,$F5,$D7,$CD,$D9,$CF,$0F,$D9,$5F,$3B,$F6,$4E,$4F,$F5,$14
	.dc.b	$66,$7D,$11,$EB,$C3,$CB,$7D,$6F,$8B,$6B,$71,$E4,$9F,$16,$BF,$E4
	.dc.b	$3C,$3D,$4F,$F9,$17,$0F,$55,$F9,$31,$FF,$27,$4E,$DB,$7E,$8E,$0B
	.dc.b	$7E,$69,$AB,$0F,$F6,$75,$7F,$8E,$DF,$E8,$AF,$F9,$0E,$DF,$50,$EE
	.dc.b	$E3,$5A,$BF,$F5,$5D,$6A,$73,$D5,$E0,$8C,$61,$D6,$31,$87,$38,$F0
	.dc.b	$57,$43,$CF,$F1,$9F,$EA,$1F,$13,$13,$A3,$FC,$AD,$0F,$A8,$7C,$46
	.dc.b	$B3,$FF,$5E,$CE,$63,$19,$FF,$9C,$E3,$17,$CE,$B7,$7C,$A3,$FC,$B9
	.dc.b	$E3,$BA,$7E,$3B,$63,$FC,$CC,$3E,$9D,$86,$AA,$4E,$A7,$EC,$BF,$CE
	.dc.b	$EB,$96,$1F,$ED,$28,$E5,$C1,$C2,$31,$5C,$1F,$F7,$75,$70,$CB,$FD
	.dc.b	$19,$7C,$0F,$F3,$97,$D4,$7F,$BD,$5B,$ED,$7E,$01,$F3,$71,$FE,$02
	.dc.b	$75,$87,$FB,$51,$B7,$D3,$1B,$7F,$D4,$66,$D7,$FB,$0A,$6A,$8C,$CF
	.dc.b	$C3,$C3,$B1,$E1,$AD,$C7,$FB,$20,$FF,$44,$16,$BD,$FD,$DF,$94,$8C
	.dc.b	$61,$FE,$A3,$68,$9F,$F6,$9F,$D4,$7F,$95,$A3,$9F,$5F,$BE,$31,$3F
	.dc.b	$EF,$09,$E2,$F0,$F2,$BB,$56,$DF,$D1,$FC,$DF,$E4,$E7,$FD,$EE,$DC
	.dc.b	$3E,$26,$1A,$E5,$98,$FD,$DF,$8A,$B8,$7D,$86,$DF,$CB,$6C,$FA,$4F
	.dc.b	$C4,$3F,$B2,$FA,$57,$A1,$4F,$DA,$FD,$43,$EA,$31,$7F,$EA,$FA,$DB
	.dc.b	$7D,$DC,$E5,$FE,$0A,$8F,$F0,$44,$A7,$B1,$FC,$03,$EA,$1C,$D8,$AF
	.dc.b	$F5,$D4,$61,$ED,$C1,$FC,$1B,$A5,$E6,$79,$E3,$0C,$9F,$F6,$14,$FF
	.dc.b	$96,$9B,$1C,$7D,$1F,$E1,$FF,$51,$FE,$66,$9F,$F6,$47,$BE,$F7,$33
	.dc.b	$EA,$FF,$14,$8C,$4F,$FB,$37,$EA,$3F,$D3,$AD,$F6,$FF,$8D,$FD,$43
	.dc.b	$EA,$1F,$51,$FC,$72,$F7,$C9,$F0,$CF,$C4,$38,$E9,$FF,$7A,$79,$4F
	.dc.b	$A8,$FF,$39,$E7,$D4,$FF,$71,$46,$7D,$AE,$9A,$7F,$2A,$8C,$4E,$F1
	.dc.b	$ED,$18,$7C,$1D,$FF,$97,$FA,$8F,$E6,$2F,$E6,$7B,$EC,$A6,$C6,$1B
	.dc.b	$7F,$EF,$3A,$E4,$DA,$FC,$A3,$FA,$38,$FD,$27,$D4,$74,$3A,$F2,$D9
	.dc.b	$E1,$16,$BC,$A7,$D7,$2B,$FD,$13,$E9,$9E,$C5,$3F,$99,$FD,$47,$D4
	.dc.b	$AF,$98,$27,$77,$FB,$5E,$03,$FE,$59,$8C,$CF,$F3,$4A,$FE,$35,$39
	.dc.b	$BE,$BE,$67,$78,$FE,$7B,$3A,$AF,$E0,$17,$CB,$0F,$E7,$53,$9C,$FF
	.dc.b	$C4,$17,$F3,$BD,$6F,$C7,$61,$F4,$9F,$1C,$7C,$B7,$37,$E6,$95,$72
	.dc.b	$AD,$74,$73,$7A,$3D,$7E,$A2,$47,$3F,$8C,$B7,$DE,$6E,$1D,$70,$FF
	.dc.b	$1A,$56,$2F,$E2,$7C,$7F,$CF,$3F,$19,$F0,$6F,$96,$C4,$FC,$B3,$67
	.dc.b	$F7,$3F,$A8,$FA,$58,$FA,$73,$FA,$58,$7F,$6F,$9C,$D7,$09,$FE,$EB
	.dc.b	$18,$7E,$5C,$EF,$BF,$E4,$7C,$74,$BF,$8D,$F5,$16,$FE,$54,$DC,$7D
	.dc.b	$9C,$0A,$C5,$FE,$4E,$1F,$2F,$F4,$FD,$9F,$A8,$D9,$FA,$4D,$C7,$EB
	.dc.b	$37,$0F,$67,$D4,$7F,$BE,$2B,$E1,$57,$F4,$18,$C3,$FB,$0D,$FF,$91
	.dc.b	$BF,$94,$8C,$D7,$2C,$3F,$C1,$7F,$50,$F9,$6F,$F1,$0D,$6B,$1F,$E9
	.dc.b	$9E,$4A,$B5,$1E,$0C,$55,$B7,$FB,$FC,$E9,$F0,$61,$DB,$0F,$C9,$7D
	.dc.b	$47,$E4,$FE,$A2,$E3,$F9,$16,$DF,$E5,$5B,$9F,$CC,$7F,$35,$C3,$F9
	.dc.b	$BE,$BB,$7F,$99,$6D,$FE,$75,$B7,$F5,$1F,$F1,$EF,$D4,$7C,$4F,$84
	.dc.b	$F8,$15,$FD,$A6,$1F,$21,$FA,$19,$CC,$A7,$86,$9F,$ED,$17,$78,$FF
	.dc.b	$9E,$A3,$1E,$B7,$17,$58,$5B,$ED,$35,$E9,$16,$F9,$C7,$FA,$3A,$B8
	.dc.b	$DF,$58,$E2,$D7,$8B,$67,$8E,$1E,$85,$F5,$1D,$70,$EC,$F7,$F3,$BC
	.dc.b	$5F,$4C,$ED,$FF,$1B,$E5,$E1,$3E,$24,$EA,$33,$8F,$58,$C7,$2F,$85
	.dc.b	$B5,$C4,$5B,$EA,$3F,$B6,$C7,$FA,$E2,$31,$EA,$FF,$AD,$1F,$15,$DA
	.dc.b	$DD,$7F,$B0,$27,$6F,$56,$E1,$F5,$9B,$B7,$F8,$8E,$5F,$4C,$FA,$8F
	.dc.b	$F4,$DC,$E6,$B7,$5F,$E2,$96,$BE,$8C,$3C,$F8,$7E,$0B,$0E,$53,$F4
	.dc.b	$6B,$64,$AD,$DF,$86,$B7,$5E,$79,$CB,$DD,$AF,$7C,$BE,$E4,$F0,$C3
	.dc.b	$D9,$5A,$C3,$B5,$6B,$1F,$5E,$1C,$E3,$C9,$5B,$3A,$E1,$CF,$1F,$81
	.dc.b	$19,$9D,$9C,$EA,$9E,$76,$E3,$F0,$70,$EF,$3E,$73,$44,$8D,$3C,$2F
	.dc.b	$BD,$58,$FF,$2B,$61,$E3,$7D,$07,$19,$DB,$F8,$98,$D3,$DE,$BE,$1D
	.dc.b	$73,$8F,$BB,$7E,$4C,$7E,$2C,$73,$3B,$C6,$71,$F8,$11,$93,$89,$CE
	.dc.b	$78,$DF,$AA,$35,$18,$C7,$EE,$5F,$3C,$38,$61,$F3,$A3,$17,$C2,$30
	.dc.b	$EA,$7D,$18,$FB,$30,$E9,$3F,$0B,$1F,$01,$EE,$79,$57,$D5,$69,$E9
	.dc.b	$AE,$57,$EC,$AE,$07,$2C,$7E,$5B,$86,$9E,$AA,$F2,$1C,$F1,$F4,$CE
	.dc.b	$E3,$C9,$3B,$7C,$31,$97,$E4,$98,$8E,$AF,$D0,$3A,$C6,$EF,$CF,$86
	.dc.b	$2B,$A4,$F3,$C7,$E1,$DF,$48,$C6,$3E,$DA,$D9,$DA,$7F,$08,$DC,$3A
	.dc.b	$C7,$8C,$F1,$14,$D9,$8A,$EB,$87,$D4,$C7,$E0,$DF,$5C,$7F,$81,$E1
	.dc.b	$F9,$23,$AD,$FA,$EB,$F4,$71,$96,$98,$FA,$8C,$FC,$3C,$CB,$F7,$AF
	.dc.b	$EA,$BE,$A1,$F0,$D9,$9F,$D0,$4D,$C7,$FA,$8D,$6A,$7F,$A0,$35,$D9
	.dc.b	$D0,$CD,$75,$3F,$91,$73,$FF,$88,$E7,$83,$63,$3C,$BF,$C9,$E5,$F0
	.dc.b	$B8,$BF,$F6,$67,$C0,$BC,$5C,$42,$1F,$51,$86,$C7,$F8,$15,$E0,$D9
	.dc.b	$D4,$C8,$C1,$BF,$73,$97,$D0,$C8,$C1,$0F,$F3,$6D,$A3,$51,$89,$F2
	.dc.b	$FD,$3B,$EA,$3C,$9E,$94,$AF,$F0,$C4,$E5,$B3,$C0,$8F,$A7,$6A,$FF
	.dc.b	$9F,$27,$83,$67,$E0,$4F,$C4,$FA,$87,$5D,$F8,$39,$7F,$E2,$79,$CB
	.dc.b	$67,$2F,$FA,$22,$3F,$DA,$8E,$4E,$1F,$29,$D4,$8F,$3F,$89,$AF,$18
	.dc.b	$BC,$9F,$E6,$1A,$57,$D8,$FA,$8F,$F2,$8D,$F8,$FD,$2F,$41,$9F,$3A
	.dc.b	$47,$F8,$F6,$76,$FF,$A7,$1B,$87,$5E,$FD,$CE,$6A,$CE,$63,$FC,$8D
	.dc.b	$F5,$0F,$A8,$FC,$4B,$63,$31,$8F,$9A,$CB,$23,$27,$28,$E3,$F3,$DC
	.dc.b	$BE,$96,$4F,$68,$CB,$E0,$8C,$BF,$EB,$4B,$4F,$E6,$BC,$7F,$F1,$EE
	.dc.b	$3F,$E8,$7F,$A7,$7D,$47,$E7,$9B,$87,$F9,$7E,$B7,$E0,$7F,$83,$5B
	.dc.b	$03,$FC,$55,$E3,$F5,$39,$77,$D8,$E5,$15,$FF,$63,$DF,$E9,$9B,$D9
	.dc.b	$E1,$23,$FE,$6F,$8F,$D0,$36,$32,$F0,$35,$DB,$E9,$EA,$F5,$78,$12
	.dc.b	$79,$46,$5E,$91,$97,$C7,$1F,$17,$AF,$F7,$FE,$C3,$EA,$29,$9A,$F8
	.dc.b	$93,$5F,$F7,$0E,$BF,$9A,$7F,$DC,$75,$FA,$07,$33,$BE,$0F,$F5,$B5
	.dc.b	$6D,$FF,$5C,$56,$DF,$F5,$D7,$D4,$39,$3F,$88,$B7,$0F,$F4,$93,$5F
	.dc.b	$F7,$8A,$FA,$8D,$FA,$72,$76,$FF,$1A,$FA,$7F,$D2,$C7,$FA,$6D,$AF
	.dc.b	$FB,$E9,$B8,$7F,$B2,$DB,$C9,$FE,$4D,$FA,$8F,$B1,$39,$F8,$9E,$84
	.dc.b	$BF,$F5,$37,$D4,$7F,$30,$BF,$F2,$A3,$63,$F1,$0D,$4D,$FA,$87,$C9
	.dc.b	$FF,$30,$6C,$F4,$E3,$FE,$67,$5B,$AF,$E6,$BE,$C3,$BB,$87,$AF,$30
	.dc.b	$FF,$5C,$4E,$5B,$3C,$08,$FF,$AF,$71,$FF,$38,$E9,$FE,$EB,$AE,$12
	.dc.b	$8C,$57,$08,$C6,$7F,$E7,$1B,$E2,$FF,$9C,$FD,$FF,$4C,$E8,$FD,$47
	.dc.b	$D2,$3D,$EF,$E9,$AD,$8F,$F6,$74,$F0,$C7,$FD,$F8,$DB,$E9,$A7,$F5
	.dc.b	$48,$C5,$74,$F8,$6C,$4C,$8E,$6F,$68,$FE,$B7,$E1,$3E,$3B,$FA,$EF
	.dc.b	$D3,$3E,$32,$9A,$89,$87,$FB,$86,$B8,$36,$38,$FC,$76,$A3,$DC,$7F
	.dc.b	$10,$8C,$68,$D3,$FD,$2D,$8A,$FF,$B6,$FD,$3B,$E8,$BC,$7F,$51,$32
	.dc.b	$FF,$78,$4E,$5B,$3C,$08,$FF,$BD,$6D,$E2,$E2,$3C,$75,$C5,$B5,$F0
	.dc.b	$27,$FB,$EF,$D4,$7D,$EC,$FC,$D5,$FC,$77,$0F,$F0,$03,$D7,$D8,$DF
	.dc.b	$34,$8F,$E4,$55,$C1,$5F,$3D,$7F,$D6,$8D,$8E,$9F,$3F,$C5,$F8,$F1
	.dc.b	$97,$8D,$FD,$49,$F2,$FC,$37,$D6,$F6,$A5,$78,$30,$EB,$3A,$8C,$1F
	.dc.b	$CD,$70,$46,$63,$EA,$E7,$FB,$1F,$A8,$74,$7F,$87,$9B,$3C,$83,$87
	.dc.b	$EE,$FD,$43,$D9,$FD,$24,$3E,$A1,$F5,$1F,$ED,$3B,$E1,$C1,$F1,$5B
	.dc.b	$AE,$72,$FF,$16,$E5,$CE,$BF,$A6,$61,$CB,$1F,$E9,$D3,$97,$B6,$3C
	.dc.b	$5D,$B5,$C9,$C3,$2F,$2D,$73,$85,$73,$8C,$65,$ED,$6E,$3F,$EE,$B7
	.dc.b	$84,$EF,$0F,$F7,$73,$6F,$FB,$15,$F1,$9F,$15,$70,$C3,$C3,$1C,$1B
	.dc.b	$5C,$30,$ED,$3C,$0D,$63,$EC,$9F,$53,$5F,$A7,$3C,$1E,$46,$71,$F2
	.dc.b	$3E,$EB,$EE,$F0,$39,$63,$E7,$7A,$9F,$20,$E1,$7A,$BE,$D1,$C4,$D5
	.dc.b	$F6,$3E,$A3,$FE,$03,$39,$94,$E6,$52,$99,$7F,$C6,$FF,$1F,$98,$FA
	.dc.b	$BE,$DA,$31,$3A,$9E,$4D,$78,$7C,$17,$17,$FC,$28,$D4,$D2,$3E,$FF
	.dc.b	$98,$7D,$37,$F0,$0D,$BF,$0D,$3C,$8D,$46,$76,$F5,$5F,$D3,$BE,$3B
	.dc.b	$FD,$5B,$E1,$87,$E3,$E7,$31,$F3,$4C,$DF,$C7,$8C,$CF,$82,$7C,$E7
	.dc.b	$82,$BA,$38,$8E,$98,$78,$E3,$E7,$61,$E3,$8C,$47,$D3,$8C,$57,$26
	.dc.b	$C7,$8C,$E9,$87,$8E,$B8,$C7,$98,$FF,$1F,$61,$E6,$9F,$83,$1E,$53
	.dc.b	$95,$74,$C3,$CB,$87,$F0,$9B,$F8,$3A,$FF,$25,$BE,$82,$33,$3E,$13
	.dc.b	$87,$8F,$FC,$CA,$BF,$96,$4E,$AB,$C7,$39,$AF,$95,$D3,$67,$FB,$B7
	.dc.b	$D4,$7F,$4D,$9F,$14,$E7,$04,$E6,$BF,$AB,$E1,$FD,$46,$FF,$94,$D6
	.dc.b	$BB,$7A,$57,$9F,$0E,$58,$71,$AF,$2D,$6B,$0C,$61,$FE,$65,$F8,$CF
	.dc.b	$F5,$2F,$8B,$9A,$93,$FC,$DA,$5F,$50,$F8,$8C,$E9,$FE,$07,$9C,$C6
	.dc.b	$32,$FE,$AB,$18,$BF,$9D,$87,$C1,$8C,$5F,$F7,$2F,$80,$FE,$83,$5F
	.dc.b	$CE,$DB,$1F,$E1,$38,$7D,$3A,$BD,$F5,$AC,$3F,$AE,$36,$BF,$D0,$07
	.dc.b	$0E,$3D,$F5,$FD,$18,$C1,$DD,$C6,$1F,$67,$15,$F3,$7F,$C0,$91,$9F
	.dc.b	$A8,$FF,$79,$DF,$F5,$16,$CF,$9E,$5C,$5F,$52,$73,$97,$F7,$16,$DF
	.dc.b	$9A,$AD,$FF,$A5,$E6,$CF,$C0,$F6,$B9,$1E,$62,$BF,$DA,$8D,$1F,$D8
	.dc.b	$CF,$F6,$69,$7A,$7D,$BE,$07,$E1,$A3,$37,$FE,$22,$6E,$3F,$E0,$C5
	.dc.b	$F5,$1F,$E5,$E7,$91,$DD,$C4,$F1,$BF,$99,$18,$AF,$F0,$75,$74,$3E
	.dc.b	$9C,$65,$FE,$65,$F3,$3F,$92,$BF,$F5,$B3,$63,$EF,$CF,$E7,$23,$52
	.dc.b	$75,$9B,$C4,$D7,$17,$DA,$6E,$1F,$16,$74,$FF,$81,$EB,$8C,$FE,$A3
	.dc.b	$E9,$9E,$85,$3F,$57,$F5,$0F,$A8,$C6,$1F,$E0,$FA,$DB,$7D,$EC,$E9
	.dc.b	$FE,$4C,$94,$75,$3F,$6B,$F5,$1F,$8C,$AE,$07,$F8,$65,$C5,$71,$F4
	.dc.b	$BF,$77,$D3,$FE,$46,$AF,$04,$63,$CC,$91,$FE,$21,$9F,$F3,$23,$63
	.dc.b	$8F,$A3,$FC,$13,$EA,$1E,$9F,$F0,$6F,$8D,$7B,$99,$8C,$3F,$C2,$BE
	.dc.b	$95,$ED,$D3,$F8,$C7,$D4,$3E,$A1,$F5,$19,$D7,$FC,$65,$7C,$1F,$F1
	.dc.b	$EF,$90,$FA,$8F,$F3,$D4,$6B,$1F,$E4,$53,$9F,$6B,$97,$1F,$FA,$46
	.dc.b	$31,$5B,$39,$DF,$29,$FF,$4B,$75,$FF,$BF,$7E,$A3,$F9,$63,$FE,$FF
	.dc.b	$EE,$B2,$9B,$18,$6D,$FF,$94,$AB,$93,$6B,$F2,$90,$5F,$51,$D0,$EB
	.dc.b	$C9,$6B,$84,$59,$F2,$9F,$50,$AF,$F3,$6F,$A8,$7B,$54,$FE,$49,$F5
	.dc.b	$1F,$52,$FF,$CB,$93,$A6,$CE,$BC,$BE,$11,$FC,$D6,$BF,$8B,$60,$BE
	.dc.b	$0F,$F3,$58,$CC,$63,$C5,$FE,$6B,$1F,$C7,$73,$FF,$0D,$5F,$CA,$F5
	.dc.b	$BF,$0E,$FC,$57,$FD,$48,$DE,$8D,$3B,$6B,$93,$87,$D4,$73,$3A,$7C
	.dc.b	$5D,$8C,$93,$3F,$B7,$69,$D4,$7F,$AD,$B3,$7C,$87,$C9,$3F,$13,$F0
	.dc.b	$6F,$96,$C6,$3F,$22,$7F,$B5,$7D,$47,$CB,$C3,$94,$63,$BB,$71,$FE
	.dc.b	$DF,$E0,$56,$DF,$F5,$36,$5E,$0C,$3F,$B7,$CE,$5F,$E2,$FA,$3E,$3B
	.dc.b	$E4,$1F,$27,$EA,$2C,$FC,$98,$B8,$FB,$A6,$E3,$FE,$BA,$E1,$FF,$25
	.dc.b	$7C,$BF,$D1,$F6,$7E,$93,$71,$FA,$4D,$C3,$E3,$B6,$FF,$D7,$9F,$51
	.dc.b	$FE,$F7,$AF,$F5,$23,$6B,$F5,$9C,$3F,$C6,$1F,$F7,$0C,$66,$B8,$E1
	.dc.b	$FD,$D7,$EA,$1F,$2D,$FE,$0C,$AD,$63,$FE,$07,$9C,$BD,$AB,$85,$6D
	.dc.b	$B1,$F8,$9F,$A8,$FC,$67,$D4,$7E,$35,$B1,$93,$F1,$D1,$96,$9F,$91
	.dc.b	$E3,$FC,$A7,$57,$E5,$5B,$19,$3F,$2D,$19,$69,$FE,$C5,$FA,$87,$C8
	.dc.b	$7F,$BA,$3E,$13,$DB,$BF,$EF,$67,$35,$FE,$74,$D3,$FD,$0D,$1F,$1E
	.dc.b	$B7,$48,$DB,$FE,$78,$9D,$DF,$57,$8F,$D4,$59,$ED,$35,$E5,$16,$8D
	.dc.b	$31,$37,$0F,$25,$79,$E7,$0F,$84,$EB,$1A,$7A,$B7,$E9,$FF,$59,$E8
	.dc.b	$5E,$DF,$F1,$55,$71,$3D,$9E,$77,$96,$B7,$19,$5F,$51,$FE,$24,$8F
	.dc.b	$F5,$34,$63,$E0,$7F,$D4,$8F,$08,$FF,$56,$39,$7F,$D5,$9F,$50,$FA
	.dc.b	$8F,$F4,$06,$FF,$EB,$3A,$D7,$07,$CE,$6C,$7F,$AD,$EB,$85,$FB,$5A
	.dc.b	$F9,$71,$F2,$5F,$16,$D7,$1C,$3B,$D7,$F3,$BB,$E3,$7B,$C7,$E3,$61
	.dc.b	$C2,$B5,$7C,$15,$F0,$C3,$D1,$5B,$96,$3E,$87,$89,$E7,$BE,$51,$98
	.dc.b	$DC,$F1,$BE,$91,$C0,$F6,$61,$E4,$39,$30,$C9,$31,$FC,$AC,$F2,$8F
	.dc.b	$3E,$1E,$0B,$F3,$61,$E1,$AD,$1B,$7C,$86,$67,$E0,$46,$67,$89,$E7
	.dc.b	$8C,$9D,$70,$F3,$5F,$E2,$EB,$AC,$E6,$3C,$D8,$FD,$37,$D5,$8F,$D1
	.dc.b	$72,$7C,$17,$D7,$3C,$5D,$18,$8E,$33,$F4,$AF,$C3,$39,$9F,$15,$71
	.dc.b	$AD,$46,$0E,$27,$F7,$DC,$7E,$0E,$3E,$D8,$C9,$F5,$B0,$EB,$5E,$98
	.dc.b	$E4,$7A,$0D,$18,$C3,$C4,$7A,$23,$C2,$E1,$E4,$E6,$3C,$55,$E1,$79
	.dc.b	$4F,$02,$3F,$15,$E2,$61,$F0,$CF,$57,$47,$4A,$F4,$B8,$3F,$0A,$60
	.dc.b	$E3,$3D,$1D,$5F,$C4,$9F,$87,$8F,$E0,$DE,$0F,$DE,$9E,$53,$F8,$97
	.dc.b	$E4,$BC,$5F,$59,$D3,$0F,$C5,$D7,$4C,$7E,$B4,$F9,$CD,$9E,$63,$E7
	.dc.b	$57,$57,$2F,$5C,$7E,$04,$6D,$F9,$8F,$B0,$F8,$18,$7A,$5F,$B0,$FA
	.dc.b	$5E,$B8,$7D,$19,$E7,$5D,$71,$F8,$95,$D5,$FA,$11,$E3,$9F,$D8,$FD
	.dc.b	$47,$E9,$2B,$E3,$BF,$AA,$30,$FB,$30,$47,$F0,$DF,$A8,$FF,$63,$CF
	.dc.b	$F1,$DF,$87,$FD,$89,$9F,$86,$91,$FE,$53,$EC,$71,$7F,$95,$61,$7C
	.dc.b	$94,$E9,$FF,$2D,$46,$63,$11,$C3,$E1,$79,$46,$25,$18,$AE,$D1,$C5
	.dc.b	$B7,$9B,$EB,$55,$E8,$79,$E7,$6D,$3C,$73,$BA,$EF,$19,$3F,$D7,$9E
	.dc.b	$37,$3F,$81,$7D,$45,$31,$12,$78,$4E,$9B,$39,$7E,$1E,$1F,$1E,$5E
	.dc.b	$DF,$5F,$A0,$95,$CA,$3C,$06,$1E,$27,$FA,$0E,$53,$A8,$F4,$7D,$43
	.dc.b	$EA,$3D,$1E,$8C,$9C,$C7,$C5,$9D,$46,$08,$76,$3F,$DB,$9A,$EE,$64
	.dc.b	$6B,$7F,$B3,$8F,$D3,$8F,$F4,$B7,$D4,$3A,$BE,$93,$FE,$D2,$8F,$98
	.dc.b	$D7,$B3,$B7,$31,$C3,$E0,$A4,$60,$E3,$F2,$BD,$08,$F4,$F8,$AA,$F4
	.dc.b	$8B,$8F,$E3,$B0,$57,$E1,$7E,$A1,$E6,$79,$BD,$39,$8F,$E0,$67,$2D
	.dc.b	$9E,$04,$7D,$7E,$05,$7E,$44,$F0,$6C,$FC,$29,$FC,$A7,$D3,$BE,$A2
	.dc.b	$92,$BF,$21,$0F,$9B,$C4,$7D,$38,$E0,$EA,$79,$CF,$4F,$9E,$EB,$F4
	.dc.b	$B2,$B9,$C6,$5E,$B1,$97,$A4,$7E,$9F,$C6,$FF,$26,$EF,$FB,$4F,$A7
	.dc.b	$7D,$47,$EA,$DB,$7F,$A5,$AD,$F6,$7F,$01,$6C,$7F,$B5,$DA,$FF,$AA
	.dc.b	$FC,$6E,$5D,$FD,$D9,$1C,$DC,$BF,$B1,$BF,$D2,$B7,$BB,$C4,$78,$FB
	.dc.b	$AC,$64,$87,$E6,$67,$F4,$AF,$3E,$CF,$57,$83,$23,$04,$38,$39,$79
	.dc.b	$47,$C5,$EA,$FE,$19,$EB,$BE,$A2,$99,$AF,$89,$35,$FD,$56,$BF,$9F
	.dc.b	$E0,$FE,$3F,$7F,$9F,$9C,$D6,$E7,$F9,$1D,$6F,$7F,$E5,$97,$B9,$FE
	.dc.b	$4F,$F5,$0E,$3F,$E3,$CD,$C7,$F8,$C3,$5F,$DD,$2F,$A8,$DF,$A5,$27
	.dc.b	$6F,$F2,$7F,$A8,$FD,$2C,$7F,$21,$6B,$FC,$0D,$B8,$7F,$38,$6E,$5F
	.dc.b	$D0,$7E,$A3,$FD,$CB,$5C,$7E,$2B,$A1,$1F,$E5,$9F,$51,$FC,$1D,$34
	.dc.b	$9C,$3F,$CE,$7E,$A1,$F2,$54,$FE,$79,$3A,$AF,$E6,$D3,$A7,$F9,$C7
	.dc.b	$CB,$FE,$A5,$CF,$AF,$CC,$FE,$89,$19,$8F,$E3,$CD,$D7,$FA,$9B,$74
	.dc.b	$FE,$B8,$D9,$0B,$FF,$49,$3C,$BB,$BA,$91,$F2,$F8,$3F,$D4,$FD,$FF
	.dc.b	$34,$7F,$2D,$FA,$37,$BD,$FD,$5D,$B1,$FD,$46,$77,$9F,$F6,$66,$E7
	.dc.b	$FD,$BD,$B3,$E4,$6A,$A7,$C3,$F8,$77,$93,$D5,$F0,$47,$F6,$7F,$0D
	.dc.b	$F1,$FF,$DA,$7E,$95,$F1,$94,$D4,$4C,$3F,$B3,$57,$06,$C7,$1F,$8E
	.dc.b	$D4,$7F,$4B,$9D,$36,$74,$FF,$6F,$FA,$77,$D1,$78,$7E,$A2,$4E,$63
	.dc.b	$FB,$E4,$EA,$30,$43,$FA,$C6,$5F,$C2,$1F,$EB,$51,$B6,$CF,$23,$DD
	.dc.b	$E3,$FF,$81,$FE,$A3,$EF,$4A,$DE,$AF,$C3,$35,$CD,$27,$FC,$0D,$5C
	.dc.b	$0E,$34,$FA,$0D,$4F,$F6,$B9,$D5,$7C,$D9,$F0,$7C,$37,$CF,$F6,$A5
	.dc.b	$F4,$8C,$B4,$87,$BB,$BB,$FC,$3D,$F5,$1F,$7E,$3F,$C3,$0D,$7F,$BD
	.dc.b	$37,$A7,$FC,$5F,$F5,$0F,$67,$FC,$6B,$4F,$A7,$7D,$47,$F8,$EF,$0E
	.dc.b	$1D,$B7,$B3,$B9,$FE,$2D,$C7,$96,$5F,$A4,$C3,$8E,$3F,$A6,$9C,$BD
	.dc.b	$71,$E4,$ED,$AE,$4E,$19,$79,$2B,$9C,$2B,$9C,$63,$2F,$73,$71,$FF
	.dc.b	$29,$3C,$27,$78,$7F,$95,$5B,$7F,$BA,$BE,$33,$E2,$AE,$18,$78,$63
	.dc.b	$83,$6B,$86,$1D,$A7,$81,$AC,$7D,$93,$EA,$6A,$C3,$BB,$C8,$CE,$3E
	.dc.b	$47,$DD,$87,$77,$86,$1D,$DE,$07,$C7,$37,$7C,$A7,$85,$F6,$9E,$26
	.dc.b	$66,$3F,$E7,$73,$CC,$53,$99,$6F,$EC,$BB,$7F,$C5,$3F,$23,$B8,$FA
	.dc.b	$9E,$DA,$31,$3A,$9E,$53,$97,$80,$F8,$5A,$D3,$2E,$08,$FB,$FE,$65
	.dc.b	$F4,$DF,$E8,$BD,$BF,$0B,$3F,$CD,$2B,$E7,$D6,$B7,$FA,$33,$93,$95
	.dc.b	$F1,$DF,$E0,$5F,$08,$C4,$FC,$58,$C5,$72,$8F,$36,$1F,$94,$C3,$B4
	.dc.b	$7D,$33,$B4,$66,$F9,$C7,$FA,$96,$F9,$BB,$39,$E1,$FE,$A9,$BE,$77
	.dc.b	$EC,$9C,$9F,$D4,$23,$33,$E8,$8F,$5E,$1E,$5B,$EB,$7C,$5B,$5B,$8F
	.dc.b	$24,$F8,$B4,$FF,$58,$E1,$EA,$7F,$D6,$78,$7A,$AF,$C9,$8F,$FA,$E6
	.dc.b	$76,$DB,$F4,$70,$5B,$F3,$4D,$58,$27,$0F,$F3,$EA,$FF,$59,$5B,$EA
	.dc.b	$1D,$FA,$EB,$E9,$39,$8C,$57,$0A,$DB,$FE,$B7,$BE,$98,$FF,$A9,$AF
	.dc.b	$84,$E6,$FC,$27,$9F,$E3,$3F,$D2,$DE,$2E,$27,$47,$FB,$16,$1F,$50
	.dc.b	$F8,$8D,$63,$FE,$B2,$9C,$C6,$31,$FF,$6F,$46,$2F,$9D,$6E,$F9,$47
	.dc.b	$FB,$37,$C8,$74,$FC,$76,$C7,$FB,$54,$7D,$33,$0D,$54,$9D,$4F,$D9
	.dc.b	$7F,$DE,$55,$CB,$0F,$F6,$3C,$72,$E1,$E1,$18,$AE,$0F,$F8,$72,$B8
	.dc.b	$65,$FE,$F7,$BE,$07,$FB,$8F,$EA,$3F,$DE,$6D,$F6,$BF,$00,$F9,$B8
	.dc.b	$FF,$01,$3A,$C3,$FD,$94,$DB,$E9,$8D,$BF,$F1,$5D,$2B,$0A,$3E,$0E
	.dc.b	$1D,$0E,$F5,$A9,$E1,$18,$3F,$21,$3F,$EB,$E3,$FD,$F4,$58,$F8,$1D
	.dc.b	$DA,$E8,$93,$F3,$FE,$A3,$FC,$9E,$F0,$8F,$6C,$62,$BF,$DC,$93,$CF
	.dc.b	$2F,$8D,$39,$7D,$13,$FE,$46,$9D,$16,$74,$FC,$0F,$97,$FC,$CC,$FF
	.dc.b	$BB,$DB,$5F,$1A,$BF,$55,$18,$AE,$1C,$9F,$0B,$C4,$CB,$63,$FC,$AB
	.dc.b	$7E,$2C,$17,$C5,$F9,$9A,$7F,$87,$2B,$67,$43,$F1,$78,$7E,$E7,$E9
	.dc.b	$9E,$85,$3F,$79,$F5,$0F,$A8,$C5,$F2,$C7,$FC,$CB,$7C,$23,$10,$87
	.dc.b	$BD,$C2,$1A,$79,$E7,$35,$C1,$B3,$A8,$FC,$3F,$D4,$39,$7C,$D8,$7F
	.dc.b	$0F,$8C,$4F,$1F,$4B,$F8,$97,$4B,$CC,$F6,$46,$19,$3E,$69,$FF,$2F
	.dc.b	$B6,$38,$FA,$3F,$C6,$FE,$A3,$FC,$D7,$3E,$CF,$8D,$7B,$99,$F5,$7F
	.dc.b	$90,$46,$27,$E0,$FD,$2B,$DB,$FE,$48,$D8,$F9,$1F,$50,$FA,$87,$D4
	.dc.b	$6A,$E5,$7C,$DF,$21,$F5,$1F,$E7,$68,$EB,$3F,$CC,$67,$3E,$D7,$0E
	.dc.b	$6F,$90,$7F,$32,$BE,$0E,$06,$7E,$37,$75,$CB,$0F,$B9,$DD,$FE,$A0
	.dc.b	$FA,$8F,$E7,$CF,$FA,$8B,$C0,$B6,$9B,$18,$6D,$72,$9D,$36,$78,$FD
	.dc.b	$44,$8D,$72,$5A,$F4,$E1,$FE,$B0,$AC,$76,$7E,$BA,$3F,$D6,$75,$8B
	.dc.b	$FE,$17,$F4,$CF,$63,$FA,$93,$63,$F9,$E7,$D4,$7D,$5C,$BE,$14,$E5
	.dc.b	$B5,$FD,$16,$7F,$91,$D7,$48,$C7,$99,$D1,$B1,$F9,$EB,$E5,$97,$F4
	.dc.b	$2B,$E3,$87,$EC,$BD,$6F,$E3,$97,$F3,$1F,$C5,$C7,$B9,$CB,$7A,$3E
	.dc.b	$7C,$62,$FE,$03,$C1,$B5,$A9,$DF,$D4,$48,$ED,$F1,$77,$0E,$D3,$6F
	.dc.b	$9C,$DA,$F9,$3F,$21,$F2,$0F,$C6,$7C,$1B,$E5,$B1,$3E,$F9,$FE,$EB
	.dc.b	$F5,$1F,$2F,$0F,$E2,$1B,$27,$3D,$FC,$34,$FE,$E9,$7B,$C3,$B9,$FE
	.dc.b	$02,$71,$1F,$1F,$E3,$BE,$09,$F0,$BE,$A2,$D7,$C9,$8B,$3F,$26,$6E
	.dc.b	$5F,$1B,$E5,$FE,$4F,$B3,$F2,$9B,$87,$D1,$6D,$FD,$16,$D7,$F2,$9F
	.dc.b	$A8,$FF,$7A,$D7,$F1,$A6,$D7,$E9,$B7,$FE,$3C,$FE,$AE,$33,$5C,$70
	.dc.b	$FF,$07,$7D,$43,$E5,$BF,$C4,$95,$AC,$58,$7F,$7D,$AD,$47,$7B,$E3
	.dc.b	$7C,$1B,$1F,$89,$FA,$8F,$C2,$FD,$47,$E1,$9B,$19,$3F,$1D,$19,$69
	.dc.b	$F9,$1E,$0F,$C6,$F5,$7E,$39,$B1,$93,$F2,$D1,$96,$9F,$CF,$3E,$A1
	.dc.b	$F2,$1F,$E0,$7D,$3F,$BD,$C6,$27,$FB,$14,$62,$74,$FC,$09,$D3,$FE
	.dc.b	$50,$6F,$C2,$FC,$78,$D7,$D4,$7F,$6C,$8C,$42,$77,$AF,$82,$B4,$DC
	.dc.b	$3F,$D0,$8D,$AF,$F4,$83,$E3,$8C,$C7,$66,$BD,$67,$CE,$D8,$F5,$BB
	.dc.b	$3D,$11,$AF,$A6,$7D,$3F,$F5,$59,$FF,$56,$46,$3D,$5F,$F5,$42,$EC
	.dc.b	$FF,$1E,$4E,$AB,$83,$FE,$B5,$9E,$0F,$A0,$FC,$3B,$8C,$FE,$1B,$72
	.dc.b	$FF,$15,$FD,$3B,$EA,$3F,$D2,$11,$86,$DF,$2C,$3C,$D8,$72,$6D,$6E
	.dc.b	$7E,$1D,$F3,$7E,$2E,$1D,$DE,$59,$6B,$1E,$B0,$AE,$0F,$A4,$57,$16
	.dc.b	$E1,$E0,$7F,$47,$3C,$31,$FB,$ED,$C7,$FC,$8F,$7C,$4E,$51,$9A,$E3
	.dc.b	$87,$8A,$75,$87,$59,$F2,$4F,$9A,$B8,$47,$8B,$1F,$6D,$6E,$72,$6A
	.dc.b	$F9,$CF,$BE,$A1,$ED,$8C,$BE,$F9,$E3,$7E,$3B,$F1,$57,$A2,$3C,$46
	.dc.b	$6F,$C2,$74,$C7,$E0,$C6,$70,$ED,$7E,$58,$E3,$1C,$0C,$D7,$1C,$7E
	.dc.b	$D4,$60,$97,$D1,$D1,$F6,$8C,$C7,$08,$D4,$60,$D3,$82,$47,$D4,$BF
	.dc.b	$69,$E5,$9E,$A7,$CD,$39,$1F,$48,$CE,$1F,$22,$B9,$C7,$DA,$3F,$5E
	.dc.b	$6C,$FF,$6A,$E3,$F8,$CD,$3F,$D1,$31,$96,$C6,$63,$E0,$C6,$32,$FF
	.dc.b	$47,$B6,$3A,$CF,$D6,$7C,$EB,$EA,$3E,$6C,$7D,$BA,$B7,$F9,$49,$B9
	.dc.b	$7C,$72,$57,$E6,$A3,$2F,$AC,$E4,$7E,$B1,$CB,$ED,$9F,$82,$FE,$A6
	.dc.b	$33,$7E,$58,$FC,$64,$62,$F2,$D8,$F4,$B5,$EE,$70,$3D,$B1,$C1,$A7
	.dc.b	$02,$9C,$2B,$DC,$E1,$F2,$D7,$C5,$8C,$5F,$C0,$7D,$06,$0F,$81,$3C
	.dc.b	$9E,$91,$B9,$FD,$6D,$F9,$EB,$AB,$67,$4F,$ED,$CC,$57,$CD,$3E,$51
	.dc.b	$F6,$9F,$BA,$FF,$22,$6B,$B3,$FD,$48,$3E,$A1,$F0,$BF,$8A,$8F,$F6
	.dc.b	$9C,$61,$C1,$B8,$AD,$33,$1B,$7F,$CD,$8D,$E9,$FF,$59,$37,$3E,$BF
	.dc.b	$09,$FE,$E0,$9C,$B6,$3C,$37,$6F,$FC,$0B,$7C,$5B,$59,$7D,$4F,$92
	.dc.b	$0D,$24,$7A,$BE,$03,$FC,$68,$FF,$AD,$17,$D4,$73,$6C,$67,$C2,$C1
	.dc.b	$99,$91,$93,$97,$B3,$C7,$E8,$64,$60,$87,$FA,$3E,$D1,$A8,$C4,$FA
	.dc.b	$FE,$A1,$F5,$1E,$3F,$36,$57,$FB,$2A,$72,$D9,$E0,$47,$D9,$AD,$7F
	.dc.b	$D3,$D3,$C1,$B3,$EF,$9F,$95,$F5,$0F,$72,$FC,$0C,$BF,$F6,$CC,$E5
	.dc.b	$B3,$97,$FD,$55,$1F,$ED,$C7,$27,$0F,$94,$EA,$47,$9F,$C4,$D7,$8C
	.dc.b	$5E,$4F,$F7,$C5,$2B,$EF,$7D,$43,$CA,$F3,$F9,$F3,$1F,$F7,$6C,$E5
	.dc.b	$B3,$C0,$8F,$AB,$BE,$BF,$EC,$19,$E0,$D9,$F8,$73,$F8,$DF,$A8,$7D
	.dc.b	$45,$25,$7F,$BF,$21,$F3,$38,$8F,$F6,$64,$70,$75,$3C,$E7,$A7,$CF
	.dc.b	$75,$FA,$59,$5C,$E3,$2F,$58,$CB,$D2,$3F,$41,$E2,$FF,$90,$B7,$77
	.dc.b	$7F,$9A,$A9,$F4,$EF,$A8,$FD,$33,$63,$FC,$DF,$5C,$1F,$F3,$8D,$F3
	.dc.b	$BE,$4D,$8C,$F8,$5E,$03,$9F,$7D,$8E,$75,$C2,$71,$1F,$EE,$DC,$3F
	.dc.b	$5E,$DE,$CF,$19,$1E,$27,$FA,$0A,$3F,$4A,$E0,$F5,$C6,$DE,$46,$63
	.dc.b	$6F,$7E,$9F,$7F,$81,$23,$51,$F0,$8C,$3E,$58,$C3,$DF,$2F,$CD,$2C
	.dc.b	$FE,$32,$F5,$1F,$51,$FC,$45,$B1,$C7,$B3,$EA,$91,$C1,$EB,$8F,$D5
	.dc.b	$3D,$75,$F9,$F3,$B8,$C1,$F5,$1E,$DD,$9F,$C7,$FE,$A1,$C7,$49,$5F
	.dc.b	$1E,$1F,$51,$BF,$42,$4E,$5F,$E5,$1F,$51,$FA,$D8,$F9,$CD,$7C,$21
	.dc.b	$C6,$BF,$09,$B9,$FE,$3F,$EA,$3E,$EC,$E7,$E2,$7A,$92,$FE,$D7,$D4
	.dc.b	$7F,$39,$4D,$27,$1F,$F3,$BF,$A8,$7C,$8D,$3F,$9F,$4E,$AB,$F0,$93
	.dc.b	$A7,$F0,$DF,$2D,$FB,$27,$3E,$B7,$33,$FA,$34,$66,$3D,$AD,$CB,$EE
	.dc.b	$B7,$4F,$D5,$36,$7F,$DA,$B7,$FE,$94,$AD,$F7,$74,$24,$F9,$3C,$1F
	.dc.b	$EA,$9E,$EF,$96,$3E,$5F,$D2,$3D,$CF,$EB,$0D,$8F,$CF,$4E,$F0,$FC
	.dc.b	$7B,$73,$FD,$E3,$67,$FD,$CE,$D5,$4F,$83,$F0,$EF,$27,$AB,$E0,$8F
	.dc.b	$ED,$1E,$1B,$E3,$BF,$B5,$7D,$33,$E2,$29,$A8,$98,$7E,$CE,$B8,$36
	.dc.b	$38,$FC,$5E,$A3,$F1,$33,$A6,$CE,$9F,$DE,$7D,$3B,$E8,$BC,$3F,$51
	.dc.b	$27,$31,$FD,$F6,$75,$18,$21,$F9,$6C,$9E,$3E,$EA,$4E,$B8,$7F,$73
	.dc.b	$5A,$8F,$E1,$9F,$51,$F7,$54,$3D,$8F,$86,$7F,$83,$67,$44,$E6,$C1
	.dc.b	$F2,$3E,$7E,$DF,$48,$AF,$96,$39,$7C,$27,$CE,$F6,$64,$F8,$A3,$2D
	.dc.b	$21,$ED,$EE,$FF,$0F,$7D,$47,$DE,$8F,$E4,$8D,$7F,$6E,$DD,$3F,$8B
	.dc.b	$FD,$43,$D9,$FE,$35,$4F,$A8,$7D,$47,$F4,$7C,$38,$76,$DE,$CE,$E7
	.dc.b	$F8,$B7,$1E,$59,$31,$E5,$87,$F4,$F9,$CC,$60,$D6,$1E,$5C,$7B,$C6
	.dc.b	$30,$FE,$AD,$18,$85,$72,$8C,$36,$FC,$93,$E1,$9E,$17,$D6,$7C,$59
	.dc.b	$7B,$A7,$8B,$63,$8D,$79,$45,$71,$7D,$97,$DD,$B5,$C5,$F4,$1E,$1A
	.dc.b	$E7,$7D,$B1,$C4,$F0,$6B,$DB,$1F,$7D,$70,$3C,$F3,$AC,$3C,$15,$B3
	.dc.b	$8E,$3E,$57,$E2,$E3,$E3,$6D,$6B,$0E,$B1,$E6,$3D,$D7,$D0,$FA,$AF
	.dc.b	$FB,$19,$AF,$88,$FF,$40,$38,$7E,$1F,$B4,$F9,$1F,$F4,$0C,$FF,$8C
	.dc.b	$DB,$19,$9E,$2D,$7F,$D0,$FF,$07,$FE,$86,$4D,$23,$FE,$8A,$7E,$3F
	.dc.b	$CB,$3E,$97,$FC,$29,$B7,$E1,$6B,$F2,$15,$A9,$CE,$DF,$E9,$78,$C6
	.dc.b	$FF,$42,$B6,$7D,$FC,$3A,$46,$63,$C6,$66,$F9,$BF,$EE,$AA,$E6,$FE
	.dc.b	$1C,$E7,$87,$E5,$EF,$9C,$64,$E7,$19,$AE,$71,$FE,$33,$AE,$73,$EB
	.dc.b	$8D,$E1,$FE,$37,$BE,$73,$9A,$DC,$FB,$6B,$CD,$87,$A6,$73,$1E,$73
	.dc.b	$24,$D7,$FC,$2F,$87,$99,$FF,$5F,$C7,$80,$FD,$D4,$66,$BC,$58,$FF
	.dc.b	$90,$EB,$8F,$05,$9C,$9F,$E3,$3A,$FF,$63,$B7,$A3,$58,$7F,$89,$FE
	.dc.b	$A1,$DF,$88,$5F,$F3,$FB,$FF,$1F,$E2,$BD,$C7,$33,$2E,$2B,$FC,$27
	.dc.b	$5C,$5F,$F1,$2D,$71,$BE,$D3,$B8,$F8,$31,$E4,$F9,$FF,$F4,$E7,$8B
	.dc.b	$A8,$98,$FF,$94,$25,$F4,$EF,$A8,$CE,$5F,$E3,$29,$CC,$63,$1F,$F3
	.dc.b	$44,$62,$FE,$26,$1D,$F0,$E5,$1F,$E5,$9F,$21,$D3,$F0,$DB,$1F,$E6
	.dc.b	$01,$F4,$CC,$35,$52,$75,$3F,$5F,$3F,$E6,$57,$C9,$FF,$23,$CF,$2E
	.dc.b	$1E,$11,$8A,$E0,$AB,$86,$7F,$E8,$1B,$E0,$7F,$9B,$FE,$A3,$FD,$D2
	.dc.b	$DF,$6F,$EE,$BE,$6E,$3F,$B9,$3A,$C3,$FC,$8E,$DB,$E9,$8D,$BF,$F4
	.dc.b	$A4,$DA,$FF,$4B,$CD,$51,$99,$F8,$78,$76,$3C,$35,$B8,$FF,$62,$1F
	.dc.b	$E8,$52,$D7,$BF,$BB,$F2,$11,$8C,$3F,$CF,$2D,$3F,$AD,$0F,$A8,$FF
	.dc.b	$27,$BC,$FA,$7D,$F1,$8B,$FF,$31,$CF,$17,$FD,$5F,$19,$AE,$B5,$C3
	.dc.b	$CB,$FD,$07,$CE,$7E,$4F,$0F,$F5,$A3,$70,$F8,$38,$6B,$8E,$57,$DE
	.dc.b	$F1,$2D,$7D,$96,$E5,$F3,$1B,$3E,$AD,$3F,$C3,$0F,$FA,$9F,$E9,$9E
	.dc.b	$87,$EC,$5B,$1F,$EA,$BF,$A8,$7D,$46,$6A,$4F,$1F,$76,$E7,$FE,$82
	.dc.b	$C1,$1C,$CF,$DE,$7D,$47,$E2,$EB,$81,$FE,$B7,$71,$5C,$78,$3F,$81
	.dc.b	$74,$FF,$8D,$AB,$CD,$18,$F3,$24,$7F,$AF,$E7,$FC,$A4,$D8,$E3,$E8
	.dc.b	$FF,$0B,$FA,$8F,$F2,$F4,$FF,$B1,$7D,$F7,$B9,$9F,$57,$F8,$84,$62
	.dc.b	$7F,$D9,$5F,$48,$F6,$FF,$8A,$36,$3F,$DA,$3F,$50,$FA,$87,$D4,$6A
	.dc.b	$E5,$7F,$B5,$BC,$87,$D4,$7F,$9A,$63,$AC,$FF,$22,$9C,$FB,$5C,$38
	.dc.b	$FF,$D7,$71,$8A,$D9,$CE,$F9,$4F,$FB,$03,$BF,$FC,$5D,$F5,$1F,$CB
	.dc.b	$1F,$F1,$87,$7D,$94,$D8,$C3,$6F,$FD,$D7,$5C,$9B,$5F,$91,$8F,$F1
	.dc.b	$D7,$D4,$73,$3A,$72,$5A,$DC,$5A,$F2,$1F,$50,$AF,$F3,$DF,$A6,$7B
	.dc.b	$14,$FE,$61,$F5,$1F,$46,$BF,$D9,$73,$AB,$FF,$7E,$36,$3F,$97,$4F
	.dc.b	$F1,$49,$FF,$27,$42,$75,$E1,$FF,$30,$AD,$4F,$F9,$4E,$73,$5F,$E5
	.dc.b	$5E,$A5,$6A,$B9,$61,$E4,$F3,$3E,$67,$AA,$C2,$F4,$E1,$A7,$CA,$8D
	.dc.b	$36,$34,$D3,$5F,$51,$D0,$EB,$F1,$96,$32,$4C,$7E,$F5,$A7,$51,$F1
	.dc.b	$BE,$3F,$E6,$1F,$8B,$F8,$37,$CA,$E3,$0F,$93,$3F,$DC,$38,$1F,$51
	.dc.b	$F3,$DF,$66,$1F,$16,$75,$39,$AF,$9B,$C2,$9C,$F8,$1D,$6F,$FB,$6D
	.dc.b	$ED,$72,$FF,$90,$FE,$3B,$E4,$9F,$2B,$EA,$2E,$5F,$1E,$2D,$7B,$A6
	.dc.b	$E7,$E4,$BF,$F2,$67,$CB,$7D,$2F,$67,$E9,$B7,$1F,$A0,$DC,$3E,$3B
	.dc.b	$6F,$DB,$0F,$A8,$FF,$76,$57,$FA,$56,$73,$18,$9F,$D8,$F0,$7F,$22
	.dc.b	$53,$AC,$3F,$C2,$D5,$B8,$FE,$FB,$F5,$0F,$95,$CD,$7F,$82,$E3,$18
	.dc.b	$F7,$3F,$2F,$7C,$5F,$C9,$46,$B0,$EF,$1A,$AE,$57,$F9,$2F,$A8,$FC
	.dc.b	$9F,$D4,$5A,$FC,$83,$6F,$F0,$ED,$CB,$E5,$73,$23,$F3,$7C,$5F,$9C
	.dc.b	$E9,$B8,$7E,$65,$B8,$7E,$41,$B8,$72,$7F,$C6,$DF,$50,$F2,$BF,$C0
	.dc.b	$95,$FB,$28,$C5,$FF,$68,$8C,$4E,$9F,$F6,$24,$E9,$FA,$A1,$C7,$C1
	.dc.b	$7C,$B8,$FF,$65,$7D,$37,$E8,$7C,$0F,$F3,$EE,$FD,$31,$B5,$AE,$CF
	.dc.b	$F1,$23,$FE,$8A,$9F,$E0,$B7,$C6,$30,$6E,$F8,$BE,$1C,$BF,$47,$8F
	.dc.b	$86,$15,$C6,$30,$BE,$9F,$FC,$4D,$3F,$C4,$8F,$3B,$87,$11,$E1,$6C
	.dc.b	$65,$FE,$33,$F4,$CF,$9A,$FE,$37,$A7,$38,$C4,$FC,$B3,$FD,$D8,$E1
	.dc.b	$F9,$2D,$8E,$13,$FE,$8B,$CF,$AB,$6B,$8B,$55,$7F,$96,$DF,$86,$D7
	.dc.b	$E3,$9A,$8E,$11,$E1,$C7,$E3,$3B,$6F,$C2,$6A,$65,$74,$8E,$26,$1E
	.dc.b	$B5,$CE,$FE,$1C,$EA,$7C,$75,$BB,$EB,$1C,$CC,$90,$E7,$3C,$5E,$B3
	.dc.b	$C6,$BA,$C7,$13,$27,$1A,$EA,$F2,$31,$1F,$29,$E1,$1A,$3E,$0C,$F5
	.dc.b	$8F,$AB,$5E,$19,$CC,$72,$32,$72,$AE,$8F,$88,$F0,$12,$38,$B9,$7C
	.dc.b	$11,$EA,$8C,$3C,$A3,$31,$D8,$C9,$F8,$03,$C0,$65,$EA,$48,$E8,$F2
	.dc.b	$8F,$C0,$1E,$93,$F0,$F3,$ED,$9F,$01,$23,$A3,$C8,$C3,$C8,$F6,$3E
	.dc.b	$9C,$3F,$3D,$3F,$A0,$9C,$10,$F5,$46,$9F,$D1,$1E,$22,$E1,$F9,$78
	.dc.b	$D4,$76,$3C,$60,$7D,$2B,$E0,$BD,$51,$A3,$A1,$FA,$08,$D3,$F2,$0C
	.dc.b	$E3,$F1,$63,$31,$DC,$CE,$3F,$05,$F0,$46,$A7,$E6,$61,$FA,$7A,$D5
	.dc.b	$79,$2F,$C6,$F2,$3C,$97,$C6,$77,$5C,$67,$35,$C2,$31,$1C,$4E,$87
	.dc.b	$A5,$C9,$AA,$E9,$1F,$06,$33,$8F,$F0,$59,$CC,$F3,$9C,$BF,$CA,$23
	.dc.b	$31,$88,$F8,$51,$89,$F2,$1C,$EA,$CF,$CB,$9D,$3E,$59,$D3,$4D,$4F
	.dc.b	$07,$AB,$70,$FA,$4F,$08,$FA,$57,$ED,$8C,$47,$F3,$5B,$FA,$ED,$8F
	.dc.b	$E7,$0F,$F9,$2A,$32,$FF,$93,$3F,$10,$7C,$0E,$A3,$F1,$51,$87,$0D
	.dc.b	$79,$B5,$FE,$88,$E7,$93,$FC,$26,$DD,$3B,$57,$F9,$6A,$7F,$C2,$F7
	.dc.b	$FE,$66,$BF,$F0,$41,$F8,$AF,$1F,$57,$C2,$32,$30,$E8,$EC,$CD,$73
	.dc.b	$7D,$ED,$7E,$25,$7B,$5F,$D2,$B7,$DF,$FF,$20,$47,$65,$F4,$DE,$28
	.dc.b	$FD,$64,$61,$D7,$8B,$95,$1C,$9F,$F0,$E9,$C2,$75,$1C,$63,$27,$09
	.dc.b	$CF,$B1,$AF,$A6,$93,$E3,$8C,$BE,$58,$CB,$CA,$93,$E3,$FA,$67,$D3
	.dc.b	$78,$32,$97,$F2,$A7,$2D,$9E,$04,$7E,$17,$05,$7C,$F3,$C1,$B3,$EC
	.dc.b	$9F,$85,$F4,$DF,$E2,$79,$FA,$EF,$FB,$99,$BE,$FE,$BA,$A4,$60,$87
	.dc.b	$C4,$9E,$4D,$F9,$2E,$A4,$7A,$FC,$3D,$74,$78,$23,$C7,$1F,$57,$E9
	.dc.b	$9E,$57,$9B,$D0,$97,$F8,$C9,$CB,$67,$81,$1F,$5F,$C5,$D7,$EC,$CF
	.dc.b	$06,$CF,$C2,$9F,$C2,$7D,$33,$E9,$A9,$2B,$F3,$70,$F9,$7C,$47,$E1
	.dc.b	$23,$83,$A9,$E7,$3D,$3E,$73,$AF,$D2,$4A,$E7,$19,$7A,$C6,$5E,$91
	.dc.b	$F9,$7F,$1F,$FC,$81,$DD,$FE,$88,$FA,$57,$D3,$7E,$6D,$B7,$FB,$0A
	.dc.b	$DF,$81,$FE,$D2,$6C,$7F,$4F,$6B,$FE,$B9,$F1,$79,$77,$77,$64,$73
	.dc.b	$72,$FE,$F2,$FF,$32,$DE,$CF,$11,$E3,$EE,$B1,$92,$1F,$AA,$94,$7A
	.dc.b	$7A,$BD,$5F,$23,$2B,$B4,$65,$F0,$46,$5E,$51,$F1,$7A,$FF,$79,$EB
	.dc.b	$FF,$94,$77,$FF,$56,$B7,$6F,$85,$3B,$BF,$8C,$D7,$F7,$B3,$F9,$AA
	.dc.b	$FF,$5A,$ED,$FA,$39,$FE,$02,$D8,$CD,$78,$3B,$3F,$86,$7D,$33,$8E
	.dc.b	$93,$0F,$E4,$70,$FA,$6D,$FA,$12,$72,$FF,$15,$FA,$67,$75,$24,$FF
	.dc.b	$2C,$8C,$BF,$CB,$BE,$9B,$F8,$D6,$9F,$EC,$C9,$7C,$47,$32,$57,$F3
	.dc.b	$4F,$A6,$FF,$02,$D7,$FB,$49,$BC,$3F,$E1,$39,$46,$1F,$F0,$B3,$57
	.dc.b	$D2,$BE,$43,$F9,$63,$63,$B7,$27,$F2,$FA,$E1,$39,$9F,$E5,$7F,$27
	.dc.b	$FA,$B7,$FA,$3C,$3E,$9B,$99,$FC,$DE,$75,$3C,$3B,$F0,$47,$8F,$85
	.dc.b	$5D,$1F,$D4,$A7,$8E,$9F,$EF,$0C,$3F,$D4,$4F,$FB,$CB,$DF,$F2,$3F
	.dc.b	$EF,$4F,$A1,$7B,$DF,$D1,$DB,$87,$F5,$CA,$DE,$FF,$1D,$B1,$FE,$3B
	.dc.b	$6A,$6C,$BE,$1B,$FA,$94,$68,$9C,$1D,$E3,$FA,$B7,$88,$F8,$EF,$EB
	.dc.b	$1F,$48,$F8,$BA,$6A,$26,$5F,$DC,$2B,$83,$63,$8F,$C6,$EA,$3F,$AD
	.dc.b	$CE,$9B,$3A,$7F,$BC,$FD,$2B,$E8,$3C,$3F,$4D,$27,$31,$FD,$AE,$75
	.dc.b	$18,$21,$FD,$A3,$2F,$E0,$EF,$F6,$A8,$DB,$67,$91,$EE,$F2,$3F,$BB
	.dc.b	$7D,$37,$D8,$E1,$FC,$A6,$1F,$DE,$2F,$FC,$1A,$D7,$FB,$AC,$ED,$FF
	.dc.b	$36,$35,$34,$C4,$7F,$85,$23,$3C,$FE,$02,$3D,$BD,$BF,$1F,$C2,$7D
	.dc.b	$2F,$62,$4F,$F8,$7B,$04,$EA,$30,$D7,$E1,$76,$FF,$84,$7E,$99,$CF
	.dc.b	$49,$3F,$E2,$8F,$A6,$7A,$EE,$BF,$F4,$37,$D3,$3E,$9B,$F6,$D8,$F2
	.dc.b	$C3,$FC,$5E,$D7,$FC,$2E,$DA,$70,$FE,$F7,$1E,$B1,$D2,$F1,$8F,$7C
	.dc.b	$18,$F7,$7F,$D2,$D8,$F5,$8C,$42,$31,$5C,$AF,$A4,$FF,$A7,$70,$EE
	.dc.b	$F4,$9F,$4E,$5F,$C3,$F0,$E7,$87,$2A,$E1,$8F,$8E,$7B,$E1,$C9,$DD
	.dc.b	$F1,$C3,$9D,$F8,$05,$6E,$3C,$91,$E1,$3D,$73,$C5,$A7,$19,$E0,$F8
	.dc.b	$B1,$E9,$8F,$96,$3E,$05,$71,$7B,$E3,$E7,$C3,$BE,$3E,$69,$E3,$5B
	.dc.b	$C7,$AC,$70,$C7,$2C,$3C,$C6,$4B,$1F,$29,$FE,$6C,$64,$AE,$5F,$79
	.dc.b	$8F,$19,$F1,$5F,$E8,$79,$FE,$9A,$D8,$CC,$FF,$A4,$27,$2F,$F9,$C7
	.dc.b	$E0,$FF,$CE,$49,$A4,$74,$7A,$3E,$5D,$F4,$9F,$90,$8F,$59,$FE,$DF
	.dc.b	$9F,$AD,$3A,$DF,$EC,$4E,$4E,$55,$C7,$4F,$C7,$C6,$5F,$C8,$D7,$18
	.dc.b	$EA,$79,$4E,$35,$D1,$C4,$F0,$AE,$9A,$FF,$AB,$B0,$F1,$C6,$2B,$E6
	.dc.b	$C6,$2F,$93,$63,$BB,$6F,$CB,$1D,$AF,$FB,$56,$1E,$59,$CD,$6E,$7C
	.dc.b	$15,$E6,$39,$96,$7C,$D7,$FE,$AA,$BF,$76,$9F,$2E,$BF,$BB,$61,$E3
	.dc.b	$8C,$4E,$AB,$C3,$C1,$F4,$67,$39,$59,$C9,$D7,$A3,$18,$FF,$0F,$FA
	.dc.b	$67,$76,$21,$97,$F4,$3B,$FF,$60,$E2,$BD,$C7,$33,$2E,$2B,$FD,$53
	.dc.b	$5C,$5F,$F5,$7D,$71,$BE,$D3,$B8,$F5,$C7,$93,$E2,$DE,$8F,$F7,$4C
	.dc.b	$7E,$24,$F0,$35,$F4,$CF,$89,$C4,$FF,$B5,$AD,$39,$8E,$26,$5C,$65
	.dc.b	$E6,$BE,$77,$F0,$7C,$AF,$E7,$B7,$FE,$DC,$6D,$7F,$B2,$21,$F4,$8C
	.dc.b	$78,$75,$49,$FF,$73,$9A,$E0,$C5,$F1,$BF,$B0,$86,$1D,$B5,$FD,$55
	.dc.b	$F1,$3C,$58,$7D,$87,$15,$FE,$D6,$FA,$67,$B3,$F6,$5B,$3E,$2C,$5C
	.dc.b	$79,$A9,$39,$6D,$F9,$5B,$8F,$8A,$F9,$CF,$79,$F1,$46,$27,$F2,$D5
	.dc.b	$FE,$5A,$3D,$8E,$3A,$37,$DF,$F8,$D9,$D6,$1F,$EE,$E6,$D7,$C0,$FA
	.dc.b	$6F,$F2,$74,$72,$3F,$CA,$C6,$8F,$90,$1D,$CF,$3B,$F8,$D1,$8A,$F1
	.dc.b	$57,$D4,$8C,$3F,$E6,$1F,$95,$FC,$93,$78,$BD,$B3,$FA,$39,$CE,$1F
	.dc.b	$5D,$B5,$F8,$2F,$89,$D5,$FC,$46,$C7,$D1,$6D,$7E,$05,$FD,$77,$D2
	.dc.b	$3D,$1A,$7E,$C7,$E9,$9F,$4D,$8C,$3E,$16,$1F,$87,$BE,$11,$88,$43
	.dc.b	$DD,$E1,$1F,$E0,$B9,$FF,$3F,$4E,$5E,$0D,$8D,$1F,$93,$FA,$6F,$C9
	.dc.b	$CE,$AB,$F8,$24,$62,$76,$78,$3D,$2F,$E1,$1D,$2F,$33,$DB,$18,$66
	.dc.b	$1E,$79,$FF,$2E,$36,$38,$FA,$3F,$C4,$7E,$9B,$FC,$D1,$3E,$DE,$F7
	.dc.b	$B9,$9F,$53,$F8,$B4,$62,$7E,$17,$D3,$7F,$A8,$9B,$ED,$FF,$1E,$FA
	.dc.b	$67,$D3,$3E,$9B,$F8,$F5,$EF,$83,$D6,$DE,$BF,$EB,$5E,$3B,$E9,$BF
	.dc.b	$CE,$99,$F4,$3E,$9C,$67,$D9,$E7,$A7,$F2,$C8,$C4,$EF,$1F,$9B,$18
	.dc.b	$7E,$7F,$6F,$F0,$1F,$A6,$FE,$66,$FF,$03,$EE,$B2,$9B,$18,$6D,$FE
	.dc.b	$12,$B9,$36,$BF,$2F,$1F,$C3,$7E,$9B,$99,$D3,$96,$D6,$E2,$D7,$90
	.dc.b	$FA,$A5,$7F,$A2,$7D,$23,$D8,$A7,$F3,$7F,$A6,$FA,$55,$F8,$49,$D5
	.dc.b	$FE,$55,$B1,$F8,$EE,$A5,$7F,$1F,$AF,$F6,$95,$6B,$87,$FC,$15,$5A
	.dc.b	$F2,$BB,$D7,$F3,$DB,$E1,$87,$F2,$9C,$38,$4F,$F2,$BB,$E0,$FC,$5F
	.dc.b	$57,$FC,$C6,$DE,$9F,$9F,$18,$BF,$A6,$FE,$41,$B3,$A9,$DB,$4E,$9F
	.dc.b	$4D,$C8,$E7,$F1,$56,$32,$4A,$EB,$19,$7E,$37,$C7,$FC,$C3,$F1,$1F
	.dc.b	$04,$F9,$5C,$5F,$C9,$9F,$EE,$FB,$BE,$9B,$E7,$BF,$E0,$5D,$93,$A9
	.dc.b	$CD,$7C,$C8,$C7,$81,$D3,$2F,$EE,$75,$A5,$CB,$FE,$48,$F8,$EF,$8A
	.dc.b	$7C,$6F,$A6,$B8,$FC,$B8,$B5,$F2,$26,$E7,$FE,$41,$BF,$F2,$8F,$CB
	.dc.b	$7C,$EF,$88,$F9,$ED,$C7,$EA,$37,$0F,$A2,$DB,$FF,$10,$7D,$37,$F8
	.dc.b	$06,$7F,$87,$B6,$3F,$6B,$C1,$FC,$8A,$73,$87,$F8,$66,$B5,$87,$E0
	.dc.b	$FE,$99,$F2,$B9,$AF,$F0,$74,$62,$7F,$95,$57,$0B,$FE,$F9,$39,$7A
	.dc.b	$CF,$F8,$36,$31,$8F,$BF,$99,$F4,$DF,$8C,$FA,$6F,$C6,$B6,$FF,$24
	.dc.b	$DC,$3F,$1E,$DB,$E5,$C5,$F9,$4E,$BF,$CA,$B6,$FF,$34,$DB,$FC,$BB
	.dc.b	$6B,$FC,$8D,$F4,$DF,$E2,$8A,$FF,$38,$6D,$FE,$0D,$B7,$D3,$3A,$7F
	.dc.b	$C3,$11,$FE,$7D,$9D,$5F,$18,$C1,$FE,$15,$AE,$1D,$9D,$23,$FC,$31
	.dc.b	$3B,$F8,$3F,$9B,$1B,$C3,$AC,$62,$13,$BC,$3C,$98,$74,$9E,$E7,$9B
	.dc.b	$E9,$BF,$C4,$6F,$FA,$80,$C7,$BF,$5F,$5E,$3E,$A7,$FD,$4B,$4A,$E3
	.dc.b	$18,$79,$E3,$E3,$9D,$5F,$AE,$BA,$9E,$2F,$A5,$7D,$47,$F9,$C3,$2F
	.dc.b	$95,$0F,$A7,$C5,$7F,$8E,$63,$0F,$36,$BC,$FE,$13,$6A,$33,$97,$C5
	.dc.b	$6D,$F6,$BD,$C7,$F9,$76,$BA,$5E,$08,$FA,$7E,$9B,$FC,$7A,$E8,$87
	.dc.b	$84,$D7,$C1,$F0,$70,$4C,$7E,$B4,$F2,$8C,$63,$F0,$2F,$85,$6A,$EC
	.dc.b	$E6,$BB,$CE,$5B,$19,$79,$9F,$E5,$FB,$ED,$3F,$06,$F9,$CF,$CB,$7E
	.dc.b	$3C,$EA,$B9,$56,$EB,$A9,$CE,$F8,$BF,$36,$ED,$F1,$8E,$87,$03,$47
	.dc.b	$C4,$31,$87,$CF,$7C,$E2,$39,$3C,$0C,$C6,$EF,$E0,$46,$2B,$C3,$7E
	.dc.b	$4A,$D6,$1E,$A3,$F6,$98,$74,$8C,$DF,$98,$F0,$C6,$6F,$91,$F1,$2E
	.dc.b	$CF,$AE,$74,$FA,$A7,$4D,$35,$3C,$1E,$AD,$C3,$E6,$47,$EC,$AB,$9B
	.dc.b	$EB,$3E,$31,$4E,$27,$52,$99,$8E,$67,$C4,$2E,$1F,$06,$3D,$D1,$98
	.dc.b	$C4,$7C,$D8,$DB,$F7,$4E,$CE,$1F,$A9,$5E,$99,$E3,$5E,$8A,$D6,$1F
	.dc.b	$32,$F6,$72,$AD,$DF,$18,$F1,$C7,$D4,$8C,$47,$8E,$30,$F1,$36,$79
	.dc.b	$4F,$80,$F4,$74,$71,$35,$5E,$4C,$3F,$0B,$1A,$BF,$36,$1F,$8E,$3F
	.dc.b	$65,$C5,$FF,$00,$5B,$C5,$E2,$AF,$F3,$8C,$7F,$B0,$1F,$F4,$13,$A7
	.dc.b	$C9,$3C,$72,$FE,$84,$F3,$F8,$3B,$7F,$D1,$7E,$9D,$F0,$78,$C3,$FD
	.dc.b	$A9,$18,$7F,$D1,$CD,$7C,$3E,$56,$D4,$67,$05,$F5,$BD,$4E,$63,$F4
	.dc.b	$37,$C7,$2F,$F2,$66,$1F,$CF,$5A,$FF,$41,$6B,$FD,$6C,$FD,$A4,$7F
	.dc.b	$43,$69,$F0,$9E,$87,$5F,$8B,$F3,$7C,$17,$F8,$9A,$7F,$D5,$0B,$E9
	.dc.b	$F2,$6B,$C8,$93,$A8,$CB,$C2,$39,$FB,$9D,$3E,$7E,$4F,$48,$CB,$DA
	.dc.b	$32,$FF,$AA,$2F,$FD,$4C,$FF,$AE,$7E,$9D,$F4,$FE,$6C,$E4,$E6,$3E
	.dc.b	$74,$EA,$30,$43,$A7,$16,$A6,$46,$78,$7E,$9A,$18,$7E,$1D,$7D,$3B
	.dc.b	$DF,$FB,$8F,$FB,$2E,$3F,$D9,$CD,$7F,$D8,$2E,$DC,$C7,$0F,$83,$91
	.dc.b	$83,$8F,$C9,$F4,$23,$D3,$E2,$AB,$D2,$2E,$3E,$88,$F2,$CF,$E3,$FE
	.dc.b	$9D,$E7,$79,$BD,$49,$5F,$EE,$39,$CB,$67,$81,$1F,$5F,$83,$5F,$F6
	.dc.b	$DC,$F0,$6C,$FC,$29,$FC,$F7,$D3,$BE,$9E,$92,$BF,$DE,$50,$F9,$9C
	.dc.b	$47,$FB,$9A,$38,$3A,$9E,$73,$D3,$E7,$7A,$FD,$2C,$AE,$71,$97,$AC
	.dc.b	$65,$E9,$1F,$B1,$F2,$7F,$C5,$DB,$7C,$7F,$A6,$7D,$3F,$ED,$5B,$7F
	.dc.b	$81,$9F,$99,$5C,$3C,$0F,$F2,$33,$67,$F0,$CD,$7F,$C9,$1E,$3F,$3E
	.dc.b	$EF,$03,$23,$A5,$72,$8D,$46,$0F,$C8,$61,$FA,$05,$87,$9F,$BB,$CC
	.dc.b	$79,$FB,$AC,$F1,$15,$C6,$B8,$46,$61,$3E,$FE,$CF,$81,$E2,$4A,$D4
	.dc.b	$65,$E1,$19,$7C,$79,$7F,$99,$EF,$F4,$9A,$7E,$7F,$1F,$C9,$7B,$3F
	.dc.b	$E3,$C6,$F9,$3F,$47,$AE,$63,$E7,$BE,$18,$FC,$C9,$FE,$4D,$87,$E9
	.dc.b	$EB,$B9,$F3,$A3,$8B,$83,$FC,$EC,$F0,$EC,$FE,$5D,$F4,$EE,$4A,$4A
	.dc.b	$F8,$D0,$FA,$7D,$FA,$32,$72,$FF,$37,$FA,$67,$7D,$24,$FC,$F8,$CB
	.dc.b	$F4,$7E,$9F,$F9,$43,$E5,$87,$C4,$73,$24,$7D,$7F,$A7,$FE,$18,$4E
	.dc.b	$5F,$E8,$CF,$F0,$FF,$A6,$7C,$97,$F4,$96,$D7,$1A,$D9,$99,$FE,$9D
	.dc.b	$F2,$FF,$AD,$9F,$C2,$53,$E9,$F9,$1A,$F0,$24,$FE,$26,$B8,$3F,$35
	.dc.b	$BD,$1F,$C3,$6F,$FD,$0E,$FF,$0F,$8C,$F3,$7F,$AA,$5B,$5F,$CA,$5A
	.dc.b	$AB,$8C,$FF,$AB,$FE,$07,$D7,$3F,$6F,$E8,$9F,$05,$FD,$9D,$B1,$F9
	.dc.b	$D9,$DE,$3F,$C7,$DB,$7C,$F4,$FE,$DD,$18,$AF,$E7,$56,$F8,$8C,$D4
	.dc.b	$9E,$8F,$78,$FE,$ED,$E2,$3E,$3F,$FB,$C7,$D2,$BE,$36,$9A,$89,$87
	.dc.b	$EC,$6B,$83,$63,$8F,$C7,$EA,$3F,$31,$3A,$6C,$E9,$FD,$D7,$D3,$3E
	.dc.b	$8B,$C3,$F4,$F2,$73,$1F,$E0,$B9,$D4,$60,$87,$E9,$72,$FE,$05,$BF
	.dc.b	$FB,$76,$77,$3F,$A9,$6F,$07,$F4,$96,$CF,$FB,$8E,$B8,$D7,$F6,$3F
	.dc.b	$A7,$FB,$71,$FD,$92,$3F,$88,$35,$FD,$A4,$7E,$00,$9C,$BF,$E2,$17
	.dc.b	$8F,$BB,$62,$4F,$F1,$99,$CB,$7E,$1B,$E8,$FB,$52,$30,$43,$BC,$75
	.dc.b	$3E,$BF,$57,$F8,$BF,$E9,$DC,$D4,$95,$FC,$9F,$E9,$DE,$C7,$F2,$E8
	.dc.b	$7D,$3B,$DF,$FF,$01,$43,$0E,$6F,$F2,$F6,$BF,$C4,$DB,$C5,$FD,$CF
	.dc.b	$0E,$AF,$F8,$2E,$FA,$57,$F8,$36,$72,$F8,$2F,$BD,$F4,$AD,$CF,$87
	.dc.b	$1F,$1C,$62,$F8,$B7,$1F,$53,$C2,$F9,$61,$D2,$F8,$BE,$CB,$E2,$DB
	.dc.b	$E2,$EF,$0F,$43,$5F,$31,$E9,$8D,$B7,$1D,$BF,$E2,$9C,$3C,$2D,$C3
	.dc.b	$C2,$EC,$E6,$30,$F1,$BB,$39,$9F,$04,$AE,$9E,$17,$E2,$74,$74,$AF
	.dc.b	$0B,$AB,$E3,$3B,$9C,$9A,$AE,$87,$D4,$AF,$F3,$53,$5F,$8D,$5F,$E6
	.dc.b	$D7,$D4,$6B,$C4,$7C,$66,$63,$D9,$18,$78,$57,$F4,$C9,$D3,$63,$FC
	.dc.b	$ED,$F0,$BF,$E7,$74,$D2,$3A,$3E,$4F,$CB,$BE,$97,$FC,$B3,$B7,$E2
	.dc.b	$27,$FC,$B5,$5F,$2A,$B5,$CD,$F3,$27,$27,$2B,$E3,$BF,$B6,$F8,$46
	.dc.b	$27,$FD,$19,$18,$AE,$51,$E6,$39,$38,$8D,$47,$DE,$3B,$E1,$F9,$CC
	.dc.b	$3C,$71,$93,$94,$66,$FC,$11,$FB,$2A,$E7,$3E,$D8,$DE,$1F,$B6,$BE
	.dc.b	$73,$9A,$DC,$FA,$EB,$CA,$77,$8C,$BF,$E7,$FA,$F1,$E9,$FE,$6B,$6E
	.dc.b	$3C,$70,$F2,$3F,$E2,$F8,$CD,$71,$E0,$FF,$44,$DF,$1C,$AD,$F9,$66
	.dc.b	$D7,$7E,$8B,$6F,$A7,$76,$EA,$99,$7F,$43,$C7,$FC,$8D,$7B,$AD,$4E
	.dc.b	$6B,$FC,$4F,$39,$AF,$36,$1D,$B0,$EB,$1E,$9C,$38,$9E,$4F,$8D,$FF
	.dc.b	$4E,$78,$78,$9D,$1F,$E5,$58,$7D,$3B,$E2,$B5,$9F,$F8,$DA,$73,$18
	.dc.b	$CF,$FD,$57,$18,$BE,$75,$BB,$E5,$1F,$E5,$9F,$7D,$D3,$F0,$5B,$1F
	.dc.b	$E6,$51,$F4,$CC,$35,$52,$75,$3F,$53,$05,$F2,$7F,$C9,$73,$CB,$87
	.dc.b	$84,$62,$B8,$56,$F2,$FF,$61,$DE,$CF,$F3,$77,$D3,$BD,$AF,$AC,$D8
	.dc.b	$F0,$F1,$E2,$64,$79,$1B,$8F,$76,$DF,$6C,$79,$4F,$F9,$77,$0B,$39
	.dc.b	$AF,$C2,$5F,$F9,$50,$FF,$40,$16,$BD,$5E,$06,$BA,$A4,$FC,$9F,$A7
	.dc.b	$FF,$25,$BC,$23,$F2,$6A,$33,$5F,$E6,$D8,$E9,$87,$C2,$9C,$BF,$EC
	.dc.b	$F8,$C9,$FE,$E7,$39,$79,$3F,$99,$F9,$6F,$C7,$DF,$F9,$DD,$B5,$EF
	.dc.b	$BD,$72,$CC,$3E,$CF,$C4,$5C,$3E,$A3,$6F,$E3,$36,$7C,$1B,$7F,$BD
	.dc.b	$9F,$F5,$17,$D2,$BC,$FF,$D5,$36,$3F,$D4,$DF,$4E,$FA,$7C,$D4,$9E
	.dc.b	$3E,$E5,$CF,$FD,$1D,$82,$39,$9F,$B5,$FA,$7F,$C6,$D7,$03,$FD,$64
	.dc.b	$E2,$B8,$F0,$7E,$EF,$A5,$F0,$7E,$48,$C3,$2F,$FD,$75,$3F,$E4,$F6
	.dc.b	$C7,$1F,$47,$F8,$27,$D3,$FF,$96,$E7,$FD,$81,$EF,$BD,$CC,$FA,$BF
	.dc.b	$C2,$63,$13,$FE,$C5,$FA,$4F,$F4,$9B,$7D,$BF,$E2,$7F,$4E,$FA,$77
	.dc.b	$D3,$FF,$11,$BD,$F0,$7F,$AB,$DB,$D2,$C7,$F9,$47,$D3,$FF,$99,$DE
	.dc.b	$87,$FB,$6F,$CB,$46,$6F,$96,$7F,$C8,$A7,$3E,$D7,$78,$C5,$6F,$0E
	.dc.b	$D6,$EF,$2F,$A7,$FE,$57,$18,$EF,$B3,$9C,$FF,$DD,$93,$96,$C6,$4F
	.dc.b	$F6,$6C,$65,$A6,$6F,$C3,$19,$6C,$3E,$9F,$B9,$E0,$E5,$B7,$DA,$2C
	.dc.b	$F4,$9B,$5E,$C8,$7D,$2B,$D8,$FE,$80,$DC,$3F,$98,$7D,$3F,$DA,$AF
	.dc.b	$F7,$04,$E5,$B5,$FC,$E6,$7F,$8C,$57,$F9,$62,$B5,$7F,$CC,$A7,$37
	.dc.b	$F2,$A7,$35,$F3,$3A,$95,$AA,$F2,$E1,$D7,$CC,$F9,$FE,$B7,$E0,$70
	.dc.b	$F9,$AF,$E4,$EB,$E5,$B9,$6F,$4F,$09,$E2,$D7,$8B,$D3,$E9,$E4,$74
	.dc.b	$F8,$BB,$5D,$26,$EB,$F5,$30,$56,$2F,$E0,$7C,$87,$CA,$3F,$1D,$F0
	.dc.b	$6F,$96,$C4,$FE,$2B,$0F,$A5,$C0,$BF,$EE,$7F,$4F,$F3,$E7,$31,$8F
	.dc.b	$03,$8E,$BF,$DB,$AB,$58,$78,$0F,$EE,$AE,$23,$F3,$1A,$3E,$3B,$E4
	.dc.b	$9F,$2B,$E9,$ED,$FC,$A8,$B5,$F2,$A6,$E5,$EF,$F9,$7F,$A3,$EC,$FD
	.dc.b	$26,$E1,$F4,$9B,$7F,$49,$B5,$F8,$FF,$A7,$FF,$76,$D7,$C1,$3F,$D3
	.dc.b	$AE,$29,$3F,$B0,$F0,$1C,$3F,$CA,$67,$37,$C3,$1F,$F0,$B6,$1C,$70
	.dc.b	$FE,$FF,$F4,$EF,$96,$CD,$7F,$83,$63,$17,$E1,$C3,$F7,$78,$71,$8F
	.dc.b	$C9,$4E,$A7,$B9,$AA,$E5,$7F,$92,$FA,$6F,$C9,$FD,$3D,$C3,$F2,$6D
	.dc.b	$BF,$C9,$B7,$2F,$C4,$3F,$9A,$E2,$FC,$DF,$4D,$BF,$CD,$B6,$FF,$36
	.dc.b	$DB,$FC,$83,$FE,$35,$FA,$5F,$C8,$F6,$FF,$9B,$B3,$FF,$63,$C6,$2F
	.dc.b	$FC,$F3,$1F,$EC,$AF,$A7,$79,$7F,$C7,$3E,$13,$E3,$57,$F1,$CB,$E1
	.dc.b	$18,$84,$3E,$9F,$FC,$37,$1F,$E9,$18,$C7,$BF,$FE,$8F,$7C,$11,$FE
	.dc.b	$D8,$38,$CF,$07,$1F,$4C,$FA,$7F,$F3,$66,$7E,$8A,$EB,$8F,$87,$0E
	.dc.b	$B8,$7A,$8F,$E6,$38,$71,$7E,$01,$8C,$39,$BE,$BB,$E7,$87,$5C,$77
	.dc.b	$2B,$E4,$F5,$31,$F0,$9C,$9D,$7C,$DD,$93,$66,$4F,$49,$CB,$67,$2F
	.dc.b	$5F,$A6,$7C,$26,$35,$F9,$D7,$FE,$5B,$AF,$2D,$69,$B5,$BC,$BF,$A8
	.dc.b	$C7,$98,$C9,$0F,$55,$73,$6E,$3E,$BB,$E9,$1E,$3B,$E8,$FC,$C9,$D4
	.dc.b	$48,$F6,$46,$94,$F2,$C3,$F0,$CD,$C3,$CB,$87,$86,$7D,$B7,$CE,$FC
	.dc.b	$07,$42,$18,$24,$70,$72,$F6,$C3,$93,$98,$F6,$9F,$05,$D3,$C2,$FC
	.dc.b	$75,$C1,$F9,$38,$79,$E7,$E0,$1B,$3F,$BE,$B9,$7D,$F7,$F0,$70,$F6
	.dc.b	$57,$28,$E9,$3C,$8F,$19,$EA,$8D,$3D,$4F,$B0,$64,$EF,$1D,$6F,$9C
	.dc.b	$78,$0E,$15,$E0,$76,$7B,$EB,$DA,$F5,$BF,$41,$E9,$9E,$11,$E0,$C3
	.dc.b	$E2,$BE,$13,$15,$C9,$C5,$71,$BF,$C0,$3D,$30,$F3,$4F,$84,$F1,$9F
	.dc.b	$02,$FE,$A1,$D8,$97,$D5,$CB,$F1,$23,$C5,$87,$98,$F3,$DF,$78,$F8
	.dc.b	$38,$7C,$48,$F2,$47,$BF,$0F,$8F,$5C,$6B,$51,$E6,$9F,$5B,$70,$F1
	.dc.b	$C7,$A1,$F1,$BF,$65,$CB,$F4,$4F,$66,$1E,$93,$D1,$1A,$BF,$04,$FC
	.dc.b	$33,$84,$74,$7E,$69,$DE,$78,$9C,$DF,$03,$87,$46,$08,$74,$74,$F4
	.dc.b	$26,$1F,$1D,$F3,$9C,$1C,$4F,$8E,$BE,$6D,$FF,$03,$F8,$3F,$E1,$7C
	.dc.b	$7F,$E4,$3B,$FE,$0A,$D9,$CF,$27,$EA,$2A,$E3,$FA,$66,$BF,$1E,$B8
	.dc.b	$3F,$DB,$4F,$E0,$F1,$AB,$FF,$27,$46,$7E,$9D,$F0,$9C,$9B,$1C,$67
	.dc.b	$6F,$F7,$23,$F8,$8F,$95,$FD,$35,$FD,$DB,$5F,$DC,$B5,$D9,$98,$7C
	.dc.b	$5F,$8E,$64,$F7,$8F,$31,$87,$C1,$3E,$2C,$7E,$34,$FC,$0F,$83,$FF
	.dc.b	$1C,$57,$BD,$7D,$3E,$4D,$79,$12,$75,$19,$78,$47,$3F,$73,$A7,$D0
	.dc.b	$49,$E9,$19,$7B,$46,$5F,$89,$7F,$EB,$C7,$E8,$7D,$3B,$E9,$FC,$DA
	.dc.b	$C9,$CC,$7D,$89,$D4,$60,$87,$4E,$2D,$4C,$8C,$EB,$FE,$89,$C3,$F0
	.dc.b	$FC,$5F,$D5,$3E,$9F,$FC,$1B,$5F,$81,$7F,$DB,$0D,$F8,$3F,$C3,$C6
	.dc.b	$A3,$EE,$36,$FF,$D2,$8F,$08,$CC,$F0,$EE,$E3,$F2,$72,$7B,$3D,$3E
	.dc.b	$2F,$11,$D2,$2E,$1D,$27,$F2,$FF,$4E,$F3,$7C,$7E,$9C,$C3,$F3,$13
	.dc.b	$96,$CF,$02,$3E,$DF,$02,$BF,$95,$9E,$0D,$9F,$7C,$FE,$9B,$E9,$DF
	.dc.b	$4F,$49,$5F,$A7,$87,$CC,$E2,$3F,$41,$1C,$1D,$4F,$39,$E9,$F3,$BD
	.dc.b	$7E,$96,$57,$38,$CB,$D6,$32,$F4,$8F,$DC,$FB,$3F,$E3,$FD,$BF,$D8
	.dc.b	$9F,$4C,$FA,$7F,$DE,$B6,$FF,$79,$5B,$E5,$FF,$09,$B6,$3F,$DC,$AD
	.dc.b	$7F,$D4,$BE,$37,$2E,$EE,$EC,$8E,$6E,$5F,$E1,$17,$F9,$B7,$F8,$03
	.dc.b	$71,$F2,$F7,$79,$3B,$BC,$AC,$AE,$2E,$5E,$7D,$9E,$AF,$12,$46,$08
	.dc.b	$76,$72,$F8,$23,$E2,$F5,$7F,$1E,$F6,$3F,$C7,$B3,$FC,$4D,$B8,$FF
	.dc.b	$BA,$F5,$FC,$C3,$6B,$8D,$7C,$87,$97,$5E,$23,$F8,$E5,$70,$DF,$E5
	.dc.b	$D6,$A7,$F9,$57,$D3,$B8,$FF,$97,$B6,$33,$18,$FA,$7C,$FA,$52,$78
	.dc.b	$BF,$CD,$BE,$99,$DD,$49,$3F,$CD,$E3,$2F,$F3,$9F,$A7,$FE,$3A,$FF
	.dc.b	$31,$87,$C4,$F3,$24,$7F,$3F,$FA,$7F,$E0,$68,$38,$6F,$17,$F4,$88
	.dc.b	$C7,$D3,$3E,$4B,$FA,$5B,$6B,$A5,$6C,$CC,$FF,$50,$F9,$5F,$D3,$CF
	.dc.b	$F4,$FA,$7D,$3F,$23,$5D,$F2,$7F,$AA,$57,$07,$FA,$63,$79,$9C,$9F
	.dc.b	$E6,$EC,$3F,$D3,$11,$8A,$F6,$56,$EB,$FC,$E6,$D8,$CD,$FA,$BC,$2F
	.dc.b	$EC,$BF,$01,$EB,$9F,$EB,$5F,$46,$F7,$FF,$B4,$B6,$3F,$B4,$4E,$F2
	.dc.b	$F0,$47,$FB,$8B,$5E,$64,$7F,$B7,$53,$E2,$31,$1F,$D9,$EF,$96,$9F
	.dc.b	$E8,$86,$C7,$D5,$AE,$0F,$F7,$7F,$11,$F2,$1F,$DF,$3E,$95,$F1,$94
	.dc.b	$FE,$FF,$3A,$AE,$15,$82,$3F,$E0,$1A,$7C,$7E,$23,$FB,$D5,$71,$6D
	.dc.b	$77,$9F,$F0,$5F,$D3,$3E,$8B,$C5,$F4,$F2,$FF,$C1,$93,$96,$CF,$02
	.dc.b	$3E,$1F,$1A,$BF,$E0,$99,$E0,$D9,$F7,$4F,$F8,$6B,$E9,$FE,$D4,$B3
	.dc.b	$7A,$9F,$0C,$D7,$34,$9F,$F0,$F5,$70,$38,$D7,$F8,$6A,$3F,$C3,$11
	.dc.b	$99,$E1,$1F,$E1,$B9,$D7,$27,$DE,$A7,$C3,$7D,$1F,$62,$5F,$68,$F3
	.dc.b	$11,$C1,$F3,$FB,$3F,$C6,$7F,$4E,$E4,$A4,$9F,$F1,$A7,$D3,$BD,$76
	.dc.b	$7F,$B7,$FA,$7F,$F1,$2D,$7E,$E7,$0E,$4F,$F8,$ED,$AF,$F8,$CD,$B3
	.dc.b	$F9,$6B,$E8,$F3,$F7,$7F,$DA,$F1,$8A,$5F,$65,$7D,$23,$0D,$AE,$38
	.dc.b	$79,$AB,$8C,$63,$0E,$D1,$E7,$C7,$FD,$CB,$87,$26,$D6,$E7,$89,$FC
	.dc.b	$57,$1F,$25,$6E,$39,$9E,$4A,$E2,$EF,$0F,$14,$FA,$6B,$8D,$FF,$BD
	.dc.b	$31,$F4,$5F,$8C,$E7,$3A,$79,$63,$DA,$77,$1D,$CD,$E1,$DA,$FD,$18
	.dc.b	$79,$67,$D2,$6C,$B8,$75,$4D,$BE,$A7,$67,$47,$D3,$8F,$E6,$C7,$90
	.dc.b	$FF,$36,$39,$7E,$41,$A3,$AF,$8C,$F9,$0F,$F3,$84,$FF,$4C,$6C,$66
	.dc.b	$79,$B5,$FF,$3B,$7C,$1F,$F9,$D5,$34,$8F,$F9,$E5,$FA,$7F,$1E,$FA
	.dc.b	$5F,$E6,$59,$7E,$26,$BF,$A7,$56,$A7,$39,$FF,$A1,$63,$1D,$FF,$72
	.dc.b	$B6,$7C,$AC,$3A,$46,$63,$C6,$66,$F9,$BF,$06,$B9,$BF,$48,$E7,$87
	.dc.b	$E6,$EF,$9C,$64,$E7,$19,$AE,$71,$F4,$2B,$9C,$FA,$E3,$78,$7D,$4B
	.dc.b	$E7,$39,$AD,$CF,$B6,$BC,$D8,$7A,$67,$31,$E7,$32,$4D,$BF,$A5,$E1
	.dc.b	$E6,$7F,$A6,$C7,$AA,$FC,$18,$FF,$5F,$9D,$B7,$0F,$27,$05,$BF,$EF
	.dc.b	$D8,$73,$9C,$9F,$EC,$17,$F9,$D3,$FE,$03,$BE,$9F,$4F,$FD,$12,$35
	.dc.b	$D3,$CE,$53,$9B,$FF,$60,$5F,$27,$F8,$BD,$F3,$9D,$57,$B2,$F9,$46
	.dc.b	$27,$E5,$76,$79,$ED,$86,$57,$C6,$BD,$3F,$EE,$97,$E9,$BE,$17,$BA
	.dc.b	$7D,$3B,$E2,$73,$5F,$0E,$73,$18,$BF,$80,$7F,$88,$E3,$2F,$53,$38
	.dc.b	$F8,$AF,$A4,$7F,$BA,$7C,$8F,$E8,$6D,$E9,$FA,$0B,$E9,$7D,$2E,$49
	.dc.b	$7F,$18,$FF,$0D,$C7,$D5,$7F,$C6,$95,$D1,$D4,$F3,$E1,$F8,$D3,$9E
	.dc.b	$0F,$F1,$04,$E4,$EA,$E2,$FA,$5B,$E9,$DE,$CF,$CF,$6D,$FB,$8F,$47
	.dc.b	$1E,$66,$46,$6F,$77,$FE,$FF,$BB,$87,$7B,$DC,$FA,$2B,$37,$FC,$6A
	.dc.b	$EC,$FC,$0A,$FC,$04,$FF,$98,$0F,$9F,$1F,$E3,$EE,$7F,$97,$6F,$8A
	.dc.b	$DC,$4A,$F4,$7D,$3F,$F9,$55,$E2,$73,$BF,$93,$19,$7E,$81,$D6,$38
	.dc.b	$F9,$9C,$E3,$6F,$E4,$FE,$73,$F2,$50,$7F,$29,$3F,$76,$B5,$C9,$23
	.dc.b	$E3,$78,$B6,$FE,$13,$76,$FA,$8D,$9F,$F3,$91,$F8,$77,$F4,$5F,$4A
	.dc.b	$F4,$29,$FA,$5F,$A7,$7D,$3E,$2F,$E3,$D6,$DB,$EE,$E7,$2F,$7C,$A3
	.dc.b	$A9,$FA,$FF,$A7,$73,$62,$FE,$8C,$61,$EB,$C1,$FB,$5E,$A7,$99,$E6
	.dc.b	$8C,$32,$7E,$CC,$FF,$95,$1B,$1C,$7D,$1F,$DF,$7D,$3F,$F9,$8E,$7E
	.dc.b	$FF,$80,$F7,$33,$EA,$FF,$03,$8C,$4F,$E1,$3E,$93,$FD,$3A,$DF,$6F
	.dc.b	$F8,$F7,$D3,$BE,$9D,$F4,$FF,$C2,$EF,$7C,$9F,$80,$6F,$4F,$FA,$1F
	.dc.b	$C7,$7D,$3F,$F9,$B3,$3E,$87,$F0,$08,$CF,$B3,$CF,$4F,$E5,$91,$89
	.dc.b	$DE,$3F,$BB,$8C,$3F,$BD,$EF,$FF,$57,$7D,$3F,$F3,$37,$FD,$61,$DF
	.dc.b	$65,$36,$30,$DB,$FE,$23,$5C,$9B,$5F,$C9,$63,$FD,$75,$F4,$FC,$CE
	.dc.b	$9C,$B6,$B7,$16,$BC,$87,$D5,$2B,$FC,$C7,$E9,$5E,$BD,$3F,$94,$7D
	.dc.b	$3F,$D2,$AF,$D4,$CE,$AF,$F6,$ED,$8F,$E3,$9D,$4A,$FE,$2D,$5F,$E1
	.dc.b	$AA,$D7,$0F,$FA,$DE,$B5,$E6,$77,$AF,$E5,$F1,$87,$AE,$3E,$29,$F9
	.dc.b	$5E,$A7,$FB,$0A,$7E,$2B,$F8,$BC,$BE,$83,$96,$F4,$F0,$9E,$2D,$78
	.dc.b	$BD,$3E,$9E,$47,$4F,$8A,$B5,$D2,$6E,$7F,$EC,$FE,$0F,$F1,$75,$62
	.dc.b	$FD,$DF,$1F,$F2,$CF,$C3,$FC,$13,$E5,$71,$3F,$22,$7F,$B6,$7D,$3F
	.dc.b	$CD,$C7,$FD,$97,$C2,$9C,$F8,$1C,$35,$FE,$91,$4B,$E1,$87,$80,$FE
	.dc.b	$AF,$19,$AF,$8F,$F1,$D2,$7E,$3F,$D3,$DC,$3E,$3C,$DB,$F6,$70,$AB
	.dc.b	$16,$BF,$E4,$52,$F9,$6F,$A7,$F1,$1F,$51,$B5,$F5,$1B,$5F,$51,$B8
	.dc.b	$7F,$BA,$3E,$9F,$FB,$DC,$FF,$34,$6C,$7E,$D7,$83,$F8,$E4,$E7,$0F
	.dc.b	$F0,$35,$6B,$0F,$EF,$1F,$4E,$F9,$5C,$D7,$F7,$F8,$C4,$FF,$4C,$AE
	.dc.b	$17,$FE,$0C,$9C,$BD,$67,$F2,$31,$8C,$7E,$1F,$D3,$FE,$43,$E9,$FF
	.dc.b	$22,$DA,$FC,$8B,$6F,$F2,$2D,$AF,$97,$C1,$F9,$8E,$9F,$CC,$B6,$FF
	.dc.b	$32,$DB,$FC,$CB,$69,$3F,$E8,$2F,$A6,$7D,$2F,$F3,$8B,$FF,$40,$08
	.dc.b	$CF,$A1,$88,$C7,$4F,$F8,$A9,$DC,$E9,$B5,$CF,$E1,$3E,$54,$7D,$F9
	.dc.b	$D4,$62,$3B,$4E,$F8,$7F,$D2,$BF,$48,$FA,$6F,$F3,$44,$61,$B5,$CE
	.dc.b	$B8,$37,$2F,$F4,$2D,$6E,$29,$F8,$FB,$E8,$FF,$66,$34,$7F,$99,$1C
	.dc.b	$B6,$FF,$D4,$F9,$FF,$A8,$E7,$FC,$85,$3C,$4E,$9F,$07,$A7,$D2,$D7
	.dc.b	$26,$C4,$72,$7F,$D4,$77,$C1,$F3,$9A,$9F,$7A,$FA,$47,$C1,$61,$B1
	.dc.b	$C2,$FA,$69,$F3,$5B,$87,$FB,$02,$B7,$5E,$03,$F6,$4E,$2B,$86,$1E
	.dc.b	$0B,$E3,$5A,$C3,$C0,$66,$64,$7A,$2F,$96,$1D,$EB,$33,$DC,$F6,$1B
	.dc.b	$79,$BE,$9A,$DE,$1F,$85,$BE,$73,$E1,$34,$43,$A3,$A7,$E3,$57,$2B
	.dc.b	$F9,$91,$A8,$EE,$F8,$F0,$F6,$BE,$43,$47,$94,$ED,$7F,$11,$C1,$A3
	.dc.b	$96,$1D,$DF,$31,$98,$E0,$E8,$E2,$FB,$6B,$84,$78,$63,$0E,$AF,$D3
	.dc.b	$1C,$8C,$C7,$58,$C9,$1F,$0C,$68,$E1,$19,$8F,$8A,$75,$3F,$2A,$79
	.dc.b	$4E,$87,$27,$85,$FC,$6C,$3E,$71,$F3,$DF,$92,$66,$FE,$6D,$7F,$3D
	.dc.b	$3E,$19,$4F,$09,$FC,$3A,$7F,$26,$E5,$AF,$E2,$4F,$E2,$0E,$1E,$66
	.dc.b	$4F,$F3,$8B,$F9,$CD,$FF,$8E,$C7,$D2,$6B,$FA,$9F,$81,$F2,$46,$27
	.dc.b	$86,$7C,$E3,$17,$A8,$FD,$45,$5B,$E9,$1B,$23,$FA,$A8,$D3,$EE,$AE
	.dc.b	$6F,$FA,$4A,$FB,$BB,$31,$7C,$23,$31,$C6,$FC,$2E,$88,$7C,$09,$E1
	.dc.b	$7C,$59,$1A,$BF,$74,$E2,$38,$19,$AE,$AE,$09,$3E,$87,$E6,$CF,$18
	.dc.b	$C1,$27,$8B,$67,$E2,$90,$FA,$07,$73,$17,$EE,$3F,$D9,$2E,$4F,$80
	.dc.b	$6A,$BC,$15,$F3,$A3,$27,$43,$CE,$E6,$7A,$4F,$09,$EF,$7F,$19,$ED
	.dc.b	$3E,$18,$C4,$72,$36,$7D,$79,$CD,$FB,$63,$C3,$5E,$87,$E2,$9F,$96
	.dc.b	$9F,$EC,$CF,$94,$FC,$E3,$88,$FF,$31,$C7,$E7,$3C,$4C,$11,$C1,$F8
	.dc.b	$77,$F0,$87,$E1,$7E,$1E,$46,$48,$7E,$89,$CB,$FA,$D7,$C0,$47,$FB
	.dc.b	$D7,$BD,$FE,$43,$7F,$DC,$AB,$E9,$3F,$32,$D8,$CF,$85,$83,$33,$23
	.dc.b	$27,$2F,$63,$8F,$CE,$48,$C1,$0F,$F7,$3D,$A3,$51,$89,$FD,$4F,$D2
	.dc.b	$3E,$93,$C7,$9C,$AF,$F7,$DC,$E5,$B3,$C0,$8F,$B3,$7A,$FF,$BD,$A7
	.dc.b	$83,$67,$DF,$3F,$B8,$FA,$47,$B7,$7D,$F9,$7F,$E5,$7E,$2F,$E1,$75
	.dc.b	$A8,$C7,$1F,$F1,$F8,$C4,$27,$75,$CA,$31,$5B,$BF,$07,$C8,$49,$E6
	.dc.b	$7F,$99,$70,$47,$97,$E1,$B5,$3C,$8D,$46,$31,$F2,$CF,$F1,$2F,$A4
	.dc.b	$FF,$36,$57,$41,$7F,$E8,$97,$0D,$7A,$08,$E2,$E1,$B5,$FC,$B1,$B8
	.dc.b	$FF,$3F,$6C,$FA,$7B,$3C,$5E,$44,$8F,$14,$E9,$F1,$CF,$F2,$8F,$A4
	.dc.b	$7D,$25,$24,$FA,$61,$F2,$F8,$8F,$3C,$70,$75,$3C,$E7,$A7,$CD,$75
	.dc.b	$FA,$19,$5C,$E3,$2F,$58,$CB,$D2,$3F,$9D,$F8,$BF,$E5,$7D,$7F,$A5
	.dc.b	$FD,$1B,$E9,$3F,$9F,$B6,$FE,$75,$6F,$7F,$EC,$4D,$8F,$F5,$03,$53
	.dc.b	$67,$F6,$91,$F4,$A3,$44,$F0,$BB,$F7,$78,$3A,$E5,$7D,$DB,$47,$93
	.dc.b	$B3,$33,$F7,$27,$2E,$E7,$4F,$1E,$9F,$5F,$83,$23,$04,$39,$39,$7A
	.dc.b	$47,$C6,$EA,$FE,$C3,$E9,$3E,$9A,$99,$AF,$8B,$35,$F7,$3C,$3B,$31
	.dc.b	$C3,$FD,$E3,$74,$E3,$83,$FC,$19,$8F,$EB,$AB,$FB,$EF,$D2,$B9,$3F
	.dc.b	$B8,$B7,$2F,$D0,$35,$FC,$EA,$FA,$5D,$F9,$92,$76,$FF,$7A,$FA,$47
	.dc.b	$7D,$24,$FE,$AE,$32,$FE,$B7,$E9,$7E,$CE,$5F,$E1,$09,$7C,$2F,$32
	.dc.b	$57,$EC,$FE,$97,$FB,$F9,$38,$FF,$C1,$2F,$F8,$0B,$E9,$1F,$1F,$FE
	.dc.b	$0C,$6D,$71,$AD,$99,$9F,$F0,$8F,$CA,$7E,$9A,$7F,$80,$53,$E9,$79
	.dc.b	$1A,$F0,$24,$FF,$05,$AE,$0F,$F0,$26,$E9,$FE,$2A,$BF,$F6,$34,$E6
	.dc.b	$B8,$B4,$EF,$E1,$FF,$88,$BD,$BF,$3C,$7F,$0C,$FA,$17,$B9,$FE,$26
	.dc.b	$6C,$7F,$15,$9D,$E3,$FE,$38,$6F,$47,$F9,$2D,$B2,$0E,$15,$3E,$13
	.dc.b	$C5,$B4,$9E,$CF,$86,$3F,$C6,$FE,$2B,$E3,$3F,$C7,$3F,$46,$F8,$AA
	.dc.b	$6A,$26,$1F,$CA,$EB,$83,$63,$8F,$C6,$6A,$3F,$95,$CE,$9B,$3A,$7F
	.dc.b	$9A,$7D,$23,$E8,$3C,$3F,$4B,$27,$31,$FE,$4C,$9D,$46,$08,$7F,$38
	.dc.b	$CB,$F7,$AF,$F3,$A8,$DB,$67,$91,$EE,$F1,$FF,$CA,$FF,$4B,$F7,$B8
	.dc.b	$7F,$CF,$18,$3D,$7F,$86,$6B,$9A,$63,$FD,$2E,$F8,$3F,$4D,$AA,$9F
	.dc.b	$3B,$C2,$3F,$A6,$CE,$8D,$CE,$A3,$A7,$C1,$7D,$2F,$66,$46,$08,$72
	.dc.b	$8F,$49,$F6,$3B,$7F,$CE,$3F,$4B,$F8,$29,$FE,$C0,$D7,$FA,$EB,$7A
	.dc.b	$7F,$D0,$BF,$4B,$FD,$B9,$FC,$67,$D2,$BD,$EF,$F4,$D5,$6F,$82,$F5
	.dc.b	$F6,$3F,$CF,$F8,$71,$C5,$87,$17,$F2,$D1,$98,$C4,$2F,$C9,$18,$6E
	.dc.b	$1F,$EA,$9A,$E0,$DF,$6F,$FC,$19,$7D,$A3,$C6,$FA,$1A,$E4,$F0,$C7
	.dc.b	$0B,$F0,$D7,$49,$FD,$55,$F8,$1B,$87,$FA,$E6,$B7,$7C,$70,$F1,$65
	.dc.b	$DE,$F6,$53,$D9,$7C,$67,$71,$E3,$18,$F8,$5D,$9C,$71,$F1,$3A,$C3
	.dc.b	$C0,$FA,$B0,$D6,$1D,$63,$C2,$75,$BE,$87,$5B,$E6,$74,$73,$7F,$2A
	.dc.b	$7F,$DA,$91,$F6,$5B,$19,$7D,$26,$BC,$77,$C5,$7F,$B6,$E3,$EE,$C6
	.dc.b	$1C,$57,$F4,$89,$D4,$7C,$09,$D3,$8F,$82,$FF,$43,$A6,$91,$FF,$45
	.dc.b	$BF,$37,$E5,$5F,$47,$FE,$DE,$DB,$F1,$75,$FC,$8E,$B5,$39,$DB,$CD
	.dc.b	$18,$DF,$EB,$56,$CF,$A9,$87,$48,$CC,$78,$CC,$DF,$37,$FD,$F5,$5C
	.dc.b	$DF,$B8,$73,$C3,$F3,$B7,$CE,$32,$73,$8C,$D7,$38,$FE,$74,$67,$0B
	.dc.b	$5C,$6B,$DB,$39,$3F,$A0,$46,$67,$DB,$1E,$6C,$3D,$38,$79,$6F,$8B
	.dc.b	$6B,$71,$E2,$9F,$56,$E7,$C3,$8D,$C3,$FC,$C6,$7B,$6F,$33,$FE,$6B
	.dc.b	$C7,$DD,$7B,$A4,$38,$39,$4E,$63,$75,$EE,$9C,$BF,$03,$A6,$DF,$FA
	.dc.b	$27,$E9,$5D,$F8,$8E,$B3,$AC,$55,$AB,$F2,$4E,$63,$15,$FE,$7C,$C7
	.dc.b	$FA,$75,$7F,$A3,$AF,$8E,$1D,$67,$71,$E7,$8F,$1F,$CA,$7F,$A9,$3C
	.dc.b	$5D,$44,$BF,$F4,$BC,$BE,$95,$F1,$19,$CB,$FD,$2D,$39,$8C,$63,$F3
	.dc.b	$63,$17,$F2,$F0,$EF,$87,$28,$FF,$4C,$79,$0E,$9F,$84,$D8,$FF,$53
	.dc.b	$8F,$A4,$61,$AA,$93,$A9,$FA,$39,$FF,$82,$6F,$93,$FE,$AB,$9E,$5C
	.dc.b	$5C,$23,$15,$C1,$57,$0C,$FF,$0D,$7C,$0F,$F5,$77,$D2,$BD,$BF,$B0
	.dc.b	$D8,$F1,$71,$E2,$64,$79,$5B,$87,$81,$B7,$DB,$1E,$53,$FE,$B6,$C2
	.dc.b	$CE,$6B,$F0,$D7,$FE,$58,$3F,$D8,$05,$AF,$5F,$81,$AE,$89,$3F,$2B
	.dc.b	$E9,$7F,$C9,$EF,$08,$FC,$9A,$8C,$D7,$FB,$1A,$3A,$61,$F0,$A7,$2F
	.dc.b	$E8,$E3,$27,$F8,$DC,$E5,$E5,$FE,$6F,$E5,$FF,$1F,$1F,$C0,$9B,$7F
	.dc.b	$B3,$D3,$C1,$3F,$ED,$99,$CC,$9D,$C7,$E0,$7C,$5E,$9C,$9E,$8B,$FF
	.dc.b	$0D,$4F,$18,$C4,$EA,$3F,$08,$FE,$B7,$E8,$DE,$85,$3F,$61,$F4,$AF
	.dc.b	$A5,$C5,$FF,$BA,$EB,$6D,$F7,$73,$A7,$FB,$AA,$7F,$D3,$33,$93,$FD
	.dc.b	$E5,$F4,$BF,$8A,$71,$3F,$EF,$08,$C3,$C7,$6F,$E0,$1D,$3F,$EB,$CA
	.dc.b	$F0,$46,$3C,$C9,$1F,$EF,$D9,$FF,$2C,$36,$38,$FA,$3F,$C2,$7E,$97
	.dc.b	$FC,$C9,$3F,$E6,$9E,$3F,$E3,$1F,$02,$F7,$35,$EA,$FF,$0F,$8C,$56
	.dc.b	$CF,$F3,$87,$37,$F9,$C7,$E8,$3F,$8B,$BF,$E7,$3E,$6B,$BA,$6C,$61
	.dc.b	$B5,$CE,$7E,$D1,$B7,$15,$CD,$B1,$91,$97,$F3,$59,$D4,$62,$7C,$AE
	.dc.b	$E7,$4F,$E3,$27,$FD,$11,$F4,$1E,$63,$CF,$AD,$BE,$F1,$6B,$A1,$E6
	.dc.b	$9C,$5F,$F3,$FF,$9D,$7A,$3F,$CD,$5B,$87,$F2,$4F,$A0,$F9,$38,$72
	.dc.b	$13,$A7,$0D,$BF,$E4,$93,$FC,$36,$BF,$B2,$CE,$5F,$F5,$07,$12,$75
	.dc.b	$7F,$A3,$F3,$BF,$AD,$57,$CA,$7E,$FC,$7A,$B4,$FF,$0C,$CF,$D8,$9C
	.dc.b	$E3,$F2,$DE,$35,$B7,$33,$F8,$07,$33,$BE,$0E,$53,$CD,$AF,$85,$ED
	.dc.b	$F4,$12,$3C,$9F,$0D,$70,$ED,$36,$FC,$53,$6B,$E0,$FC,$57,$CD,$3F
	.dc.b	$B2,$F9,$9F,$EC,$0F,$A0,$C4,$FE,$2F,$1F,$9B,$5F,$D9,$7E,$83,$E8
	.dc.b	$E3,$F5,$E3,$1C,$FF,$07,$E2,$66,$1F,$1F,$E8,$2E,$3F,$06,$6E,$1F
	.dc.b	$16,$6D,$7C,$39,$B5,$F4,$66,$F0,$7F,$91,$BE,$4F,$E7,$7A,$3F,$3D
	.dc.b	$B8,$7D,$56,$DF,$CF,$6E,$3F,$E0,$BF,$A0,$FF,$73,$D7,$B5,$B5,$F9
	.dc.b	$4D,$FF,$DC,$3E,$2F,$F8,$16,$73,$7C,$71,$FE,$F3,$F4,$0F,$91,$D4
	.dc.b	$61,$FC,$17,$CF,$FE,$0F,$E8,$2E,$9F,$89,$6E,$9F,$83,$6E,$3F,$87
	.dc.b	$6C,$FE,$15,$B3,$F9,$37,$F1,$DC,$3F,$8F,$F5,$2D,$7E,$55,$B5,$F8
	.dc.b	$F6,$DF,$E4,$DF,$F1,$07,$CC,$FF,$3A,$9F,$F3,$83,$5F,$F2,$48,$70
	.dc.b	$FF,$93,$1A,$87,$D0,$3E,$83,$FB,$DC,$7F,$9E,$E3,$1E,$B7,$F9,$E1
	.dc.b	$75,$7F,$83,$27,$55,$C1,$F0,$F9,$EE,$9F,$F2,$DE,$7F,$E4,$38,$7C
	.dc.b	$EE,$35,$FF,$2F,$C6,$2F,$F3,$0F,$16,$DE,$DF,$44,$6D,$FE,$14,$DB
	.dc.b	$ED,$3B,$9F,$F7,$C3,$67,$55,$E0,$BC,$4C,$AF,$0C,$6C,$C3,$DA,$30
	.dc.b	$E8,$F2,$90,$F1,$C6,$9E,$06,$6F,$E1,$46,$5F,$3C,$66,$BC,$8F,$53
	.dc.b	$89,$23,$8B,$97,$8B,$98,$E2,$F8,$8E,$87,$C8,$AF,$04,$6A,$78,$91
	.dc.b	$C1,$27,$81,$C8,$5F,$CA,$7E,$9D,$FC,$6A,$E5,$18,$3C,$D5,$F7,$AF
	.dc.b	$4F,$58,$E7,$58,$7B,$9E,$92,$57,$47,$ED,$9D,$0E,$C7,$C1,$38,$BF
	.dc.b	$19,$E3,$3E,$17,$D5,$7E,$97,$C2,$62,$FE,$2B,$D2,$BC,$13,$F9,$38
	.dc.b	$E6,$66,$BA,$3E,$13,$99,$A3,$CE,$66,$3C,$04,$AF,$81,$1C,$2B,$D0
	.dc.b	$78,$23,$C7,$19,$AE,$F3,$E4,$BC,$DF,$96,$31,$5F,$14,$F0,$B8,$7C
	.dc.b	$91,$BA,$E6,$7D,$37,$C9,$1F,$1A,$BA,$DF,$9A,$7D,$31,$63,$C2,$D9
	.dc.b	$F0,$4F,$27,$E0,$47,$27,$A1,$A9,$E9,$19,$8D,$99,$36,$7A,$23,$6E
	.dc.b	$4D,$99,$37,$7F,$2A,$FD,$B1,$F8,$47,$DD,$5E,$78,$F8,$6D,$9E,$11
	.dc.b	$F7,$4E,$04,$3E,$91,$93,$E6,$CF,$67,$24,$BF,$9C,$F1,$7E,$4B,$F8
	.dc.b	$F3,$E3,$D7,$86,$7C,$B1,$93,$CA,$7E,$63,$CA,$FE,$35,$F2,$F2,$3F
	.dc.b	$93,$C7,$F1,$D3,$0F,$F2,$48,$C3,$FE,$03,$F9,$DF,$E8,$8F,$F2,$EF
	.dc.b	$1B,$F9,$21,$FC,$9E,$33,$1F,$E3,$56,$BF,$D3,$1A,$FF,$49,$0E,$1F
	.dc.b	$E9,$8D,$5F,$3A,$F3,$3F,$9A,$2F,$46,$98,$89,$3F,$E1,$99,$D3,$67
	.dc.b	$2F,$F9,$26,$3A,$39,$21,$99,$EF,$F0,$92,$3A,$3E,$2F,$5F,$11,$CE
	.dc.b	$2E,$ED,$7D,$D8,$2F,$FA,$67,$CE,$BC,$1F,$5F,$8D,$32,$FF,$18,$4E
	.dc.b	$5B,$3C,$08,$FB,$79,$FF,$AB,$7C,$EB,$E7,$69,$27,$FC,$75,$0F,$8C
	.dc.b	$C4,$7F,$9A,$63,$83,$A9,$E7,$3A,$D7,$D5,$3F,$DF,$EB,$6F,$1A,$DB
	.dc.b	$DB,$E4,$3C,$7F,$35,$23,$04,$3C,$0E,$5F,$14,$7F,$6F,$EB,$FF,$16
	.dc.b	$47,$F9,$47,$E7,$1F,$3B,$FD,$CD,$B1,$FE,$55,$9D,$BF,$E8,$36,$E5
	.dc.b	$E8,$8F,$F0,$7F,$89,$D4,$E5,$DB,$CF,$AE,$4F,$77,$B7,$5E,$23,$FD
	.dc.b	$1D,$8A,$71,$87,$58,$56,$39,$BF,$39,$1B,$BE,$87,$AE,$7D,$5F,$07
	.dc.b	$EB,$F8,$29,$3E,$F8,$CB,$F0,$63,$2F,$8A,$3E,$37,$5F,$F8,$57,$CE
	.dc.b	$7C,$ED,$33,$5F,$16,$6B,$F1,$5E,$1D,$BC,$7B,$79,$76,$4A,$FF,$40
	.dc.b	$60,$9D,$46,$27,$4F,$FA,$C6,$74,$F0,$9C,$ED,$F4,$67,$F4,$13,$93
	.dc.b	$FD,$19,$F3,$AE,$6F,$F1,$73,$63,$31,$8F,$9D,$CF,$91,$27,$8B,$FE
	.dc.b	$35,$F9,$CE,$B1,$FE,$C3,$6E,$3F,$81,$6F,$1F,$F0,$F6,$CF,$F9,$67
	.dc.b	$E7,$7F,$96,$A8,$7C,$EF,$52,$47,$FA,$8F,$E7,$5C,$7F,$E4,$56,$C6
	.dc.b	$63,$07,$F9,$27,$E7,$1F,$07,$4C,$CE,$CF,$F5,$74,$69,$FF,$58,$7A
	.dc.b	$9F,$B7,$73,$F3,$BC,$CF,$F2,$A4,$66,$3F,$DC,$4D,$D7,$F4,$0D,$D7
	.dc.b	$F9,$73,$67,$FC,$F1,$7F,$E6,$FA,$DF,$87,$D0,$93,$E4,$F1,$3F,$CC
	.dc.b	$5F,$3B,$E5,$8F,$F7,$87,$CC,$3D,$5F,$F3,$33,$63,$FD,$91,$3B,$C3
	.dc.b	$F6,$4D,$CF,$FA,$03,$67,$FD,$14,$D5,$4F,$77,$C3,$B4,$9E,$AF,$82
	.dc.b	$3F,$CE,$DE,$33,$E7,$7F,$CF,$1F,$34,$F8,$4A,$6A,$26,$1F,$ED,$DA
	.dc.b	$E0,$D8,$E3,$F0,$FA,$86,$BF,$C2,$EB,$71,$8D,$7F,$AD,$46,$27,$F8
	.dc.b	$86,$2B,$FF,$45,$7C,$E3,$E5,$BC,$BF,$3B,$32,$FF,$77,$4E,$5B,$3C
	.dc.b	$08,$FF,$BC,$EF,$F8,$63,$E1,$3F,$D6,$53,$C5,$C0,$71,$87,$F2,$36
	.dc.b	$E5,$FD,$C6,$B9,$CF,$F2,$7F,$9D,$FC,$35,$A7,$FC,$EE,$F8,$7D,$8F
	.dc.b	$F5,$24,$E8,$9C,$B8,$3E,$8F,$C9,$DA,$FF,$63,$61,$D6,$71,$1E,$6C
	.dc.b	$15,$D3,$DD,$FB,$3E,$CC,$AE,$39,$FF,$81,$F2,$56,$E7,$27,$FB,$4E
	.dc.b	$32,$E0,$38,$7C,$1D,$9F,$EB,$DF,$9D,$FC,$7C,$6D,$AF,$CA,$6E,$9F
	.dc.b	$D2,$FE,$77,$F9,$23,$FE,$E1,$F9,$C7,$CE,$7F,$86,$EB,$79,$DE,$AE
	.dc.b	$C7,$FB,$2F,$0E,$38,$B0,$E2,$FF,$BA,$E3,$31,$88,$5F,$92,$30,$DC
	.dc.b	$3F,$C5,$55,$C1,$BE,$57,$E6,$2F,$B4,$78,$DF,$43,$5C,$9E,$18,$E1
	.dc.b	$7E,$1A,$E9,$3F,$EF,$FB,$F0,$37,$0F,$F1,$D5,$6E,$F8,$E1,$E2,$CB
	.dc.b	$BD,$EC,$A7,$B2,$F8,$CE,$E3,$C6,$31,$F0,$BB,$38,$E3,$E2,$75,$87
	.dc.b	$81,$F5,$61,$AC,$3A,$C7,$84,$EB,$7D,$0E,$B7,$CC,$E8,$E6,$FE,$54
	.dc.b	$FF,$95,$CE,$F7,$5E,$E5,$9F,$BA,$ED,$FF,$20,$7C,$5E,$E3,$D9,$E6
	.dc.b	$23,$13,$A9,$E5,$39,$78,$0F,$9D,$AD,$32,$E0,$8F,$CA,$F8,$87,$CD
	.dc.b	$7F,$8E,$B6,$FC,$64,$8E,$4F,$E3,$E7,$7B,$FD,$69,$D5,$FD,$89,$CC
	.dc.b	$7D,$63,$27,$7B,$ED,$BF,$C5,$AE,$11,$89,$F9,$51,$8B,$ED,$1E,$2A
	.dc.b	$FF,$42,$5F,$6C,$3C,$F3,$93,$D1,$19,$9E,$D1,$E1,$AE,$77,$D6,$3C
	.dc.b	$C7,$11,$5C,$5F,$19,$C4,$DC,$71,$BF,$0B,$6F,$CB,$87,$F9,$93,$4F
	.dc.b	$F2,$76,$1F,$11,$FF,$28,$61,$F1,$30,$F9,$95,$BC,$78,$9F,$C5,$6F
	.dc.b	$FC,$16,$D8,$D5,$7B,$B8,$2B,$E8,$E9,$CE,$2F,$9D,$FE,$9D,$87,$39
	.dc.b	$CC,$62,$BF,$A1,$E3,$FD,$46,$BF,$93,$DF,$1E,$CE,$B3,$9A,$FE,$99
	.dc.b	$39,$AF,$35,$68,$F2,$FC,$87,$FA,$CF,$C4,$C4,$E8,$FE,$67,$0F,$9C
	.dc.b	$7C,$EE,$B1,$FE,$B1,$39,$8C,$63,$FE,$8E,$8C,$5F,$3A,$DD,$F2,$8F
	.dc.b	$E7,$9E,$3F,$F4,$26,$F5,$7C,$85,$F3,$6C,$3E,$4D,$6F,$0F,$EC,$AD
	.dc.b	$C7,$FD,$3B,$7C,$38,$77,$8F,$3C,$7E,$E4,$FF,$B5,$EF,$83,$FA,$93
	.dc.b	$81,$E2,$3F,$A9,$98,$C3,$FB,$9F,$CE,$BD,$9F,$B6,$D9,$DD,$FE,$0A
	.dc.b	$B7,$5A,$3C,$45,$9D,$B6,$3F,$C0,$51,$F0,$23,$19,$7B,$2B,$B5,$7F
	.dc.b	$81,$2B,$FD,$A4,$7F,$65,$2B,$CA,$3F,$B9,$C6,$79,$FF,$22,$EA,$BE
	.dc.b	$3F,$CE,$FF,$97,$23,$47,$C0,$71,$7D,$EB,$EA,$46,$2B,$FD,$69,$5C
	.dc.b	$CF,$95,$19,$7F,$8A,$53,$E4,$7F,$35,$5F,$EB,$B6,$CF,$D1,$C1,$7F
	.dc.b	$A6,$9C,$DF,$F1,$D6,$D7,$E1,$3C,$4D,$71,$FD,$D6,$C7,$D2,$6D,$7C
	.dc.b	$37,$F6,$3F,$36,$F3,$E9,$FB,$4F,$9D,$7C,$EE,$30,$FF,$08,$56,$DB
	.dc.b	$EE,$E7,$1F,$F1,$AC,$A3,$A9,$FB,$EF,$9D,$71,$E2,$FF,$C3,$51,$87
	.dc.b	$AF,$A3,$FC,$13,$9F,$FD,$A9,$18,$3F,$C4,$0E,$31,$FF,$21,$B7,$8B
	.dc.b	$FD,$C8,$FF,$99,$67,$7F,$05,$CF,$D1,$FE,$1F,$F3,$BF,$E7,$19,$FF
	.dc.b	$15,$7C,$03,$DC,$CF,$AB,$FC,$52,$31,$3F,$E3,$2E,$6F,$E9,$9F,$3B
	.dc.b	$FC,$69,$FE,$9D,$CF,$76,$4D,$8C,$36,$B9,$4E,$9B,$3A,$69,$A9,$E1
	.dc.b	$3F,$E6,$3A,$FE,$B9,$F3,$BD,$CF,$06,$96,$FA,$C5,$AE,$07,$C3,$2B
	.dc.b	$FC,$C3,$E7,$5E,$8D,$3F,$90,$FC,$EF,$C0,$AF,$F3,$7C,$EA,$FF,$CA
	.dc.b	$0D,$8F,$E3,$F3,$FC,$21,$FE,$E1,$19,$E2,$FE,$41,$39,$CB,$FC,$09
	.dc.b	$7F,$27,$CE,$FA,$58,$7B,$54,$67,$7D,$34,$FA,$51,$A6,$C6,$9A,$68
	.dc.b	$E1,$3D,$7E,$77,$C0,$78,$7E,$2E,$CF,$48,$B3,$E3,$9B,$97,$92,$3D
	.dc.b	$BF,$15,$F2,$4F,$EB,$7E,$53,$DD,$7C,$8E,$72,$F8,$F5,$FD,$73,$E7
	.dc.b	$7E,$66,$3C,$F1,$70,$2B,$5C,$FF,$17,$E2,$66,$3F,$03,$E7,$6E,$1F
	.dc.b	$1E,$6E,$3E,$5E,$2F,$F1,$85,$62,$FF,$83,$71,$2F,$18,$7F,$11,$85
	.dc.b	$62,$FF,$C7,$1F,$25,$F5,$7D,$0F,$AC,$DC,$7E,$8B,$70,$FB,$0D,$CB
	.dc.b	$D9,$F3,$BF,$EE,$9A,$FF,$4B,$36,$BF,$D7,$9C,$5F,$1F,$C7,$FF,$00
	.dc.b	$CE,$6F,$8E,$3F,$DC,$BE,$75,$F2,$3A,$8C,$3F,$89,$F9,$DF,$C5,$7C
	.dc.b	$ED,$D3,$F0,$AD,$D3,$F1,$8D,$C7,$E3,$70,$23,$15,$F1,$9B,$5F,$15
	.dc.b	$FC,$B7,$17,$E5,$FD,$2B,$5F,$92,$6D,$7E,$65,$B8,$76,$7F,$C3,$BF
	.dc.b	$38,$F9,$CF,$E4,$35,$FE,$6D,$8C,$79,$5F,$E6,$95,$D9,$FD,$F6,$75
	.dc.b	$5C,$1F,$F3,$B7,$D7,$F3,$7F,$CE,$0D,$7F,$CF,$0D,$77,$1F,$E6,$EF
	.dc.b	$43,$F1,$D1,$FE,$C9,$5F,$3B,$FD,$53,$3F,$CA,$46,$27,$5D,$79,$D7
	.dc.b	$FA,$CC,$F2,$9F,$F7,$9B,$B6,$DE,$DF,$F3,$94,$6E,$FA,$DE,$65,$3C
	.dc.b	$67,$C3,$59,$8E,$17,$E0,$BC,$4C,$BF,$29,$0F,$09,$ED,$79,$44,$8C
	.dc.b	$10,$F4,$B9,$78,$92,$FC,$EF,$57,$CA,$F1,$72,$F2,$31,$5F,$1A,$BA
	.dc.b	$4E,$8E,$33,$D6,$7D,$8F,$57,$83,$C8,$F9,$35,$CA,$33,$1C,$4C,$9C
	.dc.b	$4F,$9F,$5F,$15,$F2,$12,$BC,$24,$9F,$0C,$65,$EB,$19,$8E,$A6,$4E
	.dc.b	$A7,$94,$FB,$35,$E5,$72,$79,$4C,$9E,$53,$EF,$D7,$96,$7E,$25,$78
	.dc.b	$9F,$04,$E8,$E1,$7F,$82,$7A,$C6,$A3,$11,$D8,$F0,$3F,$8B,$9E,$84
	.dc.b	$8E,$2F,$B8,$C3,$C6,$7A,$57,$AE,$31,$3B,$78,$9F,$12,$7D,$2E,$0F
	.dc.b	$41,$F1,$CE,$26,$6B,$C8,$F1,$9F,$0C,$62,$38,$CF,$47,$89,$89,$F1
	.dc.b	$38,$9F,$82,$71,$25,$FA,$5C,$5D,$8F,$A1,$5C,$A7,$8C,$62,$BC,$CF
	.dc.b	$1A,$C4,$77,$8B,$3D,$A3,$EA,$CE,$27,$84,$7E,$16,$7A,$B6,$7C,$4D
	.dc.b	$3E,$EC,$F4,$8C,$C7,$DA,$6D,$FD,$48,$D9,$C4,$F9,$D5,$F0,$A7,$C5
	.dc.b	$1E,$43,$C5,$5E,$47,$F0,$AD,$AF,$C5,$36,$FE,$23,$67,$E1,$9C,$CE
	.dc.b	$25,$9E,$2F,$F2,$BF,$2B,$F8,$C7,$CB,$48,$C1,$0F,$E4,$EE,$5F,$E5
	.dc.b	$31,$FC,$8F,$E7,$5F,$3D,$FC,$87,$C7,$94,$18,$7F,$82,$6B,$71,$98
	.dc.b	$C5,$73,$24,$7F,$3F,$F7,$FF,$98,$7C,$F3,$CE,$BE,$9C,$AE,$5B,$7F
	.dc.b	$A5,$AB,$51,$83,$C4,$E1,$F1,$C6,$E7,$94,$60,$DD,$78,$0C,$E3,$FE
	.dc.b	$21,$BF,$F4,$34,$F5,$0F,$58,$D7,$C2,$FA,$09,$3B,$97,$B5,$88,$F0
	.dc.b	$1C,$A7,$15,$DE,$7F,$A8,$7C,$F3,$C4,$F8,$7E,$54,$C3,$C5,$39,$6C
	.dc.b	$F0,$23,$E1,$EA,$FE,$B5,$F3,$CF,$9E,$A4,$9F,$64,$3E,$3B,$11,$E6
	.dc.b	$8E,$0E,$A7,$9C,$EB,$3F,$F2,$35,$6A,$B8,$56,$9E,$BF,$23,$E2,$F9
	.dc.b	$A9,$18,$21,$DD,$CB,$E1,$8F,$EE,$5D,$DF,$E2,$E8,$FA,$5F,$3A,$F9
	.dc.b	$EF,$EE,$CD,$8F,$AD,$3B,$7E,$7B,$72,$FF,$2E,$37,$97,$FD,$BE,$D9
	.dc.b	$FC,$D5,$7F,$38,$1D,$5D,$89,$1E,$2E,$BA,$FD,$C9,$DB,$E6,$9F,$2C
	.dc.b	$EB,$0F,$F3,$65,$6A,$B8,$56,$9F,$6F,$47,$B3,$E0,$E4,$60,$87,$77
	.dc.b	$2F,$86,$3E,$3F,$4F,$F8,$67,$D0,$7C,$F5,$33,$5F,$1A,$6B,$F1,$9E
	.dc.b	$1D,$7C,$7A,$B9,$75,$4A,$FD,$16,$09,$D4,$62,$74,$FE,$7A,$74,$F0
	.dc.b	$9C,$EB,$FE,$8D,$9F,$D2,$CE,$4F,$D7,$7C,$F3,$93,$FC,$64,$D8,$CC
	.dc.b	$63,$E7,$B3,$E4,$49,$E2,$FF,$8E,$7E,$77,$AC,$7E,$C1,$B8,$FF,$A7
	.dc.b	$5B,$CB,$FC,$66,$51,$87,$F8,$E7,$CF,$7F,$7A,$87,$CF,$75,$24,$FF
	.dc.b	$05,$F9,$E7,$1F,$F9,$25,$B1,$98,$C1,$FE,$4D,$F9,$D7,$C2,$D3,$33
	.dc.b	$B3,$F8,$84,$69,$FE,$25,$EB,$7E,$DD,$CF,$CF,$73,$3F,$CB,$11,$98
	.dc.b	$FE,$1E,$DD,$7F,$D7,$8D,$D7,$E0,$36,$7F,$9D,$5F,$F9,$EE,$B7,$E0
	.dc.b	$F4,$24,$F9,$3C,$2F,$F3,$37,$BB,$E5,$8F,$E3,$BF,$30,$F5,$FF,$CD
	.dc.b	$4D,$8F,$E5,$53,$BC,$3F,$D9,$CD,$CF,$E9,$36,$7F,$A6,$35,$53,$E0
	.dc.b	$3C,$3B,$49,$EA,$F8,$23,$FC,$F5,$E2,$3E,$7B,$FC,$F9,$F3,$8F,$85
	.dc.b	$A6,$A2,$61,$FC,$FE,$B8,$36,$38,$FC,$3E,$A3,$F9,$DC,$E9,$B3,$A7
	.dc.b	$FA,$47,$CE,$BE,$5B,$C3,$F3,$D2,$73,$1F,$E8,$F9,$D4,$60,$87,$F4
	.dc.b	$8C,$BF,$84,$70,$7F,$BD,$67,$73,$FD,$31,$B1,$9A,$F7,$F8,$FF,$E9
	.dc.b	$BF,$9D,$FB,$D6,$C3,$FD,$3A,$FF,$5D,$6B,$FD,$5A,$36,$4E,$5E,$47
	.dc.b	$CA,$F9,$EB,$3F,$D9,$67,$4F,$38,$F0,$7B,$BF,$43,$D7,$93,$B8,$CB
	.dc.b	$48,$7B,$FB,$3F,$D6,$5F,$3D,$F8,$08,$FE,$E4,$D7,$FB,$5B,$75,$74
	.dc.b	$7E,$EF,$E7,$BF,$8B,$47,$F8,$57,$E7,$1F,$3B,$FB,$FB,$E1,$77,$AB
	.dc.b	$B9,$FE,$C1,$C3,$8E,$2C,$38,$BF,$E1,$F8,$CC,$62,$17,$E5,$8C,$37
	.dc.b	$0F,$27,$95,$FE,$26,$AE,$91,$8C,$BC,$CF,$8E,$29,$FE,$2D,$AE,$4F
	.dc.b	$86,$B9,$5F,$87,$0F,$1D,$F1,$AE,$A7,$1C,$7F,$8F,$56,$EF,$8D,$78
	.dc.b	$6B,$67,$AA,$B7,$1C,$EB,$86,$3E,$4A,$61,$E0,$7B,$19,$C7,$CE,$EB
	.dc.b	$1E,$EF,$A6,$B5,$8F,$68,$E2,$78,$AF,$A3,$C6,$76,$C3,$55,$D1,$D1
	.dc.b	$FC,$EC,$EC,$53,$26,$67,$EE,$B7,$CB,$7C,$1E,$E3,$DB,$18,$79,$4F
	.dc.b	$F8,$F2,$74,$D8,$E0,$3E,$7A,$B4,$CB,$82,$3F,$43,$E2,$5F,$37,$F9
	.dc.b	$A8,$FA,$67,$F9,$B6,$7E,$84,$EB,$7F,$A5,$39,$39,$57,$1D,$BF,$29
	.dc.b	$19,$7F,$2D,$5C,$63,$A9,$E5,$8C,$CF,$1D,$FE,$46,$1E,$08,$C5,$7C
	.dc.b	$D8,$C6,$1E,$28,$E5,$5F,$DA,$F0,$ED,$87,$8A,$72,$7F,$A1,$E3,$33
	.dc.b	$DA,$3D,$35,$CE,$FA,$C7,$99,$B7,$CD,$B5,$C6,$3D,$35,$C3,$5F,$95
	.dc.b	$6C,$3E,$04,$62,$3E,$25,$F9,$F1,$FA,$13,$C2,$B7,$7E,$BE,$0C,$BB
	.dc.b	$3F,$D0,$91,$FE,$00,$6F,$46,$B1,$FF,$51,$FC,$F3,$BB,$11,$E3,$9D
	.dc.b	$62,$AD,$5F,$23,$FD,$39,$19,$BE,$38,$7F,$4B,$9C,$D7,$0C,$3F,$D1
	.dc.b	$B3,$6F,$D5,$7E,$3F,$95,$FF,$56,$78,$9C,$A2,$63,$FE,$A9,$97,$CE
	.dc.b	$BE,$7B,$39,$7F,$A9,$E7,$31,$8C,$7F,$2D,$18,$BF,$87,$87,$7C,$39
	.dc.b	$47,$FA,$A7,$C7,$74,$FC,$26,$C7,$FA,$E4,$7C,$EB,$0D,$54,$9D,$4F
	.dc.b	$D4,$CF,$FD,$A1,$7C,$9F,$F5,$E4,$F2,$E1,$E1,$18,$AE,$0A,$B8,$67
	.dc.b	$FA,$FB,$E0,$7F,$AF,$FE,$79,$EC,$FD,$86,$C7,$8B,$8F,$13,$23,$CA
	.dc.b	$DC,$3C,$0D,$BE,$D8,$F2,$9F,$F6,$1E,$16,$73,$5F,$86,$BF,$F6,$49
	.dc.b	$FE,$D1,$2D,$7A,$FB,$B5,$D5,$27,$E5,$7C,$F7,$F9,$31,$E1,$1F,$93
	.dc.b	$51,$9A,$FF,$69,$C7,$4C,$3E,$14,$E5,$FD,$4C,$64,$FF,$7F,$1C,$BC
	.dc.b	$AF,$CD,$FC,$97,$E3,$E3,$FC,$D4,$DB,$FE,$E7,$A7,$82,$7F,$DD,$53
	.dc.b	$99,$3B,$8F,$C0,$FC,$57,$4E,$4F,$45,$A3,$13,$F8,$26,$DF,$DD,$7F
	.dc.b	$5B,$F3,$8F,$42,$9F,$B0,$F9,$E7,$CF,$63,$0F,$F7,$E5,$6D,$BE,$F6
	.dc.b	$71,$FF,$7B,$CA,$3A,$9F,$BA,$F9,$EF,$C5,$57,$01,$5F,$EF,$E8,$CB
	.dc.b	$CB,$6F,$E0,$3D,$3F,$EC,$CA,$F0,$C6,$3C,$CC,$FA,$32,$79,$F1,$7F
	.dc.b	$1A,$BE,$18,$23,$15,$FC,$33,$DF,$7B,$74,$FE,$1F,$F3,$DF,$E6,$7C
	.dc.b	$79,$9F,$E9,$F9,$DB,$8C,$3E,$98,$F9,$DF,$E3,$5C,$DE,$C6,$C7,$F9
	.dc.b	$A6,$78,$4E,$5B,$1C,$5B,$3C,$31,$E2,$C8,$C9,$C2,$31,$3E,$D8,$FF
	.dc.b	$5A,$7C,$D3,$CF,$FE,$52,$D8,$FE,$31,$F3,$BE,$AA,$E0,$7D,$29,$D3
	.dc.b	$80,$E3,$5F,$E3,$33,$FC,$0A,$BB,$DF,$F2,$78,$C5,$FE,$5B,$CD,$FF
	.dc.b	$26,$D7,$C0,$7E,$64,$78,$33,$FF,$04,$CF,$C8,$9C,$E1,$FD,$EE,$73
	.dc.b	$3C,$7D,$8D,$3B,$78,$7C,$EC,$8E,$DF,$11,$71,$ED,$8A,$B1,$8F,$58
	.dc.b	$56,$2F,$F2,$78,$FF,$87,$2F,$18,$7A,$EA,$BF,$13,$E4,$BE,$69,$FD
	.dc.b	$67,$CB,$7B,$8F,$8F,$CE,$5F,$87,$BE,$0F,$F5,$6F,$9D,$FA,$5A,$FF
	.dc.b	$70,$E1,$65,$FD,$7A,$73,$5E,$4D,$9F,$11,$F1,$CF,$91,$F3,$B6,$7E
	.dc.b	$74,$5C,$BE,$0C,$DC,$BE,$07,$C8,$7C,$FF,$47,$E8,$37,$0F,$AC,$DB
	.dc.b	$F9,$2D,$AE,$33,$F6,$EB,$8B,$4F,$77,$CE,$FF,$B6,$DB,$1F,$17,$7F
	.dc.b	$E1,$18,$7E,$36,$B4,$D9,$FF,$07,$3F,$C1,$23,$2F,$F9,$55,$AF,$F0
	.dc.b	$80,$E1,$FC,$87,$CD,$FE,$2F,$E6,$FF,$18,$DC,$3F,$24,$D9,$FC,$2B
	.dc.b	$71,$FC,$27,$6F,$E4,$FD,$2F,$CA,$36,$FF,$34,$DB,$FC,$6B,$6B,$8C
	.dc.b	$FE,$7E,$B8,$B4,$FC,$17,$D8,$3E,$C7,$DF,$82,$31,$3F,$E6,$B6,$C7
	.dc.b	$F9,$CD,$AF,$F9,$DF,$E7,$5F,$39,$FD,$D6,$3F,$CF,$11,$8F,$57,$FC
	.dc.b	$E8,$FE,$7B,$C4,$B5,$BA,$FF,$40,$79,$BF,$90,$7F,$2C,$BE,$73,$E7
	.dc.b	$6B,$FE,$44,$8C,$4E,$BA,$73,$A7,$D3,$9E,$53,$FC,$25,$DB,$6F,$6F
	.dc.b	$F4,$08,$DD,$F5,$F1,$F3,$3F,$D6,$7E,$56,$D1,$B3,$27,$CF,$1A,$30
	.dc.b	$F8,$63,$0E,$8F,$6B,$AA,$F6,$4E,$A7,$81,$A3,$81,$EB,$78,$61,$EF
	.dc.b	$AE,$15,$CA,$B8,$1C,$8E,$07,$27,$DC,$E9,$E0,$E8,$E0,$E9,$C3,$A3
	.dc.b	$04,$9F,$A1,$5E,$27,$C9,$3D,$5C,$BE,$17,$2F,$67,$2F,$67,$33,$D6
	.dc.b	$3E,$B4,$F5,$7E,$C5,$7C,$27,$27,$53,$27,$84,$C9,$D4,$C9,$E1,$33
	.dc.b	$3E,$DA,$F0,$92,$32,$F9,$A3,$C9,$1F,$E1,$D9,$F2,$3F,$87,$3C,$06
	.dc.b	$4F,$21,$93,$C0,$64,$F2,$19,$8E,$B6,$9E,$AF,$E4,$67,$AB,$97,$AB
	.dc.b	$97,$AB,$97,$AB,$99,$EB,$3D,$2F,$E5,$57,$8E,$7E,$13,$F0,$0F,$8E
	.dc.b	$F9,$23,$C7,$3C,$2B,$A5,$FC,$53,$E4,$5F,$37,$0F,$33,$07,$47,$2F
	.dc.b	$C1,$78,$C7,$A6,$FC,$73,$93,$81,$0C,$12,$38,$4F,$2B,$F3,$CF,$C8
	.dc.b	$6B,$F1,$9B,$5E,$D8,$F9,$77,$ED,$9C,$D7,$39,$C9,$CC,$A7,$32,$D7
	.dc.b	$37,$E8,$D7,$37,$33,$CD,$C9,$CC,$A7,$32,$CF,$33,$E1,$DF,$BA,$3E
	.dc.b	$13,$F4,$9F,$9C,$E5,$F9,$E7,$C5,$BF,$B1,$7F,$81,$8F,$C5,$B5,$FC
	.dc.b	$4B,$6B,$F3,$FF,$38,$F9,$CF,$E4,$7E,$34,$BF,$F0,$64,$E5,$B3,$97
	.dc.b	$F9,$94,$FF,$2C,$9D,$47,$F8,$F7,$E7,$1E,$57,$F3,$07,$FC,$F1,$1F
	.dc.b	$E1,$76,$BB,$8D,$1B,$F3,$A4,$6E,$39,$7C,$24,$8E,$4F,$5F,$5F,$11
	.dc.b	$E1,$8B,$9B,$7F,$EF,$98,$AF,$31,$8A,$B7,$FE,$88,$AD,$3F,$E2,$DF
	.dc.b	$9C,$78,$1E,$BF,$1A,$4E,$63,$FA,$6C,$EA,$3F,$C1,$47,$F8,$2D,$C3
	.dc.b	$80,$E1,$DB,$5F,$F0,$97,$CE,$3E,$73,$FA,$CB,$63,$FC,$80,$D7,$FC
	.dc.b	$D8,$BE,$2F,$73,$27,$6F,$28,$E7,$F2,$1D,$3E,$62,$57,$48,$CB,$DA
	.dc.b	$32,$F6,$8C,$BE,$08,$CE,$FF,$3E,$3F,$D6,$38,$2B,$FB,$8F,$9A,$F9
	.dc.b	$EF,$EE,$7F,$38,$F9,$DA,$6A,$26,$1F,$E5,$DA,$E0,$D8,$E3,$DB,$CB
	.dc.b	$AA,$4F,$F9,$8F,$0F,$D3,$1F,$A9,$4E,$23,$71,$FE,$93,$8D,$0F,$84
	.dc.b	$F2,$FC,$1C,$9E,$91,$E8,$30,$F1,$8C,$3A,$3C,$4E,$B5,$F8,$91,$E2
	.dc.b	$9F,$87,$D3,$FE,$14,$F3,$9F,$3B,$4C,$D7,$C1,$9A,$FF,$AA,$75,$FD
	.dc.b	$4E,$1D,$2F,$F5,$F3,$FE,$83,$5D,$1D,$7A,$A4,$E6,$3C,$06,$1F,$F5
	.dc.b	$A4,$FE,$81,$B1,$E5,$F1,$64,$75,$7D,$1F,$5B,$88,$E3,$17,$0F,$24
	.dc.b	$DE,$CF,$F3,$66,$1F,$E4,$59,$F4,$57,$2B,$FF,$4D,$7C,$F3,$E7,$BF
	.dc.b	$C6,$EB,$A2,$98,$8C,$F9,$92,$B8,$BF,$E3,$EF,$9D,$F5,$63,$F4,$1B
	.dc.b	$9F,$F9,$F5,$B4,$0E,$17,$CF,$7F,$92,$A7,$87,$CF,$75,$24,$FF,$AC
	.dc.b	$7E,$7B,$F1,$0D,$E5,$FF,$27,$C6,$23,$FD,$C0,$DC,$7F,$07,$F3,$AF
	.dc.b	$83,$FF,$2A,$B6,$B9,$10,$FF,$74,$39,$7F,$DD,$5F,$11,$FB,$F7,$FD
	.dc.b	$85,$0F,$9E,$E6,$7F,$98,$27,$53,$C3,$C2,$C1,$1E,$3C,$75,$D1,$FE
	.dc.b	$CD,$9E,$38,$FA,$27,$FC,$D7,$EC,$F8,$E3,$F3,$BF,$30,$F6,$BF,$CD
	.dc.b	$ED,$C3,$FD,$AD,$5B,$E0,$FF,$57,$36,$3E,$93,$5F,$F6,$FE,$BC,$BE
	.dc.b	$0B,$BB,$23,$9B,$97,$FD,$C5,$E2,$BE,$7B,$FC,$FF,$F3,$8F,$86,$A7
	.dc.b	$FA,$0A,$75,$5C,$2B,$04,$7F,$DD,$94,$F8,$EC,$43,$6F,$DD,$DF,$28
	.dc.b	$C6,$9F,$EC,$98,$C4,$FF,$00,$BE,$55,$FE,$F7,$F9,$D7,$CB,$F9,$BE
	.dc.b	$7A,$4E,$63,$FD,$2F,$3A,$8C,$10,$EE,$7E,$66,$BF,$C0,$91,$B8,$C8
	.dc.b	$71,$7F,$C4,$9B,$8F,$FB,$7E,$BF,$C1,$9F,$3D,$F7,$A3,$1E,$CF,$CA
	.dc.b	$3F,$D5,$31,$9C,$3F,$50,$D4,$D3,$14,$F9,$4E,$3C,$F2,$BF,$61,$97
	.dc.b	$F2,$4B,$DD,$76,$F8,$1F,$B3,$EC,$CB,$DE,$7F,$EF,$5B,$56,$E7,$27
	.dc.b	$91,$C3,$F0,$B2,$FD,$85,$7F,$BF,$7E,$79,$CF,$FE,$C4,$6D,$72,$0E
	.dc.b	$1F,$97,$F3,$DF,$CB,$34,$FE,$11,$F3,$AF,$9D,$E9,$87,$32,$5F,$3B
	.dc.b	$E0,$AB,$83,$FC,$46,$33,$18,$8F,$0C,$FF,$15,$78,$57,$67,$C2,$E6
	.dc.b	$B8,$B7,$CB,$FC,$B4,$F0,$3F,$8F,$63,$EB,$C3,$B3,$6B,$86,$3D,$2B
	.dc.b	$C9,$7C,$AF,$78,$F8,$DB,$3B,$18,$F8,$9E,$37,$9C,$7B,$47,$8A,$F9
	.dc.b	$63,$E0,$8F,$31,$CB,$1F,$04,$7B,$E3,$35,$C9,$D5,$F8,$70,$F0,$47
	.dc.b	$C8,$3C,$18,$76,$7C,$67,$68,$DD,$73,$3A,$3C,$8F,$AF,$1F,$E3,$A6
	.dc.b	$BE,$73,$18,$7C,$E7,$5E,$63,$E2,$BF,$C7,$D1,$F1,$23,$0F,$F9,$06
	.dc.b	$BF,$CA,$53,$A6,$C7,$01,$F3,$D5,$A6,$5C,$11,$F9,$3F,$1C,$F9,$BF
	.dc.b	$F1,$96,$DF,$8B,$9F,$EB,$D5,$F5,$EB,$5B,$FD,$99,$C9,$CA,$F8,$EF
	.dc.b	$FE,$57,$BE,$11,$89,$FF,$2E,$C6,$2B,$94,$79,$B0,$FC,$D6,$1D,$A3
	.dc.b	$EE,$9D,$A3,$37,$CE,$3F,$3D,$7C,$DD,$9C,$F0,$FD,$15,$F3,$BF,$64
	.dc.b	$E4,$FE,$F5,$19,$9F,$44,$7A,$F0,$F2,$DF,$5B,$E2,$DA,$DC,$79,$27
	.dc.b	$C5,$BF,$F8,$E3,$0F,$53,$FE,$3A,$C3,$D5,$7E,$4C,$7F,$CC,$13,$B6
	.dc.b	$DF,$A3,$C7,$B7,$E6,$9A,$B0,$FF,$28,$57,$F1,$17,$FA,$2B,$FC,$72
	.dc.b	$DF,$3D,$FD,$4E,$35,$D5,$D2,$73,$DB,$D2,$F8,$E1,$FD,$1A,$73,$7C
	.dc.b	$0E,$98,$73,$9F,$E8,$F5,$6B,$E0,$5F,$A3,$E1,$FF,$D7,$7E,$27,$58
	.dc.b	$98,$FF,$30,$97,$CE,$BE,$7B,$39,$7F,$56,$9C,$C6,$31,$FF,$43,$46
	.dc.b	$2F,$E0,$61,$DF,$0E,$51,$FC,$CF,$C7,$FE,$7E,$DE,$AF,$96,$BE,$71
	.dc.b	$87,$CC,$AD,$E7,$FD,$89,$B9,$7F,$A5,$6F,$87,$06,$F1,$E7,$8F,$DE
	.dc.b	$9F,$E6,$17,$C1,$FF,$23,$1C,$0F,$11,$FD,$40,$C6,$1F,$DC,$3E,$79
	.dc.b	$ED,$7D,$E6,$CE,$EF,$EF,$D6,$EB,$47,$88,$B3,$B6,$C7,$F8,$02,$3E
	.dc.b	$04,$63,$2F,$65,$76,$AF,$F0,$0D,$7F,$B7,$CF,$EB,$85,$79,$47,$F7
	.dc.b	$58,$CF,$4F,$E5,$1D,$57,$C7,$F9,$EF,$F2,$B4,$68,$F8,$0E,$2F,$BD
	.dc.b	$7D,$48,$C5,$7F,$A8,$2B,$99,$F2,$A3,$2F,$F5,$DF,$97,$FC,$83,$78
	.dc.b	$3E,$9C,$FE,$9A,$73,$87,$F9,$DD,$B5,$F8,$1F,$13,$57,$F6,$9B,$1F
	.dc.b	$3D,$B5,$F0,$1F,$D8,$FC,$E3,$CE,$A7,$ED,$3E,$79,$F3,$D8,$C3,$FC
	.dc.b	$09,$5B,$6F,$BD,$9C,$7F,$C5,$F2,$8E,$A7,$EF,$BE,$79,$C7,$8B,$FF
	.dc.b	$07,$46,$1E,$BE,$87,$F0,$4E,$77,$C1,$79,$A3,$0E,$7D,$19,$5F,$E1
	.dc.b	$7A,$DB,$63,$F8,$6F,$C6,$3D,$BA,$7F,$10,$F9,$E7,$AB,$8B,$FF,$14
	.dc.b	$F2,$FF,$A8,$FE,$7B,$F8,$BB,$63,$3C,$3F,$E6,$B6,$CE,$A0,$D3,$19
	.dc.b	$59,$D9,$27,$FC,$B3,$5F,$EB,$19,$D4,$7F,$96,$3E,$71,$E8,$FF,$26
	.dc.b	$6D,$FF,$0F,$F9,$EF,$63,$FE,$37,$8C,$D7,$1A,$FE,$31,$5F,$B9,$8F
	.dc.b	$F5,$F4,$EB,$AF,$F8,$8C,$E7,$1F,$EF,$18,$7C,$0F,$37,$E2,$BF,$11
	.dc.b	$39,$D3,$4D,$34,$5C,$7D,$B2,$F9,$EE,$87,$5F,$86,$B5,$D6,$2D,$70
	.dc.b	$9B,$97,$82,$18,$FF,$83,$6B,$37,$FA,$3C,$18,$62,$FD,$FF,$13,$F2
	.dc.b	$4F,$EA,$3E,$4B,$DC,$7C,$86,$71,$F9,$55,$FD,$2F,$E7,$BE,$6E,$1F
	.dc.b	$4E,$31,$CB,$F1,$7E,$22,$5F,$C5,$F9,$EB,$7F,$02,$6F,$07,$A7,$83
	.dc.b	$FC,$49,$58,$BF,$47,$C9,$7C,$FF,$3F,$E8,$36,$3E,$83,$71,$F9,$6D
	.dc.b	$C3,$A7,$0A,$31,$5D,$0F,$B1,$7F,$DB,$BE,$7B,$FD,$A9,$1F,$E7,$D6
	.dc.b	$E1,$FE,$AA,$E1,$F8,$9E,$2E,$EE,$5F,$F5,$DF,$A0,$7C,$86,$30,$FE
	.dc.b	$B9,$18,$9F,$C3,$7C,$F7,$E2,$7E,$7F,$F1,$4D,$8C,$9F,$8B,$8C,$B4
	.dc.b	$FC,$77,$17,$E4,$7D,$1F,$C9,$36,$32,$7E,$4E,$32,$D3,$38,$7E,$42
	.dc.b	$32,$D8,$FF,$4E,$7C,$EB,$E7,$7F,$89,$CF,$FA,$34,$08,$D7,$91,$FE
	.dc.b	$5A,$7F,$1B,$E2,$DA,$E3,$5C,$7E,$79,$EF,$FF,$96,$5B,$1F,$E6,$26
	.dc.b	$BD,$C3,$87,$C5,$E7,$BA,$FF,$D4,$1D,$1F,$A1,$94,$BE,$7B,$3A,$FF
	.dc.b	$A9,$63,$13,$FE,$B9,$79,$36,$F6,$FB,$A3,$6F,$FB,$99,$B7,$DD,$E3
	.dc.b	$19,$F2,$B8,$BE,$2F,$9E,$B2,$6C,$C9,$F1,$46,$8C,$3C,$CC,$90,$C1
	.dc.b	$2B,$C5,$1C,$63,$D9,$3C,$A7,$31,$C5,$F4,$46,$70,$F8,$95,$C6,$BA
	.dc.b	$47,$C7,$9C,$57,$29,$C9,$F1,$63,$33,$CA,$3C,$24,$3E,$45,$69,$E0
	.dc.b	$66,$BA,$12,$7C,$11,$97,$A3,$C8,$C1,$C8,$C1,$23,$DA,$E8,$FA,$47
	.dc.b	$C1,$8E,$AF,$A8,$C9,$C5,$CC,$72,$8C,$1F,$83,$9F,$90,$F8,$63,$D0
	.dc.b	$E4,$F0,$5F,$F7,$47,$C0,$64,$F0,$39,$7A,$9E,$28,$F0,$12,$BE,$0B
	.dc.b	$DA,$3F,$3B,$7E,$27,$84,$E8,$F4,$91,$F0,$4F,$1A,$F0,$3D,$63,$33
	.dc.b	$C5,$F6,$3C,$5E,$A6,$8C,$90,$C9,$27,$AC,$71,$7F,$41,$1F,$08,$F4
	.dc.b	$9F,$1A,$FF,$08,$E0,$F8,$AE,$23,$9C,$77,$79,$BF,$21,$CB,$C6,$BD
	.dc.b	$D3,$F1,$67,$27,$08,$F8,$57,$99,$EC,$DA,$F3,$57,$08,$F1,$4F,$36
	.dc.b	$CF,$DE,$6B,$F7,$1B,$5F,$36,$3C,$97,$F3,$5F,$8D,$3D,$DF,$53,$67
	.dc.b	$CB,$1F,$14,$F1,$C6,$67,$95,$F9,$67,$CF,$7E,$56,$9E,$52,$9E,$52
	.dc.b	$DF,$96,$7E,$2C,$FC,$29,$F0,$1E,$42,$CF,$91,$FA,$A7,$90,$CD,$79
	.dc.b	$2F,$F9,$0F,$CE,$BE,$7B,$F9,$D5,$7F,$48,$8C,$38,$6B,$FC,$FD,$AF
	.dc.b	$F4,$F7,$37,$FC,$F5,$B7,$FD,$4A,$51,$87,$F9,$C7,$BF,$FC,$CB,$E7
	.dc.b	$9E,$7D,$F4,$A5,$7F,$5A,$C7,$FA,$45,$6A,$31,$8F,$58,$C4,$27,$75
	.dc.b	$CA,$31,$5B,$BF,$07,$0C,$9E,$67,$F6,$8C,$3F,$AE,$E1,$E0,$F6,$35
	.dc.b	$3C,$8D,$56,$27,$51,$C8,$57,$2C,$3C,$D5,$B7,$D1,$7F,$D4,$BE,$7B
	.dc.b	$FC,$41,$7C,$6F,$FC,$76,$DF,$23,$5E,$1C,$8E,$D5,$C1,$F2,$07,$19
	.dc.b	$7C,$8E,$9F,$EB,$FF,$3A,$F9,$EA,$4B,$F4,$43,$E3,$71,$1D,$23,$83
	.dc.b	$A9,$E7,$3D,$3E,$43,$AF,$CC,$CA,$E7,$19,$7A,$C6,$5E,$91,$97,$B4
	.dc.b	$67,$1F,$F0,$36,$0A,$FE,$EF,$E7,$FF,$8E,$2F,$E7,$7C,$EB,$E7,$BF
	.dc.b	$BD,$B6,$3E,$94,$ED,$F8,$AD,$C3,$FC,$24,$DE,$2F,$CB,$36,$7F,$C5
	.dc.b	$35,$FE,$08,$AF,$F0,$C3,$FC,$68,$9D,$3E,$0E,$AF,$0B,$2B,$CA,$E5
	.dc.b	$F3,$F4,$79,$BE,$0A,$46,$08,$7D,$AB,$46,$A3,$13,$EE,$31,$87,$92
	.dc.b	$71,$3F,$03,$A7,$FC,$3B,$E8,$3E,$7A,$9F,$2A,$24,$EC,$87,$E2,$A7
	.dc.b	$F5,$77,$FE,$37,$7F,$35,$1F,$E4,$C5,$D3,$D8,$EF,$D5,$27,$F4,$53
	.dc.b	$A6,$CE,$5E,$DC,$DE,$62,$3C,$3E,$7A,$BA,$39,$C7,$B2,$33,$87,$96
	.dc.b	$2F,$0F,$F0,$FC,$15,$FE,$3E,$F9,$E7,$CF,$5E,$79,$97,$ED,$E7,$2D
	.dc.b	$F2,$78,$11,$FD,$E7,$CF,$7A,$DF,$D7,$B7,$4F,$F2,$E3,$73,$FE,$40
	.dc.b	$D9,$FF,$39,$FC,$F7,$E2,$E7,$3F,$3D,$FE,$52,$A3,$4F,$F2,$AF,$CF
	.dc.b	$39,$69,$30,$FE,$1D,$19,$7F,$88,$7C,$F3,$E1,$3F,$CB,$AD,$8D,$D7
	.dc.b	$17,$F8,$3C,$65,$FE,$13,$EA,$FF,$01,$79,$FC,$F7,$23,$FC,$CF,$19
	.dc.b	$8D,$F8,$38,$23,$E2,$E5,$AE,$8F,$E4,$93,$C7,$7F,$F4,$16,$1F,$E9
	.dc.b	$07,$FD,$09,$EE,$F8,$DF,$F4,$37,$CC,$3D,$8F,$F3,$BB,$70,$FE,$5D
	.dc.b	$5B,$E1,$FE,$9C,$D8,$FF,$50,$B5,$FF,$73,$6D,$CB,$E0,$3F,$D0,$33
	.dc.b	$92,$46,$67,$F9,$D7,$88,$F9,$EF,$F4,$3F,$CD,$3E,$1A,$9F,$E8,$A9
	.dc.b	$D5,$70,$AC,$11,$FE,$89,$4F,$8D,$C4,$7F,$38,$AE,$2D,$AE,$F3,$FE
	.dc.b	$96,$F9,$D7,$CB,$78,$BE,$7A,$5F,$F4,$F9,$CB,$67,$81,$1F,$0F,$91
	.dc.b	$5F,$E8,$F3,$C1,$B3,$EE,$9F,$F5,$2F,$CE,$FD,$7E,$0F,$EF,$5A,$7F
	.dc.b	$AE,$2F,$D7,$EB,$6B,$A2,$47,$F6,$1B,$E0,$FF,$B4,$9A,$A9,$F3,$DC
	.dc.b	$23,$FA,$CC,$E8,$DC,$EA,$3A,$7B,$DF,$4B,$E3,$24,$60,$87,$28,$F4
	.dc.b	$9F,$63,$B7,$FD,$7F,$F3,$CE,$8A,$4A,$FE,$F1,$F3,$DF,$EE,$0E,$BF
	.dc.b	$F7,$4F,$CE,$BE,$77,$FC,$35,$87,$1B,$FE,$FC,$D7,$FB,$8B,$6B,$FD
	.dc.b	$BD,$7C,$E5,$7C,$DF,$F7,$AC,$66,$31,$0A,$E9,$18,$6D,$FF,$8A,$EB
	.dc.b	$83,$7C,$DF,$F1,$05,$F6,$8F,$3E,$3F,$E3,$1C,$39,$36,$38,$63,$EE
	.dc.b	$6D,$71,$07,$8E,$78,$38,$C7,$C6,$FA,$B0,$F0,$E1,$EA,$C7,$CB,$3B
	.dc.b	$BE,$2F,$33,$C0,$71,$C6,$A9,$E2,$EC,$F2,$16,$BB,$5F,$83,$1F,$4C
	.dc.b	$7B,$A7,$9E,$3E,$67,$A5,$6D,$F0,$1D,$23,$46,$EF,$C2,$7D,$98,$FF
	.dc.b	$2D,$1C,$B0,$FF,$33,$C6,$63,$E4,$9A,$7F,$B2,$FC,$56,$BC,$E3,$49
	.dc.b	$3F,$E7,$01,$F3,$D5,$A6,$5C,$11,$F8,$7F,$14,$F9,$B7,$3F,$CA,$9F
	.dc.b	$B3,$5F,$32,$B5,$39,$9E,$BA,$FE,$42,$73,$1F,$38,$E2,$76,$BE,$AE
	.dc.b	$4E,$3B,$7F,$9F,$67,$8D,$7E,$6A,$78,$BD,$6B,$8B,$98,$E2,$7E,$3A
	.dc.b	$78,$D7,$CE,$8C,$88,$CB,$F4,$B0,$F6,$5F,$4A,$E1,$18,$C3,$DC,$DA
	.dc.b	$F4,$37,$0F,$13,$6B,$8C,$FB,$EB,$86,$9F,$02,$3C,$58,$7C,$28,$C4
	.dc.b	$7B,$AB,$E1,$4E,$6A,$9F,$E8,$3C,$7F,$BD,$36,$35,$5F,$07,$82,$BE
	.dc.b	$FE,$9C,$E2,$F9,$EF,$E9,$18,$73,$9C,$C6,$2B,$FD,$03,$8F,$F4,$6A
	.dc.b	$FF,$4D,$DF,$1E,$DE,$B3,$9A,$FF,$4D,$4E,$6B,$CD,$5A,$3C,$BF,$0F
	.dc.b	$FE,$B4,$F1,$71,$3A,$3F,$D4,$F0,$F9,$D7,$CF,$6B,$1F,$F5,$34,$E6
	.dc.b	$31,$8F,$E2,$A3,$17,$CE,$B7,$7C,$A3,$FD,$55,$E5,$3A,$7E,$2B,$63
	.dc.b	$FD,$6E,$3E,$75,$86,$AA,$4E,$A7,$F3,$B8,$FD,$BB,$E5,$8F,$FA,$EE
	.dc.b	$B9,$70,$F0,$8C,$5F,$09,$FF,$6D,$57,$0B,$FD,$1C,$70,$3F,$D8,$1F
	.dc.b	$3D,$FE,$78,$6F,$B5,$F7,$9F,$37,$1F,$DE,$9D,$63,$FE,$B8,$6D,$F4
	.dc.b	$C6,$DF,$FB,$A6,$6D,$7E,$FA,$6A,$8C,$CF,$C3,$C3,$B1,$E1,$AD,$C7
	.dc.b	$F9,$E0,$FF,$6B,$16,$BD,$FE,$07,$E4,$A3,$18,$7E,$85,$A7,$F4,$41
	.dc.b	$F3,$DF,$E4,$F7,$9F,$57,$BE,$31,$7F,$ED,$49,$E2,$FF,$16,$8C,$D7
	.dc.b	$5A,$E1,$E5,$7E,$8B,$E5,$3F,$39,$3F,$B1,$6E,$1E,$1C,$BC,$2F,$FB
	.dc.b	$BE,$33,$8F,$F5,$86,$C7,$E1,$7C,$5E,$9C,$DF,$0D,$B1,$F8,$56,$D7
	.dc.b	$E0,$1F,$D9,$FC,$E3,$D0,$A7,$ED,$BE,$79,$F3,$D8,$C3,$FE,$00,$56
	.dc.b	$DB,$EF,$67,$1F,$F7,$AC,$A3,$A9,$FC,$03,$E7,$9C,$D8,$BF,$EE,$3B
	.dc.b	$FF,$11,$9C,$C7,$6E,$0F,$E0,$FD,$2F,$33,$CF,$18,$73,$E8,$CA,$E6
	.dc.b	$7F,$7C,$D3,$F4,$F3,$95,$8F,$F1,$FF,$8B,$7B,$7F,$C5,$1B,$7C,$AB
	.dc.b	$FC,$0B,$F3,$CF,$5B,$57,$FC,$6B,$E7,$5E,$7D,$3F,$85,$FC,$EF,$2C
	.dc.b	$3B,$E3,$E3,$AD,$BD,$61,$18,$76,$7F,$85,$1C,$67,$FC,$2E,$BA,$3F
	.dc.b	$CD,$E3,$35,$D6,$FC,$3E,$17,$8E,$3F,$8C,$DF,$EE,$DF,$C4,$79,$9F
	.dc.b	$DA,$9F,$77,$5F,$F7,$8D,$3F,$9B,$BF,$DE,$A2,$CF,$C1,$76,$F1,$CF
	.dc.b	$CF,$3C,$67,$27,$0A,$E3,$F3,$B2,$38,$FC,$45,$BE,$38,$2B,$1A,$FF
	.dc.b	$72,$C7,$FC,$17,$78,$C3,$E5,$7C,$97,$D1,$3F,$A6,$F9,$4F,$6D,$F1
	.dc.b	$F9,$AF,$9D,$C2,$BC,$6B,$F3,$B0,$E1,$78,$CF,$E6,$57,$F5,$4F,$9D
	.dc.b	$F9,$F8,$F8,$0F,$E9,$EE,$30,$69,$FE,$20,$F8,$8F,$8E,$7C,$8F,$9D
	.dc.b	$B7,$F2,$22,$D7,$C0,$9B,$8F,$BF,$81,$F2,$3F,$43,$D0,$FA,$2D,$C7
	.dc.b	$E8,$B7,$0F,$92,$DB,$FE,$F5,$F3,$BF,$EC,$BA,$FC,$73,$6B,$E0,$70
	.dc.b	$FE,$E5,$FC,$78,$8D,$57,$2C,$3F,$BB,$C6,$2F,$FC,$35,$7C,$6B,$51
	.dc.b	$DA,$BF,$B5,$FC,$E3,$E7,$78,$C6,$1F,$C5,$FC,$EF,$E3,$3E,$76,$DF
	.dc.b	$E3,$1B,$9F,$E1,$1B,$8F,$C5,$7F,$27,$D9,$F9,$4E,$BB,$7F,$94,$6D
	.dc.b	$FE,$31,$B7,$F3,$5F,$F0,$8F,$CD,$F9,$6B,$FC,$65,$48,$C3,$FE,$35
	.dc.b	$6B,$FE,$64,$6B,$FC,$35,$7C,$FF,$F8,$9A,$3F,$D1,$71,$89,$EA,$D8
	.dc.b	$E9,$3B,$5F,$3C,$F9,$EF,$E4,$B1,$FE,$45,$8C,$79,$1E,$65,$DB,$FE
	.dc.b	$02,$9D,$57,$07,$FC,$F3,$E7,$7E,$7D,$FE,$1C,$BE,$7B,$F9,$26,$BF
	.dc.b	$E3,$F8,$C4,$EB,$B3,$3A,$FF,$29,$9E,$53,$F8,$67,$6D,$BD,$BF,$EA
	.dc.b	$A8,$DD,$F5,$F1,$B3,$29,$7C,$BE,$1B,$1E,$69,$C1,$99,$91,$93,$F9
	.dc.b	$47,$CE,$BE,$7B,$97,$8F,$27,$31,$E0,$9D,$47,$C2,$9D,$3F,$A0,$8C
	.dc.b	$B6,$73,$8F,$C5,$AE,$51,$27,$91,$0E,$07,$80,$C4,$F1,$70,$68,$F9
	.dc.b	$53,$DB,$3F,$EA,$D7,$E2,$AF,$F3,$93,$63,$E9,$4E,$EB,$FD,$AF,$39
	.dc.b	$6D,$E5,$F9,$D3,$E7,$7C,$17,$E3,$7C,$27,$91,$EF,$3C,$72,$FE,$B7
	.dc.b	$39,$C3,$CB,$39,$BE,$97,$C2,$BD,$F5,$E6,$78,$1E,$43,$27,$E1,$67
	.dc.b	$8C,$74,$8C,$9E,$83,$88,$C3,$DC,$E4,$F4,$9E,$C8,$E4,$E2,$BD,$C7
	.dc.b	$95,$D1,$E1,$3B,$CF,$43,$AE,$1F,$02,$39,$57,$4A,$E4,$74,$7D,$D1
	.dc.b	$D0,$F3,$61,$F1,$5F,$4D,$F4,$C3,$EA,$3D,$E7,$53,$C6,$34,$F9,$23
	.dc.b	$E2,$BE,$73,$C6,$7C,$B3,$C1,$87,$9F,$0F,$24,$65,$F3,$E1,$E6,$73
	.dc.b	$1E,$63,$E6,$CF,$19,$D1,$EF,$24,$78,$67,$8D,$FA,$63,$EB,$D7,$17
	.dc.b	$C1,$5C,$1F,$AD,$5C,$2F,$D7,$18,$9E,$EF,$02,$19,$3C,$0F,$3B,$F8
	.dc.b	$B1,$F0,$8F,$51,$F0,$B0,$F9,$B5,$DB,$0F,$24,$F3,$37,$87,$C4,$7A
	.dc.b	$90,$CE,$1E,$99,$E2,$72,$C3,$E0,$4F,$86,$31,$7D,$F0,$F0,$CF,$CC
	.dc.b	$6B,$E5,$3E,$E1,$92,$99,$3D,$95,$DB,$0F,$AF,$7D,$A7,$81,$AB,$F0
	.dc.b	$3C,$A7,$D8,$FB,$E7,$83,$D6,$3E,$93,$C9,$AE,$70,$FA,$6F,$C2,$03
	.dc.b	$E7,$9E,$57,$F3,$A7,$FC,$F9,$1F,$E5,$96,$BF,$E0,$A7,$6E,$63,$87
	.dc.b	$9F,$23,$07,$1F,$86,$E8,$47,$A7,$AF,$5E,$91,$71,$63,$FE,$7D,$C5
	.dc.b	$7F,$D3,$BE,$7B,$FC,$53,$87,$97,$C8,$F4,$19,$F0,$E4,$64,$87,$F8
	.dc.b	$8A,$78,$B6,$36,$D3,$E0,$74,$7F,$5C,$F9,$D7,$CF,$52,$4F,$F9,$F6
	.dc.b	$1F,$1B,$88,$FF,$17,$C7,$07,$53,$CE,$7A,$7C,$8F,$5F,$99,$95,$CE
	.dc.b	$32,$F5,$8C,$BD,$23,$FB,$7F,$67,$F8,$C6,$3F,$D1,$DF,$3A,$F9,$EF
	.dc.b	$EE,$6D,$8F,$F4,$94,$ED,$FF,$21,$B7,$8B,$E7,$C7,$E1,$27,$83,$FE
	.dc.b	$48,$6C,$6A,$3F,$C2,$BE,$37,$88,$8F,$67,$C3,$DD,$E2,$ED,$93,$E6
	.dc.b	$7C,$BD,$B8,$8E,$D1,$6F,$B6,$1F,$E0,$89,$F1,$7C,$27,$B3,$E0,$E4
	.dc.b	$7C,$08,$F7,$98,$7D,$11,$87,$E0,$F5,$7F,$85,$BC,$F7,$CF,$53,$E5
	.dc.b	$C4,$9F,$F5,$C4,$E9,$B3,$97,$E4,$CD,$C7,$E2,$6D,$FA,$67,$E3,$5F
	.dc.b	$E9,$AB,$FC,$E2,$BB,$3C,$5D,$52,$7B,$C7,$94,$C3,$C2,$30,$F9,$39
	.dc.b	$FD,$44,$7C,$FF,$3D,$5E,$F1,$70,$F1,$CF,$F8,$DB,$F2,$27,$C6,$7F
	.dc.b	$8E,$3F,$17,$7D,$85,$35,$13,$1F,$F6,$D4,$E9,$B3,$97,$FD,$1B,$5D
	.dc.b	$1F,$F7,$03,$5C,$D7,$7E,$BF,$F2,$3F,$D0,$3E,$86,$F6,$4B,$FF,$74
	.dc.b	$4E,$5B,$E7,$F0,$23,$FE,$ED,$FA,$0F,$E0,$AD,$E0,$FF,$29,$C6,$23
	.dc.b	$FD,$42,$DD,$FF,$35,$F4,$3F,$E5,$AF,$1B,$F9,$1D,$29,$F4,$3C,$89
	.dc.b	$1F,$EF,$BF,$A1,$75,$FF,$98,$1B,$19,$8C,$1F,$E6,$4F,$A0,$7C,$35
	.dc.b	$33,$3B,$3F,$99,$E7,$FA,$D9,$DC,$63,$2F,$2F,$C6,$7F,$0E,$8C,$1F
	.dc.b	$CE,$A5,$F4,$3D,$CF,$F3,$9C,$EE,$7F,$D8,$2D,$BF,$DD,$36,$FA,$F8
	.dc.b	$FE,$7E,$AE,$2C,$BE,$E2,$B8,$3E,$19,$FF,$40,$FB,$9E,$88,$FF,$67
	.dc.b	$FC,$E3,$DC,$FF,$42,$B6,$39,$CE,$F2,$FE,$18,$DC,$BE,$9B,$67,$FA
	.dc.b	$BB,$55,$3E,$1B,$C3,$B4,$9E,$AF,$82,$3F,$D2,$BE,$5B,$E8,$7F,$D2
	.dc.b	$FF,$3C,$F8,$BA,$6A,$26,$1F,$02,$B8,$36,$38,$FC,$8E,$A3,$FD,$DB
	.dc.b	$3A,$6C,$E9,$F9,$3F,$41,$FA,$3D,$BF,$95,$62,$FA,$1F,$19,$FE,$AA
	.dc.b	$9D,$6F,$F9,$46,$DF,$F7,$86,$A6,$D3,$CD,$E5,$E1,$4A,$F0,$BF,$02
	.dc.b	$3F,$D6,$FF,$43,$F6,$A5,$5F,$EB,$97,$EF,$B5,$65,$FD,$06,$78,$13
	.dc.b	$9B,$99,$F3,$3E,$5E,$D7,$E2,$27,$4F,$48,$F0,$FC,$1F,$D2,$F7,$A4
	.dc.b	$EE,$32,$D2,$1E,$2D,$7F,$6F,$97,$EC,$23,$07,$F8,$73,$E8,$7F,$0A
	.dc.b	$FE,$61,$AE,$83,$8C,$FF,$AE,$7D,$0F,$F3,$57,$FC,$4B,$F4,$2F,$A0
	.dc.b	$FE,$0B,$5B,$C6,$F4,$FF,$B8,$23,$18,$FF,$8B,$31,$E4,$B0,$E4,$E7
	.dc.b	$CF,$F9,$6E,$71,$E6,$F8,$1A,$E4,$E1,$96,$5A,$64,$F5,$E3,$FC,$6A
	.dc.b	$FB,$C6,$0E,$03,$0F,$2C,$69,$B7,$E0,$8E,$38,$FC,$87,$58,$FA,$1B
	.dc.b	$8F,$49,$D5,$71,$3F,$95,$DF,$4C,$3C,$F7,$D3,$1F,$35,$6E,$72,$F0
	.dc.b	$C7,$B3,$98,$F5,$1E,$FB,$E8,$FA,$E7,$6F,$E0,$6B,$AB,$4F,$83,$3C
	.dc.b	$0F,$AA,$FF,$40,$39,$63,$5F,$9A,$66,$36,$FB,$3C,$67,$C6,$6E,$3E
	.dc.b	$64,$61,$E5,$3F,$E5,$89,$D3,$63,$80,$FA,$1A,$D3,$2E,$08,$FD,$FF
	.dc.b	$91,$7C,$F7,$F8,$E3,$5F,$BF,$3F,$E7,$CA,$FB,$F5,$AD,$FF,$03,$39
	.dc.b	$39,$5F,$1D,$BE,$05,$F0,$8C,$4F,$C5,$8C,$57,$28,$F3,$61,$F8,$EC
	.dc.b	$3B,$47,$E4,$CE,$D1,$9B,$E7,$1F,$DB,$AF,$9B,$B3,$9E,$1F,$DC,$AF
	.dc.b	$9D,$FB,$27,$27,$FA,$62,$33,$3E,$88,$F5,$E1,$E5,$BE,$B7,$C5,$B5
	.dc.b	$B8,$F2,$4F,$8B,$4F,$F2,$E6,$1E,$A7,$FC,$BD,$87,$AA,$FC,$98,$FF
	.dc.b	$99,$A7,$6D,$BF,$47,$8D,$6F,$CD,$35,$61,$FE,$08,$AF,$ED,$6F,$F8
	.dc.b	$C9,$FE,$F7,$6F,$A1,$77,$71,$AD,$5F,$F1,$EA,$D4,$E7,$AB,$C1,$18
	.dc.b	$C3,$AC,$63,$0E,$71,$E0,$AE,$87,$9F,$E6,$3F,$D8,$3E,$26,$27,$47
	.dc.b	$F8,$3A,$1F,$40,$FA,$1D,$67,$FC,$C2,$73,$18,$CF,$FD,$17,$18,$BE
	.dc.b	$75,$BB,$E5,$1F,$E0,$CF,$1D,$D5,$F1,$DB,$1F,$E1,$E1,$F3,$EC,$35
	.dc.b	$52,$75,$3F,$65,$FF,$1D,$D7,$2C,$3F,$A2,$C7,$2E,$0E,$11,$8A,$E0
	.dc.b	$FF,$B9,$2B,$86,$5F,$EA,$4B,$E0,$7F,$88,$BE,$85,$EF,$7D,$D6,$C7
	.dc.b	$8B,$8F,$13,$23,$CA,$DC,$3C,$0D,$BE,$D8,$F2,$9F,$EA,$D8,$59,$CD
	.dc.b	$7E,$2E,$FF,$DC,$67,$F8,$D0,$B5,$EB,$EF,$D7,$54,$9F,$95,$F4,$3F
	.dc.b	$E5,$27,$84,$7E,$5D,$46,$6B,$FB,$34,$74,$C3,$E1,$4E,$5F,$F6,$14
	.dc.b	$64,$4F,$FB,$E2,$75,$1F,$E6,$CF,$98,$FC,$8C,$FF,$71,$6D,$7C,$0A
	.dc.b	$FD,$34,$62,$B8,$72,$79,$74,$7C,$66,$A3,$0D,$9F,$C0,$B7,$1F,$B8
	.dc.b	$FE,$C7,$E7,$FF,$CB,$13,$FE,$5A,$6C,$FF,$80,$FE,$85,$F4,$3B,$F4
	.dc.b	$65,$EF,$E0,$ED,$FF,$82,$70,$47,$EF,$7E,$85,$CB,$4F,$04,$5B,$FF
	.dc.b	$34,$F4,$7F,$8D,$5B,$E5,$F9,$7D,$2F,$E0,$D3,$9C,$BF,$C2,$CF,$F9
	.dc.b	$56,$BF,$A7,$36,$3F,$AB,$35,$FE,$B4,$D7,$FD,$37,$F2,$0F,$6F,$F8
	.dc.b	$DC,$7D,$68,$C5,$7F,$59,$6C,$7F,$61,$6B,$FD,$97,$E7,$9E,$6F,$F1
	.dc.b	$B6,$D7,$F0,$5F,$A0,$E3,$D0,$8F,$F5,$5C,$EA,$FF,$C6,$0D,$7A,$5F
	.dc.b	$91,$46,$5E,$31,$98,$E5,$5F,$C3,$2F,$F6,$B7,$A3,$C5,$DF,$FC,$C7
	.dc.b	$A3,$FB,$AC,$5F,$2B,$FD,$0C,$FF,$AA,$23,$2D,$8C,$E5,$FE,$4A,$9C
	.dc.b	$B6,$33,$87,$B6,$13,$A8,$C3,$FE,$04,$F5,$BC,$1F,$41,$30,$ED,$F1
	.dc.b	$17,$0E,$93,$70,$E3,$8A,$B1,$7F,$2B,$E3,$3E,$81,$FD,$37,$CA,$7B
	.dc.b	$4F,$92,$CD,$7F,$AD,$B8,$17,$F2,$49,$1F,$E1,$78,$CB,$63,$FA,$4F
	.dc.b	$D0,$7D,$A8,$F0,$1F,$D5,$1C,$6B,$F0,$33,$7C,$5F,$CD,$3E,$77,$D0
	.dc.b	$5B,$F9,$91,$6B,$E5,$4D,$DB,$DF,$F2,$7F,$57,$CE,$FA,$CD,$C3,$EA
	.dc.b	$36,$FE,$93,$6B,$E9,$43,$E8,$3F,$D9,$D5,$FE,$D6,$6C,$7F,$A1,$38
	.dc.b	$3E,$3F,$83,$FD,$CE,$73,$7C,$9D,$D7,$F9,$0D,$AF,$F1,$C3,$B6,$3F
	.dc.b	$97,$8C,$C7,$F2,$07,$9E,$1F,$C8,$BA,$DF,$40,$F9,$3F,$06,$1F,$DB
	.dc.b	$E3,$15,$F9,$1F,$A0,$FC,$B7,$D0,$7E,$5D,$B1,$93,$F2,$D1,$96,$9F
	.dc.b	$99,$EF,$FC,$F7,$67,$E7,$DB,$19,$3F,$3D,$19,$69,$FE,$FC,$FA,$EF
	.dc.b	$E3,$2F,$F9,$6D,$BA,$FF,$98,$DA,$FF,$99,$DA,$FC,$5F,$9F,$7C,$FF
	.dc.b	$F2,$28,$FF,$35,$46,$3C,$6F,$F3,$33,$F9,$DD,$ED,$6E,$BF,$CE,$5F
	.dc.b	$42,$F7,$FF,$CC,$AD,$8F,$F3,$73,$5F,$F3,$B3,$5F,$F3,$D7,$9B,$FA
	.dc.b	$07,$FA,$9F,$27,$E1,$A1,$F4,$1F,$6E,$30,$DD,$7E,$F4,$F1,$AF,$F6
	.dc.b	$33,$6B,$71,$D2,$77,$5C,$63,$FB,$0D,$6E,$79,$F9,$39,$9F,$EB,$3F
	.dc.b	$35,$68,$D9,$93,$D4,$E7,$1B,$32,$FF,$B6,$7E,$81,$F4,$1C,$FC,$A9
	.dc.b	$3C,$44,$E9,$C3,$5F,$6C,$70,$7A,$4F,$9E,$75,$9F,$C9,$AD,$3F,$B0
	.dc.b	$8C,$E5,$FA,$56,$DF,$C2,$AD,$CE,$6F,$A1,$89,$D1,$E4,$72,$FA,$72
	.dc.b	$FE,$9E,$DA,$CB,$E6,$30,$43,$04,$C3,$D0,$E5,$F4,$57,$27,$31,$F0
	.dc.b	$8F,$11,$0C,$11,$E2,$E5,$E2,$4B,$F7,$46,$5F,$C5,$D7,$BE,$3E,$9D
	.dc.b	$78,$67,$31,$E5,$78,$BE,$97,$2F,$85,$CD,$78,$A3,$07,$CA,$38,$1F
	.dc.b	$E3,$2C,$3E,$9B,$FE,$28,$BF,$14,$60,$CE,$1F,$3A,$79,$9E,$63,$D4
	.dc.b	$75,$AF,$73,$D7,$0F,$53,$9C,$3C,$F3,$E5,$23,$82,$47,$13,$A9,$E5
	.dc.b	$3E,$11,$93,$D6,$64,$F2,$DF,$A6,$BE,$65,$FA,$6F,$BD,$71,$7C,$EF
	.dc.b	$17,$2F,$B4,$FF,$7F,$57,$E8,$DF,$E7,$C6,$4A,$E5,$FA,$A6,$2B,$FD
	.dc.b	$F3,$19,$8F,$41,$93,$DA,$FC,$EF,$67,$E3,$C7,$86,$BC,$76,$9C,$D7
	.dc.b	$B6,$BF,$45,$18,$BE,$35,$B9,$EB,$18,$9D,$47,$CC,$9D,$3F,$2A,$BA
	.dc.b	$3F,$C7,$27,$6F,$CE,$9D,$DF,$C6,$AD,$BF,$CA,$6B,$6F,$A1,$B8,$7E
	.dc.b	$50,$FE,$52,$E2,$FE,$94,$61,$F3,$9E,$58,$D3,$F0,$E3,$4F,$9F,$0F
	.dc.b	$63,$C0,$F3,$4F,$0B,$F9,$91,$9B,$F2,$36,$FC,$8F,$F3,$4A,$E1,$7F
	.dc.b	$CC,$5B,$1F,$38,$E0,$7A,$0E,$06,$0F,$31,$83,$CC,$7A,$CF,$E6,$1F
	.dc.b	$40,$F3,$7F,$A0,$3F,$E8,$68,$CF,$9F,$89,$91,$82,$0C,$7F,$C3,$B5
	.dc.b	$CE,$3A,$F0,$F4,$23,$E0,$F5,$EB,$C7,$3F,$F0,$C5,$7F,$67,$BE,$77
	.dc.b	$E4,$BF,$EA,$9F,$40,$F0,$FC,$FE,$54,$CB,$FA,$54,$E5,$B3,$FA,$67
	.dc.b	$FC,$49,$19,$70,$D7,$F5,$6D,$77,$1E,$02,$3F,$23,$A7,$FB,$37,$D0
	.dc.b	$3E,$66,$92,$7F,$AE,$43,$E2,$F1,$1E,$F3,$FD,$A3,$1C,$5D,$CF,$49
	.dc.b	$EB,$F1,$FD,$BE,$62,$57,$48,$CB,$DA,$32,$F5,$16,$9F,$EF,$9E,$8F
	.dc.b	$F8,$F6,$7F,$B8,$7C,$FB,$E8,$3F,$C0,$0D,$8F,$EE,$93,$B7,$A3,$5F
	.dc.b	$E2,$2D,$7F,$CA,$0D,$7F,$DD,$AD,$7F,$D0,$DE,$4F,$88,$8F,$8B,$D1
	.dc.b	$F0,$F7,$48,$F2,$BE,$4E,$EC,$47,$B7,$04,$E3,$97,$FC,$DB,$1F,$E1
	.dc.b	$69,$F6,$F5,$7B,$3E,$06,$46,$08,$73,$72,$F5,$8F,$8F,$D7,$FE,$23
	.dc.b	$F4,$1F,$41,$4C,$D7,$C6,$9F,$C7,$3F,$01,$AF,$FA,$95,$BB,$7E,$B6
	.dc.b	$3F,$28,$DC,$BF,$D0,$AD,$EC,$F9,$75,$E1,$ED,$F1,$76,$78,$FB,$26
	.dc.b	$3F,$E2,$9C,$FF,$51,$87,$FA,$05,$B0,$3F,$D4,$33,$C1,$FC,$34,$69
	.dc.b	$FF,$52,$F4,$FB,$C8,$F5,$FA,$0A,$F0,$8B,$A7,$FA,$72,$95,$FE,$49
	.dc.b	$F9,$E7,$C5,$7F,$92,$FE,$81,$F4,$14,$D4,$4C,$BE,$4D,$70,$6C,$7C
	.dc.b	$A8,$C4,$7E,$ED,$B8,$FF,$AB,$1B,$A7,$FA,$85,$B3,$F0,$70,$FD,$73
	.dc.b	$FA,$5F,$9F,$7D,$07,$D6,$6C,$7D,$A6,$BF,$C2,$5A,$FF,$70,$5E,$87
	.dc.b	$F9,$9D,$B1,$FE,$5B,$6B,$FB,$3F,$9F,$70,$7F,$9B,$7E,$83,$FC,$E1
	.dc.b	$3F,$56,$72,$FF,$98,$DA,$FE,$EE,$36,$BE,$7F,$FB,$0E,$9F,$EB,$FA
	.dc.b	$7D,$07,$F9,$E5,$B5,$FE,$6D,$AD,$BF,$83,$6B,$CF,$A7,$FD,$01,$18
	.dc.b	$FA,$07,$C3,$7F,$A0,$DB,$5C,$EB,$66,$67,$FD,$11,$EE,$7F,$1F,$9F
	.dc.b	$F3,$E5,$3E,$83,$91,$AF,$22,$4F,$FA,$06,$B8,$3F,$C6,$5B,$C9,$FE
	.dc.b	$E8,$BF,$F5,$83,$FE,$EA,$8C,$EB,$FE,$DF,$6D,$7E,$7D,$AA,$B8,$CF
	.dc.b	$FB,$97,$E0,$FD,$73,$FC,$9F,$E6,$5E,$F7,$FA,$85,$B1,$FE,$95,$9D
	.dc.b	$E3,$FE,$FA,$6D,$F3,$D3,$FD,$51,$18,$AF,$D8,$5B,$E1,$B3,$52,$7A
	.dc.b	$3D,$E3,$FD,$61,$E5,$3E,$7F,$FD,$65,$F3,$8F,$A0,$A6,$A2,$61,$FE
	.dc.b	$A9,$AE,$0D,$8E,$3F,$21,$A8,$FE,$89,$3A,$6C,$E9,$FF,$59,$7C,$FF
	.dc.b	$EA,$16,$0F,$A0,$F1,$9F,$EC,$59,$D1,$3C,$0C,$1E,$8F,$32,$DF,$48
	.dc.b	$AF,$F5,$5B,$FE,$72,$FF,$B0,$7E,$83,$EB,$C3,$DB,$F8,$27,$FB,$4E
	.dc.b	$3E,$D9,$39,$F0,$7C,$6F,$96,$B7,$CE,$2B,$DE,$39,$7C,$27,$CC,$F7
	.dc.b	$A4,$F8,$63,$2D,$21,$E7,$EC,$FF,$72,$7D,$03,$9A,$92,$7F,$DB,$5F
	.dc.b	$40,$F3,$7F,$C4,$90,$F9,$F7,$CF,$FF,$33,$86,$1C,$DF,$F7,$13,$5F
	.dc.b	$EF,$4D,$ED,$76,$F1,$C7,$B4,$7F,$3D,$C3,$AD,$FF,$3F,$9C,$BE,$1C
	.dc.b	$3C,$18,$75,$AD,$CF,$8B,$2F,$24,$62,$F8,$B7,$2F,$5B,$C2,$F9,$61
	.dc.b	$D2,$F8,$BE,$5A,$DB,$71,$F3,$3B,$C6,$9F,$D5,$6F,$9B,$E9,$AE,$2D
	.dc.b	$AD,$BE,$72,$CF,$05,$8F,$A2,$FC,$11,$F0,$4E,$D8,$79,$9A,$74,$9E
	.dc.b	$0F,$73,$81,$DC,$B5,$DE,$37,$1E,$0C,$3C,$EE,$F1,$F8,$D7,$DE,$35
	.dc.b	$8F,$7A,$FC,$2C,$7F,$AE,$AF,$EB,$B6,$32,$FA,$8D,$78,$CF,$8F,$FF
	.dc.b	$07,$D7,$F4,$C6,$C6,$67,$FC,$C1,$39,$7F,$C2,$7F,$41,$FE,$14,$4D
	.dc.b	$23,$A3,$E0,$FC,$7B,$E7,$7F,$23,$1F,$10,$FF,$06,$4F,$D4,$9D,$6F
	.dc.b	$F5,$A7,$27,$2A,$E3,$AF,$E4,$23,$2F,$E4,$AB,$8C,$75,$3C,$A7,$1A
	.dc.b	$E8,$E2,$78,$57,$4D,$FF,$CF,$D8,$78,$E3,$15,$FE,$E6,$8C,$5F,$26
	.dc.b	$C7,$76,$DF,$96,$3B,$5F,$ED,$F0,$F2,$CE,$6B,$73,$E0,$AF,$31,$CC
	.dc.b	$B3,$E6,$BF,$F1,$96,$BF,$2A,$58,$7C,$08,$C4,$7C,$4B,$F0,$63,$F4
	.dc.b	$27,$85,$6E,$FC,$9C,$19,$76,$76,$32,$5C,$3F,$C8,$53,$FE,$34,$8C
	.dc.b	$53,$E8,$1D,$FC,$EF,$7D,$56,$B5,$39,$9F,$4C,$E6,$B8,$DF,$79,$CB
	.dc.b	$DA,$33,$1F,$E9,$5C,$2E,$1E,$CC,$3C,$1F,$17,$FE,$D3,$F1,$7B,$56
	.dc.b	$B1,$FA,$55,$AF,$9F,$7D,$06,$A3,$0D,$9F,$81,$82,$B7,$1E,$5B,$E9
	.dc.b	$92,$B5,$7E,$DF,$5D,$D3,$4F,$89,$83,$E7,$B8,$3E,$F9,$DB,$67,$73
	.dc.b	$F5,$F0,$FF,$39,$56,$9F,$F6,$15,$73,$E1,$EC,$FD,$73,$07,$33,$15
	.dc.b	$CE,$5F,$40,$F7,$BE,$A3,$6B,$7C,$78,$99,$1C,$9B,$5E,$2B,$E0,$FF
	.dc.b	$B2,$6E,$F0,$FF,$00,$BF,$81,$7C,$23,$17,$A9,$FE,$B8,$7F,$97,$8F
	.dc.b	$C1,$16,$FB,$77,$7E,$26,$31,$5F,$A3,$69,$FE,$97,$1F,$41,$FE,$52
	.dc.b	$79,$F5,$7D,$28,$C5,$FF,$B7,$E7,$8B,$FD,$9A,$33,$5D,$6B,$87,$97
	.dc.b	$F9,$9F,$9B,$FC,$0D,$9F,$C1,$E1,$F7,$70,$D7,$2C,$8F,$A3,$E2,$5B
	.dc.b	$F9,$0D,$CB,$E1,$B6,$7D,$5A,$7F,$86,$DF,$D2,$7C,$F3,$D0,$FD,$53
	.dc.b	$63,$F5,$1F,$40,$FA,$0C,$D6,$BE,$06,$57,$28,$B9,$FF,$BF,$FE,$81
	.dc.b	$CF,$FB,$56,$D7,$82,$F1,$3F,$B9,$EA,$FF,$74,$CF,$ED,$DB,$6D,$BF
	.dc.b	$8B,$69,$F2,$AB,$FC,$31,$5A,$9F,$E3,$4F,$F9,$56,$3F,$80,$7D,$17
	.dc.b	$F9,$7D,$E9,$E9,$7F,$07,$8C,$61,$D6,$7F,$C4,$5F,$07,$FE,$AA,$7F
	.dc.b	$84,$7D,$CB,$E3,$FF,$87,$4E,$6B,$81,$FE,$DB,$9D,$57,$F0,$FF,$A0
	.dc.b	$79,$DF,$C5,$5B,$1F,$C0,$BE,$8B,$96,$BF,$CC,$67,$38,$74,$0E,$23
	.dc.b	$FA,$FE,$AA,$FD,$C4,$71,$C3,$F8,$BC,$63,$0F,$F1,$E5,$F1,$7F,$DE
	.dc.b	$17,$C7,$C1,$F2,$5F,$07,$F9,$77,$57,$F7,$BA,$BE,$67,$C6,$CB,$FB
	.dc.b	$7C,$E6,$7F,$9D,$3F,$E0,$1F,$63,$71,$C1,$AE,$AB,$93,$FF,$00,$67
	.dc.b	$FC,$11,$F4,$5D,$8E,$FF,$13,$67,$84,$5D,$BE,$7D,$7F,$83,$EB,$18
	.dc.b	$FC,$9F,$8D,$F9,$C7,$F4,$7F,$21,$EE,$3E,$53,$39,$FD,$5A,$FE,$95
	.dc.b	$F4,$5F,$4A,$3E,$3C,$63,$E1,$B7,$A7,$DE,$C9,$7B,$81,$F5,$AB,$85
	.dc.b	$FB,$FE,$33,$E4,$9F,$2B,$E8,$AC,$64,$95,$EC,$E0,$4E,$A3,$1D,$3F
	.dc.b	$4B,$D0,$FA,$6D,$BF,$98,$D9,$DC,$7C,$F9,$FE,$C9,$F4,$5F,$EC,$A8
	.dc.b	$FE,$C4,$D9,$FC,$CF,$07,$C6,$F0,$B7,$72,$FF,$AB,$FD,$13,$E4,$F1
	.dc.b	$87,$E8,$3A,$93,$9A,$EB,$7F,$DA,$23,$15,$F9,$BC,$BC,$1C,$3F,$9A
	.dc.b	$9C,$BF,$8C,$D9,$F4,$5F,$90,$FA,$2F,$C8,$B7,$2F,$C5,$37,$1F,$90
	.dc.b	$DB,$E5,$C9,$F9,$8E,$BF,$CC,$B6,$FF,$24,$DB,$FA,$0D,$AF,$F1,$17
	.dc.b	$D0,$BE,$87,$F8,$CD,$7F,$A9,$4F,$BF,$19,$F1,$B1,$18,$EB,$FE,$F0
	.dc.b	$E2,$74,$EB,$E9,$FF,$A4,$BC,$03,$87,$5E,$7E,$5C,$B7,$E8,$9E,$95
	.dc.b	$24,$F3,$8C,$BD,$7D,$07,$6F,$FA,$F7,$5F,$F1,$F4,$3E,$8B,$1C,$1F
	.dc.b	$EC,$18,$C5,$FF,$89,$DE,$2D,$BD,$BF,$A0,$8D,$BF,$C5,$DB,$7D,$A7
	.dc.b	$73,$FE,$E9,$F8,$7B,$4F,$9A,$A6,$22,$4F,$86,$32,$F8,$E3,$9F,$D1
	.dc.b	$3E,$8A,$92,$7D,$1E,$DD,$7F,$95,$4E,$DF,$F2,$13,$5F,$44,$6F,$2F
	.dc.b	$F3,$14,$E5,$B5,$E8,$9D,$67,$FE,$E7,$AD,$36,$75,$7D,$9D,$BF,$C1
	.dc.b	$A7,$6F,$C9,$AC,$A8,$CB,$67,$9B,$EA,$77,$1A,$C9,$58,$9E,$71,$8A
	.dc.b	$F1,$10,$E4,$7C,$63,$18,$7B,$9B,$87,$E0,$EF,$A4,$60,$F4,$14,$F3
	.dc.b	$16,$BC,$0F,$62,$57,$57,$E7,$9C,$70,$F7,$9E,$BA,$EF,$1C,$2F,$C4
	.dc.b	$EF,$0F,$6C,$EA,$FA,$1F,$9C,$71,$5F,$4E,$BD,$C7,$33,$E9,$57,$27
	.dc.b	$A1,$E9,$9E,$2E,$8E,$35,$DA,$BA,$47,$31,$87,$C0,$9D,$47,$13,$51
	.dc.b	$99,$E3,$3A,$9F,$92,$7C,$43,$A6,$1F,$13,$0F,$C0,$CE,$E3,$13,$E4
	.dc.b	$AE,$64,$7C,$13,$F0,$DF,$69,$E2,$C3,$DA,$F8,$23,$4F,$82,$FE,$23
	.dc.b	$4F,$88,$72,$8F,$46,$1E,$B8,$EF,$7E,$17,$A3,$F6,$1E,$47,$84,$E4
	.dc.b	$EE,$39,$47,$03,$99,$0F,$9B,$7E,$8C,$3F,$0D,$1C,$8D,$4F,$28,$F8
	.dc.b	$B1,$DC,$F0,$DF,$A2,$B9,$61,$F3,$30,$F4,$5F,$0A,$F8,$67,$CF,$9F
	.dc.b	$86,$DB,$F8,$67,$4A,$F5,$5F,$48,$FA,$87,$A6,$FA,$4F,$C3,$AE,$53
	.dc.b	$E5,$69,$E5,$7D,$87,$C3,$3E,$09,$FE,$D9,$FA,$27,$9F,$FC,$B1,$7A
	.dc.b	$74,$C4,$49,$E1,$19,$79,$47,$1F,$86,$91,$D5,$E9,$C5,$88,$FF,$29
	.dc.b	$5A,$71,$83,$1F,$F1,$0D,$2B,$FA,$8B,$FE,$59,$FA,$2F,$F3,$43,$FE
	.dc.b	$F1,$6B,$9F,$1B,$D5,$D3,$FD,$1A,$73,$87,$5A,$FF,$16,$C6,$30,$EE
	.dc.b	$D8,$FF,$18,$E1,$FD,$D9,$E8,$77,$C3,$A1,$D9,$FF,$1B,$DF,$55,$F4
	.dc.b	$0F,$A2,$FE,$AE,$D8,$DE,$3E,$E8,$C4,$7F,$9D,$57,$C6,$70,$A9,$3C
	.dc.b	$0F,$F2,$2C,$F0,$BF,$03,$C2,$34,$78,$BE,$4B,$C3,$F4,$52,$33,$29
	.dc.b	$F2,$99,$8F,$14,$E1,$E7,$DD,$FE,$37,$7F,$C1,$CD,$7F,$C2,$ED,$7F
	.dc.b	$C3,$8D,$7E,$DB,$BB,$F8,$CD,$CB,$3E,$4F,$F7,$CF,$A1,$FF,$02,$DF
	.dc.b	$FA,$5E,$B4,$FF,$99,$1A,$9B,$1C,$2F,$E9,$2E,$9E,$84,$AF,$8B,$1F
	.dc.b	$E0,$76,$C6,$5F,$F5,$34,$71,$77,$D5,$E5,$E8,$F3,$7C,$1C,$9E,$71
	.dc.b	$97,$AC,$65,$E7,$1F,$E1,$8C,$7F,$5B,$1F,$5B,$CF,$7D,$0F,$F8,$71
	.dc.b	$B1,$F4,$3A,$B8,$11,$C1,$C7,$AB,$0F,$C7,$8B,$AF,$E4,$EB,$F7,$B3
	.dc.b	$DF,$A7,$B7,$54,$8C,$10,$FF,$63,$4F,$EC,$2B,$94,$F9,$6B,$CD,$D5
	.dc.b	$2F,$D4,$FA,$7E,$87,$11,$CA,$2D,$F8,$E7,$FC,$73,$F4,$0F,$8C,$FF
	.dc.b	$1D,$FD,$03,$E8,$69,$A8,$98,$FE,$72,$B8,$36,$38,$F5,$EA,$3F,$DB
	.dc.b	$D3,$A6,$CE,$9F,$D2,$FD,$03,$E8,$7F,$C9,$6D,$8C,$C6,$3C,$CF,$F2
	.dc.b	$7F,$CF,$FC,$3E,$CF,$E2,$B6,$9D,$5F,$EC,$2B,$4F,$FA,$EB,$E8,$7F
	.dc.b	$29,$7F,$C7,$27,$32,$6C,$EA,$3F,$D8,$3F,$41,$FD,$6A,$5F,$43,$E2
	.dc.b	$3F,$CC,$31,$FD,$78,$9D,$58,$3F,$CC,$BF,$40,$F8,$FA,$67,$0F,$E0
	.dc.b	$B5,$FD,$9A,$75,$1F,$C2,$3D,$BF,$E1,$4B,$97,$F8,$E4,$BE,$87,$A9
	.dc.b	$FE,$72,$AD,$CF,$F2,$16,$EB,$FC,$E9,$B7,$FE,$DC,$6A,$BF,$E5,$D5
	.dc.b	$C7,$C7,$E8,$49,$FE,$37,$3A,$BF,$E8,$BF,$03,$E8,$9F,$34,$7F,$48
	.dc.b	$F9,$B7,$BB,$FE,$86,$6C,$FF,$24,$AE,$18,$FF,$37,$6D,$F4,$D7,$FD
	.dc.b	$1B,$18,$AF,$F7,$9D,$BE,$17,$35,$27,$A3,$DE,$3F,$D2,$BE,$53,$E8
	.dc.b	$7F,$D2,$FF,$3E,$F8,$DA,$6A,$26,$1F,$CE,$2B,$83,$63,$8F,$C9,$EA
	.dc.b	$3D,$E0,$FE,$75,$19,$D7,$FA,$96,$0A,$FF,$53,$FD,$07,$E8,$E7,$3F
	.dc.b	$43,$E4,$33,$DB,$31,$FE,$93,$5B,$E0,$FE,$E6,$D8,$F0,$63,$FC,$95
	.dc.b	$F1,$88,$D3,$6B,$A3,$A9,$E1,$87,$F6,$7F,$A1,$FC,$1C,$7F,$98,$27
	.dc.b	$FA,$CB,$5E,$B1,$F8,$82,$72,$7F,$B0,$9E,$3F,$2F,$62,$4F,$F6,$29
	.dc.b	$CB,$7E,$07,$E9,$FC,$74,$8C,$10,$EF,$1D,$4F,$B3,$D7,$FE,$D0,$FA
	.dc.b	$17,$35,$25,$7F,$70,$FA,$1F,$F0,$E6,$BF,$57,$E8,$1F,$41,$FE,$1C
	.dc.b	$C3,$8D,$FF,$76,$6B,$F2,$9B,$5F,$E0,$BB,$E7,$2B,$E6,$FE,$06,$33
	.dc.b	$18,$85,$74,$8C,$36,$FF,$C5,$B5,$C1,$BE,$67,$EF,$EF,$B4,$79,$F1
	.dc.b	$FF,$19,$E1,$C9,$B1,$C3,$1F,$73,$6B,$89,$F9,$5C,$7C,$0D,$AE,$2F
	.dc.b	$A7,$0F,$06,$1E,$9C,$7C,$93,$B3,$A9,$4E,$A1,$DB,$67,$9E,$1E,$98
	.dc.b	$E2,$7C,$4B,$F0,$BB,$C7,$37,$CB,$1F,$3C,$EC,$CE,$1D,$1F,$45,$74
	.dc.b	$9D,$9C,$2B,$C6,$7D,$89,$FF,$2D,$9D,$CA,$53,$2E,$DF,$AB,$E6,$BE
	.dc.b	$27,$71,$EF,$8C,$3C,$A5,$18,$8E,$6D,$7B,$FD,$0F,$27,$F8,$43,$53
	.dc.b	$48,$FC,$EF,$8E,$7C,$F7,$F9,$A7,$6F,$C7,$4F,$23,$51,$9D,$BB,$C6
	.dc.b	$3A,$3E,$8B,$BA,$FC,$B5,$F4,$9C,$CF,$88,$E0,$7D,$E3,$85,$7F,$A1
	.dc.b	$27,$84,$64,$E1,$5D,$2F,$C5,$1F,$43,$0F,$C0,$D7,$0A,$F4,$CF,$0B
	.dc.b	$E9,$7C,$23,$FD,$23,$7C,$23,$11,$E2,$3F,$98,$5F,$3A,$F7,$DF,$87
	.dc.b	$0F,$5E,$1C,$5B,$7E,$C9,$D3,$FC,$A2,$FD,$7A,$FF,$A2,$63,$B0,$8C
	.dc.b	$D7,$98,$FF,$3E,$AE,$0F,$F4,$E5,$7F,$A4,$63,$75,$E6,$9C,$D7,$C9
	.dc.b	$E7,$B3,$FD,$F7,$E8,$7F,$A3,$4F,$49,$CE,$09,$CD,$79,$E7,$31,$89
	.dc.b	$F0,$46,$3A,$FA,$D7,$A5,$B8,$EF,$0F,$EB,$1F,$21,$FE,$D0,$F0,$F3
	.dc.b	$52,$7F,$D3,$D2,$FA,$17,$D0,$E7,$1F,$EC,$33,$98,$C6,$7F,$EB,$D8
	.dc.b	$C5,$FC,$6C,$3D,$F1,$8B,$FE,$AF,$EA,$BA,$7C,$ED,$BF,$F5,$70,$FA
	.dc.b	$05,$EA,$A4,$EA,$7E,$7D,$AF,$93,$FD,$DA,$79,$72,$70,$8C,$57,$0A
	.dc.b	$DE,$5F,$ED,$5B,$D9,$FE,$BA,$FA,$17,$BB,$F3,$1B,$1E,$1E,$3C,$4C
	.dc.b	$8F,$23,$71,$EE,$DB,$ED,$8F,$29,$FE,$F3,$85,$9C,$D7,$DA,$BF,$E1
	.dc.b	$A7,$FB,$30,$B5,$EA,$F0,$35,$D1,$27,$E4,$FC,$6F,$F9,$45,$E1,$1F
	.dc.b	$90,$51,$9A,$FF,$04,$C7,$4C,$3E,$14,$E5,$FF,$76,$46,$4F,$F1,$01
	.dc.b	$CB,$C9,$FC,$7F,$CC,$7E,$36,$FF,$C4,$8D,$9F,$B5,$7A,$E2,$91,$F6
	.dc.b	$7E,$2E,$DF,$D4,$6E,$3F,$19,$B3,$E0,$D9,$1F,$9E,$F9,$E7,$9F,$4F
	.dc.b	$D0,$FD,$0B,$E8,$71,$87,$FB,$D2,$9F,$01,$C4,$CE,$7F,$E2,$A8,$CC
	.dc.b	$FF,$BE,$BE,$87,$F1,$8D,$E6,$E0,$D3,$FD,$FF,$D3,$FE,$DF,$6F,$C3
	.dc.b	$F9,$78,$FF,$67,$39,$C7,$FC,$6E,$FF,$94,$EB,$FC,$95,$C0,$7F,$C9
	.dc.b	$92,$FA,$1F,$F3,$04,$76,$F4,$BF,$7B,$18,$9F,$F2,$0F,$C2,$7F,$A8
	.dc.b	$EB,$99,$CB,$E8,$5F,$1B,$FC,$12,$75,$5F,$E4,$89,$CD,$70,$87,$47
	.dc.b	$F6,$18,$FE,$2E,$FF,$63,$F5,$AF,$97,$E7,$9E,$13,$96,$C6,$6B,$FC
	.dc.b	$CF,$39,$6C,$67,$4F,$81,$1E,$08,$F6,$46,$27,$C5,$5D,$2F,$C2,$F8
	.dc.b	$9F,$1F,$CF,$C8,$F0,$FC,$4D,$C7,$BC,$DC,$3A,$62,$AC,$5F,$D0,$FA
	.dc.b	$07,$CA,$7F,$2A,$6C,$7C,$5F,$25,$ED,$BE,$83,$57,$2B,$FC,$0F,$7C
	.dc.b	$34,$FF,$07,$B5,$FF,$09,$B5,$FD,$3E,$4F,$8B,$F9,$A7,$CE,$F9,$FB
	.dc.b	$7F,$32,$2D,$7C,$A9,$BA,$FB,$26,$D7,$49,$B5,$F3,$78,$FE,$CF,$9B
	.dc.b	$F6,$9B,$5F,$61,$B3,$F5,$9B,$87,$D6,$87,$CF,$FF,$AF,$AB,$F1,$11
	.dc.b	$8B,$FE,$BB,$BF,$C6,$F0,$7F,$A9,$DE,$EB,$FD,$2E,$D7,$FC,$6A,$D7
	.dc.b	$F4,$2D,$7F,$D5,$CD,$E4,$FF,$1A,$7C,$EF,$E4,$BE,$7B,$F2,$6D,$CB
	.dc.b	$F2,$2D,$BF,$C7,$B7,$3F,$C0,$B6,$7E,$8B,$67,$F2,$5F,$11,$F9,$EF
	.dc.b	$3F,$F3,$ED,$AF,$CE,$B6,$7F,$36,$DB,$F8,$DF,$47,$FC,$22,$7F,$C9
	.dc.b	$ED,$7F,$CA,$6D,$7F,$7E,$D7,$FD,$99,$F3,$CF,$9F,$FE,$B9,$1F,$E5
	.dc.b	$88,$C7,$87,$FE,$54,$7F,$35,$E0,$DA,$DD,$7F,$98,$3C,$9F,$CD,$BF
	.dc.b	$59,$7C,$EF,$F0,$EC,$BE,$E4,$62,$75,$D5,$9D,$7F,$89,$CF,$29,$FF
	.dc.b	$67,$BB,$6D,$ED,$FF,$72,$46,$EF,$AD,$E6,$7F,$9C,$D5,$AE,$55,$A9
	.dc.b	$F3,$78,$B9,$F9,$59,$18,$21,$E2,$72,$F9,$23,$B7,$CE,$3E,$77,$B7
	.dc.b	$C1,$4A,$FC,$BC,$E5,$B3,$97,$DB,$3C,$A3,$D1,$F3,$8F,$1E,$F9,$B2
	.dc.b	$FF,$43,$39,$6C,$E5,$FF,$7A,$C7,$F5,$07,$24,$33,$3B,$CF,$F5,$D3
	.dc.b	$CE,$78,$D6,$77,$FE,$8B,$3C,$E7,$27,$3A,$E3,$7B,$80,$9E,$93,$DA
	.dc.b	$F5,$39,$AF,$7C,$48,$F6,$C6,$AB,$8D,$F8,$AB,$13,$BA,$F1,$9E,$C3
	.dc.b	$89,$E5,$9E,$0E,$0D,$1C,$2F,$C8,$E5,$F5,$BC,$4F,$5D,$F8,$9B,$5E
	.dc.b	$5B,$F1,$57,$09,$EA,$F4,$7B,$5F,$CA,$AE,$11,$D4,$F8,$47,$86,$3C
	.dc.b	$EF,$03,$B9,$93,$C7,$5E,$08,$E9,$19,$3E,$84,$65,$E2,$79,$23,$AD
	.dc.b	$79,$4F,$41,$1C,$92,$38,$C6,$5F,$09,$E9,$BF,$C2,$46,$63,$C2,$66
	.dc.b	$FD,$EF,$AC,$F7,$9E,$33,$99,$99,$E7,$5E,$C8,$C5,$FE,$01,$F0,$35
	.dc.b	$C9,$F0,$67,$8B,$D2,$FD,$D5,$CD,$F3,$4F,$DB,$9F,$39,$E3,$3C,$B1
	.dc.b	$C1,$E5,$3C,$48,$E0,$87,$07,$2F,$48,$E4,$E4,$F9,$71,$F3,$5D,$3C
	.dc.b	$27,$D3,$18,$3A,$B8,$25,$7D,$D8,$E8,$48,$F6,$12,$FE,$8D,$F8,$A3
	.dc.b	$94,$71,$75,$3F,$26,$B1,$3C,$2F,$CD,$5C,$2B,$E8,$9F,$3E,$79,$57
	.dc.b	$D3,$BE,$CD,$BE,$2D,$AF,$5B,$5F,$71,$2F,$B3,$F7,$0F,$A8,$FC,$07
	.dc.b	$F1,$87,$9C,$A6,$A3,$D2,$E4,$FE,$47,$F0,$5F,$CB,$08,$FF,$2A,$F5
	.dc.b	$2B,$A3,$F9,$6C,$7F,$32,$8F,$E9,$5F,$3B,$FE,$5C,$9F,$E6,$3E,$37
	.dc.b	$F3,$63,$3E,$14,$8C,$90,$67,$FE,$1B,$AE,$51,$8C,$FF,$CE,$3F,$3A
	.dc.b	$F9,$6F,$EB,$0D,$9F,$F0,$FB,$5E,$07,$F2,$E5,$F1,$1C,$26,$4F,$07
	.dc.b	$9C,$62,$38,$89,$E9,$87,$F8,$BD,$B7,$E4,$AF,$1F,$C6,$F9,$3E,$76
	.dc.b	$61,$E4,$8C,$BE,$68,$CB,$DA,$D3,$FD,$CF,$CD,$7C,$EF,$F7,$5F,$9D
	.dc.b	$7D,$1D,$35,$13,$0F,$F4,$A9,$FE,$96,$AE,$2E,$0F,$E9,$EE,$1C,$35
	.dc.b	$E3,$4E,$9F,$01,$23,$FC,$97,$0E,$AC,$47,$C6,$8E,$0E,$A7,$D1,$3A
	.dc.b	$DB,$FC,$AD,$5A,$AE,$15,$A7,$AF,$C0,$FB,$FE,$02,$46,$08,$77,$72
	.dc.b	$F8,$63,$E4,$74,$FF,$86,$FC,$F7,$CF,$53,$35,$FE,$64,$D3,$FB,$B3
	.dc.b	$F2,$1A,$FC,$9A,$F9,$0F,$C8,$6B,$FE,$83,$E9,$ED,$D3,$23,$FC,$DF
	.dc.b	$82,$7C,$46,$63,$F1,$70,$9C,$C7,$09,$D6,$BF,$E7,$BA,$D4,$E4,$EB
	.dc.b	$E2,$FB,$48,$F8,$3E,$7A,$BD,$A2,$D7,$39,$FF,$1F,$7C,$E3,$E2,$3F
	.dc.b	$C8,$1F,$3C,$F9,$EA,$6A,$26,$3F,$0A,$B8,$36,$3E,$1C,$62,$3F,$1D
	.dc.b	$5F,$9A,$7F,$D2,$2B,$A7,$B9,$E0,$E8,$93,$A8,$CB,$C2,$32,$FE,$CA
	.dc.b	$79,$3B,$34,$7F,$A7,$BE,$75,$F3,$DF,$E5,$B6,$C6,$63,$1E,$77,$F9
	.dc.b	$7F,$E7,$7E,$6C,$FE,$FA,$93,$A8,$FF,$54,$CE,$B6,$FF,$57,$7C,$F7
	.dc.b	$F9,$A7,$C0,$FF,$1C,$D2,$B5,$18,$6C,$7F,$8F,$A7,$83,$FE,$40,$F9
	.dc.b	$D7,$4F,$8F,$E7,$A4,$FF,$AE,$27,$2D,$9F,$F3,$B4,$62,$3F,$8A,$B7
	.dc.b	$1F,$F6,$17,$CE,$BE,$17,$FC,$F8,$DA,$E4,$43,$F8,$EB,$97,$F8,$FF
	.dc.b	$B5,$FC,$99,$FF,$64,$C3,$E7,$B9,$9F,$E8,$79,$D4,$F0,$F1,$B0,$47
	.dc.b	$8F,$4D,$74,$7F,$B5,$27,$8E,$3E,$89,$FF,$49,$7C,$0F,$8E,$3F,$DB
	.dc.b	$9F,$32,$F7,$FF,$D2,$CD,$C3,$FD,$BB,$5B,$E2,$FA,$6D,$8F,$F3,$6B
	.dc.b	$53,$65,$F0,$9E,$0D,$A5,$74,$7B,$C7,$FA,$97,$A1,$F3,$DF,$EA,$7F
	.dc.b	$9C,$7C,$4D,$35,$13,$0F,$F7,$8D,$70,$6C,$71,$F8,$FD,$47,$F4,$E9
	.dc.b	$D3,$67,$4F,$FB,$EB,$E7,$7F,$5D,$BF,$FB,$FF,$17,$CF,$78,$CF,$F5
	.dc.b	$D4,$EB,$7F,$CA,$B6,$FF,$D1,$ED,$4D,$AF,$E6,$96,$AE,$3E,$0F,$42
	.dc.b	$5F,$FA,$62,$7C,$5E,$5F,$FB,$27,$E7,$BE,$DD,$FE,$93,$07,$B7,$F1
	.dc.b	$CD,$73,$CC,$79,$57,$EA,$AF,$8C,$7F,$A9,$1A,$AD,$F2,$9C,$63,$FB
	.dc.b	$AC,$E8,$DC,$EA,$3A,$9C,$FE,$0F,$EC,$7B,$92,$32,$29,$3E,$B2,$46
	.dc.b	$A3,$2F,$0A,$FD,$86,$1F,$C0,$BE,$79,$CF,$FE,$EB,$6D,$76,$6B,$FE
	.dc.b	$0C,$F9,$EF,$E8,$DB,$FF,$B0,$BE,$71,$F3,$DC,$F0,$E4,$4B,$F3,$DF
	.dc.b	$05,$5C,$1F,$F6,$5C,$66,$31,$0C,$3C,$31,$86,$DF,$F1,$AA,$E0,$DF
	.dc.b	$3F,$FC,$67,$7D,$A3,$C8,$F9,$DA,$E4,$F0,$C7,$0B,$F0,$D7,$49,$FF
	.dc.b	$6F,$DF,$81,$B8,$7F,$2A,$AD,$DF,$1C,$3C,$58,$F7,$BD,$94,$F6,$5F
	.dc.b	$19,$DC,$78,$C6,$3E,$17,$67,$1C,$7C,$4E,$B0,$F0,$3E,$AC,$35,$87
	.dc.b	$58,$F0,$9D,$6F,$A1,$D6,$F9,$9D,$1C,$DF,$CA,$9F,$F1,$AC,$7D,$96
	.dc.b	$C6,$5F,$49,$AF,$41,$F1,$DF,$E1,$2A,$FF,$04,$36,$33,$3F,$E3,$D9
	.dc.b	$CB,$FE,$16,$F9,$EF,$F0,$BA,$69,$1D,$1F,$1F,$E3,$9F,$37,$FC,$06
	.dc.b	$3E,$50,$39,$1B,$75,$B7,$E3,$A3,$55,$F5,$6B,$9E,$FF,$5E,$72,$71
	.dc.b	$70,$F4,$AF,$CB,$CF,$08,$F0,$99,$DF,$FB,$7C,$66,$7C,$13,$93,$CD
	.dc.b	$19,$9F,$F3,$05,$F8,$F0,$F3,$C6,$2B,$C4,$76,$9D,$61,$E4,$6C,$64
	.dc.b	$DE,$1E,$08,$CB,$63,$35,$E6,$BF,$95,$BF,$CB,$96,$3E,$D8,$C4,$7C
	.dc.b	$4B,$F8,$35,$C3,$1E,$47,$F1,$6B,$FF,$03,$B6,$35,$5F,$03,$C8,$AF
	.dc.b	$6E,$9C,$E2,$F9,$E7,$7E,$B2,$FF,$21,$65,$FC,$9A,$B5,$39,$BE,$37
	.dc.b	$C2,$BC,$51,$89,$E4,$48,$FE,$5F,$F1,$4F,$84,$FE,$F1,$1E,$99,$D6
	.dc.b	$5F,$EB,$4A,$7C,$F3,$E7,$B3,$9F,$E4,$67,$31,$8C,$3F,$25,$18,$7A
	.dc.b	$1E,$0A,$E4,$FF,$3C,$F2,$7F,$9D,$B7,$AF,$E3,$AF,$9C,$5F,$C8,$AD
	.dc.b	$E5,$FE,$C0,$6E,$5F,$A3,$BE,$1C,$3B,$C7,$9E,$3F,$6E,$7F,$C8,$77
	.dc.b	$C1,$FD,$41,$C0,$F1,$1F,$D5,$0C,$61,$FE,$CE,$F9,$E7,$BF,$F6,$9B
	.dc.b	$3B,$BF,$C0,$D6,$EB,$47,$88,$B3,$B6,$C7,$FB,$8A,$3E,$04,$63,$2F
	.dc.b	$65,$76,$AF,$F6,$FD,$7F,$A5,$4F,$EC,$A5,$79,$47,$F7,$28,$CF,$47
	.dc.b	$E4,$1D,$57,$C7,$F9,$EF,$F2,$94,$68,$F8,$0E,$2F,$BD,$7D,$48,$C5
	.dc.b	$7E,$9E,$B9,$9F,$2A,$32,$FF,$9C,$3E,$53,$F2,$97,$FE,$F6,$6D,$7B
	.dc.b	$2B,$F4,$71,$8A,$E1,$C3,$F6,$3C,$4C,$B6,$3F,$C0,$73,$BC,$3F,$C4
	.dc.b	$59,$7F,$86,$A7,$35,$B9,$63,$FB,$DF,$9C,$7A,$1F,$B2,$6D,$FF,$83
	.dc.b	$BE,$79,$F3,$D9,$AD,$7C,$14,$AF,$F0,$9D,$F0,$C7,$AC,$7F,$A3,$A1
	.dc.b	$3F,$BE,$F9,$E7,$1D,$39,$5F,$03,$F9,$C4,$6A,$FF,$C3,$FC,$DF,$EF
	.dc.b	$D6,$F9,$7E,$7F,$4B,$F8,$4C,$E6,$F9,$1F,$A6,$F4,$7F,$86,$FC,$F7
	.dc.b	$F9,$A3,$0E,$71,$8F,$4F,$F8,$8F,$77,$FA,$AA,$7F,$C5,$BF,$3C,$F8
	.dc.b	$FC,$FB,$DF,$C5,$A3,$13,$FE,$36,$E8,$FE,$BE,$FF,$1C,$7F,$B0,$E1
	.dc.b	$7C,$AF,$F2,$A4,$ED,$F1,$69,$EF,$AE,$33,$9A,$E0,$F1,$6B,$93,$91
	.dc.b	$93,$A7,$CF,$77,$3C,$1F,$11,$63,$24,$BF,$06,$09,$D4,$7C,$8F,$9E
	.dc.b	$7C,$BF,$F3,$46,$DF,$F9,$53,$C9,$FF,$49,$FC,$9E,$6A,$4F,$F3,$3F
	.dc.b	$A2,$F9,$58,$73,$9F,$E7,$71,$8E,$EF,$EE,$DB,$7F,$81,$BE,$33,$E3
	.dc.b	$1F,$1F,$E7,$EE,$3F,$1E,$2E,$3E,$89,$B7,$CB,$65,$62,$FF,$C0,$18
	.dc.b	$7F,$84,$AF,$18,$23,$FC,$2B,$C7,$F5,$7E,$03,$EB,$36,$FE,$B3,$6B
	.dc.b	$E3,$B7,$2F,$EC,$9F,$3F,$FD,$1A,$7F,$1E,$D8,$FF,$09,$6F,$FE,$B8
	.dc.b	$F0,$3F,$AE,$CE,$70,$FF,$78,$B6,$FF,$9D,$FD,$13,$E4,$F7,$9F,$F6
	.dc.b	$6E,$C4,$E7,$0F,$C5,$7C,$FF,$E3,$7E,$83,$F1,$CD,$8C,$9F,$8F,$8C
	.dc.b	$B4,$CE,$3F,$82,$8C,$B6,$3F,$2D,$D7,$F9,$9F,$3B,$F3,$4D,$8C,$9F
	.dc.b	$9B,$8C,$B4,$FF,$4A,$FC,$FB,$E7,$FF,$7F,$7F,$E3,$32,$46,$7C,$5C
	.dc.b	$46,$3B,$3F,$AE,$B8,$13,$B7,$1F,$44,$F3,$B5,$1F,$E3,$F8,$C3,$86
	.dc.b	$BC,$23,$6F,$FA,$FB,$B3,$C9,$CD,$FE,$1F,$F3,$9F,$41,$8D,$7D,$11
	.dc.b	$8B,$FF,$58,$BF,$A6,$6D,$ED,$FF,$28,$46,$DF,$E2,$6D,$BE,$D3,$B9
	.dc.b	$F1,$1D,$2F,$15,$BB,$F0,$D6,$3C,$8C,$FC,$CC,$C3,$C7,$19,$7C,$B1
	.dc.b	$97,$C3,$1D,$FE,$81,$F4,$14,$93,$EC,$F6,$EB,$FC,$AA,$76,$FF,$B4
	.dc.b	$1A,$FB,$23,$D5,$F4,$1F,$E2,$98,$FF,$6A,$BF,$D0,$1B,$E7,$EB,$CB
	.dc.b	$91,$82,$1F,$CE,$27,$93,$63,$AC,$C8,$EB,$1C,$9E,$11,$97,$81,$A3
	.dc.b	$AE,$FF,$E5,$BA,$F2,$46,$89,$19,$9E,$4E,$EB,$B3,$99,$E7,$AF,$E6
	.dc.b	$EF,$CF,$3B,$9C,$9F,$E9,$18,$EF,$19,$8E,$F3,$C6,$FB,$BC,$EF,$BB
	.dc.b	$99,$E6,$2F,$BC,$6E,$33,$1D,$A3,$11,$D8,$F1,$57,$67,$C9,$5D,$9C
	.dc.b	$9C,$9C,$3D,$E3,$83,$E1,$8C,$D7,$78,$E2,$77,$34,$79,$27,$73,$CA
	.dc.b	$78,$4E,$8E,$0E,$09,$1D,$A7,$83,$67,$CC,$7D,$D3,$F0,$26,$0E,$71
	.dc.b	$C1,$D5,$FD,$97,$67,$82,$7C,$93,$E6,$8F,$41,$C2,$FE,$5D,$71,$23
	.dc.b	$82,$47,$C2,$8C,$BF,$38,$E4,$66,$3E,$E3,$E1,$3B,$94,$F2,$C7,$27
	.dc.b	$E3,$3F,$89,$31,$7F,$1A,$3A,$9F,$5C,$F8,$51,$B9,$EE,$F0,$BF,$95
	.dc.b	$3D,$A7,$D2,$F6,$76,$47,$D4,$7C,$C3,$AB,$C2,$BE,$C5,$FC,$99,$E8
	.dc.b	$7B,$CE,$37,$E3,$70,$79,$CF,$80,$FB,$23,$93,$63,$DB,$3F,$A0,$78
	.dc.b	$BF,$05,$F2,$1E,$03,$D4,$78,$2B,$D4,$F7,$9F,$13,$A3,$11,$A9,$ED
	.dc.b	$19,$7D,$AF,$8A,$FF,$0E,$D3,$EC,$D7,$49,$E2,$70,$8F,$6D,$75,$BF
	.dc.b	$D4,$FC,$0F,$F2,$82,$4F,$F2,$8F,$5A,$BF,$C8,$E2,$E1,$FC,$66,$7F
	.dc.b	$A4,$7D,$07,$F9,$56,$BF,$B0,$B5,$FE,$D0,$D7,$FB,$6B,$5F,$EE,$4E
	.dc.b	$E7,$FD,$FC,$DC,$33,$E2,$7F,$3C,$E8,$FE,$AD,$F3,$EF,$A0,$A4,$C4
	.dc.b	$7F,$9C,$21,$F1,$79,$8C,$4D,$C9,$8F,$F8,$0E,$F7,$5C,$A7,$2F,$48
	.dc.b	$F0,$39,$9E,$15,$CD,$E5,$7E,$6F,$8F,$F3,$FC,$D4,$BE,$07,$38,$D3
	.dc.b	$81,$82,$78,$47,$AE,$7F,$BB,$76,$FF,$8D,$63,$FC,$1E,$D7,$FC,$2C
	.dc.b	$D7,$FC,$36,$D7,$FC,$40,$EF,$2F,$6C,$7F,$9C,$7C,$6E,$A7,$F8,$0F
	.dc.b	$E7,$FF,$C1,$91,$FE,$9A,$5D,$19,$24,$7A,$61,$D3,$88,$E1,$36,$FB
	.dc.b	$D6,$A7,$84,$E7,$5F,$F1,$A4,$F6,$9E,$97,$E3,$E7,$F2,$7B,$D2,$7C
	.dc.b	$11,$97,$C5,$19,$7D,$11,$F2,$39,$BF,$C3,$DE,$73,$E7,$E9,$9A,$FA
	.dc.b	$5B,$7F,$44,$25,$70,$EA,$E3,$D3,$CB,$A6,$5F,$D9,$C5,$3A,$8C,$4E
	.dc.b	$9F,$F6,$1C,$E9,$E1,$39,$E0,$FA,$72,$AE,$55,$C3,$9F,$D2,$47,$C1
	.dc.b	$F3,$F5,$D1,$E3,$8F,$7C,$7F,$8F,$7E,$79,$F0,$DF,$E3,$FF,$9E,$7C
	.dc.b	$FD,$35,$13,$0F,$CA,$57,$06,$C7,$1E,$CE,$5D,$3C,$FA,$A5,$7E,$6F
	.dc.b	$14,$EA,$31,$3A,$7F,$DC,$B3,$A7,$84,$E7,$6F,$CA,$CF,$E8,$E7,$27
	.dc.b	$E9,$FE,$79,$F3,$FF,$E5,$96,$C6,$63,$1E,$67,$F9,$77,$E7,$FE,$64
	.dc.b	$F1,$8C,$BF,$B1,$8C,$BF,$EF,$8F,$9F,$FD,$0E,$3F,$A6,$9C,$B7,$8B
	.dc.b	$FC,$D1,$3A,$ED,$FF,$46,$7C,$F3,$AF,$C1,$F3,$F3,$2F,$DF,$4E,$5B
	.dc.b	$3F,$E7,$18,$C4,$31,$FD,$AC,$61,$FD,$BF,$CF,$BE,$17,$FC,$F0,$DA
	.dc.b	$E6,$43,$96,$3F,$E9,$F8,$D4,$60,$FE,$4F,$EC,$FF,$20,$70,$21,$F3
	.dc.b	$FD,$8F,$F4,$34,$EF,$0F,$E1,$2D,$BF,$F5,$63,$6F,$AF,$8F,$E6,$EB
	.dc.b	$E2,$C8,$FE,$41,$3B,$7C,$13,$FE,$93,$F9,$FF,$3E,$1F,$C5,$BE,$61
	.dc.b	$ED,$FF,$A5,$DB,$7F,$CA,$6B,$79,$7F,$AF,$5B,$1F,$D1,$1A,$9B,$2F
	.dc.b	$85,$F0,$6D,$2B,$A3,$DE,$3F,$D4,$DD,$2F,$9F,$FF,$54,$7C,$D3,$E2
	.dc.b	$69,$A8,$98,$7F,$3D,$AE,$0D,$8E,$3F,$1F,$A8,$FA,$33,$A6,$CE,$9F
	.dc.b	$E8,$DF,$3D,$FB,$37,$EB,$E0,$F9,$FF,$19,$FE,$BB,$9D,$70,$7F,$3E
	.dc.b	$6E,$FF,$EE,$16,$DF,$F6,$86,$AB,$FE,$64,$FE,$12,$32,$4F,$0B,$E0
	.dc.b	$F9,$FF,$EC,$9F,$9D,$FB,$56,$F6,$7E,$29,$9E,$79,$8F,$F5,$FA,$FC
	.dc.b	$01,$C3,$E7,$F3,$1F,$92,$8C,$CF,$CE,$9E,$DF,$01,$F4,$3E,$7E,$5F
	.dc.b	$38,$CB,$48,$7C,$6E,$EF,$F6,$F7,$CF,$BA,$29,$27,$FB,$97,$CF,$FF
	.dc.b	$3A,$E2,$FF,$05,$FC,$EB,$E7,$BF,$C1,$B4,$C3,$93,$FD,$E5,$AF,$EA
	.dc.b	$9B,$5F,$DB,$AF,$A4,$AF,$A3,$FE,$18,$8C,$C6,$23,$B6,$1E,$08,$C5
	.dc.b	$F6,$75,$5C,$30,$F2,$B6,$FF,$C4,$75,$B9,$F4,$B6,$3C,$1E,$8F,$F0
	.dc.b	$F7,$C9,$3C,$A3,$17,$E0,$C3,$C5,$3C,$9B,$5C,$B1,$F8,$F3,$C5,$B1
	.dc.b	$E2,$9D,$BE,$C3,$23,$1F,$43,$70,$F0,$DF,$78,$DE,$3E,$97,$58,$72
	.dc.b	$BE,$33,$93,$4F,$43,$38,$78,$A3,$38,$76,$BF,$51,$F7,$8A,$5A,$FB
	.dc.b	$2D,$8F,$90,$E7,$C6,$7C,$77,$18,$FB,$91,$87,$15,$FD,$F6,$75,$1F
	.dc.b	$06,$74,$E3,$E7,$F2,$FF,$97,$1A,$9A,$47,$EC,$9F,$E6,$2F,$8D,$7C
	.dc.b	$E7,$F9,$6B,$6F,$C6,$4F,$33,$62,$77,$D5,$F5,$23,$53,$F9,$39,$CC
	.dc.b	$70,$33,$BF,$E9,$23,$33,$DA,$7C,$C7,$68,$CC,$FC,$5C,$3F,$03,$18
	.dc.b	$9F,$A1,$18,$C3,$D1,$5B,$9E,$51,$87,$66,$30,$F4,$3E,$0C,$3D,$13
	.dc.b	$B8,$FD,$CD,$F5,$AD,$B6,$3D,$87,$1B,$F6,$57,$18,$F7,$9F,$35,$E9
	.dc.b	$87,$C4,$DF,$FC,$D6,$DC,$3D,$18,$F8,$9F,$F3,$F4,$66,$FE,$B7,$07
	.dc.b	$FA,$3A,$F8,$E1,$6F,$E6,$4D,$AF,$B9,$D3,$6D,$F3,$EE,$FD,$57,$DB
	.dc.b	$AD,$4E,$70,$F0,$C6,$2B,$79,$7F,$91,$27,$FD,$2B,$87,$49,$C9,$23
	.dc.b	$FD,$33,$F1,$CF,$63,$FB,$9C,$7B,$27,$59,$7F,$83,$69,$F3,$CF,$9F
	.dc.b	$CE,$3F,$EA,$C9,$CC,$63,$0F,$F5,$CC,$61,$E8,$78,$2B,$93,$FE,$A7
	.dc.b	$F2,$5D,$3F,$01,$B1,$FE,$AE,$1F,$3C,$BD,$54,$9D,$4F,$D3,$C9,$7C
	.dc.b	$9F,$F1,$34,$F2,$E1,$E1,$18,$AE,$15,$BC,$FF,$D8,$F7,$B3,$FD,$79
	.dc.b	$F3,$EF,$80,$FA,$AD,$8F,$0F,$1E,$26,$47,$91,$B8,$F7,$6D,$F6,$C7
	.dc.b	$94,$FF,$8E,$30,$B3,$9A,$FC,$1D,$FF,$A2,$CF,$F6,$71,$6B,$D5,$DD
	.dc.b	$AE,$89,$3F,$27,$E3,$3F,$C9,$CF,$08,$FC,$92,$8C,$D7,$F9,$22,$3A
	.dc.b	$61,$F0,$A7,$2F,$FB,$3A,$32,$7F,$BD,$8E,$5E,$5F,$E6,$3E,$53,$F2
	.dc.b	$D7,$FE,$52,$6D,$7B,$EF,$5C,$93,$0F,$B3,$E2,$5C,$3E,$A3,$6F,$E3
	.dc.b	$36,$7C,$1B,$7F,$85,$5F,$F7,$87,$CE,$BD,$0F,$D4,$B6,$3F,$DE,$9F
	.dc.b	$3E,$F9,$FC,$D6,$BE,$0E,$57,$28,$B9,$FF,$9A,$BE,$7F,$F1,$73,$FE
	.dc.b	$FD,$6C,$7F,$9B,$B1,$5F,$EE,$27,$3A,$7F,$BD,$3A,$7F,$DD,$F8,$30
	.dc.b	$FF,$3A,$C6,$1E,$5C,$3F,$C0,$23,$15,$FE,$FC,$F9,$FF,$F2,$E5,$F1
	.dc.b	$8C,$7A,$7B,$F4,$FF,$83,$D6,$B1,$FF,$40,$FC,$FB,$E3,$7F,$85,$B7
	.dc.b	$1E,$47,$F8,$8B,$C1,$7B,$DA,$AF,$E2,$DD,$1F,$D8,$BE,$7E,$F9,$46
	.dc.b	$FC,$4E,$19,$8F,$4C,$D7,$C6,$30,$FF,$6C,$AD,$CF,$FA,$66,$34,$DB
	.dc.b	$ED,$5C,$15,$70,$DD,$3F,$E3,$AF,$9F,$7C,$A7,$F2,$86,$C7,$8F,$C8
	.dc.b	$FF,$43,$7C,$FE,$6A,$4F,$F2,$6F,$9F,$F8,$2F,$A7,$36,$1C,$9F,$F5
	.dc.b	$76,$2A,$DC,$79,$FE,$37,$E2,$9F,$1B,$E7,$EE,$1E,$98,$B3,$ED,$9B
	.dc.b	$7E,$A8,$F8,$FA,$FD,$57,$E1,$7C,$07,$D0,$E1,$FA,$2D,$C7,$E2,$36
	.dc.b	$77,$1F,$1E,$76,$D3,$67,$B7,$E7,$FF,$9F,$CF,$FB,$19,$B8,$78,$F7
	.dc.b	$FF,$56,$77,$7F,$52,$9C,$DF,$F9,$A1,$B7,$FC,$CF,$E7,$DF,$29,$BC
	.dc.b	$FF,$AF,$76,$27,$38,$7D,$BF,$9F,$FC,$5F,$CF,$FE,$31,$B1,$93,$F0
	.dc.b	$11,$96,$9F,$83,$E9,$FC,$9F,$A1,$F9,$46,$C6,$4F,$C4,$46,$5A,$67
	.dc.b	$3F,$C7,$46,$5B,$1F,$81,$FA,$1F,$DE,$C7,$F9,$25,$AF,$FA,$21,$AF
	.dc.b	$EE,$DA,$BE,$7D,$F4,$1F,$D4,$A3,$FC,$95,$18,$F4,$BF,$C8,$EF,$E7
	.dc.b	$35,$B5,$BA,$FF,$2C,$79,$7F,$9C,$7F,$28,$BE,$7F,$FD,$33,$CB,$F8
	.dc.b	$99,$FF,$5B,$46,$33,$FF,$4E,$D7,$19,$56,$E3,$0D,$AE,$13,$FC,$56
	.dc.b	$7B,$3E,$0E,$CC,$BF,$EC,$3F,$84,$B2,$F9,$9A,$62,$24,$FA,$23,$2F
	.dc.b	$AA,$32,$F9,$31,$F3,$CE,$A3,$07,$8E,$7C,$3F,$3C,$F9,$FF,$0F,$C2
	.dc.b	$4A,$FD,$C4,$E5,$B3,$FE,$70,$8C,$47,$F9,$ED,$AF,$FA,$15,$AF,$FB
	.dc.b	$13,$E7,$FF,$C4,$8F,$F0,$37,$F9,$EB,$7C,$EF,$87,$1A,$8F,$38,$71
	.dc.b	$B7,$D5,$78,$C6,$9B,$19,$F2,$39,$FC,$1C,$AF,$03,$DB,$D6,$C4,$71
	.dc.b	$8B,$5D,$66,$ED,$E9,$9B,$5F,$D3,$F0,$56,$23,$CD,$36,$73,$1E,$02
	.dc.b	$1D,$CF,$85,$3B,$9E,$0D,$8D,$DF,$8D,$B7,$E3,$69,$E3,$9E,$0F,$53
	.dc.b	$E2,$38,$9F,$91,$5F,$06,$78,$9E,$82,$DF,$A1,$E5,$3C,$1C,$D7,$3B
	.dc.b	$F1,$47,$63,$8D,$FA,$9A,$70,$29,$D2,$3F,$09,$3C,$E3,$E6,$4F,$95
	.dc.b	$D1,$83,$91,$83,$D0,$5B,$F4,$5F,$9E,$31,$3C,$E3,$04,$BF,$4B,$F3
	.dc.b	$67,$67,$67,$4F,$84,$ED,$7E,$77,$C3,$1C,$4D,$B8,$9E,$A7,$B9,$F8
	.dc.b	$31,$AB,$F3,$CE,$A7,$B4,$E6,$BB,$52,$FD,$57,$BB,$EC,$70,$7B,$3E
	.dc.b	$A3,$CE,$78,$1E,$13,$E0,$8E,$04,$39,$9F,$AF,$AF,$0D,$78,$23,$53
	.dc.b	$E3,$7E,$35,$74,$9F,$C0,$47,$86,$FD,$4F,$84,$E3,$7D,$1F,$D0,$99
	.dc.b	$3C,$37,$E5,$6D,$74,$AF,$90,$DB,$F0,$4F,$88,$F2,$C6,$5A,$64,$E6
	.dc.b	$5A,$F5,$3E,$79,$F9,$35,$DA,$35,$5F,$4A,$FD,$11,$F9,$F3,$81,$EC
	.dc.b	$3F,$C3,$DF,$3F,$FE,$45,$7F,$8F,$79,$1F,$C7,$0F,$E7,$F1,$98,$FE
	.dc.b	$3A,$DB,$FE,$39,$3B,$9E,$7E,$1F,$F4,$6F,$9F,$7C,$D5,$24,$FF,$8B
	.dc.b	$61,$F1,$38,$8F,$E5,$31,$C1,$D4,$F3,$9E,$9F,$1F,$D7,$E6,$A5,$73
	.dc.b	$8C,$BD,$63,$2F,$48,$CB,$DA,$33,$97,$F6,$9C,$15,$FD,$7B,$CE,$FF
	.dc.b	$0D,$DF,$F9,$0F,$E7,$DF,$43,$FD,$89,$B1,$FE,$47,$9D,$BF,$D0,$9A
	.dc.b	$FF,$9F,$CF,$F4,$03,$87,$0D,$73,$E4,$6D,$92,$E9,$F0,$12,$3C,$9D
	.dc.b	$55,$FE,$A3,$3B,$7D,$13,$E7,$9F,$47,$C1,$7A,$7E,$06,$57,$B6,$32
	.dc.b	$FB,$E3,$2F,$48,$CB,$DA,$33,$AF,$F8,$47,$05,$7C,$9E,$9F,$F0,$3F
	.dc.b	$A0,$FA,$0A,$67,$1F,$91,$35,$F9,$0F,$0E,$BE,$3D,$3C,$BA,$65,$7F
	.dc.b	$9D,$70,$FD,$1E,$7F,$E2,$16,$F5,$7C,$F7,$F6,$84,$BF,$1F,$8B,$2F
	.dc.b	$93,$E6,$FA,$0C,$47,$F7,$AA,$FF,$19,$61,$D6,$B5,$3C,$08,$6D,$CB
	.dc.b	$FE,$02,$F9,$F7,$C4,$FF,$8A,$3E,$7D,$F4,$14,$FF,$15,$CE,$AB,$FD
	.dc.b	$20,$D7,$95,$68,$E5,$DB,$CF,$AB,$A7,$64,$AF,$F4,$C6,$3F,$9A,$6C
	.dc.b	$E5,$FF,$0A,$C2,$79,$47,$FA,$7F,$F1,$6F,$F2,$1F,$07,$E5,$DB,$1F
	.dc.b	$ED,$76,$A7,$FC,$8A,$FF,$88,$1C,$93,$C7,$FF,$23,$BA,$DB,$FC,$AC
	.dc.b	$E6,$7C,$13,$AF,$23,$FC,$9B,$F4,$3F,$ED,$29,$FF,$59,$2F,$A1,$C9
	.dc.b	$AF,$43,$FC,$A7,$F4,$3F,$CC,$67,$85,$7F,$8C,$9B,$97,$F9,$96,$7F
	.dc.b	$9B,$7D,$0F,$A9,$FE,$70,$63,$2F,$F2,$EF,$D0,$7E,$6B,$0F,$F6,$14
	.dc.b	$BE,$87,$B9,$AE,$E9,$3F,$EC,$7A,$E0,$FF,$90,$9B,$8F,$F9,$D7,$E8
	.dc.b	$5F,$0F,$FE,$6A,$6D,$71,$BE,$6F,$F9,$2A,$32,$FF,$93,$3D,$CF,$E4
	.dc.b	$EF,$FB,$56,$1F,$43,$CC,$FF,$3A,$CE,$A7,$87,$97,$82,$3C,$7B,$2B
	.dc.b	$A3,$FD,$C1,$3C,$72,$F4,$4F,$FA,$03,$E8,$7C,$71,$FE,$8F,$F9,$C7
	.dc.b	$BD,$FE,$84,$6E,$1F,$EE,$9A,$DF,$47,$F2,$16,$C0,$38,$6E,$DC,$FE
	.dc.b	$2B,$C0,$C8,$E8,$E5,$FF,$7B,$76,$3E,$87,$FD,$2B,$F3,$EF,$8C,$A7
	.dc.b	$FA,$62,$75,$5C,$2B,$04,$7F,$E0,$05,$3E,$4B,$11,$FE,$7C,$AE,$2D
	.dc.b	$AE,$F3,$FE,$A4,$FA,$0F,$DC,$6F,$FE,$B3,$C1,$F4,$3E,$43,$5E,$24
	.dc.b	$C7,$E1,$57,$FA,$E6,$F8,$EF,$FD,$1D,$B3,$F1,$F0,$FE,$78,$FF,$A3
	.dc.b	$A3,$29,$A6,$23,$75,$FE,$C5,$FA,$1F,$B9,$7F,$D4,$E5,$EF,$7C,$83
	.dc.b	$FD,$7B,$3A,$C3,$EB,$35,$34,$C4,$BE,$5F,$8F,$3C,$AF,$6E,$1F,$D8
	.dc.b	$6F,$75,$DB,$E1,$3E,$BF,$C0,$4B,$D9,$E0,$8D,$38,$3C,$2E,$1F,$F5
	.dc.b	$7E,$3F,$AE,$C3,$FB,$77,$D0,$B9,$FF,$DA,$ED,$AE,$4D,$7F,$D6,$FF
	.dc.b	$43,$FE,$57,$DB,$F2,$9F,$3E,$FA,$0E,$78,$72,$25,$F8,$EF,$82,$AE
	.dc.b	$0F,$E7,$63,$31,$88,$61,$E1,$8C,$36,$FF,$C0,$D5,$C1,$BE,$97,$FB
	.dc.b	$E2,$FB,$47,$91,$F3,$B5,$C9,$E1,$8E,$17,$E1,$AE,$93,$FB,$1B,$F0
	.dc.b	$37,$0F,$F0,$B5,$6E,$F8,$E1,$E2,$C7,$BD,$EC,$A7,$B2,$F8,$CE,$E3
	.dc.b	$C6,$31,$F0,$BB,$38,$E3,$E2,$75,$87,$81,$F5,$61,$AC,$3A,$C7,$84
	.dc.b	$EB,$7D,$0E,$B7,$CC,$E8,$E6,$FE,$54,$FF,$8B,$63,$EC,$B6,$32,$FA
	.dc.b	$4D,$7A,$6F,$8E,$FF,$1A,$47,$DD,$8C,$38,$AF,$F7,$6C,$EA,$3E,$04
	.dc.b	$E9,$C7,$D0,$E5,$FE,$44,$D4,$D2,$3F,$2B,$E3,$DF,$3C,$DF,$E8,$CF
	.dc.b	$23,$51,$97,$E8,$99,$DF,$F2,$55,$CA,$B8,$CF,$80,$E2,$E2,$FE,$EC
	.dc.b	$EE,$79,$9B,$AE,$77,$DD,$FC,$F4,$77,$DF,$FC,$C1,$5C,$67,$33,$C6
	.dc.b	$73,$5C,$5F,$9B,$7E,$08,$C4,$FF,$98,$E3,$18,$79,$63,$CE,$79,$23
	.dc.b	$2D,$33,$7E,$08,$CB,$E8,$9F,$E1,$98,$7B,$77,$F9,$17,$FE,$77,$BE
	.dc.b	$71,$89,$D5,$7B,$3C,$6F,$9D,$39,$C6,$CE,$4E,$BD,$38,$C5,$4F,$A1
	.dc.b	$FE,$9D,$5A,$3F,$90,$C6,$7B,$38,$53,$2F,$F2,$5D,$6A,$F8,$CF,$48
	.dc.b	$F9,$F8,$77,$C3,$94,$7B,$B0,$E7,$87,$09,$FC,$37,$C5,$BD,$AF,$EF
	.dc.b	$77,$B2,$63,$D4,$4B,$E8,$1F,$43,$D3,$2F,$74,$E6,$31,$7F,$CF,$23
	.dc.b	$0F,$4C,$7C,$37,$CA,$3F,$9F,$79,$0E,$AF,$82,$D8,$FE,$8A,$3E,$7F
	.dc.b	$C6,$F0,$9D,$9F,$3D,$D7,$06,$1E,$33,$F6,$B4,$FF,$09,$5F,$57,$DD
	.dc.b	$3C,$0E,$87,$0A,$FA,$7F,$42,$F8,$2A,$77,$C3,$FD,$43,$5F,$D5,$E7
	.dc.b	$4D,$9D,$1C,$1F,$1B,$6F,$C3,$5C,$1F,$A0,$2B,$19,$FE,$F6,$FE,$15
	.dc.b	$FE,$19,$73,$D3,$C9,$DF,$63,$FB,$51,$A3,$E5,$FD,$0F,$F9,$51,$CC
	.dc.b	$FC,$98,$C6,$3E,$29,$E0,$FC,$68,$CE,$1F,$A8,$9C,$BE,$58,$CC,$7F
	.dc.b	$B8,$7E,$6B,$F2,$0D,$E2,$F9,$13,$F9,$E9,$CD,$FF,$91,$5B,$5F,$7F
	.dc.b	$C4,$D5,$FC,$F6,$C7,$DC,$6D,$7C,$A7,$F5,$3F,$3E,$F3,$E9,$FA,$CF
	.dc.b	$A1,$7D,$0E,$30,$FF,$07,$53,$E1,$78,$99,$C7,$E9,$C6,$67,$FC,$29
	.dc.b	$F4,$2E,$5F,$DB,$B6,$39,$DE,$27,$F7,$7D,$1F,$EF,$B9,$FF,$0E,$37
	.dc.b	$0E,$BE,$86,$BE,$2B,$F8,$0F,$D0,$BD,$3A,$72,$F4,$EE,$3F,$E2,$5F
	.dc.b	$A1,$7C,$8F,$F0,$86,$C7,$F8,$A7,$C1,$7C,$06,$6B,$F8,$A7,$3F,$F5
	.dc.b	$DF,$A1,$BE,$59,$BF,$21,$86,$63,$FE,$36,$AE,$0D,$A5,$6B,$83,$F8
	.dc.b	$45,$69,$B3,$D6,$B8,$56,$FB,$3F,$98,$FD,$0B,$E0,$7F,$99,$E1,$FE
	.dc.b	$87,$11,$9B,$FE,$DC,$D8,$FE,$EA,$D7,$81,$C2,$7F,$92,$7D,$0F,$C2
	.dc.b	$AF,$F2,$65,$71,$BF,$E7,$73,$9B,$F3,$FC,$6F,$C4,$3E,$2F,$D0,$DC
	.dc.b	$3E,$04,$59,$F6,$CD,$BF,$EE,$5A,$7F,$82,$76,$F9,$BE,$FF,$CE,$6E
	.dc.b	$3F,$25,$B8,$7C,$66,$DF,$2D,$BE,$DC,$FC,$AB,$E5,$7F,$45,$FE,$8F
	.dc.b	$F4,$3F,$CE,$A7,$FC,$2E,$D7,$E1,$B5,$F9,$8D,$7F,$D6,$8F,$1B,$FE
	.dc.b	$8B,$F4,$2F,$95,$EB,$7F,$CF,$A3,$18,$7E,$07,$E8,$7F,$13,$F4,$3F
	.dc.b	$8A,$6C,$64,$FC,$34,$65,$A7,$E2,$79,$7F,$23,$E7,$7E,$49,$B1,$93
	.dc.b	$F1,$D1,$96,$99,$D3,$F1,$B1,$96,$C7,$FA,$27,$E8,$1F,$41,$FB,$79
	.dc.b	$FF,$1D,$C6,$3C,$1F,$F1,$C3,$F9,$AF,$06,$D6,$EB,$F1,$7F,$44,$F2
	.dc.b	$FF,$C7,$4D,$8F,$F2,$03,$5F,$F2,$3B,$5F,$C9,$F9,$5F,$9D,$7F,$D3
	.dc.b	$8B,$E4,$3F,$3D,$C7,$FE,$12,$8C,$4E,$BB,$33,$AF,$E8,$E7,$94,$FF
	.dc.b	$46,$76,$DB,$DB,$FE,$EF,$8D,$DF,$5F,$13,$33,$FE,$C2,$F9,$AB,$46
	.dc.b	$CC,$9F,$3C,$68,$C3,$E1,$8C,$3A,$3D,$AF,$6F,$A0,$7D,$0D,$24,$FC
	.dc.b	$2F,$6E,$BF,$D8,$27,$6F,$FB,$05,$AF,$BE,$3D,$3F,$43,$FE,$19,$8F
	.dc.b	$F6,$33,$FC,$E5,$BE,$5E,$BC,$69,$18,$21,$FD,$BA,$79,$37,$E0,$7A
	.dc.b	$91,$EB,$E9,$D7,$47,$82,$3C,$71,$9D,$BF,$C9,$74,$9C,$70,$7F,$1B
	.dc.b	$8F,$EE,$F7,$C4,$87,$43,$C9,$7C,$10,$8E,$8F,$0B,$EF,$3B,$8D,$5D
	.dc.b	$AE,$F1,$D6,$BC,$B3,$9A,$D3,$67,$A5,$F9,$E3,$11,$C6,$30,$E8,$F2
	.dc.b	$DF,$6B,$EB,$19,$3C,$67,$CA,$31,$1E,$A7,$A1,$CA,$FC,$6E,$9A,$E8
	.dc.b	$F8,$4E,$27,$BC,$7C,$83,$AD,$FA,$23,$24,$BF,$4C,$EC,$E9,$7E,$39
	.dc.b	$D4,$78,$CE,$67,$8C,$A7,$0B,$F1,$57,$49,$FA,$73,$A7,$9B,$E8,$31
	.dc.b	$1B,$3F,$09,$3C,$A3,$E5,$DF,$E3,$1D,$98,$3E,$51,$E8,$7D,$D3,$CE
	.dc.b	$34,$79,$4E,$27,$D2,$3E,$05,$79,$2F,$E0,$4F,$19,$F3,$1D,$4F,$80
	.dc.b	$53,$A1,$67,$95,$FA,$63,$B4,$7C,$27,$53,$D6,$BF,$0C,$DB,$F4,$CF
	.dc.b	$A2,$BC,$91,$F0,$63,$89,$99,$F8,$EF,$B2,$BD,$2F,$D7,$36,$78,$A3
	.dc.b	$6E,$23,$6F,$BA,$BB,$BE,$E6,$BE,$E6,$9E,$E7,$C1,$5E,$58,$CD,$7B
	.dc.b	$6F,$E1,$5F,$63,$DF,$7E,$92,$1E,$53,$D1,$5C,$9C,$C7,$07,$44,$7A
	.dc.b	$4F,$6B,$F1,$5F,$F8,$9F,$E8,$7F,$C8,$77,$FC,$23,$0F,$F6,$EC,$62
	.dc.b	$3F,$D1,$AD,$8F,$E2,$FD,$1F,$D1,$23,$15,$CF,$C2,$C1,$FC,$C2,$7F
	.dc.b	$A4,$FD,$03,$E6,$A9,$31,$F0,$5F,$F8,$E6,$5F,$15,$99,$FF,$4E,$C7
	.dc.b	$17,$67,$F1,$E7,$0E,$A3,$BF,$C7,$78,$3E,$66,$4F,$01,$49,$DC,$66
	.dc.b	$13,$C2,$3C,$33,$A7,$C7,$19,$D1,$87,$A6,$BF,$B3,$F9,$BF,$E2,$3B
	.dc.b	$FF,$26,$9F,$CD,$1C,$38,$6B,$FE,$52,$6B,$FE,$5B,$77,$5D,$A3,$97
	.dc.b	$8D,$D8,$FE,$EB,$F3,$FF,$DE,$30,$FF,$31,$2E,$8C,$92,$BE,$14,$7F
	.dc.b	$79,$6C,$65,$FF,$65,$C7,$17,$7D,$3D,$B9,$FB,$FB,$92,$79,$C6,$5E
	.dc.b	$B1,$97,$9C,$65,$EB,$19,$CB,$D9,$0B,$57,$CA,$E8,$F9,$33,$F2,$5A
	.dc.b	$FF,$A2,$9B,$A7,$F7,$B8,$F9,$91,$C5,$FC,$04,$66,$FB,$F9,$BF,$DE
	.dc.b	$9B,$CD,$FE,$1E,$8F,$D5,$11,$C1,$9E,$79,$7F,$8B,$9F,$E0,$A7,$F9
	.dc.b	$06,$BA,$F2,$F3,$3A,$7C,$F5,$AE,$11,$5F,$F7,$CD,$27,$15,$CA,$76
	.dc.b	$77,$8F,$F1,$9F,$CE,$BE,$17,$E1,$B6,$7E,$3B,$5F,$C8,$B5,$FF,$4F
	.dc.b	$B9,$AF,$A2,$DE,$2F,$F5,$53,$67,$FC,$17,$F3,$BF,$AE,$AF,$AE,$FE
	.dc.b	$A6,$3F,$58,$FE,$A1,$BC,$DF,$E4,$88,$D4,$27,$8F,$36,$6A,$4F,$28
	.dc.b	$FF,$27,$FC,$EF,$E6,$E3,$FC,$36,$D4,$D3,$FC,$B7,$5F,$8A,$71,$E1
	.dc.b	$F2,$39,$F8,$76,$BF,$81,$4E,$9B,$1F,$E5,$EF,$9C,$FF,$70,$C7,$F0
	.dc.b	$75,$F3,$B9,$35,$E6,$7F,$99,$3E,$73,$F9,$84,$F0,$AF,$76,$FF,$ED
	.dc.b	$38,$C6,$7F,$9E,$AF,$E6,$DF,$3B,$EB,$7F,$9C,$18,$D3,$FC,$DF,$F3
	.dc.b	$9F,$A2,$C7,$F8,$C4,$BE,$77,$C0,$6B,$AE,$4F,$F1,$DA,$E0,$F8,$8F
	.dc.b	$F2,$1B,$8C,$7F,$5F,$F3,$AF,$81,$FF,$3E,$B6,$B9,$5F,$47,$90,$8D
	.dc.b	$38,$F8,$6F,$E6,$EF,$F2,$D8,$7C,$EF,$43,$FD,$13,$3A,$9E,$1E,$36
	.dc.b	$08,$F1,$EC,$AE,$8F,$E7,$13,$C7,$2F,$4C,$FF,$A5,$3E,$77,$C7,$1F
	.dc.b	$C2,$BE,$5D,$ED,$7F,$A6,$1B,$87,$F4,$1A,$DF,$27,$F9,$9A,$11,$87
	.dc.b	$FC,$D2,$D7,$3B,$73,$F8,$4F,$03,$23,$A3,$97,$FA,$77,$53,$E7,$7F
	.dc.b	$D5,$3F,$32,$F8,$6A,$7F,$AB,$27,$55,$C2,$B0,$47,$FA,$C5,$3E,$33
	.dc.b	$11,$F5,$2B,$8B,$6B,$BC,$FF,$AE,$3E,$73,$F7,$52,$B7,$CE,$F9,$0D
	.dc.b	$78,$92,$7F,$B3,$57,$03,$8D,$7F,$43,$8F,$C1,$C6,$67,$84,$7E,$1A
	.dc.b	$75,$C3,$FC,$F3,$E7,$7E,$C7,$07,$CF,$97,$B3,$F1,$0F,$F6,$6C,$EA
	.dc.b	$FF,$D3,$2D,$4D,$31,$2F,$90,$E3,$CF,$2B,$FB,$F5,$6A,$7A,$FC,$07
	.dc.b	$D2,$F9,$D9,$5A,$8C,$B4,$87,$AF,$B3,$FD,$C3,$F3,$AE,$8A,$49,$FF
	.dc.b	$06,$7C,$EF,$FA,$E7,$7F,$F5,$8F,$CE,$3E,$77,$F1,$D8,$71,$BF,$F0
	.dc.b	$93,$5F,$D6,$B6,$BF,$B1,$DF,$39,$5F,$37,$FD,$7B,$19,$8C,$42,$BA
	.dc.b	$46,$1B,$7F,$9F,$AE,$0D,$F4,$7F,$9A,$5F,$68,$F3,$E3,$FA,$5C,$39
	.dc.b	$36,$38,$63,$EE,$6D,$71,$3F,$DA,$38,$F8,$1B,$5C,$5F,$4E,$1E,$0C
	.dc.b	$3D,$38,$F9,$27,$67,$52,$9D,$43,$B6,$CF,$3C,$3D,$31,$C4,$F8,$97
	.dc.b	$E1,$77,$8E,$6F,$96,$3E,$79,$D9,$9C,$3A,$3E,$8A,$E9,$3B,$38,$57
	.dc.b	$8C,$FB,$13,$FC,$34,$EE,$7F,$90,$9C,$BF,$20,$D2,$F4,$5F,$15,$FE
	.dc.b	$20,$8C,$36,$33,$3C,$DA,$FF,$89,$BE,$77,$FC,$4A,$9A,$47,$FC,$52
	.dc.b	$FC,$5F,$9C,$7C,$D7,$F9,$37,$6F,$C5,$D0,$E4,$7D,$19,$DF,$3F,$CD
	.dc.b	$8D,$5F,$3B,$E5,$BF,$F9,$82,$31,$5B,$AE,$78,$7E,$63,$0E,$71,$F7
	.dc.b	$4E,$67,$7C,$39,$BF,$2E,$F9,$BB,$39,$E1,$FC,$EE,$B9,$DF,$AA,$72
	.dc.b	$7A,$E3,$33,$E5,$8E,$D8,$79,$6F,$AD,$F1,$6D,$6E,$3D,$18,$79,$2F
	.dc.b	$FA,$55,$69,$ED,$1F,$E7,$7D,$7F,$CB,$B8,$FB,$9F,$F2,$FC,$FB,$AF
	.dc.b	$9E,$3F,$E9,$5A,$D3,$EB,$E0,$E4,$FB,$70,$B7,$EB,$9A,$AB,$FC,$89
	.dc.b	$5F,$EA,$6C,$1F,$3B,$FD,$26,$3B,$74,$F2,$9C,$F6,$72,$BE,$38,$7F
	.dc.b	$82,$67,$37,$C0,$E9,$87,$39,$FF,$02,$D5,$AF,$7D,$F9,$BE,$25,$F0
	.dc.b	$9F,$DD,$31,$F7,$4F,$73,$5F,$38,$F9,$DC,$4F,$FA,$FA,$D3,$98,$E2
	.dc.b	$65,$C6,$5E,$2B,$E7,$7F,$13,$C8,$74,$D3,$E4,$E0,$F9,$D6,$59,$8F
	.dc.b	$F6,$35,$70,$7F,$D9,$E6,$F8,$75,$97,$F8,$B6,$30,$70,$C3,$AE,$1F
	.dc.b	$6A,$7F,$55,$7C,$DE,$11,$95,$F3,$AF,$80,$FB,$6D,$C3,$C3,$5F,$EA
	.dc.b	$47,$FD,$B5,$19,$AF,$F1,$AB,$77,$FF,$63,$4F,$B6,$F9,$9E,$42,$BD
	.dc.b	$1B,$5E,$47,$F1,$71,$FC,$4C,$FF,$74,$16,$3D,$3D,$DA,$E8,$93,$F1
	.dc.b	$BE,$33,$FC,$98,$F0,$8F,$9F,$18,$AF,$F2,$5C,$F3,$C3,$CF,$39,$7F
	.dc.b	$D8,$71,$93,$F8,$90,$8D,$3F,$E3,$9F,$93,$FC,$83,$78,$7E,$8C,$FE
	.dc.b	$92,$73,$7F,$C6,$DB,$5F,$81,$F1,$35,$7F,$61,$B1,$F4,$1B,$5F,$61
	.dc.b	$FD,$7F,$CD,$BC,$FA,$7E,$CB,$E7,$5F,$3B,$8C,$3F,$95,$69,$FE,$DE
	.dc.b	$B7,$C2,$72,$35,$97,$F9,$BA,$33,$5C,$0F,$F2,$B7,$CE,$B9,$3F,$7E
	.dc.b	$D8,$D5,$7F,$37,$AE,$2F,$F9,$E2,$74,$FF,$9E,$B9,$DF,$11,$FC,$21
	.dc.b	$B7,$CA,$57,$F3,$9F,$E8,$53,$A7,$FD,$07,$F3,$AF,$4F,$F8,$8B,$6F
	.dc.b	$95,$7F,$97,$DE,$91,$9A,$FF,$45,$FC,$EB,$E3,$BF,$8C,$B6,$BC,$5D
	.dc.b	$EF,$80,$CD,$7F,$1E,$F9,$C7,$BF,$4F,$E4,$7E,$1F,$F9,$EB,$E8,$71
	.dc.b	$8F,$AE,$B6,$DA,$DF,$B4,$AD,$E9,$FE,$9E,$BD,$BF,$C8,$7E,$77,$C5
	.dc.b	$87,$13,$F9,$63,$88,$FF,$57,$76,$7F,$78,$F8,$BF,$E6,$AD,$AF,$AA
	.dc.b	$FF,$7C,$F9,$DC,$F9,$EA,$A4,$AD,$C2,$71,$7D,$A3,$E3,$37,$DE,$E7
	.dc.b	$BF,$46,$DF,$26,$CE,$E3,$EF,$7C,$EF,$F3,$38,$FF,$5E,$B7,$B1,$3F
	.dc.b	$D9,$F7,$FC,$F3,$C2,$FF,$A7,$D6,$E3,$FD,$54,$DB,$FE,$A5,$5C,$1E
	.dc.b	$B5,$C0,$FE,$89,$F3,$AF,$92,$C6,$1F,$CF,$E3,$0F,$C2,$F9,$DF,$8F
	.dc.b	$F3,$BF,$21,$B1,$93,$E3,$C6,$5A,$7C,$AE,$6F,$9F,$E6,$7D,$06,$C6
	.dc.b	$4F,$9F,$19,$69,$FA,$39,$7D,$07,$F7,$8C,$FF,$C4,$91,$89,$FF,$1F
	.dc.b	$46,$27,$4A,$74,$EB,$E8,$BF,$66,$FF,$91,$5A,$FF,$B4,$DA,$FF,$8C
	.dc.b	$5A,$BE,$7D,$F3,$FF,$C0,$E3,$C3,$18,$F1,$3C,$71,$BF,$96,$FE,$BA
	.dc.b	$EA,$B8,$3F,$E5,$5F,$2B,$F5,$6F,$F0,$15,$F2,$3F,$EE,$5E,$5F,$C1
	.dc.b	$CF,$F9,$0E,$72,$EF,$BB,$5B,$7F,$BA,$E7,$9C,$AB,$8B,$FE,$C0,$BD
	.dc.b	$BC,$DA,$FF,$B1,$2E,$D7,$FB,$2A,$FA,$7C,$C6,$0C,$F8,$F2,$FD,$11
	.dc.b	$97,$D5,$1D,$3E,$75,$F3,$DD,$3E,$0E,$57,$F1,$F9,$CB,$67,$2F,$A7
	.dc.b	$0F,$65,$73,$99,$3F,$1E,$37,$A7,$BA,$F9,$CF,$13,$FC,$D5,$F3,$DF
	.dc.b	$E2,$27,$F9,$73,$FD,$09,$BE,$6F,$C2,$8D,$47,$10,$E2,$BE,$5B,$C6
	.dc.b	$34,$D8,$CF,$8F,$CF,$E0,$E5,$78,$1E,$DE,$A6,$23,$8C,$5A,$EB,$3F
	.dc.b	$53,$E7,$BF,$9B,$57,$A3,$C9,$F3,$19,$F0,$E4,$64,$87,$A6,$78,$B6
	.dc.b	$38,$BA,$BF,$C2,$CE,$9B,$3A,$9F,$6D,$76,$79,$1C,$77,$FF,$4B,$3F
	.dc.b	$1A,$35,$5C,$A3,$47,$D2,$8C,$CF,$0B,$F0,$3C,$B3,$FE,$89,$5C,$04
	.dc.b	$E9,$FF,$52,$35,$F0,$1B,$70,$66,$39,$C7,$A8,$FE,$DD,$1A,$C3,$DD
	.dc.b	$87,$DC,$C3,$CB,$5D,$A7,$38,$79,$67,$33,$E4,$9C,$BE,$D3,$AB,$AC
	.dc.b	$3C,$37,$D6,$72,$43,$C8,$47,$E4,$3F,$38,$C5,$F8,$5E,$11,$A3,$A3
	.dc.b	$A3,$D0,$6A,$38,$39,$3D,$06,$4F,$41,$9C,$3E,$23,$F4,$6B,$D3,$3F
	.dc.b	$5E,$FD,$98,$7C,$0A,$E6,$79,$F0,$F1,$92,$B8,$47,$73,$EB,$9D,$EB
	.dc.b	$9B,$F6,$5A,$E4,$E1,$7E,$27,$CF,$5F,$06,$BA,$C7,$B6,$74,$F9,$27
	.dc.b	$4F,$09,$F8,$71,$88,$D4,$60,$8F,$CA,$3D,$8E,$1F,$4C,$71,$C3,$F1
	.dc.b	$D8,$7C,$37,$E4,$CF,$8E,$3F,$0F,$1A,$8F,$71,$EC,$8E,$18,$7C,$0A
	.dc.b	$F9,$13,$E1,$7C,$D8,$7B,$0F,$6D,$F6,$8D,$CF,$93,$0F,$87,$1C,$30
	.dc.b	$EB,$3D,$CF,$09,$EC,$3E,$21,$0F,$86,$78,$67,$E0,$5F,$83,$0F,$34
	.dc.b	$7C,$3C,$3C,$4E,$AF,$DF,$3F,$85,$AE,$17,$E1,$3E,$E1,$E1,$3D,$07
	.dc.b	$86,$F9,$BE,$38,$E2,$ED,$F8,$55,$DA,$3F,$2C,$D7,$E8,$1F,$8F,$9F
	.dc.b	$A0,$F8,$30,$FE,$C7,$F3,$AF,$9E,$FE,$C2,$D9,$FF,$39,$35,$FE,$84
	.dc.b	$BE,$27,$73,$27,$6F,$28,$E7,$F1,$DD,$3E,$6A,$57,$48,$CB,$DA,$32
	.dc.b	$F6,$8F,$EE,$7D,$7F,$E3,$58,$FF,$42,$7C,$F3,$E8,$BF,$BB,$B6,$3F
	.dc.b	$D1,$13,$B7,$FA,$C3,$73,$FF,$00,$CE,$E7,$FA,$E3,$63,$FC,$0B,$3B
	.dc.b	$E0,$FE,$76,$F3,$AF,$E7,$A1,$C3,$AF,$1F,$FC,$3E,$E5,$B7,$FE,$09
	.dc.b	$80,$70,$DE,$9F,$5F,$57,$B1,$91,$F0,$1C,$BF,$0B,$E1,$BE,$0F,$BF
	.dc.b	$23,$04,$3D,$AE,$5F,$7C,$7C,$CE,$9F,$F0,$FF,$9F,$FE,$4C,$9F,$F5
	.dc.b	$83,$5F,$93,$1B,$73,$E3,$70,$30,$E8,$FA,$31,$8D,$BF,$C5,$F5,$FE
	.dc.b	$07,$DB,$FB,$2A,$EA,$F0,$F5,$4C,$3C,$31,$E4,$30,$FF,$84,$65,$39
	.dc.b	$8F,$3F,$87,$EA,$23,$EA,$F9,$FA,$F1,$8B,$8F,$92,$7F,$C7,$FF,$3E
	.dc.b	$F8,$8F,$F2,$0F,$CF,$BE,$8A,$9A,$89,$8F,$FB,$56,$74,$D9,$CB,$FE
	.dc.b	$27,$AE,$8F,$FB,$6D,$AE,$6B,$BF,$5F,$83,$A7,$C3,$D5,$2B,$B4,$65
	.dc.b	$F0,$46,$5F,$F1,$A6,$29,$FF,$2C,$FD,$07,$F9,$83,$4F,$F2,$FB,$78
	.dc.b	$7F,$C2,$AD,$BF,$C9,$BF,$E6,$A7,$FC,$7C,$E5,$34,$9E,$47,$F9,$9B
	.dc.b	$F1,$77,$D1,$F8,$3E,$5E,$5F,$FB,$E6,$72,$D9,$CB,$DA,$79,$44,$8D
	.dc.b	$10,$F1,$F5,$7F,$9D,$3E,$87,$FD,$E5,$1F,$AA,$CF,$FC,$F7,$0F,$A2
	.dc.b	$D1,$BF,$4F,$FC,$F7,$F4,$3F,$88,$BE,$1B,$7F,$8D,$AF,$9C,$7F,$97
	.dc.b	$5B,$87,$FA,$1A,$B8,$4F,$F9,$87,$E8,$BE,$24,$62,$75,$D3,$FE,$88
	.dc.b	$9F,$91,$F4,$2E,$FF,$0F,$D1,$4C,$7C,$27,$F0,$29,$D3,$80,$E1,$FF
	.dc.b	$38,$37,$0F,$F4,$CB,$73,$FF,$27,$FD,$13,$E8,$7F,$D3,$2D,$BE,$95
	.dc.b	$B7,$C1,$8E,$AB,$FD,$3F,$F1,$BF,$CE,$E7,$C5,$6F,$A2,$E6,$6B,$CB
	.dc.b	$93,$E7,$AE,$0F,$FA,$15,$B9,$FF,$AB,$AF,$FD,$7A,$FF,$AC,$23,$39
	.dc.b	$FF,$99,$9B,$5F,$C9,$1A,$AB,$8C,$FF,$9B,$3D,$FF,$64,$FF,$A4,$FE
	.dc.b	$75,$F0,$1F,$EB,$86,$C7,$CB,$9D,$E3,$FE,$C0,$6D,$F3,$D3,$FD,$7F
	.dc.b	$18,$AF,$E6,$B6,$F8,$BC,$D4,$9E,$8F,$78,$FF,$63,$F6,$BE,$87,$FD
	.dc.b	$93,$F4,$0F,$A2,$A6,$A2,$61,$F7,$EB,$83,$63,$8F,$C9,$EA,$3F,$D6
	.dc.b	$13,$A6,$CE,$9F,$C4,$7D,$17,$F0,$CD,$BF,$DB,$B8,$BB,$FC,$67,$FB
	.dc.b	$72,$75,$BF,$FA,$59,$B7,$FD,$51,$A9,$B5,$FD,$59,$C9,$3C,$6F,$7F
	.dc.b	$7F,$FB,$A3,$E8,$BE,$CD,$3D,$FF,$86,$67,$9E,$63,$F9,$FA,$FB,$E7
	.dc.b	$0F,$97,$CC,$7F,$B3,$E3,$33,$F3,$27,$B7,$C3,$7C,$EF,$80,$97,$CE
	.dc.b	$32,$D2,$1F,$ED,$9C,$7F,$44,$BE,$89,$CF,$FF,$00,$1B,$1F,$B3,$6B
	.dc.b	$FE,$E1,$FA,$2F,$EB,$1C,$5F,$DE,$BE,$81,$F4,$5F,$EC,$0B,$DC,$B8
	.dc.b	$F1,$8F,$53,$FD,$8B,$19,$C7,$FC,$07,$8F,$45,$8F,$47,$3E,$AA,$BC
	.dc.b	$19,$74,$9A,$7F,$83,$32,$F3,$DF,$18,$61,$D6,$30,$DA,$E5,$3E,$06
	.dc.b	$CE,$B0,$F0,$4E,$4F,$1E,$1E,$07,$E5,$63,$FE,$E0,$AF,$06,$3E,$2C
	.dc.b	$3F,$DC,$78,$F7,$BE,$27,$33,$89,$CE,$78,$3A,$AF,$16,$19,$C3,$B6
	.dc.b	$1E,$63,$AE,$3E,$6A,$DD,$F5,$C3,$9B,$D8,$F0,$63,$E3,$7E,$2D,$71
	.dc.b	$BE,$91,$F2,$F0,$EE,$F8,$E3,$BE,$3F,$E4,$28,$F5,$9F,$E4,$4B,$FB
	.dc.b	$4D,$BF,$92,$E7,$C5,$7C,$77,$18,$FA,$B1,$87,$A5,$7F,$87,$A7,$4D
	.dc.b	$8F,$F2,$6F,$D1,$7F,$93,$93,$48,$E8,$F9,$DF,$28,$F9,$FF,$C8,$6D
	.dc.b	$F8,$59,$FF,$29,$D7,$DB,$AD,$6F,$F7,$67,$27,$2B,$E3,$AF,$C1,$BE
	.dc.b	$11,$89,$FF,$32,$C6,$2B,$94,$79,$8E,$4E,$23,$51,$F8,$33,$BE,$1F
	.dc.b	$99,$C3,$C7,$19,$39,$46,$6F,$C1,$1F,$E6,$C3,$38,$5A,$E3,$5E,$F9
	.dc.b	$C9,$FE,$72,$8C,$CF,$AE,$3C,$A7,$7B,$F2,$D7,$16,$CF,$5B,$FD,$D6
	.dc.b	$BF,$BB,$B6,$1E,$88,$C4,$7B,$2B,$86,$3F,$DF,$EB,$4F,$9F,$C6,$E6
	.dc.b	$EA,$BF,$80,$39,$38,$74,$EA,$3F,$82,$7D,$13,$BF,$11,$E2,$9D,$64
	.dc.b	$AD,$5F,$23,$FC,$27,$19,$BE,$38,$FF,$98,$E7,$35,$C3,$0F,$19,$F2
	.dc.b	$6B,$DA,$D9,$FC,$E5,$6E,$3F,$D4,$ED,$7E,$A4,$FA,$7E,$65,$F4,$4F
	.dc.b	$A2,$F0,$CE,$63,$18,$7C,$78,$C4,$EA,$3F,$C5,$75,$CA,$3E,$0F,$B8
	.dc.b	$E9,$F2,$E2,$8C,$5F,$EC,$5B,$1F,$BB,$6B,$FE,$BE,$73,$A7,$F3,$79
	.dc.b	$7D,$17,$D8,$AF,$F5,$DD,$6A,$7F,$D7,$A6,$B8,$7C,$78,$7D,$78,$C1
	.dc.b	$FE,$3F,$9E,$15,$FD,$1E,$FA,$2F,$A2,$7C,$17,$A9,$B7,$FC,$25,$B3
	.dc.b	$FC,$4D,$AF,$FB,$59,$CE,$5E,$9B,$B5,$FA,$79,$B5,$E1,$BF,$C4,$C6
	.dc.b	$27,$DF,$18,$3D,$55,$C2,$BF,$AF,$74,$7C,$EA,$FF,$5A,$9F,$C7,$4B
	.dc.b	$1F,$C9,$DA,$FF,$BA,$9C,$E1,$F3,$7E,$8B,$FC,$B5,$1F,$E6,$33,$27
	.dc.b	$CC,$73,$3F,$92,$3C,$31,$FE,$6C,$11,$A9,$FD,$A4,$72,$9F,$EE,$93
	.dc.b	$98,$FE,$11,$F3,$5F,$95,$9F,$F7,$AB,$53,$49,$C1,$F0,$BC,$3F,$D7
	.dc.b	$C7,$E9,$70,$F9,$D1,$89,$D5,$FF,$9F,$6B,$4F,$E1,$B2,$FF,$13,$3F
	.dc.b	$E8,$1F,$9F,$7A,$1F,$98,$6C,$7F,$47,$6E,$5F,$D4,$5A,$FF,$21,$0B
	.dc.b	$E1,$39,$12,$79,$3F,$A4,$FA,$0F,$D2,$47,$F5,$36,$D7,$F5,$D6,$BD
	.dc.b	$5A,$FF,$2F,$E5,$79,$1F,$AE,$6D,$7F,$5D,$6C,$FF,$67,$6B,$E7,$FA
	.dc.b	$0F,$F3,$74,$FF,$63,$6C,$7F,$6D,$6B,$BF,$3F,$F7,$7F,$40,$F9,$2A
	.dc.b	$4B,$FE,$D9,$19,$7F,$B7,$7D,$03,$E5,$FF,$92,$B6,$3F,$D6,$7D,$CF
	.dc.b	$55,$F2,$7A,$B9,$5C,$9F,$E1,$77,$C7,$3F,$F0,$13,$5F,$A4,$E7,$9F
	.dc.b	$FA,$2F,$D0,$3E,$03,$9D,$7F,$2A,$94,$E7,$0F,$F6,$2F,$3F,$F8,$1B
	.dc.b	$E3,$3F,$9F,$36,$33,$E5,$BE,$83,$2A,$A9,$27,$73,$71,$FE,$73,$7F
	.dc.b	$E1,$2E,$1E,$3E,$DF,$26,$DF,$66,$D7,$0C,$7E,$63,$6B,$FA,$7F,$D0
	.dc.b	$7F,$AC,$AF,$FD,$B4,$D9,$FE,$8F,$C5,$F3,$7A,$F7,$72,$FF,$C4,$57
	.dc.b	$BB,$FC,$D3,$5F,$EF,$8D,$41,$ED,$B7,$F9,$AF,$9D,$F4,$0F,$93,$ED
	.dc.b	$7F,$D6,$27,$31,$F2,$3E,$83,$E4,$FD,$05,$CB,$E6,$B6,$FE,$03,$71
	.dc.b	$FF,$3C,$70,$7F,$8E,$A3,$E8,$F2,$7D,$2F,$22,$E1,$F5,$9B,$87,$C9
	.dc.b	$6E,$1C,$B8,$7F,$A0,$CB,$EE,$DF,$45,$FD,$61,$B8,$7F,$91,$DB,$3F
	.dc.b	$E5,$16,$BF,$C0,$7E,$85,$F4,$3F,$D7,$E3,$FC,$AD,$18,$F3,$3F,$CA
	.dc.b	$0B,$E6,$9C,$58,$BE,$17,$FC,$2F,$E9,$3F,$95,$BF,$E5,$76,$BF,$E6
	.dc.b	$16,$BB,$F2,$FF,$CB,$8F,$FB,$07,$BF,$FD,$3F,$A7,$F8,$D7,$CB,$7C
	.dc.b	$8E,$2B,$ED,$C6,$1F,$BB,$C3,$FE,$2B,$8F,$DC,$46,$2B,$95,$F1,$8F
	.dc.b	$F6,$75,$DA,$E5,$1F,$D1,$5E,$2D,$F1,$F5,$F3,$52,$3A,$1E,$28,$E4
	.dc.b	$65,$E5,$18,$76,$3E,$85,$F4,$5E,$1F,$81,$93,$A9,$CB,$67,$E2,$C6
	.dc.b	$23,$FD,$14,$D7,$F9,$D3,$5F,$F2,$E3,$BD,$3F,$C9,$CD,$D3,$35,$F0
	.dc.b	$7D,$9F,$89,$F4,$2F,$A2,$BE,$84,$CB,$D3,$39,$6C,$E5,$FE,$D9,$1F
	.dc.b	$D2,$5C,$9C,$3E,$03,$A9,$1E,$7E,$AD,$78,$C5,$D3,$FC,$C9,$4A,$FB
	.dc.b	$1F,$45,$FC,$E6,$FC,$7E,$77,$A0,$CF,$95,$23,$24,$3F,$C0,$13,$C5
	.dc.b	$B1,$C5,$D6,$1F,$E7,$39,$D3,$67,$53,$EE,$AE,$CF,$23,$8E,$DF,$E8
	.dc.b	$BC,$7F,$A1,$C7,$E0,$30,$E7,$3F,$E8,$F6,$C7,$F1,$49,$DB,$D4,$E1
	.dc.b	$7D,$B0,$F8,$78,$78,$9F,$24,$E8,$F0,$9C,$E3,$86,$1D,$EB,$83,$E5
	.dc.b	$8C,$CF,$F2,$4C,$3C,$6B,$0F,$1E,$1E,$08,$EC,$7B,$CF,$31,$1F,$1B
	.dc.b	$97,$83,$9B,$F1,$3C,$A3,$24,$8C,$C7,$17,$CC,$FB,$05,$79,$8F,$3D
	.dc.b	$79,$A7,$73,$C4,$F6,$61,$F0,$1F,$88,$78,$63,$BD,$F8,$63,$27,$13
	.dc.b	$BC,$F1,$7E,$26,$1F,$21,$C9,$E5,$BF,$13,$E0,$3B,$1D,$2B,$B3,$C0
	.dc.b	$ED,$1C,$1E,$A7,$03,$51,$C2,$35,$87,$C1,$BF,$24,$61,$F2,$3C,$CC
	.dc.b	$57,$D8,$3F,$C7,$D3,$FA,$16,$CF,$F3,$08,$C4,$7D,$23,$27,$F9,$14
	.dc.b	$F6,$15,$DB,$8A,$FC,$09,$30,$FD,$17,$67,$C7,$AF,$45,$7F,$0A,$8C
	.dc.b	$5F,$A1,$B1,$F0,$70,$FE,$97,$87,$38,$ED,$87,$D8,$3A,$BF,$1F,$0E
	.dc.b	$AF,$F1,$2B,$E6,$FC,$E6,$D7,$38,$FC,$F4,$62,$BA,$47,$EC,$E3,$0F
	.dc.b	$80,$C1,$DF,$0F,$43,$4F,$44,$70,$7E,$CC,$F9,$9F,$C6,$5F,$9A,$FF
	.dc.b	$8E,$D6,$E3,$F3,$6D,$BF,$E6,$6D,$AF,$90,$77,$3D,$85,$AF,$66,$1F
	.dc.b	$23,$0F,$5C,$F8,$8F,$51,$F0,$0E,$0E,$A3,$84,$6B,$0F,$34,$FA,$6B
	.dc.b	$E4,$46,$27,$E1,$9E,$CC,$3E,$1B,$FE,$AD,$FA,$27,$CE,$7F,$6F,$6C
	.dc.b	$7F,$62,$6B,$BF,$88,$D4,$C8,$D3,$C6,$39,$7C,$6F,$3F,$9C,$95,$CE
	.dc.b	$32,$F5,$8C,$BE,$38,$FF,$00,$78,$1F,$E3,$D8,$FE,$E7,$F4,$2F,$A2
	.dc.b	$FF,$02,$36,$3F,$BB,$CE,$DF,$F6,$43,$5F,$F5,$43,$5F,$F2,$69,$FE
	.dc.b	$25,$F3,$3F,$C3,$71,$98,$FF,$28,$BC,$F6,$6F,$C2,$30,$DA,$F1,$F7
	.dc.b	$F9,$3B,$66,$1E,$87,$CF,$DB,$88,$FF,$6E,$65,$FE,$AC,$74,$7B,$A3
	.dc.b	$DB,$D5,$EE,$F7,$A4,$FC,$28,$CB,$F1,$23,$2F,$48,$F9,$3D,$7F,$E2
	.dc.b	$9F,$51,$F4,$54,$CD,$7C,$89,$F8,$24,$7F,$85,$1F,$E4,$97,$0F,$CC
	.dc.b	$78,$BF,$EF,$59,$FF,$07,$37,$B3,$C1,$D9,$23,$FC,$57,$29,$F2,$19
	.dc.b	$8E,$53,$87,$7D,$7E,$92,$3C,$BE,$8A,$BE,$18,$BC,$7F,$5A,$3F,$D1
	.dc.b	$38,$2B,$FC,$95,$F4,$2F,$87,$FF,$26,$7D,$0B,$E8,$A9,$A8,$98,$7C
	.dc.b	$2A,$E0,$D8,$FF,$30,$C6,$23,$F9,$83,$71,$F1,$42,$31,$CD,$F9,$06
	.dc.b	$C7,$FA,$7B,$0F,$D6,$D7,$05,$D3,$E6,$EA,$95,$FE,$54,$9C,$B6,$72
	.dc.b	$AB,$95,$7B,$A7,$FC,$D5,$F4,$5F,$E6,$69,$F2,$B7,$1F,$D1,$B6,$FF
	.dc.b	$D6,$2F,$F9,$BA,$77,$D1,$C8,$93,$FE,$66,$9D,$57,$EB,$3E,$89,$F4
	.dc.b	$5E,$0F,$97,$97,$FE,$6C,$9C,$B6,$7F,$D1,$71,$88,$FE,$B4,$D7,$F6
	.dc.b	$E1,$C3,$DE,$3F,$D0,$D1,$C7,$1F,$A2,$DB,$F6,$F5,$FF,$A2,$FE,$87
	.dc.b	$FD,$ED,$87,$E7,$9A,$FF,$69,$6B,$E0,$6B,$FC,$23,$E8,$7F,$CB,$CF
	.dc.b	$E9,$9A,$FF,$71,$6B,$E8,$6B,$FC,$43,$E8,$7F,$D2,$EF,$B1,$AE,$75
	.dc.b	$FF,$56,$C6,$23,$FD,$C7,$F4,$0E,$CF,$F5,$8B,$63,$F6,$ED,$7F,$C0
	.dc.b	$4D,$7E,$4B,$9D,$3F,$2A,$DC,$FF,$91,$B6,$7F,$DE,$3F,$3E,$F9,$FF
	.dc.b	$F5,$93,$63,$F8,$2B,$5F,$F0,$83,$5F,$AE,$D7,$F9,$7F,$C1,$3C,$6F
	.dc.b	$F6,$4B,$63,$F8,$7B,$5F,$F0,$D3,$5F,$C0,$B9,$C7,$F5,$AD,$C7,$AD
	.dc.b	$79,$BB,$3F,$D9,$13,$9B,$07,$0D,$9E,$0F,$FB,$03,$E1,$FD,$0F,$06
	.dc.b	$A6,$93,$2F,$F6,$B3,$F0,$F1,$FE,$02,$DC,$7F,$A7,$FC,$C3,$E2,$B7
	.dc.b	$73,$0F,$03,$FE,$DF,$9D,$47,$FB,$4B,$A1,$F3,$DF,$EF,$06,$C7,$F3
	.dc.b	$66,$BF,$E4,$16,$BF,$A8,$73,$97,$F1,$46,$E3,$FD,$89,$B3,$F1,$BE
	.dc.b	$6D,$F2,$3F,$EF,$19,$CC,$FF,$14,$C3,$07,$FB,$D6,$B7,$1F,$EE,$8F
	.dc.b	$9E,$FE,$48,$FE,$E1,$AF,$F7,$06,$BF,$51,$A9,$FE,$4A,$DF,$03,$FB
	.dc.b	$BE,$5F,$CB,$27,$35,$FD,$A6,$BF,$90,$65,$D6,$F6,$D8,$F2,$47,$9F
	.dc.b	$A7,$31,$F8,$4F,$9E,$FC,$73,$FE,$04,$6B,$9E,$1F,$EC,$6A,$32,$D8
	.dc.b	$E6,$D3,$F1,$F9,$7F,$42,$9F,$C7,$E2,$F9,$ED,$C7,$E4,$89,$7F,$2E
	.dc.b	$33,$F0,$9E,$09,$34,$98,$7F,$76,$8C,$BE,$0C,$FD,$53,$CA,$30,$7D
	.dc.b	$DE,$8F,$3F,$CF,$38,$A9,$27,$63,$E7,$BF,$BC,$6B,$FA,$AF,$9E,$7C
	.dc.b	$EF,$F8,$AB,$0E,$57,$C1,$AF,$D4,$3F,$62,$E2,$BF,$B2,$DF,$49,$5F
	.dc.b	$47,$F6,$F1,$98,$C4,$2B,$AC,$61,$B7,$FE,$3C,$AE,$0D,$F3,$3F,$B2
	.dc.b	$5F,$68,$F4,$63,$FE,$42,$C3,$93,$63,$86,$3E,$F6,$D7,$13,$D7,$3B
	.dc.b	$69,$B3,$D3,$7E,$0B,$F4,$E1,$E4,$9D,$9C,$CF,$65,$73,$7E,$85,$71
	.dc.b	$BF,$0D,$F9,$A7,$B1,$C0,$B7,$D5,$E3,$86,$B1,$F8,$B1,$E0,$AF,$1E
	.dc.b	$1D,$63,$CF,$5B,$C7,$C5,$7C,$27,$99,$E7,$C3,$EC,$3F,$DD,$1A,$FC
	.dc.b	$63,$17,$F4,$1D,$79,$8F,$88,$FF,$26,$D2,$31,$3F,$ED,$98,$C4,$EA
	.dc.b	$73,$3F,$E5,$4F,$9E,$FD,$9A,$69,$1F,$DB,$BF,$1B,$E4,$5F,$37,$FE
	.dc.b	$45,$D7,$F1,$75,$FD,$12,$B5,$39,$DB,$FD,$D5,$18,$E1,$F9,$F5,$B3
	.dc.b	$EC,$61,$D2,$33,$1E,$33,$37,$CD,$FF,$0C,$D7,$37,$F0,$07,$3C,$3F
	.dc.b	$3B,$7C,$E3,$27,$38,$CD,$73,$8F,$F4,$CD,$73,$9F,$5C,$6F,$0F,$F4
	.dc.b	$ED,$F3,$9C,$D6,$E7,$DB,$5E,$6C,$3D,$33,$98,$F3,$99,$26,$FF,$E5
	.dc.b	$CC,$3C,$CF,$F9,$7A,$3D,$57,$E0,$C7,$FC,$E5,$3B,$6E,$1E,$4E,$0B
	.dc.b	$7F,$DD,$30,$E7,$39,$3F,$88,$BF,$E7,$77,$F8,$CD,$BE,$7B,$FA,$AC
	.dc.b	$6B,$A3,$9C,$E7,$B3,$AE,$0C,$39,$3F,$C4,$E7,$57,$C4,$EB,$87,$49
	.dc.b	$FE,$30,$71,$BF,$3F,$C4,$BE,$79,$F3,$DD,$B2,$E1,$97,$FA,$46,$B5
	.dc.b	$39,$BE,$37,$C3,$2F,$6C,$62,$75,$1F,$CD,$AB,$8B,$A3,$15,$C7,$1F
	.dc.b	$2F,$89,$FD,$12,$7F,$9E,$8F,$9D,$F9,$73,$FE,$20,$9D,$3F,$CC,$E7
	.dc.b	$5B,$FC,$69,$CE,$7F,$EA,$99,$C9,$FD,$0A,$32,$FF,$8A,$8C,$D7,$8B
	.dc.b	$A7,$ED,$C6,$2B,$FD,$43,$5C,$2B,$FD,$D3,$87,$91,$7C,$F3,$D9,$E3
	.dc.b	$37,$0F,$EA,$33,$6B,$D1,$37,$0F,$F5,$6C,$DA,$F2,$4D,$AF,$F1,$FC
	.dc.b	$D9,$F8,$97,$F8,$48,$C5,$6E,$BC,$73,$E2,$3F,$13,$5F,$C5,$4F,$9B
	.dc.b	$F1,$7F,$E6,$37,$FB,$29,$99,$F9,$91,$97,$F1,$31,$9C,$3E,$64,$E5
	.dc.b	$FF,$62,$C6,$4F,$99,$19,$7F,$CA,$F2,$74,$D8,$D3,$F4,$F9,$FE,$8C
	.dc.b	$62,$1F,$27,$F9,$4A,$F8,$BF,$17,$FA,$48,$C4,$FD,$48,$C7,$17,$D2
	.dc.b	$6D,$7E,$33,$34,$62,$BE,$9B,$72,$FC,$44,$EA,$BF,$C1,$3F,$3D,$FE
	.dc.b	$6F,$6F,$B9,$FB,$48,$C6,$1F,$AE,$51,$93,$F7,$3F,$3C,$E0,$C5,$FF
	.dc.b	$85,$23,$0D,$8F,$DF,$FC,$EB,$EB,$F8,$C6,$1F,$E0,$7C,$CF,$22,$FD
	.dc.b	$06,$1B,$87,$F8,$83,$E8,$3F,$CE,$CF,$F8,$8B,$E3,$9F,$31,$FC,$32
	.dc.b	$73,$BF,$F0,$69,$CD,$7F,$8A,$34,$7C,$67,$F1,$2F,$C0,$55,$E6,$8C
	.dc.b	$3F,$C5,$BE,$85,$ED,$53,$F9,$2F,$83,$FE,$8F,$F9,$1C,$65,$FE,$34
	.dc.b	$AD,$BF,$C9,$3E,$87,$C3,$A7,$F1,$89,$CB,$FC,$7E,$33,$D1,$F8,$B8
	.dc.b	$FF,$51,$CE,$A3,$F9,$BC,$EA,$BF,$23,$39,$7F,$CB,$71,$9A,$FC,$6F
	.dc.b	$94,$F8,$8F,$E7,$7D,$BF,$DF,$5B,$0F,$A1,$B1,$30,$FF,$29,$D6,$DB
	.dc.b	$5B,$6C,$F2,$DF,$9F,$D0,$D9,$DB,$71,$FF,$27,$61,$FE,$0D,$8C,$69
	.dc.b	$EF,$8C,$57,$F9,$73,$E8,$55,$FD,$66,$FF,$D8,$B1,$89,$FD,$8C,$7E
	.dc.b	$CA,$30,$E0,$38,$76,$EA,$FF,$A1,$7D,$0B,$E4,$3F,$A9,$CE,$6F,$FD
	.dc.b	$A7,$D4,$AD,$47,$97,$1F,$E9,$F1,$8C,$7F,$A9,$DF,$15,$7C,$70,$F9
	.dc.b	$FF,$43,$F4,$3E,$86,$E5,$F3,$1B,$7F,$1B,$74,$63,$4F,$8D,$18,$AF
	.dc.b	$A7,$D5,$F6,$BC,$9F,$B6,$D8,$C9,$F6,$23,$2D,$3F,$D1,$3F,$3E,$F9
	.dc.b	$FF,$DD,$65,$F2,$63,$1E,$0F,$CB,$84,$3E,$57,$DB,$1B,$AE,$2F,$D4
	.dc.b	$F9,$FF,$E2,$6F,$16,$BF,$55,$AE,$FC,$8F,$F5,$CF,$DD,$F4,$8F,$B4
	.dc.b	$D7,$EF,$B5,$FE,$62,$D4,$7F,$22,$EE,$FF,$5F,$CB,$E8,$29,$88,$93
	.dc.b	$FE,$AA,$9D,$36,$72,$FF,$83,$A3,$FD,$B5,$C1,$FB,$58,$FE,$5F,$F2
	.dc.b	$FD,$8E,$FE,$2C,$8E,$91,$97,$B4,$7F,$B4,$BE,$7D,$F3,$F4,$93,$E3
	.dc.b	$F6,$2B,$FE,$1E,$9D,$BF,$EC,$06,$BE,$58,$DE,$1E,$59,$CB,$6B,$E1
	.dc.b	$CF,$FB,$8B,$E7,$FF,$8F,$DF,$FB,$29,$FE,$7A,$DF,$37,$5E,$3C,$8C
	.dc.b	$10,$FF,$18,$CF,$26,$FB,$DD,$48,$F5,$F4,$EB,$A3,$C1,$1E,$38,$FF
	.dc.b	$7A,$FC,$FF,$F3,$49,$FF,$6E,$B5,$FF,$1F,$B7,$3F,$A5,$7F,$90,$7E
	.dc.b	$A4,$67,$A3,$FB,$AB,$6B,$F0,$AD,$55,$C6,$77,$E2,$78,$89,$3F,$17
	.dc.b	$A6,$BE,$28,$B8,$78,$A6,$D7,$2A,$F4,$D7,$CF,$9C,$C7,$8A,$2B,$E2
	.dc.b	$8B,$87,$94,$F3,$61,$E5,$9E,$07,$42,$1F,$4A,$78,$D7,$49,$E1,$2A
	.dc.b	$E9,$39,$AD,$CF,$2C,$3D,$4D,$C3,$D4,$FB,$CF,$84,$64,$F1,$99,$3C
	.dc.b	$46,$4F,$05,$77,$71,$7E,$27,$99,$C6,$BE,$CC,$EE,$31,$5E,$29,$CC
	.dc.b	$70,$6B,$A3,$C6,$FA,$9C,$1E,$23,$C9,$87,$C4,$BF,$14,$6A,$31,$7F
	.dc.b	$12,$3D,$67,$29,$DD,$75,$3C,$24,$BF,$31,$FE,$44,$BE,$75,$F0,$2F
	.dc.b	$F0,$4D,$3C,$78,$77,$BE,$51,$FC,$6A,$B8,$BE,$1A,$E2,$D3,$89,$CC
	.dc.b	$F5,$57,$3B,$F1,$C7,$38,$C3,$F1,$A3,$CD,$18,$8F,$8C,$7C,$0B,$F1
	.dc.b	$B8,$6D,$F9,$1C,$E1,$F9,$6A,$E4,$DA,$F2,$4F,$09,$CE,$1E,$A8,$E2
	.dc.b	$FD,$F6,$CF,$B0,$F7,$1E,$C3,$9E,$1E,$C7,$F5,$06,$0F,$61,$EF,$30
	.dc.b	$7A,$5F,$7B,$8B,$F3,$3D,$A7,$CC,$E6,$BC,$D5,$DE,$BF,$02,$F2,$9E
	.dc.b	$0F,$8F,$0F,$88,$EE,$B8,$CF,$90,$F1,$16,$7F,$6B,$F3,$EF,$9F,$FF
	.dc.b	$45,$E3,$FD,$36,$31,$5F,$D7,$23,$13,$AC,$3F,$D2,$34,$F8,$6C,$5F
	.dc.b	$F6,$28,$E2,$FF,$B9,$9B,$7C,$A3,$A5,$72,$AE,$30,$A7,$C5,$61,$7C
	.dc.b	$C5,$31,$12,$79,$9E,$18,$D3,$83,$A4,$76,$ED,$FF,$18,$46,$5A,$FF
	.dc.b	$BF,$1B,$8F,$F6,$FF,$9E,$FF,$38,$CE,$FC,$5E,$44,$7B,$46,$9E,$51
	.dc.b	$96,$9A,$EC,$F7,$CF,$F7,$CD,$3F,$D7,$6D,$9F,$F0,$5B,$52,$EA,$E3
	.dc.b	$D5,$27,$C2,$F4,$8E,$BD,$1D,$BD,$B9,$5E,$68,$CB,$E8,$8C,$AC,$FE
	.dc.b	$04,$FC,$7E,$99,$3F,$1D,$FA,$FE,$67,$F7,$1A,$FF,$0A,$6D,$FD,$EA
	.dc.b	$BE,$55,$70,$7A,$69,$FE,$D8,$BE,$0F,$FB,$6A,$B8,$5F,$2B,$E0,$FF
	.dc.b	$87,$EB,$77,$FE,$2E,$5D,$3E,$2E,$99,$3F,$7E,$72,$D9,$CB,$E2,$39
	.dc.b	$4F,$AA,$3D,$7D,$32,$79,$3E,$EF,$9D,$C4,$71,$3F,$D3,$91,$8C,$3C
	.dc.b	$D3,$F9,$AF,$9C,$7C,$1F,$F9,$66,$D1,$87,$FC,$B8,$D7,$FD,$10,$D7
	.dc.b	$FD,$1A,$E6,$BF,$C8,$0D,$D7,$FE,$00,$B6,$7F,$CA,$3F,$39,$FA,$E9
	.dc.b	$CD,$F8,$C9,$3F,$A3,$9D,$3C,$39,$BF,$63,$1F,$AE,$8D,$73,$E2,$64
	.dc.b	$60,$87,$C3,$9F,$DB,$BF,$AF,$F9,$CF,$D1,$65,$FE,$57,$6E,$3F,$E6
	.dc.b	$69,$46,$1F,$F3,$4B,$FE,$62,$7E,$73,$97,$7C,$FF,$C1,$A3,$06,$EB
	.dc.b	$FC,$C9,$F3,$4F,$9B,$FA,$0D,$BF,$F4,$53,$5F,$F5,$B3,$5F,$F5,$DB
	.dc.b	$99,$E7,$3F,$C5,$23,$18,$7A,$6F,$F6,$91,$8B,$FF,$40,$B5,$36,$3F
	.dc.b	$75,$F3,$6F,$33,$F8,$FB,$6B,$F8,$B3,$5F,$C7,$7C,$DB,$93,$F9,$2B
	.dc.b	$63,$F8,$E3,$5D,$C7,$F2,$AF,$9B,$F9,$04,$D7,$07,$F2,$EF,$DC,$5F
	.dc.b	$57,$FC,$C3,$E7,$5D,$94,$98,$FF,$29,$F9,$D7,$CD,$7F,$36,$6C,$7F
	.dc.b	$2D,$AE,$2D,$AD,$54,$8E,$4D,$7F,$65,$F0,$6F,$BE,$7A,$BF,$D0,$BE
	.dc.b	$7B,$FA,$65,$7F,$39,$6B,$FB,$16,$F3,$FF,$A8,$F3,$FF,$27,$C6,$39
	.dc.b	$BF,$1C,$DA,$FF,$5A,$B5,$4F,$0F,$27,$C0,$48,$F8,$1D,$FF,$D5,$3E
	.dc.b	$7B,$DF,$1F,$C0,$5B,$8F,$FA,$CD,$B8,$FE,$77,$E6,$1F,$3D,$C7,$79
	.dc.b	$87,$F5,$3B,$DB,$FE,$C8,$AD,$B9,$9D,$BD,$DF,$EC,$B1,$A8,$FE,$25
	.dc.b	$DC,$F9,$EF,$ED,$2D,$8F,$EC,$0D,$7F,$8C,$7C,$EB,$E4,$37,$52,$76
	.dc.b	$FF,$72,$9D,$46,$08,$78,$23,$53,$FD,$E3,$E7,$BF,$89,$49,$A6,$2B
	.dc.b	$FA,$D3,$8F,$13,$81,$DF,$CD,$B5,$FD,$DE,$74,$FF,$79,$AF,$EC,$5F
	.dc.b	$3B,$F6,$49,$C5,$E7,$F9,$EF,$F0,$4C,$62,$BF,$C0,$0F,$DF,$A4,$EA
	.dc.b	$A5,$69,$E4,$F0,$E6,$E6,$47,$C5,$D3,$5F,$E8,$58,$7E,$61,$E0,$68
	.dc.b	$FF,$08,$7C,$F3,$83,$FC,$3A,$D8,$CC,$7F,$88,$3E,$7B,$FA,$DF,$32
	.dc.b	$5F,$3C,$F9,$DF,$E5,$18,$71,$9F,$EA,$4D,$A7,$46,$B1,$E9,$E7,$7F
	.dc.b	$7B,$BE,$78,$7F,$31,$CB,$BC,$60,$F0,$63,$D7,$2F,$04,$63,$0F,$E7
	.dc.b	$11,$88,$57,$28,$C3,$6F,$C5,$3E,$B9,$E1,$7D,$67,$C5,$87,$96,$78
	.dc.b	$B6,$38,$D7,$94,$57,$17,$E3,$5F,$76,$D7,$17,$D0,$78,$6B,$9D,$F6
	.dc.b	$CB,$13,$C1,$AF,$6C,$7D,$F5,$C0,$F3,$CE,$B0,$F0,$56,$CE,$38,$F9
	.dc.b	$5F,$8B,$8F,$8D,$B5,$AC,$3A,$C7,$98,$F7,$5F,$43,$EA,$BF,$D9,$CE
	.dc.b	$25,$29,$97,$6F,$DD,$F1,$1F,$09,$B8,$FA,$F1,$87,$94,$FF,$4D,$9D
	.dc.b	$36,$3F,$C3,$1F,$3D,$FE,$19,$4D,$23,$A3,$EB,$7C,$8B,$E6,$FF,$27
	.dc.b	$1F,$64,$FC,$E4,$FC,$E9,$D6,$FF,$42,$72,$72,$AE,$3B,$FE,$3E,$32
	.dc.b	$FE,$46,$B8,$C7,$53,$CA,$71,$AE,$8E,$27,$85,$74,$DF,$E5,$61,$E3
	.dc.b	$8C,$57,$CF,$8C,$5F,$26,$C7,$76,$DF,$96,$3B,$5F,$F8,$77,$0F,$2C
	.dc.b	$E6,$B7,$3E,$0A,$F3,$1C,$CB,$3E,$6B,$FF,$4C,$5F,$BB,$5F,$97,$5F
	.dc.b	$E2,$9C,$3C,$71,$89,$D5,$78,$78,$7E,$8C,$E7,$2B,$39,$3A,$F4,$63
	.dc.b	$1F,$F4,$07,$CF,$7F,$45,$9C,$9F,$EA,$CB,$46,$67,$FC,$C5,$8F,$F9
	.dc.b	$7A,$B5,$39,$AF,$25,$F5,$78,$F5,$77,$C1,$86,$F0,$FE,$41,$F1,$6F
	.dc.b	$9D,$7C,$F6,$EB,$56,$AD,$4E,$70,$F5,$46,$30,$F9,$31,$8C,$39,$C6
	.dc.b	$0F,$F6,$3C,$EE,$FD,$75,$BB,$F3,$78,$7F,$E1,$4B,$FF,$66,$8F,$9C
	.dc.b	$F9,$F7,$FC,$DE,$74,$FF,$A3,$67,$5B,$72,$EA,$F9,$D1,$83,$FD,$A9
	.dc.b	$5C,$35,$FD,$3C,$66,$FF,$D1,$D3,$98,$FE,$88,$66,$7F,$07,$19,$9F
	.dc.b	$E8,$38,$79,$17,$CF,$3D,$CF,$0C,$DC,$3F,$DC,$93,$6B,$D1,$37,$0F
	.dc.b	$C7,$CD,$AF,$0C,$DA,$FE,$AD,$36,$7E,$25,$FD,$D8,$C5,$6E,$BC,$73
	.dc.b	$E2,$3F,$05,$5F,$E5,$D3,$E6,$FC,$F7,$F9,$39,$FF,$7C,$19,$9F,$99
	.dc.b	$19,$7F,$D6,$51,$9C,$3E,$4C,$E5,$FD,$3C,$64,$F9,$31,$97,$FD,$24
	.dc.b	$66,$7F,$B4,$C6,$5F,$9F,$D3,$F6,$23,$10,$F9,$4F,$C9,$D7,$E0,$FE
	.dc.b	$2B,$F3,$91,$8C,$91,$8E,$0F,$A2,$DA,$FC,$5C,$E9,$FF,$00,$65,$FB
	.dc.b	$EB,$E5,$5B,$C3,$EA,$BF,$A9,$F9,$E7,$99,$7D,$FC,$37,$1E,$67,$FA
	.dc.b	$BE,$53,$FB,$1F,$9E,$70,$EF,$1F,$F6,$E4,$62,$F8,$C7,$F8,$4E,$30
	.dc.b	$FF,$B8,$7E,$71,$F3,$DC,$EF,$F7,$91,$8B,$E3,$5F,$B7,$F2,$BF,$80
	.dc.b	$74,$3E,$7A,$C6,$1F,$E0,$9E,$9F,$F9,$A2,$F8,$FC,$6F,$FA,$9D,$E7
	.dc.b	$19,$C3,$FD,$E5,$39,$7C,$9F,$19,$FC,$37,$E7,$93,$E4,$8C,$3F,$C7
	.dc.b	$F8,$FF,$B2,$FC,$35,$F2,$0D,$F9,$BC,$33,$3F,$87,$5C,$1B,$4A,$D6
	.dc.b	$4E,$7F,$EC,$37,$B8,$C5,$FF,$8F,$63,$13,$AB,$E5,$3D,$30,$F2,$65
	.dc.b	$E5,$F9,$BB,$3E,$06,$E3,$D0,$FC,$A5,$A3,$38,$75,$7F,$98,$7C,$D6
	.dc.b	$6F,$EA,$B6,$F9,$78,$8F,$5F,$F1,$B5,$AB,$E2,$7F,$52,$AF,$F7,$D5
	.dc.b	$F1,$7F,$C2,$CD,$73,$1B,$7C,$BF,$35,$F3,$FE,$6B,$E8,$36,$FE,$5B
	.dc.b	$67,$D9,$B7,$F8,$6A,$31,$9F,$C7,$6D,$7C,$BD,$11,$8B,$F8,$7F,$09
	.dc.b	$F7,$7C,$5F,$BC,$DB,$FB,$2D,$9E,$F7,$F8,$BC,$3E,$A5,$FF,$57,$FA
	.dc.b	$47,$D2,$FA,$ED,$18,$C7,$FC,$5C,$D8,$FF,$1B,$35,$FF,$1D,$7C,$E3
	.dc.b	$E7,$3F,$63,$1F,$E3,$B8,$C7,$6F,$F8,$DD,$7C,$87,$F5,$F9,$D5,$70
	.dc.b	$7F,$C8,$9E,$23,$E1,$BF,$D1,$11,$F8,$33,$F8,$8F,$CD,$63,$6F,$F0
	.dc.b	$4D,$23,$1A,$70,$8E,$55,$FB,$86,$BC,$DE,$0E,$67,$8B,$FE,$97,$8D
	.dc.b	$D7,$6B,$CC,$F8,$8E,$07,$57,$32,$EF,$BF,$25,$2B,$C5,$1A,$30,$F8
	.dc.b	$E3,$0F,$F0,$EF,$9A,$7C,$D7,$4F,$A5,$93,$98,$F1,$CE,$A3,$FC,$C8
	.dc.b	$D7,$FC,$D8,$D7,$FC,$E0,$D7,$E6,$FC,$D7,$F0,$C7,$F8,$EB,$FC,$B1
	.dc.b	$BE,$4F,$C0,$8D,$47,$D3,$6E,$9F,$E3,$A7,$84,$66,$78,$78,$52,$39
	.dc.b	$90,$E4,$F5,$F7,$64,$78,$9F,$07,$A1,$88,$F0,$45,$C3,$C1,$3F,$53
	.dc.b	$E6,$DE,$1F,$9F,$E8,$66,$1F,$CE,$27,$2D,$9C,$BE,$B9,$E5,$1F,$7B
	.dc.b	$E6,$DF,$39,$49,$5F,$D0,$A1,$F0,$98,$8F,$CB,$47,$07,$53,$CE,$75
	.dc.b	$97,$F9,$82,$B5,$5C,$2B,$4F,$02,$19,$3C,$15,$CB,$9F,$F4,$B5,$CA
	.dc.b	$FA,$5F,$2B,$E0,$48,$CB,$C8,$EA,$E1,$D5,$74,$BE,$D1,$FE,$F9,$BE
	.dc.b	$D3,$E0,$BE,$D7,$C4,$91,$9A,$EC,$74,$39,$4F,$45,$5F,$29,$F1,$CF
	.dc.b	$17,$81,$AB,$E1,$3E,$28,$E6,$76,$BF,$0B,$83,$8C,$75,$6D,$F2,$AE
	.dc.b	$6D,$8F,$09,$C6,$B9,$BE,$13,$27,$A2,$38,$3D,$0C,$D7,$79,$E2,$F9
	.dc.b	$E3,$81,$E7,$BF,$74,$EA,$BE,$5C,$60,$93,$E3,$7E,$A3,$DA,$3C,$33
	.dc.b	$C6,$3C,$37,$EC,$69,$B2,$9E,$73,$E2,$BB,$69,$B2,$9B,$8F,$5C,$6D
	.dc.b	$B1,$B6,$BB,$AF,$4B,$F2,$4C,$1E,$B3,$F1,$93,$CE,$37,$7F,$25,$B7
	.dc.b	$EE,$6D,$7A,$AF,$E0,$3E,$66,$9E,$22,$C7,$4B,$ED,$3E,$13,$C9,$3A
	.dc.b	$71,$7E,$16,$DF,$85,$A7,$86,$FC,$6F,$85,$B5,$E1,$69,$E1,$8F,$8C
	.dc.b	$D7,$E1,$9F,$84,$3F,$84,$DF,$99,$C4,$FF,$07,$8C,$5F,$E2,$CF,$34
	.dc.b	$FE,$AB,$E2,$7F,$8C,$7C,$C4,$8F,$E2,$D1,$FC,$6C,$C3,$FC,$5A,$30
	.dc.b	$FF,$97,$3A,$DF,$39,$FD,$7B,$E7,$1F,$4B,$4F,$EC,$33,$AA,$E1,$59
	.dc.b	$7F,$CC,$B5,$C1,$C9,$8E,$AE,$84,$77,$D5,$5F,$EF,$D3,$FD,$A5,$E0
	.dc.b	$68,$A6,$A7,$86,$AD,$BF,$C0,$F4,$BF,$57,$C1,$7A,$FD,$C9,$8F,$9A
	.dc.b	$32,$FA,$23,$2F,$8A,$3D,$FD,$1F,$E0,$0F,$3D,$F4,$B4,$CD,$7C,$99
	.dc.b	$FF,$0B,$B3,$6F,$EC,$6D,$8C,$BF,$E1,$C8,$CB,$FE,$1E,$A7,$4F,$7E
	.dc.b	$99,$3F,$E9,$09,$CB,$67,$2F,$F8,$96,$13,$E7,$99,$3C,$63,$7B,$7C
	.dc.b	$78,$1F,$E8,$69,$F7,$78,$92,$3C,$0F,$C1,$F9,$EC,$47,$78,$B5,$CE
	.dc.b	$3F,$D1,$F3,$FE,$26,$F9,$E7,$C2,$7F,$8A,$3E,$79,$F4,$B4,$D4,$4B
	.dc.b	$F8,$A2,$B8,$B8,$6B,$F2,$9A,$FD,$0A,$74,$F7,$E9,$93,$FE,$B8,$9C
	.dc.b	$B6,$72,$FF,$93,$61,$3E,$79,$93,$C6,$37,$97,$BE,$B9,$CF,$F9,$57
	.dc.b	$E7,$7F,$C8,$FC,$FF,$E5,$86,$C7,$CE,$6A,$7F,$CA,$2F,$1F,$1F,$FC
	.dc.b	$9B,$1A,$27,$91,$FE,$50,$F9,$E7,$CF,$EB,$E9,$66,$1F,$ED,$49,$CB
	.dc.b	$67,$E5,$A8,$CB,$F9,$26,$BF,$98,$6B,$E9,$E9,$F1,$F5,$48,$FF,$70
	.dc.b	$E2,$9F,$39,$98,$F8,$53,$87,$71,$9D,$7F,$0D,$4A,$D5,$7F,$9F,$7E
	.dc.b	$81,$E9,$FF,$9B,$9B,$1F,$EF,$06,$BF,$E8,$3F,$A0,$FC,$FB,$FE,$F4
	.dc.b	$6B,$9E,$9F,$F3,$BC,$63,$1F,$CB,$7D,$07,$D2,$27,$07,$F9,$F1,$FA
	.dc.b	$BF,$40,$FA,$0F,$F3,$FF,$D0,$7B,$F2,$FF,$80,$2D,$7F,$D2,$8D,$C3
	.dc.b	$F4,$5F,$3E,$F9,$EF,$F4,$3B,$6B,$8F,$0F,$43,$33,$23,$F6,$75,$CA
	.dc.b	$31,$1F,$EE,$3F,$84,$7D,$03,$DB,$FF,$49,$FC,$67,$F4,$22,$79,$DE
	.dc.b	$0F,$0F,$2C,$C3,$AF,$67,$FA,$B2,$BF,$A4,$EB,$FE,$B1,$D3,$F7,$51
	.dc.b	$CA,$B6,$7F,$06,$77,$87,$FA,$8B,$E0,$FD,$93,$FE,$B5,$6D,$7F,$03
	.dc.b	$6E,$1F,$EB,$AF,$99,$7D,$07,$1D,$E6,$1D,$8F,$F7,$B3,$FE,$AE,$9D
	.dc.b	$C6,$6F,$9B,$FC,$61,$AF,$03,$94,$70,$7F,$D9,$1E,$13,$E8,$3F,$D7
	.dc.b	$2D,$8F,$03,$5F,$F6,$77,$CF,$BE,$4B,$75,$27,$6F,$FB,$0A,$75,$18
	.dc.b	$21,$E0,$8D,$4F,$FB,$23,$E8,$3F,$8C,$EB,$FC,$9D,$BA,$FF,$B7,$9B
	.dc.b	$7F,$CD,$30,$FE,$C4,$FF,$B8,$63,$29,$A6,$23,$75,$D7,$C8,$F0,$12
	.dc.b	$BE,$74,$EA,$FF,$DD,$BF,$3F,$FE,$0A,$27,$1F,$B7,$E8,$3F,$DB,$F1
	.dc.b	$8B,$FA,$CF,$F8,$42,$D3,$AA,$95,$A7,$93,$C3,$97,$99,$1F,$1F,$45
	.dc.b	$7F,$DF,$D8,$7E,$75,$E0,$68,$FC,$37,$D0,$38,$7F,$DE,$6D,$8C,$C7
	.dc.b	$FB,$DB,$E8,$3F,$B5,$ED,$FD,$7B,$E7,$DF,$41,$FD,$82,$D8,$71,$55
	.dc.b	$FD,$6E,$30,$DC,$3A,$7A,$3A,$A6,$1D,$1F,$ED,$78,$F8,$1C,$E3,$DF
	.dc.b	$1E,$D5,$B9,$F0,$E1,$E3,$8C,$5F,$16,$E1,$DD,$E1,$7C,$B0,$E9,$7C
	.dc.b	$5F,$65,$F1,$6D,$F1,$77,$87,$C3,$6B,$E6,$3D,$31,$B6,$E3,$B7,$FC
	.dc.b	$0F,$87,$85,$B8,$78,$5D,$9C,$C6,$1E,$37,$67,$33,$E0,$95,$D3,$C2
	.dc.b	$FC,$4E,$8E,$95,$E1,$75,$7C,$67,$73,$93,$55,$D0,$FA,$95,$FE,$36
	.dc.b	$6B,$F1,$AB,$FC,$23,$18,$8F,$B7,$E2,$3E,$2B,$2D,$8F,$F1,$DC,$FF
	.dc.b	$8C,$27,$27,$F7,$08,$7D,$05,$23,$FD,$D1,$BA,$7E,$0F,$E4,$5F,$3B
	.dc.b	$FE,$35,$8F,$C3,$57,$F8,$DA,$BE,$45,$6B,$97,$E5,$4E,$4F,$A1,$19
	.dc.b	$7E,$91,$9A,$ED,$7D,$77,$FF,$29,$E1,$D2,$31,$5F,$E5,$A8,$C6,$1D
	.dc.b	$A3,$D6,$76,$9D,$BF,$AB,$9D,$BE,$03,$77,$E0,$AE,$53,$A8,$F6,$CE
	.dc.b	$AF,$F6,$98,$7A,$67,$35,$B9,$F3,$D7,$86,$F9,$B4,$9A,$7F,$94,$F0
	.dc.b	$F8,$4F,$F9,$56,$3E,$15,$76,$C7,$FC,$C5,$3B,$6D,$F8,$B8,$6D,$FF
	.dc.b	$99,$B0,$E7,$39,$3F,$AC,$3F,$E5,$E7,$FB,$3D,$BE,$83,$FA,$54,$6B
	.dc.b	$9F,$9C,$E7,$A7,$AE,$0C,$39,$3F,$C4,$AF,$FC,$07,$87,$28,$C1,$E3
	.dc.b	$C3,$9C,$67,$0F,$81,$87,$A3,$E5,$DF,$40,$FA,$0E,$F9,$EF,$3F,$F4
	.dc.b	$0D,$6A,$73,$7C,$6F,$85,$7B,$A3,$13,$A8,$FF,$02,$D7,$17,$46,$2B
	.dc.b	$8E,$7E,$81,$E1,$7F,$86,$67,$FC,$1A,$3E,$7B,$E6,$CF,$FA,$52,$74
	.dc.b	$FC,$09,$D6,$FF,$22,$73,$A7,$FA,$6E,$72,$7F,$82,$E3,$2F,$FA,$78
	.dc.b	$CD,$78,$E3,$2F,$FB,$26,$FC,$2F,$9B,$AB,$F0,$35,$C2,$FF,$D3,$DF
	.dc.b	$41,$FE,$59,$8F,$4D,$DB,$F7,$4D,$AF,$24,$DA,$FE,$A1,$36,$BC,$1E
	.dc.b	$8F,$E0,$A3,$13,$F0,$E3,$07,$8A,$B8,$52,$B7,$18,$AF,$F6,$09,$F3
	.dc.b	$3E,$83,$FC,$A4,$FF,$8B,$4C,$D7,$CB,$8C,$BF,$56,$33,$87,$CB,$9C
	.dc.b	$BF,$EB,$D8,$C9,$F2,$E3,$2F,$F6,$B3,$33,$FE,$F3,$8C,$BF,$47,$A7
	.dc.b	$E5,$C6,$21,$F2,$DF,$93,$7E,$97,$C8,$FE,$92,$31,$7F,$85,$8C,$71
	.dc.b	$7C,$36,$C7,$D2,$6D,$7E,$1F,$24,$63,$2F,$A6,$FE,$C3,$E8,$1E,$55
	.dc.b	$F8,$2C,$37,$1F,$F2,$A4,$A7,$F6,$FF,$41,$F9,$A9,$CC,$63,$8A,$47
	.dc.b	$F9,$6F,$E7,$DF,$41,$8D,$7F,$45,$18,$9F,$F3,$07,$3B,$E2,$7F,$81
	.dc.b	$AF,$A0,$B1,$87,$F8,$3F,$A7,$FE,$6A,$AF,$F3,$47,$C9,$7F,$AA,$1F
	.dc.b	$F3,$54,$67,$1F,$D8,$4E,$5F,$0F,$C8,$7F,$10,$FA,$04,$FF,$12,$CD
	.dc.b	$39,$8F,$E4,$3C,$DF,$D8,$7E,$26,$F8,$E6,$FD,$06,$19,$A7,$FA,$22
	.dc.b	$B8,$36,$95,$AC,$FF,$8D,$D6,$9B,$3D,$6B,$85,$6F,$9D,$0A,$FC,$4D
	.dc.b	$7A,$23,$B7,$CF,$F9,$B3,$F3,$B6,$32,$78,$E3,$2F,$FA,$47,$E8,$13
	.dc.b	$FD,$1E,$FB,$46,$27,$FC,$91,$1F,$E4,$98,$C3,$80,$E1,$DB,$AC,$3D
	.dc.b	$30,$E2,$7D,$1B,$E4,$BF,$A1,$DE,$F0,$FE,$6B,$39,$7E,$57,$D0,$7D
	.dc.b	$2F,$A1,$FA,$6D,$8C,$9F,$3E,$32,$D3,$38,$FC,$E8,$CB,$63,$EB,$7C
	.dc.b	$37,$DE,$F1,$7E,$FB,$63,$27,$DB,$8C,$B4,$FF,$5D,$7C,$FB,$E7,$FF
	.dc.b	$9F,$DF,$C5,$8C,$76,$7C,$78,$43,$E4,$FE,$24,$6E,$B8,$BF,$43,$E8
	.dc.b	$FF,$76,$F1,$6B,$F4,$5A,$EF,$C3,$EC,$FF,$02,$F8,$8F,$F4,$07,$1F
	.dc.b	$D2,$8F,$E0,$BF,$1F,$8D,$3F,$D0,$96,$8C,$4F,$E3,$E1,$78,$8E,$87
	.dc.b	$FB,$66,$38,$36,$75,$1C,$23,$53,$89,$91,$83,$97,$81,$C3,$E5,$A4
	.dc.b	$60,$87,$F6,$BB,$46,$A3,$13,$E5,$F9,$E7,$CF,$D2,$4F,$9F,$D5,$AF
	.dc.b	$F7,$59,$DB,$FE,$F3,$6B,$DE,$3E,$07,$CF,$FF,$08,$8F,$F7,$C3,$FC
	.dc.b	$95,$BE,$3E,$BC,$29,$18,$21,$FE,$03,$9E,$4D,$8E,$B3,$23,$AC,$78
	.dc.b	$3D,$C9,$1E,$07,$C7,$E7,$E2,$3A,$45,$E1,$FB,$D8,$2B,$EC,$7C,$FF
	.dc.b	$F1,$EB,$FE,$15,$5F,$83,$8C,$47,$F8,$5D,$B1,$CB,$9F,$8F,$83,$2B
	.dc.b	$8F,$27,$FB,$2E,$F6,$FF,$B3,$71,$FC,$CE,$1F,$8F,$F9,$D7,$CF,$7E
	.dc.b	$25,$B8,$72,$3F,$8E,$38,$7F,$C5,$0B,$E0,$B8,$4C,$9E,$02,$78,$57
	.dc.b	$7A,$D4,$F2,$AB,$39,$8E,$35,$C3,$0F,$F7,$23,$63,$C7,$5E,$7B,$E3
	.dc.b	$5A,$8B,$3E,$89,$AF,$8E,$2D,$7A,$63,$BC,$60,$E8,$E0,$87,$9C,$F2
	.dc.b	$CE,$DC,$DF,$A5,$D1,$1F,$9D,$3D,$23,$17,$E7,$78,$BE,$EA,$E9,$1A
	.dc.b	$3D,$46,$8F,$82,$53,$E7,$3C,$2B,$BC,$79,$8D,$5F,$C9,$7E,$11,$93
	.dc.b	$B1,$82,$BF,$15,$B7,$F2,$2B,$B5,$FA,$67,$F0,$71,$87,$CA,$60,$F0
	.dc.b	$CF,$16,$9E,$48,$E0,$F8,$6F,$E5,$1E,$A7,$C2,$79,$DC,$12,$BC,$4F
	.dc.b	$06,$92,$3C,$71,$96,$C6,$5F,$03,$F4,$5F,$29,$F0,$A7,$CB,$7F,$24
	.dc.b	$F8,$95,$E1,$9E,$64,$3E,$99,$C9,$D3,$83,$D6,$75,$35,$5E,$39,$EA
	.dc.b	$DB,$F9,$13,$C2,$B9,$D6,$A3,$C0,$F9,$67,$F1,$D3,$A7,$8B,$67,$E8
	.dc.b	$1E,$03,$8D,$FB,$A7,$8D,$F8,$67,$67,$85,$F3,$4E,$E3,$EA,$46,$DF
	.dc.b	$61,$B3,$A9,$DC,$F9,$E7,$E7,$0D,$18,$3E,$35,$79,$DE,$6D,$3E,$0C
	.dc.b	$F8,$63,$F1,$C7,$2B,$FA,$D1,$FC,$6F,$E1,$FF,$8E,$7C,$C4,$AF,$E1
	.dc.b	$B1,$97,$F8,$7C,$65,$FE,$3B,$1F,$DA,$FB,$3F,$C4,$51,$FC,$71,$AF
	.dc.b	$FB,$25,$B8,$7F,$79,$F9,$DF,$ED,$93,$9F,$17,$91,$1F,$E4,$B1,$A9
	.dc.b	$E1,$D1,$FD,$E6,$3F,$CE,$13,$A8,$C7,$4E,$59,$19,$39,$C6,$1F,$F6
	.dc.b	$E4,$F2,$6C,$F2,$7C,$3C,$FD,$FD,$89,$18,$21,$C9,$CB,$D2,$3E,$27
	.dc.b	$44,$8F,$88,$FF,$84,$7C,$CF,$ED,$39,$7F,$86,$F4,$FE,$E5,$5F,$1E
	.dc.b	$B8,$3F,$EF,$1A,$DF,$0F,$F9,$A6,$3F,$CF,$5D,$1C,$FA,$24,$7F,$4C
	.dc.b	$A4,$F7,$33,$1C,$E7,$0E,$E3,$38,$FF,$8B,$6B,$FB,$84,$E4,$E9,$D1
	.dc.b	$E9,$23,$DF,$E7,$2B,$E7,$8B,$5C,$E7,$FC,$61,$F3,$6F,$80,$FF,$1E
	.dc.b	$B6,$FF,$C9,$8D,$7F,$CE,$8D,$7F,$CF,$2E,$6B,$FC,$84,$DC,$FE,$23
	.dc.b	$67,$FD,$1D,$F3,$7F,$CE,$C7,$2F,$F9,$05,$73,$53,$11,$27,$FB,$AC
	.dc.b	$E9,$B3,$95,$87,$F9,$4E,$7A,$C6,$88,$6B,$0F,$B1,$1A,$8C,$1F,$EA
	.dc.b	$3F,$98,$FC,$F4,$FD,$D6,$C6,$7C,$3F,$F2,$C4,$7F,$95,$A3,$88,$76
	.dc.b	$D8,$D9,$C3,$C3,$FF,$30,$7C,$CB,$E6,$BF,$CD,$AD,$AF,$F3,$E3,$5F
	.dc.b	$F5,$23,$5F,$F5,$4B,$9C,$BF,$CE,$4D,$C7,$F2,$8D,$9F,$67,$47,$A3
	.dc.b	$9F,$D3,$D1,$27,$FC,$43,$A2,$75,$18,$9D,$3E,$E9,$D3,$C2,$73,$8F
	.dc.b	$FA,$0E,$55,$CA,$BE,$2F,$CD,$3C,$BF,$F4,$03,$63,$FC,$62,$D7,$E5
	.dc.b	$FC,$D7,$E8,$1B,$D3,$FE,$87,$8C,$63,$FE,$91,$6E,$BF,$B3,$F9,$AF
	.dc.b	$A2,$4E,$1F,$F4,$73,$F5,$3E,$69,$F3,$5F,$E9,$1F,$9A,$F5,$E7,$F6
	.dc.b	$9B,$97,$FA,$75,$B8,$FE,$FF,$E6,$5F,$31,$FE,$99,$6D,$FF,$92,$AB
	.dc.b	$6F,$FB,$B5,$AA,$36,$FE,$21,$B7,$FE,$AB,$6D,$F7,$BF,$F5,$4F,$BC
	.dc.b	$F9,$A7,$A9,$FE,$A9,$F8,$4F,$ED,$99,$1A,$63,$1F,$F3,$4E,$3F,$C5
	.dc.b	$E3,$3E,$4F,$13,$D5,$E1,$5B,$FF,$31,$CE,$9F,$F3,$2D,$F2,$43,$E6
	.dc.b	$7C,$EE,$1A,$E7,$4F,$F5,$F7,$CA,$3E,$6B,$FD,$81,$5A,$24,$EA,$32
	.dc.b	$CA,$FF,$3A,$CF,$4A,$E5,$1F,$E7,$8F,$09,$F3,$5F,$EC,$B6,$C6,$63
	.dc.b	$FD,$9D,$F3,$2F,$8B,$D4,$49,$D1,$FE,$82,$9D,$38,$3F,$7B,$5D,$1B
	.dc.b	$3D,$2B,$FD,$B5,$F3,$5F,$C7,$E4,$D3,$17,$FD,$77,$93,$FD,$D1,$19
	.dc.b	$F2,$38,$9E,$6F,$0A,$DF,$FA,$52,$74,$F9,$63,$FD,$D1,$F3,$3F,$80
	.dc.b	$27,$1F,$BB,$E6,$B2,$CA,$FF,$4E,$CF,$E1,$B2,$56,$67,$F8,$CC,$F2
	.dc.b	$7E,$14,$76,$E6,$91,$C9,$F9,$7E,$26,$23,$8C,$5D,$1C,$FF,$D9,$71
	.dc.b	$FD,$0F,$CD,$38,$BF,$E0,$03,$70,$FF,$58,$B5,$FE,$57,$F3,$5F,$DA
	.dc.b	$30,$FC,$7F,$CD,$3E,$67,$FB,$85,$EE,$7B,$DF,$E5,$23,$19,$6B,$1F
	.dc.b	$EE,$D5,$AA,$63,$D5,$63,$D5,$FC,$EC,$66,$31,$0C,$3C,$D1,$86,$E3
	.dc.b	$FE,$03,$AE,$0D,$F4,$7F,$AA,$DF,$68,$F2,$3E,$C6,$B9,$3C,$31,$C2
	.dc.b	$FC,$35,$D2,$7D,$D8,$FF,$84,$67,$C4,$DB,$F0,$63,$D6,$B9,$B5,$F5
	.dc.b	$4E,$DE,$66,$6F,$B3,$F1,$B0,$F1,$5F,$9B,$0E,$F3,$B9,$CB,$D8,$F5
	.dc.b	$1D,$8F,$06,$3F,$19,$D4,$FB,$27,$8C,$6C,$E3,$3B,$7C,$58,$F7,$3F
	.dc.b	$C6,$0F,$B8,$A6,$8B,$87,$DD,$76,$BC,$57,$C4,$F0,$9F,$C1,$C6,$1E
	.dc.b	$2D,$78,$B5,$EF,$F3,$5C,$5F,$E4,$6D,$4D,$23,$F1,$3E,$65,$F2,$FF
	.dc.b	$98,$D3,$F1,$B3,$C8,$D4,$67,$5E,$F1,$8D,$FE,$D3,$BA,$FC,$AD,$F4
	.dc.b	$9C,$CF,$88,$E0,$7D,$43,$85,$7F,$96,$67,$84,$64,$E1,$5D,$2F,$C5
	.dc.b	$1F,$43,$0F,$C5,$57,$0A,$F4,$CF,$0B,$E9,$7C,$23,$FC,$CD,$7C,$23
	.dc.b	$11,$E2,$3F,$A5,$DF,$3A,$F7,$DF,$87,$0F,$5E,$1C,$5B,$7E,$C9,$D3
	.dc.b	$FE,$32,$BF,$5E,$9F,$BF,$8E,$C2,$33,$5E,$63,$FA,$DC,$FF,$05,$8F
	.dc.b	$F1,$14,$60,$D4,$F5,$E2,$FE,$1F,$5B,$E8,$B5,$CB,$E6,$BF,$A7,$3C
	.dc.b	$6F,$DB,$3A,$7F,$CA,$F3,$A7,$D1,$19,$A4,$E5,$ED,$5C,$DC,$F5,$73
	.dc.b	$BF,$63,$D7,$E3,$1F,$32,$F9,$AD,$69,$CF,$1F,$EF,$35,$A9,$CD,$F1
	.dc.b	$BE,$15,$E9,$C3,$AB,$EB,$BE,$0E,$70,$E9,$29,$CB,$F4,$EF,$A6,$1E
	.dc.b	$7F,$17,$FC,$33,$87,$C5,$1F,$31,$F5,$2F,$FC,$1B,$3A,$7F,$CE,$B3
	.dc.b	$AE,$2E,$5D,$5F,$2E,$30,$79,$EB,$86,$BF,$D1,$23,$37,$FE,$76,$9C
	.dc.b	$C7,$F8,$68,$CC,$FF,$45,$8C,$CF,$F5,$AC,$3C,$8B,$E6,$BF,$CB,$B7
	.dc.b	$E9,$9B,$3F,$E8,$29,$B7,$E1,$9B,$5F,$EC,$79,$B5,$D7,$D2,$FB,$51
	.dc.b	$89,$F8,$71,$83,$DD,$5C,$2B,$C5,$18,$3F,$05,$3F,$C7,$4F,$99,$F3
	.dc.b	$5F,$E5,$67,$ED,$19,$9F,$97,$19,$7F,$D3,$31,$9C,$3E,$44,$E5,$FE
	.dc.b	$D1,$19,$3E,$44,$65,$FF,$71,$19,$9F,$F0,$0C,$65,$F9,$DC,$FF,$5E
	.dc.b	$31,$0F,$94,$FC,$4D,$7E,$0B,$E2,$FF,$39,$18,$C9,$18,$E0,$F8,$6D
	.dc.b	$AF,$A4,$DB,$FC,$3D,$70,$9F,$CD,$5F,$3A,$FD,$47,$CD,$7F,$99,$E7
	.dc.b	$F3,$CB,$E0,$33,$E6,$E9,$93,$FA,$3F,$9A,$70,$E3,$3F,$F6,$0C,$62
	.dc.b	$75,$3F,$B5,$F9,$77,$CD,$7E,$DE,$B5,$7F,$EC,$89,$CD,$FE,$BF,$9D
	.dc.b	$F3,$5F,$BD,$6C,$67,$CA,$C4,$FF,$01,$F8,$EF,$F5,$64,$FE,$DD,$B8
	.dc.b	$7F,$B5,$3D,$4F,$F3,$84,$E7,$DC,$FE,$11,$F3,$4B,$A7,$9F,$E3,$ED
	.dc.b	$7F,$19,$E4,$FE,$CF,$1F,$C6,$5F,$ED,$3F,$33,$7C,$84,$D8,$C3,$6B
	.dc.b	$94,$E9,$B3,$A6,$9A,$9E,$19,$FF,$BB,$3E,$65,$5F,$C7,$DB,$5F,$CB
	.dc.b	$7C,$2F,$F4,$97,$36,$6A,$4F,$F7,$DA,$DD,$FF,$81,$9A,$FF,$82,$9A
	.dc.b	$FF,$A7,$32,$FF,$02,$FC,$C7,$B7,$E6,$3D,$CD,$BF,$5B,$69,$DF,$FE
	.dc.b	$52,$BE,$35,$F0,$2F,$E3,$70,$7C,$7F,$0A,$CF,$C4,$6E,$1C,$A3,$C5
	.dc.b	$BF,$F1,$2C,$3F,$C2,$F7,$C6,$7E,$5B,$6B,$81,$FA,$BA,$FC,$0E,$3E
	.dc.b	$37,$F8,$DC,$3E,$85,$F4,$5F,$D9,$63,$FD,$77,$18,$C3,$FC,$48,$D8
	.dc.b	$FF,$16,$B5,$FF,$18,$FC,$D3,$E6,$BF,$47,$1F,$E3,$28,$C7,$6F,$F8
	.dc.b	$B5,$7C,$87,$F4,$59,$D5,$70,$7F,$C7,$7E,$0B,$E1,$7F,$CE,$F9,$7F
	.dc.b	$B3,$A7,$F8,$0F,$C5,$E1,$FF,$68,$D2,$31,$1F,$C1,$AA,$CF,$38,$FE
	.dc.b	$B4,$ED,$B3,$A9,$DF,$6F,$0F,$92,$95,$C2,$32,$F2,$8C,$BF,$40,$FE
	.dc.b	$2B,$1A,$71,$69,$DF,$07,$FA,$46,$3F,$D8,$58,$FB,$7E,$61,$F3,$3E
	.dc.b	$8F,$A2,$98,$FF,$71,$9C,$B6,$72,$F6,$9F,$83,$3A,$8F,$27,$CC,$FE
	.dc.b	$F5,$FE,$F4,$FF,$1D,$6F,$8F,$F2,$23,$51,$C3,$C0,$C4,$C8,$E3,$1B
	.dc.b	$F6,$A4,$74,$79,$79,$B8,$8F,$0C,$5D,$7F,$D6,$18,$7F,$9C,$67,$81
	.dc.b	$0D,$9E,$6A,$FB,$7F,$35,$FC,$32,$3D,$3F,$45,$E7,$33,$E2,$48,$C9
	.dc.b	$0F,$A9,$3C,$5F,$AD,$F3,$4F,$9B,$FC,$3B,$63,$FC,$3C,$D7,$7F,$07
	.dc.b	$A9,$91,$A7,$8C,$72,$F8,$6E,$7F,$2D,$2B,$9C,$65,$EB,$19,$7C,$71
	.dc.b	$97,$CB,$19,$D7,$FD,$AD,$82,$B5,$18,$9D,$72,$7E,$96,$B5,$7E,$03
	.dc.b	$B9,$0E,$E6,$4E,$E6,$6B,$BD,$F6,$AE,$97,$DA,$82,$FB,$BD,$0E,$64
	.dc.b	$BE,$6F,$77,$D0,$F7,$AF,$45,$F7,$8E,$B4,$BF,$04,$77,$9D,$3E,$18
	.dc.b	$CC,$F8,$AF,$BC,$F8,$6F,$D2,$F8,$5E,$EF,$84,$EF,$5D,$2B,$94,$F0
	.dc.b	$9E,$47,$92,$F5,$5E,$06,$C7,$33,$CC,$77,$3A,$99,$AE,$AF,$33,$13
	.dc.b	$E1,$9F,$5B,$63,$27,$33,$A9,$E4,$3E,$11,$99,$EA,$E0,$D3,$82,$47
	.dc.b	$A6,$38,$35,$FB,$57,$F5,$9E,$35,$EE,$69,$C0,$F6,$39,$8E,$07,$D2
	.dc.b	$25,$FC,$B7,$2F,$D2,$9E,$EE,$48,$FC,$53,$27,$CA,$3C,$26,$0F,$74
	.dc.b	$74,$78,$3E,$07,$83,$97,$E4,$B9,$78,$C7,$81,$CB,$E5,$8E,$17,$F9
	.dc.b	$39,$F9,$D5,$F5,$9A,$7D,$63,$C9,$1F,$39,$F8,$A7,$C7,$3D,$33,$E1
	.dc.b	$8E,$91,$8A,$F0,$1E,$38,$CC,$EE,$BC,$15,$D9,$B1,$98,$F1,$35,$FC
	.dc.b	$0D,$FE,$2D,$A7,$A0,$A7,$33,$8D,$F8,$9F,$96,$F8,$E3,$94,$FC,$47
	.dc.b	$F6,$7D,$6F,$9B,$FE,$E3,$F3,$6F,$A2,$A7,$F7,$49,$D6,$3F,$EC,$B6
	.dc.b	$BF,$DA,$67,$83,$FE,$CF,$6B,$FD,$B6,$37,$83,$1F,$F0,$25,$7F,$6A
	.dc.b	$AF,$E2,$AD,$4D,$3F,$B9,$3D,$7A,$FC,$E4,$7C,$1D,$7E,$AF,$7F,$D7
	.dc.b	$EC,$C9,$F2,$46,$5F,$34,$65,$F5,$46,$5F,$64,$67,$1F,$44,$79,$A7
	.dc.b	$E2,$74,$FF,$86,$3E,$71,$F4,$54,$CD,$7C,$79,$C3,$FE,$F3,$9D,$B9
	.dc.b	$31,$D5,$CF,$A6,$46,$08,$7F,$85,$E7,$FB,$93,$63,$C5,$3E,$3F,$12
	.dc.b	$4F,$77,$CD,$F3,$B8,$8E,$51,$73,$F6,$C3,$05,$67,$83,$FC,$75,$1F
	.dc.b	$E3,$FA,$DB,$63,$33,$FE,$3A,$F9,$D7,$C1,$FF,$8F,$3E,$75,$F4,$54
	.dc.b	$D4,$4C,$BF,$C8,$C7,$F9,$1E,$B8,$B8,$3F,$AC,$B8,$70,$D7,$8D,$7F
	.dc.b	$2A,$3F,$63,$87,$F9,$45,$B1,$FD,$89,$D7,$4F,$97,$A6,$4F,$F6,$6D
	.dc.b	$3E,$1D,$7F,$34,$F9,$CF,$F2,$94,$FF,$91,$DB,$3F,$21,$B8,$7F,$97
	.dc.b	$DF,$F2,$B4,$EF,$C7,$FF,$2E,$C6,$A3,$FB,$9B,$53,$C2,$BA,$4F,$02
	.dc.b	$4F,$96,$BB,$79,$1F,$E6,$8F,$9D,$7C,$F7,$6F,$A2,$97,$B3,$FC,$03
	.dc.b	$3A,$70,$1C,$3F,$3D,$AF,$D4,$6B,$FE,$7B,$6B,$EA,$E8,$F6,$F4,$C8
	.dc.b	$EB,$A2,$7E,$09,$98,$F9,$53,$87,$C7,$F4,$9F,$E5,$AD,$7F,$26,$DD
	.dc.b	$3F,$D1,$4D,$9F,$F0,$DB,$FE,$63,$7F,$CF,$4E,$53,$49,$D3,$FE,$8E
	.dc.b	$71,$E4,$53,$11,$FE,$92,$FA,$07,$A7,$49,$7F,$23,$E8,$3F,$63,$97
	.dc.b	$EA,$1B,$97,$FA,$7D,$B3,$FE,$30,$6A,$FA,$0F,$C0,$13,$8F,$FD,$42
	.dc.b	$FE,$0F,$E8,$1F,$41,$FE,$A4,$FA,$0F,$95,$7F,$61,$AF,$FA,$65,$B9
	.dc.b	$7E,$EB,$E7,$DF,$3D,$FE,$AC,$6D,$71,$E3,$E8,$66,$64,$7E,$12,$76
	.dc.b	$FF,$A9,$5B,$5F,$C8,$BE,$11,$F4,$0F,$7B,$FD,$73,$F0,$6F,$42,$93
	.dc.b	$1E,$9E,$35,$7F,$D6,$13,$B7,$F2,$6D,$7C,$11,$BC,$7F,$8A,$CE,$5B
	.dc.b	$5E,$D9,$FF,$64,$FC,$FF,$9B,$93,$FD,$A0,$DC,$3F,$CC,$2D,$5F,$34
	.dc.b	$FA,$0E,$3B,$C9,$FD,$2D,$6E,$B5,$18,$9D,$3F,$EC,$99,$D5,$7F,$B6
	.dc.b	$BC,$47,$D0,$52,$57,$EC,$3E,$75,$F2,$58,$8F,$F6,$95,$70,$7F,$6A
	.dc.b	$D7,$FD,$AD,$3C,$1C,$D7,$1D,$BF,$9C,$D6,$DF,$E7,$75,$BA,$FF,$6F
	.dc.b	$7D,$07,$F2,$99,$FE,$7E,$DD,$7F,$DE,$CD,$9F,$F4,$3D,$7F,$73,$8C
	.dc.b	$41,$A6,$23,$5E,$57,$FB,$EA,$32,$FF,$BF,$23,$FB,$E7,$D0,$7D,$9B
	.dc.b	$FF,$49,$DA,$30,$FF,$A5,$9A,$A5,$F4,$1C,$31,$FE,$AF,$5A,$27,$17
	.dc.b	$79,$D6,$1D,$A7,$38,$7F,$20,$7A,$E1,$D2,$31,$C9,$C1,$F1,$35,$F3
	.dc.b	$3E,$BE,$8C,$57,$E8,$67,$89,$0F,$AF,$1C,$7E,$83,$F1,$F1,$FC,$C1
	.dc.b	$AB,$3F,$EE,$31,$87,$FB,$9F,$D0,$3D,$2D,$53,$E7,$9F,$41,$F2,$B0
	.dc.b	$E1,$AD,$E8,$F0,$1E,$9C,$79,$65,$FE,$C2,$C3,$8E,$3F,$EC,$49,$CB
	.dc.b	$D7,$1E,$4E,$DA,$E4,$E1,$97,$92,$B9,$C2,$B9,$C6,$32,$F7,$B7,$1F
	.dc.b	$C0,$BC,$27,$78,$7E,$11,$B7,$FE,$D8,$BE,$33,$E2,$AE,$18,$78,$63
	.dc.b	$83,$6B,$86,$1D,$A7,$81,$AC,$7D,$93,$EA,$6A,$C3,$BB,$C8,$CE,$3E
	.dc.b	$47,$DD,$87,$77,$86,$1D,$DE,$07,$C7,$37,$7C,$A7,$85,$F6,$9E,$26
	.dc.b	$66,$3F,$E2,$1B,$FA,$AF,$F8,$99,$AF,$AC,$D7,$AA,$F8,$BF,$F1,$44
	.dc.b	$7D,$38,$C3,$A8,$CC,$A7,$33,$C3,$E8,$38,$A0,$E1,$A6,$9C,$1F,$2B
	.dc.b	$E4,$9F,$3B,$FE,$68,$DB,$F0,$B5,$CE,$B5,$39,$DB,$FC,$7D,$18,$DF
	.dc.b	$E9,$D6,$CF,$95,$87,$48,$CC,$78,$B0,$FC,$8C,$E6,$3E,$81,$C6,$33
	.dc.b	$87,$78,$FF,$27,$DF,$77,$67,$3C,$3F,$8A,$DF,$3B,$F4,$CE,$4F,$89
	.dc.b	$19,$9F,$4C,$79,$AF,$B3,$FC,$7A,$FB,$5F,$26,$D7,$08,$F2,$E1,$EC
	.dc.b	$9C,$D7,$6B,$FF,$48,$E9,$FE,$76,$C7,$DC,$FF,$9E,$27,$DD,$8F,$A2
	.dc.b	$B7,$85,$3F,$C7,$F7,$FE,$33,$6C,$6A,$BD,$7C,$15,$57,$F9,$EA,$BF
	.dc.b	$C9,$58,$3E,$83,$FA,$CC,$74,$E7,$E5,$39,$EB,$E5,$7C,$70,$FE,$3D
	.dc.b	$7F,$E6,$DC,$38,$C6,$0F,$97,$87,$28,$CE,$1E,$FC,$3C,$DF,$2E,$FA
	.dc.b	$07,$D0,$76,$9C,$C6,$32,$F7,$C6,$27,$51,$FE,$63,$AE,$51,$8C,$BC
	.dc.b	$51,$89,$E2,$6E,$FA,$4F,$9C,$78,$5F,$E2,$19,$FF,$37,$0F,$9E,$FA
	.dc.b	$AF,$F5,$E8,$D4,$FB,$E3,$5B,$FC,$F9,$CE,$7F,$EC,$E9,$C9,$FE,$76
	.dc.b	$8C,$BF,$B3,$33,$5E,$38,$CB,$FE,$ED,$BF,$0B,$FB,$98,$CC,$F0,$AF
	.dc.b	$F5,$FF,$D0,$3D,$4F,$C1,$4F,$A5,$FC,$14,$66,$3D,$B8,$5C,$3E,$14
	.dc.b	$DB,$F4,$4D,$AF,$EF,$B3,$6B,$C9,$7D,$0E,$F5,$98,$4E,$EB,$C4,$7F
	.dc.b	$A5,$8F,$E2,$C7,$CE,$FA,$0F,$F2,$EB,$9E,$6F,$4C,$62,$BE,$B4,$F0
	.dc.b	$7E,$84,$67,$1F,$F6,$FC,$E5,$FA,$11,$98,$FF,$0A,$99,$3E,$83,$99
	.dc.b	$FA,$55,$E0,$F1,$BF,$51,$F2,$DF,$98,$BF,$AD,$93,$E4,$B3,$C5,$F0
	.dc.b	$1B,$3F,$31,$B8,$FE,$16,$B8,$4F,$FA,$DA,$FA,$4F,$DD,$AD,$E7,$F4
	.dc.b	$1F,$DC,$7D,$03,$CA,$BE,$B6,$1B,$8F,$FB,$02,$53,$FB,$FF,$9F,$FC
	.dc.b	$DC,$E6,$31,$C5,$23,$FD,$8F,$F3,$EF,$A0,$C6,$9F,$A3,$8C,$4F,$FB
	.dc.b	$33,$9D,$F4,$1F,$C2,$DB,$19,$F8,$AC,$4F,$F0,$FF,$73,$FD,$5F,$3F
	.dc.b	$ED,$46,$E1,$FA,$FF,$4F,$FC,$ED,$39,$F6,$FF,$8B,$7D,$02,$E9,$FC
	.dc.b	$67,$37,$C9,$E2,$FF,$DC,$5C,$BF,$D9,$63,$F9,$03,$FD,$9F,$E8,$2F
	.dc.b	$93,$C2,$B6,$A3,$35,$A6,$CE,$9B,$A7,$F1,$BB,$E9,$3C,$AB,$83,$FE
	.dc.b	$F0,$FA,$0E,$8F,$F7,$66,$D7,$F9,$90,$FE,$E0,$E1,$C3,$5F,$F3,$4B
	.dc.b	$C3,$A1,$3F,$DD,$F8,$5F,$44,$F9,$3F,$E7,$97,$BA,$D4,$E5,$F7,$FD
	.dc.b	$0F,$C2,$FA,$1F,$86,$D8,$C9,$F1,$23,$2D,$3E,$3F,$17,$CA,$F4,$FE
	.dc.b	$5B,$63,$27,$CC,$8C,$B4,$CE,$9F,$27,$83,$F8,$DE,$FF,$E1,$BB,$E5
	.dc.b	$3F,$CC,$7E,$7D,$F3,$FF,$CC,$A3,$E4,$08,$CF,$57,$F8,$89,$7C,$B3
	.dc.b	$87,$17,$C6,$FE,$87,$D1,$7E,$DD,$FF,$14,$B5,$EA,$D7,$7E,$0F,$F8
	.dc.b	$98,$7C,$37,$FA,$1D,$7C,$75,$31,$12,$7F,$C5,$9C,$DF,$61,$FF,$17
	.dc.b	$57,$07,$3C,$1F,$EA,$68,$C4,$7F,$51,$7B,$BA,$38,$F7,$78,$7E,$5A
	.dc.b	$46,$4F,$E2,$51,$E5,$32,$F3,$8C,$3B,$94,$EA,$FC,$D3,$FC,$6E,$B6
	.dc.b	$FF,$63,$F9,$E7,$D0,$7F,$92,$1B,$1F,$E5,$26,$BF,$59,$AF,$E7,$DC
	.dc.b	$D7,$66,$E3,$FE,$C6,$6C,$FC,$6F,$53,$FD,$69,$F3,$FF,$C0,$63,$DC
	.dc.b	$FF,$1C,$6F,$8B,$AF,$06,$46,$08,$7F,$7B,$9E,$4D,$F6,$FA,$91,$EB
	.dc.b	$E6,$D7,$47,$82,$3C,$71,$9C,$FE,$3D,$7F,$AE,$6B,$47,$23,$EA,$7C
	.dc.b	$F7,$F1,$66,$F8,$5F,$EC,$F8,$C4,$79,$7A,$71,$32,$36,$7C,$08,$FF
	.dc.b	$6B,$7C,$EB,$E7,$A9,$27,$F0,$51,$FE,$EC,$34,$E2,$34,$FF,$87,$63
	.dc.b	$5F,$0B,$D3,$E5,$E4,$F2,$8E,$E6,$1E,$11,$87,$47,$81,$D6,$7F,$8A
	.dc.b	$A5,$6E,$B8,$9B,$38,$9E,$37,$8F,$17,$FA,$4E,$12,$BE,$57,$D6,$F7
	.dc.b	$5F,$E9,$9A,$D3,$D2,$74,$F7,$3C,$27,$73,$57,$DE,$78,$C6,$A7,$8B
	.dc.b	$B3,$13,$BB,$F1,$BC,$6F,$C7,$3E,$89,$CB,$E5,$8C,$DF,$6B,$EA,$4B
	.dc.b	$E9,$5E,$78,$C3,$C4,$C4,$69,$F2,$46,$9F,$01,$A3,$81,$A8,$E1,$1F
	.dc.b	$56,$FC,$B3,$F3,$AF,$CB,$18,$BF,$B3,$3C,$CE,$A7,$33,$AD,$73,$75
	.dc.b	$1C,$1D,$1C,$23,$4E,$23,$51,$82,$5F,$CB,$9F,$AA,$7D,$DA,$F9,$37
	.dc.b	$E5,$6C,$68,$ED,$3E,$E3,$93,$DE,$36,$D3,$CF,$3C,$DA,$73,$BF,$AD
	.dc.b	$5C,$E3,$DE,$F4,$BF,$3C,$79,$4E,$15,$F2,$9A,$7C,$4B,$F9,$91,$98
	.dc.b	$F8,$4D,$9F,$B5,$7F,$12,$FC,$15,$EC,$3C,$24,$30,$48,$E0,$E5,$F4
	.dc.b	$BE,$53,$B9,$E3,$3C,$A7,$E7,$0D,$4F,$8A,$BC,$95,$EF,$24,$E9,$F2
	.dc.b	$D7,$82,$3F,$14,$75,$3B,$9D,$4F,$34,$FC,$F7,$11,$C4,$97,$F5,$AF
	.dc.b	$E1,$B6,$FC,$35,$EF,$8F,$96,$F7,$9F,$43,$67,$75,$D2,$3E,$E9,$F8
	.dc.b	$13,$C3,$7E,$3A,$FB,$71,$EA,$3F,$D8,$5D,$9F,$E2,$D7,$FB,$73,$5C
	.dc.b	$F8,$DF,$C5,$E7,$FB,$4F,$CE,$FF,$78,$9C,$9C,$27,$58,$FF,$1D,$BE
	.dc.b	$57,$C3,$AB,$36,$18,$7F,$76,$7F,$BF,$35,$FF,$6B,$36,$7C,$3D,$1C
	.dc.b	$A7,$8F,$4D,$AD,$63,$FC,$B5,$B3,$DD,$56,$F9,$F8,$7A,$D2,$BC,$31
	.dc.b	$97,$C7,$19,$7C,$71,$97,$CB,$19,$C7,$E2,$CF,$97,$3F,$EF,$55,$FE
	.dc.b	$3C,$6B,$FE,$65,$6B,$F1,$FA,$3F,$C8,$2F,$C5,$8E,$35,$FE,$18,$6E
	.dc.b	$5B,$9F,$F1,$27,$CE,$7F,$80,$2B,$FC,$3A,$FF,$77,$8F,$F7,$F2,$E7
	.dc.b	$F9,$0D,$8E,$3D,$18,$33,$32,$32,$76,$E5,$E2,$47,$C3,$F3,$95,$EF
	.dc.b	$17,$0F,$1C,$DC,$7F,$B0,$63,$FE,$57,$9F,$3D,$79,$2B,$EE,$7C,$DB
	.dc.b	$DE,$F8,$78,$A3,$0F,$C6,$6B,$FE,$68,$6B,$FE,$91,$73,$7F,$E4,$06
	.dc.b	$E3,$FD,$C5,$B3,$F9,$3F,$9B,$FF,$05,$AA,$FF,$23,$C7,$F0,$13,$F8
	.dc.b	$7C,$1A,$67,$9B,$F2,$8E,$3A,$2C,$7E,$5F,$E6,$3F,$51,$7F,$E5,$26
	.dc.b	$E3,$FE,$03,$6E,$1F,$A5,$7F,$CC,$F3,$BE,$7E,$44,$9F,$F3,$24,$EA
	.dc.b	$B9,$4E,$67,$D5,$18,$77,$E1,$7E,$A7,$E6,$5F,$37,$F3,$9B,$3F,$92
	.dc.b	$6B,$FE,$94,$6B,$FE,$BD,$73,$8F,$F9,$B1,$B7,$FE,$1D,$6C,$FA,$FA
	.dc.b	$BE,$17,$3F,$C3,$E8,$93,$EA,$8C,$BE,$C8,$CB,$EA,$3E,$24,$7F,$01
	.dc.b	$F9,$AF,$D0,$61,$FE,$2B,$6C,$FF,$0D,$6A,$7F,$52,$E7,$C2,$E4,$47
	.dc.b	$FD,$1D,$1A,$78,$46,$5F,$11,$83,$7D,$5F,$C4,$FE,$69,$E5,$D2,$4F
	.dc.b	$FA,$67,$E6,$BF,$60,$F2,$6E,$DF,$E9,$46,$E5,$FE,$45,$6C,$FF,$27
	.dc.b	$F9,$AF,$C2,$A6,$93,$8B,$F9,$27,$CD,$3E,$6B,$F9,$37,$CD,$38,$29
	.dc.b	$30,$FF,$56,$7C,$D3,$E1,$3F,$96,$36,$3F,$D6,$15,$C5,$B5,$AA,$91
	.dc.b	$C9,$AF,$D7,$F7,$5F,$34,$F5,$FF,$9C,$7C,$1B,$E5,$A9,$2F,$C7,$E2
	.dc.b	$57,$E7,$CE,$DF,$F6,$13,$5F,$A7,$A7,$F0,$BE,$2F,$F3,$64,$E7,$8F
	.dc.b	$FD,$7E,$D8,$CE,$1E,$EC,$3F,$A5,$7C,$D7,$BF,$1F,$BA,$DC,$3F,$D8
	.dc.b	$ED,$BF,$F3,$C7,$CA,$3E,$6B,$8E,$F3,$0F,$F6,$B5,$ED,$FE,$BD,$5B
	.dc.b	$73,$3B,$7B,$BF,$D6,$E3,$51,$F8,$EF,$09,$F3,$5F,$D7,$DB,$1F,$EE
	.dc.b	$36,$BF,$94,$F9,$87,$C5,$EE,$A4,$ED,$FE,$D1,$3A,$8C,$10,$F0,$46
	.dc.b	$A7,$FB,$77,$CD,$7F,$26,$93,$4C,$57,$F6,$E8,$CF,$95,$C0,$8F,$FB
	.dc.b	$E6,$35,$3C,$67,$33,$E9,$F0,$FF,$BE,$7C,$CF,$D7,$8F,$F0,$33,$53
	.dc.b	$6D,$F3,$5C,$38,$E5,$7F,$83,$78,$BF,$DF,$F7,$BA,$CC,$FE,$EA,$3A
	.dc.b	$19,$78,$71,$7F,$AD,$63,$F3,$11,$83,$C7,$CD,$E6,$3E,$5F,$4D,$8D
	.dc.b	$1E,$68,$C3,$F5,$63,$FC,$31,$F3,$4E,$2A,$49,$E3,$F3,$4F,$4B,$FA
	.dc.b	$94,$3E,$69,$F3,$3F,$EC,$9C,$39,$4F,$81,$AF,$F1,$66,$EC,$D7,$FD
	.dc.b	$A1,$8F,$58,$FF,$19,$61,$D2,$FF,$C6,$93,$97,$C1,$87,$7C,$BF,$DB
	.dc.b	$15,$B9,$63,$E0,$8C,$5F,$16,$E5,$EA,$78,$4E,$F0,$E9,$7E,$27,$CB
	.dc.b	$5B,$6E,$39,$9F,$F2,$45,$71,$6E,$3E,$4A,$DC,$7A,$0F,$5D,$F3,$76
	.dc.b	$73,$3D,$33,$C5,$DD,$E7,$1F,$55,$F9,$A7,$A1,$C0,$D5,$F5,$53,$C1
	.dc.b	$EE,$70,$C3,$D0,$EA,$78,$4E,$E3,$A1,$BC,$3C,$8E,$F0,$EE,$7F,$91
	.dc.b	$A7,$EB,$47,$F9,$AC,$F4,$94,$A5,$7F,$C8,$5F,$0D,$C7,$D5,$34,$93
	.dc.b	$FE,$30,$F9,$AF,$F1,$8A,$69,$1D,$1F,$13,$E6,$9F,$2F,$F8,$F8,$F8
	.dc.b	$E7,$F9,$FE,$7E,$84,$EB,$7F,$A5,$39,$39,$57,$1D,$FF,$1D,$19,$7F
	.dc.b	$21,$5C,$63,$A9,$E5,$38,$D7,$47,$13,$C2,$BA,$6F,$F1,$67,$2F,$C9
	.dc.b	$BE,$95,$E0,$9C,$9F,$E9,$78,$CC,$F1,$C3,$D7,$7D,$2B,$E8,$46,$30
	.dc.b	$F1,$36,$3C,$ED,$C3,$B3,$6F,$A4,$FA,$89,$A7,$CB,$58,$FC,$18,$C4
	.dc.b	$7C,$59,$F5,$63,$F4,$67,$85,$6F,$1F,$77,$06,$5D,$9F,$C6,$23,$FC
	.dc.b	$0C,$DE,$7D,$63,$FE,$63,$F9,$AF,$E9,$B3,$93,$F9,$26,$08,$CC,$FF
	.dc.b	$89,$AF,$FC,$F5,$5A,$9C,$D7,$92,$FA,$BC,$7A,$BB,$E1,$EE,$7C,$3F
	.dc.b	$10,$F9,$97,$CD,$6B,$3E,$79,$FF,$B2,$EB,$53,$9B,$E3,$7C,$2B,$DD
	.dc.b	$18,$9D,$47,$F3,$DA,$E2,$E8,$C5,$71,$C3,$FC,$83,$E1,$FF,$86,$E7
	.dc.b	$FA,$20,$F9,$8F,$9D,$3F,$ED,$D9,$D3,$FE,$59,$9D,$6F,$EE,$9C,$E9
	.dc.b	$FA,$79,$C9,$FD,$2A,$32,$FF,$B9,$CC,$D7,$8B,$A7,$EA,$C6,$2B,$F0
	.dc.b	$D5,$C2,$BF,$A0,$61,$E4,$5F,$34,$F5,$78,$CD,$C3,$FA,$CC,$DA,$F4
	.dc.b	$4D,$C3,$F2,$53,$6B,$C9,$36,$BF,$DF,$93,$67,$E2,$5F,$DC,$8C,$56
	.dc.b	$EB,$C7,$3E,$23,$F0,$35,$FE,$D7,$3E,$6F,$CD,$7F,$95,$5F,$ED,$A6
	.dc.b	$67,$E6,$46,$5F,$F3,$AC,$67,$0F,$99,$39,$7F,$55,$19,$3E,$64,$65
	.dc.b	$FE,$54,$66,$7F,$B4,$46,$5F,$A5,$D1,$F4,$23,$10,$F8,$FF,$C9,$BF
	.dc.b	$43,$E3,$3F,$37,$18,$BF,$C3,$46,$38,$BE,$8B,$63,$F1,$79,$A3,$17
	.dc.b	$F4,$9B,$97,$E1,$E7,$55,$FE,$10,$F9,$A7,$97,$FA,$A5,$EB,$D8,$C3
	.dc.b	$27,$FC,$2F,$F3,$4E,$1C,$67,$FE,$9E,$8C,$4E,$A7,$F6,$BF,$2E,$F9
	.dc.b	$AF,$DB,$D6,$AF,$FD,$4B,$39,$BF,$F1,$1F,$3B,$E6,$BF,$7A,$D8,$CF
	.dc.b	$95,$89,$FE,$03,$EE,$7F,$AA,$A7,$FC,$56,$DC,$3F,$D6,$3E,$A7,$F9
	.dc.b	$C6,$73,$ED,$FF,$08,$F9,$A5,$D3,$CF,$F1,$D6,$BF,$C6,$BF,$32,$F9
	.dc.b	$AF,$E1,$ED,$8F,$F1,$F7,$84,$F5,$BE,$5C,$70,$AD,$91,$DC,$6B,$4F
	.dc.b	$F6,$2D,$F3,$6D,$7F,$36,$42,$34,$FF,$77,$6B,$FD,$ED,$AF,$E3,$7C
	.dc.b	$A7,$CC,$FF,$34,$ED,$FE,$F3,$F3,$36,$73,$97,$F9,$5E,$72,$D8,$CD
	.dc.b	$FF,$BA,$E7,$2D,$8C,$F6,$7F,$79,$9C,$B6,$3E,$93,$7C,$3F,$07,$0F
	.dc.b	$85,$B8,$77,$6E,$1D,$5F,$E7,$BF,$43,$FC,$92,$3F,$C2,$CD,$7F,$C3
	.dc.b	$4D,$7E,$3B,$52,$FA,$2F,$99,$18,$6C,$66,$BF,$BF,$4E,$5B,$F3,$6F
	.dc.b	$9B,$FD,$04,$FF,$80,$E3,$1D,$BA,$F9,$17,$5F,$0B,$DD,$FF,$8B,$7C
	.dc.b	$3F,$F7,$98,$F8,$BF,$F1,$96,$FF,$DD,$E3,$13,$FE,$84,$77,$F0,$BA
	.dc.b	$D3,$FC,$71,$1A,$95,$71,$BD,$47,$E2,$27,$BF,$76,$23,$F8,$37,$C9
	.dc.b	$59,$36,$64,$F8,$63,$46,$1E,$F8,$7F,$92,$7E,$65,$F3,$5C,$FE,$8A
	.dc.b	$57,$FA,$72,$72,$D9,$CB,$C8,$F3,$4F,$96,$77,$19,$0E,$33,$F8,$D3
	.dc.b	$C9,$FF,$0F,$46,$6F,$FC,$B9,$F3,$7F,$BC,$BF,$F5,$6B,$FC,$71,$BE
	.dc.b	$3F,$C3,$8D,$47,$76,$DF,$F8,$A5,$E1,$19,$9E,$1E,$0F,$1F,$6E,$4F
	.dc.b	$67,$A7,$9D,$88,$E9,$17,$0E,$93,$F4,$FE,$6D,$E2,$78,$FE,$8A,$61
	.dc.b	$FE,$C4,$9C,$B6,$72,$F9,$E7,$94,$48,$F7,$C6,$AF,$FC,$85,$5C,$A7
	.dc.b	$E6,$7C,$DB,$E7,$3F,$08,$D8,$FF,$69,$B5,$DF,$C1,$EA,$64,$69,$E3
	.dc.b	$1C,$BE,$1F,$9F,$CB,$4A,$E7,$19,$7A,$C6,$5F,$1C,$7E,$57,$B1,$F3
	.dc.b	$9F,$97,$F9,$C7,$D1,$53,$51,$2F,$FD,$D3,$3A,$6C,$E5,$FC,$55,$74
	.dc.b	$72,$43,$35,$BD,$7F,$CD,$17,$CE,$7F,$24,$D8,$FF,$36,$57,$39,$E2
	.dc.b	$73,$8E,$37,$CF,$97,$FD,$5D,$5B,$9C,$D6,$E7,$38,$6E,$3F,$9F,$E1
	.dc.b	$DE,$7F,$D6,$CD,$8F,$E8,$77,$CE,$78,$C7,$33,$8E,$1C,$D4,$ED,$C9
	.dc.b	$B3,$37,$BB,$ED,$06,$CF,$18,$D9,$C6,$33,$3C,$6B,$EB,$4F,$F5,$8C
	.dc.b	$3E,$3C,$E6,$3C,$17,$CC,$E9,$7E,$03,$9C,$71,$C3,$DA,$FD,$2C,$3A
	.dc.b	$BF,$40,$98,$7C,$C7,$47,$13,$E3,$E1,$E1,$72,$70,$AE,$B3,$C4,$8E
	.dc.b	$09,$1F,$20,$E3,$1E,$C7,$B1,$F1,$0C,$C7,$D8,$AF,$A5,$87,$B9,$F4
	.dc.b	$D7,$1A,$F1,$1D,$5F,$0B,$E8,$3A,$CF,$17,$E4,$DF,$AE,$77,$1E,$17
	.dc.b	$A1,$8C,$3E,$5C,$76,$3E,$E5,$F8,$B0,$F6,$C7,$17,$D0,$F6,$3C,$37
	.dc.b	$B7,$CD,$1C,$30,$F2,$47,$C0,$3D,$07,$7B,$FB,$97,$F0,$AF,$94,$F8
	.dc.b	$8E,$A4,$39,$19,$3D,$15,$DB,$0F,$7B,$F8,$99,$E7,$7E,$E9,$F8,$38
	.dc.b	$7A,$AF,$9D,$ED,$FC,$C6,$1E,$36,$9E,$1A,$E2,$FB,$88,$7C,$53,$DA
	.dc.b	$FC,$D8,$E5,$1E,$8B,$F1,$5F,$CC,$7E,$0C,$75,$8C,$4F,$C2,$BF,$24
	.dc.b	$F1,$8E,$77,$D6,$38,$1E,$0A,$F6,$47,$C2,$C3,$F0,$51,$83,$EA,$3F
	.dc.b	$12,$B9,$C6,$6B,$D7,$7E,$0B,$ED,$1F,$3A,$F8,$3E,$03,$F8,$17,$57
	.dc.b	$F3,$32,$3F,$CC,$3A,$AB,$FC,$9A,$2D,$7F,$2A,$9F,$E7,$5F,$01,$FC
	.dc.b	$EF,$DA,$97,$FC,$E2,$32,$FF,$3A,$8C,$BF,$CC,$23,$FA,$27,$47,$F8
	.dc.b	$67,$E7,$5F,$45,$4C,$D7,$E9,$A7,$0F,$ED,$27,$6E,$48,$7F,$83,$5C
	.dc.b	$B6,$79,$B9,$E9,$F0,$74,$C8,$E9,$1E,$33,$0F,$38,$C3,$E1,$F1,$3D
	.dc.b	$04,$7C,$9F,$3D,$5F,$14,$5D,$5C,$3F,$E8,$09,$64,$BF,$F1,$D7,$CF
	.dc.b	$3E,$0F,$FC,$79,$F3,$CF,$A2,$A6,$A2,$61,$FC,$5A,$74,$D9,$CB,$FE
	.dc.b	$2F,$AE,$8F,$F9,$89,$AF,$8C,$FF,$49,$B8,$70,$D7,$D1,$2E,$8F,$2F
	.dc.b	$4C,$9F,$14,$7A,$0C,$3F,$E3,$F9,$4E,$63,$F9,$6F,$CF,$BC,$8F,$F2
	.dc.b	$D3,$63,$31,$8F,$23,$2C,$8C,$9C,$A3,$FC,$C5,$F3,$CF,$9F,$E7,$F4
	.dc.b	$52,$BF,$9D,$CE,$5B,$39,$7F,$CA,$98,$2B,$9C,$C9,$FF,$44,$D7,$17
	.dc.b	$E4,$07,$0F,$CB,$6B,$F7,$E9,$D1,$EE,$E9,$93,$F0,$A3,$E0,$98,$7C
	.dc.b	$71,$87,$FA,$9F,$D0,$7F,$96,$B9,$3F,$14,$D8,$FF,$65,$B5,$3F,$E5
	.dc.b	$EE,$5F,$C8,$C6,$BC,$7F,$F4,$44,$EC,$9D,$7F,$2B,$D7,$CF,$D3,$4B
	.dc.b	$F9,$31,$97,$C7,$19,$7F,$CF,$9A,$FF,$B5,$1D,$90,$CC,$F8,$3B,$3F
	.dc.b	$D3,$1F,$42,$F5,$A9,$2B,$FB,$8F,$D0,$BB,$FF,$D3,$CD,$8C,$C6,$0F
	.dc.b	$F5,$17,$D0,$FE,$26,$3F,$D1,$CD,$DF,$E0,$C2,$31,$C1,$FB,$36,$C7
	.dc.b	$FB,$D3,$E8,$5F,$43,$FE,$AD,$FA,$1F,$B3,$3F,$E0,$46,$BF,$E9,$B6
	.dc.b	$E1,$DB,$E8,$1F,$3F,$FE,$B5,$6D,$71,$E7,$E8,$66,$64,$7F,$84,$27
	.dc.b	$6F,$FA,$9D,$B5,$F6,$BE,$19,$F4,$2F,$7F,$FD,$85,$F1,$8F,$46,$93
	.dc.b	$1E,$9E,$55,$7F,$D6,$53,$B7,$FC,$3E,$D7,$C1,$1B,$C7,$E4,$CE,$5B
	.dc.b	$5E,$D9,$FF,$69,$7C,$37,$9B,$8B,$F8,$F3,$70,$07,$0B,$E6,$9F,$43
	.dc.b	$CB,$79,$3F,$E3,$2A,$DD,$6A,$31,$3A,$7F,$D9,$93,$AA,$FF,$72,$79
	.dc.b	$0F,$A1,$A4,$AF,$F1,$EF,$D0,$3E,$4F,$11,$FE,$D5,$AE,$0F,$F9,$09
	.dc.b	$AF,$FB,$66,$78,$39,$AE,$3B,$7E,$3A,$B6,$FE,$42,$B7,$5F,$EE,$1F
	.dc.b	$A1,$FE,$63,$3F,$94,$6E,$BF,$D1,$5B,$3F,$22,$9E,$A7,$FB,$FE,$73
	.dc.b	$26,$98,$9D,$C7,$F9,$6A,$33,$E8,$71,$3F,$A9,$61,$E5,$BB,$87,$F5
	.dc.b	$5F,$A0,$FB,$CD,$E2,$CE,$3F,$6A,$72,$F9,$7E,$87,$2A,$32,$D9,$CC
	.dc.b	$63,$35,$62,$65,$7F,$9C,$2B,$53,$C3,$96,$4E,$5E,$F7,$FB,$0C,$3D
	.dc.b	$11,$A9,$65,$FC,$03,$0F,$D3,$47,$E0,$3A,$7C,$FF,$43,$F9,$18,$FF
	.dc.b	$3F,$B5,$E3,$87,$F7,$78,$C5,$FF,$0A,$FA,$17,$89,$DA,$DF,$42,$FA
	.dc.b	$0E,$1D,$3D,$89,$3F,$E8,$CB,$E3,$87,$E9,$2F,$84,$FE,$9A,$72,$F4
	.dc.b	$BE,$6E,$DA,$E4,$E1,$87,$8E,$B9,$C2,$B9,$C6,$31,$F7,$37,$1F,$F0
	.dc.b	$A3,$C2,$77,$87,$F8,$5D,$B7,$FB,$AB,$E3,$3E,$2A,$E1,$87,$86,$38
	.dc.b	$36,$B8,$61,$DA,$78,$1A,$C7,$D9,$3E,$A6,$AC,$3B,$BC,$8C,$E3,$E4
	.dc.b	$7D,$D8,$77,$78,$61,$DD,$E0,$7C,$73,$77,$CA,$78,$5F,$69,$E2,$66
	.dc.b	$63,$FE,$00,$BF,$AA,$FF,$8A,$5A,$FA,$CD,$78,$6F,$8E,$FF,$03,$C7
	.dc.b	$D3,$8C,$3A,$8C,$CF,$F5,$F7,$71,$FC,$A3,$E8,$6A,$69,$1D,$BF,$23
	.dc.b	$E5,$1F,$3D,$FE,$00,$DB,$F0,$75,$CC,$F8,$53,$AD,$FE,$7C,$66,$7F
	.dc.b	$13,$39,$8F,$90,$67,$7F,$F2,$D4,$66,$7A,$CF,$94,$EB,$7E,$1A,$FF
	.dc.b	$13,$E1,$E5,$8C,$4F,$F9,$46,$31,$87,$8A,$39,$5F,$8A,$B7,$1F,$E6
	.dc.b	$9A,$DB,$E4,$37,$7E,$49,$DC,$79,$CF,$EA,$38,$79,$E7,$C7,$1E,$6C
	.dc.b	$3C,$17,$D3,$0F,$5B,$6B,$A4,$FF,$5A,$D3,$FC,$17,$87,$C5,$7F,$C1
	.dc.b	$B8,$77,$11,$99,$E3,$1F,$C3,$0F,$1F,$07,$F8,$D2,$B9,$E5,$6B,$E6
	.dc.b	$CD,$AF,$4F,$3D,$B7,$D0,$FF,$59,$BF,$21,$F5,$27,$51,$9A,$F0,$1E
	.dc.b	$48,$CF,$67,$58,$C5,$72,$7F,$94,$D7,$1B,$F7,$D7,$8F,$E3,$DF,$40
	.dc.b	$FA,$1E,$5A,$EF,$2F,$EF,$F5,$A9,$CD,$F1,$BE,$15,$DF,$2F,$8F,$39
	.dc.b	$AD,$CF,$8E,$F9,$3B,$C3,$B5,$A7,$2F,$BB,$C3,$FF,$12,$57,$AC,$7C
	.dc.b	$FF,$D3,$9F,$F0,$9C,$E9,$FA,$93,$AD,$FE,$24,$E7,$4F,$F3,$4C,$E4
	.dc.b	$F4,$46,$5F,$F0,$E9,$9A,$F1,$75,$7D,$98,$C5,$7F,$94,$AB,$85,$7F
	.dc.b	$B1,$30,$F2,$2F,$A1,$7A,$FC,$66,$E1,$F1,$E6,$D7,$A2,$6E,$1F,$E5
	.dc.b	$E9,$B5,$E4,$9B,$5F,$E3,$29,$B3,$F1,$2F,$F0,$51,$8A,$DD,$78,$E7
	.dc.b	$C4,$7E,$1E,$BF,$CE,$87,$CD,$FA,$1F,$F2,$FB,$F6,$CC,$CF,$CC,$8C
	.dc.b	$BF,$92,$8C,$E1,$F3,$27,$2F,$F9,$DA,$32,$7C,$C8,$CB,$FE,$E9,$33
	.dc.b	$3F,$EE,$C8,$CB,$F4,$BA,$7E,$84,$62,$1F,$31,$F8,$8A,$FC,$1F,$CA
	.dc.b	$7E,$86,$31,$92,$31,$C5,$F4,$5B,$5F,$8B,$9D,$3F,$99,$AE,$33,$AB
	.dc.b	$FA,$8F,$EB,$7E,$85,$E5,$DF,$67,$0D,$C3,$F4,$5F,$40,$E1,$FD,$A3
	.dc.b	$63,$F4,$D6,$BC,$D4,$9F,$D5,$7D,$03,$E8,$71,$9F,$F0,$08,$C4,$FE
	.dc.b	$BF,$99,$F4,$3F,$C6,$2B,$F6,$ED,$7F,$58,$D8,$DC,$FF,$5A,$F7,$FF
	.dc.b	$D7,$0D,$F6,$B8,$46,$2F,$FD,$CA,$D7,$FD,$02,$7F,$A0,$BE,$85,$1D
	.dc.b	$DB,$5F,$D9,$9B,$3F,$EE,$E6,$BC,$87,$D0,$3E,$87,$F8,$D3,$6B,$F8
	.dc.b	$D7,$6F,$FA,$3F,$8B,$35,$27,$FB,$9D,$6E,$FF,$DF,$6D,$7B,$35,$F4
	.dc.b	$BB,$E9,$FF,$75,$E8,$FA,$37,$C8,$F5,$BF,$E7,$D3,$98,$FE,$75,$CF
	.dc.b	$FE,$03,$F8,$1B,$E4,$BE,$8A,$C6,$1B,$A7,$F3,$3A,$FF,$05,$46,$31
	.dc.b	$FE,$95,$18,$9D,$5F,$D1,$CF,$FD,$FF,$3F,$E1,$1F,$07,$BF,$1F,$81
	.dc.b	$B7,$C1,$B8,$74,$C3,$F8,$AA,$FA,$17,$D1,$7E,$15,$B7,$FE,$1C,$6C
	.dc.b	$FD,$53,$84,$FE,$2F,$E8,$9F,$45,$F9,$D9,$D4,$63,$AF,$FC,$4A,$BE
	.dc.b	$5F,$F9,$E4,$EA,$B8,$3F,$81,$FA,$3F,$DB,$3F,$E2,$C6,$BE,$16,$BB
	.dc.b	$F0,$7F,$C5,$67,$FB,$F9,$FD,$F7,$C7,$FF,$A6,$20,$DF,$90,$93,$FD
	.dc.b	$56,$72,$D9,$CB,$FD,$B2,$3F,$D5,$ED,$F9,$8E,$87,$5F,$72,$47,$38
	.dc.b	$CB,$D6,$3F,$D5,$3F,$42,$FA,$3F,$27,$D1,$CA,$FE,$CD,$39,$6C,$FF
	.dc.b	$91,$A3,$11,$F4,$5A,$FE,$9D,$AF,$EC,$9D,$F3,$B8,$F1,$5E,$23,$E2
	.dc.b	$7A,$9F,$EB,$CF,$A3,$FD,$BE,$1F,$DD,$DF,$E2,$4D,$F2,$35,$E2,$C8
	.dc.b	$C1,$0F,$F0,$8C,$F2,$6F,$B7,$D4,$8F,$5F,$3E,$BA,$3C,$11,$E3,$8F
	.dc.b	$F6,$7F,$D0,$BE,$8F,$CD,$F4,$72,$FF,$C1,$93,$96,$CF,$F9,$B2,$31
	.dc.b	$1F,$97,$6B,$FC,$35,$AF,$F1,$87,$7B,$7C,$78,$08,$CE,$DB,$AF,$91
	.dc.b	$D5,$FE,$E4,$FA,$37,$D1,$D2,$61,$FE,$21,$8F,$F7,$39,$A7,$11,$A7
	.dc.b	$FC,$73,$1A,$F8,$7E,$9F,$37,$27,$94,$77,$30,$F0,$8C,$3F,$E2,$FE
	.dc.b	$DF,$E2,$EF,$F8,$C7,$E8,$DF,$47,$FE,$FA,$6C,$6F,$C9,$FF,$7E,$D7
	.dc.b	$08,$FF,$25,$CE,$9B,$5A,$A9,$1D,$63,$A1,$87,$FC,$A7,$5C,$F7,$F9
	.dc.b	$90,$BE,$91,$86,$ED,$FD,$2A,$B8,$37,$0E,$B7,$DC,$F4,$10,$F2,$99
	.dc.b	$3C,$2F,$8E,$FB,$61,$CA,$72,$71,$8C,$BC,$5F,$3D,$F8,$A3,$F0,$6D
	.dc.b	$7D,$53,$CE,$77,$5D,$EF,$FB,$1E,$1E,$CB,$FC,$5B,$5F,$E4,$53,$B7
	.dc.b	$37,$CE,$BE,$04,$79,$5E,$71,$98,$F1,$CF,$17,$C6,$7A,$67,$8B,$EE
	.dc.b	$3E,$29,$2F,$CE,$F5,$3C,$27,$13,$A9,$93,$A9,$C6,$3A,$C7,$28,$C1
	.dc.b	$C8,$F0,$39,$3E,$6B,$A3,$E0,$9D,$6F,$E2,$1F,$17,$0F,$9C,$7C,$F7
	.dc.b	$38,$7B,$E3,$0F,$C3,$9E,$67,$82,$FE,$3B,$EB,$33,$7E,$38,$CC,$78
	.dc.b	$EB,$BE,$1C,$F0,$F1,$3F,$9B,$9F,$F0,$A4,$FE,$95,$B1,$FE,$FA,$8C
	.dc.b	$47,$D0,$32,$7F,$67,$3C,$A6,$0F,$ED,$4F,$81,$F8,$5D,$DE,$D9,$FB
	.dc.b	$95,$FD,$6E,$B5,$5E,$D3,$F1,$75,$B7,$89,$F5,$E3,$78,$7E,$B6,$75
	.dc.b	$3C,$4F,$D9,$38,$8F,$89,$1D,$0E,$57,$E2,$9D,$61,$F8,$03,$F3,$AE
	.dc.b	$27,$F0,$E7,$F0,$87,$0F,$D0,$8D,$BD,$0F,$39,$E3,$BF,$3C,$6A,$31
	.dc.b	$5E,$17,$81,$E9,$C3,$E2,$47,$66,$C6,$B0,$F8,$0D,$AF,$43,$ED,$3A
	.dc.b	$9E,$28,$F8,$2F,$9C,$F5,$BC,$9D,$9E,$1C,$3D,$B8,$7E,$55,$B8,$7F
	.dc.b	$2D,$6C,$FF,$31,$7F,$98,$75,$C8,$FE,$68,$FF,$33,$EC,$C4,$7F,$30
	.dc.b	$8B,$87,$F2,$59,$FE,$71,$D3,$FC,$E7,$D9,$97,$FC,$F6,$32,$FF,$3F
	.dc.b	$8C,$BF,$CE,$A3,$F9,$EF,$57,$F9,$DA,$FF,$99,$FD,$13,$E8,$F1,$B7
	.dc.b	$F9,$EA,$31,$87,$F4,$BB,$E2,$FE,$6F,$6F,$F0,$2B,$F9,$A8,$FD,$11
	.dc.b	$1E,$04,$18,$78,$3B,$3B,$C7,$F6,$EB,$75,$E2,$10,$EB,$C3,$63,$FB
	.dc.b	$AD,$FA,$AB,$26,$AA,$4E,$B2,$F6,$4F,$2E,$89,$39,$3C,$4F,$6F,$A3
	.dc.b	$CC,$7A,$2B,$E2,$4E,$30,$F9,$35,$E7,$FA,$27,$C1,$FA,$7E,$89,$F4
	.dc.b	$74,$D4,$4C,$BF,$82,$57,$06,$C7,$F0,$75,$19,$7F,$CE,$0D,$C3,$FB
	.dc.b	$5B,$72,$FF,$58,$B6,$7F,$C3,$38,$7F,$83,$DF,$51,$8E,$BF,$37,$5C
	.dc.b	$8F,$84,$25,$3E,$B3,$51,$C8,$D4,$E5,$E3,$F4,$3F,$E7,$12,$74,$FD
	.dc.b	$B7,$FC,$DB,$1C,$5C,$BC,$23,$5E,$3F,$E0,$7E,$7D,$F4,$5C,$7E,$8E
	.dc.b	$5F,$9A,$72,$D9,$F9,$11,$88,$FF,$4B,$35,$FF,$66,$35,$FF,$6B,$BB
	.dc.b	$DB,$FC,$2A,$DD,$B3,$5E,$9E,$DF,$57,$5F,$AF,$B2,$63,$F0,$23,$2F
	.dc.b	$C2,$8C,$BE,$F8,$FC,$E7,$D1,$FF,$99,$72,$FF,$11,$B7,$1F,$F7,$4B
	.dc.b	$6B,$FC,$A2,$FE,$BA,$77,$D5,$C8,$93,$F6,$27,$55,$EC,$F5,$BB,$FD
	.dc.b	$1C,$AF,$91,$1F,$34,$F8,$E4,$7F,$AB,$6A,$8C,$47,$DF,$8C,$3C,$8F
	.dc.b	$07,$57,$ED,$FE,$8D,$EB,$52,$4F,$E3,$3E,$8F,$F7,$0F,$E2,$7E,$8F
	.dc.b	$FD,$6D,$18,$DB,$FC,$8A,$DD,$3F,$CE,$34,$AF,$F5,$E1,$39,$7F,$83
	.dc.b	$3F,$EB,$FF,$A3,$7D,$1F,$F0,$9F,$A3,$71,$D2,$61,$F9,$EF,$A3,$7D
	.dc.b	$0F,$F0,$F6,$C7,$E8,$AB,$8B,$6B,$55,$23,$93,$5F,$D6,$FD,$1B,$E8
	.dc.b	$DE,$FF,$F1,$BF,$86,$7C,$FD,$25,$F8,$FC,$8A,$FE,$B6,$76,$FE,$D1
	.dc.b	$AF,$EC,$74,$FF,$73,$F1,$FF,$9B,$27,$2D,$BE,$6F,$F9,$C2,$77,$C5
	.dc.b	$F3,$3E,$8F,$DF,$87,$EE,$DB,$5F,$E7,$66,$CF,$D2,$F9,$D7,$D1,$FF
	.dc.b	$35,$9C,$DF,$FA,$99,$A9,$B3,$C2,$3F,$86,$C6,$89,$BF,$F3,$F9,$D4
	.dc.b	$FF,$0D,$F1,$9F,$47,$FD,$0D,$B1,$FC,$59,$AF,$F1,$3F,$A1,$7C,$AE
	.dc.b	$EA,$4E,$DF,$E9,$93,$A8,$C1,$0F,$04,$6A,$7F,$A9,$7D,$1F,$F2,$B9
	.dc.b	$34,$C5,$3C,$5D,$F9,$B2,$BF,$95,$D6,$A7,$11,$FC,$A6,$39,$4F,$92
	.dc.b	$7F,$B0,$7D,$17,$DA,$DF,$F4,$2D,$BF,$F6,$9B,$56,$0F,$A3,$E3,$C7
	.dc.b	$27,$F9,$DD,$6E,$B5,$18,$9D,$3F,$CE,$E7,$55,$E2,$E6,$93,$C9,$F9
	.dc.b	$3D,$38,$8E,$11,$79,$3F,$D7,$B8,$FE,$A2,$7F,$A3,$7D,$1B,$8B,$FB
	.dc.b	$E3,$63,$FA,$7B,$5F,$E9,$9F,$47,$FE,$02,$BF,$F7,$8F,$D1,$BE,$8B
	.dc.b	$F8,$1D,$EE,$7F,$D9,$6D,$DF,$38,$FF,$82,$70,$E5,$8B,$1E,$4F,$FB
	.dc.b	$F2,$33,$18,$85,$F8,$E3,$0D,$C7,$F8,$9D,$70,$6F,$8D,$FE,$1C,$BE
	.dc.b	$D1,$E2,$7D,$2D,$72,$78,$63,$85,$F8,$6B,$A4,$8F,$1E,$1F,$C8,$A7
	.dc.b	$C6,$DB,$F0,$E5,$DA,$BA,$35,$E2,$5C,$7D,$4E,$8F,$18,$C3,$C2,$F0
	.dc.b	$33,$1C,$1D,$63,$E8,$BF,$0C,$6B,$0F,$23,$E9,$AE,$17,$D0,$EA,$F3
	.dc.b	$33,$7C,$E7,$C8,$7F,$40,$3B,$E1,$FE,$A9,$AF,$B7,$19,$6B,$F5,$3D
	.dc.b	$67,$C7,$6D,$B1,$CC,$F7,$38,$79,$35,$D7,$D1,$FF,$8C,$53,$48,$FF
	.dc.b	$8D,$5F,$9D,$F2,$CF,$A0,$6F,$F4,$6B,$ED,$56,$A7,$31,$F4,$0C,$9F
	.dc.b	$4A,$B9,$57,$18,$F0,$99,$9E,$3B,$FE,$5E,$38,$BF,$99,$AE,$31,$D4
	.dc.b	$E3,$19,$9E,$3B,$FB,$A7,$A5,$F7,$C3,$CC,$D8,$C9,$CE,$32,$D3,$37
	.dc.b	$E3,$7B,$D7,$F7,$8C,$3C,$73,$9A,$E2,$DA,$E3,$5B,$C3,$E2,$E9,$F1
	.dc.b	$A5,$87,$B2,$31,$1F,$02,$B8,$63,$FE,$58,$AD,$3E,$1E,$2E,$6E,$AB
	.dc.b	$FC,$B1,$3F,$E0,$48,$E1,$CF,$A9,$FF,$4A,$FD,$1F,$F4,$A9,$C9,$FE
	.dc.b	$85,$C5,$19,$9E,$75,$C6,$73,$D5,$CE,$7D,$33,$9A,$E1,$5E,$DC,$2E
	.dc.b	$1F,$BF,$F9,$57,$D1,$BE,$8F,$95,$6B,$25,$6A,$73,$9F,$A6,$31,$87
	.dc.b	$C3,$8C,$61,$CE,$30,$7F,$A7,$E7,$77,$F1,$2B,$77,$E9,$F1,$7F,$C3
	.dc.b	$37,$FE,$A5,$1F,$43,$F3,$AF,$F8,$B4,$E9,$FF,$46,$CE,$B8,$39,$75
	.dc.b	$7C,$D8,$C1,$FE,$8E,$AE,$1A,$FE,$92,$33,$7F,$E8,$D9,$CC,$7F,$23
	.dc.b	$33,$3F,$85,$8C,$CF,$EC,$30,$F2,$2F,$A3,$7A,$FE,$19,$B8,$7F,$A7
	.dc.b	$A6,$D7,$A2,$6E,$1F,$91,$9B,$5E,$19,$B5,$FC,$D6,$6C,$FC,$4B,$FC
	.dc.b	$04,$62,$B7,$5E,$39,$F1,$1F,$85,$AF,$F6,$E1,$F3,$7E,$8F,$FC,$A6
	.dc.b	$FF,$AC,$8C,$CF,$CC,$8C,$BF,$EB,$18,$CE,$1F,$26,$72,$FE,$A6,$32
	.dc.b	$7C,$98,$CB,$FE,$6C,$33,$3F,$C6,$23,$2F,$CF,$E7,$FB,$11,$88,$7C
	.dc.b	$CF,$E4,$EB,$F0,$7F,$29,$F9,$F8,$C6,$48,$C7,$07,$D1,$6D,$7E,$2E
	.dc.b	$74,$FF,$B4,$2B,$8C,$EA,$FE,$A3,$FA,$BF,$A2,$7C,$35,$31,$E6,$DC
	.dc.b	$3F,$DB,$3F,$46,$E1,$FD,$93,$63,$FD,$BB,$7C,$9D,$CF,$ED,$FE,$89
	.dc.b	$F4,$74,$E3,$16,$7F,$DC,$BC,$EF,$A3,$FD,$F7,$A7,$FE,$64,$9C,$C6
	.dc.b	$3D,$1F,$E0,$5F,$01,$FE,$A4,$9F,$F7,$6E,$4F,$93,$CF,$B9,$E1,$7F
	.dc.b	$84,$7D,$1A,$FF,$85,$7E,$05,$F9,$C8,$C6,$1F,$EF,$68,$FC,$FF,$D1
	.dc.b	$BE,$93,$F8,$83,$67,$FC,$15,$D9,$FD,$43,$C3,$FF,$3A,$72,$EA,$E5
	.dc.b	$7F,$03,$FA,$47,$CA,$63,$A9,$8F,$F5,$8E,$45,$6A,$FA,$9F,$EB,$6C
	.dc.b	$FC,$17,$C2,$3F,$8E,$4E,$E3,$07,$F8,$76,$77,$7D,$67,$4F,$F6,$6E
	.dc.b	$BF,$EF,$7F,$0B,$FC,$D5,$FE,$FB,$F4,$97,$CE,$FF,$6B,$65,$E2,$CB
	.dc.b	$F2,$B0,$9D,$5E,$DB,$5B,$BE,$6D,$AF,$AD,$7E,$69,$FF,$05,$F8,$BD
	.dc.b	$F9,$7C,$0D,$BE,$0D,$AE,$17,$FC,$51,$7D,$13,$E8,$BF,$29,$1F,$E1
	.dc.b	$38,$C7,$67,$F8,$56,$37,$F2,$BF,$CF,$9D,$57,$07,$FC,$49,$F4,$5F
	.dc.b	$AD,$78,$35,$FF,$0F,$35,$DF,$87,$D5,$43,$CE,$C3,$63,$9C,$60,$E1
	.dc.b	$2F,$87,$FF,$79,$F1,$7F,$75,$87,$C7,$E3,$7F,$F7,$BC,$63,$0F,$A8
	.dc.b	$7F,$03,$BD,$3F,$E5,$D8,$E8,$E9,$E0,$FE,$18,$FE,$15,$5C,$1E,$46
	.dc.b	$FD,$DC,$4B,$E6,$29,$88,$93,$D6,$32,$F1,$3C,$BF,$42,$FA,$1F,$8C
	.dc.b	$D8,$FA,$8D,$7F,$C9,$4D,$7F,$CA,$CE,$73,$FF,$0C,$37,$3F,$F3,$D4
	.dc.b	$A3,$0F,$C1,$F5,$7D,$91,$88,$D5,$E0,$F6,$CF,$C1,$FA,$1F,$E0,$93
	.dc.b	$F3,$1F,$E4,$6D,$F3,$35,$E2,$48,$C1,$0F,$99,$3C,$9B,$ED,$F5,$23
	.dc.b	$D7,$D1,$AE,$8F,$04,$78,$E3,$EA,$7D,$07,$F1,$B9,$FB,$ED,$7E,$D3
	.dc.b	$75,$FF,$1E,$5F,$E6,$DF,$F1,$F4,$67,$5F,$8E,$DA,$FF,$51,$35,$57
	.dc.b	$19,$DF,$91,$E2,$24,$FC,$5E,$8F,$C4,$7C,$FB,$E8,$29,$27,$F2,$B1
	.dc.b	$FE,$EB,$34,$E2,$34,$FE,$4E,$35,$F0,$FD,$3E,$62,$4F,$28,$EE,$61
	.dc.b	$E1,$18,$7F,$47,$D9,$FC,$B9,$BE,$3F,$E7,$63,$11,$FA,$1F,$9F,$FD
	.dc.b	$2B,$FA,$38,$CE,$FF,$E6,$56,$EB,$F8,$E6,$CF,$FB,$26,$B8,$D6,$EF
	.dc.b	$B1,$27,$F6,$B3,$AB,$E5,$39,$9F,$03,$FB,$19,$D4,$7F,$B5,$6F,$BA
	.dc.b	$69,$88,$DC,$66,$7A,$5F,$8E,$3C,$3D,$78,$C3,$C3,$1F,$10,$C9,$4F
	.dc.b	$19,$67,$FD,$0D,$7D,$B0,$F0,$E1,$CE,$BE,$16,$15,$F1,$1C,$70,$FD
	.dc.b	$AD,$F1,$AD,$5F,$A6,$EC,$71,$9D,$5F,$13,$E1,$47,$C5,$9D,$C6,$44
	.dc.b	$F0,$7D,$24,$C3,$D3,$3C,$23,$18,$79,$23,$13,$F1,$E7,$F0,$91,$89
	.dc.b	$F8,$F1,$83,$C1,$87,$C2,$7E,$49,$D0,$F3,$9E,$5B,$F2,$5F,$52,$57
	.dc.b	$86,$79,$E1,$F0,$70,$F0,$47,$23,$C0,$75,$3E,$69,$6F,$AD,$79,$AF
	.dc.b	$AC,$FD,$43,$A9,$E3,$AE,$AF,$28,$F8,$AE,$8E,$91,$AC,$3E,$B4,$66
	.dc.b	$7A,$BF,$42,$75,$87,$D5,$8D,$5F,$8C,$F7,$3E,$9B,$F9,$0F,$B2,$F9
	.dc.b	$C6,$1F,$1C,$6E,$BE,$23,$D0,$F8,$91,$A3,$8C,$6A,$38,$1E,$F3,$81
	.dc.b	$EF,$AF,$1C,$F0,$8D,$1C,$0F,$82,$70,$AF,$82,$E6,$3E,$0E,$1F,$42
	.dc.b	$3C,$75,$EF,$7E,$81,$B7,$0F,$C7,$3A,$5F,$C7,$74,$72,$38,$D7,$95
	.dc.b	$F1,$9A,$3D,$46,$A7,$AD,$FC,$97,$DA,$7E,$C3,$A7,$F8,$61,$27,$F8
	.dc.b	$9F,$4D,$7F,$91,$45,$BF,$DF,$CF,$F1,$3E,$7F,$E2,$9E,$C4,$BF,$E3
	.dc.b	$71,$97,$F8,$EC,$65,$FE,$59,$1F,$CB,$FA,$24,$7F,$34,$7F,$C1,$DF
	.dc.b	$3D,$FD,$97,$0F,$E5,$D1,$A9,$FE,$64,$EB,$A7,$13,$A9,$E4,$7F,$88
	.dc.b	$2B,$96,$7F,$E0,$25,$D3,$DB,$A6,$56,$E3,$C2,$61,$FF,$11,$4A,$73
	.dc.b	$1D,$79,$BC,$E4,$7C,$1F,$3D,$5E,$31,$71,$63,$E3,$AF,$F1,$67,$CE
	.dc.b	$FF,$90,$6F,$FC,$60,$DA,$DC,$04,$66,$BF,$C3,$BF,$3B,$FD,$A2,$78
	.dc.b	$F8,$3D,$08,$FF,$90,$A3,$57,$D2,$39,$CE,$9E,$26,$FA,$BF,$C8,$14
	.dc.b	$E9,$BD,$32,$BA,$46,$8C,$3F,$E4,$99,$4E,$63,$FC,$A7,$F3,$BF,$D5
	.dc.b	$09,$D5,$FE,$4D,$7F,$CA,$39,$79,$5C,$CF,$09,$D7,$85,$FE,$53,$F9
	.dc.b	$97,$CE,$7F,$9B,$5B,$5F,$E7,$B6,$BF,$E4,$23,$E5,$38,$70,$EB,$0F
	.dc.b	$88,$DD,$BF,$C9,$2D,$9F,$A5,$8B,$AB,$C5,$84,$9F,$F3,$7D,$6D,$56
	.dc.b	$DF,$83,$D1,$EF,$E9,$91,$82,$1F,$E7,$19,$46,$A3,$13,$FE,$73,$F9
	.dc.b	$CF,$CE,$C7,$E1,$5A,$9A,$7F,$95,$9F,$F3,$E3,$8E,$9E,$47,$3E,$AF
	.dc.b	$99,$E7,$7C,$B2,$57,$CB,$6B,$FE,$85,$C9,$1C,$23,$FD,$1F,$7F,$AE
	.dc.b	$7E,$1C,$65,$E0,$FA,$7A,$FF,$D1,$9F,$35,$FA,$D9,$FF,$4B,$FC,$DB
	.dc.b	$CE,$FF,$48,$36,$3F,$D3,$73,$B7,$FD,$2D,$F3,$7F,$E1,$E8,$C7,$1F
	.dc.b	$E7,$5B,$C7,$FE,$82,$9F,$F1,$1F,$CD,$BE,$6F,$FD,$39,$1F,$E2,$63
	.dc.b	$2F,$FA,$7F,$E6,$DC,$54,$99,$7F,$AB,$3E,$69,$F3,$3F,$EA,$56,$C7
	.dc.b	$FA,$C2,$B8,$B6,$B5,$52,$39,$35,$FF,$7A,$FC,$13,$E6,$DE,$B7,$FA
	.dc.b	$C7,$DD,$7C,$C5,$25,$F8,$FC,$7A,$FF,$AE,$67,$6F,$FB,$09,$AF,$FA
	.dc.b	$F7,$4F,$F3,$D3,$FC,$36,$33,$C5,$FC,$15,$BC,$3F,$EA,$A6,$CF,$F1
	.dc.b	$3A,$E2,$BE,$6F,$E0,$BF,$EC,$96,$DF,$F0,$F6,$DF,$FA,$D3,$E5,$9F
	.dc.b	$0B,$FE,$CD,$9C,$DF,$F1,$F6,$BD,$77,$FF,$69,$57,$08,$FF,$6E,$CE
	.dc.b	$89,$3A,$AF,$F6,$C7,$8C,$F9,$BA,$4B,$FF,$72,$FC,$C3,$E3,$71,$1F
	.dc.b	$EE,$1A,$E0,$FF,$BA,$9A,$FF,$B9,$67,$83,$9A,$E3,$A7,$F2,$DA,$DB
	.dc.b	$FC,$BE,$B7,$5F,$EE,$DF,$9B,$FE,$74,$FF,$3D,$6A,$6C,$7F,$09,$8E
	.dc.b	$1E,$67,$FB,$CE,$76,$6B,$C7,$C4,$49,$FF,$80,$33,$97,$FA,$55,$FF
	.dc.b	$41,$F9,$BF,$B7,$06,$98,$C1,$F3,$7F,$F0,$06,$33,$CB,$4F,$F7,$6D
	.dc.b	$FB,$67,$17,$DB,$8B,$FD,$E3,$7C,$23,$E7,$CF,$3E,$69,$1B,$3F,$AF
	.dc.b	$4F,$E8,$1F,$AB,$D3,$A5,$87,$A6,$B0,$FC,$D9,$F1,$E1,$FF,$00,$3E
	.dc.b	$6D,$CB,$8C,$3F,$B5,$46,$1F,$00,$70,$F6,$9F,$EE,$0B,$E6,$FF,$C1
	.dc.b	$38,$7F,$73,$F9,$A7,$CD,$77,$C7,$B5,$F3,$6E,$3D,$2F,$FB,$86,$1C
	.dc.b	$A9,$87,$27,$FB,$F4,$66,$31,$1C,$CF,$06,$1D,$DA,$70,$F1,$7E,$2B
	.dc.b	$D2,$3D,$35,$CD,$E1,$19,$3E,$1E,$3E,$EB,$EF,$5B,$9E,$98,$76,$7C
	.dc.b	$47,$12,$CF,$0C,$3C,$2D,$9E,$18,$FC,$C8,$F5,$B7,$1F,$2D,$70,$6D
	.dc.b	$70,$C7,$C9,$1E,$8B,$E9,$87,$87,$1F,$33,$E2,$9D,$1C,$9D,$61,$D5
	.dc.b	$F7,$9D,$4F,$64,$F2,$3E,$A3,$FE,$07,$6B,$CC,$C4,$7C,$E7,$5E,$AB
	.dc.b	$E1,$7F,$C1,$D1,$F6,$A3,$0E,$2B,$FC,$95,$3A,$8F,$C6,$4E,$9C,$7C
	.dc.b	$DE,$5F,$F2,$03,$53,$48,$FD,$3F,$8D,$7C,$C7,$F9,$23,$3F,$C4,$CF
	.dc.b	$23,$51,$9C,$FC,$F7,$F7,$6F,$8E,$FF,$7E,$F8,$61,$F9,$29,$CC,$7E
	.dc.b	$04,$CD,$FE,$9A,$33,$3E,$09,$F3,$9E,$0A,$E8,$E2,$3A,$61,$E3,$8F
	.dc.b	$F1,$B6,$1E,$38,$C4,$7E,$CA,$31,$5C,$9B,$1E,$33,$A6,$1E,$3A,$E3
	.dc.b	$1E,$63,$FC,$D5,$87,$9A,$7E,$0C,$79,$4E,$55,$D3,$0F,$2E,$1F,$C1
	.dc.b	$AF,$E0,$ED,$FE,$69,$BE,$82,$33,$3E,$13,$87,$07,$F9,$EA,$BF,$CE
	.dc.b	$13,$AA,$F1,$CE,$6B,$E5,$73,$D9,$FE,$3F,$F3,$7F,$D5,$67,$C5,$39
	.dc.b	$C1,$39,$AF,$E3,$07,$BA,$75,$19,$9F,$0C,$63,$B3,$B5,$7A,$5F,$07
	.dc.b	$C7,$3E,$6D,$F3,$7A,$D3,$96,$7F,$E9,$3A,$D4,$E6,$F8,$DF,$0A,$F9
	.dc.b	$11,$89,$D4,$7F,$3F,$AE,$2E,$8C,$57,$1C,$3D,$63,$C5,$FF,$11,$CF
	.dc.b	$F4,$71,$F3,$3F,$2E,$75,$D3,$F2,$67,$3C,$7E,$EB,$E3,$5F,$1A,$72
	.dc.b	$7F,$42,$8C,$ED,$FE,$B0,$8C,$CF,$8A,$73,$1F,$EB,$33,$33,$FE,$AB
	.dc.b	$8C,$CF,$FA,$FB,$0F,$27,$CD,$FF,$9E,$AF,$D1,$16,$7D,$B3,$6F,$C1
	.dc.b	$36,$BF,$A8,$CD,$AE,$9E,$B7,$D8,$8C,$4F,$C2,$8C,$1F,$06,$B8,$52
	.dc.b	$B7,$18,$AF,$F6,$E1,$F2,$FE,$6F,$FC,$B6,$FF,$64,$33,$5F,$1E,$32
	.dc.b	$FD,$48,$CE,$1F,$1E,$72,$FF,$B3,$63,$27,$C7,$8C,$BF,$DB,$0C,$CF
	.dc.b	$FB,$92,$32,$FC,$DE,$7F,$B5,$18,$87,$C8,$7E,$49,$F9,$DF,$1B,$F9
	.dc.b	$A8,$C5,$FE,$12,$31,$C5,$F0,$DB,$1F,$49,$B5,$F8,$3C,$91,$8C,$BE
	.dc.b	$9B,$FA,$6F,$9A,$7B,$34,$C7,$9B,$71,$FF,$05,$FC,$DF,$E5,$DB,$C3
	.dc.b	$9C,$FF,$37,$39,$7F,$3B,$F3,$2F,$9B,$FD,$9B,$6F,$89,$FE,$16,$71
	.dc.b	$3F,$A4,$E7,$7C,$DF,$EE,$BD,$3F,$F3,$5D,$FF,$87,$27,$2F,$F8,$7A
	.dc.b	$33,$5F,$AC,$F7,$BF,$D5,$D3,$D7,$E3,$BF,$81,$C6,$27,$CB,$1F,$C1
	.dc.b	$BE,$6D,$3F,$C2,$3E,$6F,$F6,$11,$8B,$FF,$15,$C7,$F1,$DE,$6F,$EC
	.dc.b	$9E,$FF,$F1,$B7,$FB,$37,$CC,$5F,$23,$FD,$93,$7F,$E4,$19,$CB,$63
	.dc.b	$39,$FF,$1C,$9C,$B6,$33,$CE,$A4,$FE,$1A,$57,$DF,$0F,$03,$6B,$B3
	.dc.b	$E3,$F8,$0F,$27,$81,$67,$C4,$DC,$3C,$0D,$C7,$C1,$1D,$E3,$F9,$5F
	.dc.b	$D2,$3E,$93,$FA,$34,$F9,$23,$17,$FE,$01,$6C,$7F,$82,$1A,$FF,$94
	.dc.b	$07,$CD,$3E,$6B,$F9,$14,$7F,$83,$A3,$1D,$3F,$E0,$B5,$F2,$4E,$6C
	.dc.b	$5F,$0B,$E9,$DC,$F8,$BF,$F2,$BA,$F8,$BA,$62,$24,$FF,$99,$A7,$4D
	.dc.b	$9C,$BF,$CD,$A3,$A3,$92,$19,$9E,$FE,$D7,$83,$E4,$A4,$F5,$8C,$BD
	.dc.b	$E3,$2F,$58,$FF,$36,$FC,$CB,$E6,$7D,$3F,$49,$2B,$FC,$F7,$39,$6C
	.dc.b	$FF,$8A,$E3,$11,$FE,$36,$6B,$F3,$DA,$FF,$A0,$3E,$67,$F3,$EF,$FA
	.dc.b	$19,$FE,$02,$DF,$17,$FC,$FF,$1A,$8F,$EA,$ED,$E6,$FC,$2C,$23,$8B
	.dc.b	$A9,$E3,$DD,$23,$A1,$0E,$6F,$6F,$62,$47,$8D,$F0,$F9,$78,$8F,$0C
	.dc.b	$5C,$3C,$38,$2B,$FD,$2B,$F3,$4F,$0B,$D3,$F4,$93,$1F,$F5,$1C,$E5
	.dc.b	$B3,$FE,$53,$8C,$47,$F9,$79,$AF,$E5,$DA,$FF,$AA,$3E,$6D,$F2,$DF
	.dc.b	$EA,$46,$C7,$FA,$C1,$AF,$F7,$E7,$FD,$A6,$63,$2F,$92,$DE,$CE,$95
	.dc.b	$CF,$E0,$7A,$32,$FA,$39,$7B,$7C,$27,$87,$E5,$64,$60,$87,$85,$CB
	.dc.b	$E3,$8F,$F5,$DF,$63,$E6,$BF,$D7,$BF,$36,$FA,$4A,$6A,$25,$FF,$B3
	.dc.b	$27,$4D,$9C,$BF,$E1,$C8,$FF,$7E,$6B,$F8,$66,$E9,$FE,$D0,$6C,$FF
	.dc.b	$1C,$5D,$5E,$2E,$A9,$3D,$DF,$2C,$62,$3B,$45,$BE,$D3,$6F,$B4,$DA
	.dc.b	$ED,$35,$5F,$89,$B3,$D9,$F0,$10,$EE,$7C,$19,$E5,$3C,$48,$E2,$72
	.dc.b	$D9,$E5,$7E,$66,$DF,$99,$A7,$98,$A7,$90,$B7,$F1,$9B,$7D,$9A,$6A
	.dc.b	$FC,$0F,$A0,$FA,$8E,$2F,$EA,$4F,$BE,$75,$18,$79,$1D,$0C,$DF,$D9
	.dc.b	$AE,$4F,$A6,$3A,$9B,$9E,$4E,$0E,$27,$DF,$AF,$43,$83,$DC,$5B,$F7
	.dc.b	$36,$37,$3D,$09,$5E,$17,$E6,$4E,$CF,$49,$4E,$97,$E8,$7C,$47,$DB
	.dc.b	$3D,$13,$CC,$F7,$38,$36,$7A,$E7,$C2,$E1,$AF,$E3,$27,$B3,$EC,$8F
	.dc.b	$29,$AB,$FC,$45,$FD,$18,$C3,$EE,$BF,$15,$75,$6B,$D4,$F2,$1D,$5F
	.dc.b	$3B,$EF,$9E,$D7,$E2,$BF,$CA,$CF,$06,$BB,$AE,$B1,$F4,$2B,$C7,$3F
	.dc.b	$29,$AF,$CA,$39,$5F,$A6,$B8,$BE,$98,$F4,$10,$F3,$1F,$9D,$3C,$06
	.dc.b	$6B,$C1,$3C,$A3,$C0,$D7,$C0,$70,$39,$9F,$0C,$C4,$7B,$2B,$CE,$7C
	.dc.b	$28,$F3,$9F,$7C,$FE,$2D,$EF,$7F,$20,$F6,$24,$7F,$1D,$8F,$E4,$66
	.dc.b	$1F,$E3,$91,$87,$FC,$67,$3F,$D8,$B4,$FE,$C2,$FF,$64,$F9,$C7,$D2
	.dc.b	$7F,$8E,$2B,$FD,$7F,$18,$78,$E5,$9B,$E2,$DE,$BE,$18,$73,$C2,$CF
	.dc.b	$3C,$39,$61,$FE,$CA,$C3,$8B,$FE,$46,$87,$57,$08,$C7,$4D,$53,$FD
	.dc.b	$A5,$F1,$9C,$67,$51,$CA,$32,$70,$9C,$9A,$99,$1A,$9E,$8F,$C0,$9C
	.dc.b	$BF,$0B,$C5,$EA,$47,$87,$CE,$D7,$47,$38,$F1,$C7,$A7,$E7,$5F,$03
	.dc.b	$EB,$F9,$D7,$D2,$53,$51,$30,$FF,$28,$1F,$E6,$AA,$E2,$E0,$38,$76
	.dc.b	$D7,$FC,$B1,$4E,$8F,$07,$4C,$9F,$8B,$39,$6C,$E5,$FF,$7F,$C2,$7D
	.dc.b	$13,$27,$8C,$6F,$87,$FC,$5B,$5C,$E5,$1F,$E3,$0F,$9D,$FF,$29,$C6
	.dc.b	$38,$7F,$1D,$18,$9D,$78,$DF,$7E,$3F,$CB,$B5,$B3,$FD,$16,$EE,$B8
	.dc.b	$CE,$CE,$1E,$47,$E1,$FE,$75,$F3,$DC,$FE,$92,$63,$F8,$49,$CB,$67
	.dc.b	$E6,$28,$CB,$F4,$9A,$FF,$A0,$DA,$FA,$BA,$BD,$1D,$52,$3F,$23,$9A
	.dc.b	$7D,$86,$63,$D8,$21,$3A,$8E,$33,$BD,$3E,$54,$E6,$78,$FC,$FF,$F9
	.dc.b	$77,$5F,$DC,$36,$3F,$D6,$2D,$4F,$EC,$1B,$E4,$7E,$B6,$34,$4E,$CF
	.dc.b	$95,$EA,$E7,$E9,$25,$FC,$98,$CB,$F9,$3D,$51,$F2,$9B,$19,$8F,$9B
	.dc.b	$7F,$EA,$1C,$FF,$CE,$6B,$A3,$C4,$78,$FA,$65,$FE,$EA,$74,$D9,$CB
	.dc.b	$F0,$A7,$8B,$93,$83,$FC,$27,$E7,$FF,$7B,$7F,$E8,$2F,$A0,$7A,$FF
	.dc.b	$C3,$5B,$7F,$C2,$2B,$7B,$FF,$23,$F9,$FF,$C8,$4F,$FB,$72,$72,$DE
	.dc.b	$EF,$E2,$D3,$A8,$7D,$03,$E8,$33,$D5,$27,$F8,$AC,$E5,$FE,$2F,$F4
	.dc.b	$1F,$87,$6F,$3F,$F2,$18,$C4,$7E,$23,$E7,$DF,$3D,$FC,$95,$B1,$C0
	.dc.b	$86,$E3,$44,$7F,$92,$C7,$07,$F5,$DF,$06,$FA,$07,$BF,$FC,$C7,$E2
	.dc.b	$5E,$95,$25,$F2,$E6,$AF,$E7,$27,$6F,$F3,$26,$BE,$98,$DF,$37,$FA
	.dc.b	$D2,$72,$DA,$F6,$4F,$F4,$1F,$A0,$F3,$4F,$FA,$E9,$BC,$BF,$D5,$DB
	.dc.b	$80,$7C,$CB,$E8,$39,$6F,$27,$FA,$2D,$70,$77,$39,$7F,$6B,$19,$35
	.dc.b	$5C,$E3,$FA,$6F,$90,$FA,$0F,$EA,$CD,$8C,$C7,$F5,$9F,$9F,$7C,$8E
	.dc.b	$A2,$4E,$8F,$EA,$F3,$A7,$07,$F0,$6A,$E8,$D9,$E9,$5F,$D9,$7E,$83
	.dc.b	$FA,$16,$BF,$ED,$E6,$EB,$FD,$ED,$B7,$F1,$71,$FF,$45,$46,$20,$D3
	.dc.b	$11,$AF,$3B,$B1,$27,$CF,$EA,$FF,$75,$F9,$FF,$B3,$87,$F8,$25,$B3
	.dc.b	$F4,$DA,$AD,$F4,$1C,$78,$E4,$FF,$78,$AD,$D6,$A3,$13,$A7,$F9,$4C
	.dc.b	$EA,$BC,$5C,$B2,$79,$3F,$47,$AB,$11,$C2,$2E,$AC,$BF,$C4,$19,$7E
	.dc.b	$BA,$31,$5F,$8B,$FA,$0F,$C5,$BF,$E0,$C6,$BF,$CE,$9B,$1C,$21,$F4
	.dc.b	$1F,$E1,$C8,$FC,$A7,$CF,$3E,$83,$FC,$5D,$7C,$31,$BC,$BF,$E1,$F8
	.dc.b	$C6,$3F,$9B,$C7,$92,$C3,$93,$9F,$63,$FB,$7B,$9C,$79,$BE,$06,$B9
	.dc.b	$38,$65,$96,$99,$18,$7B,$AF,$BC,$63,$1F,$F2,$34,$63,$1F,$2B,$71
	.dc.b	$F1,$BC,$2F,$96,$3E,$28,$D6,$3E,$1C,$7E,$3D,$F0,$AD,$12,$3C,$B5
	.dc.b	$C1,$F3,$5F,$47,$EB,$63,$E4,$AF,$F2,$D6,$3E,$46,$E1,$D6,$B6,$78
	.dc.b	$6F,$93,$E2,$C7,$C9,$7C,$63,$67,$83,$0E,$CF,$93,$0E,$B5,$FE,$35
	.dc.b	$AF,$AF,$3F,$E6,$B3,$A1,$FE,$3C,$72,$F9,$BC,$47,$C7,$F1,$6C,$6D
	.dc.b	$AF,$39,$46,$23,$FC,$8F,$F4,$1F,$E1,$A4,$D2,$3F,$E1,$E7,$E8,$7C
	.dc.b	$FB,$E7,$7F,$3F,$B7,$DE,$AF,$E5,$55,$A9,$CE,$7D,$A7,$F0,$D7,$C7
	.dc.b	$7F,$B5,$1C,$0C,$D7,$07,$E1,$5F,$39,$DC,$79,$8D,$B8,$8D,$47,$E4
	.dc.b	$CE,$98,$7E,$4A,$FC,$11,$93,$94,$66,$FC,$11,$FE,$9B,$AE,$73,$ED
	.dc.b	$8D,$E1,$FE,$9F,$BE,$73,$9A,$DC,$FB,$E3,$A6,$1E,$5B,$F3,$93,$4F
	.dc.b	$E0,$18,$79,$1F,$E0,$51,$E8,$BE,$D8,$FF,$0C,$9D,$B6,$FC,$5C,$56
	.dc.b	$FF,$AD,$61,$CE,$72,$7F,$A7,$1F,$F4,$5B,$FE,$A2,$B7,$D0,$7F,$4E
	.dc.b	$8D,$74,$73,$9C,$F5,$75,$C1,$87,$27,$FC,$DD,$3A,$BE,$27,$5C,$3A
	.dc.b	$4F,$F9,$62,$AD,$7F,$26,$F8,$F7,$CF,$BE,$83,$B5,$6B,$1F,$EE,$95
	.dc.b	$A9,$CE,$5E,$F8,$C4,$FB,$63,$17,$CE,$30,$7F,$B0,$27,$78,$F8,$EB
	.dc.b	$77,$ED,$F0,$FF,$C3,$37,$FE,$C6,$1F,$3D,$F2,$EF,$FC,$0D,$3A,$7F
	.dc.b	$D1,$D3,$AD,$F9,$75,$7D,$38,$C1,$FE,$CC,$AE,$1A,$FF,$46,$8C,$DF
	.dc.b	$F9,$F2,$73,$1F,$E1,$33,$33,$FD,$16,$33,$3F,$D4,$70,$F2,$2F,$A0
	.dc.b	$FF,$41,$DF,$A6,$6C,$FF,$A2,$26,$DF,$86,$6D,$7F,$BE,$66,$D7,$5F
	.dc.b	$67,$F0,$11,$89,$F8,$71,$83,$DD,$5C,$2B,$C5,$18,$3F,$11,$3F,$C4
	.dc.b	$0F,$99,$F4,$1F,$E5,$27,$FD,$DE,$66,$7E,$5C,$65,$FF,$4F,$C6,$70
	.dc.b	$F9,$13,$97,$FB,$44,$64,$F9,$11,$97,$FC,$79,$3F,$DD,$F9,$FE,$A4
	.dc.b	$E6,$3E,$74,$63,$C7,$FC,$F7,$CC,$FE,$46,$7E,$77,$CA,$63,$87,$DE
	.dc.b	$DC,$7E,$73,$67,$F0,$7A,$23,$19,$FC,$F6,$D7,$DD,$9D,$3F,$E5,$19
	.dc.b	$FF,$01,$7D,$03,$DC,$FD,$83,$67,$47,$B1,$79,$9A,$9F,$DB,$7D,$03
	.dc.b	$8A,$99,$BE,$39,$FF,$97,$E7,$51,$88,$7C,$FB,$E8,$3F,$80,$36,$75
	.dc.b	$5C,$2B,$FD,$A7,$39,$7F,$DA,$BD,$0F,$A0,$FE,$11,$E9,$7F,$9A,$F0
	.dc.b	$ED,$39,$7C,$11,$9A,$FF,$6E,$FC,$0F,$FA,$BA,$7C,$98,$3E,$53,$F8
	.dc.b	$94,$E6,$BC,$31,$FC,$57,$E8,$13,$FC,$5F,$E8,$3F,$4B,$18,$C3,$DD
	.dc.b	$1F,$BB,$E6,$FE,$D3,$F0,$5F,$CA,$5F,$ED,$7F,$41,$7C,$AF,$F6,$AD
	.dc.b	$FC,$89,$CB,$63,$39,$B9,$BF,$C8,$35,$A8,$C5,$7F,$A3,$63,$BB,$C6
	.dc.b	$FB,$61,$DD,$B1,$93,$B7,$C0,$F8,$FC,$0F,$23,$63,$27,$8E,$32,$D3
	.dc.b	$EE,$7C,$F3,$E7,$BF,$09,$97,$F8,$34,$46,$79,$F1,$18,$F9,$17,$2E
	.dc.b	$2B,$53,$C7,$E8,$7F,$3C,$F0,$6B,$9E,$EE,$6A,$32,$FF,$87,$63,$FD
	.dc.b	$C0,$3E,$37,$F6,$FC,$3F,$E6,$38,$C4,$FE,$55,$DF,$C1,$6B,$5F,$DE
	.dc.b	$C6,$A7,$E7,$9F,$EB,$B7,$15,$CE,$B8,$4E,$5A,$FE,$DE,$ED,$70,$BE
	.dc.b	$B5,$8F,$67,$3F,$2B,$30,$F2,$46,$5F,$34,$65,$F1,$9F,$C0,$E3,$BF
	.dc.b	$CE,$BE,$7B,$E2,$B6,$7E,$8B,$5F,$AC,$D7,$FC,$88,$E7,$3F,$F4,$43
	.dc.b	$71,$FE,$3E,$D9,$F8,$5E,$8F,$BB,$E7,$BF,$4B,$1F,$BF,$7F,$87,$37
	.dc.b	$C3,$D7,$6C,$8C,$10,$F5,$CF,$26,$C7,$59,$91,$D6,$3C,$1E,$CC,$8F
	.dc.b	$03,$E3,$F2,$B1,$1D,$22,$E7,$FE,$A3,$C3,$FC,$AF,$3F,$C5,$7E,$77
	.dc.b	$F8,$13,$7C,$1F,$B3,$18,$8F,$2F,$46,$26,$46,$CF,$81,$1F,$80,$F9
	.dc.b	$D7,$CE,$D2,$4F,$F2,$38,$FF,$68,$9A,$71,$1A,$7E,$EC,$6A,$31,$8F
	.dc.b	$FA,$EE,$38,$4E,$BE,$03,$C1,$F0,$7E,$1F,$96,$93,$DE,$32,$F8,$63
	.dc.b	$2F,$68,$FC,$BF,$6B,$E7,$69,$27,$F9,$D7,$CE,$3A,$B1,$1F,$92,$AE
	.dc.b	$0F,$F3,$F6,$BF,$96,$9E,$0F,$2E,$9F,$D2,$C1,$A6,$3A,$AC,$EC,$87
	.dc.b	$37,$2F,$58,$BA,$FF,$B7,$E9,$58,$E9,$FE,$EF,$1D,$EF,$9D,$6E,$B5
	.dc.b	$01,$3B,$7B,$1F,$0A,$36,$EA,$EC,$F8,$23,$AD,$F8,$27,$35,$A6,$CE
	.dc.b	$8F,$05,$77,$BF,$54,$FC,$39,$D1,$DC,$B3,$DD,$F2,$92,$78,$BE,$73
	.dc.b	$73,$E0,$75,$7E,$57,$9D,$F9,$5A,$70,$29,$C0,$F3,$CF,$3A,$EF,$19
	.dc.b	$3C,$C7,$D7,$39,$19,$BF,$44,$70,$25,$F8,$E7,$81,$BA,$ED,$19,$9D
	.dc.b	$CF,$88,$F7,$1B,$3C,$EE,$9F,$1B,$6B,$C3,$3C,$C9,$3C,$CF,$0C,$EE
	.dc.b	$BC,$31,$98,$C3,$E0,$6B,$ED,$BF,$94,$D3,$E0,$C7,$86,$32,$D8,$CD
	.dc.b	$78,$9F,$C8,$CE,$A3,$17,$E8,$9F,$14,$62,$FF,$0D,$1F,$04,$F0,$16
	.dc.b	$FB,$36,$79,$DF,$58,$E6,$7E,$49,$C3,$F2,$23,$F5,$31,$8B,$FC,$B4
	.dc.b	$62,$7D,$27,$23,$D2,$60,$E4,$75,$31,$1C,$9C,$1E,$53,$93,$EF,$78
	.dc.b	$91,$E0,$F6,$38,$19,$9F,$35,$FC,$4B,$FA,$A7,$C2,$2D,$FC,$2A,$F1
	.dc.b	$CE,$E3,$E6,$47,$E2,$9F,$E1,$3D,$1F,$C6,$3D,$49,$3F,$C9,$23,$F8
	.dc.b	$D9,$87,$F9,$04,$61,$FE,$49,$D3,$FC,$B8,$8F,$F9,$97,$E6,$FF,$B0
	.dc.b	$62,$C3,$FD,$1B,$A7,$F6,$AA,$FE,$5D,$7C,$1F,$F0,$0D,$F0,$AF,$F4
	.dc.b	$9A,$E8,$FF,$08,$BF,$D7,$A7,$7D,$38,$99,$1E,$18,$D9,$87,$AC,$61
	.dc.b	$D1,$E7,$7A,$F3,$48,$F5,$3E,$0F,$92,$C4,$79,$A2,$E7,$ED,$8F,$E5
	.dc.b	$D8,$2B,$FC,$4F,$F3,$4F,$7A,$92,$BF,$D1,$5F,$35,$FD,$83,$0E,$50
	.dc.b	$AE,$0F,$FA,$3D,$AF,$F8,$99,$B1,$C2,$79,$69,$FE,$BA,$5C,$DF,$E3
	.dc.b	$85,$CF,$4C,$44,$C3,$C7,$19,$7C,$B1,$97,$FC,$6B,$3C,$9D,$9A,$3F
	.dc.b	$D4,$9F,$31,$FD,$91,$FF,$1D,$B7,$0F,$6B,$70,$FF,$67,$BF,$E4,$99
	.dc.b	$FE,$AC,$E5,$BC,$FF,$E5,$08,$C4,$1E,$0B,$E5,$DF,$33,$FE,$6A,$6C
	.dc.b	$7F,$9D,$9A,$FF,$9E,$DA,$FF,$B8,$0E,$51,$FE,$5E,$8D,$E1,$F5,$1B
	.dc.b	$87,$A3,$9F,$C9,$CF,$E5,$E7,$98,$FC,$28,$CB,$F1,$23,$2F,$C4,$8C
	.dc.b	$BF,$1A,$33,$87,$C6,$9F,$F3,$7F,$CC,$7E,$8E,$BF,$BD,$35,$34,$FD
	.dc.b	$64,$FE,$29,$C7,$85,$C8,$E7,$D1,$ED,$F3,$BE,$31,$2F,$E3,$35,$FF
	.dc.b	$39,$E8,$8E,$11,$FE,$E1,$BF,$DA,$37,$9F,$CF,$CD,$23,$04,$3F,$CF
	.dc.b	$B3,$FA,$57,$94,$7B,$A6,$46,$08,$70,$7F,$D2,$3F,$2F,$FC,$07,$3F
	.dc.b	$D3,$FC,$C3,$CF,$FF,$4A,$B6,$FF,$DE,$D5,$BE,$1F,$F1,$14,$3E,$63
	.dc.b	$F1,$73,$98,$FF,$13,$35,$D7,$2F,$FA,$83,$E5,$DF,$31,$FE,$A2,$BF
	.dc.b	$C9,$9F,$94,$8C,$D7,$F8,$BF,$3F,$DD,$FC,$C3,$93,$FD,$52,$D9,$D6
	.dc.b	$7E,$48,$C4,$7F,$A8,$BE,$61,$F0,$7F,$EB,$06,$C7,$03,$FC,$79,$3C
	.dc.b	$30,$EF,$86,$EE,$56,$C5,$F2,$7B,$47,$FA,$FB,$DC,$7C,$C3,$D7,$FF
	.dc.b	$5F,$FC,$0B,$CB,$A4,$C3,$A6,$3F,$EE,$67,$FD,$73,$3B,$7C,$0D,$73
	.dc.b	$3E,$BE,$5E,$51,$83,$71,$97,$D1,$F2,$FE,$7D,$7F,$92,$B7,$7F,$B6
	.dc.b	$D9,$FF,$2F,$7C,$8B,$E6,$3F,$DA,$F3,$99,$34,$C4,$EE,$3B,$EF,$C0
	.dc.b	$8F,$C9,$8D,$3F,$ED,$6F,$11,$F3,$1F,$EE,$46,$C7,$CF,$6B,$FE,$DC
	.dc.b	$F9,$57,$C4,$EE,$A4,$ED,$FF,$76,$CE,$A3,$04,$3C,$11,$A9,$FF,$79
	.dc.b	$FC,$C7,$F4,$2D,$3F,$3A,$DA,$FF,$40,$35,$5F,$F8,$3A,$30,$4F,$2F
	.dc.b	$C3,$E3,$74,$65,$7E,$26,$79,$57,$F8,$5B,$E6,$3E,$E7,$07,$F5,$86
	.dc.b	$E3,$FA,$E6,$DF,$FA,$46,$9F,$31,$FB,$2B,$E1,$C5,$70,$E0,$DB,$FF
	.dc.b	$4C,$61,$C9,$F4,$DE,$54,$64,$93,$F2,$A7,$51,$F9,$B9,$FD,$13,$55
	.dc.b	$FA,$79,$BE,$8F,$4F,$06,$4F,$07,$2F,$EA,$FE,$61,$C5,$E3,$6C,$66
	.dc.b	$3C,$DF,$31,$FE,$19,$C3,$FD,$65,$F2,$EF,$97,$FE,$27,$87,$0C,$3C
	.dc.b	$7B,$63,$1E,$87,$2F,$4F,$CF,$2C,$39,$3F,$EB,$DC,$BB,$39,$C7,$93
	.dc.b	$E0,$6B,$93,$85,$74,$AE,$70,$AE,$71,$8C,$7C,$CD,$C3,$F9,$5B,$C2
	.dc.b	$77,$87,$F2,$E6,$DF,$FB,$46,$F8,$CF,$8A,$B8,$61,$E1,$8E,$0D,$AE
	.dc.b	$18,$76,$9E,$06,$B2,$F6,$4F,$A9,$AB,$0E,$EF,$23,$38,$F9,$1F,$76
	.dc.b	$1D,$DE,$18,$77,$78,$1F,$1C,$DD,$F2,$9E,$17,$DA,$78,$99,$98,$FF
	.dc.b	$83,$4E,$25,$BF,$AE,$FF,$84,$5A,$FC,$9F,$51,$F1,$1C,$63,$EA,$46
	.dc.b	$1E,$4D,$79,$35,$F0,$7C,$C7,$14,$1C,$34,$D3,$83,$E5,$FC,$BB,$E5
	.dc.b	$7F,$CB,$DB,$7E,$1A,$B9,$D6,$A7,$3B,$78,$27,$E4,$5F,$1D,$FE,$AC
	.dc.b	$70,$33,$5C,$1F,$F1,$95,$F3,$9D,$C7,$97,$0F,$CA,$DF,$38,$FA,$87
	.dc.b	$38,$CD,$F3,$8F,$EF,$F7,$CD,$D9,$CF,$0F,$F0,$1D,$F3,$BF,$64,$E4
	.dc.b	$F8,$F1,$99,$F6,$4F,$6C,$3C,$B7,$D6,$38,$B6,$76,$76,$C3,$C9,$7E
	.dc.b	$9D,$3F,$CF,$ED,$C3,$78,$F8,$2B,$FD,$19,$5A,$6C,$FA,$38,$6D,$6E
	.dc.b	$3F,$BF,$5F,$F8,$3D,$BC,$FA,$C1,$F3,$1F,$D4,$A7,$27,$F8,$2E,$33
	.dc.b	$D5,$D3,$06,$1F,$E9,$9A,$D5,$F1,$9F,$E2,$E7,$C6,$C3,$B4,$67,$0F
	.dc.b	$6E,$1E,$6F,$89,$7C,$C3,$E6,$3B,$67,$BC,$FF,$2B,$5A,$9C,$DF,$1B
	.dc.b	$E1,$5E,$E8,$C4,$EA,$3F,$C4,$95,$C5,$D1,$8A,$E3,$9F,$A0,$78,$7F
	.dc.b	$E1,$F9,$FF,$16,$0F,$96,$FA,$93,$FA,$89,$D3,$F0,$27,$5B,$FC,$79
	.dc.b	$CE,$9F,$EC,$39,$C9,$FE,$31,$8C,$BF,$B2,$33,$5E,$38,$CB,$FE,$E7
	.dc.b	$BF,$0B,$E6,$E9,$FB,$F5,$C2,$FF,$D8,$FF,$30,$F6,$33,$85,$AF,$F2
	.dc.b	$04,$DC,$3D,$13,$6F,$8D,$F0,$3F,$01,$59,$BF,$35,$5B,$FE,$13,$36
	.dc.b	$7E,$35,$70,$39,$3F,$E4,$F1,$CD,$E8,$F9,$8F,$F2,$DC,$66,$31,$E2
	.dc.b	$FE,$3E,$36,$7D,$29,$DE,$1F,$5E,$32,$FD,$18,$CE,$5F,$EE,$89,$CB
	.dc.b	$F4,$63,$31,$FE,$80,$32,$7D,$17,$33,$FC,$93,$E4,$7F,$24,$FD,$1F
	.dc.b	$8C,$FD,$14,$62,$BF,$09,$18,$E3,$F9,$ED,$8F,$C4,$EA,$8C,$5F,$D0
	.dc.b	$6E,$3F,$85,$9D,$57,$F9,$D7,$E6,$1E,$DF,$EC,$5B,$19,$F3,$31,$3F
	.dc.b	$B5,$F9,$7F,$CC,$CF,$F9,$F5,$B9,$7E,$76,$F8,$BF,$9F,$8C,$E2,$AF
	.dc.b	$F5,$1F,$CB,$3E,$63,$F8,$03,$63,$A7,$2E,$27,$F8,$27,$A5,$FE,$6D
	.dc.b	$9F,$F4,$63,$71,$FD,$67,$4B,$E6,$37,$E8,$F4,$F8,$D9,$87,$FA,$4F
	.dc.b	$E0,$7F,$D7,$0F,$FA,$57,$44,$FF,$11,$F9,$7F,$D0,$9F,$A2,$F9,$85
	.dc.b	$E7,$BB,$F9,$0F,$47,$F6,$6F,$7E,$F9,$3F,$AA,$8C,$67,$FE,$A7,$7F
	.dc.b	$B5,$F8,$1A,$AD,$B6,$75,$5C,$4F,$2F,$07,$F1,$EB,$E3,$3A,$BE,$2F
	.dc.b	$FA,$B7,$E8,$1F,$43,$F3,$6F,$FC,$01,$18,$9F,$EF,$4D,$8C,$CE,$DF
	.dc.b	$F6,$A8,$F9,$87,$CC,$FF,$AB,$63,$FD,$B3,$18,$E7,$FF,$02,$2F,$8E
	.dc.b	$73,$62,$F8,$5F,$F8,$03,$BD,$F0,$5F,$E6,$95,$F3,$14,$C4,$49,$FF
	.dc.b	$64,$4E,$9B,$39,$7F,$95,$43,$DA,$E9,$F2,$12,$78,$C7,$73,$0F,$08
	.dc.b	$C3,$A3,$BB,$AE,$EF,$89,$5D,$67,$FD,$B5,$F2,$EF,$98,$F5,$7D,$0C
	.dc.b	$9C,$C7,$FA,$16,$75,$1F,$E2,$96,$BF,$E3,$36,$BF,$41,$AF,$F8,$98
	.dc.b	$7C,$CF,$E4,$1F,$F7,$53,$FB,$66,$F8,$BF,$E9,$38,$D4,$7F,$4D,$6E
	.dc.b	$3F,$75,$E1,$19,$9E,$1D,$7C,$7D,$99,$3D,$9E,$9E,$66,$23,$A4,$5C
	.dc.b	$3A,$4D,$E8,$F4,$62,$AC,$5F,$F9,$02,$3F,$99,$BF,$E4,$2B,$FC,$87
	.dc.b	$CC,$FE,$C2,$3D,$BF,$43,$EB,$3F,$D5,$D1,$98,$FE,$D0,$D9,$F6,$4E
	.dc.b	$E7,$A6,$1F,$26,$B4,$FC,$BF,$99,$7C,$D7,$FA,$E5,$B5,$C0,$FF,$2A
	.dc.b	$B8,$7F,$BB,$3F,$EC,$13,$18,$7E,$9D,$B5,$D6,$BA,$7C,$0F,$5F,$85
	.dc.b	$ED,$F2,$D3,$1E,$C3,$04,$EE,$33,$09,$E1,$1E,$29,$D3,$E4,$8C,$E3
	.dc.b	$E3,$9F,$F6,$A7,$5B,$E6,$BF,$DA,$DF,$34,$FA,$1A,$6A,$26,$3E,$49
	.dc.b	$D3,$67,$2F,$F8,$5E,$3F,$44,$DE,$AE,$A7,$6E,$AB,$1C,$E2,$BC,$63
	.dc.b	$C7,$EF,$79,$3D,$89,$3E,$18,$CB,$E3,$8C,$BE,$18,$CB,$E3,$8C,$E9
	.dc.b	$FC,$67,$05,$6A,$31,$3A,$E1,$FD,$65,$6A,$FC,$07,$72,$1D,$CC,$9D
	.dc.b	$CC,$D7,$7B,$ED,$5D,$2F,$B5,$7F,$A7,$AF,$B3,$CC,$E4,$4B,$E4,$F6
	.dc.b	$7C,$EF,$6A,$F3,$DF,$68,$E9,$7D,$95,$F6,$BE,$71,$D6,$FC,$8F,$57
	.dc.b	$9B,$D4,$E6,$64,$F4,$5F,$5A,$FF,$1E,$46,$4E,$F1,$9B,$F2,$46,$0E
	.dc.b	$87,$91,$CC,$F9,$6B,$C7,$3E,$53,$B9,$D6,$3B,$B9,$3A,$1F,$26,$FE
	.dc.b	$2B,$98,$E8,$F2,$31,$5E,$0B,$F3,$47,$56,$BC,$EF,$E1,$BC,$0D,$1E
	.dc.b	$AA,$F1,$DF,$CC,$24,$FA,$9F,$19,$89,$F1,$38,$8D,$1D,$9D,$47,$C5
	.dc.b	$3B,$57,$C6,$BF,$A0,$D8,$C9,$F1,$63,$2D,$32,$7A,$CF,$B6,$7C,$58
	.dc.b	$E4,$D3,$81,$4E,$07,$8C,$C4,$F9,$AF,$E0,$DF,$AA,$7D,$AF,$AC,$C3
	.dc.b	$C6,$78,$57,$CE,$6C,$7C,$BB,$FA,$8D,$3E,$79,$F4,$4F,$1D,$FD,$C7
	.dc.b	$A5,$F6,$8C,$12,$7E,$1C,$F0,$6D,$7A,$5B,$5E,$79,$F8,$B3,$F5,$5F
	.dc.b	$59,$0E,$E6,$A3,$83,$F9,$28,$E0,$E8,$C4,$6A,$BE,$3C,$71,$3A,$3E
	.dc.b	$77,$EB,$CF,$95,$B3,$FD,$FF,$4F,$EE,$B3,$FE,$33,$F9,$B7,$D0,$FF
	.dc.b	$80,$E3,$0F,$F3,$EE,$A9,$87,$F4,$CA,$DB,$67,$2F,$FB,$3E,$3F,$BA
	.dc.b	$3B,$E9,$E9,$8F,$F8,$3E,$7F,$BA,$CE,$FA,$72,$84,$72,$7A,$56,$9C
	.dc.b	$B6,$35,$8F,$AF,$C4,$C1,$2B,$D1,$F2,$B5,$F3,$45,$9F,$85,$37,$1F
	.dc.b	$34,$DA,$F2,$5A,$BC,$5F,$38,$F8,$2F,$27,$CE,$3E,$86,$9A,$89,$87
	.dc.b	$F7,$49,$D3,$67,$2F,$FB,$DE,$3F,$BB,$BB,$E9,$EA,$76,$E9,$91,$92
	.dc.b	$1C,$5F,$8D,$F3,$BF,$E6,$28,$FF,$01,$B5,$6D,$FE,$73,$8C,$37,$C6
	.dc.b	$E3,$E3,$CA,$E2,$F4,$8C,$47,$2E,$0F,$F2,$55,$F2,$6D,$6E,$7F,$C2
	.dc.b	$7F,$38,$F9,$DF,$2F,$D0,$C9,$CC,$7D,$D9,$D4,$7F,$A1,$8F,$C2,$38
	.dc.b	$70,$D7,$FD,$16,$D7,$C7,$D7,$F0,$BA,$E4,$7C,$28,$F8,$A6,$1F,$51
	.dc.b	$F9,$1C,$BF,$5A,$9F,$E0,$DF,$3B,$FE,$5E,$CF,$FC,$D0,$D9,$FC,$C3
	.dc.b	$53,$FA,$E8,$DF,$8F,$F9,$98,$D1,$3A,$F1,$12,$BF,$C7,$33,$97,$E4
	.dc.b	$FC,$F7,$1F,$A1,$91,$82,$1F,$12,$3E,$43,$F2,$1A,$B4,$46,$29,$D5
	.dc.b	$E5,$E9,$97,$E2,$8F,$41,$87,$E3,$9E,$E8,$CB,$FE,$50,$EC,$FF,$2B
	.dc.b	$BF,$BB,$6B,$FE,$77,$F1,$BF,$75,$0F,$1A,$C4,$9E,$4E,$A3,$11,$E0
	.dc.b	$9E,$64,$8F,$E0,$9F,$3D,$FC,$33,$9F,$FD,$9D,$F3,$EF,$57,$F8,$43
	.dc.b	$70,$FF,$32,$D6,$F8,$BF,$D4,$5F,$3D,$F9,$99,$FE,$25,$39,$6F,$85
	.dc.b	$FC,$46,$75,$0F,$9E,$7C,$FE,$7B,$24,$FF,$9C,$A7,$2F,$F9,$D3,$E7
	.dc.b	$FF,$A8,$13,$AB,$F8,$E7,$CF,$3E,$36,$9F,$C7,$E7,$55,$FE,$7C,$9D
	.dc.b	$36,$76,$D3,$FD,$1D,$F0,$CF,$9F,$7C,$07,$F2,$9F,$8B,$FA,$15,$FE
	.dc.b	$86,$6B,$F7,$1B,$C9,$FE,$F7,$C7,$F0,$B3,$BE,$5E,$44,$7C,$9D,$35
	.dc.b	$E5,$16,$B9,$4F,$F3,$9F,$9F,$F4,$D7,$E2,$9B,$2E,$5F,$F6,$A4,$61
	.dc.b	$FF,$6B,$7C,$D3,$E2,$79,$6F,$30,$FF,$4F,$DE,$DF,$E9,$55,$B7,$33
	.dc.b	$B7,$C0,$FF,$4B,$8D,$47,$E7,$3C,$87,$CF,$FF,$4F,$6C,$7F,$AB,$1A
	.dc.b	$FE,$8B,$E7,$9F,$23,$BA,$93,$B7,$FA,$C4,$EA,$30,$43,$C1,$1A,$9F
	.dc.b	$EB,$DF,$3F,$FD,$36,$7F,$D7,$6D,$7F,$62,$DD,$7E,$44,$7F,$69,$BF
	.dc.b	$F5,$4C,$62,$13,$BD,$7F,$E0,$03,$63,$FB,$83,$55,$72,$9D,$FA,$1E
	.dc.b	$42,$4F,$FB,$3A,$74,$FF,$79,$F9,$EF,$C4,$D7,$F7,$B8,$46,$1D,$70
	.dc.b	$FC,$4F,$9F,$FE,$F9,$39,$9D,$CE,$0D,$3F,$ED,$C8,$D1,$23,$31,$FC
	.dc.b	$52,$3F,$3E,$EF,$97,$A9,$F4,$BA,$6C,$64,$87,$DB,$7F,$C1,$DF,$3E
	.dc.b	$E1,$A4,$9F,$F7,$6F,$CF,$BC,$4F,$ED,$70,$F9,$F7,$CF,$7C,$1C,$39
	.dc.b	$4F,$FB,$D1,$AF,$F2,$A6,$F1,$7D,$7C,$3A,$3F,$E2,$3B,$E7,$5F,$E2
	.dc.b	$59,$CB,$DE,$FB,$63,$F3,$EB,$72,$C3,$BC,$62,$F8,$B7,$1F,$4B,$C2
	.dc.b	$77,$87,$4B,$F1,$3F,$6E,$F8,$B6,$F8,$BB,$9F,$21,$FE,$3B,$AE,$6F
	.dc.b	$A6,$76,$DC,$7A,$BB,$C7,$D5,$7C,$F1,$F3,$56,$A9,$8F,$89,$E0,$67
	.dc.b	$1E,$EE,$B1,$EE,$DA,$D6,$3D,$A3,$89,$E8,$34,$5B,$EC,$FC,$B8,$E0
	.dc.b	$D3,$C9,$1A,$3E,$AB,$FA,$13,$99,$5C,$BF,$26,$7F,$BB,$C3,$C3,$7C
	.dc.b	$76,$1B,$5F,$E2,$39,$D4,$E4,$FE,$FB,$0F,$9F,$A4,$7F,$C0,$2D,$CF
	.dc.b	$F0,$BF,$3C,$F9,$1F,$E1,$31,$F8,$8A,$FF,$39,$57,$CF,$AD,$72,$7D
	.dc.b	$19,$C9,$F1,$E3,$2F,$C9,$33,$5D,$AF,$AE,$FF,$E3,$BC,$3A,$46,$2B
	.dc.b	$E3,$46,$30,$ED,$1E,$B3,$B4,$ED,$FE,$19,$3B,$7C,$06,$EF,$C1,$5C
	.dc.b	$A7,$51,$ED,$9D,$5F,$F1,$5C,$3D,$33,$9A,$DC,$F9,$EB,$C3,$7C,$DA
	.dc.b	$4D,$3F,$97,$61,$F0,$9F,$E6,$11,$F0,$AB,$B6,$3F,$CD,$A7,$6D,$BF
	.dc.b	$17,$05,$BF,$F5,$16,$1C,$E7,$27,$FA,$96,$FB,$F3,$6D,$7C,$FF,$F5
	.dc.b	$29,$D1,$ED,$9D,$46,$6F,$FC,$73,$7F,$E9,$9A,$DC,$7F,$A1,$27,$57
	.dc.b	$DA,$72,$F6,$E9,$F3,$52,$F7,$5F,$EB,$9F,$8E,$7C,$F3,$E7,$F7,$5A
	.dc.b	$B5,$6A,$73,$87,$C1,$8C,$61,$E4,$AE,$D3,$9C,$3A,$4E,$70,$E9,$7C
	.dc.b	$2F,$DB,$7B,$C3,$CF,$E1,$7F,$88,$B0,$F7,$8F,$9D,$F9,$D7,$FE,$D1
	.dc.b	$9D,$3F,$E8,$49,$D6,$FC,$BA,$7E,$B4,$60,$F3,$57,$0D,$7F,$B9,$C6
	.dc.b	$6F,$FC,$CF,$39,$8F,$F6,$F1,$99,$FE,$ED,$19,$9F,$F1,$16,$1E,$45
	.dc.b	$F3,$EF,$6B,$C3,$37,$0F,$8D,$36,$BD,$13,$70,$FF,$01,$4D,$AF,$0C
	.dc.b	$DA,$FF,$78,$4D,$9F,$89,$7F,$7A,$31,$5B,$AF,$1C,$F8,$8F,$C1,$D7
	.dc.b	$F1,$93,$E6,$FC,$FF,$F9,$6D,$FB,$46,$67,$E6,$46,$5F,$F4,$5C,$67
	.dc.b	$0F,$93,$39,$7F,$C3,$51,$93,$E4,$C6,$5F,$F0,$71,$99,$FF,$25,$46
	.dc.b	$5F,$9F,$D3,$F6,$23,$10,$F9,$7F,$CA,$57,$E0,$FE,$4F,$F3,$B1,$8C
	.dc.b	$91,$8E,$0F,$A2,$DA,$FC,$5C,$E9,$FC,$C5,$71,$9D,$5F,$D4,$7F,$53
	.dc.b	$F3,$EF,$76,$98,$F3,$2E,$1F,$A1,$F9,$FF,$F5,$AB,$78,$73,$8F,$FA
	.dc.b	$CA,$72,$FF,$AD,$3E,$75,$F3,$FF,$B6,$6D,$F1,$3F,$58,$E2,$7F,$D7
	.dc.b	$7E,$8F,$F9,$B6,$7F,$60,$D8,$FF,$5F,$F4,$3E,$7F,$F8,$0C,$E7,$93
	.dc.b	$FC,$9D,$83,$E7,$FC,$07,$F0,$6F,$82,$FF,$59,$65,$E0,$8F,$E1,$3F
	.dc.b	$3F,$F9,$D3,$F3,$DF,$3E,$AC,$F6,$7F,$23,$E5,$FE,$C7,$EF,$5F,$1F
	.dc.b	$F4,$D1,$8C,$BF,$8C,$BF,$DA,$3B,$B5,$5B,$6C,$EA,$B8,$9E,$4E,$1F
	.dc.b	$F8,$03,$7C,$67,$57,$C5,$FE,$25,$F3,$AF,$9D,$FF,$51,$DF,$F7,$D8
	.dc.b	$C7,$37,$FB,$C1,$7C,$8F,$F1,$E9,$FF,$01,$08,$DB,$8F,$A0,$FC,$6B
	.dc.b	$FE,$F8,$6B,$9E,$CF,$F7,$CC,$62,$3F,$C1,$91,$FE,$D2,$1F,$19,$FE
	.dc.b	$6F,$E5,$FF,$40,$46,$27,$F9,$53,$BF,$7F,$5A,$7F,$9D,$63,$53,$FE
	.dc.b	$FC,$6D,$72,$AD,$C8,$FD,$1F,$CA,$E1,$FD,$2F,$CA,$DB,$36,$64,$F8
	.dc.b	$A3,$46,$1F,$04,$61,$D1,$E8,$7B,$7C,$E3,$E7,$7E,$5C,$F8,$E3,$0E
	.dc.b	$1A,$FD,$06,$BF,$E3,$37,$38,$FF,$A9,$5B,$97,$FA,$5D,$B3,$EF,$C5
	.dc.b	$E7,$7B,$FE,$77,$F2,$D3,$FD,$55,$FD,$D3,$7C,$3D,$74,$C8,$C1,$06
	.dc.b	$5F,$EB,$5A,$E7,$18,$CB,$FD,$51,$1F,$E8,$D8,$ED,$EA,$F8,$4F,$17
	.dc.b	$93,$63,$24,$39,$1F,$5E,$7F,$C9,$8E,$6F,$C1,$5C,$63,$62,$76,$E3
	.dc.b	$E7,$3F,$78,$FD,$06,$BF,$9C,$6B,$FE,$5F,$6B,$FE,$C1,$77,$5F,$ED
	.dc.b	$36,$E1,$9E,$FF,$99,$CF,$F8,$BF,$9C,$7C,$5D,$26,$3F,$E0,$38,$FF
	.dc.b	$63,$1A,$71,$1A,$7D,$71,$AF,$80,$E9,$F2,$92,$79,$47,$73,$0F,$08
	.dc.b	$C3,$A3,$C0,$EB,$3F,$F7,$65,$2B,$F4,$5D,$6F,$9C,$A4,$C3,$FC,$35
	.dc.b	$F3,$6E,$AC,$47,$D0,$AE,$0F,$F8,$7D,$AF,$D4,$9E,$0F,$2E,$9F,$D8
	.dc.b	$47,$FB,$C2,$0D,$3F,$62,$EC,$87,$38,$E3,$D5,$CB,$9F,$9F,$A3,$2B
	.dc.b	$CF,$19,$7D,$31,$97,$C9,$19,$7C,$D1,$9D,$5C,$5F,$F0,$07,$1F,$D4
	.dc.b	$D7,$43,$99,$0E,$66,$4E,$66,$67,$9B,$FC,$46,$39,$BC,$0D,$92,$36
	.dc.b	$FC,$4B,$F3,$46,$27,$FC,$55,$3D,$6B,$CD,$5A,$7D,$17,$EB,$73,$3E
	.dc.b	$5B,$E5,$39,$8F,$25,$71,$7C,$D1,$97,$C4,$71,$AE,$AF,$84,$E2,$64
	.dc.b	$E2,$64,$E2,$7B,$EB,$C2,$E6,$FE,$1C,$7B,$6F,$C1,$1E,$E9,$E1,$5E
	.dc.b	$A7,$E3,$3E,$27,$27,$88,$C9,$E2,$32,$78,$AB,$C2,$E8,$C4,$75,$8C
	.dc.b	$91,$C1,$23,$A4,$65,$F0,$3E,$97,$D9,$7F,$2D,$EA,$70,$9F,$44,$EE
	.dc.b	$3C,$B5,$EF,$8E,$91,$C0,$91,$99,$E2,$78,$E3,$35,$F2,$2F,$C1,$5B
	.dc.b	$9F,$01,$E6,$AE,$11,$98,$FD,$8D,$70,$AF,$A2,$D8,$F5,$9F,$74,$EE
	.dc.b	$7D,$98,$EF,$1E,$18,$D4,$F5,$24,$EA,$38,$1A,$9E,$0F,$C2,$6D,$FC
	.dc.b	$53,$E1,$D7,$29,$E0,$72,$38,$16,$38,$5F,$C6,$8F,$82,$7C,$AB,$F6
	.dc.b	$BE,$66,$BF,$49,$B3,$CA,$3E,$41,$CA,$BC,$4F,$80,$FC,$53,$F9,$AA
	.dc.b	$F9,$B7,$EA,$6C,$F9,$5A,$F0,$BF,$88,$D3,$DA,$53,$B5,$FC,$28,$DB
	.dc.b	$63,$F9,$87,$47,$F3,$A2,$3F,$ED,$1F,$9A,$FE,$EB,$3F,$C3,$34,$FE
	.dc.b	$FF,$3F,$CE,$6B,$6F,$F6,$AA,$DE,$1F,$E2,$78,$47,$F7,$B7,$5D,$1F
	.dc.b	$E2,$98,$FE,$DF,$87,$F1,$87,$4D,$E8,$D4,$CA,$F2,$47,$1E,$59,$1E
	.dc.b	$77,$A7,$C8,$62,$3C,$91,$6F,$B4,$DC,$7C,$D8,$FF,$97,$27,$C5,$5C
	.dc.b	$AF,$FD,$E5,$F3,$4F,$6B,$FC,$63,$08,$C3,$FE,$34,$6B,$FE,$60,$6B
	.dc.b	$F0,$CE,$11,$FE,$49,$8D,$D7,$F3,$86,$E1,$F2,$7E,$67,$FB,$F5,$FF
	.dc.b	$85,$9F,$EE,$0F,$F9,$4E,$BF,$9F,$C3,$A3,$FC,$A8,$D9,$D5,$73,$E7
	.dc.b	$C9,$AA,$93,$A3,$FC,$BD,$5D,$9F,$F1,$2F,$CB,$FF,$72,$6F,$3F,$F9
	.dc.b	$86,$31,$8F,$F5,$07,$FC,$D3,$8F,$E0,$5C,$B6,$78,$1C,$7C,$2B,$7E
	.dc.b	$11,$5C,$1F,$F3,$83,$5C,$C7,$F9,$C7,$E5,$9F,$31,$F5,$1B,$3F,$E6
	.dc.b	$66,$BF,$E9,$26,$BF,$4D,$CE,$3F,$DA,$1B,$8F,$E6,$5B,$3E,$2E,$AF
	.dc.b	$81,$CF,$F0,$7A,$24,$FA,$73,$FD,$D3,$E8,$9C,$BF,$2A,$78,$BE,$FF
	.dc.b	$96,$FD,$3B,$FE,$8D,$6A,$69,$FB,$0C,$BF,$BE,$B8,$F0,$BF,$D2,$4E
	.dc.b	$BA,$6C,$4A,$EC,$EA,$3E,$47,$CB,$FC,$72,$4F,$C7,$6B,$FE,$64,$D1
	.dc.b	$1C,$23,$D1,$D3,$E0,$E7,$F0,$F4,$C9,$F6,$46,$5F,$74,$65,$FF,$37
	.dc.b	$E5,$FA,$FE,$0F,$E0,$D7,$F9,$ED,$3F,$C3,$2D,$8C,$F8,$1F,$EA,$F8
	.dc.b	$7C,$45,$89,$87,$E2,$67,$83,$FE,$B7,$76,$E6,$36,$FF,$A0,$23,$69
	.dc.b	$DB,$FE,$84,$F9,$6F,$E2,$B7,$FC,$6F,$E5,$DE,$6F,$FB,$05,$B5,$F9
	.dc.b	$9A,$DB,$FE,$C7,$F9,$7F,$CC,$46,$20,$D8,$FC,$E7,$CB,$3E,$5F,$FD
	.dc.b	$99,$1F,$9F,$32,$FF,$B3,$FE,$5F,$F1,$B6,$69,$8F,$96,$7C,$A7,$FB
	.dc.b	$55,$B3,$FA,$D9,$DF,$4E,$6A,$4F,$EC,$6B,$6F,$FA,$91,$BC,$3F,$E5
	.dc.b	$2F,$6D,$F2,$EF,$63,$FD,$C5,$F0,$3F,$22,$BF,$CA,$2D,$E2,$FE,$84
	.dc.b	$DC,$7F,$DD,$58,$FE,$05,$FF,$76,$46,$53,$4C,$46,$E3,$3C,$BE,$02
	.dc.b	$3E,$9E,$8F,$F7,$A7,$CB,$7A,$30,$FE,$A4,$D9,$FF,$7D,$35,$7C,$83
	.dc.b	$E5,$F8,$EF,$27,$F8,$A5,$6E,$B5,$18,$9D,$3F,$EC,$A9,$D5,$7F,$61
	.dc.b	$D3,$FC,$E7,$E4,$3E,$5F,$19,$7F,$65,$8C,$61,$FC,$85,$AF,$FB,$53
	.dc.b	$E5,$9F,$13,$C2,$07,$FA,$C6,$F0,$D3,$93,$89,$E3,$18,$21,$FE,$E0
	.dc.b	$AE,$F8,$7F,$A2,$27,$73,$FE,$E4,$F9,$7F,$EA,$6F,$81,$A9,$B1,$FE
	.dc.b	$B3,$72,$4F,$3B,$C5,$E4,$79,$A3,$15,$FC,$FD,$E3,$1E,$AF,$96,$FC
	.dc.b	$11,$38,$BD,$5F,$2F,$96,$5F,$F4,$79,$FC,$46,$2A,$CC,$FF,$BF,$27
	.dc.b	$93,$FD,$36,$74,$F1,$E6,$F0,$11,$D7,$4D,$79,$45,$D3,$E3,$4F,$D0
	.dc.b	$F9,$7F,$C4,$D7,$F5,$96,$A9,$FF,$59,$C6,$1F,$F5,$AF,$CB,$FF,$0F
	.dc.b	$7E,$DF,$CB,$BE,$5B,$FC,$51,$5B,$E4,$BD,$1D,$CF,$BF,$87,$1C,$58
	.dc.b	$71,$7F,$0B,$19,$8C,$42,$FC,$B1,$86,$E1,$FE,$37,$AE,$0D,$F1,$7D
	.dc.b	$57,$DA,$3C,$6F,$A1,$AE,$4F,$0C,$70,$BF,$0D,$74,$9F,$CC,$DF,$81
	.dc.b	$B8,$7F,$91,$EB,$77,$C7,$0F,$16,$5D,$EF,$65,$3D,$97,$C6,$77,$1E
	.dc.b	$31,$8F,$85,$D9,$C7,$1F,$13,$AC,$3C,$0F,$AB,$0D,$61,$D6,$3C,$27
	.dc.b	$5B,$E8,$75,$BE,$67,$47,$37,$F2,$A7,$FC,$85,$1F,$65,$B1,$97,$D2
	.dc.b	$6B,$D6,$7C,$37,$F9,$26,$3E,$EC,$61,$C5,$23,$11,$C6,$7F,$CA,$EB
	.dc.b	$E5,$E9,$1F,$F2,$DB,$74,$FC,$3F,$CA,$BE,$25,$BF,$CF,$BF,$E2,$37
	.dc.b	$F4,$6B,$53,$99,$EC,$7D,$18,$CB,$F4,$CE,$26,$6B,$AB,$93,$8D,$F8
	.dc.b	$A3,$2F,$E6,$A3,$8B,$9A,$F0,$5F,$8E,$B7,$BF,$C7,$8C,$4E,$AF,$C0
	.dc.b	$DB,$E5,$1E,$19,$DB,$4D,$96,$36,$78,$DF,$3D,$7F,$A7,$F0,$F7,$CF
	.dc.b	$4D,$FD,$B4,$9F,$87,$18,$8F,$5D,$71,$C7,$E1,$CF,$0A,$DE,$3D,$78
	.dc.b	$32,$EC,$FF,$34,$42,$33,$87,$F4,$68,$FF,$4C,$4F,$F9,$F6,$9F,$2F
	.dc.b	$FD,$32,$38,$F4,$F0,$9C,$F6,$72,$C3,$D1,$39,$BE,$11,$FC,$76,$FE
	.dc.b	$04,$E6,$B8,$61,$AC,$3E,$EF,$C7,$BE,$5D,$F2,$FA,$BF,$3C,$E6,$31
	.dc.b	$8F,$FA,$4E,$30,$FB,$A7,$75,$CA,$30,$6D,$C6,$5E,$53,$E1,$4F,$0B
	.dc.b	$F6,$78,$7F,$E1,$9B,$FF,$50,$0F,$95,$F9,$F7,$FD,$BA,$74,$F3,$9D
	.dc.b	$6F,$CB,$A7,$E7,$C6,$0F,$F4,$D5,$70,$D3,$F3,$B1,$9B,$F2,$09,$D4
	.dc.b	$7F,$7C,$35,$3F,$88,$8C,$CF,$EC,$B0,$F2,$AF,$97,$7B,$1E,$29,$B8
	.dc.b	$7F,$AA,$E6,$D7,$A6,$6E,$1F,$94,$9B,$5E,$19,$B5,$FE,$0E,$9B,$3F
	.dc.b	$16,$FF,$09,$18,$AD,$D7,$8E,$7C,$47,$E2,$6B,$F8,$A1,$F3,$BE,$5F
	.dc.b	$FC,$AA,$FF,$AF,$CC,$CF,$CD,$8C,$BF,$1A,$33,$87,$CA,$9C,$BF,$AC
	.dc.b	$8C,$9F,$26,$32,$FF,$89,$A7,$F8,$CF,$47,$D6,$9C,$C7,$CF,$8C,$78
	.dc.b	$FF,$9F,$F9,$0F,$CC,$5F,$C3,$F8,$BC,$70,$FC,$D6,$C7,$E2,$34,$46
	.dc.b	$31,$F9,$CD,$CF,$F0,$33,$AA,$FF,$6F,$61,$DA,$FE,$E5,$FE,$BF,$E5
	.dc.b	$DE,$DD,$31,$E6,$DC,$3F,$DC,$DF,$2F,$F8,$96,$F1,$67,$1F,$C6,$CE
	.dc.b	$5F,$C7,$FC,$AB,$E5,$FF,$78,$DB,$E2,$7F,$BC,$5C,$4F,$E5,$BD,$2F
	.dc.b	$F3,$24,$FF,$BD,$5B,$1F,$9A,$EA,$7C,$BF,$F0,$79,$CF,$0F,$F2,$0C
	.dc.b	$1F,$1D,$E0,$3F,$86,$7C,$0F,$FA,$AB,$2F,$04,$7F,$0F,$F9,$7F,$CE
	.dc.b	$9F,$9E,$F9,$75,$67,$B7,$F8,$A7,$C1,$3E,$3F,$F9,$56,$FF,$E6,$48
	.dc.b	$C6,$5F,$D9,$E3,$FD,$B3,$3A,$6C,$67,$E8,$1F,$41,$96,$C6,$63,$07
	.dc.b	$1F,$98,$7C,$CF,$EE,$6F,$FB,$A4,$63,$9B,$C0,$BE,$45,$CB,$8B,$E1
	.dc.b	$7F,$95,$3F,$13,$F1,$BF,$E7,$AE,$C7,$C6,$D3,$11,$27,$FD,$03,$CD
	.dc.b	$FD,$32,$B7,$18,$D7,$F4,$11,$88,$FE,$1E,$AB,$B7,$B3,$DF,$E4,$E5
	.dc.b	$71,$3F,$D1,$B4,$9D,$C6,$4F,$CF,$CF,$07,$AC,$7F,$A4,$BE,$5D,$F3
	.dc.b	$1E,$BF,$A0,$95,$C6,$72,$D9,$CB,$E6,$9E,$51,$23,$ED,$4F,$07,$FC
	.dc.b	$44,$D7,$E5,$35,$FE,$01,$F3,$3F,$88,$7D,$0F,$EB,$1B,$E1,$7F,$AA
	.dc.b	$23,$51,$FC,$D1,$BA,$FF,$78,$78,$46,$67,$87,$4F,$1F,$62,$4F,$67
	.dc.b	$A7,$93,$88,$E9,$17,$0E,$93,$FE,$B8,$F9,$9F,$D2,$D7,$9B,$E8,$3C
	.dc.b	$86,$7C,$09,$19,$21,$FD,$2A,$78,$B6,$38,$BA,$BF,$F0,$8C,$E9,$FF
	.dc.b	$09,$FC,$D3,$E4,$3F,$D9,$2D,$AF,$BC,$D7,$FA,$BB,$FE,$F5,$31,$3C
	.dc.b	$A7,$89,$CB,$E0,$B9,$FC,$27,$4F,$97,$98,$74,$8C,$BD,$A3,$2F,$68
	.dc.b	$FF,$6D,$F6,$7E,$F6,$3F,$21,$F3,$4F,$A0,$FF,$6F,$B6,$3F,$29,$3B
	.dc.b	$7F,$B6,$B7,$2F,$F1,$8C,$EE,$7F,$B8,$B6,$3F,$C6,$B3,$BE,$BF,$E9
	.dc.b	$0D,$AF,$F7,$7D,$6E,$7F,$A6,$36,$3F,$DE,$73,$B5,$1F,$A9,$7D,$3D
	.dc.b	$1E,$83,$D3,$D3,$63,$24,$3C,$AF,$BB,$DC,$F7,$FA,$B2,$7E,$1C,$65
	.dc.b	$F8,$B1,$97,$DB,$1F,$2B,$9F,$FB,$16,$7F,$E5,$2F,$9B,$7D,$06,$35
	.dc.b	$FE,$CD,$18,$C3,$E5,$DF,$17,$F8,$04,$65,$FF,$0C,$B6,$FE,$4C,$EE
	.dc.b	$73,$7D,$50,$AE,$17,$F3,$AB,$85,$F9,$6F,$85,$ED,$CD,$6E,$3C,$71
	.dc.b	$AB,$E7,$1F,$4A,$F6,$F8,$23,$2F,$33,$DA,$43,$CE,$74,$31,$7E,$C7
	.dc.b	$06,$8F,$C6,$CF,$67,$A1,$98,$E0,$E8,$E2,$FB,$5E,$B7,$EB,$78,$3F
	.dc.b	$4A,$FA,$6B,$FC,$42,$B8,$56,$8F,$71,$C8,$F7,$4F,$C3,$38,$DD,$8F
	.dc.b	$0D,$F7,$AF,$29,$D0,$8E,$49,$1D,$0E,$27,$80,$F6,$1E,$33,$E0,$C7
	.dc.b	$48,$E3,$18,$7D,$8F,$9C,$FE,$C9,$3E,$57,$95,$FB,$E7,$8E,$1F,$60
	.dc.b	$FB,$31,$99,$F1,$5F,$A1,$F3,$1E,$47,$C4,$E1,$D3,$E0,$3C,$53,$D7
	.dc.b	$0F,$8F,$1E,$E3,$D0,$76,$BF,$1C,$79,$EB,$AC,$F3,$3E,$15,$79,$B0
	.dc.b	$FC,$23,$D6,$3E,$06,$1E,$97,$A9,$1C,$19,$8E,$11,$EB,$C3,$C2,$F9
	.dc.b	$67,$75,$E3,$9F,$9B,$3B,$3D,$07,$C1,$73,$7E,$8A,$ED,$87,$1A,$D9
	.dc.b	$EB,$9F,$8F,$87,$AE,$3E,$3C,$ED,$D5,$F9,$1F,$B3,$87,$C0,$9D,$61
	.dc.b	$E6,$9F,$99,$7C,$DF,$31,$F0,$9C,$61,$F3,$1F,$82,$F9,$EF,$C1,$1A
	.dc.b	$C3,$C1,$3F,$72,$FB,$5F,$D2,$C3,$ED,$B4,$FA,$38,$72,$8C,$D7,$B9
	.dc.b	$B7,$F1,$89,$5F,$12,$FE,$3B,$F4,$6F,$DC,$D3,$E2,$94,$FE,$4B,$D1
	.dc.b	$FC,$57,$A2,$47,$F2,$A8,$FE,$30,$61,$FE,$57,$18,$7F,$97,$F8,$9F
	.dc.b	$C7,$48,$FF,$32,$F9,$7A,$FF,$32,$8B,$9F,$F3,$D9,$FE,$53,$F3,$8F
	.dc.b	$7F,$F9,$5F,$CE,$3E,$82,$9A,$89,$8F,$F8,$F6,$74,$D9,$CB,$FE,$D8
	.dc.b	$AE,$8E,$48,$66,$BB,$F3,$F8,$3A,$24,$F5,$8C,$BD,$E3,$2F,$FB,$93
	.dc.b	$04,$FF,$42,$F9,$CF,$F2,$C4,$FF,$94,$5A,$FF,$BA,$9B,$9B,$87,$FD
	.dc.b	$81,$1F,$E6,$18,$C4,$FF,$52,$8D,$CE,$BC,$6E,$C7,$7F,$1F,$FA,$8F
	.dc.b	$CE,$3E,$77,$C5,$F4,$13,$1F,$F3,$14,$E5,$B3,$97,$C9,$3C,$A2,$47
	.dc.b	$FA,$4E,$78,$3F,$E8,$10,$E1,$FF,$42,$35,$F5,$F4,$FC,$3E,$A9,$1F
	.dc.b	$0E,$3E,$31,$87,$E2,$46,$1F,$F3,$C7,$CF,$7F,$97,$33,$FF,$72,$36
	.dc.b	$3F,$BA,$B5,$3F,$C5,$CC,$13,$C6,$FE,$EA,$FF,$1B,$C1,$E9,$63,$E8
	.dc.b	$25,$FC,$68,$F9,$66,$17,$07,$D4,$7B,$CE,$9E,$07,$C7,$9F,$F5,$D3
	.dc.b	$F1,$9B,$5E,$08,$FF,$5F,$5F,$FA,$B2,$FA,$2E,$7F,$39,$E8,$E8,$95
	.dc.b	$F1,$23,$2F,$C6,$8F,$F0,$8F,$57,$F9,$7B,$4F,$F0,$CC,$A3,$0F,$F8
	.dc.b	$71,$AE,$7C,$3E,$3E,$24,$8F,$F5,$44,$E8,$4F,$F8,$7B,$F1,$87,$6F
	.dc.b	$83,$DB,$97,$FE,$AF,$9C,$B6,$72,$FC,$AC,$7C,$B8,$7F,$BF,$63,$44
	.dc.b	$34,$7F,$8C,$AB,$9B,$80,$E3,$5F,$A9,$3F,$B2,$9F,$F6,$07,$CF,$66
	.dc.b	$FE,$1F,$CF,$BE,$83,$FC,$6E,$DA,$DE,$5F,$E3,$AF,$9F,$FD,$7C,$7F
	.dc.b	$90,$B0,$E4,$D9,$D7,$5F,$F9,$03,$E8,$1F,$41,$8A,$FF,$68,$61,$C9
	.dc.b	$4E,$9F,$F6,$A7,$D0,$3A,$FF,$C9,$2D,$8C,$C7,$F9,$33,$E7,$DF,$15
	.dc.b	$4D,$C4,$9F,$F6,$F4,$E9,$B3,$B6,$9F,$EE,$7F,$83,$7D,$03,$E3,$BF
	.dc.b	$CB,$3F,$11,$F7,$AB,$FD,$D0,$D7,$E8,$37,$97,$F4,$58,$FE,$67,$93
	.dc.b	$F1,$F3,$96,$F4,$74,$3D,$5D,$57,$0E,$B1,$5F,$B9,$7C,$DF,$F3,$8F
	.dc.b	$D0,$7A,$DF,$C0,$37,$1F,$D9,$37,$0F,$CF,$FC,$D3,$E8,$3F,$CE,$73
	.dc.b	$9B,$E7,$BE,$0F,$F3,$BD,$6E,$3F,$CF,$F8,$74,$9D,$46,$30,$FD,$5D
	.dc.b	$F3,$85,$F1,$7F,$29,$E4,$BE,$83,$FD,$06,$D8,$E4,$7F,$A3,$1C,$3F
	.dc.b	$9B,$F9,$D7,$C8,$F0,$A9,$3C,$0F,$E9,$AF,$FA,$3A,$77,$19,$AE,$79
	.dc.b	$FF,$15,$6E,$5F,$C0,$9B,$3F,$E9,$CA,$E2,$FE,$AF,$E8,$3F,$B2,$B7
	.dc.b	$D0,$FF,$4D,$C6,$30,$FE,$41,$5F,$EC,$5A,$D1,$E5,$9D,$61,$FC,$41
	.dc.b	$B1,$DF,$D8,$E7,$84,$8E,$0F,$AE,$3F,$D5,$3F,$3F,$F8,$99,$FF,$58
	.dc.b	$B5,$37,$07,$D0,$70,$E3,$95,$F1,$EB,$53,$88,$FE,$11,$1C,$CC,$BB
	.dc.b	$E6,$F0,$91,$FA,$17,$FB,$37,$8C,$65,$A6,$BA,$7F,$D8,$3F,$41,$F8
	.dc.b	$48,$FB,$0D,$7F,$8C,$37,$6F,$E9,$5F,$41,$FC,$D5,$FF,$67,$FC,$F3
	.dc.b	$E7,$BF,$9C,$D6,$F8,$6F,$2F,$63,$FD,$9F,$87,$1C,$58,$71,$7F,$DB
	.dc.b	$31,$98,$C4,$2F,$C9,$18,$6E,$1F,$D2,$2B,$83,$7C,$7F,$EC,$37,$DA
	.dc.b	$3C,$6F,$A1,$AE,$4F,$0C,$70,$BF,$0D,$74,$9F,$F7,$6D,$F8,$1B,$87
	.dc.b	$F5,$8A,$DD,$F1,$C3,$C5,$97,$7B,$D9,$4F,$65,$F1,$9D,$C7,$8C,$63
	.dc.b	$E1,$76,$71,$C7,$C4,$EB,$0F,$03,$EA,$C3,$58,$75,$8F,$09,$D6,$FA
	.dc.b	$1D,$6F,$99,$D1,$CD,$FC,$A9,$FE,$EA,$77,$BF,$F1,$63,$4F,$F1,$74
	.dc.b	$FD,$D7,$6F,$F7,$0F,$A0,$DB,$63,$7E,$2A,$72,$D8,$FF,$09,$FD,$07
	.dc.b	$F8,$51,$34,$8E,$8F,$91,$F3,$EF,$93,$FF,$29,$ED,$F8,$A9,$1C,$8D
	.dc.b	$C6,$B6,$FF,$23,$5F,$D5,$BE,$5B,$FD,$7B,$E3,$39,$BE,$11,$F1,$23
	.dc.b	$13,$B8,$ED,$87,$E6,$AF,$9C,$7D,$D3,$99,$E2,$9E,$15,$CA,$34,$72
	.dc.b	$C3,$FC,$4B,$5C,$AF,$D9,$39,$3D,$B1,$99,$F5,$C7,$53,$9D,$75,$AE
	.dc.b	$2D,$9D,$C7,$9F,$0F,$6D,$2F,$D3,$A7,$F9,$2D,$B8,$F0,$C3,$CB,$7F
	.dc.b	$E8,$2A,$D3,$67,$D3,$C1,$70,$F3,$CD,$7B,$45,$C3,$FD,$59,$5F,$D4
	.dc.b	$AD,$F4,$1F,$D4,$23,$FA,$74,$EA,$FF,$C6,$35,$AE,$CE,$5D,$3D,$E3
	.dc.b	$17,$C2,$31,$87,$3B,$FF,$1B,$61,$6F,$FD,$37,$F2,$0F,$A0,$7D,$07
	.dc.b	$97,$3F,$54,$E6,$31,$9F,$F6,$88,$C3,$E6,$9D,$D7,$28,$C1,$B7,$18
	.dc.b	$F8,$A7,$77,$F1,$7C,$4F,$F1,$05,$FF,$74,$1F,$3D,$F4,$AF,$F3,$B3
	.dc.b	$A7,$FC,$A7,$3A,$DF,$97,$4F,$D1,$8C,$1F,$DD,$EB,$86,$9F,$EB,$A8
	.dc.b	$CD,$FF,$96,$67,$31,$FA,$A3,$33,$FE,$8D,$8C,$CF,$FB,$17,$0F,$22
	.dc.b	$FA,$07,$B7,$E1,$9B,$87,$F8,$1E,$6D,$7A,$26,$E1,$FE,$99,$9B,$5E
	.dc.b	$19,$B5,$FB,$C9,$B3,$F1,$2F,$F1,$91,$8A,$DD,$78,$E7,$C4,$7E,$46
	.dc.b	$BF,$D0,$A7,$CD,$FA,$0F,$F2,$CB,$FE,$1B,$33,$3F,$32,$32,$FF,$9F
	.dc.b	$63,$38,$7C,$99,$CB,$FE,$B1,$8C,$9F,$26,$32,$FF,$A8,$0C,$CF,$FB
	.dc.b	$B6,$32,$FC,$FE,$7F,$B1,$18,$87,$D0,$7E,$5E,$BE,$27,$CA,$FE,$AE
	.dc.b	$31,$3F,$4E,$31,$C1,$F4,$5B,$5F,$8B,$CD,$18,$AF,$A4,$DC,$BF,$09
	.dc.b	$3A,$AF,$F1,$F7,$D0,$3E,$23,$FD,$9D,$87,$FA,$EA,$31,$3F,$D1,$9B
	.dc.b	$1F,$EF,$36,$BF,$D3,$3E,$81,$C1,$CA,$3F,$9B,$C6,$2B,$FA,$7B,$63
	.dc.b	$FD,$F8,$D7,$FA,$AF,$CF,$BD,$BE,$50,$8C,$57,$F5,$86,$C7,$F5,$C3
	.dc.b	$53,$C3,$D2,$FF,$3A,$57,$F5,$E6,$CE,$DA,$FF,$64,$6B,$FE,$DA,$A7
	.dc.b	$C9,$F8,$9B,$5E,$18,$C1,$E6,$F9,$FF,$E8,$B5,$FD,$B9,$AF,$FB,$91
	.dc.b	$AF,$F5,$A6,$AF,$9D,$FE,$00,$FF,$76,$6B,$FE,$EC,$6B,$FD,$89,$AB
	.dc.b	$E7,$1F,$39,$F4,$A3,$E2,$46,$35,$FE,$FE,$BE,$41,$DD,$8B,$E1,$7F
	.dc.b	$E0,$6F,$A1,$FC,$5B,$FE,$06,$6B,$F3,$DA,$EF,$97,$FC,$16,$FE,$4B
	.dc.b	$DE,$FE,$EE,$FF,$BA,$61,$F1,$98,$E7,$70,$6A,$73,$5F,$E9,$17,$FD
	.dc.b	$B4,$4A,$E1,$3B,$6C,$FF,$81,$27,$7E,$96,$23,$F4,$7F,$29,$64,$D9
	.dc.b	$93,$DE,$34,$61,$FF,$80,$1F,$36,$F9,$CE,$5F,$3B,$27,$FD,$4D,$39
	.dc.b	$6C,$FC,$28,$C4,$7E,$55,$AF,$F8,$C9,$AF,$C6,$77,$9F,$C4,$8F,$F0
	.dc.b	$FC,$63,$3D,$57,$AF,$CA,$F7,$7C,$EB,$E3,$EF,$74,$C7,$FD,$75,$39
	.dc.b	$6C,$E5,$65,$DD,$4F,$F0,$37,$67,$2F,$4B,$B9,$1E,$BE,$25,$79,$E7
	.dc.b	$F8,$0A,$56,$6F,$B4,$2F,$EB,$7C,$E7,$EB,$B0,$F3,$FC,$EF,$B0,$CF
	.dc.b	$6C,$8F,$F2,$BC,$ED,$FE,$00,$D7,$FC,$BA,$D7,$F2,$2E,$F0,$F7,$37
	.dc.b	$2C,$D7,$CA,$EB,$FC,$47,$CE,$BE,$36,$93,$1F,$F7,$24,$7F,$B1,$4D
	.dc.b	$38,$8D,$3F,$08,$FF,$2B,$C6,$FD,$EE,$BF,$2F,$27,$9C,$78,$0C,$3C
	.dc.b	$44,$A7,$51,$FE,$F5,$EE,$7C,$FF,$E7,$FE,$75,$F3,$D4,$FD,$14,$EA
	.dc.b	$B8,$57,$ED,$A3,$11,$FC,$A5,$AF,$FA,$19,$AF,$EE,$9D,$EF,$F8,$76
	.dc.b	$ED,$9A,$F1,$77,$78,$FB,$26,$1F,$E7,$8C,$3F,$1D,$97,$FB,$ED,$BD
	.dc.b	$9A,$9F,$81,$39,$C7,$F8,$4B,$63,$FD,$02,$FE,$E9,$E9,$1A,$E9,$F8
	.dc.b	$5E,$74,$9C,$9C,$23,$E3,$19,$7B,$C6,$1F,$99,$D5,$FC,$27,$E7,$9F
	.dc.b	$3D,$4F,$97,$12,$7A,$89,$DB,$86,$BF,$6F,$7F,$D6,$36,$7A,$3B,$AE
	.dc.b	$8E,$4C,$10,$C1,$34,$FD,$C5,$F6,$6C,$F2,$7E,$B5,$76,$AF,$C6,$56
	.dc.b	$F2,$FB,$93,$B7,$F9,$BC,$66,$7F,$D5,$ED,$7F,$A4,$4E,$DC,$DF,$5B
	.dc.b	$EF,$1F,$28,$F4,$3E,$47,$07,$A6,$FB,$CF,$17,$C7,$1A,$9E,$0F,$A2
	.dc.b	$7C,$0F,$88,$EB,$3C,$9E,$0F,$27,$81,$A3,$FD,$93,$87,$D2,$AF,$3C
	.dc.b	$68,$8E,$09,$3D,$63,$2F,$58,$CC,$7C,$23,$B1,$E5,$3D,$C6,$4E,$27
	.dc.b	$8A,$FD,$78,$7C,$D3,$E7,$5F,$96,$FB,$1E,$1B,$F8,$67,$82,$FB,$38
	.dc.b	$8D,$1C,$23,$B1,$E4,$AE,$B7,$E0,$C3,$E8,$4E,$8F,$39,$F9,$B7,$2F
	.dc.b	$6A,$E8,$E4,$F7,$E1,$E7,$79,$1F,$C2,$6B,$F8,$71,$1F,$F1,$B1,$F1
	.dc.b	$4D,$18,$8F,$D2,$B8,$3F,$8A,$47,$E0,$23,$B4,$7E,$BF,$B3,$13,$F2
	.dc.b	$9B,$1F,$82,$3F,$32,$E1,$F6,$E1,$FD,$8D,$F7,$D7,$43,$AC,$67,$1F
	.dc.b	$9D,$19,$BF,$E0,$97,$CE,$7E,$95,$6F,$0F,$8B,$7B,$C3,$F8,$15,$EE
	.dc.b	$7E,$4D,$6E,$7E,$85,$F6,$C3,$AC,$7A,$4F,$A0,$6A,$BF,$6B,$7C,$E3
	.dc.b	$07,$B8,$E1,$87,$85,$FD,$FC,$6D,$E6,$7C,$93,$75,$F2,$5A,$78,$70
	.dc.b	$F3,$3C,$2B,$A6,$1F,$C4,$2F,$A4,$7C,$86,$DF,$AA,$BE,$6B,$D5,$A7
	.dc.b	$58,$DB,$F1,$8F,$09,$F5,$9C,$BE,$A3,$F9,$67,$67,$F3,$5E,$B9,$18
	.dc.b	$21,$FC,$79,$CB,$FC,$82,$3F,$9E,$74,$48,$FE,$6C,$FF,$3F,$F9,$4C
	.dc.b	$47,$F2,$68,$B9,$7F,$2E,$9F,$F2,$67,$CE,$BD,$AF,$F2,$77,$CE,$BE
	.dc.b	$7A,$9A,$89,$8F,$E8,$2B,$83,$63,$FC,$CF,$18,$8F,$F4,$D3,$70,$FF
	.dc.b	$1A,$37,$83,$FC,$D6,$D9,$0C,$3F,$C2,$6F,$F8,$25,$76,$79,$BA,$E4
	.dc.b	$FF,$58,$9C,$B6,$72,$F1,$9F,$6C,$7F,$9A,$FE,$79,$D5,$49,$5F,$D8
	.dc.b	$A3,$FC,$F2,$FF,$86,$5D,$B9,$8D,$B9,$F1,$FF,$CE,$DF,$34,$F9,$DE
	.dc.b	$DF,$3D,$2B,$FB,$6C,$E5,$B3,$FE,$8E,$8C,$47,$FB,$05,$AF,$FA,$40
	.dc.b	$38,$7F,$D2,$6F,$0D,$BF,$CC,$0D,$DB,$35,$EF,$EE,$F8,$1D,$9F,$07
	.dc.b	$B2,$63,$EF,$8C,$BF,$06,$32,$FC,$18,$FF,$49,$7C,$E7,$EA,$89,$D3
	.dc.b	$8E,$AC,$B2,$BF,$C1,$73,$B9,$F8,$BF,$3B,$CB,$E7,$A5,$7C,$48,$CB
	.dc.b	$FC,$B3,$54,$7C,$52,$46,$5D,$BF,$E4,$D8,$DB,$E0,$EB,$F3,$F5,$FA
	.dc.b	$3A,$E4,$F9,$23,$2F,$9A,$32,$FF,$95,$F2,$FF,$64,$F0,$FF,$A2,$EF
	.dc.b	$F2,$6D,$F1,$7F,$D6,$91,$8E,$2F,$F6,$2A,$F8,$EC,$CC,$C3,$FC,$5B
	.dc.b	$3B,$7F,$3E,$ED,$FF,$18,$FC,$F7,$FA,$2D,$DF,$B1,$D8,$FF,$61,$46
	.dc.b	$63,$FC,$D8,$D7,$F0,$CD,$7F,$DA,$ED,$7F,$5A,$D4,$F3,$C7,$FD,$50
	.dc.b	$DC,$78,$F8,$1E,$EE,$BF,$F6,$97,$CE,$71,$24,$7F,$B9,$9E,$9F,$37
	.dc.b	$FA,$6C,$7F,$30,$1C,$3F,$9B,$6B,$FE,$EC,$6B,$FC,$0B,$E6,$DF,$39
	.dc.b	$FE,$F2,$6C,$7F,$70,$6B,$D9,$AF,$E9,$FE,$71,$D3,$FE,$E9,$6C,$7F
	.dc.b	$97,$9A,$FF,$A5,$7E,$69,$F3,$3F,$EE,$F6,$C7,$F7,$E6,$BE,$C6,$BF
	.dc.b	$B5,$6B,$FC,$6B,$DF,$7C,$E3,$E2,$3A,$5C,$AF,$F3,$DD,$6A,$7F,$DF
	.dc.b	$DF,$39,$F8,$39,$FF,$54,$37,$8F,$FD,$B8,$DE,$2F,$E4,$F2,$CB,$F2
	.dc.b	$F1,$89,$D4,$FF,$2C,$6A,$6C,$F0,$8D,$73,$F8,$88,$FF,$A1,$23,$4B
	.dc.b	$E6,$FE,$1E,$1F,$EE,$F6,$E1,$E4,$6C,$FF,$3C,$F9,$67,$CE,$7A,$2F
	.dc.b	$86,$B7,$0E,$0D,$BF,$E8,$78,$72,$58,$72,$7F,$D2,$F1,$92,$4F,$18
	.dc.b	$DE,$17,$99,$F3,$94,$93,$FE,$A0,$F9,$97,$C5,$E2,$3F,$DB,$15,$C1
	.dc.b	$FF,$52,$B5,$FF,$6E,$4F,$07,$35,$C7,$46,$9F,$5A,$F8,$46,$2B,$FA
	.dc.b	$FD,$F1,$7F,$DC,$DF,$39,$FD,$8D,$34,$C4,$7F,$B2,$63,$3E,$8F,$BE
	.dc.b	$74,$FC,$19,$D4,$62,$7F,$D7,$AF,$2F,$27,$E3,$7C,$DF,$E1,$09,$C5
	.dc.b	$C7,$E7,$32,$CB,$FF,$62,$CF,$E2,$B0,$56,$67,$FD,$FD,$3C,$9F,$99
	.dc.b	$1D,$B9,$64,$72,$7C,$DD,$38,$8E,$31,$74,$F8,$D0,$C3,$F6,$F1,$8F
	.dc.b	$9C,$FC,$1C,$FF,$B6,$1A,$A3,$5C,$BF,$83,$F9,$C7,$87,$9A,$7C,$D3
	.dc.b	$E6,$BF,$57,$8F,$2C,$78,$D7,$5C,$79,$61,$C7,$0E,$13,$FE,$18,$C7
	.dc.b	$AD,$FF,$86,$A7,$32,$C3,$CB,$18,$6E,$3F,$BA,$BE,$35,$E4,$8C,$61
	.dc.b	$DD,$F2,$E3,$FC,$03,$0E,$4D,$8D,$E3,$E1,$6D,$F3,$3F,$C5,$B8,$F8
	.dc.b	$1B,$5C,$5F,$56,$1D,$DB,$7C,$DD,$94,$F6,$4F,$15,$8F,$AB,$1F,$82
	.dc.b	$DC,$7D,$33,$C7,$1F,$13,$C3,$1F,$03,$71,$D6,$3D,$6B,$78,$73,$AF
	.dc.b	$29,$ED,$3C,$67,$03,$CC,$ED,$E8,$7F,$54,$7E,$7B,$FE,$2D,$6B,$F2
	.dc.b	$0C,$78,$6F,$9C,$FF,$5F,$CA,$31,$3C,$DA,$F0,$9F,$F6,$2B,$FE,$5C
	.dc.b	$F9,$CA,$9A,$47,$6F,$CD,$F9,$B7,$CC,$FF,$07,$DB,$F1,$75,$CC,$F8
	.dc.b	$13,$AE,$9F,$A9,$19,$9F,$C8,$CE,$63,$E3,$19,$DF,$FA,$4C,$66,$7A
	.dc.b	$CF,$94,$EB,$7E,$1A,$FE,$F7,$87,$96,$31,$3F,$E5,$58,$C6,$1E,$28
	.dc.b	$E5,$7E,$2A,$DC,$7F,$58,$AD,$BE,$43,$77,$E4,$9D,$C7,$9C,$FF,$42
	.dc.b	$E1,$E7,$9F,$1C,$79,$B0,$F0,$5F,$4C,$3D,$6D,$AE,$93,$FE,$8E,$DB
	.dc.b	$F8,$6E,$1F,$15,$FE,$1F,$87,$71,$19,$9E,$31,$FC,$35,$F9,$3C,$5F
	.dc.b	$CE,$EB,$9E,$56,$72,$77,$3E,$7F,$46,$6F,$F8,$B7,$CE,$7F,$50,$9D
	.dc.b	$47,$BA,$75,$6A,$D4,$F8,$4D,$57,$09,$D7,$67,$59,$F4,$DF,$2B,$E3
	.dc.b	$5F,$02,$BC,$5F,$20,$F9,$C7,$CE,$71,$D7,$79,$FF,$AD,$2B,$53,$9B
	.dc.b	$E3,$7C,$2B,$D3,$9F,$C8,$9C,$D6,$E7,$DB,$7C,$9D,$E1,$DA,$D3,$97
	.dc.b	$DC,$3D,$1F,$F1,$55,$7C,$61,$F3,$3F,$56,$7F,$D9,$33,$A7,$E3,$4E
	.dc.b	$B6,$F8,$93,$9D,$3F,$B9,$4E,$4F,$4C,$65,$FF,$69,$99,$AF,$1C,$65
	.dc.b	$FE,$FB,$7E,$17,$CD,$D5,$F7,$6B,$85,$FF,$5A,$F9,$C7,$B3,$9C,$2D
	.dc.b	$7C,$89,$B8,$7A,$26,$DF,$1B,$E0,$7D,$DA,$CD,$F9,$AA,$DF,$FB,$AA
	.dc.b	$6C,$FC,$6A,$E0,$72,$7E,$B9,$FC,$6C,$F9,$9F,$39,$FE,$62,$73,$D1
	.dc.b	$EB,$8C,$5F,$D4,$9E,$0F,$CE,$8C,$E3,$FD,$E2,$72,$FC,$E8,$CC,$7F
	.dc.b	$96,$0C,$9F,$39,$CC,$FD,$0A,$F0,$78,$DF,$9B,$F9,$CF,$C3,$5F,$E0
	.dc.b	$3E,$53,$1C,$7F,$2D,$B1,$F8,$5D,$51,$8C,$7E,$63,$73,$FC,$0C,$EA
	.dc.b	$BF,$33,$87,$68,$C5,$FE,$4B,$E7,$1E,$E7,$EB,$7E,$73,$F2,$98,$7E
	.dc.b	$82,$72,$FE,$8A,$33,$5F,$9A,$6E,$3F,$E1,$CF,$9C,$7B,$1C,$F9,$3F
	.dc.b	$71,$18,$C3,$F5,$6D,$C3,$F8,$07,$A3,$FE,$8C,$FA,$5D,$E1,$FC,$2E
	.dc.b	$31,$7F,$B2,$7F,$CD,$EF,$ED,$3E,$01,$F3,$FF,$C1,$67,$38,$7E,$BA
	.dc.b	$73,$5E,$0F,$93,$FE,$15,$F4,$29,$F2,$C6,$1F,$E1,$DF,$3F,$FB,$1C
	.dc.b	$3F,$D3,$FF,$47,$8E,$DF,$27,$B2,$8F,$E2,$9F,$3E,$F9,$FF,$DE,$6B
	.dc.b	$FE,$C6,$8C,$71,$7F,$B2,$21,$0F,$91,$FE,$0B,$1B,$AE,$2F,$FA,$88
	.dc.b	$7C,$FB,$C8,$FE,$E2,$D8,$E8,$D7,$FD,$AA,$D7,$97,$D7,$7D,$47,$C4
	.dc.b	$D7,$FD,$B6,$D7,$7C,$9F,$DF,$BB,$7F,$DC,$1D,$1F,$E5,$85,$F1,$BF
	.dc.b	$E9,$36,$D7,$F2,$67,$7C,$DF,$EB,$A6,$C0,$FC,$CC,$F1,$39,$7C,$77
	.dc.b	$3F,$67,$A7,$CB,$CA,$E9,$19,$7B,$46,$5F,$E5,$77,$FE,$89,$F9,$F7
	.dc.b	$CF,$FF,$A9,$DB,$3F,$CE,$9A,$F0,$F4,$35,$32,$34,$FB,$63,$11,$FC
	.dc.b	$E2,$78,$E9,$FE,$7A,$6E,$9E,$FA,$FF,$5B,$7C,$FF,$E3,$6F,$FA,$4B
	.dc.b	$FA,$E6,$F8,$7A,$EC,$91,$82,$1F,$D1,$A7,$93,$7D,$9E,$A4,$7A,$F9
	.dc.b	$35,$D1,$E0,$8F,$1C,$7F,$B2,$7E,$79,$E0,$52,$57,$9B,$AE,$BF,$D5
	.dc.b	$A7,$6F,$F6,$26,$BE,$58,$DE,$DF,$E9,$C9,$CB,$6B,$E1,$CF,$FB,$67
	.dc.b	$E7,$9F,$1B,$49,$87,$F6,$E8,$FF,$7D,$9A,$71,$1A,$7F,$B4,$C6,$BE
	.dc.b	$0B,$A7,$CB,$49,$E5,$1D,$CC,$3C,$23,$0F,$F7,$CE,$E7,$CE,$FF,$BB
	.dc.b	$5B,$19,$8F,$F7,$87,$CE,$FE,$6A,$3F,$71,$1B,$23,$FE,$06,$8D,$4F
	.dc.b	$28,$FE,$FE,$DD,$3F,$D8,$2D,$E4,$FD,$24,$AB,$B1,$FE,$E3,$27,$4F
	.dc.b	$8B,$A7,$C6,$C9,$F0,$39,$7C,$5C,$DE,$8F,$4A,$46,$08,$7F,$84,$AF
	.dc.b	$FB,$CB,$E9,$8F,$7C,$FC,$3E,$79,$3F,$1D,$F3,$E7,$FE,$DA,$F9,$CF
	.dc.b	$CA,$BE,$C8,$DC,$FC,$07,$79,$FB,$AB,$6F,$C0,$D3,$F4,$95,$B3,$B1
	.dc.b	$FE,$30,$9D,$BF,$C2,$63,$77,$DE,$36,$7F,$8B,$27,$6F,$F2,$DB,$F1
	.dc.b	$CE,$E3,$5D,$18,$C3,$C5,$56,$7F,$8B,$5F,$2C,$3B,$D6,$A7,$BE,$1F
	.dc.b	$07,$0E,$11,$DC,$CC,$FB,$E3,$35,$CA,$7D,$58,$7C,$47,$BC,$EA,$39
	.dc.b	$E1,$E1,$71,$3F,$0E,$6B,$F1,$CF,$59,$E0,$3A,$94,$4F,$37,$15,$E3
	.dc.b	$73,$3D,$30,$F8,$C7,$B6,$37,$7E,$49,$D6,$1F,$0E,$7C,$27,$8E,$B8
	.dc.b	$C6,$5E,$51,$98,$C1,$1F,$82,$76,$3D,$18,$79,$DE,$07,$B0,$EF,$3D
	.dc.b	$9C,$35,$F5,$9C,$CE,$26,$0B,$87,$C1,$3E,$14,$F7,$9E,$51,$8B,$F7
	.dc.b	$D7,$6B,$FB,$8F,$5A,$F0,$5F,$19,$F6,$4F,$E0,$CF,$3B,$8B,$F0,$61
	.dc.b	$F8,$19,$F2,$5F,$82,$3E,$74,$75,$AE,$77,$D6,$3C,$44,$38,$C7,$59
	.dc.b	$E3,$1E,$E6,$D7,$58,$F9,$B8,$7B,$1F,$34,$7D,$63,$4F,$C8,$BF,$03
	.dc.b	$AB,$F1,$D7,$1C,$3E,$34,$F9,$CF,$25,$70,$8F,$7C,$7E,$64,$D1,$CC
	.dc.b	$FC,$09,$E7,$3C,$A6,$9E,$AE,$4F,$24,$F8,$70,$F9,$55,$DD,$F7,$DF
	.dc.b	$C3,$AE,$67,$1C,$3E,$BD,$F1,$8F,$21,$FC,$27,$9F,$9F,$3C,$8F,$E2
	.dc.b	$51,$D8,$C3,$FC,$3E,$30,$FF,$17,$E5,$F1,$91,$FE,$35,$F2,$75,$FE
	.dc.b	$2D,$17,$0F,$E5,$53,$F8,$3F,$9B,$FF,$1D,$43,$2F,$E8,$B1,$86,$F8
	.dc.b	$7F,$89,$F9,$BF,$EC,$D8,$FF,$B3,$2B,$4F,$F8,$BD,$A9,$B1,$C2,$FB
	.dc.b	$4E,$EB,$C4,$47,$FD,$C3,$3B,$BF,$EA,$EB,$9F,$C5,$CF,$2B,$71,$E5
	.dc.b	$30,$FF,$B8,$25,$39,$8F,$F7,$6F,$CD,$FF,$6D,$27,$83,$FA,$57,$FC
	.dc.b	$9B,$1F,$EE,$87,$2F,$08,$D7,$85,$FA,$BF,$97,$7C,$D5,$25,$76,$EA
	.dc.b	$AF,$FB,$CA,$76,$FF,$BF,$9A,$F7,$8D,$ED,$F1,$A1,$3A,$8C,$37,$5F
	.dc.b	$EF,$79,$2F,$D5,$D1,$EB,$E9,$98,$FC,$08,$CB,$F0,$A3,$2F,$BE,$3F
	.dc.b	$83,$7C,$C7,$F5,$C2,$73,$E3,$A3,$2C,$AF,$F3,$76,$9E,$7A,$E1,$5F
	.dc.b	$17,$E6,$65,$7B,$DE,$BD,$38,$8F,$89,$16,$F9,$4D,$D7,$FC,$2F,$B2
	.dc.b	$B1,$7F,$E7,$DC,$57,$E8,$E8,$F4,$F4,$4A,$F2,$E3,$FE,$20,$BF,$F4
	.dc.b	$FE,$5E,$7B,$D5,$23,$D5,$7C,$E3,$89,$FE,$8B,$C3,$F2,$8D,$F0,$3F
	.dc.b	$96,$46,$23,$C7,$F1,$B8,$99,$1C,$8F,$3B,$C2,$3F,$9A,$7C,$CB,$B2
	.dc.b	$92,$B8,$5F,$E8,$5E,$53,$B6,$CE,$DF,$56,$9F,$E3,$EA,$E1,$18,$3F
	.dc.b	$D4,$95,$C5,$C0,$71,$97,$E4,$5B,$5C,$FA,$BF,$A5,$7C,$CE,$2F,$FD
	.dc.b	$5C,$D4,$DF,$99,$FF,$2D,$13,$AB,$FA,$93,$FE,$5D,$F9,$87,$CC,$FF
	.dc.b	$55,$6D,$FC,$76,$BE,$66,$E7,$FC,$2B,$D5,$7A,$8F,$A3,$FE,$BD,$F3
	.dc.b	$0F,$82,$FE,$C1,$F3,$4E,$DA,$6A,$26,$7F,$56,$B8,$36,$3F,$C5,$18
	.dc.b	$F6,$F8,$6B,$91,$E3,$5F,$2A,$F8,$BF,$3E,$B8,$3F,$81,$F9,$A7,$3F
	.dc.b	$F7,$36,$C6,$63,$D3,$D5,$96,$46,$4F,$34,$61,$FB,$33,$C9,$B3,$C9
	.dc.b	$FC,$87,$CD,$79,$37,$FE,$18,$DE,$6F,$F4,$1B,$67,$FD,$D1,$F2,$8F
	.dc.b	$9A,$FF,$04,$4E,$64,$D3,$13,$B8,$EF,$BF,$02,$3F,$A1,$8D,$3F,$90
	.dc.b	$E8,$7C,$D7,$F8,$55,$B1,$FA,$76,$BF,$96,$F9,$77,$C5,$6E,$A4,$ED
	.dc.b	$FF,$0F,$4E,$A3,$04,$3C,$11,$A9,$FF,$12,$7C,$D7,$F5,$82,$79,$DC
	.dc.b	$BC,$9C,$B2,$FF,$75,$3F,$D7,$B0,$FE,$61,$18,$9D,$3F,$AC,$9D,$57
	.dc.b	$F8,$C7,$E6,$BE,$B7,$07,$F3,$D6,$F0,$7F,$AD,$1B,$61,$8B,$E6,$B9
	.dc.b	$71,$49,$FE,$19,$5C,$1D,$CE,$5F,$DE,$C6,$4D,$57,$38,$E3,$CF,$E5
	.dc.b	$23,$FC,$5E,$3F,$50,$E8,$87,$48,$F9,$FD,$3F,$E4,$DF,$9A,$71,$52
	.dc.b	$57,$F2,$1F,$9A,$FE,$6F,$AF,$B7,$E6,$5F,$31,$FE,$D4,$C3,$8D,$FF
	.dc.b	$27,$6B,$FC,$59,$B5,$FD,$96,$F9,$CA,$F9,$BF,$26,$33,$18,$85,$74
	.dc.b	$8C,$36,$FF,$DC,$B5,$C1,$BE,$37,$F4,$5B,$ED,$1E,$7C,$7F,$DD,$78
	.dc.b	$72,$6C,$70,$C7,$DC,$DA,$E2,$7D,$DC,$7C,$0D,$AE,$2F,$A7,$0F,$06
	.dc.b	$1E,$9C,$7C,$93,$B3,$A9,$4E,$A1,$DB,$67,$9E,$1E,$98,$E2,$7C,$4B
	.dc.b	$F0,$BB,$C7,$37,$CB,$1F,$3C,$EC,$CE,$1D,$1F,$45,$74,$9D,$9C,$2B
	.dc.b	$C6,$7D,$89,$FC,$F1,$C0,$A5,$2B,$B7,$EA,$F8,$EF,$9A,$DC,$7D,$F8
	.dc.b	$C3,$89,$FE,$D9,$3A,$8F,$89,$3A,$71,$F3,$59,$7F,$60,$D4,$D2,$3F
	.dc.b	$3B,$E5,$DF,$17,$FE,$5C,$DB,$F1,$D3,$C8,$D4,$67,$6F,$3D,$FD,$0B
	.dc.b	$E3,$CB,$F4,$AF,$86,$1F,$97,$9C,$C7,$DC,$33,$7F,$22,$33,$3E,$09
	.dc.b	$F3,$9E,$0A,$E8,$E2,$3A,$61,$E3,$8F,$9F,$87,$8E,$31,$1F,$52,$31
	.dc.b	$5C,$9B,$1E,$33,$A6,$1E,$3A,$E3,$1E,$63,$F8,$D6,$1E,$69,$F8,$31
	.dc.b	$E5,$39,$57,$4C,$3C,$B8,$7F,$89,$EF,$E0,$EB,$FE,$63,$BE,$82,$33
	.dc.b	$3E,$13,$87,$07,$FA,$3A,$BF,$CD,$53,$AA,$F1,$CE,$6B,$E5,$74,$59
	.dc.b	$FF,$21,$7C,$D7,$F4,$E9,$F1,$4E,$70,$4E,$6B,$FC,$53,$82,$B5,$39
	.dc.b	$AF,$0C,$63,$AB,$B5,$FA,$5F,$07,$CC,$3E,$69,$F3,$5A,$CB,$96,$7F
	.dc.b	$D0,$EB,$53,$9B,$E3,$7C,$2B,$E4,$46,$27,$51,$FE,$5A,$AE,$2E,$8C
	.dc.b	$57,$1C,$3F,$C8,$7E,$27,$F8,$76,$7F,$CC,$23,$E5,$FE,$54,$A7,$4F
	.dc.b	$F9,$36,$75,$C5,$EE,$9C,$EB,$FE,$C0,$9C,$9F,$E6,$38,$CB,$FD,$7C
	.dc.b	$CD,$78,$BA,$7E,$A4,$62,$BF,$D4,$95,$C2,$BF,$DB,$18,$79,$3E,$69
	.dc.b	$EC,$F0,$8B,$87,$F9,$C6,$6D,$79,$E6,$E1,$FE,$AF,$9B,$5E,$39,$B5
	.dc.b	$FD,$CE,$6C,$FC,$3B,$FB,$71,$8A,$DD,$23,$13,$F8,$0D,$9F,$39,$4F
	.dc.b	$C8,$79,$2F,$87,$FE,$FD,$97,$FA,$1E,$B6,$DA,$E0,$FF,$A5,$2B,$87
	.dc.b	$83,$8C,$BF,$CF,$77,$C1,$FE,$4A,$D7,$F9,$6B,$5F,$E6,$2D,$7F,$C2
	.dc.b	$1F,$84,$FF,$9A,$6F,$FD,$30,$D8,$FF,$2E,$62,$F7,$FF,$37,$39,$DB
	.dc.b	$FD,$B0,$DB,$FE,$7A,$DB,$65,$EF,$BE,$3E,$4C,$BF,$E6,$D7,$DE,$FF
	.dc.b	$A1,$B5,$E7,$3F,$C0,$DF,$F1,$37,$DF,$7F,$A0,$6B,$F9,$F3,$63,$FA
	.dc.b	$2B,$5F,$E9,$AD,$7F,$A8,$FC,$DF,$57,$FC,$60,$D4,$D3,$1C,$DF,$EE
	.dc.b	$FA,$FE,$98,$D9,$FE,$A6,$D7,$FA,$DB,$5F,$EB,$FF,$6E,$FA,$DF,$54
	.dc.b	$23,$15,$FD,$5D,$B1,$FD,$79,$AF,$F6,$6F,$AD,$7B,$3C,$9B,$5F,$D7
	.dc.b	$DB,$3F,$D9,$9A,$FF,$6E,$F9,$77,$CB,$FE,$EE,$3F,$B5,$C6,$3A,$7F
	.dc.b	$B4,$2F,$8B,$69,$8B,$E1,$7F,$DE,$3E,$0B,$FC,$EF,$C9,$FE,$B5,$F6
	.dc.b	$1F,$37,$FE,$A8,$3F,$BF,$C3,$5F,$F0,$0D,$70,$6F,$A9,$FD,$CE,$3F
	.dc.b	$D0,$38,$75,$8C,$11,$FF,$76,$47,$17,$FC,$D0,$D5,$F8,$0F,$F7,$67
	.dc.b	$73,$E2,$AE,$F1,$83,$FD,$ED,$3B,$7F,$DC,$B5,$CC,$95,$C6,$3F,$CE
	.dc.b	$D3,$BB,$F6,$46,$3C,$1D,$2B,$E2,$F6,$8C,$D7,$1C,$3C,$73,$E0,$C3
	.dc.b	$35,$23,$CC,$64,$E4,$43,$FC,$47,$5E,$97,$FC,$56,$D7,$FC,$5E,$D7
	.dc.b	$CC,$71,$C3,$D2,$FC,$89,$F4,$FE,$00,$F9,$6F,$14,$AB,$77,$E9,$8E
	.dc.b	$66,$71,$FF,$1E,$DF,$18,$ED,$1E,$2C,$39,$DF,$82,$B5,$5C,$4F,$F5
	.dc.b	$13,$88,$EF,$3C,$2B,$AE,$1E,$63,$17,$E2,$AE,$91,$98,$F0,$99,$9E
	.dc.b	$2F,$FA,$BE,$76,$FF,$91,$EF,$C6,$FF,$95,$1A,$F0,$3F,$CA,$2F,$96
	.dc.b	$71,$3C,$A7,$D4,$3E,$FD,$F4,$BE,$38,$CD,$F1,$7D,$F8,$7E,$22,$B7
	.dc.b	$1E,$5C,$3A,$91,$D9,$93,$FD,$91,$1C,$1F,$0B,$70,$F2,$12,$7F,$CE
	.dc.b	$31,$96,$E1,$E5,$9D,$57,$48,$F2,$19,$9F,$14,$7B,$63,$2F,$F9,$BE
	.dc.b	$BC,$4F,$F9,$F1,$AF,$63,$51,$C1,$F3,$BF,$A8,$FA,$B7,$C7,$E7,$0F
	.dc.b	$CA,$DF,$9E,$FC,$87,$C1,$15,$C1,$C4,$6D,$F3,$18,$BE,$D7,$23,$04
	.dc.b	$3E,$CD,$73,$6C,$78,$F0,$F5,$46,$0F,$14,$EF,$0F,$3D,$EE,$7F,$DF
	.dc.b	$93,$9A,$FF,$50,$B6,$3D,$67,$C2,$39,$DF,$95,$C1,$2F,$F6,$F3,$C6
	.dc.b	$3E,$0C,$F7,$7D,$E7,$94,$93,$E4,$7D,$CF,$28,$EC,$E6,$3B,$57,$9A
	.dc.b	$72,$FC,$09,$D1,$C6,$7C,$E7,$48,$D1,$C5,$C9,$D0,$F7,$C7,$17,$CB
	.dc.b	$1C,$1C,$BC,$5C,$98,$21,$F0,$0E,$B1,$E4,$8E,$51,$93,$89,$F1,$23
	.dc.b	$4F,$D4,$9F,$6B,$D3,$0F,$C4,$5F,$CD,$3E,$73,$F8,$B3,$33,$C2,$7A
	.dc.b	$5F,$82,$B8,$4F,$D3,$AD,$D7,$9A,$BC,$B1,$88,$F2,$BD,$70,$E9,$7E
	.dc.b	$07,$EE,$3D,$27,$E7,$1F,$09,$C3,$F3,$4F,$19,$ED,$3C,$07,$B6,$3A
	.dc.b	$DF,$95,$B1,$FE,$E7,$9F,$D0,$3F,$EF,$46,$BF,$4C,$C4,$FF,$BA,$63
	.dc.b	$31,$DC,$FF,$7A,$B8,$7E,$96,$1F,$7C,$98,$7E,$8B,$EC,$7C,$33,$F5
	.dc.b	$A9,$18,$AD,$CF,$EA,$A3,$15,$F8,$59,$FD,$1C,$62,$7C,$51,$87,$E0
	.dc.b	$9E,$29,$E0,$78,$5C,$D7,$CA,$9E,$11,$E3,$BF,$A7,$7D,$63,$91,$E3
	.dc.b	$38,$C6,$9A,$74,$8E,$15,$F8,$B6,$CF,$F1,$F6,$CF,$F0,$A6,$BF,$C4
	.dc.b	$9B,$5E,$43,$E1,$C7,$88,$95,$F2,$1F,$09,$F2,$E7,$8D,$FB,$E3,$D8
	.dc.b	$79,$4D,$D7,$4A,$E5,$1A,$39,$1A,$9F,$92,$FC,$13,$53,$C6,$BB,$C6
	.dc.b	$63,$A5,$FC,$58,$F6,$1C,$CD,$57,$96,$77,$1F,$0C,$87,$68,$D1,$EB
	.dc.b	$7F,$2D,$5F,$93,$23,$F9,$B7,$E7,$3C,$A7,$F9,$4F,$06,$23,$C6,$48
	.dc.b	$CB,$BA,$EE,$F9,$8F,$E7,$73,$FD,$5C,$E5,$3F,$56,$7F,$9E,$C6,$67
	.dc.b	$91,$C4,$FE,$AE,$E5,$F0,$47,$67,$27,$F4,$C8,$FD,$71,$1F,$DB,$35
	.dc.b	$FE,$AE,$F3,$9F,$F1,$DF,$C3,$BD,$3C,$5F,$92,$FA,$1F,$DA,$67,$4E
	.dc.b	$0E,$65,$AE,$75,$CA,$3E,$09,$9A,$32,$B9,$BC,$89,$5F,$31,$FF,$12
	.dc.b	$C6,$A3,$04,$9F,$34,$EA,$BC,$0F,$C4,$39,$C7,$81,$57,$38,$C3,$63
	.dc.b	$CE,$70,$3C,$D3,$FE,$76,$8D,$47,$5A,$F1,$3A,$3E,$34,$F8,$B2,$FF
	.dc.b	$4E,$C6,$08,$75,$39,$5F,$8A,$7B,$1C,$70,$B3,$C2,$FB,$BC,$0F,$01
	.dc.b	$F0,$4F,$71,$CC,$A6,$8C,$E1,$F4,$1B,$5D,$9E,$47,$F9,$53,$E1,$FF
	.dc.b	$D9,$2F,$F8,$A5,$7D,$6F,$48,$FB,$71,$87,$0D,$7C,$2D,$7C,$3D,$DC
	.dc.b	$5F,$A6,$DB,$FF,$70,$47,$56,$BE,$AF,$85,$7A,$7C,$5B,$1E,$36,$BB
	.dc.b	$8E,$DF,$B9,$3D,$7A,$49,$F5,$46,$5B,$F5,$1F,$EB,$48,$C4,$7F,$9D
	.dc.b	$9B,$AF,$FA,$89,$4E,$CE,$E4,$9E,$67,$6F,$81,$FF,$62,$79,$4F,$D1
	.dc.b	$BF,$DC,$B8,$7F,$B1,$23,$1E,$1F,$86,$A5,$FF,$9E,$67,$2D,$BE,$3C
	.dc.b	$7F,$9D,$C9,$7C,$2B,$F0,$31,$87,$C4,$43,$3F,$73,$E4,$B7,$C7,$D3
	.dc.b	$E1,$44,$9F,$04,$65,$FF,$36,$52,$75,$1D,$0F,$9A,$FD,$23,$F0,$1F
	.dc.b	$0D,$FE,$F1,$27,$81,$FE,$E2,$F8,$37,$A3,$FE,$E4,$9C,$DF,$F9,$E5
	.dc.b	$FF,$7C,$4E,$A3,$15,$B2,$3F,$E7,$F8,$D3,$FE,$AE,$F8,$67,$B9,$FE
	.dc.b	$F0,$6C,$7F,$A1,$5A,$FF,$AD,$27,$E4,$3C,$67,$81,$1E,$DA,$D7,$47
	.dc.b	$38,$F6,$47,$FB,$FB,$E1,$5F,$0D,$49,$5E,$7F,$5A,$BF,$EC,$49,$DB
	.dc.b	$FE,$99,$6B,$E5,$8D,$F2,$7F,$BD,$27,$2D,$AF,$99,$3E,$1F,$6E,$5F
	.dc.b	$37,$E8,$74,$62,$3C,$11,$6F,$FD,$FF,$9F,$EB,$1F,$F8,$03,$F0,$8F
	.dc.b	$86,$FE,$59,$18,$6F,$37,$F2,$F9,$D6,$1F,$EB,$1F,$85,$7C,$33,$0F
	.dc.b	$E8,$7F,$09,$FA,$EB,$DD,$FF,$B9,$1B,$3D,$63,$95,$71,$55,$FB,$6B
	.dc.b	$EB,$5D,$3B,$7C,$59,$7F,$41,$AE,$CD,$8F,$E8,$95,$C2,$F9,$C6,$30
	.dc.b	$ED,$7C,$27,$A1,$FD,$6B,$1F,$3C,$78,$67,$75,$C7,$1D,$DA,$F7,$3E
	.dc.b	$48,$CE,$1E,$1C,$39,$5F,$0B,$DD,$77,$C7,$D0,$FF,$6A,$BE,$18,$77
	.dc.b	$8F,$49,$B3,$94,$F0,$53,$C1,$E5,$8F,$A1,$E0,$72,$3D,$13,$C6,$78
	.dc.b	$47,$29,$E0,$E8,$EE,$7B,$1C,$C7,$CD,$8F,$EA,$AD,$7D,$06,$23,$EE
	.dc.b	$3A,$F0,$DF,$0D,$FD,$3E,$BF,$C2,$CD,$8C,$CF,$F5,$E9,$CB,$FD,$37
	.dc.b	$E1,$BF,$A7,$26,$91,$D1,$EF,$F8,$57,$9E,$DF,$EC,$CF,$F8,$62,$BE
	.dc.b	$DD,$6A,$73,$3D,$4E,$55,$C5,$F9,$47,$13,$31,$C2,$B9,$CE,$F7,$FE
	.dc.b	$9F,$5B,$7F,$A8,$DF,$78,$E6,$77,$8C,$CF,$1D,$7F,$BF,$DF,$4C,$3B
	.dc.b	$5F,$80,$E3,$19,$69,$9B,$F0,$BD,$6B,$FC,$51,$87,$86,$73,$5C,$5B
	.dc.b	$5C,$6B,$78,$7C,$1D,$FE,$14,$B0,$F5,$46,$23,$DD,$5C,$31,$FF,$22
	.dc.b	$CF,$0A,$DC,$78,$B8,$BA,$3B,$BF,$E1,$4E,$4E,$27,$A3,$9F,$71,$FC
	.dc.b	$3F,$E1,$BF,$A6,$CE,$A3,$FC,$A7,$7F,$D1,$23,$18,$2B,$75,$D8,$DD
	.dc.b	$F4,$9D,$76,$76,$9F,$74,$E6,$B8,$D7,$AE,$BC,$7E,$AB,$E1,$5F,$0D
	.dc.b	$C7,$3D,$E7,$FE,$66,$AD,$4E,$6F,$8D,$F0,$AF,$56,$7F,$36,$73,$5B
	.dc.b	$9F,$75,$F2,$77,$87,$6B,$4E,$5F,$70,$F1,$5D,$3F,$2A,$DF,$0D,$F5
	.dc.b	$2B,$FC,$F7,$5A,$9F,$98,$6B,$87,$E2,$46,$67,$E5,$4E,$4F,$4C,$67
	.dc.b	$5F,$D4,$46,$67,$C7,$39,$8F,$F4,$49,$99,$FC,$6C,$66,$7F,$71,$7E
	.dc.b	$55,$F0,$CF,$86,$F5,$4D,$C3,$E4,$CD,$AF,$2C,$DC,$3F,$33,$36,$BC
	.dc.b	$33,$6B,$FD,$37,$36,$7E,$2D,$FE,$0E,$31,$5B,$AF,$1C,$F8,$8F,$C4
	.dc.b	$6C,$F8,$6A,$7E,$63,$C9,$7C,$33,$C0,$CE,$7F,$80,$BE,$0F,$F9,$CD
	.dc.b	$B8,$FF,$A0,$1A,$FF,$A1,$1A,$FF,$A8,$DD,$E9,$FE,$51,$BF,$F1,$D5
	.dc.b	$75,$C7,$C5,$18,$BF,$C2,$FC,$2B,$E1,$7F,$40,$DB,$FF,$2B,$53,$E1
	.dc.b	$B3,$52,$7F,$D1,$15,$BB,$FF,$4A,$B5,$FF,$4C,$B5,$FF,$5C,$3B,$C7
	.dc.b	$FC,$CD,$EF,$3E,$03,$A5,$7E,$D7,$64,$E7,$0F,$D2,$FC,$2F,$3A,$FF
	.dc.b	$68,$B5,$34,$C7,$37,$F9,$AA,$7F,$D3,$CD,$9F,$F5,$5B,$5F,$F5,$7B
	.dc.b	$5F,$F6,$75,$BE,$23,$F8,$F4,$7F,$B8,$23,$15,$FE,$A9,$6C,$7F,$AD
	.dc.b	$9A,$FF,$AE,$BE,$11,$F0,$BC,$A1,$18,$AF,$F5,$93,$63,$FD,$7E,$D7
	.dc.b	$FD,$87,$F0,$AF,$85,$FD,$E4,$7F,$B1,$23,$1D,$3F,$EB,$B7,$E6,$67
	.dc.b	$6B,$75,$FE,$CC,$F9,$DF,$C9,$BF,$EC,$16,$BF,$EC,$E6,$BB,$E4,$FF
	.dc.b	$5F,$7C,$23,$E1,$3F,$C8,$B2,$1F,$0B,$FD,$C1,$FE,$36,$D7,$F8,$9C
	.dc.b	$F1,$7F,$8F,$B5,$FF,$6E,$B5,$FF,$70,$B5,$FF,$80,$0D,$7C,$0F,$1E
	.dc.b	$4F,$EA,$AD,$DB,$77,$E6,$A7,$C2,$BE,$17,$17,$B9,$E7,$58,$38,$56
	.dc.b	$E3,$CF,$19,$69,$FC,$E2,$34,$FF,$30,$6E,$1E,$49,$E1,$39,$BE,$A6
	.dc.b	$27,$51,$CE,$32,$F6,$3E,$4C,$E1,$E4,$7F,$BE,$6B,$C2,$FD,$33,$EA
	.dc.b	$38,$70,$D7,$33,$CE,$BF,$BE,$CF,$A3,$0E,$5F,$06,$F8,$5D,$47,$2B
	.dc.b	$B8,$7E,$06,$B4,$FA,$48,$79,$8F,$EA,$6E,$9B,$5C,$9F,$AB,$3F,$5E
	.dc.b	$75,$87,$F8,$41,$B7,$E1,$AE,$B3,$9B,$E0,$73,$C3,$D5,$3E,$58,$CB
	.dc.b	$F1,$4F,$6B,$5A,$51,$7E,$77,$83,$E2,$7A,$1F,$DE,$1C,$3C,$3E,$17
	.dc.b	$BB,$EE,$35,$3E,$D8,$CD,$FA,$67,$27,$38,$D3,$83,$91,$F2,$9C,$B8
	.dc.b	$8E,$4E,$4E,$47,$17,$27,$CB,$8E,$31,$EC,$7E,$03,$E9,$3E,$15,$23
	.dc.b	$2D,$8E,$A7,$73,$EE,$FC,$2F,$92,$BB,$46,$9F,$B5,$19,$C7,$F1,$B7
	.dc.b	$6B,$F1,$53,$F8,$13,$0F,$8C,$EA,$7C,$F7,$37,$E3,$8C,$92,$FC,$2F
	.dc.b	$95,$F0,$BC,$4F,$25,$F8,$9C,$1C,$5C,$3C,$5F,$D1,$D7,$96,$38,$9F
	.dc.b	$A1,$BC,$3C,$4C,$1A,$3F,$17,$1A,$AF,$29,$2B,$AC,$F8,$A3,$0F,$23
	.dc.b	$A1,$C8,$E9,$7F,$06,$3D,$B1,$A8,$C4,$FA,$A7,$24,$70,$49,$FA,$27
	.dc.b	$85,$C3,$F4,$63,$83,$A3,$E2,$1E,$57,$2F,$94,$C9,$C6,$3D,$AF,$96
	.dc.b	$FE,$CD,$FC,$39,$EF,$7E,$8A,$ED,$7D,$23,$11,$CC,$CC,$F4,$9F,$8F
	.dc.b	$19,$3C,$4F,$1B,$EF,$1E,$D3,$D1,$3E,$0B,$F1,$46,$08,$72,$8E,$C7
	.dc.b	$9A,$77,$19,$37,$7E,$61,$5A,$71,$36,$3F,$32,$D9,$FD,$7C,$7E,$B8
	.dc.b	$8E,$08,$7F,$17,$9E,$AE,$0F,$E2,$B1,$CD,$FF,$40,$FC,$13,$D1,$C6
	.dc.b	$3F,$5A,$31,$3F,$0C,$F5,$11,$CB,$F8,$BA,$F7,$C6,$27,$EA,$46,$2B
	.dc.b	$AE,$3F,$1A,$FF,$C0,$97,$B8,$C4,$76,$C3,$D2,$F9,$AB,$8B,$F7,$0E
	.dc.b	$2E,$1F,$23,$63,$8E,$1F,$ED,$4B,$ED,$5C,$4E,$D8,$7F,$80,$63,$31
	.dc.b	$83,$E0,$38,$AE,$78,$79,$9B,$3E,$23,$E3,$38,$7C,$66,$2B,$E1,$34
	.dc.b	$EB,$87,$9E,$3C,$4D,$8F,$F5,$FD,$F8,$67,$DC,$D8,$E4,$7C,$52,$E1
	.dc.b	$F6,$6B,$94,$64,$FA,$C7,$80,$DE,$1F,$42,$BA,$46,$8E,$15,$F6,$9E
	.dc.b	$06,$5E,$CF,$23,$E6,$47,$28,$E1,$1C,$A7,$E4,$5F,$C2,$8C,$1F,$4F
	.dc.b	$0F,$91,$7F,$48,$E5,$5E,$5C,$3E,$2D,$77,$13,$B8,$F8,$A6,$A3,$CD
	.dc.b	$3F,$E1,$F9,$D4,$71,$6B,$FD,$F5,$FC,$DB,$FE,$01,$6B,$C6,$7F,$B0
	.dc.b	$C7,$0C,$3F,$19,$1F,$E4,$9F,$84,$FF,$3C,$D7,$D8,$8F,$14,$EA,$3F
	.dc.b	$DF,$33,$A7,$E0,$9E,$A9,$DD,$FF,$8D,$67,$55,$C6,$B7,$7F,$96,$9C
	.dc.b	$D7,$96,$7C,$04,$3F,$DF,$B1,$A7,$F2,$C7,$2B,$F4,$74,$48,$E1,$87
	.dc.b	$71,$7E,$39,$D4,$F1,$3D,$6E,$27,$84,$F8,$E7,$C3,$39,$8C,$4E,$A7
	.dc.b	$FD,$09,$7E,$54,$DC,$3E,$75,$78,$67,$B9,$E6,$79,$99,$7B,$99,$3D
	.dc.b	$F5,$EA,$7B,$DF,$A9,$FF,$1F,$5F,$D3,$6C,$7F,$87,$A3,$11,$D6,$79
	.dc.b	$3E,$69,$F6,$3F,$E9,$DF,$8D,$7C,$77,$F0,$BF,$86,$FC,$39,$27,$FC
	.dc.b	$91,$92,$3F,$C5,$7F,$0C,$F4,$A9,$2B,$FC,$9D,$19,$6C,$7F,$A5,$E3
	.dc.b	$10,$CB,$CB,$18,$7C,$FF,$1B,$FE,$F9,$9F,$F5,$94,$7A,$BE,$11,$F0
	.dc.b	$FF,$EA,$15,$EB,$53,$11,$2B,$FC,$C1,$3C,$1B,$1F,$E4,$48,$C4,$7F
	.dc.b	$99,$DB,$5D,$E7,$94,$F9,$F8,$3F,$D2,$5E,$9F,$FB,$32,$BC,$9F,$03
	.dc.b	$FE,$C8,$F1,$1F,$0F,$FE,$D4,$9F,$B5,$18,$F1,$3D,$15,$31,$F2,$BE
	.dc.b	$32,$38,$3E,$0E,$3F,$EA,$1C,$D3,$C6,$7C,$82,$FC,$3F,$5D,$F1,$7E
	.dc.b	$1F,$C9,$9B,$E2,$29,$9B,$EA,$7F,$B7,$F1,$57,$C5,$24,$7C,$A6,$BE
	.dc.b	$01,$1C,$1E,$66,$23,$C3,$F3,$BF,$ED,$7F,$89,$7C,$4D,$25,$70,$F8
	.dc.b	$67,$AB,$88,$ED,$5C,$1E,$AD,$7C,$53,$C1,$CD,$71,$E1,$FD,$0D,$6D
	.dc.b	$FD,$1D,$6E,$BD,$5F,$12,$F8,$0F,$F7,$63,$63,$D6,$D7,$BF,$0E,$A6
	.dc.b	$46,$9F,$F7,$A5,$7D,$68,$C1,$0E,$F1,$A9,$FF,$7D,$7C,$4B,$E2,$69
	.dc.b	$2B,$E2,$7B,$55,$FA,$13,$B7,$E7,$B5,$F8,$71,$BC,$BF,$80,$CE,$5B
	.dc.b	$5E,$69,$F0,$FB,$F2,$F9,$BE,$AE,$DC,$47,$82,$2D,$FC,$A9,$FF,$7A
	.dc.b	$3F,$C3,$BE,$25,$F1,$38,$E9,$FF,$7C,$46,$30,$FC,$2B,$5F,$C5,$7C
	.dc.b	$3B,$E2,$7F,$82,$43,$E1,$FF,$7A,$85,$71,$C7,$C1,$8F,$79,$D4,$E7
	.dc.b	$1E,$10,$C7,$B5,$F3,$C3,$AC,$E7,$1F,$0C,$63,$0F,$33,$71,$E6,$F0
	.dc.b	$C3,$9D,$F9,$FC,$0E,$F1,$89,$E1,$3D,$E7,$E1,$D7,$27,$78,$78,$1B
	.dc.b	$8E,$EB,$D7,$5B,$69,$93,$D1,$17,$0F,$05,$76,$9D,$63,$EA,$8E,$B1
	.dc.b	$E7,$3E,$64,$F0,$7A,$99,$3E,$7D,$F8,$5E,$E6,$63,$83,$A3,$BC,$6E
	.dc.b	$32,$6A,$BC,$53,$DB,$1F,$95,$8F,$C1,$8F,$1B,$F8,$27,$EC,$9E,$02
	.dc.b	$D7,$D0,$6C,$6D,$FA,$3E,$4B,$E2,$78,$C7,$BA,$30,$E2,$BF,$D1,$53
	.dc.b	$A8,$FC,$34,$E9,$C7,$C4,$E5,$FC,$93,$53,$48,$FC,$9F,$87,$7C,$3F
	.dc.b	$F7,$BC,$BF,$1F,$3C,$8D,$46,$75,$F3,$DF,$DD,$BE,$3B,$FD,$FB,$E1
	.dc.b	$87,$E6,$27,$31,$F5,$CC,$DF,$E8,$23,$33,$E0,$9F,$39,$E0,$AE,$8E
	.dc.b	$23,$A6,$1E,$38,$FA,$38,$78,$E3,$11,$FA,$C8,$C5,$72,$6C,$78,$CE
	.dc.b	$98,$78,$EB,$8C,$79,$8F,$DE,$61,$E6,$9F,$83,$1E,$53,$95,$74,$C3
	.dc.b	$CB,$87,$F4,$EB,$F8,$3B,$FF,$87,$AF,$A0,$8C,$CF,$84,$E1,$C1,$FE
	.dc.b	$2D,$AF,$F1,$2C,$EA,$BC,$73,$9A,$F9,$5C,$D6,$7F,$C3,$1F,$13,$FD
	.dc.b	$5A,$7C,$53,$9C,$13,$9A,$FE,$CD,$87,$F4,$DB,$FE,$DF,$5A,$EC,$E9
	.dc.b	$5E,$7C,$39,$61,$C6,$BC,$B5,$E1,$F6,$DF,$12,$F1,$F8,$E7,$BC,$FF
	.dc.b	$93,$56,$A7,$37,$C6,$F8,$57,$C9,$8C,$4E,$A3,$FC,$07,$5C,$5D,$18
	.dc.b	$AE,$39,$FF,$80,$7C,$77,$47,$C5,$A7,$C4,$7C,$BA,$56,$A7,$FC,$1E
	.dc.b	$6B,$8F,$E0,$C6,$67,$FC,$37,$39,$3F,$C1,$71,$9D,$7F,$CC,$91,$99
	.dc.b	$F1,$61,$E1,$C3,$C1,$7B,$AF,$E9,$06,$EF,$FC,$C7,$F1,$2F,$89,$E0
	.dc.b	$7D,$27,$13,$E6,$BB,$8F,$F8,$7E,$6D,$F8,$A6,$D7,$F4,$E9,$B5,$FE
	.dc.b	$92,$9A,$A3,$33,$F1,$AF,$C0,$78,$9F,$F1,$4E,$CF,$89,$FC,$E5,$FF
	.dc.b	$61,$8C,$4F,$F1,$76,$C7,$F1,$E6,$BF,$CA,$1C,$E3,$FE,$52,$F7,$BF
	.dc.b	$C7,$B5,$DA,$FF,$39,$E6,$27,$38,$7F,$8F,$70,$E3,$7F,$DC,$6B,$5E
	.dc.b	$47,$E6,$27,$35,$FE,$51,$BE,$38,$FF,$A4,$3E,$1D,$F1,$1E,$AA,$46
	.dc.b	$32,$FE,$60,$D8,$FE,$6E,$D7,$FA,$0B,$9A,$FF,$35,$FC,$3B,$E2,$3F
	.dc.b	$53,$39,$BF,$34,$E5,$FF,$29,$C6,$6F,$FC,$CD,$39,$7F,$CB,$51,$9D
	.dc.b	$13,$9C,$3F,$D5,$3C,$BF,$E7,$0C,$3F,$A3,$37,$0F,$E9,$AD,$7F,$AB
	.dc.b	$B5,$FF,$08,$DB,$E1,$FE,$13,$6B,$FA,$83,$67,$FA,$BB,$5F,$EC,$3F
	.dc.b	$53,$F5,$67,$FA,$B3,$63,$FA,$EB,$5D,$EB,$F3,$FE,$1D,$F0,$FF,$B2
	.dc.b	$9F,$EC,$71,$8E,$6F,$EC,$0F,$CB,$CE,$D6,$EB,$FB,$8F,$C3,$FE,$15
	.dc.b	$FE,$CC,$D7,$FB,$6B,$5D,$E1,$FD,$9F,$81,$F2,$D4,$93,$FD,$BE,$32
	.dc.b	$FF,$71,$F8,$07,$C1,$FF,$8E,$E1,$F0,$8F,$85,$C3,$63,$FD,$1F,$5C
	.dc.b	$1B,$5F,$DD,$E3,$11,$FE,$02,$6B,$FE,$0C,$6B,$FE,$46,$77,$C7,$FE
	.dc.b	$E8,$6F,$0E,$6B,$C9,$0F,$85,$B1,$2F,$C2,$79,$23,$0F,$FA,$9E,$79
	.dc.b	$BF,$E5,$B6,$A7,$A3,$C4,$E6,$71,$9F,$0C,$C8,$CD,$F9,$E7,$11,$DE
	.dc.b	$7C,$33,$98,$E8,$64,$FA,$11,$F1,$C8,$E0,$DC,$7C,$38,$F8,$3F,$0B
	.dc.b	$DA,$31,$3A,$C3,$E5,$4E,$5F,$99,$56,$7E,$04,$F2,$78,$1F,$0E,$30
	.dc.b	$FC,$78,$E0,$FC,$58,$CD,$7D,$59,$CE,$1F,$22,$B8,$1F,$84,$8F,$C1
	.dc.b	$11,$FC,$73,$5D,$92,$57,$FE,$88,$F8,$17,$97,$E0,$8F,$6D,$75,$3F
	.dc.b	$DB,$F3,$A7,$07,$38,$DB,$FE,$E7,$8D,$39,$3E,$19,$C2,$BC,$2F,$FA
	.dc.b	$7E,$37,$1E,$F8,$D4,$F4,$8D,$C2,$77,$39,$24,$76,$3C,$55,$EB,$71
	.dc.b	$5D,$5C,$9E,$C8,$D3,$F9,$5B,$FD,$A5,$F6,$8F,$34,$EB,$05,$6A,$24
	.dc.b	$FA,$0E,$F8,$7C,$D8,$CC,$7C,$33,$B1,$E1,$21,$9A,$EA,$E3,$0F,$83
	.dc.b	$87,$8E,$38,$9C,$DC,$19,$9E,$31,$9A,$F0,$D7,$C1,$25,$F9,$67,$B3
	.dc.b	$F2,$9E,$B1,$E0,$9E,$33,$9A,$EB,$39,$AE,$27,$C0,$21,$F5,$EB,$8B
	.dc.b	$89,$E3,$5E,$18,$FA,$67,$C1,$C3,$6F,$13,$E6,$1E,$48,$E1,$7E,$18
	.dc.b	$E0,$78,$8E,$06,$0E,$AE,$CF,$82,$F9,$B0,$F9,$38,$7D,$DB,$F0,$C7
	.dc.b	$CB,$7F,$0A,$EA,$7C,$78,$7A,$9F,$C2,$9E,$17,$0F,$9C,$C9,$E2,$C3
	.dc.b	$E5,$57,$6B,$F0,$4F,$A6,$FC,$0F,$C0,$7C,$0F,$AF,$0F,$55,$FC,$23
	.dc.b	$9C,$71,$AE,$6F,$73,$C3,$3F,$01,$F2,$1F,$19,$E6,$FD,$5B,$F2,$3F
	.dc.b	$EC,$7B,$FD,$DB,$FE,$D6,$6B,$F6,$4C,$57,$FB,$26,$33,$1F,$00,$C9
	.dc.b	$CE,$3A,$AA,$FD,$EF,$C2,$BE,$17,$31,$E4,$94,$E5,$F7,$1F,$7D,$C3
	.dc.b	$EF,$3D,$98,$FC,$54,$72,$AE,$B1,$98,$C1,$D0,$E5,$3D,$2B,$C5,$87
	.dc.b	$D2,$7C,$F7,$CA,$BE,$6D,$F6,$79,$9E,$3A,$E1,$19,$BE,$8F,$86,$33
	.dc.b	$3E,$69,$CB,$D0,$87,$B0,$D5,$78,$5C,$1E,$0B,$F3,$E1,$FA,$AA,$E7
	.dc.b	$3A,$BE,$EE,$F0,$F1,$56,$F0,$E9,$5F,$C3,$EF,$8B,$C0,$C4,$6A,$7F
	.dc.b	$C3,$53,$A7,$C8,$DB,$F9,$98,$7C,$86,$E1,$F4,$67,$51,$F1,$EB,$9D
	.dc.b	$F5,$8F,$B3,$7C,$5D,$1F,$81,$8D,$3F,$8B,$BF,$9D,$18,$9E,$07,$C4
	.dc.b	$9E,$07,$88,$F9,$B7,$E2,$8D,$47,$33,$E7,$9E,$2A,$EA,$F1,$3E,$51
	.dc.b	$A3,$CA,$71,$32,$4C,$3E,$8D,$7D,$43,$C1,$3F,$0E,$FE,$0D,$75,$9D
	.dc.b	$47,$CC,$33,$1D,$AF,$DB,$5E,$48,$FC,$91,$E3,$2D,$7E,$99,$B5,$A7
	.dc.b	$85,$72,$7D,$4F,$F7,$0F,$85,$7C,$2F,$82,$35,$3E,$C9,$CB,$F1,$4E
	.dc.b	$38,$74,$23,$FB,$28,$DB,$F1,$70,$F6,$BC,$4E,$97,$DE,$B8,$C6,$0D
	.dc.b	$4F,$07,$96,$1F,$E6,$BB,$EB,$87,$9A,$B4,$FA,$F0,$D1,$D6,$FC,$E7
	.dc.b	$95,$CB,$DA,$F4,$AF,$AE,$1F,$BF,$6C,$72,$3C,$C7,$E0,$5C,$BF,$82
	.dc.b	$C3,$F6,$71,$83,$2F,$9A,$7A,$5F,$8A,$BD,$A7,$8C,$F8,$D5,$C6,$3E
	.dc.b	$D4,$7A,$4B,$1F,$39,$FC,$1B,$5E,$A7,$28,$ED,$87,$FB,$43,$E4,$BF
	.dc.b	$D2,$9F,$0A,$F3,$BA,$DA,$30,$F8,$1A,$FE,$2D,$AF,$E4,$9C,$D7,$F8
	.dc.b	$E9,$B8,$FF,$27,$6C,$FF,$36,$F8,$5F,$F7,$5B,$F8,$36,$BF,$97,$6B
	.dc.b	$F9,$E6,$BF,$CC,$5A,$FF,$40,$F8,$17,$C2,$FA,$9B,$1F,$21,$AF,$E8
	.dc.b	$DA,$FE,$A9,$CE,$1F,$E5,$26,$E1,$FC,$FD,$B3,$FD,$3B,$CE,$FF,$64
	.dc.b	$57,$F9,$61,$FF,$64,$47,$F4,$56,$BB,$EA,$F9,$B1,$88,$53,$DD,$F4
	.dc.b	$12,$3D,$0D,$7F,$87,$46,$DF,$D5,$35,$FE,$35,$E1,$3E,$11,$F0,$BF
	.dc.b	$ED,$1A,$D6,$5E,$DC,$30,$7F,$5B,$7F,$CE,$F7,$C6,$B8,$79,$5F,$2C
	.dc.b	$98,$FC,$B6,$BE,$47,$6F,$F0,$59,$F1,$7C,$26,$23,$47,$A2,$7C,$1A
	.dc.b	$FF,$78,$AD,$CF,$F0,$5F,$83,$7C,$27,$FB,$A5,$B1,$98,$FF,$76,$7C
	.dc.b	$1B,$CE,$D4,$49,$D1,$FC,$3E,$74,$E0,$FE,$71,$5D,$1B,$3D,$2B,$FD
	.dc.b	$F1,$F0,$8F,$66,$92,$BA,$6F,$5F,$E8,$53,$B7,$F8,$F3,$5F,$E8,$D8
	.dc.b	$FD,$2C,$FF,$C1,$53,$9D,$FF,$D4,$2D,$8C,$E1,$EE,$C3,$E3,$E1,$FE
	.dc.b	$1E,$F8,$47,$C2,$62,$11,$87,$F9,$73,$5F,$0F,$C2,$6E,$31,$97,$FA
	.dc.b	$C1,$B8,$FF,$87,$5B,$5F,$E2,$CC,$D7,$C7,$09,$3D,$9F,$A7,$1D,$7D
	.dc.b	$89,$3E,$27,$EB,$F5,$E2,$3F,$B4,$60,$9C,$65,$FE,$C2,$C7,$F8,$3D
	.dc.b	$3E,$0D,$F0,$9E,$F6,$C7,$F4,$B6,$BF,$DC,$FE,$0D,$F0,$9F,$E5,$18
	.dc.b	$7C,$0B,$E0,$FF,$C7,$57,$C3,$2F,$F6,$8B,$78,$33,$97,$CD,$CB,$9D
	.dc.b	$7F,$93,$B0,$E5,$8F,$F9,$42,$72,$F6,$C7,$8B,$B6,$B9,$38,$65,$E5
	.dc.b	$AE,$70,$AE,$71,$8C,$BD,$AD,$C7,$FC,$A4,$F0,$9D,$E1,$FE,$55,$6D
	.dc.b	$FF,$99,$2F,$8C,$F8,$AB,$86,$1E,$18,$E0,$DA,$E1,$87,$69,$E0,$6B
	.dc.b	$1F,$64,$FA,$9A,$FD,$79,$E0,$F2,$33,$8F,$91,$F7,$5F,$77,$81,$CB
	.dc.b	$1F,$3B,$D4,$F9,$07,$0B,$D5,$F6,$8E,$26,$AF,$B1,$F5,$1F,$F1,$3B
	.dc.b	$5F,$51,$FE,$25,$70,$FD,$BF,$21,$F0,$9F,$DC,$25,$18,$9E,$0D,$7A
	.dc.b	$B5,$FF,$18,$7C,$27,$F6,$C4,$D2,$3F,$DC,$1F,$93,$F0,$6F,$82,$6F
	.dc.b	$F3,$2B,$FD,$3B,$5A,$9C,$C7,$C4,$32,$7C,$6A,$E5,$5C,$63,$C2,$66
	.dc.b	$78,$EF,$F9,$08,$E2,$FE,$4A,$B8,$C7,$53,$8C,$66,$78,$EF,$EE,$9E
	.dc.b	$97,$DF,$0F,$33,$63,$27,$38,$CB,$4C,$DF,$8D,$EF,$5F,$EA,$FC,$3C
	.dc.b	$73,$9A,$E2,$DA,$E3,$5B,$C3,$E2,$EB,$F1,$A5,$87,$B2,$31,$1F,$02
	.dc.b	$B8,$63,$FE,$55,$AD,$3E,$1E,$3E,$6E,$AB,$FD,$01,$3F,$E2,$98,$E1
	.dc.b	$CF,$A9,$FF,$70,$7C,$27,$F5,$59,$C9,$FE,$69,$C5,$19,$9E,$75,$C6
	.dc.b	$73,$D5,$CE,$7D,$33,$9A,$E1,$5E,$DC,$2E,$1F,$EE,$2F,$61,$F0,$8F
	.dc.b	$17,$95,$6B,$25,$6A,$73,$9F,$A6,$31,$87,$C3,$8C,$61,$CE,$30,$7F
	.dc.b	$A0,$E7,$77,$F1,$2B,$77,$E9,$F6,$FF,$9F,$5F,$FA,$24,$7C,$27,$CE
	.dc.b	$BF,$F7,$CC,$E9,$FF,$5A,$4E,$B8,$39,$75,$7C,$D8,$C1,$FE,$8B,$AE
	.dc.b	$1A,$FE,$92,$33,$7F,$EB,$59,$CC,$23,$F5,$B8,$78,$54,$E9,$CE,$1F
	.dc.b	$47,$E1,$1F,$09,$E1,$BB,$87,$FA,$6E,$6D,$FA,$26,$E3,$F9,$19,$B5
	.dc.b	$E1,$3E,$0D,$F5,$3A,$15,$F8,$D5,$C2,$FD,$15,$67,$89,$FE,$AB,$D1
	.dc.b	$F7,$1F,$87,$6C,$7F,$8C,$FC,$BF,$F3,$2B,$B9,$D1,$1D,$79,$97,$0F
	.dc.b	$F6,$E5,$F2,$7F,$D9,$4D,$7F,$DA,$2D,$7E,$61,$FE,$33,$AF,$D6,$BF
	.dc.b	$3F,$F0,$BF,$9C,$7D,$FF,$AD,$1F,$E6,$1C,$3F,$D8,$91,$AD,$BF,$40
	.dc.b	$EE,$75,$5F,$EE,$F8,$D5,$7F,$A9,$1F,$E2,$F1,$FE,$E4,$6B,$E4,$6B
	.dc.b	$F8,$1F,$92,$7C,$57,$9A,$11,$8A,$FF,$57,$B6,$3F,$DD,$4D,$7F,$DE
	.dc.b	$3F,$1C,$F6,$39,$36,$BF,$D6,$ED,$9F,$F7,$93,$5F,$F7,$CF,$C2,$3E
	.dc.b	$13,$F2,$31,$FE,$F7,$8C,$72,$FF,$B0,$17,$C1,$7E,$E2,$75,$5C,$1F
	.dc.b	$A2,$3E,$35,$F1,$BF,$EB,$C6,$C7,$FB,$29,$AF,$03,$84,$7F,$B0,$BE
	.dc.b	$01,$F0,$54,$93,$FE,$D2,$8C,$B7,$E0,$78,$11,$C1,$FE,$D8,$F9,$2F
	.dc.b	$F7,$4F,$1B,$B1,$F5,$FF,$E5,$C8,$C3,$FE,$5B,$9F,$F7,$59,$FE,$DB
	.dc.b	$AF,$DC,$35,$34,$9F,$01,$C4,$91,$FB,$FB,$FE,$85,$F0,$4F,$93,$CC
	.dc.b	$F8,$F9,$3F,$33,$87,$83,$0F,$F3,$65,$EE,$B9,$56,$A7,$8E,$1F,$D3
	.dc.b	$6F,$FD,$FB,$86,$2B,$57,$C0,$F4,$DF,$49,$D4,$60,$D1,$EC,$8D,$38
	.dc.b	$AF,$23,$63,$2F,$A4,$FE,$39,$87,$9D,$C1,$CC,$E0,$76,$75,$87,$C4
	.dc.b	$BF,$1C,$EC,$E3,$3A,$C3,$F7,$AD,$8F,$41,$DB,$0F,$82,$F5,$AE,$71
	.dc.b	$88,$E7,$18,$C3,$D3,$87,$86,$7C,$A7,$38,$CE,$1E,$0A,$E1,$3D,$48
	.dc.b	$7E,$67,$0F,$80,$FE,$89,$AF,$13,$F3,$C1,$F5,$36,$3B,$BF,$D1,$7E
	.dc.b	$13,$F8,$2B,$DB,$0F,$33,$EA,$3E,$BC,$69,$C1,$23,$67,$F8,$46,$34
	.dc.b	$E2,$FC,$93,$B7,$F9,$4C,$E9,$CC,$EE,$7E,$ED,$7F,$AF,$EB,$51,$83
	.dc.b	$CD,$5C,$0E,$7D,$1E,$C8,$F7,$1A,$3D,$CF,$33,$58,$7C,$4A,$EB,$26
	.dc.b	$C7,$C8,$35,$3C,$67,$A1,$D8,$FC,$8C,$F2,$38,$DF,$8D,$C4,$F8,$23
	.dc.b	$94,$F0,$8C,$3F,$39,$CE,$1F,$0E,$3B,$3F,$36,$33,$87,$C1,$9F,$88
	.dc.b	$73,$39,$1E,$6C,$3D,$2E,$0F,$A4,$70,$AE,$8E,$5F,$04,$70,$8E,$87
	.dc.b	$03,$C7,$3F,$41,$E0,$F8,$DC,$9F,$40,$C9,$DD,$D6,$1F,$12,$3C,$D8
	.dc.b	$7B,$1E,$07,$7A,$F1,$5F,$9A,$31,$3B,$7E,$C6,$1F,$29,$F3,$CE,$CF
	.dc.b	$5C,$EE,$FE,$AC,$FD,$A7,$C4,$F4,$3C,$E7,$43,$CA,$61,$E2,$FC,$37
	.dc.b	$8C,$F6,$8D,$57,$8E,$3C,$58,$7C,$79,$EC,$79,$9C,$47,$18,$C1,$F0
	.dc.b	$9C,$1C,$9C,$47,$CF,$C3,$E2,$BD,$63,$FB,$2D,$FE,$DD,$B7,$FC,$11
	.dc.b	$F6,$1F,$D8,$E3,$4F,$12,$1F,$48,$E2,$7B,$0F,$E7,$3F,$04,$F2,$BE
	.dc.b	$65,$7D,$38,$C4,$EC,$F5,$87,$2E,$2B,$E4,$DF,$F8,$B2,$73,$8F,$BE
	.dc.b	$BC,$ED,$8C,$CF,$C8,$24,$FC,$43,$FC,$C1,$1A,$70,$72,$8D,$BE,$3C
	.dc.b	$3D,$2F,$E0,$27,$83,$B3,$8D,$FF,$06,$8C,$BE,$88,$EB,$18,$C3,$9C
	.dc.b	$7F,$54,$8C,$3F,$0F,$0F,$F5,$B4,$62,$3E,$14,$FF,$93,$27,$37,$C9
	.dc.b	$F8,$0D,$C3,$D3,$7C,$67,$E3,$61,$E5,$AD,$61,$E4,$6B,$EB,$3E,$01
	.dc.b	$E3,$BE,$71,$A9,$E5,$39,$39,$14,$E5,$7E,$1A,$F9,$73,$C2,$7D,$67
	.dc.b	$CB,$AE,$98,$7B,$6B,$EF,$46,$27,$F1,$6D,$7E,$43,$70,$F8,$71,$E1
	.dc.b	$BE,$D1,$88,$EE,$70,$39,$9E,$7C,$3E,$D9,$E7,$7D,$66,$CE,$31,$CC
	.dc.b	$97,$EC,$C3,$D1,$3F,$8C,$C2,$BF,$28,$E3,$3E,$C7,$C2,$60,$E0,$60
	.dc.b	$FF,$2C,$DF,$E8,$9B,$3F,$B5,$8C,$47,$33,$CD,$7E,$47,$FD,$79,$F0
	.dc.b	$4F,$8E,$F1,$D7,$BA,$E9,$E6,$9D,$57,$92,$F9,$3E,$95,$87,$FA,$72
	.dc.b	$77,$3E,$08,$F1,$4F,$09,$CB,$CA,$FC,$15,$FE,$3C,$9C,$D7,$02,$47
	.dc.b	$C2,$3F,$C6,$F1,$A7,$15,$CE,$FD,$18,$7A,$A7,$D1,$7C,$95,$F1,$7D
	.dc.b	$51,$9C,$3F,$73,$39,$7C,$B3,$EB,$3E,$4D,$75,$70,$72,$32,$78,$0F
	.dc.b	$2D,$78,$6B,$A5,$F1,$C3,$E2,$5D,$AF,$14,$FA,$AF,$E2,$4E,$5E,$67
	.dc.b	$63,$FC,$B3,$F0,$4F,$3B,$F9,$B3,$63,$FC,$76,$D7,$FD,$34,$D7,$FD
	.dc.b	$46,$BC,$3F,$CA,$BB,$C7,$EE,$C8,$46,$71,$FF,$6D,$BC,$63,$FC,$EA
	.dc.b	$D7,$BD,$79,$08,$FF,$93,$BF,$00,$79,$FF,$D3,$5B,$19,$8C,$1F,$9D
	.dc.b	$E1,$FD,$4C,$EA,$7F,$CF,$2D,$C7,$F9,$2C,$FE,$70,$E1,$3F,$E5,$CD
	.dc.b	$BF,$CB,$EF,$F3,$E8,$FE,$B9,$E5,$BE,$17,$F6,$FF,$02,$F8,$57,$C0
	.dc.b	$E1,$B8,$7F,$A4,$AF,$83,$6F,$55,$23,$04,$3F,$D9,$53,$C1,$B3,$E2
	.dc.b	$3F,$B7,$7E,$E6,$F8,$6F,$07,$CC,$4A,$FF,$50,$CE,$5B,$39,$7F,$D3
	.dc.b	$D9,$7A,$6B,$F8,$7D,$6E,$7F,$D3,$ED,$7F,$DC,$6D,$7E,$F0,$70,$FE
	.dc.b	$05,$AB,$E1,$7E,$1B,$FE,$AA,$6E,$DF,$88,$8D,$67,$FC,$83,$E0,$1E
	.dc.b	$7F,$F2,$2B,$DF,$2F,$81,$B7,$F8,$F6,$AA,$E2,$FF,$B0,$A3,$24,$BE
	.dc.b	$D7,$FC,$BB,$E1,$7F,$09,$87,$B1,$B7,$F9,$B6,$BD,$7E,$17,$87,$2C
	.dc.b	$8F,$F6,$7C,$EB,$25,$66,$7F,$D9,$33,$C8,$D4,$7F,$3E,$F8,$37,$C2
	.dc.b	$D2,$57,$FB,$6A,$DF,$0B,$88,$FF,$69,$D7,$07,$FD,$BE,$D7,$FD,$AF
	.dc.b	$3C,$1C,$D7,$1D,$BE,$85,$6D,$FA,$55,$BA,$EB,$EC,$F9,$48,$F8,$FB
	.dc.b	$AB,$A3,$C3,$1F,$66,$3F,$AC,$FC,$23,$E1,$29,$2B,$FD,$EB,$F0,$6F
	.dc.b	$85,$65,$FC,$27,$E0,$BF,$89,$61,$C6,$FF,$DF,$6D,$7F,$DE,$6D,$9F
	.dc.b	$C5,$5F,$37,$97,$85,$F9,$68,$C5,$2F,$AE,$3F,$9A,$6C,$70,$3F,$8D
	.dc.b	$4F,$08,$FE,$48,$DA,$FD,$0E,$1E,$39,$E3,$5D,$67,$C1,$8F,$B2,$30
	.dc.b	$74,$71,$3C,$09,$3C,$31,$F1,$53,$0F,$03,$FC,$B7,$0F,$05,$78,$B1
	.dc.b	$F5,$3C,$67,$B9,$C7,$1E,$F1,$B8,$EB,$5C,$9F,$26,$1E,$0C,$3B,$C2
	.dc.b	$74,$DC,$3B,$BC,$0E,$F3,$C2,$76,$70,$79,$9F,$4A,$3F,$84,$9E,$5A
	.dc.b	$AF,$33,$31,$B7,$ED,$7C,$23,$E1,$77,$1F,$16,$30,$F2,$94,$62,$39
	.dc.b	$B5,$EF,$F0,$BF,$E9,$74,$D2,$3F,$E9,$B7,$E9,$FC,$23,$DD,$FE,$BF
	.dc.b	$B7,$F6,$5A,$E2,$7D,$09,$D7,$47,$DA,$8F,$EC,$97,$CA,$F8,$ED,$E2
	.dc.b	$9E,$11,$87,$E2,$CE,$EB,$94,$76,$C3,$FB,$B6,$1D,$A3,$EF,$9D,$8E
	.dc.b	$73,$C2,$B9,$46,$8E,$58,$7F,$3A,$AE,$57,$EB,$AD,$B6,$B9,$9C,$63
	.dc.b	$35,$E9,$9F,$1E,$1E,$9A,$FE,$8B,$5A,$7F,$C0,$97,$E7,$D3,$FA,$03
	.dc.b	$71,$E5,$8F,$8F,$0F,$F0,$5D,$69,$B3,$E7,$E0,$B8,$74,$9A,$B0,$FF
	.dc.b	$64,$57,$F8,$4E,$FB,$F4,$70,$5F,$0B,$FD,$3A,$74,$7C,$39,$D4,$66
	.dc.b	$FF,$8A,$CE,$4D,$BF,$C6,$23,$55,$DA,$72,$F6,$EA,$F2,$E1,$A9,$FF
	.dc.b	$13,$7C,$23,$E1,$5E,$26,$AF,$DB,$39,$8C,$57,$82,$30,$F8,$B2,$F8
	.dc.b	$15,$C2,$F9,$CE,$70,$F0,$52,$F8,$BE,$C3,$EA,$DF,$1C,$3E,$5F,$B8
	.dc.b	$E8,$F9,$F8,$BE,$13,$EC,$61,$FE,$3F,$AD,$4F,$C4,$35,$BF,$2B,$E3
	.dc.b	$7C,$2D,$7C,$1F,$15,$F0,$3D,$91,$9D,$7F,$C9,$B1,$99,$F4,$1C,$AF
	.dc.b	$AE,$1F,$82,$F8,$47,$C2,$F8,$0C,$16,$FE,$44,$69,$F4,$65,$70,$EB
	.dc.b	$37,$1F,$0C,$DA,$FF,$10,$CD,$AF,$F4,$EC,$D5,$19,$9F,$91,$7F,$96
	.dc.b	$D1,$F0,$94,$FC,$DF,$8E,$F8,$5F,$F3,$75,$FD,$DA,$DB,$6B,$FC,$51
	.dc.b	$8F,$F8,$CA,$72,$FF,$2A,$6B,$FC,$B9,$AE,$E3,$BD,$7E,$12,$3F,$C5
	.dc.b	$F1,$89,$FD,$5F,$C2,$FD,$73,$EF,$FB,$43,$FC,$3B,$87,$E4,$63,$5B
	.dc.b	$79,$30,$FF,$41,$47,$F1,$57,$F9,$FB,$5F,$F2,$4B,$5D,$C7,$C2,$F9
	.dc.b	$07,$C0,$F8,$A3,$0F,$EF,$BE,$11,$F0,$BD,$ED,$18,$C7,$FA,$2B,$63
	.dc.b	$FA,$6B,$5F,$EA,$3F,$0B,$DB,$97,$FA,$5B,$63,$FA,$9A,$F8,$5F,$D4
	.dc.b	$46,$31,$FC,$CB,$FD,$63,$E3,$BE,$C3,$FD,$51,$AE,$78,$7F,$AB,$A8
	.dc.b	$CB,$FD,$8B,$E0,$DF,$09,$A6,$C7,$F5,$E6,$BF,$D9,$9A,$F1,$F8,$4E
	.dc.b	$04,$78,$3E,$0F,$80,$FE,$C7,$1F,$DA,$1A,$FF,$70,$6B,$FD,$D1,$AF
	.dc.b	$F9,$FB,$E0,$5E,$D7,$FA,$03,$B3,$FC,$F1,$F0,$6F,$83,$C7,$1F,$FA
	.dc.b	$12,$31,$87,$F1,$5C,$38,$B7,$0F,$87,$1F,$EA,$E8,$34,$91,$D2,$3F
	.dc.b	$D9,$91,$C9,$FE,$02,$DC,$3F,$D1,$58,$BE,$0F,$85,$74,$BB,$7B,$C7
	.dc.b	$D9,$1A,$BE,$15,$E5,$21,$DA,$35,$7C,$99,$1A,$C3,$C3,$38,$8E,$06
	.dc.b	$6F,$FD,$4F,$7D,$A3,$24,$AF,$F5,$16,$1E,$58,$FF,$75,$4E,$E7,$38
	.dc.b	$75,$6C,$FF,$41,$9D,$5F,$6C,$34,$FB,$A7,$51,$EA,$7C,$87,$CC,$BC
	.dc.b	$3F,$CB,$2B,$9B,$93,$61,$CB,$67,$BB,$B3,$93,$D2,$FC,$F1,$EF,$BF
	.dc.b	$33,$E4,$3A,$E1,$F0,$5E,$86,$4F,$CB,$57,$E6,$C8,$FE,$6D,$ED,$3D
	.dc.b	$23,$FA,$2F,$BC,$F2,$33,$1F,$32,$AA,$9D,$3D,$F0,$FB,$6D,$BF,$F6
	.dc.b	$FC,$EA,$3D,$A6,$4E,$87,$F7,$37,$53,$CA,$36,$47,$FD,$C1,$1A,$9F
	.dc.b	$44,$6C,$8F,$82,$FF,$37,$3B,$AE,$E7,$39,$D3,$E1,$9F,$6E,$3F,$5A
	.dc.b	$B8,$4F,$BE,$73,$87,$46,$47,$AC,$FA,$27,$CA,$72,$F7,$3C,$38,$7A
	.dc.b	$A3,$2F,$A2,$33,$87,$91,$E4,$75,$C3,$C9,$5C,$27,$E0,$92,$47,$9C
	.dc.b	$F6,$1C,$6F,$83,$AC,$3D,$B7,$F9,$43,$AC,$70,$3D,$07,$B3,$0F,$43
	.dc.b	$97,$D0,$E6,$FD,$11,$A8,$E3,$3D,$1F,$59,$8B,$F1,$BE,$73,$07,$5A
	.dc.b	$E6,$7B,$5E,$C7,$A7,$0E,$35,$E9,$AE,$78,$7B,$1F,$CB,$BF,$68,$C5
	.dc.b	$7D,$77,$15,$F7,$23,$C0,$75,$3E,$41,$83,$CD,$87,$C0,$BF,$63,$EF
	.dc.b	$31,$7E,$2A,$E9,$5F,$02,$75,$87,$AE,$F8,$BE,$E9,$FC,$09,$0E,$EE
	.dc.b	$5F,$82,$4C,$3D,$2F,$98,$ED,$1B,$9F,$89,$1E,$99,$F8,$2F,$DC,$78
	.dc.b	$BF,$EC,$A7,$F6,$CF,$FB,$39,$AF,$E4,$CC,$5F,$F7,$48,$CC,$7C,$73
	.dc.b	$94,$64,$F8,$4F,$F9,$BF,$E0,$DF,$07,$88,$F3,$9F,$98,$8C,$BE,$79
	.dc.b	$FC,$2C,$62,$36,$7C,$43,$C1,$8B,$1F,$D0,$C2,$78,$4F,$D0,$8D,$CF
	.dc.b	$CB,$9C,$BF,$37,$0F,$87,$5C,$EB,$C9,$87,$9E,$39,$47,$58,$C5,$73
	.dc.b	$6C,$76,$38,$9E,$B3,$F1,$57,$E0,$74,$7A,$C8,$E5,$C5,$F9,$2B,$DB
	.dc.b	$39,$8C,$4F,$28,$C6,$1E,$28,$F3,$1E,$13,$E4,$C6,$9C,$61,$EF,$76
	.dc.b	$7C,$EA,$EA,$E2,$3E,$0C,$62,$79,$4E,$EB,$FC,$13,$18,$AD,$C6,$30
	.dc.b	$F2,$5F,$F2,$CA,$D6,$1F,$46,$B5,$87,$92,$3E,$63,$4F,$75,$F2,$9D
	.dc.b	$C7,$84,$F9,$B1,$97,$F2,$0D,$7F,$D4,$35,$E0,$7E,$8D,$F8,$E7,$58
	.dc.b	$7C,$3A,$D5,$7D,$2B,$EF,$3E,$C8,$EE,$79,$4F,$B3,$5C,$B0,$FA,$ED
	.dc.b	$8F,$C1,$9E,$73,$E9,$B9,$7E,$74,$70,$7A,$5F,$E0,$CF,$1E,$1E,$78
	.dc.b	$F6,$9A,$AF,$2D,$7C,$B8,$C9,$F0,$23,$F0,$95,$CD,$A7,$08,$F8,$78
	.dc.b	$57,$C6,$60,$E7,$87,$C5,$9E,$31,$8B,$F0,$3D,$CC,$1C,$4C,$12,$7E
	.dc.b	$53,$63,$E0,$3F,$16,$3B,$3E,$A3,$F7,$A7,$E5,$8B,$7F,$C1,$5B,$5F
	.dc.b	$E8,$88,$F7,$47,$8B,$C0,$76,$7F,$BD,$63,$FB,$3D,$FE,$87,$E0,$DF
	.dc.b	$07,$FE,$62,$5E,$6D,$31,$12,$BC,$67,$F7,$09,$E2,$E0,$38,$76,$D7
	.dc.b	$D6,$E6,$BC,$53,$D6,$7F,$DE,$FE,$47,$FA,$92,$7F,$C4,$3C,$5F,$E7
	.dc.b	$6F,$83,$79,$D4,$FF,$0F,$C4,$BF,$17,$0F,$FA,$1C,$91,$FC,$E4,$D4
	.dc.b	$7F,$90,$1F,$F4,$46,$AF,$C4,$BF,$82,$63,$FC,$17,$27,$C1,$EB,$E1
	.dc.b	$1F,$01,$4F,$F4,$4C,$EB,$2F,$9B,$C9,$FC,$D2,$31,$1F,$21,$AF,$DC
	.dc.b	$6B,$BB,$FF,$60,$C7,$0C,$3C,$51,$D8,$F1,$7B,$3F,$E9,$8F,$82,$7C
	.dc.b	$27,$97,$E1,$26,$3E,$23,$FC,$4B,$3A,$70,$1C,$3F,$71,$AF,$E4,$1A
	.dc.b	$FF,$4E,$6B,$D2,$7C,$44,$7C,$1B,$35,$FF,$5E,$C3,$2F,$E6,$1F,$09
	.dc.b	$F2,$A3,$C6,$D7,$5A,$FF,$5C,$8C,$4F,$F3,$7F,$7D,$E7,$FF,$AE,$27
	.dc.b	$35,$BE,$0C,$1F,$EB,$CA,$DE,$1D,$A3,$2D,$3E,$1C,$69,$F0,$9F,$E4
	.dc.b	$CF,$84,$FC,$7B,$F2,$1A,$E8,$38,$C7,$CD,$F0,$9F,$EC,$C8,$C6,$1F
	.dc.b	$DD,$9B,$5D,$79,$BC,$D5,$B6,$47,$77,$2F,$D8,$F8,$27,$C1,$FF,$B6
	.dc.b	$1B,$19,$8F,$81,$F0,$99,$64,$64,$F7,$C6,$1F,$54,$F2,$6C,$F2,$7C
	.dc.b	$1E,$CF,$72,$3F,$23,$C2,$AF,$28,$BA,$7D,$79,$FF,$76,$FC,$1B,$9E
	.dc.b	$92,$FF,$27,$F0,$4F,$84,$FC,$34,$3E,$0B,$F8,$9B,$FE,$7E,$BE,$4F
	.dc.b	$E6,$9A,$FD,$36,$E3,$FD,$F6,$FA,$3F,$E8,$6A,$E7,$49,$CB,$DF,$C3
	.dc.b	$FF,$15,$DF,$4A,$DB,$63,$FC,$5F,$5C,$27,$9B,$63,$8E,$3F,$E3,$49
	.dc.b	$E2,$D8,$E3,$8F,$BA,$78,$B6,$38,$E1,$E7,$6B,$E5,$3F,$D3,$F7,$DD
	.dc.b	$B3,$C5,$F8,$D7,$DD,$A7,$33,$78,$D7,$E1,$4F,$15,$8F,$BD,$E2,$78
	.dc.b	$4D,$CF,$1B,$F1,$47,$0B,$F0,$34,$D5,$F5,$9D,$DF,$59,$D1,$D6,$39
	.dc.b	$9F,$51,$FE,$C2,$D7,$E4,$DF,$F4,$F8,$C4,$7C,$4F,$83,$7C,$26,$5B
	.dc.b	$1F,$D6,$25,$18,$8E,$AD,$7F,$AE,$FC,$27,$F5,$B7,$FD,$9A,$D4,$D2
	.dc.b	$3F,$17,$E0,$9E,$DF,$F9,$AF,$6F,$C7,$CF,$23,$51,$9D,$BB,$C6,$39
	.dc.b	$FE,$A3,$BA,$FC,$BD,$F4,$9C,$CF,$88,$E0,$7D,$53,$85,$7F,$A1,$27
	.dc.b	$84,$64,$E1,$5D,$2F,$C5,$1F,$E0,$EC,$3F,$09,$5C,$2B,$D3,$3C,$2F
	.dc.b	$A5,$F0,$8F,$F4,$8D,$F0,$8C,$47,$88,$FF,$7A,$DF,$3A,$F7,$DF,$87
	.dc.b	$0F,$5E,$1C,$5B,$7E,$C9,$D3,$FC,$BE,$FD,$7A,$7F,$9E,$63,$B0,$8C
	.dc.b	$D7,$98,$FE,$BE,$B8,$3F,$D3,$D5,$FE,$81,$8D,$D7,$9A,$73,$5F,$27
	.dc.b	$96,$CF,$F5,$9F,$84,$FE,$A3,$3D,$27,$38,$27,$35,$E7,$9C,$C6,$27
	.dc.b	$C1,$18,$EC,$EB,$5E,$97,$BF,$C1,$BE,$11,$E2,$EB,$4E,$59,$B0,$5E
	.dc.b	$EB,$58,$72,$C3,$8D,$FC,$28,$C4,$EA,$3F,$C0,$D5,$C5,$D1,$8A,$E3
	.dc.b	$7F,$D7,$FD,$C7,$47,$C4,$A7,$C2,$7C,$AA,$56,$A7,$FC,$24,$6B,$8F
	.dc.b	$DF,$19,$9F,$ED,$F3,$93,$FC,$2B,$19,$DB,$FD,$A9,$19,$9F,$16,$1E
	.dc.b	$1C,$3C,$17,$BA,$F8,$86,$EF,$FD,$7D,$F0,$8F,$84,$E0,$7D,$07,$13
	.dc.b	$E6,$BB,$8F,$F7,$D9,$B7,$E2,$9B,$5F,$EF,$09,$B5,$FE,$F3,$9A,$A3
	.dc.b	$33,$F1,$AF,$C0,$78,$9F,$F1,$9E,$CF,$84,$FC,$E5,$FD,$D8,$C4,$FF
	.dc.b	$9B,$5B,$1F,$E7,$26,$BF,$E8,$67,$38,$FF,$BE,$FC,$67,$BB,$F8,$8A
	.dc.b	$E9,$87,$E7,$27,$37,$F9,$23,$8E,$1F,$8D,$8C,$BE,$69,$E1,$7C,$ED
	.dc.b	$F0,$7F,$C3,$9F,$F4,$BB,$5F,$CB,$1A,$DF,$DA,$DB,$FF,$44,$4F,$EC
	.dc.b	$B0,$FF,$4B,$FB,$FF,$A4,$8C,$7A,$88,$C6,$1E,$98,$C5,$FF,$97,$2B
	.dc.b	$75,$FA,$CF,$81,$E1,$C3,$FE,$95,$6E,$5F,$E9,$A5,$F0,$3F,$AA,$8C
	.dc.b	$57,$E6,$9F,$F4,$FF,$C0,$FC,$A7,$FD,$3C,$D7,$3A,$FF,$47,$8C,$47
	.dc.b	$FA,$E2,$9F,$1F,$FD,$2D,$B1,$FE,$A7,$6B,$FE,$AC,$6B,$FE,$C0,$F8
	.dc.b	$17,$C0,$FF,$53,$6C,$7F,$AB,$DA,$FF,$AD,$1A,$FF,$B2,$3D,$DF,$EB
	.dc.b	$EF,$FA,$D5,$AF,$FA,$E9,$AE,$FC,$3F,$EC,$1E,$FB,$E1,$29,$27,$FD
	.dc.b	$7D,$19,$7A,$7C,$07,$DE,$E8,$71,$FE,$02,$9E,$D6,$63,$0D,$BF,$E3
	.dc.b	$59,$30,$C4,$23,$0F,$62,$1E,$98,$FE,$ED,$1A,$8F,$F3,$A3,$70,$F1
	.dc.b	$C7,$69,$EB,$ED,$FF,$43,$8F,$F5,$1D,$6F,$0F,$CB,$B6,$38,$1F,$EA
	.dc.b	$89,$E0,$F6,$AF,$3C,$EE,$3D,$11,$DB,$C5,$CB,$23,$BB,$96,$90,$E8
	.dc.b	$77,$8E,$0E,$09,$87,$B1,$5F,$92,$30,$DC,$3D,$77,$CA,$73,$7E,$1B
	.dc.b	$E7,$5A,$7C,$04,$32,$7C,$23,$FD,$41,$7D,$5F,$EA,$4D,$7E,$09,$EB
	.dc.b	$AE,$EF,$37,$D1,$87,$C7,$3E,$44,$F1,$39,$39,$C3,$CD,$3C,$AB,$C6
	.dc.b	$76,$3C,$78,$7A,$DF,$E1,$E7,$C4,$7F,$C7,$57,$F6,$23,$2D,$7F,$8A
	.dc.b	$BE,$18,$EF,$87,$F1,$8F,$7D,$F0,$18,$BF,$7D,$55,$3A,$7D,$D5,$D2
	.dc.b	$75,$39,$9F,$39,$A3,$99,$FE,$F4,$75,$87,$BC,$95,$CE,$75,$7E,$1C
	.dc.b	$3D,$11,$83,$A1,$F0,$4B,$7D,$2B,$94,$78,$89,$7F,$8B,$9D,$47,$90
	.dc.b	$CC,$F0,$9D,$5F,$87,$0F,$6B,$6B,$BB,$5E,$27,$F8,$5E,$FA,$BE,$53
	.dc.b	$25,$AE,$AF,$AB,$0F,$6B,$FE,$EF,$9E,$31,$89,$F8,$B3,$D6,$BC,$11
	.dc.b	$88,$FA,$B1,$8C,$3D,$47,$DD,$71,$1E,$F9,$F0,$BE,$4A,$ED,$5B,$8F
	.dc.b	$26,$1F,$01,$A7,$C0,$29,$EC,$C3,$CF,$1D,$A3,$C6,$6A,$FC,$D3,$A8
	.dc.b	$C5,$77,$7C,$55,$DD,$A7,$72,$9D,$F0,$ED,$87,$EC,$EF,$A4,$EB,$0F
	.dc.b	$69,$F1,$67,$DB,$1F,$24,$F0,$9F,$0F,$0F,$14,$EF,$0F,$13,$67,$C4
	.dc.b	$D7,$C4,$72,$3E,$E4,$F2,$3E,$8B,$D4,$87,$C4,$3C,$CE,$9C,$47,$A2
	.dc.b	$31,$5E,$DC,$3F,$A3,$E1,$EE,$AF,$97,$18,$3E,$84,$7E,$40,$D5,$75
	.dc.b	$71,$1A,$38,$5F,$A6,$32,$D8,$CB,$E6,$3E,$41,$F9,$57,$93,$FB,$33
	.dc.b	$47,$AA,$7E,$13,$F0,$2F,$D4,$EE,$FE,$51,$EF,$7A,$B8,$9E,$31,$CC
	.dc.b	$E5,$87,$92,$78,$47,$D0,$3E,$6E,$1E,$A7,$C6,$D8,$E9,$1D,$AB,$AD
	.dc.b	$FE,$12,$3B,$11,$C5,$71,$9F,$A7,$7E,$17,$8B,$F8,$66,$D7,$E1,$9A
	.dc.b	$FE,$18,$F9,$07,$CD,$3E,$13,$D1,$A7,$94,$F7,$98,$3C,$91,$DD,$D1
	.dc.b	$F0,$49,$3E,$77,$89,$E0,$25,$FC,$27,$DF,$1F,$86,$7E,$4E,$1E,$D7
	.dc.b	$8C,$FC,$E9,$D4,$76,$9D,$3E,$1C,$3E,$EB,$DB,$0F,$63,$4F,$3C,$7C
	.dc.b	$C9,$FC,$6C,$7F,$5A,$AF,$E9,$0D,$8F,$EB,$AF,$F0,$53,$07,$C3,$8F
	.dc.b	$E7,$3D,$6E,$AF,$F5,$BC,$E6,$BD,$53,$9F,$07,$CF,$7F,$16,$31,$5B
	.dc.b	$3E,$8B,$87,$D0,$7C,$6A,$FF,$1D,$46,$72,$F8,$91,$DB,$0F,$0D,$70
	.dc.b	$BE,$D1,$8A,$DC,$F6,$8C,$65,$F9,$EA,$DE,$3E,$6C,$3A,$36,$F8,$35
	.dc.b	$F9,$98,$F8,$EF,$95,$70,$9F,$F7,$46,$3F,$EA,$5C,$7B,$C7,$E2,$31
	.dc.b	$F9,$31,$8C,$39,$4E,$6F,$DF,$3C,$21,$F0,$0F,$80,$FA,$B1,$FE,$F5
	.dc.b	$8C,$38,$6B,$F6,$9A,$FF,$86,$1C,$E3,$FE,$3D,$6E,$5F,$EA,$A6,$CF
	.dc.b	$FC,$01,$F1,$BF,$D2,$55,$FE,$43,$7F,$D2,$31,$FE,$B0,$6B,$BE,$AF
	.dc.b	$2A,$8C,$BF,$E4,$AE,$2F,$F1,$39,$1F,$5E,$FF,$E7,$A6,$C7,$F9,$A8
	.dc.b	$8F,$CC,$C7,$FD,$91,$EF,$FF,$95,$71,$FF,$29,$D6,$AB,$FC,$92,$FE
	.dc.b	$22,$75,$5E,$7F,$7F,$CB,$89,$A4,$BF,$9B,$19,$7C,$07,$C3,$8F,$F4
	.dc.b	$4C,$66,$18,$7F,$B8,$FD,$C7,$A9,$AB,$95,$C8,$F8,$8F,$E7,$27,$71
	.dc.b	$9F,$79,$EF,$7C,$A6,$C7,$E0,$1A,$FE,$D9,$AF,$F9,$51,$CE,$1F,$E7
	.dc.b	$96,$E5,$FE,$E6,$6C,$B8,$D8,$3E,$0B,$FD,$4B,$5A,$A3,$4E,$35,$C2
	.dc.b	$7F,$D5,$7E,$FB,$DF,$A4,$BF,$81,$EF,$BC,$BC,$47,$C8,$CF,$FD,$F3
	.dc.b	$7C,$63,$E3,$35,$D0,$15,$CD,$DD,$F4,$D3,$FD,$31,$5C,$5F,$F4,$D5
	.dc.b	$6E,$B9,$8F,$7D,$E9,$FF,$B1,$1B,$1F,$5D,$AF,$09,$7B,$FB,$A9,$3B
	.dc.b	$7F,$D9,$B3,$A8,$C1,$0F,$0D,$74,$BF,$2D,$FF,$B5,$3D,$E7,$BD,$49
	.dc.b	$7F,$2B,$DF,$AF,$AA,$76,$FE,$35,$AF,$9E,$37,$7F,$EB,$B9,$CB,$6B
	.dc.b	$CF,$3E,$1F,$02,$5F,$37,$D7,$E0,$E2,$3C,$11,$6F,$E6,$CF,$FB,$B3
	.dc.b	$DE,$FC,$DD,$7C,$E6,$CF,$FB,$2D,$BC,$1F,$6E,$5E,$EB,$DF,$FD,$5C
	.dc.b	$3D,$C7,$BD,$BC,$38,$CF,$E0,$DA,$9B,$7C,$F1,$FC,$45,$F2,$AF,$C5
	.dc.b	$4E,$5E,$D7,$D7,$B3,$C9,$5B,$9E,$91,$89,$F4,$57,$1A,$D3,$63,$95
	.dc.b	$78,$1A,$F6,$3C,$38,$FB,$AB,$8B,$67,$71,$E1,$C7,$E5,$BA,$C7,$A5
	.dc.b	$F2,$6B,$E6,$2C,$F0,$AF,$3D,$F5,$9E,$07,$8C,$E0,$59,$E0,$EA,$FC
	.dc.b	$0E,$71,$F8,$4F,$D1,$6B,$F6,$AB,$B3,$70,$F1,$CE,$AF,$AC,$E4,$F1
	.dc.b	$63,$F4,$DF,$F7,$6B,$5F,$4D,$7F,$60,$8C,$47,$DA,$F2,$DE,$FE,$5B
	.dc.b	$1F,$E1,$A9,$46,$23,$AB,$5F,$F0,$FF,$BF,$FD,$61,$34,$8F,$F5,$C7
	.dc.b	$E7,$FB,$CF,$61,$BF,$D2,$AF,$E2,$D5,$A9,$CC,$7C,$F3,$27,$D1,$AE
	.dc.b	$55,$C6,$3C,$26,$67,$8E,$DF,$93,$8E,$2F,$E5,$6B,$8C,$75,$38,$C6
	.dc.b	$67,$8E,$FE,$E9,$E9,$7D,$F0,$F3,$36,$32,$73,$8C,$B4,$CD,$F8,$DE
	.dc.b	$F5,$FB,$BC,$3C,$73,$9A,$E2,$DA,$E3,$5B,$C3,$E2,$E9,$F1,$A5,$87
	.dc.b	$B2,$31,$1F,$02,$B8,$63,$FD,$AE,$B4,$F8,$78,$39,$BA,$AF,$E7,$F2
	.dc.b	$9C,$C7,$F4,$B8,$FE,$83,$3F,$DF,$AD,$EF,$FF,$57,$8E,$DD,$3C,$27
	.dc.b	$3D,$9C,$B0,$F3,$CE,$6F,$84,$78,$23,$13,$BC,$33,$7F,$84,$F7,$DE
	.dc.b	$FF,$F8,$9F,$2F,$65,$F9,$A7,$3E,$2F,$18,$C5,$6F,$2F,$6C,$62,$75
	.dc.b	$7F,$E1,$8B,$E2,$F4,$31,$7C,$70,$FA,$1F,$8A,$BE,$C7,$D8,$DA,$FF
	.dc.b	$10,$F8,$6F,$C3,$FF,$BF,$47,$F8,$6E,$75,$A7,$F6,$C6,$E1,$D3,$0F
	.dc.b	$9F,$87,$1F,$0F,$AC,$4C,$3F,$CB,$77,$CB,$0F,$E8,$81,$C3,$FD,$1D
	.dc.b	$AF,$F8,$0F,$F1,$27,$D7,$7D,$36,$D7,$A3,$37,$E1,$19,$A9,$3F,$E3
	.dc.b	$3A,$DD,$FF,$83,$B3,$F7,$5E,$A3,$FC,$0F,$3A,$7C,$AD,$7F,$AD,$35
	.dc.b	$F2,$5F,$E6,$1F,$F0,$C7,$C5,$FE,$03,$0F,$EC,$0D,$8F,$E1,$3F,$0D
	.dc.b	$FE,$7D,$9F,$F2,$34,$E6,$4D,$38,$42,$78,$4F,$F8,$72,$75,$C5,$E1
	.dc.b	$8E,$2D,$8F,$CA,$C7,$CE,$6B,$FD,$CF,$E3,$DE,$C7,$16,$E1,$FE,$90
	.dc.b	$6C,$FF,$B1,$9A,$FD,$8F,$8C,$7B,$9E,$16,$D7,$FA,$61,$B3,$FE,$CE
	.dc.b	$6B,$F8,$0F,$83,$7C,$1B,$E1,$3F,$D3,$F1,$86,$FC,$13,$A7,$35,$BB
	.dc.b	$FF,$6C,$FE,$1F,$FE,$56,$F7,$5F,$55,$FE,$AC,$3F,$C2,$1D,$7F,$C3
	.dc.b	$E7,$FC,$BC,$6E,$35,$5F,$E7,$78,$CF,$AF,$FE,$9C,$7F,$CD,$F1,$FE
	.dc.b	$19,$6B,$FE,$55,$6A,$F9,$17,$DC,$7F,$AF,$3E,$92,$4F,$FA,$0E,$72
	.dc.b	$D9,$CB,$F9,$3C,$95,$E8,$E1,$FF,$14,$5F,$0A,$FF,$2D,$B5,$FD,$4B
	.dc.b	$5D,$DF,$FA,$D2,$31,$1F,$E6,$BF,$8E,$FF,$21,$D7,$F8,$CF,$E1,$1F
	.dc.b	$71,$DE,$BF,$CE,$11,$87,$FD,$35,$5C,$1E,$5E,$AE,$D4,$ED,$FF,$37
	.dc.b	$B7,$7F,$E7,$D5,$C6,$4D,$9D,$61,$E3,$C3,$A4,$EC,$D3,$DE,$77,$5E
	.dc.b	$23,$CB,$86,$66,$46,$08,$79,$1C,$BF,$E8,$4A,$ED,$38,$C3,$FA,$A5
	.dc.b	$F6,$AF,$3C,$F9,$0D,$61,$E7,$8E,$B5,$EE,$9C,$C7,$CA,$C3,$83,$AA
	.dc.b	$E0,$F8,$CC,$1E,$B3,$04,$AE,$8E,$8F,$9F,$3D,$9F,$29,$9B,$F4,$1E
	.dc.b	$9C,$3D,$B1,$FE,$D3,$8C,$CF,$92,$3E,$16,$1C,$E7,$83,$9B,$F8,$75
	.dc.b	$DE,$F9,$DF,$97,$0F,$5D,$79,$A3,$99,$DE,$B8,$5F,$3A,$D9,$27,$E0
	.dc.b	$57,$13,$BB,$C4,$F8,$17,$F8,$27,$07,$62,$47,$07,$C8,$74,$9F,$F4
	.dc.b	$84,$F1,$8C,$9F,$20,$E3,$1E,$37,$E6,$9E,$0A,$F8,$97,$E9,$8E,$87
	.dc.b	$DD,$3C,$77,$E9,$AE,$F5,$C2,$79,$1A,$8E,$D7,$F2,$AB,$D0,$48,$E3
	.dc.b	$3F,$6E,$FE,$4B,$4E,$47,$88,$F5,$4F,$17,$C8,$7B,$0D,$61,$F8,$07
	.dc.b	$C6,$7C,$27,$EF,$BD,$4F,$8B,$7E,$5B,$F8,$93,$A9,$E8,$7C,$63,$78
	.dc.b	$7C,$57,$17,$E3,$7C,$B5,$EC,$7C,$18,$6F,$0F,$A9,$5B,$7C,$13,$EA
	.dc.b	$8C,$5F,$5C,$3D,$53,$E2,$AE,$38,$7C,$CA,$D6,$1F,$22,$BC,$F7,$C3
	.dc.b	$0F,$AD,$1F,$12,$75,$18,$8F,$4D,$F3,$7E,$24,$78,$9B,$19,$7A,$CF
	.dc.b	$28,$F7,$4F,$C1,$AF,$90,$F5,$7C,$51,$A2,$57,$BD,$E9,$3B,$AE,$93
	.dc.b	$E7,$3F,$02,$7C,$C3,$07,$D2,$3C,$A7,$33,$F0,$D8,$7C,$67,$07,$12
	.dc.b	$5F,$91,$DE,$1F,$99,$7E,$73,$6F,$E3,$57,$39,$F8,$E7,$03,$CB,$5F
	.dc.b	$03,$0F,$6C,$68,$93,$F7,$9C,$4E,$DF,$5D,$7C,$F7,$E6,$9F,$02,$78
	.dc.b	$4F,$9E,$3D,$51,$E3,$8E,$51,$EC,$C3,$E4,$B4,$E5,$87,$E1,$9A,$7C
	.dc.b	$83,$CA,$E5,$F9,$87,$88,$B7,$F5,$5E,$26,$5F,$41,$97,$78,$7C,$2B
	.dc.b	$F4,$61,$EA,$C3,$CA,$FA,$5F,$D2,$61,$F3,$EB,$E4,$CE,$6F,$F7,$38
	.dc.b	$7E,$36,$F7,$1F,$41,$AF,$EB,$8D,$D7,$EE,$A3,$33,$C6,$72,$7E,$13
	.dc.b	$0F,$93,$7D,$E3,$F0,$27,$F9,$D8,$FE,$AC,$7F,$9E,$16,$19,$8C,$9E
	.dc.b	$87,$F7,$1C,$7E,$0C,$7E,$13,$67,$F8,$AB,$5F,$C9,$B6,$BF,$C0,$CD
	.dc.b	$AF,$F0,$13,$67,$FA,$E3,$5F,$F0,$8B,$70,$FE,$F6,$DB,$FF,$10,$36
	.dc.b	$7F,$C3,$ED,$7F,$C3,$6D,$C3,$FC,$9D,$1F,$08,$FE,$17,$19,$3F,$C4
	.dc.b	$5F,$06,$F6,$3F,$C5,$5E,$0B,$DE,$E3,$93,$26,$2A,$D6,$3C,$B0,$93
	.dc.b	$FE,$B2,$C3,$8D,$F0,$BD,$CF,$2C,$3F,$DD,$58,$F8,$63,$15,$D9,$46
	.dc.b	$5F,$B4,$D7,$FD,$62,$D7,$83,$5F,$3F,$90,$F7,$FD,$B8,$BC,$8A,$7F
	.dc.b	$90,$E7,$55,$E2,$38,$D6,$9F,$64,$EE,$24,$7E,$26,$78,$BF,$8F,$6B
	.dc.b	$FE,$C6,$6B,$ED,$FA,$47,$C3,$7F,$94,$FD,$8F,$F6,$E4,$F8,$43,$85
	.dc.b	$A7,$D3,$7E,$9D,$72,$78,$35,$DE,$9F,$2E,$B9,$57,$19,$EB,$E5,$FF
	.dc.b	$98,$EF,$84,$E5,$E5,$3F,$ED,$E6,$BE,$9C,$3F,$50,$A3,$2F,$E1,$BF
	.dc.b	$74,$7B,$9F,$43,$DB,$93,$ED,$9C,$B6,$72,$FF,$B7,$77,$57,$F9,$DE
	.dc.b	$B7,$3E,$C3,$B5,$78,$A3,$31,$FE,$F4,$6B,$F8,$B6,$BE,$5F,$47,$F0
	.dc.b	$7E,$C4,$8F,$24,$77,$30,$F8,$E3,$0E,$8F,$E1,$12,$9F,$11,$23,$3F
	.dc.b	$41,$FE,$8F,$F8,$0F,$CD,$6B,$F4,$5B,$C1,$E5,$9F,$71,$BE,$69,$7F
	.dc.b	$1E,$32,$F4,$21,$FC,$77,$14,$7D,$4F,$7F,$FD,$30,$4E,$AF,$16,$BF
	.dc.b	$EA,$18,$C5,$FE,$3D,$FF,$4F,$65,$F3,$B3,$FC,$34,$6A,$7E,$80,$77
	.dc.b	$5C,$FD,$F9,$3C,$DF,$81,$EC,$62,$38,$D2,$71,$7F,$0E,$BF,$D6,$9E
	.dc.b	$FF,$E8,$2F,$F4,$4D,$7C,$2D,$9F,$CD,$7B,$DF,$86,$7F,$8A,$FB,$AF
	.dc.b	$73,$F8,$55,$6F,$86,$F4,$F6,$3F,$D8,$58,$71,$C5,$87,$17,$F8,$FC
	.dc.b	$66,$31,$0B,$F2,$46,$1B,$87,$F1,$9A,$E0,$DE,$5F,$DC,$DF,$68,$F1
	.dc.b	$BE,$86,$B9,$3C,$31,$C2,$FC,$35,$D2,$7F,$99,$DF,$81,$B8,$7F,$29
	.dc.b	$AD,$DF,$1C,$3C,$59,$77,$BD,$94,$F6,$5F,$19,$DC,$78,$C6,$3E,$17
	.dc.b	$67,$1C,$7C,$4E,$B0,$F0,$3E,$AC,$35,$87,$58,$F0,$9D,$6F,$A1,$D6
	.dc.b	$F9,$9D,$1C,$DF,$CA,$9F,$E8,$C7,$7B,$FE,$F5,$3F,$6E,$32,$D7,$EA
	.dc.b	$72,$BD,$FD,$B6,$3F,$C0,$D2,$8C,$47,$26,$BF,$E0,$BF,$7F,$F9,$0A
	.dc.b	$69,$1F,$E4,$AF,$CD,$F7,$5E,$FF,$F3,$9D,$BF,$17,$5F,$DB,$2B,$53
	.dc.b	$9D,$BB,$4F,$D1,$BE,$3B,$FD,$78,$E0,$66,$B8,$3F,$12,$F9,$CE,$E3
	.dc.b	$CC,$6D,$C4,$6A,$3E,$F1,$D3,$0F,$CF,$DF,$82,$32,$72,$8C,$DF,$82
	.dc.b	$3F,$BC,$57,$39,$F6,$C6,$F0,$FE,$F9,$7C,$E7,$35,$B9,$F7,$C7,$4C
	.dc.b	$3C,$B7,$E7,$26,$9F,$D6,$70,$F2,$3F,$D6,$E3,$D1,$7D,$B1,$FE,$CD
	.dc.b	$3B,$6D,$F8,$B8,$2D,$FF,$87,$30,$E7,$39,$3F,$B9,$3F,$CF,$5F,$F0
	.dc.b	$25,$BD,$F7,$B1,$CA,$D7,$B7,$FA,$B5,$6E,$75,$39,$8F,$2D,$72,$BE
	.dc.b	$FD,$7E,$49,$E1,$1E,$2B,$F3,$FA,$2F,$7D,$E1,$74,$9C,$C6,$2B,$81
	.dc.b	$FE,$1B,$AE,$15,$B8,$C4,$EF,$3F,$E7,$B9,$F8,$DF,$ED,$D7,$C2,$B9
	.dc.b	$C6,$2B,$86,$3F,$E0,$BF,$7D,$E1,$FC,$36,$DF,$F8,$AF,$C7,$7B,$FF
	.dc.b	$E3,$18,$FF,$0F,$4E,$B1,$FE,$9A,$D7,$F7,$AD,$7F,$C8,$8D,$7F,$CA
	.dc.b	$6D,$4F,$3C,$3F,$C4,$D7,$E0,$E5,$F2,$63,$F8,$08,$C6,$1F,$E2,$FF
	.dc.b	$79,$EF,$7D,$B6,$C7,$A7,$37,$BF,$9A,$93,$FE,$89,$AD,$E9,$FE,$8F
	.dc.b	$6B,$FC,$59,$AF,$F9,$75,$DE,$7E,$6F,$87,$FF,$36,$D7,$8E,$FC,$D1
	.dc.b	$8B,$FE,$E3,$EF,$FD,$5B,$FF,$1D,$36,$F8,$FC,$03,$E2,$3F,$29,$39
	.dc.b	$BF,$F3,$CB,$F7,$5F,$E5,$2D,$7F,$CE,$0D,$77,$CB,$FA,$D6,$DF,$F9
	.dc.b	$94,$5B,$EC,$FB,$D7,$EB,$27,$31,$FA,$5F,$7D,$F0,$1C,$F0,$46,$2B
	.dc.b	$FD,$64,$D8,$FF,$5C,$35,$FE,$7F,$F0,$0F,$7F,$FD,$3B,$1F,$E8,$78
	.dc.b	$C7,$AB,$FE,$BB,$5F,$00,$EA,$C5,$F0,$BF,$F4,$DF,$C7,$B8,$BF,$CB
	.dc.b	$6D,$8F,$F6,$23,$5F,$F6,$53,$5F,$EA,$3E,$FB,$E0,$3F,$D5,$3F,$00
	.dc.b	$F8,$1A,$7F,$A6,$F9,$BF,$CE,$15,$BB,$FF,$3F,$35,$FF,$6B,$35,$EB
	.dc.b	$7F,$E6,$E8,$C4,$7F,$A9,$5E,$13,$D2,$7B,$70,$65,$B4,$E5,$7C,$9B
	.dc.b	$E4,$B3,$2F,$81,$7C,$17,$AB,$E0,$E4,$FF,$A4,$E7,$2D,$9F,$F4,$0C
	.dc.b	$62,$3F,$DD,$CD,$7F,$BA,$B5,$FF,$5E,$3B,$CF,$FD,$24,$DC,$F3,$5F
	.dc.b	$2B,$E7,$BC,$BF,$07,$FE,$FB,$BF,$F4,$C3,$FE,$F8,$C3,$FD,$4C,$E6
	.dc.b	$FF,$DF,$11,$FE,$A9,$F8,$37,$C2,$7B,$9B,$1C,$3D,$BE,$A6,$66,$46
	.dc.b	$48,$7F,$98,$27,$8B,$63,$FE,$00,$46,$30,$FF,$56,$36,$FC,$75,$B3
	.dc.b	$C9,$5C,$0E,$EF,$FB,$D6,$B8,$C6,$1F,$7D,$F9,$A7,$89,$0E,$15,$FE
	.dc.b	$E7,$BF,$04,$7F,$BD,$9A,$B0,$F9,$0F,$F9,$E9,$F9,$0F,$FB,$F1,$CD
	.dc.b	$F8,$23,$B3,$E4,$7E,$1B,$97,$D8,$47,$04,$AF,$26,$1E,$B7,$E1,$C7
	.dc.b	$39,$E1,$18,$36,$F8,$1C,$3E,$B9,$E4,$F9,$F0,$F8,$8F,$23,$47,$9A
	.dc.b	$4E,$CE,$86,$8F,$92,$5A,$F2,$1E,$97,$81,$D4,$61,$F3,$30,$F8,$F5
	.dc.b	$DA,$7E,$75,$F9,$9D,$E1,$F7,$DF,$09,$A3,$CC,$78,$67,$E0,$BE,$43
	.dc.b	$89,$C2,$BB,$BE,$4C,$3C,$F3,$C1,$F4,$47,$53,$D1,$87,$9F,$0F,$8F
	.dc.b	$3C,$27,$C0,$73,$AF,$55,$FE,$12,$39,$4F,$4B,$FC,$59,$CE,$3D,$27
	.dc.b	$98,$F8,$D3,$E6,$78,$D7,$94,$8F,$72,$1C,$6B,$C8,$F0,$33,$5E,$0C
	.dc.b	$3E,$54,$76,$C3,$D7,$7D,$23,$93,$F8,$86,$9E,$E3,$A1,$2B,$A3,$EA
	.dc.b	$C3,$E0,$1E,$92,$D7,$68,$E5,$3F,$05,$FD,$54,$7D,$F8,$C6,$1F,$1F
	.dc.b	$0F,$C1,$5F,$3A,$D4,$7A,$A7,$DF,$39,$BF,$C2,$DF,$2C,$3E,$9E,$1C
	.dc.b	$27,$CC,$D7,$F0,$D3,$EA,$6E,$1E,$09,$F4,$47,$C2,$25,$F7,$7F,$D5
	.dc.b	$27,$E1,$CF,$F5,$5B,$FC,$19,$B1,$A7,$F4,$66,$0C,$46,$DF,$39,$88
	.dc.b	$FE,$4B,$D5,$E2,$9F,$06,$3F,$C9,$A3,$13,$E6,$C1,$18,$BF,$C6,$C6
	.dc.b	$1F,$82,$62,$34,$FD,$23,$FA,$03,$8A,$FC,$C3,$71,$F2,$CF,$F3,$B8
	.dc.b	$C3,$F8,$E6,$BF,$85,$6E,$1F,$9B,$BF,$14,$FE,$D6,$31,$3A,$C3,$FA
	.dc.b	$73,$67,$C6,$7A,$0F,$31,$E1,$31,$87,$99,$B8,$79,$A3,$E3,$46,$0F
	.dc.b	$31,$4F,$31,$6F,$C6,$DB,$F1,$B4,$F1,$9E,$97,$2D,$33,$3F,$C2,$A7
	.dc.b	$EB,$D7,$F4,$DB,$F6,$57,$A4,$E0,$5A,$E1,$7F,$6E,$77,$1E,$A9,$DB
	.dc.b	$4D,$BF,$6C,$F3,$1C,$EB,$F7,$57,$EC,$8E,$31,$E3,$77,$3C,$E3,$77
	.dc.b	$EC,$8E,$B3,$F4,$1D,$3F,$C4,$2B,$DA,$FF,$67,$BF,$3D,$F5,$8E,$E6
	.dc.b	$4E,$EE,$6B,$A9,$F6,$63,$67,$F4,$69,$E6,$F1,$33,$87,$E2,$A7,$8D
	.dc.b	$7D,$F3,$A1,$EA,$9F,$A6,$D7,$E2,$1E,$9C,$3E,$73,$E4,$3B,$4F,$47
	.dc.b	$F0,$E7,$82,$FF,$C7,$FF,$08,$F7,$FF,$C8,$BE,$3B,$E0,$3C,$99,$7F
	.dc.b	$68,$8C,$5F,$0A,$98,$7F,$03,$C3,$93,$63,$F8,$32,$8C,$BF,$C3,$9B
	.dc.b	$8B,$0F,$ED,$11,$86,$D7,$93,$1F,$F2,$EC,$7F,$6D,$F3,$9F,$07,$9F
	.dc.b	$84,$99,$7F,$7E,$CB,$A3,$FE,$00,$AD,$B9,$BE,$57,$E6,$9C,$CF,$F2
	.dc.b	$26,$BF,$DE,$CF,$F7,$6C,$6C,$FF,$33,$C7,$0A,$EF,$3A,$3F,$DB,$D8
	.dc.b	$3D,$CF,$F3,$6F,$C1,$BE,$13,$FC,$DF,$1F,$EE,$6A,$F2,$1F,$E1,$67
	.dc.b	$0A,$FA,$D7,$18,$D8,$70,$FF,$3A,$6B,$E9,$6B,$FE,$15,$6B,$FE,$24
	.dc.b	$78,$C7,$6A,$F3,$9E,$6F,$A0,$FF,$42,$FC,$2B,$E0,$FE,$A7,$C0,$4C
	.dc.b	$7C,$73,$96,$CF,$E0,$A3,$11,$FD,$39,$AF,$CD,$6B,$FE,$29,$77,$87
	.dc.b	$CC,$6E,$D9,$AF,$C2,$73,$F9,$3D,$8F,$C3,$7A,$B3,$1F,$1C,$65,$FE
	.dc.b	$B1,$82,$3C,$84,$8C,$BB,$7E,$86,$1E,$38,$FC,$97,$E1,$5F,$0B,$E1
	.dc.b	$BF,$D4,$FF,$0D,$F9,$7B,$FF,$26,$B5,$60,$9E,$11,$E2,$7E,$21,$1C
	.dc.b	$1F,$29,$DB,$E2,$1D,$3F,$B5,$F8,$8F,$F5,$BF,$C3,$FF,$8B,$49,$D9
	.dc.b	$E4,$DF,$2C,$BF,$C3,$CF,$F8,$CF,$2F,$97,$1F,$E6,$48,$E2,$EA,$79
	.dc.b	$3C,$3E,$1F,$B9,$1F,$C8,$D3,$E0,$31,$1E,$71,$6A,$CF,$17,$E9,$AF
	.dc.b	$85,$7E,$6B,$E1,$FF,$4C,$DE,$8F,$F6,$9C,$62,$3A,$7C,$33,$9B,$F4
	.dc.b	$90,$F8,$57,$C2,$FF,$9C,$6F,$86,$5F,$B4,$6F,$36,$72,$FF,$6F,$E5
	.dc.b	$CE,$BF,$D0,$B8,$72,$C7,$FD,$0F,$39,$7B,$63,$C5,$DB,$5C,$9C,$32
	.dc.b	$F2,$D7,$38,$57,$38,$C6,$5E,$D6,$E3,$FE,$89,$78,$4E,$F0,$FF,$46
	.dc.b	$36,$FF,$D3,$77,$C6,$7C,$55,$C3,$0F,$0C,$70,$6D,$70,$C3,$B4,$F0
	.dc.b	$35,$8F,$B2,$7D,$4D,$58,$77,$79,$19,$C7,$C8,$FB,$B0,$EE,$F0,$C3
	.dc.b	$BB,$C0,$F8,$E6,$EF,$94,$F0,$BE,$D3,$C4,$CC,$C7,$FC,$AF,$7F,$55
	.dc.b	$FE,$6C,$D7,$D6,$6B,$9D,$F0,$FF,$E5,$99,$46,$27,$53,$98,$C4,$7F
	.dc.b	$97,$FE,$1F,$FA,$DA,$69,$1F,$EB,$EF,$C7,$F8,$57,$C3,$FF,$52,$DB
	.dc.b	$F0,$55,$FE,$D2,$AD,$4E,$76,$FF,$38,$C6,$37,$FA,$15,$B3,$E3,$E1
	.dc.b	$D2,$33,$1E,$33,$37,$CD,$FE,$B1,$5C,$DF,$9E,$73,$C3,$F2,$97,$CE
	.dc.b	$32,$73,$8C,$D7,$38,$FF,$6E,$57,$39,$F5,$C6,$F0,$FF,$70,$5F,$39
	.dc.b	$CD,$6E,$7D,$B5,$E6,$C3,$D3,$39,$8F,$39,$92,$69,$FD,$B3,$0F,$33
	.dc.b	$FD,$BA,$3D,$57,$E0,$C7,$FB,$FC,$ED,$B8,$79,$38,$ED,$86,$1D,$27
	.dc.b	$47,$F6,$C7,$FA,$33,$FD,$D2,$DF,$0E,$F7,$B9,$DF,$2E,$DE,$B5,$A9
	.dc.b	$CE,$1C,$AB,$F9,$D4,$E7,$0E,$D7,$C3,$0E,$91,$E1,$C3,$CF,$EC,$BE
	.dc.b	$1D,$E2,$F4,$9C,$C6,$2B,$81,$FE,$0A,$AE,$15,$B8,$C4,$EF,$2F,$E6
	.dc.b	$D9,$F8,$1F,$E9,$D7,$C2,$B9,$C6,$2B,$86,$7F,$E1,$EF,$87,$7C,$3F
	.dc.b	$D3,$AF,$99,$18,$BF,$F6,$33,$63,$FD,$9A,$D7,$FC,$58,$E6,$BF,$C6
	.dc.b	$FD,$EF,$87,$FD,$AD,$FF,$87,$AB,$4F,$F9,$4E,$33,$7F,$D3,$E7,$2F
	.dc.b	$F6,$68,$CD,$F9,$F0,$FC,$0E,$1D,$79,$7E,$CF,$C3,$3E,$1B,$D5,$8A
	.dc.b	$31,$9F,$FB,$7D,$B1,$FE,$E6,$6B,$FE,$45,$73,$8F,$AB,$DE,$FC,$75
	.dc.b	$7F,$8D,$A7,$2F,$F9,$AE,$33,$5F,$D9,$E7,$2F,$F7,$E8,$CD,$7F,$93
	.dc.b	$27,$38,$A7,$37,$F9,$BB,$7B,$3F,$6D,$FF,$2C,$35,$D3,$5F,$D1,$35
	.dc.b	$E3,$C5,$FA,$69,$7C,$2B,$E1,$7F,$3F,$7F,$87,$CF,$FC,$DB,$5A,$9C
	.dc.b	$DF,$1B,$E1,$3F,$E5,$6B,$E0,$FC,$48,$CE,$1F,$E0,$7F,$83,$7C,$17
	.dc.b	$FB,$B2,$3D,$91,$8F,$57,$FC,$E5,$1F,$B9,$87,$C1,$B9,$B3,$29,$E0
	.dc.b	$F0,$F8,$3F,$90,$E9,$AE,$77,$FF,$4B,$C6,$23,$FD,$01,$6F,$94,$FF
	.dc.b	$4D,$B6,$3B,$B5,$F2,$B5,$FF,$44,$7C,$13,$E0,$BF,$D4,$6D,$8F,$3B
	.dc.b	$5F,$6B,$5E,$31,$F7,$78,$BF,$D1,$B3,$FE,$A2,$9F,$E3,$4D,$E0,$7C
	.dc.b	$13,$E0,$B8,$71,$7E,$16,$B8,$59,$A7,$1A,$E1,$1F,$E8,$CF,$81,$7C
	.dc.b	$0F,$FA,$E9,$B1,$F3,$9A,$FD,$46,$BF,$EA,$77,$39,$7F,$A9,$1B,$87
	.dc.b	$E5,$DB,$3F,$E5,$CF,$84,$7C,$17,$A2,$3F,$9B,$4E,$93,$4E,$3D,$2A
	.dc.b	$E3,$3F,$EA,$4F,$82,$7A,$7C,$E1,$F0,$4F,$6B,$0D,$8E,$13,$BA,$E3
	.dc.b	$1F,$EA,$E9,$D3,$FE,$5D,$6E,$9F,$EE,$D8,$EB,$7D,$0E,$57,$23,$94
	.dc.b	$65,$EE,$43,$A3,$97,$FC,$D5,$7E,$26,$CE,$6B,$FD,$9F,$7E,$2D,$3F
	.dc.b	$81,$5E,$E7,$FD,$A6,$D8,$FE,$0D,$3B,$BF,$EC,$4D,$53,$B7,$FB,$33
	.dc.b	$5F,$69,$EA,$3D,$15,$DA,$FC,$97,$DB,$0F,$81,$5E,$83,$81,$EE,$38
	.dc.b	$C7,$A2,$78,$C7,$92,$78,$D7,$58,$E2,$FC,$08,$CD,$71,$9C,$BE,$17
	.dc.b	$A9,$EA,$AE,$AF,$A8,$F2,$12,$BA,$3C,$8E,$07,$D5,$3F,$97,$1F,$1A
	.dc.b	$3F,$87,$B4,$FE,$23,$7F,$5D,$DB,$F2,$49,$1F,$90,$E9,$F4,$CF,$63
	.dc.b	$C5,$7F,$5A,$33,$1D,$4C,$9F,$61,$CD,$74,$8F,$CB,$46,$2B,$8E,$28
	.dc.b	$C4,$FC,$43,$F4,$6E,$1E,$A6,$0D,$57,$57,$F4,$11,$BC,$3E,$A4,$60
	.dc.b	$DD,$F8,$27,$58,$7D,$59,$D1,$DA,$FA,$5F,$E6,$AB,$73,$F0,$E3,$0F
	.dc.b	$43,$D1,$87,$AD,$C1,$E8,$70,$F5,$C3,$C4,$EA,$7C,$E7,$47,$47,$4A
	.dc.b	$F0,$BF,$3E,$34,$7E,$F6,$FF,$45,$7D,$A3,$17,$F6,$EB,$B6,$1E,$F9
	.dc.b	$E2,$D2,$47,$B5,$FB,$CF,$03,$07,$DD,$30,$78,$DD,$BE,$83,$E0,$61
	.dc.b	$E8,$C3,$CA,$DB,$F2,$4E,$EF,$E0,$DF,$27,$D3,$1F,$88,$3B,$DF,$26
	.dc.b	$D7,$29,$DC,$7D,$A6,$9E,$E7,$D2,$79,$CB,$87,$C7,$BF,$83,$19,$24
	.dc.b	$65,$E3,$7F,$84,$9F,$31,$C1,$C4,$7C,$53,$D9,$1D,$27,$94,$75,$9D
	.dc.b	$BF,$9B,$36,$79,$30,$F8,$CF,$C1,$32,$79,$1C,$BE,$FA,$F1,$61,$F8
	.dc.b	$C8,$DD,$7C,$76,$C7,$C7,$9D,$C7,$E3,$4F,$96,$77,$C3,$E4,$C7,$BA
	.dc.b	$3E,$0C,$F5,$9C,$C7,$BD,$AF,$AD,$B8,$7C,$0B,$F7,$9E,$38,$CE,$1E
	.dc.b	$A7,$EB,$57,$C5,$BE,$D8,$79,$A3,$F1,$67,$98,$F8,$07,$90,$F5,$1E
	.dc.b	$E3,$D4,$F1,$9F,$C6,$C7,$9E,$30,$7D,$5A,$F1,$3E,$F8,$F1,$1D,$27
	.dc.b	$CE,$F8,$5F,$3D,$FB,$2F,$E6,$4F,$8C,$F3,$9E,$33,$CE,$74,$33,$3D
	.dc.b	$1D,$3D,$23,$F7,$F8,$7C,$96,$E1,$E6,$AF,$51,$DE,$BD,$AF,$D1,$C3
	.dc.b	$DA,$FE,$2C,$F4,$3E,$DC,$3D,$11,$F1,$8E,$51,$F0,$DD,$9D,$8F,$44
	.dc.b	$F6,$C3,$C9,$1F,$C0,$9A,$FE,$59,$B1,$FE,$11,$6C,$F9,$27,$CC,$78
	.dc.b	$0F,$31,$F0,$9F,$A5,$F0,$4F,$82,$FE,$B6,$D9,$FF,$35,$B5,$FF,$38
	.dc.b	$B5,$FE,$C6,$E6,$D5,$FE,$93,$8C,$36,$BF,$C7,$9A,$BE,$03,$FC,$A7
	.dc.b	$5A,$C3,$FD,$E8,$D4,$E6,$78,$73,$7F,$8E,$FC,$97,$C0,$FF,$72,$6D
	.dc.b	$FF,$A0,$9A,$FF,$A2,$1A,$FF,$7A,$73,$9F,$A2,$07,$FB,$BA,$7B,$1E
	.dc.b	$8C,$5E,$B6,$27,$75,$C5,$E0,$63,$4F,$F5,$A6,$1D,$9E,$B9,$7C,$1B
	.dc.b	$FF,$2A,$7C,$0B,$E0,$7F,$C1,$6D,$C3,$FD,$36,$D7,$FD,$42,$D7,$FC
	.dc.b	$26,$E7,$1F,$13,$70,$FF,$62,$36,$7D,$B6,$F8,$1F,$0E,$12,$7D,$07
	.dc.b	$C2,$AE,$0F,$C0,$9F,$F3,$57,$C0,$3D,$EF,$F1,$0B,$67,$FD,$62,$D7
	.dc.b	$FD,$6E,$D7,$FC,$52,$E7,$2F,$90,$DC,$BF,$DB,$4D,$96,$7F,$87,$E6
	.dc.b	$F0,$E1,$27,$CE,$2B,$85,$6E,$33,$0F,$4B,$CE,$48,$F3,$B5,$63,$E4
	.dc.b	$8F,$79,$F8,$1F,$81,$B3,$D2,$2B,$C3,$2F,$BB,$C0,$AF,$8F,$F0,$3F
	.dc.b	$18,$95,$F1,$9A,$F1,$3F,$23,$8A,$38,$C7,$D4,$CB,$F5,$1F,$03,$88
	.dc.b	$E0,$7C,$19,$C6,$7F,$9B,$AD,$CF,$C6,$F8,$1F,$F3,$D1,$3A,$FF,$D4
	.dc.b	$9F,$01,$F3,$E5,$39,$78,$4E,$B7,$FF,$55,$4E,$67,$C2,$FF,$A1,$6B
	.dc.b	$92,$F8,$1D,$92,$7E,$BC,$BD,$9C,$47,$29,$B9,$7E,$AE,$B5,$3F,$7B
	.dc.b	$E0,$3F,$62,$DE,$7F,$F5,$E4,$62,$3E,$1F,$C0,$3E,$07,$ED,$C3,$E0
	.dc.b	$1E,$FF,$EF,$A1,$7C,$72,$FD,$D3,$79,$B4,$D9,$FE,$0B,$8F,$5A,$FE
	.dc.b	$0F,$87,$49,$CE,$3E,$08,$C4,$76,$CB,$C1,$18,$C3,$D0,$EB,$0E,$95
	.dc.b	$E7,$CB,$CD,$18,$9E,$17,$CA,$7C,$AD,$AF,$0E,$3E,$49,$E3,$8F,$8D
	.dc.b	$DC,$E4,$FE,$3F,$8F,$8A,$37,$59,$96,$3E,$19,$E3,$1E,$53,$8E,$1E
	.dc.b	$17,$78,$77,$7C,$66,$CF,$84,$75,$C3,$C0,$EB,$1F,$85,$3C,$63,$67
	.dc.b	$19,$DB,$DB,$1F,$36,$3E,$58,$F1,$9D,$70,$F2,$46,$63,$81,$FE,$BA
	.dc.b	$29,$FE,$BC,$7E,$8B,$6F,$F1,$2E,$7D,$87,$C0,$F1,$8F,$8B,$18,$7A
	.dc.b	$52,$31,$1C,$67,$FD,$8E,$BE,$06,$91,$FF,$65,$B7,$3F,$C4,$FB,$EF
	.dc.b	$81,$FE,$AD,$1F,$8C,$AF,$EA,$F5,$F5,$EB,$5B,$FD,$99,$C9,$F4,$63
	.dc.b	$2F,$D3,$33,$5D,$AF,$AE,$FF,$17,$0E,$91,$8A,$FF,$70,$C6,$30,$ED
	.dc.b	$1E,$B3,$B4,$ED,$FD,$74,$ED,$F0,$1B,$BF,$05,$72,$9D,$47,$B6,$75
	.dc.b	$7F,$B7,$C3,$D3,$39,$AD,$CF,$9E,$BC,$37,$CD,$A4,$DB,$FB,$B6,$1F
	.dc.b	$09,$FE,$F1,$1F,$0A,$BB,$63,$FE,$03,$9D,$B6,$FC,$5C,$56,$FF,$C0
	.dc.b	$F8,$73,$9C,$9F,$D5,$1F,$EA,$0F,$F6,$CB,$7C,$0B,$DA,$E5,$6B,$DB
	.dc.b	$FD,$0E,$B7,$3A,$9C,$C7,$96,$B9,$5F,$7E,$EF,$24,$F0,$8F,$15,$F9
	.dc.b	$FD,$87,$C0,$BC,$5E,$93,$98,$C5,$70,$3F,$C0,$75,$C2,$B7,$18,$9D
	.dc.b	$E7,$FD,$4B,$3F,$1B,$FD,$6E,$F8,$57,$38,$C5,$70,$C7,$FC,$51,$F0
	.dc.b	$2E,$CF,$86,$DB,$FE,$C1,$F0,$2F,$37,$35,$27,$FC,$03,$5B,$C7,$FB
	.dc.b	$57,$7B,$0E,$15,$AB,$E1,$5F,$C0,$6F,$83,$FD,$0C,$38,$7F,$A3,$35
	.dc.b	$FF,$22,$FB,$EF,$6F,$EC,$B6,$FC,$FF,$57,$89,$91,$FE,$16,$9D,$57
	.dc.b	$9B,$15,$EA,$3F,$BD,$CE,$9F,$23,$5F,$EB,$0D,$7C,$77,$F9,$A7,$FC
	.dc.b	$B3,$F1,$4F,$17,$10,$8C,$63,$FC,$65,$B1,$FC,$8D,$AF,$C3,$F8,$77
	.dc.b	$C4,$72,$6D,$7F,$21,$6C,$FF,$2B,$6B,$F2,$FE,$01,$F0,$2F,$5B,$F9
	.dc.b	$44,$61,$BE,$FB,$9B,$35,$BB,$FE,$6D,$C9,$F8,$57,$FC,$7F,$EB,$BD
	.dc.b	$0F,$EF,$5C,$FF,$E7,$18,$C4,$FF,$A9,$E7,$3A,$FF,$7F,$9C,$CF,$F8
	.dc.b	$A2,$FA,$37,$9B,$F9,$E3,$F8,$F6,$C6,$A3,$F0,$CD,$7F,$C1,$5E,$EF
	.dc.b	$E5,$5F,$F1,$B7,$B2,$FB,$FF,$F4,$FB,$63,$FC,$AF,$3B,$7F,$C7,$2D
	.dc.b	$CF,$FC,$71,$3B,$9F,$CD,$B6,$3F,$D1,$D3,$BC,$BC,$38,$7A,$23,$A5
	.dc.b	$FE,$6D,$AE,$8F,$EB,$1C,$EF,$AE,$F2,$E7,$27,$FC,$D7,$39,$6C,$E5
	.dc.b	$F1,$CF,$CC,$7F,$D4,$F3,$B7,$FC,$A6,$D7,$F5,$8D,$7F,$C6,$0D,$78
	.dc.b	$CF,$FA,$83,$D7,$E9,$1F,$EA,$EF,$87,$BE,$69,$FF,$4B,$47,$F9,$83
	.dc.b	$D8,$FF,$4A,$CE,$63,$FD,$06,$D7,$55,$C3,$5F,$F5,$B4,$E5,$B7,$9D
	.dc.b	$7F,$D4,$F5,$E0,$89,$7F,$E8,$EC,$3F,$D4,$67,$82,$33,$1D,$8D,$4E
	.dc.b	$2F,$AC,$60,$87,$F9,$E3,$0F,$16,$1F,$E5,$36,$C7,$FA,$C6,$BC,$8F
	.dc.b	$FA,$E2,$FA,$57,$9A,$39,$DF,$67,$A4,$F0,$58,$79,$DF,$8B,$19,$8C
	.dc.b	$47,$03,$9D,$75,$70,$4A,$FF,$66,$DF,$5C,$3C,$F3,$F5,$CE,$27,$A4
	.dc.b	$EE,$43,$E3,$12,$7D,$04,$AF,$03,$DA,$38,$39,$7A,$1C,$63,$E4,$61
	.dc.b	$F1,$23,$33,$C5,$F4,$4E,$9F,$29,$F1,$2F,$83,$E9,$3A,$1C,$A3,$83
	.dc.b	$C8,$C1,$0E,$13,$E5,$7A,$1D,$AF,$E2,$9F,$1B,$0F,$C4,$DF,$82,$30
	.dc.b	$FC,$BC,$3C,$73,$AA,$F8,$38,$78,$B0,$F6,$11,$C1,$23,$C9,$3F,$72
	.dc.b	$7C,$8F,$38,$F5,$1E,$89,$F9,$E4,$7F,$1C,$78,$0E,$77,$E7,$BF,$C4
	.dc.b	$36,$3E,$5D,$77,$AF,$14,$7C,$23,$E5,$9D,$CE,$A6,$23,$83,$8B,$F6
	.dc.b	$3F,$37,$0F,$93,$87,$82,$3C,$87,$33,$B4,$F0,$71,$3F,$31,$FA,$53
	.dc.b	$F3,$30,$F9,$D1,$CA,$FE,$25,$FD,$9B,$E1,$3E,$87,$C2,$FB,$AB,$E0
	.dc.b	$3E,$96,$CF,$86,$BD,$73,$E5,$7A,$E1,$F1,$2F,$AD,$79,$DE,$B5,$CC
	.dc.b	$F6,$BA,$9F,$A9,$1E,$CC,$3E,$13,$F6,$A7,$4F,$C7,$8F,$73,$F0,$0F
	.dc.b	$76,$1F,$01,$FC,$0E,$1E,$47,$E8,$5F,$D8,$8C,$47,$73,$E0,$1E,$4A
	.dc.b	$F0,$E1,$F1,$DB,$5E,$D6,$CF,$42,$1F,$5A,$3D,$78,$7C,$0A,$F9,$18
	.dc.b	$7C,$A9,$F7,$E1,$F2,$AB,$DB,$7C,$4E,$6E,$23,$58,$78,$27,$D0,$7C
	.dc.b	$B3,$E3,$39,$9F,$05,$76,$BF,$05,$7E,$78,$F7,$1F,$00,$E2,$5C,$3E
	.dc.b	$03,$67,$E0,$34,$F2,$DF,$73,$E9,$BF,$8B,$6C,$FE,$05,$B3,$F1,$1B
	.dc.b	$3F,$6C,$FB,$27,$DB,$3E,$CD,$7D,$BA,$FD,$EB,$6B,$ED,$36,$7E,$8E
	.dc.b	$1F,$73,$0F,$83,$1E,$F3,$DA,$7C,$7B,$F6,$B6,$FD,$AD,$3D,$A7,$BC
	.dc.b	$F9,$F1,$F6,$9F,$92,$6C,$CC,$7C,$F7,$31,$B7,$FD,$E5,$4B,$FF,$01
	.dc.b	$47,$FB,$B6,$30,$D8,$FB,$5E,$C7,$DF,$9F,$F0,$0B,$63,$FC,$0E,$D9
	.dc.b	$FF,$01,$B5,$FE,$E0,$DA,$FF,$01,$B6,$FF,$B9,$36,$43,$04,$66,$30
	.dc.b	$D8,$CC,$7C,$5C,$3C,$AD,$C3,$CA,$D3,$CA,$53,$CA,$5C,$3C,$AF,$89
	.dc.b	$61,$E4,$87,$B6,$F9,$FF,$F1,$A7,$84,$F6,$9F,$4D,$86,$DF,$3C,$39
	.dc.b	$61,$C6,$30,$7C,$07,$0F,$FA,$B2,$78,$B6,$BF,$C6,$F3,$AC,$7C,$B3
	.dc.b	$C7,$1F,$F3,$5C,$F2,$8F,$F5,$73,$5F,$0C,$7F,$9C,$3E,$49,$EF,$79
	.dc.b	$F9,$65,$79,$04,$E9,$C3,$5F,$34,$70,$58,$79,$1E,$35,$FE,$4E,$AE
	.dc.b	$11,$FE,$C1,$6B,$F3,$5A,$FE,$09,$AB,$DC,$7B,$7F,$37,$D7,$93,$E7
	.dc.b	$9C,$B6,$72,$FC,$B9,$FF,$2F,$CE,$A3,$CF,$B7,$DD,$8C,$47,$FB,$55
	.dc.b	$AE,$6B,$FD,$2F,$18,$8F,$CB,$FE,$38,$F7,$7F,$07,$F2,$9F,$94,$9F
	.dc.b	$F7,$2B,$74,$F8,$95,$B9,$F9,$CD,$77,$CD,$C8,$FF,$3A,$C6,$67,$8D
	.dc.b	$FF,$AA,$23,$85,$7F,$BA,$DA,$FE,$5D,$AF,$EB,$3D,$2F,$F4,$E6,$7F
	.dc.b	$AE,$E4,$FC,$C7,$E0,$FF,$97,$3F,$D0,$73,$AA,$E1,$DF,$97,$F2,$F5
	.dc.b	$C1,$FC,$34,$ED,$3F,$EA,$B7,$FD,$FC,$D7,$F5,$4D,$7A,$9F,$AD,$33
	.dc.b	$F4,$BF,$A9,$F2,$E4,$76,$8F,$29,$87,$F4,$D1,$87,$47,$33,$E0,$47
	.dc.b	$FB,$3A,$4D,$3F,$DA,$3F,$01,$FE,$D4,$D7,$F3,$0D,$CF,$F8,$03,$6F
	.dc.b	$F8,$AF,$BF,$F8,$9A,$DF,$83,$FE,$AA,$AD,$51,$B5,$BB,$E7,$5B,$25
	.dc.b	$7E,$BE,$74,$FF,$22,$A7,$C0,$75,$24,$6F,$D8,$AF,$E6,$8E,$11,$8C
	.dc.b	$7F,$65,$6A,$FF,$5F,$FC,$07,$F2,$DB,$FD,$F3,$5E,$41,$C5,$7E,$EF
	.dc.b	$E0,$1D,$1F,$BE,$A7,$BC,$F7,$7F,$93,$61,$C3,$6B,$E2,$78,$0F,$F6
	.dc.b	$7E,$3C,$B2,$FE,$7F,$87,$1C,$7F,$A0,$CE,$5E,$B8,$F2,$76,$D7,$27
	.dc.b	$0C,$BC,$95,$CE,$15,$CE,$31,$97,$BD,$B8,$FF,$3D,$78,$4E,$F0,$FE
	.dc.b	$80,$DB,$FE,$AB,$7C,$67,$C5,$5C,$30,$F0,$C7,$06,$D7,$0C,$3B,$4F
	.dc.b	$03,$58,$FB,$27,$D4,$D7,$E9,$4F,$07,$91,$9C,$7C,$8F,$BA,$FB,$BC
	.dc.b	$0E,$58,$F9,$DE,$A7,$C8,$38,$5E,$AF,$B4,$71,$35,$7D,$8F,$A8,$FE
	.dc.b	$DD,$AF,$A8,$FD,$83,$87,$ED,$FC,$03,$E0,$3F,$69,$1F,$52,$30,$F0
	.dc.b	$6B,$D5,$AF,$EE,$FE,$03,$8B,$FE,$06,$6A,$69,$1F,$77,$B8,$F8,$0F
	.dc.b	$EC,$FB,$7E,$16,$79,$1A,$8C,$ED,$DE,$31,$BF,$D3,$77,$5F,$8D,$BE
	.dc.b	$93,$99,$F1,$1C,$0F,$9C,$70,$AF,$F0,$54,$F0,$8C,$9C,$2B,$A5,$F8
	.dc.b	$A3,$F9,$26,$1F,$83,$AE,$15,$E9,$9E,$17,$D2,$F8,$47,$F8,$5E,$F8
	.dc.b	$46,$23,$C4,$7F,$8B,$AF,$9D,$7B,$EF,$C3,$87,$AF,$0E,$2D,$BF,$64
	.dc.b	$E9,$FE,$0F,$7E,$BD,$7F,$C0,$11,$D8,$46,$6B,$CC,$7F,$83,$5F,$E2
	.dc.b	$8D,$C7,$FC,$37,$5F,$E0,$99,$CD,$76,$F8,$0C,$DF,$F8,$D7,$9D,$F0
	.dc.b	$18,$AF,$65,$5C,$3F,$A9,$CE,$67,$C7,$39,$AE,$37,$E3,$8C,$56,$F0
	.dc.b	$CD,$FF,$2D,$F5,$1E,$FB,$E0,$35,$7E,$19,$CC,$63,$1F,$F1,$EC,$61
	.dc.b	$F8,$13,$BA,$E5,$18,$36,$E2,$FE,$6C,$EE,$FD,$5F,$00,$EB,$A7,$C1
	.dc.b	$EE,$7C,$03,$CA,$CE,$7F,$E3,$BB,$E0,$FE,$C9,$B9,$7F,$67,$6B,$FD
	.dc.b	$CD,$AF,$F9,$85,$DF,$07,$F9,$43,$0E,$FE,$BF,$6A,$FE,$A5,$18,$BF
	.dc.b	$F2,$57,$C0,$3E,$03,$E9,$36,$3F,$C1,$79,$BD,$6C,$D4,$9F,$E0,$F5
	.dc.b	$BC,$3F,$BE,$35,$FF,$04,$35,$FF,$3B,$3B,$CB,$CD,$4E,$1E,$95,$E6
	.dc.b	$8C,$5F,$F9,$7F,$DE,$7B,$FF,$9A,$AF,$F4,$9C,$62,$7F,$8C,$B6,$3F
	.dc.b	$C1,$CD,$7F,$C3,$5E,$FB,$DD,$7B,$FF,$C8,$23,$0D,$F7,$FC,$95,$FE
	.dc.b	$17,$9C,$BF,$E1,$9F,$8B,$F9,$0F,$F2,$86,$B9,$CF,$F9,$1C,$62,$3F
	.dc.b	$C5,$5C,$9F,$E1,$2F,$80,$72,$7F,$A6,$78,$3F,$CE,$91,$8A,$FF,$3F
	.dc.b	$57,$F8,$60,$FC,$63,$FC,$D9,$AF,$F8,$AD,$AF,$6E,$6F,$E6,$71,$C5
	.dc.b	$ED,$BF,$FA,$9E,$3C,$5A,$E2,$BF,$D7,$BE,$FF,$E5,$5F,$F2,$5F,$C0
	.dc.b	$7F,$93,$F9,$CF,$F9,$2E,$75,$3F,$E1,$46,$BF,$E4,$76,$BF,$EB,$56
	.dc.b	$BD,$AF,$FA,$3C,$72,$CF,$FD,$28,$DC,$78,$4F,$FA,$EA,$FE,$6D,$7F
	.dc.b	$A6,$79,$9F,$01,$9D,$26,$5F,$EA,$59,$CB,$67,$FA,$AC,$62,$3F,$CA
	.dc.b	$4D,$7F,$CB,$CD,$7F,$D9,$8E,$EB,$C1,$3F,$1E,$7F,$48,$F9,$7E,$83
	.dc.b	$1F,$01,$FE,$F6,$9F,$F1,$BF,$92,$F8,$1E,$0D,$8F,$F5,$D4,$EF,$0F
	.dc.b	$F5,$A3,$75,$FF,$6E,$56,$EB,$9D,$F6,$F8,$19,$1F,$EC,$39,$D6,$1F
	.dc.b	$EB,$D6,$DF,$FB,$97,$0E,$95,$FE,$FA,$6B,$FE,$FB,$C3,$CA,$74,$7F
	.dc.b	$C9,$F1,$AA,$E7,$03,$81,$EA,$C3,$86,$1E,$1B,$E7,$39,$21,$EA,$3D
	.dc.b	$0F,$0C,$3C,$D5,$C2,$74,$78,$9D,$3E,$1B,$F1,$BE,$23,$47,$6B,$F1
	.dc.b	$3D,$8C,$57,$67,$07,$68,$FA,$35,$DA,$3A,$47,$27,$A9,$1C,$12,$7A
	.dc.b	$46,$5F,$B3,$7E,$B9,$EB,$3E,$23,$D5,$19,$9E,$33,$A9,$F2,$47,$27
	.dc.b	$80,$C3,$EA,$57,$18,$FB,$97,$E9,$C3,$E2,$9F,$1A,$BE,$F9,$C2,$7E
	.dc.b	$0E,$1D,$C9,$87,$78,$F0,$9D,$0F,$0D,$74,$C3,$E3,$5F,$39,$D9,$E5
	.dc.b	$C3,$C5,$3E,$58,$C1,$C9,$C4,$F9,$27,$55,$E7,$78,$D7,$9A,$FE,$29
	.dc.b	$1F,$92,$64,$F3,$1F,$44,$C9,$E4,$AE,$CE,$CF,$BB,$1C,$DF,$29,$EA
	.dc.b	$AF,$13,$E7,$38,$4F,$27,$57,$F0,$AF,$D2,$D8,$ED,$5C,$9F,$21,$E0
	.dc.b	$3C,$B8,$7D,$07,$D2,$FE,$2D,$CE,$1E,$5C,$3E,$74,$E6,$7C,$EF,$17
	.dc.b	$A1,$23,$CD,$1B,$AF,$13,$B7,$ED,$4F,$E3,$27,$AD,$7E,$06,$BA,$46
	.dc.b	$27,$C6,$79,$A7,$E4,$57,$28,$F6,$CE,$DF,$16,$1F,$69,$FA,$4D,$BF
	.dc.b	$54,$F9,$AF,$C0,$4A,$EC,$43,$DB,$7E,$A8,$F0,$E1,$F8,$AC,$3A,$47
	.dc.b	$1A,$F8,$AE,$9F,$B3,$1A,$7D,$66,$8F,$C4,$9F,$C7,$4A,$7E,$F0,$C1
	.dc.b	$F8,$62,$DF,$F3,$F6,$C7,$FA,$8A,$31,$1C,$4F,$B5,$1F,$B5,$F6,$3F
	.dc.b	$0D,$1F,$84,$C1,$18,$C7,$EB,$37,$0F,$9E,$DA,$F8,$95,$C1,$FB,$57
	.dc.b	$D5,$A7,$EE,$27,$83,$EF,$13,$C5,$FB,$33,$C5,$A7,$13,$99,$E7,$9E
	.dc.b	$78,$7F,$45,$6C,$FF,$5E,$6C,$FF,$55,$6B,$FD,$91,$B8,$7F,$57,$6E
	.dc.b	$1F,$D8,$1B,$3F,$CF,$DA,$FA,$1B,$87,$A6,$BF,$2F,$7E,$96,$D7,$A4
	.dc.b	$F5,$9E,$93,$D6,$FC,$E6,$DF,$CE,$69,$F3,$AF,$B9,$1C,$9F,$2E,$FC
	.dc.b	$53,$F7,$A3,$C1,$5E,$27,$C0,$78,$9F,$03,$63,$C0,$D3,$C0,$76,$C3
	.dc.b	$E7,$37,$0F,$9C,$D3,$E7,$14,$F9,$C5,$AF,$9C,$DA,$F9,$CD,$3E,$71
	.dc.b	$4F,$9C,$5B,$F6,$BF,$B3,$BF,$DC,$3C,$0F,$41,$F0,$F0,$F8,$CE,$23
	.dc.b	$D2,$DC,$3D,$2D,$3D,$25,$3D,$25,$C3,$D2,$DC,$3D,$2A,$11,$FB,$9F
	.dc.b	$7D,$EF,$FF,$97,$78,$DF,$02,$F8,$1C,$37,$1F,$F0,$E6,$3C,$9F,$F0
	.dc.b	$F6,$1C,$9C,$E3,$D7,$1F,$E0,$71,$4A,$27,$83,$A6,$BF,$C2,$9C,$D7
	.dc.b	$78,$E1,$3F,$E4,$D9,$CD,$7F,$8B,$F1,$F4,$FD,$27,$F9,$47,$E0,$5F
	.dc.b	$05,$EA,$E7,$99,$79,$8F,$F1,$B4,$E9,$C0,$70,$FA,$9A,$FF,$1E,$6B
	.dc.b	$FC,$8D,$AF,$49,$F1,$11,$F0,$5F,$CE,$F9,$8F,$F3,$37,$C1,$3E,$0F
	.dc.b	$E7,$FB,$12,$FC,$73,$96,$CF,$CA,$51,$97,$E7,$B5,$FE,$66,$D7,$F9
	.dc.b	$C3,$BD,$FE,$63,$75,$CD,$7E,$03,$BF,$FC,$F5,$F0,$2F,$82,$FC,$37
	.dc.b	$C1,$FE,$7B,$53,$FA,$0C,$D5,$B7,$E5,$B5,$FC,$1B,$5F,$E8,$ED,$7F
	.dc.b	$A6,$35,$FF,$35,$3C,$73,$FC,$13,$72,$DE,$1E,$4E,$CF,$F4,$87,$A7
	.dc.b	$FE,$B3,$E6,$7C,$1F,$E6,$8C,$FC,$1C,$CF,$F2,$74,$8F,$98,$D8,$CC
	.dc.b	$60,$E0,$FE,$AF,$1F,$EC,$53,$DF,$F0,$BF,$0F,$C1,$FE,$A3,$E1,$25
	.dc.b	$FE,$32,$93,$A8,$F0,$BE,$02,$38,$38,$3B,$7C,$03,$83,$F6,$FF,$2F
	.dc.b	$FE,$AF,$F8,$5F,$F6,$FF,$0F,$E9,$DB,$9F,$F7,$76,$DF,$FA,$57,$E1
	.dc.b	$3F,$00,$FF,$A5,$E3,$29,$A6,$23,$71,$9F,$1F,$C0,$47,$F5,$31,$A9
	.dc.b	$7C,$2F,$94,$93,$AF,$85,$AF,$B0,$F0,$46,$31,$FE,$01,$6A,$FF,$66
	.dc.b	$7C,$2F,$F7,$4B,$FD,$BB,$5E,$4D,$AF,$E1,$1F,$0A,$EC,$FE,$45,$4F
	.dc.b	$81,$7C,$1F,$F8,$6F,$0E,$1B,$5F,$2B,$B9,$FE,$DB,$C7,$96,$5F,$EB
	.dc.b	$DC,$38,$E3,$FE,$BF,$9C,$BD,$71,$E4,$ED,$AE,$4E,$19,$79,$2B,$9C
	.dc.b	$2B,$9C,$63,$2F,$73,$71,$FF,$19,$3C,$27,$78,$7F,$8D,$5B,$7F,$ED
	.dc.b	$3B,$E3,$3E,$2A,$E1,$87,$86,$38,$36,$B8,$61,$DA,$78,$1A,$C7,$D9
	.dc.b	$3E,$A6,$AC,$3B,$BC,$8C,$E3,$E4,$7D,$D8,$77,$78,$61,$DD,$E0,$7C
	.dc.b	$73,$77,$CA,$78,$5F,$69,$E2,$66,$63,$FD,$4A,$FE,$AB,$FE,$56,$6B
	.dc.b	$EB,$35,$F0,$8F,$85,$FE,$AF,$28,$C4,$EA,$73,$18,$8F,$EB,$BF,$0B
	.dc.b	$FC,$D5,$34,$8F,$F3,$97,$E3,$FC,$1B,$E1,$7F,$B6,$6D,$F8,$2A,$1C
	.dc.b	$8F,$85,$3B,$DF,$E7,$C6,$AF,$9D,$F2,$DF,$FB,$8C,$62,$B7,$5C,$F0
	.dc.b	$FC,$7E,$1C,$E3,$E7,$1C,$CE,$F8,$73,$7F,$CF,$B7,$CD,$D9,$CF,$0F
	.dc.b	$F4,$25,$73,$BF,$54,$E4,$F5,$C6,$67,$CB,$1D,$B0,$F2,$DF,$5B,$E2
	.dc.b	$DA,$DC,$7A,$30,$F2,$5F,$FA,$4E,$B4,$F6,$8F,$F0,$B6,$9F,$E0,$9C
	.dc.b	$7D,$CF,$F8,$2E,$7D,$D7,$CF,$1F,$F0,$0D,$69,$F5,$F0,$72,$7D,$B8
	.dc.b	$5B,$F5,$CD,$54,$F8,$5D,$62,$E9,$7C,$26,$EB,$59,$7F,$84,$EB,$53
	.dc.b	$9B,$E3,$7C,$2B,$CB,$18,$9E,$47,$8F,$E0,$1F,$0A,$F8,$5C,$DF,$F8
	.dc.b	$7E,$73,$18,$CF,$FA,$5C,$62,$7E,$0D,$F3,$BE,$51,$83,$93,$8A,$FE
	.dc.b	$71,$9F,$95,$E7,$19,$CF,$FC,$71,$F0,$AE,$BF,$A1,$5F,$9C,$8C,$61
	.dc.b	$FE,$44,$6C,$7F,$92,$1A,$FF,$B3,$1C,$D7,$F9,$47,$B5,$F0,$AF,$27
	.dc.b	$E9,$56,$B0,$FF,$22,$4E,$5F,$F1,$EC,$67,$0F,$ED,$53,$97,$FA,$94
	.dc.b	$66,$FA,$DF,$DD,$F8,$57,$C0,$7A,$AD,$18,$C7,$FC,$BA,$D8,$FF,$30
	.dc.b	$B5,$FF,$72,$B9,$CB,$D5,$4D,$FB,$3F,$8D,$C7,$BC,$63,$0F,$F2,$4C
	.dc.b	$ED,$FF,$2B,$DF,$37,$FC,$C5,$19,$C3,$FB,$67,$C0,$BE,$0D,$F0,$1F
	.dc.b	$E7,$28,$C3,$7E,$0F,$C5,$5F,$E7,$89,$CB,$FE,$7A,$F8,$3F,$82,$FF
	.dc.b	$9E,$9A,$E7,$0F,$F2,$A4,$62,$3F,$8C,$9F,$C6,$A9,$F2,$1F,$E5,$96
	.dc.b	$C7,$FA,$11,$AF,$FA,$21,$AF,$21,$EB,$FE,$2D,$FF,$45,$B5,$FF,$47
	.dc.b	$B5,$E0,$D7,$C6,$FF,$87,$89,$C7,$E6,$E0,$FF,$47,$7C,$27,$C6,$AF
	.dc.b	$F4,$1D,$69,$B5,$BD,$BD,$B5,$B9,$FF,$0D,$7C,$27,$99,$FF,$4F,$B5
	.dc.b	$FF,$51,$B5,$F6,$35,$F8,$EE,$F0,$FF,$50,$36,$B3,$87,$FA,$DF,$E1
	.dc.b	$3E,$8D,$6A,$3C,$97,$FD,$22,$F8,$36,$38,$6B,$A8,$7C,$23,$D3,$FF
	.dc.b	$46,$36,$3F,$D6,$ED,$7F,$D7,$6D,$7E,$A3,$9C,$3F,$D6,$CD,$CB,$F0
	.dc.b	$2D,$9F,$F2,$C6,$4F,$84,$F1,$4E,$67,$C7,$86,$0F,$74,$F9,$FE,$09
	.dc.b	$F0,$5E,$9F,$31,$F0,$94,$D4,$4C,$3F,$D7,$73,$A6,$CE,$5F,$F2,$85
	.dc.b	$74,$72,$43,$FD,$EF,$3C,$5F,$F2,$B3,$74,$FF,$6A,$4F,$FB,$FA,$53
	.dc.b	$9D,$3F,$42,$D8,$FE,$F7,$3B,$8C,$47,$0B,$F2,$5F,$4C,$3D,$93,$E3
	.dc.b	$8C,$3F,$ED,$6B,$F4,$3E,$4A,$DB,$E8,$6C,$F0,$8F,$7D,$63,$0F,$2D
	.dc.b	$78,$1F,$DB,$DF,$4B,$FF,$78,$61,$E6,$3C,$F1,$E9,$AE,$58,$7B,$30
	.dc.b	$E4,$FF,$86,$2B,$84,$E2,$15,$D7,$0F,$0C,$7B,$5F,$0D,$EE,$F8,$CF
	.dc.b	$BE,$3E,$C3,$FC,$0E,$BE,$0B,$63,$FD,$C9,$18,$8E,$67,$12,$1F,$6B
	.dc.b	$E1,$3C,$91,$F4,$DB,$3E,$16,$CF,$0C,$7E,$1C,$62,$3E,$59,$F8,$E7
	.dc.b	$11,$F3,$CD,$C7,$1C,$3C,$6F,$03,$C4,$64,$F9,$45,$3F,$17,$3D,$23
	.dc.b	$8C,$F4,$9E,$4F,$78,$C9,$F9,$09,$E1,$3D,$0E,$E7,$08,$DB,$99,$E8
	.dc.b	$73,$71,$1A,$7C,$67,$96,$FE,$43,$6F,$E4,$34,$F5,$60,$C3,$C7,$1F
	.dc.b	$82,$9F,$DD,$46,$2B,$F0,$93,$F8,$98,$C4,$EA,$3A,$1E,$93,$46,$30
	.dc.b	$F4,$3E,$63,$47,$73,$C7,$3D,$1F,$B0,$71,$C3,$DD,$87,$9E,$30,$7C
	.dc.b	$23,$B9,$D3,$0F,$82,$DC,$3E,$0B,$4E,$E7,$23,$C4,$7D,$2C,$3C,$F3
	.dc.b	$C9,$B3,$C9,$B3,$C8,$E8,$7B,$2B,$CC,$E2,$3E,$0C,$61,$E3,$1B,$C3
	.dc.b	$E0,$3E,$1C,$3E,$E6,$1E,$16,$E1,$E1,$6C,$F8,$6F,$99,$DA,$F9,$BF
	.dc.b	$88,$9D,$CF,$73,$E0,$5F,$DC,$7E,$A4,$F2,$9F,$9F,$5C,$9F,$CC,$4F
	.dc.b	$0B,$FC,$03,$ED,$8D,$B9,$3E,$C9,$93,$EB,$1D,$9F,$1D,$79,$E3,$81
	.dc.b	$F0,$4D,$DF,$C1,$6D,$7C,$16,$9F,$06,$B9,$9E,$27,$C0,$7D,$CA,$F8
	.dc.b	$4F,$03,$E1,$1A,$AF,$85,$87,$D5,$AE,$73,$E8,$8C,$3C,$6B,$D0,$F1
	.dc.b	$BF,$87,$1F,$26,$38,$61,$F2,$5F,$CA,$C6,$5F,$25,$FC,$C7,$91,$F2
	.dc.b	$23,$58,$79,$2B,$94,$FC,$93,$C5,$5E,$FA,$DD,$F8,$A3,$D3,$3D,$8F
	.dc.b	$0B,$8B,$E9,$1F,$42,$BE,$2B,$B2,$46,$9F,$29,$F5,$4C,$D7,$D3,$69
	.dc.b	$F4,$CF,$29,$F0,$8E,$D5,$EF,$AF,$3C,$7B,$E3,$C9,$3E,$98,$E2,$78
	.dc.b	$F0,$FA,$2F,$63,$EC,$D7,$27,$C9,$7F,$DF,$CF,$44,$64,$F6,$B9,$AE
	.dc.b	$D3,$CE,$7E,$D9,$BB,$F3,$1E,$48,$F3,$1E,$4A,$EB,$1F,$E6,$AF,$82
	.dc.b	$7C,$1F,$F9,$8E,$B0,$D4,$23,$6E,$5A,$F1,$6D,$AB,$9C,$F2,$8C,$3D
	.dc.b	$27,$73,$DB,$B5,$ED,$3E,$0F,$FC,$99,$5A,$BE,$18,$E0,$FF,$2A,$DF
	.dc.b	$0A,$FF,$29,$7C,$1B,$E0,$FD,$56,$8C,$3E,$D6,$BF,$05,$AF,$FA,$15
	.dc.b	$CE,$7E,$88,$E5,$3D,$8F,$0E,$4F,$83,$C4,$FB,$F4,$FF,$4B,$E1,$C9
	.dc.b	$B7,$C2,$BF,$CD,$1F,$06,$F8,$3F,$3B,$70,$FA,$4D,$7E,$B3,$5F,$F4
	.dc.b	$D3,$9C,$BC,$CD,$C3,$FD,$4E,$D9,$69,$F3,$BC,$5F,$0E,$12,$7D,$07
	.dc.b	$E0,$2B,$85,$6E,$32,$3E,$09,$F0,$5F,$39,$B1,$F8,$96,$BF,$8D,$6B
	.dc.b	$FE,$B1,$73,$97,$CD,$6E,$5F,$EB,$E6,$CF,$A3,$E0,$FF,$CF,$91,$88
	.dc.b	$F1,$5F,$FA,$46,$37,$82,$BF,$D0,$5E,$83,$E0,$BF,$06,$DC,$3F,$3C
	.dc.b	$D7,$F4,$4D,$7F,$D8,$CE,$72,$FC,$13,$72,$FF,$69,$B6,$7E,$6F,$0B
	.dc.b	$E0,$BF,$D4,$93,$99,$34,$E1,$97,$E8,$27,$7F,$03,$E6,$25,$79,$9A
	.dc.b	$B1,$F1,$C7,$08,$FB,$35,$FA,$AF,$81,$C4,$6F,$2F,$3D,$62,$FF,$67
	.dc.b	$7B,$AF,$BB,$F0,$3F,$EE,$19,$FD,$03,$6F,$FD,$E2,$D9,$FE,$0B,$F0
	.dc.b	$1F,$8D,$51,$96,$F8,$FF,$EC,$58,$C4,$ED,$D1,$23,$FD,$95,$19,$F8
	.dc.b	$1E,$04,$AF,$C8,$C3,$D8,$C4,$7B,$4F,$E2,$F6,$AC,$E3,$FB,$9B,$E1
	.dc.b	$5F,$9A,$F8,$0F,$F5,$6B,$7C,$0F,$F6,$D4,$62,$3A,$0F,$80,$7C,$0F
	.dc.b	$D1,$87,$BA,$F7,$FF,$93,$C2,$F9,$65,$FC,$25,$BB,$E9,$B3,$FC,$BB
	.dc.b	$1E,$B5,$FC,$C3,$0E,$93,$9C,$7C,$11,$88,$ED,$97,$82,$31,$87,$A5
	.dc.b	$D6,$1D,$2B,$CF,$97,$9A,$31,$3C,$2F,$94,$F9,$5B,$5E,$1C,$7C,$93
	.dc.b	$C7,$1F,$1B,$B9,$C9,$FD,$23,$1F,$14,$6E,$B3,$3F,$2E,$B8,$BB,$BE
	.dc.b	$EF,$B0,$DE,$3E,$6B,$E5,$1E,$A3,$81,$D3,$0E,$EF,$9C,$D1,$DF,$0D
	.dc.b	$5F,$69,$DC,$E4,$DD,$F7,$7A,$9C,$A3,$71,$98,$E0,$7E,$00,$A7,$98
	.dc.b	$C3,$F4,$5B,$F0,$0F,$81,$FE,$DD,$28,$C4,$FF,$73,$8C,$4E,$A7,$33
	.dc.b	$FD,$D7,$E0,$7F,$AE,$A6,$91,$FE,$C2,$FD,$0F,$75,$F0,$3F,$DC,$B6
	.dc.b	$FC,$55,$7F,$7E,$AD,$4E,$75,$FF,$02,$46,$37,$FB,$35,$B3,$E6,$E1
	.dc.b	$D2,$33,$1E,$33,$37,$CD,$F8,$D5,$CD,$FA,$A7,$3C,$3F,$39,$7C,$E3
	.dc.b	$27,$38,$CD,$73,$8F,$D7,$D7,$39,$F5,$C6,$F0,$FD,$9D,$F3,$9C,$D6
	.dc.b	$E7,$DB,$5E,$6C,$3D,$33,$98,$F3,$99,$26,$DF,$E0,$3C,$3C,$CF,$F8
	.dc.b	$16,$3D,$57,$E0,$C7,$FC,$3F,$3B,$6E,$1E,$4E,$3B,$7F,$E2,$EC,$39
	.dc.b	$CE,$4F,$ED,$FF,$8D,$7F,$8D,$9B,$E3,$F0,$E7,$FE,$BD,$E8,$F2,$C1
	.dc.b	$7B,$7F,$8C,$B5,$FC,$B3,$5F,$E2,$ED,$7F,$C7,$ED,$5F,$2C,$F8,$2E
	.dc.b	$CD,$C7,$FA,$7F,$82,$F4,$33,$52,$7F,$C0,$F5,$BB,$FE,$6D,$97,$F7
	.dc.b	$5B,$D4,$7F,$2E,$9D,$3F,$AB,$6B,$FC,$B1,$AF,$F9,$51,$AB,$E4,$1F
	.dc.b	$45,$F0,$DB,$87,$F8,$70,$7C,$A7,$F4,$97,$FC,$37,$1A,$AF,$E7,$ED
	.dc.b	$E2,$FF,$20,$36,$7F,$CC,$67,$2E,$1E,$D8,$48,$FF,$34,$4F,$3B,$FE
	.dc.b	$08,$D7,$FA,$2B,$5F,$F3,$6F,$C6,$3E,$23,$F1,$51,$86,$A8,$D5,$7F
	.dc.b	$9D,$5B,$1F,$E7,$9F,$8B,$7C,$67,$06,$D7,$F9,$ED,$B3,$B9,$EB,$F0
	.dc.b	$2F,$83,$7C,$1F,$FA,$06,$30,$DF,$81,$7C,$2E,$6B,$78,$7F,$A2,$7E
	.dc.b	$25,$F5,$BF,$E7,$4E,$A7,$CB,$53,$FB,$2F,$27,$F9,$AE,$B7,$7F,$E4
	.dc.b	$C6,$BF,$86,$AE,$2F,$F8,$71,$AF,$F7,$48,$DC,$F3,$BF,$F3,$FC,$72
	.dc.b	$9F,$E6,$8D,$7F,$BA,$B5,$FF,$4E,$FD,$1F,$F6,$BA,$FF,$11,$DB,$E5
	.dc.b	$7D,$47,$A3,$87,$FB,$ED,$6D,$BE,$0E,$B0,$F3,$56,$DE,$B1,$55,$7F
	.dc.b	$16,$72,$7F,$AB,$DC,$3A,$8F,$F4,$CF,$CE,$BE,$6B,$FC,$4B,$F3,$D2
	.dc.b	$BF,$CE,$73,$96,$CE,$5F,$F0,$CF,$9A,$AF,$15,$6E,$7F,$C3,$AD,$D3
	.dc.b	$FC,$4D,$3C,$5F,$F5,$04,$67,$3F,$F0,$EB,$6B,$FC,$83,$5B,$9C,$E1
	.dc.b	$D8,$C4,$E8,$F2,$90,$F2,$CE,$95,$75,$6C,$E7,$0F,$54,$71,$C3,$CB
	.dc.b	$39,$BF,$04,$E7,$0F,$F1,$9D,$F0,$8E,$E6,$4F,$F6,$EE,$1D,$E3,$58
	.dc.b	$7C,$29,$D1,$E2,$3F,$CA,$37,$E7,$AF,$31,$C5,$C6,$1E,$29,$EE,$4A
	.dc.b	$F3,$BD,$2F,$BE,$1E,$6B,$E1,$5A,$AE,$13,$97,$C7,$87,$77,$C6,$63
	.dc.b	$0F,$84,$D3,$27,$23,$DE,$F6,$9E,$91,$93,$DC,$7E,$12,$39,$B4,$F4
	.dc.b	$5F,$B9,$A7,$6B,$F4,$12,$BA,$BF,$2E,$39,$11,$C1,$23,$F1,$31,$D0
	.dc.b	$F2,$5F,$CE,$BF,$1D,$75,$9F,$59,$E1,$AE,$E7,$27,$B3,$87,$F2,$46
	.dc.b	$0E,$27,$C1,$C3,$E9,$B8,$9E,$CA,$32,$DC,$3E,$53,$BA,$EC,$E4,$FA
	.dc.b	$47,$58,$DB,$F3,$AB,$E0,$3E,$47,$BE,$1F,$76,$3E,$0E,$1C,$30,$F0
	.dc.b	$9E,$29,$F8,$93,$F0,$4E,$58,$7C,$1B,$F4,$D7,$0C,$3C,$37,$D6,$7C
	.dc.b	$37,$E3,$BE,$07,$AA,$FB,$61,$E1,$C3,$F0,$15,$F2,$4F,$71,$E6,$3E
	.dc.b	$C5,$F4,$C3,$CB,$1F,$04,$F8,$71,$E0,$BE,$B1,$C3,$0F,$93,$5C,$2B
	.dc.b	$57,$E8,$C3,$F0,$35,$E6,$BF,$A0,$7C,$02,$1E,$B3,$27,$9A,$BC,$93
	.dc.b	$EB,$8F,$90,$6C,$98,$7C,$77,$CF,$5F,$0E,$FC,$51,$C2,$7D,$2F,$C3
	.dc.b	$78,$3F,$17,$0F,$9D,$5E,$CB,$F0,$4F,$C7,$38,$5F,$CC,$C3,$EB,$B6
	.dc.b	$BD,$D3,$E9,$79,$3F,$1C,$F7,$DF,$C7,$8F,$34,$7A,$0F,$87,$87,$9A
	.dc.b	$FA,$C7,$C0,$AE,$0E,$AF,$DA,$FD,$73,$D5,$87,$D0,$7E,$5C,$FB,$1F
	.dc.b	$0C,$7B,$2B,$E6,$C7,$94,$F4,$1C,$23,$93,$F3,$E7,$93,$A3,$11,$A8
	.dc.b	$F2,$C6,$6F,$D5,$87,$C4,$7E,$7D,$FE,$A2,$F8,$1E,$39,$F8,$F5,$F3
	.dc.b	$8F,$81,$19,$7D,$27,$9E,$BA,$BF,$88,$7A,$CF,$D2,$BF,$3C,$6A,$7F
	.dc.b	$0F,$87,$D7,$78,$35,$C9,$E0,$3B,$3F,$3B,$0F,$CE,$C7,$D7,$3D,$D1
	.dc.b	$9B,$FB,$71,$F4,$8F,$C8,$46,$5F,$B0,$71,$AF,$3B,$F3,$CD,$99,$8F
	.dc.b	$25,$FA,$9E,$51,$E0,$34,$79,$CF,$E6,$E7,$FB,$76,$3F,$C1,$0D,$C3
	.dc.b	$FC,$2A,$47,$C5,$AF,$49,$FE,$D8,$D9,$FF,$11,$5A,$31,$43,$FC,$2E
	.dc.b	$E1,$A7,$C5,$AE,$6D,$AE,$73,$BC,$58,$FF,$80,$23,$0D,$AE,$07,$79
	.dc.b	$DB,$E1,$C3,$C8,$E8,$F0,$9F,$44,$FD,$F5,$74,$78,$35,$FE,$8C,$D8
	.dc.b	$E7,$1C,$67,$F9,$9C,$E4,$91,$DE,$3C,$6F,$23,$15,$E9,$79,$3F,$A1
	.dc.b	$9F,$F2,$67,$B2,$F9,$1F,$F2,$87,$82,$F5,$DE,$5E,$1B,$8F,$C1,$C7
	.dc.b	$E8,$2C,$39,$39,$3E,$A3,$87,$FD,$73,$3C,$9B,$5E,$2C,$24,$74,$3A
	.dc.b	$D7,$17,$0D,$73,$1B,$7C,$7E,$BB,$D8,$F4,$7C,$94,$9F,$20,$9D,$38
	.dc.b	$6B,$E6,$8E,$0B,$5F,$9B,$3A,$6D,$78,$67,$F2,$53,$C9,$F5,$35,$F2
	.dc.b	$C6,$D7,$CF,$7C,$5A,$FB,$3B,$BD,$CF,$CF,$9F,$E8,$19,$D6,$DF,$41
	.dc.b	$B7,$F9,$19,$DB,$EE,$8C,$BD,$B3,$F0,$47,$C9,$8F,$D4,$35,$D1,$EB
	.dc.b	$F5,$9E,$FF,$E0,$7E,$26,$4F,$CA,$9C,$B6,$72,$FD,$E9,$FF,$4C,$4E
	.dc.b	$A3,$E5,$63,$F8,$88,$C4,$7D,$F6,$B9,$AF,$8D,$18,$8F,$DF,$2F,$80
	.dc.b	$FC,$A7,$C7,$49,$F2,$47,$80,$C3,$F9,$08,$C3,$A3,$DD,$1D,$88,$FC
	.dc.b	$1F,$1F,$F4,$1F,$4D,$23,$E0,$47,$B0,$C3,$F9,$C8,$C3,$F8,$EF,$65
	.dc.b	$ED,$7F,$AE,$5B,$19,$8F,$F5,$E7,$B4,$F6,$FF,$1D,$2F,$69,$E0,$7E
	.dc.b	$BB,$0E,$33,$F2,$8F,$E4,$0E,$39,$7F,$80,$B7,$83,$F6,$B8,$F5,$9F
	.dc.b	$E4,$D8,$74,$BF,$E5,$13,$97,$C1,$87,$7C,$BF,$7B,$5B,$96,$3E,$28
	.dc.b	$C5,$F1,$6E,$5E,$A7,$84,$EF,$0E,$97,$E2,$7C,$B5,$B6,$E3,$99,$F6
	.dc.b	$E3,$E2,$9E,$31,$E6,$3D,$35,$C5,$B5,$C5,$F3,$D7,$17,$99,$9C,$7D
	.dc.b	$17,$E0,$8E,$87,$03,$96,$1E,$97,$A9,$A9,$E2,$F4,$38,$C6,$D4,$69
	.dc.b	$B8,$79,$DE,$C7,$0B,$F3,$3A,$C3,$CD,$3F,$84,$8F,$E4,$95,$F5,$DB
	.dc.b	$19,$7D,$46,$BA,$9E,$DF,$F1,$5A,$FE,$76,$D8,$CC,$FF,$29,$9C,$BF
	.dc.b	$C5,$7D,$BF,$E2,$C9,$A4,$74,$7C,$8F,$69,$ED,$FF,$64,$DB,$F1,$13
	.dc.b	$FD,$D6,$BE,$A5,$6B,$7F,$AD,$39,$39,$5F,$1D,$FF,$9E,$5F,$08,$C4
	.dc.b	$FF,$45,$8C,$57,$28,$F3,$1C,$9C,$46,$A3,$EE,$9D,$F0,$FC,$E6,$1E
	.dc.b	$38,$C9,$CA,$33,$7E,$08,$FE,$BF,$5C,$E7,$DB,$1B,$C3,$FB,$1D,$F3
	.dc.b	$9C,$D6,$E7,$D7,$5E,$53,$BC,$65,$FF,$0D,$57,$8F,$5F,$EF,$6D,$C7
	.dc.b	$8E,$1E,$47,$FA,$DC,$66,$B8,$FA,$9F,$E1,$4B,$E3,$95,$BF,$2C,$DA
	.dc.b	$FF,$14,$7B,$4F,$6F,$FC,$13,$1F,$DF,$23,$15,$FD,$75,$B1,$FD,$BD
	.dc.b	$AF,$F7,$77,$38,$7F,$89,$3C,$37,$B6,$F8,$AF,$05,$AF,$78,$7F,$8E
	.dc.b	$6F,$75,$AF,$69,$EE,$53,$D1,$3D,$3D,$CC,$5F,$F7,$CA,$DB,$FD,$CD
	.dc.b	$B8,$FF,$81,$5A,$FF,$83,$5A,$FF,$86,$9D,$E7,$E9,$CD,$F2,$FD,$AB
	.dc.b	$F9,$86,$3E,$C8,$F5,$FB,$6F,$72,$9F,$37,$27,$BA,$E1,$CC,$7F,$84
	.dc.b	$AB,$83,$FE,$09,$6E,$3F,$E1,$B6,$BF,$E2,$36,$BF,$E3,$07,$79,$FF
	.dc.b	$94,$A7,$BE,$1D,$AB,$FA,$34,$60,$FF,$10,$FB,$8F,$73,$EF,$CF,$F9
	.dc.b	$A6,$31,$3F,$E1,$B6,$C7,$F8,$B5,$AF,$F8,$D7,$DC,$7B,$2F,$73,$FC
	.dc.b	$45,$18,$6F,$B8,$F7,$33,$5B,$BF,$F1,$C7,$C8,$BD,$BF,$F1,$23,$63
	.dc.b	$FC,$58,$D7,$71,$FE,$29,$F5,$9E,$E7,$F9,$62,$1E,$ED,$3F,$CA,$BC
	.dc.b	$3F,$E6,$4A,$DD,$FF,$90,$9A,$FF,$92,$5A,$FF,$94,$5A,$FF,$97,$9A
	.dc.b	$FF,$A3,$DE,$31,$D6,$FB,$EF,$B7,$C9,$F3,$78,$BF,$F4,$B7,$BB,$FD
	.dc.b	$AE,$FF,$4D,$93,$DE,$F5,$1B,$A5,$6F,$1F,$F3,$7B,$FC,$42,$39,$F8
	.dc.b	$BF,$E4,$B8,$C4,$70,$8C,$1C,$E3,$94,$70,$3F,$CE,$DD,$4F,$8F,$7B
	.dc.b	$DF,$ED,$D9,$FF,$76,$3F,$EE,$AF,$63,$FD,$8B,$1F,$EE,$E3,$FD,$8A
	.dc.b	$7F,$A3,$0F,$F6,$59,$4F,$F6,$21,$5E,$31,$CF,$DF,$FF,$65,$FB,$FF
	.dc.b	$EF,$0A,$FF,$6E,$9F,$EC,$A3,$FC,$3C,$78,$9F,$25,$7F,$BA,$6B,$FD
	.dc.b	$A5,$3B,$DB,$FC,$D3,$7F,$ED,$4A,$EF,$7C,$23,$07,$82,$BD,$35,$FF
	.dc.b	$00,$5B,$A7,$FB,$DE,$BC,$91,$FF,$00,$4F,$1D,$F8,$63,$35,$FE,$B0
	.dc.b	$C3,$C7,$5E,$FC,$3B,$3C,$1B,$3F,$EF,$88,$D1,$FE,$E7,$79,$57,$4C
	.dc.b	$3F,$D5,$B7,$D2,$7C,$D5,$CC,$E1,$3A,$AE,$11,$FE,$E3,$AE,$13,$EB
	.dc.b	$8F,$69,$FE,$FD,$39,$12,$BA,$9E,$52,$C7,$FA,$FB,$0F,$97,$3F,$58
	.dc.b	$F4,$9C,$63,$B3,$4F,$F7,$26,$1F,$41,$F3,$9E,$5C,$3C,$EF,$E4,$1B
	.dc.b	$87,$A2,$3C,$D8,$78,$EF,$AB,$63,$31,$E5,$BF,$9A,$D8,$EE,$75,$3E
	.dc.b	$09,$F9,$93,$CC,$7E,$34,$F6,$4F,$C0,$7F,$36,$D7,$F3,$6D,$C3,$E1
	.dc.b	$C7,$E1,$63,$F2,$2F,$E4,$E7,$C4,$FC,$68,$CB,$63,$31,$CE,$BD,$33
	.dc.b	$E0,$9F,$96,$FF,$B8,$ED,$87,$FB,$25,$B1,$FE,$D1,$6B,$FE,$E3,$6B
	.dc.b	$C6,$38,$11,$E0,$FE,$9F,$87,$F8,$84,$F2,$C3,$DD,$18,$9D,$57,$FB
	.dc.b	$AA,$B5,$7F,$50,$FC,$35,$6E,$75,$55,$FD,$4C,$5B,$EB,$1A,$21,$FC
	.dc.b	$02,$FA,$A1,$8A,$B9,$61,$E1,$9D,$4E,$5E,$77,$D9,$FE,$29,$7D,$1E
	.dc.b	$77,$C9,$FC,$59,$C9,$C3,$F8,$79,$FF,$7A,$C6,$27,$F9,$23,$70,$DD
	.dc.b	$F8,$6F,$BC,$7F,$3B,$8C,$5F,$E2,$EF,$C5,$18,$C3,$D5,$7F,$97,$3E
	.dc.b	$39,$EA,$73,$1B,$8F,$39,$BA,$F2,$BB,$33,$1B,$7D,$53,$FC,$1D,$F9
	.dc.b	$93,$C1,$F3,$98,$8F,$B6,$F3,$38,$9F,$14,$C5,$7C,$26,$DF,$C2,$C3
	.dc.b	$E2,$47,$1C,$3C,$55,$F4,$4E,$11,$E0,$3D,$E5,$C3,$C8,$7C,$7B,$F9
	.dc.b	$57,$D2,$7F,$0D,$18,$7C,$27,$78,$D4,$FF,$2C,$8F,$61,$E9,$3B,$61
	.dc.b	$EE,$69,$98,$F0,$3C,$0F,$C3,$9D,$8F,$09,$9C,$3D,$B7,$FA,$68,$E6
	.dc.b	$70,$2C,$69,$E5,$3E,$EA,$FC,$0E,$1F,$84,$C3,$D3,$1F,$5E,$BE,$2D
	.dc.b	$79,$1B,$1C,$30,$F8,$51,$DA,$B8,$BF,$CB,$CF,$33,$87,$F1,$13,$F0
	.dc.b	$6B,$A1,$E3,$8F,$C6,$DF,$8D,$F3,$CF,$8E,$75,$1F,$84,$39,$9F,$26
	.dc.b	$3C,$8F,$80,$E2,$78,$CF,$49,$DA,$FD,$33,$E2,$9F,$21,$CA,$76,$77
	.dc.b	$AF,$23,$FE,$06,$F7,$9E,$D7,$F8,$BF,$57,$BE,$F3,$F0,$DB,$FB,$97
	.dc.b	$C1,$B7,$FC,$3D,$87,$F1,$18,$D3,$80,$E1,$DB,$5E,$4F,$0C,$99,$62
	.dc.b	$BF,$C7,$D3,$98,$FF,$51,$5B,$E6,$3F,$C8,$5E,$FB,$DF,$F4,$FC,$04
	.dc.b	$CF,$F2,$93,$96,$CF,$95,$FE,$51,$19,$7D,$0D,$7D,$AD,$7E,$03,$BC
	.dc.b	$FC,$F1,$CA,$7C,$67,$CA,$F8,$6F,$F2,$B7,$C0,$7A,$67,$F9,$8F,$C0
	.dc.b	$FF,$1F,$33,$B4,$8F,$24,$FC,$57,$F9,$E3,$53,$4E,$8E,$23,$93,$FE
	.dc.b	$32,$F9,$9F,$F3,$37,$C0,$3E,$07,$EE,$FC,$14,$AF,$DC,$4E,$5B,$3F
	.dc.b	$15,$46,$5F,$B8,$D7,$F0,$2D,$7F,$08,$EF,$4F,$8B,$1C,$A7,$C6,$7E
	.dc.b	$3B,$D2,$F2,$FC,$0F,$E4,$3E,$0A,$57,$92,$32,$FD,$7C,$11,$E5,$6C
	.dc.b	$66,$30,$70,$7F,$CC,$B1,$A7,$FD,$AB,$C6,$F8,$FF,$8B,$F0,$7F,$A0
	.dc.b	$F8,$39,$5F,$12,$32,$F8,$5F,$01,$1F,$C5,$7C,$1B,$E0,$FF,$D2,$8D
	.dc.b	$8F,$E4,$8D,$7F,$CE,$FF,$02,$F8,$37,$07,$F7,$9F,$81,$78,$FF,$AD
	.dc.b	$C3,$84,$FC,$A6,$F6,$E7,$2F,$F5,$16,$3C,$EB,$95,$FE,$D6,$31,$93
	.dc.b	$1E,$AF,$F8,$1F,$1F,$24,$62,$11,$8A,$E5,$97,$9A,$79,$46,$30,$EF
	.dc.b	$1C,$F1,$FE,$09,$87,$3C,$39,$4F,$F8,$57,$1F,$1C,$F7,$6E,$1E,$B9
	.dc.b	$E6,$78,$AF,$9C,$77,$8F,$09,$CF,$1F,$25,$F7,$AD,$CF,$8C,$E7,$7D
	.dc.b	$9E,$18,$F6,$C7,$CB,$1D,$CD,$E1,$D3,$1F,$06,$3D,$67,$86,$3D,$96
	.dc.b	$3D,$5B,$7C,$5F,$11,$E8,$3B,$16,$38,$1F,$59,$FE,$56,$D7,$E0,$18
	.dc.b	$8F,$A4,$EB,$BD,$F0,$7F,$C6,$EB,$F9,$D3,$63,$33,$FC,$CE,$72,$FE
	.dc.b	$F7,$E0,$FF,$7C,$9A,$47,$47,$A7,$E0,$DF,$07,$FC,$83,$3F,$C4,$4F
	.dc.b	$F9,$46,$BE,$CD,$6B,$7F,$B7,$39,$39,$5F,$1D,$7F,$A3,$5F,$08,$C4
	.dc.b	$FF,$4C,$8C,$57,$28,$F3,$1C,$9C,$46,$A3,$E7,$1D,$F0,$FC,$E6,$1E
	.dc.b	$38,$C9,$CA,$33,$7E,$08,$FE,$B9,$5C,$E7,$DB,$1B,$C3,$FB,$05,$F3
	.dc.b	$9C,$D6,$E7,$D7,$5E,$53,$BC,$65,$FF,$3B,$57,$8F,$7F,$E4,$AD,$C7
	.dc.b	$8E,$1E,$47,$FC,$59,$19,$AE,$3E,$DF,$F5,$0B,$E3,$95,$BF,$2C,$DA
	.dc.b	$FF,$42,$7C,$13,$E0,$F1,$E3,$BE,$01,$F0,$7B,$B5,$EF,$1F,$F2,$1D
	.dc.b	$EE,$B5,$39,$AF,$E3,$35,$FE,$91,$AD,$3F,$E4,$39,$D6,$7F,$D9,$A7
	.dc.b	$35,$C0,$FF,$4C,$7C,$13,$BB,$FA,$3C,$7F,$83,$63,$18,$7F,$77,$6C
	.dc.b	$7F,$83,$5A,$FF,$84,$9C,$DF,$7C,$9F,$05,$ED,$9C,$DF,$9E,$72,$FF
	.dc.b	$96,$63,$36,$C7,$E3,$5F,$1C,$7E,$4E,$1C,$2B,$D4,$3E,$05,$F0,$3E
	.dc.b	$A8,$FF,$5A,$46,$32,$FF,$08,$36,$3F,$C4,$CD,$7F,$C5,$6E,$73,$FF
	.dc.b	$06,$CB,$E0,$B0,$FF,$8C,$66,$D7,$F8,$A2,$6B,$E3,$BF,$C1,$57,$17
	.dc.b	$EF,$57,$08,$C1,$FE,$79,$F8,$07,$BA,$F8,$0F,$F1,$34,$61,$BF,$00
	.dc.b	$F8,$0C,$D6,$F2,$FF,$20,$7C,$03,$DD,$FF,$05,$36,$3F,$C6,$0D,$77
	.dc.b	$1F,$E0,$FB,$79,$34,$93,$FE,$37,$8C,$BF,$E3,$9F,$61,$EF,$FF,$85
	.dc.b	$DB,$19,$8C,$11,$FF,$4B,$46,$9F,$B2,$DE,$1F,$CB,$60,$F7,$FF,$CB
	.dc.b	$D1,$FE,$1A,$9D,$6F,$FE,$61,$6D,$FF,$BC,$5B,$CD,$F0,$6F,$8D,$7F
	.dc.b	$99,$FD,$EF,$E1,$77,$FE,$F7,$6A,$69,$3C,$1E,$31,$B6,$BB,$7F,$CF
	.dc.b	$5E,$F3,$DB,$CD,$7F,$7B,$CF,$FC,$62,$DC,$3C,$57,$C2,$BF,$DE,$7E
	.dc.b	$F7,$FB,$C4,$FF,$50,$7B,$C7,$19,$7F,$A5,$04,$65,$FF,$5E,$47,$03
	.dc.b	$6F,$FB,$00,$98,$FF,$90,$A7,$84,$7E,$A1,$B1,$FE,$D1,$8D,$E1,$FE
	.dc.b	$8D,$9C,$B6,$BB,$4F,$80,$F2,$4E,$91,$E6,$8D,$BD,$DB,$5E,$36,$D7
	.dc.b	$59,$E7,$1E,$A9,$F2,$12,$75,$3C,$CD,$9E,$33,$D2,$7F,$AE,$63,$B3
	.dc.b	$97,$93,$98,$E7,$1A,$8E,$27,$FB,$18,$F5,$1C,$63,$FD,$98,$47,$DF
	.dc.b	$5F,$05,$D9,$CA,$DE,$F7,$AE,$BB,$9C,$67,$B5,$71,$39,$57,$08,$FF
	.dc.b	$7E,$35,$F8,$04,$BF,$5B,$F9,$1B,$FF,$44,$4E,$E3,$FD,$8A,$D7,$FD
	.dc.b	$EC,$62,$3F,$1F,$B7,$FB,$BE,$7F,$DA,$F5,$FE,$DD,$AF,$F7,$31,$FE
	.dc.b	$E1,$3F,$DC,$C7,$FB,$E4,$FC,$A4,$70,$C3,$FD,$C5,$3F,$EF,$08,$F4
	.dc.b	$3E,$67,$53,$CC,$FF,$73,$BA,$9F,$88,$EA,$3F,$DD,$57,$FE,$F0,$9E
	.dc.b	$A7,$FB,$C2,$39,$57,$79,$F5,$C7,$91,$E0,$EA,$78,$3D,$0F,$5C,$FB
	.dc.b	$DF,$D6,$DF,$C3,$AE,$31,$B3,$CE,$F6,$7E,$11,$27,$96,$0A,$E5,$3D
	.dc.b	$AF,$D2,$FC,$C3,$A5,$7A,$5F,$D5,$CF,$07,$B9,$83,$D8,$73,$3D,$87
	.dc.b	$3C,$3E,$53,$F9,$D2,$1E,$C3,$89,$D0,$FC,$B6,$1E,$E8,$E3,$1F,$A1
	.dc.b	$7E,$91,$82,$46,$CF,$85,$5F,$02,$BD,$0F,$AE,$BA,$9E,$73,$DF,$7F
	.dc.b	$4A,$F9,$C7,$E7,$AF,$A3,$EF,$3E,$64,$FC,$87,$D4,$60,$F8,$26,$2B
	.dc.b	$C4,$FB,$9F,$11,$F9,$19,$EB,$1E,$43,$E1,$46,$5F,$35,$79,$1F,$E0
	.dc.b	$78,$7D,$B7,$A5,$FA,$6F,$E0,$BE,$59,$D4,$F9,$4F,$79,$F5,$67,$6F
	.dc.b	$A2,$37,$7E,$88,$E1,$7E,$18,$FD,$EC,$EA,$FB,$61,$F3,$9F,$15,$FC
	.dc.b	$7A,$E4,$48,$DC,$7A,$0F,$11,$B7,$2F,$83,$0F,$2B,$E8,$3E,$64,$6A
	.dc.b	$7A,$47,$A6,$BC,$4F,$43,$E6,$47,$D0,$9F,$DD,$D7,$D0,$8F,$9D,$3F
	.dc.b	$42,$72,$FD,$09,$F0,$E1,$F0,$67,$F9,$36,$1F,$06,$BC,$07,$AC,$F2
	.dc.b	$98,$3D,$47,$B4,$C5,$FA,$1F,$64,$73,$6E,$1E,$A7,$98,$F7,$9E,$BF
	.dc.b	$F5,$C8,$87,$F5,$E8,$CB,$83,$11,$5A,$58,$DB,$FE,$58,$6D,$B0,$ED
	.dc.b	$18,$7F,$CF,$9E,$93,$DE,$FF,$08,$CE,$6F,$A9,$CF,$0F,$F0,$BF,$12
	.dc.b	$F8,$4F,$F7,$4C,$7B,$3F,$DD,$BD,$D7,$BB,$E7,$84,$61,$F5,$35,$F7
	.dc.b	$35,$F8,$6E,$67,$AC,$72,$9F,$F1,$3C,$E7,$3F,$5F,$BB,$89,$95,$E5
	.dc.b	$9F,$9D,$5B,$8F,$F4,$16,$0B,$FF,$18,$FB,$9F,$2A,$7E,$9B,$5F,$AC
	.dc.b	$D7,$FC,$26,$D4,$FC,$96,$E7,$D0,$FF,$1E,$EE,$F7,$31,$3E,$53,$9C
	.dc.b	$E1,$FF,$15,$47,$F9,$13,$D9,$7B,$7F,$2D,$B3,$F8,$36,$BF,$87,$6B
	.dc.b	$F8,$C7,$3B,$7C,$A6,$EB,$F9,$16,$CF,$C4,$F1,$FC,$17,$23,$CC,$27
	.dc.b	$7C,$1F,$83,$AF,$37,$B7,$E5,$25,$79,$5A,$FF,$AB,$29,$1C,$23,$FD
	.dc.b	$A7,$BB,$DB,$C4,$6E,$6E,$7F,$EC,$FA,$D4,$F9,$2D,$ED,$7C,$52,$47
	.dc.b	$C5,$6B,$E1,$7E,$61,$F9,$CF,$D2,$7F,$63,$7F,$8E,$F6,$DE,$DF,$F9
	.dc.b	$D1,$B7,$F0,$9A,$FF,$B1,$BD,$87,$B6,$BF,$F2,$BF,$B0,$F0,$3F,$7D
	.dc.b	$87,$09,$ED,$8F,$53,$FD,$03,$87,$28,$E3,$5F,$C1,$63,$18,$B0,$E8
	.dc.b	$FF,$99,$30,$F1,$C6,$21,$18,$AE,$58,$F9,$67,$94,$63,$0E,$F1,$CF
	.dc.b	$1F,$E2,$98,$73,$C3,$94,$FF,$9C,$B1,$F1,$CF,$76,$E1,$EE,$BE,$2D
	.dc.b	$AE,$A2,$76,$F3,$C7,$C9,$7D,$EF,$8C,$F8,$CE,$35,$C1,$E5,$5C,$2F
	.dc.b	$B4,$7B,$0E,$05,$37,$8F,$29,$E1,$1A,$C3,$C0,$E6,$63,$C7,$1F,$82
	.dc.b	$FC,$CC,$7C,$AF,$43,$B6,$3F,$59,$FE,$76,$D7,$E7,$1F,$CE,$DC,$3F
	.dc.b	$2F,$DA,$7B,$7F,$CF,$65,$18,$9E,$0D,$7A,$B5,$FE,$8D,$ED,$FF,$2F
	.dc.b	$4D,$23,$FC,$CD,$FA,$7E,$DB,$DB,$FE,$73,$A7,$E2,$6B,$FD,$5F,$5A
	.dc.b	$9C,$ED,$DA,$7E,$65,$F1,$E8,$FB,$B1,$C0,$CD,$70,$7F,$AE,$DF,$39
	.dc.b	$DC,$79,$8D,$B8,$8D,$47,$D9,$3A,$61,$F9,$DB,$F0,$46,$4E,$51,$9B
	.dc.b	$F0,$47,$F6,$DA,$E7,$3E,$D8,$DE,$1F,$DC,$6F,$9C,$E6,$B7,$3E,$F8
	.dc.b	$E9,$87,$96,$FC,$E4,$DB,$FA,$AE,$1E,$47,$FA,$BC,$7A,$2F,$B6,$3F
	.dc.b	$D9,$A7,$6D,$BF,$17,$B7,$6F,$FD,$BF,$87,$39,$C9,$FC,$CB,$D9,$FF
	.dc.b	$7D,$37,$C7,$E1,$D4,$F5,$DF,$11,$CF,$16,$1C,$1F,$E2,$D8,$7F,$84
	.dc.b	$63,$11,$FE,$12,$6B,$9C,$3D,$6F,$FB,$CA,$32,$FF,$BD,$3F,$1B,$7A
	.dc.b	$9D,$DB,$7F,$D7,$72,$F2,$7D,$F6,$6A,$4F,$F4,$CA,$DD,$F9,$05,$F0
	.dc.b	$9D,$57,$09,$FF,$12,$57,$07,$FC,$52,$D7,$FC,$58,$D4,$3E,$23,$FA
	.dc.b	$3D,$7F,$8A,$DB,$1F,$E2,$F6,$BF,$E3,$46,$BF,$E3,$7F,$86,$7C,$3F
	.dc.b	$C4,$84,$62,$BF,$C6,$CD,$8F,$F1,$D3,$5F,$F1,$FF,$B4,$F6,$5E,$D7
	.dc.b	$F8,$F6,$30,$DF,$63,$E6,$D7,$F9,$0E,$72,$FF,$91,$7D,$2F,$F5,$6F
	.dc.b	$97,$FE,$FF,$E2,$FF,$26,$74,$7F,$99,$0F,$F1,$4C,$66,$BF,$D4,$84
	.dc.b	$39,$7A,$7F,$EF,$E2,$3F,$E4,$FF,$C5,$DF,$29,$FE,$42,$F6,$7F,$C8
	.dc.b	$DC,$A7,$FC,$41,$3A,$AF,$F0,$3B,$5F,$B9,$3B,$F3,$F9,$1E,$6E,$2F
	.dc.b	$F3,$15,$6F,$3F,$F1,$FB,$C6,$FF,$CC,$AD,$77,$8F,$F9,$36,$31,$1F
	.dc.b	$7B,$D7,$7C,$4F,$F9,$63,$D9,$FF,$24,$E1,$FE,$56,$CD,$18,$CB,$FC
	.dc.b	$E5,$7C,$30,$FF,$0F,$B5,$F4,$CE,$DF,$CA,$B5,$46,$EF,$FC,$AE,$F1
	.dc.b	$8F,$F3,$EB,$5E,$F8,$7F,$99,$63,$11,$FE,$42,$3F,$C5,$FB,$3E,$1F
	.dc.b	$FC,$92,$7F,$91,$6F,$FC,$C7,$EA,$7F,$93,$E7,$55,$C2,$B5,$97,$F9
	.dc.b	$63,$6E,$33,$93,$FC,$77,$7F,$E5,$73,$95,$6A,$3F,$64,$7F,$96,$8D
	.dc.b	$CE,$70,$FC,$5C,$F6,$9F,$41,$C3,$0F,$1B,$63,$4F,$33,$07,$8C,$EF
	.dc.b	$7E,$3B,$ED,$87,$FA,$96,$F9,$E1,$F0,$27,$9C,$62,$75,$87,$6B,$F5
	.dc.b	$E1,$D7,$0E,$0F,$A8,$C5,$70,$70,$78,$B0,$F0,$BF,$C2,$30,$EC,$FB
	.dc.b	$AB,$9B,$EA,$23,$82,$57,$58,$CB,$F4,$DE,$F1,$F2,$9F,$19,$D6,$34
	.dc.b	$F0,$3C,$58,$7B,$6F,$CB,$0B,$F2,$BF,$36,$FC,$2F,$03,$A9,$9B,$EB
	.dc.b	$1E,$18,$EC,$64,$F1,$1C,$4C,$E1,$F5,$5E,$75,$E0,$8C,$47,$28,$C5
	.dc.b	$FD,$6C,$3C,$F5,$AB,$F0,$13,$0F,$19,$30,$F0,$46,$5F,$19,$E5,$3E
	.dc.b	$05,$74,$BE,$D1,$EE,$BE,$D7,$F4,$6F,$94,$F1,$9F,$C2,$BE,$CA,$E2
	.dc.b	$D3,$EA,$D7,$12,$3E,$29,$E3,$5D,$1E,$F1,$E8,$9D,$BD,$EB,$EB,$E1
	.dc.b	$F1,$23,$A9,$B2,$1B,$3E,$6E,$1F,$7D,$A7,$03,$E1,$47,$E0,$0E,$8F
	.dc.b	$E6,$4D,$94,$98,$7C,$28,$DB,$F8,$68,$F9,$67,$C9,$AF,$87,$1E,$D6
	.dc.b	$DF,$B3,$0F,$7D,$70,$3E,$41,$DA,$3E,$0D,$F6,$7E,$E4,$6E,$FA,$3E
	.dc.b	$B9,$F5,$61,$EF,$8F,$C5,$36,$7D,$0D,$AE,$AD,$9F,$93,$5D,$DF,$C3
	.dc.b	$BD,$1F,$91,$7F,$62,$BD,$58,$7B,$27,$E8,$CF,$B2,$B8,$47,$AC,$E7
	.dc.b	$7F,$8F,$7D,$0D,$74,$7B,$6F,$AB,$A3,$E6,$14,$F8,$F8,$7D,$1A,$EE
	.dc.b	$FD,$3B,$F3,$4F,$93,$0F,$33,$D4,$FC,$93,$F8,$CC,$3E,$14,$64,$93
	.dc.b	$EB,$8F,$35,$77,$6E,$1F,$53,$0F,$0B,$F5,$AB,$C3,$7C,$B0,$F0,$D7
	.dc.b	$CD,$8F,$C8,$DF,$27,$E2,$9E,$E3,$E1,$C7,$47,$CE,$7E,$53,$0F,$75
	.dc.b	$7A,$63,$EA,$B5,$F9,$77,$E2,$AF,$03,$6F,$DF,$1D,$67,$F5,$07,$E2
	.dc.b	$0E,$67,$7C,$3D,$97,$D7,$0F,$0C,$E4,$F2,$9C,$4F,$0C,$6E,$78,$C6
	.dc.b	$CE,$F3,$F9,$47,$B9,$C5,$EA,$D8,$E2,$D3,$BC,$71,$7B,$9D,$70,$FB
	.dc.b	$0F,$CB,$8F,$B0,$D9,$FB,$18,$7D,$18,$FA,$66,$4F,$EE,$67,$F9,$80
	.dc.b	$A7,$F4,$C3,$07,$F7,$32,$D7,$EF,$0F,$F0,$67,$A9,$F8,$C6,$C7,$BA
	.dc.b	$FF,$31,$39,$C3,$FC,$75,$3F,$D9,$E3,$0F,$F8,$69,$AF,$F8,$15,$B3
	.dc.b	$FE,$54,$6E,$1F,$DF,$DB,$3C,$CF,$04,$7E,$5E,$33,$1F,$D8,$5B,$89
	.dc.b	$A4,$8E,$27,$F0,$47,$0F,$D0,$3F,$2E,$E2,$BE,$C9,$EE,$8E,$27,$F3
	.dc.b	$08,$D5,$7C,$DF,$63,$F9,$94,$7C,$29,$F2,$1F,$E9,$0A,$56,$F2,$D5
	.dc.b	$F0,$AF,$8C,$6E,$74,$D3,$FD,$33,$1F,$E2,$C7,$EC,$B5,$D9,$F7,$03
	.dc.b	$97,$FD,$43,$EB,$FF,$B1,$1F,$F5,$A7,$A4,$F4,$7E,$CF,$0A,$31,$3E
	.dc.b	$11,$5C,$5F,$F2,$2B,$72,$FF,$36,$4F,$09,$FF,$3F,$B6,$3E,$0C,$78
	.dc.b	$BC,$3F,$8D,$5C,$E3,$F2,$6D,$7F,$30,$D7,$81,$ED,$1E,$93,$D6,$F4
	.dc.b	$E0,$F8,$7A,$7D,$39,$D5,$7A,$DA,$F9,$A3,$6E,$6B,$81,$DD,$EE,$D9
	.dc.b	$E3,$1F,$AA,$6B,$FA,$FF,$8B,$7C,$AF,$CC,$F2,$5E,$05,$3F,$0D,$3A
	.dc.b	$AF,$9C,$D7,$FD,$4F,$3C,$1F,$F7,$53,$5F,$F3,$AC,$6F,$5F,$9B,$08
	.dc.b	$E4,$FE,$F9,$AF,$96,$FD,$91,$88,$FF,$72,$FE,$EB,$FC,$12,$BF,$D4
	.dc.b	$DE,$BF,$FB,$42,$7F,$D7,$2D,$8E,$F3,$BE,$3F,$E1,$CD,$AF,$F7,$CD
	.dc.b	$6E,$7F,$DE,$3C,$28,$C4,$F8,$0F,$D4,$57,$CF,$4D,$23,$F4,$D7,$9D
	.dc.b	$F9,$89,$B3,$FA,$7A,$ED,$3D,$09,$19,$9D,$BE,$C8,$CB,$FE,$A0,$F5
	.dc.b	$9E,$B7,$EE,$DB,$1F,$A3,$6B,$FE,$A6,$F5,$5E,$BB,$44,$BD,$47,$A9
	.dc.b	$FC,$CF,$0E,$35,$F6,$1B,$B6,$72,$FE,$15,$97,$4A,$E7,$87,$2C,$7A
	.dc.b	$CE,$5E,$D8,$F1,$76,$D7,$27,$0C,$BC,$CE,$04,$F3,$AE,$D0,$CB,$DF
	.dc.b	$18,$6D,$FF,$4D,$AE,$0D,$BE,$B5,$BC,$7F,$A9,$37,$0F,$29,$E3,$C6
	.dc.b	$BE,$2A,$E3,$7E,$9A,$E1,$3E,$7C,$3B,$BC,$4F,$56,$3E,$47,$E8,$B5
	.dc.b	$E9,$87,$69,$E3,$1E,$DC,$7C,$B8,$77,$3D,$55,$BC,$7C,$75,$D2,$B6
	.dc.b	$D3,$53,$1F,$E6,$57,$F5,$5F,$ED,$ED,$7A,$9A,$F5,$9E,$BF,$F1,$D9
	.dc.b	$46,$27,$53,$98,$C4,$7F,$23,$F5,$FF,$50,$47,$F8,$C3,$74,$FB,$7E
	.dc.b	$B7,$EA,$36,$F9,$F5,$F9,$6A,$FA,$33,$98,$FC,$21,$93,$F0,$CE,$6B
	.dc.b	$F0,$B1,$97,$AE,$FF,$CF,$9C,$C2,$BA,$CE,$76,$FE,$AD,$18,$C3,$C7
	.dc.b	$5B,$3B,$D6,$DF,$F0,$BC,$ED,$F2,$9B,$BF,$0D,$72,$9D,$47,$89,$B7
	.dc.b	$E0,$AE,$B5,$B6,$9D,$27,$6F,$EE,$B0,$F8,$9A,$7F,$2E,$6D,$73,$C3
	.dc.b	$C3,$5F,$C0,$63,$FC,$61,$3A,$8E,$5E,$BF,$F6,$2B,$E5,$95,$BD,$CD
	.dc.b	$AF,$E1,$3E,$BB,$CD,$FE,$7B,$1F,$E3,$B8,$C5,$7F,$63,$6C,$7F,$80
	.dc.b	$5A,$FF,$8F,$9C,$E1,$FD,$CB,$91,$D0,$F8,$7F,$05,$AF,$78,$7F,$38
	.dc.b	$AD,$3F,$E0,$3F,$59,$EB,$F0,$6C,$7F,$56,$F6,$31,$32,$3F,$BC,$4E
	.dc.b	$B0,$FF,$0B,$35,$FF,$29,$B5,$FF,$2C,$3B,$CF,$D1,$F3,$8F,$57,$A5
	.dc.b	$7C,$F1,$AA,$75,$87,$F8,$67,$D7,$7B,$1F,$2A,$FF,$9A,$C6,$27,$FC
	.dc.b	$18,$D8,$FF,$15,$35,$FF,$33,$FA,$EF,$63,$C5,$1F,$E2,$F8,$C6,$5F
	.dc.b	$E1,$57,$FB,$FE,$B6,$B7,$5F,$E7,$0F,$84,$7A,$FF,$E0,$D6,$C7,$F8
	.dc.b	$75,$AF,$F8,$ED,$AF,$F9,$E3,$D9,$7B,$3F,$E4,$FF,$4B,$F0,$D3,$FE
	.dc.b	$37,$6B,$F1,$44,$F0,$7F,$C7,$8D,$7F,$C9,$0D,$7B,$4F,$F8,$86,$31
	.dc.b	$1F,$E8,$B7,$87,$0F,$F8,$41,$BB,$66,$BC,$9F,$03,$FE,$63,$F6,$5E
	.dc.b	$D7,$F9,$3B,$47,$AD,$4D,$44,$CF,$FC,$7D,$5C,$1B,$1F,$E3,$38,$C4
	.dc.b	$7F,$97,$DB,$9F,$F8,$89,$B9,$7F,$A8,$5B,$3F,$86,$C3,$E2,$D7,$F8
	.dc.b	$A3,$D8,$7B,$3F,$E6,$2F,$6B,$FC,$57,$3F,$A7,$6E,$3F,$EA,$BA,$DE
	.dc.b	$1F,$E6,$56,$BF,$E7,$16,$BF,$EA,$E6,$BC,$6F,$FC,$99,$1C,$2F,$C5
	.dc.b	$1F,$E8,$98,$C5,$7F,$8A,$2D,$F5,$7F,$EF,$2F,$67,$FD,$C2,$7F,$B0
	.dc.b	$7D,$BF,$F7,$26,$5F,$E9,$6E,$2F,$F7,$24,$7F,$B2,$A3,$F0,$51,$FE
	.dc.b	$EA,$DF,$FD,$B9,$1F,$ED,$2C,$7F,$D8,$D3,$C2,$7B,$4F,$3B,$ED,$1C
	.dc.b	$23,$FD,$C3,$5C,$27,$31,$FE,$D9,$9E,$2F,$FA,$B7,$4F,$F6,$F1,$FE
	.dc.b	$AC,$AE,$11,$FE,$D7,$3F,$DB,$27,$9A,$BD,$AE,$70,$F6,$BC,$E7,$8B
	.dc.b	$B3,$CC,$6E,$FC,$D3,$F0,$6F,$FC,$CB,$87,$C1,$AF,$31,$FE,$9D,$9E
	.dc.b	$B3,$E5,$8F,$F1,$CC,$FC,$37,$55,$D0,$F6,$CE,$E1,$3B,$7F,$61,$5D
	.dc.b	$2F,$FE,$00,$D7,$C2,$69,$EA,$7C,$B8,$7C,$0A,$F6,$1E,$F3,$C6,$FC
	.dc.b	$E7,$31,$C1,$F8,$26,$27,$AB,$C0,$F8,$87,$C0,$3C,$B8,$7A,$DF,$69
	.dc.b	$99,$FF,$45,$DF,$A6,$3F,$0C,$73,$3B,$18,$3E,$0D,$F9,$67,$DD,$19
	.dc.b	$C3,$E2,$47,$4B,$F0,$4F,$88,$F0,$61,$FE,$68,$7F,$07,$3E,$D7,$E1
	.dc.b	$1D,$8F,$59,$E2,$30,$6A,$BB,$CF,$3C,$3D,$53,$F9,$83,$D5,$5D,$4F
	.dc.b	$B5,$5C,$23,$FC,$F3,$3C,$0F,$C5,$57,$D7,$3E,$6C,$78,$2B,$94,$62
	.dc.b	$B9,$4F,$AA,$38,$3D,$2F,$DD,$5F,$0E,$3F,$DB,$97,$FC,$91,$B3,$FE
	.dc.b	$D1,$8C,$47,$FB,$F5,$AF,$C7,$0E,$1F,$81,$1C,$48,$FF,$05,$F6,$FC
	.dc.b	$37,$EA,$AE,$B7,$FA,$69,$CD,$7D,$D9,$CD,$7F,$06,$9C,$CF,$CB,$3C
	.dc.b	$37,$FD,$4E,$33,$5C,$27,$F0,$51,$89,$91,$F9,$03,$F8,$AC,$61,$FB
	.dc.b	$4D,$BF,$AA,$DB,$F9,$CD,$9E,$A7,$39,$DB,$67,$6D,$36,$7E,$32,$FC
	.dc.b	$CF,$EE,$AB,$A3,$ED,$C1,$18,$9D,$9C,$23,$33,$D4,$E1,$3B,$C3,$9C
	.dc.b	$EA,$7C,$11,$F1,$DE,$15,$D2,$31,$1A,$3F,$BE,$57,$37,$B5,$73,$C3
	.dc.b	$F8,$95,$F8,$EB,$55,$FA,$D3,$E6,$9D,$AF,$C7,$1A,$AF,$C9,$4E,$8F
	.dc.b	$01,$5E,$B3,$E4,$6C,$79,$0E,$87,$92,$7A,$61,$F9,$28,$D1,$D2,$7F
	.dc.b	$1C,$F0,$3C,$CE,$1D,$47,$E3,$DD,$47,$98,$D1,$E2,$7C,$D1,$F9,$BA
	.dc.b	$F3,$36,$3C,$33,$B3,$8C,$66,$38,$9F,$9F,$9E,$2D,$38,$C6,$E7,$8C
	.dc.b	$64,$E2,$7E,$8E,$FE,$93,$E8,$9F,$84,$EA,$3E,$F3,$F1,$CD,$1F,$C1
	.dc.b	$4E,$E7,$59,$EE,$F5,$AF,$2B,$4D,$94,$EC,$FD,$27,$FA,$DD,$D7,$F6
	.dc.b	$E7,$0A,$F1,$BF,$4A,$FE,$83,$F4,$8E,$27,$98,$E2,$76,$3F,$86,$03
	.dc.b	$FC,$A1,$ED,$B2,$F3,$E5,$F9,$7A,$FE,$7D,$39,$AE,$19,$27,$37,$FC
	.dc.b	$76,$69,$45,$71,$74,$D7,$F9,$23,$9A,$E9,$1E,$58,$C5,$74,$8F,$F4
	.dc.b	$57,$6B,$D8,$7B,$7F,$C7,$BD,$A7,$B7,$FC,$92,$B5,$36,$B7,$7E,$AC
	.dc.b	$3F,$AD,$61,$C2,$30,$7C,$57,$0F,$B5,$AF,$F3,$B6,$BC,$B0,$F8,$0A
	.dc.b	$38,$E5,$EC,$8C,$4F,$9C,$F2,$7C,$8F,$F3,$6F,$69,$ED,$F9,$FD,$C9
	.dc.b	$7E,$41,$3A,$70,$D7,$E9,$35,$FE,$A0,$D7,$FB,$C3,$5F,$C1,$3C,$70
	.dc.b	$F3,$37,$2D,$DF,$C9,$F8,$0F,$E9,$7E,$DB,$DC,$F9,$DE,$5B,$DD,$A6
	.dc.b	$A2,$67,$F8,$3A,$E0,$D8,$F8,$F1,$88,$FC,$53,$73,$F9,$0D,$CB,$FC
	.dc.b	$20,$D9,$7E,$26,$FC,$79,$E1,$7F,$60,$F7,$FF,$B8,$63,$FE,$EB,$F7
	.dc.b	$FF,$DC,$78,$7D,$EE,$74,$62,$3F,$D6,$55,$C3,$3F,$F0,$F3,$F7,$20
	.dc.b	$D2,$3F,$80,$78,$3F,$AD,$6B,$C0,$FC,$A4,$BD,$4F,$37,$AB,$82,$4F
	.dc.b	$DF,$A4,$6A,$3F,$59,$18,$6C,$78,$4F,$C8,$FB,$EF,$7B,$FC,$02,$D9
	.dc.b	$FC,$53,$5F,$F6,$87,$BC,$F7,$DA,$FF,$05,$F6,$DE,$EF,$F8,$FF,$0E
	.dc.b	$13,$F9,$B6,$ED,$9C,$BF,$C1,$F8,$F3,$AE,$57,$FE,$47,$8C,$64,$C7
	.dc.b	$AB,$FC,$4F,$1F,$24,$62,$11,$8A,$E5,$97,$9A,$79,$46,$30,$EF,$1C
	.dc.b	$F1,$FF,$2B,$61,$CF,$0E,$53,$FC,$87,$1F,$1C,$F7,$6E,$1E,$B9,$E6
	.dc.b	$78,$AF,$9C,$77,$8F,$09,$CF,$1F,$25,$F7,$AD,$CF,$8C,$E7,$7D,$9E
	.dc.b	$18,$F6,$C7,$CB,$1D,$CD,$E1,$D3,$1F,$06,$3D,$67,$86,$3D,$96,$3D
	.dc.b	$5B,$7C,$5F,$11,$E8,$3B,$16,$38,$1F,$59,$FF,$2C,$35,$F8,$06,$23
	.dc.b	$E9,$3A,$F7,$9E,$FF,$F4,$1A,$46,$27,$FA,$34,$62,$75,$39,$9F,$E9
	.dc.b	$1E,$FF,$EB,$C8,$FF,$2C,$6E,$3F,$85,$F7,$FF,$4B,$B7,$D6,$AF,$EB
	.dc.b	$95,$F6,$27,$31,$EF,$32,$7C,$17,$35,$F8,$F8,$CB,$D7,$7F,$EA,$6E
	.dc.b	$61,$5D,$67,$3A,$FF,$61,$8C,$61,$E3,$AD,$9D,$EB,$6F,$FA,$0E,$76
	.dc.b	$F9,$4D,$DF,$86,$B9,$4E,$A3,$C4,$DB,$F0,$57,$5A,$DB,$4E,$93,$B7
	.dc.b	$FC,$01,$87,$C4,$DF,$FA,$A3,$6B,$9E,$1E,$1A,$FE,$07,$1F,$EB,$F9
	.dc.b	$D4,$72,$E1,$FE,$EF,$7C,$B2,$B7,$B9,$B5,$FE,$0F,$F7,$5E,$FE,$39
	.dc.b	$5D,$8F,$7F,$76,$BD,$E3,$FC,$DE,$B4,$FF,$63,$8C,$CF,$F8,$12,$72
	.dc.b	$FA,$0F,$06,$1D,$E3,$39,$FF,$73,$F7,$9E,$CF,$F7,$5B,$FF,$6E,$46
	.dc.b	$31,$FF,$0D,$36,$3F,$C3,$CD,$7F,$D7,$8E,$6B,$FC,$33,$9B,$DE,$7B
	.dc.b	$9E,$CA,$D6,$1F,$E3,$1A,$D4,$E7,$1F,$89,$7C,$2B,$FC,$01,$7C,$31
	.dc.b	$61,$C2,$7F,$C0,$9E,$E3,$DC,$EF,$5F,$E3,$38,$C6,$5F,$E3,$37,$FB
	.dc.b	$B6,$56,$B7,$5F,$ED,$5F,$71,$EC,$7F,$86,$9B,$1F,$E3,$B6,$BF,$E3
	.dc.b	$F6,$BF,$ED,$DF,$15,$F1,$1F,$E2,$26,$C7,$F9,$0D,$AF,$F9,$1D,$AF
	.dc.b	$FB,$97,$DD,$7B,$BF,$E2,$A6,$C7,$F9,$2D,$AF,$F9,$3D,$AF,$18,$FF
	.dc.b	$6F,$C6,$B8,$7F,$C6,$0D,$DB,$FC,$C1,$4F,$53,$FC,$0E,$FF,$8B,$67
	.dc.b	$4F,$F9,$95,$A9,$B1,$C2,$3F,$1C,$3D,$C7,$BB,$FE,$42,$6C,$7F,$98
	.dc.b	$DA,$FF,$99,$DA,$F9,$0E,$51,$FE,$F8,$8D,$E1,$FE,$49,$6E,$1F,$89
	.dc.b	$D7,$F5,$77,$FE,$45,$9C,$BF,$E7,$D6,$BD,$0F,$F1,$04,$70,$7D,$3E
	.dc.b	$E3,$DD,$FF,$2C,$B6,$3F,$CF,$4D,$7F,$CF,$CD,$7D,$0E,$70,$FF,$2F
	.dc.b	$37,$0F,$9E,$D9,$FC,$CF,$BA,$F7,$7E,$9D,$F0,$C2,$E1,$C1,$B7,$FE
	.dc.b	$97,$C3,$97,$33,$1E,$53,$FE,$EE,$F6,$B6,$FD,$D7,$FD,$55,$B7,$FA
	.dc.b	$2A,$3F,$D4,$0D,$7F,$D4,$B9,$7F,$B5,$F2,$FC,$2C,$E4,$D7,$B7,$FE
	.dc.b	$B0,$9F,$F3,$D4,$66,$FA,$9E,$07,$7E,$EF,$8A,$8E,$66,$61,$FE,$B5
	.dc.b	$AE,$4D,$8C,$BF,$EB,$D3,$FD,$9C,$7E,$3A,$36,$E4,$F4,$E1,$E8,$9E
	.dc.b	$38,$7A,$E7,$8B,$E8,$8C,$CF,$FB,$7A,$35,$3F,$1C,$FF,$47,$57,$A9
	.dc.b	$F9,$86,$2B,$DE,$D3,$D0,$53,$C9,$87,$C0,$71,$5E,$17,$67,$AE,$79
	.dc.b	$B4,$CC,$6D,$E7,$87,$A1,$FB,$33,$C2,$74,$7A,$AB,$C1,$3F,$ED,$36
	.dc.b	$BF,$AD,$6B,$FE,$FE,$31,$5F,$86,$F7,$7F,$DE,$13,$CA,$BF,$DF,$6F
	.dc.b	$E0,$71,$F9,$07,$FB,$C2,$3C,$15,$F7,$CF,$F7,$FC,$EE,$FE,$8D,$6C
	.dc.b	$E2,$E0,$87,$F1,$3B,$F0,$D7,$23,$C9,$1F,$EF,$BA,$F2,$57,$38,$E3
	.dc.b	$1D,$DE,$46,$E7,$93,$FF,$00,$0F,$96,$7F,$B2,$E3,$9B,$FA,$4A,$F0
	.dc.b	$BF,$82,$6B,$F9,$16,$DF,$C4,$AF,$2C,$F4,$3F,$29,$8A,$B7,$87,$9E
	.dc.b	$7F,$2E,$7A,$2B,$CA,$F1,$BE,$8F,$9D,$F9,$AF,$9E,$37,$18,$8F,$21
	.dc.b	$EF,$3D,$05,$32,$75,$9F,$A2,$FC,$63,$07,$D2,$BF,$43,$C6,$7C,$6E
	.dc.b	$8E,$0F,$84,$FD,$4C,$70,$8D,$61,$F3,$1D,$D7,$82,$3E,$B4,$EE,$B9
	.dc.b	$4F,$92,$30,$EE,$FF,$19,$7E,$C3,$AB,$F0,$A3,$51,$9A,$E6,$FA,$F0
	.dc.b	$FA,$2F,$DD,$3D,$93,$F4,$F0,$FC,$41,$E7,$C3,$C3,$3F,$54,$FA,$07
	.dc.b	$59,$EE,$FD,$EA,$EA,$FB,$8E,$06,$4E,$E7,$84,$F4,$9C,$CD,$DF,$CA
	.dc.b	$8E,$75,$E2,$9F,$C3,$92,$FC,$D3,$A2,$47,$63,$C4,$6F,$0F,$92,$FE
	.dc.b	$0A,$FE,$2B,$D2,$FB,$5F,$C2,$AF,$99,$3E,$4C,$3E,$34,$7E,$7A,$F8
	.dc.b	$D7,$8A,$BC,$17,$CA,$B5,$7F,$1A,$FD,$95,$AC,$3C,$EF,$C2,$3C,$93
	.dc.b	$B3,$F0,$51,$A9,$EF,$5E,$4A,$F6,$1F,$24,$A7,$CF,$3A,$47,$E6,$0F
	.dc.b	$7D,$F9,$63,$F5,$97,$DD,$FC,$31,$E5,$21,$C9,$F8,$2F,$03,$E2,$1C
	.dc.b	$0D,$1F,$76,$7C,$EF,$97,$0F,$98,$DB,$F8,$AF,$23,$E9,$61,$EF,$BF
	.dc.b	$24,$60,$F1,$DF,$47,$E3,$9C,$6F,$E9,$3E,$23,$F0,$2F,$81,$D9,$F1
	.dc.b	$A3,$E9,$BB,$C3,$E2,$3F,$BD,$6C,$7C,$8B,$F2,$48,$7B,$AC,$BF,$C7
	.dc.b	$11,$FE,$4C,$8C,$38,$37,$15,$A6,$63,$FC,$24,$48,$FF,$0F,$E8,$F7
	.dc.b	$1E,$EF,$F8,$72,$B5,$7E,$3C,$5A,$78,$71,$E3,$87,$0A,$D4,$FF,$98
	.dc.b	$3D,$D7,$BB,$E3,$6C,$F9,$DA,$FB,$1A,$FF,$95,$5C,$E1,$E0,$8F,$89
	.dc.b	$18,$AE,$D5,$E2,$F7,$7F,$C6,$73,$9A,$DC,$E0,$D3,$E6,$3E,$77,$B8
	.dc.b	$F7,$3E,$43,$63,$E6,$35,$FA,$2D,$7F,$CD,$4E,$71,$F3,$42,$31,$8F
	.dc.b	$FA,$09,$B1,$FE,$AC,$C9,$EE,$7F,$92,$A7,$33,$C7,$0C,$1F,$3A,$7F
	.dc.b	$C9,$FE,$DB,$DC,$FB,$CD,$AF,$C1,$35,$FC,$3B,$5F,$F4,$03,$9C,$BC
	.dc.b	$ED,$CB,$FD,$2E,$D9,$7B,$8F,$73,$FD,$25,$7C,$30,$F1,$92,$BE,$65
	.dc.b	$6F,$DC,$FF,$32,$FB,$7F,$93,$93,$48,$FE,$5D,$FF,$4A,$5F,$CA,$6E
	.dc.b	$5F,$EA,$A6,$CF,$FB,$53,$DC,$7A,$3D,$A2,$1F,$74,$EE,$7E,$7A,$36
	.dc.b	$E6,$5E,$97,$84,$91,$FA,$B8,$CB,$63,$D6,$47,$F1,$5D,$0A,$7C,$57
	.dc.b	$FA,$13,$DD,$7B,$B4,$97,$E5,$F7,$1E,$EB,$6F,$E0,$1E,$CB,$DB,$FF
	.dc.b	$64,$E3,$CB,$0F,$63,$5F,$F6,$1B,$6B,$F2,$38,$74,$79,$D7,$FB,$4A
	.dc.b	$31,$4B,$EC,$FF,$0F,$BE,$91,$88,$46,$2B,$96,$1E,$79,$E5,$18,$C3
	.dc.b	$BC,$73,$C7,$FD,$C3,$87,$3C,$39,$4F,$F1,$DC,$7C,$73,$DD,$B8,$7C
	.dc.b	$0B,$E2,$DA,$EA,$27,$6F,$3C,$7C,$97,$DE,$F8,$CF,$8C,$E3,$5C,$1E
	.dc.b	$55,$C2,$FB,$47,$B0,$E0,$53,$78,$F2,$9E,$11,$AC,$3C,$0E,$66,$3C
	.dc.b	$71,$F8,$2F,$CC,$C7,$CA,$F4,$3B,$63,$F5,$9F,$F7,$B3,$5F,$9C,$7F
	.dc.b	$42,$70,$FC,$BF,$71,$EE,$FF,$43,$94,$62,$78,$35,$EA,$D7,$FA,$6F
	.dc.b	$BB,$FB,$72,$3F,$D2,$DB,$97,$E1,$7D,$DF,$D4,$F2,$7D,$4A,$FE,$B1
	.dc.b	$5F,$5A,$73,$1F,$24,$C9,$F2,$DC,$D7,$E3,$E3,$2F,$5E,$2F,$E8,$0E
	.dc.b	$61,$5D,$67,$3B,$FA,$27,$A4,$61,$FE,$D4,$74,$8B,$7E,$07,$85,$F7
	.dc.b	$C3,$D5,$5D,$8E,$6E,$30,$F0,$36,$32,$7B,$4E,$13,$CD,$C4,$F0,$69
	.dc.b	$C2,$35,$B7,$F4,$8C,$3E,$3B,$FD,$2A,$7D,$07,$F8,$0E,$33,$7E,$AC
	.dc.b	$7F,$B2,$57,$1E,$4B,$39,$3F,$91,$57,$F2,$66,$F2,$EB,$0E,$8D,$F5
	.dc.b	$5E,$F3,$E3,$75,$4C,$17,$86,$E1,$FE,$11,$6C,$FF,$85,$9A,$FD,$23
	.dc.b	$94,$FB,$5F,$F0,$F7,$C1,$FF,$84,$2B,$FC,$34,$D8,$FF,$0F,$B5,$E4
	.dc.b	$D7,$EE,$FC,$0B,$E0,$BA,$42,$31,$5F,$E2,$56,$C7,$F8,$A9,$AF,$97
	.dc.b	$DB,$7B,$6F,$73,$FC,$59,$18,$6F,$B4,$DF,$35,$BB,$FF,$1A,$FC,$FB
	.dc.b	$EB,$3F,$C5,$BF,$01,$FC,$D7,$4F,$F1,$E7,$82,$8C,$5F,$F8,$FE,$F8
	.dc.b	$3F,$D6,$5A,$FF,$45,$AE,$33,$C2,$B9,$76,$FE,$7B,$9B,$FC,$23,$87
	.dc.b	$06,$E3,$FE,$43,$8E,$78,$FD,$36,$BF,$A6,$6B,$FE,$51,$FA,$07,$CB
	.dc.b	$7A,$BE,$02,$4F,$F6,$D9,$CB,$67,$2F,$9C,$65,$EF,$AF,$36,$9F,$B4
	.dc.b	$BE,$35,$B0,$E1,$FC,$43,$5E,$17,$FE,$5D,$8C,$47,$F9,$9F,$C9,$E7
	.dc.b	$9F,$F8,$CB,$D5,$7B,$5E,$0A,$FF,$36,$46,$1F,$F0,$35,$70,$C7,$FC
	.dc.b	$43,$8E,$E5,$5B,$7C,$CD,$C3,$FC,$53,$3F,$E3,$43,$84,$66,$BF,$88
	.dc.b	$B6,$BC,$D3,$B7,$FC,$2D,$19,$7B,$E1,$97,$E2,$61,$E5,$7D,$31,$D8
	.dc.b	$D6,$1E,$28,$E7,$19,$7D,$B7,$AC,$3C,$73,$99,$EB,$39,$7A,$9C,$8E
	.dc.b	$24,$3E,$0C,$71,$AF,$04,$72,$38,$1A,$38,$3A,$3E,$A4,$71,$AF,$9F
	.dc.b	$3A,$7C,$AF,$29,$E1,$1A,$8E,$84,$8C,$BF,$30,$E1,$1C,$9F,$21,$23
	.dc.b	$C2,$E8,$E0,$66,$38,$46,$A3,$06,$9C,$12,$3C,$8F,$8A,$3E,$C9,$D4
	.dc.b	$5F,$D9,$6D,$FC,$17,$E7,$9D,$8C,$9C,$70,$FC,$24,$F1,$39,$3F,$87
	.dc.b	$33,$3F,$66,$7C,$27,$C5,$BF,$29,$1E,$F7,$E9,$7D,$91,$97,$8C,$65
	.dc.b	$F1,$9E,$5A,$F0,$61,$DF,$0F,$1B,$EB,$9D,$D7,$C3,$AD,$C7,$DD,$3B
	.dc.b	$9E,$2B,$F9,$E7,$86,$F9,$C7,$C9,$7E,$06,$1D,$6B,$DF,$3C,$CF,$3D
	.dc.b	$72,$7A,$9E,$27,$0E,$8C,$D7,$67,$E3,$18,$AE,$D8,$7B,$5B,$1D,$E7
	.dc.b	$86,$1F,$15,$F2,$C7,$D7,$AE,$0D,$38,$57,$27,$EF,$4F,$2B,$F2,$BC
	.dc.b	$A3,$C6,$79,$CF,$51,$EC,$C3,$E5,$BC,$AB,$A6,$1F,$91,$AE,$93,$F3
	.dc.b	$4F,$61,$E3,$38,$16,$38,$4F,$2B,$FA,$4D,$9E,$27,$92,$FE,$8B,$F8
	.dc.b	$3C,$3F,$01,$87,$A7,$0F,$C4,$56,$9F,$C1,$4E,$AF,$F2,$D3,$EA,$9F
	.dc.b	$0D,$F8,$F0,$F0,$B4,$F0,$C7,$4B,$F9,$B1,$E3,$6C,$66,$3E,$79,$C4
	.dc.b	$F9,$24,$AF,$05,$76,$8E,$47,$02,$9F,$18,$B3,$E6,$6C,$71,$8F,$04
	.dc.b	$6D,$A6,$CB,$1B,$3F,$27,$3F,$15,$FD,$83,$E4,$9F,$7C,$6A,$BD,$AD
	.dc.b	$8F,$35,$74,$76,$7D,$F3,$F1,$18,$7D,$39,$E1,$1E,$56,$CF,$95,$A7
	.dc.b	$96,$39,$5F,$95,$FC,$85,$FD,$97,$F6,$11,$E7,$7E,$6D,$FE,$41,$F0
	.dc.b	$E1,$F7,$AB,$84,$7E,$1C,$FD,$E4,$65,$A7,$48,$DE,$1F,$88,$8C,$9C
	.dc.b	$4F,$D2,$D7,$B5,$F3,$CF,$C6,$69,$F1,$B0,$F6,$BF,$18,$FD,$66,$1E
	.dc.b	$D7,$F0,$93,$E0,$9F,$A6,$D8,$F8,$27,$8A,$FC,$8F,$8B,$0F,$53,$4F
	.dc.b	$51,$BA,$E6,$D3,$99,$B3,$D9,$3F,$C2,$E7,$9B,$63,$99,$B7,$C2,$65
	.dc.b	$FE,$BA,$7F,$9B,$70,$FF,$11,$BF,$CE,$0C,$4F,$F9,$C5,$FB,$C7,$E8
	.dc.b	$FE,$63,$E4,$34,$F9,$07,$F0,$F7,$33,$FD,$41,$B3,$FE,$1E,$6C,$FF
	.dc.b	$86,$9A,$FF,$89,$9B,$5F,$E3,$36,$DF,$F8,$91,$B3,$FE,$34,$6B,$FE
	.dc.b	$30,$6E,$26,$92,$3C,$E7,$A4,$FC,$F1,$E9,$33,$87,$A5,$B8,$7A,$5A
	.dc.b	$7A,$4A,$7A,$4B,$87,$A5,$B8,$7A,$5A,$7A,$4A,$7A,$4B,$87,$A5,$B8
	.dc.b	$7A,$5F,$F3,$07,$9E,$F8,$9F,$F2,$B6,$CF,$45,$BE,$1B,$63,$8D,$7F
	.dc.b	$95,$EF,$94,$67,$36,$7F,$E9,$0B,$E3,$3F,$E7,$5C,$7B,$63,$C7,$19
	.dc.b	$3F,$87,$18,$F4,$7F,$14,$D7,$FD,$78,$D7,$FD,$DF,$E6,$7F,$AF,$F2
	.dc.b	$F0,$C7,$FA,$0B,$CC,$FF,$54,$3E,$5C,$FF,$D4,$71,$88,$FF,$41,$57
	.dc.b	$19,$E7,$1F,$12,$75,$5D,$32,$46,$2B,$75,$D4,$F8,$D7,$EB,$69,$5C
	.dc.b	$26,$EF,$F0,$23,$FC,$EB,$F8,$37,$F9,$F0,$FF,$42,$46,$63,$FC,$EB
	.dc.b	$96,$5D,$9F,$E8,$C7,$67,$86,$38,$47,$ED,$E3,$07,$8F,$EA,$7D,$D5
	.dc.b	$27,$D4,$79,$63,$FD,$29,$67,$51,$F6,$F1,$56,$A7,$FD,$35,$E6,$7F
	.dc.b	$B3,$A8,$D3,$1E,$5B,$CD,$6D,$F8,$DF,$25,$E5,$FE,$3F,$0E,$3D,$7D
	.dc.b	$08,$F9,$B0,$E5,$7C,$6B,$84,$62,$57,$D1,$FC,$D6,$2C,$3B,$4E,$67
	.dc.b	$B9,$CC,$FD,$05,$73,$8D,$36,$FC,$D7,$CE,$B8,$1C,$8F,$0C,$79,$31
	.dc.b	$F7,$C6,$0E,$AE,$2F,$C5,$5C,$E7,$47,$38,$F0,$CE,$EB,$A6,$3E,$0B
	.dc.b	$E1,$5C,$B1,$F6,$D6,$B0,$E7,$8F,$A5,$F8,$73,$A6,$9E,$A8,$D2,$6D
	.dc.b	$F6,$AF,$83,$5C,$6B,$58,$F9,$5F,$6E,$3E,$59,$D5,$F1,$9C,$9E,$4A
	.dc.b	$F8,$AF,$E0,$4E,$75,$FC,$17,$0F,$B9,$19,$6B,$F8,$0F,$31,$E6,$ED
	.dc.b	$B1,$FD,$4E,$51,$88,$E4,$D7,$FA,$C7,$9B,$FB,$62,$3F,$BB,$6E,$9F
	.dc.b	$87,$F2,$5E,$6F,$F1,$08,$FC,$55,$7F,$27,$AF,$C1,$D6,$B9,$7F,$0B
	.dc.b	$39,$3E,$EC,$65,$FB,$E6,$6B,$B5,$F5,$DF,$CB,$3C,$23,$0F,$C7,$C3
	.dc.b	$BC,$62,$3F,$55,$18,$C3,$C3,$08,$C3,$E0,$30,$75,$8D,$C6,$AF,$DB
	.dc.b	$3A,$7C,$B3,$A8,$E0,$73,$9E,$17,$E8,$AD,$CE,$B4,$FE,$6E,$B1,$F8
	.dc.b	$51,$88,$F8,$55,$FC,$0E,$73,$85,$C7,$F9,$FD,$F2,$97,$9B,$8A,$D1
	.dc.b	$FC,$72,$32,$DF,$2F,$B6,$3F,$D7,$65,$18,$8F,$23,$63,$33,$FD,$5D
	.dc.b	$B3,$FC,$6A,$3F,$B3,$78,$AF,$A3,$7A,$7E,$3A,$ED,$85,$AD,$4E,$65
	.dc.b	$E5,$BC,$CF,$13,$71,$FE,$CC,$D9,$FF,$17,$35,$FD,$47,$98,$F2,$9E
	.dc.b	$67,$F6,$D8,$C3,$7C,$CF,$EC,$55,$FC,$BE,$72,$FF,$8E,$7E,$19,$F0
	.dc.b	$5F,$D6,$DB,$1F,$DD,$DA,$EE,$3F,$B0,$79,$8F,$37,$FC,$1B,$E7,$7F
	.dc.b	$1B,$E5,$FE,$D4,$FF,$45,$9D,$5F,$F5,$06,$BF,$E4,$96,$BF,$C0,$DA
	.dc.b	$FF,$45,$6B,$FE,$15,$78,$CF,$E3,$1B,$B6,$EF,$C7,$F2,$1F,$D3,$6B
	.dc.b	$F2,$3E,$7B,$CE,$F3,$F9,$73,$2F,$EC,$73,$96,$CF,$F7,$D8,$FF,$30
	.dc.b	$91,$C1,$D7,$1F,$EB,$51,$C6,$7F,$0C,$7F,$8A,$DC,$5F,$F8,$37,$4F
	.dc.b	$E1,$73,$FE,$72,$8F,$F3,$81,$FE,$72,$AF,$F7,$CD,$FF,$9E,$1B,$3B
	.dc.b	$6B,$FE,$7E,$6C,$F1,$8F,$F5,$C7,$93,$FE,$9B,$7F,$D2,$BE,$7F,$F9
	.dc.b	$D3,$CF,$FF,$57,$4E,$5B,$87,$7A,$E2,$D7,$FD,$62,$78,$4C,$94,$C9
	.dc.b	$E3,$32,$53,$37,$E5,$6E,$1F,$E9,$A7,$FD,$24,$D8,$F8,$6D,$3D,$A5
	.dc.b	$8E,$B1,$CE,$BF,$D6,$17,$FE,$C7,$6D,$FD,$5A,$FF,$5C,$9F,$E7,$A3
	.dc.b	$E1,$9F,$E7,$E8,$F8,$2F,$AD,$E2,$FF,$BF,$E7,$E3,$61,$FE,$B9,$8F
	.dc.b	$F5,$E9,$FE,$B1,$29,$EA,$3F,$D6,$77,$EA,$6E,$1E,$A6,$9E,$A7,$D8
	.dc.b	$FC,$43,$25,$3A,$C7,$36,$C7,$36,$9C,$CE,$37,$EA,$7F,$26,$7C,$03
	.dc.b	$E1,$3F,$01,$FA,$CE,$E7,$F9,$BB,$97,$E7,$57,$E2,$4F,$85,$7F,$3D
	.dc.b	$C5,$7C,$38,$C3,$F3,$63,$78,$7D,$38,$F1,$57,$F4,$4A,$F9,$B1,$C6
	.dc.b	$75,$7E,$1C,$3E,$2B,$E5,$7F,$DE,$32,$AE,$2F,$D1,$3E,$19,$F2,$5F
	.dc.b	$1B,$F0,$30,$F1,$BF,$76,$FF,$3D,$7D,$63,$F0,$F8,$79,$67,$35,$EE
	.dc.b	$9E,$D7,$D2,$B5,$1F,$E6,$C8,$FF,$73,$92,$BF,$CE,$C4,$8F,$F5,$14
	.dc.b	$65,$B1,$97,$EA,$F9,$FF,$EB,$BA,$EB,$18,$3F,$D7,$E4,$3F,$39,$5F
	.dc.b	$ED,$36,$D7,$FB,$4D,$A7,$FB,$4C,$A7,$FB,$4C,$B3,$FE,$D3,$6E,$3F
	.dc.b	$ED,$36,$9F,$ED,$32,$9F,$ED,$32,$E3,$FE,$D3,$6E,$1E,$99,$FF,$3F
	.dc.b	$61,$FC,$4C,$F3,$5F,$B0,$91,$AC,$3D,$11,$8B,$FF,$6A,$57,$F1,$7B
	.dc.b	$E9,$18,$AF,$E3,$91,$88,$FC,$24,$61,$F8,$D5,$CD,$C9,$F5,$E3,$BD
	.dc.b	$F8,$63,$F8,$25,$73,$BF,$45,$6A,$7C,$06,$8E,$B5,$C5,$FA,$C7,$C2
	.dc.b	$32,$70,$33,$7E,$3A,$F6,$46,$8F,$29,$A9,$E9,$3B,$AF,$14,$EA,$30
	.dc.b	$FD,$D6,$BF,$C0,$2B,$E7,$D7,$22,$47,$8E,$7F,$00,$70,$3E,$36,$1F
	.dc.b	$6D,$C4,$F0,$9F,$C0,$46,$1E,$51,$F8,$78,$C4,$79,$63,$07,$1A,$F7
	.dc.b	$C7,$C5,$24,$F0,$8F,$3C,$74,$21,$CC,$FC,$94,$6E,$FE,$1C,$FD,$69
	.dc.b	$F8,$55,$DE,$7F,$DD,$A6,$8F,$C2,$47,$82,$34,$FE,$0C,$E2,$78,$27
	.dc.b	$CF,$5D,$E3,$27,$8C,$91,$E8,$6D,$78,$1D,$C7,$95,$FD,$84,$F3,$6D
	.dc.b	$78,$9F,$B1,$5E,$77,$F1,$C7,$F3,$AF,$31,$E7,$7F,$5D,$F3,$9E,$7B
	.dc.b	$83,$0D,$BF,$F7,$FD,$F0,$6D,$FF,$AC,$5F,$F7,$94,$65,$C3,$5F,$F5
	.dc.b	$B3,$5D,$C7,$F3,$B8,$E1,$97,$F9,$F9,$B9,$29,$8F,$9B,$E3,$BF,$A1
	.dc.b	$7A,$1F,$EC,$1C,$17,$BA,$7A,$0F,$03,$DC,$DC,$7A,$5F,$FA,$33,$0E
	.dc.b	$58,$77,$94,$7B,$5C,$35,$F2,$35,$D9,$97,$9B,$A3,$CE,$78,$BC,$4F
	.dc.b	$EB,$5C,$0D,$3F,$AB,$FA,$12,$BF,$A1,$46,$8F,$04,$EE,$3F,$A0,$C6
	.dc.b	$08,$F0,$21,$FD,$E3,$8B,$FD,$3E,$FC,$0C,$7D,$7E,$87,$C4,$C2,$4E
	.dc.b	$4F,$6C,$7C,$E3,$D8,$47,$E2,$64,$8F,$89,$E7,$FF,$9E,$6F,$FD,$FF
	.dc.b	$5C,$C4,$65,$FF,$7D,$F9,$4F,$2B,$71,$FE,$0E,$9D,$3C,$3D,$0F,$F0
	.dc.b	$07,$9C,$F4,$1C,$3F,$85,$C7,$95,$F4,$C7,$9C,$62,$39,$1B,$8F,$BD
	.dc.b	$7C,$EB,$13,$C4,$FC,$86,$2C,$7F,$09,$87,$38,$C4,$F0,$17,$CD,$4F
	.dc.b	$86,$30,$F1,$AE,$78,$79,$8F,$1E,$3E,$CB,$E3,$18,$C7,$B6,$3E,$09
	.dc.b	$C9,$E6,$BE,$6F,$A4,$F1,$DF,$3C,$3C,$11,$CA,$76,$EA,$FD,$58,$F9
	.dc.b	$B1,$F2,$4F,$53,$38,$73,$8F,$24,$78,$8F,$9A,$5C,$7C,$D8,$FC,$1A
	.dc.b	$E1,$3F,$BB,$BF,$8C,$FE,$BC,$F1,$1F,$C2,$DC,$BD,$8D,$2F,$3D,$E8
	.dc.b	$7E,$D6,$30,$D8,$CC,$F3,$6B,$FB,$BF,$43,$F7,$04,$7F,$78,$DC,$3F
	.dc.b	$11,$E7,$BD,$0F,$E1,$B5,$EF,$3F,$8A,$57,$E0,$67,$5B,$FE,$0E,$72
	.dc.b	$7C,$58,$CB,$C4,$4F,$1A,$EB,$7F,$97,$9E,$2F,$8A,$33,$AF,$F2,$FA
	.dc.b	$E1,$87,$7B,$E6,$77,$AD,$BF,$36,$76,$F8,$5B,$7D,$E3,$8B,$DE,$BA
	.dc.b	$4E,$AF,$C1,$5D,$67,$2D,$3A,$16,$B9,$3F,$E8,$8C,$3E,$0E,$DF,$2A
	.dc.b	$3F,$97,$61,$E4,$8C,$4E,$AF,$F9,$92,$E6,$FA,$33,$AC,$AD,$6A,$6E
	.dc.b	$1F,$D1,$ED,$E7,$E3,$A5,$E8,$3D,$0D,$DA,$F7,$8F,$F2,$1B,$DD,$6A
	.dc.b	$73,$5F,$EF,$49,$CD,$7F,$1D,$9C,$E7,$F8,$69,$CD,$7F,$AA,$3C,$B7
	.dc.b	$9C,$F3,$BF,$A6,$46,$1B,$E7,$7F,$55,$AF,$D2,$4E,$5F,$ED,$DE,$73
	.dc.b	$C9,$FE,$A4,$D8,$FE,$AC,$D7,$71,$FD,$5E,$BF,$BD,$47,$F7,$76,$BF
	.dc.b	$B1,$6B,$FD,$85,$AF,$F7,$1F,$35,$E6,$FF,$61,$6C,$7F,$65,$6B,$FE
	.dc.b	$03,$6B,$FB,$D7,$3A,$7E,$05,$BB,$7F,$6F,$6C,$FF,$7E,$F3,$9E,$77
	.dc.b	$FA,$FE,$73,$26,$9C,$3C,$65,$70,$9F,$EA,$1E,$73,$CC,$FE,$F0,$D8
	.dc.b	$FE,$F8,$D7,$FC,$34,$D7,$F8,$A3,$9C,$BF,$30,$DC,$BF,$C0,$CD,$9F
	.dc.b	$F0,$AE,$BF,$C1,$DF,$0D,$FF,$B5,$E7,$50,$78,$63,$F6,$FA,$3F,$D8
	.dc.b	$B3,$FE,$DF,$3F,$DB,$47,$FB,$34,$FF,$26,$1F,$DE,$FC,$EF,$F6,$8B
	.dc.b	$4D,$94,$FF,$66,$DF,$FB,$8A,$37,$97,$FB,$D1,$B9,$7F,$B8,$BC,$8F
	.dc.b	$F7,$0C,$E6,$FC,$47,$23,$C4,$7F,$BF,$AB,$FD,$B5,$1B,$8F,$F7,$21
	.dc.b	$E2,$70,$4A,$FF,$2A,$36,$3F,$C7,$9E,$3F,$FB,$9D,$FF,$7D,$CE,$8F
	.dc.b	$26,$1F,$12,$3A,$9E,$33,$DD,$5C,$DF,$F3,$39,$E6,$7F,$D2,$37,$EE
	.dc.b	$8F,$F7,$9D,$75,$69,$FE,$75,$AF,$8C,$FC,$A3,$FA,$A6,$1F,$EE,$D7
	.dc.b	$FC,$BC,$DB,$FF,$4A,$E1,$FE,$9D,$BE,$95,$FE,$7B,$8C,$3D,$4F,$B3
	.dc.b	$3F,$EA,$69,$DD,$7F,$9F,$CF,$F3,$E4,$66,$78,$1F,$E7,$A3,$FD,$03
	.dc.b	$7F,$E9,$08,$FF,$40,$1B,$8F,$82,$6C,$FF,$41,$9C,$9F,$F5,$9F,$9D
	.dc.b	$FE,$D3,$9E,$57,$E2,$7F,$D0,$D5,$FE,$C7,$38,$47,$FA,$DF,$0F,$5C
	.dc.b	$76,$BE,$6F,$FB,$3F,$0F,$4D,$7F,$AD,$A3,$89,$FE,$B0,$AE,$33,$A8
	.dc.b	$FF,$58,$CF,$85,$F6,$9F,$E9,$6A,$F6,$C7,$38,$F0,$3E,$3C,$3D,$51
	.dc.b	$B8,$E4,$FC,$03,$67,$86,$36,$FB,$4C,$9F,$5D,$CC,$FD,$C6,$CF,$DC
	.dc.b	$BF,$4E,$1F,$22,$FD,$A7,$13,$E1,$39,$7A,$E1,$F7,$63,$0F,$C1,$8F
	.dc.b	$F7,$0D,$7A,$30,$FB,$53,$9B,$EF,$5E,$D3,$BC,$73,$70,$76,$BF,$B3
	.dc.b	$5E,$98,$F1,$9A,$9E,$98,$7B,$23,$D3,$7D,$9F,$89,$3A,$78,$98,$BF
	.dc.b	$49,$1E,$8D,$74,$60,$86,$09,$7E,$57,$FD,$D0,$7B,$23,$D3,$87,$CF
	.dc.b	$C3,$CD,$3F,$5B,$0F,$35,$F1,$3F,$0B,$1F,$8A,$39,$1F,$8A,$3B,$D7
	.dc.b	$37,$D7,$1C,$63,$E8,$1F,$8D,$C3,$E0,$4F,$90,$F9,$87,$39,$E0,$F6
	.dc.b	$3F,$1D,$1C,$23,$C7,$1F,$70,$D3,$F7,$30,$F8,$CF,$28,$EA,$72,$8C
	.dc.b	$D7,$9E,$3C,$55,$D6,$3C,$F1,$EF,$38,$D7,$57,$E7,$BD,$67,$84,$78
	.dc.b	$A7,$F5,$5E,$4F,$FB,$C6,$3F,$DD,$8D,$7F,$DA,$5E,$67,$FB,$76,$30
	.dc.b	$7F,$BC,$A3,$71,$FE,$F7,$3F,$DC,$D7,$FE,$64,$D7,$FD,$E9,$3F,$EB
	.dc.b	$FA,$E0,$F4,$31,$5C,$2F,$FD,$ED,$E6,$BC,$DF,$F7,$2D,$7F,$B3,$0F
	.dc.b	$F7,$51,$0C,$1F,$EC,$BF,$2F,$FD,$DA,$FF,$BC,$0F,$F6,$C6,$3F,$F0
	.dc.b	$03,$C1,$E7,$1F,$EC,$6C,$3F,$DF,$2E,$33,$F4,$79,$BF,$EE,$48,$D0
	.dc.b	$1A,$FF,$BB,$DF,$E8,$3A,$FF,$BB,$9A,$7F,$BC,$8F,$F7,$7D,$7F,$97
	.dc.b	$67,$83,$FE,$5F,$DF,$B3,$9A,$FF,$7E,$57,$0C,$3C,$0A,$B7,$1F,$EF
	.dc.b	$68,$FF,$77,$F9,$9F,$EF,$17,$FD,$50,$34,$FF,$79,$3A,$3E,$34,$7F
	.dc.b	$BE,$5F,$8B,$97,$FB,$8F,$1F,$F3,$97,$97,$FE,$EE,$27,$99,$FE,$EE
	.dc.b	$F3,$3F,$D6,$F7,$FE,$BC,$AD,$CF,$D9,$8F,$F6,$21,$F2,$0F,$F5,$CF
	.dc.b	$91,$FE,$BF,$9F,$F6,$74,$43,$FD,$8D,$E6,$7F,$B3,$5F,$F6,$B6,$1F
	.dc.b	$EC,$6F,$2B,$FD,$9F,$1C,$9F,$F6,$6D,$FF,$B6,$23,$FD,$99,$87,$FB
	.dc.b	$62,$F9,$46,$CF,$1C,$EC,$F1,$3E,$0C,$7D,$78,$77,$8D,$E1,$D8,$8E
	.dc.b	$08,$75,$24,$71,$7F,$DB,$F5,$D8,$EE,$FF,$B5,$AB,$FD,$F4,$DC,$7F
	.dc.b	$DA,$D1,$88,$FB,$F8,$7F,$B6,$1F,$F6,$B7,$99,$FE,$D9,$23,$FE,$D4
	.dc.b	$21,$FE,$D9,$29,$FE,$B4,$BF,$F5,$79,$FE,$BB,$F2,$FF,$D7,$11,$FE
	.dc.b	$C1,$3F,$D6,$99,$6C,$FF,$56,$F5,$FF,$B1,$A3,$FD,$66,$68,$E1,$5D
	.dc.b	$27,$84,$7F,$AB,$67,$83,$FE,$D6,$8E,$7C,$1F,$EC,$4A,$E1,$3C,$A7
	.dc.b	$2F,$FB,$1A,$32,$F8,$1A,$F3,$3B,$4F,$39,$E3,$3F,$EC,$CC,$FF,$D8
	.dc.b	$53,$FB,$77,$FD,$99,$5C,$4F,$F5,$E5,$FF,$AE,$4F,$F5,$D7,$25,$AF
	.dc.b	$F5,$C3,$FE,$C2,$F3,$DE,$87,$2B,$FE,$05,$18,$AF,$D7,$C2,$30,$FE
	.dc.b	$C9,$AF,$FA,$E3,$C8,$FF,$5F,$3F,$B6,$F1,$FF,$D7,$4F,$FA,$F3,$C6
	.dc.b	$FF,$5F,$3F,$C0,$23,$2F,$F8,$BF,$C0,$FF,$2C,$3F,$E6,$43,$FC,$86
	.dc.b	$7F,$95,$3B,$FC,$6F,$F1,$66,$F5,$FF,$A6,$63,$F8,$5F,$E3,$9F,$C2
	.dc.b	$9F,$F5,$67,$B3,$FB,$9E,$EF,$F5,$3B,$FB,$B8,$4E,$DB,$BE,$FC,$55
	.dc.b	$53,$85,$6C,$FE,$55,$5F,$C6,$8E,$27,$C8,$72,$F9,$2B,$D3,$8B,$E6
	.dc.b	$3B,$12,$7E,$E7,$07,$F2,$49,$D4,$60,$FE,$6C,$E1,$FE,$73,$97,$FB
	.dc.b	$0A,$B9,$47,$F1,$59,$D7,$1A,$FA,$57,$88,$95,$F7,$AB,$A3,$FE,$82
	.dc.b	$F0,$3E,$37,$80,$F4,$A9,$FC,$EA,$7F,$91,$6F,$FD,$03,$1F,$E7,$75
	.dc.b	$B9,$F4,$1F,$CF,$6B,$86,$5F,$D0,$E3,$33,$B7,$F9,$63,$63,$51,$DB
	.dc.b	$0F,$0B,$C2,$B9,$DF,$F4,$AC,$3C,$53,$FE,$AE,$9D,$BC,$4F,$EA,$06
	.dc.b	$4E,$46,$0F,$EC,$35,$C2,$FF,$AD,$4F,$69,$FE,$CD,$5D,$9F,$61,$C8
	.dc.b	$EC,$49,$E2,$E5,$F5,$92,$FE,$3D,$77,$7D,$57,$DE,$BA,$9F,$12,$71
	.dc.b	$7E,$1C,$3E,$35,$F2,$AF,$54,$E6,$BA,$47,$A2,$FA,$5F,$08,$F8,$47
	.dc.b	$9E,$FC,$2F,$9D,$E8,$E4,$E4,$64,$E3,$87,$AD,$C9,$E7,$34,$7C,$C3
	.dc.b	$E7,$4F,$85,$F6,$C7,$17,$47,$12,$39,$C3,$D9,$1D,$1F,$A4,$61,$DD
	.dc.b	$7A,$9F,$41,$D9,$F5,$CF,$3B,$F8,$B3,$93,$D3,$1E,$3B,$F3,$1E,$7B
	.dc.b	$EF,$5C,$23,$0F,$CC,$C3,$9C,$EA,$FE,$0D,$69,$F8,$33,$AA,$E1,$1E
	.dc.b	$B6,$BE,$A3,$C7,$87,$AF,$0F,$C5,$C6,$4F,$8E,$F9,$EB,$A4,$76,$3A
	.dc.b	$14,$E1,$87,$95,$E5,$87,$95,$F8,$EF,$C0,$9F,$C7,$E1,$EF,$6C,$FB
	.dc.b	$1A,$F2,$6D,$F9,$89,$87,$A4,$93,$F8,$87,$E9,$98,$7D,$6E,$1D,$1F
	.dc.b	$08,$87,$D7,$26,$1E,$B9,$EF,$1C,$0E,$67,$0C,$3D,$8F,$E9,$B0,$F6
	.dc.b	$34,$F6,$14,$F6,$16,$7A,$BE,$08,$FA,$F1,$8A,$F1,$5F,$A6,$B4,$7D
	.dc.b	$19,$D5,$70,$9D,$57,$0A,$F4,$5F,$D3,$BF,$B9,$1F,$68,$FA,$B8,$7D
	.dc.b	$E6,$99,$8D,$BF,$7B,$0E,$CF,$C3,$BE,$CF,$C9,$BE,$D8,$7D,$17,$E2
	.dc.b	$C7,$EA,$27,$F5,$97,$E8,$8F,$55,$7A,$2F,$AC,$7B,$8F,$19,$E1,$36
	.dc.b	$78,$5D,$DF,$82,$3E,$A6,$1E,$BB,$F4,$9D,$E3,$9B,$4C,$DF,$86,$3A
	.dc.b	$9E,$18,$CB,$DC,$CD,$77,$7F,$06,$7C,$1A,$FC,$BD,$7D,$79,$D6,$1F
	.dc.b	$02,$3B,$57,$89,$B7,$DA,$3C,$11,$F2,$6F,$D4,$D8,$FA,$8D,$7E,$21
	.dc.b	$CC,$F3,$14,$E6,$7C,$43,$89,$4E,$33,$E1,$BF,$3E,$1F,$9A,$9F,$B0
	.dc.b	$D9,$FE,$3D,$1F,$9C,$32,$57,$2F,$E7,$8C,$95,$C9,$FA,$09,$EF,$1C
	.dc.b	$6B,$F7,$18,$7C,$78,$F5,$D7,$97,$0F,$9B,$1F,$9C,$6B,$F4,$4F,$D6
	.dc.b	$57,$D1,$8C,$DF,$E4,$AB,$C3,$39,$AF,$DA,$1F,$E1,$53,$FD,$A4,$7F
	.dc.b	$B5,$09,$1C,$1F,$F6,$E4,$FF,$8F,$6F,$F4,$27,$F6,$9D,$3C,$6D,$9F
	.dc.b	$19,$FC,$1E,$33,$8F,$F8,$89,$B1,$FE,$18,$6B,$FE,$18,$6D,$FF,$B8
	.dc.b	$5B,$87,$F8,$91,$B3,$E4,$6B,$FE,$E1,$6E,$1E,$7B,$F8,$71,$82,$4F
	.dc.b	$98,$FC,$DC,$ED,$C3,$5F,$B7,$7E,$37,$86,$1E,$F7,$38,$7F,$01,$BE
	.dc.b	$B1,$83,$E1,$1C,$30,$F8,$4D,$3E,$11,$4F,$84,$5B,$F0,$3C,$CF,$3C
	.dc.b	$FF,$9C,$65,$F0,$1F,$E7,$5F,$13,$FC,$AF,$82,$DB,$27,$31,$9B,$E2
	.dc.b	$FF,$93,$9A,$FF,$99,$23,$73,$C7,$6F,$06,$12,$BF,$D3,$B7,$C1,$FF
	.dc.b	$3D,$35,$F1,$07,$0F,$97,$C3,$74,$FF,$A3,$7C,$2B,$3F,$E7,$7C,$51
	.dc.b	$88,$6B,$FE,$73,$79,$57,$FA,$4E,$F5,$39,$BE,$33,$D4,$D1,$2B,$FD
	.dc.b	$6B,$5D,$5F,$F4,$93,$5F,$4B,$5F,$47,$87,$E7,$E1,$91,$FE,$76,$94
	.dc.b	$F8,$CC,$C7,$FA,$3B,$1F,$F1,$C4,$EA,$38,$4F,$FA,$D2,$31,$1F,$EA
	.dc.b	$30,$E1,$F2,$C7,$FA,$EB,$EB,$FF,$DA,$31,$FE,$9D,$C9,$F8,$F3,$E0
	.dc.b	$FE,$8C,$7C,$73,$2F,$FA,$E6,$B8,$D7,$CC,$E1,$FA,$F1,$8A,$E1,$7C
	.dc.b	$EB,$B3,$5E,$07,$E3,$E7,$85,$7E,$03,$C5,$78,$DF,$EC,$76,$D7,$FA
	.dc.b	$71,$AF,$21,$E2,$BC,$6F,$BF,$0F,$0D,$E2,$75,$C3,$A1,$27,$FD,$4F
	.dc.b	$7C,$74,$FD,$15,$F0,$9F,$D2,$4E,$5E,$97,$CD,$DB,$5C,$9C,$30,$F1
	.dc.b	$D7,$38,$F3,$C6,$27,$84,$E7,$1F,$6B,$BB,$F0,$36,$B8,$5F,$ED,$EB
	.dc.b	$77,$C8,$FD,$4D,$F0,$6D,$70,$A5,$70,$79,$63,$FB,$0C,$3B,$63,$EC
	.dc.b	$9F,$53,$5E,$33,$BC,$3C,$11,$A3,$E0,$9E,$E9,$E1,$7D,$E3,$89,$CF
	.dc.b	$1F,$33,$C7,$0F,$85,$87,$58,$F2,$9B,$C7,$D5,$5F,$4E,$3F,$83,$9D
	.dc.b	$CF,$D1,$39,$7E,$0E,$1F,$A5,$7F,$0F,$EB,$3C,$7C,$52,$31,$3C,$9A
	.dc.b	$F6,$8D,$4F,$38,$FE,$51,$E3,$D4,$D2,$3B,$7E,$57,$A7,$FC,$6B,$6F
	.dc.b	$99,$3C,$8F,$93,$19,$7E,$59,$9D,$3F,$15,$19,$7F,$19,$19,$DB,$F9
	.dc.b	$4C,$E5,$FE,$09,$7E,$0E,$6F,$E4,$57,$C2,$FC,$E7,$82,$33,$5C,$23
	.dc.b	$18,$79,$9D,$D7,$8A,$B7,$1F,$4A,$FB,$56,$DB,$1E,$A3,$8D,$F9,$AB
	.dc.b	$8C,$7F,$40,$9D,$DF,$95,$B5,$E3,$8F,$05,$F4,$AF,$C0,$E1,$F1,$DB
	.dc.b	$87,$F3,$F6,$CF,$E0,$67,$E6,$69,$95,$D9,$FC,$7E,$7C,$87,$F6,$39
	.dc.b	$D4,$66,$BF,$B3,$78,$8F,$19,$E9,$7E,$36,$30,$DF,$1D,$E9,$E6,$B7
	.dc.b	$7F,$CD,$FD,$37,$A1,$F8,$B6,$C7,$E5,$DA,$EE,$3F,$21,$E2,$3C,$6F
	.dc.b	$E6,$FB,$7F,$68,$F1,$9E,$56,$30,$FF,$19,$B5,$5F,$17,$DD,$5F,$DC
	.dc.b	$23,$11,$FD,$2D,$AF,$EB,$9A,$EE,$FC,$67,$58,$E3,$C1,$F4,$5B,$5E
	.dc.b	$5F,$3F,$C5,$E4,$FD,$0C,$3F,$BC,$79,$6F,$2F,$D1,$E3,$A3,$13,$E6
	.dc.b	$3F,$C2,$D5,$C5,$FE,$84,$D7,$F8,$1B,$5F,$F0,$5B,$5E,$3C,$BF,$B9
	.dc.b	$8E,$59,$FE,$1D,$B9,$70,$9F,$1D,$FF,$87,$AB,$F1,$9A,$7F,$87,$A2
	.dc.b	$79,$5F,$36,$3F,$C4,$F8,$FD,$58,$FE,$0D,$06,$92,$3F,$8A,$47,$02
	.dc.b	$47,$C2,$E0,$FF,$26,$5F,$FA,$0F,$C8,$FF,$3B,$9F,$EA,$0B,$FF,$4B
	.dc.b	$47,$FA,$DF,$2F,$F5,$03,$63,$31,$FE,$AA,$32,$57,$27,$FA,$29,$CB
	.dc.b	$4C,$9F,$EA,$63,$FC,$C3,$3E,$16,$9F,$EA,$82,$9F,$EA,$88,$E6,$F1
	.dc.b	$DF,$FD,$30,$DF,$27,$FD,$51,$3F,$E9,$C6,$C7,$F9,$4E,$7F,$DD,$3C
	.dc.b	$5F,$E9,$5A,$D4,$FF,$A8,$CD,$57,$FB,$7A,$FE,$2C,$78,$EF,$BD,$FF
	.dc.b	$AA,$A3,$FD,$5D,$7C,$9A,$72,$3F,$DB,$35,$C9,$F5,$1F,$0A,$79,$61
	.dc.b	$F4,$E3,$FD,$03,$3F,$0E,$FB,$BC,$67,$E9,$4F,$C4,$6C,$EA,$38,$9E
	.dc.b	$7C,$3E,$0B,$FE,$E8,$8D,$3F,$E7,$BB,$F8,$4F,$9C,$FF,$6D,$3E,$4C
	.dc.b	$3F,$D7,$11,$BC,$3C,$97,$D2,$FE,$5C,$62,$7E,$51,$DE,$74,$7C,$B7
	.dc.b	$CA,$7C,$23,$D5,$5F,$89,$69,$F8,$72,$9F,$74,$EB,$5E,$07,$DC,$62
	.dc.b	$7B,$3E,$73,$27,$D4,$3C,$86,$CF,$AA,$79,$30,$F9,$71,$E8,$8C,$9C
	.dc.b	$CF,$44,$FC,$A9,$FF,$6B,$47,$FB,$8C,$D3,$C9,$EC,$7F,$93,$63,$FD
	.dc.b	$24,$FF,$B8,$0F,$F7,$11,$FE,$E1,$29,$0F,$E5,$53,$FC,$D8,$8F,$FB
	.dc.b	$1B,$5F,$F5,$FD,$78,$AB,$FD,$88,$DA,$FF,$5F,$D7,$FA,$E6,$73,$7F
	.dc.b	$97,$AF,$E5,$95,$A7,$85,$7F,$B1,$0E,$11,$99,$E0,$7F,$B3,$E3,$FD
	.dc.b	$8C,$FF,$B9,$5A,$7F,$B9,$4A,$7F,$B9,$4B,$8F,$F1,$28,$E7,$1B,$3B
	.dc.b	$9E,$1A,$F1,$BF,$ED,$43,$25,$72,$7F,$B5,$8B,$1E,$38,$DC,$66,$3B
	.dc.b	$C7,$FB,$68,$F1,$E1,$F2,$A3,$38,$7C,$69,$FD,$25,$7D,$EB,$F4,$46
	.dc.b	$2B,$E1,$3D,$9F,$04,$F8,$A3,$D8,$D7,$E1,$1B,$33,$1E,$0C,$3E,$74
	.dc.b	$75,$21,$E7,$38,$1C,$A7,$A5,$79,$23,$F3,$A7,$C5,$AE,$8F,$CB,$6B
	.dc.b	$93,$AD,$FF,$24,$C3,$E1,$46,$30,$FE,$E4,$79,$1E,$B8,$7E,$21,$E8
	.dc.b	$6C,$FE,$D8,$79,$DF,$2C,$F4,$8E,$37,$F2,$27,$31,$F4,$88,$7B,$B0
	.dc.b	$F6,$61,$E5,$8C,$CF,$07,$57,$D6,$7F,$5B,$1F,$C7,$63,$18,$7A,$A3
	.dc.b	$CB,$3F,$75,$F1,$9F,$3A,$FB,$CE,$E3,$E3,$1A,$3D,$98,$7C,$AC,$3F
	.dc.b	$19,$1F,$04,$FD,$F4,$FC,$7B,$F1,$CF,$E2,$70,$F8,$53,$FE,$01,$B5
	.dc.b	$B7,$79,$7F,$DF,$63,$FC,$2D,$18,$D7,$FC,$61,$7C,$67,$8A,$F2,$FF
	.dc.b	$D1,$4D,$8F,$F4,$B3,$5F,$F7,$E3,$5F,$F4,$FB,$CD,$34,$ED,$4F,$2F
	.dc.b	$FD,$4B,$9F,$F9,$8F,$CB,$77,$7F,$8B,$A6,$E1,$E8,$6B,$FE,$ED,$9B
	.dc.b	$5C,$6B,$FD,$41,$1A,$9F,$F5,$8B,$5F,$8A,$1C,$3F,$21,$E7,$A6,$9B
	.dc.b	$19,$9F,$27,$B7,$FE,$57,$97,$99,$E7,$E2,$97,$F9,$89,$CB,$67,$CA
	.dc.b	$47,$FD,$81,$1A,$7F,$D8,$E7,$96,$3B,$11,$F8,$5E,$17,$F9,$9F,$CA
	.dc.b	$79,$DF,$E6,$BD,$1E,$63,$CD,$C3,$70,$FD,$5D,$F0,$6D,$FF,$B1,$E3
	.dc.b	$11,$E2,$6B,$FE,$DA,$6B,$F2,$BC,$D7,$9B,$FE,$79,$6C,$7E,$E1,$AB
	.dc.b	$1F,$F5,$07,$97,$FE,$EC,$8F,$F7,$79,$9E,$1E,$97,$CC,$FF,$42,$4E
	.dc.b	$CE,$17,$FE,$F7,$3F,$DF,$0E,$7C,$DE,$D7,$FE,$FE,$9D,$57,$FB,$DE
	.dc.b	$3F,$DE,$5E,$53,$6F,$F7,$8C,$6E,$75,$5D,$2F,$FD,$F6,$74,$BF,$F8
	.dc.b	$03,$5F,$AE,$6A,$7A,$3F,$B2,$C3,$C9,$96,$F1,$F6,$CD,$3F,$6F,$87
	.dc.b	$79,$C4,$79,$A7,$6F,$94,$CD,$74,$70,$4B,$F0,$BC,$4F,$16,$1F,$11
	.dc.b	$FD,$08,$94,$66,$3E,$19,$A8,$CD,$FE,$52,$38,$79,$AF,$37,$14,$8C
	.dc.b	$4F,$E6,$A3,$13,$A9,$CC,$F3,$F3,$7A,$3F,$C6,$1A,$FF,$C0,$03,$46
	.dc.b	$9F,$F7,$7B,$FE,$FB,$3F,$E0,$09,$FE,$F9,$72,$FF,$BD,$CC,$E3,$FE
	.dc.b	$F7,$6E,$5F,$EF,$73,$FD,$D6,$F1,$32,$75,$8F,$F7,$DB,$96,$9C,$CB
	.dc.b	$1C,$E3,$F9,$8F,$9B,$FC,$6B,$6F,$B7,$1E,$D3,$A9,$E5,$AE,$2F,$7A
	.dc.b	$F1,$C7,$DA,$A7,$C2,$AF,$E9,$46,$1B,$19,$BF,$E1,$33,$97,$F8,$5F
	.dc.b	$9D,$FF,$00,$1B,$E5,$FF,$BF,$9F,$C1,$C3,$CD,$79,$BF,$7E,$38,$4F
	.dc.b	$E2,$38,$BF,$77,$C3,$DE,$10,$FA,$4F,$E0,$47,$EE,$63,$F8,$17,$26
	.dc.b	$1F,$E0,$91,$97,$F8,$01,$FA,$28,$D3,$83,$AC,$6E,$BC,$15,$E2,$D7
	.dc.b	$F8,$94,$F0,$9C,$DE,$E7,$37,$E0,$3C,$64,$3B,$5F,$A5,$D1,$89,$E4
	.dc.b	$FE,$DA,$FB,$57,$ED,$2F,$BC,$F0,$7C,$B3,$C2,$3F,$8E,$C6,$1D,$4F
	.dc.b	$12,$3D,$4C,$9F,$0C,$F3,$47,$17,$CF,$3D,$6F,$E1,$47,$13,$27,$0A
	.dc.b	$EA,$FC,$A3,$A9,$E4,$73,$1F,$06,$38,$9F,$04,$E4,$75,$AF,$11,$F0
	.dc.b	$1D,$1C,$1C,$CF,$4B,$F8,$F3,$C6,$30,$F8,$AB,$C1,$5C,$C9,$3F,$62
	.dc.b	$7A,$3C,$1A,$73,$3E,$31,$E2,$8F,$0B,$93,$DC,$69,$CB,$F0,$CE,$11
	.dc.b	$F0,$AF,$CF,$1C,$09,$5E,$69,$F2,$57,$9A,$7A,$46,$08,$E0,$87,$78
	.dc.b	$F0,$CF,$D8,$8E,$B5,$E0,$8C,$11,$C5,$FD,$F6,$CF,$D4,$3A,$5F,$81
	.dc.b	$F3,$CF,$39,$FC,$33,$63,$85,$75,$7E,$94,$75,$69,$A9,$EE,$FC,$63
	.dc.b	$E6,$4F,$C3,$7D,$AF,$37,$EA,$9E,$23,$EF,$47,$89,$A7,$8A,$7C,$8F
	.dc.b	$53,$E1,$9E,$23,$13,$DD,$F7,$9E,$08,$E3,$7F,$4E,$38,$9E,$29,$F9
	.dc.b	$73,$AB,$FC,$55,$7D,$2B,$F2,$5E,$DB,$19,$8F,$31,$E8,$3C,$75,$CD
	.dc.b	$F1,$19,$8F,$15,$FC,$37,$EF,$CF,$CD,$9F,$23,$63,$31,$F3,$CF,$80
	.dc.b	$F2,$AF,$A2,$D8,$CD,$FE,$5D,$A7,$D2,$9D,$C6,$63,$B9,$D4,$A7,$CC
	.dc.b	$8D,$BC,$4E,$67,$CA,$9F,$2B,$B3,$37,$ED,$AF,$C8,$3F,$3A,$3E,$14
	.dc.b	$7C,$E3,$35,$F3,$1B,$1F,$30,$FA,$05,$AF,$A0,$FC,$E3,$99,$F3,$8D
	.dc.b	$99,$2B,$F4,$2F,$EA,$47,$E4,$4F,$F3,$5D,$FF,$94,$9F,$EB,$51,$FE
	.dc.b	$10,$8C,$47,$F8,$B8,$C9,$FE,$1F,$72,$FF,$0E,$32,$78,$67,$FC,$B7
	.dc.b	$C9,$FE,$45,$E0,$F8,$13,$FC,$03,$93,$FC,$AF,$87,$C1,$8F,$F1,$9D
	.dc.b	$FF,$8F,$B0,$F9,$11,$FA,$1A,$DB,$F0,$F1,$FF,$2A,$F0,$FF,$93,$AF
	.dc.b	$DD,$3F,$EE,$4A,$DB,$E3,$AD,$E3,$E2,$8F,$E3,$53,$AA,$E3,$1F,$DB
	.dc.b	$63,$15,$B6,$C7,$CF,$9F,$69,$E7,$C7,$D5,$1F,$E4,$D3,$E7,$47,$EC
	.dc.b	$67,$8C,$62,$BC,$6F,$37,$FA,$9C,$6B,$1F,$83,$5D,$AB,$FC,$B6,$F9
	.dc.b	$8C,$63,$F0,$6F,$CD,$97,$CF,$AE,$87,$CF,$70,$F8,$4C,$57,$A5,$F9
	.dc.b	$87,$D2,$3B,$CF,$48,$D4,$7C,$B3,$55,$E1,$7B,$9F,$06,$FD,$35,$EC
	.dc.b	$BF,$0C,$62,$7C,$27,$63,$8E,$1F,$07,$1F,$F0,$33,$63,$F9,$E3,$5F
	.dc.b	$C2,$57,$49,$FA,$33,$9C,$7E,$84,$E6,$BD,$A7,$96,$B7,$19,$8F,$61
	.dc.b	$93,$75,$CF,$0F,$55,$FC,$23,$E9,$3F,$6A,$7E,$61,$F8,$CA,$E9,$3A
	.dc.b	$8F,$8F,$97,$E4,$71,$FE,$BD,$87,$59,$F8,$79,$7C,$B7,$9C,$65,$E9
	.dc.b	$7E,$1C,$7E,$A4,$7C,$93,$78,$F8,$23,$37,$CF,$0F,$A1,$5E,$D3,$DF
	.dc.b	$87,$58,$F1,$DF,$19,$FA,$6D,$7A,$E1,$E1,$72,$D7,$27,$8E,$7B,$D7
	.dc.b	$5C,$BF,$DA,$F7,$CE,$72,$D8,$FB,$ED,$AF,$BE,$D7,$EF,$DF,$7A,$E2
	.dc.b	$DA,$E2,$D9,$E2,$D7,$89,$E5,$8D,$B4,$D9,$4D,$E1,$F2,$63,$6D,$8D
	.dc.b	$B5,$DE,$5F,$76,$7D,$F5,$E1,$3F,$26,$5C,$BF,$17,$5D,$E3,$2D,$8C
	.dc.b	$DF,$FB,$63,$2F,$F6,$7C,$BE,$9F,$FD,$DA,$4C,$7F,$DE,$EE,$F3,$FC
	.dc.b	$E5,$FF,$BC,$8C,$63,$FE,$E6,$8C,$D7,$FB,$AA,$73,$7F,$EF,$78,$FC
	.dc.b	$E6,$6C,$FF,$E0,$08,$9E,$58,$47,$FD,$DD,$5C,$CF,$D2,$65,$FC,$CB
	.dc.b	$0F,$14,$63,$1F,$0D,$7F,$95,$A7,$C2,$49,$E5,$19,$78,$5F,$F1,$29
	.dc.b	$E7,$3F,$AC,$3F,$8C,$1F,$EF,$5A,$FF,$7C,$38,$3F,$DF,$27,$A4,$FF
	.dc.b	$7B,$F2,$FF,$BC,$48,$FF,$BD,$79,$7F,$DE,$6E,$CF,$F6,$B1,$FE,$F0
	.dc.b	$21,$93,$FD,$E8,$7F,$B7,$C7,$1F,$FB,$7D,$FF,$73,$63,$FE,$DA,$35
	.dc.b	$19,$C3,$FD,$D1,$19,$9F,$F7,$01,$FE,$E8,$EE,$FF,$6D,$4E,$A3,$1B
	.dc.b	$FF,$BA,$97,$37,$FB,$F9,$FF,$76,$4F,$FB,$A9,$FF,$76,$5F,$56,$9C
	.dc.b	$8F,$29,$63,$B3,$EA,$BF,$0C,$FA,$1B,$1E,$63,$57,$F8,$5A,$57,$CF
	.dc.b	$E2,$FF,$73,$D5,$E5,$FF,$77,$39,$E5,$FF,$76,$BF,$EF,$2F,$43,$E6
	.dc.b	$2F,$1D,$E1,$E2,$30,$D8,$CC,$70,$6B,$FE,$1E,$E4,$FF,$0E,$11,$FF
	.dc.b	$0D,$35,$FE,$47,$C5,$FC,$3F,$8F,$F6,$7A,$FE,$F5,$BA,$FE,$F6,$BF
	.dc.b	$C1,$D5,$BD,$BF,$6F,$39,$9F,$F0,$54,$E5,$FE,$7B,$19,$AF,$EF,$2D
	.dc.b	$AF,$D7,$E1,$FD,$52,$7F,$83,$57,$03,$FB,$6F,$2F,$F3,$68,$CC,$7F
	.dc.b	$44,$3F,$84,$1B,$3F,$A9,$C6,$DC,$9B,$C7,$D9,$7E,$88,$EB,$7F,$C6
	.dc.b	$1D,$D7,$1A,$DC,$75,$8F,$E9,$66,$70,$F5,$47,$EC,$CF,$0D,$73,$AE
	.dc.b	$31,$E1,$9D,$BF,$D4,$F0,$FE,$CC,$DC,$3F,$AC,$CE,$B9,$7F,$A5,$4F
	.dc.b	$D0,$AF,$83,$8F,$86,$3D,$26,$4F,$4E,$3E,$1B,$E5,$97,$D3,$9F,$11
	.dc.b	$C0,$FA,$F7,$D2,$B8,$46,$63,$EB,$B7,$1F,$AE,$FF,$81,$AB,$A6,$1E
	.dc.b	$38,$F8,$27,$8F,$2F,$83,$19,$3E,$4E,$1D,$A7,$67,$88,$F4,$9C,$71
	.dc.b	$F4,$E3,$E5,$8C,$1E,$33,$E8,$57,$47,$27,$48,$F4,$B7,$1F,$4B,$FA
	.dc.b	$08,$E8,$F8,$70,$F3,$E1,$E5,$BE,$93,$9C,$3C,$6F,$98,$C1,$E3,$3E
	.dc.b	$D1,$FE,$06,$35,$7E,$29,$E6,$DA,$FD,$04,$EB,$0F,$2C,$FE,$13,$2F
	.dc.b	$C0,$CF,$63,$D4,$7C,$0A,$E3,$8F,$A2,$3E,$0E,$1D,$B0,$E6,$7E,$32
	.dc.b	$3C,$D7,$D3,$0F,$14,$7D,$B7,$E1,$98,$7E,$31,$E5,$3C,$06,$CC,$1F
	.dc.b	$00,$FF,$11,$DF,$8D,$F4,$5F,$8D,$E2,$74,$8D,$4F,$73,$E0,$65,$F8
	.dc.b	$CC,$7E,$91,$EE,$CB,$E8,$D7,$0C,$BD,$B8,$7C,$F8,$C3,$DB,$0E,$67
	.dc.b	$B2,$77,$1E,$6A,$E6,$47,$E2,$E1,$E1,$7E,$5E,$1E,$17,$37,$CD,$ED
	.dc.b	$7F,$35,$FA,$A7,$C3,$C7,$FC,$6E,$FC,$6C,$BE,$84,$F8,$A7,$39,$7A
	.dc.b	$2B,$86,$1F,$E3,$0C,$BC,$91,$D3,$0E,$2F,$80,$F8,$11,$AB,$F9,$D1
	.dc.b	$C8,$FF,$1B,$C6,$6B,$D5,$1C,$F2,$F7,$3F,$5A,$33,$5C,$AF,$BC,$F3
	.dc.b	$3F,$C7,$E7,$C2,$3D,$95,$CB,$2F,$B5,$97,$C3,$BE,$53,$98,$F8,$18
	.dc.b	$7B,$6F,$96,$5F,$07,$1E,$E4,$BE,$13,$EF,$3E,$C1,$C3,$2F,$63,$F9
	.dc.b	$3A,$E5,$3C,$23,$D1,$97,$F5,$A7,$81,$0F,$84,$FD,$B7,$78,$FF,$45
	.dc.b	$72,$7B,$A3,$58,$79,$5F,$F2,$15,$FF,$8D,$25,$97,$F9,$25,$CE,$3F
	.dc.b	$C4,$27,$57,$FE,$5C,$7F,$C8,$13,$FE,$3F,$21,$FE,$EB,$8F,$F2,$54
	.dc.b	$FF,$91,$23,$18,$FF,$92,$9E,$F9,$7F,$8F,$6F,$9C,$7F,$A7,$DE,$F3
	.dc.b	$FE,$E2,$BE,$73,$D9,$F1,$5F,$94,$EC,$E2,$3F,$6F,$5C,$B1,$FF,$30
	.dc.b	$47,$F9,$3D,$FF,$0B,$CF,$F9,$DA,$30,$7F,$92,$F9,$7F,$C9,$AF,$F9
	.dc.b	$2E,$1C,$BF,$E4,$E8,$FF,$36,$46,$27,$FC,$3D,$1F,$E4,$C3,$FC,$A5
	.dc.b	$C9,$82,$3F,$E2,$37,$FC,$80,$FF,$90,$8F,$F3,$D1,$FE,$3B,$C7,$FC
	.dc.b	$71,$AF,$F9,$1E,$76,$49,$D9,$FE,$3A,$3F,$C8,$65,$CB,$FC,$87,$CB
	.dc.b	$FE,$57,$BE,$73,$A8,$C3,$E6,$3F,$D2,$D1,$FE,$6F,$8F,$F3,$17,$37
	.dc.b	$F9,$12,$74,$D7,$C0,$DE,$5F,$F2,$3C,$7F,$A3,$DA,$97,$27,$F9,$22
	.dc.b	$3F,$CE,$D1,$AD,$3E,$EC,$7E,$10,$FC,$89,$F6,$B0,$FB,$BB,$FE,$0C
	.dc.b	$FC,$57,$0F,$E5,$E7,$E1,$D7,$E0,$5F,$C5,$57,$0E,$0E,$71,$D4,$FF
	.dc.b	$50,$9F,$E2,$19,$F0,$D7,$3D,$3F,$27,$18,$E1,$FC,$B6,$1F,$E9,$CB
	.dc.b	$E2,$73,$AD,$E3,$E4,$AD,$CF,$58,$C3,$E2,$BE,$8D,$8C,$BE,$A7,$DD
	.dc.b	$8F,$FA,$A7,$0F,$54,$E6,$30,$FE,$8E,$76,$F9,$31,$F5,$57,$9F,$1F
	.dc.b	$DF,$46,$27,$C9,$7C,$9F,$55,$6C,$FD,$C4,$EE,$3F,$73,$3B,$8C,$61
	.dc.b	$F3,$CF,$5C,$F9,$49,$7E,$0C,$7E,$1C,$72,$3C,$CE,$27,$96,$3F,$0A
	.dc.b	$BE,$5E,$1D,$EB,$53,$EE,$9E,$97,$F2,$6B,$4D,$3A,$96,$36,$F9,$31
	.dc.b	$FE,$10,$FC,$99,$FC,$2E,$3E,$F3,$DB,$7D,$EB,$E1,$E1,$E2,$AD,$BF
	.dc.b	$EE,$EB,$ED,$3A,$69,$FC,$92,$F9,$D7,$08,$FA,$18,$FC,$08,$F4,$C7
	.dc.b	$94,$FD,$01,$E5,$30,$7D,$63,$18,$FF,$28,$AF,$46,$1E,$59,$DC,$61
	.dc.b	$F1,$9C,$49,$7E,$17,$E3,$9D,$9F,$B7,$5F,$4A,$32,$D3,$E5,$97,$0F
	.dc.b	$81,$1C,$71,$FB,$51,$F8,$B8,$F8,$58,$79,$6B,$D0,$FC,$98,$E1,$8F
	.dc.b	$C2,$9E,$91,$C0,$F4,$E3,$EB,$76,$79,$2F,$B3,$ED,$8F,$9E,$FC,$2B
	.dc.b	$ED,$5D,$23,$51,$E3,$AF,$C1,$C7,$C4,$C3,$F0,$0F,$84,$EF,$5F,$02
	.dc.b	$7B,$DF,$8E,$3A,$36,$33,$7E,$06,$D7,$E0,$2F,$AB,$E2,$3D,$B3,$B7
	.dc.b	$F6,$A7,$B4,$B7,$E1,$6D,$7D,$08,$F0,$CF,$19,$FD,$74,$6D,$B3,$B3
	.dc.b	$C6,$FF,$1A,$C7,$F0,$53,$B6,$CE,$E7,$B9,$F0,$E7,$BB,$EB,$C7,$F3
	.dc.b	$8F,$3C,$3D,$B8,$FC,$38,$E2,$7B,$CF,$97,$7D,$5F,$29,$D5,$F2,$DF
	.dc.b	$F7,$98,$FB,$47,$72,$9D,$EF,$C3,$1B,$3F,$1B,$1B,$6B,$B3,$8B,$B9
	.dc.b	$FC,$0D,$7E,$22,$36,$66,$7F,$66,$7E,$16,$BB,$D7,$F3,$49,$DB,$4E
	.dc.b	$78,$FE,$36,$B8,$DF,$59,$CC,$7E,$00,$F3,$38,$7E,$29,$F0,$CF,$93
	.dc.b	$1F,$2D,$F8,$E7,$CB,$AF,$A6,$FB,$5F,$1B,$FC,$1A,$7A,$46,$8F,$64
	.dc.b	$7F,$66,$34,$79,$AB,$D0,$E4,$F8,$07,$C8,$9F,$73,$B3,$D4,$78,$5E
	.dc.b	$E7,$DA,$36,$F3,$AF,$84,$F1,$3D,$E7,$AE,$38,$3E,$43,$84,$F9,$31
	.dc.b	$FE,$2B,$3A,$9F,$2D,$FC,$06,$C7,$89,$F9,$35,$FA,$5A,$FC,$EC,$7C
	.dc.b	$D3,$E5,$DF,$CD,$8C,$C7,$8C,$E2,$7D,$63,$D1,$7F,$2E,$FD,$B3,$E4
	.dc.b	$9D,$9F,$0C,$DC,$79,$EB,$E0,$C7,$B0,$F9,$35,$F5,$2B,$94,$79,$2F
	.dc.b	$D1,$8F,$EB,$1B,$8F,$E4,$A7,$E3,$57,$03,$D3,$87,$AE,$F8,$4F,$A4
	.dc.b	$F0,$1F,$42,$35,$5C,$E3,$E4,$9E,$83,$D7,$87,$96,$BF,$9E,$C6,$31
	.dc.b	$F8,$31,$88,$F4,$1D,$F0,$F2,$47,$8A,$3E,$21,$E2,$AF,$94,$FA,$A3
	.dc.b	$94,$FA,$B1,$F7,$CF,$E4,$4D,$61,$FD,$1F,$0F,$99,$3F,$85,$BE,$27
	.dc.b	$7C,$3D,$15,$BB,$FE,$5A,$7F,$A0,$08,$0F,$F3,$EC,$A3,$FD,$07,$19
	.dc.b	$18,$7F,$A0,$5F,$F4,$41,$0D,$1B,$DB,$FD,$02,$E0,$EC,$E1,$FF,$3A
	.dc.b	$DF,$F9,$E3,$99,$C1,$FE,$81,$9F,$F3,$FF,$07,$F9,$E1,$E8,$62,$76
	.dc.b	$FF,$9F,$5B,$5C,$EB,$93,$FE,$79,$AE,$55,$E2,$39,$4E,$B4,$FF,$41
	.dc.b	$5F,$27,$CA,$7C,$1D,$BF,$CF,$A4,$DF,$FC,$F9,$79,$E6,$73,$7F,$AB
	.dc.b	$B7,$F9,$53,$FE,$8E,$3E,$77,$0F,$F9,$E7,$8B,$FD,$18,$DC,$BE,$77
	.dc.b	$02,$F8,$D7,$EF,$6B,$F5,$33,$F8,$CE,$1F,$CC,$46,$2B,$F0,$31,$8E
	.dc.b	$2F,$C2,$3F,$81,$C7,$F2,$A7,$E4,$AF,$93,$F8,$D3,$F4,$27,$E5,$71
	.dc.b	$FC,$6B,$FB,$67,$F3,$F3,$CE,$7F,$2D,$3F,$9F,$BF,$0B,$CF,$0F,$25
	.dc.b	$6E,$7F,$29,$7C,$AB,$FB,$F1,$E4,$E1,$FC,$BB,$8C,$3A,$BE,$7A,$EB
	.dc.b	$3F,$0F,$1F,$6C,$EE,$7A,$BF,$A6,$31,$87,$C3,$7D,$46,$6B,$FC,$01
	.dc.b	$3C,$23,$E0,$DF,$C4,$9E,$D1,$C0,$CC,$F8,$B0,$FE,$5D,$1E,$97,$D9
	.dc.b	$3D,$2F,$C1,$3F,$C3,$F0,$FD,$DC,$F4,$8C,$1F,$26,$7E,$31,$C2,$3F
	.dc.b	$82,$C7,$82,$32,$FC,$D9,$F8,$33,$DE,$30,$79,$F1,$FA,$2F,$C2,$32
	.dc.b	$78,$0F,$99,$5D,$27,$C8,$78,$27,$8B,$D4,$CD,$FE,$12,$B9,$47,$D8
	.dc.b	$35,$1F,$13,$1F,$B3,$19,$8F,$F1,$04,$61,$F8,$E6,$31,$FA,$57,$E0
	.dc.b	$8C,$61,$F0,$DF,$05,$F9,$71,$F8,$31,$AC,$7D,$B3,$97,$E1,$4E,$9F
	.dc.b	$EC,$18,$F8,$5F,$C1,$57,$1A,$F6,$C7,$F3,$D3,$27,$0C,$7C,$B7,$CE
	.dc.b	$3F,$B3,$1F,$81,$36,$62,$FB,$3F,$64,$FA,$55,$D9,$E0,$65,$F9,$2F
	.dc.b	$F5,$13,$38,$FC,$37,$C5,$8F,$C3,$7C,$C7,$1C,$7D,$47,$D1,$C7,$CF
	.dc.b	$1F,$08,$87,$D8,$8F,$A2,$F4,$3E,$4B,$F3,$88,$F6,$AF,$13,$ED,$9E
	.dc.b	$AF,$33,$E0,$C7,$F6,$C7,$99,$C7,$1F,$91,$19,$C3,$E1,$63,$F1,$A3
	.dc.b	$A9,$9C,$3D,$58,$FB,$F0,$F8,$78,$72,$3F,$57,$3A,$8E,$E6,$8F,$7E
	.dc.b	$1F,$83,$7C,$C7,$C7,$3E,$03,$E2,$C7,$D6,$EC,$93,$F0,$A7,$A1,$CB
	.dc.b	$0F,$3D,$F8,$A3,$EC,$63,$EB,$8F,$34,$7E,$F4,$FC,$0D,$79,$EF,$C1
	.dc.b	$8F,$B6,$F8,$57,$A2,$BC,$11,$88,$F0,$5F,$47,$F7,$27,$43,$C3,$8F
	.dc.b	$AA,$B8,$B6,$3E,$91,$D4,$A7,$43,$07,$BB,$0F,$82,$E9,$AE,$67,$FA
	.dc.b	$BC,$FF,$5E,$94,$63,$0F,$EC,$44,$AF,$F0,$16,$1F,$D7,$63,$FC,$1B
	.dc.b	$1F,$E1,$18,$FF,$06,$3F,$E0,$A6,$BF,$D9,$39,$1C,$9F,$E0,$98,$FF
	.dc.b	$07,$F2,$57,$FC,$0A,$7F,$83,$C3,$FE,$1D,$D7,$FC,$05,$C5,$FE,$08
	.dc.b	$7C,$F3,$C8,$FE,$EC,$39,$13,$BC,$FF,$C0,$73,$BD,$39,$F2,$7F,$85
	.dc.b	$F0,$B3,$C8,$FF,$07,$38,$3B,$BF,$E1,$A7,$15,$FE,$25,$C7,$CC,$FF
	.dc.b	$84,$AA,$D7,$F8,$D7,$93,$FC,$23,$C7,$FE,$0E,$8F,$F1,$64,$39,$3F
	.dc.b	$C2,$71,$B9,$D3,$F2,$F8,$7F,$AC,$3F,$86,$E1,$FC,$84,$7E,$16,$3E
	.dc.b	$E1,$F7,$F1,$FC,$4C,$E8,$91,$FA,$EE,$1F,$CA,$6B,$F8,$46,$D7,$09
	.dc.b	$D5,$F5,$9C,$C7,$52,$1E,$02,$3A,$6C,$71,$25,$F8,$AF,$F5,$33,$DA
	.dc.b	$31,$7F,$AD,$24,$7B,$CD,$12,$78,$92,$37,$1F,$E2,$13,$CB,$7E,$32
	.dc.b	$5F,$8A,$39,$4F,$09,$FF,$04,$36,$7E,$7C,$F3,$3D,$77,$E6,$7C,$97
	.dc.b	$E1,$6D,$71,$9F,$EE,$8D,$8F,$DF,$1C,$EF,$C7,$3C,$E3,$17,$E5,$BE
	.dc.b	$91,$89,$D5,$79,$A3,$B9,$0D,$1E,$C3,$0E,$DC,$CF,$48,$C1,$2B,$DE
	.dc.b	$DA,$F4,$D7,$29,$C9,$23,$31,$D2,$FD,$D7,$FD,$8A,$30,$71,$9D,$DF
	.dc.b	$58,$F8,$6D,$9E,$0D,$AF,$EC,$73,$C4,$DD,$F8,$5B,$7D,$23,$17,$C9
	.dc.b	$B5,$CA,$BC,$64,$BE,$CF,$F4,$A8,$F4,$9B,$8C,$C7,$F6,$7B,$FA,$35
	.dc.b	$D2,$38,$1E,$5B,$F0,$4F,$8C,$91,$E6,$24,$69,$B7,$DE,$BA,$47,$73
	.dc.b	$ED,$1E,$77,$3C,$7F,$DB,$63,$C1,$3C,$6F,$BD,$F5,$24,$66,$7B,$9E
	.dc.b	$83,$A4,$F0,$8D,$4F,$09,$D6,$1F,$0A,$FC,$11,$C0,$E6,$79,$E7,$F1
	.dc.b	$98,$79,$B8,$FF,$B6,$4E,$E7,$A3,$BC,$3E,$7B,$DA,$B8,$3F,$DB,$E7
	.dc.b	$77,$E0,$C3,$C5,$87,$46,$D6,$F0,$F7,$3F,$EE,$43,$FD,$C7,$8F,$F5
	.dc.b	$97,$FA,$A7,$1F,$F5,$27,$FD,$C0,$7F,$55,$E3,$FE,$B6,$FF,$52,$3F
	.dc.b	$9E,$1F,$D4,$CF,$EA,$65,$E0,$C3,$8C,$FF,$A4,$9F,$ED,$97,$FD,$B9
	.dc.b	$C1,$FE,$D5,$7F,$DE,$DC,$7F,$ED,$A7,$9F,$17,$F4,$A7,$FA,$D3,$57
	.dc.b	$17,$F4,$8E,$4F,$EC,$8F,$F4,$7E,$8B,$C7,$FD,$1A,$BF,$BC,$1F,$D8
	.dc.b	$B8,$BF,$DA,$8F,$FB,$63,$88,$FF,$54,$1C,$5F,$DE,$DF,$ED,$2F,$CC
	.dc.b	$7F,$C2,$BA,$FD,$29,$FA,$1C,$7F,$5A,$73,$C7,$F8,$82,$3F,$6F,$8F
	.dc.b	$F1,$8D,$3F,$0F,$8F,$DB,$AE,$58,$7F,$A3,$67,$4F,$77,$EE,$18,$BE
	.dc.b	$AF,$E2,$4F,$D0,$99,$21,$93,$F2,$0F,$0A,$F0,$5F,$92,$75,$8F,$BE
	.dc.b	$FC,$51,$8C,$3C,$B8,$7F,$9B,$A3,$17,$C7,$0E,$71,$EC,$C7,$E2,$CE
	.dc.b	$5B,$19,$3D,$67,$23,$07,$CA,$3D,$33,$D3,$0F,$85,$19,$38,$38,$3C
	.dc.b	$F1,$BB,$F7,$57,$82,$75,$1E,$99,$D5,$76,$AD,$4F,$08,$F2,$CE,$9E
	.dc.b	$F1,$97,$C5,$19,$24,$F2,$8E,$87,$27,$0F,$23,$27,$27,$31,$CA,$38
	.dc.b	$10,$D1,$6B,$DF,$3C,$09,$19,$AF,$74,$62,$38,$92,$78,$BB,$3F,$D3
	.dc.b	$F1,$B8,$CC,$6E,$79,$BB,$3F,$D2,$07,$AC,$95,$E9,$78,$1D,$0E,$0F
	.dc.b	$FA,$39,$B5,$FE,$81,$AE,$E4,$AE,$F3,$C8,$EF,$5D,$67,$51,$D4,$D4
	.dc.b	$75,$6D,$75,$AE,$75,$C2,$79,$CF,$BE,$72,$48,$E9,$3B,$AE,$8D,$AE
	.dc.b	$93,$AA,$E0,$F4,$3D,$A7,$FA,$4B,$1F,$F0,$EB,$FD,$69,$FD,$74,$70
	.dc.b	$BE,$AD,$AE,$B5,$CC,$F5,$9C,$F1,$FF,$12,$4F,$F4,$88,$FA,$33,$EB
	.dc.b	$7C,$84,$39,$61,$E9,$79,$63,$FD,$E8,$FD,$B3,$E8,$3F,$D5,$44,$8F
	.dc.b	$F3,$FE,$3F,$C9,$5B,$82,$BF,$A6,$57,$F3,$03,$FA,$81,$FC,$B3,$8F
	.dc.b	$FA,$73,$FD,$43,$8F,$FA,$6E,$1F,$EA,$48,$FF,$4C,$1F,$E9,$FE,$3F
	.dc.b	$F5,$13,$FC,$D9,$FE,$9A,$7F,$38,$C7,$FA,$6E,$7F,$E9,$A9,$FE,$76
	.dc.b	$7F,$51,$7F,$D2,$0F,$F5,$01,$1F,$E9,$8D,$FF,$A5,$47,$FA,$43,$93
	.dc.b	$FD,$1E,$FF,$4E,$DF,$FA,$53,$FD,$A0,$FE,$95,$B7,$6E,$2F,$F5,$0F
	.dc.b	$07,$F4,$C9,$FE,$93,$C1,$F0,$DF,$E9,$2D,$9D,$D7,$1C,$FC,$F7,$C7
	.dc.b	$5F,$EA,$15,$FE,$08,$E0,$FA,$78,$FE,$00,$FA,$57,$FE,$93,$D7,$51
	.dc.b	$FC,$A1,$FB,$8E,$8F,$F3,$FE,$5F,$86,$7F,$A2,$9F,$86,$3E,$F4,$7F
	.dc.b	$A7,$E3,$51,$FD,$18,$FE,$2B,$1D,$0D,$57,$E1,$E3,$C1,$5F,$97,$7F
	.dc.b	$16,$7E,$76,$BF,$A3,$BF,$91,$C7,$F3,$AD,$CB,$F3,$B1,$88,$FC,$AD
	.dc.b	$73,$7D,$47,$FA,$38,$E9,$7F,$13,$0F,$24,$FC,$33,$F4,$55,$FC,$15
	.dc.b	$E5,$1E,$8A,$E2,$7A,$0F,$11,$E8,$7E,$59,$FD,$19,$C9,$E2,$CB,$F7
	.dc.b	$31,$A9,$FE,$84,$F8,$A7,$E7,$DF,$C5,$8E,$51,$EC,$39,$BE,$33,$93
	.dc.b	$F3,$A7,$CE,$7A,$DF,$F3,$F6,$5F,$88,$7F,$C2,$D8,$7E,$1F,$2F,$C6
	.dc.b	$57,$F8,$7A,$B5,$5F,$96,$3F,$C3,$98,$FE,$32,$7E,$2E,$5F,$C1,$67
	.dc.b	$C8,$4C,$3E,$91,$23,$EA,$92,$3E,$34,$F2,$7F,$12,$F9,$DF,$C4,$1F
	.dc.b	$8D,$3F,$31,$8F,$BE,$38,$46,$1E,$17,$FC,$8A,$3E,$2C,$63,$1F,$C4
	.dc.b	$C7,$96,$FC,$4F,$D3,$8F,$CB,$D7,$18,$D1,$E6,$3F,$3F,$87,$E0,$DE
	.dc.b	$A7,$D6,$3F,$07,$87,$D6,$8F,$81,$1C,$4F,$5B,$C5,$EA,$78,$72,$FB
	.dc.b	$D8,$FE,$08,$FB,$31,$D6,$BE,$EB,$F0,$DF,$C3,$3F,$9B,$73,$5D,$4F
	.dc.b	$0C,$75,$3C,$B3,$D5,$F4,$9C,$5F,$97,$3D,$2F,$F2,$B1,$83,$C7,$3B
	.dc.b	$3F,$16,$FF,$10,$C7,$F8,$94,$7F,$98,$CF,$C3,$E3,$F8,$C7,$E4,$C7
	.dc.b	$F8,$E1,$FE,$77,$5F,$E6,$56,$F0,$7F,$3C,$7F,$D1,$07,$FA,$28,$FF
	.dc.b	$36,$C7,$F3,$D7,$FD,$14,$7F,$3E,$3F,$D1,$67,$F3,$F3,$FD,$14,$33
	.dc.b	$FF,$40,$3F,$D0,$4C,$61,$FC,$F2,$3F,$9D,$69,$FC,$DA,$7F,$9C,$71
	.dc.b	$7F,$A1,$36,$FE,$77,$BA,$B5,$8F,$E1,$F6,$F9,$FB,$7E,$57,$1F,$79
	.dc.b	$F9,$C3,$ED,$67,$F2,$DF,$A3,$5F,$39,$D9,$F9,$B3,$66,$27,$EA,$6D
	.dc.b	$F9,$63,$F3,$77,$E0,$BF,$CF,$6F,$FC,$6D,$FA,$F8,$75,$C3,$F2,$3B
	.dc.b	$FE,$DE,$F8,$47,$63,$27,$9C,$FB,$C6,$0F,$21,$89,$F8,$13,$FA,$AC
	.dc.b	$7E,$04,$ED,$F3,$E1,$F9,$C8,$FE,$28,$7C,$0C,$7D,$51,$E7,$8C,$3B
	.dc.b	$7F,$3D,$8F,$AA,$77,$1F,$CB,$CF,$D4,$D7,$83,$0F,$1C,$7E,$BD,$AF
	.dc.b	$E8,$63,$F4,$B8,$7C,$13,$D1,$7F,$22,$7A,$63,$F1,$1F,$76,$1F,$3A
	.dc.b	$B7,$1F,$1F,$0E,$B8,$78,$6B,$53,$E1,$6B,$E1,$C3,$F0,$0F,$A1,$F0
	.dc.b	$4F,$F0,$7B,$EB,$3E,$0C,$7E,$23,$B3,$C1,$7D,$2B,$C1,$3A,$C7,$D1
	.dc.b	$8F,$DF,$3C,$B8,$F6,$3F,$80,$61,$CC,$EC,$7C,$0C,$7C,$33,$C7,$1E
	.dc.b	$D5,$C6,$BE,$C9,$C6,$75,$87,$86,$F8,$DF,$D6,$C7,$D9,$7F,$0E,$B4
	.dc.b	$D3,$C0,$58,$DB,$F1,$CF,$D0,$61,$F8,$19,$FC,$26,$3F,$C7,$70,$EF
	.dc.b	$18,$7E,$E5,$73,$AF,$7D,$7C,$4C,$3C,$91,$8A,$E3,$5F,$C9,$0F,$6D
	.dc.b	$6E,$FF,$C8,$95,$FE,$51,$98,$7F,$8E,$4F,$E2,$FC,$3F,$E5,$0E,$0F
	.dc.b	$E3,$11,$FC,$59,$AF,$F1,$62,$1F,$C4,$4F,$E2,$C7,$F9,$38,$FE,$1E
	.dc.b	$52,$99,$D5,$B7,$F1,$2D,$BF,$88,$E9,$FC,$B6,$99,$FF,$92,$4F,$EF
	.dc.b	$99,$FF,$32,$3E,$8C,$7F,$96,$33,$FF,$2E,$ED,$FE,$71,$7F,$C0,$07
	.dc.b	$F3,$B8,$FF,$02,$EB,$FC,$F1,$FF,$29,$D7,$37,$EC,$9F,$5C,$FE,$01
	.dc.b	$3F,$5D,$F9,$87,$EE,$8F,$DE,$98,$BF,$B4,$FD,$73,$E8,$D7,$D4,$DB
	.dc.b	$F6,$F1,$FC,$4F,$0F,$C7,$BE,$08,$FD,$0B,$6F,$F2,$CF,$73,$F2,$D8
	.dc.b	$FE,$01,$EA,$77,$35,$87,$DF,$7C,$98,$7E,$29,$F6,$61,$FA,$A7,$F6
	.dc.b	$38,$FC,$AC,$7C,$27,$E6,$0B,$8F,$E3,$67,$B6,$1E,$E7,$9E,$3F,$AC
	.dc.b	$79,$9E,$92,$D7,$3A,$F5,$D7,$18,$EB,$3E,$A3,$9D,$F8,$EB,$78,$FA
	.dc.b	$6B,$75,$EE,$8F,$99,$3F,$C4,$AF,$E0,$57,$CE,$AF,$E7,$B8,$FC,$E8
	.dc.b	$F1,$47,$92,$BD,$8F,$80,$F4,$1F,$C2,$A3,$6D,$8D,$C7,$8A,$BD,$51
	.dc.b	$E6,$9F,$B9,$8F,$D5,$7C,$C7,$C8,$C3,$D5,$3B,$8F,$2C,$EE,$B9,$C7
	.dc.b	$94,$F8,$97,$DE,$F8,$C6,$4E,$A7,$A6,$3C,$F8,$7B,$F0,$F3,$9D,$E3
	.dc.b	$FA,$CC,$FE,$10,$F3,$E3,$FC,$3D,$A6,$4F,$2B,$96,$99,$8E,$67,$CF
	.dc.b	$2C,$FD,$3B,$FA,$B3,$FC,$80,$FD,$3C,$EE,$7F,$58,$FF,$AD,$A3,$FC
	.dc.b	$53,$8F,$FB,$15,$FE,$26,$7F,$89,$CF,$F7,$36,$DF,$C5,$42,$F8,$07
	.dc.b	$95,$FE,$2E,$6C,$7F,$19,$9D,$C6,$23,$99,$C6,$7F,$91,$BF,$C1,$0F
	.dc.b	$E3,$19,$FF,$8E,$DC,$B5,$FE,$50,$62,$FD,$0D,$CF,$FD,$FE,$E8,$87
	.dc.b	$53,$DE,$E9,$C1,$C4,$F0,$C7,$47,$8C,$F8,$5D,$98,$8D,$38,$24,$6C
	.dc.b	$91,$A6,$CF,$85,$B3,$E0,$8C,$4E,$C8,$E0,$91,$B2,$1E,$C2,$9E,$B3
	.dc.b	$04,$30,$43,$D4,$F4,$74,$60,$E6,$53,$27,$BA,$3C,$AE,$8C,$47,$37
	.dc.b	$B1,$83,$53,$F0,$1E,$6D,$76,$D7,$C0,$62,$7B,$BC,$1A,$E6,$7C,$0F
	.dc.b	$02,$1E,$13,$83,$A7,$07,$13,$CB,$1D,$1E,$33,$E3,$76,$62,$34,$E0
	.dc.b	$91,$B2,$46,$9B,$3E,$16,$CF,$82,$31,$3B,$23,$82,$46,$C8,$7B,$0A
	.dc.b	$7A,$CC,$10,$C1,$0F,$53,$D1,$D1,$83,$99,$4C,$9E,$E8,$F2,$BA,$31
	.dc.b	$1C,$DE,$C6,$0D,$4F,$C0,$79,$B5,$DB,$5F,$01,$89,$EE,$F0,$6B,$99
	.dc.b	$F0,$3C,$08,$78,$4E,$0E,$9C,$1C,$4F,$2C,$74,$78,$CF,$8D,$D9,$88
	.dc.b	$D3,$82,$46,$C9,$1A,$6C,$F8,$5B,$3E,$08,$C4,$EC,$8E,$09,$1B,$21
	.dc.b	$EC,$29,$EB,$30,$43,$04,$3D,$4F,$47,$46,$0E,$65,$32,$7B,$A3,$CA
	.dc.b	$E8,$C4,$73,$7B,$18,$35,$3F,$01,$E6,$D7,$6D,$7C,$06,$27,$BB,$C1
	.dc.b	$AE,$67,$C0,$F0,$21,$E1,$38,$3A,$70,$71,$3C,$B1,$D1,$E3,$3E,$37
	.dc.b	$66,$23,$4E,$09,$1B,$24,$69,$B3,$E1,$6C,$F8,$23,$13,$B2,$38,$24
	.dc.b	$6C,$87,$B0,$A7,$AC,$C1,$0C,$10,$F5,$3D,$1D,$18,$3F,$9B,$F0,$7F
	.dc.b	$A4,$A8,$D2,$63,$FE,$05,$9F,$F0,$44,$EA,$31,$1F,$E9,$58,$C3,$A9
	.dc.b	$E2,$D9,$DB,$FE,$8C,$8C,$BC,$5A,$E6,$36,$FF,$83,$CE,$44,$38,$3A
	.dc.b	$7F,$C2,$26,$0D,$1C,$CA,$73,$34,$43,$45,$34,$60,$86,$08,$78,$0C
	.dc.b	$94,$CC,$FC,$47,$A1,$92,$39,$78,$B5,$E3,$3F,$11,$B3,$EF,$9D,$47
	.dc.b	$29,$D3,$67,$51,$89,$D9,$1C,$12,$36,$49,$F8,$E7,$32,$CF,$18,$D4
	.dc.b	$7B,$23,$0E,$A7,$A3,$67,$6F,$BA,$32,$F1,$6B,$98,$DB,$F0,$0E,$44
	.dc.b	$38,$3A,$7C,$C6,$0D,$1C,$CA,$73,$34,$43,$45,$34,$60,$86,$08,$78
	.dc.b	$0C,$94,$CC,$FC,$47,$A1,$92,$39,$78,$B5,$E3,$3F,$11,$B3,$EF,$9D
	.dc.b	$47,$29,$D3,$67,$51,$89,$D9,$1C,$12,$36,$49,$F8,$E7,$32,$CF,$18
	.dc.b	$D4,$7B,$23,$0E,$A7,$A3,$67,$6F,$BA,$32,$F1,$6B,$98,$DB,$F0,$0E
	.dc.b	$44,$38,$3A,$7C,$C6,$0D,$1C,$CA,$73,$34,$43,$45,$34,$60,$86,$08
	.dc.b	$78,$0C,$94,$CC,$FC,$47,$A1,$92,$39,$78,$B5,$E3,$3F,$11,$B3,$EF
	.dc.b	$9D,$47,$29,$D3,$67,$51,$89,$D9,$1C,$12,$36,$49,$F8,$E7,$32,$CF
	.dc.b	$18,$D4,$7B,$23,$0E,$A7,$A3,$67,$6F,$BA,$32,$F1,$6B,$98,$DB,$F0
	.dc.b	$0E,$44,$38,$3A,$7C,$C6,$0D,$1C,$CA,$73,$34,$43,$45,$34,$60,$86
	.dc.b	$08,$78,$0C,$94,$CC,$FC,$47,$A1,$92,$39,$78,$B5,$E3,$3F,$11,$B3
	.dc.b	$EF,$9D,$47,$29,$D3,$67,$FD,$F6,$FF,$5D,$FC,$54,$40

;--------------------------------------------------------------------------------
;	FSCALE.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fscale_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FSCALE.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSCALE.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fscale_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fscale.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSCALE.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fscale.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSCALE.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fscale_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3D,$61,$F3,$7F,$DA,$BD,$E7,$99,$C6,$30,$DA,$FE
	.dc.b	$50,$D9,$DC,$7F,$72,$F8,$17,$97,$FC,$F3,$E5,$7D,$72,$9C,$FC,$A7
	.dc.b	$F7,$89,$CC,$7F,$33,$F8,$3C,$79,$BF,$C3,$AA,$DF,$5C,$3A,$4E,$61
	.dc.b	$87,$2C,$7F,$5C,$F4,$AD,$CF,$1B,$E6,$F8,$E3,$2F,$96,$B8,$DF,$5B
	.dc.b	$E5,$7C,$31,$F4,$DF,$82,$BF,$92,$63,$9C,$79,$D6,$B0,$E3,$7B,$3E
	.dc.b	$36,$3D,$9F,$21,$93,$C9,$8F,$33,$C1,$1B,$8C,$9B,$AE,$D8,$F9,$CF
	.dc.b	$1C,$67,$0F,$19,$2F,$9E,$3D,$67,$51,$71,$F9,$53,$6F,$A5,$79,$A7
	.dc.b	$2F,$02,$3E,$E3,$25,$8E,$31,$E8,$8C,$C6,$30,$F4,$CF,$58,$F4,$CF
	.dc.b	$18,$C1,$0F,$AD,$5E,$57,$15,$C9,$C9,$E4,$AE,$D8,$7D,$4C,$7E,$55
	.dc.b	$6C,$91,$9B,$E9,$7E,$7B,$F3,$5F,$0A,$F9,$53,$D4,$FB,$75,$C2,$F9
	.dc.b	$E3,$F3,$A7,$47,$33,$CF,$7C,$E3,$85,$F8,$67,$75,$C9,$FB,$D5,$B7
	.dc.b	$33,$BC,$7E,$34,$75,$25,$7E,$AD,$DD,$F9,$6F,$95,$70,$AF,$85,$18
	.dc.b	$C7,$E3,$D7,$A6,$B5,$87,$C0,$7E,$CE,$1F,$42,$31,$5F,$60,$E3,$1E
	.dc.b	$AB,$E8,$EE,$FA,$4F,$E1,$3E,$43,$F8,$7E,$5F,$E7,$8A,$FF,$1A,$E7
	.dc.b	$FC,$C5,$FF,$19,$C6,$54,$F0,$78,$2F,$47,$84,$7F,$8F,$9A,$F0,$6C
	.dc.b	$6E,$7A,$7C,$8B,$E1,$F6,$D8,$CC,$60,$CF,$C3,$C8,$E2,$DC,$3C,$71
	.dc.b	$D8,$87,$A6,$FF,$C5,$8D,$F8,$7F,$19,$1C,$1B,$D7,$CE,$47,$07,$3D
	.dc.b	$BD,$64,$70,$77,$F8,$7F,$79,$1C,$1E,$3F,$1B,$E1,$91,$C1,$E7,$F4
	.dc.b	$3E,$39,$1C,$1E,$BF,$87,$F9,$64,$70,$7B,$FE,$1F,$E7,$91,$C1,$F0
	.dc.b	$FD,$9F,$A6,$47,$07,$C7,$E2,$FA,$E4,$70,$7C,$BE,$2F,$B6,$47,$07
	.dc.b	$D8,$26,$FF,$8D,$E2,$FB,$8D,$E2,$FA,$0F,$E2,$B8,$2B,$FE,$CA,$E1
	.dc.b	$FF,$4F,$5F,$D0,$8F,$C7,$F0,$FD,$36,$C7,$C4,$7F,$27,$B7,$C7,$6C
	.dc.b	$7C,$47,$F2,$FB,$7C,$76,$C7,$B1,$FC,$DE,$9E,$F6,$C7,$B1,$FC,$FE
	.dc.b	$9E,$F6,$C7,$E3,$7C,$4F,$CE,$91,$C1,$F9,$1F,$0F,$F4,$24,$70,$7E
	.dc.b	$57,$E1,$FF,$4A,$47,$07,$E6,$7D,$AF,$D4,$91,$C1,$FB,$5D,$3F,$D4
	.dc.b	$4D,$F9,$1F,$D8,$C6,$31,$FD,$F3,$5F,$F7,$BF,$C8,$FF,$BC,$6B,$F7
	.dc.b	$CD,$8F,$E0,$8D,$77,$E1,$7E,$F2,$3F,$D5,$93,$C3,$E1,$F0,$47,$9F
	.dc.b	$81,$B2,$38,$3B,$FC,$8F,$32,$38,$3C,$7E,$7F,$72,$38,$3C,$FE,$E7
	.dc.b	$8C,$8E,$0F,$E3,$1F,$21,$FE,$BB,$6F,$C8,$FF,$14,$8C,$69,$FC,$85
	.dc.b	$A8,$FF,$60,$CF,$28,$E0,$47,$AF,$C8,$F3,$23,$83,$C3,$F2,$3D,$C8
	.dc.b	$E0,$F2,$FC,$8F,$8C,$8E,$0F,$4F,$C8,$F9,$C8,$E0,$F6,$FC,$57,$AC
	.dc.b	$8E,$0F,$83,$F1,$5E,$F2,$38,$3E,$2F,$BF,$F0,$C8,$E0,$F9,$3F,$23
	.dc.b	$F1,$C8,$E0,$F9,$BF,$17,$F2,$C8,$E0,$FA,$3F,$17,$FE,$B1,$7E,$66
	.dc.b	$DF,$37,$64,$E6,$3F,$DC,$B3,$F4,$4D,$A9,$DC,$66,$3F,$DD,$4D,$7A
	.dc.b	$7C,$5F,$36,$C6,$5E,$51,$97,$A7,$89,$6B,$F8,$AD,$78,$1B,$3F,$EF
	.dc.b	$68,$E5,$0F,$2F,$81,$23,$2F,$5F,$3F,$FA,$7D,$7F,$6B,$6C,$FB,$1B
	.dc.b	$1F,$F0,$02,$38,$78,$D1,$F4,$4F,$F2,$29,$E5,$3D,$7B,$25,$69,$B3
	.dc.b	$DF,$CC,$79,$DF,$DE,$70,$FF,$34,$CF,$86,$72,$FF,$80,$63,$2F,$D6
	.dc.b	$5E,$6B,$C0,$E4,$D9,$E0,$D7,$D8,$D7,$97,$9B,$F8,$0C,$BE,$03,$63
	.dc.b	$27,$B9,$C3,$FE,$9B,$F2,$6B,$FE,$99,$6E,$1F,$EE,$28,$F8,$B1,$8F
	.dc.b	$26,$47,$C8,$6B,$F4,$7C,$DE,$24,$7E,$6B,$D3,$CD,$91,$F4,$1A,$FD
	.dc.b	$9F,$2F,$C2,$47,$EA,$BE,$4F,$2E,$47,$D8,$6B,$F8,$1F,$2F,$FC,$54
	.dc.b	$FF,$AC,$DB,$97,$FB,$76,$3E,$B4,$63,$DF,$91,$F6,$5A,$FD,$7F,$6B
	.dc.b	$89,$1F,$BC,$F4,$F3,$64,$7E,$05,$AF,$DF,$F3,$7C,$24,$7F,$0C,$F9
	.dc.b	$3E,$46,$47,$E2,$5A,$FE,$1F,$DB,$F4,$91,$FC,$73,$EE,$F9,$1F,$F6
	.dc.b	$8C,$7E,$E9,$AF,$F0,$26,$BF,$CB,$DA,$9F,$F7,$8B,$FE,$55,$6B,$CF
	.dc.b	$E0,$7B,$1D,$FD,$09,$1F,$E5,$79,$46,$23,$97,$9F,$E5,$23,$B7,$D1
	.dc.b	$F0,$F2,$39,$35,$F1,$7C,$5F,$B4,$8F,$77,$E0,$7C,$1C,$8F,$13,$5F
	.dc.b	$57,$C3,$7C,$52,$3E,$77,$F0,$BF,$22,$F4,$29,$27,$FC,$E5,$19,$6F
	.dc.b	$89,$C0,$8E,$0F,$F3,$C7,$C8,$3E,$47,$F2,$2D,$8C,$C6,$0C,$FC,$8C
	.dc.b	$8E,$2D,$7F,$D0,$D2,$F3,$F9,$92,$3B,$BD,$BC,$F9,$1E,$26,$BC,$BC
	.dc.b	$DF,$19,$1F,$3B,$E6,$F3,$64,$7A,$9A,$F8,$BC,$DF,$59,$1F,$7B,$FA
	.dc.b	$EF,$91,$FF,$48,$C1,$A4,$F1,$B5,$E6,$49,$FF,$4F,$4A,$31,$1F,$EA
	.dc.b	$58,$78,$FD,$49,$1C,$1F,$07,$8F,$23,$9B,$5E,$5E,$2F,$94,$8F,$81
	.dc.b	$FE,$0B,$E5,$3D,$5A,$49,$FF,$59,$46,$5B,$F2,$1C,$08,$E0,$FF,$5C
	.dc.b	$7D,$5B,$D6,$FE,$22,$D8,$CC,$60,$CF,$C2,$C8,$E2,$D7,$FD,$8D,$2E
	.dc.b	$2E,$64,$8E,$EF,$6F,$56,$47,$89,$AF,$FB,$3F,$EB,$3E,$73,$FE,$CD
	.dc.b	$6B,$9E,$0F,$E4,$F1,$88,$FF,$01,$AF,$4F,$24,$8E,$4F,$0D,$24,$75
	.dc.b	$6B,$E1,$F3,$7A,$91,$F1,$3E,$0F,$32,$47,$95,$AF,$A7,$DE,$F2,$91
	.dc.b	$F5,$3F,$CF,$BE,$B7,$FB,$14,$7F,$88,$1A,$FF,$93,$5A,$9A,$3E,$1B
	.dc.b	$91,$23,$FD,$E5,$19,$7F,$DE,$9E,$4B,$CE,$FE,$9A,$D8,$CC,$60,$CF
	.dc.b	$9D,$23,$8B,$5F,$F1,$78,$F8,$7F,$17,$CF,$7F,$98,$FE,$85,$CB,$FE
	.dc.b	$3A,$97,$C5,$7F,$8F,$27,$37,$FE,$69,$9C,$BC,$6B,$86,$8C,$31,$3B
	.dc.b	$8C,$2B,$E6,$F8,$F0,$ED,$7E,$16,$C7,$1B,$E6,$FA,$6F,$9D,$F2,$C3
	.dc.b	$C7,$3D,$70,$E4,$F0,$8F,$7B,$93,$A1,$4F,$11,$BC,$3C,$2F,$7C,$39
	.dc.b	$61,$E0,$BE,$B3,$C6,$FC,$38,$78,$EB,$53,$DA,$3C,$D8,$74,$AD,$DF
	.dc.b	$C8,$C3,$A9,$E1,$B8,$75,$C3,$9B,$E3,$3C,$17,$CC,$E9,$1A,$BF,$7E
	.dc.b	$1E,$0A,$E6,$F9,$6F,$BC,$72,$33,$87,$8D,$F0,$46,$1F,$C0,$61,$DA
	.dc.b	$38,$D7,$59,$E4,$70,$23,$92,$57,$57,$C5,$1C,$1C,$BF,$1E,$34,$7C
	.dc.b	$E8,$CC,$7C,$17,$89,$8B,$F9,$71,$C4,$87,$DC,$3B,$4F,$47,$31,$C1
	.dc.b	$D4,$F4,$9D,$46,$1F,$11,$89,$F9,$AE,$23,$E2,$9F,$11,$D5,$FD,$58
	.dc.b	$C1,$E0,$70,$FC,$68,$DB,$F3,$CF,$73,$87,$D2,$7A,$30,$FB,$D3,$CA
	.dc.b	$FD,$13,$D4,$E5,$7E,$67,$07,$23,$A9,$88,$F9,$55,$F8,$93,$AC,$72
	.dc.b	$71,$87,$4B,$FC,$14,$FA,$4F,$82,$75,$BF,$03,$D8,$F9,$46,$AF,$CE
	.dc.b	$FF,$13,$F8,$37,$C4,$FF,$BD,$78,$DE,$BF,$F1,$39,$CD,$7F,$45,$8C
	.dc.b	$56,$DB,$19,$3F,$8B,$4F,$07,$FD,$9F,$82,$31,$1C,$0C,$B8,$78,$D7
	.dc.b	$56,$9D,$4F,$43,$96,$99,$03,$E2,$7F,$D0,$AD,$3F,$D0,$EE,$08,$E0
	.dc.b	$87,$F3,$83,$81,$5D,$B4,$D9,$FC,$F9,$CB,$4C,$9F,$EF,$07,$2D,$32
	.dc.b	$7F,$BE,$9C,$B4,$C9,$EC,$72,$D3,$27,$98,$F0,$1E,$52,$18,$21,$E1
	.dc.b	$32,$53,$27,$84,$C9,$4C,$9E,$13,$25,$32,$78,$CE,$84,$39,$14,$E0
	.dc.b	$53,$81,$A2,$1A,$29,$A3,$04,$30,$47,$E3,$B4,$F8,$EE,$08,$E0,$87
	.dc.b	$B0,$E0,$57,$6D,$36,$7C,$07,$2D,$32,$7B,$5C,$B4,$C9,$EC,$72,$D3
	.dc.b	$27,$B5,$CB,$4C,$9E,$63,$C0,$79,$48,$60,$87,$84,$C9,$4C,$9E,$13
	.dc.b	$25,$32,$78,$4C,$94,$C9,$E3,$3A,$10,$E4,$53,$81,$4E,$06,$88,$68
	.dc.b	$A6,$8C,$10,$C1,$1F,$8E,$D3,$E3,$B8,$23,$82,$1E,$C3,$81,$5D,$B4
	.dc.b	$D9,$F0,$1C,$B4,$C9,$ED,$72,$D3,$27,$B1,$CB,$4C,$9E,$D7,$2D,$32
	.dc.b	$79,$8F,$01,$E5,$21,$82,$1E,$13,$25,$32,$78,$4C,$94,$C9,$E1,$32
	.dc.b	$53,$27,$8C,$E8,$43,$91,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04
	.dc.b	$7E,$3B,$4F,$8E,$E0,$8E,$08,$7B,$0E,$05,$76,$D3,$67,$C0,$72,$D3
	.dc.b	$27,$B5,$CB,$4C,$9E,$C7,$2D,$32,$7B,$5C,$B4,$C9,$E6,$3C,$07,$94
	.dc.b	$86,$08,$78,$4C,$94,$C9,$E1,$32,$53,$27,$84,$C9,$4C,$9E,$33,$A1
	.dc.b	$0E,$45,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$11,$F8,$ED,$3E,$3B
	.dc.b	$82,$38,$21,$EC,$38,$1F,$E1,$68,$FF,$08,$57,$F8,$6D,$B5,$FE,$26
	.dc.b	$C3,$FC,$53,$5F,$E1,$86,$BF,$E1,$63,$91,$FE,$C4,$AE,$4D,$9F,$F1
	.dc.b	$13,$FE,$1C,$9E,$33,$A6,$CF,$7C,$3F,$C6,$31,$DB,$D8,$73,$3D,$DF
	.dc.b	$F0,$47,$C4,$FF,$89,$32,$56,$A7,$31,$FE,$28,$85,$BC,$5C,$CF,$0C
	.dc.b	$3A,$3F,$E2,$DC,$3A,$5F,$38,$C1,$CE,$77,$AA,$B7,$87,$5A,$D9,$D7
	.dc.b	$1E,$F8,$73,$BE,$13,$CE,$73,$5C,$67,$33,$C4,$E7,$5C,$6F,$CB,$87
	.dc.b	$89,$F0,$E3,$DA,$31,$5D,$4E,$2E,$1F,$8F,$87,$7A,$E7,$8F,$12,$4F
	.dc.b	$82,$74,$F3,$C7,$CB,$8F,$8E,$31,$7E,$27,$BD,$F2,$AD,$E1,$E1,$BF
	.dc.b	$A7,$8E,$70,$F0,$E1,$C0,$F0,$46,$5E,$A4,$3D,$2E,$5F,$8D,$7D,$B0
	.dc.b	$F9,$73,$9B,$F3,$56,$AF,$CD,$7B,$8F,$03,$4F,$6E,$3F,$02,$38,$1E
	.dc.b	$1A,$E6,$7A,$9E,$35,$EA,$AE,$4F,$63,$04,$8E,$55,$D6,$B9,$E3,$F1
	.dc.b	$4F,$8D,$19,$71,$8F,$AE,$30,$FB,$63,$78,$D8,$D3,$6F,$CA,$7E,$8E
	.dc.b	$B8,$B8,$3D,$A7,$CC,$31,$7E,$4A,$F9,$71,$8B,$F1,$BF,$58,$C4,$78
	.dc.b	$1A,$72,$AE,$F8,$FC,$48,$F8,$38,$79,$63,$11,$E1,$6E,$1E,$18,$ED
	.dc.b	$18,$C3,$C2,$FD,$6A,$E5,$8F,$CD,$AE,$4D,$BE,$55,$B9,$F8,$7F,$3E
	.dc.b	$FD,$21,$E9,$FE,$E6,$7F,$86,$4E,$63,$11,$FE,$B2,$8C,$63,$FC,$46
	.dc.b	$31,$5C,$A3,$10,$9D,$BB,$6B,$DA,$78,$39,$6B,$CB,$E1,$BC,$2F,$F1
	.dc.b	$F6,$BF,$EC,$46,$A6,$92,$BF,$D8,$11,$96,$C6,$63,$FD,$88,$D7,$C1
	.dc.b	$1A,$23,$A6,$BA,$30,$43,$04,$3E,$34,$75,$6C,$72,$69,$C4,$A7,$13
	.dc.b	$E0,$3B,$69,$B2,$9B,$3E,$23,$B6,$9B,$29,$B3,$E2,$19,$29,$98,$F4
	.dc.b	$3D,$8F,$29,$0C,$10,$EA,$64,$A6,$4F,$EB,$51,$F5,$35,$FB,$24,$70
	.dc.b	$48,$FF,$7C,$46,$5B,$19,$7F,$1F,$19,$6C,$65,$FC,$B4,$65,$B1,$97
	.dc.b	$FB,$54,$65,$B1,$94,$7A,$0F,$09,$0E,$E5,$3A,$94,$E6,$53,$9C,$77
	.dc.b	$6C,$77,$7B,$1D,$08,$74,$29,$D0,$A7,$23,$B1,$C8,$E0,$43,$81,$4E
	.dc.b	$07,$23,$81,$A2,$1A,$38,$1A,$30,$68,$C1,$23,$DC,$D8,$F7,$46,$08
	.dc.b	$E0,$91,$EA,$8C,$B6,$32,$F9,$E3,$2D,$8C,$BE,$68,$CB,$63,$2F,$9E
	.dc.b	$32,$D8,$CB,$EC,$21,$E9,$3C,$44,$3C,$05,$3B,$14,$E8,$53,$9C,$78
	.dc.b	$1B,$1E,$07,$B1,$D0,$87,$42,$9D,$0A,$72,$3B,$1C,$8E,$04,$38,$14
	.dc.b	$E0,$72,$38,$1A,$21,$A3,$81,$A3,$06,$8C,$12,$3D,$CD,$8F,$74,$60
	.dc.b	$8E,$09,$1E,$A8,$CB,$63,$2F,$9E,$32,$D8,$CB,$E6,$8C,$B6,$32,$F9
	.dc.b	$E3,$2D,$8C,$BE,$C2,$1E,$93,$C4,$43,$C0,$53,$B1,$4E,$85,$39,$C7
	.dc.b	$81,$B1,$FD,$59,$63,$FC,$43,$89,$CD,$FC,$31,$B3,$FD,$C6,$75,$39
	.dc.b	$3F,$B8,$B8,$8F,$E2,$33,$C1,$FF,$44,$46,$5F,$E1,$11,$97,$F8,$FC
	.dc.b	$65,$FE,$41,$19,$7C,$A1,$E1,$7F,$E8,$17,$53,$C6,$31,$5E,$4A,$E0
	.dc.b	$F9,$DF,$51,$83,$CE,$EE,$BA,$DF,$8E,$7C,$31,$87,$81,$EC,$29,$E8
	.dc.b	$9E,$47,$77,$51,$C4,$F3,$BC,$2B,$CE,$F9,$23,$53,$CA,$3C,$A6,$5E
	.dc.b	$8F,$C8,$39,$57,$9E,$FE,$13,$63,$27,$98,$F5,$CF,$26,$9D,$8F,$3B
	.dc.b	$99,$E7,$1C,$5B,$5E,$A3,$CF,$1E,$53,$91,$E5,$72,$F4,$32,$78,$0F
	.dc.b	$C1,$CF,$BD,$A7,$32,$9C,$4A,$75,$7C,$CF,$BA,$3C,$91,$87,$C3,$1B
	.dc.b	$3C,$33,$EA,$9F,$04,$7D,$93,$C0,$53,$C0,$53,$91,$E2,$3D,$46,$4B
	.dc.b	$7F,$71,$E2,$D7,$89,$B9,$F1,$3E,$08,$D3,$CA,$FE,$D5,$FA,$A3,$07
	.dc.b	$D3,$3E,$05,$FA,$AB,$8D,$F4,$9F,$0D,$F2,$AF,$88,$48,$F8,$67,$23
	.dc.b	$C5,$1B,$73,$7E,$1A,$E9,$1E,$AB,$EE,$F1,$3B,$9E,$59,$E0,$E4,$E1
	.dc.b	$5D,$5F,$95,$5D,$1F,$6C,$FD,$A7,$C3,$7E,$77,$67,$33,$6F,$C1,$AE
	.dc.b	$91,$F3,$4D,$57,$52,$38,$25,$79,$89,$5D,$E3,$94,$61,$F1,$90,$F1
	.dc.b	$B9,$7F,$2F,$C4,$E2,$71,$FF,$8F,$38,$9C,$7F,$E3,$EB,$5E,$EB,$53
	.dc.b	$98,$C4,$AB,$8B,$FE,$46,$8C,$C6,$23,$9D,$F8,$1F,$F5,$2C,$F2,$7F
	.dc.b	$D5,$4D,$7B,$46,$DC,$BC,$67,$A4,$F6,$AD,$DF,$9A,$FC,$57,$B9,$FF
	.dc.b	$61,$71,$FF,$B9,$A7,$FD,$AB,$9F,$FB,$3B,$1F,$F6,$AF,$1E,$B6,$FF
	.dc.b	$6F,$CF,$FB,$3D,$FF,$6E,$36,$BF,$DB,$9B,$7F,$B4,$1F,$F6,$D9,$D2
	.dc.b	$BF,$DA,$17,$D2,$3F,$CD,$3C,$7F,$ED,$29,$F9,$CF,$C9,$AE,$53,$FE
	.dc.b	$F2,$8C,$3C,$CE,$B7,$E1,$3D,$13,$C2,$58,$FB,$67,$57,$E2,$AD,$4E
	.dc.b	$63,$DD,$7E,$A3,$B1,$E1,$C3,$D0,$F8,$E7,$83,$EA,$7F,$DE,$33,$D6
	.dc.b	$38,$E1,$E2,$C7,$E4,$CF,$6C,$7C,$F5,$AB,$E6,$F5,$AF,$9B,$5B,$3A
	.dc.b	$57,$AA,$7E,$0C,$65,$A7,$8E,$7E,$16,$1E,$8C,$3E,$74,$FD,$08,$C6
	.dc.b	$1E,$88,$C6,$3F,$12,$FC,$F1,$CC,$F5,$D7,$08,$F2,$C6,$71,$F2,$2C
	.dc.b	$3E,$BC,$31,$FB,$B8,$AF,$FC,$17,$1F,$EA,$AE,$27,$13,$C4,$FF,$7E
	.dc.b	$C6,$1B,$19,$D3,$F6,$99,$F1,$8C,$38,$AE,$13,$A8,$46,$1D,$4F,$07
	.dc.b	$F7,$D1,$97,$87,$88,$AE,$51,$FE,$C0,$8C,$38,$6B,$C2,$36,$A3,$2F
	.dc.b	$A1,$AF,$F0,$D6,$BD,$63,$8C,$62,$39,$46,$08,$F8,$5A,$F8,$4C,$10
	.dc.b	$C1,$0F,$13,$C1,$A6,$CA,$6C,$F3,$3B,$69,$B2,$9B,$3D,$46,$4A,$64
	.dc.b	$F5,$99,$29,$98,$F8,$04,$8F,$04,$72,$3C,$A4,$3C,$25,$34,$60,$86
	.dc.b	$08,$78,$0E,$05,$3F,$A0,$7A,$7F,$44,$91,$F7,$1B,$1F,$84,$6B,$F8
	.dc.b	$F6,$BF,$96,$6B,$D2,$39,$11,$E4,$D7,$93,$5E,$4D,$79,$1C,$08,$70
	.dc.b	$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$80,$47,$B9,$C3,$5C,$C7
	.dc.b	$A5,$CB,$5C,$9E,$63,$25,$32,$79,$4C,$94,$C9,$E6,$32,$53,$27,$A0
	.dc.b	$A6,$8C,$47,$9C,$91,$E5,$6C,$78,$DA,$F8,$5A,$F7,$6B,$D2,$39,$11
	.dc.b	$E4,$D7,$93,$5E,$4D,$79,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53
	.dc.b	$46,$08,$60,$87,$C7,$21,$F2,$89,$1F,$01,$CB,$5C,$C7,$AD,$CB,$5C
	.dc.b	$9E,$83,$25,$32,$79,$CC,$94,$C9,$E8,$32,$53,$27,$A8,$A6,$8C,$47
	.dc.b	$9C,$91,$E5,$6C,$78,$DA,$F8,$5A,$F7,$6B,$D2,$39,$11,$E4,$D7,$93
	.dc.b	$5E,$4D,$79,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60
	.dc.b	$F9,$07,$F9,$E8,$FF,$48,$9F,$E8,$9E,$4F,$F3,$96,$7F,$E8,$EE,$4F
	.dc.b	$F3,$74,$FF,$9D,$27,$35,$FE,$80,$AD,$46,$27,$91,$FE,$77,$38,$16
	.dc.b	$36,$D9,$E4,$47,$89,$FE,$7E,$9E,$95,$FE,$95,$8C,$1C,$4D,$9D,$8C
	.dc.b	$90,$C1,$2B,$C2,$E5,$AE,$6B,$C5,$3D,$E3,$07,$69,$D9,$27,$75,$FE
	.dc.b	$A3,$9D,$46,$1F,$9B,$5C,$EB,$84,$68,$91,$99,$E1,$1F,$EB,$C8,$C3
	.dc.b	$AA,$ED,$18,$23,$82,$47,$81,$AF,$36,$BF,$EC,$43,$A1,$B2,$1B,$29
	.dc.b	$B8,$EE,$E0,$FC,$51,$B2,$1B,$8E,$EE,$8C,$1C,$76,$FF,$62,$CF,$E3
	.dc.b	$AF,$E2,$BD,$A7,$99,$C4,$93,$C2,$BF,$0A,$76,$71,$1F,$85,$94,$FF
	.dc.b	$AA,$A3,$FD,$79,$C3,$78,$7F,$D5,$FC,$5F,$EA,$B9,$FF,$56,$F1,$FF
	.dc.b	$AB,$DF,$F5,$EB,$FE,$48,$E4,$FF,$5D,$46,$8F,$F5,$FC,$FF,$AE,$DE
	.dc.b	$97,$1D,$5F,$6B,$E9,$C9,$FE,$7A,$9F,$F3,$74,$FF,$9C,$DF,$F4,$C6
	.dc.b	$7F,$E7,$29,$FF,$48,$42,$3F,$CE,$4F,$F9,$F8,$FF,$48,$69,$FE,$7E
	.dc.b	$7F,$D2,$07,$2E,$3F,$F4,$63,$FE,$A1,$DB,$FD,$08,$F3,$8F,$F5,$BE
	.dc.b	$1E,$1B,$FF,$4D,$9E,$1E,$4F,$F4,$DE,$3E,$73,$C7,$1E,$2B,$F4,$E1
	.dc.b	$D2,$31,$3F,$E8,$8C,$3C,$2F,$07,$BC,$ED,$B5,$EA,$3C,$87,$B6,$BC
	.dc.b	$58,$FB,$6F,$C5,$87,$90,$FF,$50,$57,$B4,$E5,$5E,$2C,$7D,$98,$7F
	.dc.b	$AC,$0F,$1D,$71,$C3,$FD,$5B,$8F,$8F,$0F,$05,$6A,$7C,$15,$DB,$1F
	.dc.b	$55,$68,$FF,$54,$57,$BB,$0E,$F1,$FE,$BA,$3E,$9E,$3F,$32,$7A,$47
	.dc.b	$B8,$F9,$27,$A1,$F8,$71,$F0,$23,$55,$F2,$DE,$E4,$3E,$04,$6C,$F8
	.dc.b	$18,$FF,$AD,$23,$7F,$02,$FD,$37,$FC,$52,$DE,$2F,$F4,$CB,$6B,$FD
	.dc.b	$5C,$DA,$FF,$5A,$B5,$FF,$23,$07,$0F,$F9,$25,$AE,$6B,$FC,$6F,$C9
	.dc.b	$FE,$C4,$F2,$3F,$D8,$24,$BF,$F5,$E4,$65,$BC,$C9,$D5,$FF,$AF,$9A
	.dc.b	$FF,$AF,$A3,$53,$C2,$73,$3F,$EC,$B6,$BF,$EC,$89,$DB,$CE,$78,$11
	.dc.b	$C1,$0E,$D3,$CA,$BC,$D1,$93,$B4,$F1,$76,$D7,$26,$CF,$14,$F3,$AF
	.dc.b	$1C,$70,$35,$5D,$9C,$10,$E6,$53,$9D,$77,$69,$CA,$7A,$BC,$08,$68
	.dc.b	$E5,$3D,$DC,$B5,$D1,$88,$E6,$E0,$87,$FA,$27,$D4,$FF,$6F,$12,$BF
	.dc.b	$DD,$2D,$8F,$F7,$8B,$5F,$F7,$E3,$5F,$C6,$87,$0F,$E4,$1A,$F7,$8E
	.dc.b	$A4,$7A,$B5,$EA,$D7,$AC,$72,$7B,$46,$5E,$CD,$7B,$1D,$08,$74,$29
	.dc.b	$D1,$E0,$F4,$32,$74,$29,$D0,$E4,$43,$91,$A3,$91,$93,$91,$4E,$47
	.dc.b	$03,$07,$03,$04,$38,$94,$E2,$60,$86,$08,$73,$32,$53,$31,$F2,$C9
	.dc.b	$1F,$09,$B1,$EB,$6B,$E6,$6B,$E3,$3A,$11,$CB,$E0,$6B,$E0,$8E,$C4
	.dc.b	$7B,$35,$EC,$D7,$B4,$72,$7B,$46,$5E,$CD,$7B,$1D,$08,$74,$29,$D1
	.dc.b	$E0,$F4,$32,$74,$29,$D0,$E4,$43,$91,$A3,$91,$93,$91,$4E,$47,$03
	.dc.b	$07,$03,$04,$38,$94,$E2,$60,$86,$08,$73,$32,$53,$31,$F2,$C9,$1F
	.dc.b	$09,$B1,$EB,$6B,$E6,$6B,$E3,$3A,$11,$CB,$E0,$6B,$E0,$8E,$C4,$7B
	.dc.b	$35,$EC,$D7,$B4,$72,$7B,$46,$5E,$CD,$7B,$1D,$08,$74,$29,$D1,$E0
	.dc.b	$F4,$3E,$D1,$FE,$58,$1C,$DF,$E5,$F8,$FF,$3F,$C7,$F9,$90,$FF,$2F
	.dc.b	$E9,$FE,$5A,$7F,$D2,$D1,$A7,$94,$F1,$7F,$CF,$27,$03,$AF,$37,$F9
	.dc.b	$91,$FF,$37,$18,$3F,$CC,$8F,$F9,$9D,$C1,$D8,$E6,$78,$C8,$78,$1E
	.dc.b	$8F,$78,$F1,$1F,$E7,$77,$75,$E9,$7F,$D5,$31,$97,$91,$E2,$3C,$AF
	.dc.b	$29,$F2,$36,$33,$1E,$07,$A9,$E8,$3A,$BC,$0D,$3D,$8F,$64,$78,$67
	.dc.b	$C9,$C5,$FE,$8E,$AE,$15,$E3,$3D,$8F,$29,$DD,$FD,$16,$CF,$C6,$3B
	.dc.b	$1E,$23,$CE,$73,$3C,$F7,$EB,$7A,$99,$38,$DF,$8D,$E2,$7A,$2F,$CA
	.dc.b	$D3,$FD,$17,$C9,$FE,$8D,$7F,$D2,$A7,$FA,$70,$C9,$FE,$95,$3F,$B5
	.dc.b	$F3,$7F,$A6,$1D,$98,$E6,$FF,$4C,$3C,$0E,$7C,$9F,$E9,$77,$A1,$8A
	.dc.b	$FF,$2F,$67,$FE,$67,$8F,$F2,$CE,$7F,$E5,$1E,$4F,$F2,$C9,$FE,$5E
	.dc.b	$7F,$CA,$71,$FE,$5D,$3F,$CD,$07,$F9,$92,$3F,$CE,$8F,$F9,$6F,$93
	.dc.b	$FD,$1D,$87,$F9,$AA,$3F,$CB,$B3,$CC,$EA,$EF,$97,$FC,$E9,$3B,$AF
	.dc.b	$F3,$AC,$F9,$2B,$83,$E0,$8F,$F4,$4E,$BF,$E7,$C3,$FD,$03,$1E,$53
	.dc.b	$81,$D5,$FF,$35,$3E,$23,$07,$FA,$B4,$B8,$7F,$9F,$5B,$5E,$C7,$DC
	.dc.b	$7F,$A0,$EB,$C2,$FF,$9F,$2F,$D1,$87,$FA,$4E,$3F,$CF,$EF,$B3,$1F
	.dc.b	$9D,$5F,$E8,$19,$E6,$78,$CA,$FC,$03,$47,$6C,$3E,$89,$F5,$63,$E1
	.dc.b	$57,$FA,$6C,$F0,$47,$FB,$3E,$FC,$78,$FC,$5A,$F1,$BF,$0A,$FB,$4B
	.dc.b	$0C,$60,$AF,$F4,$87,$23,$F5,$0D,$57,$FA,$31,$B8,$FF,$A3,$1A,$FF
	.dc.b	$A3,$27,$84,$FF,$A5,$27,$33,$FE,$8D,$6B,$FE,$91,$6C,$F0,$9F,$F4
	.dc.b	$4F,$5F,$96,$BF,$C9,$7C,$8E,$44,$E6,$7F,$D1,$4D,$47,$FA,$1D,$C3
	.dc.b	$6B,$A9,$1E,$B3,$D2,$31,$1D,$1A,$F4,$AF,$15,$74,$6C,$EA,$7B,$1D
	.dc.b	$27,$55,$D6,$31,$3B,$AE,$31,$82,$1E,$12,$9E,$19,$E4,$E4,$86,$6B
	.dc.b	$B9,$2B,$B4,$F0,$9D,$46,$27,$51,$D8,$F4,$9D,$0A,$70,$35,$5E,$87
	.dc.b	$4D,$74,$62,$BC,$0F,$22,$46,$DB,$1A,$8C,$11,$E3,$1F,$73,$AF,$ED
	.dc.b	$B6,$FF,$D1,$EF,$FB,$72,$32,$9A,$48,$DC,$71,$2C,$65,$B1,$A7,$93
	.dc.b	$63,$51,$E0,$6B,$FE,$A0,$6B,$E3,$8E,$04,$7A,$90,$E4,$E5,$A6,$4E
	.dc.b	$8E,$5A,$64,$EC,$53,$D8,$72,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D
	.dc.b	$18,$21,$82,$1F,$20,$F2,$14,$EE,$53,$A1,$4E,$65,$39,$C7,$CC,$6C
	.dc.b	$7C,$37,$71,$97,$81,$1C,$12,$38,$47,$23,$CB,$19,$6C,$69,$E4,$D8
	.dc.b	$D4,$7A,$9A,$FA,$9A,$F9,$23,$81,$1E,$A4,$39,$39,$69,$93,$A3,$96
	.dc.b	$99,$3B,$14,$F6,$1C,$88,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08
	.dc.b	$60,$87,$C8,$3C,$85,$3B,$94,$E8,$53,$99,$4E,$71,$F3,$1B,$1F,$0D
	.dc.b	$DC,$65,$E0,$47,$04,$8E,$11,$C8,$F2,$C6,$5B,$1A,$79,$36,$35,$1E
	.dc.b	$A6,$BE,$A6,$BE,$48,$E0,$47,$A9,$0E,$4E,$5A,$64,$E8,$E5,$A6,$4E
	.dc.b	$C5,$3D,$87,$23,$E7,$E5,$FC,$90,$91,$FC,$75,$B1,$FC,$91,$AF,$F2
	.dc.b	$C6,$BF,$DB,$0F,$ED,$AE,$1C,$35,$CC,$75,$23,$D5,$AF,$56,$BD,$4F
	.dc.b	$F0,$93,$87,$AB,$5E,$A7,$32,$1C,$CA,$73,$8F,$23,$D0,$C9,$D0,$A7
	.dc.b	$43,$91,$0E,$51,$E1,$79,$19,$39,$14,$E4,$70,$3A,$9C,$0C,$1B,$30
	.dc.b	$72,$32,$43,$04,$3F,$D2,$27,$FA,$2C,$A7,$FA,$64,$B1,$FE,$93,$6C
	.dc.b	$7F,$A8,$5F,$F5,$84,$65,$FF,$4E,$F1,$FF,$A9,$DC,$CE,$F9,$3F,$D4
	.dc.b	$11,$A3,$FD,$41,$C1,$FE,$AA,$8F,$F5,$A4,$70,$3F,$D5,$91,$9A,$FF
	.dc.b	$50,$C7,$94,$E4,$7A,$4A,$64,$FF,$55,$D7,$86,$3C,$95,$E2,$9C,$9D
	.dc.b	$8F,$03,$E2,$35,$5E,$18,$FF,$59,$9D,$5C,$3F,$10,$E5,$1C,$1E,$73
	.dc.b	$FE,$0C,$9F,$F2,$5C,$A3,$EE,$36,$7F,$D6,$1C,$8E,$47,$4F,$F9,$4B
	.dc.b	$8B,$FD,$64,$4E,$5F,$D5,$57,$FA,$EC,$95,$FE,$BA,$BF,$F5,$E9,$FE
	.dc.b	$C5,$8E,$97,$CE,$B1,$5D,$AB,$FD,$7B,$C9,$FE,$B0,$D3,$FD,$1F,$8F
	.dc.b	$FA,$0A,$7F,$D2,$D1,$82,$3F,$EA,$79,$FF,$47,$F2,$7F,$A2,$F6,$FF
	.dc.b	$44,$6D,$FE,$8B,$BD,$BF,$E9,$1B,$FF,$56,$9F,$E9,$8A,$FF,$48,$F2
	.dc.b	$FF,$B3,$23,$A6,$3E,$1A,$FF,$4C,$CF,$FA,$EA,$78,$E3,$FE,$9D,$8F
	.dc.b	$F6,$1C,$73,$C7,$C3,$1C,$EB,$C1,$5E,$69,$FF,$5F,$C6,$2F,$B5,$72
	.dc.b	$AE,$97,$FE,$A6,$BD,$CF,$86,$79,$4F,$FA,$90,$FF,$61,$1E,$13,$E3
	.dc.b	$9E,$AC,$7E,$0C,$F1,$BE,$F3,$D4,$E5,$87,$FB,$25,$5B,$1F,$F5,$64
	.dc.b	$25,$7F,$E9,$6E,$47,$7B,$E1,$3F,$D2,$F1,$86,$C6,$73,$FF,$2F,$CE
	.dc.b	$5B,$19,$C3,$F9,$54,$E5,$B1,$97,$FD,$6C,$D7,$FD,$19,$DB,$FE,$96
	.dc.b	$25,$7F,$9B,$39,$BF,$D6,$11,$89,$FF,$53,$B5,$FE,$89,$1A,$9E,$1C
	.dc.b	$CA,$D5,$76,$9C,$C6,$21,$5C,$5E,$B3,$A2,$38,$21,$E0,$AE,$D1,$A8
	.dc.b	$C5,$75,$AE,$91,$89,$D3,$E3,$39,$94,$D1,$83,$CB,$5D,$9C,$4F,$47
	.dc.b	$81,$0D,$D7,$86,$78,$47,$42,$1B,$9E,$4F,$0A,$F2,$B4,$EF,$3C,$9E
	.dc.b	$0D,$74,$74,$29,$AA,$F9,$CF,$FB,$93,$B7,$FD,$4A,$4A,$FF,$4F,$36
	.dc.b	$3F,$D4,$AD,$7F,$D5,$AD,$7F,$BB,$35,$FF,$02,$35,$ED,$1D,$08,$F4
	.dc.b	$6B,$D1,$AF,$46,$BD,$1A,$F4,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C
	.dc.b	$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$82,$3E,$43,$63
	.dc.b	$DE,$D3,$D0,$53,$C6,$53,$C0,$53,$B9,$4E,$F1,$D8,$91,$D9,$B1,$D9
	.dc.b	$AF,$66,$BD,$9A,$F6,$6B,$DA,$3A,$11,$E8,$D7,$A3,$5E,$8D,$7A,$35
	.dc.b	$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E,$06
	.dc.b	$88,$68,$A6,$8C,$10,$C1,$0F,$A8,$48,$F9,$2D,$8F,$80,$D3,$D2,$53
	.dc.b	$C8,$53,$C2,$53,$C0,$53,$C1,$1D,$89,$1D,$9B,$1D,$9A,$F6,$6B,$D9
	.dc.b	$AF,$66,$BD,$A3,$A1,$1E,$8D,$7A,$35,$E8,$D7,$A3,$5E,$87,$22,$1C
	.dc.b	$8A,$72,$29,$C8,$A7,$23,$81,$FA,$AE,$D4,$7F,$A4,$63,$FD,$3D,$18
	.dc.b	$70,$1C,$3B,$6B,$FE,$99,$6B,$CA,$38,$BC,$5B,$1A,$81,$93,$BB,$97
	.dc.b	$07,$78,$E2,$47,$6D,$74,$60,$86,$08,$79,$DE,$0D,$36,$53,$67,$AC
	.dc.b	$C9,$4C,$9E,$A8,$F4,$12,$3A,$C6,$CE,$84,$30,$43,$93,$C1,$A7,$80
	.dc.b	$A7,$80,$F8,$A7,$33,$25,$32,$75,$29,$A3,$11,$F3,$DF,$F4,$6F,$37
	.dc.b	$FA,$2F,$9F,$FD,$22,$48,$FF,$50,$9F,$E8,$63,$FD,$1B,$C7,$FE,$AB
	.dc.b	$78,$C7,$FA,$82,$31,$C1,$FE,$90,$9F,$F4,$FC,$61,$E4,$7F,$A4,$EB
	.dc.b	$C0,$F6,$7F,$D2,$84,$AF,$14,$7F,$AA,$4E,$E7,$8B,$0F,$F5,$43,$C6
	.dc.b	$36,$FF,$A8,$4E,$E7,$8A,$36,$F3,$BF,$33,$D0,$F2,$4F,$AE,$77,$1C
	.dc.b	$EB,$C8,$FF,$BA,$DF,$F5,$22,$E7,$FF,$4C,$AE,$4F,$F5,$13,$FE,$A8
	.dc.b	$E5,$FF,$4D,$43,$9B,$FD,$4B,$1F,$EA,$3A,$FF,$52,$F3,$FE,$C9,$B7
	.dc.b	$CE,$BF,$D4,$73,$FE,$A5,$8F,$D5,$CF,$82,$FA,$47,$EB,$B9,$FF,$D4
	.dc.b	$CF,$F9,$F4,$9A,$7F,$A2,$F3,$FF,$3F,$C6,$F9,$FF,$D4,$8F,$FA,$12
	.dc.b	$74,$F0,$E7,$FF,$42,$57,$08,$FF,$45,$4F,$FA,$81,$E5,$BF,$FA,$49
	.dc.b	$A7,$FA,$22,$BF,$80,$BF,$E8,$8B,$FF,$55,$BA,$AF,$F4,$66,$1F,$E9
	.dc.b	$C8,$FF,$54,$CF,$92,$7B,$47,$82,$BF,$D4,$53,$96,$DF,$86,$74,$7F
	.dc.b	$A6,$6B,$FD,$7C,$FA,$CF,$04,$FF,$AC,$4F,$6B,$93,$89,$E2,$9E,$97
	.dc.b	$E5,$3F,$D5,$B5,$D0,$F8,$E4,$AE,$6F,$FA,$8A,$BB,$8F,$9D,$7E,$4C
	.dc.b	$FC,$0F,$F9,$5C,$61,$B1,$9C,$FF,$9C,$4E,$5B,$19,$AF,$9A,$75,$70
	.dc.b	$FD,$09,$E0,$E9,$AF,$FA,$73,$F5,$27,$8D,$FC,$F3,$A5,$18,$C3,$FD
	.dc.b	$4C,$DA,$EF,$3A,$AE,$13,$99,$E9,$18,$25,$72,$6C,$72,$9E,$0F,$F4
	.dc.b	$6A,$E9,$5B,$24,$6E,$B8,$BD,$A3,$2D,$34,$E1,$E9,$5D,$9F,$F5,$9B
	.dc.b	$6B,$BC,$70,$21,$CC,$87,$2A,$F0,$B6,$BB,$3D,$23,$2D,$8C,$CF,$68
	.dc.b	$E6,$64,$EE,$73,$29,$93,$6F,$17,$46,$2B,$D4,$4A,$F0,$CE,$E7,$91
	.dc.b	$C0,$86,$0F,$EE,$BE,$23,$E0,$3F,$BC,$37,$0F,$F0,$2B,$67,$FC,$14
	.dc.b	$D7,$F2,$2D,$7F,$7E,$D7,$AC,$F4,$24,$F4,$6C,$74,$6B,$D1,$AF,$46
	.dc.b	$BD,$23,$91,$1E,$4D,$79,$35,$E4,$D7,$91,$C0,$87,$02,$9C,$0A,$70
	.dc.b	$34,$43,$45,$34,$60,$86,$08,$27,$E4,$36,$7D,$ED,$9F,$43,$4F,$19
	.dc.b	$4F,$01,$4E,$E5,$3B,$CF,$62,$4F,$66,$CF,$66,$CF,$66,$BD,$9A,$F6
	.dc.b	$6B,$DA,$7A,$12,$7A,$36,$3A,$35,$E8,$D7,$A3,$5E,$91,$C8,$8F,$26
	.dc.b	$BC,$9A,$F2,$6B,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$3E,$A1,$27,$E4,$B6,$7E,$03,$67,$D2,$D3,$C8,$53,$C2,$53
	.dc.b	$C0,$53,$C1,$3D,$89,$3D,$9B,$3D,$9B,$3D,$9A,$F6,$6B,$D9,$AF,$69
	.dc.b	$E8,$49,$E8,$D8,$E8,$D7,$A3,$5E,$8D,$7A,$47,$22,$3C,$9A,$F2,$6B
	.dc.b	$C9,$AF,$23,$81,$0E,$05,$38,$14,$E0,$68,$87,$F8,$7B,$F3,$A7,$C4
	.dc.b	$FF,$A8,$5B,$3F,$EA,$E6,$CF,$FA,$CD,$AF,$F7,$A6,$BC,$E7,$91,$27
	.dc.b	$93,$63,$93,$5E,$4D,$79,$47,$02,$3C,$1A,$F0,$6B,$C0,$D1,$0D,$14
	.dc.b	$D1,$82,$18,$21,$FE,$C2,$9F,$6B,$67,$CC,$D9,$F0,$B4,$EC,$53,$A9
	.dc.b	$4E,$A2,$7A,$B6,$7A,$B6,$7A,$B4,$EA,$53,$A9,$4E,$B3,$D0,$93,$FE
	.dc.b	$36,$F7,$65,$7F,$8D,$1B,$5F,$E2,$FD,$51,$88,$FF,$1A,$46,$27,$81
	.dc.b	$FE,$EA,$8F,$F4,$A1,$FE,$99,$1D,$3F,$EA,$87,$FD,$34,$7F,$A7,$8D
	.dc.b	$93,$83,$FD,$3A,$47,$4D,$8F,$F4,$DC,$72,$6B,$FE,$9F,$EA,$FF,$55
	.dc.b	$F2,$7F,$AE,$67,$2D,$8C,$DF,$FA,$9A,$31,$5F,$EB,$32,$1D,$67,$6F
	.dc.b	$A9,$FF,$57,$34,$FF,$5E,$14,$FF,$59,$98,$AF,$80,$F9,$4E,$A2,$54
	.dc.b	$8E,$5C,$FF,$EA,$5E,$9F,$F5,$4C,$7F,$AA,$CF,$F5,$A7,$4F,$FA,$BF
	.dc.b	$AB,$FD,$64,$47,$07,$FA,$BE,$3F,$D7,$D1,$FE,$BA,$9E,$7D,$5F,$EB
	.dc.b	$09,$FF,$4E,$63,$E1,$D3,$FD,$37,$D5,$FE,$95,$9F,$F4,$97,$4F,$FA
	.dc.b	$3D,$FF,$41,$9C,$0F,$F5,$31,$A3,$3D,$49,$FF,$53,$C6,$EB,$AD,$FF
	.dc.b	$A5,$27,$B4,$FF,$A5,$CE,$66,$EF,$FD,$4A,$E2,$FA,$57,$9E,$73,$7F
	.dc.b	$E9,$B7,$FD,$32,$4A,$F1,$DF,$CF,$79,$61,$F4,$1F,$F5,$46,$1F,$EA
	.dc.b	$58,$C5,$FF,$A8,$E7,$8E,$1D,$DF,$6C,$EB,$0E,$D1,$FE,$B0,$BF,$F5
	.dc.b	$6C,$7C,$78,$C4,$ED,$E6,$78,$B5,$FF,$6A,$57,$85,$B8,$FC,$48,$E7
	.dc.b	$3C,$31,$FF,$65,$C6,$E3,$D3,$3B,$BE,$D1,$E0,$BF,$43,$F2,$C7,$33
	.dc.b	$D4,$7C,$7F,$FA,$A5,$B5,$FE,$B4,$AD,$C6,$1F,$F5,$4B,$5F,$F7,$2D
	.dc.b	$73,$AE,$44,$AE,$51,$97,$FC,$E7,$B7,$F9,$D8,$8E,$09,$5F,$E7,$68
	.dc.b	$CB,$63,$2F,$F9,$E2,$32,$D8,$CB,$FE,$7A,$8C,$B6,$32,$FF,$A2,$23
	.dc.b	$2D,$8C,$BF,$E5,$B8,$CB,$63,$3D,$09,$CB,$63,$31,$82,$38,$24,$FA
	.dc.b	$C8,$7A,$4A,$79,$CA,$79,$4A,$78,$CB,$3E,$16,$CF,$81,$B3,$E8,$F8
	.dc.b	$3F,$F5,$29,$1F,$F5,$2B,$63,$FD,$4A,$D7,$FD,$4A,$D7,$FD,$4A,$D7
	.dc.b	$FD,$62,$D7,$FD,$04,$DA,$34,$93,$E0,$24,$F8,$1A,$F8,$1A,$F8,$1A
	.dc.b	$F8,$1A,$F8,$1B,$3E,$06,$CF,$82,$7B,$10,$EC,$53,$B1,$4E,$C5,$3B
	.dc.b	$16,$7B,$36,$3B,$3D,$08,$74,$29,$D0,$A7,$42,$C7,$46,$C7,$47,$91
	.dc.b	$0E,$45,$39,$16,$39,$36,$39,$3C,$08,$70,$2C,$70,$6C,$70,$74,$48
	.dc.b	$D3,$63,$51,$82,$38,$24,$FD,$42,$4F,$C9,$6B,$F0,$1A,$FA,$5A,$F9
	.dc.b	$1A,$F8,$5B,$3E,$06,$CF,$82,$7B,$10,$EC,$53,$B1,$4E,$C5,$3B,$16
	.dc.b	$7B,$36,$3B,$3D,$08,$74,$29,$D0,$A7,$42,$C7,$46,$C7,$47,$91,$0E
	.dc.b	$45,$39,$16,$39,$36,$39,$3C,$08,$70,$2C,$70,$6C,$70,$74,$48,$D3
	.dc.b	$63,$51,$82,$38,$24,$FD,$42,$4F,$C9,$6B,$F0,$1A,$FA,$5A,$F9,$1A
	.dc.b	$F8,$5B,$3E,$06,$CF,$82,$7B,$1F,$E9,$9F,$63,$FD,$3C,$49,$FF,$4F
	.dc.b	$36,$35,$18,$23,$82,$1F,$EA,$18,$E0,$D8,$DB,$4D,$9F,$EA,$F8,$DB
	.dc.b	$63,$6D,$36,$7F,$A4,$63,$6D,$8D,$B4,$D9,$FE,$8B,$8D,$B6,$36,$D3
	.dc.b	$7D,$29,$DB,$67,$6D,$9D,$CE,$89,$3A,$6C,$6A,$30,$47,$D7,$DA,$E8
	.dc.b	$7B,$FF,$E3,$39,$FF,$46,$B5,$FF,$32,$4E,$E3,$10,$8C,$57,$F8,$5D
	.dc.b	$B1,$FE,$99,$9D,$C6,$23,$AC,$FF,$A5,$F8,$BF,$D1,$9D,$3F,$E7,$EE
	.dc.b	$9F,$F3,$BF,$37,$FA,$12,$FF,$D0,$47,$F9,$E1,$C3,$FE,$7B,$32,$43
	.dc.b	$04,$76,$4B,$FF,$48,$12,$3F,$D1,$07,$FA,$9E,$7C,$31,$96,$C6,$8F
	.dc.b	$F5,$31,$FE,$A4,$23,$FE,$9E,$8E,$15,$FE,$96,$8E,$24,$70,$49,$FF
	.dc.b	$4B,$35,$FF,$43,$F4,$7F,$A2,$7A,$7F,$D2,$51,$FE,$90,$3F,$D2,$46
	.dc.b	$8F,$F2,$33,$FE,$23,$8F,$F4,$8C,$63,$87,$FD,$28,$F0,$9E,$6F,$1E
	.dc.b	$5F,$F4,$94,$63,$9F,$FC,$FB,$3F,$E6,$53,$FC,$BB,$9F,$FA,$1C,$8F
	.dc.b	$FA,$0F,$9F,$FD,$1B,$3C,$35,$FF,$46,$CE,$6B,$FC,$C3,$3F,$E6,$A8
	.dc.b	$FF,$31,$11,$C1,$C7,$A3,$FD,$65,$19,$C7,$FC,$D7,$3F,$E6,$FB,$FF
	.dc.b	$42,$4F,$F9,$B4,$F2,$9D,$C3,$FE,$76,$C3,$FC,$E9,$B7,$F9,$F3,$1F
	.dc.b	$4C,$FF,$A5,$EB,$FC,$F9,$8F,$83,$0F,$14,$7F,$AA,$CF,$36,$1E,$E3
	.dc.b	$FD,$35,$5E,$E3,$94,$78,$DF,$2C,$F9,$EB,$FD,$2A,$D9,$EE,$7F,$A6
	.dc.b	$B0,$F2,$3E,$9A,$F4,$61,$E6,$BF,$F4,$D5,$F9,$37,$FD,$C7,$CB,$FE
	.dc.b	$B3,$88,$D2,$65,$FB,$38,$CB,$63,$39,$7C,$E3,$F7,$53,$A7,$1C,$EF
	.dc.b	$88,$7D,$07,$F0,$08,$C3,$63,$38,$7F,$A3,$9B,$5C,$9A,$FF,$A5,$9A
	.dc.b	$FF,$BB,$EB,$9C,$F1,$AE,$84,$70,$49,$E9,$19,$6C,$65,$ED,$19,$6C
	.dc.b	$65,$F0,$57,$89,$FF,$59,$B6,$BC,$71,$D0,$87,$22,$9C,$AB,$B3,$6B
	.dc.b	$B3,$D2,$32,$D8,$CC,$F6,$8E,$84,$39,$1D,$8E,$05,$78,$3A,$38,$1A
	.dc.b	$2B,$A7,$06,$C8,$60,$8E,$C9,$5F,$16,$3F,$DA,$FA,$7F,$40,$6D,$7F
	.dc.b	$B5,$5A,$7F,$B3,$A3,$54,$69,$27,$84,$65,$B1,$99,$E5,$18,$23,$82
	.dc.b	$1F,$96,$8E,$0D,$8D,$B4,$D9,$FE,$E6,$8D,$B6,$36,$D3,$67,$E5,$63
	.dc.b	$6D,$8D,$B4,$D9,$F0,$8F,$01,$0E,$85,$38,$14,$E0,$68,$86,$8A,$68
	.dc.b	$C1,$0C,$12,$7E,$63,$E4,$6B,$DD,$AF,$46,$BC,$DA,$F3,$9F,$82,$D9
	.dc.b	$F8,$2D,$3B,$46,$A7,$89,$1C,$12,$78,$C6,$5B,$19,$9E,$71,$82,$38
	.dc.b	$21,$EB,$8E,$0D,$8D,$B4,$D9,$EF,$8D,$B6,$36,$D3,$67,$C3,$8D,$B6
	.dc.b	$36,$D3,$67,$C3,$3C,$04,$3A,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04
	.dc.b	$30,$49,$F9,$8F,$91,$AF,$76,$BD,$1A,$F3,$6B,$CE,$7E,$0B,$67,$E0
	.dc.b	$B4,$ED,$1A,$9E,$24,$70,$49,$E3,$19,$6C,$66,$79,$C6,$08,$E0,$87
	.dc.b	$AE,$38,$36,$36,$D3,$67,$BE,$36,$D8,$DB,$4D,$9F,$0E,$36,$D8,$DB
	.dc.b	$4D,$9F,$0C,$FB,$9C,$EF,$73,$F9,$43,$67,$F9,$43,$67,$FB,$90,$70
	.dc.b	$FF,$75,$6B,$FE,$24,$6A,$38,$4F,$62,$4F,$66,$C7,$69,$E2,$F0,$9D
	.dc.b	$11,$D4,$F0,$7C,$4D,$7C,$4D,$7C,$53,$D0,$91,$98,$EC,$74,$9E,$CF
	.dc.b	$46,$BD,$1A,$F4,$8E,$53,$D2,$73,$3D,$0E,$47,$52,$18,$21,$D4,$C9
	.dc.b	$4C,$9D,$DE,$B1,$DE,$36,$64,$A7,$84,$E4,$7F,$A9,$B2,$7C,$87,$F8
	.dc.b	$D4,$95,$FE,$33,$6C,$FF,$8C,$75,$FF,$59,$46,$23,$FC,$C3,$18,$9F
	.dc.b	$F5,$34,$7F,$A3,$B9,$BF,$D4,$C7,$FA,$1D,$FF,$4C,$35,$FF,$4A,$F3
	.dc.b	$FF,$A4,$DF,$F4,$C7,$17,$FA,$B9,$FF,$55,$57,$FA,$59,$FF,$48,$90
	.dc.b	$FF,$38,$C6,$DF,$F5,$66,$9F,$EA,$97,$99,$E3,$39,$98,$AF,$01,$FE
	.dc.b	$9F,$7A,$1E,$62,$9E,$6B,$F5,$B8,$AF,$03,$FE,$B9,$BF,$33,$E4,$3C
	.dc.b	$07,$E8,$6D,$1F,$A3,$B5,$FF,$A2,$B9,$7F,$D3,$E4,$E6,$73,$FF,$AB
	.dc.b	$5B,$8F,$FA,$95,$FF,$53,$BF,$E6,$59,$FF,$56,$3F,$EA,$B9,$FF,$56
	.dc.b	$F3,$FF,$AA,$2B,$8B,$63,$A1,$FB,$55,$DD,$FE,$7A,$E8,$FF,$49,$DF
	.dc.b	$FA,$63,$5F,$F4,$4F,$3F,$FA,$16,$3F,$D1,$5D,$1F,$E8,$8A,$DD,$7F
	.dc.b	$A3,$1B,$19,$DF,$FD,$0D,$5F,$E8,$C1,$7F,$EA,$CA,$FF,$4B,$D7,$FA
	.dc.b	$75,$FF,$4D,$57,$FA,$D1,$B5,$FE,$BF,$BE,$D3,$D2,$3F,$D4,$2F,$FA
	.dc.b	$EB,$1F,$F5,$9D,$78,$A7,$C1,$87,$FA,$A1,$F0,$CE,$AB,$94,$E4,$91
	.dc.b	$99,$FF,$50,$61,$FE,$B1,$8C,$61,$F1,$30,$ED,$5F,$10,$EE,$FA,$23
	.dc.b	$FD,$65,$5F,$37,$0E,$F3,$F0,$2B,$FD,$5C,$7F,$B1,$27,$E1,$4F,$F2
	.dc.b	$9A,$47,$EE,$7E,$43,$FD,$58,$4E,$0F,$F5,$53,$6F,$F7,$1C,$68,$C4
	.dc.b	$7E,$E2,$31,$3F,$E5,$EF,$1F,$FC,$C6,$47,$04,$9F,$E8,$3C,$E9,$D4
	.dc.b	$61,$AF,$FA,$D1,$AF,$FB,$9A,$38,$4F,$38,$D3,$5D,$4F,$07,$FD,$60
	.dc.b	$D5,$3C,$9F,$04,$F1,$76,$D7,$73,$C6,$78,$47,$80,$E2,$43,$04,$3C
	.dc.b	$73,$CE,$38,$C6,$0E,$05,$38,$1E,$23,$25,$32,$79,$23,$A4,$F9,$48
	.dc.b	$E0,$91,$C8,$E0,$7D,$2F,$1F,$FD,$32,$48,$FF,$4B,$B6,$35,$18,$23
	.dc.b	$82,$1F,$EE,$A8,$E0,$D8,$DB,$4D,$9F,$6E,$36,$D8,$DB,$4D,$9F,$EE
	.dc.b	$98,$DB,$63,$6D,$36,$7D,$E8,$DB,$63,$6D,$37,$49,$DB,$67,$6D,$9D
	.dc.b	$CE,$89,$3A,$6C,$6A,$30,$47,$04,$3E,$21,$0F,$80,$53,$D8,$53,$D0
	.dc.b	$59,$F2,$36,$7C,$2D,$9F,$03,$4F,$01,$D8,$87,$62,$9D,$8B,$3D,$9B
	.dc.b	$3D,$9B,$3D,$9A,$76,$3A,$10,$E8,$59,$E8,$D9,$E8,$D9,$E8,$D3,$A1
	.dc.b	$C8,$93,$C9,$B3,$C9,$B3,$C9,$A7,$29,$E0,$49,$E0,$D9,$E0,$D9,$E1
	.dc.b	$3A,$24,$E9,$B1,$A8,$C1,$1C,$10,$FA,$84,$3E,$49,$4F,$80,$53,$D2
	.dc.b	$59,$F2,$36,$7C,$2D,$9F,$03,$4F,$01,$D8,$87,$62,$9D,$8B,$3D,$9B
	.dc.b	$3D,$9B,$3D,$9A,$76,$3A,$10,$E8,$59,$E8,$D9,$E8,$D9,$E8,$D3,$A1
	.dc.b	$C8,$93,$C9,$B3,$C9,$B3,$C9,$A7,$29,$E0,$71,$E7,$7C,$77,$FA,$68
	.dc.b	$95,$FE,$9A,$6D,$7F,$A3,$5A,$FF,$A6,$5A,$FF,$A6,$5A,$FF,$A2,$1A
	.dc.b	$FF,$A8,$5B,$3E,$09,$EC,$49,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D8
	.dc.b	$ED,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$F4,$79,$10,$E4,$53,$91
	.dc.b	$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF,$05,$CE,$C9,$F0,$9F,$E6,$89
	.dc.b	$FF,$0D,$B5,$FF,$0D,$35,$FF,$4E,$35,$E7,$1C,$AF,$FD,$B3,$5D,$6B
	.dc.b	$4D,$8C,$CE,$EB,$FD,$BE,$7F,$A0,$CF,$F3,$E7,$47,$FA,$23,$A3,$FD
	.dc.b	$35,$C1,$FE,$82,$E1,$FF,$48,$4F,$FA,$2E,$B7,$5F,$E8,$86,$9F,$E9
	.dc.b	$28,$D3,$D4,$87,$FA,$3C,$D4,$76,$7F,$D1,$67,$FA,$3C,$FF,$4D,$3D
	.dc.b	$5E,$35,$E3,$C3,$FD,$27,$18,$38,$DF,$9A,$BA,$5F,$AE,$7F,$D3,$86
	.dc.b	$AB,$FD,$37,$D1,$FE,$93,$E7,$FF,$43,$CF,$FA,$45,$AF,$FA,$47,$A3
	.dc.b	$FD,$26,$FF,$A6,$DF,$F4,$BB,$98,$E3,$CD,$FE,$91,$8E,$41,$FF,$48
	.dc.b	$BF,$E7,$67,$FC,$DB,$CB,$FE,$67,$8F,$F3,$88,$3F,$CF,$38,$7F,$A1
	.dc.b	$73,$FF,$35,$96,$3F,$CD,$46,$F9,$FF,$D1,$D1,$9D,$7F,$CD,$73,$9C
	.dc.b	$3B,$BE,$1B,$D6,$1E,$1A,$FF,$3F,$5F,$6A,$FF,$49,$4F,$FA,$75,$F2
	.dc.b	$9E,$43,$FD,$50,$F1,$23,$D4,$C9,$67,$FD,$36,$F2,$C3,$CE,$F6,$3F
	.dc.b	$D1,$75,$FE,$9D,$BF,$2C,$71,$BF,$15,$F9,$A3,$96,$1E,$C7,$FD,$4A
	.dc.b	$7C,$43,$CD,$1E,$03,$81,$27,$CF,$1F,$E7,$0F,$82,$FD,$63,$6F,$F4
	.dc.b	$BC,$89,$CB,$63,$31,$82,$38,$38,$FB,$FC,$08,$E0,$95,$FE,$90,$8C
	.dc.b	$B6,$32,$FE,$0C,$FC,$25,$27,$71,$93,$FD,$12,$E1,$FF,$4A,$35,$E2
	.dc.b	$D7,$73,$D9,$DB,$5D,$B5,$D9,$E5,$21,$E1,$9F,$24,$6D,$B1,$A3,$A9
	.dc.b	$4E,$A5,$3A,$9C,$E7,$CD,$1C,$A3,$13,$D6,$72,$47,$04,$8E,$93,$C5
	.dc.b	$F0,$9C,$8C,$CF,$6E,$FF,$BA,$4A,$FC,$0B,$67,$FD,$8A,$D7,$E0,$35
	.dc.b	$FF,$61,$35,$FA,$ED,$7F,$D7,$AD,$A3,$49,$3E,$02,$4F,$81,$AF,$81
	.dc.b	$AF,$81,$AF,$81,$AF,$81,$B3,$E0,$6C,$F8,$27,$B1,$0E,$C5,$3B,$14
	.dc.b	$EC,$53,$B1,$67,$B3,$63,$B3,$D0,$87,$42,$9D,$0A,$74,$2C,$74,$6C
	.dc.b	$74,$79,$10,$E4,$53,$91,$63,$93,$63,$93,$C0,$87,$02,$C7,$06,$C7
	.dc.b	$07,$44,$8D,$36,$35,$18,$23,$82,$4F,$D4,$24,$FC,$96,$BF,$01,$AF
	.dc.b	$A5,$AF,$91,$AF,$85,$B3,$E0,$6C,$F8,$27,$B1,$0E,$C5,$3B,$14,$EC
	.dc.b	$53,$B1,$67,$B3,$63,$B3,$D0,$87,$42,$9D,$0A,$74,$2C,$74,$6C,$74
	.dc.b	$79,$10,$E4,$53,$91,$63,$93,$63,$93,$C0,$87,$02,$C7,$06,$C7,$07
	.dc.b	$44,$8D,$36,$35,$18,$23,$82,$4F,$D4,$24,$FC,$96,$BF,$01,$AF,$A5
	.dc.b	$AF,$91,$AF,$85,$B3,$E0,$6C,$F8,$27,$B1,$FD,$2C,$72,$BE,$3B,$FB
	.dc.b	$2B,$6B,$FC,$92,$D9,$E0,$D7,$FA,$A3,$5F,$EB,$4D,$7F,$C3,$CD,$7B
	.dc.b	$CF,$62,$4F,$66,$C7,$66,$BD,$9A,$F6,$6B,$D9,$AF,$68,$E8,$47,$A3
	.dc.b	$5E,$8D,$7A,$35,$E8,$D7,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0
	.dc.b	$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$1C,$BF,$E9,$5F,$81
	.dc.b	$FF,$44,$F3,$1A,$4A,$FF,$45,$B6,$BF,$C2,$55,$C6,$72,$7F,$A3,$A3
	.dc.b	$FC,$ED,$CB,$FE,$73,$7F,$D0,$E7,$F9,$7F,$87,$FC,$F0,$FF,$9D,$A3
	.dc.b	$4F,$0E,$6F,$F4,$13,$FE,$89,$38,$57,$FA,$09,$C1,$FE,$89,$9F,$F4
	.dc.b	$8B,$8A,$EA,$F8,$0D,$57,$FA,$18,$EB,$C3,$FE,$8D,$8F,$F4,$89,$FE
	.dc.b	$96,$3C,$91,$A7,$FD,$14,$68,$F1,$94,$EA,$70,$3F,$C7,$53,$FE,$BF
	.dc.b	$A4,$FE,$BE,$BF,$D0,$11,$8A,$DD,$E2,$BF,$D2,$1C,$9F,$E8,$EE,$DB
	.dc.b	$5F,$E8,$EE,$4F,$F4,$83,$FE,$93,$3F,$75,$CE,$E7,$FF,$4C,$C7,$FA
	.dc.b	$36,$3F,$D4,$0F,$1A,$E7,$5F,$E9,$68,$FF,$35,$E7,$FE,$8A,$7F,$CD
	.dc.b	$7C,$DF,$E7,$23,$FC,$D3,$CE,$ED,$6F,$FE,$78,$AD,$9F,$E7,$C3,$FC
	.dc.b	$E9,$3F,$E8,$0B,$E8,$F1,$9F,$F4,$04,$FF,$A2,$0F,$0D,$FF,$98,$67
	.dc.b	$83,$6F,$BC,$F4,$C3,$FD,$05,$5C,$E3,$FD,$18,$72,$3F,$D4,$D3,$C3
	.dc.b	$0F,$0D,$70,$8C,$5F,$5C,$3C,$75,$FE,$89,$8F,$7C,$EA,$F9,$D6,$A7
	.dc.b	$E2,$47,$FA,$5C,$F8,$15,$DB,$6F,$89,$8F,$96,$3E,$0E,$1C,$EB,$E3
	.dc.b	$9E,$71,$3B,$BF,$85,$8F,$9F,$1E,$6D,$AF,$7D,$6D,$F1,$1C,$DC,$3D
	.dc.b	$9A,$4B,$F1,$3F,$3A,$3D,$A7,$CD,$C7,$D3,$3A,$37,$5E,$1C,$7D,$78
	.dc.b	$7C,$3B,$F8,$38,$FA,$30,$E3,$3E,$86,$CF,$A3,$0E,$93,$98,$FC,$3C
	.dc.b	$7C,$FE,$67,$ED,$0F,$D4,$18,$7F,$57,$9C,$C6,$2F,$FD,$29,$5B,$AD
	.dc.b	$36,$33,$3B,$7F,$D5,$D5,$CA,$B7,$39,$9F,$F5,$8D,$71,$7F,$D4,$1F
	.dc.b	$25,$FE,$C4,$25,$7F,$B2,$1B,$1E,$56,$BC,$23,$44,$74,$D7,$46,$08
	.dc.b	$60,$87,$A2,$3A,$B6,$39,$34,$E2,$53,$89,$FE,$CC,$8E,$2D,$8E,$2D
	.dc.b	$38,$94,$E2,$2D,$3C,$A3,$0D,$8C,$CF,$49,$CB,$63,$31,$D8,$C9,$5C
	.dc.b	$BE,$06,$BE,$B2,$1E,$52,$9C,$8E,$13,$E3,$F0,$7F,$DF,$64,$AF,$F7
	.dc.b	$FB,$67,$F1,$8D,$7F,$DB,$CD,$7F,$0E,$D7,$FD,$EA,$D7,$E2,$36,$8D
	.dc.b	$24,$F8,$09,$3E,$06,$BE,$06,$BE,$06,$BE,$06,$BE,$06,$CF,$81,$B3
	.dc.b	$E0,$9E,$C4,$3B,$14,$EC,$53,$B1,$4E,$C5,$9E,$CD,$8E,$CF,$42,$1D
	.dc.b	$0A,$74,$29,$D0,$B1,$D1,$B1,$D1,$E4,$43,$91,$4E,$45,$8E,$4D,$8E
	.dc.b	$4F,$02,$1C,$0B,$1C,$1B,$1C,$1D,$12,$34,$D8,$D4,$60,$8E,$09,$3F
	.dc.b	$50,$93,$F2,$5A,$FC,$06,$BE,$96,$BE,$46,$BE,$16,$CF,$81,$B3,$E0
	.dc.b	$9E,$C4,$3B,$14,$EC,$53,$B1,$4E,$C5,$9E,$CD,$8E,$CF,$42,$1D,$0A
	.dc.b	$74,$29,$D0,$B1,$D1,$B1,$D1,$E4,$43,$91,$4E,$45,$8E,$4D,$8E,$4F
	.dc.b	$02,$1C,$0B,$1C,$1B,$1C,$1D,$12,$34,$D8,$D4,$60,$8E,$09,$3F,$50
	.dc.b	$93,$F2,$5A,$FC,$06,$BE,$96,$BE,$46,$BE,$16,$CF,$81,$B3,$E0,$9E
	.dc.b	$C7,$FA,$E3,$CA,$FF,$5E,$92,$7F,$D7,$AD,$8D,$46,$08,$E0,$87,$FB
	.dc.b	$02,$38,$36,$36,$D3,$67,$FB,$12,$36,$D8,$DB,$4D,$9F,$EC,$88,$DB
	.dc.b	$63,$6D,$36,$7F,$B6,$A3,$6D,$8D,$B4,$DF,$52,$76,$D9,$DB,$67,$73
	.dc.b	$A2,$4E,$9B,$1A,$8C,$11,$F5,$F3,$BA,$5D,$9F,$EB,$02,$57,$FB,$CE
	.dc.b	$75,$18,$AF,$F5,$7D,$70,$AD,$36,$33,$3F,$E1,$97,$FD,$24,$7F,$A5
	.dc.b	$BA,$7F,$D4,$BC,$5F,$E9,$F6,$C7,$FA,$71,$A7,$FA,$C9,$FF,$55,$5F
	.dc.b	$FA,$BE,$78,$74,$FF,$AA,$67,$FD,$4C,$7F,$AB,$78,$BF,$D5,$EF,$FA
	.dc.b	$E4,$C9,$4F,$14,$71,$69,$E6,$3F,$D5,$27,$3A,$F0,$3E,$58,$E0,$E8
	.dc.b	$E8,$70,$E7,$FF,$4E,$F2,$FF,$A6,$DB,$CF,$FE,$A8,$9F,$F5,$31,$0C
	.dc.b	$1F,$EA,$26,$3F,$EA,$7E,$87,$4F,$FA,$C6,$3F,$CD,$BD,$3F,$E6,$9C
	.dc.b	$FF,$CB,$1A,$7F,$A1,$E7,$3D,$1F,$E5,$C7,$FC,$BD,$3F,$E6,$09,$FF
	.dc.b	$42,$47,$FA,$7E,$B9,$3F,$E6,$F3,$15,$C6,$76,$75,$8F,$F3,$E6,$7F
	.dc.b	$E7,$1E,$9F,$F4,$C4,$71,$9D,$5F,$47,$8C,$FF,$A5,$A7,$8E,$DE,$38
	.dc.b	$E3,$3A,$9F,$F3,$C8,$3F,$D1,$18,$78,$E7,$67,$F9,$F6,$3C,$F8,$FC
	.dc.b	$28,$ED,$87,$85,$D5,$F9,$2F,$95,$F5,$AF,$F4,$83,$FE,$A0,$8F,$F4
	.dc.b	$6E,$1E,$58,$FF,$49,$D7,$95,$FF,$4E,$4E,$9F,$09,$F3,$8F,$F4,$4E
	.dc.b	$1E,$37,$8E,$1E,$BE,$2F,$D8,$73,$FE,$95,$B8,$FE,$AA,$31,$C7,$FE
	.dc.b	$4E,$B5,$69,$FD,$8C,$E8,$95,$FE,$93,$FD,$9D,$E4,$FE,$C7,$A5,$18
	.dc.b	$9D,$4E,$63,$17,$F3,$CF,$A1,$5C,$27,$44,$AE,$24,$9E,$07,$F9,$CE
	.dc.b	$B9,$BF,$E9,$C6,$BD,$1A,$EE,$7A,$3C,$AB,$C1,$5D,$63,$94,$E9,$B1
	.dc.b	$97,$A4,$65,$B1,$97,$B3,$67,$AC,$77,$AF,$1C,$62,$3A,$99,$AE,$8F
	.dc.b	$23,$15,$CE,$B8,$92,$77,$5C,$EB,$C6,$70,$21,$D8,$86,$8A,$77,$2B
	.dc.b	$CD,$C1,$B2,$18,$23,$B7,$FD,$4D,$D9,$FC,$CD,$B7,$F6,$1F,$BF,$19
	.dc.b	$93,$49,$3B,$9E,$27,$FB,$86,$32,$D8,$D4,$F2,$6C,$EA,$7E,$F3,$5F
	.dc.b	$F7,$0B,$5F,$1C,$70,$23,$D4,$87,$27,$2D,$32,$74,$72,$D3,$27,$62
	.dc.b	$9E,$C3,$91,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$7E
	.dc.b	$43,$E4,$6B,$DD,$AF,$46,$BC,$DA,$F3,$9F,$98,$D9,$F8,$6E,$E3,$33
	.dc.b	$C0,$8E,$09,$3C,$27,$91,$E5,$8C,$B6,$35,$3C,$9B,$3A,$9F,$53,$5F
	.dc.b	$53,$5F,$24,$70,$23,$D4,$87,$27,$2D,$32,$74,$72,$D3,$27,$62,$9E
	.dc.b	$C3,$91,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$7E,$43
	.dc.b	$E4,$6B,$DD,$AF,$46,$BC,$DA,$F3,$9F,$98,$D9,$F8,$6E,$E3,$33,$C0
	.dc.b	$8E,$09,$3C,$27,$91,$E5,$8C,$B6,$35,$3C,$9B,$3A,$9F,$53,$5F,$53
	.dc.b	$5F,$24,$70,$23,$D4,$87,$27,$2D,$32,$74,$72,$D3,$27,$62,$9E,$C3
	.dc.b	$91,$F3,$FC,$8F,$E4,$84,$8F,$E3,$AD,$8F,$E5,$54,$8C,$47,$F5,$B6
	.dc.b	$BF,$DB,$63,$53,$C8,$91,$CA,$76,$FF,$72,$6A,$9E,$2E,$E7,$8B,$DD
	.dc.b	$AF,$79,$E3,$3C,$23,$47,$12,$18,$21,$C6,$79,$BE,$36,$BE,$39,$EA
	.dc.b	$48,$CC,$F7,$9E,$47,$62,$9D,$8E,$86,$4A,$64,$ED,$1D,$27,$B4,$6C
	.dc.b	$C9,$C6,$5E,$DB,$CA,$FF,$11,$36,$FE,$BB,$6B,$FC,$20,$D7,$EB,$35
	.dc.b	$FF,$11,$B5,$FA,$4D,$AE,$F5,$FD,$B6,$7F,$D0,$7B,$FF,$A4,$9F,$F4
	.dc.b	$C8,$E9,$FF,$4E,$F4,$FF,$8A,$27,$5C,$7F,$E9,$CA,$FF,$49,$46,$1F
	.dc.b	$F4,$81,$8C,$BF,$D2,$77,$FE,$A5,$AE,$13,$FE,$97,$3B,$5F,$FA,$5A
	.dc.b	$F9,$56,$E7,$99,$B7,$C6,$FF,$A9,$8F,$3B,$87,$FD,$4A,$6E,$7A,$5F
	.dc.b	$89,$F5,$3F,$E7,$79,$47,$FA,$6A,$5D,$3F,$EA,$27,$FD,$39,$CF,$FE
	.dc.b	$7D,$E7,$FF,$53,$74,$7F,$A7,$A7,$FC,$29,$48,$C4,$7F,$A9,$CE,$26
	.dc.b	$9E,$13,$CA,$BF,$D4,$D3,$E0,$9C,$F4,$FF,$A2,$23,$FC,$E6,$7F,$A0
	.dc.b	$6B,$FD,$34,$E0,$FF,$47,$E7,$FE,$81,$69,$FE,$7C,$3F,$CF,$DA,$39
	.dc.b	$FA,$47,$FA,$26,$7F,$D3,$13,$FE,$87,$E9,$FF,$49,$4F,$57,$FD,$1E
	.dc.b	$3A,$7F,$D1,$53,$FE,$95,$AF,$F4,$45,$7F,$A3,$9E,$77,$FE,$9A,$3F
	.dc.b	$D3,$B5,$FE,$9A,$9F,$2D,$78,$DF,$F6,$0D,$7A,$6F,$BD,$76,$BE,$33
	.dc.b	$DA,$7C,$B5,$E9,$3F,$D7,$53,$A7,$DE,$FB,$A3,$D1,$3E,$49,$F9,$5B
	.dc.b	$F7,$BE,$D1,$FE,$BA,$3C,$B7,$C6,$33,$1E,$D3,$B9,$BB,$F3,$DF,$58
	.dc.b	$E2,$7B,$0F,$A5,$87,$BD,$F4,$3F,$3E,$BE,$25,$79,$A7,$A9,$DE,$75
	.dc.b	$5F,$17,$16,$3D,$EF,$85,$75,$E1,$7C,$9F,$F9,$94,$95,$FE,$5E,$6E
	.dc.b	$3F,$D0,$9A,$FF,$A6,$DF,$F3,$67,$87,$27,$FC,$CA,$D9,$FE,$92,$D7
	.dc.b	$FD,$44,$D7,$FA,$43,$5F,$F5,$AB,$5F,$E8,$8D,$A3,$49,$3E,$02,$4F
	.dc.b	$81,$AF,$81,$AF,$81,$AF,$81,$AF,$81,$B3,$E0,$6C,$F8,$27,$B1,$0E
	.dc.b	$C5,$3B,$14,$EC,$53,$B1,$67,$B3,$63,$B3,$D0,$87,$42,$9D,$0F,$F6
	.dc.b	$77,$67,$FA,$7C,$95,$FE,$98,$6C,$7F,$7C,$6B,$C2,$34,$47,$4D,$74
	.dc.b	$60,$86,$08,$7F,$80,$E3,$AB,$63,$93,$4E,$25,$38,$9F,$EF,$68,$E2
	.dc.b	$D8,$E2,$D3,$89,$4E,$27,$F8,$2E,$38,$B6,$38,$B4,$E2,$53,$8D,$27
	.dc.b	$8B,$67,$8B,$67,$8B,$5E,$33,$C0,$93,$C1,$B1,$C1,$AF,$08,$D1,$1D
	.dc.b	$35,$D1,$82,$18,$21,$F2,$C8,$7C,$52,$9E,$F2,$CF,$A5,$B3,$E4,$6C
	.dc.b	$F8,$5A,$78,$0A,$78,$0E,$C4,$3B,$16,$7B,$36,$7B,$36,$7B,$34,$EC
	.dc.b	$53,$B1,$D0,$93,$D1,$B3,$D1,$B3,$D1,$A7,$42,$9D,$27,$91,$27,$93
	.dc.b	$67,$93,$67,$93,$5E,$53,$C0,$93,$C1,$B1,$C1,$AF,$08,$D1,$1D,$35
	.dc.b	$D1,$82,$18,$21,$F5,$08,$7C,$92,$9F,$00,$B3,$E9,$6C,$F9,$1B,$3E
	.dc.b	$16,$9E,$02,$9E,$03,$B1,$0E,$C5,$9E,$CD,$9E,$CD,$9E,$CD,$3B,$14
	.dc.b	$EC,$74,$24,$F4,$6C,$F4,$7F,$0B,$E1,$FF,$62,$7F,$D3,$6D,$47,$FA
	.dc.b	$65,$C3,$6B,$2C,$3F,$D2,$D1,$A7,$8D,$23,$11,$A9,$E0,$EA,$33,$3B
	.dc.b	$9C,$CE,$CF,$2B,$88,$D4,$60,$87,$8D,$CB,$67,$A9,$27,$93,$63,$89
	.dc.b	$ED,$29,$E5,$8D,$91,$C9,$E5,$3B,$39,$69,$A9,$F1,$BD,$4C,$90,$C1
	.dc.b	$0F,$01,$C6,$7C,$CE,$CC,$94,$F1,$1C,$4A,$68,$C4,$78,$9C,$10,$D9
	.dc.b	$1D,$2E,$F7,$87,$F6,$9B,$7F,$59,$B3,$F7,$1A,$FE,$15,$AF,$E2,$1A
	.dc.b	$FE,$39,$AF,$79,$FF,$56,$3F,$E9,$4E,$5F,$F4,$8B,$FE,$9E,$3F,$D4
	.dc.b	$85,$3F,$D2,$C6,$39,$BF,$D3,$4B,$83,$FD,$47,$3F,$E9,$88,$E0,$7F
	.dc.b	$AE,$30,$FF,$4E,$3F,$EA,$3D,$BF,$D5,$AE,$4D,$9E,$33,$B4,$F9,$2F
	.dc.b	$BD,$F6,$9E,$47,$FA,$C0,$E5,$1A,$BE,$F1,$FE,$AE,$33,$5F,$EA,$9E
	.dc.b	$6F,$F5,$44,$7F,$A6,$F9,$BF,$D5,$1C,$BF,$B0,$9F,$F5,$34,$7E,$99
	.dc.b	$B3,$FE,$A5,$8C,$73,$7F,$AA,$A3,$A3,$FE,$A4,$7F,$D5,$7C,$9F,$EA
	.dc.b	$94,$AF,$FC,$F1,$9F,$F9,$96,$3F,$D4,$F3,$FE,$70,$E6,$FF,$4A,$3F
	.dc.b	$E7,$98,$CF,$2F,$F9,$AA,$73,$3F,$E7,$16,$C7,$FA,$40,$D4,$7F,$A2
	.dc.b	$8F,$F3,$EC,$78,$1F,$F3,$CC,$6F,$9B,$FD,$12,$DB,$F2,$6B,$FE,$91
	.dc.b	$6D,$72,$3F,$D1,$1B,$FF,$A1,$EB,$C1,$8F,$9B,$0E,$67,$FA,$20,$B7
	.dc.b	$FE,$85,$C3,$A6,$1E,$29,$D4,$F6,$C3,$FD,$71,$3F,$EA,$63,$BE,$3E
	.dc.b	$98,$FF,$47,$C7,$8C,$F9,$07,$BD,$E2,$79,$6B,$C3,$1B,$C7,$E2,$BF
	.dc.b	$EA,$23,$89,$1F,$F5,$3B,$E5,$BF,$43,$C6,$7F,$D4,$78,$FB,$E7,$D1
	.dc.b	$87,$FA,$F7,$0F,$04,$62,$FF,$DC,$BD,$CE,$F7,$C5,$7F,$3B,$8C,$36
	.dc.b	$33,$A7,$F3,$A9,$CB,$63,$35,$FD,$16,$72,$D8,$CB,$FD,$2A,$32,$FF
	.dc.b	$4C,$E5,$78,$6F,$D2,$7F,$A6,$46,$1B,$19,$C3,$FA,$5C,$E5,$B1,$9A
	.dc.b	$FE,$88,$D7,$B3,$5F,$EA,$ED,$53,$CD,$E3,$5E,$02,$57,$66,$C6,$A3
	.dc.b	$04,$70,$43,$C3,$1C,$1B,$1B,$69,$B3,$CB,$1B,$6C,$6D,$A6,$CF,$4D
	.dc.b	$7A,$EB,$D5,$1D,$08,$70,$29,$C2,$BB,$36,$BB,$34,$ED,$3D,$1E,$44
	.dc.b	$34,$74,$74,$E0,$D3,$82,$57,$9C,$95,$FE,$04,$C6,$5F,$F8,$05,$B7
	.dc.b	$F8,$06,$BF,$E0,$16,$BF,$8B,$6B,$FE,$14,$6B,$F7,$9B,$7F,$94,$8E
	.dc.b	$13,$DC,$8F,$76,$BD,$DA,$F0,$34,$43,$45,$34,$60,$86,$08,$79,$DE
	.dc.b	$AD,$39,$14,$E2,$53,$89,$ED,$78,$B4,$E2,$53,$89,$4E,$31,$F0,$DE
	.dc.b	$84,$34,$57,$4E,$08,$E0,$93,$F3,$89,$3E,$48,$CB,$63,$33,$E6,$9C
	.dc.b	$B6,$32,$FA,$23,$2D,$8C,$BE,$A6,$BE,$76,$BE,$66,$C7,$28,$E1,$3E
	.dc.b	$32,$3E,$16,$BD,$DA,$F0,$34,$43,$45,$34,$60,$86,$08,$79,$DE,$AD
	.dc.b	$39,$14,$E2,$53,$89,$ED,$78,$B4,$E2,$53,$89,$4E,$31,$F0,$DE,$84
	.dc.b	$34,$57,$4E,$08,$E0,$93,$F3,$89,$3E,$48,$CB,$63,$33,$E6,$9C,$B6
	.dc.b	$32,$FA,$23,$2D,$8C,$BE,$A6,$BE,$76,$BE,$66,$C7,$28,$E1,$3E,$32
	.dc.b	$3E,$16,$BD,$DA,$F0,$34,$43,$45,$34,$60,$86,$08,$7F,$AA,$BC,$AF
	.dc.b	$F5,$29,$23,$FD,$44,$D8,$FE,$C2,$D7,$FD,$44,$D7,$FB,$63,$5E,$91
	.dc.b	$C8,$8F,$26,$BC,$9A,$F2,$6B,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21
	.dc.b	$A2,$9A,$30,$43,$04,$3F,$BB,$C7,$B5,$B1,$E6,$69,$E1,$29,$D8,$A7
	.dc.b	$52,$9D,$65,$3D,$5B,$3D,$5B,$3D,$5A,$FF,$86,$FB,$9E,$5F,$F8,$69
	.dc.b	$B5,$FE,$18,$6C,$FF,$B6,$5A,$F1,$9E,$04,$9E,$0F,$FA,$9B,$6F,$F4
	.dc.b	$BC,$FF,$A5,$4F,$F5,$19,$0F,$F4,$69,$FE,$98,$21,$93,$81,$A3,$0F
	.dc.b	$42,$3F,$E9,$B3,$64,$73,$D0,$E8,$FF,$57,$3F,$E9,$F9,$FF,$54,$CE
	.dc.b	$A7,$27,$FA,$A4,$91,$E4,$3C,$40,$45,$9E,$D1,$93,$FD,$7D,$79,$AC
	.dc.b	$47,$1A,$C3,$BE,$05,$62,$3F,$D5,$C7,$FA,$8B,$97,$FD,$4F,$CF,$FE
	.dc.b	$B0,$8F,$F5,$3C,$7F,$AC,$79,$FF,$D5,$91,$E5,$E8,$FF,$58,$BE,$06
	.dc.b	$BB,$AF,$F5,$99,$CD,$EF,$CD,$FE,$B1,$25,$FF,$A2,$79,$BF,$D4,$FC
	.dc.b	$FF,$E9,$0A,$FF,$3F,$1F,$E7,$DC,$FF,$CF,$0F,$FA,$B4,$FF,$46,$4F
	.dc.b	$F9,$F4,$F8,$9D,$1F,$E8,$A3,$FD,$0F,$AF,$FA,$50,$FF,$40,$CF,$FA
	.dc.b	$64,$FF,$46,$5F,$FA,$4A,$32,$4A,$F1,$DF,$FA,$3A,$3F,$D3,$31,$DA
	.dc.b	$78,$3E,$9A,$F1,$9F,$EC,$5D,$7F,$D5,$75,$E1,$9D,$4F,$FA,$77,$0F
	.dc.b	$F5,$AD,$6A,$7B,$CF,$BC,$F8,$35,$E3,$7F,$09,$5E,$DC,$BF,$D3,$D1
	.dc.b	$AA,$EB,$3F,$01,$E7,$5C,$67,$FD,$55,$5C,$23,$C9,$1F,$EB,$A2,$1F
	.dc.b	$13,$1F,$9F,$7F,$ED,$18,$EC,$69,$F4,$BF,$ED,$1E,$7F,$F5,$57,$C5
	.dc.b	$FF,$A9,$C9,$97,$FA,$95,$B3,$FE,$A3,$6B,$FE,$A3,$E9,$47,$E5,$A3
	.dc.b	$F3,$63,$EA,$9E,$F6,$63,$F3,$F1,$8B,$DC,$6E,$74,$FE,$9E,$74,$F4
	.dc.b	$6B,$CA,$36,$FE,$C6,$32,$F1,$8C,$BD,$67,$BC,$EA,$38,$4E,$A3,$11
	.dc.b	$B3,$C1,$19,$24,$F2,$6C,$6A,$30,$47,$04,$3D,$0E,$5A,$64,$F4,$39
	.dc.b	$69,$93,$CE,$E5,$A6,$4F,$41,$D0,$87,$22,$9C,$0A,$70,$34,$43,$45
	.dc.b	$34,$60,$86,$09,$3F,$28,$93,$F1,$A3,$F8,$FF,$B8,$F3,$7E,$AB,$6F
	.dc.b	$F0,$6D,$9F,$C5,$B5,$E3,$3C,$09,$3C,$1B,$1C,$1A,$F0,$8D,$11,$D3
	.dc.b	$5D,$18,$21,$82,$1F,$CE,$E7,$C8,$D9,$EE,$D9,$E8,$D3,$99,$4E,$67
	.dc.b	$F4,$99,$E6,$D9,$E6,$D9,$E6,$D3,$99,$4E,$62,$79,$B6,$79,$B6,$79
	.dc.b	$B4,$E6,$53,$9C,$F2,$24,$F2,$6C,$F2,$6C,$F2,$6B,$CA,$78,$12,$78
	.dc.b	$36,$38,$35,$E1,$1A,$23,$A6,$BA,$30,$43,$04,$3E,$81,$0F,$90,$53
	.dc.b	$E0,$16,$7D,$2D,$9F,$23,$67,$C2,$D3,$C0,$53,$C0,$76,$21,$D8,$B3
	.dc.b	$D9,$B3,$D9,$B3,$D9,$A7,$62,$9D,$8E,$84,$9E,$8D,$9E,$8D,$9E,$8D
	.dc.b	$3A,$14,$E9,$3C,$89,$3C,$9B,$3C,$9B,$3C,$9A,$F2,$9E,$04,$9E,$0D
	.dc.b	$8E,$0D,$78,$46,$88,$E9,$AE,$8C,$10,$C1,$0F,$A8,$43,$E4,$94,$F8
	.dc.b	$05,$9F,$4B,$67,$C8,$D9,$F0,$B4,$F0,$14,$F0,$1D,$88,$76,$2C,$F6
	.dc.b	$6C,$F6,$6C,$F6,$69,$FD,$BB,$BD,$C5,$FD,$79,$B3,$FD,$B1,$B3,$B9
	.dc.b	$D1,$27,$4D,$8D,$46,$08,$E0,$87,$F9,$62,$7A,$B6,$79,$36,$78,$B4
	.dc.b	$E2,$7F,$98,$E7,$8B,$67,$8B,$67,$8B,$4E,$27,$F9,$BE,$78,$B6,$78
	.dc.b	$B6,$78,$B4,$E2,$27,$8B,$67,$8B,$67,$8B,$4E,$33,$C0,$93,$C1,$B3
	.dc.b	$C1,$B3,$C2,$74,$49,$FF,$13,$E8,$D1,$E1,$7F,$87,$23,$0D,$8C,$CF
	.dc.b	$06,$DF,$FA,$2A,$B7,$39,$AE,$04,$7F,$C5,$F3,$FE,$98,$3F,$D4,$87
	.dc.b	$FA,$9F,$83,$FD,$40,$FF,$A4,$63,$FD,$20,$EE,$32,$FF,$A7,$F8,$3F
	.dc.b	$C8,$1D,$1F,$EA,$92,$74,$7F,$A9,$A7,$FD,$6A,$75,$24,$7F,$A9,$1E
	.dc.b	$57,$D9,$F1,$C6,$5F,$F5,$39,$FE,$B4,$38,$CF,$FA,$D1,$FF,$59,$1C
	.dc.b	$1C,$3E,$08,$E5,$5E,$69,$F2,$4F,$AC,$E6,$7B,$CE,$D3,$C2,$BC,$97
	.dc.b	$F0,$27,$C7,$3E,$9E,$77,$43,$A3,$FD,$51,$18,$6C,$66,$BF,$D7,$07
	.dc.b	$FA,$D8,$FF,$5A,$3F,$EB,$67,$FD,$7A,$7F,$AD,$5C,$4F,$FA,$DA,$38
	.dc.b	$1F,$EB,$4D,$BF,$D8,$2F,$43,$FD,$04,$7F,$A0,$0B,$87,$FA,$21,$BC
	.dc.b	$FF,$E8,$E3,$FD,$23,$CF,$FE,$8D,$71,$1F,$E8,$7E,$86,$DF,$ED,$59
	.dc.b	$FF,$55,$3F,$E8,$E8,$FF,$60,$5F,$5E,$8F,$F4,$6B,$FE,$B0,$BF,$F5
	.dc.b	$0C,$FF,$AC,$1E,$95,$CD,$CB,$CC,$FF,$5C,$5F,$FB,$16,$32,$7A,$EF
	.dc.b	$AC,$66,$3F,$D4,$BB,$7F,$AC,$DF,$F5,$D9,$FE,$CF,$3F,$D7,$17,$EA
	.dc.b	$7F,$D9,$55,$C9,$FF,$60,$1F,$EB,$9C,$3F,$D6,$11,$FE,$B3,$8E,$6F
	.dc.b	$FA,$AE,$3D,$D3,$CD,$DE,$3F,$26,$B7,$3F,$13,$0F,$97,$1F,$40,$CE
	.dc.b	$3E,$E7,$DB,$69,$44,$FA,$A7,$89,$FE,$9A,$6E,$7F,$EA,$86,$CF,$FA
	.dc.b	$C1,$AF,$23,$94,$F1,$24,$F1,$7A,$7B,$F2,$3A,$36,$7A,$35,$E8,$29
	.dc.b	$19,$8F,$04,$62,$79,$92,$39,$B5,$E7,$1B,$72,$43,$33,$D5,$F8,$8D
	.dc.b	$7C,$4D,$8E,$51,$C2,$7B,$3E,$49,$D1,$23,$31,$D1,$AF,$46,$CF,$48
	.dc.b	$E5,$18,$9E,$51,$E7,$32,$53,$A1,$4E,$85,$9E,$8E,$8F,$41,$98,$EA
	.dc.b	$E0,$FF,$54,$FC,$EF,$FA,$A0,$95,$FE,$A8,$6C,$7F,$AA,$1A,$F0,$8D
	.dc.b	$11,$D3,$5D,$18,$21,$82,$1F,$EA,$18,$EA,$D8,$E4,$D3,$89,$4E,$27
	.dc.b	$FA,$7E,$38,$B6,$38,$B4,$E2,$53,$89,$FE,$9D,$8E,$2D,$8E,$2D,$38
	.dc.b	$94,$E3,$D2,$9E,$2D,$9E,$2D,$9E,$2D,$78,$CF,$02,$4F,$06,$C7,$06
	.dc.b	$BC,$23,$44,$74,$D7,$46,$08,$60,$87,$CB,$21,$F1,$4A,$7B,$CB,$3E
	.dc.b	$96,$CF,$91,$B3,$E1,$69,$E0,$29,$E0,$3B,$10,$EC,$59,$EC,$D9,$EC
	.dc.b	$D9,$EC,$D3,$B1,$4E,$C7,$42,$4F,$46,$CF,$46,$CF,$46,$9D,$0A,$74
	.dc.b	$9E,$44,$9E,$4D,$9E,$4D,$9E,$4D,$79,$4F,$02,$4F,$06,$C7,$06,$BC
	.dc.b	$23,$44,$74,$D7,$46,$08,$60,$87,$D4,$21,$F2,$4A,$7C,$02,$CF,$A5
	.dc.b	$B3,$E4,$6C,$F8,$5A,$78,$0A,$78,$0E,$C4,$3B,$16,$7B,$36,$7B,$36
	.dc.b	$7B,$34,$EC,$53,$B1,$D0,$93,$D1,$B3,$D1,$FC,$2F,$8E,$AF,$EA,$B1
	.dc.b	$FD,$AA,$93,$98,$CC,$62,$77,$08,$C3,$A9,$E0,$FF,$60,$8C,$B6,$32
	.dc.b	$FF,$6F,$8C,$B6,$32,$F8,$DA,$F7,$6C,$70,$8D,$4F,$92,$31,$1B,$8C
	.dc.b	$4F,$02,$38,$24,$7A,$5A,$F9,$DA,$F9,$5A,$F8,$0E,$04,$38,$14,$E0
	.dc.b	$53,$81,$A2,$1A,$29,$A3,$04,$30,$43,$DC,$75,$29,$C8,$A7,$12,$9C
	.dc.b	$67,$E3,$92,$7E,$2C,$73,$87,$B1,$F4,$09,$5F,$3D,$B3,$F8,$66,$BF
	.dc.b	$81,$7E,$CB,$FE,$89,$E5,$FF,$43,$3F,$E8,$8E,$6F,$F4,$8B,$96,$BF
	.dc.b	$E9,$53,$FD,$1D,$1F,$EA,$67,$FD,$33,$B7,$FA,$37,$7F,$F4,$B4,$60
	.dc.b	$FF,$4A,$9F,$EA,$63,$B1,$CA,$BF,$D3,$6F,$53,$55,$D5,$FF,$51,$47
	.dc.b	$8A,$BB,$61,$FE,$9C,$9E,$A7,$FA,$72,$79,$CE,$A3,$B9,$D2,$32,$F0
	.dc.b	$BF,$3B,$FE,$4F,$9F,$F6,$3C,$63,$93,$FD,$45,$3F,$E9,$56,$F1,$39
	.dc.b	$3F,$D3,$2D,$9F,$F5,$1F,$37,$ED,$2B,$FD,$41,$1F,$EA,$43,$FC,$B8
	.dc.b	$E0,$E7,$5C,$8C,$F3,$7F,$AB,$EF,$A5,$7F,$9C,$4F,$F4,$1E,$1F,$E9
	.dc.b	$17,$FD,$12,$7F,$A0,$79,$3F,$D3,$AF,$FA,$3F,$97,$FC,$FD,$18,$CF
	.dc.b	$FD,$03,$5C,$79,$95,$FE,$9D,$BE,$D1,$FE,$89,$7B,$9F,$E9,$08,$E6
	.dc.b	$FF,$A0,$2F,$FD,$0B,$3F,$E9,$38,$FF,$4C,$3C,$49,$87,$FA,$36,$39
	.dc.b	$E1,$FE,$90,$DB,$FD,$41,$18,$BF,$31,$D3,$0E,$F1,$EE,$AF,$F5,$A3
	.dc.b	$67,$FD,$55,$87,$AB,$1F,$F5,$94,$FC,$33,$FD,$2E,$53,$B1,$E9,$33
	.dc.b	$5F,$EA,$8B,$F8,$55,$EB,$3F,$D5,$17,$E3,$BF,$91,$3A,$3C,$B3,$E1
	.dc.b	$BF,$A1,$3A,$3C,$77,$D5,$B7,$FE,$A6,$F0,$5D,$CF,$85,$FE,$5F,$18
	.dc.b	$6C,$67,$3F,$E8,$B3,$96,$C6,$71,$FF,$4E,$CE,$5B,$19,$7F,$D5,$91
	.dc.b	$97,$FD,$33,$E4,$7F,$AD,$49,$5F,$EB,$56,$C7,$FA,$D5,$AF,$08,$D1
	.dc.b	$1D,$35,$D1,$82,$18,$21,$FE,$B0,$8E,$AD,$8E,$4D,$38,$94,$E2,$7F
	.dc.b	$B3,$E3,$8B,$63,$8B,$4E,$25,$38,$9F,$EC,$38,$E2,$D8,$E2,$D3,$89
	.dc.b	$4E,$3D,$29,$E2,$D9,$E2,$D9,$E2,$D7,$8C,$F0,$24,$F0,$7F,$B7,$DB
	.dc.b	$D8,$A4,$9F,$F0,$0C,$65,$B1,$9A,$FF,$05,$CE,$5B,$19,$7F,$DC,$71
	.dc.b	$96,$C6,$5F,$F7,$94,$65,$B1,$97,$FD,$FF,$19,$6C,$65,$FC,$EB,$63
	.dc.b	$51,$89,$F3,$91,$F2,$B5,$F1,$B5,$F0,$B5,$EE,$D7,$A1,$C8,$87,$22
	.dc.b	$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$8F,$8E,$49,$F9,$44,$9F,$81,$19,$6C,$66,$7D,$73,$96,$C6
	.dc.b	$5F,$44,$65,$B1,$97,$CF,$19,$6C,$65,$F4,$46,$5B,$19,$7D,$4D,$8D
	.dc.b	$46,$27,$CE,$47,$CA,$D7,$C6,$D7,$C2,$D7,$BB,$5E,$87,$22,$1C,$8A
	.dc.b	$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1
	.dc.b	$0C,$12,$3E,$39,$27,$E5,$12,$7E,$04,$65,$B1,$99,$F5,$CE,$5B,$19
	.dc.b	$7D,$11,$96,$C6,$5F,$3C,$65,$B1,$97,$D1,$19,$6C,$65,$F5,$36,$35
	.dc.b	$18,$9F,$39,$1F,$2B,$5F,$1B,$5F,$0B,$5E,$ED,$7A,$1C,$8F,$B7,$DE
	.dc.b	$F8,$AF,$F4,$CB,$67,$FD,$52,$D9,$FF,$58,$B5,$10,$39,$70,$D7,$6D
	.dc.b	$77,$3E,$02,$4F,$81,$B1,$E0,$6B,$DC,$E3,$1B,$23,$B6,$BB,$9F,$01
	.dc.b	$27,$9C,$79,$9A,$F9,$9A,$F6,$3A,$10,$E8,$EA,$3A,$4F,$82,$7A,$35
	.dc.b	$E8,$D7,$A1,$C9,$C4,$72,$8E,$C7,$22,$9C,$8A,$72,$34,$73,$32,$F3
	.dc.b	$70,$7D,$7E,$F7,$80,$EF,$FF,$6D,$C6,$1B,$19,$BF,$F7,$8C,$E5,$B1
	.dc.b	$9A,$FA,$93,$96,$C6,$5F,$03,$5F,$F4,$7F,$57,$FA,$B8,$FF,$4E,$3F
	.dc.b	$EA,$D3,$FD,$44,$7F,$A9,$DD,$75,$70,$3F,$D4,$11,$FE,$97,$BF,$F5
	.dc.b	$04,$EA,$38,$4E,$A3,$11,$FE,$A8,$3A,$94,$FF,$51,$1C,$0C,$9E,$33
	.dc.b	$26,$CF,$F5,$61,$B3,$26,$CC,$9B,$E2,$FF,$55,$B9,$6B,$FE,$BD,$36
	.dc.b	$66,$BF,$5B,$2A,$FF,$51,$4F,$FA,$A8,$FF,$58,$73,$FF,$A8,$FA,$3F
	.dc.b	$D4,$BD,$5F,$EA,$D8,$34,$93,$FA,$91,$D5,$FE,$AE,$7A,$75,$7F,$AB
	.dc.b	$23,$11,$AA,$EC,$FF,$A1,$4F,$F4,$71,$7A,$7F,$D0,$9D,$5F,$E8,$02
	.dc.b	$6B,$FE,$7C,$9D,$47,$FA,$16,$3F,$D0,$51,$FE,$84,$3F,$D5,$4E,$3A
	.dc.b	$BF,$D6,$71,$82,$3F,$EA,$57,$FD,$31,$5F,$E8,$D8,$D9,$93,$FD,$46
	.dc.b	$E7,$AB,$FD,$2D,$3C,$9F,$F5,$5B,$FE,$9E,$6B,$EC,$3F,$D6,$F9,$7F
	.dc.b	$AD,$67,$FD,$6F,$BF,$FA,$D1,$B3,$E2,$C7,$D8,$F5,$3E,$25,$71,$25
	.dc.b	$71,$6C,$71,$C7,$E3,$3F,$EC,$E8,$CE,$2C,$7C,$FC,$3F,$B8,$F8,$2F
	.dc.b	$DA,$93,$1F,$D9,$B7,$0F,$D8,$B5,$FD,$8B,$FE,$AA,$F6,$9F,$0D,$49
	.dc.b	$5F,$EA,$A8,$CB,$63,$35,$FE,$AA,$9C,$B6,$32,$FF,$AA,$A3,$2D,$8C
	.dc.b	$BF,$EB,$D8,$CB,$63,$2B,$B1,$3A,$8C,$36,$33,$3C,$1B,$5E,$72,$57
	.dc.b	$95,$AF,$8D,$AF,$85,$AF,$69,$E9,$5D,$C8,$E0,$93,$C2,$BC,$35,$DC
	.dc.b	$87,$72,$9D,$CA,$77,$8E,$55,$D9,$B5,$DA,$77,$5D,$9E,$87,$FA,$5B
	.dc.b	$E4,$BF,$D4,$64,$AF,$F5,$0B,$63,$FD,$3E,$D7,$84,$68,$8E,$9A,$E8
	.dc.b	$C1,$0C,$10,$FF,$51,$C7,$56,$C7,$26,$9C,$4A,$71,$3F,$D4,$F1,$C5
	.dc.b	$B1,$C5,$A7,$12,$9C,$47,$52,$79,$46,$1B,$19,$9E,$93,$96,$C6,$63
	.dc.b	$B1,$92,$B9,$7C,$0D,$7D,$64,$3C,$A5,$39,$1C,$27,$C6,$49,$F0,$B6
	.dc.b	$3B,$B5,$E1,$1A,$23,$A6,$BA,$30,$43,$04,$3C,$93,$E9,$7C,$E4,$3C
	.dc.b	$E6,$A7,$C4,$D9,$EE,$D3,$B1,$4E,$C7,$39,$EC,$F4,$31,$3D,$1B,$3D
	.dc.b	$1A,$74,$29,$D0,$E4,$49,$EA,$49,$E9,$19,$6C,$66,$7B,$4E,$5B,$19
	.dc.b	$8F,$01,$92,$B9,$7C,$4D,$7E,$01,$0F,$39,$4E,$47,$09,$F1,$92,$7C
	.dc.b	$2D,$8E,$ED,$78,$46,$88,$E9,$AE,$8C,$10,$C1,$0F,$24,$FA,$5F,$39
	.dc.b	$0F,$39,$A9,$F1,$36,$7B,$B4,$EC,$53,$B1,$CE,$7B,$3D,$0C,$4F,$46
	.dc.b	$CF,$46,$9D,$0A,$74,$39,$12,$7A,$92,$7F,$03,$D2,$F8,$BF,$F4,$E9
	.dc.b	$2F,$FD,$3A,$DB,$FF,$4E,$B5,$FF,$44,$35,$FF,$4D,$B5,$FF,$4D,$B5
	.dc.b	$FF,$50,$B6,$7C,$13,$D8,$93,$D9,$AF,$66,$BD,$9A,$F6,$6B,$D9,$B1
	.dc.b	$DA,$3A,$10,$E8,$53,$A1,$4E,$85,$3A,$15,$E8,$F2,$21,$C8,$A7,$22
	.dc.b	$9C,$8A,$F2,$78,$10,$E0,$53,$81,$5E,$0F,$F8,$0B,$F5,$4F,$F4,$43
	.dc.b	$7A,$D5,$FE,$5B,$D1,$5A,$6C,$EA,$BF,$D9,$11,$8A,$E1,$5B,$25,$6E
	.dc.b	$BF,$CC,$C7,$F9,$D7,$8B,$FC,$EF,$C3,$FE,$74,$E9,$FF,$46,$12,$7F
	.dc.b	$D0,$E6,$F0,$FF,$49,$C6,$63,$FD,$0A,$64,$75,$7F,$A8,$1C,$36,$BB
	.dc.b	$47,$46,$BC,$9A,$FF,$A2,$0E,$33,$FE,$A5,$78,$1F,$EA,$53,$BB,$97
	.dc.b	$91,$99,$F0,$36,$34,$F5,$9E,$F7,$F0,$A3,$47,$FB,$CE,$33,$7F,$02
	.dc.b	$3F,$D5,$BD,$1F,$E8,$A2,$74,$7F,$A2,$A3,$FD,$19,$D2,$7F,$D3,$6F
	.dc.b	$FA,$59,$FF,$7F,$4A,$7F,$D3,$71,$FE,$99,$8C,$4F,$FA,$73,$9F,$FD
	.dc.b	$31,$5D,$67,$BC,$F5,$97,$57,$F9,$FB,$AB,$FD,$09,$87,$FA,$04,$FF
	.dc.b	$4A,$E7,$FE,$72,$8F,$F4,$DB,$FE,$93,$E9,$E1,$7F,$E7,$47,$97,$57
	.dc.b	$FA,$7A,$FF,$D3,$5A,$FF,$A4,$67,$37,$C6,$7B,$1F,$E7,$DD,$7F,$D0
	.dc.b	$44,$7F,$D0,$47,$29,$EF,$19,$AF,$F4,$DD,$FF,$A3,$9E,$D5,$E8,$3F
	.dc.b	$D2,$07,$FA,$67,$1F,$6B,$DC,$EB,$1F,$EA,$6B,$EB,$18,$7F,$D4,$B1
	.dc.b	$D2,$3F,$D4,$37,$E1,$9F,$5C,$76,$AE,$A7,$FA,$8A,$BC,$B8,$7C,$9C
	.dc.b	$3E,$24,$7F,$A6,$6B,$E0,$F7,$BA,$BF,$5D,$1F,$81,$11,$9C,$BF,$62
	.dc.b	$D8,$FF,$6C,$35,$E0,$70,$9E,$6F,$E2,$E3,$4F,$56,$B9,$8D,$BE,$0E
	.dc.b	$2B,$3F,$C3,$A3,$15,$E2,$9C,$91,$C1,$23,$CB,$19,$6C,$65,$F3,$CF
	.dc.b	$52,$4E,$E3,$2D,$7D,$0D,$7D,$8D,$7C,$A7,$62,$1C,$A7,$C0,$F2,$6B
	.dc.b	$C9,$AF,$23,$81,$C8,$E0,$53,$81,$4E,$06,$8E,$24,$30,$43,$89,$92
	.dc.b	$99,$39,$98,$21,$82,$4F,$C8,$8F,$4A,$F1,$7E,$C9,$23,$F1,$2D,$9F
	.dc.b	$C4,$B5,$FA,$6D,$7F,$0C,$D7,$F1,$4D,$7F,$2A,$D9,$F0,$4F,$62,$4F
	.dc.b	$66,$BD,$9A,$F6,$6B,$D9,$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$85,$3A
	.dc.b	$14,$E8,$57,$A3,$C8,$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E
	.dc.b	$05,$78,$3A,$21,$A2,$BA,$70,$47,$04,$9F,$A8,$49,$F9,$2D,$9F,$80
	.dc.b	$D7,$D2,$D7,$C8,$D7,$C2,$D7,$C0,$D9,$F0,$4F,$62,$4F,$66,$BD,$9A
	.dc.b	$F6,$6B,$D9,$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57
	.dc.b	$A3,$C8,$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A
	.dc.b	$21,$A2,$BA,$70,$47,$04,$9F,$A8,$49,$F9,$2D,$9F,$80,$D7,$D2,$D7
	.dc.b	$C8,$D7,$C2,$D7,$C0,$D9,$F0,$4F,$62,$4F,$66,$BD,$9A,$F6,$6B,$D9
	.dc.b	$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$87,$DF,$F0,$DE,$2F,$F5,$E6,$D7
	.dc.b	$F9,$71,$B3,$FD,$A8,$38,$7F,$B6,$B5,$FE,$EC,$D7,$FC,$70,$D7,$8C
	.dc.b	$F7,$24,$F7,$6C,$77,$9E,$04,$8C,$C7,$81,$AF,$81,$AF,$81,$AF,$82
	.dc.b	$7B,$11,$ED,$1A,$9E,$D3,$99,$EC,$D7,$B3,$5E,$CD,$7B,$1D,$1C,$47
	.dc.b	$48,$C1,$0E,$A5,$3A,$94,$EA,$53,$A9,$C0,$87,$02,$9C,$0A,$7D,$8F
	.dc.b	$17,$EF,$12,$7E,$F3,$63,$E8,$35,$FB,$AF,$F8,$89,$FF,$39,$67,$FE
	.dc.b	$7A,$7F,$D0,$E7,$F9,$F4,$FF,$3F,$9F,$E3,$77,$5C,$DF,$E8,$81,$CD
	.dc.b	$FE,$9F,$9F,$F4,$41,$C0,$EA,$53,$FD,$2E,$75,$8E,$2F,$FA,$61,$AF
	.dc.b	$FA,$59,$AF,$46,$EF,$FE,$93,$6E,$1F,$E9,$19,$E1,$5E,$3A,$D3,$67
	.dc.b	$53,$CE,$30,$FC,$03,$DC,$79,$EF,$DF,$1F,$EA,$4A,$E7,$3E,$78,$E3
	.dc.b	$7F,$0D,$A6,$EF,$BD,$70,$AF,$29,$C4,$86,$08,$79,$4C,$94,$CB,$FE
	.dc.b	$DD,$AF,$8F,$CB,$FE,$A0,$9F,$F5,$59,$38,$FF,$D4,$BC,$BF,$EA,$D8
	.dc.b	$FF,$38,$4F,$FA,$A9,$FF,$56,$73,$7F,$AB,$9F,$F5,$8B,$FE,$AB,$24
	.dc.b	$FF,$AA,$F9,$B8,$C7,$FA,$EF,$97,$FD,$4F,$4D,$7F,$D0,$0F,$FA,$45
	.dc.b	$FF,$42,$BF,$E9,$D3,$FD,$81,$1F,$E8,$AC,$7F,$D5,$91,$FE,$9B,$9F
	.dc.b	$F4,$94,$FF,$A3,$B9,$7F,$D1,$7C,$DF,$E9,$63,$FD,$33,$3D,$0F,$F4
	.dc.b	$E5,$74,$9F,$F5,$34,$E8,$FF,$4C,$CF,$48,$E7,$3B,$D7,$FD,$30,$7F
	.dc.b	$AD,$D5,$FF,$A7,$6F,$89,$CE,$3C,$CD,$7F,$D4,$E7,$FA,$AC,$FF,$5F
	.dc.b	$C7,$C4,$BF,$86,$D9,$E7,$3A,$BF,$1B,$E0,$C3,$E1,$BD,$2F,$DD,$5F
	.dc.b	$06,$3D,$07,$CB,$3F,$D8,$15,$CA,$FE,$66,$DF,$ED,$1E,$F7,$B0,$8F
	.dc.b	$F5,$3C,$61,$B1,$97,$FA,$24,$65,$B1,$9B,$FF,$58,$CE,$5B,$19,$7F
	.dc.b	$D7,$91,$97,$FD,$7D,$DF,$E1,$F9,$3F,$F5,$BB,$6F,$FA,$93,$67,$FD
	.dc.b	$7A,$D7,$8C,$27,$8B,$4E,$25,$8E,$2D,$78,$C7,$02,$47,$06,$C7,$06
	.dc.b	$BC,$23,$44,$74,$D7,$46,$08,$60,$80,$4F,$B5,$C3,$5C,$C7,$A5,$CB
	.dc.b	$5C,$CF,$99,$CB,$5C,$C7,$95,$CB,$5C,$9E,$63,$25,$32,$7A,$0A,$68
	.dc.b	$C4,$F9,$C9,$1E,$56,$CF,$8D,$B1,$E1,$7D,$D3,$FF,$00,$3D,$8F,$F7
	.dc.b	$89,$27,$FE,$00,$B6,$3F,$E0,$0B,$5F,$F7,$5B,$5F,$F7,$F3,$5F,$9D
	.dc.b	$D0,$8C,$47,$79,$EA,$47,$AB,$5E,$AD,$7A,$B5,$EB,$1D,$27,$B4,$E6
	.dc.b	$7B,$1D,$08,$74,$29,$D0,$A7,$47,$94,$74,$8C,$9D,$0E,$44,$39,$14
	.dc.b	$E4,$70,$79,$39,$39,$1C,$08,$70,$34,$F0,$72,$70,$34,$61,$D3,$82
	.dc.b	$1B,$23,$A6,$CF,$C9,$24,$FC,$06,$C7,$A5,$AF,$91,$AF,$85,$AF,$82
	.dc.b	$7B,$12,$33,$1E,$09,$EC,$47,$B3,$5E,$CD,$7B,$35,$ED,$1D,$27,$B4
	.dc.b	$E6,$7B,$1D,$08,$74,$29,$D0,$A7,$47,$94,$74,$8C,$9D,$0E,$44,$39
	.dc.b	$14,$E4,$70,$79,$39,$39,$1C,$08,$70,$34,$F0,$72,$70,$34,$61,$D3
	.dc.b	$82,$1B,$23,$A6,$CF,$C9,$24,$FC,$06,$C7,$A5,$AF,$91,$AF,$85,$AF
	.dc.b	$82,$7B,$12,$33,$1E,$09,$EC,$47,$B3,$5E,$CD,$7B,$35,$ED,$1D,$27
	.dc.b	$B4,$E6,$7B,$1D,$08,$74,$29,$D0,$A7,$47,$F1,$5D,$CF,$98,$FE,$F1
	.dc.b	$3F,$EB,$78,$C4,$FF,$AE,$61,$18,$75,$3C,$1F,$EA,$07,$F5,$18,$D3
	.dc.b	$83,$8C,$6D,$DC,$69,$B1,$97,$93,$5E,$33,$E3,$9E,$11,$A8,$C4,$EE
	.dc.b	$73,$3C,$0F,$3C,$E8,$8E,$08,$79,$E3,$89,$27,$A1,$27,$8C,$EE,$30
	.dc.b	$F2,$3D,$71,$A6,$99,$3D,$05,$3C,$C7,$37,$83,$A3,$06,$E7,$C9,$1A
	.dc.b	$8E,$27,$63,$64,$30,$43,$A1,$4E,$26,$08,$63,$BD,$FC,$00,$EE,$FA
	.dc.b	$D1,$86,$C6,$6B,$FE,$00,$4E,$5B,$19,$AF,$C2,$4E,$5B,$19,$7F,$17
	.dc.b	$19,$7F,$25,$C5,$FE,$A6,$7F,$D7,$9D,$5F,$EA,$26,$9F,$E9,$AE,$5F
	.dc.b	$F5,$13,$FE,$9B,$23,$82,$61,$FE,$A8,$24,$7F,$AD,$4F,$F6,$5F,$57
	.dc.b	$FB,$01,$C1,$FE,$CC,$3F,$D7,$2E,$4F,$F5,$A3,$E1,$8E,$71,$88,$FF
	.dc.b	$55,$C6,$2B,$FD,$71,$1D,$2F,$D8,$F4,$8F,$63,$4E,$86,$CF,$F6,$41
	.dc.b	$2B,$C5,$1D,$1F,$04,$7C,$0A,$F0,$3F,$18,$F1,$9C,$E3,$8B,$93,$81
	.dc.b	$99,$F1,$BF,$11,$FD,$E5,$FF,$B1,$5F,$F5,$BF,$4F,$FA,$E6,$7F,$CA
	.dc.b	$5D,$5F,$EC,$7E,$8F,$F6,$2C,$7F,$B2,$0F,$F6,$0C,$FF,$B1,$DD,$F4
	.dc.b	$FF,$B1,$5E,$67,$FB,$37,$AA,$D7,$20,$FF,$A7,$A3,$FD,$39,$9F,$FA
	.dc.b	$B7,$A3,$FD,$45,$1F,$EA,$A3,$FD,$45,$9E,$6F,$FD,$43,$3F,$EC,$18
	.dc.b	$C5,$71,$E9,$FF,$4F,$56,$9F,$F5,$01,$FE,$C0,$BF,$F5,$2E,$DF,$EC
	.dc.b	$47,$FD,$4B,$7C,$4F,$F5,$FB,$FE,$D7,$EA,$FF,$5D,$63,$DE,$30,$72
	.dc.b	$8F,$F5,$83,$C8,$C9,$FE,$DD,$7D,$D1,$FE,$B0,$8F,$F6,$11,$F0,$0A
	.dc.b	$FF,$B2,$DF,$80,$73,$AF,$F6,$2D,$FB,$31,$F3,$4F,$6A,$E9,$5F,$EC
	.dc.b	$28,$F8,$E6,$6B,$E0,$3C,$EF,$C7,$8F,$C5,$C7,$9D,$FC,$3C,$39,$D7
	.dc.b	$A3,$05,$7F,$A5,$6B,$FD,$AF,$7F,$D4,$1F,$2C,$66,$BF,$D3,$5D,$3F
	.dc.b	$EC,$CF,$91,$FF,$64,$93,$3F,$F6,$3B,$73,$FF,$62,$B5,$FF,$62,$F5
	.dc.b	$A3,$FA,$57,$D5,$BD,$9A,$4B,$FE,$BD,$19,$6C,$66,$FF,$AD,$09,$D3
	.dc.b	$86,$BF,$D7,$DA,$FF,$64,$6B,$FE,$B5,$6B,$CE,$3B,$3C,$5A,$F1,$6B
	.dc.b	$C5,$AF,$19,$F3,$91,$F2,$46,$A7,$B9,$1C,$12,$78,$4F,$85,$AF,$85
	.dc.b	$AF,$85,$AF,$84,$EE,$E2,$7B,$36,$7B,$4E,$88,$77,$29,$DC,$A7,$72
	.dc.b	$9D,$CF,$F0,$67,$D6,$3E,$6B,$FC,$0C,$DA,$FF,$0C,$36,$7E,$CB,$5E
	.dc.b	$33,$C0,$93,$C1,$B1,$C1,$AF,$08,$D1,$1D,$35,$D1,$82,$18,$21,$FE
	.dc.b	$28,$9F,$23,$67,$BB,$67,$A3,$4E,$65,$39,$9F,$F0,$06,$79,$B6,$79
	.dc.b	$B6,$79,$B4,$E6,$53,$98,$9E,$6D,$9E,$6D,$9E,$6D,$39,$94,$E7,$3C
	.dc.b	$89,$3C,$9B,$3C,$9B,$3C,$9A,$F2,$9E,$04,$9E,$0D,$8E,$0D,$78,$46
	.dc.b	$88,$E9,$AE,$8C,$10,$C1,$0F,$A0,$43,$E4,$14,$F8,$05,$9F,$4B,$67
	.dc.b	$C8,$D9,$F0,$B4,$F0,$14,$F0,$1D,$88,$76,$2C,$F6,$6C,$F6,$6C,$F6
	.dc.b	$69,$D8,$A7,$63,$A1,$27,$A3,$67,$A3,$67,$A3,$4E,$85,$3A,$4F,$22
	.dc.b	$4F,$26,$CF,$26,$CF,$26,$BC,$A7,$81,$27,$83,$63,$83,$5E,$11,$A2
	.dc.b	$3A,$6B,$A3,$04,$30,$43,$EA,$10,$F9,$25,$3E,$01,$67,$D2,$D9,$F2
	.dc.b	$36,$7C,$2D,$3C,$05,$3C,$07,$62,$1D,$8B,$3D,$9B,$3D,$9B,$3D,$9A
	.dc.b	$7F,$AF,$FE,$47,$FD,$50,$48,$FF,$57,$36,$35,$18,$23,$82,$1F,$EA
	.dc.b	$98,$E0,$D8,$DB,$4D,$9F,$EC,$28,$DB,$63,$6D,$36,$7F,$B3,$63,$6D
	.dc.b	$8D,$B4,$D9,$FE,$D8,$8D,$B6,$36,$D3,$7D,$69,$DB,$67,$6D,$9D,$CE
	.dc.b	$89,$3A,$6C,$6A,$30,$47,$D7,$A3,$47,$07,$FB,$86,$30,$D8,$CD,$7F
	.dc.b	$84,$A7,$2D,$8C,$D7,$FB,$BE,$72,$D8,$CB,$FE,$24,$AF,$F7,$63,$FE
	.dc.b	$7B,$3F,$D1,$BD,$1F,$E8,$8E,$0F,$F5,$2C,$7F,$A0,$4C,$F4,$7F,$A3
	.dc.b	$48,$FF,$A1,$E3,$73,$C7,$1F,$F4,$7C,$FF,$A6,$4F,$F4,$7C,$65,$E8
	.dc.b	$62,$B9,$BB,$3F,$D2,$31,$B2,$46,$A3,$D0,$71,$3C,$53,$E0,$72,$72
	.dc.b	$8E,$0F,$FA,$E7,$A3,$FD,$29,$CD,$FE,$8C,$24,$7F,$A4,$BA,$3F,$D2
	.dc.b	$8A,$5D,$1F,$E9,$58,$D1,$FE,$44,$8E,$6E,$33,$FF,$4B,$5F,$F9,$9B
	.dc.b	$1F,$F3,$2F,$2F,$FA,$0B,$9F,$FD,$17,$AF,$FA,$1B,$9F,$FC,$C3,$8F
	.dc.b	$F9,$CF,$1F,$F3,$63,$FE,$71,$9F,$F4,$AB,$FE,$7E,$8C,$CF,$02,$5F
	.dc.b	$FA,$2B,$A3,$FC,$E3,$5F,$E8,$26,$A7,$FD,$25,$3C,$23,$FC,$FE,$D7
	.dc.b	$FD,$18,$FF,$A0,$DB,$3F,$E7,$BB,$F4,$3E,$B7,$DB,$1C,$A3,$0F,$A4
	.dc.b	$93,$FE,$85,$C3,$FD,$17,$7E,$43,$C7,$5C,$27,$C5,$19,$9F,$8F,$5F
	.dc.b	$17,$1E,$EF,$FA,$4C,$F7,$63,$F1,$30,$FA,$11,$F1,$E7,$77,$CF,$1B
	.dc.b	$7F,$E9,$1B,$5F,$FA,$CC,$FF,$4B,$7A,$4C,$3F,$D0,$30,$C9,$39,$8F
	.dc.b	$F4,$2C,$62,$76,$D8,$CC,$ED,$EE,$76,$9D,$C6,$63,$33,$C5,$FF,$46
	.dc.b	$7B,$9F,$E9,$A2,$57,$FA,$61,$B1,$FE,$93,$6B,$C2,$34,$47,$4D,$74
	.dc.b	$60,$86,$08,$7F,$A6,$E3,$AB,$63,$93,$4E,$25,$38,$9F,$EA,$18,$E2
	.dc.b	$D8,$E2,$D3,$89,$4E,$27,$FA,$C6,$38,$B6,$38,$B4,$E2,$53,$8F,$4A
	.dc.b	$78,$B6,$78,$B6,$78,$B5,$E3,$3C,$09,$3C,$1F,$B1,$1F,$8A,$8F,$E0
	.dc.b	$E7,$E2,$0A,$FF,$B2,$9A,$FE,$08,$FC,$76,$FC,$9B,$1C,$63,$0F,$08
	.dc.b	$FC,$4B,$5F,$04,$68,$91,$C4,$91,$E1,$8C,$B6,$32,$F8,$DE,$8F,$88
	.dc.b	$D1,$0D,$47,$26,$C7,$27,$CB,$19,$6C,$65,$F3,$B5,$EC,$72,$3B,$1C
	.dc.b	$0A,$70,$29,$C0,$D1,$C4,$86,$08,$71,$32,$53,$27,$33,$04,$30,$47
	.dc.b	$E2,$92,$3D,$D1,$B8,$C1,$ED,$2B,$E8,$6B,$E2,$3A,$47,$26,$C7,$18
	.dc.b	$C3,$C2,$3C,$ED,$7C,$11,$A2,$47,$12,$47,$86,$32,$D8,$CB,$E3,$7A
	.dc.b	$3E,$23,$44,$35,$1C,$9B,$1C,$9F,$2C,$65,$B1,$97,$CE,$D7,$B1,$C8
	.dc.b	$EC,$70,$29,$C0,$A7,$03,$47,$12,$18,$21,$C4,$C9,$4C,$9C,$CC,$10
	.dc.b	$C1,$1F,$8A,$48,$F7,$46,$E3,$07,$B4,$AF,$A1,$AF,$88,$E9,$1C,$9B
	.dc.b	$1C,$63,$0F,$08,$F3,$B5,$F0,$46,$89,$1C,$49,$1E,$18,$CB,$63,$2F
	.dc.b	$8D,$E8,$F8,$8D,$10,$D4,$72,$6C,$72,$7C,$B1,$96,$C6,$5F,$3B,$5E
	.dc.b	$C7,$23,$B1,$FD,$9B,$A1,$F0,$BF,$C9,$5B,$3F,$DD,$1B,$3F,$CA,$C3
	.dc.b	$87,$F9,$73,$5E,$93,$CC,$93,$CD,$B1,$CE,$78,$3F,$E0,$66,$BD,$A3
	.dc.b	$6F,$66,$BD,$A7,$A1,$1E,$91,$A7,$A3,$5E,$91,$B7,$A3,$5E,$87,$27
	.dc.b	$0F,$26,$BC,$8D,$10,$E6,$53,$99,$82,$18,$21,$D4,$E0,$53,$65,$36
	.dc.b	$78,$4C,$94,$CC,$FC,$A2,$4F,$C1,$5F,$1D,$F7,$C9,$5F,$7D,$B5,$F5
	.dc.b	$5A,$FD,$E7,$FD,$3A,$FF,$8D,$C7,$2B,$97,$FB,$87,$07,$FA,$61,$FF
	.dc.b	$42,$EB,$FE,$42,$9D,$57,$FA,$80,$FF,$46,$9F,$E9,$83,$FD,$18,$61
	.dc.b	$FF,$4A,$B6,$79,$BE,$07,$BE,$1F,$E9,$58,$F2,$4F,$67,$FD,$34,$7F
	.dc.b	$A8,$6F,$C8,$F3,$21,$99,$E5,$7E,$7B,$EB,$3E,$5A,$E2,$EC,$C5,$70
	.dc.b	$7C,$87,$B4,$D5,$FA,$9B,$5E,$A8,$F3,$47,$C0,$32,$79,$0F,$F5,$7D
	.dc.b	$78,$1B,$5D,$E7,$8C,$E6,$7E,$0D,$7A,$27,$C1,$C9,$FE,$AA,$9F,$F5
	.dc.b	$86,$8E,$5F,$F3,$AF,$17,$FA,$B2,$7F,$D6,$27,$F8,$A6,$BF,$D6,$B3
	.dc.b	$FE,$B8,$6C,$7F,$AD,$6B,$93,$FE,$EE,$E5,$FF,$5F,$BD,$AB,$AC,$63
	.dc.b	$93,$FD,$27,$3F,$E9,$B3,$FD,$08,$7F,$A8,$8F,$F5,$67,$2F,$FA,$68
	.dc.b	$9A,$7F,$A9,$23,$73,$FE,$F4,$DB,$FD,$54,$DE,$5F,$F5,$54,$FF,$A6
	.dc.b	$CF,$F4,$B1,$27,$7A,$7F,$A6,$E3,$FD,$3B,$BF,$FA,$6B,$1E,$F1,$C6
	.dc.b	$B7,$39,$C7,$C5,$3F,$EA,$53,$FD,$81,$48,$FF,$65,$D7,$09,$F0,$61
	.dc.b	$E4,$8F,$F5,$B4,$F2,$C7,$FD,$51,$5F,$EB,$AA,$FF,$5B,$1F,$EA,$EA
	.dc.b	$E4,$7F,$B2,$F9,$DC,$EC,$7F,$CC,$B1,$86,$C6,$72,$FE,$71,$39,$6C
	.dc.b	$67,$0F,$F3,$34,$E5,$B1,$97,$FD,$57,$19,$7F,$CD,$9C,$CF,$27,$FC
	.dc.b	$D8,$DB,$FF,$5D,$B6,$7F,$A4,$35,$E3,$3C,$09,$3C,$1B,$1C,$1A,$F0
	.dc.b	$8D,$11,$D3,$5D,$18,$21,$82,$1F,$EC,$69,$F2,$36,$7B,$B6,$7A,$34
	.dc.b	$E6,$53,$99,$FE,$DE,$9E,$6D,$9E,$6D,$9E,$6D,$39,$94,$E6,$27,$9B
	.dc.b	$67,$9B,$67,$9B,$4E,$65,$39,$CF,$22,$4F,$26,$CF,$27,$53,$F7,$09
	.dc.b	$1F,$71,$A7,$DC,$29,$F8,$42,$9F,$34,$A7,$C5,$29,$ED,$2C,$7F,$AF
	.dc.b	$5E,$91,$DC,$8E,$08,$77,$72,$D3,$27,$85,$CB,$4C,$9E,$37,$2D,$32
	.dc.b	$79,$5C,$B4,$CC,$79,$DF,$01,$0E,$C5,$3A,$14,$E4,$57,$93,$C0,$87
	.dc.b	$02,$9C,$0A,$F0,$74,$43,$45,$74,$E0,$8E,$08,$FD,$22,$47,$C0,$8C
	.dc.b	$B6,$32,$7C,$22,$9E,$B2,$9E,$62,$9E,$32,$9E,$22,$C7,$67,$A4,$77
	.dc.b	$23,$82,$1D,$DC,$B4,$C9,$E1,$72,$D3,$27,$8D,$CB,$4C,$9E,$57,$2D
	.dc.b	$33,$1E,$77,$C0,$43,$B1,$4E,$85,$39,$15,$E4,$F0,$21,$C0,$A7,$02
	.dc.b	$BC,$1D,$10,$D1,$5D,$38,$23,$82,$3F,$48,$91,$F0,$23,$2D,$8C,$9F
	.dc.b	$08,$A7,$AC,$A7,$98,$A7,$8C,$A7,$88,$B1,$D9,$E9,$1D,$C8,$E0,$87
	.dc.b	$77,$2D,$32,$78,$5C,$B4,$C9,$E3,$72,$D3,$27,$95,$FF,$50,$2E,$AF
	.dc.b	$F3,$4C,$61,$B1,$FE,$A5,$6A,$21,$C7,$95,$3C,$5B,$1C,$5A,$F0,$38
	.dc.b	$46,$C8,$ED,$AE,$E7,$C2,$F8,$1A,$F8,$DA,$F4,$39,$10,$E5,$3E,$27
	.dc.b	$93,$5E,$4D,$79,$1C,$27,$C6,$F0,$6B,$C1,$AF,$03,$47,$84,$A7,$33
	.dc.b	$07,$12,$9C,$4C,$1B,$21,$C8,$86,$8A,$71,$24,$FC,$42,$4F,$B5,$B3
	.dc.b	$E8,$6C,$F8,$9A,$FF,$89,$B1,$7A,$9F,$E3,$22,$57,$F8,$C1,$B3,$FE
	.dc.b	$2E,$6B,$C5,$FF,$56,$E9,$FE,$91,$8F,$F4,$51,$FE,$80,$3F,$D1,$E4
	.dc.b	$E5,$FF,$46,$39,$6B,$FE,$82,$38,$94,$FF,$51,$9B,$E6,$FF,$49,$1F
	.dc.b	$E8,$72,$BC,$27,$C4,$F0,$39,$1F,$E9,$73,$91,$FE,$97,$33,$C3,$FE
	.dc.b	$93,$8E,$31,$B3,$38,$7F,$A6,$6B,$C1,$1C,$67,$75,$C6,$3F,$D5,$67
	.dc.b	$86,$33,$5C,$63,$CC,$68,$ED,$1A,$7E,$35,$7B,$1E,$46,$EB,$FD,$77
	.dc.b	$CD,$FE,$A5,$27,$37,$FA,$72,$7F,$D4,$4D,$7F,$77,$1F,$EA,$27,$3C
	.dc.b	$DF,$C1,$A3,$8B,$FE,$A0,$8F,$E0,$51,$FE,$A8,$E5,$73,$7F,$9F,$79
	.dc.b	$7F,$D0,$87,$F0,$6E,$6F,$F4,$D5,$FF,$A8,$B2,$FF,$43,$73,$7F,$A5
	.dc.b	$35,$FF,$44,$CF,$FA,$56,$F8,$BF,$EA,$B3,$FD,$29,$7F,$EA,$E7,$FD
	.dc.b	$58,$7F,$A5,$AF,$F8,$A5,$2B,$C5,$3D,$CE,$85,$8E,$E4,$8F,$36,$3F
	.dc.b	$EA,$99,$E8,$7F,$A6,$4F,$15,$FF,$A7,$2B,$B4,$7F,$AB,$8F,$F5,$19
	.dc.b	$D6,$B9,$ED,$FE,$AD,$8F,$09,$D2,$78,$B6,$7F,$D4,$AF,$38,$FF,$52
	.dc.b	$C7,$C2,$8F,$7D,$7C,$13,$C5,$1E,$E8,$E4,$71,$78,$47,$D0,$8C,$91
	.dc.b	$E2,$EC,$FE,$78,$FF,$AC,$70,$F9,$73,$FC,$FF,$9B,$FD,$6B,$E3,$FF
	.dc.b	$AC,$89,$8F,$FA,$C1,$B8,$7F,$AB,$9A,$FF,$09,$7F,$97,$78,$CF,$9E
	.dc.b	$79,$58,$6D,$FF,$45,$B5,$6E,$B4,$49,$D4,$65,$B5,$CA,$BF,$A7,$4E
	.dc.b	$5B,$19,$7F,$AC,$C6,$5B,$19,$7F,$AF,$46,$5B,$19,$7C,$AD,$8D,$46
	.dc.b	$2B,$CC,$F8,$1A,$F6,$6B,$D1,$AF,$1A,$F5,$C6,$27,$A9,$2B,$C0,$48
	.dc.b	$D3,$67,$89,$2B,$C4,$D7,$C4,$D7,$C4,$D7,$C4,$70,$21,$C0,$A7,$02
	.dc.b	$9C,$0D,$10,$D1,$4D,$18,$3F,$DB,$1F,$23,$FE,$F0,$24,$7F,$BC,$1B
	.dc.b	$1F,$EF,$06,$BF,$EF,$86,$BC,$A3,$81,$1E,$0D,$78,$35,$E0,$68,$86
	.dc.b	$8A,$68,$C1,$0C,$10,$FF,$74,$47,$91,$B1,$DD,$A7,$42,$9C,$CA,$73
	.dc.b	$3F,$DD,$11,$CD,$B1,$CD,$A7,$32,$9C,$CA,$73,$EA,$4F,$36,$CF,$36
	.dc.b	$CF,$36,$BC,$DA,$F3,$9E,$44,$9E,$4D,$8E,$4D,$79,$35,$E5,$1C,$08
	.dc.b	$F0,$6B,$C1,$AF,$03,$44,$34,$53,$46,$08,$60,$87,$D0,$21,$F2,$0B
	.dc.b	$3F,$01,$B3,$E9,$6C,$F9,$1A,$78,$4A,$78,$0A,$78,$0E,$C4,$9E,$CD
	.dc.b	$9E,$CD,$9E,$CD,$3B,$14,$EC,$53,$B4,$F4,$24,$F4,$6C,$F4,$6C,$F4
	.dc.b	$6B,$D1,$AF,$49,$E4,$49,$E4,$D8,$E4,$D7,$93,$5E,$51,$C0,$8F,$06
	.dc.b	$BC,$1A,$F0,$34,$43,$45,$34,$60,$86,$08,$7D,$42,$1F,$24,$B3,$F0
	.dc.b	$1B,$3E,$96,$CF,$91,$A7,$84,$A7,$80,$A7,$80,$EC,$4E,$97,$6F,$F9
	.dc.b	$DC,$93,$FE,$B5,$6C,$FF,$AD,$5A,$FF,$A9,$DA,$FF,$A9,$DA,$FF,$A9
	.dc.b	$DA,$FF,$B0,$5B,$3E,$09,$EC,$49,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC
	.dc.b	$D8,$ED,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$F4,$79,$10,$E4,$53
	.dc.b	$91,$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF,$05,$E2,$3C,$67,$CE,$FF
	.dc.b	$B3,$E3,$0D,$8C,$DF,$F8,$86,$72,$D8,$CD,$7F,$BA,$67,$2D,$8C,$BF
	.dc.b	$EF,$48,$CB,$FE,$FE,$E3,$FF,$44,$75,$FF,$AA,$9B,$D7,$FE,$89,$9F
	.dc.b	$F4,$69,$FE,$89,$C3,$FD,$24,$FF,$A3,$29,$3F,$E9,$C8,$FF,$80,$35
	.dc.b	$C9,$FF,$4F,$1C,$4F,$F4,$A9,$89,$E2,$FF,$A7,$1A,$F5,$9E,$8F,$FA
	.dc.b	$C0,$EA,$73,$8F,$43,$63,$D1,$C9,$FE,$AB,$9E,$73,$EA,$6C,$66,$7A
	.dc.b	$CF,$FA,$D8,$F2,$9D,$A7,$77,$F0,$23,$94,$F0,$BF,$74,$6E,$30,$F8
	.dc.b	$CF,$73,$8A,$E7,$1D,$8F,$61,$0F,$13,$E8,$6C,$7F,$98,$67,$BF,$4F
	.dc.b	$FA,$C2,$5C,$FF,$EA,$E7,$FD,$63,$D6,$EB,$FF,$5A,$42,$BF,$D7,$4F
	.dc.b	$FA,$F2,$7F,$D7,$CF,$09,$FF,$5E,$46,$23,$83,$CD,$EF,$D7,$C6,$7A
	.dc.b	$D7,$FA,$FE,$31,$3A,$3F,$D1,$47,$FA,$37,$AD,$AF,$FA,$2A,$3F,$D1
	.dc.b	$B1,$FE,$BF,$EA,$FF,$52,$C7,$FA,$A0,$E1,$3B,$7F,$D7,$F3,$C9,$FF
	.dc.b	$59,$1F,$EA,$27,$19,$FF,$AA,$BA,$FF,$D4,$9A,$7F,$AE,$5B,$BF,$FA
	.dc.b	$C6,$7F,$D7,$D7,$FE,$BA,$BF,$13,$70,$FF,$69,$91,$FF,$5B,$DF,$FA
	.dc.b	$BE,$77,$1E,$19,$F3,$4F,$68,$F5,$3B,$8E,$55,$E0,$9E,$07,$AA,$7F
	.dc.b	$D6,$2F,$C0,$3B,$D7,$B3,$1F,$F5,$E4,$F3,$9C,$DF,$82,$B5,$1F,$32
	.dc.b	$3B,$D7,$CC,$BF,$F6,$4C,$FF,$B4,$CF,$F5,$EE,$2C,$79,$5F,$CF,$A7
	.dc.b	$C0,$E6,$FF,$24,$D4,$7F,$00,$70,$DA,$FE,$1D,$18,$8F,$F7,$9B,$54
	.dc.b	$F6,$9D,$9F,$C5,$7C,$3D,$41,$A4,$8D,$12,$7F,$4D,$19,$6C,$65,$F3
	.dc.b	$9E,$89,$E9,$19,$8F,$E3,$33,$C5,$D9,$E6,$70,$EA,$38,$3C,$E3,$2D
	.dc.b	$8C,$BC,$5B,$3E,$08,$F2,$90,$EE,$E5,$B3,$D1,$E4,$6C,$F1,$14,$F0
	.dc.b	$14,$D9,$CA,$7C,$31,$CE,$7B,$4F,$23,$47,$17,$2F,$43,$24,$38,$10
	.dc.b	$D1,$4F,$09,$5E,$6B,$A1,$E1,$FE,$0D,$B7,$FE,$FF,$9D,$4E,$49,$19
	.dc.b	$9D,$CF,$D6,$37,$19,$9D,$C6,$67,$6F,$DC,$6B,$F8,$46,$BF,$9C,$6B
	.dc.b	$D0,$E5,$1D,$1E,$4D,$79,$35,$E4,$D7,$91,$C2,$7C,$C4,$7A,$B5,$D1
	.dc.b	$82,$18,$21,$E0,$78,$34,$D9,$4D,$9E,$47,$6D,$36,$53,$67,$A0,$E4
	.dc.b	$43,$45,$34,$60,$86,$09,$3F,$30,$93,$F1,$5B,$3E,$49,$DC,$E4,$91
	.dc.b	$99,$DC,$FA,$0D,$C6,$67,$71,$99,$DB,$EC,$6B,$E9,$6B,$E8,$6B,$D0
	.dc.b	$E5,$1D,$1E,$4D,$79,$35,$E4,$D7,$91,$C2,$7C,$C4,$7A,$B5,$D1,$82
	.dc.b	$18,$21,$E0,$78,$34,$D9,$4D,$9E,$47,$6D,$36,$53,$67,$A0,$E4,$43
	.dc.b	$45,$34,$60,$86,$09,$3F,$30,$93,$F1,$5B,$3E,$49,$DC,$E4,$91,$99
	.dc.b	$DC,$FA,$0D,$C6,$67,$71,$99,$DB,$EC,$6B,$E9,$6B,$E8,$6B,$D0,$E5
	.dc.b	$1D,$1E,$4D,$79,$35,$E4,$D7,$91,$C2,$7C,$C4,$7A,$B5,$D1,$82,$18
	.dc.b	$21,$E0,$7E,$40,$E1,$FE,$DC,$D3,$F9,$61,$63,$F9,$43,$5F,$EB,$ED
	.dc.b	$7F,$B3,$B5,$FE,$FE,$D7,$FC,$11,$1D,$89,$1D,$9B,$1D,$9A,$F6,$6B
	.dc.b	$D9,$AF,$66,$BD,$A3,$A1,$1E,$8D,$7A,$35,$E8,$D7,$A3,$5E,$87,$22
	.dc.b	$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A
	.dc.b	$68,$C1,$0C,$0F,$0F,$FD,$9C,$DE,$9F,$F1,$74,$60,$8E,$09,$3F,$DB
	.dc.b	$E3,$8C,$FF,$A4,$F1,$FF,$4E,$B4,$FF,$4E,$1F,$E9,$F3,$FD,$71,$C9
	.dc.b	$FE,$9D,$23,$82,$1F,$EB,$B2,$9F,$EA,$A2,$EB,$FE,$B3,$6D,$FF,$AB
	.dc.b	$E3,$11,$C4,$87,$12,$BA,$70,$47,$04,$8E,$84,$8F,$F5,$91,$FE,$AE
	.dc.b	$3F,$D6,$E7,$88,$CD,$7C,$27,$47,$C1,$3C,$31,$FE,$62,$E5,$FF,$5C
	.dc.b	$4F,$FA,$DB,$2F,$F5,$46,$FF,$EB,$5B,$FF,$55,$F2,$FF,$AA,$27,$3C
	.dc.b	$9F,$EA,$8A,$D7,$2F,$FA,$DA,$FF,$D6,$87,$F8,$75,$C3,$C8,$C5,$72
	.dc.b	$77,$3C,$A3,$11,$B8,$F1,$72,$FF,$A5,$71,$FF,$41,$CF,$FA,$3D,$FF
	.dc.b	$54,$72,$FF,$A7,$27,$FD,$9B,$A7,$FA,$9F,$97,$FD,$14,$7F,$9A,$A3
	.dc.b	$FD,$25,$5F,$E9,$B7,$77,$FE,$A0,$BF,$F4,$9B,$D6,$B8,$BF,$E9,$13
	.dc.b	$FD,$43,$5E,$0D,$BF,$D2,$31,$FE,$9C,$3F,$D4,$95,$E3,$C7,$FD,$74
	.dc.b	$F4,$C3,$FD,$4E,$7A,$F0,$F2,$46,$49,$7F,$EA,$2B,$FF,$5E,$1D,$70
	.dc.b	$F0,$3F,$EB,$1C,$3A,$CF,$8A,$7D,$91,$89,$D6,$3F,$0D,$E7,$3A,$8C
	.dc.b	$47,$B3,$0F,$F5,$BE,$1E,$28,$E2,$7F,$A8,$23,$CF,$87,$FB,$8F,$BD
	.dc.b	$E0,$BD,$CF,$F5,$44,$61,$B1,$9C,$7F,$CC,$13,$96,$C6,$70,$FF,$4C
	.dc.b	$CE,$5B,$19,$7F,$D4,$71,$97,$FD,$4B,$E0,$3E,$E9,$F1,$7F,$EB,$98
	.dc.b	$C3,$63,$38,$7F,$9C,$67,$2D,$8C,$D7,$FA,$AA,$72,$D8,$CB,$E0,$6B
	.dc.b	$FE,$CC,$8D,$B9,$AF,$01,$2B,$B3,$63,$A3,$5E,$11,$A2,$3A,$6B,$A3
	.dc.b	$04,$30,$43,$C9,$5E,$65,$5E,$67,$35,$E1,$6D,$76,$69,$D4,$A7,$53
	.dc.b	$9D,$75,$AE,$91,$8A,$E8,$DA,$E8,$D3,$A1,$4E,$93,$C8,$95,$D4,$95
	.dc.b	$D2,$32,$D8,$F7,$4F,$FB,$DB,$C3,$FF,$79,$92,$FF,$D6,$DB,$23,$13
	.dc.b	$FE,$B3,$8C,$4E,$A7,$33,$FE,$DE,$6B,$FE,$F4,$9D,$BC,$63,$2F,$FB
	.dc.b	$DD,$AF,$8E,$7A,$11,$E0,$D7,$B3,$5E,$D1,$CA,$7B,$C6,$8E,$24,$30
	.dc.b	$43,$8C,$F3,$7C,$8D,$7C,$87,$13,$65,$36,$53,$67,$53,$25,$32,$72
	.dc.b	$31,$1E,$87,$04,$38,$94,$E2,$60,$86,$08,$F3,$72,$D7,$33,$F1,$C9
	.dc.b	$3E,$F9,$D1,$23,$33,$ED,$8C,$4E,$E7,$33,$EA,$6B,$E7,$9D,$BC,$63
	.dc.b	$2F,$91,$AF,$92,$7A,$11,$E0,$D7,$B3,$5E,$D1,$CA,$7B,$C6,$8E,$24
	.dc.b	$30,$43,$8C,$F3,$7C,$8D,$7C,$87,$13,$65,$36,$53,$67,$53,$25,$32
	.dc.b	$72,$31,$1E,$87,$04,$38,$94,$E2,$60,$86,$08,$F3,$72,$D7,$33,$F1
	.dc.b	$C9,$3E,$F9,$D1,$23,$33,$ED,$8C,$4E,$E7,$33,$EA,$6B,$E7,$9D,$BC
	.dc.b	$63,$2F,$91,$AF,$92,$7A,$11,$E0,$D7,$B3,$5E,$D1,$CA,$7B,$C6,$8E
	.dc.b	$24,$30,$43,$8C,$F3,$7C,$8D,$7E,$67,$80,$F9,$E4,$A6,$12,$75,$18
	.dc.b	$8C,$4F,$07,$02,$76,$E1,$AF,$16,$BC,$A7,$8B,$C6,$7A,$3B,$6B,$B3
	.dc.b	$A3,$87,$D1,$3D,$1B,$19,$7A,$4F,$69,$D4,$60,$E7,$19,$9E,$13,$A2
	.dc.b	$38,$24,$6C,$93,$C8,$87,$A5,$CB,$4C,$9E,$B3,$04,$30,$43,$D6,$F2
	.dc.b	$70,$7A,$4E,$06,$8E,$A5,$32,$6C,$E3,$3E,$C8,$E9,$1B,$33,$3E,$E7
	.dc.b	$99,$C4,$77,$BC,$07,$C9,$7D,$38,$C3,$63,$37,$F8,$69,$CB,$63,$35
	.dc.b	$F8,$D9,$CB,$63,$2F,$E5,$23,$2F,$F7,$C3,$FD,$7E,$7F,$B0,$7A,$9D
	.dc.b	$7F,$EB,$E8,$FF,$50,$9F,$EA,$63,$67,$FA,$98,$E0,$ED,$FF,$5A,$5F
	.dc.b	$FA,$C6,$7F,$D4,$F1,$C3,$3F,$F5,$71,$1C,$12,$3C,$2F,$FA,$D5,$E2
	.dc.b	$61,$F1,$B8,$AF,$07,$1F,$FB,$06,$7C,$D3,$B7,$C6,$62,$77,$7E,$57
	.dc.b	$FC,$35,$1E,$17,$2E,$CF,$13,$A9,$E9,$5D,$DF,$14,$E8,$FF,$66,$9E
	.dc.b	$69,$E0,$74,$8D,$C6,$4D,$89,$FE,$09,$1E,$FE,$97,$4F,$FA,$F9,$BD
	.dc.b	$7F,$EC,$D2,$5B,$A7,$FD,$89,$1F,$C1,$27,$FD,$96,$4A,$46,$5E,$84
	.dc.b	$79,$CF,$FB,$2B,$A7,$FD,$39,$A7,$FA,$87,$0F,$F6,$06,$5F,$EA,$9E
	.dc.b	$9F,$F4,$EC,$7F,$AB,$DF,$F5,$1C,$66,$3F,$92,$75,$7F,$AE,$AF,$9D
	.dc.b	$7F,$B1,$A7,$FD,$90,$6A,$3F,$D8,$5B,$7F,$AB,$AB,$FD,$9F,$87,$FB
	.dc.b	$69,$FF,$69,$DF,$16,$DF,$FA,$DA,$31,$7D,$A3,$A6,$1F,$EC,$57,$FD
	.dc.b	$70,$60,$FF,$5B,$DF,$59,$E1,$0B,$F4,$57,$AE,$38,$9E,$82,$C7,$FB
	.dc.b	$55,$FF,$70,$3E,$8B,$F8,$B7,$FE,$D2,$23,$E6,$AE,$93,$DA,$BC,$73
	.dc.b	$C0,$F9,$6F,$FB,$9A,$7D,$53,$D6,$3B,$E3,$FD,$2E,$FA,$C3,$A3,$FD
	.dc.b	$9F,$E0,$7F,$B3,$09,$9F,$FB,$29,$B3,$FE,$C8,$6B,$FE,$C8,$EA,$47
	.dc.b	$F4,$3C,$9F,$1B,$49,$7F,$D4,$A3,$2D,$8C,$DF,$F5,$E9,$CB,$63,$2F
	.dc.b	$F6,$58,$CB,$63,$2F,$F6,$C8,$CB,$63,$2F,$F7,$18,$CB,$63,$2F,$EB
	.dc.b	$5B,$1A,$8C,$4F,$9C,$8F,$95,$AF,$8D,$AF,$85,$AF,$76,$BD,$0E,$44
	.dc.b	$39,$14,$E4,$53,$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14
	.dc.b	$D1,$82,$0C,$5E,$57,$F8,$0D,$B3,$FE,$C4,$8C,$11,$C1,$27,$FD,$8D
	.dc.b	$19,$6C,$65,$FF,$12,$4F,$47,$24,$33,$3C,$5B,$3C,$5F,$F6,$64,$65
	.dc.b	$B1,$97,$FC,$69,$3D,$5E,$27,$53,$65,$37,$3C,$DB,$3C,$DF,$89,$19
	.dc.b	$6C,$66,$7B,$47,$43,$B9,$0C,$10,$EA,$78,$4E,$65,$79,$BA,$39,$14
	.dc.b	$E4,$70,$32,$53,$2F,$27,$07,$12,$18,$23,$B2,$4F,$BD,$B3,$EF,$8C
	.dc.b	$11,$C1,$27,$D5,$19,$6C,$65,$F3,$CF,$47,$24,$33,$3C,$5B,$3C,$5F
	.dc.b	$6C,$65,$B1,$97,$DB,$3D,$5E,$27,$53,$65,$37,$3C,$DB,$3C,$DF,$8B
	.dc.b	$19,$6C,$66,$7B,$47,$43,$B9,$0C,$10,$EA,$78,$4E,$65,$79,$BA,$39
	.dc.b	$14,$E4,$70,$32,$53,$2F,$27,$07,$12,$18,$23,$B2,$4F,$BD,$B3,$EF
	.dc.b	$8C,$11,$C1,$27,$D5,$19,$6C,$65,$F3,$CF,$47,$24,$33,$3C,$5B,$3C
	.dc.b	$5F,$6C,$65,$B1,$97,$DB,$3D,$5E,$27,$53,$65,$37,$3C,$DB,$3C,$DF
	.dc.b	$8B,$19,$6C,$7E,$17,$A1,$E5,$FF,$A8,$89,$1F,$EA,$96,$CF,$FA,$F5
	.dc.b	$AF,$FA,$F5,$AF,$FB,$25,$AF,$FA,$D9,$AF,$FB,$1D,$B3,$E0,$9E,$C4
	.dc.b	$9E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$CD,$8E,$D1,$D0,$87,$42,$9D,$0A
	.dc.b	$74,$29,$D0,$AF,$47,$91,$0E,$45,$39,$14,$E4,$57,$93,$C0,$87,$02
	.dc.b	$9C,$0A,$F0,$7F,$C5,$19,$BC,$4F,$F7,$36,$A8,$C5,$FF,$BE,$23,$13
	.dc.b	$A9,$CD,$7F,$BF,$E7,$35,$B9,$CC,$FF,$8B,$6B,$81,$FE,$EF,$7F,$D5
	.dc.b	$67,$FA,$2C,$18,$3F,$D1,$06,$0F,$F5,$91,$32,$FF,$56,$72,$FF,$A5
	.dc.b	$AF,$FD,$3D,$3F,$E8,$ED,$3F,$D3,$1C,$DF,$E9,$48,$EC,$75,$3F,$D4
	.dc.b	$F5,$FE,$96,$6C,$6D,$F0,$9B,$9E,$6F,$5A,$EE,$F9,$6F,$C7,$5F,$EA
	.dc.b	$13,$94,$74,$38,$1F,$EA,$F3,$A9,$EB,$3F,$D6,$87,$98,$C5,$FB,$5E
	.dc.b	$07,$33,$C0,$70,$3F,$C8,$95,$F2,$A3,$FD,$59,$CB,$FE,$9F,$9F,$F5
	.dc.b	$47,$81,$F2,$F9,$FF,$D5,$B1,$83,$E8,$B8,$E3,$FF,$56,$C7,$D1,$33
	.dc.b	$5F,$EA,$C7,$02,$7F,$D6,$1C,$FF,$EA,$F3,$FD,$5B,$7D,$EB,$C5,$7F
	.dc.b	$EA,$FD,$3F,$D0,$6F,$F9,$EF,$9B,$FD,$05,$8F,$FA,$5F,$5F,$F3,$EB
	.dc.b	$FE,$90,$E7,$FF,$4B,$F3,$FF,$A0,$E3,$FD,$28,$FF,$A2,$E7,$6E,$4F
	.dc.b	$F5,$3C,$71,$C3,$BE,$DF,$EA,$FC,$7C,$13,$FE,$8C,$8C,$11,$FF,$60
	.dc.b	$9F,$E9,$33,$FD,$80,$79,$71,$F3,$BD,$4F,$F4,$D9,$CC,$E3,$87,$93
	.dc.b	$1F,$F5,$FC,$62,$7F,$D3,$F3,$9C,$3F,$D3,$F8,$79,$2B,$DB,$7C,$2B
	.dc.b	$FD,$71,$5F,$EA,$A8,$C5,$F8,$B1,$FF,$5A,$5F,$0A,$E7,$8F,$FA,$EF
	.dc.b	$89,$82,$3E,$CD,$7F,$AA,$63,$0F,$EF,$DA,$FE,$F5,$AF,$EC,$E7,$9C
	.dc.b	$F2,$6C,$71,$8C,$4F,$FA,$AF,$8A,$47,$FA,$A5,$B3,$FE,$A7,$6B,$FE
	.dc.b	$A5,$6B,$FE,$A3,$6B,$FE,$A1,$6A,$BF,$E3,$13,$D4,$91,$D5,$AF,$56
	.dc.b	$BD,$5A,$F5,$6B,$D6,$7B,$CF,$68,$E8,$43,$A1,$4E,$85,$3A,$14,$E9
	.dc.b	$3D,$A3,$A4,$72,$21,$C8,$A7,$22,$9C,$8E,$8F,$27,$81,$0E,$05,$38
	.dc.b	$1C,$9E,$0E,$88,$68,$FE,$65,$F5,$AF,$90,$FC,$1C,$7F,$38,$8C,$57
	.dc.b	$E3,$1B,$1F,$C1,$9A,$FF,$0B,$6B,$D2,$79,$A8,$CB,$D6,$33,$18,$8D
	.dc.b	$46,$09,$3E,$16,$C7,$85,$AF,$85,$AF,$86,$38,$11,$E0,$D7,$83,$5E
	.dc.b	$06,$88,$68,$A6,$8C,$10,$C1,$0F,$73,$D5,$A7,$22,$9C,$4A,$71,$3E
	.dc.b	$1B,$C5,$A7,$12,$9C,$4A,$71,$9F,$98,$49,$F8,$B1,$D8,$87,$02,$C6
	.dc.b	$E7,$C0,$D8,$F0,$35,$F0,$35,$F0,$4F,$67,$A1,$23,$4F,$68,$D4,$62
	.dc.b	$35,$18,$24,$F8,$9B,$1E,$26,$BE,$26,$BE,$28,$E0,$47,$83,$5E,$0D
	.dc.b	$78,$1A,$21,$A2,$9A,$30,$43,$04,$3D,$CF,$56,$9C,$8A,$71,$29,$C4
	.dc.b	$F8,$6F,$16,$9C,$4A,$71,$29,$C6,$7E,$61,$27,$E2,$C7,$62,$1C,$0B
	.dc.b	$1B,$9F,$03,$63,$C0,$D7,$C0,$D7,$C1,$3D,$9E,$84,$8D,$3D,$A3,$51
	.dc.b	$88,$D4,$60,$93,$E2,$6C,$78,$9A,$F8,$9A,$F8,$A3,$81,$1E,$0D,$78
	.dc.b	$35,$E0,$68,$86,$8A,$68,$C1,$0C,$1F,$67,$47,$B1,$FC,$89,$B3,$FD
	.dc.b	$D9,$B3,$FE,$30,$6B,$FE,$35,$6B,$FE,$53,$9E,$44,$9E,$4D,$8E,$4D
	.dc.b	$79,$35,$E5,$1C,$08,$F0,$6B,$C1,$AF,$03,$44,$34,$53,$46,$08,$60
	.dc.b	$87,$FA,$12,$7D,$AD,$9F,$33,$67,$C2,$D3,$B1,$4E,$A5,$3A,$89,$EA
	.dc.b	$D9,$EA,$D9,$EA,$D3,$A9,$4E,$A5,$3A,$CF,$42,$4F,$DD,$F7,$65,$7D
	.dc.b	$D6,$D7,$DD,$6B,$F7,$5A,$FF,$A3,$31,$FF,$49,$BF,$E8,$2E,$4F,$F4
	.dc.b	$31,$FE,$9B,$8C,$C7,$FA,$66,$BF,$D2,$ED,$3F,$D1,$71,$A7,$FD,$1F
	.dc.b	$CB,$FE,$99,$7F,$D2,$F5,$CE,$B8,$C6,$8F,$F4,$B6,$3F,$EA,$06,$9F
	.dc.b	$E9,$33,$FD,$2C,$77,$9E,$4F,$82,$79,$CF,$18,$FF,$51,$B5,$E2,$77
	.dc.b	$32,$75,$BF,$44,$6D,$F8,$7C,$BF,$E9,$9E,$4F,$F4,$87,$2F,$FA,$4A
	.dc.b	$FF,$D3,$2F,$FB,$3B,$93,$FD,$3B,$1F,$E9,$DA,$FF,$4C,$BF,$E9,$C6
	.dc.b	$C7,$FA,$88,$95,$D2,$7F,$D5,$07,$FA,$4A,$3F,$CF,$1C,$7F,$E9,$03
	.dc.b	$FD,$25,$C5,$FE,$98,$3F,$CE,$F3,$FE,$85,$CB,$FC,$F6,$FF,$A3,$57
	.dc.b	$27,$FB,$7E,$B8,$E7,$FE,$7F,$3F,$CF,$75,$FE,$89,$3F,$D2,$44,$3F
	.dc.b	$D2,$55,$FE,$A2,$C3,$C7,$87,$86,$B8,$CF,$80,$FF,$48,$57,$FA,$70
	.dc.b	$F5,$69,$FE,$A1,$8D,$63,$FE,$96,$24,$F7,$C7,$FD,$61,$1F,$E9,$1C
	.dc.b	$7B,$E1,$E7,$AE,$95,$FE,$94,$3E,$24,$EA,$3E,$05,$7B,$E3,$17,$E4
	.dc.b	$3E,$1C,$8F,$6C,$F7,$C3,$F8,$23,$FE,$C8,$E4,$72,$3E,$33,$FC,$AB
	.dc.b	$18,$6C,$66,$BF,$DB,$33,$96,$C6,$6F,$FD,$C1,$39,$6C,$65,$FF,$51
	.dc.b	$35,$F3,$1F,$E9,$3F,$87,$FF,$2D,$E8,$69,$2F,$FD,$63,$19,$6C,$66
	.dc.b	$7F,$DD,$B3,$96,$C6,$5F,$F7,$A4,$65,$B1,$97,$C7,$19,$6C,$65,$F9
	.dc.b	$91,$96,$C6,$5F,$54,$F9,$08,$F8,$9A,$F8,$1A,$F6,$6B,$D1,$AF,$43
	.dc.b	$91,$0E,$45,$39,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43
	.dc.b	$45,$34,$60,$86,$09,$CC,$F8,$6F,$F4,$A9,$27,$FD,$44,$D9,$FF,$51
	.dc.b	$35,$FF,$51,$35,$FF,$56,$B5,$FF,$56,$35,$FF,$52,$36,$7C,$13,$D8
	.dc.b	$93,$D9,$AF,$66,$BD,$9A,$F6,$6B,$D9,$B1,$DA,$3A,$10,$E8,$53,$A1
	.dc.b	$4E,$85,$3A,$15,$E8,$F2,$21,$C8,$A7,$22,$9C,$8A,$F2,$78,$10,$E0
	.dc.b	$53,$81,$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$27,$EA,$12,$7E,$4B,$67
	.dc.b	$E0,$35,$F4,$B5,$F2,$35,$F0,$B5,$F0,$36,$7C,$13,$D8,$93,$D9,$AF
	.dc.b	$66,$BD,$9A,$F6,$6B,$D9,$B1,$DA,$3A,$10,$E8,$53,$A1,$4E,$85,$3A
	.dc.b	$15,$E8,$F2,$21,$C8,$A7,$22,$9C,$8A,$F2,$78,$10,$E0,$53,$81,$5E
	.dc.b	$0E,$88,$68,$AE,$9C,$11,$C1,$27,$EA,$12,$7E,$4B,$67,$E0,$35,$F4
	.dc.b	$B5,$F2,$35,$F0,$B5,$F0,$36,$7C,$13,$D8,$93,$D9,$AF,$66,$BD,$9A
	.dc.b	$F6,$6B,$D9,$B1,$DA,$3A,$10,$E8,$53,$A1,$F7,$FA,$9E,$7A,$BF,$B2
	.dc.b	$CF,$F6,$69,$CC,$62,$3F,$BF,$46,$1C,$35,$E2,$D7,$94,$F5,$76,$D7
	.dc.b	$6D,$77,$3D,$D4,$F3,$7C,$33,$B6,$CE,$E7,$51,$C2,$74,$D9,$D4,$62
	.dc.b	$76,$47,$04,$8D,$92,$7D,$2D,$7D,$2D,$7D,$26,$88,$68,$A6,$8C,$10
	.dc.b	$C1,$0F,$69,$C0,$A6,$CA,$6C,$F7,$1B,$29,$B2,$9B,$9F,$81,$1E,$98
	.dc.b	$E6,$43,$66,$33,$FF,$1C,$F1,$7F,$8D,$89,$3F,$E3,$46,$CF,$F8,$C9
	.dc.b	$AF,$F8,$C9,$FA,$6F,$FA,$3F,$9B,$FD,$4D,$3F,$E8,$E8,$FF,$44,$37
	.dc.b	$9B,$FD,$4F,$3A,$9F,$F4,$A1,$23,$FD,$2E,$48,$D3,$63,$FD,$4B,$18
	.dc.b	$23,$82,$1F,$EB,$57,$FD,$58,$D3,$FD,$4F,$1E,$27,$67,$A4,$A7,$FA
	.dc.b	$84,$FF,$51,$F0,$FF,$A7,$EB,$D5,$5D,$27,$81,$D4,$FF,$53,$D7,$57
	.dc.b	$07,$23,$E1,$9E,$F8,$F4,$BA,$DB,$F5,$92,$9F,$F5,$07,$17,$FA,$8A
	.dc.b	$BF,$D5,$67,$FA,$8B,$97,$FD,$60,$E0,$FF,$56,$75,$39,$FF,$61,$3F
	.dc.b	$EA,$E9,$CC,$7F,$AD,$54,$72,$6C,$F1,$7F,$D5,$FC,$DF,$E8,$BE,$7F
	.dc.b	$F4,$E4,$7F,$A3,$96,$7F,$E8,$48,$FF,$4B,$63,$FE,$8C,$E7,$FF,$4C
	.dc.b	$C7,$FA,$78,$FF,$55,$17,$6F,$F4,$FB,$FE,$9D,$BF,$F5,$BA,$C3,$FD
	.dc.b	$33,$B7,$FA,$96,$FF,$D6,$D5,$E2,$BF,$F4,$BD,$FF,$A5,$8F,$0D,$6D
	.dc.b	$FF,$60,$57,$0C,$7F,$D4,$D5,$C3,$0F,$F5,$FD,$7F,$A9,$0F,$F5,$03
	.dc.b	$88,$FF,$52,$57,$FA,$9E,$FA,$57,$97,$0F,$14,$E6,$3D,$B7,$F1,$F0
	.dc.b	$EF,$39,$C3,$DF,$5F,$06,$3C,$8F,$84,$F5,$1D,$67,$FD,$93,$1A,$9F
	.dc.b	$1C,$F9,$B1,$F9,$CB,$83,$FD,$67,$D1,$FE,$B0,$26,$7F,$EA,$E6,$CF
	.dc.b	$FA,$B1,$AF,$FA,$B1,$FE,$65,$E3,$BA,$DD,$B8,$6D,$FF,$49,$9D,$4E
	.dc.b	$49,$19,$9D,$DF,$F5,$2A,$DC,$E6,$B7,$39,$AD,$BF,$EB,$CA,$DB,$99
	.dc.b	$DB,$99,$DB,$FE,$D7,$9D,$B9,$9D,$B9,$9D,$A9,$DB,$99,$DB,$99,$DC
	.dc.b	$EA,$31,$3A,$8C,$12,$76,$48,$D3,$6B,$E0,$12,$BD,$8D,$7D,$0D,$7C
	.dc.b	$8D,$7C,$2D,$AF,$05,$76,$24,$66,$BC,$15,$D8,$87,$62,$9F,$EE,$AE
	.dc.b	$77,$B3,$FD,$B9,$B5,$FD,$E5,$B3,$FE,$09,$6B,$FE,$0E,$6B,$CE,$79
	.dc.b	$12,$79,$36,$39,$35,$E4,$D7,$94,$70,$23,$C1,$AF,$06,$BC,$0D,$10
	.dc.b	$D1,$4D,$18,$21,$82,$01,$3E,$E7,$0D,$73,$3E,$97,$2D,$73,$3E,$67
	.dc.b	$2D,$72,$79,$4C,$94,$C9,$E6,$32,$53,$27,$A0,$A6,$8C,$4F,$9C,$93
	.dc.b	$E5,$6C,$F8,$DB,$3E,$16,$BD,$DA,$F4,$9E,$44,$9E,$4D,$8E,$4D,$79
	.dc.b	$35,$E5,$1C,$08,$F0,$6B,$C1,$AF,$03,$44,$34,$53,$46,$08,$60,$87
	.dc.b	$C7,$21,$F2,$89,$3F,$01,$CB,$5C,$CF,$AD,$CB,$5C,$CF,$A1,$CB,$5C
	.dc.b	$9E,$73,$25,$32,$7A,$0C,$94,$C9,$EA,$29,$A3,$13,$E7,$24,$F9,$5B
	.dc.b	$3E,$36,$CF,$85,$AF,$76,$BD,$27,$91,$27,$93,$63,$93,$5E,$4D,$79
	.dc.b	$47,$02,$3C,$1A,$F0,$6B,$C0,$D1,$0D,$14,$D1,$82,$18,$21,$F1,$C8
	.dc.b	$7C,$A2,$4F,$C0,$72,$D7,$33,$F7,$F9,$DF,$31,$FE,$95,$25,$FF,$A7
	.dc.b	$5B,$5F,$EB,$06,$BF,$EB,$06,$BF,$EB,$96,$BF,$EB,$86,$BF,$EA,$C6
	.dc.b	$CF,$82,$7B,$12,$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$36,$3B,$47,$42
	.dc.b	$1D,$0A,$74,$29,$D0,$A7,$42,$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E
	.dc.b	$4F,$02,$1C,$0A,$70,$2B,$C1,$FF,$16,$7E,$54,$F5,$7F,$DA,$DB,$23
	.dc.b	$17,$FE,$EB,$8C,$4E,$A7,$35,$FE,$F2,$9C,$D6,$E7,$33,$FD,$34,$EF
	.dc.b	$5C,$4F,$F5,$6B,$FE,$93,$E2,$FF,$42,$F4,$7F,$A2,$23,$FD,$22,$43
	.dc.b	$57,$FE,$96,$7F,$D3,$F1,$FE,$97,$72,$7F,$A4,$4F,$F4,$FF,$17,$FA
	.dc.b	$75,$C1,$FE,$93,$9E,$4F,$FA,$72,$79,$4E,$FA,$7F,$D6,$93,$B3,$89
	.dc.b	$E3,$8D,$4F,$9E,$3C,$33,$C1,$D1,$FE,$A4,$70,$EA,$38,$4F,$28,$F3
	.dc.b	$9B,$8C,$A9,$FF,$58,$74,$7F,$A6,$C9,$D1,$FE,$9B,$E5,$FF,$55,$57
	.dc.b	$FA,$7B,$A7,$FD,$4B,$3F,$E9,$F8,$FF,$52,$9F,$4C,$FF,$56,$1F,$EB
	.dc.b	$2E,$A8,$FF,$AC,$AB,$9F,$4F,$F9,$FE,$3F,$CF,$4F,$FA,$27,$A7,$FD
	.dc.b	$42,$E7,$4F,$F3,$CC,$FF,$9F,$A1,$D5,$FE,$A0,$7F,$D3,$F1,$FE,$8A
	.dc.b	$8F,$F5,$6A,$9F,$F4,$C4,$6C,$FF,$48,$C7,$FA,$0F,$6F,$F4,$AD,$72
	.dc.b	$AF,$F4,$7B,$E1,$BF,$1D,$F4,$9E,$7B,$FF,$A3,$B0,$FF,$53,$90,$FF
	.dc.b	$49,$1E,$CC,$7F,$D1,$EF,$FA,$64,$D6,$1D,$F0,$FF,$4B,$1D,$AB,$FD
	.dc.b	$89,$3C,$AF,$FD,$61,$7C,$AF,$C9,$3F,$EC,$1A,$F1,$3B,$BF,$F6,$4F
	.dc.b	$13,$D4,$67,$F5,$6B,$FC,$69,$39,$8C,$47,$FA,$8E,$76,$FE,$D5,$AF
	.dc.b	$EC,$DA,$FE,$C6,$79,$3B,$9E,$71,$88,$FD,$D6,$7C,$1F,$0B,$5F,$F5
	.dc.b	$93,$63,$EE,$CF,$28,$C1,$1C,$12,$7F,$85,$46,$5B,$19,$7C,$2D,$7D
	.dc.b	$11,$B7,$27,$02,$18,$21,$C2,$39,$3E,$A6,$BE,$A6,$CF,$AA,$39,$9C
	.dc.b	$4A,$71,$29,$C4,$AF,$17,$C2,$6C,$A6,$CA,$ED,$EC,$62,$7D,$CE,$08
	.dc.b	$73,$29,$CC,$B1,$CD,$D1,$06,$0E,$FF,$B4,$DB,$FC,$5B,$67,$F2,$0D
	.dc.b	$7F,$83,$35,$FE,$74,$D7,$FD,$10,$D7,$BC,$F6,$24,$F6,$6C,$76,$6B
	.dc.b	$D9,$AF,$66,$BD,$9A,$F6,$8E,$84,$7A,$35,$E8,$D7,$A3,$5E,$8D,$7A
	.dc.b	$1C,$88,$72,$29,$C8,$A7,$22,$9C,$8E,$04,$38,$14,$E0,$53,$81,$A2
	.dc.b	$1A,$29,$A3,$04,$30,$49,$FA,$84,$9F,$92,$D9,$F8,$0D,$9F,$4B,$5F
	.dc.b	$23,$5F,$0B,$5F,$03,$5F,$04,$F6,$24,$F6,$6C,$76,$6B,$D9,$AF,$66
	.dc.b	$BD,$9A,$F6,$8E,$84,$7A,$35,$E8,$D7,$A3,$5E,$8D,$7A,$1C,$88,$72
	.dc.b	$29,$C8,$A7,$22,$9C,$8E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3
	.dc.b	$04,$30,$49,$FA,$84,$9F,$92,$D9,$F8,$0D,$9F,$4B,$5F,$23,$5F,$0B
	.dc.b	$5F,$03,$5F,$04,$F6,$24,$F6,$6C,$76,$6B,$D9,$AF,$66,$BD,$9A,$F6
	.dc.b	$8E,$84,$7A,$35,$E8,$D7,$A3,$5E,$8D,$7A,$1C,$88,$7D,$5E,$87,$4F
	.dc.b	$F6,$36,$CF,$F7,$46,$CF,$F7,$96,$BF,$E2,$F6,$BF,$E6,$36,$A0,$E1
	.dc.b	$B3,$DC,$93,$DD,$B1,$DD,$AF,$76,$BD,$DA,$F7,$9E,$C4,$8C,$C7,$82
	.dc.b	$7B,$11,$EC,$D7,$B3,$5E,$CD,$7B,$47,$49,$ED,$39,$9E,$C7,$42,$1D
	.dc.b	$0A,$74,$29,$D1,$E5,$1D,$23,$27,$43,$91,$0E,$45,$39,$1C,$1F,$A1
	.dc.b	$B3,$C1,$FF,$10,$36,$BF,$C6,$2D,$AF,$F1,$B3,$5F,$F3,$63,$5F,$F5
	.dc.b	$03,$5F,$BC,$D7,$FC,$C1,$3F,$EA,$28,$FF,$41,$47,$FA,$3F,$93,$FD
	.dc.b	$0F,$18,$8F,$F4,$5C,$39,$7F,$D2,$F1,$FE,$87,$3F,$D1,$31,$97,$FD
	.dc.b	$4A,$43,$FD,$3C,$53,$FC,$0B,$7F,$EA,$26,$9F,$E9,$E3,$FD,$2C,$74
	.dc.b	$9E,$F3,$D6,$39,$90,$E2,$FA,$1F,$21,$EB,$8F,$F4,$DC,$71,$8D,$C6
	.dc.b	$0E,$A4,$7C,$D9,$FF,$B1,$1F,$11,$FE,$12,$AF,$D4,$57,$FA,$DB,$97
	.dc.b	$FD,$47,$CB,$FA,$BE,$3F,$F5,$1C,$62,$FF,$CD,$8B,$97,$FD,$49,$1F
	.dc.b	$ED,$58,$C5,$7F,$AA,$09,$0A,$E7,$5F,$EA,$FE,$5F,$F5,$3C,$FF,$A1
	.dc.b	$CF,$F3,$E1,$FE,$7F,$3F,$D3,$9C,$9F,$EE,$1E,$5F,$F5,$44,$67,$6F
	.dc.b	$F4,$3E,$9F,$E9,$CE,$4F,$F4,$FC,$FF,$AA,$5F,$F5,$16,$DF,$E8,$EB
	.dc.b	$FF,$45,$CF,$77,$FD,$28,$FF,$A5,$A7,$8C,$76,$8D,$46,$33,$FF,$4A
	.dc.b	$3F,$EA,$5C,$7F,$D3,$73,$FE,$AD,$3F,$D6,$97,$DE,$7C,$31,$FE,$A5
	.dc.b	$C3,$CD,$39,$24,$7B,$8E,$58,$FF,$AD,$2B,$B4,$7F,$1D,$C7,$FD,$87
	.dc.b	$5F,$EA,$27,$D3,$1F,$20,$F3,$C7,$4A,$FF,$5C,$C7,$53,$E2,$47,$1A
	.dc.b	$F6,$5F,$92,$7C,$5C,$9F,$E5,$FE,$1F,$F4,$D3,$67,$FD,$58,$D9,$FF
	.dc.b	$71,$35,$FF,$53,$72,$7E,$DE,$31,$1F,$EB,$28,$C1,$FB,$BE,$C7,$0B
	.dc.b	$DF,$FF,$54,$46,$1B,$19,$C3,$FD,$79,$39,$6C,$66,$FF,$DE,$93,$96
	.dc.b	$C6,$5F,$E8,$C2,$34,$E1,$AF,$F4,$B6,$BB,$8E,$4E,$5A,$E6,$BC,$E4
	.dc.b	$AF,$2B,$63,$C6,$D7,$C1,$3C,$AB,$B9,$1C,$12,$78,$4F,$85,$AF,$86
	.dc.b	$BB,$91,$EE,$D7,$BC,$70,$AE,$CD,$AE,$D5,$BA,$EC,$D7,$B1,$D0,$87
	.dc.b	$47,$55,$D1,$B3,$D2,$76,$74,$29,$FE,$F6,$F6,$3F,$DC,$84,$9F,$F7
	.dc.b	$6B,$63,$F2,$2D,$7F,$22,$D7,$FD,$EC,$D7,$A4,$72,$23,$C9,$AF,$26
	.dc.b	$BC,$9A,$F2,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1
	.dc.b	$06,$29,$F7,$46,$1B,$19,$9F,$4C,$E5,$B1,$98,$F3,$19,$2B,$97,$CA
	.dc.b	$64,$AE,$5F,$31,$92,$B9,$7D,$0D,$74,$62,$7C,$E4,$9F,$2B,$63,$C6
	.dc.b	$D7,$C2,$D7,$BB,$5E,$91,$C8,$8F,$26,$BC,$9A,$F2,$6B,$C8,$E0,$43
	.dc.b	$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3E,$39,$27,$E5,$12
	.dc.b	$7E,$04,$65,$B1,$99,$F5,$CE,$5B,$19,$8F,$41,$92,$B9,$7C,$E6,$4A
	.dc.b	$E5,$F4,$19,$2B,$97,$D4,$D7,$46,$27,$CE,$49,$F2,$B6,$3C,$6D,$7C
	.dc.b	$2D,$7B,$B5,$E9,$1C,$88,$F2,$6B,$C9,$AF,$26,$BC,$8E,$04,$38,$14
	.dc.b	$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$7E,$07,$F4,$17,$BC,$9F,$EB
	.dc.b	$33,$FD,$6A,$73,$18,$8F,$F5,$14,$61,$C3,$5E,$2D,$79,$4F,$57,$6D
	.dc.b	$76,$D7,$73,$DD,$4F,$37,$C3,$3B,$6C,$EE,$75,$1C,$27,$4D,$9D,$46
	.dc.b	$27,$64,$70,$48,$D9,$27,$D2,$D7,$D2,$D7,$D2,$68,$86,$8A,$68,$C1
	.dc.b	$0C,$10,$F6,$9C,$0A,$6C,$A6,$CF,$71,$B2,$9B,$29,$B9,$F8,$11,$E9
	.dc.b	$8E,$64,$36,$62,$D9,$BD,$CF,$F1,$CB,$6B,$FC,$72,$DA,$FF,$1C,$B5
	.dc.b	$FF,$1C,$CF,$02,$4F,$07,$FD,$5D,$1F,$E9,$43,$FD,$57,$D1,$FE,$8F
	.dc.b	$7F,$D3,$5C,$1F,$E8,$F7,$FD,$30,$7F,$A4,$0C,$D7,$FA,$9E,$35,$1C
	.dc.b	$8F,$F4,$BC,$64,$8F,$FA,$AF,$87,$FD,$74,$FF,$AB,$0E,$A4,$39,$1D
	.dc.b	$8E,$07,$9C,$E0,$68,$FF,$54,$9E,$52,$9C,$CD,$91,$FF,$53,$D7,$EC
	.dc.b	$AB,$DD,$AF,$FA,$97,$93,$FD,$3B,$CF,$FE,$A9,$E6,$FF,$52,$73,$FF
	.dc.b	$A9,$B2,$1F,$E6,$07,$0D,$9F,$F5,$3B,$FB,$EB,$E9,$5F,$EA,$67,$FD
	.dc.b	$5B,$1F,$EA,$B8,$FF,$48,$BF,$E8,$1E,$6F,$F4,$C1,$FE,$9E,$27,$37
	.dc.b	$FA,$2A,$35,$8F,$FA,$6B,$2F,$F4,$37,$3F,$FA,$B6,$31,$3F,$E8,$7B
	.dc.b	$FF,$47,$B6,$33,$AF,$FA,$30,$FF,$58,$C7,$FA,$80,$E1,$87,$FA,$82
	.dc.b	$35,$7E,$6B,$F2,$CF,$8D,$EB,$3C,$8F,$F5,$06,$1E,$4B,$FF,$51,$DF
	.dc.b	$1B,$E0,$75,$71,$3F,$E9,$B3,$84,$FB,$B6,$FF,$4E,$5F,$8A,$3F,$D5
	.dc.b	$57,$C5,$FF,$65,$5F,$C1,$AD,$BD,$23,$E0,$46,$1D,$9E,$43,$FD,$91
	.dc.b	$7E,$0C,$3D,$75,$CE,$7E,$26,$1E,$15,$3F,$EC,$DF,$65,$9F,$F2,$26
	.dc.b	$E3,$FC,$B5,$B5,$FC,$F5,$AF,$F4,$26,$BF,$E9,$C6,$BF,$8E,$6A,$6F
	.dc.b	$D1,$BE,$2B,$F9,$8B,$6B,$F9,$E3,$6B,$FA,$63,$5F,$EA,$2D,$7F,$D5
	.dc.b	$6D,$7F,$3A,$D4,$D2,$4F,$80,$93,$E0,$6C,$F8,$1A,$F8,$1A,$F8,$1A
	.dc.b	$F8,$1A,$F8,$1B,$3E,$09,$EC,$49,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC
	.dc.b	$D8,$ED,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$F4,$79,$10,$E4,$53
	.dc.b	$F8,$6F,$D7,$BD,$5F,$ED,$AD,$AF,$EF,$0D,$9F,$F0,$43,$5F,$F0,$6B
	.dc.b	$5F,$F7,$3B,$5E,$B3,$D0,$93,$D1,$B1,$D1,$AF,$46,$BD,$1A,$F4,$8E
	.dc.b	$44,$79,$35,$E4,$D7,$93,$5E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D
	.dc.b	$14,$D1,$82,$18,$20,$9F,$90,$D9,$F7,$B6,$7D,$0D,$3C,$65,$3C,$05
	.dc.b	$3B,$94,$EF,$3D,$89,$3D,$9B,$3D,$9B,$3D,$9A,$F6,$6B,$D9,$AF,$69
	.dc.b	$E8,$49,$E8,$D8,$E8,$D7,$A3,$5E,$8D,$7A,$47,$22,$3C,$9A,$F2,$6B
	.dc.b	$C9,$AF,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$10
	.dc.b	$FA,$84,$9F,$92,$D9,$F8,$0D,$9F,$4B,$4F,$21,$4F,$09,$4F,$01,$4F
	.dc.b	$04,$F6,$24,$F6,$6C,$F6,$6C,$F6,$6B,$D9,$AF,$66,$BD,$A7,$A1,$27
	.dc.b	$A3,$63,$A3,$5E,$8D,$7A,$35,$E9,$1C,$88,$F2,$6B,$C9,$AF,$26,$BC
	.dc.b	$8E,$04,$38,$14,$E0,$53,$81,$A2,$1F,$E3,$61,$F7,$2F,$8C,$FF,$4B
	.dc.b	$B6,$7F,$D4,$4D,$9F,$F5,$93,$5F,$F5,$BB,$5E,$8D,$7F,$C8,$6D,$7B
	.dc.b	$CF,$62,$4F,$66,$C7,$66,$BD,$9A,$F6,$6B,$D9,$AF,$68,$E8,$47,$A3
	.dc.b	$5E,$8D,$7A,$35,$E8,$D7,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0
	.dc.b	$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$03,$ED,$9F,$23,$FE
	.dc.b	$D6,$6C,$FF,$B8,$5B,$3F,$EF,$16,$BF,$EF,$66,$BF,$8B,$6B,$FE,$82
	.dc.b	$6B,$DE,$7F,$D5,$BE,$0B,$A3,$FD,$59,$3F,$E9,$43,$FD,$31,$C7,$FE
	.dc.b	$A8,$7F,$D2,$87,$FA,$77,$8B,$FD,$4A,$47,$04,$AE,$8F,$FA,$B8,$FF
	.dc.b	$52,$B8,$7F,$D5,$0D,$7B,$E3,$FE,$B2,$8C,$47,$81,$D5,$78,$EB,$C1
	.dc.b	$18,$AE,$17,$FA,$BB,$E6,$78,$1F,$6D,$F1,$8F,$34,$62,$78,$C7,$FA
	.dc.b	$D6,$75,$5D,$60,$F9,$2F,$FD,$6B,$1F,$B9,$9F,$F5,$AF,$43,$A3,$FD
	.dc.b	$69,$3F,$EB,$2E,$8F,$F5,$6B,$67,$4F,$FA,$B4,$C4,$FF,$AD,$1F,$F5
	.dc.b	$DB,$5C,$9B,$3F,$D7,$45,$9E,$AF,$FA,$D7,$9F,$FD,$20,$DB,$FF,$4C
	.dc.b	$3F,$EA,$87,$FD,$15,$D1,$FE,$AE,$E8,$FF,$54,$36,$7F,$D3,$33,$C3
	.dc.b	$2F,$F4,$67,$47,$FA,$42,$7F,$D2,$87,$53,$A1,$FE,$9A,$3F,$D2,$77
	.dc.b	$D3,$5F,$F4,$E4,$6E,$BF,$D6,$53,$FE,$A1,$7A,$E1,$E1,$9F,$F5,$61
	.dc.b	$C8,$F1,$9D,$63,$FD,$72,$F8,$1E,$6F,$FA,$98,$C4,$F8,$5E,$47,$48
	.dc.b	$E2,$78,$2B,$CB,$7F,$ED,$AC,$3E,$2C,$7A,$39,$DE,$5F,$FA,$A4,$91
	.dc.b	$FE,$A9,$6E,$1F,$EA,$96,$BF,$EA,$97,$FD,$43,$F1,$D2,$7F,$D4,$2D
	.dc.b	$9F,$F5,$0B,$5F,$F5,$0B,$5F,$F5,$0B,$5F,$F5,$0B,$5F,$F5,$0B,$7A
	.dc.b	$0D,$24,$F8,$09,$3E,$06,$BE,$06,$BE,$06,$BE,$06,$BE,$06,$CF,$81
	.dc.b	$B3,$E0,$9E,$C4,$3B,$14,$EC,$53,$B1,$4E,$C5,$9E,$CD,$8E,$CF,$42
	.dc.b	$1D,$0A,$74,$3F,$D2,$3F,$49,$FE,$98,$25,$7F,$A6,$1B,$1F,$E9,$86
	.dc.b	$BC,$23,$44,$74,$D7,$46,$08,$60,$87,$FA,$82,$3A,$B6,$39,$34,$E2
	.dc.b	$53,$89,$FE,$B2,$8E,$2D,$8E,$2D,$38,$94,$E2,$7F,$A2,$63,$8B,$63
	.dc.b	$8B,$4E,$25,$38,$F4,$A7,$8B,$67,$8B,$67,$8B,$5E,$33,$C0,$93,$C1
	.dc.b	$B1,$C1,$AF,$08,$D1,$1D,$35,$D1,$82,$18,$21,$F2,$C8,$7C,$52,$9E
	.dc.b	$F2,$CF,$A5,$B3,$E4,$6C,$F8,$5A,$78,$0A,$78,$0E,$C4,$3B,$16,$7B
	.dc.b	$36,$7B,$36,$7B,$34,$EC,$53,$B1,$D0,$93,$D1,$B3,$D1,$B3,$D1,$A7
	.dc.b	$42,$9D,$27,$91,$27,$93,$67,$93,$67,$93,$5E,$53,$C0,$93,$C1,$B1
	.dc.b	$C1,$AF,$08,$D1,$1D,$35,$D1,$82,$18,$21,$F5,$08,$7C,$92,$9F,$00
	.dc.b	$B3,$E9,$6C,$F9,$1B,$3E,$16,$9E,$02,$9E,$03,$B1,$0E,$C5,$9E,$CD
	.dc.b	$9E,$CD,$9E,$CD,$3B,$14,$EC,$74,$24,$F4,$6C,$F4,$7F,$0D,$D0,$F3
	.dc.b	$3F,$D2,$24,$BF,$F4,$8B,$67,$FD,$22,$D7,$FD,$22,$D7,$FD,$22,$D7
	.dc.b	$FD,$22,$D7,$FD,$3E,$D9,$F0,$4F,$62,$4F,$66,$BD,$9A,$F6,$6B,$D9
	.dc.b	$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8,$87
	.dc.b	$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3F,$E7,$FF,$7F
	.dc.b	$FD,$14,$49,$FF,$45,$36,$3F,$D1,$4D,$7F,$D1,$4F,$FA,$13,$9F,$FC
	.dc.b	$C9,$1F,$E6,$93,$FC,$F5,$CF,$FE,$61,$7F,$CD,$26,$23,$FC,$E0,$FF
	.dc.b	$9D,$DE,$31,$B3,$FC,$D4,$57,$27,$0E,$1F,$F3,$DB,$89,$FF,$39,$5F
	.dc.b	$F9,$D1,$FF,$73,$4E,$9F,$01,$A3,$CC,$7F,$A3,$27,$A3,$E6,$30,$6E
	.dc.b	$BC,$4F,$23,$A9,$FE,$7E,$8F,$F4,$57,$07,$F9,$F3,$9B,$FC,$F7,$3F
	.dc.b	$E7,$0E,$47,$6F,$F9,$DA,$3F,$CE,$5D,$1F,$E7,$C9,$CC,$7F,$9E,$0F
	.dc.b	$F4,$A7,$42,$7F,$D0,$31,$86,$BC,$CC,$5F,$82,$7F,$CF,$A7,$17,$C7
	.dc.b	$CD,$FE,$57,$9F,$F2,$F9,$FE,$74,$7F,$CE,$CF,$F9,$BF,$9C,$E7,$A1
	.dc.b	$9E,$36,$FF,$3E,$C7,$FA,$22,$FA,$47,$F9,$96,$BF,$CD,$31,$D0,$67
	.dc.b	$FE,$77,$70,$DA,$E4,$76,$9F,$28,$9E,$3B,$FF,$9C,$27,$57,$DC,$FF
	.dc.b	$3A,$E3,$FE,$79,$6D,$FF,$A0,$9F,$F3,$FB,$ED,$31,$1E,$E7,$B6,$1E
	.dc.b	$3C,$7F,$D3,$53,$FE,$8F,$35,$8F,$8A,$BD,$B8,$7A,$2B,$FD,$03,$18
	.dc.b	$78,$BF,$2B,$A1,$E3,$33,$FD,$0C,$A7,$31,$86,$C7,$E9,$9A,$FF,$B5
	.dc.b	$DA,$FA,$5A,$F7,$9E,$CF,$16,$BF,$ED,$FE,$87,$B1,$88,$F8,$51,$8B
	.dc.b	$F2,$B6,$3F,$6E,$D7,$94,$F1,$7F,$0F,$19,$79,$C6,$54,$65,$DC,$66
	.dc.b	$31,$1A,$8C,$12,$7C,$6D,$8F,$1B,$5F,$1C,$68,$8E,$9A,$E8,$C1,$0C
	.dc.b	$10,$F4,$BC,$1A,$6C,$A6,$CF,$5B,$B6,$9B,$29,$B3,$DC,$ED,$A6,$CA
	.dc.b	$6E,$7E,$29,$27,$E0,$47,$88,$87,$62,$9C,$8B,$1F,$3B,$A1,$E3,$FD
	.dc.b	$56,$DF,$E5,$1B,$5F,$CD,$9A,$FE,$9C,$38,$7F,$54,$D7,$F8,$4B,$5D
	.dc.b	$CF,$72,$4F,$76,$C7,$76,$BD,$E7,$91,$23,$31,$E0,$6B,$E0,$6B,$E0
	.dc.b	$9E,$C4,$7B,$35,$ED,$1C,$27,$B4,$E6,$7B,$35,$EC,$D7,$B1,$D0,$87
	.dc.b	$47,$51,$D2,$32,$74,$29,$D0,$A7,$43,$91,$87,$93,$82,$1C,$CA,$73
	.dc.b	$29,$CC,$D1,$0D,$14,$D1,$82,$18,$23,$E2,$78,$35,$DB,$5D,$CF,$CD
	.dc.b	$24,$FC,$46,$C7,$B5,$AF,$A2,$79,$12,$33,$1E,$06,$BE,$06,$BE,$09
	.dc.b	$EC,$47,$B3,$5E,$D1,$C2,$7B,$4E,$67,$B3,$5E,$CD,$7B,$1D,$08,$74
	.dc.b	$75,$1D,$23,$27,$42,$9D,$0A,$74,$39,$18,$79,$38,$21,$CC,$A7,$32
	.dc.b	$9C,$CD,$10,$D1,$4D,$18,$21,$82,$3E,$27,$83,$5D,$B5,$DC,$FC,$D2
	.dc.b	$4F,$C4,$6C,$7B,$5A,$FA,$27,$91,$23,$31,$E0,$6B,$E0,$6B,$E0,$9E
	.dc.b	$C4,$7B,$35,$ED,$1C,$27,$B4,$E6,$7B,$35,$EC,$D5,$F7,$5F,$E2,$88
	.dc.b	$FE,$FA,$D8,$FF,$2F,$B5,$FE,$54,$D7,$FC,$24,$D7,$FC,$62,$D7,$A9
	.dc.b	$FD,$AD,$C1,$1C,$12,$3C,$2D,$3C,$25,$3C,$25,$3C,$25,$3C,$25,$3C
	.dc.b	$2F,$42,$3D,$1A,$F4,$6B,$D1,$AF,$46,$BD,$0E,$44,$39,$14,$E4,$53
	.dc.b	$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$83,$FC,$CB
	.dc.b	$DC,$F8,$8F,$F0,$F3,$67,$FC,$7A,$D9,$FF,$4F,$B5,$FE,$D4,$D7,$FC
	.dc.b	$A4,$D7,$AC,$F4,$3F,$D5,$CF,$FA,$44,$FF,$39,$1F,$E9,$73,$FC,$EB
	.dc.b	$3F,$EA,$17,$3D,$1F,$EA,$AE,$8F,$F4,$EB,$63,$FD,$39,$18,$3F,$D5
	.dc.b	$4F,$FA,$72,$34,$D8,$CB,$D1,$AF,$FA,$E4,$F2,$14,$FF,$54,$9C,$4C
	.dc.b	$F2,$7F,$AC,$DC,$9C,$CC,$9E,$BE,$0F,$F2,$CD,$7F,$AC,$23,$67,$53
	.dc.b	$DE,$7F,$AE,$8A,$AB,$FD,$5D,$C9,$FE,$AC,$9F,$F5,$97,$46,$08,$FF
	.dc.b	$AB,$CF,$F5,$51,$0C,$1F,$EA,$2E,$7F,$F5,$5F,$47,$FA,$DE,$78,$CF
	.dc.b	$FA,$2C,$FF,$44,$E7,$FE,$95,$CB,$FD,$07,$9F,$FA,$4A,$7F,$D3,$5C
	.dc.b	$BF,$E8,$77,$FD,$3B,$CD,$FE,$8B,$9D,$3F,$EA,$4B,$E5,$3A,$CF,$FD
	.dc.b	$28,$AB,$FD,$37,$5F,$EA,$DB,$FF,$51,$F3,$FF,$A8,$4F,$F4,$C5,$7F
	.dc.b	$A7,$8F,$F5,$D6,$3F,$EB,$98,$CC,$7A,$0F,$F4,$C1,$D5,$FF,$4F,$E1
	.dc.b	$E0,$C3,$BD,$F6,$AE,$87,$B2,$FF,$D3,$EF,$C3,$C3,$D9,$87,$67,$D9
	.dc.b	$1D,$AF,$9D,$78,$E3,$9C,$EA,$7E,$1D,$FC,$77,$FD,$7D,$7F,$EB,$28
	.dc.b	$F1,$78,$AF,$15,$EB,$FF,$A4,$E3,$0D,$8C,$E9,$FE,$59,$C9,$5A,$8C
	.dc.b	$36,$3F,$CB,$CD,$7F,$D8,$CD,$7F,$DC,$AD,$7F,$D4,$3E,$77,$FA,$6C
	.dc.b	$93,$FE,$AA,$6C,$7F,$AA,$9A,$FF,$AC,$DA,$FF,$AC,$9A,$F4,$8E,$44
	.dc.b	$79,$35,$E4,$D7,$93,$5E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14
	.dc.b	$D1,$82,$18,$21,$FE,$C2,$8F,$6B,$63,$CC,$D3,$C2,$53,$B1,$4E,$A5
	.dc.b	$3A,$F4,$27,$AB,$67,$AB,$67,$AB,$5F,$F4,$77,$C1,$FF,$A8,$C9,$3F
	.dc.b	$EA,$36,$C7,$FA,$8D,$AF,$08,$D1,$1D,$35,$D1,$82,$18,$21,$FE,$AA
	.dc.b	$8E,$AD,$8E,$4D,$38,$94,$E2,$7F,$AC,$23,$8B,$63,$8B,$4E,$25,$38
	.dc.b	$9F,$EB,$C8,$E2,$D8,$E2,$D3,$89,$4E,$3D,$09,$E2,$D9,$E2,$D9,$E2
	.dc.b	$D7,$8C,$F0,$24,$F0,$6C,$70,$6B,$C2,$34,$47,$4D,$74,$60,$86,$08
	.dc.b	$7C,$B2,$1F,$14,$A7,$BC,$B3,$E9,$6C,$F9,$1B,$3E,$16,$9E,$02,$9E
	.dc.b	$03,$B1,$0E,$C5,$9E,$CD,$9E,$CD,$9E,$CD,$3B,$14,$EC,$74,$24,$F4
	.dc.b	$6C,$F4,$6C,$F4,$69,$D0,$A7,$49,$E4,$49,$E4,$D9,$E4,$D9,$E4,$D7
	.dc.b	$94,$F0,$24,$F0,$6C,$70,$6B,$C2,$34,$47,$4D,$74,$60,$86,$08,$7D
	.dc.b	$42,$1F,$24,$A7,$C0,$2C,$FA,$5B,$3E,$46,$CF,$85,$A7,$80,$A7,$80
	.dc.b	$EC,$43,$B1,$67,$B3,$67,$B3,$67,$B3,$4E,$C5,$3B,$1D,$09,$3D,$1B
	.dc.b	$3D,$1F,$C3,$74,$3E,$67,$FD,$20,$4B,$FF,$4E,$B6,$7F,$D3,$AD,$7F
	.dc.b	$D3,$AD,$7F,$D3,$AD,$7F,$D3,$AD,$7F,$D5,$2D,$9F,$04,$F6,$24,$F6
	.dc.b	$6B,$D9,$AF,$66,$BD,$9A,$F6,$6C,$76,$8E,$84,$3A,$14,$E8,$53,$A1
	.dc.b	$4E,$85,$7A,$3C,$88,$72,$29,$C8,$A7,$22,$BC,$9E,$04,$38,$14,$E0
	.dc.b	$57,$82,$1F,$70,$F1,$55,$FE,$23,$9C,$C6,$1F,$F1,$34,$EE,$74,$D8
	.dc.b	$CC,$ED,$EC,$77,$8D,$38,$3A,$CF,$29,$FF,$4D,$C7,$F9,$F3,$87,$FD
	.dc.b	$08,$FF,$9E,$BA,$3F,$D2,$8F,$FA,$13,$A7,$FD,$26,$FF,$A1,$CE,$07
	.dc.b	$FA,$1E,$FF,$D4,$0F,$FA,$17,$97,$FD,$13,$5C,$A3,$75,$C9,$D5,$70
	.dc.b	$BF,$0D,$F0,$9F,$F5,$29,$0F,$F4,$D1,$CA,$FF,$D4,$33,$C2,$B9,$4F
	.dc.b	$8C,$FF,$4A,$57,$37,$FD,$22,$7F,$A2,$FA,$3F,$D1,$4D,$F0,$3E,$B7
	.dc.b	$47,$FA,$5F,$A9,$5A,$A4,$62,$BE,$C3,$63,$FC,$B8,$D7,$E8,$35,$FF
	.dc.b	$4B,$BF,$E9,$A8,$EF,$D5,$FE,$75,$BF,$F3,$F7,$4F,$FA,$03,$2F,$F4
	.dc.b	$8E,$7F,$E7,$0C,$7F,$CE,$BD,$3F,$E8,$17,$FD,$00,$7F,$A1,$CF,$F4
	.dc.b	$69,$FE,$9C,$7F,$CF,$15,$FE,$93,$8E,$3A,$7F,$A9,$23,$FD,$09,$3A
	.dc.b	$9F,$F4,$0E,$1E,$27,$9D,$AB,$CD,$87,$91,$E9,$3F,$E8,$FB,$E4,$F9
	.dc.b	$2B,$FD,$1A,$F8,$F0,$FF,$4A,$F5,$7F,$A9,$DB,$5E,$B8,$FF,$48,$5F
	.dc.b	$92,$72,$79,$EB,$D2,$7F,$AC,$F0,$F1,$3E,$86,$BE,$19,$F9,$4F,$C5
	.dc.b	$30,$7F,$A4,$3A,$5F,$0F,$FE,$95,$25,$7F,$A4,$A3,$2D,$8C,$CF,$FA
	.dc.b	$52,$72,$D8,$FD,$93,$FB,$2F,$07,$F7,$B5,$FB,$D6,$C7,$FA,$85,$AF
	.dc.b	$ED,$5A,$F1,$8E,$6F,$F0,$98,$CB,$63,$2A,$32,$D8,$CC,$60,$8E,$08
	.dc.b	$79,$4B,$1E,$56,$BE,$56,$BE,$57,$81,$1E,$0D,$78,$35,$E0,$68,$86
	.dc.b	$8A,$68,$C1,$0C,$10,$F8,$4F,$56,$9C,$8A,$71,$29,$C4,$F8,$CF,$16
	.dc.b	$9C,$4A,$71,$29,$C6,$3E,$5C,$71,$E8,$7B,$1F,$51,$B1,$FC,$F5,$B5
	.dc.b	$FE,$E1,$6B,$FC,$D5,$AF,$F3,$A6,$BF,$D2,$5A,$9A,$49,$F0,$12,$7C
	.dc.b	$0D,$9F,$03,$5F,$03,$5F,$03,$5F,$03,$5F,$03,$67,$C1,$3D,$89,$3D
	.dc.b	$9A,$F6,$6B,$D9,$AF,$66,$BD,$9B,$1D,$A3,$A1,$0E,$85,$3A,$14,$E8
	.dc.b	$53,$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8,$AF,$27,$81,$0E,$05,$38
	.dc.b	$15,$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$7E,$A1,$27,$E4,$B6,$7E,$03
	.dc.b	$5F,$4B,$5F,$23,$5F,$0B,$5F,$03,$67,$C1,$3D,$89,$3D,$9A,$F6,$6B
	.dc.b	$D9,$AF,$66,$BD,$9B,$1D,$A3,$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$5E
	.dc.b	$8F,$22,$1C,$8A,$72,$29,$C8,$AF,$27,$81,$0E,$05,$38,$15,$E0,$E8
	.dc.b	$86,$8A,$E9,$C1,$1C,$12,$7E,$A1,$27,$E4,$B6,$7E,$03,$5F,$4B,$5F
	.dc.b	$23,$5F,$0B,$5F,$03,$67,$C1,$3D,$89,$3D,$9A,$F6,$6B,$D9,$AF,$66
	.dc.b	$AE,$77,$C4,$7F,$5E,$6C,$FF,$99,$DB,$3F,$E5,$D3,$84,$EC,$93,$B6
	.dc.b	$C6,$E7,$8B,$FE,$65,$6B,$FE,$6A,$6B,$FE,$79,$6B,$DA,$7A,$11,$E9
	.dc.b	$3D,$9E,$8D,$7A,$35,$E8,$D7,$A1,$CA,$3A,$3C,$9A,$F2,$6B,$C9,$AF
	.dc.b	$23,$81,$CC,$86,$08,$73,$32,$53,$27,$53,$25,$32,$77,$38,$10,$D1
	.dc.b	$4D,$18,$21,$82,$74,$3C,$3F,$F0,$EB,$67,$FD,$43,$18,$23,$82,$4F
	.dc.b	$FA,$A2,$3F,$D3,$AF,$FA,$6C,$FF,$45,$1F,$E9,$93,$07,$FA,$30,$A6
	.dc.b	$76,$FF,$44,$36,$7F,$D1,$BD,$1F,$E9,$9E,$2F,$F4,$84,$FF,$A3,$A7
	.dc.b	$3D,$1F,$E9,$48,$FF,$49,$52,$3F,$D4,$51,$D8,$FF,$50,$9C,$49,$1B
	.dc.b	$7C,$E6,$9C,$3E,$7B,$F2,$B9,$9E,$B1,$8B,$F2,$D7,$0A,$F5,$1D,$AB
	.dc.b	$77,$E8,$8F,$F5,$F4,$7F,$B1,$EF,$FD,$51,$CF,$FE,$97,$97,$2F,$FA
	.dc.b	$97,$A3,$FD,$49,$3F,$E9,$D9,$FF,$55,$1F,$EA,$28,$74,$7F,$A9,$9B
	.dc.b	$19,$8F,$F5,$0B,$C2,$BA,$46,$23,$FD,$6E,$7F,$9E,$09,$CD,$E2,$CB
	.dc.b	$FC,$FE,$FF,$A2,$23,$FD,$0B,$A7,$FA,$9D,$FF,$4A,$65,$FE,$8A,$7F
	.dc.b	$D1,$07,$FA,$4C,$B1,$FE,$9D,$78,$F3,$7F,$A7,$2B,$FD,$2D,$1A,$8F
	.dc.b	$F4,$A3,$FE,$8B,$8F,$13,$CE,$7C,$11,$C1,$BC,$FF,$E9,$D2,$5F,$8E
	.dc.b	$32,$AB,$95,$FF,$A6,$27,$C9,$1E,$A3,$C1,$97,$FA,$D5,$61,$E2,$9D
	.dc.b	$BE,$1A,$F5,$61,$E0,$AF,$51,$EE,$3F,$D5,$07,$80,$B5,$CA,$38,$3E
	.dc.b	$68,$D1,$E0,$BF,$15,$F8,$6F,$E2,$D7,$FB,$00,$E5,$87,$FA,$E6,$FD
	.dc.b	$51,$99,$FF,$65,$C7,$4A,$FB,$4E,$A3,$E1,$1F,$15,$FC,$17,$2B,$CD
	.dc.b	$FF,$4E,$93,$3F,$F5,$73,$67,$FD,$5C,$D7,$FD,$5C,$FF,$A9,$3D,$09
	.dc.b	$3F,$EB,$56,$CF,$FA,$D5,$AF,$FA,$D5,$AF,$FA,$F5,$AF,$FA,$D5,$AF
	.dc.b	$FA,$CD,$B4,$69,$27,$C0,$49,$F0,$35,$F0,$35,$F0,$35,$F0,$35,$F0
	.dc.b	$36,$7C,$0D,$9F,$04,$F6,$21,$D8,$A7,$62,$9D,$8A,$76,$2C,$F6,$6C
	.dc.b	$76,$7A,$10,$E8,$53,$A1,$F7,$47,$77,$E8,$EB,$FB,$E4,$60,$FF,$6F
	.dc.b	$C6,$5F,$F0,$03,$5E,$4D,$7F,$10,$D7,$F5,$CD,$7A,$47,$68,$C1,$23
	.dc.b	$B9,$23,$4D,$3C,$25,$3C,$25,$3C,$25,$3C,$25,$3C,$27,$27,$A9,$1C
	.dc.b	$10,$EA,$E5,$A6,$4E,$EE,$5A,$64,$F0,$B9,$69,$93,$C6,$74,$21,$C8
	.dc.b	$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$47,$D1,$8F,$34,$60,$F9
	.dc.b	$11,$97,$D0,$D7,$C6,$D7,$C0,$D7,$BB,$5E,$91,$DA,$30,$48,$EE,$48
	.dc.b	$D3,$4F,$09,$4F,$09,$4F,$09,$4F,$09,$4F,$09,$C9,$EA,$47,$04,$3A
	.dc.b	$B9,$69,$93,$BB,$96,$99,$3C,$2E,$5A,$64,$F1,$9D,$08,$72,$29,$C0
	.dc.b	$A7,$03,$44,$34,$53,$46,$08,$60,$91,$F4,$63,$CD,$18,$3E,$44,$65
	.dc.b	$F4,$35,$F1,$B5,$F0,$35,$EE,$D7,$A4,$76,$8C,$12,$3B,$92,$34,$D3
	.dc.b	$C2,$53,$C2,$53,$C2,$53,$C2,$53,$C2,$72,$7A,$91,$C1,$0E,$AE,$5A
	.dc.b	$64,$EE,$FC,$8E,$67,$B5,$FE,$A9,$24,$7F,$AA,$5B,$3F,$EA,$36,$BF
	.dc.b	$EA,$36,$BF,$EA,$36,$BF,$EA,$36,$BF,$EB,$76,$CF,$82,$7B,$12,$7B
	.dc.b	$35,$EC,$D7,$B3,$5E,$CD,$7B,$36,$3B,$47,$42,$1D,$0A,$74,$29,$D0
	.dc.b	$A7,$42,$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A,$70
	.dc.b	$2B,$C1,$FF,$22,$7D,$CB,$89,$5F,$D9,$23,$0D,$8C,$CF,$FB,$8E,$72
	.dc.b	$D8,$CB,$FD,$BE,$32,$D8,$CB,$FE,$2F,$8C,$BF,$E8,$63,$FD,$10,$7F
	.dc.b	$A5,$36,$FF,$46,$F2,$67,$97,$FD,$1D,$3F,$E9,$33,$47,$FA,$8A,$34
	.dc.b	$47,$31,$FE,$AB,$74,$62,$3F,$D5,$51,$92,$38,$24,$7F,$A6,$08,$7F
	.dc.b	$9A,$F8,$7F,$D4,$4F,$43,$FD,$51,$8F,$FA,$7D,$D3,$5F,$F6,$1F,$2F
	.dc.b	$FA,$7F,$8F,$FD,$27,$1F,$EA,$0E,$DF,$F2,$94,$FF,$A7,$E0,$D2,$69
	.dc.b	$FE,$9E,$8F,$F0,$E4,$73,$9E,$33,$FE,$A2,$A4,$7F,$98,$B3,$FF,$33
	.dc.b	$DF,$F9,$8E,$7F,$CD,$A4,$C3,$FD,$3F,$CA,$CB,$FD,$15,$C9,$FE,$7E
	.dc.b	$7F,$D1,$58,$72,$3F,$CF,$05,$8D,$61,$D6,$F9,$D7,$28,$FF,$40,$CF
	.dc.b	$F9,$E8,$FF,$38,$BF,$E8,$73,$FD,$63,$87,$8D,$FF,$3C,$61,$E3,$7F
	.dc.b	$D2,$B7,$D1,$A6,$AF,$B3,$CA,$35,$87,$FA,$0E,$3B,$72,$FF,$A2,$DB
	.dc.b	$7D,$67,$51,$FE,$C2,$8C,$63,$E6,$8F,$F4,$41,$FE,$AB,$AE,$4E,$0F
	.dc.b	$F4,$C5,$FC,$69,$F7,$46,$F1,$FD,$4E,$1F,$1E,$3F,$D6,$9C,$8F,$0F
	.dc.b	$F4,$B1,$86,$A8,$D6,$1F,$E5,$86,$C7,$F9,$74,$D4,$F0,$8C,$C7,$F9
	.dc.b	$85,$AF,$FA,$7D,$AF,$8A,$F8,$12,$BB,$36,$3B,$4F,$FB,$51,$E9,$19
	.dc.b	$6C,$F0,$23,$C2,$34,$F8,$5A,$F8,$63,$6F,$23,$83,$87,$83,$5E,$06
	.dc.b	$88,$71,$21,$C0,$C9,$4C,$9C,$8A,$7C,$F8,$F2,$B6,$3C,$0D,$39,$9C
	.dc.b	$4E,$84,$30,$43,$81,$DA,$7E,$53,$B3,$25,$72,$F1,$6B,$A3,$13,$FE
	.dc.b	$98,$CE,$57,$FA,$3D,$B3,$FE,$9A,$6B,$FE,$9A,$6B,$FE,$9A,$6B,$FE
	.dc.b	$9A,$6B,$FE,$A8,$6F,$27,$F0,$59,$EC,$49,$EC,$D7,$B3,$5E,$CD,$7B
	.dc.b	$35,$EC,$D8,$ED,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$F4,$79,$10
	.dc.b	$E4,$53,$91,$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF,$07,$44,$34,$57
	.dc.b	$4E,$08,$E0,$93,$F5,$09,$3F,$25,$B3,$F0,$1A,$FA,$5A,$F9,$1A,$F8
	.dc.b	$5A,$F8,$1B,$3E,$09,$EC,$49,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D8
	.dc.b	$ED,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$F4,$79,$10,$E4,$53,$91
	.dc.b	$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF,$07,$44,$34,$57,$4E,$08,$E0
	.dc.b	$93,$F5,$09,$3F,$25,$B3,$F0,$1A,$FA,$5A,$F9,$1A,$F8,$5A,$F8,$1B
	.dc.b	$3E,$09,$EC,$49,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D8,$ED,$1D,$08
	.dc.b	$74,$29,$D0,$A0,$71,$BD,$0F,$ED,$6D,$AF,$ED,$13,$A9,$C9,$23,$33
	.dc.b	$B9,$FE,$DC,$D7,$B1,$DD,$C3,$86,$B9,$9E,$8F,$26,$BC,$8F,$14,$F8
	.dc.b	$23,$31,$CD,$AF,$38,$D1,$1E,$8D,$7A,$4F,$89,$EA,$68,$86,$A7,$AB
	.dc.b	$A3,$13,$CD,$F1,$35,$F1,$1C,$8C,$94,$C9,$C8,$C1,$D4,$D9,$92,$18
	.dc.b	$21,$C8,$A7,$13,$04,$30,$49,$F8,$07,$C1,$E4,$7C,$8F,$F8,$B1,$B5
	.dc.b	$FE,$9F,$8C,$11,$C1,$2B,$FD,$41,$1F,$E3,$47,$FC,$C6,$7F,$A0,$4B
	.dc.b	$CB,$FE,$8A,$7F,$D0,$E7,$F5,$A7,$FD,$33,$B7,$FA,$72,$3F,$D2,$46
	.dc.b	$B3,$FF,$45,$CF,$18,$FF,$44,$9F,$EA,$43,$FD,$1A,$53,$BF,$2F,$FA
	.dc.b	$59,$DC,$76,$7C,$87,$FA,$9C,$E2,$E4,$EC,$F9,$63,$FD,$50,$E9,$AF
	.dc.b	$63,$CD,$87,$FA,$6E,$BC,$F3,$E1,$9E,$47,$B0,$F1,$BC,$95,$FE,$D6
	.dc.b	$D3,$FA,$1A,$FF,$66,$C6,$2B,$F7,$5C,$BF,$E9,$D6,$F2,$FF,$A7,$79
	.dc.b	$7F,$D3,$FD,$3F,$BA,$AF,$F5,$2B,$67,$FD,$4D,$5A,$3F,$D5,$C7,$28
	.dc.b	$FE,$0D,$C3,$FE,$B2,$25,$FF,$A9,$F9,$3F,$CF,$35,$FE,$7C,$3F,$D1
	.dc.b	$C7,$FA,$43,$4F,$F3,$D3,$FE,$87,$CF,$FD,$23,$CA,$E6,$FF,$44,$CF
	.dc.b	$FA,$07,$7F,$F4,$9E,$1F,$E8,$BA,$D5,$FF,$A1,$EB,$FD,$2B,$97,$FA
	.dc.b	$87,$1E,$B5,$A9,$F9,$37,$FE,$AE,$9C,$E3,$DA,$7C,$D1,$8C,$3A,$47
	.dc.b	$FA,$8C,$F4,$C7,$FA,$7A,$FA,$3F,$EB,$AB,$F8,$13,$FE,$A4,$AF,$5D
	.dc.b	$F5,$BF,$F4,$F9,$C2,$B7,$5F,$EA,$18,$CC,$EC,$95,$C8,$93,$F2,$6B
	.dc.b	$E6,$D7,$09,$F3,$BD,$67,$FD,$8E,$78,$CE,$05,$BF,$55,$F6,$8C,$9E
	.dc.b	$3C,$3F,$D5,$F8,$7F,$36,$BE,$9E,$9B,$E8,$5E,$5F,$FA,$C2,$30,$D8
	.dc.b	$CE,$9F,$EA,$D9,$CB,$63,$38,$FF,$39,$9C,$B6,$32,$FF,$46,$8C,$BF
	.dc.b	$EB,$3C,$BF,$D6,$C4,$AF,$F5,$E3,$63,$FD,$5E,$D7,$84,$68,$8E,$9A
	.dc.b	$E8,$C1,$0C,$10,$FF,$61,$47,$56,$C7,$26,$9C,$4A,$71,$3F,$DB,$31
	.dc.b	$C5,$B1,$C5,$A7,$12,$9C,$4F,$F6,$E4,$71,$6C,$71,$69,$C4,$A7,$1E
	.dc.b	$8F,$D7,$46,$DB,$1B,$6B,$B9,$ED,$18,$8F,$59,$0F,$29,$4E,$E5,$E8
	.dc.b	$7C,$77,$FB,$D1,$B7,$FE,$0A,$6C,$FF,$BC,$DA,$FF,$7D,$6B,$FE,$08
	.dc.b	$6B,$FE,$17,$6A,$69,$27,$C0,$49,$F0,$36,$7C,$0D,$7C,$0D,$7C,$0D
	.dc.b	$7C,$0D,$7C,$0D,$9F,$04,$F6,$24,$F6,$6B,$D9,$AF,$66,$BD,$9A,$F6
	.dc.b	$6C,$76,$8E,$84,$3A,$14,$E8,$53,$A1,$4E,$85,$7A,$3C,$88,$72,$29
	.dc.b	$C8,$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04
	.dc.b	$70,$49,$FA,$84,$9F,$92,$D9,$F8,$0D,$7D,$2D,$7C,$8D,$7C,$2D,$7C
	.dc.b	$0D,$9F,$04,$F6,$24,$F6,$6B,$D9,$AF,$66,$BD,$9A,$F6,$6C,$76,$8E
	.dc.b	$84,$3A,$14,$E8,$53,$A1,$4E,$85,$7A,$3C,$88,$72,$29,$C8,$A7,$22
	.dc.b	$BC,$9E,$04,$38,$14,$E0,$57,$83,$A2,$1A,$2B,$A7,$04,$70,$49,$FA
	.dc.b	$84,$9F,$92,$D9,$F8,$0D,$7D,$2D,$7C,$8D,$7C,$2D,$7C,$0D,$9F,$04
	.dc.b	$F6,$24,$F6,$6B,$D9,$AF,$66,$BD,$9A,$BE,$D1,$ED,$27,$FD,$67,$18
	.dc.b	$6C,$66,$7F,$D4,$C7,$17,$0F,$FA,$A6,$78,$3A,$6B,$98,$DB,$FE,$B8
	.dc.b	$8C,$B6,$32,$FF,$B2,$63,$2D,$8C,$BE,$19,$F2,$91,$F0,$46,$A7,$8C
	.dc.b	$6C,$CC,$ED,$F2,$35,$F1,$35,$F0,$35,$E4,$70,$39,$90,$C1,$0E,$66
	.dc.b	$4A,$64,$EA,$64,$A6,$4E,$E7,$02,$1A,$29,$A3,$04,$30,$48,$F8,$64
	.dc.b	$9F,$8C,$49,$F7,$47,$C8,$F9,$DF,$F4,$2C,$7F,$BD,$9B,$5F,$82,$6B
	.dc.b	$FE,$E9,$6B,$FE,$F4,$6A,$0E,$1B,$D1,$F7,$23,$11,$FE,$9F,$3F,$D3
	.dc.b	$27,$FA,$6C,$FF,$4E,$9F,$EA,$53,$1C,$7F,$EA,$27,$FD,$4A,$7F,$AD
	.dc.b	$4F,$F5,$81,$CC,$E2,$57,$83,$D8,$E8,$43,$55,$FE,$AF,$23,$D7,$8F
	.dc.b	$FD,$94,$EC,$EE,$47,$2F,$23,$81,$FE,$C4,$72,$FF,$B1,$79,$FF,$D5
	.dc.b	$C9,$73,$FF,$AB,$E3,$FD,$50,$7E,$CA,$3F,$D5,$ED,$8F,$F5,$7B,$AE
	.dc.b	$5F,$F5,$5B,$79,$BF,$D6,$50,$7F,$D1,$3C,$BF,$E7,$F9,$FF,$55,$1F
	.dc.b	$E7,$A7,$04,$E2,$FF,$49,$57,$FA,$97,$9B,$FD,$12,$DA,$FF,$42,$47
	.dc.b	$FA,$B6,$7F,$D0,$99,$7F,$A7,$67,$FD,$5B,$9F,$FA,$C9,$FF,$5A,$1F
	.dc.b	$EA,$4A,$E1,$A7,$FA,$4A,$F8,$C6,$27,$FD,$62,$72,$38,$1C,$AF,$FD
	.dc.b	$60,$E8,$FF,$58,$38,$3C,$4F,$B9,$FF,$4C,$D7,$73,$FD,$4A,$E2,$39
	.dc.b	$63,$F0,$B1,$F7,$47,$FA,$77,$1F,$53,$FE,$C0,$C3,$E1,$C6,$2B,$FD
	.dc.b	$59,$3C,$B1,$F5,$BE,$7A,$DB,$EA,$39,$1F,$A7,$9F,$F5,$97,$B3,$F5
	.dc.b	$FD,$7F,$E4,$AD,$C3,$F9,$9B,$6B,$F5,$C7,$EB,$DC,$38,$6B,$9D,$15
	.dc.b	$D1,$B5,$D1,$B3,$FD,$03,$97,$F9,$59,$38,$FF,$99,$B6,$BF,$A2,$CA
	.dc.b	$30,$F1,$9D,$92,$77,$3A,$9F,$49,$C6,$74,$F1,$9D,$C6,$23,$73,$C5
	.dc.b	$F8,$2D,$7F,$88,$35,$E9,$3C,$88,$F5,$21,$D2,$7B,$3D,$1A,$F4,$6B
	.dc.b	$D0,$E2,$6E,$3A,$BC,$DA,$E4,$D9,$CD,$CB,$4C,$9D,$4E,$46,$27,$D8
	.dc.b	$49,$F1,$BF,$D8,$39,$1F,$07,$FE,$A7,$25,$FF,$A9,$DB,$7F,$EA,$76
	.dc.b	$BF,$EA,$76,$BF,$EA,$76,$BF,$EA,$76,$BF,$EA,$76,$CF,$82,$7B,$12
	.dc.b	$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$36,$3B,$47,$42,$1D,$0A,$74,$29
	.dc.b	$D0,$A7,$42,$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A
	.dc.b	$70,$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$24,$FD,$42,$4F,$C9,$6C,$FC
	.dc.b	$06,$BE,$96,$BE,$46,$BE,$16,$BE,$06,$CF,$82,$7B,$12,$7B,$35,$EC
	.dc.b	$D7,$B3,$5E,$CD,$7B,$36,$3B,$47,$42,$1D,$0A,$74,$29,$D0,$A7,$42
	.dc.b	$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1
	.dc.b	$D1,$0D,$15,$D3,$82,$38,$24,$FD,$42,$4F,$C9,$6C,$FC,$06,$BE,$96
	.dc.b	$BE,$46,$BE,$16,$BE,$06,$CF,$82,$7B,$12,$7B,$35,$EC,$D7,$B3,$5E
	.dc.b	$CD,$7B,$36,$3B,$47,$42,$1D,$0A,$74,$3E,$FF,$2B,$E2,$3F,$D0,$C4
	.dc.b	$AF,$F5,$0B,$67,$FD,$34,$D7,$FD,$3C,$D7,$FD,$34,$D7,$FD,$30,$D7
	.dc.b	$FD,$5C,$D9,$F0,$4F,$62,$4F,$66,$BD,$9A,$F6,$6B,$D9,$AF,$66,$C7
	.dc.b	$68,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8,$87,$22,$9C,$8A
	.dc.b	$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3F,$E8,$BE,$57,$D0,$7F,$A2
	.dc.b	$49,$5F,$E9,$76,$CF,$FA,$5D,$AF,$FA,$5D,$FF,$65,$BF,$E7,$5E,$D7
	.dc.b	$47,$FB,$7B,$97,$FC,$EA,$49,$FF,$3C,$12,$76,$FF,$A4,$1A,$FF,$A0
	.dc.b	$E3,$84,$62,$3F,$D2,$27,$FA,$14,$E5,$3F,$E2,$0E,$6F,$F4,$9C,$6A
	.dc.b	$B8,$CE,$70,$FF,$48,$DF,$3A,$E4,$F5,$BE,$24,$9D,$D7,$AE,$BB,$9E
	.dc.b	$77,$15,$C6,$77,$3F,$E3,$2B,$FF,$5C,$4E,$60,$F8,$63,$35,$E5,$9C
	.dc.b	$BF,$E8,$EE,$09,$1F,$E9,$12,$1F,$E9,$DE,$7F,$F4,$BB,$7B,$7F,$C9
	.dc.b	$3D,$3F,$E9,$C8,$FF,$21,$BF,$EA,$43,$FD,$47,$1F,$EC,$BE,$9F,$F4
	.dc.b	$D4,$FF,$A0,$F1,$FF,$3F,$74,$7F,$98,$BA,$3F,$CF,$1A,$FF,$9C,$EF
	.dc.b	$FC,$F3,$1F,$E9,$BD,$BF,$D0,$18,$7F,$A7,$FA,$7F,$CE,$AF,$FA,$23
	.dc.b	$0F,$F3,$AE,$3D,$23,$FC,$E9,$3F,$EA,$2C,$7A,$12,$3F,$CF,$58,$7F
	.dc.b	$A0,$2B,$84,$7F,$9F,$0F,$F4,$41,$AC,$7C,$77,$E0,$AF,$3C,$E6,$FA
	.dc.b	$5F,$29,$CB,$6B,$C5,$5B,$24,$67,$1F,$F4,$DE,$3F,$EA,$98,$E5,$8F
	.dc.b	$82,$7E,$0C,$7F,$AD,$7A,$DF,$44,$F8,$C6,$9F,$E7,$4C,$95,$A9,$CC
	.dc.b	$7F,$9E,$23,$13,$B6,$C6,$6B,$A4,$E5,$B1,$9A,$E6,$FF,$A6,$AB,$8B
	.dc.b	$FE,$7E,$E6,$FF,$47,$92,$FF,$D1,$2D,$8F,$F4,$B3,$5E,$11,$A2,$3A
	.dc.b	$6B,$A3,$04,$30,$43,$FD,$1B,$1D,$5B,$1C,$9A,$71,$29,$C4,$FF,$49
	.dc.b	$C7,$16,$C7,$16,$9C,$4A,$71,$3F,$D5,$91,$C5,$B1,$C5,$A7,$12,$9C
	.dc.b	$7A,$7F,$94,$B8,$3C,$A4,$3B,$94,$D1,$82,$18,$23,$D0,$93,$E5,$E8
	.dc.b	$FE,$57,$5F,$EC,$46,$D7,$E2,$DA,$FF,$B0,$1A,$FF,$B6,$9A,$83,$86
	.dc.b	$F2,$7E,$21,$E0,$D7,$84,$6C,$91,$98,$E3,$3E,$58,$C4,$EA,$30,$49
	.dc.b	$F4,$4E,$89,$19,$8F,$54,$61,$DC,$65,$F2,$B5,$F2,$C6,$DE,$31,$97
	.dc.b	$CA,$72,$21,$A2,$9D,$0A,$74,$79,$4F,$AE,$32,$74,$39,$10,$E4,$53
	.dc.b	$91,$C1,$E4,$E4,$E4,$70,$21,$C0,$D3,$C1,$C9,$C0,$D1,$87,$4E,$08
	.dc.b	$6C,$8E,$9B,$3F,$25,$E4,$D7,$8C,$6C,$91,$98,$E3,$3F,$1A,$31,$3A
	.dc.b	$8C,$12,$7E,$14,$E8,$91,$98,$F7,$C6,$1D,$C6,$5F,$2B,$5F,$2C,$6D
	.dc.b	$E3,$19,$7C,$A7,$22,$1A,$29,$D0,$A7,$47,$94,$FA,$E3,$27,$43,$91
	.dc.b	$0E,$45,$39,$1C,$1E,$4E,$4E,$47,$02,$1C,$0D,$3C,$1C,$9C,$0D,$18
	.dc.b	$74,$E0,$86,$C8,$E9,$B3,$F2,$5E,$4D,$78,$C6,$C9,$19,$8E,$33,$F1
	.dc.b	$A3,$13,$A8,$C1,$27,$E1,$4E,$89,$19,$8F,$7C,$61,$DC,$65,$F2,$B5
	.dc.b	$F2,$C6,$DE,$31,$97,$CA,$72,$21,$A3,$F7,$7D,$1F,$DA,$9F,$E4,$C7
	.dc.b	$F2,$77,$0E,$1A,$E7,$2F,$E4,$51,$88,$34,$91,$A2,$4F,$F8,$0E,$77
	.dc.b	$18,$7F,$C0,$D3,$C2,$74,$D8,$CC,$ED,$F3,$9C,$A7,$71,$98,$F4,$35
	.dc.b	$F4,$46,$88,$F4,$6B,$D2,$7C,$0D,$9E,$4F,$73,$81,$0D,$4F,$27,$C2
	.dc.b	$D7,$C2,$6C,$C9,$4C,$9C,$8C,$1D,$CC,$10,$E2,$53,$89,$82,$18,$20
	.dc.b	$E8,$7A,$1F,$D6,$DB,$3F,$55,$B3,$FE,$39,$6B,$F4,$1A,$FE,$21,$AF
	.dc.b	$FA,$B6,$7A,$1F,$EA,$67,$FD,$1F,$87,$F9,$FF,$A3,$FD,$05,$1F,$E8
	.dc.b	$C3,$FB,$B7,$07,$FA,$45,$FF,$41,$9F,$E8,$B3,$47,$FA,$74,$AE,$E3
	.dc.b	$99,$FE,$8F,$2C,$F7,$3F,$D2,$E5,$9E,$91,$B8,$CC,$7F,$A7,$4C,$BB
	.dc.b	$7F,$D6,$87,$83,$0F,$F4,$FB,$89,$E4,$F0,$39,$19,$9E,$4F,$FB,$1D
	.dc.b	$EE,$FC,$09,$E6,$E4,$E4,$64,$F8,$67,$33,$A9,$B3,$06,$E7,$E0,$BF
	.dc.b	$BC,$AF,$F3,$7C,$7F,$B3,$B9,$BF,$D4,$9C,$EE,$7F,$F5,$27,$47,$F0
	.dc.b	$29,$FF,$55,$61,$6B,$FD,$5F,$1F,$EA,$A8,$FF,$3E,$3C,$AB,$8C,$7F
	.dc.b	$AA,$F9,$BF,$D4,$24,$7F,$D4,$38,$FF,$A2,$23,$FD,$5D,$CF,$FE,$92
	.dc.b	$E6,$FF,$48,$57,$FA,$AE,$B4,$F0,$30,$23,$FD,$27,$A7,$FA,$7A,$FB
	.dc.b	$4F,$FA,$9A,$79,$88,$FF,$4D,$1E,$0A,$EB,$3F,$EA,$DE,$8F,$F5,$E3
	.dc.b	$D6,$74,$F8,$2B,$BB,$FE,$A6,$8E,$F3,$C2,$7B,$4F,$9D,$AF,$5C,$3F
	.dc.b	$D6,$C7,$BE,$B8,$B8,$BF,$F5,$7C,$FF,$95,$EF,$FD,$85,$5E,$2B,$E7
	.dc.b	$1F,$EB,$A8,$E3,$5E,$A8,$C5,$78,$67,$81,$1E,$9D,$1F,$EB,$97,$DC
	.dc.b	$7D,$73,$81,$4F,$95,$1D,$9D,$8E,$67,$9D,$FE,$63,$6E,$7F,$CF,$5B
	.dc.b	$3F,$EC,$16,$BF,$CE,$1A,$FF,$50,$6B,$FE,$77,$6B,$DE,$7B,$72,$4A
	.dc.b	$EC,$D9,$EC,$D7,$B3,$5E,$CD,$7B,$35,$F6,$36,$BF,$AE,$11,$C1,$27
	.dc.b	$C2,$49,$F0,$B5,$F0,$B5,$F0,$B5,$F0,$B5,$F0,$B6,$7C,$0D,$9F,$04
	.dc.b	$F6,$21,$D8,$A7,$62,$9D,$8A,$76,$2C,$F6,$6C,$76,$7A,$1F,$EA,$DF
	.dc.b	$33,$FD,$3A,$4A,$FF,$5A,$36,$3F,$D6,$8D,$78,$46,$88,$E9,$AE,$8C
	.dc.b	$10,$C1,$0F,$F5,$E4,$75,$6C,$72,$69,$C4,$A7,$13,$FD,$77,$1C,$5B
	.dc.b	$1C,$5A,$71,$29,$C4,$FF,$64,$C7,$16,$C7,$16,$9C,$4A,$71,$E7,$4F
	.dc.b	$16,$CF,$16,$CF,$16,$BC,$67,$81,$27,$83,$63,$83,$5E,$11,$A2,$3A
	.dc.b	$6B,$A3,$04,$30,$43,$E5,$90,$F8,$A5,$3D,$E5,$9F,$4B,$67,$C8,$D9
	.dc.b	$F0,$B4,$F0,$14,$F0,$1D,$88,$76,$2C,$F6,$6C,$F6,$6C,$F6,$69,$D8
	.dc.b	$A7,$63,$A1,$27,$A3,$67,$A3,$67,$A3,$4E,$85,$3A,$4F,$22,$4F,$26
	.dc.b	$CF,$26,$CF,$26,$BC,$A7,$81,$27,$83,$63,$83,$5E,$11,$A2,$3A,$6B
	.dc.b	$A3,$04,$30,$43,$EA,$10,$F9,$25,$3E,$01,$67,$D2,$D9,$F2,$36,$7C
	.dc.b	$2D,$3C,$05,$3C,$07,$62,$1D,$8B,$3D,$9B,$3D,$9B,$3D,$9A,$76,$29
	.dc.b	$D8,$E8,$49,$E8,$D9,$E8,$FE,$1B,$9D,$E3,$FF,$A8,$89,$7F,$E9,$86
	.dc.b	$CF,$FA,$71,$AF,$FA,$35,$AF,$FA,$B5,$AF,$FA,$C5,$AF,$FA,$B1,$B3
	.dc.b	$E0,$9E,$C4,$9E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$CD,$8E,$D1,$D0,$87
	.dc.b	$42,$9D,$0A,$74,$29,$D0,$AF,$47,$91,$0E,$45,$39,$14,$E4,$57,$93
	.dc.b	$C0,$87,$02,$9C,$0A,$F0,$43,$EE,$9C,$2A,$FF,$6B,$46,$1B,$19,$9F
	.dc.b	$F6,$89,$C5,$C3,$FE,$D4,$9E,$0E,$9A,$E6,$36,$FF,$BC,$23,$2F,$FA
	.dc.b	$10,$FF,$4F,$73,$7F,$A3,$1F,$F3,$E9,$8E,$7F,$F4,$5B,$FE,$AC,$3F
	.dc.b	$D0,$65,$3F,$D0,$4F,$3E,$2F,$F5,$33,$70,$FF,$47,$CE,$CE,$87,$2E
	.dc.b	$3F,$8B,$7E,$18,$FF,$47,$10,$E1,$7D,$1F,$F4,$B9,$C8,$F0,$98,$74
	.dc.b	$FA,$9A,$F3,$9E,$F7,$EC,$6C,$78,$CF,$F5,$81,$FE,$A2,$38,$18,$AF
	.dc.b	$F4,$E7,$37,$FA,$6E,$1C,$6E,$7F,$F4,$D7,$37,$FA,$5A,$5C,$FF,$E9
	.dc.b	$BA,$D4,$7E,$12,$31,$5C,$23,$FD,$3F,$7F,$EA,$22,$5F,$E0,$A3,$9C
	.dc.b	$78,$6F,$9C,$62,$7C,$17,$D2,$3B,$56,$EF,$FC,$ED,$0E,$7F,$F3,$DB
	.dc.b	$FE,$72,$3F,$CE,$A6,$77,$6B,$FE,$81,$CB,$FC,$F9,$A7,$FA,$58,$FF
	.dc.b	$50,$F3,$FF,$A8,$2F,$FC,$F9,$87,$FA,$5E,$3B,$BF,$E8,$E8,$CE,$1F
	.dc.b	$E9,$37,$FD,$41,$3A,$8F,$F4,$C4,$78,$A7,$93,$FE,$8C,$DF,$FD,$64
	.dc.b	$E6,$7C,$AF,$FA,$22,$7D,$67,$07,$D5,$8F,$FA,$81,$FF,$53,$3F,$E9
	.dc.b	$B9,$EF,$1F,$04,$93,$C4,$E7,$1E,$7C,$3F,$D4,$B8,$FC,$2C,$7D,$18
	.dc.b	$75,$BF,$76,$1E,$03,$C9,$87,$68,$E8,$D3,$A5,$7C,$99,$E1,$87,$FA
	.dc.b	$CA,$75,$7E,$73,$F8,$5C,$78,$B1,$FF,$66,$F3,$3E,$3F,$FD,$4C,$48
	.dc.b	$FF,$53,$37,$0F,$F5,$33,$5F,$F5,$33,$FE,$AD,$F8,$A9,$3F,$EA,$D6
	.dc.b	$CF,$FA,$B5,$AF,$FA,$B5,$AF,$FA,$B5,$AF,$FA,$B5,$AF,$FA,$ED,$B4
	.dc.b	$69,$27,$C0,$49,$F0,$35,$F0,$35,$F0,$35,$F0,$35,$F0,$36,$7C,$0D
	.dc.b	$9F,$04,$F6,$21,$D8,$A7,$62,$9D,$8A,$76,$2C,$F6,$6C,$76,$7A,$10
	.dc.b	$E8,$53,$A1,$FD,$20,$FE,$93,$CE,$F9,$8F,$E7,$71,$86,$A8,$D5,$FF
	.dc.b	$07,$6C,$7F,$4A,$6B,$D4,$E1,$3C,$A3,$31,$E1,$6B,$DE,$7A,$4E,$C8
	.dc.b	$E0,$91,$DE,$32,$D8,$CB,$E1,$9E,$AF,$8A,$32,$F8,$67,$8B,$B6,$B9
	.dc.b	$8D,$BE,$53,$B1,$0E,$0E,$9E,$C6,$4E,$45,$39,$1C,$0C,$1C,$0C,$10
	.dc.b	$E2,$53,$89,$82,$18,$21,$CC,$C9,$4C,$CF,$C4,$24,$FC,$08,$CB,$5E
	.dc.b	$04,$8D,$CF,$BD,$B1,$EA,$6B,$E7,$38,$4F,$38,$CC,$79,$1A,$F8,$27
	.dc.b	$A4,$EC,$8E,$09,$1D,$E3,$2D,$8C,$BE,$19,$EA,$F8,$A3,$2F,$86,$78
	.dc.b	$BB,$6B,$98,$DB,$E5,$3B,$10,$E0,$E9,$EC,$64,$E4,$53,$91,$C0,$C1
	.dc.b	$C0,$C1,$0E,$25,$38,$98,$21,$82,$1C,$CC,$94,$CC,$FC,$42,$4F,$C0
	.dc.b	$8C,$B5,$E0,$48,$DC,$FB,$DB,$1E,$A6,$BE,$73,$84,$F3,$8C,$C7,$91
	.dc.b	$AF,$82,$7A,$4E,$C8,$E0,$91,$DE,$32,$D8,$CB,$E1,$9E,$AF,$8A,$32
	.dc.b	$F8,$67,$8B,$B6,$B9,$8D,$BE,$53,$B1,$0E,$0E,$9E,$C6,$4F,$A3,$CC
	.dc.b	$EA,$FE,$56,$D9,$FF,$18,$36,$7F,$CE,$6D,$7F,$95,$35,$FE,$70,$D7
	.dc.b	$FD,$02,$D4,$D2,$4F,$80,$93,$E0,$6C,$F8,$1A,$F8,$1A,$F8,$1A,$F8
	.dc.b	$1A,$F8,$1B,$3E,$09,$EC,$49,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D8
	.dc.b	$ED,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$F4,$79,$10,$E4,$53,$FD
	.dc.b	$63,$CE,$F0,$BF,$B3,$B6,$BF,$D6,$31,$82,$38,$24,$FF,$AB,$63,$FB
	.dc.b	$7B,$FE,$84,$E7,$FF,$4C,$3F,$E9,$69,$FF,$4F,$ED,$FE,$8C,$8F,$F4
	.dc.b	$A9,$FE,$93,$8C,$BF,$E8,$BE,$8F,$F5,$33,$FE,$8F,$39,$16,$79,$3F
	.dc.b	$E9,$F2,$70,$7F,$A7,$1D,$DA,$B9,$B6,$3F,$D4,$97,$E0,$8F,$3D,$F4
	.dc.b	$BE,$4D,$9D,$5F,$FA,$96,$FC,$C7,$23,$C8,$60,$F6,$18,$BE,$AF,$FA
	.dc.b	$A6,$FA,$B6,$BE,$27,$3F,$FA,$56,$3F,$D3,$07,$FA,$AB,$9B,$FD,$45
	.dc.b	$CD,$FE,$9C,$9D,$0E,$7F,$F5,$1D,$7F,$A9,$BA,$3F,$D4,$93,$96,$CF
	.dc.b	$FA,$7E,$7F,$D4,$31,$FE,$A4,$8C,$47,$82,$BF,$D5,$3D,$1F,$E7,$F7
	.dc.b	$FC,$F9,$1F,$E9,$6C,$BF,$CE,$0F,$FA,$58,$C1,$FE,$7C,$7F,$D4,$99
	.dc.b	$7F,$A5,$23,$FD,$13,$CF,$FE,$7C,$7F,$D3,$06,$C9,$B7,$F9,$F6,$BF
	.dc.b	$D1,$17,$E0,$9F,$F4,$0F,$47,$FA,$A2,$31,$8F,$97,$1E,$CF,$FA,$2A
	.dc.b	$F8,$C6,$31,$F0,$61,$FE,$9E,$3B,$E1,$CE,$75,$5E,$CC,$7F,$D1,$55
	.dc.b	$A3,$DA,$FF,$AF,$6B,$FD,$3B,$19,$3A,$42,$10,$F5,$DE,$A7,$FA,$4F
	.dc.b	$0F,$F2,$74,$E6,$30,$FF,$94,$67,$7D,$5F,$E5,$C6,$D7,$06,$CE,$EB
	.dc.b	$8B,$FE,$65,$8C,$B6,$32,$FF,$9A,$7A,$9F,$17,$FE,$66,$6D,$FF,$BA
	.dc.b	$DB,$3E,$66,$BF,$E6,$56,$BC,$E7,$91,$27,$93,$63,$93,$5E,$4D,$79
	.dc.b	$47,$02,$3C,$1A,$F0,$6B,$C0,$D1,$0D,$14,$D1,$82,$18,$21,$F6,$27
	.dc.b	$DA,$D9,$F3,$36,$7C,$2D,$3B,$14,$EA,$53,$A8,$9E,$AD,$9E,$AD,$9E
	.dc.b	$AD,$3A,$94,$EA,$53,$AC,$F4,$24,$FF,$A6,$7E,$06,$57,$FA,$85,$B5
	.dc.b	$FE,$A1,$6B,$FE,$91,$6B,$FE,$9F,$6B,$FE,$B0,$6B,$FE,$AD,$6F,$59
	.dc.b	$A4,$9F,$01,$27,$C0,$D7,$C0,$D7,$C0,$D7,$C0,$D7,$C0,$D9,$F0,$36
	.dc.b	$7C,$13,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$B3,$D9,$B1,$D9,$E8,$43
	.dc.b	$A1,$4E,$85,$3A,$16,$3A,$36,$3A,$3C,$88,$72,$29,$C8,$B1,$C9,$B1
	.dc.b	$C9,$E0,$43,$81,$63,$83,$63,$83,$A2,$46,$9B,$1A,$8C,$11,$C1,$27
	.dc.b	$EA,$12,$7E,$4B,$5F,$80,$D7,$D2,$D7,$C8,$D7,$C2,$D9,$F0,$36,$7C
	.dc.b	$13,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$B3,$D9,$B1,$D9,$E8,$43,$A1
	.dc.b	$4E,$85,$3A,$16,$3A,$36,$3A,$3C,$88,$72,$29,$C8,$B1,$C9,$B1,$C9
	.dc.b	$E0,$43,$81,$63,$83,$63,$83,$A2,$46,$9B,$1A,$8C,$11,$C1,$27,$EA
	.dc.b	$12,$7E,$4B,$5F,$80,$D7,$D2,$D7,$C8,$D7,$C2,$D9,$F0,$36,$7C,$13
	.dc.b	$D8,$FD,$DF,$AB,$FE,$8E,$9F,$EE,$30,$8C,$3F,$DD,$1A,$FF,$7A,$3F
	.dc.b	$BD,$B8,$70,$D7,$3D,$49,$EB,$39,$9E,$D3,$97,$B3,$5E,$C6,$A7,$8C
	.dc.b	$66,$3B,$B5,$EF,$1A,$23,$D1,$AF,$49,$F1,$12,$79,$47,$73,$81,$0E
	.dc.b	$0E,$A3,$84,$F4,$9F,$1B,$5F,$19,$C4,$C9,$4C,$9D,$1E,$0F,$84,$C1
	.dc.b	$B3,$07,$23,$24,$30,$43,$A9,$4E,$26,$08,$60,$91,$EC,$7C,$0F,$D7
	.dc.b	$F7,$9D,$1F,$19,$B5,$F8,$86,$D7,$F8,$E9,$AF,$DA,$6B,$F8,$96,$BF
	.dc.b	$E4,$46,$A6,$BF,$EA,$9E,$9F,$F4,$75,$7F,$A2,$08,$E0,$91,$F9,$5E
	.dc.b	$2F,$F4,$5C,$EA,$31,$7F,$E9,$86,$9F,$E9,$3A,$E2,$FF,$AC,$8F,$01
	.dc.b	$A9,$FF,$53,$3F,$E9,$A6,$BA,$38,$1F,$EA,$E8,$FF,$56,$91,$FF,$52
	.dc.b	$47,$07,$46,$23,$4F,$56,$BD,$4F,$F5,$DC,$FA,$5E,$FC,$5F,$EB,$37
	.dc.b	$D4,$79,$E7,$8B,$D8,$E8,$77,$21,$82,$1D,$EF,$F6,$97,$F0,$3A,$3F
	.dc.b	$D5,$BC,$FF,$EA,$46,$F4,$7F,$A9,$CF,$F5,$59,$2F,$FD,$60,$4E,$8F
	.dc.b	$F5,$B5,$71,$24,$FF,$A1,$23,$FD,$5B,$CE,$C7,$FD,$4F,$7F,$E8,$BE
	.dc.b	$8F,$F4,$E3,$67,$4E,$6B,$FD,$43,$18,$D3,$FD,$51,$3F,$EB,$33,$FD
	.dc.b	$4E,$7F,$A4,$A9,$3B,$E8,$FE,$39,$0B,$E7,$87,$8B,$0E,$67,$FA,$7F
	.dc.b	$0E,$58,$7F,$AA,$AF,$86,$DF,$EA,$89,$F3,$5F,$FA,$C1,$A7,$9E,$35
	.dc.b	$7C,$E3,$FD,$87,$3F,$EA,$FC,$7C,$78,$F8,$A7,$81,$CA,$BC,$73,$AB
	.dc.b	$E6,$FF,$AC,$A3,$FD,$6E,$6A,$7F,$D6,$F1,$93,$83,$CD,$FF,$60,$34
	.dc.b	$F6,$DF,$FA,$E2,$79,$CF,$19,$C9,$FE,$DF,$E7,$FF,$5D,$FC,$77,$FA
	.dc.b	$D7,$A0,$D2,$63,$FE,$B7,$8C,$B6,$33,$5F,$EB,$97,$FD,$65,$F1,$92
	.dc.b	$7F,$D6,$4D,$9F,$F5,$93,$5F,$F5,$93,$5F,$F5,$E3,$5F,$F4,$D3,$5F
	.dc.b	$F4,$83,$7A,$4D,$24,$F8,$09,$3E,$06,$BE,$06,$BE,$06,$BE,$06,$BE
	.dc.b	$06,$CF,$81,$B3,$E0,$9E,$C4,$3B,$14,$EC,$53,$B1,$4E,$C5,$9E,$CD
	.dc.b	$8E,$CF,$42,$1D,$0A,$74,$3F,$B9,$F4,$3E,$03,$FB,$3B,$6F,$FC,$12
	.dc.b	$D9,$FF,$7D,$B5,$E3,$3C,$09,$3C,$1B,$1C,$1A,$F0,$8D,$11,$D3,$5D
	.dc.b	$18,$21,$82,$1F,$E1,$99,$F2,$36,$7B,$B6,$7A,$34,$E6,$53,$98,$13
	.dc.b	$D1,$C3,$5C,$CF,$67,$2D,$73,$3E,$07,$2D,$72,$78,$8C,$94,$C9,$E4
	.dc.b	$29,$E7,$21,$C0,$D4,$F9,$49,$3E,$36,$CF,$85,$B3,$DD,$AF,$29,$E0
	.dc.b	$49,$E0,$D8,$E0,$D7,$84,$68,$8E,$9A,$E8,$C1,$0C,$10,$F6,$9F,$00
	.dc.b	$F7,$18,$9F,$3B,$67,$C4,$D9,$EE,$D3,$B1,$4E,$C7,$42,$1D,$C9,$3D
	.dc.b	$9C,$B5,$CC,$F8,$1C,$B5,$CC,$F8,$9C,$B5,$C9,$E4,$32,$53,$27,$98
	.dc.b	$A7,$A4,$87,$03,$53,$E5,$24,$F8,$DB,$3E,$16,$CF,$76,$BC,$A7,$81
	.dc.b	$27,$83,$63,$83,$5E,$11,$A2,$3A,$6B,$A3,$04,$30,$43,$DA,$7C,$03
	.dc.b	$DC,$62,$7C,$ED,$9F,$13,$67,$BB,$4E,$C5,$3B,$1D,$08,$77,$24,$F6
	.dc.b	$72,$D7,$33,$E0,$78,$F3,$B4,$FF,$46,$12,$FF,$D4,$ED,$9F,$F5,$93
	.dc.b	$5F,$F4,$D3,$5F,$F5,$8B,$5F,$F5,$AB,$5F,$F5,$E3,$67,$C1,$3D,$89
	.dc.b	$3D,$9A,$F6,$6B,$D9,$AF,$66,$BD,$9B,$1D,$A3,$A1,$0E,$85,$3A,$14
	.dc.b	$E8,$53,$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8,$AF,$27,$81,$0E,$05
	.dc.b	$38,$15,$E0,$FF,$A6,$7A,$3F,$D4,$84,$9F,$F5,$23,$63,$FD,$48,$D7
	.dc.b	$FD,$48,$FF,$B4,$1F,$F3,$84,$7F,$9F,$5F,$F3,$F9,$FE,$7C,$3F,$CF
	.dc.b	$E7,$FA,$CB,$9F,$FD,$2E,$EB,$9F,$FD,$19,$C1,$FE,$83,$6C,$6C,$FF
	.dc.b	$42,$1F,$E8,$03,$B9,$D4,$E5,$3F,$E8,$89,$E4,$D8,$E4,$78,$A7,$FD
	.dc.b	$58,$D8,$F1,$F0,$FF,$AC,$1B,$7E,$B8,$C4,$F8,$8F,$F4,$79,$FE,$95
	.dc.b	$3C,$75,$E8,$3A,$BE,$43,$D0,$6C,$97,$E6,$72,$D7,$AB,$5F,$F2,$D1
	.dc.b	$EA,$3F,$D1,$D5,$FD,$E6,$7F,$D3,$DC,$DF,$E9,$49,$FF,$41,$76,$7F
	.dc.b	$7D,$5D,$1F,$EA,$07,$27,$F8,$05,$4B,$A3,$FD,$45,$3C,$4E,$31,$97
	.dc.b	$FD,$4A,$66,$7A,$73,$7F,$A9,$AB,$FD,$59,$3D,$8F,$F4,$C9,$39,$FF
	.dc.b	$D1,$EF,$F9,$DC,$C4,$FF,$A1,$F3,$FF,$40,$42,$77,$18,$7F,$D5,$6D
	.dc.b	$E8,$FF,$51,$4F,$F9,$F2,$3F,$CF,$ED,$E1,$FF,$40,$3C,$EF,$B4,$7F
	.dc.b	$A2,$27,$96,$DF,$EA,$78,$DE,$1F,$EA,$E8,$FF,$51,$4E,$AF,$FD,$36
	.dc.b	$D3,$C9,$7D,$23,$C0,$7F,$AD,$DC,$3E,$ED,$BF,$D3,$EF,$FA,$70,$87
	.dc.b	$C5,$C7,$D6,$F9,$E3,$6E,$6F,$C7,$8F,$9A,$B7,$8F,$C6,$8C,$4F,$FA
	.dc.b	$85,$FF,$53,$E3,$E7,$7E,$01,$FE,$C8,$AE,$FC,$BF,$E9,$CC,$3F,$00
	.dc.b	$1C,$3F,$83,$6B,$F8,$D6,$BF,$E4,$B6,$BC,$79,$99,$74,$8C,$36,$76
	.dc.b	$37,$DC,$F8,$A3,$13,$B9,$CC,$F8,$DA,$F8,$DA,$F8,$DA,$FA,$DB,$3D
	.dc.b	$23,$B4,$60,$93,$DC,$91,$A6,$BE,$16,$BE,$16,$BE,$16,$BE,$16,$CF
	.dc.b	$86,$39,$1D,$48,$60,$87,$53,$25,$32,$77,$32,$53,$27,$84,$C9,$4C
	.dc.b	$BE,$37,$A1,$0E,$45,$38,$15,$E0,$E8,$86,$8A,$E9,$C1,$39,$5E,$A7
	.dc.b	$FA,$D0,$95,$FE,$B4,$6C,$FF,$AD,$1A,$FF,$AF,$1A,$FF,$A9,$5A,$FF
	.dc.b	$A6,$9A,$FF,$A7,$9B,$3E,$09,$EC,$49,$EC,$D7,$B3,$5E,$CD,$7B,$35
	.dc.b	$EC,$D8,$ED,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$F4,$79,$10,$E4
	.dc.b	$53,$91,$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF,$07,$44,$34,$57,$4E
	.dc.b	$08,$E0,$93,$F5,$09,$3F,$25,$B3,$F0,$1A,$FA,$5A,$F9,$1A,$F8,$5A
	.dc.b	$F8,$1B,$3E,$09,$EC,$49,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D8,$ED
	.dc.b	$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$F4,$79,$10,$E4,$53,$91,$4E
	.dc.b	$45,$79,$3C,$08,$70,$29,$C0,$AF,$07,$44,$34,$57,$4E,$08,$E0,$93
	.dc.b	$F5,$09,$3F,$25,$B3,$F0,$1A,$FA,$5A,$F9,$1A,$F8,$5A,$F8,$1B,$3E
	.dc.b	$09,$EC,$49,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D8,$ED,$1D,$08,$74
	.dc.b	$29,$D0,$FB,$FC,$AF,$84,$FE,$B2,$DA,$FE,$D4,$D9,$FE,$EE,$D7,$FC
	.dc.b	$70,$D7,$FC,$84,$D4,$1C,$36,$7B,$92,$7B,$B6,$3B,$B5,$EE,$D7,$BB
	.dc.b	$5E,$F3,$D8,$91,$98,$F0,$4F,$62,$3D,$9A,$F6,$6B,$D9,$AF,$68,$E9
	.dc.b	$3D,$A7,$33,$D8,$E8,$43,$A1,$4E,$85,$3A,$3C,$A3,$A4,$64,$E8,$72
	.dc.b	$21,$C8,$A7,$23,$83,$F4,$3A,$DE,$77,$F8,$59,$B5,$FE,$27,$6D,$7F
	.dc.b	$8D,$1A,$FF,$9D,$5A,$FF,$9F,$DA,$FD,$E6,$A6,$BF,$E9,$8E,$7F,$F5
	.dc.b	$37,$3F,$FA,$63,$8B,$FD,$3F,$3F,$EA,$53,$FD,$40,$6C,$FF,$4B,$9F
	.dc.b	$E9,$37,$6F,$FA,$7C,$FF,$4C,$99,$BF,$F5,$51,$1F,$F5,$C4,$70,$24
	.dc.b	$66,$7F,$D5,$55,$FE,$B2,$AD,$CE,$63,$CE,$66,$77,$3E,$63,$FD,$5A
	.dc.b	$78,$4E,$24,$76,$FB,$48,$7A,$4F,$71,$FE,$A9,$E2,$FF,$5A,$73,$7F
	.dc.b	$A7,$B8,$3F,$D5,$8D,$9F,$F5,$57,$3F,$FA,$95,$FF,$52,$35,$DB,$5D
	.dc.b	$F3,$FF,$98,$E7,$94,$7F,$AB,$A3,$FD,$53,$CF,$FE,$63,$8F,$F4,$2F
	.dc.b	$3F,$FA,$83,$9F,$FC,$FB,$0C,$BF,$D3,$EF,$FA,$14,$FF,$4A,$C6,$A3
	.dc.b	$33,$FE,$9B,$9F,$F5,$83,$FE,$87,$7F,$D2,$E7,$18,$DF,$3F,$FA,$30
	.dc.b	$9B,$FF,$A4,$27,$75,$E3,$6C,$E9,$F2,$5F,$FA,$40,$F1,$15,$DD,$7F
	.dc.b	$AB,$DA,$7A,$0B,$7F,$EA,$CD,$FF,$D4,$38,$77,$9F,$35,$FF,$A6,$9D
	.dc.b	$B5,$C9,$E6,$C7,$FD,$63,$3F,$06,$FD,$B8,$7F,$A9,$6B,$E6,$E3,$E9
	.dc.b	$9C,$C7,$C6,$BE,$33,$FE,$AE,$6D,$71,$9F,$F5,$B1,$F3,$27,$58,$E7
	.dc.b	$0F,$06,$3F,$EA,$7F,$8B,$FF,$48,$92,$7F,$D4,$4D,$BF,$F5,$03,$5F
	.dc.b	$F5,$83,$FE,$98,$F7,$64,$FF,$AA,$5B,$3F,$EA,$76,$BF,$EB,$76,$BF
	.dc.b	$EB,$76,$BF,$EB,$76,$BF,$EB,$F6,$F4,$1A,$49,$F0,$12,$7C,$0D,$7C
	.dc.b	$0D,$7C,$0D,$7C,$0D,$7C,$0D,$9F,$03,$67,$C1,$3D,$88,$76,$29,$D8
	.dc.b	$A7,$62,$9D,$8B,$3D,$9B,$1D,$9E,$84,$3A,$14,$E8,$7F,$6A,$3F,$B5
	.dc.b	$F3,$BE,$17,$FD,$9E,$DB,$FF,$70,$36,$7F,$DD,$AD,$79,$07,$0F,$56
	.dc.b	$BD,$A7,$A9,$27,$AB,$63,$AB,$5E,$B3,$C9,$ED,$19,$7B,$35,$ED,$3D
	.dc.b	$08,$F4,$6B,$D2,$38,$3D,$23,$2F,$46,$BD,$0E,$44,$39,$3A,$79,$19
	.dc.b	$39,$14,$E4,$70,$30,$70,$30,$43,$89,$4E,$26,$08,$60,$87,$33,$25
	.dc.b	$33,$3F,$2C,$93,$F0,$9B,$3E,$B6,$CF,$99,$AF,$8C,$E8,$47,$2F,$81
	.dc.b	$AF,$82,$7B,$12,$7B,$36,$3B,$35,$ED,$3C,$9E,$D1,$97,$B3,$5E,$D3
	.dc.b	$D0,$8F,$46,$BD,$23,$83,$D2,$32,$F4,$6B,$D0,$E4,$43,$93,$A7,$91
	.dc.b	$93,$91,$4E,$47,$03,$07,$03,$04,$38,$94,$E2,$60,$86,$08,$73,$32
	.dc.b	$53,$33,$F2,$C9,$3F,$09,$B3,$EB,$6C,$F9,$9A,$F8,$CE,$84,$72,$F8
	.dc.b	$1A,$F8,$27,$B1,$27,$B3,$63,$B3,$5E,$D3,$C9,$ED,$19,$7B,$35,$ED
	.dc.b	$3D,$08,$F4,$6B,$D2,$38,$3D,$23,$2F,$46,$BD,$0E,$44,$39,$3A,$7E
	.dc.b	$1F,$3B,$E1,$7F,$D4,$24,$9F,$F5,$0B,$67,$FD,$42,$D7,$FD,$42,$D7
	.dc.b	$FD,$42,$D7,$FD,$26,$D7,$FD,$44,$D9,$F0,$4F,$62,$4F,$66,$BD,$9A
	.dc.b	$F6,$6B,$D9,$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57
	.dc.b	$A3,$C8,$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3F
	.dc.b	$E9,$1F,$85,$FF,$4E,$12,$7F,$D3,$8D,$8F,$F4,$E3,$5F,$F4,$E2,$8F
	.dc.b	$EC,$B1,$FE,$90,$E6,$FF,$4B,$47,$FA,$0E,$31,$CF,$FE,$83,$3F,$D0
	.dc.b	$AF,$FA,$0F,$8F,$FD,$BF,$5C,$23,$FD,$05,$3A,$7F,$D0,$3C,$5F,$E9
	.dc.b	$27,$27,$43,$91,$FE,$8B,$9F,$F4,$4B,$FE,$87,$3F,$D1,$EF,$53,$31
	.dc.b	$CD,$B7,$E7,$9F,$F5,$2C,$6E,$BC,$51,$CE,$FC,$F1,$EB,$8F,$11,$E6
	.dc.b	$AE,$4F,$FA,$58,$87,$FA,$0C,$F3,$47,$FA,$43,$93,$FD,$17,$CF,$65
	.dc.b	$CB,$FE,$8D,$7F,$D2,$06,$23,$4F,$FA,$6A,$7F,$D2,$DC,$FF,$E9,$98
	.dc.b	$D9,$93,$81,$FE,$9C,$29,$D2,$BB,$73,$A5,$4A,$DF,$37,$F9,$F4,$FF
	.dc.b	$33,$BF,$E7,$4D,$7F,$D0,$CF,$F9,$F0,$46,$9F,$F3,$8E,$8E,$6F,$F4
	.dc.b	$7C,$75,$9F,$F4,$F3,$C5,$E0,$60,$BC,$FF,$E8,$19,$FF,$4B,$4E,$E3
	.dc.b	$FC,$F6,$FF,$9F,$6F,$FD,$44,$E5,$B5,$C9,$F3,$B6,$3F,$D3,$86,$CF
	.dc.b	$F5,$29,$E5,$CF,$FD,$16,$E0,$F5,$DF,$A9,$A7,$FA,$78,$FF,$4E,$6D
	.dc.b	$FE,$A4,$C3,$D1,$1C,$E3,$18,$78,$A7,$33,$F2,$1F,$5E,$3F,$1E,$3A
	.dc.b	$19,$9F,$5C,$F0,$C7,$E3,$E3,$FB,$3A,$F3,$47,$D1,$E6,$7C,$66,$67
	.dc.b	$FC,$B7,$18,$CB,$F7,$30,$8C,$3F,$BC,$6B,$C2,$79,$3F,$E6,$D8,$CB
	.dc.b	$63,$2F,$E2,$63,$2D,$F7,$F2,$F7,$6C,$1A,$4A,$F2,$C6,$5B,$19,$9F
	.dc.b	$33,$5F,$6B,$5F,$33,$5F,$14,$72,$23,$C9,$AF,$26,$BC,$9A,$F2,$38
	.dc.b	$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$94,$F9,$1A
	.dc.b	$77,$29,$D0,$A7,$32,$9C,$E7,$E7,$3C,$5A,$F1,$6B,$C7,$9D,$EC,$FF
	.dc.b	$AB,$09,$5F,$EA,$C6,$CF,$FA,$B1,$AF,$FA,$81,$AF,$FA,$AD,$AF,$FA
	.dc.b	$E1,$AF,$FA,$C5,$B3,$E0,$9E,$C4,$9E,$CD,$7B,$35,$EC,$D7,$B3,$5E
	.dc.b	$CD,$8E,$D1,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$AF,$47,$91,$0E,$45
	.dc.b	$39,$14,$E4,$57,$93,$C0,$87,$02,$9C,$0A,$F0,$74,$43,$45,$74,$E0
	.dc.b	$8E,$09,$3F,$50,$93,$F2,$5B,$3F,$01,$AF,$A5,$AF,$91,$AF,$85,$AF
	.dc.b	$81,$B3,$E0,$9E,$C4,$9E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$CD,$8E,$D1
	.dc.b	$D0,$87,$42,$9D,$0A,$74,$29,$D0,$AF,$47,$91,$0E,$45,$39,$14,$E4
	.dc.b	$57,$93,$C0,$87,$02,$9C,$0A,$F0,$74,$43,$45,$74,$E0,$8E,$09,$3F
	.dc.b	$50,$93,$F2,$5B,$3F,$01,$AF,$A5,$AF,$91,$AF,$85,$AF,$81,$B3,$E0
	.dc.b	$9E,$C4,$9E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$CD,$8E,$D1,$D0,$87,$42
	.dc.b	$9D,$0F,$BF,$CC,$F8,$6F,$EB,$ED,$AF,$ED,$ED,$9F,$F3,$4B,$5F,$F1
	.dc.b	$81,$FE,$31,$70,$E1,$AE,$67,$A9,$27,$AB,$63,$AB,$5E,$A2,$79,$B6
	.dc.b	$34,$F7,$6B,$DE,$7A,$91,$EA,$D7,$AC,$F4,$24,$F2,$8E,$D3,$B7,$B3
	.dc.b	$5E,$C7,$42,$1D,$23,$94,$F4,$9E,$13,$D0,$D9,$D0,$A7,$43,$93,$C2
	.dc.b	$39,$46,$8E,$46,$8C,$3A,$70,$43,$EB,$FB,$BF,$E6,$79,$FF,$15,$B6
	.dc.b	$3F,$D4,$AD,$7E,$8B,$5E,$27,$F8,$91,$C3,$FE,$71,$6A,$6B,$FE,$8C
	.dc.b	$E0,$FF,$49,$BF,$EA,$2E,$7F,$F4,$C3,$FE,$94,$3F,$D3,$07,$FA,$40
	.dc.b	$C7,$0F,$FA,$87,$9F,$FD,$3A,$D8,$CE,$1F,$E9,$A7,$FD,$53,$3F,$E9
	.dc.b	$F7,$FD,$4A,$D7,$C4,$75,$32,$78,$DD,$3C,$8C,$10,$FF,$54,$1F,$EB
	.dc.b	$C3,$89,$8A,$FF,$5A,$B9,$3B,$BE,$E7,$62,$BF,$CC,$50,$9F,$F5,$67
	.dc.b	$0F,$FA,$7B,$97,$FD,$3B,$7F,$EA,$2E,$6F,$F5,$07,$37,$FA,$CA,$3F
	.dc.b	$D4,$C7,$FA,$A4,$87,$FA,$C0,$E7,$CF,$1F,$F5,$65,$F5,$9F,$F5,$43
	.dc.b	$FE,$B9,$E5,$FF,$54,$73,$7F,$9D,$CF,$F3,$EF,$3F,$FA,$26,$3F,$D3
	.dc.b	$D1,$9D,$7F,$CF,$A7,$FA,$C7,$7F,$F4,$16,$3F,$EA,$5C,$3F,$D2,$71
	.dc.b	$8B,$E3,$5B,$C3,$F8,$4C,$78,$23,$17,$FE,$B7,$3F,$D3,$F1,$FE,$A1
	.dc.b	$AE,$58,$79,$27,$FD,$1F,$3F,$E8,$CD,$FF,$D2,$F1,$C2,$7C,$CF,$FA
	.dc.b	$3B,$1F,$1E,$3F,$EB,$4B,$E8,$F8,$AB,$A5,$7F,$AB,$70,$F4,$5F,$08
	.dc.b	$FF,$52,$1F,$EB,$93,$A6,$1E,$09,$F2,$E3,$F1,$27,$FD,$3B,$7F,$EA
	.dc.b	$1B,$FF,$50,$56,$A7,$35,$FE,$AD,$3E,$0F,$8C,$F1,$9D,$DF,$E9,$68
	.dc.b	$C3,$63,$39,$7F,$26,$B5,$6A,$30,$D8,$FE,$56,$D7,$F9,$CB,$5F,$F7
	.dc.b	$53,$5F,$F5,$27,$C0,$7F,$A8,$09,$3F,$E9,$E6,$C7,$FA,$C5,$AF,$FA
	.dc.b	$E1,$AF,$FA,$C5,$AF,$48,$E4,$47,$93,$5E,$4D,$79,$35,$E4,$70,$21
	.dc.b	$C0,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$ED,$38,$F6,$B6
	.dc.b	$3C,$CD,$3C,$25,$3B,$14,$EA,$53,$AF,$4A,$7A,$B6,$7A,$B6,$7A,$B5
	.dc.b	$FF,$74,$FC,$07,$FB,$60,$93,$FE,$D6,$6C,$7F,$BA,$5A,$F0,$8D,$11
	.dc.b	$D3,$5D,$18,$21,$82,$1F,$EF,$18,$EA,$D8,$E4,$D3,$89,$4E,$27,$FB
	.dc.b	$D6,$38,$B6,$38,$B4,$E2,$53,$88,$94,$F2,$8C,$36,$33,$3D,$27,$2D
	.dc.b	$8C,$C7,$63,$25,$72,$F8,$1A,$FA,$C8,$79,$4A,$72,$38,$4F,$8C,$93
	.dc.b	$E1,$6C,$77,$6B,$C2,$34,$47,$4D,$74,$60,$86,$08,$79,$27,$D2,$F9
	.dc.b	$C8,$79,$CD,$4F,$89,$B3,$DD,$A7,$62,$9D,$8E,$73,$D9,$E8,$62,$7A
	.dc.b	$36,$7A,$34,$E8,$53,$A1,$C8,$93,$D4,$93,$D2,$32,$D8,$CC,$F6,$9C
	.dc.b	$B6,$33,$1E,$03,$25,$72,$F8,$9A,$FC,$02,$1E,$72,$9C,$8E,$13,$E3
	.dc.b	$24,$F8,$5B,$1D,$DA,$F0,$8D,$11,$D3,$5D,$18,$21,$82,$1E,$49,$F4
	.dc.b	$BE,$72,$1E,$73,$53,$E2,$6C,$F7,$69,$D8,$A7,$63,$9C,$F6,$7A,$18
	.dc.b	$9E,$8D,$9E,$8D,$3A,$14,$E8,$72,$24,$F5,$24,$FE,$07,$A7,$FC,$AE
	.dc.b	$49,$57,$F6,$B5,$19,$8C,$36,$3F,$B8,$35,$FF,$4B,$B5,$FF,$52,$B5
	.dc.b	$1D,$0E,$E7,$22,$9C,$8A,$72,$29,$C9,$E8,$F3,$8F,$2B,$A3,$07,$57
	.dc.b	$24,$70,$43,$C0,$E5,$A6,$4F,$13,$96,$99,$3C,$87,$86,$3C,$A4,$70
	.dc.b	$48,$E4,$F0,$6B,$E8,$3C,$24,$3B,$94,$E0,$78,$4E,$06,$88,$7F,$A5
	.dc.b	$7E,$03,$FD,$30,$48,$FF,$50,$B6,$3F,$D4,$2D,$7F,$D4,$2F,$CB,$7F
	.dc.b	$CF,$7C,$DF,$E8,$6D,$BF,$D3,$71,$FE,$81,$8F,$F4,$AD,$7F,$A4,$E7
	.dc.b	$33,$FE,$9C,$E7,$FF,$41,$3C,$8F,$F4,$61,$9A,$E8,$FF,$A5,$4D,$1D
	.dc.b	$4F,$F4,$F9,$FE,$8D,$3F,$D2,$C6,$4F,$F4,$6E,$FF,$EA,$56,$DF,$8D
	.dc.b	$E8,$70,$3C,$07,$8A,$38,$3F,$E7,$B3,$77,$EA,$8F,$F4,$BD,$7F,$A2
	.dc.b	$E5,$CD,$FE,$94,$7F,$D2,$DC,$FF,$E9,$8E,$2F,$F4,$74,$63,$9B,$F5
	.dc.b	$95,$B8,$FF,$4B,$9F,$E9,$A3,$A3,$CE,$78,$93,$9F,$FC,$ED,$8F,$F9
	.dc.b	$9E,$BF,$CD,$27,$F9,$FD,$FF,$44,$BF,$E9,$1C,$FF,$CE,$9C,$DC,$23
	.dc.b	$FC,$EE,$7F,$9E,$74,$FF,$3E,$CE,$A3,$F6,$BC,$FF,$E8,$B7,$55,$FE
	.dc.b	$87,$E7,$FF,$45,$DF,$09,$C9,$FE,$84,$7F,$CF,$84,$9F,$03,$B3,$11
	.dc.b	$FE,$9D,$6A,$AF,$F5,$2B,$EC,$8C,$C6,$1B,$5F,$E9,$68,$FF,$4C,$CF
	.dc.b	$A2,$30,$79,$AB,$D1,$87,$C4,$AF,$F5,$75,$75,$9F,$29,$A7,$07,$39
	.dc.b	$FF,$4D,$5F,$81,$B3,$C2,$74,$7C,$C9,$D3,$F4,$CE,$67,$B6,$FC,$B5
	.dc.b	$E3,$7F,$D4,$98,$58,$FF,$50,$F3,$3B,$7F,$D3,$C4,$9F,$F4,$DC,$65
	.dc.b	$B1,$9B,$FF,$4E,$CE,$5B,$1F,$C8,$F9,$5E,$95,$25,$FF,$3F,$8C,$B6
	.dc.b	$33,$5F,$EA,$D9,$CB,$63,$2F,$F3,$C3,$F9,$EC,$69,$C1,$FE,$51,$8D
	.dc.b	$BB,$8D,$36,$32,$F2,$6B,$C6,$7C,$84,$7C,$4D,$7B,$CF,$47,$94,$65
	.dc.b	$EC,$D7,$27,$22,$1C,$A7,$C3,$3C,$E3,$C4,$68,$E6,$53,$99,$C6,$7B
	.dc.b	$3C,$0C,$1B,$30,$72,$32,$43,$04,$3A,$94,$E2,$60,$86,$09,$0D,$BF
	.dc.b	$04,$49,$FC,$03,$67,$EF,$35,$F9,$6D,$7E,$C3,$5F,$BE,$D7,$FD,$1C
	.dc.b	$D9,$F0,$4F,$62,$4F,$66,$BD,$9A,$F6,$6B,$D9,$AF,$66,$C7,$68,$E8
	.dc.b	$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8,$87,$22,$9C,$8A,$72,$2B
	.dc.b	$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA,$70,$47,$04,$9F,$A8
	.dc.b	$49,$F9,$2D,$9F,$80,$D7,$D2,$D7,$C8,$D7,$C2,$D7,$C0,$D9,$F0,$4F
	.dc.b	$62,$4F,$66,$BD,$9A,$F6,$6B,$D9,$AF,$66,$C7,$68,$E8,$43,$A1,$4E
	.dc.b	$85,$3A,$14,$E8,$57,$A3,$C8,$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43
	.dc.b	$81,$4E,$05,$78,$3A,$21,$A2,$BA,$70,$47,$04,$9F,$A8,$49,$F9,$2D
	.dc.b	$9F,$80,$D7,$D2,$D7,$C8,$D7,$C2,$D7,$C0,$D9,$F0,$4F,$62,$4F,$66
	.dc.b	$BD,$9A,$F6,$6B,$D9,$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$87,$DF,$E5
	.dc.b	$65,$FE,$77,$25,$7F,$A7,$9B,$3F,$E9,$E6,$BF,$E9,$E6,$BF,$E9,$E6
	.dc.b	$BF,$E9,$56,$BF,$EA,$A6,$CF,$82,$7B,$12,$7B,$35,$EC,$D7,$B3,$5E
	.dc.b	$CD,$7B,$36,$3B,$47,$42,$1D,$0A,$74,$29,$D0,$A7,$42,$BD,$1E,$44
	.dc.b	$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1,$FF,$65,$F5
	.dc.b	$3C,$6F,$F6,$03,$6B,$FD,$B4,$D9,$FF,$09,$87,$0F,$F8,$59,$AF,$F8
	.dc.b	$6D,$AF,$69,$EA,$7F,$A8,$DF,$F4,$55,$FF,$A2,$1B,$5F,$E8,$FE,$6F
	.dc.b	$F4,$FC,$60,$FF,$4C,$6F,$FE,$8B,$9D,$73,$7F,$A8,$2B,$73,$9A,$D9
	.dc.b	$FE,$92,$24,$69,$B3,$C4,$8F,$FA,$6E,$BA,$B4,$E7,$3C,$49,$3C,$23
	.dc.b	$FD,$55,$C9,$FE,$91,$E4,$FF,$49,$F3,$7F,$A6,$DB,$CD,$FE,$9A,$83
	.dc.b	$49,$5F,$3A,$B9,$3F,$F0,$01,$AF,$FA,$A2,$7A,$39,$6B,$FE,$99,$3F
	.dc.b	$D1,$A7,$FA,$17,$81,$97,$FA,$5E,$3F,$CD,$9A,$7F,$9D,$1F,$F4,$67
	.dc.b	$2F,$F9,$D5,$E1,$3B,$E6,$FF,$4E,$6D,$FE,$76,$3F,$D1,$B8,$74,$AE
	.dc.b	$F3,$FE,$8F,$BE,$4F,$06,$F3,$7F,$9E,$27,$75,$C6,$FA,$57,$08,$C3
	.dc.b	$C3,$1F,$F5,$C4,$EA,$BF,$D5,$13,$97,$FD,$11,$19,$7D,$27,$6C,$7D
	.dc.b	$58,$7F,$A2,$A3,$FD,$1B,$8F,$95,$FF,$49,$D7,$18,$F5,$47,$C4,$33
	.dc.b	$7F,$E9,$F9,$D6,$3E,$69,$C9,$0F,$9A,$7F,$A7,$67,$E6,$71,$BE,$33
	.dc.b	$FD,$18,$4C,$BF,$D1,$4D,$AF,$F4,$43,$5F,$F3,$AB,$FE,$93,$F4,$64
	.dc.b	$FF,$A4,$5B,$3F,$E8,$F6,$BF,$E7,$C6,$BF,$E8,$96,$BF,$E9,$06,$AA
	.dc.b	$46,$1B,$3D,$C9,$3D,$DA,$F7,$6B,$DD,$AF,$76,$BD,$E7,$B1,$23,$33
	.dc.b	$E0,$9E,$C4,$3B,$14,$EC,$53,$B1,$4E,$D1,$D2,$7B,$46,$67,$B3,$D0
	.dc.b	$87,$42,$9D,$0A,$74,$39,$47,$FB,$1F,$D5,$95,$FE,$C5,$6C,$FF,$B0
	.dc.b	$DA,$FF,$AC,$1A,$FF,$AF,$5A,$FF,$B1,$1A,$FD,$56,$D1,$A4,$9F,$01
	.dc.b	$27,$C0,$D7,$C0,$D7,$C0,$D7,$C0,$D7,$C0,$D9,$F0,$36,$7C,$13,$D8
	.dc.b	$87,$62,$9D,$8A,$76,$29,$D8,$B3,$D9,$B1,$D9,$E8,$43,$A1,$4E,$85
	.dc.b	$3A,$16,$3A,$36,$3A,$3C,$88,$72,$29,$C8,$B1,$C9,$B1,$C9,$E0,$43
	.dc.b	$81,$63,$83,$63,$83,$A2,$46,$9B,$1A,$8C,$11,$C1,$27,$EA,$12,$7E
	.dc.b	$4B,$5F,$80,$D7,$D2,$D7,$C8,$D7,$C2,$D9,$F0,$36,$7C,$13,$D8,$87
	.dc.b	$62,$9D,$8A,$76,$29,$D8,$B3,$D9,$B1,$D9,$E8,$43,$A1,$4E,$85,$3A
	.dc.b	$16,$3A,$36,$3A,$3C,$88,$72,$29,$C8,$B1,$C9,$B1,$C9,$E0,$43,$81
	.dc.b	$63,$83,$63,$83,$A2,$46,$9B,$1A,$8C,$11,$C1,$27,$EA,$12,$7E,$4B
	.dc.b	$5F,$80,$D7,$D2,$D7,$C8,$D7,$C2,$D9,$F0,$36,$7C,$13,$D8,$FD,$DF
	.dc.b	$33,$D5,$FF,$03,$40,$82,$35,$5F,$E4,$E8,$C4,$F0,$77,$3A,$7F,$CA
	.dc.b	$6D,$79,$9C,$27,$94,$66,$38,$CF,$37,$C5,$19,$6C,$65,$E4,$76,$71
	.dc.b	$1B,$8C,$9E,$07,$04,$8D,$92,$7C,$6D,$9E,$91,$82,$38,$21,$D6,$38
	.dc.b	$36,$36,$D3,$67,$B9,$CB,$4C,$9E,$E7,$2D,$32,$7B,$0E,$07,$52,$18
	.dc.b	$21,$C4,$C9,$4C,$9C,$CC,$10,$C1,$27,$E6,$79,$8E,$9F,$A4,$DA,$FC
	.dc.b	$23,$6B,$FD,$26,$D7,$F0,$2D,$7F,$1A,$D7,$F2,$0D,$4D,$7F,$29,$B7
	.dc.b	$FA,$27,$9D,$CF,$FE,$A0,$8C,$36,$7F,$D1,$27,$FA,$16,$76,$FE,$4A
	.dc.b	$76,$D3,$67,$E8,$78,$FF,$D4,$11,$A8,$FD,$1B,$5F,$19,$C0,$FF,$69
	.dc.b	$CF,$67,$66,$67,$9B,$70,$FF,$4C,$BD,$DA,$F2,$BF,$55,$7F,$A6,$A3
	.dc.b	$D5,$18,$3F,$D3,$8E,$CF,$54,$6D,$E8,$7F,$A9,$23,$75,$E2,$9D,$4F
	.dc.b	$13,$CE,$E2,$35,$3C,$2B,$F8,$24,$FB,$6B,$E0,$BB,$8C,$CF,$96,$7C
	.dc.b	$E7,$FA,$87,$9B,$F8,$67,$13,$9F,$FD,$5F,$28,$C3,$B3,$11,$FC,$4B
	.dc.b	$9D,$CF,$FE,$B4,$9F,$F5,$93,$5F,$E1,$B1,$AB,$FF,$5B,$C6,$27,$FD
	.dc.b	$63,$3F,$EB,$62,$3F,$EB,$4E,$67,$3F,$FA,$87,$6F,$F4,$73,$FE,$93
	.dc.b	$7F,$D1,$7A,$7F,$A5,$B6,$FF,$5A,$63,$FE,$BB,$9D,$C6,$2F,$9B,$FE
	.dc.b	$A4,$E7,$FF,$51,$49,$FF,$4B,$18,$2D,$7F,$A6,$CF,$F4,$AC,$6D,$FF
	.dc.b	$55,$9F,$E9,$9A,$EA,$FF,$AB,$27,$93,$A3,$8B,$87,$FD,$5F,$3E,$67
	.dc.b	$FD,$40,$73,$C3,$E0,$4E,$EB,$FD,$61,$3C,$8F,$23,$83,$C3,$5F,$12
	.dc.b	$3F,$D7,$44,$7B,$1E,$97,$C3,$3D,$6B,$E6,$3D,$F5,$FF,$5B,$4F,$73
	.dc.b	$E6,$47,$FB,$7E,$1E,$CF,$F4,$16,$E5,$FD,$4D,$B5,$FD,$38,$FC,$64
	.dc.b	$EC,$93,$B6,$C6,$F9,$9E,$97,$F4,$D6,$D7,$F5,$F6,$CE,$E7,$44,$9D
	.dc.b	$36,$35,$18,$23,$82,$1F,$D9,$27,$AB,$67,$93,$67,$8B,$4E,$27,$F7
	.dc.b	$69,$E2,$D9,$E2,$D9,$E2,$D3,$89,$FD,$FE,$78,$B6,$78,$B6,$78,$B4
	.dc.b	$E2,$7F,$B4,$E3,$6D,$8D,$B4,$DC,$F6,$70,$7A,$C8,$79,$4A,$77,$29
	.dc.b	$C0,$D1,$0D,$14,$D1,$82,$1F,$EE,$9F,$5B,$EF,$12,$7E,$F3,$63,$F0
	.dc.b	$8D,$7E,$B3,$5F,$A0,$D7,$E5,$35,$ED,$1D,$08,$F4,$6B,$D1,$AF,$46
	.dc.b	$BD,$1A,$F4,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70,$29,$C0
	.dc.b	$A7,$03,$44,$34,$53,$46,$08,$60,$92,$9F,$90,$D9,$F7,$B6,$7D,$0D
	.dc.b	$7C,$6D,$7C,$0D,$7B,$B5,$EF,$3D,$89,$3D,$9B,$1D,$9A,$F6,$6B,$D9
	.dc.b	$AF,$66,$BD,$A3,$A1,$1E,$8D,$7A,$35,$E8,$D7,$A3,$5E,$87,$22,$1C
	.dc.b	$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68
	.dc.b	$C1,$0C,$12,$7E,$A1,$27,$E4,$B6,$7E,$03,$67,$D2,$D7,$C8,$D7,$C2
	.dc.b	$D7,$C0,$D7,$C1,$3D,$89,$3D,$9B,$1D,$9A,$F6,$6B,$D9,$AF,$66,$BD
	.dc.b	$A3,$A1,$1E,$8D,$7A,$35,$E8,$D7,$A3,$5E,$87,$22,$1C,$8A,$72,$29
	.dc.b	$C8,$A7,$23,$81,$FB,$CE,$67,$AD,$FE,$AA,$24,$FF,$AD,$9B,$3F,$EB
	.dc.b	$C6,$BF,$EA,$D6,$BF,$EC,$26,$BF,$EB,$A6,$BF,$EC,$B6,$CF,$82,$7B
	.dc.b	$12,$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$36,$3B,$47,$42,$1D,$0A,$74
	.dc.b	$29,$D0,$A7,$42,$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C
	.dc.b	$0A,$70,$2B,$C1,$FF,$54,$F5,$7F,$AD,$09,$3F,$EB,$46,$C7,$FA,$D1
	.dc.b	$AF,$FA,$D1,$FF,$76,$3F,$E7,$CE,$5F,$F4,$03,$FE,$89,$3F,$CF,$67
	.dc.b	$FA,$19,$CF,$37,$FA,$0E,$1C,$1F,$EA,$56,$9F,$E9,$F3,$91,$FE,$94
	.dc.b	$8E,$11,$B2,$46,$E3,$2D,$9F,$F4,$63,$C5,$AF,$FA,$64,$F4,$1E,$62
	.dc.b	$9B,$3F,$D4,$AF,$A0,$E3,$C5,$FE,$9E,$8E,$07,$A2,$7F,$D3,$8F,$A0
	.dc.b	$EC,$66,$79,$3D,$0E,$35,$EA,$9E,$27,$A8,$B3,$E0,$C3,$FD,$7F,$7E
	.dc.b	$1B,$F3,$1F,$34,$E6,$7C,$13,$FD,$49,$BF,$FA,$99,$72,$7F,$A6,$B9
	.dc.b	$7F,$D5,$BC,$DF,$EA,$C9,$FF,$50,$47,$FA,$89,$FF,$55,$73,$7F,$A9
	.dc.b	$DF,$F5,$8D,$F3,$BE,$51,$FE,$B0,$3F,$CF,$FC,$BF,$E7,$66,$E3,$FE
	.dc.b	$9D,$8F,$F4,$3C,$7F,$A2,$63,$FD,$3C,$E8,$FF,$46,$9F,$E7,$BE,$6F
	.dc.b	$F4,$3F,$32,$3F,$D3,$58,$7F,$A0,$27,$31,$C7,$7F,$F4,$44,$7F,$B1
	.dc.b	$75,$FF,$58,$4F,$48,$DB,$FE,$86,$3C,$47,$FA,$8C,$E5,$7F,$EA,$C7
	.dc.b	$47,$8F,$1F,$4B,$71,$F4,$BA,$AE,$A4,$9F,$85,$19,$53,$EB,$C3,$FD
	.dc.b	$44,$7F,$A8,$2F,$C1,$7C,$67,$E4,$1F,$1B,$99,$FA,$FF,$FB,$0F,$2F
	.dc.b	$F4,$C4,$62,$D1,$89,$D6,$5F,$24,$F9,$55,$B9,$D5,$70,$9D,$1F,$EA
	.dc.b	$8A,$E2,$F1,$AE,$0E,$9A,$FF,$A4,$FD,$1F,$F5,$11,$2F,$FD,$44,$D8
	.dc.b	$FF,$54,$35,$E1,$1A,$23,$A6,$BA,$30,$43,$04,$3F,$D3,$91,$D5,$B1
	.dc.b	$C9,$A7,$12,$9C,$4F,$F4,$E4,$71,$6C,$71,$69,$C4,$A7,$11,$CE,$9E
	.dc.b	$51,$86,$C6,$67,$A4,$E5,$B1,$98,$EC,$64,$AE,$5F,$03,$5F,$59,$0F
	.dc.b	$29,$4E,$47,$09,$F1,$F0,$FF,$B6,$49,$5F,$ED,$96,$CF,$FB,$7D,$AF
	.dc.b	$FB,$45,$AF,$FB,$2D,$AF,$FB,$15,$AF,$C4,$6D,$1A,$49,$F0,$12,$7C
	.dc.b	$0D,$7C,$0D,$7C,$0D,$7C,$0D,$7C,$0D,$9F,$03,$67,$C1,$3D,$88,$76
	.dc.b	$29,$D8,$A7,$62,$9D,$8B,$3D,$9B,$1D,$9E,$84,$3A,$14,$E8,$53,$A1
	.dc.b	$63,$A3,$63,$A3,$C8,$87,$22,$9C,$8B,$1C,$9B,$1C,$9E,$04,$38,$16
	.dc.b	$38,$36,$38,$3A,$24,$69,$B1,$A8,$C1,$1C,$12,$7E,$A1,$27,$E4,$B5
	.dc.b	$F8,$0D,$7D,$2D,$7C,$8D,$7C,$2D,$9F,$03,$67,$C1,$3D,$88,$76,$29
	.dc.b	$D8,$A7,$62,$9D,$8B,$3D,$9B,$1D,$9E,$84,$3A,$14,$E8,$53,$A1,$63
	.dc.b	$A3,$63,$A3,$C8,$87,$22,$9C,$8B,$1C,$9B,$1C,$9E,$04,$38,$16,$38
	.dc.b	$36,$38,$3A,$24,$69,$B1,$A8,$C1,$1C,$12,$7E,$A1,$27,$E4,$B5,$F8
	.dc.b	$0D,$7D,$2D,$7C,$8D,$7C,$2D,$9F,$03,$67,$C1,$3D,$8F,$EC,$23,$95
	.dc.b	$DD,$FD,$96,$3F,$B2,$C6,$2B,$71,$B9,$D3,$FD,$BE,$74,$F4,$6B,$CA
	.dc.b	$36,$FF,$87,$63,$2F,$18,$CB,$D6,$7B,$CE,$A3,$84,$EA,$31,$1B,$3C
	.dc.b	$11,$92,$4F,$26,$C6,$A3,$04,$70,$43,$D0,$E5,$A6,$4F,$43,$96,$99
	.dc.b	$3C,$EE,$5A,$64,$F4,$1D,$08,$72,$29,$C0,$A7,$03,$44,$34,$53,$46
	.dc.b	$08,$60,$93,$F2,$89,$3F,$1A,$3F,$C4,$BE,$3B,$D1,$FA,$AD,$BF,$C1
	.dc.b	$B6,$7F,$16,$D7,$8C,$F0,$24,$F0,$7F,$D5,$71,$FE,$87,$3F,$CF,$7C
	.dc.b	$FF,$E8,$27,$FD,$10,$7F,$A0,$0F,$F4,$B9,$FE,$8E,$E1,$FF,$45,$73
	.dc.b	$FF,$A3,$A3,$FD,$08,$6C,$E8,$58,$DB,$DA,$38,$3D,$38,$7F,$D1,$CF
	.dc.b	$53,$FD,$18,$74,$31,$5E,$57,$11,$C0,$98,$7F,$A6,$09,$1F,$EA,$58
	.dc.b	$CB,$63,$33,$FE,$A1,$AF,$57,$2F,$FA,$3B,$97,$FD,$21,$CC,$E7,$FF
	.dc.b	$4C,$E3,$FE,$9E,$E0,$FF,$4B,$C7,$FA,$66,$3F,$D1,$16,$AF,$F4,$D4
	.dc.b	$61,$AF,$FA,$91,$E8,$FF,$A6,$CC,$4F,$66,$99,$9E,$91,$9E,$6F,$F3
	.dc.b	$C4,$FF,$A3,$F9,$FF,$CF,$EF,$FB,$7B,$9F,$FC,$E5,$3A,$DF,$FD,$2F
	.dc.b	$9F,$FA,$6E,$F7,$1F,$E8,$4A,$DC,$61,$BB,$FF,$A4,$2F,$94,$7F,$A2
	.dc.b	$27,$FD,$2B,$18,$C3,$FD,$4F,$BF,$FA,$8B,$0E,$ED,$BE,$18,$73,$C3
	.dc.b	$89,$DE,$FF,$D4,$67,$FA,$C7,$0F,$F5,$63,$EB,$3F,$D5,$38,$7F,$AC
	.dc.b	$27,$FD,$3F,$3E,$19,$3C,$27,$FD,$3D,$1C,$F1,$F8,$17,$FE,$A3,$C3
	.dc.b	$DB,$3F,$EB,$33,$FD,$51,$19,$7C,$67,$49,$E2,$F8,$CD,$96,$FC,$47
	.dc.b	$86,$BA,$DF,$CD,$9F,$4C,$60,$F8,$6F,$86,$78,$D7,$6A,$E1,$3F,$EC
	.dc.b	$13,$CD,$0D,$FF,$9B,$B7,$2F,$E8,$AD,$9F,$E9,$AD,$7F,$D5,$ED,$7F
	.dc.b	$D6,$CD,$7F,$DC,$0D,$4D,$E4,$79,$1F,$D1,$9B,$5F,$D5,$1B,$5F,$D7
	.dc.b	$1A,$FF,$AF,$9A,$FF,$B1,$5A,$FF,$BB,$DA,$FF,$A7,$27,$B1,$27,$B3
	.dc.b	$63,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B4,$74,$23,$D1,$AF,$46,$BD,$1A
	.dc.b	$F4,$6B,$D0,$E4,$43,$91,$4E,$45,$39,$14,$E4,$70,$21,$C0,$A7,$02
	.dc.b	$9C,$0D,$10,$D1,$4D,$18,$21,$83,$FD,$9B,$E7,$FD,$F2,$47,$DF,$6C
	.dc.b	$7D,$F6,$BF,$7D,$AF,$DF,$6B,$F7,$DA,$F6,$8E,$84,$7A,$35,$E8,$D7
	.dc.b	$A3,$5E,$8D,$7A,$1C,$88,$72,$29,$C8,$A7,$22,$9C,$8E,$04,$38,$14
	.dc.b	$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$4A,$4F,$C8,$6C,$FB,$DB,$3E
	.dc.b	$86,$BE,$36,$BE,$06,$BD,$DA,$F7,$9E,$C4,$9E,$CD,$8E,$CD,$7B,$35
	.dc.b	$EC,$D7,$B3,$5E,$D1,$D0,$8F,$46,$BD,$1A,$F4,$6B,$D1,$AF,$43,$91
	.dc.b	$0E,$45,$39,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45
	.dc.b	$34,$60,$86,$09,$3F,$50,$93,$F2,$5B,$3F,$01,$B3,$E9,$6B,$E4,$6B
	.dc.b	$E1,$6B,$E0,$6B,$E0,$9E,$C4,$9E,$CD,$8E,$CD,$7B,$35,$EC,$D7,$B3
	.dc.b	$5E,$D1,$D0,$8F,$46,$BD,$1A,$F4,$6B,$D1,$AF,$43,$91,$0E,$45,$39
	.dc.b	$14,$E4,$53,$91,$C0,$FD,$E7,$33,$CF,$FF,$51,$12,$7F,$D4,$ED,$9F
	.dc.b	$F5,$93,$5F,$F5,$DB,$5F,$F5,$C3,$5F,$F5,$A3,$5F,$F6,$23,$67,$C1
	.dc.b	$3D,$89,$3D,$9A,$F6,$6B,$D9,$AF,$66,$BD,$9B,$1D,$A3,$A1,$0E,$85
	.dc.b	$3A,$14,$E8,$53,$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8,$AF,$27,$81
	.dc.b	$0E,$05,$38,$15,$E0,$87,$33,$CF,$FF,$56,$92,$BF,$D5,$AD,$9F,$F5
	.dc.b	$6B,$5F,$F5,$AB,$FE,$AA,$D7,$FD,$3B,$1F,$E9,$F3,$FD,$32,$7F,$A7
	.dc.b	$67,$FD,$1F,$CD,$FF,$00,$27,$73,$FE,$9C,$8F,$F5,$1B,$FE,$A8,$30
	.dc.b	$7F,$A9,$C8,$75,$DB,$FD,$44,$F5,$3F,$D3,$84,$7F,$D3,$8F,$03,$99
	.dc.b	$C6,$3B,$CF,$8B,$87,$FD,$4B,$5C,$2B,$47,$18,$CD,$7A,$63,$DA,$79
	.dc.b	$E3,$2F,$FB,$20,$EE,$47,$2F,$FA,$E2,$FE,$95,$77,$8D,$46,$2B,$FD
	.dc.b	$5B,$1F,$EA,$1E,$5F,$F5,$3F,$07,$FA,$D3,$93,$FD,$4D,$CB,$FE,$A4
	.dc.b	$AD,$47,$FA,$CD,$FF,$54,$9F,$7C,$FF,$59,$35,$EB,$3C,$DE,$24,$5C
	.dc.b	$DF,$E7,$78,$FF,$4E,$9F,$E7,$FE,$6F,$F4,$C3,$FE,$82,$D7,$FC,$EF
	.dc.b	$CB,$FE,$82,$9F,$F4,$A9,$FE,$9E,$DB,$FD,$19,$5B,$E6,$E8,$FF,$A6
	.dc.b	$27,$73,$FE,$A5,$24,$7F,$A1,$63,$93,$E1,$8C,$DF,$FA,$8D,$EB,$7F
	.dc.b	$E9,$19,$FF,$47,$9E,$89,$FF,$51,$BF,$E9,$BD,$BF,$D6,$8F,$FA,$95
	.dc.b	$AE,$4E,$C4,$8E,$CF,$FB,$3E,$BB,$DF,$92,$7E,$0B,$F1,$2F,$FD,$37
	.dc.b	$7D,$1F,$25,$FF,$A6,$EB,$99,$FE,$C6,$7B,$61,$F1,$25,$7F,$12,$7E
	.dc.b	$16,$1E,$03,$DF,$97,$FA,$9E,$FF,$79,$D4,$FD,$89,$EF,$34,$FC,$06
	.dc.b	$2A,$D4,$E6,$3F,$09,$18,$9D,$B6,$33,$5D,$27,$2D,$8C,$D7,$37,$F2
	.dc.b	$15,$C5,$FF,$51,$79,$1F,$EA,$92,$5F,$FA,$A5,$B1,$FE,$A9,$6B,$C2
	.dc.b	$34,$47,$4D,$74,$60,$86,$08,$7F,$AC,$63,$AB,$63,$93,$4E,$25,$38
	.dc.b	$9F,$EB,$78,$E2,$D8,$E2,$D3,$89,$4E,$23,$A5,$3C,$A3,$0D,$8C,$CF
	.dc.b	$49,$CB,$63,$31,$D8,$C9,$5C,$BE,$06,$BE,$B2,$1E,$52,$9C,$8E,$13
	.dc.b	$E3,$F4,$3F,$DC,$04,$AF,$F7,$03,$67,$FD,$C0,$D7,$FD,$C0,$D7,$FD
	.dc.b	$C0,$D7,$FD,$C0,$D7,$E2,$36,$8D,$24,$F8,$09,$3E,$06,$BE,$06,$BE
	.dc.b	$06,$BE,$06,$BE,$06,$CF,$81,$B3,$E0,$9E,$C4,$3B,$14,$EC,$53,$B1
	.dc.b	$4E,$C5,$9E,$CD,$8E,$CF,$42,$1D,$0A,$74,$29,$D0,$B1,$D1,$B1,$D1
	.dc.b	$E4,$43,$91,$4E,$45,$8E,$4D,$8E,$4F,$02,$1C,$0B,$1C,$1B,$1C,$1D
	.dc.b	$12,$34,$D8,$D4,$60,$8E,$09,$3F,$50,$93,$F2,$5A,$FC,$06,$BE,$96
	.dc.b	$BE,$46,$BE,$16,$CF,$81,$B3,$E0,$9E,$C4,$3B,$14,$EC,$53,$B1,$4E
	.dc.b	$C5,$9E,$CD,$8E,$CF,$42,$1D,$0A,$74,$29,$D0,$B1,$D1,$B1,$D1,$E4
	.dc.b	$43,$91,$4E,$45,$8E,$4D,$8E,$4F,$02,$1C,$0B,$1C,$1B,$1C,$1D,$12
	.dc.b	$34,$D8,$D4,$60,$8E,$09,$3F,$50,$93,$F2,$5A,$FC,$06,$BE,$96,$BE
	.dc.b	$46,$BE,$16,$CF,$81,$B3,$E0,$9E,$C7,$F6,$AE,$87,$A1,$FD,$7E,$3F
	.dc.b	$C5,$D1,$8A,$FE,$DF,$08,$C3,$C6,$78,$3F,$E7,$38,$CB,$FE,$35,$6B
	.dc.b	$D2,$36,$F3,$6C,$6A,$39,$46,$09,$3E,$29,$D4,$F1,$8C,$4E,$A3,$C9
	.dc.b	$18,$23,$82,$1E,$58,$E0,$D8,$DB,$4D,$9E,$97,$2D,$32,$7A,$1C,$B4
	.dc.b	$C9,$E7,$2C,$FA,$89,$3E,$28,$E8,$78,$48,$60,$87,$53,$25,$32,$77
	.dc.b	$32,$53,$31,$E0,$9F,$BB,$D2,$F4,$3E,$AB,$63,$31,$82,$38,$25,$7F
	.dc.b	$8D,$27,$83,$F8,$83,$FD,$19,$1F,$EA,$87,$FD,$19,$C1,$FE,$8A,$7F
	.dc.b	$D4,$1D,$3F,$E8,$98,$C4,$70,$3F,$D4,$5D,$3F,$E9,$79,$FF,$53,$C6
	.dc.b	$5E,$15,$D6,$3F,$D2,$87,$03,$BB,$97,$BC,$F3,$3F,$D5,$2F,$FA,$8B
	.dc.b	$97,$FD,$55,$3B,$8E,$D3,$B7,$FD,$55,$1D,$FA,$3F,$D3,$1C,$9F,$E9
	.dc.b	$12,$73,$FF,$A5,$0F,$F5,$0D,$7F,$A5,$E7,$FD,$3C,$DE,$97,$4F,$FA
	.dc.b	$7E,$79,$D7,$1A,$FF,$4F,$9F,$E9,$B7,$FC,$D9,$D1,$FE,$6B,$CF,$FC
	.dc.b	$E3,$D1,$FE,$A5,$E9,$FF,$3A,$9F,$E9,$9D,$FF,$CD,$BA,$FF,$9E,$75
	.dc.b	$FF,$37,$65,$FE,$8C,$3F,$D0,$D8,$7F,$A3,$2F,$8E,$3F,$E8,$13,$FD
	.dc.b	$09,$8F,$83,$1E,$77,$FE,$7E,$39,$57,$08,$FF,$3E,$47,$F9,$F8,$97
	.dc.b	$E6,$7F,$D2,$C7,$13,$FD,$54,$72,$C3,$CA,$E0,$95,$FE,$92,$78,$36
	.dc.b	$7D,$13,$C2,$31,$7E,$7A,$F0,$BF,$E9,$39,$D6,$1E,$C9,$F0,$63,$F1
	.dc.b	$48,$F9,$65,$E3,$7F,$18,$6C,$FF,$24,$6C,$FF,$A5,$1A,$FF,$B5,$5A
	.dc.b	$FF,$B6,$5A,$83,$86,$CF,$F2,$6E,$67,$95,$49,$7F,$CD,$23,$2D,$8C
	.dc.b	$D7,$FA,$9A,$72,$D8,$CB,$FE,$E9,$8C,$B6,$32,$FF,$BB,$E3,$2D,$8C
	.dc.b	$BE,$98,$CB,$63,$2F,$FA,$15,$B1,$A8,$C4,$F9,$C8,$F9,$5A,$F8,$DA
	.dc.b	$F8,$5A,$F7,$6B,$D0,$E4,$43,$91,$4E,$45,$39,$14,$E4,$70,$21,$C0
	.dc.b	$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$FE,$AB,$F1,$3F,$00,$48,$FB
	.dc.b	$CD,$8F,$B8,$D7,$ED,$35,$FB,$0D,$7E,$B3,$5E,$D1,$D0,$8F,$46,$BD
	.dc.b	$1A,$F4,$6B,$D1,$AF,$43,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0,$87
	.dc.b	$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$09,$29,$F9,$0D,$9F,$7B
	.dc.b	$67,$D0,$D7,$C6,$D7,$C0,$D7,$BB,$5E,$F3,$D8,$93,$D9,$B1,$D9,$AF
	.dc.b	$66,$BD,$9A,$F6,$6B,$DA,$3A,$11,$E8,$D7,$A3,$5E,$8D,$7A,$35,$E8
	.dc.b	$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E,$06,$88
	.dc.b	$68,$A6,$8C,$10,$C1,$27,$EA,$12,$7E,$4B,$67,$E0,$36,$7D,$2D,$7C
	.dc.b	$8D,$7C,$2D,$7C,$0D,$7C,$13,$D8,$93,$D9,$B1,$D9,$AF,$66,$BD,$9A
	.dc.b	$F6,$6B,$DA,$3A,$11,$E8,$D7,$A3,$5E,$8D,$7A,$35,$E8,$72,$21,$C8
	.dc.b	$A7,$22,$9C,$8A,$72,$38,$1F,$AA,$F2,$BF,$B1,$12,$3F,$B0,$36,$3F
	.dc.b	$AF,$35,$FE,$B8,$D7,$FA,$D3,$5F,$EB,$0D,$54,$8C,$36,$7B,$92,$7B
	.dc.b	$B5,$EE,$D7,$BB,$5E,$ED,$7B,$CF,$62,$46,$67,$C1,$3D,$88,$76,$29
	.dc.b	$D8,$A7,$62,$9D,$A3,$A4,$F6,$8C,$CF,$67,$A1,$0E,$85,$3A,$14,$E8
	.dc.b	$76,$E7,$79,$5F,$EB,$A2,$5F,$FA,$DD,$B3,$FE,$CF,$6B,$FE,$C5,$7F
	.dc.b	$D8,$6F,$F9,$D0,$FF,$3B,$9F,$E7,$93,$FC,$F7,$CD,$FE,$7E,$70,$6C
	.dc.b	$CF,$37,$F9,$D5,$FF,$16,$71,$7F,$A2,$B8,$BF,$CF,$71,$C0,$FF,$3D
	.dc.b	$9D,$AB,$A1,$2B,$94,$2B,$93,$FE,$85,$24,$FF,$A4,$AB,$AB,$C6,$3A
	.dc.b	$57,$2A,$E3,$3B,$9E,$33,$FE,$90,$8C,$11,$C1,$23,$FD,$1E,$21,$1F
	.dc.b	$E8,$58,$72,$7F,$A2,$1F,$F4,$3F,$07,$FA,$49,$FF,$44,$9F,$E8,$FE
	.dc.b	$5F,$F4,$77,$37,$FA,$42,$3F,$CA,$47,$3A,$E4,$F4,$8D,$57,$38,$FF
	.dc.b	$39,$F1,$FF,$9A,$5B,$97,$FA,$03,$37,$27,$FA,$4E,$FF,$CD,$F3,$98
	.dc.b	$C6,$DF,$E6,$EA,$FF,$3C,$B6,$3F,$CF,$11,$FE,$87,$BF,$F4,$AC,$60
	.dc.b	$FF,$3D,$4E,$E3,$FD,$23,$87,$FA,$8E,$73,$1F,$E8,$68,$F3,$72,$FF
	.dc.b	$A1,$23,$78,$7F,$A6,$B0,$F0,$C6,$E3,$CC,$FF,$A5,$A3,$24,$BF,$14
	.dc.b	$7F,$9E,$69,$7E,$18,$C3,$54,$6A,$FC,$15,$CA,$73,$5B,$AD,$42,$FE
	.dc.b	$5B,$E3,$AE,$13,$A9,$CC,$FC,$63,$C3,$7F,$E8,$D8,$CE,$3F,$53,$1F
	.dc.b	$F4,$C6,$1F,$E9,$B2,$38,$25,$FF,$A6,$E3,$2F,$FA,$77,$B3,$FD,$44
	.dc.b	$49,$FF,$51,$36,$35,$18,$23,$82,$1F,$EA,$58,$E0,$D8,$DB,$4D,$9F
	.dc.b	$EA,$98,$DB,$63,$6D,$36,$7F,$AB,$69,$3C,$23,$0D,$7F,$D6,$2D,$8D
	.dc.b	$CF,$49,$DB,$5D,$B6,$37,$1E,$03,$25,$72,$FA,$C8,$79,$CA,$76,$38
	.dc.b	$4F,$8C,$93,$DD,$B1,$A8,$FF,$61,$79,$3F,$EC,$72,$38,$24,$7F,$B0
	.dc.b	$E3,$2D,$8C,$BF,$EC,$38,$CB,$63,$2F,$FB,$0E,$32,$D8,$CB,$FE,$C3
	.dc.b	$8C,$B6,$32,$FF,$6E,$8C,$B6,$33,$C6,$9C,$B6,$33,$18,$23,$82,$4F
	.dc.b	$AC,$87,$A4,$A7,$9C,$A7,$94,$A7,$8C,$B3,$E1,$6C,$F8,$1B,$3E,$07
	.dc.b	$B1,$0E,$C5,$3B,$14,$EC,$59,$EC,$D9,$EC,$D9,$EC,$F4,$21,$D0,$A7
	.dc.b	$42,$CF,$46,$CF,$46,$CF,$47,$91,$0E,$45,$9E,$4D,$9E,$4D,$9E,$4F
	.dc.b	$02,$4F,$06,$CF,$06,$CF,$09,$D1,$27,$4D,$8D,$46,$08,$E0,$93,$F5
	.dc.b	$08,$7C,$92,$9F,$00,$A7,$A4,$A7,$90,$B3,$E1,$6C,$F8,$1B,$3E,$07
	.dc.b	$B1,$0E,$C5,$3B,$14,$EC,$59,$EC,$D9,$EC,$D9,$EC,$F4,$21,$D0,$A7
	.dc.b	$42,$CF,$46,$CF,$46,$CF,$47,$91,$0E,$45,$9E,$4D,$9E,$4D,$9E,$4F
	.dc.b	$02,$4F,$06,$CF,$06,$CF,$09,$D1,$27,$4D,$8D,$46,$08,$E0,$93,$FE
	.dc.b	$1A,$07,$F9,$F3,$C9,$46,$63,$34,$9C,$C6,$63,$13,$B8,$E4,$D8,$CC
	.dc.b	$6E,$78,$BE,$18,$CB,$63,$2F,$F8,$BA,$32,$D8,$CB,$E4,$6B,$E0,$3B
	.dc.b	$4F,$86,$33,$3C,$A7,$84,$E4,$8F,$02,$1E,$37,$2D,$32,$78,$DC,$B4
	.dc.b	$C9,$D8,$A7,$94,$E4,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$8F,$84,$D9,$F8,$A4,$9F,$5C,$71,$9D,$9C,$39,$1E,$4F,$D0,$6D
	.dc.b	$7E,$05,$B5,$F8,$96,$BF,$8F,$6B,$FE,$6A,$6B,$F9,$26,$A6,$BF,$E9
	.dc.b	$92,$1F,$E9,$43,$FD,$36,$7F,$A2,$88,$60,$87,$0D,$3F,$D2,$3C,$7F
	.dc.b	$E8,$FA,$FD,$0F,$27,$FA,$4E,$B6,$75,$38,$1F,$E9,$23,$FD,$1A,$ED
	.dc.b	$CB,$FE,$93,$73,$3D,$5F,$F5,$55,$78,$1F,$29,$C2,$3F,$D3,$0F,$FA
	.dc.b	$CA,$BC,$71,$C6,$BB,$4F,$8C,$95,$CD,$F3,$B5,$E8,$78,$8E,$E7,$03
	.dc.b	$D4,$77,$3C,$A6,$0F,$F5,$D5,$78,$9F,$97,$C7,$FE,$9E,$27,$1F,$FA
	.dc.b	$7B,$AB,$FD,$44,$FF,$A9,$67,$FD,$50,$7F,$09,$94,$66,$3F,$D5,$31
	.dc.b	$8E,$5F,$E1,$0F,$47,$FC,$A5,$5D,$1B,$3F,$E7,$63,$FC,$FD,$C5,$FE
	.dc.b	$98,$C7,$FD,$1D,$1F,$E9,$63,$27,$FA,$04,$C3,$FE,$80,$E4,$71,$FF
	.dc.b	$AB,$A7,$67,$FA,$33,$93,$FD,$1E,$F0,$BF,$F5,$8B,$FE,$9F,$37,$5F
	.dc.b	$E8,$B8,$CD,$F8,$A7,$91,$B7,$C1,$7D,$DF,$F5,$B5,$78,$B4,$FF,$5A
	.dc.b	$57,$09,$CF,$2F,$FA,$74,$98,$F9,$27,$58,$FF,$A7,$31,$E8,$FC,$33
	.dc.b	$FD,$3F,$5C,$27,$CF,$3A,$7E,$55,$F9,$30,$F0,$E3,$EB,$9C,$E1,$CD
	.dc.b	$F8,$47,$CB,$3C,$26,$88,$7F,$AA,$CC,$9F,$ED,$E3,$E3,$E1,$E3,$BF
	.dc.b	$04,$7C,$59,$EC,$F0,$AF,$F6,$83,$DB,$D0,$79,$AE,$BF,$E8,$31,$86
	.dc.b	$C6,$72,$FE,$B1,$39,$6C,$66,$BF,$B0,$4E,$5B,$19,$7F,$DF,$B1,$97
	.dc.b	$FB,$36,$6E,$CF,$E9,$ED,$BF,$EC,$0D,$9F,$ED,$0D,$78,$CF,$02,$4F
	.dc.b	$06,$C7,$06,$BC,$23,$44,$74,$D7,$46,$08,$60,$87,$F7,$E3,$F4,$31
	.dc.b	$D9,$C4,$F9,$DB,$19,$7F,$D8,$8D,$78,$CF,$85,$DB,$5D,$B5,$D9,$E4
	.dc.b	$36,$53,$65,$36,$7A,$0C,$94,$C9,$C8,$C4,$FA,$5D,$99,$21,$82,$1C
	.dc.b	$8A,$F1,$5C,$DF,$83,$7F,$C2,$58,$FF,$82,$89,$5F,$E1,$D6,$C7,$F8
	.dc.b	$8D,$AF,$F8,$75,$AF,$28,$E0,$47,$83,$5E,$0D,$78,$1A,$21,$A2,$9A
	.dc.b	$30,$43,$04,$3F,$C7,$71,$E4,$6C,$77,$69,$D0,$A7,$32,$9C,$CF,$8F
	.dc.b	$1C,$DB,$1C,$DA,$73,$29,$CC,$A7,$39,$FA,$0F,$62,$1A,$2B,$A7,$04
	.dc.b	$70,$49,$F2,$36,$7C,$8D,$9F,$23,$5F,$23,$5F,$24,$F2,$24,$F2,$6C
	.dc.b	$72,$6B,$C9,$AF,$28,$E0,$47,$83,$5E,$0D,$78,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$3E,$4C,$79,$1B,$1D,$DA,$74,$29,$CC,$A7,$33,$E8,$47,$36
	.dc.b	$C7,$36,$9C,$CA,$73,$29,$CE,$7E,$83,$D8,$86,$8A,$E9,$C1,$1C,$12
	.dc.b	$7C,$8D,$9F,$23,$67,$C8,$D7,$C8,$D7,$C9,$3C,$89,$3C,$9B,$1C,$9A
	.dc.b	$F2,$6B,$CA,$38,$11,$E0,$D7,$83,$5E,$06,$88,$68,$A6,$8C,$10,$C1
	.dc.b	$0F,$93,$1E,$46,$C7,$76,$9F,$EB,$DE,$CF,$F5,$49,$23,$FD,$78,$D8
	.dc.b	$D4,$60,$8E,$08,$7F,$AF,$E3,$83,$63,$6D,$36,$7F,$B0,$63,$6D,$8D
	.dc.b	$B4,$D9,$FE,$D6,$8D,$B6,$36,$D3,$67,$F3,$E8,$DB,$63,$6D,$37,$49
	.dc.b	$DB,$67,$6D,$9D,$CE,$89,$3A,$6C,$6A,$30,$47,$BF,$67,$FB,$60,$91
	.dc.b	$FE,$EB,$6C,$7F,$BC,$5A,$FF,$BA,$DF,$F7,$1B,$FE,$82,$3F,$CF,$DC
	.dc.b	$BF,$E8,$AD,$FF,$D0,$D1,$FE,$83,$8F,$F4,$C3,$AE,$6F,$F4,$81,$1F
	.dc.b	$F4,$4C,$F0,$23,$FE,$87,$E2,$61,$FE,$87,$8F,$F4,$99,$97,$0F,$8C
	.dc.b	$FF,$46,$9F,$E9,$0B,$F0,$B8,$7C,$CF,$13,$57,$FE,$A6,$74,$F9,$5E
	.dc.b	$F1,$CA,$BC,$71,$C0,$FF,$51,$1A,$9E,$51,$A9,$EE,$71,$AF,$3D,$7F
	.dc.b	$B3,$79,$7F,$D2,$8F,$FA,$23,$97,$FD,$28,$D0,$23,$FD,$38,$FF,$A8
	.dc.b	$79,$BF,$D4,$0E,$A3,$2F,$FA,$98,$FF,$53,$EE,$AE,$AD,$8E,$15,$FE
	.dc.b	$AF,$7F,$CD,$FC,$7F,$E7,$77,$FC,$E0,$D3,$FC,$D2,$5E,$56,$7E,$6D
	.dc.b	$7F,$D2,$38,$7F,$A0,$EF,$FC,$E0,$DB,$E4,$DB,$FF,$3C,$72,$FF,$A7
	.dc.b	$8F,$F4,$6E,$1F,$E9,$3A,$EC,$F5,$8F,$F3,$DC,$62,$FF,$D0,$33,$BA
	.dc.b	$F1,$C7,$FA,$04,$FF,$46,$4E,$F0,$F2,$E9,$FE,$94,$AD,$C7,$FA,$92
	.dc.b	$F9,$1D,$EF,$E1,$CF,$99,$E6,$F0,$C3,$FD,$4D,$7C,$EB,$85,$7F,$A5
	.dc.b	$6F,$BB,$E3,$3E,$1C,$ED,$C6,$1F,$EB,$76,$E1,$DD,$FA,$2F,$77,$C7
	.dc.b	$7E,$67,$FD,$A7,$C8,$FC,$C9,$D9,$FE,$9A,$6E,$3F,$E9,$86,$D7,$FA
	.dc.b	$59,$AF,$FA,$51,$AF,$E4,$2B,$A5,$73,$25,$7F,$A5,$33,$FF,$4B,$91
	.dc.b	$C1,$2B,$FD,$2F,$19,$6C,$65,$FF,$50,$C6,$5B,$19,$7F,$D1,$31,$96
	.dc.b	$C6,$5F,$D2,$E8,$9D,$46,$1A,$FF,$BD,$1B,$1B,$9E,$53,$96,$C6,$63
	.dc.b	$CE,$43,$CA,$53,$C6,$53,$C1,$1C,$A7,$B9,$1C,$12,$38,$1E,$12,$CF
	.dc.b	$85,$EE,$43,$B9,$4E,$E7,$09,$FF,$64,$D4,$AF,$F6,$4B,$67,$FD,$92
	.dc.b	$D7,$FD,$A2,$D7,$FD,$6E,$D7,$E6,$B5,$57,$E1,$9C,$CF,$62,$3D,$9A
	.dc.b	$F6,$6B,$D9,$AF,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68,$C1
	.dc.b	$0C,$10,$F6,$CF,$BE,$7D,$D1,$89,$F3,$B4,$F1,$14,$EE,$53,$B1,$4E
	.dc.b	$D1,$D0,$93,$DC,$93,$DA,$32,$D8,$CC,$F8,$27,$2D,$8C,$BE,$28,$CB
	.dc.b	$63,$2F,$92,$32,$D8,$CB,$E6,$6B,$E9,$24,$70,$75,$3E,$52,$3E,$36
	.dc.b	$BE,$16,$BD,$DA,$F2,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$C1,$0F,$6C,$FC,$09,$F7,$46,$27,$CE,$D3,$C4,$53,$B9,$4E,$C5
	.dc.b	$3B,$47,$42,$4F,$72,$4F,$68,$CB,$63,$33,$E0,$9C,$B6,$32,$F8,$A3
	.dc.b	$2D,$8C,$BE,$48,$CB,$63,$2F,$99,$AF,$A4,$91,$C1,$D4,$F9,$48,$F8
	.dc.b	$DA,$F8,$5A,$F7,$6B,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$3F,$63,$C8
	.dc.b	$EC,$FF,$1B,$C7,$F8,$A8,$FF,$15,$C6,$67,$11,$98,$C3,$80,$E1,$DB
	.dc.b	$5E,$13,$D9,$B3,$C5,$E3,$19,$7C,$2D,$73,$1B,$9E,$91,$A8,$C4,$EE
	.dc.b	$73,$3C,$0F,$33,$82,$46,$C9,$3C,$89,$3C,$23,$2D,$8C,$BE,$E8,$CB
	.dc.b	$63,$2F,$B8,$E0,$43,$45,$34,$60,$86,$08,$7B,$CE,$05,$36,$53,$67
	.dc.b	$C1,$32,$53,$33,$F0,$A3,$CF,$1C,$48,$6C,$FC,$1F,$67,$CD,$24,$7D
	.dc.b	$36,$C7,$CE,$6B,$F8,$27,$FD,$46,$FF,$A7,$F9,$3F,$D0,$33,$FE,$7E
	.dc.b	$E0,$FF,$4D,$F2,$7F,$A0,$6B,$FD,$10,$7F,$A8,$CF,$F5,$09,$FE,$93
	.dc.b	$7F,$D0,$F5,$CA,$36,$7F,$A5,$0D,$C7,$33,$6F,$FA,$48,$FF,$51,$6F
	.dc.b	$FE,$91,$9F,$F4,$D4,$62,$FC,$8F,$78,$CB,$E9,$39,$14,$F3,$1F,$E9
	.dc.b	$D3,$37,$EC,$8F,$35,$F8,$A7,$67,$70,$A3,$FD,$23,$C7,$27,$FD,$23
	.dc.b	$D5,$FA,$7E,$5F,$F4,$CC,$7E,$DD,$AE,$8F,$F5,$04,$7F,$90,$27,$94
	.dc.b	$7F,$A9,$79,$7F,$D3,$4F,$F9,$C7,$87,$FD,$0F,$C9,$FE,$69,$8C,$1F
	.dc.b	$E9,$EE,$1F,$F4,$96,$DF,$E6,$BD,$7F,$D0,$52,$7F,$C9,$8F,$FA,$02
	.dc.b	$3F,$CF,$BC,$BF,$EA,$2B,$E8,$D8,$CC,$60,$FF,$3D,$47,$F9,$F6,$BA
	.dc.b	$9F,$E8,$42,$E1,$E6,$AE,$24,$BF,$14,$7F,$A9,$5F,$F4,$09,$FE,$A0
	.dc.b	$32,$79,$4F,$F4,$5E,$1E,$C7,$27,$FA,$2A,$7A,$57,$86,$BF,$D5,$F3
	.dc.b	$AB,$F3,$46,$27,$E0,$CF,$13,$AE,$FF,$E9,$98,$F4,$C7,$85,$AE,$DB
	.dc.b	$5D,$23,$D3,$1E,$16,$D7,$C1,$7A,$9F,$EA,$2B,$FA,$91,$FE,$AE,$8C
	.dc.b	$1E,$7C,$3F,$DE,$75,$E6,$B6,$3F,$EA,$7C,$5F,$8D,$BA,$7F,$99,$46
	.dc.b	$1B,$19,$8F,$E8,$06,$4A,$E7,$1F,$E8,$F1,$96,$C6,$5F,$E9,$D1,$97
	.dc.b	$FA,$6F,$D7,$BC,$1F,$E8,$0D,$9F,$E8,$CD,$9F,$EA,$0D,$78,$CF,$02
	.dc.b	$4F,$06,$C7,$06,$BC,$23,$44,$74,$D7,$46,$08,$60,$87,$F6,$99,$F2
	.dc.b	$36,$7B,$B6,$7A,$34,$E6,$53,$98,$13,$D1,$C3,$5C,$CF,$67,$2D,$73
	.dc.b	$3E,$07,$2D,$72,$78,$8C,$94,$C9,$E4,$29,$E7,$21,$C0,$D4,$F9,$49
	.dc.b	$3E,$37,$1F,$1F,$F7,$89,$5F,$79,$B5,$F7,$9A,$FE,$19,$AF,$CF,$6B
	.dc.b	$F1,$1A,$FC,$86,$C9,$A4,$9F,$01,$27,$C0,$D7,$C0,$D7,$C0,$D7,$C0
	.dc.b	$D7,$C0,$D9,$F0,$36,$7C,$13,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$B3
	.dc.b	$D9,$B1,$D9,$E8,$43,$A1,$4E,$85,$3A,$16,$3A,$36,$3A,$3C,$88,$72
	.dc.b	$29,$C8,$B1,$C9,$B1,$C9,$E0,$43,$81,$63,$83,$63,$83,$A2,$46,$9B
	.dc.b	$1A,$8C,$11,$C1,$27,$EA,$12,$7E,$4B,$5F,$80,$D7,$D2,$D7,$C8,$D7
	.dc.b	$C2,$D9,$F0,$36,$7C,$13,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$B3,$D9
	.dc.b	$B1,$D9,$E8,$43,$A1,$4E,$85,$3A,$16,$3A,$36,$3A,$3C,$88,$72,$29
	.dc.b	$C8,$B1,$C9,$B1,$C9,$E0,$43,$81,$63,$83,$63,$83,$A2,$46,$9B,$1A
	.dc.b	$8C,$11,$C1,$27,$EA,$12,$7E,$4B,$5F,$80,$D7,$D2,$D7,$C8,$D7,$C2
	.dc.b	$D9,$F0,$36,$7C,$13,$D8,$FE,$B7,$DB,$FD,$80,$93,$FD,$81,$B1,$A8
	.dc.b	$C1,$1C,$10,$FE,$D7,$1C,$1B,$1B,$69,$B3,$FA,$C4,$6D,$B1,$B6,$9B
	.dc.b	$3F,$A8,$C6,$DB,$1B,$69,$B3,$FA,$CC,$6D,$B1,$B6,$9B,$A4,$ED,$B3
	.dc.b	$B6,$CE,$E7,$44,$9D,$36,$35,$18,$23,$EB,$EA,$75,$FF,$9C,$A7,$FD
	.dc.b	$B9,$18,$E7,$FF,$6A,$36,$33,$3B,$AF,$F7,$2C,$E5,$B1,$97,$FD,$D9
	.dc.b	$19,$6C,$65,$FF,$46,$EF,$FE,$87,$7F,$D1,$27,$FA,$44,$FF,$4D,$F3
	.dc.b	$FF,$A2,$8D,$B9,$3F,$D2,$07,$FA,$30,$FF,$4A,$1F,$EA,$33,$25,$3F
	.dc.b	$D2,$18,$7F,$A4,$DC,$1C,$38,$FF,$D4,$8F,$80,$F1,$C7,$FA,$AE,$7C
	.dc.b	$2D,$8D,$1E,$1A,$F5,$34,$F3,$1E,$09,$E2,$7A,$DE,$27,$8A,$38,$B4
	.dc.b	$FF,$52,$C7,$F8,$8D,$FF,$57,$F3,$B9,$7F,$D3,$6D,$E7,$FF,$50,$4F
	.dc.b	$FA,$8F,$95,$CF,$FE,$A7,$BD,$8A,$FF,$56,$36,$BF,$D5,$24,$3C,$07
	.dc.b	$FA,$A6,$3A,$B4,$57,$67,$C1,$5D,$E7,$7C,$DF,$EA,$19,$FF,$3C,$E5
	.dc.b	$FE,$80,$E6,$FF,$50,$E1,$FE,$93,$8F,$F3,$B3,$79,$FF,$D3,$94,$BF
	.dc.b	$F3,$CE,$DF,$E9,$69,$FF,$42,$5F,$FA,$2A,$F8,$5F,$FA,$3E,$7F,$D0
	.dc.b	$93,$DE,$BF,$D0,$F3,$FE,$8C,$34,$53,$FD,$29,$3C,$0F,$F4,$F1,$E2
	.dc.b	$BF,$0C,$7A,$A3,$2F,$23,$FD,$92,$79,$F0,$F6,$F0,$FF,$A8,$E7,$A4
	.dc.b	$F7,$AE,$78,$7C,$0A,$D4,$E7,$1F,$5B,$E4,$31,$1F,$05,$A7,$FA,$FE
	.dc.b	$7F,$DA,$7C,$3F,$BC,$E2,$FD,$CF,$17,$F0,$8A,$FD,$8C,$60,$97,$FB
	.dc.b	$C6,$C7,$EE,$94,$7E,$0A,$3F,$11,$F8,$F3,$CB,$CC,$7E,$1B,$74,$E6
	.dc.b	$B8,$46,$2B,$75,$F9,$B8,$C4,$EA,$B8,$4F,$36,$BB,$6B,$FE,$85,$9E
	.dc.b	$2E,$EB,$AB,$99,$DB,$E8,$8C,$B6,$BC,$11,$8A,$E7,$3D,$E3,$15,$CA
	.dc.b	$73,$3D,$67,$24,$AE,$64,$8F,$23,$5E,$8D,$79,$35,$F2,$1C,$08,$70
	.dc.b	$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$93,$EE,$6D,$7C,$12,$57
	.dc.b	$BE,$3C,$33,$B2,$4E,$AB,$FD,$B3,$F4,$7F,$ED,$C2,$38,$25,$7F,$B7
	.dc.b	$23,$2D,$8C,$BF,$EE,$68,$CB,$63,$2F,$FB,$46,$32,$D8,$CB,$FE,$C8
	.dc.b	$8C,$B6,$32,$FF,$B3,$A3,$2D,$8C,$F3,$27,$2D,$8C,$C6,$08,$E0,$93
	.dc.b	$EB,$21,$E9,$29,$E7,$29,$E5,$29,$E3,$2C,$F8,$5B,$3E,$06,$CF,$81
	.dc.b	$EC,$43,$B1,$4E,$C5,$3B,$16,$7B,$36,$7B,$36,$7B,$3D,$08,$74,$29
	.dc.b	$D0,$B3,$D1,$B3,$D1,$B3,$D1,$E4,$43,$91,$67,$93,$67,$93,$67,$93
	.dc.b	$C0,$93,$C1,$B3,$C1,$B3,$C2,$74,$49,$D3,$63,$51,$82,$38,$24,$FD
	.dc.b	$42,$1F,$24,$A7,$C0,$29,$E9,$29,$E4,$2C,$F8,$5B,$3E,$06,$CF,$81
	.dc.b	$EC,$43,$B1,$4E,$C5,$3B,$16,$7B,$36,$7B,$36,$7B,$3D,$08,$74,$29
	.dc.b	$D0,$B3,$D1,$B3,$D1,$B3,$D1,$E4,$43,$91,$67,$93,$67,$93,$67,$93
	.dc.b	$C0,$93,$C1,$B3,$C1,$B3,$C2,$74,$49,$D3,$63,$51,$82,$38,$24,$FF
	.dc.b	$69,$F9,$0F,$F3,$02,$0E,$1B,$CF,$FD,$59,$B3,$FD,$C5,$A7,$38,$D4
	.dc.b	$F1,$23,$82,$4F,$18,$CB,$63,$33,$CE,$30,$47,$04,$3F,$C8,$73,$C9
	.dc.b	$C1,$FE,$55,$70,$EA,$7A,$B6,$7A,$34,$E2,$7C,$39,$E2,$F3,$9E,$2D
	.dc.b	$9E,$2D,$38,$9F,$10,$E8,$43,$45,$3B,$47,$46,$C7,$46,$9D,$27,$C6
	.dc.b	$E0,$E4,$77,$3A,$9F,$6F,$5F,$F5,$5F,$0F,$F8,$C0,$95,$FE,$A8,$6C
	.dc.b	$FF,$A9,$9A,$FF,$A9,$1F,$C0,$3F,$E8,$73,$FD,$0B,$C7,$FE,$94,$7F
	.dc.b	$D3,$6D,$74,$62,$7F,$19,$87,$FA,$76,$75,$3F,$E9,$D9,$CF,$3F,$FA
	.dc.b	$4A,$3F,$D3,$AF,$FA,$4D,$E4,$7F,$A4,$0F,$0C,$70,$69,$93,$75,$DE
	.dc.b	$35,$18,$9E,$CF,$FA,$8E,$7F,$D5,$0E,$4E,$6F,$17,$A1,$EC,$21,$E9
	.dc.b	$32,$57,$81,$23,$BC,$7F,$AA,$63,$BF,$3F,$FA,$96,$3F,$D2,$FC,$DF
	.dc.b	$E9,$56,$9F,$E9,$82,$1F,$E9,$D3,$FD,$43,$1C,$5B,$CF,$FE,$A7,$6F
	.dc.b	$07,$FA,$A8,$FF,$52,$CF,$7E,$5F,$F5,$24,$FF,$A9,$DB,$19,$BF,$F4
	.dc.b	$1C,$FF,$A1,$F3,$FF,$3E,$34,$72,$FF,$A1,$5F,$F3,$F6,$BC,$39,$BF
	.dc.b	$D1,$2D,$E7,$5F,$29,$CD,$7F,$A5,$5F,$F5,$95,$FF,$A9,$F7,$FF,$47
	.dc.b	$CF,$FA,$62,$76,$FF,$A3,$CC,$5F,$FA,$69,$F0,$9F,$E9,$0A,$FF,$56
	.dc.b	$3E,$6C,$3F,$D4,$D3,$E2,$8F,$F5,$31,$FE,$9F,$79,$4F,$03,$D8,$59
	.dc.b	$E3,$87,$A6,$77,$1F,$EA,$EA,$F2,$5F,$AA,$3E,$05,$6C,$FF,$5E,$CF
	.dc.b	$A1,$B3,$B8,$F4,$4F,$FB,$AE,$3E,$2F,$03,$9D,$F4,$FF,$EA,$82,$67
	.dc.b	$FE,$A1,$9D,$46,$30,$FF,$72,$C6,$27,$55,$C9,$B5,$FE,$90,$F0,$1F
	.dc.b	$B6,$BE,$5F,$0D,$AF,$F5,$74,$EA,$72,$48,$CC,$EF,$1F,$F7,$AD,$6E
	.dc.b	$73,$5B,$9C,$D6,$DF,$EA,$02,$B8,$3A,$9E,$0E,$A3,$2D,$9E,$4F,$F5
	.dc.b	$98,$CB,$63,$2F,$56,$C6,$A3,$13,$D9,$E0,$D7,$35,$EA,$23,$E5,$6B
	.dc.b	$E3,$8E,$13,$D4,$95,$E0,$24,$69,$B5,$CA,$BC,$2D,$7C,$27,$72,$1D
	.dc.b	$DD,$57,$6A,$E8,$48,$CD,$76,$AE,$27,$D0,$E8,$72,$7F,$7D,$6D,$7F
	.dc.b	$7D,$6C,$FF,$BB,$DA,$FD,$80,$E1,$E9,$3C,$C9,$3C,$DB,$1C,$DA,$F3
	.dc.b	$9E,$4F,$F8,$4D,$AF,$68,$DB,$DA,$7A,$11,$E8,$D7,$A4,$70,$7A,$35
	.dc.b	$E9,$1B,$7A,$1C,$88,$72,$74,$F2,$6B,$C8,$D9,$C8,$E0,$60,$E0,$53
	.dc.b	$81,$A2,$1C,$48,$70,$32,$53,$27,$22,$CF,$C8,$24,$FB,$DB,$3E,$86
	.dc.b	$CF,$8D,$AF,$63,$A1,$DC,$86,$08,$F0,$7C,$33,$DC,$93,$DD,$B1,$DD
	.dc.b	$AF,$79,$E4,$F6,$6B,$DA,$36,$F6,$9E,$84,$7A,$35,$E9,$1C,$1E,$8D
	.dc.b	$7A,$46,$DE,$87,$22,$1C,$9D,$3C,$9A,$F2,$36,$72,$38,$18,$38,$14
	.dc.b	$E0,$68,$87,$12,$1C,$0C,$94,$C9,$C8,$B3,$F2,$09,$3E,$F6,$CF,$A1
	.dc.b	$B3,$E3,$6B,$D8,$E8,$77,$21,$82,$3C,$1F,$0C,$F7,$24,$F7,$6C,$77
	.dc.b	$6B,$DE,$79,$3D,$9A,$F6,$8D,$BD,$A7,$A1,$1E,$8D,$7A,$47,$07,$A3
	.dc.b	$5E,$91,$B7,$A1,$C8,$87,$27,$4F,$26,$BF,$0F,$9D,$D7,$FE,$A2,$24
	.dc.b	$FF,$A7,$DB,$3F,$EB,$16,$BF,$E9,$A6,$BF,$E9,$96,$BF,$E9,$36,$BF
	.dc.b	$EB,$66,$CF,$82,$7B,$12,$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$36,$3B
	.dc.b	$47,$42,$1D,$0A,$74,$29,$D0,$A7,$42,$BD,$1E,$44,$39,$14,$E4,$53
	.dc.b	$91,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1,$6B,$FE,$31,$EC,$57,$F8,$A9
	.dc.b	$B3,$FE,$2E,$D9,$18,$8F,$F0,$DD,$71,$AE,$04,$8E,$13,$B7,$FD,$6D
	.dc.b	$5C,$A7,$FD,$D9,$1F,$E8,$EE,$3F,$F4,$CB,$FE,$89,$E9,$FF,$52,$39
	.dc.b	$3F,$D2,$BC,$7F,$EA,$C9,$DB,$63,$35,$C9,$FF,$4B,$57,$FA,$52,$75
	.dc.b	$D3,$FE,$A2,$9E,$8D,$76,$7F,$A9,$27,$85,$73,$BE,$ED,$9E,$E7,$FA
	.dc.b	$92,$32,$D3,$35,$FE,$A5,$7A,$10,$FF,$59,$10,$F1,$D7,$C0,$7C,$66
	.dc.b	$4A,$68,$FF,$64,$9F,$E4,$CB,$FF,$6B,$73,$FF,$A7,$63,$FD,$56,$D7
	.dc.b	$FD,$45,$CF,$FE,$B1,$7F,$D5,$27,$F8,$74,$FF,$58,$19,$3F,$D6,$DD
	.dc.b	$1F,$EA,$8E,$07,$4F,$FA,$B9,$B3,$FE,$BA,$3A,$9F,$E7,$E3,$FD,$21
	.dc.b	$87,$FA,$0B,$A7,$FD,$29,$A3,$A7,$FD,$05,$1F,$E8,$EA,$D9,$FE,$9B
	.dc.b	$3F,$D5,$0F,$08,$DD,$73,$3F,$D5,$D8,$FE,$5A,$FF,$D3,$57,$C2,$73
	.dc.b	$B7,$FA,$BB,$A7,$FD,$25,$3F,$E9,$53,$B1,$FE,$9C,$C7,$B2,$3F,$D8
	.dc.b	$57,$FE,$A6,$C7,$FD,$45,$4C,$7C,$11,$86,$D7,$AA,$B8,$36,$7F,$D5
	.dc.b	$AF,$FA,$D0,$EA,$FF,$AA,$9C,$57,$FB,$08,$F3,$92,$FC,$B1,$C0,$F7
	.dc.b	$1F,$EC,$43,$D5,$8F,$C0,$71,$3C,$63,$17,$F1,$A7,$66,$B0,$F4,$BD
	.dc.b	$78,$BF,$D6,$9D,$5F,$B9,$8F,$E0,$FC,$48,$C4,$EB,$4F,$F5,$8C,$62
	.dc.b	$75,$18,$25,$7F,$0B,$9E,$9F,$12,$F3,$1F,$19,$86,$DF,$E5,$67,$53
	.dc.b	$92,$46,$67,$77,$FE,$77,$AD,$CE,$6B,$73,$9A,$DB,$FE,$B4,$AD,$B9
	.dc.b	$9D,$B9,$9D,$BF,$85,$9D,$B9,$9D,$B9,$9D,$A9,$DB,$99,$DB,$99,$DC
	.dc.b	$EA,$31,$3A,$8C,$12,$76,$48,$D3,$6B,$E0,$12,$BD,$8D,$7D,$0D,$7C
	.dc.b	$8D,$7C,$2D,$AF,$05,$76,$24,$66,$BC,$15,$D8,$87,$62,$9F,$EB,$39
	.dc.b	$FF,$59,$12,$7F,$D6,$4D,$8F,$F5,$93,$5F,$F5,$43,$5E,$51,$C0,$8F
	.dc.b	$06,$BC,$1A,$F0,$34,$43,$45,$34,$60,$86,$08,$7F,$AE,$63,$C8,$D8
	.dc.b	$EE,$D3,$A1,$4E,$65,$39,$9F,$EC,$E8,$E6,$D8,$E6,$D3,$99,$4E,$65
	.dc.b	$39,$F4,$FE,$49,$C1,$DC,$86,$08,$EC,$93,$E2,$6C,$F8,$9B,$3E,$26
	.dc.b	$BE,$26,$BE,$29,$E4,$49,$E4,$D8,$E4,$D7,$93,$5E,$51,$C0,$8F,$06
	.dc.b	$BC,$1A,$F0,$34,$43,$45,$34,$60,$86,$08,$7C,$89,$F6,$B6,$7C,$CD
	.dc.b	$9F,$0B,$4E,$C5,$3A,$94,$EA,$7D,$28,$E6,$D8,$E6,$D3,$99,$4E,$65
	.dc.b	$39,$CF,$99,$C1,$DC,$86,$08,$EC,$93,$E2,$6C,$F8,$9B,$3E,$26,$BE
	.dc.b	$26,$BE,$29,$E4,$49,$E4,$D8,$E4,$D7,$93,$5E,$51,$C0,$8F,$06,$BC
	.dc.b	$1A,$F0,$34,$43,$45,$34,$60,$86,$08,$7C,$89,$F6,$B6,$7C,$CD,$9F
	.dc.b	$0B,$F3,$BA,$3F,$C4,$11,$FE,$2F,$84,$61,$FF,$19,$35,$FF,$09,$1F
	.dc.b	$E1,$37,$0E,$1A,$E7,$87,$FB,$B3,$D1,$B1,$A8,$C4,$F8,$63,$04,$9D
	.dc.b	$92,$34,$D9,$F2,$B5,$F2,$9D,$0C,$C7,$02,$38,$21,$C2,$7C,$11,$A8
	.dc.b	$C4,$F5,$7D,$CD,$7D,$D1,$B7,$2D,$72,$76,$32,$53,$27,$67,$83,$E9
	.dc.b	$30,$6C,$C1,$C8,$C9,$0C,$10,$EA,$53,$89,$82,$68,$D1,$F4,$1F,$E2
	.dc.b	$72,$57,$F9,$CE,$75,$18,$AF,$F5,$93,$5F,$F5,$8D,$72,$AF,$F5,$5C
	.dc.b	$7F,$A1,$89,$BF,$FA,$53,$A3,$FD,$25,$1F,$E8,$B8,$C7,$47,$FA,$7A
	.dc.b	$3F,$D3,$13,$FE,$A3,$9C,$C7,$FA,$5C,$FF,$48,$1F,$E6,$5E,$2F,$F5
	.dc.b	$5D,$7F,$A5,$E7,$73,$D4,$FF,$50,$C7,$5A,$F0,$C7,$23,$FD,$7C,$7F
	.dc.b	$A6,$8E,$97,$FE,$AD,$8F,$14,$72,$3C,$47,$B0,$C1,$DD,$D9,$27,$F2
	.dc.b	$B7,$E6,$9F,$F5,$35,$75,$9C,$BF,$E7,$FA,$47,$FA,$83,$9E,$57,$FA
	.dc.b	$AB,$9B,$FD,$3F,$C9,$FE,$A3,$E7,$FD,$D7,$42,$7F,$D4,$75,$FE,$AC
	.dc.b	$21,$C8,$FF,$52,$92,$3F,$D6,$9C,$FF,$E7,$58,$FF,$39,$67,$FE,$97
	.dc.b	$E8,$FF,$3F,$CF,$FA,$0F,$1F,$F3,$AB,$FE,$85,$8C,$9F,$EA,$9A,$FF
	.dc.b	$42,$1F,$E9,$DD,$BF,$D2,$2E,$27,$B4,$7F,$AC,$63,$FD,$3D,$1D,$6F
	.dc.b	$C1,$B7,$FA,$1B,$8B,$FD,$39,$18,$AF,$29,$FE,$8C,$C3,$C3,$5F,$EB
	.dc.b	$38,$CE,$1F,$E8,$DC,$3F,$D4,$D7,$C6,$3C,$13,$E3,$C7,$D7,$87,$28
	.dc.b	$E3,$8F,$81,$ED,$8F,$8D,$FF,$55,$1E,$8A,$E3,$1F,$EB,$7B,$F6,$71
	.dc.b	$7E,$B7,$E8,$BF,$D2,$44,$CB,$F5,$9C,$48,$C6,$3F,$C6,$AB,$85,$6A
	.dc.b	$73,$3F,$E8,$5F,$CF,$1E,$47,$FA,$61,$B5,$FE,$95,$6C,$FF,$A4,$5A
	.dc.b	$F1,$9E,$04,$9E,$0D,$8E,$0D,$78,$46,$88,$E9,$AE,$8C,$10,$C1,$0F
	.dc.b	$F4,$C4,$F9,$1B,$3D,$DB,$3D,$1A,$73,$29,$CC,$FF,$4E,$4F,$36,$CF
	.dc.b	$36,$CF,$36,$9C,$CA,$73,$EA,$57,$36,$D7,$36,$D7,$36,$D7,$36,$BC
	.dc.b	$EB,$91,$2B,$93,$6B,$93,$E9,$97,$4B,$E4,$FF,$B5,$B6,$BF,$BC,$36
	.dc.b	$7F,$B0,$9C,$27,$64,$9D,$B6,$37,$3C,$5F,$ED,$2D,$7F,$BF,$B5,$1D
	.dc.b	$67,$99,$1E,$73,$D5,$E6,$D7,$9B,$5E,$73,$D5,$E8,$72,$9E,$8F,$26
	.dc.b	$BC,$9A,$F2,$9E,$8F,$23,$81,$CC,$86,$08,$73,$32,$53,$27,$58,$EE
	.dc.b	$F5,$34,$43,$47,$22,$9C,$8E,$06,$0E,$06,$09,$3F,$0C,$93,$EC,$6C
	.dc.b	$F8,$9B,$3D,$CE,$33,$B2,$4E,$DB,$1B,$9E,$2F,$95,$AF,$91,$AF,$9C
	.dc.b	$87,$69,$E8,$47,$A4,$F6,$7A,$35,$E8,$D7,$9C,$F6,$7A,$1C,$A7,$A3
	.dc.b	$C9,$AF,$26,$BC,$A7,$A3,$C8,$E0,$73,$21,$82,$1C,$CC,$94,$C9,$D6
	.dc.b	$3B,$BD,$4D,$10,$D1,$C8,$A7,$23,$81,$83,$81,$82,$4F,$C3,$24,$FB
	.dc.b	$1B,$3E,$26,$CF,$73,$8C,$EC,$93,$B6,$C6,$E7,$8B,$E5,$6B,$E4,$6B
	.dc.b	$E7,$21,$DA,$7A,$11,$E9,$3D,$9E,$8D,$7A,$35,$E7,$3D,$9E,$87,$29
	.dc.b	$E8,$F2,$6B,$C9,$AF,$29,$E8,$F2,$38,$1C,$C8,$7B,$FE,$A3,$F9,$59
	.dc.b	$27,$F9,$5B,$63,$F9,$2B,$5F,$E4,$AD,$7F,$92,$B5,$FE,$36,$D7,$F9
	.dc.b	$CB,$68,$D2,$4F,$80,$93,$E0,$6B,$E0,$6B,$E0,$6B,$E0,$6B,$E0,$6C
	.dc.b	$F8,$1B,$3E,$09,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$59,$EC,$D8,$EC
	.dc.b	$F4,$21,$D0,$A3,$3F,$F4,$97,$22,$FF,$C4,$AD,$9F,$F1,$26,$BF,$E5
	.dc.b	$D9,$C9,$23,$33,$B7,$FD,$26,$D7,$FA,$CD,$76,$9F,$F4,$AE,$DF,$E8
	.dc.b	$69,$FF,$41,$6F,$FE,$75,$7F,$D0,$27,$FA,$4B,$9F,$FD,$18,$FF,$9F
	.dc.b	$8D,$D7,$F9,$E5,$FF,$44,$1F,$E8,$A3,$8F,$47,$FA,$39,$FF,$43,$3F
	.dc.b	$E8,$37,$75,$DD,$E0,$47,$04,$9E,$D1,$D2,$BF,$D3,$44,$7F,$D2,$77
	.dc.b	$EB,$7B,$C7,$A2,$FC,$11,$C4,$D9,$EB,$36,$79,$2B,$BA,$BF,$A5,$3F
	.dc.b	$E9,$48,$FF,$49,$73,$FF,$81,$38,$FF,$D2,$50,$69,$87,$FD,$22,$7F
	.dc.b	$82,$B9,$FF,$D2,$D3,$FE,$98,$D3,$FD,$2F,$3F,$E9,$59,$CB,$CF,$A3
	.dc.b	$13,$FE,$A0,$7F,$C2,$71,$FE,$8D,$E6,$BD,$1F,$E7,$E8,$FF,$3B,$9F
	.dc.b	$E7,$A3,$FC,$EE,$1D,$6D,$FE,$85,$CB,$FC,$F3,$3C,$23,$FD,$2F,$3F
	.dc.b	$E9,$DD,$3F,$D0,$13,$FE,$99,$E8,$FF,$43,$56,$9B,$87,$89,$FF,$51
	.dc.b	$5F,$1A,$F1,$1C,$6F,$9C,$66,$FF,$D2,$6F,$FA,$86,$7C,$57,$CE,$BF
	.dc.b	$D3,$95,$AB,$EA,$78,$2B,$FD,$52,$E2,$7F,$D2,$55,$C6,$7C,$37,$DE
	.dc.b	$31,$7E,$9A,$DD,$72,$7C,$B7,$D6,$3F,$D4,$DB,$7F,$AA,$6F,$89,$F4
	.dc.b	$31,$F9,$38,$79,$70,$E2,$74,$BF,$AC,$7F,$A9,$EF,$18,$F8,$E3,$C0
	.dc.b	$2B,$C8,$FC,$23,$81,$4F,$19,$BC,$3C,$6F,$F0,$BE,$17,$5F,$F0,$18
	.dc.b	$C3,$78,$93,$AC,$7F,$D6,$73,$98,$C5,$7F,$10,$8C,$12,$7F,$CD,$FE
	.dc.b	$13,$F6,$A7,$8D,$DE,$30,$D8,$CC,$F0,$6D,$FE,$68,$57,$09,$D5,$71
	.dc.b	$23,$82,$4F,$19,$E0,$DA,$EA,$FE,$9A,$32,$D8,$CB,$FE,$7C,$8C,$B6
	.dc.b	$32,$F9,$1B,$1A,$8C,$57,$91,$F0,$35,$EC,$D7,$A3,$5E,$35,$EA,$76
	.dc.b	$D7,$6D,$76,$76,$71,$3E,$32,$57,$98,$95,$DE,$32,$D8,$CC,$F2,$6D
	.dc.b	$75,$21,$E4,$29,$E4,$29,$E4,$3F,$D5,$3F,$2B,$FE,$B6,$24,$FF,$AD
	.dc.b	$9B,$1F,$EB,$06,$BC,$23,$44,$74,$D7,$46,$08,$60,$87,$FA,$E2,$3A
	.dc.b	$B6,$39,$34,$E2,$53,$89,$FE,$B4,$8E,$2D,$8E,$2D,$38,$94,$E2,$7F
	.dc.b	$B6,$23,$8B,$63,$8B,$4E,$25,$38,$F4,$27,$8B,$67,$8B,$67,$8B,$5E
	.dc.b	$33,$C0,$93,$C1,$B1,$C1,$AF,$08,$D1,$1D,$35,$D1,$82,$18,$21,$F2
	.dc.b	$C8,$7C,$52,$9E,$F2,$CF,$A5,$B3,$E4,$6C,$F8,$5A,$78,$0A,$78,$0E
	.dc.b	$C4,$3B,$16,$7B,$36,$7B,$36,$7B,$34,$EC,$53,$B1,$D0,$93,$D1,$B3
	.dc.b	$D1,$B3,$D1,$A7,$42,$9D,$27,$91,$27,$93,$67,$93,$67,$93,$5E,$53
	.dc.b	$C0,$93,$C1,$B1,$C1,$AF,$08,$D1,$1D,$35,$D1,$82,$18,$21,$F5,$08
	.dc.b	$7C,$92,$9F,$00,$B3,$E9,$6C,$F9,$1B,$3E,$16,$9E,$02,$9E,$03,$B1
	.dc.b	$0E,$C5,$9E,$CD,$9E,$CD,$9E,$CD,$3B,$14,$EC,$74,$24,$F4,$6C,$F4
	.dc.b	$7F,$0D,$DE,$F1,$BF,$B9,$42,$31,$7F,$D8,$60,$66,$77,$3A,$8D,$35
	.dc.b	$FE,$D0,$D7,$FC,$86,$D7,$9B,$5E,$93,$DD,$A7,$49,$EE,$48,$D4,$F1
	.dc.b	$24,$66,$78,$4E,$A3,$11,$E4,$6B,$E4,$6B,$E4,$6B,$E2,$38,$47,$32
	.dc.b	$3D,$88,$68,$A7,$57,$6E,$48,$64,$D9,$E1,$76,$E4,$D9,$93,$67,$94
	.dc.b	$D1,$0D,$1C,$0D,$18,$34,$60,$CE,$BF,$E3,$0F,$6F,$FC,$5C,$49,$FC
	.dc.b	$14,$65,$BB,$27,$55,$FE,$C2,$6B,$F6,$F8,$3F,$D4,$04,$7F,$D3,$8D
	.dc.b	$E9,$FF,$50,$C7,$FA,$B6,$FF,$D3,$BC,$7F,$EA,$D8,$FF,$5F,$1F,$EA
	.dc.b	$12,$9F,$EA,$73,$FD,$1F,$5C,$FA,$BF,$D5,$84,$8E,$17,$FE,$BF,$9D
	.dc.b	$46,$27,$FD,$6C,$F2,$3C,$A4,$9F,$2D,$78,$DE,$51,$99,$EC,$F8,$CF
	.dc.b	$F6,$89,$E6,$3D,$85,$9E,$51,$E0,$E9,$FF,$55,$F3,$7F,$AD,$3A,$3F
	.dc.b	$D5,$7D,$5F,$EB,$97,$FD,$57,$3F,$EB,$0C,$3F,$D6,$E1,$C3,$AE,$AF
	.dc.b	$F5,$BB,$FE,$C2,$3F,$D6,$C5,$B5,$7F,$9D,$23,$FC,$EF,$9F,$FA,$BF
	.dc.b	$4F,$F4,$54,$7F,$A0,$23,$FC,$FE,$7F,$A2,$5F,$F4,$BF,$57,$FA,$47
	.dc.b	$AB,$FD,$49,$3C,$67,$FD,$05,$18,$EA,$EC,$FF,$A2,$E3,$FD,$7D,$8F
	.dc.b	$FA,$52,$73,$7F,$EA,$59,$E3,$39,$7C,$53,$CA,$31,$1F,$E9,$83,$9E
	.dc.b	$1E,$87,$D7,$3D,$36,$FF,$4F,$C7,$49,$D6,$1E,$58,$FF,$68,$BF,$ED
	.dc.b	$0A,$E1,$1E,$EA,$FF,$51,$E1,$EF,$C3,$C3,$7E,$B3,$85,$72,$8F,$47
	.dc.b	$7B,$F2,$B7,$C0,$7F,$A7,$E3,$0D,$8C,$E7,$FC,$BE,$72,$D8,$CD,$FF
	.dc.b	$A4,$E7,$2D,$8C,$BF,$EF,$08,$CB,$FE,$99,$FA,$17,$C7,$FF,$3E,$6D
	.dc.b	$FF,$2F,$6C,$FF,$3F,$6B,$C6,$78,$12,$78,$36,$38,$35,$E1,$1A,$23
	.dc.b	$A6,$BA,$30,$43,$04,$3F,$A8,$89,$F2,$B8,$6B,$FD,$59,$AE,$E7,$C8
	.dc.b	$ED,$AE,$DA,$EE,$7C,$AE,$DA,$ED,$AE,$CF,$41,$B2,$9B,$29,$B3,$D8
	.dc.b	$64,$A6,$4E,$46,$27,$D6,$49,$F3,$B6,$7C,$6B,$E4,$3E,$F1,$23,$EF
	.dc.b	$36,$BE,$F3,$5F,$C3,$35,$F9,$8D,$7E,$63,$5F,$A4,$D9,$F0,$4F,$62
	.dc.b	$4F,$66,$BD,$9A,$F6,$6B,$D9,$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$85
	.dc.b	$3A,$14,$E8,$57,$A3,$C8,$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81
	.dc.b	$4E,$05,$78,$3A,$21,$A2,$BA,$70,$47,$04,$9F,$A8,$49,$F9,$2D,$9F
	.dc.b	$80,$D7,$D2,$D7,$C8,$D7,$C2,$D7,$C0,$D9,$F0,$4F,$62,$4F,$66,$BD
	.dc.b	$9A,$F6,$6B,$D9,$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$85,$3A,$14,$E8
	.dc.b	$57,$A3,$C8,$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78
	.dc.b	$3A,$21,$A2,$BA,$70,$47,$04,$9F,$A8,$49,$F9,$2D,$9F,$80,$D7,$D2
	.dc.b	$D7,$C8,$D7,$C2,$D7,$C0,$D9,$F0,$4F,$62,$4F,$66,$BD,$9A,$F6,$6B
	.dc.b	$D9,$AF,$66,$C7,$68,$E8,$43,$A1,$4E,$87,$DF,$EC,$7C,$47,$FA,$79
	.dc.b	$B5,$FE,$8F,$6C,$FF,$A7,$DA,$FF,$50,$0E,$1F,$EA,$6D,$7F,$D4,$AD
	.dc.b	$77,$3D,$C9,$3D,$DB,$1D,$DA,$F7,$9E,$44,$8C,$C7,$81,$AF,$81,$AF
	.dc.b	$82,$7B,$11,$EC,$D7,$B4,$70,$9E,$D3,$99,$EC,$D7,$B3,$5E,$C7,$42
	.dc.b	$1D,$1D,$47,$48,$C9,$D0,$A7,$42,$9D,$0E,$46,$1E,$4E,$08,$73,$3F
	.dc.b	$C0,$BD,$4E,$A7,$BE,$BF,$F4,$BB,$6B,$EE,$56,$E3,$11,$CA,$B8,$D7
	.dc.b	$02,$47,$08,$CB,$67,$FD,$D2,$FF,$A0,$3A,$BF,$D0,$A7,$FA,$22,$BF
	.dc.b	$D3,$64,$8F,$F4,$13,$63,$7D,$5F,$E8,$A9,$FF,$43,$C6,$1F,$F4,$B1
	.dc.b	$F9,$3E,$6F,$F4,$CB,$FE,$8F,$0E,$1E,$33,$FE,$9C,$7C,$8D,$76,$7F
	.dc.b	$A6,$4E,$A7,$FA,$86,$7D,$0F,$98,$8F,$8D,$F0,$4F,$7E,$5F,$F4,$DC
	.dc.b	$79,$4E,$86,$8A,$79,$0D,$4A,$BF,$5D,$3F,$56,$31,$3F,$E9,$7E,$7F
	.dc.b	$F5,$1B,$FE,$9A,$3F,$D4,$1D,$1F,$E9,$E6,$F5,$7E,$DD,$18,$31,$D5
	.dc.b	$FE,$A6,$7A,$10,$FF,$45,$74,$7F,$9E,$93,$FE,$7D,$E9,$FF,$4A,$69
	.dc.b	$FE,$A1,$27,$57,$F9,$52,$78,$1B,$D3,$C3,$3F,$EA,$89,$FF,$45,$B6
	.dc.b	$FF,$D3,$64,$3F,$D3,$27,$FA,$A2,$FB,$CF,$03,$B7,$57,$FA,$42,$3F
	.dc.b	$D2,$5B,$FF,$A6,$9D,$D7,$3C,$7F,$D3,$8F,$AA,$3F,$D6,$C6,$9F,$7D
	.dc.b	$73,$C7,$FD,$4B,$8F,$B2,$79,$4E,$5F,$F5,$11,$FE,$A9,$3C,$D3,$EE
	.dc.b	$C7,$CB,$8F,$8E,$31,$3F,$4B,$1F,$8B,$8F,$8A,$73,$3F,$03,$1F,$07
	.dc.b	$17,$EE,$7D,$0F,$D9,$CF,$DA,$8C,$63,$F7,$5B,$1F,$B3,$9D,$A8,$CB
	.dc.b	$FB,$B8,$CC,$62,$3A,$71,$A7,$8E,$CF,$C8,$DE,$56,$1B,$8F,$E3,$67
	.dc.b	$53,$92,$46,$67,$78,$7F,$49,$AD,$CE,$6B,$73,$9A,$DB,$FE,$9E,$AD
	.dc.b	$B9,$9D,$B9,$9D,$BF,$9E,$9D,$B9,$9D,$B9,$9D,$A9,$DB,$99,$DB,$99
	.dc.b	$DC,$EA,$31,$3A,$8C,$12,$76,$48,$D3,$6B,$E0,$12,$BD,$8D,$7D,$0D
	.dc.b	$7C,$8D,$7C,$2D,$AF,$05,$76,$24,$66,$BC,$15,$D8,$87,$62,$9F,$EC
	.dc.b	$9E,$8F,$F7,$21,$27,$FD,$C8,$D8,$FF,$72,$35,$FF,$76,$35,$E5,$1C
	.dc.b	$08,$F0,$6B,$C1,$AF,$03,$44,$34,$53,$46,$08,$60,$87,$FB,$86,$3C
	.dc.b	$8D,$8E,$ED,$3A,$14,$E6,$53,$99,$FE,$EF,$8E,$6D,$8E,$6D,$39,$94
	.dc.b	$E6,$53,$9C,$7F,$BE,$63,$8B,$63,$8B,$5E,$2D,$78,$BE,$23,$0F,$9C
	.dc.b	$87,$72,$BA,$70,$47,$04,$3A,$12,$3C,$AD,$8F,$2B,$5F,$2B,$5F,$2C
	.dc.b	$70,$23,$C1,$AF,$06,$BC,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$02,$3C
	.dc.b	$8D,$8E,$ED,$3A,$14,$E6,$53,$99,$F1,$E3,$9B,$63,$9B,$4E,$65,$39
	.dc.b	$94,$E7,$1F,$3E,$38,$B6,$38,$B5,$E2,$D7,$8B,$E2,$30,$F9,$C8,$77
	.dc.b	$2B,$A7,$04,$70,$43,$A1,$23,$CA,$D8,$F2,$B5,$F2,$B5,$F2,$C7,$02
	.dc.b	$3C,$1A,$F0,$6B,$C0,$D1,$0D,$14,$D1,$82,$18,$21,$F0,$23,$C8,$D8
	.dc.b	$EE,$D3,$A1,$4E,$67,$F8,$A3,$C0,$79,$BF,$DB,$21,$18,$9F,$F1,$74
	.dc.b	$62,$75,$39,$8F,$F1,$93,$5F,$F5,$53,$5F,$F1,$A9,$C6,$79,$C6,$63
	.dc.b	$C5,$3D,$09,$3A,$6C,$76,$6B,$D9,$AF,$69,$E8,$F7,$38,$4F,$32,$38
	.dc.b	$21,$CD,$CB,$4C,$9D,$67,$BB,$E8,$8D,$B9,$6B,$93,$89,$4E,$84,$38
	.dc.b	$1A,$3D,$06,$4E,$06,$8C,$1A,$30,$43,$64,$34,$59,$F8,$A4,$9F,$F1
	.dc.b	$3E,$6F,$A9,$A4,$BF,$F1,$94,$65,$B1,$9B,$FF,$53,$35,$FC,$3E,$DF
	.dc.b	$E8,$79,$FF,$50,$74,$7F,$A4,$E3,$13,$FE,$8E,$8C,$3F,$E8,$8E,$0F
	.dc.b	$F4,$6B,$FE,$95,$8F,$F5,$34,$62,$3F,$D3,$A7,$FA,$32,$33,$5D,$A3
	.dc.b	$07,$FA,$78,$D9,$CC,$C1,$FE,$A4,$E8,$FF,$50,$9F,$E9,$F7,$FD,$30
	.dc.b	$6C,$FF,$4E,$3A,$9F,$3B,$63,$B9,$A8,$FF,$4E,$CF,$8E,$3F,$D2,$9C
	.dc.b	$AE,$8F,$F4,$EB,$63,$FD,$33,$CE,$E8,$FF,$1C,$47,$FA,$86,$3F,$D3
	.dc.b	$4F,$FA,$74,$FD,$59,$4F,$D6,$BA,$39,$71,$FF,$A9,$2B,$B4,$F8,$AF
	.dc.b	$C2,$7F,$A3,$DF,$F3,$C4,$7F,$A0,$74,$FF,$42,$47,$F9,$DF,$A3,$4F
	.dc.b	$FA,$37,$5F,$F3,$DF,$47,$F9,$F9,$BD,$1F,$E9,$69,$FF,$43,$9F,$ED
	.dc.b	$6A,$FF,$43,$B8,$BF,$F5,$23,$FE,$94,$D7,$FD,$14,$EC,$F1,$B8,$8D
	.dc.b	$E0,$BE,$D1,$E6,$8C,$61,$FE,$9C,$C3,$CF,$1E,$5C,$3F,$D3,$55,$BB
	.dc.b	$E3,$5A,$7F,$D4,$D5,$CA,$BF,$D4,$35,$BB,$F7,$E3,$F1,$70,$FF,$61
	.dc.b	$61,$C0,$E7,$7F,$EB,$36,$BF,$1C,$F0,$C7,$7A,$E7,$8F,$C8,$64,$F8
	.dc.b	$27,$FD,$2F,$E4,$7E,$D0,$95,$FB,$26,$CF,$E0,$BA,$5C,$69,$CC,$FF
	.dc.b	$30,$9C,$C7,$FA,$A3,$E9,$DE,$03,$D2,$C3,$71,$FF,$76,$4E,$A7,$24
	.dc.b	$8C,$CE,$F0,$FC,$9D,$6E,$73,$5B,$9C,$D6,$DF,$E8,$75,$B7,$33,$B7
	.dc.b	$33,$B7,$FD,$FB,$3B,$73,$3B,$73,$3B,$53,$B7,$33,$B7,$33,$B9,$D4
	.dc.b	$62,$75,$18,$24,$EC,$91,$A6,$D7,$C0,$25,$7B,$1A,$FA,$1A,$F9,$1A
	.dc.b	$F8,$5B,$5E,$0A,$EC,$48,$CD,$78,$2B,$B1,$0E,$C5,$3F,$B8,$F4,$BE
	.dc.b	$5F,$FB,$F3,$6B,$FB,$EB,$67,$F8,$68,$70,$FF,$10,$6B,$D2,$79,$92
	.dc.b	$79,$B6,$39,$CF,$04,$75,$8D,$3D,$5A,$F5,$9E,$64,$79,$C6,$9E,$73
	.dc.b	$D5,$E9,$3B,$7A,$35,$E8,$72,$70,$F2,$9E,$8F,$23,$44,$39,$94,$E6
	.dc.b	$60,$86,$08,$75,$8E,$EF,$53,$66,$4E,$85,$32,$6C,$E6,$53,$46,$27
	.dc.b	$E1,$12,$7D,$6D,$9F,$0B,$67,$B9,$CC,$8E,$5F,$0B,$5E,$D3,$D0,$93
	.dc.b	$D1,$B1,$D2,$78,$3E,$62,$1D,$A3,$6F,$66,$BD,$A7,$A1,$1E,$91,$A7
	.dc.b	$9C,$F6,$7A,$4E,$DE,$8D,$7A,$1C,$9C,$3C,$A7,$A3,$C8,$D1,$0E,$65
	.dc.b	$39,$98,$21,$82,$1D,$63,$BB,$D4,$D9,$93,$A1,$4C,$9B,$39,$94,$D1
	.dc.b	$89,$F8,$44,$9F,$5B,$67,$C2,$D9,$EE,$73,$23,$97,$C2,$D7,$B4,$F4
	.dc.b	$24,$F4,$6C,$74,$9E,$0F,$98,$87,$68,$DB,$D9,$AF,$69,$E8,$47,$A4
	.dc.b	$69,$E7,$3D,$9E,$93,$B7,$A3,$5E,$87,$27,$0F,$29,$E8,$F2,$34,$43
	.dc.b	$99,$F2,$3A,$9E,$8F,$FA,$90,$95,$FE,$9E,$6C,$FF,$A9,$9A,$FF,$A9
	.dc.b	$9A,$FF,$A6,$5A,$FF,$AD,$5A,$FF,$AD,$1B,$3E,$09,$EC,$49,$EC,$D7
	.dc.b	$B3,$5E,$CD,$7B,$35,$EC,$D8,$ED,$1D,$08,$74,$29,$D0,$A7,$42,$9D
	.dc.b	$0A,$F4,$79,$10,$E4,$53,$91,$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF
	.dc.b	$05,$A7,$F8,$AB,$CD,$FF,$51,$92,$BF,$D4,$2D,$9F,$F1,$86,$C8,$C4
	.dc.b	$7F,$8B,$E3,$13,$FE,$E8,$8F,$F4,$89,$0F,$F4,$87,$4F,$FA,$61,$FF
	.dc.b	$51,$1F,$E8,$A3,$FD,$22,$6F,$93,$FD,$4F,$D5,$FE,$95,$E3,$FF,$51
	.dc.b	$4F,$23,$FD,$37,$3A,$AF,$F4,$A9,$D6,$B8,$5F,$FA,$4E,$BA,$1F,$EA
	.dc.b	$F7,$0F,$33,$FD,$34,$74,$34,$79,$8F,$F5,$21,$E6,$38,$16,$BC,$4F
	.dc.b	$AE,$FC,$F1,$88,$FF,$49,$1F,$EA,$0E,$AF,$F4,$FF,$37,$FA,$5B,$AB
	.dc.b	$FD,$33,$3F,$EA,$19,$CF,$4F,$FA,$79,$FF,$52,$C7,$FA,$8A,$7F,$D4
	.dc.b	$11,$88,$D7,$57,$F8,$69,$EE,$49,$EB,$7D,$EB,$AC,$FF,$9E,$71,$FF
	.dc.b	$3A,$BF,$E9,$7E,$8F,$F4,$CE,$5F,$E8,$38,$FF,$3B,$1B,$3F,$D3,$07
	.dc.b	$FA,$9C,$C3,$FE,$95,$BF,$F4,$44,$7F,$88,$3A,$BF,$D1,$D3,$FE,$9C
	.dc.b	$EA,$FF,$43,$C6,$2F,$FD,$33,$7C,$2F,$FD,$15,$7B,$C3,$FD,$3D,$1D
	.dc.b	$AB,$CA,$7F,$A3,$5F,$1D,$FF,$A8,$6B,$D0,$71,$8F,$49,$FE,$B1,$3D
	.dc.b	$C7,$FA,$C0,$A7,$AC,$F0,$CF,$93,$5F,$F5,$A3,$E0,$8F,$11,$ED,$C7
	.dc.b	$FD,$49,$87,$9A,$3B,$5F,$8A,$BF,$D5,$33,$F3,$A3,$13,$C6,$31,$3E
	.dc.b	$B9,$F4,$1F,$2E,$FF,$D4,$EA,$3F,$C9,$7C,$5F,$EA,$2F,$25,$5F,$E5
	.dc.b	$A8,$C3,$78,$93,$AC,$7F,$76,$D7,$F7,$4D,$7A,$C7,$09,$E5,$39,$9F
	.dc.b	$E1,$33,$97,$FD,$3F,$E1,$FF,$AB,$09,$5F,$EA,$86,$C6,$A3,$04,$70
	.dc.b	$43,$FD,$47,$1C,$1B,$1B,$69,$B3,$FD,$47,$1B,$6C,$6D,$A6,$CF,$F5
	.dc.b	$1C,$6D,$B1,$B6,$9B,$EB,$FF,$64,$CE,$DB,$3B,$6C,$EF,$A9,$5B,$6D
	.dc.b	$6D,$B3,$B9,$D1,$27,$4D,$8D,$46,$08,$FA,$CF,$F4,$C7,$C1,$7F,$A7
	.dc.b	$89,$1F,$E9,$E6,$C7,$FA,$51,$AF,$FA,$75,$AF,$FA,$A1,$AF,$FA,$39
	.dc.b	$AF,$68,$E8,$47,$A3,$5E,$8D,$7A,$35,$E8,$D7,$A1,$C8,$87,$22,$9C
	.dc.b	$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$EB,$4F,$C8,$6C,$FB,$DB,$3E,$86,$BE,$36,$BE,$06,$BD,$DA,$F7
	.dc.b	$9E,$C4,$9E,$CD,$8E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$D1,$D0,$8F,$46
	.dc.b	$BD,$1A,$F4,$6B,$D1,$AF,$43,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0
	.dc.b	$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$09,$3F,$50,$93,$F2
	.dc.b	$5B,$3F,$01,$B3,$E9,$6B,$E4,$6B,$E1,$6B,$E0,$6B,$E0,$9E,$C4,$9E
	.dc.b	$CD,$8E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$D1,$D0,$8F,$46,$BD,$1A,$F4
	.dc.b	$6B,$D1,$AF,$43,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0,$FD,$E7,$4B
	.dc.b	$C1,$FE,$B1,$08,$C4,$FF,$89,$DB,$02,$15,$CB,$83,$94,$F4,$8C,$C7
	.dc.b	$06,$BC,$63,$44,$9F,$03,$63,$A1,$D2,$32,$47,$33,$D1,$D1,$89,$E4
	.dc.b	$F9,$63,$44,$8C,$C7,$95,$AF,$96,$7B,$3C,$63,$2F,$63,$11,$A7,$93
	.dc.b	$5E,$47,$8D,$C3,$C4,$D9,$92,$1D,$0A,$68,$C1,$0C,$10,$EE,$64,$A6
	.dc.b	$4F,$01,$82,$18,$D5,$A3,$B7,$FC,$39,$18,$6C,$66,$BF,$C5,$51,$82
	.dc.b	$38,$25,$7F,$A8,$78,$3F,$D0,$0F,$FA,$43,$A5,$D5,$FE,$95,$3F,$D2
	.dc.b	$F1,$FE,$8F,$8C,$B6,$34,$FF,$A2,$EB,$FD,$2F,$5D,$23,$FD,$23,$C7
	.dc.b	$FE,$8E,$7F,$D2,$F7,$FE,$A0,$7F,$D3,$07,$86,$F9,$BF,$E9,$A9,$E1
	.dc.b	$3E,$08,$FF,$50,$90,$FF,$4E,$1D,$27,$CE,$E6,$78,$C7,$48,$FD,$45
	.dc.b	$7E,$A6,$BA,$42,$BF,$CA,$11,$DF,$A3,$FD,$27,$D5,$FE,$98,$AF,$F5
	.dc.b	$07,$57,$EC,$7A,$7F,$D2,$CD,$9D,$4F,$FA,$87,$AB,$FD,$40,$FF,$AA
	.dc.b	$E7,$FD,$41,$7F,$EA,$47,$84,$64,$EA,$4B,$5F,$F9,$F6,$3F,$D0,$3D
	.dc.b	$5F,$E8,$9E,$9C,$69,$FE,$76,$9F,$F4,$19,$FE,$87,$71,$3F,$E9,$98
	.dc.b	$C4,$FF,$9E,$4F,$F4,$33,$C0,$FF,$41,$B9,$76,$7F,$A2,$8D,$47,$85
	.dc.b	$C3,$5E,$26,$3A,$BF,$D1,$EF,$FA,$72,$3C,$4F,$2C,$3F,$D5,$51,$83
	.dc.b	$91,$67,$8D,$FF,$A3,$6B,$E0,$D7,$09,$CD,$FF,$A6,$27,$FD,$5B,$39
	.dc.b	$D7,$FD,$2F,$1E,$36,$BE,$AA,$FF,$5E,$9F,$81,$7F,$D6,$AE,$23,$FD
	.dc.b	$37,$5D,$6F,$FD,$3E,$D3,$E4,$E1,$E7,$9F,$F6,$0F,$17,$EE,$3E,$3B
	.dc.b	$F6,$A4,$E1,$FF,$4F,$37,$2F,$E6,$D3,$B9,$C9,$23,$F9,$D7,$9C,$F6
	.dc.b	$1F,$05,$FE,$9D,$8C,$36,$33,$5F,$CD,$A7,$2D,$8C,$D7,$46,$BF,$D2
	.dc.b	$A7,$6E,$6B,$A1,$2B,$93,$63,$51,$82,$38,$21,$D6,$BB,$BF,$D6,$6B
	.dc.b	$BB,$9A,$E8,$DA,$E6,$D3,$99,$C6,$B9,$AA,$E6,$E6,$B9,$B6,$B9,$B4
	.dc.b	$E6,$71,$AE,$75,$CA,$31,$5C,$9B,$5C,$9A,$72,$9E,$04,$AE,$64,$AE
	.dc.b	$51,$96,$C6,$6B,$A4,$E5,$ED,$4E,$A7,$BD,$FE,$06,$6C,$FF,$74,$9D
	.dc.b	$4E,$49,$19,$9D,$CF,$F7,$A6,$BF,$E0,$76,$BF,$E0,$C6,$BF,$E0,$F6
	.dc.b	$BD,$A3,$A4,$F6,$8E,$8D,$7A,$35,$E8,$D7,$A3,$5E,$87,$28,$EA,$47
	.dc.b	$04,$3A,$B9,$69,$93,$BB,$96,$99,$3C,$2E,$5A,$64,$F1,$9D,$08,$72
	.dc.b	$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$93,$F4,$09,$3F,$1D,$B3
	.dc.b	$E9,$9D,$CE,$49,$19,$9D,$CF,$B5,$AF,$A1,$AF,$95,$AF,$91,$AF,$68
	.dc.b	$E9,$3D,$A3,$A3,$5E,$8D,$7A,$35,$E8,$D7,$A1,$CA,$3A,$91,$C1,$0E
	.dc.b	$AE,$5A,$64,$EE,$E5,$A6,$4F,$0B,$96,$99,$3C,$67,$42,$1C,$8A,$70
	.dc.b	$29,$C0,$D1,$0D,$14,$D1,$82,$18,$24,$FD,$02,$4F,$C7,$6C,$FA,$67
	.dc.b	$73,$92,$46,$67,$73,$ED,$6B,$E8,$6B,$E5,$6B,$E4,$6B,$DA,$3A,$4F
	.dc.b	$68,$E8,$D7,$A3,$5E,$8D,$7A,$35,$E8,$72,$8E,$A4,$70,$43,$AB,$96
	.dc.b	$99,$3B,$B9,$69,$F0,$7E,$03,$F9,$51,$23,$F9,$33,$63,$F9,$33,$5F
	.dc.b	$E4,$CD,$7F,$93,$35,$FE,$4C,$D7,$F9,$C3,$68,$D2,$4F,$80,$93,$E0
	.dc.b	$6B,$E0,$6B,$E0,$6B,$E0,$6B,$E0,$6C,$F8,$1B,$3E,$09,$EC,$43,$B1
	.dc.b	$4E,$C5,$3B,$14,$EC,$59,$EC,$D8,$EC,$F4,$21,$D0,$A3,$B9,$F2,$2F
	.dc.b	$8C,$61,$FE,$D0,$9C,$C6,$2B,$FD,$BD,$18,$9D,$57,$16,$D6,$EB,$FD
	.dc.b	$CF,$5B,$78,$D6,$DB,$3F,$EF,$27,$FD,$28,$7F,$A7,$C8,$7F,$A3,$6B
	.dc.b	$FD,$3E,$FF,$A4,$4F,$F5,$11,$FE,$9C,$3F,$D4,$2E,$BA,$7F,$D1,$FC
	.dc.b	$1F,$E9,$49,$E6,$7F,$A6,$A3,$33,$BE,$AF,$F4,$C4,$F2,$3A,$72,$7F
	.dc.b	$AB,$48,$F8,$4F,$F5,$44,$F1,$7A,$DF,$77,$D0,$72,$38,$1E,$E3,$37
	.dc.b	$FE,$B5,$3F,$D4,$5D,$4E,$9F,$F4,$BC,$BA,$BF,$D4,$AF,$FA,$67,$A7
	.dc.b	$FD,$46,$FF,$A6,$8F,$F5,$14,$FF,$A8,$2B,$FD,$47,$1C,$5B,$5C,$A3
	.dc.b	$13,$DA,$30,$67,$9F,$FC,$DD,$3F,$E7,$EC,$BF,$CE,$AF,$FA,$4B,$0F
	.dc.b	$F4,$03,$93,$FD,$0F,$1A,$7F,$D2,$A7,$FA,$07,$2C,$57,$FA,$43,$A7
	.dc.b	$FD,$01,$3F,$E8,$29,$CD,$F4,$D3,$FD,$03,$23,$FD,$35,$87,$2B,$E3
	.dc.b	$3F,$EB,$0A,$E1,$D3,$FE,$86,$9F,$F4,$74,$E6,$B8,$61,$D9,$B8,$7F
	.dc.b	$A9,$EB,$9D,$FF,$AF,$8F,$F5,$A4,$31,$F4,$5F,$18,$F7,$93,$0F,$F6
	.dc.b	$0D,$FF,$A9,$A3,$D1,$5F,$EA,$FC,$7C,$B1,$FE,$BE,$3E,$21,$E1,$BE
	.dc.b	$97,$C9,$FF,$54,$D7,$D2,$8F,$F5,$A1,$99,$FA,$0B,$0F,$F5,$34,$3A
	.dc.b	$1D,$8F,$03,$F6,$93,$F2,$E7,$31,$88,$E1,$3B,$D3,$E8,$44,$AD,$C6
	.dc.b	$23,$94,$ED,$FF,$56,$B5,$E0,$D7,$75,$FE,$AB,$F9,$B7,$C3,$52,$5F
	.dc.b	$FA,$AE,$32,$D8,$CD,$7F,$AD,$A7,$2D,$8C,$BF,$EA,$D8,$CB,$63,$2F
	.dc.b	$FA,$8E,$32,$D8,$CB,$FE,$9D,$8C,$B6,$33,$D4,$9C,$B6,$33,$18,$23
	.dc.b	$82,$57,$AC,$95,$E9,$6B,$E7,$6B,$E5,$6B,$E3,$6D,$78,$5B,$5E,$06
	.dc.b	$D7,$82,$BB,$10,$EC,$53,$B1,$4E,$C5,$AE,$CE,$4F,$F5,$37,$D4,$FF
	.dc.b	$A9,$49,$5F,$EA,$16,$C7,$FA,$85,$AF,$FA,$85,$AF,$FA,$85,$AF,$FA
	.dc.b	$85,$AF,$68,$E8,$47,$A3,$5E,$8D,$7A,$35,$E8,$D7,$A1,$C8,$87,$22
	.dc.b	$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$EA,$4F,$C8,$6C,$FB,$DB,$3E,$86,$BE,$36,$BE,$06,$BD,$DA
	.dc.b	$F7,$9E,$C4,$9E,$CD,$8E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$D1,$D0,$8F
	.dc.b	$46,$BD,$1A,$F4,$6B,$D1,$AF,$43,$91,$0E,$45,$39,$14,$E4,$53,$91
	.dc.b	$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$09,$3F,$50,$93
	.dc.b	$F2,$5B,$3F,$01,$B3,$E9,$6B,$E4,$6B,$E1,$6B,$E0,$6B,$E0,$9E,$C4
	.dc.b	$9E,$CD,$8E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$D1,$D0,$8F,$46,$BD,$1A
	.dc.b	$F4,$6B,$D1,$AF,$43,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0,$FD,$E7
	.dc.b	$84,$F6,$FF,$B6,$42,$31,$3F,$DA,$63,$FC,$39,$18,$70,$D7,$FB,$73
	.dc.b	$5F,$EE,$6D,$78,$CF,$68,$C4,$77,$8C,$12,$7C,$11,$96,$CF,$02,$4E
	.dc.b	$E7,$C6,$D7,$C6,$D7,$C6,$D7,$C7,$3D,$23,$11,$DC,$E7,$19,$23,$82
	.dc.b	$1D,$9C,$B4,$C9,$E0,$72,$D3,$27,$88,$A7,$84,$E4,$43,$81,$4E,$05
	.dc.b	$38,$1A,$21,$A2,$9A,$30,$4D,$5F,$9B,$7F,$96,$67,$F0,$11,$89,$FF
	.dc.b	$18,$36,$33,$3B,$9F,$F1,$74,$E5,$B1,$97,$FC,$68,$D7,$FC,$78,$D7
	.dc.b	$FD,$4B,$D1,$FE,$8D,$E9,$FC,$87,$17,$FA,$6A,$3F,$D4,$33,$FE,$95
	.dc.b	$69,$B7,$8C,$F0,$7F,$D3,$27,$FA,$96,$34,$FF,$A8,$8E,$51,$FE,$AA
	.dc.b	$9E,$CD,$8E,$31,$87,$FD,$4F,$87,$FA,$A5,$FF,$51,$CF,$7A,$F1,$B4
	.dc.b	$E4,$E2,$3C,$A4,$3C,$4F,$91,$FF,$57,$1F,$EB,$F8,$FF,$2C,$72,$7F
	.dc.b	$B6,$E7,$FD,$5F,$D1,$FE,$A1,$E9,$FF,$58,$5B,$9F,$FD,$5A,$4F,$0B
	.dc.b	$FC,$C7,$D4,$AF,$F5,$59,$2F,$FD,$60,$4B,$E5,$5C,$27,$AD,$F3,$9F
	.dc.b	$F5,$D4,$FF,$A3,$DF,$F4,$3F,$57,$FA,$63,$A7,$11,$FE,$86,$DB,$FD
	.dc.b	$03,$1F,$E9,$4E,$AF,$F4,$9C,$FF,$A2,$49,$3A,$79,$EB,$FE,$9E,$8F
	.dc.b	$F5,$5C,$7F,$AA,$8F,$F4,$C5,$73,$C7,$FD,$2B,$1F,$E9,$F6,$DF,$6A
	.dc.b	$E9,$3E,$3A,$FF,$47,$BB,$DF,$B4,$62,$B8,$36,$F9,$92,$77,$5E,$63
	.dc.b	$FD,$37,$3E,$38,$CC,$79,$EF,$FD,$40,$7A,$AF,$FD,$59,$3A,$6C,$FB
	.dc.b	$0E,$93,$FE,$C1,$70,$7A,$4C,$D7,$FB,$1F,$0E,$58,$79,$67,$76,$C7
	.dc.b	$EA,$47,$FB,$4F,$97,$FD,$05,$97,$FB,$57,$2F,$F5,$94,$E7,$A1,$39
	.dc.b	$CF,$FD,$57,$6F,$07,$36,$C3,$8A,$9E,$AE,$6B,$8E,$1C,$2B,$75,$A6
	.dc.b	$DE,$6B,$78,$78,$70,$F4,$7C,$F7,$F5,$4C,$0D,$24,$FF,$46,$8C,$B6
	.dc.b	$33,$3F,$D6,$67,$2D,$8C,$BF,$EC,$58,$CB,$63,$2F,$C2,$6B,$E4,$9E
	.dc.b	$C4,$7A,$35,$E4,$D7,$93,$5E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D
	.dc.b	$14,$D1,$82,$18,$21,$F6,$63,$DA,$D8,$F3,$34,$F0,$94,$EC,$53,$A9
	.dc.b	$4E,$B3,$F3,$1C,$B5,$FC,$A7,$D6,$FD,$D2,$4F,$DB,$6C,$7D,$96,$BF
	.dc.b	$5D,$AF,$E5,$9A,$F4,$8E,$44,$79,$35,$E4,$D7,$93,$5E,$47,$02,$1C
	.dc.b	$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18,$21,$FA,$D8,$F6,$B6,$3C
	.dc.b	$CD,$3C,$25,$3B,$14,$EA,$53,$AC,$A7,$AB,$67,$AB,$67,$AB,$5E,$AD
	.dc.b	$7A,$B5,$EB,$3D,$09,$3D,$1B,$1D,$1A,$F4,$6B,$D1,$AF,$48,$E4,$47
	.dc.b	$93,$5E,$4D,$79,$35,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D
	.dc.b	$18,$21,$82,$1F,$4C,$93,$F2,$5B,$3F,$01,$B3,$E9,$69,$E4,$29,$E1
	.dc.b	$29,$E0,$29,$E0,$9E,$C4,$9E,$CD,$9E,$CD,$9E,$CD,$7B,$35,$EC,$D7
	.dc.b	$B4,$F4,$24,$F4,$6C,$74,$6B,$D1,$AF,$46,$BD,$23,$91,$1E,$4D,$79
	.dc.b	$35,$E4,$D7,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86
	.dc.b	$08,$7E,$07,$A3,$FC,$6D,$1F,$E9,$EE,$1F,$F5,$31,$1C,$12,$7F,$D6
	.dc.b	$31,$96,$C6,$5F,$F5,$9C,$65,$B1,$97,$FD,$6B,$19,$6C,$65,$FF,$5F
	.dc.b	$C6,$5B,$19,$7F,$D8,$D1,$96,$C6,$67,$D7,$1E,$42,$1E,$22,$9E,$02
	.dc.b	$9D,$8A,$74,$2B,$D1,$E4,$43,$91,$4E,$45,$39,$15,$E4,$F0,$21,$C0
	.dc.b	$A7,$02,$BC,$1D,$13,$9F,$FD,$53,$C3,$FE,$A6,$E7,$34,$97,$FE,$A8
	.dc.b	$6D,$7F,$A9,$9A,$FF,$B8,$36,$FF,$48,$3F,$E8,$E3,$FD,$10,$7F,$A1
	.dc.b	$A3,$FD,$09,$CF,$FE,$8F,$8F,$F4,$A9,$FE,$9C,$2F,$3F,$FA,$35,$FF
	.dc.b	$51,$46,$DE,$87,$FA,$49,$C3,$FE,$91,$38,$B8,$9E,$F1,$E2,$39,$96
	.dc.b	$BC,$EF,$13,$A9,$E7,$BF,$F5,$43,$92,$4F,$47,$C4,$73,$3F,$D6,$24
	.dc.b	$BF,$F5,$53,$A2,$1F,$E5,$B0,$EB,$9F,$FD,$41,$CF,$FE,$A7,$24,$FF
	.dc.b	$A7,$8F,$F4,$D8,$11,$C1,$C1,$FE,$A3,$8F,$F5,$33,$83,$7A,$FF,$AA
	.dc.b	$DE,$86,$23,$BF,$37,$FA,$A6,$3F,$CF,$47,$FA,$44,$FF,$3B,$F2,$FF
	.dc.b	$9D,$9C,$F3,$7F,$A4,$9F,$F3,$F9,$E5,$70,$FF,$A3,$79,$3F,$D2,$0F
	.dc.b	$2C,$FF,$D0,$7A,$7F,$A3,$6F,$A5,$F0,$3F,$D1,$FA,$FF,$A5,$30,$FF
	.dc.b	$45,$61,$FE,$77,$3B,$9D,$4B,$3F,$E8,$BC,$3F,$CE,$91,$93,$CF,$7D
	.dc.b	$B1,$FF,$49,$61,$D6,$7D,$11,$8C,$3C,$F3,$99,$F8,$4D,$C3,$CD,$7C
	.dc.b	$F1,$F5,$63,$E0,$9C,$B7,$1E,$98,$79,$AF,$DE,$71,$7E,$21,$ED,$3E
	.dc.b	$6C,$70,$7D,$33,$CE,$FB,$C7,$FA,$F0,$E9,$1E,$1A,$E9,$87,$9E,$FF
	.dc.b	$D3,$15,$F2,$38,$FF,$D4,$38,$7F,$A2,$F9,$BF,$D4,$ED,$3F,$1C,$59
	.dc.b	$FF,$55,$9F,$EB,$17,$2F,$FA,$88,$C9,$FE,$AD,$8E,$6E,$63,$FD,$31
	.dc.b	$DD,$FE,$A5,$23,$FE,$A3,$6C,$7F,$A8,$5A,$F0,$8D,$11,$D3,$5D,$18
	.dc.b	$21,$82,$1F,$EC,$68,$EA,$D8,$E4,$D3,$89,$4E,$27,$FB,$56,$38,$B6
	.dc.b	$38,$B4,$E2,$53,$89,$FE,$D1,$8E,$2D,$8E,$2D,$38,$94,$E3,$49,$E2
	.dc.b	$D9,$E2,$D9,$E2,$D7,$8C,$F0,$24,$F0,$7F,$DB,$1D,$F2,$3F,$DA,$CD
	.dc.b	$9F,$F6,$A3,$5F,$F6,$93,$5F,$F7,$8B,$5F,$F7,$D3,$5F,$F7,$83,$68
	.dc.b	$D2,$4F,$80,$93,$E0,$6B,$E0,$6B,$E0,$6B,$E0,$6B,$E0,$6C,$F8,$1B
	.dc.b	$3E,$09,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$59,$EC,$D8,$EC,$F4,$21
	.dc.b	$D0,$A7,$42,$9D,$0B,$1D,$1B,$1D,$1E,$44,$39,$14,$E4,$58,$E4,$D8
	.dc.b	$E4,$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D,$8D,$46,$08,$E0,$93
	.dc.b	$F5,$09,$3F,$25,$AF,$C0,$6B,$E9,$6B,$E4,$6B,$E1,$6C,$F8,$1B,$3E
	.dc.b	$09,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$59,$EC,$D8,$EC,$F4,$21,$D0
	.dc.b	$A7,$42,$9D,$0B,$1D,$1B,$1D,$1E,$44,$39,$14,$E4,$58,$E4,$D8,$E4
	.dc.b	$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D,$8D,$46,$08,$E0,$93,$F5
	.dc.b	$09,$3F,$25,$AF,$C0,$6B,$E9,$6B,$E4,$6B,$E1,$6C,$F8,$1B,$3E,$09
	.dc.b	$EC,$7E,$EF,$95,$EE,$7F,$5C,$8F,$F1,$84,$62,$BF,$C5,$91,$89,$D2
	.dc.b	$9D,$3A,$6B,$FE,$33,$6B,$FE,$38,$6B,$CE,$7A,$BC,$E7,$AC,$EA,$31
	.dc.b	$3A,$8C,$12,$76,$48,$D3,$67,$C8,$D7,$C8,$D7,$C8,$D7,$C8,$70,$8E
	.dc.b	$64,$70,$43,$9B,$96,$99,$3A,$B9,$69,$93,$B9,$C0,$86,$8A,$68,$C1
	.dc.b	$0C,$10,$F8,$E7,$52,$9C,$8A,$71,$29,$CB,$99,$CC,$F5,$7F,$D5,$04
	.dc.b	$AF,$C1,$CE,$A3,$15,$FE,$71,$8C,$4E,$AB,$93,$6B,$F1,$35,$FE,$91
	.dc.b	$3F,$D4,$87,$FA,$50,$FF,$4C,$9F,$E8,$A2,$9B,$E6,$FF,$54,$6F,$FE
	.dc.b	$9C,$E1,$FF,$4B,$4E,$79,$FF,$D4,$98,$7F,$A7,$27,$33,$C8,$95,$C6
	.dc.b	$74,$48,$F1,$9F,$EA,$27,$17,$E0,$8F,$F5,$05,$FF,$A3,$EF,$AD,$F3
	.dc.b	$9E,$31,$87,$67,$FA,$E0,$FF,$55,$57,$FA,$A1,$FF,$53,$35,$F8,$3C
	.dc.b	$FF,$E9,$FE,$2F,$F4,$B3,$7A,$FF,$D4,$13,$FE,$9C,$8F,$F5,$23,$FE
	.dc.b	$A1,$6B,$A3,$1D,$1F,$EA,$3A,$E6,$BA,$3F,$D4,$77,$FE,$AC,$E7,$FF
	.dc.b	$46,$5F,$F9,$D5,$FF,$4A,$F4,$7F,$A1,$89,$1F,$E7,$23,$5D,$1F,$E7
	.dc.b	$28,$CC,$7F,$9E,$8F,$F3,$C1,$75,$47,$FA,$79,$B1,$CD,$AF,$FA,$4A
	.dc.b	$39,$D7,$92,$72,$D8,$F4,$57,$86,$B9,$46,$35,$FF,$45,$47,$FA,$42
	.dc.b	$F8,$D7,$FA,$52,$73,$1F,$E8,$9D,$FF,$D1,$EF,$6D,$3F,$D2,$98,$F7
	.dc.b	$C3,$A9,$FE,$AA,$AD,$E3,$EB,$3C,$11,$F1,$30,$EB,$1E,$AC,$7C,$13
	.dc.b	$93,$E2,$DF,$57,$FD,$47,$7D,$1E,$C7,$A6,$BA,$E3,$EA,$C3,$AC,$F9
	.dc.b	$0F,$F6,$15,$F3,$26,$1F,$16,$3E,$95,$F9,$DE,$13,$FE,$BA,$7C,$73
	.dc.b	$E6,$BE,$D1,$8B,$FF,$6D,$BE,$98,$FA,$FC,$CE,$8F,$F5,$8C,$B0,$65
	.dc.b	$8A,$D3,$FE,$C9,$31,$5C,$27,$71,$FE,$DC,$9D,$D7,$E4,$E3,$97,$47
	.dc.b	$3F,$8E,$B1,$B6,$CF,$FB,$36,$91,$8C,$7C,$ED,$D1,$52,$B7,$7E,$1C
	.dc.b	$3B,$34,$EC,$59,$ED,$7C,$DE,$4D,$55,$D1,$F2,$57,$46,$D7,$46,$9D
	.dc.b	$1E,$D8,$7A,$1B,$87,$A2,$7A,$E1,$DF,$0E,$D7,$F5,$30,$EC,$DC,$3B
	.dc.b	$36,$FB,$1E,$07,$B3,$70,$ED,$7D,$1F,$F6,$44,$E8,$E4,$57,$27,$0C
	.dc.b	$3C,$84,$BF,$4D,$FC,$66,$BE,$A7,$F0,$1F,$2B,$FE,$0E,$C0,$D2,$47
	.dc.b	$F8,$22,$32,$D8,$CC,$FF,$86,$27,$2D,$8C,$BF,$24,$FA,$AE,$1F,$7B
	.dc.b	$5F,$C2,$35,$DC,$F7,$23,$D5,$AF,$59,$F0,$B6,$79,$46,$23,$BC,$F0
	.dc.b	$69,$CC,$87,$17,$4F,$91,$AF,$90,$D9,$C8,$E0,$60,$E0,$53,$81,$A2
	.dc.b	$1C,$48,$70,$32,$53,$27,$22,$C7,$B5,$D1,$0D,$15,$D3,$82,$38,$24
	.dc.b	$FC,$82,$4F,$C0,$8C,$B6,$33,$3E,$D9,$CB,$63,$2F,$A8,$EE,$E1,$F4
	.dc.b	$B5,$F3,$35,$DC,$F7,$23,$D5,$AF,$59,$F0,$B6,$79,$46,$23,$BC,$F0
	.dc.b	$69,$CC,$87,$17,$4F,$91,$AF,$90,$D9,$C8,$E0,$60,$E0,$53,$81,$A2
	.dc.b	$1C,$48,$70,$32,$53,$27,$22,$C7,$B5,$D1,$0D,$15,$D3,$82,$38,$24
	.dc.b	$FC,$82,$4F,$C0,$8C,$B6,$33,$3E,$D9,$CB,$63,$2F,$A8,$EE,$E1,$F4
	.dc.b	$B5,$F3,$35,$DC,$F7,$23,$D5,$AF,$59,$F0,$B6,$79,$46,$23,$BC,$F0
	.dc.b	$69,$CC,$87,$17,$4F,$91,$AF,$90,$FD,$E7,$4B,$DB,$FF,$55,$92,$7F
	.dc.b	$D5,$4D,$9F,$F5,$C3,$5F,$F5,$C3,$5F,$F5,$C3,$5F,$F6,$2B,$5F,$F6
	.dc.b	$13,$67,$C1,$3D,$89,$3D,$9A,$F6,$6B,$D9,$AF,$66,$BD,$9B,$1D,$A3
	.dc.b	$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8
	.dc.b	$AF,$27,$81,$0E,$05,$38,$15,$E0,$BA,$7F,$D6,$77,$FE,$F5,$6D,$7F
	.dc.b	$AB,$63,$1D,$3F,$E0,$E8,$C1,$2B,$FD,$69,$19,$6C,$FF,$AC,$9F,$F5
	.dc.b	$19,$FE,$96,$3F,$D6,$5A,$7F,$A6,$F9,$FF,$D4,$07,$FA,$55,$C7,$47
	.dc.b	$FA,$BA,$7F,$D3,$F8,$7F,$A9,$A3,$15,$C1,$FF,$53,$46,$67,$FD,$43
	.dc.b	$18,$78,$92,$78,$3F,$EA,$73,$FD,$85,$5E,$07,$89,$93,$FD,$80,$60
	.dc.b	$F4,$9E,$B3,$FD,$74,$21,$3E,$63,$FD,$57,$C9,$FE,$A4,$E6,$FF,$51
	.dc.b	$D7,$F9,$13,$A3,$FD,$56,$DE,$8F,$F5,$43,$FE,$09,$11,$A6,$C7,$FA
	.dc.b	$BF,$A3,$A9,$23,$95,$7F,$94,$2D,$3F,$E7,$CE,$39,$5F,$E8,$0C,$FF
	.dc.b	$D2,$8F,$FA,$04,$FF,$3B,$D7,$FA,$53,$A3,$FC,$ED,$5C,$27,$FD,$50
	.dc.b	$7F,$9F,$23,$FD,$01,$CD,$FE,$82,$E8,$FF,$2F,$E1,$FE,$7C,$3F,$D3
	.dc.b	$DB,$FF,$A3,$EF,$89,$FE,$86,$3C,$58,$7F,$A6,$6B,$FD,$03,$5E,$42
	.dc.b	$46,$67,$FD,$08,$F8,$EF,$FD,$19,$3D,$A7,$D4,$7F,$A7,$8B,$1E,$57
	.dc.b	$FD,$1C,$72,$3F,$CE,$AF,$FB,$49,$F6,$57,$FA,$81,$F3,$1E,$88,$E5
	.dc.b	$3E,$DA,$FF,$4F,$4F,$78,$F3,$BC,$8E,$EE,$DE,$F8,$7C,$83,$E6,$B8
	.dc.b	$BF,$05,$F6,$BF,$76,$3F,$4E,$7E,$6C,$7F,$A3,$3C,$17,$EB,$2F,$2B
	.dc.b	$FD,$47,$8B,$3F,$F7,$0D,$6A,$73,$18,$A7,$97,$8C,$3A,$47,$3D,$7B
	.dc.b	$E2,$C7,$94,$61,$B1,$DE,$B8,$DF,$38,$E1,$7C,$DF,$F5,$47,$67,$FA
	.dc.b	$D3,$0F,$F5,$97,$85,$67,$31,$FE,$B5,$9F,$F5,$E4,$E7,$2F,$F6,$1C
	.dc.b	$5C,$7F,$D1,$33,$6F,$FD,$5D,$35,$FF,$52,$C5,$E6,$FF,$5F,$DF,$2C
	.dc.b	$3C,$27,$72,$57,$88,$FF,$5F,$57,$1B,$E4,$D9,$DD,$7F,$B5,$A3,$15
	.dc.b	$C6,$31,$3A,$C3,$FD,$3F,$7C,$A3,$15,$E8,$6B,$E8,$6E,$1E,$8A,$E3
	.dc.b	$1E,$89,$ED,$18,$69,$FE,$CF,$C3,$C7,$87,$BE,$3D,$75,$D3,$C2,$FF
	.dc.b	$51,$11,$FF,$50,$36,$3F,$D3,$CD,$7F,$D2,$0D,$7F,$D7,$2D,$7F,$DB
	.dc.b	$CD,$7F,$DA,$AD,$C0,$D2,$4F,$80,$93,$E0,$6B,$E0,$6B,$E0,$6B,$E0
	.dc.b	$6B,$E0,$6C,$F8,$1B,$3E,$09,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$59
	.dc.b	$EC,$D8,$EC,$F4,$21,$D0,$A7,$42,$9D,$0B,$1D,$1B,$1D,$1E,$44,$39
	.dc.b	$14,$E4,$58,$E4,$D8,$E4,$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D
	.dc.b	$8D,$46,$08,$E0,$93,$F5,$09,$3F,$25,$AF,$C0,$6B,$E9,$6B,$E4,$6B
	.dc.b	$E1,$6C,$F8,$1B,$3E,$09,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$59,$EC
	.dc.b	$D8,$EC,$F4,$21,$D0,$A7,$42,$9D,$0B,$1D,$1B,$1D,$1E,$44,$39,$14
	.dc.b	$E4,$58,$E4,$D8,$E4,$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D,$8D
	.dc.b	$46,$08,$E0,$93,$F5,$09,$3F,$25,$AF,$C0,$6B,$E9,$6B,$E4,$6B,$E1
	.dc.b	$6C,$F8,$1B,$3E,$09,$EC,$7F,$8B,$E5,$E1,$7F,$6A,$81,$04,$6A,$BF
	.dc.b	$C6,$D1,$89,$E0,$EE,$74,$F5,$8C,$BD,$23,$2F,$53,$94,$F4,$8C,$C7
	.dc.b	$39,$E2,$E8,$E9,$3B,$8C,$CF,$09,$ED,$18,$24,$EC,$91,$C8,$87,$8E
	.dc.b	$79,$3D,$CC,$4F,$27,$67,$96,$79,$BA,$6B,$93,$67,$91,$CB,$4C,$9D
	.dc.b	$CE,$07,$33,$33,$E4,$24,$F8,$E3,$44,$38,$90,$EC,$64,$A6,$4F,$01
	.dc.b	$82,$18,$27,$5B,$9B,$E9,$B6,$FF,$06,$D9,$FF,$3D,$35,$FB,$CD,$7F
	.dc.b	$18,$D7,$F1,$ED,$4D,$7F,$D2,$67,$FA,$60,$FF,$49,$70,$FF,$90,$3A
	.dc.b	$7F,$D3,$B1,$FE,$96,$7F,$D5,$8E,$AB,$FD,$4C,$F1,$8C,$93,$A7,$FD
	.dc.b	$50,$F2,$8F,$F4,$EB,$6B,$B4,$FF,$A7,$67,$FD,$48,$69,$C3,$D5,$E6
	.dc.b	$61,$F4,$90,$FF,$62,$CF,$85,$FF,$58,$18,$8E,$2F,$69,$F0,$B6,$BF
	.dc.b	$6B,$5F,$EB,$08,$FD,$56,$98,$BE,$31,$FE,$A8,$E8,$FF,$58,$B4,$72
	.dc.b	$FF,$AD,$5F,$F5,$87,$4F,$FA,$BC,$9A,$7F,$AB,$E3,$FD,$5A,$70,$E9
	.dc.b	$FF,$58,$3C,$EB,$FD,$5E,$FF,$AA,$B3,$FF,$50,$74,$7F,$A5,$B8,$FF
	.dc.b	$D0,$EF,$F9,$F6,$7F,$D4,$27,$FA,$1F,$4F,$F4,$D1,$FE,$8A,$7F,$D2
	.dc.b	$31,$C5,$D9,$8B,$FF,$42,$3F,$E9,$B1,$1F,$E9,$B7,$57,$E2,$9D,$D7
	.dc.b	$FA,$4E,$77,$D3,$FE,$93,$DF,$FD,$63,$8F,$FA,$BE,$B5,$8F,$33,$D1
	.dc.b	$7C,$2B,$FD,$2F,$7F,$E9,$A8,$C1,$FE,$A0,$BF,$F5,$B4,$75,$8F,$35
	.dc.b	$73,$8E,$0D,$AF,$0C,$FC,$3A,$F8,$F5,$AC,$3D,$31,$D2,$BC,$27,$0A
	.dc.b	$F7,$9F,$EC,$0A,$E1,$8F,$FA,$6A,$3C,$B8,$FF,$A8,$7A,$7F,$D4,$F3
	.dc.b	$FC,$F7,$A7,$FD,$36,$DE,$8F,$F4,$C3,$67,$77,$FE,$A6,$6B,$FE,$9E
	.dc.b	$6D,$72,$8D,$4F,$F4,$A3,$F9,$67,$47,$08,$FE,$96,$DE,$63,$49,$3C
	.dc.b	$67,$FC,$FC,$F2,$30,$FF,$A2,$4F,$EA,$33,$CA,$33,$1D,$9B,$19,$8D
	.dc.b	$CF,$66,$CE,$A3,$0F,$78,$C9,$1C,$11,$F2,$90,$F2,$39,$69,$93,$C8
	.dc.b	$73,$31,$3E,$58,$E1,$1A,$6C,$66,$37,$3E,$66,$CF,$A4,$8F,$53,$C2
	.dc.b	$43,$A1,$C0,$D1,$83,$53,$DD,$B3,$DD,$CF,$7F,$F7,$02,$57,$F7,$96
	.dc.b	$C7,$F8,$35,$AF,$F6,$D6,$BF,$E0,$76,$BF,$E0,$56,$BF,$E1,$06,$D1
	.dc.b	$A4,$9F,$01,$27,$C0,$D7,$C0,$D7,$C0,$D7,$C0,$D7,$C0,$D9,$F0,$36
	.dc.b	$7C,$13,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$B3,$D9,$B1,$D9,$E8,$43
	.dc.b	$A1,$4E,$85,$3A,$16,$3A,$36,$3A,$3C,$88,$72,$29,$C8,$B1,$C9,$B1
	.dc.b	$C9,$E0,$43,$81,$63,$83,$63,$83,$A2,$46,$9B,$1A,$8C,$11,$C1,$27
	.dc.b	$EA,$12,$7E,$4B,$5F,$80,$D7,$D2,$D7,$C8,$D7,$C2,$D9,$F0,$36,$7C
	.dc.b	$13,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$B3,$D9,$B1,$D9,$E8,$43,$A1
	.dc.b	$4E,$85,$3A,$16,$3A,$36,$3A,$3C,$88,$72,$29,$C8,$B1,$C9,$B1,$C9
	.dc.b	$E0,$43,$81,$63,$83,$63,$83,$A2,$46,$9B,$1A,$8C,$11,$C1,$27,$EA
	.dc.b	$12,$7E,$4B,$5F,$80,$D7,$D2,$D7,$C8,$D7,$C2,$D9,$F0,$36,$7C,$13
	.dc.b	$D8,$FF,$55,$78,$5F,$E9,$22,$4F,$FA,$6D,$B1,$A8,$C1,$1C,$10,$FF
	.dc.b	$49,$47,$06,$C6,$DA,$6C,$FF,$58,$46,$DB,$1B,$69,$B3,$FD,$65,$1B
	.dc.b	$6C,$6D,$A6,$CF,$F6,$0C,$6D,$B1,$B6,$9B,$A4,$ED,$B3,$B6,$CE,$E7
	.dc.b	$44,$9D,$36,$35,$18,$23,$E5,$E7,$78,$5F,$E9,$42,$4F,$FA,$51,$B3
	.dc.b	$FE,$94,$6B,$FE,$94,$7F,$D9,$AF,$F9,$D7,$5F,$F3,$53,$FE,$74,$6B
	.dc.b	$FE,$8D,$E7,$FF,$36,$BB,$3F,$D1,$C3,$A3,$FD,$11,$7F,$E9,$08,$D4
	.dc.b	$7F,$9F,$DE,$47,$53,$8B,$C1,$E5,$3F,$E8,$97,$FD,$08,$6E,$39,$3F
	.dc.b	$E9,$13,$FD,$08,$70,$2B,$CC,$FF,$43,$3F,$EF,$BA,$F5,$AE,$1F,$F4
	.dc.b	$6B,$FE,$7B,$E5,$FF,$3E,$F3,$FF,$AF,$FA,$3F,$D1,$2F,$FA,$11,$FF
	.dc.b	$80,$2D,$9F,$F4,$4F,$42,$76,$D8,$FF,$44,$1F,$E8,$AE,$7F,$F4,$3B
	.dc.b	$FE,$54,$E7,$33,$A3,$FC,$C8,$FF,$9A,$3A,$3F,$CD,$5A,$7F,$9D,$1B
	.dc.b	$5F,$E6,$39,$FF,$38,$1F,$E9,$13,$FC,$C5,$A7,$F9,$86,$F9,$DF,$18
	.dc.b	$FF,$34,$CF,$56,$C6,$63,$99,$E5,$3C,$97,$E0,$7B,$47,$FA,$5E,$BF
	.dc.b	$D0,$A7,$6B,$F2,$47,$FA,$71,$FF,$46,$18,$3F,$CF,$B7,$E5,$BE,$97
	.dc.b	$DC,$97,$FE,$A1,$7F,$D3,$B8,$7F,$A0,$9F,$0D,$F4,$69,$FE,$89,$3F
	.dc.b	$D4,$F3,$E4,$AE,$AF,$9A,$B9,$4F,$69,$C9,$FE,$88,$7E,$49,$FE,$98
	.dc.b	$86,$FE,$7C,$3E,$BE,$3F,$E8,$B8,$B8,$FD,$AF,$39,$8B,$E0,$78,$E4
	.dc.b	$C3,$83,$6F,$85,$6F,$0E,$73,$FE,$8D,$6E,$1D,$30,$E7,$87,$21,$87
	.dc.b	$28,$CB,$F8,$CB,$FF,$37,$CE,$9F,$C7,$5E,$9F,$F4,$54,$65,$FF,$65
	.dc.b	$C6,$5E,$33,$CF,$3F,$87,$3F,$EA,$DE,$5F,$F4,$65,$F1,$9C,$DF,$B4
	.dc.b	$EB,$7C,$6F,$A5,$F0,$8C,$47,$3E,$7F,$05,$F2,$C3,$C9,$7B,$C3,$D3
	.dc.b	$7B,$3F,$3C,$74,$BE,$2D,$32,$6E,$3C,$87,$B9,$C5,$7A,$5E,$11,$AB
	.dc.b	$F1,$61,$F2,$F0,$F1,$4F,$13,$9B,$89,$DC,$7C,$03,$CE,$6B,$0F,$2D
	.dc.b	$F8,$2F,$E0,$CE,$63,$CE,$D3,$C4,$F9,$30,$F3,$3C,$63,$AC,$7C,$6A
	.dc.b	$EB,$1C,$63,$6D,$AE,$F1,$BA,$EE,$D9,$EF,$1F,$77,$C3,$FF,$6B,$11
	.dc.b	$FF,$65,$B6,$3F,$DA,$2D,$8F,$C8,$43,$14,$E6,$3E,$53,$5F,$E4,$31
	.dc.b	$A9,$EA,$48,$EA,$D3,$AC,$6E,$32,$48,$CC,$F5,$8F,$13,$67,$84,$E8
	.dc.b	$E9,$3E,$46,$BE,$49,$E0,$48,$CC,$74,$6C,$F4,$8E,$47,$8C,$B1,$DD
	.dc.b	$C1,$D5,$E3,$3D,$63,$27,$52,$CF,$57,$81,$88,$E1,$3D,$49,$1E,$02
	.dc.b	$1C,$1C,$B4,$C9,$DD,$CB,$4C,$CF,$85,$C1,$C4,$86,$08,$EC,$93,$E6
	.dc.b	$8E,$6D,$78,$C6,$A7,$E1,$92,$3D,$CD,$3C,$B1,$B8,$C9,$23,$33,$D6
	.dc.b	$3C,$6D,$9E,$13,$A3,$A4,$F9,$1A,$F9,$27,$81,$23,$31,$D1,$B3,$D2
	.dc.b	$39,$1E,$32,$C7,$77,$07,$57,$8C,$F5,$8C,$9D,$4B,$3D,$5E,$06,$23
	.dc.b	$84,$F5,$24,$78,$08,$70,$72,$D3,$27,$77,$2D,$33,$3E,$17,$07,$12
	.dc.b	$18,$23,$B2,$4F,$9A,$39,$B5,$E3,$1A,$9F,$86,$48,$F7,$34,$F2,$C6
	.dc.b	$E3,$24,$8C,$CF,$58,$F1,$B6,$78,$4E,$8E,$93,$E4,$6B,$E4,$9E,$04
	.dc.b	$8C,$C7,$46,$CF,$C5,$E6,$7B,$3F,$E0,$0A,$FE,$47,$18,$AF,$E4,$B1
	.dc.b	$FE,$40,$8C,$38,$6B,$C2,$79,$3F,$DE,$E3,$2D,$8C,$BC,$DA,$A3,$6E
	.dc.b	$67,$C2,$E6,$77,$3A,$8C,$4E,$A3,$04,$9D,$92,$34,$D9,$F5,$35,$F5
	.dc.b	$46,$88,$E9,$AE,$8C,$10,$C1,$0F,$80,$F0,$69,$B2,$9B,$3E,$01,$92
	.dc.b	$99,$3D,$E6,$4A,$66,$7E,$04,$77,$8E,$27,$62,$18,$21,$C8,$CE,$AF
	.dc.b	$83,$79,$1F,$62,$30,$D8,$CD,$7E,$1A,$72,$D8,$CD,$7F,$6E,$9C,$B6
	.dc.b	$32,$FF,$90,$63,$2F,$E0,$33,$FF,$50,$F0,$FF,$A8,$23,$FD,$6B,$D1
	.dc.b	$FE,$BC,$7F,$D5,$E7,$FA,$D8,$87,$FA,$BC,$FF,$55,$1F,$EB,$87,$FD
	.dc.b	$6D,$3D,$30,$FF,$60,$4F,$23,$81,$D8,$FF,$5D,$4F,$79,$E1,$1F,$EB
	.dc.b	$93,$AF,$47,$FA,$FA,$FC,$D1,$CD,$B3,$C2,$7F,$D7,$D5,$E2,$AF,$21
	.dc.b	$FE,$C2,$71,$1C,$EF,$CB,$5C,$67,$27,$8E,$B8,$9E,$27,$15,$FE,$CA
	.dc.b	$8F,$F6,$45,$7F,$B9,$39,$FF,$D8,$D1,$FE,$BE,$3F,$D9,$47,$FA,$FF
	.dc.b	$A3,$FD,$97,$CF,$FE,$C3,$8F,$F1,$2C,$FF,$B2,$5B,$19,$77,$D1,$FE
	.dc.b	$CA,$3F,$D8,$BD,$1F,$EC,$8E,$6F,$F4,$D6,$3F,$E9,$3A,$FF,$5D,$1F
	.dc.b	$EB,$88,$CB,$FE,$9A,$D3,$FC,$E9,$D1,$FE,$9E,$DB,$84,$7F,$AD,$67
	.dc.b	$FD,$50,$7F,$AB,$AF,$3C,$DF,$EB,$78,$FF,$55,$DF,$FB,$9D,$EF,$5C
	.dc.b	$5F,$F5,$55,$F0,$76,$D7,$FD,$77,$8F,$FA,$C3,$1F,$2D,$7F,$B2,$6F
	.dc.b	$A3,$FE,$D1,$3A,$47,$FA,$E6,$FF,$DA,$B8,$78,$F0,$F3,$1E,$D7,$FD
	.dc.b	$29,$7E,$D8,$FF,$5C,$63,$E1,$BE,$F5,$FE,$C7,$9F,$39,$F1,$EB,$FD
	.dc.b	$81,$87,$8F,$0E,$11,$CE,$3E,$4C,$78,$A7,$FD,$9E,$72,$8C,$BC,$E3
	.dc.b	$C1,$8F,$BE,$BA,$DF,$9A,$7F,$D7,$BD,$8E,$E7,$77,$FA,$CF,$8F,$FD
	.dc.b	$7D,$AE,$A7,$2F,$FB,$0B,$1F,$F6,$36,$7F,$EA,$DC,$39,$61,$C6,$73
	.dc.b	$A7,$FB,$17,$3F,$F6,$26,$1F,$E8,$C3,$FC,$EB,$AF,$FA,$D9,$A7,$FB
	.dc.b	$58,$A3,$1F,$EA,$31,$B7,$FD,$2F,$1A,$72,$F1,$72,$43,$FD,$42,$EC
	.dc.b	$8E,$4F,$F5,$53,$87,$66,$63,$C0,$E0,$87,$03,$51,$E1,$73,$1C,$1D
	.dc.b	$18,$8D,$38,$24,$6C,$91,$A6,$BF,$0C,$F3,$47,$07,$44,$34,$F9,$1A
	.dc.b	$F8,$5A,$F1,$6C,$76,$8F,$97,$18,$3A,$38,$23,$E2,$8F,$77,$43,$E3
	.dc.b	$FD,$AD,$9E,$ED,$AF,$F0,$5B,$63,$FD,$57,$3C,$08,$F0,$6C,$F0,$6C
	.dc.b	$70,$34,$48,$D3,$4D,$38,$23,$82,$3F,$EE,$D8,$EA,$D8,$E4,$D3,$89
	.dc.b	$5E,$2F,$FB,$2A,$38,$B6,$38,$B4,$E2,$57,$8B,$FE,$D7,$8E,$2D,$8E
	.dc.b	$2D,$38,$95,$E3,$3F,$2E,$3C,$84,$3B,$14,$E0,$57,$83,$A2,$1A,$2B
	.dc.b	$A7,$04,$70,$49,$F4,$B6,$7D,$2D,$9F,$4B,$63,$D3,$3C,$08,$F0,$6C
	.dc.b	$F0,$6C,$70,$34,$48,$D3,$4D,$38,$23,$82,$3F,$0E,$3A,$B6,$39,$34
	.dc.b	$E2,$57,$8B,$F1,$A3,$8B,$63,$8B,$4E,$25,$78,$BF,$2E,$38,$B6,$38
	.dc.b	$B4,$E2,$57,$8C,$FC,$B8,$F2,$10,$EC,$53,$81,$5E,$0E,$88,$68,$AE
	.dc.b	$9C,$11,$C1,$27,$D2,$D9,$F4,$B6,$7D,$2D,$8F,$4C,$F0,$23,$C1,$B3
	.dc.b	$C1,$B1,$C0,$D1,$23,$4D,$34,$E0,$8E,$08,$FC,$38,$EA,$D8,$E4,$D3
	.dc.b	$89,$5E,$2F,$C6,$8E,$2D,$8E,$2D,$3E,$CF,$43,$DB,$FF,$50,$92,$7F
	.dc.b	$D4,$0D,$9F,$F5,$A3,$5F,$F5,$DB,$5F,$F5,$A3,$5F,$F5,$E3,$5F,$F5
	.dc.b	$93,$67,$C1,$3D,$89,$3D,$9A,$F6,$6B,$D9,$AF,$66,$BD,$9B,$1D,$A3
	.dc.b	$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8
	.dc.b	$AF,$27,$81,$0E,$05,$38,$15,$E0,$BC,$67,$8C,$F5,$FF,$DC,$71,$86
	.dc.b	$C6,$6F,$FD,$E3,$39,$6C,$66,$BF,$C4,$07,$F8,$86,$74,$E0,$F0,$46
	.dc.b	$DD,$C6,$9F,$F5,$71,$FE,$8A,$3F,$D4,$07,$FA,$AB,$8F,$FD,$32,$E4
	.dc.b	$FF,$4E,$57,$FA,$51,$C6,$DF,$E9,$57,$FD,$34,$43,$FD,$28,$7F,$AA
	.dc.b	$88,$75,$8E,$5D,$3F,$E9,$F8,$FF,$56,$1E,$23,$FD,$66,$78,$0C,$95
	.dc.b	$FF,$55,$BD,$CE,$14,$9F,$F5,$07,$3B,$9F,$FD,$2D,$5F,$EA,$1E,$8F
	.dc.b	$F5,$1F,$43,$A7,$E5,$C7,$FA,$7E,$6D,$FF,$A7,$DF,$F2,$74,$7F,$AA
	.dc.b	$9F,$F5,$1D,$F4,$BE,$51,$FE,$AB,$3F,$CE,$3C,$7F,$E8,$57,$FC,$E0
	.dc.b	$7F,$A0,$0F,$F4,$21,$FE,$59,$5C,$BF,$E7,$BE,$5F,$F3,$94,$F0,$DB
	.dc.b	$FC,$F2,$7F,$A4,$27,$FD,$27,$3F,$E8,$FA,$5F,$F9,$EF,$5F,$F3,$EE
	.dc.b	$BF,$E8,$66,$E3,$E0,$3F,$D3,$F1,$97,$C9,$8F,$FA,$35,$FF,$54,$E1
	.dc.b	$F8,$EC,$3A,$D6,$EB,$C7,$39,$C3,$C1,$7C,$A3,$FD,$35,$7F,$E9,$C8
	.dc.b	$FF,$5C,$BC,$9F,$13,$F0,$30,$FF,$4C,$47,$FA,$5E,$7A,$61,$DD,$F8
	.dc.b	$78,$7C,$1A,$C5,$7F,$AE,$8F,$39,$63,$FD,$6D,$3C,$A3,$BE,$3F,$35
	.dc.b	$FA,$0F,$CA,$78,$47,$EA,$F8,$1E,$A7,$E2,$E5,$6A,$D4,$E6,$31,$3F
	.dc.b	$B0,$EB,$C5,$6B,$2F,$F4,$B5,$6B,$8B,$F8,$1E,$1A,$C3,$F6,$ED,$8F
	.dc.b	$E0,$F8,$5C,$7C,$78,$7F,$AB,$FA,$38,$74,$FF,$AB,$23,$FD,$65,$E4
	.dc.b	$63,$AF,$FD,$67,$9F,$FA,$C5,$BA,$7F,$AB,$E1,$7F,$EB,$DA,$E1,$84
	.dc.b	$C3,$B1,$D0,$B5,$CE,$34,$F5,$9B,$7D,$6F,$C9,$39,$BE,$33,$9C,$3C
	.dc.b	$7A,$FF,$AE,$6F,$C5,$5A,$8C,$4E,$AF,$CE,$DA,$DC,$78,$EB,$83,$AC
	.dc.b	$7E,$16,$3E,$4C,$39,$C5,$9E,$73,$6F,$CF,$8F,$96,$31,$8F,$72,$38
	.dc.b	$BC,$B7,$1F,$0D,$F9,$2F,$DD,$5A,$9F,$2E,$0E,$B7,$91,$FE,$A2,$24
	.dc.b	$FF,$A8,$9B,$3F,$EA,$26,$C7,$F3,$D9,$E2,$EC,$93,$B6,$9B,$9E,$31
	.dc.b	$FD,$22,$78,$BB,$24,$EC,$E2,$EE,$78,$BF,$EE,$F9,$E3,$3B,$8E,$26
	.dc.b	$DE,$2E,$E7,$8B,$F0,$5B,$3C,$A3,$81,$C8,$E1,$3C,$9E,$0D,$8E,$11
	.dc.b	$A3,$81,$A2,$BA,$71,$1E,$72,$3E,$06,$BC,$DA,$F0,$34,$43,$45,$34
	.dc.b	$60,$86,$09,$1E,$32,$4F,$C2,$24,$FB,$5B,$3E,$A6,$C7,$92,$78,$BB
	.dc.b	$24,$ED,$A6,$E7,$8C,$79,$E7,$8B,$B2,$4E,$CE,$2E,$E7,$8B,$EC,$9E
	.dc.b	$33,$B8,$E2,$6D,$E2,$EE,$78,$BF,$09,$B3,$CA,$38,$1C,$8E,$13,$C9
	.dc.b	$E0,$D8,$E1,$1A,$38,$1A,$2B,$A7,$11,$E7,$23,$E0,$6B,$CD,$AF,$03
	.dc.b	$44,$34,$53,$46,$08,$60,$91,$E3,$24,$FC,$22,$4F,$B5,$B3,$EA,$6C
	.dc.b	$79,$27,$8B,$B2,$4E,$DA,$6E,$78,$C7,$9E,$78,$BB,$24,$EC,$E2,$EE
	.dc.b	$78,$BE,$C9,$E3,$3B,$8E,$26,$DE,$2E,$E7,$8B,$F0,$9B,$3C,$A3,$81
	.dc.b	$C8,$E1,$3F,$81,$EA,$7B,$3F,$DC,$E7,$FC,$BD,$18,$9F,$F3,$04,$04
	.dc.b	$65,$CB,$5E,$33,$CD,$FF,$35,$C6,$5B,$19,$7A,$46,$5B,$19,$7B,$B6
	.dc.b	$35,$18,$9F,$2B,$4F,$1C,$EC,$91,$A9,$E2,$48,$CC,$F0,$9D,$46,$23
	.dc.b	$D7,$18,$23,$B2,$1A,$29,$EE,$76,$E4,$86,$4D,$9F,$01,$DB,$93,$66
	.dc.b	$4D,$9E,$F3,$25,$38,$10,$E0,$6A,$7E,$04,$76,$8C,$9B,$39,$18,$3F
	.dc.b	$01,$F0,$2F,$2B,$EA,$36,$7E,$8B,$67,$F0,$ED,$7F,$D5,$0D,$7F,$D5
	.dc.b	$AD,$7A,$CF,$43,$F0,$0F,$FA,$42,$7F,$D3,$7C,$3F,$EB,$1E,$5F,$F5
	.dc.b	$4F,$67,$FA,$74,$FF,$4B,$D7,$FA,$90,$FF,$51,$1F,$E9,$CD,$BF,$D5
	.dc.b	$4E,$4F,$F5,$0B,$87,$B1,$B3,$FD,$62,$48,$ED,$5F,$A2,$9E,$13,$FE
	.dc.b	$A7,$8C,$3B,$7C,$D7,$E8,$8F,$F5,$A9,$E0,$34,$62,$FD,$73,$C2,$3C
	.dc.b	$57,$E0,$7F,$D4,$03,$AB,$FD,$5B,$D9,$FE,$AA,$EB,$FF,$58,$CF,$FA
	.dc.b	$A0,$FD,$BC,$FF,$AB,$DE,$2F,$06,$F6,$7F,$AE,$5F,$F4,$ED,$7F,$A7
	.dc.b	$FB,$3F,$D1,$FA,$7F,$A4,$A7,$FC,$F2,$7F,$AC,$3A,$FF,$CF,$8F,$FA
	.dc.b	$2C,$C9,$47,$67,$FA,$4E,$7F,$D4,$07,$FA,$22,$37,$1F,$E9,$97,$83
	.dc.b	$FE,$97,$7F,$D6,$9D,$3F,$E9,$43,$FD,$1D,$3C,$9F,$36,$FF,$E9,$69
	.dc.b	$FF,$57,$9F,$E9,$2D,$FF,$D3,$F7,$D4,$FF,$63,$5F,$FA,$76,$7F,$D3
	.dc.b	$67,$9B,$1F,$7E,$1D,$E3,$CB,$8F,$BA,$FC,$97,$C4,$F5,$4F,$1C,$7C
	.dc.b	$53,$AB,$F3,$9F,$EB,$6B,$E4,$7F,$A9,$6B,$CF,$7E,$B7,$E6,$63,$CB
	.dc.b	$1F,$1C,$7F,$AA,$A3,$FD,$3B,$7D,$23,$07,$FA,$A1,$FF,$51,$3F,$CE
	.dc.b	$7A,$FF,$D4,$39,$7F,$37,$3F,$9C,$47,$FA,$89,$C3,$C1,$CC,$70,$8E
	.dc.b	$5D,$7F,$EA,$46,$F6,$7F,$AD,$EB,$73,$97,$AF,$57,$FA,$51,$BD,$9F
	.dc.b	$EA,$66,$F5,$F1,$C3,$FD,$0F,$1E,$27,$FD,$18,$DA,$E5,$1A,$3F,$AA
	.dc.b	$38,$76,$62,$BB,$3B,$9E,$E7,$38,$CC,$70,$8C,$BC,$6B,$C8,$E4,$E4
	.dc.b	$43,$04,$AF,$11,$2B,$AB,$67,$9B,$6B,$C5,$18,$9D,$91,$C1,$23,$64
	.dc.b	$AF,$33,$4F,$49,$1F,$35,$77,$76,$48,$D3,$E9,$AF,$04,$73,$8E,$31
	.dc.b	$87,$81,$8A,$F8,$AF,$FA,$B7,$C5,$BD,$66,$92,$FE,$E4,$ED,$B3,$B6
	.dc.b	$CE,$E7,$FC,$0D,$39,$6C,$66,$7F,$D6,$8D,$8F,$F5,$73,$5F,$F5,$43
	.dc.b	$5F,$01,$CA,$3B,$91,$C1,$1E,$A6,$4A,$E4,$EE,$E5,$A6,$4F,$0B,$96
	.dc.b	$99,$3C,$6F,$42,$3C,$9A,$F0,$6B,$C0,$D1,$0D,$14,$D1,$82,$18,$24
	.dc.b	$FC,$D8,$EA,$D8,$E4,$D7,$8B,$5E,$33,$F3,$63,$04,$70,$49,$F8,$53
	.dc.b	$C1,$B3,$B6,$CE,$E7,$E3,$4E,$5B,$19,$9F,$3B,$63,$CE,$D7,$CE,$D7
	.dc.b	$C0,$72,$8E,$E4,$70,$47,$A9,$92,$B9,$3B,$B9,$69,$93,$C2,$E5,$A6
	.dc.b	$4F,$1B,$D0,$8F,$26,$BC,$1A,$F0,$34,$43,$45,$34,$60,$86,$09,$3F
	.dc.b	$36,$3A,$B6,$39,$35,$E2,$D7,$8C,$FC,$D8,$C1,$1C,$12,$7E,$14,$F0
	.dc.b	$6C,$ED,$B3,$B9,$F8,$D3,$96,$C6,$67,$CE,$D8,$F3,$B5,$F3,$B5,$F0
	.dc.b	$1C,$A3,$B9,$1C,$11,$EA,$64,$AE,$4E,$EE,$5A,$64,$F0,$B9,$69,$EC
	.dc.b	$F2,$7F,$D3,$54,$69,$27,$FD,$33,$19,$6C,$66,$7F,$D6,$33,$96,$C6
	.dc.b	$5F,$EF,$A2,$34,$E1,$AF,$F8,$09,$AF,$FA,$ED,$AF,$18,$E8,$ED,$AE
	.dc.b	$DA,$EE,$7C,$E4,$7C,$AD,$7C,$51,$C2,$7B,$91,$C1,$27,$84,$F8,$5A
	.dc.b	$F8,$5A,$F8,$4E,$E4,$3B,$BA,$9E,$CD,$9E,$D3,$B3,$B1,$4E,$C5,$3B
	.dc.b	$1F,$2F,$C9,$FB,$64,$9F,$B2,$D8,$FA,$ED,$78,$47,$FB,$59,$FF,$47
	.dc.b	$F4,$7F,$A6,$1B,$D3,$FE,$8A,$8F,$F4,$04,$7F,$A5,$F8,$7F,$D0,$AF
	.dc.b	$FA,$1E,$BF,$D1,$8F,$FA,$7A,$79,$3F,$E9,$33,$F0,$7C,$7F,$E9,$62
	.dc.b	$3F,$E9,$03,$85,$F7,$3F,$D2,$13,$CA,$76,$F8,$A7,$CA,$74,$39,$1F
	.dc.b	$E9,$B7,$4A,$3F,$D5,$FD,$1F,$E8,$F8,$FF,$4C,$74,$7F,$A3,$F9,$5D
	.dc.b	$1F,$E9,$99,$FF,$48,$9F,$E9,$4E,$9F,$F7,$DD,$FF,$A6,$63,$8C,$62
	.dc.b	$F9,$57,$1B,$E7,$5C,$4E,$67,$FA,$31,$FF,$35,$BF,$E7,$0E,$7F,$F3
	.dc.b	$A6,$BF,$E6,$C8,$FF,$35,$E5,$FE,$7D,$9F,$F4,$6C,$7F,$A2,$CF,$F4
	.dc.b	$99,$A7,$FC,$E4,$FF,$A6,$3A,$3F,$CF,$9D,$3F,$E8,$18,$FF,$44,$DF
	.dc.b	$57,$C0,$71,$3F,$CF,$E7,$FA,$4A,$3F,$D4,$D5,$E4,$8F,$F3,$E1,$B3
	.dc.b	$FD,$1C,$78,$EF,$CC,$F2,$7F,$D2,$73,$CA,$38,$3F,$EA,$1B,$FF,$48
	.dc.b	$DF,$8E,$3C,$D3,$D2,$BF,$D2,$4F,$C7,$DF,$FD,$37,$7E,$C2,$5F,$FA
	.dc.b	$B1,$E7,$3F,$10,$F6,$9F,$0E,$34,$77,$8E,$31,$EC,$3C,$18,$7F,$AB
	.dc.b	$FA,$7F,$D2,$18,$21,$29,$74,$3E,$75,$97,$D9,$9E,$37,$C1,$DC,$EA
	.dc.b	$FE,$5C,$69,$E9,$7C,$EF,$93,$FE,$9F,$C3,$C5,$3A,$BE,$51,$8E,$97
	.dc.b	$4B,$A7,$39,$FA,$3C,$46,$AF,$27,$39,$FF,$A6,$32,$EB,$5A,$9C,$E5
	.dc.b	$D3,$1D,$CE,$75,$FC,$B4,$E5,$FC,$C4,$67,$0F,$C3,$CE,$5F,$C4,$C6
	.dc.b	$70,$F4,$67,$E7,$BE,$39,$79,$5E,$B1,$9B,$E3,$39,$7A,$C6,$6B,$D3
	.dc.b	$39,$7A,$E3,$E8,$8C,$67,$FE,$B9,$9D,$E7,$EE,$BE,$53,$F1,$EB,$78
	.dc.b	$7C,$58,$C5,$6F,$0E,$8D,$C7,$D2,$DA,$F3,$B7,$1E,$F3,$EC,$AE,$2D
	.dc.b	$38,$96,$B8,$B6,$78,$C7,$C0,$9E,$2D,$9E,$2D,$38,$96,$78,$9F,$12
	.dc.b	$36,$D8,$DB,$4D,$DF,$C7,$3E,$0C,$FE,$DC,$FF,$58,$F0,$7E,$81,$FF
	.dc.b	$56,$B6,$BF,$3C,$7E,$A1,$C4,$7F,$81,$1B,$1F,$AC,$0E,$1E,$11,$D8
	.dc.b	$87,$62,$BD,$8E,$2E,$0E,$C5,$7B,$47,$57,$71,$CC,$8F,$38,$D1,$23
	.dc.b	$94,$79,$1D,$9D,$8A,$F6,$76,$F6,$3A,$38,$8E,$8F,$08,$E8,$69,$C3
	.dc.b	$C5,$EC,$E4,$8E,$08,$F8,$5E,$31,$DA,$36,$64,$A7,$82,$38,$B4,$D3
	.dc.b	$87,$C4,$60,$86,$C8,$68,$B1,$E1,$8E,$4D,$8E,$26,$C8,$E5,$E3,$1F
	.dc.b	$1C,$87,$B0,$AF,$A8,$E2,$E0,$EE,$57,$BC,$75,$77,$1C,$C8,$F3,$8D
	.dc.b	$12,$39,$47,$91,$D9,$D8,$AF,$67,$6F,$63,$A3,$88,$E8,$F0,$8E,$86
	.dc.b	$9C,$3C,$5E,$CE,$48,$E0,$8F,$85,$E3,$1D,$A3,$66,$4A,$78,$23,$8B
	.dc.b	$4D,$38,$7C,$46,$08,$6C,$86,$8B,$1E,$18,$E4,$D8,$E2,$6C,$8E,$5E
	.dc.b	$31,$F1,$C8,$7B,$0A,$FA,$8E,$2E,$0E,$E5,$7B,$C7,$57,$71,$CC,$8F
	.dc.b	$38,$D1,$23,$94,$79,$1D,$9D,$8A,$F6,$76,$F6,$3A,$38,$8E,$8F,$08
	.dc.b	$E8,$7F,$1B,$EA,$78,$BF,$E0,$79,$FF,$54,$36,$7F,$D5,$0D,$7F,$D5
	.dc.b	$0D,$7F,$D6,$0D,$9E,$73,$FE,$1A,$39,$96,$79,$B5,$E6,$D7,$9B,$63
	.dc.b	$9C,$79,$E7,$B1,$27,$4D,$8D,$46,$08,$E0,$8F,$91,$A7,$90,$A7,$90
	.dc.b	$A7,$90,$B3,$E4,$8E,$44,$8E,$4D,$79,$35,$E4,$D8,$E5,$1C,$08,$70
	.dc.b	$29,$C0,$AF,$07,$44,$34,$57,$4E,$08,$E0,$93,$F6,$FC,$07,$8B,$FD
	.dc.b	$7E,$BF,$CA,$71,$8B,$FF,$77,$C6,$27,$4A,$74,$E9,$AF,$FB,$D9,$AF
	.dc.b	$F7,$36,$BC,$E7,$FD,$8C,$FF,$AE,$CF,$F6,$13,$FE,$C0,$7F,$D6,$1D
	.dc.b	$7F,$EB,$83,$FD,$74,$FF,$B1,$47,$5F,$FA,$ED,$FF,$5B,$F3,$7F,$B2
	.dc.b	$9F,$F6,$09,$FE,$BB,$8F,$F5,$CE,$BF,$EC,$0A,$E9,$39,$6B,$E2,$3A
	.dc.b	$BE,$07,$99,$99,$F0,$1D,$1F,$0D,$FA,$6F,$C5,$1E,$A3,$D1,$3C,$2B
	.dc.b	$BC,$7F,$A0,$AF,$FD,$97,$D3,$FE,$C0,$8F,$F5,$F7,$47,$FB,$01,$FF
	.dc.b	$62,$9F,$EC,$0E,$BF,$F6,$41,$27,$84,$6A,$78,$4F,$FB,$54,$FF,$6B
	.dc.b	$6F,$FE,$A7,$CB,$FD,$30,$4A,$FF,$4F,$47,$FA,$A0,$FF,$57,$3F,$E9
	.dc.b	$F3,$3C,$7F,$E9,$EC,$7F,$D6,$B1,$FE,$AB,$3F,$D7,$1C,$3F,$EC,$8A
	.dc.b	$E5,$5F,$EA,$AA,$DC,$78,$6B,$FD,$89,$8F,$FB,$62,$3C,$0D,$AE,$D7
	.dc.b	$CD,$E4,$FF,$8C,$1C,$D7,$FB,$48,$FF,$58,$57,$2D,$7F,$D7,$B8,$FF
	.dc.b	$AD,$A7,$35,$C1,$FF,$5A,$9F,$EC,$A7,$AB,$CF,$1F,$F5,$2B,$FE,$C0
	.dc.b	$BF,$5D,$7C,$79,$F5,$D7,$0A,$E6,$7C,$78,$CC,$EF,$0F,$53,$67,$FD
	.dc.b	$B3,$8F,$BE,$3C,$D5,$E0,$C3,$D5,$5F,$EE,$9C,$15,$FE,$DD,$BF,$F5
	.dc.b	$7C,$FF,$97,$A7,$2F,$C2,$8F,$F3,$59,$31,$FF,$35,$B6,$3F,$CA,$8D
	.dc.b	$8F,$F3,$53,$63,$FC,$BA,$D7,$FC,$E0,$D7,$FD,$EF,$1C,$0E,$A4,$7A
	.dc.b	$B4,$D3,$82,$38,$23,$E0,$78,$34,$D9,$5D,$BE,$47,$6D,$36,$57,$6F
	.dc.b	$A0,$E4,$43,$45,$34,$60,$86,$09,$69,$F8,$8D,$9F,$1C,$EA,$72,$48
	.dc.b	$FF,$4E,$FC,$AF,$FA,$80,$8E,$09,$3F,$E8,$98,$CB,$63,$27,$FA,$86
	.dc.b	$32,$D3,$31,$FE,$96,$8C,$B6,$32,$FF,$AA,$E3,$2D,$8C,$BF,$EA,$A8
	.dc.b	$CB,$63,$34,$9C,$B6,$33,$18,$23,$82,$4F,$AC,$8F,$A5,$A7,$9C,$A7
	.dc.b	$94,$A7,$8C,$B3,$E1,$6C,$F8,$1B,$3E,$03,$B1,$1E,$CD,$7B,$35,$EC
	.dc.b	$D9,$EC,$D9,$EC,$D3,$B4,$F4,$21,$D0,$A7,$42,$CF,$46,$CF,$46,$CF
	.dc.b	$47,$91,$0E,$45,$9E,$4D,$9E,$4D,$9E,$4F,$02,$4F,$06,$CF,$06,$CF
	.dc.b	$09,$D1,$27,$4D,$8D,$46,$08,$E0,$93,$F5,$08,$FC,$96,$9F,$00,$A7
	.dc.b	$A4,$A7,$90,$B3,$E1,$6C,$F8,$1B,$3E,$03,$B1,$1E,$CD,$7B,$35,$EC
	.dc.b	$D9,$EC,$D9,$EC,$D3,$B4,$F4,$21,$D0,$A7,$42,$CF,$46,$CF,$46,$CF
	.dc.b	$47,$91,$0E,$45,$9E,$4D,$9E,$4D,$9E,$4F,$02,$4F,$06,$CF,$06,$CF
	.dc.b	$09,$D1,$27,$4D,$8D,$46,$08,$E0,$92,$7F,$C9,$1E,$E7,$F9,$54,$8F
	.dc.b	$FA,$1F,$81,$18,$9E,$07,$F9,$35,$C4,$F2,$38,$38,$8D,$46,$08,$F2
	.dc.b	$3F,$CC,$31,$B6,$C6,$8E,$8E,$CF,$F3,$14,$6D,$B1,$B7,$B9,$B9,$F4
	.dc.b	$C6,$63,$A9,$99,$E5,$3A,$8C,$4E,$DE,$A6,$08,$F9,$08,$F8,$1A,$F6
	.dc.b	$6B,$C9,$B1,$DD,$B3,$E2,$23,$E1,$8D,$4F,$52,$1D,$4A,$75,$3F,$D8
	.dc.b	$9E,$8B,$C8,$FF,$66,$52,$30,$FF,$B1,$4C,$46,$A3,$33,$F5,$67,$26
	.dc.b	$E3,$2A,$3B,$CF,$13,$FD,$B0,$FF,$9B,$4F,$F3,$A9,$FE,$7C,$3F,$D0
	.dc.b	$07,$F9,$D3,$87,$FC,$FC,$FF,$9D,$CF,$F3,$F9,$C7,$8B,$FC,$FC,$48
	.dc.b	$FF,$3C,$C7,$FA,$64,$D3,$8D,$FF,$D1,$D1,$D4,$EE,$74,$38,$11,$DB
	.dc.b	$96,$BE,$32,$6F,$FE,$8C,$26,$1F,$E8,$F2,$47,$B4,$FF,$4B,$57,$B2
	.dc.b	$77,$19,$3C,$E7,$C9,$3F,$C9,$B3,$F2,$ED,$5C,$63,$0D,$9E,$DC,$3F
	.dc.b	$E8,$DE,$2F,$F4,$8B,$7B,$FE,$A7,$0F,$FA,$4A,$51,$FE,$CC,$AF,$F4
	.dc.b	$D9,$E1,$29,$E3,$A4,$F0,$8F,$F6,$9C,$75,$7F,$D4,$51,$E0,$3F,$D4
	.dc.b	$3C,$3F,$E8,$DD,$3F,$CD,$5C,$5F,$E7,$58,$FF,$4D,$9F,$E7,$AA,$FF
	.dc.b	$41,$EB,$FE,$70,$E3,$FF,$46,$36,$3F,$CE,$BA,$FF,$A2,$9F,$F5,$0C
	.dc.b	$65,$FF,$42,$5F,$53,$FD,$19,$7D,$9C,$4E,$E3,$FC,$F7,$87,$92,$3F
	.dc.b	$D0,$55,$C9,$B3,$AC,$3C,$73,$FE,$A6,$C3,$C5,$7F,$E8,$C8,$F4,$9E
	.dc.b	$42,$39,$3B,$38,$9F,$F4,$8B,$CC,$FF,$5B,$9F,$E9,$F7,$BC,$71,$69
	.dc.b	$C1,$E6,$47,$06,$A3,$E3,$9D,$0F,$F5,$1D,$7C,$F9,$E1,$1D,$CD,$B8
	.dc.b	$7F,$D9,$4D,$8F,$B5,$3D,$67,$E3,$1E,$99,$D6,$1F,$77,$4F,$F5,$54
	.dc.b	$FB,$CE,$C7,$09,$DE,$3F,$4E,$BE,$BD,$70,$9F,$81,$5E,$4B,$F0,$61
	.dc.b	$E2,$8F,$F4,$0B,$FE,$7D,$7F,$D1,$A7,$F9,$F7,$1F,$F3,$FC,$65,$FF
	.dc.b	$47,$B6,$3F,$D1,$8D,$79,$DF,$FA,$0D,$E7,$19,$7A,$18,$8D,$4F,$46
	.dc.b	$C7,$46,$BD,$23,$C7,$18,$8E,$71,$C4,$C9,$1E,$CD,$34,$E0,$8E,$08
	.dc.b	$F8,$5E,$0D,$36,$57,$6F,$94,$C9,$5C,$BE,$63,$81,$0D,$14,$D1,$82
	.dc.b	$18,$24,$FC,$B2,$4F,$C3,$6C,$FA,$27,$8C,$61,$F2,$9C,$A7,$81,$27
	.dc.b	$84,$65,$B3,$C5,$F5,$B6,$3D,$0D,$7B,$CF,$37,$BC,$65,$E8,$62,$35
	.dc.b	$3D,$1B,$1D,$1A,$F4,$8F,$1C,$62,$39,$C7,$13,$24,$7B,$34,$D3,$82
	.dc.b	$38,$23,$E1,$78,$34,$D9,$5D,$BE,$53,$25,$72,$F9,$8E,$04,$34,$53
	.dc.b	$46,$08,$60,$93,$F2,$C9,$3F,$0D,$B3,$E8,$9E,$31,$87,$CA,$7F,$98
	.dc.b	$0F,$F4,$49,$23,$FC,$FC,$D3,$FC,$EA,$53,$FC,$DC,$53,$93,$C0,$8F
	.dc.b	$06,$BC,$1A,$F0,$34,$7F,$A2,$38,$7F,$C6,$6F,$F8,$C4,$87,$F8,$C4
	.dc.b	$AE,$9C,$10,$FF,$00,$9F,$E0,$F3,$FC,$4B,$BF,$F8,$01,$FF,$0B,$9F
	.dc.b	$E1,$09,$FF,$06,$C6,$63,$FC,$1A,$6B,$3F,$F0,$6C,$66,$3A,$9C,$4F
	.dc.b	$F0,$81,$FE,$1F,$3B,$EF,$FE,$0B,$8E,$11,$97,$FC,$5C,$FF,$8A,$4E
	.dc.b	$E4,$3A,$CF,$82,$79,$12,$3C,$95,$E8,$8C,$47,$13,$B3,$FE,$37,$7C
	.dc.b	$67,$03,$CF,$9F,$F8,$CA,$BB,$46,$6F,$CF,$39,$BF,$F1,$2C,$FF,$94
	.dc.b	$F6,$FF,$0B,$67,$FE,$21,$DF,$FC,$47,$DF,$FE,$1D,$E0,$FF,$0C,$47
	.dc.b	$F8,$8E,$BF,$C3,$ED,$8F,$F1,$05,$FF,$89,$67,$2F,$F8,$8A,$38,$BF
	.dc.b	$E2,$29,$EA,$F8,$63,$FC,$47,$C0,$D3,$FC,$61,$BA,$31,$5F,$E3,$11
	.dc.b	$19,$9D,$CE,$A7,$FC,$6D,$39,$78,$35,$CC,$6D,$FF,$0C,$98,$9E,$64
	.dc.b	$78,$C6,$DF,$F3,$D4,$F3,$78,$90,$CC,$6D,$F3,$B6,$38,$46,$A7,$B3
	.dc.b	$E1,$3B,$B8,$7A,$1C,$9C,$3C,$4D,$90,$E4,$43,$8C,$78,$1E,$46,$CC
	.dc.b	$CF,$94,$93,$DA,$36,$73,$3C,$05,$32,$72,$78,$4F,$72,$4F,$49,$F0
	.dc.b	$1C,$8E,$06,$0E,$66,$48,$60,$91,$D2,$7B,$CF,$59,$F2,$4E,$5E,$8D
	.dc.b	$73,$1B,$7C,$26,$27,$99,$1E,$31,$B7,$C7,$3C,$DE,$24,$33,$1B,$7C
	.dc.b	$ED,$8E,$11,$A9,$EC,$F8,$4E,$EE,$1E,$87,$27,$0F,$13,$64,$39,$10
	.dc.b	$E3,$1E,$07,$91,$B3,$33,$E5,$24,$F6,$8D,$9C,$CF,$01,$4C,$9C,$9E
	.dc.b	$13,$DC,$93,$D2,$7C,$07,$23,$81,$83,$99,$92,$18,$24,$74,$9E,$F3
	.dc.b	$D6,$7C,$93,$97,$A3,$5C,$C6,$DF,$09,$89,$E6,$47,$8C,$6D,$F1,$CF
	.dc.b	$37,$89,$0C,$C6,$DF,$3B,$63,$84,$6A,$7B,$3E,$13,$BB,$87,$A1,$C9
	.dc.b	$C3,$C4,$D9,$0E,$44,$38,$C7,$81,$E4,$6C,$CC,$F9,$49,$3D,$A3,$67
	.dc.b	$33,$C0,$53,$27,$27,$84,$F7,$24,$F4,$9F,$01,$C8,$E0,$60,$E6,$64
	.dc.b	$86,$09,$1D,$27,$BC,$F5,$9F,$24,$E5,$E8,$D7,$31,$B7,$C2,$62,$79
	.dc.b	$91,$E3,$1B,$7C,$73,$CD,$E2,$43,$31,$B7,$CE,$D8,$E1,$1A,$9E,$CF
	.dc.b	$84,$EE,$E1,$E8,$72,$70,$F1,$36,$43,$91,$0E,$31,$E0,$79,$1B,$33
	.dc.b	$3E,$52,$4F,$68,$D9,$CC,$F0,$14,$C9,$C9,$E1,$3D,$C9,$3D,$27,$C0
	.dc.b	$72,$38,$18,$39,$99,$21,$82,$47,$49,$EF,$3D,$67,$C9,$39,$7A,$35
	.dc.b	$CC,$6D,$F0,$98,$9F,$D0,$61,$FE,$24,$C7,$FC,$45,$8F,$F8,$62,$93
	.dc.b	$FE,$1E,$CB,$FC,$51,$1F,$E1,$F9,$FF,$12,$11,$C1,$FE,$21,$24,$70
	.dc.b	$14,$8F,$F0,$6F,$CD,$3B,$7F,$C4,$6D,$8F,$F0,$DB,$5F,$F0,$FB,$5E
	.dc.b	$47,$F8,$99,$C3,$D2,$39,$91,$E6,$D7,$9B,$5E,$79,$A7,$9C,$75,$9D
	.dc.b	$3D,$4E,$64,$39,$94,$E7,$1C,$A7,$A4,$F2,$9E,$86,$CE,$87,$22,$1C
	.dc.b	$9E,$11,$CA,$38,$1C,$8D,$9C,$8E,$06,$9E,$0E,$8E,$06,$8C,$3A,$70
	.dc.b	$43,$A1,$0E,$25,$78,$36,$7E,$01,$27,$D2,$D8,$F2,$35,$F0,$B5,$F0
	.dc.b	$1D,$89,$3D,$9B,$1B,$7C,$13,$D8,$8F,$66,$BD,$9A,$F6,$9E,$84,$9E
	.dc.b	$91,$DA,$76,$F6,$3A,$10,$E8,$53,$A4,$72,$9E,$93,$CA,$7A,$1B,$3A
	.dc.b	$1C,$88,$72,$78,$47,$28,$E0,$72,$36,$72,$38,$1A,$78,$3A,$38,$1A
	.dc.b	$30,$E9,$C1,$0E,$84,$38,$95,$E0,$D9,$F8,$04,$9F,$4B,$63,$C8,$D7
	.dc.b	$C2,$D7,$C0,$76,$24,$F6,$6C,$6D,$F0,$4F,$62,$3D,$9A,$F6,$6B,$DA
	.dc.b	$7A,$12,$7A,$47,$69,$DB,$D8,$E8,$43,$A1,$4E,$91,$CA,$7A,$4F,$29
	.dc.b	$E8,$6C,$E8,$72,$21,$C9,$E1,$1C,$A3,$81,$C8,$D9,$C8,$E0,$69,$E0
	.dc.b	$E8,$E0,$68,$C3,$A7,$04,$3A,$10,$E2,$57,$83,$67,$E0,$12,$7D,$2D
	.dc.b	$8F,$23,$5F,$0B,$5F,$01,$D8,$93,$D9,$B1,$B7,$C1,$3D,$88,$F6,$6B
	.dc.b	$D9,$AF,$69,$E8,$49,$E9,$1D,$A7,$6F,$63,$A1,$0E,$85,$3A,$47,$29
	.dc.b	$E9,$3C,$A7,$A1,$B3,$A1,$C8,$87,$27,$84,$72,$8E,$07,$23,$67,$23
	.dc.b	$81,$A7,$83,$A3,$81,$A3,$0E,$9C,$10,$E8,$43,$89,$5E,$0D,$9F,$80
	.dc.b	$49,$F4,$B6,$3C,$8D,$7C,$2D,$7C,$07,$62,$4F,$66,$C6,$DF,$04,$F6
	.dc.b	$23,$D9,$AF,$67,$FB,$96,$3F,$E2,$18,$FF,$0A,$35,$FF,$10,$E8,$CB
	.dc.b	$FC,$3D,$1F,$E1,$03,$2E,$1F,$F1,$49,$FE,$20,$C6,$E5,$FE,$21,$7F
	.dc.b	$B9,$EC,$E2,$A4,$AF,$F1,$26,$8F,$9C,$FF,$0E,$D6,$E7,$FC,$20,$D7
	.dc.b	$FC,$12,$D9,$FF,$02,$36,$78,$D7,$5A,$DB,$6B,$B4,$7F,$8C,$EB,$96
	.dc.b	$BF,$E3,$16,$EB,$9B,$E1,$87,$4B,$F1,$C7,$3B,$E0,$FF,$8A,$A7,$53
	.dc.b	$E0,$36,$41,$87,$89,$61,$E0,$6D,$75,$C3,$C5,$1F,$1A,$FA,$DF,$C0
	.dc.b	$AD,$4F,$03,$47,$03,$C4,$48,$E0,$E8,$E0,$67,$0F,$5B,$E7,$8F,$36
	.dc.b	$1E,$52,$57,$C4,$9E,$D3,$D1,$E5,$86,$30,$F3,$E1,$D2,$72,$78,$2B
	.dc.b	$AD,$FA,$B0,$EB,$3E,$56,$BC,$1F,$BC,$F8,$2B,$A5,$78,$49,$7E,$0C
	.dc.b	$34,$4A,$F0,$46,$5F,$15,$72,$39,$C6,$6F,$C3,$7D,$2B,$94,$62,$B7
	.dc.b	$18,$C3,$D4,$7E,$0D,$C1,$F2,$E3,$6F,$83,$0C,$BF,$57,$0F,$05,$72
	.dc.b	$8F,$84,$D7,$47,$0C,$3E,$05,$F4,$C3,$C9,$86,$70,$E7,$7E,$18,$FA
	.dc.b	$8F,$1A,$F8,$11,$88,$F2,$12,$BE,$0D,$72,$8F,$A5,$87,$B5,$D3,$E5
	.dc.b	$C3,$ED,$C7,$73,$27,$A8,$C5,$74,$7C,$57,$E7,$69,$C0,$FD,$3D,$F8
	.dc.b	$6B,$DF,$3E,$7A,$F7,$D6,$E7,$C2,$F9,$8F,$CD,$57,$1C,$3D,$F1,$F8
	.dc.b	$23,$BD,$78,$DE,$51,$A8,$FC,$7C,$62,$FC,$CD,$3C,$15,$D6,$33,$3E
	.dc.b	$43,$A1,$70,$FD,$45,$71,$8C,$C7,$CB,$E0,$FE,$5D,$8F,$F0,$56,$E1
	.dc.b	$FE,$C0,$6D,$FF,$B2,$DA,$9A,$47,$A1,$0F,$E0,$0E,$5A,$64,$F9,$E7
	.dc.b	$53,$13,$E3,$6C,$F8,$DA,$6C,$E6,$43,$99,$E5,$36,$53,$51,$D5,$B1
	.dc.b	$D5,$A7,$53,$9C,$FA,$5A,$74,$71,$3C,$E3,$27,$23,$C0,$60,$8E,$C9
	.dc.b	$3E,$18,$CB,$63,$31,$E3,$6B,$D1,$B3,$E2,$9E,$04,$9E,$0D,$78,$36
	.dc.b	$38,$46,$88,$68,$AE,$9C,$11,$C1,$1F,$88,$43,$E1,$39,$69,$93,$DE
	.dc.b	$76,$31,$3E,$C6,$CF,$9D,$A6,$CE,$64,$39,$9E,$53,$65,$35,$1D,$5B
	.dc.b	$1D,$5A,$75,$39,$CF,$A5,$A7,$47,$13,$CE,$32,$72,$3C,$06,$08,$EC
	.dc.b	$93,$E1,$8C,$B6,$33,$1E,$36,$BD,$1B,$3E,$29,$E0,$49,$E0,$D7,$83
	.dc.b	$63,$84,$68,$86,$8A,$E9,$C1,$1C,$11,$F8,$84,$3E,$13,$96,$99,$3D
	.dc.b	$E7,$63,$13,$EC,$6C,$F9,$DA,$6C,$E6,$43,$99,$E5,$36,$53,$51,$D5
	.dc.b	$B1,$D5,$A7,$53,$9C,$FA,$5A,$7E,$5B,$E3,$BF,$6C,$47,$F6,$CD,$7F
	.dc.b	$6C,$D8,$FE,$5E,$D8,$1F,$E5,$47,$0D,$39,$90,$E6,$58,$E6,$D3,$99
	.dc.b	$FE,$40,$8E,$6D,$34,$75,$39,$92,$39,$B4,$E7,$49,$E6,$D8,$E6,$75
	.dc.b	$38,$1D,$67,$99,$1E,$73,$C8,$8F,$29,$E8,$F2,$3A,$4F,$17,$A1,$C8
	.dc.b	$E1,$3C,$9E,$13,$CA,$78,$47,$23,$E2,$F2,$3D,$EF,$E6,$0D,$8F,$E8
	.dc.b	$0D,$7F,$9C,$B6,$7F,$D2,$8D,$8E,$73,$C8,$8F,$F4,$9E,$17,$3B,$E9
	.dc.b	$7E,$FF,$33,$9F,$EB,$D3,$DE,$FB,$33,$9C,$3F,$D4,$F1,$8A,$DE,$1C
	.dc.b	$E5,$5B,$57,$E0,$C3,$BD,$F6,$C3,$95,$74,$32,$7F,$AD,$30,$E9,$87
	.dc.b	$3C,$39,$4F,$6B,$E4,$F0,$3B,$DF,$47,$C3,$7C,$A7,$CF,$86,$EF,$DD
	.dc.b	$56,$F8,$47,$6C,$3C,$18,$71,$AF,$31,$E1,$2C,$6A,$FA,$4E,$E7,$C6
	.dc.b	$75,$33,$11,$E4,$68,$B5,$E4,$6C,$F3,$8F,$26,$1F,$02,$77,$87,$A2
	.dc.b	$B4,$6D,$F5,$99,$BF,$06,$1F,$2E,$33,$58,$BF,$8A,$F4,$3D,$07,$8A
	.dc.b	$B4,$FC,$08,$F1,$57,$C1,$8D,$1C,$48,$E4,$91,$C0,$91,$F7,$E7,$AB
	.dc.b	$F0,$E7,$CC,$E3,$0F,$82,$F8,$48,$7D,$E9,$D5,$FB,$23,$C7,$5E,$07
	.dc.b	$75,$F3,$63,$11,$C4,$F3,$DF,$91,$C4,$6A,$78,$4E,$5F,$1C,$EA,$31
	.dc.b	$5F,$3E,$3C,$31,$87,$D5,$87,$D5,$7E,$6D,$F5,$8C,$47,$88,$F8,$87
	.dc.b	$23,$D3,$5D,$67,$C8,$FB,$DE,$58,$74,$BF,$35,$FE,$12,$31,$7F,$0A
	.dc.b	$30,$F7,$3E,$84,$FE,$0A,$BA,$C7,$03,$DC,$79,$1F,$AA,$F1,$C3,$E1
	.dc.b	$93,$0F,$25,$73,$3F,$86,$F3,$39,$BF,$A7,$F3,$39,$B3,$18,$6E,$29
	.dc.b	$D3,$FD,$5B,$15,$F3,$8D,$57,$07,$FA,$89,$FC,$4A,$34,$E2,$78,$DE
	.dc.b	$A3,$A4,$E9,$F0,$DF,$81,$E5,$3E,$57,$6D,$72,$78,$8F,$F5,$FF,$1F
	.dc.b	$FA,$83,$93,$FD,$45,$1F,$EB,$43,$FD,$63,$19,$3F,$D9,$5C,$DF,$ED
	.dc.b	$B6,$F3,$7F,$AF,$DB,$3F,$EB,$93,$FD,$54,$F1,$7B,$9D,$4C,$96,$BC
	.dc.b	$2F,$FB,$A8,$C9,$5C,$F2,$7F,$BF,$E3,$2D,$8C,$DF,$96,$7B,$9C,$6B
	.dc.b	$CB,$3F,$EC,$83,$77,$E9,$9D,$DF,$3A,$D9,$FE,$EB,$AD,$CF,$B5,$A7
	.dc.b	$38,$DC,$7F,$BA,$CF,$F6,$F1,$E1,$3A,$DF,$95,$CD,$77,$73,$7E,$37
	.dc.b	$33,$B9,$FF,$76,$C7,$FB,$70,$FA,$A7,$C9,$AF,$0B,$FE,$BC,$BF,$2C
	.dc.b	$7A,$4C,$CF,$FB,$C8,$E7,$1F,$ED,$E9,$F0,$9E,$D9,$FF,$21,$F2,$7F
	.dc.b	$8E,$EB,$FC,$85,$3F,$E3,$C3,$FC,$82,$7F,$91,$CC,$11,$D9,$39,$BF
	.dc.b	$C9,$11,$96,$C6,$63,$FC,$94,$D7,$A3,$67,$FC,$2A,$FF,$85,$E3,$FC
	.dc.b	$38,$7F,$86,$B9,$BF,$C3,$93,$A6,$C6,$A3,$0F,$F8,$88,$87,$12,$BA
	.dc.b	$70,$47,$04,$7A,$10,$FF,$17,$90,$FF,$15,$16,$7D,$0D,$9F,$2B,$4F
	.dc.b	$09,$CC,$E2,$48,$E2,$D8,$E2,$D3,$89,$CC,$F4,$1C,$E3,$89,$23,$8B
	.dc.b	$63,$89,$CC,$E2,$73,$9F,$80,$D3,$A4,$F2,$23,$CA,$7A,$3C,$8E,$93
	.dc.b	$C9,$A7,$23,$84,$F2,$78,$4F,$29,$E0,$D8,$E0,$6A,$78,$4E,$9B,$1A
	.dc.b	$8C,$3D,$88,$71,$2B,$A7,$04,$70,$47,$A1,$0F,$88,$43,$DA,$59,$F4
	.dc.b	$36,$7C,$AD,$3C,$27,$33,$89,$23,$8B,$63,$8B,$4E,$27,$33,$D0,$73
	.dc.b	$8E,$24,$8E,$2D,$8E,$27,$33,$89,$CE,$7E,$03,$4E,$93,$C8,$8F,$29
	.dc.b	$E8,$F2,$3A,$4F,$26,$9C,$8E,$13,$C9,$E1,$3C,$A7,$F1,$1C,$DF,$E8
	.dc.b	$B9,$FF,$61,$B5,$34,$9C,$DF,$E8,$E8,$FF,$4A,$B6,$7F,$D2,$AD,$8F
	.dc.b	$F4,$A9,$FE,$9C,$70,$F6,$6B,$D9,$B3,$DA,$38,$10,$EE,$59,$EE,$D3
	.dc.b	$B9,$A2,$3A,$69,$A7,$04,$70,$43,$CC,$75,$29,$C8,$AF,$16,$9C,$4F
	.dc.b	$61,$C4,$A7,$12,$BC,$5A,$71,$8F,$84,$ED,$B1,$FE,$EE,$1F,$12,$F9
	.dc.b	$2C,$57,$FB,$F2,$73,$18,$84,$61,$D1,$B9,$E4,$E0,$DB,$87,$89,$B9
	.dc.b	$E9,$3C,$E3,$27,$08,$F0,$3F,$E4,$A3,$FC,$9E,$7F,$94,$FA,$3F,$CF
	.dc.b	$2F,$F9,$3F,$8F,$FC,$97,$D1,$FE,$64,$E8,$FF,$27,$9F,$E5,$E9,$FF
	.dc.b	$28,$74,$7F,$98,$FA,$3F,$C9,$D1,$FE,$66,$3F,$CD,$F8,$73,$9D,$47
	.dc.b	$F9,$66,$7F,$CE,$8F,$F9,$B4,$F2,$61,$FE,$7E,$7C,$86,$5C,$3E,$83
	.dc.b	$78,$79,$67,$FC,$DF,$7E,$A9,$FF,$32,$4F,$F9,$BB,$0F,$4D,$6A,$7A
	.dc.b	$61,$FE,$66,$7C,$84,$3B,$61,$E0,$3E,$1E,$1E,$B8,$F6,$9F,$08,$B7
	.dc.b	$D9,$F4,$BF,$E7,$A7,$C7,$1E,$58,$D4,$7A,$A3,$A3,$EF,$C3,$D8,$F2
	.dc.b	$C3,$E6,$BE,$4C,$3E,$5B,$F2,$1F,$53,$CC,$F2,$1C,$EB,$AD,$FD,$93
	.dc.b	$C3,$87,$B7,$0F,$8D,$87,$D4,$9D,$34,$F0,$CF,$FA,$36,$FF,$0D,$5C
	.dc.b	$1F,$C0,$1D,$CF,$0D,$72,$72,$FF,$AE,$F9,$FF,$D1,$1C,$FF,$E7,$DB
	.dc.b	$FF,$41,$D7,$F9,$E4,$FF,$3D,$C6,$67,$7D,$1F,$E8,$D9,$87,$FA,$08
	.dc.b	$FF,$45,$57,$27,$FD,$1D,$3C,$1D,$57,$3E,$8F,$F4,$45,$78,$23,$FD
	.dc.b	$1D,$C5,$FE,$48,$E5,$FF,$3E,$5F,$F9,$5B,$A3,$FC,$B1,$39,$8F,$F2
	.dc.b	$A1,$FE,$4C,$E7,$FF,$33,$36,$BF,$CA,$8D,$3F,$CD,$13,$C5,$FF,$34
	.dc.b	$1D,$4F,$F2,$A1,$B3,$FC,$BD,$1F,$E5,$CE,$74,$7F,$99,$0E,$33,$CC
	.dc.b	$FF,$37,$3C,$EB,$C6,$DA,$F1,$B4,$F1,$F4,$30,$E9,$7F,$E6,$98,$C4
	.dc.b	$FF,$A4,$B0,$F0,$BF,$E9,$48,$FF,$3C,$1A,$C3,$FD,$33,$3A,$7F,$CE
	.dc.b	$B3,$DA,$FA,$CE,$70,$F6,$4F,$82,$F9,$D6,$A3,$0F,$D0,$8E,$D3,$E2
	.dc.b	$C3,$E0,$5F,$3B,$F9,$11,$E9,$BE,$51,$E7,$79,$18,$C3,$CB,$1F,$38
	.dc.b	$E3,$5C,$FC,$77,$8E,$F8,$6C,$3F,$CC,$6D,$97,$D4,$AD,$F3,$6A,$E6
	.dc.b	$3F,$62,$F6,$AB,$6E,$6F,$96,$1D,$5B,$7E,$DC,$9F,$AB,$3D,$2F,$81
	.dc.b	$E4,$BE,$D7,$86,$48,$C6,$BC,$6F,$84,$F7,$3F,$00,$E1,$6B,$E6,$AE
	.dc.b	$31,$B6,$BA,$CF,$F2,$79,$79,$32,$EC,$79,$32,$F1,$DF,$0C,$3E,$0D
	.dc.b	$6F,$0F,$13,$5E,$87,$A6,$B8,$9E,$C8,$D4,$F2,$C7,$D0,$F9,$E7,$79
	.dc.b	$78,$AF,$AD,$6B,$2F,$D4,$E3,$E3,$6D,$F5,$9D,$57,$C4,$CB,$D5,$84
	.dc.b	$8E,$71,$D5,$F3,$E1,$CD,$F4,$4E,$71,$F3,$57,$19,$DE,$1D,$E3,$58
	.dc.b	$F8,$AF,$57,$C2,$FB,$9F,$46,$FD,$AE,$0F,$8F,$97,$C4,$BE,$51,$BB
	.dc.b	$EA,$FA,$CC,$10,$E3,$97,$C0,$AE,$51,$EA,$35,$87,$8D,$FA,$A3,$77
	.dc.b	$C8,$FF,$83,$1B,$1F,$E1,$73,$51,$92,$39,$8D,$BF,$E1,$63,$66,$63
	.dc.b	$6E,$63,$7D,$9F,$C4,$A3,$66,$67,$71,$99,$DC,$FF,$07,$9C,$B4,$EC
	.dc.b	$43,$91,$67,$94,$70,$8F,$E6,$33,$D1,$B1,$97,$C7,$3C,$1E,$47,$73
	.dc.b	$A9,$1D,$34,$D3,$82,$38,$21,$E0,$3A,$94,$E4,$57,$8B,$4E,$27,$98
	.dc.b	$B3,$EB,$70,$6C,$86,$08,$EC,$93,$F0,$5B,$1E,$43,$71,$92,$39,$8D
	.dc.b	$BE,$83,$66,$63,$6E,$63,$73,$EC,$8D,$99,$9D,$C6,$67,$73,$EE,$9C
	.dc.b	$B4,$EC,$43,$91,$67,$94,$70,$8F,$87,$3D,$1B,$19,$7C,$73,$C1,$E4
	.dc.b	$77,$3A,$91,$D3,$4D,$38,$23,$82,$1E,$03,$A9,$4E,$45,$78,$B4,$E2
	.dc.b	$79,$8B,$3E,$B7,$06,$C8,$60,$8E,$C9,$3F,$05,$B1,$E4,$37,$19,$23
	.dc.b	$98,$DB,$E8,$36,$66,$36,$E6,$37,$3E,$C8,$D9,$99,$DC,$66,$77,$3E
	.dc.b	$E9,$CB,$4E,$C4,$39,$16,$79,$47,$08,$F8,$73,$D1,$B1,$97,$C7,$3C
	.dc.b	$1E,$47,$73,$A9,$1D,$34,$FA,$7C,$FF,$CA,$C8,$FF,$27,$6B,$FC,$8D
	.dc.b	$B1,$FC,$7D,$B1,$FE,$25,$6B,$FC,$61,$AF,$F3,$96,$D1,$A4,$8F,$01
	.dc.b	$0F,$01,$67,$C0,$D3,$C0,$53,$C0,$53,$C0,$59,$F0,$36,$7C,$0F,$62
	.dc.b	$4F,$66,$9D,$8A,$76,$29,$D8,$B3,$D9,$B3,$DA,$7A,$12,$3A,$35,$FF
	.dc.b	$37,$F7,$7F,$89,$5F,$EA,$0D,$7F,$CC,$6D,$77,$B7,$F5,$57,$2D,$72
	.dc.b	$58,$E0,$9B,$CD,$FE,$AD,$8F,$F6,$39,$FE,$D6,$3F,$D5,$FD,$9F,$EC
	.dc.b	$9E,$CF,$F6,$D7,$5F,$FB,$34,$FF,$5D,$F6,$7F,$AC,$EB,$FD,$8C,$7F
	.dc.b	$B5,$27,$FD,$74,$FF,$B7,$6F,$B5,$70,$BE,$93,$FE,$CF,$35,$DB,$FE
	.dc.b	$CC,$AF,$F7,$0C,$7F,$AF,$2F,$FD,$87,$87,$FB,$5A,$3F,$DC,$EF,$86
	.dc.b	$7F,$DD,$15,$FE,$D6,$7B,$D7,$A0,$FF,$65,$C7,$59,$E0,$DB,$F1,$DF
	.dc.b	$46,$DF,$FB,$5D,$B3,$D6,$BC,$53,$9A,$F9,$07,$CA,$C3,$E3,$61,$E1
	.dc.b	$8F,$69,$F3,$0D,$C7,$C3,$6C,$6D,$FF,$77,$1B,$8E,$6F,$C7,$32,$77
	.dc.b	$8F,$6B,$B6,$B9,$3C,$25,$38,$C7,$84,$EA,$FC,$10,$A3,$FD,$67,$3F
	.dc.b	$ED,$9D,$2F,$6F,$FB,$7E,$FF,$DD,$31,$8E,$DF,$F6,$FC,$9F,$CA,$57
	.dc.b	$FB,$9E,$BF,$DD,$61,$C3,$FE,$E8,$AE,$4F,$FB,$D6,$3C,$0E,$4F,$F5
	.dc.b	$49,$FE,$B8,$3F,$DB,$67,$FB,$3C,$FF,$57,$17,$7F,$F5,$B3,$FE,$CD
	.dc.b	$38,$70,$FF,$B6,$23,$FD,$B9,$D9,$FE,$C4,$9D,$1F,$EB,$B3,$AC,$6B
	.dc.b	$8F,$FD,$7B,$1A,$AF,$F5,$EC,$F6,$7F,$DA,$B1,$DB,$9B,$FD,$C7,$3C
	.dc.b	$63,$9D,$FF,$B3,$9F,$F1,$FB,$5F,$F7,$7B,$63,$FD,$A3,$1C,$CF,$F6
	.dc.b	$F1,$E1,$9F,$F6,$EB,$E7,$AF,$65,$71,$25,$FA,$23,$CB,$3B,$AF,$2D
	.dc.b	$6A,$3F,$CE,$37,$FE,$FE,$3D,$15,$D2,$FF,$DB,$97,$E0,$8F,$83,$D9
	.dc.b	$FE,$E5,$BF,$A0,$DB,$E3,$5D,$AF,$8D,$7F,$B5,$7A,$FF,$DC,$08,$72
	.dc.b	$D7,$5D,$76,$31,$D9,$FE,$E8,$8C,$4F,$FB,$68,$E5,$3A,$AE,$B1,$D0
	.dc.b	$D1,$4E,$A7,$03,$31,$E1,$BF,$F6,$43,$7A,$7F,$DB,$4D,$E8,$F0,$D7
	.dc.b	$FA,$84,$E9,$5C,$1F,$F6,$CD,$70,$6C,$F0,$9D,$47,$1A,$E8,$F8,$CC
	.dc.b	$1C,$0C,$47,$02,$57,$58,$D9,$23,$35,$DE,$75,$18,$9E,$71,$82,$4E
	.dc.b	$C9,$1A,$6B,$EE,$8D,$B9,$21,$98,$DD,$7B,$23,$06,$9C,$11,$F7,$1A
	.dc.b	$23,$97,$D9,$3E,$07,$27,$08,$D5,$7B,$63,$27,$0A,$F1,$BF,$8D,$CD
	.dc.b	$F3,$DF,$EC,$D2,$47,$FB,$25,$AF,$FB,$15,$B7,$FE,$C1,$6C,$73,$6B
	.dc.b	$FE,$BA,$6B,$FE,$DB,$6C,$F8,$23,$B1,$0E,$C5,$9E,$CD,$3B,$14,$EC
	.dc.b	$53,$B1,$63,$B3,$D0,$91,$D1,$A7,$42,$9D,$0A,$74,$2C,$74,$8E,$44
	.dc.b	$8E,$4D,$79,$35,$E4,$D8,$E5,$1C,$08,$70,$29,$C0,$AF,$07,$44,$34
	.dc.b	$57,$4E,$08,$E0,$93,$F5,$09,$1F,$25,$AF,$C0,$6C,$FA,$5B,$1E,$46
	.dc.b	$BE,$16,$BE,$06,$CF,$82,$3B,$10,$EC,$59,$EC,$D3,$B1,$4E,$C5,$3B
	.dc.b	$16,$3B,$3D,$09,$1D,$1A,$74,$29,$D0,$A7,$42,$C7,$48,$E4,$48,$E4
	.dc.b	$D7,$93,$5E,$4D,$8E,$51,$C0,$87,$02,$9C,$0A,$F0,$74,$43,$45,$74
	.dc.b	$E0,$8E,$09,$3F,$50,$91,$F2,$5A,$FC,$06,$CF,$A5,$B1,$E4,$6B,$E1
	.dc.b	$6B,$E0,$6C,$F8,$23,$B1,$0E,$C5,$9E,$CD,$3B,$14,$EC,$53,$B1,$63
	.dc.b	$B3,$D0,$91,$D1,$A7,$43,$EE,$FC,$7F,$FA,$2D,$4A,$30,$D8,$FF,$48
	.dc.b	$35,$FF,$4A,$35,$FE,$F4,$72,$8E,$24,$78,$B5,$E2,$D7,$8C,$73,$7F
	.dc.b	$9C,$46,$5A,$72,$21,$C8,$A7,$27,$83,$FE,$0F,$8E,$8D,$8C,$BE,$73
	.dc.b	$89,$D8,$E8,$43,$A1,$A3,$91,$4E,$46,$CE,$47,$03,$07,$02,$9C,$0D
	.dc.b	$10,$E2,$43,$81,$92,$99,$39,$16,$3D,$EE,$0D,$90,$FB,$FF,$29,$F7
	.dc.b	$88,$FD,$E6,$BF,$79,$AF,$D6,$7F,$11,$C5,$FE,$B2,$8F,$F4,$C3,$5F
	.dc.b	$F4,$77,$3F,$FA,$42,$3F,$D5,$4D,$E9,$FF,$57,$75,$7F,$90,$E3,$FD
	.dc.b	$25,$5F,$E4,$5A,$FF,$5F,$BF,$EC,$F3,$15,$FE,$BB,$7F,$D5,$67,$83
	.dc.b	$AB,$FD,$6F,$1D,$23,$FD,$98,$7F,$B5,$63,$2F,$33,$FD,$B6,$7F,$A8
	.dc.b	$BA,$BF,$DA,$38,$78,$DF,$F6,$EC,$7F,$A6,$67,$FD,$96,$FF,$AC,$4F
	.dc.b	$F5,$7E,$1F,$EA,$D9,$FF,$59,$B6,$33,$1F,$EB,$FC,$3E,$21,$F0,$9D
	.dc.b	$99,$AF,$2C,$F8,$CF,$45,$F8,$DF,$82,$78,$4F,$F5,$EC,$F1,$8D,$47
	.dc.b	$B2,$FE,$5B,$E9,$8F,$F6,$EB,$F0,$6F,$E6,$BD,$CD,$1F,$30,$FB,$35
	.dc.b	$E6,$7E,$74,$F6,$9E,$35,$D9,$FF,$7D,$B7,$0F,$AC,$FD,$FA,$F0,$46
	.dc.b	$2F,$A4,$76,$3E,$39,$F8,$AE,$AF,$F6,$85,$FF,$B1,$23,$FD,$A2,$7F
	.dc.b	$B1,$4F,$F6,$64,$7F,$B4,$5F,$E2,$26,$3A,$BF,$DA,$71,$98,$FE,$2B
	.dc.b	$3F,$ED,$27,$99,$1E,$C7,$7D,$3F,$D3,$71,$FE,$B2,$E8,$FF,$55,$9F
	.dc.b	$EA,$27,$FD,$6C,$7F,$B4,$A3,$6D,$E9,$FF,$57,$4B,$A3,$FC,$A5,$1F
	.dc.b	$ED,$18,$E7,$1F,$EA,$D2,$4F,$FB,$80,$FF,$5C,$1E,$17,$8F,$27,$FA
	.dc.b	$A6,$7A,$4F,$FB,$32,$3F,$D8,$13,$A8,$FF,$5E,$D7,$78,$C5,$70,$9F
	.dc.b	$F6,$8D,$7F,$AD,$9F,$F5,$D9,$EC,$BF,$F6,$F5,$70,$24,$78,$CF,$01
	.dc.b	$DA,$7C,$2E,$2B,$B5,$7F,$B3,$6F,$E1,$DF,$2A,$F8,$D7,$E1,$AF,$92
	.dc.b	$78,$23,$94,$EE,$31,$1F,$D5,$AB,$9B,$E3,$31,$5F,$16,$B8,$74,$FF
	.dc.b	$B5,$6B,$75,$F0,$0F,$F7,$61,$F3,$F0,$FB,$8F,$D0,$3C,$A6,$BA,$7F
	.dc.b	$DA,$4F,$53,$FD,$96,$23,$E9,$C6,$A3,$17,$FE,$DF,$87,$43,$E1,$7F
	.dc.b	$DC,$17,$C2,$3F,$DB,$2D,$EA,$CD,$FF,$B6,$A3,$55,$8C,$3F,$DA,$51
	.dc.b	$F8,$2C,$DD,$4E,$AD,$46,$1B,$0F,$36,$C4,$BF,$65,$6D,$B5,$B6,$CE
	.dc.b	$DF,$69,$E2,$C7,$FD,$C3,$79,$C3,$C9,$7E,$3A,$E2,$74,$9D,$57,$FB
	.dc.b	$0A,$B7,$7E,$69,$CD,$75,$26,$1E,$E8,$CB,$E7,$8F,$1B,$6B,$C3,$3A
	.dc.b	$C3,$C5,$5A,$6C,$EA,$3C,$A6,$6F,$85,$71,$6C,$6A,$7C,$ED,$C3,$E0
	.dc.b	$B4,$F3,$DF,$68,$CC,$75,$35,$87,$9E,$33,$5D,$27,$44,$7C,$CF,$D0
	.dc.b	$3F,$C6,$FF,$3C,$67,$5F,$F8,$AA,$31,$08,$C3,$A3,$FD,$C9,$1A,$8C
	.dc.b	$C6,$CC,$9B,$3F,$C5,$51,$B7,$26,$CC,$9B,$3C,$A5,$E1,$47,$18,$D9
	.dc.b	$C4,$D9,$63,$6F,$18,$E0,$EA,$7C,$E4,$7C,$6D,$7A,$B5,$E0,$68,$86
	.dc.b	$8A,$68,$C1,$0C,$12,$7C,$AF,$9E,$7C,$B1,$89,$F0,$36,$7A,$B5,$E8
	.dc.b	$D7,$A4,$F2,$23,$F0,$48,$77,$2C,$F6,$8D,$C6,$49,$19,$76,$78,$A3
	.dc.b	$71,$98,$D9,$93,$67,$9A,$36,$E4,$D9,$93,$67,$A8,$B3,$ED,$24,$72
	.dc.b	$8E,$07,$23,$81,$63,$6F,$28,$E0,$EA,$7D,$04,$7C,$8D,$7A,$B5,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$12,$7C,$AF,$9E,$7C,$B1,$89,$F0,$36,$7A
	.dc.b	$B5,$E8,$D7,$A4,$F2,$23,$F0,$48,$77,$2C,$F6,$8D,$C6,$49,$19,$76
	.dc.b	$78,$A3,$71,$98,$D9,$93,$67,$9A,$36,$E4,$D9,$93,$67,$A8,$B3,$ED
	.dc.b	$24,$72,$8E,$07,$23,$81,$63,$6F,$28,$E0,$EA,$7D,$04,$7C,$8D,$7A
	.dc.b	$B5,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$7C,$AF,$9E,$7C,$B1,$F6,$FB
	.dc.b	$9F,$13,$FE,$ED,$6D,$7F,$95,$83,$87,$FC,$B2,$D7,$FD,$D6,$D7,$F9
	.dc.b	$DB,$5F,$E8,$0D,$79,$C7,$72,$3D,$E3,$44,$8C,$C7,$81,$AF,$81,$AF
	.dc.b	$81,$AF,$81,$AF,$80,$EC,$E2,$3B,$46,$08,$77,$29,$DC,$A7,$72,$9D
	.dc.b	$CA,$77,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70,$29,$C0,$A7
	.dc.b	$03,$44,$3F,$B9,$7D,$7B,$E0,$BE,$F4,$23,$0F,$E0,$5A,$FF,$A6,$1A
	.dc.b	$FF,$78,$38,$47,$28,$CC,$7F,$7C,$6B,$E0,$8F,$F7,$9F,$6F,$FB,$2B
	.dc.b	$B3,$FD,$68,$FF,$B6,$27,$FD,$7D,$D1,$FE,$D1,$7F,$D5,$E7,$FA,$C4
	.dc.b	$FF,$65,$9F,$EB,$E3,$FD,$D2,$7F,$B1,$27,$FD,$74,$F2,$9E,$91,$FE
	.dc.b	$DF,$8E,$27,$39,$FF,$6F,$F6,$7F,$BC,$E7,$FD,$89,$3F,$ED,$99,$CB
	.dc.b	$FE,$C4,$31,$5E,$9E,$CF,$F7,$9C,$78,$A7,$FD,$9B,$1F,$EF,$46,$C7
	.dc.b	$FB,$90,$F6,$DF,$79,$D5,$79,$9F,$F7,$34,$F3,$8C,$3E,$FB,$F2,$BC
	.dc.b	$8F,$21,$4F,$81,$DB,$FE,$EA,$69,$FE,$E7,$C3,$E6,$3F,$09,$A7,$B9
	.dc.b	$DC,$FC,$A3,$EA,$CF,$36,$C7,$3C,$3D,$4F,$B6,$3E,$75,$73,$76,$73
	.dc.b	$C3,$FC,$7F,$0C,$18,$7F,$BD,$FB,$3F,$DD,$64,$ED,$FF,$24,$78,$CF
	.dc.b	$0B,$8C,$E7,$3F,$F7,$5E,$3F,$EE,$72,$47,$FB,$AE,$37,$7C,$AA,$C2
	.dc.b	$FA,$3F,$EF,$7B,$FF,$76,$77,$7F,$AC,$B8,$BF,$D8,$9C,$FF,$EC,$18
	.dc.b	$C1,$FE,$D7,$AF,$F6,$09,$0F,$F5,$9F,$37,$FB,$A1,$FF,$5C,$73,$FF
	.dc.b	$BA,$1F,$F6,$C9,$FE,$B9,$BF,$F6,$38,$AF,$07,$7F,$FB,$0F,$0F,$01
	.dc.b	$2B,$FD,$86,$72,$C3,$FD,$86,$4C,$3B,$61,$FE,$D6,$BE,$55,$C2,$30
	.dc.b	$FF,$BA,$63,$35,$E7,$8E,$95,$CB,$0F,$7E,$1F,$EE,$89,$FF,$6C,$9F
	.dc.b	$EE,$B3,$FC,$FD,$87,$AA,$7D,$31,$D4,$87,$BF,$0E,$F7,$FE,$E7,$9F
	.dc.b	$95,$3D,$30,$FF,$79,$C3,$0F,$F6,$D4,$B0,$61,$FE,$B0,$F1,$9E,$33
	.dc.b	$C6,$FF,$68,$46,$1F,$F5,$25,$FF,$B8,$E3,$0D,$9E,$0F,$FA,$8C,$ED
	.dc.b	$E3,$7F,$99,$88,$FF,$A1,$5A,$FF,$A0,$5B,$D9,$FC,$8F,$14,$62,$7F
	.dc.b	$D3,$B1,$87,$91,$0E,$45,$9E,$53,$C6,$3F,$D5,$13,$C5,$D9,$27,$6F
	.dc.b	$13,$BC,$F1,$7D,$66,$4E,$6E,$67,$A4,$F2,$9E,$87,$23,$81,$C8,$C9
	.dc.b	$C8,$CC,$72,$78,$1A,$38,$47,$52,$3A,$6B,$CC,$E1,$1C,$DC,$46,$9E
	.dc.b	$67,$12,$1A,$29,$A3,$07,$CA,$E0,$FF,$50,$91,$FF,$60,$35,$FF,$5E
	.dc.b	$36,$3E,$E3,$63,$ED,$B5,$E8,$72,$21,$C8,$B1,$C9,$A7,$22,$9C,$8E
	.dc.b	$04,$78,$34,$E0,$53,$83,$A2,$3A,$6B,$A3,$04,$30,$43,$F5,$D1,$ED
	.dc.b	$69,$E6,$29,$E1,$2C,$76,$69,$D4,$A7,$59,$FE,$5E,$F1,$6C,$71,$6C
	.dc.b	$71,$6B,$C6,$7C,$91,$88,$EA,$47,$04,$8D,$92,$3D,$51,$96,$C6,$5F
	.dc.b	$23,$67,$C8,$D8,$F2,$35,$F2,$1C,$88,$72,$2C,$72,$69,$C8,$A7,$23
	.dc.b	$81,$1E,$0D,$38,$14,$E0,$E8,$8E,$9A,$E8,$C1,$0C,$10,$F9,$51,$ED
	.dc.b	$69,$E6,$29,$E1,$2C,$76,$69,$D4,$A7,$59,$FA,$0F,$16,$C7,$16,$C7
	.dc.b	$16,$BC,$67,$C9,$18,$8E,$A4,$70,$48,$D9,$23,$D5,$19,$6C,$65,$F2
	.dc.b	$36,$7C,$8D,$8F,$23,$5F,$21,$C8,$87,$22,$C7,$26,$9C,$8A,$72,$38
	.dc.b	$11,$E0,$D3,$81,$4E,$0E,$88,$E9,$AE,$8F,$DA,$75,$3B,$BF,$D8,$04
	.dc.b	$8F,$F5,$93,$5F,$F5,$FB,$5F,$F5,$FB,$5F,$F5,$FB,$5F,$F5,$FB,$5F
	.dc.b	$F6,$4B,$63,$C1,$1D,$88,$76,$29,$D8,$A7,$62,$9D,$8A,$76,$2B,$D9
	.dc.b	$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8,$87,$22,$9C,$8A,$72
	.dc.b	$2B,$C9,$E0,$43,$81,$4E,$05,$78,$2F,$0D,$E1,$BE,$27,$FD,$F1,$18
	.dc.b	$6C,$66,$7F,$D7,$07,$17,$0F,$FA,$EE,$78,$3A,$6B,$98,$DA,$24,$69
	.dc.b	$FF,$62,$72,$7F,$A0,$7A,$BF,$D4,$BD,$7F,$E9,$83,$FD,$0B,$D7,$FE
	.dc.b	$A9,$AE,$55,$FE,$A6,$4B,$B3,$B4,$7F,$A7,$6B,$FD,$17,$7F,$E8,$AB
	.dc.b	$E5,$3F,$EC,$18,$C7,$67,$96,$BC,$11,$FE,$AC,$6C,$FF,$AA,$E3,$FD
	.dc.b	$63,$3F,$E9,$D7,$CE,$62,$3F,$D3,$C7,$FA,$BA,$7D,$33,$D6,$75,$1F
	.dc.b	$E9,$D9,$F5,$D7,$FA,$E1,$FF,$67,$3E,$6C,$3D,$C7,$B6,$3F,$D9,$77
	.dc.b	$CD,$FF,$55,$5F,$9C,$EA,$7B,$1F,$F6,$24,$79,$DE,$ED,$7E,$11,$FE
	.dc.b	$E8,$9F,$F5,$ED,$78,$4F,$23,$89,$F0,$4F,$D0,$C3,$E7,$E1,$EE,$8E
	.dc.b	$E7,$C2,$3F,$D8,$B8,$78,$E3,$E5,$C7,$BA,$B9,$DF,$B6,$3E,$C5,$7C
	.dc.b	$C7,$DA,$7D,$12,$9F,$EF,$3E,$CF,$F5,$E7,$5F,$F8,$1A,$91,$8E,$1F
	.dc.b	$F5,$CD,$7E,$3A,$3F,$D7,$CF,$E2,$9A,$FE,$29,$B3,$CB,$B3,$FD,$85
	.dc.b	$3F,$EB,$F9,$E8,$FF,$AD,$88,$7F,$A8,$8B,$C7,$FE,$A8,$7F,$D3,$47
	.dc.b	$FA,$73,$AF,$FD,$60,$EB,$AF,$FD,$4E,$FF,$A9,$3A,$BF,$D3,$EF,$FB
	.dc.b	$16,$BF,$D9,$6F,$FA,$CE,$77,$4E,$CF,$F4,$3D,$F0,$AF,$F5,$24,$7F
	.dc.b	$A2,$AB,$FD,$4F,$7F,$EC,$7E,$CF,$F5,$2B,$D3,$0F,$F5,$D3,$E7,$7C
	.dc.b	$B7,$EB,$24,$66,$FF,$D7,$D5,$FE,$CD,$AE,$37,$E4,$9F,$1D,$FB,$DD
	.dc.b	$BE,$9A,$E1,$1F,$EB,$B7,$FD,$8D,$7E,$59,$FF,$63,$5F,$B5,$FF,$4D
	.dc.b	$E1,$ED,$9D,$4F,$FB,$02,$BD,$73,$FE,$BD,$EA,$FF,$5A,$DA,$B9,$75
	.dc.b	$FF,$AE,$A3,$18,$FF,$13,$C1,$E0,$FF,$41,$BF,$04,$63,$AB,$FD,$8B
	.dc.b	$8F,$49,$E6,$DB,$E5,$7C,$5F,$F5,$CF,$8F,$FC,$4A,$BD,$58,$FF,$A8
	.dc.b	$DB,$A7,$FA,$D5,$B3,$BA,$E3,$5F,$5B,$B7,$0F,$FA,$7E,$B6,$FD,$AA
	.dc.b	$CB,$D6,$BF,$D5,$73,$A8,$FF,$77,$CE,$AF,$C1,$5A,$3B,$CE,$8F,$01
	.dc.b	$0C,$12,$BD,$37,$CC,$E9,$72,$78,$4E,$9F,$31,$0F,$F8,$03,$3C,$5C
	.dc.b	$47,$9C,$E0,$E8,$CD,$79,$A3,$24,$70,$48,$F8,$CF,$B5,$FF,$71,$D7
	.dc.b	$95,$E8,$49,$F0,$4F,$18,$F9,$47,$23,$89,$E2,$30,$68,$5A,$BE,$77
	.dc.b	$C9,$7F,$52,$EE,$FF,$7F,$36,$BF,$DF,$AD,$7F,$DE,$ED,$79,$B6,$8D
	.dc.b	$24,$7F,$6B,$9E,$AD,$3A,$94,$EA,$53,$A9,$67,$AB,$67,$AA,$9E,$AD
	.dc.b	$3A,$94,$EA,$53,$A9,$67,$AB,$67,$AC,$F4,$24,$74,$6B,$D1,$AF,$46
	.dc.b	$CF,$46,$CF,$49,$E4,$43,$91,$4E,$45,$9E,$4D,$8E,$4F,$02,$1C,$0B
	.dc.b	$1C,$1B,$1C,$1D,$12,$34,$D8,$D4,$60,$8E,$09,$1F,$4C,$87,$C9,$2C
	.dc.b	$FC,$06,$9E,$92,$9E,$42,$9E,$12,$CF,$81,$B3,$E0,$7B,$12,$7B,$34
	.dc.b	$EC,$53,$B1,$4E,$C5,$9E,$CD,$9E,$D3,$D0,$91,$D1,$AF,$46,$BD,$1B
	.dc.b	$3D,$1B,$3D,$27,$91,$0E,$45,$39,$16,$79,$36,$39,$3C,$08,$70,$2C
	.dc.b	$70,$6C,$70,$74,$48,$D3,$63,$51,$82,$38,$24,$7D,$42,$1F,$24,$B3
	.dc.b	$F0,$1A,$7A,$4A,$79,$0A,$78,$4B,$3E,$06,$CF,$81,$EC,$49,$EC,$D3
	.dc.b	$B1,$4E,$C5,$3B,$16,$7B,$36,$7B,$4F,$42,$47,$46,$BD,$1A,$F4,$6C
	.dc.b	$F4,$6C,$F4,$9E,$44,$3F,$D2,$1D,$CF,$13,$FD,$37,$18,$6A,$8D,$3A
	.dc.b	$6B,$FE,$A2,$6B,$FE,$8E,$38,$47,$28,$CC,$7F,$A4,$9A,$F7,$3A,$46
	.dc.b	$C8,$E0,$87,$77,$2D,$32,$78,$63,$AB,$E2,$32,$78,$63,$8B,$B6,$B9
	.dc.b	$36,$79,$4E,$C4,$38,$1A,$3B,$19,$39,$14,$E4,$70,$30,$70,$30,$43
	.dc.b	$89,$4E,$26,$08,$60,$87,$33,$25,$33,$1F,$10,$91,$F8,$0E,$07,$CE
	.dc.b	$BE,$13,$0D,$BF,$F6,$FC,$EA,$72,$48,$CC,$ED,$FC,$14,$ED,$CC,$FE
	.dc.b	$33,$9F,$FD,$B9,$1F,$EB,$03,$FD,$49,$D9,$FE,$D4,$7F,$D6,$1D,$BF
	.dc.b	$EA,$AE,$CF,$F5,$04,$7F,$AB,$4F,$F5,$F9,$FE,$AC,$9F,$F6,$43,$FE
	.dc.b	$B8,$25,$7F,$AA,$63,$FD,$64,$7F,$AB,$9F,$F5,$C4,$71,$9F,$F5,$91
	.dc.b	$FE,$C6,$7C,$27,$FB,$5D,$FF,$65,$3F,$EA,$E8,$E3,$3C,$EB,$CD,$3E
	.dc.b	$B8,$F1,$CF,$FB,$5A,$73,$3F,$EB,$D3,$C0,$76,$3F,$D8,$65,$9F,$84
	.dc.b	$FA,$A3,$8C,$F8,$2B,$CD,$39,$3E,$1D,$78,$DF,$41,$FE,$C5,$3E,$41
	.dc.b	$FE,$CA,$3C,$D1,$F0,$1F,$65,$76,$EE,$FF,$6E,$B7,$BB,$FD,$C9,$87
	.dc.b	$AE,$BB,$DF,$8A,$3E,$DC,$7D,$0B,$FF,$73,$57,$FB,$6B,$AF,$FD,$8B
	.dc.b	$3F,$EC,$BE,$CF,$F6,$4B,$47,$77,$FB,$5E,$7F,$D9,$E4,$9F,$E2,$B8
	.dc.b	$7F,$B4,$6B,$F8,$71,$C6,$B8,$4F,$49,$D4,$62,$7A,$F7,$7F,$AC,$39
	.dc.b	$FF,$D6,$BD,$9F,$EB,$09,$FF,$66,$77,$7F,$B1,$04,$7F,$B1,$1C,$34
	.dc.b	$FF,$5A,$BA,$8F,$F5,$AF,$4F,$FA,$F9,$FF,$6E,$92,$BF,$D9,$3D,$1F
	.dc.b	$EC,$78,$FF,$6F,$5F,$85,$FF,$6A,$9F,$ED,$DE,$EF,$F5,$E6,$1D,$6F
	.dc.b	$FD,$79,$3F,$EE,$33,$FD,$94,$72,$3F,$D9,$44,$7D,$AD,$C3,$FD,$98
	.dc.b	$F1,$3C,$8F,$C3,$33,$1F,$EB,$2A,$F0,$DF,$BE,$B8,$4F,$AA,$FF,$DB
	.dc.b	$91,$FE,$D2,$3C,$67,$FB,$91,$CB,$FE,$82,$AE,$6F,$C6,$38,$1F,$ED
	.dc.b	$03,$E7,$F6,$3B,$BF,$D0,$58,$7F,$B3,$E7,$FC,$F0,$D7,$FC,$DC,$DE
	.dc.b	$B4,$7F,$B4,$27,$8D,$73,$8E,$2D,$78,$B6,$7F,$AC,$67,$FE,$97,$9F
	.dc.b	$F4,$17,$67,$FA,$3C,$8E,$08,$7F,$9E,$DC,$B4,$CC,$F4,$8C,$B6,$32
	.dc.b	$7F,$B3,$9C,$3C,$48,$E0,$8E,$C8,$7F,$AC,$8D,$98,$3F,$D6,$31,$E0
	.dc.b	$71,$3E,$D7,$67,$19,$EA,$D3,$73,$C5,$C1,$0E,$A7,$A8,$C4,$F5,$9E
	.dc.b	$71,$A3,$B9,$5C,$C6,$E3,$B9,$CC,$D1,$83,$8B,$B7,$EF,$AF,$2B,$FC
	.dc.b	$2E,$D8,$FF,$0D,$B5,$FE,$0E,$D7,$EF,$C7,$02,$3C,$1A,$F0,$6B,$C0
	.dc.b	$D1,$0D,$14,$D1,$82,$18,$21,$F9,$83,$FD,$CE,$F6,$71,$1E,$76,$C6
	.dc.b	$5F,$82,$D7,$89,$E1,$36,$53,$65,$36,$79,$0D,$94,$D9,$4D,$9E,$83
	.dc.b	$25,$32,$72,$31,$1E,$97,$66,$48,$60,$87,$22,$BC,$5C,$11,$C1,$23
	.dc.b	$E1,$36,$3D,$CD,$7D,$4D,$7C,$B1,$C0,$8F,$06,$BC,$1A,$F0,$34,$43
	.dc.b	$45,$34,$60,$86,$08,$7C,$23,$D8,$F6,$71,$1E,$76,$C6,$5F,$8A,$D7
	.dc.b	$89,$E1,$36,$53,$65,$36,$79,$0D,$94,$D9,$4D,$9E,$83,$25,$32,$72
	.dc.b	$31,$1E,$97,$66,$48,$60,$87,$22,$BC,$5C,$11,$C1,$23,$E1,$36,$3D
	.dc.b	$CD,$7D,$4D,$7C,$B1,$C0,$8F,$06,$BC,$1A,$F0,$34,$43,$45,$34,$60
	.dc.b	$86,$08,$7C,$23,$D8,$F6,$71,$1E,$76,$C6,$5F,$8A,$D7,$89,$E1,$36
	.dc.b	$53,$65,$3E,$57,$95,$FC,$A0,$8F,$F2,$86,$BF,$CA,$1A,$FF,$66,$6B
	.dc.b	$FC,$59,$AF,$F3,$46,$A9,$46,$1B,$1D,$C8,$77,$29,$DC,$A7,$72,$9D
	.dc.b	$CA,$77,$8E,$C4,$73,$1E,$07,$B1,$0E,$C5,$3B,$14,$EC,$53,$B1,$D2
	.dc.b	$3B,$39,$8E,$CF,$42,$1D,$0A,$74,$29,$D0,$ED,$FA,$47,$FB,$C9,$34
	.dc.b	$8F,$FB,$D9,$AF,$FB,$E9,$AF,$F9,$9B,$8B,$FC,$EA,$7F,$9E,$5F,$F4
	.dc.b	$11,$FE,$77,$E5,$FF,$14,$F4,$FF,$A2,$63,$FD,$26,$7F,$A3,$4D,$CF
	.dc.b	$18,$FF,$55,$F4,$7F,$A0,$7A,$3F,$CF,$F5,$FE,$BA,$8F,$F5,$BD,$7F
	.dc.b	$AC,$5F,$F5,$4C,$7F,$A4,$AB,$8B,$FE,$90,$3F,$D2,$E7,$FA,$95,$F1
	.dc.b	$1F,$EA,$29,$F4,$D7,$FA,$52,$77,$1F,$EA,$37,$91,$E8,$77,$7F,$E9
	.dc.b	$89,$E1,$39,$7C,$F5,$FE,$B6,$72,$F7,$73,$7F,$EA,$59,$F4,$3E,$AB
	.dc.b	$F6,$47,$FA,$B5,$AF,$A6,$3E,$25,$FC,$C6,$C7,$FB,$00,$F9,$97,$F0
	.dc.b	$9E,$A7,$FB,$36,$D5,$FE,$A9,$E7,$FF,$4E,$74,$7F,$AA,$B1,$FF,$4E
	.dc.b	$F3,$FF,$82,$5F,$F5,$26,$1F,$E0,$B3,$FD,$58,$57,$2E,$E7,$FD,$58
	.dc.b	$FF,$A9,$0F,$F3,$4C,$74,$6C,$7F,$AE,$34,$FF,$4C,$47,$F9,$F7,$8F
	.dc.b	$FD,$31,$0E,$5F,$F4,$B7,$3F,$F8,$61,$FF,$53,$D7,$F9,$FC,$FF,$41
	.dc.b	$1F,$E8,$69,$E6,$48,$FF,$43,$AE,$6F,$F4,$3F,$3F,$FA,$80,$FF,$4A
	.dc.b	$57,$1A,$DC,$EA,$FC,$18,$7F,$AD,$0F,$F4,$C5,$F0,$BF,$F4,$F9,$FE
	.dc.b	$A7,$7F,$D4,$A4,$9F,$F5,$09,$FE,$B9,$7F,$D3,$D3,$EE,$BF,$6D,$7F
	.dc.b	$AD,$4F,$14,$64,$EF,$5C,$4F,$F5,$64,$7F,$B3,$23,$C8,$FC,$3C,$3D
	.dc.b	$CF,$23,$FD,$6C,$6C,$FF,$57,$CF,$91,$F0,$1F,$EA,$79,$FF,$51,$5A
	.dc.b	$7F,$D4,$13,$FE,$A2,$8C,$3F,$EA,$CE,$7F,$F2,$74,$62,$9C,$EE,$DC
	.dc.b	$E1,$CB,$0E,$D3,$FE,$A7,$BE,$0D,$9F,$F5,$7D,$F2,$BE,$37,$FE,$93
	.dc.b	$BF,$0B,$FE,$81,$6E,$3F,$E9,$16,$F9,$3C,$27,$58,$7F,$98,$3A,$71
	.dc.b	$7B,$8F,$F3,$1D,$62,$65,$7B,$E7,$83,$4F,$F4,$2C,$E9,$FF,$67,$1D
	.dc.b	$8F,$34,$74,$78,$19,$25,$78,$5E,$66,$9C,$3A,$AF,$3D,$77,$BF,$7C
	.dc.b	$5B,$ED,$5E,$2B,$E2,$49,$E3,$1D,$5E,$0F,$98,$F0,$9C,$5C,$BE,$6B
	.dc.b	$B3,$CE,$3C,$D1,$89,$CB,$63,$B1,$A2,$9D,$4B,$7F,$59,$F8,$1F,$3F
	.dc.b	$FE,$AE,$23,$FE,$AE,$6B,$FE,$AE,$6B,$FE,$EC,$6B,$FE,$9C,$6B,$FE
	.dc.b	$BA,$6B,$FE,$07,$6D,$9A,$48,$F0,$10,$F0,$14,$F0,$14,$F0,$14,$F0
	.dc.b	$14,$F0,$16,$3C,$0D,$8F,$03,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$B1
	.dc.b	$D9,$B1,$D9,$E8,$43,$A1,$4E,$85,$3A,$16,$3A,$36,$3A,$3C,$88,$72
	.dc.b	$29,$C8,$B1,$C9,$B1,$C9,$E0,$43,$81,$63,$83,$63,$83,$A2,$46,$9B
	.dc.b	$1A,$8C,$11,$C1,$23,$EA,$10,$F9,$25,$3E,$01,$4F,$49,$4F,$21,$4F
	.dc.b	$09,$63,$C0,$D8,$F0,$3D,$88,$76,$29,$D8,$A7,$62,$9D,$8B,$1D,$9B
	.dc.b	$1D,$9E,$84,$3A,$14,$E8,$53,$A1,$63,$A3,$63,$A3,$C8,$87,$22,$9C
	.dc.b	$8B,$1C,$9B,$1C,$9E,$04,$38,$16,$38,$36,$38,$3A,$24,$69,$B1,$A8
	.dc.b	$C1,$1C,$12,$3E,$A1,$0F,$92,$53,$E0,$14,$F4,$94,$F2,$14,$F0,$96
	.dc.b	$3C,$0D,$8F,$03,$D8,$FF,$17,$1F,$E9,$DF,$62,$A3,$06,$25,$1A,$8C
	.dc.b	$9C,$A3,$2E,$E3,$51,$88,$D4,$60,$8E,$C8,$68,$A7,$F9,$0E,$39,$36
	.dc.b	$38,$BB,$21,$93,$89,$EB,$3C,$E6,$4D,$99,$8E,$AD,$8E,$AE,$C8,$64
	.dc.b	$EA,$7C,$78,$E2,$D8,$E2,$EC,$EE,$64,$E2,$79,$0D,$1D,$8D,$1C,$0D
	.dc.b	$18,$34,$60,$91,$E2,$6C,$78,$9E,$07,$62,$18,$3F,$01,$D0,$F4,$7F
	.dc.b	$D6,$84,$8F,$F5,$A3,$5F,$C4,$C6,$DC,$90,$FF,$5C,$73,$FF,$A4,$9F
	.dc.b	$F3,$AB,$5F,$F4,$4F,$4F,$FA,$4F,$BB,$EB,$75,$7F,$A2,$F9,$FF,$D2
	.dc.b	$67,$FA,$96,$BF,$D1,$A7,$FA,$54,$FF,$53,$C7,$FA,$7F,$AB,$FD,$2D
	.dc.b	$18,$BF,$F5,$CD,$F4,$7B,$46,$8E,$47,$4B,$EC,$78,$27,$84,$7A,$2B
	.dc.b	$FD,$64,$47,$FD,$8C,$64,$A7,$FB,$4C,$FF,$53,$4F,$FA,$E9,$CB,$FE
	.dc.b	$D1,$AF,$1C,$7F,$B5,$30,$FF,$5C,$C7,$A2,$3C,$D3,$E4,$BF,$1C,$7A
	.dc.b	$23,$CC,$FF,$B2,$A3,$31,$D6,$3F,$D7,$91,$83,$35,$DA,$3D,$18,$7F
	.dc.b	$B2,$70,$F6,$36,$0B,$0F,$F5,$87,$57,$FA,$E8,$93,$FE,$B1,$18,$FF
	.dc.b	$AC,$5F,$F5,$C9,$C0,$FF,$7D,$74,$FF,$AE,$1F,$F5,$E3,$FE,$C3,$3F
	.dc.b	$D6,$DD,$1F,$EA,$48,$FF,$46,$70,$FF,$A7,$5F,$F4,$63,$FE,$A3,$3F
	.dc.b	$D3,$AF,$FA,$40,$70,$7F,$AE,$F9,$BF,$D3,$71,$FB,$5E,$9F,$F5,$73
	.dc.b	$FE,$92,$AE,$27,$57,$8D,$FF,$A9,$9B,$5F,$EA,$09,$FF,$5D,$73,$7F
	.dc.b	$A9,$67,$FD,$64,$FF,$A7,$5F,$F6,$29,$FE,$B6,$3F,$D9,$C7,$97,$0F
	.dc.b	$23,$FE,$A9,$C3,$C6,$F6,$3F,$D4,$E7,$43,$DF,$3D,$DE,$75,$E7,$AE
	.dc.b	$6F,$92,$3D,$C7,$C8,$9F,$24,$FF,$AD,$A3,$C8,$FF,$B5,$EF,$FD,$81
	.dc.b	$1E,$33,$81,$E0,$39,$9F,$E6,$B3,$FD,$01,$E2,$7F,$A7,$0F,$F3,$93
	.dc.b	$83,$FD,$18,$FF,$9E,$BB,$7F,$D4,$3E,$24,$9F,$F4,$84,$7F,$A9,$0C
	.dc.b	$3F,$E8,$B8,$C3,$A3,$FD,$2C,$EB,$1F,$F5,$8C,$EA,$78,$1A,$38,$1F
	.dc.b	$E9,$57,$86,$3F,$CF,$6B,$85,$72,$AE,$07,$23,$81,$C9,$FF,$5F,$C7
	.dc.b	$28,$FE,$7D,$5C,$AB,$A4,$72,$AE,$8F,$23,$A3,$C9,$E9,$1F,$D3,$E7
	.dc.b	$A4,$7E,$62,$BA,$3D,$8E,$93,$D9,$E9,$5D,$8E,$8F,$68,$E8,$F6,$3F
	.dc.b	$AA,$4F,$6C,$57,$DA,$7C,$07,$63,$C0,$7F,$AF,$7E,$53,$FB,$C9,$1F
	.dc.b	$F0,$13,$5F,$F0,$03,$5F,$EE,$CD,$7F,$B6,$35,$FE,$CC,$D7,$F6,$ED
	.dc.b	$93,$49,$1E,$02,$1E,$02,$9E,$02,$9E,$02,$9E,$02,$9E,$02,$C7,$81
	.dc.b	$B1,$E0,$7B,$10,$EC,$53,$B1,$4E,$C5,$3B,$16,$3B,$36,$3B,$3D,$08
	.dc.b	$74,$29,$D0,$A7,$42,$C7,$46,$C7,$47,$91,$0E,$45,$39,$16,$39,$36
	.dc.b	$39,$3C,$08,$70,$2C,$70,$6C,$70,$74,$48,$D3,$63,$51,$82,$38,$24
	.dc.b	$7D,$42,$1F,$24,$A7,$C0,$29,$E9,$29,$E4,$29,$E1,$2C,$78,$1B,$1E
	.dc.b	$07,$B1,$0E,$C5,$3B,$14,$EC,$53,$B1,$63,$B3,$63,$B3,$D0,$87,$42
	.dc.b	$9D,$0A,$74,$2C,$74,$6C,$74,$79,$10,$E4,$53,$91,$63,$93,$63,$93
	.dc.b	$C0,$87,$02,$C7,$06,$C7,$07,$44,$8D,$36,$35,$18,$23,$82,$47,$D4
	.dc.b	$21,$F2,$4A,$7C,$02,$9E,$92,$9E,$42,$9E,$12,$C7,$81,$B1,$E0,$7B
	.dc.b	$1F,$EA,$6F,$15,$F1,$9F,$EA,$16,$C7,$FA,$A1,$AE,$E3,$44,$74,$D7
	.dc.b	$46,$08,$60,$87,$FA,$64,$47,$67,$0D,$7F,$D3,$CD,$7F,$D7,$4D,$78
	.dc.b	$C7,$85,$E2,$D7,$8B,$5E,$2D,$78,$9E,$63,$89,$4E,$25,$38,$94,$E2
	.dc.b	$7B,$0D,$94,$D9,$4D,$9D,$8C,$47,$AC,$91,$E5,$6C,$77,$6B,$C2,$34
	.dc.b	$7B,$7C,$87,$92,$F8,$6F,$F7,$0C,$61,$B1,$9A,$FF,$73,$CE,$5B,$19
	.dc.b	$7E,$B4,$65,$B1,$97,$FD,$A5,$19,$7E,$E7,$1F,$FA,$17,$9B,$FC,$E0
	.dc.b	$4E,$6F,$F6,$C4,$65,$BD,$1F,$EA,$98,$C5,$7F,$9E,$A7,$7D,$3F,$E7
	.dc.b	$F8,$FF,$42,$10,$FF,$3D,$C7,$FA,$35,$FF,$48,$74,$FF,$A0,$23,$AB
	.dc.b	$FE,$90,$9E,$53,$C0,$EC,$43,$FD,$50,$FF,$A3,$23,$94,$66,$BF,$D2
	.dc.b	$AF,$13,$FD,$49,$87,$FA,$2D,$FF,$5B,$1B,$C3,$D8,$4A,$F1,$44,$39
	.dc.b	$57,$FA,$75,$FF,$51,$5F,$9A,$FF,$D8,$51,$FE,$B0,$BE,$4D,$BE,$53
	.dc.b	$FE,$CD,$AE,$18,$7C,$38,$C5,$6A,$30,$7F,$AD,$A3,$31,$F2,$0F,$F5
	.dc.b	$DD,$AF,$FD,$6F,$D1,$FE,$9B,$E9,$FF,$52,$F8,$CF,$13,$FD,$45,$87
	.dc.b	$FA,$82,$7F,$78,$8F,$F5,$39,$74,$5A,$77,$3F,$EA,$E8,$EC,$FF,$AC
	.dc.b	$67,$FD,$61,$C7,$FE,$7D,$8F,$F4,$87,$5F,$FA,$6E,$3F,$D0,$FC,$FF
	.dc.b	$E8,$B7,$5D,$1F,$E8,$B9,$D1,$BE,$BF,$F4,$AC,$F0,$BF,$F4,$1D,$7F
	.dc.b	$AA,$2F,$FD,$1B,$5A,$BE,$15,$FE,$8A,$3F,$D3,$95,$FE,$95,$7F,$D4
	.dc.b	$AF,$17,$FD,$1C,$62,$BF,$D2,$F3,$CA,$3F,$D2,$EF,$FB,$0D,$BD,$7F
	.dc.b	$E9,$8A,$E9,$87,$FA,$75,$E9,$3E,$0B,$F9,$8D,$3D,$2F,$6A,$FF,$5E
	.dc.b	$3A,$AF,$74,$F8,$CF,$F5,$91,$C4,$8F,$78,$E5,$1F,$EB,$83,$FD,$6A
	.dc.b	$FC,$47,$E1,$47,$42,$61,$F1,$EB,$F0,$17,$FE,$DB,$BF,$17,$83,$E8
	.dc.b	$F1,$BF,$A0,$60,$8C,$63,$A3,$F1,$13,$A8,$CC,$7E,$2E,$78,$3A,$38
	.dc.b	$09,$E0,$E4,$E1,$3C,$9F,$F7,$5F,$5F,$E4,$1A,$7E,$4A,$BF,$D1,$6D
	.dc.b	$8D,$46,$23,$F3,$AD,$7D,$71,$A2,$3C,$48,$70,$72,$D3,$27,$22,$9F
	.dc.b	$EA,$03,$44,$34,$53,$46,$08,$60,$87,$FA,$88,$C9,$4C,$9F,$08,$E8
	.dc.b	$64,$86,$4E,$25,$38,$C7,$C3,$6C,$7B,$9D,$C6,$5E,$6D,$79,$18,$8E
	.dc.b	$0F,$23,$E1,$99,$3E,$B7,$8D,$F5,$08,$E0,$87,$D6,$72,$D3,$27,$E4
	.dc.b	$9C,$B4,$C9,$FA,$97,$2D,$32,$7E,$4D,$CB,$4C,$89,$46,$A3,$0D,$8C
	.dc.b	$C7,$06,$BE,$72,$1E,$52,$9E,$32,$9E,$12,$9D,$8E,$91,$DC,$8E,$08
	.dc.b	$70,$8F,$0B,$DC,$87,$72,$9D,$CA,$77,$39,$47,$66,$9D,$8D,$C7,$67
	.dc.b	$A1,$0E,$85,$3A,$1C,$23,$A3,$4E,$86,$E3,$A3,$C8,$87,$23,$51,$C9
	.dc.b	$A7,$23,$71,$C9,$E0,$62,$38,$34,$E0,$68,$91,$C4,$91,$C2,$32,$D8
	.dc.b	$CC,$72,$6B,$F2,$08,$7B,$CA,$7A,$0A,$78,$CA,$76,$3A,$47,$72,$38
	.dc.b	$21,$C2,$3C,$2F,$72,$1D,$CA,$77,$29,$DC,$E5,$1D,$9A,$76,$37,$1D
	.dc.b	$9E,$84,$3A,$14,$E8,$70,$8E,$8D,$3A,$1B,$8E,$8F,$22,$1C,$8D,$47
	.dc.b	$26,$9C,$8D,$C7,$27,$81,$88,$E0,$D3,$81,$A2,$47,$12,$47,$08,$CB
	.dc.b	$63,$31,$C9,$AF,$C8,$21,$EF,$29,$E8,$29,$EE,$EF,$FE,$B8,$9A,$47
	.dc.b	$FB,$03,$5F,$F1,$8B,$5F,$F1,$CB,$5F,$F5,$B3,$5F,$F0,$8B,$5F,$F1
	.dc.b	$F3,$5F,$01,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$A7,$62,$9D,$8E,$84
	.dc.b	$3A,$14,$E8,$53,$A1,$4E,$85,$3A,$1C,$88,$72,$29,$C8,$A7,$22,$9C
	.dc.b	$8E,$04,$38,$14,$E0,$5E,$8F,$F5,$CF,$C4,$FF,$AD,$89,$1F,$EB,$46
	.dc.b	$BF,$EB,$26,$BF,$EB,$2E,$A4,$7F,$87,$78,$7F,$D5,$72,$69,$39,$FF
	.dc.b	$D1,$7D,$1F,$E9,$99,$FF,$50,$9F,$EA,$D7,$FD,$22,$66,$7F,$D4,$2E
	.dc.b	$3A,$7F,$D6,$B3,$FE,$8D,$3F,$D4,$1D,$35,$FF,$4D,$96,$BF,$D3,$C3
	.dc.b	$AB,$FD,$5D,$1F,$E9,$09,$FF,$57,$BE,$7C,$3C,$B7,$DA,$3F,$D2,$D8
	.dc.b	$7F,$B2,$DF,$F5,$A5,$FF,$A8,$1F,$09,$FE,$A5,$24,$7A,$63,$51,$FE
	.dc.b	$B8,$9E,$93,$FE,$A2,$23,$FE,$A5,$24,$7F,$A6,$67,$CF,$3E,$18,$F3
	.dc.b	$5F,$B6,$73,$5F,$ED,$09,$F0,$BD,$8E,$6F,$C4,$3D,$6E,$EB,$CD,$1F
	.dc.b	$1E,$FF,$D6,$4F,$FB,$1B,$9F,$FD,$4B,$D5,$FE,$B1,$8F,$F5,$64,$7F
	.dc.b	$AA,$DF,$F5,$43,$5F,$F5,$A1,$FE,$B4,$CB,$FD,$63,$D3,$FE,$B0,$6C
	.dc.b	$66,$7A,$F5,$7F,$BD,$9E,$97,$DE,$7F,$D0,$67,$FA,$7B,$87,$FD,$10
	.dc.b	$FF,$A3,$1F,$F4,$27,$57,$FA,$7F,$A3,$FD,$33,$18,$E9,$FF,$4A,$D7
	.dc.b	$F9,$EE,$B5,$D3,$FE,$90,$BF,$F4,$44,$F3,$9F,$F5,$CD,$70,$8F,$F4
	.dc.b	$BF,$57,$FA,$3D,$FF,$4B,$61,$FE,$C0,$C3,$FD,$2D,$1F,$EA,$C3,$FD
	.dc.b	$8F,$5F,$E9,$48,$FF,$52,$5F,$82,$7B,$4E,$5F,$23,$DE,$BF,$D3,$B1
	.dc.b	$FE,$A0,$9F,$1B,$E0,$7F,$D7,$58,$7B,$9C,$BE,$A8,$F8,$E7,$FA,$BA
	.dc.b	$3C,$0E,$27,$E2,$3C,$EB,$FD,$85,$7E,$67,$C5,$5C,$2B,$FD,$43,$4B
	.dc.b	$5F,$FA,$89,$B1,$FE,$A0,$3D,$87,$F9,$4F,$C4,$FF,$64,$C6,$A3,$FC
	.dc.b	$AD,$D5,$8A,$E4,$48,$FF,$58,$56,$5F,$F5,$97,$89,$FE,$A4,$23,$FE
	.dc.b	$A9,$6B,$FE,$A6,$6B,$FE,$C0,$6B,$FE,$D5,$6B,$D0,$E4,$43,$91,$4E
	.dc.b	$45,$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21
	.dc.b	$82,$1F,$E8,$17,$DA,$D3,$CC,$53,$C2,$53,$B1,$4E,$A5,$3A,$E0,$8E
	.dc.b	$AD,$8E,$AD,$7A,$B5,$FF,$78,$F8,$9F,$ED,$C2,$3F,$EE,$16,$BF,$ED
	.dc.b	$E6,$BC,$0D,$10,$D1,$4D,$18,$21,$82,$03,$FD,$F8,$F6,$70,$7F,$AA
	.dc.b	$5C,$3A,$3B,$1A,$84,$6A,$38,$1A,$38,$1E,$23,$83,$B8,$E3,$1B,$38
	.dc.b	$9B,$38,$9E,$82,$9C,$8E,$0F,$27,$81,$C8,$E0,$53,$81,$A7,$83,$A6
	.dc.b	$BA,$31,$1E,$72,$3E,$06,$BC,$DA,$F0,$34,$43,$45,$34,$60,$86,$08
	.dc.b	$78,$C8,$7C,$22,$1E,$D2,$C7,$A9,$B1,$E4,$78,$9B,$24,$6D,$B1,$B7
	.dc.b	$89,$E7,$38,$C6,$C9,$1B,$8E,$26,$CE,$27,$B0,$E2,$EE,$38,$C6,$CE
	.dc.b	$26,$CE,$27,$C2,$29,$C8,$E0,$F2,$78,$1C,$8E,$05,$38,$1A,$78,$3A
	.dc.b	$6B,$A3,$11,$E7,$23,$E0,$6B,$CD,$AF,$03,$44,$34,$53,$46,$08,$60
	.dc.b	$87,$8C,$87,$C2,$21,$ED,$2C,$7A,$9B,$1E,$47,$89,$B2,$46,$DB,$1B
	.dc.b	$78,$9E,$73,$8C,$6C,$91,$B8,$E2,$6C,$E2,$7B,$0E,$2E,$E3,$8C,$6C
	.dc.b	$E2,$7D,$AE,$A7,$89,$FE,$A8,$24,$7F,$AA,$1A,$FF,$AA,$1A,$FF,$AA
	.dc.b	$1A,$FF,$AA,$1A,$FF,$AA,$1A,$FF,$AD,$DB,$1E,$08,$EC,$43,$B1,$4E
	.dc.b	$C5,$3B,$14,$EC,$53,$B1,$5E,$CF,$42,$1D,$0A,$74,$29,$D0,$A7,$42
	.dc.b	$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1
	.dc.b	$75,$BA,$DE,$FF,$F4,$E9,$FF,$54,$35,$FE,$A9,$3B,$8C,$47,$29,$E2
	.dc.b	$EE,$74,$D8,$CB,$D6,$32,$D8,$CB,$E1,$E5,$FF,$59,$75,$FF,$A2,$63
	.dc.b	$FD,$61,$D5,$FE,$83,$23,$FE,$91,$3F,$D0,$91,$FE,$A6,$9E,$31,$FE
	.dc.b	$8E,$34,$7F,$A3,$0F,$F5,$29,$C0,$FF,$57,$13,$AB,$FD,$35,$1C,$9F
	.dc.b	$F4,$8C,$71,$3F,$D3,$A7,$53,$FD,$8A,$1F,$F6,$13,$E0,$7F,$D5,$2D
	.dc.b	$AF,$4B,$CA,$3A,$9A,$7F,$D4,$4D,$9F,$F6,$24,$7C,$23,$FD,$77,$1C
	.dc.b	$27,$FD,$53,$3A,$9F,$14,$76,$8F,$53,$FE,$AF,$8F,$11,$C9,$C3,$E0
	.dc.b	$79,$F5,$FF,$AE,$23,$E3,$9F,$06,$7F,$DC,$FD,$1F,$EA,$48,$FF,$57
	.dc.b	$F5,$7F,$86,$7A,$7F,$D4,$95,$FE,$AA,$9F,$F5,$79,$0F,$F5,$54,$FF
	.dc.b	$AB,$EB,$94,$7F,$BC,$DF,$F5,$6F,$57,$FA,$87,$97,$FC,$39,$D5,$FE
	.dc.b	$9D,$E8,$FF,$45,$72,$7F,$A7,$70,$FF,$58,$4E,$6B,$71,$FE,$95,$EA
	.dc.b	$FF,$3F,$D7,$FA,$62,$3F,$D4,$57,$FE,$7F,$8F,$F4,$7B,$D3,$AB,$FD
	.dc.b	$33,$7E,$1A,$FF,$1A,$57,$FA,$F1,$FF,$4B,$4F,$FA,$72,$7B,$C7,$47
	.dc.b	$FD,$5F,$5F,$EA,$1B,$F0,$D7,$FA,$5C,$EB,$5F,$EA,$E3,$FD,$31,$5F
	.dc.b	$E9,$D7,$FD,$40,$7F,$AD,$8C,$9F,$EB,$23,$35,$E9,$9F,$14,$79,$B0
	.dc.b	$F8,$4D,$3E,$11,$C8,$FF,$63,$98,$9E,$57,$FE,$AC,$24,$76,$BF,$59
	.dc.b	$F0,$D6,$1F,$ED,$EE,$3F,$E1,$5F,$15,$FC,$11,$B3,$FC,$06,$31,$C8
	.dc.b	$69,$23,$44,$3F,$89,$60,$FD,$FB,$FC,$50,$DE,$9F,$E2,$2D,$9F,$E2
	.dc.b	$CD,$78,$CF,$02,$4F,$06,$C7,$06,$BC,$23,$44,$74,$D7,$46,$08,$60
	.dc.b	$87,$F9,$E2,$7C,$8D,$9E,$ED,$9E,$8D,$39,$94,$E6,$7F,$A1,$67,$9B
	.dc.b	$67,$9B,$67,$9B,$4E,$65,$39,$89,$E6,$D9,$E6,$D9,$E6,$D3,$99,$4E
	.dc.b	$73,$C8,$93,$C9,$B3,$C9,$EF,$3F,$8A,$F9,$FF,$E9,$6D,$3F,$A6,$E4
	.dc.b	$8C,$B6,$33,$18,$23,$82,$1F,$D6,$63,$83,$63,$6D,$36,$7F,$61,$8D
	.dc.b	$B6,$36,$D3,$67,$FA,$BD,$CB,$4C,$9F,$C2,$1C,$B4,$C9,$EF,$3C,$44
	.dc.b	$3B,$14,$E8,$53,$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14
	.dc.b	$D1,$82,$18,$24,$7C,$87,$83,$5D,$B5,$DC,$7C,$46,$9F,$12,$34,$48
	.dc.b	$D3,$63,$51,$82,$38,$21,$F1,$63,$83,$63,$6D,$36,$7C,$48,$DB,$63
	.dc.b	$6D,$36,$7B,$9C,$B4,$C9,$F0,$1C,$B4,$C9,$F0,$0F,$11,$0E,$C5,$3A
	.dc.b	$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86
	.dc.b	$09,$1F,$21,$E0,$D7,$6D,$77,$1F,$11,$A7,$C4,$8D,$12,$34,$D8,$D4
	.dc.b	$60,$8E,$08,$7C,$58,$E0,$D8,$DB,$4D,$9F,$12,$36,$D8,$DB,$4D,$9E
	.dc.b	$E7,$2D,$32,$7C,$07,$2D,$32,$7C,$03,$C4,$43,$B1,$4E,$85,$3E,$8F
	.dc.b	$4B,$E3,$7F,$94,$B6,$3F,$CE,$AD,$7F,$CF,$4D,$7F,$CF,$CD,$44,$70
	.dc.b	$D8,$EA,$47,$AB,$5E,$AD,$7A,$B5,$EA,$23,$AB,$63,$4F,$73,$A9,$0E
	.dc.b	$A5,$3A,$94,$EB,$1D,$09,$1D,$23,$B1,$B3,$B1,$D0,$87,$42,$9D,$1E
	.dc.b	$51,$D2,$39,$1D,$0D,$9D,$0E,$44,$39,$1C,$1E,$4F,$03,$91,$B3,$91
	.dc.b	$C0,$EB,$A7,$FA,$DF,$E1,$FF,$C5,$64,$8F,$F5,$73,$5F,$F5,$63,$5F
	.dc.b	$A3,$1C,$5F,$F5,$CF,$2F,$FA,$11,$FF,$41,$9F,$E8,$12,$AF,$01,$D7
	.dc.b	$FE,$AD,$6C,$FF,$A8,$67,$FD,$21,$D7,$FE,$97,$EA,$FF,$44,$4F,$FA
	.dc.b	$77,$AF,$FD,$1B,$5A,$7F,$D2,$97,$CA,$FF,$D3,$55,$E0,$8C,$4F,$FA
	.dc.b	$66,$FF,$13,$5D,$6B,$FD,$44,$75,$3F,$D2,$AE,$5F,$F5,$09,$FE,$9D
	.dc.b	$3D,$71,$E4,$BF,$81,$1F,$EB,$89,$FF,$5B,$B6,$BC,$B8,$7F,$B4,$63
	.dc.b	$99,$DC,$95,$EB,$7D,$11,$F1,$4E,$15,$D2,$75,$3F,$34,$8E,$0F,$9D
	.dc.b	$3C,$63,$E6,$5F,$C0,$BF,$F5,$9C,$7C,$0A,$DB,$FE,$D7,$8F,$8C,$FF
	.dc.b	$B6,$ED,$1F,$EA,$E8,$75,$FF,$AD,$3A,$BF,$D5,$F7,$FE,$B5,$9D,$1F
	.dc.b	$EB,$53,$FD,$76,$7F,$C0,$15,$1C,$0F,$F5,$AC,$7F,$A6,$79,$FF,$D1
	.dc.b	$2F,$FA,$1C,$FF,$59,$9F,$E8,$DE,$7F,$F4,$E4,$7F,$AB,$78,$FF,$D4
	.dc.b	$64,$9E,$07,$FA,$33,$A7,$FD,$36,$FF,$A3,$BA,$7F,$D3,$6D,$3A,$1F
	.dc.b	$EA,$3A,$E8,$FF,$A9,$DF,$19,$82,$DF,$88,$FF,$5B,$BF,$EB,$58,$FF
	.dc.b	$52,$72,$FF,$B1,$63,$07,$08,$FF,$55,$1F,$EB,$D3,$DA,$53,$D1,$87
	.dc.b	$AE,$3C,$15,$FE,$A8,$9F,$8C,$F8,$5E,$71,$DA,$31,$1E,$D3,$A5,$7F
	.dc.b	$B5,$23,$CE,$7F,$AC,$5F,$F6,$91,$FE,$E1,$8F,$09,$EA,$C3,$FD,$9D
	.dc.b	$7E,$3A,$EB,$5F,$EB,$7E,$97,$C2,$FF,$AD,$09,$87,$F3,$63,$FA,$1F
	.dc.b	$1A,$77,$19,$11,$99,$E3,$1C,$63,$0E,$DA,$FF,$AF,$3F,$7A,$7A,$BF
	.dc.b	$EB,$A6,$C7,$FA,$E9,$AE,$E3,$44,$74,$D7,$46,$08,$60,$87,$FA,$CE
	.dc.b	$3A,$B6,$39,$34,$E2,$53,$89,$FE,$B1,$8E,$2D,$8E,$2D,$38,$94,$E2
	.dc.b	$7F,$AB,$E3,$8B,$63,$8B,$4E,$25,$38,$E4,$9E,$2D,$9E,$2D,$9E,$2D
	.dc.b	$78,$CF,$02,$4F,$06,$C7,$06,$BC,$23,$44,$7F,$D6,$BF,$41,$FE,$B4
	.dc.b	$23,$82,$1F,$EB,$17,$2D,$32,$7F,$AB,$DC,$B4,$C9,$FE,$08,$C9,$1A
	.dc.b	$8C,$35,$FF,$0C,$C6,$9E,$26,$23,$4A,$34,$F8,$C8,$78,$4A,$76,$39
	.dc.b	$9C,$63,$9C,$72,$8E,$04,$38,$1C,$8F,$11,$C9,$E0,$E8,$E0,$68,$E0
	.dc.b	$77,$38,$47,$80,$8E,$08,$70,$8E,$AF,$33,$A9,$CC,$D4,$72,$69,$C8
	.dc.b	$DC,$72,$78,$18,$8E,$0D,$38,$1E,$B2,$1E,$12,$9C,$8B,$1D,$49,$1D
	.dc.b	$23,$2D,$79,$10,$E4,$58,$E2,$F8,$8C,$47,$17,$64,$8D,$BF,$00,$87
	.dc.b	$AC,$A7,$88,$E8,$71,$8E,$91,$CA,$38,$10,$E0,$72,$3C,$67,$27,$83
	.dc.b	$A3,$81,$A3,$81,$DC,$E1,$1E,$02,$38,$21,$C2,$3A,$BC,$CE,$A7,$33
	.dc.b	$51,$C9,$A7,$23,$71,$C9,$E0,$62,$38,$34,$E0,$7A,$C8,$78,$4A,$72
	.dc.b	$2C,$75,$24,$74,$8C,$B5,$E4,$43,$91,$63,$8B,$E2,$31,$1C,$5D,$92
	.dc.b	$36,$FC,$02,$1E,$B2,$9E,$23,$A1,$C6,$3A,$47,$28,$E0,$43,$81,$F4
	.dc.b	$FE,$43,$FB,$43,$FC,$59,$AF,$F1,$36,$C1,$A4,$8E,$31,$96,$C6,$5F
	.dc.b	$EE,$91,$96,$C6,$5F,$EF,$91,$96,$C6,$5F,$E5,$CD,$7C,$C7,$62,$1D
	.dc.b	$0A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68
	.dc.b	$C1,$0C,$10,$F9,$C7,$90,$A7,$72,$9D,$0A,$73,$29,$CE,$3E,$53,$96
	.dc.b	$AF,$92,$7E,$A0,$F7,$3F,$03,$18,$6C,$66,$7F,$C5,$93,$96,$C6,$5F
	.dc.b	$F1,$A4,$65,$B1,$97,$FD,$DB,$19,$7F,$C7,$5C,$9F,$E9,$E3,$FD,$1D
	.dc.b	$D1,$FE,$A1,$EA,$FF,$56,$CF,$FA,$90,$FF,$46,$38,$7F,$D3,$B1,$FE
	.dc.b	$91,$7F,$D3,$A7,$FA,$3C,$FF,$57,$F4,$7F,$A5,$1F,$F5,$2D,$7F,$A5
	.dc.b	$3A,$7F,$D2,$31,$A3,$8B,$87,$A9,$FE,$97,$9F,$F4,$CB,$CA,$BC,$DD
	.dc.b	$5F,$EA,$A9,$FF,$59,$CF,$FA,$90,$E0,$75,$3F,$D5,$73,$C1,$D1,$CC
	.dc.b	$F6,$CF,$FA,$B6,$BF,$D9,$F1,$C0,$C9,$ED,$3C,$A7,$13,$DB,$87,$C5
	.dc.b	$C3,$E6,$5F,$C5,$9E,$46,$AF,$B6,$1F,$EE,$09,$F2,$61,$E0,$BF,$F6
	.dc.b	$06,$1C,$CE,$93,$CE,$33,$6C,$3E,$6F,$4F,$FA,$C7,$C3,$C7,$5F,$FA
	.dc.b	$ED,$FF,$7D,$C1,$A4,$9F,$F5,$D4,$71,$7F,$D8,$7E,$1F,$FA,$2F,$B3
	.dc.b	$FD,$5B,$C3,$FE,$8C,$EB,$FF,$58,$BF,$EA,$3E,$BD,$5F,$FA,$C0,$FF
	.dc.b	$47,$F6,$7F,$A7,$9F,$F5,$1B,$FE,$9B,$7F,$D6,$91,$FE,$98,$7F,$D4
	.dc.b	$26,$0F,$F5,$0B,$D1,$FF,$4F,$D7,$82,$FC,$64,$8F,$05,$F8,$DE,$67
	.dc.b	$8A,$BB,$34,$F1,$BD,$6F,$C9,$19,$3F,$D5,$CE,$27,$A4,$65,$FF,$58
	.dc.b	$3E,$63,$BB,$A3,$FD,$60,$69,$E7,$3D,$CF,$F5,$4F,$63,$E4,$BF,$A0
	.dc.b	$F6,$7F,$AA,$DB,$19,$AF,$F5,$33,$5F,$F5,$53,$5F,$C1,$1F,$EB,$07
	.dc.b	$0F,$68,$E6,$7F,$4E,$EC,$7F,$00,$BE,$27,$0D,$AF,$F5,$D4,$EA,$72
	.dc.b	$48,$CC,$ED,$FF,$62,$4E,$DC,$CE,$DC,$CE,$DF,$3C,$ED,$CC,$ED,$CC
	.dc.b	$ED,$FC,$F4,$ED,$CC,$ED,$CC,$ED,$4E,$DC,$CE,$DC,$CE,$E7,$51,$89
	.dc.b	$D4,$60,$93,$B2,$46,$9B,$3F,$00,$87,$B0,$A7,$A0,$A7,$90,$A7,$84
	.dc.b	$B3,$E0,$9E,$C4,$8C,$CF,$81,$EC,$43,$B1,$4F,$C2,$FD,$EB,$2F,$F0
	.dc.b	$6B,$63,$FC,$22,$D7,$FD,$F4,$D7,$FE,$00,$35,$E7,$1C,$88,$F2,$6B
	.dc.b	$C9,$AF,$26,$BC,$8E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04
	.dc.b	$30,$43,$F9,$54,$7B,$5B,$1E,$66,$9E,$12,$9D,$8A,$75,$29,$D4,$47
	.dc.b	$56,$C7,$56,$9D,$4A,$75,$29,$D4,$A7,$58,$E8,$48,$E8,$D8,$E8,$D7
	.dc.b	$A3,$5E,$8D,$7A,$47,$22,$3C,$9A,$F2,$6B,$C9,$AF,$23,$81,$0E,$05
	.dc.b	$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$10,$FA,$64,$3E,$49,$63,$E0
	.dc.b	$36,$3D,$2D,$3C,$85,$3C,$25,$3C,$05,$3C,$07,$62,$47,$66,$C7,$66
	.dc.b	$9D,$8A,$76,$29,$D8,$A7,$68,$E8,$48,$E8,$D8,$E8,$D7,$A3,$5E,$8D
	.dc.b	$7A,$47,$22,$3C,$9A,$F2,$6B,$C9,$AF,$23,$81,$0E,$05,$38,$14,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$10,$FA,$84,$3E,$49,$63,$E0,$36,$3D,$2D
	.dc.b	$3C,$87,$F8,$BF,$C2,$7C,$5F,$FA,$D9,$B1,$FE,$BA,$6B,$FD,$D0,$E1
	.dc.b	$1B,$23,$B6,$BB,$11,$C9,$E4,$D7,$FB,$FB,$5E,$A7,$32,$1C,$DD,$3D
	.dc.b	$23,$BB,$D1,$AF,$46,$BD,$0E,$4E,$1E,$47,$63,$91,$4E,$45,$39,$1A
	.dc.b	$39,$99,$39,$98,$21,$B2,$1D,$0C,$94,$C9,$D8,$C1,$0C,$10,$EE,$70
	.dc.b	$29,$B2,$9B,$8F,$92,$7C,$FD,$1A,$7F,$83,$DB,$97,$F8,$E0,$95,$FE
	.dc.b	$38,$6B,$FE,$38,$6B,$FE,$B2,$E8,$FF,$3E,$BF,$E9,$83,$FD,$1B,$D3
	.dc.b	$FE,$83,$8F,$F4,$17,$3F,$FA,$7E,$7F,$D0,$1D,$1F,$EA,$BE,$AF,$F5
	.dc.b	$2D,$EE,$BF,$D1,$B5,$FE,$B9,$3F,$D2,$07,$FA,$66,$BF,$D2,$AF,$76
	.dc.b	$BF,$E9,$B9,$FF,$59,$47,$58,$FF,$55,$3F,$E9,$07,$95,$FA,$1E,$17
	.dc.b	$FE,$B2,$AF,$F4,$D7,$57,$FA,$72,$3F,$D3,$D3,$FE,$AA,$9E,$B5,$C5
	.dc.b	$F4,$47,$4C,$3D,$CB,$0F,$13,$F1,$2F,$FD,$99,$5C,$AB,$DF,$3B,$C3
	.dc.b	$FD,$4C,$78,$5F,$85,$3B,$7B,$1F,$EC,$27,$C2,$64,$F4,$BE,$7A,$F9
	.dc.b	$A7,$80,$FF,$37,$DF,$90,$74,$7F,$AA,$7A,$BF,$D6,$15,$FE,$AC,$EA
	.dc.b	$FF,$57,$3F,$EF,$09,$FF,$59,$39,$3F,$D5,$E7,$DC,$73,$3C,$89,$D5
	.dc.b	$FE,$89,$9F,$F4,$7F,$57,$F9,$79,$74,$FF,$AB,$DF,$F4,$B4,$66,$31
	.dc.b	$D3,$FE,$93,$3F,$D7,$5D,$1F,$EC,$98,$FF,$20,$46,$7A,$BF,$D5,$97
	.dc.b	$FE,$B0,$AD,$73,$FF,$A3,$AB,$FD,$3D,$87,$FB,$08,$ED,$1F,$EA,$E8
	.dc.b	$FF,$50,$C7,$FA,$DE,$BF,$D6,$AE,$9F,$F5,$41,$E6,$AE,$AF,$FB,$04
	.dc.b	$FF,$53,$9E,$E7,$2E,$0E,$07,$C0,$AF,$24,$F5,$8D,$57,$8A,$B8,$BE
	.dc.b	$86,$CF,$96,$7D,$C7,$33,$C5,$3F,$19,$C1,$E0,$C3,$E5,$CF,$17,$FD
	.dc.b	$8C,$F8,$67,$84,$7C,$C8,$D5,$FD,$9A,$E4,$FA,$2F,$CB,$87,$D0,$BE
	.dc.b	$11,$E6,$25,$78,$B0,$57,$FB,$02,$BF,$CF,$5C,$9F,$C5,$3E,$4B,$FD
	.dc.b	$92,$4B,$FF,$64,$B5,$FF,$64,$B5,$FF,$64,$C7,$17,$FD,$7B,$1F,$EC
	.dc.b	$42,$38,$21,$FE,$BC,$72,$D3,$27,$FA,$E9,$CB,$4C,$9F,$EB,$87,$2D
	.dc.b	$32,$7F,$AD,$9C,$B4,$CE,$1F,$ED,$A8,$CB,$63,$3D,$49,$CB,$63,$31
	.dc.b	$82,$38,$21,$EB,$21,$E9,$29,$E7,$29,$E5,$2C,$F8,$DB,$3E,$16,$CF
	.dc.b	$81,$A7,$C0,$3F,$D0,$FF,$39,$FE,$B6,$23,$FE,$B6,$6B,$FE,$B6,$6B
	.dc.b	$FE,$B6,$6B,$FE,$C0,$6B,$FE,$9E,$6B,$D8,$E8,$43,$A1,$4E,$85,$3A
	.dc.b	$14,$E8,$53,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E
	.dc.b	$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$EB,$47,$C8,$6C,$7B,$DA,$FA
	.dc.b	$1A,$F8,$DA,$F8,$1A,$F7,$6B,$DE,$3B,$11,$EC,$D7,$B3,$5E,$CD,$7B
	.dc.b	$35,$EC,$D7,$B1,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0E
	.dc.b	$45,$39,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34
	.dc.b	$60,$86,$09,$1F,$50,$91,$F2,$5B,$1F,$01,$AF,$A5,$AF,$91,$AF,$85
	.dc.b	$AF,$81,$AF,$82,$3B,$11,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B1
	.dc.b	$D0,$87,$42,$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0E,$45,$39,$14,$E4
	.dc.b	$53,$91,$C0,$FD,$E7,$53,$E2,$BF,$D6,$04,$8F,$F5,$83,$5F,$F5,$03
	.dc.b	$5F,$F5,$4B,$5F,$F5,$4B,$5F,$F5,$E3,$5F,$F5,$8B,$63,$C1,$1D,$88
	.dc.b	$76,$29,$D8,$A7,$62,$9D,$8A,$76,$2B,$D9,$E8,$43,$A1,$4E,$85,$3A
	.dc.b	$14,$E8,$57,$A3,$C8,$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E
	.dc.b	$05,$78,$2C,$BF,$D5,$3F,$21,$FE,$DF,$8C,$76,$23,$13,$A9,$FF,$72
	.dc.b	$35,$FF,$76,$35,$FF,$0C,$CF,$17,$6D,$76,$D7,$FD,$61,$D5,$FE,$8E
	.dc.b	$E7,$FF,$3D,$75,$FF,$A4,$FA,$7F,$CF,$D5,$FE,$AE,$8F,$F5,$8B,$FE
	.dc.b	$87,$8C,$BF,$E9,$58,$CF,$57,$FA,$22,$3F,$D5,$EF,$FA,$2E,$7F,$D4
	.dc.b	$31,$87,$FD,$16,$FF,$AE,$BA,$FF,$D2,$CF,$FA,$64,$FF,$5A,$57,$8D
	.dc.b	$D1,$FE,$93,$3F,$D6,$71,$C2,$CF,$FA,$82,$34,$7F,$A9,$CE,$27,$84
	.dc.b	$E2,$6A,$3F,$D8,$37,$FE,$A2,$BF,$2C,$F8,$CE,$07,$33,$11,$ED,$79
	.dc.b	$9A,$9E,$73,$F1,$9F,$9B,$87,$CF,$C3,$E3,$CF,$C4,$BE,$57,$D6,$FF
	.dc.b	$D9,$D1,$87,$DB,$1F,$12,$F9,$07,$FD,$B9,$D5,$FE,$B1,$E9,$FF,$5A
	.dc.b	$BF,$EA,$E8,$FF,$58,$75,$FF,$AC,$9C,$BF,$91,$BF,$F5,$CB,$CF,$AF
	.dc.b	$FD,$5F,$3F,$EB,$38,$C5,$F0,$8F,$F5,$B9,$FE,$7A,$3F,$D0,$FD,$1F
	.dc.b	$E7,$9E,$AF,$F3,$F7,$57,$FA,$46,$7F,$CF,$1D,$7F,$E8,$5A,$D4,$7F
	.dc.b	$A7,$4F,$F3,$E3,$87,$A0,$9F,$F4,$C3,$CB,$AB,$FD,$07,$D7,$FE,$92
	.dc.b	$6E,$1F,$EB,$1C,$3F,$D6,$91,$FE,$8C,$AF,$F5,$8B,$5F,$F4,$FD,$7F
	.dc.b	$A3,$9F,$09,$A2,$9A,$AF,$F4,$B3,$88,$FF,$4F,$CF,$36,$C6,$67,$E0
	.dc.b	$D7,$8A,$30,$7B,$E7,$D0,$FA,$CC,$9F,$02,$FD,$F1,$E5,$25,$7F,$B4
	.dc.b	$70,$FF,$50,$3F,$81,$E4,$7E,$7C,$F5,$BF,$84,$B7,$1F,$E1,$2D,$7F
	.dc.b	$84,$B5,$FF,$50,$B5,$E8,$D7,$F8,$C3,$67,$BC,$FF,$19,$D5,$F9,$6B
	.dc.b	$DC,$C3,$70,$FC,$B1,$27,$73,$A2,$46,$8F,$CE,$46,$9C,$07,$0E,$DA
	.dc.b	$F5,$9F,$03,$96,$B9,$9E,$EE,$5A,$E4,$EF,$3D,$1F,$1C,$F2,$6C,$F0
	.dc.b	$69,$C2,$34,$F9,$DB,$3C,$E3,$13,$C5,$A7,$17,$11,$B2,$4F,$22,$46
	.dc.b	$9B,$3C,$48,$7B,$4A,$7B,$4C,$10,$C1,$0F,$59,$92,$99,$3C,$E6,$4A
	.dc.b	$66,$7C,$F3,$ED,$87,$6B,$E6,$FF,$B0,$46,$1B,$19,$8F,$EC,$8D,$7F
	.dc.b	$02,$D7,$F0,$CD,$7F,$1A,$D7,$F9,$23,$5E,$C7,$48,$EE,$47,$04,$3B
	.dc.b	$B9,$69,$93,$C2,$E5,$A6,$4F,$1B,$96,$99,$3C,$AE,$5A,$64,$F3,$9E
	.dc.b	$02,$1D,$8A,$74,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86
	.dc.b	$8A,$68,$C1,$0C,$12,$3E,$91,$23,$E0,$46,$5B,$19,$8F,$84,$D7,$D6
	.dc.b	$D7,$CC,$D7,$C6,$D7,$C4,$D7,$B1,$D2,$3B,$91,$C1,$0E,$EE,$5A,$64
	.dc.b	$F0,$B9,$69,$93,$C6,$E5,$A6,$4F,$2B,$96,$99,$3C,$E7,$80,$87,$62
	.dc.b	$9D,$0A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$8F,$A4,$48,$F8,$11,$96,$C6,$63,$E1,$35,$F5,$B5,$F3,$35
	.dc.b	$F1,$B5,$F1,$35,$EC,$74,$8E,$E4,$70,$43,$BB,$96,$99,$3C,$2E,$5A
	.dc.b	$64,$F1,$B9,$69,$93,$CA,$E5,$A6,$4F,$39,$E0,$21,$D8,$FB,$BD,$8F
	.dc.b	$94,$FE,$56,$D8,$FF,$46,$35,$FE,$D2,$D7,$FB,$73,$5F,$EE,$AD,$7F
	.dc.b	$C9,$8D,$4D,$24,$78,$09,$1E,$06,$BE,$06,$BE,$06,$BE,$06,$BE,$06
	.dc.b	$BE,$06,$C7,$82,$3B,$10,$EC,$53,$B1,$4E,$C5,$3B,$14,$EC,$57,$B3
	.dc.b	$D0,$87,$42,$9D,$0A,$74,$29,$D0,$AF,$47,$91,$0E,$45,$3F,$C5,$FF
	.dc.b	$23,$FE,$3A,$23,$FE,$3A,$6B,$FE,$3A,$6B,$FE,$3A,$7F,$B6,$F3,$7F
	.dc.b	$A4,$A3,$FD,$43,$D3,$FE,$8B,$EB,$FF,$56,$47,$FA,$6B,$AF,$FC,$6B
	.dc.b	$5F,$E9,$68,$FF,$5F,$1C,$23,$33,$FE,$98,$8C,$10,$FF,$61,$75,$FF
	.dc.b	$A4,$EF,$FD,$31,$D9,$FE,$B3,$9F,$F5,$25,$78,$30,$FF,$4E,$E1,$FE
	.dc.b	$AE,$6C,$FF,$AA,$AF,$CD,$87,$FA,$7A,$3F,$D5,$98,$73,$9C,$C7,$98
	.dc.b	$F5,$1F,$ED,$79,$F1,$5F,$5A,$FF,$63,$1C,$5F,$5C,$6A,$7C,$11,$87
	.dc.b	$B9,$FE,$AF,$3B,$5F,$99,$FF,$5B,$57,$47,$57,$FE,$E1,$3C,$A5,$AE
	.dc.b	$8F,$C4,$8D,$C7,$FB,$04,$C9,$E7,$7F,$DA,$98,$7C,$7A,$ED,$0A,$FF
	.dc.b	$6B,$3E,$83,$E0,$1F,$EC,$2D,$FF,$D6,$EF,$FA,$E4,$FF,$5B,$E1,$FE
	.dc.b	$BE,$6F,$67,$F9,$AD,$B3,$FE,$C5,$78,$1F,$EC,$83,$FC,$E3,$3C,$7B
	.dc.b	$3F,$D8,$B0,$8C,$76,$7F,$B1,$64,$FF,$A8,$BB,$3F,$D6,$B5,$FE,$C6
	.dc.b	$3F,$D5,$DD,$5F,$EA,$23,$FD,$3A,$FF,$B6,$FA,$DD,$7F,$EC,$37,$FD
	.dc.b	$95,$3A,$9F,$F6,$24,$E7,$B3,$FD,$5D,$3C,$FB,$3F,$D6,$ED,$AF,$01
	.dc.b	$FE,$BB,$2B,$FE,$B8,$34,$FF,$B3,$E7,$FD,$AC,$F9,$A7,$FD,$99,$87
	.dc.b	$AA,$34,$FF,$B0,$63,$DD,$1F,$EB,$BA,$F3,$61,$EA,$8E,$47,$B7,$0F
	.dc.b	$45,$F3,$9F,$F5,$FD,$FA,$09,$5B,$38,$9F,$00,$B1,$D5,$F0,$47,$07
	.dc.b	$CA,$7F,$B4,$3C,$66,$CF,$2B,$FD,$57,$29,$CC,$61,$B1,$FE,$B0,$6B
	.dc.b	$FE,$B5,$6B,$FE,$BC,$6A,$3A,$4F,$77,$93,$5F,$F3,$DE,$1F,$E8,$02
	.dc.b	$3F,$E8,$06,$BA,$30,$43,$04,$3F,$D0,$6F,$06,$9B,$29,$B3,$FD,$0E
	.dc.b	$ED,$A6,$CA,$6C,$FF,$45,$BB,$69,$B2,$9B,$3F,$D3,$EE,$DA,$6C,$A6
	.dc.b	$FA,$51,$B6,$C6,$DA,$EE,$34,$47,$4D,$74,$60,$87,$7F,$9B,$FF,$57
	.dc.b	$91,$FF,$57,$B5,$FF,$57,$B5,$FF,$57,$B5,$FF,$57,$B5,$FF,$57,$B5
	.dc.b	$FF,$5F,$B6,$4D,$24,$78,$08,$78,$0A,$78,$0A,$78,$0A,$78,$0A,$78
	.dc.b	$0B,$1E,$06,$C7,$81,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$58,$EC,$D8
	.dc.b	$EC,$F4,$21,$D0,$A7,$42,$9D,$0B,$1D,$1B,$1D,$1E,$44,$39,$14,$E4
	.dc.b	$58,$E4,$D8,$E4,$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D,$8D,$46
	.dc.b	$08,$E0,$91,$F5,$08,$7C,$92,$9F,$00,$A7,$A4,$A7,$90,$A7,$84,$B1
	.dc.b	$E0,$6C,$78,$1E,$C4,$3B,$14,$EC,$53,$B1,$4E,$C5,$8E,$CD,$8E,$CF
	.dc.b	$42,$1D,$0A,$74,$29,$D0,$B1,$D1,$B1,$D1,$E4,$43,$91,$4E,$45,$8E
	.dc.b	$4D,$8E,$4F,$02,$1C,$0B,$1C,$1B,$1C,$1D,$12,$34,$D8,$D4,$60,$8E
	.dc.b	$09,$1F,$50,$87,$C9,$29,$F0,$0A,$7A,$4A,$79,$0A,$78,$4B,$1E,$06
	.dc.b	$C7,$81,$EC,$7F,$57,$24,$7F,$66,$7F,$B2,$B5,$D1,$FD,$75,$C3,$86
	.dc.b	$BC,$AB,$FB,$6A,$8C,$BC,$63,$31,$88,$D4,$60,$8F,$89,$AF,$88,$D9
	.dc.b	$92,$19,$8E,$84,$8E,$31,$E7,$6B,$E7,$38,$99,$29,$93,$A3,$C1,$F5
	.dc.b	$1A,$29,$A3,$07,$98,$C1,$0E,$C5,$3B,$18,$21,$82,$1D,$8C,$94,$C9
	.dc.b	$DC,$C9,$4C,$C7,$C3,$7C,$AF,$D4,$EE,$77,$B8,$7F,$C4,$B1,$86,$C6
	.dc.b	$6B,$E9,$4E,$5B,$19,$7F,$0F,$19,$6C,$65,$FC,$84,$65,$FC,$6F,$17
	.dc.b	$FA,$DD,$FF,$4D,$9F,$EA,$C3,$FD,$26,$7F,$A0,$79,$FF,$D6,$FC,$DF
	.dc.b	$E8,$4E,$6F,$F4,$1C,$FF,$A5,$0F,$F4,$4F,$47,$FA,$A6,$3F,$D2,$11
	.dc.b	$C4,$FF,$5C,$5F,$FA,$FE,$3F,$D2,$D1,$FE,$8A,$76,$74,$8F,$21,$FE
	.dc.b	$A1,$2D,$78,$6F,$C7,$5E,$03,$FD,$4F,$19,$7B,$F4,$7F,$B2,$5F,$6D
	.dc.b	$73,$8F,$F4,$F4,$71,$BF,$1B,$F0,$5F,$1D,$F6,$8F,$F6,$5E,$1E,$D8
	.dc.b	$FF,$54,$46,$0F,$09,$CD,$FF,$57,$3E,$D7,$73,$FE,$B6,$73,$1D,$DF
	.dc.b	$F4,$D7,$3C,$8F,$F5,$56,$9F,$EA,$0C,$7F,$D5,$51,$FE,$A5,$21,$FE
	.dc.b	$EA,$DB,$FD,$58,$FF,$AA,$67,$84,$7F,$AB,$DF,$F5,$4F,$27,$FA,$73
	.dc.b	$93,$FC,$F0,$FF,$9E,$F9,$BF,$D2,$91,$FE,$90,$BF,$F3,$E9,$FE,$98
	.dc.b	$7F,$D1,$F1,$FE,$8E,$E7,$FF,$4B,$CF,$FA,$74,$FF,$40,$C7,$FA,$72
	.dc.b	$BF,$D1,$B1,$FE,$AD,$9F,$F2,$6C,$FF,$A5,$B9,$FF,$D2,$73,$D8,$E3
	.dc.b	$7E,$19,$ED,$CB,$FE,$9F,$9E,$4F,$FA,$66,$FC,$33,$E8,$9F,$F4,$C1
	.dc.b	$E3,$C3,$C2,$F4,$3D,$47,$23,$C0,$E8,$4F,$9A,$FE,$2E,$1E,$6B,$F1
	.dc.b	$D9,$F7,$57,$8A,$BE,$29,$F3,$1C,$4F,$79,$E2,$70,$C3,$D9,$1E,$52
	.dc.b	$61,$E3,$AF,$14,$7A,$0C,$B8,$77,$87,$F4,$BB,$61,$FE,$A3,$E2,$FE
	.dc.b	$1B,$F2,$3F,$C2,$89,$7F,$C2,$1A,$FF,$06,$6B,$FC,$1B,$91,$1F,$EB
	.dc.b	$8C,$DF,$B6,$BE,$33,$0D,$BF,$E8,$D3,$A9,$C9,$23,$33,$B7,$FA,$EC
	.dc.b	$ED,$CC,$ED,$CC,$ED,$FE,$D1,$3B,$73,$3B,$73,$3B,$7F,$D8,$93,$B7
	.dc.b	$33,$B7,$33,$B7,$FD,$46,$D9,$E1,$1A,$9E,$0E,$9B,$1A,$8C,$1E,$A2
	.dc.b	$1E,$62,$9E,$22,$9D,$4A,$72,$38,$10,$E0,$53,$81,$4E,$06,$88,$68
	.dc.b	$A6,$8C,$10,$C1,$1F,$61,$27,$E2,$12,$7F,$C1,$7F,$35,$2B,$FC,$32
	.dc.b	$D7,$FC,$08,$D7,$FC,$30,$D7,$FC,$42,$D7,$FC,$3C,$D7,$FC,$34,$DE
	.dc.b	$83,$49,$1E,$02,$1E,$02,$9E,$02,$9E,$02,$9E,$02,$9E,$02,$C7,$81
	.dc.b	$B1,$E0,$7B,$10,$EC,$53,$B1,$4E,$C5,$3B,$16,$3B,$36,$3B,$3D,$08
	.dc.b	$74,$29,$D0,$A7,$42,$C7,$46,$C7,$47,$91,$0E,$45,$39,$16,$39,$36
	.dc.b	$39,$3C,$08,$70,$2C,$70,$6C,$70,$74,$48,$D3,$63,$51,$82,$38,$24
	.dc.b	$7D,$42,$1F,$24,$A7,$C0,$29,$E9,$29,$E4,$29,$E1,$2C,$78,$1B,$1E
	.dc.b	$07,$B1,$0E,$C5,$3B,$14,$EC,$53,$B1,$63,$B3,$63,$B3,$D0,$87,$42
	.dc.b	$9D,$0A,$74,$2C,$74,$6C,$74,$79,$10,$E4,$53,$91,$63,$93,$63,$93
	.dc.b	$C0,$87,$02,$C7,$06,$C7,$07,$44,$8D,$36,$35,$18,$23,$82,$47,$D4
	.dc.b	$21,$F2,$4A,$7C,$02,$9E,$92,$9E,$42,$9E,$12,$C7,$81,$B1,$E0,$7B
	.dc.b	$1F,$BB,$EE,$72,$FF,$AC,$9B,$1F,$EA,$90,$E1,$FF,$56,$35,$FF,$5F
	.dc.b	$B5,$FF,$63,$35,$1C,$A3,$B1,$1E,$D1,$B7,$46,$23,$4F,$81,$AF,$81
	.dc.b	$AF,$81,$AF,$82,$3A,$12,$33,$1D,$8E,$87,$72,$18,$21,$DC,$C9,$4C
	.dc.b	$9E,$13,$25,$32,$78,$DF,$0C,$78,$E3,$66,$4A,$79,$4E,$C4,$3A,$14
	.dc.b	$E0,$76,$38,$1A,$21,$A3,$86,$9F,$EB,$AE,$0F,$F5,$B9,$23,$FD,$6A
	.dc.b	$D7,$FD,$66,$D7,$FC,$5C,$FF,$BE,$39,$3F,$D3,$B1,$FE,$89,$3F,$D4
	.dc.b	$FD,$3F,$E9,$B7,$FD,$33,$D3,$FE,$9B,$94,$FE,$12,$7F,$D5,$C6,$BA
	.dc.b	$7F,$0F,$D3,$FE,$9A,$8F,$F5,$1C,$63,$3F,$F4,$99,$C6,$FF,$D6,$8F
	.dc.b	$FA,$51,$B3,$E2,$3C,$87,$FA,$AA,$FC,$91,$FE,$AB,$AF,$F5,$24,$FF
	.dc.b	$A7,$CE,$07,$FA,$96,$FF,$D5,$B3,$D2,$39,$1D,$1F,$5B,$E5,$3C,$E7
	.dc.b	$FB,$4A,$3F,$D8,$98,$7F,$AE,$23,$A6,$1E,$78,$F8,$98,$7C,$3A,$DE
	.dc.b	$1E,$4B,$FF,$6B,$1C,$0F,$2C,$7F,$AC,$DB,$7F,$0E,$F9,$C7,$8C,$F9
	.dc.b	$87,$8E,$3F,$D8,$93,$C6,$3D,$67,$52,$B9,$C3,$DB,$1F,$59,$FF,$6E
	.dc.b	$74,$7F,$AD,$7A,$7F,$D8,$12,$94,$E6,$3F,$D8,$73,$FE,$BF,$ED,$FE
	.dc.b	$19,$5F,$EC,$49,$C1,$F8,$33,$F8,$3B,$97,$89,$0E,$D5,$D5,$B3,$FC
	.dc.b	$3B,$A9,$7F,$E9,$C9,$FF,$4D,$73,$FF,$A9,$1F,$F5,$97,$2F,$F1,$4E
	.dc.b	$8F,$F5,$E4,$60,$FF,$56,$F4,$FF,$AA,$6B,$FD,$49,$5F,$EA,$0E,$AF
	.dc.b	$F6,$95,$7F,$A7,$5A,$FF,$A7,$0E,$57,$D1,$FF,$5E,$71,$7F,$AB,$1F
	.dc.b	$F5,$1D,$78,$6F,$C9,$1A,$8E,$D3,$FE,$A7,$BF,$F5,$EC,$6E,$7C,$13
	.dc.b	$92,$47,$03,$C3,$5F,$ED,$33,$D3,$3F,$EC,$17,$0F,$C4,$39,$C7,$07
	.dc.b	$FD,$A7,$7F,$ED,$55,$7F,$EE,$FA,$F9,$91,$EA,$3A,$1C,$1C,$C7,$A2
	.dc.b	$3E,$59,$D4,$F8,$C7,$BC,$E2,$76,$C3,$FD,$89,$3E,$8A,$F9,$47,$FB
	.dc.b	$1A,$5C,$6E,$A5,$FF,$57,$8C,$36,$33,$3F,$8D,$38,$B8,$7F,$21,$3C
	.dc.b	$1D,$35,$CC,$6D,$FF,$67,$4F,$F6,$5E,$C7,$E8,$8F,$0B,$0D,$AF,$04
	.dc.b	$EA,$72,$48,$CC,$ED,$FD,$14,$ED,$CC,$ED,$CC,$ED,$FE,$55,$3B,$73
	.dc.b	$3B,$73,$3B,$7F,$B9,$CE,$DC,$CE,$DC,$CE,$D4,$ED,$CC,$ED,$CC,$EE
	.dc.b	$75,$18,$9D,$46,$09,$3B,$24,$69,$B3,$F0,$08,$7B,$0A,$7A,$0A,$79
	.dc.b	$0A,$78,$4B,$3E,$09,$EC,$48,$CC,$F8,$1E,$C4,$3B,$14,$FE,$B1,$D4
	.dc.b	$F9,$6F,$F1,$34,$23,$0F,$F8,$A5,$AF,$E4,$5A,$FF,$1F,$6B,$D2,$39
	.dc.b	$C6,$23,$FC,$52,$D7,$B4,$68,$8F,$76,$BD,$DA,$F7,$6B,$DC,$E0,$43
	.dc.b	$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3D,$AF,$91,$A7,$72
	.dc.b	$9D,$0A,$73,$29,$CC,$F8,$A7,$12,$9C,$4A,$71,$29,$C6,$3E,$71,$23
	.dc.b	$C9,$1A,$3B,$90,$C1,$23,$83,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$8E,$F1
	.dc.b	$88,$EC,$D7,$B4,$68,$8F,$76,$BD,$DA,$F7,$6B,$DC,$E0,$43,$81,$4E
	.dc.b	$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3D,$AF,$91,$A7,$72,$9D,$0A
	.dc.b	$73,$29,$CC,$F8,$A7,$12,$9C,$4A,$71,$29,$C6,$3E,$71,$23,$C9,$1A
	.dc.b	$3B,$90,$C1,$23,$83,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$8E,$F1,$88,$EC
	.dc.b	$D7,$B4,$68,$8F,$76,$BD,$DA,$F7,$6B,$DC,$E0,$43,$81,$4E,$05,$38
	.dc.b	$1A,$21,$A2,$9A,$30,$43,$04,$3D,$BF,$35,$FC,$80,$8F,$F2,$96,$BF
	.dc.b	$CA,$5A,$FF,$29,$6B,$FC,$D1,$AF,$F2,$86,$BF,$CD,$5B,$26,$92,$3C
	.dc.b	$04,$3C,$05,$3C,$05,$3C,$05,$3C,$05,$3C,$05,$8F,$03,$63,$C0,$F6
	.dc.b	$21,$D8,$A7,$62,$9D,$8A,$76,$2C,$76,$6C,$76,$7A,$10,$E8,$53,$FC
	.dc.b	$4B,$F2,$DF,$E2,$92,$3F,$E2,$4C,$D1,$88,$FF,$12,$B5,$FF,$5E,$69
	.dc.b	$FE,$35,$8F,$F6,$37,$4F,$FA,$87,$A3,$FD,$19,$D7,$FE,$9F,$EA,$FF
	.dc.b	$49,$9F,$E9,$9E,$BF,$F4,$C7,$5F,$FB,$33,$2F,$F4,$A3,$FE,$92,$BF
	.dc.b	$F5,$2C,$7F,$A4,$EF,$93,$FE,$B0,$8F,$F6,$43,$FE,$A0,$6C,$FF,$AF
	.dc.b	$E3,$81,$DA,$7F,$D8,$64,$7F,$D7,$8F,$47,$95,$FF,$B4,$1F,$F6,$31
	.dc.b	$FE,$B4,$AF,$5B,$CC,$E2,$53,$D3,$3C,$EB,$C5,$3F,$EC,$CC,$3F,$D6
	.dc.b	$F5,$E4,$36,$71,$8E,$AE,$5A,$7C,$23,$96,$1F,$22,$7E,$76,$1F,$2E
	.dc.b	$BF,$D8,$07,$C5,$7D,$4F,$FA,$DE,$3C,$35,$CE,$7F,$DA,$95,$CC,$FF
	.dc.b	$69,$5A,$BA,$DF,$C2,$8F,$F5,$FF,$57,$FA,$C7,$A7,$FD,$87,$18,$97
	.dc.b	$57,$FA,$D1,$FF,$5D,$47,$FA,$F5,$FF,$5C,$9F,$E7,$57,$04,$EB,$FF
	.dc.b	$5F,$3C,$E7,$94,$F0,$6C,$66,$BF,$D2,$3D,$7F,$EA,$4A,$FF,$46,$F5
	.dc.b	$7F,$A6,$5F,$F5,$01,$FE,$99,$E9,$FF,$46,$3F,$E9,$13,$FD,$2D,$1F
	.dc.b	$E9,$98,$CB,$C6,$38,$75,$7F,$AE,$27,$FD,$4D,$39,$3F,$D8,$67,$18
	.dc.b	$FF,$57,$CF,$03,$FD,$4C,$78,$A3,$4F,$FA,$C8,$F3,$D7,$A6,$BC,$3D
	.dc.b	$5F,$ED,$39,$E1,$3F,$EC,$BA,$EE,$D3,$85,$F9,$4F,$87,$D7,$FE,$C9
	.dc.b	$AF,$F5,$8E,$1E,$8C,$3A,$47,$94,$F8,$47,$FA,$D6,$3C,$2F,$C4,$BE
	.dc.b	$D5,$F1,$AB,$8E,$1E,$AA,$EB,$87,$81,$FF,$60,$9E,$CA,$E0,$FF,$B6
	.dc.b	$F9,$3F,$D7,$5F,$13,$FE,$B7,$24,$7F,$0F,$6B,$FE,$B2,$6B,$FC,$3B
	.dc.b	$99,$1F,$EB,$3F,$D2,$1F,$17,$49,$5F,$EB,$28,$CB,$63,$2F,$FA,$C2
	.dc.b	$32,$D8,$CB,$FE,$AE,$8C,$B6,$32,$FF,$AB,$23,$2D,$8C,$BF,$EA,$A8
	.dc.b	$CB,$63,$3D,$C9,$CB,$63,$31,$82,$38,$24,$FA,$C8,$7A,$4A,$79,$CA
	.dc.b	$79,$4A,$78,$CB,$3E,$16,$CF,$81,$B3,$E0,$7B,$10,$EC,$53,$B1,$4E
	.dc.b	$C5,$9E,$CE,$4F,$F5,$54,$7F,$A8,$88,$FF,$AA,$DA,$FF,$AA,$DA,$FF
	.dc.b	$AA,$DA,$FF,$AD,$9A,$FF,$AA,$9A,$F6,$3A,$10,$E8,$53,$A1,$4E,$85
	.dc.b	$3A,$14,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81
	.dc.b	$4E,$06,$88,$68,$A6,$8C,$10,$C1,$3A,$FF,$2E,$D8,$F9,$31,$82,$38
	.dc.b	$24,$7C,$38,$CB,$63,$2F,$BA,$32,$D8,$CB,$EB,$8C,$B6,$32,$FA,$A3
	.dc.b	$2D,$8C,$BE,$B8,$CB,$63,$2F,$AC,$F2,$10,$F1,$14,$F0,$14,$EC,$53
	.dc.b	$A1,$4E,$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14
	.dc.b	$E0,$68,$86,$8A,$68,$C1,$0C,$12,$3E,$4B,$63,$E4,$C6,$08,$E0,$91
	.dc.b	$F0,$E3,$2D,$8C,$BE,$E8,$CB,$63,$2F,$AE,$32,$D8,$CB,$EA,$8C,$B6
	.dc.b	$32,$FA,$E3,$2D,$8C,$BE,$B3,$C8,$43,$C4,$53,$C0,$53,$B1,$4E,$85
	.dc.b	$3A,$1C,$88,$72,$29,$C8,$FB,$FD,$EF,$9A,$44,$20,$8D,$38,$36,$E1
	.dc.b	$C3,$5E,$2D,$7A,$35,$E3,$1C,$DA,$58,$D3,$E1,$8D,$34,$D3,$88,$D9
	.dc.b	$1C,$12,$36,$43,$CE,$E5,$A6,$4F,$4B,$96,$99,$3D,$6E,$5A,$64,$F5
	.dc.b	$9D,$08,$72,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$87,$C7,$3A
	.dc.b	$94,$E4,$53,$89,$4E,$59,$B3,$7C,$2F,$D7,$8C,$36,$33,$3F,$EB,$78
	.dc.b	$C1,$1C,$10,$FF,$38,$74,$7F,$9F,$1F,$F4,$27,$4F,$FA,$05,$FF,$48
	.dc.b	$75,$7F,$AD,$9F,$F4,$04,$EF,$9F,$FD,$09,$5B,$EB,$FF,$49,$D7,$FA
	.dc.b	$40,$FF,$54,$9F,$E8,$7E,$CF,$F5,$11,$FE,$AF,$7F,$D3,$E7,$63,$AC
	.dc.b	$FF,$A4,$2F,$B1,$FE,$96,$C3,$C5,$3E,$63,$B6,$1F,$EA,$C8,$FF,$4C
	.dc.b	$C6,$2F,$B6,$1F,$EA,$89,$E3,$5E,$8A,$D1,$FE,$B8,$3C,$47,$59,$FF
	.dc.b	$64,$D7,$2A,$FF,$56,$46,$27,$83,$FE,$A9,$7F,$73,$5F,$ED,$79,$EF
	.dc.b	$7E,$99,$FF,$59,$1F,$0C,$FF,$5F,$47,$43,$FD,$6B,$3E,$5A,$ED,$3F
	.dc.b	$19,$EC,$A3,$B3,$D0,$FF,$6A,$4F,$37,$FD,$5F,$D7,$FE,$AA,$27,$67
	.dc.b	$FA,$DE,$3C,$F0,$8F,$E0,$55,$FE,$B1,$6C,$66,$7F,$D6,$CF,$27,$FD
	.dc.b	$7B,$D7,$FE,$C0,$8D,$76,$7F,$AD,$BA,$7F,$D2,$BD,$1F,$E9,$17,$FD
	.dc.b	$4D,$D7,$FE,$96,$C3,$FD,$41,$D9,$FE,$AE,$8C,$47,$FA,$4A,$3F,$D2
	.dc.b	$CF,$FA,$46,$B9,$76,$7F,$A4,$2B,$77,$D2,$7F,$D1,$B5,$FE,$99,$EC
	.dc.b	$FF,$58,$CF,$FA,$54,$97,$C6,$BC,$97,$FE,$9C,$BF,$F4,$B1,$FE,$98
	.dc.b	$C3,$C2,$F3,$3F,$D7,$27,$FA,$9E,$F9,$BE,$53,$FD,$9F,$7D,$5E,$58
	.dc.b	$7F,$A8,$EB,$DD,$87,$AA,$78,$47,$FA,$B9,$EC,$7C,$19,$E0,$7F,$AE
	.dc.b	$DF,$F6,$0C,$FA,$0E,$4A,$53,$FD,$2B,$97,$F8,$47,$C2,$FF,$05,$25
	.dc.b	$7F,$49,$8C,$B7,$99,$3A,$8F,$EA,$33,$A6,$CE,$9F,$EA,$58,$3E,$51
	.dc.b	$EB,$FF,$4B,$8C,$36,$33,$3C,$1A,$FF,$5E,$9D,$B9,$9E,$04,$70,$43
	.dc.b	$84,$F2,$7F,$B4,$4F,$27,$33,$C5,$A7,$13,$73,$C5,$FE,$E3,$3C,$5C
	.dc.b	$CF,$16,$9C,$4D,$CF,$15,$3C,$5C,$CF,$16,$9C,$4D,$CF,$19,$E1,$18
	.dc.b	$9E,$0D,$38,$46,$89,$3C,$49,$3C,$23,$2D,$8C,$CF,$26,$BF,$10,$87
	.dc.b	$B8,$A7,$A0,$A7,$8C,$A3,$B1,$F1,$7F,$E0,$66,$C7,$F8,$35,$AF,$F8
	.dc.b	$65,$AF,$F8,$79,$AF,$F8,$85,$AF,$F8,$91,$AF,$78,$EC,$47,$B3,$5E
	.dc.b	$CD,$7B,$35,$EC,$D7,$B3,$5E,$C7,$42,$1D,$0A,$74,$29,$D0,$A7,$42
	.dc.b	$9D,$0E,$44,$39,$14,$E4,$53,$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0
	.dc.b	$D1,$0D,$14,$D1,$82,$18,$24,$7D,$42,$47,$C9,$6C,$7C,$06,$BE,$96
	.dc.b	$BE,$46,$BE,$16,$BE,$06,$BE,$08,$EC,$47,$B3,$5E,$CD,$7B,$35,$EC
	.dc.b	$D7,$B3,$5E,$C7,$42,$1D,$0A,$74,$29,$D0,$A7,$42,$9D,$0E,$44,$39
	.dc.b	$14,$E4,$53,$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1
	.dc.b	$82,$18,$24,$7D,$42,$47,$C9,$6C,$7C,$06,$BE,$96,$BE,$46,$BE,$16
	.dc.b	$BE,$06,$BE,$08,$EC,$47,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$C7
	.dc.b	$42,$1D,$0A,$74,$29,$D0,$A7,$42,$9D,$0E,$44,$3E,$AF,$63,$E5,$BF
	.dc.b	$D5,$84,$8F,$F5,$A3,$5F,$F5,$FB,$5F,$F5,$FB,$5F,$F5,$FB,$5F,$F6
	.dc.b	$3B,$5F,$F6,$63,$63,$C1,$1D,$88,$76,$29,$D8,$A7,$62,$9D,$8A,$76
	.dc.b	$2B,$D9,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8,$87,$22,$9C
	.dc.b	$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$2C,$FF,$D7,$1C,$9F,$EB
	.dc.b	$52,$47,$F8,$BD,$AF,$F8,$B1,$AF,$FC,$01,$8E,$2F,$F9,$D3,$A7,$FD
	.dc.b	$47,$D7,$FE,$AF,$E9,$FF,$4F,$F5,$7F,$A1,$8F,$F4,$F5,$7F,$A0,$FA
	.dc.b	$BF,$D1,$53,$A3,$FD,$1B,$7F,$E9,$39,$FF,$5A,$9F,$E8,$83,$FD,$62
	.dc.b	$7F,$A8,$A7,$FD,$58,$FF,$A7,$3A,$FF,$D7,$2F,$FA,$82,$BF,$D3,$B8
	.dc.b	$78,$E3,$47,$FA,$70,$E7,$87,$FB,$10,$F3,$4F,$FA,$56,$3C,$07,$FA
	.dc.b	$BD,$F2,$CF,$38,$FF,$4F,$57,$81,$FF,$51,$4F,$81,$FF,$5C,$98,$3E
	.dc.b	$06,$1F,$19,$CD,$FC,$47,$DF,$87,$C4,$7C,$E7,$BE,$7E,$24,$7C,$08
	.dc.b	$F0,$9E,$B3,$CC,$7F,$AE,$4E,$18,$79,$E3,$FD,$67,$1D,$A7,$C5,$5F
	.dc.b	$02,$3E,$1D,$FD,$37,$FD,$6B,$D7,$FE,$B3,$7F,$D7,$9D,$7F,$EA,$DE
	.dc.b	$B6,$A9,$FF,$5B,$D7,$FA,$D0,$8F,$FA,$D1,$B3,$C2,$3A,$9F,$EB,$63
	.dc.b	$89,$4F,$F4,$D7,$5F,$FA,$87,$A7,$FD,$2B,$1F,$E9,$83,$FD,$1E,$53
	.dc.b	$7C,$DF,$E9,$08,$FF,$52,$93,$A7,$FD,$2F,$3F,$EB,$C7,$D0,$FF,$A9
	.dc.b	$FA,$BF,$D4,$93,$D8,$FF,$50,$1C,$63,$BB,$99,$FF,$5C,$DF,$FA,$FA
	.dc.b	$B8,$C7,$FA,$8E,$BC,$67,$FA,$C0,$FF,$55,$BC,$2F,$FD,$4D,$19,$3F
	.dc.b	$D5,$BD,$7F,$EC,$77,$FD,$7F,$87,$77,$FD,$7B,$1F,$ED,$87,$84,$79
	.dc.b	$5E,$EF,$A2,$BD,$F0,$9F,$F6,$1B,$FE,$B3,$8F,$F5,$97,$53,$C6,$F8
	.dc.b	$73,$FB,$D8,$C3,$FE,$B1,$8C,$BC,$1A,$FF,$0B,$8D,$BF,$EB,$28,$CF
	.dc.b	$2A,$73,$3D,$67,$8E,$AF,$CC,$5E,$66,$1B,$5E,$09,$D4,$E4,$91,$99
	.dc.b	$DB,$EA,$9D,$B9,$9D,$B9,$9D,$BF,$82,$9D,$B9,$9D,$B9,$9D,$BF,$9F
	.dc.b	$9D,$B9,$9D,$B9,$9D,$A9,$DB,$99,$DB,$99,$DC,$EA,$31,$3A,$8C,$12
	.dc.b	$76,$48,$D3,$67,$E0,$10,$F6,$14,$F4,$14,$F2,$14,$F0,$96,$7C,$13
	.dc.b	$D8,$91,$99,$F0,$3D,$88,$76,$29,$FB,$0F,$C0,$DF,$07,$F8,$A8,$46
	.dc.b	$1F,$C6,$B5,$FC,$8B,$5F,$CB,$B5,$E9,$1C,$E3,$11,$FC,$55,$AF,$68
	.dc.b	$D1,$1E,$ED,$7B,$B5,$EE,$D7,$B9,$C0,$87,$02,$9C,$0A,$70,$34,$43
	.dc.b	$45,$34,$60,$86,$08,$7B,$5F,$23,$4E,$E5,$3A,$14,$E6,$53,$99,$F1
	.dc.b	$4E,$25,$38,$94,$E2,$53,$8C,$7C,$E2,$47,$92,$34,$77,$21,$82,$47
	.dc.b	$07,$C2,$D7,$C2,$D7,$C2,$D7,$C3,$1D,$E3,$11,$D9,$AF,$68,$D1,$1E
	.dc.b	$ED,$7B,$B5,$EE,$D7,$B9,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34
	.dc.b	$60,$86,$08,$7B,$5F,$23,$4E,$E5,$3A,$14,$E6,$53,$99,$F1,$4E,$25
	.dc.b	$38,$94,$E2,$53,$8C,$7C,$E2,$47,$92,$34,$77,$21,$82,$47,$07,$C2
	.dc.b	$D7,$C2,$D7,$C2,$D7,$C3,$1D,$E3,$11,$D9,$AF,$68,$D1,$1E,$ED,$7B
	.dc.b	$B5,$EE,$D7,$B9,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86
	.dc.b	$08,$7C,$AE,$A7,$C0,$7F,$2B,$6C,$7F,$76,$6B,$FD,$E9,$AF,$F8,$01
	.dc.b	$AF,$F8,$91,$AF,$F8,$F5,$A9,$A4,$8F,$01,$23,$C0,$D7,$C0,$D7,$C0
	.dc.b	$D7,$C0,$D7,$C0,$D7,$C0,$D8,$F0,$47,$62,$1D,$8A,$76,$29,$D8,$A7
	.dc.b	$62,$9D,$8A,$F6,$7A,$10,$E8,$53,$A1,$4E,$85,$3A,$15,$E8,$F2,$21
	.dc.b	$C8,$A7,$F8,$B7,$E3,$7F,$C5,$7A,$1A,$48,$FF,$1A,$EB,$FE,$2F,$8C
	.dc.b	$47,$F8,$F2,$30,$EA,$79,$3F,$EB,$9E,$BF,$F4,$57,$5F,$FA,$34,$FF
	.dc.b	$43,$97,$B3,$FD,$3D,$D7,$FE,$97,$8C,$57,$0E,$CF,$F4,$AD,$7F,$9B
	.dc.b	$2B,$4D,$EC,$FF,$56,$5F,$FA,$76,$FF,$D3,$27,$FA,$58,$FF,$51,$C6
	.dc.b	$AF,$FD,$2A,$FF,$AC,$23,$27,$FA,$58,$FF,$52,$DA,$3F,$D6,$0F,$FA
	.dc.b	$6E,$3F,$D4,$67,$FA,$EA,$BC,$04,$BF,$5D,$77,$C3,$E0,$C7,$9E,$BB
	.dc.b	$1E,$6B,$EC,$FF,$AE,$A7,$6F,$FB,$3C,$C1,$FE,$D5,$9E,$4F,$8A,$FC
	.dc.b	$0F,$C7,$9F,$F5,$A4,$F8,$A3,$BC,$FD,$08,$E2,$FA,$30,$F8,$73,$FE
	.dc.b	$BD,$BF,$1B,$FE,$BB,$BF,$1B,$FE,$CF,$30,$78,$CF,$31,$FE,$CE,$EC
	.dc.b	$FF,$5D,$76,$7F,$A8,$B0,$46,$27,$FD,$66,$7F,$A1,$23,$3C,$9F,$EB
	.dc.b	$C8,$C4,$FF,$97,$5F,$F5,$F1,$FE,$B9,$9D,$B6,$3A,$3F,$E9,$6E,$6F
	.dc.b	$F5,$13,$FE,$B9,$E7,$FF,$4E,$74,$7F,$A6,$63,$FD,$5F,$3F,$EA,$B9
	.dc.b	$CB,$C2,$7F,$D4,$93,$FE,$9D,$EA,$FF,$58,$BF,$EB,$D3,$89,$FE,$A8
	.dc.b	$EB,$75,$FF,$AC,$2B,$8E,$1F,$EA,$F6,$CF,$FB,$3F,$0F,$F6,$2C,$A3
	.dc.b	$FD,$7D,$7F,$EB,$A3,$A5,$F3,$AF,$34,$F6,$9F,$F5,$CC,$62,$75,$1E
	.dc.b	$53,$FD,$A0,$73,$3D,$45,$AF,$24,$F8,$AB,$C3,$3F,$EC,$6C,$3D,$EF
	.dc.b	$FA,$F6,$3D,$0F,$C0,$39,$9B,$3A,$47,$0A,$FF,$61,$E0,$BF,$F7,$D7
	.dc.b	$AC,$FD,$29,$F1,$7F,$EB,$E8,$C3,$63,$38,$FF,$B1,$27,$2D,$8C,$BF
	.dc.b	$EC,$C8,$CB,$63,$2F,$FB,$CA,$32,$FF,$B4,$BC,$27,$CD,$FF,$A1,$5B
	.dc.b	$1F,$ED,$06,$BF,$ED,$46,$BC,$63,$81,$1E,$0D,$78,$35,$E0,$68,$86
	.dc.b	$8A,$68,$C1,$0C,$10,$12,$3C,$AE,$03,$87,$51,$E3,$75,$19,$77,$19
	.dc.b	$76,$79,$0D,$B9,$76,$64,$D9,$E7,$36,$64,$D9,$93,$67,$AC,$A7,$03
	.dc.b	$47,$03,$45,$34,$62,$3D,$44,$8F,$33,$63,$FD,$7F,$F0,$3F,$EB,$92
	.dc.b	$38,$21,$FE,$C2,$72,$D3,$27,$FB,$11,$CB,$4C,$9F,$ED,$17,$2D,$32
	.dc.b	$7F,$B1,$DC,$B4,$C9,$FE,$D6,$72,$D3,$3D,$C8,$CB,$63,$31,$82,$38
	.dc.b	$21,$EB,$21,$E9,$29,$E7,$29,$E5,$29,$E3,$2C,$78,$5B,$1E,$06,$9E
	.dc.b	$03,$B1,$0E,$C5,$3B,$14,$EC,$58,$EC,$D8,$EC,$D3,$B1,$D0,$87,$42
	.dc.b	$9D,$0B,$1D,$1B,$1D,$1A,$74,$39,$10,$E4,$58,$E4,$D8,$E4,$D3,$91
	.dc.b	$C0,$91,$C1,$B1,$C1,$A7,$08,$D1,$23,$4D,$8D,$46,$08,$E0,$87,$D4
	.dc.b	$21,$F2,$4A,$7C,$02,$9E,$92,$9E,$42,$C7,$85,$B1,$E0,$69,$E0,$3B
	.dc.b	$10,$EC,$53,$B1,$4E,$C5,$8E,$CD,$8E,$CD,$3B,$1D,$08,$74,$29,$D0
	.dc.b	$B1,$D1,$B1,$D1,$A7,$43,$91,$0E,$45,$8E,$4D,$8E,$4D,$39,$1C,$09
	.dc.b	$1C,$1B,$1C,$1A,$70,$8D,$12,$34,$D8,$D4,$60,$8E,$08,$7F,$45,$3F
	.dc.b	$A4,$C7,$F8,$F9,$C3,$5F,$E9,$4D,$77,$D1,$C5,$CB,$5C,$96,$38,$26
	.dc.b	$92,$38,$46,$5B,$19,$8E,$51,$82,$38,$21,$E3,$8E,$0D,$8D,$B4,$D9
	.dc.b	$EE,$72,$D3,$27,$B0,$A7,$B0,$E4,$43,$81,$4E,$05,$38,$1A,$21,$A2
	.dc.b	$9A,$30,$43,$04,$3E,$31,$E4,$29,$DC,$A7,$42,$9C,$CA,$3A,$9D,$4F
	.dc.b	$8D,$FF,$60,$92,$7F,$D7,$4D,$7F,$D7,$D1,$B7,$24,$3F,$09,$CB,$FE
	.dc.b	$92,$7F,$D6,$1D,$3F,$EA,$78,$FF,$47,$74,$FF,$A2,$A3,$FD,$73,$3F
	.dc.b	$E8,$89,$FF,$5F,$BC,$31,$FF,$5A,$4F,$1E,$AF,$F6,$0C,$EF,$AF,$FD
	.dc.b	$33,$18,$8F,$F4,$A5,$F2,$AF,$F4,$CB,$6F,$FD,$72,$74,$24,$FF,$A7
	.dc.b	$DC,$B4,$FF,$57,$CF,$2B,$FF,$55,$47,$FA,$BE,$FF,$D4,$EF,$FB,$05
	.dc.b	$F1,$4F,$B2,$3C,$73,$E6,$C3,$FD,$67,$1F,$EB,$53,$FD,$9E,$73,$7E
	.dc.b	$25,$FF,$B1,$E3,$FD,$58,$72,$3F,$C7,$F5,$D4,$F6,$DA,$B3,$5F,$EA
	.dc.b	$FE,$9F,$F5,$1E,$3F,$EA,$6F,$0D,$D9,$9E,$CF,$F2,$4C,$63,$3F,$F5
	.dc.b	$99,$93,$FD,$5E,$7F,$AC,$A3,$F7,$D5,$CE,$B5,$5C,$69,$39,$7F,$CA
	.dc.b	$AD,$7F,$D6,$87,$FA,$70,$FF,$4C,$9F,$E9,$7E,$9F,$F5,$17,$67,$FA
	.dc.b	$6A,$3F,$D1,$BD,$5F,$EA,$88,$C4,$FF,$A6,$27,$FD,$24,$D7,$FD,$3E
	.dc.b	$7F,$A4,$C9,$3C,$5F,$F6,$17,$67,$FA,$A6,$30,$7F,$B0,$E3,$CA,$78
	.dc.b	$3A,$FF,$D4,$97,$E0,$AF,$F5,$0C,$F6,$AE,$35,$FE,$B1,$8F,$59,$FE
	.dc.b	$A7,$AF,$F3,$AC,$F5,$BF,$F6,$6C,$FC,$18,$E2,$DA,$EE,$F9,$AB,$C0
	.dc.b	$72,$8F,$F6,$A1,$DE,$78,$CF,$9C,$F8,$77,$E6,$BF,$97,$1C,$CF,$2B
	.dc.b	$FE,$B4,$AE,$F5,$CA,$3E,$51,$C1,$FF,$6C,$F6,$7F,$AE,$EF,$A1,$FE
	.dc.b	$C6,$8F,$F5,$E5,$FF,$0B,$E3,$7C,$97,$F4,$3E,$6F,$E1,$AD,$8C,$CF
	.dc.b	$F4,$99,$CB,$63,$2F,$F0,$E9,$E0,$47,$FA,$77,$E0,$8F,$6F,$FA,$4B
	.dc.b	$63,$31,$82,$38,$21,$FD,$66,$38,$36,$36,$D3,$67,$F6,$78,$DB,$63
	.dc.b	$6D,$36,$7F,$AC,$63,$6D,$8D,$B4,$D9,$FE,$BA,$8D,$B6,$36,$D3,$62
	.dc.b	$36,$D8,$DB,$4D,$C6,$89,$1A,$6C,$6A,$30,$47,$04,$3E,$21,$0F,$80
	.dc.b	$53,$D8,$53,$D0,$53,$D1,$E3,$FD,$82,$3F,$71,$AF,$D5,$6B,$F4,$5A
	.dc.b	$FC,$D6,$BF,$25,$AF,$63,$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$4E,$87
	.dc.b	$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86
	.dc.b	$8A,$68,$C1,$0C,$12,$51,$F2,$1B,$1E,$F6,$BE,$86,$BE,$36,$BE,$06
	.dc.b	$BD,$DA,$F7,$8E,$C4,$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$74
	.dc.b	$21,$D0,$A7,$42,$9D,$0A,$74,$29,$D0,$E4,$43,$91,$4E,$45,$39,$14
	.dc.b	$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$47,$D4
	.dc.b	$24,$7C,$96,$C7,$C0,$6B,$E9,$6B,$E4,$6B,$E1,$6B,$E0,$6B,$E0,$8E
	.dc.b	$C4,$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$74,$21,$D0,$A7,$42
	.dc.b	$9D,$0A,$74,$29,$D0,$E4,$43,$91,$4E,$45,$39,$14,$E4,$70,$3F,$55
	.dc.b	$F2,$1F,$D6,$88,$FF,$5E,$6B,$FD,$5D,$AF,$F5,$36,$BF,$D3,$DA,$FF
	.dc.b	$4B,$6B,$FD,$1D,$B2,$69,$23,$C0,$43,$C0,$53,$C0,$53,$C0,$53,$C0
	.dc.b	$53,$C0,$58,$F0,$36,$3C,$0F,$62,$1D,$8A,$76,$29,$D8,$A7,$62,$C7
	.dc.b	$66,$C7,$67,$A1,$0E,$85,$03,$E6,$5E,$07,$FB,$59,$B1,$FE,$DF,$6B
	.dc.b	$FE,$E3,$6B,$FE,$ED,$6B,$D0,$E9,$1C,$C8,$FF,$9B,$3A,$7F,$D1,$7C
	.dc.b	$7F,$EA,$9E,$AF,$F4,$84,$FF,$A7,$C8,$7F,$A0,$C8,$FF,$A0,$FA,$7F
	.dc.b	$D1,$0F,$08,$FF,$55,$1F,$E9,$B3,$FC,$EB,$D7,$FE,$93,$7F,$D2,$35
	.dc.b	$FE,$B4,$7F,$D1,$F3,$CD,$FF,$3C,$DF,$FA,$72,$FA,$9E,$BA,$5F,$FA
	.dc.b	$91,$E5,$7E,$12,$4F,$0B,$FF,$41,$D6,$EF,$A4,$76,$EB,$FF,$52,$BE
	.dc.b	$83,$FD,$58,$E2,$F9,$C6,$CF,$4C,$F2,$7F,$D5,$47,$AA,$FC,$0F,$C0
	.dc.b	$3F,$D6,$13,$FE,$CB,$AE,$93,$FE,$C0,$9E,$E7,$48,$CB,$FE,$C2,$8F
	.dc.b	$81,$7E,$99,$FF,$57,$CF,$FA,$7B,$AF,$FD,$5E,$4E,$BF,$F5,$3A,$69
	.dc.b	$23,$FD,$CF,$1B,$5D,$5F,$EA,$BC,$FF,$D5,$1D,$1F,$E8,$89,$FF,$3F
	.dc.b	$1F,$E9,$67,$FC,$EF,$CF,$FE,$73,$E9,$FF,$3D,$92,$7F,$D5,$91,$97
	.dc.b	$FD,$0B,$D5,$FE,$8B,$7F,$D2,$07,$FA,$AB,$AB,$FC,$FD,$19,$8E,$F7
	.dc.b	$E1,$3B,$C7,$FA,$CA,$77,$2E,$AF,$F4,$F5,$FF,$AB,$70,$F1,$4F,$FA
	.dc.b	$8C,$D3,$FE,$9A,$7C,$B7,$CD,$FF,$4F,$9F,$E9,$7A,$FF,$4A,$5F,$67
	.dc.b	$AC,$E9,$B3,$FE,$9C,$BF,$F4,$AC,$F3,$8F,$F6,$0D,$F8,$B0,$FF,$4E
	.dc.b	$3F,$EA,$93,$15,$C2,$7C,$11,$8A,$F8,$55,$F0,$27,$FD,$53,$87,$EE
	.dc.b	$F0,$FF,$50,$F5,$3A,$9F,$25,$FE,$9F,$25,$FF,$A7,$5A,$FF,$03,$3F
	.dc.b	$D5,$0E,$1F,$E0,$D3,$CE,$3F,$D3,$BC,$52,$7F,$D4,$2D,$7F,$D3,$4D
	.dc.b	$7F,$D2,$CD,$7F,$D2,$4D,$7F,$D1,$CD,$7F,$D1,$4D,$E8,$34,$91,$E0
	.dc.b	$21,$E0,$29,$E0,$29,$E0,$29,$E0,$29,$E0,$2C,$78,$1B,$1E,$07,$B1
	.dc.b	$0E,$C5,$3B,$14,$EC,$53,$B1,$63,$B3,$63,$B3,$D0,$87,$42,$9D,$0F
	.dc.b	$F6,$17,$8D,$FE,$D0,$23,$FE,$D4,$6B,$FE,$CD,$6B,$C0,$D1,$0D,$14
	.dc.b	$D1,$82,$18,$21,$FE,$CD,$7A,$B4,$E4,$53,$89,$4E,$27,$FB,$3D,$E2
	.dc.b	$D3,$89,$4E,$25,$38,$9F,$ED,$17,$8B,$4E,$25,$38,$94,$E3,$28,$E2
	.dc.b	$D8,$E2,$D7,$8B,$5E,$31,$C0,$8F,$06,$BC,$1A,$F0,$34,$43,$45,$34
	.dc.b	$60,$86,$08,$7C,$B2,$1F,$14,$A7,$BC,$B1,$E9,$6C,$79,$1A,$78,$4A
	.dc.b	$78,$0A,$78,$0E,$C4,$3B,$16,$3B,$36,$3B,$34,$EC,$53,$B1,$4E,$C7
	.dc.b	$42,$47,$46,$C7,$46,$9D,$0A,$74,$29,$D2,$39,$12,$39,$36,$39,$35
	.dc.b	$E4,$D7,$94,$70,$23,$C1,$AF,$06,$BC,$0D,$10,$D1,$4D,$18,$21,$82
	.dc.b	$1F,$50,$87,$C9,$29,$F0,$0B,$1E,$96,$C7,$91,$A7,$84,$A7,$80,$A7
	.dc.b	$80,$EC,$43,$B1,$63,$B3,$63,$B3,$4E,$C5,$3B,$14,$EC,$74,$24,$74
	.dc.b	$6C,$74,$7F,$0B,$F2,$5F,$D6,$DF,$F2,$F4,$A3,$11,$FE,$6C,$8C,$38
	.dc.b	$6B,$C5,$AF,$F6,$56,$BD,$23,$9A,$8C,$BD,$63,$31,$88,$D4,$60,$8F
	.dc.b	$85,$AF,$85,$AF,$85,$AF,$84,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2
	.dc.b	$9A,$30,$43,$04,$3D,$C7,$52,$9C,$8A,$71,$29,$C4,$F8,$67,$12,$9C
	.dc.b	$4A,$71,$29,$C6,$3E,$61,$23,$E2,$C7,$FA,$A3,$A5,$EF,$FF,$A9,$C9
	.dc.b	$1F,$EA,$76,$BF,$EA,$76,$BF,$EA,$77,$EE,$F1,$7F,$A8,$71,$FF,$3A
	.dc.b	$36,$7F,$CE,$47,$FA,$28,$FF,$3D,$C7,$FA,$4F,$A7,$FD,$52,$4E,$7F
	.dc.b	$F3,$AC,$FF,$A0,$4F,$F4,$65,$73,$71,$D3,$FE,$9F,$7F,$D1,$2F,$FA
	.dc.b	$39,$CD,$FF,$A2,$A7,$FD,$3F,$1E,$03,$FD,$58,$E1,$FF,$59,$C6,$FA
	.dc.b	$7F,$D0,$D1,$FE,$8B,$25,$7F,$A4,$5B,$3E,$9B,$F3,$DF,$3A,$F1,$3F
	.dc.b	$EA,$C9,$FF,$48,$D7,$84,$FF,$4B,$CF,$32,$47,$38,$FF,$51,$1F,$EA
	.dc.b	$78,$CB,$E8,$BF,$F6,$14,$62,$39,$CF,$18,$F0,$CE,$A3,$07,$8A,$7E
	.dc.b	$5E,$1F,$3D,$EE,$7C,$A8,$D5,$7A,$A5,$87,$BB,$A3,$FD,$41,$D1,$FE
	.dc.b	$9F,$7F,$D4,$D8,$7F,$A8,$BB,$FF,$D9,$D1,$88,$FF,$34,$66,$8C,$46
	.dc.b	$BA,$BF,$D5,$8E,$A3,$AD,$66,$3A,$1F,$EA,$23,$FD,$26,$7F,$A8,$8F
	.dc.b	$F4,$B7,$2B,$A7,$FD,$0D,$1F,$EA,$53,$FD,$20,$E1,$DB,$7A,$3F,$D1
	.dc.b	$B1,$FE,$9C,$3F,$D1,$67,$13,$FD,$26,$78,$67,$FD,$32,$7F,$A7,$9E
	.dc.b	$A7,$17,$C4,$D9,$ED,$D5,$FE,$BB,$8F,$F6,$17,$4F,$FA,$96,$7F,$D5
	.dc.b	$91,$E4,$3D,$97,$E4,$70,$7F,$AF,$2B,$FD,$6C,$F9,$6F,$BB,$D0,$F1
	.dc.b	$9C,$27,$FD,$45,$1A,$7A,$18,$AF,$2F,$57,$FA,$C6,$75,$7D,$AF,$93
	.dc.b	$FE,$AF,$8C,$CF,$FA,$BB,$0F,$55,$7A,$2B,$DD,$CC,$EE,$FE,$7A,$49
	.dc.b	$FE,$94,$D7,$FA,$53,$5F,$E9,$4F,$F4,$4E,$E9,$1F,$D4,$5A,$FF,$51
	.dc.b	$6B,$FD,$45,$AF,$F5,$C6,$BF,$D6,$DA,$FF,$53,$6C,$9A,$48,$F0,$10
	.dc.b	$F0,$14,$F0,$14,$F0,$14,$F0,$14,$F0,$16,$3C,$0D,$8F,$03,$D8,$87
	.dc.b	$62,$9D,$8A,$76,$29,$D8,$B1,$D9,$B1,$D9,$E8,$43,$A1,$4E,$87,$F7
	.dc.b	$DE,$EF,$EE,$E4,$7F,$C0,$8D,$7F,$C0,$8D,$78,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$3F,$C2,$EF,$56,$9C,$8A,$71,$29,$C4,$FF,$0E,$3C,$5A,$71
	.dc.b	$29,$C4,$A7,$13,$FC,$34,$F1,$69,$C4,$A7,$12,$9C,$65,$1C,$5B,$1C
	.dc.b	$5A,$F1,$6B,$C6,$38,$11,$E0,$D7,$83,$5E,$06,$88,$68,$A6,$8C,$10
	.dc.b	$C1,$0F,$96,$43,$E2,$94,$F7,$96,$3D,$2D,$8F,$23,$4F,$09,$4F,$01
	.dc.b	$4F,$01,$D8,$87,$62,$C7,$66,$C7,$66,$9D,$8A,$76,$29,$D8,$E8,$48
	.dc.b	$E8,$D8,$E8,$D3,$A1,$4E,$85,$3A,$47,$22,$47,$26,$C7,$26,$BC,$9A
	.dc.b	$F2,$8E,$04,$78,$35,$E0,$D7,$81,$A2,$1A,$29,$A3,$04,$30,$43,$EA
	.dc.b	$10,$F9,$25,$3E,$01,$63,$D2,$D8,$F2,$34,$F0,$94,$F0,$14,$F0,$1D
	.dc.b	$88,$76,$2C,$76,$6C,$76,$69,$D8,$A7,$62,$9D,$8E,$84,$8E,$8D,$8E
	.dc.b	$8F,$E1,$BA,$1D,$DF,$E9,$E6,$C7,$FA,$A9,$AF,$FA,$BD,$A8,$81,$CB
	.dc.b	$86,$BF,$E7,$78,$EC,$47,$B3,$5E,$CD,$7A,$9D,$5C,$3D,$5A,$F5,$39
	.dc.b	$90,$E6,$53,$9C,$79,$1E,$86,$4E,$85,$3A,$1C,$88,$72,$8F,$0B,$C8
	.dc.b	$C9,$C8,$A7,$23,$81,$D4,$E0,$60,$D9,$83,$91,$92,$18,$21,$D4,$A7
	.dc.b	$13,$04,$30,$48,$F8,$43,$E2,$FF,$04,$48,$FC,$13,$5F,$C1,$35,$F9
	.dc.b	$6F,$FA,$AB,$9B,$FC,$E9,$CF,$FE,$7A,$9F,$F4,$6C,$7F,$9C,$9F,$F4
	.dc.b	$89,$FE,$83,$33,$CF,$1F,$F3,$DF,$37,$FA,$45,$FF,$58,$1F,$E8,$08
	.dc.b	$FF,$4B,$F4,$FF,$A7,$AF,$FD,$15,$5F,$EA,$D3,$FD,$35,$D3,$FE,$A3
	.dc.b	$BF,$F4,$C4,$66,$BF,$D1,$91,$E1,$3F,$D5,$27,$FA,$5E,$7F,$D1,$EF
	.dc.b	$82,$BF,$D1,$57,$E5,$7D,$91,$C0,$FF,$48,$1B,$9F,$F6,$4E,$1F,$09
	.dc.b	$FF,$49,$98,$AF,$4C,$F9,$27,$4F,$13,$FD,$9E,$7F,$B0,$8B,$87,$B1
	.dc.b	$B8,$7C,$1B,$F3,$C6,$C9,$3F,$25,$B3,$F0,$1F,$F5,$79,$F0,$70,$FF
	.dc.b	$69,$4E,$CE,$D1,$CA,$FE,$C4,$FC,$EE,$7F,$F4,$EC,$FF,$01,$AF,$F5
	.dc.b	$27,$47,$F0,$3E,$9F,$F5,$63,$FE,$A4,$9F,$F5,$11,$FE,$A5,$7F,$76
	.dc.b	$F1,$AE,$6A,$B9,$BF,$EA,$83,$FD,$08,$7F,$A6,$8B,$CD,$FE,$7A,$7F
	.dc.b	$D5,$67,$F9,$FB,$9B,$FC,$F8,$FF,$A2,$BA,$3F,$D0,$FC,$FF,$E7,$E8
	.dc.b	$FF,$49,$4E,$63,$13,$FE,$8B,$9F,$F5,$01,$0F,$F4,$94,$68,$8F,$5E
	.dc.b	$7F,$F4,$67,$47,$FA,$4A,$3F,$D6,$27,$8C,$F2,$D7,$57,$99,$C6,$FC
	.dc.b	$EF,$B6,$3F,$D3,$B5,$FE,$AF,$6C,$FF,$A7,$6B,$F9,$7C,$62,$BB,$B7
	.dc.b	$0F,$F5,$EC,$EA,$F9,$1F,$EA,$39,$D4,$72,$C3,$F9,$CD,$7F,$AC,$B9
	.dc.b	$9C,$3F,$5B,$1F,$E7,$D1,$87,$0D,$7E,$DB,$5F,$E6,$ED,$7F,$A2,$B5
	.dc.b	$E7,$1D,$54,$65,$BD,$99,$83,$49,$1E,$26,$BE,$26,$BE,$26,$BE,$26
	.dc.b	$BE,$C6,$BE,$03,$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$4E,$87,$22,$1C
	.dc.b	$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68
	.dc.b	$C1,$0C,$12,$3E,$9B,$ED,$6B,$E6,$6B,$F8,$2E,$CF,$B4,$47,$F1,$2D
	.dc.b	$7F,$12,$D7,$81,$A2,$1A,$29,$A3,$04,$30,$43,$F2,$2F,$56,$9C,$8A
	.dc.b	$71,$29,$C4,$FD,$0B,$C5,$A7,$12,$9C,$4A,$71,$3F,$3A,$F1,$69,$C4
	.dc.b	$A7,$12,$9C,$61,$1C,$5B,$1C,$5A,$F1,$6B,$C6,$38,$11,$E0,$D7,$83
	.dc.b	$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$96,$43,$E2,$94,$F7,$96,$3D
	.dc.b	$2D,$8F,$23,$4F,$09,$4F,$01,$4F,$01,$D8,$87,$62,$C7,$66,$C7,$66
	.dc.b	$9D,$8A,$76,$29,$D8,$E8,$48,$E8,$D8,$E8,$D3,$A1,$4E,$85,$3A,$47
	.dc.b	$22,$47,$26,$C7,$26,$BC,$9A,$F2,$8E,$04,$78,$35,$E0,$D7,$81,$A2
	.dc.b	$1A,$29,$A3,$04,$30,$43,$EA,$10,$F9,$25,$3E,$01,$63,$D2,$D8,$F2
	.dc.b	$34,$F0,$94,$F0,$14,$F0,$1D,$88,$76,$2C,$76,$6C,$76,$69,$D8,$A7
	.dc.b	$62,$9D,$8E,$84,$8E,$8D,$8E,$8F,$E1,$B9,$9D,$9F,$EA,$F2,$47,$FA
	.dc.b	$BD,$AF,$FA,$BD,$AF,$FA,$61,$AF,$FA,$B5,$AF,$FA,$CE,$3A,$BC,$C8
	.dc.b	$73,$29,$CC,$A7,$32,$9C,$E3,$AB,$FE,$BD,$6C,$76,$8E,$84,$3A,$14
	.dc.b	$E8,$53,$A1,$D8,$E8,$57,$A3,$C8,$87,$22,$9C,$8E,$87,$22,$BC,$9E
	.dc.b	$04,$38,$1C,$8E,$05,$78,$3A,$38,$1A,$2A,$FA,$2F,$F7,$41,$23,$FD
	.dc.b	$D0,$D7,$FD,$D0,$D7,$FD,$A2,$FF,$A9,$F9,$FF,$D0,$71,$FE,$9D,$3F
	.dc.b	$D4,$3C,$DF,$E7,$0C,$FF,$CF,$BC,$5F,$E7,$42,$46,$67,$FC,$F5,$3F
	.dc.b	$E9,$F7,$97,$37,$FA,$58,$FF,$3E,$CE,$E3,$FD,$02,$4A,$FF,$52,$3F
	.dc.b	$E9,$DA,$EF,$CF,$FE,$8D,$8E,$A7,$FA,$94,$A7,$FA,$80,$FF,$46,$9C
	.dc.b	$4C,$CF,$FA,$2E,$39,$47,$7A,$FF,$4F,$BC,$0F,$F5,$54,$ED,$FF,$52
	.dc.b	$61,$F1,$30,$FF,$58,$C7,$FB,$20,$E5,$19,$C3,$FD,$59,$1F,$22,$3C
	.dc.b	$4F,$59,$F8,$31,$8A,$F9,$58,$7F,$A8,$DF,$83,$5E,$2C,$3E,$34,$62
	.dc.b	$FF,$CB,$CA,$33,$CB,$FE,$98,$8F,$F4,$BF,$3F,$FA,$82,$FF,$D3,$B1
	.dc.b	$FE,$9A,$7F,$D4,$06,$25,$CF,$FE,$6D,$9E,$3D,$AA,$FF,$52,$4F,$FA
	.dc.b	$AE,$31,$1F,$EA,$5A,$E2,$FF,$A9,$6B,$8B,$C7,$9B,$FC,$FA,$FF,$A6
	.dc.b	$5F,$F5,$0B,$FE,$8F,$3F,$D2,$7C,$BF,$E9,$A7,$FD,$21,$CF,$FE,$85
	.dc.b	$3F,$D0,$FD,$1F,$EA,$58,$E1,$CF,$FE,$8C,$7B,$4E,$9F,$F4,$EB,$63
	.dc.b	$FD,$68,$4E,$8F,$F4,$74,$F2,$6B,$E0,$AE,$CD,$39,$BF,$E9,$F7,$BB
	.dc.b	$A9,$E9,$18,$3F,$D4,$35,$E1,$AF,$F5,$05,$71,$9F,$F6,$3C,$61,$EA
	.dc.b	$70,$8D,$61,$F2,$23,$FD,$6A,$78,$EF,$BB,$86,$C6,$C6,$1F,$EA,$AB
	.dc.b	$FB,$56,$F8,$C7,$9D,$FE,$AB,$8C,$36,$33,$8F,$FB,$92,$72,$D8,$CB
	.dc.b	$F8,$21,$1A,$70,$D7,$F0,$ED,$7F,$D6,$3F,$39,$FE,$AB,$23,$FE,$BA
	.dc.b	$6B,$FE,$BA,$6B,$FE,$BA,$6B,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21
	.dc.b	$A2,$9A,$30,$43,$04,$3F,$DA,$6F,$91,$A7,$72,$9D,$0A,$73,$29,$CC
	.dc.b	$FF,$68,$BC,$DA,$73,$29,$CC,$A7,$32,$9C,$FA,$91,$CD,$B1,$CD,$AF
	.dc.b	$36,$BC,$DA,$F3,$F8,$FF,$F7,$11,$1F,$F7,$8B,$5F,$F7,$8B,$5F,$F7
	.dc.b	$8B,$5F,$F7,$8B,$5F,$F7,$DB,$5F,$F7,$8B,$60,$D2,$47,$80,$87,$80
	.dc.b	$A7,$80,$A7,$80,$A7,$80,$A7,$80,$B1,$E0,$6C,$78,$1E,$C4,$3B,$14
	.dc.b	$EC,$53,$B1,$4E,$C5,$8E,$CD,$8E,$CF,$42,$1D,$0A,$74,$29,$D0,$B1
	.dc.b	$D1,$B1,$D1,$E4,$43,$91,$4E,$45,$8E,$4D,$8E,$4F,$02,$1C,$0B,$1C
	.dc.b	$1B,$1C,$1D,$12,$34,$D8,$D4,$60,$8E,$09,$1F,$50,$87,$C9,$29,$F0
	.dc.b	$0A,$7A,$4A,$79,$0A,$78,$4B,$1E,$06,$C7,$81,$EC,$43,$B1,$4E,$C5
	.dc.b	$3B,$14,$EC,$58,$EC,$D8,$EC,$F4,$21,$D0,$A7,$42,$9D,$0B,$1D,$1B
	.dc.b	$1D,$1E,$44,$39,$14,$E4,$58,$E4,$D8,$E4,$F0,$21,$C0,$B1,$C1,$B1
	.dc.b	$C1,$D1,$23,$4D,$8D,$46,$08,$E0,$91,$F5,$08,$7C,$92,$9F,$00,$A7
	.dc.b	$A4,$A7,$90,$A7,$84,$B1,$E0,$6C,$78,$1E,$C7,$F6,$11,$D2,$F5,$71
	.dc.b	$1F,$EA,$16,$BA,$6C,$6E,$36,$D7,$A3,$5F,$ED,$CD,$7F,$C6,$2D,$7B
	.dc.b	$1C,$A3,$A9,$1C,$10,$EA,$E5,$A6,$4E,$EE,$5A,$64,$F0,$B9,$69,$93
	.dc.b	$C6,$74,$21,$C8,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$3C
	.dc.b	$F2,$14,$EE,$53,$A1,$4E,$65,$39,$C7,$CF,$70,$4F,$83,$79,$9F,$79
	.dc.b	$B1,$F8,$26,$BF,$87,$6B,$FE,$29,$6B,$FE,$6C,$6B,$FE,$71,$6A,$6F
	.dc.b	$17,$FA,$0F,$AF,$FD,$5D,$D3,$FE,$94,$3F,$D0,$BD,$3F,$E9,$C9,$FF
	.dc.b	$46,$DF,$FA,$3E,$31,$D5,$FE,$A2,$9E,$53,$FE,$AB,$3F,$D3,$67,$FA
	.dc.b	$64,$FF,$4F,$3A,$7F,$D5,$CF,$47,$FD,$2D,$D7,$FE,$9B,$9E,$95,$DA
	.dc.b	$7F,$D5,$27,$3B,$F2,$D7,$FA,$9E,$3F,$D1,$D7,$FE,$CA,$8E,$51,$C2
	.dc.b	$7F,$D9,$18,$7B,$8F,$F5,$6B,$88,$F4,$57,$94,$FF,$5C,$3F,$ED,$09
	.dc.b	$FF,$4B,$C7,$9E,$7F,$D6,$33,$EC,$BF,$33,$B3,$C7,$87,$AB,$0F,$34
	.dc.b	$2F,$CC,$FC,$4E,$8F,$F5,$81,$FE,$A2,$EB,$FF,$58,$47,$FA,$B8,$87
	.dc.b	$8A,$3F,$D4,$EF,$FA,$CA,$3F,$D6,$C4,$EA,$D7,$5F,$FA,$BE,$7F,$D1
	.dc.b	$9D,$5F,$E7,$EE,$BF,$F3,$7C,$7F,$AA,$0F,$F4,$D1,$FE,$AF,$EA,$FF
	.dc.b	$55,$4E,$F9,$FF,$D1,$75,$FE,$AB,$69,$FE,$99,$EA,$FF,$49,$4E,$EF
	.dc.b	$A1,$FE,$8F,$8C,$91,$EC,$78,$07,$5F,$FA,$41,$C3,$5E,$47,$AA,$39
	.dc.b	$3C,$30,$F3,$C6,$27,$FD,$35,$18,$9E,$04,$3F,$D4,$13,$BC,$3F,$D5
	.dc.b	$D1,$9A,$FF,$66,$BE,$03,$CB,$1A,$6E,$1F,$09,$FF,$58,$E1,$FE,$AA
	.dc.b	$C3,$C1,$1F,$EB,$63,$FD,$C5,$D0,$F6,$3F,$9D,$B6,$3F,$9F,$B5,$FE
	.dc.b	$88,$D7,$FD,$4C,$D7,$FD,$D2,$D7,$AC,$74,$3A,$FC,$E4,$8E,$AD,$7A
	.dc.b	$B5,$EA,$D7,$AB,$5F,$4B,$54,$A3,$0D,$8E,$E4,$3B,$94,$EE,$53,$B9
	.dc.b	$4E,$E5,$3B,$C7,$62,$39,$8F,$03,$D8,$87,$62,$9D,$8A,$76,$29,$D8
	.dc.b	$E9,$1D,$9C,$C7,$67,$A1,$0E,$85,$3A,$14,$E8,$72,$8F,$C0,$FB,$12
	.dc.b	$3F,$02,$D7,$F0,$2D,$7F,$02,$D7,$F0,$2D,$7F,$1A,$D7,$EA,$B6,$4D
	.dc.b	$24,$78,$08,$78,$0A,$78,$0A,$78,$0A,$78,$0A,$78,$0B,$1E,$06,$C7
	.dc.b	$81,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$58,$EC,$D8,$EC,$F4,$21,$D0
	.dc.b	$A7,$42,$9D,$0B,$1D,$1B,$1D,$1E,$44,$39,$14,$E4,$58,$E4,$D8,$E4
	.dc.b	$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D,$8D,$46,$08,$E0,$91,$F5
	.dc.b	$08,$7C,$92,$9F,$00,$A7,$A4,$A7,$90,$A7,$84,$B1,$E0,$6C,$78,$1E
	.dc.b	$C4,$3B,$14,$EC,$53,$B1,$4E,$C5,$8E,$CD,$8E,$CF,$42,$1D,$0A,$74
	.dc.b	$29,$D0,$B1,$D1,$B1,$D1,$E4,$43,$91,$4E,$45,$8E,$4D,$8E,$4F,$02
	.dc.b	$1C,$0B,$1C,$1B,$1C,$1D,$12,$34,$D8,$D4,$60,$8E,$09,$1F,$50,$87
	.dc.b	$C9,$29,$F0,$0A,$7A,$4A,$79,$0A,$78,$4B,$1E,$06,$C7,$81,$EC,$7E
	.dc.b	$EF,$A5,$BF,$FA,$70,$91,$FE,$9C,$6B,$FE,$9C,$6B,$FE,$9C,$6B,$FE
	.dc.b	$9C,$6A,$3F,$D5,$AE,$1A,$75,$21,$D4,$A7,$52,$9D,$4A,$75,$8F,$4B
	.dc.b	$83,$A9,$CC,$87,$32,$9C,$CA,$73,$8F,$34,$74,$8C,$9D,$0E,$44,$39
	.dc.b	$14,$E4,$78,$DE,$4E,$4E,$47,$02,$1C,$0E,$EF,$07,$27,$03,$47,$37
	.dc.b	$4E,$0F,$F4,$97,$6F,$FA,$6C,$8F,$FA,$6D,$AF,$FA,$6D,$AF,$FA,$6D
	.dc.b	$FF,$6F,$F1,$FF,$9C,$F9,$9C,$BF,$E7,$CA,$FF,$37,$1F,$E6,$BE,$6F
	.dc.b	$F3,$73,$FE,$85,$E8,$FF,$3B,$D7,$F9,$D3,$A3,$FD,$39,$18,$BF,$F3
	.dc.b	$9E,$1F,$E7,$29,$FF,$45,$DF,$13,$FD,$07,$87,$48,$FF,$43,$47,$FA
	.dc.b	$62,$31,$7F,$E8,$46,$C7,$FA,$00,$FF,$48,$46,$5A,$64,$F4,$D7,$FA
	.dc.b	$2E,$3F,$D2,$91,$9A,$FF,$46,$9E,$43,$C3,$3F,$E9,$47,$D7,$1E,$63
	.dc.b	$A9,$F0,$CB,$1D,$5F,$F4,$EE,$1F,$0E,$BF,$D4,$27,$96,$3C,$77,$EA
	.dc.b	$79,$18,$39,$B9,$AE,$91,$F0,$CF,$97,$5F,$59,$5A,$FF,$D2,$70,$E7
	.dc.b	$FF,$4B,$3F,$EB,$8E,$8F,$F5,$06,$BA,$9F,$F5,$0C,$F0,$3F,$D4,$26
	.dc.b	$FA,$3F,$78,$FF,$A8,$67,$81,$1D,$1D,$DF,$F3,$F7,$2F,$F9,$DF,$97
	.dc.b	$FD,$2C,$3A,$39,$F4,$7F,$A3,$E3,$31,$FE,$8B,$E7,$FF,$3D,$CB,$A3
	.dc.b	$FD,$45,$8F,$FA,$02,$B8,$D6,$A7,$37,$FE,$83,$AF,$F4,$DC,$E7,$0F
	.dc.b	$F5,$44,$7F,$A3,$EB,$FD,$19,$87,$FA,$62,$36,$43,$FD,$24,$E5,$B1
	.dc.b	$C0,$93,$ED,$C3,$DF,$3F,$EB,$2C,$3D,$6F,$FA,$77,$0F,$13,$E9,$30
	.dc.b	$77,$8D,$BF,$EA,$61,$4F,$D9,$9E,$97,$D2,$9F,$A7,$39,$8C,$47,$F3
	.dc.b	$18,$C3,$86,$BC,$5A,$F2,$9E,$AE,$DA,$ED,$AF,$FA,$93,$E5,$FF,$D5
	.dc.b	$64,$7F,$D5,$6D,$7F,$D5,$6D,$78,$1A,$21,$A2,$9A,$30,$43,$04,$3F
	.dc.b	$D6,$6F,$56,$9C,$8A,$71,$29,$C4,$FF,$62,$BC,$5A,$71,$29,$C4,$A7
	.dc.b	$13,$FD,$66,$F1,$69,$C4,$A7,$12,$9C,$7A,$51,$C5,$B1,$C5,$AF,$16
	.dc.b	$BC,$63,$81,$1E,$0F,$FB,$7B,$E4,$24,$7F,$B7,$9A,$FF,$B7,$9A,$FF
	.dc.b	$B7,$9A,$FF,$B7,$9A,$FF,$BA,$9A,$FF,$B4,$9B,$26,$92,$3C,$04,$3C
	.dc.b	$05,$3C,$05,$3C,$05,$3C,$05,$3C,$05,$8F,$03,$63,$C0,$F6,$21,$D8
	.dc.b	$A7,$62,$9D,$8A,$76,$2C,$76,$6C,$76,$7A,$10,$E8,$53,$A1,$4E,$85
	.dc.b	$8E,$8D,$8E,$8F,$22,$1C,$8A,$72,$2C,$72,$6C,$72,$78,$10,$E0,$58
	.dc.b	$E0,$D8,$E0,$E8,$91,$A6,$C6,$A3,$04,$70,$48,$FA,$84,$3E,$49,$4F
	.dc.b	$80,$53,$D2,$53,$C8,$53,$C2,$58,$F0,$36,$3C,$0F,$62,$1D,$8A,$76
	.dc.b	$29,$D8,$A7,$62,$C7,$66,$C7,$67,$A1,$0E,$85,$3A,$14,$E8,$58,$E8
	.dc.b	$D8,$E8,$F2,$21,$C8,$A7,$22,$C7,$26,$C7,$27,$81,$0E,$05,$8E,$0D
	.dc.b	$8E,$0E,$89,$1A,$6C,$6A,$30,$47,$04,$8F,$A8,$43,$E4,$94,$F8,$05
	.dc.b	$3D,$25,$3C,$85,$3C,$25,$8F,$03,$63,$C0,$F6,$3F,$77,$D2,$F6,$7F
	.dc.b	$B2,$C7,$FA,$25,$AA,$34,$E9,$AF,$F6,$C6,$BF,$DC,$5A,$FF,$89,$1A
	.dc.b	$F5,$39,$47,$68,$CC,$77,$8C,$11,$D9,$0F,$03,$96,$99,$3C,$4E,$5A
	.dc.b	$64,$F2,$39,$69,$93,$CC,$74,$21,$C8,$A7,$02,$9C,$0D,$10,$D1,$4D
	.dc.b	$18,$21,$82,$1F,$0C,$F2,$14,$EE,$53,$A1,$4E,$65,$39,$C7,$CE,$3E
	.dc.b	$7F,$73,$E6,$BE,$FB,$67,$F0,$6D,$7F,$C5,$2D,$7F,$C5,$CD,$7F,$CC
	.dc.b	$2D,$7F,$DB,$CD,$4D,$E9,$FF,$55,$75,$FF,$AA,$7A,$7F,$D1,$34,$9C
	.dc.b	$BF,$E7,$8E,$9F,$F5,$33,$FE,$A3,$35,$D5,$FE,$89,$7F,$D2,$D1,$FE
	.dc.b	$98,$7F,$D0,$F5,$FE,$98,$8E,$2F,$FA,$62,$7F,$D0,$8F,$FA,$6E,$3F
	.dc.b	$D2,$0F,$8A,$BC,$2E,$6B,$A9,$0F,$F4,$71,$FE,$93,$23,$C6,$BC,$91
	.dc.b	$E2,$AF,$25,$7F,$AF,$5F,$19,$E0,$3F,$D4,$31,$D6,$34,$FF,$AA,$CE
	.dc.b	$B1,$A7,$A1,$B9,$F5,$C7,$48,$E5,$D7,$FE,$94,$BF,$85,$1C,$DF,$36
	.dc.b	$1F,$ED,$09,$D5,$79,$9B,$5F,$EB,$69,$D2,$3E,$A1,$93,$EA,$61,$F2
	.dc.b	$2B,$C5,$1E,$62,$43,$0F,$3C,$7F,$A9,$3A,$BF,$D6,$2F,$E0,$7A,$FF
	.dc.b	$D6,$2E,$7A,$BF,$D6,$0C,$3F,$D5,$67,$FA,$C8,$E1,$D7,$E5,$AE,$78
	.dc.b	$2F,$9D,$68,$FF,$59,$1F,$E9,$FE,$5F,$F4,$B3,$4F,$F4,$B7,$4F,$FA
	.dc.b	$63,$A3,$FD,$51,$1F,$E9,$E9,$DA,$EB,$FF,$51,$4F,$FA,$59,$FF,$5A
	.dc.b	$4F,$C7,$8F,$F5,$E4,$FC,$99,$FF,$5E,$57,$53,$FD,$79,$1F,$E9,$F7
	.dc.b	$0F,$FA,$88,$F0,$1F,$EA,$53,$C1,$D5,$FE,$A8,$AF,$03,$67,$8C,$E8
	.dc.b	$E8,$7B,$67,$D6,$79,$FA,$FF,$D5,$71,$F0,$0D,$4F,$91,$F0,$CF,$08
	.dc.b	$EB,$7E,$E9,$E0,$7C,$3A,$FF,$67,$CF,$1A,$ED,$87,$C3,$8F,$F6,$AD
	.dc.b	$79,$0E,$11,$FE,$D2,$5D,$0F,$33,$FA,$9B,$63,$FA,$BB,$5F,$F5,$EB
	.dc.b	$5F,$F6,$13,$5F,$F7,$AB,$5F,$CD,$35,$EF,$1D,$BC,$C9,$1D,$9A,$F6
	.dc.b	$6B,$D9,$AF,$66,$BD,$9A,$FB,$1B,$26,$92,$3C,$04,$3C,$05,$3C,$05
	.dc.b	$3C,$05,$3C,$05,$3C,$05,$8F,$03,$63,$C0,$F6,$21,$D8,$A7,$62,$9D
	.dc.b	$8A,$76,$2C,$76,$6C,$76,$7A,$10,$E8,$53,$A1,$FE,$0B,$F3,$3F,$C0
	.dc.b	$C4,$7F,$C2,$4D,$7F,$C2,$4D,$78,$1A,$21,$A2,$9A,$30,$43,$04,$3F
	.dc.b	$C3,$4F,$56,$9C,$8A,$71,$29,$C4,$FE,$51,$D2,$8E,$51,$86,$BF,$E2
	.dc.b	$F6,$C6,$E3,$B1,$B2,$9B,$2B,$B7,$C4,$6C,$A6,$CA,$ED,$F3,$19,$2B
	.dc.b	$97,$D6,$43,$A1,$A8,$F3,$91,$F1,$B5,$EE,$D7,$81,$A2,$1A,$29,$A3
	.dc.b	$04,$30,$43,$C8,$7A,$4B,$1E,$97,$CE,$62,$3C,$4D,$3B,$94,$EC,$53
	.dc.b	$B1,$C8,$87,$72,$C7,$77,$04,$70,$48,$F0,$C7,$06,$BB,$6C,$6E,$3C
	.dc.b	$A6,$CA,$6C,$AE,$DF,$49,$B2,$9B,$2B,$B7,$DA,$64,$AE,$5F,$84,$43
	.dc.b	$A1,$A8,$F3,$91,$F1,$B5,$EE,$D7,$81,$A2,$1A,$29,$A3,$04,$30,$43
	.dc.b	$C8,$7A,$4B,$1E,$97,$CE,$62,$3C,$4D,$3B,$94,$EC,$53,$B1,$C8,$87
	.dc.b	$72,$C7,$77,$04,$70,$48,$F0,$C7,$06,$BB,$6C,$6E,$3C,$A6,$CF,$BF
	.dc.b	$E6,$7F,$AA,$48,$FF,$AD,$5A,$FF,$AD,$5A,$FF,$AD,$5A,$FF,$AD,$5A
	.dc.b	$FF,$16,$6B,$FE,$C1,$6C,$1A,$48,$F0,$10,$F0,$14,$F0,$14,$F0,$14
	.dc.b	$F0,$14,$F0,$16,$3C,$0D,$8F,$03,$D8,$87,$62,$9D,$8A,$76,$29,$D8
	.dc.b	$B1,$D9,$B1,$D9,$E8,$43,$A1,$4F,$F3,$27,$E0,$AF,$33,$FD,$D2,$D8
	.dc.b	$FF,$76,$B5,$FF,$35,$11,$C3,$80,$E1,$D4,$74,$23,$FE,$F8,$E2,$FF
	.dc.b	$38,$3F,$E7,$DE,$7F,$F4,$34,$3A,$7F,$CF,$73,$FE,$72,$E8,$FF,$38
	.dc.b	$BF,$EA,$AE,$9F,$F3,$EB,$FE,$7C,$3F,$D3,$BD,$5F,$E8,$67,$FC,$FB
	.dc.b	$3F,$E9,$CB,$FF,$4A,$C7,$13,$35,$CA,$BF,$D1,$F1,$97,$FD,$61,$1F
	.dc.b	$EB,$FC,$3C,$B8,$79,$2F,$99,$EA,$BE,$13,$FE,$A5,$7F,$D3,$D5,$C5
	.dc.b	$A7,$9E,$78,$13,$0F,$35,$F5,$7D,$F1,$9A,$F3,$5F,$FA,$7E,$FF,$D3
	.dc.b	$98,$7A,$2B,$51,$EC,$24,$7A,$9F,$05,$F6,$78,$D7,$37,$37,$87,$FD
	.dc.b	$34,$43,$FD,$25,$D1,$FE,$96,$E8,$74,$FD,$B6,$CF,$E0,$CF,$F4,$C1
	.dc.b	$FE,$9F,$CB,$FD,$39,$3F,$E9,$FE,$AF,$C5,$C6,$27,$FD,$3A,$76,$8C
	.dc.b	$D7,$FA,$8B,$A7,$FC,$F1,$D1,$FE,$88,$AF,$F3,$94,$63,$A3,$FC,$FB
	.dc.b	$D5,$FE,$8B,$9F,$F4,$19,$FE,$99,$70,$FF,$9E,$0F,$F4,$A4,$FF,$A7
	.dc.b	$1D,$99,$AF,$F5,$0B,$FE,$93,$3F,$D1,$75,$FE,$85,$3F,$D1,$7D,$5F
	.dc.b	$E9,$29,$DD,$73,$AE,$55,$E4,$7C,$F5,$FE,$8E,$7F,$D4,$D5,$FE,$8E
	.dc.b	$7F,$D2,$75,$EB,$7C,$6F,$33,$0F,$43,$8D,$78,$BA,$BF,$D2,$D1,$EF
	.dc.b	$3B,$BF,$E9,$59,$E2,$7F,$B0,$27,$84,$F8,$2F,$C8,$F5,$3E,$05,$F9
	.dc.b	$4F,$6E,$1F,$2A,$F9,$4F,$9A,$BE,$15,$FF,$A9,$A3,$E6,$C7,$49,$F8
	.dc.b	$9D,$2F,$37,$FD,$4A,$4A,$FF,$52,$B5,$FF,$52,$B5,$FF,$52,$BF,$EA
	.dc.b	$CE,$59,$1F,$EA,$C6,$BF,$EA,$C6,$BF,$EA,$C6,$BF,$EA,$C6,$BF,$EA
	.dc.b	$C6,$BF,$EB,$A6,$CF,$F1,$B8,$D1,$DC,$87,$72,$9D,$CA,$77,$29,$C8
	.dc.b	$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3D,$C7,$90
	.dc.b	$A7,$72,$9D,$0A,$73,$29,$CD,$F8,$CE,$08,$E0,$CF,$43,$CD,$FE,$0D
	.dc.b	$3F,$E8,$E6,$A8,$D3,$A6,$BF,$C3,$5A,$FF,$51,$6B,$FD,$59,$AF,$53
	.dc.b	$94,$76,$8C,$C7,$78,$C1,$1D,$90,$F0,$39,$69,$93,$C4,$E5,$A6,$4F
	.dc.b	$23,$96,$99,$3C,$C7,$42,$1C,$8A,$70,$29,$C0,$D1,$0D,$14,$D1,$82
	.dc.b	$18,$21,$F0,$CF,$21,$4E,$E5,$3A,$14,$E6,$53,$9C,$7C,$E2,$47,$C1
	.dc.b	$8C,$B5,$E0,$48,$DB,$E2,$6B,$E2,$6B,$E2,$6B,$E2,$6B,$D8,$E5,$1D
	.dc.b	$E3,$31,$DE,$30,$47,$64,$3C,$0E,$5A,$64,$F1,$39,$69,$93,$C8,$E5
	.dc.b	$A6,$4F,$31,$D0,$87,$22,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08
	.dc.b	$7C,$33,$C8,$53,$B9,$4E,$85,$39,$94,$E7,$1F,$38,$91,$F0,$63,$2D
	.dc.b	$78,$12,$36,$F8,$9A,$F8,$9A,$F8,$9A,$F8,$9A,$F6,$39,$47,$78,$CC
	.dc.b	$77,$8C,$11,$D9,$0F,$03,$96,$99,$3C,$4E,$5A,$64,$F2,$39,$69,$93
	.dc.b	$CC,$74,$21,$C8,$A7,$02,$9F,$3B,$CD,$FE,$38,$47,$F9,$43,$5F,$E5
	.dc.b	$0D,$7F,$94,$35,$FE,$50,$D4,$7F,$33,$70,$D3,$A9,$0E,$A5,$3A,$94
	.dc.b	$EA,$53,$AC,$A3,$AB,$63,$47,$73,$A9,$0E,$A5,$3A,$94,$EB,$1D,$08
	.dc.b	$F4,$8E,$CE,$CE,$C7,$42,$1D,$0A,$74,$39,$47,$47,$94,$74,$76,$74
	.dc.b	$39,$1F,$EA,$9F,$99,$FF,$57,$91,$FF,$57,$B5,$FF,$35,$75,$23,$11
	.dc.b	$FE,$AF,$8C,$3F,$EA,$BE,$8F,$F3,$E3,$FE,$7A,$27,$4F,$FA,$04,$74
	.dc.b	$FF,$9C,$27,$71,$FE,$89,$6C,$7F,$A0,$E3,$1D,$3F,$E9,$0A,$FC,$1F
	.dc.b	$3A,$FF,$D2,$A7,$08,$CB,$89,$E1,$1F,$EB,$4B,$FF,$54,$BF,$E8,$D3
	.dc.b	$FD,$79,$1F,$EA,$56,$9E,$88,$FF,$55,$38,$BF,$4B,$FE,$C7,$3F,$D5
	.dc.b	$6E,$1F,$F5,$89,$E6,$EA,$FF,$5B,$1F,$EA,$BA,$ED,$87,$FA,$E3,$0F
	.dc.b	$45,$79,$4E,$78,$7C,$5A,$F4,$D7,$BD,$E3,$7C,$EB,$CC,$77,$3D,$93
	.dc.b	$D6,$78,$DF,$C3,$9F,$F5,$9D,$FB,$2B,$51,$83,$91,$FE,$C5,$BE,$D8
	.dc.b	$7B,$67,$FD,$4D,$D5,$FE,$B1,$EA,$FF,$55,$5F,$FA,$D6,$7F,$D5,$4F
	.dc.b	$FA,$BF,$A7,$FD,$66,$D9,$DB,$FE,$AD,$38,$47,$FA,$B7,$97,$FC,$EC
	.dc.b	$7F,$9E,$A3,$FD,$14,$7F,$A8,$0F,$F4,$79,$FE,$7F,$E9,$FF,$47,$BF
	.dc.b	$E8,$6E,$7F,$F4,$5F,$4F,$FA,$1A,$3F,$D2,$27,$FA,$0A,$77,$D1,$FE
	.dc.b	$9B,$BE,$F1,$8E,$9F,$F4,$64,$7F,$A1,$DB,$1C,$23,$C9,$5D,$CF,$11
	.dc.b	$FE,$99,$8F,$F4,$F4,$74,$8C,$1E,$12,$DF,$FB,$25,$F6,$1F,$E9,$FB
	.dc.b	$F2,$57,$22,$47,$19,$F6,$3C,$27,$FD,$57,$7E,$68,$FF,$65,$BF,$CA
	.dc.b	$14,$B0,$FE,$57,$E8,$BD,$27,$CC,$7F,$98,$A3,$0D,$8C,$DF,$FB,$AE
	.dc.b	$72,$D8,$CB,$DC,$46,$9C,$35,$F2,$35,$F3,$75,$7F,$0A,$7F,$DF,$0D
	.dc.b	$7A,$35,$DF,$3F,$F9,$DA,$31,$1F,$D1,$9A,$9A,$48,$E0,$48,$FF,$40
	.dc.b	$C6,$5B,$19,$7C,$6D,$7C,$6D,$7C,$67,$02,$1C,$0A,$70,$29,$C0,$D1
	.dc.b	$0D,$14,$D1,$82,$18,$21,$F0,$5F,$23,$4E,$E5,$3A,$14,$E6,$53,$99
	.dc.b	$F2,$5E,$6D,$39,$94,$E6,$53,$99,$4E,$7D,$0C,$FF,$D4,$04,$8F,$F5
	.dc.b	$63,$5F,$F5,$63,$5F,$F5,$63,$5F,$F5,$63,$5F,$EF,$2D,$7F,$D7,$0D
	.dc.b	$8F,$04,$76,$21,$D8,$A7,$62,$9D,$8A,$76,$29,$D8,$AF,$67,$A1,$0E
	.dc.b	$85,$3A,$14,$E8,$53,$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8,$AF,$27
	.dc.b	$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$3E,$A1,$23
	.dc.b	$E4,$B5,$F8,$0D,$7D,$2D,$7C,$8D,$7C,$2D,$7C,$0D,$8F,$04,$76,$21
	.dc.b	$D8,$A7,$62,$9D,$8A,$76,$29,$D8,$AF,$67,$A1,$0E,$85,$3A,$14,$E8
	.dc.b	$53,$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8,$AF,$27,$81,$0E,$05,$38
	.dc.b	$15,$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$3E,$A1,$23,$E4,$B5,$F8,$0D
	.dc.b	$7D,$2D,$7C,$8D,$7C,$2D,$7C,$0D,$8F,$04,$76,$21,$D8,$A7,$62,$9D
	.dc.b	$8A,$76,$29,$D8,$AF,$67,$A1,$0E,$85,$3A,$1F,$7F,$C2,$72,$FF,$69
	.dc.b	$8F,$EF,$11,$87,$01,$C3,$B6,$BF,$DC,$5A,$F2,$8E,$2F,$16,$C6,$A3
	.dc.b	$FD,$5A,$D7,$47,$81,$C1,$23,$89,$23,$4D,$7C,$8D,$7C,$8D,$7C,$87
	.dc.b	$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18,$21,$F0,$1F,$23
	.dc.b	$4E,$E5,$3A,$14,$E6,$53,$99,$F2,$0D,$94,$D9,$4D,$C7,$CB,$7E,$EF
	.dc.b	$E0,$AF,$17,$EB,$B6,$33,$18,$23,$82,$1F,$89,$8E,$0F,$E2,$F9,$7F
	.dc.b	$D1,$0F,$F9,$F4,$3F,$8A,$E7,$FF,$3E,$61,$FE,$7E,$E8,$FF,$48,$56
	.dc.b	$EB,$51,$FE,$A8,$8F,$F4,$21,$FE,$8E,$E9,$FF,$46,$5F,$FA,$F6,$78
	.dc.b	$B6,$7F,$D4,$C7,$FA,$A6,$32,$7F,$A4,$6F,$C1,$1A,$BF,$F4,$9D,$71
	.dc.b	$8F,$F5,$59,$E8,$9F,$F5,$5B,$E1,$EA,$FF,$51,$B4,$FF,$51,$9B,$3F
	.dc.b	$D4,$71,$D9,$E7,$5E,$C7,$FD,$5E,$78,$63,$51,$F2,$2B,$C4,$DB,$F4
	.dc.b	$57,$79,$FF,$52,$61,$FE,$BA,$AD,$4F,$33,$55,$F0,$1F,$F5,$6D,$7D
	.dc.b	$07,$26,$CF,$4C,$F2,$7D,$47,$D0,$33,$5D,$CF,$A3,$5E,$48,$CE,$1F
	.dc.b	$EC,$03,$E2,$B8,$3E,$E6,$1F,$29,$C6,$1E,$EE,$87,$4F,$FA,$DB,$2F
	.dc.b	$F5,$FF,$81,$FE,$8B,$6D,$FF,$AE,$27,$F8,$AC,$7F,$B0,$23,$55,$FE
	.dc.b	$C0,$6C,$EA,$3A,$C3,$AF,$FD,$94,$DB,$DB,$4F,$F5,$6F,$47,$FA,$80
	.dc.b	$FF,$4D,$38,$E8,$FF,$4F,$74,$FF,$AA,$23,$FD,$4D,$18,$E4,$FF,$52
	.dc.b	$57,$09,$FF,$57,$46,$27,$8F,$5F,$FA,$F2,$F9,$5F,$FA,$D4,$E0,$A3
	.dc.b	$FD,$79,$1E,$39,$FF,$67,$9D,$EF,$FD,$9F,$1C,$63,$31,$FE,$DA,$8E
	.dc.b	$73,$CA,$78,$4F,$FA,$EE,$3C,$57,$E5,$8F,$F6,$F9,$FE,$C5,$8F,$F6
	.dc.b	$A3,$DA,$7D,$3D,$7F,$ED,$53,$FD,$B5,$19,$07,$17,$2F,$76,$DF,$91
	.dc.b	$FF,$67,$BE,$98,$FF,$5D,$F5,$3E,$67,$FD,$88,$48,$FF,$61,$35,$FF
	.dc.b	$5F,$C6,$DC,$90,$FF,$5E,$FC,$CF,$FA,$E4,$8F,$FA,$C9,$AE,$8C,$10
	.dc.b	$C1,$0F,$F6,$2B,$C1,$A6,$CA,$6C,$FF,$67,$BB,$69,$B2,$9B,$3F,$D9
	.dc.b	$8E,$DA,$6C,$A6,$CF,$F6,$83,$B6,$9B,$29,$BE,$B4,$6D,$B1,$B6,$BB
	.dc.b	$8D,$11,$D3,$5D,$18,$21,$E5,$EA,$7C,$47,$F8,$11,$B1,$FE,$0E,$6B
	.dc.b	$FE,$13,$6B,$FE,$FD,$6B,$F8,$C0,$E1,$FC,$7B,$5C,$C7,$72,$3D,$DA
	.dc.b	$F7,$6B,$DD,$AF,$78,$E8,$48,$CC,$78,$1A,$F8,$0E,$C4,$3B,$14,$EC
	.dc.b	$53,$B3,$CA,$3B,$46,$4E,$C5,$3B,$1D,$08,$74,$29,$D0,$E0,$F4,$72
	.dc.b	$74,$29,$D0,$E4,$43,$91,$A7,$93,$93,$91,$4E,$47,$03,$0F,$07,$04
	.dc.b	$38,$94,$E2,$60,$86,$08,$F3,$72,$D7,$31,$F2,$C8,$FC,$26,$BE,$B6
	.dc.b	$BE,$66,$BE,$38,$E8,$48,$CC,$78,$1A,$F8,$0E,$C4,$3B,$14,$EC,$53
	.dc.b	$B3,$CA,$3B,$46,$4E,$C5,$3B,$1D,$08,$74,$29,$D0,$E0,$F4,$72,$74
	.dc.b	$29,$D0,$E4,$43,$91,$A7,$93,$93,$91,$4E,$47,$03,$0F,$07,$04,$38
	.dc.b	$94,$E2,$60,$86,$08,$F3,$72,$D7,$31,$F2,$C8,$FC,$26,$BE,$B6,$BE
	.dc.b	$66,$BE,$38,$E8,$48,$CC,$78,$1A,$F8,$0E,$C4,$3B,$14,$EC,$53,$B3
	.dc.b	$CA,$3B,$46,$4E,$C5,$3F,$D6,$9D,$1F,$EA,$C2,$3F,$EA,$96,$BF,$EB
	.dc.b	$76,$BF,$EB,$76,$BF,$EC,$06,$BF,$EB,$46,$BD,$8E,$84,$3A,$14,$E8
	.dc.b	$53,$A1,$4E,$85,$3A,$1C,$88,$72,$29,$C8,$A7,$22,$9C,$8E,$04,$38
	.dc.b	$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$4E,$B4,$7C,$86,$C7,$BD
	.dc.b	$AF,$A3,$A9,$EF,$7F,$AC,$09,$1F,$EB,$46,$BF,$EA,$D6,$BF,$D8,$44
	.dc.b	$72,$7F,$CC,$75,$FE,$75,$EA,$FF,$38,$F4,$FF,$A1,$9B,$D7,$FE,$83
	.dc.b	$AF,$F4,$9F,$5F,$F9,$CE,$B5,$1F,$E8,$28,$C4,$EF,$AF,$FD,$08,$7F
	.dc.b	$A6,$67,$FD,$58,$7F,$A2,$6B,$89,$27,$FD,$08,$D8,$D4,$73,$C3,$CD
	.dc.b	$1D,$CF,$F5,$39,$FE,$94,$3C,$A6,$0F,$F4,$3C,$74,$39,$61,$FE,$8F
	.dc.b	$8E,$47,$FA,$5E,$BC,$0F,$23,$13,$C1,$FF,$49,$9D,$63,$77,$EE,$C3
	.dc.b	$E0,$CF,$F9,$6A,$3E,$0E,$1E,$8A,$DD,$6A,$BF,$D7,$B3,$9B,$F9,$13
	.dc.b	$9B,$E3,$87,$9F,$0E,$38,$78,$6F,$C9,$1F,$EC,$0C,$39,$4F,$FA,$7C
	.dc.b	$FF,$4A,$1F,$EE,$AE,$AF,$F4,$DF,$57,$FA,$6A,$3F,$09,$8F,$FA,$8D
	.dc.b	$F4,$4F,$FA,$8D,$BD,$7F,$EA,$33,$FD,$5D,$1C,$BA,$FF,$D4,$36,$69
	.dc.b	$FE,$8A,$A7,$5F,$FA,$5B,$A7,$FD,$03,$0E,$8F,$F3,$D3,$63,$FD,$1B
	.dc.b	$18,$3F,$CF,$91,$FE,$83,$34,$7F,$A3,$A7,$FD,$3D,$D3,$FE,$82,$AE
	.dc.b	$B5,$C5,$E6,$78,$E7,$C5,$5F,$E8,$E7,$8B,$D0,$FF,$57,$9F,$E9,$33
	.dc.b	$C1,$3F,$EA,$97,$FD,$74,$7F,$A9,$67,$A7,$5F,$FA,$86,$38,$9F,$E9
	.dc.b	$73,$A5,$78,$1E,$73,$E4,$79,$9F,$1A,$79,$4F,$92,$BC,$13,$C0,$F8
	.dc.b	$1D,$7F,$EB,$37,$C4,$72,$9E,$C7,$1B,$F8,$67,$78,$E4,$79,$AB,$E3
	.dc.b	$3F,$0E,$BF,$D6,$EF,$A6,$3C,$3D,$2F,$57,$F8,$54,$FF,$9A,$E3,$0E
	.dc.b	$03,$87,$6D,$7F,$89,$35,$E5,$1C,$5E,$2D,$8D,$47,$FB,$11,$7A,$B9
	.dc.b	$24,$78,$63,$2D,$8C,$BE,$26,$BE,$26,$BE,$B6,$BD,$DA,$FB,$1B,$1D
	.dc.b	$A3,$A1,$DC,$86,$08,$77,$32,$53,$27,$84,$C9,$4C,$9E,$33,$25,$32
	.dc.b	$79,$4C,$94,$CB,$E7,$7C,$04,$3B,$14,$E8,$53,$91,$5E,$4F,$02,$1C
	.dc.b	$0A,$70,$2B,$C1,$D1,$FE,$B0,$F5,$7F,$D7,$84,$7F,$D7,$8D,$7F,$D7
	.dc.b	$8D,$7F,$D7,$8D,$7F,$D5,$CD,$7A,$1C,$88,$72,$29,$C8,$A7,$22,$9C
	.dc.b	$8E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$43,$FD,$AE
	.dc.b	$FB,$5A,$79,$8A,$78,$4A,$76,$29,$D4,$A7,$5E,$A4,$75,$6C,$75,$6B
	.dc.b	$D5,$AF,$56,$BD,$5A,$F5,$8E,$84,$7A,$35,$E8,$D7,$A3,$5E,$8D,$7A
	.dc.b	$1C,$88,$72,$29,$C8,$A7,$22,$9C,$8E,$04,$38,$14,$E0,$53,$81,$A2
	.dc.b	$1A,$29,$A3,$04,$30,$43,$E9,$92,$3E,$4B,$63,$E0,$34,$F4,$94,$F2
	.dc.b	$14,$F0,$94,$F0,$14,$F0,$47,$62,$47,$66,$C7,$66,$BD,$9A,$F6,$6B
	.dc.b	$D9,$AF,$68,$E8,$47,$A3,$5E,$8D,$7A,$35,$E8,$D7,$A1,$C8,$87,$22
	.dc.b	$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$3F,$03,$D4,$F5,$7F,$B1,$36,$3F,$B6,$35,$FE,$E2,$D7,$FC
	.dc.b	$F8,$D4,$47,$0D,$8E,$A4,$7A,$B5,$EA,$D7,$AB,$5E,$A2,$3A,$B6,$34
	.dc.b	$F7,$3A,$90,$EA,$53,$A9,$4E,$B1,$D0,$91,$D2,$3B,$1B,$3B,$1D,$08
	.dc.b	$74,$29,$D1,$E5,$1D,$23,$91,$D0,$D9,$D0,$E4,$43,$91,$C1,$E4,$F0
	.dc.b	$39,$1B,$39,$1C,$0D,$7E,$0B,$FE,$BC,$7F,$D6,$8D,$7E,$6B,$5F,$B6
	.dc.b	$D4,$D2,$73,$7F,$8A,$5D,$73,$7F,$A4,$E7,$FD,$0C,$7F,$A4,$E5,$D3
	.dc.b	$FE,$BC,$8C,$35,$FF,$49,$BF,$E8,$FE,$9F,$F4,$9C,$72,$32,$7F,$A8
	.dc.b	$8F,$F5,$29,$C8,$FF,$55,$F4,$A3,$F0,$F7,$FE,$C1,$3F,$D3,$73,$C6
	.dc.b	$74,$FF,$A9,$FA,$BF,$1B,$1F,$EA,$CC,$3F,$D3,$73,$FE,$B3,$8D,$BF
	.dc.b	$EC,$A3,$17,$DA,$B8,$BF,$EA,$43,$FD,$81,$1E,$59,$FF,$54,$BE,$83
	.dc.b	$D6,$FF,$AA,$70,$F8,$11,$FE,$D0,$38,$4F,$C2,$78,$61,$EA,$AE,$11
	.dc.b	$EB,$9F,$13,$FE,$D3,$9F,$A4,$F0,$C3,$E8,$BF,$1A,$76,$FF,$B5,$8F
	.dc.b	$84,$7C,$F9,$ED,$1E,$A8,$F3,$B6,$33,$1C,$0E,$D7,$F6,$E3,$30,$7E
	.dc.b	$96,$1F,$1A,$3D,$A7,$BC,$D5,$F9,$B0,$F9,$F4,$E9,$FF,$60,$57,$FB
	.dc.b	$37,$A7,$FD,$94,$FD,$E8,$FF,$64,$3F,$EC,$FE,$AF,$F6,$43,$FE,$88
	.dc.b	$AE,$51,$FE,$CA,$24,$7F,$B5,$A3,$2F,$FB,$63,$9F,$FD,$34,$FF,$A5
	.dc.b	$7A,$7F,$8E,$F3,$7F,$A6,$23,$FD,$50,$7F,$AD,$E7,$6F,$FA,$77,$A3
	.dc.b	$FD,$6A,$E7,$A3,$FD,$3B,$5C,$6B,$64,$E9,$FF,$58,$CB,$0E,$F8,$7F
	.dc.b	$AA,$6B,$73,$CA,$3A,$E1,$FE,$A7,$75,$1F,$EB,$58,$F2,$99,$3F,$D6
	.dc.b	$E7,$FA,$CA,$7C,$0F,$96,$BB,$BF,$EB,$E7,$FD,$B8,$FC,$13,$FD,$97
	.dc.b	$3E,$57,$FD,$9B,$7E,$89,$E3,$1E,$0C,$3D,$D5,$E7,$8F,$F6,$35,$6D
	.dc.b	$E2,$70,$E7,$7A,$BF,$EB,$12,$47,$FA,$B5,$AF,$FA,$F1,$AF,$FA,$F1
	.dc.b	$FF,$5C,$7A,$B2,$3F,$D6,$8D,$7F,$D8,$6D,$7F,$D8,$6D,$7F,$D9,$8D
	.dc.b	$7F,$D8,$0D,$7F,$D9,$AD,$93,$49,$1E,$02,$1E,$02,$9E,$02,$9E,$02
	.dc.b	$9E,$02,$9E,$02,$C7,$81,$B1,$E0,$7B,$10,$EC,$53,$B1,$4E,$C5,$3B
	.dc.b	$16,$3B,$36,$3B,$3D,$08,$74,$29,$D0,$FE,$EE,$3A,$5D,$9F,$EE,$F6
	.dc.b	$C7,$FB,$E5,$AF,$FB,$F9,$AF,$26,$BF,$E0,$26,$BF,$E1,$06,$BD,$E3
	.dc.b	$B1,$1E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$CD,$7B,$1D,$08,$74,$29,$D0
	.dc.b	$A7,$42,$9D,$0A,$74,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70
	.dc.b	$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$91,$F5,$09,$1F,$25,$B1
	.dc.b	$F0,$1A,$FA,$5A,$F9,$1A,$F8,$5A,$F8,$1A,$F8,$23,$B1,$1E,$CD,$7B
	.dc.b	$35,$EC,$D7,$B3,$5E,$CD,$7B,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A
	.dc.b	$74,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70,$29,$C0,$A7,$03
	.dc.b	$44,$34,$53,$46,$08,$60,$91,$F5,$09,$1F,$25,$B1,$F0,$1A,$FA,$5A
	.dc.b	$F9,$1A,$F8,$5A,$F8,$1A,$F8,$23,$B1,$1E,$CD,$7B,$35,$EC,$D7,$B3
	.dc.b	$5E,$CD,$7B,$1D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$74,$39,$10,$FA
	.dc.b	$BD,$0F,$57,$FD,$6E,$48,$FF,$5B,$B5,$FF,$5B,$B5,$FF,$5B,$B5,$FF
	.dc.b	$4F,$B5,$FF,$5C,$35,$FF,$57,$36,$3C,$11,$D8,$87,$62,$9D,$8A,$76
	.dc.b	$29,$D8,$A7,$62,$BD,$9E,$84,$3A,$14,$E8,$53,$A1,$4E,$85,$7A,$3C
	.dc.b	$88,$72,$29,$C8,$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83,$FE,$AC
	.dc.b	$E2,$FF,$54,$11,$FE,$CE,$64,$A7,$FA,$C1,$D3,$AE,$9F,$F3,$23,$A6
	.dc.b	$BF,$E5,$FE,$6F,$F4,$CF,$1F,$FA,$0B,$9B,$FC,$D7,$CF,$FE,$71,$AF
	.dc.b	$F3,$D9,$FE,$79,$8C,$BF,$E8,$9A,$E2,$4E,$8F,$F4,$AC,$7F,$A8,$0E
	.dc.b	$1D,$3F,$E9,$F6,$90,$EA,$7F,$A0,$6F,$C1,$87,$FA,$CE,$FF,$CF,$D5
	.dc.b	$FE,$9D,$C3,$C3,$7F,$E9,$33,$81,$FE,$93,$9D,$BF,$E8,$7A,$FF,$46
	.dc.b	$9F,$E8,$C7,$D1,$87,$9A,$3E,$05,$FF,$A7,$AF,$FD,$79,$39,$BF,$65
	.dc.b	$FF,$AA,$4F,$45,$F4,$C3,$CE,$7A,$27,$C3,$3F,$04,$FF,$61,$4F,$67
	.dc.b	$8C,$6A,$31,$1A,$8E,$F7,$FE,$B8,$26,$1F,$EA,$B8,$E1,$97,$FA,$7A
	.dc.b	$BF,$D3,$47,$FA,$55,$C5,$FF,$A5,$BA,$7F,$C4,$EF,$FA,$62,$7F,$C5
	.dc.b	$ED,$7F,$D4,$9D,$3F,$EA,$27,$35,$C8,$91,$FE,$A2,$AF,$F1,$8F,$47
	.dc.b	$FA,$3B,$A3,$FC,$F6,$FF,$A1,$AF,$FD,$14,$7F,$9F,$7A,$7F,$D2,$3D
	.dc.b	$3F,$E7,$EA,$DF,$4F,$FA,$16,$13,$FE,$94,$9F,$F4,$5D,$7F,$A4,$7A
	.dc.b	$7F,$D0,$55,$A8,$FF,$4A,$E1,$E1,$7F,$D1,$F5,$E2,$AD,$DF,$39,$C9
	.dc.b	$2F,$93,$E5,$3F,$D1,$E7,$FA,$74,$D4,$71,$8F,$44,$7F,$A4,$8F,$6D
	.dc.b	$AB,$BC,$79,$30,$FF,$56,$4E,$63,$A5,$7F,$AE,$EB,$71,$E1,$1D,$0F
	.dc.b	$4F,$F7,$71,$86,$C6,$63,$F7,$ED,$7F,$81,$B5,$FF,$35,$B5,$11,$C3
	.dc.b	$63,$AF,$A7,$23,$AB,$5E,$AD,$7A,$B5,$EA,$7A,$9C,$3D,$DA,$E4,$EA
	.dc.b	$43,$A9,$4E,$A5,$3A,$CA,$39,$B6,$34,$77,$29,$DC,$EA,$43,$A9,$4E
	.dc.b	$B1,$D0,$8F,$28,$EC,$EC,$EC,$53,$B1,$D0,$87,$43,$94,$74,$78,$47
	.dc.b	$47,$67,$42,$9D,$0E,$47,$7E,$97,$8D,$FE,$A5,$24,$7F,$A9,$5A,$FF
	.dc.b	$A9,$5A,$FF,$A9,$5A,$FF,$A9,$5A,$FF,$A9,$5A,$FF,$AC,$DB,$1E,$08
	.dc.b	$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$53,$B1,$5E,$CF,$42,$1D,$0A,$74
	.dc.b	$29,$D0,$A7,$42,$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C
	.dc.b	$0A,$70,$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$24,$7D,$42,$47,$C9,$6B
	.dc.b	$F0,$1A,$FA,$5A,$F9,$1A,$F8,$5A,$F8,$1B,$1E,$08,$EC,$43,$B1,$4E
	.dc.b	$C5,$3B,$14,$EC,$53,$B1,$5E,$CF,$42,$1D,$0A,$74,$29,$D0,$A7,$42
	.dc.b	$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1
	.dc.b	$D1,$0D,$15,$D3,$82,$38,$24,$7D,$42,$47,$C9,$6B,$F0,$1A,$FA,$5A
	.dc.b	$F9,$1A,$F8,$5A,$F8,$1B,$1E,$08,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC
	.dc.b	$53,$B1,$5E,$CF,$42,$1D,$0A,$74,$3E,$FF,$43,$D3,$FE,$D2,$D8,$FF
	.dc.b	$32,$B5,$FF,$34,$B5,$FE,$B2,$1C,$3F,$D7,$1A,$FF,$6A,$6B,$B8,$EE
	.dc.b	$47,$BB,$5E,$ED,$7B,$C7,$22,$46,$63,$C0,$D7,$C0,$D7,$C0,$76,$21
	.dc.b	$D8,$A7,$67,$84,$76,$8C,$9D,$8A,$76,$29,$D8,$E8,$43,$A1,$A7,$A3
	.dc.b	$93,$A1,$4E,$85,$3A,$1C,$8C,$3C,$9C,$10,$E6,$7D,$6E,$97,$CF,$7F
	.dc.b	$89,$9B,$1F,$E9,$D8,$C1,$1C,$10,$FF,$52,$3F,$E2,$EE,$5F,$F3,$89
	.dc.b	$FE,$8F,$EA,$FF,$13,$F4,$BA,$7F,$D2,$07,$FA,$14,$FF,$4D,$5F,$FA
	.dc.b	$40,$47,$FA,$50,$FF,$44,$BF,$E8,$73,$FC,$F9,$3F,$E8,$B2,$39,$AF
	.dc.b	$03,$CA,$BB,$BF,$E8,$69,$F1,$C6,$23,$99,$C7,$AB,$FD,$21,$3A,$8F
	.dc.b	$F5,$07,$57,$FA,$34,$FF,$45,$D7,$FA,$26,$75,$3F,$E9,$5B,$F0,$E1
	.dc.b	$EF,$8E,$33,$A9,$F8,$B8,$76,$9F,$31,$A2,$9C,$DD,$57,$38,$C1,$FE
	.dc.b	$B0,$3D,$05,$7F,$D3,$AF,$FA,$B2,$38,$DF,$BE,$BC,$AD,$8C,$CF,$86
	.dc.b	$BB,$57,$9A,$3A,$57,$BD,$F2,$C7,$FA,$8F,$B3,$FD,$3F,$1F,$EA,$7E
	.dc.b	$AF,$F4,$E4,$62,$3F,$D4,$31,$FE,$A3,$70,$70,$EA,$FF,$55,$F5,$FF
	.dc.b	$A9,$23,$90,$8F,$F5,$2F,$4F,$FA,$0B,$9F,$FC,$FF,$D3,$FE,$83,$AF
	.dc.b	$F4,$4C,$60,$8F,$FA,$1B,$AB,$FC,$D7,$3F,$EA,$BE,$AF,$F4,$89,$FE
	.dc.b	$89,$9D,$1F,$E7,$C8,$D4,$FF,$AB,$89,$1F,$E9,$C3,$B9,$C0,$47,$07
	.dc.b	$53,$DA,$7C,$37,$FE,$B5,$7F,$D4,$46,$3A,$FF,$D7,$27,$FA,$BE,$BA
	.dc.b	$5F,$37,$AD,$7C,$17,$FD,$23,$7F,$EB,$9F,$29,$E4,$3E,$47,$F9,$8D
	.dc.b	$7F,$A5,$A7,$31,$88,$46,$1D,$B5,$FE,$72,$D7,$FD,$26,$D7,$A4,$F7
	.dc.b	$78,$B5,$FF,$54,$FC,$C7,$FA,$B8,$8F,$FA,$B9,$AE,$8C,$10,$C1,$0F
	.dc.b	$F5,$8B,$C1,$A6,$CA,$6C,$FF,$4E,$BB,$69,$B2,$9B,$3F,$D7,$4E,$DA
	.dc.b	$6C,$A6,$CF,$F6,$6B,$B6,$9B,$29,$BE,$D4,$6D,$B1,$B6,$BB,$8D,$11
	.dc.b	$D3,$5D,$18,$21,$E5,$EC,$7B,$BF,$EE,$06,$C7,$E0,$5A,$FE,$19,$AF
	.dc.b	$F7,$86,$BF,$ED,$26,$BF,$EE,$06,$A6,$92,$3C,$04,$8F,$03,$5F,$03
	.dc.b	$5F,$03,$5F,$03,$5F,$03,$5F,$03,$63,$C1,$1D,$88,$76,$29,$D8,$A7
	.dc.b	$62,$9D,$8A,$76,$2B,$D9,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3
	.dc.b	$C8,$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21
	.dc.b	$A2,$BA,$70,$47,$04,$8F,$A8,$48,$F9,$2D,$7E,$03,$5F,$4B,$5F,$23
	.dc.b	$5F,$0B,$5F,$03,$63,$C1,$1D,$88,$76,$29,$D8,$A7,$62,$9D,$8A,$76
	.dc.b	$2B,$D9,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8,$87,$22,$9C
	.dc.b	$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA,$70,$47
	.dc.b	$04,$8F,$A8,$48,$F9,$2D,$7E,$03,$5F,$4B,$5F,$23,$5F,$0B,$5F,$03
	.dc.b	$63,$C1,$1D,$88,$76,$29,$D8,$A7,$62,$9D,$8A,$3C,$F7,$B9,$FD,$A6
	.dc.b	$30,$D8,$CC,$7F,$6D,$6B,$FD,$C9,$AF,$FA,$91,$A8,$8E,$1B,$1D,$48
	.dc.b	$F5,$6B,$D5,$AF,$56,$BD,$4F,$39,$C4,$87,$12,$9C,$4A,$71,$8F,$23
	.dc.b	$C8,$E0,$43,$81,$4E,$11,$E2,$7D,$67,$42,$9B,$29,$C8,$87,$23,$81
	.dc.b	$E1,$32,$70,$34,$73,$34,$60,$D9,$83,$81,$0C,$12,$3D,$E4,$8F,$C0
	.dc.b	$7C,$BF,$F9,$C6,$7F,$DB,$AD,$7E,$73,$5F,$B8,$D4,$D2,$74,$74,$3F
	.dc.b	$D5,$1D,$9F,$E8,$C9,$FF,$3C,$1F,$E7,$02,$9F,$E7,$07,$FD,$25,$D9
	.dc.b	$FE,$A0,$EB,$FF,$3F,$CF,$FA,$A5,$FF,$54,$57,$FA,$B3,$AD,$D9,$F8
	.dc.b	$39,$CD,$7E,$46,$58,$74,$8F,$F4,$59,$DF,$0F,$F4,$C4,$79,$63,$B6
	.dc.b	$1F,$EB,$97,$94,$71,$BF,$1D,$F8,$27,$FD,$41,$87,$FA,$39,$FF,$4C
	.dc.b	$9D,$49,$3C,$23,$CD,$7D,$65,$5E,$6C,$3C,$67,$C1,$9F,$F4,$A5,$FF
	.dc.b	$A6,$30,$F3,$57,$86,$FF,$D2,$F8,$78,$63,$FD,$39,$1E,$0B,$FF,$51
	.dc.b	$3F,$3C,$F6,$9F,$EC,$73,$47,$D2,$BF,$F6,$99,$E7,$7D,$A7,$19,$F0
	.dc.b	$47,$CE,$8F,$4D,$FC,$5A,$F3,$9E,$08,$EA,$7B,$BA,$BF,$D4,$5D,$5F
	.dc.b	$EA,$4E,$BF,$F5,$5C,$FD,$79,$47,$FA,$AE,$30,$D7,$93,$C1,$A7,$FA
	.dc.b	$B2,$7F,$D5,$7D,$3F,$E9,$08,$73,$FF,$A8,$DF,$F4,$E9,$1F,$F4,$D3
	.dc.b	$FE,$A5,$E8,$FF,$4D,$3F,$EA,$7E,$BF,$F5,$7B,$BE,$BF,$F5,$3B,$FE
	.dc.b	$98,$BE,$D7,$D4,$E7,$19,$3C,$33,$FE,$A4,$7B,$4F,$08,$C4,$6E,$79
	.dc.b	$47,$6E,$B6,$1F,$EA,$5A,$F1,$1F,$EB,$29,$D4,$66,$BC,$35,$FE,$AD
	.dc.b	$38,$DF,$7B,$F0,$9E,$43,$DC,$7F,$AD,$2B,$E0,$BF,$EC,$0C,$3F,$97
	.dc.b	$5F,$49,$CD,$FF,$B5,$63,$FD,$6D,$D2,$F5,$BF,$D6,$5D,$7F,$EB,$86
	.dc.b	$C6,$71,$FF,$55,$C6,$08,$E0,$87,$FA,$FB,$AD,$F2,$FF,$D5,$5B,$1F
	.dc.b	$EB,$F6,$BB,$8D,$11,$D3,$5D,$18,$21,$82,$1F,$D3,$44,$76,$70,$D7
	.dc.b	$FA,$93,$5F,$EC,$0D,$78,$C7,$85,$E2,$D7,$8B,$5E,$2D,$78,$9E,$63
	.dc.b	$89,$4E,$25,$38,$94,$E2,$7B,$0D,$94,$D9,$4D,$9D,$8C,$47,$AC,$91
	.dc.b	$E5,$6C,$77,$6B,$C2,$34,$7A,$FA,$9E,$FF,$FA,$E4,$93,$FE,$B9,$6B
	.dc.b	$FE,$B9,$6B,$FE,$B9,$6B,$FE,$B9,$6B,$FE,$B9,$6B,$FE,$B9,$6C,$78
	.dc.b	$23,$B1,$0E,$C5,$3B,$14,$EC,$53,$B1,$4E,$C5,$7B,$3D,$08,$74,$29
	.dc.b	$D0,$A7,$42,$9D,$0A,$F4,$79,$10,$E4,$53,$91,$4E,$45,$79,$3C,$08
	.dc.b	$70,$29,$C0,$AF,$07,$44,$34,$57,$4E,$08,$E0,$91,$F5,$09,$1F,$25
	.dc.b	$AF,$C0,$6B,$E9,$6B,$E4,$6B,$E1,$6B,$E0,$6C,$78,$23,$B1,$0E,$C5
	.dc.b	$3B,$14,$EC,$53,$B1,$4E,$C5,$7B,$3D,$08,$74,$29,$D0,$A7,$42,$9D
	.dc.b	$0A,$F4,$79,$10,$E4,$53,$91,$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF
	.dc.b	$07,$44,$34,$57,$4E,$08,$E0,$91,$F5,$09,$1F,$25,$AF,$C0,$6B,$E9
	.dc.b	$6B,$E4,$6B,$E1,$6B,$E0,$6C,$78,$23,$B1,$0E,$C5,$3B,$14,$EC,$53
	.dc.b	$B1,$4E,$C5,$7B,$3D,$08,$74,$29,$D0,$FB,$FD,$7F,$E6,$B2,$72,$7F
	.dc.b	$AC,$49,$1F,$EB,$16,$BF,$EB,$16,$BF,$EB,$16,$BF,$E9,$D6,$BF,$EB
	.dc.b	$26,$BF,$EC,$16,$C7,$82,$3B,$10,$EC,$53,$B1,$4E,$C5,$3B,$14,$EC
	.dc.b	$57,$B3,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$AF,$47,$91,$0E,$45,$39
	.dc.b	$14,$E4,$57,$93,$C0,$FF,$50,$FB,$DF,$EA,$62,$3F,$E0,$53,$FD,$46
	.dc.b	$E1,$C7,$5A,$31,$1F,$EA,$C7,$FD,$47,$D7,$FE,$7D,$8F,$F3,$85,$FF
	.dc.b	$A6,$FA,$FF,$D0,$B1,$FE,$75,$EA,$FF,$40,$56,$E3,$FC,$F2,$4E,$BF
	.dc.b	$F4,$43,$FE,$6E,$7F,$CE,$B7,$FE,$8D,$73,$5C,$A7,$FD,$42,$FF,$9F
	.dc.b	$A7,$FC,$EF,$1F,$E9,$EA,$ED,$5C,$A3,$C5,$3F,$E7,$BA,$E1,$1E,$83
	.dc.b	$81,$E1,$BF,$4B,$E5,$EB,$FF,$42,$3B,$3F,$D0,$F8,$7F,$A7,$9F,$1D
	.dc.b	$FC,$17,$FD,$1E,$7A,$E7,$FD,$3F,$5D,$30,$F2,$4F,$23,$C0,$70,$C3
	.dc.b	$CA,$FC,$59,$FF,$52,$57,$28,$D5,$F7,$7F,$D5,$27,$FA,$76,$7B,$1C
	.dc.b	$30,$F8,$6E,$23,$47,$23,$E0,$DF,$CD,$3F,$D4,$98,$79,$8F,$A3,$69
	.dc.b	$FF,$4D,$9F,$E9,$7F,$03,$0F,$FA,$6B,$AB,$FD,$48,$9A,$FF,$94,$E9
	.dc.b	$D9,$FE,$A6,$8F,$C9,$46,$3B,$3F,$D4,$CD,$8E,$C7,$82,$BA,$9F,$EA
	.dc.b	$67,$FD,$04,$7F,$A4,$7B,$3C,$7D,$3F,$E8,$9E,$AF,$F4,$64,$7F,$A3
	.dc.b	$BB,$3F,$D1,$55,$B3,$FD,$12,$7F,$A5,$0C,$12,$B9,$47,$06,$C7,$81
	.dc.b	$FF,$4A,$B6,$BF,$D1,$BD,$9F,$EB,$0E,$CF,$F4,$85,$FC,$95,$87,$FA
	.dc.b	$A3,$0E,$31,$FE,$A8,$C3,$FD,$48,$43,$C5,$87,$FA,$60,$EA,$FF,$A7
	.dc.b	$EF,$A5,$F3,$9D,$46,$27,$DB,$5E,$27,$D1,$87,$96,$FE,$24,$7A,$6B
	.dc.b	$DD,$39,$8F,$16,$1E,$6A,$F9,$B8,$76,$9E,$67,$73,$FD,$1F,$ED,$FF
	.dc.b	$A7,$C8,$FF,$A7,$DA,$FF,$A9,$5A,$FF,$A5,$1F,$E2,$8F,$E3,$F9,$DE
	.dc.b	$DE,$21,$18,$76,$D7,$F8,$E3,$5F,$F5,$13,$5F,$CB,$B5,$ED,$1D,$63
	.dc.b	$11,$DA,$30,$47,$BB,$5E,$ED,$7B,$B5,$EE,$D7,$B9,$C8,$87,$22,$9C
	.dc.b	$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$3E,$31,$E4,$29,$DC,$A7,$42,$9C,$CA,$73,$8F,$A2,$7F,$BB,$DF
	.dc.b	$B7,$D9,$F5,$48,$FE,$01,$AE,$8C,$10,$C1,$0F,$C1,$BC,$1A,$6C,$A6
	.dc.b	$CF,$C3,$BB,$69,$B2,$9B,$3F,$30,$ED,$A6,$CA,$6C,$FC,$DB,$B6,$9B
	.dc.b	$29,$B8,$F8,$8F,$98,$87,$88,$A7,$62,$9C,$8A,$F2,$78,$10,$E0,$53
	.dc.b	$81,$5E,$0E,$88,$68,$AE,$9C,$11,$C1,$23,$E0,$36,$3E,$03,$5F,$81
	.dc.b	$1A,$23,$A6,$BA,$30,$43,$04,$3E,$0B,$C1,$A6,$CA,$6C,$F8,$0E,$DA
	.dc.b	$6C,$A6,$CF,$82,$ED,$A6,$CA,$6C,$F8,$8E,$DA,$6C,$A6,$E3,$E2,$3E
	.dc.b	$62,$1E,$22,$9D,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21
	.dc.b	$A2,$BA,$70,$47,$04,$8F,$80,$D8,$F8,$0D,$7E,$04,$68,$8E,$9A,$E8
	.dc.b	$C1,$0C,$10,$F8,$2F,$06,$9B,$29,$B3,$E0,$3B,$69,$B2,$9B,$3E,$0B
	.dc.b	$B6,$9B,$29,$B3,$E2,$3B,$69,$B3,$F5,$3D,$2F,$6B,$FB,$13,$63,$FB
	.dc.b	$3B,$5F,$F3,$E3,$5F,$F1,$A8,$70,$FF,$8E,$1A,$FF,$8F,$5A,$EE,$3B
	.dc.b	$91,$EE,$D7,$BB,$5E,$F1,$C8,$91,$98,$F0,$35,$F0,$35,$F0,$1D,$88
	.dc.b	$76,$29,$D9,$E1,$1D,$A3,$27,$62,$9D,$8A,$76,$3A,$10,$E8,$69,$E8
	.dc.b	$E4,$E8,$53,$A1,$4E,$87,$23,$0F,$27,$04,$39,$9F,$5B,$F1,$77,$8F
	.dc.b	$FE,$1F,$6C,$7F,$89,$1A,$FF,$AE,$5A,$F1,$8E,$04,$78,$3F,$EB,$87
	.dc.b	$FD,$2E,$7F,$A0,$39,$BF,$D4,$87,$FA,$2F,$9F,$FD,$03,$D5,$FE,$A0
	.dc.b	$9F,$F4,$91,$FE,$88,$3F,$D1,$E7,$FA,$77,$AB,$FD,$1A,$FF,$A2,$CF
	.dc.b	$F4,$E4,$FF,$A2,$EB,$FD,$6B,$18,$6F,$37,$FA,$32,$3F,$D5,$8D,$74
	.dc.b	$62,$FC,$51,$FE,$93,$EA,$FF,$4E,$4F,$A8,$D9,$23,$4F,$9E,$FD,$78
	.dc.b	$7F,$AD,$DF,$F4,$ED,$72,$6C,$EF,$0F,$F6,$86,$1E,$0A,$FF,$54,$4F
	.dc.b	$BC,$F0,$61,$E7,$24,$6A,$FF,$D7,$31,$E2,$9E,$11,$D9,$B7,$F0,$AF
	.dc.b	$FD,$79,$1F,$32,$78,$DF,$3A,$FA,$87,$93,$0F,$F5,$94,$A3,$FD,$63
	.dc.b	$87,$FA,$9F,$9F,$FD,$4B,$D5,$8E,$AF,$F5,$74,$63,$2F,$F5,$C9,$FE
	.dc.b	$B3,$8F,$F5,$93,$7A,$BF,$D6,$31,$C2,$B9,$47,$FB,$06,$B9,$57,$FA
	.dc.b	$FE,$7F,$D1,$3D,$5F,$E9,$C8,$FF,$63,$74,$7F,$A1,$DB,$CF,$FE,$98
	.dc.b	$7F,$D5,$D1,$97,$FD,$3B,$D5,$FE,$89,$87,$57,$FA,$46,$72,$7F,$A5
	.dc.b	$4F,$F4,$79,$CA,$BF,$D7,$AF,$FA,$59,$AF,$0B,$FF,$56,$37,$AB,$FD
	.dc.b	$33,$87,$7B,$ED,$87,$8C,$98,$79,$F0,$FF,$54,$1E,$DC,$13,$E6,$7F
	.dc.b	$D6,$47,$92,$37,$87,$FA,$CF,$0F,$F5,$5C,$EA,$79,$57,$18,$FF,$68
	.dc.b	$77,$FF,$30,$27,$3F,$FA,$7D,$B1,$FE,$A4,$6B,$FC,$F0,$3F,$CE,$3A
	.dc.b	$5E,$DF,$FA,$6C,$93,$FE,$A6,$6B,$FE,$A6,$6B,$FE,$A6,$6B,$FE,$A6
	.dc.b	$6B,$FE,$B3,$6B,$FE,$B2,$6C,$78,$23,$B1,$0E,$C5,$3B,$14,$EC,$53
	.dc.b	$B1,$4E,$C5,$7B,$3D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$F4,$79,$10
	.dc.b	$E4,$53,$91,$4E,$45,$79,$3C,$08,$70,$29,$C0,$AF,$07,$FD,$C1,$ED
	.dc.b	$FF,$B4,$C8,$FF,$B7,$9A,$FF,$B7,$9A,$FF,$B7,$9A,$FF,$B7,$9A,$F4
	.dc.b	$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70,$29,$C0,$A7,$03,$44
	.dc.b	$34,$53,$46,$08,$60,$82,$11,$EE,$8C,$36,$33,$1E,$93,$25,$72,$F9
	.dc.b	$8C,$95,$CB,$E5,$32,$57,$2F,$98,$C9,$5C,$BE,$86,$BA,$31,$1E,$72
	.dc.b	$3E,$56,$BE,$36,$BE,$16,$BD,$DA,$F4,$39,$10,$E4,$53,$91,$4E,$45
	.dc.b	$39,$1C,$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$87,$C7
	.dc.b	$24,$7C,$A2,$47,$C0,$8C,$B6,$33,$1E,$B3,$25,$72,$FA,$0C,$95,$CB
	.dc.b	$E7,$32,$57,$2F,$A0,$C9,$5C,$BE,$A6,$BA,$31,$1E,$72,$3E,$56,$BE
	.dc.b	$36,$BE,$16,$BD,$DA,$F4,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08
	.dc.b	$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$FC,$0F,$53,$DA,$FF
	.dc.b	$50,$92,$3F,$D4,$2D,$7F,$D4,$2D,$7F,$D2,$6D,$7F,$D4,$0D,$7F,$D4
	.dc.b	$0D,$7F,$D4,$CD,$8F,$04,$76,$21,$D8,$A7,$62,$9D,$8A,$76,$29,$D8
	.dc.b	$AF,$67,$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$5E,$8F,$22,$1C,$8A,$72
	.dc.b	$29,$C8,$AF,$27,$81,$0E,$05,$38,$15,$E0,$87,$5B,$7F,$F4,$F1,$23
	.dc.b	$FD,$3C,$D7,$FD,$3C,$D7,$FD,$4C,$FF,$A6,$F9,$FF,$CD,$87,$F9,$A2
	.dc.b	$3F,$D2,$5C,$DF,$E7,$5E,$AF,$F3,$6C,$FF,$A0,$A4,$FF,$A5,$5F,$F4
	.dc.b	$77,$4F,$FA,$17,$A7,$FC,$E8,$FF,$A7,$23,$FC,$E9,$1F,$E7,$27,$D5
	.dc.b	$6E,$AF,$F4,$54,$F0,$9F,$F4,$2E,$1F,$E8,$97,$C6,$D7,$99,$FE,$B1
	.dc.b	$21,$FE,$AE,$3F,$D0,$A7,$90,$F3,$C7,$FA,$D9,$DD,$7F,$A4,$1E,$B7
	.dc.b	$E9,$78,$61,$ED,$7A,$3F,$15,$F0,$9F,$EA,$7B,$F8,$B7,$FE,$8D,$8E
	.dc.b	$E7,$6C,$3F,$D2,$4F,$82,$76,$FC,$28,$F8,$2F,$82,$7C,$CF,$5B,$FA
	.dc.b	$73,$F8,$18,$EA,$7F,$AD,$FA,$BF,$D1,$CF,$CC,$6B,$FE,$8D,$31,$D5
	.dc.b	$FE,$91,$E9,$FF,$49,$C7,$F9,$87,$35,$7F,$A6,$A3,$27,$FA,$6C,$D3
	.dc.b	$D2,$3F,$D4,$D3,$CB,$AB,$FD,$37,$CD,$FE,$80,$E8,$FF,$3B,$E1,$FE
	.dc.b	$8D,$7F,$D2,$9B,$9C,$F4,$7F,$9F,$A7,$FD,$3F,$D5,$FE,$8A,$AF,$F4
	.dc.b	$8D,$F3,$9F,$F4,$C5,$7F,$A0,$2B,$57,$D0,$E7,$D5,$FE,$AA,$3F,$D2
	.dc.b	$B5,$C1,$FF,$45,$34,$EE,$7F,$A2,$DC,$D7,$FA,$7A,$FC,$0D,$C3,$C6
	.dc.b	$4B,$F0,$5F,$33,$FD,$37,$87,$91,$C5,$FF,$A8,$E3,$EC,$5F,$BE,$76
	.dc.b	$78,$4F,$18,$7F,$D8,$54,$F0,$5D,$FF,$E9,$7A,$F9,$51,$87,$FD,$35
	.dc.b	$3B,$9D,$36,$33,$3B,$7F,$14,$72,$9D,$C6,$63,$F1,$D3,$C5,$FF,$53
	.dc.b	$74,$B4,$F1,$47,$F9,$25,$B1,$98,$F3,$35,$E3,$1A,$23,$A6,$BA,$30
	.dc.b	$43,$04,$3D,$31,$D5,$B1,$C9,$A7,$12,$9C,$4F,$D5,$47,$16,$C7,$16
	.dc.b	$9C,$4A,$71,$3F,$99,$47,$16,$C7,$16,$9C,$4A,$71,$3E,$41,$92,$99
	.dc.b	$8E,$8E,$0F,$49,$0F,$19,$4E,$A5,$38,$1A,$21,$A2,$9A,$30,$43,$03
	.dc.b	$B5,$F4,$DF,$80,$6C,$7E,$4D,$AF,$FB,$AD,$AF,$E3,$9A,$FF,$19,$6B
	.dc.b	$FD,$6D,$AF,$78,$EC,$47,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$C7
	.dc.b	$42,$1D,$0A,$74,$29,$D0,$A7,$42,$9D,$0E,$44,$39,$14,$E4,$53,$91
	.dc.b	$4E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18,$24,$7D
	.dc.b	$42,$47,$C9,$6C,$7C,$06,$BE,$96,$BE,$46,$BE,$16,$BE,$06,$BE,$08
	.dc.b	$EC,$47,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$C7,$42,$1D,$0A,$74
	.dc.b	$29,$D0,$A7,$42,$9D,$0E,$44,$39,$14,$E4,$53,$91,$4E,$47,$02,$1C
	.dc.b	$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18,$24,$7D,$42,$47,$C9,$6C
	.dc.b	$7C,$06,$BE,$96,$BE,$46,$BE,$16,$BE,$06,$BE,$08,$EC,$47,$B3,$5E
	.dc.b	$CD,$7B,$35,$EC,$D7,$B3,$5E,$C7,$42,$1D,$0A,$74,$29,$D0,$A7,$42
	.dc.b	$9D,$0E,$44,$3E,$AF,$AE,$ED,$FE,$D2,$D8,$FE,$F8,$D7,$FB,$A8,$70
	.dc.b	$FF,$78,$6B,$FE,$43,$6B,$FE,$85,$6B,$C6,$3B,$91,$EE,$D7,$BC,$70
	.dc.b	$24,$66,$3C,$0D,$7C,$0D,$7C,$0D,$7C,$07,$62,$1D,$9D,$47,$68,$C9
	.dc.b	$D8,$A7,$62,$9D,$8A,$76,$3A,$18,$7A,$38,$21,$D4,$A7,$52,$9D,$4A
	.dc.b	$75,$38,$10,$E0,$53,$81,$47,$E3,$1F,$E5,$F7,$FC,$70,$D7,$EE,$B5
	.dc.b	$FF,$19,$B5,$FF,$3F,$35,$FF,$60,$35,$EB,$97,$F8,$CA,$31,$C7,$FE
	.dc.b	$9B,$9F,$F4,$AB,$FE,$B9,$E7,$FF,$59,$75,$7F,$AE,$4F,$F4,$E4,$7F
	.dc.b	$AC,$67,$FD,$42,$7F,$A3,$1E,$51,$BE,$9F,$F4,$9C,$73,$9F,$F5,$AD
	.dc.b	$7F,$A9,$DF,$F5,$2C,$FF,$AF,$54,$F5,$7F,$D6,$66,$8F,$F5,$14,$F9
	.dc.b	$27,$FD,$51,$5E,$4E,$9F,$F5,$3D,$7F,$A8,$4E,$E7,$97,$AB,$FD,$52
	.dc.b	$F6,$8F,$F5,$4B,$D4,$E1,$3F,$ED,$CA,$F2,$61,$E3,$8F,$F5,$D1,$ED
	.dc.b	$9F,$F5,$B1,$F0,$9F,$13,$67,$E0,$DF,$FB,$46,$78,$4F,$BA,$3C,$75
	.dc.b	$D6,$BB,$DF,$92,$31,$5F,$EB,$F3,$E4,$3F,$EA,$A7,$FD,$6A,$7F,$9B
	.dc.b	$7A,$3F,$D5,$84,$72,$7F,$AC,$C8,$7F,$AD,$8D,$1F,$EB,$F2,$47,$55
	.dc.b	$CD,$FE,$9C,$E8,$FF,$52,$F3,$7F,$AA,$E7,$FD,$24,$7F,$A5,$0F,$F5
	.dc.b	$C1,$FE,$9A,$E8,$FC,$D4,$6E,$32,$A7,$99,$23,$FD,$37,$1F,$E9,$FA
	.dc.b	$74,$7F,$A9,$63,$84,$8F,$2D,$F6,$BE,$51,$E3,$9E,$55,$C6,$73,$7F
	.dc.b	$EA,$DA,$F4,$BD,$9B,$3B,$7B,$47,$C1,$7C,$B3,$FE,$C7,$73,$7E,$16
	.dc.b	$F0,$79,$70,$F8,$07,$8A,$31,$5E,$9F,$8F,$7E,$82,$F7,$FF,$D6,$D1
	.dc.b	$86,$C6,$6B,$F9,$EC,$E5,$B1,$97,$FD,$6B,$19,$6C,$65,$F9,$02,$34
	.dc.b	$FF,$A9,$39,$FF,$D6,$64,$7F,$D6,$6D,$7F,$D7,$6D,$78,$1A,$21,$A2
	.dc.b	$9A,$30,$43,$04,$3F,$D5,$AF,$56,$9C,$8A,$71,$29,$C4,$FF,$56,$BC
	.dc.b	$5A,$71,$29,$C4,$A7,$11,$D1,$FE,$99,$78,$18,$8E,$8D,$33,$1B,$7A
	.dc.b	$19,$2B,$97,$B3,$5F,$49,$0F,$19,$4E,$47,$08,$F0,$C7,$18,$D9,$FE
	.dc.b	$A3,$C3,$FD,$5C,$47,$FD,$5C,$D7,$FD,$5C,$D7,$FD,$5C,$D7,$FD,$5C
	.dc.b	$D7,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38
	.dc.b	$1A,$21,$A2,$9A,$30,$43,$04,$3F,$DA,$8F,$B5,$A7,$98,$A7,$84,$A7
	.dc.b	$62,$9D,$4A,$75,$9F,$E2,$AF,$16,$BC,$5A,$F1,$6B,$C6,$3C,$91,$88
	.dc.b	$EA,$47,$04,$8D,$92,$3D,$51,$96,$C6,$5F,$23,$5F,$23,$5F,$23,$5F
	.dc.b	$21,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A
	.dc.b	$21,$A2,$9A,$30,$43,$04,$3E,$53,$ED,$69,$E6,$29,$E1,$29,$D8,$A7
	.dc.b	$52,$9D,$63,$E8,$3C,$5A,$F1,$6B,$C5,$AF,$18,$F2,$46,$23,$A9,$1C
	.dc.b	$12,$36,$48,$F5,$46,$5B,$19,$7C,$8D,$7C,$8D,$7C,$8D,$7C,$87,$22
	.dc.b	$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A
	.dc.b	$68,$FD,$A7,$33,$A7,$FD,$4E,$48,$FF,$53,$B5,$FF,$4C,$B5,$FF,$53
	.dc.b	$35,$FF,$53,$35,$FF,$53,$35,$FF,$57,$36,$3C,$11,$D8,$87,$62,$9D
	.dc.b	$8A,$76,$29,$D8,$A7,$62,$BD,$9E,$84,$3A,$14,$E8,$53,$A1,$4E,$85
	.dc.b	$7A,$3C,$88,$72,$29,$C8,$A7,$22,$BC,$9E,$04,$38,$14,$E0,$57,$83
	.dc.b	$FE,$98,$E5,$FF,$4B,$11,$FF,$51,$B5,$FF,$51,$B5,$FF,$51,$BF,$EE
	.dc.b	$27,$FC,$EA,$7F,$A7,$39,$7F,$CE,$5C,$BF,$E8,$83,$FC,$EB,$CD,$FE
	.dc.b	$6C,$E6,$FF,$44,$F2,$FF,$A5,$F0,$FF,$41,$61,$FE,$90,$3F,$D1,$6F
	.dc.b	$F9,$F9,$E9,$5C,$5F,$F4,$01,$E0,$8F,$F3,$FC,$FF,$A9,$2F,$C1,$39
	.dc.b	$6B,$FE,$83,$AF,$F4,$E3,$FE,$7F,$3F,$D4,$33,$FE,$82,$7F,$D3,$07
	.dc.b	$FA,$0B,$0F,$F4,$4E,$1F,$E8,$73,$D2,$FB,$A3,$96,$1F,$E9,$8C,$3B
	.dc.b	$DF,$8E,$73,$1F,$E9,$07,$A6,$1E,$27,$B9,$E2,$C3,$AC,$7B,$E3,$E0
	.dc.b	$9F,$EA,$D7,$85,$79,$CF,$95,$5F,$E5,$07,$15,$E7,$AF,$8E,$7C,$13
	.dc.b	$FD,$A2,$FF,$95,$AF,$FD,$9F,$D8,$ED,$FF,$40,$68,$E8,$FF,$4C,$E3
	.dc.b	$FE,$97,$25,$6F,$1F,$F2,$F9,$FE,$A2,$3F,$D4,$05,$3F,$D5,$35,$CD
	.dc.b	$B5,$D2,$79,$4E,$67,$FD,$1F,$CB,$FE,$7A,$7F,$CF,$E3,$97,$FC,$03
	.dc.b	$CF,$FE,$8D,$8F,$F4,$FF,$0F,$FA,$43,$9F,$FD,$0D,$1F,$E9,$58,$FF
	.dc.b	$45,$3D,$0E,$4F,$FA,$5B,$A3,$FD,$25,$19,$BE,$B5,$CA,$FF,$D3,$13
	.dc.b	$D8,$F1,$CE,$B0,$F3,$47,$73,$38,$78,$E3,$35,$CA,$BF,$D2,$D5,$CC
	.dc.b	$F0,$57,$8D,$FF,$52,$90,$D4,$FF,$A9,$5E,$C7,$03,$B9,$0C,$BB,$3E
	.dc.b	$34,$F8,$0F,$F5,$ED,$3F,$5D,$7C,$07,$EE,$E3,$0D,$8C,$DF,$DE,$13
	.dc.b	$A7,$0D,$7F,$06,$D7,$FD,$46,$D7,$FD,$98,$D7,$FD,$9D,$CE,$CB,$F8
	.dc.b	$64,$7F,$A1,$9B,$19,$8F,$E2,$0D,$7C,$8D,$7F,$26,$D7,$FD,$66,$D7
	.dc.b	$BC,$74,$23,$D1,$AF,$46,$BD,$1A,$F4,$6B,$D0,$E4,$43,$91,$4E,$45
	.dc.b	$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82
	.dc.b	$1F,$44,$F2,$14,$EE,$53,$A1,$4E,$65,$39,$C7,$D1,$70,$39,$DF,$3B
	.dc.b	$F8,$06,$C7,$F4,$D6,$BF,$91,$6B,$FC,$41,$AF,$FB,$C1,$A8,$ED,$1D
	.dc.b	$48,$F5,$6B,$D5,$AF,$56,$BD,$5A,$F5,$8E,$EF,$63,$A1,$0E,$85,$3A
	.dc.b	$14,$E8,$53,$A4,$76,$7A,$1C,$88,$72,$29,$C8,$A7,$23,$A1,$C8,$E0
	.dc.b	$43,$81,$4E,$07,$23,$81,$A2,$1A,$38,$1A,$30,$68,$C1,$23,$E3,$92
	.dc.b	$3D,$CD,$8F,$3B,$5F,$13,$5E,$ED,$7B,$35,$F0,$90,$F0,$47,$62,$3D
	.dc.b	$9A,$F6,$6B,$D9,$AF,$66,$BD,$63,$C0,$F6,$3A,$10,$E8,$53,$A1,$4E
	.dc.b	$85,$3A,$47,$67,$A1,$C8,$87,$22,$9C,$8A,$72,$3A,$1C,$8E,$04,$38
	.dc.b	$14,$E0,$72,$38,$1A,$21,$A3,$81,$A3,$06,$8C,$12,$3E,$39,$23,$DC
	.dc.b	$D8,$F3,$B5,$F1,$35,$EE,$D7,$B3,$5F,$09,$0F,$04,$76,$23,$D9,$AF
	.dc.b	$66,$BD,$9A,$F6,$6B,$D6,$3C,$0F,$63,$A1,$0E,$85,$3A,$14,$E8,$53
	.dc.b	$A4,$76,$7A,$1C,$88,$72,$29,$F3,$FA,$3F,$96,$3F,$E7,$56,$BF,$DE
	.dc.b	$5A,$FF,$8E,$1A,$FF,$27,$0E,$1F,$E5,$4D,$73,$3F,$DD,$A3,$04,$70
	.dc.b	$48,$F1,$35,$F1,$35,$F1,$35,$F1,$47,$37,$89,$0E,$25,$38,$94,$E3
	.dc.b	$1C,$DF,$4B,$5F,$49,$B3,$25,$32,$76,$21,$D8,$A7,$63,$84,$79,$DC
	.dc.b	$9D,$0A,$74,$39,$10,$E4,$69,$E4,$E4,$E4,$51,$D8,$EE,$FF,$13,$B6
	.dc.b	$3F,$D6,$8D,$7F,$C7,$0D,$7F,$CE,$ED,$7F,$00,$D7,$FB,$43,$5E,$F1
	.dc.b	$FE,$34,$E4,$FF,$4E,$47,$FA,$63,$9B,$FD,$40,$FF,$A7,$8F,$F5,$D3
	.dc.b	$FE,$BC,$7F,$D2,$0F,$FA,$EC,$FF,$5B,$F5,$7F,$AB,$23,$FD,$46,$7F
	.dc.b	$B0,$3A,$7F,$D8,$EF,$FA,$7C,$C1,$FE,$BB,$3B,$1D,$1D,$CF,$57,$FD
	.dc.b	$63,$3D,$67,$FD,$58,$7F,$A8,$AB,$D2,$7F,$AA,$63,$FD,$75,$1E,$B9
	.dc.b	$FF,$5B,$3C,$0F,$F5,$8D,$75,$7C,$35,$FE,$B0,$6F,$4F,$FA,$BD,$DD
	.dc.b	$79,$08,$FA,$4E,$C7,$C6,$7F,$D7,$11,$FE,$C0,$8F,$81,$5F,$26,$BB
	.dc.b	$47,$02,$2B,$75,$7F,$87,$E9,$87,$FA,$B7,$AB,$FD,$77,$D5,$71,$74
	.dc.b	$FF,$AB,$F2,$75,$7F,$AE,$E7,$FD,$77,$18,$BF,$F7,$F3,$63,$FD,$04
	.dc.b	$D7,$73,$FE,$BA,$BE,$D7,$D6,$31,$3F,$E9,$5E,$8F,$F4,$A7,$57,$FA
	.dc.b	$C1,$B1,$FE,$9F,$E9,$FF,$4B,$75,$7F,$AE,$27,$33,$FE,$A2,$E7,$FF
	.dc.b	$5A,$1F,$EB,$5B,$E4,$E3,$A3,$FD,$5F,$18,$7F,$D6,$31,$B9,$ED,$7D
	.dc.b	$63,$84,$7F,$B1,$0F,$F2,$45,$FF,$AD,$E7,$55,$C8,$EB,$3C,$8F,$65
	.dc.b	$72,$C3,$FD,$AB,$87,$F2,$B7,$D7,$3F,$E5,$3F,$15,$E2,$7F,$A7,$9B
	.dc.b	$1F,$EA,$D0,$E1,$FF,$58,$35,$FF,$77,$B5,$FE,$70,$D7,$B4,$75,$3F
	.dc.b	$D5,$35,$23,$FD,$52,$D7,$FD,$52,$D7,$FD,$52,$D7,$FD,$52,$D7,$FD
	.dc.b	$6E,$D5,$D5,$FE,$E6,$78,$9D,$88,$76,$29,$A3,$04,$30,$43,$C2,$70
	.dc.b	$29,$B2,$9B,$3C,$A6,$CA,$6C,$A6,$CF,$4C,$7A,$E3,$D5,$1D,$08,$70
	.dc.b	$29,$C0,$EC,$53,$B1,$4E,$CF,$47,$91,$0D,$1D,$1D,$38,$35,$D2,$E6
	.dc.b	$FF,$53,$12,$3F,$D4,$CD,$7F,$D4,$CD,$7F,$D4,$CD,$7F,$D4,$CD,$7F
	.dc.b	$D3,$4D,$7F,$D6,$8D,$8F,$04,$76,$21,$D8,$A7,$62,$9D,$8A,$76,$29
	.dc.b	$D8,$AF,$67,$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$5E,$8F,$22,$1C,$8A
	.dc.b	$72,$29,$C8,$AF,$27,$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9,$C1
	.dc.b	$1C,$12,$3E,$A1,$23,$E4,$B5,$F8,$0D,$7D,$2D,$7C,$8D,$7C,$2D,$7C
	.dc.b	$0D,$8F,$04,$76,$21,$D8,$A7,$62,$9D,$8A,$76,$29,$D8,$AF,$67,$A1
	.dc.b	$0E,$85,$3A,$14,$E8,$53,$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8,$AF
	.dc.b	$27,$81,$0E,$05,$38,$15,$E0,$E8,$86,$8A,$E9,$C1,$1C,$12,$3E,$A1
	.dc.b	$23,$E4,$B5,$F8,$0D,$7D,$2D,$7C,$8D,$7C,$2D,$7C,$0D,$8F,$04,$76
	.dc.b	$21,$D8,$A7,$62,$9D,$8A,$76,$29,$D8,$AF,$67,$A1,$0E,$85,$3A,$1F
	.dc.b	$7F,$A5,$EA,$FF,$A4,$C9,$1F,$EA,$16,$BF,$EA,$16,$BF,$EA,$16,$BF
	.dc.b	$E9,$06,$BF,$EA,$06,$BF,$EA,$46,$C7,$82,$3B,$10,$EC,$53,$B1,$4E
	.dc.b	$C5,$3B,$14,$EC,$57,$B3,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$AF,$47
	.dc.b	$91,$0E,$45,$39,$14,$E4,$57,$93,$C0,$87,$02,$9C,$0A,$F0,$7F,$D1
	.dc.b	$FF,$25,$FE,$9A,$23,$FE,$9A,$6B,$FE,$9A,$6B,$FE,$9A,$7F,$D2,$9D
	.dc.b	$3F,$E7,$FE,$8F,$F4,$9C,$FF,$99,$A3,$FC,$FB,$CD,$FE,$7F,$E9,$FF
	.dc.b	$33,$74,$FF,$9D,$2B,$FD,$14,$7F,$9E,$8F,$F4,$9D,$FF,$9C,$6F,$8E
	.dc.b	$9F,$E7,$D8,$FF,$4C,$DF,$2A,$E0,$7F,$A1,$E3,$2F,$FA,$60,$EA,$7F
	.dc.b	$9E,$27,$95,$FF,$A5,$AB,$94,$E4,$F0,$C7,$FA,$11,$FF,$4C,$5F,$B6
	.dc.b	$3F,$D2,$B7,$EE,$C3,$D9,$5F,$E8,$89,$E0,$FF,$A4,$8F,$F4,$84,$6A
	.dc.b	$7F,$D3,$51,$F1,$6F,$C3,$3F,$EA,$4C,$3C,$B3,$FE,$9E,$C3,$E0,$BE
	.dc.b	$43,$DF,$7E,$19,$DC,$73,$11,$FE,$92,$42,$BF,$D1,$5C,$FF,$E8,$5E
	.dc.b	$9F,$F4,$AC,$EA,$73,$D3,$FE,$97,$8F,$F4,$6C,$FF,$A4,$BA,$7F,$D2
	.dc.b	$D1,$D0,$E8,$E1,$E4,$FF,$A4,$1C,$C7,$66,$9F,$EA,$29,$E5,$5E,$08
	.dc.b	$FF,$3D,$F3,$FF,$9F,$DF,$F4,$20,$E9,$FF,$42,$3A,$32,$BA,$3F,$D1
	.dc.b	$D1,$FE,$97,$9E,$5D,$1F,$E8,$48,$FF,$44,$1C,$E7,$FD,$11,$D3,$EB
	.dc.b	$BE,$57,$FE,$8E,$AE,$47,$6A,$EB,$7F,$E8,$C8,$E5,$3D,$AF,$95,$69
	.dc.b	$FF,$4C,$CF,$19,$E9,$5C,$6F,$BC,$E4,$91,$E7,$7C,$2A,$BA,$D7,$27
	.dc.b	$AC,$66,$7F,$D7,$DC,$FF,$EB,$17,$9C,$FB,$63,$FD,$48,$F0,$39,$C7
	.dc.b	$CA,$8C,$BF,$36,$7A,$3E,$E3,$47,$E3,$1C,$F4,$BE,$0B,$FD,$58,$49
	.dc.b	$FF,$56,$46,$5B,$19,$7F,$D5,$AD,$7F,$18,$7F,$A7,$3E,$0B,$FD,$68
	.dc.b	$D3,$FD,$6D,$D4,$8C,$B6,$33,$18,$23,$82,$1D,$8E,$85,$8E,$4D,$8E
	.dc.b	$2D,$38,$9E,$68,$DB,$63,$6D,$36,$7E,$9E,$36,$D8,$DB,$4D,$9F,$B8
	.dc.b	$8D,$B6,$36,$D3,$67,$A8,$F0,$1E,$72,$18,$21,$DC,$C9,$4C,$9D,$4C
	.dc.b	$94,$C9,$DC,$E0,$43,$45,$34,$60,$86,$09,$1E,$E7,$1D,$4E,$5F,$C0
	.dc.b	$B6,$3F,$DE,$CD,$7E,$D3,$5F,$C4,$B5,$FE,$2C,$D7,$F9,$1B,$53,$49
	.dc.b	$1E,$02,$47,$81,$AF,$81,$AF,$81,$AF,$81,$AF,$81,$AF,$81,$B1,$E0
	.dc.b	$8E,$C4,$3B,$14,$EC,$53,$B1,$4E,$C5,$3B,$15,$EC,$F4,$21,$D0,$A7
	.dc.b	$42,$9D,$0A,$74,$2B,$D1,$E4,$43,$91,$4E,$45,$39,$15,$E4,$F0,$21
	.dc.b	$C0,$A7,$02,$BC,$1D,$10,$D1,$5D,$38,$23,$82,$47,$D4,$24,$7C,$96
	.dc.b	$BF,$01,$AF,$A5,$AF,$91,$AF,$85,$AF,$81,$B1,$E0,$8E,$C4,$3B,$14
	.dc.b	$EC,$53,$B1,$4E,$C5,$3B,$15,$EC,$F4,$21,$D0,$A7,$42,$9D,$0A,$74
	.dc.b	$2B,$D1,$E4,$43,$91,$4E,$45,$39,$15,$E4,$F0,$21,$C0,$A7,$02,$BC
	.dc.b	$1D,$10,$D1,$5D,$38,$23,$82,$47,$D4,$24,$7C,$96,$BF,$01,$AF,$A5
	.dc.b	$AF,$91,$AF,$85,$AF,$81,$B1,$E0,$8E,$C4,$3B,$14,$EC,$53,$B1,$4E
	.dc.b	$C5,$1D,$2F,$82,$FE,$D6,$D8,$FE,$C4,$6A,$32,$47,$31,$B7,$FB,$43
	.dc.b	$5F,$EE,$ED,$7F,$C8,$8D,$7F,$CA,$2D,$7B,$1D,$23,$B3,$D1,$AF,$46
	.dc.b	$BD,$1A,$F4,$6B,$D0,$E4,$75,$21,$82,$1D,$4C,$94,$C9,$DC,$C9,$4C
	.dc.b	$9E,$13,$25,$32,$78,$CE,$84,$39,$14,$E0,$53,$81,$A2,$1A,$29,$A3
	.dc.b	$04,$30,$4E,$E7,$C8,$7F,$8A,$9B,$1F,$80,$6B,$FE,$23,$6B,$FE,$32
	.dc.b	$6B,$FE,$7F,$6B,$FE,$8C,$6A,$6F,$1F,$FA,$6E,$3F,$D3,$07,$FA,$B4
	.dc.b	$FF,$5B,$75,$FF,$AC,$5F,$F5,$B4,$6D,$C9,$FE,$A1,$EB,$FF,$52,$E7
	.dc.b	$FE,$B1,$8F,$F4,$B3,$FE,$A4,$76,$7F,$A8,$7A,$BF,$D5,$B5,$FE,$AF
	.dc.b	$BF,$F6,$73,$D6,$3F,$D5,$13,$E4,$87,$5F,$FA,$6E,$7F,$D9,$27,$FB
	.dc.b	$47,$0F,$09,$FE,$B6,$8D,$5F,$A6,$75,$7F,$EC,$13,$FD,$6B,$1F,$EC
	.dc.b	$62,$3E,$8A,$E6,$FC,$0C,$3D,$8F,$FB,$5B,$0F,$35,$70,$8C,$4A,$BC
	.dc.b	$51,$FE,$D4,$6B,$FE,$CF,$3F,$D5,$BD,$7F,$EB,$D7,$FD,$5C,$7F,$AD
	.dc.b	$3A,$BF,$D6,$58,$7F,$AB,$E3,$FD,$66,$7F,$AF,$8F,$F3,$22,$EB,$FF
	.dc.b	$5E,$46,$8F,$F5,$BB,$FE,$BD,$8E,$44,$7F,$D8,$1D,$3F,$E9,$CE,$6F
	.dc.b	$F4,$4B,$7A,$3F,$D1,$F1,$FE,$9D,$3F,$D4,$5D,$5F,$EB,$2C,$3F,$D3
	.dc.b	$27,$FA,$DA,$B8,$3F,$E9,$3A,$FF,$4D,$39,$9E,$0F,$FA,$56,$BA,$4F
	.dc.b	$FA,$76,$39,$9F,$EA,$A7,$C2,$72,$70,$6A,$3C,$7D,$5F,$EB,$B8,$C4
	.dc.b	$F3,$7B,$3C,$67,$71,$FE,$AE,$36,$79,$CA,$70,$30,$7F,$B6,$FA,$BF
	.dc.b	$D6,$D1,$F1,$AF,$FD,$91,$3A,$C3,$E0,$BB,$3F,$D6,$17,$FE,$FC,$F4
	.dc.b	$DE,$9B,$CD,$FE,$77,$18,$6C,$66,$BF,$D4,$13,$96,$C6,$5F,$F5,$8C
	.dc.b	$65,$B1,$97,$FE,$00,$46,$5F,$F5,$17,$CA,$7F,$AC,$88,$FF,$AD,$9A
	.dc.b	$FF,$AB,$5A,$F0,$34,$43,$45,$34,$60,$86,$08,$7F,$AB,$5E,$AD,$39
	.dc.b	$14,$E2,$53,$89,$FE,$AF,$78,$B4,$E2,$53,$89,$4E,$23,$AD,$1C,$A3
	.dc.b	$0D,$8C,$C7,$43,$25,$72,$F6,$32,$57,$2F,$81,$AF,$AC,$87,$94,$A7
	.dc.b	$23,$84,$79,$7A,$DF,$31,$FE,$AE,$24,$FF,$AB,$9A,$FF,$AB,$9A,$FF
	.dc.b	$AB,$9A,$FF,$AB,$9A,$FF,$AB,$9A,$FF,$AB,$9B,$1E,$08,$EC,$43,$B1
	.dc.b	$4E,$C5,$3B,$14,$EC,$53,$B1,$5E,$CF,$42,$1D,$0A,$74,$29,$D0,$A7
	.dc.b	$42,$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A,$70,$2B
	.dc.b	$C1,$D1,$0D,$15,$D3,$82,$38,$24,$7D,$42,$47,$C9,$6B,$F0,$1A,$FA
	.dc.b	$5A,$F9,$1A,$F8,$5A,$F8,$1B,$1E,$08,$EC,$43,$B1,$4E,$C5,$3B,$14
	.dc.b	$EC,$53,$B1,$5E,$CF,$42,$1D,$0A,$74,$29,$D0,$A7,$42,$BD,$1E,$44
	.dc.b	$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1,$D1,$0D,$15
	.dc.b	$D3,$82,$38,$24,$7D,$42,$47,$C9,$6B,$F0,$1A,$FA,$5A,$F9,$1A,$F8
	.dc.b	$5A,$F8,$1B,$1E,$08,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$53,$B1,$5E
	.dc.b	$CF,$42,$1D,$0A,$74,$3E,$FF,$67,$F5,$F7,$FD,$37,$97,$FA,$A0,$8F
	.dc.b	$FA,$91,$AF,$FA,$91,$AF,$FA,$91,$AF,$23,$81,$0E,$05,$38,$14,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$10,$FF,$59,$BE,$46,$9D,$CA,$74,$29,$CC
	.dc.b	$A7,$33,$FD,$9A,$F3,$69,$CC,$A7,$32,$9C,$CA,$73,$8F,$A0,$F6,$21
	.dc.b	$A2,$BA,$70,$47,$1F,$8D,$BE,$17,$FD,$BB,$28,$C3,$FE,$E0,$6A,$21
	.dc.b	$5C,$B8,$39,$47,$48,$CC,$70,$6B,$C6,$3C,$3C,$9F,$E7,$C8,$FF,$41
	.dc.b	$1F,$E8,$73,$FD,$49,$D5,$FE,$94,$7F,$D3,$71,$FE,$99,$77,$1F,$E8
	.dc.b	$E3,$FD,$3B,$D7,$FE,$92,$8C,$F5,$FF,$A7,$BA,$FF,$D1,$76,$7F,$D4
	.dc.b	$D7,$D4,$93,$FE,$8D,$9D,$75,$FF,$A5,$27,$B4,$F3,$8F,$F4,$6C,$7F
	.dc.b	$A9,$23,$E2,$CF,$FA,$A2,$FF,$D2,$47,$FA,$6E,$B7,$3F,$EA,$53,$78
	.dc.b	$79,$E7,$FD,$6D,$1D,$5E,$04,$BF,$0B,$EC,$3F,$D5,$65,$3F,$D4,$D3
	.dc.b	$CE,$FE,$34,$60,$F1,$CF,$27,$FD,$47,$E1,$7F,$A6,$3A,$FF,$D4,$3D
	.dc.b	$7F,$E9,$92,$3F,$76,$7F,$D4,$C7,$FA,$8A,$3F,$D4,$37,$FE,$A3,$9D
	.dc.b	$4F,$43,$7D,$9F,$EA,$F2,$5F,$FA,$1F,$A3,$FC,$F2,$3A,$FF,$CF,$FD
	.dc.b	$3F,$E9,$2B,$FC,$84,$E5,$FF,$4E,$76,$7F,$A4,$5F,$F4,$41,$FE,$81
	.dc.b	$3F,$D2,$E6,$23,$83,$87,$BD,$76,$8F,$F4,$8C,$78,$BA,$FF,$D4,$EF
	.dc.b	$0E,$CF,$F4,$C9,$FE,$91,$BF,$04,$67,$0F,$55,$7F,$A5,$2F,$C0,$F2
	.dc.b	$C3,$B4,$78,$DF,$D4,$E1,$CE,$3F,$D4,$86,$4F,$F4,$C8,$EB,$7C,$D3
	.dc.b	$2F,$F4,$CB,$67,$FD,$2E,$D7,$FD,$2A,$D7,$FD,$26,$D7,$FD,$48,$D7
	.dc.b	$B3,$67,$BC,$FF,$04,$EC,$7C,$B5,$25,$FD,$E8,$CB,$63,$2F,$E1,$E3
	.dc.b	$2D,$8C,$BF,$C1,$E3,$2D,$8C,$BF,$C3,$A3,$2D,$8C,$BF,$C4,$E3,$2D
	.dc.b	$8C,$BF,$E8,$F6,$C6,$A3,$07,$9C,$87,$94,$A7,$8C,$A7,$84,$A7,$72
	.dc.b	$9D,$0E,$44,$39,$14,$E4,$53,$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0
	.dc.b	$D1,$0D,$14,$D1,$82,$0E,$B7,$CC,$FF,$44,$6C,$7E,$F1,$AF,$F0,$36
	.dc.b	$BF,$D0,$5A,$FF,$46,$6B,$FD,$25,$AF,$78,$EC,$47,$B3,$5E,$CD,$7B
	.dc.b	$35,$EC,$D7,$B3,$5E,$C7,$42,$1D,$0A,$74,$29,$D0,$A7,$42,$9D,$0E
	.dc.b	$44,$39,$14,$E4,$53,$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D
	.dc.b	$14,$D1,$82,$18,$24,$7D,$42,$47,$C9,$6C,$7C,$06,$BE,$96,$BE,$46
	.dc.b	$BE,$16,$BE,$06,$BE,$08,$EC,$47,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B3
	.dc.b	$5E,$C7,$42,$1D,$0A,$74,$29,$D0,$A7,$42,$9D,$0E,$44,$39,$14,$E4
	.dc.b	$53,$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82,$18
	.dc.b	$24,$7D,$42,$47,$C9,$6C,$7C,$06,$BE,$96,$BE,$46,$BE,$16,$BE,$06
	.dc.b	$BE,$08,$EC,$47,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B3,$5E,$C7,$42,$1D
	.dc.b	$0A,$74,$29,$D0,$A7,$42,$9D,$0E,$44,$3E,$AF,$C0,$7F,$6D,$4D,$26
	.dc.b	$1F,$E7,$06,$C7,$F8,$A1,$AF,$F8,$BD,$AF,$F9,$B5,$AF,$F9,$C9,$AF
	.dc.b	$F9,$D5,$AF,$82,$3B,$11,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B1
	.dc.b	$D0,$87,$42,$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0E,$45,$39,$14,$E4
	.dc.b	$53,$91,$C0,$87,$02,$9C,$0A,$7F,$A8,$FE,$17,$FD,$34,$47,$FD,$48
	.dc.b	$D7,$FC,$C9,$D8,$8C,$47,$FA,$8E,$30,$FF,$99,$39,$7F,$D3,$C9,$FF
	.dc.b	$46,$BF,$E9,$F7,$FD,$10,$B9,$FF,$D1,$8F,$F9,$F8,$E1,$D7,$FE,$98
	.dc.b	$EA,$FF,$4B,$36,$3F,$D0,$55,$D8,$EB,$5F,$EA,$48,$FF,$4D,$57,$FA
	.dc.b	$36,$3F,$D0,$53,$BA,$E5,$5D,$BA,$F6,$F8,$67,$9D,$F1,$3F,$D3,$31
	.dc.b	$FE,$8C,$6B,$E9,$7F,$D3,$CF,$1A,$EB,$3C,$A3,$CF,$D7,$FE,$97,$3F
	.dc.b	$D3,$13,$A2,$5F,$8A,$7F,$D8,$77,$F0,$23,$17,$F2,$30,$F8,$F5,$A3
	.dc.b	$E1,$DF,$FB,$00,$E6,$FC,$FC,$3C,$0F,$28,$5F,$FA,$DA,$3F,$DC,$17
	.dc.b	$FE,$AA,$3E,$34,$7F,$B8,$ED,$87,$FA,$A6,$9D,$3F,$E9,$E7,$FD,$43
	.dc.b	$D3,$98,$3B,$8F,$F7,$46,$1F,$EA,$97,$FD,$59,$D5,$F8,$7E,$BF,$F5
	.dc.b	$69,$FE,$B2,$8E,$B5,$CC,$8F,$FA,$9B,$A7,$FD,$32,$7F,$A4,$FA,$FF
	.dc.b	$D3,$9C,$3F,$E9,$8E,$AF,$F4,$9C,$7F,$A8,$AB,$FD,$56,$7F,$A4,$9C
	.dc.b	$3F,$EA,$7A,$E4,$74,$AF,$F4,$AC,$7F,$AB,$FA,$BF,$D4,$BD,$7F,$EB
	.dc.b	$78,$FF,$4D,$1C,$67,$77,$FE,$A7,$8F,$34,$7F,$AA,$AF,$AB,$4D,$E1
	.dc.b	$E6,$8C,$CF,$95,$B8,$7F,$AA,$49,$82,$7F,$D8,$F5,$C6,$F9,$C7,$A3
	.dc.b	$C7,$76,$FF,$B5,$A1,$18,$7F,$DB,$2D,$7F,$DC,$0D,$7B,$1D,$DC,$38
	.dc.b	$6B,$98,$ED,$18,$87,$53,$E2,$3F,$D5,$C4,$8F,$F5,$63,$5F,$F5,$53
	.dc.b	$5F,$F5,$43,$5F,$F5,$33,$5F,$F5,$23,$5E,$ED,$8F,$04,$76,$21,$D8
	.dc.b	$A7,$62,$9D,$8A,$76,$29,$D8,$AF,$67,$A1,$0E,$85,$3A,$14,$E8,$53
	.dc.b	$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8,$AF,$27,$81,$0E,$05,$38,$15
	.dc.b	$E0,$FF,$A6,$FE,$23,$FD,$4C,$47,$FD,$4C,$D7,$FD,$4C,$D7,$FD,$4C
	.dc.b	$D7,$FD,$4C,$D7,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81
	.dc.b	$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3F,$D9,$8F,$B5,$A7,$98
	.dc.b	$A7,$84,$A7,$62,$9D,$4A,$75,$EC,$47,$56,$C7,$56,$BD,$5A,$F5,$6B
	.dc.b	$D5,$AF,$58,$E8,$47,$A3,$5E,$8D,$7A,$35,$E8,$D7,$A1,$C8,$87,$22
	.dc.b	$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$3E,$99,$23,$E4,$B6,$3E,$03,$4F,$49,$4F,$21,$4F,$09,$4F
	.dc.b	$01,$4F,$04,$76,$24,$76,$6C,$76,$6B,$D9,$AF,$66,$BD,$9A,$F6,$8E
	.dc.b	$84,$7A,$35,$E8,$D7,$A3,$5E,$8D,$7A,$1C,$88,$72,$29,$C8,$A7,$22
	.dc.b	$9C,$8E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$43,$F0
	.dc.b	$3D,$6E,$8F,$F4,$E1,$23,$FD,$38,$D7,$FD,$38,$D7,$FD,$38,$D7,$FD
	.dc.b	$38,$D7,$FD,$38,$D7,$FD,$56,$D8,$F0,$47,$62,$1D,$8A,$76,$29,$D8
	.dc.b	$A7,$62,$9D,$8A,$F6,$7A,$10,$E8,$53,$A1,$4E,$85,$3A,$15,$E8,$F2
	.dc.b	$21,$C8,$A7,$22,$9C,$8A,$F2,$78,$10,$E0,$53,$81,$5E,$0F,$F8,$A7
	.dc.b	$CE,$7C,$7F,$F8,$95,$B1,$FE,$20,$3F,$C4,$2E,$1C,$07,$0E,$DA,$F0
	.dc.b	$8E,$84,$7F,$D4,$95,$FE,$75,$EA,$FF,$3D,$F5,$FF,$9D,$27,$FD,$49
	.dc.b	$39,$7F,$CF,$07,$FA,$3C,$FF,$46,$4F,$07,$07,$FA,$04,$FF,$4D,$1C
	.dc.b	$0F,$F4,$A1,$FE,$7C,$9F,$F4,$7C,$71,$8C,$9F,$EB,$58,$CB,$C7,$AF
	.dc.b	$FD,$1B,$D3,$FE,$8F,$EB,$FF,$4E,$CF,$78,$FF,$4E,$3E,$09,$FF,$56
	.dc.b	$E1,$E9,$7F,$D6,$51,$EC,$C3,$FD,$1C,$F7,$3F,$D2,$4E,$2B,$FD,$21
	.dc.b	$5C,$6F,$FD,$61,$5F,$00,$F2,$3E,$29,$F2,$47,$B4,$E6,$7B,$1C,$BD
	.dc.b	$CF,$11,$9C,$3E,$86,$1F,$E9,$C8,$FF,$65,$F4,$FF,$A4,$7A,$FF,$D3
	.dc.b	$91,$FE,$D9,$8F,$F4,$A3,$FE,$9A,$7F,$D2,$FD,$7F,$E9,$92,$47,$FA
	.dc.b	$60,$FF,$51,$90,$E4,$FF,$9E,$67,$FD,$4B,$1F,$E8,$63,$FD,$01,$CB
	.dc.b	$FE,$7A,$7F,$D1,$83,$A1,$D1,$FE,$74,$EA,$FF,$47,$49,$FF,$50,$BF
	.dc.b	$E9,$DE,$9F,$F4,$23,$C6,$78,$46,$2B,$9C,$FF,$A2,$8F,$F5,$2B,$FE
	.dc.b	$8E,$7C,$8E,$9F,$F5,$14,$FF,$A8,$6B,$BC,$E5,$F1,$98,$8D,$57,$77
	.dc.b	$C3,$5D,$9E,$5D,$5F,$E9,$E6,$E1,$F0,$70,$FF,$57,$D7,$8A,$79,$BC
	.dc.b	$4C,$10,$F2,$9F,$E9,$7E,$DF,$F4,$99,$1F,$F4,$FB,$5F,$F4,$FB,$5F
	.dc.b	$E0,$1D,$48,$FF,$54,$74,$FD,$69,$FE,$02,$D7,$F8,$33,$5F,$F5,$3B
	.dc.b	$5F,$F5,$B3,$5F,$F5,$1F,$67,$42,$3F,$C3,$0C,$96,$38,$12,$37,$1E
	.dc.b	$26,$BE,$26,$BE,$26,$BE,$26,$BD,$8E,$47,$77,$31,$DE,$30,$43,$64
	.dc.b	$3C,$06,$4A,$64,$F1,$19,$29,$93,$C8,$64,$A6,$4F,$31,$D0,$87,$22
	.dc.b	$9C,$0A,$70,$34,$43,$45,$34,$7F,$BA,$FA,$9F,$11,$FE,$E7,$6C,$7D
	.dc.b	$56,$BF,$D0,$9A,$FF,$47,$6B,$FE,$E8,$6B,$D6,$3A,$11,$E8,$D7,$A3
	.dc.b	$5E,$8D,$7A,$35,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10,$E0
	.dc.b	$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$04,$7C,$86,$C7,$BD,$A7
	.dc.b	$A0,$A7,$8C,$A7,$80,$A7,$72,$9D,$E3,$B1,$23,$B3,$63,$B3,$5E,$CD
	.dc.b	$7B,$35,$EC,$D7,$B4,$74,$23,$D1,$AF,$46,$BD,$1A,$F4,$6B,$D0,$E4
	.dc.b	$43,$91,$4E,$45,$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1
	.dc.b	$4D,$18,$21,$82,$1F,$50,$91,$F2,$5B,$1F,$01,$A7,$A4,$A7,$90,$A7
	.dc.b	$84,$A7,$80,$A7,$82,$3B,$12,$3B,$36,$3B,$35,$EC,$D7,$B3,$5E,$CD
	.dc.b	$7B,$47,$42,$3D,$1A,$F4,$6B,$D1,$AF,$46,$BD,$0E,$44,$39,$14,$E4
	.dc.b	$53,$91,$4E,$47,$02,$1C,$0A,$70,$29,$C0,$D1,$01,$3A,$9C,$5F,$D7
	.dc.b	$61,$18,$7F,$B0,$35,$FF,$36,$B5,$FF,$39,$35,$EA,$72,$8E,$91,$98
	.dc.b	$F0,$46,$08,$F6,$6B,$D9,$AF,$66,$BD,$A3,$A3,$DC,$E4,$43,$91,$4E
	.dc.b	$45,$39,$47,$47,$CA,$64,$A7,$22,$1C,$8A,$72,$38,$1E,$33,$27,$23
	.dc.b	$81,$0E,$06,$8E,$06,$4E,$06,$8C,$1A,$30,$43,$64,$34,$58,$F8,$E7
	.dc.b	$C8,$EA,$7C,$17,$FA,$B4,$93,$FE,$1F,$8C,$B6,$32,$FF,$AB,$A3,$2D
	.dc.b	$8F,$C5,$73,$FF,$A4,$0F,$F3,$E9,$FE,$88,$EA,$FF,$45,$74,$7F,$A6
	.dc.b	$BA,$3F,$D2,$27,$F9,$F6,$FF,$D2,$DD,$3F,$E8,$C7,$FD,$35,$D4,$BF
	.dc.b	$F5,$25,$7F,$A8,$2F,$FD,$0B,$7F,$90,$C3,$C1,$7F,$E9,$5C,$3F,$3F
	.dc.b	$1E,$3A,$DC,$62,$7A,$3F,$E9,$83,$FD,$2C,$7F,$A4,$AF,$FD,$4F,$87
	.dc.b	$A2,$3F,$D3,$31,$87,$FD,$2C,$7F,$AF,$AF,$DF,$7E,$DC,$3D,$51,$89
	.dc.b	$E3,$1E,$79,$F1,$C7,$C0,$3C,$97,$DD,$F1,$4E,$DE,$D3,$C2,$BD,$F5
	.dc.b	$BB,$F9,$51,$8A,$F2,$92,$33,$87,$C0,$7E,$8E,$1F,$26,$30,$FA,$30
	.dc.b	$ED,$18,$AF,$29,$CE,$7F,$D7,$D3,$AC,$3E,$AC,$62,$7F,$D9,$15,$C5
	.dc.b	$FF,$56,$9F,$EB,$12,$F4,$FF,$AE,$B1,$FE,$27,$D5,$FE,$B1,$E2,$FB
	.dc.b	$9B,$FF,$AE,$AF,$FD,$FB,$5A,$6C,$66,$F9,$3F,$EC,$5A,$FF,$4B,$CF
	.dc.b	$FA,$BC,$FF,$4A,$1F,$EB,$3E,$87,$47,$F2,$4E,$8F,$F5,$64,$FF,$A5
	.dc.b	$8D,$46,$7A,$39,$CF,$FA,$AF,$A7,$FD,$51,$5F,$EB,$53,$FD,$87,$19
	.dc.b	$7B,$F5,$7F,$AB,$DD,$47,$FA,$A1,$EA,$79,$4E,$35,$FE,$B2,$69,$C1
	.dc.b	$F0,$9F,$EA,$E7,$27,$9E,$7B,$3D,$27,$C4,$FF,$B3,$4F,$E8,$3D,$4F
	.dc.b	$87,$FE,$80,$D8,$FF,$4F,$35,$FE,$76,$70,$8D,$91,$DB,$5F,$F5,$8F
	.dc.b	$A1,$FE,$B4,$23,$FE,$A7,$6B,$A3,$04,$30,$43,$FD,$72,$F0,$69,$B2
	.dc.b	$9B,$3F,$D7,$AE,$DA,$6C,$A6,$CF,$F6,$0B,$B6,$9B,$29,$B1,$D6,$8E
	.dc.b	$11,$86,$C6,$63,$91,$92,$B9,$7A,$35,$F5,$90,$F3,$94,$F1,$94,$E8
	.dc.b	$72,$8F,$09,$1E,$ED,$7F,$D4,$9F,$0F,$FE,$A9,$23,$FE,$A9,$6B,$A3
	.dc.b	$04,$30,$43,$FD,$58,$F0,$69,$B2,$9B,$3F,$D6,$0E,$DA,$6C,$A6,$CF
	.dc.b	$F6,$03,$B6,$9B,$29,$B3,$FD,$2A,$ED,$A6,$CA,$6F,$B1,$1B,$6C,$6D
	.dc.b	$AE,$E3,$44,$74,$D7,$46,$08,$60,$87,$C4,$21,$F0,$0A,$7B,$0A,$7A
	.dc.b	$0B,$1E,$46,$C7,$85,$A7,$80,$A7,$80,$EC,$43,$B1,$4E,$C5,$8E,$CD
	.dc.b	$8E,$CD,$3B,$14,$EC,$74,$21,$D0,$B1,$D1,$B1,$D1,$A7,$42,$9D,$0E
	.dc.b	$44,$8E,$4D,$8E,$4D,$39,$14,$E5,$1C,$09,$1C,$1B,$1C,$1A,$F0,$8D
	.dc.b	$11,$D3,$5D,$18,$21,$82,$1F,$50,$87,$C9,$29,$F0,$0A,$7A,$4B,$1E
	.dc.b	$46,$C7,$85,$A7,$80,$A7,$80,$EC,$43,$B1,$4E,$C5,$8E,$CD,$8E,$CD
	.dc.b	$3B,$14,$EC,$74,$21,$D0,$B1,$D1,$B1,$D1,$A7,$42,$9D,$0E,$44,$8E
	.dc.b	$4D,$8E,$4D,$39,$14,$E5,$1C,$0E,$3D,$6F,$87,$FF,$4B,$92,$7F,$D4
	.dc.b	$2D,$7F,$D4,$2D,$7F,$D4,$2D,$7F,$D2,$8D,$7F,$D5,$CD,$7F,$D5,$AD
	.dc.b	$8F,$04,$76,$21,$D8,$A7,$62,$9D,$8A,$76,$29,$D8,$AF,$67,$A1,$0E
	.dc.b	$85,$3A,$14,$E8,$53,$A1,$5E,$8F,$22,$1C,$8A,$72,$29,$C8,$AF,$27
	.dc.b	$81,$0E,$05,$38,$15,$E0,$FF,$B7,$3B,$1E,$DF,$FB,$6D,$B1,$FE,$9B
	.dc.b	$8C,$11,$C1,$0F,$F4,$E3,$FE,$9A,$E8,$FF,$45,$47,$F9,$D4,$FF,$49
	.dc.b	$F5,$7F,$97,$9F,$F4,$3F,$67,$F9,$B2,$31,$0E,$9F,$F5,$03,$FE,$67
	.dc.b	$37,$D7,$FE,$6F,$9D,$CE,$AB,$FD,$18,$49,$FF,$40,$57,$13,$C5,$3F
	.dc.b	$E9,$B7,$FD,$06,$43,$4F,$FA,$06,$79,$47,$80,$87,$2B,$F6,$CF,$38
	.dc.b	$E3,$1F,$E7,$78,$C4,$FF,$AB,$E7,$B3,$5D,$DF,$AB,$B3,$FD,$41,$1F
	.dc.b	$E8,$69,$F0,$BF,$20,$FF,$57,$1F,$16,$36,$FB,$1F,$F4,$53,$AA,$F8
	.dc.b	$98,$7C,$2A,$F0,$1F,$E9,$EA,$E3,$3F,$E8,$3E,$BF,$F4,$0F,$67,$FA
	.dc.b	$27,$0F,$F6,$0F,$62,$31,$1F,$E9,$23,$FC,$BB,$5F,$E8,$B6,$CF,$FA
	.dc.b	$3D,$B3,$C2,$7B,$76,$7F,$A5,$23,$15,$D6,$BF,$D3,$9D,$1F,$E7,$21
	.dc.b	$D9,$FE,$8E,$8F,$F3,$91,$FE,$77,$71,$D3,$FE,$90,$9D,$F6,$7F,$A3
	.dc.b	$EB,$FD,$00,$0F,$F3,$E7,$5F,$FA,$09,$C9,$83,$53,$E1,$EC,$FF,$4A
	.dc.b	$C7,$7A,$E8,$EC,$92,$9D,$B6,$3F,$D3,$38,$79,$AF,$FD,$21,$87,$FA
	.dc.b	$66,$BB,$56,$B0,$EB,$1E,$A8,$F7,$BC,$DF,$F4,$BB,$67,$95,$F5,$9F
	.dc.b	$4B,$5F,$41,$C9,$C4,$F5,$9F,$F4,$E9,$CA,$BC,$33,$AC,$3E,$45,$F1
	.dc.b	$9C,$91,$F2,$C6,$AB,$E9,$57,$63,$94,$FD,$3F,$1D,$E4,$38,$7F,$D4
	.dc.b	$B3,$F5,$67,$31,$88,$46,$1D,$B5,$FF,$56,$B5,$F3,$B5,$E9,$3D,$DE
	.dc.b	$2D,$7F,$D6,$BE,$C3,$F0,$C7,$A3,$E1,$8C,$36,$33,$3E,$89,$CB,$63
	.dc.b	$2F,$46,$BF,$2A,$36,$E6,$7A,$11,$E4,$D7,$46,$08,$60,$87,$59,$EE
	.dc.b	$FE,$7E,$7B,$B9,$9E,$8D,$39,$94,$E6,$71,$9E,$6A,$79,$B9,$9E,$6D
	.dc.b	$39,$94,$E6,$71,$9E,$73,$CA,$31,$3C,$9A,$72,$29,$CA,$38,$12,$79
	.dc.b	$92,$79,$46,$5B,$19,$9E,$93,$97,$B0,$FC,$51,$EA,$7E,$46,$30,$D8
	.dc.b	$CC,$7E,$55,$AF,$E6,$9A,$FF,$C0,$06,$BF,$AA,$6B,$FC,$4D,$AF,$63
	.dc.b	$A4,$77,$23,$82,$1D,$DC,$B4,$C9,$E1,$72,$D3,$27,$8D,$CB,$4C,$9E
	.dc.b	$57,$2D,$32,$79,$CF,$01,$0E,$C5,$3A,$14,$E4,$53,$91,$C0,$87,$02
	.dc.b	$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$09,$1F,$48,$91,$F0,$23,$2D
	.dc.b	$8C,$C7,$C2,$6B,$EB,$6B,$E6,$6B,$E3,$6B,$E2,$6B,$D8,$E9,$1D,$C8
	.dc.b	$E0,$87,$77,$2D,$32,$78,$5C,$B4,$C9,$E3,$72,$D3,$27,$95,$CB,$4C
	.dc.b	$9E,$73,$C0,$43,$B1,$4E,$85,$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C
	.dc.b	$0D,$10,$D1,$4D,$18,$21,$82,$47,$D2,$24,$7C,$08,$CB,$63,$31,$F0
	.dc.b	$9A,$FA,$DA,$F9,$9A,$F8,$DA,$F8,$9A,$F6,$3A,$47,$72,$38,$21,$DD
	.dc.b	$CB,$4C,$9E,$17,$2D,$32,$78,$DC,$B4,$C9,$E5,$72,$D3,$27,$9C,$F0
	.dc.b	$10,$EC,$7D,$DF,$C9,$5E,$5F,$F2,$B6,$C7,$F7,$D6,$BF,$E0,$26,$BF
	.dc.b	$E1,$00,$E1,$FF,$09,$B5,$FF,$1E,$35,$DC,$77,$23,$DD,$AF,$76,$BD
	.dc.b	$E3,$91,$23,$31,$E0,$6B,$E0,$6B,$E0,$3B,$10,$EC,$53,$B3,$C2,$3B
	.dc.b	$46,$4E,$C5,$3B,$14,$EC,$74,$21,$D0,$D3,$D1,$C9,$D0,$A7,$42,$9D
	.dc.b	$0E,$46,$1E,$4E,$08,$73,$3F,$C5,$F9,$3D,$9F,$F1,$C6,$9F,$E3,$E6
	.dc.b	$C6,$67,$FC,$63,$18,$23,$82,$1F,$E4,$4E,$7F,$F5,$4F,$57,$FA,$64
	.dc.b	$FF,$60,$47,$FA,$77,$AB,$FD,$33,$D9,$FE,$B2,$27,$67,$F9,$FF,$8B
	.dc.b	$FD,$53,$82,$BF,$D4,$B1,$FE,$CF,$BF,$F5,$FC,$72,$AE,$18,$77,$9F
	.dc.b	$F5,$FD,$FF,$B0,$E3,$11,$FE,$B1,$36,$7F,$AD,$DF,$2D,$F8,$2F,$BC
	.dc.b	$60,$E3,$5F,$EB,$4A,$EB,$3E,$17,$C6,$74,$3F,$D9,$D7,$67,$D5,$1F
	.dc.b	$EC,$AA,$E7,$87,$FB,$3C,$F8,$31,$C8,$FF,$37,$47,$85,$F3,$1E,$03
	.dc.b	$A4,$71,$AE,$D7,$E8,$9D,$9F,$ED,$1C,$3E,$41,$ED,$7C,$47,$C6,$AE
	.dc.b	$4F,$86,$37,$7E,$18,$E4,$FF,$B0,$57,$67,$FB,$41,$BD,$99,$83,$B8
	.dc.b	$FF,$65,$9F,$85,$EB,$FF,$67,$3B,$88,$FF,$B6,$3B,$3F,$D9,$A7,$FB
	.dc.b	$63,$AF,$FD,$99,$D1,$FE,$94,$9F,$F5,$5F,$57,$FA,$E0,$8F,$FA,$CC
	.dc.b	$FF,$50,$1F,$E9,$E3,$1D,$5F,$EC,$F8,$DC,$72,$33,$D9,$FE,$BE,$3F
	.dc.b	$D9,$CF,$FB,$1A,$3F,$D4,$D1,$8B,$FF,$59,$3F,$EA,$E7,$FD,$78,$FF
	.dc.b	$91,$FC,$1F,$F2,$81,$2B,$FC,$A0,$D7,$FC,$A0,$D7,$FC,$A0,$FF,$95
	.dc.b	$F0,$91,$FE,$57,$6B,$FE,$57,$6B,$FE,$57,$6B,$FE,$57,$6B,$FE,$57
	.dc.b	$6B,$FE,$67,$6F,$4F,$FB,$DA,$3B,$10,$EC,$53,$B1,$4E,$C5,$3B,$14
	.dc.b	$EC,$57,$B3,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$AF,$47,$91,$0E,$45
	.dc.b	$39,$14,$E4,$57,$93,$C0,$87,$02,$9C,$0A,$F0,$74,$7F,$A1,$7E,$2B
	.dc.b	$FD,$20,$47,$FD,$20,$D7,$FD,$20,$D7,$FD,$20,$D7,$FD,$20,$D7,$A1
	.dc.b	$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21
	.dc.b	$A2,$9A,$30,$43,$04,$3F,$D7,$0F,$B5,$A7,$98,$A7,$84,$A7,$62,$9D
	.dc.b	$4A,$75,$84,$75,$6C,$75,$6B,$D5,$AF,$56,$BD,$5A,$F5,$8E,$84,$7A
	.dc.b	$35,$E8,$D7,$A3,$5E,$8D,$7A,$1C,$88,$72,$29,$C8,$A7,$22,$9C,$8E
	.dc.b	$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$43,$E9,$92,$3E
	.dc.b	$4B,$63,$E0,$34,$F4,$94,$F2,$14,$F0,$94,$F0,$14,$F0,$47,$62,$47
	.dc.b	$66,$C7,$66,$BD,$9A,$F6,$6B,$D9,$AF,$68,$E8,$47,$A3,$5E,$8D,$7A
	.dc.b	$35,$E8,$D7,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E
	.dc.b	$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3F,$03,$F8,$A3,$E5,$7F,$D0
	.dc.b	$90,$24,$65,$C8,$70,$F0,$6B,$FE,$8D,$6B,$CC,$E1,$1C,$9A,$74,$8D
	.dc.b	$BE,$08,$D4,$62,$36,$77,$8C,$91,$E4,$D7,$46,$08,$60,$87,$9D,$E0
	.dc.b	$D3,$65,$36,$7A,$E3,$C8,$48,$E3,$19,$6B,$EB,$39,$99,$29,$93,$A9
	.dc.b	$4F,$01,$0E,$06,$8F,$86,$64,$E0,$68,$C1,$A3,$04,$36,$43,$45,$8F
	.dc.b	$F6,$01,$3A,$FE,$99,$23,$EC,$35,$FC,$13,$F7,$79,$FF,$D1,$B1,$FE
	.dc.b	$59,$E7,$FF,$31,$9F,$E6,$27,$FC,$BA,$7F,$9F,$CF,$F3,$59,$FE,$64
	.dc.b	$8E,$2E,$79,$7F,$CB,$EE,$79,$FF,$CD,$4A,$76,$7F,$A0,$5F,$F3,$19
	.dc.b	$B3,$BB,$93,$FC,$FE,$F8,$0E,$DC,$DF,$E6,$B8,$E7,$3F,$E8,$9A,$FF
	.dc.b	$3A,$56,$E7,$94,$F8,$9D,$C7,$B8,$FF,$55,$16,$7D,$2F,$90,$F4,$73
	.dc.b	$FF,$A3,$23,$A8,$BF,$F4,$8E,$1C,$5F,$F4,$EE,$1E,$36,$DE,$4F,$71
	.dc.b	$E5,$8D,$3F,$E9,$33,$D1,$7F,$E8,$AB,$EF,$5D,$A7,$CE,$6A,$FA,$3E
	.dc.b	$7C,$3D,$AE,$4D,$D7,$B2,$1C,$CE,$5F,$F4,$2D,$7F,$A4,$0F,$F4,$41
	.dc.b	$4F,$B1,$1B,$71,$3C,$A7,$F8,$7B,$63,$FD,$29,$18,$9E,$11,$FE,$98
	.dc.b	$E7,$FF,$4F,$1F,$E9,$C7,$BB,$57,$27,$FA,$5B,$C0,$FE,$3D,$CE,$E6
	.dc.b	$FF,$43,$73,$FF,$A4,$23,$17,$FC,$8E,$F8,$3F,$E9,$73,$FD,$28,$48
	.dc.b	$DB,$DB,$99,$87,$6C,$3F,$D2,$98,$75,$7F,$D2,$D3,$FE,$A7,$BE,$38
	.dc.b	$76,$6D,$6F,$0F,$15,$74,$7F,$D5,$33,$FE,$A7,$7F,$D4,$A6,$B0,$F0
	.dc.b	$3E,$9A,$FF,$52,$B9,$38,$E1,$FE,$AB,$71,$5E,$2B,$EC,$F5,$8D,$9A
	.dc.b	$9E,$1D,$6E,$BF,$83,$3F,$CF,$E3,$13,$FD,$21,$B1,$FD,$39,$AF,$F4
	.dc.b	$50,$E1,$E5,$3D,$1F,$EA,$91,$96,$C6,$57,$73,$A5,$F0,$DF,$D6,$23
	.dc.b	$0D,$8C,$CF,$F6,$09,$CB,$63,$2F,$96,$32,$D8,$CB,$FD,$76,$32,$D8
	.dc.b	$CB,$FD,$AE,$32,$D8,$CB,$E8,$6C,$6A,$31,$3E,$72,$3E,$56,$BE,$36
	.dc.b	$BE,$16,$BD,$DA,$F4,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70
	.dc.b	$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$91,$F1,$C9,$3F,$28,$93
	.dc.b	$F0,$25,$F8,$AF,$F1,$67,$FC,$36,$D7,$FC,$42,$D7,$F2,$8D,$7F,$C3
	.dc.b	$4D,$7F,$C4,$CD,$7A,$F0,$7F,$87,$63,$04,$70,$48,$F0,$B5,$F0,$B5
	.dc.b	$F0,$B5,$F0,$B5,$F0,$B5,$F0,$9D,$08,$74,$29,$D0,$A7,$42,$9D,$0A
	.dc.b	$74,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C,$08,$70,$29,$C0,$A7,$03
	.dc.b	$44,$34,$53,$46,$08,$60,$91,$F4,$DF,$6B,$5F,$33,$5F,$0B,$5E,$CD
	.dc.b	$7A,$B5,$EB,$1F,$4E,$30,$47,$04,$8F,$0B,$5F,$0B,$5F,$0B,$5F,$0B
	.dc.b	$5F,$0B,$5F,$09,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0E
	.dc.b	$45,$39,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34
	.dc.b	$60,$86,$09,$1F,$4D,$F6,$B5,$F3,$35,$F0,$B5,$EC,$D7,$AB,$5E,$B1
	.dc.b	$F4,$E3,$04,$70,$48,$F0,$B5,$F0,$B5,$F0,$B5,$F0,$B5,$F0,$B5,$F0
	.dc.b	$9D,$08,$74,$29,$D0,$A7,$42,$9D,$0A,$74,$3E,$D7,$E4,$AF,$0F,$FD
	.dc.b	$74,$D8,$FF,$62,$B5,$FF,$65,$B5,$FE,$FC,$D7,$FD,$88,$D7,$FD,$A0
	.dc.b	$D4,$D2,$47,$80,$91,$E0,$6B,$E0,$6B,$E0,$6B,$E0,$6B,$E0,$6B,$E0
	.dc.b	$6C,$78,$23,$B1,$0E,$C5,$3B,$14,$EC,$53,$B1,$4E,$C5,$7B,$3D,$08
	.dc.b	$74,$29,$D0,$A7,$42,$9D,$0A,$F4,$79,$10,$E4,$53,$FC,$63,$FA,$13
	.dc.b	$E2,$BF,$C7,$0D,$8F,$F1,$C3,$5F,$F1,$C6,$A8,$C4,$7F,$90,$5A,$FF
	.dc.b	$8E,$1A,$F6,$9E,$A7,$FB,$1F,$AB,$FD,$45,$D5,$FE,$9C,$9F,$F4,$77
	.dc.b	$4F,$FA,$37,$AF,$FD,$5D,$5A,$9F,$F4,$E4,$FF,$A4,$63,$11,$FE,$97
	.dc.b	$6F,$57,$FA,$FA,$7F,$D5,$93,$FE,$D3,$3F,$D3,$37,$FE,$BF,$AD,$DF
	.dc.b	$6E,$BF,$F6,$B5,$7F,$AA,$CF,$F5,$1E,$1F,$EA,$2C,$3C,$13,$A7,$AC
	.dc.b	$FF,$B2,$8F,$F6,$D5,$F4,$7F,$DA,$37,$FE,$B4,$7C,$D5,$C7,$0F,$F5
	.dc.b	$95,$F2,$C3,$FD,$41,$D5,$FE,$9C,$F0,$58,$7F,$A7,$1A,$FF,$AA,$BB
	.dc.b	$3F,$D4,$84,$7F,$D4,$E4,$9E,$4F,$13,$FC,$F4,$43,$FD,$31,$D7,$FE
	.dc.b	$88,$7F,$D4,$FD,$7F,$E9,$17,$27,$FA,$08,$BD,$5F,$E9,$67,$FD,$22
	.dc.b	$63,$AB,$FD,$56,$E3,$B1,$D9,$FE,$9F,$9F,$F5,$0E,$1D,$E5,$87,$77
	.dc.b	$FD,$45,$3C,$AF,$FD,$5C,$F0,$7F,$D6,$58,$7F,$A5,$EF,$96,$1F,$E9
	.dc.b	$8A,$FF,$4E,$57,$8C,$F7,$5F,$8A,$BC,$18,$77,$BE,$15,$A7,$CC,$D9
	.dc.b	$EA,$70,$69,$F0,$EB,$A3,$E4,$BF,$F4,$AF,$8F,$FE,$A1,$ED,$34,$93
	.dc.b	$FE,$A3,$8C,$B6,$32,$FE,$BA,$32,$FF,$A9,$CF,$F4,$F7,$BD,$5F,$C2
	.dc.b	$F6,$A3,$11,$A3,$51,$96,$C6,$A3,$11,$D5,$AF,$FA,$D5,$AF,$E4,$1A
	.dc.b	$F8,$23,$91,$1E,$A4,$3A,$39,$69,$93,$B3,$96,$99,$3C,$0E,$5A,$64
	.dc.b	$F1,$14,$F8,$07,$62,$1D,$0A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38
	.dc.b	$14,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$3E,$7B,$8E,$C7,$B7,$FE,$8F
	.dc.b	$24,$7F,$A7,$1A,$FF,$A7,$1A,$FF,$A7,$1A,$FF,$A7,$1A,$FF,$A7,$1A
	.dc.b	$FF,$AA,$1B,$1E,$08,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$53,$B1,$5E
	.dc.b	$CF,$42,$1D,$0A,$74,$29,$D0,$A7,$42,$BD,$1E,$44,$39,$14,$E4,$53
	.dc.b	$91,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$24
	.dc.b	$7D,$42,$47,$C9,$6B,$F0,$1A,$FA,$5A,$F9,$1A,$F8,$5A,$F8,$1B,$1E
	.dc.b	$08,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$53,$B1,$5E,$CF,$42,$1D,$0A
	.dc.b	$74,$29,$D0,$A7,$42,$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02
	.dc.b	$1C,$0A,$70,$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$24,$7D,$42,$47,$C9
	.dc.b	$6B,$F0,$1A,$FA,$5A,$F9,$1A,$F8,$5A,$F8,$1B,$1E,$08,$EC,$43,$B1
	.dc.b	$4E,$C5,$3B,$14,$EC,$53,$B1,$5E,$CF,$42,$1D,$0A,$74,$3E,$FF,$CD
	.dc.b	$BE,$4F,$FB,$0B,$63,$FB,$33,$5F,$ED,$4D,$7F,$C5,$61,$C3,$FE,$2E
	.dc.b	$6B,$FE,$32,$6B,$B8,$EE,$47,$BB,$5E,$ED,$7B,$C7,$22,$46,$63,$C0
	.dc.b	$D7,$C0,$D7,$C0,$76,$21,$D8,$A7,$67,$84,$76,$8C,$9D,$8A,$76,$29
	.dc.b	$D8,$E8,$43,$A1,$A7,$A3,$93,$A1,$4E,$85,$3A,$1C,$8C,$3C,$9C,$10
	.dc.b	$E6,$7F,$98,$3B,$19,$3E,$8F,$FC,$5C,$49,$FF,$16,$35,$FF,$15,$35
	.dc.b	$E5,$1C,$5F,$F1,$57,$37,$F9,$F7,$A3,$FD,$09,$D3,$FE,$75,$9F,$F4
	.dc.b	$47,$4F,$FA,$4D,$FF,$3D,$9F,$E8,$4E,$BF,$F3,$F3,$FE,$87,$3F,$D3
	.dc.b	$35,$FE,$76,$BE,$91,$FE,$84,$BE,$8F,$FA,$1C,$42,$93,$FE,$90,$EB
	.dc.b	$FF,$4A,$57,$5C,$3F,$D0,$D1,$FE,$8B,$BF,$F4,$F6,$1F,$E8,$A3,$C1
	.dc.b	$87,$92,$B7,$7E,$7C,$3C,$37,$C2,$3C,$87,$5B,$E4,$FF,$A4,$23,$D2
	.dc.b	$F9,$AF,$CA,$FF,$AA,$8E,$37,$D9,$E4,$62,$BF,$D7,$91,$FE,$A8,$8C
	.dc.b	$9F,$06,$BE,$05,$FB,$E3,$E6,$9F,$28,$F1,$5F,$AE,$FC,$B1,$E3,$9F
	.dc.b	$5E,$1F,$0C,$FF,$52,$E1,$E3,$3F,$D5,$97,$E8,$7F,$D5,$DE,$03,$AF
	.dc.b	$FD,$5B,$D7,$FE,$A6,$C7,$FD,$58,$17,$67,$DB,$AE,$13,$FE,$AE,$3F
	.dc.b	$84,$E8,$8C,$C6,$1A,$F1,$BE,$F5,$63,$B3,$6F,$BC,$7F,$AD,$CF,$F4
	.dc.b	$D1,$3A,$3F,$D1,$FD,$9F,$EB,$18,$FF,$52,$9F,$E9,$73,$FD,$29,$D9
	.dc.b	$FE,$9D,$EB,$FF,$4B,$CF,$FA,$9E,$7F,$D2,$C6,$DC,$1C,$CB,$D7,$FE
	.dc.b	$CE,$BE,$EF,$FA,$B1,$FF,$5A,$CF,$FB,$51,$FF,$5C,$BF,$EB,$FE,$CF
	.dc.b	$F5,$4B,$E7,$7F,$D5,$AF,$A7,$0F,$23,$FE,$A9,$8D,$CF,$48,$FF,$68
	.dc.b	$9C,$F0,$FF,$58,$3E,$E3,$A3,$EC,$BF,$F6,$36,$07,$FD,$2D,$7F,$E9
	.dc.b	$D2,$57,$FA,$89,$AF,$FA,$7D,$AF,$FA,$7D,$FF,$52,$78,$D2,$3F,$D5
	.dc.b	$2D,$7F,$D4,$CD,$7F,$D4,$CD,$7F,$D4,$CD,$7F,$D6,$AD,$5D,$68,$C3
	.dc.b	$63,$B9,$0E,$E5,$3B,$94,$EE,$53,$B9,$4E,$F1,$D8,$8E,$63,$C0,$F6
	.dc.b	$21,$D8,$A7,$62,$9D,$8A,$76,$3A,$47,$67,$31,$D9,$E8,$43,$A1,$4E
	.dc.b	$85,$3A,$1C,$A3,$FD,$BB,$E2,$BC,$6A,$49,$FF,$72,$46,$5B,$19,$7F
	.dc.b	$DD,$51,$96,$C6,$5F,$F0,$1C,$65,$B1,$94,$48,$D3,$80,$E1,$D4,$71
	.dc.b	$69,$E4,$21,$E2,$29,$E0,$29,$D6,$3A,$3D,$A3,$A3,$4D,$9D,$8E,$84
	.dc.b	$3A,$14,$E8,$F2,$8E,$A4,$70,$48,$EA,$77,$39,$1D,$CE,$A4,$3A,$9C
	.dc.b	$23,$A3,$63,$A3,$C8,$E8,$71,$3A,$1C,$8D,$47,$26,$C7,$27,$81,$C8
	.dc.b	$E0,$62,$38,$36,$38,$3A,$21,$E0,$21,$CC,$B1,$C9,$D9,$92,$39,$77
	.dc.b	$1D,$9A,$7C,$12,$1E,$A2,$9E,$52,$9D,$A3,$A3,$DA,$3A,$34,$D9,$D8
	.dc.b	$E8,$43,$A1,$4E,$8F,$28,$EA,$47,$04,$8E,$A7,$73,$91,$DC,$EA,$43
	.dc.b	$A9,$C2,$3A,$36,$3A,$3C,$8E,$87,$13,$A1,$C8,$D4,$72,$6C,$72,$78
	.dc.b	$1C,$8E,$06,$23,$83,$63,$83,$A2,$1E,$02,$1C,$CB,$1C,$9D,$99,$23
	.dc.b	$97,$71,$D9,$A7,$C1,$21,$EA,$29,$E5,$29,$DA,$3A,$3D,$A3,$A3,$4D
	.dc.b	$9D,$8E,$84,$3A,$14,$E8,$FE,$07,$C6,$FE,$5A,$48,$FE,$56,$D7,$F9
	.dc.b	$4B,$5F,$E4,$ED,$7F,$8B,$35,$FE,$4C,$D7,$F9,$2B,$64,$D2,$47,$80
	.dc.b	$87,$80,$A7,$80,$A7,$80,$A7,$80,$A7,$80,$B1,$E0,$6C,$78,$1E,$C4
	.dc.b	$3B,$14,$EC,$53,$B1,$4E,$C5,$8E,$CD,$8E,$CF,$42,$1D,$0A,$7F,$4D
	.dc.b	$F5,$BF,$AF,$91,$FE,$BA,$D7,$FA,$DB,$5F,$EB,$2F,$FB,$71,$FF,$41
	.dc.b	$F3,$FF,$99,$9F,$F4,$11,$FE,$72,$E4,$FF,$41,$3F,$E6,$B3,$FC,$F8
	.dc.b	$5E,$5F,$F5,$03,$BE,$6F,$F3,$57,$47,$48,$E5,$D1,$FE,$71,$7F,$CF
	.dc.b	$35,$FE,$7F,$3F,$CE,$53,$FE,$78,$BF,$F4,$59,$FE,$80,$9E,$ED,$9E
	.dc.b	$E7,$08,$FF,$45,$1F,$E9,$3A,$FF,$51,$C6,$63,$FD,$15,$7E,$47,$FD
	.dc.b	$0F,$87,$BD,$FF,$45,$1E,$BC,$3F,$D2,$67,$FA,$F9,$F4,$9F,$12,$BA
	.dc.b	$61,$F1,$0F,$24,$FB,$E3,$38,$7B,$EB,$E2,$5F,$FA,$65,$A7,$B3,$0F
	.dc.b	$23,$E1,$3D,$47,$C6,$BF,$15,$F4,$8F,$14,$79,$9F,$82,$7F,$A9,$EF
	.dc.b	$C1,$3E,$6A,$EB,$87,$CC,$69,$FE,$A7,$AF,$F6,$CD,$FF,$A3,$79,$A4
	.dc.b	$7F,$A3,$6B,$FD,$31,$CF,$FE,$94,$8F,$F6,$07,$47,$FA,$61,$B1,$FE
	.dc.b	$E6,$8C,$47,$FA,$98,$FF,$4F,$47,$17,$FD,$53,$D1,$FE,$AB,$E7,$FF
	.dc.b	$3E,$1F,$E8,$68,$FF,$3D,$38,$E8,$FF,$41,$F3,$FF,$A4,$27,$FD,$19
	.dc.b	$CD,$FE,$84,$3F,$D1,$35,$FE,$92,$8C,$F4,$7F,$A1,$2B,$84,$7F,$A1
	.dc.b	$0F,$F4,$E9,$5C,$CF,$85,$B5,$D0,$ED,$5F,$E8,$BA,$EA,$72,$AF,$33
	.dc.b	$FE,$93,$31,$3C,$67,$84,$7F,$A6,$0F,$6C,$65,$FF,$59,$9C,$6B,$FD
	.dc.b	$5D,$D1,$FE,$AD,$9D,$47,$9B,$0F,$F2,$53,$E6,$BE,$B8,$7A,$6F,$84
	.dc.b	$62,$FE,$2D,$F0,$9F,$F2,$AF,$5B,$A9,$F3,$3F,$C0,$F3,$4E,$63,$0D
	.dc.b	$8F,$E1,$0D,$7F,$CB,$87,$09,$E6,$E9,$AE,$A3,$83,$FE,$63,$6B,$FE
	.dc.b	$65,$EF,$FF,$5B,$3E,$36,$BF,$C5,$DA,$F8,$1A,$F1,$BF,$52,$8C,$BF
	.dc.b	$E7,$46,$B9,$8D,$C6,$08,$E0,$91,$E3,$6B,$E3,$6B,$E3,$6B,$E3,$38
	.dc.b	$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$80,$F9,$1A
	.dc.b	$77,$29,$D0,$A7,$32,$9C,$CF,$90,$71,$29,$C4,$A7,$12,$9C,$63,$E6
	.dc.b	$BF,$EB,$2F,$85,$FF,$5B,$91,$C1,$0F,$F5,$9B,$96,$99,$3F,$D6,$6E
	.dc.b	$5A,$64,$FF,$52,$39,$69,$93,$FD,$6C,$E5,$A6,$4F,$F5,$BB,$96,$99
	.dc.b	$EA,$46,$5B,$19,$8C,$11,$C1,$0F,$59,$0F,$49,$4F,$39,$4F,$29,$4F
	.dc.b	$19,$63,$C2,$D8,$F0,$34,$F0,$1D,$88,$76,$29,$D8,$A7,$62,$C7,$66
	.dc.b	$C7,$66,$9D,$8E,$84,$3A,$14,$E8,$58,$E8,$D8,$E8,$D3,$A1,$C8,$87
	.dc.b	$22,$C7,$26,$C7,$26,$9C,$8E,$04,$8E,$0D,$8E,$0D,$38,$46,$89,$1A
	.dc.b	$6C,$6A,$30,$47,$04,$3E,$A1,$0F,$92,$53,$E0,$14,$F4,$94,$F2,$16
	.dc.b	$3C,$2D,$8F,$03,$4F,$01,$D8,$87,$62,$9D,$8A,$76,$2C,$76,$6C,$76
	.dc.b	$69,$D8,$E8,$43,$A1,$4E,$85,$8E,$8D,$8E,$8D,$3A,$1C,$88,$72,$2C
	.dc.b	$72,$6C,$72,$69,$C8,$E0,$48,$E0,$D8,$E0,$D3,$84,$68,$91,$A6,$C6
	.dc.b	$A3,$04,$70,$43,$FC,$E5,$DE,$F9,$5F,$EB,$2D,$8F,$EE,$8D,$77,$1A
	.dc.b	$23,$A6,$BA,$30,$43,$04,$3F,$BC,$C7,$56,$C7,$26,$9C,$4A,$71,$04
	.dc.b	$8E,$4E,$03,$87,$51,$D5,$D4,$65,$DC,$65,$D9,$E1,$36,$E5,$D9,$93
	.dc.b	$67,$94,$A7,$A0,$87,$23,$81,$C8,$E0,$53,$67,$23,$81,$A8,$F4,$12
	.dc.b	$3C,$8D,$8E,$AD,$7F,$C4,$59,$3A,$7F,$C5,$AD,$8F,$F1,$44,$60,$8E
	.dc.b	$08,$72,$7F,$C3,$FC,$BF,$E7,$DE,$AF,$F4,$77,$57,$FA,$05,$BD,$3F
	.dc.b	$EA,$DE,$AF,$F4,$06,$1F,$EA,$49,$CB,$FE,$82,$3F,$D4,$A7,$FA,$80
	.dc.b	$FF,$52,$DF,$57,$FD,$31,$1C,$1E,$46,$09,$7F,$EB,$6B,$F0,$4F,$FA
	.dc.b	$BE,$3F,$D8,$11,$FE,$99,$9F,$F5,$01,$2B,$FD,$59,$87,$FA,$6A,$78
	.dc.b	$CF,$8E,$B8,$BF,$EA,$63,$FD,$56,$F7,$7D,$47,$39,$FF,$61,$CF,$7A
	.dc.b	$FF,$5B,$CF,$BA,$BF,$D5,$8F,$BB,$0F,$8B,$3F,$EB,$D7,$FD,$4F,$87
	.dc.b	$A6,$FF,$D6,$67,$38,$CD,$79,$63,$FD,$84,$7F,$AA,$9F,$F4,$FF,$55
	.dc.b	$BF,$F5,$5C,$3A,$13,$FA,$79,$CC,$FF,$AA,$0F,$F7,$B3,$87,$5D,$5F
	.dc.b	$EA,$E8,$E7,$5C,$A7,$FD,$71,$2E,$AF,$F5,$85,$6A,$FF,$D0,$B4,$EA
	.dc.b	$FF,$45,$3F,$E9,$8E,$9F,$F4,$41,$FE,$A9,$E5,$DC,$E8,$FF,$4A,$CF
	.dc.b	$FA,$89,$C4,$7E,$FF,$AB,$FD,$25,$39,$8E,$24,$7F,$83,$C6,$9F,$F5
	.dc.b	$94,$65,$FF,$53,$4F,$47,$91,$C3,$A7,$FD,$87,$3B,$AF,$F6,$04,$F0
	.dc.b	$8C,$75,$7F,$AA,$1F,$25,$70,$9E,$62,$B8,$9E,$33,$FD,$76,$FF,$AD
	.dc.b	$B0,$F8,$6F,$C5,$3E,$45,$76,$8F,$8D,$1D,$27,$9D,$FF,$B2,$23,$66
	.dc.b	$5C,$61,$ED,$8F,$41,$9A,$FF,$5D,$E1,$FE,$B7,$8C,$1F,$EC,$63,$99
	.dc.b	$F2,$0F,$03,$B7,$07,$B6,$BF,$D8,$2F,$F3,$6A,$FF,$56,$F4,$FF,$B0
	.dc.b	$78,$3F,$88,$92,$7F,$D7,$0D,$7F,$8B,$B5,$FF,$5C,$72,$23,$FD,$91
	.dc.b	$E5,$3E,$6D,$EE,$61,$B7,$FE,$C6,$13,$B9,$D1,$23,$51,$96,$CF,$27
	.dc.b	$FD,$9F,$19,$6C,$65,$FE,$A1,$19,$6C,$65,$FE,$AF,$19,$6C,$65,$F2
	.dc.b	$B6,$35,$18,$9F,$33,$E0,$6B,$D9,$AF,$46,$BC,$4F,$5B,$88,$EA,$49
	.dc.b	$F0,$12,$34,$D9,$E2,$43,$C4,$53,$C4,$53,$C4,$53,$C4,$70,$21,$C0
	.dc.b	$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$3F,$C0,$5E,$1F,$F1,$F7,$F0,$41
	.dc.b	$C3,$F8,$56,$BF,$E0,$93,$5C,$9F,$EF,$C8,$CC,$74,$8C,$11,$EB,$1A
	.dc.b	$7F,$C2,$AD,$79,$9A,$8E,$0F,$85,$AF,$86,$38,$12,$33,$1E,$13,$91
	.dc.b	$0E,$45,$39,$3C,$1F,$1B,$5F,$19,$C0,$87,$03,$91,$DC,$C9,$4D,$18
	.dc.b	$3A,$9C,$0A,$64,$EE,$70,$21,$CC,$87,$03,$25,$32,$72,$2C,$7C,$57
	.dc.b	$C0,$64,$E0,$43,$04,$78,$3E,$56,$BE,$03,$51,$F1,$E3,$31,$E0,$8C
	.dc.b	$11,$F0,$C6,$9F,$03,$5E,$66,$A3,$83,$E1,$6B,$E1,$8E,$04,$8C,$C7
	.dc.b	$84,$E4,$43,$91,$4E,$4F,$07,$C6,$D7,$C6,$70,$21,$C0,$E4,$77,$32
	.dc.b	$53,$46,$0E,$A7,$02,$99,$3B,$9C,$08,$73,$21,$C0,$C9,$4C,$9C,$8B
	.dc.b	$1F,$15,$F0,$19,$38,$10,$C1,$1E,$0F,$95,$AF,$80,$D4,$7C,$78,$CC
	.dc.b	$78,$23,$04,$7C,$31,$A7,$C0,$D7,$99,$A8,$E0,$F8,$5A,$F8,$63,$81
	.dc.b	$23,$31,$E1,$39,$10,$E4,$53,$93,$C1,$F1,$B5,$F1,$9C,$08,$70,$39
	.dc.b	$1D,$CF,$B5,$D2,$F9,$6F,$F6,$39,$23,$FD,$8E,$D7,$FD,$9A,$D7,$FD
	.dc.b	$7E,$D7,$FD,$62,$D7,$FD,$74,$D7,$FD,$72,$D8,$F0,$47,$62,$1D,$8A
	.dc.b	$76,$29,$D8,$A7,$62,$9D,$8A,$F6,$7A,$10,$E8,$53,$A1,$4E,$85,$3A
	.dc.b	$15,$E8,$F2,$21,$C8,$A7,$22,$9C,$8A,$F2,$78,$10,$E0,$53,$81,$5E
	.dc.b	$0B,$A1,$FB,$C3,$C0,$FE,$CB,$18,$6C,$66,$7F,$C5,$53,$96,$C6,$5F
	.dc.b	$ED,$11,$96,$C6,$5F,$F7,$A4,$65,$FE,$DD,$CF,$FE,$BD,$EB,$FF,$55
	.dc.b	$47,$FA,$F1,$B1,$FE,$A3,$9F,$F6,$03,$67,$FD,$7F,$D7,$FE,$A0,$7F
	.dc.b	$D6,$95,$FE,$D1,$72,$7F,$AB,$E3,$FD,$51,$5C,$BB,$3F,$D7,$D3,$FE
	.dc.b	$B8,$BF,$15,$73,$84,$7F,$B1,$11,$FE,$B7,$AF,$F5,$A3,$D4,$87,$0A
	.dc.b	$FF,$60,$BE,$4A,$FF,$65,$DF,$FB,$00,$FF,$65,$C6,$CE,$AF,$A4,$F8
	.dc.b	$33,$E6,$8F,$89,$3E,$5B,$F2,$47,$8C,$FF,$61,$1E,$27,$D6,$D9,$F1
	.dc.b	$BD,$CF,$F6,$CC,$F3,$8C,$CF,$4A,$F2,$B6,$7F,$DA,$35,$EB,$3E,$47
	.dc.b	$0F,$FB,$62,$FE,$2C,$7F,$BE,$3B,$3F,$D8,$4F,$FB,$3F,$AF,$FD,$A5
	.dc.b	$18,$DF,$FD,$91,$3F,$ED,$63,$E6,$E3,$FE,$CF,$6C,$FF,$B3,$C9,$D9
	.dc.b	$FE,$3A,$9F,$F6,$64,$FF,$B4,$23,$FD,$9F,$3D,$6F,$FD,$4D,$3F,$E9
	.dc.b	$9E,$7F,$F5,$07,$5F,$FA,$C3,$AB,$FD,$69,$3F,$E9,$88,$FF,$59,$75
	.dc.b	$7F,$AE,$63,$FD,$3A,$7F,$AC,$A9,$D9,$FE,$C1,$BE,$37,$FE,$B0,$EC
	.dc.b	$FF,$59,$DF,$3C,$3F,$D5,$F7,$E2,$C3,$FD,$B3,$39,$AD,$DF,$FB,$02
	.dc.b	$7C,$B7,$C2,$BF,$D6,$31,$C2,$75,$7C,$CF,$14,$F9,$5A,$FF,$B2,$4F
	.dc.b	$F5,$BC,$79,$5F,$F6,$54,$FC,$2C,$3F,$D7,$13,$D2,$7B,$D7,$2A,$F1
	.dc.b	$E1,$EC,$EF,$7C,$AF,$F9,$AA,$D1,$87,$FC,$DA,$D7,$FD,$60,$D7,$FC
	.dc.b	$D8,$D7,$F0,$ED,$7F,$5A,$D7,$AC,$7F,$9C,$7C,$B9,$1F,$E6,$E6,$BF
	.dc.b	$E7,$96,$BF,$E6,$96,$BF,$E8,$26,$BF,$E7,$46,$BF,$E8,$36,$F5,$1A
	.dc.b	$48,$F0,$10,$F0,$14,$F0,$14,$F0,$14,$F0,$14,$F0,$16,$3C,$0D,$8F
	.dc.b	$03,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$B1,$D9,$B1,$D9,$E8,$43,$A1
	.dc.b	$4E,$87,$FA,$57,$DD,$FF,$52,$11,$FF,$50,$B5,$FF,$56,$35,$E0,$68
	.dc.b	$86,$8A,$68,$C1,$0C,$10,$FF,$5E,$BD,$5A,$72,$29,$C4,$A7,$13,$FD
	.dc.b	$6A,$F1,$69,$C4,$A7,$12,$9C,$4F,$F6,$53,$C5,$A7,$12,$9C,$4A,$71
	.dc.b	$94,$71,$6C,$71,$6B,$C5,$AF,$18,$E0,$47,$83,$5E,$0D,$78,$1A,$21
	.dc.b	$A2,$9A,$30,$43,$04,$3E,$59,$0F,$8A,$53,$DE,$58,$F4,$B6,$3C,$8D
	.dc.b	$3C,$25,$3C,$05,$3C,$07,$62,$1D,$8B,$1D,$9B,$1D,$9A,$76,$29,$D8
	.dc.b	$A7,$63,$A1,$23,$A3,$63,$A3,$4E,$85,$3A,$14,$E9,$1C,$89,$1C,$9B
	.dc.b	$1C,$9A,$F2,$6B,$CA,$38,$11,$E0,$D7,$83,$5E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$C1,$0F,$A8,$43,$E4,$94,$F8,$05,$8F,$4B,$63,$C8,$D3,$C2,$53
	.dc.b	$C0,$53,$C0,$76,$21,$D8,$B1,$D9,$B1,$D9,$A7,$62,$9D,$8A,$76,$3A
	.dc.b	$12,$3A,$36,$3A,$3F,$86,$FC,$61,$E8,$7F,$65,$6C,$7F,$89,$0F,$F1
	.dc.b	$2B,$87,$01,$C3,$B6,$BC,$23,$A1,$1E,$27,$17,$0F,$48,$E2,$FF,$96
	.dc.b	$1A,$EC,$E2,$E0,$91,$CA,$3C,$4D,$7C,$47,$07,$0F,$56,$BD,$4F,$13
	.dc.b	$88,$EB,$1C,$48,$68,$A7,$80,$A7,$80,$D1,$0D,$14,$D1,$82,$18,$21
	.dc.b	$E7,$3A,$94,$E4,$53,$89,$4E,$27,$B4,$C9,$4F,$F3,$0F,$E3,$0F,$92
	.dc.b	$FF,$13,$36,$3E,$BB,$5F,$C0,$B5,$FC,$33,$5E,$71,$C8,$8F,$F9,$87
	.dc.b	$A3,$FC,$E0,$FF,$A2,$3A,$7F,$CE,$AF,$F9,$F7,$AB,$FD,$15,$1F,$E7
	.dc.b	$D8,$D1,$FE,$78,$EB,$FF,$4C,$BF,$E9,$EA,$FF,$62,$3F,$EA,$83,$11
	.dc.b	$AE,$94,$EA,$FF,$D2,$B5,$D0,$D9,$C6,$7F,$D2,$2E,$23,$FD,$41,$7E
	.dc.b	$69,$FF,$46,$4F,$FA,$5E,$BD,$6F,$FA,$6E,$B6,$E6,$FC,$53,$E1,$BF
	.dc.b	$64,$FF,$A6,$8F,$F4,$F1,$E3,$BF,$F6,$35,$F9,$A3,$A4,$FF,$B0,$9F
	.dc.b	$35,$78,$27,$FD,$3F,$D7,$FE,$A2,$24,$F3,$7F,$D3,$B0,$AF,$F5,$97
	.dc.b	$4F,$FA,$3E,$BF,$D2,$47,$FA,$7C,$A7,$FB,$79,$FF,$4C,$BF,$C0,$30
	.dc.b	$FF,$4F,$3F,$EA,$12,$3F,$E9,$C7,$1C,$DF,$E7,$57,$FD,$10,$7F,$9E
	.dc.b	$B9,$BF,$D0,$4F,$FB,$52,$3F,$D3,$3C,$BF,$E8,$38,$C7,$3F,$FA,$5B
	.dc.b	$A3,$FD,$39,$0A,$DB,$FE,$86,$8C,$C6,$3A,$3F,$D0,$D3,$CC,$E1,$3F
	.dc.b	$E9,$57,$FD,$51,$5D,$6F,$FD,$CB,$5F,$EA,$69,$FF,$54,$9F,$E9,$37
	.dc.b	$9B,$70,$F4,$BE,$3A,$FF,$49,$DF,$88,$ED,$87,$FA,$66,$FF,$D4,$51
	.dc.b	$9C,$3F,$D5,$2D,$BF,$F4,$FD,$FF,$B4,$AB,$C5,$E2,$3C,$46,$9F,$CA
	.dc.b	$23,$0D,$8C,$DF,$F3,$49,$CB,$63,$2F,$F3,$B8,$CB,$63,$2F,$FB,$8A
	.dc.b	$32,$FF,$B9,$3C,$47,$DF,$38,$3F,$9C,$C6,$1B,$19,$9F,$E8,$B3,$96
	.dc.b	$C6,$5F,$E9,$91,$96,$C6,$5F,$03,$50,$8E,$0E,$A7,$C2,$47,$BB,$5E
	.dc.b	$AD,$78,$1A,$21,$A2,$9A,$30,$43,$04,$3C,$B1,$E4,$6C,$FA,$5F,$43
	.dc.b	$63,$51,$89,$F2,$3C,$1A,$E4,$F1,$99,$29,$93,$C8,$64,$A6,$4F,$31
	.dc.b	$4D,$18,$8F,$19,$27,$CC,$49,$F1,$C6,$5F,$2C,$FF,$80,$7A,$7F,$C1
	.dc.b	$04,$7F,$BA,$B5,$FE,$F8,$D7,$FC,$0C,$D7,$FC,$1A,$D7,$FC,$00,$D7
	.dc.b	$B1,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0E,$45,$39,$14
	.dc.b	$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$09
	.dc.b	$CE,$8F,$90,$D8,$F7,$B5,$F4,$35,$F1,$B5,$F0,$35,$EE,$D7,$BC,$76
	.dc.b	$23,$D9,$AF,$66,$BD,$9A,$F6,$6B,$D9,$AF,$63,$A1,$0E,$85,$3A,$14
	.dc.b	$E8,$53,$A1,$4E,$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05
	.dc.b	$38,$14,$E0,$68,$86,$8A,$68,$C1,$0C,$12,$3E,$A1,$23,$E4,$B6,$3E
	.dc.b	$03,$5F,$4B,$5F,$23,$5F,$0B,$5F,$03,$5F,$04,$76,$23,$D9,$AF,$66
	.dc.b	$BD,$9A,$F6,$6B,$D9,$AF,$63,$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$4E
	.dc.b	$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81,$FB,$CF,$0B,$FC,$C4,$FF
	.dc.b	$58,$6A,$69,$39,$FF,$D4,$71,$96,$C6,$63,$FD,$4E,$D7,$FD,$62,$D7
	.dc.b	$FD,$6C,$D7,$FB,$39,$CC,$E3,$1D,$08,$E0,$87,$47,$2D,$32,$76,$72
	.dc.b	$D3,$27,$82,$3C,$4F,$C1,$8D,$B9,$6B,$93,$89,$4F,$01,$0E,$C5,$39
	.dc.b	$1C,$0F,$71,$93,$91,$C0,$87,$03,$47,$03,$27,$03,$46,$0D,$1F,$E1
	.dc.b	$2E,$77,$83,$FE,$8C,$8F,$F7,$1C,$63,$A3,$FD,$BA,$D8,$CC,$ED,$FF
	.dc.b	$76,$46,$5B,$19,$7E,$CC,$65,$B1,$97,$FD,$13,$97,$FA,$99,$FF,$44
	.dc.b	$9F,$E8,$B3,$FD,$31,$CF,$FE,$87,$8F,$F4,$51,$FE,$86,$8F,$F4,$21
	.dc.b	$FE,$8F,$BF,$F4,$A3,$CA,$3A,$3C,$8F,$F1,$8F,$3F,$FA,$81,$C1,$1C
	.dc.b	$1F,$EA,$EC,$3A,$1F,$EA,$02,$C7,$F9,$26,$FC,$87,$A8,$FC,$BC,$FF
	.dc.b	$A8,$1C,$47,$FA,$FC,$E0,$78,$BA,$3F,$D3,$AF,$BE,$7F,$D6,$E4,$C3
	.dc.b	$FD,$49,$5F,$EB,$7A,$DE,$1E,$CA,$D6,$1E,$5C,$3F,$D5,$6D,$9F,$8F
	.dc.b	$7C,$67,$E3,$9F,$EB,$23,$C4,$62,$3E,$63,$E3,$C3,$E5,$BF,$ED,$B9
	.dc.b	$F6,$E1,$FE,$B2,$7A,$47,$A7,$8B,$FD,$64,$FF,$AA,$9B,$8F,$FA,$AC
	.dc.b	$FF,$57,$91,$FF,$59,$C7,$FA,$D7,$C0,$FF,$32,$F8,$2F,$07,$E5,$D7
	.dc.b	$22,$4E,$A3,$FD,$33,$D4,$E9,$FF,$4B,$9F,$EA,$AE,$9F,$F5,$26,$3F
	.dc.b	$EA,$DE,$AF,$F4,$B9,$FE,$AB,$AF,$F5,$77,$5F,$2A,$DF,$5F,$FA,$A6
	.dc.b	$FF,$D5,$F7,$C1,$E4,$48,$FF,$51,$C7,$39,$E3,$5F,$C6,$2F,$C9,$3C
	.dc.b	$6B,$FD,$62,$75,$8F,$F6,$61,$30,$FF,$55,$47,$4C,$3B,$5F,$39,$F6
	.dc.b	$DF,$FA,$FA,$30,$F5,$7D,$53,$ED,$9D,$BC,$8F,$F6,$99,$99,$E0,$FF
	.dc.b	$B6,$63,$E4,$3D,$47,$57,$FA,$E7,$DC,$FF,$5B,$12,$3F,$D6,$8D,$7F
	.dc.b	$D6,$4D,$7F,$D6,$5D,$88,$FE,$8B,$D2,$F8,$7A,$4A,$FE,$7E,$7F,$40
	.dc.b	$8D,$38,$3F,$D5,$2E,$1C,$35,$E2,$D7,$94,$75,$76,$D7,$6D,$76,$77
	.dc.b	$3F,$CA,$4E,$5F,$0B,$5C,$9B,$8F,$03,$A3,$11,$B8,$CC,$70,$3C,$66
	.dc.b	$88,$60,$87,$98,$C9,$4C,$9E,$82,$9E,$12,$9D,$CE,$04,$38,$14,$E0
	.dc.b	$53,$81,$A2,$1A,$29,$A3,$04,$30,$47,$E0,$F5,$FF,$41,$8F,$F7,$1B
	.dc.b	$53,$49,$CD,$FE,$0C,$6C,$7D,$56,$BF,$85,$6B,$F8,$A6,$BF,$E0,$A6
	.dc.b	$BE,$08,$E8,$47,$A3,$5E,$8D,$7A,$35,$E8,$D7,$A1,$C8,$87,$22,$9C
	.dc.b	$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43
	.dc.b	$04,$3E,$9B,$ED,$69,$E6,$29,$E1,$29,$D8,$A7,$52,$9D,$63,$E9,$B8
	.dc.b	$23,$82,$47,$85,$B1,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$8E,$84
	.dc.b	$7A,$35,$E8,$D7,$A3,$5E,$8D,$7A,$1C,$88,$72,$29,$C8,$A7,$22,$9C
	.dc.b	$8E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$43,$E9,$BE
	.dc.b	$D6,$9E,$62,$9E,$12,$9D,$8A,$75,$29,$D6,$3E,$9B,$82,$38,$24,$78
	.dc.b	$5B,$1E,$16,$BE,$16,$BE,$16,$BE,$16,$BE,$18,$E8,$47,$A3,$5E,$8D
	.dc.b	$7A,$35,$E8,$D7,$A1,$C8,$87,$22,$9C,$8A,$72,$29,$FB,$1E,$B7,$CB
	.dc.b	$7F,$AA,$89,$1F,$EA,$A6,$BF,$EA,$A6,$BF,$EA,$A6,$BF,$E9,$B6,$BF
	.dc.b	$EB,$B6,$BF,$EB,$B6,$C7,$82,$3B,$10,$EC,$53,$B1,$4E,$C5,$3B,$14
	.dc.b	$EC,$57,$B3,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$AF,$47,$91,$0E,$45
	.dc.b	$39,$14,$E4,$57,$93,$C0,$87,$02,$9C,$0A,$F0,$7F,$C5,$39,$0F,$F0
	.dc.b	$66,$75,$A6,$63,$FC,$3D,$19,$6C,$65,$FF,$12,$C6,$5B,$19,$7F,$C5
	.dc.b	$EF,$FB,$CB,$9B,$FD,$0B,$1F,$E9,$03,$FC,$F4,$7F,$A4,$FA,$5D,$1F
	.dc.b	$E9,$A8,$FF,$54,$3F,$E7,$E7,$FD,$3F,$1F,$E8,$A7,$FD,$53,$97,$FA
	.dc.b	$12,$BA,$CF,$29,$FF,$45,$9F,$E8,$9E,$AF,$F4,$95,$FF,$A8,$A3,$C2
	.dc.b	$7F,$AB,$EF,$FD,$3A,$F4,$9E,$F3,$E3,$8F,$F5,$A7,$57,$FA,$82,$3C
	.dc.b	$91,$FE,$99,$39,$E1,$FE,$AF,$7D,$27,$B6,$39,$47,$99,$B7,$E0,$9F
	.dc.b	$15,$F9,$9E,$27,$48,$F7,$1B,$AF,$F6,$24,$7C,$B7,$FD,$6F,$D3,$FE
	.dc.b	$94,$87,$57,$FA,$65,$FF,$67,$C6,$63,$10,$9F,$F5,$17,$57,$F9,$3D
	.dc.b	$B1,$FE,$5F,$7F,$D5,$15,$C7,$AB,$FD,$47,$5F,$EA,$1A,$FF,$3C,$73
	.dc.b	$7F,$A4,$5F,$F3,$E0,$E8,$FF,$45,$1F,$E9,$0E,$8F,$F5,$3F,$47,$FA
	.dc.b	$12,$3F,$D2,$4D,$8F,$F4,$5C,$F1,$7F,$D3,$5D,$5F,$E9,$C8,$EE,$64
	.dc.b	$FF,$45,$9F,$E8,$D3,$FD,$4E,$5E,$AF,$F5,$6B,$C2,$FB,$E1,$FE,$90
	.dc.b	$8D,$C6,$1E,$82,$78,$BA,$BF,$2C,$FB,$1F,$F4,$FC,$77,$C3,$FD,$32
	.dc.b	$DB,$FF,$4D,$3C,$4F,$3D,$77,$3E,$23,$E5,$AE,$71,$96,$9F,$22,$7D
	.dc.b	$17,$FE,$BC,$AF,$76,$1F,$15,$EC,$7A,$53,$F5,$2F,$D0,$F7,$21,$B3
	.dc.b	$57,$F4,$1A,$7A,$1F,$F5,$27,$85,$FE,$AA,$24,$7F,$A4,$5A,$FF,$A9
	.dc.b	$1A,$FF,$A9,$1F,$F5,$8F,$85,$23,$FD,$30,$D7,$FD,$56,$D7,$FD,$56
	.dc.b	$D7,$FD,$68,$D7,$FD,$66,$D5,$D4,$8C,$36,$3B,$90,$EE,$53,$B9,$4E
	.dc.b	$E5,$3B,$94,$EF,$1D,$88,$E6,$3C,$0F,$62,$1D,$8A,$76,$29,$D8,$A7
	.dc.b	$63,$A4,$76,$73,$1D,$9E,$84,$3A,$14,$E8,$53,$A1,$CA,$3F,$D3,$DE
	.dc.b	$14,$8F,$F4,$F3,$5F,$F4,$F3,$5F,$F4,$F3,$5F,$F4,$F3,$5F,$F5,$6B
	.dc.b	$5F,$F5,$6B,$7A,$CD,$24,$78,$08,$78,$0A,$78,$0A,$78,$0A,$78,$0A
	.dc.b	$78,$0B,$1E,$06,$C7,$81,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$58,$EC
	.dc.b	$D8,$EC,$F4,$21,$D0,$A7,$42,$9D,$0B,$1D,$1B,$1D,$1E,$44,$39,$14
	.dc.b	$E4,$58,$E4,$D8,$E4,$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23,$4D,$8D
	.dc.b	$46,$08,$E0,$91,$F5,$08,$7C,$92,$9F,$00,$A7,$A4,$A7,$90,$A7,$84
	.dc.b	$B1,$E0,$6C,$78,$1E,$C4,$3B,$14,$EC,$53,$B1,$4E,$C5,$8E,$CD,$8E
	.dc.b	$CF,$42,$1D,$0A,$74,$29,$D0,$B1,$D1,$B1,$D1,$E4,$43,$91,$4E,$45
	.dc.b	$8E,$4D,$8E,$4F,$02,$1C,$0B,$1C,$1B,$1C,$1D,$12,$34,$D8,$D4,$60
	.dc.b	$8E,$09,$1F,$50,$87,$C9,$29,$F0,$0A,$7A,$4A,$79,$0A,$78,$4B,$1E
	.dc.b	$06,$C7,$81,$EC,$7E,$EF,$C2,$78,$1F,$DA,$A2,$1F,$DB,$23,$2E,$0C
	.dc.b	$45,$69,$63,$6E,$DA,$F4,$6B,$DD,$AF,$23,$84,$75,$51,$A6,$C6,$63
	.dc.b	$04,$70,$48,$F2,$35,$EA,$60,$86,$08,$79,$DE,$0D,$36,$53,$67,$AD
	.dc.b	$DB,$4D,$94,$D9,$EF,$39,$10,$D1,$4D,$18,$21,$82,$1F,$10,$EE,$62
	.dc.b	$3E,$23,$83,$47,$88,$A7,$73,$E7,$F6,$FC,$D2,$3F,$75,$AF,$DD,$6B
	.dc.b	$F7,$5F,$F5,$27,$37,$F9,$EC,$FF,$3E,$F3,$FF,$A0,$67,$FD,$1E,$7F
	.dc.b	$8E,$3A,$BF,$D3,$8F,$FA,$0E,$3F,$D2,$AF,$FA,$7A,$33,$CF,$FE,$91
	.dc.b	$8E,$27,$FA,$84,$FF,$43,$75,$7F,$A6,$5F,$F4,$A1,$C7,$AB,$FD,$37
	.dc.b	$5C,$B0,$F1,$CF,$03,$C5,$3A,$3F,$45,$7F,$E9,$2A,$FF,$5A,$4F,$FA
	.dc.b	$68,$E0,$78,$5F,$2D,$F8,$63,$D4,$7F,$A6,$6F,$E1,$61,$EC,$8F,$F5
	.dc.b	$5C,$FE,$BE,$BC,$93,$9B,$4F,$37,$E0,$1F,$EA,$DA,$F6,$1E,$E3,$99
	.dc.b	$C6,$FB,$BE,$1A,$FA,$13,$FE,$AA,$E7,$FF,$4E,$4F,$FA,$97,$A7,$FD
	.dc.b	$40,$FF,$AA,$A3,$33,$FE,$A6,$8F,$E0,$6D,$7F,$D4,$A7,$FA,$CB,$A6
	.dc.b	$F5,$7F,$AC,$27,$FD,$57,$5D,$BA,$3F,$CF,$5C,$FF,$E8,$2C,$3F,$CF
	.dc.b	$31,$8E,$7F,$F4,$79,$FE,$73,$8F,$F4,$61,$FE,$80,$3F,$CE,$A7,$F9
	.dc.b	$E4,$CC,$FF,$A2,$5F,$F3,$C1,$FE,$7E,$9E,$8D,$E8,$FF,$44,$46,$23
	.dc.b	$94,$F1,$7C,$7D,$5F,$EB,$57,$81,$E0,$9F,$F5,$29,$C4,$95,$D2,$78
	.dc.b	$4F,$B2,$31,$D5,$FE,$9A,$9F,$F5,$3D,$F7,$BE,$51,$C1,$E7,$7D,$EF
	.dc.b	$F9,$95,$6E,$B8,$E1,$FE,$A5,$AF,$F5,$CD,$AF,$73,$FE,$A2,$EC,$76
	.dc.b	$3D,$2F,$E6,$D1,$86,$C6,$70,$FE,$79,$39,$6C,$65,$FE,$87,$19,$6C
	.dc.b	$65,$FE,$91,$19,$7F,$A4,$F6,$3E,$2B,$F9,$D3,$63,$FA,$43,$5F,$E9
	.dc.b	$8D,$78,$C7,$02,$3C,$1A,$F0,$6B,$C0,$D1,$0D,$14,$D1,$82,$18,$21
	.dc.b	$FD,$8E,$3C,$8D,$8E,$ED,$3A,$14,$E6,$53,$99,$FE,$D6,$8E,$6D,$8E
	.dc.b	$6D,$39,$94,$E6,$53,$99,$FE,$C2,$78,$B4,$E2,$53,$89,$4E,$31,$E2
	.dc.b	$70,$79,$C8,$77,$29,$A3,$04,$3F,$BA,$F9,$7F,$DC,$C8,$FF,$80,$DA
	.dc.b	$FF,$80,$DA,$FF,$80,$DA,$FF,$80,$DA,$FF,$84,$5A,$F6,$3A,$10,$E8
	.dc.b	$53,$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38
	.dc.b	$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$3A,$91,$F2,$1B
	.dc.b	$1E,$F6,$BE,$86,$BE,$36,$BE,$06,$BD,$DA,$F7,$8E,$C4,$7B,$35,$EC
	.dc.b	$D7,$B3,$5E,$CD,$7B,$35,$EC,$74,$21,$D0,$A7,$42,$9D,$0A,$74,$29
	.dc.b	$D0,$E4,$43,$91,$4E,$45,$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D
	.dc.b	$10,$D1,$4D,$18,$21,$82,$47,$D4,$24,$7C,$96,$C7,$C0,$6B,$E9,$6B
	.dc.b	$E4,$6B,$E1,$6B,$E0,$6B,$E0,$8E,$C4,$7B,$35,$EC,$D7,$B3,$5E,$CD
	.dc.b	$7B,$35,$EC,$74,$21,$D0,$A7,$42,$9D,$0A,$74,$29,$D0,$E4,$43,$91
	.dc.b	$4E,$45,$39,$14,$E4,$70,$3F,$79,$E0,$3C,$CF,$F4,$D3,$63,$FD,$50
	.dc.b	$D7,$FD,$58,$D7,$FD,$62,$D7,$FD,$68,$D4,$1C,$36,$3B,$91,$EE,$D7
	.dc.b	$BB,$5E,$ED,$7B,$B5,$EF,$1D,$89,$19,$8F,$01,$D8,$87,$62,$9D,$8A
	.dc.b	$76,$29,$D9,$E9,$1D,$A3,$27,$63,$A1,$0E,$85,$3A,$14,$E8,$72,$7A
	.dc.b	$39,$3A,$1C,$88,$72,$29,$C8,$E0,$FD,$3C,$7F,$C5,$FF,$13,$FE,$2C
	.dc.b	$24,$FF,$8A,$9A,$FF,$8A,$1A,$FF,$8A,$33,$47,$FA,$A3,$A3,$FC,$EF
	.dc.b	$D3,$FE,$7A,$EB,$FF,$48,$F3,$FF,$A4,$FB,$3F,$D2,$B3,$96,$D7,$02
	.dc.b	$63,$FE,$96,$BF,$F4,$AB,$FE,$85,$8C,$C6,$23,$FD,$35,$3F,$EA,$17
	.dc.b	$FD,$19,$7F,$E9,$27,$FD,$4A,$63,$B3,$FD,$28,$7F,$A6,$67,$C3,$1F
	.dc.b	$E8,$F3,$C0,$F3,$7F,$D5,$15,$DC,$91,$C1,$F6,$E1,$FE,$B4,$73,$3D
	.dc.b	$CF,$09,$FE,$AC,$77,$87,$FA,$9E,$FC,$73,$ED,$3C,$27,$13,$DA,$7F
	.dc.b	$AA,$0E,$57,$E1,$3C,$D8,$7F,$AB,$B0,$5F,$FA,$A3,$B3,$FD,$4B,$7F
	.dc.b	$EA,$7E,$AF,$F5,$2C,$1A,$7E,$D2,$3F,$D4,$F1,$97,$07,$FA,$B4,$A7
	.dc.b	$32,$DF,$FA,$A1,$FF,$56,$1B,$EB,$FF,$56,$4F,$35,$D9,$FE,$83,$EB
	.dc.b	$74,$7F,$AA,$5F,$F4,$47,$67,$F0,$67,$FD,$21,$D5,$FE,$94,$C7,$FD
	.dc.b	$12,$D8,$E3,$1F,$EA,$A3,$FD,$47,$3C,$6B,$FD,$6B,$19,$EB,$FF,$4C
	.dc.b	$CE,$63,$FD,$4E,$7F,$AD,$27,$6F,$FA,$B9,$F3,$1E,$52,$BC,$FB,$13
	.dc.b	$CE,$3C,$27,$98,$FF,$52,$10,$EC,$78,$2F,$DE,$FF,$B1,$CC,$9F,$EB
	.dc.b	$33,$83,$A3,$E1,$B8,$AF,$1B,$E7,$3F,$D9,$55,$FE,$B0,$6D,$79,$6F
	.dc.b	$FD,$6F,$1D,$9B,$7D,$6F,$E0,$C7,$FB,$22,$F9,$72,$FF,$AD,$3E,$97
	.dc.b	$FD,$62,$4C,$3F,$D5,$ED,$7F,$D5,$AD,$7F,$83,$3F,$88,$5E,$3B,$F0
	.dc.b	$CF,$F6,$14,$7E,$32,$31,$A6,$DB,$19,$9D,$BF,$94,$8C,$B6,$32,$FE
	.dc.b	$6E,$32,$D8,$CB,$FD,$4A,$32,$D8,$CB,$E6,$6C,$6A,$31,$3E,$59,$C9
	.dc.b	$1C,$12,$3D,$0D,$7C,$CD,$7C,$8D,$7C,$4D,$7B,$9C,$88,$72,$29,$C8
	.dc.b	$A7,$22,$9C,$8E,$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30
	.dc.b	$49,$F9,$44,$9F,$91,$1F,$85,$FC,$61,$F1,$5F,$87,$84,$61,$D4,$66
	.dc.b	$31,$1F,$C5,$1A,$FF,$18,$6B,$FC,$75,$AF,$F8,$49,$AF,$68,$E8,$47
	.dc.b	$B9,$0E,$CE,$5A,$64,$F0,$39,$69,$93,$C4,$E5,$A6,$4F,$23,$96,$99
	.dc.b	$3C,$C5,$8F,$49,$23,$84,$68,$F2,$90,$F1,$94,$F0,$94,$EE,$53,$91
	.dc.b	$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$09,$1E,$D7,$E0
	.dc.b	$47,$BA,$31,$1E,$76,$BE,$26,$BD,$DA,$F6,$6B,$DA,$3A,$11,$EE,$43
	.dc.b	$B3,$96,$99,$3C,$0E,$5A,$64,$F1,$39,$69,$93,$C8,$E5,$A6,$4F,$31
	.dc.b	$63,$D2,$48,$E1,$1A,$3C,$A4,$3C,$65,$3C,$25,$3B,$94,$E4,$70,$21
	.dc.b	$C0,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$47,$B5,$F8,$11,$EE
	.dc.b	$8C,$47,$9D,$AF,$89,$AF,$76,$BD,$9A,$F6,$8E,$84,$7B,$90,$EC,$E5
	.dc.b	$A6,$4F,$03,$96,$99,$3C,$4E,$5A,$64,$F2,$39,$69,$93,$CC,$58,$F4
	.dc.b	$92,$38,$47,$EA,$7F,$27,$78,$7F,$CA,$DB,$1F,$DC,$DA,$FF,$8F,$1A
	.dc.b	$FF,$90,$1A,$FF,$91,$5A,$FF,$AE,$9A,$9A,$48,$F0,$12,$3C,$0D,$7C
	.dc.b	$0D,$7C,$0D,$7C,$0D,$7C,$0D,$7C,$0D,$8F,$04,$76,$21,$D8,$A7,$62
	.dc.b	$9D,$8A,$76,$29,$D8,$AF,$67,$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$5E
	.dc.b	$8F,$22,$1C,$8A,$7F,$9F,$BF,$10,$FF,$76,$73,$9A,$4E,$FF,$F1,$84
	.dc.b	$65,$B1,$99,$FF,$3D,$CE,$5F,$F1,$B7,$3F,$FA,$DD,$FF,$48,$3F,$E8
	.dc.b	$D8,$FF,$58,$F5,$7F,$A4,$3A,$FF,$D5,$55,$FE,$FC,$AE,$09,$A7,$FA
	.dc.b	$CD,$FF,$4D,$99,$8F,$F5,$0B,$FE,$B1,$EC,$FF,$52,$B6,$BB,$3F,$EA
	.dc.b	$58,$FF,$56,$BF,$EA,$23,$C1,$D9,$FE,$CB,$6E,$1E,$67,$D3,$7E,$67
	.dc.b	$FD,$77,$7D,$1E,$C7,$03,$FD,$62,$75,$3D,$55,$D5,$B5,$D5,$F5,$1F
	.dc.b	$EB,$EC,$3E,$33,$C9,$B8,$7A,$1B,$7E,$18,$3F,$EB,$B9,$F1,$46,$F0
	.dc.b	$F4,$57,$CA,$C3,$C4,$FC,$C8,$C9,$A8,$FF,$60,$D2,$BA,$BF,$EB,$3E
	.dc.b	$9F,$F5,$F5,$FF,$AD,$FA,$BF,$D6,$BD,$9F,$EC,$2A,$FF,$5D,$F6,$7F
	.dc.b	$B0,$5F,$F6,$1C,$7E,$12,$31,$1C,$9F,$F6,$6D,$73,$BF,$F4,$9F,$4F
	.dc.b	$FA,$27,$B3,$FD,$21,$D5,$FE,$A6,$C3,$FD,$48,$7F,$A4,$5C,$76,$7F
	.dc.b	$AE,$A3,$FD,$34,$7F,$AA,$9F,$F4,$FB,$B9,$FF,$4D,$74,$FF,$B0,$7B
	.dc.b	$3F,$D6,$93,$FE,$9E,$3F,$D5,$55,$FE,$9B,$7F,$D5,$D1,$E4,$9E,$11
	.dc.b	$FE,$B9,$AF,$29,$1F,$2D,$71,$7F,$D6,$47,$84,$CD,$73,$7F,$D6,$47
	.dc.b	$86,$3C,$E7,$1C,$3F,$D5,$B3,$B9,$E5,$3E,$F9,$F8,$D1,$E8,$9D,$4F
	.dc.b	$1C,$3E,$2C,$7B,$A7,$E2,$3F,$EB,$6C,$3F,$DF,$15,$BC,$3E,$2D,$7F
	.dc.b	$B4,$FA,$1E,$57,$F9,$9C,$91,$FE,$6F,$6B,$FE,$69,$6B,$FE,$69,$7F
	.dc.b	$CD,$DF,$1B,$23,$FC,$EC,$D7,$FC,$E0,$D7,$FC,$E0,$D7,$FC,$E0,$D7
	.dc.b	$FC,$E0,$D7,$FC,$FA,$D9,$34,$91,$E0,$21,$E0,$29,$E0,$29,$E0,$29
	.dc.b	$E0,$29,$E0,$2C,$78,$1B,$1E,$07,$B1,$0E,$C5,$3B,$14,$EC,$53,$B1
	.dc.b	$63,$B3,$63,$B3,$D0,$87,$42,$9D,$01,$06,$0F,$77,$FD,$50,$D8,$FF
	.dc.b	$78,$35,$E0,$72,$70,$E1,$AF,$56,$BB,$8E,$A4,$7A,$B5,$EA,$79,$1C
	.dc.b	$3D,$5A,$F5,$6B,$D4,$E6,$43,$9C,$78,$9E,$86,$4E,$85,$3A,$14,$E8
	.dc.b	$72,$8E,$EF,$23,$06,$CC,$1D,$0C,$90,$C1,$0E,$E6,$4A,$64,$F0,$94
	.dc.b	$EA,$68,$86,$8A,$68,$C1,$0C,$12,$3E,$21,$23,$DA,$D8,$F6,$35,$F3
	.dc.b	$1C,$DC,$3E,$56,$BE,$26,$BB,$8E,$A4,$7A,$B5,$EA,$79,$5C,$3D,$5A
	.dc.b	$F5,$6B,$D4,$E6,$43,$9C,$78,$9E,$86,$4E,$85,$3A,$14,$E8,$72,$8E
	.dc.b	$EF,$23,$06,$CC,$1D,$0C,$90,$C1,$0E,$E6,$4A,$64,$F0,$94,$EA,$68
	.dc.b	$86,$8A,$68,$C1,$0C,$12,$3E,$21,$23,$DA,$D8,$F6,$35,$F3,$1C,$DC
	.dc.b	$3E,$56,$BE,$26,$BB,$8E,$A4,$7A,$B5,$EA,$79,$5C,$3D,$5A,$F5,$6B
	.dc.b	$D4,$E6,$43,$9C,$78,$9E,$86,$4E,$85,$3A,$14,$E8,$72,$8E,$EF,$23
	.dc.b	$06,$CC,$1F,$5F,$F1,$97,$BB,$FC,$72,$3F,$98,$C6,$1C,$35,$FE,$42
	.dc.b	$D7,$F9,$7B,$5F,$E7,$0D,$79,$C7,$55,$19,$6C,$66,$30,$47,$04,$8F
	.dc.b	$1B,$5F,$1B,$5F,$1B,$5F,$1B,$5F,$19,$C8,$87,$22,$9C,$8A,$72,$29
	.dc.b	$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30,$43,$04,$3E,$49
	.dc.b	$E4,$29,$DC,$A7,$42,$9C,$C7,$E3,$2E,$4F,$AC,$D8,$FC,$23,$5F,$EC
	.dc.b	$2D,$7F,$B7,$B5,$FE,$F0,$D7,$FB,$DB,$5E,$F1,$F8,$3E,$2F,$F3,$F4
	.dc.b	$7F,$A3,$79,$FF,$CF,$DD,$3F,$E8,$FE,$8F,$F4,$19,$1F,$F4,$A5,$7F
	.dc.b	$AB,$FA,$3F,$D2,$13,$FE,$84,$7F,$D5,$A4,$E9,$FF,$48,$61,$FE,$9A
	.dc.b	$3F,$D6,$B5,$C6,$3F,$D0,$F8,$78,$27,$FD,$4E,$6E,$BA,$E1,$E3,$3C
	.dc.b	$33,$FE,$BC,$8F,$F4,$8C,$FF,$AC,$0F,$F5,$99,$FE,$BB,$AF,$F5,$11
	.dc.b	$E2,$BF,$7C,$78,$AB,$94,$E7,$0F,$74,$FF,$AB,$67,$C9,$7E,$B8,$FF
	.dc.b	$54,$D7,$46,$90,$F4,$3E,$C7,$D0,$7F,$A9,$DF,$F4,$E7,$4F,$FA,$8F
	.dc.b	$2F,$F5,$2C,$FE,$07,$A3,$2D,$3F,$DC,$1D,$3F,$EA,$46,$D7,$FA,$9C
	.dc.b	$8F,$FA,$B1,$E9,$3C,$AB,$A8,$7F,$D0,$3C,$F2,$BF,$D1,$8F,$F9,$FD
	.dc.b	$FF,$47,$9F,$E8,$23,$FD,$3A,$68,$A6,$4F,$F4,$CF,$3F,$FA,$15,$FF
	.dc.b	$4C,$9B,$33,$1D,$DE,$1D,$3F,$E9,$87,$4D,$73,$D3,$FE,$94,$7F,$D5
	.dc.b	$42,$BF,$D4,$1C,$BF,$EA,$E8,$FF,$5D,$D2,$B8,$47,$69,$D4,$FF,$A8
	.dc.b	$AB,$FD,$53,$7E,$38,$E1,$D3,$FE,$B5,$9C,$DF,$93,$0E,$75,$C6,$FE
	.dc.b	$34,$FF,$AB,$9F,$F5,$44,$72,$8F,$34,$7F,$9A,$63,$15,$D3,$C2,$78
	.dc.b	$8F,$B9,$FE,$81,$18,$6C,$67,$0F,$F2,$FC,$E5,$B1,$97,$FC,$DF,$19
	.dc.b	$6C,$65,$FF,$3D,$46,$5F,$F3,$E7,$87,$FE,$7E,$E1,$FE,$A0,$D8,$FF
	.dc.b	$36,$B5,$FF,$3D,$B5,$E2,$7E,$C0,$B1,$C9,$C1,$FE,$8A,$8E,$0D,$32
	.dc.b	$1C,$B4,$CB,$82,$38,$24,$77,$23,$E8,$32,$57,$2F,$A8,$C9,$5C,$BE
	.dc.b	$A3,$81,$0D,$14,$D1,$82,$18,$21,$EF,$7A,$B4,$E4,$53,$89,$4E,$27
	.dc.b	$C3,$78,$B4,$E2,$53,$89,$4E,$31,$F2,$23,$8B,$F3,$A3,$FD,$77,$EC
	.dc.b	$FF,$AE,$88,$FF,$AE,$9A,$FF,$A9,$1A,$FF,$AE,$5A,$FF,$AE,$5A,$FF
	.dc.b	$AF,$DA,$F6,$3A,$10,$E8,$53,$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8
	.dc.b	$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$C1,$36,$47,$C8,$6C,$7B,$DA,$FA,$1A,$F8,$DA,$F8,$1A,$F7,$6B
	.dc.b	$DE,$3B,$11,$EC,$D7,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B1,$D0,$87,$42
	.dc.b	$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0
	.dc.b	$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$09,$1F,$50,$91,$F2
	.dc.b	$5B,$1F,$01,$AF,$A5,$AF,$91,$AF,$85,$AF,$81,$AF,$82,$3B,$11,$EC
	.dc.b	$D7,$B3,$5E,$CD,$7B,$35,$EC,$D7,$B1,$D0,$87,$42,$9D,$0A,$74,$29
	.dc.b	$D0,$A7,$43,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0,$FD,$E7,$82,$F6
	.dc.b	$3F,$D4,$CD,$8F,$F5,$A3,$51,$0A,$E5,$C0,$70,$ED,$AF,$08,$EE,$47
	.dc.b	$BB,$5E,$67,$37,$0F,$68,$E6,$E4,$E2,$43,$8B,$A2,$47,$28,$F2,$35
	.dc.b	$F2,$1C,$4E,$CE,$5E,$C7,$47,$11,$D2,$38,$1D,$0A,$74,$38,$18,$39
	.dc.b	$9D,$8C,$3C,$DD,$90,$D1,$4E,$C5,$3B,$1A,$21,$A2,$9A,$30,$7D,$CE
	.dc.b	$97,$CB,$38,$3F,$DE,$91,$86,$C6,$67,$EA,$CE,$5B,$19,$7F,$07,$19
	.dc.b	$6C,$65,$FC,$54,$65,$FC,$4F,$57,$FA,$5B,$A3,$FD,$43,$1F,$EA,$48
	.dc.b	$FF,$54,$3F,$92,$6B,$FE,$A1,$3F,$D2,$26,$7A,$7F,$D3,$3B,$7F,$A8
	.dc.b	$5F,$F4,$FC,$7F,$A5,$E7,$B4,$FF,$A5,$E3,$B7,$57,$FA,$8A,$FF,$D5
	.dc.b	$B5,$D8,$91,$FE,$BF,$AE,$95,$FE,$C9,$3F,$D6,$07,$FA,$C8,$E7,$87
	.dc.b	$FA,$E1,$D1,$83,$FD,$5C,$70,$3D,$24,$A5,$77,$8F,$46,$1E,$A7,$E0
	.dc.b	$4F,$96,$3D,$F1,$A3,$BE,$1E,$AA,$E5,$0A,$F8,$2F,$C7,$3C,$4F,$36
	.dc.b	$D7,$78,$F1,$DF,$FA,$F2,$FE,$2C,$E6,$FA,$D7,$4B,$F5,$46,$0C,$CE
	.dc.b	$EF,$96,$1F,$EC,$BA,$F5,$CF,$F0,$DC,$3F,$02,$7D,$2E,$A7,$57,$FA
	.dc.b	$FA,$BF,$D8,$67,$FB,$1C,$F6,$46,$9C,$64,$8F,$F6,$54,$6A,$7F,$D9
	.dc.b	$2F,$E4,$23,$33,$B9,$ED,$3B,$39,$CF,$5E,$AF,$F6,$64,$7F,$A9,$FA
	.dc.b	$7F,$92,$F3,$FE,$87,$A3,$FD,$51,$3F,$EC,$87,$FD,$47,$C1,$FE,$A5
	.dc.b	$E9,$FF,$56,$9F,$EA,$EE,$5F,$F5,$B9,$2B,$FD,$A7,$5C,$1F,$F6,$3C
	.dc.b	$65,$FF,$59,$1F,$EC,$6B,$FF,$62,$BC,$2B,$AE,$1F,$EB,$DC,$3F,$D6
	.dc.b	$F1,$89,$FF,$5B,$9E,$57,$15,$FE,$C2,$8F,$F6,$9D,$7B,$2B,$8B,$CE
	.dc.b	$38,$35,$FF,$6F,$9F,$0C,$FF,$71,$9C,$8B,$87,$AD,$FF,$6B,$1D,$4F
	.dc.b	$F5,$24,$3C,$1F,$F5,$AB,$63,$FD,$6A,$D7,$FD,$6A,$D7,$FD,$7A,$D7
	.dc.b	$FD,$64,$D7,$AC,$74,$3F,$A6,$BF,$D7,$3C,$17,$C2,$62,$11,$87,$6D
	.dc.b	$7F,$AA,$B5,$FE,$C2,$D7,$FB,$33,$5E,$D1,$D6,$31,$1D,$A3,$04,$7B
	.dc.b	$B5,$EE,$D7,$BB,$5E,$ED,$7B,$9C,$88,$72,$29,$C8,$A7,$22,$9C,$8E
	.dc.b	$04,$38,$14,$E0,$53,$81,$A2,$1A,$29,$A3,$04,$30,$43,$E3,$1E,$42
	.dc.b	$9D,$CA,$74,$29,$CC,$A7,$38,$FA,$39,$BE,$5F,$F0,$0D,$9F,$C2,$B5
	.dc.b	$FF,$09,$B5,$FF,$0F,$B5,$FF,$12,$35,$FF,$13,$B5,$34,$91,$E0,$24
	.dc.b	$78,$1A,$F8,$1A,$F8,$1A,$F8,$1A,$F8,$1A,$F8,$1B,$1E,$08,$EC,$43
	.dc.b	$B1,$4E,$C5,$3B,$14,$EC,$53,$B1,$5E,$CF,$42,$1D,$0A,$74,$29,$D0
	.dc.b	$A7,$42,$BD,$1E,$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A,$70
	.dc.b	$2B,$C1,$D1,$0D,$15,$D3,$82,$38,$24,$7D,$42,$47,$C9,$6B,$F0,$1A
	.dc.b	$FA,$5A,$F9,$1A,$F8,$5A,$F8,$1B,$1E,$08,$EC,$43,$B1,$4E,$C5,$3B
	.dc.b	$14,$EC,$53,$B1,$5E,$CF,$42,$1D,$0A,$74,$29,$D0,$A7,$42,$BD,$1E
	.dc.b	$44,$39,$14,$E4,$53,$91,$5E,$4F,$02,$1C,$0A,$70,$2B,$C1,$D1,$0D
	.dc.b	$15,$D3,$82,$38,$24,$7D,$42,$47,$C9,$6B,$F0,$1A,$FA,$5A,$F9,$1A
	.dc.b	$F8,$5A,$F8,$1B,$1E,$08,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC,$51,$E5
	.dc.b	$FF,$A2,$9F,$ED,$AD,$7F,$D7,$6D,$7F,$D8,$ED,$7F,$D9,$8D,$7F,$D9
	.dc.b	$ED,$7A,$F3,$FF,$75,$8C,$11,$C1,$23,$C2,$D7,$C2,$D7,$C2,$D7,$C2
	.dc.b	$D7,$C2,$D7,$C2,$74,$21,$D0,$A7,$42,$9D,$0A,$74,$29,$D0,$E4,$43
	.dc.b	$91,$4E,$45,$39,$14,$E4,$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D
	.dc.b	$18,$1D,$8E,$D7,$67,$F8,$B2,$30,$D8,$CC,$FF,$BF,$27,$2D,$8C,$A2
	.dc.b	$1D,$9C,$B8,$8E,$2E,$8D,$4F,$F8,$E7,$97,$FD,$24,$7F,$AE,$3A,$7F
	.dc.b	$D1,$F3,$FE,$B5,$E7,$FF,$52,$46,$23,$FD,$61,$D1,$FE,$AA,$7F,$D4
	.dc.b	$DD,$5F,$EC,$3E,$AF,$F4,$B6,$1F,$EA,$29,$FF,$6A,$57,$FA,$6D,$A7
	.dc.b	$FA,$C2,$7D,$33,$BC,$3B,$CF,$FB,$00,$FF,$61,$3F,$EA,$A7,$81,$FE
	.dc.b	$D5,$BF,$03,$E7,$AE,$71,$FE,$AB,$C3,$FD,$99,$2A,$FF,$53,$D7,$FB
	.dc.b	$52,$30,$FF,$B7,$63,$C2,$47,$D7,$5C,$E7,$C9,$39,$7F,$D9,$47,$23
	.dc.b	$FD,$8C,$7A,$27,$8D,$7C,$9B,$FF,$64,$4E,$4F,$41,$EB,$8F,$3D,$79
	.dc.b	$A3,$AB,$5F,$9F,$1A,$AF,$41,$F0,$1F,$C2,$4F,$FB,$0F,$AB,$FD,$6B
	.dc.b	$7F,$EB,$7B,$FF,$5A,$C7,$E2,$7A,$7F,$D6,$B1,$FE,$B9,$3F,$D7,$D1
	.dc.b	$FE,$C0,$7E,$AF,$57,$FB,$00,$FF,$5E,$77,$E2,$7A,$C6,$27,$FD,$39
	.dc.b	$D5,$FE,$BD,$E3,$FF,$59,$3F,$EB,$FE,$BF,$F5,$4F,$5F,$FA,$F7,$AB
	.dc.b	$FD,$39,$6E,$BF,$F6,$1B,$FE,$A6,$3F,$D4,$97,$CD,$FF,$5C,$4F,$FA
	.dc.b	$AE,$7F,$D4,$C7,$02,$E1,$E0,$BE,$D5,$FE,$BA,$3F,$D5,$67,$FB,$10
	.dc.b	$F0,$46,$DF,$4C,$7F,$AE,$9E,$13,$DB,$0F,$F6,$94,$78,$8C,$BE,$C7
	.dc.b	$FD,$9D,$3E,$0C,$3F,$D8,$D5,$C7,$0F,$46,$1F,$B8,$9F,$89,$1E,$BC
	.dc.b	$3A,$E1,$E5,$BF,$F3,$1B,$F1,$BA,$5E,$27,$FA,$94,$91,$FE,$B0,$6B
	.dc.b	$FE,$B0,$6B,$FE,$B0,$7F,$CF,$3E,$24,$8F,$F3,$CB,$5F,$F3,$CB,$5F
	.dc.b	$F3,$CB,$5F,$F3,$CB,$5F,$F3,$CB,$5F,$F3,$CB,$72,$34,$91,$E0,$21
	.dc.b	$E0,$29,$E0,$29,$E0,$29,$E0,$29,$E0,$2C,$78,$1B,$1E,$07,$B1,$0E
	.dc.b	$C5,$3B,$14,$EC,$53,$B1,$63,$B3,$63,$B3,$D0,$87,$42,$9D,$0F,$E6
	.dc.b	$9E,$23,$C4,$FF,$76,$36,$3F,$DE,$2D,$7F,$9D,$07,$0F,$28,$E2,$47
	.dc.b	$8B,$5E,$31,$C1,$FE,$92,$D7,$FA,$63,$5E,$D1,$C5,$EC,$74,$21,$D1
	.dc.b	$D3,$D1,$AF,$46,$BD,$0E,$27,$43,$91,$83,$91,$4E,$45,$39,$1C,$08
	.dc.b	$73,$21,$C8,$C9,$4C,$9D,$0C,$94,$C9,$D8,$B1,$F1,$49,$1E,$C6,$C7
	.dc.b	$95,$AF,$43,$91,$E1,$21,$DC,$A6,$8C,$10,$C1,$1E,$AF,$92,$3C,$44
	.dc.b	$7C,$4D,$7C,$51,$C1,$EE,$D7,$B3,$5E,$D1,$C5,$EC,$74,$21,$D1,$D3
	.dc.b	$D1,$AF,$46,$BD,$0E,$27,$43,$91,$83,$91,$4E,$45,$39,$1C,$08,$73
	.dc.b	$21,$C8,$C9,$4C,$9D,$0C,$94,$C9,$D8,$B1,$F1,$49,$1E,$C6,$C7,$95
	.dc.b	$AF,$43,$91,$E1,$21,$DC,$A6,$8C,$10,$C1,$1E,$AF,$92,$3C,$44,$7C
	.dc.b	$4D,$7C,$51,$C1,$EE,$D7,$B3,$5E,$D1,$C5,$EC,$74,$21,$D1,$D3,$D1
	.dc.b	$AF,$46,$BD,$0E,$27,$43,$91,$83,$91,$4E,$45,$39,$1C,$0F,$9D,$E2
	.dc.b	$38,$BF,$93,$C7,$F9,$9E,$30,$E1,$AF,$F2,$C6,$BF,$E6,$76,$BF,$E6
	.dc.b	$E6,$BC,$E3,$AA,$8C,$B6,$33,$18,$23,$82,$47,$8D,$AF,$8D,$AF,$8D
	.dc.b	$AF,$8D,$AF,$8C,$E4,$43,$91,$4E,$45,$39,$14,$E4,$70,$21,$C0,$A7
	.dc.b	$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$24,$F2,$14,$EE,$53,$A1
	.dc.b	$4E,$67,$D3,$F8,$6F,$A2,$47,$F0,$0D,$7F,$00,$D7,$F0,$0F,$FA,$97
	.dc.b	$A3,$FD,$1B,$1F,$EA,$03,$FC,$F4,$E3,$97,$FC,$F7,$CD,$FE,$83,$9F
	.dc.b	$F3,$A9,$FE,$86,$E8,$FF,$3A,$3F,$EA,$03,$FD,$0C,$27,$FD,$60,$E6
	.dc.b	$7F,$CF,$D1,$FE,$B4,$E9,$FF,$57,$36,$7C,$3D,$3F,$E8,$17,$FD,$1F
	.dc.b	$5E,$19,$D4,$7F,$A3,$4F,$41,$5F,$3D,$F9,$27,$85,$FF,$AD,$5B,$5F
	.dc.b	$EA,$E3,$D6,$E1,$F4,$D7,$A2,$30,$7A,$30,$F8,$95,$CE,$FF,$D5,$D3
	.dc.b	$EC,$E7,$FF,$43,$61,$F0,$5B,$87,$FA,$57,$3F,$F4,$9B,$7A,$7F,$D1
	.dc.b	$D5,$FE,$20,$8F,$F4,$D3,$FE,$96,$31,$5F,$E9,$5E,$9F,$F4,$99,$FE
	.dc.b	$9D,$AE,$37,$FE,$9E,$E5,$FF,$39,$5F,$F9,$E3,$9F,$FD,$23,$1F,$E8
	.dc.b	$9E,$7F,$F3,$8F,$47,$F9,$D2,$B4,$47,$FD,$24,$FF,$9D,$79,$BF,$D1
	.dc.b	$0F,$FA,$5B,$A7,$FD,$11,$3C,$4F,$F4,$0D,$7F,$A1,$67,$FD,$39,$7E
	.dc.b	$29,$CE,$1F,$E8,$13,$C7,$87,$49,$D6,$1F,$E8,$B3,$A9,$6C,$74,$C3
	.dc.b	$B4,$FF,$A8,$A3,$FD,$2D,$87,$8E,$B8,$47,$FA,$80,$CB,$89,$E1,$1E
	.dc.b	$D3,$27,$67,$C6,$FF,$A7,$8C,$57,$82,$FF,$D3,$95,$CE,$31,$3E,$D3
	.dc.b	$E5,$47,$A1,$F1,$0B,$47,$B3,$AD,$D8,$F9,$2F,$E7,$71,$86,$C6,$6F
	.dc.b	$FC,$CB,$39,$6C,$65,$F0,$08,$D3,$86,$BE,$56,$BE,$7C,$5E,$C7,$F4
	.dc.b	$06,$C7,$F4,$E6,$BF,$E7,$66,$BE,$06,$BC,$E3,$91,$1E,$4D,$79,$35
	.dc.b	$E4,$D7,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08
	.dc.b	$7E,$4A,$3D,$AD,$8F,$33,$4F,$09,$4E,$C5,$3A,$94,$EA,$23,$AB,$63
	.dc.b	$AB,$4E,$A5,$3A,$94,$EA,$53,$AC,$74,$24,$7F,$78,$F6,$24,$FF,$81
	.dc.b	$DA,$FF,$81,$DA,$FF,$81,$DA,$FF,$81,$DA,$FF,$84,$5A,$FF,$81,$DB
	.dc.b	$D2,$69,$23,$C0,$43,$C0,$53,$C0,$53,$C0,$53,$C0,$53,$C0,$58,$F0
	.dc.b	$36,$3C,$0F,$62,$1D,$8A,$76,$29,$D8,$A7,$62,$C7,$66,$C7,$67,$A1
	.dc.b	$0E,$85,$3A,$14,$E8,$58,$E8,$D8,$E8,$F2,$21,$C8,$A7,$22,$C7,$26
	.dc.b	$C7,$27,$81,$0E,$05,$8E,$0D,$8E,$0E,$89,$1A,$6C,$6A,$30,$47,$04
	.dc.b	$8F,$A8,$43,$E4,$94,$F8,$05,$3D,$25,$3C,$85,$3C,$25,$8F,$03,$63
	.dc.b	$C0,$F6,$21,$D8,$A7,$62,$9D,$8A,$76,$2C,$76,$6C,$76,$7A,$10,$E8
	.dc.b	$53,$A1,$4E,$85,$8E,$8D,$8E,$8F,$22,$1C,$8A,$72,$2C,$72,$6C,$72
	.dc.b	$78,$10,$E0,$58,$E0,$D8,$E0,$E8,$91,$A6,$C6,$A3,$04,$70,$48,$FA
	.dc.b	$84,$3E,$49,$4F,$80,$53,$D2,$53,$C8,$53,$C2,$58,$F0,$36,$3C,$0F
	.dc.b	$63,$F7,$7E,$0B,$E0,$7F,$D3,$ED,$8F,$F5,$6B,$5F,$ED,$A1,$C3,$FD
	.dc.b	$C1,$AF,$F8,$05,$AF,$F8,$69,$AF,$18,$EE,$47,$BB,$5E,$F1,$C0,$91
	.dc.b	$98,$F0,$35,$F0,$35,$F0,$35,$F0,$1D,$88,$76,$75,$1D,$A3,$27,$62
	.dc.b	$9D,$8A,$76,$29,$D8,$E8,$61,$E8,$E0,$87,$52,$9D,$4A,$75,$29,$D4
	.dc.b	$E0,$43,$81,$4E,$05,$C7,$FC,$5F,$F1,$DF,$E2,$C2,$47,$F8,$A9,$AF
	.dc.b	$F8,$9F,$34,$62,$3F,$C5,$11,$87,$FD,$E3,$CD,$FE,$94,$E7,$FF,$4B
	.dc.b	$75,$7F,$A1,$A7,$FD,$1D,$28,$75,$FF,$97,$2B,$FD,$1B,$D5,$FE,$47
	.dc.b	$AF,$F4,$83,$FE,$8C,$31,$D7,$FE,$A6,$BE,$98,$7F,$A5,$A3,$18,$7F
	.dc.b	$A9,$E3,$FD,$61,$18,$7F,$D6,$C7,$F9,$AA,$FC,$53,$E1,$BE,$D3,$FE
	.dc.b	$AE,$7F,$D5,$06,$1E,$67,$FA,$D0,$E6,$7F,$AA,$1D,$4F,$A2,$3B,$5F
	.dc.b	$9B,$0F,$F5,$45,$F9,$6B,$55,$C4,$F0,$10,$C1,$1F,$F5,$DB,$D2,$3D
	.dc.b	$0D,$3C,$71,$EC,$7F,$D9,$D7,$F2,$1F,$45,$23,$11,$F2,$4E,$67,$7E
	.dc.b	$BF,$F5,$B9,$3A,$FF,$D5,$F1,$FE,$B2,$7E,$FC,$7F,$AC,$E3,$03,$AB
	.dc.b	$FD,$7B,$1C,$63,$1D,$7F,$EB,$63,$FD,$11,$D7,$FE,$B4,$EA,$FF,$59
	.dc.b	$F4,$E2,$BF,$D1,$27,$FA,$44,$75,$FF,$A5,$1F,$F4,$5F,$5F,$FA,$7D
	.dc.b	$FF,$4A,$DF,$33,$FD,$5A,$FF,$A7,$0E,$37,$FE,$96,$7F,$D4,$04,$9E
	.dc.b	$4E,$A7,$93,$FE,$97,$BF,$F5,$9B,$D8,$F2,$9E,$27,$D1,$3F,$EB,$E8
	.dc.b	$CD,$FF,$AB,$A3,$AB,$E9,$3F,$DA,$55,$E6,$9F,$15,$72,$3F,$D6,$DD
	.dc.b	$5F,$EC,$5C,$3F,$D7,$E7,$C6,$9F,$16,$1D,$23,$11,$FE,$D5,$3E,$46
	.dc.b	$1D,$EB,$96,$1F,$EB,$5B,$FF,$6E,$9C,$5F,$7B,$8C,$3F,$D6,$9C,$6F
	.dc.b	$13,$FD,$62,$4C,$7F,$D5,$ED,$7F,$83,$35,$FE,$0A,$FF,$46,$5E,$4F
	.dc.b	$F0,$AF,$8F,$C3,$63,$FA,$5B,$5F,$F5,$CB,$5F,$F6,$53,$5F,$F7,$1B
	.dc.b	$5E,$BC,$E9,$EA,$D8,$EA,$D7,$AB,$5E,$AD,$7A,$B5,$EB,$1D,$08,$F4
	.dc.b	$6B,$D1,$AF,$46,$BD,$1A,$F4,$39,$10,$E4,$53,$91,$4E,$45,$39,$1C
	.dc.b	$08,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$08,$60,$93,$F4,$C9,$3F
	.dc.b	$25,$B3,$F0,$1F,$B1,$2E,$C7,$C3,$FE,$1E,$30,$D8,$CC,$7E,$2D,$AF
	.dc.b	$F8,$49,$AF,$FC,$00,$6A,$23,$86,$C7,$52,$3D,$5A,$F5,$6B,$D5,$AF
	.dc.b	$53,$CE,$71,$21,$C4,$A7,$12,$9C,$63,$C8,$F2,$38,$10,$E0,$53,$84
	.dc.b	$78,$9F,$59,$D0,$A6,$CA,$72,$21,$C8,$E0,$78,$4C,$9C,$0D,$1C,$CD
	.dc.b	$18,$36,$60,$E0,$43,$04,$8F,$79,$23,$D5,$19,$6C,$66,$3C,$AD,$7C
	.dc.b	$4D,$7C,$2D,$7C,$84,$3C,$4E,$5B,$1D,$88,$F6,$6B,$D9,$AF,$66,$BD
	.dc.b	$4F,$41,$C4,$87,$12,$9C,$4A,$71,$8F,$23,$C8,$E0,$43,$81,$4E,$11
	.dc.b	$E2,$7D,$67,$42,$9B,$29,$C8,$87,$23,$81,$E1,$32,$70,$34,$73,$34
	.dc.b	$60,$D9,$83,$81,$0C,$12,$3D,$E4,$8F,$54,$65,$B1,$98,$F2,$B5,$F1
	.dc.b	$35,$F0,$B5,$F2,$10,$F1,$39,$6C,$76,$23,$D9,$AF,$66,$BD,$9A,$F5
	.dc.b	$3D,$07,$12,$1C,$4A,$71,$29,$C6,$3C,$8F,$23,$81,$0E,$05,$38,$47
	.dc.b	$89,$F5,$9D,$0A,$7C,$6F,$99,$FE,$54,$47,$F9,$53,$5F,$E5,$4D,$7F
	.dc.b	$95,$35,$FE,$2C,$D7,$F9,$5B,$5F,$E7,$4D,$93,$49,$1E,$02,$1E,$02
	.dc.b	$9E,$02,$9E,$02,$9E,$02,$9E,$02,$C7,$81,$B1,$E0,$7B,$10,$EC,$53
	.dc.b	$B1,$4E,$C5,$3B,$16,$3B,$36,$3B,$3D,$08,$74,$29,$FE,$22,$FD,$A1
	.dc.b	$E4,$7F,$88,$9B,$1F,$E2,$26,$BF,$E2,$26,$BF,$E2,$26,$BF,$E2,$26
	.dc.b	$BD,$63,$A1,$FD,$AF,$A7,$FD,$69,$D1,$FE,$9C,$9F,$F4,$47,$4F,$FA
	.dc.b	$0D,$FF,$48,$4F,$FA,$50,$DB,$A9,$FF,$4E,$F6,$7F,$AA,$EB,$84,$FF
	.dc.b	$A9,$0F,$F4,$50,$7F,$D8,$3D,$7F,$E9,$F9,$E1,$5F,$E9,$09,$FF,$51
	.dc.b	$D7,$FA,$63,$B3,$FD,$86,$FF,$A6,$0F,$1B,$88,$FF,$62,$C6,$2F,$FD
	.dc.b	$2F,$87,$9A,$3F,$D4,$D5,$FE,$AF,$BE,$C7,$BE,$3F,$D5,$58,$7F,$77
	.dc.b	$C3,$C2,$7A,$AF,$FD,$5B,$3C,$B0,$F4,$3F,$EA,$8B,$E6,$F8,$4F,$31
	.dc.b	$F0,$6B,$DE,$FA,$E7,$CB,$87,$8E,$38,$4F,$A9,$E1,$5E,$5B,$F5,$1F
	.dc.b	$EB,$99,$D3,$F4,$FA,$BF,$D5,$5D,$7F,$EA,$CE,$CF,$F5,$9E,$1F,$E1
	.dc.b	$11,$4A,$FF,$5A,$46,$63,$13,$FE,$B8,$3F,$D7,$75,$CE,$7F,$D6,$A7
	.dc.b	$FA,$33,$9B,$FD,$24,$FF,$A2,$BA,$FF,$D5,$1D,$1F,$E8,$78,$FF,$4E
	.dc.b	$F5,$7F,$A3,$1F,$F4,$FC,$FF,$A4,$BA,$FF,$D3,$CD,$3F,$D3,$3D,$7F
	.dc.b	$E9,$FB,$E4,$7F,$A9,$CB,$87,$FA,$BA,$FF,$D6,$47,$63,$FD,$75,$7F
	.dc.b	$E9,$B6,$DF,$92,$7C,$F1,$C2,$BF,$D3,$4F,$AC,$86,$63,$B5,$73,$8F
	.dc.b	$F5,$89,$A7,$13,$E4,$8C,$61,$EF,$8C,$47,$FB,$04,$E1,$7E,$36,$E1
	.dc.b	$FE,$AC,$C3,$FC,$8D,$83,$0F,$F2,$46,$09,$FF,$21,$78,$FF,$E5,$F2
	.dc.b	$61,$FE,$5F,$6B,$FE,$5F,$6B,$FE,$5F,$7F,$CC,$DD,$B2,$3F,$CC,$CD
	.dc.b	$7F,$CC,$CD,$7F,$CC,$CD,$7F,$CA,$4D,$7F,$CC,$ED,$7F,$CE,$CD,$E7
	.dc.b	$34,$91,$E0,$21,$E0,$29,$E0,$29,$E0,$29,$E0,$29,$E0,$2C,$78,$1B
	.dc.b	$1E,$07,$B1,$0E,$C5,$3B,$14,$EC,$53,$B1,$63,$B3,$63,$B3,$D0,$87
	.dc.b	$42,$9D,$0F,$F4,$97,$BD,$FE,$99,$23,$FE,$99,$6B,$FE,$99,$6B,$C0
	.dc.b	$D1,$0D,$14,$D1,$82,$18,$21,$FE,$8D,$7A,$B4,$E4,$53,$89,$4E,$27
	.dc.b	$FA,$A1,$E2,$D3,$89,$4E,$25,$38,$9F,$EB,$D7,$8B,$4E,$25,$38,$94
	.dc.b	$E3,$28,$E2,$D8,$E2,$D7,$8B,$5E,$31,$C0,$8F,$06,$BC,$1A,$F0,$34
	.dc.b	$43,$45,$34,$60,$86,$08,$7C,$B2,$1F,$14,$A7,$BC,$B1,$E9,$6C,$79
	.dc.b	$1A,$78,$4A,$78,$0A,$78,$0E,$C4,$3B,$16,$3B,$36,$3B,$34,$EC,$53
	.dc.b	$B1,$4E,$C7,$42,$47,$46,$C7,$46,$9D,$0A,$74,$29,$D2,$39,$12,$39
	.dc.b	$36,$39,$35,$E4,$D7,$94,$70,$23,$C1,$AF,$06,$BC,$0D,$10,$D1,$4D
	.dc.b	$18,$21,$82,$1F,$50,$87,$C9,$29,$F0,$0B,$1E,$96,$C7,$91,$A7,$84
	.dc.b	$A7,$80,$A7,$80,$EC,$43,$B1,$63,$B3,$63,$B3,$4E,$C5,$3B,$14,$EC
	.dc.b	$74,$24,$74,$6C,$74,$7F,$0B,$E5,$FF,$85,$D1,$25,$19,$8C,$35,$46
	.dc.b	$9D,$35,$E8,$D7,$FC,$48,$D7,$B9,$CA,$3A,$46,$63,$B9,$D2,$32,$47
	.dc.b	$04,$3B,$B9,$69,$93,$C2,$E5,$A6,$4F,$1C,$78,$5F,$21,$93,$C6,$74
	.dc.b	$21,$C8,$A7,$28,$EC,$D3,$91,$D0,$D9,$D0,$E4,$43,$81,$A3,$81,$93
	.dc.b	$81,$A3,$06,$8C,$10,$D9,$0D,$16,$3E,$01,$F0,$7C,$17,$C9,$FD,$F6
	.dc.b	$CF,$E0,$DA,$FE,$25,$AF,$F9,$85,$AF,$DE,$6B,$F8,$E6,$A6,$F1,$7F
	.dc.b	$9C,$39,$BF,$D1,$DC,$DF,$E9,$EA,$FF,$40,$1F,$EA,$4E,$8F,$F4,$CB
	.dc.b	$4C,$9F,$E7,$A3,$FC,$FB,$1C,$5F,$F5,$13,$FE,$7D,$9F,$F4,$14,$6C
	.dc.b	$FF,$47,$1F,$E8,$73,$FD,$17,$5F,$EB,$18,$DB,$7A,$3F,$D2,$73,$D0
	.dc.b	$FF,$5A,$10,$E9,$1B,$7A,$74,$7F,$A2,$B0,$F4,$C6,$9B,$3F,$EA,$03
	.dc.b	$FD,$21,$87,$A6,$BF,$D3,$67,$FA,$96,$7F,$D5,$11,$FE,$A2,$AF,$4C
	.dc.b	$E5,$FF,$59,$CF,$43,$6F,$C4,$3C,$D5,$DF,$0F,$9F,$18,$9D,$BE,$A8
	.dc.b	$CB,$FE,$B8,$3A,$9F,$EA,$CC,$3F,$D8,$3D,$1F,$EA,$4D,$FF,$7B,$87
	.dc.b	$FA,$7A,$BF,$D3,$67,$FA,$92,$5D,$1F,$C4,$3A,$3F,$D4,$13,$FE,$A5
	.dc.b	$9F,$F5,$3B,$C9,$DB,$FC,$5E,$BA,$4F,$FA,$9C,$97,$BB,$E8,$FF,$A4
	.dc.b	$B9,$9D,$1F,$EA,$1E,$8F,$F4,$16,$3F,$E8,$A3,$FD,$32,$E1,$FF,$45
	.dc.b	$1F,$E9,$53,$FD,$12,$39,$FF,$D3,$6E,$49,$5D,$A7,$FD,$59,$CF,$E1
	.dc.b	$BF,$05,$7F,$AD,$27,$2F,$4B,$FF,$58,$57,$9F,$A3,$FD,$38,$D9,$F2
	.dc.b	$9F,$EB,$13,$FD,$4A,$78,$2F,$FD,$3C,$DB,$F5,$61,$FE,$9D,$8E,$38
	.dc.b	$76,$7C,$17,$FE,$B0,$BE,$35,$FE,$C6,$BF,$2D,$73,$3C,$17,$FE,$AE
	.dc.b	$9F,$74,$7C,$3B,$E5,$5E,$0C,$3F,$D5,$71,$F0,$63,$9D,$75,$AE,$07
	.dc.b	$FB,$6E,$38,$3F,$D4,$AF,$FD,$64,$BC,$27,$84,$F7,$BF,$AA,$46,$1B
	.dc.b	$19,$9F,$EB,$93,$96,$C6,$5F,$F7,$B4,$65,$B1,$97,$FA,$AC,$65,$FE
	.dc.b	$AF,$E0,$3F,$6D,$7C,$EF,$F6,$58,$C3,$63,$33,$FD,$B6,$72,$D8,$CB
	.dc.b	$FD,$88,$46,$9C,$35,$F0,$C7,$06,$CF,$72,$3D,$5A,$F2,$8E,$0F,$F7
	.dc.b	$E9,$E8,$D8,$CB,$E2,$6C,$6A,$31,$3D,$1F,$09,$DC,$87,$77,$4F,$56
	.dc.b	$BC,$8D,$9C,$8E,$06,$0E,$05,$38,$1A,$21,$C4,$87,$03,$25,$32,$72
	.dc.b	$2C,$7B,$49,$3F,$00,$93,$E8,$8C,$B6,$33,$3F,$E2,$0F,$93,$FF,$0C
	.dc.b	$91,$C1,$0F,$F1,$1B,$96,$99,$3F,$C4,$AE,$5A,$64,$FF,$0A,$B9,$69
	.dc.b	$93,$FC,$70,$E5,$A6,$4F,$F1,$BB,$96,$99,$EC,$46,$5B,$19,$8C,$11
	.dc.b	$C1,$0F,$59,$0F,$49,$4F,$39,$4F,$29,$4F,$19,$63,$C2,$D8,$F0,$34
	.dc.b	$F0,$1D,$88,$76,$29,$D8,$A7,$62,$C7,$66,$C7,$66,$9D,$8E,$84,$3A
	.dc.b	$14,$E8,$58,$E8,$D8,$E8,$D3,$A1,$C8,$87,$22,$C7,$26,$C7,$26,$9C
	.dc.b	$8E,$04,$8E,$0D,$8E,$0D,$38,$46,$89,$1A,$6C,$6A,$30,$47,$04,$3E
	.dc.b	$A1,$0F,$92,$53,$E0,$14,$F4,$94,$F2,$16,$3C,$2D,$8F,$03,$4F,$01
	.dc.b	$D8,$87,$62,$9D,$8A,$76,$2C,$76,$6C,$76,$69,$D8,$E8,$43,$A1,$4E
	.dc.b	$85,$8E,$8D,$8E,$8D,$3A,$1C,$88,$72,$2C,$72,$6C,$72,$69,$C8,$E0
	.dc.b	$48,$E0,$D8,$E0,$D3,$84,$68,$91,$A6,$C6,$A3,$04,$70,$43,$FD,$7F
	.dc.b	$EF,$FF,$AE,$88,$FF,$B1,$1A,$E8,$C1,$0C,$10,$FF,$63,$BC,$1A,$6C
	.dc.b	$A6,$CF,$F5,$CB,$B6,$9B,$29,$B3,$FD,$BC,$ED,$A6,$CA,$6C,$FF,$6F
	.dc.b	$BB,$69,$B2,$9B,$94,$6D,$B1,$B6,$BB,$8D,$11,$D3,$5D,$18,$21,$EB
	.dc.b	$C1,$E1,$3C,$BF,$F0,$A3,$67,$FC,$24,$D7,$FC,$1A,$D7,$FC,$3A,$D7
	.dc.b	$FC,$0C,$D7,$FC,$2A,$D9,$EF,$3F,$E1,$99,$FF,$4B,$76,$FF,$AB,$7A
	.dc.b	$BF,$D4,$BD,$BF,$E9,$88,$FF,$4B,$75,$7F,$A8,$E3,$FD,$2D,$5F,$EA
	.dc.b	$93,$FD,$1D,$5F,$E8,$B7,$FD,$4F,$5D,$1F,$F4,$BF,$67,$FA,$CB,$B7
	.dc.b	$FD,$39,$1F,$E9,$23,$96,$1F,$EA,$87,$FD,$34,$67,$0F,$13,$FE,$C7
	.dc.b	$37,$5F,$E9,$C9,$FF,$68,$61,$E1,$AF,$29,$30,$F6,$5F,$90,$EA,$78
	.dc.b	$30,$F1,$B4,$FF,$69,$DF,$FA,$B6,$FA,$57,$C2,$AD,$DF,$5B,$F6,$D7
	.dc.b	$C8,$AE,$F3,$99,$F8,$A7,$FB,$52,$B8,$CF,$CC,$C3,$D4,$7A,$E3,$FD
	.dc.b	$73,$87,$CF,$9F,$05,$79,$27,$27,$27,$FD,$88,$7F,$AD,$7A,$FF,$D5
	.dc.b	$28,$C1,$FE,$BA,$AF,$F5,$8F,$5F,$FA,$BE,$31,$DB,$FE,$B8,$78,$4F
	.dc.b	$FB,$05,$FF,$5E,$57,$E1,$DE,$7D,$BF,$EB,$87,$75,$FE,$C9,$7F,$D2
	.dc.b	$A7,$FA,$D4,$FF,$4F,$F5,$7F,$A2,$7A,$7F,$D6,$2D,$EB,$FF,$48,$CF
	.dc.b	$FA,$59,$FF,$48,$3F,$E8,$E3,$FD,$77,$1B,$7F,$D2,$06,$CE,$E7,$FA
	.dc.b	$EB,$B3,$FD,$62,$E2,$3F,$D5,$4F,$8B,$B5,$DB,$FE,$AE,$7F,$D7,$B8
	.dc.b	$7F,$A9,$6F,$8D,$7F,$AA,$63,$CF,$3C,$23,$FD,$84,$79,$6F,$AC,$FF
	.dc.b	$AB,$67,$9E,$1F,$EC,$1A,$F0,$61,$DC,$FF,$55,$CF,$3A,$F0,$4F,$FA
	.dc.b	$FA,$BE,$C3,$EB,$3F,$D5,$8F,$FA,$07,$C7,$78,$DF,$E9,$96,$CF,$FA
	.dc.b	$5D,$AF,$FA,$49,$AF,$FA,$89,$AF,$38,$E4,$47,$EF,$74,$BD,$9F,$C0
	.dc.b	$42,$30,$EA,$33,$18,$8F,$E1,$8D,$7F,$D4,$CD,$7D,$AD,$7F,$2C,$D7
	.dc.b	$B4,$74,$23,$DC,$87,$67,$2D,$32,$78,$1C,$B4,$C9,$E2,$72,$D3,$27
	.dc.b	$91,$CB,$4C,$9E,$62,$C7,$A4,$91,$C2,$34,$79,$48,$78,$CA,$78,$4A
	.dc.b	$77,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9F,$77,$E3,$3E
	.dc.b	$D9,$1F,$C1,$B5,$FB,$AD,$7E,$EB,$5F,$BA,$D7,$EE,$B5,$EC,$74,$21
	.dc.b	$D0,$A7,$42,$9D,$0A,$74,$29,$D0,$E4,$43,$91,$4E,$45,$39,$14,$E4
	.dc.b	$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$4A,$3E,$43
	.dc.b	$63,$DE,$D7,$D0,$D7,$C6,$D7,$C0,$D7,$BB,$5E,$F1,$D8,$8F,$66,$BD
	.dc.b	$9A,$F6,$6B,$D9,$AF,$66,$BD,$8E,$84,$3A,$14,$E8,$53,$A1,$4E,$85
	.dc.b	$3A,$1C,$88,$72,$29,$C8,$A7,$22,$9C,$8E,$04,$38,$14,$E0,$53,$81
	.dc.b	$A2,$1A,$29,$A3,$04,$30,$48,$FA,$84,$8F,$92,$D8,$F8,$0D,$7D,$2D
	.dc.b	$7C,$8D,$7C,$2D,$7C,$0D,$7C,$11,$D8,$8F,$66,$BD,$9A,$F6,$6B,$D9
	.dc.b	$AF,$66,$BD,$8E,$84,$3A,$14,$E8,$53,$A1,$4E,$85,$3A,$1C,$88,$72
	.dc.b	$29,$C8,$A7,$22,$9C,$8E,$07,$EF,$3C,$27,$C5,$FF,$62,$6C,$7F,$88
	.dc.b	$DA,$FF,$9A,$DA,$FF,$7A,$0E,$1F,$EF,$AD,$7F,$C6,$CD,$77,$1D,$C8
	.dc.b	$F7,$6B,$DD,$AF,$78,$E4,$48,$CC,$78,$1A,$F8,$1A,$F8,$0E,$C4,$3B
	.dc.b	$14,$EC,$F0,$8E,$D1,$93,$B1,$4E,$C5,$3B,$1D,$08,$74,$34,$F4,$72
	.dc.b	$74,$29,$D0,$A7,$43,$91,$87,$93,$82,$1C,$CF,$F1,$7F,$82,$FD,$23
	.dc.b	$FC,$3E,$D8,$FF,$30,$35,$FF,$54,$35,$FB,$CD,$79,$C7,$22,$3F,$E2
	.dc.b	$AE,$5F,$F4,$BB,$FE,$83,$3F,$CF,$87,$FA,$7F,$A3,$FD,$12,$7F,$A0
	.dc.b	$1F,$F4,$FF,$5F,$FA,$49,$FF,$51,$75,$FF,$A9,$15,$7F,$A5,$DE,$87
	.dc.b	$02,$C6,$E3,$4A,$BF,$D6,$D3,$CB,$AF,$FD,$27,$7F,$E9,$B8,$FF,$4E
	.dc.b	$9F,$EA,$93,$4E,$5A,$76,$9F,$4B,$E6,$6B,$E6,$EB,$FF,$4C,$BF,$00
	.dc.b	$FF,$58,$1E,$7B,$F2,$3F,$EB,$67,$FD,$55,$87,$FB,$16,$3B,$C6,$27
	.dc.b	$FD,$45,$1F,$EA,$E8,$E2,$EA,$BE,$54,$7C,$19,$F0,$CF,$79,$F8,$35
	.dc.b	$D4,$FA,$B5,$C9,$B1,$AC,$3E,$2D,$7C,$39,$CB,$E7,$9F,$F5,$E6,$1F
	.dc.b	$5E,$3E,$01,$F7,$4F,$F5,$6D,$7F,$B6,$7A,$3F,$D5,$11,$FE,$AC,$EB
	.dc.b	$FF,$5A,$75,$1F,$F5,$B4,$7F,$AC,$8F,$F7,$C6,$3F,$EB,$57,$27,$08
	.dc.b	$8F,$FA,$D3,$AF,$FD,$86,$F5,$3F,$D4,$9D,$1F,$E9,$67,$FD,$4D,$D7
	.dc.b	$FE,$A0,$EB,$4F,$FA,$57,$A7,$FD,$27,$18,$EB,$FF,$4C,$9F,$E9,$BB
	.dc.b	$FF,$51,$5F,$1B,$E1,$1F,$EC,$1E,$BF,$F5,$25,$70,$8F,$F5,$4C,$7F
	.dc.b	$AB,$23,$93,$63,$38,$7F,$B1,$20,$F8,$AB,$D1,$3F,$EB,$58,$D1,$9A
	.dc.b	$ED,$5F,$EA,$C9,$FF,$58,$9D,$0F,$F6,$4B,$D1,$F3,$4F,$FB,$42,$3C
	.dc.b	$47,$13,$D3,$1E,$E3,$FD,$6C,$70,$3E,$5C,$FF,$AD,$9F,$F7,$86,$1F
	.dc.b	$EC,$4A,$FF,$4F,$E5,$FE,$A6,$24,$FF,$AB,$9A,$FF,$AA,$1A,$FF,$AA
	.dc.b	$1F,$F5,$6F,$8F,$23,$FD,$6A,$D7,$FD,$5E,$D7,$FD,$5E,$D7,$FD,$5E
	.dc.b	$D7,$FD,$5E,$D7,$FD,$78,$DE,$93,$49,$1E,$02,$1E,$02,$9E,$02,$9E
	.dc.b	$02,$9E,$02,$9E,$02,$C7,$81,$B1,$E0,$7B,$10,$EC,$53,$B1,$4E,$C5
	.dc.b	$3B,$16,$3B,$36,$3B,$3D,$08,$74,$29,$D0,$FF,$6D,$F8,$FF,$EE,$02
	.dc.b	$3F,$EE,$86,$BF,$EE,$26,$BC,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$EE
	.dc.b	$A7,$AB,$4E,$45,$38,$94,$E2,$7F,$BC,$1E,$2D,$38,$94,$E2,$53,$88
	.dc.b	$94,$72,$8C,$36,$33,$1D,$0C,$95,$CB,$D8,$C9,$5C,$BE,$06,$BE,$B2
	.dc.b	$1E,$52,$9C,$8E,$11,$E3,$23,$E1,$6B,$DD,$AF,$03,$44,$34,$53,$46
	.dc.b	$08,$60,$87,$92,$3D,$2F,$9C,$87,$9C,$D4,$78,$9A,$77,$29,$D8,$A7
	.dc.b	$63,$9C,$76,$7A,$18,$8E,$8D,$3A,$14,$E8,$53,$A1,$C8,$91,$D4,$91
	.dc.b	$D2,$32,$D8,$CC,$76,$32,$57,$2F,$80,$C9,$5C,$BE,$26,$BF,$00,$87
	.dc.b	$9C,$A7,$23,$84,$78,$C8,$F8,$5A,$F7,$6B,$C0,$D1,$0D,$14,$D1,$82
	.dc.b	$18,$21,$E4,$8F,$4B,$E7,$21,$E7,$35,$1E,$26,$9D,$CA,$76,$29,$D8
	.dc.b	$E7,$1D,$9E,$86,$23,$A3,$4E,$85,$3A,$14,$E8,$72,$24,$75,$24,$7E
	.dc.b	$07,$A1,$E3,$FF,$A9,$C9,$5F,$EA,$76,$BF,$EA,$76,$BF,$E9,$96,$BF
	.dc.b	$EA,$66,$BF,$EA,$66,$BF,$EA,$C6,$C7,$82,$3B,$10,$EC,$53,$B1,$4E
	.dc.b	$C5,$3B,$14,$EC,$57,$B3,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$AF,$47
	.dc.b	$91,$0E,$45,$39,$14,$E4,$57,$93,$C0,$87,$02,$9C,$0A,$F0,$5D,$3F
	.dc.b	$EA,$3F,$A9,$FF,$50,$F5,$1A,$49,$FF,$51,$C6,$5B,$19,$7F,$D4,$B1
	.dc.b	$97,$FC,$B3,$CC,$E7,$FF,$45,$72,$FF,$9C,$63,$13,$FE,$6D,$6F,$4F
	.dc.b	$F9,$FE,$3F,$D1,$0F,$FA,$12,$3F,$CE,$6E,$63,$FC,$F8,$7F,$9C,$C1
	.dc.b	$FE,$85,$E9,$FF,$3D,$5F,$F9,$D6,$FF,$D4,$71,$89,$FF,$4D,$1F,$EA
	.dc.b	$0A,$E8,$7F,$A2,$5F,$F3,$D1,$FE,$8D,$3B,$1D,$4B,$2E,$AF,$F4,$AB
	.dc.b	$63,$D3,$5F,$E8,$67,$B9,$CC,$91,$E4,$79,$CF,$22,$3C,$8F,$6C,$75
	.dc.b	$7E,$09,$CA,$3F,$D3,$F5,$E9,$8F,$F5,$7B,$4F,$95,$7F,$0E,$77,$7F
	.dc.b	$EB,$8B,$F1,$5F,$33,$E0,$BF,$5E,$BE,$CE,$1F,$4A,$BF,$D5,$E7,$D0
	.dc.b	$AF,$5F,$57,$FA,$7A,$CD,$1D,$3B,$9F,$F4,$D9,$FE,$99,$3F,$D3,$F1
	.dc.b	$A6,$C7,$FA,$72,$79,$B6,$7F,$D5,$7C,$BF,$E9,$7A,$FF,$40,$07,$FC
	.dc.b	$E8,$FF,$9E,$0F,$F3,$E1,$FE,$86,$E7,$FF,$40,$73,$7F,$A0,$CF,$F4
	.dc.b	$6C,$7F,$9F,$CF,$F4,$81,$67,$B4,$F0,$8E,$8F,$FA,$2A,$7F,$D2,$90
	.dc.b	$8F,$F4,$83,$E3,$E8,$FF,$4D,$D7,$2A,$DD,$71,$AF,$F5,$A9,$FE,$90
	.dc.b	$38,$74,$7F,$A7,$A7,$64,$8F,$F4,$EB,$5F,$81,$D3,$FE,$B5,$9F,$F4
	.dc.b	$B4,$72,$3E,$0D,$7F,$A7,$8F,$03,$EE,$38,$9E,$1B,$F3,$3D,$8F,$85
	.dc.b	$87,$FB,$01,$E1,$1F,$39,$B1,$EE,$35,$1C,$A3,$A9,$F4,$70,$F4,$46
	.dc.b	$4F,$AF,$C6,$E2,$7D,$67,$F1,$42,$5F,$F1,$26,$BF,$C4,$1A,$FE,$4E
	.dc.b	$79,$47,$E3,$E3,$F2,$DF,$2D,$F5,$1F,$C9,$35,$CE,$1B,$7F,$85,$C6
	.dc.b	$5B,$19,$51,$96,$C6,$63,$04,$70,$48,$ED,$19,$6C,$65,$F4,$46,$5B
	.dc.b	$19,$7D,$2D,$7D,$27,$22,$1C,$0A,$70,$29,$C0,$D1,$0D,$14,$D1,$82
	.dc.b	$18,$21,$F1,$8E,$A5,$39,$14,$E2,$53,$89,$F2,$8E,$25,$38,$94,$E2
	.dc.b	$53,$8C,$7D,$0F,$8D,$FA,$44,$8F,$9E,$D7,$E6,$B5,$FC,$1B,$5F,$C3
	.dc.b	$B5,$FC,$1B,$5F,$C0,$B6,$0D,$24,$78,$08,$78,$0A,$78,$0A,$78,$0A
	.dc.b	$78,$0A,$78,$0B,$1E,$06,$C7,$81,$EC,$43,$B1,$4E,$C5,$3B,$14,$EC
	.dc.b	$58,$EC,$D8,$EC,$F4,$21,$D0,$A7,$42,$9D,$0B,$1D,$1B,$1D,$1E,$44
	.dc.b	$39,$14,$E4,$58,$E4,$D8,$E4,$F0,$21,$C0,$B1,$C1,$B1,$C1,$D1,$23
	.dc.b	$4D,$8D,$46,$08,$E0,$91,$F5,$08,$7C,$92,$9F,$00,$A7,$A4,$A7,$90
	.dc.b	$A7,$84,$B1,$E0,$6C,$78,$1E,$C4,$3B,$14,$EC,$53,$B1,$4E,$C5,$8E
	.dc.b	$CD,$8E,$CF,$42,$1D,$0A,$74,$29,$D0,$B1,$D1,$B1,$D1,$E4,$43,$91
	.dc.b	$4E,$45,$8E,$4D,$8E,$4F,$02,$1C,$0B,$1C,$1B,$1C,$1D,$12,$34,$D8
	.dc.b	$D4,$60,$8E,$09,$1F,$50,$87,$C9,$29,$F0,$0A,$7A,$4A,$79,$0A,$78
	.dc.b	$4B,$1E,$06,$C7,$81,$EC,$7F,$61,$E7,$7D,$3F,$F5,$96,$C7,$F5,$F6
	.dc.b	$BB,$8D,$11,$D3,$5D,$18,$21,$82,$1F,$E3,$F8,$EA,$D8,$E4,$D3,$89
	.dc.b	$4E,$27,$F9,$36,$38,$B6,$38,$B4,$E2,$53,$89,$FE,$84,$8E,$2D,$8E
	.dc.b	$2D,$38,$94,$E2,$23,$8B,$63,$8B,$4E,$25,$38,$C7,$02,$47,$06,$C7
	.dc.b	$06,$BC,$23,$44,$7F,$D6,$3F,$23,$FE,$B6,$23,$82,$1F,$EB,$27,$2D
	.dc.b	$3F,$C3,$9C,$6E,$5F,$F5,$7F,$2F,$F9,$EA,$7F,$D4,$4D,$7F,$D2,$B1
	.dc.b	$A7,$FD,$23,$CF,$FE,$88,$76,$60,$FF,$43,$C7,$FA,$15,$C1,$FE,$92
	.dc.b	$9F,$F4,$74,$F1,$8E,$BC,$FF,$E8,$C3,$FD,$28,$F2,$6B,$C4,$EE,$2F
	.dc.b	$FD,$2E,$7A,$E3,$94,$7F,$AB,$A3,$FD,$30,$FF,$B3,$63,$3D,$0B,$FF
	.dc.b	$61,$36,$7F,$D6,$27,$FA,$BC,$FF,$5F,$1F,$EA,$33,$C4,$7A,$0E,$84
	.dc.b	$7F,$D9,$31,$EE,$8F,$51,$E3,$C3,$E2,$3E,$9C,$3E,$16,$1E,$6B,$E3
	.dc.b	$3E,$E8,$F6,$D7,$D1,$8E,$07,$49,$F4,$3E,$57,$FD,$7A,$F7,$C3,$FD
	.dc.b	$C7,$1A,$AF,$F5,$4C,$BB,$DD,$FF,$EB,$99,$CC,$62,$7F,$D6,$8D,$7E
	.dc.b	$87,$4F,$FA,$D6,$31,$3D,$27,$FD,$78,$D7,$FD,$7A,$70,$EA,$FC,$8D
	.dc.b	$7F,$B0,$DF,$F5,$D1,$FE,$F8,$AE,$DC,$DF,$E8,$3E,$7F,$F4,$49,$FE
	.dc.b	$8B,$E9,$FF,$4D,$C7,$FA,$1F,$AB,$FD,$3D,$8F,$FA,$BF,$9B,$FD,$17
	.dc.b	$18,$EA,$FF,$58,$D6,$EB,$FD,$39,$87,$37,$81,$FE,$9C,$C3,$AC,$78
	.dc.b	$1F,$F5,$84,$7F,$AD,$0F,$F4,$C1,$6B,$FD,$55,$87,$FB,$22,$3F,$D5
	.dc.b	$D1,$FE,$AC,$BF,$55,$72,$9F,$F5,$3E,$1E,$5B,$F3,$C7,$82,$63,$E4
	.dc.b	$39,$C7,$FB,$B6,$BF,$D4,$BD,$4F,$A6,$FF,$4B,$75,$FF,$A9,$DB,$19
	.dc.b	$BF,$F5,$04,$E5,$B1,$97,$FD,$E5,$19,$6C,$65,$FF,$55,$35,$0F,$1D
	.dc.b	$FB,$2F,$F8,$A5,$BD,$7F,$EA,$96,$CF,$FA,$C5,$AF,$FC,$01,$6B,$D0
	.dc.b	$E9,$3C,$C9,$3C,$DB,$1C,$DA,$F3,$6B,$CE,$7A,$BE,$86,$BD,$A7,$A1
	.dc.b	$1E,$8D,$7A,$35,$E9,$3D,$9E,$8D,$7A,$1C,$88,$72,$29,$CA,$3A,$3C
	.dc.b	$9A,$F2,$38,$10,$E0,$72,$38,$14,$E0,$68,$E0,$68,$A6,$8C,$1B,$21
	.dc.b	$82,$1B,$21,$FE,$E0,$E7,$FF,$6F,$11,$FF,$6D,$B5,$FF,$6C,$B5,$FF
	.dc.b	$77,$B5,$FF,$79,$B5,$FF,$77,$B5,$EC,$74,$21,$D0,$A7,$42,$9D,$0A
	.dc.b	$74,$29,$D0,$E4,$43,$91,$4E,$45,$39,$14,$E4,$70,$21,$C0,$A7,$02
	.dc.b	$9C,$0D,$10,$D1,$4D,$18,$21,$82,$76,$23,$E4,$36,$3D,$ED,$7D,$0D
	.dc.b	$7C,$6D,$7C,$0D,$7B,$B5,$EF,$1D,$88,$F6,$6B,$D9,$AF,$66,$BD,$9A
	.dc.b	$F6,$6B,$D8,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$53,$A1,$C8,$87,$22
	.dc.b	$9C,$8A,$72,$29,$C8,$E0,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$30
	.dc.b	$43,$04,$8F,$A8,$48,$F9,$2D,$8F,$80,$D7,$D2,$D7,$C8,$D7,$C2,$D7
	.dc.b	$C0,$D7,$C1,$1D,$88,$F6,$6B,$D9,$AF,$66,$BD,$9A,$F6,$6B,$D8,$E8
	.dc.b	$43,$A1,$4E,$85,$3A,$14,$E8,$53,$A1,$C8,$87,$22,$9C,$8A,$72,$29
	.dc.b	$C8,$E0,$7E,$F3,$B1,$F2,$1F,$EA,$D2,$47,$FA,$B5,$AF,$FA,$CD,$AF
	.dc.b	$FA,$95,$AF,$FA,$7D,$AF,$FA,$65,$AF,$FA,$4D,$B1,$E0,$8E,$C4,$3B
	.dc.b	$14,$EC,$53,$B1,$4E,$C5,$3B,$15,$EC,$F4,$21,$D0,$A7,$42,$9D,$0A
	.dc.b	$74,$2B,$D1,$E4,$43,$91,$4E,$45,$39,$15,$E4,$F0,$21,$C0,$A7,$02
	.dc.b	$BC,$16,$6C,$DF,$51,$FE,$2C,$24,$FF,$8A,$9A,$FF,$8A,$1A,$FF,$8A
	.dc.b	$1F,$F0,$E3,$FE,$64,$EB,$FF,$3D,$74,$7F,$9F,$3A,$FF,$D2,$7D,$9F
	.dc.b	$E9,$22,$57,$FA,$48,$76,$7F,$9E,$6F,$FD,$13,$7C,$67,$FD,$21,$5F
	.dc.b	$EB,$78,$FF,$45,$36,$77,$AF,$FA,$39,$B5,$FE,$89,$9E,$EF,$FA,$25
	.dc.b	$FF,$41,$4F,$F8,$CE,$FC,$18,$7F,$AC,$9F,$F4,$69,$8A,$F0,$9F,$EB
	.dc.b	$E7,$35,$D3,$0F,$F5,$A4,$E8,$FF,$4C,$9F,$EA,$A7,$CE,$F1,$8F,$0B
	.dc.b	$FE,$B1,$8F,$EB,$13,$DC,$93,$ED,$AF,$23,$C2,$BE,$3B,$FE,$B1,$7B
	.dc.b	$61,$F0,$1B,$3C,$1F,$01,$FE,$AF,$C3,$E3,$46,$63,$FD,$83,$87,$AB
	.dc.b	$0F,$F5,$AC,$7F,$B0,$5B,$1E,$A8,$F6,$C7,$C2,$7F,$D5,$3D,$9F,$EA
	.dc.b	$DE,$CF,$F5,$04,$7F,$AC,$63,$FD,$4B,$5F,$EA,$78,$FF,$36,$B5,$FF
	.dc.b	$54,$1F,$8A,$2F,$5F,$FA,$C2,$B9,$BF,$EA,$8E,$BF,$F4,$C4,$FF,$A2
	.dc.b	$79,$FF,$D4,$0F,$FA,$60,$FF,$48,$9F,$EA,$2E,$9F,$F4,$A4,$7F,$A5
	.dc.b	$A3,$87,$57,$FA,$55,$FF,$52,$1C,$EB,$AF,$5F,$FA,$75,$BD,$7F,$EA
	.dc.b	$1B,$EF,$3F,$EA,$4A,$FF,$43,$C2,$FA,$47,$FA,$70,$CB,$88,$E3,$5F
	.dc.b	$E9,$FB,$EB,$5E,$2B,$FF,$5D,$C6,$21,$5D,$5F,$61,$EA,$3F,$D2,$F2
	.dc.b	$AF,$F4,$17,$A4,$CF,$FD,$2C,$DA,$FF,$4B,$35,$FF,$4B,$35,$FF,$46
	.dc.b	$35,$E7,$1C,$88,$FD,$71,$D0,$F4,$B1,$18,$6C,$66,$38,$35,$FC,$03
	.dc.b	$5F,$E1,$6D,$7F,$D5,$0D,$7F,$D6,$AD,$7B,$1D,$23,$B9,$1C,$10,$EE
	.dc.b	$E5,$A6,$4F,$0B,$96,$99,$3C,$6E,$5A,$64,$F2,$B9,$69,$93,$CE,$78
	.dc.b	$08,$76,$29,$D0,$A7,$22,$9C,$8E,$04,$38,$14,$E0,$53,$81,$A2,$1A
	.dc.b	$29,$A3,$04,$30,$4E,$47,$83,$F8,$06,$C7,$E1,$5A,$FE,$F5,$AF,$F4
	.dc.b	$86,$BF,$EE,$56,$BF,$EF,$26,$A6,$92,$3C,$04,$8F,$03,$5F,$03,$5F
	.dc.b	$03,$5F,$03,$5F,$03,$5F,$03,$63,$C1,$1D,$88,$76,$29,$D8,$A7,$62
	.dc.b	$9D,$8A,$76,$2B,$D9,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8
	.dc.b	$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2
	.dc.b	$BA,$70,$47,$04,$8F,$A8,$48,$F9,$2D,$7E,$03,$5F,$4B,$5F,$23,$5F
	.dc.b	$0B,$5F,$03,$63,$C1,$1D,$88,$76,$29,$D8,$A7,$62,$9D,$8A,$76,$2B
	.dc.b	$D9,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8,$87,$22,$9C,$8A
	.dc.b	$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA,$70,$47,$04
	.dc.b	$8F,$A8,$48,$F9,$2D,$7E,$03,$5F,$4B,$5F,$23,$5F,$0B,$5F,$03,$63
	.dc.b	$C1,$1D,$88,$76,$29,$D8,$A7,$62,$9D,$8A,$3E,$25,$EC,$FF,$6A,$6C
	.dc.b	$7F,$6F,$6B,$FE,$27,$6B,$FE,$72,$6A,$23,$86,$C7,$52,$3D,$5A,$F5
	.dc.b	$6B,$D5,$AF,$53,$D0,$E1,$EA,$73,$21,$CC,$A7,$32,$9C,$E3,$CC,$F4
	.dc.b	$32,$74,$39,$10,$E4,$53,$94,$78,$DE,$46,$4E,$47,$02,$1C,$0E,$E7
	.dc.b	$03,$27,$03,$47,$33,$46,$0D,$98,$38,$10,$C1,$23,$DE,$7D,$5E,$AF
	.dc.b	$F5,$A5,$7F,$8B,$5B,$3F,$E6,$56,$BF,$EB,$06,$BF,$4D,$AF,$E0,$5A
	.dc.b	$FF,$B0,$3A,$BF,$B5,$3C,$DA,$FF,$9C,$79,$3F,$D0,$55,$FE,$91,$8F
	.dc.b	$F4,$79,$FE,$B7,$E6,$FF,$4A,$B6,$3F,$D2,$51,$FE,$83,$8F,$F5,$71
	.dc.b	$FE,$90,$3F,$D1,$51,$CE,$38,$C7,$FA,$2A,$3F,$D3,$91,$89,$F0,$CF
	.dc.b	$17,$FD,$4B,$3D,$8D,$96,$35,$D3,$FE,$A1,$AF,$15,$72,$9E,$C6,$BA
	.dc.b	$3F,$D4,$11,$D2,$3F,$D4,$87,$FB,$1C,$F6,$18,$3F,$D5,$67,$C2,$3D
	.dc.b	$A7,$FA,$A2,$BF,$DB,$FD,$3F,$EA,$C8,$D4,$F8,$4F,$F6,$D5,$FF,$AA
	.dc.b	$AB,$CD,$87,$AA,$BB,$4F,$A2,$79,$C7,$8C,$EF,$5E,$77,$E5,$9F,$16
	.dc.b	$FF,$DB,$57,$F0,$E3,$C9,$87,$FB,$C1,$F5,$9C,$0F,$86,$68,$FF,$6C
	.dc.b	$57,$52,$5F,$FB,$EB,$83,$FD,$63,$D3,$FE,$BD,$F0,$7F,$86,$75,$7F
	.dc.b	$AD,$5F,$F4,$15,$FF,$AE,$C9,$48,$CB,$CB,$0F,$F6,$39,$FE,$A8,$E8
	.dc.b	$FF,$4F,$74,$7F,$A9,$23,$1D,$5F,$E9,$CE,$8F,$F5,$33,$63,$FD,$4D
	.dc.b	$18,$7F,$D7,$11,$FE,$BE,$AE,$7D,$5F,$EA,$F5,$19,$39,$4F,$FA,$D4
	.dc.b	$FF,$56,$CE,$E3,$51,$D0,$FF,$57,$1F,$EC,$DB,$FF,$65,$75,$7F,$B1
	.dc.b	$67,$9C,$7A,$67,$A3,$AC,$3F,$D9,$18,$78,$A7,$9D,$FF,$B2,$5F,$09
	.dc.b	$FE,$BC,$7A,$3E,$16,$9F,$EC,$4B,$F2,$B6,$39,$D7,$C3,$8C,$11,$F0
	.dc.b	$CF,$07,$47,$F3,$9B,$FF,$66,$52,$BF,$D9,$4F,$F0,$7F,$A8,$FE,$25
	.dc.b	$C8,$69,$23,$F8,$53,$5F,$E2,$91,$B7,$27,$FB,$F7,$C1,$7D,$F3,$3F
	.dc.b	$F6,$54,$61,$B1,$99,$E0,$D7,$FA,$80,$9E,$0E,$A7,$89,$1C,$10,$E3
	.dc.b	$1C,$1B,$3D,$5F,$EB,$91,$96,$C6,$5F,$ED,$91,$96,$C6,$5F,$23,$63
	.dc.b	$51,$89,$F2,$3E,$06,$BD,$9A,$F4,$6B,$C4,$F5,$1B,$29,$B2,$9B,$3B
	.dc.b	$18,$8F,$19,$27,$CC,$49,$EF,$19,$6C,$66,$79,$35,$EA,$43,$C8,$53
	.dc.b	$C8,$53,$C8,$7F,$B0,$7C,$4F,$F6,$41,$1F,$F6,$43,$5F,$F6,$43,$5E
	.dc.b	$06,$88,$68,$A6,$8C,$10,$C1,$0F,$F6,$83,$D5,$A7,$22,$9C,$4A,$71
	.dc.b	$3F,$DA,$CF,$16,$9C,$4A,$71,$29,$C4,$FF,$6E,$3C,$5A,$71,$29,$C4
	.dc.b	$A7,$1E,$54,$71,$6C,$71,$6B,$C5,$AF,$18,$E0,$47,$83,$5E,$0D,$78
	.dc.b	$1A,$21,$A2,$9A,$30,$43,$04,$3E,$59,$0F,$8A,$53,$DE,$58,$F4,$B6
	.dc.b	$3C,$8D,$3C,$25,$3C,$05,$3C,$07,$62,$1D,$8B,$1D,$9B,$1D,$9A,$76
	.dc.b	$29,$D8,$A7,$63,$A1,$23,$A3,$63,$A3,$4E,$85,$3A,$14,$E9,$1C,$89
	.dc.b	$1C,$9B,$1C,$9A,$F2,$6B,$CA,$38,$11,$E0,$D7,$83,$5E,$06,$88,$68
	.dc.b	$A6,$8C,$10,$C1,$0F,$A8,$43,$E4,$94,$F8,$05,$8F,$4B,$63,$C8,$D3
	.dc.b	$C2,$53,$C0,$53,$C0,$76,$21,$D8,$B1,$D9,$B1,$D9,$A7,$62,$9D,$8A
	.dc.b	$76,$3A,$12,$3A,$36,$3A,$3F,$86,$E9,$7D,$3F,$FB,$04,$91,$FE,$C1
	.dc.b	$6B,$FE,$C1,$6B,$FE,$C1,$6B,$FE,$AB,$6B,$FE,$C2,$6B,$FE,$CF,$6C
	.dc.b	$78,$23,$B1,$0E,$C5,$3B,$14,$EC,$53,$B1,$4E,$C5,$7B,$3D,$08,$74
	.dc.b	$29,$D0,$A7,$42,$9D,$0A,$F4,$79,$10,$E4,$53,$91,$4E,$45,$79,$3C
	.dc.b	$08,$70,$29,$C0,$AF,$07,$FD,$5D,$F5,$7F,$E2,$82,$3F,$EA,$F6,$BF
	.dc.b	$E2,$66,$BF,$EB,$9C,$D1,$FE,$1E,$8F,$F7,$77,$27,$FA,$2F,$A7,$FD
	.dc.b	$1D,$D5,$FE,$A7,$6E,$7A,$E9,$FF,$44,$DF,$FA,$62,$B4,$FF,$A2,$EF
	.dc.b	$93,$67,$73,$FE,$91,$EA,$FF,$44,$4F,$FA,$43,$0F,$F4,$6B,$FE,$96
	.dc.b	$C3,$FD,$30,$FF,$A2,$F0,$FF,$4E,$C7,$FA,$F4,$F1,$57,$28,$EF,$3E
	.dc.b	$39,$CB,$FE,$90,$AF,$4B,$FE,$96,$9F,$F5,$46,$1E,$4A,$EF,$18,$BE
	.dc.b	$93,$9A,$FF,$5E,$1F,$EB,$07,$0F,$FA,$77,$0F,$F5,$84,$FF,$B0,$30
	.dc.b	$F5,$CF,$A5,$F0,$BE,$F0,$73,$8E,$0E,$27,$FD,$AD,$5D,$AF,$FD,$4F
	.dc.b	$5A,$AE,$04,$AF,$0B,$70,$F8,$CF,$FA,$88,$53,$9F,$FD,$4D,$18,$7F
	.dc.b	$D5,$4F,$FB,$D5,$AF,$F9,$8A,$3F,$D5,$FD,$1F,$EA,$E3,$FD,$6B,$D5
	.dc.b	$FE,$B7,$97,$57,$FA,$DD,$B1,$98,$3F,$E8,$48,$FF,$58,$9F,$E9,$1E
	.dc.b	$8F,$F4,$4B,$FE,$8E,$33,$CF,$FE,$8D,$7F,$D6,$66,$23,$93,$FE,$9C
	.dc.b	$3F,$D5,$1C,$BF,$EB,$47,$FD,$3D,$1D,$27,$FD,$3A,$F0,$E8,$FF,$49
	.dc.b	$C7,$FA,$95,$B1,$9A,$E5,$5D,$3A,$7F,$D6,$31,$FE,$AD,$BE,$73,$E4
	.dc.b	$3F,$D6,$71,$9A,$F0,$12,$3E,$FD,$F6,$7D,$44,$3F,$D7,$7D,$5F,$EC
	.dc.b	$57,$E1,$1D,$30,$F0,$9F,$EA,$D7,$C1,$3F,$EC,$17,$AC,$FE,$31,$F4
	.dc.b	$1F,$34,$C9,$F1,$0A,$7F,$AC,$EB,$F2,$74,$F5,$5E,$57,$F0,$5A,$FF
	.dc.b	$6E,$C6,$1F,$E1,$13,$B9,$D3,$63,$33,$B7,$FD,$57,$19,$6C,$65,$F4
	.dc.b	$46,$5F,$F5,$2F,$EB,$EF,$1B,$FD,$72,$D8,$FF,$5F,$35,$FF,$57,$B5
	.dc.b	$E3,$1C,$08,$F0,$6B,$C1,$AF,$03,$44,$34,$53,$46,$08,$60,$87,$FA
	.dc.b	$EA,$3C,$8D,$8E,$ED,$3A,$14,$E6,$53,$99,$FE,$D3,$8E,$6D,$8E,$6D
	.dc.b	$39,$94,$E6,$53,$9F,$6A,$79,$B6,$79,$B6,$79,$B5,$E6,$D7,$9C,$F2
	.dc.b	$24,$F2,$6C,$72,$7D,$2F,$F1,$4C,$7F,$1C,$D3,$F2,$1C,$28,$CB,$63
	.dc.b	$31,$82,$38,$21,$FD,$76,$38,$36,$36,$D3,$67,$F1,$F7,$2D,$32,$7F
	.dc.b	$66,$72,$D3,$27,$F3,$C8,$EE,$F1,$3B,$10,$C1,$0E,$46,$4A,$64,$E9
	.dc.b	$1E,$06,$9E,$03,$E4,$1C,$4C,$94,$C9,$CC,$A7,$62,$1C,$0D,$47,$A9
	.dc.b	$C9,$C0,$D1,$87,$4E,$08,$6C,$8E,$9B,$1E,$B6,$9E,$B8,$D1,$23,$4D
	.dc.b	$8D,$46,$08,$E0,$87,$C1,$8E,$0D,$8D,$B4,$D9,$F0,$1C,$B4,$C9,$EF
	.dc.b	$72,$D3,$27,$C0,$8E,$EF,$13,$B1,$0C,$10,$E4,$64,$A6,$4E,$91,$E0
	.dc.b	$69,$E0,$3E,$49,$C4,$C9,$4C,$9C,$CA,$76,$21,$C0,$D4,$7A,$9C,$9C
	.dc.b	$0D,$18,$74,$E0,$86,$C8,$E9,$B1,$EB,$69,$EB,$8D,$12,$34,$D8,$D4
	.dc.b	$60,$8E,$08,$7C,$18,$E0,$D8,$DB,$4D,$9F,$01,$CB,$4C,$9E,$F7,$2D
	.dc.b	$32,$7C,$08,$EE,$F1,$3B,$10,$C1,$0E,$46,$4A,$64,$E9,$1E,$07,$B7
	.dc.b	$8F,$FC,$90,$8F,$F2,$86,$BF,$C8,$5A,$FF,$8E,$DA,$FF,$A0,$1A,$FF
	.dc.b	$19,$6B,$FE,$48,$6C,$9A,$48,$F0,$10,$F0,$14,$F0,$14,$F0,$14,$F0
	.dc.b	$14,$F0,$16,$3C,$0D,$8F,$03,$D8,$87,$62,$9D,$8A,$76,$29,$D8,$B1
	.dc.b	$D9,$B1,$D9,$E8,$43,$A1,$4F,$F4,$67,$3B,$9F,$FC,$ED,$CF,$FE,$7A
	.dc.b	$7F,$D1,$64,$9F,$F4,$A1,$FE,$A5,$71,$1F,$E7,$A3,$FD,$57,$1F,$E9
	.dc.b	$B3,$AC,$74,$E7,$FF,$4E,$CE,$8F,$F5,$64,$F4,$38,$1F,$E8,$F3,$FD
	.dc.b	$08,$72,$29,$FE,$84,$E8,$FF,$4A,$E3,$FE,$AA,$9E,$67,$8E,$BB,$9F
	.dc.b	$E8,$CA,$E9,$1E,$D7,$C0,$F0,$8D,$4F,$37,$FD,$5B,$87,$C5,$AF,$0C
	.dc.b	$FF,$AF,$1F,$ED,$17,$E2,$9F,$F5,$03,$FE,$AA,$9F,$F5,$A4,$F6,$AE
	.dc.b	$78,$7C,$38,$FE,$E1,$7D,$70,$57,$FA,$47,$96,$57,$FA,$8F,$0F,$F5
	.dc.b	$17,$3F,$FA,$6A,$3F,$BC,$74,$7F,$A8,$1F,$F4,$E1,$FE,$A1,$70,$EA
	.dc.b	$3F,$D4,$4F,$FA,$78,$E9,$1F,$EA,$AE,$8F,$F5,$11,$FE,$A0,$8F,$F3
	.dc.b	$F7,$37,$F9,$DD,$FF,$4C,$74,$7F,$A4,$79,$FF,$D0,$0F,$FA,$32,$3F
	.dc.b	$D0,$67,$F9,$DC,$DF,$0F,$FA,$57,$97,$FD,$07,$1F,$E8,$6B,$E8,$3A
	.dc.b	$3F,$D2,$51,$FE,$7E,$BE,$CF,$39,$FF,$49,$3F,$E8,$D3,$FD,$0C,$76
	.dc.b	$3C,$F3,$FE,$A2,$9E,$31,$FE,$8C,$7F,$D6,$8E,$CE,$4E,$27,$A9,$E4
	.dc.b	$9F,$1C,$EE,$31,$5F,$E9,$B8,$F7,$9D,$6F,$FD,$6D,$87,$FA,$6F,$0F
	.dc.b	$09,$23,$E1,$4F,$A2,$FB,$E1,$D9,$F8,$77,$C6,$B7,$7E,$58,$C5,$FF
	.dc.b	$A7,$3C,$AF,$F4,$CF,$41,$A4,$8F,$F4,$DB,$5F,$F4,$CB,$5F,$F5,$87
	.dc.b	$43,$F6,$07,$87,$FE,$A9,$8C,$36,$33,$3F,$EB,$70,$27,$6E,$4C,$47
	.dc.b	$07,$84,$69,$B1,$97,$C6,$D7,$AC,$6D,$CC,$F8,$C8,$F7,$8D,$3C,$27
	.dc.b	$93,$E2,$9D,$BE,$06,$BD,$0E,$51,$C2,$79,$C7,$13,$33,$C1,$EC,$D7
	.dc.b	$47,$53,$64,$30,$43,$C0,$53,$99,$B3,$24,$32,$6C,$EE,$64,$A6,$8C
	.dc.b	$47,$A8,$93,$ED,$24,$FA,$63,$2D,$8C,$CF,$FA,$73,$E1,$3F,$D3,$64
	.dc.b	$70,$43,$FD,$30,$E5,$A6,$4F,$C4,$71,$23,$51,$86,$BF,$8C,$11,$B7
	.dc.b	$91,$97,$0F,$62,$1C,$CD,$9F,$E9,$D8,$EA,$D3,$27,$67,$2D,$33,$1D
	.dc.b	$9E,$87,$63,$C5,$1C,$23,$51,$83,$8C,$64,$8E,$08,$73,$3C,$85,$3C
	.dc.b	$45,$8E,$AE,$8F,$19,$0F,$01,$B8,$EE,$D3,$B9,$B3,$91,$C0,$C1,$E0
	.dc.b	$2C,$78,$1C,$1B,$21,$C8,$86,$8A,$F1,$24,$79,$23,$AB,$5E,$67,$12
	.dc.b	$47,$07,$D2,$69,$C3,$ED,$21,$D0,$D9,$E2,$8E,$CD,$32,$76,$72,$D3
	.dc.b	$31,$D9,$E8,$76,$3C,$51,$C2,$35,$18,$38,$C6,$48,$E0,$87,$33,$C8
	.dc.b	$53,$C4,$58,$EA,$E8,$F1,$90,$F0,$1B,$8E,$ED,$3B,$9B,$39,$1C,$0C
	.dc.b	$1E,$02,$C7,$81,$C1,$B2,$1C,$88,$68,$AF,$12,$47,$92,$3A,$B5,$E6
	.dc.b	$71,$24,$70,$7D,$26,$9C,$3E,$D2,$1D,$0D,$9E,$28,$EC,$D3,$27,$67
	.dc.b	$2D,$33,$1D,$9E,$87,$63,$C5,$1C,$23,$51,$E5,$E8,$FE,$C4,$FF,$91
	.dc.b	$8B,$08,$CC,$7F,$96,$CC,$B8,$8D,$46,$08,$7F,$6D,$72,$D8,$E0,$47
	.dc.b	$67,$F3,$07,$0E,$CC,$9F,$CA,$DC,$BB,$32,$7A,$4E,$04,$34,$53,$91
	.dc.b	$C0,$F6,$19,$39,$1C,$08,$70,$34,$70,$32,$70,$34,$60,$D1,$82,$1B
	.dc.b	$21,$A2,$C7,$C8,$79,$35,$E2,$6C,$8E,$5E,$31,$F5,$79,$FF,$CF,$F1
	.dc.b	$FE,$22,$E7,$FF,$46,$1F,$E9,$17,$FD,$30,$FF,$A7,$F9,$FF,$D0,$31
	.dc.b	$FE,$8A,$AF,$F4,$47,$47,$FA,$AA,$7F,$D2,$37,$D2,$BF,$D1,$ED,$BE
	.dc.b	$B3,$AA,$E1,$5F,$56,$7F,$D3,$4D,$9E,$F1,$E1,$8C,$1F,$E9,$09,$FF
	.dc.b	$49,$3F,$DB,$AB,$FD,$25,$5F,$EA,$88,$D9,$F6,$AB,$C2,$E3,$0F,$C8
	.dc.b	$C6,$23,$53,$E0,$39,$38,$38,$57,$67,$E2,$C6,$AB,$FD,$59,$1B,$23
	.dc.b	$FE,$AA,$AF,$99,$1F,$EB,$A9,$E1,$5F,$15,$B7,$F1,$27,$8E,$1F,$EB
	.dc.b	$3A,$F0,$47,$13,$E1,$1F,$4C,$FF,$6E,$1F,$EB,$73,$CF,$3F,$76,$BF
	.dc.b	$D6,$71,$FE,$14,$E6,$FF,$54,$74,$63,$2F,$F5,$AC,$FF,$AB,$4D,$74
	.dc.b	$7F,$AC,$1F,$F5,$A3,$5E,$15,$C9,$FF,$59,$C7,$FA,$F5,$FF,$4A,$F2
	.dc.b	$7F,$A3,$1F,$F5,$17,$1F,$FA,$C2,$3F,$C4,$DC,$BF,$E8,$AE,$7F,$F5
	.dc.b	$6F,$37,$FA,$1E,$F6,$43,$FD,$2C,$7F,$AF,$2F,$93,$FE,$AF,$BF,$F4
	.dc.b	$79,$DD,$D9,$FE,$9E,$9F,$F4,$D3,$FE,$AE,$AE,$F3,$E2,$AF,$F4,$AC
	.dc.b	$FA,$0E,$C7,$86,$B8,$B9,$3A,$4F,$FA,$2A,$9F,$14,$F8,$FF,$E5,$31
	.dc.b	$86,$C6,$6B,$FD,$1D,$39,$6C,$65,$FF,$4B,$46,$5B,$19,$7F,$C9,$D1
	.dc.b	$97,$FD,$23,$E1,$7F,$A7,$C8,$FF,$A1,$1A,$FF,$A6,$1A,$F0,$34,$43
	.dc.b	$45,$34,$60,$86,$08,$7F,$A6,$5E,$AD,$39,$14,$E2,$53,$89,$FE,$B0
	.dc.b	$78,$B4,$E2,$53,$89,$4E,$3C,$FF,$EC,$38,$E2,$D8,$E2,$D7,$8B,$5E
	.dc.b	$3C,$C9,$E2,$D8,$E2,$D7,$8B,$5E,$31,$C0,$8F,$07,$FB,$27,$C9,$48
	.dc.b	$FE,$D6,$D7,$FB,$23,$5F,$F4,$7B,$5F,$F4,$43,$5F,$F6,$CF,$02,$31
	.dc.b	$1F,$E9,$06,$C6,$63,$B9,$0E,$E5,$3B,$94,$EE,$53,$BC,$74,$23,$93
	.dc.b	$C0,$58,$F0,$3D,$88,$76,$29,$D8,$A7,$63,$94,$76,$72,$76,$2C,$76
	.dc.b	$7A,$10,$E8,$53,$A1,$C2,$3A,$39,$3A,$16,$3A,$3C,$88,$72,$35,$1C
	.dc.b	$9C,$9C,$8B,$1C,$9E,$06,$23,$83,$82,$1C,$4B,$1C,$5C,$11,$C1,$23
	.dc.b	$9C,$65,$B1,$98,$F9,$64,$3E,$11,$4F,$59,$4F,$31,$4F,$1C,$74,$23
	.dc.b	$93,$C0,$58,$F0,$3D,$88,$76,$29,$D8,$A7,$63,$94,$76,$72,$76,$2C
	.dc.b	$76,$7A,$10,$E8,$53,$A1,$C2,$3A,$39,$3A,$16,$3A,$3C,$88,$72,$35
	.dc.b	$1C,$9C,$9C,$8B,$1C,$9E,$06,$23,$83,$82,$1C,$4B,$1C,$5C,$11,$C1
	.dc.b	$23,$9C,$65,$B1,$98,$F9,$64,$3E,$11,$4F,$59,$4F,$31,$4F,$1C,$74
	.dc.b	$23,$93,$C0,$58,$F0,$3D,$88,$76,$29,$FE,$29,$E5,$FF,$25,$BF,$E1
	.dc.b	$93,$FC,$53,$CB,$FE,$20,$93,$FE,$27,$8F,$F1,$0F,$37,$F9,$15,$DF
	.dc.b	$37,$F9,$12,$BF,$C6,$1C,$DF,$E4,$F9,$E9,$1F,$E4,$08,$FF,$17,$C6
	.dc.b	$30,$FF,$17,$57,$1C,$3C,$07,$79,$FF,$27,$46,$48,$FF,$8C,$46,$1F
	.dc.b	$E5,$A7,$FC,$72,$7F,$8C,$4F,$3D,$FB,$DF,$F1,$BC,$7F,$91,$30,$F6
	.dc.b	$D7,$F9,$05,$FF,$25,$93,$0F,$8B,$3B,$C3,$C9,$3E,$EB,$FF,$2E,$57
	.dc.b	$A9,$F2,$57,$8E,$79,$47,$83,$0F,$89,$3C,$F0,$F0,$5F,$69,$C9,$2B
	.dc.b	$A9,$EB,$3E,$19,$D9,$FF,$34,$3F,$E6,$42,$1F,$E5,$73,$75,$F5,$EF
	.dc.b	$E9,$47,$C4,$36,$FC,$27,$EA,$1D,$DF,$73,$F3,$0F,$F4,$6D,$E6,$D1
	.dc.b	$4E,$18,$7E,$27,$97,$FC,$AD,$1F,$E5,$C3,$FC,$BD,$7F,$E6,$0E,$C7
	.dc.b	$3F,$F9,$73,$9F,$FC,$BD,$5F,$DB,$AD,$18,$8E,$55,$C1,$FF,$30,$9D
	.dc.b	$27,$BF,$3B,$B1,$C7,$FE,$50,$AF,$F2,$69,$FE,$56,$9D,$47,$AB,$9D
	.dc.b	$CF,$FE,$55,$AF,$F2,$93,$FE,$5E,$78,$C7,$F9,$52,$B8,$C2,$38,$BF
	.dc.b	$E5,$F3,$13,$CD,$F0,$5F,$91,$C9,$FE,$6B,$38,$4F,$F9,$CD,$DD,$F9
	.dc.b	$6B,$94,$F8,$6B,$93,$E0,$AF,$F3,$35,$FF,$9F,$5E,$A6,$3A,$3F,$CF
	.dc.b	$2A,$35,$3C,$5F,$5B,$E1,$3B,$61,$F2,$67,$FC,$E2,$79,$B0,$FF,$40
	.dc.b	$4E,$A3,$BD,$72,$9C,$9F,$26,$38,$9E,$6C,$3D,$31,$EC,$35,$3C,$67
	.dc.b	$E2,$1B,$BF,$84,$FC,$7E,$67,$73,$BF,$F2,$73,$04,$EA,$31,$1F,$98
	.dc.b	$8C,$38,$6B,$C5,$AF,$46,$BC,$67,$C0,$F1,$6B,$DB,$83,$EA,$3E,$36
	.dc.b	$BE,$66,$BF,$E4,$CC,$7C,$AD,$2C,$69,$34,$91,$C2,$32,$D8,$CC,$72
	.dc.b	$8C,$11,$C1,$0F,$43,$96,$99,$3D,$4E,$5A,$64,$F6,$39,$69,$93,$D8
	.dc.b	$74,$21,$C8,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$21,$82,$1F,$18,$F2
	.dc.b	$14,$EE,$53,$A1,$4E,$65,$39,$C3,$BF,$E8,$12,$3E,$CB,$5F,$C2,$35
	.dc.b	$FC,$23,$5F,$C2,$35,$FF,$47,$B5,$FC,$0B,$63,$C1,$1D,$88,$76,$29
	.dc.b	$D8,$A7,$62,$9D,$8A,$76,$2B,$D9,$E8,$43,$A1,$4E,$85,$3A,$14,$E8
	.dc.b	$57,$A3,$C8,$87,$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78
	.dc.b	$3A,$21,$A2,$BA,$70,$47,$04,$8F,$A8,$48,$F9,$2D,$7E,$03,$5F,$4B
	.dc.b	$5F,$23,$5F,$0B,$5F,$03,$63,$C1,$1D,$88,$76,$29,$D8,$A7,$62,$9D
	.dc.b	$8A,$76,$2B,$D9,$E8,$43,$A1,$4E,$85,$3A,$14,$E8,$57,$A3,$C8,$87
	.dc.b	$22,$9C,$8A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA
	.dc.b	$70,$47,$04,$8F,$A8,$48,$F9,$2D,$7E,$03,$5F,$4B,$5F,$23,$5F,$0B
	.dc.b	$5F,$03,$63,$C1,$1D,$88,$76,$29,$D8,$A7,$62,$9D,$8A,$76,$2B,$D9
	.dc.b	$E8,$43,$A1,$4E,$87,$DD,$E2,$FF,$33,$9F,$E7,$CE,$4F,$F2,$FC,$7F
	.dc.b	$98,$CF,$F2,$E9,$FE,$4C,$3F,$CA,$25,$E5,$FF,$33,$38,$5C,$DF,$E8
	.dc.b	$A7,$24,$7F,$CD,$11,$FE,$57,$8D,$BF,$E5,$F3,$FC,$C9,$7F,$E5,$C3
	.dc.b	$FC,$EA,$FF,$A2,$A7,$FC,$D6,$EA,$FD,$0F,$86,$3C,$07,$69,$E3,$5F
	.dc.b	$E7,$FE,$6F,$F3,$D5,$7B,$23,$C2,$FF,$9C,$AB,$FD,$24,$F7,$3F,$CD
	.dc.b	$35,$FE,$74,$74,$7F,$9D,$2F,$E4,$E1,$FE,$A0,$BF,$F3,$D6,$1F,$E8
	.dc.b	$F6,$D7,$F9,$C2,$75,$87,$69,$F2,$4F,$CF,$3E,$24,$7A,$9F,$33,$5C
	.dc.b	$9F,$1F,$0F,$F4,$24,$FC,$D8,$D9,$9C,$3C,$AF,$FA,$4C,$F5,$E1,$E5
	.dc.b	$BF,$11,$F3,$2F,$9B,$F2,$63,$E3,$CF,$80,$F8,$27,$D1,$C3,$D6,$FF
	.dc.b	$A5,$B8,$FF,$D1,$77,$FE,$89,$E6,$FF,$44,$4F,$FA,$07,$9B,$FD,$0D
	.dc.b	$2A,$FF,$44,$CF,$F8,$CD,$AF,$F8,$CD,$E6,$49,$FF,$4D,$F3,$7F,$A3
	.dc.b	$67,$FD,$1F,$7F,$E5,$4E,$4F,$F2,$BC,$7F,$95,$39,$3F,$CA,$CF,$F9
	.dc.b	$B4,$FF,$43,$9F,$E6,$53,$FD,$0F,$C1,$C7,$93,$FC,$B3,$1F,$E7,$1E
	.dc.b	$6F,$F3,$9D,$7F,$9C,$AB,$FC,$C3,$18,$24,$F4,$3F,$CC,$47,$F9,$B9
	.dc.b	$FF,$2F,$F3,$7F,$9C,$4F,$F3,$D6,$1E,$69,$FF,$37,$DF,$98,$91,$97
	.dc.b	$C1,$1E,$73,$C8,$7F,$9B,$B9,$5D,$2E,$DF,$F2,$AC,$61,$B1,$99,$FF
	.dc.b	$35,$35,$FF,$35,$35,$FF,$35,$35,$FF,$39,$35,$FF,$3B,$B5,$FF,$2F
	.dc.b	$F6,$FF,$96,$C8,$FF,$99,$1A,$FF,$9B,$DA,$FF,$9B,$DA,$FF,$9B,$DA
	.dc.b	$FE,$C1,$AF,$63,$A1,$0E,$85,$3A,$14,$E8,$53,$A1,$4E,$87,$22,$1C
	.dc.b	$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0,$68,$86,$8A,$68
	.dc.b	$C1,$0C,$13,$85,$1F,$21,$B1,$EF,$6B,$E8,$5C,$7F,$E9,$32,$47,$FA
	.dc.b	$71,$AF,$FA,$9D,$AF,$FA,$9D,$AF,$FA,$9D,$AF,$F0,$E3,$91,$0E,$45
	.dc.b	$39,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60
	.dc.b	$86,$08,$7F,$AF,$1F,$6B,$4F,$31,$4F,$09,$4E,$C5,$3A,$94,$EB,$1F
	.dc.b	$49,$E6,$D7,$9B,$5E,$6D,$79,$B5,$E7,$1E,$68,$C4,$77,$23,$82,$46
	.dc.b	$C9,$1E,$26,$BE,$26,$BE,$26,$BE,$26,$BE,$23,$91,$0E,$45,$39,$14
	.dc.b	$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08
	.dc.b	$7C,$87,$DA,$D3,$CC,$53,$C2,$53,$B1,$4E,$A5,$3A,$C7,$D2,$79,$B5
	.dc.b	$E6,$D7,$9B,$5E,$6D,$79,$C7,$9A,$31,$1D,$C8,$E0,$91,$B2,$47,$89
	.dc.b	$AF,$89,$AF,$89,$AF,$89,$AF,$88,$E4,$43,$91,$4E,$45,$39,$14,$E4
	.dc.b	$70,$21,$C0,$A7,$02,$9C,$0D,$10,$D1,$4D,$18,$3F,$C4,$67,$F1,$0E
	.dc.b	$6F,$F2,$2C,$FF,$94,$E7,$FC,$46,$FF,$93,$39,$7F,$C5,$1C,$DF,$E2
	.dc.b	$EA,$FF,$1F,$DF,$F9,$96,$7F,$C5,$4F,$F8,$F6,$B8,$3D,$63,$FC,$79
	.dc.b	$18,$8F,$E1,$53,$E0,$AF,$F2,$EB,$FE,$3D,$8F,$F1,$94,$F8,$DE,$97
	.dc.b	$E4,$23,$82,$1F,$E5,$B3,$AD,$F8,$23,$71,$99,$FF,$2F,$1D,$5C,$47
	.dc.b	$F9,$1D,$FF,$30,$73,$7F,$A1,$9C,$9F,$E5,$D3,$FC,$9C,$76,$3F,$CA
	.dc.b	$4E,$EF,$FC,$BF,$1C,$0F,$8C,$7F,$96,$0F,$F2,$6D,$7F,$94,$9C,$3E
	.dc.b	$38,$E2,$4B,$FA,$35,$D5,$B5,$FE,$83,$3C,$25,$3A,$E1,$F5,$E3,$EF
	.dc.b	$C7,$9C,$CC,$9E,$17,$E3,$7F,$C9,$E7,$FA,$13,$9B,$FC,$A8,$FF,$97
	.dc.b	$48,$7F,$96,$79,$BF,$CB,$F1,$FE,$5C,$8F,$F3,$1F,$37,$F9,$66,$BF
	.dc.b	$CC,$D1,$CE,$BF,$CB,$BC,$DE,$18,$4B,$26,$39,$9D,$4E,$63,$11,$FE
	.dc.b	$98,$9D,$B8,$9D,$36,$32,$F3,$8C,$B6,$32,$F6,$6B,$D6,$7B,$BE,$68
	.dc.b	$CB,$63,$33,$D2,$39,$46,$27,$73,$99,$E8,$79,$67,$44,$70,$43,$CC
	.dc.b	$E5,$A6,$4E,$E5,$3A,$9A,$21,$A2,$9A,$30,$43,$04,$3D,$C7,$52,$9C
	.dc.b	$8A,$71,$29,$C6,$3E,$03,$82,$38,$24,$FC,$62,$4F,$BE,$7A,$4E,$A3
	.dc.b	$11,$D6,$76,$F9,$67,$4D,$8C,$BE,$78,$CB,$63,$2F,$66,$BD,$67,$BB
	.dc.b	$E8,$8C,$B6,$33,$3D,$23,$94,$62,$77,$39,$9E,$87,$96,$74,$47,$04
	.dc.b	$3C,$CE,$5A,$64,$EE,$53,$A9,$A2,$1A,$29,$A3,$04,$30,$43,$DC,$75
	.dc.b	$29,$C8,$A7,$12,$9C,$63,$E0,$38,$23,$82,$4F,$C6,$24,$FB,$E7,$A4
	.dc.b	$EA,$31,$1D,$67,$6F,$96,$74,$D8,$CB,$E7,$8C,$B6,$32,$F6,$6B,$D6
	.dc.b	$7B,$BE,$88,$CB,$63,$33,$D2,$39,$46,$27,$73,$99,$E8,$79,$67,$44
	.dc.b	$70,$43,$CC,$E5,$A6,$4E,$E5,$3A,$9A,$21,$A2,$9A,$30,$43,$04,$3D
	.dc.b	$C7,$52,$9C,$8A,$71,$29,$C6,$3E,$03,$82,$38,$24,$FC,$62,$4F,$BE
	.dc.b	$7A,$4E,$A3,$11,$D6,$76,$F9,$67,$4D,$8C,$BE,$78,$CB,$63,$2F,$66
	.dc.b	$BD,$67,$BB,$E8,$8C,$B6,$33,$3D,$23,$94,$62,$77,$39,$9E,$87,$96
	.dc.b	$74,$47,$04,$3C,$CE,$5A,$64,$EE,$53,$A9,$A2,$1A,$29,$A3,$04,$30
	.dc.b	$43,$DC,$75,$29,$C8,$A7,$12,$9C,$63,$E0,$38,$23,$82,$4F,$C6,$24
	.dc.b	$FB,$E7,$A4,$EA,$31,$1D,$67,$6F,$96,$74,$D8,$CB,$E7,$8C,$B6,$32
	.dc.b	$F6,$6B,$D6,$7B,$BE,$88,$FD,$DE,$3F,$E2,$4C,$7F,$C3,$CD,$C3,$FC
	.dc.b	$2E,$FF,$88,$EB,$FC,$40,$7F,$85,$5C,$3F,$E2,$06,$E0,$69,$31,$FF
	.dc.b	$16,$92,$79,$35,$FF,$19,$18,$21,$82,$1E,$47,$83,$4D,$94,$D9,$E8
	.dc.b	$29,$E7,$78,$B9,$38,$10,$C1,$0E,$07,$63,$13,$EF,$8E,$71,$A3,$BC
	.dc.b	$65,$E0,$64,$EC,$53,$9C,$FC,$77,$71,$C5,$E7,$19,$6C,$F8,$27,$51
	.dc.b	$D1,$AF,$48,$D4,$F1,$8D,$35,$D1,$E0,$70,$F9,$49,$3D,$1B,$1A,$8C
	.dc.b	$11,$C1,$27,$C2,$49,$EC,$D7,$D6,$60,$86,$08,$7A,$1E,$0D,$36,$53
	.dc.b	$67,$B0,$A7,$A1,$E2,$E4,$E0,$43,$04,$38,$1D,$8C,$4F,$BE,$39,$C6
	.dc.b	$8E,$F1,$97,$81,$93,$B1,$4E,$73,$F1,$DD,$C7,$17,$9C,$65,$B3,$E0
	.dc.b	$9D,$47,$46,$BD,$23,$53,$C6,$34,$D7,$47,$81,$C3,$E5,$24,$F4,$6C
	.dc.b	$6A,$30,$47,$04,$9F,$09,$27,$B3,$5F,$59,$82,$18,$21,$E8,$78,$34
	.dc.b	$D9,$4D,$9E,$C2,$9E,$87,$8B,$93,$81,$0C,$10,$E0,$76,$31,$3E,$F8
	.dc.b	$E7,$1A,$3B,$C6,$5E,$06,$4E,$C5,$39,$CF,$C7,$77,$1C,$5E,$71,$96
	.dc.b	$CF,$82,$75,$1D,$1A,$F4,$8D,$4F,$18,$D3,$5D,$1E,$07,$0F,$94,$93
	.dc.b	$D1,$B1,$A8,$C1,$1C,$12,$7C,$24,$9E,$CD,$7D,$66,$08,$60,$87,$A1
	.dc.b	$E0,$D3,$65,$36,$7B,$0A,$7A,$1E,$2E,$4E,$04,$30,$43,$81,$D8,$C4
	.dc.b	$FB,$E3,$9C,$68,$EF,$19,$78,$19,$3B,$14,$E7,$3F,$1D,$DC,$71,$79
	.dc.b	$C6,$5B,$3E,$09,$D4,$74,$6B,$D2,$35,$3C,$63,$4D,$74,$78,$1C,$3E
	.dc.b	$52,$4F,$F4,$EF,$C7,$C4,$00

;--------------------------------------------------------------------------------
;	FScc.B Dr
;	FScc.B <mem>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fscc060_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#FPSP060,-(sp)
	peamsg	'FSCC.B <EA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FScc.B <ea>',13,10
;------------------------------------------------
;	d1	actual result
;	d3	fill=$00000000,$FFFFFFFF
;	d4	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result,expected status,...
;	a1	expected result
;	a4	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	fscc060_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FScc.B Dr=fill
	moveq.l	#$00000000,d3		;fill=$00000000,$FFFFFFFF
33:
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	move.l	d3,d1			;fill
	jsr	([fscc060_execute_dr,za0,d5.l*4])	;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FS'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.B Dr='
	puthex8	d3			;fill
	putmsg	' @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result
;
	not.l	d3			;fill
	bne	33b
;FScc.B <mem>=fill
	moveq.l	#$00000000,d3		;fill=$00000000,$FFFFFFFF
33:
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	move.l	d3,-(sp)		;fill
	jsr	([fscc060_execute_mem,za0,d5.l*4])	;EXECUTE
	move.l	(sp)+,d1		;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FS'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.B <mem>='
	puthex8	d3			;fill
	putmsg	' @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result
;
	not.l	d3			;fill
	bne	33b
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
fscc060_execute_dr::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fscc060_execute_dr_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
fscc060_execute_dr_&cc::
	FS&cc.B	d1			;EXECUTE
					;actual result
	rts
  .endm
fscc060_execute_mem::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fscc060_execute_mem_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
fscc060_execute_mem_&cc::
	FS&cc.B	(4,sp)			;EXECUTE
					;actual result
	rts
  .endm
	.cpu	68000

	.align	4
fscc060_expected_compressed::
	.dc.b	$00,$00,$40,$00,$10,$85,$FE,$01,$11,$C9,$0A,$E8,$C3,$1C,$38,$30
	.dc.b	$F3,$79,$3C,$8D,$BE,$13,$6E,$DE,$E7,$99,$F0,$18,$79,$BC,$9F,$01
	.dc.b	$EE,$70,$E0,$C3,$D5,$DB,$A7,$4D,$79,$1E,$17,$99,$ED,$3C,$6F,$C1
	.dc.b	$34,$F5,$34,$73,$7C,$EF,$8D,$F8,$4D,$7C,$AD,$7E,$A3,$E7,$7E,$31
	.dc.b	$87,$B3,$F2,$1F,$23,$E2,$7E,$33,$D9,$E8,$78,$DF,$9E,$78,$5C,$3E
	.dc.b	$63,$D8,$77,$79,$BD,$4C,$9E,$87,$2E,$CE,$8F,$A5,$F9,$ED,$7B,$1E
	.dc.b	$47,$07,$9D,$FC,$01,$ED,$3B,$9F,$82,$7B,$BE,$13,$BB,$F2,$5F,$96
	.dc.b	$60,$C3,$F4,$CE,$CF,$CC,$29,$EE,$7E,$EB,$93,$E0,$1F,$65,$F8,$EF
	.dc.b	$DA,$3B,$1E,$B7,$E9,$BF,$0C,$E4,$53,$B3,$F8,$C7,$DC,$75,$7E,$33
	.dc.b	$F5,$8C,$3E,$D7,$CA,$72,$7F,$06,$FE,$39,$F0,$BF,$14,$F0,$94,$FD
	.dc.b	$E4,$7E,$61,$3F,$B9,$3F,$2A,$57,$F2,$06,$8F,$E0,$2E,$9A,$69,$E0
	.dc.b	$FE,$5D,$AF,$EA,$1A,$FE,$F1,$CC,$7F,$0A,$72,$FF,$0F,$72,$FF,$15
	.dc.b	$72,$FF,$18,$72,$FF,$1E,$72,$F9,$E3,$DE,$79,$4F,$39,$E5,$3F,$8A
	.dc.b	$C7,$EA,$4F,$C8,$BC,$CF,$D1,$3F,$C7,$CD,$3F,$89,$6B,$FC,$99,$AF
	.dc.b	$F1,$B6,$BF,$CB,$DC,$C7,$F3,$27,$31,$FC,$65,$CB,$FC,$DD,$CB,$EF
	.dc.b	$3C,$67,$84,$A7,$84,$A7,$84,$A7,$84,$CC,$78,$5C,$9D,$C7,$E0,$72
	.dc.b	$3E,$A4,$62,$3F,$A2,$CF,$F4,$C7,$71,$F6,$CD,$1D,$5E,$53,$F6,$63
	.dc.b	$27,$78,$E3,$1F,$7E,$32,$7D,$18,$E3,$15,$F0,$C7,$68,$E1,$1E,$28
	.dc.b	$E8,$FE,$3D,$E6,$7E,$65,$EF,$1F,$2D,$E1,$1D,$63,$DF,$18,$8E,$0F
	.dc.b	$F3,$69,$F8,$0F,$88,$F2,$C6,$8F,$80,$ED,$F0,$C7,$09,$EB,$3C,$4D
	.dc.b	$1D,$E3,$51,$F1,$E3,$13,$EF,$8F,$44,$7F,$65,$9D,$C6,$0F,$09,$F4
	.dc.b	$E3,$DA,$FA,$8C,$3F,$85,$3E,$6B,$E9,$9F,$C9,$1F,$7D,$D9,$D0,$F9
	.dc.b	$F1,$E1,$8D,$46,$0D,$1E,$67,$F2,$11,$DA,$7E,$14,$7D,$69,$F1,$BA
	.dc.b	$9F,$E7,$D3,$A8,$FA,$93,$E0,$74,$F9,$1E,$4E,$5F,$4C,$F7,$7D,$46
	.dc.b	$A3,$C8,$FF,$88,$EB,$F9,$11,$FE,$19,$3F,$BA,$95,$FE,$42,$D7,$FB
	.dc.b	$93,$FE,$21,$3F,$B8,$1F,$C2,$13,$87,$FC,$42,$D7,$FC,$1A,$7F,$8D
	.dc.b	$1C,$3F,$E3,$07,$C6,$FF,$8B,$9F,$1B,$4F,$19,$4F,$19,$D8,$FE,$F0
	.dc.b	$F8,$CE,$AF,$F8,$FD,$EF,$1F,$E4,$27,$31,$F0,$CE,$A7,$F9,$05,$F0
	.dc.b	$B4,$C9,$E7,$3F,$C4,$2F,$C3,$7C,$27,$F6,$D7,$AB,$FE,$4E,$7C,$2F
	.dc.b	$73,$FC,$14,$FC,$D7,$FC,$5A,$D7,$FC,$82,$E6,$3F,$C9,$EE,$5F,$F2
	.dc.b	$A3,$E1,$8E,$67,$84,$F1,$9F,$E5,$A7,$C2,$FD,$13,$27,$AC,$FF,$04
	.dc.b	$BD,$63,$D1,$1F,$3C,$F2,$94,$F2,$9F,$44,$F0,$9F,$78,$FC,$19,$D8
	.dc.b	$FA,$47,$81,$FA,$4E,$4F,$88,$76,$3E,$81,$F7,$8A,$64,$F4,$1F,$E4
	.dc.b	$A8,$F8,$4F,$F8,$9D,$FC,$13,$D4,$F5,$9E,$63,$FC,$E0,$FD,$67,$CE
	.dc.b	$7C,$23,$DD,$1F,$39,$F8,$4F,$D2,$72,$F4,$7F,$CF,$11,$F3,$9E,$67
	.dc.b	$8C,$F8,$C5,$3E,$33,$E1,$3F,$42,$FB,$8F,$9A,$7C,$08,$F0,$39,$7C
	.dc.b	$8F,$DD,$3E,$61,$F2,$9E,$2F,$90,$C9,$F8,$13,$E1,$1F,$EA,$4B,$FF
	.dc.b	$12,$1F,$E9,$B3,$FC,$EE,$57,$FC,$46,$D7,$FC,$E8,$D7,$FD,$0A,$D7
	.dc.b	$FD,$32,$E6,$3F,$D4,$EE,$5F,$F5,$63,$97,$FD,$64,$E5,$FF,$5A,$B9
	.dc.b	$7F,$D7,$6E,$53,$98,$A7,$FA,$D9,$F4,$B4,$F4,$94,$F4,$9F,$EB,$67
	.dc.b	$FD,$14,$7F,$1B,$2B,$93,$FD,$08,$E5,$A6,$4F,$F4,$23,$96,$99,$3F
	.dc.b	$CF,$4E,$5A,$66,$3E,$3B,$FE,$75,$8F,$0B,$FE,$D2,$6B,$FE,$C7,$6B
	.dc.b	$FE,$DB,$73,$1F,$6E,$3C,$A7,$12,$9C,$4A,$6E,$BF,$BD,$CF,$F7,$E8
	.dc.b	$C4,$7F,$AD,$A3,$11,$FE,$EB,$7F,$B8,$BF,$42,$3F,$C1,$11,$C6,$7E
	.dc.b	$EC,$FF,$84,$A3,$67,$27,$ED,$9F,$DF,$63,$94,$FF,$BC,$A7,$73,$C4
	.dc.b	$D3,$C5,$DC,$FE,$31,$D4,$FF,$87,$A3,$9C,$7A,$9D,$47,$9A,$39,$9C
	.dc.b	$1F,$CC,$B4,$FA,$6E,$CF,$8A,$E6,$3D,$B1,$FE,$0E,$9F,$F0,$9C,$79
	.dc.b	$27,$E3,$1F,$E0,$67,$F3,$6F,$F8,$F9,$F4,$9C,$A7,$EA,$1F,$E1,$49
	.dc.b	$F8,$D1,$A3,$8C,$7F,$95,$67,$AB,$B7,$A4,$F8,$8F,$F0,$53,$C6,$7E
	.dc.b	$1A,$39,$CF,$E4,$8E,$8F,$D4,$7A,$1E,$D7,$D5,$1E,$D7,$51,$C4,$E8
	.dc.b	$F5,$8F,$04,$F8,$8F,$24,$79,$8E,$6E,$5A,$7F,$96,$30,$FF,$5B,$1D
	.dc.b	$18,$1B,$22,$3A,$8F,$13,$C6,$34,$6A,$3F,$DC,$51,$98,$CB,$C6,$3E
	.dc.b	$33,$E2,$70,$F8,$23,$94,$78,$A3,$94,$62,$3C,$87,$C4,$8C,$C7,$A6
	.dc.b	$39,$9F,$EC,$E8,$F0,$39,$3E,$5C,$79,$1F,$63,$97,$D6,$F7,$3E,$34
	.dc.b	$76,$39,$9E,$98,$F8,$8F,$A6,$39,$47,$68,$E0,$FC,$43,$C3,$1E,$17
	.dc.b	$89,$E7,$3D,$26,$9E,$6F,$D0,$30,$F9,$CF,$2B,$88,$F2,$BC,$CE,$A7
	.dc.b	$37,$E9,$47,$13,$C5,$1F,$30,$FA,$A7,$33,$C6,$7C,$43,$C6,$68,$E8
	.dc.b	$7C,$A8,$F7,$BF,$E1,$76,$BF,$8E,$7E,$43,$4F,$BA,$F9,$0D,$C7,$B2
	.dc.b	$3F,$40,$F9,$9C,$1E,$38,$F7,$BE,$23,$F1,$07,$A4,$F1,$14,$F0,$1D
	.dc.b	$23,$D4,$E0,$C4,$78,$DE,$11,$E0,$3F,$0B,$1E,$08,$F1,$1F,$18,$C9
	.dc.b	$F0,$CE,$27,$E5,$CF,$11,$CE,$3F,$37,$1C,$E3,$89,$EA,$3C,$D1,$F7
	.dc.b	$DE,$46,$23,$C5,$1F,$B2,$8F,$AA,$7E,$49,$F9,$0F,$E2,$4E,$07,$E4
	.dc.b	$CA,$7E,$18,$F5,$94,$F1,$9F,$BF,$8F,$AA,$F8,$8F,$D3,$C7,$82,$3E
	.dc.b	$EC,$7D,$63,$89,$FA,$33,$E1,$9F,$42,$32,$F3,$7D,$51,$F0,$DF,$4C
	.dc.b	$7E,$CD,$FC,$B9,$F2,$4F,$E0,$8F,$F1,$BC,$7F,$83,$A7,$F8,$09,$FB
	.dc.b	$53,$F4,$27,$F0,$83,$FD,$5C,$57,$F8,$B3,$C8,$E2,$FF,$12,$78,$BF
	.dc.b	$B2,$78,$B4,$E2,$7E,$E4,$F1,$18,$3F,$78,$F2,$7E,$EB,$C9,$A6,$CF
	.dc.b	$D8,$3E,$27,$89,$98,$FC,$04,$66,$3F,$38,$F5,$7E,$AB,$D5,$A7,$F2
	.dc.b	$07,$E0,$3F,$E8,$D6,$BF,$CF,$5B,$1F,$36,$3D,$C7,$F4,$68,$CB,$FD
	.dc.b	$59,$E2,$FF,$2E,$72,$FF,$65,$72,$FF,$5B,$72,$FC,$63,$E0,$9F,$D3
	.dc.b	$DF,$A4,$FC,$33,$F9,$AB,$ED,$8B,$1F,$CF,$0F,$11,$E6,$3F,$AE,$C7
	.dc.b	$F8,$B0,$FF,$0F,$1F,$E2,$B8,$FE,$BC,$FF,$81,$0F,$F0,$79,$FD,$F8
	.dc.b	$B0,$B8,$FC,$37,$13,$F4,$A3,$51,$E0,$7F,$C6,$53,$D5,$FE,$FE,$F5
	.dc.b	$69,$C0,$E9,$1F,$80,$70,$7B,$1F,$F1,$94,$73,$7C,$33,$FD,$CA,$3A
	.dc.b	$46,$23,$DB,$3C,$8F,$A4,$7F,$8A,$4F,$CA,$C7,$B1,$EC,$FE,$29,$E0
	.dc.b	$69,$F2,$1B,$79,$BE,$47,$FC,$65,$3F,$22,$3F,$C4,$31,$C6,$7E,$BC
	.dc.b	$71,$70,$7A,$63,$E4,$BF,$08,$DC,$7E,$4E,$3E,$3B,$D2,$7B,$39,$3E
	.dc.b	$04,$F7,$9F,$5B,$C9,$FC,$A1,$CA,$3F,$B0,$3E,$28,$F8,$F1,$E5,$9E
	.dc.b	$B1,$E3,$3C,$D1,$F0,$67,$81,$F1,$4F,$B2,$77,$3F,$3C,$77,$8E,$47
	.dc.b	$68,$FF,$18,$E5,$FC,$60,$FF,$0B,$9F,$E0,$B3,$FB,$B9,$FE,$1E,$3F
	.dc.b	$C6,$13,$FE,$44,$70,$FF,$8B,$5E,$4D,$3F,$C5,$2F,$27,$FC,$12,$F2
	.dc.b	$7F,$CB,$0F,$27,$FC,$7C,$F2,$69,$C8,$FF,$13,$BB,$7E,$4B,$5F,$F2
	.dc.b	$63,$C5,$F6,$9C,$4F,$F2,$6B,$C6,$77,$1E,$43,$D0,$7F,$75,$11,$F1
	.dc.b	$1E,$47,$F9,$45,$E4,$FC,$33,$91,$E8,$39,$3C,$5F,$F1,$63,$F0,$DF
	.dc.b	$ED,$AF,$C3,$7B,$9C,$8F,$49,$C8,$F2,$9F,$E0,$97,$93,$FD,$89,$E4
	.dc.b	$7C,$37,$C8,$71,$3C,$87,$13,$E5,$9D,$E3,$E1,$BE,$43,$BB,$FD,$F6
	.dc.b	$3D,$51,$C4,$F9,$67,$13,$B9,$E3,$38,$C7,$F9,$A2,$38,$C7,$F9,$79
	.dc.b	$F4,$C5,$8F,$E5,$47,$E1,$0F,$84,$79,$0F,$F2,$14,$72,$8F,$F4,$0B
	.dc.b	$C9,$B1,$C8,$F9,$67,$AC,$F8,$2F,$AD,$F9,$67,$84,$F5,$9B,$3F,$1A
	.dc.b	$78,$8F,$6B,$C5,$F9,$67,$13,$F0,$87,$C3,$3D,$47,$33,$D4,$F1,$7E
	.dc.b	$E9,$F2,$CE,$27,$85,$D9,$EC,$7D,$2F,$CB,$7D,$2E,$5E,$27,$D1,$3F
	.dc.b	$D7,$07,$A1,$D3,$D1,$F2,$BE,$23,$D4,$7F,$A5,$A3,$F3,$AE,$A3,$2F
	.dc.b	$FB,$13,$3F,$F0,$F9,$FE,$9E,$3F,$D2,$47,$F9,$F8,$FF,$52,$9F,$9D
	.dc.b	$2B,$FE,$B3,$78,$B4,$FF,$57,$BC,$5F,$F4,$7B,$C5,$A6,$CF,$DB,$3E
	.dc.b	$17,$89,$FE,$AC,$7B,$BF,$08,$FF,$73,$38,$7F,$DA,$2F,$89,$FF,$67
	.dc.b	$3E,$A7,$B1,$FE,$91,$7B,$3F,$EA,$57,$CC,$FF,$99,$9F,$33,$4F,$11
	.dc.b	$E8,$3C,$41,$F1,$3F,$EC,$47,$C9,$1F,$EE,$57,$31,$FE,$F1,$72,$FD
	.dc.b	$13,$A1,$FE,$E3,$72,$FB,$8C,$9D,$4D,$C7,$57,$27,$33,$E5,$9F,$00
	.dc.b	$FF,$69,$C7,$B9,$FF,$37,$BE,$E6,$9F,$66,$BF,$C0,$0F,$B4,$FE,$F3
	.dc.b	$1F,$E1,$F9,$C4,$FF,$71,$84,$FF,$70,$9F,$F7,$DC,$62,$7F,$BC,$B5
	.dc.b	$EE,$7F,$61,$71,$1F,$E9,$17,$53,$ED,$9F,$F1,$3C,$F4,$8F,$F0,$93
	.dc.b	$E2,$8E,$67,$F7,$29,$E7,$1B,$8E,$D3,$FE,$17,$9E,$AF,$D1,$8E,$B2
	.dc.b	$3A,$47,$CF,$8F,$F0,$1C,$F9,$A3,$84,$78,$1F,$33,$F8,$D7,$E3,$9F
	.dc.b	$E3,$39,$F1,$47,$E5,$A3,$C4,$69,$F0,$C7,$82,$3F,$0C,$7C,$B7,$C9
	.dc.b	$3E,$18,$FC,$24,$F9,$1F,$87,$1F,$02,$77,$1F,$E2,$23,$CE,$FE,$09
	.dc.b	$D9,$D5,$F9,$47,$A8,$D4,$FD,$D8,$D4,$F8,$5D,$CF,$B1,$F8,$D3,$F4
	.dc.b	$A3,$91,$98,$F5,$C7,$9C,$FC,$1B,$D0,$F8,$4F,$C9,$3E,$B9,$F1,$27
	.dc.b	$EB,$C7,$B2,$39,$1F,$7D,$E4,$FF,$8A,$F4,$FF,$5C,$9F,$94,$88,$56
	.dc.b	$BF,$7E,$BF,$01,$18,$8D,$1F,$EB,$17,$0F,$E7,$9E,$31,$C5,$CC,$1F
	.dc.b	$0B,$C1,$DC,$76,$8C,$1D,$4A,$79,$DE,$8F,$8D,$E8,$E5,$CB,$FC,$4A
	.dc.b	$3D,$B1,$EB,$37,$1E,$68,$CC,$70,$8F,$F2,$BC,$62,$3C,$91,$B3,$B3
	.dc.b	$87,$66,$1E,$AF,$82,$39,$46,$0F,$9E,$E0,$F8,$A7,$C4,$8F,$9C,$FC
	.dc.b	$48,$E7,$1C,$0F,$88,$68,$F3,$47,$27,$27,$DA,$79,$19,$7F,$C8,$0D
	.dc.b	$7F,$8A,$C7,$C1,$7E,$74,$79,$5F,$F0,$73,$5F,$84,$70,$8F,$90,$79
	.dc.b	$5C,$3F,$15,$E6,$7C,$58,$FF,$21,$1C,$DF,$8C,$64,$EE,$53,$DC,$7A
	.dc.b	$0F,$C1,$9D,$9F,$54,$7E,$5E,$3E,$39,$F9,$63,$EB,$C6,$A3,$F8,$D4
	.dc.b	$68,$FC,$C1,$F2,$8E,$AF,$38,$F8,$D1,$F8,$97,$F3,$F1,$F6,$CF,$83
	.dc.b	$1D,$CF,$C7,$3D,$4D,$1F,$87,$3E,$33,$A8,$E6,$7C,$38,$C9,$DE,$3F
	.dc.b	$3D,$19,$7B,$46,$5F,$8A,$7E,$74,$F6,$1F,$88,$7E,$E3,$C1,$F8,$31
	.dc.b	$EC,$73,$1F,$08,$F9,$C7,$43,$B1,$F8,$B3,$F1,$A7,$A8,$C9,$F7,$8F
	.dc.b	$C0,$BF,$9F,$7C,$0F,$E0,$DF,$C7,$C7,$AD,$F0,$1F,$0C,$F7,$9F,$81
	.dc.b	$3F,$5A,$74,$7E,$3B,$98,$F8,$E7,$A9,$6B,$FB,$96,$9F,$C5,$9F,$D6
	.dc.b	$1F,$C5,$0D,$15,$FE,$00,$D7,$FD,$4C,$D7,$F3,$AD,$8F,$C8,$39,$8F
	.dc.b	$E1,$CE,$63,$F1,$EE,$5F,$E5,$8E,$5F,$F0,$2B,$97,$FB,$73,$98,$F4
	.dc.b	$9F,$CC,$5F,$4B,$EB,$3F,$89,$9A,$3F,$9C,$BA,$7F,$99,$BF,$C4,$8F
	.dc.b	$F5,$91,$5F,$E5,$4D,$7F,$C7,$2D,$7F,$94,$B6,$3F,$A3,$B9,$8F,$E6
	.dc.b	$8E,$5F,$E9,$8E,$5F,$E1,$AE,$5F,$79,$FC,$9D,$FD,$59,$97,$D5,$1F
	.dc.b	$29,$F2,$9B,$3F,$D4,$11,$D5,$F8,$E6,$4F,$49,$4F,$86,$7F,$81,$CF
	.dc.b	$C1,$D0,$C9,$FE,$0E,$8F,$ED,$71,$A8,$FD,$64,$60,$FF,$1F,$16,$3A
	.dc.b	$36,$39,$47,$57,$A9,$C9,$C1,$F2,$8F,$F1,$14,$FA,$A3,$C9,$19,$7F
	.dc.b	$C9,$F3,$D5,$FC,$8C,$72,$3E,$79,$E2,$30,$7C,$E7,$C6,$74,$76,$F4
	.dc.b	$76,$7B,$CF,$A0,$ED,$E4,$FA,$4C,$3B,$9F,$CD,$BF,$E5,$09,$FF,$17
	.dc.b	$47,$C9,$3D,$53,$FC,$92,$3F,$C2,$F1,$93,$8C,$7D,$37,$67,$89,$FC
	.dc.b	$8C,$FF,$8C,$E7,$FC,$5C,$76,$39,$95,$FF,$1E,$1E,$89,$E4,$FE,$58
	.dc.b	$E4,$68,$F1,$3F,$46,$7B,$9E,$37,$CA,$FF,$48,$8E,$91,$87,$A4,$6E
	.dc.b	$39,$1E,$A3,$51,$FA,$23,$C3,$1E,$F7,$E9,$BF,$15,$FA,$7B,$7F,$19
	.dc.b	$3F,$C5,$27,$F8,$0C,$AF,$F0,$06,$BF,$E1,$46,$BF,$D0,$DA,$FF,$88
	.dc.b	$5C,$CF,$F6,$77,$2F,$F8,$FD,$FE,$BA,$7F,$86,$0F,$F1,$4C,$7F,$91
	.dc.b	$DC,$3F,$E3,$16,$BF,$EC,$06,$BF,$E3,$06,$C7,$F9,$39,$CC,$7F,$8E
	.dc.b	$DC,$BD,$4F,$F1,$53,$E6,$8F,$EE,$EF,$99,$FF,$49,$39,$7F,$D5,$AE
	.dc.b	$53,$98,$A7,$F9,$59,$F9,$4D,$3E,$51,$4F,$94,$7B,$4F,$19,$FE,$51
	.dc.b	$3C,$C7,$F5,$E7,$E3,$BA,$3E,$11,$4F,$84,$58,$F8,$4E,$4F,$84,$64
	.dc.b	$EA,$7F,$82,$A3,$D4,$7B,$4A,$F9,$CF,$F1,$7B,$87,$FC,$EC,$FB,$9A
	.dc.b	$7B,$8A,$7B,$8F,$F3,$BB,$F4,$9F,$96,$79,$4F,$F2,$2B,$97,$E5,$9F
	.dc.b	$04,$F8,$C5,$3E,$09,$4F,$82,$58,$F8,$2E,$4A,$7E,$5C,$F6,$1F,$97
	.dc.b	$3D,$CF,$E5,$DC,$BF,$97,$72,$FE,$19,$CB,$F8,$67,$2F,$B1,$F8,$27
	.dc.b	$C8,$3D,$27,$CF,$3E,$F1,$B3,$07,$C6,$3E,$B1,$97,$CC,$E4,$F7,$1F
	.dc.b	$69,$CB,$F0,$8E,$67,$F0,$B3,$E5,$98,$3F,$D6,$7B,$FF,$86,$4F,$F5
	.dc.b	$49,$FE,$83,$2B,$FE,$00,$6B,$FE,$94,$6B,$FE,$43,$6B,$FE,$A1,$73
	.dc.b	$1F,$E6,$77,$2F,$FA,$FD,$CB,$FE,$66,$72,$FF,$C0,$17,$2A,$11,$A8
	.dc.b	$B1,$FE,$C4,$8F,$53,$4F,$F5,$63,$C9,$D9,$FE,$CD,$7D,$CF,$FA,$F5
	.dc.b	$FF,$54,$9F,$C4,$9F,$03,$87,$47,$FA,$C5,$D3,$D0,$FF,$39,$BF,$E7
	.dc.b	$83,$83,$FE,$A0,$3F,$DC,$4E,$1F,$F6,$3B,$63,$E9,$C7,$53,$B9,$FE
	.dc.b	$5E,$8E,$E7,$17,$FD,$AC,$F1,$7A,$9C,$4F,$F7,$2B,$97,$D6,$64,$F0
	.dc.b	$96,$3C,$2F,$73,$E5,$57,$F4,$F7,$E7,$BF,$DD,$27,$EA,$3F,$E0,$A9
	.dc.b	$FB,$93,$FD,$FA,$73,$3F,$85,$8C,$1F,$E1,$29,$E7,$1F,$76,$3D,$2F
	.dc.b	$79,$FF,$12,$9C,$A3,$F1,$87,$28,$D3,$5F,$F6,$1C,$70,$8F,$F1,$93
	.dc.b	$63,$CD,$3F,$8D,$9E,$0F,$E5,$5E,$47,$68,$F7,$3C,$8C,$C7,$BA,$32
	.dc.b	$78,$8A,$78,$A3,$FC,$4B,$1E,$E3,$A4,$7E,$CE,$7F,$C3,$27,$F8,$58
	.dc.b	$B3,$FE,$07,$6C,$7F,$89,$67,$91,$E1,$79,$4F,$F9,$06,$3A,$BF,$84
	.dc.b	$30,$EC,$FC,$6C,$7F,$82,$5F,$A3,$1B,$8C,$4F,$D8,$9D,$C7,$E7,$27
	.dc.b	$D7,$3E,$48,$F0,$9C,$23,$27,$82,$7E,$74,$7F,$84,$63,$F2,$2E,$CF
	.dc.b	$64,$F3,$9E,$E7,$AA,$7F,$09,$C1,$FE,$AD,$3F,$0C,$C0,$D8,$11,$FF
	.dc.b	$74,$9C,$1C,$99,$8F,$E6,$27,$23,$F2,$4E,$8C,$3D,$CA,$7F,$B8,$63
	.dc.b	$0F,$F1,$57,$C5,$1D,$63,$C9,$1E,$18,$F0,$BF,$C2,$0C,$3B,$8F,$64
	.dc.b	$7B,$DE,$AF,$48,$F0,$C7,$9A,$38,$C7,$90,$F8,$AE,$0F,$2C,$66,$3E
	.dc.b	$2B,$C8,$F9,$A7,$8A,$3E,$3B,$F0,$A3,$CE,$FC,$68,$E8,$F9,$0F,$F6
	.dc.b	$1B,$D8,$F3,$1C,$9F,$F7,$AB,$EF,$7A,$47,$BD,$F3,$9C,$8E,$86,$A3
	.dc.b	$93,$F4,$E3,$DC,$FA,$0E,$A7,$03,$EA,$14,$FB,$51,$CA,$2C,$7A,$CF
	.dc.b	$1B,$F0,$DF,$B4,$78,$CF,$C0,$47,$B1,$FC,$14,$7C,$B3,$EC,$1E,$A7
	.dc.b	$C6,$70,$39,$1F,$70,$F9,$C7,$28,$F9,$6F,$F4,$53,$C2,$72,$38,$1F
	.dc.b	$D8,$63,$F3,$2F,$9C,$FC,$1B,$F5,$9F,$C7,$9E,$C8,$FA,$0F,$33,$CC
	.dc.b	$7C,$63,$CC,$7C,$F3,$C8,$71,$3E,$31,$C4,$FA,$07,$13,$CA,$79,$63
	.dc.b	$EE,$3E,$78,$F6,$BF,$E1,$16,$BE,$28,$F5,$C7,$E2,$63,$E2,$9E,$A3
	.dc.b	$E1,$46,$CF,$73,$E0,$78,$BF,$7D,$CB,$FA,$53,$D4,$7C,$D3,$EC,$3F
	.dc.b	$26,$3F,$5D,$1F,$08,$F5,$47,$E0,$CD,$C7,$CA,$8F,$C9,$3E,$77,$67
	.dc.b	$33,$D2,$7F,$0F,$7E,$0C,$7A,$4F,$8B,$1F,$E1,$BE,$1F,$E0,$29,$FD
	.dc.b	$B9,$FB,$F3,$F8,$6C,$66,$3F,$2F,$1F,$E1,$03,$F8,$09,$C1,$FD,$69
	.dc.b	$B3,$F8,$C3,$CD,$FC,$CB,$C1,$FE,$42,$D7,$67,$27,$A3,$FB,$87,$C6
	.dc.b	$6D,$FF,$26,$35,$FE,$6A,$FB,$0E,$2F,$CC,$6C,$7E,$69,$CC,$7F,$2F
	.dc.b	$7A,$BE,$E3,$CA,$53,$F9,$F3,$E1,$8E,$F1,$93,$F9,$F4,$65,$FF,$06
	.dc.b	$3E,$57,$F5,$4E,$5F,$A4,$71,$3F,$A0,$B9,$7E,$51,$FC,$45,$F2,$BF
	.dc.b	$86,$8C,$9E,$27,$11,$EE,$8F,$E5,$EF,$B9,$FE,$A2,$F2,$8E,$87,$B8
	.dc.b	$E2,$7F,$4C,$7F,$B9,$47,$28,$5C,$7E,$0B,$FE,$1C,$8F,$C3,$3F,$E0
	.dc.b	$79,$E3,$1F,$D9,$1F,$EF,$93,$F1,$0F,$EE,$D1,$C2,$7F,$B6,$4F,$09
	.dc.b	$C1,$DA,$3E,$D3,$FE,$1C,$8E,$B3,$C4,$D1,$F6,$E3,$CF,$3D,$1F,$F0
	.dc.b	$21,$D0,$F1,$3E,$99,$FF,$1F,$47,$D6,$3C,$AF,$A1,$FA,$F1,$FE,$19
	.dc.b	$8D,$C6,$23,$FB,$C3,$D4,$C1,$E3,$7E,$43,$E0,$77,$1C,$A3,$DF,$3F
	.dc.b	$59,$F7,$1E,$23,$FC,$5C,$71,$7A,$4F,$E4,$23,$4F,$F8,$7D,$F8,$91
	.dc.b	$CE,$3E,$A4,$69,$FD,$4B,$4F,$C8,$99,$8F,$96,$71,$8E,$C7,$C1,$8F
	.dc.b	$19,$F0,$63,$C6,$F2,$8D,$1E,$39,$FA,$E7,$BC,$FB,$C5,$8F,$F3,$A3
	.dc.b	$F3,$27,$D2,$7E,$26,$3B,$98,$7E,$72,$E2,$FE,$36,$7F,$84,$0F,$F0
	.dc.b	$C1,$FC,$68,$FF,$0B,$9F,$EA,$82,$BF,$E2,$A7,$8B,$4F,$F1,$2B,$C5
	.dc.b	$FE,$EE,$F1,$69,$B3,$F9,$9B,$E1,$78,$9F,$DE,$1E,$EF,$F4,$36,$BF
	.dc.b	$E4,$B7,$C2,$FF,$8F,$DF,$4B,$C4,$EE,$71,$29,$C4,$FE,$62,$F8,$5F
	.dc.b	$F0,$FB,$E9,$7F,$84,$3E,$96,$9E,$93,$F3,$2F,$17,$FA,$9B,$C5,$FF
	.dc.b	$11,$3C,$5A,$71,$8F,$F0,$2C,$71,$8F,$F2,$D3,$C6,$2C,$7F,$98,$23
	.dc.b	$E0,$BF,$E6,$C7,$E0,$BE,$B3,$D2,$7C,$13,$CE,$7C,$12,$9B,$3D,$27
	.dc.b	$AC,$EE,$7F,$9D,$5F,$9F,$1D,$E3,$E4,$94,$FB,$27,$17,$FC,$EF,$1E
	.dc.b	$96,$C7,$A4,$E2,$7F,$B2,$63,$D2,$F3,$3E,$49,$E9,$3E,$39,$E9,$29
	.dc.b	$E9,$8F,$92,$F8,$5F,$F4,$5C,$78,$5B,$1E,$13,$E0,$9D,$DE,$2F,$E2
	.dc.b	$4F,$C8,$9F,$04,$FC,$3B,$B3,$D6,$EC,$F8,$25,$3D,$27,$13,$E7,$BC
	.dc.b	$4F,$1B,$B7,$C2,$FD,$33,$E8,$9F,$EA,$18,$FA,$2F,$C7,$3E,$48,$3F
	.dc.b	$74,$7E,$68,$F8,$51,$87,$C4,$7C,$87,$C4,$65,$FF,$5A,$71,$FF,$87
	.dc.b	$0F,$F4,$89,$FE,$99,$3F,$C3,$67,$FA,$60,$FF,$5A,$3F,$F0,$05,$C3
	.dc.b	$FE,$AC,$79,$34,$FF,$53,$BC,$9F,$F3,$CB,$C9,$F5,$3C,$9F,$F5,$C3
	.dc.b	$C9,$A7,$23,$FC,$F6,$ED,$FF,$52,$9C,$8E,$A7,$84,$FF,$1F,$BE,$C7
	.dc.b	$91,$D0,$C9,$B3,$FC,$96,$F9,$DF,$F0,$43,$E7,$4F,$85,$FF,$68,$BE
	.dc.b	$38,$FF,$6F,$B9,$8F,$F4,$13,$97,$E4,$1E,$63,$FD,$C4,$E5,$F5,$1E
	.dc.b	$03,$DE,$6C,$F7,$9E,$23,$07,$88,$D4,$78,$9C,$9E,$03,$DE,$7C,$13
	.dc.b	$DE,$76,$3C,$07,$F9,$C6,$3D,$4F,$CD,$BF,$F0,$DC,$FF,$A7,$27,$FB
	.dc.b	$91,$F9,$38,$FF,$0E,$3F,$D4,$CF,$F0,$79,$FE,$03,$9F,$44,$72,$9F
	.dc.b	$F7,$EC,$F2,$9F,$F0,$7B,$67,$9C,$7F,$85,$0F,$73,$C6,$47,$47,$E2
	.dc.b	$CF,$82,$78,$1F,$42,$38,$3B,$3E,$33,$D8,$FF,$07,$BD,$23,$94,$7F
	.dc.b	$81,$5A,$FE,$A2,$3C,$E7,$63,$E3,$3F,$30,$E6,$F9,$67,$E4,$3B,$3C
	.dc.b	$4F,$C2,$79,$3B,$31,$1E,$18,$FF,$20,$CF,$CD,$69,$FD,$F9,$F2,$BF
	.dc.b	$1A,$3F,$DE,$51,$F4,$4C,$C7,$E5,$0E,$31,$D4,$F9,$93,$F1,$63,$51
	.dc.b	$89,$E5,$1C,$23,$F8,$93,$A7,$BB,$E3,$34,$7C,$E7,$FC,$7F,$1F,$75
	.dc.b	$F8,$C7,$28,$FA,$07,$59,$E0,$7D,$E3,$30,$00

;--------------------------------------------------------------------------------
;	FScc.B Dr
;	FScc.B <mem>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fscc88x_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+FPSP040,-(sp)
	peamsg	'FSCC.B <EA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FScc.B <ea>',13,10
;------------------------------------------------
;	d1	actual result
;	d3	fill=$00000000,$FFFFFFFF
;	d4	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result,expected status,...
;	a1	expected result
;	a4	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	fscc88x_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FScc.B Dr=fill
	moveq.l	#$00000000,d3		;fill=$00000000,$FFFFFFFF
33:
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	move.l	d3,d1			;fill
	jsr	([fscc88x_execute_dr,za0,d5.l*4])	;EXECUTE
					;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FS'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.B Dr='
	puthex8	d3			;fill
	putmsg	' @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result
;
	not.l	d3			;fill
	bne	33b
;FScc.B <mem>=fill
	moveq.l	#$00000000,d3		;fill=$00000000,$FFFFFFFF
33:
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	move.l	d3,-(sp)		;fill
	jsr	([fscc88x_execute_mem,za0,d5.l*4])	;EXECUTE
	move.l	(sp)+,d1		;actual result
	fmove.l	fpsr,d4			;actual status
;
	movea.l	(a0),a1			;expected result
	movea.l	(4,a0),a4		;expected status
;
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	test_single
	lea.l	16(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FS'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.B <mem>='
	puthex8	d3			;fill
	putmsg	' @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1/d4/a1/a4,-(sp)	;actual result,actual status,expected result,expected status
	jbsr	output_single
	lea.l	20(sp),sp
@@:
;
	addq.l	#8,a0			;expected result
;
	not.l	d3			;fill
	bne	33b
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
fscc88x_execute_dr::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fscc88x_execute_dr_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
fscc88x_execute_dr_&cc::
	FS&cc.B	d1			;EXECUTE
					;actual result
	rts
  .endm
fscc88x_execute_mem::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	fscc88x_execute_mem_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
fscc88x_execute_mem_&cc::
	FS&cc.B	(4,sp)			;EXECUTE
					;actual result
	rts
  .endm
	.cpu	68000

	.align	4
fscc88x_expected_compressed::
	.dc.b	$00,$00,$40,$00,$10,$85,$FE,$01,$11,$C9,$0A,$E8,$C3,$1C,$38,$30
	.dc.b	$F3,$79,$3C,$8D,$BE,$13,$6E,$DE,$E7,$99,$F0,$18,$79,$BC,$9F,$01
	.dc.b	$EE,$70,$E0,$C3,$D5,$DB,$A7,$4D,$79,$1E,$17,$99,$ED,$3C,$6F,$C1
	.dc.b	$34,$F5,$34,$73,$7C,$EF,$8D,$F8,$4D,$7C,$AD,$7E,$A3,$E7,$7E,$31
	.dc.b	$87,$B3,$F2,$1F,$23,$E2,$7E,$33,$D9,$E8,$78,$DF,$9E,$78,$5C,$3E
	.dc.b	$63,$D8,$77,$79,$BD,$4C,$9E,$87,$2E,$CE,$8F,$A5,$F9,$ED,$7B,$1E
	.dc.b	$47,$07,$9D,$FC,$01,$ED,$3B,$9F,$82,$7B,$BE,$13,$BB,$F2,$5F,$96
	.dc.b	$60,$C3,$F4,$CE,$CF,$CC,$29,$EE,$7E,$EB,$93,$E0,$1F,$65,$F8,$EF
	.dc.b	$DA,$3B,$1E,$B7,$E9,$BF,$0C,$E4,$53,$B3,$F8,$C7,$DC,$75,$7E,$33
	.dc.b	$F5,$8C,$3E,$D7,$CA,$72,$7F,$06,$FE,$39,$F0,$BF,$14,$F0,$94,$FD
	.dc.b	$E4,$7E,$61,$3F,$B9,$3F,$2A,$57,$F2,$06,$8F,$E0,$2E,$9A,$69,$E0
	.dc.b	$FE,$5D,$AF,$EA,$1A,$FE,$F1,$CC,$7F,$0A,$72,$FF,$0F,$72,$FF,$15
	.dc.b	$72,$FF,$18,$72,$FF,$1E,$72,$F9,$E3,$DE,$79,$4F,$39,$E5,$3F,$8A
	.dc.b	$C7,$EA,$4F,$C8,$BC,$CF,$D1,$3F,$C7,$CD,$3F,$89,$6B,$FC,$99,$AF
	.dc.b	$F1,$B6,$BF,$CB,$DC,$C7,$F3,$27,$31,$FC,$65,$CB,$FC,$DD,$CB,$EF
	.dc.b	$3C,$67,$84,$A7,$84,$A7,$84,$A7,$84,$CC,$78,$5C,$9D,$C7,$E0,$72
	.dc.b	$3E,$A4,$62,$3F,$A2,$CF,$F4,$C7,$71,$F6,$CD,$1D,$5E,$53,$F6,$63
	.dc.b	$27,$78,$E3,$1F,$7E,$32,$7D,$18,$E3,$15,$F0,$C7,$68,$E1,$1E,$28
	.dc.b	$E8,$FE,$3D,$E6,$7E,$65,$EF,$1F,$2D,$E1,$1D,$63,$DF,$18,$8E,$0F
	.dc.b	$F3,$69,$F8,$0F,$88,$F2,$C6,$8F,$80,$ED,$F0,$C7,$09,$EB,$3C,$4D
	.dc.b	$1D,$E3,$51,$F1,$E3,$13,$EF,$8F,$44,$7F,$65,$9D,$C6,$0F,$09,$F4
	.dc.b	$E3,$DA,$FA,$8C,$3F,$85,$3E,$6B,$E9,$9F,$C9,$1F,$7D,$D9,$D0,$F9
	.dc.b	$F1,$E1,$8D,$46,$0D,$1E,$67,$F2,$11,$DA,$7E,$14,$7D,$69,$F1,$BA
	.dc.b	$9F,$E7,$D3,$A8,$FA,$93,$E0,$74,$F9,$1E,$4E,$5F,$4C,$F7,$7D,$46
	.dc.b	$A3,$C8,$FF,$88,$EB,$F9,$11,$FE,$19,$3F,$BA,$95,$FE,$42,$D7,$FB
	.dc.b	$93,$FE,$21,$3F,$B8,$1F,$C2,$13,$87,$FC,$42,$D7,$FC,$1A,$7F,$8D
	.dc.b	$1C,$3F,$E3,$07,$C6,$FF,$8B,$9F,$1B,$4F,$19,$4F,$19,$D8,$FE,$F0
	.dc.b	$F8,$CE,$AF,$F8,$FD,$EF,$1F,$E4,$27,$31,$F0,$CE,$A7,$F9,$05,$F0
	.dc.b	$B4,$C9,$E7,$3F,$C4,$2F,$C3,$7C,$27,$F6,$D7,$AB,$FE,$4E,$7C,$2F
	.dc.b	$73,$FC,$14,$FC,$D7,$FC,$5A,$D7,$FC,$82,$E6,$3F,$C9,$EE,$5F,$F2
	.dc.b	$A3,$E1,$8E,$67,$84,$F1,$9F,$E5,$A7,$C2,$FD,$13,$27,$AC,$FF,$04
	.dc.b	$BD,$63,$D1,$1F,$3C,$F2,$94,$F2,$9F,$44,$F0,$9F,$78,$FC,$19,$D8
	.dc.b	$FA,$47,$81,$FA,$4E,$4F,$88,$76,$3E,$81,$F7,$8A,$64,$F4,$1F,$E4
	.dc.b	$A8,$F8,$4F,$F8,$9D,$FC,$13,$D4,$F5,$9E,$63,$FC,$E0,$FD,$67,$CE
	.dc.b	$7C,$23,$DD,$1F,$39,$F8,$4F,$D2,$72,$F4,$7F,$CF,$11,$F3,$9E,$67
	.dc.b	$8C,$F8,$C5,$3E,$33,$E1,$3F,$42,$FB,$8F,$9A,$7C,$08,$F0,$39,$7C
	.dc.b	$8F,$DD,$3E,$61,$F2,$9E,$2F,$90,$C9,$F8,$13,$E1,$1F,$EA,$4B,$FF
	.dc.b	$12,$1F,$E9,$B3,$FC,$EE,$57,$FC,$46,$D7,$FC,$E8,$D7,$FD,$0A,$D7
	.dc.b	$FD,$32,$E6,$3F,$D4,$EE,$5F,$F5,$63,$97,$FD,$64,$E5,$FF,$5A,$B9
	.dc.b	$7F,$D7,$6E,$53,$98,$A7,$FA,$D9,$F4,$B4,$F4,$94,$F4,$9F,$EB,$67
	.dc.b	$FD,$14,$7F,$1B,$2B,$93,$FD,$08,$E5,$A6,$4F,$F4,$23,$96,$99,$3F
	.dc.b	$CF,$4E,$5A,$66,$3E,$3B,$FE,$75,$8F,$0B,$FE,$D2,$6B,$FE,$C7,$6B
	.dc.b	$FE,$DB,$73,$1F,$6E,$3C,$A7,$12,$9C,$4A,$6E,$BF,$BD,$CF,$F7,$E8
	.dc.b	$C4,$7F,$AD,$A3,$11,$FE,$EB,$7F,$B8,$BF,$42,$3F,$C1,$11,$C6,$7E
	.dc.b	$EC,$FF,$84,$A3,$67,$27,$ED,$9F,$DF,$63,$94,$FF,$BC,$A7,$73,$C4
	.dc.b	$D3,$C5,$DC,$FE,$31,$D4,$FF,$87,$A3,$9C,$7A,$9D,$47,$9A,$39,$9C
	.dc.b	$1F,$CC,$B4,$FA,$6E,$CF,$8A,$E6,$3D,$B1,$FE,$0E,$9F,$F0,$9C,$79
	.dc.b	$27,$E3,$1F,$E0,$67,$F3,$6F,$F8,$F9,$F4,$9C,$A7,$EA,$1F,$E1,$49
	.dc.b	$F8,$D1,$A3,$8C,$7F,$95,$67,$AB,$B7,$A4,$F8,$8F,$F0,$53,$C6,$7E
	.dc.b	$1A,$39,$CF,$E4,$8E,$8F,$D4,$7A,$1E,$D7,$D5,$1E,$D7,$51,$C4,$E8
	.dc.b	$F5,$8F,$04,$F8,$8F,$24,$79,$8E,$6E,$5A,$7F,$96,$30,$FF,$5B,$1D
	.dc.b	$18,$1B,$22,$3A,$8F,$13,$C6,$34,$6A,$3F,$DC,$51,$98,$CB,$C6,$3E
	.dc.b	$33,$E2,$70,$F8,$23,$94,$78,$A3,$94,$62,$3C,$87,$C4,$8C,$C7,$A6
	.dc.b	$39,$9F,$EC,$E8,$F0,$39,$3E,$5C,$79,$1F,$63,$97,$D6,$F7,$3E,$34
	.dc.b	$76,$39,$9E,$98,$F8,$8F,$A6,$39,$47,$68,$E0,$FC,$43,$C3,$1E,$17
	.dc.b	$89,$E7,$3D,$26,$9E,$6F,$D0,$30,$F9,$CF,$2B,$88,$F2,$BC,$CE,$A7
	.dc.b	$37,$E9,$47,$13,$C5,$1F,$30,$FA,$A7,$33,$C6,$7C,$43,$C6,$68,$E8
	.dc.b	$7C,$A8,$F7,$BF,$E1,$76,$BF,$8E,$7E,$43,$4F,$BA,$F9,$0D,$C7,$B2
	.dc.b	$3F,$40,$F9,$9C,$1E,$38,$F7,$BE,$23,$F1,$07,$A4,$F1,$14,$F0,$1D
	.dc.b	$23,$D4,$E0,$C4,$78,$DE,$11,$E0,$3F,$0B,$1E,$08,$F1,$1F,$18,$C9
	.dc.b	$F0,$CE,$27,$E5,$CF,$11,$CE,$3F,$37,$1C,$E3,$89,$EA,$3C,$D1,$F7
	.dc.b	$DE,$46,$23,$C5,$1F,$B2,$8F,$AA,$7E,$49,$F9,$0F,$E2,$4E,$07,$E4
	.dc.b	$CA,$7E,$18,$F5,$94,$F1,$9F,$BF,$8F,$AA,$F8,$8F,$D3,$C7,$82,$3E
	.dc.b	$EC,$7D,$63,$89,$FA,$33,$E1,$9F,$42,$32,$F3,$7D,$51,$F0,$DF,$4C
	.dc.b	$7E,$CD,$FC,$B9,$F2,$4F,$E0,$8F,$F1,$BC,$7F,$83,$A7,$F8,$09,$FB
	.dc.b	$53,$F4,$27,$F0,$83,$FD,$5C,$57,$F8,$B3,$C8,$E2,$FF,$12,$78,$BF
	.dc.b	$B2,$78,$B4,$E2,$7E,$E4,$F1,$18,$3F,$78,$F2,$7E,$EB,$C9,$A6,$CF
	.dc.b	$D8,$3E,$27,$89,$FC,$61,$F4,$3F,$E8,$56,$BF,$CD,$5B,$1F,$16,$3C
	.dc.b	$C7,$F3,$E8,$CB,$FD,$39,$E2,$FF,$26,$72,$FF,$5D,$72,$FF,$53,$72
	.dc.b	$FB,$8F,$49,$D4,$EE,$7F,$0D,$34,$7E,$21,$F8,$4D,$3F,$9A,$9E,$33
	.dc.b	$89,$E3,$39,$C7,$AD,$EE,$7B,$CC,$9C,$4C,$9F,$06,$C7,$CC,$6B,$FD
	.dc.b	$8A,$7F,$C5,$D3,$FD,$85,$FF,$03,$9F,$E1,$23,$FC,$02,$59,$E2,$75
	.dc.b	$9E,$87,$67,$07,$F4,$C8,$E6,$FF,$80,$1E,$6D,$39,$9F,$06,$3C,$F3
	.dc.b	$FB,$D9,$FF,$1B,$1E,$67,$F1,$B1,$DE,$7E,$A1,$CD,$FF,$08,$BA,$7E
	.dc.b	$71,$B7,$EE,$1F,$E2,$B3,$47,$A2,$3F,$BE,$CF,$A2,$7C,$71,$89,$F9
	.dc.b	$2F,$C1,$3D,$CF,$F8,$CE,$7E,$34,$F8,$9A,$FD,$73,$9C,$FC,$F9,$E6
	.dc.b	$7C,$18,$F3,$9B,$2C,$77,$3C,$4F,$AD,$F7,$1E,$27,$C1,$1F,$0D,$E5
	.dc.b	$19,$7F,$29,$3E,$49,$FC,$74,$79,$27,$9C,$78,$DE,$0E,$A3,$C6,$E6
	.dc.b	$3C,$71,$D9,$A7,$9C,$F7,$3F,$46,$19,$7F,$18,$3F,$C2,$E7,$F8,$2C
	.dc.b	$FE,$EE,$7F,$87,$8F,$F1,$83,$FE,$44,$70,$FF,$8B,$5E,$4D,$3F,$C5
	.dc.b	$2F,$27,$FC,$12,$F2,$7F,$CB,$0F,$27,$FC,$7C,$F2,$69,$C8,$FF,$13
	.dc.b	$BB,$7F,$78,$D7,$FC,$98,$F1,$7F,$C9,$AF,$AD,$E2,$78,$8F,$DF,$9E
	.dc.b	$23,$07,$80,$FE,$EC,$7C,$63,$D8,$7F,$89,$5F,$63,$FD,$99,$F6,$3B
	.dc.b	$30,$70,$3F,$C3,$4F,$CE,$7D,$87,$F9,$81,$F6,$4F,$AA,$3C,$67,$A8
	.dc.b	$FE,$79,$1C,$5F,$11,$EC,$3B,$9F,$E7,$73,$EB,$1C,$8E,$87,$B1,$E4
	.dc.b	$FF,$84,$DF,$5B,$FD,$B1,$F5,$9F,$35,$E6,$71,$3D,$87,$F9,$61,$E3
	.dc.b	$16,$3E,$69,$F0,$CE,$87,$C3,$7F,$C1,$11,$E2,$8E,$07,$C3,$3E,$0C
	.dc.b	$7A,$1E,$27,$D5,$3E,$59,$E2,$7F,$D0,$71,$E2,$6C,$78,$A3,$D4,$FD
	.dc.b	$F3,$D4,$73,$3C,$27,$FA,$35,$FA,$6F,$C0,$3C,$27,$D3,$38,$94,$F9
	.dc.b	$47,$D3,$37,$1E,$93,$AB,$5F,$F4,$5C,$7C,$17,$CF,$1F,$04,$FB,$67
	.dc.b	$E7,$8F,$B4,$78,$DF,$92,$FC,$17,$FD,$51,$1F,$05,$B1,$F0,$4E,$27
	.dc.b	$C1,$3E,$49,$F0,$5F,$0B,$F0,$4F,$A2,$77,$72,$FF,$AE,$73,$FF,$0F
	.dc.b	$9F,$E9,$E3,$FD,$24,$7F,$9F,$8F,$F5,$29,$FA,$42,$BF,$EB,$37,$8B
	.dc.b	$4F,$F5,$7B,$C5,$FF,$47,$BC,$5A,$6C,$FD,$AB,$E1,$78,$9F,$EA,$C7
	.dc.b	$BB,$FA,$33,$FD,$CC,$E1,$FF,$68,$BE,$27,$FD,$9C,$FA,$9E,$47,$80
	.dc.b	$E4,$1E,$4F,$FA,$F1,$E9,$1F,$16,$38,$1F,$ED,$78,$CB,$FE,$EC,$78
	.dc.b	$BF,$EE,$D7,$2F,$98,$E2,$75,$39,$C7,$57,$FC,$EB,$1F,$1D,$A7,$EC
	.dc.b	$1E,$AF,$FB,$AD,$EA,$FA,$4E,$A7,$84,$EA,$79,$4C,$9C,$4C,$9E,$53
	.dc.b	$B9,$F2,$0F,$EF,$49,$FF,$05,$9E,$BA,$FF,$0A,$4F,$FB,$E2,$75,$1F
	.dc.b	$DA,$5A,$FF,$9F,$1B,$1F,$2E,$3F,$B2,$46,$A7,$11,$FE,$C5,$8C,$47
	.dc.b	$A6,$3B,$CF,$A5,$ED,$1F,$E0,$88,$F1,$3F,$E0,$03,$CD,$1E,$17,$D2
	.dc.b	$7C,$77,$FC,$0E,$D7,$F1,$2F,$F8,$CA,$39,$C7,$63,$CF,$1D,$4E,$11
	.dc.b	$FE,$FE,$9F,$3C,$78,$23,$FD,$55,$18,$8E,$C6,$CF,$97,$1C,$8E,$F3
	.dc.b	$FE,$5D,$9F,$8F,$18,$9F,$C4,$4F,$88,$FF,$19,$CE,$A7,$E2,$47,$E5
	.dc.b	$67,$E2,$46,$0F,$9A,$78,$63,$D8,$7C,$53,$9C,$70,$8F,$3B,$F8,$B8
	.dc.b	$F0,$1E,$57,$53,$F0,$27,$D4,$77,$8F,$B1,$3D,$4F,$21,$CA,$3C,$67
	.dc.b	$B6,$3B,$46,$E3,$27,$B2,$3E,$0C,$65,$EB,$1E,$B8,$FC,$B4,$4D,$3F
	.dc.b	$D7,$26,$18,$1B,$22,$3A,$8E,$47,$FA,$C5,$C3,$FA,$37,$8C,$70,$8C
	.dc.b	$C7,$81,$A7,$86,$3B,$1E,$17,$27,$F8,$52,$3A,$B9,$3D,$71,$C2,$3C
	.dc.b	$2E,$63,$FD,$CF,$1F,$E1,$63,$B4,$6E,$3D,$4E,$CF,$09,$5F,$69,$FB
	.dc.b	$D7,$31,$D8,$E6,$FB,$63,$F9,$3C,$62,$3C,$51,$E6,$8C,$1E,$37,$D3
	.dc.b	$1E,$37,$E6,$47,$C5,$69,$D9,$F8,$26,$5F,$AE,$FC,$E8,$EF,$1E,$83
	.dc.b	$D6,$60,$FA,$87,$33,$E6,$9A,$8E,$AE,$4F,$C0,$BE,$87,$FC,$3A,$D7
	.dc.b	$27,$AA,$3E,$13,$FE,$D1,$8D,$3F,$75,$B1,$F8,$F7,$31,$F1,$4E,$51
	.dc.b	$F4,$5F,$39,$F4,$5F,$8C,$FC,$03,$F2,$07,$A8,$FA,$A7,$B1,$FC,$5B
	.dc.b	$F0,$4F,$A6,$F2,$7C,$47,$23,$E7,$9E,$B8,$F8,$6F,$E4,$4E,$D1,$C5
	.dc.b	$FC,$24,$7B,$1F,$A1,$1E,$27,$E2,$9E,$97,$E2,$9E,$07,$0F,$D5,$3F
	.dc.b	$24,$7B,$8D,$3F,$9A,$3F,$2F,$1D,$DF,$8A,$64,$F4,$9F,$6C,$F3,$C7
	.dc.b	$D4,$79,$BF,$12,$3D,$B1,$E5,$6C,$66,$38,$BF,$26,$3F,$21,$1E,$43
	.dc.b	$DC,$7D,$D3,$EC,$BF,$9D,$7D,$AF,$E9,$DC,$9F,$8A,$29,$F1,$8F,$51
	.dc.b	$F8,$93,$27,$53,$99,$C4,$A7,$F1,$3D,$7F,$66,$9F,$E0,$E7,$E9,$4A
	.dc.b	$FE,$5C,$D1,$FC,$61,$D3,$4D,$3C,$1F,$CF,$35,$FD,$6B,$5F,$E1,$2E
	.dc.b	$67,$F2,$2E,$5F,$E3,$AE,$5F,$E5,$91,$E2,$3D,$47,$F0,$C8,$D4,$7F
	.dc.b	$83,$5E,$4F,$F3,$37,$31,$C4,$FE,$72,$FA,$9F,$E7,$2F,$F1,$33,$FD
	.dc.b	$6E,$57,$F9,$0B,$5F,$F1,$FB,$5F,$E6,$8D,$8F,$E8,$CE,$63,$F9,$8B
	.dc.b	$97,$FA,$6B,$97,$F9,$43,$97,$D4,$7E,$F9,$FE,$6E,$65,$F5,$47,$CA
	.dc.b	$7C,$A6,$CF,$F5,$6C,$75,$7E,$11,$93,$D2,$53,$E1,$9F,$DF,$8F,$BF
	.dc.b	$43,$27,$F8,$A2,$72,$E0,$FB,$B1,$FE,$24,$9F,$F1,$94,$FD,$A8,$C1
	.dc.b	$D6,$3F,$C3,$46,$A3,$CB,$3D,$E7,$E4,$9B,$38,$BE,$19,$E3,$19,$8E
	.dc.b	$0F,$CC,$3F,$C6,$13,$D2,$3C,$CF,$E3,$A3,$9C,$7F,$0F,$9E,$6F,$F8
	.dc.b	$E9,$AE,$63,$73,$C6,$3B,$C7,$D6,$7C,$8F,$F8,$CA,$7F,$C5,$53,$F1
	.dc.b	$8F,$5C,$FC,$B7,$EC,$3C,$9D,$1F,$1D,$D4,$FA,$DE,$11,$97,$F1,$F1
	.dc.b	$FE,$2D,$9F,$F3,$1B,$B8,$FE,$EB,$3F,$5A,$3F,$22,$6C,$FF,$28,$4F
	.dc.b	$98,$FF,$18,$15,$C9,$B9,$F3,$46,$5F,$2C,$7C,$17,$BC,$72,$8E,$11
	.dc.b	$FB,$87,$53,$F0,$23,$D2,$7B,$1F,$AC,$7C,$67,$9C,$7F,$9B,$B6,$FE
	.dc.b	$2E,$7F,$87,$0F,$EF,$45,$7F,$8B,$35,$FE,$DC,$D7,$FB,$EB,$5F,$F0
	.dc.b	$CB,$99,$FE,$C2,$E5,$FF,$16,$BF,$E2,$63,$FC,$2C,$7F,$8C,$63,$FC
	.dc.b	$8A,$E1,$FF,$12,$B5,$FF,$61,$B5,$FF,$1C,$36,$3F,$C9,$4E,$63,$FC
	.dc.b	$6A,$E5,$EA,$7F,$7E,$7C,$D1,$FC,$89,$F3,$3F,$E4,$F7,$2F,$FA,$B5
	.dc.b	$CA,$73,$14,$FF,$28,$BF,$29,$A7,$CA,$29,$F2,$8F,$69,$E3,$3F,$C9
	.dc.b	$67,$98,$FE,$EA,$FC,$77,$47,$C2,$29,$F0,$8B,$1F,$09,$C9,$F0,$8C
	.dc.b	$9D,$4F,$CD,$C7,$A8,$F6,$95,$F3,$9F,$E0,$97,$0F,$F9,$C5,$F7,$34
	.dc.b	$F7,$14,$F7,$1F,$E7,$A7,$E9,$3F,$2C,$F2,$9F,$E5,$D7,$2F,$CB,$3E
	.dc.b	$09,$F1,$8A,$7C,$12,$9F,$04,$B1,$F0,$5C,$94,$FC,$B9,$EC,$3F,$2E
	.dc.b	$7B,$9F,$CB,$B9,$7F,$2E,$E5,$FC,$33,$97,$F0,$CE,$5F,$63,$F0,$4F
	.dc.b	$90,$7A,$4F,$9E,$7D,$E3,$66,$0F,$8C,$7D,$63,$2F,$99,$C9,$EE,$3E
	.dc.b	$D3,$97,$E1,$1C,$CF,$E0,$C7,$CB,$30,$7F,$AB,$F7,$FF,$0B,$9F,$E9
	.dc.b	$C3,$FC,$F4,$57,$FC,$2C,$D7,$FC,$DC,$D7,$FC,$FA,$D7,$FD,$32,$E6
	.dc.b	$3F,$CC,$2E,$5F,$F5,$6B,$97,$FC,$22,$E5,$FF,$5F,$39,$50,$8D,$45
	.dc.b	$8F,$F5,$FC,$7A,$9A,$7F,$A9,$1E,$4E,$CF,$F6,$43,$EE,$7F,$D6,$2F
	.dc.b	$FA,$9C,$FE,$24,$F8,$1C,$3A,$3F,$D0,$AE,$9E,$87,$F4,$47,$FD,$82
	.dc.b	$70,$7F,$D7,$07,$FB,$81,$C3,$FE,$CF,$6C,$7D,$38,$EA,$77,$3F,$CE
	.dc.b	$B1,$DC,$E2,$FF,$B4,$1E,$2F,$53,$89,$FE,$E8,$72,$FA,$CC,$9E,$12
	.dc.b	$C7,$85,$EE,$7C,$AA,$FE,$94,$FC,$F7,$FB,$AB,$5F,$F4,$64,$FD,$67
	.dc.b	$2D,$7F,$0B,$1F,$E1,$F7,$FD,$FB,$1F,$E1,$59,$E5,$3F,$06,$7B,$CF
	.dc.b	$18,$C4,$76,$79,$18,$9F,$C6,$B8,$3E,$14,$7F,$8A,$E7,$94,$62,$3C
	.dc.b	$67,$99,$FF,$0D,$CF,$89,$D3,$D2,$34,$73,$8E,$C7,$CB,$7C,$EE,$0F
	.dc.b	$5C,$7E,$79,$C4,$FC,$38,$E9,$3F,$0A,$7C,$33,$E6,$7F,$42,$7C,$79
	.dc.b	$FD,$5B,$4F,$2B,$D6,$7F,$C3,$D3,$FE,$24,$3F,$C2,$F3,$E5,$8F,$CA
	.dc.b	$3E,$89,$FB,$07,$48,$FF,$07,$C6,$23,$A4,$7E,$14,$FF,$05,$3C,$23
	.dc.b	$C2,$F9,$A3,$C3,$1E,$78,$E2,$7D,$53,$9B,$F9,$73,$4F,$18,$FE,$2B
	.dc.b	$3E,$99,$F3,$36,$3C,$AD,$7C,$87,$B0,$CC,$74,$9F,$B4,$FF,$08,$E0
	.dc.b	$FF,$59,$9F,$C2,$98,$56,$A8,$82,$33,$1F,$C4,$A3,$0F,$37,$66,$1F
	.dc.b	$DF,$9E,$13,$4F,$48,$E7,$19,$3F,$DE,$87,$13,$C6,$F9,$23,$84,$78
	.dc.b	$9D,$1E,$08,$FE,$24,$7B,$1C,$47,$68,$F8,$2D,$35,$1E,$F8,$F1,$BD
	.dc.b	$0E,$F1,$EA,$30,$ED,$E4,$7A,$23,$D6,$FA,$4F,$09,$E9,$7B,$C7,$18
	.dc.b	$E9,$18,$8E,$11,$F4,$CF,$F7,$24,$74,$3E,$14,$71,$8F,$74,$71,$3E
	.dc.b	$0B,$F3,$A3,$4F,$DE,$8F,$82,$FD,$A7,$9C,$78,$23,$A3,$E8,$7F,$B3
	.dc.b	$C7,$57,$DE,$7D,$E3,$81,$EF,$8F,$8E,$F9,$0F,$13,$F5,$9F,$61,$F8
	.dc.b	$97,$6F,$94,$F7,$9D,$8F,$11,$F8,$D8,$F1,$3F,$20,$EA,$FC,$57,$89
	.dc.b	$ED,$3E,$7B,$F7,$0E,$2F,$B0,$E2,$7B,$23,$F3,$EF,$B2,$3E,$29,$C5
	.dc.b	$D9,$E1,$3D,$51,$F3,$9F,$69,$F1,$63,$C8,$FD,$37,$93,$C2,$3E,$23
	.dc.b	$FA,$13,$FD,$57,$1C,$1E,$47,$C3,$8F,$8E,$FE,$C6,$3D,$71,$F2,$63
	.dc.b	$A4,$71,$8F,$5B,$CE,$3E,$59,$C4,$F5,$9F,$0C,$E2,$7C,$38,$FE,$04
	.dc.b	$FC,$33,$E5,$9F,$B1,$8F,$0B,$F0,$23,$AB,$EB,$7C,$2F,$A5,$F8,$CF
	.dc.b	$E3,$9F,$C3,$9E,$97,$E0,$9E,$17,$E2,$39,$3A,$BF,$B2,$3E,$4B,$63
	.dc.b	$F6,$3C,$3F,$AE,$83,$87,$F6,$2F,$EF,$8F,$D8,$1F,$E1,$92,$BF,$C5
	.dc.b	$1E,$47,$17,$F5,$8F,$17,$F3,$AF,$16,$9C,$4F,$D9,$1E,$23,$07,$23
	.dc.b	$F8,$D3,$E0,$7F,$87,$35,$FE,$66,$F8,$9E,$27,$E4,$1F,$13,$F8,$B7
	.dc.b	$C4,$D3,$A9,$E4,$3A,$94,$FE,$4C,$F7,$9F,$E8,$6F,$86,$3F,$9E,$39
	.dc.b	$7F,$A4,$39,$7F,$A1,$39,$7F,$CA,$AE,$5F,$E9,$4E,$5F,$84,$79,$4F
	.dc.b	$E7,$8E,$4E,$11,$A8,$FE,$96,$D3,$CA,$FA,$4F,$29,$EB,$72,$F5,$3E
	.dc.b	$49,$FD,$B8,$FA,$55,$FE,$38,$9F,$F1,$81,$FE,$0F,$9C,$C7,$82,$3F
	.dc.b	$C1,$33,$89,$FE,$BB,$1F,$E1,$58,$C4,$7E,$9A,$30,$74,$8F,$F0,$73
	.dc.b	$FE,$1F,$8F,$ED,$6F,$43,$0F,$C6,$8F,$F1,$44,$61,$F1,$9F,$02,$47
	.dc.b	$E0,$5E,$A7,$F7,$F8,$E8,$D3,$99,$F0,$5F,$2B,$FE,$3E,$7E,$01,$E2
	.dc.b	$9F,$74,$7D,$E3,$94,$F8,$1F,$9E,$FF,$88,$A7,$BB,$63,$C6,$D3,$FC
	.dc.b	$57,$1D,$4E,$F1,$D2,$3F,$D6,$71,$98,$E3,$1F,$45,$FF,$1F,$1F,$8C
	.dc.b	$38,$1E,$27,$47,$67,$27,$E5,$0E,$D3,$F2,$63,$67,$E8,$23,$27,$BD
	.dc.b	$F4,$19,$2C,$71,$3F,$03,$1F,$E5,$DE,$2F,$E1,$C7,$F8,$00,$FE,$D6
	.dc.b	$7F,$82,$8F,$EF,$87,$FA,$94,$AF,$F8,$81,$E2,$D3,$FC,$0A,$F1,$7F
	.dc.b	$B4,$BC,$5A,$6C,$FF,$13,$BE,$17,$89,$FE,$25,$7B,$BF,$E1,$96,$BF
	.dc.b	$E3,$E7,$C2,$FF,$8F,$1F,$4B,$C4,$EE,$71,$29,$C4,$FE,$3A,$F8,$5F
	.dc.b	$F0,$83,$E9,$7F,$AF,$3E,$96,$9E,$93,$FC,$14,$F1,$7F,$B5,$BC,$5F
	.dc.b	$E0,$2F,$16,$9C,$63,$FB,$BC,$71,$8F,$F2,$8B,$C6,$2C,$7F,$95,$4F
	.dc.b	$F3,$04,$7C,$27,$FC,$C2,$FC,$27,$D8,$7A,$8F,$84,$7A,$0F,$84,$53
	.dc.b	$67,$A8,$F6,$1E,$03,$FC,$B0,$FD,$08,$EF,$1F,$28,$A7,$DA,$38,$BF
	.dc.b	$E7,$08,$F5,$36,$3D,$40,$47,$AD,$F6,$9F,$50,$E7,$1F,$E7,$B8,$E6
	.dc.b	$D8,$E6,$72,$3C,$87,$23,$C8,$7E,$14,$FC,$69,$EC,$3F,$06,$FA,$9F
	.dc.b	$AA,$6C,$E2,$7C,$63,$E2,$1F,$E7,$58,$F8,$8F,$73,$89,$F3,$23,$EE
	.dc.b	$BE,$A7,$CC,$F8,$4F,$A4,$F8,$4E,$AD,$7F,$C3,$31,$F8,$57,$E5,$47
	.dc.b	$13,$BB,$B3,$E2,$3E,$93,$EE,$14,$EA,$7F,$A8,$63,$D2,$FD,$37,$E9
	.dc.b	$3F,$57,$8F,$FC,$24,$7F,$A1,$0F,$F3,$79,$FE,$8E,$3F,$D0,$07,$FA
	.dc.b	$5D,$FF,$5A,$38,$7F,$D4,$AF,$26,$9F,$E8,$A7,$93,$FE,$6E,$79,$3E
	.dc.b	$13,$07,$03,$FD,$5A,$F8,$DE,$47,$FA,$AD,$F0,$3F,$E9,$F6,$BF,$EC
	.dc.b	$47,$C4,$FF,$B0,$9F,$5B,$C4,$EE,$71,$29,$FE,$B3,$79,$CF,$FB,$41
	.dc.b	$EB,$1F,$EC,$97,$2F,$C2,$38,$99,$39,$C7,$17,$27,$FB,$4E,$38,$BF
	.dc.b	$04,$FF,$37,$3F,$19,$A7,$E8,$1F,$0B,$C4,$EA,$78,$4D,$9F,$ED,$17
	.dc.b	$AA,$78,$BF,$EE,$77,$26,$1D,$1F,$E1,$17,$FC,$10,$7E,$E6,$BF,$B6
	.dc.b	$4F,$FB,$BA,$3F,$C2,$47,$F7,$33,$FC,$22,$59,$FF,$00,$CF,$D3,$7A
	.dc.b	$18,$3E,$23,$D2,$7C,$91,$93,$DD,$1F,$E0,$A8,$FF,$7B,$CF,$69,$EB
	.dc.b	$18,$78,$1F,$8B,$9E,$11,$FE,$84,$9F,$5C,$F8,$63,$D4,$F0,$7E,$0C
	.dc.b	$F0,$7C,$6F,$86,$3E,$F3,$98,$DB,$ED,$38,$46,$0F,$6B,$FE,$B0,$9F
	.dc.b	$8F,$1E,$89,$F0,$CF,$58,$F4,$3F,$E4,$A9,$DB,$DD,$F8,$0D,$8F,$44
	.dc.b	$7C,$F8,$E5,$1E,$49,$F8,$91,$FE,$23,$8F,$96,$F5,$7F,$C2,$E6,$E3
	.dc.b	$C7,$1F,$A1,$8F,$34,$7C,$28,$FF,$56,$C7,$9A,$3A,$C6,$CF,$21,$E2
	.dc.b	$9F,$23,$93,$C0,$66,$7A,$CF,$C7,$7C,$AF,$43,$E4,$9D,$9F,$D6,$47
	.dc.b	$63,$C6,$FA,$63,$EB,$C0

;--------------------------------------------------------------------------------
;	FSDIV.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsdiv_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FSDIV.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSDIV.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsdiv_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsdiv.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSDIV.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsdiv.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSDIV.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsdiv_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3D,$61,$F3,$7F,$DA,$BD,$E7,$99,$C6,$30,$DA,$FE
	.dc.b	$50,$D9,$DC,$7E,$72,$3F,$B8,$BF,$98,$29,$F9,$E6,$B4,$CF,$9D,$C9
	.dc.b	$E2,$D7,$FB,$DB,$5E,$86,$A3,$9C,$66,$3C,$47,$48,$C9,$1C,$10,$E4
	.dc.b	$53,$BB,$B7,$24,$32,$6C,$F1,$98,$34,$60,$87,$90,$C9,$4E,$04,$36
	.dc.b	$7C,$53,$65,$34,$71,$37,$1F,$05,$D1,$88,$F5,$3C,$0C,$9C,$48,$60
	.dc.b	$91,$CC,$91,$CA,$3A,$1F,$04,$E3,$1A,$70,$72,$8E,$0E,$9A,$E4,$DC
	.dc.b	$71,$7D,$ED,$7D,$0D,$7C,$26,$A3,$9C,$66,$3B,$1D,$23,$24,$70,$43
	.dc.b	$91,$4E,$EE,$DC,$90,$C9,$B3,$C6,$60,$D1,$82,$1E,$43,$25,$38,$10
	.dc.b	$D9,$F1,$4D,$94,$D1,$C4,$DC,$7C,$17,$46,$23,$D4,$F0,$32,$71,$21
	.dc.b	$82,$47,$32,$47,$28,$E8,$7C,$13,$8C,$69,$C1,$CA,$38,$3A,$6B,$93
	.dc.b	$71,$C5,$F7,$B5,$F4,$35,$F0,$9A,$8E,$71,$98,$EC,$74,$8C,$91,$C1
	.dc.b	$0E,$45,$3B,$BB,$72,$43,$26,$CF,$19,$83,$46,$08,$79,$0C,$94,$E0
	.dc.b	$43,$67,$C5,$36,$53,$47,$13,$71,$F0,$5D,$18,$8F,$53,$F8,$1C,$BF
	.dc.b	$D1,$17,$FC,$F2,$3F,$A1,$35,$FE,$97,$87,$F4,$47,$F9,$F9,$A2,$9A
	.dc.b	$8E,$13,$D2,$32,$DC,$3F,$9E,$36,$FC,$93,$F2,$4F,$51,$0F,$49,$70
	.dc.b	$F6,$12,$3D,$71,$C8,$F8,$B1,$97,$A1,$C8,$8E,$5D,$C7,$27,$B3,$5D
	.dc.b	$B6,$38,$47,$D0,$32,$76,$72,$F8,$CD,$38,$7A,$9B,$8E,$84,$70,$43
	.dc.b	$CA,$53,$9B,$B7,$24,$32,$6C,$F1,$10,$F0,$9B,$30,$7A,$8C,$1A,$39
	.dc.b	$98,$34,$60,$87,$28,$F7,$3A,$31,$1E,$77,$67,$53,$27,$9C,$C1,$0E
	.dc.b	$45,$8E,$A4,$8E,$91,$C8,$F3,$46,$5E,$87,$22,$39,$77,$1C,$9E,$CD
	.dc.b	$76,$D9,$FC,$C4,$7E,$86,$BF,$37,$87,$E9,$5B,$87,$E8,$A3,$F5,$12
	.dc.b	$3F,$C5,$05,$3F,$54,$53,$F5,$47,$F8,$A2,$BB,$B8,$AE,$AF,$EB,$27
	.dc.b	$C1,$7F,$E4,$2C,$9F,$37,$FE,$2E,$C3,$FC,$CA,$D8,$CD,$F5,$6B,$E1
	.dc.b	$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7C,$51,$B7,$2D,$72,$71,$29,$A3
	.dc.b	$07,$F9,$61,$E0,$D3,$27,$F9,$7A,$3C,$0E,$4E,$44,$3A,$90,$D1,$4E
	.dc.b	$51,$D9,$A7,$69,$F8,$6F,$33,$89,$0D,$99,$29,$D5,$E2,$D9,$F3,$BA
	.dc.b	$39,$46,$9B,$1A,$8C,$46,$C8,$E0,$91,$B2,$4F,$AE,$32,$D8,$CC,$FB
	.dc.b	$5A,$FA,$67,$6E,$67,$81,$1C,$10,$E1,$3C,$9F,$74,$6D,$CB,$5C,$9C
	.dc.b	$4A,$68,$C1,$F0,$9E,$0D,$32,$7C,$28,$F0,$39,$39,$10,$EA,$43,$45
	.dc.b	$39,$47,$66,$9D,$A7,$E1,$BC,$CE,$24,$36,$64,$A7,$57,$8B,$67,$CE
	.dc.b	$E8,$E5,$1A,$6C,$6A,$31,$1B,$23,$82,$46,$C9,$3E,$B8,$CB,$63,$33
	.dc.b	$ED,$6B,$E9,$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7D,$D1,$B7,$2D,$72
	.dc.b	$71,$29,$A3,$07,$C2,$78,$34,$C9,$F0,$A3,$C0,$E4,$E4,$43,$A9,$0D
	.dc.b	$14,$E5,$1D,$9A,$76,$9F,$86,$F3,$38,$90,$D9,$92,$9D,$5E,$2D,$9F
	.dc.b	$3B,$A3,$94,$7E,$F9,$FE,$8D,$F2,$DF,$EE,$44,$D2,$3F,$EE,$86,$BF
	.dc.b	$EE,$C6,$BF,$EF,$06,$BF,$D3,$5A,$FF,$51,$6B,$E0,$3A,$10,$E8,$53
	.dc.b	$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10
	.dc.b	$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$A6,$7B,$4A,$79
	.dc.b	$8A,$78,$4A,$76,$29,$D4,$A7,$57,$E9,$B8,$23,$82,$3E,$16,$BE,$16
	.dc.b	$BE,$16,$BE,$16,$BE,$16,$BE,$13,$A1,$0E,$85,$3A,$14,$E8,$53,$A1
	.dc.b	$4E,$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$10,$FA,$67,$B4,$A7,$98,$A7,$84,$A7,$62
	.dc.b	$9D,$4A,$75,$7E,$9B,$82,$38,$23,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$6B
	.dc.b	$E1,$6B,$E1,$3A,$10,$E8,$53,$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8
	.dc.b	$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$FE,$09,$1F,$E7,$22,$61,$FE,$70,$8F,$F4,$16,$1F,$E7,$C6,$1F
	.dc.b	$E7,$C2,$8F,$83,$7C,$0B,$E3,$FF,$D0,$13,$FE,$85,$9C,$C6,$23,$84
	.dc.b	$ED,$DC,$E9,$FF,$4C,$9F,$E8,$DC,$DF,$09,$FE,$8D,$DD,$8F,$FA,$3E
	.dc.b	$73,$5F,$E9,$39,$CB,$C6,$BF,$D2,$9F,$88,$F7,$8F,$F4,$5C,$7F,$A5
	.dc.b	$FD,$27,$9F,$CE,$3F,$D3,$5F,$06,$F4,$31,$7F,$E9,$F8,$C3,$D4,$FF
	.dc.b	$51,$5F,$27,$01,$C3,$F1,$1A,$FF,$A9,$5A,$F8,$EF,$E3,$3C,$5F,$0B
	.dc.b	$4E,$B1,$D3,$2F,$83,$0A,$78,$7C,$38,$7F,$D4,$3F,$42,$E9,$FF,$5B
	.dc.b	$75,$BE,$07,$FD,$71,$39,$C7,$FD,$79,$39,$7F,$D7,$D1,$97,$FD,$16
	.dc.b	$DB,$EB,$7D,$2F,$9B,$6B,$8B,$FE,$A6,$AE,$1C,$FF,$EB,$87,$A6,$1D
	.dc.b	$E9,$18,$BE,$71,$FE,$A7,$8E,$39,$F4,$8F,$3C,$E9,$57,$3A,$DC,$74
	.dc.b	$C3,$C4,$79,$5C,$1C,$8E,$87,$B9,$FF,$5E,$FD,$07,$8E,$3F,$DC,$0D
	.dc.b	$4F,$63,$BC,$E4,$93,$E0,$9F,$53,$E1,$8C,$C6,$08,$7A,$AF,$D9,$7E
	.dc.b	$BA,$E5,$3E,$53,$D0,$7A,$8E,$07,$AB,$2F,$AA,$F4,$9E,$4F,$0B,$F3
	.dc.b	$D7,$49,$CC,$F2,$3A,$57,$27,$C0,$66,$BD,$2E,$23,$53,$C8,$E6,$F1
	.dc.b	$33,$7E,$C3,$EC,$39,$CB,$F0,$98,$7C,$52,$5F,$C4,$9E,$2F,$D6,$BE
	.dc.b	$B5,$C2,$3D,$D7,$D5,$F8,$51,$97,$D3,$3E,$59,$F0,$CE,$6B,$AD,$7C
	.dc.b	$BA,$EB,$97,$23,$C1,$7D,$67,$E2,$DF,$13,$B5,$75,$8E,$06,$AB,$8B
	.dc.b	$0C,$13,$1F,$97,$3E,$08,$C6,$3F,$09,$FA,$0F,$E0,$B1,$F3,$D7,$D2
	.dc.b	$7E,$55,$F5,$8C,$61,$E5,$8C,$57,$3A,$DD,$FB,$F1,$F6,$61,$F8,$48
	.dc.b	$09,$60,$C3,$F4,$78,$7E,$AF,$2F,$D2,$1F,$A3,$7F,$19,$8F,$47,$D1
	.dc.b	$5F,$1A,$31,$87,$DD,$8C,$3E,$63,$17,$5F,$82,$60,$F4,$19,$C3,$D7
	.dc.b	$8F,$CA,$BF,$BB,$5F,$4A,$31,$57,$0F,$2B,$C7,$0F,$8C,$FE,$F2,$3D
	.dc.b	$84,$7E,$1E,$1F,$0A,$FE,$0C,$7E,$40,$E4,$F5,$7F,$5D,$3F,$3E,$BA
	.dc.b	$DF,$8A,$7E,$E9,$C4,$F5,$98,$3E,$84,$7C,$F7,$C8,$7F,$00,$C7,$E7
	.dc.b	$D7,$F1,$63,$D7,$1C,$AB,$ED,$3F,$A0,$C7,$E7,$BF,$B7,$BF,$9E,$EE
	.dc.b	$BE,$AE,$1E,$C8,$FE,$05,$87,$67,$E0,$D7,$8E,$3A,$E5,$F9,$F9,$F4
	.dc.b	$65,$F9,$4C,$3E,$06,$1F,$A2,$BE,$55,$D2,$7D,$0F,$C1,$CB,$F2,$F5
	.dc.b	$B9,$FA,$13,$B3,$C3,$1E,$5B,$ED,$1F,$38,$FC,$25,$F6,$C3,$FA,$63
	.dc.b	$F9,$38,$ED,$7E,$47,$F6,$31,$BC,$BF,$3F,$1F,$38,$FA,$38,$77,$CB
	.dc.b	$E3,$E5,$E6,$9F,$AD,$8F,$C1,$8F,$9F,$97,$92,$7F,$2E,$7C,$FB,$F9
	.dc.b	$95,$E4,$CB,$DB,$5F,$8B,$AF,$03,$C6,$7F,$81,$5F,$A6,$7E,$11,$27
	.dc.b	$86,$5F,$4B,$2F,$BB,$1F,$A2,$C3,$F8,$46,$1C,$9F,$C2,$E1,$E9,$3A
	.dc.b	$47,$8F,$0F,$C6,$4F,$D4,$7F,$10,$7E,$70,$F7,$DF,$6C,$BF,$4F,$8F
	.dc.b	$C9,$9F,$8D,$94,$7F,$18,$7D,$53,$E1,$18,$3E,$2E,$5E,$77,$58,$F9
	.dc.b	$DF,$AE,$71,$8D,$3F,$32,$34,$F9,$8E,$91,$A8,$FA,$98,$FE,$72,$7C
	.dc.b	$11,$C2,$7E,$76,$3F,$C1,$A7,$F9,$9F,$37,$F7,$C7,$F1,$E7,$F9,$77
	.dc.b	$1F,$C2,$D7,$D7,$9D,$4F,$E1,$8D,$1F,$8A,$3F,$DD,$AE,$9F,$E5,$7D
	.dc.b	$1F,$CB,$23,$11,$AA,$FF,$34,$CE,$5B,$1F,$DF,$7E,$79,$F2,$3F,$E6
	.dc.b	$0D,$BF,$C6,$93,$D0,$D4,$67,$4E,$33,$FE,$39,$AE,$3C,$9F,$E1,$EB
	.dc.b	$E1,$39,$BE,$11,$E9,$C7,$FC,$47,$7C,$23,$FC,$1C,$70,$C3,$F4,$F1
	.dc.b	$9B,$F1,$CA,$B8,$DF,$48,$FD,$6C,$63,$0F,$14,$6C,$F4,$B8,$8D,$47
	.dc.b	$80,$D5,$79,$E7,$51,$C0,$D4,$F0,$9F,$1B,$70,$F8,$67,$A2,$FC,$95
	.dc.b	$EB,$3F,$AF,$E1,$EB,$9D,$C6,$27,$6F,$8A,$39,$61,$E2,$C7,$C9,$3F
	.dc.b	$C2,$4F,$A8,$E0,$F1,$BE,$7C,$69,$C0,$B5,$D2,$3D,$66,$DF,$C1,$CF
	.dc.b	$36,$9B,$78,$C7,$9E,$BE,$2D,$75,$AF,$5C,$B0,$F7,$3F,$7C,$EC,$59
	.dc.b	$E0,$D9,$F2,$47,$AA,$FC,$53,$C5,$AF,$DE,$39,$96,$39,$B9,$8D,$BF
	.dc.b	$1A,$7D,$47,$9A,$BC,$B1,$C6,$3C,$26,$70,$FA,$78,$FD,$D9,$F3,$1F
	.dc.b	$04,$F9,$38,$78,$1E,$C6,$70,$F5,$5F,$82,$38,$9F,$16,$FA,$BE,$A3
	.dc.b	$89,$9C,$7F,$B3,$38,$C7,$F4,$D7,$F4,$A3,$47,$8A,$FD,$2F,$C7,$7C
	.dc.b	$0F,$C3,$6B,$F9,$33,$8E,$1C,$EB,$E1,$C7,$7C,$3F,$DC,$37,$EA,$8F
	.dc.b	$65,$FC,$5C,$3F,$02,$F8,$23,$E5,$1F,$17,$0E,$97,$EE,$BF,$A7,$3F
	.dc.b	$B3,$6E,$3F,$36,$38,$9E,$4B,$EC,$FB,$4E,$8E,$2F,$AE,$1F,$A5,$8F
	.dc.b	$6C,$F5,$6D,$75,$8F,$81,$8F,$B9,$FD,$71,$D0,$FA,$AE,$63,$84,$68
	.dc.b	$F2,$1A,$C7,$EC,$E1,$F8,$18,$F0,$1F,$18,$FA,$45,$3D,$B5,$D5,$FC
	.dc.b	$01,$D5,$C4,$73,$AF,$09,$27,$6D,$8D,$46,$08,$E0,$97,$F5,$63,$E0
	.dc.b	$DF,$A2,$31,$1E,$83,$F1,$78,$FD,$F6,$C7,$CF,$23,$F0,$71,$FC,$0B
	.dc.b	$F2,$4F,$6D,$75,$70,$78,$23,$6F,$CA,$36,$E3,$1F,$83,$5E,$4C,$7D
	.dc.b	$CD,$AF,$6B,$71,$F6,$C7,$63,$E5,$61,$E1,$7B,$1E,$13,$F1,$46,$0D
	.dc.b	$E1,$E1,$8E,$18,$78,$A7,$A6,$1E,$26,$9E,$22,$DF,$89,$E9,$1B,$69
	.dc.b	$B2,$C6,$E7,$B4,$63,$0F,$61,$1E,$AD,$74,$60,$87,$6C,$7F,$56,$FF
	.dc.b	$93,$67,$F8,$3C,$FD,$88,$F9,$E7,$C2,$8F,$09,$30,$F8,$24,$C7,$E2
	.dc.b	$12,$78,$47,$F6,$FF,$97,$7A,$6A,$FF,$5D,$F0,$5B,$FF,$5E,$D6,$A7
	.dc.b	$27,$FB,$03,$E7,$1F,$7A,$F8,$BE,$18,$71,$C1,$87,$17,$FD,$89,$87
	.dc.b	$F9,$AA,$31,$1F,$E2,$36,$B9,$C7,$93,$67,$FC,$50,$F9,$23,$BC,$7F
	.dc.b	$74,$FB,$87,$9F,$8A,$D4,$FA,$1F,$EB,$93,$A8,$C4,$3E,$65,$F0,$5C
	.dc.b	$63,$0F,$3D,$DE,$BB,$DF,$D5,$F0,$C3,$9B,$70,$F3,$B5,$FF,$20,$35
	.dc.b	$FF,$45,$B5,$FF,$75,$F8,$6E,$E7,$CE,$7C,$09,$4E,$63,$0D,$8F,$8A
	.dc.b	$D7,$FC,$9D,$D4,$F1,$9E,$47,$38,$C3,$63,$33,$F3,$A7,$2F,$4F,$1D
	.dc.b	$DB,$E0,$6C,$66,$30,$79,$FC,$67,$8F,$FE,$A8,$E7,$FC,$27,$8F,$F2
	.dc.b	$6F,$E8,$B5,$FC,$28,$15,$C9,$FA,$ED,$7F,$06,$D7,$FC,$D8,$D7,$FD
	.dc.b	$6E,$D7,$A6,$CF,$1B,$C9,$C3,$FE,$BC,$F1,$9E,$3E,$2F,$C5,$18,$7F
	.dc.b	$0F,$7C,$5B,$7B,$53,$B7,$D6,$62,$B5,$5C,$63,$2D,$7F,$D0,$8D,$7F
	.dc.b	$D9,$ED,$7D,$35,$F8,$37,$FD,$39,$E1,$7F,$AD,$FC,$1F,$F5,$C9,$FE
	.dc.b	$D9,$7F,$D8,$11,$FE,$B5,$6B,$FE,$A2,$F1,$9A,$FC,$F9,$C9,$CE,$7F
	.dc.b	$D5,$A7,$FA,$CA,$76,$FF,$AF,$8E,$95,$E0,$7F,$D6,$51,$FE,$C0,$8F
	.dc.b	$F7,$AF,$8D,$FE,$C9,$9F,$F5,$17,$89,$DA,$7C,$C7,$18,$ED,$87,$8E
	.dc.b	$31,$5F,$EB,$3B,$FF,$76,$C7,$0A,$F2,$CE,$6B,$84,$7F,$B1,$BC,$6F
	.dc.b	$F5,$C6,$3D,$E3,$CB,$3A,$7E,$4D,$F3,$69,$A3,$DC,$7C,$78,$E1,$C9
	.dc.b	$FE,$C4,$8B,$8F,$C4,$BE,$D1,$EA,$38,$BF,$2D,$EA,$65,$DE,$3E,$98
	.dc.b	$FF,$68,$E3,$E8,$C7,$E0,$B6,$3D,$ED,$79,$CF,$1A,$57,$FB,$F2,$B5
	.dc.b	$87,$9E,$D5,$A7,$FD,$B2,$F8,$A7,$78,$7C,$E9,$F8,$53,$97,$DF,$8C
	.dc.b	$7C,$E5,$8D,$3D,$B0,$FF,$76,$47,$78,$FF,$7A,$5F,$6A,$FA,$58,$FF
	.dc.b	$BC,$A7,$C7,$7F,$01,$E8,$D7,$A1,$A3,$E1,$9A,$29,$A3,$FD,$E1,$7E
	.dc.b	$43,$C5,$8F,$FC,$01,$C3,$F8,$74,$FD,$D3,$EB,$D7,$E0,$E7,$E4,$57
	.dc.b	$0C,$7E,$F4,$F8,$E3,$18,$F8,$6F,$DA,$5B,$E3,$5F,$08,$FA,$91,$9B
	.dc.b	$F9,$CF,$E0,$E3,$A1,$B7,$F3,$46,$EB,$E3,$E1,$E8,$C3,$F1,$98,$76
	.dc.b	$AF,$88,$7E,$12,$7E,$73,$F5,$8F,$89,$1F,$05,$FD,$C9,$F2,$27,$D6
	.dc.b	$FC,$23,$F0,$37,$E4,$7A,$CF,$77,$E6,$C6,$5A,$73,$7D,$4F,$C4,$3C
	.dc.b	$38,$7C,$A7,$47,$47,$E1,$E3,$F9,$B9,$F8,$F3,$C3,$1F,$9D,$3F,$87
	.dc.b	$8E,$E7,$BC,$F8,$C7,$BF,$0F,$59,$F8,$1A,$F1,$CF,$1C,$18,$7F,$17
	.dc.b	$BF,$EA,$13,$FC,$5C,$FE,$31,$3A,$C3,$FA,$A6,$5C,$EB,$D1,$1F,$51
	.dc.b	$F9,$67,$8C,$F4,$1F,$2C,$FB,$C7,$C0,$3D,$33,$F2,$CE,$CF,$0E,$35
	.dc.b	$7E,$5A,$73,$18,$9F,$EC,$D1,$89,$D6,$1F,$CD,$AB,$55,$FE,$EF,$6B
	.dc.b	$FD,$B7,$C6,$FE,$DD,$18,$84,$62,$BF,$C5,$ED,$8F,$F4,$41,$FE,$4A
	.dc.b	$E1,$FF,$26,$69,$FE,$BA,$8C,$11,$C1,$34,$FF,$2D,$47,$F9,$EE,$30
	.dc.b	$F4,$7F,$D3,$1D,$BF,$E6,$08,$FF,$58,$9F,$E9,$A3,$53,$C9,$FF,$40
	.dc.b	$9D,$DC,$5F,$F9,$E6,$7F,$CC,$97,$E6,$9C,$D7,$F9,$EE,$72,$F1,$3A
	.dc.b	$90,$FF,$37,$DF,$99,$B3,$CA,$3F,$D0,$5E,$37,$F9,$BE,$79,$1A,$39
	.dc.b	$1E,$F9,$E0,$7F,$A0,$EB,$C3,$5D,$67,$47,$03,$DA,$E5,$B3,$C9,$FF
	.dc.b	$44,$CF,$28,$C4,$7F,$A3,$89,$19,$7A,$61,$F1,$6F,$D1,$7D,$63,$18
	.dc.b	$7C,$7A,$D9,$F1,$EB,$78,$76,$C3,$A4,$FB,$C8,$7A,$8A,$7A,$0E,$78
	.dc.b	$7C,$8B,$ED,$1A,$C3,$C2,$FC,$A3,$91,$0E,$46,$CF,$24,$72,$75,$1C
	.dc.b	$A7,$D9,$1A,$3A,$1C,$8D,$99,$76,$79,$5D,$C7,$3B,$F8,$18,$7C,$29
	.dc.b	$F5,$C7,$DB,$8C,$47,$96,$B9,$1D,$27,$57,$E7,$9F,$11,$2F,$B3,$D0
	.dc.b	$FB,$B7,$D1,$F3,$DF,$4B,$F1,$DF,$3B,$FE,$2B,$1F,$D7,$30,$FD,$55
	.dc.b	$FF,$5E,$9C,$BF,$E6,$D8,$CB,$D2,$7C,$AF,$C0,$3E,$23,$D9,$C5,$F8
	.dc.b	$5F,$31,$F8,$EC,$3C,$CF,$D6,$F0,$FF,$D7,$6F,$FA,$C8,$FF,$5C,$B8
	.dc.b	$7F,$D8,$BE,$1F,$FA,$F9,$D6,$FF,$ED,$1C,$5E,$17,$FA,$DE,$7F,$CD
	.dc.b	$77,$FE,$CB,$AD,$4E,$5F,$F6,$84,$67,$4F,$F6,$27,$87,$FE,$BB,$8C
	.dc.b	$5F,$FB,$0A,$3F,$D7,$3E,$1F,$86,$3F,$D7,$F5,$FE,$AF,$F0,$7F,$D6
	.dc.b	$91,$83,$FD,$77,$3F,$EC,$DC,$1F,$86,$BC,$EE,$78,$7F,$B3,$E7,$31
	.dc.b	$FE,$CB,$F0,$BF,$D6,$98,$7F,$84,$1A,$FF,$B0,$3A,$BF,$D7,$EF,$FA
	.dc.b	$E9,$AF,$FA,$EB,$C1,$FF,$5A,$B7,$C2,$FF,$5C,$11,$FF,$59,$10,$FF
	.dc.b	$0E,$9F,$EC,$8F,$0B,$FD,$6E,$E2,$D1,$88,$FF,$11,$FC,$53,$E1,$BF
	.dc.b	$D3,$8D,$8F,$F7,$3B,$5F,$F4,$EF,$C5,$3E,$63,$75,$27,$6F,$FA,$96
	.dc.b	$75,$1F,$EB,$AF,$1B,$FD,$35,$E2,$FF,$AD,$63,$FD,$7F,$18,$23,$B2
	.dc.b	$1F,$EA,$CF,$1B,$D7,$C0,$E1,$FF,$5F,$D7,$FA,$CE,$FF,$D8,$D3,$9B
	.dc.b	$FF,$66,$46,$2B,$FC,$83,$7C,$9F,$F0,$D7,$83,$FE,$25,$AF,$F2,$27
	.dc.b	$67,$F7,$F7,$FB,$BF,$83,$FE,$16,$F0,$FF,$C5,$BE,$0F,$F9,$77,$C3
	.dc.b	$FF,$04,$47,$F9,$32,$3F,$C5,$67,$1C,$3A,$61,$FE,$38,$BF,$F5,$5C
	.dc.b	$F1,$AE,$1E,$0F,$F9,$53,$0F,$F0,$8D,$6A,$3C,$AF,$F8,$A7,$1F,$F0
	.dc.b	$76,$1F,$E1,$E9,$FF,$00,$EB,$FB,$E9,$5F,$91,$C6,$5F,$E0,$EC,$BC
	.dc.b	$98,$F3,$8F,$F0,$93,$FE,$23,$9E,$39,$79,$1F,$F1,$2B,$5F,$F5,$DE
	.dc.b	$3E,$4C,$7F,$C6,$B3,$FE,$55,$8F,$F3,$55,$7A,$6B,$89,$E5,$C7,$DB
	.dc.b	$7E,$49,$F9,$93,$FE,$3A,$CB,$E2,$3F,$E2,$CC,$7D,$18,$7F,$B3,$31
	.dc.b	$F1,$46,$27,$FC,$58,$FC,$E9,$F0,$E1,$DE,$BD,$18,$F8,$2F,$FC,$5B
	.dc.b	$5E,$97,$FC,$6D,$1F,$EB,$3C,$BC,$31,$A3,$D2,$53,$E0,$9E,$0C,$BE
	.dc.b	$51,$EA,$7F,$C8,$39,$7E,$0A,$FD,$11,$EF,$C6,$3F,$E5,$78,$FF,$2A
	.dc.b	$E3,$F2,$63,$FC,$E2,$75,$9F,$9A,$FC,$3B,$ED,$7F,$81,$8D,$1E,$AA
	.dc.b	$FF,$40,$C7,$A8,$FC,$85,$7F,$95,$E7,$F3,$B7,$F0,$67,$F0,$73,$DA
	.dc.b	$72,$74,$8F,$53,$F0,$B0,$F8,$A6,$1F,$C9,$D7,$A9,$F3,$57,$7A,$F0
	.dc.b	$C7,$BC,$FF,$34,$94,$FD,$6C,$6E,$7E,$CC,$7C,$FC,$7F,$34,$FD,$1C
	.dc.b	$BF,$00,$F8,$0E,$77,$EF,$AE,$2D,$8F,$96,$6C,$EE,$78,$67,$FC,$E3
	.dc.b	$1E,$BA,$EF,$87,$B6,$7F,$4B,$87,$C5,$AF,$59,$D6,$FF,$CF,$A7,$E1
	.dc.b	$0F,$D4,$9F,$91,$3E,$FD,$7C,$E7,$FD,$0A,$FC,$AC,$3F,$80,$63,$F9
	.dc.b	$CC,$BF,$51,$5E,$7C,$7F,$38,$FA,$9F,$1B,$CA,$BF,$1A,$FC,$59,$E3
	.dc.b	$3A,$C3,$C0,$7B,$6F,$F4,$67,$7C,$7F,$27,$1E,$BC,$BF,$D6,$58,$7F
	.dc.b	$B4,$8F,$CD,$E3,$FC,$39,$FA,$97,$C9,$F6,$C7,$E2,$C8,$FE,$25,$E1
	.dc.b	$97,$BE,$3F,$15,$18,$36,$FC,$9A,$FA,$31,$98,$FF,$4F,$E5,$F1,$27
	.dc.b	$31,$8C,$7F,$57,$E1,$FF,$A8,$33,$FF,$52,$4E,$44,$66,$FF,$D4,$D8
	.dc.b	$FF,$A7,$A7,$2D,$9F,$F5,$27,$87,$FE,$A1,$9F,$F5,$54,$EC,$65,$FE
	.dc.b	$2F,$F0,$FF,$A7,$47,$FA,$69,$B1,$B3,$5B,$7F,$35,$AF,$F5,$91,$FC
	.dc.b	$2F,$C2,$FF,$19,$BF,$E4,$31,$BF,$67,$FC,$71,$7F,$D6,$E3,$B5,$73
	.dc.b	$9F,$F5,$C9,$CE,$7C,$2F,$F8,$DA,$BF,$CE,$EF,$F8,$FC,$FF,$26,$C9
	.dc.b	$EF,$7E,$77,$BC,$75,$C7,$FA,$65,$79,$9F,$F2,$BC,$FF,$97,$EB,$DA
	.dc.b	$F2,$3F,$D7,$33,$B7,$C9,$5C,$DF,$F2,$69,$E5,$BF,$65,$78,$63,$FD
	.dc.b	$80,$7F,$96,$4E,$27,$F3,$13,$D4,$7F,$94,$67,$FC,$C4,$F4,$38,$63
	.dc.b	$F2,$5E,$67,$D0,$71,$7E,$5C,$7F,$CB,$93,$C2,$FD,$75,$F3,$A3,$C0
	.dc.b	$7D,$59,$DB,$E2,$AF,$9A,$FD,$29,$EB,$87,$D2,$8E,$87,$CC,$39,$E3
	.dc.b	$F2,$9E,$51,$FE,$6A,$8C,$47,$7C,$3F,$CE,$B8,$78,$EE,$47,$F9,$D6
	.dc.b	$3F,$D1,$D3,$F7,$0F,$F3,$BC,$70,$34,$47,$EA,$DF,$95,$F8,$91,$F0
	.dc.b	$0D,$3F,$74,$E9,$3D,$9F,$A0,$76,$33,$1C,$1F,$5E,$3F,$52,$7E,$3C
	.dc.b	$7A,$CD,$9F,$81,$8D,$4F,$9E,$3C,$B1,$F8,$7B,$F9,$6F,$E2,$6B,$9E
	.dc.b	$1F,$18,$FC,$AE,$1E,$BB,$FA,$58,$7A,$DE,$C7,$9C,$62,$9F,$C8,$C7
	.dc.b	$80,$FA,$A7,$E4,$31,$F8,$13,$F7,$AF,$F5,$18,$79,$2B,$F4,$07,$DE
	.dc.b	$3F,$1E,$78,$8F,$C0,$57,$2B,$FB,$D1,$C2,$FC,$58,$F4,$C7,$F4,$51
	.dc.b	$FA,$9B,$EC,$71,$9F,$8F,$5F,$8B,$C3,$F0,$6F,$A6,$74,$FB,$EB,$AB
	.dc.b	$E6,$3E,$F9,$E4,$BE,$F8,$7C,$9C,$7E,$64,$FE,$AC,$D9,$EC,$7E,$7E
	.dc.b	$3F,$4E,$BE,$C4,$FE,$1C,$F9,$B5,$F3,$1F,$89,$7E,$B9,$FA,$91,$B8
	.dc.b	$F9,$8F,$73,$F0,$C7,$E3,$B1,$FD,$D7,$7B,$BF,$FD,$5F,$2D,$FF,$D9
	.dc.b	$1E,$4B,$C9,$CC,$7F,$A0,$67,$FD,$74,$DB,$E5,$A7,$F2,$18,$E3,$5C
	.dc.b	$23,$F9,$25,$7F,$B0,$4F,$F5,$5D,$7F,$B1,$09,$87,$FB,$27,$E6,$DF
	.dc.b	$11,$FE,$8F,$C7,$C1,$18,$C3,$B1,$FE,$F1,$C9,$E9,$6E,$3B,$CE,$F2
	.dc.b	$F6,$46,$30,$D4,$7F,$37,$9D,$61,$F0,$2F,$32,$9C,$DF,$FB,$3F,$C2
	.dc.b	$FF,$65,$CD,$F3,$3F,$D8,$AF,$FB,$45,$AB,$C1,$FF,$6A,$AE,$77,$87
	.dc.b	$FE,$D2,$8F,$8B,$18,$F0,$FF,$DA,$51,$FE,$D5,$3F,$DA,$47,$FB,$7F
	.dc.b	$C3,$FF,$67,$3F,$D3,$EF,$FD,$A3,$D3,$E0,$6C,$79,$3C,$3F,$F6,$73
	.dc.b	$FE,$DA,$6B,$FE,$D6,$F0,$BC,$EA,$32,$FF,$5B,$7F,$DA,$DC,$1F,$81
	.dc.b	$6E,$DC,$67,$E9,$9F,$D8,$7C,$2F,$F6,$AB,$F4,$0F,$F6,$3F,$85,$F8
	.dc.b	$2B,$FC,$57,$83,$F8,$D9,$E2,$1C,$3C,$4F,$F6,$D9,$FE,$D8,$3F,$DE
	.dc.b	$7E,$17,$FB,$7C,$95,$FE,$DE,$8E,$11,$FE,$D5,$7C,$27,$69,$FF,$6B
	.dc.b	$3F,$E2,$B2,$1F,$E4,$5F,$03,$FD,$23,$7F,$E6,$98,$FF,$23,$47,$FA
	.dc.b	$1B,$B7,$FC,$71,$D9,$FE,$41,$9F,$F2,$A1,$FE,$47,$AF,$F2,$0C,$FF
	.dc.b	$94,$A3,$FC,$F2,$7F,$99,$6B,$AB,$67,$FC,$C5,$3F,$E6,$E3,$84,$7F
	.dc.b	$A2,$0F,$24,$69,$F1,$BE,$57,$FC,$CF,$C5,$FE,$55,$9F,$F2,$0B,$63
	.dc.b	$DA,$D7,$99,$FE,$9D,$3F,$D8,$D1,$E0,$BF,$F3,$9B,$4F,$24,$7F,$96
	.dc.b	$63,$FD,$4C,$F2,$7F,$CC,$71,$E0,$BF,$8D,$3F,$E8,$A8,$E4,$FA,$0F
	.dc.b	$F2,$F9,$F2,$3C,$0F,$F4,$24,$F0,$8F,$F4,$5D,$7F,$9A,$7C,$2F,$F5
	.dc.b	$8D,$F2,$9F,$7D,$6F,$0F,$F3,$DC,$63,$1F,$89,$E1,$7F,$9C,$23,$FD
	.dc.b	$12,$7C,$B9,$F8,$47,$FA,$82,$3F,$CC,$31,$FE,$8F,$8C,$D7,$53,$EB
	.dc.b	$E3,$E3,$9F,$8E,$7C,$3C,$3F,$06,$7F,$9B,$5E,$F9,$7F,$9D,$5E,$95
	.dc.b	$CF,$1F,$B5,$1E,$73,$F0,$47,$FA,$44,$FF,$3D,$90,$E2,$FB,$32,$F8
	.dc.b	$71,$E1,$3F,$D8,$35,$DF,$2F,$F7,$A3,$F5,$23,$FC,$F0,$E6,$7F,$D2
	.dc.b	$D1,$F9,$73,$C3,$8F,$E6,$DF,$AA,$66,$79,$3F,$E8,$B9,$FF,$4B,$39
	.dc.b	$C7,$F0,$18,$79,$63,$E2,$9F,$42,$FF,$DA,$F7,$E1,$8F,$F4,$E9,$F9
	.dc.b	$4C,$7F,$D2,$58,$74,$9F,$C1,$DF,$3A,$FF,$53,$9B,$71,$7F,$92,$CB
	.dc.b	$ED,$5F,$1C,$7E,$94,$7D,$AA,$DC,$7D,$6B,$F8,$6F,$FA,$FB,$0E,$F1
	.dc.b	$F0,$A3,$D6,$D3,$C0,$53,$F4,$0E,$B0,$FC,$24,$79,$8C,$9F,$EA,$7E
	.dc.b	$3F,$F5,$35,$FF,$A8,$70,$FF,$55,$CF,$F4,$3C,$BF,$D5,$2F,$F6,$63
	.dc.b	$FD,$23,$5F,$EA,$87,$FD,$56,$67,$C2,$FF,$55,$3F,$EA,$08,$D6,$1F
	.dc.b	$E3,$A8,$FF,$00,$4F,$F8,$42,$30,$4C,$7F,$C1,$B3,$FE,$4C,$3F,$C3
	.dc.b	$9E,$17,$F8,$AA,$7F,$C6,$47,$F8,$8A,$3F,$C1,$F1,$FE,$1B,$7A,$93
	.dc.b	$3F,$F1,$5D,$7F,$88,$38,$BF,$C3,$53,$FE,$54,$3F,$C2,$95,$D6,$7F
	.dc.b	$D3,$B7,$D6,$7F,$C4,$7E,$17,$FA,$26,$7C,$B1,$FE,$3F,$3F,$C4,$C7
	.dc.b	$F9,$25,$FF,$32,$3F,$E1,$C8,$FF,$0F,$CF,$A2,$7C,$53,$FE,$50,$C3
	.dc.b	$CA,$FF,$96,$27,$C0,$F5,$AF,$03,$E7,$3C,$77,$EF,$AE,$31,$EA,$C3
	.dc.b	$FC,$98,$FF,$91,$70,$EF,$3F,$E4,$6C,$24,$60,$93,$E1,$7D,$38,$FF
	.dc.b	$91,$2B,$AC,$F8,$2F,$E4,$BF,$E3,$7A,$DB,$F4,$2F,$EF,$57,$08,$F5
	.dc.b	$61,$F5,$67,$EF,$47,$AC,$F8,$D1,$F4,$5C,$61,$F8,$5B,$F9,$ED,$8F
	.dc.b	$F3,$14,$7A,$1F,$C0,$E1,$E4,$C7,$E7,$CF,$09,$F2,$37,$0F,$0C,$7A
	.dc.b	$4F,$80,$43,$E5,$1F,$90,$AF,$F3,$8B,$B3,$31,$F2,$9E,$07,$E5,$DC
	.dc.b	$3A,$9F,$8B,$1E,$0A,$F2,$CF,$B8,$F2,$9E,$C3,$EC,$10,$E4,$76,$3E
	.dc.b	$3D,$79,$23,$9C,$7F,$9C,$A3,$F0,$2F,$F9,$89,$FF,$2E,$F8,$3F,$E6
	.dc.b	$28,$FF,$10,$4B,$C0,$FF,$35,$57,$F8,$7A,$B5,$BF,$F9,$A7,$C1,$FF
	.dc.b	$36,$61,$C9,$A8,$FF,$34,$19,$AF,$F2,$6D,$72,$7F,$CE,$1E,$2B,$C5
	.dc.b	$FF,$38,$57,$F9,$EE,$31,$3F,$E2,$EF,$07,$FC,$B7,$3F,$E6,$1F,$0F
	.dc.b	$FC,$C7,$A7,$9F,$C3,$FF,$30,$D6,$AB,$FC,$C5,$5F,$E7,$52,$5F,$F9
	.dc.b	$7A,$7F,$CD,$D3,$FC,$67,$BB,$FC,$BE,$4F,$0F,$FC,$CF,$1F,$E6,$53
	.dc.b	$FC,$CF,$CF,$FE,$5F,$6F,$85,$FE,$63,$7E,$04,$65,$FF,$31,$1D,$7C
	.dc.b	$1F,$F3,$0B,$FE,$6D,$6D,$FF,$98,$23,$5D,$1F,$E7,$2F,$03,$FC,$AF
	.dc.b	$1F,$E5,$5F,$98,$FF,$29,$69,$FE,$94,$F0,$51,$89,$FF,$30,$78,$55
	.dc.b	$FF,$32,$3F,$E6,$6B,$FF,$32,$C6,$0E,$D1,$FE,$51,$E6,$FF,$2E,$F8
	.dc.b	$BC,$7C,$1F,$F2,$DC,$7F,$AA,$7C,$2F,$F2,$A5,$FF,$97,$0F,$F2,$F4
	.dc.b	$EB,$3F,$F5,$3B,$63,$FC,$B2,$7F,$97,$3C,$3F,$F2,$F3,$FE,$62,$6B
	.dc.b	$FD,$4F,$B7,$FC,$4D,$E1,$FF,$88,$47,$87,$FE,$B2,$EF,$FF,$1E,$D7
	.dc.b	$F8,$24,$FF,$17,$D7,$09,$FF,$06,$C7,$F7,$98,$FF,$20,$B4,$FF,$1D
	.dc.b	$4F,$F6,$03,$FB,$97,$87,$FD,$EA,$3F,$C7,$93,$FD,$CA,$3F,$C4,$11
	.dc.b	$FE,$4F,$3F,$B3,$72,$FF,$75,$7F,$B6,$36,$BA,$57,$F7,$79,$FE,$D6
	.dc.b	$49,$D4,$79,$E3,$FC,$1F,$87,$B9,$FF,$37,$D7,$27,$FC,$A4,$7B,$EB
	.dc.b	$FC,$8F,$3F,$17,$0F,$89,$7E,$E8,$E5,$5F,$E4,$7C,$3D,$38,$7F,$88
	.dc.b	$B0,$F0,$CF,$39,$FF,$22,$DF,$8E,$3D,$B1,$C2,$3E,$A4,$6F,$5F,$F2
	.dc.b	$0E,$1E,$18,$F3,$CF,$1C,$3E,$5D,$7F,$86,$27,$D2,$7D,$13,$FC,$91
	.dc.b	$1E,$D7,$EC,$9D,$4F,$9C,$F2,$7E,$6D,$F8,$DF,$F0,$DC,$71,$C3,$E1
	.dc.b	$E3,$F8,$0C,$7E,$0C,$FE,$1A,$BC,$11,$E7,$7E,$26,$1E,$57,$C2,$6A
	.dc.b	$B9,$1F,$31,$FC,$3E,$3E,$D3,$EE,$61,$F2,$A3,$81,$F6,$A7,$83,$EE
	.dc.b	$9E,$0F,$C2,$BF,$2C,$7D,$79,$DC,$79,$70,$F4,$3A,$AF,$09,$F1,$63
	.dc.b	$ED,$63,$F0,$E3,$F1,$8D,$C3,$C9,$5C,$67,$C4,$71,$C3,$ED,$BE,$83
	.dc.b	$EF,$E3,$F0,$E9,$7F,$EB,$EF,$0F,$FC,$C1,$5F,$E6,$73,$E8,$DF,$B1
	.dc.b	$F9,$38,$77,$3D,$51,$ED,$8F,$C2,$9E,$13,$FC,$99,$3F,$D1,$5F,$F2
	.dc.b	$56,$1F,$E4,$DB,$FF,$25,$C6,$0F,$F2,$49,$BF,$0B,$FC,$92,$FF,$93
	.dc.b	$AB,$FC,$A0,$D3,$FC,$9C,$70,$3F,$B8,$5F,$F6,$E3,$FB,$F1,$1F,$ED
	.dc.b	$D9,$7F,$85,$23,$C9,$DA,$F0,$FF,$C5,$51,$87,$FC,$2C,$F1,$AF,$F0
	.dc.b	$01,$D8,$FE,$E5,$87,$F8,$4D,$C9,$FD,$D0,$FF,$04,$12,$FF,$C0,$6F
	.dc.b	$13,$FB,$D9,$0C,$10,$E4,$F0,$6C,$FB,$5F,$5B,$E5,$BF,$F0,$09,$2F
	.dc.b	$FC,$55,$3E,$4B,$F3,$4E,$78,$3F,$C1,$15,$A8,$C5,$79,$EF,$B3,$C4
	.dc.b	$FF,$0A,$97,$0F,$F0,$D5,$71,$25,$F4,$24,$F3,$3E,$39,$CA,$FB,$D7
	.dc.b	$B7,$0F,$2D,$F2,$AE,$07,$C5,$3E,$09,$CA,$7C,$37,$D9,$E5,$1D,$67
	.dc.b	$94,$60,$DC,$74,$3F,$C3,$15,$FE,$26,$BF,$C1,$5F,$F8,$A2,$7F,$CB
	.dc.b	$C7,$F9,$D7,$0F,$C2,$47,$F9,$37,$0E,$B1,$F1,$AB,$D6,$D8,$CD,$79
	.dc.b	$E3,$FD,$16,$7F,$89,$5F,$CD,$F8,$3F,$E3,$37,$FD,$58,$D7,$46,$3C
	.dc.b	$26,$1F,$E2,$FD,$6C,$FF,$8A,$7C,$2F,$F1,$5C,$3C,$2F,$F1,$5C,$A3
	.dc.b	$FC,$75,$5C,$4F,$F1,$9D,$70,$AD,$CF,$4A,$FF,$15,$D7,$F8,$A7,$C1
	.dc.b	$FF,$15,$C7,$EC,$BC,$1F,$F1,$5B,$FB,$D7,$FC,$6F,$E2,$BF,$5B,$7C
	.dc.b	$4F,$F8,$DA,$FC,$F8,$FA,$2B,$53,$8C,$7F,$83,$78,$3F,$E3,$39,$FD
	.dc.b	$FE,$48,$C3,$FE,$32,$F1,$F3,$D1,$FE,$31,$8F,$E1,$CD,$7F,$C6,$BE
	.dc.b	$27,$F8,$DD,$FF,$19,$78,$7D,$2F,$FC,$65,$1E,$0F,$13,$FC,$6A,$49
	.dc.b	$FC,$2E,$79,$6C,$FF,$97,$E9,$49,$CC,$BE,$45,$F1,$5F,$C6,$F0,$E5
	.dc.b	$2E,$9F,$8A,$F3,$C3,$93,$99,$DC,$EA,$3A,$57,$37,$86,$1E,$0C,$7C
	.dc.b	$6D,$C3,$9D,$F1,$C7,$C7,$79,$A8,$F7,$39,$94,$EE,$6A,$7C,$57,$DA
	.dc.b	$34,$D8,$ED,$87,$96,$7F,$C9,$07,$F9,$04,$FB,$7E,$0B,$C0,$FF,$21
	.dc.b	$C7,$F9,$23,$87,$FC,$A1,$1F,$E5,$B9,$FF,$28,$F8,$3F,$12,$BF,$35
	.dc.b	$C5,$FD,$73,$C0,$FE,$A5,$E0,$FF,$7A,$9F,$EA,$27,$F5,$FE,$FF,$F0
	.dc.b	$61,$FE,$09,$F0,$7F,$C4,$31,$FD,$54,$FE,$D7,$9F,$F6,$38,$CC,$71
	.dc.b	$EE,$FF,$00,$C7,$F6,$F2,$4E,$CF,$F1,$2D,$F8,$6B,$FC,$1E,$7F,$86
	.dc.b	$A3,$4F,$F8,$7A,$B9,$C6,$2F,$FB,$EB,$63,$D9,$5F,$E0,$63,$A4,$FF
	.dc.b	$81,$9C,$57,$F8,$44,$F7,$E1,$F0,$9C,$1E,$29,$F8,$17,$E6,$8C,$C7
	.dc.b	$37,$8D,$74,$7F,$C0,$A7,$A4,$91,$C9,$F5,$1F,$E1,$D8,$E1,$1F,$3A
	.dc.b	$FF,$C0,$DC,$BF,$E3,$6A,$EF,$1F,$E0,$57,$C8,$EB,$1F,$9F,$8F,$A6
	.dc.b	$7C,$37,$DC,$F8,$AF,$8A,$B8,$63,$EB,$8D,$C6,$1F,$29,$FE,$16,$C7
	.dc.b	$CB,$1F,$42,$39,$1F,$81,$C7,$FC,$53,$3A,$8F,$61,$F1,$08,$7A,$0E
	.dc.b	$18,$FE,$09,$E6,$7E,$1C,$E2,$6C,$C1,$E3,$39,$E1,$F4,$5D,$63,$F0
	.dc.b	$63,$E5,$47,$6C,$7F,$08,$FF,$8C,$8C,$1E,$09,$E1,$8F,$B6,$BE,$43
	.dc.b	$F8,$A9,$FF,$80,$33,$FE,$45,$C7,$FC,$7F,$7F,$EF,$7B,$5F,$F2,$8C
	.dc.b	$3F,$C6,$75,$AC,$3F,$C6,$6D,$E1,$78,$3B,$C7,$FC,$6B,$C3,$2F,$FC
	.dc.b	$64,$7F,$8D,$4F,$F1,$D1,$87,$FB,$EE,$BF,$DD,$F2,$FE,$CC,$D9,$FE
	.dc.b	$CC,$7F,$87,$08,$7F,$83,$E3,$FC,$3D,$E0,$7F,$6C,$8F,$F0,$25,$74
	.dc.b	$7F,$BE,$1F,$E0,$63,$27,$F8,$0E,$BB,$38,$D3,$FC,$06,$7F,$81,$6F
	.dc.b	$AC,$EA,$7F,$C6,$55,$D6,$B5,$1E,$4F,$07,$FB,$FB,$E8,$3C,$C6,$70
	.dc.b	$F3,$3C,$8F,$11,$E6,$AE,$98,$7C,$28,$C1,$AB,$EE,$F4,$25,$F8,$1F
	.dc.b	$54,$FA,$1E,$33,$EC,$8E,$8F,$29,$E1,$3E,$38,$D3,$E9,$8C,$BF,$24
	.dc.b	$D4,$F5,$9E,$57,$EC,$9D,$12,$FD,$6F,$38,$D4,$62,$1E,$0B,$C1,$FF
	.dc.b	$00,$40,$CB,$8C,$5E,$4B,$C3,$DD,$6B,$4F,$F0,$D3,$C7,$0F,$F1,$43
	.dc.b	$70,$FF,$0D,$3F,$E2,$A2,$1F,$DE,$7C,$87,$7F,$F6,$96,$C7,$F6,$1F
	.dc.b	$9C,$7A,$DF,$E2,$ED,$7F,$B0,$D6,$9B,$7C,$4F,$F0,$0D,$70,$CB,$18
	.dc.b	$7F,$66,$BE,$1D,$FF,$E1,$E6,$C7,$C1,$AE,$0E,$EF,$8E,$1D,$A3,$C9
	.dc.b	$8F,$8F,$0E,$55,$A6,$CE,$B1,$C3,$EF,$9F,$F1,$06,$3E,$E8,$FF,$5A
	.dc.b	$1F,$E2,$D7,$0F,$23,$3C,$BF,$E2,$B4,$D2,$78,$1F,$E2,$DE,$EF,$F1
	.dc.b	$5C,$FF,$82,$8E,$AE,$1F,$F1,$89,$FE,$BA,$7F,$C1,$18,$FF,$87,$89
	.dc.b	$1F,$83,$EF,$FF,$0B,$BF,$E0,$EF,$03,$FC,$37,$3A,$8E,$7A,$FF,$84
	.dc.b	$1F,$F0,$B4,$66,$FA,$C7,$3F,$03,$FC,$29,$18,$8E,$33,$CF,$0E,$F7
	.dc.b	$FE,$1E,$C3,$BD,$7F,$89,$23,$1D,$5F,$E3,$EA,$DD,$FF,$B7,$7B,$FF
	.dc.b	$C2,$FC,$9F,$ED,$FE,$AF,$F3,$57,$7F,$E6,$72,$FC,$B8,$AF,$F1,$13
	.dc.b	$86,$BF,$ED,$E6,$BF,$E2,$93,$1E,$07,$F9,$91,$FF,$0C,$9F,$D0,$FC
	.dc.b	$0F,$EB,$4F,$F4,$F3,$FA,$6B,$8E,$1F,$EE,$91,$FD,$A0,$FE,$97,$D9
	.dc.b	$FD,$21,$FD,$94,$7F,$41,$8C,$47,$FB,$D7,$BB,$FA,$3B,$FB,$7F,$01
	.dc.b	$7F,$EF,$D9,$CC,$A7,$2F,$EF,$63,$37,$FD,$35,$FE,$FB,$7D,$A3,$FB
	.dc.b	$9C,$FF,$66,$EF,$FF,$04,$4F,$F6,$72,$38,$24,$78,$F0,$FF,$13,$C6
	.dc.b	$23,$C5,$18,$74,$7F,$85,$0A,$7C,$33,$FB,$BD,$7C,$6A,$FE,$E7,$18
	.dc.b	$F0,$3F,$C2,$F5,$F2,$DB,$1E,$73,$9C,$F0,$C7,$FB,$AC,$7F,$81,$23
	.dc.b	$D2,$EB,$0F,$EF,$07,$F8,$66,$3E,$1B,$4F,$EF,$51,$F6,$8E,$C7,$B0
	.dc.b	$F5,$14,$F5,$4F,$81,$FC,$14,$F8,$1D,$3E,$5B,$F5,$6B,$FE,$2B,$8C
	.dc.b	$4F,$6A,$FF,$04,$61,$EE,$8F,$11,$27,$A4,$EA,$B9,$61,$EC,$AF,$C6
	.dc.b	$BE,$B7,$FC,$34,$7C,$B8,$DB,$F1,$63,$EC,$4F,$38,$FF,$1D,$E3,$F4
	.dc.b	$E7,$A4,$71,$33,$87,$97,$0F,$1E,$3F,$7E,$7D,$F5,$C3,$0F,$89,$87
	.dc.b	$39,$F0,$1F,$4C,$A7,$02,$C7,$8E,$FE,$AC,$7B,$08,$78,$DF,$F2,$BF
	.dc.b	$0F,$F8,$9A,$3F,$57,$8F,$F8,$B0,$99,$7F,$89,$E7,$53,$FE,$26,$6B
	.dc.b	$FD,$8F,$8B,$5A,$7F,$7A,$CE,$4F,$F6,$DE,$9F,$F0,$B3,$FD,$C4,$FF
	.dc.b	$04,$46,$9F,$F0,$29,$FE,$13,$8D,$D7,$F7,$38,$FE,$FA,$64,$FF,$04
	.dc.b	$CF,$F7,$F3,$93,$A3,$FC,$23,$E0,$7F,$7C,$7B,$9E,$33,$89,$CF,$3F
	.dc.b	$F0,$3D,$F8,$23,$67,$90,$E6,$E6,$3F,$C2,$71,$9F,$03,$FC,$27,$87
	.dc.b	$9E,$3D,$D3,$A7,$C9,$3A,$7A,$98,$BF,$34,$EE,$38,$E1,$ED,$7D,$11
	.dc.b	$E6,$7D,$27,$90,$B5,$D6,$39,$5F,$8C,$FE,$39,$5F,$E1,$9A,$52,$D8
	.dc.b	$6A,$7F,$B7,$DD,$77,$3A,$7F,$CD,$B7,$F4,$1F,$39,$F2,$DC,$4F,$86
	.dc.b	$3D,$E1,$FF,$3E,$47,$8E,$FC,$57,$8A,$FF,$12,$3F,$E0,$06,$97,$BF
	.dc.b	$FC,$52,$FF,$8B,$4C,$77,$FF,$8D,$B5,$78,$AF,$1B,$FC,$67,$7F,$E3
	.dc.b	$BC,$3F,$C6,$75,$A9,$FF,$1B,$F5,$BB,$BF,$C3,$F9,$7F,$83,$FB,$FF
	.dc.b	$C4,$FE,$07,$F8,$6E,$3F,$C2,$C7,$F8,$84,$9B,$BB,$3E,$77,$83,$FE
	.dc.b	$26,$8F,$F1,$21,$23,$FC,$42,$EE,$7F,$C5,$95,$D5,$F0,$1D,$27,$67
	.dc.b	$48,$D6,$3E,$7C,$3F,$C5,$35,$C2,$FA,$2C,$3C,$58,$72,$8F,$49,$FE
	.dc.b	$20,$C7,$FC,$56,$E0,$FF,$16,$1F,$87,$E8,$FF,$14,$F8,$7F,$4E,$3F
	.dc.b	$C5,$BE,$12,$31,$3F,$E3,$1C,$FF,$C5,$44,$F0,$5D,$3C,$63,$0F,$F8
	.dc.b	$8B,$0F,$F0,$C6,$5F,$95,$F0,$7F,$C2,$F1,$82,$1F,$E2,$23,$FC,$4C
	.dc.b	$71,$30,$47,$8B,$8A,$FF,$15,$1F,$E2,$B8,$CB,$8B,$FF,$14,$CF,$1F
	.dc.b	$0B,$FC,$63,$18,$9D,$61,$FE,$2B,$9C,$BE,$18,$E9,$1F,$AD,$DB,$E8
	.dc.b	$F8,$1F,$E2,$4C,$3F,$C6,$1E,$07,$F8,$8E,$31,$1F,$E8,$B6,$BF,$E2
	.dc.b	$F8,$FF,$1A,$72,$F0,$8F,$F1,$84,$61,$FF,$49,$78,$3F,$31,$FF,$1B
	.dc.b	$FD,$E3,$CA,$D4,$23,$0E,$A3,$3E,$FB,$DC,$DD,$6A,$9D,$5F,$BF,$79
	.dc.b	$5E,$8E,$35,$E3,$97,$ED,$CF,$9C,$F3,$7D,$03,$F2,$97,$A3,$86,$E5
	.dc.b	$FB,$2B,$E0,$DB,$E9,$18,$8F,$1E,$FF,$EB,$1C,$7B,$D7,$BB,$F3,$AF
	.dc.b	$F5,$B5,$71,$21,$C0,$F5,$DF,$3A,$F6,$4F,$09,$D4,$7A,$BE,$37,$E7
	.dc.b	$7D,$33,$E6,$1E,$66,$1B,$7E,$9A,$F3,$91,$C1,$FE,$C0,$AE,$5F,$09
	.dc.b	$98,$EF,$92,$B1,$8F,$F2,$DB,$DD,$7C,$5F,$9E,$F1,$FE,$0D,$E7,$AD
	.dc.b	$37,$6F,$89,$8A,$BE,$49,$2F,$91,$E3,$A7,$D0,$66,$3C,$73,$70,$EB
	.dc.b	$3F,$77,$F6,$F7,$7F,$D9,$F8,$C9,$87,$F2,$A9,$CB,$67,$81,$1F,$85
	.dc.b	$F8,$0F,$95,$D1,$0F,$A3,$3C,$9F,$B1,$AA,$78,$46,$BB,$3F,$15,$C9
	.dc.b	$23,$C5,$1E,$72,$38,$3F,$DE,$3C,$FF,$93,$FA,$3C,$C7,$33,$47,$E1
	.dc.b	$E3,$4F,$0E,$77,$CF,$7E,$64,$91,$F0,$7E,$87,$E3,$11,$C1,$F1,$7C
	.dc.b	$EF,$94,$47,$07,$5F,$9F,$FD,$27,$3C,$8F,$03,$FA,$7E,$2F,$D7,$7E
	.dc.b	$E1,$FE,$32,$E4,$FE,$E4,$1C,$74,$7E,$95,$B6,$E2,$7E,$10,$FD,$1B
	.dc.b	$F6,$5F,$45,$F9,$E8,$C5,$7E,$5D,$AF,$C3,$E0,$4F,$F0,$CF,$73,$FA
	.dc.b	$6E,$7F,$C1,$5F,$F6,$45,$7E,$BC,$FF,$56,$7B,$DB,$F3,$A4,$7F,$B3
	.dc.b	$27,$8E,$BF,$E1,$36,$C7,$FA,$DF,$E1,$F6,$F3,$8C,$BF,$EB,$7A,$F1
	.dc.b	$4F,$F0,$BF,$84,$93,$F5,$DF,$E1,$DE,$D6,$23,$D5,$17,$2F,$F0,$D7
	.dc.b	$3F,$F1,$5A,$F9,$BE,$A4,$BF,$F0,$2F,$12,$3F,$8C,$78,$19,$9F,$E0
	.dc.b	$F3,$97,$73,$A7,$8F,$D4,$FF,$1F,$FA,$79,$18,$21,$CA,$3E,$E9,$FC
	.dc.b	$A3,$F2,$EF,$E6,$19,$FD,$EC,$13,$AA,$E3,$3E,$AA,$DF,$02,$5A,$25
	.dc.b	$F5,$1D,$49,$3F,$A4,$87,$CA,$E2,$3F,$45,$1C,$1D,$4F,$96,$7F,$27
	.dc.b	$EE,$F9,$FD,$B9,$5F,$22,$32,$D2,$1D,$3C,$6F,$C5,$7C,$2F,$E2,$FE
	.dc.b	$2E,$4F,$E1,$63,$2D,$21,$F8,$DF,$73,$FE,$00,$7C,$1F,$E2,$67,$73
	.dc.b	$F8,$77,$FB,$0C,$FF,$AA,$63,$FD,$9B,$3F,$EB,$89,$FF,$58,$74,$FF
	.dc.b	$A9,$E3,$FD,$B3,$F0,$BF,$EB,$4A,$FF,$6B,$CE,$5F,$F6,$1B,$FE,$B4
	.dc.b	$F4,$7F,$D7,$55,$CE,$77,$7D,$08,$E0,$87,$FB,$81,$FF,$6D,$BF,$EC
	.dc.b	$4E,$2F,$F6,$6B,$E6,$3F,$D7,$C5,$3F,$D7,$C5,$BF,$22,$C7,$E1,$BF
	.dc.b	$EB,$E8,$F3,$4F,$49,$F8,$44,$C3,$D3,$5D,$A3,$13,$B9,$CC,$F8,$27
	.dc.b	$2F,$9C,$98,$B1,$FF,$23,$57,$FB,$22,$7F,$C7,$FF,$0D,$8C,$3F,$29
	.dc.b	$EF,$7F,$B2,$B1,$FE,$D7,$1F,$EC,$8A,$FF,$47,$9C,$0B,$3F,$EC,$E7
	.dc.b	$FD,$10,$64,$FF,$58,$7C,$37,$FA,$A3,$1F,$F5,$34,$63,$97,$FD,$6D
	.dc.b	$18,$8F,$F5,$74,$61,$FF,$4B,$1F,$EA,$C8,$FF,$4B,$BF,$E9,$6F,$77
	.dc.b	$FD,$31,$A7,$FA,$AA,$30,$47,$04,$9E,$D1,$FE,$A2,$3F,$D6,$91,$FE
	.dc.b	$B7,$24,$76,$7E,$44,$7E,$DF,$E1,$BF,$D4,$2A,$32,$FF,$92,$CC,$4F
	.dc.b	$FA,$9F,$0F,$F5,$94,$EA,$B8,$E5,$FE,$37,$6E,$3F,$EA,$9C,$7F,$D5
	.dc.b	$98,$FF,$AE,$A3,$FC,$54,$73,$2B,$E3,$C3,$FA,$6C,$71,$8F,$09,$97
	.dc.b	$6F,$33,$FA,$05,$3E,$33,$8C,$6C,$CB,$FE,$B5,$FD,$C5,$F4,$94,$DF
	.dc.b	$57,$F1,$7F,$9E,$C6,$3F,$EB,$6A,$DB,$FC,$5E,$B3,$87,$F1,$67,$8F
	.dc.b	$5B,$E5,$A9,$8F,$96,$B7,$FC,$7E,$2D,$FF,$1E,$9B,$BF,$F6,$CC,$5F
	.dc.b	$1D,$E7,$3A,$FD,$3D,$C7,$C9,$17,$0F,$E4,$F3,$6F,$94,$F9,$3D,$6C
	.dc.b	$5F,$F3,$5D,$BF,$9A,$47,$C3,$F9,$C9,$3F,$CC,$EA,$E3,$FE,$54,$C9
	.dc.b	$58,$C7,$FA,$05,$5B,$E9,$AB,$E7,$7E,$19,$F1,$3E,$82,$CF,$F3,$F8
	.dc.b	$B9,$7F,$3E,$9B,$87,$E2,$7E,$17,$EC,$CB,$93,$FD,$19,$87,$E0,$17
	.dc.b	$D5,$69,$CC,$FE,$0A,$B0,$66,$6D,$71,$8E,$1D,$38,$AF,$C4,$BF,$6B
	.dc.b	$5F,$C4,$FA,$8F,$C3,$9C,$39,$C7,$FC,$D9,$18,$7F,$DF,$7F,$B9,$3B
	.dc.b	$FF,$23,$B2,$31,$3C,$2F,$31,$8C,$FF,$AF,$FD,$0F,$E2,$8F,$C5,$FB
	.dc.b	$96,$FA,$F8,$BC,$0F,$C7,$FD,$65,$8F,$0F,$C3,$74,$31,$87,$FA,$33
	.dc.b	$14,$7E,$5B,$DE,$95,$C0,$FE,$E4,$F9,$3E,$3E,$47,$F7,$5F,$8B,$B8
	.dc.b	$73,$9B,$B7,$F8,$06,$6D,$7F,$A9,$71,$57,$E9,$3E,$AF,$38,$7F,$81
	.dc.b	$3A,$91,$F0,$BE,$96,$61,$FE,$97,$F3,$1E,$BE,$30,$F5,$55,$D3,$FD
	.dc.b	$2D,$5A,$BF,$A3,$39,$7F,$3D,$F9,$8F,$FC,$00,$72,$7F,$9C,$9C,$5F
	.dc.b	$FA,$C3,$F7,$3F,$C5,$CF,$F1,$2C,$11,$89,$FF,$41,$F0,$7F,$AC,$EB
	.dc.b	$53,$9A,$FF,$31,$B6,$3F,$D8,$CD,$AF,$F3,$15,$71,$7C,$1D,$9F,$86
	.dc.b	$9D,$74,$FD,$6F,$80,$C4,$FE,$4A,$FF,$CD,$38,$78,$3E,$2B,$81,$F9
	.dc.b	$6F,$C5,$FF,$DB,$B1,$92,$65,$ED,$9D,$7C,$87,$09,$CE,$9F,$EB,$FB
	.dc.b	$F2,$3E,$AA,$E9,$F1,$FE,$5F,$8A,$97,$FC,$C3,$F3,$F7,$97,$8C,$FE
	.dc.b	$55,$5C,$7F,$DF,$15,$AB,$F8,$F3,$97,$EA,$F8,$16,$FF,$C8,$FC,$5F
	.dc.b	$A3,$C3,$E5,$FC,$4F,$C5,$3F,$7D,$A5,$C3,$E8,$C5,$BF,$E8,$33,$69
	.dc.b	$5F,$E5,$39,$7C,$DF,$E1,$DF,$F4,$33,$67,$71,$FC,$3F,$E7,$5F,$0F
	.dc.b	$FC,$4A,$B5,$7C,$DE,$55,$D4,$EB,$D3,$C9,$C4,$69,$E1,$19,$7C,$9F
	.dc.b	$30,$F7,$9F,$39,$FE,$69,$F6,$1F,$13,$A9,$CD,$A7,$2F,$F9,$BF,$E7
	.dc.b	$1F,$03,$EA,$9C,$CB,$1E,$D8,$F5,$C7,$A3,$67,$FC,$F1,$3A,$7B,$4F
	.dc.b	$C9,$C3,$AE,$9B,$9E,$99,$79,$31,$EF,$87,$16,$C6,$70,$E5,$97,$95
	.dc.b	$F0,$DF,$63,$C3,$3B,$6E,$5D,$F2,$F0,$C6,$CF,$0C,$78,$F1,$F0,$5F
	.dc.b	$4C,$7E,$0D,$F0,$7C,$75,$BC,$3B,$61,$C8,$E9,$5B,$BE,$44,$8F,$19
	.dc.b	$AC,$7D,$4E,$30,$EA,$FE,$0C,$E8,$7E,$17,$1F,$98,$DC,$3E,$DB,$F8
	.dc.b	$9F,$7B,$FD,$7D,$F3,$9C,$63,$FC,$D9,$F1,$08,$C4,$EA,$7B,$35,$ED
	.dc.b	$F0,$6F,$81,$E5,$8A,$31,$3E,$58,$C4,$EA,$73,$3C,$A3,$E6,$47,$0C
	.dc.b	$BE,$EB,$74,$FD,$67,$CE,$7F,$3E,$BD,$DC,$7E,$B6,$1D,$9F,$B1,$3B
	.dc.b	$73,$7C,$DF,$C8,$B5,$FC,$C3,$63,$86,$1E,$2B,$EC,$DA,$EB,$E4,$FE
	.dc.b	$FE,$30,$4C,$7F,$5F,$87,$39,$D4,$62,$F9,$B6,$F9,$98,$3C,$B1,$56
	.dc.b	$1E,$49,$FE,$1D,$F3,$9F,$E0,$69,$FF,$2A,$FC,$E7,$F0,$A6,$F3,$7F
	.dc.b	$13,$E8,$FE,$0D,$5B,$94,$63,$D3,$FC,$14,$27,$4F,$F0,$F6,$BE,$08
	.dc.b	$DF,$CD,$BD,$8A,$49,$FC,$64,$65,$B1,$E3,$E8,$FE,$4D,$39,$9F,$1C
	.dc.b	$62,$75,$1F,$EF,$DF,$9C,$7A,$38,$C9,$F3,$6F,$98,$DF,$C0,$BE,$51
	.dc.b	$E5,$E1,$B9,$7F,$6F,$BE,$0D,$BF,$D1,$61,$D1,$B3,$A9,$FE,$61,$39
	.dc.b	$7F,$60,$D7,$D2,$7C,$C8,$E1,$1F,$CF,$8D,$1F,$CD,$FE,$73,$FC,$1F
	.dc.b	$3E,$3F,$17,$DF,$E8,$3C,$C7,$C5,$E1,$B5,$B2,$1E,$4C,$57,$E8,$24
	.dc.b	$79,$A3,$2F,$0C,$7C,$91,$88,$F4,$E1,$E1,$F3,$FE,$74,$BE,$4B,$CB
	.dc.b	$F3,$9E,$8B,$DB,$70,$E7,$3C,$0F,$92,$49,$F1,$35,$F8,$F1,$F6,$A3
	.dc.b	$84,$7F,$AD,$30,$53,$E6,$3F,$01,$F3,$9F,$C3,$6B,$FD,$73,$C5,$F7
	.dc.b	$FC,$DF,$BC,$6B,$C1,$93,$F5,$AB,$83,$F6,$9A,$FE,$95,$A8,$FD,$2B
	.dc.b	$C3,$3F,$B4,$DE,$8D,$42,$31,$2C,$DF,$39,$F2,$63,$FC,$7D,$3F,$B1
	.dc.b	$AF,$F0,$0E,$5F,$97,$E4,$FC,$B1,$EE,$EB,$97,$E9,$7F,$2B,$7F,$B2
	.dc.b	$8F,$2C,$60,$DB,$F3,$23,$B6,$5F,$EB,$1E,$67,$A3,$FA,$3F,$9C,$97
	.dc.b	$F2,$A5,$1F,$3B,$E0,$64,$FC,$97,$C2,$47,$E6,$7B,$FE,$42,$38,$3F
	.dc.b	$AA,$69,$FB,$3E,$7F,$D9,$12,$3C,$1C,$7F,$BB,$F9,$6F,$F4,$D4,$7E
	.dc.b	$B1,$AF,$EC,$1A,$FF,$1B,$6B,$ED,$F9,$AF,$D4,$FB,$6F,$9B,$A7,$F0
	.dc.b	$18,$FE,$17,$2F,$95,$FF,$0B,$F2,$7F,$8A,$5F,$E5,$11,$FB,$A3,$5E
	.dc.b	$37,$08,$DF,$17,$F8,$F5,$FF,$0F,$CF,$F8,$CD,$AE,$7C,$8F,$E5,$71
	.dc.b	$88,$FC,$63,$C9,$B1,$D4,$ED,$F3,$7F,$C5,$FE,$6E,$4F,$D5,$7F,$8D
	.dc.b	$7A,$F8,$8F,$E1,$FC,$3F,$C7,$E7,$19,$FE,$33,$97,$F6,$99,$FE,$5B
	.dc.b	$81,$E9,$FF,$26,$F6,$BD,$F7,$F9,$26,$D6,$E7,$EA,$13,$2F,$F4,$B5
	.dc.b	$F2,$7F,$11,$E0,$FD,$9F,$9A,$FE,$5D,$F3,$72,$77,$19,$7E,$E3,$FC
	.dc.b	$BA,$8D,$24,$70,$21,$EE,$F8,$6F,$F2,$F6,$7F,$A8,$5F,$37,$FD,$01
	.dc.b	$B3,$FC,$11,$AF,$1F,$9B,$D4,$C8,$D3,$C6,$3F,$25,$ED,$F3,$F6,$E5
	.dc.b	$7C,$08,$CB,$48,$79,$7C,$6F,$C3,$7C,$3F,$E1,$FE,$2E,$4F,$F0,$6E
	.dc.b	$1F,$C4,$4E,$A3,$FC,$61,$D2,$8F,$C3,$92,$32,$ED,$FC,$A4,$6D,$E8
	.dc.b	$7F,$17,$F7,$3F,$A3,$ED,$FC,$67,$E6,$DF,$21,$FE,$93,$9C,$D2,$72
	.dc.b	$FE,$53,$0F,$F4,$C4,$62,$3F,$4D,$18,$AE,$5E,$8F,$F8,$22,$79,$46
	.dc.b	$30,$F0,$3E,$E8,$A7,$52,$8C,$3C,$78,$77,$6C,$F1,$9E,$A7,$12,$9C
	.dc.b	$6F,$C5,$5C,$5B,$3C,$63,$D5,$8F,$C7,$AE,$37,$D9,$F9,$17,$C1,$ED
	.dc.b	$5C,$31,$F8,$11,$AC,$7D,$37,$F2,$BE,$13,$53,$FD,$1D,$A9,$F2,$4E
	.dc.b	$72,$F8,$D3,$EA,$CB,$C9,$8F,$03,$71,$E6,$CB,$C8,$FD,$23,$E4,$CF
	.dc.b	$0A,$E7,$5C,$0E,$4F,$C1,$32,$49,$F6,$B9,$7D,$87,$7B,$EB,$97,$C0
	.dc.b	$8F,$01,$92,$63,$EE,$25,$7B,$1F,$13,$D6,$38,$DE,$1E,$EF,$99,$F1
	.dc.b	$98,$8F,$33,$C1,$E5,$97,$D7,$8F,$C0,$61,$F2,$DF,$C2,$5F,$89,$F5
	.dc.b	$BD,$DE,$47,$63,$89,$EA,$AF,$C7,$47,$18,$CB,$C7,$1F,$9D,$1C,$6F
	.dc.b	$C4,$78,$E3,$31,$D8,$EB,$87,$1C,$7E,$54,$EA,$B8,$E5,$FA,$9C,$3C
	.dc.b	$13,$F9,$57,$F0,$27,$1C,$3E,$99,$F9,$09,$FC,$1D,$78,$6F,$A9,$F5
	.dc.b	$DE,$B9,$7C,$5A,$FB,$37,$ED,$9F,$CC,$65,$F3,$B1,$F4,$9E,$5C,$7A
	.dc.b	$9F,$90,$73,$8F,$58,$CE,$3D,$67,$D1,$5F,$28,$FC,$11,$D6,$BC,$B8
	.dc.b	$FA,$6A,$3F,$A1,$8C,$BF,$8D,$CB,$E0,$39,$36,$FC,$4A,$ED,$1F,$80
	.dc.b	$7B,$BF,$89,$3E,$F5,$FA,$EF,$BE,$5F,$3B,$0F,$F1,$E5,$7E,$D6,$F8
	.dc.b	$BF,$80,$7A,$BF,$21,$CE,$5F,$CA,$3D,$CF,$F0,$9B,$6F,$F7,$71,$8C
	.dc.b	$FF,$92,$3F,$E1,$4F,$7B,$FC,$94,$DC,$3F,$DC,$D5,$FE,$F7,$9C,$B6
	.dc.b	$32,$79,$84,$7D,$BF,$91,$7C,$DB,$7F,$BD,$5F,$E5,$7A,$D4,$E6,$3D
	.dc.b	$C7,$38,$CC,$F0,$8E,$A6,$85,$76,$8D,$4F,$F3,$88,$C4,$FF,$32,$3C
	.dc.b	$07,$F3,$47,$3B,$FF,$49,$8C,$E1,$FD,$2E,$78,$B9,$8F,$0B,$9C,$3D
	.dc.b	$0E,$6F,$C1,$1E,$BB,$C5,$77,$8C,$69,$F2,$67,$96,$3F,$D3,$AB,$CB
	.dc.b	$18,$7D,$07,$C0,$70,$74,$2C,$FD,$3C,$38,$1E,$02,$BD,$CF,$C5,$E3
	.dc.b	$F0,$1C,$57,$FA,$26,$30,$FD,$13,$D1,$7E,$57,$C5,$3C,$B0,$F5,$C6
	.dc.b	$EF,$BB,$67,$55,$F5,$23,$99,$EC,$C7,$E3,$C7,$56,$C6,$4D,$E3,$EB
	.dc.b	$AE,$71,$F8,$13,$EB,$B8,$BF,$88,$EA,$6C,$FB,$70,$F3,$BF,$AC,$BE
	.dc.b	$F3,$A8,$FC,$5C,$E9,$FA,$75,$CC,$F8,$EE,$B0,$F4,$3F,$93,$C7,$F0
	.dc.b	$F3,$F4,$CF,$21,$C1,$F9,$B1,$E1,$3D,$B1,$B7,$CD,$3F,$1F,$1F,$97
	.dc.b	$1B,$C3,$EF,$BF,$55,$AE,$6B,$CC,$E5,$AE,$4E,$A7,$CD,$8E,$2F,$CA
	.dc.b	$33,$3F,$3D,$B3,$E9,$C3,$E2,$47,$E0,$8E,$75,$F2,$70,$F6,$C7,$C8
	.dc.b	$33,$5C,$70,$F8,$F5,$E0,$9E,$47,$CB,$3C,$93,$E8,$79,$19,$8E,$D7
	.dc.b	$E2,$C7,$F8,$2C,$FE,$0C,$FD,$AD,$7D,$07,$E1,$1E,$22,$9C,$DF,$23
	.dc.b	$E0,$34,$7C,$73,$47,$0C,$3D,$AF,$A1,$B3,$F8,$A9,$E7,$7E,$6C,$3F
	.dc.b	$37,$B7,$F1,$DC,$3D,$CD,$7D,$B8,$7F,$16,$8D,$3F,$C6,$23,$F9,$63
	.dc.b	$5F,$D0,$98,$C7,$93,$F7,$27,$8F,$CD,$FF,$2B,$79,$CD,$BF,$E0,$33
	.dc.b	$69,$4F,$97,$7C,$D6,$63,$0E,$FC,$27,$53,$E6,$F5,$7C,$32,$FF,$3E
	.dc.b	$E1,$C1,$FF,$4E,$61,$A3,$95,$EF,$1F,$F5,$37,$CD,$E0,$CF,$C6,$DB
	.dc.b	$FF,$52,$45,$9F,$F5,$64,$DC,$3D,$D8,$BE,$6B,$C6,$76,$F9,$BB,$8F
	.dc.b	$92,$74,$FF,$57,$C3,$E0,$C6,$1F,$EB,$57,$F0,$EF,$E6,$B6,$B8,$36
	.dc.b	$7F,$D7,$DE,$B7,$66,$CF,$7F,$99,$F8,$A7,$C0,$BB,$19,$3F,$D4,$7B
	.dc.b	$BC,$77,$99,$AC,$3F,$DC,$F1,$8A,$FF,$63,$7C,$DF,$CA,$3E,$5F,$CD
	.dc.b	$D8,$C9,$2B,$FD,$A5,$19,$7E,$FC,$7F,$42,$F8,$6F,$C1,$47,$F6,$A6
	.dc.b	$F3,$BE,$52,$C4,$AF,$91,$5B,$6D,$27,$F2,$73,$93,$E7,$46,$67,$FD
	.dc.b	$83,$1F,$DE,$3A,$BA,$3E,$43,$FA,$93,$F8,$6D,$BF,$2D,$EF,$3E,$6D
	.dc.b	$C3,$9C,$3F,$D9,$73,$98,$FC,$C7,$CC,$FF,$A6,$E7,$EB,$B7,$1F,$BA
	.dc.b	$D7,$FC,$5B,$C6,$8C,$40,$71,$7E,$83,$E6,$FF,$32,$7E,$6B,$DC,$B8
	.dc.b	$7E,$82,$DE,$3F,$E2,$1B,$1E,$28,$C1,$9F,$85,$91,$E6,$6E,$3F,$E2
	.dc.b	$BC,$7F,$5F,$EE,$FE,$A0,$F6,$CD,$AF,$07,$C7,$FC,$03,$F5,$BF,$15
	.dc.b	$63,$27,$C2,$8F,$21,$1C,$1F,$92,$EF,$7C,$D7,$1C,$BE,$A6,$5F,$C0
	.dc.b	$23,$E6,$FC,$D4,$BF,$A3,$E6,$3D,$AC,$61,$F6,$F8,$BE,$65,$FC,$32
	.dc.b	$5F,$D4,$6C,$FE,$7B,$CB,$E3,$5F,$E4,$3C,$BF,$49,$1F,$C4,$FE,$6B
	.dc.b	$F2,$98,$7F,$0E,$6D,$FE,$27,$99,$C0,$F8,$BE,$1D,$B3,$0F,$DC,$E1
	.dc.b	$C3,$0F,$C0,$B5,$FC,$06,$68,$C4,$7F,$26,$7B,$C7,$F9,$5F,$1E,$FF
	.dc.b	$03,$D6,$BC,$59,$23,$D1,$F1,$33,$1F,$CB,$FC,$D3,$E3,$B1,$87,$C7
	.dc.b	$CD,$7F,$C4,$1B,$7E,$88,$C1,$1E,$7A,$F2,$D3,$C9,$8F,$C1,$9F,$8D
	.dc.b	$F1,$1F,$24,$FD,$FF,$CD,$3C,$BC,$4E,$8F,$8B,$3E,$12,$38,$3F,$8B
	.dc.b	$F8,$FF,$AB,$C3,$FA,$76,$6A,$FE,$8D,$39,$AF,$97,$2F,$88,$FB,$47
	.dc.b	$F0,$ED,$2C,$FD,$B8,$F8,$04,$7E,$BB,$FC,$33,$E6,$7F,$1D,$1F,$C4
	.dc.b	$9B,$B7,$F3,$B6,$BF,$E9,$36,$BF,$EA,$A7,$79,$7F,$65,$F9,$D7,$77
	.dc.b	$F2,$29,$CD,$FE,$0A,$72,$FF,$18,$F9,$77,$BA,$F9,$AF,$F4,$D7,$9D
	.dc.b	$FE,$B6,$9C,$C6,$33,$FE,$15,$F3,$4F,$7F,$FD,$3D,$ED,$7F,$3E,$9F
	.dc.b	$E1,$D7,$C1,$FF,$54,$E1,$CF,$1E,$D1,$83,$6E,$2B,$94,$30,$ED,$97
	.dc.b	$86,$18,$75,$AD,$B7,$0E,$D9,$6F,$0E,$37,$DC,$E3,$8F,$7A,$DC,$66
	.dc.b	$F9,$3E,$EC,$BD,$B3,$AC,$7B,$DF,$1A,$F4,$E3,$DE,$F7,$18,$3A,$4E
	.dc.b	$DF,$0C,$EF,$2F,$66,$3E,$09,$C9,$C2,$74,$D2,$1F,$ED,$58,$F9,$CD
	.dc.b	$AF,$9A,$D9,$FE,$BB,$EE,$BE,$19,$F3,$58,$6D,$7F,$82,$67,$53,$93
	.dc.b	$FB,$2E,$0F,$81,$78,$38,$6D,$FF,$84,$67,$53,$92,$47,$CD,$9D,$EB
	.dc.b	$F6,$1B,$B7,$F5,$59,$7C,$D3,$C9,$FF,$4E,$4F,$E4,$EB,$75,$F8,$E6
	.dc.b	$BF,$84,$6D,$71,$8F,$9B,$1A,$C7,$CB,$1E,$3C,$7F,$19,$5B,$8C,$13
	.dc.b	$1F,$34,$EB,$1F,$1C,$F8,$63,$A9,$1F,$C7,$57,$3A,$E3,$1C,$F1,$F5
	.dc.b	$57,$38,$E1,$3A,$24,$FC,$3A,$E4,$FF,$79,$C9,$2F,$B1,$FE,$FF,$C1
	.dc.b	$F2,$7E,$C7,$FA,$36,$3F,$E1,$46,$BF,$E1,$3B,$E3,$5C,$23,$F0,$77
	.dc.b	$CD,$E1,$E0,$65,$5F,$2E,$86,$3E,$3B,$E1,$5F,$E1,$F1,$F3,$7F,$E1
	.dc.b	$C7,$FC,$5A,$D7,$FC,$AC,$D7,$F0,$07,$0F,$4B,$FC,$66,$D3,$FC,$5B
	.dc.b	$1C,$6F,$FD,$20,$DB,$FC,$47,$CE,$7E,$4A,$BC,$5C,$CF,$A6,$F5,$78
	.dc.b	$4F,$A7,$A7,$B6,$75,$5F,$E3,$A6,$BF,$E6,$B6,$BE,$66,$BF,$97,$6B
	.dc.b	$FE,$2D,$78,$E5,$FE,$A5,$6E,$3B,$C3,$C7,$D5,$F3,$7F,$0C,$FE,$63
	.dc.b	$87,$F5,$2F,$92,$FA,$9E,$6F,$F9,$A5,$D6,$32,$7F,$CA,$F5,$C1,$FF
	.dc.b	$40,$35,$F9,$6D,$7F,$58,$EF,$2F,$F5,$C3,$72,$CD,$78,$3C,$5F,$F1
	.dc.b	$FD,$FF,$97,$2B,$F8,$03,$FD,$9B,$D0,$D1,$E9,$EC,$91,$E4,$7E,$CC
	.dc.b	$F0,$8F,$2C,$60,$DB,$F8,$78,$D6,$3F,$E2,$5B,$FF,$0F,$CF,$F9,$B3
	.dc.b	$E8,$1F,$5F,$F9,$3C,$7F,$D4,$93,$FE,$70,$6B,$F9,$19,$DB,$C6,$7F
	.dc.b	$17,$18,$8F,$F5,$33,$5F,$C8,$B5,$FE,$24,$EE,$FF,$DB,$CD,$CF,$35
	.dc.b	$E3,$F8,$EF,$D3,$7D,$7B,$E6,$BE,$CF,$E5,$89,$A6,$A2,$65,$FE,$90
	.dc.b	$AE,$0D,$8F,$D0,$C6,$23,$FD,$78,$DC,$BF,$DE,$0D,$C7,$F9,$43,$67
	.dc.b	$FC,$FB,$83,$E4,$7E,$E7,$C8,$49,$FF,$0F,$BE,$8F,$95,$91,$E3,$7C
	.dc.b	$24,$7F,$D9,$9C,$9F,$EB,$BF,$1F,$71,$FB,$FC,$55,$F8,$99,$FC,$57
	.dc.b	$81,$2B,$FC,$5A,$FE,$37,$E4,$FE,$3F,$CA,$FB,$89,$3F,$0D,$FF,$59
	.dc.b	$E3,$FE,$C9,$FA,$BF,$CA,$12,$7F,$C6,$FE,$47,$F1,$5F,$A3,$7C,$37
	.dc.b	$F1,$7F,$A4,$FF,$29,$57,$FA,$D1,$AF,$F1,$59,$DB,$FE,$B8,$6B,$FE
	.dc.b	$E9,$6B,$FC,$51,$AF,$F5,$56,$BF,$EA,$B7,$8F,$1F,$F2,$48,$C5,$79
	.dc.b	$4F,$27,$B5,$FA,$DF,$9D,$FD,$77,$EA,$72,$BF,$CA,$31,$97,$FC,$A7
	.dc.b	$19,$7F,$DC,$F1,$FB,$5F,$07,$F6,$DE,$14,$9F,$F2,$CC,$65,$FF,$2D
	.dc.b	$C6,$5F,$F7,$7C,$7E,$F7,$C3,$FD,$F7,$C0,$49,$FF,$30,$C6,$5F,$F3
	.dc.b	$1C,$65,$FF,$7B,$C6,$5F,$F7,$CC,$67,$3F,$DC,$4F,$F4,$B5,$FB,$AB
	.dc.b	$E5,$BF,$86,$77,$CC,$3F,$D2,$7D,$BF,$E9,$4A,$D4,$60,$5F,$2B,$E9
	.dc.b	$3A,$9F,$F5,$93,$67,$FD,$59,$F2,$3F,$D2,$9B,$EA,$FF,$2A,$87,$CB
	.dc.b	$D3,$11,$2F,$C1,$CE,$AD,$C6,$1A,$FF,$B2,$1F,$F1,$6D,$FD,$F8,$72
	.dc.b	$3E,$23,$F2,$5F,$21,$27,$F1,$D1,$F9,$73,$FA,$EC,$EE,$3F,$B0,$9F
	.dc.b	$8B,$8C,$91,$E2,$43,$FB,$77,$3F,$C1,$DB,$FD,$F9,$97,$A7,$6F,$81
	.dc.b	$F0,$D2,$FD,$51,$F0,$CC,$3E,$B8,$C3,$F4,$47,$EE,$BF,$DF,$EF,$AF
	.dc.b	$E6,$6D,$3F,$AA,$B6,$B8,$56,$D1,$CB,$D5,$FF,$05,$46,$63,$07,$D0
	.dc.b	$97,$C8,$FF,$83,$FE,$A3,$FD,$E1,$3F,$83,$E0,$7D,$43,$E9,$BF,$1C
	.dc.b	$D9,$F4,$57,$07,$FC,$24,$D7,$FC,$1A,$D7,$FC,$9C,$D7,$CE,$D4,$F3
	.dc.b	$AF,$BC,$DD,$38,$FC,$6F,$E3,$75,$FC,$AE,$3F,$8F,$CA,$63,$E9,$E7
	.dc.b	$F3,$3F,$98,$9D,$B9,$7F,$2D,$DF,$F9,$EF,$AA,$FC,$C7,$04,$AE,$51
	.dc.b	$96,$90,$FC,$A7,$C6,$7F,$8D,$FE,$09,$E6,$7E,$AA,$5F,$3F,$F3,$9F
	.dc.b	$AB,$F4,$0F,$53,$FC,$E1,$39,$94,$E5,$FB,$71,$9B,$EB,$C7,$8D,$FF
	.dc.b	$CF,$11,$8A,$7D,$07,$3C,$39,$46,$2B,$FC,$FD,$18,$9E,$99,$79,$DB
	.dc.b	$87,$E2,$9E,$59,$7F,$A1,$5B,$86,$70,$F0,$DF,$16,$DE,$67,$DF,$7C
	.dc.b	$67,$C2,$D8,$FC,$BE,$3E,$9C,$3C,$31,$F2,$2B,$6F,$BC,$E6,$51,$3F
	.dc.b	$07,$2E,$35,$E6,$9F,$80,$70,$29,$C3,$1F,$43,$E2,$7E,$C4,$F1,$25
	.dc.b	$FB,$27,$C8,$6F,$1F,$73,$F6,$31,$F6,$34,$CE,$3E,$DA,$E5,$8F,$94
	.dc.b	$8E,$09,$3E,$78,$CB,$DC,$F4,$CF,$BE,$7E,$2C,$79,$0F,$01,$63,$4F
	.dc.b	$19,$E6,$E9,$F8,$F3,$E0,$8C,$3D,$23,$31,$E1,$C3,$E1,$BF,$2D,$FA
	.dc.b	$E6,$9E,$27,$A6,$7A,$57,$C1,$8C,$10,$FA,$58,$7C,$6C,$7E,$8D,$7B
	.dc.b	$2F,$BD,$68,$F6,$4F,$D8,$7B,$92,$FB,$3B,$C3,$DF,$97,$E7,$F0,$F1
	.dc.b	$DF,$18,$C5,$78,$EF,$89,$F2,$1F,$97,$87,$8A,$30,$76,$9E,$33,$E1
	.dc.b	$AF,$24,$F1,$9F,$D8,$D7,$19,$F6,$D7,$CB,$BF,$71,$1C,$1F,$7F,$E1
	.dc.b	$3F,$84,$B6,$BF,$25,$18,$8F,$88,$D7,$FA,$0F,$C0,$7F,$42,$8C,$47
	.dc.b	$FB,$60,$38,$8D,$4F,$F6,$26,$CF,$FB,$11,$FD,$77,$CC,$FF,$7A,$E6
	.dc.b	$FD,$84,$F6,$3F,$21,$19,$7F,$24,$66,$BF,$8E,$46,$5F,$E3,$D1,$9C
	.dc.b	$3F,$AB,$A9,$D4,$78,$E3,$F6,$F1,$D7,$4E,$31,$EB,$7A,$F8,$7F,$E4
	.dc.b	$17,$38,$FA,$23,$FC,$89,$86,$2B,$CF,$18,$C3,$B5,$F2,$C7,$F7,$75
	.dc.b	$ED,$8C,$3F,$C2,$0C,$4F,$A3,$1F,$23,$C8,$C5,$F8,$DC,$4F,$F7,$49
	.dc.b	$DB,$F3,$4F,$6C,$ED,$C6,$1E,$03,$D0,$5A,$E7,$1F,$1E,$78,$3F,$2B
	.dc.b	$1F,$E2,$77,$DA,$30,$76,$C7,$DA,$FC,$9B,$F3,$C6,$2B,$E4,$3F,$7A
	.dc.b	$7F,$11,$87,$85,$F4,$63,$F1,$0F,$98,$5B,$F2,$E1,$E1,$9F,$8E,$7C
	.dc.b	$1C,$7C,$EF,$E0,$4C,$61,$DF,$1F,$0C,$7E,$28,$F3,$1A,$3E,$91,$83
	.dc.b	$E1,$47,$A5,$F7,$1A,$8F,$87,$3E,$08,$E1,$5E,$56,$C6,$67,$C3,$8F
	.dc.b	$D0,$C3,$E2,$D7,$8E,$79,$E1,$DF,$1F,$81,$3F,$62,$B7,$7C,$63,$D0
	.dc.b	$FC,$68,$E7,$36,$7D,$8D,$7D,$67,$A4,$F5,$B9,$C7,$D1,$87,$8A,$75
	.dc.b	$7F,$55,$FB,$EF,$D7,$D9,$FB,$B3,$E7,$BF,$D4,$D1,$23,$19,$FF,$9B
	.dc.b	$FE,$3B,$2F,$F9,$2A,$78,$D6,$30,$FF,$42,$DF,$07,$FC,$E9,$CF,$9A
	.dc.b	$DD,$FF,$9E,$63,$14,$F9,$DA,$73,$F0,$BF,$D0,$1F,$23,$8B,$FF,$28
	.dc.b	$56,$DF,$F4,$17,$B7,$9C,$BF,$D0,$4F,$8D,$7C,$93,$E1,$30,$DA,$FF
	.dc.b	$45,$75,$E2,$64,$7F,$A0,$2B,$8C,$7F,$A2,$DF,$E2,$75,$C5,$FF,$42
	.dc.b	$B5,$FE,$2E,$70,$9F,$F3,$37,$C9,$32,$EB,$3F,$EA,$A9,$F9,$53,$97
	.dc.b	$FC,$FD,$EA,$BE,$5B,$D4,$D8,$FF,$51,$FB,$EF,$84,$CD,$49,$CF,$D0
	.dc.b	$DC,$7B,$0E,$06,$09,$F9,$DE,$83,$E5,$2F,$D3,$E1,$B8,$FF,$AB,$DB
	.dc.b	$8F,$FA,$BD,$BC,$DF,$EA,$FE,$B7,$4F,$D1,$6D,$FF,$B2,$7D,$37,$E7
	.dc.b	$8F,$67,$57,$2B,$E8,$5F,$0C,$BF,$D6,$2D,$7E,$6B,$5F,$E5,$D3,$F8
	.dc.b	$DE,$3F,$55,$7F,$A9,$EB,$96,$5F,$9C,$9C,$DF,$FB,$1F,$EA,$3E,$0F
	.dc.b	$D3,$FC,$26,$E1,$FE,$CD,$6C,$EE,$3F,$DA,$13,$F9,$8E,$39,$1F,$ED
	.dc.b	$3F,$3A,$D7,$D0,$C7,$F5,$B5,$F3,$BE,$13,$31,$FE,$E4,$7E,$8F,$C0
	.dc.b	$FE,$7F,$F3,$96,$38,$3F,$DC,$2E,$1F,$F7,$04,$6D,$D6,$1F,$7B,$9D
	.dc.b	$5B,$8F,$F7,$25,$6E,$B8,$9D,$0F,$07,$C4,$78,$7E,$22,$C7,$81,$B5
	.dc.b	$D6,$D1,$8C,$BC,$F3,$F8,$D9,$CD,$7A,$3E,$33,$D9,$D7,$ED,$6C,$64
	.dc.b	$F6,$C6,$5A,$7D,$9E,$7F,$C5,$1F,$C3,$79,$DF,$02,$EA,$CD,$6C,$FC
	.dc.b	$4D,$7E,$DF,$2F,$ED,$8D,$7F,$C1,$4D,$E6,$FC,$1E,$9E,$AC,$7F,$2B
	.dc.b	$F0,$3D,$2F,$F5,$31,$8C,$3E,$9F,$67,$D7,$3E,$C7,$D3,$58,$FF,$7B
	.dc.b	$7D,$A7,$F0,$3B,$FF,$0C,$B5,$FF,$78,$B5,$39,$E4,$FB,$3F,$4F,$F8
	.dc.b	$33,$F0,$9E,$25,$BF,$C1,$F9,$5F,$C7,$6F,$F9,$76,$BF,$4B,$0F,$C0
	.dc.b	$CE,$49,$5F,$C6,$EB,$F9,$A6,$3F,$C5,$9A,$F9,$A7,$E4,$57,$E2,$BE
	.dc.b	$4F,$F2,$87,$29,$FC,$13,$E1,$FA,$39,$87,$F4,$31,$3F,$C3,$FF,$33
	.dc.b	$76,$EA,$3F,$89,$9F,$02,$BE,$A6,$46,$92,$FD,$0F,$F4,$3B,$76,$73
	.dc.b	$D3,$E0,$46,$2B,$C5,$87,$89,$7C,$77,$F4,$36,$9A,$1F,$BF,$3D,$CD
	.dc.b	$4C,$9F,$E8,$47,$0A,$FE,$7A,$7E,$DA,$3F,$19,$19,$8F,$E8,$2E,$F0
	.dc.b	$FD,$2A,$F7,$1F,$11,$D6,$BF,$9A,$CE,$63,$FA,$C7,$A6,$EE,$A7,$F0
	.dc.b	$5E,$CB,$5E,$EB,$56,$E3,$FA,$C3,$63,$CE,$D7,$FA,$AB,$5F,$F3,$6B
	.dc.b	$BC,$7F,$15,$DF,$FD,$B6,$7E,$BC,$E6,$BE,$EF,$27,$F0,$EF,$4B,$CF
	.dc.b	$E0,$49,$FE,$2F,$F5,$17,$0E,$33,$FB,$1C,$3F,$AC,$57,$9A,$30,$47
	.dc.b	$F8,$FF,$C6,$F1,$CF,$C6,$FF,$7B,$FB,$47,$C7,$D3,$FC,$21,$D0,$FB
	.dc.b	$67,$76,$71,$F8,$D7,$C1,$FE,$F8,$DD,$3E,$DB,$5F,$EF,$2D,$7F,$D2
	.dc.b	$2E,$F2,$FE,$3F,$E1,$BE,$D9,$F3,$1D,$EF,$FB,$7C,$E6,$3F,$C2,$5F
	.dc.b	$45,$FA,$2C,$7E,$C3,$6B,$7F,$2E,$FA,$FD,$F8,$32,$FF,$C2,$57,$BB
	.dc.b	$FF,$06,$B5,$FF,$52,$B5,$7C,$F3,$47,$A1,$FE,$23,$FA,$2F,$F5,$EE
	.dc.b	$1F,$A6,$AD,$4E,$7E,$13,$FC,$75,$18,$9F,$C7,$57,$0C,$3F,$C1,$73
	.dc.b	$97,$F7,$71,$9C,$3F,$A1,$C3,$E8,$BF,$35,$E6,$7F,$B4,$3E,$87,$FD
	.dc.b	$77,$18,$97,$C1,$A7,$FD,$7F,$87,$F3,$EC,$B8,$57,$F0,$CB,$E1,$7B
	.dc.b	$8C,$78,$2C,$77,$1E,$18,$C3,$DB,$1E,$BE,$8F,$F5,$8B,$E5,$3B,$BE
	.dc.b	$38,$76,$BF,$3E,$1C,$AF,$79,$79,$70,$E8,$78,$B0,$E3,$5B,$BE,$31
	.dc.b	$E4,$AE,$17,$E0,$C3,$B6,$1C,$4E,$67,$8C,$B3,$C2,$F9,$C6,$A7,$8B
	.dc.b	$BA,$F2,$E3,$ED,$AD,$C7,$7C,$7C,$95,$DE,$37,$97,$A3,$29,$3F,$C7
	.dc.b	$B1,$FA,$CF,$F9,$71,$AF,$88,$FF,$2F,$FC,$4F,$F2,$ED,$7F,$CD,$0D
	.dc.b	$8F,$E4,$ED,$77,$E3,$7E,$59,$ED,$DC,$E0,$FF,$38,$61,$FE,$5D,$6C
	.dc.b	$66,$77,$18,$3E,$BC,$67,$3F,$98,$DD,$7F,$A9,$D3,$E8,$9E,$FF,$FB
	.dc.b	$8E,$7F,$27,$5B,$AF,$C7,$35,$FC,$D3,$6B,$8C,$78,$30,$8F,$D3,$9D
	.dc.b	$5F,$58,$E7,$3A,$C3,$CA,$FE,$5A,$F9,$CF,$18,$D6,$1E,$47,$57,$CE
	.dc.b	$7C,$71,$A7,$A9,$F8,$8A,$E7,$7D,$A3,$A1,$1F,$CF,$C6,$AF,$FB,$BE
	.dc.b	$9F,$B2,$F9,$CF,$8F,$C7,$FB,$AF,$A2,$FE,$B1,$8F,$F4,$F6,$BF,$C0
	.dc.b	$B1,$57,$0E,$1E,$71,$89,$E1,$E5,$65,$FF,$02,$DF,$2C,$3F,$C9,$38
	.dc.b	$F8,$67,$FA,$E7,$D1,$3D,$1E,$10,$FA,$2F,$E4,$CF,$F6,$16,$BF,$E0
	.dc.b	$C9,$E0,$F2,$9F,$F5,$D2,$8C,$BF,$DE,$9A,$FF,$AE,$1A,$FE,$23,$9B
	.dc.b	$FD,$2C,$FF,$6F,$F5,$1F,$39,$E4,$6C,$7F,$73,$9D,$BF,$E1,$E6,$E5
	.dc.b	$FC,$92,$77,$3F,$E2,$26,$C7,$F2,$89,$DE,$DE,$1F,$81,$F8,$31,$8B
	.dc.b	$FF,$16,$B5,$FF,$6E,$75,$7C,$6F,$A2,$FF,$04,$47,$F8,$8B,$CE,$7D
	.dc.b	$17,$F5,$68,$C3,$5F,$F1,$9C,$F0,$6D,$FD,$7A,$E3,$E2,$5B,$E1,$9F
	.dc.b	$F3,$EC,$38,$36,$FC,$44,$C3,$BF,$D4,$F9,$7A,$F9,$7C,$9C,$9F,$F0
	.dc.b	$D6,$1F,$ED,$B6,$CE,$5F,$F4,$E6,$2A,$E5,$1A,$21,$EA,$9F,$37,$D3
	.dc.b	$FE,$23,$C7,$FF,$17,$CF,$F8,$9E,$BF,$D1,$0D,$F5,$35,$EA,$C8,$C1
	.dc.b	$0F,$F2,$BD,$72,$6C,$6D,$A7,$5F,$81,$FC,$BF,$D2,$3E,$3B,$E5,$7C
	.dc.b	$5A,$69,$A8,$95,$FE,$FC,$8C,$BF,$E3,$DA,$E4,$D9,$DB,$FE,$6A,$8C
	.dc.b	$BF,$EC,$1A,$FF,$5A,$BF,$7B,$E5,$3E,$C1,$1E,$9E,$F7,$98,$C3,$F3
	.dc.b	$A3,$0F,$E1,$BC,$8A,$F7,$9F,$C1,$47,$E0,$FC,$39,$3F,$8C,$7F,$0D
	.dc.b	$F4,$9F,$0B,$E1,$EC,$FE,$42,$30,$FE,$4B,$D8,$FC,$61,$1F,$CA,$F9
	.dc.b	$5F,$C0,$3E,$89,$ED,$7F,$02,$FA,$17,$CC,$D3,$51,$2F,$FC,$C9,$3A
	.dc.b	$6C,$E5,$FF,$48,$D7,$47,$FC,$F6,$1C,$3F,$E7,$E6,$BF,$D2,$8D,$57
	.dc.b	$0E,$FF,$D6,$FD,$1F,$E9,$7E,$92,$57,$EB,$A3,$F5,$06,$1F,$CD,$C6
	.dc.b	$1F,$D9,$F8,$5F,$AD,$F0,$E4,$60,$87,$E8,$9C,$BF,$A4,$8F,$DA,$78
	.dc.b	$DF,$B5,$F8,$A9,$3F,$BE,$8C,$BF,$C0,$23,$2F,$EB,$23,$2F,$EB,$A3
	.dc.b	$3E,$27,$FB,$2F,$C1,$FE,$2B,$E3,$7F,$AF,$21,$F3,$8F,$A3,$FE,$09
	.dc.b	$DF,$31,$FF,$58,$D6,$A7,$FA,$03,$63,$FD,$52,$D7,$E6,$35,$FE,$A3
	.dc.b	$5D,$5F,$F5,$7E,$FF,$EF,$FF,$5B,$F8,$A7,$84,$F9,$CA,$62,$25,$7F
	.dc.b	$21,$9D,$36,$72,$FE,$0B,$E9,$3F,$19,$F3,$F2,$3F,$D5,$9C,$C9,$FC
	.dc.b	$91,$98,$FF,$6C,$F4,$2B,$31,$C2,$7F,$A7,$D7,$18,$FC,$6B,$5F,$C6
	.dc.b	$47,$E2,$70,$F7,$47,$F9,$DF,$7F,$F5,$6C,$3E,$93,$DE,$7C,$0D,$E5
	.dc.b	$7B,$63,$2F,$BE,$3F,$B3,$FD,$27,$F6,$DB,$FF,$5B,$FD,$27,$9E,$BF
	.dc.b	$A2,$35,$FF,$79,$CE,$DF,$F7,$CB,$FD,$F9,$FD,$03,$5F,$F1,$5B,$5D
	.dc.b	$FA,$DF,$DB,$BE,$8F,$FD,$FF,$1F,$82,$FA,$47,$D0,$FE,$09,$B1,$FE
	.dc.b	$D9,$9D,$BF,$04,$07,$3C,$AC,$BF,$25,$5C,$AB,$84,$A3,$0F,$FA,$4A
	.dc.b	$B8,$D7,$F9,$09,$EF,$5F,$BB,$6B,$9F,$83,$F5,$1F,$C1,$2B,$F3,$F8
	.dc.b	$7E,$43,$E8,$E5,$7E,$7B,$9F,$BE,$3F,$A5,$7C,$91,$F9,$AA,$E5,$18
	.dc.b	$8D,$46,$3E,$87,$F3,$7A,$49,$E9,$1F,$A8,$A3,$B8,$FD,$4C,$61,$B1
	.dc.b	$FE,$1E,$F8,$27,$97,$FE,$C1,$97,$D1,$B3,$FF,$31,$FD,$27,$F6,$B9
	.dc.b	$F9,$F3,$97,$E8,$C6,$5E,$33,$FE,$6A,$BE,$BA,$F0,$8C,$61,$DF,$E3
	.dc.b	$F9,$5B,$0E,$51,$86,$D7,$0B,$FB,$F3,$C3,$0E,$B7,$E0,$3C,$7E,$A7
	.dc.b	$F9,$EB,$2E,$4D,$C3,$94,$EC,$57,$2A,$DC,$79,$CE,$65,$37,$97,$C7
	.dc.b	$8F,$13,$6B,$E3,$DF,$7A,$E2,$77,$BE,$D3,$E2,$C3,$BC,$9A,$74,$C7
	.dc.b	$DF,$3E,$8C,$76,$47,$07,$33,$C3,$5C,$0F,$0D,$73,$7B,$47,$07,$CB
	.dc.b	$3C,$DF,$D5,$DF,$99,$B8,$FB,$5D,$D7,$67,$D9,$19,$7E,$2E,$1F,$1F
	.dc.b	$2F,$C2,$61,$D6,$72,$70,$9F,$33,$F8,$1C,$BF,$D6,$39,$7C,$2C,$7D
	.dc.b	$93,$E2,$AE,$95,$F3,$63,$81,$F5,$72,$F4,$3D,$B0,$F0,$63,$E2,$C3
	.dc.b	$E0,$4F,$8B,$1E,$CF,$C6,$8E,$EF,$17,$DD,$5E,$9B,$EA,$76,$BE,$95
	.dc.b	$F1,$0F,$31,$ED,$3F,$1C,$E9,$F3,$99,$9F,$A8,$79,$B0,$F6,$57,$6C
	.dc.b	$BE,$65,$F1,$39,$46,$72,$F8,$78,$43,$A6,$3E,$EA,$E1,$87,$4C,$3E
	.dc.b	$B6,$1E,$98,$F9,$53,$D2,$72,$78,$23,$27,$1C,$3E,$74,$7F,$07,$F8
	.dc.b	$4F,$CC,$63,$F0,$1A,$FD,$E6,$E1,$FB,$67,$31,$FC,$E3,$E0,$BF,$9C
	.dc.b	$C6,$21,$18,$D3,$FC,$55,$18,$3F,$E0,$03,$88,$D4,$7F,$8B,$DF,$DB
	.dc.b	$7C,$57,$F7,$FE,$2F,$DC,$4F,$73,$F8,$04,$65,$FE,$04,$67,$6F,$E2
	.dc.b	$71,$97,$F8,$AC,$67,$6F,$E9,$EA,$75,$1E,$38,$EB,$DF,$FD,$45,$EA
	.dc.b	$7A,$6B,$AC,$65,$EB,$19,$9E,$B5,$D2,$3D,$87,$4B,$F5,$BB,$3F,$B0
	.dc.b	$C6,$DF,$2E,$1F,$0B,$0E,$D7,$E5,$C3,$C3,$87,$2A,$D3,$6F,$B1,$E8
	.dc.b	$8C,$B4,$CC,$FA,$23,$2D,$8C,$D7,$73,$D7,$7E,$28,$F5,$E3,$E2,$7F
	.dc.b	$89,$98,$9D,$E1,$FC,$12,$3E,$81,$EE,$AE,$4F,$B8,$EE,$7C,$6B,$E8
	.dc.b	$78,$CB,$5F,$56,$FC,$07,$9C,$AF,$9E,$7B,$5F,$C3,$AE,$31,$D6,$7C
	.dc.b	$AF,$C3,$8C,$BE,$36,$BE,$83,$F0,$D7,$E8,$7F,$25,$7E,$07,$E0,$1E
	.dc.b	$03,$E1,$9C,$EB,$BD,$7A,$DA,$7A,$CF,$1C,$FA,$DA,$7A,$E7,$C3,$7E
	.dc.b	$B8,$E3,$1E,$B3,$C1,$5D,$5B,$1C,$71,$FC,$64,$7A,$DA,$FE,$CE,$FD
	.dc.b	$4F,$D3,$6B,$BC,$3E,$03,$F9,$73,$C6,$7F,$77,$D5,$F4,$8F,$98,$FD
	.dc.b	$44,$FE,$BA,$73,$18,$8F,$F4,$64,$EF,$3F,$F4,$2B,$63,$FD,$0A,$DA
	.dc.b	$FD,$97,$86,$EE,$FF,$26,$DF,$FA,$42,$B4,$FF,$9A,$B9,$FF,$71,$72
	.dc.b	$5F,$27,$E1,$C5,$3E,$88,$C3,$6F,$C7,$18,$AE,$31,$FE,$8D,$F6,$3A
	.dc.b	$BE,$A9,$B6,$CB,$CD,$58,$87,$D2,$3E,$0B,$C5,$18,$6C,$67,$0F,$F6
	.dc.b	$AB,$67,$F9,$4B,$72,$F6,$FD,$03,$E9,$3B,$4E,$6F,$FD,$71,$39,$7F
	.dc.b	$D7,$47,$F2,$08,$D5,$FC,$29,$D3,$FE,$73,$3C,$93,$FE,$BC,$8C,$DF
	.dc.b	$DC,$F5,$1F,$41,$EA,$6D,$7F,$BB,$1B,$7F,$D0,$1A,$FF,$BA,$DC,$D7
	.dc.b	$DC,$F7,$1F,$3D,$F1,$E7,$37,$FE,$84,$D9,$5A,$8F,$F5,$F7,$CF,$FD
	.dc.b	$09,$FC,$27,$D2,$5E,$44,$D8,$C3,$6B,$89,$FE,$C7,$8F,$AD,$D0,$FA
	.dc.b	$3F,$AA,$D9,$FE,$B3,$AF,$F4,$79,$C9,$23,$E3,$5F,$28,$FC,$1E,$AD
	.dc.b	$FC,$75,$F4,$AF,$F9,$FD,$6A,$73,$1F,$47,$0F,$03,$FE,$9E,$8C,$DF
	.dc.b	$E8,$B2,$7D,$1F,$BF,$E5,$3E,$03,$72,$FF,$71,$37,$0F,$F5,$76,$BF
	.dc.b	$A9,$8C,$61,$FA,$5E,$1F,$CF,$12,$FF,$66,$DB,$E3,$8B,$E0,$BD,$BD
	.dc.b	$1F,$5A,$7F,$41,$39,$BF,$F7,$97,$C0,$7D,$5F,$9D,$FF,$5D,$46,$0F
	.dc.b	$85,$A2,$33,$7D,$AB,$85,$7D,$B6,$BC,$63,$D7,$3C,$33,$F8,$95,$A9
	.dc.b	$CF,$C2,$F8,$7E,$8F,$C4,$DC,$3C,$38,$23,$15,$C2,$73,$5E,$9F,$8C
	.dc.b	$F5,$7C,$55,$9E,$2D,$AF,$1B,$72,$FE,$2F,$8A,$3F,$1B,$CB,$F8,$C2
	.dc.b	$5F,$E3,$39,$3F,$03,$87,$F8,$50,$FF,$0A,$4F,$F0,$CF,$65,$F4,$8F
	.dc.b	$77,$9E,$1F,$8B,$9C,$C6,$2F,$FD,$D9,$18,$9F,$C8,$47,$47,$85,$73
	.dc.b	$7E,$1F,$67,$D7,$3E,$C7,$D2,$58,$E3,$F3,$FF,$C2,$38,$FE,$C7,$83
	.dc.b	$FC,$76,$31,$BC,$F9,$EF,$C0,$1F,$81,$9B,$7F,$43,$C6,$E0,$7E,$14
	.dc.b	$AF,$67,$F1,$1F,$27,$8D,$7E,$FB,$89,$FB,$2E,$78,$FE,$7B,$E3,$FA
	.dc.b	$49,$3F,$8E,$FA,$47,$AF,$8C,$3F,$8F,$6A,$BF,$E1,$55,$F9,$99,$CC
	.dc.b	$63,$B7,$86,$5F,$D4,$AB,$4F,$F5,$EB,$E5,$8F,$7A,$FE,$80,$BE,$3A
	.dc.b	$9F,$D1,$97,$D2,$3E,$93,$31,$F5,$8E,$07,$F0,$71,$ED,$3E,$17,$7A
	.dc.b	$FF,$5A,$92,$77,$1A,$CF,$FC,$B5,$18,$3B,$63,$E3,$BE,$91,$F8,$6F
	.dc.b	$45,$DD,$FD,$65,$B1,$FA,$8F,$0B,$13,$23,$F5,$7E,$7D,$CF,$F6,$AD
	.dc.b	$CB,$F4,$6D,$9F,$E7,$9C,$FF,$DB,$2B,$FA,$86,$3F,$A7,$9F,$17,$81
	.dc.b	$FC,$27,$E9,$3C,$4C,$8F,$DC,$FC,$6D,$BF,$E0,$6D,$CB,$F6,$2D,$9E
	.dc.b	$EF,$F7,$3F,$A3,$7C,$85,$3F,$BC,$F2,$BE,$91,$DB,$9C,$3F,$53,$7C
	.dc.b	$1F,$E7,$5E,$2B,$E9,$1F,$35,$BD,$7E,$A1,$FE,$8D,$8C,$D7,$2C,$FF
	.dc.b	$C0,$F7,$C6,$F8,$6B,$FE,$8F,$8C,$57,$EE,$3E,$8F,$F3,$2F,$F5,$26
	.dc.b	$E3,$FE,$AC,$6B,$FC,$25,$AF,$FA,$79,$DD,$7F,$89,$3E,$95,$F4,$DF
	.dc.b	$E1,$59,$CD,$FE,$2A,$95,$A8,$FE,$B1,$F4,$1F,$E2,$8F,$4B,$FC,$39
	.dc.b	$F4,$AF,$81,$FF,$10,$CE,$6B,$F8,$7C,$E5,$E3,$82,$F8,$DF,$08,$C4
	.dc.b	$F0,$8F,$F1,$47,$D2,$FE,$4B,$CD,$FF,$67,$7D,$27,$FA,$E2,$31,$B7
	.dc.b	$FB,$3A,$3F,$D7,$79,$3E,$5B,$FD,$71,$F0,$89,$E1,$DE,$C7,$9E,$1F
	.dc.b	$A8,$C7,$9D,$EE,$31,$5B,$7B,$1F,$AE,$C7,$B6,$7C,$A1,$7C,$B1,$EF
	.dc.b	$8E,$63,$13,$C7,$2F,$2D,$6A,$3A,$9C,$B1,$F0,$65,$E3,$8C,$61,$DF
	.dc.b	$2F,$0C,$70,$C7,$AB,$A3,$17,$CB,$0F,$8B,$8F,$82,$7D,$B8,$79,$6B
	.dc.b	$51,$83,$C7,$87,$6C,$7C,$58,$F6,$25,$70,$A5,$6F,$1F,$65,$F0,$AF
	.dc.b	$95,$5F,$D7,$6F,$EC,$B6,$32,$FC,$33,$FB,$17,$C2,$FF,$29,$DF,$FC
	.dc.b	$62,$D8,$FE,$CC,$1C,$3C,$3C,$34,$FF,$8E,$2F,$FB,$96,$D7,$B7,$F3
	.dc.b	$17,$CA,$72,$D8,$CB,$E7,$77,$97,$C2,$6E,$7F,$93,$6C,$FF,$32,$FA
	.dc.b	$57,$BF,$FE,$FA,$AD,$63,$F8,$3B,$DD,$FE,$21,$B7,$C2,$35,$31,$FB
	.dc.b	$D3,$AB,$E9,$1F,$99,$8E,$78,$FA,$B1,$F3,$5F,$E5,$70,$E9,$7C,$A3
	.dc.b	$A4,$EA,$3F,$23,$3A,$8C,$13,$1F,$4C,$EB,$1F,$0B,$E9,$BE,$73,$AA
	.dc.b	$FC,$C5,$F3,$8F,$E7,$FF,$57,$FE,$C3,$E2,$FE,$AF,$83,$F6,$E7,$C4
	.dc.b	$E6,$3F,$AD,$C6,$1F,$F1,$BD,$67,$D3,$FF,$24,$3B,$8E,$55,$6B,$84
	.dc.b	$7F,$91,$A7,$5F,$6F,$FC,$46,$3F,$C4,$6D,$7F,$D4,$0D,$E7,$FF,$6C
	.dc.b	$FC,$F3,$E5,$38,$F5,$4B,$E2,$F8,$67,$51,$FE,$2E,$FA,$67,$AB,$FE
	.dc.b	$88,$84,$61,$FF,$45,$B5,$FC,$DB,$5F,$F2,$13,$9C,$3F,$DD,$6D,$DB
	.dc.b	$E4,$B6,$7F,$C0,$3F,$4F,$F9,$8A,$DD,$FC,$5A,$D5,$78,$A3,$E4,$56
	.dc.b	$EF,$E9,$FD,$33,$DF,$FA,$5E,$53,$EE,$FF,$D6,$B8,$FF,$A6,$63,$1E
	.dc.b	$EF,$D1,$BE,$13,$FE,$4E,$6B,$FE,$CB,$6E,$9F,$E1,$58,$FF,$0B,$C6
	.dc.b	$31,$FF,$0C,$DF,$7F,$77,$A6,$52,$FA,$3E,$4C,$93,$E1,$F0,$3F,$09
	.dc.b	$F3,$DF,$DC,$AF,$FD,$1A,$D7,$C5,$1B,$7C,$CD,$7F,$D4,$B0,$FA,$8C
	.dc.b	$69,$E7,$96,$9F,$E2,$47,$89,$DA,$7F,$24,$3D,$0F,$CA,$7E,$30,$FA
	.dc.b	$6A,$6A,$25,$7F,$9D,$2B,$83,$63,$8F,$7E,$A3,$FD,$E9,$3A,$6C,$E9
	.dc.b	$FF,$3E,$FD,$47,$F9,$15,$FF,$3F,$FC,$B3,$D9,$FD,$33,$63,$7E,$DF
	.dc.b	$EA,$6B,$84,$7F,$C0,$16,$EB,$F0,$5B,$C9,$FB,$8B,$FF,$54,$57,$22
	.dc.b	$1C,$6F,$BF,$C3,$FE,$DF,$DA,$FF,$37,$4F,$FA,$C3,$E0,$70,$7E,$E7
	.dc.b	$EA,$3F,$CE,$B1,$FE,$96,$6B,$FE,$2A,$0E,$33,$FA,$8D,$CB,$FC,$97
	.dc.b	$F6,$FD,$CE,$9F,$09,$2F,$FD,$7F,$1E,$3F,$56,$47,$FB,$09,$FC,$4F
	.dc.b	$8D,$E3,$F8,$39,$3E,$37,$F1,$BE,$84,$8F,$C6,$B5,$FF,$69,$FD,$0F
	.dc.b	$F1,$3F,$A2,$FD,$D4,$7F,$1E,$6B,$FE,$76,$6A,$69,$FB,$62,$78,$7F
	.dc.b	$27,$E3,$BA,$32,$7F,$D7,$73,$FC,$07,$CC,$FF,$75,$57,$F2,$5F,$CF
	.dc.b	$3F,$79,$49,$43,$E0,$F4,$6F,$B2,$57,$FB,$D2,$74,$D9,$CB,$FE,$5C
	.dc.b	$AE,$7F,$5B,$DF,$D0,$93,$C6,$3C,$46,$1E,$17,$FE,$1B,$E5,$FE,$05
	.dc.b	$7C,$A7,$C1,$EF,$7A,$7B,$E4,$60,$87,$27,$2F,$48,$F8,$06,$2B,$AD
	.dc.b	$FE,$EE,$78,$C7,$F1,$11,$FB,$D7,$FA,$D6,$3F,$CC,$1C,$FF,$E6,$18
	.dc.b	$C4,$7F,$9E,$A7,$7C,$5F,$37,$B7,$FA,$B4,$62,$78,$87,$0F,$1A,$E7
	.dc.b	$1E,$49,$FA,$DF,$5B,$F5,$F9,$A5,$FD,$18,$CB,$F4,$E3,$2F,$CA,$8F
	.dc.b	$C7,$78,$3F,$D3,$79,$3F,$82,$4F,$F6,$C3,$F1,$91,$92,$3F,$88,$78
	.dc.b	$BE,$81,$F5,$1F,$ED,$37,$44,$3F,$65,$3F,$C4,$71,$FE,$49,$3B,$9F
	.dc.b	$E5,$3E,$F7,$F3,$2F,$AA,$91,$FC,$CA,$3F,$9A,$98,$7F,$97,$C6,$1F
	.dc.b	$93,$4F,$A3,$FE,$D7,$1F,$45,$B1,$F6,$5B,$BF,$FB,$83,$8F,$F8,$E5
	.dc.b	$72,$9C,$BF,$4D,$B5,$D7,$D4,$FF,$05,$78,$DF,$1E,$3F,$9F,$68,$EC
	.dc.b	$F9,$84,$C3,$FC,$9E,$7E,$DD,$C3,$F1,$1F,$E0,$07,$F8,$4F,$5F,$0C
	.dc.b	$6C,$FD,$BC,$69,$FA,$11,$AF,$A6,$FF,$7E,$C7,$D3,$6B,$FC,$85,$B7
	.dc.b	$FA,$8F,$A7,$7D,$17,$F8,$86,$73,$B7,$F8,$B1,$B6,$0E,$36,$5F,$29
	.dc.b	$F9,$1F,$43,$F9,$CD,$CD,$3E,$3F,$C8,$59,$D9,$0F,$E0,$9B,$FE,$0F
	.dc.b	$4F,$ED,$1F,$07,$FE,$3C,$7C,$5F,$47,$DE,$9F,$48,$E9,$FF,$21,$CE
	.dc.b	$63,$13,$C3,$83,$FC,$8F,$87,$38,$57,$38,$C3,$C6,$F9,$D7,$F9,$3A
	.dc.b	$F9,$DF,$08,$E7,$D3,$FB,$78,$C1,$F1,$F0,$E7,$3C,$6F,$9C,$FC,$BB
	.dc.b	$E7,$87,$87,$0E,$4D,$C7,$D1,$5B,$BE,$2F,$C0,$6E,$3D,$A3,$C1,$7C
	.dc.b	$B1,$F0,$47,$23,$81,$67,$83,$FE,$A1,$9E,$0F,$73,$86,$3E,$6A,$F7
	.dc.b	$E1,$C6,$B7,$5D,$0F,$82,$5C,$7D,$4E,$30,$EF,$87,$38,$EC,$66,$B9
	.dc.b	$CE,$B1,$F7,$D6,$EB,$CD,$87,$84,$F2,$CE,$8F,$87,$7C,$8F,$2D,$71
	.dc.b	$39,$5F,$C0,$3B,$5F,$23,$A1,$E3,$BF,$D0,$78,$3F,$83,$8B,$97,$CC
	.dc.b	$7F,$B1,$3F,$11,$A7,$D2,$9E,$8F,$C4,$BF,$0E,$1E,$59,$CE,$5D,$6B
	.dc.b	$C0,$4C,$30,$4C,$3B,$3F,$32,$78,$12,$3C,$4E,$5F,$33,$99,$F2,$1D
	.dc.b	$9C,$3F,$5E,$38,$91,$FA,$B1,$AC,$7E,$46,$1F,$33,$2F,$A1,$5F,$22
	.dc.b	$BC,$0F,$0B,$F0,$9C,$DF,$B2,$7C,$39,$E2,$F8,$E3,$59,$7D,$79,$FB
	.dc.b	$79,$7D,$4F,$0B,$FC,$17,$7C,$9B,$97,$EE,$DB,$97,$EC,$DD,$CF,$F1
	.dc.b	$9F,$81,$FE,$35,$18,$84,$63,$4F,$EC,$11,$81,$3B,$9D,$47,$F6,$57
	.dc.b	$F5,$9F,$14,$FA,$4F,$F4,$A6,$DF,$C5,$6B,$C0,$7E,$92,$75,$CD,$FB
	.dc.b	$38,$CC,$FF,$1B,$9C,$C7,$EA,$8C,$EF,$FE,$DE,$8C,$CF,$59,$F3,$3C
	.dc.b	$6F,$C3,$5E,$AC,$3C,$B1,$8A,$F4,$1E,$37,$0F,$FB,$AA,$37,$87,$96
	.dc.b	$3A,$D7,$86,$B7,$1E,$43,$FA,$26,$1E,$49,$F0,$47,$5A,$E5,$7D,$23
	.dc.b	$D8,$74,$38,$DF,$47,$02,$FA,$3E,$8C,$3D,$D7,$D2,$3B,$1D,$2F,$F6
	.dc.b	$AF,$48,$AF,$B4,$FD,$A9,$D6,$FF,$D9,$59,$7C,$BA,$F3,$5D,$AE,$AD
	.dc.b	$3C,$2F,$73,$38,$FC,$0A,$C9,$C4,$AF,$C4,$9E,$27,$37,$E6,$61,$EB
	.dc.b	$8C,$C6,$CC,$D7,$78,$CB,$F0,$E3,$31,$CE,$3E,$01,$B8,$C9,$CD,$CC
	.dc.b	$6D,$F8,$53,$E1,$9D,$C7,$53,$6F,$17,$67,$53,$26,$CC,$CF,$5C,$7E
	.dc.b	$CC,$7D,$18,$F4,$B6,$7B,$B4,$E6,$53,$9B,$E1,$9E,$71,$F9,$13,$27
	.dc.b	$85,$CB,$F8,$A7,$F1,$3A,$BE,$71,$F2,$7F,$E7,$9C,$7F,$C3,$93,$98
	.dc.b	$C5,$FF,$A7,$F8,$9F,$47,$8A,$D5,$FF,$93,$B0,$E6,$DC,$38,$F4,$7F
	.dc.b	$B9,$EF,$85,$6B,$1F,$F6,$97,$C3,$BC,$AE,$F0,$8C,$63,$FE,$82,$C1
	.dc.b	$F4,$4F,$8A,$DD,$6A,$F9,$E1,$83,$C3,$5B,$C3,$FC,$E9,$5B,$D1,$1F
	.dc.b	$C2,$FE,$87,$C3,$B9,$F8,$3F,$1C,$F9,$C7,$8B,$A9,$97,$FE,$7F,$BE
	.dc.b	$0F,$FA,$62,$F8,$56,$B4,$FF,$78,$56,$9F,$E5,$AD,$7F,$D1,$CD,$7C
	.dc.b	$95,$FC,$CB,$1F,$B1,$A7,$CA,$3E,$6F,$13,$D4,$C5,$7D,$39,$B7,$FE
	.dc.b	$4A,$9B,$7F,$EC,$6F,$5D,$EC,$7D,$65,$F0,$56,$30,$DC,$7E,$94,$E9
	.dc.b	$B3,$A7,$E1,$7C,$F7,$CA,$3E,$4F,$96,$D3,$13,$A2,$67,$FE,$BD,$7F
	.dc.b	$CD,$9B,$AB,$84,$6B,$3F,$AF,$5F,$CB,$B0,$AF,$C8,$6B,$FD,$15,$AE
	.dc.b	$E7,$FA,$DF,$2F,$FB,$E3,$CF,$FC,$AF,$A0,$F7,$6C,$61,$B5,$FE,$F9
	.dc.b	$6D,$FF,$BC,$1F,$D1,$7B,$9F,$E8,$89,$FF,$57,$B6,$7F,$DE,$BF,$5B
	.dc.b	$AE,$E9,$5A,$7C,$31,$89,$FF,$7D,$FC,$A7,$93,$AB,$CA,$DB,$FD,$16
	.dc.b	$68,$C5,$70,$9C,$D7,$06,$9C,$27,$7B,$7F,$BE,$BE,$B7,$E0,$FD,$6F
	.dc.b	$C2,$6D,$76,$6E,$1F,$54,$6E,$C9,$3A,$C3,$EA,$CE,$78,$7E,$6F,$57
	.dc.b	$CE,$6E,$3E,$A6,$E3,$D6,$B8,$4F,$BA,$BF,$67,$D1,$23,$E0,$FD,$6D
	.dc.b	$BF,$D5,$CD,$C7,$C3,$9A,$B1,$7F,$E0,$7A,$FC,$1D,$FF,$6B,$6F,$5F
	.dc.b	$EA,$67,$FD,$A6,$D9,$FF,$80,$19,$3A,$5F,$41,$FA,$73,$F1,$3F,$45
	.dc.b	$70,$FF,$80,$1F,$17,$A6,$CE,$70,$F9,$42,$74,$FC,$1D,$1C,$DF,$94
	.dc.b	$3F,$69,$EF,$DC,$3F,$91,$7E,$7D,$FE,$D1,$7B,$C7,$F2,$E8,$C5,$F6
	.dc.b	$C1,$3F,$CB,$1B,$5C,$0E,$73,$99,$E5,$D9,$27,$53,$F5,$6D,$3F,$CD
	.dc.b	$A5,$D7,$4F,$E8,$EB,$E8,$5E,$96,$63,$FA,$69,$FE,$E9,$C3,$F9,$8D
	.dc.b	$65,$FC,$74,$71,$9C,$72,$7E,$FA,$3B,$9C,$A7,$F9,$CF,$D0,$FC,$88
	.dc.b	$C4,$74,$74,$23,$51,$F4,$BF,$03,$78,$1F,$D4,$1B,$5C,$32,$E0,$F4
	.dc.b	$8C,$D7,$1A,$E0,$7C,$0A,$DC,$66,$B8,$4E,$7E,$4F,$71,$D5,$FE,$BF
	.dc.b	$E7,$7F,$55,$AF,$C2,$B7,$2F,$E6,$DF,$45,$AF,$2E,$61,$EE,$AD,$2B
	.dc.b	$FD,$AF,$3A,$F7,$1D,$2F,$F9,$F7,$BF,$8C,$7F,$A1,$3E,$2F,$92,$98
	.dc.b	$FE,$3B,$D3,$B8,$79,$26,$DF,$EF,$B9,$3F,$CE,$AF,$F3,$2D,$13,$AC
	.dc.b	$3C,$9F,$82,$3F,$67,$7C,$9E,$72,$4E,$63,$FC,$03,$F4,$1F,$B1,$CB
	.dc.b	$F9,$EB,$70,$FF,$36,$FC,$83,$E2,$B7,$E1,$C9,$FC,$0D,$6F,$2F,$F3
	.dc.b	$45,$EE,$73,$5B,$9F,$F4,$ED,$7E,$FD,$FC,$A0,$70,$FE,$5D,$AF,$CF
	.dc.b	$FB,$FF,$EA,$AF,$EF,$1B,$8F,$F5,$6F,$11,$8B,$E2,$FF,$C3,$B5,$AC
	.dc.b	$BF,$44,$DC,$3A,$7A,$1F,$E3,$0C,$38,$DF,$08,$FD,$F7,$C4,$7F,$89
	.dc.b	$BE,$71,$F1,$58,$CB,$F4,$31,$87,$F8,$3C,$EE,$75,$8F,$F1,$2A,$D3
	.dc.b	$FC,$53,$E7,$9F,$2F,$FB,$E8,$FF,$61,$FC,$7F,$ED,$E7,$3E,$2F,$FB
	.dc.b	$26,$3F,$75,$F4,$1F,$DD,$F3,$C5,$FE,$4E,$F6,$F2,$31,$87,$5F,$82
	.dc.b	$61,$D6,$30,$F4,$F4,$7F,$BF,$CE,$AB,$8E,$5D,$96,$1D,$A7,$51,$8C
	.dc.b	$39,$BF,$A5,$CB,$CC,$F8,$CD,$5F,$3B,$E5,$8F,$8E,$3B,$57,$0C,$3A
	.dc.b	$D6,$AF,$C6,$73,$70,$A7,$85,$F3,$8D,$61,$E4,$7C,$27,$8E,$76,$F4
	.dc.b	$BE,$F3,$BB,$EB,$5A,$AE,$35,$AC,$3F,$DE,$F1,$D0,$FC,$DD,$7D,$26
	.dc.b	$DF,$DD,$7F,$3F,$F0,$BF,$CE,$BE,$83,$8B,$63,$FE,$00,$B5,$DF,$94
	.dc.b	$9E,$F7,$FE,$73,$F8,$5B,$E1,$9A,$4B,$FD,$69,$F6,$23,$4F,$C9,$6E
	.dc.b	$7F,$39,$BA,$FF,$47,$FA,$0F,$F4,$25,$7D,$79,$CB,$F8,$F6,$BD,$6F
	.dc.b	$F0,$D3,$C2,$3B,$47,$12,$3F,$83,$8D,$5F,$86,$72,$D9,$D7,$B3,$FD
	.dc.b	$66,$30,$4C,$3F,$4D,$87,$8B,$0E,$71,$8C,$39,$B7,$0E,$66,$0F,$54
	.dc.b	$55,$3C,$9F,$E7,$BF,$41,$FD,$A7,$AF,$FC,$C3,$C8,$F8,$8C,$72,$FF
	.dc.b	$6D,$8C,$5F,$F9,$E2,$F2,$F2,$3F,$83,$D6,$14,$F0,$7F,$C6,$9E,$4A
	.dc.b	$B5,$F4,$0F,$A0,$FC,$7D,$23,$0F,$E4,$9A,$FE,$01,$AF,$F2,$E7,$39
	.dc.b	$7F,$9E,$9B,$97,$F8,$15,$B3,$FD,$77,$CC,$75,$F9,$FE,$83,$FC,$3B
	.dc.b	$3F,$E0,$0F,$A0,$FE,$88,$DF,$17,$5E,$A4,$8F,$D2,$D7,$07,$F2,$6D
	.dc.b	$7F,$A0,$B5,$FF,$0B,$3B,$CF,$FD,$1E,$DC,$F3,$5F,$27,$0F,$CD,$DF
	.dc.b	$F8,$FB,$E7,$BF,$D5,$6F,$F8,$5F,$D6,$7D,$07,$F6,$68,$C3,$5E,$1F
	.dc.b	$0D,$F4,$EB,$8C,$7E,$A1,$AF,$F5,$66,$BF,$E2,$B6,$BF,$E0,$37,$8E
	.dc.b	$7F,$EA,$66,$E5,$BA,$FC,$1E,$1E,$4F,$2F,$C1,$E0,$7E,$1B,$DD,$7D
	.dc.b	$03,$E6,$B0,$DD,$3F,$C5,$F9,$FF,$48,$6D,$FE,$12,$31,$1E,$98,$C1
	.dc.b	$1D,$90,$FF,$41,$4F,$F6,$77,$FC,$77,$E4,$BD,$5F,$CB,$7D,$07,$FA
	.dc.b	$DE,$7F,$C7,$ED,$7F,$CD,$B3,$B7,$8C,$FE,$5A,$31,$1F,$C4,$9A,$FF
	.dc.b	$80,$1A,$FF,$95,$5D,$EF,$FE,$CE,$6E,$F9,$AF,$1F,$77,$EA,$7C,$EF
	.dc.b	$F3,$CE,$1F,$E5,$2F,$92,$FF,$70,$37,$CD,$D7,$BD,$23,$F9,$BD,$70
	.dc.b	$7F,$92,$B5,$FF,$09,$B5,$FF,$34,$3B,$CF,$FD,$C2,$DC,$B3,$5F,$BB
	.dc.b	$C3,$FD,$FB,$7F,$EF,$8F,$A0,$FB,$9E,$01,$A7,$F0,$1F,$A0,$7A,$94
	.dc.b	$FD,$EB,$8A,$FE,$9A,$D7,$F9,$DB,$5F,$F3,$E7,$2F,$FA,$9F,$E8,$39
	.dc.b	$47,$F9,$D2,$78,$6B,$FE,$C6,$C5,$5E,$4F,$A0,$97,$FE,$7D,$C3,$FD
	.dc.b	$61,$1F,$90,$F1,$BC,$DF,$41,$E4,$25,$79,$1F,$EC,$35,$F9,$8F,$3F
	.dc.b	$F2,$E4,$9E,$98,$BE,$7B,$F9,$07,$CF,$FE,$EE,$7F,$D3,$BF,$18,$E3
	.dc.b	$FE,$4F,$EA,$23,$13,$FE,$8E,$AE,$09,$FE,$10,$FF,$9C,$DA,$E7,$C4
	.dc.b	$FE,$4F,$18,$38,$91,$FA,$BF,$25,$FB,$1F,$88,$FE,$17,$C5,$FE,$DC
	.dc.b	$F8,$0C,$19,$EB,$98,$FC,$7A,$4E,$A3,$13,$CF,$C9,$E9,$E8,$C9,$E1
	.dc.b	$19,$79,$46,$5F,$F6,$A4,$78,$FC,$4F,$27,$81,$27,$C1,$19,$7C,$51
	.dc.b	$97,$FD,$B9,$19,$7F,$DB,$D1,$9C,$BF,$DE,$F3,$FD,$51,$7D,$07,$FB
	.dc.b	$32,$BF,$DB,$FF,$63,$FC,$4F,$E0,$CE,$E7,$52,$AF,$AA,$72,$E6,$D1
	.dc.b	$23,$E3,$77,$FC,$DF,$77,$E7,$75,$C9,$F9,$51,$97,$E6,$46,$5F,$F7
	.dc.b	$B4,$7E,$23,$C4,$FE,$D1,$F4,$2F,$46,$99,$AF,$C3,$4F,$E0,$08,$FF
	.dc.b	$8D,$B9,$FF,$87,$BF,$87,$8E,$0F,$AF,$AB,$FD,$CD,$3F,$C4,$23,$FD
	.dc.b	$7D,$7F,$EE,$C7,$F4,$DE,$43,$DE,$FE,$6B,$F1,$F2,$73,$1F,$CE,$0C
	.dc.b	$3F,$CB,$A3,$0F,$FB,$AF,$77,$CF,$FF,$6E,$8E,$BE,$47,$F4,$03,$EB
	.dc.b	$46,$63,$CC,$7D,$17,$13,$F4,$9B,$9F,$FB,$56,$3A,$FA,$9F,$D3,$7C
	.dc.b	$BF,$F5,$7C,$7E,$B2,$FF,$55,$E1,$7C,$CF,$B2,$91,$FA,$F8,$FD,$B9
	.dc.b	$FA,$A2,$3F,$2D,$AF,$90,$E6,$71,$3C,$58,$7F,$B9,$D6,$DE,$67,$FD
	.dc.b	$D5,$F4,$4F,$8E,$FE,$C5,$F4,$4C,$69,$FE,$27,$9D,$57,$43,$8D,$64
	.dc.b	$8F,$F1,$BA,$7C,$BF,$CA,$FA,$2F,$84,$C9,$FA,$5B,$A7,$51,$86,$BF
	.dc.b	$43,$CA,$6F,$F8,$38,$F8,$1F,$0B,$FE,$43,$E0,$47,$F0,$4C,$7E,$07
	.dc.b	$D1,$3B,$3E,$17,$8D,$E4,$AD,$CE,$5E,$31,$95,$87,$57,$A1,$D9,$C7
	.dc.b	$D1,$6E,$B9,$4A,$31,$7F,$45,$E9,$5C,$30,$FA,$8D,$AF,$1D,$6E,$7D
	.dc.b	$2D,$8F,$07,$77,$F0,$17,$B4,$F2,$8C,$5F,$82,$7E,$1B,$5D,$9C,$B2
	.dc.b	$CD,$71,$7E,$0B,$E0,$8A,$79,$0A,$7C,$4A,$F1,$65,$EC,$7C,$33,$C6
	.dc.b	$7B,$8C,$BE,$2E,$1B,$C3,$C3,$3E,$33,$E6,$9B,$C7,$EB,$DF,$3C,$78
	.dc.b	$4E,$B2,$F9,$0E,$CE,$65,$3C,$2E,$8D,$E3,$E2,$AF,$36,$3D,$08,$E0
	.dc.b	$95,$E5,$8F,$0D,$FC,$B2,$4F,$C9,$9F,$26,$5F,$09,$F9,$07,$C2,$BF
	.dc.b	$03,$CE,$78,$5F,$D5,$23,$83,$A5,$F6,$9E,$6F,$8A,$35,$1D,$A7,$8B
	.dc.b	$D4,$87,$C3,$72,$FD,$9C,$BE,$3E,$3F,$9D,$BE,$D3,$F2,$5E,$CF,$2C
	.dc.b	$3D,$2F,$3C,$3C,$2F,$33,$DD,$97,$E5,$DD,$1E,$2B,$FA,$31,$9A,$E7
	.dc.b	$1F,$57,$2F,$C5,$DF,$DF,$25,$FE,$17,$0F,$AA,$FB,$E3,$EA,$BE,$1C
	.dc.b	$BC,$EF,$A6,$38,$BE,$59,$CE,$1F,$B6,$7D,$F8,$7C,$1A,$FB,$5E,$3F
	.dc.b	$F8,$32,$FF,$28,$D7,$DC,$D9,$FE,$08,$E6,$3F,$88,$FC,$27,$F1,$28
	.dc.b	$C4,$23,$17,$FC,$CE,$36,$27,$73,$A8,$D3,$FB,$0F,$89,$7C,$7F,$F8
	.dc.b	$D2,$FF,$8B,$57,$27,$F8,$C4,$EB,$A3,$F6,$F3,$9C,$BF,$A1,$4E,$63
	.dc.b	$F5,$E6,$4F,$E5,$2E,$6B,$AC,$7A,$4F,$DD,$C7,$1D,$33,$8F,$EC,$23
	.dc.b	$17,$ED,$AE,$17,$E4,$C7,$D1,$39,$BF,$64,$62,$BD,$E7,$67,$FC,$A3
	.dc.b	$87,$68,$C6,$3F,$BE,$7E,$4C,$75,$8C,$47,$9E,$30,$FF,$12,$37,$7F
	.dc.b	$0A,$BA,$BE,$38,$EA,$6C,$F4,$BB,$AF,$97,$18,$3E,$3C,$FA,$5F,$ED
	.dc.b	$D7,$F0,$E3,$AC,$7B,$2F,$C6,$7A,$CB,$7E,$16,$C7,$C2,$C3,$D2,$7C
	.dc.b	$12,$CE,$CF,$24,$79,$EF,$C1,$87,$C2,$6C,$64,$F6,$DF,$3A,$E3,$76
	.dc.b	$7D,$77,$EE,$BE,$37,$E7,$BE,$ED,$AD,$9D,$5F,$4B,$F0,$8E,$F1,$E7
	.dc.b	$77,$1D,$D6,$3F,$8F,$9F,$A5,$87,$BA,$7E,$44,$78,$0E,$4F,$0C,$7E
	.dc.b	$6B,$DE,$78,$36,$7D,$A7,$0A,$FC,$AC,$74,$3C,$71,$9A,$E7,$1F,$86
	.dc.b	$32,$6C,$F0,$63,$F3,$5F,$69,$F5,$E7,$BB,$4F,$8B,$7F,$E0,$FD,$DF
	.dc.b	$40,$FA,$2F,$F3,$CE,$1F,$E4,$78,$C3,$FE,$C3,$E2,$7D,$16,$B1,$FF
	.dc.b	$40,$56,$9F,$F2,$9E,$1D,$30,$D7,$27,$F8,$9B,$EB,$F8,$4E,$72,$E9
	.dc.b	$1C,$ED,$F1,$6F,$47,$0D,$AF,$F6,$A7,$D1,$BE,$33,$35,$27,$3F,$29
	.dc.b	$70,$FF,$1B,$5A,$3B,$71,$BE,$8A,$9E,$BF,$61,$F4,$6F,$23,$39,$7F
	.dc.b	$93,$72,$FC,$2D,$FF,$99,$DF,$E3,$2D,$7F,$8B,$B5,$DC,$7F,$11,$69
	.dc.b	$F1,$BE,$B5,$F3,$3D,$A3,$0F,$C6,$E4,$F8,$11,$F3,$1B,$C5,$FE,$E2
	.dc.b	$E3,$FF,$28,$FC,$0E,$F8,$66,$5F,$5F,$DE,$B8,$FF,$95,$FA,$11,$F5
	.dc.b	$3D,$E7,$D2,$5F,$8A,$C3,$70,$FF,$7B,$B7,$0F,$F3,$05,$70,$9F,$F7
	.dc.b	$DD,$7C,$CD,$A4,$7F,$9C,$F8,$FF,$D4,$99,$63,$0F,$F4,$27,$27,$C0
	.dc.b	$AF,$88,$D8,$FE,$6C,$D7,$FA,$A3,$5F,$E8,$F7,$F5,$F9,$1F,$3F,$D7
	.dc.b	$85,$87,$F9,$C3,$2F,$B7,$3F,$92,$F7,$DE,$C5,$FA,$4C,$36,$FF,$DF
	.dc.b	$4D,$CF,$FC,$FD,$5C,$27,$FD,$FD,$5F,$9A,$F4,$FF,$D1,$71,$FE,$B9
	.dc.b	$6D,$70,$FA,$4D,$78,$72,$BF,$D2,$DC,$EF,$77,$FD,$39,$1C,$67,$F4
	.dc.b	$F1,$8B,$F0,$FD,$27,$A7,$C1,$F5,$36,$F9,$E8,$8C,$57,$09,$CD,$70
	.dc.b	$69,$C2,$77,$C1,$F8,$BB,$7D,$27,$C7,$F8,$5F,$90,$DB,$EE,$DC,$7E
	.dc.b	$2B,$6F,$97,$B3,$F3,$FC,$1F,$A0,$DB,$F3,$B6,$FE,$6B,$6B,$FD,$A5
	.dc.b	$CE,$EE,$FD,$B9,$F6,$7E,$92,$E1,$FB,$39,$D3,$F8,$A9,$FD,$84,$63
	.dc.b	$0F,$C5,$C6,$0F,$DF,$7E,$83,$FE,$B4,$AE,$51,$89,$FD,$27,$D2,$C9
	.dc.b	$F9,$BD,$16,$7F,$DB,$9F,$24,$AF,$DF,$36,$BF,$7C,$D9,$FF,$08,$35
	.dc.b	$FF,$70,$72,$B8,$5C,$FF,$95,$3F,$69,$F1,$37,$0F,$C3,$F3,$79,$F0
	.dc.b	$F0,$CF,$E1,$9F,$E5,$18,$A2,$9E,$42,$57,$91,$FE,$2D,$8F,$E4,$7A
	.dc.b	$FF,$1E,$70,$9E,$EF,$F3,$37,$F4,$17,$FE,$FB,$6F,$0F,$EB,$A3,$F5
	.dc.b	$FF,$46,$FA,$5E,$1D,$B2,$FF,$90,$1C,$AB,$FC,$9C,$D7,$FD,$CA,$7E
	.dc.b	$E2,$3F,$94,$47,$63,$EB,$7B,$7F,$19,$FE,$64,$66,$7F,$7F,$19,$58
	.dc.b	$79,$57,$D2,$FF,$0D,$8C,$36,$BF,$05,$8A,$FF,$A9,$4E,$6F,$FD,$F2
	.dc.b	$FF,$11,$6F,$83,$FC,$F6,$31,$1D,$A3,$07,$F6,$5B,$5B,$E1,$7A,$CE
	.dc.b	$63,$13,$FD,$37,$D2,$FE,$B8,$DF,$3F,$F9,$B7,$D2,$FF,$65,$9C,$E3
	.dc.b	$EE,$AF,$93,$18,$8F,$F3,$D3,$5C,$D7,$F5,$38,$C4,$03,$D3,$4F,$E9
	.dc.b	$9E,$43,$E1,$BC,$17,$FD,$1D,$F1,$7A,$13,$1F,$CA,$FA,$B7,$2F,$21
	.dc.b	$D7,$2F,$EE,$13,$FD,$85,$B1,$FE,$73,$6B,$E8,$AF,$31,$FE,$0D,$FA
	.dc.b	$47,$D2,$BE,$63,$B6,$BF,$91,$9C,$C7,$F8,$27,$E9,$7F,$80,$4F,$E0
	.dc.b	$5B,$97,$FA,$2B,$E5,$9F,$23,$BF,$0A,$4F,$F7,$5A,$FE,$11,$A7,$FA
	.dc.b	$8D,$AF,$E5,$9A,$FC,$97,$AB,$F9,$0F,$A6,$7C,$A7,$48,$FF,$11,$C6
	.dc.b	$2B,$FC,$15,$F4,$CF,$4F,$FC,$40,$D8,$FA,$9A,$7E,$B7,$C8,$6A,$F9
	.dc.b	$4D,$E1,$2F,$FC,$19,$87,$1C,$BF,$D3,$CD,$7F,$D7,$8D,$7F,$01,$E0
	.dc.b	$FF,$51,$F8,$EF,$F1,$87,$D1,$3E,$4B,$FA,$8C,$61,$B8,$7F,$88,$67
	.dc.b	$4F,$F1,$49,$D3,$FE,$25,$FA,$47,$CE,$3C,$6F,$F6,$8F,$CC,$7E,$FA
	.dc.b	$73,$18,$CF,$F2,$5F,$4C,$F4,$BF,$86,$7C,$43,$0E,$18,$F3,$BF,$F6
	.dc.b	$C4,$63,$0E,$50,$C7,$AB,$A3,$AE,$1D,$2B,$84,$64,$E1,$7D,$70,$F0
	.dc.b	$47,$E9,$6F,$9C,$62,$18,$F9,$F2,$F0,$D6,$B0,$EB,$7B,$8C,$5F,$8A
	.dc.b	$B7,$87,$7B,$D9,$27,$47,$96,$B8,$BE,$6C,$B8,$9F,$03,$1F,$14,$6E
	.dc.b	$FA,$3E,$96,$BC,$4F,$81,$1C,$31,$F3,$4E,$AF,$E4,$D7,$F3,$43,$81
	.dc.b	$5C,$BF,$28,$D7,$C5,$7F,$8B,$3E,$9B,$FC,$BC,$D8,$FF,$78,$35,$DF
	.dc.b	$96,$9F,$F3,$25,$7F,$9E,$FE,$2E,$F8,$A6,$92,$FF,$DF,$53,$B7,$E3
	.dc.b	$4E,$DB,$3F,$84,$FA,$6F,$F4,$5C,$7D,$69,$D6,$7F,$55,$BB,$7E,$55
	.dc.b	$B3,$FC,$96,$B8,$BD,$09,$5C,$6F,$B7,$B9,$FE,$EE,$C3,$94,$77,$25
	.dc.b	$7E,$3E,$F9,$DF,$18,$E7,$5C,$1D,$5F,$37,$F2,$58,$79,$1F,$CF,$37
	.dc.b	$1F,$85,$7F,$95,$C3,$AE,$1C,$E3,$B1,$1F,$ED,$F1,$AB,$72,$BC,$87
	.dc.b	$D3,$7F,$74,$ED,$FE,$D5,$CE,$F8,$FC,$73,$7F,$78,$8C,$63,$FE,$5C
	.dc.b	$C3,$8B,$70,$DB,$FD,$3E,$76,$FF,$46,$0E,$30,$FE,$A3,$3C,$25,$F4
	.dc.b	$AF,$A6,$C7,$A0,$F9,$7A,$6E,$75,$5F,$EB,$76,$BD,$E7,$83,$FE,$BB
	.dc.b	$6B,$E3,$8D,$E1,$FE,$09,$FA,$6F,$F1,$64,$FF,$B0,$7E,$9B,$FA,$4B
	.dc.b	$7C,$9F,$34,$EA,$3D,$8D,$AF,$F0,$95,$70,$9F,$80,$D8,$ED,$EB,$F2
	.dc.b	$3F,$D8,$D2,$AE,$8F,$EC,$DA,$FF,$8A,$1A,$FF,$7B,$FA,$57,$C4,$FD
	.dc.b	$0F,$69,$F4,$DF,$EE,$1A,$FE,$9B,$C7,$FE,$01,$9C,$FA,$5F,$3F,$0E
	.dc.b	$35,$FE,$DF,$6B,$FD,$F2,$76,$F2,$0E,$1F,$B9,$1C,$33,$FF,$19,$36
	.dc.b	$FF,$C7,$ED,$9F,$46,$3E,$7C,$7F,$BF,$47,$87,$C1,$FC,$4F,$3B,$E7
	.dc.b	$9F,$45,$86,$D6,$C8,$7F,$84,$A7,$84,$EC,$F7,$54,$9D,$E7,$E6,$8E
	.dc.b	$9E,$5F,$E4,$4F,$F1,$14,$FF,$C0,$1F,$A6,$7C,$67,$E5,$5B,$1F,$0C
	.dc.b	$FE,$ED,$EC,$F1,$35,$32,$35,$5D,$1F,$D1,$37,$0F,$F2,$C3,$77,$FF
	.dc.b	$32,$B6,$7F,$C7,$18,$7F,$1F,$7F,$55,$E7,$FF,$A2,$9F,$17,$CD,$7F
	.dc.b	$83,$5B,$E8,$7E,$AE,$75,$1C,$3E,$0B,$F5,$D5,$B8,$FD,$A3,$72,$FF
	.dc.b	$35,$57,$06,$D7,$32,$57,$83,$E3,$FE,$BF,$82,$FA,$6A,$49,$F8,$DF
	.dc.b	$4D,$FE,$BD,$6F,$B3,$97,$EC,$57,$17,$27,$F0,$DF,$A6,$FE,$09,$5C
	.dc.b	$63,$F8,$53,$72,$FF,$2C,$CF,$8B,$D9,$95,$FC,$55,$FC,$47,$8F,$CB
	.dc.b	$E4,$EC,$FF,$BD,$3A,$7F,$89,$72,$7F,$A5,$32,$57,$E3,$FD,$19,$7B
	.dc.b	$7D,$FF,$47,$FC,$5F,$E9,$3F,$80,$61,$FE,$6D,$F5,$9F,$23,$FE,$93
	.dc.b	$DF,$FD,$51,$18,$9E,$1E,$2C,$CB,$F5,$93,$96,$DF,$1A,$FF,$53,$E1
	.dc.b	$DA,$7F,$CF,$4D,$AE,$11,$FE,$72,$6E,$1F,$EA,$39,$E1,$DF,$FE,$D1
	.dc.b	$F5,$3F,$5B,$F1,$7F,$E8,$2C,$3F,$79,$3F,$ED,$5F,$83,$CD,$49,$CD
	.dc.b	$6D,$FD,$73,$70,$FF,$62,$5F,$59,$DF,$95,$DB,$D2,$91,$82,$1C,$9C
	.dc.b	$BD,$23,$CB,$E2,$79,$BC,$19,$3E,$28,$CB,$E4,$8C,$BE,$48,$CB,$E6
	.dc.b	$8C,$E1,$E6,$C7,$F8,$4C,$FF,$A6,$21,$F4,$AE,$3F,$E2,$3F,$18,$FA
	.dc.b	$6A,$7F,$5C,$AD,$CD,$AE,$17,$FE,$F8,$AD,$BF,$C9,$EB,$6F,$CA,$F7
	.dc.b	$FE,$8F,$54,$8C,$10,$F9,$AE,$5F,$9F,$1F,$8F,$F1,$FF,$B3,$7D,$37
	.dc.b	$FB,$36,$7F,$89,$35,$FE,$DD,$C9,$FC,$25,$36,$35,$87,$FB,$9D,$BC
	.dc.b	$DF,$8E,$BE,$D7,$D3,$D2,$EE,$61,$FE,$5D,$DB,$FE,$79,$7F,$55,$F2
	.dc.b	$7F,$CD,$4F,$E6,$DF,$01,$23,$24,$38,$BF,$2F,$E9,$1F,$05,$FC,$F7
	.dc.b	$C8,$95,$F2,$67,$2D,$9C,$BE,$29,$E5,$1F,$B3,$F0,$3E,$2F,$AF,$2B
	.dc.b	$F9,$B7,$02,$75,$1F,$22,$74,$FC,$88,$CB,$FE,$17,$DF,$FC,$0F,$C1
	.dc.b	$F1,$EB,$2B,$83,$FD,$D9,$4F,$A6,$7C,$87,$F5,$AF,$A6,$E3,$5F,$EB
	.dc.b	$26,$D7,$FB,$CE,$B6,$FF,$4B,$6B,$FD,$6E,$37,$DE,$78,$C7,$7B,$F9
	.dc.b	$5F,$4D,$FD,$A3,$D0,$7C,$D5,$33,$5F,$2E,$70,$FF,$00,$9D,$B9,$3F
	.dc.b	$AE,$7C,$37,$F9,$07,$44,$3E,$97,$6B,$E9,$9E,$A7,$C2,$9C,$C6,$27
	.dc.b	$8C,$70,$BE,$91,$CF,$2F,$93,$18,$9E,$37,$CE,$7E,$6D,$F3,$BE,$07
	.dc.b	$92,$F8,$4F,$2E,$D6,$3E,$08,$E3,$87,$3B,$F5,$E1,$CF,$1F,$26,$1C
	.dc.b	$27,$A8,$C7,$AB,$C2,$F9,$3F,$05,$B1,$A8,$E1,$3C,$EB,$D7,$1E,$33
	.dc.b	$91,$4D,$E3,$6B,$96,$1E,$5C,$7D,$35,$9C,$3A,$47,$BE,$F9,$56,$EB
	.dc.b	$A9,$EA,$BF,$21,$D3,$1F,$4C,$EE,$7C,$26,$F0,$EA,$F2,$C7,$D2,$E7
	.dc.b	$0F,$11,$E0,$BE,$67,$C2,$AE,$33,$D2,$B8,$93,$1F,$44,$73,$3E,$7E
	.dc.b	$1E,$33,$C9,$8F,$A2,$F9,$C6,$70,$ED,$8F,$8A,$F8,$C7,$89,$C6,$3F
	.dc.b	$72,$7C,$D1,$0F,$04,$65,$E2,$65,$C1,$31,$FA,$31,$A3,$8B,$EA,$F2
	.dc.b	$3E,$29,$1C,$1E,$E9,$F8,$13,$F2,$E3,$04,$70,$43,$E4,$39,$7F,$1C
	.dc.b	$79,$5C,$65,$F6,$E3,$07,$E8,$F2,$FA,$31,$98,$F4,$57,$AD,$D6,$1C
	.dc.b	$1E,$A7,$B0,$C9,$FE,$F3,$C3,$EB,$F9,$1F,$E0,$1B,$FD,$CB,$70,$FD
	.dc.b	$43,$FE,$05,$AF,$E0,$FF,$09,$FC,$22,$31,$08,$C5,$FF,$28,$8D,$9D
	.dc.b	$CC,$46,$9A,$7E,$D3,$E2,$DF,$23,$FC,$C6,$3F,$6E,$72,$CF,$F8,$AC
	.dc.b	$EB,$97,$F5,$D3,$9D,$7F,$9B,$CE,$63,$89,$E5,$3B,$5F,$5B,$FE,$3F
	.dc.b	$1C,$4F,$E4,$73,$C6,$33,$1C,$5F,$0C,$F1,$AF,$74,$E5,$F0,$C6,$63
	.dc.b	$DB,$3C,$2F,$CF,$1B,$3B,$CE,$DF,$E9,$F8,$7F,$09,$C3,$C5,$87,$68
	.dc.b	$C5,$6E,$BD,$02,$B6,$FC,$CC,$7C,$09,$F5,$D7,$F4,$07,$26,$23,$E2
	.dc.b	$E1,$E6,$9E,$ED,$8C,$9F,$32,$32,$D3,$38,$F3,$3C,$25,$C7,$E2,$D7
	.dc.b	$38,$CC,$7B,$67,$8B,$63,$31,$F2,$A3,$8B,$DE,$32,$FC,$83,$89,$B3
	.dc.b	$E6,$CF,$79,$E3,$3E,$39,$C9,$CE,$32,$78,$4A,$F3,$3E,$76,$3F,$45
	.dc.b	$FB,$46,$2B,$B3,$D2,$BB,$39,$3B,$19,$AE,$CD,$3B,$63,$F6,$9F,$64
	.dc.b	$FB,$9F,$BF,$19,$6D,$F8,$F0,$FA,$75,$E0,$AD,$57,$99,$B3,$AB,$F0
	.dc.b	$CF,$96,$B8,$5F,$E6,$75,$7D,$03,$E9,$7F,$69,$1F,$E8,$4E,$27,$D2
	.dc.b	$E3,$5F,$F0,$5C,$F0,$AC,$64,$E7,$61,$C2,$FF,$CD,$58,$F5,$7F,$C3
	.dc.b	$18,$F0,$C7,$94,$60,$FF,$7A,$7C,$63,$E8,$BA,$E2,$FA,$57,$C2,$F8
	.dc.b	$34,$5E,$F0,$FF,$51,$D6,$9E,$78,$72,$8C,$57,$F8,$3A,$7F,$C8,$35
	.dc.b	$C6,$30,$7A,$E7,$78,$FF,$17,$8F,$D9,$7D,$2F,$A7,$E9,$5F,$37,$E3
	.dc.b	$E5,$4E,$63,$0D,$CB,$E4,$9F,$C6,$9C,$5F,$80,$EB,$1F,$E9,$1B,$EA
	.dc.b	$F0,$8F,$E3,$33,$C3,$0F,$AF,$BF,$A5,$FB,$1E,$CF,$CD,$8C,$5F,$F9
	.dc.b	$0B,$7F,$AB,$39,$DB,$E6,$F0,$FF,$BA,$2B,$EB,$C6,$27,$E8,$FB,$AF
	.dc.b	$80,$BF,$0D,$86,$DF,$FB,$4A,$74,$D9,$D3,$4F,$56,$BF,$C6,$F7,$FC
	.dc.b	$4F,$B4,$FA,$57,$D0,$7C,$B9,$CC,$62,$F8,$F1,$FC,$4C,$3F,$DD,$13
	.dc.b	$97,$E5,$46,$78,$7F,$DF,$33,$97,$ED,$46,$6B,$FE,$00,$4E,$5F,$E9
	.dc.b	$B7,$E0,$C3,$F2,$3E,$C7,$FB,$36,$7F,$DD,$7F,$0D,$8F,$63,$2D,$8C
	.dc.b	$9F,$EF,$08,$CB,$FE,$CB,$F3,$FF,$CF,$AF,$FB,$E3,$D9,$79,$9F,$A3
	.dc.b	$9C,$ED,$F9,$2E,$65,$6A,$F9,$78,$3F,$A4,$8C,$57,$FB,$8E,$F8,$F0
	.dc.b	$7F,$BF,$27,$2F,$FB,$A7,$E9,$3D,$7D,$BE,$C6,$E3,$F8,$DC,$9B,$27
	.dc.b	$39,$78,$E3,$15,$B7,$F6,$2D,$9D,$E1,$DE,$35,$87,$CA,$FA,$4F,$97
	.dc.b	$F0,$16,$79,$37,$1E,$CD,$BF,$F8,$00,$FD,$2F,$57,$E9,$FA,$D6,$FC
	.dc.b	$4D,$BF,$33,$6F,$FC,$1D,$A7,$B2,$3F,$6D,$F0,$BF,$90,$FA,$3F,$F5
	.dc.b	$2E,$1F,$8C,$8C,$61,$A8,$FD,$87,$4A,$B7,$3F,$8E,$AD,$D7,$13,$73
	.dc.b	$AC,$7F,$1F,$3A,$AF,$B9,$D6,$FA,$3F,$D6,$9F,$8E,$E8,$B8,$FE,$9B
	.dc.b	$E3,$B4,$FE,$B7,$E1,$7F,$90,$46,$30,$E7,$1F,$96,$E6,$95,$B3,$EB
	.dc.b	$3F,$87,$9E,$33,$67,$B3,$87,$F2,$9A,$D9,$FC,$97,$32,$3F,$20,$49
	.dc.b	$E2,$70,$3F,$7A,$7E,$9E,$7F,$8E,$37,$5F,$F2,$02,$FA,$17,$D1,$FF
	.dc.b	$3B,$8D,$1F,$EF,$42,$D7,$FB,$DF,$D6,$67,$BE,$CF,$EA,$35,$FE,$F8
	.dc.b	$8C,$3F,$C9,$CC,$1E,$48,$DC,$F2,$8F,$E2,$7F,$47,$FC,$0E,$FF,$DC
	.dc.b	$ED,$8F,$34,$FF,$4F,$8D,$49,$D6,$2F,$80,$D7,$81,$E4,$8C,$61,$DA
	.dc.b	$31,$3A,$8F,$34,$EA,$FF,$A1,$7A,$4F,$9A,$FE,$BC,$DA,$FE,$5D,$F4
	.dc.b	$78,$99,$1F,$CC,$7E,$0E,$DF,$E3,$9B,$C1,$FA,$2E,$74,$62,$BF,$A2
	.dc.b	$46,$1F,$E6,$B8,$FF,$60,$57,$FE,$4B,$F4,$3F,$88,$7A,$7E,$46,$4F
	.dc.b	$F5,$5F,$A2,$7D,$1B,$E8,$F3,$87,$F6,$C9,$CC,$63,$4F,$F4,$14,$62
	.dc.b	$FC,$38,$7F,$83,$23,$17,$FD,$97,$E8,$DD,$DF,$DC,$E3,$CA,$23,$37
	.dc.b	$FE,$78,$8C,$4E,$DF,$5C,$E9,$FF,$3E,$B7,$0F,$F1,$0F,$D1,$BE,$47
	.dc.b	$FC,$1D,$39,$C3,$FC,$43,$39,$7F,$AD,$C6,$6D,$6A,$D4,$7A,$70,$FF
	.dc.b	$10,$46,$27,$FC,$19,$F4,$7F,$E2,$8C,$3F,$87,$FE,$00,$F9,$2F,$F0
	.dc.b	$F4,$E7,$1F,$F0,$AF,$72,$B5,$7F,$DB,$2B,$58,$7F,$AB,$6B,$58,$72
	.dc.b	$F5,$BF,$B7,$E1,$C3,$1F,$DE,$FC,$57,$F8,$AF,$E8,$1F,$1D,$FD,$36
	.dc.b	$30,$DD,$3F,$BC,$CE,$9F,$EB,$53,$A7,$FB,$EF,$D1,$3E,$6B,$F8,$46
	.dc.b	$1F,$EC,$9C,$FE,$BC,$E7,$4F,$F6,$7C,$7D,$AF,$A4,$79,$DF,$EC,$5F
	.dc.b	$84,$F8,$58,$7E,$4B,$0E,$0F,$33,$18,$F6,$9D,$C2,$77,$18,$BE,$B9
	.dc.b	$78,$2B,$8E,$55,$E1,$87,$3C,$BC,$ED,$8F,$D1,$DF,$27,$C4,$70,$AE
	.dc.b	$58,$F8,$2F,$8C,$75,$AE,$18,$F9,$AB,$57,$E2,$39,$B8,$53,$C2,$F9
	.dc.b	$C6,$B0,$F2,$3E,$13,$C7,$3B,$7A,$5F,$79,$DD,$F5,$AD,$57,$1A,$D6
	.dc.b	$1F,$A9,$AF,$9F,$3F,$EF,$F6,$BE,$23,$F6,$23,$E1,$DE,$33,$E9,$3F
	.dc.b	$65,$18,$6C,$66,$7F,$DD,$B3,$97,$FD,$DF,$5F,$C8,$7E,$1F,$C5,$E1
	.dc.b	$1A,$49,$DC,$65,$F9,$2D,$CB,$E7,$37,$5F,$E0,$FF,$49,$FE,$88,$AF
	.dc.b	$AF,$39,$7F,$1C,$D7,$F1,$4D,$8E,$11,$DA,$F8,$91,$FC,$1C,$6A,$FC
	.dc.b	$33,$96,$CE,$BD,$6F,$E3,$31,$82,$61,$FA,$6C,$3C,$13,$A8,$C5,$F3
	.dc.b	$6D,$F3,$30,$7A,$A2,$AC,$3C,$93,$FE,$83,$F3,$5E,$CF,$F2,$AE,$9F
	.dc.b	$F2,$47,$9D,$79,$34,$DC,$7F,$D7,$D7,$C1,$FF,$2D,$5E,$7D,$5F,$F2
	.dc.b	$BD,$5A,$FE,$FB,$5A,$3F,$D0,$3F,$89,$BE,$9B,$A4,$23,$0E,$A3,$C1
	.dc.b	$C8,$FC,$2F,$33,$FE,$A7,$AB,$A7,$FA,$D5,$BA,$7F,$44,$6C,$FE,$4E
	.dc.b	$AB,$DF,$D5,$F0,$CF,$FA,$42,$3C,$B1,$E0,$F8,$27,$95,$A8,$93,$A3
	.dc.b	$FD,$69,$97,$FA,$96,$73,$1E,$59,$B9,$32,$FE,$D7,$7B,$AF,$F5,$1F
	.dc.b	$D1,$BD,$6F,$EB,$10,$8C,$3F,$D6,$DA,$FE,$C0,$D4,$70,$8C,$DF,$FB
	.dc.b	$5D,$B7,$FD,$99,$B5,$FD,$F3,$E9,$3F,$50,$FF,$B4,$63,$2F,$43,$CB
	.dc.b	$7F,$5A,$77,$7F,$EE,$3E,$E7,$B7,$F6,$BE,$BF,$E6,$9F,$3A,$33,$87
	.dc.b	$FB,$01,$AF,$FB,$7D,$B5,$CE,$B9,$78,$3F,$84,$9C,$E7,$FD,$F9,$B7
	.dc.b	$FE,$14,$6A,$C3,$FA,$4B,$FE,$CE,$FC,$1F,$FA,$93,$FE,$AF,$F6,$7F
	.dc.b	$28,$7E,$42,$33,$1F,$EF,$16,$E5,$FE,$FF,$6E,$1F,$E0,$E6,$CB,$9D
	.dc.b	$F5,$5F,$D4,$2F,$86,$5D,$49,$5F,$EA,$D9,$D6,$1C,$FB,$BF,$51,$0F
	.dc.b	$9F,$A7,$96,$25,$7A,$DA,$FF,$83,$AB,$FC,$6B,$F0,$EC,$DF,$33,$AC
	.dc.b	$FD,$58,$5B,$DD,$6B,$0F,$F6,$97,$B7,$FB,$72,$3F,$EF,$6F,$A6,$7D
	.dc.b	$8F,$ED,$9B,$19,$8C,$1F,$BF,$F8,$F9,$19,$21,$FE,$8A,$7F,$82,$7C
	.dc.b	$C4,$8F,$F7,$13,$FC,$37,$E0,$7F,$9F,$7D,$5F,$F0,$4F,$A2,$FD,$45
	.dc.b	$7F,$94,$B6,$AA,$2C,$61,$BA,$FF,$9E,$BC,$87,$C8,$F5,$6D,$F3,$9D
	.dc.b	$4E,$4F,$17,$DD,$7D,$69,$FF,$30,$FB,$BF,$55,$E0,$63,$97,$F8,$CC
	.dc.b	$66,$3E,$A1,$FE,$68,$F9,$E7,$C3,$6A,$A4,$F2,$15,$C7,$4F,$6D,$F0
	.dc.b	$9D,$4E,$72,$FF,$2F,$B6,$3C,$47,$39,$E0,$FF,$AE,$3E,$B7,$F9,$97
	.dc.b	$CC,$48,$FF,$7F,$79,$69,$FE,$6C,$66,$3F,$D7,$F3,$87,$7F,$6D,$FC
	.dc.b	$EF,$49,$18,$21,$FE,$C6,$72,$FF,$B2,$23,$FA,$17,$DA,$FF,$43,$FB
	.dc.b	$59,$3E,$28,$CB,$E4,$8C,$BF,$ED,$08,$CB,$FE,$D2,$8C,$EB,$FE,$FC
	.dc.b	$C9,$5F,$D5,$63,$EF,$F8,$57,$87,$F8,$26,$1F,$62,$72,$FC,$2F,$B0
	.dc.b	$D5,$48,$D5,$70,$7F,$DC,$33,$FD,$6F,$E0,$BF,$AE,$7C,$CC,$9F,$97
	.dc.b	$19,$7E,$6C,$65,$FF,$74,$C7,$F5,$EF,$C1,$5D,$74,$93,$F6,$7A,$EB
	.dc.b	$F8,$A1,$3C,$1F,$C4,$35,$D7,$1A,$78,$E9,$FA,$6A,$D3,$6F,$FB,$7D
	.dc.b	$6E,$FF,$07,$D9,$FD,$E7,$93,$F5,$31,$FC,$75,$BE,$2B,$87,$F8,$0C
	.dc.b	$FE,$7A,$31,$5C,$79,$E4,$F4,$3F,$3D,$E1,$7F,$A3,$1B,$F0,$1F,$E0
	.dc.b	$08,$C4,$73,$FC,$F7,$FB,$93,$F0,$63,$3A,$FF,$00,$6D,$FF,$B0,$1B
	.dc.b	$3F,$CC,$EB,$8D,$7F,$06,$FA,$67,$CE,$7F,$83,$9B,$5F,$B1,$6B,$FB
	.dc.b	$8E,$E7,$E6,$BF,$AA,$BE,$57,$C4,$95,$FC,$0A,$B7,$D9,$88,$FB,$18
	.dc.b	$74,$BE,$31,$BC,$3A,$67,$FC,$6F,$A3,$FD,$E7,$3F,$90,$F9,$0F,$CC
	.dc.b	$1F,$4F,$D5,$B1,$92,$1F,$47,$8D,$1F,$9F,$FA,$F9,$3F,$9A,$7E,$E6
	.dc.b	$D8,$8F,$E2,$B9,$FF,$24,$9C,$3F,$CB,$B8,$FF,$82,$57,$F8,$D3,$E0
	.dc.b	$5F,$49,$FE,$DC,$C1,$F4,$9F,$E4,$15,$F4,$CF,$A6,$FF,$79,$CE,$63
	.dc.b	$12,$38,$FC,$07,$FB,$E3,$0E,$B9,$E6,$FA,$46,$2B,$67,$1B,$E6,$E0
	.dc.b	$F0,$E1,$E0,$C3,$A6,$3E,$7C,$39,$B6,$3C,$55,$C0,$F0,$38,$BE,$B5
	.dc.b	$C3,$1C,$D7,$A7,$1F,$24,$FB,$71,$F0,$4F,$95,$AF,$FA,$8F,$1F,$03
	.dc.b	$C5,$AF,$84,$F5,$61,$E0,$C7,$D3,$8F,$82,$79,$36,$75,$1D,$CD,$63
	.dc.b	$E4,$AE,$31,$D3,$0F,$1D,$70,$8E,$67,$C9,$AE,$0F,$33,$81,$D3,$0F
	.dc.b	$3C,$6A,$3D,$18,$FC,$17,$58,$C7,$89,$23,$E5,$3D,$EB,$E0,$DF,$8C
	.dc.b	$F2,$5F,$53,$B3,$9C,$3E,$5C,$66,$78,$C7,$B0,$CB,$E6,$33,$5F,$1E
	.dc.b	$33,$87,$97,$0F,$44,$F9,$71,$F8,$F1,$E6,$32,$47,$24,$8F,$8D,$19
	.dc.b	$7E,$E4,$F5,$9F,$C5,$57,$96,$70,$68,$F0,$1C,$A3,$A3,$F5,$2B,$A3
	.dc.b	$F3,$A3,$38,$7C,$4A,$E9,$3F,$2E,$FA,$BB,$3A,$1D,$EF,$D9,$7E,$48
	.dc.b	$F8,$87,$0B,$E8,$E0,$E8,$4F,$25,$F2,$9F,$8C,$6C,$7A,$DA,$FE,$31
	.dc.b	$AF,$EF,$FE,$17,$F8,$04,$62,$11,$89,$FF,$4D,$36,$3F,$6A,$68,$FF
	.dc.b	$80,$11,$A7,$F8,$07,$A7,$FE,$B4,$E5,$FD,$34,$77,$33,$EF,$7F,$92
	.dc.b	$33,$FE,$2D,$19,$7F,$8C,$57,$18,$FE,$14,$67,$3F,$E0,$B1,$99,$E3
	.dc.b	$3E,$37,$8D,$75,$AF,$24,$62,$FC,$75,$B9,$E7,$5B,$8F,$6E,$3F,$BD
	.dc.b	$C3,$B5,$72,$37,$7C,$A7,$57,$C6,$FF,$8C,$C6,$27,$51,$F0,$1B,$87
	.dc.b	$8E,$3D,$A7,$8C,$FF,$43,$D7,$37,$E0,$18,$9E,$37,$E2,$8E,$37,$DD
	.dc.b	$A7,$72,$DF,$73,$E3,$C6,$DB,$1B,$69,$B9,$F8,$11,$E1,$39,$BE,$18
	.dc.b	$DB,$4D,$C3,$0F,$92,$FC,$C3,$B4,$6A,$33,$1F,$20,$C9,$B9,$F1,$36
	.dc.b	$7B,$34,$F2,$47,$89,$B1,$D7,$0F,$95,$1F,$34,$F2,$38,$BF,$85,$3B
	.dc.b	$C3,$E6,$D7,$23,$B1,$C1,$CB,$4F,$86,$6C,$F5,$5F,$A1,$B3,$E8,$69
	.dc.b	$B3,$EE,$9C,$E3,$D4,$D8,$F2,$3D,$8C,$0D,$1F,$BF,$3E,$23,$FD,$67
	.dc.b	$12,$31,$87,$FA,$93,$8F,$FD,$19,$EC,$6A,$3F,$D7,$13,$AC,$3F,$C2
	.dc.b	$2D,$BE,$58,$6E,$7F,$CA,$17,$67,$C3,$36,$BF,$DD,$F7,$FE,$7B,$6D
	.dc.b	$FF,$88,$FD,$7D,$7B,$12,$B4,$F5,$CB,$FD,$41,$39,$AF,$F3,$77,$0F
	.dc.b	$FA,$CA,$FF,$D6,$8D,$8F,$F2,$27,$E6,$FD,$B3,$FF,$68,$CE,$5B,$E9
	.dc.b	$7A,$69,$39,$BF,$F1,$07,$82,$C3,$94,$7F,$84,$A7,$55,$FC,$9A,$F9
	.dc.b	$46,$2F,$86,$78,$9E,$DF,$A7,$3E,$42,$9E,$DF,$CF,$BF,$9B,$E5,$FE
	.dc.b	$CD,$9D,$36,$76,$FD,$79,$DE,$5F,$E8,$FA,$DB,$FE,$4D,$BF,$91,$DE
	.dc.b	$F9,$7B,$18,$6E,$1F,$81,$9D,$36,$74,$FC,$CF,$1F,$DA,$7D,$F5,$F3
	.dc.b	$6F,$8F,$CC,$6C,$98,$FF,$BB,$1D,$C6,$A3,$EC,$D5,$E1,$57,$F9,$3C
	.dc.b	$D9,$C9,$FF,$2F,$62,$FB,$AF,$6F,$AB,$EE,$6D,$7A,$DB,$87,$F9,$BD
	.dc.b	$B7,$FA,$1F,$83,$FC,$A1,$F1,$BE,$7E,$DF,$E5,$62,$DF,$F9,$DE,$6D
	.dc.b	$78,$BE,$1B,$E6,$F1,$7C,$E6,$E1,$F2,$DB,$7F,$E7,$F6,$D7,$19,$FA
	.dc.b	$15,$C5,$A7,$D3,$FA,$BF,$B3,$F3,$DF,$69,$B1,$93,$EC,$46,$5A,$7D
	.dc.b	$BF,$2F,$F0,$3F,$03,$F8,$26,$C6,$4F,$C0,$46,$5A,$7F,$9F,$F0,$FA
	.dc.b	$79,$7F,$A0,$7E,$71,$F6,$1F,$A7,$6D,$6C,$FB,$1C,$8F,$97,$D5,$49
	.dc.b	$D5,$F3,$8F,$C5,$56,$E7,$32,$BE,$AF,$FB,$AD,$B8,$79,$23,$57,$CB
	.dc.b	$0E,$D1,$FC,$86,$FF,$D3,$1F,$0A,$FC,$39,$F0,$DF,$C5,$E3,$0D,$87
	.dc.b	$CA,$58,$FE,$09,$E4,$FE,$3E,$FC,$AD,$C3,$B0,$C3,$9C,$EB,$9B,$F6
	.dc.b	$3E,$94,$8F,$83,$3E,$3F,$96,$C5,$FF,$B0,$FE,$D7,$BF,$57,$EE,$1B
	.dc.b	$7E,$08,$B3,$B8,$C1,$FB,$9F,$9F,$FD,$F9,$AE,$1F,$D3,$BF,$C7,$7C
	.dc.b	$87,$97,$4F,$E7,$1E,$7F,$C4,$9C,$5F,$D2,$AD,$B6,$B7,$C0,$AD,$E3
	.dc.b	$F0,$85,$F0,$7E,$2A,$F8,$27,$77,$0C,$FF,$DB,$11,$8C,$BC,$1E,$5F
	.dc.b	$9C,$FE,$81,$1F,$C7,$27,$F0,$16,$9D,$57,$03,$FA,$D5,$6D,$FF,$63
	.dc.b	$CE,$DD,$61,$FE,$F1,$F0,$B8,$B6,$3F,$23,$F7,$6F,$A8,$FE,$88,$D8
	.dc.b	$FC,$CF,$D2,$BE,$F3,$F0,$5A,$7E,$4E,$B5,$87,$C3,$6C,$67,$0F,$EC
	.dc.b	$D7,$C7,$9F,$AC,$4C,$7F,$4B,$C4,$FB,$C7,$C7,$E7,$3F,$C7,$CE,$63
	.dc.b	$FA,$C7,$D2,$3E,$2A,$9F,$73,$EE,$5F,$71,$8C,$BF,$5B,$5B,$6D,$6E
	.dc.b	$75,$8F,$DB,$AD,$3F,$E7,$C0,$E1,$FF,$3F,$B5,$FF,$7A,$B5,$0F,$80
	.dc.b	$7D,$7F,$F6,$F6,$E1,$FC,$03,$D8,$FE,$94,$FF,$02,$8C,$D7,$E3,$5B
	.dc.b	$C3,$FC,$05,$B3,$DA,$7F,$C3,$FE,$87,$5C,$24,$F3,$7F,$BF,$61,$DA
	.dc.b	$BA,$47,$8F,$D5,$95,$E3,$FA,$0B,$5E,$19,$B9,$7F,$11,$C1,$58,$BF
	.dc.b	$7F,$43,$D2,$7C,$27,$F8,$67,$F4,$3F,$F0,$6D,$7F,$6F,$8C,$7B,$5F
	.dc.b	$DD,$9B,$19,$3F,$C6,$93,$C3,$4F,$D7,$7D,$1F,$EE,$B7,$FF,$5B,$79
	.dc.b	$5F,$06,$31,$3F,$EB,$68,$F8,$9F,$48,$F9,$CF,$E7,$7E,$67,$F1,$EA
	.dc.b	$FA,$B7,$C2,$31,$08,$C5,$F5,$F1,$FF,$BF,$4F,$53,$EE,$E1,$D7,$3F
	.dc.b	$1C,$74,$9D,$E1,$D3,$0E,$73,$9C,$B6,$B0,$E4,$F5,$3C,$57,$CB,$1F
	.dc.b	$15,$70,$BE,$75,$C7,$2F,$6D,$6A,$39,$4E,$8D,$E1,$DB,$1F,$26,$3D
	.dc.b	$6B,$C8,$70,$70,$F5,$CB,$C1,$18,$BE,$8E,$AF,$A6,$1E,$1C,$3A,$E5
	.dc.b	$FA,$9C,$7D,$6D,$C3,$EB,$C7,$FB,$A1,$AF,$E6,$FE,$13,$FC,$F1,$F3
	.dc.b	$FC,$5B,$1F,$B3,$6B,$FE,$F0,$6B,$F8,$3F,$1B,$FC,$B6,$FE,$E6,$BF
	.dc.b	$94,$70,$78,$67,$33,$E1,$6B,$C6,$3E,$74,$72,$C7,$ED,$37,$2F,$D8
	.dc.b	$7D,$23,$CD,$FF,$6E,$C7,$1C,$3E,$8D,$6D,$FC,$7B,$55,$71,$7B,$E1
	.dc.b	$1E,$2F,$E1,$A3,$51,$82,$61,$E4,$9D,$61,$E2,$9E,$F1,$CE,$75,$1F
	.dc.b	$8D,$AE,$71,$82,$61,$E9,$7F,$47,$1A,$8C,$12,$7A,$9C,$EF,$C1,$C8
	.dc.b	$F9,$17,$C2,$7F,$15,$E9,$FF,$15,$CE,$7A,$31,$C9,$FC,$22,$FF,$85
	.dc.b	$E3,$FE,$B3,$78,$B7,$0D,$BF,$DB,$A3,$6F,$F6,$E6,$E1,$D9,$E3,$1F
	.dc.b	$DD,$7E,$89,$F4,$9F,$93,$84,$61,$FC,$B3,$5F,$E8,$0D,$7F,$97,$1C
	.dc.b	$23,$C6,$D3,$FC,$73,$1C,$6B,$FC,$93,$C6,$FA,$4C,$65,$FC,$DE,$9F
	.dc.b	$49,$EC,$D5,$D4,$F8,$BC,$37,$3E,$B5,$E0,$BE,$31,$E8,$8F,$64,$62
	.dc.b	$3D,$31,$83,$81,$1F,$F2,$9E,$4F,$8F,$7C,$67,$B3,$E9,$1F,$09,$EE
	.dc.b	$FA,$4F,$ED,$78,$7F,$95,$9A,$FF,$88,$67,$6F,$F9,$69,$AF,$F6,$96
	.dc.b	$BF,$29,$AF,$F5,$96,$BF,$DA,$1E,$3C,$1F,$E5,$F6,$EF,$BB,$F2,$74
	.dc.b	$FD,$2E,$F7,$C3,$7D,$4F,$67,$E7,$E5,$FD,$C9,$AF,$F9,$0A,$38,$1F
	.dc.b	$83,$9E,$31,$FC,$5D,$AF,$F8,$05,$AF,$E0,$9C,$E5,$FE,$73,$6E,$3F
	.dc.b	$DC,$DB,$3E,$1F,$0B,$F0,$FF,$47,$FE,$52,$8F,$CF,$7D,$27,$E0,$7A
	.dc.b	$4E,$AE,$4F,$F2,$9A,$E0,$FF,$84,$DA,$FE,$51,$AF,$F8,$0D,$DE,$3F
	.dc.b	$E8,$76,$E3,$9A,$FC,$87,$91,$FD,$82,$B5,$E0,$E3,$CF,$98,$FF,$A5
	.dc.b	$31,$FF,$28,$36,$7F,$55,$18,$8F,$24,$60,$8E,$C8,$7F,$8E,$A1,$EB
	.dc.b	$FC,$1F,$A4,$93,$FC,$05,$FD,$6F,$D2,$3E,$6E,$9F,$B9,$71,$3F,$D3
	.dc.b	$5A,$FF,$8D,$DA,$FF,$98,$B7,$7C,$EF,$28,$FF,$57,$CF,$0D,$3F,$84
	.dc.b	$4F,$D3,$F7,$25,$7F,$13,$7F,$88,$FC,$0B,$99,$F1,$1F,$C0,$BE,$07
	.dc.b	$FD,$69,$7F,$C1,$9A,$FF,$8D,$5A,$FF,$91,$1B,$5C,$FD,$0E,$B7,$FE
	.dc.b	$28,$AF,$F5,$3B,$5F,$F5,$7C,$BE,$47,$EC,$73,$7F,$90,$7E,$3B,$FC
	.dc.b	$CF,$5F,$EC,$AF,$9D,$7D,$17,$F1,$66,$C7,$FB,$3A,$76,$FF,$A6,$5A
	.dc.b	$FF,$1E,$6B,$D7,$9F,$F9,$44,$62,$3F,$CB,$AF,$2E,$0F,$F6,$A3,$70
	.dc.b	$E1,$3E,$3F,$9C,$FC,$F7,$AD,$F9,$FF,$A4,$93,$FC,$FE,$32,$FF,$41
	.dc.b	$8C,$BF,$CF,$63,$F4,$DF,$49,$FA,$7A,$93,$FD,$1E,$32,$FF,$49,$8C
	.dc.b	$BF,$D1,$63,$F5,$DF,$49,$FA,$FD,$64,$FF,$4F,$8C,$BF,$D4,$63,$2F
	.dc.b	$F4,$D8,$CB,$FD,$3E,$33,$9F,$F4,$F9,$FE,$AA,$FF,$98,$3C,$2F,$E7
	.dc.b	$3E,$F4,$BF,$F8,$03,$39,$6F,$1A,$B8,$B6,$33,$1F,$C1,$7E,$93,$F8
	.dc.b	$37,$C0,$7F,$07,$FA,$49,$5F,$D9,$E3,$2F,$F6,$98,$CB,$FD,$96,$3F
	.dc.b	$AB,$FD,$27,$E8,$A3,$F0,$DC,$1F,$85,$33,$D7,$23,$FC,$71,$3B,$7F
	.dc.b	$D4,$4D,$7F,$B1,$B5,$FF,$49,$3B,$CD,$C5,$8A,$E9,$5F,$EE,$7E,$5F
	.dc.b	$F7,$23,$FD,$FF,$BB,$FB,$17,$27,$08,$FF,$54,$E1,$FE,$08,$8C,$47
	.dc.b	$F7,$F9,$FE,$50,$70,$C2,$47,$33,$8B,$FE,$A5,$F0,$1F,$49,$FD,$BF
	.dc.b	$E9,$3F,$D0,$95,$A3,$FD,$F2,$E1,$FF,$7E,$4F,$07,$47,$F1,$89,$FF
	.dc.b	$00,$C6,$63,$71,$D7,$D0,$FF,$02,$FD,$27,$FB,$06,$7B,$7A,$AF,$B4
	.dc.b	$FE,$67,$08,$C7,$95,$C3,$E1,$64,$FF,$9A,$6F,$93,$FE,$D3,$6B,$FE
	.dc.b	$14,$6B,$FE,$C3,$77,$BF,$C8,$6E,$B9,$AF,$F1,$5F,$80,$FA,$57,$7F
	.dc.b	$E4,$BE,$67,$FC,$23,$EF,$CC,$FF,$2F,$19,$69,$06,$7F,$5F,$D6,$FC
	.dc.b	$DF,$CA,$7F,$86,$BC,$59,$5F,$A0,$8C,$B4,$87,$2D,$7F,$D9,$7C,$FE
	.dc.b	$88,$FF,$75,$FC,$23,$E9,$3B,$C7,$F8,$9F,$E0,$3F,$DC,$5F,$4A,$FA
	.dc.b	$5F,$F7,$27,$C0,$7F,$90,$A3,$15,$FE,$F4,$8F,$F2,$2E,$1C,$F6,$C6
	.dc.b	$1F,$E9,$1C,$7A,$FD,$2F,$1C,$BF,$DD,$CD,$C7,$FD,$FA,$F2,$9D,$D7
	.dc.b	$1B,$CB,$E4,$BE,$2D,$BC,$9E,$CA,$E2,$DA,$CA,$CB,$D2,$B0,$F1,$36
	.dc.b	$FC,$0D,$39,$1F,$14,$1C,$F0,$F8,$B1,$E3,$8D,$0C,$7E,$23,$86,$D7
	.dc.b	$80,$93,$E0,$7E,$16,$1F,$2A,$BB,$E1,$E8,$CB,$CE,$6A,$3A,$65,$F0
	.dc.b	$1F,$21,$F0,$27,$DF,$7B,$26,$1E,$77,$C3,$3E,$E7,$2F,$83,$2F,$9E
	.dc.b	$D3,$E5,$E5,$EA,$AF,$8F,$8F,$91,$F8,$4F,$94,$E7,$3E,$43,$E6,$16
	.dc.b	$F3,$5C,$E3,$B1,$AC,$3D,$37,$F1,$4E,$E7,$62,$47,$D6,$BE,$F3,$C2
	.dc.b	$39,$DF,$6C,$7E,$34,$E6,$7C,$57,$D5,$E4,$F4,$7D,$17,$EB,$8D,$47
	.dc.b	$03,$AB,$8C,$3D,$51,$92,$38,$33,$5E,$13,$C4,$E6,$FF,$67,$7F,$2E
	.dc.b	$BE,$1B,$E1,$7E,$B5,$FD,$92,$38,$3D,$27,$07,$D5,$5F,$92,$9E,$53
	.dc.b	$EC,$9C,$D7,$81,$FD,$0F,$9D,$FE,$A7,$9F,$CD,$4F,$F0,$68,$C5,$FE
	.dc.b	$71,$B1,$9B,$FF,$5E,$D3,$E1,$FF,$4F,$5C,$E3,$F5,$31,$89,$DB,$63
	.dc.b	$FD,$8A,$D7,$F5,$E7,$33,$F5,$BF,$1F,$FE,$AE,$DF,$F4,$D1,$A9,$B3
	.dc.b	$98,$FD,$99,$E0,$38,$FC,$07,$F6,$5B,$FF,$6E,$C6,$5F,$F6,$FC,$F1
	.dc.b	$73,$1E,$1D,$73,$5F,$DB,$5C,$DF,$F0,$EB,$F0,$46,$2B,$BE,$1E,$67
	.dc.b	$B6,$9E,$CA,$EC,$FC,$28,$FD,$2C,$72,$8C,$57,$82,$BF,$8D,$D7,$AE
	.dc.b	$30,$FE,$B0,$C4,$F5,$C3,$CC,$F2,$31,$5C,$A7,$D9,$18,$79,$1F,$3F
	.dc.b	$1E,$A7,$AC,$B1,$CE,$3F,$91,$E1,$E7,$7E,$C9,$EA,$BF,$0B,$8C,$3C
	.dc.b	$E7,$C5,$2D,$7C,$1B,$EB,$3C,$9A,$FD,$D3,$A6,$1E,$9C,$7E,$86,$1D
	.dc.b	$A7,$58,$74,$7C,$18,$7B,$63,$D4,$FB,$F0,$B3,$EA,$AF,$C2,$4F,$B6
	.dc.b	$18,$FB,$5E,$B1,$CC,$D4,$66,$3E,$89,$CF,$1F,$33,$E8,$3E,$6E,$3E
	.dc.b	$88,$CE,$1E,$E9,$FC,$29,$C4,$FC,$C9,$F5,$8C,$63,$F2,$70,$F1,$47
	.dc.b	$C7,$7D,$87,$CB,$2B,$F0,$8E,$2F,$A3,$1F,$91,$1E,$83,$DE,$E1,$F0
	.dc.b	$18,$BF,$5D,$7C,$D8,$F8,$67,$43,$C4,$7A,$6F,$E6,$BC,$4F,$ED,$5B
	.dc.b	$BE,$31,$F4,$BF,$E5,$78,$91,$8C,$3F,$C7,$FC,$4E,$DD,$47,$F9,$72
	.dc.b	$75,$87,$F8,$7C,$FF,$14,$B8,$BF,$F3,$17,$73,$E1,$5E,$AF,$2B,$DE
	.dc.b	$76,$B5,$39,$BF,$37,$BB,$71,$3B,$D1,$EE,$E6,$26,$9F,$E5,$6D,$3F
	.dc.b	$C5,$4F,$FA,$62,$F8,$B9,$E8,$F1,$6B,$FE,$E3,$F5,$AF,$31,$F3,$78
	.dc.b	$2F,$A5,$E3,$5E,$F3,$86,$1F,$E6,$CB,$E0,$FF,$32,$8C,$9F,$CD,$23
	.dc.b	$2D,$BF,$F7,$5C,$63,$6F,$8F,$F7,$CF,$4B,$C7,$5F,$16,$31,$87,$F9
	.dc.b	$EB,$E9,$5F,$49,$EA,$6C,$7F,$9D,$BE,$99,$D1,$9A,$93,$9E,$EB,$AF
	.dc.b	$D9,$E7,$61,$FE,$55,$AF,$C0,$E7,$F3,$FE,$9B,$E8,$36,$32,$7E,$1A
	.dc.b	$32,$D3,$E2,$79,$DF,$30,$FC,$02,$FA,$47,$C9,$E6,$36,$7C,$9E,$25
	.dc.b	$7C,$62,$46,$5E,$75,$FD,$3F,$8B,$FD,$5B,$5C,$2B,$FD,$48,$77,$EF
	.dc.b	$FC,$47,$CD,$7E,$27,$DA,$B3,$F9,$E6,$EF,$F0,$FB,$7F,$DB,$93,$FA
	.dc.b	$2A,$E0,$FE,$2F,$E2,$7F,$28,$7E,$4F,$E7,$AC,$64,$FC,$97,$A9,$8C
	.dc.b	$3E,$77,$77,$FA,$EA,$7F,$5B,$F1,$7F,$9B,$E7,$FC,$E3,$63,$27,$EC
	.dc.b	$A3,$2D,$33,$5E,$78,$CB,$63,$F7,$3F,$4D,$FA,$7F,$A6,$FD,$43,$63
	.dc.b	$27,$EF,$A3,$2D,$3F,$80,$FA,$1F,$AF,$F8,$6F,$D8,$36,$32,$7F,$06
	.dc.b	$8C,$B4,$FF,$67,$FD,$33,$E9,$BF,$5E,$D8,$FA,$BC,$CF,$A6,$CD,$49
	.dc.b	$FD,$55,$6F,$3F,$F0,$EB,$5F,$F0,$A8,$70,$FF,$86,$1E,$1C,$FF,$EF
	.dc.b	$39,$FC,$85,$75,$AF,$E3,$11,$83,$FD,$D5,$F1,$2F,$A6,$FE,$2F,$18
	.dc.b	$6C,$FF,$1F,$9D,$7A,$FF,$C6,$DE,$1E,$67,$F0,$CF,$45,$D1,$FF,$00
	.dc.b	$2F,$FC,$4C,$D7,$FC,$66,$D7,$FC,$70,$E7,$C2,$75,$3D,$5F,$E4,$07
	.dc.b	$7C,$DF,$33,$8D,$7F,$9B,$FD,$37,$87,$BB,$F9,$33,$63,$FE,$00,$E9
	.dc.b	$F9,$CD,$3F,$C8,$35,$E1,$8C,$47,$F9,$0A,$3C,$59,$3E,$9B,$F9,$81
	.dc.b	$CE,$7F,$4C,$FF,$2A,$D3,$FA,$36,$1F,$EB,$6E,$23,$FE,$4A,$8C,$D7
	.dc.b	$F4,$7F,$5B,$FA,$1D,$F1,$F5,$25,$FF,$29,$BD,$DF,$DE,$6B,$CE,$7E
	.dc.b	$44,$74,$C3,$FA,$7F,$C7,$3C,$1E,$77,$FC,$76,$31,$87,$94,$F2,$7A
	.dc.b	$9E,$93,$FA,$94,$7F,$2A,$9F,$E7,$D8,$7F,$4E,$AE,$11,$FE,$D2,$8C
	.dc.b	$3F,$E7,$06,$BE,$7A,$F2,$92,$7F,$B5,$F8,$DD,$DA,$74,$97,$D2,$BE
	.dc.b	$93,$FA,$AB,$67,$C3,$F4,$AF,$A6,$FC,$6C,$71,$9D,$5F,$F5,$56,$BF
	.dc.b	$A6,$6B,$FE,$7C,$6B,$FE,$EC,$6B,$E2,$BF,$EE,$DC,$0F,$A6,$7B,$FE
	.dc.b	$49,$CC,$63,$0F,$57,$D2,$BE,$3B,$FB,$6B,$63,$E4,$E0,$FA,$67,$C9
	.dc.b	$6A,$E5,$7F,$69,$BE,$19,$FF,$A7,$1A,$FF,$00,$6B,$FE,$96,$77,$C9
	.dc.b	$FE,$1B,$FA,$AF,$CC,$57,$92,$FF,$C1,$91,$8B,$FE,$F5,$F1,$0F,$A7
	.dc.b	$FF,$01,$B6,$FE,$A7,$C0,$7F,$56,$7E,$B4,$66,$B8,$D7,$EE,$63,$11
	.dc.b	$FC,$51,$AF,$FA,$A9,$A8,$3C,$F1,$FF,$13,$FA,$1F,$DC,$67,$AD,$FF
	.dc.b	$86,$FD,$9F,$2F,$C0,$C8,$F2,$FD,$25,$AF,$5C,$F8,$C9,$3F,$E1,$3C
	.dc.b	$7F,$C3,$33,$FB,$3E,$F7,$CA,$FF,$8C,$72,$7B,$1F,$E2,$D8,$C3,$70
	.dc.b	$FF,$16,$36,$FF,$AD,$BF,$E2,$AF,$A6,$FF,$23,$7B,$5F,$EC,$EF,$37
	.dc.b	$FD,$93,$18,$AF,$F6,$A4,$7F,$B3,$3E,$9D,$F4,$1F,$C1,$A7,$27,$C5
	.dc.b	$AE,$1E,$42,$F9,$F8,$DF,$E0,$8A,$D8,$AD,$CF,$5B,$E9,$7C,$6B,$A5
	.dc.b	$F1,$8C,$47,$18,$C5,$76,$C7,$D5,$3F,$5F,$1F,$3E,$78,$8E,$13,$72
	.dc.b	$F0,$C7,$68,$D5,$70,$8F,$3C,$E9,$E2,$63,$0F,$0B,$DF,$2F,$4E,$3E
	.dc.b	$38,$D6,$1D,$EB,$70,$C3,$B6,$5E,$CC,$7B,$12,$B7,$3A,$C3,$A5,$6E
	.dc.b	$7E,$65,$FF,$6C,$BF,$B0,$D8,$CB,$E3,$3F,$B8,$FC,$57,$FA,$2B,$E8
	.dc.b	$FF,$CC,$0D,$8F,$EE,$A1,$C3,$C3,$E2,$53,$FE,$67,$BF,$F3,$DE,$D7
	.dc.b	$C6,$FC,$9D,$F2,$9C,$B6,$32,$F9,$DD,$E5,$F0,$5B,$9F,$E4,$9B,$3F
	.dc.b	$B1,$FA,$7F,$ED,$2F,$D9,$8C,$BF,$8B,$6D,$F8,$2F,$F0,$F5,$C2,$BE
	.dc.b	$DD,$68,$98,$7E,$16,$75,$5E,$38,$EC,$4A,$FC,$CC,$EB,$0F,$34,$E6
	.dc.b	$B8,$56,$BC,$EF,$E1,$0F,$82,$75,$5C,$23,$46,$0F,$D5,$61,$E4,$72
	.dc.b	$7E,$BB,$C9,$7C,$FB,$E3,$7A,$F6,$7F,$AE,$21,$D7,$8E,$6F,$EC,$11
	.dc.b	$8C,$3F,$B1,$E1,$CB,$0D,$42,$30,$EA,$78,$3F,$E8,$F6,$E1,$DE,$B8
	.dc.b	$CB,$E9,$5F,$4F,$D3,$8B,$FB,$E7,$CD,$5F,$40,$F0,$AD,$DF,$F6,$B6
	.dc.b	$BF,$ED,$5A,$E2,$FF,$7A,$6B,$FE,$BB,$8D,$E3,$FC,$4E,$3C,$5F,$2F
	.dc.b	$EC,$D3,$FD,$B5,$F1,$57,$BB,$FD,$5B,$5D,$EB,$C4,$66,$64,$64,$87
	.dc.b	$FA,$09,$CB,$E1,$8F,$83,$F4,$AF,$85,$F8,$7F,$4D,$FD,$9A,$BF,$C0
	.dc.b	$CD,$7F,$DA,$53,$B7,$FC,$14,$D7,$FD,$AD,$1B,$CB,$D8,$DE,$2F,$C5
	.dc.b	$D6,$E7,$E0,$3C,$6B,$FC,$DA,$D7,$77,$FE,$DD,$51,$97,$FA,$E7,$6B
	.dc.b	$E2,$3E,$DF,$B1,$F3,$31,$FC,$CB,$6B,$73,$F8,$0A,$E1,$1F,$E2,$39
	.dc.b	$D2,$6C,$6A,$3F,$DE,$4D,$C7,$EC,$36,$FF,$4E,$D9,$FE,$E1,$5F,$D4
	.dc.b	$1F,$F1,$87,$D2,$A7,$EF,$73,$3E,$9A,$9F,$94,$9D,$57,$0A,$CB,$F9
	.dc.b	$D9,$D4,$48,$D1,$0F,$F7,$77,$91,$FD,$43,$2F,$C7,$5B,$AB,$F1,$D3
	.dc.b	$FE,$2E,$AF,$D6,$B5,$57,$09,$F2,$1C,$3D,$49,$3E,$C8,$CB,$48,$7F
	.dc.b	$9C,$6F,$FD,$DA,$FE,$D7,$E9,$BE,$09,$1F,$F2,$A7,$D3,$3E,$7F,$F6
	.dc.b	$AD,$8E,$04,$3C,$BB,$FF,$16,$8D,$47,$EF,$FA,$B8,$D4,$8F,$E0,$91
	.dc.b	$F4,$3D,$C9,$1F,$C1,$DF,$E1,$7F,$07,$FE,$14,$F8,$8F,$E0,$BF,$09
	.dc.b	$FB,$5D,$FF,$C1,$13,$9B,$FF,$5C,$F8,$8B,$DE,$9C,$A3,$FC,$3F,$3A
	.dc.b	$F3,$B0,$76,$F9,$6F,$A9,$5F,$CB,$BE,$47,$E8,$65,$FC,$0F,$E8,$1F
	.dc.b	$4B,$FC,$59,$B8,$FF,$9E,$6B,$77,$C7,$97,$57,$2B,$51,$FC,$7A,$B7
	.dc.b	$1F,$E4,$3B,$E6,$FC,$A3,$15,$BA,$FC,$AF,$A7,$F9,$6F,$A3,$95,$FC
	.dc.b	$B6,$32,$FF,$2F,$8C,$BF,$CB,$E3,$F3,$BF,$4D,$F9,$EF,$9F,$93,$FC
	.dc.b	$D6,$32,$FF,$37,$8C,$BF,$CD,$E3,$F4,$BF,$4D,$FA,$6C,$64,$FF,$3D
	.dc.b	$8C,$BF,$CF,$E3,$2F,$F3,$F8,$CB,$FD,$06,$33,$AF,$F4,$F9,$FE,$92
	.dc.b	$FF,$95,$3B,$FF,$97,$7B,$F2,$FF,$D6,$33,$96,$F0,$FF,$05,$7F,$A9
	.dc.b	$F8,$FF,$C7,$5F,$EA,$4D,$9E,$11,$FD,$47,$DE,$FD,$FF,$D3,$48,$FE
	.dc.b	$AD,$1F,$C0,$8C,$3F,$D5,$E3,$0F,$FA,$DF,$E9,$BF,$A8,$BF,$EB,$9E
	.dc.b	$27,$0F,$E1,$4F,$EB,$73,$A9,$FB,$81,$C6,$1F,$8D,$9E,$13,$D7,$AF
	.dc.b	$89,$87,$8D,$F6,$74,$7F,$6D,$EC,$FE,$AD,$C7,$FA,$5E,$6F,$F6,$B3
	.dc.b	$67,$FB,$A7,$89,$FB,$4C,$BF,$BF,$5F,$0B,$95,$CC,$FF,$6A,$7C,$4B
	.dc.b	$E5,$FF,$B1,$7D,$37,$F2,$99,$FC,$B7,$9C,$8C,$63,$FD,$A6,$F8,$57
	.dc.b	$2B,$E3,$EF,$48,$C1,$FE,$E1,$FA,$5F,$F5,$1B,$FE,$0C,$FA,$6F,$E5
	.dc.b	$EF,$66,$E1,$FD,$D6,$77,$3F,$31,$B1,$FE,$8B,$C9,$5C,$33,$FF,$7C
	.dc.b	$36,$FF,$D6,$57,$C2,$7F,$C5,$AD,$8F,$F0,$D4,$EE,$B9,$F9,$5F,$E1
	.dc.b	$6A,$D6,$1F,$EA,$AF,$97,$FF,$02,$7B,$33,$1F,$CA,$C7,$E7,$3A,$4E
	.dc.b	$63,$E4,$F4,$FF,$10,$7F,$9E,$CE,$9F,$CE,$FC,$97,$F8,$3F,$C1,$91
	.dc.b	$82,$1C,$9C,$BF,$C2,$71,$FF,$5E,$CF,$F8,$B3,$E0,$1F,$09,$FD,$2A
	.dc.b	$9F,$47,$FE,$41,$7F,$DA,$9F,$48,$FA,$4F,$F2,$7C,$E6,$53,$97,$FD
	.dc.b	$B5,$19,$BE,$B8,$E3,$DE,$FF,$2B,$C6,$2B,$BC,$7F,$96,$AF,$94,$60
	.dc.b	$F0,$E1,$E0,$9E,$38,$F9,$F1,$EC,$D8,$E5,$3B,$AE,$27,$84,$FF,$77
	.dc.b	$61,$E1,$BE,$38,$E6,$7D,$75,$C5,$A6,$6B,$85,$F8,$AB,$86,$3E,$8C
	.dc.b	$3E,$23,$67,$57,$EB,$3F,$D4,$18,$F9,$DF,$29,$99,$E0,$F9,$A3,$4F
	.dc.b	$0B,$EF,$8F,$96,$B7,$39,$3B,$63,$E6,$7A,$E3,$E4,$AF,$64,$E6,$78
	.dc.b	$4E,$5E,$A7,$BF,$0F,$0C,$EE,$FB,$9E,$07,$91,$9B,$F2,$1D,$A7,$89
	.dc.b	$EA,$9E,$31,$9C,$3B,$1E,$0B,$F8,$51,$97,$D7,$5C,$63,$AC,$70,$33
	.dc.b	$3C,$70,$F5,$46,$08,$E0,$98,$FD,$47,$F0,$31,$E0,$39,$46,$48,$FC
	.dc.b	$A7,$F1,$AF,$86,$38,$9D,$1D,$1F,$66,$75,$3C,$23,$B1,$E4,$7A,$BF
	.dc.b	$8E,$33,$3D,$23,$38,$7C,$89,$E6,$75,$37,$1D,$67,$84,$65,$F2,$3D
	.dc.b	$5C,$BF,$87,$F2,$BF,$D2,$ED,$F6,$BF,$2F,$18,$CF,$F6,$CE,$A4,$3E
	.dc.b	$06,$C6,$63,$01,$C4,$6A,$BF,$E0,$0B,$67,$FD,$B2,$70,$BF,$E1,$9F
	.dc.b	$1A,$F9,$E7,$4F,$ED,$6B,$C3,$7F,$B7,$AD,$4E,$67,$96,$7F,$EE,$49
	.dc.b	$CC,$7E,$C4,$E2,$7E,$CD,$CE,$1F,$C6,$E3,$2F,$8B,$0F,$0C,$71,$3A
	.dc.b	$9F,$0A,$78,$D7,$B2,$3A,$4F,$0A,$E9,$1C,$05,$70,$7E,$0E,$1E,$BA
	.dc.b	$E1,$7C,$EF,$C5,$87,$B6,$B4,$DB,$EE,$D9,$DC,$79,$E7,$6D,$36,$58
	.dc.b	$D9,$E6,$8D,$B6,$36,$D3,$75,$E5,$3E,$75,$FA,$27,$D6,$70,$8D,$3F
	.dc.b	$1E,$34,$FE,$28,$F8,$98,$7C,$37,$17,$E2,$AE,$31,$F2,$2B,$8B,$F5
	.dc.b	$A3,$C3,$87,$C1,$6C,$7B,$5B,$87,$A5,$A7,$4C,$3C,$13,$CC,$F3,$57
	.dc.b	$36,$9C,$CB,$5C,$DB,$5C,$EF,$CB,$1E,$C6,$E1,$EB,$8F,$8B,$5F,$29
	.dc.b	$B5,$E4,$69,$D0,$A7,$C1,$BF,$23,$4F,$21,$67,$C8,$D9,$F2,$4F,$0C
	.dc.b	$3D,$F3,$E5,$8C,$61,$F6,$75,$7C,$5B,$E1,$FF,$43,$1F,$E2,$9E,$57
	.dc.b	$A7,$8C,$3F,$81,$CF,$0A,$C6,$3F,$E6,$EA,$B3,$9A,$FF,$39,$C6,$30
	.dc.b	$ED,$18,$3F,$CF,$6B,$D9,$E7,$D8,$CB,$FC,$25,$5A,$9C,$C6,$31,$FF
	.dc.b	$0F,$5F,$1A,$FF,$40,$F2,$77,$BD,$BE,$6A,$FE,$2D,$E8,$F6,$FA,$27
	.dc.b	$D1,$E7,$95,$39,$8C,$37,$1F,$82,$7E,$6E,$76,$E2,$DA,$30,$EB,$39
	.dc.b	$9F,$0B,$B9,$D6,$1D,$A3,$15,$C2,$BF,$D0,$FF,$46,$FA,$1F,$8B,$F4
	.dc.b	$7F,$E3,$AB,$E9,$39,$7A,$46,$6B,$8D,$70,$EB,$F8,$57,$B9,$FF,$3C
	.dc.b	$D6,$F8,$BF,$CD,$1E,$57,$B2,$7F,$CF,$AB,$E8,$FE,$AB,$99,$DC,$63
	.dc.b	$7C,$B7,$3F,$75,$71,$7E,$16,$EF,$23,$F1,$30,$FA,$1F,$E7,$38,$FF
	.dc.b	$32,$CB,$FA,$65,$6A,$73,$5F,$46,$31,$D1,$F6,$5B,$1D,$9B,$5F,$EB
	.dc.b	$0C,$3C,$13,$F5,$CF,$05,$7B,$3E,$7B,$DB,$EA,$D9,$F4,$B6,$FE,$1B
	.dc.b	$72,$F8,$EF,$E7,$7E,$0F,$F2,$BE,$65,$BF,$F6,$9B,$6F,$FC,$FE,$DB
	.dc.b	$FD,$73,$F3,$3E,$17,$E6,$F2,$5B,$F9,$2D,$BF,$A6,$DB,$FA,$ED,$9F
	.dc.b	$F5,$FB,$67,$F2,$7B,$FD,$FA,$FB,$BF,$47,$F6,$BE,$8F,$ED,$B7,$0F
	.dc.b	$AE,$DB,$FC,$0B,$6B,$F0,$BE,$57,$E0,$BD,$FF,$C1,$B6,$FE,$FB,$6F
	.dc.b	$F1,$2D,$AF,$B9,$F4,$4F,$A2,$FB,$87,$ED,$5C,$5F,$D5,$E5,$B8,$7F
	.dc.b	$BA,$26,$DF,$F7,$6A,$FC,$45,$F1,$39,$BF,$AF,$9D,$BF,$E1,$A3,$66
	.dc.b	$63,$74,$AF,$F0,$0F,$C1,$3E,$8B,$F9,$0D,$FE,$F6,$31,$E2,$7F,$20
	.dc.b	$75,$E8,$5A,$DE,$1F,$EF,$EF,$29,$CF,$9F,$92,$FE,$25,$18,$CB,$A5
	.dc.b	$A7,$F5,$BE,$7C,$9C,$C5,$C3,$F6,$7F,$45,$CB,$AB,$F6,$6D,$AF,$E1
	.dc.b	$D8,$A7,$F6,$FF,$45,$9B,$FD,$DB,$8A,$FD,$AB,$9D,$FF,$80,$BF,$C9
	.dc.b	$FC,$67,$AB,$4F,$E7,$FE,$6F,$F4,$49,$C5,$FF,$1D,$AD,$BF,$C7,$7E
	.dc.b	$1D,$F2,$5A,$CB,$E7,$4F,$F9,$32,$95,$A8,$E5,$8F,$F4,$28,$C5,$70
	.dc.b	$38,$46,$6F,$CD,$F1,$92,$FC,$1D,$D6,$F8,$1F,$CD,$27,$F8,$CF,$89
	.dc.b	$9D,$7F,$8F,$B6,$FC,$9D,$FF,$C1,$AF,$8E,$8C,$38,$D7,$F3,$BF,$A0
	.dc.b	$7D,$07,$F4,$D6,$C7,$F4,$3F,$A2,$7D,$17,$DF,$C7,$C7,$5A,$6D,$FC
	.dc.b	$CF,$A2,$7B,$7C,$E3,$0D,$C7,$FA,$77,$22,$B7,$1E,$3A,$DB,$FD,$3F
	.dc.b	$E8,$5F,$13,$FC,$AF,$1F,$F1,$94,$62,$7F,$CE,$71,$FE,$7E,$8C,$38
	.dc.b	$6B,$FE,$78,$73,$C3,$FE,$06,$FA,$17,$3F,$F6,$D9,$CD,$FF,$81,$A7
	.dc.b	$2F,$8F,$0F,$F0,$2C,$63,$0F,$AB,$5C,$1F,$D9,$46,$70,$FE,$85,$F4
	.dc.b	$0F,$A1,$FF,$00,$FA,$9F,$DE,$E3,$19,$7F,$6B,$F6,$BF,$AD,$3F,$E4
	.dc.b	$F8,$CE,$3F,$89,$E5,$FF,$06,$5F,$1C,$39,$DF,$F7,$5B,$DE,$3F,$E2
	.dc.b	$1A,$D4,$F5,$7C,$1E,$CF,$F8,$83,$E6,$FF,$7B,$F2,$F8,$AF,$F0,$14
	.dc.b	$61,$E9,$3B,$F1,$55,$BD,$3F,$39,$7B,$7F,$B7,$FC,$1F,$F8,$7B,$D5
	.dc.b	$FE,$DE,$DC,$7F,$B5,$B6,$77,$1F,$06,$7F,$C3,$DF,$3F,$FB,$8F,$A1
	.dc.b	$FF,$54,$F9,$5F,$E7,$18,$C5,$FF,$AE,$E3,$FC,$E9,$F4,$2F,$89,$FF
	.dc.b	$50,$F9,$09,$E1,$E3,$31,$E7,$87,$FB,$13,$1E,$77,$B8,$C5,$6D,$EC
	.dc.b	$7F,$B2,$B1,$ED,$9F,$28,$5F,$2C,$7B,$E3,$98,$C4,$F1,$CB,$CB,$5A
	.dc.b	$8E,$A7,$2C,$7C,$19,$78,$E3,$18,$77,$CB,$C3,$1C,$31,$EA,$E8,$C5
	.dc.b	$F2,$C3,$E2,$E3,$E0,$9F,$6E,$1E,$5A,$D4,$60,$F1,$E1,$DB,$1F,$16
	.dc.b	$3D,$89,$5C,$29,$5B,$C7,$D9,$7C,$2B,$E5,$57,$FB,$A2,$FE,$CB,$63
	.dc.b	$2F,$C3,$3F,$DD,$BF,$0D,$FE,$7B,$DB,$F7,$B0,$8C,$3F,$C0,$1A,$F0
	.dc.b	$F8,$34,$F0,$BF,$E4,$1B,$5F,$13,$FD,$FD,$7F,$C2,$27,$2D,$8C,$BE
	.dc.b	$77,$79,$7C,$26,$E7,$F9,$36,$C8,$7D,$0F,$F6,$37,$F0,$51,$A7,$F0
	.dc.b	$ED,$BF,$0D,$FD,$FA,$E1,$5F,$76,$B4,$4C,$3F,$21,$3A,$AF,$24,$76
	.dc.b	$25,$7E,$32,$75,$87,$9E,$73,$5C,$2B,$5E,$6F,$FB,$6A,$B5,$1E,$1A
	.dc.b	$DC,$7F,$B8,$6F,$9C,$FE,$A7,$0F,$23,$B3,$99,$FA,$EE,$AF,$F5,$16
	.dc.b	$9F,$EE,$9E,$8F,$F5,$A7,$55,$E4,$FE,$D5,$D7,$FC,$97,$0F,$E4,$57
	.dc.b	$97,$FD,$27,$1F,$EB,$D7,$F9,$35,$67,$0F,$F5,$73,$67,$53,$FD,$A3
	.dc.b	$E7,$DE,$E7,$38,$46,$1F,$F4,$E3,$5F,$E3,$7F,$4F,$F4,$B3,$FF,$3D
	.dc.b	$4E,$5B,$5C,$E3,$19,$FE,$07,$8D,$7C,$63,$F0,$98,$78,$3E,$87,$81
	.dc.b	$1F,$E5,$F2,$8C,$47,$A5,$B8,$AB,$B5,$6E,$BD,$35,$A9,$EF,$86,$67
	.dc.b	$D7,$3F,$07,$EF,$BF,$D8,$F1,$FE,$BB,$6B,$FE,$0E,$6E,$1F,$E9,$9F
	.dc.b	$9E,$7D,$2F,$1F,$88,$97,$C5,$F9,$93,$A8,$C1,$0E,$91,$A9,$FA,$5F
	.dc.b	$42,$F8,$2F,$DE,$D2,$30,$FF,$00,$6B,$FD,$F9,$AF,$CC,$38,$47,$42
	.dc.b	$3F,$ED,$7F,$A2,$FD,$DE,$3F,$EB,$3A,$E6,$DA,$DD,$7F,$8B,$AB,$7A
	.dc.b	$FD,$8F,$A2,$78,$FF,$8D,$F1,$A5,$FF,$B8,$E7,$2D,$9F,$C7,$46,$23
	.dc.b	$FC,$72,$DC,$7E,$D4,$7E,$73,$CC,$DD,$48,$E4,$7F,$80,$E3,$9F,$83
	.dc.b	$F9,$FF,$96,$FE,$E3,$3F,$EF,$16,$BD,$A3,$6F,$95,$AF,$F8,$27,$1F
	.dc.b	$E9,$98,$7F,$B5,$71,$FF,$32,$CE,$C8,$6F,$B7,$F5,$9D,$F2,$3F,$C2
	.dc.b	$CF,$EA,$FE,$89,$F3,$94,$93,$FC,$8F,$0F,$F7,$14,$EA,$3F,$C3,$DF
	.dc.b	$63,$FB,$73,$6F,$F9,$7A,$34,$FF,$87,$FB,$13,$FB,$DF,$B1,$93,$FE
	.dc.b	$27,$7E,$27,$3F,$8F,$E6,$A4,$F8,$78,$FF,$63,$8A,$7E,$57,$CE,$4A
	.dc.b	$F9,$4D,$7C,$1E,$5F,$F0,$7F,$C3,$BF,$D0,$31,$CC,$FE,$7C,$E1,$FF
	.dc.b	$3A,$B7,$5F,$BB,$E1,$B6,$FE,$0B,$5C,$3D,$EE,$84,$7F,$C8,$1E,$1D
	.dc.b	$79,$C5,$BE,$75,$CA,$FA,$78,$1F,$C7,$FE,$9A,$61,$FE,$49,$8F,$E4
	.dc.b	$46,$1F,$95,$18,$7F,$C9,$DF,$63,$FC,$A7,$ED,$7F,$55,$8F,$A1,$BC
	.dc.b	$3F,$CD,$27,$F9,$61,$BA,$87,$23,$E2,$62,$AF,$EF,$F7,$CF,$E4,$7B
	.dc.b	$1D,$FD,$19,$7C,$29,$3A,$8C,$4E,$9F,$F3,$04,$E9,$E1,$39,$AF,$D5
	.dc.b	$FB,$6F,$39,$F1,$FF,$D2,$63,$0D,$8C,$DF,$87,$C6,$C4,$C8,$FF,$4A
	.dc.b	$CE,$9F,$0C,$FC,$7F,$8B,$F9,$1F,$6B,$27,$D5,$19,$7D,$91,$97,$D3
	.dc.b	$1F,$D4,$7E,$53,$FD,$4B,$1F,$8B,$0E,$1F,$F5,$FB,$72,$FE,$0F,$0F
	.dc.b	$A1,$7D,$47,$3F,$82,$93,$CE,$F8,$71,$FF,$50,$BD,$CE,$6B,$6F,$67
	.dc.b	$FB,$24,$6A,$3F,$D1,$DE,$3B,$EA,$3F,$88,$63,$27,$F8,$84,$7F,$13
	.dc.b	$30,$F0,$8C,$3F,$A7,$C9,$ED,$FF,$BE,$E3,$0D,$7F,$DC,$91,$AF,$8B
	.dc.b	$FE,$3D,$E8,$7F,$80,$63,$13,$C2,$71,$5C,$67,$F5,$F8,$FF,$58,$D3
	.dc.b	$FB,$7F,$D2,$BC,$1F,$F0,$03,$71,$FA,$0D,$7F,$DD,$BA,$FF,$BE,$9C
	.dc.b	$CF,$02,$3F,$5E,$34,$FF,$BD,$27,$96,$1F,$59,$B7,$E2,$BF,$9B,$F1
	.dc.b	$9F,$E1,$1F,$8B,$7C,$BD,$26,$3F,$AA,$FA,$47,$7E,$23,$FE,$00,$D7
	.dc.b	$07,$F1,$0D,$7F,$D8,$93,$C1,$F9,$BF,$2D,$F1,$88,$FE,$DB,$E3,$9F
	.dc.b	$5B,$88,$FA,$3C,$2A,$C6,$FF,$C6,$AF,$75,$F9,$7F,$83,$7C,$3A,$BF
	.dc.b	$CC,$1F,$42,$FA,$1F,$F0,$FF,$D1,$BE,$87,$FC,$43,$39,$C7,$FC,$CD
	.dc.b	$39,$78,$D7,$F8,$9B,$1E,$B8,$62,$7F,$CD,$91,$85,$7C,$D5,$F3,$73
	.dc.b	$7E,$2C,$3C,$2F,$5C,$7D,$13,$A3,$AB,$8B,$E9,$7C,$DB,$7C,$AB,$78
	.dc.b	$F9,$5F,$F4,$A5,$F1,$C3,$C1,$19,$C3,$94,$78,$6B,$86,$3E,$BC,$7A
	.dc.b	$DE,$F1,$EC,$F4,$AD,$E3,$D6,$3E,$4C,$EB,$1E,$B1,$E4,$3D,$87,$12
	.dc.b	$BA,$7E,$3C,$76,$C3,$C5,$8F,$96,$77,$3D,$4D,$5F,$86,$3A,$63,$E8
	.dc.b	$76,$79,$B1,$8E,$09,$1F,$5F,$0E,$F1,$E7,$8E,$44,$C3,$C4,$EE,$FC
	.dc.b	$18,$FA,$4F,$55,$F2,$BF,$1D,$FD,$07,$BB,$C8,$8F,$C8,$3E,$71,$E6
	.dc.b	$BE,$D8,$62,$38,$9E,$59,$E2,$FC,$83,$53,$C1,$F9,$F5,$E3,$BF,$36
	.dc.b	$3F,$23,$0E,$D8,$79,$70,$F7,$61,$D4,$87,$53,$C2,$7B,$0C,$DF,$47
	.dc.b	$99,$8B,$FA,$CE,$0E,$07,$39,$EE,$E3,$0F,$34,$F0,$8F,$4E,$3F,$76
	.dc.b	$73,$1E,$EC,$3E,$0C,$7E,$C7,$C1,$7D,$1F,$E1,$DB,$5F,$91,$6B,$F9
	.dc.b	$16,$BF,$EE,$5F,$88,$FF,$73,$46,$21,$18,$9F,$F3,$E4,$60,$FE,$F8
	.dc.b	$E2,$35,$1F,$E8,$17,$F6,$5E,$2F,$F8,$2B,$A3,$F6,$93,$DC,$FD,$1C
	.dc.b	$65,$FD,$29,$9D,$3F,$85,$46,$5F,$E1,$91,$9C,$BF,$D7,$33,$97,$C5
	.dc.b	$1D,$3C,$6F,$F3,$DD,$70,$8F,$43,$D2,$33,$3C,$23,$17,$E6,$AD,$CF
	.dc.b	$6A,$DC,$7B,$4E,$D3,$BC,$3D,$F3,$A9,$FF,$66,$C6,$27,$51,$C0,$F7
	.dc.b	$9D,$80,$C3,$D8,$F7,$C3,$CA,$F0,$AE,$8D,$3A,$47,$BE,$32,$D8,$CE
	.dc.b	$1E,$F3,$D9,$7E,$49,$EA,$6C,$B1,$B9,$F2,$DF,$91,$F6,$9D,$0F,$34
	.dc.b	$74,$7C,$93,$C9,$C9,$CB,$0F,$8C,$FD,$B3,$17,$E3,$AE,$55,$E3,$69
	.dc.b	$E3,$9F,$8D,$5C,$9F,$94,$62,$B9,$3D,$0F,$1C,$F2,$7C,$66,$CC,$9B
	.dc.b	$AF,$03,$63,$C1,$7F,$06,$3D,$07,$19,$DB,$67,$77,$E5,$C3,$E9,$C6
	.dc.b	$30,$F2,$D6,$CF,$1E,$1D,$5F,$5B,$6B,$8C,$6A,$FA,$3F,$47,$77,$D1
	.dc.b	$BD,$BF,$F6,$DC,$48,$C6,$3F,$E0,$FF,$88,$CB,$FE,$C3,$8C,$DF,$F9
	.dc.b	$01,$B5,$C6,$F7,$2F,$31,$C8,$C7,$37,$E2,$BD,$DB,$E8,$9F,$2F,$86
	.dc.b	$E7,$FE,$96,$FA,$2F,$E3,$2F,$F9,$C6,$33,$5F,$E1,$B8,$FD,$2B,$CE
	.dc.b	$B9,$1E,$4C,$D7,$CA,$78,$1F,$AC,$BE,$78,$FF,$9E,$38,$DD,$9E,$66
	.dc.b	$E5,$FE,$CB,$1F,$3E,$F4,$35,$52,$7F,$D0,$D5,$C3,$1E,$F1,$C2,$B5
	.dc.b	$1E,$23,$FD,$45,$58,$C1,$5F,$C0,$0E,$07,$48,$F9,$7C,$0F,$56,$9E
	.dc.b	$FE,$17,$AB,$8B,$FF,$3E,$56,$DB,$5B,$9D,$63,$F1,$37,$FB,$3E,$8F
	.dc.b	$3F,$A1,$E8,$DC,$3F,$CA,$8D,$AE,$19,$7F,$AE,$A7,$FC,$B1,$5F,$7B
	.dc.b	$E8,$1E,$8D,$3F,$0F,$3F,$2B,$D3,$C3,$FE,$AD,$9D,$B6,$77,$C4,$AD
	.dc.b	$E3,$F2,$6F,$6D,$AD,$CE,$B8,$3F,$DC,$DD,$7F,$83,$F5,$1F,$57,$63
	.dc.b	$0D,$CF,$FD,$BC,$DC,$7F,$DB,$6F,$E7,$BB,$5F,$56,$F8,$3C,$36,$FF
	.dc.b	$DC,$DF,$29,$89,$91,$FC,$EB,$3F,$F7,$0D,$6F,$1F,$F7,$3D,$E9,$FF
	.dc.b	$74,$FC,$07,$9F,$CF,$F4,$37,$0F,$F7,$93,$67,$71,$FE,$F5,$9D,$B4
	.dc.b	$D9,$63,$FD,$EB,$E7,$FC,$3E,$4F,$88,$D9,$FF,$7F,$36,$77,$1F,$F0
	.dc.b	$06,$7E,$57,$D4,$7C,$BF,$5E,$CF,$B4,$FE,$AF,$D0,$E2,$FF,$6A,$46
	.dc.b	$70,$FE,$0D,$0B,$E5,$39,$AE,$11,$8A,$FB,$DF,$09,$F8,$B3,$ED,$7B
	.dc.b	$16,$FF,$11,$16,$7F,$35,$87,$FB,$C2,$72,$48,$E6,$F7,$8D,$46,$0F
	.dc.b	$CD,$FC,$FF,$EA,$CF,$C4,$7D,$55,$AF,$54,$FF,$88,$BD,$9F,$E3,$7E
	.dc.b	$51,$D4,$7E,$F6,$31,$3D,$FE,$AF,$F2,$47,$ED,$FE,$76,$C7,$8F,$DC
	.dc.b	$E7,$E1,$7E,$5D,$B1,$E0,$BF,$DE,$B6,$74,$D3,$D9,$F3,$BF,$9F,$38
	.dc.b	$E3,$F1,$1F,$E6,$3C,$3F,$CD,$67,$D6,$D9,$FE,$05,$C9,$FC,$92,$7F
	.dc.b	$96,$CE,$7A,$7F,$8A,$D7,$DE,$F3,$79,$9F,$CB,$EF,$84,$66,$FA,$CF
	.dc.b	$F1,$5F,$1F,$71,$FE,$DD,$7C,$7E,$FC,$8F,$A9,$C9,$6F,$F9,$36,$4A
	.dc.b	$C6,$5E,$28,$FC,$75,$62,$FC,$F5,$FB,$FF,$86,$C6,$1F,$90,$7F,$A8
	.dc.b	$FC,$EB,$5A,$7E,$F7,$EB,$1F,$59,$8C,$7E,$FD,$6D,$B5,$B9,$D6,$1F
	.dc.b	$CE,$EB,$4F,$F9,$91,$AF,$F5,$60,$E1,$FE,$B2,$D7,$FC,$7F,$F6,$3F
	.dc.b	$29,$3F,$33,$C3,$62,$FA,$8F,$ED,$73,$B2,$7A,$7E,$B8,$CB,$33,$FC
	.dc.b	$CC,$F3,$AE,$51,$89,$FE,$EB,$F2,$7F,$DF,$A3,$FB,$6B,$71,$FF,$29
	.dc.b	$FA,$1F,$CA,$A7,$F4,$73,$99,$34,$C4,$EE,$3F,$B0,$46,$7D,$5E,$E7
	.dc.b	$F8,$12,$75,$C7,$F8,$EA,$E7,$3E,$4F,$93,$98,$7E,$CA,$FF,$C1,$EF
	.dc.b	$8A,$33,$5D,$9E,$9E,$26,$30,$FC,$47,$DA,$7F,$6E,$BF,$E2,$0B,$E0
	.dc.b	$7F,$C1,$0E,$70,$FE,$C8,$E6,$76,$6A,$76,$D3,$F8,$2F,$CF,$BE,$57
	.dc.b	$FC,$0D,$7F,$EA,$8F,$37,$FC,$2F,$39,$F5,$FF,$D5,$D1,$FE,$1A,$FA
	.dc.b	$07,$C8,$FE,$BB,$C2,$FE,$9D,$87,$FA,$0B,$0E,$0F,$33,$18,$F6,$9D
	.dc.b	$C2,$77,$18,$BE,$B9,$78,$2B,$8E,$55,$E1,$87,$3C,$BC,$ED,$8F,$F4
	.dc.b	$9D,$F2,$7C,$47,$0A,$E5,$8F,$82,$F8,$C7,$5A,$E1,$8F,$9A,$B5,$7E
	.dc.b	$23,$9B,$85,$3C,$2F,$9C,$6B,$0F,$23,$E1,$3C,$73,$B7,$A5,$F7,$9D
	.dc.b	$DF,$5A,$D5,$71,$AD,$61,$FE,$DC,$8E,$87,$EE,$6B,$E8,$B6,$FE,$EB
	.dc.b	$FB,$DF,$80,$7C,$FB,$E7,$B9,$46,$1B,$19,$9F,$F7,$4C,$E5,$FF,$75
	.dc.b	$DF,$F9,$A7,$DD,$F1,$36,$33,$E2,$27,$2D,$8C,$C7,$D2,$8C,$67,$F3
	.dc.b	$DB,$AF,$F9,$A3,$E8,$1F,$31,$FE,$D4,$8F,$CA,$CE,$B0,$FC,$83,$5F
	.dc.b	$C6,$36,$B8,$C7,$7C,$23,$F6,$A7,$57,$D6,$39,$CE,$B0,$F2,$3F,$91
	.dc.b	$BE,$73,$C6,$35,$87,$91,$D5,$F3,$9F,$1C,$69,$EA,$7E,$52,$B9,$DF
	.dc.b	$68,$E8,$47,$F4,$F1,$AB,$FF,$0C,$FC,$A7,$DA,$AF,$F0,$A2,$FA,$0F
	.dc.b	$F4,$7F,$37,$F5,$18,$C4,$EB,$6C,$F1,$F2,$6C,$7F,$A9,$67,$6F,$F9
	.dc.b	$BD,$B5,$CE,$77,$2F,$9F,$FF,$3F,$3F,$93,$6B,$FD,$2C,$D7,$C1,$7E
	.dc.b	$5D,$A6,$DE,$71,$C2,$BF,$96,$37,$05,$7E,$5F,$E8,$3E,$8D,$6B,$E5
	.dc.b	$FD,$3C,$9F,$E8,$CF,$66,$F9,$E7,$AC,$CC,$3C,$CF,$8F,$23,$49,$1F
	.dc.b	$E2,$A8,$CB,$63,$BE,$38,$8F,$F5,$37,$D0,$3D,$1F,$8B,$ED,$BE,$5E
	.dc.b	$9F,$22,$75,$5F,$E3,$66,$BF,$1D,$AF,$F7,$36,$BF,$EC,$A6,$BF,$11
	.dc.b	$E3,$AF,$F5,$36,$ED,$BC,$3C,$7F,$03,$F6,$3E,$81,$F4,$1F,$6B,$E8
	.dc.b	$3F,$C1,$79,$7F,$92,$1A,$FF,$B0,$67,$6F,$F9,$31,$AF,$DD,$6B,$FE
	.dc.b	$09,$6B,$FE,$E0,$6B,$F6,$9E,$39,$7F,$69,$6E,$3B,$BF,$27,$57,$E3
	.dc.b	$BE,$83,$FB,$BE,$1F,$9E,$F2,$7F,$2F,$EB,$3A,$51,$BF,$2E,$57,$F2
	.dc.b	$5B,$E2,$FE,$4D,$AF,$F8,$7D,$AF,$FB,$D9,$DE,$9F,$DF,$5B,$9E,$6B
	.dc.b	$C3,$83,$CA,$F3,$78,$9E,$4F,$9A,$7C,$1D,$33,$9F,$F9,$D3,$2F,$EE
	.dc.b	$0F,$96,$BF,$49,$18,$8F,$34,$60,$E2,$FF,$BC,$E3,$59,$FF,$93,$74
	.dc.b	$FE,$D6,$FE,$FF,$C3,$F2,$92,$BF,$D0,$7F,$3C,$F6,$7F,$72,$D8,$FE
	.dc.b	$94,$D7,$F7,$0D,$7F,$C8,$8D,$7F,$88,$FD,$07,$29,$91,$FC,$4A,$3E
	.dc.b	$2F,$D0,$48,$FE,$2A,$FC,$5E,$BF,$0F,$B5,$FC,$0F,$03,$49,$5F,$C2
	.dc.b	$79,$1C,$1F,$BB,$8F,$F5,$93,$56,$9F,$E6,$F8,$C3,$FC,$87,$CE,$FE
	.dc.b	$2F,$F4,$0F,$0F,$F1,$3E,$8F,$DC,$C1,$8A,$31,$0A,$E0,$FF,$AB,$1A
	.dc.b	$FF,$1D,$6B,$FE,$69,$6B,$FC,$60,$E5,$5F,$C8,$63,$8E,$1F,$E4,$96
	.dc.b	$F4,$6D,$B5,$FC,$CF,$C6,$FC,$7F,$87,$F9,$0F,$9B,$93,$FD,$02,$32
	.dc.b	$FF,$42,$8C,$BF,$CE,$63,$F2,$FF,$41,$FA,$8F,$86,$FF,$3F,$65,$FE
	.dc.b	$95,$6E,$5F,$D3,$27,$77,$FE,$D1,$38,$18,$3F,$DB,$3E,$97,$3F,$94
	.dc.b	$91,$FE,$D7,$8E,$C6,$1F,$F6,$FC,$61,$D1,$FE,$DD,$75,$87,$8E,$7F
	.dc.b	$A9,$FB,$9F,$E0,$9F,$92,$FA,$FD,$B3,$0F,$F7,$B4,$E5,$BE,$12,$B8
	.dc.b	$B6,$33,$1F,$07,$C6,$F8,$5F,$1F,$F0,$FC,$A9,$5F,$EE,$F8,$CB,$FE
	.dc.b	$F1,$8C,$BF,$EF,$88,$FE,$C3,$F0,$4F,$A0,$FE,$C7,$F4,$0F,$88,$A6
	.dc.b	$A2,$5F,$F8,$DE,$B8,$36,$3F,$B2,$46,$23,$FB,$43,$77,$FF,$47,$37
	.dc.b	$8B,$F1,$8D,$9F,$EC,$B8,$3E,$33,$F8,$97,$C7,$7F,$13,$64,$FE,$A2
	.dc.b	$32,$FE,$AA,$32,$FE,$AA,$FF,$B7,$F6,$7E,$C2,$BF,$C0,$B8,$BE,$75
	.dc.b	$F1,$9F,$C8,$BE,$22,$63,$FD,$DA,$BF,$DE,$26,$E7,$8D,$70,$9D,$63
	.dc.b	$B8,$C4,$78,$4F,$DD,$B8,$AF,$DE,$7A,$5F,$E1,$0F,$A0,$FD,$ED,$FE
	.dc.b	$B7,$57,$C8,$7D,$23,$5E,$C4,$CB,$FC,$CD,$7F,$C1,$DF,$F6,$F3,$5F
	.dc.b	$DB,$35,$FF,$07,$BD,$9B,$1C,$DF,$30,$A7,$C6,$FF,$89,$3E,$35,$F4
	.dc.b	$1F,$D5,$BE,$83,$FD,$4F,$7F,$00,$FE,$1E,$E1,$E7,$3C,$1D,$07,$0E
	.dc.b	$A7,$B1,$D7,$C3,$F9,$BF,$41,$FD,$53,$E0,$65,$77,$E0,$FB,$13,$A8
	.dc.b	$C7,$0A,$31,$1F,$47,$95,$F1,$3F,$EC,$F8,$F1,$FC,$3B,$E7,$BF,$97
	.dc.b	$C3,$E7,$9F,$31,$FC,$7F,$E8,$1F,$41,$FC,$86,$73,$5F,$E3,$A9,$CB
	.dc.b	$FE,$3C,$8C,$AC,$3B,$6B,$9C,$3F,$95,$46,$25,$F4,$1C,$30,$E7,$18
	.dc.b	$AF,$16,$3F,$CC,$6B,$AD,$F1,$6E,$5E,$A7,$94,$EE,$B8,$DF,$F9,$42
	.dc.b	$77,$8D,$9F,$1D,$F1,$77,$5E,$13,$DD,$7C,$DF,$F2,$CE,$1E,$19,$E3
	.dc.b	$7E,$D9,$D9,$46,$3E,$CC,$BC,$31,$F3,$EB,$8B,$DC,$DE,$1D,$DA,$79
	.dc.b	$CF,$0E,$3E,$76,$E1,$DF,$0E,$D7,$EA,$C3,$B9,$E0,$CB,$E2,$DF,$33
	.dc.b	$A9,$E1,$CB,$D8,$F4,$3C,$97,$F5,$27,$8C,$6F,$1F,$13,$F7,$DB,$3B
	.dc.b	$C3,$E5,$D7,$85,$F7,$E1,$C8,$8E,$0F,$87,$1D,$88,$E0,$97,$F3,$8F
	.dc.b	$2E,$5F,$1F,$1F,$4C,$78,$F2,$F8,$93,$F9,$9A,$F8,$87,$79,$F7,$5F
	.dc.b	$67,$A3,$E7,$73,$1E,$A2,$38,$3E,$36,$1E,$A7,$C8,$FD,$08,$F3,$C7
	.dc.b	$CE,$9E,$CF,$94,$F9,$57,$F4,$F1,$F9,$11,$E2,$AE,$38,$FB,$27,$33
	.dc.b	$CE,$73,$97,$EE,$B1,$F1,$4F,$DC,$7C,$4F,$9C,$E0,$75,$31,$7F,$B3
	.dc.b	$8E,$37,$F8,$A9,$F6,$DF,$C7,$7A,$3F,$29,$C9,$F8,$7F,$1B,$F8,$53
	.dc.b	$6B,$F3,$B1,$88,$F8,$0D,$7F,$D8,$7F,$17,$FE,$C4,$8C,$42,$31,$3F
	.dc.b	$ED,$16,$C7,$F8,$CC,$D0,$9E,$31,$FB,$1F,$21,$E5,$BB,$7F,$69,$7E
	.dc.b	$03,$F6,$47,$FB,$52,$77,$1A,$8E,$B8,$7F,$1C,$9C,$C7,$03,$FC,$89
	.dc.b	$7D,$EF,$B4,$64,$ED,$B7,$18,$CB,$FC,$9A,$33,$97,$F9,$36,$FA,$CE
	.dc.b	$67,$C3,$39,$AE,$B8,$67,$0F,$06,$3E,$77,$F7,$98,$F2,$BF,$91,$87
	.dc.b	$F0,$D8,$C3,$F2,$67,$78,$79,$2B,$78,$F9,$F0,$FE,$0D,$7E,$63,$E3
	.dc.b	$56,$DE,$47,$57,$0F,$C4,$3E,$36,$1F,$23,$1F,$44,$62,$7D,$78,$78
	.dc.b	$0F,$75,$F8,$27,$EE,$E1,$E3,$7F,$93,$1E,$31,$87,$B1,$F0,$46,$CE
	.dc.b	$F8,$71,$2D,$F6,$9F,$AC,$76,$71,$3E,$63,$93,$E5,$8F,$83,$7E,$5C
	.dc.b	$7E,$4B,$63,$D8,$FD,$36,$9D,$A3,$6F,$A8,$F0,$1D,$0C,$3A,$AF,$15
	.dc.b	$79,$0F,$14,$FC,$0C,$7E,$7B,$F0,$F1,$F8,$4F,$E7,$71,$F8,$15,$F8
	.dc.b	$26,$C7,$D8,$BF,$6B,$67,$B3,$F7,$27,$6F,$DB,$6B,$DE,$FC,$2F,$A8
	.dc.b	$F2,$63,$F0,$EF,$AC,$7D,$4B,$EA,$D3,$E0,$47,$B9,$B1,$E7,$3E,$B5
	.dc.b	$FD,$88,$F8,$CF,$F8,$4B,$67,$C8,$3D,$DF,$F2,$BE,$3F,$E5,$98,$C3
	.dc.b	$FE,$37,$E3,$FF,$11,$FC,$6E,$B1,$FF,$67,$D6,$9F,$F2,$C6,$1D,$30
	.dc.b	$D5,$BE,$0B,$FC,$47,$8F,$F8,$CF,$0E,$17,$FC,$02,$B5,$C9,$D7,$17
	.dc.b	$CF,$BE,$4B,$0D,$C7,$C9,$F4,$1F,$E3,$17,$FC,$8F,$19,$AF,$F2,$AB
	.dc.b	$71,$ED,$5F,$E3,$9A,$DD,$F9,$A3,$17,$FE,$5C,$E4,$FF,$28,$57,$F9
	.dc.b	$49,$B1,$FE,$75,$87,$CF,$BD,$3F,$45,$6B,$1F,$E3,$77,$FC,$B7,$C0
	.dc.b	$AF,$FB,$8E,$79,$BF,$CC,$DA,$E6,$36,$FF,$3D,$7E,$86,$1E,$0E,$5E
	.dc.b	$F1,$8C,$7F,$96,$E1,$FE,$90,$DD,$EB,$FD,$78,$FB,$EF,$D4,$EC,$7B
	.dc.b	$39,$AD,$E5,$F5,$6B,$EC,$F1,$FF,$9E,$A3,$07,$E0,$DC,$69,$FE,$67
	.dc.b	$C3,$FD,$A1,$3F,$87,$5F,$41,$DF,$D5,$F0,$36,$74,$DA,$E4,$FE,$07
	.dc.b	$E8,$1E,$A5,$3D,$DE,$B5,$BF,$F4,$2C,$E9,$FE,$B9,$61,$19,$72,$D7
	.dc.b	$FB,$13,$5E,$07,$0A,$F9,$87,$77,$F0,$1B,$F3,$DF,$F3,$1E,$CF,$FA
	.dc.b	$03,$0F,$D3,$2F,$A0,$FD,$0B,$9B,$FF,$46,$39,$9D,$F7,$6A,$B6,$D8
	.dc.b	$FF,$7E,$FD,$03,$C8,$FD,$87,$84,$8C,$4F,$83,$E2,$73,$18,$CB,$FD
	.dc.b	$95,$5B,$BF,$D9,$FB,$EF,$0B,$71,$87,$CF,$F0,$9E,$8F,$46,$E7,$EC
	.dc.b	$6E,$3E,$A6,$EF,$F0,$9B,$3F,$EE,$8D,$D1,$8A,$78,$69,$CD,$7C,$8F
	.dc.b	$43,$E3,$F4,$7C,$86,$C6,$4F,$9B,$19,$69,$F4,$3D,$3F,$9F,$EB,$FD
	.dc.b	$06,$C6,$4F,$AB,$19,$69,$F8,$AF,$87,$FC,$81,$F6,$3D,$9B,$19,$3F
	.dc.b	$1B,$9F,$E3,$27,$F6,$AD,$8C,$F4,$A7,$2D,$8F,$E5,$5C,$4F,$6B,$AD
	.dc.b	$FF,$B3,$E3,$13,$FA,$9F,$A0,$FD,$B1,$F8,$AF,$A0,$B1,$FE,$E4,$F3
	.dc.b	$DE,$DF,$89,$B1,$E6,$6E,$1F,$E4,$19,$FE,$17,$C5,$F4,$7D,$EF,$CA
	.dc.b	$9F,$C0,$FE,$82,$E1,$F9,$AF,$77,$F9,$96,$3F,$C1,$A3,$07,$87,$C8
	.dc.b	$C9,$E1,$23,$E1,$7B,$F3,$3E,$83,$F4,$67,$09,$FD,$3B,$FC,$FF,$2F
	.dc.b	$CA,$47,$FC,$00,$6E,$9F,$EF,$3E,$7F,$E7,$53,$FB,$69,$CF,$4F,$F2
	.dc.b	$E7,$75,$FE,$5D,$6B,$BE,$BF,$AF,$18,$8F,$E0,$CF,$2A,$BD,$3F,$CE
	.dc.b	$5E,$07,$F1,$BF,$2B,$CD,$CB,$23,$F7,$FC,$F7,$06,$5E,$3A,$EE,$4A
	.dc.b	$FC,$4E,$3F,$7A,$BF,$8B,$B6,$BF,$85,$FC,$47,$36,$D7,$F1,$B7,$FA
	.dc.b	$5F,$D0,$7C,$B8,$C1,$FC,$61,$C5,$FF,$00,$F9,$E7,$C1,$7F,$6B,$9C
	.dc.b	$DF,$2E,$0F,$CC,$C7,$0B,$FF,$42,$35,$E9,$1A,$3F,$A5,$7D,$13,$9B
	.dc.b	$A4,$61,$FE,$F1,$F4,$1F,$5A,$BF,$99,$37,$C4,$66,$FA,$1D,$FA,$D3
	.dc.b	$6F,$D6,$DF,$F7,$9C,$3F,$D2,$CD,$7F,$9A,$B5,$EB,$3F,$E1,$3F,$55
	.dc.b	$F6,$5F,$C8,$77,$FE,$E3,$7E,$39,$CF,$91,$FE,$0D,$F9,$5F,$F0,$8E
	.dc.b	$5F,$D0,$1B,$97,$F4,$9F,$57,$F9,$B3,$F7,$27,$5E,$DF,$13,$F8,$0D
	.dc.b	$7F,$88,$67,$FD,$58,$D7,$73,$FC,$22,$3A,$FC,$E3,$E8,$DE,$7F,$3A
	.dc.b	$FB,$76,$AD,$4F,$9F,$E5,$E5,$6A,$BF,$C7,$3D,$D8,$7C,$C7,$F6,$1C
	.dc.b	$55,$E6,$25,$78,$DF,$E0,$7E,$4F,$FA,$F6,$3F,$8A,$FD,$1F,$F4,$67
	.dc.b	$FA,$3B,$5C,$9F,$C6,$8C,$9E,$3F,$A2,$7C,$7F,$F8,$8F,$1F,$F6,$97
	.dc.b	$07,$F6,$C9,$CF,$0F,$FB,$62,$3F,$B7,$FD,$1B,$E5,$3F,$D8,$3E,$4F
	.dc.b	$F5,$FC,$3F,$BD,$61,$C1,$E6,$63,$1E,$D3,$B8,$4E,$E3,$17,$D7,$2F
	.dc.b	$05,$71,$CA,$BC,$30,$E7,$97,$9D,$B1,$FE,$0B,$BE,$4F,$88,$E1,$5C
	.dc.b	$B1,$F0,$5F,$18,$EB,$5C,$31,$F3,$56,$AF,$C4,$73,$70,$A7,$85,$F3
	.dc.b	$8D,$61,$E4,$7C,$27,$8E,$76,$F4,$BE,$F3,$BB,$EB,$5A,$AE,$35,$AC
	.dc.b	$3F,$AE,$D7,$CF,$9F,$F1,$19,$D8,$AE,$5F,$F1,$3F,$C3,$BE,$87,$F3
	.dc.b	$11,$FE,$32,$8C,$71,$FF,$68,$6C,$66,$76,$F4,$F2,$1F,$05,$E2,$6C
	.dc.b	$7F,$8F,$30,$FF,$0C,$4E,$49,$1F,$26,$77,$97,$CE,$6E,$BF,$E3,$CF
	.dc.b	$A3,$7C,$DF,$F8,$FA,$B5,$87,$D9,$BD,$E1,$F9,$06,$AB,$E2,$EE,$63
	.dc.b	$E1,$7F,$17,$1A,$8C,$12,$BB,$BF,$89,$8D,$46,$09,$87,$E9,$EF,$C3
	.dc.b	$3A,$8C,$5F,$82,$B7,$7F,$E5,$FA,$D3,$E9,$9D,$57,$08,$F0,$61,$F0
	.dc.b	$63,$FB,$0F,$C1,$7D,$99,$FF,$3E,$7D,$1F,$F9,$8B,$BB,$FD,$69,$28
	.dc.b	$C7,$1F,$0D,$AD,$FF,$8F,$2B,$93,$FC,$4B,$CA,$FF,$60,$5E,$9B,$1F
	.dc.b	$E8,$9A,$E2,$BE,$8F,$FD,$2E,$FF,$22,$E2,$FF,$4A,$FC,$B7,$66,$C7
	.dc.b	$F2,$19,$DB,$FC,$80,$E5,$F0,$DB,$F4,$D7,$4E,$F5,$C3,$5F,$F5,$AA
	.dc.b	$F9,$8E,$1E,$F4,$C7,$B4,$70,$30,$FF,$2E,$CB,$FD,$53,$39,$8D,$9E
	.dc.b	$9C,$7F,$CC,$2D,$8D,$3E,$7F,$A2,$7A,$FF,$0B,$E0,$3C,$F3,$FC,$F5
	.dc.b	$AE,$E3,$E3,$CF,$08,$D6,$9C,$4C,$CC,$8C,$90,$E4,$7F,$49,$FA,$17
	.dc.b	$D0,$FD,$2F,$A2,$FF,$03,$CF,$F4,$D6,$BF,$36,$76,$F1,$9E,$1E,$AC
	.dc.b	$8C,$10,$FF,$59,$CF,$06,$CF,$F9,$A9,$46,$5F,$E1,$FF,$44,$F1,$FF
	.dc.b	$11,$CE,$F2,$E9,$F8,$89,$D5,$7F,$63,$6B,$F9,$19,$E0,$FF,$B5,$E7
	.dc.b	$85,$48,$E0,$D7,$87,$91,$D7,$C5,$FC,$C7,$CF,$FF,$76,$9F,$EE,$BD
	.dc.b	$48,$C4,$7E,$47,$2F,$EC,$2F,$8E,$B8,$F9,$72,$30,$43,$84,$77,$7A
	.dc.b	$E5,$FD,$6F,$A7,$F6,$3D,$FF,$08,$93,$FE,$01,$FA,$0F,$F4,$73,$7D
	.dc.b	$BC,$BF,$06,$7F,$5B,$3A,$8F,$89,$F4,$58,$A9,$1F,$EF,$B8,$F9,$FF
	.dc.b	$45,$23,$FD,$FC,$FC,$2E,$9E,$7E,$E5,$9F,$87,$BA,$73,$18,$EB,$F9
	.dc.b	$24,$7E,$4F,$27,$F7,$F7,$FC,$35,$F4,$4E,$DF,$E0,$DE,$87,$DA,$E0
	.dc.b	$78,$BF,$E7,$68,$C3,$FE,$F6,$AE,$17,$CA,$F8,$F4,$48,$C1,$0F,$F5
	.dc.b	$73,$97,$F8,$44,$F2,$7F,$11,$E0,$FE,$43,$E7,$24,$7A,$23,$F2,$66
	.dc.b	$1F,$C7,$46,$1F,$6F,$D1,$7F,$1F,$F8,$3F,$ED,$3A,$7A,$A7,$FD,$C3
	.dc.b	$18,$CF,$F9,$8D,$7E,$78,$E1,$30,$E6,$7E,$42,$7F,$C5,$8F,$F2,$D5
	.dc.b	$F2,$9D,$CF,$07,$A3,$27,$84,$65,$E5,$19,$79,$46,$5E,$91,$FC,$F3
	.dc.b	$DD,$FE,$E7,$F2,$9F,$53,$B2,$57,$F9,$5A,$72,$DF,$39,$5A,$8F,$F6
	.dc.b	$E5,$73,$9C,$C7,$AF,$E4,$BE,$37,$91,$23,$04,$3D,$CE,$5F,$81,$1F
	.dc.b	$D2,$3E,$11,$F4,$5F,$D2,$BE,$85,$F1,$34,$D4,$4B,$FF,$34,$4E,$9B
	.dc.b	$39,$7F,$A5,$57,$47,$24,$33,$5F,$C1,$3E,$3F,$F8,$2F,$C7,$C9,$FD
	.dc.b	$1C,$65,$FD,$2C,$65,$FE,$C9,$DC,$9F,$ED,$3C,$AF,$9F,$FF,$7D,$D7
	.dc.b	$E9,$BC,$DF,$E4,$26,$A3,$31,$FE,$7E,$9E,$1F,$1B,$A8,$E0,$7F,$A2
	.dc.b	$A3,$2F,$FA,$1F,$E8,$BF,$A9,$BF,$D9,$F6,$7C,$97,$C9,$3F,$B9,$CE
	.dc.b	$AB,$F3,$14,$8C,$75,$22,$5F,$27,$E6,$DF,$2F,$0A,$46,$08,$7F,$77
	.dc.b	$BE,$6F,$3F,$8E,$7D,$0F,$F8,$1F,$E8,$BF,$83,$4F,$F8,$2D,$B9,$F9
	.dc.b	$2B,$73,$FE,$A3,$6B,$BF,$83,$E4,$7F,$83,$A3,$33,$FB,$33,$E8,$7D
	.dc.b	$17,$F4,$EF,$7A,$5F,$C9,$E0,$4F,$D1,$33,$1F,$D3,$67,$0F,$FA,$CF
	.dc.b	$E0,$DF,$44,$C3,$E3,$7D,$0B,$E6,$7E,$45,$BE,$89,$F4,$38,$AD,$4F
	.dc.b	$CD,$9C,$BC,$AB,$FC,$DF,$87,$58,$57,$05,$8F,$8B,$1E,$11,$CF,$1E
	.dc.b	$D8,$78,$B1,$EB,$3C,$A3,$AD,$70,$8E,$78,$F7,$C7,$B5,$F0,$6E,$1D
	.dc.b	$2B,$F0,$98,$F8,$5E,$86,$2F,$96,$3E,$1A,$E1,$3F,$0D,$B5,$C6,$FD
	.dc.b	$8D,$AE,$15,$E5,$C3,$C5,$1E,$88,$EE,$7A,$31,$D6,$1D,$AB,$73,$E3
	.dc.b	$9F,$75,$F5,$8F,$8C,$70,$33,$8F,$A1,$E0,$76,$AE,$6F,$A4,$DD,$47
	.dc.b	$57,$E1,$3D,$B5,$CE,$3C,$C3,$E8,$BE,$5B,$FA,$D6,$B9,$C7,$DD,$1F
	.dc.b	$12,$79,$61,$E2,$3C,$78,$FA,$EF,$AE,$5C,$89,$87,$C7,$79,$65,$EF
	.dc.b	$9F,$39,$E9,$8C,$B8,$23,$82,$4F,$D4,$8C,$BD,$48,$FC,$C3,$27,$CA
	.dc.b	$8E,$2E,$5F,$2B,$98,$E2,$FC,$D3,$89,$EC,$73,$7F,$07,$2F,$9B,$7E
	.dc.b	$99,$F2,$5F,$4B,$E0,$75,$3C,$11,$A3,$BE,$1F,$27,$0E,$93,$CC,$E0
	.dc.b	$78,$EB,$A4,$FC,$F3,$86,$1F,$7E,$78,$1F,$7B,$C1,$FB,$CF,$E2,$9A
	.dc.b	$E7,$3F,$CD,$46,$23,$F9,$77,$C3,$FF,$2F,$8C,$42,$31,$3F,$EF,$68
	.dc.b	$C1,$FE,$EA,$71,$1A,$8F,$F7,$DB,$FB,$DF,$17,$FC,$13,$D3,$FB,$F9
	.dc.b	$EE,$7E,$AA,$32,$FE,$B0,$CE,$5F,$C3,$E3,$2F,$F1,$18,$CE,$9F,$C7
	.dc.b	$A7,$2F,$8A,$3A,$78,$FF,$E2,$3A,$E1,$1E,$87,$A4,$66,$78,$46,$2F
	.dc.b	$CD,$5B,$9E,$D5,$B8,$F6,$9D,$A7,$78,$7B,$E7,$53,$FF,$00,$4D,$1E
	.dc.b	$D3,$A1,$FD,$77,$0F,$4B,$D3,$0F,$13,$BA,$E9,$7D,$A3,$9C,$FB,$27
	.dc.b	$31,$B3,$38,$7B,$0F,$4D,$71,$AE,$A7,$19,$D9,$E1,$9D,$BC,$E3,$71
	.dc.b	$CE,$3D,$18,$7B,$9F,$A6,$62,$BB,$4F,$2A,$EC,$72,$9E,$C6,$CE,$D1
	.dc.b	$B3,$ED,$63,$F0,$8F,$31,$5F,$14,$31,$F9,$2F,$CF,$AE,$CD,$9E,$8F
	.dc.b	$80,$E1,$87,$A5,$FA,$66,$30,$F4,$BE,$5A,$F4,$D7,$28,$E8,$65,$F1
	.dc.b	$DF,$8A,$3E,$19,$CC,$F6,$9B,$3C,$4E,$DF,$0B,$71,$FB,$75,$E0,$9E
	.dc.b	$07,$80,$E8,$E5,$B5,$E2,$C3,$F2,$DB,$3E,$3B,$FD,$B9,$97,$FA,$93
	.dc.b	$5F,$F0,$A3,$FE,$2D,$9D,$7B,$38,$C5,$EE,$3D,$3F,$F0,$4D,$FF,$8F
	.dc.b	$F1,$E7,$5F,$EE,$1C,$78,$63,$CA,$BB,$1D,$4F,$F0,$9F,$D1,$3E,$4B
	.dc.b	$B6,$4F,$A2,$FF,$0D,$D6,$F1,$FF,$20,$5E,$E3,$FD,$C7,$3A,$9F,$F1
	.dc.b	$04,$74,$BE,$73,$9C,$3F,$C9,$77,$C6,$F6,$7B,$27,$57,$FE,$26,$E3
	.dc.b	$75,$7C,$18,$7D,$07,$F1,$4C,$7F,$CA,$B5,$A9,$F3,$1A,$E0,$F9,$18
	.dc.b	$A7,$4F,$82,$74,$7B,$5D,$61,$DF,$0E,$D3,$E3,$8C,$5F,$4B,$E7,$87
	.dc.b	$FA,$F6,$B9,$E1,$FE,$75,$D9,$F1,$3F,$33,$83,$FC,$BB,$5F,$5E,$31
	.dc.b	$DF,$F5,$AB,$73,$FE,$E6,$8C,$4E,$A3,$E9,$70,$3E,$47,$FD,$5F,$18
	.dc.b	$5F,$45,$F6,$9C,$E5,$F3,$AB,$84,$E7,$D1,$D3,$6F,$93,$F2,$3E,$7D
	.dc.b	$E8,$7C,$33,$05,$9F,$91,$1A,$7E,$8F,$15,$C7,$EE,$89,$C6,$FF,$1E
	.dc.b	$6D,$7F,$99,$26,$D7,$FC,$00,$9A,$F1,$BF,$03,$FE,$E1,$E8,$FC,$BF
	.dc.b	$A8,$FA,$1B,$18,$6E,$3F,$EE,$96,$D7,$FB,$A5,$FD,$67,$6B,$E8,$5E
	.dc.b	$1F,$E8,$BD,$D7,$C9,$6A,$73,$8F,$D2,$9C,$BF,$EE,$38,$CE,$3F,$3E
	.dc.b	$72,$FE,$7B,$0E,$D8,$75,$C3,$D3,$EF,$FA,$BC,$CB,$97,$FB,$FD,$B8
	.dc.b	$7F,$BF,$DB,$C2,$9F,$F5,$97,$23,$75,$6A,$73,$18,$7F,$12,$B0,$F0
	.dc.b	$46,$3E,$7F,$E5,$F3,$7C,$C6,$C6,$4E,$31,$96,$9D,$3C,$EF,$A7,$E9
	.dc.b	$7D,$46,$C6,$4F,$0C,$65,$A7,$E2,$BE,$13,$E8,$7A,$FF,$91,$6C,$64
	.dc.b	$FA,$11,$96,$99,$DB,$E8,$46,$5B,$1F,$A5,$FA,$1F,$D2,$1F,$88,$FA
	.dc.b	$1B,$1F,$C0,$7C,$A7,$AB,$FA,$9F,$9E,$FE,$13,$18,$9F,$D7,$47,$E4
	.dc.b	$3D,$B9,$18,$3E,$6B,$F9,$4F,$90,$E3,$F0,$B6,$7B,$38,$7E,$9F,$CF
	.dc.b	$DD,$3F,$6D,$C8,$8F,$CE,$92,$78,$9C,$0F,$E3,$1C,$DF,$8E,$E7,$FF
	.dc.b	$7B,$B6,$3F,$0B,$CB,$FC,$BE,$73,$D1,$FC,$92,$B5,$8F,$F4,$CE,$6F
	.dc.b	$E3,$78,$7F,$B5,$A7,$27,$F0,$FF,$2B,$85,$FF,$0D,$6C,$FF,$11,$7A
	.dc.b	$72,$48,$FC,$7F,$2D,$C7,$F8,$1E,$0A,$FD,$CF,$C3,$E7,$3F,$CE,$57
	.dc.b	$F5,$ED,$93,$9C,$7F,$36,$D9,$D3,$FC,$FB,$E8,$7E,$44,$FF,$98,$E5
	.dc.b	$18,$7F,$CC,$CD,$7F,$DE,$66,$B7,$F0,$B8,$8E,$18,$79,$07,$D0,$B9
	.dc.b	$3C,$17,$FD,$BA,$31,$87,$F4,$BB,$E2,$FF,$7B,$8C,$E1,$FC,$92,$72
	.dc.b	$B0,$EB,$87,$F5,$8F,$9E,$FB,$18,$29,$F0,$AF,$9F,$FE,$1F,$87,$A2
	.dc.b	$F7,$5A,$9C,$DF,$18,$4E,$AF,$97,$81,$DF,$D3,$FF,$07,$46,$2F,$FB
	.dc.b	$25,$6F,$0E,$78,$78,$E3,$18,$7F,$6E,$F8,$BF,$F0,$8E,$5F,$6B,$E5
	.dc.b	$DF,$3E,$F9,$2D,$7B,$3F,$E1,$B9,$CC,$63,$4F,$F0,$74,$61,$FE,$E9
	.dc.b	$87,$47,$E1,$46,$73,$FE,$9B,$39,$51,$9C,$7E,$DC,$E6,$31,$5F,$0E
	.dc.b	$1F,$29,$E3,$C9,$E2,$7A,$E3,$0D,$AF,$2C,$E9,$F2,$CE,$9F,$E8,$87
	.dc.b	$F4,$9F,$9F,$FE,$94,$D3,$27,$F4,$97,$2D,$3F,$C2,$5F,$3C,$F8,$DF
	.dc.b	$BF,$A7,$FA,$FF,$7F,$B5,$39,$E2,$FF,$62,$C7,$DD,$F9,$F7,$C3,$7E
	.dc.b	$AB,$C6,$FE,$B7,$87,$E2,$70,$E0,$F3,$31,$8F,$69,$DC,$27,$71,$8B
	.dc.b	$EB,$97,$82,$B8,$E5,$5E,$18,$73,$CB,$CE,$D8,$FC,$D5,$F2,$7C,$47
	.dc.b	$0A,$E5,$8F,$82,$F8,$C7,$5A,$E1,$8F,$9A,$B5,$7E,$23,$9B,$85,$3C
	.dc.b	$2F,$9C,$6B,$0F,$23,$E1,$3C,$73,$B7,$A5,$F7,$9D,$DF,$5A,$D5,$71
	.dc.b	$AD,$61,$FC,$46,$BE,$7C,$FF,$0E,$6B,$E2,$3F,$88,$FC,$13,$E7,$7F
	.dc.b	$33,$1F,$EF,$B8,$C7,$17,$F1,$B6,$C6,$67,$6F,$5F,$0F,$FC,$BC,$FF
	.dc.b	$1C,$3E,$D3,$87,$5E,$EF,$96,$31,$1F,$8C,$78,$E5,$F3,$5B,$AE,$6B
	.dc.b	$EF,$7C,$FB,$E6,$3F,$86,$5E,$E7,$EC,$DF,$0C,$3F,$20,$D5,$7C,$5D
	.dc.b	$CC,$7E,$CC,$EE,$75,$18,$9D,$57,$78,$FC,$4C,$EA,$30,$4C,$3D,$13
	.dc.b	$AC,$3C,$F1,$89,$D4,$F9,$AF,$DB,$5A,$8C,$1E,$12,$4F,$E9,$AF,$C1
	.dc.b	$7F,$E3,$2F,$4D,$F5,$9F,$CB,$FA,$7F,$AD,$FA,$AF,$CA,$5E,$96,$79
	.dc.b	$BF,$9E,$4E,$63,$1A,$7F,$AF,$B1,$E6,$FF,$93,$F1,$CD,$6E,$3F,$D6
	.dc.b	$8D,$7F,$CA,$11,$C9,$FF,$2D,$F8,$57,$3F,$EB,$11,$FE,$9F,$6B,$FE
	.dc.b	$B2,$E3,$7E,$25,$FE,$60,$8D,$57,$12,$3F,$CD,$A3,$4F,$F9,$FA,$BF
	.dc.b	$D6,$EF,$5F,$33,$C0,$78,$FD,$47,$A3,$89,$D1,$0F,$F4,$3C,$69,$C7
	.dc.b	$07,$F7,$4B,$EB,$1D,$EF,$A5,$FB,$FE,$85,$F1,$74,$95,$FD,$33,$F1
	.dc.b	$37,$CB,$62,$3F,$D2,$B5,$C1,$FF,$61,$35,$F8,$93,$C1,$CD,$71,$CB
	.dc.b	$FC,$0F,$5B,$7F,$B0,$56,$EB,$AF,$57,$94,$8F,$8F,$DA,$AE,$8F,$0C
	.dc.b	$7E,$0B,$81,$5F,$E7,$C9,$C9,$0E,$07,$9A,$BF,$01,$F3,$EF,$A0,$FF
	.dc.b	$11,$4A,$30,$FF,$89,$5A,$FF,$07,$6B,$FE,$24,$73,$8F,$F8,$91,$BC
	.dc.b	$FF,$E0,$B6,$CF,$F8,$57,$E8,$7F,$87,$56,$EF,$C6,$49,$FF,$77,$4E
	.dc.b	$AF,$FC,$29,$E1,$FF,$B9,$A7,$FD,$E5,$ED,$BE,$73,$F3,$4D,$8F,$F7
	.dc.b	$B4,$ED,$FF,$66,$B7,$1F,$F1,$CB,$6F,$AF,$C3,$F2,$F1,$FF,$49,$39
	.dc.b	$BF,$F1,$63,$5F,$EE,$13,$B7,$FC,$61,$7F,$E4,$4F,$81,$FB,$8D,$E4
	.dc.b	$FD,$74,$63,$1F,$F1,$77,$E4,$BF,$EA,$DB,$E3,$18,$F8,$1F,$DA,$4E
	.dc.b	$67,$FD,$CA,$DC,$3B,$7D,$3F,$43,$75,$2B,$99,$FE,$43,$C1,$3F,$C2
	.dc.b	$3D,$C7,$83,$FE,$30,$AF,$F6,$B7,$9F,$98,$C4,$7F,$B5,$78,$3F,$83
	.dc.b	$7E,$06,$F3,$E9,$27,$FD,$BD,$F0,$3C,$48,$E0,$FF,$71,$73,$7F,$0F
	.dc.b	$FA,$77,$CB,$52,$4F,$A0,$7E,$07,$F8,$17,$FC,$B9,$3C,$1D,$35,$FF
	.dc.b	$32,$71,$FF,$96,$63,$1C,$5F,$E7,$6A,$E5,$49,$CD,$6E,$FC,$33,$C8
	.dc.b	$FF,$35,$57,$87,$7F,$8F,$ED,$FC,$5F,$8D,$95,$FE,$FD,$8F,$90,$61
	.dc.b	$FF,$7E,$46,$1F,$A7,$3F,$43,$E3,$7E,$76,$12,$73,$C2,$9F,$A4,$66
	.dc.b	$3C,$95,$F6,$2B,$31,$C3,$EC,$BE,$C7,$D8,$48,$C1,$0E,$38,$23,$51
	.dc.b	$89,$FC,$01,$8D,$BF,$D5,$B8,$7E,$1A,$B3,$5F,$8C,$F8,$BF,$E1,$FD
	.dc.b	$7F,$D4,$9B,$19,$F0,$BF,$51,$7F,$8E,$F3,$F5,$1F,$AB,$AD,$D7,$CF
	.dc.b	$6E,$3E,$0C,$FF,$B0,$CF,$C4,$F3,$7F,$29,$E5,$48,$C1,$0F,$6B,$97
	.dc.b	$DF,$1F,$D7,$7E,$AD,$F4,$74,$93,$F8,$FF,$98,$FC,$E3,$F9,$19,$DB
	.dc.b	$67,$6F,$E1,$44,$F0,$71,$3C,$1B,$3C,$31,$FF,$63,$D6,$DF,$77,$CE
	.dc.b	$7E,$CF,$DE,$FC,$71,$F8,$FC,$D1,$FB,$33,$5E,$CC,$C7,$F0,$F1,$FD
	.dc.b	$E7,$F0,$A7,$87,$49,$3F,$90,$F9,$E7,$D6,$62,$3D,$55,$C1,$FD,$83
	.dc.b	$5F,$C7,$CF,$07,$F6,$9F,$82,$3D,$6F,$F0,$43,$63,$31,$FC,$37,$ED
	.dc.b	$DF,$0B,$A8,$93,$A3,$F4,$73,$A7,$07,$CD,$AE,$8D,$9E,$95,$F8,$FF
	.dc.b	$9F,$7C,$B5,$25,$7E,$BB,$81,$F4,$18,$8F,$C7,$57,$07,$F3,$4D,$7F
	.dc.b	$5B,$3C,$1C,$F9,$BD,$C8,$F5,$FA,$4A,$F4,$8B,$C9,$FA,$8A,$57,$F8
	.dc.b	$B7,$E1,$7F,$C1,$F3,$FE,$F5,$C3,$C1,$F4,$5F,$8A,$C7,$19,$FF,$BC
	.dc.b	$63,$10,$9F,$35,$F2,$8C,$57,$FB,$DE,$31,$3D,$39,$FF,$DF,$4D,$AE
	.dc.b	$A0,$4F,$0A,$E5,$EE,$31,$DD,$31,$F1,$E1,$CE,$3C,$27,$3C,$39,$36
	.dc.b	$F9,$3A,$39,$63,$E4,$C7,$DC,$D8,$E1,$8F,$82,$F9,$C7,$03,$C7,$8F
	.dc.b	$8A,$FC,$F5,$BC,$3B,$E3,$E5,$24,$EA,$FC,$58,$74,$C3,$94,$7C,$36
	.dc.b	$BF,$12,$F9,$D7,$1B,$E7,$3B,$C7,$D0,$EF,$1F,$11,$E6,$8C,$E1,$D3
	.dc.b	$1F,$59,$E1,$8C,$9C,$9C,$E3,$E0,$BE,$C7,$33,$A9,$EA,$74,$FC,$53
	.dc.b	$E7,$57,$78,$CC,$F7,$9E,$47,$1B,$E7,$87,$B2,$39,$3F,$8E,$AE,$27
	.dc.b	$6B,$F8,$A7,$5A,$E3,$8F,$DA,$BE,$B7,$C6,$73,$1E,$A3,$CB,$8F,$BF
	.dc.b	$0E,$B3,$C2,$F9,$57,$94,$97,$E5,$C3,$C1,$18,$9D,$CF,$A7,$1F,$93
	.dc.b	$7C,$EB,$C1,$3E,$37,$81,$8A,$E2,$7D,$D2,$39,$21,$F7,$4E,$27,$C6
	.dc.b	$78,$BF,$98,$3D,$1C,$FF,$C2,$5E,$0D,$BF,$DC,$37,$1F,$D2,$BB,$9F
	.dc.b	$DA,$7C,$27,$ED,$63,$10,$8C,$6F,$FE,$F7,$6C,$7E,$EC,$E2,$27,$8C
	.dc.b	$7E,$EB,$B9,$F4,$0F,$2F,$F7,$B7,$C3,$4F,$E4,$55,$A9,$CC,$F8,$4F
	.dc.b	$DD,$46,$5F,$DE,$1C,$74,$F1,$46,$6B,$F9,$64,$E4,$FE,$21,$19,$9F
	.dc.b	$54,$74,$BF,$23,$D0,$F0,$57,$92,$32,$F4,$8C,$CF,$82,$72,$F4,$9E
	.dc.b	$17,$EA,$C3,$E1,$56,$AF,$E4,$1C,$8F,$2B,$97,$E6,$47,$09,$E5,$1E
	.dc.b	$76,$D7,$68,$F4,$4E,$DA,$6C,$B1,$B3,$D5,$1B,$6C,$6D,$A6,$EF,$D8
	.dc.b	$7B,$EB,$C2,$D9,$EA,$D9,$E2,$D3,$8C,$7B,$A7,$8B,$4E,$25,$9E,$2D
	.dc.b	$9E,$2B,$0F,$B4,$FC,$AB,$F5,$4E,$A7,$33,$F2,$8F,$64,$67,$0F,$9D
	.dc.b	$1F,$85,$3A,$38,$7E,$A1,$8C,$3C,$ED,$3C,$E5,$C3,$CE,$DA,$ED,$3C
	.dc.b	$2F,$C7,$3D,$DA,$F6,$6E,$1E,$A8,$F6,$1E,$82,$DF,$D3,$6D,$FC,$D6
	.dc.b	$9F,$1C,$A7,$8C,$F9,$1F,$5E,$F8,$8F,$CF,$EF,$FA,$0C,$53,$9C,$BF
	.dc.b	$C3,$5F,$05,$A7,$F5,$31,$9C,$3F,$CD,$0D,$BE,$4D,$9F,$EF,$DE,$33
	.dc.b	$95,$87,$3C,$3C,$79,$5C,$BC,$3E,$AD,$C8,$EF,$1E,$5F,$51,$98,$99
	.dc.b	$FF,$A0,$EB,$6F,$F8,$C6,$73,$8F,$F9,$9F,$EC,$FF,$D2,$13,$FE,$BD
	.dc.b	$6C,$7F,$A7,$3E,$D9,$F2,$DE,$58,$4E,$B7,$FE,$0B,$18,$F3,$75,$39
	.dc.b	$C3,$DF,$87,$27,$B4,$E8,$8F,$FA,$CA,$35,$5E,$59,$F0,$7E,$AE,$F7
	.dc.b	$A9,$F0,$7D,$77,$D0,$63,$1F,$F6,$B5,$6D,$B5,$B9,$D6,$1F,$E4,$CF
	.dc.b	$63,$E5,$72,$FF,$A2,$2F,$FC,$AD,$18,$EF,$CB,$63,$27,$F9,$6E,$3E
	.dc.b	$22,$F8,$6A,$7D,$B5,$F4,$3F,$E8,$27,$FD,$D4,$6C,$B1,$C1,$FF,$21
	.dc.b	$70,$7F,$30,$78,$F9,$79,$CB,$FD,$3D,$1C,$5B,$5D,$EF,$B4,$62,$7F
	.dc.b	$CE,$1F,$47,$EE,$F0,$7D,$ED,$8C,$9F,$E7,$58,$CB,$4F,$C5,$7A,$9F
	.dc.b	$91,$3E,$3E,$16,$32,$4B,$FF,$3B,$C6,$5F,$F3,$CF,$9B,$F3,$BE,$5B
	.dc.b	$E7,$B6,$BF,$D0,$8D,$9D,$C7,$FA,$1E,$76,$D3,$65,$8F,$F4,$2E,$FF
	.dc.b	$6B,$E4,$7E,$DB,$67,$FD,$20,$D9,$DC,$7F,$A4,$E7,$F0,$3F,$5F,$F8
	.dc.b	$2F,$AF,$B3,$FE,$97,$6D,$7F,$A4,$9B,$C1,$FE,$8B,$7F,$49,$F6,$EF
	.dc.b	$AE,$A7,$D3,$F9,$AB,$87,$DC,$3F,$C0,$93,$B7,$18,$7F,$A0,$63,$0F
	.dc.b	$F9,$FE,$3F,$D0,$87,$E2,$A7,$9C,$7F,$B7,$E7,$78,$7E,$C9,$B5,$C3
	.dc.b	$89,$7C,$2B,$FD,$45,$F1,$5F,$A0,$3F,$43,$EE,$DB,$F1,$E3,$FE,$B6
	.dc.b	$F9,$3F,$E0,$1E,$69,$D4,$7F,$BB,$E3,$13,$FE,$BD,$F8,$CF,$D5,$9F
	.dc.b	$AC,$FA,$FB,$1E,$EF,$99,$7E,$CC,$F9,$CE,$BE,$EF,$F2,$19,$CC,$62
	.dc.b	$BC,$78,$FF,$BC,$1B,$5B,$BE,$51,$83,$B3,$F0,$B2,$61,$E0,$E2,$E9
	.dc.b	$1F,$34,$CC,$ED,$7D,$B7,$F3,$5B,$F8,$6D,$8E,$1E,$3B,$ED,$FF,$96
	.dc.b	$C7,$33,$8F,$4F,$F2,$7A,$E1,$C7,$FC,$99,$B8,$7F,$AF,$F0,$E5,$E6
	.dc.b	$6E,$E4,$FE,$13,$CE,$7D,$BB,$CD,$CE,$7E,$99,$CC,$7F,$35,$FA,$2F
	.dc.b	$E5,$3A,$7F,$34,$6E,$1F,$E3,$EF,$84,$FF,$70,$BF,$7A,$33,$87,$FB
	.dc.b	$45,$A9,$CE,$0A,$E3,$3E,$F9,$D9,$FC,$E1,$FF,$1C,$B5,$FF,$31,$35
	.dc.b	$ED,$E7,$FF,$48,$7F,$CC,$DE,$EB,$E1,$3F,$A7,$37,$1F,$C9,$FE,$FE
	.dc.b	$F0,$73,$52,$7F,$2F,$5B,$BF,$A5,$7B,$9C,$D6,$E7,$EB,$56,$F4,$FE
	.dc.b	$B3,$7B,$9C,$D7,$F5,$FF,$C7,$BF,$63,$1F,$75,$B5,$C2,$75,$33,$0F
	.dc.b	$D4,$D6,$9F,$F4,$15,$F7,$7F,$D0,$91,$99,$FF,$80,$0D,$4D,$9E,$11
	.dc.b	$A3,$C9,$F3,$32,$BF,$6F,$FC,$01,$FA,$59,$F3,$4E,$5B,$3A,$EA,$C6
	.dc.b	$5F,$DE,$31,$5E,$2B,$FD,$31,$87,$3A,$D4,$FC,$3F,$AC,$95,$FC,$1B
	.dc.b	$F8,$07,$F0,$75,$CE,$73,$97,$C7,$7F,$C1,$13,$AA,$F8,$31,$FD,$C3
	.dc.b	$4F,$F1,$07,$D0,$BF,$2F,$FE,$E5,$87,$ED,$7D,$BF,$F1,$35,$6A,$73
	.dc.b	$97,$FA,$52,$73,$8F,$F2,$AC,$39,$7A,$7D,$63,$17,$C1,$FF,$5A,$7D
	.dc.b	$2F,$F8,$2F,$C1,$FF,$5E,$FB,$D8,$F6,$3F,$83,$EB,$FE,$BD,$8F,$F5
	.dc.b	$D7,$D3,$3C,$BF,$F5,$1F,$C1,$7F,$8E,$28,$31,$E8,$F1,$32,$E2,$38
	.dc.b	$46,$F0,$EF,$18,$BE,$D9,$79,$2B,$8D,$6B,$1F,$36,$59,$9E,$4D,$BE
	.dc.b	$31,$E1,$32,$5C,$7C,$55,$CA,$35,$3C,$31,$F2,$D6,$A7,$C3,$7D,$E3
	.dc.b	$10,$AE,$17,$CE,$F8,$E1,$E7,$AE,$27,$6C,$BE,$4D,$6C,$91,$AA,$F8
	.dc.b	$53,$A8,$C4,$F9,$DF,$F6,$2E,$1F,$39,$B7,$F6,$1B,$8F,$FB,$29,$FA
	.dc.b	$BF,$0B,$FE,$35,$FA,$1C,$42,$30,$ED,$AF,$FB,$49,$AF,$E2,$3E,$0B
	.dc.b	$FD,$F4,$F5,$C3,$FD,$11,$F0,$DD,$1B,$19,$9D,$BE,$87,$8E,$9F,$39
	.dc.b	$B9,$F5,$BF,$B9,$F4,$CF,$5B,$FD,$BF,$39,$C3,$ED,$5F,$07,$F1,$AD
	.dc.b	$55,$C5,$FB,$11,$9C,$7C,$B8,$F8,$71,$EF,$8F,$6B,$E7,$1C,$08,$FE
	.dc.b	$2F,$0E,$8F,$8A,$33,$5C,$E3,$55,$CD,$FD,$26,$1E,$12,$57,$82,$38
	.dc.b	$E3,$ED,$BE,$47,$E9,$B0,$EE,$FE,$BF,$91,$F2,$1F,$E1,$3B,$FE,$01
	.dc.b	$F4,$CF,$73,$F6,$71,$FD,$F7,$9D,$39,$8F,$F1,$7D,$F2,$F5,$37,$D3
	.dc.b	$FC,$2F,$1E,$95,$C6,$F8,$37,$1D,$F3,$FF,$AE,$AB,$77,$FE,$A2,$D5
	.dc.b	$F2,$5F,$8E,$C3,$FD,$7A,$D8,$CC,$7E,$49,$AF,$E5,$DB,$3C,$E7,$BC
	.dc.b	$FF,$0E,$9C,$DF,$F1,$0E,$47,$D3,$7F,$B6,$6B,$FC,$73,$4F,$94,$E3
	.dc.b	$4F,$CF,$78,$D2,$31,$8F,$68,$FF,$69,$D5,$BE,$51,$E7,$8F,$51,$1C
	.dc.b	$1E,$07,$81,$CF,$E4,$5F,$4F,$D7,$E9,$DF,$09,$E4,$FC,$F9,$F5,$F4
	.dc.b	$D4,$4C,$7F,$6F,$5C,$1B,$1F,$36,$31,$1F,$BF,$6F,$07,$FB,$51,$BC
	.dc.b	$1F,$84,$6C,$FC,$1F,$9A,$7D,$3F,$97,$E2,$3E,$3F,$E7,$FF,$EC,$FA
	.dc.b	$FF,$3A,$F1,$23,$11,$FE,$3F,$C3,$FD,$80,$FF,$9A,$1A,$F8,$EB,$CC
	.dc.b	$47,$D6,$F7,$78,$3F,$5B,$0F,$EE,$71,$FE,$72,$FA,$A7,$C8,$7D,$5F
	.dc.b	$AA,$7C,$CD,$3E,$C4,$EA,$B8,$57,$E4,$63,$11,$FC,$8D,$AF,$E9,$83
	.dc.b	$87,$F5,$2F,$0D,$BF,$DF,$AD,$DB,$35,$E3,$F0,$7F,$11,$E3,$BE,$1F
	.dc.b	$F1,$5F,$50,$F8,$8A,$6A,$26,$7F,$CC,$AB,$83,$63,$F5,$11,$88,$FE
	.dc.b	$74,$DC,$D9,$7E,$FA,$30,$DA,$F0,$F9,$7F,$9B,$F9,$0F,$BB,$87,$FA
	.dc.b	$71,$AF,$F9,$DD,$B8,$74,$8F,$EF,$7F,$54,$F9,$CF,$D2,$36,$3F,$A4
	.dc.b	$B5,$FE,$9C,$D7,$E6,$35,$FE,$27,$4F,$AA,$E7,$52,$7C,$91,$FB,$EF
	.dc.b	$1F,$F8,$EF,$C2,$5F,$14,$D3,$1C,$1F,$B7,$FA,$A7,$D5,$7F,$0F,$6E
	.dc.b	$3F,$D7,$9A,$FF,$63,$6B,$FC,$9B,$0F,$E5,$7F,$55,$DC,$91,$FC,$9B
	.dc.b	$1F,$C0,$3F,$E1,$EF,$97,$7D,$37,$F0,$7F,$AA,$7C,$C5,$3F,$85,$4E
	.dc.b	$AB,$FD,$96,$D7,$FB,$A3,$5F,$C9,$35,$FE,$72,$D7,$C1,$5D,$08,$FF
	.dc.b	$B5,$AB,$FB,$6F,$17,$F3,$9F,$AA,$F9,$1F,$55,$2B,$D5,$1F,$2C,$C3
	.dc.b	$FE,$4D,$C9,$39,$8F,$81,$F5,$5F,$47,$19,$18,$21,$C5,$CB,$CE,$3E
	.dc.b	$C7,$D5,$7D,$9F,$A0,$93,$F2,$E3,$2F,$CD,$8C,$BE,$38,$CB,$E5,$8C
	.dc.b	$E5,$FD,$56,$7F,$A1,$FC,$6B,$C3,$FD,$0D,$B8,$5F,$A4,$FE,$F7,$0F
	.dc.b	$F3,$33,$7B,$7F,$CA,$F8,$71,$ED,$E7,$8F,$28,$C6,$3F,$ED,$18,$EB
	.dc.b	$8F,$F7,$0F,$53,$F2,$9E,$97,$E5,$7D,$19,$7F,$87,$8C,$BF,$89,$8C
	.dc.b	$BF,$3E,$3F,$A7,$7C,$77,$FA,$FE,$3F,$CD,$3A,$FE,$C5,$7C,$13,$97
	.dc.b	$FA,$A3,$63,$7B,$7F,$7E,$BE,$51,$FE,$F1,$6B,$FC,$75,$AE,$FE,$0F
	.dc.b	$FB,$A4,$62,$3F,$CF,$2F,$2C,$3C,$4D,$AE,$13,$E0,$3F,$CF,$F7,$FA
	.dc.b	$CA,$FF,$04,$7C,$4F,$EE,$7E,$47,$F2,$77,$FE,$84,$D9,$18,$C3,$EB
	.dc.b	$74,$2F,$F7,$C7,$1F,$22,$4F,$EC,$63,$FB,$A7,$D5,$BC,$5F,$E0,$DF
	.dc.b	$57,$FD,$6B,$23,$FD,$72,$D5,$B7,$D0,$62,$B9,$91,$FF,$11,$CF,$EE
	.dc.b	$9D,$FD,$53,$D9,$FE,$29,$F5,$7E,$FD,$BD,$8D,$AF,$F5,$0D,$6D,$E3
	.dc.b	$5F,$E2,$18,$C4,$7F,$9C,$DA,$FF,$51,$6B,$FE,$38,$77,$BF,$E0,$DB
	.dc.b	$79,$AF,$1F,$C8,$7E,$8B,$EA,$5F,$53,$FC,$A7,$B7,$FD,$05,$8F,$E8
	.dc.b	$3E,$2B,$53,$33,$FB,$5D,$AC,$38,$C6,$1A,$FF,$B4,$79,$5F,$3B,$DF
	.dc.b	$CE,$ED,$F5,$52,$B8,$46,$5A,$43,$97,$AD,$FE,$2D,$F8,$CF,$F1,$0E
	.dc.b	$BF,$EC,$8C,$3B,$FD,$5F,$E8,$2F,$1A,$FF,$96,$A3,$11,$E2,$8F,$F2
	.dc.b	$ED,$F2,$8C,$1F,$E9,$EF,$7D,$87,$FB,$53,$5E,$BD,$BF,$E6,$46,$D6
	.dc.b	$70,$E7,$7C,$23,$E2,$46,$2F,$51,$FE,$DE,$C7,$C7,$7C,$A3,$67,$86
	.dc.b	$76,$FF,$AB,$6F,$A4,$EC,$97,$D6,$78,$9B,$AF,$82,$F7,$C3,$FC,$F7
	.dc.b	$97,$73,$EA,$E5,$F0,$5B,$8F,$6C,$3E,$45,$7B,$DB,$7E,$4B,$F9,$18
	.dc.b	$F0,$C7,$BC,$B2,$F9,$F9,$77,$C3,$CC,$EE,$7C,$F3,$F0,$A7,$AC,$7A
	.dc.b	$B0,$EB,$1E,$DC,$3D,$F3,$EB,$3C,$59,$79,$DF,$90,$74,$CB,$D0,$EC
	.dc.b	$E9,$1B,$7E,$31,$F8,$03,$D1,$1F,$1A,$BC,$78,$F9,$CF,$45,$F8,$AB
	.dc.b	$67,$8B,$28,$F8,$4F,$B3,$97,$B5,$F2,$CF,$3B,$EA,$48,$F4,$DF,$68
	.dc.b	$F8,$73,$DF,$2F,$87,$5C,$A7,$24,$70,$72,$7A,$3E,$63,$DD,$1F,$7E
	.dc.b	$3E,$61,$C6,$35,$3E,$17,$99,$88,$D4,$F8,$A3,$E4,$CF,$C3,$BF,$74
	.dc.b	$62,$BE,$F9,$E2,$8F,$C1,$B8,$8E,$B1,$C0,$E3,$5D,$4F,$34,$7D,$9B
	.dc.b	$FC,$53,$ED,$21,$ED,$72,$FD,$C7,$F1,$DD,$9F,$C4,$23,$F0,$AD,$7F
	.dc.b	$02,$D9,$FE,$04,$E6,$03,$E2,$6C,$66,$30,$1C,$46,$AB,$FD,$09,$3A
	.dc.b	$9C,$BF,$8F,$E1,$70,$7F,$99,$EB,$F9,$3C,$F8,$2F,$F5,$75,$AE,$6F
	.dc.b	$D6,$CE,$4E,$57,$C7,$3E,$F7,$C2,$31,$3F,$E9,$58,$FE,$17,$7C,$E3
	.dc.b	$F8,$39,$CF,$4F,$E6,$F1,$97,$C7,$1E,$EC,$73,$7D,$23,$E2,$99,$C2
	.dc.b	$DF,$5A,$E3,$87,$82,$FF,$D0,$98,$63,$0F,$E2,$F7,$DF,$0F,$4C,$62
	.dc.b	$B7,$18,$AF,$0E,$3F,$C9,$2B,$D1,$3E,$0A,$DC,$7B,$CE,$53,$B7,$CC
	.dc.b	$7F,$8E,$23,$D4,$FC,$BC,$7A,$BF,$C0,$CE,$A6,$E7,$96,$3E,$B6,$C7
	.dc.b	$63,$E5,$9F,$84,$C3,$99,$EF,$AE,$A5,$C3,$D1,$5F,$72,$B8,$46,$1E
	.dc.b	$77,$DD,$A7,$7B,$F9,$47,$69,$F9,$37,$D9,$A7,$37,$6F,$58,$B3,$ED
	.dc.b	$3A,$BC,$DF,$94,$D7,$EF,$D7,$8E,$38,$C3,$1F,$B4,$FD,$1B,$F1,$36
	.dc.b	$FC,$47,$DB,$3C,$45,$7C,$55,$F1,$5F,$24,$D7,$B6,$3F,$61,$F8,$E7
	.dc.b	$0A,$E4,$EB,$1F,$AB,$7E,$98,$FB,$66,$4F,$7C,$F8,$9F,$7B,$5F,$79
	.dc.b	$FA,$E9,$FA,$D8,$FE,$B5,$FE,$DB,$8B,$EA,$DF,$19,$FE,$E1,$DF,$FD
	.dc.b	$C5,$39,$8C,$57,$F9,$6F,$91,$F0,$D8,$CB,$E5,$63,$CA,$BF,$C5,$AD
	.dc.b	$8F,$F0,$9C,$7F,$98,$F9,$FF,$DC,$4E,$B0,$E5,$DF,$FE,$98,$F5,$FA
	.dc.b	$67,$E5,$F5,$EE,$5F,$EE,$2C,$3F,$CE,$57,$FE,$6C,$F9,$AC,$D4,$9C
	.dc.b	$E4,$FA,$9C,$69,$E2,$3F,$D4,$D1,$99,$FF,$32,$FC,$EB,$D0,$F6,$52
	.dc.b	$31,$3E,$0F,$AB,$7C,$E6,$A7,$35,$D4,$E1,$7F,$E9,$B7,$F8,$7C,$69
	.dc.b	$FE,$22,$6A,$FF,$97,$B4,$FE,$63,$1D,$38,$D3,$FE,$8B,$C1,$F1,$3E
	.dc.b	$1C,$3E,$4C,$62,$7F,$CF,$FF,$56,$F8,$1F,$5B,$63,$E4,$FC,$1F,$C1
	.dc.b	$CF,$D1,$39,$6D,$7D,$49,$7C,$BF,$28,$C3,$F6,$3A,$1E,$E5,$F0,$B0
	.dc.b	$DC,$BE,$BC,$E9,$B3,$A6,$9F,$E9,$61,$F0,$FF,$49,$A7,$FA,$C4,$7D
	.dc.b	$3F,$FA,$8D,$E1,$1B,$3E,$86,$7F,$D8,$A3,$8B,$FD,$71,$AF,$D2,$6B
	.dc.b	$FD,$11,$DF,$07,$D0,$8F,$A3,$F4,$EF,$53,$B5,$78,$E3,$0F,$BB,$EA
	.dc.b	$FE,$27,$91,$F1,$5B,$19,$3E,$1C,$65,$A7,$EA,$3D,$9F,$CF,$9F,$37
	.dc.b	$2B,$19,$3F,$3B,$F1,$D8,$CB,$F3,$9D,$BF,$F0,$02,$3E,$A7,$C1,$7D
	.dc.b	$5F,$AB,$B7,$F4,$9B,$7F,$E9,$D6,$ED,$EF,$C3,$FD,$AF,$18,$AF,$E0
	.dc.b	$0D,$A4,$EA,$7E,$EF,$0F,$E0,$FE,$63,$F0,$8D,$8C,$9F,$82,$8C,$B4
	.dc.b	$FC,$2F,$D5,$FE,$2F,$E8,$BF,$18,$D8,$C9,$F8,$A8,$CB,$4F,$F6,$EF
	.dc.b	$D5,$BE,$A7,$F7,$2D,$8F,$C4,$72,$B9,$33,$52,$7F,$63,$5B,$D3,$FC
	.dc.b	$42,$D7,$F6,$ED,$7F,$C1,$2E,$F6,$FC,$34,$FE,$53,$D3,$E9,$5F,$C3
	.dc.b	$E3,$07,$E8,$BE,$3B,$F5,$07,$EA,$7E,$12,$C7,$FB,$DB,$D1,$7C,$D7
	.dc.b	$E6,$AF,$FD,$A2,$D8,$CC,$FF,$08,$6C,$FF,$0C,$E2,$71,$3E,$43,$F6
	.dc.b	$C7,$ED,$FE,$AE,$E1,$FE,$46,$D7,$F3,$DE,$CB,$EA,$9F,$57,$FC,$8B
	.dc.b	$47,$C1,$FF,$3F,$AD,$4E,$62,$61,$E3,$7C,$C4,$7F,$C9,$73,$D7,$8F
	.dc.b	$96,$1F,$26,$7F,$D7,$B5,$89,$FE,$63,$F5,$5E,$A9,$F3,$1F,$E3,$87
	.dc.b	$1B,$FF,$3B,$F2,$DF,$57,$FD,$16,$3F,$99,$4E,$B5,$FD,$D4,$23,$8B
	.dc.b	$FC,$C9,$AF,$5E,$0F,$A1,$18,$8F,$F3,$5B,$CF,$0F,$F3,$6F,$A6,$FA
	.dc.b	$B7,$A7,$E8,$C3,$FA,$CC,$62,$BD,$38,$7F,$9D,$3E,$AF,$FA,$4C,$62
	.dc.b	$11,$87,$FA,$B6,$0F,$89,$7C,$0F,$F5,$4A,$D6,$9E,$C7,$85,$FF,$4C
	.dc.b	$6B,$FE,$5E,$6B,$E1,$F4,$FF,$19,$1C,$B1,$FC,$17,$D7,$3E,$33,$B5
	.dc.b	$7F,$7D,$C1,$39,$AF,$47,$C2,$7F,$20,$C7,$FB,$03,$6B,$E7,$FD,$63
	.dc.b	$C4,$FE,$D7,$39,$BF,$F3,$8B,$C6,$FF,$B1,$B5,$CF,$9D,$FA,$18,$C4
	.dc.b	$7F,$A7,$1E,$4B,$E7,$9F,$9B,$F7,$BF,$2D,$FF,$85,$67,$38,$7F,$7C
	.dc.b	$FA,$DF,$E8,$98,$7F,$77,$6E,$7E,$8B,$D4,$CC,$3F,$BA,$D7,$08,$FE
	.dc.b	$F2,$D7,$FD,$16,$D7,$FD,$62,$F4,$8F,$F5,$9E,$1E,$6A,$E9,$5F,$E1
	.dc.b	$38,$C4,$FE,$0E,$9F,$59,$E7,$3F,$C2,$7F,$5B,$FB,$29,$D4,$65,$F2
	.dc.b	$BF,$C1,$E7,$8C,$60,$97,$DF,$F1,$57,$83,$C6,$30,$FC,$5F,$5E,$57
	.dc.b	$E5,$7E,$BB,$F2,$2F,$96,$33,$5F,$EC,$6C,$3F,$C0,$B3,$F1,$A7,$C4
	.dc.b	$7F,$72,$F1,$BF,$89,$EB,$FE,$47,$FA,$A7,$D7,$3D,$9F,$F0,$54,$E6
	.dc.b	$31,$87,$F8,$22,$31,$3A,$8F,$D5,$D7,$28,$FF,$1B,$FD,$6F,$F8,$7F
	.dc.b	$C7,$FF,$69,$7C,$47,$FA,$FA,$31,$AF,$FB,$8E,$3F,$D8,$3F,$5C,$F3
	.dc.b	$FF,$BF,$7C,$57,$F9,$72,$BF,$DD,$37,$C2,$31,$08,$C5,$F5,$F7,$D8
	.dc.b	$F8,$2B,$FD,$E3,$8F,$5D,$31,$1C,$6B,$6F,$8C,$EE,$E3,$1E,$F8,$F6
	.dc.b	$9E,$79,$79,$B2,$F0,$65,$D2,$B5,$5D,$63,$BE,$5E,$18,$EC,$64,$B3
	.dc.b	$C0,$E3,$1B,$C3,$C3,$8F,$8F,$0E,$F7,$C2,$30,$F9,$32,$F1,$46,$23
	.dc.b	$B5,$72,$C7,$C7,$97,$A6,$7D,$F1,$C0,$FF,$30,$E5,$F5,$A3,$FC,$CA
	.dc.b	$D7,$C2,$7F,$9A,$3E,$3F,$FC,$8B,$D1,$FE,$68,$6C,$7F,$9B,$03,$87
	.dc.b	$87,$B9,$FC,$1D,$D5,$7F,$A8,$3E,$3F,$C3,$F1,$86,$92,$77,$19,$7E
	.dc.b	$4B,$73,$FA,$ED,$D9,$33,$EB,$9E,$E7,$FB,$4A,$7F,$29,$5B,$AF,$C7
	.dc.b	$B5,$FC,$73,$6B,$8C,$7D,$F8,$D6,$3E,$98,$F1,$E3,$F8,$EA,$DC,$60
	.dc.b	$98,$F9,$A7,$58,$F8,$E7,$C3,$1D,$48,$FE,$22,$B9,$D7,$18,$E7,$8F
	.dc.b	$AA,$B9,$C7,$09,$D1,$27,$E1,$C2,$B9,$C7,$C6,$E3,$7D,$73,$E0,$77
	.dc.b	$E0,$7F,$9C,$67,$31,$8E,$4E,$15,$BC,$7F,$D8,$7F,$33,$FE,$28,$7F
	.dc.b	$82,$C6,$5F,$F6,$05,$F2,$73,$7C,$DF,$F2,$DD,$71,$7F,$D7,$58,$FF
	.dc.b	$63,$7F,$BC,$F1,$3E,$B5,$F4,$1C,$63,$0F,$39,$FF,$72,$F9,$56,$F9
	.dc.b	$47,$F3,$FB,$E2,$F4,$6B,$E0,$9E,$1F,$3B,$D6,$78,$ED,$FE,$93,$C7
	.dc.b	$C5,$5A,$BF,$F0,$56,$0C,$38,$4C,$9C,$C6,$DF,$14,$65,$F7,$C7,$A7
	.dc.b	$EA,$9F,$0F,$ED,$FA,$CF,$F6,$7D,$7C,$26,$BF,$ED,$F9,$DB,$FE,$1B
	.dc.b	$6B,$FD,$76,$37,$97,$A5,$BC,$3F,$84,$AD,$CF,$F8,$D9,$B1,$FD,$A2
	.dc.b	$77,$5F,$DF,$DA,$A7,$6F,$EC,$1F,$F6,$4B,$F6,$3E,$B3,$FD,$C4,$DE
	.dc.b	$DF,$77,$54,$8C,$10,$F6,$39,$7F,$BB,$4F,$26,$CE,$71,$FF,$26,$FD
	.dc.b	$5B,$E3,$7F,$05,$F5,$9F,$E3,$8B,$FF,$1F,$35,$57,$09,$FC,$45,$71
	.dc.b	$9D,$FA,$9C,$8C,$CC,$8F,$E3,$C2,$78,$3F,$C8,$5A,$FE,$C0,$D4,$70
	.dc.b	$58,$3E,$13,$F3,$1F,$57,$FE,$87,$9F,$F2,$A3,$5F,$3E,$FF,$BA,$AE
	.dc.b	$11,$FE,$59,$6B,$A3,$F9,$0C,$F1,$CB,$FC,$E2,$DA,$E5,$7F,$A7,$BE
	.dc.b	$35,$FE,$76,$6C,$EA,$FC,$7E,$1F,$EB,$7E,$85,$F1,$DF,$B9,$F2,$9F
	.dc.b	$2A,$F9,$AC,$37,$0F,$F3,$75,$F0,$7F,$CE,$1F,$58,$F8,$BE,$98,$C9
	.dc.b	$CC,$7E,$EE,$75,$18,$21,$EF,$14,$F9,$BF,$27,$80,$F9,$AF,$1C,$CA
	.dc.b	$FE,$2F,$1F,$C1,$FE,$AD,$F5,$9C,$39,$25,$7F,$A1,$67,$2F,$F1,$EF
	.dc.b	$AC,$E1,$52,$30,$7F,$A2,$FE,$63,$F8,$5F,$2F,$FA,$83,$EB,$1F,$31
	.dc.b	$FC,$55,$B5,$BF,$1F,$F8,$C5,$F0,$E7,$FE,$6F,$5A,$AF,$73,$70,$FF
	.dc.b	$55,$D6,$E7,$FA,$FB,$54,$EE,$B9,$E1,$DB,$0F,$8B,$F5,$1F,$1B,$EB
	.dc.b	$26,$1F,$CE,$A3,$2F,$F3,$C8,$CB,$F8,$DE,$87,$12,$BE,$7F,$D6,$7D
	.dc.b	$0C,$25,$FF,$44,$8C,$BF,$D1,$A3,$2F,$3A,$4F,$D8,$FA,$CF,$B3,$F4
	.dc.b	$12,$BF,$A6,$C6,$5F,$E9,$F1,$97,$94,$65,$E9,$19,$DD,$AF,$F8,$13
	.dc.b	$3F,$4D,$FF,$51,$F8,$B7,$D6,$7E,$83,$15,$7E,$71,$FC,$D9,$B8,$D4
	.dc.b	$F0,$F0,$33,$3F,$EC,$EA,$E0,$FF,$B4,$27,$4F,$F6,$6F,$2F,$F2,$BE
	.dc.b	$5C,$8C,$10,$F8,$8E,$5F,$8D,$1F,$D7,$7E,$31,$EE,$FF,$5F,$F7,$7E
	.dc.b	$95,$7F,$B6,$DA,$FE,$96,$76,$FF,$B7,$DA,$FE,$A6,$37,$B7,$FB,$D9
	.dc.b	$BB,$7F,$82,$2B,$73,$FE,$FA,$6C,$7F,$83,$27,$75,$FE,$30,$6B,$FE
	.dc.b	$E9,$BF,$D9,$FC,$A7,$F0,$7F,$8F,$FF,$1D,$4F,$E2,$DB,$8F,$D6,$8F
	.dc.b	$F0,$1C,$F0,$7E,$64,$EB,$0F,$B4,$D8,$D9,$FE,$FB,$FA,$CF,$E0,$6F
	.dc.b	$D3,$FA,$CF,$EA,$BB,$9F,$EA,$F0,$9D,$BF,$C2,$DB,$8F,$8E,$B8,$7A
	.dc.b	$5F,$C7,$EF,$8D,$6B,$1F,$F1,$6F,$D5,$BE,$AF,$F9,$27,$D6,$7F,$08
	.dc.b	$CF,$F5,$0D,$CF,$DD,$0A,$E0,$FF,$96,$23,$31,$88,$E3,$E0,$74,$3F
	.dc.b	$C3,$D1,$99,$DD,$E6,$BF,$25,$7D,$F0,$F2,$CF,$E7,$FE,$AB,$F9,$A5
	.dc.b	$7E,$EF,$CB,$FF,$51,$B7,$E1,$35,$E6,$C8,$FC,$E4,$EB,$2F,$88,$DC
	.dc.b	$7F,$CD,$FD,$0C,$7A,$D7,$DB,$8C,$79,$9F,$D1,$3E,$AE,$63,$FA,$5B
	.dc.b	$FE,$7A,$FF,$59,$AD,$BE,$09,$DB,$99,$DD,$FC,$AF,$85,$FF,$13,$BF
	.dc.b	$E8,$08,$F6,$FD,$67,$FB,$76,$B1,$A7,$F9,$8A,$31,$1E,$78,$FF,$32
	.dc.b	$DF,$28,$C1,$FE,$8A,$71,$5D,$3A,$7F,$CD,$6D,$9E,$A6,$F9,$7F,$C1
	.dc.b	$AF,$23,$FD,$46,$E2,$38,$46,$70,$F1,$C7,$D4,$BE,$F8,$79,$2F,$83
	.dc.b	$E9,$C3,$C2,$DC,$3C,$27,$6C,$6D,$F2,$8F,$F5,$1E,$3D,$AB,$84,$79
	.dc.b	$67,$4F,$13,$C9,$7D,$9E,$2D,$75,$8F,$9F,$1F,$2C,$68,$F1,$E1,$DD
	.dc.b	$F8,$98,$F9,$27,$70,$9D,$36,$FB,$61,$E4,$3C,$B3,$A3,$6E,$6B,$9C
	.dc.b	$66,$78,$C7,$69,$E3,$1E,$99,$E3,$39,$38,$C6,$4E,$37,$E7,$3D,$6E
	.dc.b	$6F,$07,$C9,$9E,$2E,$8C,$E1,$F1,$6B,$11,$F3,$0C,$E3,$F7,$63,$33
	.dc.b	$F8,$43,$84,$F5,$8C,$CF,$D8,$9C,$91,$C1,$27,$E7,$9C,$4F,$01,$23
	.dc.b	$83,$A7,$C1,$1C,$63,$EE,$BE,$57,$83,$9C,$7E,$85,$79,$27,$F0,$07
	.dc.b	$7A,$EA,$FD,$B3,$47,$43,$C1,$5D,$F1,$FA,$75,$F3,$3E,$0D,$F1,$9E
	.dc.b	$16,$DF,$B1,$AF,$E5,$DA,$FF,$C0,$0F,$84,$FF,$80,$31,$88,$46,$27
	.dc.b	$FC,$E9,$18,$13,$B9,$D4,$7F,$9E,$5F,$D6,$EC,$DB,$FA,$B6,$7F,$CA
	.dc.b	$AB,$C0,$7E,$FE,$75,$C1,$FB,$48,$CC,$FF,$30,$9C,$C7,$F0,$53,$3A
	.dc.b	$7F,$BD,$A3,$33,$D6,$7C,$CF,$1B,$F0,$D7,$AB,$0F,$2C,$62,$BD,$07
	.dc.b	$8D,$C2,$3B,$9F,$CE,$1C,$B8,$17,$E2,$AE,$D1,$CF,$0F,$63,$E4,$8F
	.dc.b	$45,$72,$8C,$1D,$67,$77,$E6,$6C,$73,$3E,$11,$C0,$F0,$CF,$39,$E1
	.dc.b	$1A,$6E,$1F,$15,$B7,$E6,$8F,$8D,$7D,$9B,$5C,$4F,$03,$C5,$B3,$C4
	.dc.b	$F5,$CF,$16,$C7,$19,$E6,$71,$2C,$71,$8F,$80,$6D,$E3,$1B,$69,$B8
	.dc.b	$F7,$CF,$77,$8C,$ED,$B3,$BC,$3E,$A4,$7D,$E9,$F0,$B6,$7A,$B4,$E2
	.dc.b	$53,$B9,$E5,$32,$7E,$14,$FA,$53,$D9,$C5,$FC,$E7,$94,$65,$FC,$43
	.dc.b	$5E,$38,$7D,$97,$E6,$9C,$1C,$5F,$A2,$3A,$9A,$2B,$EC,$97,$D6,$3E
	.dc.b	$23,$FC,$8D,$AF,$F8,$4B,$D3,$BA,$7F,$84,$AB,$4F,$F8,$62,$33,$8F
	.dc.b	$E7,$27,$2F,$F8,$86,$33,$87,$6C,$37,$D9,$FE,$DC,$9C,$DF,$FB,$2F
	.dc.b	$0F,$F1,$1F,$D6,$63,$D3,$DF,$3A,$31,$3A,$3F,$D0,$55,$C7,$3F,$F1
	.dc.b	$84,$F9,$17,$CE,$BC,$9F,$0F,$D5,$BE,$B3,$53,$9C,$7F,$D1,$D3,$97
	.dc.b	$FC,$90,$7D,$98,$D5,$FF,$AD,$E7,$4F,$FB,$9E,$33,$C1,$FE,$5D,$9C
	.dc.b	$BF,$E5,$48,$CD,$7F,$A9,$E7,$2F,$5C,$3C,$B8,$7A,$7E,$B1,$EE,$7C
	.dc.b	$8F,$75,$F2,$19,$8C,$37,$0F,$F5,$34,$E9,$F4,$E9,$FC,$56,$F9,$C6
	.dc.b	$AB,$E9,$F8,$EF,$5E,$F8,$18,$6D,$7F,$A1,$9B,$87,$FA,$35,$BB,$FF
	.dc.b	$9C,$25,$F0,$7F,$84,$5F,$51,$FE,$AC,$7C,$77,$FE,$C8,$9D,$4E,$61
	.dc.b	$5C,$E7,$3E,$67,$DB,$BE,$A7,$B0,$EA,$E6,$37,$1D,$8D,$D7,$47,$F9
	.dc.b	$B1,$C0,$FF,$50,$FD,$67,$C2,$F1,$3E,$1B,$63,$27,$FA,$9A,$32,$D3
	.dc.b	$F2,$3F,$17,$F9,$6F,$81,$FC,$DB,$63,$27,$F9,$FA,$32,$D3,$FD,$63
	.dc.b	$ED,$FD,$2F,$AC,$FA,$6D,$8C,$9F,$EB,$68,$CB,$4C,$E3,$FE,$BF,$8C
	.dc.b	$B6,$3F,$D7,$BC,$1F,$7B,$E4,$3E,$FB,$63,$27,$FB,$0A,$32,$D3,$FD
	.dc.b	$8B,$F5,$9F,$86,$F9,$AF,$C3,$B6,$32,$7F,$B2,$A3,$2D,$3E,$FF,$D5
	.dc.b	$BE,$9B,$F0,$9F,$58,$F3,$FF,$6D,$39,$AF,$B9,$39,$78,$D7,$0A,$DB
	.dc.b	$63,$27,$08,$CB,$4F,$F6,$EF,$C4,$FF,$03,$F3,$9F,$59,$FA,$08,$C3
	.dc.b	$27,$FD,$C7,$F5,$77,$7F,$F7,$53,$FA,$6F,$AC,$FE,$27,$D5,$6B,$FD
	.dc.b	$72,$F5,$F6,$E4,$7E,$C3,$D5,$7D,$5B,$EB,$33,$97,$E4,$CF,$F0,$46
	.dc.b	$AC,$3F,$95,$DF,$0D,$F7,$DD,$FE,$F9,$8C,$65,$D7,$2F,$91,$8F,$FB
	.dc.b	$3F,$83,$8C,$62,$FF,$D9,$71,$FC,$8F,$EA,$DC,$F4,$FE,$73,$E4,$BE
	.dc.b	$B3,$F9,$55,$FF,$BA,$AB,$79,$7E,$E1,$B9,$7F,$B2,$FE,$AD,$F5,$9C
	.dc.b	$EB,$F8,$7F,$A8,$9C,$C7,$1C,$FF,$A1,$C6,$2F,$E1,$0C,$3A,$E3,$FE
	.dc.b	$DD,$9D,$3F,$EF,$CF,$AC,$7D,$67,$FB,$FA,$FF,$B6,$46,$31,$FF,$28
	.dc.b	$C2,$30,$FF,$95,$1A,$FF,$73,$35,$3F,$D6,$A9,$F5,$8F,$86,$FE,$95
	.dc.b	$7C,$A7,$FB,$2C,$E7,$0F,$E3,$73,$97,$FC,$85,$3C,$AB,$8D,$F5,$BE
	.dc.b	$18,$7C,$0F,$AB,$FE,$41,$6C,$3F,$C9,$5F,$3C,$FA,$B7,$83,$FD,$6E
	.dc.b	$F7,$5A,$9C,$BF,$CB,$23,$38,$2F,$95,$F1,$F2,$FF,$BB,$C6,$2F,$84
	.dc.b	$63,$0E,$77,$E5,$8C,$5F,$F6,$BF,$A7,$FE,$53,$3F,$99,$C3,$CD,$59
	.dc.b	$8E,$4D,$C3,$FA,$8C,$E9,$FE,$83,$5C,$B2,$F8,$17,$E7,$C7,$D2,$A7
	.dc.b	$55,$EF,$C3,$A4,$F8,$6B,$74,$FA,$AF,$6F,$D5,$7F,$A7,$B0,$EF,$18
	.dc.b	$ED,$F5,$36,$32,$79,$A7,$86,$1F,$94,$BE,$93,$9B,$E2,$D3,$D7,$E8
	.dc.b	$F8,$39,$FE,$33,$63,$27,$58,$CB,$4F,$EE,$BF,$48,$FA,$97,$A7,$FE
	.dc.b	$2C,$8C,$3F,$DD,$B6,$C4,$EA,$B8,$B7,$3F,$E1,$97,$C2,$7F,$C0,$3E
	.dc.b	$FF,$F0,$0B,$E3,$87,$F8,$97,$EA,$3F,$C1,$BE,$0F,$FA,$F7,$E9,$7F
	.dc.b	$CA,$D1,$8C,$BF,$D7,$B1,$FE,$59,$FA,$97,$97,$F6,$3E,$19,$3C,$3D
	.dc.b	$B6,$3C,$F0,$FF,$52,$63,$CE,$F7,$18,$AD,$BD,$8F,$F5,$56,$3D,$B5
	.dc.b	$E5,$0B,$E5,$8F,$7C,$73,$18,$9E,$39,$79,$6B,$51,$D4,$E5,$8F,$83
	.dc.b	$2F,$1C,$63,$0E,$F9,$78,$63,$86,$3D,$5D,$18,$BE,$58,$7C,$5C,$7C
	.dc.b	$13,$ED,$C3,$CB,$5A,$8C,$1E,$3C,$3B,$63,$E2,$C7,$B1,$2B,$85,$2B
	.dc.b	$78,$FB,$2F,$85,$7C,$AA,$FE,$39,$7F,$65,$B1,$97,$E1,$9F,$C8,$7E
	.dc.b	$2B,$FC,$6F,$D3,$FC,$A1,$B1,$FE,$FA,$6B,$BF,$67,$F8,$63,$FE,$FE
	.dc.b	$BF,$F4,$47,$C4,$FF,$30,$6C,$67,$E1,$BF,$E0,$0B,$67,$51,$8C,$BE
	.dc.b	$03,$73,$FC,$8B,$6B,$FD,$D9,$F5,$2F,$5B,$FD,$4F,$5A,$C3,$F0,$37
	.dc.b	$BB,$FC,$33,$6F,$84,$6A,$63,$F7,$27,$57,$D2,$3F,$2F,$1C,$F1,$F5
	.dc.b	$E3,$E6,$BF,$C9,$E1,$D2,$F9,$47,$49,$D4,$7E,$3E,$75,$18,$26,$3E
	.dc.b	$99,$D6,$3E,$17,$D3,$7C,$E7,$55,$F9,$6B,$E7,$0A,$72,$3E,$17,$F9
	.dc.b	$14,$7F,$0C,$EE,$FE,$BD,$DF,$FC,$BB,$A3,$FA,$64,$FF,$B4,$E3,$F9
	.dc.b	$8D,$F1,$EF,$C7,$3F,$58,$C5,$73,$32,$2F,$9B,$FC,$3B,$1F,$0B,$FD
	.dc.b	$F3,$27,$B7,$C2,$3F,$83,$B6,$33,$1F,$D1,$BE,$B9,$F2,$19,$A9,$3F
	.dc.b	$E8,$8A,$DB,$FD,$46,$B6,$E7,$C7,$E8,$47,$A7,$25,$7C,$91,$76,$FE
	.dc.b	$A3,$CE,$BF,$F7,$15,$73,$21,$C0,$E6,$7F,$A6,$BE,$9D,$F0,$1E,$D6
	.dc.b	$C6,$63,$E3,$D3,$E5,$FE,$9F,$0F,$D9,$A5,$F0,$7F,$C4,$6D,$C7,$FB
	.dc.b	$2C,$6B,$0E,$77,$CA,$3F,$05,$5B,$BE,$8D,$8C,$D7,$07,$CD,$F0,$31
	.dc.b	$F5,$3D,$3E,$07,$11,$E2,$8B,$8F,$C6,$C5,$58,$C7,$BD,$5B,$FF,$08
	.dc.b	$56,$9F,$F5,$FF,$E0,$4F,$AF,$FC,$13,$63,$31,$F8,$5F,$59,$F1,$BA
	.dc.b	$89,$3A,$3F,$D9,$33,$A7,$07,$F9,$2A,$BA,$36,$7A,$57,$E4,$7E,$99
	.dc.b	$F1,$1F,$E4,$6A,$46,$1F,$F2,$4B,$5F,$F2,$31,$A8,$E1,$19,$D7,$EA
	.dc.b	$B7,$4F,$B8,$DA,$FD,$B7,$D4,$7F,$22,$7F,$CB,$F1,$97,$A1,$FE,$E4
	.dc.b	$71,$1B,$AF,$C0,$FE,$3A,$F9,$BF,$D5,$36,$FF,$DD,$6D,$7F,$CD,$7C
	.dc.b	$DF,$D7,$2B,$F1,$2D,$C7,$F2,$0D,$9F,$E1,$4D,$55,$FA,$EA,$F0,$F4
	.dc.b	$7E,$E6,$B4,$6E,$75,$7F,$C1,$FA,$9F,$29,$FC,$1B,$B7,$F8,$A3,$63
	.dc.b	$39,$7F,$99,$AB,$F4,$5F,$4F,$F8,$78,$FF,$41,$FD,$47,$FB,$15,$D1
	.dc.b	$FC,$7D,$C3,$AE,$4F,$E1,$93,$F8,$38,$FF,$45,$B7,$5F,$D4,$B6,$FB
	.dc.b	$F1,$78,$FD,$1F,$E2,$9F,$1B,$FA,$EB,$FD,$7B,$6B,$F6,$CD,$C1,$F1
	.dc.b	$4F,$1F,$FD,$71,$1E,$61,$3B,$6D,$76,$3F,$94,$D7,$1F,$84,$C4,$F1
	.dc.b	$AE,$07,$79,$D4,$7F,$2D,$FB,$6F,$95,$F3,$32,$30,$43,$FD,$99,$C5
	.dc.b	$FD,$0E,$35,$18,$E1,$57,$D3,$F8,$7F,$A9,$F0,$F2,$BF,$9D,$46,$5F
	.dc.b	$E7,$91,$97,$94,$7D,$BF,$B8,$FB,$9F,$71,$27,$FA,$14,$65,$FE,$89
	.dc.b	$19,$7C,$51,$97,$C9,$19,$E0,$FE,$A7,$8A,$BF,$A7,$FC,$33,$F2,$7F
	.dc.b	$F5,$A6,$1F,$EC,$E6,$BF,$EE,$B8,$D7,$93,$F8,$FF,$07,$8D,$6B,$1F
	.dc.b	$F4,$7D,$6A,$F9,$5E,$DF,$EB,$5F,$07,$F9,$7F,$94,$91,$82,$1F,$2D
	.dc.b	$CB,$F3,$63,$F1,$7F,$4D,$23,$EE,$3F,$D8,$7E,$A7,$F7,$9C,$FF,$8C
	.dc.b	$8D,$4F,$FB,$8B,$99,$1B,$E8,$CE,$9F,$94,$8B,$7F,$D2,$70,$E7,$87
	.dc.b	$F8,$0B,$E9,$DF,$4F,$FB,$17,$9C,$7F,$A0,$9A,$FE,$19,$DB,$98,$E1
	.dc.b	$72,$30,$7F,$BF,$3C,$97,$D3,$7D,$46,$C6,$63,$C1,$E2,$E5,$91,$93
	.dc.b	$C5,$1F,$E0,$3F,$A6,$7C,$15,$24,$FE,$AF,$E0,$BF,$BE,$AD,$BF,$B1
	.dc.b	$D7,$08,$FD,$83,$5D,$6D,$FD,$B6,$31,$3F,$DA,$AB,$9B,$CA,$FB,$DF
	.dc.b	$03,$FC,$59,$5C,$2F,$FB,$B5,$F0,$7F,$BC,$7D,$2B,$C4,$FF,$0E,$36
	.dc.b	$FF,$81,$35,$FD,$9F,$D3,$3E,$6F,$75,$27,$6F,$F2,$79,$D4,$7F,$20
	.dc.b	$E8,$7D,$17,$F2,$88,$FC,$F3,$75,$FC,$DB,$72,$FF,$1C,$7C,$93,$E2
	.dc.b	$BF,$53,$39,$B6,$1D,$3C,$CC,$9F,$CD,$EF,$84,$FF,$1B,$F8,$47,$AB
	.dc.b	$F4,$63,$81,$FE,$51,$FA,$4F,$F6,$BB,$FE,$53,$C7,$CD,$18,$8D,$8A
	.dc.b	$E7,$A6,$F1,$FF,$11,$46,$27,$75,$C6,$F9,$F0,$FF,$89,$DF,$85,$5B
	.dc.b	$6E,$3E,$57,$77,$D4,$F3,$57,$1C,$3B,$D6,$DB,$8E,$E9,$8F,$86,$7F
	.dc.b	$C6,$CD,$8F,$F3,$86,$3E,$19,$E3,$1E,$8C,$3B,$3E,$03,$A4,$ED,$E3
	.dc.b	$87,$85,$D1,$C4,$DB,$97,$E4,$36,$FC,$51,$E6,$AE,$EE,$B0,$E2,$7A
	.dc.b	$E7,$67,$0B,$F1,$4E,$CF,$01,$31,$F9,$75,$CA,$31,$8F,$C2,$C3,$BD
	.dc.b	$F2,$39,$E1,$DA,$33,$8F,$A6,$73,$1E,$69,$F1,$4E,$4F,$64,$64,$E3
	.dc.b	$87,$87,$1F,$B7,$8F,$88,$F9,$31,$9C,$7C,$91,$EF,$AE,$2F,$E4,$31
	.dc.b	$F0,$B9,$38,$1C,$8E,$87,$CB,$32,$7C,$5B,$F0,$92,$BE,$FB,$F1,$A3
	.dc.b	$8E,$3E,$CC,$38,$BF,$0E,$39,$DF,$89,$EE,$FC,$9C,$7E,$0C,$79,$4F
	.dc.b	$43,$F1,$6B,$17,$E1,$9F,$43,$E1,$30,$FA,$4E,$31,$F6,$7A,$FF,$77
	.dc.b	$1F,$7D,$B7,$FA,$F6,$DF,$EF,$5D,$CF,$FB,$37,$E0,$BF,$D9,$D1,$88
	.dc.b	$46,$36,$FF,$32,$B6,$3F,$D3,$86,$8F,$F1,$34,$69,$FE,$07,$B3,$E9
	.dc.b	$1C,$1F,$B1,$9F,$01,$A8,$CB,$FC,$28,$CE,$9F,$C8,$EB,$95,$71,$9F
	.dc.b	$04,$E6,$B8,$BF,$CA,$63,$8E,$DF,$EB,$1B,$E3,$1D,$4E,$31,$9A,$E2
	.dc.b	$FA,$6B,$A5,$F7,$C7,$F8,$46,$1C,$A7,$B6,$1C,$8F,$F7,$B4,$66,$F9
	.dc.b	$1D,$6F,$94,$FB,$DB,$5D,$A3,$C9,$7D,$B0,$F1,$C7,$CE,$8C,$57,$0B
	.dc.b	$E9,$3C,$0E,$85,$BE,$87,$9A,$B8,$C6,$67,$8C,$E6,$38,$B4,$E3,$1E
	.dc.b	$A8,$F6,$DF,$91,$FF,$6E,$57,$4A,$E1,$5D,$0E,$13,$FD,$4A,$B8,$3F
	.dc.b	$50,$E7,$7E,$B6,$9E,$52,$D7,$46,$D7,$4C,$3E,$74,$7C,$53,$D3,$19
	.dc.b	$BE,$F1,$F5,$4F,$15,$F7,$69,$DC,$B7,$DD,$B5,$CE,$78,$5F,$82,$7E
	.dc.b	$09,$E0,$3C,$55,$F2,$1B,$5F,$0D,$A7,$A8,$A7,$62,$CF,$57,$E3,$19
	.dc.b	$3B,$CF,$56,$CF,$56,$9D,$65,$F4,$6F,$A4,$FD,$6E,$FF,$E6,$48,$C6
	.dc.b	$3F,$EB,$0F,$35,$C6,$F8,$ED,$CF,$E7,$2B,$79,$7F,$9C,$5B,$87,$3C
	.dc.b	$77,$2C,$79,$E1,$8C,$BC,$75,$AA,$7D,$0B,$E2,$F0,$DC,$FC,$9E,$E3
	.dc.b	$E9,$33,$52,$7F,$D6,$55,$BB,$FF,$51,$DE,$E7,$35,$DD,$7C,$65,$3C
	.dc.b	$EB,$E8,$3F,$C7,$CF,$FA,$F4,$D9,$63,$83,$FE,$D6,$38,$76,$63,$0F
	.dc.b	$F2,$61,$FC,$7A,$71,$BF,$F1,$CA,$D9,$CC,$FF,$20,$70,$3C,$1F,$8C
	.dc.b	$D8,$FA,$9C,$5E,$2C,$3E,$85,$68,$93,$AC,$2C,$FF,$B1,$2B,$ED,$CF
	.dc.b	$2F,$8C,$FA,$BC,$7F,$7A,$68,$8A,$E1,$B3,$C4,$FF,$70,$7A,$CF,$53
	.dc.b	$ED,$36,$3E,$26,$CF,$9D,$CD,$49,$E2,$FD,$09,$D4,$60,$99,$7C,$AB
	.dc.b	$E8,$FF,$9D,$E3,$EB,$5F,$38,$FF,$6F,$7D,$BF,$B3,$F4,$AF,$E9,$B5
	.dc.b	$F9,$BF,$55,$39,$8F,$83,$18,$7F,$41,$3C,$2B,$9C,$60,$FC,$C7,$C8
	.dc.b	$BE,$82,$9D,$FE,$36,$CF,$F9,$E6,$74,$D9,$D3,$4E,$FF,$33,$E7,$E7
	.dc.b	$F4,$36,$32,$79,$23,$2D,$33,$87,$FC,$00,$C7,$F4,$1C,$DF,$DA,$38
	.dc.b	$D0,$C3,$A5,$6F,$E1,$3E,$37,$DD,$7C,$76,$E3,$F0,$9B,$3B,$8F,$8B
	.dc.b	$3F,$37,$E8,$FE,$77,$75,$9F,$92,$DA,$F2,$37,$3F,$26,$5F,$C1,$AB
	.dc.b	$FA,$3C,$FF,$01,$F1,$31,$6E,$47,$E0,$CF,$5F,$57,$30,$5F,$EB,$B0
	.dc.b	$E3,$87,$0B,$D4,$E6,$EE,$5F,$E0,$CA,$D6,$1C,$A3,$FC,$1F,$5F,$E1
	.dc.b	$EA,$FE,$33,$F2,$EF,$A3,$7C,$0E,$1B,$87,$E2,$3E,$C3,$13,$FC,$6F
	.dc.b	$CB,$7C,$2D,$33,$5D,$CF,$4E,$5F,$EB,$78,$FC,$9F,$C4,$49,$D4,$B2
	.dc.b	$F1,$77,$E6,$3F,$DB,$10,$FD,$A9,$E3,$FF,$2B,$F3,$5F,$21,$EA,$9C
	.dc.b	$E3,$D3,$0F,$F6,$6C,$EC,$91,$E0,$C3,$A6,$18,$AE,$F6,$FA,$EE,$86
	.dc.b	$78,$BF,$C6,$F7,$E3,$8F,$E7,$9F,$44,$F8,$2A,$7F,$41,$F6,$DF,$65
	.dc.b	$8D,$3C,$95,$B6,$D6,$E7,$59,$FF,$2C,$AD,$3F,$E5,$00,$E1,$FF,$29
	.dc.b	$B5,$F9,$ED,$7F,$13,$F7,$5E,$BB,$F8,$8D,$86,$7F,$E5,$2F,$49,$97
	.dc.b	$E4,$EF,$79,$FF,$59,$C3,$85,$FE,$29,$B3,$FE,$42,$AE,$18,$7F,$44
	.dc.b	$9C,$BE,$2B,$E4,$7C,$5E,$AB,$7F,$5F,$E8,$7F,$3E,$DA,$FE,$D1,$F7
	.dc.b	$4F,$80,$FE,$D5,$5A,$9F,$EA,$2D,$CF,$F3,$AD,$BF,$C8,$DF,$17,$A1
	.dc.b	$2B,$89,$DA,$FA,$BD,$B8,$24,$FE,$2F,$CB,$FE,$DD,$3E,$09,$CB,$67
	.dc.b	$5E,$AE,$32,$F1,$D5,$9F,$D1,$DF,$29,$CC,$7C,$1F,$8F,$F6,$9F,$E0
	.dc.b	$7F,$B9,$FC,$4C,$64,$98,$FC,$09,$D7,$1F,$09,$CD,$FC,$29,$CC,$60
	.dc.b	$FD,$9E,$AF,$8F,$F9,$67,$7D,$EC,$FD,$38,$B8,$FF,$7E,$CB,$EE,$E3
	.dc.b	$F3,$A7,$FA,$97,$DA,$7F,$88,$BE,$79,$F4,$1F,$D7,$B0,$FE,$71,$39
	.dc.b	$8C,$5F,$F2,$68,$C7,$A7,$CB,$2F,$E0,$78,$72,$8F,$F1,$47,$D1,$7F
	.dc.b	$89,$F8,$7F,$D7,$1F,$41,$FC,$32,$31,$9F,$FA,$E2,$3F,$87,$FA,$6F
	.dc.b	$A3,$FF,$56,$F9,$69,$E1,$E9,$FF,$0F,$BE,$07,$F2,$EC,$39,$D6,$E3
	.dc.b	$13,$B8,$C4,$30,$ED,$A7,$92,$37,$95,$58,$74,$C7,$C3,$39,$BE,$35
	.dc.b	$E2,$37,$8F,$7A,$DC,$66,$F9,$3F,$D1,$2B,$8E,$3E,$49,$CC,$F2,$C3
	.dc.b	$C5,$18,$AD,$DF,$4B,$E1,$87,$7C,$3D,$27,$8E,$B6,$E2,$77,$87,$4A
	.dc.b	$E1,$3F,$0F,$2F,$9F,$87,$53,$D3,$1F,$EE,$8C,$7E,$BB,$6B,$E9,$36
	.dc.b	$FF,$8F,$2F,$84,$FF,$2C,$FD,$06,$1B,$1F,$C8,$DA,$FF,$BD,$1A,$FE
	.dc.b	$17,$D3,$7C,$27,$4C,$51,$89,$FE,$59,$18,$9D,$4E,$67,$D0,$EF,$4F
	.dc.b	$9C,$DC,$FB,$57,$DF,$FA,$3F,$F2,$65,$FD,$0A,$D3,$F8,$D6,$BF,$84
	.dc.b	$6C,$70,$8F,$BD,$1A,$C3,$C9,$18,$AE,$DE,$E7,$FB,$4B,$0F,$0C,$E4
	.dc.b	$98,$FE,$5B,$1E,$8F,$A2,$FC,$84,$9F,$C8,$4E,$AB,$9D,$6E,$79,$C7
	.dc.b	$F7,$A9,$D4,$7A,$EB,$95,$6C,$F0,$5F,$96,$7F,$D4,$9E,$DB,$E8,$FF
	.dc.b	$C0,$7C,$CE,$3F,$F0,$EF,$2F,$F0,$8B,$CC,$63,$9F,$FC,$13,$1F,$D5
	.dc.b	$A3,$FC,$47,$ED,$31,$E9,$3F,$E6,$1C,$78,$AC,$3A,$3F,$C3,$30,$E7
	.dc.b	$8F,$F9,$26,$F8,$CF,$FA,$6B,$6F,$F4,$A3,$F9,$26,$BF,$DC,$5A,$F9
	.dc.b	$1A,$FF,$87,$9D,$CF,$F8,$B9,$B9,$67,$E0,$3C,$BC,$6F,$A3,$7B,$DE
	.dc.b	$8C,$7F,$8B,$7C,$AE,$BE,$E1,$F2,$54,$D6,$69,$A7,$EA,$2B,$89,$E9
	.dc.b	$71,$1B,$8C,$1F,$E2,$57,$0F,$C1,$8F,$2E,$1F,$A6,$8F,$F1,$57,$D2
	.dc.b	$3E,$93,$D3,$F7,$1F,$EA,$39,$FF,$1C,$E8,$8C,$47,$FB,$12,$B8,$3C
	.dc.b	$AB,$CD,$18,$8F,$F0,$E3,$5F,$BE,$D7,$FC,$92,$EF,$7F,$F2,$CB,$77
	.dc.b	$CD,$7C,$CF,$95,$FB,$B1,$C3,$E3,$FC,$FF,$33,$31,$F5,$C7,$CF,$C0
	.dc.b	$F9,$09,$1F,$C4,$A3,$2D,$8E,$04,$7E,$0D,$FF,$94,$9F,$F2,$87,$D2
	.dc.b	$BE,$73,$EE,$7B,$99,$9F,$F7,$3B,$74,$FF,$39,$D6,$E7,$8D,$7D,$B8
	.dc.b	$C4,$7F,$90,$5A,$FE,$91,$AF,$F9,$B9,$DC,$F4,$9F,$C6,$46,$33,$FF
	.dc.b	$39,$DB,$E8,$3F,$21,$F4,$BF,$E6,$CC,$7F,$CD,$1F,$4B,$FE,$58,$6F
	.dc.b	$A1,$AF,$72,$47,$F3,$6A,$E0,$FF,$95,$5A,$FE,$E1,$AF,$FA,$09,$DE
	.dc.b	$9F,$E9,$26,$E5,$9A,$FD,$16,$1F,$CC,$6B,$FD,$2F,$F4,$8F,$A5,$FD
	.dc.b	$33,$71,$FF,$3F,$35,$FE,$0C,$7F,$AA,$3C,$7F,$EE,$F5,$FE,$A0,$7F
	.dc.b	$BC,$C7,$F0,$C6,$BF,$E9,$66,$BC,$7A,$7F,$59,$5F,$BA,$8C,$1F,$E8
	.dc.b	$D9,$EB,$97,$FA,$7A,$75,$E1,$7F,$15,$F9,$6F,$E2,$6D,$BF,$F7,$B4
	.dc.b	$7E,$FF,$E5,$3F,$05,$7F,$EA,$6F,$A5,$71,$7E,$ED,$B7,$FD,$91,$AF
	.dc.b	$FA,$0D,$AF,$F2,$6A,$FB,$8F,$FA,$85,$AF,$C4,$0E,$34,$FE,$45,$EB
	.dc.b	$7F,$11,$F5,$7F,$D8,$B1,$FE,$C9,$F2,$9F,$6B,$FC,$B3,$E3,$3F,$B5
	.dc.b	$BB,$D2,$65,$FD,$E3,$0F,$F6,$6B,$FC,$E1,$AF,$FB,$09,$AF,$FB,$71
	.dc.b	$EF,$19,$39,$C7,$DA,$AF,$EF,$3F,$0F,$F2,$BC,$8F,$F2,$C4,$FF,$B2
	.dc.b	$38,$3F,$A1,$7D,$36,$4D,$7B,$12,$75,$49,$D4,$62,$7A,$7C,$57,$5F
	.dc.b	$89,$93,$C2,$32,$F2,$8C,$BF,$8E,$C3,$FA,$94,$F9,$7E,$97,$CD,$F4
	.dc.b	$D2,$BC,$31,$97,$C7,$19,$79,$64,$9D,$46,$27,$5C,$3F,$D5,$E7,$FA
	.dc.b	$EF,$C3,$3E,$9B,$F9,$CC,$FF,$6B,$B6,$4F,$92,$FD,$6C,$61,$B7,$C7
	.dc.b	$CD,$E7,$87,$2E,$25,$AB,$4F,$FB,$D9,$AF,$A3,$C5,$FA,$9F,$15,$F5
	.dc.b	$7C,$A9,$3F,$42,$32,$FD,$28,$CB,$F0,$23,$F2,$1E,$7F,$F6,$1F,$A6
	.dc.b	$7D,$15,$33,$5F,$EF,$BE,$9F,$EC,$6F,$E3,$EB,$FB,$04,$62,$3F,$DC
	.dc.b	$CD,$7F,$25,$1C,$9F,$D1,$46,$F8,$3E,$44,$2F,$F8,$86,$9F,$E0,$EF
	.dc.b	$A6,$7D,$37,$EE,$9E,$11,$9C,$71,$32,$FF,$29,$5A,$7E,$9B,$78,$8F
	.dc.b	$48,$F9,$9E,$6B,$E9,$BF,$3D,$10,$CB,$F8,$08,$D4,$60,$87,$E4,$E7
	.dc.b	$F8,$B3,$FE,$FA,$F8,$AF,$F0,$67,$D2,$3E,$36,$9F,$E0,$FF,$8F,$FF
	.dc.b	$76,$57,$E0,$37,$F8,$71,$88,$FF,$18,$9C,$CF,$F3,$AB,$97,$FC,$6A
	.dc.b	$D7,$F1,$26,$CF,$E2,$0F,$38,$EB,$7E,$4F,$6B,$FC,$41,$F4,$DF,$C5
	.dc.b	$2F,$F7,$9F,$4D,$FD,$C1,$FD,$73,$6F,$EB,$CE,$E7,$F8,$E3,$63,$3E
	.dc.b	$47,$43,$F9,$8D,$7E,$32,$BF,$35,$F2,$AF,$33,$F8,$67,$C3,$C9,$CC
	.dc.b	$7E,$BA,$75,$18,$21,$FC,$07,$E8,$1F,$2F,$FE,$39,$F8,$9F,$F0,$5E
	.dc.b	$7F,$AA,$8E,$34,$FA,$67,$17,$C6,$9C,$C6,$27,$86,$5F,$2B,$0E,$71
	.dc.b	$DA,$31,$EE,$31,$FD,$BE,$7C,$B2,$FA,$2D,$BC,$DF,$0A,$B1,$FB,$EC
	.dc.b	$3A,$D6,$E0,$D3,$D3,$1A,$AE,$53,$A7,$EF,$E3,$E9,$8C,$1F,$C2,$9C
	.dc.b	$57,$2C,$7D,$13,$A3,$FD,$7C,$E3,$2F,$11,$2F,$BB,$C7,$1F,$4D,$7C
	.dc.b	$BC,$BD,$F5,$E2,$6C,$7B,$6F,$BE,$1C,$63,$E5,$1E,$59,$FA,$55,$C6
	.dc.b	$3E,$89,$C6,$EB,$E6,$7C,$38,$FB,$27,$84,$FA,$F2,$F8,$CF,$13,$DD
	.dc.b	$3B,$7C,$07,$AA,$36,$F5,$37,$7E,$A7,$DC,$6F,$1F,$5D,$FC,$43,$C3
	.dc.b	$97,$C7,$9F,$03,$EF,$CB,$ED,$4F,$99,$C3,$F6,$9C,$CF,$90,$F5,$57
	.dc.b	$67,$9C,$F6,$23,$83,$DD,$3C,$9F,$39,$CB,$2F,$41,$F2,$30,$F6,$65
	.dc.b	$F0,$E7,$E0,$57,$D5,$C7,$D5,$1C,$9C,$61,$E7,$3D,$44,$8F,$0B,$97
	.dc.b	$D5,$3E,$3C,$BE,$69,$F3,$AF,$C0,$78,$5C,$E5,$F9,$1B,$F6,$DF,$85
	.dc.b	$F9,$CF,$63,$AB,$E1,$7D,$2E,$9E,$07,$CB,$C3,$D2,$FC,$73,$24,$AF
	.dc.b	$DE,$C7,$E5,$7C,$FF,$E1,$73,$F8,$96,$BF,$85,$6C,$FF,$04,$73,$1F
	.dc.b	$DE,$3E,$23,$FB,$CC,$62,$11,$8A,$FE,$76,$D8,$FF,$76,$B5,$FE,$72
	.dc.b	$FE,$83,$C3,$7D,$37,$F8,$36,$BF,$8D,$D7,$73,$F5,$67,$F8,$1E,$35
	.dc.b	$8F,$68,$CD,$72,$EE,$FD,$5C,$71,$BF,$E4,$B3,$98,$E6,$75,$D7,$38
	.dc.b	$7F,$87,$5F,$0C,$7B,$0F,$E1,$53,$C6,$33,$8F,$A6,$BC,$75,$EE,$3C
	.dc.b	$6E,$1F,$79,$C9,$ED,$8F,$A6,$BC,$38,$FF,$10,$AF,$24,$FF,$5E,$C3
	.dc.b	$C0,$EC,$C5,$F2,$9F,$05,$F2,$7F,$90,$1C,$8F,$F1,$7D,$72,$7C,$D8
	.dc.b	$EC,$F5,$BE,$8C,$3E,$4B,$F0,$6B,$A1,$A2,$E1,$E3,$8E,$B8,$78,$DF
	.dc.b	$7C,$E9,$EF,$3E,$8A,$EB,$8F,$C9,$6C,$FC,$57,$E1,$3E,$19,$E0,$FB
	.dc.b	$0C,$1E,$08,$DC,$FF,$38,$C7,$DD,$3F,$43,$1F,$75,$FE,$1D,$B1,$F2
	.dc.b	$23,$AE,$1E,$D6,$9E,$D8,$FC,$86,$1D,$9F,$31,$E3,$C3,$C1,$8F,$C1
	.dc.b	$8F,$A0,$70,$34,$7C,$D3,$F2,$27,$E3,$70,$F1,$BF,$22,$FC,$2F,$94
	.dc.b	$F7,$1C,$A3,$6F,$E7,$63,$E2,$C7,$A6,$37,$85,$7C,$4D,$C7,$E2,$C7
	.dc.b	$99,$AF,$CC,$37,$5F,$06,$33,$87,$D7,$9F,$4B,$6F,$FC,$33,$2F,$A4
	.dc.b	$7C,$EF,$FA,$1F,$87,$FD,$11,$39,$8C,$69,$FE,$82,$EE,$71,$BE,$9B
	.dc.b	$3A,$7F,$A1,$72,$E7,$7F,$E8,$56,$C7,$FA,$73,$E5,$F7,$97,$5A,$FF
	.dc.b	$59,$C7,$29,$7D,$33,$E4,$70,$DA,$FF,$5F,$7C,$02,$F3,$52,$73,$EC
	.dc.b	$DC,$3F,$D6,$96,$8E,$D2,$FA,$6A,$7A,$97,$D3,$7F,$B9,$1F,$F3,$11
	.dc.b	$FE,$5A,$38,$18,$BF,$F5,$2B,$87,$FC,$CA,$D7,$FC,$C2,$1C,$3C,$E7
	.dc.b	$FD,$C3,$1C,$78,$7F,$D7,$0B,$E5,$9E,$17,$9A,$3C,$06,$5F,$97,$BB
	.dc.b	$E3,$A9,$F5,$38,$FE,$BE,$38,$BF,$A9,$39,$F5,$9F,$3F,$9D,$3F,$DA
	.dc.b	$2F,$1E,$5F,$BF,$7F,$EE,$D5,$F2,$79,$74,$D8,$D3,$6B,$FC,$C8,$FD
	.dc.b	$DF,$79,$ED,$D3,$E8,$70,$FB,$EF,$E6,$5A,$B7,$18,$6C,$7F,$A4,$9A
	.dc.b	$FC,$E6,$BF,$E9,$06,$BF,$ED,$A7,$8F,$0F,$CE,$A7,$C0,$F8,$EF,$E6
	.dc.b	$C6,$1F,$BD,$BB,$57,$CD,$FC,$DF,$6B,$EE,$56,$A7,$2F,$08,$CE,$7F
	.dc.b	$EF,$CA,$E1,$3F,$45,$AF,$FB,$AF,$E6,$1F,$4D,$FA,$66,$C7,$3F,$37
	.dc.b	$13,$23,$B7,$9F,$70,$FF,$41,$47,$ED,$F1,$FF,$5F,$3E,$3F,$A6,$F4
	.dc.b	$76,$7A,$5B,$19,$3D,$11,$96,$99,$AE,$A3,$55,$FE,$D6,$33,$18,$BE
	.dc.b	$B1,$89,$7C,$3F,$C9,$FA,$6F,$94,$DC,$3E,$43,$6B,$83,$78,$BF,$55
	.dc.b	$1E,$1B,$7D,$37,$D3,$F1,$3E,$A3,$70,$FA,$4D,$AF,$13,$6F,$FD,$FF
	.dc.b	$E3,$7F,$16,$6D,$FF,$1C,$E6,$7D,$33,$DB,$FE,$2D,$7B,$CB,$FC,$4B
	.dc.b	$7E,$4C,$BF,$C5,$2D,$73,$1F,$EE,$87,$FC,$63,$19,$7F,$C6,$CF,$7B
	.dc.b	$FD,$6C,$FE,$8A,$BB,$57,$E8,$E7,$2F,$D2,$F4,$5F,$4C,$F8,$5F,$E6
	.dc.b	$71,$87,$F2,$DF,$4D,$89,$D6,$3F,$AE,$F3,$1F,$13,$E3,$C7,$FD,$B2
	.dc.b	$D8,$CC,$FA,$5B,$3F,$C7,$39,$7F,$49,$DD,$FC,$2C,$ED,$DF,$6F,$F9
	.dc.b	$36,$8F,$A6,$79,$5F,$CF,$FD,$0F,$E2,$35,$A9,$CB,$FC,$0A,$33,$97
	.dc.b	$E4,$0F,$F6,$9F,$8B,$E7,$8C,$C7,$F2,$E6,$BF,$EE,$08,$F9,$D8,$78
	.dc.b	$7B,$7E,$2E,$1F,$A4,$AE,$97,$DB,$9F,$F1,$B7,$BC,$7F,$89,$FD,$33
	.dc.b	$E0,$FF,$AB,$36,$3F,$A1,$7A,$0F,$71,$F2,$7A,$B9,$5F,$81,$BE,$19
	.dc.b	$FF,$44,$0E,$1F,$E8,$ED,$44,$78,$E5,$FD,$DF,$EA,$9F,$13,$DA,$BF
	.dc.b	$6B,$29,$CD,$FF,$2C,$FA,$7F,$81,$3F,$97,$6C,$7F,$94,$33,$FC,$F3
	.dc.b	$BE,$0F,$EE,$93,$F1,$F0,$FE,$B6,$D7,$CA,$1C,$3E,$86,$B9,$8F,$CA
	.dc.b	$FD,$DB,$E4,$FA,$C6,$1F,$F0,$17,$D3,$FF,$6C,$AF,$C0,$37,$2F,$D7
	.dc.b	$7D,$47,$F6,$4C,$BF,$A3,$CE,$5F,$1B,$C6,$BF,$B8,$B5,$C9,$F9,$F7
	.dc.b	$0F,$CE,$79,$2F,$9E,$7D,$0F,$78,$F1,$9F,$E2,$08,$CF,$AF,$E8,$E5
	.dc.b	$98,$7F,$5A,$F4,$3F,$C2,$CF,$9A,$33,$7F,$81,$C5,$1E,$32,$4F,$F0
	.dc.b	$78,$CB,$FE,$0F,$C1,$F9,$D7,$4B,$F8,$B7,$E3,$8C,$69,$FC,$4F,$0F
	.dc.b	$F1,$06,$1F,$1B,$DA,$FF,$17,$7D,$4F,$E5,$A3,$04,$C7,$C9,$3F,$14
	.dc.b	$91,$E3,$7E,$07,$D5,$FF,$90,$63,$86,$1F,$CD,$33,$4E,$6F,$EC,$FC
	.dc.b	$B7,$DC,$3C,$3E,$5D,$C3,$EC,$C7,$D3,$23,$F0,$5F,$E4,$DE,$1F,$F4
	.dc.b	$EF,$AA,$FF,$27,$FD,$37,$D4,$F7,$FF,$B9,$46,$39,$3F,$C1,$59,$7F
	.dc.b	$4F,$8C,$4E,$B1,$FF,$1C,$5F,$17,$FC,$1F,$F5,$3F,$E4,$FE,$3F,$F7
	.dc.b	$47,$3F,$E7,$23,$18,$FF,$BA,$23,$F3,$FE,$CB,$EA,$BF,$A9,$FA,$1F
	.dc.b	$9B,$AF,$DB,$DF,$08,$C4,$23,$17,$D7,$D8,$FC,$0C,$F5,$3F,$7F,$87
	.dc.b	$5C,$BC,$71,$D2,$77,$87,$4C,$39,$CE,$72,$DA,$C3,$93,$D4,$F1,$5F
	.dc.b	$2C,$7C,$55,$C2,$F9,$D7,$1C,$BD,$B5,$A8,$E5,$3A,$37,$87,$6C,$7C
	.dc.b	$98,$F5,$AF,$21,$C1,$C3,$D7,$2F,$04,$62,$FA,$3A,$BE,$98,$78,$70
	.dc.b	$EB,$97,$F6,$BB,$FA,$B8,$FF,$9B,$9A,$F9,$CF,$F3,$8F,$C4,$7F,$9F
	.dc.b	$F9,$3F,$B9,$36,$3F,$99,$35,$DF,$C0,$FF,$01,$7F,$9B,$57,$F8,$63
	.dc.b	$E2,$3C,$1E,$C9,$A4,$9D,$C6,$5F,$8E,$DC,$BE,$C3,$73,$FE,$53,$F5
	.dc.b	$5F,$E6,$CA,$FA,$53,$97,$F1,$6D,$7F,$14,$D8,$E1,$1D,$AF,$89,$1F
	.dc.b	$C0,$46,$AF,$C3,$39,$6C,$EB,$E0,$7F,$9D,$C6,$09,$87,$E8,$F0,$F0
	.dc.b	$4E,$A3,$17,$CD,$B7,$CC,$C1,$EA,$8A,$B0,$F2,$4F,$F5,$EE,$2F,$E5
	.dc.b	$F3,$FD,$33,$AB,$FC,$7F,$18,$8F,$DC,$F3,$7F,$AB,$27,$FC,$87,$CF
	.dc.b	$FE,$75,$FA,$AE,$1C,$BC,$EB,$4D,$BC,$8C,$3A,$3F,$E5,$8F,$81,$5E
	.dc.b	$A3,$FA,$B7,$C2,$70,$E4,$7D,$53,$C6,$C3,$6B,$FD,$33,$B7,$FA,$CD
	.dc.b	$B7,$AA,$91,$83,$7E,$5F,$2F,$96,$91,$82,$1F,$E7,$0C,$7C,$31,$A8
	.dc.b	$FF,$12,$61,$FE,$02,$77,$C1,$FC,$CF,$25,$6E,$3F,$CA,$D5,$BA,$FF
	.dc.b	$16,$7D,$3B,$EA,$3D,$DF,$53,$FE,$B6,$9F,$EA,$0D,$77,$1F,$16,$78
	.dc.b	$46,$BD,$8E,$26,$66,$46,$48,$72,$3F,$D7,$F2,$F9,$0F,$0F,$CE,$49
	.dc.b	$CC,$77,$30,$FC,$3C,$13,$98,$D9,$E8,$77,$97,$F9,$D6,$FB,$4F,$FB
	.dc.b	$3B,$EA,$5F,$3D,$F7,$FD,$DF,$F7,$1C,$FF,$B4,$9A,$FF,$94,$67,$6F
	.dc.b	$19,$E1,$EE,$48,$C1,$0F,$F2,$C5,$72,$C3,$FD,$0C,$DE,$8F,$F1,$4E
	.dc.b	$1D,$2B,$FC,$BD,$F5,$1F,$E6,$67,$FD,$C7,$F5,$1F,$E4,$C6,$FB,$7F
	.dc.b	$95,$9D,$47,$0F,$7F,$F3,$15,$B8,$FF,$34,$37,$2F,$F4,$B5,$70,$6D
	.dc.b	$7F,$A5,$69,$3D,$1F,$F1,$E9,$88,$D4,$66,$3F,$C8,$3F,$4E,$FA,$8F
	.dc.b	$D6,$F8,$1F,$EC,$4F,$B8,$BF,$18,$7F,$59,$5B,$BF,$F7,$FF,$8D,$FD
	.dc.b	$9B,$3F,$F5,$54,$E6,$7F,$D0,$0D,$8F,$F5,$74,$EF,$7F,$F2,$8B,$6B
	.dc.b	$FC,$CE,$D7,$FD,$9F,$D3,$E0,$EF,$F2,$F8,$2F,$97,$F2,$FD,$0C,$BF
	.dc.b	$F7,$D4,$7F,$01,$FA,$97,$37,$F0,$3F,$9E,$FB,$15,$FE,$74,$6A,$9D
	.dc.b	$BF,$E7,$8D,$7F,$D8,$92,$FA,$AE,$97,$2B,$5B,$7F,$9A,$AB,$51,$83
	.dc.b	$FD,$03,$EB,$FF,$B3,$70,$FF,$70,$FC,$87,$F5,$1F,$AA,$FE,$36,$DA
	.dc.b	$E0,$7F,$9E,$8B,$5C,$70,$FF,$45,$CE,$E7,$FD,$F6,$DB,$FF,$6D,$56
	.dc.b	$E7,$C3,$5D,$BC,$D6,$14,$FE,$4D,$5B,$E6,$70,$3D,$9F,$9F,$F5,$53
	.dc.b	$2F,$65,$6A,$7E,$99,$98,$FE,$65,$38,$77,$5F,$ED,$D6,$F8,$DF,$68
	.dc.b	$FB,$7F,$21,$23,$24,$3F,$9E,$3F,$80,$FA,$9F,$C0,$FD,$54,$9E,$B1
	.dc.b	$97,$BC,$65,$FE,$89,$19,$7F,$A3,$46,$77,$FE,$A3,$82,$BF,$A6,$FC
	.dc.b	$63,$EA,$BF,$97,$57,$F5,$6E,$EF,$D8,$71,$7F,$AF,$37,$46,$33,$61
	.dc.b	$C7,$CF,$E7,$8F,$2C,$38,$5E,$F0,$E3,$E1,$6A,$FF,$3B,$F1,$BF,$9E
	.dc.b	$F2,$A6,$7F,$2A,$32,$FC,$C8,$CB,$FD,$86,$3E,$DF,$9D,$FD,$87,$EA
	.dc.b	$5F,$55,$4C,$D7,$E3,$27,$0F,$E4,$27,$6E,$48,$7F,$BE,$5C,$B6,$79
	.dc.b	$AD,$FF,$BD,$FD,$43,$C1,$FE,$16,$FF,$5A,$9F,$C4,$B5,$E1,$BF,$CB
	.dc.b	$8E,$11,$A9,$E3,$8C,$8C,$1F,$EF,$8F,$A9,$7D,$57,$F0,$48,$86,$5F
	.dc.b	$B7,$1A,$8F,$C9,$9A,$3C,$27,$D8,$74,$E0,$7C,$83,$E3,$7F,$C1,$9F
	.dc.b	$51,$FE,$07,$9D,$37,$46,$9F,$EE,$CB,$E1,$5F,$BC,$6B,$C3,$D9,$E6
	.dc.b	$7F,$85,$E3,$35,$CA,$47,$8B,$0D,$BF,$C1,$B0,$F0,$61,$DE,$7F,$C4
	.dc.b	$1F,$54,$F1,$3F,$9D,$FD,$57,$F6,$8C,$3D,$FC,$E8,$C6,$5D,$E5,$7C
	.dc.b	$5F,$E2,$6D,$78,$CF,$0C,$7F,$44,$D8,$F3,$CE,$EB,$D5,$D4,$F9,$6F
	.dc.b	$E7,$15,$FC,$13,$E1,$BF,$9D,$9F,$B0,$8C,$C7,$D4,$9E,$1E,$5E,$A3
	.dc.b	$81,$FC,$B6,$32,$FF,$28,$F8,$6F,$F0,$96,$1F,$22,$F9,$FD,$53,$83
	.dc.b	$E5,$DF,$01,$19,$78,$63,$AC,$39,$C6,$DB,$7D,$2F,$9B,$67,$89,$F6
	.dc.b	$27,$75,$CF,$DC,$FD,$34,$F8,$70,$F0,$37,$0E,$75,$BC,$79,$CB,$0E
	.dc.b	$6F,$E1,$2F,$8E,$3E,$18,$F3,$63,$DA,$3E,$5D,$70,$9F,$7E,$3D,$9E
	.dc.b	$4D,$7B,$57,$1C,$7C,$51,$E5,$3C,$78,$78,$23,$38,$76,$AE,$15,$D0
	.dc.b	$DD,$F5,$C3,$C4,$78,$F0,$E5,$3B,$3D,$38,$FC,$09,$D5,$79,$6B,$75
	.dc.b	$E6,$AD,$5F,$82,$3D,$A7,$80,$CC,$47,$D4,$4C,$3D,$11,$97,$E5,$13
	.dc.b	$1F,$69,$27,$CD,$19,$7E,$F5,$78,$71,$F3,$C7,$84,$F5,$92,$78,$3E
	.dc.b	$63,$15,$D5,$C4,$6A,$7E,$3D,$7B,$70,$F8,$27,$C2,$8C,$E1,$DE,$B8
	.dc.b	$57,$8C,$E1,$8F,$C9,$3E,$93,$99,$E1,$8F,$E9,$30,$E9,$39,$C7,$E3
	.dc.b	$4E,$6F,$8D,$FC,$78,$C4,$F2,$3E,$1C,$65,$F8,$F1,$A7,$DE,$68,$EA
	.dc.b	$68,$C6,$1F,$51,$FF,$7B,$7C,$CF,$E4,$BC,$B3,$49,$1F,$E5,$EF,$86
	.dc.b	$FE,$4F,$3A,$8C,$1C,$8A,$E5,$FF,$73,$99,$2D,$7F,$12,$D5,$F5,$5F
	.dc.b	$CC,$E3,$F8,$AC,$FF,$34,$BF,$DA,$D6,$B8,$BF,$6F,$39,$3F,$87,$46
	.dc.b	$5F,$E2,$06,$6B,$B5,$F5,$CB,$FC,$EF,$2C,$3A,$CE,$63,$D7,$5C,$30
	.dc.b	$EF,$1B,$3B,$CE,$DF,$5D,$FF,$20,$C3,$C5,$7E,$A9,$E0,$F3,$7C,$A6
	.dc.b	$27,$85,$73,$8F,$44,$62,$FC,$38,$79,$9B,$3E,$26,$DF,$53,$C1,$3B
	.dc.b	$9E,$ED,$8C,$C7,$13,$25,$72,$73,$AF,$8F,$3C,$5B,$3B,$6C,$EE,$3E
	.dc.b	$4C,$7A,$5E,$B1,$97,$E3,$9F,$DC,$30,$F8,$F3,$C8,$E1,$3E,$26,$99
	.dc.b	$9E,$18,$7C,$77,$E5,$18,$C3,$E1,$CF,$08,$F1,$99,$3C,$6E,$5F,$19
	.dc.b	$FE,$4C,$9E,$98,$7B,$E3,$89,$9A,$EA,$F3,$BF,$33,$6F,$CC,$EC,$EE
	.dc.b	$64,$E7,$87,$C0,$6E,$1E,$07,$E6,$18,$BE,$EF,$3B,$EE,$DB,$EF,$7C
	.dc.b	$E3,$D6,$D7,$F1,$67,$85,$7D,$3B,$E8,$3F,$C4,$5A,$FF,$9D,$BE,$A2
	.dc.b	$E5,$F5,$6B,$4F,$FA,$3A,$33,$87,$F9,$CA,$72,$FF,$88,$3C,$94,$EB
	.dc.b	$1E,$F8,$F0,$E7,$FF,$50,$CF,$2B,$7D,$4B,$E3,$F5,$E4,$3E,$53,$53
	.dc.b	$9D,$7F,$D4,$B3,$97,$FD,$77,$19,$BF,$F1,$8C,$E5,$FF,$27,$46,$76
	.dc.b	$FF,$20,$64,$AD,$47,$F8,$D6,$DF,$1F,$EE,$5F,$25,$FE,$55,$7F,$D9
	.dc.b	$26,$8F,$F6,$91,$AE,$FF,$63,$9C,$BF,$CC,$2E,$4E,$35,$D6,$BA,$4F
	.dc.b	$F9,$8E,$78,$5F,$F9,$55,$E1,$19,$AE,$17,$F3,$78,$5E,$D7,$CD,$E2
	.dc.b	$7C,$26,$63,$0D,$C7,$FD,$8E,$DC,$7E,$C3,$F5,$7E,$A4,$5F,$90,$C3
	.dc.b	$6B,$FD,$AA,$E9,$B1,$A6,$9F,$EE,$EF,$69,$EB,$7C,$BF,$A9,$7B,$9F
	.dc.b	$81,$9C,$D7,$CA,$9C,$BC,$6B,$85,$6D,$B1,$93,$84,$65,$A7,$FB,$93
	.dc.b	$E6,$BF,$21,$EA,$FF,$A2,$23,$07,$FB,$B9,$C6,$9F,$99,$9D,$CF,$CB
	.dc.b	$6B,$FE,$63,$F9,$47,$D4,$74,$F3,$7F,$3D,$18,$9E,$91,$89,$D4,$71
	.dc.b	$6D,$FF,$A1,$B0,$7D,$47,$9B,$B3,$CE,$DB,$F3,$36,$B8,$37,$9F,$FD
	.dc.b	$D0,$D9,$FF,$70,$B6,$5D,$5F,$EE,$AF,$84,$F8,$7F,$51,$F1,$1B,$7F
	.dc.b	$0D,$B5,$E5,$6E,$1F,$EF,$5F,$A8,$F9,$7E,$0F,$CC,$6D,$FC,$B6,$DF
	.dc.b	$B5,$B5,$FE,$A1,$F1,$BF,$83,$BF,$EB,$1E,$57,$D4,$39,$FF,$82,$56
	.dc.b	$B1,$FF,$5A,$D6,$A7,$39,$7F,$12,$8C,$4F,$FA,$EB,$0E,$D3,$9C,$3A
	.dc.b	$4E,$4F,$F4,$AE,$1D,$72,$61,$C9,$FF,$6A,$46,$63,$18,$7F,$B1,$FC
	.dc.b	$E7,$D3,$BE,$07,$F8,$D7,$98,$FA,$7F,$E4,$B5,$A9,$CC,$4C,$FC,$59
	.dc.b	$AB,$9F,$5E,$63,$FD,$CF,$18,$9D,$F1,$BC,$0F,$D2,$D7,$E2,$7D,$8F
	.dc.b	$E0,$1D,$BF,$9E,$7F,$96,$79,$9F,$C0,$B0,$EF,$5F,$EC,$5F,$07,$CB
	.dc.b	$F4,$FF,$C8,$67,$35,$BA,$FF,$7A,$37,$87,$E0,$53,$BB,$AE,$1C,$99
	.dc.b	$87,$EA,$FE,$1E,$E3,$F9,$D7,$F9,$F7,$D3,$BE,$03,$FD,$C5,$3F,$EC
	.dc.b	$68,$C5,$7F,$8D,$21,$18,$7F,$C6,$ED,$7F,$D9,$A6,$A7,$FA,$9E,$4F
	.dc.b	$A7,$7C,$1F,$82,$73,$18,$C3,$E1,$E1,$D6,$FF,$B0,$CE,$70,$47,$FB
	.dc.b	$DE,$75,$5F,$8D,$AD,$4F,$FB,$EF,$E9,$FE,$24,$FD,$CE,$37,$D3,$BE
	.dc.b	$3B,$FA,$F5,$6B,$2F,$BB,$39,$BF,$AB,$39,$C3,$F1,$D3,$9A,$F1,$D7
	.dc.b	$FB,$B2,$B4,$A3,$F1,$53,$BA,$F2,$6D,$F9,$28,$C5,$71,$8C,$61,$FD
	.dc.b	$C7,$E9,$79,$4F,$E5,$3E,$71,$F3,$FF,$D5,$31,$4F,$FB,$F2,$B7,$97
	.dc.b	$90,$5F,$0F,$43,$A4,$7F,$81,$A3,$15,$F9,$2A,$E3,$7D,$2B,$8E,$1D
	.dc.b	$EB,$B4,$71,$8C,$3E,$FA,$E3,$3F,$06,$FA,$CF,$8C,$4F,$EB,$67,$D9
	.dc.b	$F4,$DF,$E1,$47,$C5,$B7,$96,$31,$5C,$6B,$83,$ED,$B4,$EB,$1F,$44
	.dc.b	$F6,$BE,$5E,$BF,$CB,$6E,$3D,$8F,$0C,$7C,$FF,$5F,$E2,$7D,$2B,$EA
	.dc.b	$B1,$5F,$46,$30,$FC,$79,$DC,$EB,$0F,$25,$69,$F4,$4E,$AB,$81,$CE
	.dc.b	$BE,$F7,$C8,$F8,$3E,$7E,$CF,$76,$DF,$46,$E7,$F5,$75,$64,$9F,$F2
	.dc.b	$0F,$D1,$7E,$1B,$DA,$BC,$3F,$E2,$6C,$BF,$B6,$7D,$36,$26,$47,$F2
	.dc.b	$0A,$E0,$E8,$C5,$71,$CF,$ED,$7D,$2F,$F9,$0F,$87,$FD,$85,$CB,$F9
	.dc.b	$F9,$4E,$5F,$F5,$D4,$67,$D6,$7D,$37,$E3,$7C,$DF,$D9,$57,$E9,$EF
	.dc.b	$84,$62,$39,$FA,$BF,$CC,$DE,$66,$DC,$46,$EF,$AE,$7E,$4A,$D6,$55
	.dc.b	$E6,$DC,$3B,$E5,$E5,$CB,$C8,$FA,$E3,$35,$C8,$F0,$9C,$8B,$87,$81
	.dc.b	$CC,$70,$75,$8F,$AE,$31,$3B,$8C,$65,$EA,$BF,$03,$6F,$C1,$3E,$DC
	.dc.b	$3C,$B1,$8A,$DE,$3E,$BB,$E1,$2C,$3B,$56,$A3,$CB,$3F,$2D,$FE,$6F
	.dc.b	$8F,$D7,$6C,$65,$E0,$7F,$3C,$F8,$7F,$F3,$8F,$17,$F4,$36,$C7,$FB
	.dc.b	$E1,$AE,$FD,$EF,$E2,$2F,$FB,$F2,$DE,$AB,$E1,$3F,$A6,$46,$1B,$19
	.dc.b	$9E,$0D,$7C,$EE,$76,$F9,$8D,$CF,$EC,$36,$7F,$99,$7D,$37,$F8,$65
	.dc.b	$FC,$04,$65,$FC,$BB,$70,$F1,$47,$DE,$9E,$0F,$DC,$9D,$13,$0F,$C6
	.dc.b	$4E,$B0,$F1,$C7,$62,$57,$3F,$5B,$FC,$3B,$8F,$9A,$73,$5C,$2B,$46
	.dc.b	$27,$4F,$86,$7B,$D7,$28,$FD,$46,$1E,$37,$47,$23,$F4,$FE,$73,$DA
	.dc.b	$7D,$AF,$ED,$FA,$FF,$D5,$F1,$8A,$FF,$38,$4E,$79,$7F,$AD,$5F,$1F
	.dc.b	$80,$C5,$7F,$96,$2F,$83,$FE,$67,$BC,$D7,$76,$CF,$FA,$8A,$7F,$C3
	.dc.b	$FA,$7F,$AF,$5B,$F0,$5D,$1F,$EE,$8F,$6E,$37,$E4,$8F,$87,$C3,$6B
	.dc.b	$89,$0F,$E8,$8E,$5F,$EA,$17,$FE,$25,$7C,$3E,$3F,$78,$C4,$70,$F3
	.dc.b	$F1,$32,$36,$7F,$71,$AF,$F5,$CB,$67,$4F,$69,$F2,$C7,$AF,$E9,$7F
	.dc.b	$31,$C1,$FD,$51,$BC,$1F,$C9,$DB,$7F,$DD,$BE,$B5,$F4,$1C,$79,$64
	.dc.b	$FF,$91,$AB,$7B,$2F,$73,$FE,$05,$6B,$FD,$BE,$7A,$A7,$8C,$6B,$1F
	.dc.b	$27,$B7,$E7,$25,$73,$E2,$AF,$38,$B3,$FD,$46,$BF,$02,$DA,$CE,$7E
	.dc.b	$DA,$FB,$55,$A8,$FE,$8D,$F4,$BF,$E4,$6B,$FE,$EE,$D8,$CE,$1F,$8E
	.dc.b	$F8,$A7,$E1,$3F,$87,$8F,$F0,$55,$6D,$FF,$17,$35,$39,$9E,$3C,$9C
	.dc.b	$EB,$96,$1E,$0B,$FC,$85,$EE,$73,$87,$9B,$C1,$C9,$2B,$FD,$A9,$EA
	.dc.b	$D7,$C7,$17,$0F,$D5,$CF,$E8,$0C,$57,$3A,$E2,$7F,$A0,$3E,$93,$D8
	.dc.b	$FF,$9A,$1A,$FF,$6B,$6B,$FE,$4B,$6A,$7C,$EE,$72,$FD,$54,$62,$1F
	.dc.b	$4B,$F8,$22,$71,$7E,$BD,$FC,$8E,$2F,$D8,$DF,$51,$FB,$3F,$AE,$FF
	.dc.b	$75,$E1,$FE,$96,$6B,$FE,$3B,$6D,$7F,$8C,$3C,$67,$CD,$7E,$ED,$B5
	.dc.b	$FE,$9E,$6B,$FE,$43,$6C,$FF,$8E,$2B,$FC,$0B,$3F,$F0,$07,$EA,$3C
	.dc.b	$5E,$54,$8C,$1F,$EA,$8E,$2F,$F4,$5B,$78,$7F,$85,$46,$23,$FC,$A1
	.dc.b	$F9,$2B,$C9,$CD,$49,$E0,$7F,$AC,$A7,$4D,$9D,$3F,$E5,$9A,$E9,$97
	.dc.b	$F9,$3A,$B7,$3F,$E6,$FF,$0F,$F9,$77,$C2,$FF,$20,$24,$7F,$B1,$BF
	.dc.b	$4D,$FD,$FC,$EA,$3F,$D9,$3E,$2F,$FB,$55,$B5,$FC,$AA,$B6,$3B,$30
	.dc.b	$49,$E9,$F5,$75,$F1,$4E,$DF,$04,$F3,$9E,$9E,$2C,$9F,$23,$FC,$F3
	.dc.b	$D0,$7E,$95,$EC,$D7,$FD,$05,$A3,$2F,$F1,$A4,$61,$BD,$9F,$D1,$9B
	.dc.b	$8B,$0E,$7C,$1D,$E3,$CB,$95,$C3,$FD,$1D,$5A,$9F,$8D,$EB,$FF,$01
	.dc.b	$FC,$CB,$F1,$B9,$7F,$AA,$FC,$14,$62,$9A,$79,$6F,$E4,$1C,$7E,$4A
	.dc.b	$4F,$43,$E9,$7D,$57,$CD,$F9,$A9,$18,$21,$F2,$9C,$BF,$32,$3F,$AB
	.dc.b	$7E,$1C,$FA,$CA,$49,$FF,$7D,$FD,$3B,$E0,$F1,$1F,$EA,$1A,$E0,$FF
	.dc.b	$C0,$06,$BF,$EA,$59,$E0,$F2,$9E,$3C,$1F,$E9,$CA,$DB,$FE,$9E,$AD
	.dc.b	$D7,$E0,$3D,$DF,$C4,$11,$FC,$87,$CF,$57,$47,$E4,$E3,$F1,$D1,$FD
	.dc.b	$C7,$EE,$FF,$84,$7E,$48,$F7,$FF,$BA,$36,$FF,$DB,$98,$FF,$AB,$AB
	.dc.b	$F8,$71,$C2,$B6,$73,$F2,$FE,$3F,$D6,$CB,$F8,$0F,$F8,$1B,$F0,$0F
	.dc.b	$F5,$BE,$1E,$D8,$D4,$FC,$77,$5C,$F2,$38,$90,$FD,$85,$72,$8F,$C9
	.dc.b	$75,$7F,$49,$FC,$CF,$D1,$5F,$EE,$3E,$94,$63,$1F,$CD,$57,$EB,$CE
	.dc.b	$19,$CB,$FD,$C4,$7E,$CF,$EB,$3A,$7C,$3C,$9F,$DE,$46,$5F,$8B,$97
	.dc.b	$E5,$F1,$56,$E7,$F6,$B5,$FE,$1F,$F8,$27,$0B,$35,$3E,$49,$F4,$9F
	.dc.b	$EA,$DF,$A5,$75,$FF,$AB,$E7,$35,$C2,$72,$F2,$8C,$AC,$7B,$E1,$9C
	.dc.b	$3F,$D7,$11,$89,$F0,$47,$7C,$3B,$36,$38,$1E,$69,$E1,$7D,$9B,$3B
	.dc.b	$9E,$67,$96,$B9,$BC,$71,$F8,$0E,$8F,$33,$8B,$F4,$E3,$EC,$9D,$36
	.dc.b	$3C,$13,$BB,$F5,$B6,$BC,$07,$A0,$57,$19,$DE,$39,$C7,$D3,$5C,$0E
	.dc.b	$F8,$FA,$1F,$2C,$EF,$0E,$D1,$D2,$75,$87,$58,$F7,$56,$CF,$14,$66
	.dc.b	$38,$19,$3E,$15,$F8,$9C,$9E,$23,$58,$FC,$0A,$E7,$7E,$03,$CD,$5C
	.dc.b	$4E,$57,$DC,$F0,$39,$C3,$D3,$19,$9E,$31,$EC,$9E,$31,$D7,$0F,$0E
	.dc.b	$3F,$21,$F2,$13,$1F,$23,$EA,$BE,$B5,$C6,$31,$8F,$27,$E7,$57,$DE
	.dc.b	$9F,$0C,$E6,$7E,$0C,$E5,$C6,$1E,$47,$E0,$98,$BE,$EE,$0E,$E4,$9E
	.dc.b	$4F,$9F,$0F,$54,$7A,$0C,$91,$C9,$23,$CE,$FE,$76,$38,$39,$8F,$3B
	.dc.b	$93,$15,$E7,$8F,$01,$F8,$43,$CB,$E9,$FF,$77,$7F,$1A,$7F,$B9,$7C
	.dc.b	$6F,$9D,$19,$9F,$DB,$3B,$9F,$E0,$FF,$07,$FC,$22,$31,$08,$C5,$7F
	.dc.b	$0C,$89,$18,$70,$6C,$F0,$18,$7F,$51,$C7,$FE,$B1,$E1,$FD,$54,$F8
	.dc.b	$0F,$DC,$46,$5F,$DD,$19,$CB,$F8,$8C,$65,$FE,$27,$19,$CB,$C9,$39
	.dc.b	$7C,$71,$E3,$9E,$3F,$00,$BF,$0C,$E6,$B8,$CE,$63,$C2,$64,$FF,$04
	.dc.b	$39,$AF,$0C,$64,$5F,$85,$F8,$07,$B2,$78,$57,$93,$0F,$87,$18,$9D
	.dc.b	$DF,$9B,$0F,$2D,$69,$B3,$CD,$B3,$B8,$F3,$CE,$DA,$6C,$B1,$B3,$D3
	.dc.b	$1B,$6C,$6D,$A6,$EF,$D1,$19,$6C,$66,$3D,$31,$F2,$AB,$CA,$D8,$F0
	.dc.b	$B6,$7A,$B4,$E6,$58,$E6,$7D,$09,$E6,$D3,$99,$67,$9B,$67,$9B,$4E
	.dc.b	$67,$D6,$BF,$A0,$FE,$68,$F6,$1A,$C3,$ED,$39,$AF,$73,$6B,$CC,$D3
	.dc.b	$B1,$4E,$D7,$E8,$69,$E8,$2C,$FA,$1B,$3E,$86,$9E,$3B,$F3,$BF,$9F
	.dc.b	$3B,$94,$D9,$E1,$BF,$4B,$97,$F2,$9E,$17,$F9,$C7,$F0,$0F,$CF,$6F
	.dc.b	$F9,$F9,$78,$6F,$9F,$CE,$9F,$E7,$3B,$B8,$FE,$7A,$B5,$3F,$E9,$39
	.dc.b	$CD,$F2,$BD,$E3,$DE,$BF,$C3,$78,$71,$9C,$F9,$0F,$A4,$E3,$93,$E2
	.dc.b	$F3,$18,$6E,$1C,$B2,$F3,$C6,$31,$FF,$1A,$5F,$1A,$EF,$93,$F0,$17
	.dc.b	$D2,$7A,$AF,$F8,$E7,$C1,$F9,$A7,$37,$E6,$9C,$B7,$8D,$5C,$5B,$1D
	.dc.b	$AB,$8D,$7F,$8C,$63,$F9,$2D,$6C,$8E,$EE,$CE,$CF,$F1,$BF,$CC,$7F
	.dc.b	$B8,$27,$C4,$BC,$3F,$43,$99,$E0,$1D,$4F,$0F,$6B,$75,$C1,$B1,$FE
	.dc.b	$C9,$F8,$AF,$A2,$7D,$2F,$C0,$6C,$64,$98,$7F,$B0,$A3,$2F,$83,$F1
	.dc.b	$F7,$AB,$F2,$FE,$8D,$E8,$7C,$D9,$CE,$1F,$2E,$72,$D9,$E0,$6E,$AC
	.dc.b	$64,$97,$DE,$32,$FF,$B5,$3E,$FB,$F1,$3E,$E7,$E2,$9B,$5F,$EE,$56
	.dc.b	$CE,$E3,$FD,$D1,$3F,$90,$D3,$F1,$F7,$67,$F1,$9C,$9E,$8B,$FF,$70
	.dc.b	$D6,$A7,$FD,$E0,$D8,$FF,$72,$FA,$7F,$9A,$3F,$4B,$EA,$BF,$1B,$7C
	.dc.b	$66,$70,$FF,$7C,$55,$BF,$C0,$5E,$EB,$1D,$FF,$EE,$3A,$B7,$FE,$7A
	.dc.b	$AD,$3C,$8F,$CE,$F6,$FE,$03,$0F,$F7,$BF,$27,$EB,$3D,$5F,$D6,$B2
	.dc.b	$B6,$E5,$A7,$E9,$BA,$3F,$D0,$ED,$AF,$DC,$C3,$E1,$AC,$49,$D1,$C6
	.dc.b	$31,$1F,$94,$E4,$FF,$6C,$4F,$EA,$E7,$FD,$32,$FE,$0F,$0F,$D9,$46
	.dc.b	$2B,$F3,$6D,$D3,$FB,$94,$78,$7C,$CF,$C7,$57,$06,$5F,$57,$2D,$3D
	.dc.b	$9E,$1F,$F1,$8F,$C0,$DF,$81,$BE,$A3,$DB,$39,$8C,$57,$F0,$1F,$C0
	.dc.b	$FF,$8A,$37,$C5,$D5,$FF,$09,$9C,$E7,$FE,$C3,$FC,$0F,$F9,$4D,$FF
	.dc.b	$15,$6C,$7E,$DA,$77,$AF,$F2,$F9,$CB,$FC,$C7,$EC,$DF,$03,$FC,$A9
	.dc.b	$B8,$7F,$28,$FC,$29,$F6,$2F,$B6,$D5,$FF,$1B,$FA,$57,$D3,$53,$38
	.dc.b	$7F,$1B,$8C,$B6,$3E,$6E,$BF,$E3,$B9,$CB,$FE,$28,$6B,$FE,$3C,$6B
	.dc.b	$FE,$CC,$F2,$FF,$91,$FC,$1F,$F2,$BF,$B0,$7C,$17,$E2,$23,$0C,$C7
	.dc.b	$F9,$64,$E9,$FC,$84,$E6,$FF,$8D,$CE,$5B,$5C,$EB,$8B,$F1,$7D,$BF
	.dc.b	$07,$B3,$67,$AB,$86,$9F,$D1,$F8,$13,$9B,$EF,$57,$87,$E9,$4D,$BF
	.dc.b	$F2,$F7,$E1,$3F,$D2,$EF,$F9,$7F,$D2,$BE,$DB,$FA,$54,$E7,$0F,$CB
	.dc.b	$CE,$6F,$FA,$53,$6F,$8D,$F0,$F5,$3F,$A9,$4E,$6F,$8C,$E0,$D5,$F2
	.dc.b	$7B,$F1,$F0,$F8,$1B,$3D,$9C,$34,$FE,$B5,$97,$E1,$DC,$65,$EA,$9D
	.dc.b	$BF,$D2,$A7,$4E,$67,$FB,$17,$E5,$3F,$E7,$3C,$7D,$BF,$27,$FD,$A6
	.dc.b	$31,$3F,$E8,$3C,$D3,$EC,$24,$72,$EE,$B8,$7C,$A9,$B8,$7E,$FB,$0E
	.dc.b	$72,$18,$74,$C3,$C7,$F3,$9F,$23,$9F,$C4,$D8,$F9,$F1,$6B,$FB,$2D
	.dc.b	$AB,$18,$7F,$9F,$79,$BF,$BD,$FD,$37,$F1,$E3,$F8,$FC,$DC,$7F,$92
	.dc.b	$45,$C3,$FB,$B5,$7F,$80,$AB,$17,$FC,$5B,$E8,$9F,$2D,$FC,$9E,$FF
	.dc.b	$D5,$1F,$1F,$FE,$A7,$9C,$F8,$FF,$EA,$E8,$FF,$55,$7D,$1B,$D4,$FF
	.dc.b	$48,$7B,$A9,$E1,$8F,$38,$C6,$3F,$17,$1E,$6F,$63,$13,$BA,$EB,$7D
	.dc.b	$1E,$78,$F8,$6B,$6D,$57,$D3,$0E,$F0,$C7,$CD,$B6,$EB,$58,$74,$AF
	.dc.b	$11,$CB,$2F,$13,$F6,$EF,$8B,$AA,$E2,$F8,$8C,$4F,$07,$C1,$97,$A5
	.dc.b	$D6,$3E,$4A,$E3,$1D,$F2,$F5,$61,$DA,$B5,$87,$3B,$D9,$C2,$75,$8F
	.dc.b	$A6,$B7,$7F,$EC,$8C,$3A,$1F,$EC,$AB,$F9,$EF,$FB,$79,$AF,$FB,$8F
	.dc.b	$E1,$5F,$12,$F9,$FE,$51,$86,$C6,$67,$FD,$AB,$39,$7B,$7D,$13,$E2
	.dc.b	$3C,$58,$23,$13,$A9,$CC,$62,$3C,$8E,$B2,$F8,$CD,$CF,$F2,$4D,$9F
	.dc.b	$D7,$7D,$1B,$E0,$3F,$DF,$75,$AC,$3E,$F5,$EE,$FF,$0A,$DB,$E1,$1A
	.dc.b	$98,$FD,$D9,$D5,$F4,$8F,$CC,$47,$3C,$7D,$95,$C3,$1F,$C7,$E1,$D2
	.dc.b	$F9,$4E,$71,$F2,$47,$E3,$A7,$55,$EE,$AD,$4F,$67,$F3,$93,$AA,$E3
	.dc.b	$5B,$AE,$A7,$3C,$7C,$DC,$8F,$95,$7D,$07,$F1,$3A,$FF,$3B,$7C,$33
	.dc.b	$F3,$57,$D3,$E7,$A3,$FD,$55,$18,$AF,$F4,$8E,$1C,$9B,$86,$FA,$3F
	.dc.b	$8F,$63,$CF,$1C,$52,$30,$F0,$3F,$D2,$9A,$BE,$4F,$F2,$D1,$FC,$1D
	.dc.b	$B1,$98,$FC,$D3,$5F,$51,$FC,$B5,$C6,$7F,$E2,$E6,$E1,$EF,$6C,$FE
	.dc.b	$6F,$E9,$1E,$0F,$9A,$3F,$8B,$7B,$FE,$8E,$87,$E8,$AF,$B1,$C3,$6B
	.dc.b	$D3,$5C,$71,$FF,$32,$C7,$9A,$3D,$24,$70,$70,$21,$FE,$A7,$B6,$3F
	.dc.b	$CF,$63,$FD,$71,$F4,$DF,$A3,$D1,$A7,$F4,$BB,$FD,$AF,$C4,$79,$FB
	.dc.b	$3F,$D2,$F8,$70,$E9,$F6,$63,$FA,$B8,$F8,$ED,$7F,$04,$D7,$FA,$B3
	.dc.b	$53,$E1,$C3,$73,$D0,$F8,$DF,$4D,$FE,$CA,$AF,$3F,$C9,$F8,$F0,$91
	.dc.b	$82,$1F,$60,$EB,$1F,$80,$C0,$D2,$47,$0C,$FF,$CD,$33,$A8,$FF,$78
	.dc.b	$E6,$C3,$F2,$91,$FE,$CA,$FA,$6F,$C2,$DF,$C8,$8F,$C9,$7E,$6F,$FD
	.dc.b	$71,$FF,$78,$35,$E0,$7E,$0E,$7F,$26,$9A,$47,$F2,$EF,$6D,$3F,$58
	.dc.b	$D8,$FF,$00,$37,$95,$F4,$3F,$92,$9E,$9E,$17,$E9,$A5,$83,$E9,$30
	.dc.b	$DA,$FF,$7D,$61,$FD,$CD,$F7,$5F,$88,$8F,$F3,$F8,$D3,$FA,$67,$86
	.dc.b	$5F,$E9,$26,$E5,$E8,$AF,$DB,$F6,$FF,$07,$FA,$3F,$D2,$7D,$37,$F9
	.dc.b	$EB,$1F,$F0,$D6,$3F,$E9,$A8,$C4,$7F,$BE,$1B,$1F,$E9,$EF,$A7,$F7
	.dc.b	$1F,$B7,$9F,$DB,$B5,$FE,$24,$D7,$FC,$4C,$D5,$3F,$91,$73,$D9,$FB
	.dc.b	$79,$F9,$53,$9F,$79,$FA,$FB,$ED,$FF,$88,$C6,$1B,$19,$C3,$D7,$AF
	.dc.b	$F2,$3E,$07,$C7,$7F,$03,$F3,$1F,$21,$FC,$51,$B7,$FD,$AD,$AF,$E6
	.dc.b	$DA,$FF,$14,$AF,$F1,$13,$D8,$FF,$21,$38,$5B,$7F,$B1,$23,$0F,$F8
	.dc.b	$EF,$C5,$7D,$47,$F1,$5E,$77,$15,$F8,$D9,$8F,$53,$F1,$B1,$B3,$F9
	.dc.b	$5C,$F0,$8F,$F0,$13,$5F,$D9,$35,$FE,$56,$E7,$0F,$F6,$BB,$75,$FE
	.dc.b	$90,$D9,$FF,$2E,$7D,$47,$F8,$19,$FE,$8B,$F2,$9F,$8A,$C5,$F3,$B4
	.dc.b	$FC,$6E,$89,$A6,$6B,$FD,$F1,$38,$7E,$47,$3B,$E4,$BF,$80,$47,$F5
	.dc.b	$8E,$BF,$8A,$FE,$EB,$7F,$CC,$CE,$B6,$FF,$32,$37,$6F,$EA,$B7,$C2
	.dc.b	$F9,$5F,$0F,$03,$8F,$8D,$FB,$3F,$47,$85,$7F,$59,$6E,$7E,$39,$DC
	.dc.b	$FD,$DE,$FF,$F0,$E5,$F1,$33,$1F,$E6,$EC,$FF,$CA,$3A,$7F,$13,$AD
	.dc.b	$CF,$1C,$7A,$D6,$B4,$FE,$D7,$F4,$FF,$94,$AF,$87,$C3,$FE,$B7,$6F
	.dc.b	$BF,$FC,$26,$75,$1F,$DF,$1B,$97,$76,$DF,$4E,$9F,$8F,$F2,$DF,$23
	.dc.b	$E8,$3E,$07,$C3,$4C,$7E,$8E,$89,$D4,$79,$E7,$4F,$B6,$3F,$B5,$7D
	.dc.b	$3B,$E2,$7F,$B6,$74,$BE,$42,$9A,$89,$7F,$E5,$5A,$E0,$D8,$FF,$01
	.dc.b	$C6,$23,$FA,$43,$76,$FE,$79,$08,$C6,$BF,$E1,$D6,$C7,$FA,$C7,$E9
	.dc.b	$DF,$2B,$F8,$6D,$3E,$F7,$C4,$CA,$F7,$C6,$5F,$A9,$19,$7E,$D4,$7F
	.dc.b	$81,$FE,$9D,$F1,$17,$B6,$57,$EC,$E7,$2D,$9E,$04,$7F,$71,$F3,$1F
	.dc.b	$D0,$F3,$7D,$3F,$F8,$4F,$F2,$57,$E0,$F4,$F9,$B1,$2F,$E4,$3F,$18
	.dc.b	$8E,$0F,$E0,$95,$D1,$B1,$FC,$FB,$EA,$7E,$4F,$0F,$E4,$FE,$A7,$CF
	.dc.b	$5F,$EC,$4E,$CB,$3F,$EC,$D6,$F1,$FF,$8F,$AA,$D7,$F8,$F5,$BD,$C6
	.dc.b	$9A,$F0,$7F,$75,$EB,$7F,$5F,$87,$CE,$7F,$07,$FB,$59,$87,$EA,$E7
	.dc.b	$2D,$9E,$04,$7F,$2B,$CD,$FE,$39,$F8,$77,$C9,$FF,$2C,$9F,$F0,$5F
	.dc.b	$3F,$F9,$7F,$EA,$BF,$35,$1F,$EF,$3F,$1B,$F8,$AB,$FE,$F5,$8C,$BC
	.dc.b	$0F,$F7,$C5,$F4,$E1,$E1,$7F,$EA,$5C,$3B,$7D,$57,$2B,$61,$CA,$30
	.dc.b	$DA,$E1,$6A,$57,$1C,$3B,$46,$2B,$D1,$E0,$FF,$9D,$EB,$9B,$63,$9C
	.dc.b	$EC,$F3,$E3,$E4,$8E,$31,$DE,$B8,$B6,$38,$C7,$86,$7E,$46,$5A,$AF
	.dc.b	$84,$7B,$2F,$FD,$11,$1B,$C7,$E1,$CF,$81,$E8,$FC,$B9,$F3,$34,$C8
	.dc.b	$C7,$DF,$87,$38,$EF,$97,$D2,$C3,$C4,$64,$D4,$F9,$49,$97,$A8,$95
	.dc.b	$DC,$F8,$87,$A4,$F0,$DF,$82,$37,$19,$37,$7F,$5D,$F6,$1C,$A7,$EA
	.dc.b	$DF,$92,$BE,$6C,$FC,$69,$CC,$74,$AE,$53,$D6,$FD,$33,$9C,$3E,$2C
	.dc.b	$E6,$3C,$77,$CA,$3D,$58,$77,$7C,$6F,$27,$DD,$7F,$09,$F3,$B9,$7A
	.dc.b	$1A,$AE,$B9,$7D,$E7,$D3,$5D,$67,$EA,$4E,$5E,$73,$EE,$AE,$D1,$AC
	.dc.b	$BE,$5D,$7E,$3E,$3B,$10,$D1,$F2,$5D,$3E,$42,$61,$F4,$DF,$6D,$F9
	.dc.b	$08,$E0,$F1,$9D,$72,$F9,$B3,$C6,$3B,$E3,$EB,$9F,$C5,$57,$E9,$A3
	.dc.b	$F1,$4F,$49,$FC,$04,$61,$FB,$FE,$EF,$F8,$3A,$FF,$66,$DC,$3F,$86
	.dc.b	$37,$1F,$DD,$3F,$E1,$4A,$FD,$E7,$C4,$63,$0F,$F4,$F3,$63,$FD,$AE
	.dc.b	$D7,$FD,$E4,$D7,$89,$FB,$0E,$3F,$F6,$AF,$27,$EB,$A3,$66,$7E,$13
	.dc.b	$FD,$39,$7F,$C4,$63,$8B,$FC,$4E,$B8,$C7,$EE,$8C,$E9,$FF,$00,$63
	.dc.b	$33,$C6,$58,$7E,$DA,$78,$ED,$98,$C6,$1F,$EA,$B8,$C5,$78,$F1,$F3
	.dc.b	$3D,$8F,$E3,$D7,$DB,$1F,$87,$1F,$B4,$8E,$B1,$8A,$F4,$5F,$58,$C1
	.dc.b	$E7,$71,$5F,$28,$E4,$E1,$F8,$B3,$CA,$FF,$87,$C7,$CC,$8C,$3F,$C1
	.dc.b	$0F,$30,$BF,$33,$D7,$1D,$9E,$E9,$E2,$5C,$3C,$97,$DA,$BE,$49,$E4
	.dc.b	$71,$8F,$C8,$6C,$7C,$73,$EA,$1C,$2F,$BD,$7C,$08,$F9,$38,$FB,$5F
	.dc.b	$54,$78,$1F,$51,$E0,$C3,$CB,$1C,$63,$D9,$3C,$5E,$D1,$C6,$BE,$C9
	.dc.b	$E0,$8F,$99,$5C,$9B,$3D,$E3,$E0,$47,$92,$FD,$71,$B8,$FA,$33,$B3
	.dc.b	$C5,$8F,$DE,$8F,$99,$1E,$13,$F0,$77,$E1,$8E,$27,$53,$DB,$8F,$C9
	.dc.b	$69,$F2,$63,$D3,$3C,$E3,$CE,$71,$BF,$3B,$F0,$B0,$F3,$CF,$96,$36
	.dc.b	$64,$EB,$3E,$B7,$E7,$9C,$E3,$CE,$77,$7F,$67,$5F,$27,$07,$8C,$FA
	.dc.b	$96,$3F,$E8,$58,$C3,$FE,$6C,$FA,$AC,$E1,$FE,$C2,$8F,$D3,$CE,$A3
	.dc.b	$07,$E5,$5C,$6D,$FE,$33,$ED,$E2,$FF,$A2,$E3,$35,$FE,$0D,$E9,$75
	.dc.b	$7F,$92,$A3,$C5,$F2,$FC,$1E,$ED,$C5,$5F,$EA,$FF,$1F,$1A,$FB,$2A
	.dc.b	$E2,$78,$57,$6E,$9C,$C6,$32,$FF,$1E,$E5,$FE,$D5,$E2,$F1,$B6,$38
	.dc.b	$FD,$53,$EA,$BC,$5A,$FF,$12,$9B,$7C,$23,$F9,$5C,$7F,$A0,$1F,$E5
	.dc.b	$2D,$7F,$98,$B5,$E5,$7F,$EE,$5F,$A8,$FF,$75,$BE,$13,$1E,$97,$83
	.dc.b	$EA,$9E,$35,$31,$ED,$DA,$FF,$77,$4E,$9F,$F7,$87,$E0,$5F,$E7,$87
	.dc.b	$35,$EA,$8C,$63,$FE,$C1,$F9,$0F,$97,$F8,$6F,$D6,$8F,$9F,$E9,$27
	.dc.b	$3D,$5F,$32,$31,$8F,$FB,$36,$F8,$72,$7F,$B8,$EF,$73,$FF,$00,$7E
	.dc.b	$A1,$F1,$7F,$29,$B1,$F0,$7D,$BC,$4C,$8F,$F2,$96,$70,$57,$07,$06
	.dc.b	$27,$E6,$46,$63,$83,$BE,$BF,$F5,$9A,$E0,$ED,$3F,$9A,$3A,$BE,$DF
	.dc.b	$96,$F8,$7D,$17,$1F,$73,$6B,$F4,$3D,$6D,$BF,$21,$5F,$ED,$4B,$E2
	.dc.b	$FE,$B7,$4F,$CA,$9F,$27,$1B,$19,$3F,$DD,$3F,$23,$8C,$7F,$D3,$CF
	.dc.b	$CF,$F6,$3E,$AF,$2D,$AF,$A0,$DA,$F1,$E4,$8C,$6D,$FE,$FB,$8C,$57
	.dc.b	$FB,$A2,$F9,$B4,$E6,$7F,$BB,$39,$7F,$07,$F0,$1F,$7D,$B1,$93,$F0
	.dc.b	$11,$96,$9F,$E8,$9B,$FF,$5A,$63,$FA,$FF,$C1,$3F,$60,$D9,$D1,$FC
	.dc.b	$0F,$AB,$FD,$20,$E2,$3F,$7F,$09,$E3,$39,$E5,$FE,$13,$1F,$B4,$F8
	.dc.b	$AF,$DD,$DF,$DB,$6E,$1F,$80,$F6,$35,$E0,$4C,$7F,$13,$6B,$E2,$FE
	.dc.b	$3D,$AF,$E4,$9A,$FF,$A9,$5E,$B1,$9F,$4F,$A4,$F8,$5F,$E3,$1F,$52
	.dc.b	$FA,$97,$D4,$E1,$B7,$EB,$DB,$FC,$47,$A7,$F3,$0B,$E1,$7F,$C4,$3E
	.dc.b	$A5,$E3,$FF,$09,$6E,$1E,$A6,$E7,$FA,$46,$BF,$AA,$F8,$B7,$EA,$CF
	.dc.b	$BB,$FE,$59,$18,$7F,$95,$FD,$5F,$F4,$6C,$3E,$94,$E5,$FA,$9D,$CF
	.dc.b	$A7,$7D,$57,$F2,$4F,$AA,$7D,$2F,$F3,$49,$CD,$FD,$BD,$15,$A8,$C3
	.dc.b	$63,$F7,$6D,$7F,$80,$35,$FE,$06,$D5,$C2,$FB,$87,$5F,$F4,$78,$C3
	.dc.b	$E5,$C7,$13,$AC,$3C,$9F,$56,$F8,$6F,$E7,$AD,$8F,$A9,$F1,$18,$99
	.dc.b	$1F,$88,$9D,$65,$FC,$49,$AF,$F1,$76,$BF,$C6,$9D,$E5,$FD,$B3,$67
	.dc.b	$03,$D5,$EB,$7F,$C3,$23,$18,$79,$BD,$BF,$29,$F3,$7D,$FB,$19,$3C
	.dc.b	$91,$FD,$37,$23,$49,$1F,$6A,$9E,$4F,$2C,$3C,$0F,$F5,$EF,$A8,$7C
	.dc.b	$65,$3F,$B0,$FD,$57,$F1,$EA,$FE,$51,$3A,$C3,$EF,$36,$3F,$9A,$B5
	.dc.b	$FE,$74,$D7,$F9,$EB,$50,$79,$6D,$FE,$12,$F9,$C7,$C5,$3E,$6B,$C9
	.dc.b	$87,$F0,$79,$CC,$7A,$66,$4F,$E4,$71,$7D,$5E,$32,$F4,$D7,$F7,$3C
	.dc.b	$BA,$DF,$F0,$46,$CF,$AA,$74,$7F,$84,$F1,$F1,$FD,$D7,$3C,$FC,$D2
	.dc.b	$8F,$95,$F5,$93,$1F,$D1,$7D,$67,$EF,$9F,$83,$19,$C7,$E3,$BE,$32
	.dc.b	$47,$8D,$F0,$7E,$82,$FA,$DE,$19,$7C,$58,$C5,$7D,$2E,$BF,$B4,$7F
	.dc.b	$86,$31,$7D,$6E,$27,$47,$BE,$7E,$11,$1F,$ED,$BE,$7F,$F8,$DB,$1F
	.dc.b	$F1,$7D,$7D,$49,$FE,$71,$18,$E3,$FF,$16,$36,$3F,$8E,$37,$4F,$ED
	.dc.b	$DF,$58,$F9,$EF,$E6,$11,$FE,$BC,$F9,$AF,$EF,$F3,$9F,$3F,$FD,$AF
	.dc.b	$1F,$E0,$2F,$AD,$FC,$BE,$38,$BF,$C3,$DE,$DE,$46,$30,$EB,$F0,$8C
	.dc.b	$3A,$C6,$1E,$9D,$5F,$E5,$E9,$D5,$71,$CB,$B2,$C3,$B4,$EA,$31,$87
	.dc.b	$37,$F3,$39,$79,$9F,$19,$AB,$E7,$7C,$B1,$F1,$C7,$6A,$E1,$87,$5A
	.dc.b	$D5,$F8,$CE,$6E,$14,$F0,$BE,$71,$AC,$3C,$8F,$84,$F1,$CE,$DE,$97
	.dc.b	$DE,$77,$7D,$6B,$55,$C6,$B5,$87,$F6,$8A,$FA,$13,$FE,$35,$6B,$E2
	.dc.b	$3F,$C7,$1F,$1C,$F9,$37,$47,$F7,$08,$C3,$63,$33,$FE,$3C,$9C,$BF
	.dc.b	$E3,$EA,$7C,$DB,$E4,$3C,$4D,$9D,$35,$F2,$35,$FC,$6B,$BC,$BE,$73
	.dc.b	$75,$CD,$7E,$03,$EB,$7F,$CC,$17,$F6,$2B,$4F,$E4,$1A,$FE,$15,$B1
	.dc.b	$C2,$3E,$BC,$6B,$0F,$4D,$73,$3B,$7C,$27,$F9,$7B,$1F,$1C,$66,$F8
	.dc.b	$BA,$71,$1A,$72,$7E,$3B,$0F,$04,$64,$98,$FB,$27,$57,$DE,$31,$3A
	.dc.b	$9F,$65,$F7,$7F,$57,$E3,$3E,$B5,$F0,$1F,$E7,$4E,$CF,$F6,$2F,$4F
	.dc.b	$FA,$EB,$D1,$C7,$3F,$F8,$ED,$B8,$B1,$E5,$86,$1B,$1F,$E7,$A6,$BF
	.dc.b	$E7,$F6,$BF,$E8,$26,$BF,$D1,$B8,$DF,$57,$E4,$8F,$F4,$1C,$62,$BF
	.dc.b	$C7,$AD,$8C,$CE,$DF,$F4,$64,$65,$FF,$6D,$B7,$2F,$EA,$13,$B7,$FD
	.dc.b	$C7,$3F,$C2,$FC,$5F,$6D,$78,$BE,$B3,$DF,$A7,$FB,$71,$BE,$C9,$DE
	.dc.b	$37,$0F,$EB,$0D,$7C,$6D,$7F,$D4,$37,$D5,$F4,$CF,$B2,$78,$63,$E6
	.dc.b	$8E,$A7,$D0,$9F,$E4,$1F,$59,$E6,$E8,$FF,$58,$C7,$F8,$EB,$E5,$BF
	.dc.b	$4D,$7F,$E1,$36,$F0,$7C,$DA,$DD,$7C,$46,$BE,$46,$BB,$F8,$9F,$9F
	.dc.b	$18,$8F,$DF,$BC,$AF,$FB,$D3,$6B,$84,$FF,$91,$7E,$AF,$C2,$F0,$9F
	.dc.b	$93,$F1,$B2,$3F,$BF,$47,$F0,$B6,$BE,$02,$3F,$82,$74,$FD,$E6,$DF
	.dc.b	$89,$F8,$87,$C7,$A7,$CA,$7E,$57,$E8,$BF,$AB,$5F,$C7,$F4,$BF,$31
	.dc.b	$A1,$D7,$85,$27,$F3,$55,$C1,$FC,$D3,$5F,$CD,$35,$FF,$0F,$F3,$23
	.dc.b	$86,$7F,$E4,$56,$EB,$A6,$D7,$F8,$9B,$EA,$BF,$DC,$95,$BF,$A6,$FC
	.dc.b	$7F,$C2,$4A,$FF,$3B,$DA,$7D,$87,$FA,$83,$1F,$03,$63,$53,$EA,$8C
	.dc.b	$1C,$1E,$6F,$E3,$70,$FE,$F3,$1F,$E7,$2C,$7F,$D8,$1F,$3C,$F8,$AF
	.dc.b	$E0,$39,$A3,$15,$FE,$BD,$8C,$4F,$E5,$FE,$A7,$FD,$17,$3F,$E8,$76
	.dc.b	$BF,$EF,$E6,$BF,$E4,$63,$FC,$8E,$E1,$C3,$5F,$F2,$5F,$D5,$7F,$91
	.dc.b	$DF,$F5,$FB,$5E,$46,$BB,$3F,$D8,$6A,$34,$F7,$EC,$76,$3D,$1F,$DE
	.dc.b	$92,$7F,$D4,$FF,$52,$FA,$AF,$D9,$70,$7C,$09,$EE,$7D,$3D,$3B,$C6
	.dc.b	$63,$C6,$D7,$DC,$FF,$9D,$9F,$F5,$57,$23,$E4,$7F,$7C,$D8,$FF,$57
	.dc.b	$4E,$DF,$A8,$7F,$9B,$3E,$AB,$D2,$FF,$AC,$D7,$CB,$FF,$07,$6C,$67
	.dc.b	$4C,$1F,$12,$38,$C7,$A9,$AF,$CA,$6B,$F5,$5D,$F0,$FF,$28,$6F,$1E
	.dc.b	$6B,$C1,$F0,$7F,$C9,$BE,$7A,$61,$FE,$E3,$7F,$95,$FD,$53,$E5,$B3
	.dc.b	$3E,$CE,$3F,$A7,$58,$CF,$FA,$9D,$EE,$BF,$43,$F3,$5F,$CC,$BD,$99
	.dc.b	$18,$21,$F1,$BB,$3F,$65,$3F,$C0,$4F,$E7,$7F,$35,$FC,$12,$7F,$DC
	.dc.b	$3E,$9D,$7C,$5B,$3F,$05,$FE,$7E,$7F,$5D,$9D,$67,$FE,$BF,$78,$57
	.dc.b	$F5,$B6,$BF,$C0,$9A,$FF,$67,$6E,$3F,$E2,$0F,$0F,$A1,$27,$FD,$4F
	.dc.b	$5F,$DC,$B1,$FF,$2A,$3F,$DD,$31,$FF,$5C,$CE,$B8,$7F,$D5,$BB,$FE
	.dc.b	$6B,$DA,$FE,$AA,$7F,$73,$FA,$BF,$CB,$61,$C6,$33,$7C,$45,$71,$71
	.dc.b	$9A,$F8,$E1,$EB,$F8,$8F,$EC,$3F,$3D,$2F,$E4,$C7,$C5,$23,$83,$B6
	.dc.b	$0D,$1F,$49,$FD,$6F,$EA,$DF,$29,$FD,$77,$F3,$B7,$93,$4D,$44,$C7
	.dc.b	$F5,$E2,$B8,$B8,$6B,$FB,$46,$BF,$C1,$1C,$F0,$7F,$80,$1B,$AF,$F1
	.dc.b	$16,$CF,$FB,$63,$EA,$5F,$1F,$F8,$1F,$33,$FC,$07,$FA,$87,$F8,$23
	.dc.b	$1F,$F1,$3B,$71,$FC,$04,$EF,$0F,$C6,$B5,$FC,$F6,$CE,$0F,$EF,$9F
	.dc.b	$58,$ED,$A4,$AF,$AC,$23,$FD,$CE,$E1,$E0,$D8,$E0,$FF,$06,$38,$7C
	.dc.b	$57,$F8,$27,$EB,$1F,$51,$FE,$1C,$F9,$3F,$E6,$7B,$7F,$90,$1B,$AF
	.dc.b	$F3,$3A,$DE,$1E,$26,$BF,$C8,$1A,$FF,$30,$6B,$C6,$FF,$98,$C7,$0B
	.dc.b	$FF,$17,$B6,$B5,$5E,$EE,$0F,$C0,$BF,$88,$EB,$F7,$BF,$B4,$CB,$FC
	.dc.b	$62,$7F,$93,$FE,$AF,$38,$FE,$96,$33,$3C,$5D,$1F,$E3,$8F,$AC,$91
	.dc.b	$D5,$E5,$F4,$18,$8F,$E0,$31,$69,$97,$F7,$CC,$17,$FE,$33,$F8,$87
	.dc.b	$CA,$AD,$8F,$F8,$43,$83,$FA,$E7,$B1,$FA,$08,$CF,$85,$F6,$14,$65
	.dc.b	$E0,$72,$C3,$AD,$F0,$BF,$EC,$D1,$89,$7D,$5F,$3B,$E4,$DE,$EF,$ED
	.dc.b	$8F,$7A,$E0,$DC,$BD,$17,$CB,$0F,$3D,$F0,$6C,$7B,$2B,$86,$36,$7C
	.dc.b	$77,$CA,$7C,$15,$93,$E1,$E3,$E4,$AD,$DE,$65,$8F,$AF,$2E,$31,$F0
	.dc.b	$8F,$8B,$97,$F8,$16,$3C,$31,$99,$F5,$DF,$9A,$3D,$47,$82,$BD,$4F
	.dc.b	$CB,$CB,$CD,$3E,$28,$ED,$8F,$D1,$C3,$C2,$76,$3B,$E5,$E4,$25,$71
	.dc.b	$72,$F8,$5C,$D7,$7C,$BD,$D9,$79,$A3,$C8,$7C,$1C,$3E,$23,$B3,$D2
	.dc.b	$7C,$4C,$7E,$AE,$3E,$4C,$7D,$CF,$95,$F1,$1E,$12,$4F,$3C,$7C,$AF
	.dc.b	$80,$EB,$7E,$07,$B6,$3E,$6C,$3B,$DF,$62,$4F,$97,$0E,$6F,$D4,$3D
	.dc.b	$59,$7C,$E7,$99,$EA,$CB,$E0,$3B,$3C,$31,$B8,$C9,$F8,$1A,$EF,$3B
	.dc.b	$AF,$73,$EB,$8F,$6B,$CF,$2F,$1D,$76,$CB,$D8,$7B,$5E,$B8,$FE,$02
	.dc.b	$BE,$15,$7D,$67,$CE,$F6,$7C,$17,$E2,$AF,$24,$78,$4E,$31,$E2,$8E
	.dc.b	$B9,$7E,$D6,$7E,$76,$3F,$5D,$F9,$15,$F8,$F8,$C4,$76,$32,$7E,$07
	.dc.b	$E1,$7F,$C2,$57,$FB,$D6,$E3,$FB,$66,$E1,$FC,$39,$FF,$07,$FC,$35
	.dc.b	$C3,$FC,$A6,$DC,$3F,$CC,$6D,$7F,$CD,$CD,$7F,$8B,$BE,$97,$FD,$F9
	.dc.b	$F5,$9F,$C8,$1F,$D9,$1A,$3F,$69,$E2,$A3,$55,$FC,$6E,$74,$FF,$1D
	.dc.b	$9D,$70,$F5,$9C,$E9,$FF,$00,$6F,$8B,$E8,$8F,$D0,$D7,$07,$F4,$67
	.dc.b	$87,$5E,$95,$E5,$7C,$10,$C7,$D5,$5C,$71,$CC,$62,$FB,$D7,$C4,$3B
	.dc.b	$BC,$F0,$F1,$61,$DB,$1F,$E2,$D5,$E1,$9F,$44,$62,$3C,$07,$27,$17
	.dc.b	$FC,$77,$1F,$55,$F8,$AB,$47,$B0,$E3,$8F,$77,$E7,$98,$C3,$BB,$8B
	.dc.b	$E9,$8F,$9E,$3D,$57,$C4,$EC,$5A,$F5,$9C,$9F,$15,$F7,$2C,$76,$78
	.dc.b	$BB,$38,$CF,$FB,$72,$BC,$0F,$E2,$8F,$5D,$70,$75,$8F,$C2,$7F,$0F
	.dc.b	$3F,$41,$E1,$7E,$1B,$EB,$1F,$90,$AE,$17,$E9,$8F,$6C,$F8,$DA,$76
	.dc.b	$3E,$56,$1F,$12,$B8,$CE,$E7,$BB,$ED,$3F,$1D,$7E,$18,$E6,$78,$4E
	.dc.b	$67,$B2,$36,$71,$7F,$11,$1C,$5F,$85,$7E,$76,$CF,$85,$FC,$41,$E7
	.dc.b	$8F,$8C,$F7,$6E,$3F,$86,$9F,$86,$D9,$EF,$1E,$17,$D6,$7C,$1A,$FF
	.dc.b	$20,$F8,$7F,$ED,$AE,$8F,$DC,$6B,$FE,$87,$6E,$5F,$E1,$4C,$5F,$59
	.dc.b	$FE,$C6,$8F,$F1,$74,$EB,$1F,$F4,$1F,$3F,$F9,$1A,$72,$7F,$A1,$2B
	.dc.b	$84,$FF,$A9,$AB,$86,$3F,$EE,$AE,$4B,$87,$F8,$9B,$0E,$1F,$53,$89
	.dc.b	$91,$FE,$3D,$7C,$19,$79,$27,$38,$FF,$96,$7E,$AD,$E6,$79,$61,$18
	.dc.b	$C3,$FC,$CB,$C3,$FE,$FA,$7F,$CC,$F1,$99,$E5,$5F,$C6,$DF,$F1,$A9
	.dc.b	$FE,$36,$8D,$38,$3A,$C6,$DD,$C6,$9B,$1F,$03,$17,$CC,$79,$6B,$C1
	.dc.b	$BF,$F2,$9D,$7F,$C8,$F3,$E9,$F8,$3F,$F4,$EE,$5F,$E7,$C6,$E5,$FE
	.dc.b	$77,$FA,$A7,$D4,$EF,$CA,$97,$B7,$E1,$4E,$A3,$07,$FA,$E3,$EA,$BE
	.dc.b	$AE,$7F,$CE,$78,$51,$89,$FF,$63,$F8,$CF,$90,$FA,$33,$9C,$D7,$C3
	.dc.b	$D8,$FB,$BE,$53,$F0,$FD,$C6,$1F,$A9,$F5,$8F,$76,$98,$F8,$AB,$AF
	.dc.b	$FA,$6A,$74,$FF,$AE,$9B,$97,$D1,$0E,$1F,$A8,$D7,$EC,$BC,$36,$FB
	.dc.b	$9F,$A9,$BC,$2E,$F5,$FE,$A1,$8C,$5F,$C5,$EB,$F7,$9F,$98,$E1,$7C
	.dc.b	$C6,$27,$47,$FB,$63,$77,$D0,$E7,$1F,$F3,$F4,$FF,$BC,$27,$2F,$FA
	.dc.b	$5F,$E4,$BF,$42,$7E,$73,$C1,$7B,$D8,$9D,$12,$F9,$9F,$42,$77,$19
	.dc.b	$DB,$ED,$72,$B9,$FE,$B9,$E9,$F8,$3B,$5F,$81,$8F,$CC,$61,$F6,$AB
	.dc.b	$84,$FC,$DE,$7F,$C1,$D6,$9B,$19,$9D,$D7,$93,$B5,$AB,$E0,$FB,$CE
	.dc.b	$63,$17,$F9,$1F,$81,$FC,$59,$F3,$7B,$6C,$64,$F9,$9E,$F6,$2F,$93
	.dc.b	$FC,$23,$EB,$64,$FF,$B0,$66,$9E,$3F,$19,$38,$E0,$FD,$74,$7F,$AC
	.dc.b	$2B,$FD,$91,$3F,$C1,$FE,$67,$F5,$B5,$FC,$05,$B3,$FC,$09,$FE,$2D
	.dc.b	$7F,$EE,$77,$2F,$FA,$E9,$FD,$AB,$FC,$15,$AF,$1F,$07,$E7,$A8,$CB
	.dc.b	$C6,$3F,$8C,$6D,$FA,$9A,$FF,$7D,$D7,$6A,$4E,$5F,$C1,$FD,$5B,$EB
	.dc.b	$1F,$59,$FC,$87,$EB,$1F,$01,$FC,$DE,$B5,$39,$8F,$E5,$7F,$4D,$FC
	.dc.b	$FB,$3F,$C0,$37,$1F,$E3,$DC,$C8,$C4,$7F,$24,$8C,$74,$FE,$2B,$4F
	.dc.b	$C0,$FC,$5F,$F0,$23,$FA,$37,$D5,$BE,$B1,$F3,$99,$CF,$FD,$C3,$C7
	.dc.b	$FC,$BF,$EA,$9F,$19,$F1,$9B,$3F,$BF,$D5,$18,$7F,$81,$35,$FE,$6F
	.dc.b	$4F,$AA,$75,$FE,$7A,$30,$D8,$CD,$FF,$0B,$9C,$BF,$A1,$F6,$FF,$7E
	.dc.b	$F9,$8F,$EC,$8E,$2F,$F6,$73,$C1,$FD,$AF,$07,$FB,$5B,$C5,$F9,$BF
	.dc.b	$07,$30,$E5,$5F,$F0,$03,$C6,$7C,$2E,$63,$8F,$2A,$FF,$60,$DB,$FA
	.dc.b	$B1,$82,$3F,$C4,$7D,$0E,$59,$FF,$4B,$F0,$BF,$9B,$69,$F6,$74,$FB
	.dc.b	$E7,$F6,$5D,$1E,$EE,$31,$F2,$D6,$DF,$C1,$D7,$88,$8E,$0D,$FE,$36
	.dc.b	$FA,$6E,$5B,$7C,$28,$C4,$FC,$0F,$A0,$7C,$2B,$E5,$FF,$C1,$51,$87
	.dc.b	$FB,$C7,$D5,$3E,$47,$35,$BB,$E4,$DC,$7F,$B9,$61,$C6,$7F,$A2,$36
	.dc.b	$33,$3B,$7F,$BD,$C6,$78,$3E,$37,$83,$DB,$2F,$EB,$2F,$83,$EA,$A6
	.dc.b	$3F,$63,$06,$7E,$A8,$F1,$CE,$A7,$BB,$FC,$B3,$1F,$7B,$72,$34,$FF
	.dc.b	$0D,$E8,$FA,$1E,$5A,$7F,$73,$8C,$63,$F2,$7E,$A9,$F5,$3F,$21,$B1
	.dc.b	$E5,$D1,$F3,$79,$A9,$3F,$12,$B7,$AF,$C8,$6B,$F1,$1A,$86,$A8,$E3
	.dc.b	$97,$E3,$7E,$6B,$AD,$7F,$3D,$7C,$9C,$33,$0F,$27,$81,$6B,$FA,$CE
	.dc.b	$3C,$EB,$E3,$63,$F7,$EF,$CD,$18,$3F,$A8,$71,$E6,$FF,$89,$B5,$FE
	.dc.b	$24,$D7,$FC,$B0,$D7,$FC,$2A,$F1,$CB,$F4,$9E,$C7,$F9,$2A,$31,$7F
	.dc.b	$DD,$63,$FC,$91,$F5,$1F,$D6,$7C,$DF,$F6,$FF,$D3,$FF,$90,$A3,$18
	.dc.b	$FF,$B2,$63,$FC,$8B,$F5,$3F,$C0,$FE,$21,$1B,$F5,$BF,$9F,$56,$CF
	.dc.b	$99,$7C,$AB,$71,$89,$DC,$62,$17,$D6,$FA,$61,$E4,$9E,$0D,$58,$75
	.dc.b	$9F,$2F,$5F,$FB,$D7,$0E,$93,$9A,$E2,$79,$4E,$38,$F8,$AF,$8E,$7D
	.dc.b	$6B,$85,$6B,$1C,$5E,$E3,$F1,$19,$79,$DE,$B8,$F8,$27,$51,$C4,$F0
	.dc.b	$4E,$DC,$4E,$EB,$A9,$ED,$9D,$5F,$88,$FA,$75,$FE,$73,$6B,$F1,$8F
	.dc.b	$F2,$33,$87,$F3,$9E,$23,$E9,$7F,$3B,$1F,$E4,$B8,$C7,$17,$F4,$F6
	.dc.b	$C6,$67,$6F,$7F,$5B,$FC,$94,$F6,$6B,$9F,$7F,$FA,$C6,$09,$CC,$FE
	.dc.b	$2D,$E3,$9F,$D4,$6E,$D9,$BF,$B9,$F5,$3F,$E7,$1B,$FA,$55,$A7,$F1
	.dc.b	$6D,$7F,$02,$D8,$E1,$1F,$77,$0F,$15,$72,$24,$7E,$06,$77,$87,$8E
	.dc.b	$72,$D9,$D7,$C2,$FF,$A9,$23,$04,$C7,$D1,$3A,$38,$B8,$8D,$47,$F0
	.dc.b	$DB,$EF,$39,$24,$7B,$2F,$E0,$DF,$79,$FF,$35,$78,$3F,$DF,$B8,$7F
	.dc.b	$CD,$9F,$4F,$70,$FF,$12,$F8,$C8,$C7,$A5,$FE,$54,$BB,$3F,$D4,$6B
	.dc.b	$FC,$F1,$58,$33,$3C,$F8,$7F,$C5,$FF,$5D,$FE,$11,$6D,$7F,$4E,$C1
	.dc.b	$39,$6C,$66,$39,$46,$23,$F9,$0B,$7A,$7F,$BB,$7B,$7D,$FE,$C9,$F2
	.dc.b	$5E,$BB,$FE,$17,$95,$DB,$D1,$1F,$CA,$AB,$6F,$F6,$D6,$BF,$D9,$1A
	.dc.b	$FF,$60,$6B,$FC,$F7,$34,$62,$3F,$9F,$BC,$B1,$6D,$B8,$C3,$71,$F9
	.dc.b	$7F,$17,$F1,$FE,$C9,$E1,$79,$FE,$12,$65,$FE,$01,$9C,$B6,$7E,$64
	.dc.b	$62,$3F,$BD,$B5,$FE,$EE,$D7,$D0,$EF,$2F,$2C,$7F,$56,$8C,$63,$AA
	.dc.b	$FB,$FE,$4F,$97,$7F,$99,$F5,$5F,$B6,$C7,$FD,$FC,$DB,$EB,$3B,$C7
	.dc.b	$FC,$17,$8A,$3C,$E4,$8C,$BB,$7F,$AE,$C6,$DF,$F6,$27,$C2,$7E,$57
	.dc.b	$CE,$7D,$23,$DC,$F1,$F3,$4B,$FF,$18,$CE,$5B,$3F,$E6,$1A,$E2,$D9
	.dc.b	$D4,$FF,$9D,$9A,$FF,$9D,$5C,$E9,$FE,$FD,$6E,$7F,$DF,$6F,$9D,$7F
	.dc.b	$B6,$7E,$CF,$CD,$5F,$8E,$BF,$9B,$2F,$AA,$FF,$02,$3F,$E2,$36,$BF
	.dc.b	$E6,$09,$E1,$87,$57,$F2,$F1,$E3,$E7,$F4,$11,$FF,$67,$47,$05,$8F
	.dc.b	$C6,$9F,$19,$FB,$8F,$BB,$FD,$26,$1F,$E1,$EA,$F1,$79,$D8,$3F,$51
	.dc.b	$F3,$8F,$A1,$A7,$FA,$26,$51,$97,$FD,$18,$D7,$D6,$6A,$3F,$D2,$1D
	.dc.b	$9F,$ED,$CF,$77,$9C,$FF,$90,$EB,$86,$DF,$E6,$F9,$FD,$C7,$D4,$FC
	.dc.b	$0A,$F8,$ED,$7F,$C5,$8D,$4D,$FA,$AF,$E8,$B1,$F3,$9A,$FF,$8C,$DA
	.dc.b	$9C,$D3,$EE,$FE,$24,$FF,$A5,$1B,$1F,$A6,$35,$9F,$F2,$7F,$B7,$FE
	.dc.b	$33,$E9,$7E,$03,$E6,$BF,$88,$71,$FF,$57,$C7,$FC,$C0,$D7,$EF,$46
	.dc.b	$DF,$B2,$FE,$4C,$8E,$0F,$CD,$BB,$58,$7E,$E3,$91,$F5,$9F,$E7,$F9
	.dc.b	$FF,$52,$F0,$3C,$66,$1F,$CE,$3C,$09,$5A,$9C,$B6,$72,$F6,$C3,$FC
	.dc.b	$B5,$7C,$EF,$FA,$0D,$7F,$44,$FD,$4B,$F8,$05,$FF,$BB,$17,$CF,$64
	.dc.b	$D7,$95,$27,$CB,$3A,$6C,$E5,$FF,$60,$FD,$43,$E6,$7B,$7D,$AC,$9F
	.dc.b	$F5,$6C,$E5,$B3,$FB,$D8,$C4,$7F,$BC,$1A,$FF,$BB,$DA,$FE,$E5,$DE
	.dc.b	$FF,$C8,$BB,$51,$8D,$B5,$7E,$FF,$2F,$E0,$7C,$6C,$C7,$D6,$FC,$39
	.dc.b	$91,$82,$1F,$2C,$F8,$CF,$C9,$FA,$4F,$95,$F1,$32,$7E,$2C,$65,$F8
	.dc.b	$F1,$97,$97,$13,$75,$7F,$5F,$CF,$FC,$69,$5F,$90,$EC,$FC,$99,$9F
	.dc.b	$56,$47,$ED,$8F,$7B,$FB,$D9,$E3,$1F,$BA,$6B,$FC,$59,$CE,$FF,$C7
	.dc.b	$1B,$B7,$F1,$C6,$CF,$FA,$87,$EA,$9E,$FF,$F8,$2B,$E6,$9C,$77,$92
	.dc.b	$5F,$E6,$27,$2D,$9C,$BF,$02,$3F,$10,$D8,$E9,$C5,$23,$93,$F9,$5F
	.dc.b	$87,$91,$82,$1F,$2B,$83,$FD,$F2,$FF,$B5,$7E,$DF,$FC,$19,$AF,$F3
	.dc.b	$6F,$A8,$FE,$2A,$DF,$17,$FC,$19,$3A,$AF,$F5,$F3,$FE,$15,$7F,$9D
	.dc.b	$B5,$FE,$7E,$D7,$7E,$5F,$F2,$B8,$E5,$06,$9D,$79,$5F,$79,$FB,$69
	.dc.b	$FC,$50,$F9,$AF,$E8,$0D,$F2,$FF,$C4,$11,$FD,$0E,$BF,$0D,$CD,$FB
	.dc.b	$73,$33,$2B,$F6,$71,$F5,$48,$FD,$37,$F7,$DF,$59,$23,$04,$3E,$53
	.dc.b	$FE,$2F,$F7,$DC,$1F,$C5,$A7,$F8,$1F,$D3,$FF,$60,$FA,$9F,$CB,$C7
	.dc.b	$FB,$97,$99,$39,$EC,$FE,$CB,$18,$9F,$F4,$7D,$F3,$8C,$43,$0E,$FD
	.dc.b	$7F,$EE,$FF,$A9,$CE,$5D,$A9,$18,$C7,$FB,$8B,$CA,$F8,$CE,$B2,$FE
	.dc.b	$EB,$18,$9E,$13,$D6,$7C,$CD,$3C,$B8,$57,$C9,$97,$C1,$AE,$31,$D4
	.dc.b	$0C,$7C,$D1,$E3,$9E,$17,$E2,$7E,$03,$E0,$AF,$96,$72,$AF,$76,$5C
	.dc.b	$70,$F9,$95,$C5,$F2,$4F,$68,$E5,$86,$EF,$C8,$7A,$2B,$8B,$D7,$2F
	.dc.b	$A7,$5E,$1C,$38,$1C,$6F,$DA,$48,$F0,$57,$4C,$BE,$54,$78,$09,$1F
	.dc.b	$20,$D6,$5F,$01,$F0,$1E,$8C,$3E,$BC,$F4,$BF,$0B,$E3,$24,$F3,$C3
	.dc.b	$C9,$1D,$A3,$E7,$57,$27,$C1,$5A,$7B,$9F,$0B,$2F,$8C,$EA,$BD,$77
	.dc.b	$F6,$88,$E0,$F2,$1D,$E7,$8C,$FC,$B3,$DB,$97,$AD,$86,$09,$5D,$1F
	.dc.b	$1B,$EE,$7C,$47,$9A,$FE,$73,$83,$B3,$87,$DE,$7E,$4F,$2F,$A2,$EE
	.dc.b	$FC,$A7,$9A,$33,$1E,$7A,$F1,$3C,$8F,$C7,$99,$C3,$DE,$F8,$DF,$9D
	.dc.b	$3C,$8F,$D5,$CF,$2A,$FC,$AC,$FC,$7A,$FA,$CF,$D8,$7B,$4F,$BD,$F9
	.dc.b	$1E,$17,$F3,$EB,$FD,$AB,$73,$FE,$18,$DC,$BF,$56,$FF,$44,$AF,$D6
	.dc.b	$FC,$06,$30,$FF,$29,$B6,$3F,$CD,$ED,$7F,$DE,$42,$79,$28,$FD,$1B
	.dc.b	$FE,$6C,$F6,$FF,$D4,$F5,$FC,$4A,$3B,$9F,$1B,$C4,$FF,$68,$F4,$70
	.dc.b	$8C,$CF,$F1,$89,$CC,$7C,$C3,$3A,$7F,$A1,$23,$33,$D6,$7C,$47,$ED
	.dc.b	$A3,$8E,$D9,$C7,$F8,$84,$62,$FD,$F5,$E3,$C7,$CF,$18,$C3,$C5,$18
	.dc.b	$9F,$86,$75,$7F,$D8,$78,$75,$8C,$63,$FC,$75,$F7,$C7,$28,$C4,$79
	.dc.b	$A3,$0F,$F0,$C3,$77,$ED,$AE,$AF,$9A,$76,$77,$BE,$25,$8E,$17,$F2
	.dc.b	$EF,$B4,$61,$EC,$60,$FF,$5A,$5F,$67,$DB,$7D,$A7,$C0,$71,$7B,$47
	.dc.b	$0A,$ED,$8F,$E0,$31,$FA,$77,$C2,$B4,$78,$27,$E8,$E1,$DE,$3F,$17
	.dc.b	$87,$77,$C3,$5D,$CF,$A4,$7A,$71,$F0,$C7,$E3,$71,$F0,$BF,$40,$EC
	.dc.b	$5A,$FA,$87,$57,$BD,$75,$7B,$9E,$68,$DC,$66,$3F,$3F,$5F,$3A,$3C
	.dc.b	$71,$EC,$8D,$BC,$63,$E0,$3E,$53,$C7,$1C,$5F,$C4,$B5,$F1,$9C,$4F
	.dc.b	$19,$C6,$7A,$B6,$7A,$E3,$F9,$98,$DB,$5E,$66,$CE,$F1,$F2,$4F,$4B
	.dc.b	$E5,$AF,$F0,$47,$1F,$F9,$97,$E4,$BF,$D7,$18,$FF,$A6,$BB,$9F,$33
	.dc.b	$8C,$3F,$D3,$15,$6D,$5A,$BF,$F2,$A5,$E6,$31,$7F,$E4,$9B,$C4,$F1
	.dc.b	$ED,$CF,$07,$FC,$00,$F5,$DF,$49,$9A,$F0,$60,$9C,$E5,$FE,$04,$BE
	.dc.b	$17,$A9,$CD,$7F,$95,$EF,$13,$CB,$E9,$DE,$5D,$3D,$5D,$AF,$93,$C6
	.dc.b	$7D,$4F,$F0,$8E,$2B,$E3,$1F,$E8,$36,$C7,$FB,$E9,$AF,$03,$85,$A7
	.dc.b	$83,$EC,$FA,$5F,$F5,$C4,$77,$9F,$F0,$FD,$74,$D3,$84,$7F,$88,$AB
	.dc.b	$6F,$F9,$CF,$E1,$7E,$76,$7F,$2D,$B8,$7F,$A1,$62,$CE,$E3,$07,$FA
	.dc.b	$27,$C2,$FA,$DE,$0F,$D7,$6C,$64,$FF,$47,$46,$5A,$79,$E5,$82,$BD
	.dc.b	$87,$D8,$FD,$97,$D5,$5F,$2A,$74,$F1,$6D,$7C,$73,$9D,$7C,$07,$E5
	.dc.b	$46,$9B,$19,$7A,$3F,$86,$F8,$B7,$D4,$F8,$AF,$CD,$18,$7F,$DD,$7F
	.dc.b	$7B,$F9,$1F,$BD,$FC,$93,$63,$27,$FB,$C2,$32,$D3,$FD,$E7,$F2,$1F
	.dc.b	$99,$D7,$F3,$4D,$8F,$F7,$BC,$5C,$BF,$D5,$13,$77,$FF,$74,$7A,$4F
	.dc.b	$8B,$FD,$67,$77,$FB,$E5,$FD,$27,$D3,$E2,$75,$97,$E1,$6B,$4F,$C7
	.dc.b	$6E,$3F,$8F,$6B,$FD,$96,$78,$B9,$6E,$1F,$8C,$ED,$FF,$72,$71,$F7
	.dc.b	$CB,$F6,$8F,$ED,$78,$7F,$7D,$FA,$A7,$F9,$CB,$3F,$CB,$5F,$FB,$AE
	.dc.b	$73,$96,$E7,$35,$B8,$FF,$64,$4E,$A7,$3C,$3D,$3C,$2E,$AD,$8C,$9C
	.dc.b	$A3,$2D,$3E,$16,$8F,$53,$E3,$1F,$C5,$61,$FA,$43,$A3,$31,$C2,$7F
	.dc.b	$11,$93,$81,$87,$3A,$F9,$8D,$9F,$91,$5C,$2B,$6F,$F1,$4D,$5F,$27
	.dc.b	$DA,$BF,$D9,$F1,$8C,$3F,$79,$F4,$6F,$BE,$7C,$7F,$F2,$B8,$C3,$F3
	.dc.b	$BD,$3C,$4E,$AB,$F0,$1F,$7A,$F5,$FC,$4D,$8F,$C1,$37,$2F,$C1,$35
	.dc.b	$FE,$25,$F4,$6F,$67,$E0,$C2,$31,$5F,$B1,$8C,$4E,$A7,$31,$FC,$6F
	.dc.b	$47,$33,$E6,$FF,$33,$8F,$F1,$4E,$AF,$C6,$4E,$72,$F4,$F1,$E8,$F4
	.dc.b	$91,$F4,$BE,$4F,$C1,$3F,$2B,$1F,$CC,$1B,$97,$D7,$C3,$B4,$E7,$AF
	.dc.b	$C3,$EB,$48,$FC,$8F,$CD,$DB,$67,$FC,$AB,$9D,$8B,$0F,$96,$4C,$3E
	.dc.b	$5B,$6F,$8C,$7F,$42,$FA,$1F,$F7,$BB,$F5,$27,$FA,$AC,$62,$76,$79
	.dc.b	$FC,$2F,$8A,$7E,$AB,$D1,$B7,$F7,$23,$ED,$91,$FB,$6D,$9E,$62,$9E
	.dc.b	$CF,$28,$CC,$62,$BF,$9A,$7D,$F7,$E2,$67,$8B,$74,$FA,$BF,$0F,$FD
	.dc.b	$B2,$31,$5F,$D7,$9E,$15,$FD,$65,$AE,$7A,$BF,$13,$18,$8F,$F7,$AB
	.dc.b	$C9,$FF,$7B,$FC,$FB,$E8,$BB,$57,$F7,$A8,$C5,$F9,$BE,$47,$C4,$7F
	.dc.b	$0E,$AB,$19,$3F,$29,$82,$7C,$24,$8C,$BD,$7E,$43,$8E,$3F,$23,$04
	.dc.b	$7F,$82,$BE,$F9,$C9,$4F,$E3,$B1,$70,$FE,$B9,$3A,$7F,$48,$DB,$FE
	.dc.b	$AC,$D7,$FB,$F3,$5F,$F0,$3B,$BD,$7F,$AD,$61,$FA,$A7,$FA,$E4,$66
	.dc.b	$BF,$C1,$B3,$97,$8E,$7E,$73,$F9,$8F,$D9,$D8,$F2,$C7,$EE,$AB,$E5
	.dc.b	$E1,$E4,$8C,$1F,$C7,$BA,$FF,$6F,$E7,$7F,$3A,$FB,$0F,$E7,$6D,$C3
	.dc.b	$F9,$BB,$6F,$F3,$35,$C2,$7F,$9D,$D7,$F8,$CB,$E7,$FF,$8F,$F9,$5F
	.dc.b	$EC,$8F,$8B,$FF,$12,$C6,$30,$FF,$61,$47,$F8,$A3,$E8,$5E,$87,$F8
	.dc.b	$6B,$E0,$13,$C3,$CA,$63,$CF,$00,$C7,$A5,$F0,$8C,$C6,$2F,$B6,$1D
	.dc.b	$5E,$97,$CE,$B9,$61,$D9,$F2,$9C,$9C,$61,$D6,$18,$FA,$33,$E5,$97
	.dc.b	$79,$E1,$86,$30,$E3,$8F,$8A,$F1,$32,$BB,$9E,$27,$C9,$87,$58,$CE
	.dc.b	$1D,$5D,$18,$C3,$AB,$F1,$27,$86,$3E,$7A,$E0,$4A,$DD,$7A,$B2,$F8
	.dc.b	$8F,$A0,$E9,$87,$F3,$FA,$FB,$11,$FE,$57,$6B,$F2,$8F,$F1,$A7,$C1
	.dc.b	$FF,$10,$F9,$FF,$E8,$CD,$8C,$F6,$27,$2D,$8F,$F1,$E5,$3E,$85,$F0
	.dc.b	$9E,$16,$CE,$9A,$F8,$DA,$FE,$39,$DE,$9F,$05,$BB,$E6,$BF,$01,$F4
	.dc.b	$3F,$E0,$1B,$FB,$35,$A7,$F1,$2D,$7F,$1C,$D8,$E1,$1F,$52,$35,$87
	.dc.b	$A6,$B9,$9D,$BD,$1F,$F4,$2E,$3E,$38,$CD,$F1,$74,$E2,$34,$E4,$FC
	.dc.b	$2E,$1E,$08,$C9,$31,$F6,$4E,$AF,$BC,$62,$75,$3E,$CB,$EE,$FE,$7B
	.dc.b	$8D,$F2,$3F,$E4,$F8,$FF,$1A,$FD,$7F,$FA,$73,$AF,$FB,$BB,$79,$7F
	.dc.b	$63,$7C,$3C,$8C,$5A,$30,$F2,$21,$FC,$23,$0E,$CF,$F5,$8C,$DF,$47
	.dc.b	$FE,$0D,$8F,$E0,$2D,$8C,$C7,$F8,$45,$AF,$FA,$0D,$B9,$7F,$A0,$DB
	.dc.b	$8F,$FA,$D1,$B3,$F9,$1C,$3F,$D2,$4F,$FA,$56,$1F,$19,$EA,$E3,$7D
	.dc.b	$EB,$AB,$0D,$BE,$15,$F0,$1F,$E2,$EF,$9C,$8F,$9D,$AF,$BD,$E1,$7E
	.dc.b	$08,$FF,$5E,$5B,$E9,$1E,$CE,$BE,$0B,$C9,$D0,$F8,$AA,$67,$0F,$54
	.dc.b	$FC,$72,$3F,$1D,$AF,$CF,$78,$3F,$E1,$2C,$1F,$40,$FA,$4F,$67,$D1
	.dc.b	$FE,$BB,$8B,$FD,$9B,$5F,$E0,$CF,$53,$C6,$6B,$A2,$63,$FE,$45,$BF
	.dc.b	$D9,$BF,$EE,$46,$BF,$CE,$1A,$9E,$DC,$9B,$9E,$73,$E5,$C9,$F4,$7E
	.dc.b	$4F,$80,$FC,$7F,$37,$F9,$3B,$EA,$EF,$B0,$75,$87,$FB,$47,$3F,$BE
	.dc.b	$FF,$B2,$5A,$FF,$BC,$5A,$F9,$EF,$F2,$51,$C3,$6F,$F6,$73,$74,$D5
	.dc.b	$7F,$AB,$FE,$71,$F4,$9F,$83,$FA,$3F,$F7,$E6,$1F,$E4,$BF,$BD,$FF
	.dc.b	$40,$DF,$FA,$D5,$AF,$F8,$E2,$77,$87,$F5,$C8,$47,$E8,$DA,$7E,$95
	.dc.b	$C1,$1F,$01,$0F,$27,$57,$E5,$7D,$0F,$F0,$DC,$7F,$BC,$DA,$FF,$92
	.dc.b	$1B,$87,$FB,$AE,$9F,$43,$EB,$3C,$04,$AF,$F4,$14,$69,$FF,$42,$53
	.dc.b	$E0,$F8,$47,$FC,$01,$9D,$3D,$E3,$F8,$3D,$BE,$F5,$F1,$38,$6D,$7F
	.dc.b	$7E,$96,$76,$73,$3F,$E6,$0F,$A4,$FF,$50,$3B,$9F,$95,$F4,$BF,$AD
	.dc.b	$8F,$98,$7F,$82,$1C,$38,$6B,$C3,$5F,$E2,$55,$F9,$48,$FF,$07,$B5
	.dc.b	$39,$C1,$F4,$BE,$E7,$CF,$3F,$E1,$6E,$1F,$EB,$8F,$FB,$93,$E5,$DF
	.dc.b	$3F,$F9,$A6,$C7,$A4,$E1,$E7,$FF,$0D,$AE,$31,$EA,$0E,$1F,$6B,$5F
	.dc.b	$C5,$35,$E6,$75,$38,$C7,$53,$F9,$CF,$C2,$74,$FA,$3F,$D6,$F9,$6F
	.dc.b	$0F,$F5,$BF,$55,$30,$FF,$7E,$CE,$5B,$39,$7F,$C8,$BD,$AA,$E7,$3F
	.dc.b	$B8,$F9,$AF,$E3,$55,$EF,$EF,$FF,$63,$FD,$B7,$EE,$8D,$78,$F2,$75
	.dc.b	$8A,$75,$1C,$6F,$9B,$FE,$54,$7F,$06,$FF,$AB,$BE,$9F,$F1,$DD,$3F
	.dc.b	$D5,$9B,$1C,$7E,$3B,$C8,$67,$AB,$F1,$BC,$BF,$6A,$B7,$5D,$6B,$6F
	.dc.b	$D7,$8C,$CE,$EB,$8F,$57,$F6,$6C,$7C,$73,$FE,$CD,$FC,$85,$F2,$5F
	.dc.b	$C6,$3E,$D2,$65,$F8,$1E,$55,$6A,$30,$D7,$FD,$93,$1A,$AF,$B3,$39
	.dc.b	$BF,$C8,$F0,$B1,$FC,$6E,$1F,$53,$C1,$FE,$4F,$F2,$32,$3E,$6C,$7F
	.dc.b	$2A,$30,$FD,$A8,$C3,$F8,$9E,$9E,$DA,$FE,$3B,$C8,$FC,$B1,$F9,$7D
	.dc.b	$25,$7E,$5C,$FC,$BB,$FD,$F6,$BD,$51,$F9,$88,$C1,$C1,$FF,$6D,$C7
	.dc.b	$67,$FC,$FF,$F4,$FF,$ED,$C7,$F1,$1F,$41,$F8,$0D,$7F,$97,$75,$FE
	.dc.b	$6B,$C6,$3F,$D1,$A7,$59,$FF,$86,$1B,$6E,$15,$F0,$8F,$E6,$D7,$C7
	.dc.b	$A7,$F4,$BF,$79,$F7,$27,$FC,$3D,$B3,$E3,$3F,$A9,$B6,$3E,$D4,$EF
	.dc.b	$13,$49,$D3,$E5,$F0,$7F,$C1,$5F,$4F,$FC,$A6,$FF,$6D,$C0,$FA,$1F
	.dc.b	$F0,$73,$63,$F7,$53,$B7,$FE,$00,$B5,$FE,$3E,$D7,$FA,$63,$5F,$F5
	.dc.b	$4B,$53,$CF,$87,$F8,$CC,$63,$C8,$F2,$9F,$E1,$FF,$A7,$FE,$C5,$1F
	.dc.b	$EB,$2F,$BB,$FF,$7D,$3F,$C3,$57,$C9,$FF,$89,$5B,$19,$F8,$AC,$11
	.dc.b	$C1,$FC,$57,$E7,$3F,$9C,$63,$FC,$AB,$81,$F7,$5D,$AB,$FD,$A3,$C5
	.dc.b	$6B,$F3,$ED,$E8,$FF,$62,$DF,$0C,$7F,$8E,$E1,$9C,$3F,$00,$D9,$D4
	.dc.b	$7E,$33,$DA,$FF,$1E,$7C,$3B,$E9,$D9,$FF,$B4,$3A,$3D,$76,$FA,$87
	.dc.b	$D4,$FF,$79,$9C,$C6,$27,$87,$0F,$F7,$DC,$7A,$61,$89,$FF,$6D,$C6
	.dc.b	$17,$D4,$F2,$C7,$FC,$08,$D8,$FF,$70,$BE,$29,$E0,$DC,$BC,$F3,$CB
	.dc.b	$0F,$F0,$66,$3E,$38,$C1,$FE,$EA,$71,$5C,$71,$EB,$5E,$7B,$E7,$3E
	.dc.b	$1B,$E2,$7C,$1C,$7C,$B5,$BB,$E3,$3E,$CB,$F1,$E5,$C4,$95,$B9,$FF
	.dc.b	$2E,$37,$0E,$C4,$70,$6E,$78,$CF,$A7,$0F,$A1,$39,$37,$19,$3C,$67
	.dc.b	$9C,$B9,$7C,$88,$DC,$64,$F6,$E5,$F1,$23,$33,$B9,$E8,$6A,$BE,$0B
	.dc.b	$EF,$CB,$AE,$5F,$02,$FE,$8D,$79,$AB,$8E,$3F,$3D,$F8,$27,$13,$A1
	.dc.b	$E8,$C7,$C4,$E7,$2E,$07,$C5,$7C,$6E,$2F,$E1,$47,$67,$C2,$FE,$3C
	.dc.b	$65,$F8,$FB,$E1,$1C,$A7,$6F,$DE,$9F,$80,$F0,$C7,$F0,$73,$AB,$F1
	.dc.b	$4E,$72,$F8,$27,$7C,$7F,$21,$8F,$D8,$9D,$1F,$00,$B5,$CA,$7D,$F1
	.dc.b	$8B,$FA,$B1,$83,$38,$F9,$0F,$2B,$9C,$BE,$9D,$71,$3E,$66,$1E,$19
	.dc.b	$F1,$DF,$C6,$9D,$1F,$81,$C3,$D7,$87,$5C,$BE,$D4,$FC,$B2,$1E,$AC
	.dc.b	$3E,$D1,$D7,$1F,$BD,$19,$C3,$F3,$DE,$E7,$F8,$5B,$0F,$DF,$37,$0F
	.dc.b	$DB,$B7,$35,$A3,$FC,$3A,$E3,$E2,$33,$18,$84,$63,$2F,$F2,$33,$63
	.dc.b	$FC,$C2,$D7,$FD,$1C,$FE,$0F,$6F,$F6,$7F,$0F,$E1,$67,$B1,$F5,$63
	.dc.b	$2F,$D7,$33,$5F,$C8,$A3,$2F,$F2,$48,$CE,$7F,$E9,$B9,$CB,$E4,$8F
	.dc.b	$E1,$31,$D3,$5E,$11,$EA,$33,$F0,$3F,$E6,$A8,$CE,$3E,$78,$FF,$36
	.dc.b	$E1,$8A,$F3,$46,$30,$ED,$7E,$58,$E5,$18,$3C,$8E,$27,$E2,$9E,$0C
	.dc.b	$7F,$83,$BF,$0A,$B9,$46,$23,$CD,$18,$7F,$91,$9B,$AE,$B3,$BC,$3D
	.dc.b	$0F,$CD,$9F,$76,$1C,$4F,$55,$F3,$2D,$F2,$9E,$A7,$81,$C4,$6A,$3F
	.dc.b	$AF,$E1,$EB,$7E,$E9,$F1,$2F,$C1,$87,$8A,$3E,$D1,$CC,$E3,$1E,$77
	.dc.b	$E1,$1E,$08,$FE,$53,$8F,$D7,$C3,$E5,$4F,$03,$47,$73,$AC,$F7,$C7
	.dc.b	$CD,$1F,$E2,$FC,$3C,$CF,$C7,$39,$63,$AC,$3B,$47,$D4,$8F,$89,$87
	.dc.b	$82,$39,$63,$E7,$9C,$C7,$DA,$C3,$E6,$57,$E3,$DB,$1C,$E3,$E4,$9C
	.dc.b	$EF,$BC,$7C,$89,$DB,$67,$E9,$47,$97,$0F,$4C,$FC,$93,$C6,$7C,$2B
	.dc.b	$F1,$BF,$49,$AF,$BC,$DB,$CF,$0F,$A1,$5E,$09,$FD,$49,$F3,$5F,$F1
	.dc.b	$4E,$1F,$EC,$8F,$86,$FF,$66,$E9,$FE,$25,$EC,$46,$30,$FF,$2F,$76
	.dc.b	$5C,$7F,$CB,$D5,$A9,$FF,$47,$36,$3F,$C5,$7F,$21,$BC,$7B,$57,$FA
	.dc.b	$D7,$9D,$3C,$E9,$F5,$2E,$9C,$36,$FF,$D9,$CB,$E8,$F3,$32,$7F,$DA
	.dc.b	$5E,$8B,$CE,$C6,$1F,$E2,$E9,$FE,$42,$FF,$AF,$E3,$C1,$F4,$AF,$84
	.dc.b	$A7,$8F,$E9,$FF,$DA,$D7,$FE,$DD,$9D,$36,$7F,$CB,$77,$D2,$72,$FF
	.dc.b	$BA,$DA,$F8,$03,$87,$84,$78,$B2,$F0,$D7,$77,$E4,$6F,$E2,$AF,$F5
	.dc.b	$AB,$78,$BF,$D6,$7E,$57,$C3,$BF,$A3,$EB,$70,$3E,$B7,$A2,$E2,$7C
	.dc.b	$5E,$6B,$E9,$3A,$BF,$9D,$97,$FB,$0B,$EE,$3F,$D9,$75,$FE,$A4,$E2
	.dc.b	$FB,$51,$89,$F9,$83,$15,$F2,$CF,$FC,$DD,$5B,$7C,$F7,$CB,$D0,$FC
	.dc.b	$07,$C9,$7E,$0B,$2F,$EA,$B5,$FE,$86,$8C,$65,$D8,$F9,$3F,$0D,$C3
	.dc.b	$9E,$5F,$FB,$0E,$B7,$5F,$3D,$AF,$E0,$1A,$FF,$62,$7A,$DF,$7E,$5E
	.dc.b	$73,$DE,$7F,$15,$5E,$43,$FB,$4E,$AF,$69,$F7,$BF,$39,$7D,$ED,$8C
	.dc.b	$36,$B9,$61,$F1,$EB,$71,$8B,$E9,$7C,$23,$FD,$F1,$F0,$EF,$47,$F5
	.dc.b	$4D,$8E,$1E,$CF,$E0,$71,$FF,$7C,$CE,$5B,$5F,$82,$F9,$15,$F2,$8C
	.dc.b	$3F,$86,$BF,$B0,$FF,$81,$BD,$D7,$C2,$F0,$8E,$FC,$9F,$C0,$67,$31
	.dc.b	$C6,$31,$3B,$C7,$F0,$BA,$30,$EB,$0C,$38,$E1,$D0,$F8,$AE,$23,$8C
	.dc.b	$F0,$8C,$70,$FA,$A7,$E5,$1E,$DF,$BC,$FA,$4D,$BF,$99,$BF,$F8,$81
	.dc.b	$F9,$DA,$AB,$71,$87,$ED,$78,$FF,$1F,$E8,$2C,$F0,$6E,$1E,$CB,$46
	.dc.b	$31,$F2,$68,$9C,$DF,$E0,$2F,$F4,$DF,$31,$FC,$7E,$8F,$F2,$5F,$03
	.dc.b	$F4,$D5,$FC,$5A,$72,$F1,$6D,$7F,$0B,$39,$C5,$69,$98,$DB,$FC,$45
	.dc.b	$E3,$87,$9B,$9F,$F9,$B3,$FE,$F4,$8C,$D7,$E7,$A7,$2F,$FB,$83,$E8
	.dc.b	$DF,$07,$FC,$ED,$B1,$F9,$CF,$A8,$7D,$43,$E8,$F5,$7F,$CF,$7E,$A1
	.dc.b	$F0,$9F,$CA,$35,$FF,$29,$C6,$31,$F5,$36,$3D,$CD,$7F,$C8,$1E,$AB
	.dc.b	$E7,$7F,$A6,$36,$BF,$5B,$BB,$ED,$FE,$5D,$BC,$FF,$F4,$AE,$1E,$9E
	.dc.b	$2F,$EA,$1F,$50,$F9,$CC,$65,$F5,$E3,$0D,$8F,$46,$09,$CB,$F4,$DA
	.dc.b	$FF,$3F,$6B,$FC,$3F,$F0,$57,$CE,$7F,$69,$F1,$51,$8C,$3B,$1F,$C1
	.dc.b	$A3,$33,$FD,$97,$A9,$ED,$D3,$E9,$7D,$4D,$BF,$C7,$4E,$9F,$C2,$37
	.dc.b	$1F,$C4,$35,$FE,$A8,$D7,$FB,$2B,$BE,$0F,$D7,$6E,$FB,$8F,$9D,$5F
	.dc.b	$CC,$EB,$57,$FE,$8C,$AD,$70,$79,$FE,$02,$4F,$F2,$09,$75,$3E,$AB
	.dc.b	$38,$F3,$BF,$9B,$89,$A4,$C3,$D6,$FF,$1B,$EE,$E3,$A7,$EC,$3B,$FF
	.dc.b	$33,$87,$FA,$7B,$E9,$DF,$09,$FE,$0D,$6E,$3F,$B3,$FA,$AF,$CB,$3F
	.dc.b	$B5,$8C,$D7,$F4,$F6,$BF,$DE,$5A,$FF,$82,$9A,$F7,$AF,$C2,$C7,$2C
	.dc.b	$3F,$8B,$7C,$FF,$8B,$ED,$7C,$55,$FE,$29,$8C,$61,$E1,$F9,$DF,$31
	.dc.b	$FE,$23,$C9,$F5,$78,$9D,$1F,$BF,$C9,$5E,$02,$46,$5E,$D8,$7E,$8B
	.dc.b	$35,$70,$AF,$EA,$78,$FC,$6F,$53,$E3,$78,$32,$7F,$8C,$7D,$4B,$EA
	.dc.b	$F1,$8F,$C5,$AF,$79,$2B,$C8,$D8,$FF,$02,$E0,$FB,$67,$C9,$F2,$E1
	.dc.b	$F8,$B1,$8C,$3E,$97,$AB,$F5,$CF,$F2,$26,$7F,$2E,$3E,$84,$65,$F8
	.dc.b	$4F,$D0,$23,$FC,$87,$CF,$FF,$0F,$BF,$D0,$3E,$73,$FB,$B7,$B9,$FD
	.dc.b	$8D,$B1,$93,$FA,$3C,$65,$FE,$71,$F5,$5F,$CD,$BC,$3F,$F7,$37,$87
	.dc.b	$FE,$4A,$8C,$6F,$FE,$E6,$8F,$F2,$6F,$D5,$FE,$9B,$E3,$3F,$7F,$3F
	.dc.b	$EF,$0A,$DC,$62,$11,$8A,$E9,$D8,$C3,$BD,$7F,$BE,$30,$E9,$5A,$9C
	.dc.b	$CF,$2A,$E0,$F8,$CF,$03,$8C,$3C,$17,$CA,$73,$5C,$B0,$F1,$46,$0E
	.dc.b	$42,$78,$3D,$E7,$85,$F5,$C3,$C9,$3C,$2B,$77,$D2,$38,$1D,$1C,$46
	.dc.b	$AB,$94,$F8,$70,$F8,$D5,$AF,$0B,$FC,$01,$18,$AD,$D7,$43,$FC,$DD
	.dc.b	$5F,$51,$B1,$F1,$1B,$8F,$F6,$7F,$2F,$F8,$77,$25,$EF,$34,$95,$FD
	.dc.b	$BF,$06,$EF,$8D,$FE,$E1,$18,$6C,$66,$78,$35,$F2,$B9,$CF,$DA,$DD
	.dc.b	$BF,$18,$D9,$0F,$AB,$7D,$5F,$FA,$6A,$F8,$E1,$F6,$6F,$85,$FE,$31
	.dc.b	$B7,$C2,$35,$31,$FA,$53,$AB,$E9,$87,$8A,$B5,$87,$6B,$FC,$1E,$1C
	.dc.b	$DE,$33,$AA,$E1,$1A,$BE,$6F,$E8,$70,$F2,$E3,$F1,$E3,$89,$B3,$91
	.dc.b	$BB,$AF,$C1,$AE,$7E,$AF,$F8,$B7,$DD,$FF,$5D,$F4,$FF,$41,$FB,$8B
	.dc.b	$CB,$FE,$76,$F1,$51,$8C,$7F,$D2,$D0,$BC,$61,$CE,$7F,$C8,$B5,$8A
	.dc.b	$E3,$1F,$E4,$8A,$E5,$E4,$BF,$22,$FF,$04,$4F,$F8,$CE,$31,$E8,$7F
	.dc.b	$8C,$DB,$19,$9D,$BF,$EC,$38,$CB,$FE,$AA,$6E,$7F,$DC,$27,$6F,$F4
	.dc.b	$AF,$AD,$FE,$7F,$5F,$DA,$E7,$FC,$5C,$FF,$97,$3E,$CF,$D0,$79,$A3
	.dc.b	$31,$D5,$AF,$FB,$8A,$FD,$5E,$BF,$9E,$73,$3F,$DE,$DA,$FF,$8A,$5E
	.dc.b	$78,$79,$67,$B1,$F3,$79,$98,$FF,$B3,$AF,$CD,$E4,$FC,$BF,$AD,$91
	.dc.b	$82,$1F,$0F,$1F,$EB,$EF,$C7,$9F,$21,$1D,$E3,$F1,$63,$C1,$D2,$F9
	.dc.b	$2F,$AB,$F2,$DF,$AF,$AF,$F0,$27,$C9,$FD,$B3,$3C,$F3,$1F,$B1,$5F
	.dc.b	$B5,$7E,$23,$5F,$F0,$E3,$5F,$F2,$53,$D7,$9B,$33,$CE,$7F,$AF,$E2
	.dc.b	$F6,$3C,$67,$E2,$FE,$0E,$4F,$F8,$8F,$0D,$36,$7F,$11,$18,$8F,$C3
	.dc.b	$35,$F2,$C6,$8E,$AF,$13,$FD,$3F,$F5,$9F,$0B,$D7,$FC,$D7,$34,$9F
	.dc.b	$04,$65,$F8,$2F,$80,$8E,$0D,$90,$FC,$B7,$D0,$7F,$A6,$1F,$F5,$5F
	.dc.b	$C6,$FE,$9C,$8F,$B7,$D7,$FA,$67,$B0,$8F,$D0,$7F,$D3,$0F,$F8,$6B
	.dc.b	$D0,$DB,$C4,$FF,$54,$7A,$9F,$E1,$17,$FD,$8B,$F9,$4F,$FA,$C1,$BE
	.dc.b	$C7,$EF,$7E,$19,$39,$97,$2F,$FB,$1E,$73,$E5,$FF,$03,$87,$D2,$FF
	.dc.b	$BB,$6B,$FD,$59,$18,$F8,$AF,$F5,$63,$63,$33,$B7,$FD,$65,$F5,$7F
	.dc.b	$EE,$67,$FD,$64,$D7,$3C,$BE,$E8,$C4,$7E,$3B,$E3,$BF,$87,$78,$32
	.dc.b	$3F,$09,$1F,$9A,$3F,$06,$47,$F2,$ED,$7F,$84,$73,$A3,$84,$7F,$48
	.dc.b	$C7,$FD,$A7,$F5,$2F,$93,$FE,$3F,$EA,$BE,$A3,$F8,$F7,$C0,$4C,$3F
	.dc.b	$DC,$F1,$A3,$0F,$FB,$7E,$30,$FF,$BA,$BE,$67,$F9,$31,$1F,$F7,$7F
	.dc.b	$CC,$D7,$FD,$CD,$17,$2F,$F7,$2E,$0A,$FE,$5D,$F7,$32,$FF,$DE,$AF
	.dc.b	$DB,$E1,$F9,$D7,$FE,$7D,$8D,$4F,$FB,$C5,$D7,$16,$27,$53,$C8,$F8
	.dc.b	$5E,$17,$F6,$AE,$5F,$B5,$1F,$EF,$9F,$17,$F9,$F1,$1D,$8F,$3D,$A3
	.dc.b	$E5,$3E,$49,$B9,$E4,$70,$9D,$C7,$F4,$9F,$81,$93,$9F,$41,$1F,$4B
	.dc.b	$E0,$73,$3F,$96,$CF,$FA,$C4,$E3,$4F,$8F,$5F,$D6,$69,$F4,$FF,$D5
	.dc.b	$1E,$07,$F8,$1B,$F4,$27,$1E,$DB,$1E,$C3,$8F,$A9,$CE,$B9,$47,$F8
	.dc.b	$38,$FF,$07,$B8,$70,$D7,$31,$FB,$88,$F1,$60,$CB,$73,$D4,$ED,$E2
	.dc.b	$7F,$6E,$3F,$E0,$0D,$FC,$76,$BE,$AD,$91,$8B,$FF,$68,$F8,$AF,$A8
	.dc.b	$E5,$AC,$C3,$93,$FD,$D2,$75,$1F,$3E,$7F,$BD,$11,$F0,$7C,$27,$8C
	.dc.b	$C3,$E1,$E0,$E2,$F8,$F1,$FF,$4B,$BF,$27,$EE,$3F,$04,$FC,$BE,$77
	.dc.b	$DC,$FD,$56,$C6,$F9,$3E,$C5,$70,$8F,$9C,$7F,$8F,$1C,$38,$6B,$F4
	.dc.b	$DA,$EE,$39,$91,$FB,$9F,$52,$FA,$1F,$F0,$C7,$CF,$49,$FC,$83,$F8
	.dc.b	$9F,$A6,$7D,$25,$35,$3F,$06,$70,$FE,$6F,$E8,$FF,$AF,$3F,$A3,$6F
	.dc.b	$21,$BB,$7F,$BB,$A7,$EE,$7C,$77,$F8,$B2,$76,$69,$FC,$D7,$C3,$3E
	.dc.b	$4B,$F9,$AD,$7F,$79,$CB,$F2,$5E,$4B,$E9,$7F,$29,$ED,$FA,$2B,$FC
	.dc.b	$C7,$39,$7F,$CC,$B1,$95,$87,$5E,$AC,$E1,$F9,$D8,$C4,$BE,$97,$86
	.dc.b	$1C,$E3,$15,$E2,$C7,$F4,$95,$D6,$F8,$B7,$2F,$4B,$CA,$77,$5C,$6F
	.dc.b	$FC,$F3,$3B,$C6,$CF,$8E,$FE,$0E,$5C,$6F,$CE,$7B,$B0,$F8,$35,$BB
	.dc.b	$CC,$FB,$AF,$8C,$F7,$8F,$94,$71,$37,$8F,$C0,$C3,$C5,$1E,$E3,$45
	.dc.b	$19,$7C,$3C,$B6,$70,$9F,$34,$F6,$37,$97,$99,$FF,$4B,$DF,$7C,$3D
	.dc.b	$11,$D7,$0E,$91,$E7,$C3,$AB,$A7,$A9,$1C,$12,$7C,$F1,$97,$BE,$3F
	.dc.b	$4A,$7B,$E3,$E6,$23,$83,$94,$FA,$0F,$A6,$74,$73,$87,$DE,$7C,$C4
	.dc.b	$8E,$F9,$70,$7E,$93,$F5,$31,$FC,$55,$F8,$27,$67,$C7,$BE,$8F,$72
	.dc.b	$1C,$8F,$A4,$E9,$F3,$1F,$27,$2F,$C6,$BD,$32,$FA,$98,$FE,$06,$FC
	.dc.b	$13,$CD,$F9,$AF,$6C,$3E,$8B,$C6,$7E,$1C,$62,$3B,$D7,$07,$C8,$FC
	.dc.b	$F7,$33,$F1,$B2,$FC,$8C,$7E,$2A,$FC,$38,$FE,$06,$3B,$1E,$1C,$7D
	.dc.b	$2F,$4B,$FA,$8F,$17,$D4,$61,$F5,$38,$74,$7C,$CA,$FC,$DF,$B7,$FE
	.dc.b	$81,$8F,$9E,$D7,$E6,$36,$FF,$80,$B9,$8F,$F7,$6F,$C4,$FF,$BB,$A3
	.dc.b	$10,$8C,$57,$F3,$88,$C1,$FF,$00,$5C,$46,$A3,$F9,$EB,$FB,$6F,$82
	.dc.b	$7C,$EF,$F6,$59,$FD,$41,$E0,$3F,$55,$1A,$F0,$7F,$7B,$19,$BF,$E3
	.dc.b	$93,$98,$E2,$78,$A3,$35,$D5,$FE,$45,$1D,$75,$CD,$FE,$5E,$31,$9F
	.dc.b	$FB,$06,$FC,$95,$C2,$7A,$4E,$6B,$C7,$18,$03,$0F,$26,$1C,$71,$FD
	.dc.b	$24,$79,$F0,$C5,$7F,$0D,$8C,$61,$F2,$23,$18,$F8,$A3,$15,$E8,$9F
	.dc.b	$E2,$87,$A0,$F0,$D7,$27,$E7,$9C,$B1,$F7,$3E,$73,$CD,$7F,$16,$FC
	.dc.b	$18,$F5,$3C,$B8,$77,$2C,$7C,$1A,$F0,$9F,$80,$8D,$47,$98,$D3,$EE
	.dc.b	$9E,$91,$D8,$E4,$F4,$9F,$4D,$7B,$DF,$4D,$74,$6D,$74,$3B,$3C,$2F
	.dc.b	$DB,$8F,$E3,$6F,$D9,$3F,$0B,$0F,$0B,$C8,$E0,$7A,$A7,$EA,$CF,$86
	.dc.b	$3F,$C8,$11,$8A,$EB,$1F,$2C,$E0,$5C,$3D,$4F,$CC,$8D,$DF,$82,$39
	.dc.b	$9F,$9C,$BF,$2B,$8A,$F7,$37,$0F,$73,$E9,$7E,$39,$F2,$5F,$4D,$7C
	.dc.b	$D8,$CC,$6C,$CB,$F1,$9F,$21,$E8,$9E,$EF,$E2,$4C,$BE,$AC,$7F,$15
	.dc.b	$1E,$8A,$F5,$3F,$B1,$AE,$FC,$7F,$E5,$4F,$9D,$FD,$D7,$0F,$EE,$E3
	.dc.b	$FC,$45,$78,$CF,$F4,$B1,$B7,$FC,$55,$19,$C3,$FC,$54,$D9,$E3,$3F
	.dc.b	$E6,$3B,$C7,$1A,$B8,$5F,$FA,$9F,$A7,$3F,$37,$8F,$37,$FD,$29,$8F
	.dc.b	$F8,$DE,$32,$D3,$FD,$01,$1A,$AF,$F3,$76,$0A,$EB,$F8,$03,$8A,$9E
	.dc.b	$0F,$A6,$7D,$36,$32,$FF,$45,$56,$DF,$E4,$98,$A3,$0F,$F2,$76,$BF
	.dc.b	$E6,$06,$BF,$E8,$F7,$7B,$FF,$25,$87,$EC,$EF,$8A,$F0,$47,$FA,$78
	.dc.b	$CB,$F2,$7B,$1F,$29,$F2,$3E,$43,$0D,$CB,$FC,$B5,$16,$FF,$CF,$93
	.dc.b	$71,$FB,$5F,$5E,$F9,$1F,$A8,$DC,$3F,$D8,$7D,$0F,$A7,$FF,$5C,$C7
	.dc.b	$FA,$FE,$75,$8F,$FA,$65,$B8,$74,$C3,$F0,$18,$71,$F4,$BA,$C7,$DC
	.dc.b	$E9,$7D,$25,$3E,$FF,$CA,$FD,$7C,$3E,$24,$EB,$3F,$F5,$D3,$73,$FE
	.dc.b	$A8,$D7,$FD,$1A,$D7,$FD,$78,$D7,$F2,$B7,$F8,$CF,$96,$F5,$C7,$CD
	.dc.b	$9E,$97,$F4,$E7,$2F,$E7,$FA,$1E,$EF,$E5,$8F,$CD,$7A,$EF,$A7,$C6
	.dc.b	$7F,$E8,$99,$B9,$7F,$AF,$6B,$53,$93,$FD,$C8,$7E,$83,$E3,$7F,$49
	.dc.b	$E8,$FF,$B2,$E3,$FD,$25,$19,$69,$C0,$B5,$F9,$CF,$1F,$F5,$27,$EB
	.dc.b	$FD,$B7,$C2,$73,$9D,$57,$FA,$76,$B6,$FF,$C0,$0A,$DB,$89,$D9,$35
	.dc.b	$FE,$F5,$3A,$9E,$47,$FA,$9F,$D3,$FD,$C9,$EA,$E1,$B1,$92,$5F,$FB
	.dc.b	$3E,$32,$FF,$B4,$7E,$1F,$F8,$0F,$99,$F0,$9B,$5F,$DC,$31,$FE,$21
	.dc.b	$AA,$73,$7C,$6B,$FD,$71,$7F,$90,$F3,$1F,$2B,$4F,$E2,$1E,$4B,$E4
	.dc.b	$31,$7F,$9E,$AD,$BF,$E1,$FE,$BF,$F5,$5D,$FE,$D7,$1F,$EE,$35,$A7
	.dc.b	$FD,$B7,$3A,$AE,$13,$9D,$3F,$11,$F2,$1D,$AF,$F8,$8B,$FC,$73,$E8
	.dc.b	$DE,$E7,$59,$FF,$06,$C6,$30,$FE,$1A,$D8,$FF,$19,$35,$FF,$09,$9F
	.dc.b	$E1,$3C,$1D,$5C,$F2,$FF,$61,$B6,$33,$7E,$68,$F0,$75,$FF,$B6,$FE
	.dc.b	$3F,$F9,$89,$FC,$CB,$E1,$ED,$7D,$13,$FC,$43,$DC,$FA,$17,$8B,$FB
	.dc.b	$EF,$BD,$FF,$12,$D6,$F5,$5E,$EB,$51,$27,$E7,$57,$08,$F9,$6D,$7D
	.dc.b	$BA,$A3,$11,$FD,$3B,$C1,$FF,$80,$15,$FE,$F7,$F6,$BA,$4F,$F3,$E7
	.dc.b	$C9,$F2,$12,$7F,$8D,$FD,$17,$C2,$A7,$2F,$F9,$57,$2E,$D7,$B9,$F1
	.dc.b	$BF,$2F,$23,$49,$87,$BF,$81,$9F,$E0,$63,$FC,$6F,$5C,$1D,$46,$6F
	.dc.b	$FC,$95,$2F,$B7,$F9,$47,$F0,$2F,$90,$B7,$FC,$CA,$74,$FC,$89,$F2
	.dc.b	$11,$C1,$DB,$EE,$B8,$DF,$F6,$07,$EB,$7C,$74,$9F,$AD,$F3,$4F,$7D
	.dc.b	$F1,$F9,$C3,$E2,$D7,$AA,$FE,$11,$1F,$B2,$D8,$FE,$DB,$F2,$BC,$78
	.dc.b	$7F,$94,$3F,$85,$EC,$97,$E0,$C7,$F9,$6B,$F1,$A3,$37,$F8,$27,$E8
	.dc.b	$12,$3F,$0E,$FF,$81,$BA,$B7,$87,$F3,$57,$F2,$5E,$64,$AF,$6F,$D0
	.dc.b	$BE,$23,$18,$FB,$0F,$A3,$5F,$74,$95,$F9,$46,$C7,$F8,$23,$D8,$E3
	.dc.b	$8F,$F4,$37,$FA,$EF,$97,$FC,$4E,$FF,$B3,$B6,$BB,$70,$32,$FB,$31
	.dc.b	$E1,$E8,$56,$F1,$FC,$93,$C6,$7F,$42,$D7,$FD,$02,$D7,$FD,$24,$DE
	.dc.b	$7F,$C8,$FC,$87,$F8,$AB,$EA,$BF,$C5,$6D,$CF,$F2,$2D,$BF,$F1,$1B
	.dc.b	$70,$FF,$09,$7C,$F3,$A7,$F7,$35,$FE,$C3,$EE,$FE,$1F,$39,$EF,$FF
	.dc.b	$64,$C7,$F1,$2F,$A0,$FD,$67,$C3,$7F,$04,$BF,$F5,$3D,$ED,$E4,$63
	.dc.b	$0E,$B3,$B8,$4E,$E3,$15,$D2,$FC,$37,$C1,$B8,$78,$B0,$E8,$D5,$5C
	.dc.b	$5C,$9E,$23,$BD,$73,$73,$7E,$0E,$5F,$F6,$C3,$E5,$C7,$C9,$8F,$82
	.dc.b	$B9,$46,$1F,$F6,$2E,$1D,$DE,$B8,$F8,$E7,$51,$83,$C5,$3B,$8C,$93
	.dc.b	$1F,$24,$FA,$B1,$F1,$E1,$E1,$C7,$C9,$5F,$5A,$BF,$A1,$9D,$4B,$7F
	.dc.b	$45,$B1,$B7,$FA,$4F,$8A,$F1,$DC,$9F,$E5,$C8,$C3,$63,$33,$D8,$4E
	.dc.b	$9C,$5F,$F9,$2F,$DF,$FF,$32,$B6,$33,$EB,$7F,$9B,$DB,$3A,$8C,$65
	.dc.b	$F3,$DB,$9F,$E4,$1B,$5F,$EF,$2F,$A0,$FE,$FA,$FC,$D8,$CB,$F8,$A6
	.dc.b	$E1,$E1,$BF,$C7,$57,$0A,$FB,$35,$A2,$61,$F8,$99,$D5,$78,$E3,$B1
	.dc.b	$2B,$F1,$53,$AC,$3C,$D3,$9A,$E1,$5A,$F2,$FF,$D9,$F5,$A8,$F0,$D6
	.dc.b	$E3,$FB,$DD,$F3,$9F,$D4,$E1,$E4,$76,$73,$3C,$DB,$74,$7E,$23,$54
	.dc.b	$FF,$81,$1F,$67,$4E,$63,$11,$EA,$9C,$19,$9E,$5F,$99,$3E,$6F,$FC
	.dc.b	$1F,$C9,$FE,$97,$8C,$5F,$F7,$F0,$E1,$FE,$43,$AF,$CF,$9E,$15,$CB
	.dc.b	$9B,$FC,$33,$7B,$9C,$DF,$F5,$4B,$E1,$5A,$F1,$3F,$D5,$F3,$D9,$B7
	.dc.b	$FD,$A6,$3F,$D6,$6D,$7F,$2C,$DB,$F4,$4F,$86,$47,$33,$C1,$96,$5C
	.dc.b	$1F,$EB,$1A,$E5,$7E,$38,$D9,$F0,$61,$F4,$5D,$79,$3E,$47,$CB,$DD
	.dc.b	$0F,$5A,$DD,$F1,$BC,$11,$FE,$EB,$D1,$FC,$CD,$D1,$0F,$24,$77,$8D
	.dc.b	$47,$9F,$BF,$C5,$FA,$6F,$EB,$E3,$FA,$9B,$76,$FB,$53,$E3,$24,$FA
	.dc.b	$DE,$9F,$3D,$88,$E8,$64,$F2,$BF,$1F,$91,$F2,$9F,$07,$F1,$A7,$C0
	.dc.b	$53,$51,$2F,$F1,$15,$C1,$B1,$F4,$7D,$97,$DA,$6E,$7F,$DC,$55,$B6
	.dc.b	$D6,$E7,$C5,$F2,$5F,$47,$BD,$F1,$D4,$CD,$F8,$26,$BD,$1F,$B1,$F7
	.dc.b	$4F,$A4,$CF,$07,$FB,$D7,$35,$63,$A3,$F1,$3D,$0C,$38,$63,$FE,$2C
	.dc.b	$FC,$6B,$F0,$EF,$F2,$C6,$C9,$CC,$3E,$EF,$FC,$06,$FE,$AD,$AF,$FC
	.dc.b	$01,$6E,$4C,$3F,$D0,$31,$87,$C7,$A7,$EF,$27,$55,$FC,$F2,$31,$3C
	.dc.b	$3C,$AF,$CF,$5F,$1A,$D7,$67,$E9,$69,$F6,$FF,$EF,$69,$FF,$3E,$46
	.dc.b	$3E,$0B,$53,$99,$FF,$0F,$E6,$F4,$FA,$FC,$1C,$9F,$63,$FA,$5F,$C5
	.dc.b	$BF,$AB,$ED,$E6,$DD,$18,$CF,$E6,$B6,$FF,$AA,$B7,$EC,$9F,$08,$EC
	.dc.b	$FD,$C7,$D3,$7F,$B8,$2F,$FA,$B0,$70,$FF,$59,$6B,$FD,$A1,$AF,$FA
	.dc.b	$B1,$DE,$5F,$2A,$39,$7C,$47,$63,$F8,$37,$DD,$FE,$AF,$3E,$11,$99
	.dc.b	$FE,$49,$7C,$A7,$FD,$6F,$5B,$97,$A9,$FE,$88,$FA,$6F,$E2,$8B,$E5
	.dc.b	$3E,$03,$67,$3F,$00,$AA,$70,$AD,$9C,$23,$53,$FE,$95,$8F,$F0,$36
	.dc.b	$7F,$84,$BE,$DF,$37,$C4,$E5,$E3,$CB,$E1,$19,$79,$47,$F2,$5F,$D3
	.dc.b	$5F,$5B,$7E,$92,$57,$F4,$39,$CB,$67,$2F,$F9,$C4,$42,$7A,$4C,$9E
	.dc.b	$51,$C3,$1F,$A5,$5D,$27,$97,$B1,$FE,$49,$9F,$AC,$DD,$FF,$DC,$AD
	.dc.b	$9F,$E7,$FF,$A6,$FF,$BC,$1F,$EB,$11,$97,$FC,$C1,$E1,$FF,$41,$8C
	.dc.b	$4E,$E7,$FD,$11,$AF,$D5,$FB,$AF,$A1,$F8,$04,$9F,$9D,$1F,$4C,$C3
	.dc.b	$C3,$81,$39,$8F,$0F,$AB,$FD,$73,$13,$49,$1F,$9A,$8C,$BC,$08,$7C
	.dc.b	$BF,$1F,$F7,$AF,$F5,$C8,$DB,$F1,$A7,$D8,$7F,$5D,$9D,$E3,$F8,$FF
	.dc.b	$88,$FC,$57,$D0,$4B,$FC,$44,$7E,$34,$C3,$EB,$8C,$3F,$E0,$1F,$D0
	.dc.b	$DF,$27,$FD,$A5,$B1,$98,$FE,$D9,$F8,$13,$E5,$35,$12,$74,$7F,$82
	.dc.b	$E7,$4F,$F8,$37,$F4,$4F,$F2,$B3,$FB,$C6,$BF,$E5,$36,$BF,$E6,$26
	.dc.b	$BF,$DF,$83,$C3,$93,$F7,$8D,$E4,$D3,$67,$FC,$0B,$E0,$7E,$3E,$B7
	.dc.b	$E7,$FF,$81,$6B,$55,$D2,$3F,$C1,$15,$BB,$F0,$79,$1F,$D4,$AB,$FC
	.dc.b	$4C,$D7,$F4,$4D,$C3,$F8,$77,$E2,$5F,$6A,$7F,$98,$F9,$DF,$E1,$3A
	.dc.b	$D5,$F4,$8F,$F0,$B5,$7E,$03,$E1,$DF,$7E,$F9,$2F,$F0,$F4,$61,$B1
	.dc.b	$9C,$7F,$82,$7C,$5E,$26,$47,$F8,$93,$3F,$E9,$37,$C6,$36,$7F,$8F
	.dc.b	$FE,$9D,$F3,$9F,$E2,$76,$C6,$63,$F8,$A7,$C7,$E5,$91,$93,$F8,$A4
	.dc.b	$69,$A4,$3F,$51,$F0,$59,$7F,$55,$17,$0F,$E3,$5D,$DF,$E4,$9D,$FF
	.dc.b	$D3,$7C,$29,$FF,$1E,$FD,$3B,$A7,$F0,$5A,$7F,$6C,$8C,$65,$FE,$42
	.dc.b	$8C,$4F,$2A,$E2,$D9,$57,$05,$5C,$1E,$5E,$BF,$E3,$E7,$5F,$17,$FD
	.dc.b	$5B,$0E,$75,$FE,$50,$8C,$3D,$30,$F2,$DF,$16,$C7,$03,$FD,$5F,$8F
	.dc.b	$B5,$E0,$73,$71,$87,$97,$1F,$24,$F9,$AF,$A3,$BC,$7A,$63,$E2,$BE
	.dc.b	$15,$AA,$E0,$6B,$0E,$B8,$FC,$3A,$DC,$F9,$71,$EF,$87,$3A,$D1,$EA
	.dc.b	$9F,$39,$D0,$B1,$B7,$C8,$6A,$7D,$27,$4B,$F0,$C6,$B9,$3C,$2F,$F8
	.dc.b	$FA,$FB,$47,$D0,$9F,$8C,$DC,$BC,$D9,$7B,$70,$E9,$1E,$D3,$38,$43
	.dc.b	$CB,$7E,$07,$E8,$C7,$FA,$22,$FC,$24,$C3,$F0,$CF,$79,$F5,$61,$DC
	.dc.b	$F9,$98,$FB,$E3,$99,$A3,$C1,$97,$12,$5F,$C9,$7D,$E6,$23,$E3,$B8
	.dc.b	$8D,$1F,$42,$32,$E1,$FC,$A9,$EF,$3D,$B8,$78,$23,$51,$E4,$9E,$CF
	.dc.b	$D7,$7B,$3D,$08,$66,$38,$D7,$B4,$FC,$EF,$C4,$FF,$1F,$6E,$3F,$AB
	.dc.b	$8C,$65,$F9,$B7,$F9,$25,$FF,$AD,$3E,$1F,$FD,$6B,$18,$84,$62,$BF
	.dc.b	$9E,$46,$C4,$EE,$75,$1A,$7F,$55,$C4,$F2,$FF,$CC,$B9,$FE,$EE,$B9
	.dc.b	$3F,$BD,$9D,$6F,$FB,$19,$CE,$9F,$D2,$E7,$31,$F9,$C3,$27,$FB,$31
	.dc.b	$CD,$75,$8F,$49,$FA,$28,$CC,$FA,$A3,$33,$E1,$BF,$2C,$E7,$0F,$4C
	.dc.b	$62,$BD,$87,$37,$0F,$F5,$F8,$DE,$1E,$98,$EB,$5E,$9A,$DC,$7A,$06
	.dc.b	$1E,$47,$AE,$1E,$3B,$E5,$87,$49,$F8,$07,$43,$E0,$5F,$47,$07,$47
	.dc.b	$11,$C6,$BE,$23,$FA,$BC,$3C,$97,$6F,$DB,$97,$85,$FD,$F1,$E4,$9D
	.dc.b	$BE,$2A,$E9,$1F,$1E,$3E,$25,$7A,$1A,$6C,$E6,$79,$5F,$03,$F2,$4C
	.dc.b	$E3,$F7,$AF,$CF,$1C,$B0,$A7,$94,$B7,$E4,$6C,$7C,$4B,$F1,$46,$E7
	.dc.b	$CC,$D9,$F6,$1B,$3C,$CE,$5F,$23,$5F,$B1,$7F,$5D,$CC,$F8,$4E,$A5
	.dc.b	$7A,$9E,$F3,$AB,$97,$89,$E1,$9F,$4B,$B7,$FB,$3F,$81,$FE,$95,$F8
	.dc.b	$7F,$F6,$6E,$FF,$DC,$9B,$8A,$FF,$D3,$BC,$5F,$E3,$F8,$FF,$49,$4E
	.dc.b	$A7,$2D,$BE,$93,$FE,$97,$EF,$C5,$AB,$85,$FE,$5B,$C5,$BA,$7F,$9B
	.dc.b	$30,$EA,$7F,$86,$4F,$DB,$46,$88,$EA,$7F,$D2,$53,$C5,$FF,$04,$47
	.dc.b	$EE,$7E,$A9,$E3,$F9,$E9,$18,$9F,$F5,$67,$D0,$E6,$31,$7C,$8F,$D5
	.dc.b	$B8,$D7,$FC,$56,$DA,$F8,$DA,$FF,$AF,$EF,$96,$3E,$EE,$27,$65,$3E
	.dc.b	$47,$9C,$FD,$AF,$38,$FB,$F4,$FA,$17,$87,$FD,$5C,$DD,$3F,$DB,$4D
	.dc.b	$9F,$AD,$7C,$F0,$ED,$1C,$ED,$FA,$1F,$FB,$3E,$30,$D8,$F7,$53,$F3
	.dc.b	$17,$BD,$F5,$AB,$59,$7F,$BA,$5B,$87,$F9,$CD,$B8,$30,$E5,$7C,$7D
	.dc.b	$49,$7F,$EF,$AB,$F0,$5F,$F4,$06,$BC,$E7,$FD,$EC,$FF,$9F,$FE,$A7
	.dc.b	$FD,$AB,$5E,$B6,$C7,$A3,$27,$9B,$F8,$C9,$CE,$7F,$EF,$F6,$DF,$3F
	.dc.b	$53,$E8,$DF,$08,$E9,$1F,$96,$F8,$6F,$CB,$FD,$B5,$9F,$F7,$83,$72
	.dc.b	$FF,$61,$37,$1F,$F2,$CB,$F9,$CF,$C3,$7F,$CF,$15,$F4,$5B,$5F,$8B
	.dc.b	$FC,$35,$F8,$73,$EA,$F8,$79,$73,$0E,$0F,$EB,$6B,$73,$9B,$FF,$78
	.dc.b	$61,$D1,$FE,$D2,$D7,$FD,$4C,$D7,$F7,$4D,$E9,$77,$3E,$13,$F6,$BF
	.dc.b	$5D,$FB,$66,$E7,$F9,$3A,$DC,$63,$0F,$F7,$13,$76,$57,$FA,$D7,$E9
	.dc.b	$5E,$BF,$EF,$9B,$1F,$9D,$E9,$FC,$77,$C7,$66,$A4,$FE,$5E,$B7,$8F
	.dc.b	$DD,$AC,$BF,$B3,$FC,$65,$F0,$FE,$0B,$FD,$44,$63,$B3,$F8,$A3,$A8
	.dc.b	$B3,$F9,$C7,$6F,$1A,$E8,$E5,$B1,$C2,$B8,$B4,$E2,$58,$FE,$0D,$AF
	.dc.b	$CC,$FB,$0F,$E3,$FF,$1F,$FC,$81,$B5,$F2,$F8,$51,$8B,$E3,$18,$9D
	.dc.b	$57,$F2,$8F,$9F,$FE,$53,$F2,$B6,$79,$36,$F8,$B7,$87,$FD,$56,$FD
	.dc.b	$1A,$95,$FC,$37,$C7,$7E,$00,$EF,$CE,$3F,$C7,$C7,$13,$0C,$E5,$FE
	.dc.b	$CA,$BE,$18,$FD,$79,$E3,$7F,$E3,$AB,$E1,$3F,$E2,$38,$E7,$3F,$BC
	.dc.b	$7F,$A6,$4E,$9F,$F1,$47,$CA,$FF,$39,$C3,$F1,$71,$FD,$46,$31,$7F
	.dc.b	$D4,$C7,$C2,$7F,$31,$9D,$79,$7F,$D3,$3E,$FF,$F8,$95,$7F,$21,$6C
	.dc.b	$FB,$6F,$9B,$FD,$76,$33,$87,$83,$E3,$5F,$A5,$3C,$3F,$EB,$D1,$87
	.dc.b	$FB,$07,$3B,$DE,$E7,$39,$CF,$F9,$86,$6F,$73,$C9,$E6,$7F,$60,$8C
	.dc.b	$4F,$96,$32,$D3,$8C,$E3,$5F,$EA,$58,$F7,$8F,$0E,$3C,$A7,$EC,$6B
	.dc.b	$FE,$5A,$D7,$E4,$DE,$F0,$F8,$58,$7C,$53,$FB,$8F,$E0,$2F,$9B,$C4
	.dc.b	$E8,$98,$FE,$0E,$75,$18,$C3,$EA,$CB,$1F,$14,$78,$3D,$E7,$7F,$4A
	.dc.b	$F4,$C6,$2F,$EA,$7A,$9F,$5C,$F1,$FC,$BD,$8C,$92,$BC,$51,$97,$DF
	.dc.b	$3F,$4B,$F1,$77,$B5,$9C,$BE,$0C,$63,$0F,$EC,$5F,$90,$7F,$3F,$C3
	.dc.b	$F3,$CD,$AF,$8F,$0F,$B9,$FF,$0A,$CE,$6B,$FC,$45,$28,$C3,$FE,$25
	.dc.b	$6F,$0F,$F8,$9E,$B8,$FA,$52,$FF,$BE,$4B,$F4,$DF,$85,$87,$8A,$BF
	.dc.b	$17,$18,$C7,$E6,$E2,$FB,$0F,$19,$F9,$0C,$3F,$43,$8F,$B2,$32,$D3
	.dc.b	$BC,$6B,$1F,$7C,$E9,$FC,$5F,$C0,$FC,$57,$35,$E6,$8C,$61,$E2,$F9
	.dc.b	$DF,$86,$7E,$81,$7B,$98,$AF,$9B,$36,$FF,$2F,$5A,$9C,$92,$3F,$D4
	.dc.b	$D9,$2B,$F8,$24,$7C,$D6,$C6,$67,$F8,$52,$E1,$FA,$A7,$8F,$CE,$B3
	.dc.b	$F7,$A2,$DF,$8A,$6E,$1F,$98,$FA,$57,$D2,$7F,$0E,$F8,$3F,$E1,$77
	.dc.b	$FE,$5F,$9C,$E3,$FE,$D9,$F0,$FF,$03,$1F,$DE,$AB,$74,$8D,$BF,$ED
	.dc.b	$C9,$DD,$F5,$6D,$F3,$9E,$DE,$B6,$5F,$05,$55,$F4,$AF,$92,$FF,$16
	.dc.b	$47,$53,$FC,$E9,$19,$F5,$3A,$47,$8F,$E7,$B9,$4E,$63,$81,$D2,$B9
	.dc.b	$E9,$EA,$9D,$3E,$CD,$17,$CA,$B7,$A7,$E8,$AF,$6D,$8C,$E1,$F2,$6B
	.dc.b	$86,$3E,$4B,$DC,$FA,$2B,$6F,$9F,$4D,$46,$27,$73,$E7,$AF,$6E,$3E
	.dc.b	$4B,$DC,$49,$F6,$D7,$0A,$D4,$62,$FA,$12,$B9,$E1,$E6,$9D,$47,$F5
	.dc.b	$7B,$F9,$11,$FC,$00,$F1,$1F,$EF,$D7,$2F,$F0,$4F,$88,$7C,$77,$F8
	.dc.b	$52,$38,$08,$CF,$3F,$F6,$36,$C6,$9A,$F7,$A7,$D2,$BE,$2B,$C8,$D9
	.dc.b	$F1,$35,$E0,$D7,$EB,$3B,$DF,$EF,$37,$7C,$D7,$E2,$3E,$97,$F4,$77
	.dc.b	$F5,$EB,$4F,$E3,$9A,$FE,$51,$B1,$C2,$3E,$8C,$6B,$0F,$4D,$73,$3B
	.dc.b	$70,$FF,$2A,$C7,$C7,$19,$BE,$2E,$9C,$46,$9C,$9F,$9A,$C3,$C1,$19
	.dc.b	$26,$3E,$C9,$D5,$F7,$8C,$4E,$A7,$D9,$7D,$DE,$9B,$6D,$FF,$70,$61
	.dc.b	$FA,$27,$E3,$46,$54,$3E,$57,$0E,$E3,$99,$C0,$F9,$D1,$C7,$E9,$7F
	.dc.b	$C0,$91,$FE,$25,$6B,$F8,$86,$BF,$91,$6B,$FD,$49,$DF,$2F,$F7,$26
	.dc.b	$F2,$67,$C8,$FE,$AF,$3F,$E0,$2C,$FE,$B3,$BF,$0F,$35,$FE,$FD,$E2
	.dc.b	$7C,$DF,$2C,$BF,$CD,$AD,$59,$7F,$BE,$5F,$F6,$63,$5F,$45,$FF,$8F
	.dc.b	$DD,$B6,$35,$1E,$B8,$F0,$A3,$67,$47,$34,$C7,$E6,$63,$EF,$F5,$B3
	.dc.b	$9F,$F8,$67,$0F,$ED,$AD,$7F,$CB,$58,$A7,$BC,$7A,$9F,$21,$1F,$A4
	.dc.b	$D7,$8B,$C2,$39,$47,$F7,$AE,$6F,$E6,$8F,$C2,$F3,$7C,$FF,$4B,$FB
	.dc.b	$3C,$7F,$C6,$2D,$DB,$EC,$4F,$33,$E2,$12,$BF,$04,$D7,$D2,$ED,$F0
	.dc.b	$7C,$FF,$AB,$93,$FB,$44,$7F,$83,$3E,$97,$FC,$5E,$DF,$2F,$5E,$14
	.dc.b	$8F,$F3,$95,$70,$7F,$8A,$B5,$FE,$3C,$D7,$FC,$3E,$EF,$3F,$F1,$FB
	.dc.b	$72,$CD,$7C,$CF,$49,$ED,$79,$7C,$3F,$B3,$DF,$FE,$34,$CB,$FC,$42
	.dc.b	$D7,$FD,$1D,$8F,$F7,$B7,$F0,$59,$76,$AF,$C5,$4E,$67,$CF,$18,$23
	.dc.b	$C3,$2F,$3C,$78,$63,$FC,$5F,$F4,$AF,$2B,$F1,$7F,$49,$F9,$7C,$7F
	.dc.b	$CC,$AD,$C3,$F4,$4D,$BF,$CA,$7D,$2B,$E5,$7F,$12,$D8,$FF,$4E,$35
	.dc.b	$FE,$7E,$D7,$FA,$43,$5F,$D7,$7A,$1F,$9A,$BF,$F2,$5C,$F1,$C3,$FC
	.dc.b	$D6,$D8,$DF,$81,$FA,$65,$F3,$2F,$83,$C3,$71,$F8,$47,$FB,$C9,$C3
	.dc.b	$FE,$88,$CF,$FD,$FC,$E8,$60,$FA,$5F,$DC,$36,$3D,$B1,$83,$FD,$75
	.dc.b	$DE,$FA,$5F,$EC,$9A,$7F,$A3,$9B,$9F,$F0,$B6,$DB,$DB,$7A,$2E,$BF
	.dc.b	$D9,$FD,$23,$E9,$7F,$6B,$E3,$7F,$B4,$F0,$F3,$35,$FF,$6A,$4E,$DF
	.dc.b	$53,$5F,$ED,$ED,$7F,$BC,$35,$FF,$3C,$B5,$FE,$20,$F1,$CF,$FD,$36
	.dc.b	$DC,$F7,$7E,$4F,$27,$F8,$BF,$C3,$BE,$95,$F2,$78,$F8,$8F,$7C,$9B
	.dc.b	$3A,$CF,$EA,$CF,$1A,$E1,$10,$CC,$ED,$67,$FE,$E2,$AE,$9E,$37,$7F
	.dc.b	$81,$95,$D3,$1F,$F2,$2B,$C3,$4F,$E5,$97,$C2,$35,$5C,$1D,$08,$D3
	.dc.b	$FF,$00,$7E,$91,$F2,$7F,$CA,$17,$D2,$D3,$11,$27,$F1,$93,$A6,$CE
	.dc.b	$5E,$93,$C5,$C9,$C1,$FE,$6B,$EB,$7F,$9B,$34,$FF,$66,$FD,$2B,$CB
	.dc.b	$FE,$72,$DB,$FC,$D5,$6F,$4F,$F5,$AB,$63,$FA,$63,$5F,$F3,$57,$8B
	.dc.b	$CB,$E2,$7E,$C7,$CD,$7D,$9F,$89,$95,$F2,$E3,$2F,$CD,$8C,$BF,$23
	.dc.b	$7F,$BD,$3E,$8F,$57,$FA,$6F,$8E,$FA,$6A,$66,$FC,$F3,$FD,$97,$3F
	.dc.b	$F0,$44,$7F,$91,$8F,$F3,$F4,$64,$8F,$02,$1F,$60,$FE,$39,$E3,$7E
	.dc.b	$47,$E4,$7F,$25,$F3,$F2,$7F,$09,$19,$7F,$0D,$19,$7D,$5C,$E9,$FE
	.dc.b	$CB,$F4,$CF,$A6,$FE,$CF,$F4,$CF,$4A,$9A,$89,$87,$F0,$BA,$E0,$D8
	.dc.b	$E3,$F4,$3F,$DC,$3E,$9B,$FC,$CB,$3F,$C4,$7C,$67,$CD,$7F,$74,$6C
	.dc.b	$7F,$15,$9D,$BE,$F6,$BF,$E7,$36,$BF,$EF,$46,$BF,$E0,$B6,$A7,$9F
	.dc.b	$5B,$A7,$87,$A9,$E5,$3C,$5D,$BF,$E0,$6F,$3F,$FA,$B5,$6F,$0F,$F0
	.dc.b	$B7,$C8,$FD,$D7,$F9,$53,$5F,$A9,$3C,$30,$FF,$0E,$47,$F3,$8F,$4B
	.dc.b	$A1,$1E,$3C,$BF,$E2,$2F,$96,$7D,$33,$E2,$3F,$86,$78,$72,$FF,$9E
	.dc.b	$D6,$A7,$85,$6E,$72,$D7,$F7,$B3,$B7,$F8,$37,$D0,$FF,$88,$BE,$95
	.dc.b	$F4,$DF,$E3,$76,$CF,$E4,$DA,$FF,$A6,$9A,$FF,$A8,$DC,$EF,$F1,$61
	.dc.b	$18,$D7,$E0,$B6,$3F,$87,$75,$BE,$4F,$F9,$41,$23,$F8,$D3,$FC,$AF
	.dc.b	$E1,$31,$1F,$96,$8B,$87,$F2,$3A,$FF,$2A,$7C,$5F,$F4,$C8,$FF,$25
	.dc.b	$70,$BB,$FC,$3E,$E7,$D2,$8C,$63,$FE,$4F,$CF,$FB,$C4,$62,$3F,$D6
	.dc.b	$DD,$3F,$CC,$14,$67,$1E,$D0,$FA,$6E,$17,$FE,$01,$6E,$1F,$80,$78
	.dc.b	$61,$CD,$B9,$79,$B2,$F2,$E1,$4F,$25,$F9,$E3,$FD,$99,$7E,$3C,$B8
	.dc.b	$D7,$86,$78,$B4,$E2,$7F,$9A,$F0,$D9,$23,$6F,$9A,$35,$1D,$B0,$B1
	.dc.b	$FE,$73,$CB,$DF,$5B,$8E,$C7,$DE,$C7,$DD,$7C,$63,$35,$F1,$B0,$E0
	.dc.b	$49,$D3,$E3,$9F,$39,$D3,$2F,$95,$3C,$F0,$D4,$F7,$3C,$A5,$3B,$E5
	.dc.b	$F3,$67,$A9,$F1,$2F,$C1,$3E,$B9,$F9,$53,$D4,$F8,$64,$8D,$65,$DB
	.dc.b	$1F,$83,$39,$9F,$06,$1E,$49,$F7,$4F,$72,$4F,$23,$F1,$2F,$77,$DF
	.dc.b	$3F,$2C,$E5,$7E,$98,$CC,$F7,$AF,$0C,$73,$AE,$13,$9A,$E1,$3F,$98
	.dc.b	$BF,$C5,$3F,$17,$0F,$82,$E4,$F1,$9C,$EB,$D9,$8F,$D0,$8C,$1E,$99
	.dc.b	$E3,$1F,$A6,$9F,$9B,$39,$7D,$EF,$77,$2F,$7F,$4B,$FC,$FD,$97,$EA
	.dc.b	$1B,$8F,$EB,$5B,$8F,$FB,$05,$FF,$42,$57,$FB,$0F,$E1,$F1,$97,$F2
	.dc.b	$A8,$C0,$9D,$BF,$EC,$53,$11,$B7,$11,$A9,$F8,$BC,$6F,$43,$F9,$E5
	.dc.b	$FF,$01,$BE,$AF,$F0,$39,$D7,$17,$CC,$9C,$E9,$FC,$F2,$73,$1F,$30
	.dc.b	$FE,$7F,$87,$6C,$3A,$D7,$8E,$3F,$33,$5C,$36,$F0,$5F,$A5,$F0,$5F
	.dc.b	$F4,$F7,$35,$CF,$0F,$15,$7B,$F1,$DC,$62,$7A,$DF,$1C,$7C,$17,$FB
	.dc.b	$8C,$7C,$11,$8C,$78,$CF,$EF,$0F,$9A,$E0,$F0,$BC,$6B,$A3,$F4,$EF
	.dc.b	$BE,$1F,$B4,$8F,$8A,$7F,$BE,$F0,$F8,$B8,$F2,$3B,$9D,$30,$ED,$8F
	.dc.b	$8C,$F9,$98,$F6,$BF,$96,$78,$70,$EC,$FC,$79,$DC,$63,$1F,$8D,$3F
	.dc.b	$4C,$F1,$8C,$7C,$75,$DE,$3C,$D7,$DD,$F8,$31,$D1,$F8,$87,$73,$77
	.dc.b	$DD,$F2,$1A,$BF,$0E,$3E,$FA,$E4,$FB,$DE,$98,$DA,$E6,$71,$39,$B9
	.dc.b	$7D,$18,$79,$5F,$11,$8C,$3C,$53,$F2,$A3,$EA,$1F,$93,$C7,$DE,$E3
	.dc.b	$0F,$A3,$87,$C1,$8E,$47,$86,$B9,$3F,$42,$76,$F1,$9F,$59,$FE,$8C
	.dc.b	$E0,$7C,$47,$F8,$97,$5F,$EE,$8D,$CB,$5E,$05,$C3,$FC,$03,$37,$CB
	.dc.b	$FF,$53,$37,$0E,$77,$FE,$9B,$BC,$4F,$1B,$FB,$5E,$65,$D4,$EA,$FF
	.dc.b	$DB,$91,$86,$C7,$F8,$56,$7F,$70,$D7,$F2,$CD,$7A,$C7,$F0,$99,$7D
	.dc.b	$33,$CD,$F5,$46,$1F,$47,$C8,$E2,$75,$5F,$E3,$66,$DF,$FA,$1F,$F1
	.dc.b	$4F,$F2,$84,$E6,$FE,$14,$62,$7F,$CF,$9C,$4E,$CF,$80,$DB,$FF,$7E
	.dc.b	$FA,$AF,$A7,$C5,$F0,$BD,$9F,$3E,$75,$D9,$8B,$E3,$7F,$E3,$15,$7F
	.dc.b	$E8,$BF,$A7,$FF,$6D,$C6,$39,$3F,$C7,$91,$89,$FF,$63,$E0,$FA,$77
	.dc.b	$C2,$FD,$5A,$D5,$F0,$F6,$FE,$64,$71,$C3,$F9,$D3,$5F,$F3,$2B,$5F
	.dc.b	$1B,$CB,$1F,$2F,$D9,$3F,$2F,$7D,$57,$7C,$3F,$19,$39,$8F,$C3,$7D
	.dc.b	$1F,$FB,$62,$FD,$47,$F9,$C9,$C6,$3E,$AD,$1E,$CF,$0F,$3A,$5F,$CE
	.dc.b	$97,$CD,$62,$F8,$3F,$97,$F5,$7F,$31,$F5,$56,$FF,$DE,$4D,$AF,$F5
	.dc.b	$6B,$78,$3F,$CB,$0F,$E6,$FE,$A3,$FC,$FD,$5E,$6A,$FA,$71,$87,$FD
	.dc.b	$EB,$F5,$2F,$AA,$FF,$53,$D6,$E5,$5B,$C3,$FC,$E4,$F1,$8F,$ED,$0D
	.dc.b	$7F,$D2,$6D,$7B,$FA,$1F,$A1,$8E,$55,$FC,$E7,$EC,$3F,$D0,$AF,$78
	.dc.b	$CD,$7F,$2B,$9C,$BF,$EC,$AF,$9E,$FD,$D7,$D5,$FE,$ED,$B7,$ED,$F1
	.dc.b	$DE,$3A,$73,$7C,$6B,$52,$BF,$E0,$1F,$51,$FE,$D5,$9F,$80,$DA,$FB
	.dc.b	$5F,$2F,$AF,$80,$95,$F6,$7E,$5E,$D7,$DF,$7F,$89,$7D,$5B,$E4,$BF
	.dc.b	$61,$C3,$FE,$BD,$8C,$78,$1A,$9C,$E1,$FA,$69,$CB,$F5,$DE,$4E,$E7
	.dc.b	$83,$4D,$4F,$2C,$15,$C9,$5F,$86,$75,$39,$8F,$F4,$F7,$D5,$FF,$25
	.dc.b	$F9,$9F,$E4,$CD,$C3,$E3,$B7,$0F,$F7,$DB,$6B,$FD,$4D,$F1,$3F,$CB
	.dc.b	$7E,$77,$F9,$73,$6F,$E7,$B6,$FC,$71,$FC,$F6,$31,$5F,$EB,$DD,$D1
	.dc.b	$F5,$0F,$E4,$7F,$56,$F0,$31,$7F,$5A,$7F,$96,$57,$08,$FF,$1A,$60
	.dc.b	$8C,$3F,$E3,$76,$BF,$E3,$F7,$38,$7F,$4D,$FC,$21,$C5,$DA,$BC,$E7
	.dc.b	$CA,$8C,$DF,$E9,$7E,$03,$FA,$14,$E8,$38,$BF,$F7,$E4,$BE,$37,$F9
	.dc.b	$BC,$7F,$B6,$E7,$53,$FE,$FA,$FA,$97,$95,$FD,$1D,$B1,$EC,$6E,$1F
	.dc.b	$E5,$16,$BF,$CB,$87,$CB,$BE,$B1,$E1,$FF,$65,$8C,$3F,$D9,$F9,$9F
	.dc.b	$0F,$E2,$BE,$37,$C3,$1F,$77,$C1,$FF,$65,$CB,$FC,$A7,$AF,$A6,$31
	.dc.b	$E9,$4A,$F3,$3E,$92,$3F,$E6,$7C,$BB,$CF,$87,$23,$C7,$2F,$9F,$3A
	.dc.b	$AF,$E2,$58,$E9,$B5,$FD,$DF,$EA,$DF,$59,$E5,$89,$3F,$1D,$F9,$44
	.dc.b	$70,$7F,$23,$B7,$A7,$F7,$31,$FB,$58,$F7,$7C,$71,$9D,$17,$F5,$FD
	.dc.b	$BF,$88,$78,$FE,$82,$C7,$A2,$3E,$D1,$1F,$B4,$DB,$F7,$FD,$63,$E0
	.dc.b	$38,$C6,$1F,$F0,$A7,$D5,$7F,$4A,$BF,$E6,$8D,$9F,$8E,$3E,$B3,$7E
	.dc.b	$9C,$AF,$D1,$D6,$EB,$FD,$3C,$D7,$F1,$0D,$7F,$BB,$BD,$73,$F7,$C7
	.dc.b	$F1,$BF,$AC,$F8,$55,$C2,$FF,$15,$18,$9E,$3E,$97,$A0,$FC,$77,$07
	.dc.b	$E8,$6F,$D3,$19,$7C,$AF,$F8,$97,$1F,$EC,$71,$82,$57,$E1,$3E,$1F
	.dc.b	$E3,$47,$23,$ED,$38,$8F,$D9,$4B,$C8,$F9,$27,$E7,$F3,$7C,$06,$30
	.dc.b	$F9,$73,$E1,$23,$F2,$1B,$7F,$94,$EF,$FF,$24,$BD,$A3,$F5,$51,$8B
	.dc.b	$F8,$F2,$C1,$C5,$F6,$8F,$17,$9B,$71,$FB,$71,$F0,$48,$FD,$87,$FC
	.dc.b	$91,$F5,$8F,$AA,$FA,$7F,$11,$FC,$2F,$1F,$F1,$6C,$E6,$FF,$DB,$9E
	.dc.b	$1F,$F0,$38,$FE,$FD,$5B,$A4,$6D,$46,$63,$0D,$BE,$F5,$CF,$D4,$B7
	.dc.b	$E0,$9A,$BE,$AD,$F3,$3F,$C6,$A3,$AD,$FF,$15,$9C,$F8,$BE,$69,$F4
	.dc.b	$FC,$8F,$29,$CC,$70,$3F,$C9,$2E,$32,$F5,$36,$FD,$D5,$C7,$0F,$05
	.dc.b	$69,$B3,$E1,$6D,$F8,$34,$F1,$3E,$98,$CB,$71,$CE,$3D,$E3,$81,$E5
	.dc.b	$34,$57,$4E,$30,$F9,$98,$F9,$6B,$9D,$F6,$BD,$E8,$AD,$61,$CE,$B4
	.dc.b	$79,$27,$58,$F2,$C2,$57,$F4,$13,$E3,$D7,$D1,$71,$87,$CC,$6C,$65
	.dc.b	$FE,$91,$F1,$AF,$93,$FF,$0C,$47,$31,$19,$E8,$FE,$DA,$D8,$D3,$5E
	.dc.b	$FF,$58,$F8,$FF,$EE,$78,$A3,$13,$D6,$31,$3A,$9C,$CF,$28,$FA,$31
	.dc.b	$C3,$6F,$C0,$B7,$7F,$EC,$1F,$59,$FA,$5B,$DD,$C7,$EA,$56,$DF,$F4
	.dc.b	$9C,$ED,$CC,$EE,$FA,$BF,$84,$9D,$46,$27,$64,$7F,$59,$1A,$BF,$CF
	.dc.b	$5F,$7A,$FD,$23,$70,$F7,$C7,$E2,$6B,$83,$F9,$E9,$D1,$30,$F1,$70
	.dc.b	$FF,$7D,$C7,$C4,$4C,$7B,$CE,$AF,$D1,$C3,$B9,$F8,$AD,$8F,$82,$D5
	.dc.b	$8B,$E6,$F0,$FE,$4E,$39,$90,$F4,$47,$1F,$AC,$FE,$F3,$1F,$DA,$FE
	.dc.b	$91,$F2,$5F,$E1,$3E,$AF,$E9,$3C,$29,$CD,$7F,$A3,$6F,$8B,$FE,$29
	.dc.b	$6B,$F9,$B6,$BF,$E1,$D6,$BF,$EC,$C6,$A5,$E2,$F8,$79,$BF,$DA,$37
	.dc.b	$E1,$F2,$31,$7F,$E7,$5C,$3C,$38,$FF,$38,$C7,$8C,$F9,$71,$DE,$1C
	.dc.b	$63,$DD,$87,$27,$C1,$36,$B8,$C7,$AE,$30,$72,$E0,$F9,$91,$EF,$F5
	.dc.b	$BA,$DF,$F8,$6A,$3F,$C2,$8D,$74,$7C,$78,$DF,$95,$23,$04,$36,$4C
	.dc.b	$BE,$8F,$4F,$F3,$48,$FA,$1E,$8F,$B3,$EB,$3F,$5F,$5F,$59,$BB,$FC
	.dc.b	$F9,$F6,$1B,$E5,$97,$C8,$87,$D9,$8D,$BF,$E2,$DC,$1F,$4D,$F1,$F9
	.dc.b	$BF,$BA,$4F,$F8,$C7,$EB,$3F,$CD,$ED,$F5,$75,$E2,$C8,$C1,$0F,$F3
	.dc.b	$75,$72,$6C,$6D,$A7,$FB,$67,$D4,$7B,$7E,$1F,$13,$ED,$F8,$3F,$E2
	.dc.b	$FA,$FF,$21,$35,$EB,$1B,$73,$E7,$70,$30,$E8,$FF,$73,$63,$FD,$F9
	.dc.b	$B3,$F8,$8F,$AC,$FF,$80,$31,$FE,$4F,$FA,$CF,$C2,$F2,$7E,$91,$BC
	.dc.b	$7F,$96,$6C,$87,$D6,$3E,$6F,$F2,$0D,$8E,$9D,$7F,$96,$AE,$31,$C2
	.dc.b	$BF,$33,$A7,$E4,$31,$FF,$0A,$E1,$C6,$BF,$D3,$0D,$8F,$DB,$CE,$F1
	.dc.b	$F1,$0F,$9E,$7C,$17,$EA,$A3,$0D,$8C,$E1,$EF,$F3,$3C,$FF,$59,$E5
	.dc.b	$64,$7B,$71,$F1,$CF,$EB,$BE,$B3,$FA,$D4,$7F,$08,$6E,$3F,$BA,$6D
	.dc.b	$7F,$A7,$BD,$2F,$C6,$BC,$E7,$FD,$45,$F5,$8F,$AC,$FD,$B7,$8D,$FE
	.dc.b	$C9,$AD,$1F,$35,$C3,$FE,$B9,$9E,$0E,$83,$87,$FD,$7D,$1C,$31,$FF
	.dc.b	$1F,$B6,$FC,$37,$FC,$16,$F8,$5F,$F7,$56,$BA,$3F,$DA,$EE,$17,$C5
	.dc.b	$3E,$B1,$F2,$BF,$C6,$21,$F1,$9F,$09,$B5,$9A,$DB,$EE,$8E,$55,$C6
	.dc.b	$B8,$1C,$6B,$75,$FE,$DE,$AE,$BE,$0E,$FE,$0A,$4F,$3B,$FF,$06,$6B
	.dc.b	$FE,$84,$BD,$CE,$6B,$6F,$3A,$DE,$9F,$ED,$5F,$AC,$7C,$A7,$F2,$95
	.dc.b	$F5,$94,$C4,$4B,$FA,$53,$A6,$CE,$5F,$F7,$84,$74,$72,$43,$33,$FC
	.dc.b	$DF,$D7,$FF,$31,$67,$FE,$76,$FA,$CF,$F7,$BD,$7D,$F6,$BF,$EF,$B9
	.dc.b	$DE,$3F,$EE,$D7,$FC,$D5,$1F,$E2,$F6,$B9,$F0,$3F,$C5,$0F,$EE,$63
	.dc.b	$2F,$DC,$CD,$F3,$DF,$4F,$E8,$64,$FC,$D8,$FA,$C6,$1F,$95,$18,$7F
	.dc.b	$29,$5F,$DD,$DF,$CB,$7C,$13,$EB,$3F,$AA,$36,$37,$EB,$7C,$13,$33
	.dc.b	$23,$24,$3F,$C8,$15,$D5,$B3,$EC,$7F,$CF,$AF,$BB,$8B,$F2,$7F,$25
	.dc.b	$F8,$DF,$5E,$57,$E1,$E3,$F2,$06,$1F,$C2,$46,$1F,$D7,$7D,$5B,$EA
	.dc.b	$FF,$B6,$7D,$63,$CD,$A7,$F6,$F9,$D5,$70,$AF,$24,$62,$3F,$CB,$0D
	.dc.b	$7F,$CC,$8D,$7F,$96,$0F,$AB,$FF,$30,$3F,$BF,$F0,$1F,$59,$FD,$F5
	.dc.b	$B1,$FC,$0E,$76,$B9,$FF,$D7,$F1,$87,$FD,$83,$3C,$27,$5C,$DE,$08
	.dc.b	$C4,$F0,$F3,$BF,$C1,$57,$D2,$FF,$CF,$4D,$7C,$ED,$7F,$A2,$35,$FE
	.dc.b	$69,$E4,$7F,$84,$FE,$3F,$F0,$11,$E0,$3F,$A5,$38,$BF,$F6,$AF,$91
	.dc.b	$FD,$7D,$B5,$F7,$63,$FB,$15,$7F,$A2,$DA,$FC,$86,$AF,$A7,$7D,$63
	.dc.b	$E1,$FF,$87,$7B,$12,$BF,$97,$D6,$A7,$FC,$62,$D8,$FF,$47,$35,$FF
	.dc.b	$4D,$35,$FA,$ED,$7E,$CF,$D5,$BE,$B3,$FC,$64,$FE,$DA,$3F,$B4,$35
	.dc.b	$3B,$7F,$9E,$35,$FF,$4F,$35,$FF,$55,$35,$EF,$1F,$9F,$EF,$7C,$8F
	.dc.b	$F2,$B2,$4F,$F2,$5F,$80,$AF,$F2,$18,$BB,$7F,$27,$AF,$F2,$7F,$1F
	.dc.b	$FA,$C3,$8F,$F1,$56,$FA,$CF,$F3,$2B,$E8,$F5,$3F,$2F,$18,$D3,$FB
	.dc.b	$C3,$AF,$8A,$5F,$28,$C3,$D3,$8F,$85,$FE,$81,$B8,$7F,$83,$5D,$E7
	.dc.b	$7B,$3C,$8F,$F8,$47,$0D,$D7,$96,$7E,$8D,$F9,$1D,$C6,$4F,$F0,$C6
	.dc.b	$3E,$4B,$E9,$1F,$52,$BC,$55,$CA,$32,$7F,$B3,$B1,$F4,$47,$4C,$3C
	.dc.b	$24,$70,$49,$E8,$79,$DE,$8F,$8C,$E9,$79,$7C,$B5,$F1,$E7,$75,$D6
	.dc.b	$73,$5B,$78,$8C,$7E,$19,$2B,$E0,$BF,$2A,$3C,$19,$7D,$69,$EA,$79
	.dc.b	$E3,$9D,$F7,$7E,$F1,$DF,$2F,$90,$FA,$5B,$7D,$E3,$DE,$78,$EB,$C2
	.dc.b	$75,$BF,$04,$6A,$FC,$84,$9F,$1C,$65,$F2,$5F,$82,$7C,$95,$CA,$BC
	.dc.b	$F5,$D7,$2F,$C0,$47,$13,$E3,$9E,$7C,$BE,$63,$93,$DA,$67,$2F,$74
	.dc.b	$6A,$33,$87,$C3,$C7,$C5,$87,$79,$FA,$58,$F8,$E7,$31,$E0,$AF,$3E
	.dc.b	$3E,$48,$F3,$99,$BF,$7B,$D9,$F7,$E1,$FB,$0F,$7F,$FD,$01,$87,$D0
	.dc.b	$6C,$67,$3F,$CE,$C7,$F2,$2F,$83,$FF,$3A,$37,$03,$49,$3F,$D3,$A3
	.dc.b	$FC,$F5,$3F,$8B,$E2,$79,$8D,$FF,$1D,$5A,$3F,$31,$3A,$8C,$C2,$B9
	.dc.b	$57,$18,$FC,$F1,$C7,$5F,$E1,$37,$C6,$32,$71,$3A,$CF,$19,$FE,$1F
	.dc.b	$1C,$74,$FE,$5F,$39,$7D,$51,$C6,$7A,$CE,$63,$DD,$7E,$4C,$3C,$ED
	.dc.b	$8F,$0B,$6B,$BC,$73,$BE,$F5,$C2,$3C,$AD,$9E,$ED,$9D,$CF,$43,$65
	.dc.b	$9D,$B6,$37,$1E,$0A,$F1,$47,$B1,$FE,$23,$87,$C7,$C7,$57,$FA,$D3
	.dc.b	$FD,$65,$19,$AE,$11,$E0,$BD,$9E,$02,$C7,$80,$DB,$97,$8D,$79,$8E
	.dc.b	$E5,$76,$71,$9F,$9A,$77,$72,$F9,$31,$FA,$4F,$F8,$73,$0F,$6B,$C2
	.dc.b	$FC,$0D,$AE,$8E,$CF,$71,$FD,$EE,$BF,$19,$7E,$49,$F9,$EF,$B2,$7C
	.dc.b	$18,$FD,$36,$9F,$4E,$3D,$4F,$A1,$AF,$7B,$F1,$B6,$BD,$DE,$0F,$FA
	.dc.b	$73,$D9,$FC,$E7,$0F,$F6,$76,$E3,$FE,$E3,$9F,$F5,$0F,$D6,$63,$0F
	.dc.b	$F3,$65,$59,$E5,$5F,$E9,$78,$C7,$1A,$EC,$71,$BF,$F0,$D7,$91,$FA
	.dc.b	$B7,$FC,$22,$D7,$FC,$22,$D7,$F0,$6D,$78,$DF,$FB,$1F,$77,$89,$D5
	.dc.b	$55,$8F,$44,$3E,$B3,$FD,$07,$5F,$E2,$26,$DF,$F8,$89,$AF,$F8,$89
	.dc.b	$AF,$E5,$1D,$E9,$FE,$B3,$FA,$CF,$E1,$95,$FE,$30,$9C,$BF,$26,$33
	.dc.b	$5F,$E4,$53,$F8,$9F,$D6,$7A,$F8,$8D,$25,$7F,$A7,$23,$2F,$FB,$AB
	.dc.b	$D2,$7D,$67,$F8,$92,$3F,$DD,$13,$AB,$FF,$29,$56,$BA,$FE,$87,$D6
	.dc.b	$7F,$B7,$2F,$FD,$4E,$DA,$FE,$1E,$D7,$F9,$53,$5F,$F3,$CB,$BD,$3F
	.dc.b	$91,$7D,$5B,$EB,$1F,$59,$DB,$D9,$FC,$34,$E6,$31,$8B,$1F,$F3,$34
	.dc.b	$E6,$3F,$E0,$06,$3D,$5D,$46,$67,$E4,$7D,$17,$FB,$6A,$7F,$D7,$ED
	.dc.b	$8F,$E4,$ED,$7F,$A1,$35,$FF,$48,$BB,$CB,$F9,$D6,$8F,$AC,$7C,$D7
	.dc.b	$E5,$6B,$58,$79,$EB,$53,$F9,$CF,$43,$F3,$BF,$59,$6B,$FD,$EC,$DC
	.dc.b	$BF,$DE,$4D,$CB,$FC,$A4,$FE,$87,$EA,$BF,$CF,$15,$FE,$D8,$6D,$7F
	.dc.b	$40,$6B,$FD,$6D,$AF,$FA,$A9,$DD,$7F,$53,$E7,$7D,$57,$F9,$D6,$FE
	.dc.b	$74,$FE,$B6,$B7,$1F,$CB,$27,$57,$F9,$39,$CB,$FE,$8D,$8C,$D7,$1F
	.dc.b	$2B,$F5,$31,$8B,$FE,$09,$DC,$F8,$5F,$DF,$FD,$57,$F0,$06,$CF,$88
	.dc.b	$DB,$8C,$7F,$D4,$DE,$0A,$F9,$56,$E7,$FD,$F1,$7F,$C1,$FE,$9B,$F7
	.dc.b	$B3,$FE,$FE,$6C,$FF,$66,$6B,$FE,$01,$6B,$FE,$D0,$77,$9F,$E0,$38
	.dc.b	$DF,$1D,$D7,$C0,$FE,$39,$C9,$F8,$38,$FE,$3D,$F5,$4F,$97,$FD,$CE
	.dc.b	$1F,$E0,$18,$C7,$C5,$EA,$73,$A7,$E7,$67,$2F,$AD,$E4,$EE,$78,$34
	.dc.b	$D4,$F2,$D1,$5C,$9F,$F6,$63,$CA,$77,$5F,$CB,$BE,$AB,$F9,$7F,$C8
	.dc.b	$D9,$F9,$5C,$3F,$36,$31,$87,$D0,$E1,$46,$33,$FF,$53,$3F,$CE,$3E
	.dc.b	$A7,$F9,$CF,$CC,$5C,$79,$DA,$31,$87,$36,$E1,$FE,$AF,$7E,$C5,$FF
	.dc.b	$37,$E2,$34,$98,$7D,$B7,$FA,$65,$3E,$85,$F2,$BC,$AB,$F0,$93,$98
	.dc.b	$C6,$3F,$EE,$F8,$C4,$FE,$26,$18,$78,$23,$8D,$F4,$75,$EC,$7F,$46
	.dc.b	$6F,$97,$FD,$75,$7C,$3F,$F3,$AC,$3F,$8C,$56,$DF,$E7,$5F,$4C,$E6
	.dc.b	$FE,$C9,$CA,$F7,$75,$AF,$F5,$DA,$D6,$5F,$9D,$AD,$67,$FD,$AF,$EA
	.dc.b	$5F,$52,$F0,$F0,$DB,$E1,$97,$47,$F9,$55,$70,$AF,$ED,$9A,$BD,$CF
	.dc.b	$ED,$D9,$FF,$94,$F4,$FE,$DD,$39,$C3,$E3,$4E,$5F,$E6,$9B,$23,$0F
	.dc.b	$E8,$72,$E0,$FC,$0F,$37,$DF,$18,$C3,$FC,$C9,$B7,$C1,$CB,$27,$8A
	.dc.b	$AE,$3E,$08,$F2,$9F,$E0,$3F,$A9,$7C,$EE,$27,$47,$C7,$F3,$FE,$C6
	.dc.b	$7F,$42,$FE,$7C,$63,$2F,$9D,$18,$E6,$E5,$49,$F2,$D6,$EF,$DF,$AB
	.dc.b	$D4,$F8,$E7,$8F,$E6,$AC,$FC,$68,$FB,$BF,$53,$FE,$D3,$8C,$1F,$6D
	.dc.b	$C6,$3F,$84,$9D,$FB,$5C,$A3,$C5,$5F,$D8,$78,$17,$C3,$3F,$F1,$17
	.dc.b	$D3,$3E,$3B,$FC,$4C,$DB,$FC,$46,$0F,$A9,$CD,$49,$FC,$67,$D4,$FB
	.dc.b	$6B,$FB,$75,$F1,$7C,$27,$F1,$B8,$D2,$8D,$61,$F6,$A7,$31,$8A,$F4
	.dc.b	$46,$3D,$1F,$29,$FA,$0C,$FF,$4F,$3F,$B1,$AF,$ED,$6D,$9F,$D2,$FC
	.dc.b	$0F,$BE,$3F,$1F,$3A,$C7,$CB,$6A,$D4,$7A,$AF,$8E,$BE,$0C,$3F,$55
	.dc.b	$5A,$AF,$57,$93,$F2,$0F,$D9,$E6,$FA,$9C,$5F,$96,$7F,$6F,$E0,$FE
	.dc.b	$06,$30,$7E,$D5,$C5,$FF,$8D,$F4,$FE,$8F,$39,$E1,$E7,$3E,$8A,$DE
	.dc.b	$3F,$6B,$8B,$FC,$AD,$81,$A4,$C3,$FD,$7B,$7F,$46,$3C,$7D,$9F,$80
	.dc.b	$6C,$FF,$1B,$FA,$97,$D3,$FF,$B4,$7E,$0F,$F8,$AE,$1F,$E2,$B9,$CD
	.dc.b	$7F,$B8,$3C,$4F,$DD,$C7,$F8,$32,$B7,$48,$DB,$FE,$DD,$9D,$DF,$56
	.dc.b	$DF,$39,$ED,$E9,$E5,$F0,$55,$5F,$52,$F9,$3F,$E4,$91,$D4,$FF,$20
	.dc.b	$46,$7C,$7E,$91,$E3,$FA,$8E,$53,$98,$E0,$74,$AE,$7A,$7A,$A7,$4F
	.dc.b	$B3,$45,$F2,$AD,$E9,$FC,$72,$F6,$D8,$CE,$1F,$26,$B8,$63,$E4,$BD
	.dc.b	$CF,$A2,$B6,$F9,$F4,$D4,$62,$77,$3E,$7A,$F6,$E3,$E4,$BD,$C4,$9F
	.dc.b	$6D,$70,$AD,$46,$2F,$A1,$2B,$9E,$1E,$69,$D4,$7E,$2A,$FE,$44,$7F
	.dc.b	$3A,$3C,$45,$29,$9F,$88,$7C,$7F,$F8,$76,$35,$18,$EA,$FC,$AB,$63
	.dc.b	$33,$B7,$B5,$3E,$A5,$F1,$5E,$36,$CF,$86,$75,$39,$24,$7C,$29,$DE
	.dc.b	$FF,$75,$BB,$FF,$54,$FA,$9F,$D0,$D7,$D5,$9C,$BF,$8C,$6B,$F9,$26
	.dc.b	$C7,$08,$ED,$7C,$48,$FE,$3A,$35,$7E,$19,$CB,$67,$5C,$3F,$EF,$FC
	.dc.b	$3F,$19,$87,$83,$1F,$6D,$76,$36,$72,$9D,$E1,$63,$FD,$45,$8F,$82
	.dc.b	$73,$1C,$71,$F0,$1F,$B8,$9D,$43,$D3,$7C,$3B,$D6,$F2,$4E,$61,$F2
	.dc.b	$98,$AD,$B6,$38,$4F,$EF,$67,$9D,$66,$A4,$E6,$B6,$FE,$EE,$BC,$3C
	.dc.b	$7F,$E0,$8E,$AF,$F7,$16,$D7,$9B,$FD,$C0,$DB,$F2,$55,$BE,$31,$FE
	.dc.b	$56,$C3,$F1,$4D,$AD,$47,$F0,$B8,$D7,$CD,$63,$E7,$24,$60,$87,$FA
	.dc.b	$32,$51,$A8,$C4,$F4,$31,$97,$FB,$E6,$71,$3E,$3F,$95,$F0,$F8,$D2
	.dc.b	$30,$43,$B3,$97,$C1,$1E,$83,$14,$D3,$FD,$C1,$9F,$D8,$BD,$5F,$C1
	.dc.b	$F1,$7D,$FF,$3F,$23,$04,$3D,$8E,$5F,$74,$7C,$63,$18,$77,$8F,$C1
	.dc.b	$4E,$5F,$9D,$F4,$FF,$33,$EE,$64,$60,$87,$C9,$72,$FC,$B8,$FA,$66
	.dc.b	$2B,$B6,$3F,$92,$AC,$D7,$DB,$FC,$0F,$EC,$FD,$8C,$8C,$10,$FA,$CE
	.dc.b	$5F,$B1,$1F,$80,$31,$7D,$B8,$15,$9A,$FC,$47,$E0,$5F,$86,$F3,$64
	.dc.b	$60,$FC,$67,$A7,$F8,$A9,$91,$83,$F2,$1E,$97,$E3,$A6,$46,$0F,$CA
	.dc.b	$79,$7F,$92,$F1,$E4,$60,$FF,$56,$F8,$1F,$C7,$9F,$F6,$DA,$F3,$BF
	.dc.b	$3C,$D8,$DF,$81,$8F,$8A,$CF,$83,$27,$9C,$7E,$9D,$7D,$2B,$E6,$70
	.dc.b	$BE,$E2,$98,$89,$3F,$EA,$19,$D3,$67,$2F,$F9,$08,$EB,$3C,$9D,$1C
	.dc.b	$5F,$17,$D3,$3A,$7F,$C2,$F2,$8C,$3F,$E1,$A6,$BF,$D3,$DA,$FF,$53
	.dc.b	$73,$CB,$F9,$58,$E5,$3D,$8F,$6E,$7F,$22,$FE,$4E,$12,$BE,$7C,$65
	.dc.b	$FA,$31,$97,$E8,$C7,$D0,$F9,$A9,$1F,$59,$FA,$7F,$51,$88,$FA,$F1
	.dc.b	$71,$FC,$DE,$CA,$C7,$77,$EE,$21,$58,$BF,$E6,$51,$B4,$7F,$9B,$BE
	.dc.b	$3D,$FA,$57,$FA,$72,$3F,$0F,$EB,$FE,$10,$FE,$49,$3A,$E6,$B5,$BA
	.dc.b	$E4,$DC,$FF,$79,$5B,$9F,$C8,$F9,$9F,$8E,$FB,$99,$1F,$94,$7F,$23
	.dc.b	$F7,$38,$8F,$CA,$C5,$BF,$7C,$FE,$3F,$EA,$1F,$51,$7E,$1E,$61,$FE
	.dc.b	$EB,$9C,$B6,$72,$FE,$2E,$13,$CE,$64,$F1,$8D,$DF,$E0,$6B,$9C,$FE
	.dc.b	$A3,$D7,$FD,$3E,$F2,$3F,$5B,$1F,$AA,$30,$FE,$BA,$30,$FF,$BF,$FF
	.dc.b	$48,$7C,$77,$EA,$7F,$4E,$FE,$91,$3F,$9F,$6E,$3F,$91,$AD,$CF,$1A
	.dc.b	$FD,$84,$62,$3F,$4C,$D9,$FC,$C5,$7F,$62,$7F,$35,$EE,$FF,$01,$6F
	.dc.b	$7F,$EE,$63,$F8,$05,$FE,$BD,$D7,$37,$FB,$B9,$B1,$F8,$87,$75,$CA
	.dc.b	$3F,$16,$F6,$EA,$E8,$75,$B9,$3A,$BE,$93,$FC,$27,$F4,$AF,$F6,$54
	.dc.b	$E4,$FC,$AF,$C2,$3E,$27,$F8,$73,$63,$42,$78,$BF,$C1,$DB,$8F,$EE
	.dc.b	$5B,$87,$E8,$27,$87,$9F,$D3,$E3,$7F,$3B,$FA,$67,$FA,$43,$0F,$D6
	.dc.b	$FD,$DF,$F2,$43,$3E,$34,$C3,$CD,$D6,$BE,$19,$7E,$B2,$31,$5F,$CB
	.dc.b	$3E,$EF,$F9,$6F,$C0,$FF,$2E,$FA,$D9,$8F,$F3,$48,$CB,$FC,$DA,$32
	.dc.b	$FF,$36,$8F,$E6,$BF,$5D,$FC,$DB,$E8,$24,$FF,$3C,$8C,$BF,$CF,$A3
	.dc.b	$2F,$F3,$E8,$FE,$6D,$F0,$5F,$ED,$7D,$BF,$80,$FD,$43,$EA,$3E,$27
	.dc.b	$D4,$FF,$08,$C3,$31,$8F,$23,$E4,$56,$E7,$F8,$4E,$1C,$D4,$65,$46
	.dc.b	$63,$17,$D6,$77,$7D,$5E,$E7,$4F,$21,$8E,$23,$9C,$61,$E2,$7F,$1A
	.dc.b	$AE,$97,$DB,$1F,$3E,$1F,$7F,$1F,$13,$70,$E5,$5B,$C3,$B6,$1C,$AB
	.dc.b	$58,$F8,$70,$E3,$5D,$F1,$EA,$D3,$58,$F4,$BE,$31,$E6,$18,$75,$78
	.dc.b	$19,$BE,$AF,$BB,$0F,$1E,$1D,$A3,$8B,$67,$67,$0A,$E9,$3A,$3C,$75
	.dc.b	$1F,$14,$68,$DC,$F2,$C3,$D1,$3E,$4C,$7E,$23,$E5,$8C,$BE,$53,$9E
	.dc.b	$1E,$5B,$E7,$3A,$BF,$24,$E4,$F4,$46,$4E,$73,$E3,$8C,$E1,$E7,$C3
	.dc.b	$C2,$4C,$7E,$5B,$F8,$DC,$3A,$E3,$F1,$8E,$13,$A3,$F1,$8F,$98,$C9
	.dc.b	$F8,$DA,$E0,$E4,$F7,$3A,$7E,$51,$9A,$F0,$63,$E9,$9E,$31,$E5,$9E
	.dc.b	$87,$23,$85,$78,$A3,$8C,$FD,$3E,$9F,$E5,$F8,$7E,$A5,$B8,$7E,$75
	.dc.b	$B3,$FD,$D9,$FE,$6B,$87,$F7,$8F,$84,$C7,$5F,$FB,$9D,$B1,$FE,$EB
	.dc.b	$6B,$FC,$7D,$AF,$13,$F2,$9C,$6F,$92,$FF,$39,$67,$FB,$99,$D3,$FB
	.dc.b	$B9,$D7,$2F,$EC,$E7,$3A,$FF,$2B,$9C,$C7,$E7,$CC,$9F,$C8,$EF,$AC
	.dc.b	$F9,$E3,$C3,$18,$7D,$06,$2B,$B2,$C3,$F8,$15,$F6,$C3,$CD,$39,$38
	.dc.b	$1E,$13,$0F,$A6,$FB,$47,$C3,$37,$1C,$AB,$AC,$EA,$7D,$22,$BC,$18
	.dc.b	$7A,$6F,$A6,$1E,$6C,$38,$B6,$FA,$D6,$CF,$8D,$87,$58,$F2,$C6,$27
	.dc.b	$85,$F4,$9F,$55,$F4,$69,$D0,$F8,$F1,$B3,$D4,$E4,$F7,$39,$AF,$14
	.dc.b	$7D,$3A,$EF,$1E,$18,$EB,$3B,$7A,$C6,$63,$A9,$98,$EB,$19,$3A,$BF
	.dc.b	$1E,$FD,$D7,$E9,$AE,$35,$BA,$EA,$78,$CF,$0B,$97,$A9,$98,$E2,$D8
	.dc.b	$EF,$87,$D4,$8E,$67,$87,$0F,$6B,$B3,$31,$BF,$6D,$E1,$3E,$93,$FD
	.dc.b	$A5,$B7,$FB,$1B,$F5,$7F,$F3,$2E,$3F,$EA,$89,$CB,$FE,$E4,$8C,$E3
	.dc.b	$FE,$AA,$9C,$AC,$13,$AB,$FF,$3B,$D6,$A7,$35,$69,$F1,$57,$6C,$37
	.dc.b	$1F,$F1,$25,$F4,$8C,$5F,$F6,$69,$F2,$46,$31,$FF,$38,$FD,$43,$E7
	.dc.b	$BC,$BA,$A3,$17,$D3,$3F,$F3,$5B,$FE,$1C,$8C,$E3,$FE,$85,$6E,$1F
	.dc.b	$17,$CA,$E7,$18,$AF,$89,$1F,$0A,$BF,$86,$78,$97,$5F,$F1,$D5,$7F
	.dc.b	$8F,$74,$7D,$43,$EA,$B5,$72,$B5,$2F,$93,$C6,$3E,$F3,$E3,$C6,$67
	.dc.b	$FC,$90,$3F,$0D,$FF,$68,$38,$7F,$D4,$6D,$F5,$FE,$BB,$BE,$CB,$3C
	.dc.b	$2F,$B4,$E4,$E0,$73,$9D,$D7,$06,$C6,$4F,$F2,$AF,$B6,$F8,$4F,$88
	.dc.b	$D8,$F8,$5F,$45,$FE,$D7,$77,$3A,$3F,$17,$B2,$5F,$0D,$9A,$E2,$FE
	.dc.b	$27,$9B,$F1,$5D,$56,$7E,$E3,$74,$F9,$1D,$1F,$94,$8C,$73,$7E,$7D
	.dc.b	$FC,$CF,$D1,$7F,$B1,$3E,$9E,$DB,$E9,$6C,$61,$B8,$FF,$9D,$EB,$FD
	.dc.b	$97,$18,$AF,$AD,$18,$9D,$42,$B9,$46,$1B,$1E,$6E,$AF,$CB,$4F,$E5
	.dc.b	$7A,$BF,$4D,$E8,$7E,$9D,$91,$93,$F3,$51,$8A,$FC,$C7,$D2,$FE,$58
	.dc.b	$FD,$77,$CE,$DA,$FC,$F4,$7E,$C7,$E6,$B1,$5E,$17,$F6,$9D,$92,$BF
	.dc.b	$4D,$D7,$6B,$CD,$37,$83,$F8,$44,$DA,$FE,$0D,$36,$BF,$82,$69,$F9
	.dc.b	$77,$8C,$FE,$C7,$E6,$7F,$81,$6D,$FC,$0D,$B1,$93,$F6,$D1,$96,$9F
	.dc.b	$EA,$2F,$A2,$7A,$7F,$B1,$6C,$7F,$A5,$BE,$E9,$D7,$9A,$93,$FE,$A3
	.dc.b	$C7,$FD,$49,$7C,$1C,$07,$0E,$DA,$87,$1B,$E6,$9F,$27,$F8,$C5,$F2
	.dc.b	$76,$30,$DC,$3F,$15,$3A,$6C,$E9,$A6,$A7,$85,$FE,$3A,$B8,$3F,$89
	.dc.b	$E7,$FE,$59,$C9,$FC,$B5,$B7,$FC,$7F,$DD,$FE,$61,$F4,$DF,$CC,$5B
	.dc.b	$1F,$C8,$F5,$FE,$69,$F4,$BF,$CD,$5B,$1F,$C9,$FC,$6F,$E7,$5F,$AC
	.dc.b	$23,$F9,$D4,$61,$87,$F3,$88,$7C,$DE,$34,$FD,$9C,$62,$75,$3D,$3E
	.dc.b	$6F,$AB,$63,$97,$CD,$F8,$3E,$6F,$C2,$D8,$C9,$DE,$32,$D3,$38,$78
	.dc.b	$23,$2D,$8F,$37,$CD,$FB,$38,$FD,$AD,$8C,$9E,$B8,$CB,$4C,$CF,$B2
	.dc.b	$32,$D8,$F8,$5C,$7F,$23,$E0,$7E,$4B,$63,$27,$C7,$8C,$B4,$CC,$FC
	.dc.b	$88,$CB,$63,$E7,$7C,$DF,$D4,$F9,$1F,$AA,$D8,$C9,$F4,$E3,$2D,$33
	.dc.b	$3F,$52,$32,$D8,$FB,$5E,$47,$E0,$3C,$BF,$C0,$B6,$32,$7D,$F8,$CB
	.dc.b	$4C,$CF,$E0,$23,$2D,$8F,$C3,$7C,$DF,$E3,$3E,$6B,$F1,$AD,$8C,$9F
	.dc.b	$8B,$8C,$B4,$FC,$77,$CC,$FE,$4E,$BF,$CD,$1F,$09,$FB,$5F,$31,$CA
	.dc.b	$F9,$AD,$7B,$6A,$D4,$B0,$CC,$62,$AE,$1F,$E1,$DA,$D6,$3F,$E9,$3A
	.dc.b	$D3,$67,$57,$C3,$1F,$0E,$1C,$A5,$F3,$AE,$6F,$F1,$0C,$7A,$0F,$E1
	.dc.b	$91,$9F,$4B,$A4,$3E,$79,$F0,$BC,$EB,$53,$C4,$EB,$7D,$36,$F7,$47
	.dc.b	$48,$C6,$BF,$C6,$63,$A5,$F3,$9C,$C6,$F1,$F0,$E1,$DB,$0E,$73,$C3
	.dc.b	$5C,$DF,$08,$F0,$DE,$0F,$4E,$5E,$8A,$C4,$F9,$F0,$E9,$97,$8D,$B1
	.dc.b	$FC,$B6,$B8,$3B,$C3,$B4,$EA,$3B,$9F,$06,$77,$8F,$9A,$74,$4A,$F3
	.dc.b	$4E,$AF,$C1,$1D,$0F,$A7,$7F,$D1,$6B,$E6,$B6,$32,$F0,$C7,$FA,$3A
	.dc.b	$F8,$27,$C9,$FF,$43,$8F,$E9,$71,$8F,$17,$6D,$8C,$CE,$DF,$07,$63
	.dc.b	$E0,$BA,$E4,$8C,$4F,$F5,$68,$C4,$EA,$73,$3C,$A3,$EC,$C7,$0D,$7E
	.dc.b	$23,$74,$FC,$17,$CF,$7F,$A1,$2F,$77,$1F,$AD,$5B,$7F,$B0,$4E,$DC
	.dc.b	$CE,$EF,$AB,$F9,$89,$D4,$62,$76,$47,$F1,$71,$AB,$FC,$15,$F7,$AF
	.dc.b	$D3,$B7,$0F,$7C,$7E,$7E,$B8,$3F,$A8,$9D,$13,$0F,$17,$C0,$FE,$7F
	.dc.b	$1F,$11,$31,$EF,$3A,$BF,$DD,$71,$BE,$7B,$FA,$CC,$61,$78,$5C,$EF
	.dc.b	$F8,$74,$62,$75,$E1,$65,$FF,$52,$47,$27,$F7,$D7,$D6,$F4,$D8,$E2
	.dc.b	$77,$E2,$FD,$B7,$A7,$FC,$76,$7C,$9A,$E3,$97,$C2,$DB,$EF,$76,$BD
	.dc.b	$4D,$C7,$F1,$53,$FE,$AC,$6B,$FE,$66,$7F,$8A,$39,$F0,$F8,$78,$72
	.dc.b	$30,$43,$83,$97,$94,$77,$31,$97,$A6,$BF,$C4,$B3,$C6,$31,$E1,$F9
	.dc.b	$7C,$39,$3E,$58,$F4,$18,$7C,$D1,$87,$47,$AD,$D4,$F4,$8F,$77,$87
	.dc.b	$F0,$7C,$39,$18,$21,$F0,$5C,$BF,$0E,$3E,$41,$89,$F5,$4E,$27,$E5
	.dc.b	$78,$7F,$37,$C3,$91,$82,$1F,$35,$CB,$F3,$E3,$EA,$18,$AF,$8D,$38
	.dc.b	$9F,$AD,$E1,$FD,$9F,$0E,$46,$08,$7D,$97,$2F,$DB,$8F,$C0,$18,$AF
	.dc.b	$A5,$38,$9F,$C1,$78,$5F,$85,$F0,$A4,$60,$FC,$37,$CB,$FE,$26,$64
	.dc.b	$60,$FC,$57,$CF,$7E,$36,$64,$60,$FC,$77,$CF,$7E,$47,$E7,$A4,$60
	.dc.b	$FF,$7E,$F7,$7F,$18,$CF,$F3,$CB,$E2,$3F,$34,$D9,$DF,$76,$3D,$2C
	.dc.b	$F7,$C9,$E7,$1F,$A3,$5F,$3C,$F9,$DC,$2F,$83,$A6,$22,$4F,$F9,$F7
	.dc.b	$3F,$F0,$63,$CA,$76,$E7,$C0,$FE,$D5,$7D,$63,$07,$FA,$15,$C3,$FE
	.dc.b	$F9,$39,$1A,$29,$97,$75,$E0,$F9,$E7,$47,$A3,$29,$7C,$67,$2D,$9F
	.dc.b	$EA,$6A,$32,$FF,$A7,$1A,$FF,$A8,$5A,$FF,$89,$F5,$F0,$E3,$72,$FA
	.dc.b	$27,$D2,$CE,$57,$D8,$8C,$BF,$6A,$3E,$BF,$C2,$48,$FB,$AF,$DA,$F9
	.dc.b	$EC,$47,$FB,$D3,$04,$E3,$4F,$C0,$D5,$C7,$C9,$36,$DE,$13,$05,$FF
	.dc.b	$21,$F8,$A7,$CF,$39,$BF,$13,$F1,$D3,$1F,$AB,$5A,$9E,$15,$B9,$CB
	.dc.b	$56,$9F,$E4,$DA,$E1,$1F,$98,$F3,$3F,$2B,$F3,$D2,$3F,$36,$FE,$63
	.dc.b	$E7,$B1,$1E,$F8,$B7,$F9,$E9,$FC,$B7,$CE,$BE,$7A,$FC,$84,$C3,$F0
	.dc.b	$D3,$96,$CE,$5F,$C6,$42,$79,$CC,$9E,$31,$BB,$FC,$DD,$73,$9F,$A5
	.dc.b	$F3,$DF,$AE,$E0,$91,$FB,$48,$FD,$89,$87,$EC,$46,$1F,$CD,$7C,$F3
	.dc.b	$DD,$FD,$87,$CF,$7F,$4A,$9F,$D0,$37,$1F,$D5,$D6,$E7,$8D,$7E,$DE
	.dc.b	$31,$1F,$A7,$6C,$FE,$CA,$BF,$B5,$3F,$B4,$F9,$EF,$E1,$1C,$FF,$E5
	.dc.b	$B7,$F8,$35,$FE,$BD,$B3,$BE,$FF,$E0,$91,$C2,$7F,$DF,$4E,$AB,$97
	.dc.b	$81,$AA,$95,$FB,$7A,$D3,$C2,$72,$FE,$F7,$E7,$9E,$67,$F1,$1F,$8E
	.dc.b	$7B,$F4,$FE,$29,$3A,$AF,$E0,$4D,$79,$56,$8E,$5F,$01,$F9,$DF,$9D
	.dc.b	$FF,$38,$EB,$FC,$9B,$E7,$BF,$92,$99,$F1,$E6,$1F,$9E,$AD,$EE,$87
	.dc.b	$5A,$73,$1F,$AC,$E4,$7C,$F7,$F3,$0F,$82,$FE,$63,$F3,$D2,$BF,$9B
	.dc.b	$C6,$5F,$E7,$31,$97,$F7,$91,$FC,$E3,$E7,$7F,$9C,$FC,$F4,$9F,$E7
	.dc.b	$F1,$97,$FA,$0C,$65,$FE,$09,$1F,$CE,$7E,$09,$E2,$7E,$CA,$5F,$3A
	.dc.b	$F9,$7F,$F4,$D7,$CF,$7F,$0A,$9F,$74,$E5,$F8,$11,$97,$8C,$FF,$A8
	.dc.b	$6F,$AF,$1F,$08,$C6,$1D,$F0,$E5,$6C,$39,$46,$1B,$5C,$2F,$95,$7C
	.dc.b	$FC,$3A,$BE,$03,$C7,$E5,$7F,$AC,$70,$E4,$DB,$E5,$3B,$3F,$D1,$98
	.dc.b	$F8,$E3,$DF,$3B,$9E,$67,$80,$DE,$17,$0F,$0E,$1D,$A7,$58,$FA,$1F
	.dc.b	$36,$3D,$5B,$7A,$C3,$B4,$7C,$16,$E1,$D3,$1F,$3E,$3D,$6B,$D1,$8F
	.dc.b	$57,$C6,$7A,$EB,$83,$E8,$37,$7D,$31,$F1,$47,$0B,$E8,$4C,$3B,$92
	.dc.b	$39,$46,$5F,$49,$DE,$79,$E1,$E3,$3C,$98,$78,$A7,$B5,$FC,$C7,$E0
	.dc.b	$1F,$26,$BC,$73,$BA,$E5,$87,$9F,$1D,$1C,$2F,$D6,$60,$F1,$39,$C7
	.dc.b	$CD,$1E,$03,$CE,$E2,$BC,$B8,$78,$0F,$4D,$FD,$28,$CD,$F5,$AE,$13
	.dc.b	$E3,$BE,$E7,$88,$91,$F9,$AC,$3E,$01,$EB,$9E,$87,$18,$D6,$1F,$AF
	.dc.b	$EC,$FE,$57,$5F,$9A,$6C,$67,$AF,$F2,$F1,$FC,$C3,$0F,$E1,$DF,$07
	.dc.b	$FC,$3E,$31,$08,$C4,$FF,$13,$8F,$F5,$A1,$FE,$DF,$70,$F2,$30,$62
	.dc.b	$3F,$3D,$C8,$F9,$D7,$37,$E8,$6B,$86,$7F,$B8,$AD,$4E,$67,$C4,$7E
	.dc.b	$B6,$32,$FE,$BC,$E3,$BF,$96,$33,$5F,$C0,$A7,$27,$F2,$B8,$CC,$FA
	.dc.b	$23,$AC,$F8,$6F,$CB,$1D,$4C,$CF,$58,$CC,$F5,$9C,$9D,$63,$33,$D7
	.dc.b	$0F,$83,$18,$9F,$88,$7A,$DC,$3F,$32,$BC,$58,$7C,$1A,$E2,$D9,$E0
	.dc.b	$D9,$DC,$79,$E7,$6D,$36,$58,$D9,$E9,$8D,$B6,$36,$D3,$77,$EB,$8C
	.dc.b	$B6,$33,$1E,$C6,$BE,$B8,$F9,$15,$E7,$6C,$78,$DB,$3D,$DA,$75,$2C
	.dc.b	$75,$6B,$D4,$FA,$13,$D5,$A7,$52,$CF,$56,$CF,$56,$9D,$4A,$75,$3F
	.dc.b	$B8,$DF,$D6,$7F,$16,$62,$BC,$0D,$3C,$05,$AF,$03,$6B,$FC,$F1,$F3
	.dc.b	$DF,$8E,$DF,$F1,$FE,$6B,$E5,$7F,$C0,$18,$FF,$AB,$AB,$4F,$F8,$0E
	.dc.b	$75,$86,$0F,$F0,$14,$D8,$CC,$F1,$3F,$D3,$AE,$27,$FC,$CB,$1C,$E9
	.dc.b	$F1,$5C,$5F,$D6,$E1,$E4,$73,$3C,$9C,$BC,$67,$86,$7C,$5B,$1F,$E7
	.dc.b	$8F,$9A,$7C,$DF,$9F,$74,$62,$7A,$79,$79,$8C,$67,$E6,$3F,$C4,$2E
	.dc.b	$33,$F2,$79,$7B,$C7,$D0,$FC,$3A,$FE,$1D,$E2,$5D,$3F,$C7,$79,$7F
	.dc.b	$AF,$B4,$7C,$F3,$E6,$75,$72,$B5,$2F,$9A,$C6,$3E,$3C,$38,$C7,$D1
	.dc.b	$A7,$CF,$7F,$B5,$5F,$F6,$59,$8E,$BD,$46,$6F,$D9,$39,$37,$19,$DB
	.dc.b	$F8,$8C,$66,$77,$39,$8F,$C2,$BF,$EB,$27,$8F,$AF,$DA,$FB,$B6,$3F
	.dc.b	$DB,$1E,$D3,$E0,$FE,$33,$63,$E2,$7C,$F7,$FB,$61,$DC,$E8,$FC,$76
	.dc.b	$89,$7C,$2E,$6B,$8B,$F8,$9E,$3F,$C5,$75,$59,$FC,$43,$75,$FB,$9D
	.dc.b	$3F,$9F,$8C,$6F,$F8,$F7,$F3,$9F,$31,$FE,$61,$F9,$BB,$6F,$98,$B1
	.dc.b	$86,$E3,$FE,$FA,$AF,$F3,$44,$62,$BF,$15,$18,$9D,$42,$B9,$46,$1B
	.dc.b	$1E,$6E,$CF,$CB,$CF,$E8,$7A,$BF,$4D,$E8,$FE,$9D,$91,$93,$F4,$D1
	.dc.b	$8A,$FC,$D7,$6F,$E8,$3E,$67,$F3,$C7,$EB,$FE,$7A,$DF,$E9,$E3,$F6
	.dc.b	$5F,$3D,$8B,$F1,$3F,$B5,$ED,$95,$FB,$3E,$BB,$5E,$79,$BC,$3F,$B9
	.dc.b	$9B,$5F,$BF,$9B,$5F,$EF,$4D,$7F,$3D,$3F,$BD,$9F,$E0,$5F,$3D,$FC
	.dc.b	$0F,$7B,$3F,$C0,$5B,$7F,$15,$B8,$FF,$07,$7F,$65,$F3,$AF,$53,$FB
	.dc.b	$5D,$7F,$81,$A3,$17,$FD,$C6,$11,$87,$FB,$A3,$5E,$13,$FC,$7E,$DE
	.dc.b	$BB,$E2,$7B,$CE,$63,$18,$7E,$0F,$E6,$5F,$27,$F8,$C5,$F2,$36,$30
	.dc.b	$D9,$E0,$DC,$BF,$13,$5C,$27,$A5,$70,$69,$C0,$FE,$4D,$D1,$FC,$AF
	.dc.b	$8F,$F9,$63,$63,$F9,$57,$BD,$FC,$BF,$E7,$7F,$98,$36,$3F,$97,$69
	.dc.b	$FC,$CF,$E7,$7F,$9A,$36,$3F,$99,$79,$1F,$CF,$FE,$77,$FC,$57,$18
	.dc.b	$3F,$9F,$B8,$D7,$F5,$ED,$E1,$FE,$9E,$BE,$6B,$95,$73,$63,$CB,$E6
	.dc.b	$AD,$F2,$7C,$1F,$33,$E1,$F9,$AB,$5E,$06,$D7,$81,$B8,$F8,$1B,$3F
	.dc.b	$D6,$F1,$46,$29,$E1,$A7,$35,$EC,$F9,$AF,$81,$C7,$F0,$5B,$19,$3E
	.dc.b	$0C,$65,$A6,$6F,$E1,$C6,$5B,$1F,$23,$97,$E6,$7C,$1F,$CD,$6C,$64
	.dc.b	$F9,$B1,$96,$99,$9F,$9F,$19,$6C,$7D,$4F,$99,$FB,$1F,$25,$F6,$5B
	.dc.b	$19,$3E,$CC,$65,$A6,$67,$ED,$C6,$5B,$1F,$80,$F2,$7F,$09,$E6,$7E
	.dc.b	$15,$B1,$93,$F0,$B1,$96,$99,$9F,$C3,$C6,$5B,$1F,$8C,$F9,$9F,$C8
	.dc.b	$7C,$CF,$E4,$5B,$19,$3F,$23,$19,$69,$F9,$4F,$98,$FF,$36,$FB,$4F
	.dc.b	$87,$FC,$17,$CC,$E1,$FC,$24,$DC,$7F,$07,$C4,$AC,$71,$FE,$16,$AC
	.dc.b	$E7,$1F,$F4,$DE,$1C,$9B,$87,$2C,$3A,$2F,$9C,$73,$FF,$88,$A3,$FD
	.dc.b	$41,$7F,$E9,$F9,$CF,$A7,$CA,$5F,$3A,$F8,$6E,$75,$A9,$E2,$0E,$D8
	.dc.b	$75,$DB,$E0,$46,$67,$84,$62,$3A,$D7,$4A,$E1,$8F,$8D,$B3,$C5,$AF
	.dc.b	$0C,$FC,$D8,$74,$AE,$19,$62,$F7,$97,$A6,$F1,$33,$0F,$49,$E8,$7E
	.dc.b	$04,$EE,$33,$87,$57,$C5,$3C,$72,$F5,$56,$DF,$65,$6E,$BA,$1C,$B0
	.dc.b	$F1,$E5,$EF,$C3,$C5,$3F,$EC,$CA,$F9,$8D,$AF,$AC,$D9,$FE,$95,$F0
	.dc.b	$6F,$93,$78,$D8,$6D,$7F,$AF,$E7,$53,$93,$FA,$96,$1F,$C4,$BE,$0E
	.dc.b	$F6,$9A,$4B,$FF,$64,$4E,$DF,$87,$3B,$6C,$FE,$1F,$E7,$5F,$17,$FE
	.dc.b	$E9,$DB,$E1,$DE,$DB,$5F,$62,$31,$C3,$FE,$EF,$BE,$38,$66,$A3,$D8
	.dc.b	$FA,$F1,$A7,$17,$D1,$B5,$CF,$0F,$13,$F9,$DC,$3B,$CF,$E4,$DA,$FE
	.dc.b	$95,$B7,$D6,$FA,$12,$B9,$61,$EF,$C3,$AE,$1C,$E3,$CF,$5B,$8F,$D2
	.dc.b	$5F,$3B,$FE,$85,$F2,$DC,$39,$BF,$A0,$79,$97,$9B,$FB,$F6,$1F,$BA
	.dc.b	$9F,$E8,$95,$93,$55,$27,$54,$9E,$5F,$39,$FB,$39,$FE,$C6,$DF,$0F
	.dc.b	$FB,$1B,$79,$FF,$AA,$7C,$E5,$AF,$E2,$33,$72,$FE,$A8,$DC,$4D,$AD
	.dc.b	$7E,$B3,$F8,$3B,$FF,$2F,$B7,$1F,$F5,$FB,$71,$FF,$5E,$2F,$89,$E3
	.dc.b	$C1,$2F,$FD,$73,$3B,$7F,$B2,$BB,$73,$1B,$7B,$BE,$29,$F1,$FC,$04
	.dc.b	$9E,$6F,$9B,$7C,$47,$9A,$2E,$3E,$69,$B8,$F9,$A6,$D7,$8A,$7D,$FF
	.dc.b	$37,$23,$CE,$FC,$2F,$82,$C4,$7C,$28,$B8,$7C,$29,$B8,$7C,$29,$B5
	.dc.b	$EE,$9F,$97,$F0,$72,$3E,$1B,$F3,$BE,$6F,$11,$F3,$A2,$E1,$F3,$A6
	.dc.b	$E1,$F3,$A6,$D7,$CA,$9F,$AF,$F1,$92,$3E,$7B,$F6,$BB,$31,$1F,$6A
	.dc.b	$2E,$1F,$6A,$6E,$1F,$6A,$6D,$7D,$69,$FC,$1F,$CD,$7E,$13,$E6,$A4
	.dc.b	$FD,$C8,$CB,$48,$7E,$23,$E6,$7F,$16,$47,$F0,$6F,$E3,$BA,$FF,$1F
	.dc.b	$F2,$DF,$90,$F9,$59,$5F,$88,$8F,$C9,$F8,$92,$3F,$16,$FE,$5B,$E5
	.dc.b	$7F,$2F,$F2,$92,$7F,$1E,$FE,$6B,$E5,$7F,$45,$0F,$9D,$6F,$86,$C7
	.dc.b	$F5,$FE,$1F,$F3,$FC,$62,$3F,$DF,$47,$1E,$0E,$15,$23,$87,$0F,$ED
	.dc.b	$6B,$85,$F2,$BD,$57,$29,$CD,$6E,$FC,$55,$FB,$0F,$94,$FD,$B7,$9D
	.dc.b	$FB,$FC,$3F,$D9,$ED,$C7,$F6,$33,$B9,$F0,$9F,$E5,$09,$D6,$1F,$DE
	.dc.b	$7C,$04,$62,$7F,$63,$5F,$13,$E7,$5E,$97,$FA,$6E,$51,$87,$FD,$3C
	.dc.b	$D7,$C8,$6A,$38,$46,$71,$FF,$53,$B7,$1F,$E3,$0D,$AF,$F0,$66,$9F
	.dc.b	$C0,$F2,$F2,$F9,$7F,$03,$E4,$E5,$7B,$27,$2D,$D3,$FB,$4E,$4C,$BD
	.dc.b	$D0,$BE,$35,$F2,$30,$F9,$66,$2F,$84,$7C,$79,$CB,$F2,$8F,$83,$E2
	.dc.b	$FC,$52,$39,$FB,$47,$AB,$88,$FC,$2E,$BF,$CD,$DE,$4D,$7F,$13,$3C
	.dc.b	$1F,$57,$CB,$7D,$EF,$96,$91,$82,$1F,$80,$72,$FE,$0A,$3F,$37,$F3
	.dc.b	$AF,$8A,$A4,$9F,$54,$3E,$33,$11,$F9,$A8,$E0,$EA,$79,$CE,$B5,$FF
	.dc.b	$6F,$D6,$AB,$85,$69,$F8,$DE,$A3,$AB,$F1,$0D,$8F,$F7,$23,$5F,$C3
	.dc.b	$B5,$FC,$E3,$9B,$FF,$76,$B7,$8B,$FA,$F3,$67,$FC,$A5,$ED,$FF,$95
	.dc.b	$DF,$AD,$F3,$AF,$8E,$FD,$EB,$63,$31,$FC,$03,$E7,$7F,$43,$18,$21
	.dc.b	$FB,$C9,$E0,$FE,$FA,$36,$FF,$00,$F8,$BF,$03,$7C,$0F,$E1,$71,$8C
	.dc.b	$FF,$17,$3F,$EF,$CD,$33,$52,$3F,$19,$5B,$D3,$FB,$D3,$ED,$83,$97
	.dc.b	$FC,$E0,$D7,$33,$E0,$AF,$E3,$7C,$2F,$8B,$A4,$AF,$CD,$7D,$FF,$F6
	.dc.b	$F5,$3B,$7F,$3C,$D7,$F8,$EB,$72,$F5,$CF,$1F,$C2,$5F,$09,$FC,$A9
	.dc.b	$B7,$FA,$76,$BF,$C9,$BF,$32,$F4,$46,$2F,$E2,$B7,$5F,$F0,$D3,$6B
	.dc.b	$FD,$1D,$5F,$07,$BB,$F8,$BF,$CC,$FF,$22,$24,$FF,$01,$F9,$9A,$FF
	.dc.b	$24,$8B,$8F,$EF,$2B,$F8,$FD,$7F,$2A,$F9,$89,$7F,$C2,$5F,$E5,$BF
	.dc.b	$31,$88,$FE,$5B,$16,$F9,$78,$4A,$FE,$93,$F0,$5F,$D2,$27,$FD,$49
	.dc.b	$F3,$AC,$7F,$C6,$9F,$3C,$F5,$3F,$C6,$D3,$9C,$FF,$D5,$53,$97,$8D
	.dc.b	$7F,$8E,$F0,$E9,$0A,$E0,$A3,$31,$8A,$E5,$3B,$AF,$0D,$F8,$1E,$9E
	.dc.b	$5B,$1E,$F5,$FE,$BB,$C7,$9D,$70,$8C,$8C,$3A,$E3,$E2,$AD,$CF,$1C
	.dc.b	$3A,$3E,$13,$DB,$8F,$85,$A7,$49,$E2,$FA,$E7,$8D,$7A,$EB,$84,$F8
	.dc.b	$5B,$5C,$0E,$37,$E4,$C3,$D1,$5E,$13,$A6,$3D,$F1,$EB,$3A,$C7,$B5
	.dc.b	$66,$A3,$C6,$FA,$3E,$83,$A6,$1E,$1C,$39,$C6,$6F,$9B,$A8,$CB,$E2
	.dc.b	$3A,$63,$EC,$BE,$B1,$9A,$F9,$95,$C2,$72,$6A,$FC,$18,$7D,$48,$CE
	.dc.b	$1F,$12,$BB,$D7,$C4,$8E,$C6,$6B,$DB,$7E,$62,$63,$F8,$48,$CB,$F1
	.dc.b	$49,$8F,$DD,$8C,$BE,$F8,$CB,$E9,$79,$3E,$E3,$CD,$5E,$17,$E2,$4E
	.dc.b	$18,$60,$95,$E4,$7B,$1E,$13,$35,$E0,$C7,$E2,$63,$F0,$0F,$C8,$FB
	.dc.b	$5F,$E8,$9A,$FA,$0D,$CB,$F5,$6D,$9D,$FB,$5F,$E1,$8F,$81,$5F,$FB
	.dc.b	$96,$30,$D3,$FC,$61,$1A,$AE,$D1,$C2,$31,$3C,$32,$FD,$BF,$CD,$7F
	.dc.b	$84,$79,$7F,$57,$7B,$3F,$2D,$19,$7F,$30,$66,$78,$F9,$5F,$E1,$47
	.dc.b	$F8,$0C,$67,$5F,$F6,$05,$71,$9C,$CB,$1F,$DC,$CE,$6B,$C8,$70,$BE
	.dc.b	$8E,$2B,$E0,$CE,$DF,$79,$E6,$9D,$B8,$BF,$2E,$1E,$C8,$F0,$1B,$71
	.dc.b	$87,$8A,$B8,$CF,$FB,$5E,$31,$3A,$8E,$07,$C2,$39,$D7,$67,$13,$C1
	.dc.b	$A7,$08,$F8,$31,$96,$C6,$70,$F7,$2C,$3D,$CF,$8C,$D1,$ED,$8F,$8D
	.dc.b	$7E,$5A,$F1,$1D,$0B,$1D,$2B,$9D,$7D,$53,$E5,$E1,$F0,$1C,$1E,$17
	.dc.b	$0F,$BA,$7B,$BE,$CA,$EA,$F8,$CE,$77,$E3,$9D,$CF,$8D,$A7,$8E,$BA
	.dc.b	$E1,$F0,$E3,$15,$C6,$F9,$9D,$E3,$8B,$DC,$F3,$46,$DA,$6C,$E3,$7E
	.dc.b	$16,$CF,$F9,$C3,$E6,$BF,$CD,$7B,$FF,$68,$6D,$FF,$9D,$3C,$1B,$7F
	.dc.b	$E9,$59,$BC,$BF,$E9,$46,$F2,$1B,$58,$73,$E2,$4C,$3F,$D0,$5E,$35
	.dc.b	$C7,$FD,$4B,$37,$2F,$EF,$33,$6B,$F7,$F4,$FC,$A1,$F2,$3E,$48,$C3
	.dc.b	$E6,$F3,$9F,$92,$BE,$17,$57,$C3,$0F,$F4,$33,$70,$FF,$62,$B7,$2F
	.dc.b	$F4,$6E,$1C,$AF,$8E,$3D,$FF,$4D,$FF,$59,$E3,$FE,$8F,$6E,$3F,$DD
	.dc.b	$E3,$F8,$AD,$BE,$85,$F3,$5C,$78,$65,$71,$7D,$13,$A8,$C1,$FE,$C4
	.dc.b	$1F,$42,$F8,$0F,$A7,$18,$6C,$3D,$FB,$12,$79,$4F,$16,$CE,$72,$F4
	.dc.b	$1F,$EB,$39,$D3,$89,$EC,$70,$39,$D7,$D1,$F8,$57,$C4,$53,$E9,$7D
	.dc.b	$0B,$E4,$F1,$7F,$EB,$5A,$DB,$6B,$73,$AC,$3F,$DE,$FF,$63,$EB,$AF
	.dc.b	$F5,$E3,$63,$CF,$F8,$23,$F0,$47,$CC,$7E,$0E,$B5,$CD,$FE,$7C,$78
	.dc.b	$5F,$FB,$CD,$AF,$E3,$1A,$FF,$4C,$6F,$D2,$7F,$4E,$F4,$BF,$42,$DC
	.dc.b	$BF,$DE,$EF,$F5,$2F,$9F,$CD,$FF,$9E,$E7,$2D,$3E,$6C,$F1,$68,$9E
	.dc.b	$35,$2B,$8B,$D2,$31,$3F,$5B,$C0,$72,$BE,$8F,$F6,$3F,$47,$FB,$26
	.dc.b	$E3,$FE,$EF,$6D,$F1,$AE,$13,$FE,$F3,$AF,$D5,$7D,$1F,$FB,$3E,$3F
	.dc.b	$DD,$4D,$AF,$C2,$F5,$39,$5F,$27,$C3,$C1,$98,$70,$7F,$03,$18,$BF
	.dc.b	$07,$67,$F0,$6D,$FF,$83,$B6,$FF,$BD,$F1,$7F,$9E,$23,$15,$C2,$72
	.dc.b	$7F,$80,$37,$F8,$37,$CA,$33,$29,$CC,$61,$F8,$F5,$D2,$FF,$D7,$FA
	.dc.b	$3E,$87,$E0,$9F,$C7,$BC,$A7,$D4,$BA,$F3,$97,$C4,$AF,$F6,$3E,$7F
	.dc.b	$C5,$30,$E5,$3F,$C5,$A3,$13,$A7,$FC,$2F,$3A,$75,$9F,$DA,$E4,$FE
	.dc.b	$53,$F4,$9F,$CA,$9B,$87,$A7,$93,$F0,$CD,$9D,$34,$FC,$9F,$CD,$BE
	.dc.b	$07,$F2,$ED,$8C,$FC,$3E,$26,$47,$07,$71,$8C,$BF,$2F,$5C,$5B,$3A
	.dc.b	$AE,$07,$97,$4F,$E8,$27,$F4,$2F,$A4,$B1,$E9,$FA,$4F,$E8,$A7,$F4
	.dc.b	$8F,$C8,$51,$83,$EB,$3F,$CC,$5F,$BC,$BE,$97,$87,$93,$C5,$E5,$F0
	.dc.b	$F2,$B8,$FC,$F5,$D3,$F4,$D8,$3E,$97,$C0,$67,$A3,$FD,$9D,$19,$BF
	.dc.b	$0C,$F9,$7E,$0B,$71,$E4,$7D,$1E,$24,$8F,$3F,$05,$C7,$BC,$DC,$7A
	.dc.b	$CD,$AF,$5C,$DA,$FC,$B6,$3F,$36,$FC,$73,$F1,$78,$A5,$FC,$4F,$90
	.dc.b	$B8,$7B,$26,$DF,$AA,$6D,$7C,$89,$B5,$DA,$6F,$3B,$17,$D2,$FD,$33
	.dc.b	$EA,$79,$37,$1F,$A7,$17,$0F,$97,$37,$2F,$93,$36,$BE,$BC,$DA,$F6
	.dc.b	$F4,$7D,$F3,$F0,$1F,$4B,$6B,$EF,$C5,$9F,$AF,$37,$0F,$AB,$36,$BF
	.dc.b	$07,$36,$BE,$4F,$D2,$FE,$2C,$FC,$67,$D2,$DA,$FC,$5C,$59,$FC,$1C
	.dc.b	$DC,$3F,$03,$36,$BF,$1F,$36,$BE,$AF,$C9,$FE,$5C,$FC,$C7,$A9,$6B
	.dc.b	$F2,$F1,$67,$F1,$F3,$70,$FC,$6F,$9F,$F9,$E9,$FF,$3F,$7D,$07,$E7
	.dc.b	$3C,$7C,$7B,$A8,$FC,$FD,$61,$D6,$5F,$E9,$77,$2D,$BF,$C4,$DE,$DF
	.dc.b	$C5,$D5,$E2,$FB,$D3,$99,$7D,$23,$AD,$E6,$73,$9C,$C6,$1F,$39,$FE
	.dc.b	$AE,$FA,$57,$C7,$F3,$BE,$57,$C5,$B5,$C0,$E7,$1D,$6B,$8E,$7B,$C3
	.dc.b	$B6,$BF,$06,$FF,$D7,$78,$72,$8C,$47,$23,$A5,$F2,$7B,$4E,$F5,$FF
	.dc.b	$22,$B6,$39,$DF,$69,$FF,$24,$D7,$1B,$EB,$7B,$BE,$D1,$8B,$E1,$1F
	.dc.b	$07,$2F,$81,$18,$8D,$E1,$E0,$8C,$61,$D6,$31,$97,$BC,$95,$EA,$BE
	.dc.b	$71,$88,$FE,$91,$8F,$CA,$6E,$1F,$4D,$B3,$FD,$35,$FF,$70,$FC,$2B
	.dc.b	$E6,$5E,$36,$65,$39,$8C,$36,$39,$35,$F0,$F8,$2F,$85,$E5,$92,$31
	.dc.b	$3E,$18,$C4,$EA,$73,$3C,$A3,$F0,$91,$C3,$6F,$9E,$DD,$7F,$CF,$7F
	.dc.b	$4A,$F9,$FF,$F6,$74,$7D,$A9,$D6,$1F,$84,$6B,$F9,$A6,$D7,$18,$EF
	.dc.b	$84,$7F,$0B,$3A,$BE,$B1,$CE,$75,$87,$91,$FD,$15,$F3,$9E,$31,$AC
	.dc.b	$3C,$8E,$AF,$9C,$F8,$E3,$4F,$53,$F4,$55,$CE,$FB,$47,$42,$3F,$9E
	.dc.b	$8D,$5F,$F8,$77,$E9,$5D,$1E,$8E,$54,$7F,$A6,$F0,$FE,$05,$18,$E8
	.dc.b	$D4,$F0,$9C,$B5,$FD,$DC,$ED,$CC,$F0,$AE,$3C,$FF,$EA,$EE,$8F,$E3
	.dc.b	$17,$E0,$8C,$73,$F9,$5B,$87,$F8,$BF,$1F,$F2,$1B,$FE,$2F,$7F,$D0
	.dc.b	$6D,$8D,$D7,$38,$D7,$46,$3E,$97,$F1,$38,$F8,$1B,$8F,$F8,$FD,$B5
	.dc.b	$FE,$3D,$5F,$13,$C7,$82,$4F,$F8,$FE,$77,$89,$DC,$66,$76,$FF,$9A
	.dc.b	$E3,$66,$9C,$7A,$1E,$52,$3D,$77,$AE,$8F,$04,$7A,$E3,$39,$7A,$E2
	.dc.b	$E5,$EB,$9F,$81,$F4,$B2,$FD,$0F,$C3,$FA,$5C,$47,$96,$2E,$1F,$16
	.dc.b	$6D,$FC,$59,$B5,$F1,$67,$E6,$7C,$74,$8F,$88,$FC,$FF,$A5,$C4,$7C
	.dc.b	$18,$B8,$7D,$19,$B8,$7D,$19,$B5,$F4,$67,$EC,$7C,$74,$8F,$A0,$FD
	.dc.b	$BE,$CC,$47,$CD,$8B,$87,$DD,$9B,$87,$DD,$9B,$5F,$76,$7F,$09,$F4
	.dc.b	$BF,$85,$F8,$39,$3F,$76,$32,$D2,$1F,$6F,$E1,$7F,$18,$47,$F0,$8F
	.dc.b	$E3,$FA,$FF,$21,$F4,$9F,$91,$FA,$59,$5F,$89,$8F,$CA,$78,$92,$3F
	.dc.b	$18,$FE,$5F,$E6,$BF,$31,$F4,$B2,$7F,$20,$FE,$6F,$E9,$7F,$4B,$0F
	.dc.b	$A4,$6F,$86,$C7,$FA,$60,$D4,$67,$7D,$32,$35,$3C,$1F,$F7,$B9,$D6
	.dc.b	$74,$E2,$35,$C3,$FE,$11,$8E,$A7,$29,$FD,$87,$D2,$FE,$D3,$CE,$FF
	.dc.b	$4C,$57,$F0,$06,$F1,$FE,$C6,$77,$7F,$E9,$33,$81,$83,$C7,$F4,$AF
	.dc.b	$4F,$DD,$D7,$27,$DD,$39,$6C,$FF,$57,$51,$97,$FA,$F3,$5F,$F5,$63
	.dc.b	$5F,$F6,$16,$1F,$D9,$71,$F4,$7C,$BF,$AC,$F6,$79,$92,$BF,$D5,$51
	.dc.b	$97,$FD,$59,$19,$7F,$DC,$91,$97,$FD,$CD,$1F,$6A,$FF,$95,$57,$CB
	.dc.b	$FA,$5F,$F3,$4D,$7D,$96,$BF,$0B,$C6,$E0,$47,$E0,$BF,$84,$87,$AD
	.dc.b	$F4,$7E,$97,$E9,$7D,$0C,$BF,$F6,$04,$65,$FF,$61,$46,$5F,$F7,$D4
	.dc.b	$7E,$3B,$E9,$5F,$2F,$49,$3F,$8D,$87,$C5,$E2,$3F,$13,$1C,$1D,$4F
	.dc.b	$39,$D6,$7F,$89,$AD,$57,$0A,$D3,$F9,$9F,$51,$D5,$FA,$0F,$75,$D7
	.dc.b	$4F,$D1,$CE,$AB,$85,$7E,$06,$31,$1F,$E1,$D6,$BF,$EE,$B6,$A2,$7D
	.dc.b	$2B,$E6,$BF,$5F,$F4,$AF,$6E,$9F,$B2,$9D,$57,$EB,$DA,$FE,$A9,$BA
	.dc.b	$FE,$9D,$BD,$7D,$E5,$6E,$8F,$DE,$FD,$2F,$F9,$92,$7F,$74,$D7,$F6
	.dc.b	$AD,$C3,$F6,$4F,$B9,$BE,$0F,$22,$4F,$2D,$6B,$CA,$2C,$F2,$9F,$E1
	.dc.b	$BC,$8F,$8F,$FE,$1F,$F4,$AE,$EA,$6A,$26,$3F,$C3,$EB,$83,$63,$8F
	.dc.b	$A1,$FC,$63,$E9,$5F,$0B,$FC,$6B,$E9,$7E,$DD,$FF,$17,$6B,$FC,$4A
	.dc.b	$76,$FF,$1B,$6B,$99,$E8,$73,$D7,$F8,$87,$D2,$C9,$FB,$9B,$7F,$11
	.dc.b	$8F,$E2,$9F,$43,$99,$4E,$5D,$F1,$FF,$02,$AD,$C6,$07,$AD,$F5,$FE
	.dc.b	$97,$F8,$F1,$23,$51,$F6,$FE,$97,$11,$DA,$FF,$86,$D6,$30,$FE,$49
	.dc.b	$5F,$CF,$3E,$0B,$F9,$DC,$7E,$4F,$E9,$19,$FE,$1F,$E9,$5E,$BF,$E2
	.dc.b	$67,$39,$FE,$66,$72,$F1,$AF,$C7,$61,$D2,$15,$C1,$FF,$3D,$F9,$FF
	.dc.b	$09,$CD,$77,$C3,$B3,$63,$A1,$FE,$86,$9E,$15,$CE,$FB,$3D,$70,$F1
	.dc.b	$D7,$43,$CD,$87,$47,$96,$3D,$1F,$1B,$67,$8D,$73,$AE,$15,$A3,$A1
	.dc.b	$63,$4F,$18,$DD,$F9,$31,$F8,$78,$FA,$1F,$65,$72,$C3,$BC,$6B,$1F
	.dc.b	$25,$F1,$AF,$49,$D4,$CD,$F3,$7D,$67,$3C,$3C,$F8,$74,$8C,$E3,$F1
	.dc.b	$EB,$84,$27,$4D,$AF,$4C,$EA,$FC,$53,$93,$C4,$7C,$27,$26,$AB,$99
	.dc.b	$D1,$CC,$F8,$23,$35,$DB,$0F,$59,$D3,$0F,$97,$87,$5B,$F1,$1C,$6B
	.dc.b	$DA,$71,$8D,$57,$1C,$7E,$6D,$FB,$E3,$AC,$FC,$53,$47,$C4,$9F,$A0
	.dc.b	$E4,$93,$C5,$EA,$62,$3A,$9C,$23,$47,$A0,$8E,$49,$1F,$68,$E4,$78
	.dc.b	$4F,$B3,$C7,$FE,$77,$CB,$F5,$0D,$C7,$F4,$CD,$C3,$FA,$43,$FE,$7D
	.dc.b	$AF,$E9,$5F,$05,$8F,$77,$FD,$D8,$D8,$FE,$80,$1C,$3F,$D0,$DA,$F2
	.dc.b	$3F,$3B,$F2,$EF,$39,$C3,$FA,$79,$D6,$BF,$BB,$AD,$4E,$67,$B1,$CA
	.dc.b	$B8,$BF,$A8,$38,$EB,$FD,$4E,$B8,$C6,$4E,$37,$D6,$B8,$BF,$C2,$63
	.dc.b	$89,$FD,$A6,$78,$C6,$63,$8D,$75,$AE,$2F,$C4,$8F,$25,$F9,$F0,$FE
	.dc.b	$05,$87,$2A,$D6,$1E,$1A,$D5,$F0,$3C,$D7,$C3,$0E,$B1,$F3,$A3,$13
	.dc.b	$C2,$B9,$4F,$03,$93,$E7,$9D,$C6,$63,$66,$5D,$DF,$90,$F6,$57,$56
	.dc.b	$C7,$16,$CF,$13,$9B,$C6,$3E,$45,$FA,$1F,$C1,$1D,$8F,$51,$4C,$CF
	.dc.b	$27,$2D,$73,$3E,$03,$B2,$C3,$E6,$3C,$23,$4D,$3C,$77,$EB,$69,$EB
	.dc.b	$2D,$7A,$E7,$8C,$78,$0C,$BE,$0A,$EF,$19,$3C,$57,$EF,$7F,$12,$62
	.dc.b	$FC,$0D,$3C,$05,$BF,$EB,$FC,$5F,$E2,$7F,$A5,$FC,$DE,$BF,$E8,$D6
	.dc.b	$DF,$F9,$EB,$C3,$B8,$7F,$78,$F0,$15,$8C,$39,$5F,$F7,$89,$CB,$6F
	.dc.b	$51,$8B,$FE,$F4,$73,$F1,$F2,$6B,$C9,$B7,$CE,$2C,$F0,$C3,$FC,$7D
	.dc.b	$58,$CB,$FD,$55,$F4,$8F,$92,$F3,$DA,$31,$8F,$F8,$47,$D1,$7D,$2B
	.dc.b	$E1,$B7,$5A,$BF,$F5,$CB,$71,$FD,$FD,$BF,$22,$FD,$EC,$78,$32,$F8
	.dc.b	$B3,$9E,$2F,$0F,$D3,$7F,$94,$F4,$FF,$65,$37,$2F,$83,$4F,$A6,$7C
	.dc.b	$F7,$0E,$29,$87,$9E,$7F,$C9,$4F,$28,$EF,$18,$BF,$97,$2F,$A6,$7C
	.dc.b	$1F,$D8,$8C,$36,$1F,$05,$62,$4F,$F8,$D4,$D9,$67,$6D,$8D,$D7,$82
	.dc.b	$7B,$CE,$EB,$3B,$7F,$B7,$7C,$B7,$C5,$7D,$46,$D7,$D3,$FA,$07,$CA
	.dc.b	$66,$A4,$E7,$5B,$9F,$C8,$BF,$95,$8F,$FB,$A7,$E9,$3C,$B3,$FD,$10
	.dc.b	$38,$7F,$A3,$B5,$FF,$44,$35,$FF,$4C,$3C,$A3,$F9,$EF,$73,$ED,$5F
	.dc.b	$17,$F8,$CA,$D6,$5F,$CD,$E7,$2F,$F9,$77,$E9,$7F,$D4,$5E,$8F,$EA
	.dc.b	$1B,$19,$C9,$F4,$39,$7E,$D6,$7F,$D3,$2B,$71,$FA,$66,$C6,$63,$8D
	.dc.b	$73,$BE,$44,$7B,$5D,$9F,$D4,$6A,$FA,$5F,$D9,$FD,$2F,$ED,$1B,$7F
	.dc.b	$B9,$6D,$70,$6E,$3F,$D8,$21,$F4,$9F,$ED,$8A,$FF,$3E,$46,$3C,$1F
	.dc.b	$D1,$F0,$BE,$4F,$F6,$77,$BA,$FE,$02,$6E,$FF,$69,$F8,$43,$CC,$DE
	.dc.b	$3F,$AC,$8C,$4F,$F0,$DE,$CF,$E1,$7C,$1F,$C3,$1B,$19,$3F,$89,$46
	.dc.b	$5A,$67,$0F,$87,$19,$6C,$7A,$B1,$7D,$27,$BE,$7F,$D9,$D1,$89,$FC
	.dc.b	$74,$7C,$29,$D3,$FE,$E1,$9D,$3F,$93,$6E,$BF,$CB,$BD,$CF,$E2,$17
	.dc.b	$DE,$FF,$96,$CA,$73,$7F,$63,$8B,$F9,$4F,$D2,$FF,$2A,$6E,$1F,$CB
	.dc.b	$DB,$3B,$8F,$E6,$53,$F9,$1F,$9B,$7C,$25,$31,$F0,$F6,$BF,$31,$3A
	.dc.b	$6C,$E9,$A4,$BF,$CA,$4E,$DB,$5C,$DF,$C7,$E7,$FD,$00,$FE,$83,$F4
	.dc.b	$B6,$FF,$A3,$7D,$2F,$F4,$43,$FA,$57,$D1,$7F,$8D,$23,$27,$FB,$09
	.dc.b	$FE,$64,$AF,$FD,$93,$0F,$A5,$E3,$E5,$72,$79,$FC,$44,$BE,$5F,$3D
	.dc.b	$70,$FE,$95,$83,$E9,$7C,$26,$78,$FF,$C7,$BE,$92,$75,$87,$8E,$BC
	.dc.b	$FF,$05,$C2,$3C,$CF,$AB,$E9,$64,$7A,$78,$EE,$5E,$09,$B9,$76,$C7
	.dc.b	$E5,$36,$33,$7E,$FC,$BC,$51,$F1,$B0,$CE,$1F,$0B,$8B,$E4,$9F,$2B
	.dc.b	$E4,$AE,$1F,$26,$2C,$FC,$09,$B8,$F8,$66,$D7,$7C,$95,$8B,$F9,$FF
	.dc.b	$4B,$F5,$8F,$AF,$E5,$5B,$FA,$D1,$67,$E6,$CD,$C7,$DD,$36,$BB,$4D
	.dc.b	$AF,$B7,$E2,$7E,$08,$FC,$1F,$D2,$DA,$FC,$14,$59,$FB,$33,$70,$F9
	.dc.b	$53,$6B,$D5,$36,$BF,$0F,$F1,$DF,$8E,$3F,$1F,$F4,$B6,$BF,$1D,$16
	.dc.b	$7F,$0B,$37,$0F,$AD,$36,$BE,$34,$DA,$FC,$9F,$CA,$7E,$68,$FC,$DF
	.dc.b	$B3,$6B,$F3,$51,$67,$F2,$33,$70,$FC,$17,$9F,$FE,$8C,$F8,$07,$D0
	.dc.b	$7D,$3F,$27,$0F,$D5,$9B,$87,$E8,$A6,$F3,$7E,$5E,$6F,$1F,$F3,$1D
	.dc.b	$3F,$D5,$78,$71,$6E,$1C,$5A,$BE,$91,$D6,$F6,$33,$18,$77,$F4,$AF
	.dc.b	$8C,$F1,$DF,$1B,$E0,$FF,$AD,$71,$FD,$C3,$9B,$E7,$3A,$DB,$DF,$3F
	.dc.b	$BD,$C3,$A6,$5A,$B9,$3C,$CF,$E0,$59,$78,$DC,$0A,$E2,$E2,$35,$7D
	.dc.b	$A3,$C4,$67,$0F,$16,$3E,$6B,$E3,$5B,$BE,$33,$EA,$6D,$F8,$AF,$8B
	.dc.b	$E9,$CB,$E1,$CE,$A3,$BD,$F3,$AD,$C7,$9E,$B7,$18,$AE,$07,$1C,$3C
	.dc.b	$78,$78,$67,$47,$CE,$BF,$F7,$1D,$7C,$36,$C6,$5E,$18,$FF,$58,$F8
	.dc.b	$77,$CD,$3C,$6C,$36,$BF,$C5,$53,$A9,$C9,$FD,$83,$07,$83,$FD,$DE
	.dc.b	$BF,$C6,$11,$8F,$3F,$53,$99,$F2,$B5,$F9,$4F,$1D,$BE,$73,$74,$CD
	.dc.b	$7E,$53,$E9,$5F,$41,$FE,$FD,$9C,$E1,$F7,$6F,$83,$F8,$46,$AA,$E2
	.dc.b	$FE,$6A,$33,$8F,$9E,$39,$61,$DF,$0E,$D1,$CE,$38,$11,$FC,$7D,$F4
	.dc.b	$78,$93,$1F,$5C,$F3,$8E,$11,$87,$58,$FE,$92,$B9,$4E,$49,$1E,$E7
	.dc.b	$F3,$11,$A8,$FE,$C5,$C3,$FE,$7A,$F8,$8F,$3B,$6F,$F7,$AD,$7F,$51
	.dc.b	$CA,$FC,$E3,$EA,$63,$FB,$0A,$D3,$7D,$D6,$1D,$1B,$87,$F0,$8C,$39
	.dc.b	$67,$F1,$A1,$F9,$E7,$E3,$1F,$F2,$1B,$5F,$F1,$FB,$7A,$BF,$B6,$E6
	.dc.b	$F8,$BF,$E0,$B1,$FE,$4D,$9D,$69,$FC,$21,$B7,$FE,$01,$6E,$3E,$BB
	.dc.b	$E3,$5F,$C3,$BF,$72,$FD,$26,$1D,$BE,$03,$38,$B0,$F8,$51,$88,$57
	.dc.b	$0B,$E5,$CF,$98,$C4,$FF,$84,$5B,$7F,$2F,$83,$55,$2B,$FC,$C7,$5A
	.dc.b	$D3,$F8,$86,$5F,$E8,$CF,$D4,$9E,$B7,$93,$2F,$A9,$F0,$16,$FF,$D1
	.dc.b	$F8,$FF,$87,$F0,$E1,$7E,$AF,$81,$63,$CA,$B5,$AF,$F8,$96,$73,$5B
	.dc.b	$AD,$57,$FB,$A7,$07,$D3,$E6,$A4,$F2,$17,$C9,$C6,$7F,$89,$AE,$9E
	.dc.b	$A7,$9F,$E4,$25,$71,$BF,$2C,$EA,$31,$92,$31,$1E,$3F,$A7,$BA,$FC
	.dc.b	$59,$BA,$FF,$88,$75,$57,$C1,$FC,$AB,$FA,$AE,$1E,$49,$7C,$97,$C9
	.dc.b	$35,$E1,$49,$F2,$57,$05,$97,$FA,$FA,$30,$FE,$53,$C2,$E1,$3A,$CB
	.dc.b	$E6,$56,$9B,$7F,$5E,$BC,$9E,$8F,$D9,$FA,$89,$87,$82,$32,$D2,$1F
	.dc.b	$4F,$E2,$72,$F7,$3E,$8E,$29,$CE,$5F,$A8,$AF,$C3,$76,$CB,$F8,$4F
	.dc.b	$C5,$F3,$A4,$7C,$06,$BF,$81,$F0,$BF,$1C,$62,$FE,$14,$E2,$7E,$67
	.dc.b	$D3,$FE,$4F,$EA,$24,$60,$FB,$DF,$71,$FC,$F6,$FF,$DC,$ED,$E0,$FC
	.dc.b	$93,$67,$F6,$1C,$5F,$E0,$77,$EC,$7C,$FF,$E7,$FC,$C9,$18,$3F,$D9
	.dc.b	$7F,$4D,$F3,$DF,$F4,$07,$D4,$34,$FD,$C3,$63,$7C,$BF,$BB,$AE,$10
	.dc.b	$69,$89,$70,$6E,$A5,$72,$3F,$DB,$73,$A7,$F1,$DF,$49,$FB,$52,$3F
	.dc.b	$91,$FA,$9A,$FE,$B6,$2E,$FF,$C7,$A7,$F8,$8F,$86,$FD,$7D,$18,$6D
	.dc.b	$FE,$C6,$75,$39,$FA,$AD,$32,$75,$5C,$1F,$C4,$57,$5A,$F8,$5E,$EF
	.dc.b	$FB,$5A,$7E,$13,$75,$FF,$3E,$C7,$FB,$69,$AF,$F8,$D8,$FF,$57,$38
	.dc.b	$C7,$FE,$00,$37,$8F,$FC,$76,$D9,$FA,$BE,$5F,$81,$FB,$37,$5F,$F7
	.dc.b	$FE,$BF,$83,$F0,$5C,$BF,$B8,$9F,$C5,$F7,$7F,$1D,$8C,$47,$F1,$7D
	.dc.b	$31,$32,$36,$7F,$01,$8F,$E5,$5F,$AE,$7A,$75,$FF,$5C,$B7,$5F,$F5
	.dc.b	$53,$6B,$FA,$13,$FC,$E8,$C7,$8A,$69,$37,$FE,$8D,$C7,$FC,$EA,$FF
	.dc.b	$CB,$7F,$52,$F8,$6F,$E9,$4D,$C7,$F9,$51,$FE,$6A,$F2,$F8,$91,$CD
	.dc.b	$F3,$7F,$DB,$92,$F8,$7D,$47,$F0,$33,$91,$C2,$BB,$C7,$C0,$F9,$89
	.dc.b	$3F,$CD,$B3,$47,$C4,$FA,$AC,$CF,$F0,$DD,$7D,$D3,$8D,$BE,$3D,$7F
	.dc.b	$63,$FA,$A3,$49,$7E,$CF,$85,$FE,$AA,$F9,$67,$6F,$C0,$6B,$99,$FE
	.dc.b	$DD,$3C,$29,$BF,$FA,$13,$F0,$5F,$F7,$D5,$FC,$8F,$C1,$9F,$2B,$FD
	.dc.b	$D5,$B1,$F3,$27,$6F,$C5,$6D,$F4,$80,$8C,$DF,$7F,$8C,$E9,$F4,$1F
	.dc.b	$E0,$19,$CD,$70,$6B,$C6,$76,$FD,$CF,$AA,$F9,$AD,$DB,$FC,$15,$18
	.dc.b	$8F,$B1,$F7,$3F,$0B,$2F,$24,$27,$51,$86,$E5,$FE,$A6,$6C,$FA,$AB
	.dc.b	$94,$FF,$AA,$BF,$2A,$FC,$D4,$9A,$63,$F4,$AF,$E6,$11,$93,$FC,$3D
	.dc.b	$1A,$7F,$C4,$11,$FB,$3E,$CF,$E5,$51,$FD,$47,$B7,$F6,$BE,$37,$EC
	.dc.b	$CF,$D6,$E5,$67,$FA,$0C,$57,$F1,$9C,$DE,$D7,$F2,$5F,$57,$C1,$FC
	.dc.b	$A5,$7E,$DA,$24,$60,$87,$27,$FC,$69,$F0,$9F,$DF,$36,$FB,$7F,$51
	.dc.b	$FE,$52,$A7,$D4,$BD,$9F,$F6,$6C,$E6,$31,$3C,$3E,$07,$FD,$A3,$87
	.dc.b	$38,$57,$38,$C3,$C6,$F9,$D7,$FB,$62,$F9,$DF,$08,$E7,$E2,$7F,$82
	.dc.b	$A3,$07,$E4,$30,$E7,$3C,$6F,$9C,$FE,$52,$F9,$E1,$E1,$C3,$93,$71
	.dc.b	$F4,$56,$EF,$8B,$F0,$1B,$8F,$68,$F0,$5F,$2C,$7C,$11,$C8,$E0,$59
	.dc.b	$E0,$9B,$8F,$92,$78,$CF,$B0,$D5,$75,$7B,$E1,$D6,$B8,$47,$9C,$ED
	.dc.b	$3B,$78,$CE,$F0,$F5,$D7,$09,$C9,$DF,$0F,$1C,$E8,$F7,$DF,$38,$CC
	.dc.b	$F3,$3B,$DC,$7D,$17,$D4,$ED,$3F,$00,$E3,$1A,$BE,$B5,$C7,$1F,$C1
	.dc.b	$4F,$53,$C9,$87,$84,$F6,$61,$D6,$F8,$CE,$AF,$85,$F2,$8C,$11,$C1
	.dc.b	$31,$FC,$5C,$F8,$5F,$A3,$19,$9F,$9E,$71,$8C,$91,$CC,$7E,$25,$F6
	.dc.b	$1E,$28,$EA,$F1,$3E,$B1,$A3,$07,$D2,$F1,$FF,$D6,$8F,$81,$AE,$7E
	.dc.b	$03,$9C,$62,$3F,$97,$FC,$17,$F3,$08,$C4,$23,$13,$FE,$0D,$8C,$09
	.dc.b	$DC,$EA,$3F,$C2,$4F,$E6,$78,$7F,$AD,$F2,$FE,$6E,$7B,$9F,$95,$8C
	.dc.b	$BF,$97,$33,$8F,$F0,$78,$CB,$FC,$26,$33,$8F,$F4,$67,$FD,$3F,$CF
	.dc.b	$FD,$26,$75,$1E,$67,$94,$66,$7D,$58,$79,$63,$15,$E8,$3C,$8E,$1E
	.dc.b	$D1,$BC,$3D,$51,$D6,$BC,$58,$79,$9E,$B1,$CA,$77,$5C,$63,$DC,$71
	.dc.b	$3F,$D8,$17,$DD,$F6,$98,$9E,$2F,$33,$8C,$F3,$C3,$DE,$DB,$F0,$C7
	.dc.b	$C0,$AE,$2D,$AE,$27,$37,$8B,$67,$89,$EC,$BF,$24,$79,$CF,$9F,$5D
	.dc.b	$27,$4D,$8F,$41,$DA,$BC,$07,$62,$CF,$68,$DC,$66,$39,$C7,$16,$C6
	.dc.b	$63,$71,$C5,$F1,$9B,$32,$7D,$B8,$EE,$D8,$E6,$71,$C3,$EC,$46,$4E
	.dc.b	$27,$E0,$0F,$88,$64,$EF,$3E,$36,$CF,$3B,$FC,$5F,$81,$FE,$53,$8F
	.dc.b	$F0,$E6,$9F,$E5,$CF,$AB,$B9,$7F,$97,$26,$E1,$FE,$5C,$9B,$7C,$6B
	.dc.b	$FC,$89,$18,$E4,$5D,$8E,$39,$BE,$AD,$F5,$78,$6E,$1C,$8F,$F7,$F3
	.dc.b	$8C,$BF,$C1,$0D,$CF,$FC,$65,$7C,$EF,$B5,$F0,$7C,$F6,$F9,$8F,$F7
	.dc.b	$CC,$FF,$9A,$1B,$1F,$E9,$CF,$91,$7C,$27,$09,$CE,$3F,$E2,$C6,$DF
	.dc.b	$EE,$DB,$6F,$31,$87,$2A,$ED,$CF,$E2,$FC,$0F,$FC,$15,$1F,$E7,$86
	.dc.b	$DF,$FA,$B3,$E9,$FF,$CE,$93,$9E,$8E,$04,$7F,$D7,$11,$A9,$FF,$3C
	.dc.b	$57,$2C,$FF,$C8,$3F,$10,$F9,$AF,$82,$DC,$3F,$D8,$1F,$44,$F2,$B3
	.dc.b	$52,$7F,$D1,$75,$BC,$7F,$D7,$F7,$B9,$CD,$6E,$73,$F0,$37,$1F,$F3
	.dc.b	$26,$EC,$13,$F4,$BE,$99,$F1,$34,$FC,$5D,$6F,$E5,$F1,$9F,$FA,$72
	.dc.b	$B6,$DA,$DE,$0A,$DE,$5F,$E4,$4E,$7F,$C8,$1F,$91,$E6,$B8,$7F,$A5
	.dc.b	$E2,$CF,$F9,$2A,$6E,$9F,$E6,$8F,$A0,$78,$9F,$88,$6E,$1F,$23,$E5
	.dc.b	$F1,$32,$3F,$D6,$33,$AC,$3E,$3D,$65,$FF,$56,$46,$6F,$E2,$E6,$F9
	.dc.b	$6D,$46,$1F,$D4,$7D,$64,$C7,$FD,$5D,$E8,$5C,$7F,$CC,$33,$6F,$FD
	.dc.b	$A3,$36,$BF,$D2,$7D,$BF,$AD,$F9,$FF,$D6,$1F,$B3,$FA,$4B,$3F,$AC
	.dc.b	$8B,$8F,$F9,$B2,$6E,$1F,$0F,$D7,$F8,$47,$EE,$3E,$D1,$D3,$8C,$7E
	.dc.b	$0C,$DC,$3F,$CF,$95,$A9,$C9,$23,$FB,$A6,$FF,$99,$7F,$B2,$07,$0F
	.dc.b	$F6,$76,$BF,$AD,$FC,$25,$EA,$7E,$21,$B5,$FE,$B2,$CB,$F2,$30,$F7
	.dc.b	$35,$52,$B4,$7F,$AE,$3A,$95,$CB,$06,$1C,$B0,$E1,$1F,$E9,$0F,$63
	.dc.b	$F8,$E1,$FC,$77,$E9,$EE,$3F,$EE,$38,$B3,$FE,$96,$9B,$7F,$EB,$4F
	.dc.b	$C7,$DF,$35,$FC,$51,$B8,$79,$2F,$F8,$57,$D4,$3D,$2D,$5C,$AF,$F7
	.dc.b	$FD,$F0,$BF,$3C,$7A,$EF,$83,$AA,$E1,$3A,$3A,$F1,$63,$1F,$DB,$DF
	.dc.b	$D7,$9F,$F5,$6F,$D4,$7F,$39,$F9,$8F,$E7,$4D,$C7,$CD,$CE,$FB,$F7
	.dc.b	$D3,$FF,$3C,$6C,$FA,$FD,$27,$A4,$C7,$57,$FD,$17,$EA,$3F,$A3,$76
	.dc.b	$5C,$38,$5A,$3F,$A5,$75,$F5,$FA,$6B,$5E,$2C,$D1,$98,$C3,$AD,$BF
	.dc.b	$D9,$D3,$A9,$EB,$F3,$7F,$D5,$4E,$13,$FE,$CC,$7D,$5E,$1C,$9F,$1F
	.dc.b	$C4,$5A,$E7,$37,$8D,$49,$70,$7C,$0B,$D4,$FE,$D3,$16,$1F,$DA,$3A
	.dc.b	$38,$CF,$AE,$30,$FF,$66,$FB,$37,$C4,$53,$87,$AB,$67,$E9,$CE,$9B
	.dc.b	$3A,$69,$33,$FC,$16,$6C,$7F,$41,$A7,$F6,$AC,$39,$56,$A7,$37,$FC
	.dc.b	$32,$5F,$37,$E5,$3D,$5D,$D6,$7C,$71,$6F,$94,$DC,$3B,$4F,$C1,$EB
	.dc.b	$C5,$FC,$9C,$BC,$91,$F0,$7E,$62,$4F,$E2,$F3,$FE,$A9,$87,$9E,$73
	.dc.b	$97,$E7,$7A,$5A,$27,$E7,$7A,$3A,$BE,$18,$F3,$E9,$C3,$E1,$CB,$F3
	.dc.b	$39,$7E,$9C,$7D,$33,$E4,$FF,$C1,$6D,$AF,$27,$D4,$E2,$64,$7E,$EE
	.dc.b	$B8,$C7,$DF,$9C,$C7,$5A,$E4,$FC,$FF,$57,$39,$7E,$32,$3F,$17,$8F
	.dc.b	$F7,$17,$C9,$F4,$3F,$5A,$75,$08,$C6,$5E,$BF,$5B,$FC,$43,$18,$9F
	.dc.b	$CB,$35,$E1,$CB,$F9,$39,$DB,$FE,$3E,$A7,$B5,$FE,$0F,$F8,$1F,$C9
	.dc.b	$DF,$E8,$23,$15,$F2,$7E,$4B,$D1,$5A,$BE,$3D,$F8,$AF,$D3,$46,$0F
	.dc.b	$F5,$DB,$8B,$E7,$87,$C0,$C3,$83,$DE,$EA,$FA,$3F,$E7,$DE,$67,$28
	.dc.b	$FD,$94,$61,$FF,$62,$61,$FE,$B1,$FA,$57,$CC,$F2,$AD,$69,$EC,$BE
	.dc.b	$9D,$BF,$E3,$B7,$89,$D1,$C4,$6E,$F9,$EB,$E1,$AE,$39,$5C,$3A,$E5
	.dc.b	$FE,$C1,$6C,$FF,$07,$BE,$2F,$63,$89,$63,$6E,$30,$F0,$65,$E9,$CB
	.dc.b	$C7,$87,$56,$D7,$9F,$1F,$26,$3D,$B0,$F7,$63,$D0,$97,$C2,$D5,$BC
	.dc.b	$3A,$61,$EB,$AD,$5F,$92,$B5,$7D,$CF,$44,$FC,$97,$FB,$EC,$FD,$36
	.dc.b	$C6,$5E,$A7,$FB,$DF,$E2,$3F,$86,$FD,$2F,$FB,$E6,$11,$87,$5D,$E9
	.dc.b	$D4,$E6,$7B,$F7,$7F,$BF,$DF,$F0,$69,$F8,$77,$0E,$BE,$23,$B6,$3F
	.dc.b	$DD,$1B,$3C,$75,$F8,$ED,$D7,$ED,$36,$FF,$A0,$7D,$2F,$F0,$F7,$F0
	.dc.b	$11,$97,$C1,$78,$3F,$0D,$3B,$8F,$BD,$3B,$C3,$C5,$18,$AD,$F7,$7F
	.dc.b	$BF,$30,$F0,$C7,$39,$D1,$F8,$9C,$3A,$63,$E8,$9D,$36,$75,$7C,$DF
	.dc.b	$D3,$E3,$E1,$25,$78,$23,$8E,$1E,$07,$BD,$FE,$9F,$0E,$F3,$FB,$CF
	.dc.b	$9B,$FF,$5F,$D7,$EC,$F8,$91,$88,$FD,$5C,$63,$E2,$3C,$5E,$1B,$E7
	.dc.b	$7E,$AE,$5F,$C2,$70,$E0,$DA,$F4,$5F,$1C,$3F,$80,$61,$C7,$3F,$64
	.dc.b	$7F,$23,$E1,$FC,$53,$4F,$C6,$75,$7F,$5A,$6C,$6A,$31,$1F,$CB,$A3
	.dc.b	$FB,$E3,$4F,$F0,$E6,$7F,$E3,$57,$F9,$F7,$D2,$FE,$7F,$B3,$F9,$C6
	.dc.b	$3E,$BF,$4B,$8C,$62,$FF,$CD,$37,$C3,$A3,$3A,$76,$C7,$F4,$93,$67
	.dc.b	$D3,$18,$8F,$F0,$73,$53,$E1,$9E,$11,$D4,$F3,$F0,$76,$5E,$33,$E9
	.dc.b	$7F,$C6,$91,$EC,$F8,$AD,$6B,$F4,$DE,$15,$BF,$4A,$DF,$0B,$FF,$3C
	.dc.b	$CE,$5F,$D7,$B5,$CC,$6D,$F9,$71,$9D,$7F,$DF,$33,$CD,$B5,$E2,$D7
	.dc.b	$FD,$8F,$7F,$E5,$0A,$79,$D9,$F8,$99,$39,$8F,$29,$FC,$27,$1F,$73
	.dc.b	$6B,$8C,$F9,$9A,$F9,$A3,$8B,$CC,$FB,$FF,$4B,$F0,$BE,$93,$FD,$19
	.dc.b	$1F,$82,$F1,$3C,$C6,$7C,$09,$1F,$0A,$76,$FC,$66,$BF,$09,$AF,$FA
	.dc.b	$E1,$DE,$BF,$E0,$1C,$91,$8E,$1D,$5F,$C5,$E5,$F2,$78,$BF,$1F,$E9
	.dc.b	$66,$5F,$E7,$6C,$3F,$DE,$2D,$9F,$89,$18,$8F,$24,$60,$8E,$C8,$7F
	.dc.b	$92,$38,$1F,$4B,$F5,$7B,$64,$FE,$91,$F8,$DF,$4B,$23,$E1,$3E,$02
	.dc.b	$3F,$CD,$39,$BF,$80,$78,$3B,$8F,$F1,$D6,$4A,$FC,$14,$FE,$0F,$E5
	.dc.b	$7F,$09,$F4,$B2,$FF,$67,$1F,$96,$F9,$7F,$E7,$B1,$F8,$E6,$BF,$EE
	.dc.b	$46,$BF,$A2,$6E,$3F,$C4,$7E,$97,$F2,$1E,$54,$AF,$DF,$C7,$EA,$57
	.dc.b	$D1,$FD,$37,$FD,$01,$F4,$6C,$74,$D8,$FF,$21,$4E,$DE,$3C,$9F,$C5
	.dc.b	$23,$F0,$31,$FE,$89,$6B,$A3,$FB,$13,$C4,$91,$DB,$7F,$D3,$7A,$72
	.dc.b	$BF,$8A,$BF,$A9,$FA,$5C,$47,$D9,$8B,$D3,$FC,$4E,$7F,$82,$78,$6F
	.dc.b	$A5,$46,$1B,$7F,$ED,$59,$D4,$E7,$E9,$74,$C9,$D5,$70,$7F,$DA,$15
	.dc.b	$D6,$BE,$56,$6F,$4B,$E3,$F8,$F2,$FE,$3C,$68,$F8,$C4,$7E,$1B,$5F
	.dc.b	$F3,$B4,$F3,$78,$9F,$E7,$1D,$BF,$BF,$E7,$FA,$BE,$3F,$F2,$DC,$FF
	.dc.b	$A2,$FE,$A7,$F8,$FF,$A8,$7F,$15,$E0,$C9,$FF,$6F,$D7,$07,$FD,$1C
	.dc.b	$DB,$FE,$79,$A7,$F2,$8F,$81,$FE,$8F,$5F,$EA,$E6,$CF,$F4,$46,$F8
	.dc.b	$3F,$E9,$F7,$F9,$C4,$FF,$B2,$1C,$B4,$E8,$75,$DF,$F2,$5C,$9F,$E5
	.dc.b	$7B,$FF,$40,$FD,$33,$E0,$3F,$2C,$DB,$FF,$49,$56,$F1,$FE,$B6,$FF
	.dc.b	$97,$E3,$CE,$D7,$3E,$5F,$F6,$88,$C4,$7F,$B7,$5E,$4D,$8F,$19,$E0
	.dc.b	$F5,$FE,$17,$CE,$49,$FF,$7D,$BF,$17,$E9,$71,$1F,$16,$2E,$7F,$16
	.dc.b	$7F,$5D,$F4,$C7,$F6,$1E,$DB,$CE,$A6,$A2,$61,$FE,$AF,$AE,$0D,$8E
	.dc.b	$3E,$87,$EE,$7E,$95,$F1,$DF,$BB,$FA,$67,$C7,$53,$51,$30,$FF,$29
	.dc.b	$57,$06,$C7,$F8,$32,$31,$1F,$E4,$E6,$E9,$FE,$B4,$6F,$0F,$F8,$29
	.dc.b	$B2,$18,$3E,$4F,$F8,$77,$D1,$B8,$3F,$88,$7D,$33,$E6,$69,$A8,$98
	.dc.b	$7F,$B3,$EB,$83,$63,$8E,$FF,$C6,$7E,$9B,$F1,$B2,$FA,$6F,$F0,$9B
	.dc.b	$FE,$DF,$6E,$DF,$EC,$C9,$D9,$27,$91,$FC,$8E,$7F,$5B,$B7,$F2,$DF
	.dc.b	$27,$FB,$87,$77,$F2,$1A,$FF,$02,$B5,$FE,$69,$1A,$3F,$9C,$11,$F8
	.dc.b	$98,$7E,$9F,$3F,$D6,$69,$FA,$D6,$4E,$B0,$FD,$54,$EA,$30,$7E,$AE
	.dc.b	$76,$FE,$CE,$3F,$9D,$7C,$1F,$FB,$72,$FE,$97,$CA,$7F,$95,$7E,$9B
	.dc.b	$F1,$D7,$8F,$81,$FF,$20,$C6,$21,$3F,$E5,$BB,$E5,$18,$AF,$F2,$3C
	.dc.b	$63,$E9,$BA,$75,$7F,$92,$9B,$5F,$81,$79,$78,$8C,$B1,$87,$28,$C5
	.dc.b	$75,$3F,$11,$97,$8F,$2E,$71,$93,$FD,$4F,$3C,$23,$AD,$F4,$9E,$38
	.dc.b	$F8,$F0,$F6,$B6,$3F,$29,$97,$A6,$7C,$55,$C6,$58,$FC,$3C,$BC,$27
	.dc.b	$68,$34,$CD,$76,$8E,$57,$D5,$EF,$97,$9E,$75,$1E,$BB,$F7,$47,$8A
	.dc.b	$3D,$D5,$C5,$F1,$E3,$E9,$69,$DA,$FE,$4E,$1E,$3B,$EA,$F9,$27,$C3
	.dc.b	$8E,$CF,$C0,$3F,$18,$A7,$43,$C3,$5E,$77,$89,$F0,$2B,$A9,$1C,$D7
	.dc.b	$85,$E8,$6C,$AE,$5F,$B3,$5E,$0C,$7E,$0C,$70,$CB,$E7,$1B,$8F,$A5
	.dc.b	$7D,$E7,$89,$D4,$B7,$EF,$7D,$4F,$7C,$BE,$55,$72,$9F,$1D,$7C,$FA
	.dc.b	$EE,$FD,$4B,$FC,$79,$89,$F6,$C7,$86,$B9,$47,$82,$3E,$C4,$F0,$79
	.dc.b	$18,$24,$7B,$5C,$BF,$32,$3F,$19,$F0,$5F,$EA,$F8,$F6,$B7,$0F,$D8
	.dc.b	$B7,$2F,$D3,$BB,$9F,$E5,$DF,$07,$FC,$BE,$31,$08,$C5,$FF,$78,$8C
	.dc.b	$09,$DC,$EA,$3F,$BE,$BF,$84,$F7,$7F,$9B,$70,$7E,$1A,$7B,$9F,$B3
	.dc.b	$8C,$BF,$B5,$33,$7F,$EE,$38,$CB,$FE,$E5,$8C,$E5,$FE,$A6,$E6,$5F
	.dc.b	$59,$F2,$EB,$C6,$73,$1F,$7C,$F0,$C6,$67,$F7,$11,$D0,$FE,$A1,$87
	.dc.b	$82,$FE,$06,$3E,$BC,$39,$E3,$E3,$BE,$F5,$E3,$C3,$9C,$60,$E7,$5E
	.dc.b	$2C,$78,$61,$FC,$1F,$1F,$83,$18,$AF,$CA,$46,$30,$F0,$C6,$2F,$94
	.dc.b	$62,$BD,$53,$ED,$3E,$AE,$38,$3A,$16,$78,$1E,$78,$E1,$87,$8E,$FA
	.dc.b	$C7,$DA,$3C,$18,$FB,$27,$4D,$8F,$B5,$3F,$12,$BC,$F7,$DE,$79,$61
	.dc.b	$E6,$3D,$C5,$9F,$1C,$7C,$6B,$E9,$39,$8F,$15,$73,$7F,$C0,$87,$BA
	.dc.b	$BD,$51,$D1,$F6,$4D,$9E,$87,$B8,$F5,$95,$EC,$F4,$C3,$DA,$E6,$79
	.dc.b	$9E,$C3,$B9,$5E,$B5,$DD,$63,$F6,$5F,$83,$3C,$9F,$5B,$C9,$A7,$27
	.dc.b	$FC,$C0,$7E,$63,$6F,$F3,$27,$D3,$5C,$B2,$7F,$7A,$F5,$5E,$42,$75
	.dc.b	$87,$F9,$9F,$0B,$5A,$9B,$49,$B1,$C3,$17,$D0,$3E,$83,$0D,$CB,$C1
	.dc.b	$87,$77,$FC,$17,$5C,$2B,$67,$EC,$32,$FA,$F1,$99,$FE,$14,$D8,$FF
	.dc.b	$41,$B5,$FF,$42,$FC,$8B,$8B,$D7,$08,$C5,$7F,$AB,$FE,$99,$E2,$6A
	.dc.b	$73,$5F,$E0,$7A,$FC,$3C,$63,$7F,$E0,$7E,$63,$07,$D1,$F1,$CB,$E4
	.dc.b	$CE,$63,$C3,$F4,$BF,$E6,$DC,$7F,$C5,$39,$23,$0F,$FA,$EB,$E9,$9E
	.dc.b	$5F,$0E,$C9,$7F,$EC,$59,$7C,$3E,$2F,$83,$EE,$F5,$DF,$29,$4F,$1F
	.dc.b	$D2,$3C,$FC,$63,$E2,$AD,$BF,$CD,$70,$46,$1F,$E7,$0D,$7F,$D6,$2D
	.dc.b	$7F,$56,$EF,$87,$E2,$DB,$E1,$BB,$D7,$FB,$5A,$31,$7F,$0B,$E9,$BE
	.dc.b	$4B,$78,$3C,$FF,$4D,$F6,$27,$37,$F4,$CF,$F3,$E3,$8C,$7B,$FC,$2E
	.dc.b	$EF,$DE,$FE,$5B,$AE,$63,$F9,$0E,$AB,$8F,$FB,$26,$6E,$1F,$EC,$99
	.dc.b	$B5,$FC,$1B,$E8,$BF,$D0,$8F,$D8,$6D,$FA,$FE,$6B,$F1,$B1,$8B,$FC
	.dc.b	$C3,$A9,$FF,$6B,$B5,$E9,$E4,$7E,$75,$B7,$F4,$67,$93,$FE,$99,$E8
	.dc.b	$6E,$FA,$6E,$F8,$7D,$C7,$F6,$7F,$4D,$2F,$F5,$BE,$95,$CB,$FD,$CD
	.dc.b	$37,$3F,$F7,$34,$DA,$FE,$AF,$B3,$E7,$FF,$6E,$7E,$FB,$E9,$AC,$FE
	.dc.b	$DA,$74,$FF,$09,$F8,$07,$D0,$6B,$3F,$AD,$E4,$27,$35,$F2,$7B,$FE
	.dc.b	$39,$FC,$1F,$E9,$BF,$1F,$19,$25,$FF,$81,$F8,$FF,$47,$1F,$16,$BE
	.dc.b	$44,$FF,$AF,$67,$2E,$E3,$F8,$CE,$CF,$AA,$72,$F8,$30,$FF,$4C,$CE
	.dc.b	$63,$F6,$FF,$4F,$FE,$D8,$9F,$F4,$EF,$32,$31,$8F,$E4,$63,$1F,$01
	.dc.b	$C3,$DD,$97,$C2,$F7,$8F,$F0,$65,$5A,$C3,$E2,$F9,$7F,$EE,$4B,$FE
	.dc.b	$47,$EF,$7F,$2C,$E0,$FE,$5A,$C8,$C9,$FC,$A2,$31,$5F,$A3,$1F,$4E
	.dc.b	$F9,$CF,$E3,$ED,$AF,$5F,$D3,$FF,$AB,$1F,$E2,$31,$9A,$E4,$D7,$FB
	.dc.b	$29,$D2,$BE,$7B,$67,$F8,$34,$EE,$3F,$B5,$BC,$F4,$FE,$4B,$7F,$72
	.dc.b	$7A,$DF,$F2,$48,$C4,$FF,$8B,$39,$3F,$A8,$FD,$3F,$F4,$CF,$9B,$FE
	.dc.b	$9A,$DC,$BF,$A2,$FD,$33,$E9,$7F,$03,$E9,$A3,$13,$FD,$11,$BA,$7F
	.dc.b	$44,$6B,$FE,$00,$C7,$FA,$FF,$D3,$FF,$5A,$EE,$FE,$B6,$DB,$FE,$AB
	.dc.b	$D3,$FD,$6F,$0F,$CC,$E1,$F6,$23,$1D,$92,$BC,$0F,$92,$B7,$5F,$D6
	.dc.b	$1A,$86,$3F,$DC,$3C,$DF,$ED,$67,$38,$FC,$2B,$E1,$EA,$98,$7F,$68
	.dc.b	$F8,$AB,$5F,$C8,$F5,$F8,$55,$E1,$25,$7C,$36,$CF,$ED,$39,$7F,$CB
	.dc.b	$F3,$F3,$27,$FB,$EC,$62,$7A,$7D,$3B,$E2,$BF,$BE,$36,$FF,$80,$7A
	.dc.b	$D8,$99,$1F,$DD,$27,$55,$FD,$D1,$AF,$F7,$A6,$BF,$E6,$67,$7C,$BF
	.dc.b	$83,$C3,$F5,$35,$9B,$F2,$F6,$F8,$BE,$76,$4F,$F1,$1E,$DB,$97,$A2
	.dc.b	$7F,$0B,$91,$A4,$AF,$33,$F0,$3A,$B8,$E7,$E1,$7D,$FF,$33,$2B,$F8
	.dc.b	$F7,$A7,$71,$F8,$B3,$D8,$93,$F0,$5B,$1F,$3F,$A7,$86,$1E,$D7,$FC
	.dc.b	$27,$F4,$CF,$96,$A7,$D2,$FA,$7B,$8F,$F8,$1A,$74,$FF,$75,$6E,$1F
	.dc.b	$EA,$B6,$BF,$B8,$6B,$FE,$5E,$77,$87,$CD,$F5,$BA,$CF,$E0,$DF,$F1
	.dc.b	$2F,$D1,$FD,$2B,$FF,$32,$B6,$7F,$2F,$C7,$FD,$E1,$FE,$21,$19,$4F
	.dc.b	$DC,$7F,$3C,$D7,$3C,$7F,$C2,$A3,$55,$FE,$35,$EC,$FF,$56,$FC,$0F
	.dc.b	$E3,$EF,$FD,$67,$EE,$7F,$3C,$8C,$76,$7F,$4B,$F9,$6D,$E5,$FA,$BC
	.dc.b	$79,$1F,$ED,$27,$17,$C5,$B8,$F0,$C7,$D7,$76,$F8,$C7,$FB,$5E,$BF
	.dc.b	$D0,$BF,$4B,$FC,$E3,$CA,$E9,$3F,$B9,$8C,$3B,$3F,$79,$F4,$EF,$9D
	.dc.b	$ED,$09,$DE,$BE,$D8,$C7,$5B,$0E,$B1,$86,$E1,$CF,$2C,$E1,$CB,$2F
	.dc.b	$2E,$1C,$70,$F1,$1C,$4B,$7C,$5C,$D7,$17,$D5,$3C,$71,$F2,$4E,$65
	.dc.b	$5C,$67,$35,$BB,$F0,$E3,$E4,$8C,$5F,$C2,$CB,$E0,$E3,$CC,$93,$A9
	.dc.b	$EB,$3A,$8C,$4E,$AB,$E0,$CE,$5E,$C7,$F9,$56,$3E,$8D,$7F,$56,$6B
	.dc.b	$E4,$3F,$AC,$8F,$89,$FE,$05,$F4,$FF,$E5,$D6,$C6,$9A,$FF,$60,$6B
	.dc.b	$F8,$2E,$B7,$C5,$78,$5B,$1D,$AF,$FB,$84,$E4,$91,$F4,$27,$7A,$7C
	.dc.b	$E6,$E9,$FE,$D2,$FA,$77,$67,$FB,$26,$3C,$18,$7E,$02,$B7,$87,$E4
	.dc.b	$5A,$AF,$8B,$B9,$8F,$85,$FC,$6C,$6A,$30,$4C,$3C,$93,$AC,$3C,$53
	.dc.b	$DE,$39,$CE,$A3,$F1,$F5,$CE,$30,$4C,$3D,$33,$AB,$E6,$FE,$8F,$0F
	.dc.b	$4E,$3F,$22,$3C,$55,$F0,$EB,$C3,$C0,$F9,$EF,$F6,$75,$7E,$CA,$31
	.dc.b	$F1,$3A,$9C,$CF,$EC,$F9,$9F,$47,$F5,$A3,$F6,$D7,$C1,$E5,$5C,$34
	.dc.b	$FE,$21,$39,$9E,$55,$C6,$7F,$71,$D1,$FD,$2E,$7F,$CC,$1F,$55,$FD
	.dc.b	$ED,$FE,$D8,$DE,$CF,$F6,$C4,$EE,$7F,$93,$F8,$59,$C7,$FC,$05,$9F
	.dc.b	$F9,$97,$0C,$57,$F2,$E6,$BF,$CB,$2A,$FD,$47,$EF,$E3,$FC,$DB,$9F
	.dc.b	$FA,$D9,$FF,$7D,$35,$FF,$22,$46,$F2,$FF,$75,$37,$1F,$0D,$6E,$7F
	.dc.b	$C6,$4D,$8F,$64,$EE,$BF,$C3,$0D,$4D,$35,$8F,$D0,$DF,$D7,$E4,$3E
	.dc.b	$A1,$E8,$FB,$AB,$EA,$FC,$65,$CF,$FD,$E3,$97,$F8,$97,$46,$9F,$E2
	.dc.b	$7C,$B9,$63,$D3,$2E,$77,$F0,$31,$E4,$66,$64,$73,$BE,$55,$C2,$3F
	.dc.b	$C5,$0D,$C3,$E0,$FC,$27,$5F,$3B,$C9,$F1,$52,$F9,$0C,$FD,$93,$B8
	.dc.b	$FA,$33,$B7,$AC,$65,$A7,$29,$E2,$FD,$8F,$A8,$F8,$FF,$4D,$FE,$86
	.dc.b	$8F,$B7,$E2,$FC,$53,$3E,$14,$8C,$90,$FF,$21,$4F,$16,$C6,$DA,$7E
	.dc.b	$1B,$93,$BF,$C5,$7D,$1F,$A8,$93,$E5,$BF,$F7,$C3,$67,$2F,$E4,$27
	.dc.b	$8B,$92,$1E,$28,$FB,$9F,$51,$F7,$7A,$E4,$FE,$59,$F7,$FC,$54,$8C
	.dc.b	$10,$F5,$BF,$85,$EE,$D4,$60,$FC,$FB,$F8,$9F,$97,$FC,$57,$D4,$49
	.dc.b	$FD,$1C,$7E,$4B,$E6,$7F,$81,$72,$FE,$7A,$38,$4F,$D3,$9C,$F3,$FE
	.dc.b	$93,$EA,$3F,$29,$E5,$4A,$FD,$6C,$7E,$D3,$E9,$BE,$76,$3F,$96,$9F
	.dc.b	$F3,$E7,$D3,$32,$FD,$CB,$63,$F1,$22,$B8,$E3,$FC,$25,$B1,$CA,$B8
	.dc.b	$9D,$F9,$24,$72,$BE,$8F,$FB,$05,$B8,$78,$E3,$B4,$FE,$B7,$D4,$91
	.dc.b	$FC,$19,$FD,$87,$D4,$62,$3F,$85,$C5,$AF,$E0,$73,$FC,$53,$CE,$7D
	.dc.b	$42,$30,$DB,$FF,$73,$B5,$FF,$73,$2F,$A7,$DC,$C9,$DB,$CA,$3E,$67
	.dc.b	$A5,$F2,$3E,$5A,$57,$C8,$D9,$3A,$8F,$85,$3A,$7F,$DC,$53,$A7,$84
	.dc.b	$FF,$80,$9F,$F5,$84,$65,$FF,$02,$B5,$FF,$6F,$B5,$FE,$21,$C4,$EC
	.dc.b	$FE,$31,$EA,$BE,$A2,$9F,$8D,$89,$5F,$C1,$BC,$85,$6E,$30,$D7,$FD
	.dc.b	$71,$7D,$5E,$1F,$13,$FE,$F0,$E0,$FF,$72,$37,$97,$FA,$3B,$67,$FD
	.dc.b	$92,$FF,$30,$31,$81,$A4,$E1,$FE,$97,$C7,$FE,$51,$C3,$F0,$DF,$50
	.dc.b	$F8,$2F,$E9,$ED,$C7,$F2,$F5,$BC,$BF,$AF,$36,$3F,$DB,$53,$B5,$7D
	.dc.b	$2F,$BF,$99,$27,$F4,$35,$BA,$D4,$76,$9F,$E8,$99,$3E,$7E,$92,$7E
	.dc.b	$2F,$D2,$57,$BE,$7F,$12,$71,$87,$C8,$AF,$ED,$3F,$50,$7F,$B5,$FB
	.dc.b	$AF,$2E,$9A,$89,$8F,$ED,$A7,$4D,$9C,$BE,$03,$FD,$F5,$F5,$0F,$84
	.dc.b	$FE,$EF,$F5,$0F,$8F,$A7,$F7,$A9,$D5,$7F,$02,$6B,$CC,$4F,$17,$F8
	.dc.b	$3B,$5D,$47,$0C,$FF,$7C,$DD,$7F,$C2,$55,$B9,$FE,$04,$D8,$FF,$0B
	.dc.b	$4E,$E8,$3E,$97,$E8,$3F,$C5,$BE,$A1,$F3,$7F,$E1,$26,$C7,$F1,$A9
	.dc.b	$DB,$E9,$6E,$3F,$C4,$67,$73,$ED,$6C,$7F,$15,$9D,$E1,$FE,$29,$FA
	.dc.b	$8F,$CE,$5B,$EA,$3F,$C2,$4F,$99,$AA,$36,$70,$8C,$9A,$7F,$4D,$E9
	.dc.b	$BA,$FF,$92,$DF,$F8,$C3,$B1,$F4,$3F,$BC,$3F,$A3,$79,$13,$2F,$DB
	.dc.b	$C6,$5F,$E8,$D1,$F9,$6D,$BF,$2F,$A4,$9F,$DC,$C6,$5F,$DD,$C6,$5F
	.dc.b	$C4,$77,$7E,$D6,$7F,$C7,$BF,$09,$FC,$37,$7F,$A7,$F4,$FF,$E5,$AF
	.dc.b	$A8,$FC,$8E,$08,$C7,$3F,$F9,$3D,$FF,$2F,$DF,$19,$D7,$C1,$7F,$94
	.dc.b	$E3,$13,$C2,$7A,$CF,$F9,$5E,$FA,$CF,$2F,$0F,$F7,$35,$C8,$FC,$2D
	.dc.b	$F2,$BE,$18,$78,$5B,$3D,$0D,$D7,$2B,$E9,$8E,$65,$8F,$82,$F8,$C7
	.dc.b	$73,$FD,$6D,$7C,$63,$D4,$D8,$FC,$BC,$F1,$6E,$1E,$37,$C3,$7D,$F1
	.dc.b	$F4,$46,$8F,$15,$71,$7B,$CE,$EF,$B4,$79,$AF,$A3,$4D,$4C,$7D,$15
	.dc.b	$CC,$E9,$7E,$08,$EA,$6A,$B9,$CE,$AF,$BD,$6E,$39,$4F,$82,$39,$19
	.dc.b	$78,$99,$9E,$27,$A2,$78,$D7,$97,$1F,$A2,$F8,$4F,$83,$5C,$70,$D6
	.dc.b	$3F,$4A,$BF,$0D,$5F,$4F,$1F,$0E,$3D,$70,$F0,$CF,$03,$E7,$9D,$49
	.dc.b	$5E,$02,$5F,$52,$47,$08,$CB,$F5,$4E,$31,$F8,$0C,$3B,$47,$D1,$32
	.dc.b	$71,$8D,$38,$35,$3D,$1F,$A3,$F0,$CF,$A8,$DB,$63,$D2,$D7,$71,$FE
	.dc.b	$BB,$7F,$92,$FC,$27,$F2,$69,$D4,$60,$38,$51,$B9,$E5,$18,$36,$E2
	.dc.b	$3F,$49,$CB,$FD,$7F,$93,$F4,$D5,$D7,$0F,$F6,$DC,$E6,$3F,$4C,$67
	.dc.b	$1F,$F6,$84,$66,$7F,$DC,$33,$93,$FD,$51,$E0,$7F,$95,$63,$31,$E3
	.dc.b	$7A,$C6,$6B,$C9,$3F,$B3,$AE,$91,$8A,$DD,$F9,$AB,$94,$F8,$AF,$94
	.dc.b	$7C,$9C,$3D,$B1,$88,$FF,$73,$46,$2B,$93,$63,$DA,$74,$3D,$31,$A7
	.dc.b	$C9,$3C,$DF,$71,$FE,$B3,$70,$F1,$31,$1C,$6F,$D3,$1D,$AB,$C1,$1F
	.dc.b	$11,$B7,$EF,$3D,$95,$C5,$B5,$C4,$F1,$C6,$5F,$39,$9A,$EE,$75,$76
	.dc.b	$66,$79,$C6,$4D,$BE,$18,$E7,$19,$8D,$99,$8F,$69,$CC,$DB,$9C,$3E
	.dc.b	$CC,$7D,$33,$C3,$3C,$DD,$99,$7C,$2D,$9E,$ED,$36,$7D,$A3,$9B,$DF
	.dc.b	$0F,$A3,$87,$CA,$9E,$07,$9A,$95,$FE,$5A,$3F,$D5,$19,$FF,$97,$BE
	.dc.b	$A2,$E5,$93,$FC,$C6,$E2,$FF,$CB,$F5,$78,$BF,$CB,$D3,$7C,$2E,$B7
	.dc.b	$C0,$FF,$77,$7C,$F3,$E9,$38,$74,$BD,$7E,$9C,$0A,$D5,$FF,$9D,$A7
	.dc.b	$31,$8D,$3F,$BC,$DF,$1A,$EB,$87,$4B,$F3,$53,$E4,$FF,$D1,$15,$FE
	.dc.b	$2E,$DB,$FD,$8D,$E5,$BE,$A1,$F0,$5C,$2F,$78,$7F,$93,$6B,$53,$9E
	.dc.b	$8F,$85,$18,$C7,$FD,$97,$7C,$31,$EB,$8F,$49,$FF,$67,$FD,$47,$F9
	.dc.b	$7D,$F1,$1F,$E4,$0B,$7D,$43,$E1,$3D,$35,$BD,$3C,$53,$9B,$E7,$D7
	.dc.b	$E2,$8C,$4F,$F9,$66,$B8,$65,$EE,$9C,$BF,$ED,$3F,$51,$F2,$BE,$AF
	.dc.b	$A7,$7C,$17,$C8,$9C,$E7,$E6,$C9,$AA,$F7,$3F,$EE,$AA,$DD,$7C,$79
	.dc.b	$CB,$FE,$6B,$8C,$E3,$F4,$27,$2F,$8F,$1F,$16,$3E,$16,$E1,$FE,$F6
	.dc.b	$FA,$8F,$95,$8F,$5E,$3F,$BB,$18,$C3,$A4,$63,$8F,$D8,$DB,$F5,$69
	.dc.b	$E9,$8F,$CB,$73,$49,$FF,$5F,$F2,$DC,$BF,$C8,$BD,$1F,$93,$9F,$F6
	.dc.b	$1D,$7E,$76,$B5,$97,$FB,$F7,$E8,$FF,$CE,$B5,$DB,$77,$D4,$7E,$4A
	.dc.b	$73,$97,$C5,$8C,$78,$BF,$45,$B7,$C5,$B3,$B8,$FA,$4A,$B9,$E1,$F2
	.dc.b	$EB,$F6,$1F,$51,$2F,$FD,$B3,$F5,$17,$0F,$5C,$DB,$FF,$6D,$CD,$AF
	.dc.b	$F4,$8F,$6F,$EC,$7E,$87,$F6,$BF,$41,$FB,$06,$CF,$EE,$9B,$8F,$FA
	.dc.b	$C6,$B8,$4F,$FB,$B2,$BE,$07,$CD,$23,$FE,$00,$68,$8C,$5F,$F7,$48
	.dc.b	$C4,$EA,$73,$3F,$E0,$19,$CB,$FE,$BF,$C9,$F4,$CF,$A0,$EF,$87,$E5
	.dc.b	$E7,$31,$8A,$E1,$5B,$AF,$97,$7D,$2F,$F9,$0C,$E6,$BF,$80,$7D,$37
	.dc.b	$FB,$2E,$FF,$12,$D8,$CF,$BF,$FC,$0A,$31,$7F,$C3,$1B,$5F,$BD,$6D
	.dc.b	$7E,$A9,$57,$1F,$67,$F9,$47,$B3,$2F,$F6,$1E,$43,$A1,$E3,$E7,$2F
	.dc.b	$CA,$DD,$CF,$E5,$DE,$EB,$18,$7F,$B3,$2D,$F4,$CF,$9B,$FE,$33,$F4
	.dc.b	$CF,$0F,$F9,$3C,$E6,$FF,$4F,$39,$7F,$23,$19,$BF,$E4,$93,$97,$FD
	.dc.b	$D1,$19,$CB,$F9,$6C,$E5,$FB,$71,$9A,$FE,$41,$39,$7C,$D1,$9A,$FE
	.dc.b	$1F,$39,$59,$7C,$3F,$A6,$FE,$9D,$F2,$FF,$D3,$DB,$87,$F1,$7E,$77
	.dc.b	$D2,$BE,$93,$FA,$A7,$D3,$7F,$4E,$8C,$57,$CE,$E2,$4E,$A3,$1D,$77
	.dc.b	$4F,$E0,$6F,$F5,$DE,$AE,$5D,$76,$BC,$18,$A3,$31,$D7,$CC,$B8,$7F
	.dc.b	$0C,$7F,$B3,$12,$38,$CD,$9F,$D6,$57,$E4,$F8,$BF,$4F,$AF,$B3,$B3
	.dc.b	$FB,$77,$C3,$FA,$D9,$8F,$E2,$39,$FF,$60,$F9,$23,$FB,$7B,$AC,$7F
	.dc.b	$B4,$B7,$0F,$C4,$D7,$09,$FD,$E7,$D3,$3E,$1F,$FB,$AB,$6F,$B7,$A3
	.dc.b	$89,$91,$E1,$EC,$B7,$FD,$D9,$B8,$F6,$6C,$FF,$C0,$17,$AF,$CC,$C8
	.dc.b	$F5,$77,$5B,$F1,$CF,$F8,$23,$B7,$18,$7A,$1B,$6A,$FC,$3E,$9F,$8B
	.dc.b	$DD,$AB,$1E,$57,$F8,$78,$7C,$AF,$B8,$FF,$0A,$F9,$D6,$75,$1F,$E1
	.dc.b	$8E,$9C,$65,$FE,$12,$6E,$9F,$29,$B5,$E3,$EC,$FD,$CC,$7F,$85,$3E
	.dc.b	$95,$F2,$54,$FF,$11,$FD,$35,$BF,$E7,$F3,$A7,$EE,$7C,$56,$72,$FF
	.dc.b	$10,$B7,$2F,$B2,$D9,$F0,$70,$23,$FC,$47,$F4,$5F,$66,$FF,$C5,$3E
	.dc.b	$C6,$39,$BF,$C6,$6D,$C3,$F1,$4D,$9D,$C7,$F8,$0C,$7B,$3F,$CD,$3D
	.dc.b	$DF,$C3,$4F,$F3,$F8,$C5,$F8,$BE,$47,$F9,$FD,$6A,$F8,$F8,$18,$AF
	.dc.b	$B1,$18,$3F,$D7,$EE,$2F,$9E,$1F,$73,$0E,$0F,$7B,$AB,$E8,$FF,$96
	.dc.b	$FA,$DC,$A3,$F0,$91,$87,$FD,$8B,$87,$F2,$5F,$A6,$7C,$CF,$2A,$D6
	.dc.b	$9E,$CB,$E9,$DB,$FE,$22,$78,$9D,$1C,$46,$EF,$9E,$BE,$1A,$E3,$95
	.dc.b	$C3,$AE,$5F,$CD,$5B,$3F,$99,$BE,$2F,$63,$89,$63,$6E,$30,$F0,$65
	.dc.b	$E9,$CB,$C7,$87,$56,$D7,$9F,$1F,$26,$3D,$B0,$F7,$63,$D0,$97,$C2
	.dc.b	$D5,$BC,$3A,$61,$EB,$AD,$5F,$92,$B5,$7D,$CF,$44,$FC,$97,$FD,$F9
	.dc.b	$3F,$4D,$B1,$97,$A9,$FE,$F8,$F8,$5F,$DC,$7D,$37,$FB,$EA,$11,$87
	.dc.b	$5D,$A9,$D4,$E6,$7B,$F7,$BE,$1B,$96,$3F,$D8,$DB,$19,$9F,$ED,$07
	.dc.b	$E2,$1C,$3C,$A3,$ED,$46,$F5,$F9,$2D,$D7,$FB,$87,$D3,$7F,$0C,$BD
	.dc.b	$DC,$7E,$D5,$6D,$FE,$31,$3B,$73,$3B,$BE,$AF,$E2,$E7,$51,$89,$D9
	.dc.b	$1F,$CC,$46,$AF,$F1,$D7,$DE,$BF,$4C,$DC,$3E,$14,$7E,$3E,$B8,$3F
	.dc.b	$07,$0F,$15,$F2,$7C,$5D,$DF,$E0,$DC,$7C,$44,$9D,$57,$39,$D4,$7F
	.dc.b	$14,$EA,$B3,$FC,$DB,$C4,$BD,$FF,$E1,$E6,$F5,$AA,$98,$89,$5F,$BD
	.dc.b	$AF,$F1,$2C,$F0,$F5,$FF,$C4,$1D,$BF,$D7,$3E,$AA,$F4,$7F,$24,$6E
	.dc.b	$1F,$D1,$AA,$E1,$FE,$2E,$6E,$4A,$A6,$23,$7F,$AE,$FE,$13,$1F,$F1
	.dc.b	$8A,$F8,$9C,$9A,$E0,$99,$7F,$8F,$27,$4D,$9C,$AA,$E7,$5C,$A3,$47
	.dc.b	$17,$C3,$EA,$78,$B7,$93,$D6,$32,$F7,$8C,$BD,$75,$FE,$A9,$8F,$F9
	.dc.b	$06,$38,$1B,$21,$B3,$FC,$62,$ED,$C6,$A8,$7D,$2F,$C4,$F8,$19,$5E
	.dc.b	$F8,$F8,$E6,$1E,$33,$FE,$F0,$74,$43,$33,$BC,$FE,$6D,$7D,$0F,$86
	.dc.b	$FA,$3F,$49,$2F,$E6,$46,$5F,$9D,$19,$7B,$C6,$5F,$0C,$66,$BE,$CC
	.dc.b	$FD,$CF,$0B,$EE,$F6,$4B,$FB,$11,$97,$ED,$46,$5F,$5C,$65,$F6,$C6
	.dc.b	$6B,$F0,$B3,$F8,$8F,$A3,$95,$F8,$37,$F1,$7F,$47,$23,$04,$3E,$3B
	.dc.b	$96,$9F,$8B,$7F,$27,$F4,$7F,$94,$E9,$93,$F8,$F7,$F2,$FF,$23,$F9
	.dc.b	$8F,$A3,$B3,$F9,$38,$C3,$F9,$5F,$17,$F3,$A4,$7F,$31,$F4,$7F,$A8
	.dc.b	$5F,$47,$FD,$A1,$FF,$6A,$A8,$D3,$63,$33,$83,$33,$23,$24,$3F,$20
	.dc.b	$76,$74,$E2,$7C,$0F,$E2,$BC,$EF,$D9,$FD,$14,$8F,$D7,$47,$ED,$8C
	.dc.b	$3F,$90,$8C,$3F,$B3,$FA,$2F,$E0,$3F,$AF,$7F,$40,$E1,$FF,$55,$B7
	.dc.b	$A7,$F4,$F3,$FC,$00,$DD,$43,$91,$FE,$B8,$9F,$EB,$D5,$CB,$E5,$7A
	.dc.b	$9D,$BC,$C9,$5B,$8C,$BC,$63,$2F,$EB,$A9,$3A,$8C,$4F,$D3,$FA,$37
	.dc.b	$9F,$FD,$96,$51,$87,$FB,$43,$5F,$EB,$A6,$A3,$84,$67,$0F,$F7,$8B
	.dc.b	$72,$FF,$65,$36,$BF,$CD,$DA,$FF,$21,$9F,$F2,$F6,$4F,$A0,$F9,$7F
	.dc.b	$33,$2F,$E2,$47,$CE,$30,$FB,$63,$0F,$F9,$E7,$1F,$F8,$03,$F6,$EF
	.dc.b	$63,$F2,$0F,$FA,$EE,$75,$8F,$FB,$76,$31,$1F,$88,$78,$3A,$3F,$D0
	.dc.b	$8E,$1D,$7B,$1D,$0E,$B5,$23,$80,$9E,$B7,$E4,$AD,$4F,$D5,$F8,$EF
	.dc.b	$C4,$FC,$C4,$8F,$C1,$47,$E3,$0C,$3F,$6E,$30,$F6,$F5,$3F,$4C,$FD
	.dc.b	$46,$BF,$E1,$26,$BF,$EF,$B6,$BF,$E9,$F7,$8D,$B2,$D7,$37,$ED,$E7
	.dc.b	$31,$FE,$A5,$AF,$F4,$27,$3F,$EE,$FE,$DF,$E2,$61,$C6,$7F,$D5,$51
	.dc.b	$F0,$2B,$83,$63,$3A,$F1,$39,$69,$23,$E4,$CF,$4A,$E3,$59,$9F,$DE
	.dc.b	$CF,$34,$ED,$FF,$5D,$7E,$19,$FE,$B7,$7F,$C8,$0D,$7F,$32,$D7,$FC
	.dc.b	$6E,$D7,$DC,$EF,$0F,$64,$23,$17,$AF,$07,$F8,$E7,$83,$E1,$AF,$EB
	.dc.b	$7F,$3B,$FA,$BF,$14,$D2,$47,$F9,$4E,$32,$FE,$21,$AF,$F1,$A6,$E3
	.dc.b	$E2,$FD,$5B,$DD,$E4,$B9,$DF,$39,$FE,$14,$FC,$27,$F4,$1E,$9D,$E1
	.dc.b	$FF,$11,$67,$FC,$36,$BF,$8B,$99,$99,$19,$3F,$89,$7C,$C7,$F1,$C2
	.dc.b	$3F,$C5,$7E,$62,$BF,$C3,$62,$E9,$FC,$82,$7F,$9F,$7B,$7F,$EB,$E8
	.dc.b	$FF,$76,$7C,$F3,$AB,$F1,$5F,$3E,$F3,$BF,$19,$39,$CB,$FD,$E5,$39
	.dc.b	$78,$D7,$E4,$70,$E9,$0A,$E0,$FF,$9F,$39,$3F,$C4,$4E,$6B,$BE,$1D
	.dc.b	$9B,$1D,$0F,$F8,$01,$7D,$2F,$9B,$67,$73,$CC,$F2,$D7,$37,$8D,$F8
	.dc.b	$2F,$8B,$6B,$8C,$79,$B0,$F6,$57,$33,$D1,$8F,$8D,$A6,$63,$73,$C7
	.dc.b	$0F,$04,$EA,$3B,$D7,$1A,$DD,$F8,$0F,$91,$87,$8E,$FC,$F5,$BC,$39
	.dc.b	$DF,$09,$C9,$E5,$C7,$E1,$3F,$00,$E0,$79,$DC,$BE,$3C,$3C,$0E,$CE
	.dc.b	$58,$78,$5D,$19,$C7,$D8,$F3,$8D,$57,$4B,$F2,$49,$B1,$E0,$35,$87
	.dc.b	$A8,$F8,$AE,$6B,$C1,$3C,$8E,$2F,$63,$33,$C6,$BC,$58,$C7,$E9,$46
	.dc.b	$5F,$5C,$67,$0F,$4C,$68,$C6,$3F,$42,$7E,$3C,$64,$E2,$49,$E2,$49
	.dc.b	$E1,$19,$7E,$5B,$F0,$6B,$DB,$7E,$9B,$EB,$7C,$6B,$EC,$BF,$84,$7F
	.dc.b	$DB,$1E,$AF,$4F,$57,$FD,$13,$1F,$8E,$6E,$7F,$AC,$73,$01,$ED,$58
	.dc.b	$CC,$60,$EA,$7F,$B8,$23,$55,$FE,$06,$6C,$64,$FD,$07,$C3,$7F,$85
	.dc.b	$3A,$BF,$41,$1D,$E7,$F4,$53,$98,$FD,$31,$93,$8F,$93,$FD,$3F,$2F
	.dc.b	$E0,$31,$97,$F8,$1C,$F1,$73,$1E,$77,$33,$C7,$6F,$34,$F4,$C3,$C6
	.dc.b	$EE,$3F,$B0,$5F,$37,$77,$CA,$B7,$87,$A3,$0E,$98,$7B,$EB,$4F,$A8
	.dc.b	$E8,$7C,$DC,$3A,$3B,$C3,$BD,$72,$8F,$F7,$31,$E5,$39,$5F,$7C,$3C
	.dc.b	$31,$9C,$3D,$F3,$98,$D9,$9B,$F4,$9D,$AB,$8D,$79,$8E,$33,$B3,$C8
	.dc.b	$79,$EF,$C8,$FC,$FA,$E4,$F6,$9E,$51,$DA,$38,$57,$68,$DC,$F6,$37
	.dc.b	$3D,$86,$1F,$11,$F8,$31,$A6,$CF,$23,$B3,$C2,$FC,$AF,$CE,$31,$7E
	.dc.b	$17,$C6,$79,$E7,$93,$E1,$BF,$03,$ED,$37,$7E,$79,$DC,$78,$8C,$BE
	.dc.b	$29,$E3,$1E,$16,$E1,$F4,$EB,$FA,$F7,$3F,$F7,$9F,$99,$FF,$14,$6D
	.dc.b	$FE,$88,$6E,$5F,$E6,$4F,$99,$B9,$2F,$84,$FF,$A1,$EB,$31,$89,$FF
	.dc.b	$40,$56,$3A,$55,$7E,$C3,$9F,$53,$F2,$1F,$D7,$67,$FE,$98,$9C,$C6
	.dc.b	$33,$FF,$3A,$C6,$38,$39,$47,$F9,$96,$F9,$47,$9B,$37,$EA,$4F,$95
	.dc.b	$C3,$6F,$FC,$D9,$E7,$BE,$21,$EB,$EA,$E5,$7F,$A3,$B0,$E7,$1F,$E3
	.dc.b	$27,$FC,$A5,$87,$37,$FD,$26,$D7,$FC,$AC,$3E,$89,$E9,$F9,$5B,$7E
	.dc.b	$2F,$9E,$7C,$86,$6A,$4F,$86,$B7,$7F,$EA,$06,$BD,$9A,$FF,$27,$FB
	.dc.b	$4F,$F3,$34,$FF,$A2,$5F,$E5,$5A,$FD,$A7,$32,$E2,$FF,$34,$46,$AB
	.dc.b	$84,$FF,$31,$6D,$71,$8C,$1C,$63,$C0,$A3,$4E,$DB,$1C,$8F,$F4,$AF
	.dc.b	$D0,$FD,$06,$ED,$F8,$EF,$C0,$1F,$80,$3E,$4B,$EE,$5E,$F1,$FF,$76
	.dc.b	$B5,$F9,$ED,$7F,$D0,$CD,$7F,$A5,$37,$AB,$FC,$B1,$F3,$FF,$95,$F8
	.dc.b	$AF,$CB,$37,$3F,$F7,$13,$6F,$FD,$B4,$DD,$BE,$3F,$C7,$FF,$A5,$EB
	.dc.b	$E7,$3F,$69,$A7,$CF,$71,$5F,$1C,$6E,$8E,$51,$98,$C7,$17,$1A,$E4
	.dc.b	$D7,$BB,$67,$85,$76,$BF,$F8,$03,$CB,$FB,$1F,$9D,$FD,$93,$70,$FF
	.dc.b	$7F,$36,$FF,$DE,$CD,$AF,$F4,$7F,$D1,$3E,$4F,$F5,$8D,$BF,$F6,$A7
	.dc.b	$86,$FB,$57,$C6,$EA,$E5,$7E,$A6,$F8,$65,$FE,$F8,$6B,$FA,$66,$BF
	.dc.b	$EF,$2F,$9E,$FE,$11,$F3,$DF,$C2,$5B,$5F,$EE,$3E,$3F,$F4,$84,$62
	.dc.b	$BE,$64,$62,$75,$E2,$2F,$94,$61,$B5,$C2,$75,$3C,$31,$FE,$31,$F3
	.dc.b	$DF,$C6,$5B,$19,$3A,$C6,$5A,$78,$3E,$73,$F9,$1F,$EB,$FF,$73,$83
	.dc.b	$FD,$A9,$9F,$FA,$9E,$73,$5F,$C9,$5A,$F1,$3D,$D5,$C5,$FD,$E7,$1B
	.dc.b	$E1,$FF,$19,$F6,$1F,$8D,$6C,$FE,$49,$BA,$FE,$11,$B9,$7E,$61,$B3
	.dc.b	$D1,$B3,$FC,$F7,$0F,$C5,$4F,$F8,$C7,$24,$62,$75,$F3,$DE,$EE,$7F
	.dc.b	$7F,$C1,$7C,$0F,$9A,$B8,$FA,$9F,$87,$ED,$C8,$C1,$EE,$7E,$3F,$C1
	.dc.b	$48,$F8,$3E,$05,$CB,$EE,$E6,$AF,$9B,$F3,$B9,$DF,$E6,$CA,$33,$87
	.dc.b	$CA,$74,$FA,$E3,$4F,$CE,$FA,$7F,$AF,$C1,$F6,$1B,$19,$3E,$9C,$65
	.dc.b	$A6,$71,$F2,$C6,$5B,$1F,$6B,$87,$F0,$7F,$43,$F8,$46,$C6,$4F,$BF
	.dc.b	$19,$69,$99,$F8,$31,$96,$C7,$E1,$BE,$DB,$F1,$FF,$07,$F9,$06,$C6
	.dc.b	$4F,$C5,$C6,$5A,$66,$7E,$6C,$65,$B1,$F9,$2F,$A8,$FC,$DF,$9F,$F9
	.dc.b	$C6,$C6,$4F,$CB,$C6,$5A,$66,$7E,$CC,$65,$B1,$F9,$EF,$81,$FD,$3F
	.dc.b	$CE,$FE,$A1,$B1,$93,$F4,$71,$96,$99,$9F,$C2,$C6,$5B,$1F,$AA,$F9
	.dc.b	$CF,$D9,$FC,$EF,$ED,$1B,$19,$3F,$5F,$19,$69,$FC,$3F,$E5,$5E,$FA
	.dc.b	$3F,$D0,$DF,$3B,$FA,$59,$7C,$06,$71,$FF,$4B,$CE,$6F,$1E,$FA,$38
	.dc.b	$56,$1D,$61,$D2,$FF,$C5,$D8,$74,$8C,$77,$A3,$C7,$18,$75,$3E,$03
	.dc.b	$FD,$51,$F3,$EE,$EF,$EE,$91,$91,$19,$F5,$7C,$10,$FA,$07,$C7,$F9
	.dc.b	$63,$9D,$F2,$D7,$D5,$18,$BE,$38,$73,$8E,$D8,$72,$CB,$38,$71,$D3
	.dc.b	$FD,$79,$5E,$28,$C3,$CC,$E3,$97,$85,$D6,$3C,$88,$EA,$EE,$5E,$48
	.dc.b	$D9,$EF,$CB,$C9,$19,$CB,$C9,$39,$C7,$C2,$F0,$9E,$55,$C3,$2F,$3D
	.dc.b	$F0,$9C,$DF,$0A,$D5,$FA,$67,$B6,$5E,$99,$CC,$F9,$0F,$F2,$2B,$F1
	.dc.b	$DB,$5F,$59,$B9,$7F,$94,$7E,$11,$F1,$BF,$E0,$79,$FF,$26,$88,$CF
	.dc.b	$8D,$B9,$D4,$FF,$94,$A7,$2F,$0C,$1E,$27,$E7,$23,$0D,$7C,$AD,$7E
	.dc.b	$BC,$70,$E2,$FF,$32,$36,$3E,$74,$F3,$D3,$E4,$B7,$6F,$BF,$F4,$0F
	.dc.b	$9B,$FC,$0D,$6B,$1F,$C4,$DE,$F0,$FC,$CB,$55,$F1,$77,$31,$F0,$BF
	.dc.b	$9A,$8D,$46,$09,$5D,$DF,$D1,$46,$A3,$04,$C3,$DF,$7E,$19,$D4,$62
	.dc.b	$FC,$15,$BB,$1E,$6C,$3D,$B5,$B8,$F0,$B6,$7F,$2D,$7D,$27,$FD,$59
	.dc.b	$EB,$BA,$78,$C6,$3A,$6F,$32,$91,$87,$84,$FF,$06,$6B,$FC,$0E,$78
	.dc.b	$4E,$CE,$15,$C7,$A7,$F6,$7E,$27,$F8,$5B,$07,$56,$3A,$3F,$C0,$51
	.dc.b	$8C,$BF,$C6,$4A,$F5,$1E,$A6,$E5,$E7,$BB,$5C,$6B,$FC,$6E,$F1,$7F
	.dc.b	$C7,$9F,$41,$F9,$1C,$7F,$C4,$EF,$F2,$C8,$CF,$5F,$48,$D6,$5F,$E1
	.dc.b	$56,$CF,$FB,$55,$E0,$FF,$91,$1B,$5D,$38,$79,$1C,$EE,$57,$8A,$3B
	.dc.b	$7A,$5D,$F8,$64,$F9,$A3,$2F,$A2,$32,$F7,$8C,$BE,$18,$CD,$FB,$27
	.dc.b	$D5,$F2,$1E,$BF,$A0,$97,$F0,$A3,$2F,$C4,$8C,$BE,$B8,$CB,$ED,$8C
	.dc.b	$D7,$C8,$9F,$8D,$F1,$9F,$1F,$E8,$25,$FC,$E8,$CB,$F4,$23,$2F,$C7
	.dc.b	$8C,$BF,$26,$33,$5F,$52,$7E,$97,$C6,$FD,$3E,$B9,$7F,$6A,$32,$FD
	.dc.b	$C8,$CB,$F4,$E3,$2F,$D5,$8C,$D7,$E0,$27,$EF,$7D,$04,$AF,$C2,$BF
	.dc.b	$81,$FA,$09,$18,$21,$F7,$DC,$B4,$FC,$6B,$F8,$9F,$83,$FC,$57,$5C
	.dc.b	$9F,$C8,$BF,$8D,$F1,$3F,$1D,$F4,$16,$7F,$2B,$18,$7F,$2F,$E2,$7E
	.dc.b	$4C,$8F,$E6,$BE,$63,$F4,$6B,$E7,$7F,$B5,$BF,$C1,$B5,$FF,$03,$39
	.dc.b	$9D,$C9,$B5,$27,$87,$47,$F6,$8A,$D4,$60,$FE,$22,$E1,$FF,$67,$1E
	.dc.b	$19,$E4,$E2,$35,$7E,$26,$D7,$5A,$FD,$77,$D0,$7E,$BF,$E8,$26,$1F
	.dc.b	$B6,$8C,$BF,$B8,$8C,$BF,$AF,$8F,$DB,$7C,$57,$F0,$2F,$77,$FB,$FC
	.dc.b	$FF,$AB,$CC,$9F,$C0,$23,$4C,$30,$7F,$AE,$BC,$9E,$5F,$1B,$23,$04
	.dc.b	$3F,$D7,$4E,$5F,$F5,$E4,$78,$4C,$69,$FE,$B2,$B5,$66,$BE,$50,$FA
	.dc.b	$0F,$F5,$73,$8D,$7F,$DB,$AD,$EB,$FB,$B3,$A8,$F9,$ED,$7D,$ED,$7F
	.dc.b	$C3,$ED,$7F,$A3,$B7,$C2,$60,$FA,$0F,$91,$F4,$12,$FF,$DB,$71,$F2
	.dc.b	$CC,$3F,$ED,$98,$C3,$EA,$FA,$07,$BB,$F9,$27,$FD,$51,$19,$F7,$31
	.dc.b	$32,$30,$43,$F1,$33,$C9,$E9,$1C,$E6,$46,$0F,$F7,$7F,$D0,$7D,$FF
	.dc.b	$99,$91,$82,$1F,$EE,$F7,$2F,$FB,$C6,$3F,$4F,$E9,$FE,$4A,$3F,$B5
	.dc.b	$FB,$CE,$5F,$D5,$B6,$3E,$BC,$ED,$FF,$2F,$74,$FF,$89,$1F,$EE,$71
	.dc.b	$FA,$37,$F2,$2D,$7F,$CA,$2D,$7F,$BC,$FD,$07,$BD,$FD,$4B,$F2,$1B
	.dc.b	$97,$EF,$23,$5C,$BF,$BA,$6F,$57,$F9,$71,$E1,$19,$9E,$1A,$CB,$FC
	.dc.b	$7D,$6A,$F8,$56,$67,$F7,$3F,$41,$FE,$40,$7F,$2B,$E1,$3E,$07,$F8
	.dc.b	$6B,$63,$F3,$33,$B7,$8F,$89,$F0,$D4,$65,$FD,$9B,$5F,$F3,$9B,$5F
	.dc.b	$F0,$DF,$8E,$F6,$FF,$8D,$FD,$07,$79,$FE,$1A,$DD,$BF,$8D,$D6,$E7
	.dc.b	$F5,$6D,$7F,$CA,$ED,$7F,$81,$B5,$FF,$41,$B5,$74,$FE,$93,$E8,$3F
	.dc.b	$C1,$BF,$31,$F9,$CF,$52,$F1,$7F,$87,$B1,$FD,$E5,$7F,$0F,$33,$33
	.dc.b	$0F,$E4,$D8,$73,$6F,$63,$E8,$3F,$8B,$12,$33,$F4,$15,$63,$FC,$26
	.dc.b	$70,$FF,$0B,$9F,$E7,$FE,$E3,$8B,$76,$F9,$F7,$6F,$FA,$83,$E8,$1E
	.dc.b	$87,$FA,$8A,$73,$5F,$E4,$A9,$CB,$FE,$4C,$8C,$BF,$E7,$FB,$EB,$18
	.dc.b	$C3,$C1,$96,$F1,$ED,$8F,$56,$E1,$93,$FC,$AF,$5C,$1F,$34,$7F,$AD
	.dc.b	$23,$17,$E0,$C7,$C7,$5A,$3C,$D7,$D5,$E7,$8F,$93,$0E,$4D,$8D,$9E
	.dc.b	$11,$5D,$2F,$D5,$3B,$9F,$76,$3B,$AD,$5F,$AA,$7C,$46,$8B,$8F,$A2
	.dc.b	$35,$87,$A7,$1F,$35,$6E,$3B,$E1,$D6,$73,$87,$6A,$F5,$CF,$9F,$0E
	.dc.b	$91,$EA,$3C,$33,$B7,$A6,$1F,$02,$BA,$44,$3C,$B8,$FB,$6F,$9C,$7C
	.dc.b	$2C,$3C,$27,$9A,$32,$6B,$0E,$C7,$77,$35,$F1,$E3,$33,$C6,$3B,$19
	.dc.b	$7A,$D7,$AE,$33,$7E,$5C,$7E,$34,$E6,$23,$EE,$AE,$2F,$AA,$33,$3C
	.dc.b	$23,$46,$31,$F8,$4E,$09,$1F,$41,$E4,$62,$3E,$B9,$C2,$34,$75,$23
	.dc.b	$92,$47,$D3,$8C,$BF,$60,$EB,$EA,$7F,$A0,$B2,$FC,$F3,$71,$FD,$AB
	.dc.b	$70,$FF,$5C,$BF,$E8,$8A,$FF,$5D,$FB,$98,$F5,$BF,$DD,$ED,$8F,$F0
	.dc.b	$DB,$5F,$F6,$D8,$9E,$4A,$3F,$41,$F2,$5F,$E1,$0E,$DF,$D1,$4F,$63
	.dc.b	$F3,$71,$97,$F3,$A6,$73,$FE,$03,$19,$7F,$81,$C6,$74,$FF,$68,$A9
	.dc.b	$D4,$7A,$23,$AF,$94,$BF,$0C,$79,$4C,$B8,$C3,$CF,$3B,$9F,$19,$BC
	.dc.b	$3C,$71,$97,$C7,$19,$BF,$1E,$1E,$28,$F8,$E7,$63,$9D,$F6,$C3,$C1
	.dc.b	$19,$BE,$73,$98,$D9,$9B,$F6,$4F,$8A,$B8,$B6,$76,$D3,$9B,$C4,$F1
	.dc.b	$B9,$9E,$67,$A3,$0F,$63,$89,$FF,$7E,$E1,$EB,$7E,$FD,$71,$79,$CF
	.dc.b	$1B,$F1,$B6,$3C,$6D,$9F,$19,$C6,$39,$C6,$4F,$03,$F6,$EF,$C6,$F6
	.dc.b	$3C,$6E,$1E,$26,$2B,$C2,$D3,$C2,$5A,$F0,$D7,$78,$E8,$77,$9E,$2D
	.dc.b	$38,$96,$78,$BC,$E7,$CF,$7F,$16,$3F,$28,$64,$EA,$FE,$0F,$8F,$FC
	.dc.b	$BF,$F3,$FF,$8D,$D7,$FC,$E8,$DC,$57,$FE,$02,$E4,$7D,$06,$63,$15
	.dc.b	$70,$E7,$7F,$DF,$23,$17,$C2,$30,$7F,$80,$3E,$7F,$97,$53,$E8,$3F
	.dc.b	$61,$9F,$FA,$16,$73,$18,$AF,$1E,$3F,$E7,$69,$CE,$FD,$9B,$3F,$E8
	.dc.b	$AF,$A0,$7C,$C7,$FA,$FB,$5F,$DA,$C6,$27,$FC,$AC,$D8,$FF,$52,$87
	.dc.b	$0F,$FA,$A1,$D6,$5F,$B3,$F4,$5F,$10,$F6,$3E,$25,$F1,$C7,$E3,$56
	.dc.b	$A7,$3F,$3E,$F5,$7C,$97,$F3,$A3,$19,$FF,$9A,$5B,$1E,$46,$BF,$EB
	.dc.b	$A7,$39,$7F,$05,$E0,$7E,$1A,$F5,$FB,$E3,$F3,$A3,$18,$FF,$B5,$FE
	.dc.b	$73,$F8,$BB,$72,$69,$F7,$67,$31,$FE,$30,$9F,$C1,$C7,$E1,$27,$53
	.dc.b	$F8,$69,$E5,$18,$8C,$B6,$32,$77,$8C,$BF,$EC,$0F,$9E,$7B,$BF,$12
	.dc.b	$51,$89,$FF,$46,$36,$3E,$BB,$5F,$F6,$EB,$9D,$3F,$94,$74,$BE,$E5
	.dc.b	$E4,$FD,$6A,$D6,$1F,$C3,$27,$2F,$F9,$53,$E7,$FF,$33,$F0,$DF,$9A
	.dc.b	$6C,$64,$FF,$2C,$C6,$5A,$7C,$DF,$8D,$FE,$7C,$DF,$96,$F9,$91,$8D
	.dc.b	$3E,$60,$DD,$09,$DD,$70,$6B,$CC,$EC,$64,$A6,$6B,$AF,$5F,$FB,$73
	.dc.b	$9F,$F6,$7F,$3F,$FB,$46,$D7,$F9,$C9,$B5,$C3,$BB,$F3,$F3,$FB,$05
	.dc.b	$5F,$9E,$F9,$E7,$CE,$7D,$76,$D7,$FB,$05,$B7,$FA,$C6,$BF,$D9,$83
	.dc.b	$AE,$3F,$E2,$5F,$54,$F8,$FE,$D8,$7F,$0F,$94,$E7,$0F,$F4,$4F,$CF
	.dc.b	$FF,$0E,$F9,$FF,$E1,$ED,$8C,$9F,$E8,$E8,$CB,$4C,$E5,$EF,$8C,$B6
	.dc.b	$3F,$D2,$B8,$7F,$1A,$F9,$FF,$E3,$6D,$8C,$9F,$E9,$A8,$CB,$4F,$F4
	.dc.b	$EF,$CF,$FF,$25,$F9,$EF,$C0,$67,$FD,$EF,$85,$C0,$AD,$5F,$F2,$86
	.dc.b	$BC,$8F,$F5,$4D,$F2,$7F,$67,$C6,$F9,$DF,$C6,$FC,$EF,$E3,$9B,$3F
	.dc.b	$93,$6E,$BF,$90,$E6,$FC,$7C,$63,$4F,$F0,$74,$62,$BF,$35,$7C,$DA
	.dc.b	$78,$B9,$BD,$87,$B7,$E7,$EC,$79,$BE,$0B,$E0,$7C,$CF,$C1,$6C,$7A
	.dc.b	$BD,$CF,$88,$63,$4F,$CB,$EC,$8F,$91,$F0,$52,$BE,$0F,$C1,$5B,$DC
	.dc.b	$FC,$DF,$85,$C6,$FF,$37,$14,$67,$0F,$92,$E9,$FE,$97,$1A,$7E,$6F
	.dc.b	$CF,$FD,$7E,$1F,$B0,$D8,$C9,$F4,$A3,$2D,$33,$8F,$AA,$32,$D8,$FB
	.dc.b	$3C,$1F,$83,$F9,$FF,$C2,$36,$32,$7D,$E8,$CB,$4C,$CF,$C6,$8C,$B6
	.dc.b	$3F,$0B,$F3,$FF,$8F,$F8,$4F,$C8,$36,$32,$7E,$2A,$32,$D3,$33,$F4
	.dc.b	$A3,$2D,$8F,$C8,$FC,$FF,$E6,$FD,$1F,$CE,$36,$32,$7E,$5A,$32,$D3
	.dc.b	$33,$F7,$A3,$2D,$8F,$CE,$FC,$1F,$E9,$FC,$1F,$D4,$36,$32,$7E,$8A
	.dc.b	$32,$D3,$33,$F8,$A8,$CB,$63,$F5,$3F,$39,$FB,$3F,$9C,$FD,$A3,$63
	.dc.b	$27,$EB,$A3,$2D,$3F,$C4,$7C,$5F,$B9,$F9,$5F,$F4,$74,$7F,$A5,$BE
	.dc.b	$6F,$FC,$2B,$0F,$81,$FF,$3C,$7C,$02,$73,$F0,$38,$77,$37,$37,$86
	.dc.b	$C3,$AC,$7F,$A6,$6B,$78,$F7,$8D,$CE,$B1,$ED,$18,$C3,$A2,$F9,$E7
	.dc.b	$6F,$F7,$E8,$E8,$7F,$8F,$A3,$3E,$A7,$58,$7C,$FB,$E6,$39,$D6,$A7
	.dc.b	$89,$D6,$FA,$63,$E2,$8C,$61,$CF,$0E,$59,$F5,$C3,$95,$ED,$BA,$7F
	.dc.b	$AA,$9B,$87,$F9,$4A,$F8,$BE,$08,$CB,$D8,$C6,$3E,$2A,$FF,$59,$D7
	.dc.b	$0B,$ED,$39,$AF,$80,$D7,$94,$7A,$67,$85,$78,$EB,$57,$E0,$3C,$33
	.dc.b	$AB,$E7,$8F,$B6,$B6,$75,$9D,$4F,$6B,$E9,$5F,$E5,$08,$F5,$1F,$D9
	.dc.b	$6F,$E7,$36,$FE,$C3,$FD,$A7,$E1,$9F,$1D,$FE,$08,$8F,$F2,$CC,$63
	.dc.b	$C7,$E8,$D8,$CC,$ED,$ED,$E2,$7F,$B9,$DF,$F3,$0E,$08,$C4,$6B,$8B
	.dc.b	$CD,$39,$6C,$F1,$D7,$ED,$37,$4F,$B4,$DB,$FF,$45,$FC,$FF,$F9,$9D
	.dc.b	$FC,$14,$65,$FC,$B3,$70,$ED,$1F,$8E,$9E,$0F,$E0,$67,$44,$C3,$F3
	.dc.b	$13,$AC,$3C,$71,$D8,$95,$CF,$DF,$FF,$7B,$CE,$A3,$A4,$EA,$3F,$35
	.dc.b	$5C,$F1,$F5,$CE,$5B,$5C,$CC,$1F,$AB,$C3,$C8,$E4,$FD,$7F,$27,$EF
	.dc.b	$7E,$63,$C3,$5F,$C2,$1B,$3F,$BC,$6C,$FF,$03,$9B,$CA,$9A,$4A,$FE
	.dc.b	$09,$5B,$6C,$FF,$98,$1F,$E3,$3C,$7F,$C5,$BD,$9F,$F5,$EC,$7F,$91
	.dc.b	$A7,$4D,$F9,$EC,$F4,$7F,$1F,$AB,$7C,$6B,$99,$FC,$4A,$32,$FF,$14
	.dc.b	$AB,$87,$F1,$BF,$8B,$B8,$9A,$63,$E7,$AD,$F4,$AE,$0D,$AE,$0D,$9D
	.dc.b	$BE,$9A,$B9,$7F,$35,$9E,$FF,$19,$2B,$93,$E2,$E4,$C4,$7C,$18,$B5
	.dc.b	$FC,$F6,$6E,$9F,$CF,$26,$D7,$F3,$09,$F5,$FC,$F4,$8F,$33,$EE,$F9
	.dc.b	$EC,$47,$CD,$8B,$87,$F4,$99,$B8,$7F,$48,$9B,$5F,$CF,$27,$E3,$FC
	.dc.b	$0C,$8F,$84,$FC,$AE,$6C,$47,$D9,$8B,$87,$F5,$59,$B8,$7F,$54,$9B
	.dc.b	$5F,$D2,$27,$E9,$EB,$23,$E7,$3F,$5B,$E7,$B1,$1F,$85,$8B,$87,$F6
	.dc.b	$19,$B8,$7F,$60,$9B,$5F,$D5,$27,$EF,$F7,$7E,$03,$E5,$E4,$FD,$B8
	.dc.b	$CB,$48,$7E,$4F,$E6,$3F,$0E,$69,$FC,$B4,$61,$FC,$27,$C3,$FE,$30
	.dc.b	$8F,$E1,$FE,$1F,$F1,$E4,$70,$7E,$37,$ED,$3F,$D2,$5A,$7F,$9C,$DF
	.dc.b	$E5,$AD,$8F,$CC,$C6,$B4,$47,$F7,$68,$DC,$7E,$AB,$E7,$FF,$DF,$93
	.dc.b	$FA,$40,$E1,$FF,$5F,$98,$BF,$D3,$AF,$95,$EC,$F6,$9E,$18,$2F,$B5
	.dc.b	$F0,$9C,$C7,$EE,$AF,$8D,$F5,$3C,$53,$A7,$A5,$78,$27,$BF,$3B,$17
	.dc.b	$C7,$FC,$02,$5F,$C7,$F9,$EA,$FF,$B3,$27,$6F,$CB,$9F,$8F,$3F,$23
	.dc.b	$E7,$A4,$FC,$F7,$E5,$FC,$EE,$23,$94,$5D,$7E,$94,$FD,$0F,$96,$FA
	.dc.b	$3F,$35,$2F,$EC,$46,$5F,$B5,$19,$7C,$91,$97,$CD,$19,$AF,$BD,$3F
	.dc.b	$C5,$BE,$3F,$E8,$7C,$2D,$DB,$FD,$B7,$87,$F1,$C9,$CB,$F8,$7F,$9A
	.dc.b	$D5,$48,$D5,$70,$7E,$24,$FD,$CF,$9F,$FF,$5A,$47,$F7,$D6,$BF,$EE
	.dc.b	$10,$E1,$FF,$72,$35,$FE,$6A,$F0,$C7,$F0,$51,$F9,$28,$C7,$1E,$BD
	.dc.b	$BF,$C2,$79,$1F,$8C,$F9,$F7,$B5,$49,$87,$E5,$61,$ED,$62,$3F,$2B
	.dc.b	$1C,$1D,$4F,$39,$D6,$5E,$29,$56,$DE,$35,$B7,$F5,$DA,$7E,$A8,$8F
	.dc.b	$EC,$7E,$66,$BF,$8B,$8B,$7F,$B3,$9F,$D2,$FA,$4F,$12,$92,$FF,$5D
	.dc.b	$F6,$8E,$4C,$47,$EB,$AB,$83,$FB,$36,$BF,$B2,$9E,$0F,$ED,$FE,$07
	.dc.b	$FC,$3F,$8F,$B5,$BC,$1F,$D9,$9B,$3F,$C0,$9F,$F3,$F3,$C7,$7F,$DD
	.dc.b	$C6,$A3,$FD,$A3,$18,$F5,$F2,$C8,$E0,$78,$63,$F8,$1F,$D9,$3D,$DA
	.dc.b	$49,$FE,$19,$F1,$CF,$33,$11,$FC,$32,$B8,$3F,$C4,$5A,$FF,$10,$9E
	.dc.b	$0F,$F1,$5F,$CA,$1D,$1F,$C4,$9B,$19,$8F,$E2,$BF,$6E,$F9,$ED,$44
	.dc.b	$9D,$1F,$C7,$E7,$4F,$F2,$D9,$FE,$52,$47,$F9,$87,$CE,$D7,$F8,$54
	.dc.b	$5E,$0F,$E6,$53,$FC,$C3,$E5,$FF,$98,$B2,$BF,$9C,$46,$5E,$64,$3F
	.dc.b	$8A,$3F,$CB,$3D,$CF,$8D,$5F,$CE,$3E,$7D,$E0,$FF,$88,$3E,$83,$FC
	.dc.b	$17,$82,$72,$FF,$3E,$8C,$C6,$21,$7D,$1E,$15,$FE,$2A,$8C,$4E,$EB
	.dc.b	$8C,$F7,$F5,$BF,$C5,$F3,$CE,$B7,$3C,$78,$58,$F9,$E3,$86,$34,$F2
	.dc.b	$61,$CA,$38,$4E,$67,$CB,$5C,$9F,$29,$C0,$D6,$3E,$47,$FA,$DE,$1E
	.dc.b	$26,$CF,$1C,$7D,$58,$F7,$9E,$47,$43,$BE,$1E,$1B,$E5,$0C,$3B,$36
	.dc.b	$B9,$3C,$0E,$53,$C2,$76,$70,$79,$99,$9E,$31,$9C,$3C,$D8,$FC,$2B
	.dc.b	$F2,$4F,$2A,$EF,$7C,$AF,$75,$EA,$C7,$DA,$FA,$63,$2F,$96,$32,$F4
	.dc.b	$AE,$73,$AA,$F2,$4F,$CC,$9C,$E3,$CC,$C4,$43,$DF,$5C,$4E,$F8,$79
	.dc.b	$E3,$81,$A9,$E1,$87,$88,$FC,$45,$71,$8C,$E3,$F8,$6A,$E0,$4A,$E1
	.dc.b	$19,$78,$CF,$67,$33,$F0,$E7,$E5,$46,$A3,$E0,$1D,$8F,$C2,$CF,$19
	.dc.b	$E8,$48,$C9,$E2,$E4,$FE,$4A,$FD,$F6,$E5,$FA,$96,$E1,$F9,$77,$73
	.dc.b	$FE,$2E,$F7,$7F,$C5,$F1,$88,$46,$39,$7F,$B5,$36,$3F,$D3,$66,$84
	.dc.b	$F1,$8F,$D3,$78,$9F,$E1,$EE,$8F,$D4,$57,$73,$F3,$91,$97,$F3,$C6
	.dc.b	$74,$FE,$05,$19,$7F,$82,$46,$73,$FF,$55,$CE,$5F,$2C,$74,$F2,$3F
	.dc.b	$BC,$D7,$08,$F3,$3D,$23,$33,$C2,$31,$7E,$6A,$DC,$F6,$AD,$C7,$C0
	.dc.b	$C3,$DB,$18,$8F,$F6,$E9,$E1,$AE,$53,$A6,$C6,$4F,$79,$D8,$FF,$2F
	.dc.b	$E1,$EB,$7B,$1B,$C3,$D5,$5C,$E3,$8B,$63,$31,$BC,$3D,$87,$A6,$FC
	.dc.b	$33,$EE,$9E,$26,$CE,$EE,$4E,$65,$9E,$71,$E6,$C3,$E0,$3F,$4C,$C5
	.dc.b	$76,$9F,$04,$76,$AE,$4F,$68,$CC,$76,$69,$DA,$18,$7C,$37,$E6,$CF
	.dc.b	$26,$CF,$07,$A1,$E2,$3C,$0E,$5D,$99,$8F,$03,$63,$C1,$3D,$E3,$D8
	.dc.b	$78,$AF,$DE,$FD,$73,$17,$EF,$79,$1E,$23,$27,$F4,$4F,$2B,$FD,$01
	.dc.b	$F3,$3F,$E4,$CD,$FF,$D4,$5F,$33,$6F,$FD,$3F,$5A,$7F,$D5,$11,$9B
	.dc.b	$FF,$51,$5D,$AF,$F4,$54,$DF,$31,$36,$31,$F3,$33,$0E,$67,$FA,$CD
	.dc.b	$C6,$1F,$E8,$D7,$84,$64,$FF,$56,$7E,$AB,$FB,$C6,$F0,$78,$FE,$13
	.dc.b	$FC,$51,$E8,$79,$70,$5E,$DF,$F5,$8B,$75,$FF,$2D,$37,$0F,$F5,$ED
	.dc.b	$F2,$7A,$9F,$EB,$BF,$D5,$9D,$3D,$5B,$1F,$EC,$4B,$FF,$3A,$D3,$EA
	.dc.b	$9F,$07,$BC,$25,$F3,$7D,$13,$A8,$C1,$CC,$FF,$67,$FC,$B7,$FB,$43
	.dc.b	$97,$E8,$36,$35,$A2,$3F,$DA,$BE,$46,$A3,$FC,$3F,$3A,$9E,$11,$FE
	.dc.b	$0E,$8E,$AE,$8C,$1C,$CF,$F3,$63,$C5,$D6,$7F,$E9,$29,$E6,$D8,$CC
	.dc.b	$F4,$AF,$15,$FD,$3F,$9E,$7A,$F4,$F8,$3E,$3D,$BF,$F7,$44,$E9,$B3
	.dc.b	$A6,$9F,$EF,$9F,$9C,$FC,$71,$F8,$FF,$9C,$B1,$92,$65,$FE,$D7,$8C
	.dc.b	$BE,$ED,$1F,$14,$F9,$8F,$9A,$DB,$F9,$0B,$CE,$CC,$C9,$CB,$C2,$31
	.dc.b	$7D,$6B,$93,$67,$55,$C0,$FB,$30,$F9,$D7,$B1,$F9,$96,$CF,$FB,$4F
	.dc.b	$EA,$1F,$15,$9A,$93,$FE,$CB,$BE,$51,$FE,$B5,$44,$BE,$6F,$FB,$45
	.dc.b	$AF,$F5,$8F,$C9,$FA,$D7,$FB,$85,$B1,$FE,$FF,$AF,$E0,$1E,$1A,$73
	.dc.b	$CB,$FE,$D4,$8C,$63,$A9,$CE,$1F,$C2,$BC,$1F,$0D,$EE,$73,$8B,$C2
	.dc.b	$53,$E7,$FF,$85,$7C,$F7,$F0,$B6,$E3,$FA,$0A,$DC,$62,$FF,$33,$18
	.dc.b	$9D,$5A,$F9,$46,$1B,$5C,$27,$53,$D7,$F6,$03,$FC,$75,$AE,$7E,$7F
	.dc.b	$0C,$38,$1B,$71,$5E,$49,$FC,$1E,$7F,$EF,$AC,$51,$8B,$FE,$5E,$FE
	.dc.b	$3D,$F1,$70,$F8,$48,$F9,$78,$EE,$1F,$8F,$F9,$9F,$CC,$9F,$9A,$F7
	.dc.b	$2D,$7E,$6A,$2C,$FE,$56,$6E,$5F,$BB,$E5,$56,$2F,$F8,$8C,$2B,$17
	.dc.b	$F5,$F2,$FE,$30,$FB,$7E,$7F,$E3,$78,$3F,$1D,$F9,$1F,$3D,$31,$F9
	.dc.b	$91,$6B,$E2,$7C,$FF,$CD,$31,$A7,$C5,$C5,$1F,$47,$DB,$FA,$5F,$3D
	.dc.b	$71,$FA,$CE,$1A,$7D,$98,$C6,$1F,$46,$7E,$D7,$C1,$E3,$1E,$CF,$DD
	.dc.b	$F9,$E9,$87,$E0,$78,$2E,$1F,$66,$6D,$F9,$66,$D7,$D8,$9B,$5F,$16
	.dc.b	$6F,$1F,$DE,$E0,$FC,$51,$F8,$BF,$84,$B8,$7E,$3E,$2D,$FE,$1E,$6E
	.dc.b	$3F,$0E,$6D,$7E,$1A,$6D,$7D,$3F,$9E,$FC,$B1,$F9,$7F,$16,$D7,$E6
	.dc.b	$E2,$CF,$E4,$E6,$E1,$F3,$E6,$D7,$E4,$A6,$D7,$DF,$F9,$EF,$D1,$1F
	.dc.b	$A3,$F9,$EB,$5F,$A7,$8B,$3F,$9F,$9B,$87,$DB,$9B,$5F,$9E,$9B,$5F
	.dc.b	$8B,$F9,$EF,$D7,$1F,$AF,$F9,$EB,$5F,$B3,$8B,$3F,$AB,$9B,$87,$E1
	.dc.b	$E6,$D7,$EA,$A6,$D7,$F0,$DF,$D8,$7F,$BF,$37,$D3,$FF,$0C,$43,$DA
	.dc.b	$B1,$2F,$F8,$74,$F0,$E0,$3B,$8C,$CF,$2F,$0F,$F8,$3E,$1F,$E9,$6F
	.dc.b	$9C,$FF,$14,$DF,$F2,$AF,$6D,$18,$36,$E2,$7C,$D5,$89,$E2,$D8,$CE
	.dc.b	$1F,$EA,$59,$E4,$D8,$C9,$E1,$3F,$A0,$FC,$F7,$F5,$9F,$53,$50,$8C
	.dc.b	$3B,$1F,$3E,$F9,$3F,$04,$6F,$C3,$5F,$2B,$E3,$9F,$A6,$B5,$3D,$B0
	.dc.b	$E5,$8E,$F0,$E3,$5D,$F0,$E3,$97,$F5,$68,$E2,$78,$EF,$8B,$BA,$E3
	.dc.b	$8F,$7B,$E3,$3E,$E9,$CD,$F3,$BE,$37,$B9,$CD,$6F,$1F,$21,$E7,$CB
	.dc.b	$DF,$8F,$49,$F1,$DF,$1A,$D6,$5C,$27,$C9,$16,$7E,$16,$3E,$0C,$BD
	.dc.b	$98,$73,$3E,$7D,$FF,$B5,$9A,$EB,$2F,$8A,$E5,$AF,$C5,$F8,$07,$CD
	.dc.b	$FF,$86,$23,$FD,$BE,$23,$3E,$2F,$F8,$0D,$B1,$A6,$BE,$1F,$17,$FD
	.dc.b	$CA,$F2,$C3,$FC,$AF,$C3,$FE,$0B,$6C,$66,$76,$F1,$9F,$C1,$47,$2D
	.dc.b	$3E,$A3,$73,$FB,$9F,$3E,$F9,$AF,$BD,$5A,$BE,$24,$C3,$F2,$55,$B7
	.dc.b	$13,$B7,$F1,$18,$75,$7F,$C3,$B3,$B7,$35,$C3,$0F,$A3,$5B,$95,$6E
	.dc.b	$FA,$12,$B7,$7F,$A6,$9D,$46,$2B,$A9,$2B,$F4,$13,$A8,$FD,$7D,$F7
	.dc.b	$C7,$E4,$D7,$59,$F3,$E3,$F0,$B0,$FF,$56,$FC,$D7,$F8,$F2,$7F,$75
	.dc.b	$CA,$8C,$47,$EE,$E3,$1C,$9C,$63,$FC,$51,$3F,$B1,$C3,$F8,$55,$6A
	.dc.b	$79,$CF,$29,$CB,$6F,$C2,$73,$AF,$07,$CF,$FE,$CF,$AB,$FC,$2D,$48
	.dc.b	$C7,$2F,$A1,$B8,$F8,$BE,$7E,$E1,$CA,$3F,$C6,$D7,$C5,$B5,$A8,$55
	.dc.b	$F8,$89,$5C,$1D,$7C,$FE,$23,$FC,$7B,$49,$C6,$5E,$6A,$B9,$7C,$C9
	.dc.b	$B7,$F1,$A7,$C1,$F1,$12,$3C,$AF,$8F,$9B,$11,$DA,$2E,$1F,$0A,$6E
	.dc.b	$1F,$62,$6D,$7D,$29,$F6,$7C,$94,$8F,$82,$FB,$FE,$7F,$11,$EA,$8B
	.dc.b	$87,$CE,$9B,$87,$E1,$26,$D7,$DE,$9F,$91,$F3,$F2,$3E,$6B,$F2,$F8
	.dc.b	$F1,$1F,$1A,$2E,$1F,$6A,$6E,$1F,$90,$9B,$5F,$8A,$9F,$A9,$AC,$8F
	.dc.b	$B2,$FD,$7F,$9F,$C4,$7D,$28,$B8,$7E,$1A,$6E,$1F,$9C,$9B,$5F,$96
	.dc.b	$9F,$C0,$77,$7E,$07,$E7,$E4,$FE,$1A,$32,$D2,$1F,$80,$F9,$FF,$C4
	.dc.b	$1A,$7F,$07,$18,$7F,$1F,$F0,$3F,$8D,$23,$F9,$2F,$89,$FC,$81,$1C
	.dc.b	$1F,$E8,$63,$F3,$3E,$C7,$F3,$2D,$3F,$CE,$CF,$F5,$58,$FF,$43,$1A
	.dc.b	$F0,$FF,$42,$BE,$7D,$78,$6C,$7F,$AF,$9A,$FF,$BB,$87,$CD,$75,$C3
	.dc.b	$FC,$13,$1C,$EB,$91,$1F,$F6,$3C,$EF,$7F,$ED,$17,$D2,$F8,$B6,$33
	.dc.b	$7E,$6A,$E1,$7F,$EA,$FC,$5F,$1F,$F0,$09,$1F,$EA,$8F,$9F,$AF,$FA
	.dc.b	$96,$2E,$5F,$12,$7E,$3F,$CF,$CA,$FF,$57,$BF,$2B,$E7,$F1,$1F,$EA
	.dc.b	$D8,$B5,$F3,$27,$E7,$FC,$57,$D0,$F8,$B9,$7F,$EB,$88,$CB,$FE,$BA
	.dc.b	$8C,$BF,$EB,$78,$CB,$FE,$B9,$8C,$D7,$D8,$9F,$E2,$BF,$09,$F3,$BE
	.dc.b	$17,$F8,$AF,$9D,$FC,$4F,$0F,$F6,$3B,$6F,$FD,$B5,$B7,$FB,$CA,$7F
	.dc.b	$DD,$31,$8A,$E7,$39,$94,$E5,$FF,$67,$FC,$C7,$E1,$3E,$7D,$EE,$7E
	.dc.b	$0F,$C9,$95,$FF,$00,$27,$2D,$9F,$F7,$D2,$8C,$BF,$DF,$9A,$FE,$4D
	.dc.b	$AF,$F8,$A7,$E7,$FF,$D6,$CF,$A3,$97,$1E,$DF,$E5,$A7,$33,$F9,$15
	.dc.b	$ED,$E6,$64,$F1,$02,$78,$E7,$F9,$67,$8C,$6C,$D4,$F2,$7B,$3E,$1A
	.dc.b	$FD,$6E,$92,$7F,$DE,$AF,$EC,$3E,$7F,$11,$FE,$F3,$8B,$87,$EC,$A7
	.dc.b	$F5,$FE,$A3,$C5,$FD,$8F,$CF,$FE,$AB,$0F,$13,$5F,$D7,$4E,$DF,$33
	.dc.b	$5C,$CF,$43,$9F,$2F,$EF,$3E,$07,$FC,$3F,$8F,$ED,$1B,$B7,$F9,$5D
	.dc.b	$B7,$F1,$9F,$F4,$53,$F0,$DC,$A6,$93,$D8,$ED,$EA,$C9,$E0,$F8,$63
	.dc.b	$F8,$6F,$CF,$BD,$EF,$E1,$FF,$3F,$E9,$CB,$F8,$3B,$72,$FF,$35,$56
	.dc.b	$F1,$FA,$8D,$7F,$88,$36,$BF,$87,$B6,$7F,$CE,$54,$F3,$BF,$8E,$FC
	.dc.b	$FB,$A7,$F8,$FF,$CF,$BE,$5A,$9A,$89,$9F,$E0,$EB,$83,$63,$8F,$CF
	.dc.b	$FF,$2E,$24,$7E,$3F,$D7,$78,$C9,$FE,$67,$F3,$FA,$AF,$D7,$EB,$F9
	.dc.b	$4B,$DD,$67,$7F,$E6,$F7,$FC,$EB,$3E,$38,$75,$C5,$39,$8C,$4E,$9E
	.dc.b	$A7,$F2,$58,$D3,$FC,$DF,$DE,$F8,$97,$FA,$4F,$9E,$78,$BF,$E9,$9F
	.dc.b	$9F,$FF,$0A,$5E,$63,$1E,$CF,$FA,$76,$B7,$3F,$AC,$C3,$9A,$8C,$BF
	.dc.b	$E7,$BE,$2F,$E1,$AE,$6F,$9B,$D4,$FD,$AE,$1E,$07,$47,$59,$E3,$5D
	.dc.b	$2F,$BB,$DA,$B8,$37,$0F,$36,$1F,$E8,$FB,$E2,$DC,$7C,$18,$B0,$E3
	.dc.b	$3E,$48,$F0,$9E,$99,$DE,$3E,$09,$D6,$3D,$EB,$38,$76,$7D,$17,$DB
	.dc.b	$0E,$B1,$EE,$C7,$CA,$E8,$EF,$8F,$57,$47,$18,$DB,$E3,$37,$7D,$DF
	.dc.b	$54,$6A,$BB,$C6,$F0,$F9,$15,$B9,$C9,$AB,$F0,$56,$F1,$F6,$91,$F3
	.dc.b	$5F,$2C,$7C,$91,$0F,$1B,$F0,$8C,$47,$C5,$8C,$91,$C1,$27,$EA,$46
	.dc.b	$5F,$03,$EC,$3A,$19,$3E,$29,$F5,$AF,$D5,$87,$C3,$C3,$D7,$1E,$EC
	.dc.b	$7E,$4C,$7B,$8C,$BF,$1C,$CC,$F1,$3E,$33,$9A,$F0,$E1,$F0,$5F,$04
	.dc.b	$65,$F2,$1F,$24,$F2,$C7,$B5,$F9,$38,$7B,$B9,$FF,$96,$C7,$DB,$6E
	.dc.b	$1F,$AA,$6E,$1F,$A8,$77,$3F,$D6,$BD,$DF,$EB,$71,$88,$46,$3A,$3F
	.dc.b	$DD,$6D,$8F,$F2,$03,$5F,$F6,$C9,$F9,$EF,$23,$FC,$53,$C7,$FA,$48
	.dc.b	$EC,$67,$CC,$46,$63,$B9,$97,$15,$FA,$CA,$E7,$3F,$AA,$39,$E9,$FC
	.dc.b	$12,$32,$FF,$06,$8C,$E9,$FD,$D9,$4E,$A3,$D7,$1D,$CE,$87,$8A,$7A
	.dc.b	$61,$E6,$8C,$DF,$8E,$31,$3D,$E7,$85,$77,$BE,$D1,$F1,$8E,$D3,$BB
	.dc.b	$E5,$87,$BA,$73,$7E,$4A,$F0,$CE,$5B,$3E,$23,$67,$57,$27,$57,$27
	.dc.b	$D4,$73,$3D,$4F,$36,$1E,$A7,$13,$FE,$1C,$C3,$D2,$F1,$8D,$DF,$8D
	.dc.b	$B1,$E3,$BE,$AF,$8E,$33,$1E,$33,$31,$CC,$ED,$7E,$17,$81,$8A,$EC
	.dc.b	$D3,$B5,$74,$3C,$EF,$DC,$9E,$4F,$53,$13,$C9,$A7,$23,$A5,$79,$6B
	.dc.b	$A4,$78,$8D,$C6,$5D,$C7,$82,$FE,$2C,$7E,$38,$C9,$C5,$FC,$0F,$27
	.dc.b	$F7,$8F,$9F,$FC,$57,$0F,$FA,$A3,$E7,$EE,$5F,$DD,$F9,$55,$8C,$7F
	.dc.b	$D2,$55,$70,$E5,$7F,$DD,$E3,$17,$C2,$30,$7F,$7B,$F9,$FC,$9A,$9B
	.dc.b	$19,$25,$FF,$7C,$BF,$F0,$6C,$EA,$3F,$D6,$5F,$3E,$E1,$F2,$42,$31
	.dc.b	$7F,$E6,$3F,$4F,$FC,$65,$E9,$6A,$73,$5C,$AF,$D9,$4F,$B5,$FF,$34
	.dc.b	$E1,$FE,$65,$BD,$ED,$FA,$FA,$D6,$1E,$0F,$A0,$FF,$16,$5F,$F9,$AD
	.dc.b	$B1,$FE,$14,$97,$D0,$3E,$1B,$87,$6C,$C3,$CD,$2F,$94,$C6,$1C,$63
	.dc.b	$13,$FE,$7D,$F9,$DF,$F3,$EF,$37,$D3,$6C,$66,$9E,$46,$5F,$F0,$FC
	.dc.b	$66,$B0,$66,$6C,$64,$93,$FE,$2C,$AF,$F4,$5C,$EA,$3F,$DC,$5F,$40
	.dc.b	$F6,$FE,$8B,$6F,$D7,$E4,$E2,$64,$7B,$BE,$42,$DF,$B5,$FC,$3F,$D0
	.dc.b	$49,$FF,$20,$7D,$05,$AF,$C1,$75,$AB,$1C,$9F,$EF,$EA,$B7,$FE,$A0
	.dc.b	$D5,$F1,$CF,$9A,$F8,$CD,$9F,$94,$BD,$1C,$CC,$9C,$BC,$23,$19,$FC
	.dc.b	$3A,$E4,$D9,$D5,$70,$3F,$D5,$9F,$3E,$F7,$3F,$9B,$57,$F3,$F1,$19
	.dc.b	$9C,$47,$FA,$E2,$30,$E1,$AF,$17,$39,$FE,$E3,$D0,$7C,$A7,$E8,$A7
	.dc.b	$37,$FE,$E2,$9C,$BF,$EC,$5F,$11,$F3,$EF,$7B,$F7,$91,$87,$F7,$39
	.dc.b	$E2,$75,$5C,$5B,$9F,$FB,$CB,$E3,$B5,$7F,$C0,$FE,$01,$1F,$C0,$BE
	.dc.b	$7F,$F8,$1F,$CF,$DC,$7F,$CF,$8D,$C3,$EE,$B7,$7F,$EB,$9E,$47,$FA
	.dc.b	$32,$31,$5F,$C4,$DB,$49,$D4,$FF,$A4,$FE,$77,$FB,$CB,$4F,$E2,$CB
	.dc.b	$E7,$EC,$43,$47,$17,$15,$FC,$46,$7E,$FE,$1E,$46,$D7,$FA,$6D,$FC
	.dc.b	$53,$E1,$E3,$EE,$47,$C9,$CB,$6F,$F1,$5F,$37,$F9,$43,$F2,$BF,$05
	.dc.b	$6B,$F2,$B1,$67,$F2,$13,$71,$FF,$00,$74,$7E,$7A,$B1,$7F,$A1,$E8
	.dc.b	$5E,$30,$F8,$D9,$FF,$11,$AF,$8D,$E1,$FC,$4F,$17,$E2,$BF,$1B,$E7
	.dc.b	$E4,$E8,$F9,$8F,$CA,$F9,$F9,$18,$3E,$83,$F3,$BD,$FF,$9F,$F3,$F6
	.dc.b	$7E,$A3,$86,$9F,$5E,$31,$97,$CF,$9F,$AF,$F1,$18,$E0,$FA,$0F,$DA
	.dc.b	$F9,$F9,$87,$DF,$E0,$B8,$7D,$79,$B7,$F5,$E6,$D7,$D7,$9B,$5F,$12
	.dc.b	$6F,$2B,$7F,$93,$7F,$8B,$E0,$FC,$4F,$C3,$FE,$29,$91,$93,$F2,$11
	.dc.b	$8C,$3F,$11,$37,$2F,$C4,$4D,$AF,$C4,$7C,$FF,$E5,$0F,$CA,$F9,$16
	.dc.b	$3F,$35,$17,$0F,$C9,$4D,$BF,$C9,$4D,$AF,$C9,$4D,$AF,$BF,$E3,$FE
	.dc.b	$80,$FD,$0F,$89,$6B,$F4,$D1,$67,$F3,$D3,$70,$FC,$F4,$DA,$FC,$F4
	.dc.b	$DA,$FC,$5F,$CF,$FE,$B0,$FD,$6F,$CF,$DA,$FD,$94,$59,$FD,$54,$DC
	.dc.b	$3F,$55,$36,$BF,$55,$36,$BF,$79,$E0,$7F,$70,$94,$63,$D4,$FE,$1D
	.dc.b	$EE,$7F,$0F,$64,$EB,$1F,$E1,$37,$C6,$F9,$DE,$9F,$F0,$64,$7F,$1A
	.dc.b	$F7,$1F,$39,$DB,$8D,$39,$8C,$3A,$C7,$FD,$29,$18,$F1,$7F,$CE,$57
	.dc.b	$CB,$0F,$4C,$62,$75,$5C,$27,$DB,$19,$77,$19,$5F,$3B,$FD,$57,$D5
	.dc.b	$C3,$63,$FD,$43,$5F,$E8,$9F,$9F,$7C,$A7,$39,$E3,$E2,$30,$E2,$D8
	.dc.b	$E2,$7C,$CC,$3A,$67,$E9,$86,$1D,$32,$E3,$84,$9E,$27,$67,$D1,$3B
	.dc.b	$8C,$DF,$4C,$BF,$D3,$CD,$8F,$06,$3E,$3C,$7C,$17,$C5,$F1,$56,$EB
	.dc.b	$B1,$D2,$B8,$5F,$C1,$CB,$E0,$56,$C9,$3A,$AF,$C4,$E1,$D6,$FC,$73
	.dc.b	$99,$EB,$87,$82,$31,$5B,$9F,$3E,$3F,$05,$FE,$F5,$87,$D3,$6D,$7C
	.dc.b	$D6,$CF,$F7,$C5,$F0,$4F,$9C,$FF,$0A,$C7,$F8,$02,$31,$E2,$ED,$B1
	.dc.b	$99,$DB,$E0,$F1,$7F,$CC,$0F,$F8,$23,$1F,$F2,$67,$27,$76,$C6,$67
	.dc.b	$6F,$19,$FC,$5C,$72,$D3,$E7,$B7,$5F,$F6,$A7,$CF,$BE,$6F,$FD,$AD
	.dc.b	$5A,$BE,$24,$C3,$F0,$75,$B7,$13,$B7,$F0,$78,$75,$7F,$77,$3B,$73
	.dc.b	$5C,$30,$FC,$3D,$6E,$55,$BB,$E8,$4A,$DD,$FE,$A6,$75,$18,$AE,$A4
	.dc.b	$AF,$CF,$4E,$A3,$F4,$57,$DF,$1F,$93,$5D,$67,$CF,$8F,$C2,$C3,$FD
	.dc.b	$0B,$F9,$67,$DF,$AF,$E0,$1C,$48,$C4,$EB,$7E,$18,$7F,$1C,$F8,$7B
	.dc.b	$7B,$9B,$7C,$29,$39,$7F,$84,$73,$7F,$14,$BE,$70,$FC,$79,$F1,$9F
	.dc.b	$C9,$A1,$D5,$FE,$50,$9C,$C7,$F2,$5C,$35,$18,$8D,$C7,$F2,$EE,$46
	.dc.b	$AF,$90,$E9,$87,$3C,$6D,$F2,$6E,$1F,$CD,$B1,$E9,$5E,$A8,$FF,$39
	.dc.b	$E1,$CD,$F1,$1B,$F9,$3F,$D3,$3E,$B6,$D1,$A4,$8E,$26,$CB,$1B,$7A
	.dc.b	$E5,$EA,$E0,$F2,$C6,$CF,$79,$E1,$38,$D7,$AA,$78,$47,$4F,$57,$12
	.dc.b	$F1,$29,$D3,$0F,$F3,$F5,$6C,$87,$63,$EB,$57,$5E,$1F,$F4,$1E,$7F
	.dc.b	$56,$39,$12,$7D,$33,$AD,$7E,$BF,$A7,$E2,$F0,$E5,$7C,$08,$F1,$11
	.dc.b	$C1,$CB,$AF,$0F,$8F,$24,$E3,$2F,$AD,$AF,$F4,$7B,$F8,$FC,$BF,$E8
	.dc.b	$CC,$3E,$5F,$E3,$BF,$05,$BB,$6B,$C4,$91,$F5,$2B,$83,$F0,$5B,$8F
	.dc.b	$6F,$C7,$7F,$DC,$53,$BA,$E2,$ED,$FA,$51,$A8,$E5,$E2,$FD,$5C,$79
	.dc.b	$EF,$F9,$76,$CF,$F8,$1B,$15,$7A,$89,$1E,$57,$F0,$7F,$06,$F7,$33
	.dc.b	$3D,$8D,$1F,$8E,$8D,$3F,$8C,$FC,$75,$F3,$3F,$87,$6C,$66,$3F,$1F
	.dc.b	$F8,$EF,$FA,$8A,$3F,$2E,$D7,$F0,$8D,$DB,$FD,$8A,$DD,$3F,$65,$E4
	.dc.b	$BE,$8F,$F2,$8D,$BE,$7E,$E7,$63,$33,$FB,$2F,$03,$FC,$33,$F4,$97
	.dc.b	$CC,$34,$F4,$7C,$D7,$9B,$C3,$98,$F9,$9E,$FE,$64,$8F,$D5,$E7,$FE
	.dc.b	$0E,$6C,$75,$F9,$3F,$D6,$FE,$61,$F1,$A7,$FD,$B6,$D7,$FD,$C0,$D7
	.dc.b	$F8,$B3,$5F,$F1,$EB,$C7,$83,$FD,$CE,$DD,$F7,$C9,$FB,$AF,$09,$F3
	.dc.b	$FE,$08,$98,$7F,$04,$BE,$0F,$F9,$22,$B6,$FF,$10,$E0,$FF,$1E,$F9
	.dc.b	$BF,$17,$C8,$93,$9E,$2F,$EE,$CD,$9C,$BF,$E4,$89,$F9,$91,$F3,$78
	.dc.b	$FE,$77,$D2,$CA,$E5,$19,$7A,$46,$5F,$F2,$8C,$7F,$14,$FA,$1F,$99
	.dc.b	$1F,$C1,$D7,$D1,$E4,$D7,$14,$9F,$E4,$33,$A6,$CE,$5E,$53,$C5,$C9
	.dc.b	$C1,$FE,$47,$F4,$2F,$0E,$92,$7A,$E3,$5F,$E3,$53,$B7,$F9,$8B,$5F
	.dc.b	$44,$7E,$3F,$CA,$FC,$87,$D2,$C9,$F9,$91,$97,$E7,$46,$5F,$F3,$9C
	.dc.b	$79,$3D,$DF,$E8,$76,$69,$27,$FC,$59,$19,$7C,$71,$C5,$F8,$BE,$9D
	.dc.b	$7D,$31,$79,$FF,$AA,$43,$5F,$F1,$2C,$7F,$57,$E2,$7E,$35,$FE,$3B
	.dc.b	$8C,$2F,$9B,$FD,$73,$6F,$F1,$8E,$FA,$7F,$D1,$4D,$85,$F4,$AE,$27
	.dc.b	$2F,$3B,$FA,$AF,$07,$F9,$C7,$1F,$2B,$71,$FE,$F8,$DC,$3F,$D4,$7F
	.dc.b	$45,$FD,$41,$FE,$9D,$19,$7E,$E7,$8F,$D0,$FE,$C3,$3B,$BE,$BC,$9F
	.dc.b	$DC,$2F,$F1,$57,$FE,$09,$3F,$D5,$EE,$1C,$35,$F1,$C7,$A0,$8F,$F7
	.dc.b	$4F,$82,$FC,$DD,$76,$8F,$EE,$F5,$DA,$31,$1F,$D9,$67,$78,$FF,$86
	.dc.b	$7E,$89,$F1,$1F,$DE,$1B,$5F,$E0,$06,$BF,$DB,$7F,$02,$7D,$0E,$EA
	.dc.b	$4E,$DF,$F0,$0C,$EA,$3F,$C1,$3F,$8E,$BC,$4F,$F0,$2B,$63,$31,$FE
	.dc.b	$08,$FC,$03,$CB,$1F,$ED,$29,$D3,$FD,$F5,$BA,$7E,$31,$B9,$FF,$8C
	.dc.b	$EB,$8F,$C8,$3F,$43,$F9,$D7,$FE,$25,$8C,$61,$FE,$2A,$6B,$FC,$4F
	.dc.b	$EA,$B1,$5C,$30,$FF,$71,$D6,$AB,$FC,$36,$DB,$FF,$70,$DF,$F4,$12
	.dc.b	$4F,$33,$F8,$F7,$7E,$1F,$F7,$1C,$59,$FF,$74,$CF,$F8,$A3,$E0,$5E
	.dc.b	$4F,$F9,$32,$DF,$46,$F9,$BF,$F2,$87,$D2,$3C,$DF,$F2,$94,$E6,$BF
	.dc.b	$DF,$33,$97,$FD,$F7,$19,$7F,$D4,$B7,$D6,$31,$87,$83,$2D,$E3,$DB
	.dc.b	$1E,$AD,$C3,$20,$E7,$38,$69,$E8,$8E,$0D,$3C,$B3,$D1,$E5,$5E,$17
	.dc.b	$47,$2B,$F4,$63,$EB,$8C,$B6,$3D,$38,$FA,$F0,$FF,$3A,$5F,$39,$E3
	.dc.b	$87,$7C,$18,$71,$BD,$E1,$C6,$7C,$6D,$77,$5C,$6B,$84,$6B,$1F,$81
	.dc.b	$8F,$7C,$3D,$F8,$F8,$2B,$38,$F3,$AD,$47,$C2,$3E,$46,$1D,$1F,$81
	.dc.b	$7E,$4C,$3E,$77,$0F,$39,$FD,$54,$FC,$C8,$F8,$33,$A8,$E1,$3F,$1B
	.dc.b	$2F,$41,$F0,$70,$F0,$C7,$63,$C4,$6C,$F3,$D4,$7B,$9A,$BF,$11,$E0
	.dc.b	$72,$73,$3E,$39,$EC,$AF,$06,$3F,$04,$F9,$98,$FB,$CF,$81,$87,$90
	.dc.b	$F2,$B9,$9E,$38,$78,$F2,$F8,$44,$CB,$DE,$F8,$CC,$63,$EC,$71,$1A
	.dc.b	$3E,$A9,$23,$ED,$1F,$90,$C3,$C3,$7F,$22,$3C,$13,$C5,$E8,$F1,$7E
	.dc.b	$04,$FE,$67,$C1,$FE,$2E,$DC,$FE,$9C,$63,$3F,$DE,$3F,$C7,$2B,$F5
	.dc.b	$5E,$FF,$EA,$E3,$11,$FE,$C4,$0E,$23,$55,$FD,$26,$75,$39,$7F,$53
	.dc.b	$BF,$F6,$9E,$0F,$DA,$47,$63,$3D,$DF,$EC,$7D,$7F,$82,$46,$5F,$E0
	.dc.b	$D5,$C6,$3F,$66,$67,$4F,$F7,$04,$66,$78,$CF,$9D,$E3,$5D,$6B,$CD
	.dc.b	$97,$F0,$5B,$E1,$18,$BF,$1E,$1E,$67,$94,$76,$BE,$51,$FD,$9E,$F9
	.dc.b	$46,$23,$FD,$E5,$18,$AE,$4D,$8F,$49,$D0,$FE,$77,$3C,$5F,$61,$89
	.dc.b	$E2,$F9,$0E,$33,$CF,$0F,$7B,$6F,$D5,$97,$C0,$BF,$D9,$5F,$9C,$FD
	.dc.b	$2E,$3F,$2E,$78,$C7,$5C,$29,$C5,$EE,$66,$FE,$34,$64,$E3,$3B,$2C
	.dc.b	$7C,$7B,$F0,$D7,$C4,$39,$1E,$3A,$E5,$3D,$0E,$11,$E0,$8C,$BB,$8C
	.dc.b	$C7,$99,$A7,$18,$F9,$07,$17,$2F,$13,$31,$C5,$B1,$C5,$5F,$E1,$1E
	.dc.b	$06,$31,$FA,$CF,$D9,$31,$7F,$5A,$FC,$F1,$EF,$6B,$F7,$0D,$9F,$D4
	.dc.b	$7B,$DE,$57,$FA,$02,$BF,$D2,$5D,$A7,$FD,$29,$F5,$56,$DC,$5F,$E5
	.dc.b	$7B,$E1,$18,$AF,$EC,$58,$74,$C3,$85,$ED,$B3,$AB,$61,$8C,$3F,$CA
	.dc.b	$F1,$D2,$FF,$C2,$DC,$CF,$A6,$D6,$BF,$E1,$B7,$86,$7D,$F2,$FE,$E5
	.dc.b	$7C,$27,$77,$C6,$72,$7F,$86,$9C,$63,$E9,$3C,$39,$F0,$AF,$24,$7F
	.dc.b	$9C,$7E,$7A,$EC,$7D,$1E,$1B,$E8,$B1,$97,$6B,$FF,$02,$5E,$1E,$C7
	.dc.b	$38,$D3,$4F,$15,$73,$BE,$E7,$8B,$E2,$9F,$2B,$4F,$0F,$D0,$3C,$4C
	.dc.b	$63,$D2,$B6,$DA,$DC,$EB,$1F,$1F,$D0,$3C,$FF,$7B,$6F,$D5,$F7,$4F
	.dc.b	$BA,$7C,$C6,$AE,$57,$AA,$F8,$61,$F2,$F0,$E1,$7A,$8E,$F7,$FE,$A2
	.dc.b	$7F,$C5,$02,$B8,$BF,$E8,$36,$BA,$6B,$FE,$2F,$F9,$F7,$BF,$F5,$9B
	.dc.b	$8F,$D1,$E0,$6E,$F8,$5D,$5C,$AF,$A3,$7C,$30,$FA,$37,$B9,$CD,$7E
	.dc.b	$3B,$D0,$7D,$05,$F7,$30,$DA,$FF,$59,$B7,$0F,$C6,$75,$FF,$AF,$23
	.dc.b	$1D,$7F,$EC,$2C,$DF,$19,$E9,$3F,$3B,$F5,$CE,$7C,$65,$E6,$9B,$97
	.dc.b	$F9,$C2,$B5,$39,$24,$7F,$63,$BE,$B1,$FD,$96,$11,$87,$55,$F4,$DF
	.dc.b	$D8,$FD,$1F,$EB,$3C,$07,$D3,$58,$C3,$70,$FC,$F6,$BF,$C2,$A3,$17
	.dc.b	$F0,$2F,$8D,$7E,$7B,$F0,$07,$CA,$7E,$B1,$B1,$F5,$FF,$00,$FF,$52
	.dc.b	$3B,$9D,$11,$D7,$91,$70,$FD,$C7,$D0,$BE,$43,$FD,$F1,$5F,$B1,$8C
	.dc.b	$7C,$FF,$F0,$D7,$5E,$F5,$A4,$D8,$C3,$6B,$94,$E9,$B3,$A6,$9A,$9F
	.dc.b	$E2,$7F,$23,$FE,$83,$8F,$8A,$BB,$33,$8E,$9B,$3A,$6C,$7E,$22,$78
	.dc.b	$4F,$2F,$17,$AF,$C4,$76,$6C,$64,$EF,$19,$69,$F4,$70,$FA,$07,$E3
	.dc.b	$BE,$2E,$C6,$49,$87,$E3,$63,$2F,$F1,$8E,$0F,$84,$FF,$8B,$DA,$FF
	.dc.b	$BE,$5A,$FE,$F1,$AF,$FA,$97,$E8,$DF,$13,$FC,$B1,$B7,$F1,$BE,$BF
	.dc.b	$13,$FC,$C3,$F1,$9F,$F1,$87,$6F,$F5,$1C,$11,$8B,$FE,$13,$87,$29
	.dc.b	$FE,$5B,$39,$5D,$0D,$FF,$7D,$87,$F5,$B8,$C7,$83,$FD,$47,$8D,$EE
	.dc.b	$73,$AD,$5F,$0C,$FB,$FD,$6F,$97,$8B,$D4,$D8,$FE,$77,$A2,$71,$A7
	.dc.b	$F4,$0D,$57,$8C,$7F,$B6,$67,$E0,$8F,$16,$7A,$CB,$9E,$38,$9F,$89
	.dc.b	$F1,$D2,$7F,$A4,$FE,$34,$F9,$FC,$69,$F0,$AA,$E7,$F2,$6B,$57,$E6
	.dc.b	$9C,$BF,$ED,$9C,$DF,$80,$FE,$CB,$0E,$53,$FD,$6E,$73,$D7,$F5,$7E
	.dc.b	$16,$4F,$F5,$AF,$D0,$3F,$89,$DF,$D2,$9C,$B6,$75,$E3,$63,$4F,$B9
	.dc.b	$C4,$BF,$A5,$9B,$ED,$35,$87,$96,$30,$FF,$68,$FC,$6D,$F5,$3F,$DD
	.dc.b	$9B,$1E,$7D,$9F,$1B,$9A,$93,$C5,$FC,$04,$EA,$30,$4D,$7F,$C0,$90
	.dc.b	$F8,$2F,$89,$AF,$E4,$F1,$EF,$5A,$9C,$DF,$F5,$C9,$FE,$27,$3F,$E0
	.dc.b	$06,$D7,$E0,$2D,$F0,$9E,$9F,$0E,$61,$ED,$7C,$B1,$8C,$39,$47,$E5
	.dc.b	$F1,$F8,$71,$8A,$FA,$D1,$8C,$3C,$DC,$BF,$14,$FC,$F2,$FB,$1C,$57
	.dc.b	$CB,$9B,$7F,$02,$B5,$39,$24,$7F,$AB,$F3,$FC,$5B,$6B,$FA,$52,$8C
	.dc.b	$BF,$DF,$87,$D2,$FF,$AE,$AB,$FC,$42,$D8,$F0,$7A,$1F,$E3,$08,$C6
	.dc.b	$1F,$C0,$5B,$5F,$E0,$86,$D7,$F4,$19,$57,$1F,$85,$FF,$26,$5B,$F0
	.dc.b	$3F,$3F,$C3,$7E,$13,$0F,$F2,$F4,$63,$8B,$FC,$B3,$E0,$3E,$1B,$FC
	.dc.b	$C1,$7B,$C3,$96,$D8,$BF,$E7,$D8,$71,$C7,$9C,$E5,$FE,$23,$8F,$3B
	.dc.b	$FF,$6E,$DF,$0A,$D5,$3E,$95,$F2,$1F,$D1,$E1,$97,$9D,$D4,$66,$31
	.dc.b	$BB,$E2,$F9,$DF,$2F,$37,$BE,$03,$8B,$86,$9D,$B2,$F9,$95,$CB,$1E
	.dc.b	$D1,$FE,$36,$BE,$95,$B3,$A3,$8C,$7C,$77,$C7,$0E,$D3,$9B,$E5,$87
	.dc.b	$4C,$3C,$98,$FA,$73,$DC,$E6,$7C,$B9,$7C,$1C,$39,$47,$0C,$BB,$3E
	.dc.b	$33,$78,$F6,$CB,$DD,$97,$59,$D5,$F2,$76,$7A,$B1,$F0,$4E,$F1,$F0
	.dc.b	$47,$8E,$BE,$0D,$7E,$6D,$AF,$73,$F3,$C7,$CC,$3F,$41,$F1,$4F,$97
	.dc.b	$FF,$0B,$46,$A3,$1D,$3F,$E3,$56,$C6,$67,$6F,$7F,$45,$C7,$FE,$3B
	.dc.b	$C3,$FC,$56,$D8,$CC,$F8,$A7,$2D,$8C,$E5,$F2,$DB,$AF,$DD,$6C,$FE
	.dc.b	$1F,$E9,$9F,$2D,$FE,$76,$BE,$18,$FD,$CB,$DD,$FE,$65,$B7,$C2,$35
	.dc.b	$31,$FA,$33,$AB,$E9,$87,$86,$B5,$87,$6B,$FB,$98,$73,$78,$CE,$AB
	.dc.b	$84,$6A,$F9,$BF,$94,$C3,$CB,$8F,$C2,$C7,$A1,$FA,$BC,$7A,$38,$C3
	.dc.b	$9B,$6F,$E1,$61,$D3,$85,$F4,$DF,$B5,$9F,$D1,$46,$38,$75,$39,$BF
	.dc.b	$E1,$DE,$6F,$FB,$BB,$BB,$35,$30,$FD,$AE,$1C,$5F,$DB,$D6,$DF,$DB
	.dc.b	$7D,$23,$E9,$BF,$CF,$50,$8C,$3F,$E7,$D6,$BF,$92,$6B,$FC,$C5,$CF
	.dc.b	$67,$F8,$AD,$B7,$FE,$B2,$6C,$FF,$9A,$F2,$7C,$E7,$FA,$DA,$F8,$7A
	.dc.b	$37,$0F,$F0,$47,$CC,$FF,$27,$C3,$FC,$0C,$D7,$FD,$BF,$8F,$F9,$01
	.dc.b	$FF,$6F,$B5,$F3,$57,$B2,$31,$1E,$78,$C1,$DD,$FE,$35,$1C,$6F,$C7
	.dc.b	$1F,$E1,$5C,$7F,$0D,$1E,$DF,$77,$87,$47,$F9,$43,$B6,$E3,$FE,$43
	.dc.b	$C3,$C7,$5D,$5F,$98,$F8,$C8,$FC,$D6,$BE,$C7,$86,$3F,$EB,$A8,$FF
	.dc.b	$13,$6F,$FE,$57,$C7,$E7,$7D,$2F,$83,$E9,$A5,$7C,$28,$E2,$7C,$12
	.dc.b	$3F,$6D,$AF,$CA,$76,$F6,$EC,$F4,$73,$7F,$B7,$63,$FC,$6D,$F4,$BE
	.dc.b	$D7,$FC,$32,$DC,$3F,$CD,$53,$B9,$46,$3E,$1B,$F1,$B3,$99,$FE,$F0
	.dc.b	$D7,$FD,$EE,$D7,$FA,$63,$CE,$79,$47,$F4,$F6,$BF,$E2,$AF,$A6,$FF
	.dc.b	$60,$B9,$E3,$E7,$F4,$DF,$5B,$1E,$51,$FE,$53,$6C,$F6,$9F,$79,$E6
	.dc.b	$23,$FE,$DD,$C1,$18,$8F,$6D,$76,$7C,$07,$32,$8F,$A6,$FF,$04,$BF
	.dc.b	$E5,$FF,$A6,$7B,$7F,$8A,$6C,$66,$3F,$11,$F4,$CF,$7E,$9F,$B6,$71
	.dc.b	$3F,$EF,$16,$BF,$D1,$9A,$FF,$75,$3F,$DB,$5E,$BB,$E9,$39,$C7,$9A
	.dc.b	$78,$EF,$FE,$88,$AD,$4F,$FB,$CB,$E1,$5F,$4D,$FA,$7F,$01,$18,$9D
	.dc.b	$4F,$AF,$DB,$F4,$F8,$5E,$72,$4F,$9D,$FE,$B6,$F9,$3C,$CF,$D5,$E3
	.dc.b	$FE,$0F,$8F,$F4,$23,$5F,$F8,$01,$E9,$FE,$AF,$E9,$BE,$5C,$7D,$23
	.dc.b	$FC,$25,$DF,$F2,$9C,$1F,$EC,$4E,$1E,$1E,$94,$8F,$0D,$72,$7F,$C9
	.dc.b	$4D,$7E,$43,$5F,$F0,$F3,$E0,$DF,$FD,$90,$DE,$5D,$D7,$EF,$F0,$FF
	.dc.b	$5B,$5F,$C9,$F3,$7E,$47,$07,$F2,$08,$F9,$9E,$36,$26,$46,$08,$7B
	.dc.b	$78,$BE,$7C,$72,$E4,$E7,$F2,$12,$BE,$B4,$65,$FB,$11,$97,$96,$1F
	.dc.b	$71,$F8,$BE,$3F,$CD,$DB,$FD,$C4,$BE,$67,$F8,$C3,$67,$F3,$6E,$F6
	.dc.b	$FC,$63,$63,$FC,$84,$ED,$EB,$3C,$B8,$E4,$72,$8E,$A6,$1F,$AF,$F4
	.dc.b	$BF,$E5,$67,$BF,$D3,$78,$8F,$E5,$51,$98,$DF,$85,$82,$3E,$6D,$3E
	.dc.b	$7F,$9B,$F4,$2E,$4F,$E6,$E3,$2F,$E7,$63,$2F,$C6,$8E,$FE,$CF,$C2
	.dc.b	$F7,$E4,$FE,$F3,$14,$EA,$3F,$37,$18,$6C,$7F,$6C,$8C,$47,$F8,$E9
	.dc.b	$BB,$36,$FD,$B4,$61,$B5,$F4,$BD,$3F,$D8,$7C,$8F,$F8,$16,$F8,$60
	.dc.b	$E4,$4B,$E2,$B6,$70,$F2,$E5,$7F,$16,$AD,$CE,$A3,$15,$D7,$2F,$E1
	.dc.b	$1C,$1F,$E8,$5C,$7B,$D3,$E9,$9F,$11,$FD,$75,$B1,$FC,$82,$B8,$5F
	.dc.b	$2F,$23,$FB,$14,$62,$7F,$D1,$0D,$7A,$79,$FA,$E5,$FE,$D1,$F3,$0F
	.dc.b	$9F,$A6,$6B,$B4,$FF,$86,$2F,$F2,$11,$FE,$C3,$6B,$EA,$8D,$1C,$5F
	.dc.b	$E0,$B1,$F9,$7D,$3D,$B7,$F9,$B9,$7D,$33,$E9,$BF,$BD,$7D,$37,$F3
	.dc.b	$B9,$FE,$7A,$D7,$F8,$7C,$ED,$FE,$82,$D7,$33,$FE,$03,$8C,$5F,$CC
	.dc.b	$A7,$C3,$FF,$81,$FE,$99,$E4,$7F,$82,$7E,$3B,$E1,$65,$FD,$31,$AF
	.dc.b	$F1,$E9,$DB,$FD,$41,$AE,$67,$A1,$CF,$BF,$FC,$41,$F4,$6F,$A3,$FF
	.dc.b	$20,$B6,$BF,$9A,$B5,$FF,$48,$B5,$FF,$78,$37,$3F,$C9,$B7,$93,$FD
	.dc.b	$88,$D9,$FC,$B7,$86,$FA,$6F,$E4,$64,$8F,$E9,$CF,$F2,$6C,$71,$1F
	.dc.b	$C6,$22,$E1,$FA,$0E,$BF,$EA,$35,$FE,$2D,$F8,$37,$9B,$FE,$10,$C3
	.dc.b	$E4,$FC,$3F,$F6,$1F,$A6,$7A,$1F,$D8,$F8,$3F,$C9,$51,$8A,$FF,$6D
	.dc.b	$47,$F9,$37,$0E,$79,$E3,$0F,$ED,$51,$8A,$7D,$37,$2C,$7F,$B7,$36
	.dc.b	$3F,$DC,$8F,$3C,$BC,$8D,$C3,$A4,$FF,$BA,$72,$F1,$3E,$83,$17,$E7
	.dc.b	$C7,$C3,$87,$B7,$2E,$55,$E2,$9F,$75,$78,$9B,$1E,$27,$8E,$16,$3F
	.dc.b	$DF,$19,$7A,$F2,$F4,$47,$33,$C2,$6C,$46,$EB,$9C,$6C,$65,$EA,$AE
	.dc.b	$67,$98,$F3,$E3,$D8,$95,$C9,$D3,$E8,$AE,$95,$CE,$3B,$D6,$A7,$CA
	.dc.b	$6B,$1F,$80,$FB,$32,$F9,$F5,$D4,$E9,$87,$82,$75,$97,$D2,$7D,$B5
	.dc.b	$C1,$EB,$87,$AA,$FC,$24,$8C,$BE,$78,$F2,$47,$2A,$E9,$8F,$DC,$C3
	.dc.b	$C1,$39,$3E,$F1,$6B,$B3,$D0,$F2,$9F,$37,$2F,$A0,$E4,$F9,$F9,$7C
	.dc.b	$07,$96,$3E,$57,$27,$D9,$C3,$E5,$3D,$2F,$E9,$CF,$C4,$24,$78,$30
	.dc.b	$F5,$D7,$08,$E6,$78,$F2,$F8,$CF,$AE,$7D,$B3,$EF,$AF,$A1,$3C,$DE
	.dc.b	$AF,$29,$F0,$BE,$EF,$0F,$F8,$BE,$5F,$AD,$6E,$1F,$BB,$6E,$6A,$47
	.dc.b	$F1,$E7,$1F,$03,$98,$C4,$23,$18,$7F,$98,$E3,$67,$27,$0E,$CC,$1B
	.dc.b	$8F,$C3,$70,$FF,$6D,$E2,$FC,$44,$F2,$AF,$F6,$A4,$E6,$3E,$E9,$9C
	.dc.b	$FF,$CE,$51,$99,$FF,$6D,$4E,$4F,$E4,$FE,$1F,$F7,$C8,$CC,$78,$F6
	.dc.b	$EB,$3F,$E0,$07,$38,$FE,$D6,$FC,$11,$8A,$FC,$6C,$77,$C3,$DB,$5E
	.dc.b	$AB,$ED,$18,$AF,$DD,$57,$28,$C5,$F8,$31,$FE,$05,$5F,$22,$30,$FF
	.dc.b	$01,$31,$3E,$EB,$F3,$3C,$8C,$57,$29,$F6,$46,$1E,$47,$D1,$C7,$A9
	.dc.b	$EB,$2C,$73,$8F,$8F,$3C,$F1,$F8,$57,$E0,$9F,$B0,$7B,$1C,$3F,$EA
	.dc.b	$38,$DC,$F1,$8F,$8F,$3F,$11,$E3,$7E,$07,$9E,$1F,$00,$F9,$C5,$C3
	.dc.b	$C9,$0C,$7E,$53,$C2,$3D,$D3,$EB,$7E,$05,$59,$F5,$9F,$88,$3E,$29
	.dc.b	$89,$EF,$3C,$8E,$F5,$DA,$7D,$87,$BF,$0F,$65,$79,$63,$C4,$73,$AF
	.dc.b	$C0,$5F,$B6,$38,$B5,$FB,$66,$CE,$27,$B7,$0F,$96,$FE,$64,$C5,$FE
	.dc.b	$91,$FB,$55,$EE,$3F,$71,$C3,$FE,$C8,$ED,$FF,$2C,$EB,$FD,$EF,$93
	.dc.b	$FD,$21,$87,$FB,$3F,$BB,$2F,$08,$CE,$3F,$E9,$AB,$CB,$FE,$0C,$6E
	.dc.b	$1F,$ED,$4B,$B5,$C3,$0F,$F0,$AF,$33,$E9,$B8,$EF,$FE,$21,$77,$8E
	.dc.b	$27,$59,$FF,$BF,$2B,$4F,$F8,$8A,$33,$9F,$F8,$89,$AF,$EF,$9A,$FF
	.dc.b	$BC,$75,$7D,$2F,$A2,$3F,$CE,$CF,$AB,$C6,$7D,$36,$B0,$FE,$3D,$A7
	.dc.b	$F9,$C3,$0C,$61,$FE,$DC,$6C,$7F,$A0,$AF,$8F,$C2,$F5,$9C,$DF,$F2
	.dc.b	$79,$CB,$D0,$FE,$55,$6F,$2D,$F4,$DE,$56,$CE,$E5,$4F,$A2,$78,$BB
	.dc.b	$C2,$5E,$EB,$5F,$23,$8C,$3F,$DE,$D1,$89,$E9,$0F,$A3,$7A,$5F,$15
	.dc.b	$B3,$E1,$FA,$57,$D3,$3E,$9B,$57,$2B,$F9,$CE,$1C,$E3,$FD,$2E,$FB
	.dc.b	$B0,$E6,$FF,$3E,$6B,$FD,$0D,$CF,$0F,$F9,$6B,$EE,$5F,$31,$D6,$93
	.dc.b	$98,$FB,$5F,$4A,$F8,$0A,$7D,$BE,$27,$C5,$63,$4F,$7D,$6D,$F8,$D2
	.dc.b	$F8,$8D,$65,$F1,$A3,$13,$FE,$B4,$FA,$7F,$CE,$FB,$0F,$82,$B1,$86
	.dc.b	$CF,$FA,$E9,$B7,$E8,$8F,$D8,$D7,$17,$C7,$EB,$7A,$8F,$D1,$FD,$3F
	.dc.b	$E0,$E3,$24,$AF,$EC,$B9,$AB,$E6,$C7,$F7,$38,$C4,$FD,$26,$C7,$09
	.dc.b	$FD,$BE,$EF,$B0,$F8,$B7,$D6,$3F,$81,$C6,$2B,$D0,$7F,$A2,$3E,$A3
	.dc.b	$F7,$7F,$51,$FB,$C6,$DF,$FB,$6F,$C6,$46,$2B,$8F,$90,$9C,$D6,$EF
	.dc.b	$F6,$5F,$4A,$F9,$9A,$7E,$CF,$EA,$3F,$D6,$35,$F9,$69,$D3,$67,$5E
	.dc.b	$17,$08,$FD,$05,$7C,$9C,$BD,$3F,$50,$F9,$87,$C1,$FF,$18,$8C,$7D
	.dc.b	$45,$9C,$E1,$FA,$F9,$CB,$63,$37,$D1,$57,$1B,$E4,$DA,$E0,$DA,$E0
	.dc.b	$FF,$26,$F9,$9F,$F1,$54,$FE,$F7,$74,$7E,$FA,$30,$EA,$BE,$5B,$96
	.dc.b	$9C,$FC,$4E,$8D,$9E,$3E,$47,$7F,$A7,$F0,$36,$32,$76,$8C,$B4,$FA
	.dc.b	$18,$FD,$63,$F2,$BF,$27,$63,$27,$F0,$0E,$1F,$86,$D9,$FC,$AE,$09
	.dc.b	$CB,$FE,$3E,$6F,$37,$F2,$9E,$07,$C2,$F5,$AF,$E2,$F1,$8C,$3F,$8B
	.dc.b	$FC,$3B,$E3,$FF,$9D,$36,$3E,$3F,$D4,$62,$7F,$9F,$7D,$2F,$F9,$0F
	.dc.b	$0F,$F2,$83,$5F,$E6,$2D,$7F,$CB,$8D,$E2,$FA,$7E,$2F,$D7,$3F,$AF
	.dc.b	$F2,$BE,$13,$C1,$7F,$52,$7C,$7F,$51,$E0,$E3,$BA,$7F,$2E,$C9,$19
	.dc.b	$8F,$E4,$F1,$86,$C7,$C8,$8C,$5F,$F9,$9F,$1F,$05,$F8,$B1,$3C,$E7
	.dc.b	$F6,$F3,$AC,$3C,$1F,$29,$F0,$8F,$EC,$1F,$50,$FA,$3C,$4E,$8F,$1C
	.dc.b	$F8,$48,$E0,$FE,$A5,$4F,$A7,$FD,$AE,$1F,$C1,$86,$6B,$F8,$F3,$AE
	.dc.b	$46,$1F,$43,$E2,$BE,$A9,$FD,$CF,$EA,$3F,$71,$3A,$C7,$E9,$CE,$AB
	.dc.b	$E3,$BF,$14,$8E,$0F,$2D,$FF,$48,$7E,$84,$78,$BC,$EF,$F0,$07,$D3
	.dc.b	$BD,$6A,$7D,$CD,$7F,$1D,$8F,$AE,$75,$87,$F6,$66,$C7,$FA,$59,$AF
	.dc.b	$FA,$6D,$AF,$F7,$C6,$BF,$EA,$77,$8E,$5E,$DA,$7C,$37,$C5,$BE,$18
	.dc.b	$7E,$3A,$31,$87,$1F,$07,$D2,$79,$FE,$26,$DF,$96,$3F,$5F,$97,$F1
	.dc.b	$6A,$F2,$46,$09,$3F,$6A,$FE,$34,$FB,$E7,$C1,$9F,$C6,$E1,$97,$F8
	.dc.b	$0D,$1F,$51,$8C,$3E,$35,$78,$09,$5F,$E0,$2C,$BF,$1C,$D8,$FD,$4F
	.dc.b	$94,$F9,$EE,$58,$FC,$38,$C5,$7F,$85,$3E,$9F,$FD,$9A,$DF,$8A,$FD
	.dc.b	$8D,$7F,$69,$CB,$F2,$15,$A6,$CF,$F1,$08,$C4,$7F,$B3,$DA,$FF,$8C
	.dc.b	$1A,$FE,$53,$EA,$3C,$EF,$FB,$5F,$C2,$FE,$6B,$18,$F8,$BF,$E2,$B2
	.dc.b	$EF,$FE,$6F,$F1,$3F,$97,$9C,$E6,$BE,$18,$72,$DB,$4D,$9F,$F7,$3B
	.dc.b	$FE,$74,$F2,$98,$6B,$2F,$25,$D5,$F5,$0F,$95,$FF,$21,$C7,$9C,$FF
	.dc.b	$7A,$46,$7C,$CE,$B1,$FE,$F7,$D9,$F2,$1C,$EB,$53,$C4,$EB,$7D,$36
	.dc.b	$F7,$C6,$30,$E6,$B0,$E7,$AE,$30,$E7,$7C,$1B,$AF,$F9,$79,$B7,$E3
	.dc.b	$CB,$C5,$97,$7A,$D3,$67,$D6,$D7,$C1,$3C,$32,$F0,$4E,$6F,$AD,$6B
	.dc.b	$1C,$D4,$8F,$74,$EB,$2F,$4C,$62,$F9,$12,$BD,$F3,$AC,$BD,$93,$9A
	.dc.b	$FF,$3C,$9E,$6A,$FF,$30,$DF,$CC,$8C,$B5,$FF,$3F,$FC,$7B,$E7,$3F
	.dc.b	$C2,$91,$C8,$46,$7B,$7F,$CD,$6D,$8D,$35,$EB,$E9,$39,$3C,$2D,$8E
	.dc.b	$98,$7F,$8B,$67,$24,$8F,$87,$3B,$CF,$EC,$37,$7F,$F4,$AF,$D4,$3E
	.dc.b	$6B,$FD,$F1,$1F,$6E,$75,$7F,$90,$6B,$F9,$66,$D7,$18,$FC,$34,$6B
	.dc.b	$1F,$24,$78,$B1,$FC,$6D,$6E,$30,$4C,$7C,$B3,$AC,$7C,$73,$E1,$8E
	.dc.b	$A4,$7F,$39,$5C,$EB,$8C,$73,$C7,$D5,$5C,$E3,$84,$E8,$93,$F0,$AF
	.dc.b	$BF,$03,$EA,$3F,$63,$7F,$B1,$8C,$6F,$A9,$CE,$1F,$C2,$39,$FF,$CC
	.dc.b	$5E,$06,$6A,$61,$FB,$6C,$38,$BF,$B8,$AD,$BF,$B0,$FA,$87,$C5,$FF
	.dc.b	$B2,$3A,$FF,$A5,$7C,$F7,$F8,$32,$BF,$BA,$B5,$FE,$E9,$5C,$1F,$F5
	.dc.b	$FB,$5F,$C8,$35,$FC,$DB,$5F,$F6,$AB,$50,$7F,$C6,$4F,$83,$AB,$C9
	.dc.b	$8E,$B9,$57,$FC,$72,$BC,$3E,$7E,$9E,$78,$72,$7F,$C3,$33,$B6,$E1
	.dc.b	$B7,$C1,$56,$B8,$63,$FE,$07,$9C,$9D,$8F,$F0,$BE,$DF,$8B,$8D,$4F
	.dc.b	$FA,$A6,$7F,$C3,$75,$FE,$70,$6F,$67,$5F,$72,$46,$08,$7C,$C7,$2D
	.dc.b	$3E,$6B,$ED,$CF,$FC,$ED,$1F,$4B,$E9,$F1,$6F,$A8,$A7,$AE,$24,$F9
	.dc.b	$23,$2F,$D8,$8C,$BC,$CC,$1F,$E3,$4E,$1F,$E8,$8F,$FB,$F3,$97,$FD
	.dc.b	$C2,$FF,$8E,$3E,$9F,$DC,$FF,$91,$9B,$AF,$C9,$9D,$CF,$F8,$ED,$B1
	.dc.b	$FE,$17,$9D,$F0,$FD,$D6,$D7,$E1,$2B,$73,$F8,$68,$FF,$3D,$FC,$3F
	.dc.b	$D9,$9C,$C6,$E3,$E3,$7D,$47,$FB,$06,$3D,$1F,$11,$F2,$70,$DE,$C7
	.dc.b	$8C,$FF,$A1,$A7,$A1,$BB,$91,$C8,$87,$AE,$36,$FE,$6B,$2F,$C4,$78
	.dc.b	$DF,$EF,$B8,$FF,$2E,$FD,$43,$D7,$FB,$4D,$8C,$C7,$E4,$FE,$A1,$EF
	.dc.b	$53,$71,$27,$B6,$FF,$0E,$B7,$1F,$E6,$AF,$9A,$7D,$2F,$18,$D9,$FA
	.dc.b	$E9,$DB,$FE,$7C,$9D,$BF,$EE,$FF,$A8,$FD,$57,$77,$FB,$B5,$BF,$0C
	.dc.b	$7D,$3E,$C7,$9F,$C1,$B7,$EE,$8C,$3E,$3F,$33,$F3,$18,$FF,$81,$A3
	.dc.b	$07,$FB,$F7,$D4,$FD,$97,$D4,$7C,$78,$FF,$40,$76,$FC,$56,$F0,$6B
	.dc.b	$E3,$24,$7B,$0F,$F5,$C4,$7E,$EE,$78,$72,$FE,$E6,$FA,$36,$FF,$B2
	.dc.b	$5F,$F1,$09,$F0,$3F,$DC,$1A,$FE,$3D,$AF,$F8,$D7,$6F,$ED,$CF,$FB
	.dc.b	$B5,$7D,$47,$C6,$3E,$3F,$9D,$27,$51,$97,$84,$7C,$CE,$3F,$9B,$F2
	.dc.b	$52,$7A,$46,$5E,$D1,$95,$97,$FA,$6F,$8D,$8F,$DA,$9F,$F5,$47,$8F
	.dc.b	$F3,$DB,$F3,$9F,$C7,$E0,$DE,$29,$3F,$2E,$72,$D9,$CB,$D2,$79,$44
	.dc.b	$8E,$11,$FC,$AF,$E9,$FF,$83,$C7,$8B,$EA,$3C,$06,$7D,$39,$1F,$4A
	.dc.b	$76,$FF,$88,$1A,$FE,$FD,$AF,$D1,$F3,$BF,$27,$52,$3E,$94,$7E,$58
	.dc.b	$C3,$F1,$BA,$FF,$A6,$E9,$F1,$AB,$53,$F8,$9B,$7B,$7F,$17,$E3,$24
	.dc.b	$E6,$3D,$86,$1F,$9B,$18,$74,$7A,$E3,$F1,$AA,$7A,$B9,$3F,$C6,$96
	.dc.b	$F9,$1F,$EF,$FC,$1F,$6D,$7C,$67,$EC,$9B,$5F,$BA,$AE,$1E,$4E,$4D
	.dc.b	$54,$9D,$57,$4D,$3F,$94,$6D,$FE,$70,$C7,$FA,$E7,$D4,$3E,$2B,$FB
	.dc.b	$13,$70,$FD,$65,$6E,$F8,$F9,$7E,$C8,$FF,$3C,$CF,$F9,$59,$AE,$8F
	.dc.b	$D6,$BF,$8A,$7F,$DF,$5F,$36,$FA,$1F,$EE,$0D,$8E,$FC,$BF,$DC,$EB
	.dc.b	$86,$7F,$DC,$6B,$55,$CA,$F5,$72,$3B,$BE,$52,$3F,$BE,$FA,$77,$D4
	.dc.b	$7F,$7F,$FA,$8F,$F6,$34,$FF,$09,$6B,$F8,$73,$FC,$27,$3C,$1F,$E1
	.dc.b	$CD,$7F,$CD,$8D,$7F,$CE,$ED,$7F,$AB,$35,$FF,$0E,$FD,$43,$C5,$FF
	.dc.b	$08,$7C,$77,$B3,$4F,$F0,$73,$77,$FE,$BD,$5B,$BF,$E3,$CD,$7F,$CF
	.dc.b	$8D,$7F,$D0,$ED,$78,$DF,$F1,$EF,$A4,$7D,$27,$FA,$33,$1F,$E1,$1F
	.dc.b	$AA,$FE,$32,$FE,$07,$42,$31,$2C,$BF,$C6,$B8,$72,$BF,$F1,$73,$5F
	.dc.b	$F4,$73,$5F,$F4,$CB,$5F,$EF,$0D,$7F,$67,$E0,$3E,$A7,$FA,$69,$23
	.dc.b	$F9,$67,$81,$5F,$2C,$ED,$FE,$61,$3F,$D5,$27,$FC,$75,$F0,$8F,$2F
	.dc.b	$F9,$2D,$3E,$A3,$FC,$81,$F5,$2F,$3F,$FC,$15,$18,$6F,$C0,$7F,$AA
	.dc.b	$AF,$8C,$FF,$91,$F0,$E9,$B7,$F9,$5B,$8B,$86,$57,$0F,$F0,$A6,$5D
	.dc.b	$63,$07,$F9,$72,$F8,$BF,$81,$9E,$D7,$FE,$60,$C3,$9C,$F9,$E7,$9E
	.dc.b	$5E,$48,$E5,$97,$92,$78,$C7,$33,$FC,$D3,$8F,$85,$C0,$71,$97,$AE
	.dc.b	$BE,$2B,$71,$F1,$5F,$82,$E9,$D3,$0F,$13,$FE,$CA,$BF,$13,$E4,$AD
	.dc.b	$BD,$E7,$78,$78,$F1,$ED,$3E,$D3,$81,$4E,$07,$28,$E1,$87,$BA,$78
	.dc.b	$E3,$F1,$89,$1F,$09,$EB,$1F,$28,$91,$C4,$95,$EB,$3B,$5F,$B3,$2F
	.dc.b	$AB,$5D,$AB,$E1,$57,$CA,$C3,$E4,$C7,$73,$9C,$7A,$0F,$54,$7A,$1F
	.dc.b	$6C,$F9,$27,$9B,$E4,$8C,$C7,$29,$E1,$1C,$E7,$83,$A3,$D0,$76,$BF
	.dc.b	$46,$5F,$4A,$3E,$A6,$5E,$C6,$1F,$37,$1F,$6E,$3D,$EF,$EA,$9C,$AB
	.dc.b	$EC,$CE,$AB,$BE,$1E,$77,$A9,$CF,$0F,$06,$3E,$83,$D3,$5C,$67,$75
	.dc.b	$C6,$73,$5E,$7A,$FD,$1F,$C1,$FF,$1B,$6E,$7F,$9C,$8C,$4F,$EE,$9F
	.dc.b	$E3,$9F,$03,$FE,$C6,$6E,$26,$92,$BF,$D0,$D1,$FE,$CC,$9F,$CA,$F0
	.dc.b	$7F,$6C,$E1,$FC,$BC,$E4,$FC,$94,$65,$FC,$A1,$9C,$FF,$D9,$11,$97
	.dc.b	$FD,$95,$19,$DB,$F8,$EF,$85,$FC,$9E,$38,$8C,$3C,$31,$9A,$E3,$39
	.dc.b	$3C,$31,$9A,$F4,$C6,$67,$C3,$39,$8F,$6D,$74,$C3,$D7,$1B,$BF,$1C
	.dc.b	$62,$75,$3E,$46,$DF,$89,$B5,$E0,$6C,$F6,$3C,$B5,$CD,$A7,$32,$D7
	.dc.b	$36,$D7,$36,$9C,$F1,$F5,$D7,$F2,$68,$FF,$5E,$5D,$C7,$E9,$57,$3C
	.dc.b	$37,$5D,$DA,$67,$0F,$7C,$65,$B1,$E3,$9E,$25,$8D,$BC,$67,$FA,$EE
	.dc.b	$3F,$35,$F1,$9E,$82,$D7,$82,$BD,$8D,$8F,$8C,$D7,$71,$E2,$3E,$15
	.dc.b	$78,$9A,$78,$8E,$66,$E3,$9B,$5F,$1C,$71,$7E,$54,$71,$76,$7A,$B1
	.dc.b	$FA,$D5,$E5,$E4,$FF,$42,$F5,$FE,$57,$83,$FD,$11,$C9,$FE,$82,$BF
	.dc.b	$F4,$D7,$66,$5F,$EE,$B5,$71,$FF,$50,$37,$0E,$73,$C0,$FF,$51,$46
	.dc.b	$F5,$FF,$07,$FD,$4F,$E8,$EF,$FC,$03,$D0,$8C,$47,$F8,$32,$31,$7F
	.dc.b	$E0,$A6,$C7,$F8,$29,$B9,$FF,$A7,$FF,$39,$B5,$C6,$FC,$91,$8B,$1F
	.dc.b	$E6,$AF,$AA,$EE,$E3,$85,$18,$8D,$5F,$FB,$12,$72,$D9,$F8,$5E,$33
	.dc.b	$EA,$9F,$0D,$F1,$70,$E2,$7F,$92,$67,$51,$9F,$2D,$F5,$5D,$DB,$8F
	.dc.b	$FB,$49,$B7,$FC,$01,$AF,$F1,$A7,$3A,$FF,$0D,$C1,$F4,$AF,$94,$ED
	.dc.b	$8F,$9A,$31,$8F,$FB,$97,$E9,$DE,$9F,$BD,$B1,$FE,$E2,$6E,$7F,$C4
	.dc.b	$1A,$FF,$2A,$73,$97,$F1,$DE,$37,$D3,$BE,$5B,$E1,$56,$B0,$FF,$23
	.dc.b	$CE,$5F,$16,$3F,$85,$8C,$63,$FE,$6F,$15,$C5,$61,$DB,$0F,$9D,$F4
	.dc.b	$AF,$77,$EC,$37,$2F,$F7,$DB,$6F,$F9,$4B,$5F,$E8,$6E,$73,$F9,$5A
	.dc.b	$3E,$13,$AE,$1F,$25,$FC,$F7,$A6,$FA,$7B,$EF,$61,$B9,$FF,$AB,$9B
	.dc.b	$9F,$F9,$AA,$B8,$4F,$FA,$CA,$BD,$3F,$19,$E8,$24,$7A,$1B,$5F,$E9
	.dc.b	$1C,$DF,$2C,$F8,$8E,$37,$F9,$C9,$FD,$D4,$E6,$30,$7E,$82,$9B,$30
	.dc.b	$FD,$D5,$F4,$C3,$9D,$77,$C3,$93,$FE,$6E,$FA,$6F,$DF,$78,$6F,$A6
	.dc.b	$B1,$86,$E7,$FE,$D3,$9D,$36,$74,$F9,$2B,$F6,$7F,$48,$F9,$6F,$D8
	.dc.b	$36,$7F,$AF,$CA,$30,$FF,$62,$6B,$FD,$F1,$CE,$7F,$A3,$FA,$47,$6F
	.dc.b	$EC,$E7,$37,$FC,$0B,$B5,$5A,$8F,$27,$D3,$3E,$43,$FE,$00,$C6,$17
	.dc.b	$D3,$7F,$21,$9D,$7C,$15,$9C,$D6,$E3,$15,$96,$C6,$4E,$91,$96,$9F
	.dc.b	$EF,$BF,$95,$FE,$2E,$DC,$7F,$97,$46,$35,$FF,$5A,$A8,$CC,$69,$B3
	.dc.b	$C4,$E9,$E3,$72,$FA,$5E,$6D,$8C,$9E,$18,$CB,$4F,$B1,$5F,$4F,$0F
	.dc.b	$AE,$48,$FA,$FA,$FC,$29,$FA,$D3,$FC,$9B,$57,$C0,$FD,$38,$C3,$6F
	.dc.b	$F0,$99,$7D,$48,$C5,$FF,$29,$BE,$35,$E2,$F8,$37,$C6,$7F,$2F,$6C
	.dc.b	$7C,$9F,$A6,$C4,$FF,$33,$FA,$4F,$F1,$D7,$27,$F8,$CF,$89,$F0,$1C
	.dc.b	$33,$FE,$21,$5A,$C7,$FA,$3D,$6B,$0F,$A3,$E1,$49,$FC,$D4,$E5,$FB
	.dc.b	$11,$E3,$FA,$6F,$07,$15,$9F,$93,$82,$33,$1C,$B2,$56,$32,$67,$E7
	.dc.b	$9C,$D6,$EB,$FB,$96,$5E,$A7,$85,$6D,$F8,$1F,$21,$2B,$EC,$FD,$33
	.dc.b	$E8,$31,$97,$C0,$AE,$7F,$4A,$F6,$F3,$97,$C7,$8C,$63,$EA,$AD,$D6
	.dc.b	$A3,$E3,$CE,$AF,$E5,$FD,$2F,$D1,$3F,$B2,$7D,$37,$EB,$E3,$35,$F1
	.dc.b	$27,$C3,$E6,$6E,$3E,$6D,$FF,$44,$CB,$FC,$11,$C2,$BD,$D7,$13,$77
	.dc.b	$FD,$C7,$E9,$7F,$AF,$E3,$FD,$2D,$B8,$7E,$07,$83,$F1,$73,$9F,$4F
	.dc.b	$81,$F3,$3E,$07,$D5,$5F,$E1,$7B,$E3,$8F,$E1,$27,$2F,$F2,$D8,$CD
	.dc.b	$FF,$09,$9C,$BF,$E1,$08,$CE,$09,$CC,$A7,$F8,$7B,$E7,$F0,$3E,$ED
	.dc.b	$3E,$13,$D9,$13,$0F,$C8,$E1,$F1,$5F,$14,$7A,$73,$D6,$1F,$8A,$9F
	.dc.b	$44,$63,$2F,$D1,$5F,$1A,$F3,$71,$7C,$93,$F1,$B2,$FA,$6C,$61,$FC
	.dc.b	$0B,$4F,$C5,$3F,$A2,$F2,$5F,$39,$AC,$7C,$07,$F2,$B8,$CC,$F1,$C7
	.dc.b	$F3,$31,$89,$FA,$D1,$8A,$FF,$15,$7D,$2F,$FA,$EE,$FF,$A1,$B6,$BF
	.dc.b	$D4,$6D,$7F,$D7,$AD,$7F,$B0,$3B,$CB,$F8,$06,$0F,$B6,$EB,$F0,$DE
	.dc.b	$27,$FC,$B9,$83,$E9,$FD,$B8,$FF,$AD,$FC,$3F,$E6,$71,$8F,$87,$FE
	.dc.b	$37,$2F,$84,$79,$4F,$87,$FF,$60,$56,$B1,$E9,$C1,$8C,$B9,$B7,$14
	.dc.b	$E8,$FF,$6B,$4F,$F9,$63,$1E,$B8,$6E,$75,$5E,$53,$FC,$BB,$F4,$EF
	.dc.b	$93,$FE,$89,$19,$11,$9F,$2B,$AC,$7F,$99,$38,$9F,$19,$E3,$8D,$DF
	.dc.b	$2D,$7D,$F1,$8B,$E2,$D3,$B5,$F2,$C7,$38,$F7,$9F,$13,$6F,$9B,$74
	.dc.b	$FA,$F9,$78,$63,$15,$EC,$CB,$C5,$5F,$E1,$B9,$CC,$B1,$F0,$3D,$AB
	.dc.b	$6F,$1C,$BC,$71,$88,$E1,$87,$78,$C5,$F4,$3A,$DF,$4C,$7C,$F3,$A2
	.dc.b	$61,$F0,$A7,$58,$F9,$67,$38,$FF,$BE,$30,$F8,$4F,$F8,$CC,$F9,$05
	.dc.b	$72,$FF,$8D,$FE,$2D,$F3,$5F,$E1,$38,$FF,$80,$31,$8E,$DF,$F1,$2B
	.dc.b	$63,$33,$B7,$A7,$A6,$E5,$FF,$15,$E1,$FE,$2F,$6C,$66,$7C,$CD,$78
	.dc.b	$C7,$DA,$8D,$6B,$F6,$1B,$9F,$F9,$37,$E9,$DF,$33,$FE,$F8,$8F,$AD
	.dc.b	$3A,$C3,$F2,$AD,$7F,$26,$DA,$E3,$1D,$F0,$8F,$E2,$27,$57,$D6,$39
	.dc.b	$CE,$B0,$F2,$3F,$94,$BE,$73,$C6,$35,$87,$91,$D5,$F3,$9F,$1C,$69
	.dc.b	$EA,$7E,$A6,$B9,$DF,$68,$E8,$47,$F4,$B1,$AB,$FD,$C7,$2F,$FB,$32
	.dc.b	$BF,$5F,$C9,$FC,$1B,$E1,$BD,$4D,$8C,$CE,$DF,$DE,$47,$F0,$CF,$88
	.dc.b	$DC,$FF,$B3,$A7,$75,$89,$CF,$9F,$FE,$7F,$EC,$FE,$17,$5F,$C5,$6F
	.dc.b	$FC,$4E,$FF,$93,$67,$5C,$FE,$6F,$2A,$CE,$E6,$E5,$FC,$71,$B7,$CA
	.dc.b	$7F,$CA,$35,$AF,$C6,$3F,$D5,$11,$FE,$7D,$B4,$62,$3F,$A1,$C6,$3E
	.dc.b	$0B,$87,$67,$16,$4F,$1A,$E1,$9F,$AA,$F9,$DE,$AB,$9D,$6B,$3F,$E4
	.dc.b	$CD,$8F,$E6,$8F,$F8,$7D,$FF,$36,$79,$1E,$A2,$3F,$E6,$FF,$8B,$AF
	.dc.b	$F9,$B2,$2E,$3E,$1D,$15,$8C,$3E,$ED,$5B,$F9,$D3,$F2,$3F,$18,$FF
	.dc.b	$48,$C7,$C3,$6B,$F0,$9B,$87,$B5,$FE,$5A,$DF,$6F,$91,$27,$97,$15
	.dc.b	$79,$45,$9E,$53,$F5,$BC,$49,$5F,$E8,$B7,$EC,$FA,$38,$8F,$F4,$4C
	.dc.b	$5A,$F7,$4D,$CF,$F3,$33,$6B,$F1,$53,$ED,$F2,$E4,$7B,$5F,$81,$F4
	.dc.b	$18,$8F,$74,$5C,$3D,$D3,$70,$FE,$B9,$9F,$F8,$76,$F9,$1F,$1E,$7F
	.dc.b	$1F,$E4,$FE,$43,$C7,$93,$FE,$A5,$8C,$B4,$87,$FA,$93,$E1,$7F,$2E
	.dc.b	$47,$FD,$58,$FE,$6B,$E1,$3E,$B7,$CF,$7E,$73,$E1,$25,$7D,$78,$FB
	.dc.b	$5F,$3F,$23,$ED,$3F,$77,$E3,$3E,$F7,$85,$27,$EF,$3F,$A6,$FA,$6F
	.dc.b	$D5,$C7,$FA,$5B,$BE,$F7,$FE,$F7,$8B,$F6,$B3,$93,$FD,$5B,$1F,$BD
	.dc.b	$7F,$C5,$26,$4A,$F1,$E2,$FF,$7E,$36,$07,$FA,$93,$C0,$EF,$72,$3A
	.dc.b	$BE,$28,$FD,$DF,$BD,$FB,$EF,$D8,$BC,$78,$FE,$BD,$B8,$FF,$1A,$9D
	.dc.b	$E1,$FB,$D9,$CB,$FE,$21,$6F,$81,$FE,$22,$9F,$83,$F4,$4E,$2F,$F7
	.dc.b	$DD,$A3,$0F,$FB,$F9,$AF,$F6,$53,$51,$C2,$33,$8F,$C8,$6E,$3F,$02
	.dc.b	$B6,$7C,$AF,$7B,$D9,$F6,$BE,$DF,$9E,$93,$E9,$8C,$BE,$B8,$CB,$FC
	.dc.b	$5B,$44,$EA,$31,$3A,$C7,$E3,$4F,$F2,$8F,$8F,$7E,$B6,$F6,$F0,$DC
	.dc.b	$3E,$AD,$FF,$34,$AD,$D7,$D1,$F4,$B6,$CA,$DE,$9F,$72,$B7,$1F,$46
	.dc.b	$BE,$C7,$9D,$F6,$7E,$4E,$4F,$D5,$8C,$BF,$5E,$32,$FB,$63,$F2,$BF
	.dc.b	$49,$FE,$93,$8F,$F7,$EB,$5F,$C8,$B7,$2F,$CB,$BF,$E1,$48,$DF,$B9
	.dc.b	$C8,$8F,$2F,$4E,$BC,$A2,$D7,$29,$FD,$2F,$D1,$FC,$5A,$FC,$03,$5F
	.dc.b	$F1,$33,$5F,$BC,$D7,$F3,$EF,$19,$FD,$5B,$74,$DE,$BE,$EF,$4F,$BF
	.dc.b	$D5,$4C,$3E,$69,$C9,$FD,$BF,$DE,$BE,$07,$53,$FB,$2C,$55,$8C,$BF
	.dc.b	$73,$7B,$AE,$63,$E9,$7F,$DA,$CE,$1A,$FE,$F1,$B3,$FC,$09,$FE,$9B
	.dc.b	$1B,$F4,$3F,$84,$C6,$B4,$FC,$F3,$67,$F7,$EF,$07,$A4,$F0,$23,$E1
	.dc.b	$F4,$3F,$89,$FE,$1E,$F5,$A9,$27,$D3,$F8,$43,$ED,$B1,$1F,$C4,$AB
	.dc.b	$83,$EE,$6B,$FC,$5A,$78,$3F,$0B,$E3,$7F,$D5,$CD,$F3,$BF,$91,$C6
	.dc.b	$23,$F8,$EF,$C1,$FF,$93,$F6,$FE,$47,$39,$6D,$70,$3F,$95,$CE,$F6
	.dc.b	$FF,$52,$77,$BE,$E7,$F8,$D1,$31,$FE,$2D,$F6,$75,$FD,$E4,$5C,$7F
	.dc.b	$8F,$4F,$F2,$1F,$AA,$95,$FC,$75,$FE,$49,$CD,$88,$FE,$09,$16,$BF
	.dc.b	$93,$4F,$F3,$EF,$81,$FF,$7E,$47,$F3,$2F,$A4,$FF,$61,$5F,$F3,$4F
	.dc.b	$A5,$FE,$BC,$A3,$2A,$33,$18,$AE,$73,$BF,$88,$FF,$42,$77,$30,$ED
	.dc.b	$1F,$C8,$70,$E2,$DF,$67,$F9,$26,$3C,$EF,$86,$1C,$CF,$E8,$F8,$F6
	.dc.b	$9F,$1E,$1C,$A7,$D1,$7C,$B0,$F0,$E3,$E0,$9E,$38,$F4,$C3,$8D,$EE
	.dc.b	$78,$9B,$18,$F7,$C3,$9C,$78,$CF,$85,$3C,$5A,$71,$9E,$0F,$83,$0E
	.dc.b	$98,$F9,$E3,$67,$39,$DB,$E0,$C7,$CD,$87,$C5,$8E,$38,$77,$AD,$9C
	.dc.b	$2F,$A1,$D5,$CD,$F2,$AF,$69,$E9,$9D,$46,$63,$C1,$87,$66,$1E,$58
	.dc.b	$D5,$77,$BF,$81,$7C,$AF,$E0,$57,$38,$F1,$CF,$9E,$38,$99,$78,$99
	.dc.b	$9E,$27,$83,$1F,$AB,$86,$0F,$14,$67,$1F,$89,$18,$24,$F0,$7D,$D5
	.dc.b	$D5,$CC,$F0,$9D,$1F,$8F,$AE,$CE,$23,$47,$E0,$1D,$4F,$43,$C4,$7D
	.dc.b	$D3,$A9,$F3,$9E,$4E,$5E,$3D,$AF,$8F,$FC,$5B,$67,$93,$5F,$A6,$7F
	.dc.b	$B8,$E3,$FC,$1D,$F0,$9F,$E2,$89,$D4,$FF,$8A,$A3,$13,$A8,$E0,$7F
	.dc.b	$BE,$63,$2D,$3F,$35,$E3,$3E,$5B,$FD,$63,$87,$EF,$A7,$A1,$A8,$CE
	.dc.b	$3D,$67,$F8,$1D,$71,$E4,$FD,$45,$F0,$9C,$DF,$08,$F0,$46,$27,$70
	.dc.b	$C3,$F6,$17,$CE,$3F,$5C,$73,$C3,$FC,$51,$19,$BF,$34,$F1,$3C,$D7
	.dc.b	$D2,$F9,$E1,$E1,$9F,$F1,$84,$62,$75,$1C,$0E,$E7,$3A,$F3,$38,$9E
	.dc.b	$0D,$38,$47,$8A,$32,$D8,$CE,$1F,$1C,$F4,$DF,$8E,$3D,$87,$F8,$9D
	.dc.b	$C3,$E1,$8D,$DF,$9A,$BC,$67,$52,$C7,$59,$E5,$3B,$8E,$31,$E7,$9D
	.dc.b	$BF,$4A,$33,$1C,$DA,$73,$86,$1F,$3E,$30,$D8,$E4,$78,$0E,$D1,$C9
	.dc.b	$C9,$CA,$FD,$B8,$7C,$69,$E7,$18,$9F,$8E,$70,$AE,$CF,$D8,$32,$5B
	.dc.b	$F5,$B6,$BC,$0D,$3C,$91,$ED,$6C,$7F,$72,$F0,$98,$FF,$A1,$5B,$3F
	.dc.b	$E6,$BF,$95,$BA,$FF,$A3,$66,$D7,$13,$81,$FE,$8D,$9C,$D7,$FA,$36
	.dc.b	$6F,$A0,$E7,$7D,$CE,$1B,$87,$FA,$57,$BB,$13,$23,$FD,$30,$EE,$31
	.dc.b	$87,$FA,$BB,$F1,$5F,$F1,$05,$7F,$8E,$9B,$1F,$DB,$E3,$FD,$5B,$F1
	.dc.b	$8F,$A3,$7B,$FE,$6A,$5F,$0D,$BF,$D7,$AD,$C3,$FD,$2C,$D9,$61,$CE
	.dc.b	$F9,$4E,$65,$F0,$AF,$89,$EC,$DC,$3F,$CB,$38,$3E,$87,$E0,$DF,$FB
	.dc.b	$22,$B4,$DB,$E2,$7C,$CA,$E1,$AE,$30,$FF,$31,$FC,$DF,$FB,$2B,$C3
	.dc.b	$FA,$2D,$8C,$E0,$F8,$3C,$AA,$A4,$9D,$CD,$9F,$94,$2B,$A3,$86,$ED
	.dc.b	$F8,$87,$FD,$B7,$3C,$63,$FD,$CF,$F4,$0F,$2B,$E8,$37,$0F,$87,$B3
	.dc.b	$E3,$33,$52,$7F,$CE,$B5,$BC,$BC,$7A,$FC,$9B,$D4,$7F,$BC,$25,$F0
	.dc.b	$BD,$8F,$C9,$F5,$BE,$D7,$EA,$E3,$D2,$74,$FF,$B6,$6D,$7C,$A3,$0D
	.dc.b	$8F,$F6,$F3,$5F,$F6,$D6,$CF,$90,$78,$FF,$49,$B7,$F5,$9F,$F4,$B3
	.dc.b	$63,$44,$75,$E3,$DB,$F9,$D5,$C9,$B5,$D6,$75,$C1,$FF,$00,$7E,$37
	.dc.b	$F5,$9F,$73,$FA,$D6,$E3,$FF,$00,$5B,$87,$FB,$9D,$B5,$FE,$DD,$FD
	.dc.b	$71,$F5,$7F,$21,$B7,$FE,$B0,$EF,$FD,$D7,$DD,$66,$A4,$FF,$AD,$38
	.dc.b	$97,$C3,$2F,$F7,$A5,$E9,$FF,$7B,$7A,$FF,$C0,$7E,$97,$F8,$13,$70
	.dc.b	$ED,$E1,$7F,$AE,$1B,$3A,$69,$A9,$E1,$5D,$23,$2F,$E6,$07,$D1,$3C
	.dc.b	$5F,$E0,$0D,$9F,$47,$3B,$EC,$F3,$52,$7F,$09,$5B,$B1,$0A,$E5,$C1
	.dc.b	$E0,$9F,$E5,$6E,$F7,$FE,$45,$E0,$7F,$21,$EA,$FE,$44,$DC,$3E,$0B
	.dc.b	$6B,$F6,$35,$C2,$7E,$2D,$7E,$37,$8E,$47,$E3,$79,$2D,$FE,$22,$6F
	.dc.b	$27,$ED,$79,$D5,$8B,$F5,$F4,$7E,$66,$F1,$87,$EF,$EF,$75,$F9,$99
	.dc.b	$FE,$77,$F8,$AB,$DF,$A7,$F3,$EF,$CC,$DE,$7E,$2F,$F1,$14,$F8,$1E
	.dc.b	$27,$2F,$82,$B3,$AF,$5F,$97,$A5,$D9,$B1,$F9,$1C,$7F,$73,$AF,$C4
	.dc.b	$6F,$07,$F5,$8E,$24,$3E,$5B,$C9,$E5,$79,$7E,$7A,$63,$CE,$2E,$3C
	.dc.b	$E7,$D5,$D3,$8C,$BA,$63,$F8,$5C,$3D,$7E,$17,$C2,$3E,$1F,$A1,$63
	.dc.b	$24,$C7,$EC,$F0,$A7,$51,$8A,$B8,$7A,$6B,$57,$C6,$3C,$3E,$AF,$CD
	.dc.b	$3E,$77,$CA,$59,$F8,$91,$70,$F8,$93,$6B,$8D,$68,$F9,$51,$97,$F3
	.dc.b	$7B,$B0,$FE,$BC,$DF,$4B,$EF,$F3,$FE,$01,$95,$F7,$1C,$B4,$98,$FC
	.dc.b	$08,$CB,$63,$ED,$7A,$FF,$87,$3F,$11,$F0,$96,$32,$49,$FB,$D1,$97
	.dc.b	$F0,$13,$73,$FA,$B3,$6B,$D7,$97,$F0,$9B,$FA,$3E,$17,$E6,$3C,$3F
	.dc.b	$CC,$B2,$BF,$2A,$E5,$A4,$9E,$2F,$E6,$A7,$84,$7E,$47,$E4,$FF,$40
	.dc.b	$7E,$87,$E5,$2D,$7E,$5E,$2C,$FE,$5E,$6E,$5F,$8C,$F9,$4F,$F1,$7F
	.dc.b	$B7,$F9,$7A,$70,$7E,$6B,$E5,$33,$D5,$F5,$B9,$D8,$62,$F7,$0A,$C5
	.dc.b	$FE,$8A,$AC,$E7,$4F,$F5,$36,$1C,$E7,$37,$CF,$07,$CD,$3B,$BF,$C3
	.dc.b	$F1,$FE,$95,$BF,$E4,$13,$9F,$4B,$9C,$FF,$24,$F9,$B7,$CE,$73,$AD
	.dc.b	$4F,$13,$F9,$4B,$8D,$BE,$04,$78,$31,$EF,$18,$BE,$7A,$76,$AD,$63
	.dc.b	$70,$E5,$97,$93,$1E,$2F,$F3,$5C,$38,$DF,$84,$E2,$5A,$E2,$F9,$2B
	.dc.b	$8B,$EB,$CB,$D5,$18,$8E,$39,$7A,$2B,$8C,$2F,$BC,$E6,$B8,$56,$8D
	.dc.b	$CE,$B0,$F4,$63,$EB,$AD,$9D,$E7,$58,$79,$67,$47,$F8,$5E,$FE,$6C
	.dc.b	$7F,$50,$6B,$D0,$FE,$A6,$3E,$05,$CD,$FE,$17,$8C,$C6,$3C,$6F,$F1
	.dc.b	$13,$63,$33,$B7,$B7,$8D,$FC,$F9,$EB,$7F,$D0,$FC,$6F,$24,$E6,$77
	.dc.b	$18,$3E,$C4,$71,$D7,$E5,$37,$5F,$ED,$9F,$39,$F9,$BB,$DD,$C7,$EC
	.dc.b	$56,$DF,$EE,$73,$B7,$33,$BB,$EA,$FE,$26,$75,$18,$9D,$91,$FC,$9C
	.dc.b	$6A,$FF,$3F,$7D,$EB,$F4,$AD,$C3,$E0,$47,$E3,$6B,$83,$F9,$E9,$D1
	.dc.b	$30,$F1,$70,$FF,$BC,$B0,$EB,$1E,$76,$D7,$39,$D3,$FC,$13,$B5,$F2
	.dc.b	$D8,$E6,$43,$E5,$B0,$DA,$FD,$AC,$EA,$B2,$6A,$A4,$EA,$BA,$B7,$E5
	.dc.b	$FF,$C3,$DD,$3F,$DB,$65,$18,$E7,$FE,$93,$18,$AF,$F2,$32,$F8,$BD
	.dc.b	$47,$A6,$AE,$5E,$86,$D7,$19,$FF,$0F,$BC,$5F,$F2,$87,$CE,$7F,$AA
	.dc.b	$DB,$E8,$72,$8C,$47,$F8,$99,$FF,$45,$3A,$21,$98,$E1,$E8,$71,$64
	.dc.b	$6B,$3F,$F0,$55,$72,$8F,$1F,$C6,$F9,$88,$FB,$BD,$DA,$E8,$F8,$31
	.dc.b	$EC,$8C,$D7,$C9,$8B,$AF,$C8,$9F,$87,$F3,$5F,$EC,$8B,$FF,$37,$35
	.dc.b	$FF,$20,$B6,$FF,$C3,$CF,$F2,$86,$F9,$DC,$89,$3C,$B8,$EB,$CA,$2C
	.dc.b	$F2,$9F,$A3,$F3,$92,$BE,$CB,$F5,$3E,$73,$11,$F5,$22,$D7,$E0,$E6
	.dc.b	$E5,$F8,$29,$B5,$F9,$79,$F6,$F9,$92,$3D,$AF,$C0,$F3,$31,$1E,$E8
	.dc.b	$B8,$7B,$A6,$E1,$FE,$16,$CD,$58,$BF,$EC,$F0,$AF,$C5,$FC,$37,$E3
	.dc.b	$3E,$12,$57,$E5,$23,$2D,$21,$F9,$0F,$84,$FC,$99,$1F,$CD,$BF,$96
	.dc.b	$F9,$AF,$AF,$E5,$7E,$63,$E1,$25,$7D,$88,$FB,$7E,$64,$8F,$B6,$FD
	.dc.b	$EF,$2F,$EF,$F7,$C9,$FB,$EF,$E8,$BE,$73,$F4,$F1,$FA,$EF,$02,$F6
	.dc.b	$FE,$D2,$3F,$D3,$8D,$E3,$FF,$46,$BF,$BF,$8C,$CF,$02,$3F,$E9,$89
	.dc.b	$DF,$81,$88,$D9,$9E,$4F,$F4,$E5,$F5,$6C,$FE,$DB,$E7,$3F,$73,$F0
	.dc.b	$1F,$E9,$CC,$3F,$CE,$0D,$AF,$E0,$73,$BC,$3F,$D7,$71,$C0,$C3,$FE
	.dc.b	$B5,$F9,$C7,$0F,$BB,$E9,$24,$F8,$CF,$F5,$7C,$E9,$C0,$70,$FF,$5E
	.dc.b	$6B,$FE,$B2,$6B,$FE,$B7,$6B,$FE,$2E,$C1,$F3,$7E,$CF,$9C,$93,$FE
	.dc.b	$D2,$8F,$79,$87,$FD,$67,$29,$CC,$6C,$FF,$6C,$BB,$CF,$E2,$D7,$F2
	.dc.b	$2F,$87,$7C,$E7,$FA,$0E,$FF,$D3,$9F,$69,$FC,$93,$D6,$3B,$9D,$4B
	.dc.b	$06,$DF,$ED,$2C,$39,$FB,$BB,$25,$7F,$BA,$BD,$6F,$AF,$E7,$7D,$8F
	.dc.b	$82,$93,$FE,$EE,$8C,$BF,$EF,$08,$CB,$EE,$8F,$C9,$FC,$EB,$DD,$A4
	.dc.b	$9F,$A7,$1F,$E0,$C6,$C6,$5F,$F7,$94,$71,$77,$EA,$E7,$4F,$F7,$2C
	.dc.b	$E6,$77,$5E,$0A,$FD,$1F,$CE,$FC,$2B,$DF,$95,$E4,$33,$A4,$C3,$FC
	.dc.b	$3E,$2B,$E4,$BF,$EF,$46,$BF,$EF,$96,$BF,$EF,$C5,$EE,$FE,$CB,$E7
	.dc.b	$5F,$11,$4F,$1C,$4A,$FC,$9D,$70,$6C,$71,$F8,$2F,$DD,$FC,$EB,$D1
	.dc.b	$A4,$AF,$CD,$C3,$D1,$C4,$7D,$FE,$53,$27,$93,$C2,$BA,$D7,$0D,$BF
	.dc.b	$0B,$3B,$9C,$5F,$28,$5F,$77,$F5,$7F,$3B,$FD,$BD,$F0,$1E,$0F,$9D
	.dc.b	$7C,$5F,$F1,$46,$C7,$0F,$6B,$F8,$BD,$71,$CB,$C3,$28,$C6,$3E,$16
	.dc.b	$C7,$5F,$86,$FE,$3F,$F3,$AF,$47,$F9,$0F,$CE,$FF,$5F,$C3,$D0,$DB
	.dc.b	$F8,$75,$B7,$F8,$1B,$5F,$48,$71,$8F,$91,$B5,$D3,$DE,$FE,$37,$F1
	.dc.b	$72,$FF,$88,$7B,$EF,$19,$3F,$C8,$3E,$77,$55,$FC,$0A,$2E,$5F,$C8
	.dc.b	$A7,$F9,$2F,$CA,$4A,$E7,$68,$FE,$53,$CD,$99,$FE,$17,$16,$FF,$95
	.dc.b	$CF,$F4,$2F,$85,$FD,$84,$7E,$0B,$E6,$DF,$37,$F6,$FE,$77,$FB,$1E
	.dc.b	$39,$8C,$7B,$DF,$7E,$B7,$3F,$8A,$C3,$9A,$8C,$BF,$E8,$5F,$0F,$F8
	.dc.b	$73,$9B,$E6,$F5,$3F,$27,$87,$81,$D1,$D6,$78,$D7,$4B,$EE,$F6,$AE
	.dc.b	$0D,$C3,$CD,$87,$E5,$6B,$6D,$BE,$6F,$8B,$0E,$E3,$0E,$71,$BC,$7C
	.dc.b	$D3,$EB,$6C,$EA,$3D,$38,$78,$1F,$44,$F0,$C3,$C3,$1A,$38,$CE,$DE
	.dc.b	$87,$8A,$36,$F8,$E3,$55,$DE,$37,$8F,$C3,$C7,$D7,$1D,$4F,$5C,$EE
	.dc.b	$F9,$9D,$1E,$46,$67,$8C,$F4,$C7,$E2,$5F,$CB,$8F,$54,$EA,$F9,$61
	.dc.b	$E4,$3D,$4F,$68,$E7,$87,$69,$CC,$71,$32,$F1,$AE,$B8,$C7,$D3,$19
	.dc.b	$7E,$24,$F1,$9F,$93,$10,$CB,$C7,$0F,$44,$72,$31,$8F,$DB,$70,$48
	.dc.b	$F9,$53,$A8,$F9,$51,$C8,$CB,$D8,$E1,$3F,$18,$E9,$1A,$38,$39,$8F
	.dc.b	$C1,$9C,$78,$1F,$23,$E3,$6C,$69,$AF,$E1,$1A,$FF,$38,$F8,$9F,$E7
	.dc.b	$31,$88,$46,$27,$FD,$EA,$D8,$FE,$6C,$68,$4F,$18,$FC,$EF,$94,$F9
	.dc.b	$DF,$E7,$F8,$FE,$EA,$FC,$07,$E9,$A7,$5C,$BF,$A4,$8C,$CF,$F0,$09
	.dc.b	$CC,$7E,$B0,$CE,$5F,$D1,$E3,$33,$D6,$7C,$8F,$1B,$F0,$D7,$AA,$7C
	.dc.b	$11,$87,$D7,$87,$A2,$31,$1F,$D4,$E3,$18,$79,$63,$AD,$78,$6B,$71
	.dc.b	$E4,$3F,$C4,$0E,$1D,$98,$9E,$55,$D2,$3D,$86,$86,$1E,$CB,$F4,$61
	.dc.b	$EB,$BE,$35,$EA,$6C,$66,$39,$DF,$58,$F8,$8D,$AE,$67,$A6,$B8,$B6
	.dc.b	$B8,$9D,$4B,$5D,$4E,$6E,$5A,$66,$3A,$C7,$BC,$E2,$F5,$8D,$B4,$DC
	.dc.b	$7C,$0C,$3E,$45,$F9,$E3,$EA,$9D,$E7,$78,$7C,$58,$CD,$F8,$5B,$3D
	.dc.b	$5A,$78,$E3,$C2,$D8,$E7,$7E,$08,$EC,$70,$3C,$98,$7C,$67,$81,$9B
	.dc.b	$F1,$72,$FF,$73,$F1,$7F,$C3,$BA,$FF,$A4,$1B,$7F,$E2,$69,$FE,$F1
	.dc.b	$E2,$63,$0F,$F3,$F5,$59,$E5,$5F,$DD,$23,$1C,$BF,$E3,$E8,$C1,$FD
	.dc.b	$E3,$E6,$F2,$6B,$B2,$C6,$49,$87,$FA,$12,$32,$FF,$84,$BE,$71,$1E
	.dc.b	$38,$46,$27,$FD,$3D,$F1,$CF,$9D,$7C,$1E,$EB,$57,$CF,$0F,$65,$EF
	.dc.b	$C0,$D4,$63,$0F,$F1,$56,$3F,$EA,$6F,$2D,$F1,$5E,$16,$C7,$83,$E7
	.dc.b	$5A,$E6,$A4,$E7,$0B,$9F,$F9,$13,$36,$1E,$4E,$2F,$F2,$7F,$CB,$FF
	.dc.b	$95,$B5,$FA,$0D,$C7,$FD,$74,$FF,$96,$BE,$1B,$38,$F9,$8F,$F2,$7C
	.dc.b	$6D,$A6,$6F,$AA,$9E,$37,$2F,$AC,$ED,$FF,$33,$DF,$77,$33,$F4,$7E
	.dc.b	$77,$FC,$DD,$1F,$EC,$A6,$D7,$B7,$17,$C7,$6F,$CE,$97,$F5,$BE,$9B
	.dc.b	$F9,$9B,$C6,$3E,$9C,$62,$FF,$CA,$38,$3E,$27,$FC,$C7,$1F,$E7,$18
	.dc.b	$C5,$FF,$A5,$9B,$1F,$D0,$03,$87,$FA,$1B,$AD,$7F,$43,$F4,$0F,$97
	.dc.b	$F6,$61,$F9,$F9,$4E,$70,$FA,$3F,$2B,$FE,$51,$6F,$93,$F5,$A1,$E4
	.dc.b	$D8,$93,$A3,$8C,$63,$3F,$AF,$5C,$DB,$5C,$E7,$58,$7F,$A4,$BB,$7F
	.dc.b	$CD,$5F,$27,$FA,$EF,$4B,$F5,$ED,$CB,$FC,$CA,$DC,$7F,$DD,$AD,$BF
	.dc.b	$F4,$DF,$82,$F9,$E7,$CF,$FE,$F6,$30,$FE,$EB,$E7,$F1,$3A,$C7,$FD
	.dc.b	$53,$7C,$A3,$18,$FF,$A7,$B3,$7C,$16,$2B,$58,$FF,$07,$8C,$61,$FC
	.dc.b	$37,$C3,$FF,$62,$BF,$E7,$CE,$3F,$E1,$3E,$CF,$F0,$A6,$C6,$4F,$F4
	.dc.b	$0C,$65,$A6,$72,$F1,$57,$46,$D7,$F0,$7F,$9E,$FE,$29,$1F,$85,$6B
	.dc.b	$FE,$06,$6B,$FE,$BD,$6B,$FE,$C8,$78,$E5,$FE,$C4,$EE,$74,$3E,$5B
	.dc.b	$B7,$91,$F9,$A7,$F9,$1F,$85,$FC,$93,$B6,$EB,$FE,$9A,$6E,$7E,$86
	.dc.b	$E5,$F7,$F5,$FD,$4D,$7E,$43,$97,$F1,$C7,$E3,$F9,$AC,$64,$98,$7E
	.dc.b	$32,$32,$FE,$3A,$6D,$FF,$0E,$E9,$FF,$69,$DF,$08,$FC,$77,$CE,$3E
	.dc.b	$07,$F9,$F3,$67,$FD,$77,$C9,$FB,$5A,$DD,$6B,$E0,$A5,$7F,$B0,$2D
	.dc.b	$F0,$78,$CF,$F0,$D1,$FE,$C2,$9E,$7E,$97,$6F,$63,$F0,$CF,$E1,$FC
	.dc.b	$A9,$3E,$0E,$3B,$87,$E3,$F5,$7C,$F7,$E3,$4C,$ED,$FA,$78,$CD,$FA
	.dc.b	$27,$F2,$5F,$3D,$B8,$F5,$BE,$EF,$06,$47,$A7,$D0,$B8,$FA,$26,$E3
	.dc.b	$F1,$A6,$D7,$F6,$7D,$BF,$80,$BF,$06,$BE,$4C,$EA,$DD,$FF,$A6,$F5
	.dc.b	$FF,$48,$7E,$97,$E5,$AE,$1F,$3A,$2D,$7B,$E6,$E5,$ED,$9B,$5F,$DD
	.dc.b	$F5,$56,$2F,$BE,$7F,$D3,$6B,$CD,$E7,$7D,$CE,$7F,$BA,$C8,$FB,$0E
	.dc.b	$5A,$4A,$FB,$31,$96,$C7,$E1,$3D,$8F,$DF,$1F,$BF,$F8,$5B,$19,$24
	.dc.b	$FD,$98,$CB,$F6,$E6,$E9,$F6,$66,$D7,$CD,$F0,$FF,$26,$7E,$53,$C3
	.dc.b	$B1,$F8,$F8,$B8,$7E,$3A,$6D,$FE,$72,$6D,$7D,$79,$B5,$F5,$35,$7C
	.dc.b	$AF,$F1,$73,$F8,$C7,$CB,$5B,$FD,$0C,$5A,$FC,$94,$DC,$7F,$21,$F2
	.dc.b	$CB,$FF,$4C,$73,$FD,$5F,$96,$C6,$1F,$E9,$88,$FB,$15,$8F,$73,$F8
	.dc.b	$54,$F8,$EB,$15,$AB,$E7,$39,$6F,$2F,$F3,$0C,$37,$39,$3B,$01,$F3
	.dc.b	$9F,$B4,$F2,$39,$C6,$3B,$D1,$8F,$9D,$7C,$F7,$4A,$D4,$F0,$BE,$51
	.dc.b	$D8,$C9,$6B,$8E,$5B,$C3,$AC,$EE,$B8,$EB,$F0,$23,$8B,$6F,$8C,$6E
	.dc.b	$B8,$CE,$EB,$8E,$9E,$DA,$E2,$DB,$E3,$5B,$BE,$33,$BB,$E3,$7F,$15
	.dc.b	$AA,$F8,$CE,$EB,$8C,$ED,$AF,$80,$F9,$D3,$C1,$D6,$3F,$04,$8F,$C5
	.dc.b	$CB,$E6,$61,$E0,$3E,$5C,$7F,$B6,$1A,$F8,$67,$FC,$38,$FC,$E3,$FD
	.dc.b	$BB,$F0,$4E,$57,$8B,$A8,$C3,$63,$33,$FE,$25,$9C,$BF,$E2,$7B,$FE
	.dc.b	$81,$F0,$3D,$9B,$19,$F1,$BF,$DD,$2D,$9D,$46,$36,$FA,$8D,$DB,$E7
	.dc.b	$36,$BF,$25,$F3,$CE,$3F,$F6,$C5,$6B,$0F,$C6,$5E,$EF,$F2,$ED,$BE
	.dc.b	$11,$A9,$8F,$E2,$E7,$57,$D2,$3E,$E4,$73,$C7,$D7,$8F,$9A,$FF,$43
	.dc.b	$87,$4B,$E5,$1D,$27,$51,$F9,$D9,$D4,$60,$98,$FA,$67,$58,$F8,$5F
	.dc.b	$4D,$F3,$9D,$57,$E7,$EF,$9C,$7F,$39,$DB,$D1,$C7,$FE,$2A,$6F,$3F
	.dc.b	$EC,$B1,$FD,$FA,$FA,$FC,$CC,$9C,$BC,$23,$8F,$D8,$7E,$97,$9F,$FB
	.dc.b	$27,$D4,$DE,$7F,$E9,$0D,$C3,$F9,$15,$5C,$BF,$C7,$6D,$C9,$54,$C4
	.dc.b	$6F,$E7,$FF,$D4,$B7,$FE,$46,$C3,$FD,$95,$1F,$E8,$28,$C1,$FD,$89
	.dc.b	$C3,$FE,$49,$78,$B7,$CF,$E9,$E7,$C8,$E3,$92,$B7,$9F,$F9,$95,$B8
	.dc.b	$78,$2F,$85,$F3,$3F,$C3,$3F,$57,$ED,$FD,$8F,$F8,$44,$7F,$96,$5B
	.dc.b	$87,$F8,$26,$7E,$01,$B9,$87,$22,$1C,$4C,$9F,$E1,$57,$E5,$7E,$1C
	.dc.b	$F3,$6F,$14,$AF,$5C,$E5,$B3,$97,$FC,$D9,$1F,$C9,$6F,$FD,$04,$EE
	.dc.b	$33,$3B,$9C,$D7,$82,$3D,$DD,$9E,$FF,$87,$97,$F2,$B2,$FE,$66,$FC
	.dc.b	$19,$CB,$F0,$E7,$8B,$B8,$CE,$1F,$E9,$3A,$56,$AB,$CF,$E5,$7C,$0F
	.dc.b	$2A,$47,$BA,$3E,$19,$87,$DB,$18,$74,$7C,$47,$57,$F0,$27,$EC,$7D
	.dc.b	$34,$AF,$53,$F6,$FE,$36,$46,$08,$7D,$D7,$2D,$3E,$13,$F8,$3F,$9C
	.dc.b	$FC,$27,$C5,$C9,$F8,$CF,$DA,$F2,$3E,$DF,$91,$67,$EC,$C6,$1F,$B7
	.dc.b	$E4,$7E,$00,$8F,$CF,$FA,$1F,$D7,$2F,$A1,$7C,$B6,$17,$C3,$53,$11
	.dc.b	$27,$F5,$27,$18,$D3,$89,$E6,$D9,$CA,$DF,$FC,$22,$F1,$34,$7E,$0F
	.dc.b	$E8,$7F,$49,$52,$30,$43,$F4,$CE,$5F,$D4,$47,$EA,$FE,$57,$F5,$97
	.dc.b	$27,$F1,$D1,$97,$F2,$11,$97,$F6,$11,$FB,$3C,$3F,$69,$F4,$92,$7F
	.dc.b	$2D,$19,$7F,$31,$19,$7F,$71,$19,$7F,$75,$19,$C7,$F2,$D3,$F4,$3E
	.dc.b	$81,$F4,$3F,$DA,$AD,$18,$7F,$B6,$B5,$FE,$0A,$6A,$38,$46,$6B,$E5
	.dc.b	$B7,$BB,$FC,$24,$DA,$FF,$18,$68,$E4,$FE,$21,$DF,$FC,$47,$E4,$E5
	.dc.b	$7E,$CE,$32,$FE,$D6,$32,$FF,$15,$8F,$C4,$FC,$EF,$F4,$38,$FB,$98
	.dc.b	$FF,$88,$63,$FA,$24,$7E,$19,$AE,$8F,$F2,$13,$C7,$2F,$C1,$B6,$79
	.dc.b	$78,$FD,$8E,$FE,$44,$BE,$91,$97,$B4,$7F,$1E,$F0,$BF,$A4,$7E,$0F
	.dc.b	$FC,$AA,$FF,$CA,$1E,$2A,$31,$7F,$EA,$2F,$11,$7C,$63,$F9,$EC,$EA
	.dc.b	$B8,$46,$4F,$50,$F9,$D7,$B9,$F6,$DB,$1F,$E2,$66,$BF,$69,$AF,$F4
	.dc.b	$07,$3A,$FE,$91,$B9,$7F,$92,$DB,$3F,$23,$CE,$FD,$CF,$CF,$3E,$0E
	.dc.b	$FC,$1C,$AF,$DA,$CE,$5B,$39,$7D,$30,$9E,$73,$27,$8C,$6F,$1F,$DB
	.dc.b	$D7,$39,$F8,$3E,$CF,$DC,$7F,$AC,$B5,$FF,$31,$35,$FF,$40,$35,$FF
	.dc.b	$79,$F0,$FF,$12,$8F,$C2,$D7,$F0,$46,$E7,$FE,$6B,$F8,$DF,$C2,$5F
	.dc.b	$FA,$27,$F0,$EF,$F2,$73,$FE,$52,$6B,$FA,$66,$BF,$DB,$1A,$FF,$9D
	.dc.b	$5D,$DF,$F1,$26,$D6,$7D,$9F,$E4,$BE,$F7,$BE,$BF,$A7,$7C,$9F,$F8
	.dc.b	$57,$F0,$AF,$A1,$E9,$DE,$0F,$EF,$F9,$FF,$BB,$3C,$F5,$FF,$7D,$35
	.dc.b	$52,$7F,$B0,$47,$F8,$07,$E5,$A4,$7F,$63,$7F,$C0,$9E,$0E,$23,$FC
	.dc.b	$09,$17,$2F,$EC,$13,$FC,$EB,$DD,$69,$FC,$B6,$DF,$34,$F3,$7F,$D2
	.dc.b	$9F,$36,$F4,$7F,$D2,$D3,$9A,$FF,$5A,$4E,$5F,$F5,$B4,$65,$61,$DB
	.dc.b	$2C,$E1,$FE,$9F,$8C,$4F,$82,$3B,$E1,$D9,$B1,$C0,$FF,$61,$4F,$0B
	.dc.b	$EC,$D9,$DC,$F3,$3C,$B5,$CD,$E3,$8F,$BE,$F8,$B6,$B8,$C7,$A7,$1F
	.dc.b	$65,$73,$3D,$75,$B7,$9E,$3E,$55,$3C,$2B,$96,$1E,$19,$E3,$5B,$C7
	.dc.b	$D1,$1E,$63,$95,$F1,$C7,$E3,$63,$DE,$78,$9E,$39,$D3,$C4,$F3,$57
	.dc.b	$47,$78,$F8,$9E,$B8,$F8,$E7,$46,$EB,$9C,$FA,$A7,$57,$E1,$AD,$47
	.dc.b	$AE,$B7,$1C,$A7,$CD,$1C,$8C,$BC,$4C,$CF,$13,$AE,$3F,$0E,$33,$7E
	.dc.b	$63,$11,$1F,$81,$3C,$5C,$61,$F6,$A7,$84,$68,$CE,$3F,$2A,$33,$3F
	.dc.b	$0E,$7E,$3D,$F9,$67,$35,$F8,$23,$84,$EA,$32,$FC,$FA,$F0,$39,$3C
	.dc.b	$04,$72,$49,$E9,$1C,$5E,$B1,$CB,$A5,$F1,$5E,$06,$D7,$E3,$83,$87
	.dc.b	$F0,$6D,$43,$DD,$B1,$98,$C1,$CA,$7F,$C0,$B3,$A6,$C6,$4F,$F4,$6C
	.dc.b	$7E,$77,$E4,$FF,$CC,$FC,$DF,$9F,$9E,$F3,$9F,$59,$39,$9E,$F8,$FF
	.dc.b	$03,$BE,$31,$FA,$B3,$89,$FA,$D7,$31,$FC,$26,$32,$F1,$AE,$B1,$C4
	.dc.b	$EB,$97,$C0,$55,$CA,$3D,$75,$C6,$FC,$98,$79,$DB,$7E,$1C,$3A,$D7
	.dc.b	$86,$3A,$9F,$33,$0E,$8F,$C4,$38,$57,$29,$ED,$1E,$5B,$EC,$D3,$65
	.dc.b	$8D,$9D,$E3,$6D,$8D,$B4,$DE,$1E,$C3,$E0,$5F,$92,$7E,$15,$72,$9D
	.dc.b	$47,$A7,$0F,$6C,$62,$3E,$3C,$61,$F6,$5F,$66,$CF,$69,$D4,$76,$3C
	.dc.b	$B1,$CE,$FC,$51,$F6,$23,$86,$1F,$15,$B1,$F0,$5B,$7E,$D6,$9C,$EF
	.dc.b	$BC,$F3,$3D,$55,$CD,$A7,$32,$D7,$36,$D7,$39,$E1,$1F,$01,$B7,$E8
	.dc.b	$8F,$BE,$64,$E8,$5A,$FE,$C1,$E0,$39,$5F,$2B,$F9,$4D,$BF,$D9,$8D
	.dc.b	$C7,$FC,$93,$2F,$A6,$C6,$3F,$DF,$EA,$E1,$CD,$B5,$D2,$FF,$DB,$D7
	.dc.b	$89,$CF,$D3,$72,$EA,$7E,$1B,$FC,$02,$FF,$DC,$93,$98,$C6,$5F,$E6
	.dc.b	$9C,$BD,$13,$98,$E7,$5F,$C2,$33,$FD,$6E,$1D,$27,$CE,$3F,$0C,$FF
	.dc.b	$2F,$D7,$FA,$E9,$B1,$FE,$75,$F7,$1F,$36,$FA,$7E,$35,$AC,$7F,$D7
	.dc.b	$2F,$0B,$FD,$C1,$D7,$CA,$FF,$5E,$BF,$BB,$8D,$36,$32,$FF,$1E,$FB
	.dc.b	$47,$B5,$E6,$6E,$7E,$B3,$F9,$1F,$E0,$6F,$C0,$DF,$49,$BB,$95,$FE
	.dc.b	$A6,$C1,$87,$27,$FD,$52,$D7,$CE,$D7,$F9,$7F,$D0,$BE,$3B,$F9,$0D
	.dc.b	$78,$65,$39,$DB,$F0,$2E,$E6,$D7,$C7,$78,$46,$5B,$19,$AF,$F5,$14
	.dc.b	$E5,$B1,$9D,$BF,$92,$CE,$5B,$19,$AF,$3B,$E6,$AE,$37,$F7,$BE,$87
	.dc.b	$F2,$11,$FE,$96,$6D,$FF,$30,$EA,$7D,$73,$E1,$F8,$74,$CC,$FE,$56
	.dc.b	$1C,$B1,$FB,$AD,$7F,$9B,$35,$74,$7F,$AB,$6B,$F4,$7E,$A7,$E9,$EF
	.dc.b	$FD,$57,$18,$EB,$CB,$63,$27,$FA,$C2,$3E,$5F,$A9,$F3,$3E,$72,$CF
	.dc.b	$CA,$6E,$3F,$21,$BB,$7C,$66,$CF,$C6,$6C,$FD,$AE,$04,$7E,$E3,$DA
	.dc.b	$F9,$57,$FD,$53,$44,$7C,$FB,$F6,$BA,$C3,$FD,$96,$E6,$77,$1C,$F8
	.dc.b	$7A,$34,$FC,$55,$3E,$7B,$C3,$18,$6C,$7F,$B9,$78,$BE,$AF,$CD,$7E
	.dc.b	$E2,$73,$97,$F6,$8A,$E5,$9F,$EA,$9B,$87,$E8,$6B,$9C,$7E,$D1,$AF
	.dc.b	$A3,$EB,$3E,$07,$AB,$F0,$5B,$5F,$9F,$D3,$E0,$C6,$2B,$F8,$86,$A8
	.dc.b	$C6,$5F,$1E,$31,$7C,$56,$1D,$63,$0F,$CF,$F9,$4F,$A1,$E9,$D9,$E0
	.dc.b	$DC,$7B,$B7,$0E,$CF,$D7,$F6,$7E,$C7,$65,$BF,$03,$6F,$CE,$DB,$F3
	.dc.b	$3F,$B6,$FA,$2F,$DB,$FC,$C5,$BF,$DA,$B6,$FF,$66,$DB,$FD,$8B,$67
	.dc.b	$F6,$2D,$9F,$C6,$3F,$D0,$BF,$0B,$FE,$2B,$3F,$2D,$B3,$F2,$BD,$AD
	.dc.b	$63,$BC,$B8,$45,$BE,$0F,$E5,$7E,$76,$46,$0F,$9C,$FE,$67,$A2,$47
	.dc.b	$D1,$E4,$B8,$FD,$89,$FC,$F7,$D2,$E3,$5F,$B0,$E2,$FE,$C3,$97,$F3
	.dc.b	$AF,$AE,$FD,$8D,$8F,$63,$6F,$FA,$A6,$88,$C5,$FE,$AE,$31,$5F,$0E
	.dc.b	$F9,$B4,$E6,$7E,$27,$E7,$3F,$69,$F3,$BF,$B5,$6C,$64,$FC,$74,$65
	.dc.b	$A6,$71,$FC,$BC,$65,$B1,$F4,$FE,$83,$EB,$70,$FD,$76,$C6,$4F,$B1
	.dc.b	$19,$69,$99,$F7,$C6,$5B,$1F,$A1,$9F,$E1,$93,$FC,$35,$B1,$93,$F4
	.dc.b	$D1,$96,$99,$9F,$D7,$C6,$5B,$1F,$8B,$F7,$7F,$1D,$F2,$DF,$8F,$6C
	.dc.b	$64,$FC,$84,$65,$A6,$67,$EB,$C6,$5B,$1F,$B9,$E3,$FE,$49,$F3,$DF
	.dc.b	$C9,$5B,$19,$3F,$7D,$19,$69,$FC,$06,$FF,$CF,$BE,$B3,$E5,$FF,$93
	.dc.b	$C6,$1F,$E5,$38,$DC,$7F,$96,$CD,$EC,$FE,$62,$DE,$B3,$6B,$0F,$1A
	.dc.b	$7C,$EF,$F1,$0F,$47,$FD,$37,$8F,$FA,$2E,$31,$3F,$EA,$B8,$FF,$47
	.dc.b	$7C,$F3,$E2,$79,$56,$B4,$F4,$5F,$FA,$CA,$31,$7C,$FB,$D8,$F6,$AF
	.dc.b	$F5,$BE,$3C,$F6,$C4,$79,$2B,$6F,$84,$E8,$E3,$1E,$F1,$BB,$E3,$7C
	.dc.b	$E3,$81,$92,$E3,$E4,$79,$CF,$0C,$BD,$78,$75,$CB,$D1,$3E,$4B,$E1
	.dc.b	$6C,$3B,$63,$E2,$9C,$93,$0E,$55,$EE,$C7,$AE,$1C,$E7,$33,$E5,$BF
	.dc.b	$5E,$1D,$31,$F0,$CF,$A2,$BE,$1B,$FE,$48,$C7,$EA,$B6,$32,$F0,$3F
	.dc.b	$C9,$BF,$00,$F8,$4F,$F1,$B4,$7F,$94,$04,$67,$C7,$FF,$21,$36,$34
	.dc.b	$D7,$C1,$E1,$7F,$BC,$9E,$58,$23,$11,$AF,$7B,$FC,$99,$39,$6C,$F1
	.dc.b	$D3,$E4,$37,$3F,$C8,$36,$FF,$DF,$7F,$3A,$F9,$EF,$F5,$ED,$6B,$0F
	.dc.b	$A9,$7B,$BF,$CD,$B6,$F8,$46,$A6,$3F,$82,$9D,$5F,$48,$FC,$C4,$73
	.dc.b	$C7,$D1,$1B,$AF,$CA,$DF,$48,$C1,$31,$F5,$CE,$B0,$E9,$3E,$48,$F0
	.dc.b	$CE,$A3,$F2,$B5,$CE,$30,$4C,$7D,$D3,$AB,$F3,$72,$BE,$7B,$F6,$53
	.dc.b	$FA,$B8,$C7,$16,$A7,$33,$FA,$99,$E1,$F1,$39,$A9,$3C,$6B,$78,$FE
	.dc.b	$EA,$F7,$39,$F9,$D3,$FE,$1F,$A7,$CF,$62,$3F,$D5,$D1,$87,$53,$9F
	.dc.b	$0B,$FC,$8D,$28,$C7,$3F,$4C,$3D,$98,$63,$0F,$F3,$2D,$F3,$7A,$DE
	.dc.b	$69,$F2,$1F,$C7,$A7,$FC,$64,$BA,$B4,$D8,$CF,$56,$0C,$CC,$8C,$90
	.dc.b	$FF,$64,$4F,$16,$C7,$17,$E1,$7B,$7E,$3F,$81,$FF,$5D,$EB,$FE,$63
	.dc.b	$6E,$BE,$DD,$15,$E9,$6D,$FF,$92,$6B,$6F,$FA,$11,$B9,$FD,$FA,$E0
	.dc.b	$77,$98,$78,$8F,$F2,$97,$C7,$FF,$A5,$5B,$C7,$F1,$E3,$F9,$3E,$5F
	.dc.b	$E7,$77,$53,$FE,$E5,$6C,$66,$78,$1C,$7D,$79,$5D,$08,$71,$7E,$17
	.dc.b	$77,$C3,$F8,$59,$3E,$D8,$CB,$EF,$8C,$BF,$EF,$5C,$FF,$8E,$E1,$F2
	.dc.b	$6B,$2F,$CB,$AE,$35,$E7,$F8,$8F,$67,$CF,$48,$F8,$31,$EF,$30,$FB
	.dc.b	$63,$0E,$8F,$8E,$EA,$BE,$24,$FD,$BF,$92,$95,$F5,$5F,$BD,$F3,$D2
	.dc.b	$30,$43,$D4,$E5,$A7,$DD,$7F,$0D,$F3,$DF,$87,$F8,$A9,$3F,$81,$7E
	.dc.b	$B7,$CE,$FD,$7F,$9E,$B3,$F7,$23,$0F,$DE,$F8,$4F,$B8,$47,$F1,$5F
	.dc.b	$3D,$FA,$95,$F3,$0F,$9D,$C2,$F9,$BA,$62,$24,$FF,$AC,$27,$4D,$9C
	.dc.b	$AD,$FF,$D5,$93,$D6,$34,$43,$5D,$5F,$C1,$63,$51,$F7,$FE,$7B,$F5
	.dc.b	$11,$23,$F4,$51,$FA,$B3,$0F,$E1,$63,$0F,$E9,$FE,$7B,$F6,$18,$C8
	.dc.b	$C1,$0F,$C6,$39,$7F,$1D,$1F,$B7,$C7,$F7,$1F,$3D,$27,$F6,$51,$97
	.dc.b	$F6,$91,$97,$F2,$D1,$97,$F3,$11,$9C,$BF,$80,$4F,$D1,$F9,$E7,$CC
	.dc.b	$7D,$0E,$29,$87,$FB,$72,$72,$D9,$FE,$E8,$FF,$2E,$8C,$BF,$DD,$DA
	.dc.b	$FF,$2F,$0E,$1F,$E5,$5D,$3F,$C9,$1F,$E1,$9E,$97,$F1,$53,$F8,$B7
	.dc.b	$85,$23,$24,$3F,$64,$FE,$27,$E7,$BF,$A4,$E7,$FE,$E9,$5E,$4E,$4D
	.dc.b	$79,$32,$FF,$DE,$F3,$A6,$CE,$5F,$C1,$9F,$D2,$A3,$AF,$1F,$F4,$97
	.dc.b	$84,$68,$38,$BF,$15,$FF,$1B,$F0,$BF,$A1,$7C,$57,$F2,$63,$F9,$3F
	.dc.b	$87,$FC,$B3,$0F,$E3,$93,$A6,$1F,$C9,$9F,$D3,$7C,$F3,$DF,$F8,$1E
	.dc.b	$24,$AF,$EB,$F9,$34,$FE,$BD,$7B,$9C,$C6,$2F,$EC,$C2,$74,$FF,$8B
	.dc.b	$DA,$FD,$96,$BF,$DA,$7E,$79,$F0,$BF,$B9,$5F,$09,$4C,$44,$9F,$09
	.dc.b	$FD,$D2,$76,$E0,$38,$7E,$63,$D9,$D3,$5C,$CF,$77,$33,$FC,$2F,$D8
	.dc.b	$FA,$58,$FF,$7F,$F8,$4F,$EF,$55,$DD,$AF,$D9,$9D,$BF,$E0,$87,$EE
	.dc.b	$BF,$99,$6B,$FE,$09,$6B,$BE,$2F,$97,$F3,$DF,$E5,$98,$F6,$7C,$DB
	.dc.b	$E1,$FF,$8F,$36,$3E,$04,$ED,$FC,$53,$5F,$D4,$35,$FF,$0C,$35,$FF
	.dc.b	$0E,$B7,$CE,$74,$BE,$7B,$FC,$29,$F3,$1F,$5B,$D3,$BB,$FF,$84,$71
	.dc.b	$FE,$B5,$5F,$E0,$33,$33,$23,$FB,$F4,$7F,$81,$FE,$7A,$47,$F8,$05
	.dc.b	$FF,$05,$77,$E2,$3F,$B2,$45,$E6,$FF,$06,$4F,$F3,$DF,$75,$B7,$F3
	.dc.b	$3B,$7C,$E3,$CD,$FF,$1A,$7C,$F3,$D1,$FF,$1B,$4E,$6B,$FC,$77,$39
	.dc.b	$7F,$C7,$B1,$97,$FD,$0B,$7D,$63,$18,$78,$32,$DE,$3D,$B1,$EA,$DC
	.dc.b	$32,$7F,$92,$2B,$83,$E6,$8F,$F2,$44,$62,$FC,$18,$F8,$EB,$47,$9A
	.dc.b	$FA,$BC,$F1,$F2,$61,$C9,$B1,$B3,$C2,$2B,$A5,$FA,$A7,$73,$EE,$C7
	.dc.b	$75,$AB,$F5,$4F,$88,$D1,$71,$F4,$46,$B0,$F4,$E3,$E6,$AD,$C7,$7C
	.dc.b	$3A,$CE,$70,$ED,$5E,$B9,$F3,$E1,$D2,$3D,$47,$86,$76,$F4,$C3,$E0
	.dc.b	$57,$48,$87,$97,$1F,$6D,$F3,$8F,$85,$87,$84,$F3,$46,$4D,$61,$D8
	.dc.b	$EE,$E6,$BE,$3C,$66,$78,$C7,$63,$2F,$5A,$F5,$C6,$6F,$CB,$8F,$C6
	.dc.b	$9C,$C4,$7D,$D5,$C5,$F5,$46,$67,$84,$68,$C6,$3F,$09,$C1,$23,$E8
	.dc.b	$3C,$8C,$47,$D7,$38,$46,$8E,$A4,$72,$48,$FA,$71,$97,$EC,$1D,$7D
	.dc.b	$D7,$CF,$78,$DB,$1A,$6B,$F8,$A6,$BF,$EA,$CF,$73,$FD,$5B,$18,$84
	.dc.b	$62,$7F,$DF,$0D,$8F,$F0,$13,$5F,$F5,$93,$FA,$0F,$91,$7A,$FF,$D6
	.dc.b	$67,$F3,$47,$73,$F3,$91,$AE,$CF,$D3,$C6,$71,$FE,$05,$39,$8E,$27
	.dc.b	$90,$E3,$5D,$67,$F8,$44,$71,$CB,$FD,$89,$5C,$67,$33,$C5,$EB,$5C
	.dc.b	$6B,$DB,$3D,$2F,$C6,$EE,$39,$CE,$E3,$E0,$9B,$3C,$B8,$7C,$18,$D4
	.dc.b	$7C,$43,$57,$E2,$9D,$5F,$66,$CF,$23,$77,$E4,$AF,$3D,$6E,$72,$78
	.dc.b	$A3,$2F,$16,$9D,$DD,$9E,$37,$33,$D4,$F6,$CF,$58,$CC,$F5,$9C,$BD
	.dc.b	$5B,$1D,$4D,$C7,$58,$C9,$D5,$FF,$7C,$E1,$F4,$E3,$10,$8C,$3D,$CD
	.dc.b	$E1,$F2,$9F,$B8,$62,$FC,$D5,$DA,$38,$1C,$E7,$83,$F6,$4E,$07,$A4
	.dc.b	$B7,$E9,$9E,$51,$E9,$32,$F9,$AF,$CB,$19,$3B,$5F,$89,$D9,$E2,$8F
	.dc.b	$5F,$3F,$F7,$AF,$94,$FD,$16,$BF,$EF,$56,$DF,$F9,$A3,$E5,$2E,$1F
	.dc.b	$EF,$19,$B7,$C6,$BF,$DE,$D1,$8E,$95,$D8,$C7,$CE,$F2,$EA,$7C,$F3
	.dc.b	$2C,$E3,$FE,$38,$9C,$C6,$0F,$F3,$2B,$88,$FF,$1E,$E5,$FE,$12,$BE
	.dc.b	$97,$CE,$7F,$C7,$DF,$3A,$F1,$FF,$7D,$8F,$FA,$16,$31,$3F,$E8,$C6
	.dc.b	$C7,$FA,$24,$38,$78,$4F,$A7,$F0,$C7,$C1,$FC,$49,$DC,$E6,$31,$F3
	.dc.b	$8F,$6B,$C3,$5E,$F1,$19,$C7,$FD,$3A,$D8,$F1,$35,$FF,$4E,$39,$E1
	.dc.b	$FE,$31,$C4,$FB,$97,$D1,$78,$31,$FA,$F3,$98,$FB,$3F,$3C,$F8,$CF
	.dc.b	$F3,$C6,$7F,$E6,$68,$C7,$CE,$6A,$73,$A7,$FA,$52,$73,$5B,$9C,$E7
	.dc.b	$FC,$A6,$73,$5B,$9C,$D7,$DD,$9C,$D6,$E7,$34,$9C,$D6,$E7,$31,$89
	.dc.b	$FC,$37,$CE,$B9,$BE,$5C,$61,$AA,$35,$7F,$EC,$86,$C7,$DD,$6E,$1F
	.dc.b	$A0,$FA,$77,$C9,$7D,$B9,$CD,$FF,$A7,$A7,$2F,$E4,$3D,$5C,$E0,$E8
	.dc.b	$47,$E8,$27,$59,$7E,$9A,$31,$5F,$EB,$EA,$DC,$E7,$AB,$4D,$BF,$9D
	.dc.b	$E9,$FD,$3F,$98,$FA,$8D,$8C,$9F,$4E,$32,$D3,$38,$7D,$38,$CB,$63
	.dc.b	$F5,$3E,$D7,$D2,$C4,$B8,$7F,$7A,$E4,$C4,$F0,$33,$1A,$69,$FE,$AD
	.dc.b	$F9,$D6,$1F,$51,$B1,$FE,$FA,$6E,$1F,$AF,$6B,$FE,$F9,$73,$C1,$FD
	.dc.b	$D7,$E2,$9F,$4A,$F8,$CF,$D8,$56,$B0,$FE,$15,$5A,$9F,$4F,$CF,$7A
	.dc.b	$BD,$6B,$5F,$01,$B9,$7D,$3D,$51,$8C,$FF,$C0,$9B,$7C,$48,$C5,$FF
	.dc.b	$17,$8C,$3F,$2B,$0E,$CF,$CC,$F9,$4F,$93,$EA,$7C,$A6,$C6,$4F,$A3
	.dc.b	$19,$69,$F5,$3D,$9F,$A3,$EB,$FD,$26,$C6,$4F,$B3,$19,$69,$FB,$5F
	.dc.b	$98,$FD,$D7,$CC,$7E,$ED,$B1,$93,$F7,$51,$96,$99,$DB,$F7,$51,$96
	.dc.b	$C7,$F3,$0F,$9C,$FF,$67,$36,$FF,$0F,$EE,$7F,$44,$9C,$D7,$E7,$DF
	.dc.b	$F7,$BD,$71,$9D,$63,$B7,$F1,$FF,$3D,$27,$47,$E5,$9F,$C9,$F4,$48
	.dc.b	$FC,$CF,$3D,$C7,$ED,$4F,$E6,$3E,$7B,$1C,$7F,$6B,$91,$19,$C3,$F4
	.dc.b	$2E,$9F,$B9,$1A,$7D,$D8,$7A,$E7,$D8,$D8,$C9,$F0,$E3,$2D,$33,$87
	.dc.b	$C1,$8C,$B6,$3F,$63,$F3,$DF,$AE,$F9,$EF,$D7,$B6,$32,$7E,$DA,$32
	.dc.b	$D3,$33,$F9,$98,$CB,$63,$EB,$7C,$F7,$D3,$E2,$FA,$8D,$8C,$9F,$6E
	.dc.b	$32,$D3,$33,$F6,$63,$2D,$8F,$E1,$33,$FC,$1A,$7F,$83,$B6,$32,$7F
	.dc.b	$0E,$8C,$B4,$CC,$FE,$C6,$32,$D8,$FC,$77,$BF,$F8,$BF,$98,$FC,$63
	.dc.b	$63,$27,$E4,$E3,$2D,$33,$3F,$91,$8C,$B6,$3F,$90,$F2,$7F,$1E,$F1
	.dc.b	$3F,$8F,$B6,$32,$7F,$26,$8C,$B4,$FE,$53,$7F,$E8,$9F,$5F,$F8,$A5
	.dc.b	$FF,$43,$DD,$39,$EC,$FC,$6E,$56,$B5,$37,$AF,$F9,$8B,$6F,$8E,$3F
	.dc.b	$EA,$1C,$38,$B6,$F8,$B5,$7C,$E7,$F1,$AE,$EC,$36,$3F,$D4,$F5,$FE
	.dc.b	$11,$F9,$E7,$C4,$F8,$E7,$8F,$80,$C3,$8B,$63,$89,$FE,$37,$C3,$A6
	.dc.b	$DF,$02,$18,$74,$CB,$8E,$12,$78,$9D,$9F,$44,$EE,$33,$7D,$32,$FF
	.dc.b	$13,$36,$3C,$18,$F8,$F1,$F0,$5F,$17,$C5,$5B,$AE,$C7,$4A,$E1,$7F
	.dc.b	$07,$2F,$81,$5B,$24,$EA,$BF,$CA,$98,$75,$BF,$1C,$E6,$7A,$E1,$E0
	.dc.b	$8C,$56,$E7,$CF,$8F,$C1,$7F,$C8,$07,$90,$B8,$7D,$56,$D7,$CF,$77
	.dc.b	$F0,$0F,$9E,$FF,$1C,$47,$F9,$56,$31,$E6,$7F,$95,$E1,$18,$75,$3C
	.dc.b	$1E,$FE,$13,$DE,$E1,$8A,$31,$3F,$E6,$18,$C4,$EA,$73,$3C,$A3,$F1
	.dc.b	$B1,$C3,$4F,$A2,$DD,$7F,$CB,$DF,$3C,$F9,$EF,$F7,$8C,$7D,$C9,$D6
	.dc.b	$1F,$8B,$6B,$F9,$A6,$D7,$18,$EF,$84,$7F,$0F,$3A,$BE,$B1,$CE,$75
	.dc.b	$87,$91,$FD,$35,$F3,$9E,$31,$AC,$3C,$8E,$AF,$9C,$F8,$E3,$4F,$53
	.dc.b	$F3,$75,$CE,$FB,$47,$42,$3F,$A5,$8D,$5F,$FA,$57,$F4,$37,$C4,$FA
	.dc.b	$38,$D1,$8C,$3F,$7D,$18,$9F,$6F,$33,$F3,$DF,$A7,$97,$F1,$0A,$D5
	.dc.b	$F0,$C3,$93,$FB,$88,$CB,$FB,$AB,$E5,$9F,$BB,$F1,$17,$C3,$7F,$A5
	.dc.b	$7C,$3F,$F5,$3C,$62,$FF,$D6,$D1,$89,$D6,$1F,$E2,$3F,$D4,$1E,$06
	.dc.b	$2F,$FD,$79,$87,$15,$87,$17,$9D,$FF,$A8,$E7,$53,$FE,$30,$BE,$75
	.dc.b	$A3,$FD,$8D,$F8,$43,$E6,$FF,$41,$08,$C3,$FA,$36,$BF,$EB,$B6,$BF
	.dc.b	$B2,$73,$5F,$EC,$46,$E1,$FE,$B5,$6C,$FF,$A9,$F6,$7C,$EF,$B6,$3C
	.dc.b	$18,$5A,$DC,$60,$F4,$74,$AF,$8C,$7A,$CF,$47,$C6,$7C,$79,$7C,$E7
	.dc.b	$F9,$A2,$78,$B5,$F9,$39,$79,$6B,$C0,$48,$E9,$82,$3A,$7E,$63,$FE
	.dc.b	$A7,$9C,$C6,$DE,$67,$CC,$77,$EB,$7A,$CF,$26,$0F,$22,$99,$AE,$F3
	.dc.b	$5E,$67,$D3,$7D,$9E,$07,$FB,$C3,$7F,$9D,$1B,$6C,$FC,$69,$D6,$7F
	.dc.b	$EE,$5E,$97,$8F,$F1,$7F,$08,$F8,$58,$29,$87,$FB,$6E,$73,$1C,$6B
	.dc.b	$EB,$5F,$2A,$FA,$0D,$7C,$ED,$BE,$91,$FE,$E5,$FC,$27,$FD,$44,$F0
	.dc.b	$AE,$64,$7E,$EC,$69,$E7,$E4,$F9,$F4,$91,$E0,$8F,$19,$1C,$1C,$BC
	.dc.b	$0C,$3C,$8C,$9F,$5B,$89,$7F,$CD,$E3,$F1,$BF,$84,$FF,$4C,$CB,$D3
	.dc.b	$48,$C6,$9F,$1A,$31,$5F,$99,$FC,$25,$F0,$14,$91,$F9,$58,$CB,$F9
	.dc.b	$AF,$90,$7D,$37,$E5,$DB,$1B,$AE,$2F,$9E,$32,$FF,$B7,$FC,$24,$7F
	.dc.b	$05,$F0,$7F,$8D,$36,$33,$3E,$CF,$9E,$F3,$7D,$4C,$8F,$23,$E1,$FA
	.dc.b	$59,$1E,$56,$BD,$FE,$73,$F6,$DF,$40,$FA,$4F,$E0,$ED,$9F,$ED,$0D
	.dc.b	$7F,$BA,$87,$0F,$F7,$87,$5C,$5F,$86,$8D,$4F,$63,$F8,$27,$67,$FB
	.dc.b	$AA,$3F,$7D,$5C,$38,$BF,$2F,$1C,$0C,$D7,$0F,$4B,$E1,$75,$7C,$3E
	.dc.b	$19,$5F,$C8,$78,$7F,$A0,$4E,$A3,$1C,$1F,$59,$F9,$D5,$97,$E8,$6F
	.dc.b	$F3,$7E,$62,$47,$08,$FA,$06,$1F,$AD,$18,$7F,$8B,$6B,$E9,$7F,$5E
	.dc.b	$BE,$8B,$DE,$7C,$0D,$24,$F0,$8C,$BC,$A3,$2F,$F0,$69,$E4,$EC,$D1
	.dc.b	$FC,$FF,$E2,$DF,$A6,$BD,$9F,$E9,$91,$86,$C6,$67,$E1,$F3,$E2,$64
	.dc.b	$73,$9D,$3F,$93,$9F,$C8,$79,$9F,$91,$F4,$E4,$FC,$C8,$CB,$F3,$A3
	.dc.b	$2F,$E6,$E3,$F1,$BE,$2F,$E3,$09,$1F,$8C,$6B,$FC,$93,$97,$F9,$A3
	.dc.b	$F8,$48,$FC,$3F,$D4,$72,$8C,$10,$F8,$CF,$E9,$FF,$03,$75,$5F,$8F
	.dc.b	$95,$F8,$59,$CB,$67,$2F,$F3,$EE,$74,$AB,$A5,$7F,$5B,$FA,$27,$B3
	.dc.b	$49,$87,$F5,$7F,$A4,$E0,$FF,$59,$9D,$B6,$76,$FF,$42,$9D,$BF,$D1
	.dc.b	$25,$F0,$5F,$DE,$1B,$1F,$E5,$96,$BF,$B2,$6B,$FE,$68,$73,$DD,$FB
	.dc.b	$16,$F7,$7E,$B1,$B3,$FD,$12,$FF,$CA,$F5,$BB,$F7,$12,$7D,$9E,$AF
	.dc.b	$F7,$BC,$7F,$D3,$55,$FB,$D6,$E3,$FB,$86,$D7,$F5,$0F,$B9,$FE,$D8
	.dc.b	$FF,$60,$8C,$BF,$2B,$BF,$A1,$FE,$07,$9D,$DF,$F0,$9F,$C2,$5F,$17
	.dc.b	$FE,$0B,$6D,$FF,$81,$5A,$FF,$6C,$FC,$2B,$F6,$4A,$33,$5C,$EB,$91
	.dc.b	$FE,$15,$AF,$DC,$75,$FE,$AB,$E0,$FF,$4E,$4C,$3F,$84,$FA,$F5,$FF
	.dc.b	$01,$CE,$DF,$D4,$4F,$F3,$59,$91,$C0,$87,$F3,$DE,$2C,$BF,$CF,$E2
	.dc.b	$E3,$FD,$B3,$97,$FC,$B1,$5F,$E2,$9F,$80,$FF,$27,$CF,$F7,$AF,$A0
	.dc.b	$7D,$07,$F2,$6F,$A2,$78,$BF,$CA,$27,$39,$FF,$80,$27,$2F,$1A,$FE
	.dc.b	$5B,$87,$48,$57,$07,$FD,$4B,$DD,$FE,$52,$73,$5D,$F0,$EC,$D8,$E8
	.dc.b	$7F,$83,$AF,$A5,$F3,$6C,$EE,$79,$9E,$5A,$E6,$F1,$BF,$05,$F1,$6D
	.dc.b	$71,$8F,$36,$1E,$CA,$E6,$7A,$31,$F1,$B4,$CC,$6E,$78,$E1,$E0,$9D
	.dc.b	$47,$7A,$E3,$5B,$BF,$01,$F2,$30,$F1,$DF,$9E,$B7,$87,$3B,$E1,$39
	.dc.b	$3C,$B8,$FC,$27,$E0,$1C,$0F,$3B,$97,$D1,$E6,$79,$9F,$F1,$CB,$5F
	.dc.b	$F0,$2E,$3E,$3C,$32,$D8,$F7,$E1,$E3,$3D,$D7,$D4,$ED,$87,$9A,$4D
	.dc.b	$8D,$63,$F8,$1C,$7C,$04,$C3,$B3,$F1,$EF,$B9,$EB,$3D,$F3,$F1,$9F
	.dc.b	$9B,$3F,$29,$F8,$26,$B1,$F8,$33,$BC,$3C,$87,$8D,$E5,$87,$A3,$1F
	.dc.b	$65,$F0,$B8,$F8,$89,$97,$94,$98,$F8,$63,$DF,$5F,$13,$0E,$A7,$85
	.dc.b	$EB,$1E,$52,$57,$D1,$25,$7C,$F7,$E1,$73,$FF,$8D,$27,$F2,$0D,$C7
	.dc.b	$F5,$CD,$C3,$F5,$8E,$E7,$FC,$EB,$ED,$7F,$9D,$A3,$11,$FE,$A3,$0E
	.dc.b	$23,$5A,$7F,$8C,$9B,$3F,$E2,$83,$86,$1F,$BB,$E1,$7D,$17,$F9,$E2
	.dc.b	$3F,$7D,$3E,$03,$F5,$33,$AE,$9F,$D4,$C6,$67,$95,$71,$D3,$C1,$3C
	.dc.b	$23,$0F,$87,$0F,$D8,$E1,$CE,$3F,$6A,$73,$AF,$F1,$EC,$66,$FC,$73
	.dc.b	$E8,$9C,$DF,$48,$FF,$21,$DF,$48,$C4,$73,$3F,$D3,$B7,$CE,$BC,$B7
	.dc.b	$E8,$C3,$D7,$5D,$4D,$DF,$27,$02,$F9,$3E,$B3,$DD,$3C,$B0,$F1,$36
	.dc.b	$32,$7B,$8E,$73,$C9,$F4,$9F,$B8,$3C,$31,$FB,$87,$B6,$3E,$77,$47
	.dc.b	$0C,$30,$7C,$6B,$F0,$C7,$76,$C6,$63,$C6,$D9,$E8,$5C,$7E,$6D,$F8
	.dc.b	$6B,$B4,$61,$E8,$72,$7E,$59,$9A,$EC,$79,$1C,$C7,$B0,$E3,$19,$7E
	.dc.b	$53,$5E,$E6,$DE,$71,$D4,$E7,$1F,$4F,$1F,$AD,$87,$BA,$73,$8F,$CD
	.dc.b	$AF,$C1,$1B,$C3,$EA,$61,$DE,$33,$3B,$7E,$AE,$1F,$98,$E1,$79,$7F
	.dc.b	$E3,$FC,$3F,$D0,$FD,$DF,$ED,$EE,$1F,$EE,$92,$F8,$3C,$57,$FB,$B2
	.dc.b	$F8,$DE,$31,$46,$27,$99,$FE,$92,$AD,$46,$67,$FC,$17,$DA,$FB,$5E
	.dc.b	$5A,$7F,$90,$9D,$E5,$E5,$AD,$C6,$AB,$86,$3F,$E8,$36,$CF,$FB,$2F
	.dc.b	$2D,$57,$96,$1E,$7B,$E4,$70,$DC,$12,$D7,$FD,$FF,$ED,$BE,$CD,$F1
	.dc.b	$FC,$31,$98,$70,$BF,$F6,$06,$3C,$F1,$E5,$3D,$31,$E5,$87,$09,$C9
	.dc.b	$E0,$F9,$97,$AB,$E7,$6D,$F5,$39,$66,$F6,$F5,$52,$75,$7C,$E3,$EA
	.dc.b	$63,$E9,$AD,$4F,$80,$60,$F7,$FC,$07,$D2,$FB,$1F,$8F,$8F,$4A,$DC
	.dc.b	$EA,$31,$5C,$5F,$F5,$A5,$70,$AE,$D3,$96,$D7,$FB,$95,$FF,$6F,$C6
	.dc.b	$5F,$E7,$4D,$73,$1B,$7E,$AF,$CA,$BD,$EF,$B8,$DC,$3E,$0E,$4F,$96
	.dc.b	$76,$EA,$E5,$7A,$EF,$86,$1F,$4B,$0E,$17,$A8,$FF,$52,$7C,$07,$E1
	.dc.b	$2F,$FD,$DC,$AB,$F3,$6E,$67,$71,$87,$2D,$9E,$2F,$A3,$E4,$A4,$7C
	.dc.b	$8F,$6E,$D7,$D5,$C5,$58,$E4,$5F,$F4,$0D,$D8,$66,$3F,$33,$83,$0E
	.dc.b	$31,$FD,$96,$3E,$4E,$1F,$2D,$D3,$67,$71,$FD,$AF,$AF,$F8,$2F,$DE
	.dc.b	$FE,$CB,$A3,$F6,$6D,$C3,$F1,$F3,$A6,$F5,$FE,$CB,$4F,$E1,$35,$BC
	.dc.b	$3F,$7F,$F3,$FF,$B9,$3F,$7B,$F2,$37,$0F,$DD,$45,$9F,$F6,$A6,$0A
	.dc.b	$C6,$1F,$0A,$5F,$3A,$F8,$37,$BB,$FC,$4E,$31,$F0,$D6,$73,$9F,$E9
	.dc.b	$A7,$2D,$8C,$DF,$ED,$27,$2D,$8C,$E1,$E1,$9C,$B6,$33,$5F,$C2,$A7
	.dc.b	$2D,$8C,$D7,$A1,$5F,$F1,$EF,$8F,$FE,$3F,$F6,$56,$FE,$5B,$6F,$E1
	.dc.b	$B7,$1F,$E6,$AF,$F2,$6D,$7F,$93,$FD,$0D,$BF,$A6,$DB,$F9,$6D,$BF
	.dc.b	$E7,$AF,$EA,$BB,$65,$7E,$AB,$AD,$F0,$B8,$C3,$F2,$95,$70,$FE,$4B
	.dc.b	$5A,$BF,$D6,$4E,$EB,$FC,$69,$CB,$FC,$5E,$7F,$25,$1F,$95,$9F,$C8
	.dc.b	$1F,$93,$FA,$57,$D6,$FF,$37,$6D,$7E,$3B,$F6,$4F,$D7,$65,$F8,$1C
	.dc.b	$15,$A7,$F8,$E7,$E0,$4F,$81,$FE,$80,$DC,$3F,$8D,$56,$EF,$94,$E6
	.dc.b	$FF,$99,$F4,$B1,$FD,$F6,$1E,$1E,$EF,$EB,$DC,$AF,$C1,$B9,$4F,$F6
	.dc.b	$3A,$DE,$1F,$CB,$73,$FE,$41,$9F,$6F,$B4,$F6,$78,$9E,$46,$61,$FC
	.dc.b	$B7,$54,$6A,$30,$F8,$32,$E7,$1F,$D5,$B3,$F1,$5F,$2C,$BF,$A5,$C6
	.dc.b	$27,$A7,$C5,$7C,$23,$FB,$17,$E0,$BF,$CD,$73,$F8,$31,$99,$D9,$FC
	.dc.b	$FE,$74,$FF,$0C,$CD,$3B,$C3,$BF,$E0,$BF,$81,$7E,$8C,$7C,$68,$C6
	.dc.b	$1F,$3B,$E3,$7E,$99,$E2,$F1,$2E,$3F,$06,$2E,$1F,$D5,$72,$FE,$3B
	.dc.b	$39,$24,$78,$3E,$2F,$78,$7F,$03,$D1,$92,$7F,$C0,$1F,$82,$BD,$2A
	.dc.b	$7D,$AA,$7B,$98,$CB,$FA,$F5,$6D,$B5,$B9,$D6,$5F,$DA,$6B,$4F,$D0
	.dc.b	$F6,$FD,$F7,$FC,$CA,$73,$9F,$96,$73,$5F,$E0,$E9,$FE,$5F,$3F,$9B
	.dc.b	$6D,$7D,$DC,$5F,$19,$EA,$F2,$A4,$F9,$9F,$54,$63,$0F,$24,$7E,$5B
	.dc.b	$1F,$85,$5B,$3F,$C0,$B6,$8C,$D7,$F8,$0F,$E7,$7F,$D5,$F5,$FD,$FD
	.dc.b	$B1,$F8,$CC,$9F,$19,$FE,$2B,$9C,$C7,$EC,$DB,$5F,$D4,$9B,$5F,$CD
	.dc.b	$67,$8C,$F5,$F7,$B9,$1F,$D4,$EB,$BD,$7F,$AE,$DA,$F3,$8F,$EB,$1F
	.dc.b	$33,$E8,$BF,$E0,$4D,$BF,$F2,$A7,$E9,$FF,$BB,$9C,$F0,$FF,$90,$6B
	.dc.b	$58,$7F,$8F,$9B,$DC,$BE,$18,$7F,$6A,$9D,$AF,$C9,$3F,$C7,$9F,$02
	.dc.b	$F8,$BF,$F0,$24,$61,$FE,$05,$F0,$1F,$D5,$67,$59,$7F,$78,$AD,$37
	.dc.b	$DD,$D5,$F2,$6C,$7F,$9A,$BC,$2F,$F6,$55,$F2,$C7,$C5,$5A,$A7,$CF
	.dc.b	$BE,$37,$FC,$29,$0B,$F0,$BA,$8C,$C6,$38,$BF,$B6,$61,$FE,$F2,$CB
	.dc.b	$FC,$13,$87,$3F,$17,$18,$FF,$99,$23,$0F,$59,$DE,$3D,$1F,$C0,$7A
	.dc.b	$7E,$5A,$DE,$5E,$26,$CF,$0C,$3A,$E0,$CB,$C6,$F5,$3C,$18,$F5,$6C
	.dc.b	$F0,$BE,$71,$FE,$78,$AE,$38,$77,$BE,$13,$F0,$6F,$84,$62,$B7,$87
	.dc.b	$8E,$BC,$06,$CB,$87,$67,$CD,$8F,$B6,$FA,$CE,$4D,$5F,$4A,$F3,$9C
	.dc.b	$A7,$EA,$BF,$E3,$63,$B1,$67,$E6,$B6,$36,$FF,$8F,$3E,$17,$F8,$1F
	.dc.b	$47,$F6,$A6,$C7,$F8,$F5,$AF,$56,$BF,$73,$DB,$73,$FF,$72,$C3,$FC
	.dc.b	$D4,$D8,$CC,$FF,$91,$E7,$2D,$8C,$EB,$F6,$DB,$A7,$5B,$FB,$7F,$3F
	.dc.b	$FE,$C7,$BF,$AD,$5A,$7F,$14,$D7,$F3,$2D,$8E,$11,$F4,$23,$58,$78
	.dc.b	$B0,$F0,$5F,$6F,$84,$FF,$2F,$C6,$2F,$84,$F8,$0D,$34,$C9,$23,$F4
	.dc.b	$B8,$78,$1E,$44,$C7,$D9,$3F,$84,$9D,$46,$27,$53,$EC,$AE,$EF,$E6
	.dc.b	$78,$1F,$3E,$F5,$7C,$F8,$7F,$A2,$E3,$15,$FA,$EA,$DD,$6A,$3F,$DD
	.dc.b	$7E,$CF,$D4,$7A,$57,$06,$CE,$EB,$8C,$7E,$E3,$AB,$FD,$FB,$8F,$38
	.dc.b	$74,$FF,$41,$7F,$C6,$6D,$7F,$D0,$CD,$7F,$2C,$D7,$F9,$27,$77,$FA
	.dc.b	$3B,$B7,$F9,$8B,$70,$CC,$FF,$A1,$3E,$79,$F3,$F9,$8F,$F7,$6F,$CF
	.dc.b	$3E,$7F,$1C,$EF,$94,$FF,$1C,$57,$F5,$66,$BF,$ED,$7A,$E0,$FF,$B4
	.dc.b	$DA,$FF,$A7,$5A,$FE,$B5,$AF,$FA,$91,$AF,$8A,$BC,$8A,$39,$E7,$FE
	.dc.b	$4D,$6E,$5F,$E2,$FE,$37,$C7,$F0,$9F,$EC,$77,$FD,$B3,$E6,$7F,$D6
	.dc.b	$ED,$F1,$B5,$E3,$48,$FF,$30,$D7,$07,$D1,$1F,$E6,$47,$C4,$D8,$D6
	.dc.b	$5E,$38,$E5,$3F,$E3,$EF,$9F,$FF,$7C,$3F,$0F,$D2,$F9,$78,$BE,$16
	.dc.b	$9E,$78,$93,$E2,$7B,$11,$C1,$CC,$87,$F9,$33,$C3,$FF,$38,$63,$FE
	.dc.b	$71,$EB,$FF,$4C,$CF,$F8,$33,$E7,$FE,$0B,$FE,$52,$6C,$EF,$E0,$7E
	.dc.b	$7D,$70,$AF,$F6,$A3,$5F,$E4,$8D,$7F,$DB,$6D,$47,$F7,$18,$E6,$47
	.dc.b	$FC,$3B,$F3,$EF,$A7,$EF,$EE,$FE,$3B,$1F,$89,$5E,$7D,$32,$F4,$CB
	.dc.b	$D9,$5F,$4F,$23,$49,$3F,$E9,$A8,$CB,$C7,$93,$CF,$3D,$23,$FC,$E7
	.dc.b	$93,$C4,$FC,$FF,$D0,$7F,$41,$C7,$FC,$B0,$DC,$3F,$CF,$8D,$BF,$7F
	.dc.b	$D0,$7F,$AC,$9E,$C6,$3E,$0B,$F1,$11,$98,$FF,$7F,$35,$FE,$9C,$D7
	.dc.b	$FD,$11,$F1,$7F,$E6,$57,$98,$8E,$58,$7F,$AB,$5B,$87,$0F,$4B,$F7
	.dc.b	$5E,$1B,$C6,$63,$86,$E3,$E6,$9D,$4F,$C0,$F8,$DF,$D7,$E2,$F6,$FD
	.dc.b	$24,$BF,$F6,$2C,$65,$FE,$CC,$7F,$BF,$69,$F2,$DF,$C2,$1B,$1F,$E5
	.dc.b	$C6,$BF,$EB,$AC,$5F,$23,$FB,$EF,$9E,$FF,$4C,$CF,$C0,$EB,$7D,$97
	.dc.b	$F0,$26,$C7,$C4,$9D,$AE,$0F,$F6,$84,$62,$3F,$BD,$35,$FC,$A8,$70
	.dc.b	$FE,$61,$AE,$7E,$03,$C8,$47,$C9,$F4,$3F,$2F,$A3,$E6,$72,$49,$F9
	.dc.b	$9A,$7F,$8F,$31,$72,$7E,$93,$8F,$EA,$61,$BB,$F8,$58,$73,$78,$F1
	.dc.b	$7D,$7F,$76,$46,$08,$7C,$97,$2F,$CB,$8F,$8D,$A7,$C7,$F8,$19,$3F
	.dc.b	$2E,$32,$FC,$D8,$CB,$F2,$72,$F8,$BB,$FC,$CD,$3F,$DF,$4D,$8C,$D7
	.dc.b	$09,$CD,$FF,$49,$F9,$67,$D0,$BE,$43,$F2,$7E,$E7,$F4,$6B,$F8,$79
	.dc.b	$FF,$9A,$7E,$27,$85,$EE,$3F,$53,$87,$2C,$32,$48,$F9,$BD,$7F,$9B
	.dc.b	$F4,$7F,$39,$F0,$32,$7E,$8C,$65,$FA,$71,$97,$F2,$11,$F8,$FF,$27
	.dc.b	$F1,$9D,$D2,$7F,$19,$19,$7F,$3A,$7E,$1D,$FE,$85,$D3,$FD,$16,$3F
	.dc.b	$1B,$18,$38,$3F,$2C,$FC,$63,$F9,$9D,$3F,$6B,$F4,$3F,$A2,$DB,$E1
	.dc.b	$F4,$21,$F3,$1A,$37,$DF,$31,$FD,$C4,$E9,$B3,$97,$C0,$3E,$87,$FD
	.dc.b	$20,$FF,$0A,$F4,$9F,$29,$FD,$B1,$B1,$FC,$3A,$76,$F2,$6E,$5F,$62
	.dc.b	$3A,$7B,$FF,$DD,$27,$3D,$3F,$AC,$BF,$25,$71,$F8,$7E,$DE,$1C,$8E
	.dc.b	$D1,$C4,$FE,$D9,$7F,$F0,$06,$3E,$7B,$5F,$E1,$AD,$7F,$CE,$8E,$71
	.dc.b	$F2,$37,$0F,$DE,$B6,$7F,$A9,$E1,$FE,$35,$AF,$56,$3F,$B5,$BF,$E0
	.dc.b	$7F,$42,$F8,$EF,$F0,$7B,$70,$FE,$63,$5B,$7F,$AE,$B5,$FF,$5E,$78
	.dc.b	$5F,$E1,$68,$C4,$FF,$80,$27,$E9,$FD,$0B,$E8,$7F,$C3,$7F,$42,$FB
	.dc.b	$7A,$7F,$87,$A7,$55,$FD,$01,$AF,$2A,$D1,$CB,$C1,$FD,$4F,$D1,$7E
	.dc.b	$5B,$C0,$FD,$8B,$7C,$13,$49,$7F,$AE,$8C,$B6,$3F,$A3,$7B,$52,$3F
	.dc.b	$8D,$B5,$FE,$71,$D3,$78,$BF,$33,$E4,$FF,$8C,$27,$67,$F2,$29,$FF
	.dc.b	$1C,$FC,$3B,$C1,$FE,$6D,$3F,$D5,$3E,$87,$F9,$9F,$D1,$3C,$BF,$E6
	.dc.b	$BE,$7A,$73,$18,$7F,$D5,$D7,$CA,$75,$7F,$E1,$0C,$3B,$67,$FE,$12
	.dc.b	$FA,$2C,$CB,$2E,$CE,$71,$EA,$D3,$96,$3E,$6A,$E3,$87,$49,$F4,$1F
	.dc.b	$E1,$EB,$E6,$E0,$ED,$97,$9D,$E7,$8F,$F8,$93,$2E,$78,$F5,$CB,$C3
	.dc.b	$4C,$BC,$F8,$F4,$77,$87,$96,$59,$78,$E7,$79,$7F,$5C,$6C,$EE,$75
	.dc.b	$87,$C8,$BE,$58,$75,$8F,$0D,$F6,$BF,$14,$6D,$F6,$4F,$BB,$2E,$B1
	.dc.b	$EF,$C3,$E2,$47,$1B,$C9,$2F,$93,$97,$B9,$2B,$E2,$9E,$02,$CF,$62
	.dc.b	$47,$8D,$E5,$1F,$47,$1F,$81,$3C,$E3,$E2,$E5,$F3,$5E,$B7,$E3,$78
	.dc.b	$1F,$16,$34,$F0,$CB,$EB,$57,$D4,$C3,$C3,$3E,$48,$F5,$9D,$F2,$F9
	.dc.b	$51,$DA,$32,$F6,$3D,$87,$E1,$0F,$05,$47,$33,$C9,$D1,$C3,$2F,$B2
	.dc.b	$F8,$4E,$97,$E4,$7D,$07,$D3,$C3,$D7,$5C,$CE,$95,$CA,$77,$5C,$67
	.dc.b	$EE,$61,$F3,$49,$3D,$9F,$84,$61,$EF,$F0,$BF,$E4,$A6,$D7,$E1,$E3
	.dc.b	$17,$FA,$77,$53,$FE,$84,$F8,$0F,$F4,$2C,$62,$11,$8A,$FF,$25,$B6
	.dc.b	$3F,$C6,$E6,$84,$F1,$8F,$AF,$1F,$E9,$0E,$27,$C7,$FF,$83,$AF,$F8
	.dc.b	$0C,$F8,$4F,$DB,$4E,$BB,$78,$C7,$F9,$12,$EC,$E6,$3F,$78,$71,$AF
	.dc.b	$E1,$71,$97,$AC,$79,$76,$CE,$1F,$E1,$E7,$37,$FB,$9B,$E9,$18,$AF
	.dc.b	$1D,$79,$F3,$F8,$31,$8C,$7C,$AF,$A6,$BA,$C7,$F0,$38,$C4,$FF,$97
	.dc.b	$2F,$AB,$CA,$B7,$3F,$92,$36,$7C,$7B,$F3,$4E,$CE,$4E,$23,$E8,$E1
	.dc.b	$E9,$8C,$47,$C7,$3C,$C7,$C6,$3E,$B6,$3E,$03,$DD,$7E,$22,$DF,$6C
	.dc.b	$7D,$55,$F2,$63,$0F,$9C,$C1,$E7,$8D,$E1,$F5,$B0,$F9,$13,$E6,$3C
	.dc.b	$78,$76,$78,$1C,$E7,$D5,$5C,$A3,$F6,$B8,$FC,$87,$96,$3E,$7C,$7E
	.dc.b	$9E,$1E,$B6,$9E,$67,$B4,$AB,$93,$F1,$6B,$93,$D2,$BD,$B1,$C2,$77
	.dc.b	$3F,$50,$E5,$8F,$CE,$71,$1A,$7E,$6D,$78,$4F,$1B,$EE,$BE,$95,$C6
	.dc.b	$3E,$49,$D3,$0F,$81,$3E,$B8,$F9,$0D,$F1,$7F,$D9,$1E,$27,$F9,$DB
	.dc.b	$83,$FB,$EB,$6F,$FD,$35,$C6,$F8,$AC,$62,$C3,$8D,$7F,$A8,$AF,$31
	.dc.b	$89,$FF,$60,$56,$2F,$FC,$BF,$7F,$E8,$0F,$0D,$F3,$DD,$36,$46,$27
	.dc.b	$A1,$B9,$D4,$67,$0F,$E0,$F5,$C2,$7F,$CB,$11,$FC,$16,$74,$FE,$9D
	.dc.b	$AF,$FB,$13,$C5,$7C,$7F,$B2,$3F,$DA,$AF,$BB,$E0,$1F,$42,$FA,$2D
	.dc.b	$5F,$0C,$FF,$DE,$ED,$C3,$FD,$05,$7C,$3E,$9B,$78,$F3,$C5,$1D,$FE
	.dc.b	$8D,$E8,$D3,$5F,$01,$73,$65,$FE,$7E,$AD,$C6,$39,$3F,$D5,$B9,$F9
	.dc.b	$BE,$93,$85,$F8,$B3,$47,$7F,$86,$98,$7F,$AC,$3E,$93,$FD,$E6,$F3
	.dc.b	$3F,$DC,$F9,$27,$78,$7F,$C0,$08,$C4,$FF,$A9,$9B,$19,$9F,$A2,$FA
	.dc.b	$23,$39,$7E,$37,$EB,$1A,$74,$8C,$3F,$47,$E8,$1F,$09,$4F,$A5,$F4
	.dc.b	$2E,$CC,$69,$E7,$CF,$E2,$DF,$08,$C7,$02,$FD,$B5,$FE,$A6,$AD,$3F
	.dc.b	$5A,$33,$97,$FA,$D3,$E1,$FE,$7E,$1D,$7B,$10,$C7,$F3,$AE,$B0,$FF
	.dc.b	$60,$4E,$A7,$31,$A6,$D7,$3A,$E4,$FA,$65,$8B,$E3,$7E,$11,$FA,$3E
	.dc.b	$17,$C1,$62,$F8,$56,$DF,$F6,$BB,$6F,$FB,$49,$D2,$FE,$9B,$67,$71
	.dc.b	$F3,$3D,$4E,$93,$F0,$DF,$DB,$7B,$1F,$B7,$ED,$B7,$F0,$FB,$FF,$61
	.dc.b	$BF,$FA,$0A,$B5,$39,$6E,$1F,$EE,$3E,$FF,$DD,$FD,$1F,$EE,$0F,$E0
	.dc.b	$5F,$33,$6F,$F8,$14,$7F,$08,$F8,$27,$D0,$E7,$0F,$9B,$18,$AF,$C0
	.dc.b	$7D,$2B,$E4,$DF,$0B,$FC,$52,$31,$F2,$16,$73,$5C,$14,$66,$B4,$D9
	.dc.b	$D3,$70,$FF,$7E,$5F,$49,$E5,$5C,$1F,$89,$F2,$FF,$C7,$7E,$97,$F8
	.dc.b	$F3,$63,$27,$C9,$8C,$B4,$F9,$9A,$7F,$25,$FA,$5F,$E4,$CD,$8C,$9F
	.dc.b	$46,$32,$D3,$F0,$7D,$FF,$A9,$3F,$4F,$F1,$96,$32,$7E,$1A,$3F,$07
	.dc.b	$D1,$FB,$28,$C1,$34,$72,$23,$F4,$7F,$42,$F7,$3A,$56,$A7,$31,$FC
	.dc.b	$A7,$E0,$DF,$4B,$4F,$B9,$56,$BF,$41,$F4,$8F,$89,$FC,$AB,$63,$F3
	.dc.b	$32,$8C,$3F,$9C,$6B,$FC,$90,$FE,$A5,$CD,$FA,$C7,$F0,$FE,$17,$F5
	.dc.b	$6E,$77,$D2,$F9,$30,$FD,$E5,$EF,$6F,$76,$1F,$CB,$B4,$FE,$2F,$5D
	.dc.b	$72,$FC,$84,$FB,$CD,$79,$5F,$D1,$E3,$37,$CA,$7D,$1D,$5E,$C2,$5F
	.dc.b	$B2,$B9,$CF,$A7,$3F,$2E,$1E,$7C,$FC,$33,$FD,$02,$BA,$DF,$F4,$2B
	.dc.b	$DB,$F2,$B7,$F1,$1F,$D6,$7E,$97,$F8,$DC,$7A,$63,$2F,$85,$F9,$E4
	.dc.b	$70,$7F,$2B,$97,$0B,$93,$8E,$3F,$06,$73,$1F,$07,$E4,$25,$FF,$1B
	.dc.b	$C7,$F8,$BB,$F7,$A3,$37,$F6,$5F,$BE,$48,$F8,$8F,$D7,$C1,$F2,$1C
	.dc.b	$37,$FA,$B9,$79,$23,$FB,$B7,$D2,$FF,$81,$B0,$FE,$54,$DC,$7C,$D4
	.dc.b	$F8,$1D,$FA,$F2,$FF,$1B,$7B,$BF,$F4,$7B,$5F,$C6,$35,$FE,$E2,$F5
	.dc.b	$C7,$C7,$F0,$3F,$16,$BC,$37,$F8,$F8,$C5,$FE,$1C,$5F,$F3,$5B,$FC
	.dc.b	$33,$6F,$EE,$62,$F9,$1F,$6F,$93,$27,$CE,$FF,$3A,$AF,$E7,$D8,$7A
	.dc.b	$23,$07,$87,$4E,$11,$F5,$F2,$F9,$77,$FD,$3F,$E8,$FF,$D5,$6D,$F8
	.dc.b	$1F,$D3,$68,$F9,$1F,$F1,$15,$6B,$0F,$53,$5F,$F5,$AB,$5F,$D9,$35
	.dc.b	$FF,$0B,$B5,$3C,$E7,$ED,$FA,$8F,$B4,$EF,$87,$ED,$E3,$18,$FF,$65
	.dc.b	$FA,$1F,$4E,$3F,$2D,$B8,$7E,$E3,$D3,$7D,$37,$F0,$68,$FD,$94,$EA
	.dc.b	$FF,$83,$3C,$AF,$FD,$9A,$D7,$3B,$7F,$66,$8C,$47,$F2,$1F,$A6,$FF
	.dc.b	$1E,$7C,$33,$D3,$FF,$02,$C6,$1F,$F1,$AE,$6F,$0B,$FD,$95,$5B,$BE
	.dc.b	$5E,$FF,$F0,$E8,$C6,$6C,$79,$56,$A3,$FD,$D1,$8F,$4C,$39,$BE,$0C
	.dc.b	$39,$3F,$C8,$7E,$99,$F2,$7F,$E1,$5C,$3F,$DD,$B3,$98,$C6,$5E,$39
	.dc.b	$FA,$7E,$97,$27,$FC,$9B,$7C,$BC,$7F,$0C,$62,$F8,$E1,$CE,$73,$3C
	.dc.b	$B3,$B7,$C6,$3C,$27,$19,$CD,$F2,$8F,$C2,$E1,$E1,$55,$C7,$1F,$33
	.dc.b	$71,$F0,$4F,$43,$C0,$58,$D3,$DC,$C5,$F7,$77,$5D,$23,$13,$C2,$7C
	.dc.b	$31,$BE,$56,$1E,$5A,$D1,$27,$51,$97,$99,$F4,$6B,$FA,$A1,$D3,$1F
	.dc.b	$EB,$75,$F1,$63,$2D,$7F,$AD,$7C,$4F,$F0,$1E,$CF,$EC,$2D,$8F,$F1
	.dc.b	$DB,$5E,$AD,$7F,$05,$EE,$B9,$FB,$B6,$3F,$C8,$77,$FE,$6C,$9C,$92
	.dc.b	$3E,$54,$EF,$4F,$A2,$DD,$7F,$B7,$7D,$37,$FB,$06,$BE,$FC,$E5,$FC
	.dc.b	$4B,$5F,$B6,$D8,$E1,$1D,$AF,$89,$1F,$B9,$1A,$BF,$0C,$E5,$B3,$AF
	.dc.b	$86,$FF,$00,$46,$09,$87,$C2,$AF,$CA,$DF,$38,$C5,$F8,$49,$5F,$A3
	.dc.b	$9D,$47,$AA,$BC,$15,$C2,$B5,$7E,$08,$FF,$42,$FD,$33,$D4,$C6,$1F
	.dc.b	$E2,$78,$C3,$FA,$AE,$24,$62,$3F,$4B,$CC,$F6,$BE,$B4,$7E,$A2,$B7
	.dc.b	$7D,$1B,$1C,$EF,$96,$5F,$C3,$70,$E5,$9F,$8A,$9D,$0E,$FF,$F1,$37
	.dc.b	$5F,$F2,$7E,$9F,$F6,$7D,$7F,$88,$BE,$13,$3A,$7F,$26,$BB,$3F,$ED
	.dc.b	$0C,$39,$3F,$E3,$46,$BF,$E8,$E6,$BF,$CB,$23,$FC,$72,$BE,$8D,$F4
	.dc.b	$9D,$39,$DF,$31,$FE,$41,$AF,$F0,$83,$5F,$ED,$15,$C1,$FF,$53,$35
	.dc.b	$FF,$09,$46,$F4,$FF,$02,$B7,$2F,$F2,$25,$6E,$7F,$C6,$8D,$8C,$E1
	.dc.b	$EE,$BF,$33,$E5,$DD,$F2,$39,$BF,$EC,$95,$FE,$25,$F9,$CF,$F5,$B3
	.dc.b	$7C,$6D,$78,$D2,$30,$43,$D6,$E5,$A7,$B1,$F8,$9F,$49,$FE,$EF,$8F
	.dc.b	$7F,$A7,$F1,$EF,$FD,$A3,$1E,$1F,$85,$F2,$4C,$8C,$10,$F9,$2E,$5A
	.dc.b	$74,$3F,$C7,$DE,$07,$F9,$5F,$7F,$F2,$C7,$5F,$FA,$92,$7F,$C8,$5F
	.dc.b	$47,$EE,$7F,$C8,$0D,$9D,$FC,$0F,$C9,$AE,$15,$AD,$7E,$F5,$EF,$8B
	.dc.b	$FC,$ED,$5A,$C3,$8D,$FF,$B8,$63,$13,$AC,$3F,$0D,$F4,$6F,$A4,$ED
	.dc.b	$EF,$74,$DA,$63,$C6,$32,$FA,$23,$2F,$F8,$C7,$14,$EA,$30,$DC,$BF
	.dc.b	$C9,$B8,$3C,$3F,$CB,$7D,$27,$F3,$A9,$FF,$28,$B7,$4F,$F3,$93,$70
	.dc.b	$0F,$A4,$FF,$34,$3F,$EF,$D6,$BF,$E6,$98,$E1,$D4,$79,$47,$1F,$63
	.dc.b	$F4,$DF,$19,$FE,$62,$8E,$51,$F9,$7B,$E7,$8F,$F9,$92,$B5,$5F,$E9
	.dc.b	$5F,$0B,$F8,$64,$7E,$E9,$B1,$FE,$F9,$8F,$6F,$81,$CF,$E2,$AC,$FF
	.dc.b	$A0,$72,$FE,$0A,$FF,$96,$71,$4F,$93,$E6,$24,$FF,$A1,$F0,$47,$7F
	.dc.b	$1D,$F2,$5F,$AD,$FA,$2F,$F6,$45,$FC,$6D,$7F,$D4,$DD,$8F,$A4,$FD
	.dc.b	$93,$67,$5C,$5F,$EB,$DB,$E3,$1F,$E6,$B6,$DF,$2B,$E2,$77,$F7,$E5
	.dc.b	$7F,$B2,$AB,$8B,$FD,$D1,$AF,$58,$FF,$67,$F0,$FF,$33,$E7,$F5,$71
	.dc.b	$CA,$F1,$89,$D3,$86,$BF,$47,$5F,$AB,$9F,$F9,$11,$B1,$C5,$FF,$19
	.dc.b	$67,$FE,$59,$7E,$6C,$73,$7F,$D3,$0D,$F7,$3E,$67,$0C,$8C,$10,$F0
	.dc.b	$3F,$2B,$3F,$97,$EF,$49,$F4,$8D,$FE,$4E,$5F,$EE,$96,$C7,$E6,$94
	.dc.b	$F1,$8F,$F4,$D4,$69,$E5,$59,$CB,$ED,$D7,$86,$7E,$3F,$C6,$3E,$71
	.dc.b	$F1,$9F,$93,$F3,$7F,$9B,$57,$FB,$E3,$E1,$FF,$A7,$DE,$EA,$DF,$0A
	.dc.b	$FF,$15,$89,$D3,$86,$BF,$E3,$06,$BF,$B4,$3E,$87,$9B,$F9,$5F,$80
	.dc.b	$91,$F3,$E3,$F3,$06,$1F,$B5,$18,$7E,$4F,$2F,$F4,$27,$E1,$F6,$7E
	.dc.b	$60,$FC,$F7,$93,$23,$E7,$F0,$7E,$AB,$7F,$CF,$D6,$67,$F3,$55,$FD
	.dc.b	$9A,$FA,$47,$E5,$63,$07,$E5,$3E,$8F,$F5,$B1,$F8,$F5,$F3,$1F,$BD
	.dc.b	$6C,$67,$B3,$06,$66,$46,$4F,$C1,$FD,$1F,$FA,$21,$FC,$2F,$A2,$F9
	.dc.b	$1F,$EC,$8D,$8D,$FB,$7F,$D9,$EB,$84,$7D,$46,$BF,$E6,$06,$BF,$45
	.dc.b	$AF,$F2,$65,$F0,$BD,$BC,$19,$3E,$48,$EC,$61,$FF,$75,$5F,$F0,$C7
	.dc.b	$44,$3C,$51,$FE,$10,$AE,$B1,$F8,$80,$E1,$FC,$5D,$FE,$CB,$AB,$F0
	.dc.b	$DF,$46,$F8,$CF,$F0,$23,$6B,$F5,$35,$B7,$F9,$FB,$FE,$AB,$7F,$CF
	.dc.b	$ED,$7F,$20,$D7,$7E,$17,$F8,$93,$E8,$5F,$43,$FE,$11,$FA,$37,$D1
	.dc.b	$D3,$FC,$29,$3A,$BF,$DC,$B5,$FD,$1B,$5F,$F4,$7B,$5F,$D0,$35,$FD
	.dc.b	$47,$63,$E8,$FF,$65,$F4,$32,$7F,$6F,$39,$6C,$FC,$9E,$EF,$DD,$36
	.dc.b	$75,$3F,$AF,$6B,$FA,$FF,$1F,$FC,$B9,$ED,$7F,$43,$24,$7F,$1D,$E8
	.dc.b	$AF,$F2,$78,$BD,$BF,$C8,$E7,$FC,$69,$F0,$9F,$E5,$98,$FE,$DB,$F4
	.dc.b	$4D,$3F,$79,$F4,$7F,$C9,$B1,$4E,$5F,$EE,$71,$98,$C4,$2F,$A3,$C2
	.dc.b	$BF,$D6,$3E,$1B,$0F,$EF,$7E,$77,$0C,$7B,$37,$CB,$FE,$18,$F4,$C7
	.dc.b	$9B,$6F,$96,$3D,$4F,$1D,$7F,$13,$CB,$94,$7B,$32,$F0,$E5,$63,$FC
	.dc.b	$17,$8F,$49,$F7,$DF,$81,$B3,$E6,$78,$AC,$7C,$57,$CE,$19,$7C,$3A
	.dc.b	$E3,$1E,$53,$E2,$65,$E8,$CB,$8C,$66,$B8,$3E,$A3,$A6,$5F,$2A,$76
	.dc.b	$7B,$23,$86,$5E,$FC,$3C,$B3,$D3,$2D,$12,$3D,$2F,$27,$CB,$1C,$CC
	.dc.b	$57,$C4,$C7,$CD,$3A,$24,$67,$2F,$2C,$FD,$2B,$ED,$3E,$93,$E3,$57
	.dc.b	$80,$F2,$9C,$9F,$8F,$87,$A5,$E8,$7C,$7C,$7E,$04,$EF,$2F,$8B,$3F
	.dc.b	$2B,$1F,$09,$F4,$63,$26,$AB,$8C,$F4,$CB,$DD,$7E,$19,$CC,$F3,$3E
	.dc.b	$59,$CF,$0F,$23,$9A,$F8,$AF,$A0,$D1,$E2,$3D,$19,$7C,$87,$E1,$47
	.dc.b	$81,$F1,$5F,$C2,$72,$F8,$E7,$84,$77,$8F,$81,$5E,$2B,$F3,$9F,$AC
	.dc.b	$F8,$1F,$C9,$57,$C8,$6B,$EB,$6C,$6F,$3F,$EF,$BE,$FF,$F9,$AA,$77
	.dc.b	$5F,$E6,$C8,$C4,$EA,$13,$A5,$3A,$7F,$C0,$06,$0F,$CF,$71,$3E,$8D
	.dc.b	$D9,$FA,$59,$EC,$6A,$32,$FE,$94,$CE,$BF,$C2,$AB,$95,$71,$9E,$D3
	.dc.b	$9A,$E2,$FF,$10,$8E,$3A,$7F,$84,$EF,$8C,$75,$38,$C6,$6B,$8B,$EF
	.dc.b	$AE,$97,$DF,$0F,$DE,$61,$CA,$7B,$61,$C8,$FF,$49,$C6,$6F,$91,$D6
	.dc.b	$F9,$4F,$BD,$B5,$DA,$3C,$97,$DB,$0F,$1C,$7C,$C8,$C5,$70,$BE,$93
	.dc.b	$C0,$E8,$5B,$E8,$79,$AB,$8C,$6F,$2F,$01,$97,$F6,$A7,$AB,$0F,$A7
	.dc.b	$19,$AE,$77,$EC,$AD,$61,$E3,$69,$E1,$8F,$2C,$F5,$AE,$F1,$F8,$B6
	.dc.b	$CF,$B4,$F4,$CF,$79,$E2,$77,$8F,$F2,$95,$FC,$47,$FC,$9D,$87,$B2
	.dc.b	$B8,$47,$89,$AF,$0C,$3E,$8E,$1E,$08,$E2,$7C,$03,$45,$3F,$12,$73
	.dc.b	$C3,$C5,$8F,$DD,$8F,$44,$F3,$7F,$36,$D7,$F0,$87,$4B,$F1,$CF,$1E
	.dc.b	$27,$9F,$FE,$4C,$C7,$FD,$25,$E1,$FE,$8B,$8B,$FD,$2D,$F0,$B6,$7F
	.dc.b	$D3,$93,$72,$FF,$4E,$37,$0E,$6D,$C1,$76,$38,$63,$FE,$42,$F9,$CF
	.dc.b	$D7,$E1,$FA,$5F,$09,$18,$8F,$CD,$C6,$32,$FF,$40,$36,$3F,$D0,$0D
	.dc.b	$CF,$F7,$39,$BC,$E7,$DC,$F8,$2B,$58,$7F,$BE,$2B,$53,$EB,$C3,$FD
	.dc.b	$B3,$F1,$9E,$17,$FD,$FB,$E3,$A3,$11,$AB,$FF,$27,$4E,$5B,$3F,$17
	.dc.b	$E0,$5F,$44,$FA,$3E,$D7,$BC,$3F,$C2,$18,$15,$F0,$AF,$27,$D1,$3D
	.dc.b	$0F,$0B,$6F,$FC,$BC,$DB,$FF,$53,$B5,$FE,$46,$E7,$7F,$E3,$B8,$79
	.dc.b	$3E,$8F,$AE,$1E,$7C,$D1,$E0,$F8,$6E,$E4,$CB,$BB,$72,$F4,$9F,$E2
	.dc.b	$9F,$94,$7C,$17,$1A,$FF,$74,$E6,$9C,$C7,$FB,$53,$46,$1F,$8B,$AD
	.dc.b	$5F,$1C,$3A,$C6,$2F,$FD,$B7,$F3,$AF,$82,$FA,$6D,$8F,$F4,$03,$6B
	.dc.b	$FD,$92,$D7,$FA,$2B,$9A,$F8,$7F,$46,$E1,$FC,$14,$E6,$FE,$A7,$02
	.dc.b	$B5,$1F,$F0,$07,$E1,$7E,$C3,$67,$F4,$51,$8C,$7E,$2F,$2A,$3E,$ED
	.dc.b	$71,$9D,$D7,$06,$C7,$BB,$E2,$7D,$A4,$7D,$AD,$8F,$A3,$3E,$5F,$43
	.dc.b	$D3,$3F,$91,$C5,$18,$BF,$C7,$69,$F0,$23,$18,$72,$AD,$D6,$8F,$D3
	.dc.b	$77,$7F,$B1,$5C,$4B,$3F,$E1,$3D,$7F,$BF,$F5,$BF,$80,$36,$B7,$3A
	.dc.b	$6D,$F3,$BF,$8A,$FE,$D1,$FF,$7F,$75,$BE,$83,$F6,$5D,$DF,$EC,$86
	.dc.b	$C6,$6F,$F8,$1C,$60,$FE,$25,$DC,$F9,$EF,$E0,$11,$87,$F8,$C7,$D1
	.dc.b	$3E,$21,$F0,$18,$5F,$17,$63,$0D,$C7,$E0,$D7,$18,$C6,$9F,$A2,$8C
	.dc.b	$4E,$AF,$AB,$6B,$7C,$7F,$E0,$EA,$DD,$FC,$2F,$92,$FE,$51,$F4,$5F
	.dc.b	$CA,$5B,$87,$C8,$6C,$EE,$3E,$5C,$FF,$2F,$D3,$F9,$87,$D1,$59,$F9
	.dc.b	$ED,$AF,$9E,$DC,$7F,$93,$F0,$A3,$F4,$3D,$BF,$AC,$25,$FE,$B1,$B7
	.dc.b	$FC,$02,$1F,$3C,$F8,$8E,$31,$F8,$63,$2E,$27,$84,$EE,$7F,$77,$5F
	.dc.b	$D2,$DD,$9F,$8E,$9F,$F7,$A7,$D1,$7F,$35,$F8,$03,$F7,$A3,$0F,$F3
	.dc.b	$9F,$A1,$7C,$4F,$F5,$5E,$37,$D1,$70,$C7,$A5,$69,$FC,$DC,$EA,$FF
	.dc.b	$13,$E0,$49,$CD,$F1,$C7,$F7,$7D,$1F,$BA,$7F,$90,$8C,$FF,$8D,$D7
	.dc.b	$6E,$8F,$47,$8B,$E5,$64,$74,$8B,$5F,$CA,$B3,$F3,$C7,$A7,$1E,$73
	.dc.b	$96,$F1,$FF,$68,$D7,$F3,$D8,$73,$3D,$BC,$3F,$0C,$FE,$B9,$F4,$5F
	.dc.b	$C6,$30,$F3,$D7,$D1,$6C,$FC,$7E,$27,$C7,$71,$C3,$E2,$C6,$1E,$A6
	.dc.b	$23,$55,$E9,$9D,$3D,$BE,$37,$EA,$1E,$2F,$8E,$B7,$F4,$A3,$C1,$F3
	.dc.b	$F8,$AE,$4D,$D3,$FC,$E7,$BF,$E3,$A3,$17,$FD,$9E,$30,$FE,$1F,$E8
	.dc.b	$9E,$07,$F7,$B6,$C7,$DE,$97,$BB,$9A,$93,$F8,$2F,$7B,$D3,$8F,$F8
	.dc.b	$53,$0E,$39,$7D,$59,$CD,$7F,$77,$84,$EA,$BC,$F5,$A9,$FE,$F9,$39
	.dc.b	$AF,$44,$E6,$78,$F9,$3E,$73,$F3,$19,$3E,$43,$17,$ED,$9F,$CE,$5B
	.dc.b	$4C,$E1,$E5,$6E,$5E,$26,$DF,$81,$FF,$0F,$7C,$FF,$FA,$AA,$3F,$5C
	.dc.b	$DA,$FC,$95,$3E,$3F,$7F,$01,$2F,$F4,$FE,$9B,$E8,$B1,$87,$E6,$63
	.dc.b	$13,$D2,$BF,$65,$18,$9F,$D3,$E3,$DF,$1F,$BB,$39,$7F,$C8,$7F,$39
	.dc.b	$ED,$C3,$FD,$77,$7F,$E2,$58,$C3,$86,$BF,$EC,$06,$BF,$55,$DC,$FF
	.dc.b	$91,$3B,$9F,$39,$FE,$3F,$9C,$DF,$F9,$0E,$72,$FF,$19,$FA,$1F,$F1
	.dc.b	$B7,$C1,$FF,$6A,$C7,$FD,$93,$18,$F4,$F8,$7C,$27,$FB,$22,$B5,$6A
	.dc.b	$D5,$F0,$F7,$B5,$87,$F8,$22,$73,$E9,$7F,$95,$F1,$CD,$EF,$0F,$F7
	.dc.b	$15,$FE,$3F,$E8,$1F,$21,$FE,$46,$AF,$F7,$0C,$E6,$31,$9F,$9E,$5C
	.dc.b	$2F,$A1,$EB,$0C,$39,$F9,$3D,$A3,$FC,$A5,$18,$7A,$98,$C3,$AE,$3F
	.dc.b	$A1,$C3,$A6,$5C,$6B,$A5,$69,$E1,$86,$70,$E9,$97,$97,$0E,$35,$D8
	.dc.b	$E3,$5B,$BE,$51,$83,$53,$C2,$BD,$D7,$D3,$1F,$85,$87,$13,$C9,$5B
	.dc.b	$8C,$DF,$7A,$E0,$F9,$8E,$B9,$7C,$68,$CC,$78,$70,$EF,$8F,$93,$0E
	.dc.b	$93,$C6,$73,$8F,$E6,$23,$D0,$7F,$07,$29,$E2,$31,$F1,$1F,$BF,$EC
	.dc.b	$E0,$D8,$FD,$0B,$5D,$F6,$7F,$86,$5F,$D2,$E5,$FE,$63,$E7,$F0,$36
	.dc.b	$33,$EE,$A7,$2D,$8C,$C7,$D5,$8C,$65,$F3,$1B,$B7,$F1,$FF,$A1,$FF
	.dc.b	$5D,$DE,$EE,$3F,$2E,$B6,$FF,$A6,$E7,$6E,$67,$77,$D5,$FC,$44,$EA
	.dc.b	$31,$3B,$23,$F9,$B8,$D5,$FE,$8E,$FB,$D7,$E6,$DB,$87,$C1,$BE,$1F
	.dc.b	$0D,$FE,$F6,$BE,$18,$F8,$9B,$87,$E4,$71,$EA,$FE,$8B,$1E,$64,$AF
	.dc.b	$3D,$F3,$9F,$F3,$17,$33,$E6,$31,$CB,$FE,$D1,$FB,$9B,$26,$CC,$AF
	.dc.b	$D8,$56,$DB,$5B,$6C,$F2,$F2,$7F,$A8,$76,$7F,$1B,$F1,$AF,$47,$F1
	.dc.b	$96,$DF,$92,$AE,$3F,$EE,$36,$E2,$AA,$62,$37,$F7,$4F,$5E,$FA,$F3
	.dc.b	$1F,$F3,$DC,$E5,$B3,$97,$FD,$3F,$1F,$E7,$97,$24,$33,$3B,$CF,$FD
	.dc.b	$EF,$3C,$E7,$FC,$5B,$F2,$7E,$6F,$52,$47,$FA,$2B,$34,$FA,$8C,$C7
	.dc.b	$F8,$D2,$70,$EE,$33,$7F,$D0,$67,$33,$E6,$F5,$FE,$2F,$DB,$FE,$D6
	.dc.b	$9B,$FF,$BC,$F7,$4E,$63,$19,$FF,$99,$F0,$F8,$78,$72,$61,$82,$1F
	.dc.b	$EC,$2A,$E4,$FF,$AA,$5B,$6A,$E9,$3F,$E4,$EF,$A8,$FA,$FE,$3C,$8F
	.dc.b	$85,$1F,$68,$C3,$FE,$54,$8C,$3A,$3E,$43,$AC,$BF,$B1,$4F,$C0,$EE
	.dc.b	$F8,$3E,$14,$BF,$F5,$BE,$5F,$D3,$5F,$44,$E5,$F5,$4F,$17,$71,$9B
	.dc.b	$F9,$31,$ED,$9D,$3F,$E6,$8F,$9D,$FC,$89,$1F,$B9,$E4,$FE,$50,$C3
	.dc.b	$B3,$24,$3F,$CE,$2E,$DF,$C2,$7A,$DF,$9C,$FB,$09,$1F,$87,$7E,$EF
	.dc.b	$B5,$F7,$BD,$9B,$3E,$E8,$C3,$F0,$3E,$2F,$F0,$64,$7F,$21,$ED,$FE
	.dc.b	$B1,$FD,$F6,$17,$B1,$8D,$31,$12,$7F,$D3,$53,$A6,$CE,$5F,$F4,$4C
	.dc.b	$27,$94,$68,$E2,$FE,$EB,$F6,$5F,$9B,$0D,$BE,$7F,$22,$5F,$47,$B3
	.dc.b	$87,$3C,$BF,$F7,$9C,$E9,$B3,$97,$A4,$FE,$86,$5F,$37,$E1,$F3,$25
	.dc.b	$72,$8F,$21,$87,$88,$C3,$C9,$3A,$8F,$0F,$89,$EB,$F3,$E4,$60,$87
	.dc.b	$2A,$F6,$46,$A3,$18,$2B,$E2,$19,$CF,$E1,$C2,$B3,$5F,$EC,$BF,$A0
	.dc.b	$FF,$6B,$3F,$DD,$5A,$FF,$93,$9A,$FF,$77,$6B,$FE,$7F,$77,$5F,$49
	.dc.b	$B9,$E7,$C2,$FC,$27,$57,$D7,$FA,$FF,$B1,$F6,$33,$1F,$A3,$19,$7E
	.dc.b	$9C,$65,$F6,$C7,$E3,$6B,$F1,$DE,$2C,$9F,$C5,$C6,$5F,$C6,$C6,$5F
	.dc.b	$C5,$C6,$5F,$C6,$C6,$72,$FC,$4F,$1F,$A2,$7F,$89,$78,$DF,$C8,$AE
	.dc.b	$47,$F0,$28,$FE,$4A,$61,$FE,$35,$18,$7E,$67,$57,$F8,$7D,$6B,$FE
	.dc.b	$89,$7E,$74,$66,$31,$1F,$3E,$3F,$59,$18,$C3,$E6,$47,$EB,$7A,$7F
	.dc.b	$61,$F4,$3F,$E5,$9B,$1F,$EF,$05,$F0,$BA,$37,$E9,$C9,$DE,$1F,$A8
	.dc.b	$AD,$C6,$38,$7F,$D7,$51,$88,$E9,$5C,$5E,$D7,$D5,$E0,$78,$A7,$C8
	.dc.b	$FE,$80,$7D,$0F,$FB,$49,$FF,$21,$35,$FF,$4D,$B5,$FF,$22,$35,$FF
	.dc.b	$66,$3B,$C3,$F7,$ED,$C3,$3E,$EF,$C4,$F3,$7F,$8A,$74,$FA,$70,$FF
	.dc.b	$27,$35,$FF,$6B,$35,$F9,$2D,$7F,$D1,$79,$72,$23,$EE,$1F,$59,$EE
	.dc.b	$CB,$F5,$AF,$F1,$3F,$95,$93,$FC,$32,$3F,$8B,$18,$7F,$6F,$18,$7F
	.dc.b	$88,$E3,$FC,$6C,$8F,$F1,$4F,$7A,$BF,$BE,$8B,$7F,$BC,$9F,$E6,$5E
	.dc.b	$B3,$DE,$FE,$45,$6F,$98,$7C,$D7,$FB,$D7,$E6,$FF,$DF,$F3,$F8,$89
	.dc.b	$CB,$F8,$A8,$CB,$C6,$7F,$DF,$97,$D7,$1E,$11,$8C,$3B,$E1,$CA,$D8
	.dc.b	$72,$8C,$36,$B8,$5F,$2A,$FC,$CE,$1D,$5F,$01,$E3,$F8,$3F,$CE,$B8
	.dc.b	$3A,$B8,$C3,$C1,$3C,$5B,$8F,$93,$0F,$81,$5B,$6D,$F8,$45,$F1,$8F
	.dc.b	$2C,$77,$C7,$D0,$F8,$CE,$13,$B7,$D3,$7D,$DD,$1C,$6F,$C0,$EB,$1E
	.dc.b	$33,$B8,$F5,$9E,$3B,$EE,$F9,$23,$57,$DE,$36,$75,$39,$D4,$70,$4C
	.dc.b	$7E,$1C,$65,$FA,$78,$79,$27,$9C,$69,$F9,$D5,$DF,$0F,$19,$E4,$7C
	.dc.b	$46,$6B,$94,$67,$0F,$39,$ED,$78,$C7,$79,$E3,$39,$8E,$26,$5E,$38
	.dc.b	$7C,$38,$CD,$FA,$71,$FA,$B8,$FA,$A3,$46,$4F,$9C,$7C,$18,$FB,$EF
	.dc.b	$B2,$FA,$B9,$3E,$BC,$FC,$12,$4F,$07,$B4,$71,$BE,$CF,$C8,$9D,$1E
	.dc.b	$43,$F1,$3C,$0F,$9B,$A4,$9F,$2C,$65,$FF,$71,$7B,$5F,$CE,$A3,$11
	.dc.b	$FA,$08,$C4,$FF,$B9,$E1,$18,$78,$9F,$ED,$B8,$D4,$7E,$77,$E4,$1E
	.dc.b	$77,$F9,$4B,$2F,$DB,$5F,$F4,$8A,$D4,$E7,$1E,$33,$FA,$FB,$E3,$CB
	.dc.b	$FA,$78,$E0,$66,$B8,$3E,$2C,$3F,$81,$57,$08,$FD,$C1,$C2,$33,$3C
	.dc.b	$23,$DC,$73,$AE,$CF,$B8,$C1,$C8,$C4,$F2,$8D,$E1,$EA,$6C,$F0,$9F
	.dc.b	$7B,$C3,$0F,$E1,$78,$79,$67,$37,$C6,$BB,$36,$33,$1C,$EB,$E2,$C7
	.dc.b	$86,$36,$64,$F8,$51,$DB,$0F,$74,$7C,$18,$C6,$1E,$6A,$E4,$76,$2C
	.dc.b	$76,$37,$3D,$A3,$2F,$63,$FC,$03,$87,$AE,$78,$46,$2F,$C2,$FC,$13
	.dc.b	$17,$E1,$9E,$B7,$E1,$69,$CA,$57,$D5,$F0,$18,$BE,$B8,$7C,$7A,$E1
	.dc.b	$5B,$AF,$BB,$7E,$86,$CF,$67,$9C,$ED,$F8,$C7,$98,$F8,$11,$B7,$89
	.dc.b	$99,$F4,$36,$3C,$37,$F9,$4F,$1B,$FD,$05,$F6,$1F,$EB,$2D,$7F,$CC
	.dc.b	$6D,$CF,$FC,$F9,$F6,$17,$1F,$F3,$B7,$8E,$AC,$61,$CA,$FF,$D0,$F1
	.dc.b	$89,$5D,$8C,$7C,$B6,$7E,$BE,$DF,$FA,$39,$B8,$FF,$9E,$1B,$97,$F9
	.dc.b	$D9,$F1,$43,$ED,$1F,$3F,$86,$DF,$F8,$C3,$EF,$5F,$7A,$F2,$F5,$72
	.dc.b	$BF,$84,$61,$CE,$3F,$D4,$EF,$F6,$CC,$39,$BF,$C3,$5A,$FF,$70,$1F
	.dc.b	$3D,$FE,$B3,$AF,$F1,$13,$63,$FD,$3B,$C8,$FB,$27,$CF,$FA,$AB,$5B
	.dc.b	$7F,$AD,$9E,$17,$D5,$AF,$F7,$E6,$BF,$E5,$C6,$FC,$CB,$E2,$9E,$0F
	.dc.b	$D7,$8C,$78,$36,$73,$A7,$F9,$52,$72,$D8,$CE,$1F,$EB,$49,$CB,$63
	.dc.b	$38,$FF,$BE,$27,$2D,$8C,$D7,$F9,$C2,$72,$D8,$CD,$27,$2F,$F9,$DB
	.dc.b	$E7,$D8,$7D,$E6,$DF,$F9,$AB,$5F,$95,$F2,$79,$A9,$3F,$6A,$B7,$7F
	.dc.b	$D0,$DA,$FD,$76,$BF,$E8,$4E,$97,$A7,$FE,$C5,$AF,$F3,$BC,$63,$D4
	.dc.b	$FC,$EC,$E6,$BF,$D9,$CF,$FB,$1E,$B6,$E7,$5F,$F7,$6B,$99,$DF,$AB
	.dc.b	$CA,$FE,$97,$CE,$7D,$3F,$9C,$B8,$FD,$16,$DF,$CF,$6E,$3F,$35,$B3
	.dc.b	$F2,$DB,$3F,$2F,$81,$1F,$B1,$BF,$D9,$36,$BF,$DF,$6D,$AF,$F7,$2B
	.dc.b	$6F,$FD,$C6,$FD,$1F,$9B,$95,$FE,$9A,$F8,$17,$CF,$BE,$67,$38,$FF
	.dc.b	$A6,$F8,$18,$7E,$AA,$3E,$E3,$54,$6B,$0F,$B1,$3B,$E0,$78,$59,$BF
	.dc.b	$F7,$CF,$D0,$7F,$0B,$AF,$E1,$8D,$BF,$AF,$48,$C5,$70,$9C,$D7,$06
	.dc.b	$9C,$27,$F8,$B7,$D3,$7F,$1A,$FA,$07,$97,$8A,$FC,$1C,$61,$EF,$3B
	.dc.b	$9D,$60,$E0,$57,$E3,$6F,$84,$E6,$31,$E5,$F7,$ED,$F0,$36,$3B,$B6
	.dc.b	$78,$0D,$57,$FE,$29,$AF,$2D,$FE,$F7,$4F,$DF,$69,$67,$F7,$8D,$C3
	.dc.b	$F7,$4D,$CB,$F7,$0D,$9F,$DB,$36,$7F,$18,$FE,$8B,$A2,$4F,$C4,$F9
	.dc.b	$2B,$5F,$A4,$ED,$FD,$3F,$8F,$FA,$86,$D7,$CA,$E3,$F9,$C6,$2B,$DB
	.dc.b	$B7,$B2,$3E,$97,$64,$AF,$A3,$E7,$5B,$DF,$02,$BE,$C7,$CD,$67,$7F
	.dc.b	$A3,$5F,$1E,$93,$AC,$3E,$DB,$B7,$EB,$3F,$BE,$F2,$BF,$7F,$E5,$59
	.dc.b	$FD,$EB,$6F,$F8,$0B,$70,$FD,$FB,$67,$F8,$67,$80,$8C,$57,$D9,$8C
	.dc.b	$4F,$E3,$3C,$FF,$C8,$7C,$AF,$E4,$5B,$19,$3F,$23,$19,$69,$9A,$F8
	.dc.b	$71,$96,$C7,$F1,$4F,$33,$F8,$CF,$81,$FC,$69,$B1,$93,$F8,$CC,$65
	.dc.b	$A6,$67,$F8,$FC,$65,$B1,$FA,$4F,$A5,$FD,$47,$D2,$FE,$A5,$B1,$93
	.dc.b	$F5,$31,$96,$99,$9F,$B7,$19,$6C,$7F,$2C,$F0,$BF,$98,$F5,$7F,$32
	.dc.b	$6C,$64,$FE,$63,$19,$69,$99,$FE,$6F,$19,$6C,$7E,$F3,$CB,$FE,$01
	.dc.b	$F2,$BF,$C0,$5B,$19,$3F,$80,$C6,$5A,$7F,$04,$FA,$3F,$C7,$FA,$8E
	.dc.b	$CF,$E0,$1C,$D8,$7F,$81,$4D,$C7,$F2,$93,$7A,$BF,$77,$36,$3F,$7B
	.dc.b	$87,$FA,$46,$F8,$B6,$F8,$B5,$7C,$C3,$DB,$79,$19,$8C,$3B,$F9,$97
	.dc.b	$BF,$E3,$BE,$37,$C1,$FF,$54,$E2,$66,$1C,$9C,$98,$9E,$1A,$FC,$18
	.dc.b	$E1,$87,$6C,$BC,$55,$AC,$3A,$57,$83,$2C,$9E,$5C,$BC,$99,$77,$CB
	.dc.b	$A5,$6C,$F0,$95,$CE,$5D,$E3,$33,$C7,$0E,$ED,$8F,$46,$1D,$F2,$F0
	.dc.b	$C7,$0C,$BC,$CE,$8E,$F8,$75,$C7,$C7,$87,$39,$F1,$63,$DA,$F8,$C2
	.dc.b	$F8,$4E,$71,$EF,$3F,$E4,$9B,$F9,$8D,$BF,$A4,$D9,$FF,$2A,$FB,$2E
	.dc.b	$7F,$F1,$F4,$FF,$B7,$E3,$1E,$66,$A7,$33,$FE,$55,$9C,$BF,$EE,$5C
	.dc.b	$1E,$77,$F5,$E8,$C3,$5F,$23,$5F,$92,$D7,$F0,$6F,$1D,$3E,$1B,$73
	.dc.b	$DF,$93,$F9,$5F,$99,$72,$7F,$BD,$E7,$38,$7D,$6B,$E0,$F4,$BE,$15
	.dc.b	$B8,$C1,$B8,$FC,$E4,$F0,$85,$F6,$73,$31,$E0,$7E,$22,$34,$E3,$0F
	.dc.b	$4C,$ED,$FC,$7C,$E9,$CC,$EF,$0F,$4C,$6B,$0F,$34,$FE,$CB,$0F,$26
	.dc.b	$3F,$3B,$1E,$98,$7E,$83,$1E,$8F,$F9,$47,$E5,$DF,$0F,$EF,$E6,$43
	.dc.b	$E1,$F0,$DC,$FF,$81,$4E,$AB,$26,$AA,$4E,$AB,$AB,$7E,$1F,$F6,$3E
	.dc.b	$1F,$F6,$6C,$3B,$FC,$46,$39,$FC,$8D,$BF,$F0,$86,$0C,$31,$7E,$66
	.dc.b	$C2,$A9,$88,$DF,$CA,$3E,$66,$FC,$CC,$C7,$FC,$DF,$39,$6C,$E5,$FF
	.dc.b	$02,$C2,$79,$CC,$9E,$31,$BC,$FD,$95,$CE,$7F,$C1,$B8,$BE,$67,$CF
	.dc.b	$F3,$32,$7F,$CF,$F3,$96,$CE,$5E,$33,$EE,$89,$1C,$23,$57,$F2,$EB
	.dc.b	$94,$FA,$3E,$67,$E3,$FC,$C7,$FB,$1E,$BE,$83,$77,$FF,$49,$64,$AF
	.dc.b	$8E,$70,$B8,$7C,$67,$2F,$FA,$5A,$B9,$3F,$E4,$56,$EE,$AE,$93,$EE
	.dc.b	$F2,$BE,$BF,$CC,$C8,$C1,$0F,$8E,$E5,$F9,$31,$F7,$8C,$65,$EF,$9C
	.dc.b	$4F,$8F,$BB,$DF,$F0,$B2,$30,$43,$FC,$7F,$92,$35,$18,$9F,$8E,$62
	.dc.b	$FE,$3C,$E2,$7E,$BF,$CC,$FE,$40,$8F,$DA,$F9,$0F,$C9,$98,$7F,$03
	.dc.b	$18,$74,$43,$F0,$B1,$F9,$AF,$95,$FC,$DF,$CB,$49,$FC,$2B,$F6,$BE
	.dc.b	$4B,$ED,$FC,$CD,$9F,$A5,$18,$7E,$A7,$CC,$FE,$00,$8F,$E3,$BE,$67
	.dc.b	$F5,$8F,$F0,$0C,$2F,$6B,$1A,$62,$24,$FF,$73,$D7,$FC,$BA,$FF,$08
	.dc.b	$9D,$B9,$E5,$FE,$DB,$6B,$E3,$3F,$E8,$A3,$81,$B3,$FB,$F4,$78,$5C
	.dc.b	$1E,$53,$FD,$5B,$3F,$C0,$3E,$09,$F3,$34,$C4,$49,$FF,$59,$C6,$5E
	.dc.b	$27,$FA,$DB,$E6,$79,$79,$72,$30,$43,$FD,$07,$7F,$E1,$B7,$FD,$7B
	.dc.b	$1E,$19,$F1,$78,$DE,$3F,$87,$95,$FE,$C2,$8C,$BF,$EC,$48,$CB,$D2
	.dc.b	$32,$F6,$8C,$EB,$FC,$8F,$0F,$EC,$B3,$F0,$BE,$63,$F9,$03,$F1,$34
	.dc.b	$72,$7D,$83,$ED,$CE,$A7,$E7,$35,$FF,$0F,$87,$0F,$F8,$89,$AF,$F8
	.dc.b	$BD,$BD,$7F,$EE,$2F,$99,$FA,$1F,$1F,$2B,$FD,$BF,$1F,$4C,$C3,$F2
	.dc.b	$23,$0F,$9F,$C4,$FC,$6F,$8B,$23,$04,$3F,$09,$97,$F3,$97,$DD,$1F
	.dc.b	$94,$99,$18,$21,$C1,$FE,$3B,$E3,$7F,$1E,$64,$7F,$15,$8C,$BF,$C5
	.dc.b	$E3,$2F,$F0,$28,$F9,$3D,$5F,$A2,$E8,$93,$F4,$63,$2F,$CA,$3E,$33
	.dc.b	$5A,$5D,$11,$98,$7C,$CF,$F9,$81,$FC,$EA,$F9,$8F,$D9,$B6,$33,$E9
	.dc.b	$E0,$CC,$C8,$C9,$0F,$D6,$4F,$16,$C7,$17,$F5,$7F,$32,$F8,$5E,$FF
	.dc.b	$13,$27,$31,$FC,$12,$75,$1E,$D6,$BF,$05,$AF,$E3,$1A,$FF,$9B,$3A
	.dc.b	$7C,$5C,$DF,$C1,$BE,$67,$FB,$13,$7C,$DF,$E2,$33,$AE,$DE,$19,$FC
	.dc.b	$78,$F5,$61,$F9,$86,$BA,$3E,$97,$3B,$A5,$97,$F0,$5F,$99,$95,$FC
	.dc.b	$36,$B5,$3F,$C2,$4C,$C7,$EF,$A7,$0E,$EB,$DF,$9F,$EC,$F5,$FE,$20
	.dc.b	$7F,$11,$F8,$AB,$5B,$8A,$FF,$08,$8F,$E5,$BE,$B3,$E0,$D7,$FE,$9B
	.dc.b	$F9,$87,$CC,$7F,$A4,$BE,$65,$F1,$7F,$E9,$49,$CE,$5F,$EA,$19,$CB
	.dc.b	$C6,$BF,$D3,$38,$F5,$C3,$13,$FE,$A7,$8C,$2B,$E6,$AF,$9B,$9B,$F1
	.dc.b	$61,$E1,$7A,$E3,$E8,$9D,$1D,$5C,$5F,$4B,$E6,$DB,$E5,$5B,$C7,$CA
	.dc.b	$B1,$F2,$D6,$E3,$37,$CB,$1F,$0D,$F0,$BD,$C7,$9A,$B7,$87,$67,$A6
	.dc.b	$3E,$49,$E3,$1E,$83,$E0,$4E,$F0,$ED,$1E,$53,$57,$D9,$FA,$13,$B3
	.dc.b	$B4,$66,$3B,$E1,$E2,$61,$CB,$0F,$13,$D7,$1F,$73,$E3,$C3,$C2,$F2
	.dc.b	$38,$57,$29,$D9,$F6,$2F,$C0,$78,$6B,$89,$C9,$CD,$F8,$2F,$DE,$73
	.dc.b	$C7,$E1,$DF,$39,$D5,$F2,$C7,$DB,$1C,$70,$ED,$1F,$07,$0F,$A5,$39
	.dc.b	$38,$C6,$63,$D2,$4B,$F4,$E1,$F5,$E3,$0F,$88,$C4,$6A,$3E,$34,$65
	.dc.b	$FA,$27,$A4,$C9,$0F,$69,$F2,$E7,$D3,$8F,$98,$E1,$1A,$AF,$81,$19
	.dc.b	$87,$B0,$F8,$FF,$C3,$B6,$3F,$08,$D7,$EB,$35,$0F,$72,$C6,$63,$07
	.dc.b	$69,$FE,$91,$3A,$6C,$64,$FE,$EB,$1F,$A6,$F9,$3F,$F2,$A7,$37,$EA
	.dc.b	$27,$BC,$E7,$CF,$4E,$67,$BE,$5F,$BE,$BE,$31,$FA,$13,$89,$FA,$37
	.dc.b	$31,$FC,$12,$32,$F1,$AE,$B1,$C4,$EB,$A7,$C0,$55,$CA,$3D,$75,$C6
	.dc.b	$FC,$98,$79,$DB,$7E,$1C,$3A,$D7,$86,$3A,$9F,$33,$0E,$8F,$C4,$38
	.dc.b	$57,$29,$ED,$1E,$5B,$EC,$D3,$65,$8D,$9D,$E3,$6D,$8D,$B4,$DE,$1E
	.dc.b	$C3,$E0,$5F,$92,$7E,$15,$72,$9D,$47,$A7,$0F,$6C,$62,$3E,$3C,$61
	.dc.b	$F6,$5F,$66,$CF,$69,$D4,$76,$3C,$B1,$CE,$FC,$51,$F6,$23,$86,$1F
	.dc.b	$15,$B1,$F0,$5B,$7E,$D6,$9C,$EF,$BC,$F3,$3D,$55,$CD,$A7,$32,$D7
	.dc.b	$36,$D7,$39,$E1,$1F,$01,$B7,$E8,$8F,$BE,$64,$E8,$5A,$FE,$C3,$CB
	.dc.b	$FD,$F7,$E3,$FF,$27,$AF,$F9,$95,$B8,$7E,$CA,$7F,$C0,$5F,$19,$8C
	.dc.b	$11,$C6,$7F,$BE,$56,$63,$13,$FE,$7E,$AC,$72,$AA,$FC,$66,$3E,$32
	.dc.b	$E3,$FE,$05,$6E,$1F,$E8,$66,$E5,$FE,$79,$7C,$50,$F9,$77,$C4,$7F
	.dc.b	$AA,$23,$F5,$51,$8B,$FF,$25,$36,$3F,$C9,$CD,$7F,$CF,$2E,$6F,$F6
	.dc.b	$BB,$3E,$D9,$F2,$3E,$0A,$D6,$1F,$AB,$9C,$BF,$EA,$CF,$97,$FF,$5A
	.dc.b	$3F,$E5,$96,$E3,$FE,$5D,$6B,$FE,$88,$6B,$FC,$05,$DE,$3F,$53,$E7
	.dc.b	$9F,$1B,$F0,$27,$37,$FE,$C7,$9C,$BF,$EB,$FF,$98,$7C,$5F,$F9,$4E
	.dc.b	$30,$BB,$FE,$D4,$EB,$C0,$B3,$9A,$DC,$62,$B2,$D8,$C9,$D2,$32,$D3
	.dc.b	$FD,$9B,$F2,$FC,$F4,$FF,$3E,$35,$FF,$53,$B5,$FE,$40,$D7,$FA,$0B
	.dc.b	$C7,$4F,$E4,$FE,$F3,$E6,$1F,$13,$F8,$2A,$D6,$1F,$D2,$2B,$53,$F9
	.dc.b	$3F,$83,$7B,$57,$DA,$C3,$6F,$FC,$B5,$3A,$6C,$E9,$A7,$CA,$F9,$4F
	.dc.b	$8F,$F2,$5F,$21,$B1,$93,$E7,$C6,$5A,$67,$5F,$F2,$BB,$FD,$7B,$87
	.dc.b	$EB,$CE,$E7,$22,$33,$3F,$E6,$9A,$FD,$5D,$7E,$B1,$B1,$93,$FC,$DD
	.dc.b	$19,$69,$F5,$39,$3F,$4D,$7F,$39,$AB,$5F,$C5,$BF,$62,$33,$18,$DF
	.dc.b	$FA,$CD,$3E,$81,$EC,$7D,$3B,$FE,$0D,$39,$8F,$E0,$31,$FC,$0B,$E6
	.dc.b	$2C,$FF,$A0,$DB,$8F,$C0,$F0,$BF,$D2,$11,$8C,$7F,$DF,$0D,$AF,$E2
	.dc.b	$9E,$0A,$31,$7F,$C5,$23,$13,$FE,$8F,$F9,$9F,$E2,$9F,$29,$FE,$14
	.dc.b	$8C,$1F,$E9,$97,$15,$E1,$A5,$70,$F2,$BF,$8C,$B6,$F8,$79,$5D,$3D
	.dc.b	$5E,$AD,$8C,$9D,$63,$2D,$3F,$67,$AF,$EC,$33,$FD,$8B,$63,$27,$EE
	.dc.b	$23,$2D,$33,$97,$E1,$B8,$FF,$7F,$3A,$8C,$76,$7C,$27,$E6,$73,$7E
	.dc.b	$99,$FE,$72,$DE,$06,$FF,$AC,$C7,$F4,$7D,$7F,$A9,$EC,$FD,$57,$8D
	.dc.b	$67,$F4,$EF,$CC,$F0,$A4,$60,$F9,$8F,$D0,$F9,$99,$1F,$3F,$CD,$BA
	.dc.b	$7D,$FC,$95,$F5,$BE,$67,$3C,$FF,$E2,$6D,$7C,$B1,$9C,$7E,$C3,$A7
	.dc.b	$FA,$B5,$A3,$F8,$07,$95,$FC,$07,$CA,$B5,$FB,$E6,$E5,$F1,$1B,$8F
	.dc.b	$7E,$04,$62,$BF,$B1,$46,$1F,$E0,$F7,$D5,$FC,$57,$9F,$F8,$EF,$95
	.dc.b	$FC,$7B,$63,$27,$E3,$E3,$2D,$33,$8F,$C3,$8C,$B6,$3F,$8A,$F9,$9F
	.dc.b	$C6,$FC,$1F,$E3,$8D,$8C,$9F,$C6,$A3,$2D,$33,$3F,$82,$8C,$B6,$3F
	.dc.b	$45,$F3,$3F,$A6,$F9,$9F,$D3,$B6,$32,$7E,$9E,$32,$D3,$33,$F6,$E3
	.dc.b	$2D,$8F,$E5,$BE,$1F,$F3,$3E,$BF,$E6,$8D,$8C,$9F,$CC,$A3,$2D,$33
	.dc.b	$3F,$9A,$8C,$B6,$3F,$75,$E7,$7E,$FB,$E5,$FF,$7E,$D8,$C9,$FB,$F8
	.dc.b	$CB,$4F,$E0,$5F,$33,$FE,$74,$F5,$3F,$A6,$DF,$F8,$7B,$97,$F4,$B5
	.dc.b	$FE,$86,$F2,$BF,$1F,$7B,$7F,$23,$36,$B8,$4D,$E7,$FD,$6C,$D1,$15
	.dc.b	$E1,$2F,$98,$7B,$4F,$37,$C1,$3F,$EA,$49,$CC,$62,$1F,$32,$F8,$0E
	.dc.b	$75,$A9,$FF,$55,$67,$FE,$8D,$8C,$6B,$EE,$9F,$06,$3D,$DD,$9F,$EB
	.dc.b	$2C,$3A,$6B,$DE,$F8,$E1,$CE,$F8,$61,$CE,$73,$87,$1C,$7C,$75,$AB
	.dc.b	$E1,$97,$38,$C5,$6F,$1C,$4C,$8F,$19,$E6,$7E,$06,$5E,$48,$CE,$3E
	.dc.b	$07,$46,$2F,$AB,$F1,$B1,$F1,$DF,$5C,$BD,$15,$AC,$3A,$46,$27,$C5
	.dc.b	$7F,$EB,$53,$A9,$70,$F9,$CD,$BF,$94,$EF,$D9,$73,$FF,$8D,$23,$90
	.dc.b	$8C,$F8,$98,$8C,$4E,$DE,$4D,$7E,$D7,$A1,$FE,$EE,$79,$E1,$FE,$1E
	.dc.b	$F2,$F9,$36,$33,$3B,$7D,$0F,$1C,$FE,$C3,$77,$EB,$7F,$8D,$F9,$97
	.dc.b	$27,$FA,$D2,$73,$87,$D0,$BE,$0F,$E5,$5A,$AB,$8B,$F8,$98,$CE,$3E
	.dc.b	$5C,$7C,$38,$F7,$C7,$B5,$F3,$8E,$04,$7F,$07,$87,$47,$C5,$19,$AE
	.dc.b	$71,$AA,$E6,$FE,$B7,$0F,$09,$2B,$C1,$1C,$71,$F6,$DF,$23,$F4,$D8
	.dc.b	$77,$7F,$82,$F2,$7F,$68,$E1,$F3,$B6,$FF,$6E,$D7,$F5,$6F,$F6,$EF
	.dc.b	$7B,$7C,$9F,$B7,$9D,$B5,$DB,$6F,$BA,$9E,$3E,$FA,$BF,$C3,$5F,$33
	.dc.b	$5F,$ED,$8D,$F0,$FF,$B5,$5A,$B1,$CF,$FD,$A1,$B7,$FC,$7E,$EE,$5F
	.dc.b	$E0,$46,$DF,$19,$F0,$61,$FD,$D7,$E5,$FF,$C8,$31,$8C,$7F,$D9,$10
	.dc.b	$F8,$8E,$6D,$9D,$1F,$EC,$A7,$8B,$FE,$41,$6D,$75,$9E,$27,$2F,$4E
	.dc.b	$57,$2C,$97,$D6,$39,$56,$9F,$F0,$77,$CB,$FB,$3F,$66,$FD,$D6,$3F
	.dc.b	$E0,$D6,$E5,$FE,$0A,$9F,$79,$BA,$87,$F9,$1D,$CB,$E1,$9E,$2F,$FB
	.dc.b	$91,$B8,$FF,$8B,$27,$E0,$FC,$D7,$C2,$F9,$A9,$7E,$9C,$D3,$A8,$C4
	.dc.b	$E9,$F8,$91,$97,$E3,$46,$72,$FA,$53,$ED,$ED,$F7,$7C,$1C,$BF,$34
	.dc.b	$65,$F4,$46,$5F,$F3,$86,$7F,$C6,$2F,$E2,$56,$5F,$8D,$5C,$6B,$EB
	.dc.b	$79,$5F,$73,$E6,$64,$7A,$E3,$EF,$98,$7E,$E4,$61,$D1,$F0,$5D,$57
	.dc.b	$E2,$E7,$EB,$7C,$BC,$AF,$9C,$FD,$9F,$98,$91,$82,$1E,$A7,$2D,$3E
	.dc.b	$B3,$F8,$1F,$8E,$FC,$17,$C3,$C9,$FB,$4F,$E5,$BE,$4F,$F2,$FF,$1D
	.dc.b	$67,$EB,$46,$1F,$B1,$F2,$DF,$9C,$23,$F8,$3F,$A7,$FD,$5A,$FC,$C5
	.dc.b	$F3,$18,$5F,$31,$4C,$44,$9F,$D2,$9C,$63,$4E,$27,$9B,$67,$2F,$FB
	.dc.b	$20,$E0,$64,$FC,$B7,$D1,$7E,$87,$E5,$E4,$60,$87,$E0,$9C,$BF,$84
	.dc.b	$8F,$E0,$BC,$8F,$C5,$3F,$CE,$F3,$FB,$A8,$C7,$BB,$A9,$37,$DD,$95
	.dc.b	$B9,$CB,$67,$2F,$E4,$BE,$63,$F6,$DE,$0C,$8F,$D7,$47,$EE,$4C,$3F
	.dc.b	$98,$8C,$3A,$3F,$68,$EB,$9F,$F7,$D3,$F5,$3E,$61,$C1,$FD,$BE,$D1
	.dc.b	$87,$FB,$93,$5F,$EE,$ED,$7F,$C0,$4E,$69,$E0,$FF,$31,$8C,$36,$BF
	.dc.b	$88,$78,$5F,$C4,$7B,$BF,$89,$77,$CC,$7F,$85,$46,$5F,$E1,$91,$97
	.dc.b	$F6,$B1,$F8,$CE,$17,$2D,$F9,$B9,$5F,$88,$9C,$B6,$72,$FE,$16,$13
	.dc.b	$CE,$64,$F1,$8D,$E9,$EC,$8F,$E9,$D5,$D1,$FF,$73,$73,$7F,$46,$F8
	.dc.b	$AF,$F2,$BC,$7F,$33,$6B,$A3,$11,$BF,$52,$46,$0F,$CF,$EF,$FD,$09
	.dc.b	$F9,$BC,$3F,$56,$31,$1F,$36,$30,$7D,$B2,$3F,$07,$BF,$E3,$47,$C7
	.dc.b	$DE,$4F,$C5,$8C,$BF,$1E,$32,$FC,$58,$CB,$F1,$E3,$39,$FC,$38,$67
	.dc.b	$FD,$5B,$87,$F0,$1D,$7E,$72,$4F,$E2,$9A,$FE,$D7,$24,$7E,$4C,$FB
	.dc.b	$9D,$FF,$C2,$3E,$61,$F0,$BF,$95,$6D,$FF,$91,$5A,$FF,$92,$9A,$FF
	.dc.b	$95,$5C,$E5,$E7,$6E,$3F,$E0,$56,$CF,$F8,$27,$E6,$7F,$C8,$95,$FD
	.dc.b	$8F,$6F,$EC,$DE,$1C,$AF,$EB,$D1,$FD,$A4,$C3,$FD,$22,$30,$FF,$63
	.dc.b	$F0,$3F,$B7,$11,$FE,$CD,$A5,$7F,$A7,$C5,$BF,$EE,$53,$FC,$A3,$D7
	.dc.b	$7C,$0A,$7F,$C3,$DF,$2C,$F1,$3F,$C3,$DF,$2F,$FC,$2B,$24,$E5,$FF
	.dc.b	$12,$46,$63,$11,$FE,$6F,$BE,$8F,$0A,$FF,$14,$61,$E0,$C3,$BC,$76
	.dc.b	$EF,$FF,$16,$61,$CB,$1E,$93,$C3,$1E,$73,$DB,$0E,$58,$F5,$3C,$78
	.dc.b	$72,$BE,$91,$D8,$FF,$1D,$4F,$0A,$E5,$7D,$2B,$75,$CC,$61,$E2,$78
	.dc.b	$CE,$DB,$8F,$57,$78,$F8,$5F,$19,$9C,$7C,$18,$7A,$EB,$86,$1D,$6B
	.dc.b	$89,$BC,$3B,$3F,$2E,$77,$87,$58,$D9,$E7,$C3,$AB,$E4,$3D,$57,$D4
	.dc.b	$ED,$1A,$8E,$86,$A7,$A4,$EA,$FC,$35,$B8,$E5,$3E,$18,$E4,$65,$E2
	.dc.b	$66,$78,$9E,$A9,$E3,$5E,$DB,$8F,$84,$F6,$98,$C7,$EA,$C6,$23,$47
	.dc.b	$BE,$BA,$C6,$67,$F1,$17,$E1,$8E,$58,$FD,$C9,$CD,$E0,$F8,$CE,$1E
	.dc.b	$AF,$81,$ED,$3D,$67,$91,$F6,$63,$27,$07,$2F,$47,$53,$D4,$8E,$0E
	.dc.b	$5E,$AB,$E3,$69,$2B,$ED,$46,$5F,$F7,$27,$A7,$FE,$93,$8C,$47,$FA
	.dc.b	$4E,$31,$3F,$EE,$06,$C7,$01,$3C,$1F,$F4,$E4,$7E,$67,$E5,$DE,$67
	.dc.b	$F5,$3C,$7F,$77,$5A,$3F,$49,$3A,$E5,$FD,$04,$66,$7F,$80,$CE,$63
	.dc.b	$F5,$46,$72,$FF,$54,$46,$67,$AC,$B0,$F1,$61,$E1,$BF,$54,$F8,$23
	.dc.b	$0F,$AE,$BB,$C6,$23,$E1,$1B,$C3,$CA,$F5,$30,$79,$0C,$4F,$28,$DD
	.dc.b	$F8,$9B,$3C,$27,$CF,$39,$8E,$4D,$8C,$C6,$EF,$D3,$1D,$63,$66,$4F
	.dc.b	$54,$7C,$4B,$F5,$47,$C7,$3F,$DA,$18,$7C,$79,$E6,$6F,$0F,$2D,$F5
	.dc.b	$3C,$25,$8D,$D7,$2B,$EA,$FC,$08,$E4,$F9,$27,$83,$E4,$38,$57,$69
	.dc.b	$DC,$F6,$69,$DA,$57,$E7,$79,$18,$BF,$0E,$1F,$1E,$B8,$56,$EB,$EC
	.dc.b	$DF,$89,$B3,$D9,$E7,$3B,$7E,$31,$D7,$0F,$33,$E8,$3D,$93,$C5,$E6
	.dc.b	$66,$BD,$4D,$8F,$F7,$F7,$A0,$F3,$BF,$C8,$37,$FE,$D1,$F0,$BF,$C6
	.dc.b	$DB,$7F,$B6,$FC,$2B,$3C,$2B,$78,$FF,$AF,$AF,$2F,$FB,$5D,$B8,$75
	.dc.b	$BB,$5C,$30,$FF,$73,$7C,$E6,$BF,$67,$FF,$2F,$E5,$FE,$E1,$8C,$3C
	.dc.b	$1A,$E4,$FF,$7B,$47,$0C,$BF,$D2,$DF,$2F,$D6,$51,$23,$0F,$D8,$FA
	.dc.b	$07,$C9,$7A,$21,$3A,$6E,$BF,$E3,$57,$99,$8B,$AE,$27,$E1,$B8,$6B
	.dc.b	$C3,$E5,$5F,$23,$DD,$B5,$E6,$D3,$FC,$C9,$C4,$CD,$E7,$EF,$09,$7E
	.dc.b	$0C,$7D,$B8,$F2,$70,$D7,$CA,$D7,$E1,$FC,$9B,$E1,$1F,$19,$F6,$23
	.dc.b	$1F,$15,$67,$38,$79,$4E,$2A,$35,$5B,$6C,$EE,$D7,$D2,$30,$F8,$6B
	.dc.b	$84,$62,$FB,$D7,$F8,$D6,$3B,$DF,$C7,$F8,$A9,$1F,$EC,$5F,$97,$7C
	.dc.b	$96,$2F,$FD,$AD,$9A,$FE,$C4,$7C,$66,$A8,$D5,$FC,$B8,$C4,$F0,$E8
	.dc.b	$7A,$2F,$95,$FC,$DF,$C0,$7F,$4C,$C3,$FD,$29,$18,$F0,$3F,$3F,$5B
	.dc.b	$9F,$85,$6B,$E1,$8F,$B4,$47,$6C,$39,$57,$C3,$C7,$E9,$7C,$AF,$D3
	.dc.b	$6C,$64,$F9,$11,$96,$99,$BE,$DC,$3F,$7E,$75,$18,$E0,$47,$EC,$7D
	.dc.b	$CF,$D9,$7A,$B7,$0F,$D6,$B6,$FD,$19,$BB,$13,$9E,$9F,$F5,$E4,$E6
	.dc.b	$3F,$58,$7F,$7F,$BF,$C2,$5F,$0F,$13,$F8,$07,$AB,$A6,$4F,$D3,$AE
	.dc.b	$15,$F9,$F6,$BF,$B0,$6B,$FC,$6B,$17,$C4,$7F,$0B,$F9,$7F,$E1,$8D
	.dc.b	$BF,$E0,$ED,$AF,$5D,$70,$9F,$E1,$75,$C1,$A7,$02,$CF,$D9,$F7,$5F
	.dc.b	$8E,$FE,$63,$5F,$F6,$E4,$63,$AB,$F9,$1B,$AE,$BB,$5F,$A9,$C3,$F4
	.dc.b	$D3,$96,$C6,$70,$EC,$F5,$8F,$E3,$73,$D5,$FE,$43,$F3,$FF,$CA,$7E
	.dc.b	$7F,$F9,$53,$63,$27,$F2,$68,$CB,$4F,$D8,$73,$FF,$01,$CF,$F2,$A7
	.dc.b	$FB,$D7,$9B,$F8,$34,$67,$AB,$F8,$23,$6B,$89,$FB,$CA,$E2,$FF,$06
	.dc.b	$8D,$67,$F8,$F6,$CF,$E1,$BA,$5E,$67,$EB,$0F,$E8,$5F,$3F,$70,$FD
	.dc.b	$4C,$FF,$22,$6F,$CC,$FE,$CB,$C2,$FD,$9B,$FD,$27,$E7,$E4,$E8,$FE
	.dc.b	$8A,$FF,$4D,$F6,$64,$7F,$48,$F1,$2E,$BF,$82,$9F,$EA,$5F,$37,$8D
	.dc.b	$7F,$37,$D0,$8C,$E1,$FD,$41,$D3,$F9,$17,$F8,$2F,$A5,$FC,$1B,$E4
	.dc.b	$EC,$FA,$9B,$8F,$EA,$B6,$46,$31,$F8,$8D,$AF,$EE,$B3,$A7,$E1,$DF
	.dc.b	$57,$FB,$0F,$CD,$7F,$69,$F9,$BF,$ED,$4D,$8C,$9F,$D9,$A3,$2D,$33
	.dc.b	$8F,$E8,$A3,$2D,$8F,$A3,$F3,$5F,$C8,$3E,$6B,$F9,$0B,$63,$27,$D6
	.dc.b	$8C,$B4,$CC,$FB,$E3,$2D,$8F,$EF,$3F,$37,$FE,$01,$ED,$FF,$01,$36
	.dc.b	$32,$7F,$7E,$8C,$B4,$CC,$FE,$EA,$32,$D8,$FC,$4F,$CD,$FF,$38,$F9
	.dc.b	$8F,$E7,$2D,$8C,$9F,$8E,$8C,$B4,$CC,$FD,$78,$CB,$63,$FC,$23,$E0
	.dc.b	$7F,$86,$7C,$AF,$F0,$D3,$63,$27,$F8,$5A,$32,$D3,$FC,$33,$E4,$9F
	.dc.b	$F4,$AF,$CC,$7F,$1B,$8C,$5F,$FA,$1E,$3F,$8F,$76,$E3,$DA,$46,$EF
	.dc.b	$8D,$7F,$4C,$E7,$FE,$7B,$18,$F2,$BF,$9C,$63,$C6,$F9,$CE,$6B,$77
	.dc.b	$DC,$FF,$14,$FC,$CB,$CB,$78,$DD,$A5,$39,$8C,$7C,$D3,$E6,$7B,$57
	.dc.b	$2B,$E2,$F5,$38,$B8,$D7,$D7,$1B,$BE,$75,$CA,$37,$1E,$1C,$3C,$11
	.dc.b	$CD,$B5,$CA,$18,$78,$F2,$E1,$8F,$7C,$FF,$C7,$78,$70,$6C,$66,$FF
	.dc.b	$C9,$37,$C5,$ED,$19,$7C,$66,$31,$F3,$E1,$D9,$91,$93,$E0,$57,$3C
	.dc.b	$7C,$D8,$F8,$67,$E0,$1B,$8F,$54,$EF,$1F,$2D,$F4,$AF,$F2,$8D,$FC
	.dc.b	$26,$E3,$F3,$9B,$3F,$E1,$DF,$80,$FE,$1F,$E8,$5E,$DF,$F2,$15,$7F
	.dc.b	$92,$E7,$2D,$8C,$BD,$BC,$DF,$F7,$5B,$FE,$51,$C7,$FB,$7F,$91,$FE
	.dc.b	$29,$6C,$66,$76,$F1,$9F,$B1,$1C,$B5,$F8,$2D,$D3,$FC,$DB,$F3,$4E
	.dc.b	$7F,$F7,$F4,$7D,$69,$D6,$1F,$94,$6B,$F9,$26,$D7,$18,$EF,$84,$7F
	.dc.b	$0D,$3A,$BE,$B1,$CE,$75,$87,$91,$FC,$F5,$F3,$9E,$31,$AC,$3C,$8E
	.dc.b	$AF,$9C,$F8,$E3,$4F,$53,$F4,$15,$CE,$FB,$47,$42,$3F,$99,$8D,$5F
	.dc.b	$F4,$2E,$1F,$3F,$27,$F9,$2D,$BE,$07,$EB,$F0,$FD,$F2,$F9,$AC,$CC
	.dc.b	$9C,$BC,$23,$8F,$CD,$7F,$A7,$3C,$3F,$EE,$9F,$35,$79,$FF,$A8,$B7
	.dc.b	$0F,$F2,$BE,$2B,$C6,$5F,$E5,$D6,$C2,$A9,$88,$DF,$CC,$BE,$42,$FC
	.dc.b	$84,$C7,$FC,$CF,$39,$6C,$E5,$FF,$28,$C2,$79,$CC,$9E,$31,$BC,$FF
	.dc.b	$CA,$75,$CE,$7F,$B8,$7C,$D7,$A3,$E0,$FF,$D7,$35,$FE,$58,$6E,$BF
	.dc.b	$DC,$27,$D6,$6E,$A1,$FE,$14,$72,$FF,$9F,$A7,$8B,$FE,$7C,$6E,$BF
	.dc.b	$E8,$19,$F8,$3F,$25,$F0,$BE,$4E,$5F,$C0,$8C,$BF,$0A,$32,$FC,$08
	.dc.b	$CB,$F0,$A3,$38,$FB,$A7,$D9,$D9,$ED,$F8,$99,$7E,$58,$CB,$E7,$8C
	.dc.b	$BF,$E7,$CC,$BF,$84,$5F,$C3,$AC,$BF,$16,$B8,$D7,$D2,$F8,$EF,$B7
	.dc.b	$F1,$D2,$3E,$CC,$7D,$E3,$0F,$D8,$8C,$3A,$3E,$FB,$AA,$FA,$F3,$F5
	.dc.b	$7E,$46,$57,$CD,$7E,$C7,$CD,$48,$C1,$0F,$53,$96,$9F,$55,$FC,$07
	.dc.b	$CD,$7E,$07,$E3,$64,$FD,$97,$F2,$BF,$23,$F9,$6F,$91,$B3,$F9,$48
	.dc.b	$C3,$F9,$6F,$8C,$FC,$D9,$1F,$C1,$7C,$B7,$E9,$D7,$CD,$3E,$1F,$0F
	.dc.b	$FA,$A2,$3F,$D4,$6D,$7F,$D8,$87,$08,$D9,$C3,$E1,$A4,$70,$8C,$BC
	.dc.b	$A3,$2F,$29,$F0,$3F,$95,$F9,$AF,$D0,$7C,$C4,$8C,$10,$FC,$13,$97
	.dc.b	$F0,$91,$FC,$07,$91,$F3,$4F,$82,$C3,$FE,$7A,$AF,$F4,$DB,$5F,$F5
	.dc.b	$33,$B7,$31,$C3,$E1,$24,$60,$FD,$57,$CD,$7E,$D7,$C1,$91,$82,$1F
	.dc.b	$97,$72,$FE,$66,$3F,$78,$63,$9F,$F7,$93,$89,$FF,$68,$FC,$D3,$83
	.dc.b	$F8,$CF,$E8,$3F,$D5,$AF,$FD,$B2,$DF,$05,$72,$F8,$3C,$AF,$2C,$34
	.dc.b	$D7,$31,$C1,$C3,$5E,$3E,$0B,$B9,$DF,$FC,$57,$C0,$95,$FC,$36,$3F
	.dc.b	$8C,$18,$7F,$71,$18,$7F,$DD,$BC,$6E,$7F,$C9,$2F,$97,$A6,$22,$4F
	.dc.b	$08,$CB,$CA,$32,$FE,$52,$79,$3B,$34,$7C,$6D,$BF,$E0,$0F,$C7,$FE
	.dc.b	$6D,$BE,$AF,$F4,$68,$FF,$25,$4E,$F7,$F9,$BA,$A7,$84,$61,$B2,$9D
	.dc.b	$57,$33,$A7,$4F,$EB,$F7,$FE,$7D,$8F,$F9,$A2,$D9,$3A,$FE,$19,$BE
	.dc.b	$39,$3F,$5C,$5F,$F4,$77,$EE,$35,$FB,$71,$B7,$AB,$F3,$BC,$69,$1F
	.dc.b	$4A,$3E,$89,$87,$E5,$46,$1D,$1F,$61,$D6,$1E,$9C,$3D,$7A,$FE,$2B
	.dc.b	$B7,$ED,$F7,$CC,$3F,$37,$1F,$04,$F2,$64,$69,$23,$DD,$F3,$6F,$86
	.dc.b	$FE,$2D,$F3,$6F,$98,$A7,$F1,$99,$D5,$7D,$73,$F4,$6E,$1F,$CF,$35
	.dc.b	$FD,$13,$5F,$D2,$B5,$FE,$C9,$E3,$7F,$71,$F1,$E4,$7F,$69,$8F,$EE
	.dc.b	$86,$1F,$EA,$91,$87,$FB,$77,$0F,$F7,$92,$3F,$DC,$72,$AF,$F5,$B8
	.dc.b	$BB,$FF,$7D,$9F,$E6,$9E,$C3,$E1,$D3,$FB,$EF,$9A,$79,$1F,$BC,$F9
	.dc.b	$BF,$E3,$51,$8C,$BF,$81,$C6,$27,$F8,$0E,$1C,$EF,$94,$63,$0E,$B6
	.dc.b	$C3,$92,$C3,$93,$D3,$C3,$FE,$17,$3A,$C7,$A4,$F1,$C3,$A6,$3D,$A3
	.dc.b	$81,$4E,$97,$CA,$77,$1C,$4F,$E3,$75,$D2,$FC,$78,$78,$A7,$89,$CE
	.dc.b	$75,$5C,$AB,$77,$D6,$BC,$27,$51,$5C,$1E,$F8,$FC,$2A,$E3,$1E,$7C
	.dc.b	$3C,$15,$B8,$E3,$5B,$C3,$E9,$61,$DE,$7C,$A7,$7A,$E0,$71,$BF,$51
	.dc.b	$D2,$F9,$4F,$8B,$1F,$05,$F3,$AD,$47,$BE,$B7,$5D,$B1,$F3,$BD,$E3
	.dc.b	$38,$77,$C3,$91,$E5,$7C,$71,$CB,$0F,$1C,$E6,$38,$99,$78,$D7,$87
	.dc.b	$08,$FC,$BA,$E2,$E3,$0F,$C2,$5F,$18,$F5,$99,$C7,$EB,$1F,$21,$C9
	.dc.b	$E9,$BD,$3F,$62,$BD,$A7,$48,$CC,$7C,$19,$E4,$F1,$24,$F0,$8C,$BC
	.dc.b	$63,$31,$82,$3F,$44,$E5,$EB,$BE,$53,$F2,$AD,$8D,$35,$FC,$1D,$FF
	.dc.b	$72,$8C,$47,$F7,$4F,$63,$15,$FD,$BD,$B1,$FD,$CD,$AF,$F7,$66,$BC
	.dc.b	$4F,$CC,$7C,$CB,$D0,$74,$FE,$AE,$75,$9F,$EE,$2B,$53,$99,$5F,$EC
	.dc.b	$67,$31,$FA,$33,$8E,$7F,$E0,$68,$CD,$FF,$02,$9C,$9F,$E0,$E8,$CC
	.dc.b	$FA,$E3,$CB,$87,$91,$E8,$78,$0C,$E1,$E3,$C3,$C5,$3B,$C3,$B5,$6D
	.dc.b	$F5,$46,$6F,$B4,$62,$F9,$57,$43,$BD,$F4,$8C,$CE,$E7,$26,$DE,$F3
	.dc.b	$B8,$CC,$6C,$CB,$BA,$F0,$9F,$16,$FC,$91,$F6,$4D,$16,$79,$4F,$08
	.dc.b	$F9,$13,$C1,$FB,$C7,$02,$CF,$07,$94,$F7,$BF,$5B,$63,$CE,$D9,$F3
	.dc.b	$9C,$63,$9C,$64,$F1,$2C,$3E,$CB,$D8,$F2,$D7,$66,$9D,$8B,$5D,$AB
	.dc.b	$A4,$71,$3A,$5F,$B5,$FC,$39,$8A,$E8,$D3,$A1,$6B,$A3,$DA,$BC,$F7
	.dc.b	$E6,$8F,$59,$93,$FD,$B1,$E0,$FE,$3F,$4F,$F7,$23,$70,$FF,$1B,$F8
	.dc.b	$BF,$E4,$B7,$FD,$C1,$19,$AE,$31,$83,$FD,$C9,$E1,$E7,$C2,$FF,$2F
	.dc.b	$55,$B7,$CD,$E7,$E0,$7F,$63,$87,$FB,$AE,$30,$ED,$AE,$4F,$F0,$A5
	.dc.b	$A7,$8E,$3F,$EA,$AF,$9B,$FF,$31,$57,$F8,$F5,$B1,$FE,$75,$6B,$FE
	.dc.b	$79,$6B,$FE,$7C,$77,$9F,$C3,$F9,$E7,$C3,$7B,$E9,$5A,$BF,$F3,$15
	.dc.b	$69,$FF,$14,$4B,$E6,$5F,$23,$FE,$55,$8F,$F4,$4C,$62,$7F,$CA,$4D
	.dc.b	$8F,$F4,$53,$5F,$F4,$73,$9C,$7F,$82,$7A,$4F,$9B,$7C,$07,$96,$B5
	.dc.b	$87,$FA,$6A,$B5,$3F,$5F,$E5,$9F,$0E,$F8,$DC,$2F,$8C,$B1,$86,$E1
	.dc.b	$FE,$43,$AE,$31,$8D,$3F,$C5,$73,$FE,$A8,$9C,$BB,$BE,$CD,$AE,$1A
	.dc.b	$7F,$1D,$AD,$DF,$FB,$8B,$3F,$94,$FC,$2E,$0F,$90,$FF,$AD,$23,$2F
	.dc.b	$C4,$8C,$BF,$3D,$FC,$A6,$0F,$9E,$79,$FF,$16,$73,$18,$9F,$C5,$FA
	.dc.b	$4F,$8D,$FC,$C7,$C2,$3C,$2F,$CD,$4E,$6B,$F2,$33,$97,$8D,$70,$AD
	.dc.b	$B6,$3E,$16,$1F,$3B,$E3,$7E,$7B,$63,$27,$C7,$8C,$B4,$CE,$7F,$38
	.dc.b	$6B,$F6,$A7,$71,$98,$C4,$FE,$9B,$A1,$EF,$FE,$BF,$D5,$FD,$83,$67
	.dc.b	$4D,$BE,$DD,$7F,$AD,$8C,$71,$7F,$AB,$6F,$FB,$D5,$FF,$AA,$5B,$7F
	.dc.b	$EE,$16,$BF,$EE,$56,$BF,$EE,$87,$73,$FE,$F0,$F1,$7F,$59,$18,$AF
	.dc.b	$F6,$2F,$B5,$6F,$FD,$85,$93,$E2,$BF,$84,$7C,$DF,$F0,$96,$DF,$BD
	.dc.b	$B5,$EA,$AE,$13,$F0,$EB,$83,$4E,$05,$9F,$9D,$D6,$F9,$BF,$CC,$C6
	.dc.b	$1F,$D3,$4F,$FA,$96,$75,$D7,$FC,$8D,$E1,$D5,$6B,$8E,$DF,$B8,$78
	.dc.b	$E1,$CE,$F5,$5F,$C8,$FE,$6F,$F9,$27,$CD,$DC,$BE,$D3,$6F,$D8,$DC
	.dc.b	$7F,$C0,$DC,$5F,$C8,$E3,$F7,$7C,$DF,$BF,$9F,$CB,$57,$E1,$7B,$7F
	.dc.b	$34,$FE,$1E,$32,$7E,$CE,$32,$FE,$D4,$CE,$1F,$BD,$8C,$BF,$BF,$8C
	.dc.b	$CF,$E5,$27,$2F,$E5,$A3,$3C,$BF,$C4,$2F,$C9,$D1,$FA,$EF,$3A,$63
	.dc.b	$F9,$7F,$9B,$B8,$FE,$4B,$AD,$F0,$5F,$B4,$F9,$BF,$DA,$B6,$FF,$3D
	.dc.b	$F3,$7F,$D2,$CC,$5F,$E0,$5F,$E9,$DE,$DC,$9F,$D2,$78,$D6,$BF,$0D
	.dc.b	$3F,$D4,$FE,$6F,$1C,$1F,$33,$A9,$19,$C3,$F5,$8E,$95,$FE,$5A,$3F
	.dc.b	$84,$FA,$7F,$C2,$BE,$52,$D7,$F0,$86,$D7,$B5,$B8,$7D,$AD,$91,$8A
	.dc.b	$EB,$CC,$9C,$C7,$F1,$2C,$3B,$3F,$BF,$F9,$BF,$ED,$9F,$37,$FD,$B5
	.dc.b	$B1,$93,$F8,$24,$65,$A6,$72,$FD,$44,$65,$B1,$FC,$83,$E6,$FF,$92
	.dc.b	$FC,$DF,$F2,$66,$C6,$4F,$E4,$B1,$96,$99,$9F,$B9,$19,$6C,$7F,$17
	.dc.b	$F9,$BF,$F0,$27,$77,$F8,$15,$B1,$93,$F8,$E4,$65,$A6,$67,$F8,$04
	.dc.b	$65,$B1,$FC,$E3,$E6,$FF,$9E,$FC,$CF,$F3,$E6,$C6,$4F,$E7,$B1,$96
	.dc.b	$99,$9F,$CA,$46,$5B,$1F,$CB,$FC,$0F,$F0,$E7,$95,$FE,$1D,$6C,$64
	.dc.b	$FE,$69,$19,$69,$FC,$DF,$CA,$FF,$40,$7B,$7F,$DF,$2F,$F9,$0C,$63
	.dc.b	$D0,$FC,$1F,$3F,$EE,$AB,$5C,$9F,$E2,$3A,$D5,$F0,$C3,$FD,$51,$87
	.dc.b	$09,$CD,$7F,$8A,$7C,$0B,$8A,$78,$E0,$F9,$A7,$C5,$3C,$8E,$73,$FA
	.dc.b	$09,$CC,$62,$1F,$36,$F9,$AE,$75,$A9,$FF,$54,$E9,$FA,$87,$B6,$1D
	.dc.b	$5C,$9D,$5C,$6D,$F0,$E3,$AD,$F4,$9C,$D7,$3B,$ED,$7B,$6C,$77,$C3
	.dc.b	$C9,$B7,$9E,$73,$8F,$7C,$B1,$8F,$48,$FD,$ED,$F0,$7C,$A7,$A2,$76
	.dc.b	$47,$31,$BC,$7C,$55,$CE,$78,$1C,$AF,$C2,$FB,$30,$E5,$5C,$30,$B8
	.dc.b	$FA,$72,$F8,$11,$DE,$FE,$51,$D0,$B8,$7C,$F6,$E3,$F3,$5D,$FC,$0F
	.dc.b	$F0,$EF,$23,$8C,$7F,$1A,$8C,$38,$6B,$C3,$B7,$FC,$92,$F0,$BF,$EE
	.dc.b	$9E,$3F,$28,$FC,$47,$9B,$FE,$29,$74,$D7,$CA,$D7,$EC,$3B,$D7,$EC
	.dc.b	$37,$5C,$D7,$E3,$7E,$6D,$CF,$FE,$FD,$9C,$E1,$F2,$EF,$83,$F9,$56
	.dc.b	$AA,$E2,$FE,$06,$33,$8F,$AA,$39,$61,$DF,$0E,$D1,$CE,$38,$11,$FC
	.dc.b	$1D,$F4,$78,$93,$1F,$5C,$F3,$8E,$11,$87,$58,$FE,$8E,$B9,$4E,$49
	.dc.b	$1E,$E7,$F4,$51,$A8,$C1,$27,$CC,$79,$2F,$D2,$FD,$4B,$F0,$CE,$E3
	.dc.b	$E2,$4E,$B0,$F8,$58,$78,$27,$9B,$E0,$9D,$63,$F2,$09,$87,$B6,$BD
	.dc.b	$67,$0C,$3B,$63,$E7,$8C,$10,$F2,$1E,$73,$CA,$67,$1F,$1C,$FC,$5C
	.dc.b	$3C,$37,$AA,$E3,$8F,$95,$C5,$F0,$7B,$61,$E1,$77,$87,$C1,$9D,$47
	.dc.b	$C2,$38,$E3,$EC,$7C,$87,$B0,$E2,$74,$C7,$C6,$E6,$79,$DF,$77,$A9
	.dc.b	$2B,$F2,$71,$97,$B6,$1F,$3E,$BC,$47,$38,$CE,$1F,$06,$3C,$B8,$FC
	.dc.b	$1B,$E4,$E2,$3C,$F7,$CF,$0F,$04,$FD,$58,$C6,$1D,$2B,$FC,$33,$7C
	.dc.b	$9F,$05,$6F,$0F,$04,$60,$EA,$E2,$F9,$CE,$E7,$9E,$3F,$CF,$78,$3F
	.dc.b	$8A,$E1,$FA,$56,$C6,$67,$AC,$61,$E4,$7C,$2B,$FC,$FF,$CF,$7E,$6A
	.dc.b	$FF,$A6,$4F,$F0,$3F,$07,$06,$78,$A4,$7F,$8D,$30,$F0,$C7,$78,$F7
	.dc.b	$B5,$FE,$08,$DA,$EB,$1F,$BC,$8E,$6F,$EF,$BE,$73,$F8,$1F,$43,$D1
	.dc.b	$7C,$56,$1B,$5F,$D9,$A7,$53,$9D,$F4,$C9,$D5,$70,$7F,$B0,$D7,$5A
	.dc.b	$E9,$F3,$BD,$7C,$B9,$5C,$23,$2F,$28,$CB,$CA,$D3,$E5,$F3,$A4,$F8
	.dc.b	$1F,$47,$9B,$88,$E1,$17,$87,$FC,$1B,$6A,$C6,$3F,$EB,$6A,$B6,$CB
	.dc.b	$E2,$57,$F2,$8F,$82,$7D,$CF,$F7,$9C,$B9,$F9,$5F,$1C,$D4,$67,$2F
	.dc.b	$F3,$5D,$70,$72,$4C,$7E,$3F,$4F,$CF,$F2,$24,$7C,$A7,$E9,$79,$18
	.dc.b	$8F,$85,$17,$2F,$5C,$FF,$24,$FB,$87,$CD,$FE,$37,$E7,$E6,$1F,$E1
	.dc.b	$B9,$CB,$67,$FA,$F2,$8C,$BF,$DA,$9A,$FF,$74,$6B,$FE,$1F,$77,$E7
	.dc.b	$3A,$71,$5F,$E2,$38,$C5,$7F,$A3,$2B,$75,$C7,$E7,$34,$FF,$B9,$BC
	.dc.b	$BB,$5F,$47,$97,$FA,$96,$1F,$EC,$46,$D6,$E7,$FD,$2F,$39,$AE,$3C
	.dc.b	$9F,$9B,$EE,$B8,$70,$BE,$95,$DB,$8B,$ED,$F0,$6B,$BA,$57,$4C,$FE
	.dc.b	$CF,$22,$B8,$36,$74,$2F,$9B,$86,$AD,$3D,$5D,$DE,$E8,$FA,$1F,$11
	.dc.b	$78,$1F,$27,$4C,$44,$9F,$F0,$A4,$A3,$EC,$B6,$33,$18,$38,$3E,$1C
	.dc.b	$FE,$C4,$77,$F9,$8F,$5F,$07,$C6,$F0,$25,$FA,$75,$F9,$1A,$FE,$82
	.dc.b	$B7,$39,$7E,$6F,$81,$F1,$2F,$E7,$FC,$FC,$AE,$51,$96,$90,$F2,$1F
	.dc.b	$7F,$9F,$E2,$7C,$9B,$D6,$FC,$AB,$67,$FC,$68,$D7,$FC,$7E,$D7,$FC
	.dc.b	$92,$E7,$8B,$F2,$30,$8C,$6F,$FE,$77,$6C,$7F,$B8,$3E,$53,$FC,$79
	.dc.b	$5C,$1B,$5F,$E7,$D6,$CE,$FC,$6F,$CB,$D7,$F2,$9F,$91,$FB,$95,$AA
	.dc.b	$34,$E3,$3B,$8F,$C4,$8F,$CE,$DE,$9F,$ED,$1B,$1F,$E5,$36,$BF,$E5
	.dc.b	$E6,$BF,$E6,$87,$33,$F0,$1B,$8F,$FA,$55,$B3,$FE,$FE,$FC,$EF,$FD
	.dc.b	$EA,$F8,$BC,$D7,$8B,$F9,$78,$7C,$27,$E3,$57,$C9,$BC,$9F,$F5,$04
	.dc.b	$E6,$31,$23,$8F,$57,$FA,$97,$0E,$BA,$66,$FA,$46,$29,$F2,$5C,$F1
	.dc.b	$FF,$56,$36,$3A,$3C,$32,$F3,$37,$0C,$E3,$FE,$B3,$CB,$C7,$7D,$23
	.dc.b	$18,$74,$7C,$98,$72,$6D,$F2,$AF,$55,$F2,$C3,$C5,$1C,$CF,$F4,$75
	.dc.b	$F1,$6D,$7B,$6F,$BC,$53,$9E,$1F,$22,$3B,$C7,$C6,$38,$9D,$4F,$9F
	.dc.b	$1E,$77,$89,$E5,$18,$FC,$9C,$3E,$14,$71,$C7,$D1,$7E,$28,$E9,$5B
	.dc.b	$8F,$0E,$5E,$C2,$38,$3C,$B7,$F0,$08,$E0,$95,$E4,$3C,$18,$79,$DF
	.dc.b	$2E,$5F,$15,$B9,$7C,$37,$DA,$5C,$BE,$BE,$3E,$A3,$D7,$8F,$82,$7B
	.dc.b	$38,$C7,$C3,$1C,$8F,$09,$F1,$72,$F7,$3F,$32,$79,$3B,$79,$3F,$42
	.dc.b	$BE,$0B,$EB,$7C,$55,$DA,$FE,$5C,$F8,$A3,$C3,$7F,$99,$AE,$33,$E3
	.dc.b	$78,$BD,$67,$E9,$46,$A7,$84,$7E,$0B,$87,$F5,$F5,$F8,$26,$BF,$20
	.dc.b	$38,$9F,$D4,$3A,$8F,$D8,$78,$FF,$B1,$8C,$42,$31,$7F,$EE,$66,$C7
	.dc.b	$FB,$C1,$AF,$EE,$4E,$07,$E5,$4F,$EB,$DF,$27,$FD,$F2,$7F,$82,$47
	.dc.b	$80,$D7,$5A,$75,$8F,$79,$CC,$61,$F0,$18,$E1,$FC,$23,$CF,$4F,$E2
	.dc.b	$17,$CE,$7F,$41,$39,$BF,$34,$E6,$3F,$50,$78,$F0,$E3,$87,$78,$F6
	.dc.b	$63,$F9,$1A,$E1,$5E,$A9,$F7,$E3,$C3,$1E,$F1,$8C,$7B,$E3,$FE,$1C
	.dc.b	$AE,$18,$FE,$CF,$1F,$14,$62,$FF,$75,$18,$BF,$85,$3D,$67,$CE,$75
	.dc.b	$3F,$C8,$F7,$D6,$7E,$46,$3B,$3A,$15,$E6,$79,$31,$F3,$E3,$E3,$AE
	.dc.b	$57,$F3,$B0,$F0,$1E,$52,$DF,$63,$94,$7D,$F3,$E5,$38,$C3,$83,$A9
	.dc.b	$B3,$E8,$C3,$C2,$FC,$0B,$F0,$BF,$40,$C4,$6A,$FD,$51,$F8,$26,$BB
	.dc.b	$AF,$2B,$E9,$BF,$01,$F0,$1F,$01,$F2,$B0,$F8,$17,$E2,$AE,$33,$CE
	.dc.b	$B8,$BD,$DF,$61,$F0,$5E,$77,$FD,$7F,$D4,$7C,$5F,$F9,$6B,$7F,$F0
	.dc.b	$5C,$63,$1F,$EF,$3F,$74,$FA,$1D,$47,$EA,$67,$58,$FF,$87,$83,$87
	.dc.b	$F5,$C6,$30,$FF,$0B,$BC,$5A,$FF,$6B,$8E,$0F,$FA,$33,$E8,$FF,$D6
	.dc.b	$F5,$FE,$4F,$6C,$7F,$8B,$3F,$31,$7D,$43,$E0,$FC,$B3,$FE,$57,$AD
	.dc.b	$E9,$FE,$76,$E2,$46,$31,$ED,$EB,$70,$CB,$9E,$3C,$AF,$73,$D3,$17
	.dc.b	$CA,$3E,$5F,$0D,$C7,$CD,$58,$99,$1F,$0A,$BF,$C1,$15,$B7,$0D,$7E
	.dc.b	$3B,$5F,$F5,$6F,$CF,$3E,$87,$C8,$DA,$FF,$3A,$FA,$5F,$ED,$DE,$2C
	.dc.b	$D4,$9C,$FC,$37,$FB,$B7,$0F,$F2,$9E,$8C,$79,$47,$F9,$5B,$15,$6F
	.dc.b	$83,$E4,$53,$E8,$7C,$27,$CE,$AD,$F1,$E3,$2F,$F4,$25,$6D,$F0,$F5
	.dc.b	$67,$2E,$D9,$7F,$1B,$CB,$FC,$B3,$F0,$5F,$4E,$7F,$C5,$8D,$4F,$F8
	.dc.b	$DA,$FF,$C9,$6E,$4F,$C1,$C7,$E3,$9A,$FF,$3D,$6B,$D3,$87,$F1,$EF
	.dc.b	$47,$EE,$43,$DA,$B1,$87,$E7,$FC,$EC,$BF,$97,$EC,$DC,$3E,$5E,$8A
	.dc.b	$F8,$99,$BD,$9D,$69,$FE,$C6,$DF,$E6,$C6,$BD,$4D,$F7,$5D,$3E,$BB
	.dc.b	$73,$F8,$ED,$C3,$96,$3F,$EA,$5A,$F8,$7E,$F7,$A8,$F5,$FB,$D6,$32
	.dc.b	$7E,$47,$0F,$97,$3F,$0B,$84,$D2,$4F,$F6,$EE,$0F,$95,$EC,$7F,$B7
	.dc.b	$BC,$3F,$8D,$87,$D6,$BC,$56,$AF,$84,$B7,$FC,$5F,$AD,$A8,$C6,$88
	.dc.b	$C4,$FE,$43,$E8,$EC,$E6,$F9,$CF,$26,$C6,$6F,$FC,$01,$87,$F9,$CE
	.dc.b	$B9,$56,$B0,$F0,$74,$BF,$52,$7C,$B6,$1B,$5D,$3E,$8B,$13,$23,$BF
	.dc.b	$6D,$CF,$B9,$EA,$7C,$FE,$2C,$9F,$2F,$4D,$AD,$E9,$FA,$CA,$C7,$0F
	.dc.b	$CA,$8F,$D7,$D6,$36,$F5,$6F,$FC,$B2,$B4,$D8,$CC,$EF,$0F,$97,$DD
	.dc.b	$F2,$3E,$8F,$E4,$B7,$3F,$8E,$D9,$F1,$E1,$FB,$56,$D7,$D0,$FA,$29
	.dc.b	$7F,$3B,$B2,$D7,$09,$B9,$F9,$26,$E7,$FA,$FF,$6F,$EF,$5F,$EE,$1B
	.dc.b	$D0,$F8,$0B,$13,$2F,$E2,$95,$FC,$02,$11,$88,$E4,$DA,$4E,$A7,$FA
	.dc.b	$5C,$62,$7F,$23,$1F,$C2,$3E,$03,$B3,$F6,$0C,$46,$A5,$ED,$F9,$8F
	.dc.b	$3F,$C0,$59,$D4,$77,$7F,$87,$36,$BF,$87,$35,$FE,$32,$D7,$F8,$CB
	.dc.b	$F1,$3E,$1D,$F1,$3F,$D9,$A7,$C9,$87,$8A,$FD,$35,$AF,$1F,$C3,$18
	.dc.b	$BE,$1B,$FE,$CB,$0E,$B3,$9A,$91,$EF,$BE,$4F,$6C,$7C,$58,$72,$DB
	.dc.b	$FC,$C5,$5A,$C7,$A3,$63,$DB,$8F,$49,$CD,$71,$C3,$FD,$C5,$7B,$7E
	.dc.b	$0D,$6E,$FA,$D6,$E5,$5B,$7D,$27,$9E,$35,$19,$35,$97,$C5,$9D,$65
	.dc.b	$D5,$F0,$D7,$59,$D6,$3E,$6A,$D5,$78,$2B,$59,$7C,$29,$CE,$1D,$A3
	.dc.b	$15,$D8,$97,$F3,$F0,$EB,$8F,$B3,$0E,$38,$F4,$CA,$C7,$82,$B7,$87
	.dc.b	$9A,$32,$FD,$39,$B8,$79,$E3,$07,$83,$0E,$CD,$3E,$69,$4F,$65,$77
	.dc.b	$C7,$D3,$5C,$C9,$5F,$6A,$B8,$65,$F3,$72,$F4,$46,$1F,$A2,$62,$B8
	.dc.b	$C7,$A4,$F0,$65,$EC,$7C,$27,$1C,$7F,$2E,$D3,$78,$F9,$31,$F5,$DF
	.dc.b	$49,$FC,$0B,$71,$F2,$5F,$28,$C6,$3E,$48,$C5,$F2,$25,$F9,$6B,$EB
	.dc.b	$61,$DE,$31,$3A,$7B,$B7,$0E,$F1,$F4,$2B,$BD,$FC,$BC,$BE,$BC,$FF
	.dc.b	$96,$AB,$F5,$6D,$C7,$F7,$4D,$AF,$ED,$DE,$9B,$E1,$5F,$13,$86,$D7
	.dc.b	$F9,$0E,$75,$39,$3F,$BB,$60,$F8,$9F,$F7,$1D,$7F,$92,$63,$1E,$B6
	.dc.b	$A7,$33,$E5,$6B,$FB,$C7,$8E,$9F,$BC,$6E,$B9,$AF,$E4,$FF,$13,$FC
	.dc.b	$B2,$FF,$88,$D6,$9F,$DE,$B5,$FD,$CB,$63,$84,$7F,$0F,$8D,$61,$E6
	.dc.b	$8C,$57,$6E,$3F,$F6,$AE,$1E,$19,$C9,$31,$FE,$39,$8F,$47,$D1,$7E
	.dc.b	$42,$4F,$F3,$B9,$D5,$73,$AD,$CF,$38,$FF,$7F,$4E,$A3,$D7,$5C,$AB
	.dc.b	$67,$82,$FC,$B3,$A3,$91,$63,$4F,$1B,$F0,$11,$0B,$F0,$3F,$22,$F8
	.dc.b	$3C,$E7,$CB,$7E,$08,$D1,$E0,$2D,$F8,$1C,$61,$F3,$63,$D3,$5C,$A7
	.dc.b	$A9,$C1,$F2,$3E,$3A,$ED,$5C,$67,$2D,$9F,$4E,$1E,$88,$C4,$73,$3D
	.dc.b	$06,$AB,$9E,$3F,$80,$26,$3F,$7E,$7C,$75,$C8,$E7,$7D,$6B,$73,$EF
	.dc.b	$8C,$57,$09,$FC,$75,$F6,$8C,$1D,$6B,$E1,$CF,$2C,$7E,$31,$F1,$23
	.dc.b	$CE,$7A,$4F,$84,$72,$3C,$15,$D5,$E7,$5D,$67,$94,$71,$C3,$D3,$5B
	.dc.b	$9C,$C6,$27,$A4,$70,$C3,$D7,$1D,$0F,$CC,$57,$DF,$8C,$CE,$E3,$07
	.dc.b	$5C,$7E,$CC,$FE,$28,$DB,$88,$DD,$79,$E3,$F0,$4F,$E9,$78,$FB,$70
	.dc.b	$F7,$E3,$93,$FC,$02,$32,$FF,$02,$8C,$BF,$AD,$8C,$BF,$EF,$8C,$3F
	.dc.b	$C4,$EF,$F8,$D5,$AE,$63,$F4,$2F,$F8,$D4,$FF,$41,$7C,$3B,$CD,$FD
	.dc.b	$6F,$6A,$31,$3F,$E7,$E6,$BF,$E6,$87,$FC,$3D,$A7,$F9,$9D,$CB,$6B
	.dc.b	$9C,$63,$3F,$F4,$13,$C1,$BE,$46,$BC,$79,$87,$78,$E2,$61,$F0,$46
	.dc.b	$1F,$27,$B7,$D8,$8F,$9F,$E2,$6B,$E7,$8B,$97,$9E,$6E,$1E,$49,$B5
	.dc.b	$F1,$27,$F8,$7F,$8E,$F8,$7F,$44,$FC,$8F,$77,$DA,$6A,$33,$3B,$C3
	.dc.b	$E7,$F6,$7D,$08,$C5,$FF,$9B,$1A,$FD,$36,$C7,$D6,$DF,$E9,$79,$9F
	.dc.b	$28,$95,$F4,$FE,$26,$BF,$4E,$2E,$BF,$4E,$7F,$8C,$7C,$4B,$D3,$FC
	.dc.b	$E7,$C3,$CC,$3F,$CE,$73,$96,$CE,$5F,$E7,$53,$CA,$24,$7F,$B5,$E7
	.dc.b	$83,$F8,$C3,$F9,$6B,$87,$0D,$7F,$03,$E1,$FE,$4B,$D1,$FF,$78,$72
	.dc.b	$FF,$56,$6F,$5E,$76,$DD,$FF,$38,$9D,$DE,$63,$F3,$35,$76,$FC,$C3
	.dc.b	$70,$FF,$4E,$4F,$1C,$3F,$A0,$EF,$F6,$F7,$D7,$5C,$AD,$09,$D3,$86
	.dc.b	$BE,$F6,$BF,$77,$89,$5F,$EF,$B9,$CC,$FE,$C3,$57,$A5,$E9,$84,$3D
	.dc.b	$3C,$36,$3C,$66,$A7,$26,$A6,$46,$A7,$AB,$67,$EA,$3E,$16,$BF,$59
	.dc.b	$AF,$53,$75,$F9,$FD,$3E,$57,$5C,$9F,$85,$97,$D7,$D3,$E5,$BF,$E0
	.dc.b	$E8,$F8,$55,$F0,$63,$11,$A8,$C5,$7D,$5F,$88,$93,$D2,$3E,$E6,$1F
	.dc.b	$AF,$77,$1E,$4B,$FE,$36,$FC,$F8,$F4,$FC,$3B,$E2,$3E,$67,$C3,$FF
	.dc.b	$16,$E4,$FE,$3C,$DD,$3F,$90,$D6,$F0,$F9,$AD,$7E,$3B,$5F,$CE,$35
	.dc.b	$E3,$7F,$22,$BF,$C7,$B7,$DB,$C3,$F0,$9F,$95,$7F,$95,$7C,$46,$2F
	.dc.b	$ED,$D7,$0C,$7F,$99,$B6,$15,$C1,$FC,$0F,$C3,$BE,$23,$F1,$0B,$E4
	.dc.b	$29,$88,$93,$F7,$A7,$83,$63,$EF,$C6,$23,$FC,$BD,$DF,$E0,$5F,$0F
	.dc.b	$F8,$A5,$F1,$0F,$0F,$FA,$4C,$E6,$31,$3F,$E2,$C8,$E1,$E6,$2F,$A4
	.dc.b	$61,$ED,$F1,$1C,$70,$FE,$A6,$DB,$FF,$1A,$BB,$D2,$F4,$7F,$58,$C7
	.dc.b	$A4,$77,$8F,$EB,$79,$78,$AF,$D3,$18,$C3,$CE,$F1,$C7,$C6,$DB,$E5
	.dc.b	$5F,$E4,$4B,$E5,$97,$8E,$39,$9F,$E4,$8C,$7C,$2E,$03,$88,$E0,$4B
	.dc.b	$ED,$3B,$9E,$39,$75,$24,$E9,$F8,$13,$D6,$FC,$73,$EB,$8F,$39,$E0
	.dc.b	$C7,$D6,$F9,$5A,$FD,$8C,$BE,$0B,$ED,$6B,$C7,$2E,$78,$F9,$2F,$B4
	.dc.b	$B1,$F2,$BD,$6B,$E9,$B4,$F3,$5F,$4C,$7C,$F1,$9C,$BD,$67,$C3,$7C
	.dc.b	$47,$83,$2F,$3E,$5F,$27,$2F,$0D,$6E,$BC,$C7,$C9,$AF,$6D,$F6,$8F
	.dc.b	$71,$CF,$18,$F9,$0F,$8F,$1B,$C3,$B4,$79,$70,$EE,$47,$04,$3A,$9C
	.dc.b	$CF,$15,$76,$C3,$CD,$7D,$70,$F5,$9E,$6C,$7E,$21,$F4,$31,$F0,$E1
	.dc.b	$EC,$E1,$F9,$CF,$CB,$6B,$9C,$FF,$13,$18,$8F,$F4,$CF,$87,$FE,$9A
	.dc.b	$8C,$42,$31,$3F,$D0,$1B,$1F,$DD,$5A,$FF,$90,$9F,$C5,$F3,$BE,$01
	.dc.b	$C7,$F8,$EA,$EE,$7E,$7E,$75,$19,$8F,$01,$99,$E2,$FE,$94,$E3,$AF
	.dc.b	$F0,$BA,$E3,$19,$38,$C6,$67,$8B,$FC,$46,$38,$EF,$FE,$29,$9C,$BE
	.dc.b	$D8,$F0,$5F,$8F,$0F,$5D,$6A,$7E,$01,$A3,$C1,$1A,$9E,$11,$9B,$F0
	.dc.b	$3E,$4B,$F0,$4E,$A7,$CC,$D8,$CC,$F4,$39,$3E,$59,$DC,$65,$A7,$6A
	.dc.b	$F1,$47,$1C,$76,$65,$FD,$95,$76,$8C,$57,$28,$EC,$78,$9C,$B4,$CC
	.dc.b	$FF,$88,$30,$F8,$8F,$E0,$4E,$A6,$70,$FD,$EB,$9B,$F1,$D7,$28,$FF
	.dc.b	$12,$57,$27,$EA,$9C,$4F,$74,$78,$5F,$A3,$7E,$76,$C6,$5F,$14,$65
	.dc.b	$F0,$47,$BC,$FB,$D5,$E2,$8E,$27,$79,$F7,$B6,$7F,$D3,$BE,$83,$D9
	.dc.b	$FF,$4F,$4E,$1A,$84,$6F,$1F,$F1,$2B,$63,$FC,$52,$DE,$1F,$F5,$42
	.dc.b	$FA,$2F,$CF,$C6,$2E,$DF,$8A,$6D,$FF,$92,$FE,$1F,$FD,$5D,$3F,$E4
	.dc.b	$B6,$C7,$F8,$02,$3C,$9F,$0E,$F8,$87,$C4,$72,$E1,$61,$C3,$3F,$D7
	.dc.b	$C6,$3D,$0D,$4E,$71,$E7,$87,$27,$C1,$3E,$4C,$9F,$0E,$F8,$8F,$8B
	.dc.b	$18,$6C,$66,$BF,$C9,$2D,$9F,$F2,$6B,$72,$F8,$7F,$3F,$FE,$17,$AF
	.dc.b	$F2,$CC,$E5,$FF,$3D,$46,$6B,$E0,$1F,$E6,$FF,$81,$7C,$17,$D6,$D5
	.dc.b	$18,$BF,$F2,$AB,$6B,$FC,$E4,$D7,$FD,$5C,$E7,$1F,$E2,$7E,$6B,$E0
	.dc.b	$7F,$01,$7F,$E7,$BA,$D3,$F8,$39,$D5,$FC,$F9,$7C,$17,$8D,$B5,$E2
	.dc.b	$8C,$11,$FF,$44,$D7,$7E,$9E,$53,$F8,$28,$C4,$FF,$9F,$3D,$FF,$C0
	.dc.b	$E1,$FE,$A9,$6B,$FE,$F6,$6B,$F9,$76,$A7,$EF,$39,$F5,$7E,$E4,$62
	.dc.b	$35,$18,$E1,$CB,$6B,$C4,$FB,$FE,$03,$DC,$48,$F9,$ED,$AF,$AF,$87
	.dc.b	$CB,$F8,$8E,$15,$F0,$35,$47,$BD,$FF,$4E,$46,$7A,$F1,$E8,$65,$B1
	.dc.b	$93,$94,$65,$F8,$DE,$CF,$E5,$35,$34,$91,$FE,$9C,$8C,$BF,$D9,$B4
	.dc.b	$F9,$5E,$97,$F5,$4F,$05,$1F,$36,$F0,$EA,$B8,$6B,$F3,$E7,$2F,$D1
	.dc.b	$BA,$FD,$18,$B8,$7F,$B0,$2B,$8E,$BF,$8F,$F8,$7E,$2D,$8C,$EA,$F8
	.dc.b	$8C,$AA,$A4,$9D,$CD,$C7,$FB,$F6,$7F,$ED,$AF,$88,$7C,$3F,$F7,$0B
	.dc.b	$FE,$39,$5F,$ED,$99,$CC,$7E,$82,$38,$4E,$DB,$19,$78,$47,$93,$A7
	.dc.b	$E0,$F4,$AF,$9D,$78,$86,$9F,$AD,$AF,$5F,$85,$31,$F7,$F4,$DB,$E1
	.dc.b	$7F,$B3,$AC,$71,$7D,$49,$E7,$BF,$F0,$0C,$3E,$3D,$6B,$5F,$DB,$4F
	.dc.b	$F3,$5B,$E1,$87,$EF,$71,$D6,$3C,$BA,$3E,$6F,$C3,$FC,$E6,$E7,$F4
	.dc.b	$9B,$3E,$4C,$7F,$7A,$DA,$FA,$BF,$0F,$2F,$EC,$F4,$DA,$E1,$37,$3F
	.dc.b	$1E,$0A,$C6,$9F,$BB,$F5,$FE,$86,$1F,$BB,$6B,$FB,$B6,$F2,$9A,$7E
	.dc.b	$3B,$D9,$FC,$7B,$73,$FC,$BD,$71,$6D,$2B,$71,$87,$F2,$FE,$B4,$8D
	.dc.b	$FC,$15,$BF,$B8,$7F,$0D,$9F,$CF,$FC,$1B,$E1,$FF,$B3,$CE,$B1,$56
	.dc.b	$A7,$3B,$7D,$BA,$E5,$8F,$9B,$0E,$5E,$5F,$9B,$0E,$0D,$9D,$61,$CB
	.dc.b	$5F,$E4,$55,$AC,$3C,$B8,$72,$6D,$6D,$FE,$4D,$3B,$C7,$C5,$5B,$6D
	.dc.b	$6E,$BC,$A7,$2C,$7C,$24,$7D,$D7,$CF,$0F,$D6,$E1,$CD,$B7,$CF,$2F
	.dc.b	$87,$97,$1C,$BB,$E5,$6F,$D7,$37,$2F,$F0,$96,$3C,$DB,$8F,$3C,$BC
	.dc.b	$11,$8C,$38,$3E,$8C,$7A,$BE,$92,$3F,$47,$2F,$1C,$63,$2F,$3D,$72
	.dc.b	$9C,$D7,$22,$4E,$AB,$E0,$E5,$E0,$CB,$CD,$06,$DF,$2A,$E1,$7C,$AB
	.dc.b	$D0,$72,$3C,$46,$1D,$3E,$B8,$E2,$47,$4F,$C8,$78,$9E,$1C,$BE,$D5
	.dc.b	$F8,$B2,$F8,$F3,$9C,$BD,$D1,$89,$F4,$9B,$CB,$D9,$7C,$8F,$A0,$E2
	.dc.b	$B8,$B6,$FC,$39,$7B,$DF,$29,$0F,$01,$EF,$CB,$DB,$87,$83,$0E,$B3
	.dc.b	$EA,$8C,$61,$FA,$18,$C4,$F8,$E3,$15,$CA,$31,$3D,$72,$F8,$CF,$82
	.dc.b	$7E,$5E,$1E,$58,$C4,$EB,$0F,$81,$1D,$B0,$F6,$B4,$FE,$92,$7D,$82
	.dc.b	$C7,$EB,$DB,$87,$F0,$27,$7E,$7B,$E1,$DF,$0F,$CA,$7F,$CD,$D3,$98
	.dc.b	$C4,$23,$0E,$DA,$F8,$3E,$1D,$E9,$70,$C5,$18,$9E,$D1,$89,$D4,$E6
	.dc.b	$79,$47,$F1,$18,$E1,$AF,$F0,$46,$E9,$FD,$A3,$E1,$FF,$98,$5E,$EE
	.dc.b	$3F,$C4,$2B,$6F,$FA,$26,$76,$E6,$77,$7D,$5F,$E6,$B3,$A8,$C4,$EC
	.dc.b	$8F,$F2,$78,$D5,$FF,$0E,$BE,$F5,$FA,$F6,$E1,$EF,$8F,$E6,$F5,$C1
	.dc.b	$FE,$17,$3A,$26,$1E,$2E,$3F,$F8,$01,$87,$58,$F3,$B6,$B9,$CE,$9E
	.dc.b	$73,$A6,$9A,$30,$79,$5B,$7E,$38,$C1,$EE,$AE,$D8,$FB,$EF,$B4,$EA
	.dc.b	$3E,$29,$27,$EA,$4E,$B1,$F8,$13,$9A,$F1,$BC,$DF,$1C,$6A,$BB,$9C
	.dc.b	$A3,$38,$FC,$47,$D4,$79,$4E,$71,$C1,$FB,$D7,$E1,$9E,$17,$E1,$96
	.dc.b	$1E,$FC,$3C,$11,$C8,$86,$AF,$96,$3E,$23,$EA,$45,$3C,$46,$CC,$5F
	.dc.b	$7B,$ED,$87,$8B,$1F,$35,$6C,$E3,$8F,$89,$F9,$91,$F0,$5C,$CF,$17
	.dc.b	$DA,$76,$8E,$75,$DB,$0C,$4E,$8E,$2F,$DE,$C7,$D7,$18,$3E,$5E,$3F
	.dc.b	$17,$1F,$11,$E3,$BE,$93,$B9,$FB,$0F,$0C,$3D,$75,$BB,$FC,$2D,$F0
	.dc.b	$BF,$85,$18,$25,$FC,$A3,$F4,$7C,$7C,$38,$E4,$7E,$C9,$E7,$C7,$23
	.dc.b	$04,$3F,$56,$E5,$FD,$6C,$72,$72,$7F,$9F,$BE,$05,$E4,$7E,$9F,$B3
	.dc.b	$FD,$9D,$E9,$5E,$93,$9A,$DD,$F1,$BD,$4C,$8C,$1D,$7C,$5E,$04,$7C
	.dc.b	$1F,$0F,$5F,$F3,$5D,$FF,$A9,$1E,$06,$8F,$37,$AD,$E3,$F7,$64,$60
	.dc.b	$87,$F9,$85,$CB,$D2,$3D,$06,$38,$3F,$CB,$B3,$89,$F8,$5E,$C7,$B4
	.dc.b	$8F,$FA,$AB,$C6,$7C,$07,$F9,$0E,$FF,$CE,$F8,$AF,$77,$EA,$7F,$88
	.dc.b	$5F,$0E,$6F,$91,$E5,$FC,$BF,$4E,$5F,$D1,$8C,$BF,$EA,$C8,$CB,$F1
	.dc.b	$23,$F1,$3F,$0B,$F8,$AF,$82,$93,$FC,$D2,$32,$D2,$1F,$E6,$2F,$83
	.dc.b	$7C,$2E,$A3,$F9,$7C,$DD,$7F,$CB,$D5,$A9,$FF,$5E,$FD,$17,$FA,$CB
	.dc.b	$D6,$78,$FF,$E9,$88,$C3,$64,$7F,$BB,$38,$3F,$A5,$D7,$08,$FF,$51
	.dc.b	$D7,$06,$9F,$EE,$F7,$1D,$2A,$E3,$7D,$23,$1D,$9A,$EA,$7D,$0D,$38
	.dc.b	$DF,$FA,$1F,$1F,$1E,$5E,$1C,$38,$C7,$FB,$F6,$B8,$3A,$C7,$B5,$F1
	.dc.b	$85,$FA,$57,$E6,$5F,$ED,$D7,$81,$FE,$C8,$FD,$A3,$53,$DD,$B7,$FE
	.dc.b	$DA,$AE,$15,$D2,$31,$EA,$FC,$38,$EA,$D9,$DD,$F8,$76,$7A,$D9,$A9
	.dc.b	$7E,$23,$89,$67,$6F,$9B,$4F,$99,$E0,$FC,$EC,$FF,$DD,$F3,$96,$BF
	.dc.b	$66,$3E,$59,$BF,$56,$47,$3D,$7F,$D4,$ED,$E5,$E9,$F3,$F2,$30,$43
	.dc.b	$E0,$1F,$1D,$D3,$F8,$78,$FC,$26,$AF,$CF,$D1,$86,$CF,$D0,$9D,$4F
	.dc.b	$C9,$F8,$8F,$E2,$73,$DA,$72,$EE,$B8,$B9,$6C,$7D,$D9,$E2,$FD,$FF
	.dc.b	$8A,$7C,$57,$F8,$FA,$D1,$87,$FC,$82,$D7,$FC,$8C,$D7,$FC,$A2,$E6
	.dc.b	$7A,$4F,$DF,$8C,$71,$FE,$A2,$97,$FE,$3F,$BF,$C8,$7C,$43,$E7,$BF
	.dc.b	$0C,$DA,$FC,$03,$5F,$9A,$17,$C5,$F0,$99,$3C,$1E,$71,$F9,$1F,$43
	.dc.b	$F8,$D5,$FE,$EB,$E2,$1F,$01,$E5,$F8,$A7,$93,$E7,$9C,$E5,$FC,$02
	.dc.b	$72,$F1,$AF,$66,$1D,$21,$5C,$1F,$F3,$47,$8F,$FA,$E7,$35,$DF,$0E
	.dc.b	$CD,$8E,$87,$F9,$CA,$78,$57,$3B,$EC,$F5,$C3,$C7,$5D,$0F,$36,$1D
	.dc.b	$1E,$58,$F4,$7C,$6D,$9E,$35,$CE,$B8,$56,$8E,$85,$8D,$3C,$63,$77
	.dc.b	$E4,$C7,$E1,$E3,$E8,$7D,$95,$CB,$0E,$F1,$AC,$7C,$97,$C6,$BD,$27
	.dc.b	$53,$37,$CD,$F5,$9C,$F0,$F3,$E1,$D2,$33,$8F,$C7,$AE,$10,$9D,$36
	.dc.b	$BD,$33,$AB,$F1,$4E,$4F,$11,$F0,$9C,$9A,$AE,$67,$47,$33,$E0,$8C
	.dc.b	$D7,$6C,$3D,$67,$4C,$3E,$5E,$1D,$6F,$C4,$71,$AF,$69,$C6,$35,$5C
	.dc.b	$71,$F9,$B7,$EF,$8E,$B3,$F1,$4D,$1F,$12,$7E,$83,$92,$4F,$17,$A9
	.dc.b	$88,$EA,$70,$8D,$7A,$4F,$8A,$E6,$DA,$FA,$AD,$7F,$14,$D7,$FB,$4F
	.dc.b	$87,$FD,$AA,$31,$08,$C4,$FF,$B4,$DB,$1F,$ED,$E6,$BF,$C2,$DF,$CE
	.dc.b	$79,$2F,$84,$FE,$E5,$87,$EC,$6B,$B9,$F9,$A9,$D7,$1F,$E8,$E3,$33
	.dc.b	$FB,$79,$CC,$7E,$80,$CE,$5F,$C6,$63,$33,$D6,$7D,$07,$5B,$F0,$D7
	.dc.b	$A6,$31,$7D,$DD,$D7,$7A,$DC,$7B,$4D,$9E,$13,$75,$E1,$8D,$C7,$C1
	.dc.b	$37,$3E,$3C,$3E,$1C,$EA,$7D,$27,$F8,$63,$0F,$4D,$F9,$6F,$C5,$87
	.dc.b	$97,$0E,$2D,$BE,$B5,$B3,$C5,$19,$AE,$33,$E7,$AE,$2D,$38,$9E,$D8
	.dc.b	$D9,$E7,$72,$7A,$DC,$D7,$78,$F4,$61,$EB,$7E,$19,$8B,$F1,$4F,$92
	.dc.b	$3C,$15,$C9,$F0,$46,$63,$C0,$66,$3C,$11,$93,$C0,$B0,$F8,$EF,$23
	.dc.b	$C0,$76,$8E,$0E,$4E,$17,$E7,$C3,$E1,$DF,$82,$B7,$5E,$03,$9C,$7F
	.dc.b	$47,$ED,$75,$FF,$92,$F5,$FF,$1A,$46,$33,$FF,$11,$7B,$3F,$E9,$67
	.dc.b	$FC,$49,$19,$C3,$FB,$2B,$6B,$8D,$F3,$AF,$EE,$57,$C6,$F7,$39,$7F
	.dc.b	$CB,$BF,$0C,$F2,$F1,$08,$C5,$FF,$91,$36,$7C,$47,$FA,$B2,$3F,$C6
	.dc.b	$D3,$AA,$FE,$EE,$D7,$FC,$04,$73,$BF,$F2,$83,$FE,$05,$8D,$36,$32
	.dc.b	$FD,$3E,$57,$81,$E7,$C5,$18,$C3,$FC,$51,$F0,$EF,$8B,$D4,$E7,$0F
	.dc.b	$F1,$AD,$F0,$C3,$27,$C1,$D9,$87,$1C,$7D,$D8,$71,$9F,$F3,$77,$93
	.dc.b	$FE,$82,$9F,$F2,$67,$BB,$F2,$78,$53,$9C,$FF,$C8,$70,$F1,$55,$BA
	.dc.b	$D6,$1F,$E6,$B8,$C5,$F3,$AD,$E1,$FE,$5D,$BF,$E2,$B7,$FE,$71,$F6
	.dc.b	$F9,$73,$ED,$B5,$FE,$5F,$6E,$1E,$2C,$11,$F7,$3F,$17,$7A,$54,$FB
	.dc.b	$FE,$FE,$30,$FF,$38,$4D,$C3,$FC,$62,$DC,$BF,$D1,$0D,$9F,$F1,$F5
	.dc.b	$5A,$FF,$53,$CD,$4E,$75,$F8,$9F,$43,$E1,$BF,$F3,$B3,$F1,$7E,$2B
	.dc.b	$E1,$78,$76,$7F,$CF,$4D,$C3,$D8,$DC,$BF,$D2,$8F,$C9,$FA,$09,$5F
	.dc.b	$27,$B6,$DF,$C3,$9B,$7F,$72,$6D,$7A,$BD,$6F,$F5,$E7,$7F,$C6,$C3
	.dc.b	$FD,$83,$B2,$73,$BF,$C0,$F6,$AD,$6A,$6A,$AE,$18,$7F,$62,$BE,$CD
	.dc.b	$9E,$B7,$C2,$B7,$1F,$D9,$6B,$6D,$3C,$7F,$21,$BC,$38,$3E,$2E,$0F
	.dc.b	$C9,$FA,$37,$91,$F3,$F9,$E7,$4E,$63,$0D,$D7,$C9,$7C,$9B,$7C,$9C
	.dc.b	$DB,$E2,$FA,$1F,$0B,$B5,$F1,$9F,$C5,$78,$7E,$56,$3F,$A6,$AD,$E1
	.dc.b	$C9,$E1,$5B,$DF,$FD,$C9,$8F,$5A,$EF,$F1,$9E,$BF,$9F,$F5,$36,$32
	.dc.b	$79,$A3,$2D,$33,$C3,$F2,$72,$FD,$A4,$EA,$31,$AA,$3E,$37,$C5,$CB
	.dc.b	$F7,$FC,$65,$BE,$59,$2B,$1B,$7B,$A1,$58,$D3,$D7,$F1,$9F,$4F,$D4
	.dc.b	$FA,$4D,$CB,$E5,$B7,$2E,$AD,$C7,$F2,$51,$F8,$BF,$87,$FC,$37,$BD
	.dc.b	$F8,$76,$DF,$E2,$E7,$4D,$E8,$56,$E3,$0F,$09,$FC,$61,$F9,$77,$4E
	.dc.b	$27,$8C,$F3,$F1,$BF,$2C,$7E,$5F,$E1,$EC,$64,$F1,$7C,$4F,$F2,$0F
	.dc.b	$3B,$18,$76,$8C,$3D,$6F,$F4,$F1,$8C,$FD,$F8,$FF,$39,$7E,$0C,$67
	.dc.b	$6F,$F6,$63,$6C,$1C,$3F,$B1,$BE,$BE,$3B,$1F,$2B,$67,$C8,$DC,$3C
	.dc.b	$35,$C2,$B6,$4A,$F4,$65,$6B,$CB,$7C,$71,$F2,$B6,$BF,$84,$E1,$CD
	.dc.b	$EF,$87,$1A,$FE,$53,$39,$BE,$73,$9A,$E3,$85,$7D,$D8,$75,$6C,$FC
	.dc.b	$EA,$E3,$1D,$88,$FB,$2B,$96,$3F,$18,$F8,$F9,$7C,$FC,$BD,$38,$6B
	.dc.b	$1E,$D3,$9C,$7E,$7D,$F1,$9F,$13,$CB,$2F,$56,$1C,$AB,$C2,$FB,$B1
	.dc.b	$E9,$8F,$AF,$1E,$95,$EA,$36,$76,$C7,$C3,$87,$48,$C5,$71,$8C,$47
	.dc.b	$1C,$BE,$09,$EF,$79,$99,$C7,$C9,$19,$36,$E0,$91,$F4,$0E,$39,$7B
	.dc.b	$EF,$E7,$4E,$A7,$A1,$E4,$3F,$0A,$70,$C3,$E2,$9E,$AC,$BE,$0C,$79
	.dc.b	$32,$F8,$38,$75,$8C,$65,$EC,$8C,$65,$E7,$C7,$C1,$18,$7E,$65,$77
	.dc.b	$CB,$E2,$63,$DA,$FD,$B3,$EB,$6B,$E8,$26,$3E,$2C,$BC,$F7,$CA,$7A
	.dc.b	$1F,$E3,$A3,$EB,$1F,$E3,$CC,$7F,$76,$D9,$FD,$63,$FE,$40,$F2,$1F
	.dc.b	$16,$F8,$BE,$53,$FE,$3A,$8C,$3D,$67,$73,$A6,$C7,$86,$FF,$C3,$7E
	.dc.b	$9F,$26,$C6,$7C,$A4,$E5,$B1,$98,$FE,$2B,$18,$D3,$F8,$6B,$76,$FE
	.dc.b	$17,$4F,$8B,$71,$FF,$B4,$67,$F8,$8D,$6E,$BF,$46,$D7,$F8,$93,$6B
	.dc.b	$8C,$78,$30,$8F,$F2,$39,$D5,$F5,$8E,$73,$AC,$3C,$AF,$F3,$DB,$E7
	.dc.b	$3C,$63,$58,$79,$1D,$5F,$39,$F1,$C6,$9E,$A7,$F4,$0A,$E7,$7D,$A3
	.dc.b	$A1,$1F,$EA,$31,$AB,$ED,$39,$6C,$EA,$FC,$18,$FD,$48,$E8,$49,$F9
	.dc.b	$B7,$CE,$FC,$73,$96,$CE,$AF,$9E,$3F,$37,$0E,$93,$E0,$35,$5E,$D3
	.dc.b	$DE,$79,$4F,$E8,$B8,$77,$73,$87,$76,$CF,$27,$81,$83,$C6,$72,$C3
	.dc.b	$C7,$8F,$7C,$3C,$73,$93,$18,$F9,$CF,$BA,$E2,$7C,$D8,$F8,$EB,$8C
	.dc.b	$F7,$24,$66,$B9,$57,$0B,$F0,$D7,$9A,$F8,$DE,$EB,$B4,$FD,$DA,$DE
	.dc.b	$3E,$FB,$DC,$63,$1F,$95,$5E,$E3,$F4,$11,$97,$13,$C1,$EB,$5F,$27
	.dc.b	$0F,$04,$78,$F1,$F7,$B7,$1F,$43,$C3,$1E,$EF,$63,$07,$86,$BE,$B1
	.dc.b	$D2,$FB,$63,$F9,$B8,$C5,$72,$77,$1F,$B8,$E0,$FF,$67,$C7,$E9,$9B
	.dc.b	$19,$BF,$BD,$1A,$E2,$E8,$48,$FD,$77,$17,$73,$0F,$E9,$E3,$0E,$8F
	.dc.b	$DA,$3A,$C7,$F4,$F3,$FB,$BE,$B7,$BE,$F8,$BC,$75,$3E,$2E,$9B,$9D
	.dc.b	$65,$FE,$72,$6B,$FE,$50,$8D,$B9,$AE,$1E,$3C,$8E,$0F,$3F,$8B,$C4
	.dc.b	$7F,$98,$6D,$38,$CF,$FC,$EB,$0A,$F1,$7B,$3E,$6F,$8B,$98,$78,$E3
	.dc.b	$2F,$FA,$0A,$32,$F5,$8C,$BF,$E6,$68,$CD,$F6,$9F,$77,$B7,$2B,$DC
	.dc.b	$FF,$21,$F1,$DF,$17,$FE,$60,$BF,$2C,$EE,$BD,$4E,$E7,$5C,$BF,$E4
	.dc.b	$2E,$0F,$9F,$E6,$7C,$EF,$5E,$61,$F2,$E3,$E7,$18,$7F,$CF,$71,$87
	.dc.b	$F9,$77,$C5,$7E,$3F,$E1,$24,$60,$83,$9F,$F1,$F3,$F8,$D8,$FE,$1E
	.dc.b	$D7,$FA,$53,$7A,$4F,$E6,$C9,$1F,$E7,$9F,$86,$7C,$5F,$E1,$BE,$2F
	.dc.b	$FA,$9F,$A9,$DE,$9A,$BA,$95,$AC,$FF,$D5,$2D,$AD,$F3,$FF,$48,$C3
	.dc.b	$96,$1C,$6B,$4D,$AF,$F3,$5C,$EB,$0E,$B1,$8C,$3E,$76,$1C,$DF,$F5
	.dc.b	$BF,$C5,$F2,$A3,$4E,$B3,$FE,$92,$AE,$47,$93,$C8,$D3,$8C,$7D,$95
	.dc.b	$C2,$7D,$52,$F8,$BF,$F6,$CB,$BF,$3F,$E2,$2C,$FD,$B1,$87,$33,$C2
	.dc.b	$75,$9F,$F7,$E8,$C4,$7F,$80,$0E,$DE,$9F,$F8,$59,$A7,$F8,$62,$36
	.dc.b	$F1,$8F,$2C,$75,$AE,$F1,$E8,$DD,$F1,$7F,$3B,$17,$C1,$52,$63,$F0
	.dc.b	$76,$78,$5F,$20,$CC,$7A,$0F,$C0,$D7,$27,$84,$75,$F8,$B9,$1F,$15
	.dc.b	$FA,$3C,$D2,$30,$43,$A7,$0F,$F1,$68,$FC,$46,$AF,$8B,$47,$F8,$E2
	.dc.b	$30,$D8,$CD,$7F,$91,$67,$2D,$8C,$BE,$98,$CB,$63,$31,$F6,$BE,$2B
	.dc.b	$F8,$FB,$F7,$3E,$2D,$E9,$FE,$0A,$73,$C7,$FE,$36,$6E,$1F,$EC,$2A
	.dc.b	$DB,$F8,$E0,$E1,$FF,$2D,$35,$FC,$03,$5F,$F2,$0B,$5F,$F6,$6F,$C5
	.dc.b	$BD,$FF,$C2,$36,$3F,$CB,$6D,$7F,$CC,$CD,$7F,$18,$E7,$1F,$CB,$B7
	.dc.b	$0F,$F2,$7B,$67,$BF,$C5,$FE,$83,$D0,$78,$7F,$9D,$97,$C5,$3E,$03
	.dc.b	$FC,$AF,$F1,$6F,$27,$FC,$B3,$39,$AF,$F7,$2C,$E5,$FF,$73,$C6,$56
	.dc.b	$1D,$B4,$CE,$1F,$E6,$28,$C4,$BE,$2F,$86,$1C,$E3,$15,$E2,$C7,$FC
	.dc.b	$D3,$5D,$6F,$8B,$72,$F5,$3C,$A7,$75,$C6,$FF,$DF,$73,$BC,$6C,$F8
	.dc.b	$EF,$8B,$BA,$F0,$9E,$EB,$E6,$8F,$6C,$F0,$71,$87,$B3,$2F,$14,$7C
	.dc.b	$03,$91,$4E,$38,$59,$F5,$E1,$E1,$C7,$8D,$78,$8D,$96,$76,$F4,$35
	.dc.b	$97,$C9,$9F,$74,$6F,$1F,$BD,$8F,$42,$61,$F0,$71,$F1,$4F,$AE,$98
	.dc.b	$73,$AF,$21,$B8,$E9,$5F,$29,$F1,$61,$E8,$3C,$27,$8B,$0E,$F1,$E7
	.dc.b	$71,$5C,$E7,$83,$DA,$7E,$5E,$1E,$3C,$7D,$F1,$E4,$CB,$E9,$B4,$F1
	.dc.b	$1F,$1A,$FC,$37,$CE,$34,$7A,$09,$87,$91,$CB,$E4,$7C,$38,$F4,$CB
	.dc.b	$E1,$47,$BA,$BD,$07,$C3,$AE,$53,$F9,$97,$94,$FD,$87,$38,$7E,$AA
	.dc.b	$B8,$1F,$2F,$D2,$FD,$8D,$78,$0F,$F6,$7D,$64,$8F,$F4,$BF,$2B,$F2
	.dc.b	$53,$C2,$3F,$D9,$B1,$89,$FF,$76,$36,$38,$1D,$0B,$1F,$B7,$F2,$DF
	.dc.b	$0F,$FE,$03,$8F,$DD,$4F,$E7,$EB,$F3,$15,$AE,$1F,$CD,$4E,$4F,$E1
	.dc.b	$38,$FF,$B7,$A7,$51,$83,$BC,$67,$0E,$CF,$9A,$3F,$3D,$3C,$35,$F0
	.dc.b	$C3,$0F,$BB,$39,$AE,$18,$FA,$70,$E7,$7D,$8F,$DB,$57,$8A,$79,$DF
	.dc.b	$2A,$FF,$10,$61,$8B,$F0,$C6,$30,$F2,$4F,$5C,$7F,$1B,$5E,$F8,$C3
	.dc.b	$FB,$53,$13,$ED,$9D,$C7,$A4,$EA,$E2,$FE,$5D,$FA,$27,$EC,$9F,$1C
	.dc.b	$63,$E9,$BF,$19,$D6,$3C,$07,$5C,$3C,$67,$AC,$B1,$C2,$3E,$1D,$7B
	.dc.b	$5F,$8C,$76,$7C,$4F,$DB,$C7,$DA,$73,$9F,$05,$59,$E7,$8F,$AD,$F6
	.dc.b	$9E,$4C,$7E,$74,$66,$F9,$DF,$D8,$3D,$4F,$3A,$F1,$4F,$3B,$F4,$C7
	.dc.b	$E6,$1A,$FC,$66,$BF,$40,$F1,$9E,$13,$C6,$FD,$D9,$F0,$C7,$C0,$C7
	.dc.b	$E6,$3F,$E7,$DE,$DF,$F3,$16,$DF,$E4,$0E,$9F,$F1,$5E,$3F,$DF,$FD
	.dc.b	$9F,$F5,$63,$FE,$01,$8C,$DF,$F8,$A4,$FF,$34,$DF,$F9,$21,$FE,$8C
	.dc.b	$62,$7F,$C8,$4F,$76,$B9,$33,$F1,$4F,$2F,$94,$7F,$4F,$8C,$4F,$06
	.dc.b	$C7,$09,$DA,$8C,$E3,$EC,$F8,$97,$C5,$7B,$B0,$F2,$D6,$9F,$F3,$AC
	.dc.b	$E9,$FF,$35,$E9,$FD,$EB,$C0,$E1,$8F,$E5,$39,$BF,$C9,$75,$FE,$70
	.dc.b	$C9,$87,$1A,$FF,$26,$61,$C6,$B9,$DF,$09,$FF,$32,$4E,$5B,$19,$7F
	.dc.b	$85,$35,$FF,$59,$79,$3F,$DC,$6B,$FC,$ED,$EE,$79,$37,$FF,$4C,$55
	.dc.b	$F0,$D3,$49,$5F,$E8,$3B,$E5,$5F,$E5,$6A,$DC,$E7,$9B,$97,$B3,$D6
	.dc.b	$FF,$8D,$EB,$FE,$82,$8C,$D7,$FA,$1D,$D5,$63,$0F,$C2,$7C,$53,$D2
	.dc.b	$F9,$30,$8C,$5F,$F9,$2F,$DE,$B3,$C2,$B7,$58,$8F,$F4,$53,$5F,$C6
	.dc.b	$B5,$FF,$0A,$35,$ED,$1F,$0F,$E2,$B9,$B6,$FE,$17,$C5,$7C,$7F,$0F
	.dc.b	$DC,$D8,$FF,$51,$45,$BF,$F3,$DE,$AA,$C6,$BF,$EA,$4F,$8A,$FA,$07
	.dc.b	$C7,$ED,$B8,$FF,$AA,$E2,$E1,$CA,$6D,$7F,$AB,$B5,$7A,$DF,$ED,$3E
	.dc.b	$FF,$91,$7F,$E2,$F9,$CE,$3F,$D4,$F5,$FD,$6C,$3D,$AE,$11,$8C,$2D
	.dc.b	$F1,$7F,$DB,$D7,$FE,$B3,$BE,$73,$FB,$76,$C6,$67,$73,$CE,$3C,$0F
	.dc.b	$8F,$E2,$BA,$AA,$E7,$8F,$A2,$31,$3E,$2E,$6F,$F7,$67,$A3,$79,$9F
	.dc.b	$15,$9E,$24,$E6,$30,$DC,$FC,$37,$C9,$B7,$C9,$CD,$BE,$2B,$A1,$F0
	.dc.b	$BA,$DF,$15,$FC,$63,$5F,$9B,$8F,$FB,$FE,$B7,$87,$27,$85,$6F,$83
	.dc.b	$F6,$31,$E7,$F8,$AF,$57,$C5,$5B,$F1,$B7,$83,$FB,$C5,$FF,$99,$F1
	.dc.b	$6A,$AD,$67,$E2,$8C,$61,$EA,$C7,$AB,$4E,$A7,$BB,$E2,$BE,$41,$F2
	.dc.b	$7E,$2A,$C6,$49,$9F,$88,$64,$DD,$5C,$27,$53,$9F,$8A,$FA,$DE,$9F
	.dc.b	$D5,$6E,$1F,$39,$B5,$C3,$04,$63,$4F,$C1,$72,$BE,$1F,$F0,$EB,$DC
	.dc.b	$B1,$86,$E3,$F8,$EC,$7F,$19,$18,$7D,$46,$23,$55,$CD,$B3,$BB,$F0
	.dc.b	$BE,$0F,$1B,$F3,$77,$FA,$B6,$BF,$99,$6D,$FF,$35,$BF,$D3,$47,$E7
	.dc.b	$BE,$23,$F9,$5F,$99,$E0,$8C,$78,$C8,$C5,$7A,$6B,$78,$7A,$30,$FE
	.dc.b	$81,$97,$E9,$67,$2D,$C3,$95,$7B,$A7,$3C,$BF,$BE,$CB,$BE,$5D,$6F
	.dc.b	$78,$71,$8F,$0E,$1C,$67,$37,$FD,$CB,$0E,$73,$FE,$B9,$6B,$F0,$72
	.dc.b	$E9,$72,$39,$3D,$30,$ED,$18,$23,$F7,$B1,$F2,$C6,$30,$F0,$63,$E2
	.dc.b	$8F,$A1,$87,$5B,$E6,$4A,$F7,$46,$5E,$38,$5A,$E2,$75,$3D,$57,$D5
	.dc.b	$63,$EC,$CB,$E7,$DF,$29,$E0,$6A,$B8,$4E,$A3,$9D,$F7,$74,$FC,$1A
	.dc.b	$E5,$8F,$9E,$76,$7D,$DC,$BE,$0C,$66,$38,$99,$CB,$A6,$5F,$40,$F3
	.dc.b	$57,$0B,$EA,$4A,$F7,$DF,$3A,$E1,$97,$A2,$F8,$57,$8A,$BC,$19,$7A
	.dc.b	$E3,$C3,$1C,$6B,$86,$3E,$27,$E4,$61,$E2,$8C,$65,$F0,$70,$E5,$3E
	.dc.b	$08,$C6,$3E,$0C,$7A,$E5,$E7,$7F,$14,$7C,$EA,$E0,$FC,$13,$18,$F8
	.dc.b	$0F,$1B,$8B,$EB,$8F,$8A,$3E,$BE,$3E,$2A,$E1,$5A,$9E,$06,$B1,$F2
	.dc.b	$61,$E0,$CB,$F2,$98,$F3,$AF,$F2,$99,$F5,$0B,$3F,$B6,$6E,$1F,$B2
	.dc.b	$77,$E3,$BE,$29,$F1,$5F,$E5,$88,$C3,$63,$33,$FE,$5E,$9C,$BF,$E6
	.dc.b	$0B,$79,$2F,$47,$FC,$C3,$18,$6C,$66,$78,$35,$F3,$B9,$CF,$F8,$53
	.dc.b	$77,$FD,$F3,$64,$3E,$13,$F9,$93,$FC,$56,$34,$FE,$A1,$B8,$78,$E3
	.dc.b	$F9,$7C,$F0,$7F,$8A,$4E,$89,$87,$F3,$79,$D6,$1E,$48,$EC,$4A,$E7
	.dc.b	$C5,$FE,$F6,$C7,$CF,$39,$AE,$15,$A3,$13,$A7,$C3,$3D,$EB,$94,$7F
	.dc.b	$1B,$C3,$C6,$E8,$E4,$5B,$EA,$4B,$FF,$4A,$E1,$E8,$7B,$46,$67,$A6
	.dc.b	$1D,$DF,$19,$83,$A4,$F2,$C3,$C7,$1F,$2E,$76,$D3,$C7,$5D,$5C,$36
	.dc.b	$FC,$51,$D6,$35,$8F,$D3,$7F,$C5,$93,$A9,$E3,$87,$A0,$8F,$E0,$23
	.dc.b	$55,$C6,$73,$3C,$0D,$61,$E8,$9D,$47,$86,$77,$5C,$B1,$F9,$B8,$79
	.dc.b	$F0,$E1,$18,$AD,$C6,$31,$F8,$38,$F4,$C7,$F2,$53,$97,$E4,$9D,$CF
	.dc.b	$19,$C2,$FA,$38,$36,$7A,$49,$8F,$96,$F8,$1F,$5E,$37,$7E,$13,$C5
	.dc.b	$3B,$8C,$D7,$E8,$A3,$18,$FD,$08,$C6,$1E,$78,$C6,$1D,$5B,$7E,$4B
	.dc.b	$E3,$8F,$CE,$C7,$BE,$1C,$63,$18,$F4,$BF,$39,$C6,$74,$FC,$33,$DA
	.dc.b	$E1,$F8,$15,$D1,$E3,$87,$D3,$7C,$97,$DB,$1B,$1F,$26,$7C,$E7,$E2
	.dc.b	$AF,$F2,$F8,$FA,$27,$C0,$FE,$34,$C3,$E2,$39,$9E,$2C,$7D,$4F,$E7
	.dc.b	$30,$F0,$34,$F0,$38,$BF,$04,$72,$C7,$EA,$57,$29,$CC,$FD,$4C,$7B
	.dc.b	$E1,$E2,$BF,$81,$3D,$4F,$14,$EB,$1F,$8D,$3A,$BE,$67,$C0,$C7,$F6
	.dc.b	$D7,$F8,$5D,$BF,$A1,$E3,$FC,$75,$B8,$7F,$28,$94,$61,$FF,$73,$9E
	.dc.b	$C2,$1F,$9C,$3F,$91,$71,$7C,$C9,$91,$EA,$8A,$FA,$6B,$E7,$46,$30
	.dc.b	$F3,$E1,$E6,$AF,$E6,$33,$F5,$8D,$46,$67,$F8,$1C,$66,$7E,$11,$E2
	.dc.b	$21,$EA,$3F,$8F,$BE,$DC,$3E,$0B,$63,$27,$B8,$FE,$90,$F5,$7C,$67
	.dc.b	$FB,$0B,$67,$D0,$3D,$DF,$E6,$71,$87,$86,$1E,$6A,$E1,$5B,$6C,$F0
	.dc.b	$8C,$47,$FA,$5E,$B9,$3F,$EC,$0F,$DA,$1E,$5F,$F4,$0F,$09,$F4,$4F
	.dc.b	$A1,$FE,$87,$DB,$FE,$AE,$FA,$2B,$E1,$1E,$15,$BC,$BF,$DB,$78,$72
	.dc.b	$C3,$24,$8F,$F6,$6F,$B9,$FC,$F3,$EF,$9E,$DF,$F8,$2A,$7F,$0D,$18
	.dc.b	$70,$1C,$3B,$6B,$FE,$12,$73,$E7,$7D,$58,$E1,$3F,$91,$8C,$4E,$B9
	.dc.b	$3F,$BD,$F0,$3D,$A7,$7E,$E7,$31,$8C,$7F,$C0,$38,$7E,$06,$F8,$5E
	.dc.b	$E7,$FC,$37,$5B,$55,$F8,$7E,$07,$DF,$3D,$EE,$18,$F2,$BE,$B9,$78
	.dc.b	$32,$CE,$5C,$F2,$E5,$8F,$1B,$D9,$FE,$14,$F8,$07,$8F,$F2,$21,$18
	.dc.b	$7E,$5B,$5F,$A0,$D7,$FC,$76,$E6,$BE,$53,$71,$FD,$D3,$67,$FC,$47
	.dc.b	$F0,$3F,$E2,$BA,$FA,$1E,$17,$B5,$FF,$13,$E4,$F5,$3F,$81,$7C,$07
	.dc.b	$DA,$8C,$61,$FE,$2C,$8C,$4F,$F0,$7B,$E5,$3F,$56,$31,$BF,$CC,$87
	.dc.b	$C0,$72,$57,$F8,$68,$C3,$6B,$8E,$3F,$E3,$A9,$E1,$87,$67,$7C,$7F
	.dc.b	$C5,$DB,$86,$EF,$8B,$E0,$6C,$6A,$30,$7A,$E7,$94,$E8,$E3,$96,$C8
	.dc.b	$E0,$D4,$F1,$3C,$05,$3A,$8B,$F4,$47,$08,$F8,$35,$E0,$78,$5F,$BB
	.dc.b	$2F,$90,$DC,$7E,$AD,$FD,$4B,$FE,$6D,$8F,$93,$0F,$F2,$FE,$3F,$4A
	.dc.b	$38,$47,$8B,$2F,$98,$F6,$BF,$15,$79,$23,$04,$3D,$19,$78,$A7,$E0
	.dc.b	$47,$B6,$FC,$0E,$B1,$F2,$BE,$2A,$E0,$F3,$C3,$C2,$EC,$F8,$B8,$78
	.dc.b	$5F,$86,$7C,$A2,$9A,$9E,$6F,$1C,$BE,$D1,$31,$F5,$CF,$E3,$F1,$F2
	.dc.b	$E1,$F9,$6B,$F2,$63,$E6,$79,$5F,$CB,$7C,$13,$E6,$C7,$C8,$7C,$B9
	.dc.b	$D4,$FE,$74,$F8,$4F,$C4,$7D,$97,$E0,$CB,$EA,$CE,$6B,$F4,$BE,$27
	.dc.b	$F1,$9B,$F1,$9F,$EA,$3B,$FC,$BC,$66,$3E,$33,$5F,$F4,$DF,$7F,$FA
	.dc.b	$72,$31,$08,$C4,$FF,$35,$6C,$7F,$B0,$1A,$FF,$7D,$7E,$17,$43,$DB
	.dc.b	$FE,$CD,$5F,$B5,$AE,$E7,$BC,$4E,$F8,$7E,$54,$6A,$7F,$79,$39,$8E
	.dc.b	$06,$72,$FF,$06,$46,$67,$B4,$FA,$4F,$D3,$C7,$1D,$F3,$8F,$E0,$E3
	.dc.b	$17,$ED,$AF,$24,$73,$8C,$5F,$8B,$1F,$5C,$6F,$0C,$63,$F8,$DC,$3D
	.dc.b	$B1,$87,$F6,$E6,$27,$CF,$5F,$E0,$99,$F1,$CF,$9A,$31,$3D,$4D,$D7
	.dc.b	$59,$F6,$57,$53,$8D,$F7,$2D,$70,$9F,$8E,$70,$71,$3E,$03,$93,$DE
	.dc.b	$3F,$DB,$B8,$7C,$47,$EF,$1E,$5B,$F1,$63,$F7,$B1,$F8,$73,$AC,$7C
	.dc.b	$58,$73,$8F,$BB,$3E,$C7,$9E,$3E,$06,$CF,$90,$E1,$8F,$86,$3F,$18
	.dc.b	$7D,$4A,$EE,$E2,$FB,$BC,$8F,$4C,$77,$78,$DF,$76,$DF,$77,$81,$E6
	.dc.b	$C7,$DA,$FB,$CF,$B2,$7C,$93,$AB,$AB,$F0,$C7,$8E,$7E,$84,$7F,$53
	.dc.b	$F7,$5E,$FF,$F9,$1F,$83,$FC,$29,$18,$C3,$FB,$CE,$6F,$92,$7C,$86
	.dc.b	$E7,$FC,$D1,$5B,$CF,$FC,$17,$18,$C7,$FC,$3D,$16,$3F,$C4,$F3,$B8
	.dc.b	$FF,$38,$07,$0F,$F9,$CF,$E2,$9F,$37,$DE,$02,$33,$5F,$E0,$EE,$CF
	.dc.b	$F2,$27,$BF,$BA,$D5,$DB,$DC,$79,$8E,$6E,$35,$FF,$11,$78,$0B,$E1
	.dc.b	$7C,$FE,$DB,$F7,$D5,$FE,$39,$6E,$5E,$79,$D7,$1C,$8F,$F2,$2C,$FF
	.dc.b	$05,$CB,$DE,$D7,$FD,$20,$D7,$F8,$7F,$C1,$7F,$94,$27,$FC,$AE,$D8
	.dc.b	$FE,$41,$E5,$FC,$2D,$5E,$33,$E6,$7F,$D0,$1D,$0A,$FD,$3E,$1C,$63
	.dc.b	$18,$76,$EB,$DD,$73,$C3,$FC,$D9,$97,$82,$73,$4C,$BB,$F0,$63,$C3
	.dc.b	$75,$BE,$4F,$0D,$C7,$F9,$CE,$7E,$7F,$D9,$DE,$9E,$AE,$56,$B3,$FB
	.dc.b	$BE,$B3,$E5,$B3,$1F,$00,$CB,$E4,$51,$B9,$E5,$18,$9E,$3C,$1F,$02
	.dc.b	$31,$3A,$AE,$11,$F9,$66,$C7,$E5,$1B,$1F,$EC,$56,$BE,$28,$E5,$E4
	.dc.b	$78,$E7,$9D,$77,$7F,$CE,$F3,$DE,$F5,$5B,$6C,$72,$25,$FB,$CF,$39
	.dc.b	$6B,$9B,$6B,$8B,$9A,$B7,$E1,$9B,$9F,$A6,$6D,$7C,$E9,$AF,$6A,$E1
	.dc.b	$87,$88,$F8,$53,$FE,$CB,$69,$FE,$C7,$C3,$C6,$E2,$BD,$1E,$A3,$D6
	.dc.b	$7A,$FE,$CC,$7E,$94,$E6,$31,$9F,$FB,$C6,$31,$7E,$DC,$3E,$DC,$62
	.dc.b	$FC,$B3,$DB,$0E,$AE,$CC,$61,$D1,$FE,$0F,$5C,$70,$FB,$59,$7E,$26
	.dc.b	$31,$3F,$0F,$1E,$F8,$74,$C7,$B9,$E8,$71,$87,$29,$DE,$3D,$ED,$87
	.dc.b	$47,$F2,$B8,$F7,$C3,$8C,$77,$3C,$E7,$A7,$1E,$D7,$CE,$F9,$65,$F8
	.dc.b	$09,$CD,$6F,$2F,$91,$87,$48,$EF,$97,$91,$D1,$F0,$A3,$57,$CE,$B8
	.dc.b	$E3,$DE,$73,$97,$9E,$30,$4B,$E6,$7B,$A7,$C1,$97,$BC,$F3,$D7,$19
	.dc.b	$F6,$E5,$E8,$3E,$5E,$1E,$08,$CB,$63,$31,$E7,$C3,$D5,$1C,$67,$E8
	.dc.b	$4F,$03,$45,$3E,$09,$6F,$C9,$87,$B3,$2E,$75,$DC,$95,$F4,$EF,$B3
	.dc.b	$D1,$AF,$46,$DF,$4C,$3E,$24,$63,$0E,$D5,$E1,$7E,$29,$88,$D4,$F8
	.dc.b	$A7,$C1,$7E,$27,$E6,$46,$5F,$04,$66,$FC,$11,$F4,$9B,$5E,$6C,$BE
	.dc.b	$E4,$FA,$6B,$86,$1E,$16,$E1,$D6,$B7,$5E,$03,$D1,$1E,$C9,$DC,$77
	.dc.b	$AF,$5C,$F1,$CB,$EF,$61,$EC,$AD,$4F,$A7,$1F,$C1,$E5,$70,$E3,$5A
	.dc.b	$3B,$96,$F9,$1F,$82,$7C,$8E,$31,$FC,$0B,$63,$DC,$7C,$A2,$9C,$23
	.dc.b	$E6,$65,$EE,$C3,$E3,$C6,$32,$F4,$BE,$EB,$F0,$4E,$AF,$9D,$F1,$C3
	.dc.b	$D3,$5F,$3F,$0E,$35,$E0,$BF,$7B,$EC,$9D,$63,$F0,$72,$FF,$0F,$CF
	.dc.b	$E6,$0F,$F1,$0D,$7F,$20,$6E,$1F,$C7,$1F,$F1,$2F,$8B,$FD,$A3,$87
	.dc.b	$8C,$7F,$BB,$77,$46,$27,$53,$D9,$AF,$6F,$5F,$FB,$BB,$E2,$C1,$18
	.dc.b	$8D,$71,$76,$9C,$B6,$78,$E9,$FC,$E9,$BB,$FF,$39,$6D,$FF,$8D,$BD
	.dc.b	$67,$AF,$FE,$B1,$AD,$61,$FC,$F2,$F7,$7F,$CA,$1B,$7C,$23,$53,$1F
	.dc.b	$E9,$B3,$AB,$E9,$1F,$CF,$A3,$9E,$3E,$88,$DD,$7F,$81,$2F,$A4,$60
	.dc.b	$98,$FA,$E7,$58,$74,$9F,$24,$78,$67,$51,$FD,$D6,$B9,$C6,$09,$8F
	.dc.b	$BA,$75,$7C,$F0,$F2,$56,$A7,$D1,$87,$81,$ED,$7C,$67,$50,$C3,$BC
	.dc.b	$E4,$91,$C8,$EE,$72,$31,$3C,$A7,$8C,$62,$78,$BE,$E3,$89,$F1,$1C
	.dc.b	$DF,$A1,$FF,$38,$C7,$8F,$0F,$66,$3F,$40,$98,$FA,$E7,$91,$DC,$87
	.dc.b	$13,$CC,$78,$71,$F3,$57,$49,$CE,$1E,$1A,$FC,$64,$E5,$EA,$71,$C3
	.dc.b	$CB,$8F,$BA,$30,$78,$2B,$93,$E1,$38,$B8,$33,$1C,$5C,$63,$EC,$9F
	.dc.b	$8B,$8F,$D4,$9C,$E3,$EF,$8C,$56,$DF,$F2,$85,$F3,$8C,$61,$F4,$23
	.dc.b	$15,$EE,$9F,$CE,$61,$63,$E2,$CE,$A3,$CF,$3A,$6C,$EB,$1F,$89,$87
	.dc.b	$8A,$3C,$98,$7D,$AB,$F0,$DF,$CA,$3B,$1F,$2B,$0E,$D7,$C6,$3D,$F5
	.dc.b	$C6,$FB,$47,$09,$FA,$ED,$9F,$66,$1F,$A9,$BE,$71,$89,$F1,$9F,$4F
	.dc.b	$0F,$55,$FC,$6B,$F4,$D6,$EB,$C2,$6F,$0E,$2F,$D9,$AF,$31,$C5,$CE
	.dc.b	$3F,$51,$E2,$6E,$BD,$4E,$31,$F2,$BF,$94,$7F,$10,$61,$FC,$49,$C4
	.dc.b	$F9,$07,$D0,$39,$61,$E8,$79,$3F,$04,$C3,$CA,$3A,$C6,$2B,$F1,$EE
	.dc.b	$E7,$FD,$BB,$BF,$FB,$92,$BF,$8B,$B6,$7F,$62,$D9,$D9,$FA,$08,$E4
	.dc.b	$FB,$7D,$7B,$8F,$C5,$8C,$1F,$2A,$FC,$6F,$F2,$59,$FB,$07,$0C,$7A
	.dc.b	$46,$48,$7C,$02,$D7,$E7,$A7,$84,$7C,$68,$C3,$C3,$0F,$64,$FA,$A3
	.dc.b	$F6,$ED,$7F,$A4,$37,$57,$AC,$F5,$FC,$B5,$E0,$C4,$FF,$62,$9D,$3F
	.dc.b	$EC,$BA,$E3,$5C,$E7,$32,$C3,$F8,$DD,$6A,$3E,$59,$CE,$BC,$9E,$43
	.dc.b	$D4,$75,$7F,$50,$EB,$FF,$78,$FA,$B7,$B8,$FF,$54,$BE,$15,$6F,$8E
	.dc.b	$3D,$A7,$2D,$8C,$BE,$53,$DB,$1A,$70,$7C,$0F,$51,$DF,$FC,$D3,$C5
	.dc.b	$E1,$AB,$F0,$07,$BF,$9E,$0F,$8F,$39,$8C,$63,$CF,$0A,$B1,$EA,$FF
	.dc.b	$84,$1A,$FF,$86,$9A,$FF,$44,$38,$47,$F8,$5A,$9E,$73,$BF,$BC,$61
	.dc.b	$B1,$9A,$FB,$79,$FC,$1E,$17,$AC,$F4,$F7,$84,$BD,$EB,$EC,$C7,$94
	.dc.b	$31,$E5,$87,$18,$D9,$F8,$8F,$55,$EB,$7C,$0F,$59,$E9,$53,$E1,$CE
	.dc.b	$AB,$88,$70,$FF,$8F,$1A,$FC,$36,$BF,$DA,$1A,$FE,$63,$C2,$F6,$63
	.dc.b	$F9,$AF,$55,$EA,$FF,$6E,$F5,$BE,$7C,$62,$FF,$3F,$18,$9F,$EE,$77
	.dc.b	$CA,$55,$B5,$19,$D7,$17,$F6,$9B,$C7,$FD,$F1,$E1,$3B,$C3,$AF,$0F
	.dc.b	$CB,$C3,$AC,$79,$6F,$FC,$07,$8F,$29,$F5,$4F,$0C,$AC,$7B,$EF,$95
	.dc.b	$6E,$39,$9E,$CB,$E7,$7C,$63,$27,$AA,$B8,$C7,$86,$3A,$8C,$BE,$23
	.dc.b	$DF,$0C,$E3,$EB,$8E,$C7,$F0,$8C,$BC,$47,$C0,$C3,$DC,$DA,$EB,$7D
	.dc.b	$2B,$B6,$3E,$67,$B1,$E6,$C3,$E9,$65,$ED,$26,$5E,$B7,$A4,$FE,$92
	.dc.b	$3D,$4D,$3C,$A5,$3E,$06,$1E,$78,$E4,$70,$9E,$B3,$E2,$7A,$98,$7C
	.dc.b	$E7,$C2,$CB,$E4,$BB,$9E,$44,$70,$43,$89,$F4,$B1,$F8,$58,$7B,$A7
	.dc.b	$C7,$3E,$8B,$F0,$4E,$63,$95,$7C,$93,$8D,$F8,$A3,$DA,$75,$C7,$E6
	.dc.b	$3E,$0B,$F5,$E1,$DA,$30,$F9,$DE,$31,$A7,$C5,$19,$C3,$C5,$1D,$BC
	.dc.b	$4F,$E2,$6F,$D8,$6B,$9C,$FF,$21,$18,$8F,$E9,$BE,$0F,$F4,$E8,$C4
	.dc.b	$23,$13,$FE,$BC,$6C,$7F,$23,$6B,$FE,$68,$7F,$2F,$D2,$F4,$3F,$9F
	.dc.b	$65,$FB,$7A,$EE,$7E,$82,$75,$C9,$F9,$E8,$CC,$FE,$F6,$73,$1F,$A6
	.dc.b	$33,$9F,$F9,$DA,$33,$3D,$67,$D0,$75,$BF,$0D,$7A,$63,$17,$DD,$DD
	.dc.b	$77,$AD,$C7,$B4,$D9,$E1,$37,$5E,$18,$DC,$7C,$13,$73,$E3,$C3,$E1
	.dc.b	$CE,$A7,$D2,$7F,$60,$70,$F4,$31,$7E,$59,$E2,$D8,$EB,$7E,$BB,$F4
	.dc.b	$4E,$5B,$3E,$7A,$E2,$D3,$F6,$18,$7C,$CC,$7C,$D1,$8C,$7F,$BA,$46
	.dc.b	$2F,$99,$71,$F8,$18,$F7,$6D,$F5,$9D,$C7,$CF,$AF,$19,$D8,$AF,$A4
	.dc.b	$E3,$1D,$67,$E0,$9C,$E3,$26,$DC,$C6,$D6,$1F,$82,$9D,$46,$30,$F9
	.dc.b	$B8,$FF,$0D,$8E,$B8,$78,$EB,$D2,$F3,$BE,$53,$F8,$1C,$3D,$6F,$F8
	.dc.b	$CB,$D4,$7A,$BF,$E1,$68,$FF,$63,$C6,$33,$FF,$03,$46,$E7,$4D,$8C
	.dc.b	$A9,$E0,$FF,$7B,$8F,$F0,$2F,$C4,$3E,$27,$FC,$4D,$C7,$FE,$24,$BD
	.dc.b	$B8,$9D,$57,$F7,$8F,$5D,$EB,$70,$8F,$F6,$DC,$61,$FF,$02,$F6,$7F
	.dc.b	$8F,$BD,$7D,$67,$FE,$4E,$AD,$3E,$33,$17,$A5,$1A,$C7,$FC,$35,$E1
	.dc.b	$AF,$53,$9C,$7F,$DD,$7E,$BF,$EE,$B0,$FF,$15,$35,$CF,$27,$C6,$51
	.dc.b	$97,$FC,$C2,$FF,$02,$C7,$FC,$83,$F7,$5D,$67,$35,$EA,$F0,$53,$9E
	.dc.b	$37,$AF,$FE,$4A,$CB,$FC,$A8,$DC,$3F,$C9,$F3,$AE,$AF,$F2,$A5,$7F
	.dc.b	$0D,$AD,$36,$3F,$CB,$1D,$0F,$2F,$FC,$EB,$C4,$F2,$7F,$CD,$37,$FC
	.dc.b	$B2,$7F,$CB,$B8,$77,$BE,$19,$30,$E1,$C5,$D2,$FF,$CB,$3D,$6F,$5B
	.dc.b	$C7,$78,$6A,$DD,$3B,$CF,$FC,$FB,$18,$AD,$B6,$3C,$95,$FE,$8C,$F5
	.dc.b	$5E,$B7,$86,$3A,$61,$CD,$FE,$85,$8F,$FA,$49,$57,$0B,$E7,$18,$BF
	.dc.b	$2B,$63,$FD,$1E,$D3,$F5,$13,$C5,$B3,$E1,$7A,$F8,$F6,$75,$AF,$C8
	.dc.b	$9C,$DF,$82,$73,$87,$95,$FF,$39,$46,$70,$E5,$87,$36,$C6,$4E,$51
	.dc.b	$8B,$F7,$9F,$15,$E1,$1A,$24,$68,$F1,$96,$BC,$6F,$18,$DE,$1E,$99
	.dc.b	$F2,$CE,$EF,$C1,$39,$BF,$93,$5A,$BF,$F4,$83,$6F,$DD,$3A,$3E,$47
	.dc.b	$A4,$F5,$5E,$B7,$83,$3F,$B1,$39,$8C,$5F,$FB,$32,$30,$FA,$6F,$9D
	.dc.b	$F2,$8C,$63,$F7,$AB,$79,$79,$E3,$15,$B8,$C5,$7C,$5A,$DE,$3E,$8B
	.dc.b	$DE,$3E,$2A,$56,$9F,$90,$71,$C3,$A5,$F1,$8C,$57,$1A,$E1,$1F,$4B
	.dc.b	$0F,$C3,$46,$32,$F9,$F1,$8C,$39,$D7,$68,$C5,$F3,$9C,$CF,$3C,$3E
	.dc.b	$4C,$62,$F9,$C6,$2B,$81,$2B,$E3,$9E,$62,$DF,$66,$CF,$16,$9F,$4E
	.dc.b	$FB,$65,$F4,$AF,$A5,$F1,$25,$FC,$36,$DF,$19,$F9,$AF,$43,$E6,$DF
	.dc.b	$B3,$1F,$8E,$7C,$5C,$3B,$4F,$9E,$F9,$BF,$04,$E8,$67,$0F,$44,$7D
	.dc.b	$E3,$75,$D2,$FD,$39,$78,$A7,$31,$8B,$E4,$74,$8E,$F9,$7C,$C8,$C6
	.dc.b	$1D,$88,$FD,$76,$BE,$CC,$BB,$47,$C3,$33,$8F,$8E,$3E,$4C,$63,$0F
	.dc.b	$16,$1E,$49,$E3,$7D,$4F,$09,$0E,$C7,$CB,$BE,$8F,$A8,$F8,$C6,$A7
	.dc.b	$A3,$BA,$FA,$04,$AF,$1C,$75,$AF,$A5,$8F,$8A,$39,$61,$EE,$C7,$E2
	.dc.b	$46,$0E,$18,$7A,$A7,$D9,$87,$1A,$DE,$1E,$6C,$7E,$6E,$5F,$79,$F2
	.dc.b	$9F,$66,$F8,$E3,$EB,$BE,$38,$F9,$E3,$17,$C6,$B7,$7C,$67,$C7,$8F
	.dc.b	$86,$B5,$18,$3E,$14,$7B,$71,$F7,$C7,$C9,$7C,$D7,$F6,$63,$85,$FA
	.dc.b	$32,$E1,$8F,$8E,$3C,$13,$F0,$C9,$7D,$23,$DB,$7F,$27,$1F,$1D,$6A
	.dc.b	$FE,$43,$E5,$3F,$AE,$1F,$78,$B7,$FC,$91,$B1,$FC,$8D,$DF,$86,$F5
	.dc.b	$BF,$B9,$47,$F9,$02,$31,$BF,$F8,$F9,$B1,$99,$DB,$DB,$D6,$FE,$F2
	.dc.b	$FF,$92,$30,$46,$23,$5C,$1F,$E4,$89,$CB,$67,$8E,$9F,$C8,$1B,$A7
	.dc.b	$F4,$06,$D8,$7A,$AF,$5B,$FD,$71,$5B,$C3,$F9,$F5,$F0,$BF,$E6,$AD
	.dc.b	$BE,$11,$A9,$8F,$F3,$F9,$D5,$F4,$8F,$E9,$D1,$CF,$1F,$4C,$6E,$BF
	.dc.b	$C1,$37,$D2,$30,$4C,$7D,$93,$AC,$3A,$4F,$92,$3C,$33,$A8,$FE,$E9
	.dc.b	$5C,$E3,$04,$C7,$DD,$3A,$BE,$78,$79,$2B,$53,$E8,$C3,$C0,$F6,$BE
	.dc.b	$33,$A8,$F9,$77,$CB,$1F,$71,$2F,$8C,$FA,$EF,$89,$1F,$A9,$8F,$9E
	.dc.b	$31,$1E,$D3,$81,$AB,$ED,$5E,$38,$F3,$1E,$B3,$07,$1C,$3C,$D7,$D6
	.dc.b	$37,$1E,$9C,$7E,$06,$1F,$E3,$DA,$D5,$71,$8F,$1C,$EA,$3C,$47,$67
	.dc.b	$0E,$CC,$9E,$83,$04,$9E,$B1,$C0,$98,$FE,$22,$7A,$3F,$25,$F6,$1C
	.dc.b	$E3,$78,$7C,$CC,$3D,$55,$B7,$FC,$8D,$7D,$A3,$18,$7C,$C8,$C5,$76
	.dc.b	$39,$47,$BA,$77,$87,$AA,$FA,$D7,$CC,$BE,$33,$E3,$AF,$43,$67,$58
	.dc.b	$FD,$68,$C1,$EA,$BE,$6F,$C5,$BE,$71,$83,$D3,$3C,$5C,$61,$F5,$63
	.dc.b	$0F,$9E,$FD,$37,$F7,$23,$0F,$76,$D7,$E0,$AB,$8C,$EC,$87,$E6,$8F
	.dc.b	$88,$79,$5D,$BF,$6B,$1F,$BF,$87,$A2,$FA,$9E,$59,$E3,$7D,$DB,$1F
	.dc.b	$46,$F9,$DF,$DA,$9F,$31,$E2,$71,$5F,$0A,$38,$9E,$6B,$F0,$3B,$C7
	.dc.b	$E8,$CF,$C6,$8D,$5F,$69,$FB,$71,$ED,$3E,$9C,$F9,$5B,$87,$95,$EE
	.dc.b	$7E,$5E,$B9,$63,$F7,$23,$38,$FC,$29,$CC,$7C,$73,$38,$79,$2F,$E2
	.dc.b	$C7,$90,$CD,$FD,$87,$E6,$4F,$9B,$1F,$6C,$62,$75,$87,$C4,$BF,$34
	.dc.b	$FA,$0E,$85,$3D,$07,$63,$E6,$9C,$CF,$31,$B8,$F5,$3B,$C7,$E8,$39
	.dc.b	$C3,$D4,$FF,$49,$F2,$9E,$6F,$F6,$88,$FF,$62,$46,$1C,$07,$0E,$DA
	.dc.b	$FF,$70,$73,$DC,$E2,$E7,$5F,$EC,$F8,$C5,$78,$63,$13,$FE,$D3,$F3
	.dc.b	$7F,$B3,$56,$AB,$C1,$5A,$78,$1F,$D8,$63,$57,$E0,$BE,$5A,$6E,$B8
	.dc.b	$9B,$8F,$1B,$4E,$62,$BC,$18,$78,$A7,$FD,$CD,$F1,$5F,$EF,$9A,$F8
	.dc.b	$6D,$7E,$43,$5F,$F0,$8B,$5E,$3D,$1E,$38,$E1,$7E,$D6,$D7,$FB,$CF
	.dc.b	$F3,$B7,$BF,$E9,$E4,$79,$DE,$5C,$9E,$63,$CE,$FF,$7D,$79,$EF,$0F
	.dc.b	$FD,$F9,$39,$AF,$F0,$34,$E5,$FF,$04,$46,$56,$1D,$B5,$CE,$1F,$E0
	.dc.b	$D9,$CE,$3E,$1F,$3F,$86,$3C,$E5,$18,$CB,$83,$D7,$1E,$93,$AC,$3A
	.dc.b	$46,$32,$E1,$3E,$03,$CF,$7E,$AA,$DE,$3F,$06,$BA,$36,$3E,$0C,$EE
	.dc.b	$B8,$9C,$CF,$4E,$57,$0F,$1E,$3E,$09,$F2,$1C,$0A,$70,$2C,$79,$EF
	.dc.b	$BC,$75,$CB,$E1,$65,$E4,$C3,$B6,$5E,$5A,$E9,$3D,$CD,$3F,$6A,$FD
	.dc.b	$B5,$CF,$2F,$54,$B0,$EB,$5A,$8E,$07,$F9,$07,$1F,$5C,$79,$23,$D3
	.dc.b	$3F,$57,$2E,$E7,$83,$1F,$15,$6D,$C3,$D6,$3E,$21,$1C,$12,$BE,$34
	.dc.b	$65,$F1,$65,$F8,$59,$F3,$70,$FE,$12,$73,$9F,$F9,$5E,$31,$1F,$35
	.dc.b	$B9,$7E,$19,$CC,$7F,$97,$79,$78,$63,$FC,$55,$B1,$FE,$7E,$6B,$F9
	.dc.b	$76,$BC,$4F,$B7,$E2,$BC,$7F,$E2,$35,$FA,$19,$D3,$FA,$39,$D7,$0F
	.dc.b	$F9,$B6,$35,$2B,$C6,$7F,$A5,$BD,$CF,$E2,$0E,$71,$9C,$3B,$BE,$98
	.dc.b	$FB,$22,$78,$EB,$DA,$FF,$31,$18,$AE,$18,$79,$70,$E7,$87,$C1,$3D
	.dc.b	$AE,$1F,$84,$76,$AE,$4E,$F0,$E5,$8F,$E9,$63,$E3,$46,$1F,$29,$8C
	.dc.b	$3D,$53,$B8,$F8,$47,$5B,$FE,$77,$5E,$98,$C4,$EF,$1F,$8F,$1F,$E9
	.dc.b	$7C,$3E,$3E,$3B,$3C,$07,$1B,$F1,$17,$1F,$0D,$79,$0F,$0C,$6A,$3E
	.dc.b	$63,$63,$E2,$4F,$49,$F9,$27,$27,$A4,$78,$F1,$F4,$CF,$23,$BD,$7F
	.dc.b	$25,$F9,$E7,$97,$FD,$63,$6F,$EA,$91,$8D,$3F,$A5,$0F,$D3,$1F,$2F
	.dc.b	$B8,$FE,$55,$3B,$CF,$FA,$74,$63,$0F,$EB,$D1,$63,$FB,$7C,$EE,$3F
	.dc.b	$DB,$6D,$7F,$9A,$7A,$0F,$4B,$B4,$7F,$B0,$E3,$15,$FD,$77,$77,$BE
	.dc.b	$F5,$37,$6B,$DB,$E4,$39,$61,$A8,$CE,$BF,$DE,$30,$E7,$79,$A9,$8F
	.dc.b	$F6,$BC,$3A,$47,$92,$B7,$39,$C3,$9B,$EA,$9F,$83,$5E,$08,$FF,$07
	.dc.b	$CF,$92,$3F,$DF,$6D,$7F,$AA,$78,$0F,$19,$E4,$7F,$07,$F0,$F3,$18
	.dc.b	$AF,$DC,$47,$F5,$BF,$25,$E4,$FF,$84,$BA,$13,$C3,$0E,$57,$FE,$ED
	.dc.b	$C7,$B4,$62,$11,$87,$A9,$FE,$F3,$C3,$AE,$7B,$9D,$61,$DA,$BA,$65
	.dc.b	$E5,$8C,$47,$48,$C6,$3E,$7C,$BC,$CF,$96,$7B,$9B,$CB,$B5,$70,$8F
	.dc.b	$15,$70,$7B,$CE,$88,$F2,$CB,$CE,$F4,$C7,$CF,$87,$7C,$7C,$B1,$89
	.dc.b	$EB,$97,$29,$CE,$5E,$19,$F2,$1D,$72,$F1,$38,$9D,$E5,$C7,$1F,$76
	.dc.b	$16,$BC,$E7,$8F,$0E,$99,$79,$B0,$E3,$39,$C3,$82,$9D,$61,$E9,$24
	.dc.b	$74,$AF,$9F,$3A,$8C,$4E,$9F,$33,$E4,$7B,$12,$7C,$27,$2C,$7D,$B9
	.dc.b	$7C,$59,$E2,$74,$3E,$39,$C6,$BB,$46,$EF,$C5,$3C,$09,$19,$AE,$B8
	.dc.b	$FC,$D8,$EC,$67,$2F,$82,$E2,$3E,$3C,$F0,$CB,$31,$8A,$B1,$F2,$30
	.dc.b	$F0,$DF,$18,$C5,$F1,$C3,$D0,$7A,$CF,$9C,$E7,$2F,$87,$97,$C9,$C3
	.dc.b	$9E,$5E,$72,$61,$FA,$98,$CB,$E2,$C3,$F0,$D9,$5C,$39,$C7,$C8,$6B
	.dc.b	$F1,$2F,$D4,$F9,$0C,$57,$5C,$7C,$98,$FC,$EA,$DD,$F2,$CB,$DD,$87
	.dc.b	$80,$E4,$E2,$3E,$E6,$3E,$3B,$F6,$61,$D2,$FE,$15,$FD,$17,$8E,$1E
	.dc.b	$E7,$A6,$1D,$E3,$0F,$48,$DE,$3F,$3B,$1E,$D1,$F3,$72,$F8,$AE,$8E
	.dc.b	$79,$7A,$EB,$84,$66,$FE,$2D,$70,$7C,$13,$F0,$EF,$AD,$7E,$36,$B5
	.dc.b	$7C,$E3,$A9,$EC,$C3,$CF,$1F,$A1,$C3,$9C,$FC,$FA,$E2,$EF,$0E,$CF
	.dc.b	$8E,$BE,$3E,$1E,$0C,$7E,$36,$1C,$8F,$3E,$5F,$2E,$FE,$C9,$E2,$2C
	.dc.b	$6A,$B9,$65,$F6,$B0,$F4,$4E,$6B,$C0,$7C,$9B,$E9,$87,$DB,$AF,$51
	.dc.b	$C0,$F2,$CF,$03,$E2,$4F,$0B,$F8,$31,$F9,$19,$E0,$F5,$C3,$C7,$8F
	.dc.b	$C0,$C7,$D9,$5D,$1B,$5B,$3B,$63,$F0,$A7,$9C,$78,$2F,$9B,$F0,$67
	.dc.b	$D0,$F1,$AF,$A2,$FF,$AC,$CF,$E5,$A5,$8F,$EA,$4D,$C3,$FB,$03,$BE
	.dc.b	$A7,$90,$F2,$7F,$DF,$71,$86,$C6,$67,$B1,$FE,$C4,$70,$F7,$F1,$DD
	.dc.b	$BA,$8F,$EC,$71,$87,$8E,$28,$C4,$71,$9E,$13,$97,$FB,$34,$F1,$CB
	.dc.b	$FA,$7B,$74,$FF,$69,$79,$0F,$27,$FC,$FD,$1F,$E0,$29,$D6,$3F,$DD
	.dc.b	$DA,$F9,$30,$FE,$F1,$7C,$63,$BC,$47,$FB,$04,$F0,$9D,$46,$27,$58
	.dc.b	$79,$1E,$78,$78,$A7,$C1,$1A,$C3,$C8,$EA,$30,$68,$E8,$7F,$7A,$BE
	.dc.b	$F7,$D6,$32,$D3,$FC,$0B,$1A,$AE,$B3,$A6,$CE,$AF,$C9,$3F,$04,$D1
	.dc.b	$E1,$8D,$3E,$43,$4F,$79,$F9,$F3,$CB,$1F,$AD,$5C,$8D,$CE,$9E,$51
	.dc.b	$AA,$E5,$01,$7E,$97,$C5,$5E,$82,$3B,$8D,$3D,$8D,$90,$E7,$7E,$5C
	.dc.b	$7E,$FE,$1E,$1B,$E7,$5A,$3C,$B3,$A3,$EF,$57,$1C,$72,$7D,$79,$F7
	.dc.b	$5F,$37,$C1,$87,$92,$31,$1E,$52,$46,$6F,$B5,$7C,$D8,$C3,$DA,$7F
	.dc.b	$01,$7E,$17,$BD,$79,$1D,$63,$F2,$9F,$51,$AA,$E9,$5C,$67,$58,$79
	.dc.b	$30,$EF,$7B,$3B,$9D,$31,$F8,$B1,$AA,$E3,$8F,$86,$B8,$47,$86,$75
	.dc.b	$5F,$32,$FB,$61,$E1,$8C,$61,$C6,$F9,$E3,$E6,$C7,$C0,$7C,$AB,$E0
	.dc.b	$FD,$BC,$39,$BA,$3E,$5C,$69,$F9,$D1,$A7,$CB,$87,$AA,$75,$3F,$50
	.dc.b	$D5,$F8,$AB,$F3,$18,$FA,$9F,$06,$3E,$7C,$3D,$38,$F9,$E7,$A9,$E7
	.dc.b	$3C,$98,$74,$9D,$C7,$4C,$7C,$37,$F8,$08,$F9,$E7,$23,$D6,$72,$7C
	.dc.b	$0F,$8A,$FC,$18,$F9,$CF,$A1,$5D,$23,$37,$E0,$9C,$D7,$39,$FC,$C4
	.dc.b	$71,$9F,$89,$39,$3C,$71,$97,$D8,$79,$70,$F6,$3E,$33,$D8,$71,$8F
	.dc.b	$E0,$EE,$F0,$F4,$BF,$51,$F8,$11,$A3,$B4,$7C,$03,$85,$F9,$AF,$D4
	.dc.b	$79,$A3,$27,$99,$CB,$E6,$9E,$77,$E6,$87,$90,$F2,$7F,$87,$47,$FB
	.dc.b	$64,$79,$3F,$ED,$F7,$FD,$A8,$D7,$53,$C5,$FF,$6C,$35,$FE,$C2,$D7
	.dc.b	$FB,$83,$5E,$ED,$7F,$8F,$3C,$7A,$5C,$5A,$BF,$2C,$62,$B7,$18,$9F
	.dc.b	$4D,$FF,$66,$AD,$57,$BA,$B4,$F0,$3C,$EE,$1F,$7E,$77,$C8,$79,$39
	.dc.b	$F2,$5F,$37,$4D,$44,$BF,$EF,$55,$C1,$B1,$FD,$F5,$46,$5F,$F0,$2F
	.dc.b	$2B,$C7,$67,$FC,$FF,$C4,$78,$BF,$53,$C8,$79,$3F,$5A,$73,$8F,$F4
	.dc.b	$59,$CB,$C6,$BE,$DE,$3D,$70,$C4,$FF,$4B,$8C,$2B,$E6,$AF,$9B,$9B
	.dc.b	$F1,$61,$E1,$7A,$E3,$E8,$9D,$1D,$5C,$79,$3D,$2F,$9B,$6F,$F1,$EF
	.dc.b	$83,$1F,$1B,$6B,$A6,$3B,$C3,$D2,$79,$6F,$C4,$F7,$CB,$38,$7B,$2B
	.dc.b	$84,$F9,$0E,$06,$AF,$AE,$5E,$AB,$EB,$97,$C0,$BE,$27,$8E,$38,$1A
	.dc.b	$C7,$C3,$87,$43,$B6,$3E,$1C,$3A,$47,$87,$0E,$91,$F2,$A3,$99,$C3
	.dc.b	$1F,$1B,$F6,$EF,$AD,$70,$8F,$71,$BC,$3C,$4C,$30,$4C,$7D,$38,$77
	.dc.b	$8C,$3E,$48,$CC,$77,$3C,$6F,$BD,$FF,$2A,$E5,$EA,$DC,$D2,$47,$F9
	.dc.b	$AF,$97,$FC,$2D,$3A,$8C,$1C,$8A,$E5,$FE,$1F,$1C,$27,$FC,$CE,$FD
	.dc.b	$EF,$09,$E4,$FE,$FF,$0F,$CA,$DF,$F8,$8E,$B5,$39,$C7,$8C,$FD,$FB
	.dc.b	$E3,$CB,$F8,$88,$E0,$66,$B8,$3E,$6C,$3F,$43,$5C,$23,$F1,$07,$08
	.dc.b	$CC,$F0,$8F,$49,$CE,$BB,$2C,$3C,$B1,$86,$C7,$43,$47,$13,$0E,$AB
	.dc.b	$DF,$1A,$7C,$33,$CE,$38,$46,$A7,$86,$1F,$0A,$37,$3D,$0D,$C6,$4F
	.dc.b	$E7,$F8,$7B,$E7,$94,$E8,$FB,$38,$F2,$32,$5A,$E5,$1F,$E5,$5A,$E5
	.dc.b	$18,$84,$61,$F9,$D1,$E0,$C7,$E5,$47,$CD,$8C,$63,$FB,$EF,$19,$E3
	.dc.b	$FF,$3A,$AF,$EA,$B1,$89,$FF,$54,$36,$3F,$D5,$ED,$7F,$D7,$CE,$75
	.dc.b	$FE,$91,$2F,$27,$0F,$4B,$FE,$99,$3A,$9C,$DF,$F9,$C7,$C5,$78,$DA
	.dc.b	$C3,$FB,$3C,$61,$FF,$3B,$5F,$F5,$5D,$1E,$43,$C8,$E1,$8F,$3C,$6A
	.dc.b	$C7,$93,$FD,$C5,$B8,$73,$BE,$58,$F6,$3C,$37,$CA,$B8,$DF,$FB,$75
	.dc.b	$AE,$A7,$FD,$9D,$18,$8F,$14,$73,$7F,$A8,$FA,$8F,$57,$A5,$FA,$31
	.dc.b	$4E,$6F,$85,$BC,$0F,$F7,$7A,$F2,$9E,$57,$F7,$7E,$6F,$F7,$DC,$FF
	.dc.b	$BD,$CF,$F7,$DF,$92,$F2,$BF,$79,$E2,$FF,$84,$B0,$E3,$97,$FB,$F3
	.dc.b	$04,$E7,$1E,$8F,$F8,$63,$2F,$08,$C7,$A6,$5F,$B3,$9E,$18,$73,$6C
	.dc.b	$78,$B2,$F0,$BF,$EA,$FC,$BC,$2E,$4E,$46,$71,$F0,$BF,$EB,$5C,$7C
	.dc.b	$38,$F8,$32,$F3,$CE,$F2,$F0,$34,$E9,$1A,$7E,$45,$F4,$7E,$65,$73
	.dc.b	$CB,$DC,$D8,$F0,$35,$E3,$7D,$9A,$76,$9F,$7D,$FD,$2C,$3E,$36,$5F
	.dc.b	$02,$B7,$39,$C3,$A6,$5F,$0A,$31,$5B,$8C,$61,$E4,$8E,$58,$7E,$0A
	.dc.b	$73,$87,$9A,$73,$87,$9E,$BC,$33,$F3,$0F,$0D,$F3,$71,$1C,$F2,$F8
	.dc.b	$8F,$CE,$BF,$14,$62,$FA,$12,$33,$97,$C3,$C7,$C7,$87,$CC,$9F,$19
	.dc.b	$1F,$14,$6B,$0E,$D1,$8C,$3B,$56,$F2,$FB,$15,$EA,$7E,$D5,$F1,$C7
	.dc.b	$C5,$87,$A9,$F9,$78,$72,$8C,$61,$CF,$2F,$AA,$7C,$1C,$3F,$4F,$87
	.dc.b	$39,$D5,$7C,$B3,$C9,$97,$B7,$0E,$97,$C1,$B8,$73,$25,$7E,$46,$7E
	.dc.b	$DB,$C8,$FC,$31,$98,$DB,$F1,$CF,$39,$B8,$E2,$7C,$09,$EF,$1E,$58
	.dc.b	$F8,$B8,$FC,$0B,$F3,$12,$7B,$9F,$80,$38,$5F,$83,$2F,$98,$FC,$48
	.dc.b	$C9,$0F,$E1,$57,$E2,$8D,$47,$9E,$3E,$15,$77,$75,$1E,$D3,$95,$7E
	.dc.b	$36,$75,$5E,$1B,$FB,$37,$ED,$9C,$DF,$09,$CD,$78,$2F,$F0,$F7,$BA
	.dc.b	$F6,$E5,$F6,$5F,$93,$1F,$13,$2F,$CA,$C6,$B2,$F9,$2F,$A7,$0F,$A3
	.dc.b	$7C,$6F,$D7,$3B,$CB,$C6,$EC,$F2,$61,$D2,$7D,$F1,$F3,$4D,$13,$1F
	.dc.b	$9A,$FC,$33,$19,$7B,$1C,$47,$B3,$1F,$74,$E6,$7E,$8C,$7C,$7B,$FF
	.dc.b	$64,$C7,$EE,$4B,$5F,$BE,$6D,$7F,$34,$77,$DC,$F2,$5E,$57,$29,$FF
	.dc.b	$4B,$46,$1F,$F5,$3B,$5E,$2D,$7F,$A1,$79,$0E,$EE,$51,$FD,$56,$30
	.dc.b	$F5,$C5,$18,$8F,$3B,$5F,$F7,$49,$FE,$F1,$D3,$FD,$D3,$1F,$EF,$89
	.dc.b	$E3,$5F,$C8,$23,$FD,$F3,$A7,$FB,$AB,$CA,$79,$5F,$E5,$E8,$E7,$7F
	.dc.b	$EF,$13,$8E,$3F,$C7,$AF,$8A,$17,$C9,$ED,$84,$78,$3F,$CE,$63,$51
	.dc.b	$82,$63,$FC,$5A,$BD,$07,$68,$F6,$61,$E4,$8C,$DF,$5A,$E0,$F5,$C7
	.dc.b	$D5,$7F,$D7,$EB,$96,$3E,$B6,$9C,$A3,$C3,$87,$D4,$BE,$51,$D6,$3E
	.dc.b	$6E,$3E,$8C,$7E,$6D,$6A,$B8,$9D,$DE,$0F,$C5,$9E,$15,$E9,$9E,$15
	.dc.b	$CA,$BF,$A6,$56,$9E,$47,$43,$E0,$98,$8F,$7B,$98,$E0,$FE,$26,$7C
	.dc.b	$2F,$03,$07,$3C,$3E,$14,$69,$B1,$9C,$3E,$25,$F6,$26,$1F,$87,$C7
	.dc.b	$E6,$D7,$49,$CB,$E5,$9D,$3D,$6B,$ED,$63,$E6,$C3,$F1,$77,$C2,$BC
	.dc.b	$D1,$8B,$F9,$F8,$76,$8C,$4F,$52,$1F,$90,$8D,$61,$F7,$31,$F9,$B7
	.dc.b	$F3,$EB,$E6,$D6,$A7,$C8,$6A,$39,$DF,$67,$C7,$7C,$DE,$27,$03,$E8
	.dc.b	$57,$37,$E4,$57,$37,$E2,$5F,$A9,$E1,$87,$BF,$1F,$85,$87,$58,$FF
	.dc.b	$16,$46,$2F,$77,$F2,$AF,$A9,$E0,$32,$7C,$29,$E0,$EA,$B9,$E1,$E5
	.dc.b	$7E,$76,$3F,$79,$B8,$FD,$EC,$3C,$D7,$C5,$FD,$13,$F0,$0C,$1D,$EF
	.dc.b	$A4,$70,$AF,$CD,$1E,$09,$F2,$9F,$2B,$1B,$8F,$9A,$7C,$31,$F1,$F1
	.dc.b	$FA,$04,$C7,$C9,$7F,$51,$AF,$CC,$3C,$47,$68,$F6,$61,$E3,$3F,$01
	.dc.b	$7E,$39,$F4,$9E,$32,$E1,$E3,$9E,$11,$DF,$0F,$D4,$CE,$63,$07,$C3
	.dc.b	$AF,$7D,$7D,$57,$FC,$C9,$E5,$3C,$9F,$E4,$94,$E5,$71,$63,$CA,$FF
	.dc.b	$57,$4F,$FA,$51,$AE,$E3,$8D,$F3,$8F,$F4,$CB,$5D,$1F,$CC,$67,$8D
	.dc.b	$76,$8C,$75,$F8,$A3,$10,$AE,$51,$87,$A4,$7F,$64,$6B,$E2,$C3,$FA
	.dc.b	$AB,$C5,$B1,$CB,$0D,$4F,$9A,$7C,$11,$EF,$D2,$CF,$F4,$7F,$1D,$E4
	.dc.b	$62,$1E,$53,$CA,$C3,$63,$FD,$73,$5C,$1B,$5A,$A9,$18,$3F,$D7,$FC
	.dc.b	$CF,$2B,$E9,$43,$C8,$78,$DF,$5B,$C9,$79,$5F,$62,$73,$5F,$D7,$67
	.dc.b	$2F,$F5,$F8,$CA,$C3,$B7,$0E,$70,$FC,$0C,$62,$7C,$11,$DF,$0E,$CD
	.dc.b	$8E,$07,$F6,$B9,$E1,$7D,$9B,$3B,$9E,$67,$96,$B9,$BC,$71,$F7,$DF
	.dc.b	$16,$D7,$18,$F4,$E3,$EC,$AE,$67,$AE,$B6,$F3,$C7,$CA,$A7,$85,$72
	.dc.b	$C3,$C3,$3C,$6B,$78,$FA,$23,$CC,$72,$BE,$38,$FC,$6C,$7B,$CF,$13
	.dc.b	$C7,$3A,$78,$9E,$6A,$E8,$EF,$1F,$13,$D7,$1F,$1C,$E8,$DD,$73,$9F
	.dc.b	$54,$EA,$FC,$35,$A8,$F5,$D6,$E3,$94,$F9,$A3,$91,$97,$89,$99,$E2
	.dc.b	$75,$C7,$E1,$C7,$E4,$B6,$79,$5F,$06,$7F,$C5,$51,$8C,$7E,$81,$FE
	.dc.b	$45,$9D,$34,$FF,$2C,$F2,$F0,$8F,$F2,$93,$63,$26,$CF,$E2,$6E,$5A
	.dc.b	$7E,$03,$C6,$79,$5F,$C2,$75,$FC,$EC,$FF,$8F,$6B,$F0,$35,$AE,$3F
	.dc.b	$C1,$CE,$4E,$57,$C7,$7F,$1C,$62,$76,$AF,$9D,$6E,$3C,$F8,$7E,$B3
	.dc.b	$0E,$D1,$F9,$43,$B4,$66,$78,$47,$AA,$31,$7D,$AF,$AB,$67,$C4,$73
	.dc.b	$3C,$46,$27,$DE,$E1,$E6,$EE,$79,$B4,$E1,$87,$BA,$76,$72,$9D,$B6
	.dc.b	$76,$7B,$A7,$FA,$2E,$1E,$E9,$D1,$CA,$3A,$1A,$C3,$DF,$5F,$0A,$31
	.dc.b	$87,$B6,$B8,$4F,$53,$FC,$E9,$7D,$6B,$9C,$62,$FE,$4F,$94,$F2,$BF
	.dc.b	$A4,$6B,$FD,$27,$27,$95,$97,$FC,$FD,$1F,$D2,$AB,$18,$7F,$52,$9D
	.dc.b	$DF,$27,$FB,$0C,$EA,$BA,$5D,$9C,$D6,$EF,$FB,$2C,$E5,$FF,$46,$79
	.dc.b	$2F,$23,$50,$8C,$5F,$90,$F7,$F9,$4E,$CD,$CE,$AB,$FB,$39,$C6,$FF
	.dc.b	$B1,$BF,$DD,$A3,$4F,$FB,$11,$AE,$63,$6F,$91,$F4,$64,$F2,$5E,$57
	.dc.b	$82,$73,$18,$CB,$CD,$7C,$5B,$7C,$6F,$84,$62,$B8,$57,$A3,$C2,$FF
	.dc.b	$80,$0F,$FB,$AC,$9C,$7F,$EE,$8F,$27,$FD,$E1,$E1,$7F,$B7,$67,$F5
	.dc.b	$CF,$FB,$8B,$C9,$FD,$D3,$9A,$FF,$80,$07,$1F,$2B,$FD,$F7,$87,$68
	.dc.b	$C6,$5F,$EF,$88,$FF,$76,$61,$FE,$EE,$F2,$BF,$DD,$91,$FE,$F4,$8F
	.dc.b	$F7,$C3,$A6,$C6,$86,$3F,$EF,$FC,$7C,$47,$F8,$9A,$B9,$65,$EA,$BF
	.dc.b	$F6,$64,$78,$E7,$71,$8B,$E6,$FB,$8F,$25,$FB,$9D,$35,$FF,$80,$26
	.dc.b	$0E,$02,$BB,$39,$C5,$7E,$08,$F5,$D7,$3A,$DC,$71,$BF,$74,$FB,$63
	.dc.b	$95,$71,$C3,$D9,$1B,$9E,$55,$EA,$C3,$E9,$DF,$7C,$3E,$1E,$5F,$52
	.dc.b	$B5,$3F,$1D,$B8,$7C,$03,$D7,$5E,$43,$E9,$63,$E0,$71,$3E,$69,$F8
	.dc.b	$33,$98,$E7,$1F,$18,$F3,$1D,$1C,$CF,$93,$2F,$C2,$47,$3C,$7D,$35
	.dc.b	$F0,$08,$73,$AF,$1E,$3E,$BC,$BE,$09,$F1,$E7,$BD,$7D,$07,$8D,$77
	.dc.b	$7F,$3F,$97,$C5,$C3,$CF,$97,$A7,$0E,$75,$EC,$C7,$B3,$E9,$8E,$D8
	.dc.b	$78,$B2,$F4,$C7,$0C,$BC,$15,$E7,$C7,$D9,$97,$EA,$63,$91,$27,$DE
	.dc.b	$FA,$23,$EF,$63,$E0,$9E,$78,$FA,$5F,$D5,$D7,$A8,$F6,$DF,$CC,$3D
	.dc.b	$75,$C6,$FE,$F6,$1E,$89,$F1,$D7,$DF,$BE,$B1,$C6,$74,$7A,$EB,$8D
	.dc.b	$7C,$3C,$7D,$EF,$A6,$33,$7E,$AC,$7E,$24,$F9,$4F,$61,$E6,$BE,$AF
	.dc.b	$13,$EE,$E5,$F4,$5C,$1F,$A6,$2C,$FC,$1B,$FB,$15,$CA,$FD,$2F,$9E
	.dc.b	$BD,$93,$E8,$7F,$D5,$27,$E4,$8B,$4C,$7F,$80,$C6,$1A,$FF,$9C,$3C
	.dc.b	$97,$95,$CA,$3F,$CD,$91,$87,$1E,$02,$31,$1E,$08,$C4,$F1,$8F,$E0
	.dc.b	$DE,$4B,$BB,$18,$A3,$13,$DE,$31,$3A,$9C,$CF,$A9,$FF,$80,$19,$7F
	.dc.b	$B2,$4F,$F6,$B9,$63,$FD,$9F,$A7,$E2,$A7,$FD,$C3,$7F,$EE,$5B,$FF
	.dc.b	$74,$47,$2A,$0F,$27,$FD,$B0,$FF,$B9,$CF,$F7,$05,$FF,$BC,$1F,$F6
	.dc.b	$D6,$1E,$07,$8C,$F0,$C3,$B4,$68,$F1,$CF,$F3,$7B,$F1,$9F,$02,$FC
	.dc.b	$D8,$79,$0F,$F7,$FF,$95,$FE,$50,$AE,$D7,$E5,$8C,$3D,$0F,$F7,$D9
	.dc.b	$FE,$F6,$C3,$FD,$F6,$E4,$E0,$7F,$BF,$F0,$F0,$DF,$FB,$DA,$B9,$CF
	.dc.b	$BE,$BD,$38,$7F,$C0,$13,$FD,$EF,$1E,$E9,$DC,$EB,$0F,$65,$FA,$6B
	.dc.b	$F9,$5E,$1E,$D9,$CC,$7C,$33,$B4,$65,$F4,$E1,$D9,$F3,$9E,$B8,$F3
	.dc.b	$BD,$2B,$B4,$FD,$33,$CD,$5F,$1B,$1F,$A5,$7D,$6F,$EC,$CF,$8A,$BE
	.dc.b	$7C,$78,$F0,$F5,$3E,$29,$E5,$8F,$BA,$FF,$0D,$3E,$4B,$F3,$61,$DB
	.dc.b	$1F,$95,$3F,$35,$F2,$E1,$DC,$91,$F2,$4D,$1E,$1B,$EB,$5C,$E3,$DC
	.dc.b	$66,$BF,$06,$F0,$BF,$7B,$F7,$71,$FB,$EF,$CC,$9F,$0B,$F8,$1A,$E9
	.dc.b	$5F,$00,$E3,$1B,$3B,$B8,$AF,$87,$1F,$20,$E5,$8F,$C3,$BF,$1C,$7D
	.dc.b	$57,$BB,$CE,$B9,$61,$EC,$AE,$13,$DB,$1F,$86,$FC,$B7,$E7,$BE,$D7
	.dc.b	$F7,$4F,$37,$78,$FA,$9B,$3D,$2F,$F0,$51,$C3,$1F,$49,$F1,$0F,$5D
	.dc.b	$71,$BF,$BC,$7C,$9C,$3F,$1B,$7F,$16,$F9,$47,$A6,$7C,$87,$D9,$3F
	.dc.b	$C4,$3E,$4B,$C9,$FE,$13,$1F,$E6,$6F,$2B,$FC,$D4,$FF,$96,$03,$87
	.dc.b	$FC,$D7,$3C,$5D,$35,$E5,$D5,$D3,$A1,$CB,$DA,$F9,$46,$2B,$F8,$CD
	.dc.b	$F1,$8C,$42,$31,$5C,$6F,$A3,$67,$C3,$86,$1E,$6F,$F3,$97,$93,$6B
	.dc.b	$CD,$1F,$D3,$9A,$FC,$2F,$25,$E4,$F3,$87,$94,$F2,$B0,$D8,$F8,$D5
	.dc.b	$C1,$B5,$AA,$91,$83,$E6,$72,$BC,$86,$BF,$CE,$BC,$77,$8D,$F5,$3C
	.dc.b	$A7,$95,$F5,$A7,$38,$FF,$40,$9C,$BC,$6B,$ED,$E3,$D7,$0C,$4F,$F4
	.dc.b	$68,$C2,$BE,$6A,$F9,$B9,$BF,$16,$1E,$17,$AE,$3E,$89,$D1,$D5,$C5
	.dc.b	$F4,$BE,$6D,$BE,$55,$BC,$7C,$AB,$1F,$2D,$6E,$33,$7C,$B1,$F0,$DF
	.dc.b	$0B,$DC,$79,$AB,$78,$76,$7A,$63,$E4,$9E,$31,$E8,$3E,$04,$EF,$0E
	.dc.b	$D1,$E5,$35,$7D,$9F,$A1,$3B,$3B,$46,$63,$BE,$1E,$26,$1C,$B0,$F1
	.dc.b	$3D,$71,$F7,$3E,$3C,$3C,$2F,$23,$85,$72,$9D,$9F,$62,$FC,$07,$86
	.dc.b	$B8,$9C,$9C,$DF,$82,$FD,$E7,$3C,$7E,$1D,$F3,$9F,$B9,$D2,$F2,$BE
	.dc.b	$3B,$6F,$E2,$35,$F3,$35,$FF,$09,$F3,$FF,$85,$23,$10,$8C,$4F,$F8
	.dc.b	$D5,$B1,$FE,$40,$6B,$FC,$51,$FC,$2F,$94,$F2,$BF,$84,$63,$F9,$8A
	.dc.b	$EE,$7D,$B9,$D7,$2F,$E3,$23,$33,$F9,$E9,$CC,$7E,$04,$CE,$7F,$C8
	.dc.b	$A3,$33,$D6,$7D,$07,$5B,$F0,$D7,$A6,$31,$7D,$DD,$D7,$7A,$DC,$7B
	.dc.b	$4D,$9E,$13,$75,$E1,$8D,$C7,$C1,$37,$3E,$3C,$3E,$1C,$EA,$7D,$27
	.dc.b	$F9,$0B,$0F,$4D,$F9,$6F,$C5,$87,$97,$0E,$2D,$BE,$B5,$B3,$C5,$19
	.dc.b	$AE,$33,$E7,$AE,$2D,$38,$9E,$D8,$D9,$E7,$72,$7A,$DC,$D7,$78,$F4
	.dc.b	$61,$EB,$7E,$19,$8B,$FE,$0B,$E5,$3C,$AF,$E9,$1C,$1F,$D2,$65,$E5
	.dc.b	$65,$FE,$B1,$56,$FF,$96,$C7,$F9,$98,$DD,$F2,$7F,$B0,$CE,$EF,$A5
	.dc.b	$D9,$CD,$6E,$FF,$AD,$CE,$5F,$F3,$8F,$92,$F2,$7F,$AC,$C7,$F6,$C8
	.dc.b	$C5,$FF,$A2,$DB,$1F,$E9,$56,$BF,$EA,$07,$38,$7B,$BC,$87,$93,$E1
	.dc.b	$A5,$6A,$FF,$D0,$36,$BD,$C7,$F7,$99,$D5,$7F,$4E,$9C,$BF,$10,$F4
	.dc.b	$E1,$ED,$BE,$58,$77,$BE,$31,$8C,$3A,$56,$F0,$E7,$3F,$DE,$3C,$2F
	.dc.b	$F5,$B7,$8F,$FE,$BD,$CB,$FD,$89,$C3,$FE,$C0,$7F,$D8,$FE,$47,$FB
	.dc.b	$26,$30,$7F,$B4,$FC,$8F,$F6,$FC,$F0,$F1,$3F,$DD,$BA,$7F,$B6,$AB
	.dc.b	$71,$FE,$DA,$3F,$DE,$CE,$27,$FD,$AB,$E4,$FF,$BA,$DB,$5C,$6B,$9C
	.dc.b	$7F,$B7,$F2,$FF,$70,$DF,$FB,$D8,$93,$FE,$D8,$9F,$F7,$51,$BC,$3C
	.dc.b	$B8,$FA,$67,$C5,$3F,$ED,$A3,$FD,$D0,$70,$AE,$6E,$1F,$33,$ED,$31
	.dc.b	$1E,$A3,$CB,$1C,$23,$FD,$E6,$4B,$FF,$76,$3C,$0F,$69,$E5,$CB,$E8
	.dc.b	$46,$A3,$86,$09,$FA,$98,$7A,$67,$8D,$F7,$BF,$87,$87,$C1,$8F,$A9
	.dc.b	$7D,$30,$E5,$7E,$DA,$D1,$E9,$7C,$87,$B4,$F0,$CF,$FC,$01,$8F,$69
	.dc.b	$C2,$3E,$41,$A3,$A1,$CA,$FD,$47,$91,$C5,$78,$AF,$CB,$1F,$D3,$6B
	.dc.b	$AB,$AB,$F0,$3E,$4C,$3E,$03,$F3,$1B,$87,$6C,$3E,$EE,$5F,$89,$AD
	.dc.b	$36,$75,$7F,$0C,$EB,$3E,$9C,$7C,$67,$DE,$9D,$C7,$94,$F1,$57,$3B
	.dc.b	$F5,$63,$F2,$67,$93,$6B,$F5,$71,$FA,$8A,$E1,$87,$82,$73,$87,$E3
	.dc.b	$31,$F5,$E3,$D3,$0F,$9B,$5A,$C7,$F2,$F1,$F5,$63,$0F,$E0,$A3,$EA
	.dc.b	$39,$9E,$4F,$C2,$CB,$E0,$3F,$E8,$80,$2B,$CE,$E5,$AF,$FA,$39,$AF
	.dc.b	$F9,$4F,$C9,$79,$3C,$E3,$0D,$7F,$C8,$91,$A7,$5D,$88,$C4,$77,$F1
	.dc.b	$DD,$9C,$70,$46,$27,$B4,$62,$75,$39,$9E,$51,$FE,$C6,$D7,$FD,$CF
	.dc.b	$E4,$FF,$AD,$F5,$FF,$60,$E1,$FE,$B5,$C3,$FD,$EC,$DB,$FF,$75,$E1
	.dc.b	$FE,$CD,$3F,$DC,$47,$FB,$24,$FF,$5C,$9F,$EC,$52,$C7,$FB,$02,$7F
	.dc.b	$D7,$75,$FE,$C6,$9F,$F5,$C9,$FE,$CF,$3F,$D8,$15,$FE,$D8,$AF,$F6
	.dc.b	$2D,$F9,$67,$FD,$BE,$70,$F2,$7F,$D9,$0F,$FB,$8C,$F0,$1C,$CB,$3E
	.dc.b	$97,$CC,$65,$F2,$57,$7C,$3D,$E7,$9A,$7D,$2F,$FB,$3D,$F8,$18,$FF
	.dc.b	$BA,$A7,$E5,$63,$FE,$E6,$9C,$BF,$ED,$68,$CC,$7B,$0F,$8F,$3F,$EE
	.dc.b	$47,$FD,$B0,$FD,$27,$D9,$8F,$BD,$F5,$E3,$EE,$C7,$D5,$5F,$EE,$39
	.dc.b	$E3,$87,$86,$39,$47,$88,$C9,$E1,$7D,$0F,$C2,$9E,$73,$F4,$E3,$81
	.dc.b	$F1,$71,$F8,$15,$D0,$93,$DC,$93,$D5,$F8,$0F,$59,$F9,$4F,$D5,$AF
	.dc.b	$B6,$FB,$88,$78,$2F,$DD,$3C,$27,$CB,$3F,$4B,$1F,$96,$FD,$C9,$D1
	.dc.b	$23,$ED,$1F,$96,$C3,$CF,$3E,$53,$E1,$9C,$63,$6F,$E1,$2D,$8F,$F9
	.dc.b	$D7,$1F,$E0,$32,$F9,$67,$CC,$7F,$BC,$5B,$1F,$BB,$6B,$FB,$F6,$BF
	.dc.b	$C2,$DC,$F2,$38,$3A,$C6,$1B,$5F,$EF,$FE,$0F,$EB,$9C,$3F,$EF,$7A
	.dc.b	$D6,$1F,$C5,$1A,$F9,$6B,$87,$1A,$B8,$44,$9F,$F1,$5E,$1D,$9A,$7F
	.dc.b	$16,$AE,$0E,$07,$90,$F2,$7F,$8C,$46,$1B,$19,$8F,$E3,$8D,$7F,$90
	.dc.b	$37,$3F,$81,$1F,$CB,$23,$17,$F3,$9B,$1F,$CC,$3C,$BF,$E3,$15,$C2
	.dc.b	$3F,$D9,$1E,$0F,$FB,$1B,$C1,$FF,$72,$47,$FB,$28,$9E,$1F,$FB,$40
	.dc.b	$91,$FE,$E4,$7F,$D8,$55,$FE,$C5,$9F,$F6,$F9,$FE,$EF,$F1,$3F,$D9
	.dc.b	$5D,$3F,$EC,$FF,$1B,$FD,$C5,$39,$A8,$FF,$B2,$70,$F0,$E1,$DA,$3F
	.dc.b	$DD,$8F,$67,$FD,$C1,$5C,$9F,$36,$39,$3D,$4F,$88,$CD,$7F,$B6,$0F
	.dc.b	$6C,$6A,$3E,$0C,$7A,$CF,$09,$1F,$0B,$C5,$E9,$1C,$70,$FF,$73,$E1
	.dc.b	$FD,$73,$0F,$45,$7F,$BA,$0F,$61,$C8,$86,$5D,$CB,$15,$73,$C3,$FD
	.dc.b	$D7,$52,$FF,$DD,$6D,$7F,$D8,$1C,$DF,$EE,$8E,$7F,$F7,$1C,$7F,$B7
	.dc.b	$A3,$0F,$F5,$E9,$FF,$69,$1F,$EB,$63,$FD,$60,$5A,$C9,$FE,$BC,$8F
	.dc.b	$49,$FE,$CD,$8D,$3F,$EB,$98,$D6,$5F,$EB,$93,$FD,$6F,$1F,$ED,$99
	.dc.b	$DB,$C6,$7A,$46,$DE,$27,$FA,$F6,$1E,$2F,$FB,$51,$EC,$7F,$AF,$4F
	.dc.b	$F6,$55,$76,$7F,$D8,$D3,$C9,$D9,$CB,$C5,$FF,$63,$63,$F2,$E3,$96
	.dc.b	$1F,$02,$BC,$67,$87,$0F,$04,$F4,$3C,$18,$79,$DB,$3B,$C3,$AC,$7C
	.dc.b	$6A,$47,$ED,$2F,$B4,$F8,$2B,$87,$07,$FB,$9D,$F4,$47,$FA,$DB,$C0
	.dc.b	$FF,$55,$1F,$EC,$9C,$1E,$63,$CC,$CE,$1F,$EA,$C3,$FD,$8D,$5F,$EC
	.dc.b	$AB,$FF,$5A,$C6,$0F,$F1,$27,$DD,$BE,$2F,$FD,$88,$D8,$FE,$2F,$D7
	.dc.b	$FE,$2C,$EC,$CD,$49,$FE,$31,$5B,$BF,$F6,$15,$EE,$73,$5B,$9F,$F6
	.dc.b	$30,$AE,$1C,$5F,$C5,$AF,$84,$EA,$73,$7F,$D1,$1E,$D1,$FE,$47,$6B
	.dc.b	$9B,$F7,$3D,$5B,$E0,$7F,$95,$23,$FC,$A0,$7F,$A7,$4F,$F3,$51,$FE
	.dc.b	$66,$3F,$CB,$BD,$9F,$E5,$68,$FF,$37,$F1,$FF,$9C,$63,$FC,$B3,$E1
	.dc.b	$7F,$99,$3B,$3F,$CD,$11,$B3,$3E,$1F,$F9,$B3,$0F,$F4,$89,$D4,$F0
	.dc.b	$DF,$FA,$96,$7F,$D0,$B9,$FF,$B2,$E7,$FC,$F6,$D7,$86,$1D,$DE,$78
	.dc.b	$FF,$A6,$5E,$8F,$9B,$6F,$F4,$BC,$62,$FC,$77,$CE,$BE,$11,$B9,$EE
	.dc.b	$7C,$3C,$3D,$D9,$7B,$0F,$F5,$EE,$5E,$B8,$FF,$49,$E3,$DA,$FC,$44
	.dc.b	$C7,$D7,$87,$39,$D4,$62,$BE,$06,$1F,$E8,$07,$9D,$78,$9C,$BF,$52
	.dc.b	$FC,$F5,$EA,$8E,$86,$EF,$FD,$1B,$97,$C5,$BE,$73,$F2,$31,$F4,$57
	.dc.b	$FA,$86,$FE,$B6,$3E,$DA,$FF,$51,$65,$F1,$67,$31,$F3,$9F,$45,$76
	.dc.b	$BE,$73,$F8,$6B,$F8,$B7,$C5,$B1,$E6,$3C,$51,$97,$B9,$9C,$BD,$B3
	.dc.b	$E3,$C7,$FD,$61,$7D,$AB,$FD,$51,$3D,$27,$E2,$9F,$5A,$3C,$D1,$FE
	.dc.b	$A8,$86,$1F,$EA,$5C,$7F,$DD,$B5,$FE,$A9,$6B,$9A,$FF,$4E,$46,$23
	.dc.b	$FD,$4E,$7F,$A9,$3C,$3F,$F5,$1B,$FE,$34,$EA,$FF,$53,$6F,$FE,$AB
	.dc.b	$6C,$7F,$A8,$BC,$3F,$F5,$03,$FE,$B5,$6B,$FE,$BC,$D3,$FD,$4C,$EF
	.dc.b	$AB,$FD,$4E,$7F,$93,$A3,$FD,$0B,$1F,$E7,$8C,$FF,$D4,$B5,$FE,$4F
	.dc.b	$8C,$1F,$E9,$43,$FC,$B5,$E1,$7F,$9E,$6B,$FC,$BF,$18,$3F,$CB,$FA
	.dc.b	$FF,$95,$1F,$F2,$91,$C8,$FF,$3F,$CF,$1A,$FF,$49,$E1,$FE,$56,$AF
	.dc.b	$F4,$05,$7F,$9D,$63,$CD,$3F,$E7,$6C,$3F,$CE,$73,$D2,$BF,$CF,$E7
	.dc.b	$FA,$3D,$E0,$7F,$9D,$6B,$A4,$FF,$A0,$67,$C9,$87,$C0,$78,$C7,$C0
	.dc.b	$9D,$D7,$F9,$CC,$E4,$7A,$6B,$DF,$5C,$AF,$C9,$3F,$E7,$13,$C3,$1E
	.dc.b	$3F,$0F,$FC,$FF,$7D,$27,$FD,$0D,$5F,$01,$E8,$DA,$F1,$C7,$CF,$7B
	.dc.b	$4F,$C2,$24,$79,$23,$6F,$CC,$3E,$8E,$1F,$1B,$0F,$73,$FE,$98,$9F
	.dc.b	$26,$3F,$46,$FF,$D1,$53,$BC,$3D,$78,$7C,$D8,$E0,$7C,$57,$CF,$3C
	.dc.b	$1F,$81,$8F,$B9,$C9,$FE,$8F,$BE,$CD,$3C,$A7,$07,$38,$7E,$02,$79
	.dc.b	$3E,$CC,$7E,$7E,$1E,$5C,$3E,$0C,$FA,$B0,$E6,$79,$CF,$F5,$0D,$7C
	.dc.b	$07,$FD,$47,$E1,$FF,$A3,$34,$7B,$2F,$AA,$FF,$49,$36,$BF,$D0,$95
	.dc.b	$FE,$90,$3F,$D3,$AF,$FA,$2C,$FF,$50,$78,$BF,$EA,$52,$3C,$1F,$F4
	.dc.b	$93,$FD,$D3,$8B,$FD,$1E,$FF,$72,$9E,$93,$AE,$8F,$EE,$55,$A7,$FD
	.dc.b	$05,$E2,$7F,$A1,$AD,$E2,$7F,$83,$DF,$F4,$C3,$5F,$EF,$B1,$B3,$FD
	.dc.b	$28,$FF,$AF,$0F,$F4,$D1,$FE,$98,$F1,$7F,$D3,$67,$FA,$7B,$93,$FD
	.dc.b	$05,$1F,$E8,$73,$FC,$E2,$0A,$7F,$9A,$7C,$3F,$F3,$77,$89,$FE,$8A
	.dc.b	$3F,$CC,$B1,$FE,$54,$F3,$1E,$37,$FB,$33,$C5,$FF,$35,$D6,$9F,$F4
	.dc.b	$27,$8B,$FE,$80,$25,$F7,$C7,$FC,$F2,$FF,$A3,$0F,$23,$8C,$7F,$D0
	.dc.b	$D8,$F6,$9F,$F4,$BC,$74,$7F,$D0,$75,$D1,$F4,$D7,$B9,$E8,$7F,$A4
	.dc.b	$8F,$06,$1E,$AA,$EB,$8F,$FB,$91,$E9,$3F,$0B,$0F,$03,$CA,$33,$7D
	.dc.b	$5F,$8B,$AF,$93,$1F,$15,$F0,$9D,$E3,$F2,$39,$99,$FF,$A2,$71,$FF
	.dc.b	$48,$63,$F4,$67,$39,$7F,$A3,$8F,$B9,$CD,$FE,$8C,$4D,$23,$FE,$8D
	.dc.b	$3F,$CE,$3E,$2F,$F9,$A9,$FE,$B0,$32,$FF,$3C,$38,$6B,$D0,$D6,$9F
	.dc.b	$E6,$47,$4D,$75,$1C,$2B,$F1,$0D,$8E,$27,$F9,$D8,$F1,$F8,$DF,$E7
	.dc.b	$68,$D1,$2B,$A5,$72,$6D,$72,$3F,$D1,$5E,$07,$F9,$DD,$E8,$7F,$A1
	.dc.b	$4C,$9F,$E9,$48,$E3,$5E,$3D,$3F,$D4,$33,$F0,$9A,$FA,$A7,$64,$7A
	.dc.b	$9F,$00,$EF,$E4,$3C,$5F,$F3,$0B,$70,$FF,$3F,$4F,$EC,$4F,$F4,$53
	.dc.b	$8D,$4F,$EB,$57,$90,$F2,$7F,$D1,$75,$FE,$D0,$AD,$4E,$5F,$F3,$E9
	.dc.b	$23,$FD,$03,$D9,$FE,$12,$8C,$5F,$69,$FD,$D7,$8F,$FE,$7C,$6F,$8F
	.dc.b	$99,$FF,$41,$5F,$19,$FF,$41,$4F,$FA,$25,$FF,$40,$3F,$E7,$EA,$FF
	.dc.b	$40,$C7,$2B,$FF,$4D,$CF,$37,$F8,$45,$7F,$BA,$75,$FF,$49,$61,$FE
	.dc.b	$8F,$E4,$FC,$FD,$7C,$13,$FC,$AF,$D5,$FE,$55,$F0,$3F,$C9,$0F,$F8
	.dc.b	$C5,$FF,$21,$47,$F8,$EF,$8B,$FC,$6E,$D8,$FF,$16,$67,$FE,$34,$3F
	.dc.b	$CB,$5E,$2F,$F9,$32,$BF,$CA,$BE,$0F,$F9,$02,$75,$1F,$E3,$B3,$33
	.dc.b	$BA,$F0,$CF,$F9,$22,$7F,$C8,$31,$E6,$8E,$B8,$79,$5F,$F1,$75,$7F
	.dc.b	$96,$1B,$1E,$B7,$FC,$B0,$73,$9F,$5B,$FE,$53,$9F,$03,$AA,$F6,$B9
	.dc.b	$AE,$CF,$FA,$35,$FF,$2C,$57,$C0,$9F,$4C,$6A,$FF,$D0,$8F,$86,$BF
	.dc.b	$CB,$67,$84,$F2,$9E,$A3,$07,$53,$E2,$D7,$BE,$BE,$33,$F1,$1F,$29
	.dc.b	$F1,$23,$FD,$08,$74,$76,$FA,$E3,$CC,$71,$2B,$93,$83,$96,$99,$3F
	.dc.b	$D7,$2F,$C8,$8F,$04,$7F,$9F,$8F,$8C,$7A,$AB,$E0,$BE,$D8,$F3,$CF
	.dc.b	$12,$47,$63,$94,$74,$9F,$C3,$3F,$32,$3A,$5F,$DD,$7E,$34,$64,$EA
	.dc.b	$F3,$9E,$35,$EB,$8F,$49,$9A,$FB,$EF,$D4,$AE,$2F,$49,$E1,$5F,$EC
	.dc.b	$78,$2A,$FF,$44,$46,$1B,$19,$BF,$F4,$69,$FE,$87,$F1,$7F,$D1,$75
	.dc.b	$FE,$8E,$3F,$C0,$9D,$3F,$E8,$A6,$C6,$9E,$FE,$2F,$FA,$29,$BD,$5F
	.dc.b	$E8,$87,$FD,$14,$4C,$FF,$C6,$4F,$FA,$04,$8F,$F9,$61,$FF,$2B,$67
	.dc.b	$FE,$4A,$7F,$CB,$8F,$F8,$ED,$C9,$FE,$38,$AF,$F2,$8E,$5F,$E3,$98
	.dc.b	$FF,$35,$CF,$F9,$46,$BF,$C8,$95,$FE,$67,$8E,$93,$FE,$61,$3C,$2F
	.dc.b	$59,$F2,$9E,$32,$B9,$3A,$1C,$AB,$FC,$9F,$3F,$E5,$53,$FC,$9A,$F8
	.dc.b	$FC,$4F,$F2,$AC,$EE,$FC,$55,$E9,$C3,$FC,$AB,$3D,$23,$07,$B8,$F5
	.dc.b	$9F,$08,$EC,$7A,$0B,$87,$9E,$39,$47,$03,$DD,$19,$79,$1F,$E6,$13
	.dc.b	$58,$7A,$9C,$61,$E8,$7C,$73,$F0,$9E,$F5,$F0,$DE,$75,$EF,$F1,$7F
	.dc.b	$CE,$B3,$FE,$A9,$35,$3E,$26,$CF,$73,$84,$7C,$8C,$7E,$D9,$31,$F8
	.dc.b	$B1,$DA,$30,$FC,$5C,$7C,$CF,$AA,$3E,$6B,$67,$FD,$20,$FB,$30,$F2
	.dc.b	$BF,$EB,$7C,$7D,$8E,$4F,$14,$F8,$1F,$41,$F0,$08,$7F,$99,$8F,$EF
	.dc.b	$1E,$07,$F9,$A9,$FF,$31,$19,$3F,$CD,$8E,$5F,$F3,$33,$FE,$71,$F0
	.dc.b	$37,$C5,$FE,$7A,$6F,$0F,$2F,$05,$86,$30,$E3,$55,$F1,$5F,$17,$FC
	.dc.b	$FF,$39,$E2,$7D,$3F,$F8,$36,$7F,$CE,$D1,$87,$0D,$7F,$C2,$4D,$7F
	.dc.b	$D0,$8E,$6B,$FC,$0D,$1F,$C3,$67,$B1,$E3,$F9,$07,$91,$89,$FE,$F0
	.dc.b	$78,$27,$0F,$F8,$5C,$FF,$44,$4F,$07,$67,$52,$DF,$87,$3F,$9A,$D8
	.dc.b	$E3,$7F,$0F,$F0,$D7,$27,$C5,$E7,$7C,$AD,$35,$12,$FB,$D7,$06,$C7
	.dc.b	$1F,$4F,$51,$E7,$13,$B7,$0D,$D7,$E2,$E0,$AF,$AD,$F8,$63,$D9,$A4
	.dc.b	$C3,$DF,$EB,$3E,$4F,$7F,$55,$27,$E1,$DA,$3F,$01,$4F,$67,$F0,$5D
	.dc.b	$DF,$E8,$2B,$FF,$3D,$07,$1C,$BF,$E8,$48,$C6,$9F,$E8,$BF,$03,$F2
	.dc.b	$F5,$A7,$FC,$A5,$D5,$FE,$84,$BE,$17,$FE,$82,$6B,$FE,$7F,$7A,$79
	.dc.b	$6F,$84,$FF,$1E,$4F,$FB,$0E,$31,$CF,$F8,$2F,$09,$FA,$A7,$E5,$23
	.dc.b	$EF,$CE,$AB,$E7,$2F,$73,$9C,$23,$13,$CF,$A5,$F1,$1A,$9C,$D7,$D7
	.dc.b	$FC,$35,$F5,$5C,$9B,$87,$8F,$F0,$27,$CE,$66,$A4,$FE,$4A,$B7,$7F
	.dc.b	$97,$D5,$87,$0A,$D5,$F0,$AF,$CE,$7E,$1C,$FC,$39,$F4,$B9,$EE,$FF
	.dc.b	$37,$4E,$63,$DB,$F7,$AF,$AD,$A7,$CA,$FC,$01,$F8,$07,$E6,$B4,$FD
	.dc.b	$2D,$6D,$B5,$C1,$FD,$7D,$70,$F1,$71,$A7,$E6,$6F,$83,$6F,$B6,$1D
	.dc.b	$63,$18,$74,$3F,$84,$57,$08,$FE,$3A,$1C,$3F,$C8,$1A,$8F,$F4,$D9
	.dc.b	$FA,$9F,$5B,$FD,$40,$FF,$01,$E3,$7C,$1B,$E0,$FF,$D3,$7B,$FF,$0E
	.dc.b	$BD,$B7,$D9,$FC,$6D,$F0,$AD,$E5,$FC,$C2,$FA,$E3,$E6,$6B,$9B,$FF
	.dc.b	$58,$9F,$C6,$9F,$F2,$C7,$A9,$1F,$F2,$A9,$FE,$2D,$3F,$C9,$9E,$A7
	.dc.b	$F8,$B4,$FF,$1B,$B8,$F3,$1C,$FF,$E6,$E8,$FF,$23,$7C,$1F,$F2,$2C
	.dc.b	$BF,$D2,$11,$8E,$CF,$F5,$77,$83,$FE,$49,$C3,$FC,$7B,$7B,$AF,$F2
	.dc.b	$AC,$FF,$8F,$B1,$FF,$2A,$BF,$E6,$68,$FF,$21,$65,$FE,$64,$69,$D7
	.dc.b	$1E,$EF,$F9,$28,$F5,$1A,$31,$5D,$27,$E0,$47,$F9,$71,$FF,$2F,$BC
	.dc.b	$2F,$C9,$7E,$7C,$BF,$D1,$18,$FF,$97,$70,$E5,$7F,$21,$AF,$33,$FC
	.dc.b	$B5,$3C,$0F,$F4,$75,$F4,$C3,$BD,$70,$3E,$6D,$79,$DF,$A1,$7F,$1B
	.dc.b	$2F,$93,$5C,$23,$C4,$4C,$3A,$CE,$AF,$C2,$7F,$9D,$23,$39,$7C,$6C
	.dc.b	$3A,$C7,$96,$7F,$D0,$58,$7C,$DC,$7C,$D1,$CB,$2F,$A9,$97,$78,$D9
	.dc.b	$F4,$4F,$B3,$5C,$72,$F1,$47,$CB,$6E,$3E,$FA,$F6,$65,$E3,$7F,$17
	.dc.b	$5F,$EA,$C3,$E6,$3D,$F0,$FC,$16,$3E,$49,$E9,$5C,$72,$FC,$94,$7D
	.dc.b	$EB,$E1,$97,$DE,$C7,$E6,$E1,$C5,$F1,$5F,$FA,$0A,$FF,$7A,$7F,$A1
	.dc.b	$DC,$3F,$E8,$0F,$5B,$FD,$0C,$47,$7E,$27,$FA,$15,$FF,$10,$D3,$D5
	.dc.b	$FF,$42,$37,$C4,$FF,$42,$4F,$F0,$A8,$C3,$FE,$85,$CB,$FC,$B3,$EB
	.dc.b	$7F,$8F,$09,$EB,$7F,$8E,$27,$FD,$1A,$37,$FF,$3E,$E1,$FE,$3E,$C3
	.dc.b	$FC,$97,$3F,$E6,$53,$FC,$78,$57,$33,$FE,$63,$8F,$F1,$E9,$A7,$11
	.dc.b	$FE,$55,$C7,$FC,$9F,$1F,$E4,$68,$E3,$8F,$F9,$6A,$7F,$C9,$F8,$F9
	.dc.b	$AF,$FC,$92,$7F,$95,$71,$F4,$5F,$BB,$0F,$3C,$FA,$70,$FF,$33,$3F
	.dc.b	$E4,$C9,$EA,$DC,$3C,$33,$F1,$2F,$9C,$7F,$91,$27,$91,$FE,$76,$9D
	.dc.b	$3E,$2B,$F4,$4F,$28,$FF,$35,$63,$FE,$5D,$C3,$DC,$D9,$F5,$E1,$DA
	.dc.b	$3F,$9D,$5F,$A0,$61,$E8,$C3,$E4,$3F,$0E,$7E,$95,$F7,$9F,$9B,$87
	.dc.b	$B1,$D9,$ED,$9F,$01,$C0,$AF,$B0,$FB,$F7,$E1,$BF,$05,$73,$25,$F7
	.dc.b	$7D,$75,$F4,$4F,$85,$3F,$11,$FC,$14,$EC,$F7,$BA,$3E,$29,$C4,$DD
	.dc.b	$F9,$2F,$E0,$C7,$FA,$26,$BA,$46,$23,$FC,$E7,$BF,$F9,$E7,$8F,$FC
	.dc.b	$ED,$EA,$FF,$5D,$9F,$F2,$EF,$AD,$FE,$81,$7F,$CE,$8F,$FA,$15,$FF
	.dc.b	$44,$1F,$E8,$A7,$06,$DF,$F2,$92,$F5,$BA,$E5,$FE,$6C,$E8,$FF,$3A
	.dc.b	$FA,$AF,$A6,$FF,$3A,$19,$CF,$FC,$D3,$EB,$FF,$A0,$F2,$FF,$37,$1F
	.dc.b	$E8,$07,$1C,$FF,$E8,$18,$D9,$FE,$72,$8E,$4F,$F9,$C0,$C1,$0F,$F3
	.dc.b	$34,$7F,$9D,$55,$FF,$9F,$5F,$F3,$3F,$AF,$FE,$73,$7F,$CF,$FE,$7F
	.dc.b	$F9,$D9,$FF,$38,$90,$FE,$27,$EB,$FF,$9B,$DF,$F0,$8B,$5F,$F5,$7B
	.dc.b	$5F,$F0,$83,$FE,$73,$C7,$35,$FE,$7A,$CF,$FC,$EB,$EB,$FF,$9C,$6F
	.dc.b	$F8,$FF,$AB,$FE,$75,$7F,$D4,$CD,$73,$EB,$FC,$78,$FF,$39,$46,$BD
	.dc.b	$6C,$11,$FF,$57,$75,$FF,$9A,$DF,$F1,$13,$5F,$F3,$8E,$1F,$E2,$5E
	.dc.b	$1F,$F3,$A4,$E6,$7F,$CD,$DE,$A7,$F9,$C9,$FF,$3D,$98,$3F,$C6,$3C
	.dc.b	$FC,$63,$FD,$9F,$E1,$FF,$9A,$67,$FC,$6A,$D7,$3E,$A7,$F8,$7D,$FF
	.dc.b	$3A,$35,$FF,$36,$7A,$76,$7F,$CD,$31,$CA,$30,$6F,$D1,$FF,$31,$E7
	.dc.b	$FE,$C2,$6F,$0F,$F8,$FF,$C2,$FF,$32,$FA,$5F,$E5,$F8,$FF,$71,$CF
	.dc.b	$F9,$AB,$0F,$F2,$4B,$5C,$FA,$3D,$49,$3F,$EC,$8F,$43,$FC,$CC,$FF
	.dc.b	$9B,$8C,$7A,$3F,$E6,$C7,$FC,$AD,$5F,$E5,$83,$FD,$99,$E8,$7F,$98
	.dc.b	$63,$39,$FF,$B6,$3C,$DF,$F3,$0B,$7C,$FF,$F3,$0F,$A1,$FE,$66,$8C
	.dc.b	$57,$F9,$82,$7E,$35,$7F,$97,$30,$FF,$30,$46,$0F,$F2,$9F,$DF,$BE
	.dc.b	$2B,$95,$FF,$A8,$63,$0F,$FB,$73,$D0,$FF,$2D,$BF,$E5,$33,$FD,$CB
	.dc.b	$C7,$FE,$5B,$75,$E8,$7F,$A8,$27,$FC,$B3,$8F,$F9,$AD,$AF,$F9,$B0
	.dc.b	$FF,$2F,$47,$F9,$72,$7F,$DD,$F9,$BD,$C7,$BB,$FE,$66,$B4,$E6,$3F
	.dc.b	$D5,$BE,$6F,$F8,$13,$D2,$FF,$0A,$F8,$DF,$E0,$8B,$FF,$08,$F4,$7F
	.dc.b	$8A,$1F,$F0,$E9,$FE,$1F,$EC,$FF,$22,$3F,$E2,$D3,$FC,$46,$7F,$84
	.dc.b	$63,$84,$68,$FF,$14,$B8,$F3,$7F,$C7,$4D,$9F,$F1,$74,$7F,$8A,$74
	.dc.b	$FF,$0E,$47,$F8,$8C,$FF,$26,$1F,$E2,$03,$C6,$77,$3C,$38,$FF,$8F
	.dc.b	$1F,$25,$FF,$8A,$31,$F6,$E1,$DE,$79,$4F,$97,$0F,$04,$74,$BE,$53
	.dc.b	$E8,$8E,$B8,$FF,$8D,$A3,$FA,$07,$A5,$FE,$2B,$8F,$91,$3E,$6A,$FF
	.dc.b	$16,$BE,$2B,$FF,$19,$4F,$F9,$02,$3B,$3C,$E3,$FD,$51,$1F,$E6,$18
	.dc.b	$C3,$FE,$8F,$C3,$FC,$7B,$5E,$4A,$F3,$E1,$E9,$AF,$A9,$1F,$31,$F3
	.dc.b	$D7,$0B,$E7,$1D,$63,$04,$7E,$75,$FE,$0F,$2F,$A3,$87,$BB,$2F,$93
	.dc.b	$5D,$0D,$E1,$EB,$9F,$14,$7D,$37,$E8,$57,$2C,$7D,$B9,$7C,$7C,$BE
	.dc.b	$96,$5F,$1A,$FE,$71,$C3,$0F,$8B,$3E,$58,$F1,$61,$CE,$7F,$D9,$D7
	.dc.b	$CD,$FF,$04,$5B,$1F,$F3,$5C,$7F,$93,$DA,$FF,$95,$9A,$FF,$B5,$E3
	.dc.b	$86,$1F,$E5,$D8,$FF,$30,$1F,$E5,$EF,$15,$E9,$66,$3F,$CB,$FE,$87
	.dc.b	$F9,$78,$9E,$1F,$F9,$7A,$3F,$C4,$87,$92,$3F,$C7,$1A,$7F,$86,$E3
	.dc.b	$FC,$14,$7F,$84,$1F,$F1,$BC,$FF,$83,$F5,$FF,$0F,$D7,$F8,$5E,$3F
	.dc.b	$C4,$D5,$FE,$15,$AF,$F2,$BC,$FF,$86,$A3,$FC,$2E,$66,$BC,$33,$FE
	.dc.b	$15,$CB,$FC,$A5,$1F,$8E,$F4,$BF,$CB,$15,$C2,$BC,$2F,$13,$FC,$6C
	.dc.b	$53,$FC,$72,$4A,$FF,$18,$E1,$E7,$8F,$59,$FE,$40,$C3,$D0,$FC,$18
	.dc.b	$FF,$25,$E1,$E1,$F4,$BF,$C9,$D3,$FE,$3F,$C7,$C7,$8F,$B2,$FD,$EF
	.dc.b	$90,$97,$FE,$30,$3D,$A7,$C6,$8F,$15,$7F,$92,$23,$B3,$FE,$3D,$3C
	.dc.b	$91,$AC,$7E,$1C,$7F,$9A,$4F,$A5,$3D,$5F,$83,$5E,$07,$24,$39,$9F
	.dc.b	$16,$7E,$1B,$71,$F8,$33,$F4,$5F,$C0,$D7,$92,$7E,$83,$DD,$F9,$91
	.dc.b	$C2,$3D,$91,$CE,$FD,$13,$CA,$3D,$07,$8A,$BC,$CD,$AE,$97,$E4,$9C
	.dc.b	$BF,$02,$7E,$30,$18,$31,$FE,$27,$5F,$E5,$DF,$43,$FC,$A1,$5B,$E3
	.dc.b	$70,$FF,$97,$E3,$0D,$7F,$CA,$FE,$87,$F1,$C9,$DC,$77,$AE,$6F,$F5
	.dc.b	$4A,$E1,$3A,$87,$A3,$FE,$69,$79,$9F,$E6,$AB,$FF,$2D,$5F,$F9,$99
	.dc.b	$A7,$F9,$6B,$D0,$FF,$30,$CF,$90,$F7,$7A,$1F,$E6,$57,$FC,$B1,$87
	.dc.b	$7E,$4F,$F2,$C4,$E6,$7C,$87,$0F,$47,$FC,$C2,$FF,$9A,$2B,$85,$E2
	.dc.b	$BF,$CB,$D1,$89,$EB,$9F,$F9,$A9,$E6,$7F,$99,$39,$FF,$CA,$91,$FD
	.dc.b	$03,$D1,$FF,$31,$4E,$9B,$5F,$0F,$D1,$FF,$2E,$4F,$F9,$7B,$9F,$FC
	.dc.b	$D4,$EF,$C3,$E6,$E4,$FF,$35,$57,$19,$FA,$58,$77,$F4,$7F,$CC,$53
	.dc.b	$F3,$FD,$A7,$C2,$EB,$0F,$C2,$46,$1F,$94,$3D,$2F,$F3,$35,$7D,$77
	.dc.b	$FC,$CB,$1E,$E9,$7D,$4F,$38,$46,$1E,$73,$C2,$BF,$01,$E9,$7F,$99
	.dc.b	$C9,$B7,$F8,$62,$9D,$BF,$E6,$88,$C2,$F4,$BF,$CD,$0D,$9F,$F3,$39
	.dc.b	$FE,$6A,$1D,$3F,$E6,$97,$31,$B5,$E3,$3D,$2F,$CB,$E5,$3F,$59,$7A
	.dc.b	$78,$E2,$FF,$36,$9F,$E3,$08,$CB,$F1,$AE,$E1,$F8,$AF,$11,$EE,$7F
	.dc.b	$96,$63,$F0,$73,$AC,$7F,$4B,$5A,$F6,$BF,$CD,$55,$FB,$5F,$53,$FC
	.dc.b	$DE,$A7,$FC,$DF,$8F,$F9,$96,$3F,$1B,$3A,$F4,$FF,$CD,$31,$8F,$53
	.dc.b	$D0,$A3,$38,$FF,$9E,$E7,$31,$8B,$FF,$3A,$9F,$E7,$0C,$7F,$CB,$B7
	.dc.b	$FC,$1F,$D2,$FF,$35,$CF,$F0,$9F,$4F,$FC,$E1,$3F,$E7,$3C,$7F,$CD
	.dc.b	$6D,$9D,$7B,$4E,$DE,$73,$FE,$6C,$9F,$F3,$1F,$A5,$FE,$68,$8E,$07
	.dc.b	$D4,$71,$7D,$0F,$E1,$7B,$A3,$35,$FE,$2A,$CB,$CB,$7C,$7D,$3F,$F3
	.dc.b	$67,$A5,$FE,$67,$8F,$F1,$79,$FE,$6D,$70,$FF,$9B,$2B,$FC,$D7,$E9
	.dc.b	$7F,$9A,$9F,$F3,$A9,$8F,$4B,$FC,$D9,$9F,$F9,$AE,$33,$5F,$C2,$23
	.dc.b	$0F,$F0,$0F,$3B,$FC,$58,$FF,$87,$0F,$F1,$4F,$27,$F6,$47,$FC,$72
	.dc.b	$7F,$98,$8A,$7F,$95,$BC,$FF,$F1,$E7,$5F,$F8,$94,$FF,$17,$D7,$F9
	.dc.b	$0E,$76,$FF,$8A,$E7,$FC,$8A,$70,$3F,$CA,$35,$DE,$7F,$C6,$D1,$87
	.dc.b	$FC,$AC,$7F,$A0,$0F,$F1,$F5,$7F,$8C,$1F,$43,$4F,$0F,$95,$FE,$4C
	.dc.b	$F4,$BF,$C6,$51,$E5,$AF,$05,$73,$8C,$3E,$D3,$C2,$66,$3C,$51,$FE
	.dc.b	$42,$69,$EA,$8E,$18,$7B,$DE,$27,$F9,$BE,$FF,$CE,$31,$F3,$AB,$CB
	.dc.b	$87,$F9,$26,$FF,$CA,$3E,$47,$F9,$6A,$7C,$78,$7C,$D3,$AE,$1E,$B9
	.dc.b	$D5,$78,$A7,$E6,$CF,$63,$FC,$C1,$1E,$E8,$F1,$5F,$F9,$72,$74,$FC
	.dc.b	$FC,$3D,$04,$8E,$D9,$FF,$96,$EB,$FC,$E8,$7E,$10,$FA,$67,$43,$E9
	.dc.b	$DF,$AA,$38,$65,$F4,$23,$CD,$3F,$12,$FE,$14,$79,$5B,$1E,$EC,$3E
	.dc.b	$23,$FE,$8D,$3C,$07,$F9,$9D,$FD,$2B,$FE,$65,$C3,$FC,$CE,$A7,$FC
	.dc.b	$CE,$FF,$99,$7D,$2F,$F3,$23,$FE,$0B,$F0,$9E,$47,$F9,$A2,$34,$74
	.dc.b	$A7,$A5,$FE,$68,$8C,$35,$FD,$AC,$7F,$86,$72,$FF,$0D,$CF,$F8,$96
	.dc.b	$7F,$C9,$F3,$FE,$1B,$3F,$C3,$9E,$97,$F9,$2A,$BF,$C8,$B3,$AA,$FF
	.dc.b	$33,$6F,$FE,$4D,$8F,$F2,$A1,$FE,$51,$F4,$BF,$CA,$71,$FE,$75,$8F
	.dc.b	$F1,$29,$FE,$24,$8C,$D7,$F8,$E6,$79,$1F,$E3,$DF,$13,$F5,$98,$78
	.dc.b	$6F,$FC,$93,$5E,$89,$FF,$1C,$9A,$BE,$73,$FE,$3A,$C7,$D5,$8F,$F8
	.dc.b	$F6,$3C,$07,$F9,$66,$33,$5C,$63,$13,$EC,$BF,$7C,$7A,$30,$F1,$3F
	.dc.b	$E4,$CA,$F0,$BE,$F3,$8B,$FE,$6E,$9F,$34,$7F,$93,$C9,$8F,$F9,$22
	.dc.b	$FB,$47,$B0,$FF,$31,$61,$E4,$7A,$1F,$E5,$C3,$31,$C1,$F1,$1D,$2B
	.dc.b	$BD,$75,$BF,$14,$F0,$6B,$FE,$5B,$30,$7D,$88,$FA,$D7,$F8,$0C,$7E
	.dc.b	$5C,$FC,$83,$C4,$71,$24,$79,$E7,$8E,$1F,$3F,$1F,$8F,$3F,$8F,$AE
	.dc.b	$F5,$F7,$5F,$6B,$F8,$0C,$7F,$CE,$AF,$A6,$3C,$91,$F8,$53,$FD,$8B
	.dc.b	$8F,$F0,$1C,$3E,$17,$DE,$BE,$BB,$F5,$B7,$FE,$71,$8C,$38,$6B,$FB
	.dc.b	$26,$BF,$D6,$1C,$F2,$7F,$99,$DB,$C7,$FE,$7B,$94,$61,$FF,$3F,$79
	.dc.b	$8F,$7F,$FC,$69,$39,$9D,$E1,$83,$FC,$71,$5B,$8F,$F3,$81,$E8,$F5
	.dc.b	$3F,$CE,$0D,$F5,$3F,$CE,$2E,$8F,$F3,$B9,$FE,$60,$F5,$7F,$C7,$2E
	.dc.b	$73,$FF,$3F,$57,$FA,$09,$B5,$FE,$47,$AD,$D7,$29,$CB,$C7,$3F,$74
	.dc.b	$EA,$7F,$CE,$13,$FE,$6E,$F4,$FF,$CF,$51,$FE,$8B,$6D,$7F,$9C,$E3
	.dc.b	$64,$70,$7C,$98,$FF,$3D,$B8,$9F,$F5,$37,$3B,$D4,$FF,$40,$75,$FF
	.dc.b	$97,$7D,$0F,$55,$FD,$16,$DF,$FA,$B5,$BF,$20,$FC,$35,$F7,$1F,$E7
	.dc.b	$2B,$E1,$EF,$5C,$3F,$CE,$3F,$07,$FE,$69,$EE,$FF,$31,$EF,$FE,$59
	.dc.b	$94,$61,$FF,$2F,$75,$65,$FF,$5E,$35,$FF,$30,$9A,$3F,$CD,$7E,$A7
	.dc.b	$E1,$FC,$9F,$F2,$96,$57,$D3,$FF,$23,$F8,$6E,$17,$C1,$FF,$92,$B1
	.dc.b	$FF,$33,$D5,$C1,$E4,$53,$FC,$9F,$C4,$FD,$65,$EE,$67,$4F,$F4,$5F
	.dc.b	$99,$FE,$36,$CB,$FC,$84,$7F,$AC,$3D,$0F,$F2,$5F,$17,$F8,$FE,$BF
	.dc.b	$C7,$9E,$F7,$F9,$4A,$73,$18,$39,$C7,$F9,$7B,$4F,$F2,$27,$9D,$FE
	.dc.b	$58,$6C,$FF,$8D,$AF,$FD,$9D,$EF,$7A,$23,$FD,$8D,$2F,$83,$7C,$1F
	.dc.b	$F9,$22,$78,$1F,$AD,$AE,$19,$FF,$B1,$AF,$85,$6E,$3F,$D7,$F3,$AC
	.dc.b	$3B,$F0,$79,$71,$EB,$8F,$4B,$FF,$29,$3F,$C1,$2F,$AC,$7F,$94,$CF
	.dc.b	$EE,$6E,$1F,$DF,$7D,$EF,$D4,$9F,$F5,$AB,$63,$C9,$4F,$AD,$FB,$13
	.dc.b	$9C,$7F,$DD,$CD,$BE,$7C,$9E,$5A,$DB,$23,$F8,$BC,$F5,$C3,$FC,$D2
	.dc.b	$D7,$8C,$7F,$1C,$F3,$3F,$C1,$F1,$FE,$2D,$23,$FE,$55,$7F,$C5,$7C
	.dc.b	$FF,$E2,$72,$3F,$E1,$E3,$FC,$29,$D3,$FE,$4F,$F3,$3F,$C3,$5E,$17
	.dc.b	$F8,$C2,$7F,$C3,$47,$F8,$9E,$7F,$C9,$0F,$F9,$2C,$87,$F8,$A5,$FF
	.dc.b	$1F,$DF,$86,$3F,$C4,$D3,$CD,$F2,$BF,$E3,$87,$FC,$63,$1E,$A3,$81
	.dc.b	$FE,$39,$BF,$F1,$4C,$7F,$8E,$2B,$CC,$F9,$1A,$F4,$8F,$6C,$78,$AB
	.dc.b	$B3,$E8,$9F,$35,$FF,$8E,$AF,$B4,$7B,$63,$FC,$C4,$F3,$3E,$31,$E1
	.dc.b	$7F,$C8,$8F,$F9,$D9,$B3,$FE,$62,$8E,$47,$D0,$38,$FA,$3F,$E4,$97
	.dc.b	$E3,$D7,$BE,$FC,$B7,$E0,$8F,$83,$1F,$E5,$89,$EC,$7D,$69,$EE,$7D
	.dc.b	$6C,$3F,$CC,$13,$BC,$3C,$47,$C4,$BE,$73,$F2,$B0,$F1,$61,$F0,$8F
	.dc.b	$1D,$FD,$5A,$DC,$7A,$8F,$B3,$5D,$8F,$F4,$E5,$72,$9F,$F2,$AC,$FF
	.dc.b	$96,$CF,$F2,$85,$7F,$96,$A3,$31,$FD,$9F,$B7,$FC,$AF,$E8,$FF,$96
	.dc.b	$A3,$F9,$6F,$A3,$FE,$59,$7F,$D8,$F3,$A7,$FC,$AF,$E7,$FE,$7F,$CF
	.dc.b	$FF,$2C,$47,$F8,$00,$C9,$FE,$1F,$3F,$C1,$7A,$FF,$82,$0F,$F1,$09
	.dc.b	$1C,$9F,$E5,$79,$FF,$10,$E3,$FE,$17,$9F,$F0,$64,$FF,$8C,$4F,$F0
	.dc.b	$DD,$7F,$8B,$67,$FC,$31,$18,$AE,$33,$93,$FC,$D1,$1F,$E4,$5F,$47
	.dc.b	$FC,$4F,$E8,$FF,$88,$27,$37,$FE,$2B,$AF,$F2,$29,$FE,$1C,$3F,$CA
	.dc.b	$D1,$CE,$FF,$C9,$7B,$FF,$8A,$9F,$F2,$A4,$FF,$8C,$2B,$D0,$F0,$8F
	.dc.b	$19,$1E,$87,$F8,$DE,$7F,$C5,$8F,$C1,$33,$7E,$59,$EF,$8F,$C2,$9F
	.dc.b	$F2,$5D,$7F,$8C,$4F,$F1,$B1,$E3,$7F,$C6,$F1,$FE,$64,$BF,$64,$79
	.dc.b	$70,$F5,$3F,$54,$F9,$07,$83,$1F,$8A,$47,$E1,$DF,$6B,$F1,$5F,$D0
	.dc.b	$3D,$F3,$EB,$8F,$05,$FF,$93,$6F,$94,$F8,$70,$F8,$58,$73,$8F,$8C
	.dc.b	$78,$4E,$F7,$F8,$77,$F0,$A7,$D9,$32,$7C,$78,$E4,$F8,$4F,$C3,$61
	.dc.b	$E4,$C5,$87,$F9,$16,$97,$FE,$47,$F4,$3F,$C8,$5E,$7B,$E0,$3F,$C8
	.dc.b	$27,$F9,$0F,$20,$E7,$FF,$28,$7A,$3F,$E5,$A8,$FF,$24,$9F,$E5,$97
	.dc.b	$17,$FE,$58,$8E,$D1,$CD,$E5,$CB,$87,$AD,$23,$FC,$B5,$87,$F9,$46
	.dc.b	$7F,$C8,$51,$FE,$40,$EC,$FF,$23,$1F,$E5,$FA,$FF,$22,$CF,$F9,$40
	.dc.b	$FF,$31,$0F,$47,$FC,$AE,$FF,$B7,$1B,$1C,$09,$5F,$E4,$CE,$0E,$8F
	.dc.b	$F9,$AE,$7F,$C7,$75,$CF,$D1,$CB,$F1,$23,$3E,$8F,$F9,$4E,$BF,$C9
	.dc.b	$12,$F4,$7F,$CA,$2F,$CA,$6A,$E8,$FF,$26,$3F,$0A,$7F,$CF,$51,$88
	.dc.b	$FF,$79,$B5,$F6,$B5,$DF,$A3,$F4,$E3,$86,$7F,$E5,$D9,$FF,$2C,$CF
	.dc.b	$A1,$F2,$79,$7F,$E5,$9C,$FE,$D7,$A2,$69,$27,$FD,$2F,$0F,$3F,$64
	.dc.b	$8C,$BF,$83,$EA,$FF,$2A,$6E,$F3,$BF,$97,$7B,$FF,$89,$BF,$F2,$89
	.dc.b	$F8,$CA,$46,$6B,$84,$FD,$AE,$8F,$F2,$D1,$1F,$F2,$D4,$69,$FF,$29
	.dc.b	$EF,$FE,$AC,$F1,$DE,$7F,$E2,$E7,$F0,$66,$8B,$1A,$9E,$1E,$46,$E7
	.dc.b	$F1,$33,$FE,$2F,$7F,$42,$D4,$DF,$33,$FC,$B4,$3C,$BF,$07,$1F,$87
	.dc.b	$CC,$FF,$2B,$DF,$EA,$AB,$75,$DE,$3F,$CB,$47,$9A,$39,$63,$FE,$C6
	.dc.b	$6C,$66,$7F,$CB,$DE,$B3,$CD,$FF,$2F,$36,$FF,$C9,$94,$F3,$7F,$CA
	.dc.b	$51,$EA,$F3,$7F,$CA,$33,$FE,$56,$D7,$FC,$B1,$A7,$F9,$5A,$BF,$CC
	.dc.b	$75,$A8,$C5,$6E,$7F,$CA,$D5,$E9,$8F,$F2,$C9,$FE,$56,$70,$F8,$5F
	.dc.b	$87,$1F,$BE,$F5,$9E,$B7,$82,$30,$FF,$93,$EB,$F8,$29,$FC,$43,$CD
	.dc.b	$FF,$2B,$3F,$43,$1F,$F2,$E5,$7F,$98,$3C,$EE,$13,$FE,$54,$7C,$35
	.dc.b	$FE,$55,$8E,$51,$83,$FD,$3B,$F7,$DF,$C8,$AB,$91,$FB,$19,$79,$E9
	.dc.b	$D4,$79,$0F,$F0,$B1,$3C,$BF,$F1,$44,$FF,$8F,$BC,$9F,$F1,$D7,$83
	.dc.b	$FE,$21,$7F,$C4,$95,$FE,$4C,$8F,$F0,$E1,$FE,$2E,$7F,$C6,$9D,$3F
	.dc.b	$E4,$5A,$E5,$1F,$E2,$27,$FC,$3C,$FF,$88,$6F,$FC,$8B,$3F,$E2,$BA
	.dc.b	$E5,$3F,$E2,$43,$B5,$70,$9D,$3D,$23,$D1,$3E,$28,$C5,$7A,$A7,$FC
	.dc.b	$8B,$E0,$7E,$A6,$FF,$C7,$38,$7F,$8F,$A7,$FC,$BC,$D7,$FC,$A9,$1E
	.dc.b	$97,$8E,$5E,$19,$F8,$78,$7F,$8F,$EF,$A4,$7C,$4B,$FF,$31,$61,$E1
	.dc.b	$79,$12,$38,$E1,$F0,$71,$F9,$8F,$88,$FF,$29,$10,$FF,$32,$17,$CF
	.dc.b	$FF,$22,$47,$BE,$3C,$A7,$F9,$68,$D9,$23,$78,$FA,$63,$96,$3E,$87
	.dc.b	$C0,$6D,$C3,$F0,$AF,$CF,$87,$D0,$8E,$EF,$CA,$79,$E5,$D5,$FA,$77
	.dc.b	$70,$F0,$C7,$DA,$25,$7C,$6C,$7D,$4F,$F7,$A8,$FE,$EB,$97,$F9,$71
	.dc.b	$B3,$FE,$5F,$25,$FF,$97,$7B,$7F,$CB,$53,$FE,$5C,$F3,$BF,$CB,$CF
	.dc.b	$F9,$8D,$AF,$4F,$3B,$FC,$BA,$DE,$EF,$F3,$01,$1F,$F1,$56,$3F,$E0
	.dc.b	$FF,$37,$FC,$3F,$9F,$F8,$F2,$7F,$C3,$07,$F8,$84,$FF,$0F,$15,$FF
	.dc.b	$0C,$F9,$DF,$E3,$F8,$E7,$87,$F8,$DA,$7F,$C7,$DB,$FF,$88,$E7,$FC
	.dc.b	$6F,$1C,$2F,$BC,$FF,$8C,$31,$FF,$29,$CF,$87,$1F,$F2,$0B,$70,$FF
	.dc.b	$1A,$61,$FE,$2A,$3C,$B1,$9A,$E3,$7C,$E7,$FC,$96,$7F,$8B,$2B,$C9
	.dc.b	$1F,$E7,$98,$F1,$BE,$5B,$F8,$51,$D5,$F0,$C7,$3B,$FF,$2B,$11,$FF
	.dc.b	$25,$D7,$F9,$C8,$F2,$D7,$C5,$8E,$91,$E0,$8F,$F2,$1D,$74,$7C,$97
	.dc.b	$FE,$4C,$AD,$B6,$3F,$C8,$D1,$87,$E8,$DF,$BD,$F6,$9F,$E6,$C8,$F6
	.dc.b	$57,$AA,$3C,$47,$22,$1F,$22,$3D,$87,$AA,$BF,$CE,$F1,$FE,$5C,$AF
	.dc.b	$AC,$FC,$3B,$F6,$C7,$F9,$62,$FB,$3F,$9B,$9F,$F2,$59,$FC,$BF,$CB
	.dc.b	$FF,$2A,$3F,$CD,$43,$87,$F9,$A7,$9A,$F3,$7F,$C9,$DD,$7F,$DE,$EB
	.dc.b	$4D,$F3,$B3,$C9,$FE,$51,$C3,$B4,$7F,$98,$1E,$D8,$73,$7C,$1E,$6F
	.dc.b	$F8,$FB,$CE,$FF,$28,$C6,$3C,$BF,$F1,$64,$E6,$3C,$DE,$72,$B3,$52
	.dc.b	$7F,$C2,$67,$F9,$2C,$B5,$FE,$56,$7F,$CB,$2F,$F9,$40,$91,$FE,$39
	.dc.b	$AF,$F2,$74,$FF,$8F,$0F,$89,$E7,$7F,$93,$0F,$F2,$6F,$47,$F9,$0B
	.dc.b	$3F,$F2,$9C,$67,$D1,$FF,$21,$4E,$6B,$91,$FD,$75,$C3,$FE,$55,$F3
	.dc.b	$FF,$C9,$14,$EC,$FF,$24,$4F,$F6,$7E,$B7,$CB,$7F,$8A,$78,$7F,$C6
	.dc.b	$11,$8C,$F7,$5F,$DA,$7D,$0F,$F2,$53,$F6,$FE,$0B,$ED,$B7,$93,$FC
	.dc.b	$AD,$3A,$8F,$EE,$1D,$DF,$E4,$CF,$43,$FC,$97,$97,$FA,$53,$AF,$FC
	.dc.b	$85,$3F,$E4,$A8,$FF,$23,$47,$F9,$16,$BF,$D3,$3D,$38,$BF,$F2,$1B
	.dc.b	$C5,$FF,$4E,$C7,$F9,$16,$BD,$3F,$00,$F7,$7F,$C3,$33,$FE,$42,$F4
	.dc.b	$6E,$EA,$D6,$FF,$95,$F4,$7F,$C5,$F7,$FE,$0A,$FC,$3B,$FD,$53,$39
	.dc.b	$F3,$38,$C6,$27,$FC,$19,$E8,$FF,$91,$2B,$FC,$25,$E9,$7F,$92,$25
	.dc.b	$39,$7F,$C4,$B8,$3D,$0F,$F2,$04,$72,$F4,$BF,$C8,$BE,$97,$F9,$10
	.dc.b	$98,$FF,$91,$3D,$2F,$2E,$0F,$43,$FC,$7D,$3F,$0A,$7D,$38,$7F,$8E
	.dc.b	$63,$FC,$7E,$47,$FC,$7C,$DE,$2F,$F2,$FF,$A5,$FE,$3C,$8F,$DB,$57
	.dc.b	$C2,$94,$E6,$BF,$C9,$47,$F8,$FA,$33,$E9,$7F,$90,$E3,$FD,$9D,$3B
	.dc.b	$9F,$F3,$A7,$A5,$FE,$42,$E3,$FF,$1A,$79,$9F,$E4,$79,$FC,$7F,$A3
	.dc.b	$FE,$36,$77,$E9,$7F,$8F,$A7,$5A,$7F,$B6,$AF,$8B,$FE,$40,$CB,$FC
	.dc.b	$9F,$18,$7F,$C7,$D8,$F7,$C7,$A5,$7F,$7C,$BF,$F1,$07,$9B,$FE,$B9
	.dc.b	$C3,$FC,$53,$E2,$FF,$74,$9D,$63,$FD,$F8,$FF,$07,$7A,$5F,$DF,$3B
	.dc.b	$7F,$BD,$5F,$F7,$FF,$13,$FC,$13,$8F,$3B,$FF,$03,$E1,$FE,$25,$3F
	.dc.b	$2B,$1F,$E4,$07,$C3,$8F,$F9,$02,$31,$3D,$F0,$FF,$07,$3F,$E1,$5C
	.dc.b	$BF,$C1,$D3,$FE,$2D,$BF,$2C,$E4,$F2,$E1,$E2,$C3,$A1,$FE,$2E,$8F
	.dc.b	$F0,$E1,$FE,$53,$3F,$C2,$73,$E4,$8D,$57,$A7,$0F,$F1,$C5,$6D,$F7
	.dc.b	$BC,$0F,$F2,$FD,$78,$B2,$F8,$8F,$3C,$BF,$C8,$D9,$7A,$31,$F9,$97
	.dc.b	$CE,$B4,$FF,$90,$EB,$6F,$AA,$B8,$47,$CD,$6C,$78,$4F,$F3,$36,$1E
	.dc.b	$42,$57,$47,$FC,$70,$7A,$A3,$FC,$6A,$71,$C7,$E7,$BF,$26,$3E,$7E
	.dc.b	$3E,$A8,$F6,$9E,$72,$9C,$A3,$BB,$89,$FA,$8F,$CD,$3F,$89,$87,$FC
	.dc.b	$78,$DC,$7F,$C7,$6F,$FB,$73,$C0,$FF,$1D,$BF,$E3,$9F,$57,$FC,$7D
	.dc.b	$E9,$FF,$8E,$A3,$97,$A3,$FE,$3A,$6F,$87,$FE,$39,$7B,$E2,$9F,$F0
	.dc.b	$94,$FF,$88,$A7,$FC,$1A,$7F,$82,$34,$FE,$FE,$FF,$7D,$7F,$C1,$99
	.dc.b	$FF,$7E,$8F,$F0,$39,$FE,$05,$38,$B9,$3F,$C2,$07,$F8,$85,$E4,$7F
	.dc.b	$83,$2F,$FC,$38,$FF,$84,$CF,$F0,$FD,$FF,$82,$2B,$C3,$0F,$4F,$FC
	.dc.b	$91,$7F,$E4,$68,$E5,$5B,$7C,$35,$E6,$23,$AC,$3D,$CF,$73,$FC,$76
	.dc.b	$64,$FF,$12,$E1,$F1,$63,$FC,$3B,$1E,$1F,$4F,$FC,$A2,$F6,$38,$61
	.dc.b	$ED,$74,$7F,$88,$67,$BD,$73,$8F,$F2,$FC,$7F,$8D,$DF,$90,$78,$E3
	.dc.b	$FC,$4B,$7F,$03,$1F,$7B,$FE,$36,$C7,$FC,$89,$5E,$83,$8D,$75,$C3
	.dc.b	$E2,$46,$27,$EE,$BE,$C6,$C7,$9E,$FA,$CF,$D0,$C3,$CF,$5A,$C3,$FC
	.dc.b	$7D,$3E,$83,$F0,$87,$C1,$3D,$3E,$8F,$F8,$A5,$FF,$16,$9F,$E3,$9F
	.dc.b	$23,$FC,$67,$C3,$FE,$31,$E7,$34,$FE,$E7,$1A,$91,$FE,$40,$F4,$BF
	.dc.b	$C7,$2E,$2F,$B5,$FF,$8F,$63,$B4,$E9,$FF,$17,$1F,$EF,$9F,$3F,$FC
	.dc.b	$68,$FF,$8D,$BD,$2F,$F1,$A3,$FE,$FD,$F4,$BF,$C7,$4E,$E3,$37,$E6
	.dc.b	$8F,$1C,$EA,$31,$0A,$EA,$F3,$F4,$BA,$BF,$E3,$C7,$FC,$82,$F9,$0C
	.dc.b	$1F,$E4,$4E,$6F,$F1,$8F,$A1,$FE,$37,$8F,$F3,$04,$7F,$8D,$0C,$FA
	.dc.b	$5F,$E3,$43,$FC,$77,$E9,$49,$E4,$F6,$93,$4F,$0F,$A0,$C3,$FC,$43
	.dc.b	$97,$F8,$E3,$D2,$FF,$1D,$70,$FF,$67,$6E,$3F,$4A,$F8,$3F,$E7,$37
	.dc.b	$FC,$6D,$5F,$0B,$D2,$FF,$1D,$CA,$31,$C5,$FE,$3C,$F5,$7B,$E3,$FE
	.dc.b	$3A,$A7,$37,$F8,$EB,$D5,$E1,$18,$F5,$35,$39,$C7,$FC,$75,$3F,$C9
	.dc.b	$BA,$AB,$FE,$3A,$7F,$C7,$8F,$F8,$DF,$0F,$F2,$B7,$A5,$FE,$36,$8F
	.dc.b	$24,$7F,$90,$3C,$B7,$A7,$FE,$3C,$AD,$61,$F7,$27,$FC,$77,$E2,$3D
	.dc.b	$2F,$F1,$C4,$E7,$D3,$FF,$1D,$E5,$FE,$46,$3F,$C6,$FE,$87,$E1,$1F
	.dc.b	$F2,$6F,$A4,$8C,$47,$E7,$23,$1A,$7F,$A1,$1B,$1F,$D0,$8E,$71,$FA
	.dc.b	$2B,$FC,$96,$5F,$E4,$1F,$47,$FC,$7B,$5F,$E4,$1A,$D1,$FA,$89,$D3
	.dc.b	$FE,$A9,$BE,$B7,$CB,$D2,$FF,$24,$BF,$9A,$F4,$59,$F9,$B1,$FF,$21
	.dc.b	$1F,$E4,$02,$1F,$D5,$A3,$FC,$7D,$E8,$7F,$8D,$F1,$FF,$25,$57,$F9
	.dc.b	$A2,$FF,$C7,$B1,$FE,$44,$8C,$4E,$EF,$F7,$B7,$CB,$D1,$FF,$21,$61
	.dc.b	$C2,$7F,$C8,$EF,$F9,$3A,$3F,$C7,$70,$DF,$FC,$71,$3F,$E4,$0C,$7F
	.dc.b	$D2,$9E,$7F,$F8,$E2,$BF,$84,$FA,$1F,$E3,$BF,$1F,$8E,$1C,$2F,$FD
	.dc.b	$03,$39,$7F,$C8,$98,$7F,$93,$2B,$EE,$79,$5F,$DD,$B0,$FE,$F5,$1F
	.dc.b	$E0,$73,$FC,$67,$87,$F9,$08,$FF,$08,$47,$F8,$4C,$9C,$DF,$E0,$6F
	.dc.b	$43,$FC,$5D,$3B,$9F,$F0,$67,$85,$FE,$2A,$3F,$C6,$31,$FE,$1C,$AE
	.dc.b	$19,$7F,$86,$9F,$F0,$4D,$7F,$86,$B0,$F1,$3F,$E4,$17,$FC,$90,$FF
	.dc.b	$83,$6F,$FC,$52,$FA,$0C,$5F,$9A,$BA,$C7,$9A,$B8,$BD,$4D,$9E,$69
	.dc.b	$F2,$46,$B1,$FF,$11,$FA,$1F,$E2,$1A,$E0,$48,$CE,$5F,$02,$3F,$CB
	.dc.b	$A7,$F3,$C8,$FF,$24,$57,$F9,$0E,$75,$87,$F8,$C6,$7E,$81,$31,$F5
	.dc.b	$B7,$1F,$55,$F8,$E7,$E5,$47,$CD,$AF,$8F,$5E,$2C,$BE,$55,$FC,$8C
	.dc.b	$7F,$C7,$4F,$1B,$F6,$C7,$97,$0F,$A6,$FF,$93,$C9,$87,$F9,$1A,$BD
	.dc.b	$17,$FE,$4E,$9F,$F1,$5D,$7F,$8D,$9B,$7F,$E3,$4F,$3E,$47,$EF,$71
	.dc.b	$FF,$01,$3F,$E3,$2F,$33,$FC,$6D,$18,$8F,$F0,$07,$9D,$FE,$0E,$7F
	.dc.b	$C0,$93,$FE,$00,$9F,$F0,$2C,$7F,$71,$7F,$C5,$27,$F8,$07,$2F,$F0
	.dc.b	$37,$6F,$F8,$36,$33,$5F,$DE,$EB,$FC,$62,$7F,$81,$67,$FC,$3B,$E7
	.dc.b	$7F,$80,$A3,$FC,$19,$5C,$27,$38,$78,$E3,$B4,$FF,$81,$CF,$F0,$64
	.dc.b	$E9,$FF,$10,$18,$3F,$C1,$1A,$FF,$90,$F1,$F6,$C6,$A3,$FC,$25,$7D
	.dc.b	$71,$F3,$4F,$F8,$A8,$FF,$26,$E1,$E3,$79,$DF,$F8,$F1,$F5,$18,$3F
	.dc.b	$C6,$6F,$F8,$62,$FF,$C4,$73,$C5,$B5,$FE,$47,$AF,$9B,$87,$C7,$8C
	.dc.b	$57,$BA,$7E,$1C,$60,$F2,$E3,$F1,$E7,$E9,$CF,$23,$FC,$4A,$5B,$F8
	.dc.b	$0F,$D7,$AE,$57,$F1,$E3,$DB,$18,$7D,$47,$B2,$58,$7F,$B1,$27,$FB
	.dc.b	$4C,$F6,$F3,$9F,$59,$FD,$BA,$3F,$C4,$CD,$51,$A7,$FC,$69,$E0,$66
	.dc.b	$31,$1F,$D9,$8E,$27,$F4,$C2,$99,$1E,$27,$F6,$A7,$47,$6A,$E8,$47
	.dc.b	$B9,$A9,$FF,$1C,$F0,$FF,$90,$63,$FC,$79,$EB,$79,$ED,$EE,$FF,$8E
	.dc.b	$EF,$75,$A3,$FC,$99,$19,$AF,$F2,$05,$70,$9E,$D1,$97,$67,$9B,$CF
	.dc.b	$FF,$0D,$BF,$DB,$DA,$FF,$8D,$0C,$79,$7F,$E3,$88,$DF,$47,$F9,$01
	.dc.b	$B9,$7F,$95,$AB,$FC,$9F,$3F,$E4,$4F,$3F,$EA,$C7,$F9,$0C,$F0,$5F
	.dc.b	$F9,$32,$B8,$4E,$AB,$FC,$97,$E5,$BE,$2F,$EC,$37,$0F,$F1,$84,$7D
	.dc.b	$DE,$87,$DF,$3E,$4F,$79,$CB,$FF,$1B,$3F,$83,$F7,$5F,$00,$F8,$0F
	.dc.b	$C2,$7B,$2F,$43,$FC,$63,$D7,$FE,$2F,$C7,$FC,$64,$7F,$9C,$A7,$5B
	.dc.b	$FF,$8B,$A1,$18,$76,$7F,$8C,$CA,$7F,$9E,$BA,$FF,$C6,$ED,$37,$E5
	.dc.b	$FF,$84,$38,$FF,$12,$D8,$FC,$96,$8F,$7D,$EB,$7F,$8B,$EB,$59,$7F
	.dc.b	$8D,$F8,$30,$7F,$9D,$DF,$F3,$C7,$17,$F8,$E7,$E4,$73,$9F,$F8,$4C
	.dc.b	$99,$FE,$7F,$07,$BA,$F6,$BF,$C7,$9E,$67,$6B,$FF,$0E,$9C,$32,$FF
	.dc.b	$19,$D6,$E3,$06,$B0,$FF,$13,$94,$FF,$14,$79,$DF,$E2,$DF,$2D,$EE
	.dc.b	$FF,$87,$33,$F0,$FC,$07,$F8,$7A,$D5,$AC,$3F,$C2,$A7,$F8,$F7,$15
	.dc.b	$6E,$75,$5F,$B5,$6E,$5E,$69,$D6,$3D,$EF,$8E,$1D,$30,$FF,$13,$F9
	.dc.b	$FF,$E5,$BC,$7F,$C6,$DB,$BC,$FF,$F2,$3D,$FF,$8A,$2B,$44,$7F,$C6
	.dc.b	$3E,$9D,$C3,$C5,$37,$4F,$A7,$7C,$A7,$BD,$FF,$08,$F2,$BF,$BE,$47
	.dc.b	$F8,$3C,$FE,$DF,$E5,$7F,$8B,$9F,$F0,$48,$8F,$F0,$4B,$FD,$DF,$2F
	.dc.b	$F0,$1B,$4D,$C6,$9F,$F1,$11,$FE,$15,$F3,$BF,$C0,$EE,$9F,$EF,$FD
	.dc.b	$3F,$DF,$63,$93,$FE,$0A,$8D,$BF,$E1,$57,$FC,$26,$4A,$F0,$CF,$96
	.dc.b	$39,$DF,$A6,$B8,$56,$AF,$C8,$7C,$08,$ED,$1F,$E1,$2B,$E4,$FF,$86
	.dc.b	$EB,$FC,$31,$E8,$7F,$8B,$EF,$FC,$39,$5F,$E3,$93,$CF,$3F,$E1,$47
	.dc.b	$07,$23,$DD,$5C,$DF,$F1,$ED,$F8,$1F,$9E,$7F,$92,$CE,$64,$3E,$44
	.dc.b	$79,$6B,$D3,$87,$C1,$9F,$B1,$5F,$25,$F9,$F3,$E6,$8F,$F2,$79,$2F
	.dc.b	$FC,$61,$3E,$03,$FC,$66,$77,$9F,$89,$1F,$03,$D0,$FF,$22,$BE,$DA
	.dc.b	$E5,$5B,$BF,$14,$7A,$CF,$15,$F2,$9F,$F2,$D9,$FE,$34,$3F,$C6,$53
	.dc.b	$FE,$37,$7F,$9D,$76,$3D,$E7,$A3,$FE,$38,$9F,$E7,$C7,$F2,$08,$CB
	.dc.b	$FE,$3B,$F3,$FF,$C7,$24,$F0,$3F,$BE,$65,$FD,$E2,$7F,$C0,$11,$FC
	.dc.b	$8F,$3F,$F0,$A1,$FE,$36,$8F,$F0,$3E,$7F,$E0,$F3,$FC,$16,$5B,$FF
	.dc.b	$20,$12,$BF,$C2,$C7,$39,$FF,$1D,$CF,$F8,$76,$BA,$5F,$F8,$42,$7F
	.dc.b	$C2,$51,$FE,$27,$9D,$47,$F8,$E6,$7F,$C6,$2F,$F8,$CE,$7C,$AE,$5A
	.dc.b	$FF,$88,$AB,$C0,$FF,$8D,$67,$DC,$F6,$3F,$C6,$27,$92,$FF,$C3,$91
	.dc.b	$FE,$30,$9E,$3E,$87,$F8,$A6,$7D,$95,$F0,$9F,$F1,$6F,$A3,$FE,$43
	.dc.b	$AF,$97,$87,$F9,$22,$BF,$C4,$F1,$CA,$75,$1F,$E2,$8C,$7D,$AF,$9C
	.dc.b	$FF,$29,$63,$E1,$6D,$FC,$63,$E2,$61,$E6,$71,$1E,$A3,$DC,$F8,$67
	.dc.b	$DF,$5F,$57,$0F,$A7,$5D,$0F,$90,$7C,$17,$CC,$FC,$D3,$06,$CF,$34
	.dc.b	$FD,$27,$DA,$1F,$9B,$8B,$1F,$F0,$84,$7D,$0F,$3F,$FC,$73,$C8,$E3
	.dc.b	$FF,$0A,$79,$CF,$99,$FF,$16,$1F,$E3,$3C,$BF,$C7,$57,$FE,$40,$7F
	.dc.b	$3A,$D7,$FC,$38,$D7,$FC,$82,$6B,$D1,$E5,$87,$78,$FF,$22,$B5,$FF
	.dc.b	$22,$FA,$5F,$E3,$4F,$7F,$FC,$78,$7F,$8C,$E3,$F5,$67,$F9,$53,$D2
	.dc.b	$AF,$F8,$F4,$CE,$08,$CC,$6C,$FF,$1E,$90,$FF,$2A,$65,$FE,$41,$7D
	.dc.b	$B1,$DF,$0F,$24,$E5,$FF,$1D,$79,$BF,$E6,$5F,$47,$FC,$01,$7F,$E5
	.dc.b	$A6,$BF,$E3,$92,$1F,$E4,$52,$1F,$E4,$1F,$4B,$FC,$92,$D9,$FF,$1F
	.dc.b	$23,$FC,$8B,$B7,$F9,$03,$8B,$FC,$7B,$E8,$FB,$3D,$2F,$F1,$F5,$7C
	.dc.b	$46,$BF,$C4,$1A,$FF,$93,$DE,$EF,$56,$A6,$98,$EA,$FF,$1B,$57,$F8
	.dc.b	$F4,$9E,$D7,$F2,$0F,$43,$FC,$82,$FD,$4E,$5F,$F1,$BE,$1F,$91,$3F
	.dc.b	$C7,$65,$A4,$FF,$98,$7C,$8F,$F1,$C4,$FF,$26,$6B,$9F,$27,$F0,$8D
	.dc.b	$D3,$EF,$CF,$F9,$65,$BE,$6E,$A7,$E4,$4E,$5F,$C3,$B5,$FF,$15,$CF
	.dc.b	$F8,$C1,$47,$F8,$EB,$CC,$FF,$1C,$DC,$AF,$C6,$F9,$9E,$4B,$FA,$8D
	.dc.b	$4D,$3F,$C8,$0B,$8D,$EA,$BC,$8E,$93,$FE,$A7,$9C,$C7,$F8,$BA,$FF
	.dc.b	$D5,$19,$7F,$8F,$1F,$F1,$75,$7F,$8F,$3C,$CF,$F1,$97,$9B,$B8,$FD
	.dc.b	$1F,$99,$FE,$3A,$7F,$A7,$B5,$FF,$47,$35,$FC,$63,$FE,$3E,$AF,$C7
	.dc.b	$79,$3F,$E3,$7F,$13,$FC,$75,$87,$F9,$8E,$73,$E5,$FF,$B2,$9F,$F5
	.dc.b	$BD,$FA,$DB,$9E,$F1,$63,$FE,$41,$F3,$3F,$C7,$55,$FE,$41,$3F,$6F
	.dc.b	$7C,$6B,$D3,$7F,$E5,$D2,$4F,$FB,$24,$7D,$FF,$FB,$59,$EE,$65,$C4
	.dc.b	$FF,$B3,$7C,$7F,$EF,$0F,$F8,$40,$FF,$03,$79,$BF,$E2,$1F,$33,$FC
	.dc.b	$67,$B7,$F8,$0C,$FF,$05,$F4,$FF,$80,$23,$FC,$07,$1F,$E0,$9C,$7F
	.dc.b	$AB,$E1,$FE,$18,$D7,$FC,$65,$5F,$E3,$6C,$7F,$C4,$07,$F8,$9F,$0F
	.dc.b	$F1,$BB,$FE,$1C,$C3,$FC,$74,$77,$7F,$C2,$15,$C8,$FF,$16,$3E,$79
	.dc.b	$F1,$46,$63,$FC,$62,$48,$EA,$7F,$85,$E7,$AC,$74,$3D,$FE,$6F,$F8
	.dc.b	$81,$C4,$F4,$7C,$87,$A3,$0F,$63,$83,$FC,$56,$7A,$EB,$D0,$D3,$9B
	.dc.b	$E0,$7F,$C6,$F8,$7C,$C7,$27,$57,$E3,$3D,$27,$BD,$7A,$63,$58,$7C
	.dc.b	$29,$FF,$20,$5F,$D4,$3B,$3F,$70,$CC,$F6,$3E,$44,$6A,$FE,$CC,$7F
	.dc.b	$95,$A3,$ED,$3F,$E5,$C9,$F0,$C6,$2F,$FC,$87,$3E,$D3,$D7,$7F,$E1
	.dc.b	$6C,$98,$FC,$B9,$FF,$32,$1F,$E4,$32,$1F,$EF,$BC,$18,$FE,$8A,$3F
	.dc.b	$AB,$F6,$7F,$91,$7C,$DF,$F2,$24,$7F,$92,$3C,$DF,$F2,$1B,$FE,$46
	.dc.b	$B7,$9B,$FE,$43,$8F,$F2,$37,$77,$F5,$E9,$FF,$04,$47,$F8,$7E,$3F
	.dc.b	$C3,$AF,$F9,$18,$FF,$02,$47,$F9,$0A,$7F,$C4,$71,$FE,$09,$3F,$C2
	.dc.b	$66,$F5,$FF,$04,$1F,$E1,$48,$DC,$7F,$82,$A3,$3B,$7F,$86,$49,$1C
	.dc.b	$5F,$F1,$ED,$77,$BF,$1C,$FF,$8E,$CF,$F5,$85,$FF,$89,$5F,$F0,$E9
	.dc.b	$E2,$35,$3F,$E3,$03,$93,$FE,$33,$8E,$0F,$F8,$7A,$BB,$C7,$F9,$02
	.dc.b	$FF,$CA,$D5,$F1,$1B,$1E,$A7,$D9,$7E,$49,$D5,$7F,$8E,$23,$FC,$5B
	.dc.b	$7E,$F8,$EB,$1F,$E3,$4A,$F8,$33,$EF,$3F,$C6,$2E,$1F,$4D,$F9,$AF
	.dc.b	$D3,$3F,$E4,$73,$4F,$82,$7F,$C8,$9E,$6F,$F9,$66,$FB,$61,$E0,$9F
	.dc.b	$B8,$7C,$18,$F2,$DF,$AA,$30,$7F,$8F,$67,$FC,$0C,$FC,$BF,$23,$FC
	.dc.b	$77,$C3,$FE,$3A,$F3,$3F,$C7,$33,$FE,$0B,$EB,$FF,$17,$56,$A3,$FC
	.dc.b	$63,$5B,$74,$D7,$FC,$7F,$E6,$7F,$8E,$23,$15,$FE,$24,$BC,$57,$F8
	.dc.b	$C6,$3F,$C4,$FE,$5F,$F8,$E2,$3F,$C5,$D1,$E0,$F2,$F1,$1F,$E3,$09
	.dc.b	$FF,$17,$5F,$F9,$41,$CD,$6E,$7F,$C6,$C7,$F9,$3A,$BF,$C7,$78,$FF
	.dc.b	$8F,$63,$FC,$82,$D7,$FC,$8B,$E4,$BC,$9F,$F1,$B7,$89,$27,$DA,$7F
	.dc.b	$8B,$67,$4E,$0F,$F1,$79,$FE,$31,$F2,$BF,$C7,$78,$FF,$95,$5B,$9F
	.dc.b	$F8,$DE,$78,$57,$F8,$A8,$98,$7F,$94,$1F,$F1,$5F,$95,$F5,$B0,$FF
	.dc.b	$30,$7A,$4F,$9C,$7D,$66,$72,$FF,$20,$4E,$63,$FC,$7F,$97,$97,$CD
	.dc.b	$F3,$BF,$E0,$DA,$FE,$F3,$F0,$1F,$E6,$9C,$7F,$CD,$9E,$02,$31,$3F
	.dc.b	$E4,$1D,$7B,$5F,$F9,$37,$B7,$FC,$7B,$E6,$FF,$8E,$F1,$FF,$28,$74
	.dc.b	$FF,$8F,$AB,$5E,$77,$F8,$F9,$FF,$6D,$F5,$7F,$8F,$8D,$DF,$F9,$06
	.dc.b	$15,$FE,$3E,$7F,$C6,$7E,$6F,$F8,$17,$57,$BB,$FE,$5D,$D7,$E8,$79
	.dc.b	$9F,$E2,$89,$CD,$FF,$82,$CF,$F1,$AF,$9F,$FE,$0D,$E3,$F9,$1F,$7F
	.dc.b	$9B,$FF,$1F,$6F,$F3,$39,$53,$FE,$2E,$C3,$FC,$7E,$7F,$A2,$3D,$0F
	.dc.b	$F1,$B6,$BF,$E4,$78,$FF,$1C,$BF,$E1,$9B,$FF,$19,$E9,$FE,$45,$6C
	.dc.b	$FC,$F3,$FC,$83,$3B,$F4,$3F,$C3,$53,$FE,$99,$BF,$F2,$0B,$C0,$FF
	.dc.b	$4E,$79,$9F,$E3,$1B,$FF,$1D,$61,$FE,$22,$F3,$FC,$53,$93,$FC,$69
	.dc.b	$3F,$E2,$AF,$3F,$FC,$60,$F8,$4E,$27,$8F,$83,$FC,$71,$7F,$E3,$4B
	.dc.b	$F9,$D5,$C3,$0F,$0C,$63,$C4,$FF,$1E,$D7,$E8,$AF,$8E,$1D,$71,$F0
	.dc.b	$C6,$30,$F4,$F9,$DF,$E3,$DB,$FF,$19,$F8,$BF,$E3,$CC,$3F,$D9,$1E
	.dc.b	$77,$F8,$CA,$F7,$5F,$E3,$5A,$D3,$C0,$FF,$1C,$D7,$CB,$70,$F8,$6F
	.dc.b	$B5,$F3,$9F,$F0,$8F,$91,$FD,$E2,$BF,$BE,$10,$FF,$00,$9F,$E0,$A3
	.dc.b	$FC,$4D,$9F,$0F,$33,$FC,$23,$E7,$7F,$90,$74,$FE,$F9,$E7,$7F,$7B
	.dc.b	$C3,$FC,$19,$5F,$E1,$9C,$3F,$C0,$27,$F8,$AB,$1F,$F2,$0C,$72,$AE
	.dc.b	$0F,$F8,$3A,$7C,$27,$F8,$75,$D4,$F0,$BF,$F0,$E4,$75,$BE,$F1,$E2
	.dc.b	$9D,$57,$A0,$FF,$0E,$61,$E7,$7C,$47,$F8,$66,$FA,$B6,$BA,$9C,$DC
	.dc.b	$1C,$CF,$5F,$67,$F9,$04,$FF,$18,$3E,$78,$E2,$F6,$8F,$F1,$1E,$1F
	.dc.b	$3E,$3D,$B7,$E3,$9E,$59,$7F,$8A,$4F,$F1,$B4,$FD,$18,$FF,$2A,$C6
	.dc.b	$63,$FC,$6D,$5F,$E3,$87,$FC,$5C,$FC,$27,$C1,$1C,$9F,$21,$FD,$74
	.dc.b	$FF,$71,$47,$F8,$B2,$FF,$C4,$58,$7F,$8B,$67,$FC,$62,$69,$C3,$FE
	.dc.b	$2D,$F3,$BF,$C5,$EF,$F1,$FF,$3B,$FC,$5E,$FF,$8D,$FC,$EF,$F1,$73
	.dc.b	$FE,$31,$7F,$8C,$BF,$E3,$03,$FB,$7E,$5F,$DE,$A3,$FC,$19,$97,$F8
	.dc.b	$8A,$7F,$C0,$3D,$DF,$DD,$A3,$19,$FF,$7F,$9F,$F1,$0C,$E5,$FF,$05
	.dc.b	$1F,$E2,$18,$D1,$FE,$21,$AF,$F0,$E5,$7F,$85,$67,$FC,$77,$18,$7F
	.dc.b	$C1,$3E,$9F,$F8,$03,$0F,$F0,$21,$FE,$0F,$AF,$F0,$53,$FE,$22,$C3
	.dc.b	$FC,$41,$7E,$8B,$F0,$F9,$FF,$E1,$17,$FC,$1F,$5C,$5F,$F0,$ED,$F2
	.dc.b	$3C,$51,$F0,$6F,$99,$27,$E2,$CE,$DF,$29,$FE,$4F,$AE,$D5,$E3,$8D
	.dc.b	$9D,$88,$7C,$3B,$F3,$3F,$30,$F8,$37,$FE,$26,$77,$5D,$2F,$CA,$78
	.dc.b	$63,$E7,$9C,$0F,$6B,$F7,$E7,$94,$77,$7D,$57,$F0,$63,$FC,$A9,$F3
	.dc.b	$2F,$4F,$FB,$3E,$AF,$31,$F2,$D8,$6D,$FF,$82,$6F,$83,$6F,$FC,$6A
	.dc.b	$C8,$CB,$80,$E1,$D4,$65,$1A,$7F,$BC,$B4,$EC,$7F,$87,$2D,$83,$5D
	.dc.b	$DB,$E3,$FD,$18,$4C,$3D,$11,$A3,$C0,$F9,$1B,$3C,$63,$C8,$D7,$D4
	.dc.b	$DE,$97,$2F,$F9,$52,$30,$DA,$F9,$78,$FF,$AB,$2F,$C2,$43,$FA,$F5
	.dc.b	$70,$59,$7B,$39,$BD,$7D,$3F,$47,$EE,$65,$7D,$18,$CB,$E6,$7D,$24
	.dc.b	$70,$6C,$87,$F7,$DA,$F9,$3F,$13,$F3,$BD,$0F,$F1,$DC,$FD,$36,$BF
	.dc.b	$67,$A1,$F0,$FF,$21,$B1,$F6,$5A,$EE,$3E,$6F,$99,$FE,$30,$8F,$F1
	.dc.b	$6F,$07,$F8,$A9,$A7,$F8,$AB,$CB,$FF,$1A,$34,$FF,$18,$94,$FE,$04
	.dc.b	$53,$F0,$FB,$F3,$73,$CD,$FE,$33,$5F,$06,$F8,$4F,$F1,$13,$79,$3F
	.dc.b	$20,$D9,$FC,$B3,$5F,$F1,$59,$0F,$F4,$8F,$8B,$FE,$2C,$9F,$CC,$7C
	.dc.b	$BF,$F8,$91,$FF,$4A,$B6,$7E,$DE,$7F,$AD,$9F,$F3,$84,$E6,$4B,$C3
	.dc.b	$FF,$1B,$36,$7A,$C6,$08,$FE,$87,$C7,$7C,$5F,$F9,$26,$BC,$1C,$89
	.dc.b	$CD,$78,$3E,$5B,$47,$E6,$31,$FC,$16,$4E,$7C,$D6,$C9,$3A,$8C,$BF
	.dc.b	$E3,$5C,$57,$E8,$25,$7C,$46,$CF,$E3,$F0,$FD,$A4,$F4,$C3,$C5,$8F
	.dc.b	$9F,$AB,$D5,$F0,$72,$BF,$7D,$4F,$7F,$18,$F8,$6B,$E3,$92,$BE,$9B
	.dc.b	$63,$C1,$F5,$FC,$2F,$BE,$48,$FF,$38,$78,$BF,$DD,$79,$7F,$BF,$DF
	.dc.b	$F8,$BE,$3F,$BD,$1F,$E0,$52,$F4,$FF,$85,$27,$FC,$12,$DF,$0F,$FB
	.dc.b	$E7,$91,$FE,$09,$7F,$DA,$ED,$BF,$F0,$44,$7F,$8D,$FC,$8F,$F0,$9D
	.dc.b	$F8,$27,$FC,$23,$E0,$7F,$8C,$27,$FC,$4F,$39,$C7,$C3,$1C,$2F,$AD
	.dc.b	$6E,$FF,$C1,$47,$2D,$B8,$CF,$78,$FF,$13,$5F,$BE,$38,$CE,$A7,$FC
	.dc.b	$67,$5C,$F0,$FF,$14,$E3,$E9,$9F,$F0,$BC,$7F,$88,$1F,$8D,$97,$F8
	.dc.b	$FC,$FF,$23,$63,$E6,$8E,$F1,$83,$59,$77,$8E,$79,$79,$6F,$A4,$7A
	.dc.b	$4F,$F1,$EB,$8B,$E8,$DB,$E7,$3F,$02,$73,$5C,$C9,$1E,$CC,$BC,$13
	.dc.b	$FE,$3F,$C5,$97,$C9,$AF,$F1,$68,$9F,$F1,$93,$FD,$65,$AF,$F5,$EE
	.dc.b	$C7,$91,$FE,$2E,$8F,$F1,$87,$1F,$F8,$C9,$FE,$EA,$7F,$8B,$FB,$3F
	.dc.b	$AF,$5B,$C8,$FF,$18,$CE,$4F,$F1,$96,$DF,$DD,$7C,$8F,$F0,$04,$7F
	.dc.b	$83,$8F,$74,$FF,$82,$2B,$FC,$56,$FF,$8E,$AB,$FC,$31,$3F,$E1,$03
	.dc.b	$FC,$00,$5A,$E6,$FF,$82,$23,$67,$7F,$1B,$FC,$5D,$BF,$F8,$F2,$FB
	.dc.b	$13,$1F,$F0,$4C,$7F,$83,$6F,$FC,$1E,$7F,$96,$27,$FC,$70,$F5,$8F
	.dc.b	$F0,$B6,$3F,$E3,$FA,$FF,$09,$E3,$E2,$8C,$57,$23,$81,$EF,$C7,$FC
	.dc.b	$7F,$7D,$DF,$F1,$61,$E7,$8F,$F1,$86,$3E,$87,$FC,$7E,$D7,$FC,$59
	.dc.b	$1E,$E9,$E4,$7F,$8A,$A3,$FC,$5B,$3B,$AF,$97,$87,$F9,$66,$78,$D7
	.dc.b	$AB,$1F,$9C,$D9,$D5,$70,$AF,$F1,$96,$1E,$6A,$EF,$1E,$B7,$E6,$57
	.dc.b	$CA,$8F,$90,$FC,$4C,$3C,$58,$7C,$F9,$FB,$11,$89,$E1,$1F,$18,$98
	.dc.b	$FF,$88,$5F,$EC,$77,$FE,$13,$FB,$AF,$F1,$C7,$C4,$49,$FF,$1C,$47
	.dc.b	$F8,$98,$FF,$0F,$91,$FF,$0F,$B5,$FF,$0A,$BB,$7F,$C6,$19,$3E,$D3
	.dc.b	$FC,$5F,$E4,$FF,$91,$7C,$B9,$5F,$E3,$63,$FC,$3B,$1F,$E2,$C2,$3F
	.dc.b	$E2,$C6,$BA,$38,$74,$FF,$44,$BF,$64,$64,$91,$FD,$65,$FF,$22,$1F
	.dc.b	$E1,$B8,$DC,$67,$CA,$FF,$16,$70,$FF,$96,$EF,$FC,$85,$7F,$E3,$0E
	.dc.b	$DF,$F2,$13,$FE,$3E,$7F,$C8,$24,$70,$7F,$8D,$EF,$A4,$7F,$8F,$39
	.dc.b	$7F,$CA,$BE,$F3,$DF,$FF,$25,$B6,$FF,$CB,$8D,$7F,$CC,$4D,$7F,$CC
	.dc.b	$FE,$57,$F8,$B9,$BE,$D7,$F9,$52,$0F,$F8,$AB,$DA,$65,$FE,$1F,$EA
	.dc.b	$FF,$16,$59,$A6,$1F,$F1,$29,$FE,$27,$3F,$CC,$E5,$DF,$FC,$50,$EB
	.dc.b	$9B,$FC,$51,$E4,$7F,$87,$63,$FC,$F0,$D7,$FC,$FB,$E8,$BD,$2F,$A5
	.dc.b	$1F,$E7,$81,$19,$EF,$FF,$3E,$36,$3F,$D0,$4D,$7F,$D0,$FE,$3F,$F8
	.dc.b	$2A,$7F,$C0,$87,$FA,$0B,$93,$DF,$E4,$7F,$84,$35,$FC,$05,$6E,$FF
	.dc.b	$D1,$8D,$79,$F9,$1F,$E1,$06,$E1,$FE,$27,$E5,$E4,$F5,$F1,$BF,$C2
	.dc.b	$73,$FE,$0B,$8F,$F1,$AF,$8F,$FE,$25,$C7,$FC,$2A,$7F,$84,$63,$24
	.dc.b	$9E,$07,$F8,$AB,$C8,$FF,$07,$47,$F9,$0A,$BF,$C2,$53,$EE,$6B,$98
	.dc.b	$FF,$1D,$79,$1C,$23,$FD,$03,$87,$F8,$56,$7F,$D4,$78,$7E,$7F,$C6
	.dc.b	$71,$79,$67,$FC,$1A,$DB,$FF,$27,$F4,$D5,$5C,$2B,$99,$C0,$FF,$58
	.dc.b	$39,$8F,$F5,$B3,$5F,$F5,$DB,$5E,$07,$F8,$EF,$EE,$BB,$47,$1C,$3F
	.dc.b	$C5,$B1,$FE,$2C,$F2,$3F,$C0,$1E,$4F,$F1,$4B,$FF,$10,$9F,$DC,$CF
	.dc.b	$EF,$BE,$27,$F7,$67,$FB,$84,$F0,$7F,$C3,$78,$7F,$84,$49,$DD,$FE
	.dc.b	$15,$AF,$C8,$79,$1F,$E0,$0A,$E9,$D5,$FE,$34,$C7,$FC,$4D,$1E,$03
	.dc.b	$AB,$FE,$11,$C3,$9E,$9F,$E0,$4C,$BC,$17,$FE,$2F,$7F,$C1,$AF,$F8
	.dc.b	$8B,$2F,$06,$1F,$E2,$0C,$3D,$8F,$F8,$68,$F0,$E5,$FE,$42,$9D,$C7
	.dc.b	$F8,$6E,$BC,$8F,$F8,$92,$BC,$53,$AC,$BD,$53,$E6,$3F,$C3,$98,$78
	.dc.b	$B0,$F7,$B6,$3C,$47,$C6,$AE,$51,$98,$F0,$4F,$8F,$1F,$76,$1E,$06
	.dc.b	$CF,$68,$F5,$E5,$F1,$23,$EA,$9E,$C3,$FD,$EB,$8F,$F9,$8F,$0E,$CF
	.dc.b	$A0,$FF,$26,$47,$CF,$6B,$F1,$4F,$66,$5F,$2F,$0F,$0B,$63,$B9,$AC
	.dc.b	$7D,$F8,$F9,$32,$FF,$1B,$4E,$0F,$F1,$A6,$29,$FF,$16,$B6,$3F,$8A
	.dc.b	$1F,$E3,$42,$9F,$C9,$FC,$9F,$F1,$C3,$7A,$32,$D3,$FC,$81,$E4,$FF
	.dc.b	$8D,$9F,$F1,$01,$FE,$37,$21,$82,$1F,$DF,$0E,$3C,$3F,$E4,$08,$E5
	.dc.b	$19,$38,$94,$E2,$7F,$92,$0B,$AF,$F9,$1E,$79,$12,$3A,$90,$E6,$71
	.dc.b	$3A,$10,$C1,$23,$A3,$C4,$EF,$3E,$67,$4D,$8D,$C7,$66,$BC,$27,$C8
	.dc.b	$E6,$78,$12,$33,$1B,$3C,$07,$96,$79,$B6,$78,$3A,$3C,$73,$C9,$D3
	.dc.b	$5C,$9B,$9E,$6E,$0F,$24,$6E,$32,$49,$EF,$3A,$39,$38,$7D,$06,$48
	.dc.b	$60,$87,$90,$E3,$3D,$E3,$94,$64,$E2,$53,$89,$E7,$2C,$F9,$67,$91
	.dc.b	$23,$A9,$0E,$67,$13,$A1,$0C,$12,$3A,$3C,$4E,$F3,$E6,$74,$D8,$DC
	.dc.b	$76,$6B,$C2,$7C,$8E,$67,$81,$23,$31,$B3,$C0,$79,$67,$9B,$67,$83
	.dc.b	$A3,$C7,$3C,$9D,$35,$C9,$B9,$E6,$E0,$F2,$46,$E3,$24,$9E,$F3,$A3
	.dc.b	$93,$87,$D0,$64,$86,$08,$79,$0E,$33,$DE,$39,$46,$4E,$25,$38,$9E
	.dc.b	$72,$CF,$96,$79,$12,$3A,$90,$E6,$71,$3A,$10,$C1,$23,$A3,$C4,$EF
	.dc.b	$3E,$67,$4D,$8D,$C7,$66,$BC,$27,$C8,$E6,$78,$12,$33,$1F,$C7,$BB
	.dc.b	$7F,$DD,$36,$87,$D5,$FF,$98,$1F,$E7,$0D,$7F,$9A,$36,$78,$C7,$0F
	.dc.b	$03,$FC,$A5,$18,$9D,$64,$BE,$D1,$86,$D7,$09,$D5,$F3,$3F,$AF,$E7
	.dc.b	$FD,$2D,$FE,$97,$9A,$31,$1F,$D4,$CC,$B8,$78,$E9,$FD,$50,$8F,$F5
	.dc.b	$86,$BC,$1A,$F3,$21,$C8,$E0,$43,$81,$4D,$9F,$E0,$93,$24,$33,$3E
	.dc.b	$98,$EF,$1D,$4D,$14,$DC,$FA,$DE,$86,$23,$8B,$83,$81,$0C,$12,$7C
	.dc.b	$0F,$9C,$D3,$87,$C2,$6E,$7C,$11,$98,$F0,$1C,$4D,$4F,$8A,$35,$1D
	.dc.b	$27,$89,$27,$4D,$8E,$53,$C1,$E6,$62,$34,$F9,$23,$44,$8C,$C7,$90
	.dc.b	$D3,$87,$C5,$3E,$02,$3D,$5A,$F0,$6B,$CC,$87,$23,$81,$0E,$05,$36
	.dc.b	$7A,$8C,$90,$CC,$FA,$63,$BC,$75,$34,$53,$73,$EB,$7A,$18,$8E,$2E
	.dc.b	$0E,$04,$30,$49,$F0,$3E,$73,$4E,$1F,$09,$B9,$F0,$46,$63,$C0,$71
	.dc.b	$35,$3E,$28,$D4,$74,$9E,$24,$9D,$36,$39,$4F,$07,$99,$88,$D3,$E4
	.dc.b	$8D,$12,$33,$1E,$43,$4E,$1F,$14,$F8,$08,$F5,$6B,$C1,$AF,$32,$1C
	.dc.b	$8E,$04,$38,$14,$D9,$EA,$32,$43,$33,$E9,$8E,$F1,$D4,$D1,$4D,$CF
	.dc.b	$AD,$E8,$62,$38,$B8,$38,$10,$C1,$27,$C0,$F9,$CD,$38,$7C,$26,$E7
	.dc.b	$C1,$97,$F3,$E8,$FF,$3D,$B5,$FE,$71,$8F,$F9,$E2,$31,$1D,$27,$D1
	.dc.b	$5D,$63,$D0,$78,$4E,$D7,$EA,$8D,$47,$18,$C4,$F2,$23,$82,$47,$39
	.dc.b	$E2,$F4,$31,$3C,$1B,$3C,$1F,$7C,$F0,$24,$EE,$32,$D7,$D2,$62,$79
	.dc.b	$38,$21,$E9,$3B,$90,$E0,$53,$67,$9C,$C9,$0C,$CF,$92,$3B,$C7,$53
	.dc.b	$45,$37,$3E,$77,$04,$3B,$1A,$8E,$6E,$0D,$4F,$77,$CC,$62,$7A,$3B
	.dc.b	$24,$6D,$F1,$98,$8D,$C6,$63,$C0,$74,$35,$3E,$28,$D4,$77,$8C,$4F
	.dc.b	$22,$38,$24,$73,$9E,$2F,$43,$13,$C1,$B3,$C1,$F3,$D7,$F8,$5F,$1F
	.dc.b	$CD,$C7,$F8,$44,$FF,$0F,$B8,$CB,$FC,$31,$19,$9E,$0F,$F8,$9A,$74
	.dc.b	$7F,$8C,$DF,$F1,$E1,$FE,$41,$C3,$FC,$6E,$DD,$3F,$C6,$8E,$8F,$F1
	.dc.b	$FE,$9F,$E3,$67,$4D,$72,$6E,$79,$B8,$3F,$C8,$11,$B8,$C9,$27,$BC
	.dc.b	$E8,$E4,$E1,$F4,$19,$21,$82,$1E,$43,$8C,$F7,$8E,$51,$93,$89,$4E
	.dc.b	$27,$9C,$B3,$E5,$9E,$44,$8E,$A4,$39,$9C,$4E,$84,$30,$48,$E8,$F1
	.dc.b	$3B,$CF,$99,$D3,$63,$71,$D9,$AF,$09,$F2,$39,$9E,$04,$8C,$C6,$CF
	.dc.b	$01,$E5,$9E,$6D,$9E,$0E,$8F,$1C,$F2,$74,$D7,$26,$E7,$9B,$83,$C9
	.dc.b	$1B,$8C,$92,$7B,$CE,$8F,$C8,$5F,$F3,$84,$D2,$69,$FE,$74,$8F,$DE
	.dc.b	$D7,$F9,$2B,$76,$FF,$E8,$99,$4E,$63,$1C,$2E,$07,$0F,$FA,$3F,$0E
	.dc.b	$38,$7F,$9B,$5A,$F2,$E0,$70,$6F,$1E,$D9,$75,$AF,$F4,$C6,$5D,$71
	.dc.b	$D4,$73,$AD,$E1,$CA,$18,$F6,$C3,$94,$30,$E4,$EA,$F8,$E3,$D2,$3C
	.dc.b	$8D,$AE,$35,$C2,$FA,$61,$DD,$B5,$B8,$C6,$1F,$13,$2F,$87,$8F,$73
	.dc.b	$D0,$43,$B9,$93,$CC,$F0,$BE,$B1,$9C,$7D,$67,$92,$75,$5E,$A9,$D5
	.dc.b	$F8,$9E,$D8,$FA,$67,$58,$74,$23,$E7,$CB,$D5,$94,$BE,$6F,$C5,$BE
	.dc.b	$51,$E9,$CB,$C7,$8F,$78,$F4,$DF,$1C,$7A,$C7,$8B,$2F,$3B,$C6,$B8
	.dc.b	$5F,$AE,$F8,$65,$E6,$25,$7E,$0A,$F8,$BF,$23,$0E,$AD,$9F,$53,$E8
	.dc.b	$24,$6B,$1F,$4C,$6C,$F1,$65,$ED,$BE,$B8,$FB,$23,$C7,$3E,$1C,$6B
	.dc.b	$C4,$97,$D2,$B8,$47,$83,$1F,$1E,$3E,$69,$CB,$FA,$2A,$E1,$97,$C2
	.dc.b	$26,$3E,$3C,$7B,$DF,$0C,$39,$9F,$13,$1F,$14,$7C,$97,$C0,$7B,$0F
	.dc.b	$C6,$61,$DA,$F9,$E1,$F5,$E3,$13,$C0,$F5,$E5,$F0,$A7,$C7,$19,$9D
	.dc.b	$E5,$EB,$9E,$55,$BB,$EA,$7A,$6F,$A6,$1E,$97,$E5,$18,$AE,$B7,$F0
	.dc.b	$EF,$94,$E6,$F9,$4F,$77,$E8,$E1,$E2,$C7,$E1,$63,$F0,$6B,$57,$E7
	.dc.b	$C3,$9C,$60,$ED,$97,$9A,$74,$D8,$FC,$34,$F0,$BE,$B9,$7B,$AB,$51
	.dc.b	$F4,$71,$F1,$61,$F3,$F0,$EA,$FA,$4E,$A5,$7A,$D7,$28,$FD,$23,$5F
	.dc.b	$A2,$DC,$7D,$97,$E2,$CB,$E0,$E5,$E5,$BF,$04,$FE,$11,$A7,$43,$C0
	.dc.b	$F9,$E3,$84,$7E,$23,$2F,$63,$A3,$86,$3F,$1E,$B8,$63,$E3,$AD,$CE
	.dc.b	$6B,$75,$E5,$7D,$F1,$97,$D6,$75,$2B,$9C,$7D,$B1,$E8,$33,$3C,$E7
	.dc.b	$E3,$C7,$19,$DB,$E9,$C7,$E1,$E1,$E5,$7F,$0F,$7C,$4F,$4C,$F1,$C3
	.dc.b	$D3,$19,$8F,$49,$E1,$BE,$A7,$E5,$DF,$19,$99,$E2,$DA,$EE,$FF,$2E
	.dc.b	$DB,$F9,$FF,$1F,$F4,$0E,$27,$2B,$93,$3A,$7F,$68,$BB,$82,$F7,$58
	.dc.b	$8F,$EB,$13,$FD,$29,$FE,$E4,$62,$AB,$FD,$9A,$34,$F5,$8C,$BB,$18
	.dc.b	$73,$DD,$5E,$2C,$33,$CB,$E1,$9C,$E0,$C3,$8B,$FE,$03,$EB,$74,$71
	.dc.b	$D1,$D1,$E9,$9C,$D7,$1D,$3E,$25,$6A,$30,$73,$BF,$4E,$1D,$EF,$8B
	.dc.b	$EE,$6B,$F0,$9A,$F1,$9E,$6F,$19,$F6,$5B,$A7,$C9,$12,$B7,$4B,$EF
	.dc.b	$7F,$64,$E1,$3A,$8C,$BF,$07,$AF,$E0,$75,$FB,$DB,$1B,$7D,$77,$D5
	.dc.b	$B4,$69,$27,$81,$EA,$DB,$EB,$F6,$75,$8F,$3F,$6B,$BB,$AB,$63,$EF
	.dc.b	$B7,$1F,$C2,$B5,$FC,$4F,$7B,$D6,$F2,$B6,$BF,$0E,$D9,$FC,$6B,$5F
	.dc.b	$C5,$F5,$7E,$0F,$95,$E8,$7E,$4A,$30,$D9,$FC,$6B,$67,$71,$F8,$BE
	.dc.b	$E7,$85,$F9,$48,$D9,$97,$13,$C2,$3F,$CB,$31,$8A,$FF,$2A,$57,$2E
	.dc.b	$2F,$0D,$7E,$5F,$C3,$78,$5F,$A2,$84,$61,$FD,$2B,$5F,$D5,$B5,$FD
	.dc.b	$7F,$88,$EB,$E4,$D8,$FD,$6B,$5F,$D9,$B5,$FD,$BF,$8C,$E7,$FF,$35
	.dc.b	$63,$FB,$B8,$C4,$FE,$49,$AF,$EB,$8F,$E0,$9B,$7F,$9B,$5D,$57,$AE
	.dc.b	$77,$1D,$A3,$04,$7F,$15,$EA,$3C,$7F,$0F,$17,$EE,$F9,$D3,$9A,$EA
	.dc.b	$78,$EB,$6F,$F0,$1C,$34,$F9,$43,$8C,$BF,$90,$64,$F3,$78,$D6,$AE
	.dc.b	$DE,$F2,$FF,$46,$36,$7D,$F5,$AA,$F7,$12,$3F,$8D,$CF,$89,$F9,$B1
	.dc.b	$E6,$CB,$D7,$E4,$DC,$CF,$3A,$F9,$DE,$5F,$88,$8F,$EB,$F3,$79,$78
	.dc.b	$8F,$E5,$B9,$2B,$E8,$12,$3C,$6F,$7F,$32,$47,$63,$F8,$6F,$8E,$F2
	.dc.b	$FF,$97,$C7,$F9,$ED,$B8,$FF,$A5,$DB,$E4,$FF,$AB,$FC,$DF,$95,$5F
	.dc.b	$E9,$39,$CB,$FC,$CE,$FA,$75,$E4,$FF,$4D,$57,$09,$ED,$19,$69,$FE
	.dc.b	$8F,$8D,$61,$E1,$9D,$36,$75,$C3,$FE,$C1,$8C,$13,$1F,$F5,$BE,$3E
	.dc.b	$4A,$61,$D2,$73,$87,$46,$CF,$43,$07,$AE,$B9,$63,$EC,$8D,$1E,$8C
	.dc.b	$3C,$0E,$EF,$DF,$8F,$97,$0E,$B5,$AA,$FF,$6B,$E1,$E5,$C7,$C9,$1C
	.dc.b	$4E,$95,$C0,$FA,$17,$CE,$35,$7C,$F1,$F2,$3F,$ED,$AB,$E7,$3D,$B1
	.dc.b	$F5,$B4,$F4,$CF,$1C,$7D,$A7,$8F,$0E,$D3,$B9,$D5,$7B,$23,$9E,$1E
	.dc.b	$0C,$7C,$D1,$FD,$CF,$1F,$0C,$E6,$3C,$F3,$A7,$E9,$E1,$DA,$75,$1D
	.dc.b	$E7,$55,$CB,$1F,$2E,$1E,$A8,$C3,$E2,$8F,$45,$58,$D6,$1D,$67,$8C
	.dc.b	$FC,$C3,$CD,$19,$9E,$11,$F1,$2F,$CD,$87,$90,$F2,$E1,$E0,$9E,$47
	.dc.b	$9C,$E6,$72,$C7,$EF,$46,$DF,$93,$1D,$4F,$86,$43,$91,$F1,$0E,$55
	.dc.b	$F0,$B0,$F3,$47,$82,$FC,$71,$8A,$EA,$7E,$05,$C5,$7A,$B1,$F8,$B7
	.dc.b	$C5,$F8,$75,$B8,$EE,$78,$F0,$F4,$57,$1C,$7F,$01,$3C,$5F,$9A,$F0
	.dc.b	$3D,$33,$C1,$F6,$63,$E6,$74,$D7,$91,$F2,$A3,$83,$F0,$A3,$57,$EC
	.dc.b	$BF,$85,$3C,$B0,$F5,$CE,$63,$D1,$8F,$C0,$7C,$33,$AA,$F5,$47,$79
	.dc.b	$D3,$DE,$74,$F0,$C7,$DA,$F9,$AF,$ED,$D7,$0C,$3F,$09,$7C,$2B,$95
	.dc.b	$F0,$BF,$24,$7C,$4B,$E1,$5F,$43,$0F,$DD,$CF,$AE,$FE,$6D,$F0,$C3
	.dc.b	$E0,$D7,$A2,$3C,$35,$B3,$F1,$47,$80,$F6,$19,$7E,$E3,$E9,$8F,$92
	.dc.b	$D3,$E3,$9B,$7C,$6E,$4F,$41,$93,$F3,$A6,$67,$CA,$FC,$73,$C2,$66
	.dc.b	$FE,$24,$7C,$6B,$F8,$95,$EB,$3C,$77,$EF,$AD,$9F,$16,$FC,$A7,$3A
	.dc.b	$F8,$B5,$C6,$33,$5F,$D0,$DF,$F4,$8D,$FF,$35,$6D,$FF,$5C,$6D,$FF
	.dc.b	$A4,$3A,$5E,$67,$FA,$76,$7F,$D2,$F1,$8E,$2D,$4E,$67,$FD,$3B,$39
	.dc.b	$7F,$D3,$B8,$3C,$B7,$97,$FE,$9C,$8C,$36,$33,$3C,$1A,$FA,$1C,$E7
	.dc.b	$FE,$CC,$3F,$DA,$4F,$FA,$F2,$3F,$D9,$87,$FB,$6B,$4F,$F6,$33,$FE
	.dc.b	$C2,$33,$3C,$63,$37,$FE,$C8,$AE,$B1,$C3,$CA,$79,$9F,$EF,$0A,$FF
	.dc.b	$66,$46,$3A,$3F,$DB,$18,$B1,$61,$FE,$FB,$8E,$EF,$FB,$6A,$3C,$F9
	.dc.b	$7F,$B4,$27,$27,$FB,$68,$A7,$9F,$0F,$F6,$D1,$E5,$8F,$F5,$D4,$FF
	.dc.b	$BD,$70,$F1,$57,$FB,$96,$7A,$57,$0C,$BE,$23,$63,$31,$EE,$9F,$F7
	.dc.b	$C5,$72,$BF,$2C,$7A,$F0,$F5,$61,$C6,$30,$7F,$B2,$EB,$84,$FC,$AB
	.dc.b	$EE,$75,$9F,$5C,$E9,$FF,$78,$1C,$83,$97,$E0,$D7,$BD,$D3,$F4,$32
	.dc.b	$FA,$55,$D6,$35,$87,$C3,$9C,$E3,$F1,$B1,$FA,$07,$95,$F8,$26,$70
	.dc.b	$F8,$33,$C1,$FC,$41,$E6,$AF,$B6,$FB,$E3,$51,$F8,$98,$D5,$7C,$6C
	.dc.b	$3C,$84,$BE,$F3,$EA,$8C,$3F,$07,$1F,$81,$8F,$CD,$C3,$BC,$FC,$4C
	.dc.b	$7B,$E3,$EA,$8E,$E7,$8F,$2F,$C8,$D7,$0C,$39,$E3,$DA,$3C,$D5,$F3
	.dc.b	$23,$EA,$E5,$F0,$63,$EA,$E3,$D9,$FC,$96,$5E,$3C,$BD,$95,$B8,$FA
	.dc.b	$59,$78,$27,$35,$ED,$AD,$61,$DA,$30,$7C,$C3,$C5,$1F,$13,$1F,$B1
	.dc.b	$7C,$E7,$D8,$7E,$6C,$F8,$D8,$F9,$5F,$D2,$5F,$B7,$0F,$89,$1D,$0D
	.dc.b	$D7,$37,$E5,$37,$2F,$8B,$8F,$98,$93,$C0,$F0,$63,$F2,$F1,$F0,$65
	.dc.b	$F8,$EC,$7F,$82,$46,$32,$F7,$57,$AA,$F8,$D7,$C4,$8C,$1F,$AC,$3A
	.dc.b	$65,$E3,$BE,$58,$FC,$3A,$EB,$39,$3C,$B5,$F8,$4F,$1D,$E4,$25,$8F
	.dc.b	$FB,$3F,$77,$94,$F2,$F5,$3F,$EB,$FC,$38,$DF,$0E,$DC,$E1,$C2,$F7
	.dc.b	$1C,$2D,$7C,$1E,$E4,$70,$E0,$DE,$3C,$5D,$35,$F3,$C7,$39,$EC,$DC
	.dc.b	$3F,$DB,$90,$F2,$FC,$71,$C8,$F4,$1E,$1F,$2F,$36,$CB,$E0,$F8,$18
	.dc.b	$8F,$4C,$EA,$73,$5C,$A7,$C7,$3E,$57,$71,$EB,$72,$D7,$26,$71,$4F
	.dc.b	$76,$CF,$5A,$62,$F2,$F5,$72,$BC,$58,$74,$AF,$A6,$6E,$33,$87,$49
	.dc.b	$FB,$5E,$7F,$AF,$67,$9F,$E1,$8C,$36,$FD,$16,$9F,$73,$63,$31,$83
	.dc.b	$85,$FE,$2F,$07,$87,$87,$F0,$9E,$53,$CB,$F9,$0D,$C3,$E3,$36,$FE
	.dc.b	$53,$5F,$C3,$7A,$0F,$47,$E8,$36,$BE,$73,$67,$E9,$35,$FA,$5C,$BF
	.dc.b	$A0,$EE,$79,$9F,$5E,$30,$D9,$4E,$9F,$CB,$4E,$9F,$D4,$F7,$3C,$DF
	.dc.b	$AF,$1C,$2F,$F3,$F3,$9D,$78,$4A,$30,$EC,$E7,$5C,$B0,$F0,$47,$F8
	.dc.b	$87,$D2,$FC,$53,$4F,$C6,$77,$23,$2D,$8F,$C8,$79,$AF,$3B,$8B,$63
	.dc.b	$31,$83,$AF,$92,$F5,$3F,$6F,$A7,$EE,$E3,$15,$F9,$06,$BF,$C2,$76
	.dc.b	$7A,$9C,$38,$7F,$7F,$5A,$AF,$DF,$35,$CE,$1E,$02,$3F,$A1,$F3,$BF
	.dc.b	$40,$DE,$CE,$8F,$F1,$8A,$F0,$E5,$E5,$EE,$CC,$A5,$18,$6C,$7E,$B7
	.dc.b	$3F,$F1,$AB,$EB,$9D,$11,$D7,$89,$5F,$DE,$57,$26,$CF,$C8,$23,$E9
	.dc.b	$E9,$F3,$F7,$B3,$FD,$1C,$F7,$3E,$54,$F9,$48,$F9,$7C,$9A,$F8,$23
	.dc.b	$EA,$91,$FE,$19,$87,$47,$F9,$D5,$3C,$EF,$0B,$63,$B9,$FC,$23,$C7
	.dc.b	$FF,$44,$3F,$E8,$CD,$3F,$D0,$DC,$DF,$E8,$28,$FF,$48,$47,$FA,$91
	.dc.b	$FF,$43,$1F,$EA,$49,$E3,$19,$38,$C6,$7C,$AF,$F4,$73,$FE,$9E,$38
	.dc.b	$C7,$FA,$A0,$D9,$7C,$7F,$F4,$84,$6E,$33,$1F,$EA,$86,$C7,$FA,$99
	.dc.b	$AF,$FB,$EC,$EB,$1B,$39,$D7,$FA,$AA,$3F,$D7,$53,$E7,$9C,$D7,$FA
	.dc.b	$92,$FF,$DD,$DE,$5F,$FB,$9B,$0F,$6D,$73,$C3,$E1,$BF,$EA,$C3,$8D
	.dc.b	$7C,$78,$EB,$19,$76,$7F,$AA,$9D,$BF,$16,$3B,$EB,$FE,$D8,$7F,$D7
	.dc.b	$31,$C0,$F8,$53,$C3,$1F,$8D,$19,$BF,$61,$E6,$7B,$19,$C7,$D8,$FF
	.dc.b	$B3,$F1,$F6,$47,$D4,$6E,$1E,$09,$F4,$9C,$8A,$78,$8E,$8F,$BE,$BC
	.dc.b	$11,$93,$6F,$92,$BF,$02,$D9,$F7,$CF,$FB,$7C,$FB,$17,$F6,$AF,$C7
	.dc.b	$1F,$93,$C3,$E0,$1F,$45,$EB,$87,$B2,$BF,$E0,$0D,$70,$7E,$E4,$F9
	.dc.b	$8F,$9C,$7C,$1C,$7E,$73,$D4,$D4,$7C,$C7,$AE,$1E,$DB,$F4,$9D,$F0
	.dc.b	$FC,$AE,$1D,$DF,$BD,$5C,$0E,$F7,$E4,$C3,$BB,$E4,$8D,$3D,$F1,$F9
	.dc.b	$27,$C5,$8F,$31,$F1,$6F,$B3,$F4,$E3,$78,$78,$E3,$F0,$D3,$A7,$E9
	.dc.b	$4E,$B1,$F5,$1F,$42,$3F,$4A,$7E,$A7,$1F,$9A,$D3,$E1,$BF,$3E,$7F
	.dc.b	$DF,$D1,$FE,$38,$6B,$FE,$48,$6A,$6F,$5F,$FB,$E6,$3F,$DF,$7E,$5F
	.dc.b	$FB,$F5,$FE,$C6,$D7,$FB,$4B,$5E,$5D,$5F,$EF,$87,$C0,$D7,$C0,$D7
	.dc.b	$7E,$57,$FC,$01,$3F,$CB,$F8,$7F,$9A,$4F,$F3,$1C,$7F,$9B,$0F,$F4
	.dc.b	$2E,$9F,$E6,$B8,$D4,$7F,$AE,$BA,$3F,$CD,$0E,$CF,$F4,$09,$7C,$BF
	.dc.b	$F4,$0B,$FE,$94,$AF,$F5,$2E,$7F,$E9,$F8,$EB,$7F,$EC,$77,$FD,$09
	.dc.b	$3D,$DC,$9F,$EB,$23,$AB,$96,$C7,$18,$FF,$3F,$C7,$87,$9B,$FC,$FE
	.dc.b	$F4,$9D,$1E,$A7,$FD,$67,$87,$AA,$BA,$61,$E8,$7F,$D1,$75,$C2,$BC
	.dc.b	$67,$FA,$86,$3B,$9C,$2F,$BB,$89,$E1,$3E,$63,$D5,$5D,$63,$E7,$CF
	.dc.b	$1B,$F1,$47,$09,$D5,$7C,$F3,$D9,$3A,$79,$DF,$82,$BD,$98,$7F,$B2
	.dc.b	$27,$64,$C3,$B3,$67,$B5,$71,$8F,$6D,$7D,$57,$D0,$7B,$E7,$AE,$5F
	.dc.b	$EC,$38,$DB,$5F,$F6,$3E,$5F,$8C,$7F,$DB,$B1,$9A,$FF,$5E,$47,$AF
	.dc.b	$0F,$19,$2F,$C7,$8F,$B6,$7C,$07,$6C,$3A,$3D,$8F,$C6,$4F,$FB,$52
	.dc.b	$FD,$19,$7C,$89,$FB,$A7,$BC,$F8,$F7,$D6,$77,$97,$C1,$9F,$7C,$62
	.dc.b	$75,$97,$A4,$FF,$68,$E1,$F6,$0F,$3D,$75,$3C,$E7,$28,$63,$FF,$00
	.dc.b	$7D,$1F,$F4,$14,$FF,$A5,$AF,$FD,$96,$3D,$5F,$F2,$53,$FE,$96,$9D
	.dc.b	$F4,$FF,$B3,$23,$5D,$7F,$EA,$F9,$CC,$72,$AE,$37,$D4,$FF,$6B,$CE
	.dc.b	$9F,$F6,$D4,$6A,$FF,$DA,$13,$9A,$EA,$78,$7C,$6F,$F5,$FB,$8B,$E8
	.dc.b	$25,$E9,$FF,$B1,$32,$7A,$CD,$7C,$D3,$98,$C6,$1E,$D8,$C4,$EA,$3F
	.dc.b	$D8,$D5,$CA,$31,$87,$AE,$31,$3C,$4D,$DF,$49,$F2,$E1,$E4,$9F,$F6
	.dc.b	$53,$63,$FD,$9C,$D7,$FD,$D4,$D7,$FD,$DF,$E3,$FF,$AA,$F3,$FF,$7A
	.dc.b	$F9,$7F,$26,$53,$9B,$FF,$7E,$4E,$63,$13,$FE,$D2,$8C,$79,$3F,$EE
	.dc.b	$8B,$46,$2F,$FD,$73,$3F,$EB,$C2,$3F,$EE,$1F,$65,$ED,$FF,$BC,$3D
	.dc.b	$B7,$B6,$F6,$FF,$DE,$15,$AB,$FF,$77,$56,$A7,$38,$7D,$5F,$73,$F1
	.dc.b	$FA,$7D,$AC,$3C,$FC,$09,$CE,$AC,$BF,$D7,$9D,$BF,$EA,$CC,$03,$AF
	.dc.b	$FD,$6A,$E1,$AE,$4F,$F5,$B1,$FE,$B9,$F3,$3F,$D6,$DE,$6F,$FA,$F9
	.dc.b	$B1,$FE,$BA,$C3,$FD,$7F,$DF,$FE,$BE,$C3,$FD,$77,$EF,$7F,$AC,$6B
	.dc.b	$F0,$5E,$F7,$95,$FC,$3B,$5F,$CA,$C6,$EF,$0B,$E0,$3F,$D4,$7C,$5F
	.dc.b	$2B,$E0,$2F,$21,$FF,$51,$D6,$F0,$F7,$E1,$D7,$D0,$F1,$47,$C2,$3E
	.dc.b	$1B,$87,$0D,$73,$E8,$7F,$B1,$DE,$7C,$3C,$7C,$97,$C0,$FF,$AC,$9B
	.dc.b	$3A,$38,$94,$5F,$B6,$B9,$57,$13,$FD,$77,$19,$9F,$AF,$EF,$3D,$FF
	.dc.b	$F7,$1B,$63,$D4,$7A,$DC,$38,$0E,$1D,$BA,$CB,$E8,$79,$EF,$82,$FF
	.dc.b	$65,$B6,$BD,$15,$BC,$BB,$61,$AB,$95,$A8,$FF,$69,$D6,$E3,$F1,$F7
	.dc.b	$CD,$FE,$22,$DC,$3F,$01,$3B,$9F,$3F,$8D,$E2,$F8,$39,$1E,$28,$EA
	.dc.b	$61,$E7,$18,$74,$7C,$FB,$46,$EF,$BD,$7F,$BA,$3C,$9F,$F2,$45,$FF
	.dc.b	$96,$BC,$5F,$F2,$FF,$9F,$FE,$5A,$7F,$CB,$F3,$AE,$CF,$F2,$D7,$9F
	.dc.b	$FE,$66,$EA,$FF,$27,$E1,$FE,$5A,$3F,$C9,$E7,$F9,$FA,$FF,$C9,$ED
	.dc.b	$3F,$C9,$E7,$F9,$44,$FF,$32,$BF,$E5,$19,$E9,$1E,$33,$FC,$BF,$87
	.dc.b	$8D,$FF,$32,$46,$AF,$C7,$7F,$E5,$F7,$FD,$19,$5B,$AF,$F3,$2D,$6E
	.dc.b	$FB,$5E,$DF,$2F,$9F,$FE,$6B,$C7,$FC,$E9,$8F,$F9,$EF,$0F,$F3,$A6
	.dc.b	$3C,$E7,$37,$F1,$A7,$24,$3E,$0D,$F3,$C7,$FC,$F1,$1F,$EA,$C8,$C6
	.dc.b	$3F,$23,$1E,$D3,$C3,$1E,$D7,$C3,$1F,$F5,$7E,$1C,$2B,$E2,$1C,$EB
	.dc.b	$78,$73,$C7,$C9,$3C,$CF,$26,$3F,$11,$B1,$F0,$2B,$8C,$74,$C3,$BC
	.dc.b	$62,$BE,$6C,$FD,$D7,$93,$D4,$8E,$09,$5D,$63,$B9,$F8,$03,$F0,$47
	.dc.b	$E0,$0F,$C1,$57,$E3,$63,$EA,$C7,$77,$93,$D9,$B3,$DA,$3A,$61,$F1
	.dc.b	$63,$CE,$7C,$C3,$51,$DC,$F0,$D7,$77,$C3,$8F,$E1,$5B,$8F,$E1,$5E
	.dc.b	$B8,$FF,$27,$BF,$F0,$24,$7F,$A8,$E9,$5F,$C3,$E3,$FD,$49,$18,$7F
	.dc.b	$89,$5F,$FA,$A0,$FF,$53,$F9,$FF,$EA,$1E,$DF,$F5,$44,$FF,$A9,$BD
	.dc.b	$2E,$51,$88,$FF,$50,$F8,$5F,$E9,$F7,$9F,$9D,$FE,$A3,$C0,$D3,$FC
	.dc.b	$B5,$E8,$7F,$92,$63,$FC,$D5,$A7,$F9,$56,$BF,$C9,$27,$F8,$FE,$3F
	.dc.b	$C8,$CD,$AF,$F2,$35,$FF,$63,$DF,$FC,$CE,$FF,$96,$30,$E9,$3D,$BD
	.dc.b	$0F,$F2,$84,$62,$75,$18,$27,$47,$F9,$4A,$3F,$CC,$38,$FF,$9A,$B1
	.dc.b	$FF,$2E,$E5,$E5,$7F,$CA,$38,$73,$7F,$CC,$A7,$F9,$A4,$A7,$F9,$A6
	.dc.b	$3F,$D3,$AF,$B1,$E5,$3E,$C6,$E3,$EC,$9F,$F5,$51,$BC,$BF,$CC,$33
	.dc.b	$AC,$3C,$B5,$FE,$66,$AF,$F3,$F5,$68,$8F,$12,$1E,$8C,$BD,$EF,$CB
	.dc.b	$AE,$67,$CB,$7D,$57,$E8,$78,$65,$F1,$5F,$6E,$3F,$14,$FB,$38,$78
	.dc.b	$23,$67,$53,$78,$F9,$DF,$97,$1B,$9F,$84,$D9,$F8,$57,$FE,$BD,$AF
	.dc.b	$4D,$69,$B3,$AA,$FC,$0C,$7D,$0C,$BE,$A5,$FF,$B8,$AB,$53,$EC,$AF
	.dc.b	$F4,$63,$CB,$1F,$82,$FD,$03,$D3,$5F,$3D,$F4,$C7,$C5,$7F,$D2,$0F
	.dc.b	$08,$F3,$47,$09,$F5,$4E,$5F,$2C,$F8,$E3,$FD,$2D,$E7,$FF,$A0,$3D
	.dc.b	$0F,$F4,$36,$7F,$E1,$FF,$3B,$FD,$0D,$18,$C7,$FD,$1B,$5F,$E3,$17
	.dc.b	$4D,$73,$5F,$E7,$28,$C4,$7F,$A2,$5F,$F4,$D1,$FE,$85,$F8,$4B,$3F
	.dc.b	$E7,$D9,$FF,$40,$7A,$1F,$E7,$DE,$8F,$F4,$BF,$A3,$FE,$88,$AD,$57
	.dc.b	$F9,$EA,$3A,$E5,$FE,$7B,$8F,$F5,$0C,$60,$FF,$45,$D7,$F9,$27,$CC
	.dc.b	$7C,$33,$E1,$B3,$8F,$BE,$EE,$0B,$DD,$62,$3F,$D0,$7E,$7F,$F9,$F7
	.dc.b	$D2,$FF,$3E,$47,$FA,$C7,$2F,$F3,$FE,$DF,$E5,$77,$FC,$AD,$E9,$AF
	.dc.b	$FC,$FF,$5F,$E2,$F8,$C6,$7F,$E8,$42,$47,$FA,$C2,$3F,$D1,$B9,$7A
	.dc.b	$3C,$6E,$95,$FE,$98,$BF,$F4,$1F,$A7,$F3,$AB,$FD,$09,$87,$F8,$ED
	.dc.b	$AF,$F9,$01,$AF,$F9,$6D,$F0,$BB,$9E,$44,$AF,$E6,$D1,$FE,$7F,$3F
	.dc.b	$CF,$A5,$EE,$FF,$43,$3F,$E8,$16,$F2,$FF,$9F,$3D,$0F,$F3,$FC,$FF
	.dc.b	$A0,$F4,$C1,$1F,$F4,$3F,$97,$FE,$69,$4D,$27,$97,$E2,$7F,$CE,$D1
	.dc.b	$FE,$57,$6B,$FE,$5C,$6B,$FE,$77,$6B,$FE,$6C,$3F,$DB,$7E,$57,$F9
	.dc.b	$9D,$FF,$36,$CF,$F9,$CF,$1F,$F3,$5B,$FE,$79,$8F,$F3,$24,$F9,$49
	.dc.b	$1F,$E7,$18,$FF,$3B,$79,$7F,$E7,$A7,$AC,$7F,$97,$3C,$A9,$1F,$E5
	.dc.b	$F3,$FC,$9D,$F0,$58,$7F,$DD,$BE,$5F,$F9,$6C,$FF,$2E,$79,$69,$D4
	.dc.b	$FF,$9C,$AB,$FC,$F6,$47,$FB,$55,$70,$F3,$3F,$CD,$B1,$D2,$7F,$CB
	.dc.b	$CF,$6F,$2E,$ED,$FE,$7A,$F8,$0C,$CF,$F9,$FA,$7F,$D0,$E4,$7A,$BF
	.dc.b	$E6,$3E,$2F,$F5,$16,$7F,$E6,$69,$FF,$34,$8B,$FF,$5A,$78,$BF,$E0
	.dc.b	$78,$FF,$04,$F9,$5F,$E0,$76,$F0,$FF,$81,$DF,$F0,$77,$99,$FE,$13
	.dc.b	$F1,$7F,$C3,$7C,$DF,$E1,$37,$FC,$32,$49,$FF,$06,$9F,$E2,$63,$8F
	.dc.b	$99,$FE,$31,$8E,$27,$F8,$F2,$7F,$C3,$78,$F8,$EF,$B5,$FF,$86,$0F
	.dc.b	$F1,$41,$D0,$FF,$14,$16,$78,$E3,$E9,$9E,$33,$EA,$38,$9E,$23,$C9
	.dc.b	$5D,$F0,$F2,$47,$7B,$F6,$3D,$70,$F6,$3C,$4F,$09,$63,$C3,$8F,$C7
	.dc.b	$BE,$93,$DB,$1F,$26,$1D,$A3,$C9,$7D,$9E,$58,$76,$BE,$31,$D9,$B3
	.dc.b	$DA,$3D,$71,$89,$E1,$7D,$A7,$C1,$87,$8E,$31,$1C,$DB,$7C,$CF,$15
	.dc.b	$F3,$9F,$80,$43,$CA,$57,$4E,$08,$E0,$93,$DA,$FC,$B7,$E3,$25,$7F
	.dc.b	$92,$A7,$FC,$98,$FF,$B8,$2B,$FC,$99,$18,$8F,$F2,$51,$FE,$E1,$F2
	.dc.b	$BF,$C9,$49,$A6,$27,$FC,$93,$E3,$FF,$93,$67,$A7,$91,$FE,$4B,$8F
	.dc.b	$F2,$86,$5F,$DB,$67,$FB,$B9,$FD,$EE,$BF,$C1,$0F,$F7,$53,$FC,$00
	.dc.b	$6A,$3F,$BC,$C6,$89,$5F,$E0,$89,$FE,$FE,$6F,$2F,$EF,$61,$FF,$30
	.dc.b	$9D,$E3,$87,$93,$FE,$12,$9D,$4F,$F9,$0D,$FF,$08,$BE,$32,$57,$83
	.dc.b	$0F,$F0,$44,$7F,$85,$CF,$F1,$E9,$E8,$C3,$FC,$69,$3B,$8F,$21,$B3
	.dc.b	$FC,$5A,$6E,$BB,$C6,$E7,$8E,$1E,$6F,$2B,$FC,$87,$1F,$E1,$F3,$38
	.dc.b	$78,$B0,$F8,$31,$87,$B9,$8C,$3A,$DF,$3E,$1F,$F2,$35,$6C,$93,$AB
	.dc.b	$E9,$97,$C5,$C3,$E4,$D7,$D2,$C3,$94,$7F,$92,$63,$8B,$F0,$8F,$8F
	.dc.b	$1A,$8F,$45,$F5,$9F,$2D,$75,$7E,$81,$C7,$2F,$85,$3B,$AE,$B3,$F2
	.dc.b	$0E,$B8,$79,$B2,$F8,$B5,$FE,$3C,$9E,$B1,$87,$E7,$1D,$71,$F2,$4E
	.dc.b	$EF,$AE,$1F,$E6,$6A,$D9,$DD,$FC,$EA,$8F,$F5,$CB,$FE,$39,$CB,$F1
	.dc.b	$DC,$07,$FC,$82,$60,$FC,$96,$4F,$21,$E4,$FF,$8F,$AF,$FA,$35,$70
	.dc.b	$69,$FE,$48,$3F,$C8,$31,$BB,$ED,$19,$6B,$E3,$37,$3E,$17,$4F,$F8
	.dc.b	$EC,$FF,$26,$47,$F9,$07,$C8,$FF,$20,$46,$0D,$4F,$88,$79,$2F,$27
	.dc.b	$FC,$5F,$1F,$E4,$3A,$E2,$EE,$75,$5C,$E3,$FC,$A2,$F8,$FC,$3F,$F2
	.dc.b	$4B,$98,$EB,$7F,$E4,$E9,$CC,$73,$8E,$EF,$F9,$3A,$7F,$C8,$6F,$EE
	.dc.b	$0C,$1F,$E4,$9F,$0F,$FC,$97,$8F,$F9,$39,$F3,$DF,$F9,$1E,$BE,$25
	.dc.b	$70,$AD,$46,$3C,$6F,$F2,$6E,$4F,$1B,$C3,$1F,$E4,$4A,$F9,$F8,$7F
	.dc.b	$93,$1B,$E3,$7C,$4D,$DE,$A3,$E1,$FF,$CA,$79,$F8,$30,$E4,$FF,$93
	.dc.b	$F5,$F9,$7E,$93,$E2,$5F,$13,$F0,$2F,$7C,$1E,$FE,$5F,$E3,$9D,$BF
	.dc.b	$E4,$C8,$FF,$28,$9F,$E4,$FF,$5F,$E2,$34,$FE,$45,$DA,$F3,$3F,$CA
	.dc.b	$2D,$E2,$FF,$28,$5A,$FF,$91,$CE,$6F,$FC,$A7,$E4,$FF,$26,$BF,$C7
	.dc.b	$F9,$9F,$E4,$71,$19,$E0,$FE,$61,$BB,$E2,$9F,$0B,$9C,$7F,$9A,$61
	.dc.b	$FE,$58,$C7,$9F,$B1,$99,$FF,$2F,$BB,$E6,$7C,$5B,$A3,$FC,$A7,$5F
	.dc.b	$6B,$D0,$FF,$2C,$1F,$E5,$9F,$43,$FA,$EE,$3F,$A4,$8C,$EF,$C2,$BF
	.dc.b	$39,$5A,$DF,$FA,$0D,$7F,$96,$A7,$E4,$9F,$E6,$22,$BF,$E7,$3F,$3F
	.dc.b	$FC,$A9,$E7,$7F,$95,$F1,$FD,$8F,$99,$FE,$55,$BF,$F2,$E1,$FE,$4F
	.dc.b	$2C,$6B,$D0,$E1,$8F,$F5,$0C,$39,$D7,$F9,$88,$DC,$67,$0D,$C7,$F5
	.dc.b	$BC,$3B,$3F,$E5,$93,$FC,$9F,$87,$F0,$2B,$79,$DF,$2E,$9E,$7F,$F9
	.dc.b	$7B,$1F,$04,$DC,$3A,$4D,$57,$C9,$EB,$5F,$E5,$E9,$FA,$DE,$47,$F8
	.dc.b	$43,$A7,$FC,$1B,$D7,$FE,$1C,$F3,$7F,$C2,$58,$7F,$86,$5A,$FF,$86
	.dc.b	$1B,$5C,$63,$FC,$5B,$18,$25,$74,$F2,$BF,$CB,$07,$F8,$72,$77,$19
	.dc.b	$7F,$BF,$52,$93,$FE,$01,$BF,$E6,$F3,$98,$C5,$FF,$99,$2F,$8D,$F0
	.dc.b	$7F,$9F,$47,$F8,$12,$B8,$46,$0F,$F2,$C4,$79,$1F,$75,$78,$D4,$F0
	.dc.b	$BF,$0D,$F5,$7F,$A4,$F9,$FF,$E6,$82,$79,$FF,$E5,$76,$DF,$F9,$30
	.dc.b	$F7,$92,$7E,$76,$1D,$F0,$ED,$8F,$9F,$1E,$CF,$B6,$FC,$71,$87,$DF
	.dc.b	$5F,$09,$F5,$98,$3D,$C7,$07,$4F,$F9,$AA,$7E,$03,$F6,$E3,$C5,$3D
	.dc.b	$63,$81,$CB,$0F,$5B,$FE,$63,$9D,$9D,$2F,$D3,$1E,$23,$8E,$1F,$22
	.dc.b	$7C,$35,$CA,$3F,$CD,$58,$78,$AF,$F1,$51,$F1,$2F,$E3,$4E,$70,$FC
	.dc.b	$5C,$60,$F7,$4F,$57,$CD,$8F,$D1,$23,$F4,$63,$A1,$2B,$A6,$3E,$A9
	.dc.b	$F8,$F8,$7A,$9F,$8C,$76,$9F,$A9,$3E,$AB,$EB,$3F,$0C,$F2,$1F,$02
	.dc.b	$76,$F9,$0F,$13,$E0,$77,$5D,$CF,$06,$3F,$16,$78,$CF,$F9,$F7,$1F
	.dc.b	$F3,$E3,$F7,$4C,$61,$FC,$7B,$B7,$FC,$F7,$EB,$3D,$6C,$36,$BF,$D0
	.dc.b	$4D,$7F,$CF,$BD,$5F,$E7,$F7,$F4,$DE,$7F,$F9,$FD,$FE,$A3,$A7,$F8
	.dc.b	$AD,$FF,$1B,$9F,$E3,$4E,$0F,$F2,$8E,$9F,$E3,$D2,$47,$F8,$D6,$FF
	.dc.b	$C7,$93,$A8,$C1,$C4,$FF,$28,$4F,$07,$FC,$6D,$7F,$E3,$C7,$FC,$73
	.dc.b	$7E,$13,$FC,$96,$F2,$DF,$FC,$7F,$2C,$3F,$24,$D9,$FF,$43,$47,$89
	.dc.b	$EE,$71,$BF,$05,$F5,$AF,$04,$FB,$AF,$AD,$F4,$3F,$53,$E8,$FF,$97
	.dc.b	$EF,$FC,$97,$3F,$E6,$B9,$EC,$FF,$99,$AB,$B1,$0F,$6C,$6B,$0F,$F2
	.dc.b	$C1,$E2,$C3,$D1,$5F,$EA,$D3,$C3,$5F,$94,$8C,$CF,$FA,$92,$3F,$CF
	.dc.b	$6D,$7F,$CE,$27,$63,$E7,$4F,$8D,$C3,$D0,$ED,$1E,$1A,$F7,$4F,$C8
	.dc.b	$38,$C7,$99,$F5,$5F,$C7,$9F,$54,$7D,$6B,$F2,$BE,$43,$FD,$4D,$3B
	.dc.b	$79,$9B,$9E,$F7,$F2,$63,$27,$D1,$BF,$23,$F0,$67,$E8,$FA,$3F,$E8
	.dc.b	$38,$D1,$E4,$C3,$C2,$FC,$2C,$3C,$18,$7F,$9F,$BC,$FF,$F2,$DC,$FF
	.dc.b	$9D,$30,$FF,$3D,$63,$FE,$54,$9C,$F3,$FF,$9D,$27,$3A,$7F,$98,$5B
	.dc.b	$7F,$E5,$B8,$FF,$3A,$46,$27,$FC,$F3,$5C,$A7,$FD,$0F,$E8,$B0,$ED
	.dc.b	$E8,$7F,$93,$63,$13,$FE,$57,$F2,$FF,$CE,$78,$3D,$47,$A9,$FE,$58
	.dc.b	$9C,$C6,$2B,$81,$FE,$85,$AE,$15,$B8,$C4,$EF,$4F,$F4,$17,$A1,$FE
	.dc.b	$52,$BD,$D7,$29,$FF,$4B,$C6,$1D,$E1,$FE,$94,$7F,$DD,$B4,$8C,$47
	.dc.b	$FB,$D6,$31,$5F,$E7,$C8,$FF,$79,$FC,$13,$E0,$FF,$CF,$3F,$19,$FE
	.dc.b	$A8,$AF,$F3,$CC,$E5,$FF,$46,$46,$6B,$FD,$31,$39,$7F,$D5,$7E,$7A
	.dc.b	$75,$97,$FA,$A2,$B4,$FF,$AB,$27,$5E,$B7,$FA,$42,$79,$5F,$FA,$3D
	.dc.b	$AF,$9B,$1F,$5A,$8C,$A1,$5F,$E7,$AE,$2F,$F4,$2C,$7F,$9F,$74,$FF
	.dc.b	$44,$7A,$9B,$D3,$FD,$0D,$E0,$7F,$A2,$FE,$35,$F0,$9F,$E8,$38,$C3
	.dc.b	$FE,$A4,$C3,$FC,$3C,$7F,$A0,$B9,$FF,$D0,$BE,$97,$FA,$09,$B1,$FE
	.dc.b	$89,$3F,$C4,$B1,$FE,$7E,$F4,$FF,$D0,$64,$E1,$FF,$41,$C7,$A6,$5E
	.dc.b	$83,$E3,$BF,$CE,$DF,$1E,$F2,$A9,$FE,$79,$9D,$57,$FA,$B9,$FF,$42
	.dc.b	$BF,$A0,$F8,$1F,$F6,$BE,$3F,$A7,$6E,$FD,$EB,$75,$FE,$D1,$F0,$1F
	.dc.b	$0F,$FE,$74,$84,$61,$FF,$3B,$B5,$FF,$50,$B5,$FF,$63,$39,$C3,$FD
	.dc.b	$CF,$F2,$1F,$EA,$E9,$FF,$61,$B5,$F0,$57,$FA,$4E,$B8,$61,$F0,$EB
	.dc.b	$57,$CB,$D9,$D5,$FF,$A3,$70,$FF,$6A,$7A,$3F,$E8,$B7,$FD,$66,$DA
	.dc.b	$FF,$7B,$FA,$5F,$EC,$D9,$FF,$69,$B5,$FF,$78,$4F,$FA,$61,$CF,$B9
	.dc.b	$FE,$A5,$9D,$57,$FA,$77,$D4,$FF,$4E,$3D,$27,$46,$33,$FF,$45,$5F
	.dc.b	$FB,$83,$77,$BD,$FE,$85,$3F,$D1,$B1,$FE,$80,$F5,$30,$EE,$33,$E9
	.dc.b	$A7,$32,$C7,$FD,$43,$5C,$BC,$AF,$F1,$9F,$93,$FE,$17,$E9,$FF,$14
	.dc.b	$5F,$F8,$8C,$87,$F8,$98,$FF,$10,$18,$F4,$BF,$C3,$4F,$F9,$B0,$FF
	.dc.b	$0F,$FA,$5F,$E4,$DB,$FF,$0F,$FA,$7F,$E2,$AB,$FF,$21,$D7,$49,$CB
	.dc.b	$FE,$28,$31,$5F,$E5,$5B,$EB,$1F,$E5,$93,$AF,$A7,$FE,$53,$9D,$61
	.dc.b	$D7,$1F,$F4,$86,$3E,$AB,$F3,$93,$0E,$78,$FC,$08,$EF,$8F,$F8,$FD
	.dc.b	$F1,$1C,$23,$CA,$FF,$90,$6B,$A4,$7F,$8E,$F1,$F0,$CF,$08,$C5,$73
	.dc.b	$8E,$18,$7C,$D2,$3E,$03,$FC,$90,$68,$C5,$FF,$A4,$31,$F6,$47,$CD
	.dc.b	$8F,$0E,$1F,$34,$F0,$C7,$13,$D0,$E2,$3E,$7C,$62,$FC,$91,$8C,$3C
	.dc.b	$F3,$E2,$8C,$57,$5A,$DE,$1F,$30,$F4,$D7,$33,$83,$97,$D2,$FF,$A6
	.dc.b	$EF,$AD,$FA,$E7,$E3,$5F,$57,$89,$ED,$AE,$B5,$C1,$E5,$3B,$AE,$55
	.dc.b	$EF,$AD,$3F,$E9,$36,$BF,$97,$6E,$3F,$80,$AF,$F4,$6D,$7C,$AC,$3F
	.dc.b	$14,$7B,$AF,$C4,$7A,$8F,$E8,$94,$C7,$F9,$75,$7F,$A4,$E7,$FC,$EA
	.dc.b	$7F,$9A,$AF,$FC,$F2,$FF,$AB,$0F,$E6,$83,$D3,$FF,$3C,$FA,$7F,$E7
	.dc.b	$82,$47,$F9,$CB,$A7,$FC,$E8,$FF,$9E,$0C,$78,$BC,$9F,$F1,$13,$FE
	.dc.b	$2C,$D7,$FC,$89,$AF,$F9,$59,$B5,$FE,$26,$F4,$BF,$CB,$4D,$F4,$BF
	.dc.b	$C5,$95,$FE,$6E,$C3,$FC,$8F,$7F,$E3,$2D,$3F,$C8,$F1,$D8,$FF,$1A
	.dc.b	$1F,$E4,$77,$FC,$99,$1F,$E8,$AC,$7F,$C6,$FC,$BF,$E8,$1A,$E6,$71
	.dc.b	$AE,$6E,$09,$7F,$E3,$8B,$F0,$1C,$E3,$B3,$87,$FC,$9C,$32,$F5,$E1
	.dc.b	$E0,$8F,$F3,$F9,$D3,$2F,$F2,$2D,$FC,$38,$FF,$2A,$C6,$0F,$85,$8F
	.dc.b	$9E,$7C,$53,$E7,$C3,$C5,$8F,$F9,$5C,$F7,$C7,$8B,$1F,$0B,$E6,$38
	.dc.b	$19,$CB,$E6,$E3,$FE,$8E,$C3,$E7,$1D,$A7,$CD,$97,$D9,$7C,$AD,$7D
	.dc.b	$D5,$CA,$3F,$CE,$31,$FE,$62,$38,$9F,$53,$0F,$36,$3F,$1E,$3F,$CC
	.dc.b	$98,$7B,$EB,$E3,$57,$FA,$17,$0F,$11,$27,$55,$EB,$3A,$CF,$D8,$9D
	.dc.b	$57,$97,$1F,$83,$3E,$0A,$F8,$27,$82,$75,$8F,$A6,$3F,$CF,$F7,$F0
	.dc.b	$B0,$EF,$7C,$67,$BE,$3F,$40,$FF,$0D,$62,$C1,$97,$C5,$F4,$BF,$CE
	.dc.b	$17,$FE,$C3,$F4,$7F,$CD,$AF,$F9,$DC,$FF,$40,$38,$E4,$FF,$3B,$FA
	.dc.b	$5F,$E7,$3C,$BF,$CE,$B8,$7F,$9F,$2B,$4F,$F8,$A7,$0E,$BB,$7F,$9F
	.dc.b	$E3,$17,$C5,$FF,$35,$61,$E4,$C7,$FB,$67,$A3,$FE,$71,$8C,$43,$D1
	.dc.b	$FF,$39,$4E,$6B,$FC,$DB,$E6,$7F,$9F,$BD,$2F,$F3,$D5,$FF,$9D,$2B
	.dc.b	$4F,$F9,$BE,$7F,$CA,$F7,$FE,$70,$8F,$E0,$4D,$7F,$86,$B5,$EB,$7E
	.dc.b	$63,$FC,$B9,$E7,$3D,$E7,$7F,$83,$3F,$F3,$67,$A3,$FE,$78,$F4,$BF
	.dc.b	$CD,$4D,$C3,$FC,$F2,$DC,$7F,$CE,$ED,$AE,$33,$A3,$FC,$D4,$FC,$5F
	.dc.b	$91,$7C,$9B,$E4,$FA,$D6,$8F,$F2,$C6,$9F,$E6,$FF,$53,$FC,$D9,$97
	.dc.b	$F8,$5E,$73,$BF,$F9,$C2,$3F,$CD,$91,$FE,$B4,$F5,$3F,$CD,$EF,$F9
	.dc.b	$C8,$C7,$A7,$FE,$6F,$97,$5F,$F9,$B1,$FE,$72,$7F,$9C,$49,$E1,$BD
	.dc.b	$2F,$F3,$76,$FF,$E6,$C9,$FF,$42,$78,$2F,$4F,$FC,$E1,$1F,$E7,$43
	.dc.b	$FD,$11,$E8,$7F,$9A,$9F,$41,$FE,$8B,$F4,$BF,$CE,$0F,$BC,$C7,$97
	.dc.b	$FE,$71,$F4,$FF,$CE,$5E,$6F,$F9,$E7,$2F,$F3,$76,$F4,$FF,$4A,$77
	.dc.b	$BD,$3F,$F3,$8E,$3E,$0B,$FF,$2D,$BD,$4F,$F3,$7F,$A3,$87,$C6,$3D
	.dc.b	$3F,$D6,$3B,$F2,$3F,$CC,$EF,$FA,$86,$3F,$CC,$F8,$FF,$9A,$7D,$3E
	.dc.b	$59,$FE,$72,$F7,$7C,$E7,$CC,$F1,$77,$5F,$E9,$3F,$4F,$FC,$C9,$3F
	.dc.b	$E6,$6C,$BF,$CC,$9C,$9F,$E7,$18,$D4,$7F,$A1,$E7,$FC,$E5,$1F,$E7
	.dc.b	$48,$FF,$32,$BF,$E6,$C3,$DD,$1D,$31,$FF,$35,$46,$4E,$C7,$C3,$F4
	.dc.b	$BF,$CC,$4F,$F9,$94,$87,$C4,$72,$DF,$4B,$FC,$CF,$3F,$E6,$68,$E5
	.dc.b	$1F,$E7,$0E,$1F,$F3,$9C,$72,$9F,$DF,$D7,$27,$F8,$0F,$9B,$FE,$4D
	.dc.b	$9F,$F1,$37,$47,$F8,$60,$FF,$16,$E7,$FE,$10,$8F,$F1,$09,$FE,$5D
	.dc.b	$8F,$F0,$A3,$F9,$8F,$2F,$FD,$79,$5F,$E5,$29,$E3,$39,$3F,$CE,$5E
	.dc.b	$87,$F9,$19,$E5,$5F,$E2,$AF,$43,$FC,$50,$FF,$8A,$B0,$F0,$5F,$8E
	.dc.b	$3F,$C7,$15,$E6,$74,$F5,$3F,$CD,$17,$D7,$0F,$61,$FE,$2C,$71,$87
	.dc.b	$F8,$B2,$BC,$18,$76,$79,$E0,$7D,$73,$FE,$55,$7C,$37,$CF,$0F,$F2
	.dc.b	$2D,$F1,$7F,$C7,$95,$EA,$6C,$65,$FF,$1F,$46,$5B,$19,$F4,$7F,$C8
	.dc.b	$0E,$EB,$E5,$5F,$82,$F8,$56,$AF,$FD,$1A,$7C,$1A,$DB,$FE,$44,$3E
	.dc.b	$31,$E2,$35,$3E,$B7,$51,$EB,$7F,$CA,$D1,$DB,$1F,$8F,$87,$8E,$38
	.dc.b	$4F,$D8,$3F,$CA,$07,$A8,$EB,$1F,$E5,$57,$47,$88,$D6,$3F,$35,$FC
	.dc.b	$19,$F2,$8E,$27,$E3,$AF,$EC,$D7,$8E,$35,$3C,$5F,$BE,$63,$1F,$AF
	.dc.b	$7E,$1B,$E2,$FC,$DC,$3D,$75,$F2,$2F,$95,$7C,$88,$F3,$BF,$0E,$3F
	.dc.b	$32,$6B,$0F,$83,$07,$FC,$D2,$DB,$FF,$30,$57,$F9,$AC,$F1,$9F,$E6
	.dc.b	$3E,$BF,$F3,$5B,$7D,$77,$A5,$FE,$6C,$8F,$F3,$51,$0F,$F2,$71,$4F
	.dc.b	$F0,$CD,$FF,$91,$DF,$F1,$8F,$1F,$F9,$17,$97,$FC,$2F,$AF,$F9,$B3
	.dc.b	$A7,$FC,$56,$FF,$85,$EB,$8C,$FF,$93,$1F,$F2,$9D,$FF,$8A,$1F,$F1
	.dc.b	$B4,$F1,$7F,$C5,$D1,$9D,$7F,$C5,$D5,$DA,$7F,$C8,$67,$02,$C6,$AF
	.dc.b	$FC,$5B,$87,$95,$B3,$E0,$AE,$4F,$A8,$C4,$6A,$BF,$C8,$53,$AF,$4B
	.dc.b	$FC,$77,$8F,$96,$7F,$C6,$98,$F9,$63,$DD,$87,$39,$E3,$87,$39,$CD
	.dc.b	$6E,$BF,$CC,$58,$F9,$5F,$51,$C6,$77,$8F,$87,$D2,$FF,$23,$63,$FE
	.dc.b	$6E,$AF,$F2,$51,$C3,$2F,$2B,$F0,$AF,$C7,$18,$8E,$87,$FA,$1B,$1F
	.dc.b	$13,$CA,$BE,$43,$CE,$FB,$DF,$F9,$9E,$BF,$01,$7C,$C9,$3A,$AF,$BB
	.dc.b	$97,$E0,$1E,$D8,$FA,$72,$F7,$65,$D8,$99,$79,$B0,$E5,$7F,$E5,$E3
	.dc.b	$E2,$9F,$50,$F2,$57,$29,$E0,$D8,$DD,$7D,$07,$67,$FA,$83,$D1,$FF
	.dc.b	$2B,$FA,$5F,$E5,$A8,$08,$CD,$FF,$98,$23,$FD,$95,$0F,$4B,$FC,$AF
	.dc.b	$1F,$E5,$EC,$3F,$5F,$39,$6C,$F1,$E5,$FF,$30,$79,$1F,$E4,$BF,$82
	.dc.b	$C5,$7F,$94,$30,$FF,$2D,$7A,$7F,$E4,$D8,$FF,$21,$FA,$7F,$E5,$C9
	.dc.b	$FF,$32,$79,$5F,$E5,$C2,$4F,$EF,$BD,$3F,$F2,$D4,$FF,$98,$63,$FC
	.dc.b	$B4,$B1,$FF,$2D,$7A,$5D,$63,$47,$F8,$7A,$BF,$CC,$D9,$F9,$BD,$3F
	.dc.b	$F2,$FE,$1C,$A9,$EB,$BD,$3F,$F2,$D4,$FC,$89,$CC,$61,$F9,$73,$B9
	.dc.b	$FE,$26,$D8,$FF,$7A,$35,$FF,$7F,$B5,$E8,$EE,$FF,$CC,$F1,$FE,$46
	.dc.b	$F4,$FF,$CC,$BE,$97,$F9,$92,$3F,$C6,$FA,$7C,$27,$E0,$6B,$F5,$9F
	.dc.b	$F3,$DB,$5C,$E3,$FC,$9A,$5E,$BE,$5B,$19,$8C,$1C,$67,$F0,$FC,$DF
	.dc.b	$E6,$56,$9F,$E6,$4F,$43,$E2,$38,$F0,$3F,$CC,$CF,$F9,$93,$D0,$FF
	.dc.b	$31,$C6,$21,$E0,$FF,$99,$F6,$FF,$35,$43,$D0,$FF,$33,$3D,$1A,$FD
	.dc.b	$16,$BF,$E6,$CF,$3F,$FA,$0B,$C7,$CB,$B2,$69,$23,$FA,$3F,$7F,$F9
	.dc.b	$58,$9E,$77,$F9,$5A,$3F,$CC,$5E,$07,$F9,$95,$F6,$CA,$7B,$B6,$39
	.dc.b	$C6,$0E,$5E,$77,$F9,$9A,$3F,$CC,$07,$F9,$AE,$3B,$79,$BF,$E5,$5E
	.dc.b	$2F,$F2,$EC,$FF,$9D,$7D,$76,$FA,$CE,$4F,$F9,$71,$FF,$32,$B4,$FE
	.dc.b	$C3,$7F,$E6,$47,$99,$F5,$03,$F0,$DF,$F3,$16,$DF,$E6,$47,$FC,$C6
	.dc.b	$7F,$97,$4F,$F2,$D9,$0F,$F4,$2F,$9A,$F6,$BE,$37,$9D,$FE,$64,$9F
	.dc.b	$C0,$B5,$FD,$33,$5F,$E1,$ED,$4F,$F9,$8F,$3F,$D6,$57,$F9,$A2,$7B
	.dc.b	$FB,$1D,$BD,$0F,$A9,$E5,$FF,$83,$27,$FC,$2F,$D1,$FE,$55,$7F,$C5
	.dc.b	$7E,$77,$F8,$5D,$FF,$15,$EF,$FE,$11,$7F,$C2,$C7,$F8,$80,$C9,$CB
	.dc.b	$C9,$FF,$28,$9F,$E0,$8A,$E9,$1F,$E1,$4A,$E9,$1F,$E1,$49,$D1,$C2
	.dc.b	$7A,$46,$5B,$1F,$E4,$9A,$F1,$47,$82,$7F,$C3,$91,$FE,$27,$AF,$F3
	.dc.b	$24,$FF,$8F,$3D,$0F,$F2,$75,$7F,$9F,$8D,$C7,$F8,$D6,$76,$FF,$8F
	.dc.b	$67,$77,$FE,$6C,$8F,$61,$FE,$51,$AE,$27,$B3,$0F,$94,$E2,$7D,$53
	.dc.b	$FE,$3C,$3C,$CF,$B9,$E4,$D7,$FC,$96,$DA,$F0,$C7,$A7,$D0,$FF,$1F
	.dc.b	$D7,$A6,$B5,$18,$7C,$D1,$E1,$9E,$A7,$C8,$3E,$DE,$3F,$22,$3E,$64
	.dc.b	$79,$5B,$87,$F9,$24,$F5,$D7,$69,$FB,$E6,$DF,$A4,$FC,$43,$D2,$77
	.dc.b	$3E,$F5,$77,$8D,$D7,$96,$76,$73,$3C,$A7,$3C,$3E,$0B,$F4,$F0,$FF
	.dc.b	$2F,$BF,$C9,$30,$FF,$2B,$3F,$B3,$6B,$9C,$7F,$7E,$FF,$95,$7C,$FF
	.dc.b	$F2,$A3,$FD,$3F,$AF,$FC,$B2,$FF,$94,$C5,$3A,$1E,$77,$F9,$8A,$30
	.dc.b	$D8,$CC,$F8,$5F,$F0,$A3,$FE,$0D,$CF,$FC,$26,$FF,$82,$B4,$FF,$18
	.dc.b	$D7,$F8,$E6,$3F,$C3,$6D,$7C,$35,$FE,$4A,$BF,$F1,$2F,$57,$F8,$69
	.dc.b	$FF,$16,$98,$E2,$FF,$0B,$57,$F8,$BF,$0F,$F0,$CB,$E1,$6E,$3F,$E1
	.dc.b	$93,$C7,$1B,$3C,$87,$88,$FF,$0E,$BB,$8E,$27,$A2,$32,$D8,$E7,$5E
	.dc.b	$39,$F0,$1D,$0F,$5E,$3F,$D7,$AB,$FC,$81,$8F,$F9,$DE,$FB,$56,$EF
	.dc.b	$FD,$04,$7B,$A7,$55,$DC,$F1,$CF,$9E,$BC,$33,$A8,$C4,$E9,$F0,$9F
	.dc.b	$E6,$2A,$F3,$57,$82,$3E,$11,$23,$33,$DE,$3C,$95,$F4,$CE,$CF,$4A
	.dc.b	$FF,$34,$57,$08,$F0,$9C,$2F,$E0,$3F,$E5,$A3,$FD,$23,$5F,$86,$AF
	.dc.b	$15,$71,$9E,$97,$F1,$AF,$FC,$C1,$18,$39,$57,$FA,$1A,$F9,$57,$9E
	.dc.b	$7D,$51,$E6,$BF,$96,$EA,$3C,$A6,$AF,$AC,$E9,$FC,$93,$63,$EF,$9C
	.dc.b	$1F,$31,$9B,$FA,$33,$CA,$BE,$33,$4F,$83,$5F,$E7,$13,$C0,$7C,$27
	.dc.b	$BC,$7E,$AA,$7F,$DC,$5E,$6F,$F9,$86,$BF,$CC,$FE,$6F,$F8,$BE,$31
	.dc.b	$1F,$E2,$46,$BF,$E6,$91,$CE,$CF,$0F,$F9,$4F,$CD,$FF,$33,$CF,$F9
	.dc.b	$DA,$BB,$4E,$E3,$17,$63,$85,$FF,$9F,$1D,$4F,$F9,$F6,$75,$3F,$E7
	.dc.b	$18,$FF,$1B,$F9,$5F,$E4,$92,$78,$BF,$E6,$E8,$FF,$33,$3C,$1F,$F3
	.dc.b	$67,$95,$FE,$66,$73,$5F,$E6,$7A,$FF,$35,$BF,$E8,$CF,$1B,$FC,$E2
	.dc.b	$E6,$FF,$CE,$51,$8A,$E2,$7C,$29,$FF,$30,$F9,$5F,$E6,$B8,$FF,$36
	.dc.b	$C7,$AC,$E9,$EC,$BD,$9F,$F3,$16,$9F,$E6,$63,$E2,$79,$7F,$E6,$6A
	.dc.b	$E1,$7F,$E7,$4C,$37,$1D,$8C,$CE,$E7,$C4,$6E,$38,$BF,$E6,$83,$FC
	.dc.b	$DF,$97,$C3,$F8,$07,$C0,$BD,$5F,$8D,$5B,$CB,$E8,$EB,$FE,$6C,$8F
	.dc.b	$F3,$BF,$9B,$F2,$F9,$9D,$3F,$73,$4F,$F4,$65,$EF,$0F,$F3,$D7,$37
	.dc.b	$F9,$BF,$13,$48,$FF,$91,$BC,$DF,$E8,$DE,$6F,$F9,$B9,$B5,$FE,$6C
	.dc.b	$F3,$1E,$6F,$F9,$75,$B7,$EE,$6B,$F0,$87,$AE,$F8,$DE,$1E,$64,$9E
	.dc.b	$0F,$FA,$8F,$0F,$CA,$71,$BE,$09,$F0,$5F,$E6,$BF,$03,$FC,$DB,$87
	.dc.b	$F9,$84,$F9,$BE,$8F,$C2,$70,$78,$63,$6F,$8E,$34,$E7,$D1,$FF,$36
	.dc.b	$F3,$C8,$E3,$97,$E8,$3D,$0F,$F3,$64,$7F,$B7,$9F,$F3,$03,$ED,$F4
	.dc.b	$7F,$B7,$BF,$E6,$09,$E1,$AF,$F9,$BA,$35,$E6,$59,$CC,$7F,$9B,$CC
	.dc.b	$94,$FF,$50,$DF,$F9,$C5,$F0,$57,$FA,$53,$D0,$FF,$2F,$47,$F9,$8B
	.dc.b	$CE,$FF,$2E,$3F,$8E,$F4,$3F,$CD,$8F,$F9,$BF,$0F,$DC,$D7,$08,$FF
	.dc.b	$33,$D7,$FA,$56,$B7,$1F,$E7,$73,$FC,$FD,$3F,$E6,$D2,$4F,$FA,$93
	.dc.b	$C9,$FF,$12,$CF,$F8,$7D,$AF,$F8,$6C,$FF,$10,$9F,$E1,$BF,$3B,$FC
	.dc.b	$6C,$FF,$8E,$CF,$F1,$5C,$7F,$8B,$FC,$EF,$F2,$8B,$FE,$1C,$EB,$FF
	.dc.b	$13,$79,$FF,$E2,$28,$C7,$9B,$FE,$33,$BF,$F2,$59,$E1,$BE,$27,$F9
	.dc.b	$B2,$3F,$C5,$E6,$EF,$FC,$7B,$1F,$E2,$47,$FC,$B0,$F0,$DF,$FC,$6D
	.dc.b	$5F,$E3,$29,$F0,$E3,$E8,$8F,$F1,$91,$E8,$AE,$57,$FE,$5E,$C3,$D5
	.dc.b	$87,$F4,$07,$CF,$5E,$58,$C5,$79,$63,$B5,$79,$4F,$3B,$B3,$FC,$99
	.dc.b	$5D,$1B,$3E,$07,$FC,$BA,$7C,$F8,$FA,$26,$F0,$FF,$33,$47,$36,$BB
	.dc.b	$9F,$80,$FD,$53,$D9,$7F,$E8,$93,$CE,$7C,$47,$ED,$CF,$C4,$76,$7F
	.dc.b	$A2,$63,$6F,$33,$EA,$9D,$DC,$C7,$18,$CD,$79,$EF,$FC,$F2,$7C,$38
	.dc.b	$F8,$A7,$A5,$CB,$4F,$C5,$1C,$AF,$E2,$3F,$E6,$C9,$E8,$EA,$BF,$D3
	.dc.b	$EF,$F9,$96,$FF,$CB,$95,$FB,$88,$12,$BF,$CD,$6F,$F9,$9F,$D9,$7A
	.dc.b	$1F,$E6,$A8,$DF,$A1,$FE,$6A,$78,$9F,$E3,$03,$FC,$D9,$5F,$E6,$87
	.dc.b	$FC,$CD,$E2,$7F,$9A,$4F,$F0,$09,$F8,$7E,$8F,$F0,$B4,$7F,$87,$CF
	.dc.b	$F0,$44,$FF,$84,$23,$FC,$1B,$1F,$E0,$93,$FC,$2A,$6E,$3F,$C9,$13
	.dc.b	$FE,$11,$8F,$F2,$04,$FF,$88,$CE,$39,$FF,$8A,$67,$C3,$3A,$78,$47
	.dc.b	$F8,$F2,$78,$47,$94,$DC,$65,$DE,$9F,$E3,$7A,$F2,$4F,$F8,$91,$B1
	.dc.b	$FE,$53,$BF,$F1,$43,$6F,$C6,$FC,$8F,$67,$FA,$CF,$89,$FE,$7C,$AF
	.dc.b	$81,$8F,$87,$0F,$7C,$72,$7F,$C9,$C7,$A4,$EE,$63,$1F,$33,$83,$FC
	.dc.b	$62,$FC,$17,$9E,$3E,$27,$F0,$38,$78,$9F,$F3,$D9,$C3,$1F,$37,$8B
	.dc.b	$FE,$93,$AF,$AF,$87,$8E,$B4,$72,$9D,$63,$F2,$2F,$E3,$63,$E1,$3E
	.dc.b	$31,$5F,$89,$1E,$F8,$CC,$78,$32,$F8,$98,$FD,$4C,$3E,$05,$71,$7E
	.dc.b	$05,$76,$38,$3E,$AC,$7E,$5B,$C2,$33,$3F,$E7,$2A,$E8,$FC,$83,$A4
	.dc.b	$FB,$06,$5F,$16,$38,$63,$F8,$29,$FF,$3A,$DF,$43,$FC,$E9,$87,$82
	.dc.b	$91,$FA,$6F,$37,$FC,$7B,$5F,$E5,$8F,$5D,$EC,$7F,$88,$27,$3D,$09
	.dc.b	$CE,$BF,$E2,$A9,$CC,$62,$FF,$C8,$35,$FE,$59,$8C,$3F,$E5,$89,$FF
	.dc.b	$29,$D7,$ED,$AB,$FC,$B0,$43,$FC,$6C,$7F,$8E,$3C,$87,$9F,$FE,$45
	.dc.b	$9F,$F1,$CF,$AE,$F3,$FF,$C9,$8D,$8F,$F2,$2D,$70,$F4,$3F,$C8,$B8
	.dc.b	$23,$13,$CA,$F3,$E8,$7F,$91,$F3,$FF,$28,$4F,$F9,$2E,$FF,$C5,$18
	.dc.b	$F5,$7F,$C9,$75,$B9,$D5,$7F,$9D,$3E,$59,$F2,$DF,$E6,$DF,$96,$7C
	.dc.b	$D3,$DF,$FF,$25,$D6,$B0,$FF,$2D,$D6,$A7,$37,$C6,$F8,$57,$F9,$62
	.dc.b	$F8,$65,$FE,$5C,$C3,$84,$E6,$F8,$BF,$E6,$6E,$67,$CE,$3E,$73,$38
	.dc.b	$7F,$96,$2B,$5F,$05,$DA,$30,$7F,$9A,$2D,$19,$D7,$FC,$C3,$F2,$3F
	.dc.b	$E8,$95,$F3,$5F,$E5,$DC,$3F,$CB,$F3,$98,$C6,$1F,$7F,$AB,$FC,$BB
	.dc.b	$E1,$BD,$6F,$C1,$4F,$F9,$88,$FF,$30,$1B,$28,$F0,$5E,$BF,$F9,$8E
	.dc.b	$3F,$A0,$B5,$7A,$CF,$95,$FF,$2D,$69,$FD,$13,$D7,$7C,$06,$2F,$FA
	.dc.b	$7C,$61,$FF,$34,$57,$F9,$AF,$D9,$D4,$7F,$9A,$3C,$FB,$47,$51,$FE
	.dc.b	$8C,$F9,$C7,$97,$C1,$B1,$98,$E7,$A3,$E7,$35,$18,$F8,$49,$5B,$21
	.dc.b	$FE,$79,$A4,$6A,$3F,$CF,$97,$D9,$FF,$35,$1F,$E6,$8D,$7F,$B1,$F9
	.dc.b	$6F,$5F,$D7,$EC,$7F,$9C,$3A,$3F,$D9,$73,$FE,$74,$6C,$EB,$CE,$FF
	.dc.b	$3F,$3F,$E7,$3C,$3F,$CF,$1E,$C6,$1E,$C7,$F7,$8C,$3F,$D5,$0F,$FA
	.dc.b	$0A,$34,$FF,$A4,$7D,$7F,$F3,$52,$69,$FE,$7A,$C3,$FD,$BB,$EC,$3E
	.dc.b	$17,$E8,$B6,$BF,$CE,$D3,$FE,$78,$7F,$D0,$13,$A3,$FC,$F9,$A7,$F8
	.dc.b	$EA,$3F,$C4,$07,$F8,$E3,$AB,$FC,$58,$D3,$FC,$50,$5E,$BF,$F2,$9C
	.dc.b	$7F,$89,$BD,$0F,$F1,$14,$FF,$8C,$1B,$E8,$FF,$8F,$BD,$4F,$F0,$FF
	.dc.b	$AD,$FE,$65,$7F,$C9,$95,$B3,$FC,$4C,$7F,$96,$0F,$F1,$35,$F8,$B0
	.dc.b	$FF,$21,$DF,$F8,$9A,$75,$1F,$E6,$69,$D6,$1F,$E3,$09,$F5,$CF,$0B
	.dc.b	$F0,$FA,$DF,$E6,$5C,$3F,$CB,$F5,$AB,$E1,$1D,$4C,$96,$BC,$EF,$AF
	.dc.b	$0E,$F5,$FE,$37,$AF,$95,$3B,$9F,$F2,$93,$B9,$F9,$38,$78,$E7,$8C
	.dc.b	$7A,$AB,$8D,$F9,$A3,$FC,$AF,$5F,$39,$FA,$13,$FE,$79,$9D,$47,$C6
	.dc.b	$3A,$C7,$AA,$33,$3F,$16,$72,$79,$CF,$F2,$B9,$F0,$4C,$1F,$4C,$C1
	.dc.b	$E5,$37,$7E,$D7,$FC,$D1,$1D,$4F,$31,$ED,$7B,$9C,$4E,$91,$FE,$85
	.dc.b	$7F,$CB,$87,$F1,$C7,$FC,$B2,$FF,$96,$E7,$FC,$BC,$FF,$96,$39,$FF
	.dc.b	$CB,$6F,$F2,$2E,$FF,$F2,$D3,$9F,$5B,$FC,$AC,$FF,$96,$08,$7F,$97
	.dc.b	$39,$BF,$C4,$27,$F8,$A5,$FF,$02,$F3,$7F,$88,$1F,$F1,$6F,$8B,$FE
	.dc.b	$1B,$D3,$FC,$3B,$1F,$E1,$13,$FC,$4D,$C1,$FE,$1C,$9F,$F0,$54,$7F
	.dc.b	$98,$2F,$FC,$27,$5F,$E1,$B8,$D9,$FE,$1B,$8C,$BF,$E1,$03,$3A,$FF
	.dc.b	$85,$1F,$F2,$06,$BF,$E1,$1A,$E5,$1E,$6C,$3C,$0E,$AF,$9B,$E1,$C3
	.dc.b	$C7,$7F,$E6,$C8,$D4,$62,$7E,$0B,$ED,$8F,$F1,$89,$E3,$3B,$1E,$77
	.dc.b	$81,$FE,$43,$75,$EB,$7F,$8A,$F0,$FF,$13,$D7,$8A,$FC,$18,$FF,$8C
	.dc.b	$9F,$55,$F0,$CB,$E7,$4F,$F9,$1A,$3E,$0D,$F2,$C7,$BC,$7D,$23,$E8
	.dc.b	$1E,$49,$F8,$C7,$5C,$3C,$07,$B2,$37,$8F,$95,$FF,$52,$47,$DF,$8C
	.dc.b	$47,$FA,$46,$38,$4F,$F9,$26,$FF,$C8,$53,$C2,$FE,$33,$FE,$4B,$BE
	.dc.b	$38,$FF,$9A,$4F,$6C,$7D,$13,$E7,$92,$FC,$D5,$EC,$7E,$04,$7A,$EF
	.dc.b	$E0,$3B,$7C,$2F,$96,$7E,$82,$F5,$3F,$C9,$3E,$B7,$F9,$3F,$E0,$3F
	.dc.b	$C1,$9E,$BF,$F9,$23,$6F,$F2,$25,$EE,$51,$FE,$08,$8C,$36,$FF,$C9
	.dc.b	$B1,$FE,$52,$8F,$F6,$5F,$AA,$F8,$1F,$E9,$91,$FE,$59,$BF,$F2,$2C
	.dc.b	$FF,$91,$2B,$FC,$7F,$EB,$7F,$93,$0E,$EF,$F8,$5B,$D6,$FF,$21,$3F
	.dc.b	$E5,$96,$BF,$E4,$73,$11,$FE,$4C,$F8,$9F,$EE,$F8,$7F,$93,$C9,$3D
	.dc.b	$9F,$F6,$FF,$67,$6F,$5F,$FC,$99,$3F,$E6,$28,$FF,$26,$7A,$7F,$E4
	.dc.b	$3F,$5F,$FC,$73,$1E,$4F,$57,$FC,$83,$48,$DE,$1F,$17,$D6,$FF,$23
	.dc.b	$CF,$F9,$BD,$AF,$FB,$99,$AF,$FB,$C1,$FF,$26,$CF,$28,$FF,$21,$7A
	.dc.b	$DF,$E4,$9A,$FF,$27,$1E,$83,$FD,$DF,$B3,$E0,$7E,$3D,$7F,$9F,$63
	.dc.b	$13,$F5,$DB,$1F,$E4,$06,$BB,$F5,$78,$BF,$E4,$A9,$E1,$18,$8F,$B6
	.dc.b	$D7,$FD,$FF,$D1,$FE,$4C,$F5,$3F,$C9,$6D,$8F,$F2,$2E,$47,$FC,$97
	.dc.b	$DF,$FE,$4A,$8C,$78,$5F,$E5,$39,$D9,$AF,$37,$E6,$3F,$E5,$2F,$53
	.dc.b	$FC,$98,$FE,$45,$AF,$F9,$69,$AF,$FA,$75,$AF,$84,$FF,$2A,$FA,$5F
	.dc.b	$99,$69,$CD,$C4,$1D,$79,$DF,$E3,$DC,$3E,$6B,$75,$E8,$DB,$F7,$61
	.dc.b	$FE,$4D,$7F,$C9,$27,$FA,$BE,$3F,$C8,$CD,$F1,$7A,$91,$FB,$87,$2F
	.dc.b	$3F,$FC,$9D,$87,$82,$7F,$C7,$6F,$F9,$1A,$7F,$C8,$23,$CE,$FF,$26
	.dc.b	$63,$FE,$50,$F3,$F6,$D9,$EA,$6F,$C7,$FF,$22,$BF,$EC,$3F,$1F,$FC
	.dc.b	$AC,$FE,$F5,$AF,$FA,$19,$AE,$EB,$FC,$A9,$87,$E4,$30,$FF,$28,$CF
	.dc.b	$F9,$88,$FF,$28,$7B,$BF,$E4,$47,$F0,$5E,$7F,$F9,$3F,$2F,$A9,$E7
	.dc.b	$B8,$BF,$CF,$F3,$FE,$54,$BF,$AC,$9A,$7F,$93,$EF,$D7,$5F,$E5,$57
	.dc.b	$B9,$EE,$F2,$7F,$C4,$DC,$BF,$E0,$AF,$27,$FC,$35,$8F,$F8,$9A,$7F
	.dc.b	$C6,$7E,$67,$F8,$72,$7F,$C4,$F3,$98,$FF,$1A,$9E,$5B,$FF,$13,$9F
	.dc.b	$9B,$F3,$BF,$CA,$13,$FE,$34,$9D,$61,$FE,$17,$3C,$15,$FE,$22,$C3
	.dc.b	$C7,$87,$F8,$BE,$FB,$1C,$4F,$F3,$04,$FF,$96,$27,$CB,$1A,$8F,$F1
	.dc.b	$2C,$F9,$9F,$F1,$1C,$E9,$B7,$E8,$AE,$71,$FE,$30,$9D,$5F,$92,$3F
	.dc.b	$C5,$51,$E4,$7D,$9E,$77,$F9,$6E,$F9,$47,$C0,$18,$FC,$07,$C6,$7F
	.dc.b	$8E,$A3,$E5,$37,$0F,$46,$0C,$7F,$CF,$35,$F1,$9F,$A7,$5E,$8A,$F8
	.dc.b	$4F,$7B,$ED,$7C,$63,$E2,$5F,$92,$3E,$75,$FC,$B9,$E4,$79,$0F,$F2
	.dc.b	$2C,$F9,$1B,$7E,$49,$F1,$BF,$1B,$1F,$F2,$95,$F7,$AF,$91,$7D,$EF
	.dc.b	$CF,$3F,$26,$BE,$5C,$78,$63,$13,$BA,$FC,$03,$EF,$6D,$F7,$8F,$25
	.dc.b	$FD,$19,$F9,$B8,$FD,$1A,$F8,$77,$E3,$AE,$F5,$F5,$8F,$E5,$91,$9B
	.dc.b	$FF,$2E,$C2,$7F,$CC,$2D,$8C,$C6,$3A,$7F,$CB,$F3,$FE,$FD,$E4,$FF
	.dc.b	$31,$BF,$C9,$CC,$79,$DF,$E6,$27,$FC,$C1,$E6,$D7,$FC,$97,$E7,$7F
	.dc.b	$86,$B9,$BF,$C4,$9C,$FF,$E3,$4D,$FF,$CB,$FA,$FF,$88,$EF,$FC,$6A
	.dc.b	$7F,$88,$CF,$F1,$6C,$ED,$FF,$24,$9F,$E2,$18,$FF,$0F,$BF,$E1,$B3
	.dc.b	$89,$FE,$1B,$3F,$C7,$8F,$57,$FC,$44,$75,$71,$7E,$5C,$3F,$C6,$3A
	.dc.b	$7F,$8B,$8F,$F2,$6E,$1F,$E2,$67,$FC,$65,$8F,$F9,$0A,$7F,$C4,$F1
	.dc.b	$AC,$3C,$2E,$A3,$FC,$65,$5E,$E3,$E1,$E3,$EB,$AF,$F1,$F3,$6B,$C1
	.dc.b	$7F,$E4,$6A,$DF,$27,$F9,$CB,$2F,$79,$23,$C9,$3F,$32,$75,$18,$C7
	.dc.b	$D7,$3E,$7C,$7E,$1C,$F4,$7F,$CE,$07,$C9,$3E,$29,$F2,$6F,$FC,$93
	.dc.b	$97,$C5,$8F,$A7,$87,$67,$DF,$7E,$17,$9D,$79,$5C,$E1,$E2,$3E,$1B
	.dc.b	$F6,$B1,$F9,$0F,$81,$F9,$0F,$8F,$0F,$05,$7D,$6C,$BE,$36,$1E,$B8
	.dc.b	$FD,$03,$FE,$94,$F3,$3F,$C9,$B7,$F9,$56,$BF,$DF,$57,$97,$FE,$5B
	.dc.b	$CF,$FB,$5F,$83,$FE,$51,$F3,$F1,$7F,$E5,$53,$FC,$29,$7C,$2B,$7E
	.dc.b	$6F,$F9,$5A,$74,$3C,$DF,$F2,$7B,$FE,$1B,$6B,$9F,$37,$FC,$49,$18
	.dc.b	$0F,$F9,$4F,$BF,$FC,$B7,$DD,$FE,$36,$9C,$E1,$FE,$54,$8C,$56,$A3
	.dc.b	$FC,$51,$E6,$7F,$95,$30,$ED,$1B,$3F,$C3,$97,$CA,$3F,$CA,$51,$C3
	.dc.b	$CD,$FF,$30,$C7,$F9,$7C,$C9,$FE,$4F,$F3,$7F,$CB,$93,$FE,$4C,$8C
	.dc.b	$F9,$9F,$E5,$CC,$BF,$CA,$44,$9F,$F1,$D7,$33,$CB,$FF,$2A,$5F,$41
	.dc.b	$E6,$7F,$96,$AB,$FC,$BD,$8F,$F9,$A1,$E3,$5C,$30,$E9,$7F,$E8,$6F
	.dc.b	$33,$E0,$36,$BF,$C4,$78,$B9,$DF,$07,$AA,$72,$FF,$97,$31,$FF,$30
	.dc.b	$AF,$37,$4D,$C7,$9F,$B0,$F2,$73,$5C,$F0,$FE,$56,$7F,$84,$FC,$EF
	.dc.b	$F2,$DB,$4F,$E4,$9D,$9F,$E5,$D5,$E1,$7F,$96,$DF,$F2,$C6,$7F,$E5
	.dc.b	$CA,$FF,$28,$4E,$7C,$4F,$F2,$F5,$70,$F3,$3F,$CB,$B1,$8F,$3B,$FC
	.dc.b	$B7,$87,$F9,$E9,$BA,$FF,$96,$5B,$7F,$E6,$BF,$55,$F0,$DF,$E5,$DB
	.dc.b	$E1,$E8,$5C,$38,$36,$FF,$CE,$58,$7F,$97,$5F,$F2,$1E,$0F,$83,$D4
	.dc.b	$3D,$1F,$F2,$E3,$7C,$AC,$C6,$35,$FF,$2F,$4F,$E0,$63,$64,$9E,$73
	.dc.b	$AC,$FF,$CB,$55,$FE,$52,$87,$83,$FE,$5E,$D3,$FC,$D7,$7F,$E5,$8A
	.dc.b	$FF,$44,$78,$DF,$E5,$2C,$3F,$CC,$11,$F8,$FF,$47,$FC,$BE,$EB,$97
	.dc.b	$82,$AE,$18,$7F,$A0,$6B,$73,$9A,$DD,$6A,$3F,$D2,$D3,$A8,$FF,$2F
	.dc.b	$C7,$F9,$A4,$FF,$31,$4F,$FA,$26,$33,$1F,$E5,$CF,$43,$FC,$BC,$BD
	.dc.b	$1F,$F2,$F3,$FE,$A1,$F4,$7F,$C9,$4F,$F9,$58,$99,$7F,$AE,$27,$70
	.dc.b	$9D,$CE,$6B,$FC,$C3,$3F,$EF,$1A,$FF,$32,$36,$3F,$D3,$9E,$67,$F8
	.dc.b	$2F,$D1,$FF,$14,$F9,$DF,$E1,$EA,$FF,$16,$7A,$3F,$E0,$D7,$FC,$4D
	.dc.b	$5B,$8C,$75,$FF,$82,$FA,$7F,$C4,$D5,$FE,$30,$7F,$C2,$77,$FE,$22
	.dc.b	$3F,$C4,$04,$8F,$F0,$A1,$FE,$47,$BF,$F1,$5C,$78,$6B,$FC,$6F,$3A
	.dc.b	$7B,$3E,$43,$FC,$49,$5D,$9F,$F1,$39,$93,$A7,$A3,$FE,$28,$3F,$C5
	.dc.b	$8F,$F9,$69,$B7,$FE,$26,$8F,$F1,$59,$D0,$F7,$C6,$8F,$F1,$B3,$C6
	.dc.b	$75,$5F,$E3,$A7,$C5,$19,$3C,$51,$E8,$3D,$6F,$63,$CB,$1F,$21,$FF
	.dc.b	$24,$BC,$AB,$C9,$8F,$D4,$9F,$98,$79,$27,$C3,$87,$CB,$8E,$97,$DE
	.dc.b	$3E,$84,$7C,$97,$FD,$05,$1E,$78,$E4,$79,$A3,$27,$3A,$F0,$47,$90
	.dc.b	$DC,$FA,$A3,$37,$F6,$27,$CB,$3E,$17,$C0,$7F,$9F,$4F,$F3,$7B,$97
	.dc.b	$CF,$5E,$07,$67,$F9,$C0,$DB,$F9,$37,$FC,$9D,$3F,$E5,$62,$61,$FE
	.dc.b	$5A,$EB,$FF,$2A,$9F,$E5,$68,$FD,$E7,$85,$FE,$51,$AF,$F2,$51,$FE
	.dc.b	$54,$3F,$C4,$27,$F9,$06,$FF,$C6,$8D,$3F,$C2,$9C,$1F,$E3,$17,$07
	.dc.b	$F8,$DB,$8F,$FC,$40,$FF,$85,$E3,$FC,$41,$97,$F8,$4A,$FF,$C3,$D3
	.dc.b	$FE,$19,$F0,$BF,$C2,$F1,$FE,$20,$8E,$47,$F8,$9B,$0F,$F1,$B3,$FE
	.dc.b	$2C,$30,$7F,$8A,$0C,$57,$56,$C1,$F2,$E3,$FE,$2A,$F4,$3F,$C4,$F3
	.dc.b	$FE,$3D,$AE,$07,$F8,$A7,$2F,$4C,$EA,$F8,$E5,$FE,$2B,$BF,$87,$1F
	.dc.b	$E6,$D7,$CB,$7C,$63,$CA,$7C,$1C,$7D,$F1,$FE,$3F,$8C,$1E,$63,$AC
	.dc.b	$7F,$99,$AB,$E6,$47,$53,$DD,$3D,$4F,$3D,$7A,$4F,$74,$7F,$94,$DF
	.dc.b	$2E,$3E,$FC,$BE,$6C,$FC,$FC,$BD,$B1,$8B,$F2,$65,$EF,$AF,$01,$E6
	.dc.b	$8F,$06,$1D,$B0,$FF,$28,$C7,$0B,$61,$FE,$AA,$AD,$62,$CB,$18,$FD
	.dc.b	$3F,$97,$7C,$77,$F8,$5B,$D3,$75,$BC,$2F,$F0,$A5,$6B,$3F,$B5,$CF
	.dc.b	$DB,$16,$1C,$2F,$78,$E6,$7F,$D6,$18,$71,$7F,$D6,$41,$C3,$A8,$E0
	.dc.b	$FC,$A8,$E1,$7C,$DB,$87,$E2,$F0,$F0,$E9,$E0,$8F,$29,$C7,$0C,$61
	.dc.b	$FE,$46,$AD,$CE,$A3,$E1,$4F,$F9,$AB,$0F,$F2,$DF,$AA,$F5,$B1,$7F
	.dc.b	$02,$30,$F1,$B7,$A7,$FE,$78,$7F,$CE,$3E,$AF,$A9,$B3,$F0,$1F,$C2
	.dc.b	$F1,$7E,$8F,$E7,$5F,$2D,$F8,$8F,$2D,$F2,$3F,$36,$73,$7F,$E3,$60
	.dc.b	$E1,$FF,$23,$C6,$EB,$FC,$7B,$39,$AE,$78,$77,$9C,$9D,$EB,$85,$F5
	.dc.b	$E2,$E1,$1F,$75,$ED,$1F,$E4,$A7,$35,$FB,$B7,$30,$C7,$B3,$FB,$FC
	.dc.b	$3C,$2F,$A3,$0F,$0B,$9C,$3C,$2F,$C4,$34,$67,$0E,$6E,$71,$F0,$BA
	.dc.b	$31,$5E,$B9,$DC,$7F,$98,$27,$67,$9B,$0E,$78,$79,$23,$13,$CE,$31
	.dc.b	$7E,$68,$F9,$B7,$D1,$F7,$1B,$C3,$D2,$F7,$9F,$13,$E5,$30,$7A,$CC
	.dc.b	$1C,$AF,$CB,$1D,$EF,$C3,$3F,$7E,$72,$D9,$EB,$1E,$83,$31,$F0,$8E
	.dc.b	$CF,$B3,$B3,$FC,$EB,$1F,$E8,$26,$BE,$A6,$E3,$FE,$7D,$8F,$E6,$DD
	.dc.b	$5F,$E7,$DC,$FF,$CF,$FC,$BF,$E7,$C6,$F0,$EE,$BF,$D0,$0D,$AF,$F4
	.dc.b	$46,$3F,$95,$F9,$E7,$CF,$FF,$24,$8C,$37,$4E,$78,$7D,$BB,$E3,$1F
	.dc.b	$CA,$AF,$8C,$FE,$09,$AE,$F2,$63,$F8,$6B,$E3,$3F,$D3,$EB,$78,$F8
	.dc.b	$23,$57,$C1,$E5,$7D,$71,$F2,$47,$16,$BE,$E3,$67,$E4,$6F,$AE,$3E
	.dc.b	$79,$DE,$3E,$2A,$E1,$8F,$85,$E1,$3A,$C7,$C5,$3F,$BB,$AE,$18,$74
	.dc.b	$BF,$11,$E8,$C6,$3C,$8F,$41,$8A,$E5,$7D,$31,$F2,$4E,$8F,$4C,$EA
	.dc.b	$BF,$75,$3A,$9F,$E8,$0D,$CB,$51,$83,$F0,$B7,$D5,$F2,$9E,$33,$AC
	.dc.b	$FD,$CB,$F8,$04,$70,$7F,$66,$F0,$7F,$CF,$F3,$FD,$11,$AF,$F6,$A3
	.dc.b	$FD,$21,$E0,$FF,$A1,$9D,$3F,$EB,$CF,$09,$ED,$53,$FD,$2F,$F3,$2F
	.dc.b	$93,$FF,$60,$B6,$FF,$9C,$35,$FE,$89,$DD,$FE,$AD,$73,$3C,$08,$FF
	.dc.b	$47,$EC,$FE,$D5,$DF,$88,$FE,$B7,$7C,$DF,$43,$5F,$F0,$1B,$5F,$F1
	.dc.b	$13,$E1,$E0,$FE,$D5,$4F,$80,$7D,$07,$FB,$76,$30,$F7,$9F,$ED,$9E
	.dc.b	$1F,$FA,$77,$0F,$F3,$69,$FE,$5C,$F0,$7F,$D2,$87,$F9,$B1,$FF,$2C
	.dc.b	$9F,$E5,$DE,$8F,$F3,$61,$FE,$5D,$F1,$BF,$CC,$BE,$37,$FA,$47,$AF
	.dc.b	$9D,$70,$3F,$CB,$D5,$C2,$FA,$C7,$FA,$6A,$3F,$CD,$DE,$3F,$FA,$BA
	.dc.b	$73,$3F,$E8,$D9,$CB,$D3,$1F,$F3,$93,$FE,$77,$8F,$15,$71,$8F,$51
	.dc.b	$0D,$57,$4A,$FF,$47,$CE,$E7,$26,$E3,$A9,$30,$F6,$47,$FA,$24,$F8
	.dc.b	$55,$FE,$89,$AE,$55,$BC,$3F,$D3,$CD,$AF,$F3,$E9,$D6,$F9,$E1,$EC
	.dc.b	$8F,$F5,$29,$A3,$C6,$49,$E5,$1F,$26,$39,$57,$6A,$E7,$87,$D7,$C7
	.dc.b	$E9,$CE,$49,$5D,$70,$5F,$B2,$3F,$D1,$EF,$FB,$C9,$FF,$22,$35,$FF
	.dc.b	$38,$47,$09,$FF,$09,$75,$FF,$A4,$BC,$AF,$F1,$CD,$7F,$8D,$21,$18
	.dc.b	$8F,$F4,$91,$0F,$F2,$F1,$FE,$66,$3F,$D2,$A7,$F9,$5B,$C8,$FF,$2C
	.dc.b	$5F,$F9,$7F,$C1,$FF,$2E,$92,$7F,$CE,$73,$AE,$3F,$F3,$6D,$6E,$73
	.dc.b	$E4,$7F,$98,$F5,$FF,$3D,$9E,$08,$FF,$36,$1F,$EA,$0E,$6F,$F3,$B1
	.dc.b	$1F,$1D,$73,$DF,$FC,$EB,$1E,$77,$FC,$CD,$97,$8D,$CC,$7A,$70,$F0
	.dc.b	$3F,$E7,$EC,$3C,$17,$FE,$6E,$7C,$71,$CD,$AF,$1B,$E9,$1F,$E7,$0C
	.dc.b	$3F,$D1,$EE,$AB,$C8,$E6,$BF,$D2,$07,$B0,$FF,$57,$E1,$E4,$9D,$C7
	.dc.b	$73,$E0,$1E,$EC,$7E,$49,$2F,$E1,$57,$B3,$2F,$9D,$18,$C3,$FD,$27
	.dc.b	$5B,$BE,$53,$97,$AD,$F1,$CB,$FC,$FC,$F8,$EB,$FD,$47,$E4,$7F,$9E
	.dc.b	$2B,$FC,$DD,$9F,$F9,$76,$31,$1F,$E6,$C8,$C7,$91,$FE,$72,$E3,$FE
	.dc.b	$3F,$CA,$D3,$15,$CA,$FF,$BD,$B6,$B7,$3A,$9E,$86,$8B,$1A,$9F,$F3
	.dc.b	$A6,$1F,$E8,$69,$FF,$38,$9F,$E9,$0F,$1B,$FC,$EC,$FF,$9E,$3C,$6F
	.dc.b	$F3,$B3,$6B,$FC,$F9,$29,$FF,$4F,$5F,$FA,$7F,$C5,$7B,$DF,$D4,$2F
	.dc.b	$75,$A7,$FC,$75,$E3,$FF,$91,$31,$FF,$1F,$78,$5F,$E3,$83,$FD,$43
	.dc.b	$C1,$FD,$C2,$3F,$C8,$9E,$2F,$F9,$48,$FF,$2E,$9F,$E3,$D9,$E1,$1F
	.dc.b	$E4,$13,$FC,$7A,$F6,$BF,$F2,$A5,$7F,$97,$2F,$C0,$F5,$AF,$F2,$C9
	.dc.b	$FD,$E4,$FF,$37,$5F,$6B,$E5,$1F,$E5,$49,$E5,$7D,$23,$FC,$BC,$DF
	.dc.b	$13,$FC,$9F,$1F,$E5,$C3,$AB,$E7,$7C,$A7,$FA,$2E,$BF,$CC,$77,$ED
	.dc.b	$69,$E0,$9E,$13,$CA,$3E,$2B,$E7,$CF,$FD,$13,$5C,$63,$E3,$CE,$A3
	.dc.b	$D5,$1D,$CC,$9F,$EA,$7E,$9F,$F2,$EA,$69,$23,$FC,$A2,$FF,$B7,$E3
	.dc.b	$FC,$3B,$AF,$F9,$6F,$C5,$FF,$30,$46,$0E,$55,$FE,$58,$3F,$C7,$84
	.dc.b	$BC,$12,$3F,$C4,$B1,$97,$93,$FE,$2C,$C7,$FC,$49,$1F,$E2,$A7,$13
	.dc.b	$C6,$1E,$37,$F9,$82,$B9,$47,$F9,$4D,$DB,$C6,$7F,$C7,$55,$88,$FF
	.dc.b	$28,$C6,$08,$FA,$E3,$55,$FE,$4F,$8E,$FE,$3F,$F9,$57,$1F,$F2,$7E
	.dc.b	$1E,$0A,$F1,$61,$E5,$EC,$FF,$2B,$DF,$F9,$9A,$79,$47,$F9,$84,$CE
	.dc.b	$1F,$E5,$1C,$7C,$90,$CB,$FC,$E5,$8F,$5C,$17,$F9,$9A,$D6,$56,$FF
	.dc.b	$CA,$13,$E6,$3A,$65,$FE,$4C,$78,$F8,$FF,$E4,$CA,$6D,$FE,$4F,$F1
	.dc.b	$FA,$D7,$F0,$93,$FC,$9A,$7F,$BE,$FC,$7F,$C3,$3F,$EF,$DB,$FF,$29
	.dc.b	$BC,$4F,$F7,$D3,$87,$BC,$FF,$BF,$57,$8B,$FE,$3D,$BF,$F2,$0F,$8D
	.dc.b	$FE,$39,$7D,$BD,$5F,$E4,$AA,$FF,$27,$90,$F2,$57,$15,$1B,$73,$5C
	.dc.b	$DF,$F2,$5B,$FE,$47,$EE,$FF,$2D,$46,$8F,$F3,$51,$A2,$9E,$68,$F4
	.dc.b	$57,$F9,$56,$3F,$C7,$FD,$7F,$E2,$57,$FC,$71,$E1,$7F,$85,$9F,$F2
	.dc.b	$04,$7F,$85,$FC,$4F,$F1,$4C,$E6,$7F,$C8,$13,$97,$8F,$37,$F8,$6B
	.dc.b	$A3,$FC,$45,$1F,$E2,$5A,$E1,$E1,$7F,$8F,$23,$67,$F8,$7E,$32,$7D
	.dc.b	$49,$FF,$23,$C6,$9E,$73,$FE,$2E,$BF,$44,$FF,$8D,$27,$35,$CA,$FF
	.dc.b	$1B,$1C,$A7,$55,$C2,$3F,$AA,$47,$F9,$18,$FC,$9F,$89,$FE,$34,$7C
	.dc.b	$95,$EB,$AE,$07,$F8,$C1,$FF,$34,$CE,$EB,$C3,$1B,$AE,$78,$FC,$88
	.dc.b	$C4,$7F,$9B,$F1,$FF,$3E,$3E,$43,$67,$BF,$0E,$EF,$94,$E1,$8F,$C5
	.dc.b	$8F,$A1,$1C,$4C,$D7,$92,$FC,$11,$EA,$26,$1E,$66,$D7,$67,$F0,$67
	.dc.b	$88,$E9,$8A,$BF,$57,$5F,$22,$58,$FF,$93,$DF,$F3,$81,$8C,$7F,$CA
	.dc.b	$8D,$C7,$F2,$73,$B8,$FF,$2D,$C7,$F9,$5B,$C4,$FF,$29,$47,$EE,$39
	.dc.b	$BF,$CB,$1B,$F0,$F0,$9C,$14,$95,$FA,$0C,$53,$A8,$C3,$FE,$21,$CF
	.dc.b	$FC,$3D,$C3,$FE,$38,$9F,$F1,$71,$FE,$32,$DF,$FC,$6D,$3F,$E2,$1C
	.dc.b	$3F,$13,$1F,$E3,$BB,$E4,$FF,$8B,$23,$FC,$55,$5C,$4E,$78,$78,$A3
	.dc.b	$8C,$7F,$93,$4E,$FE,$27,$F8,$D9,$FF,$19,$93,$0F,$2B,$FE,$64,$C7
	.dc.b	$CC,$FF,$91,$EF,$A2,$9F,$F2,$EB,$B2,$3E,$97,$A6,$1E,$76,$9F,$E4
	.dc.b	$07,$11,$E0,$38,$15,$F8,$67,$F9,$93,$1F,$99,$1A,$8F,$89,$8F,$C2
	.dc.b	$BF,$2C,$F9,$0D,$5F,$72,$1F,$1C,$F6,$5F,$4B,$E7,$1E,$63,$55,$FE
	.dc.b	$73,$7E,$51,$FE,$71,$3D,$F1,$E6,$7E,$6F,$89,$FE,$7C,$59,$7C,$FC
	.dc.b	$3D,$91,$F5,$B0,$F0,$3D,$30,$FF,$38,$BE,$4C,$7E,$CB,$CB,$0E,$98
	.dc.b	$73,$C7,$E4,$92,$7F,$D6,$38,$74,$78,$65,$CB,$1F,$8B,$5F,$D5,$3C
	.dc.b	$2F,$F2,$84,$FF,$62,$E8,$FF,$28,$F9,$0F,$27,$FC,$9D,$8F,$F9,$72
	.dc.b	$73,$18,$8F,$F2,$D1,$FE,$65,$84,$F1,$AF,$F2,$B9,$D4,$EC,$E6,$7F
	.dc.b	$B2,$5F,$89,$E5,$19,$AF,$ED,$71,$E1,$AE,$37,$E1,$BF,$F3,$69,$FE
	.dc.b	$6E,$FA,$37,$D1,$FA,$29,$F4,$78,$7C,$58,$7F,$80,$2B,$0F,$F9,$F6
	.dc.b	$37,$5E,$D9,$CE,$3E,$29,$C9,$CA,$93,$A8,$FF,$05,$9A,$8F,$F0,$01
	.dc.b	$FE,$59,$F1,$1E,$2F,$F9,$A2,$3F,$CB,$E7,$F9,$8C,$BE,$27,$F9,$5A
	.dc.b	$77,$5F,$E6,$3F,$07,$FC,$EB,$5F,$C8,$8D,$BF,$E5,$D2,$3F,$E7,$93
	.dc.b	$FC,$BC,$59,$E6,$F7,$AF,$F3,$6B,$FE,$88,$3F,$CD,$67,$57,$3C,$BF
	.dc.b	$E7,$68,$E1,$5E,$27,$D9,$87,$AE,$B7,$1E,$43,$C1,$19,$9E,$11,$DC
	.dc.b	$F2,$9D,$DC,$D7,$7A,$F8,$55,$F1,$FA,$7F,$CD,$57,$FE,$6F,$F1,$3F
	.dc.b	$CF,$31,$FE,$6B,$87,$4F,$F8,$E9,$FF,$34,$9F,$E6,$B8,$D4,$65,$BE
	.dc.b	$1F,$F9,$C2,$7F,$CB,$27,$F9,$74,$86,$09,$E2,$FF,$96,$4F,$F2,$8C
	.dc.b	$7F,$91,$AB,$FC,$A3,$2C,$7F,$CD,$EF,$18,$CC,$61,$FF,$2C,$4F,$F9
	.dc.b	$3A,$BA,$78,$BF,$E6,$3B,$E5,$A7,$F9,$3D,$FF,$29,$18,$3C,$B7,$CB
	.dc.b	$83,$FC,$AB,$18,$C3,$C8,$75,$BF,$F3,$41,$D7,$1F,$F3,$9B,$89,$FF
	.dc.b	$31,$C7,$F9,$A2,$78,$47,$43,$D4,$E1,$F1,$1E,$48,$E1,$5B,$C7,$E2
	.dc.b	$BD,$31,$FF,$31,$FC,$23,$E2,$BF,$CC,$B3,$FB,$EC,$1E,$A3,$E6,$B3
	.dc.b	$C7,$FE,$6A,$8C,$4F,$EC,$70,$E4,$DC,$36,$FF,$9F,$27,$6F,$F9,$C5
	.dc.b	$B8,$77,$AE,$33,$FE,$75,$F2,$3F,$CD,$CF,$F9,$D8,$FF,$2D,$B8,$27
	.dc.b	$91,$DA,$1E,$4B,$D0,$C3,$63,$FC,$F9,$3B,$AF,$F4,$36,$7F,$E8,$4A
	.dc.b	$D5,$7F,$A1,$5B,$8F,$C3,$AD,$CF,$FA,$00,$FF,$44,$61,$D8,$E9,$87
	.dc.b	$8F,$BF,$FC,$77,$CF,$FE,$25,$6F,$7F,$F9,$1E,$FF,$C4,$B1,$87,$FC
	.dc.b	$79,$E4,$7F,$92,$67,$FC,$65,$39,$9F,$F1,$99,$FE,$3C,$38,$F8,$FF
	.dc.b	$E4,$17,$FC,$EA,$43,$FC,$E6,$58,$FF,$37,$3F,$E3,$AF,$23,$FC,$87
	.dc.b	$87,$F9,$04,$97,$E0,$BE,$95,$C8,$97,$CA,$BD,$93,$A8,$C5,$7B,$67
	.dc.b	$3D,$3F,$D7,$EB,$FC,$E5,$7F,$E1,$7A,$D1,$DE,$7F,$CA,$D1,$FE,$5F
	.dc.b	$3F,$CA,$15,$C9,$F7,$C6,$9F,$51,$B3,$E2,$46,$CF,$F3,$F1,$67,$FC
	.dc.b	$CA,$D8,$F8,$95,$E9,$3C,$B1,$C2,$7C,$B3,$AA,$F5,$1F,$E7,$8C,$7E
	.dc.b	$B4,$79,$A7,$6F,$C6,$BE,$95,$BC,$7E,$7D,$7B,$EB,$D6,$4C,$3C,$8F
	.dc.b	$FA,$96,$B8,$3E,$DC,$3B,$BF,$81,$3F,$CC,$E7,$F9,$7F,$0F,$F3,$6E
	.dc.b	$3F,$E6,$B8,$FF,$34,$79,$1F,$E6,$A7,$FD,$DB,$D5,$FE,$6A,$7F,$CC
	.dc.b	$7E,$3F,$F9,$A1,$FF,$72,$9F,$E7,$AF,$17,$FC,$D0,$E2,$FF,$C9,$ED
	.dc.b	$8F,$F3,$0B,$FE,$3A,$8F,$F0,$DB,$FE,$29,$8C,$BF,$E3,$87,$FC,$45
	.dc.b	$1F,$E2,$C3,$FC,$75,$1F,$E2,$43,$FC,$72,$7F,$8C,$E7,$FC,$5F,$E3
	.dc.b	$7F,$8B,$63,$B5,$7F,$8A,$E3,$11,$D6,$7F,$C5,$93,$E0,$AF,$1C,$F5
	.dc.b	$D3,$FC,$A2,$EC,$FC,$91,$FE,$46,$E4,$FF,$36,$CF,$93,$0E,$AF,$26
	.dc.b	$DF,$87,$C7,$FF,$20,$92,$B8,$C7,$B8,$F1,$1B,$73,$8F,$98,$FF,$33
	.dc.b	$46,$70,$F8,$27,$F9,$3C,$F2,$9E,$9D,$FF,$CB,$8F,$F9,$7F,$2F,$5E
	.dc.b	$5E,$5C,$7C,$53,$CA,$3D,$51,$89,$DD,$F7,$6E,$1F,$3A,$75,$87,$6B
	.dc.b	$FA,$D7,$EB,$AF,$39,$BC,$7D,$78,$F9,$AB,$73,$ED,$8F,$F3,$C4,$79
	.dc.b	$67,$18,$FD,$2C,$99,$7F,$93,$23,$F9,$36,$3F,$E6,$7F,$17,$FC,$A5
	.dc.b	$CB,$FE,$5E,$C3,$FC,$B3,$97,$F9,$53,$C6,$FF,$28,$79,$7F,$E6,$4A
	.dc.b	$FF,$2F,$C6,$2B,$FC,$9B,$18,$78,$D7,$E7,$AF,$FC,$AF,$7F,$D7,$E3
	.dc.b	$FC,$E4,$F1,$55,$FE,$60,$F1,$1E,$67,$F4,$DC,$3C,$73,$FD,$9F,$C6
	.dc.b	$FF,$28,$5A,$7F,$CA,$13,$9F,$1F,$FC,$C5,$E4,$6A,$BF,$CA,$F1,$83
	.dc.b	$FC,$BC,$5A,$FF,$2E,$47,$F8,$F1,$FF,$28,$3F,$E4,$43,$07,$F9,$08
	.dc.b	$C1,$FE,$3B,$C7,$FC,$9B,$E2,$7F,$8E,$27,$8C,$FF,$8F,$8E,$31,$97
	.dc.b	$91,$93,$97,$91,$FE,$47,$7A,$F0,$7F,$90,$FC,$4F,$F2,$EC,$7F,$91
	.dc.b	$CC,$9F,$E4,$F3,$07,$38,$E0,$FF,$94,$AF,$C7,$87,$9A,$B9,$CF,$91
	.dc.b	$AF,$8A,$7F,$C3,$0F,$F9,$3C,$87,$F9,$C7,$AF,$FC,$A6,$FF,$94,$3C
	.dc.b	$6F,$F2,$57,$4F,$F9,$37,$C8,$4F,$F9,$45,$AE,$F7,$CD,$FF,$94,$9F
	.dc.b	$F1,$47,$91,$FE,$35,$CB,$FC,$81,$E4,$7F,$91,$2F,$C6,$7F,$8C,$AB
	.dc.b	$6E,$31,$FF,$21,$3F,$E3,$49,$E0,$FF,$8D,$A3,$3B,$FF,$95,$27,$FC
	.dc.b	$6B,$39,$CF,$FC,$79,$3F,$E5,$57,$FC,$93,$7D,$A7,$4F,$78,$DE,$1E
	.dc.b	$CB,$E7,$7F,$E4,$CB,$FF,$24,$61,$E0,$AD,$63,$FE,$4F,$9E,$64,$AF
	.dc.b	$24,$7C,$18,$FE,$93,$87,$67,$FC,$B9,$8F,$C2,$F1,$7F,$C7,$2F,$F9
	.dc.b	$27,$C6,$FF,$1E,$ED,$FE,$47,$F2,$3F,$C6,$D5,$A8,$FF,$28,$F9,$1F
	.dc.b	$E3,$9A,$DE,$1F,$E4,$7B,$DC,$FF,$8B,$23,$FC,$72,$43,$FC,$6D,$BE
	.dc.b	$5F,$F2,$2F,$8B,$FD,$69,$FB,$BE,$3F,$F9,$19,$C1,$FE,$4B,$CF,$FC
	.dc.b	$BD,$E4,$7F,$8F,$B3,$FF,$21,$79,$1C,$E7,$FC,$81,$87,$F8,$F6,$FF
	.dc.b	$C7,$E7,$86,$3F,$CC,$DE,$17,$F8,$D4,$9E,$37,$F8,$02,$7F,$92,$3F
	.dc.b	$E0,$CF,$17,$FC,$0B,$E3,$FF,$7A,$27,$91,$FE,$04,$AF,$F0,$51,$FE
	.dc.b	$42,$70,$FF,$87,$27,$FC,$32,$7F,$83,$AF,$FC,$5F,$1B,$3F,$C0,$D3
	.dc.b	$C2,$BA,$C6,$08,$7F,$8A,$09,$D7,$FE,$10,$AE,$71,$FE,$4E,$9E,$2E
	.dc.b	$AF,$C1,$87,$B6,$FF,$C8,$F3,$E8,$38,$61,$E1,$23,$E0,$8E,$D3,$C6
	.dc.b	$BF,$C4,$51,$8D,$FF,$C3,$F5,$DC,$F6,$14,$EA,$73,$9F,$1C,$7F,$8D
	.dc.b	$23,$FB,$3E,$1E,$8A,$E9,$1F,$4A,$7A,$C6,$0F,$F1,$F4,$76,$71,$5F
	.dc.b	$5B,$0F,$E0,$B1,$89,$FE,$13,$8B,$1F,$F3,$2E,$2B,$DE,$1F,$E7,$DB
	.dc.b	$FF,$1A,$90,$C1,$FE,$34,$F2,$BF,$27,$D3,$FE,$33,$8F,$F1,$67,$91
	.dc.b	$FE,$33,$7F,$8A,$E5,$FE,$30,$6F,$8D,$FE,$34,$7F,$BB,$3F,$DE,$F0
	.dc.b	$B5,$FD,$DA,$7F,$C1,$3E,$47,$F8,$62,$3F,$BC,$47,$F8,$16,$BF,$BE
	.dc.b	$3F,$DD,$9E,$4F,$F8,$C4,$E3,$E4,$7F,$81,$1C,$6C,$C3,$F9,$4E,$1D
	.dc.b	$67,$7C,$7F,$E0,$4A,$FE,$FD,$7C,$67,$FC,$39,$1C,$5F,$F2,$36,$3E
	.dc.b	$78,$C4,$7F,$87,$CE,$F8,$76,$AE,$59,$7F,$87,$30,$ED,$3F,$02,$BC
	.dc.b	$27,$B7,$0F,$F0,$7C,$7F,$8A,$31,$F3,$63,$D9,$B1,$D6,$F9,$DF,$F9
	.dc.b	$0C,$E4,$7F,$87,$72,$FF,$1D,$E1,$F2,$B0,$FF,$14,$CF,$81,$A7,$5C
	.dc.b	$BC,$F5,$F3,$A3,$FC,$3B,$5C,$70,$F4,$46,$2F,$94,$60,$F4,$5F,$2A
	.dc.b	$F5,$CF,$F9,$89,$E4,$FC,$63,$E5,$9F,$6D,$CB,$F4,$27,$DC,$FB,$2B
	.dc.b	$EA,$E1,$F0,$23,$FC,$99,$3E,$17,$B1,$E4,$3B,$1F,$54,$E8,$7F,$6C
	.dc.b	$C7,$FC,$77,$3E,$68,$78,$FF,$E3,$A8,$FE,$E3,$C5,$FE,$33,$9F,$F1
	.dc.b	$6C,$7F,$8C,$0F,$F1,$BF,$94,$F2,$7F,$C6,$53,$FE,$37,$F2,$7F,$C6
	.dc.b	$CF,$F7,$AC,$3F,$C7,$37,$FE,$32,$3F,$BE,$F8,$9F,$E3,$6E,$8F,$F1
	.dc.b	$D6,$5F,$E3,$3B,$E1,$86,$2F,$7E,$4F,$F8,$EE,$7C,$66,$A3,$CE,$7F
	.dc.b	$81,$BC,$9F,$F1,$EB,$FE,$3C,$31,$E4,$7F,$8D,$E7,$89,$FE,$3C,$8F
	.dc.b	$E9,$F2,$9F,$4C,$E4,$FE,$AD,$19,$9F,$EA,$B1,$FE,$21,$3F,$C4,$C7
	.dc.b	$F8,$D3,$2F,$F1,$33,$8F,$11,$E2,$FF,$5F,$E2,$FF,$0E,$F8,$F8,$AF
	.dc.b	$87,$E4,$7C,$18,$C4,$7F,$1C,$0E,$30,$4E,$EF,$FC,$63,$7F,$E3,$8B
	.dc.b	$FF,$1D,$D7,$89,$B5,$FE,$45,$7F,$C7,$C7,$F9,$04,$F5,$1A,$C7,$FC
	.dc.b	$7F,$1B,$9E,$53,$A2,$1F,$E3,$F3,$E0,$46,$C9,$3C,$89,$0C,$7D,$AF
	.dc.b	$F8,$FA,$FF,$C7,$78,$FC,$58,$77,$7F,$91,$9D,$18,$8F,$F2,$1F,$57
	.dc.b	$F9,$1E,$3F,$C8,$C4,$3F,$C8,$C6,$CF,$F1,$1F,$07,$F8,$CF,$0F,$9B
	.dc.b	$3F,$E2,$D9,$FF,$27,$F9,$1F,$E4,$09,$CF,$8F,$FE,$3A,$9F,$F2,$27
	.dc.b	$91,$FE,$31,$9C,$E1,$FE,$3B,$3A,$6F,$E4,$BE,$13,$FE,$34,$BD,$1F
	.dc.b	$E3,$79,$FF,$1D,$BF,$E3,$E9,$DC,$62,$3B,$9C,$27,$FC,$88,$E0,$E4
	.dc.b	$7B,$0C,$3F,$E5,$1A,$EF,$97,$E2,$B1,$63,$FE,$6A,$8C,$37,$0F,$45
	.dc.b	$7E,$51,$FF,$34,$C7,$BD,$CE,$1E,$01,$7F,$1E,$33,$87,$E6,$3C,$87
	.dc.b	$8D,$FA,$66,$DF,$F8,$F1,$AF,$F9,$05,$AF,$F9,$F8,$3F,$E5,$17,$FC
	.dc.b	$9C,$7E,$F8,$A7,$EE,$9D,$3F,$E4,$7D,$FF,$C9,$7E,$27,$F9,$15,$FF
	.dc.b	$44,$F8,$BF,$8C,$9F,$F2,$24,$69,$A7,$FA,$42,$34,$FF,$06,$F1,$BF
	.dc.b	$CA,$8F,$F9,$60,$FF,$2A,$B8,$D7,$FC,$AB,$D3,$FE,$51,$9E,$18,$7F
	.dc.b	$A6,$27,$FC,$A3,$E3,$7F,$8A,$E7,$FC,$2A,$FF,$93,$DB,$C7,$FE,$20
	.dc.b	$EF,$FF,$13,$C7,$F8,$8C,$FF,$13,$F8,$BF,$E3,$67,$FC,$5D,$E2,$7F
	.dc.b	$96,$AB,$FC,$59,$CF,$FE,$26,$9F,$F1,$89,$AC,$3C,$0A,$BF,$C4,$95
	.dc.b	$DE,$3F,$C5,$98,$78,$2F,$95,$75,$9F,$14,$62,$F9,$CF,$29,$F2,$9F
	.dc.b	$E3,$F3,$99,$0F,$1C,$FF,$8D,$5F,$14,$FC,$16,$DF,$8D,$DD,$F8,$CF
	.dc.b	$F1,$F3,$D4,$C9,$D6,$FC,$D7,$D7,$C6,$FF,$26,$57,$1A,$F9,$A7,$98
	.dc.b	$F8,$97,$F2,$23,$AC,$60,$F7,$B8,$AF,$45,$FB,$DF,$F3,$5C,$FF,$9F
	.dc.b	$CE,$AF,$C5,$AE,$91,$8A,$FF,$34,$9F,$E8,$47,$D1,$87,$CA,$3E,$15
	.dc.b	$F5,$9D,$CF,$12,$1D,$CF,$8E,$79,$8F,$F3,$E4,$7D,$77,$FC,$B4,$7C
	.dc.b	$02,$8E,$8F,$F2,$D3,$FE,$5D,$3F,$D6,$51,$FE,$59,$7F,$CD,$3E,$2F
	.dc.b	$F9,$65,$BE,$27,$F9,$7D,$FF,$18,$65,$FE,$2A,$C3,$FC,$2E,$D8,$FF
	.dc.b	$0E,$BF,$E2,$73,$FC,$5D,$C5,$FE,$19,$C1,$87,$F8,$DA,$BF,$C3,$07
	.dc.b	$F8,$58,$FF,$0C,$47,$F9,$25,$FF,$11,$5F,$F8,$7D,$CC,$FF,$8D,$E3
	.dc.b	$FC,$63,$5D,$63,$A5,$F8,$2B,$89,$E8,$AE,$57,$E2,$C3,$FC,$62,$F7
	.dc.b	$AF,$55,$76,$9F,$F1,$34,$EE,$3C,$EE,$EB,$9D,$FF,$9D,$27,$D5,$E2
	.dc.b	$7F,$96,$1F,$87,$3E,$67,$C6,$7F,$8E,$8F,$F1,$BC,$6D,$F0,$1F,$E5
	.dc.b	$BB,$FF,$38,$4F,$CE,$3B,$BE,$8A,$F8,$53,$FE,$42,$C3,$E0,$1E,$AB
	.dc.b	$E8,$4F,$17,$FC,$AD,$3D,$4F,$F2,$91,$EC,$AE,$6F,$F0,$77,$D9,$8F
	.dc.b	$BE,$3C,$F3,$CA,$FF,$CA,$B1,$EC,$9F,$8D,$1F,$81,$C7,$D1,$3C,$32
	.dc.b	$FA,$D9,$7B,$23,$E2,$9E,$99,$DB,$F2,$B0,$FF,$22,$5F,$AD,$B3,$FE
	.dc.b	$AE,$AF,$F0,$53,$FE,$55,$AF,$E9,$47,$F9,$73,$C2,$FF,$1D,$F8,$7F
	.dc.b	$E5,$28,$FF,$20,$BF,$E5,$F5,$E0,$7F,$94,$F2,$FF,$2B,$47,$F9,$34
	.dc.b	$DD,$7F,$95,$78,$FF,$CA,$D5,$B9,$CD,$70,$7C,$0F,$F5,$78,$D4,$FF
	.dc.b	$97,$67,$27,$F9,$4B,$C1,$FF,$26,$3F,$E4,$EF,$0B,$FC,$A3,$18,$8F
	.dc.b	$EC,$8D,$7F,$CB,$77,$FD,$8D,$FF,$2B,$57,$F9,$A6,$1E,$4B,$D8,$F4
	.dc.b	$47,$0A,$FF,$30,$9F,$E3,$AD,$FF,$C9,$0F,$F8,$BB,$C2,$FF,$27,$78
	.dc.b	$9F,$E4,$D8,$FE,$E9,$3F,$E0,$57,$FC,$AF,$55,$FF,$24,$1F,$E4,$27
	.dc.b	$FC,$7F,$E0,$7F,$8F,$A3,$FC,$8C,$D9,$E1,$AF,$F9,$6A,$B9,$57,$87
	.dc.b	$C1,$3D,$8F,$F3,$0B,$FE,$5E,$7F,$CB,$07,$47,$17,$E8,$9F,$0C,$7F
	.dc.b	$92,$F1,$C5,$72,$3D,$65,$3F,$03,$8F,$BD,$F5,$DF,$9D,$EF,$8F,$A2
	.dc.b	$75,$1E,$73,$CD,$5B,$3B,$96,$FD,$71,$DB,$0F,$7A,$BF,$F3,$0B,$E4
	.dc.b	$E9,$FF,$30,$39,$3F,$CB,$47,$F9,$71,$FF,$41,$BF,$E5,$FE,$8F,$F3
	.dc.b	$13,$9C,$3F,$C6,$E7,$F8,$F9,$FA,$DE,$17,$F9,$09,$FB,$12,$F1,$1C
	.dc.b	$18,$6C,$FC,$5A,$E0,$DF,$13,$FC,$90,$FF,$94,$70,$FF,$2A,$ED,$FE
	.dc.b	$51,$8F,$F3,$F6,$BE,$23,$FC,$B9,$3E,$38,$FA,$19,$78,$72,$F0,$1F
	.dc.b	$E6,$A8,$FF,$2E,$5F,$4B,$FF,$32,$3F,$E7,$08,$FF,$29,$63,$FE,$51
	.dc.b	$77,$3D,$DF,$F3,$DE,$3E,$67,$FC,$AE,$DC,$3C,$11,$E2,$CB,$DF,$87
	.dc.b	$F9,$DE,$3F,$CB,$D3,$A8,$C1,$E1,$3C,$59,$7A,$C9,$87,$86,$BF,$D2
	.dc.b	$33,$C2,$3F,$CC,$A7,$E0,$BC,$3F,$F2,$B9,$3C,$4F,$F2,$36,$FF,$86
	.dc.b	$F2,$9E,$E6,$31,$FD,$1D,$FF,$99,$9E,$57,$FE,$6D,$23,$FE,$54,$F0
	.dc.b	$FF,$CC,$0F,$EA,$BC,$54,$73,$87,$8D,$FE,$5E,$E2,$FE,$29,$5A,$9F
	.dc.b	$F3,$19,$FE,$6B,$3F,$CC,$9C,$7D,$1C,$12,$B9,$69,$FE,$6B,$8F,$F3
	.dc.b	$31,$0F,$F3,$E5,$75,$7F,$C1,$FE,$1F,$F8,$86,$FF,$C8,$1E,$0F,$F8
	.dc.b	$EB,$BF,$FC,$4D,$5F,$E3,$E2,$46,$5F,$F1,$1B,$FE,$27,$7F,$C4,$35
	.dc.b	$CD,$FF,$1F,$73,$FF,$8E,$23,$FC,$5A,$4B,$EE,$EE,$3F,$C7,$7E,$27
	.dc.b	$F9,$22,$BF,$C6,$B5,$FE,$3A,$7F,$C5,$F3,$AF,$17,$FC,$E4,$F5,$9E
	.dc.b	$66,$F0,$F2,$57,$F8,$DD,$A6,$2F,$B4,$7B,$30,$F1,$63,$FE,$40,$8E
	.dc.b	$EF,$F9,$44,$FF,$2D,$38,$3D,$F7,$DE,$7D,$91,$F0,$27,$77,$DF,$1F
	.dc.b	$3C,$62,$B7,$8F,$AE,$FF,$CD,$97,$F3,$5E,$C4,$C7,$CF,$19,$3D,$75
	.dc.b	$CA,$FD,$11,$8C,$7E,$1D,$FF,$95,$4E,$58,$F8,$9F,$9A,$7E,$0A,$3F
	.dc.b	$BB,$BF,$E5,$C3,$FC,$AF,$CD,$FE,$56,$7F,$A1,$AF,$31,$E5,$FF,$94
	.dc.b	$E7,$FC,$2D,$E2,$5C,$7F,$CB,$51,$FE,$2D,$3F,$C1,$5B,$FF,$84,$1F
	.dc.b	$F0,$FE,$DF,$E2,$6D,$3F,$C5,$71,$FE,$21,$BF,$F1,$FF,$8D,$FE,$20
	.dc.b	$3F,$C4,$78,$7F,$87,$2B,$FC,$2A,$FF,$88,$EB,$A9,$FE,$21,$9F,$F2
	.dc.b	$8C,$E9,$E2,$73,$AE,$A7,$F8,$E4,$B3,$E1,$3F,$C6,$97,$FE,$29,$7F
	.dc.b	$C4,$A7,$B8,$D3,$FE,$30,$3F,$C4,$98,$7B,$5E,$11,$FE,$4A,$75,$7E
	.dc.b	$E7,$E0,$47,$F8,$CC,$F4,$57,$F8,$CD,$E1,$3F,$E6,$FF,$1F,$FC,$91
	.dc.b	$7F,$E2,$E9,$FF,$41,$9E,$29,$F7,$57,$AE,$30,$7F,$7A,$A7,$8F,$FE
	.dc.b	$38,$8E,$38,$78,$9F,$63,$72,$F9,$D1,$C6,$78,$3A,$C3,$BB,$D8,$F3
	.dc.b	$9B,$C7,$CC,$EC,$C1,$F3,$23,$8D,$F7,$CB,$EA,$4F,$3C,$7C,$F7,$CB
	.dc.b	$1F,$1C,$F6,$24,$E9,$E8,$76,$BF,$A3,$19,$6E,$3E,$AA,$E7,$39,$C7
	.dc.b	$FD,$09,$5A,$C3,$FC,$42,$7F,$82,$30,$57,$78,$C3,$63,$FC,$1A,$D7
	.dc.b	$C5,$5F,$E6,$A8,$7C,$B7,$CA,$7C,$13,$A2,$1D,$A3,$58,$CB,$F9,$75
	.dc.b	$C2,$B5,$18,$9E,$95,$B9,$FF,$0E,$5F,$92,$31,$82,$31,$7C,$4E,$B3
	.dc.b	$A8,$CD,$7F,$B1,$30,$F4,$D6,$E3,$A9,$F8,$DA,$FF,$30,$3C,$23,$E2
	.dc.b	$C6,$23,$51,$F3,$1F,$F2,$6E,$8D,$DC,$FE,$AA,$FC,$F4,$E6,$31,$5E
	.dc.b	$08,$C4,$F3,$8F,$E2,$75,$CA,$31,$3A,$9F,$F4,$03,$63,$FC,$87,$E1
	.dc.b	$3E,$91,$E0,$FE,$97,$E4,$DF,$3A,$F7,$7F,$51,$7B,$AD,$4E,$70,$FD
	.dc.b	$A4,$E5,$F2,$37,$2F,$F4,$83,$63,$C3,$8F,$81,$F5,$4B,$E9,$BF,$CE
	.dc.b	$8F,$F9,$12,$3E,$3F,$D3,$3D,$9F,$F4,$5C,$FF,$9D,$AB,$7D,$EB,$DD
	.dc.b	$6A,$BA,$79,$3F,$EA,$17,$FC,$C7,$7D,$9F,$CF,$35,$CC,$6D,$FF,$41
	.dc.b	$FD,$3F,$86,$FF,$D4,$0D,$8F,$30,$DF,$E4,$79,$DE,$19,$D6,$5F,$EB
	.dc.b	$16,$DF,$F9,$D1,$B7,$C7,$83,$FD,$67,$5B,$CF,$D5,$E8,$7F,$B1,$E1
	.dc.b	$EA,$3D,$4C,$FA,$5F,$12,$73,$18,$CF,$FA,$ED,$FF,$B4,$72,$FE,$BF
	.dc.b	$D7,$FE,$B0,$7F,$D3,$47,$FA,$53,$CD,$FF,$44,$F9,$9F,$E9,$93,$F0
	.dc.b	$FC,$3F,$E9,$09,$D9,$FE,$96,$8F,$F5,$09,$AC,$3F,$D4,$31,$FE,$9E
	.dc.b	$F3,$7F,$D4,$27,$FA,$4D,$FF,$4B,$CF,$56,$9B,$79,$3F,$EB,$B2,$47
	.dc.b	$FA,$8C,$E7,$1C,$63,$FD,$38,$7F,$AA,$CF,$F5,$21,$8B,$E1,$1D,$26
	.dc.b	$CF,$FA,$F5,$AF,$26,$F3,$7E,$49,$B8,$7E,$36,$7C,$75,$F0,$1E,$4D
	.dc.b	$79,$36,$B9,$36,$79,$57,$4C,$3E,$3D,$FC,$B3,$E1,$1D,$0F,$F5,$A6
	.dc.b	$1F,$CC,$DF,$F5,$A7,$57,$FA,$DB,$2F,$F5,$CC,$7F,$AC,$4F,$F5,$A7
	.dc.b	$9D,$FE,$AA,$9F,$F5,$BC,$64,$FF,$42,$E5,$FE,$9C,$23,$FE,$9A,$21
	.dc.b	$FE,$98,$3F,$DC,$78,$FF,$AB,$4F,$F5,$23,$FE,$9E,$33,$1F,$E8,$E7
	.dc.b	$8F,$9F,$FE,$A5,$7B,$98,$EB,$FF,$69,$D7,$3A,$E1,$3D,$CF,$F5,$C9
	.dc.b	$D4,$FF,$51,$E1,$FE,$A6,$8F,$3C,$6E,$7F,$D4,$91,$CA,$FC,$1E,$37
	.dc.b	$FA,$A6,$31,$87,$81,$E3,$19,$BF,$F6,$54,$7F,$B1,$88,$E0,$F4,$1E
	.dc.b	$E3,$C9,$7E,$07,$27,$A0,$F2,$47,$B2,$7F,$D7,$F9,$FF,$AD,$A7,$D8
	.dc.b	$FF,$AE,$A7,$CB,$1D,$4E,$27,$D2,$9F,$33,$82,$38,$33,$87,$C0,$C7
	.dc.b	$E3,$4F,$A4,$F9,$97,$E1,$55,$E4,$BF,$0C,$7A,$6F,$FD,$DF,$87,$B6
	.dc.b	$1E,$57,$FA,$A5,$FF,$55,$F9,$9F,$E6,$87,$2D,$8D,$47,$FA,$CF,$8F
	.dc.b	$F5,$0F,$FB,$03,$D3,$FF,$61,$79,$A7,$FD,$8F,$E0,$B5,$FF,$1B,$D0
	.dc.b	$FE,$5B,$19,$F1,$BF,$DA,$8D,$8D,$61,$FB,$5E,$E7,$D4,$38,$77,$8F
	.dc.b	$4B,$E6,$DB,$FD,$E6,$1C,$E7,$F9,$DE,$1C,$AF,$FD,$B1,$F5,$3F,$D9
	.dc.b	$A7,$FD,$0E,$D7,$F6,$8D,$7F,$C5,$0D,$7D,$41,$E1,$7E,$0F,$AA,$7C
	.dc.b	$27,$B1,$B5,$EC,$E8,$E3,$7C,$1F,$89,$D4,$FA,$BF,$EE,$D5,$FD,$C1
	.dc.b	$AF,$F7,$06,$CF,$F8,$01,$B1,$C9,$DF,$57,$F8,$96,$3F,$AD,$78,$EF
	.dc.b	$87,$F4,$CE,$76,$FF,$02,$C6,$3C,$CE,$5E,$25,$C3,$FD,$4B,$F5,$8F
	.dc.b	$AC,$F8,$3A,$7C,$2E,$34,$E6,$BF,$C2,$ED,$7F,$C2,$ED,$9E,$0E,$73
	.dc.b	$F3,$4B,$D5,$7D,$6F,$86,$30,$F8,$F0,$EF,$8E,$6B,$75,$FE,$10,$6B
	.dc.b	$FE,$05,$6E,$3F,$E6,$16,$C7,$DA,$77,$87,$89,$7A,$BD,$4E,$D8,$CA
	.dc.b	$FF,$41,$CE,$9B,$39,$7F,$C9,$67,$A0,$C9,$0C,$C7,$83,$4F,$14,$F7
	.dc.b	$9D,$5F,$8E,$F1,$1F,$EF,$33,$F5,$57,$E1,$79,$C6,$A7,$D5,$56,$FD
	.dc.b	$D3,$CB,$0F,$85,$3F,$6A,$FC,$27,$C9,$70,$F6,$7E,$97,$A5,$FE,$F3
	.dc.b	$8F,$11,$F5,$E3,$3E,$B7,$F9,$86,$3F,$DE,$D9,$79,$72,$F1,$C7,$FA
	.dc.b	$86,$75,$1F,$80,$B4,$63,$2F,$1D,$76,$9F,$05,$72,$8C,$56,$DF,$E1
	.dc.b	$93,$CB,$C5,$F9,$37,$CA,$FB,$5E,$ED,$7B,$7D,$78,$79,$EF,$F7,$B3
	.dc.b	$E0,$79,$12,$7B,$D7,$39,$E5,$7E,$52,$57,$29,$D1,$E1,$BE,$D7,$F3
	.dc.b	$4F,$89,$84,$7C,$78,$FC,$C8,$C6,$1D,$27,$81,$F8,$EC,$7C,$CE,$23
	.dc.b	$4F,$BB,$0F,$0D,$7C,$48,$F4,$BD,$31,$F3,$39,$8E,$0E,$5D,$9F,$41
	.dc.b	$C3,$A3,$07,$8E,$7B,$B8,$AF,$A3,$3C,$31,$F8,$33,$C6,$72,$FE,$23
	.dc.b	$1F,$80,$F3,$AE,$11,$89,$D9,$30,$FA,$95,$CD,$F5,$BC,$CF,$3E,$3F
	.dc.b	$19,$FC,$C1,$8C,$7C,$F5,$F2,$27,$47,$DC,$AE,$AE,$27,$ED,$5F,$8E
	.dc.b	$7B,$D7,$8E,$FE,$94,$60,$8F,$CA,$8E,$D5,$E0,$9C,$CF,$90,$F8,$04
	.dc.b	$9E,$8E,$5B,$7E,$BC,$7E,$5B,$E5,$C7,$ED,$36,$BE,$95,$F6,$8F,$64
	.dc.b	$79,$8F,$24,$67,$0F,$6C,$7C,$5A,$FA,$2D,$3E,$61,$E5,$9E,$11,$E0
	.dc.b	$9F,$64,$F4,$3B,$1F,$54,$C1,$4F,$64,$F2,$7A,$E1,$EE,$9E,$71,$A3
	.dc.b	$8D,$FC,$1C,$7E,$94,$62,$7D,$2F,$94,$C6,$1E,$53,$D5,$7F,$49,$B8
	.dc.b	$7D,$EA,$EF,$87,$C6,$BF,$01,$F9,$F3,$FC,$02,$53,$F1,$E6,$1F,$E1
	.dc.b	$8E,$BC,$47,$AC,$DF,$FD,$01,$3F,$E3,$88,$C3,$FE,$83,$9D,$CE,$9B
	.dc.b	$1F,$E8,$7A,$7A,$AF,$5B,$9C,$61,$B1,$99,$E0,$D7,$CC,$E7,$4F,$E6
	.dc.b	$4D,$CF,$FC,$3B,$EB,$3C,$1F,$F5,$35,$6B,$0F,$F1,$D3,$70,$FE,$25
	.dc.b	$87,$17,$FC,$7D,$7C,$78,$77,$8F,$F2,$CA,$D5,$7F,$32,$6E,$5E,$98
	.dc.b	$FE,$D3,$7C,$5E,$B8,$F2,$7F,$9F,$D7,$03,$9C,$EF,$1F,$21,$30,$F0
	.dc.b	$5F,$1B,$56,$E7,$31,$D8,$95,$FE,$14,$9D,$47,$82,$B9,$61,$E7,$AE
	.dc.b	$D3,$F0,$32,$F8,$F8,$74,$9C,$CF,$BE,$FC,$58,$74,$8C,$1E,$0A,$E0
	.dc.b	$7D,$2B,$E7,$1E,$13,$45,$BF,$34,$6A,$39,$C7,$83,$0F,$35,$73,$CB
	.dc.b	$D3,$97,$C9,$CA,$9E,$4C,$79,$5F,$0C,$7B,$61,$E4,$7D,$D8,$70,$7C
	.dc.b	$67,$82,$3B,$38,$9E,$4E,$4E,$26,$70,$F1,$3F,$80,$BF,$0C,$63,$1F
	.dc.b	$93,$97,$C7,$BE,$64,$9F,$C0,$D7,$0C,$BC,$97,$C2,$31,$87,$79,$F7
	.dc.b	$65,$AC,$BD,$93,$E5,$3F,$0B,$5C,$2F,$E5,$61,$DA,$F8,$57,$C4,$21
	.dc.b	$C1,$CB,$CF,$1F,$83,$97,$A6,$BD,$98,$F8,$23,$19,$70,$AD,$E5,$E1
	.dc.b	$AF,$2C,$63,$1D,$61,$CB,$1F,$97,$97,$9B,$1F,$0E,$5D,$27,$E0,$1F
	.dc.b	$36,$F8,$CE,$B2,$ED,$3E,$AC,$7B,$3C,$30,$F9,$57,$C2,$3E,$FE,$1D
	.dc.b	$67,$F2,$33,$CC,$93,$AB,$EA,$FC,$53,$11,$A9,$EB,$18,$23,$82,$65
	.dc.b	$F1,$EB,$9D,$F8,$63,$F1,$F3,$AA,$FA,$73,$97,$E2,$CF,$1B,$F0,$36
	.dc.b	$3B,$B6,$B9,$47,$E0,$EB,$93,$E9,$8F,$3D,$7F,$42,$9F,$E2,$CD,$B3
	.dc.b	$FE,$DF,$30,$5A,$F3,$DC,$7F,$93,$FA,$DF,$B9,$BF,$E0,$84,$BC,$46
	.dc.b	$59,$7F,$9A,$AF,$E4,$F3,$87,$EF,$13,$0F,$E3,$51,$96,$CF,$16,$CE
	.dc.b	$F0,$FE,$83,$5B,$6D,$6D,$B3,$EE,$2C,$7C,$39,$C3,$4F,$81,$1E,$E7
	.dc.b	$F9,$0C,$7F,$1C,$D5,$C0,$F5,$BD,$F5,$FC,$FE,$73,$18,$B6,$3F,$ED
	.dc.b	$D9,$CD,$74,$38,$5F,$36,$DF,$F2,$AE,$97,$D7,$7F,$74,$9C,$9F,$D3
	.dc.b	$7B,$5E,$BB,$CD,$E1,$DC,$ED,$FE,$67,$5A,$9C,$C6,$2B,$85,$F3,$6E
	.dc.b	$DE,$8E,$03,$FE,$F4,$FA,$FE,$18,$7F,$39,$F5,$DE,$87,$CD,$AF,$F0
	.dc.b	$7C,$63,$1F,$F0,$4B,$63,$FC,$26,$1C,$3C,$27,$EB,$78,$DC,$E7,$83
	.dc.b	$D7,$D6,$7A,$DA,$C3,$CB,$87,$3C,$30,$FA,$2B,$73,$EA,$BE,$57,$AE
	.dc.b	$1F,$8D,$EB,$F0,$C1,$87,$05,$7C,$30,$E4,$78,$27,$51,$F8,$DB,$E7
	.dc.b	$3A,$AF,$16,$1D,$7D,$47,$AD,$F1,$E7,$D3,$18,$AF,$F1,$7B,$63,$E3
	.dc.b	$B5,$FF,$1D,$B9,$C3,$E8,$36,$7E,$BF,$13,$DF,$7A,$1E,$1C,$7D,$D8
	.dc.b	$FF,$89,$07,$9E,$F5,$BE,$5E,$1F,$5E,$31,$3F,$E4,$AE,$CF,$C1,$C6
	.dc.b	$21,$3B,$AF,$BB,$5F,$88,$F5,$75,$3F,$81,$78,$57,$85,$CF,$17,$ED
	.dc.b	$5C,$C7,$EA,$03,$AD,$3A,$61,$DA,$72,$FA,$2B,$93,$9A,$E1,$3A,$3D
	.dc.b	$45,$3D,$33,$C1,$CE,$3C,$0F,$1E,$3F,$02,$FB,$61,$6B,$9E,$1D,$8D
	.dc.b	$BD,$71,$F8,$15,$E3,$38,$C7,$84,$F8,$4E,$27,$8E,$1E,$18,$F7,$36
	.dc.b	$7B,$C6,$CE,$8F,$E0,$31,$F5,$3F,$A4,$C7,$E3,$B9,$C6,$B9,$3C,$D3
	.dc.b	$C1,$FC,$25,$F7,$74,$D7,$CE,$73,$3D,$47,$D7,$C3,$E1,$61,$E4,$8F
	.dc.b	$9B,$7E,$07,$89,$83,$8E,$1E,$4C,$7E,$3C,$7B,$8E,$D7,$F1,$2F,$E4
	.dc.b	$E1,$67,$C4,$FD,$09,$F0,$C7,$13,$B3,$B7,$C1,$16,$BF,$31,$1F,$80
	.dc.b	$AE,$A7,$D0,$2B,$F8,$8E,$CF,$D3,$B6,$BF,$25,$18,$8F,$98,$DF,$11
	.dc.b	$E8,$FF,$0A,$9F,$E9,$35,$FC,$62,$73,$C3,$FC,$BA,$F8,$5E,$DB,$19
	.dc.b	$AE,$13,$CD,$FD,$97,$AD,$FD,$07,$0F,$CA,$07,$0F,$EF,$B1,$E3,$7A
	.dc.b	$C7,$C3,$39,$3F,$9F,$D7,$0B,$EB,$D1,$FC,$B9,$FE,$01,$5C,$1B,$8F
	.dc.b	$B7,$4D,$E1,$F8,$EC,$3C,$31,$AC,$7F,$23,$87,$1C,$BF,$9C,$D7,$76
	.dc.b	$E5,$DD,$F3,$E1,$E5,$15,$E8,$C3,$BC,$7E,$EE,$FB,$D7,$09,$F0,$4E
	.dc.b	$6F,$FA,$3C,$70,$32,$49,$F0,$CF,$A6,$77,$1C,$CF,$1E,$5E,$02,$38
	.dc.b	$35,$3F,$2A,$78,$B4,$F8,$39,$7C,$79,$DC,$7D,$79,$DB,$4D,$E5,$1F
	.dc.b	$85,$8F,$C7,$7A,$1F,$17,$2F,$98,$FF,$78,$7E,$35,$FB,$23,$E1,$1C
	.dc.b	$F2,$F8,$0F,$27,$E2,$47,$92,$79,$CF,$8C,$F0,$E1,$F5,$5E,$D8,$FE
	.dc.b	$4E,$BD,$78,$7D,$29,$E9,$1C,$E7,$E0,$46,$08,$7D,$C7,$2D,$C7,$E9
	.dc.b	$30,$FE,$71,$87,$B1,$F0,$C7,$E8,$2F,$E2,$47,$9F,$2F,$C0,$9F,$D5
	.dc.b	$FD,$57,$AD,$FE,$08,$BF,$F0,$C4,$62,$BF,$BC,$07,$0F,$F8,$0B,$83
	.dc.b	$FC,$1F,$EB,$7F,$90,$A3,$73,$A7,$FC,$26,$71,$BF,$F0,$FB,$FE,$2A
	.dc.b	$8D,$3F,$E1,$86,$AF,$0F,$FB,$9F,$8B,$E5,$C3,$FC,$63,$18,$9F,$F0
	.dc.b	$63,$5F,$20,$EA,$7A,$DF,$C4,$23,$7A,$30,$E9,$1D,$E3,$0E,$A3,$35
	.dc.b	$CB,$7D,$CB,$D4,$FE,$17,$B7,$F8,$FD,$B9,$FF,$8C,$9B,$3D,$70,$E3
	.dc.b	$1F,$C4,$A3,$E2,$B5,$E1,$C5,$F1,$9E,$CF,$C8,$E5,$7A,$9F,$E5,$B9
	.dc.b	$FF,$30,$35,$FF,$2B,$1F,$E1,$27,$1C,$EF,$0B,$FC,$B3,$AB,$C5,$7D
	.dc.b	$87,$F6,$59,$FF,$13,$DF,$07,$72,$F5,$75,$8F,$CD,$8C,$3E,$2F,$1F
	.dc.b	$3C,$8F,$55,$EA,$E1,$B5,$FE,$85,$6B,$E8,$D1,$F6,$4F,$B2,$E1,$7E
	.dc.b	$0A,$7A,$CF,$5F,$0D,$8E,$93,$89,$91,$FE,$78,$E1,$F8,$4F,$F9,$EA
	.dc.b	$B8,$3F,$23,$3F,$45,$F0,$8F,$F2,$A4,$EB,$9B,$F3,$6D,$8F,$F3,$7B
	.dc.b	$5E,$98,$79,$3D,$79,$1E,$1E,$1D,$CF,$5C,$BC,$73,$EC,$AD,$BF,$68
	.dc.b	$F7,$CE,$27,$76,$AC,$5F,$4B,$EC,$7B,$63,$2D,$32,$74,$2C,$F3,$6C
	.dc.b	$70,$74,$78,$E3,$4D,$8D,$47,$CA,$AF,$2C,$7F,$A4,$B0,$F8,$ED,$AF
	.dc.b	$8E,$D7,$C8,$7C,$E7,$EF,$7A,$6F,$5F,$F8,$65,$FC,$68,$C7,$8B,$E6
	.dc.b	$B8,$7C,$78,$CB,$6B,$8E,$5E,$08,$DD,$F3,$9C,$C7,$FA,$F2,$B8,$CF
	.dc.b	$AA,$37,$5F,$EA,$C9,$CB,$6B,$8C,$FD,$93,$67,$37,$33,$C2,$3D,$F5
	.dc.b	$D5,$F3,$5F,$D4,$9D,$47,$33,$E7,$19,$AE,$67,$83,$2F,$A7,$7C,$CF
	.dc.b	$65,$EC,$F1,$D7,$38,$CB,$E0,$3A,$91,$C9,$30,$F9,$11,$97,$E0,$4F
	.dc.b	$58,$C1,$EC,$8C,$BD,$67,$C9,$1C,$9E,$91,$F1,$67,$CB,$87,$C0,$26
	.dc.b	$1E,$DC,$7F,$06,$FC,$09,$F0,$5F,$4B,$E4,$77,$C3,$C3,$19,$86,$1D
	.dc.b	$2B,$8D,$78,$CD,$92,$79,$D7,$1A,$E1,$1D,$4F,$71,$A9,$E0,$F6,$BF
	.dc.b	$43,$D8,$D1,$0F,$A5,$1C,$27,$AD,$77,$AF,$75,$77,$36,$71,$9D,$D7
	.dc.b	$A6,$31,$3C,$2B,$95,$F8,$67,$C8,$FC,$E9,$D5,$7A,$31,$FC,$4B,$E3
	.dc.b	$8C,$D7,$B8,$E2,$78,$F0,$FC,$AD,$70,$8F,$0D,$F7,$8C,$47,$0C,$3E
	.dc.b	$FB,$70,$F4,$34,$E2,$7E,$28,$C1,$2F,$BE,$3F,$82,$C3,$D9,$18,$9B
	.dc.b	$8F,$A1,$F8,$E7,$17,$18,$F9,$27,$E4,$CF,$D3,$8C,$4F,$19,$E9,$3E
	.dc.b	$F8,$E3,$7E,$18,$C5,$FC,$B3,$C2,$57,$27,$AE,$BB,$BE,$CB,$EF,$3C
	.dc.b	$23,$57,$E1,$3F,$12,$F2,$C7,$E4,$D7,$08,$C6,$3F,$2B,$0F,$E3,$F1
	.dc.b	$FE,$2E,$6B,$FB,$23,$17,$FE,$35,$7F,$C6,$7E,$1B,$D7,$FF,$17,$47
	.dc.b	$F8,$D8,$FF,$45,$B8,$76,$D7,$FD,$6C,$D7,$F9,$47,$AE,$F5,$FF,$C7
	.dc.b	$B8,$23,$13,$D6,$31,$3A,$9C,$CF,$A5,$DE,$7F,$CF,$DB,$92,$67,$AF
	.dc.b	$FE,$39,$DF,$F9,$45,$6A,$7C,$78,$60,$FE,$BF,$5B,$8F,$E8,$D3,$B7
	.dc.b	$F8,$2B,$71,$F2,$C7,$F7,$19,$E1,$8F,$8E,$FA,$4F,$27,$AF,$7B,$1C
	.dc.b	$D6,$EB,$47,$39,$D3,$88,$D1,$32,$F7,$43,$0F,$0C,$7F,$51,$CB,$CC
	.dc.b	$4C,$7C,$38,$F8,$27,$51,$6F,$E3,$E3,$E0,$C7,$B4,$7C,$63,$C5,$5E
	.dc.b	$0C,$7B,$C6,$67,$6F,$C6,$BE,$EF,$03,$55,$E0,$7D,$D8,$78,$5E,$06
	.dc.b	$CF,$66,$1E,$39,$E1,$7D,$B2,$FA,$2F,$F8,$93,$1F,$1D,$F2,$8D,$65
	.dc.b	$F6,$67,$35,$C2,$F9,$92,$FE,$0E,$3E,$39,$D4,$7A,$0C,$9E,$57,$2F
	.dc.b	$79,$F8,$98,$F9,$E3,$81,$92,$DF,$8B,$0F,$A9,$1E,$F8,$C6,$5E,$19
	.dc.b	$F4,$57,$3A,$E4,$4C,$3C,$37,$C2,$3C,$95,$C3,$1E,$58,$7C,$DA,$D6
	.dc.b	$5F,$12,$3E,$AB,$6F,$C5,$8F,$7C,$7D,$A4,$C7,$A4,$F6,$BE,$95,$B9
	.dc.b	$F5,$DF,$E0,$B0,$EF,$97,$CD,$C3,$D3,$18,$7E,$5C,$EE,$F9,$61,$E4
	.dc.b	$BE,$17,$BB,$FC,$1E,$5F,$BD,$3F,$D8,$F1,$FB,$76,$DF,$EA,$1A,$6E
	.dc.b	$7F,$31,$C3,$F2,$EB,$FD,$09,$3F,$C0,$38,$71,$39,$6C,$64,$FE,$0B
	.dc.b	$19,$69,$9A,$FE,$05,$1F,$81,$E6,$7D,$63,$DF,$C7,$5F,$FB,$AB,$0F
	.dc.b	$F3,$8D,$6A,$73,$E3,$F1,$BE,$15,$AB,$E1,$8F,$FA,$86,$72,$FF,$A8
	.dc.b	$E3,$3B,$FF,$A6,$31,$F0,$FD,$9F,$33,$A7,$D8,$5C,$7C,$38,$A7,$18
	.dc.b	$F5,$AB,$87,$3A,$7D,$67,$AB,$E2,$FC,$CD,$C7,$9B,$70,$F3,$36,$F9
	.dc.b	$37,$8B,$FD,$9B,$87,$4B,$F7,$61,$F0,$7D,$F9,$1F,$E9,$6F,$15,$F6
	.dc.b	$8F,$A2,$CE,$9F,$0E,$EE,$9F,$EB,$8B,$DD,$63,$3E,$FF,$27,$F4,$BE
	.dc.b	$47,$E7,$37,$3F,$86,$DC,$BE,$73,$71,$FE,$1D,$F6,$7F,$89,$FB,$5F
	.dc.b	$C5,$36,$FF,$A4,$BF,$91,$BF,$E4,$CD,$7F,$A9,$B5,$FE,$70,$DE,$93
	.dc.b	$D2,$7F,$A9,$FD,$6D,$AF,$BD,$F2,$BF,$D1,$3D,$0F,$E8,$BC,$FF,$EF
	.dc.b	$58,$C5,$FC,$CF,$5E,$DF,$FB,$C6,$6D,$6E,$BF,$B3,$5F,$2E,$95,$6B
	.dc.b	$99,$8F,$09,$CD,$F0,$3F,$DF,$FE,$4E,$4E,$1F,$01,$63,$59,$B3,$FD
	.dc.b	$1C,$E6,$19,$FA,$B2,$4E,$63,$FB,$0C,$BE,$0D,$C9,$E3,$F8,$BF,$6C
	.dc.b	$E6,$31,$97,$82,$30,$FF,$85,$DA,$7F,$86,$6B,$9C,$65,$B1,$99,$EB
	.dc.b	$18,$BF,$1F,$0D,$BF,$EE,$33,$6F,$DD,$AF,$FB,$4E,$7E,$0E,$88,$C4
	.dc.b	$FC,$09,$FD,$0F,$B9,$F3,$7C,$9F,$A0,$D8,$E3,$B2,$31,$5C,$4E,$FE
	.dc.b	$37,$D5,$3E,$C3,$6B,$8B,$4E,$18,$FF,$14,$71,$A7,$F8,$3B,$A9,$E4
	.dc.b	$FF,$88,$67,$FC,$1B,$18,$F1,$BE,$9C,$7F,$8B,$63,$18,$7F,$90,$DB
	.dc.b	$1F,$E1,$40,$E2,$FB,$5F,$F8,$B5,$FC,$D7,$C5,$3E,$2B,$F2,$F5,$BC
	.dc.b	$3F,$C3,$BF,$04,$F8,$0F,$CD,$E0,$C3,$85,$7D,$FA,$D3,$F9,$09,$D6
	.dc.b	$1F,$86,$9C,$E5,$FE,$59,$6B,$E7,$6B,$E0,$C3,$DE,$FF,$92,$F4,$FE
	.dc.b	$21,$87,$E2,$DB,$1F,$E6,$36,$BF,$05,$AF,$F8,$D9,$DE,$1F,$9C,$6C
	.dc.b	$FF,$8E,$BC,$B7,$C8,$7E,$72,$73,$7F,$2F,$BF,$FC,$B9,$EE,$7E,$D2
	.dc.b	$3E,$97,$17,$F9,$1A,$31,$1F,$E4,$49,$DF,$67,$DA,$AD,$3F,$E7,$98
	.dc.b	$FF,$33,$7C,$0F,$0E,$1F,$C7,$69,$8A,$D6,$0A,$D6,$1F,$E5,$47,$BC
	.dc.b	$7F,$9B,$F0,$E0,$F9,$2E,$BE,$33,$95,$F8,$E3,$31,$C4,$C9,$C4,$A7
	.dc.b	$1A,$F0,$C7,$08,$57,$B9,$E1,$8F,$91,$C8,$78,$B6,$FD,$B7,$83,$85
	.dc.b	$7A,$31,$F8,$75,$E2,$3C,$D3,$BA,$E2,$7A,$E3,$47,$82,$FB,$9D,$1F
	.dc.b	$01,$E6,$8E,$2F,$BE,$74,$75,$7F,$00,$7C,$2A,$F1,$C6,$A7,$C0,$F5
	.dc.b	$AE,$F1,$F6,$1B,$5C,$A7,$A3,$F2,$E7,$89,$E2,$7F,$6F,$8F,$DD,$8C
	.dc.b	$5F,$81,$B1,$99,$E9,$1C,$9F,$3D,$F8,$1F,$C0,$35,$F9,$31,$B7,$C4
	.dc.b	$7C,$17,$D0,$7C,$B8,$F6,$34,$CC,$7C,$93,$BB,$97,$E3,$FA,$5F,$7D
	.dc.b	$B3,$F9,$A8,$C4,$7E,$4D,$BC,$6F,$77,$FA,$9C,$A7,$31,$86,$F0,$7F
	.dc.b	$B3,$5B,$3B,$BF,$F5,$9D,$F3,$AF,$CA,$4B,$E0,$5E,$EF,$FA,$E2,$FF
	.dc.b	$6F,$18,$9D,$DF,$17,$F0,$E7,$F3,$4C,$7C,$B5,$C2,$F9,$65,$FE,$C2
	.dc.b	$C3,$95,$F0,$C3,$95,$FF,$04,$7A,$6B,$C6,$7F,$6B,$5C,$23,$58,$FC
	.dc.b	$26,$E5,$EA,$9E,$B8,$71,$BE,$D7,$E4,$C7,$E0,$BC,$5F,$26,$5F,$ED
	.dc.b	$6C,$78,$1D,$2B,$F7,$B9,$79,$2B,$71,$E4,$AE,$11,$D3,$2F,$E1,$75
	.dc.b	$BA,$FF,$7F,$65,$E9,$AE,$31,$D3,$1F,$12,$3E,$55,$FA,$F0,$F9,$31
	.dc.b	$C9,$AE,$8F,$6C,$78,$D8,$75,$7C,$73,$DA,$3E,$F4,$FA,$1E,$86,$A6
	.dc.b	$3F,$72,$FE,$5D,$FA,$32,$F8,$87,$C5,$CB,$E3,$C7,$B1,$C3,$F0,$EB
	.dc.b	$F9,$2D,$7C,$17,$B9,$E5,$25,$7D,$72,$4F,$17,$9C,$70,$AE,$EF,$C7
	.dc.b	$CB,$E7,$3F,$0E,$32,$FB,$8F,$5B,$C9,$C1,$E4,$C7,$E1,$3C,$6F,$C0
	.dc.b	$7B,$C8,$F7,$CB,$F1,$4F,$D7,$FB,$67,$85,$F6,$63,$FC,$40,$23,$2F
	.dc.b	$F8,$FD,$AE,$97,$BD,$C7,$5F,$F2,$5C,$6A,$7F,$C2,$CD,$7B,$35,$F1
	.dc.b	$57,$F9,$25,$7C,$0B,$D2,$C3,$63,$FC,$9C,$D7,$FC,$44,$70,$8D,$9E
	.dc.b	$3E,$4F,$67,$C1,$3E,$0F,$15,$F9,$78,$C3,$FE,$5B,$BE,$99,$7F,$A0
	.dc.b	$7E,$DD,$F4,$9C,$21,$18,$AF,$F1,$13,$5F,$93,$7D,$1F,$F2,$54,$F0
	.dc.b	$AF,$F3,$33,$63,$D2,$D7,$FC,$78,$D7,$FC,$39,$F7,$1D,$63,$FD,$15
	.dc.b	$EF,$7A,$BA,$FF,$CC,$0D,$9D,$69,$FE,$2B,$F1,$93,$B6,$E1,$FE,$68
	.dc.b	$6B,$FE,$58,$AE,$2F,$FA,$2D,$78,$7C,$7D,$29,$3F,$E7,$09,$CB,$CE
	.dc.b	$32,$FF,$98,$AB,$FD,$94,$EE,$BE,$8A,$FB,$97,$A5,$86,$C7,$FA,$35
	.dc.b	$AF,$E5,$3E,$59,$F5,$1B,$AF,$C4,$56,$AB,$FC,$C8,$D7,$F3,$8D,$7F
	.dc.b	$8B,$3D,$36,$FF,$5A,$6B,$F1,$4E,$93,$ED,$F7,$7E,$77,$96,$F8,$CA
	.dc.b	$7C,$F8,$B3,$FE,$8A,$8A,$FC,$B8,$E3,$F3,$12,$3F,$D1,$EF,$D7,$F8
	.dc.b	$0C,$47,$FA,$22,$D3,$8C,$7F,$DA,$3C,$2B,$F9,$3F,$06,$FA,$7F,$F2
	.dc.b	$97,$A9,$FB,$98,$C6,$5F,$E4,$78,$C4,$EA,$3F,$9B,$70,$BE,$5F,$10
	.dc.b	$BE,$2F,$FA,$49,$BB,$FF,$A3,$DB,$1D,$0F,$E8,$11,$9B,$E7,$3F,$A3
	.dc.b	$F2,$FF,$D9,$75,$FD,$16,$31,$8F,$18,$FF,$26,$BF,$E8,$1C,$FF,$DE
	.dc.b	$B1,$E9,$3F,$DB,$9B,$7F,$A5,$5C,$74,$23,$11,$C8,$FC,$C7,$B5,$EF
	.dc.b	$95,$74,$9D,$47,$1F,$3F,$57,$CB,$2F,$3C,$F0,$C3,$96,$1F,$E1,$9C
	.dc.b	$3C,$2D,$79,$B6,$BD,$FE,$6B,$EE,$BE,$0D,$6B,$C6,$E6,$47,$DD,$3B
	.dc.b	$D3,$BF,$91,$E4,$FA,$D9,$5E,$08,$CB,$F0,$B5,$4E,$A3,$C7,$3A,$7D
	.dc.b	$F3,$A7,$84,$E7,$2F,$7E,$6F,$73,$DE,$4A,$F7,$FC,$AD,$7D,$F1,$6B
	.dc.b	$8C,$FC,$6F,$1B,$E5,$FA,$52,$FE,$3C,$65,$F4,$46,$5F,$93,$1F,$88
	.dc.b	$F0,$3F,$13,$F2,$D2,$7F,$04,$7E,$56,$34,$E3,$95,$3B,$8F,$C1,$1B
	.dc.b	$34,$E7,$1F,$A7,$3C,$27,$95,$7E,$63,$AB,$FB,$17,$47,$E7,$DB,$3F
	.dc.b	$AD,$6B,$F9,$B6,$BF,$A9,$7F,$57,$F1,$0F,$6B,$FC,$BD,$3E,$48,$C7
	.dc.b	$A9,$E9,$AF,$04,$FB,$A7,$2E,$EB,$8E,$5E,$8A,$EB,$18,$E4,$FE,$F1
	.dc.b	$7C,$2B,$F1,$ED,$8F,$45,$F5,$AE,$0F,$F8,$DE,$74,$D8,$CB,$EB,$38
	.dc.b	$61,$1F,$23,$E9,$71,$7E,$09,$E3,$5D,$27,$91,$C0,$93,$BA,$E5,$18
	.dc.b	$BF,$5D,$F2,$3C,$79,$FF,$21,$9E,$F3,$AA,$B1,$F4,$CA,$30,$F1,$1E
	.dc.b	$3A,$E5,$1D,$CF,$81,$3C,$23,$38,$79,$67,$47,$3C,$3C,$78,$FD,$2B
	.dc.b	$EF,$87,$58,$C5,$70,$8C,$3A,$AF,$65,$7C,$0B,$EC,$4C,$7D,$95,$F9
	.dc.b	$AC,$7C,$71,$89,$D4,$FA,$0D,$63,$E0,$8C,$C7,$9E,$FB,$57,$12,$47
	.dc.b	$19,$FC,$63,$E5,$9F,$85,$5C,$EB,$73,$CC,$EE,$6C,$F8,$B7,$E6,$7A
	.dc.b	$47,$0C,$7E,$0C,$7C,$3C,$7E,$0B,$C8,$C6,$1E,$07,$A6,$3E,$AC,$3A
	.dc.b	$47,$86,$3D,$B7,$D1,$B7,$D1,$F5,$CF,$43,$CB,$8F,$AD,$F2,$90,$FC
	.dc.b	$AC,$72,$C3,$C5,$1F,$97,$3F,$CC,$87,$C1,$2D,$FE,$D9,$B7,$FB,$E7
	.dc.b	$7E,$93,$E2,$5F,$15,$FE,$56,$94,$E6,$30,$D8,$FF,$2D,$B5,$EF,$EE
	.dc.b	$3D,$9E,$58,$A3,$13,$FE,$62,$8C,$4E,$A7,$33,$CA,$3F,$87,$47,$0C
	.dc.b	$FF,$AA,$FB,$DF,$E3,$EE,$0C,$D4,$BF,$E3,$17,$BE,$0F,$F4,$35,$EE
	.dc.b	$73,$5B,$C3,$F8,$8D,$EE,$7F,$D8,$95,$B9,$CD,$6E,$FF,$8C,$DE,$ED
	.dc.b	$7B,$9C,$D6,$E7,$F8,$C5,$6E,$B5,$18,$9D,$3F,$C0,$67,$5C,$2C,$73
	.dc.b	$97,$C6,$AF,$49,$FD,$23,$0E,$6F,$1C,$37,$87,$2B,$B1,$DA,$B7,$7C
	.dc.b	$F2,$F6,$E1,$D1,$F9,$B8,$F8,$63,$19,$7B,$1A,$7C,$68,$D4,$F2,$24
	.dc.b	$ED,$E7,$8F,$8A,$B8,$9F,$22,$B8,$D6,$C9,$1F,$14,$F7,$46,$9C,$57
	.dc.b	$66,$CF,$DE,$9D,$47,$C3,$3B,$DF,$BE,$75,$7E,$3B,$E5,$5E,$38,$C4
	.dc.b	$F8,$A3,$07,$2C,$BF,$01,$7C,$E3,$E1,$57,$07,$C6,$60,$E1,$97,$D1
	.dc.b	$C7,$CB,$5C,$67,$CD,$8F,$8B,$2F,$AB,$8F,$62,$61,$FE,$0C,$8C,$BD
	.dc.b	$EB,$CB,$5E,$AB,$E8,$75,$AD,$CF,$88,$F9,$78,$F1,$9D,$63,$E4,$C7
	.dc.b	$BD,$7D,$6B,$E3,$87,$59,$FC,$64,$E6,$B8,$63,$E5,$6C,$76,$35,$8F
	.dc.b	$C1,$34,$7A,$F8,$BE,$BE,$3F,$9A,$1E,$36,$0C,$F1,$5A,$F8,$4D,$BF
	.dc.b	$D1,$37,$3F,$E0,$AD,$D7,$F8,$0B,$67,$FC,$EB,$F1,$5F,$8B,$F0,$5E
	.dc.b	$67,$FB,$E7,$1F,$F2,$B4,$E6,$31,$5F,$E5,$48,$C7,$A7,$C7,$B9,$18
	.dc.b	$BF,$F2,$F7,$C5,$64,$D7,$C5,$58,$C9,$30,$FF,$2F,$C6,$5F,$F3,$0F
	.dc.b	$C5,$78,$7E,$2B,$B3,$6B,$FC,$D4,$D9,$DC,$7F,$A3,$B3,$57,$08,$C3
	.dc.b	$67,$FD,$17,$3A,$9F,$F3,$A7,$C4,$FA,$CF,$D7,$78,$AF,$8A,$7A,$99
	.dc.b	$AD,$93,$0F,$F4,$D5,$6E,$75,$1F,$E7,$AF,$4F,$E5,$7C,$57,$C7,$6D
	.dc.b	$FF,$A0,$DB,$3B,$8F,$54,$FE,$13,$CC,$FC,$2F,$C5,$59,$FD,$13,$6B
	.dc.b	$F9,$14,$FE,$2B,$C8,$C7,$47,$F4,$68,$BD,$1F,$D0,$A6,$D3,$9F,$F0
	.dc.b	$3F,$15,$FC,$7F,$D3,$FE,$41,$18,$CD,$5F,$ED,$FF,$80,$D4,$62,$75
	.dc.b	$2A,$E3,$C6,$C3,$73,$DE,$31,$7D,$2B,$BF,$4B,$1D,$E3,$D2,$73,$F1
	.dc.b	$58,$C1,$EF,$EB,$04,$E6,$30,$DC,$7B,$1E,$79,$DB,$8C,$BC,$95,$E6
	.dc.b	$5F,$12,$DB,$C1,$F1,$59,$8C,$5F,$1D,$3F,$53,$39,$A9,$3D,$6F,$FB
	.dc.b	$2C,$61,$F8,$D1,$F3,$1E,$F1,$9D,$3F,$A8,$47,$09,$E5,$87,$82,$38
	.dc.b	$9F,$EE,$EF,$87,$F8,$9E,$5F,$C3,$6D,$79,$74,$46,$2B,$8E,$7F,$EF
	.dc.b	$38,$C6,$3F,$4A,$7E,$77,$8F,$F3,$5B,$87,$36,$9C,$B0,$FC,$2C,$7C
	.dc.b	$3D,$FF,$C3,$CF,$CD,$E9,$FF,$14,$F8,$DF,$82,$6C,$7D,$1C,$5F,$14
	.dc.b	$F8,$AD,$79,$7F,$E1,$79,$FE,$E5,$7C,$33,$F8,$2D,$8F,$CC,$5F,$1C
	.dc.b	$3C,$17,$C5,$F9,$DF,$14,$F6,$7E,$6E,$38,$6B,$F9,$E8,$D4,$FF,$90
	.dc.b	$5B,$1F,$E4,$C0,$E3,$0F,$AE,$D9,$FD,$3E,$2F,$92,$7A,$7F,$8D,$AD
	.dc.b	$61,$FA,$9F,$07,$F2,$1C,$7F,$C3,$B1,$FC,$47,$9C,$F5,$33,$52,$7F
	.dc.b	$19,$5B,$BF,$C1,$9F,$E4,$FA,$D3,$83,$F7,$3E,$07,$EE,$BE,$27,$F5
	.dc.b	$D1,$E6,$E7,$FF,$0F,$46,$23,$F2,$93,$BE,$EF,$CB,$6F,$F8,$AC,$B1
	.dc.b	$5F,$E6,$EA,$AA,$74,$AF,$A3,$E2,$31,$7D,$30,$F0,$BF,$36,$F9,$DF
	.dc.b	$2B,$E3,$69,$CD,$F0,$8F,$11,$D0,$F6,$61,$E4,$7C,$15,$ED,$BE,$98
	.dc.b	$77,$9E,$07,$A6,$FA,$3F,$08,$CC,$F0,$7B,$D7,$39,$DC,$76,$AF,$65
	.dc.b	$F8,$63,$D0,$72,$9D,$B8,$AE,$53,$BC,$3D,$AD,$9E,$0D,$9D,$C7,$A0
	.dc.b	$F0,$1F,$00,$C1,$D6,$FC,$6E,$2B,$AB,$4E,$A5,$9E,$A7,$D0,$8D,$B6
	.dc.b	$36,$D3,$73,$F2,$DB,$1F,$1A,$3C,$31,$B6,$9B,$2C,$6C,$F5,$46,$DB
	.dc.b	$1B,$69,$B3,$EC,$57,$C5,$8E,$71,$F6,$67,$C6,$D8,$CB,$F9,$3F,$51
	.dc.b	$F1,$5F,$83,$6C,$7E,$31,$AF,$E0,$DB,$C8,$F6,$FF,$B0,$4A,$73,$18
	.dc.b	$6F,$27,$F6,$A6,$CE,$EF,$FA,$15,$F3,$AF,$D5,$FC,$5B,$DC,$FE,$8D
	.dc.b	$9F,$ED,$63,$18,$7E,$C2,$F8,$3F,$9A,$6C,$EE,$B8,$E7,$FD,$3E,$F8
	.dc.b	$DE,$F3,$F4,$DF,$F0,$1C,$38,$DE,$F0,$E3,$7F,$E0,$2C,$38,$DE,$F0
	.dc.b	$E3,$7E,$A6,$D6,$EF,$8A,$BE,$35,$BC,$7D,$57,$C2,$B5,$7C,$2F,$71
	.dc.b	$FE,$12,$C7,$CA,$DA,$E1,$3A,$8E,$27,$4B,$8F,$EF,$A3,$46,$EB,$9C
	.dc.b	$7F,$BF,$AB,$9D,$F8,$4F,$4C,$E8,$DD,$72,$8C,$CF,$A6,$73,$1C,$6B
	.dc.b	$C3,$8F,$C7,$C3,$EE,$CE,$67,$E9,$46,$B1,$FA,$11,$A3,$07,$BE,$30
	.dc.b	$FA,$E7,$C3,$1A,$8F,$82,$64,$C5,$F9,$1E,$37,$E3,$7C,$26,$C8,$E4
	.dc.b	$97,$E4,$8C,$BF,$65,$E3,$1F,$8F,$7A,$C6,$63,$FA,$67,$C5,$7F,$8E
	.dc.b	$27,$FC,$48,$D7,$FA,$CB,$6F,$FC,$42,$BE,$2F,$FC,$4D,$18,$F1,$2E
	.dc.b	$FF,$E2,$9A,$E0,$D8,$CD,$FF,$96,$27,$93,$FE,$33,$F8,$B7,$A9,$A8
	.dc.b	$46,$1D,$46,$7E,$2B,$F7,$91,$83,$FC,$04,$70,$8A,$D3,$33,$CD,$FF
	.dc.b	$1F,$1F,$E6,$5F,$8B,$7C,$5F,$9B,$34,$62,$7F,$C4,$8D,$7F,$BE,$E1
	.dc.b	$FE,$0A,$73,$5C,$67,$55,$FE,$79,$6E,$7F,$E4,$DB,$E1,$7F,$E4,$0F
	.dc.b	$81,$CB,$FE,$73,$F8,$BF,$8F,$D7,$FE,$C2,$6C,$EA,$0D,$F1,$53,$C1
	.dc.b	$B5,$FE,$63,$9E,$03,$C5,$D7,$74,$9D,$47,$63,$0F,$F9,$52,$BF,$D5
	.dc.b	$EE,$8F,$F3,$57,$C5,$3D,$3F,$B0,$D8,$CC,$7D,$BF,$88,$7C,$5E,$A2
	.dc.b	$4E,$8F,$F3,$A4,$E9,$C1,$E1,$DB,$D8,$70,$7E,$47,$BD,$F1,$BC,$B7
	.dc.b	$C5,$D3,$E3,$C4,$9F,$F3,$84,$65,$F8,$91,$C7,$E2,$E4,$7F,$9D,$5F
	.dc.b	$9F,$F0,$18,$8F,$F3,$64,$5E,$5F,$F5,$9F,$0A,$BF,$31,$F0,$CF,$8A
	.dc.b	$FF,$1B,$FA,$9F,$E8,$B8,$C6,$3F,$E4,$08,$C4,$EA,$3F,$D7,$7B,$3E
	.dc.b	$2F,$10,$BE,$2F,$F8,$BE,$F8,$BC,$CF,$F6,$1C,$6B,$C8,$FF,$53,$47
	.dc.b	$4A,$E7,$39,$AF,$27,$5E,$0C,$E3,$E3,$E9,$FC,$6F,$B5,$78,$BF,$CE
	.dc.b	$D9,$77,$94,$62,$3C,$87,$FB,$67,$65,$73,$6C,$71,$9E,$1A,$79,$AF
	.dc.b	$9F,$A3,$BC,$38,$78,$F3,$0F,$06,$3F,$DF,$5F,$81,$E6,$3E,$2F,$DF
	.dc.b	$1E,$2C,$7F,$4B,$7C,$63,$17,$C5,$F3,$DF,$1D,$BC,$DE,$4F,$BF,$E2
	.dc.b	$E5,$78,$23,$2F,$A6,$32,$F4,$8C,$BE,$1D,$93,$AD,$BD,$D8,$FF,$7E
	.dc.b	$6F,$B9,$F1,$49,$7E,$EF,$8B,$AF,$A2,$2C,$F2,$9F,$99,$EA,$7D,$5F
	.dc.b	$4A,$5F,$C6,$8C,$BF,$3E,$32,$FC,$28,$FC,$1F,$85,$F8,$4F,$88,$93
	.dc.b	$F8,$B3,$F2,$11,$A7,$1C,$E9,$DC,$7E,$00,$D9,$A7,$38,$FD,$49,$E1
	.dc.b	$3C,$AB,$F2,$7D,$52,$3A,$BF,$96,$ED,$C4,$75,$F8,$97,$B7,$FD,$03
	.dc.b	$1F,$BF,$18,$F5,$38,$D4,$38,$56,$F9,$3F,$DF,$95,$B9,$F4,$67,$FB
	.dc.b	$97,$E0,$46,$70,$EB,$7C,$64,$27,$57,$FB,$DC,$BF,$6F,$18,$9D,$B5
	.dc.b	$E1,$5C,$DF,$DE,$4F,$08,$C4,$EC,$E1,$87,$96,$BD,$78,$F9,$AB,$67
	.dc.b	$39,$D6,$3E,$89,$D1,$2B,$DF,$87,$78,$C4,$EE,$72,$63,$1F,$8F,$87
	.dc.b	$6A,$F9,$95,$B2,$4E,$A3,$2D,$AF,$39,$27,$F0,$57,$E1,$8C,$4E,$E7
	.dc.b	$31,$E1,$3E,$4D,$F8,$5F,$29,$C0,$CD,$ED,$F9,$65,$C7,$E5,$4F,$C5
	.dc.b	$C3,$C1,$7E,$8A,$E2,$74,$70,$FD,$6B,$E9,$8F,$CC,$C7,$CC,$49,$E3
	.dc.b	$5E,$87,$C8,$71,$2C,$F1,$77,$3C,$5D,$4F,$13,$C3,$8F,$A2,$FE,$5C
	.dc.b	$F0,$21,$C2,$BF,$20,$FB,$F1,$F8,$D7,$CE,$7B,$5F,$3A,$F7,$46,$2B
	.dc.b	$6F,$28,$F4,$CE,$EB,$94,$78,$23,$13,$B9,$F3,$D7,$2A,$FE,$A3,$8F
	.dc.b	$CE,$AE,$D8,$7A,$A7,$D0,$73,$AF,$A8,$FA,$63,$E7,$D7,$E1,$63,$FC
	.dc.b	$6B,$87,$ED,$DB,$5F,$E4,$06,$D7,$F9,$03,$D2,$7C,$5B,$E2,$F0,$DA
	.dc.b	$FF,$1F,$89,$DC,$E8,$FF,$24,$E0,$F7,$5E,$CE,$1B,$7C,$A7,$53,$92
	.dc.b	$47,$F1,$79,$DE,$BF,$D5,$3E,$2F,$FC,$61,$E4,$7F,$0E,$9C,$EF,$FC
	.dc.b	$51,$B7,$FC,$21,$AA,$B8,$BD,$70,$E4,$47,$F9,$94,$6B,$0F,$E4,$B5
	.dc.b	$CB,$1F,$3D,$71,$37,$C4,$CB,$95,$F0,$C6,$E5,$FC,$82,$FC,$57,$B7
	.dc.b	$95,$6D,$B3,$BC,$BC,$4F,$B6,$F9,$CF,$B5,$FE,$9D,$3A,$8C,$1C,$8B
	.dc.b	$5F,$26,$FC,$0F,$72,$5F,$59,$F5,$E1,$D9,$FE,$6D,$97,$C1,$3E,$5E
	.dc.b	$1D,$E3,$38,$EF,$2F,$8B,$8F,$7C,$31,$87,$1A,$F3,$E5,$E2,$CB,$C1
	.dc.b	$7E,$2C,$7C,$F1,$F0,$67,$51,$82,$1E,$03,$CE,$7B,$CC,$4F,$37,$E7
	.dc.b	$E1,$E2,$BF,$66,$5E,$9C,$BC,$F8,$F8,$89,$7F,$D5,$A7,$59,$7A,$63
	.dc.b	$B4,$62,$FA,$47,$8B,$1F,$06,$1D,$23,$CF,$5B,$7C,$D1,$97,$CA,$62
	.dc.b	$FB,$57,$C9,$8C,$5F,$A3,$2F,$87,$1E,$EC,$7C,$ED,$3D,$11,$AC,$3B
	.dc.b	$13,$0F,$B3,$19,$7B,$D7,$2C,$7C,$D8,$FA,$A7,$79,$7B,$A3,$E6,$4E
	.dc.b	$AB,$EB,$4F,$94,$98,$FC,$6C,$3B,$65,$F8,$09,$CE,$1C,$A7,$D7,$87
	.dc.b	$E2,$CF,$F1,$24,$EE,$FF,$67,$C1,$AE,$2D,$B6,$7F,$42,$D9,$DC,$7E
	.dc.b	$92,$76,$FF,$84,$0F,$F6,$BD,$7F,$A3,$DC,$4F,$EB,$EB,$F6,$91,$A6
	.dc.b	$C7,$F8,$63,$BB,$FC,$EF,$F1,$7F,$94,$6D,$7F,$9C,$FE,$EB,$1D,$AA
	.dc.b	$A4,$9D,$1F,$E7,$97,$7F,$17,$C3,$E3,$2C,$FF,$9F,$5B,$8F,$F9,$8D
	.dc.b	$BB,$7F,$9F,$F0,$47,$83,$E3,$25,$FF,$A1,$7E,$E6,$E1,$FE,$6B,$9B
	.dc.b	$87,$19,$BC,$5E,$DC,$55,$8B,$6B,$FE,$96,$BC,$78,$6F,$8B,$7A,$9F
	.dc.b	$03,$E3,$3F,$6F,$39,$8C,$6B,$FE,$8E,$AD,$BF,$43,$15,$6E,$30,$D8
	.dc.b	$FE,$4C,$6A,$78,$B8,$C7,$FD,$53,$E9,$7D,$03,$E8,$FD,$D5,$C3,$FD
	.dc.b	$4D,$16,$BF,$D1,$93,$77,$FE,$99,$E2,$FE,$4B,$CB,$FC,$9B,$70,$FE
	.dc.b	$5A,$DB,$FE,$8A,$DA,$E3,$3F,$D9,$27,$FA,$74,$FA,$1F,$E7,$4D,$7F
	.dc.b	$86,$37,$1F,$F6,$0F,$A9,$FD,$0F,$D2,$FE,$89,$1F,$EC,$4F,$80,$C6
	.dc.b	$1F,$56,$AE,$1F,$53,$8D,$58,$BE,$77,$FE,$6E,$C3,$9D,$6A,$30,$73
	.dc.b	$2D,$F3,$BE,$BB,$B8,$9D,$1D,$70,$7C,$56,$A7,$38,$F2,$CF,$FD,$B7
	.dc.b	$39,$BF,$81,$87,$36,$DF,$0C,$BF,$B2,$1F,$82,$70,$FE,$DB,$C8,$F7
	.dc.b	$BF,$DE,$79,$EF,$62,$6C,$61,$B9,$FB,$2B,$FC,$01,$18,$5B,$7E,$32
	.dc.b	$F9,$57,$D9,$6C,$FF,$87,$9A,$F8,$E3,$FC,$43,$97,$C5,$EA,$F8,$9F
	.dc.b	$09,$F0,$DB,$3F,$32,$74,$F0,$CD,$3F,$08,$91,$9E,$9F,$9D,$87,$CD
	.dc.b	$6C,$73,$A4,$7C,$A2,$47,$13,$F2,$5F,$0A,$F4,$BF,$0F,$A3,$E1,$DF
	.dc.b	$11,$F3,$2B,$57,$F8,$2A,$D4,$E6,$FF,$31,$92,$F6,$FF,$90,$9A,$FD
	.dc.b	$A6,$BF,$EB,$ED,$1F,$11,$F3,$30,$FF,$10,$46,$38,$3F,$37,$F3,$CF
	.dc.b	$BB,$7C,$D7,$E4,$B0,$E1,$AF,$36,$E5,$FC,$21,$B7,$C6,$B8,$61,$F8
	.dc.b	$8F,$23,$F8,$DE,$1F,$83,$5E,$96,$7C,$4D,$36,$7F,$75,$3C,$32,$FF
	.dc.b	$2F,$B5,$FC,$BB,$5F,$F3,$1F,$7F,$FB,$73,$E2,$7F,$67,$1F,$07,$9F
	.dc.b	$FC,$7D,$18,$8F,$F2,$14,$EF,$BB,$FC,$89,$C0,$BD,$C7,$4A,$7C,$5F
	.dc.b	$1A,$D6,$5F,$EE,$A8,$E9,$3F,$EE,$CC,$31,$7D,$30,$FF,$27,$56,$A9
	.dc.b	$5E,$17,$CD,$75,$F2,$8C,$7C,$17,$E4,$8F,$DB,$D6,$F1,$F1,$C6,$F1
	.dc.b	$F4,$5F,$0C,$3A,$4F,$94,$D1,$E0,$29,$E0,$8E,$0E,$0E,$18,$E0,$F5
	.dc.b	$61,$EE,$8E,$51,$87,$9E,$1E,$4A,$C1,$D0,$B8,$7C,$5C,$3C,$B5,$F1
	.dc.b	$0E,$38,$FB,$DF,$87,$87,$86,$BE,$A3,$63,$E1,$9C,$2B,$9E,$3F,$16
	.dc.b	$F9,$4F,$C1,$C3,$C5,$87,$C6,$3C,$51,$EF,$3C,$51,$A7,$AB,$5F,$8D
	.dc.b	$1C,$23,$D7,$1F,$2D,$AF,$F6,$C9,$F0,$BD,$27,$81,$E1,$78,$35,$E1
	.dc.b	$7E,$A7,$D8,$79,$6F,$D9,$19,$BF,$24,$FB,$6B,$AB,$E8,$6B,$E8,$3B
	.dc.b	$9F,$23,$D1,$FC,$EB,$71,$FD,$04,$62,$3F,$26,$DF,$41,$EC,$FF,$3A
	.dc.b	$94,$E6,$30,$DE,$54,$EA,$FF,$D6,$57,$CA,$73,$5B,$AF,$C9,$FC,$5B
	.dc.b	$DA,$FF,$5C,$67,$FB,$98,$C4,$FE,$9F,$0E,$4F,$E6,$9B,$3B,$AE,$39
	.dc.b	$FF,$B0,$6F,$8D,$EF,$3F,$55,$FF,$05,$C3,$8D,$EF,$0E,$37,$EA,$C3
	.dc.b	$8D,$EF,$0E,$37,$EC,$6D,$6E,$F8,$AB,$E3,$5B,$CB,$D5,$7C,$2B,$57
	.dc.b	$C2,$F7,$1F,$2B,$1F,$2B,$6B,$84,$EA,$38,$9D,$2E,$3F,$32,$FA,$1E
	.dc.b	$4B,$E7,$06,$D7,$33,$A3,$A3,$33,$CA,$7A,$E3,$F0,$EF,$F8,$24,$78
	.dc.b	$8F,$E5,$38,$FC,$B7,$37,$F6,$33,$E1,$1E,$53,$F9,$16,$3E,$0C,$BE
	.dc.b	$25,$78,$27,$39,$7A,$67,$27,$E4,$EF,$F0,$13,$E2,$BF,$BF,$1D,$8E
	.dc.b	$F9,$6A,$7C,$07,$E0,$F2,$F5,$E5,$E7,$8C,$CF,$08,$FB,$11,$87,$51
	.dc.b	$98,$FA,$93,$E1,$C7,$F4,$31,$F0,$8F,$C2,$1F,$0B,$1F,$45,$F4,$BF
	.dc.b	$21,$1C,$1C,$BE,$27,$FC,$13,$5F,$E5,$F6,$BF,$E3,$A0,$E1,$FF,$1E
	.dc.b	$B5,$E5,$D7,$E4,$C7,$FC,$FB,$2F,$8B,$C6,$FD,$DB,$7F,$BA,$9E,$F5
	.dc.b	$FE,$84,$F8,$7F,$E0,$33,$FE,$3F,$6B,$FE,$3E,$BF,$83,$18,$9F,$F3
	.dc.b	$77,$B8,$F7,$7A,$DA,$B5,$7C,$38,$B3,$18,$9F,$F2,$4D,$F1,$7F,$C9
	.dc.b	$DF,$12,$F8,$AF,$F4,$14,$23,$0F,$FA,$15,$AF,$C7,$6B,$FE,$62,$73
	.dc.b	$97,$B7,$AF,$C2,$79,$A3,$39,$7F,$8B,$7E,$01,$EC,$7F,$A4,$7A,$91
	.dc.b	$8C,$7F,$CA,$ED,$7F,$CD,$F1,$A9,$E1,$39,$C7,$FD,$23,$9B,$E0,$BE
	.dc.b	$8F,$9A,$8C,$63,$FE,$5D,$BE,$19,$78,$FD,$8E,$64,$9F,$0B,$5F,$A8
	.dc.b	$7F,$B3,$2B,$8B,$72,$FB,$F9,$FA,$AB,$EA,$FC,$27,$F8,$66,$BF,$9D
	.dc.b	$B5,$34,$9C,$DC,$3D,$19,$3F,$E8,$39,$FF,$19,$6D,$FE,$CC,$6E,$7F
	.dc.b	$A6,$55,$C2,$BC,$E7,$6B,$F2,$E8,$F4,$FE,$49,$27,$E7,$B5,$F7,$ED
	.dc.b	$F2,$7D,$AF,$97,$B9,$E7,$58,$FA,$DC,$ED,$30,$FF,$56,$5A,$75,$18
	.dc.b	$9D,$3D,$A7,$4F,$BF,$E1,$7F,$27,$A9,$A4,$8F,$F6,$C4,$65,$FF,$25
	.dc.b	$7A,$5F,$C2,$3E,$09,$F0,$BF,$E5,$88,$C3,$67,$E9,$DD,$7E,$9B,$6B
	.dc.b	$8C,$74,$3F,$B9,$B8,$7F,$CA,$75,$F4,$67,$A6,$BF,$8F,$D2,$E5,$F3
	.dc.b	$EF,$94,$E5,$B0,$F9,$0D,$D4,$9C,$BC,$A3,$D1,$C9,$FA,$6C,$7F,$92
	.dc.b	$F1,$FF,$9C,$B8,$51,$8C,$3F,$5C,$D5,$5C,$5F,$F0,$6D,$71,$9F,$3F
	.dc.b	$83,$27,$E0,$56,$DF,$79,$FA,$1F,$0B,$DA,$47,$D7,$F7,$95,$DE,$BF
	.dc.b	$B1,$9C,$78,$FE,$69,$FD,$55,$63,$1F,$F0,$FD,$EE,$7E,$2D,$FC,$AA
	.dc.b	$F9,$7F,$0F,$F3,$3D,$B9,$5F,$2A,$32,$FC,$C8,$CB,$E2,$E0,$4F,$D4
	.dc.b	$EE,$93,$F4,$9F,$AF,$96,$23,$84,$5C,$FD,$19,$FF,$0B,$AF,$C3,$76
	.dc.b	$7C,$DB,$72,$7F,$17,$E0,$FE,$B3,$18,$9D,$4E,$63,$F2,$D1,$F8,$FF
	.dc.b	$87,$91,$F8,$F3,$81,$67,$F9,$16,$3F,$26,$3B,$CF,$07,$EA,$9C,$0B
	.dc.b	$2E,$8F,$CE,$12,$3A,$75,$57,$8F,$C2,$3E,$1F,$FA,$AE,$3D,$E3,$1E
	.dc.b	$27,$8F,$0E,$13,$9A,$DE,$BF,$DE,$2F,$78,$78,$AF,$73,$9A,$DC,$FF
	.dc.b	$97,$AB,$77,$ED,$CB,$D1,$3F,$0E,$73,$7C,$71,$C6,$1C,$AB,$E8,$E3
	.dc.b	$FB,$BB,$DC,$75,$9D,$46,$23,$67,$2A,$8F,$C2,$9E,$98,$78,$A3,$13
	.dc.b	$AC,$7E,$03,$F0,$E3,$18,$F9,$5D,$9F,$04,$C9,$E3,$8A,$F8,$8D,$CF
	.dc.b	$46,$CF,$F1,$BC,$7D,$98,$7A,$89,$7E,$1A,$D9,$D4,$A7,$5B,$E8,$4B
	.dc.b	$F3,$3F,$3E,$FA,$D7,$93,$1F,$8D,$8F,$C1,$AE,$18,$75,$6B,$D7,$0E
	.dc.b	$64,$74,$FA,$5E,$98,$FA,$1F,$88,$79,$AF,$EA,$1E,$0C,$7C,$71,$99
	.dc.b	$FA,$71,$8B,$F0,$D6,$EB,$AE,$3F,$33,$1F,$46,$3E,$39,$57,$6C,$79
	.dc.b	$C7,$03,$E3,$D7,$07,$D2,$43,$8C,$70,$7C,$97,$DF,$1F,$26,$1D,$F0
	.dc.b	$F2,$56,$8E,$B3,$A7,$A4,$6B,$1F,$5E,$1F,$15,$FA,$93,$E5,$7E,$7E
	.dc.b	$3E,$58,$F1,$E3,$9C,$7E,$74,$7E,$40,$FE,$23,$87,$EC,$9B,$87,$F9
	.dc.b	$39,$FC,$B1,$FC,$5B,$CC,$7C,$3B,$E1,$FF,$CC,$93,$FE,$9D,$8C,$3F
	.dc.b	$E6,$79,$DC,$E9,$B1,$FE,$6B,$B7,$C3,$BD,$1E,$71,$86,$C6,$67,$83
	.dc.b	$5F,$4B,$9D,$7F,$89,$37,$3F,$F4,$2F,$BD,$FE,$4A,$E3,$FE,$0D,$39
	.dc.b	$BF,$E1,$8D,$7F,$02,$D9,$E1,$1F,$C8,$23,$58,$F9,$63,$18,$FF,$33
	.dc.b	$AD,$E3,$E2,$8E,$E4,$9F,$E6,$33,$AC,$7F,$91,$61,$D0,$95,$CF,$8D
	.dc.b	$97,$1C,$7A,$63,$E0,$39,$CE,$9C,$46,$89,$87,$C2,$9D,$46,$27,$E0
	.dc.b	$DF,$5A,$D5,$D9,$C9,$D2,$32,$D3,$35,$CD,$63,$ED,$73,$3E,$EB,$E9
	.dc.b	$1A,$C3,$C7,$39,$8E,$87,$D2,$9F,$85,$87,$0C,$BE,$86,$3D,$F1,$ED
	.dc.b	$7B,$38,$CE,$B0,$F3,$D7,$96,$31,$3E,$40,$78,$EF,$B6,$5F,$4A,$3E
	.dc.b	$6C,$63,$1F,$4E,$3E,$0B,$E7,$5E,$1C,$3C,$99,$7B,$23,$B1,$9C,$BD
	.dc.b	$D3,$FD,$D6,$F9,$65,$DE,$38,$E5,$DE,$FE,$1D,$6B,$2C,$57,$1C,$7B
	.dc.b	$57,$53,$BE,$5E,$4B,$E1,$1F,$13,$1F,$CE,$57,$0A,$F4,$92,$33,$8F
	.dc.b	$6C,$BC,$E7,$CA,$2D,$F6,$9D,$BD,$72,$F8,$B8,$75,$9C,$E5,$EC,$AF
	.dc.b	$5C,$E6,$BE,$EC,$63,$0F,$F2,$7B,$F8,$E7,$FC,$4B,$97,$EC,$9B,$5F
	.dc.b	$9F,$69,$CA,$FE,$67,$1F,$C4,$AF,$C2,$9F,$E1,$D9,$FD,$B7,$1E,$78
	.dc.b	$F4,$D8,$F0,$E3,$F8,$4A,$E0,$FE,$B4,$C5,$78,$5F,$E1,$AD,$73,$3F
	.dc.b	$C0,$E3,$C5,$E1,$F5,$9F,$F2,$FF,$C0,$BE,$1F,$F0,$DC,$E8,$C6,$1F
	.dc.b	$E8,$25,$F0,$3A,$73,$7C,$AF,$06,$66,$D7,$1F,$81,$E1,$F0,$FC,$5B
	.dc.b	$19,$3B,$46,$5A,$78,$7D,$9F,$01,$E1,$F8,$2B,$19,$26,$3E,$C8,$CB
	.dc.b	$EE,$9B,$C3,$FE,$64,$CD,$58,$E5,$FD,$8F,$86,$F8,$67,$C3,$FB,$FE
	.dc.b	$13,$F5,$53,$98,$C6,$9F,$E9,$0A,$DB,$63,$E3,$7C,$DF,$09,$CE,$BF
	.dc.b	$43,$E2,$3E,$51,$F2,$FE,$22,$E1,$F3,$E2,$CF,$D6,$9B,$B7,$ED,$FC
	.dc.b	$7F,$C6,$79,$BF,$8D,$6E,$1F,$D0,$5F,$C3,$5F,$F5,$06,$BF,$8B,$6B
	.dc.b	$FC,$99,$BD,$27,$F9,$D7,$A7,$F8,$F8,$C5,$7F,$16,$8C,$79,$1A,$8E
	.dc.b	$C6,$4F,$F5,$2F,$1F,$FB,$F9,$C6,$0E,$9F,$C7,$7A,$5C,$63,$15,$B7
	.dc.b	$FD,$F6,$A3,$53,$F9,$5F,$4E,$CE,$6F,$AC,$E5,$B1,$9B,$63,$FD,$57
	.dc.b	$8F,$F0,$3D,$5C,$BE,$23,$C5,$18,$C3,$8C,$63,$A3,$F4,$31,$EA,$9D
	.dc.b	$61,$E6,$F4,$B5,$87,$AE,$3F,$A8,$F9,$97,$83,$FD,$BF,$87,$C2,$9C
	.dc.b	$B6,$33,$1E,$68,$FF,$03,$CF,$D4,$23,$FE,$1E,$94,$62,$78,$1F,$C1
	.dc.b	$DF,$07,$BD,$E1,$7D,$5D,$DF,$23,$E0,$2E,$1F,$19,$B7,$CB,$3F,$D2
	.dc.b	$D7,$CE,$F1,$B3,$BF,$C9,$D9,$DB,$F4,$B0,$F8,$CD,$CB,$E3,$4E,$9B
	.dc.b	$1F,$6B,$CF,$7D,$4F,$09,$F8,$9F,$0E,$F8,$8F,$CA,$B6,$3F,$C7,$AD
	.dc.b	$C7,$FD,$AE,$1C,$3F,$ED,$A7,$59,$7F,$27,$6C,$FE,$FB,$E1,$FE,$4B
	.dc.b	$E0,$BF,$C0,$C6,$2F,$E1,$36,$FF,$C3,$9F,$0E,$F8,$87,$C4,$70,$E3
	.dc.b	$99,$7E,$17,$0E,$0F,$F7,$CB,$E1,$5A,$C7,$EB,$9F,$E4,$FF,$1F,$F6
	.dc.b	$7C,$26,$92,$FF,$CB,$B1,$97,$F1,$4D,$9E,$7F,$7B,$F3,$27,$35,$F9
	.dc.b	$39,$FF,$06,$F8,$5F,$E0,$EE,$7F,$D6,$47,$F2,$0E,$7F,$F1,$C4,$62
	.dc.b	$3F,$C2,$62,$78,$77,$E3,$A1,$7C,$23,$E9,$CF,$F8,$6B,$E1,$F9,$56
	.dc.b	$B2,$C5,$6B,$05,$6B,$0F,$F2,$43,$E1,$8F,$39,$A9,$F0,$4E,$9E,$53
	.dc.b	$C6,$BA,$D7,$A0,$F4,$1C,$4F,$73,$A7,$DB,$85,$C3,$CD,$1B,$32,$31
	.dc.b	$EC,$DC,$3D,$51,$83,$D1,$1E,$0C,$36,$62,$FB,$63,$E9,$9F,$34,$63
	.dc.b	$1F,$24,$7C,$0A,$E3,$87,$63,$E0,$CF,$D2,$9E,$27,$8C,$F6,$94,$F4
	.dc.b	$CF,$27,$6D,$73,$3D,$1F,$80,$78,$5C,$47,$A6,$3A,$63,$F1,$B1,$C1
	.dc.b	$F0,$6B,$E9,$9C,$0B,$3E,$4A,$E4,$78,$5E,$C6,$63,$6A,$BE,$9C,$F0
	.dc.b	$9E,$93,$F0,$2B,$C0,$D8,$F5,$46,$E3,$A1,$F0,$23,$8B,$F7,$23,$8B
	.dc.b	$63,$89,$F3,$DF,$97,$1E,$E8,$CC,$7D,$0F,$5F,$F0,$8D,$9F,$CE,$46
	.dc.b	$23,$F1,$2D,$F5,$1E,$DF,$F5,$29,$4E,$63,$0D,$E5,$FF,$41,$B6,$77
	.dc.b	$7F,$B4,$BE,$75,$F8,$39,$7C,$43,$DC,$FD,$CD,$FE,$EE,$31,$3B,$BE
	.dc.b	$2F,$E1,$DB,$3B,$AE,$39,$7F,$02,$BE,$37,$BC,$BD,$57,$FC,$1F,$0E
	.dc.b	$79,$F1,$AF,$D7,$5F,$08,$D5,$FC,$16,$E5,$E9,$9E,$B8,$71,$BE,$D3
	.dc.b	$E7,$BE,$18,$F8,$F2,$FE,$2B,$8E,$CE,$77,$FB,$9C,$BC,$75,$B8,$F1
	.dc.b	$D7,$08,$F9,$57,$D2,$7F,$D6,$35,$C1,$FF,$59,$E5,$E9,$AE,$51,$D6
	.dc.b	$B8,$2A,$F9,$33,$AC,$BD,$47,$90,$F2,$E3,$E7,$8E,$6E,$1F,$1E,$5E
	.dc.b	$6B,$87,$77,$F8,$C5,$F6,$8F,$2E,$5F,$59,$A7,$C7,$98,$FF,$1B,$CB
	.dc.b	$D4,$E8,$E8,$4C,$3C,$8E,$5F,$23,$F1,$32,$FB,$F8,$FB,$49,$19,$BE
	.dc.b	$F5,$CA,$3F,$27,$87,$CF,$74,$77,$AF,$82,$47,$33,$E0,$78,$9F,$09
	.dc.b	$C6,$3F,$06,$32,$7F,$1F,$C7,$F1,$35,$F2,$A3,$F9,$17,$C3,$3E,$23
	.dc.b	$FA,$DE,$5F,$EA,$98,$C4,$FF,$8C,$DA,$FF,$74,$FB,$E7,$B9,$C3,$87
	.dc.b	$FB,$1D,$6A,$BF,$A7,$35,$CE,$1F,$E1,$47,$FD,$F7,$19,$7F,$DF,$BF
	.dc.b	$0F,$FC,$01,$FF,$21,$35,$FF,$01,$B7,$1F,$F3,$8F,$80,$F8,$8F,$15
	.dc.b	$72,$E3,$AA,$F8,$65,$FD,$E1,$B1,$FE,$60,$BE,$4F,$F9,$47,$E1,$DF
	.dc.b	$11,$C9,$B1,$98,$EB,$5C,$A1,$3A,$7F,$C6,$2D,$7F,$B8,$35,$FF,$04
	.dc.b	$B5,$FE,$79,$F0,$8F,$7B,$D7,$D0,$8C,$4F,$F9,$B5,$AF,$F9,$73,$CE
	.dc.b	$64,$F8,$6E,$35,$AD,$DA,$B0,$FF,$47,$5B,$E2,$3E,$4E,$3C,$9E,$39
	.dc.b	$F2,$F7,$A5,$F8,$5D,$DB,$E9,$F3,$3E,$38,$F0,$63,$DF,$1F,$2E,$1A
	.dc.b	$CB,$FD,$ED,$F1,$3F,$E2,$B6,$F4,$7D,$88,$C5,$FF,$8B,$97,$B3,$99
	.dc.b	$FC,$56,$49,$FF,$06,$BE,$D6,$B9,$8D,$C7,$48,$D7,$AB,$DC,$E8,$FC
	.dc.b	$DF,$63,$E4,$91,$FF,$54,$E8,$F8,$2C,$47,$BF,$E0,$BE,$19,$8D,$3E
	.dc.b	$7C,$7C,$EF,$82,$F8,$FC,$32,$39,$C7,$CA,$30,$FC,$98,$C3,$FE,$A3
	.dc.b	$CD,$EB,$FC,$52,$46,$7D,$8A,$FF,$A0,$27,$6F,$EB,$DA,$E6,$7E,$19
	.dc.b	$FE,$3E,$F8,$0F,$F2,$DF,$C3,$FE,$26,$7F,$CC,$71,$8E,$8D,$47,$CD
	.dc.b	$AB,$CB,$F9,$0E,$0B,$C0,$69,$8F,$89,$B7,$D0,$FC,$8C,$61,$FF,$35
	.dc.b	$63,$F8,$7A,$57,$83,$AB,$F4,$53,$7A,$8D,$3C,$3E,$15,$C3,$B6,$5F
	.dc.b	$A4,$F2,$3C,$A6,$63,$B9,$FA,$2F,$23,$D2,$47,$8E,$6F,$89,$C4,$70
	.dc.b	$BF,$D7,$D6,$3C,$BF,$1C,$FE,$B2,$B1,$97,$AE,$AC,$E4,$EB,$EF,$7C
	.dc.b	$8F,$7E,$46,$08,$77,$C9,$1A,$8C,$4F,$D0,$F5,$64,$7A,$1F,$A7,$86
	.dc.b	$23,$94,$5C,$BC,$F9,$FF,$08,$AF,$BB,$4E,$4F,$E1,$B5,$EC,$8C,$78
	.dc.b	$1A,$9C,$D7,$E2,$3C,$1C,$4F,$FA,$93,$C1,$FF,$47,$57,$F1,$66,$C6
	.dc.b	$6F,$F1,$ED,$7F,$8A,$4F,$46,$F7,$77,$3A,$BF,$89,$3C,$DD,$7E,$32
	.dc.b	$39,$97,$76,$23,$F6,$67,$69,$F0,$3C,$8F,$E4,$98,$7A,$23,$34,$7E
	.dc.b	$2F,$C3,$BE,$27,$FA,$B6,$BF,$16,$31,$E4,$78,$6B,$C1,$39,$AD,$ED
	.dc.b	$FE,$C8,$BD,$E5,$F9,$23,$C4,$57,$35,$C7,$0F,$81,$18,$AF,$DB,$E1
	.dc.b	$E4,$C3,$AE,$1F,$CF,$6B,$4D,$8C,$BD,$B0,$E8,$F7,$8C,$CF,$08,$D5
	.dc.b	$F9,$E7,$24,$C3,$DD,$3E,$48,$C3,$CE,$32,$D2,$5F,$80,$FC,$66,$3F
	.dc.b	$36,$39,$4C,$77,$7E,$E7,$15,$C6,$3C,$93,$DA,$31,$3E,$F8,$DC,$F1
	.dc.b	$8E,$66,$4A,$7B,$8C,$4F,$25,$87,$C3,$BF,$34,$70,$8F,$8C,$67,$1F
	.dc.b	$A5,$87,$82,$79,$E1,$E1,$C3,$A5,$F0,$C7,$E9,$E3,$D2,$73,$1F,$29
	.dc.b	$F8,$B8,$F4,$C3,$C1,$3C,$23,$13,$C3,$1F,$85,$8F,$C4,$BD,$C6,$1F
	.dc.b	$19,$0F,$6D,$F3,$7D,$06,$22,$E1,$E5,$AE,$97,$DE,$BF,$31,$39,$7A
	.dc.b	$63,$ED,$BF,$04,$73,$3C,$67,$0C,$7D,$38,$74,$24,$7E,$D4,$F1,$D7
	.dc.b	$37,$81,$8C,$7E,$34,$FC,$D3,$8C,$77,$C7,$ED,$63,$E5,$8C,$4E,$AF
	.dc.b	$EC,$47,$E3,$CF,$F1,$21,$F8,$52,$9E,$92,$BF,$B8,$73,$E7,$7F,$98
	.dc.b	$BE,$27,$FC,$B8,$D8,$E2,$7F,$A8,$7A,$53,$A6,$C6,$A9,$F1,$2F,$5B
	.dc.b	$FC,$CD,$18,$6C,$66,$78,$35,$F2,$B9,$CF,$F8,$A3,$72,$F6,$FC,$0F
	.dc.b	$F9,$37,$83,$F8,$5C,$E6,$FF,$84,$35,$FD,$0B,$67,$84,$7F,$15,$8D
	.dc.b	$63,$E5,$8C,$63,$FC,$D6,$B7,$8F,$8A,$3B,$92,$7F,$90,$4E,$B1,$FE
	.dc.b	$49,$87,$42,$57,$3E,$3F,$F1,$5E,$1C,$F0,$EE,$47,$FA,$B6,$1C,$E3
	.dc.b	$D7,$5D,$E7,$81,$AA,$E7,$87,$82,$7F,$C0,$FF,$7E,$FB,$2C,$E0,$E7
	.dc.b	$4E,$75,$DF,$DB,$EA,$73,$76,$B5,$7F,$E0,$76,$E7,$FD,$C1,$5E,$EB
	.dc.b	$87,$D5,$7F,$92,$3C,$AF,$F2,$4E,$3E,$0F,$53,$19,$FF,$9A,$F1,$5E
	.dc.b	$33,$F3,$47,$FA,$EE,$7F,$CF,$D8,$72,$8D,$C6,$0E,$57,$C7,$0F,$6E
	.dc.b	$16,$B9,$61,$FE,$8D,$EA,$79,$7B,$7F,$B2,$63,$CE,$7E,$34,$6A,$31
	.dc.b	$87,$F9,$9F,$A7,$85,$F1,$6C,$EA,$B9,$DF,$7A,$D4,$E5,$B1,$FE,$9F
	.dc.b	$C1,$F0,$FE,$7F,$BF,$F4,$36,$FC,$33,$A6,$CE,$9A,$6A,$78,$57,$56
	.dc.b	$EB,$F4,$B1,$FF,$2E,$57,$9F,$C8,$F8,$5E,$4F,$C3,$6C,$64,$F6,$46
	.dc.b	$5A,$67,$0F,$99,$19,$7D,$9E,$F7,$D1,$FA,$0F,$A4,$DA,$F5,$B6,$77
	.dc.b	$1E,$E9,$DB,$4D,$96,$3E,$D7,$93,$F6,$7B,$3E,$D3,$67,$E9,$36,$77
	.dc.b	$1F,$56,$76,$D3,$65,$8F,$A9,$F2,$1F,$89,$3F,$15,$EB,$D9,$FA,$6D
	.dc.b	$9F,$B4,$DD,$7F,$2C,$F0,$9F,$B3,$F4,$36,$BF,$01,$F3,$FF,$95,$3F
	.dc.b	$25,$E6,$DA,$FC,$57,$97,$86,$9F,$8D,$8F,$CC,$79,$B2,$3F,$0B,$F3
	.dc.b	$F7,$06,$7F,$EB,$8F,$A1,$FE,$6F,$DD,$FA,$D7,$3E,$55,$F1,$FF,$73
	.dc.b	$8F,$E8,$67,$2D,$8C,$E1,$D2,$3F,$DE,$13,$A8,$C0,$7B,$C7,$3C,$17
	.dc.b	$DF,$0F,$14,$62,$7B,$46,$27,$F7,$1F,$3B,$FB,$9F,$A0,$B5,$FA,$26
	.dc.b	$E3,$FA,$96,$E1,$FC,$15,$FE,$15,$E2,$3E,$89,$DB,$87,$F8,$4D,$FA
	.dc.b	$27,$53,$9F,$0B,$4D,$C7,$F8,$24,$E9,$B7,$CB,$1F,$E2,$9E,$6F,$F0
	.dc.b	$FF,$A1,$FE,$20,$DA,$FD,$CB,$72,$FE,$00,$DA,$E3,$3F,$BE,$AE,$2D
	.dc.b	$3F,$8A,$53,$E7,$DF,$0D,$FC,$7F,$D5,$B5,$A9,$B5,$F5,$3C,$75,$63
	.dc.b	$1F,$F4,$4D,$6A,$BF,$C0,$AD,$7F,$C1,$2D,$47,$87,$0F,$F4,$8B,$FC
	.dc.b	$E7,$E5,$FF,$97,$4F,$F2,$F6,$E3,$FC,$61,$B3,$FC,$7D,$B7,$F9,$5E
	.dc.b	$5F,$E6,$AD,$F2,$BF,$29,$E5,$7E,$55,$B8,$7E,$4D,$B8,$7E,$43,$9F
	.dc.b	$FC,$27,$18,$BE,$55,$F9,$9B,$E4,$FF,$3A,$FB,$EF,$CD,$CF,$F2,$3E
	.dc.b	$7F,$E9,$11,$89,$F7,$EE,$F1,$FF,$2D,$9F,$CB,$AD,$3F,$D1,$AF,$8F
	.dc.b	$4F,$F4,$B7,$FA,$D4,$3E,$F3,$DB,$8F,$FA,$D3,$EF,$DF,$19,$A9,$CD
	.dc.b	$29,$5A,$7F,$03,$3F,$D6,$A7,$FC,$76,$D7,$FC,$7E,$D7,$EB,$B7,$7F
	.dc.b	$F1,$BF,$4F,$EF,$FA,$BF,$80,$36,$FF,$7C,$DB,$F9,$9C,$5F,$A8,$9F
	.dc.b	$E0,$55,$C1,$A7,$02,$CF,$F0,$23,$E0,$78,$5F,$B2,$FB,$E7,$A3,$EF
	.dc.b	$8C,$36,$BF,$C9,$B5,$C1,$B9,$AB,$F1,$F3,$EC,$6A,$F9,$37,$C9,$FF
	.dc.b	$82,$79,$5F,$3F,$F3,$2B,$53,$98,$C5,$70,$C7,$C2,$DA,$FF,$2F,$B5
	.dc.b	$FF,$34,$B5,$FC,$2F,$E0,$0F,$A2,$FF,$07,$57,$F8,$46,$30,$FF,$78
	.dc.b	$31,$1A,$7F,$C1,$71,$A9,$FF,$01,$47,$1C,$23,$AF,$A2,$B3,$FB,$BA
	.dc.b	$F3,$65,$FE,$2C,$6D,$79,$2B,$73,$F2,$3D,$7F,$DC,$FC,$93,$E2,$BF
	.dc.b	$D0,$BF,$0A,$F9,$3F,$C8,$66,$AD,$3F,$E8,$8B,$E3,$7C,$2B,$75,$AA
	.dc.b	$FF,$47,$D6,$9F,$E0,$F7,$D7,$2C,$46,$8E,$D8,$73,$8C,$C7,$F0,$EA
	.dc.b	$DD,$71,$8D,$9E,$32,$87,$57,$E3,$8E,$46,$A3,$37,$FE,$9A,$AE,$B8
	.dc.b	$7A,$A3,$15,$B8,$C3,$DE,$FD,$D7,$CA,$31,$5E,$97,$C6,$78,$27,$75
	.dc.b	$CA,$3E,$04,$63,$0F,$7C,$30,$F1,$C6,$1B,$1D,$0D,$1E,$58,$D3,$4D
	.dc.b	$47,$09,$F8,$51,$C1,$F0,$3A,$38,$3A,$8E,$17,$F0,$A3,$66,$67,$71
	.dc.b	$93,$E0,$46,$4D,$B9,$8F,$C0,$DF,$C1,$7E,$24,$70,$34,$FD,$98,$D3
	.dc.b	$E7,$8D,$3F,$38,$E7,$7F,$02,$BB,$C6,$4E,$51,$99,$E3,$1F,$41,$AF
	.dc.b	$88,$E6,$7C,$39,$E6,$EC,$CB,$ED,$8E,$31,$E5,$32,$FA,$B0,$FC,$24
	.dc.b	$7E,$77,$AB,$FB,$AB,$F9,$76,$B9,$F0,$FF,$1F,$18,$F8,$07,$A9,$FE
	.dc.b	$38,$AF,$F1,$3D,$7F,$BE,$A7,$3A,$FF,$72,$BE,$17,$B6,$C6,$6B,$84
	.dc.b	$F3,$CB,$F6,$DF,$26,$F3,$FF,$BD,$E5,$FA,$A8,$C3,$FB,$6C,$BB,$BF
	.dc.b	$AE,$6D,$72,$AE,$2D,$9E,$17,$DF,$4F,$0E,$5F,$C0,$31,$EB,$5D,$F0
	.dc.b	$E3,$5A,$8E,$D7,$C5,$FF,$07,$D7,$0C,$3B,$46,$B0,$EE,$75,$C3,$9C
	.dc.b	$70,$6A,$C7,$D3,$8F,$7B,$E6,$47,$04,$9E,$47,$53,$B9,$E1,$3D,$46
	.dc.b	$4C,$47,$F2,$39,$EF,$1C,$8D,$9A,$9E,$CB,$0F,$74,$7D,$46,$C7,$F8
	.dc.b	$C2,$79,$BE,$BC,$7E,$33,$C7,$0F,$7D,$70,$8D,$63,$E8,$C7,$DD,$1E
	.dc.b	$37,$CD,$7C,$B0,$EB,$7C,$4E,$4F,$80,$CE,$1D,$8F,$9A,$E6,$7C,$73
	.dc.b	$C6,$73,$1D,$4C,$BC,$6F,$C9,$8F,$C2,$AE,$11,$83,$E0,$57,$07,$E7
	.dc.b	$61,$E8,$C7,$51,$F0,$4C,$DF,$CC,$C7,$CF,$18,$39,$CF,$F5,$5F,$0D
	.dc.b	$C4,$FA,$2F,$D1,$CF,$F6,$B6,$F0,$7F,$7E,$6D,$AF,$8F,$C9,$61,$E7
	.dc.b	$37,$0E,$95,$FD,$FA,$31,$59,$FA,$2D,$7C,$F4,$AF,$F0,$1C,$65,$FF
	.dc.b	$02,$C6,$5F,$F0,$04,$78,$97,$C9,$BD,$8F,$F1,$4C,$61,$B1,$98,$FF
	.dc.b	$16,$B5,$FE,$02,$7F,$9F,$1C,$69,$FE,$02,$87,$CA,$3E,$2F,$C0,$D8
	.dc.b	$FF,$49,$D7,$0D,$BF,$85,$FD,$EB,$4F,$1B,$6B,$FC,$7A,$D7,$C2,$1C
	.dc.b	$3E,$47,$58,$F8,$FE,$89,$F2,$9F,$05,$B5,$FE,$A4,$AD,$E1,$FC,$72
	.dc.b	$7F,$CD,$B3,$FE,$A8,$5F,$3F,$C4,$E5,$E6,$49,$D4,$65,$E1,$19,$7F
	.dc.b	$CB,$93,$C9,$D9,$A3,$C5,$1F,$00,$8F,$C0,$6B,$CB,$87,$D4,$7D,$BC
	.dc.b	$3C,$F1,$FE,$8D,$F4,$FF,$19,$E7,$BE,$2E,$98,$89,$3F,$E9,$B8,$CB
	.dc.b	$FE,$9D,$8F,$8D,$E1,$FC,$7F,$0E,$4F,$C3,$8C,$BF,$16,$32,$FB,$A3
	.dc.b	$2F,$C0,$8C,$EF,$F4,$E7,$F3,$DF,$41,$F9,$FF,$A3,$97,$FE,$5C,$8C
	.dc.b	$BF,$E5,$E8,$CB,$FE,$59,$8F,$81,$F3,$1F,$86,$24,$7E,$19,$AF,$FB
	.dc.b	$67,$D8,$78,$9F,$AF,$6C,$7E,$06,$7F,$61,$F4,$3F,$B1,$E1,$93,$FE
	.dc.b	$6E,$8C,$BF,$E7,$08,$CB,$FE,$6B,$8C,$BF,$E6,$D8,$CE,$7F,$E6,$D9
	.dc.b	$FE,$03,$EE,$BF,$01,$7C,$1E,$1F,$EB,$B5,$FE,$6A,$9D,$78,$F9,$A9
	.dc.b	$8F,$FB,$AA,$B5,$7C,$3A,$3F,$DD,$B8,$70,$AD,$46,$27,$BC,$7F,$A2
	.dc.b	$3E,$8F,$F8,$77,$44,$8C,$10,$FF,$43,$39,$7F,$D1,$11,$FC,$27,$F0
	.dc.b	$27,$BD,$79,$A5,$78,$0F,$F7,$FC,$E9,$C0,$70,$FF,$AF,$D4,$72,$8C
	.dc.b	$4F,$73,$19,$FF,$0D,$AD,$CF,$47,$9C,$7F,$A4,$BC,$57,$E0,$7F,$EB
	.dc.b	$CA,$F2,$46,$39,$FF,$98,$4E,$67,$FD,$49,$C7,$8E,$FC,$F6,$C9,$E3
	.dc.b	$5F,$C1,$9F,$E7,$5E,$37,$FA,$67,$2F,$F5,$57,$91,$FC,$E4,$E1,$3F
	.dc.b	$EA,$FF,$9D,$FE,$78,$47,$FD,$73,$E1,$57,$FD,$69,$F3,$7F,$D0,$4C
	.dc.b	$5F,$FA,$D5,$C4,$7F,$B0,$1C,$F2,$A3,$FA,$66,$4F,$23,$DB,$E4,$C9
	.dc.b	$F4,$C7,$C0,$30,$B0,$F9,$14,$AD,$4F,$03,$E0,$BC,$27,$75,$FD,$5F
	.dc.b	$A3,$FA,$C7,$A1,$2F,$FD,$AB,$19,$7F,$DA,$F1,$97,$FD,$A5,$19,$7F
	.dc.b	$DA,$91,$9A,$FF,$6A,$4F,$D6,$F0,$3E,$BF,$83,$2F,$E9,$C6,$5F,$AB
	.dc.b	$19,$7D,$99,$FC,$5E,$5F,$C1,$D6,$E7,$27,$C8,$AE,$18,$7C,$29,$FE
	.dc.b	$E9,$EF,$BE,$97,$FB,$A5,$C3,$F1,$71,$96,$C6,$5F,$F7,$71,$C9,$8F
	.dc.b	$FB,$E0,$D5,$7E,$5E,$79,$47,$93,$0F,$CC,$F8,$3F,$96,$F2,$24,$FE
	.dc.b	$4A,$3F,$32,$61,$E1,$C1,$FD,$35,$B1,$A7,$F3,$33,$BC,$3E,$FC,$FB
	.dc.b	$BE,$7F,$DF,$F3,$D2,$FD,$B1,$97,$DF,$19,$7C,$D9,$27,$E3,$FC,$BF
	.dc.b	$FB,$EB,$DC,$FA,$12,$8C,$57,$F8,$86,$30,$FF,$BD,$E7,$7F,$39,$AD
	.dc.b	$55,$AA,$FF,$6F,$57,$4B,$D4,$E7,$D8,$4F,$2A,$E8,$7E,$97,$E6,$9F
	.dc.b	$13,$FD,$02,$B7,$39,$8C,$57,$FA,$72,$B7,$E6,$72,$A7,$CD,$BE,$6B
	.dc.b	$D7,$0A,$DB,$E6,$3A,$57,$18,$CE,$3E,$09,$E7,$5C,$A7,$38,$78,$71
	.dc.b	$F1,$D7,$23,$A1,$C0,$E8,$E9,$F1,$36,$79,$C7,$F0,$5C,$7D,$0F,$0C
	.dc.b	$63,$D2,$76,$F5,$31,$5C,$C9,$3F,$0E,$B9,$CE,$A3,$9E,$3E,$DB,$E7
	.dc.b	$5A,$8C,$5F,$F7,$0B,$FF,$24,$D7,$C8,$6C,$65,$F8,$57,$FE,$3E,$7F
	.dc.b	$D6,$5F,$26,$F8,$5C,$7B,$28,$C4,$FF,$91,$23,$13,$A9,$CC,$F8,$3D
	.dc.b	$37,$67,$5C,$BF,$C4,$CD,$8C,$CF,$F9,$3A,$72,$D8,$CE,$7E,$F6,$EB
	.dc.b	$FE,$E4,$F9,$BF,$D0,$76,$FE,$06,$73,$7F,$93,$6B,$F8,$96,$CF,$08
	.dc.b	$FA,$91,$AC,$3C,$38,$77,$C7,$F2,$97,$B8,$C5,$F0,$9F,$01,$AE,$26
	.dc.b	$38,$9D,$12,$BF,$29,$7C,$F2,$FD,$16,$5E,$12,$63,$CF,$0E,$57,$E0
	.dc.b	$25,$7C,$1C,$7B,$E3,$D6,$3F,$99,$7C,$D3,$E3,$73,$7F,$B2,$E5,$4E
	.dc.b	$71,$FD,$CC,$63,$E3,$37,$87,$2C,$2E,$3F,$EA,$BC,$38,$3F,$B1,$BE
	.dc.b	$73,$A7,$96,$1D,$2B,$78,$F8,$57,$C4,$FA,$3B,$3F,$C1,$50,$F8,$DC
	.dc.b	$5F,$AA,$AE,$5E,$19,$B8,$FA,$A6,$DF,$1B,$FF,$08,$BD,$9B,$5C,$3A
	.dc.b	$5F,$37,$7A,$B0,$DA,$FE,$9E,$DC,$BF,$98,$B7,$3F,$E9,$4D,$9F,$E6
	.dc.b	$2D,$9F,$8D,$F3,$7E,$4F,$8D,$F2,$B6,$BC,$AD,$9F,$EA,$59,$A3,$18
	.dc.b	$7C,$46,$D7,$DC,$6D,$7F,$3D,$F9,$AF,$67,$CD,$FB,$5B,$5F,$DA,$1B
	.dc.b	$3F,$D2,$9B,$7F,$D8,$DB,$3F,$D2,$9B,$3F,$7F,$E6,$FE,$6F,$CD,$FC
	.dc.b	$E6,$D7,$CE,$6C,$FA,$9B,$7F,$71,$B3,$F9,$46,$CF,$F5,$5F,$88,$FA
	.dc.b	$9D,$9F,$55,$B5,$FE,$00,$6C,$FF,$62,$6D,$FF,$7B,$6C,$FF,$62,$6C
	.dc.b	$FE,$5F,$E6,$FF,$0A,$7E,$1B,$E2,$2D,$7E,$19,$B3,$F4,$9B,$7F,$94
	.dc.b	$78,$4F,$E3,$3E,$42,$D7,$D8,$F9,$BF,$C8,$9F,$8D,$F9,$BB,$5F,$E1
	.dc.b	$8F,$82,$C3,$4F,$F0,$DC,$7E,$4F,$E0,$E4,$7E,$63,$E6,$EE,$1F,$E2
	.dc.b	$8F,$96,$FE,$65,$E0,$7E,$89,$FF,$25,$79,$17,$C7,$4E,$77,$FF,$58
	.dc.b	$E3,$FE,$85,$BE,$2D,$8C,$E1,$D6,$F8,$CE,$70,$E9,$87,$36,$C6,$4F
	.dc.b	$D6,$FC,$DF,$EC,$3E,$6F,$F6,$2D,$8C,$9F,$B2,$8C,$B4,$FF,$4C,$78
	.dc.b	$CF,$9B,$76,$FE,$FA,$31,$E1,$59,$CD,$70,$7F,$79,$A7,$F9,$32,$B9
	.dc.b	$5E,$EF,$83,$6F,$F8,$1F,$99,$FC,$1B,$E6,$FF,$83,$B6,$32,$7F,$08
	.dc.b	$8C,$B4,$CE,$1F,$B6,$8C,$B6,$3F,$D5,$FF,$0C,$F9,$4F,$E6,$73,$FE
	.dc.b	$87,$3F,$D1,$11,$99,$FE,$DC,$D8,$D1,$A9,$E2,$F1,$8D,$70,$F4,$6D
	.dc.b	$7F,$3E,$F3,$3F,$92,$D7,$FA,$F2,$FF,$91,$FC,$3F,$F2,$4B,$B7,$FC
	.dc.b	$91,$B5,$FC,$41,$B8,$7F,$32,$7F,$19,$E4,$7E,$37,$CB,$B7,$F8,$BE
	.dc.b	$5F,$E8,$37,$FD,$43,$9B,$FC,$0D,$7B,$AD,$37,$1F,$C7,$36,$B8,$5F
	.dc.b	$F3,$79,$FC,$96,$1F,$CF,$3E,$3B,$F2,$71,$FE,$60,$E6,$C7,$3F,$F3
	.dc.b	$96,$E5,$FE,$84,$6C,$EE,$3F,$DC,$3F,$33,$D6,$3E,$6B,$5F,$A8,$D7
	.dc.b	$FD,$78,$D4,$F9,$72,$FB,$6D,$9F,$F8,$03,$F3,$4F,$9D,$FA,$33,$9B
	.dc.b	$6D,$FE,$29,$E5,$FD,$B7,$37,$ED,$DB,$8F,$CB,$6D,$7E,$CD,$B9,$7C
	.dc.b	$66,$CF,$FB,$46,$FA,$3F,$4E,$FC,$3E,$07,$E2,$30,$F5,$B5,$F5,$B7
	.dc.b	$13,$4E,$35,$FE,$DD,$C7,$FC,$17,$F3,$0F,$A4,$78,$3A,$C7,$FD,$76
	.dc.b	$3E,$6D,$F1,$BD,$DB,$8F,$F9,$2D,$B3,$F9,$E3,$84,$EC,$FC,$CC,$6B
	.dc.b	$1F,$0E,$0F,$98,$CD,$7F,$80,$3E,$2D,$F3,$7F,$E0,$0C,$B0,$D7,$D4
	.dc.b	$7F,$B4,$E3,$6F,$F8,$10,$D9,$4D,$E4,$F5,$53,$E3,$9E,$57,$E1,$F9
	.dc.b	$6F,$13,$23,$8D,$70,$8D,$46,$2B,$95,$FE,$FB,$D7,$2F,$9A,$7C,$77
	.dc.b	$F0,$7E,$17,$C4,$FF,$8F,$27,$35,$C4,$E1,$5B,$9D,$3F,$36,$FA,$D7
	.dc.b	$0A,$5F,$4C,$73,$7B,$9E,$67,$92,$74,$E0,$FE,$35,$7C,$1D,$61,$E0
	.dc.b	$7C,$98,$77,$AE,$2F,$86,$3D,$F5,$C5,$F1,$D7,$17,$A9,$DE,$B8,$DF
	.dc.b	$96,$FA,$56,$EB,$99,$B7,$0A,$FB,$5F,$28,$F2,$DF,$C7,$C3,$C5,$1C
	.dc.b	$B0,$EB,$18,$8E,$47,$57,$17,$D3,$0E,$ED,$8C,$CE,$CF,$90,$75,$BF
	.dc.b	$03,$83,$AD,$F8,$1C,$57,$36,$9C,$E3,$E5,$3C,$27,$9C,$62,$78,$36
	.dc.b	$78,$4F,$C9,$36,$57,$EC,$D7,$A2,$7C,$47,$12,$CE,$E7,$EB,$CE,$E3
	.dc.b	$9C,$7C,$C3,$99,$E8,$9E,$2E,$4E,$26,$63,$8B,$F8,$08,$DB,$63,$2F
	.dc.b	$C9,$3A,$C7,$96,$33,$1F,$0C,$CB,$E7,$C3,$F0,$11,$F8,$CE,$AF,$EE
	.dc.b	$8F,$EA,$5A,$E7,$C2,$FD,$14,$63,$DF,$7A,$2E,$4F,$DB,$CE,$63,$16
	.dc.b	$8C,$4E,$AF,$FC,$01,$7C,$A3,$15,$FC,$02,$30,$FE,$B7,$E6,$DE,$77
	.dc.b	$F8,$1F,$4F,$DC,$46,$27,$F5,$15,$C2,$7F,$78,$DD,$3F,$C1,$D5,$BA
	.dc.b	$E9,$7C,$E7,$C7,$9F,$A3,$0E,$2D,$8E,$33,$BA,$FE,$2B,$7C,$EF,$78
	.dc.b	$F8,$B0,$E1,$5A,$BE,$0A,$F8,$61,$E0,$BE,$07,$F0,$18,$CC,$F1,$32
	.dc.b	$7F,$0F,$C7,$D8,$E4,$F5,$9A,$C7,$C6,$E0,$8E,$09,$7E,$C8,$F0,$61
	.dc.b	$E6,$AE,$B0,$6D,$F9,$23,$D5,$3B,$C7,$DF,$1B,$6C,$EC,$E1,$5D,$0E
	.dc.b	$75,$F1,$23,$33,$CA,$73,$5E,$43,$D9,$7F,$1D,$F5,$BE,$97,$E2,$4F
	.dc.b	$13,$C3,$7D,$4F,$36,$3C,$A7,$E4,$63,$F8,$A9,$CC,$72,$9F,$03,$D8
	.dc.b	$FC,$A5,$75,$8C,$D7,$19,$E8,$7C,$18,$CD,$F9,$67,$E7,$CF,$29,$CB
	.dc.b	$E2,$9F,$64,$EA,$3E,$69,$A3,$A1,$FD,$4B,$8B,$FB,$8F,$C8,$7F,$9A
	.dc.b	$6D,$BF,$F7,$B8,$C3,$6B,$FB,$77,$C8,$E1,$FE,$F5,$C6,$AC,$61,$D2
	.dc.b	$7F,$BA,$46,$25,$57,$E3,$71,$F1,$D2,$FF,$BD,$46,$5F,$EF,$91,$97
	.dc.b	$F8,$36,$3F,$E0,$F9,$F1,$43,$E6,$9E,$B6,$1B,$1F,$E1,$03,$51,$C3
	.dc.b	$E6,$FF,$86,$46,$67,$6F,$F8,$A5,$AF,$FA,$21,$AF,$FA,$3D,$AF,$F8
	.dc.b	$3B,$E6,$FD,$AF,$76,$BF,$E3,$A6,$ED,$FE,$98,$F9,$87,$C9,$FA,$A7
	.dc.b	$75,$FE,$96,$6B,$FE,$9D,$6B,$DF,$5F,$F1,$F3,$FE,$1E,$A7,$C7,$7C
	.dc.b	$45,$F1,$F4,$C4,$49,$F8,$93,$A6,$CE,$5F,$F2,$8C,$74,$72,$43,$33
	.dc.b	$E9,$23,$F0,$5F,$16,$B8,$8F,$F2,$A6,$1C,$DB,$3E,$A5,$19,$7F,$D7
	.dc.b	$9C,$4F,$35,$E8,$FE,$39,$7A,$74,$C4,$4A,$FF,$23,$CE,$9B,$39,$7D
	.dc.b	$BE,$0F,$C7,$F0,$A4,$7C,$48,$F9,$46,$1F,$87,$18,$74,$7C,$A7,$5C
	.dc.b	$1F,$76,$7F,$41,$F2,$5F,$A1,$F9,$39,$7F,$E5,$78,$CB,$FE,$59,$8C
	.dc.b	$BF,$9E,$3F,$5B,$1F,$85,$F9,$69,$1F,$8D,$7F,$59,$C3,$88,$F9,$B1
	.dc.b	$F2,$F6,$34,$FC,$3E,$DF,$E7,$0F,$9A,$FD,$A7,$04,$BF,$F3,$54,$7E
	.dc.b	$DC,$C3,$E3,$9F,$F3,$BC,$E6,$36,$7F,$9C,$5D,$E9,$FC,$2A,$3F,$82
	.dc.b	$D7,$F0,$7F,$69,$F3,$5F,$E9,$E9,$FF,$41,$C6,$3C,$1D,$4F,$F6,$DB
	.dc.b	$E5,$3C,$3C,$DC,$9A,$A9,$7C,$AF,$A3,$E7,$EE,$7C,$D7,$F1,$1E,$89
	.dc.b	$3F,$E8,$78,$FE,$28,$61,$E1,$18,$7F,$84,$7C,$D3,$DB,$FE,$1E,$FE
	.dc.b	$4A,$33,$D3,$89,$91,$82,$1F,$EB,$A9,$E4,$F4,$8E,$73,$23,$07,$FA
	.dc.b	$7B,$E5,$FF,$D7,$ED,$E7,$FE,$5B,$18,$8F,$77,$1E,$3B,$33,$E1,$C9
	.dc.b	$E5,$1F,$CC,$7C,$59,$1F,$EA,$77,$4F,$C7,$F9,$AF,$E6,$C4,$7F,$D5
	.dc.b	$9F,$35,$5F,$99,$E0,$FF,$3B,$31,$CD,$F7,$FA,$7F,$D7,$71,$99,$FF
	.dc.b	$5A,$3A,$F1,$7F,$D6,$72,$D5,$E1,$7B,$3C,$69,$87,$C1,$8F,$79,$87
	.dc.b	$F2,$52,$9C,$C6,$CF,$90,$EF,$1F,$24,$F6,$9F,$F6,$37,$57,$F5,$3F
	.dc.b	$3E,$46,$08,$7E,$25,$CB,$F8,$B8,$FE,$B4,$62,$BE,$07,$47,$FB,$2A
	.dc.b	$78,$C7,$83,$C0,$FA,$FE,$0C,$9F,$B9,$1F,$68,$C3,$EC,$8C,$3A,$3F
	.dc.b	$04,$EA,$7A,$47,$FB,$8B,$E4,$5F,$35,$FD,$B6,$21,$97,$FB,$9C,$6A
	.dc.b	$3F,$DC,$C6,$8F,$DE,$39,$6C,$66,$7D,$B1,$CA,$3F,$23,$E0,$FE,$4B
	.dc.b	$C3,$93,$F9,$78,$CB,$F9,$98,$CB,$F6,$23,$2F,$DA,$8C,$D7,$D6,$9F
	.dc.b	$3F,$CC,$FA,$3E,$6A,$5F,$BE,$32,$FC,$18,$CB,$EA,$8F,$7F,$CB,$7F
	.dc.b	$BA,$BE,$5B,$F1,$F9,$A3,$14,$E4,$4E,$63,$FD,$D3,$5C,$3E,$6B,$75
	.dc.b	$AB,$B7,$BB,$FF,$71,$5F,$69,$CD,$F6,$97,$CC,$BE,$23,$FC,$1F,$EC
	.dc.b	$30,$54,$AD,$7A,$FE,$28,$C5,$F0,$AD,$C3,$E6,$9F,$33,$DF,$0E,$33
	.dc.b	$E4,$38,$E1,$CD,$F1,$E1,$CD,$F0,$DF,$27,$55,$CB,$1F,$34,$72,$3A
	.dc.b	$1C,$AF,$A5,$71,$8C,$D7,$1C,$7C,$ED,$8E,$57,$D1,$F5,$63,$E9,$C7
	.dc.b	$C9,$1D,$4E,$6E,$30,$ED,$5C,$6F,$A5,$6E,$31,$5B,$BE,$98,$FB,$23
	.dc.b	$AD,$F0,$C3,$9E,$3E,$69,$FF,$21,$E3,$F3,$8F,$11,$6B,$E6,$B6,$36
	.dc.b	$FF,$B4,$BE,$49,$F1,$3F,$E3,$4F,$65,$18,$9E,$87,$FA,$8D,$C3,$DE
	.dc.b	$37,$3F,$E3,$CB,$FF,$0E,$F5,$5F,$4C,$D2,$5F,$49,$DB,$F4,$A7,$6F
	.dc.b	$D8,$F9,$AF,$D1,$78,$3F,$7A,$73,$C1,$F2,$DB,$AF,$D8,$6D,$FE,$5A
	.dc.b	$F8,$BD,$71,$E4,$47,$B7,$17,$F8,$07,$0E,$93,$96,$CE,$B1,$C6,$3D
	.dc.b	$EA,$CF,$7B,$E5,$39,$8F,$01,$2F,$F1,$58,$79,$1E,$A4,$9E,$0E,$AB
	.dc.b	$9D,$7E,$BF,$1F,$6F,$27,$F0,$2F,$9E,$CB,$FC,$13,$96,$88,$AE,$1B
	.dc.b	$A7,$ED,$E7,$93,$67,$F8,$6B,$7E,$6F,$FC,$49,$D5,$FD,$9E,$7F,$73
	.dc.b	$3F,$CE,$E7,$2F,$F1,$8E,$6B,$97,$F1,$7A,$56,$30,$FE,$33,$56,$73
	.dc.b	$7F,$D1,$63,$5E,$1B,$E1,$AF,$C6,$E1,$B5,$FC,$4B,$3F,$EF,$55,$B8
	.dc.b	$C1,$FD,$E1,$57,$18,$CE,$7F,$CA,$E7,$31,$8C,$BF,$BE,$E1,$DA,$BF
	.dc.b	$C9,$55,$D5,$B7,$E9,$F9,$8F,$57,$CC,$DC,$7F,$AF,$36,$BF,$AB,$B7
	.dc.b	$2F,$94,$D9,$FE,$7E,$D9,$F9,$0F,$C2,$F9,$AF,$87,$F1,$96,$BC,$D9
	.dc.b	$23,$17,$E6,$6E,$1F,$11,$B5,$F0,$DB,$3F,$05,$FA,$5F,$05,$F4,$FA
	.dc.b	$ED,$7F,$77,$6D,$7F,$70,$6D,$FE,$0D,$B3,$FD,$61,$B3,$F8,$17,$ED
	.dc.b	$FC,$CF,$DC,$F9,$AB,$5E,$A6,$D7,$CE,$6D,$FD,$C6,$CF,$DB,$6C,$FD
	.dc.b	$97,$F1,$5F,$31,$27,$FC,$1D,$EE,$DA,$FF,$03,$B6,$FF,$34,$D9,$FE
	.dc.b	$DE,$FE,$4C,$FC,$A7,$BB,$63,$FC,$31,$F2,$1F,$97,$3F,$2B,$F3,$56
	.dc.b	$3E,$9F,$C3,$61,$A7,$D6,$8F,$CD,$FC,$7C,$8F,$F1,$17,$7D,$CF,$FD
	.dc.b	$51,$F2,$3F,$CC,$B9,$FF,$52,$FF,$34,$F9,$4B,$E2,$FE,$5A,$B6,$A3
	.dc.b	$35,$A6,$E1,$F9,$EB,$E4,$DA,$E5,$BF,$FA,$7E,$73,$87,$4C,$39,$5F
	.dc.b	$F8,$DF,$E3,$3F,$6D,$F1,$FF,$B7,$6C,$64,$FF,$1E,$46,$5A,$7F,$B6
	.dc.b	$BE,$53,$F7,$DE,$67,$F0,$1F,$8A,$46,$27,$47,$F9,$16,$75,$19,$C7
	.dc.b	$FD,$75,$5E,$8A,$FF,$39,$BF,$05,$FF,$73,$E3,$D7,$5E,$CD,$8C,$9F
	.dc.b	$EE,$C8,$CB,$4C,$D7,$68,$FE,$32,$BE,$4F,$C0,$A5,$F3,$CF,$4F,$75
	.dc.b	$A9,$CC,$63,$0F,$F6,$65,$FF,$75,$AF,$B4,$D8,$CB,$FE,$26,$8C,$BE
	.dc.b	$EE,$DF,$91,$E1,$7C,$96,$E1,$FC,$5F,$3F,$E0,$4D,$9D,$34,$FC,$04
	.dc.b	$FE,$2A,$3F,$16,$D8,$C9,$F8,$58,$CB,$4C,$D7,$E4,$23,$2D,$8F,$E2
	.dc.b	$7E,$7F,$F3,$8F,$96,$FE,$72,$D8,$C9,$FC,$62,$32,$D3,$F9,$9C,$FC
	.dc.b	$FE,$6F,$E9,$3E,$73,$E5,$BF,$CA,$15,$AD,$15,$A9,$CD,$B9,$BE,$7D
	.dc.b	$6A,$FF,$C3,$33,$D5,$B1,$98,$FF,$0E,$B5,$FF,$2B,$F3,$7F,$89,$3A
	.dc.b	$BF,$63,$8F,$EC,$9B,$7F,$AB,$6D,$7E,$D1,$B8,$7E,$B7,$A1,$18,$AF
	.dc.b	$F3,$4E,$1D,$5F,$DA,$61,$F9,$0F,$96,$7C,$37,$F6,$AC,$5F,$3D,$F9
	.dc.b	$2C,$D5,$AC,$3F,$AD,$D6,$B9,$39,$61,$F8,$CB,$FF,$3B,$DE,$DF,$F1
	.dc.b	$DB,$5F,$F2,$43,$5F,$F3,$BF,$DE,$7F,$74,$7E,$66,$1F,$E8,$0F,$96
	.dc.b	$7E,$00,$F0,$7F,$4F,$7B,$CF,$F4,$D8,$2F,$75,$AF,$43,$C9,$2B,$EC
	.dc.b	$FF,$96,$5A,$F8,$A3,$D5,$E7,$3A,$9E,$3F,$F0,$BF,$33,$F8,$63,$75
	.dc.b	$FD,$23,$70,$FE,$12,$DD,$3F,$E0,$05,$BC,$5F,$E2,$87,$F1,$5F,$26
	.dc.b	$E1,$FA,$D8,$B8,$7F,$12,$9B,$7F,$37,$D4,$FF,$51,$FC,$B3,$E2,$3F
	.dc.b	$D3,$FC,$2F,$83,$F8,$73,$9C,$7F,$CE,$D7,$C2,$F7,$5A,$7F,$CD,$F1
	.dc.b	$99,$FA,$13,$97,$88,$AE,$8F,$CF,$8D,$3F,$EB,$09,$E4,$E7,$3A,$FF
	.dc.b	$AE,$A3,$38,$79,$E7,$32,$9C,$D7,$48,$E0,$78,$70,$F4,$D7,$0A,$DB
	.dc.b	$67,$47,$A9,$C6,$1E,$8C,$3C,$13,$A3,$D0,$57,$31,$EE,$8F,$0B,$C7
	.dc.b	$0F,$24,$79,$CE,$0E,$2F,$C7,$1D,$4E,$38,$7A,$1C,$1C,$6B,$A6,$1E
	.dc.b	$69,$D4,$78,$67,$51,$83,$EB,$BE,$09,$E0,$FB,$CE,$B5,$E2,$AE,$4D
	.dc.b	$AE,$31,$E0,$9D,$9C,$CA,$E4,$E2,$E5,$A6,$63,$99,$E5,$8F,$C3,$C6
	.dc.b	$5B,$19,$7E,$7C,$65,$B1,$97,$E3,$9E,$B8,$EF,$19,$8F,$86,$65,$F0
	.dc.b	$B4,$EE,$53,$A9,$4E,$A7,$E5,$4F,$83,$E3,$BC,$9F,$9A,$D8,$FD,$0B
	.dc.b	$5F,$C3,$B7,$DA,$7A,$1F,$D0,$E5,$39,$8C,$37,$54,$EA,$EA,$BE,$73
	.dc.b	$AA,$E1,$3F,$97,$F9,$77,$9D,$C3,$3F,$D9,$C6,$27,$F6,$38,$72,$7F
	.dc.b	$3C,$D9,$DD,$71,$CF,$C3,$7C,$6F,$79,$FA,$EF,$F7,$F8,$71,$BD,$E1
	.dc.b	$C6,$FD,$78,$71,$BD,$E1,$C6,$FD,$B8,$71,$BD,$E1,$C5,$61,$C6,$B7
	.dc.b	$7C,$67,$E3,$DF,$17,$77,$C6,$F8,$46,$AB,$83,$F2,$DB,$1A,$8E,$27
	.dc.b	$C6,$8D,$D7,$59,$D1,$B7,$C0,$66,$BC,$07,$A5,$CC,$F6,$8C,$D7,$5C
	.dc.b	$A3,$F6,$63,$2F,$CE,$8C,$E3,$F5,$23,$F9,$4E,$1E,$1B,$F2,$9D,$63
	.dc.b	$37,$E1,$8C,$DE,$26,$1D,$49,$7D,$5F,$11,$E0,$C7,$E7,$CF,$09,$D1
	.dc.b	$F6,$9D,$4F,$D8,$35,$3C,$1E,$AF,$E1,$1C,$9F,$4E,$38,$BF,$40,$F0
	.dc.b	$9C,$0D,$18,$9F,$0C,$F5,$7C,$51,$98,$EC,$F1,$E5,$FE,$ED,$E0,$7E
	.dc.b	$AA,$FF,$B7,$B7,$7F,$F3,$CB,$69,$5F,$DF,$7B,$F0,$F2,$8F,$F2,$F4
	.dc.b	$E3,$97,$FC,$57,$1F,$E6,$1A,$C7,$CE,$E7,$E7,$A4,$FF,$99,$63,$2F
	.dc.b	$F9,$9E,$32,$FF,$A3,$23,$C4,$BD,$A7,$BF,$FE,$48,$8C,$36,$33,$1F
	.dc.b	$E4,$C6,$BF,$C2,$0F,$E1,$6E,$34,$FF,$40,$D3,$E0,$3B,$B6,$BF,$CC
	.dc.b	$35,$BD,$7F,$88,$FC,$C7,$BA,$7A,$07,$0F,$76,$BC,$35,$F4,$47,$1C
	.dc.b	$7F,$D2,$3F,$32,$F5,$3D,$ED,$BF,$F3,$8D,$6F,$0F,$E4,$32,$F7,$F7
	.dc.b	$AC,$BF,$F3,$BC,$E5,$B3,$97,$FC,$8E,$7F,$2C,$8F,$63,$63,$C2,$6F
	.dc.b	$1F,$2D,$62,$7F,$D0,$7F,$30,$F8,$FF,$A4,$D8,$FF,$3D,$B5,$F8,$AD
	.dc.b	$7E,$93,$9A,$FF,$58,$FC,$AB,$E3,$FE,$FB,$6B,$BD,$FE,$43,$F0,$47
	.dc.b	$CC,$DF,$62,$57,$F4,$6C,$7F,$CB,$BD,$0B,$DC,$E7,$7F,$E9,$73,$9C
	.dc.b	$38,$CF,$C4,$8E,$18,$77,$9F,$1B,$E4,$F8,$A9,$5F,$EA,$4C,$FF,$D9
	.dc.b	$CD,$9C,$BF,$38,$42,$79,$C6,$CE,$4F,$0C,$BE,$1D,$7E,$AF,$E7,$FF
	.dc.b	$59,$F3,$F2,$FF,$D9,$91,$97,$FD,$9D,$19,$7F,$DE,$71,$FA,$8D,$FC
	.dc.b	$B1,$F4,$AB,$EB,$91,$FA,$EF,$D0,$F9,$59,$1F,$7A,$3F,$79,$ED,$7E
	.dc.b	$F7,$DA,$93,$FE,$DD,$8C,$BF,$ED,$F8,$CB,$F2,$FB,$93,$A8,$C4,$EB
	.dc.b	$3F,$F3,$FC,$FF,$09,$F9,$7F,$E1,$5D,$72,$FF,$DD,$51,$97,$FD,$D9
	.dc.b	$19,$7B,$47,$F1,$1E,$2F,$E2,$5F,$31,$27,$FD,$E5,$19,$7F,$DE,$91
	.dc.b	$97,$CD,$1F,$90,$F5,$7F,$23,$F3,$92,$7E,$F4,$65,$FC,$04,$65,$FB
	.dc.b	$DC,$7F,$EC,$47,$66,$88,$6A,$7F,$35,$D1,$23,$F2,$87,$F2,$EE,$7F
	.dc.b	$D0,$3F,$93,$F9,$57,$CE,$FE,$87,$E6,$7F,$45,$F3,$32,$B8,$F5,$FE
	.dc.b	$7E,$7F,$9A,$DB,$4A,$4A,$DE,$2C,$37,$D9,$27,$9D,$FE,$D7,$6F,$D3
	.dc.b	$BE,$D8,$F2,$4F,$6E,$79,$1C,$22,$BE,$98,$BC,$9F,$B4,$AE,$93,$EF
	.dc.b	$F8,$1F,$DE,$7A,$32,$3C,$B1,$FB,$F3,$0F,$ED,$73,$FF,$5C,$3A,$21
	.dc.b	$99,$DD,$FF,$02,$95,$7C,$4F,$8B,$F8,$BF,$33,$30,$F7,$C6,$5F,$83
	.dc.b	$19,$7E,$94,$65,$FA,$91,$9B,$F6,$4F,$F6,$4F,$A8,$7C,$BF,$F1,$7B
	.dc.b	$87,$D6,$8C,$B6,$32,$FB,$4E,$4C,$7E,$B9,$AA,$FE,$3B,$3C,$A3,$F0
	.dc.b	$DF,$33,$F8,$0F,$99,$91,$F6,$A3,$F0,$66,$1F,$C6,$46,$1D,$1F,$80
	.dc.b	$75,$5F,$5E,$7D,$9E,$67,$B7,$E6,$65,$FA,$E3,$2F,$B6,$32,$FF,$2B
	.dc.b	$C3,$FB,$53,$67,$37,$EA,$AE,$33,$F9,$FF,$07,$F3,$3E,$2C,$8F,$C9
	.dc.b	$C7,$E7,$0C,$3F,$A5,$8C,$3F,$98,$EC,$FF,$76,$7C,$77,$E6,$3C,$7B
	.dc.b	$3F,$E1,$AE,$4F,$E8,$6F,$FB,$A6,$7F,$3B,$5A,$EB,$57,$F3,$8A,$FF
	.dc.b	$79,$57,$36,$9C,$EF,$84,$EE,$3A,$1F,$E9,$8F,$98,$FE,$43,$EC,$70
	.dc.b	$84,$61,$D9,$8F,$13,$FD,$46,$E3,$E6,$5F,$2D,$EA,$8D,$5F,$3B,$E5
	.dc.b	$EE,$33,$FF,$52,$C6,$1F,$F5,$3E,$1C,$AF,$86,$3D,$AF,$84,$79,$E3
	.dc.b	$15,$C9,$B1,$FE,$B2,$AE,$4F,$A2,$FB,$3C,$70,$EA,$7B,$E7,$78,$76
	.dc.b	$9F,$49,$C6,$FA,$B8,$8D,$4F,$0C,$3C,$57,$D2,$B8,$3D,$71,$F8,$98
	.dc.b	$F8,$A7,$95,$70,$7E,$54,$EB,$1F,$9B,$1F,$E4,$3A,$FA,$4D,$8C,$BC
	.dc.b	$4F,$F2,$47,$B6,$F9,$47,$C5,$FF,$90,$27,$FD,$A5,$18,$7F,$C8,$73
	.dc.b	$B9,$D3,$63,$FC,$8F,$7F,$E5,$BE,$CE,$4D,$8C,$F9,$5F,$C6,$1B,$3A
	.dc.b	$8C,$6D,$F2,$EB,$83,$F3,$7E,$67,$FC,$89,$E1,$FE,$16,$B7,$B7,$E1
	.dc.b	$1B,$8F,$E5,$5A,$AB,$8B,$F7,$63,$38,$76,$AE,$58,$FE,$6E,$F7,$8F
	.dc.b	$8E,$F6,$49,$D7,$A4,$CB,$C9,$13,$2E,$CE,$A3,$17,$CE,$2E,$3F,$F0
	.dc.b	$03,$1E,$EF,$8A,$F8,$92,$BE,$16,$1D,$E3,$13,$FE,$40,$F9,$87,$CC
	.dc.b	$E6,$FF,$43,$CC,$99,$9F,$EF,$A3,$15,$BB,$E5,$57,$0F,$F5,$2B,$D2
	.dc.b	$F8,$3C,$A3,$F8,$1D,$F2,$9C,$CF,$18,$F0,$7C,$3F,$9F,$B3,$FC,$B3
	.dc.b	$3F,$C0,$2B,$FB,$EA,$9D,$47,$AF,$92,$E1,$E2,$9B,$87,$AE,$6D,$F1
	.dc.b	$BE,$71,$AF,$09,$F2,$37,$E3,$70,$DA,$FF,$06,$CE,$9B,$3A,$69,$A9
	.dc.b	$E1,$97,$F9,$AE,$B8,$36,$BC,$7F,$33,$E4,$F9,$9B,$87,$B9,$B5,$FE
	.dc.b	$73,$C9,$18,$D7,$E2,$B6,$BE,$E3,$6B,$E9,$BE,$DF,$8B,$F7,$7C,$A5
	.dc.b	$AF,$5B,$6B,$CC,$DB,$FF,$44,$E6,$8C,$57,$FA,$22,$30,$FC,$4B,$EA
	.dc.b	$FD,$6F,$85,$F9,$FD,$7F,$41,$B1,$93,$ED,$C6,$5A,$67,$1F,$81,$19
	.dc.b	$6C,$7D,$2F,$8A,$FB,$1F,$19,$F6,$5B,$19,$3E,$BC,$65,$A6,$67,$EA
	.dc.b	$46,$5B,$1F,$8E,$F9,$9F,$C3,$9F,$88,$F9,$9B,$1F,$92,$6C,$FD,$56
	.dc.b	$EB,$F9,$87,$84,$FE,$63,$E2,$AD,$7D,$BF,$90,$FC,$99,$F9,$1F,$8F
	.dc.b	$B5,$F8,$EF,$87,$C3,$4F,$C8,$C7,$E5,$FE,$2A,$47,$E9,$3A,$EE,$1F
	.dc.b	$ED,$CF,$94,$FE,$6B,$D1,$FA,$47,$F9,$BF,$CC,$5F,$19,$36,$30,$DC
	.dc.b	$3F,$CE,$75,$C9,$B4,$AD,$C6,$1B,$7E,$19,$D4,$FE,$E3,$E4,$3F,$63
	.dc.b	$F1,$7F,$B2,$6C,$64,$FD,$EC,$65,$A7,$F9,$F3,$E1,$5F,$33,$FB,$AF
	.dc.b	$33,$F7,$B1,$89,$FE,$0B,$18,$9D,$47,$16,$E3,$FE,$B3,$D3,$8E,$7C
	.dc.b	$9B,$5D,$DB,$3B,$8F,$14,$ED,$A6,$CB,$1F,$ED,$FF,$97,$F9,$B1,$FD
	.dc.b	$95,$B0,$21,$5C,$B8,$0E,$1D,$BC,$B9,$7F,$9E,$46,$AF,$F9,$0F,$92
	.dc.b	$F9,$87,$2F,$F2,$4A,$D6,$3F,$2F,$BF,$E4,$F8,$7F,$29,$B1,$93,$E8
	.dc.b	$46,$5A,$7E,$0A,$3F,$15,$1F,$8B,$6C,$64,$FC,$3C,$65,$A6,$67,$F0
	.dc.b	$D1,$96,$C7,$F2,$EF,$3B,$F9,$A7,$C4,$7F,$35,$6C,$64,$FE,$67,$19
	.dc.b	$69,$FE,$CA,$9F,$93,$CF,$F1,$9F,$92,$D7,$E8,$35,$DC,$7D,$78,$F9
	.dc.b	$1B,$1F,$E9,$BF,$28,$F9,$F7,$6E,$A7,$F5,$9D,$9F,$AD,$6E,$1F,$A8
	.dc.b	$6E,$3F,$A5,$6F,$17,$F7,$5E,$AF,$F1,$5C,$62,$BF,$0D,$18,$7F,$63
	.dc.b	$7D,$5F,$77,$CC,$3E,$27,$D4,$D8,$FF,$1A,$37,$1F,$C6,$07,$0F,$E3
	.dc.b	$DD,$63,$F8,$C6,$CF,$E5,$F4,$F8,$1D,$5F,$8B,$9C,$DF,$E6,$FE,$5F
	.dc.b	$FB,$63,$D9,$AF,$E6,$DA,$FE,$81,$AF,$1F,$37,$F2,$D1,$CB,$1F,$F0
	.dc.b	$C7,$DF,$BE,$FD,$F2,$3A,$BF,$E0,$3E,$37,$F0,$2F,$2A,$ED,$FC,$1D
	.dc.b	$B9,$B2,$FF,$11,$7A,$0F,$09,$5A,$9C,$C6,$2F,$F8,$87,$8D,$FC,$38
	.dc.b	$FE,$1F,$E4,$D8,$C9,$31,$EC,$FE,$F7,$1E,$D3,$B9,$C9,$FB,$DF,$4B
	.dc.b	$F8,$17,$CC,$3E,$63,$F8,$1F,$2B,$E2,$3F,$C7,$13,$9B,$E0,$1C,$29
	.dc.b	$E1,$97,$4B,$FB,$F8,$5A,$FB,$58,$74,$9C,$C7,$37,$C3,$19,$8E,$F8
	.dc.b	$76,$C3,$AC,$6C,$CD,$70,$C3,$B4,$7F,$1C,$C2,$D7,$0B,$F3,$CF,$F2
	.dc.b	$19,$CE,$1D,$E3,$81,$D4,$F6,$39,$78,$19,$AE,$98,$7A,$AF,$CD,$5B
	.dc.b	$A5,$F6,$C3,$CB,$1C,$AB,$71,$88,$D9,$D6,$B9,$57,$18,$F2,$1C,$6F
	.dc.b	$E0,$38,$38,$DF,$47,$13,$C6,$3E,$0C,$6A,$78,$C6,$27,$51,$EB,$32
	.dc.b	$59,$FB,$37,$E7,$9D,$C7,$D2,$32,$78,$0F,$54,$6D,$A6,$6B,$CC,$78
	.dc.b	$5C,$B4,$CD,$78,$63,$E9,$CF,$39,$CC,$78,$4C,$9D,$67,$D2,$E7,$0F
	.dc.b	$AE,$DC,$3E,$6B,$E0,$9E,$D3,$A8,$FC,$1D,$76,$6C,$F4,$69,$F0,$23
	.dc.b	$D5,$E4,$3C,$9F,$CE,$B6,$7F,$52,$D7,$F1,$ED,$F6,$9E,$87,$F0,$79
	.dc.b	$4E,$63,$0D,$E1,$FE,$F4,$D9,$DD,$FF,$7C,$BE,$75,$FA,$4F,$99,$79
	.dc.b	$FF,$E0,$0C,$BF,$73,$18,$C3,$F5,$37,$C1,$FD,$13,$67,$75,$C7,$3F
	.dc.b	$F0,$4D,$F1,$BD,$E7,$E9,$BF,$E0,$B8,$71,$BD,$E1,$C6,$FF,$C2,$58
	.dc.b	$71,$BD,$E1,$C6,$FD,$4D,$AD,$DF,$15,$7C,$6B,$78,$FA,$AF,$85,$6A
	.dc.b	$F8,$5E,$E3,$FC,$45,$8F,$95,$B5,$C2,$75,$1C,$4E,$97,$1F,$E0,$D1
	.dc.b	$A3,$75,$CE,$0D,$8D,$5F,$A6,$34,$65,$F2,$99,$BF,$4C,$66,$BE,$D4
	.dc.b	$E6,$FC,$58,$FD,$49,$C9,$8C,$7E,$1B,$E6,$AE,$AE,$4F,$7D,$70,$74
	.dc.b	$64,$F4,$E1,$F0,$E7,$8C,$7D,$E9,$D6,$3E,$C8,$E1,$8F,$9D,$E6,$6C
	.dc.b	$98,$F9,$DC,$BF,$80,$AE,$77,$DA,$31,$3C,$31,$FC,$4B,$E7,$8C,$92
	.dc.b	$33,$1D,$1F,$83,$1F,$D6,$78,$7F,$B8,$77,$7F,$98,$30,$FE,$DE,$DE
	.dc.b	$0F,$ED,$1C,$88,$C6,$2B,$FE,$F1,$E0,$61,$E7,$1F,$E8,$59,$C5,$71
	.dc.b	$8F,$F4,$45,$63,$E6,$33,$F3,$32,$7F,$D1,$91,$97,$FD,$1D,$19,$7F
	.dc.b	$BB,$E1,$FE,$07,$9F,$1C,$3D,$C7,$CB,$E1,$B1,$FE,$51,$6B,$FE,$00
	.dc.b	$12,$F7,$38,$57,$F8,$D2,$7F,$D0,$95,$C1,$FF,$41,$35,$FF,$43,$B5
	.dc.b	$FF,$45,$7C,$CF,$BD,$EC,$DD,$BF,$D0,$AD,$DB,$FD,$1F,$F3,$3F,$C7
	.dc.b	$2B,$FC,$CD,$3C,$5F,$F4,$83,$5F,$F4,$AB,$5E,$FB,$F9,$D4,$FA,$DF
	.dc.b	$F3,$83,$F0,$2F,$FD,$2A,$DB,$F0,$E5,$F2,$A3,$85,$7C,$57,$4F,$FA
	.dc.b	$95,$AF,$68,$E0,$71,$CA,$47,$82,$3E,$AF,$CC,$3E,$46,$92,$7F,$D0
	.dc.b	$9F,$0C,$F9,$2C,$47,$F9,$92,$B8,$3F,$4D,$AE,$6B,$EB,$70,$3E,$45
	.dc.b	$F1,$1F,$93,$5F,$15,$4C,$44,$BF,$F6,$06,$FF,$E5,$2A,$DC,$63,$7F
	.dc.b	$F2,$44,$62,$3F,$22,$F6,$7D,$37,$F9,$2D,$1F,$15,$4C,$44,$9F,$7C
	.dc.b	$65,$E2,$43,$F2,$93,$C9,$D9,$9C,$3F,$CC,$7F,$33,$FA,$CF,$26,$4F
	.dc.b	$FB,$CE,$3F,$5E,$61,$FD,$3F,$4F,$F9,$D9,$D1,$FE,$B3,$DF,$D9,$7F
	.dc.b	$ED,$86,$E9,$FE,$E2,$6C,$FF,$B9,$71,$7C,$17,$DC,$D3,$F2,$15,$F1
	.dc.b	$6B,$FD,$CC,$D7,$94,$7D,$6C,$5E,$B7,$F0,$4F,$5A,$57,$CF,$3F,$86
	.dc.b	$4E,$9C,$70,$FF,$A1,$6B,$71,$EC,$AF,$E1,$B1,$23,$74,$9D,$E9,$FE
	.dc.b	$8A,$AF,$E2,$7F,$33,$FC,$53,$AA,$61,$C2,$32,$F2,$8C,$BF,$2A,$3F
	.dc.b	$8D,$ED,$FC,$73,$E6,$64,$F8,$23,$2F,$8A,$32,$FD,$28,$FC,$37,$A9
	.dc.b	$F8,$7F,$96,$93,$FA,$28,$CB,$FA,$48,$CB,$F8,$48,$CB,$F8,$68,$CE
	.dc.b	$4A,$4F,$D7,$AF,$D1,$7C,$CC,$8F,$97,$CC,$8F,$E7,$1F,$1E,$F9,$9F
	.dc.b	$D4,$7C,$CF,$EA,$7B,$26,$1C,$23,$7A,$48,$FA,$B6,$8D,$C7,$F3,$68
	.dc.b	$FD,$BE,$9F,$EB,$E6,$E9,$FE,$B3,$6C,$FF,$4B,$7F,$DD,$0F,$4E,$CF
	.dc.b	$2F,$CC,$C8,$C1,$E0,$70,$F7,$7D,$71,$FC,$07,$E0,$FF,$81,$7A,$12
	.dc.b	$BE,$0C,$65,$F8,$71,$97,$F7,$B1,$97,$F7,$F1,$9C,$7F,$83,$4F,$C8
	.dc.b	$EA,$F9,$3F,$33,$2F,$E1,$C6,$5F,$8B,$19,$7E,$24,$65,$F8,$D1,$9A
	.dc.b	$F8,$D3,$FD,$AF,$E5,$9F,$33,$FC,$6E,$E1,$FA,$48,$CB,$63,$2F,$F1
	.dc.b	$63,$93,$1F,$C0,$9A,$AF,$F7,$8C,$F2,$8F,$B1,$F3,$1F,$84,$F9,$99
	.dc.b	$1F,$7A,$3F,$0E,$61,$FB,$D1,$87,$47,$E1,$1D,$57,$E0,$67,$D9,$F3
	.dc.b	$3E,$DF,$99,$97,$EB,$8C,$BE,$D8,$CB,$FE,$00,$C1,$3A,$8C,$4E,$AF
	.dc.b	$D7,$3F,$9C,$F0,$3F,$3B,$E2,$CB,$FC,$BC,$65,$FC,$CC,$65,$FC,$C4
	.dc.b	$7E,$93,$C3,$FF,$74,$7C,$7F,$E5,$FC,$7B,$5F,$97,$6F,$88,$E6,$5E
	.dc.b	$1B,$5F,$EE,$DB,$E4,$D3,$91,$6B,$95,$76,$87,$CC,$7F,$28,$F6,$38
	.dc.b	$46,$3C,$4F,$F0,$47,$B4,$C7,$F8,$0C,$E6,$B7,$F3,$2F,$97,$E9,$8F
	.dc.b	$3A,$FF,$79,$DF,$08,$C6,$1C,$E7,$F8,$5C,$62,$7A,$DF,$39,$F1,$CE
	.dc.b	$5D,$E3,$E4,$AE,$8D,$8F,$E2,$B5,$C5,$F0,$9C,$31,$E7,$3B,$8F,$51
	.dc.b	$A1,$3C,$70,$F3,$46,$9B,$87,$8E,$31,$87,$A3,$1F,$83,$87,$5A,$D1
	.dc.b	$BC,$3A,$57,$A3,$1F,$19,$30,$F1,$4E,$9F,$41,$F4,$0A,$5A,$F9,$CD
	.dc.b	$BF,$F1,$F3,$BF,$61,$F2,$8F,$93,$FF,$28,$46,$1B,$19,$9E,$C7,$FA
	.dc.b	$A9,$C3,$DF,$CA,$76,$FF,$96,$71,$FF,$32,$36,$33,$3C,$A7,$2D,$8C
	.dc.b	$ED,$F2,$9B,$AF,$E3,$3E,$67,$FC,$CD,$E1,$7E,$02,$73,$7F,$86,$6B
	.dc.b	$F9,$86,$CF,$08,$FB,$58,$78,$30,$ED,$8F,$33,$F3,$B5,$B8,$CD,$70
	.dc.b	$BE,$B7,$C7,$1F,$27,$A3,$FD,$86,$B5,$8F,$36,$9F,$89,$C7,$A5,$7E
	.dc.b	$7F,$0E,$64,$AF,$0D,$F3,$C7,$C2,$4A,$E1,$3A,$CB,$B4,$FE,$9D,$7E
	.dc.b	$0B,$FE,$11,$7F,$7C,$63,$CE,$E9,$7F,$D8,$7C,$3B,$A7,$F7,$9B,$E0
	.dc.b	$FE,$92,$F8,$1C,$A3,$38,$7E,$D7,$9B,$F8,$85,$F2,$AD,$DF,$E9,$FF
	.dc.b	$06,$7C,$BF,$C5,$EE,$FE,$E5,$F4,$56,$FF,$83,$D6,$9F,$EF,$71,$9B
	.dc.b	$FF,$05,$CE,$5F,$C8,$B5,$FD,$03,$5E,$77,$F4,$DF,$D2,$70,$BA,$3A
	.dc.b	$72,$BE,$79,$E3,$F7,$B5,$EF,$3F,$F2,$25,$EE,$B5,$39,$A6,$9F,$81
	.dc.b	$AD,$5F,$F9,$2E,$79,$E3,$E0,$C7,$BC,$E7,$0E,$58,$79,$DB,$1F,$E2
	.dc.b	$AC,$3A,$F9,$1E,$0C,$3C,$DF,$76,$F7,$BD,$7F,$07,$F1,$72,$F8,$13
	.dc.b	$9C,$7D,$33,$97,$F9,$AC,$67,$4F,$4C,$77,$9D,$47,$F8,$EA,$75,$5E
	.dc.b	$D9,$CC,$A7,$35,$D2,$FE,$6E,$5F,$C3,$23,$E1,$4E,$AB,$A7,$B9,$E8
	.dc.b	$F9,$6F,$5B,$6B,$6D,$9F,$F7,$35,$7F,$C0,$19,$F5,$D7,$07,$67,$D6
	.dc.b	$B7,$C1,$E6,$FF,$DD,$DF,$6D,$FE,$A4,$AF,$B7,$F4,$3E,$A8,$C4,$FC
	.dc.b	$23,$F9,$14,$67,$C9,$FC,$3B,$6B,$8F,$27,$E1,$EF,$8E,$FF,$8F,$6C
	.dc.b	$FF,$62,$FC,$1D,$F2,$BB,$C7,$87,$97,$D3,$6B,$7E,$46,$E9,$E3,$D9
	.dc.b	$AA,$73,$A7,$E3,$E3,$17,$F8,$FD,$3F,$D9,$ED,$C7,$F3,$9A,$3E,$53
	.dc.b	$8C,$FE,$37,$EE,$9E,$E7,$E9,$1B,$1C,$FD,$8C,$4F,$BD,$B1,$93,$F4
	.dc.b	$B3,$C3,$6F,$94,$7F,$9E,$EB,$84,$FE,$8F,$DA,$D6,$BF,$1F,$0F,$DF
	.dc.b	$78,$1F,$E1,$9F,$16,$F9,$FF,$D2,$B0,$EB,$EA,$E2,$7E,$87,$B5,$E7
	.dc.b	$8B,$3A,$3B,$3F,$BB,$C8,$F4,$D3,$0F,$EF,$FE,$61,$F3,$9F,$C3,$EF
	.dc.b	$E7,$F1,$66,$31,$C9,$F3,$30,$F6,$DF,$0B,$FF,$64,$1C,$0F,$E0,$0E
	.dc.b	$63,$F1,$91,$97,$8D,$7F,$15,$7F,$C4,$F3,$B7,$FB,$D3,$5C,$C6,$D6
	.dc.b	$AF,$C2,$1E,$BF,$CA,$7F,$CE,$F8,$7F,$B6,$1C,$B7,$DA,$D4,$ED,$B1
	.dc.b	$93,$87,$C3,$73,$F6,$7A,$36,$32,$75,$8C,$B4,$FF,$75,$7C,$CF,$F8
	.dc.b	$2F,$E7,$3F,$9D,$3F,$E0,$DF,$9C,$BD,$C9,$B1,$86,$D7,$29,$D3,$67
	.dc.b	$4D,$35,$3C,$35,$FD,$D7,$C9,$3C,$5F,$83,$1F,$EC,$3E,$9F,$E9,$93
	.dc.b	$9D,$BF,$1F,$18,$AD,$D7,$F9,$6A,$B5,$87,$F8,$81,$B1,$FE,$2F,$6B
	.dc.b	$FE,$60,$E0,$FA,$9F,$47,$F4,$FE,$8B,$EA,$36,$FE,$A3,$6B,$F8,$85
	.dc.b	$70,$9F,$B1,$5F,$95,$F2,$3F,$B1,$75,$BE,$7F,$FC,$DD,$5F,$8E,$9C
	.dc.b	$C6,$30,$FC,$AC,$63,$D5,$E5,$5F,$EF,$4E,$FF,$F0,$B6,$1C,$E7,$37
	.dc.b	$FE,$3C,$E9,$FE,$88,$FE,$4B,$0F,$C8,$28,$CB,$FE,$FE,$CD,$F3,$3F
	.dc.b	$F0,$03,$2E,$C6,$5F,$DD,$4F,$B9,$D3,$71,$FE,$03,$18,$CB,$FC,$55
	.dc.b	$7D,$27,$35,$A7,$F7,$1E,$EF,$C0,$F1,$2E,$3E,$0A,$FF,$28,$EB,$FC
	.dc.b	$7E,$3F,$67,$5E,$AC,$CD,$24,$7F,$97,$29,$F3,$DF,$C5,$A3,$E4,$7A
	.dc.b	$5F,$BB,$D3,$FC,$19,$39,$AF,$F0,$0C,$E6,$BF,$D4,$33,$9A,$FF,$07
	.dc.b	$77,$7F,$83,$63,$15,$BB,$1E,$0F,$17,$E7,$CF,$F2,$B8,$FF,$3E,$35
	.dc.b	$EB,$1A,$F8,$8F,$C2,$FE,$02,$FA,$1E,$D1,$86,$E7,$ED,$C1,$3C,$63
	.dc.b	$0F,$DE,$DF,$FC,$83,$A7,$E1,$B0,$7C,$AF,$F8,$96,$73,$7E,$8E,$15
	.dc.b	$6A,$3F,$C4,$13,$FE,$4A,$D3,$FD,$28,$D7,$FD,$8A,$D7,$E0,$B7,$A3
	.dc.b	$F7,$1C,$FF,$D0,$BE,$9F,$F6,$8D,$BF,$BD,$3A,$7F,$69,$D4,$9F,$E4
	.dc.b	$D3,$9C,$3F,$6D,$39,$AF,$F4,$CF,$37,$F2,$9F,$9E,$FE,$50,$7F,$29
	.dc.b	$F8,$6B,$3F,$CA,$62,$DF,$ED,$BC,$6F,$F6,$D4,$E4,$FF,$1F,$47,$DA
	.dc.b	$F7,$BE,$DF,$CC,$3E,$67,$F8,$57,$C0,$7F,$A1,$63,$15,$FE,$B7,$AD
	.dc.b	$D6,$AE,$CE,$47,$CD,$F0,$5B,$31,$FF,$48,$4E,$67,$C4,$70,$BE,$8D
	.dc.b	$3A,$63,$CE,$F9,$61,$E0,$9F,$F7,$CE,$36,$73,$0A,$EC,$A7,$F3,$38
	.dc.b	$73,$C7,$C7,$8F,$FB,$1A,$72,$F8,$A7,$8C,$75,$9F,$89,$3B,$AE,$CF
	.dc.b	$A2,$F7,$18,$3F,$4F,$87,$92,$32,$78,$71,$F5,$4F,$C2,$8C,$3C,$4F
	.dc.b	$69,$67,$95,$78,$71,$D9,$E0,$2B,$E5,$AE,$67,$81,$F3,$19,$2D,$7C
	.dc.b	$E6,$BE,$09,$E6,$FD,$D6,$BE,$49,$EB,$3C,$E3,$DF,$87,$C6,$7A,$18
	.dc.b	$C3,$DE,$FD,$33,$D9,$1B,$7A,$4F,$89,$F8,$8D,$78,$57,$81,$EA,$7E
	.dc.b	$3A,$79,$3C,$0F,$01,$F2,$CC,$1B,$AF,$1E,$3F,$75,$B1,$E4,$6D,$77
	.dc.b	$8F,$29,$D1,$F9,$8D,$BF,$84,$F5,$3C,$07,$A1,$F5,$57,$78,$CB,$FC
	.dc.b	$B3,$DC,$FD,$4C,$7E,$AD,$AF,$E0,$9A,$FE,$A1,$CF,$C4,$3D,$DF,$F5
	.dc.b	$EC,$FF,$B2,$73,$FE,$6E,$A7,$5D,$1D,$E7,$31,$89,$E0,$7F,$3E,$71
	.dc.b	$7D,$AF,$AE,$5F,$BD,$F9,$BF,$F4,$CD,$FE,$CD,$AF,$E6,$6F,$A4,$FF
	.dc.b	$03,$6E,$7E,$4A,$DC,$F3,$F7,$7C,$EE,$F0,$EF,$5E,$66,$E1,$FD,$4A
	.dc.b	$F8,$CF,$3B,$E2,$EF,$3F,$E0,$D8,$F8,$2F,$85,$69,$A6,$B0,$EA,$FC
	.dc.b	$76,$E1,$DA,$36,$7B,$63,$75,$D6,$13,$B6,$D7,$77,$A5,$F7,$3E,$44
	.dc.b	$66,$78,$57,$33,$A3,$CC,$CD,$F8,$0F,$2C,$F1,$3B,$4F,$18,$FC,$1E
	.dc.b	$7F,$37,$0F,$E3,$A7,$A9,$F9,$D8,$62,$3C,$A7,$F1,$1C,$3C,$18,$74
	.dc.b	$BF,$05,$7A,$2A,$1F,$0D,$F6,$3F,$CE,$4F,$82,$75,$AF,$C2,$DF,$82
	.dc.b	$35,$87,$BB,$0F,$54,$72,$CB,$47,$A5,$F2,$47,$8F,$0D,$65,$F8,$19
	.dc.b	$F1,$D7,$08,$FC,$14,$78,$F1,$F0,$C6,$E3,$C3,$97,$C6,$8D,$D7,$72
	.dc.b	$38,$39,$E5,$E3,$23,$83,$55,$F8,$69,$E2,$FE,$AE,$BF,$33,$C7,$FE
	.dc.b	$A3,$FA,$5F,$DA,$E5,$FE,$4A,$E2,$46,$38,$3F,$CB,$11,$8C,$3F,$D4
	.dc.b	$DF,$4B,$97,$FD,$5B,$19,$7F,$C0,$15,$9C,$3B,$47,$03,$FC,$27,$F8
	.dc.b	$4B,$E9,$BF,$C9,$5C,$A8,$C4,$FF,$A2,$5A,$FF,$A2,$5A,$FF,$A3,$DC
	.dc.b	$E7,$FE,$62,$F6,$BF,$C0,$0D,$DF,$CD,$7D,$70,$F4,$60,$FC,$29,$F3
	.dc.b	$3F,$E5,$B8,$C3,$63,$31,$FE,$5F,$6B,$FC,$70,$FF,$29,$B8,$C1,$BF
	.dc.b	$F2,$0F,$C2,$DF,$2B,$E0,$6D,$FF,$9C,$4E,$E5,$F8,$F7,$A1,$E3,$6C
	.dc.b	$7F,$9B,$DA,$F8,$43,$87,$C8,$EB,$0F,$1C,$7F,$9A,$3F,$0C,$7B,$FF
	.dc.b	$09,$B1,$E1,$BF,$8B,$F4,$3F,$EC,$38,$FF,$3F,$35,$F0,$35,$F8,$4D
	.dc.b	$7F,$9F,$BB,$BF,$0C,$3E,$85,$E3,$7D,$26,$C7,$C3,$AE,$1A,$F2,$F1
	.dc.b	$35,$72,$B5,$1E,$6D,$3E,$C5,$FC,$29,$F3,$C6,$08,$FC,$FC,$1E,$AF
	.dc.b	$84,$E5,$1F,$EA,$3E,$77,$DF,$3A,$FF,$34,$BC,$5A,$62,$25,$71,$9D
	.dc.b	$36,$72,$B6,$F8,$7F,$40,$F8,$DF,$C8,$36,$7F,$D5,$8D,$7E,$F3,$5F
	.dc.b	$C8,$39,$9F,$9B,$F3,$1F,$E9,$69,$FC,$63,$5F,$07,$0F,$F7,$2A,$E0
	.dc.b	$DB,$FD,$1D,$6E,$F9,$F9,$FF,$A6,$BD,$E1,$FB,$EF,$07,$F7,$9E,$0B
	.dc.b	$E5,$E9,$88,$93,$F2,$E7,$4D,$9C,$BE,$43,$FC,$EB,$E0,$BD,$DF,$E0
	.dc.b	$6D,$8F,$E0,$51,$82,$3F,$5F,$C2,$7C,$3F,$F0,$F6,$C7,$F0,$E9,$F6
	.dc.b	$FC,$37,$BB,$E3,$24,$FC,$68,$CB,$F2,$23,$2F,$94,$6E,$9D,$C6,$61
	.dc.b	$3C,$37,$FA,$75,$F1,$BE,$37,$E3,$F9,$73,$0F,$89,$19,$7E,$34,$65
	.dc.b	$FF,$54,$47,$CE,$EE,$F9,$FF,$1B,$27,$E6,$46,$5F,$9D,$19,$7F,$D6
	.dc.b	$11,$F6,$6F,$F1,$A4,$8F,$C6,$B5,$F5,$7C,$E7,$E0,$8E,$0F,$F1,$5E
	.dc.b	$3F,$80,$FE,$32,$34,$E6,$7F,$A1,$7C,$2F,$FA,$C3,$AA,$F9,$BF,$A9
	.dc.b	$AF,$F7,$BE,$BF,$C4,$AB,$51,$F8,$5F,$0B,$F9,$9F,$DF,$36,$E7,$F8
	.dc.b	$07,$E4,$FB,$3F,$DF,$F8,$23,$18,$05,$F1,$7F,$21,$97,$E5,$88,$F8
	.dc.b	$AF,$9D,$72,$BF,$F6,$3F,$C8,$F8,$B8,$65,$EC,$4E,$9F,$E7,$18,$A7
	.dc.b	$CF,$10,$CB,$5F,$CF,$4F,$A5,$C6,$9F,$C9,$2B,$94,$7F,$43,$E1,$7C
	.dc.b	$07,$F5,$66,$C6,$63,$E7,$FE,$1A,$F8,$2D,$47,$D2,$FC,$39,$F0,$D4
	.dc.b	$CD,$F1,$3D,$99,$A7,$71,$86,$BF,$A1,$C7,$C2,$DF,$8A,$F2,$11,$E7
	.dc.b	$93,$E8,$F1,$1F,$02,$DF,$13,$E5,$33,$B7,$F4,$4C,$58,$E6,$FC,$55
	.dc.b	$AE,$2F,$F4,$4F,$DB,$BE,$1F,$FB,$CB,$6B,$FB,$53,$5F,$9F,$F0,$9F
	.dc.b	$D4,$32,$FE,$A3,$39,$7F,$25,$DF,$EE,$6F,$07,$4C,$26,$1F,$1A,$B5
	.dc.b	$8F,$E4,$9B,$1F,$91,$C3,$A4,$F9,$BC,$8F,$29,$27,$E7,$FD,$D6,$9E
	.dc.b	$11,$FA,$8B,$F8,$B1,$DF,$4E,$0F,$F1,$1C,$3F,$98,$F1,$7F,$12,$9E
	.dc.b	$06,$B4,$F6,$FC,$DF,$C4,$25,$FD,$EF,$81,$FF,$1A,$4E,$A3,$F0,$3E
	.dc.b	$87,$F2,$67,$F2,$AD,$8D,$47,$86,$BF,$7B,$39,$D7,$F6,$6D,$9F,$85
	.dc.b	$3F,$E2,$CF,$9B,$FF,$61,$47,$F8,$54,$38,$7F,$C3,$0D,$7F,$C5,$4D
	.dc.b	$7F,$D8,$EE,$EF,$FC,$79,$E6,$FF,$3D,$3B,$79,$BF,$E5,$1C,$5F,$87
	.dc.b	$3E,$4F,$FD,$9F,$A7,$F8,$A2,$73,$18,$E7,$FE,$11,$18,$CF,$F8,$55
	.dc.b	$7F,$8B,$BD,$27,$A0,$F8,$4E,$15,$FE,$3C,$CB,$B3,$C3,$1E,$71,$88
	.dc.b	$61,$E1,$8C,$3D,$6F,$A4,$F1,$3F,$8D,$7D,$07,$EE,$F0,$C4,$23,$0F
	.dc.b	$4C,$3F,$DD,$71,$8B,$FF,$21,$FD,$0B,$E2,$3C,$F7,$C7,$C2,$EF,$5F
	.dc.b	$E4,$8B,$E1,$87,$29,$CB,$FE,$F6,$C3,$95,$F8,$AB,$6F,$9A,$B6,$DC
	.dc.b	$7C,$F5,$C2,$3C,$C6,$4A,$78,$E3,$A4,$F2,$84,$F1,$C3,$CD,$39,$BF
	.dc.b	$0D,$71,$39,$D6,$F0,$F5,$61,$D0,$93,$AC,$7E,$1D,$F2,$26,$1E,$1C
	.dc.b	$39,$57,$9C,$FF,$3E,$4F,$D0,$6C,$FD,$86,$DF,$E7,$7E,$1F,$FC,$45
	.dc.b	$CF,$7D,$5F,$E8,$F5,$FE,$75,$9C,$B6,$32,$F6,$F6,$5E,$07,$E9,$F2
	.dc.b	$FF,$3D,$B6,$33,$3F,$E7,$F9,$CB,$63,$3B,$7D,$D6,$ED,$FD,$96,$5F
	.dc.b	$3A,$FA,$1F,$F4,$1D,$EF,$1F,$BD,$87,$0B,$FC,$9B,$5E,$35,$C3,$0F
	.dc.b	$C8,$CE,$71,$F1,$E1,$CE,$FE,$FE,$3D,$F9,$3A,$E5,$27,$F0,$77,$C3
	.dc.b	$1F,$35,$EE,$73,$8F,$52,$5F,$E9,$27,$58,$F7,$25,$F9,$27,$59,$79
	.dc.b	$F0,$F5,$C7,$86,$B7,$1F,$E0,$9B,$E7,$7C,$63,$F3,$39,$3E,$81,$F4
	.dc.b	$3C,$E3,$0F,$5E,$67,$B0,$F8,$CD,$5F,$0B,$FD,$F1,$C0,$AE,$70,$FD
	.dc.b	$3D,$F2,$AE,$38,$7A,$67,$F9,$4F,$9D,$FE,$AA,$7F,$92,$B5,$FF,$11
	.dc.b	$77,$23,$11,$C3,$1F,$E6,$53,$97,$FD,$A0,$D8,$FF,$39,$60,$FB,$9F
	.dc.b	$F3,$5D,$6A,$FF,$D9,$3D,$FF,$EB,$AD,$BF,$92,$4F,$F8,$2B,$E5,$1F
	.dc.b	$45,$A8,$FB,$7D,$37,$3F,$F1,$4D,$6A,$7F,$DC,$17,$CA,$7F,$DB,$4D
	.dc.b	$7F,$DD,$8D,$79,$C7,$F4,$B8,$E1,$87,$8E,$FB,$74,$FF,$AB,$7C,$7F
	.dc.b	$06,$1F,$A0,$9B,$F4,$5F,$E1,$2C,$7F,$C4,$3E,$17,$C1,$C7,$FD,$AB
	.dc.b	$39,$BF,$F1,$BC,$E6,$BF,$55,$39,$9F,$F8,$00,$D7,$3E,$37,$D0,$8C
	.dc.b	$47,$F9,$55,$E9,$7D,$A7,$CF,$97,$FA,$A6,$FF,$C9,$D6,$BD,$C7,$83
	.dc.b	$E1,$3D,$B8,$FF,$93,$E3,$72,$9C,$FA,$7E,$E7,$C4,$49,$FE,$E5,$3B
	.dc.b	$6C,$6D,$FF,$21,$E0,$F7,$31,$8F,$FA,$87,$E8,$BF,$D3,$B5,$FE,$48
	.dc.b	$F9,$FF,$89,$3F,$E7,$19,$CB,$FE,$5E,$8C,$A8,$CB,$FC,$4C,$C7,$BF
	.dc.b	$F9,$68,$CC,$7F,$17,$6B,$D6,$3E,$DC,$76,$D9,$7F,$C7,$7E,$8B,$FD
	.dc.b	$10,$FC,$A9,$D3,$CF,$DE,$FC,$2E,$3F,$27,$1F,$64,$63,$6E,$AF,$8C
	.dc.b	$95,$FE,$19,$BF,$F6,$03,$63,$6F,$D8,$C1,$EE,$E3,$0F,$F3,$FF,$D1
	.dc.b	$7F,$95,$EB,$FC,$DE,$D8,$F9,$9F,$3B,$F9,$88,$C6,$9E,$C9,$F9,$93
	.dc.b	$F1,$49,$5E,$37,$F7,$BF,$17,$C2,$3E,$44,$7E,$82,$31,$3C,$A7,$FD
	.dc.b	$CD,$E2,$7E,$CF,$B7,$F8,$BB,$E1,$FA,$2C,$78,$7F,$D4,$AB,$E5,$7C
	.dc.b	$37,$D5,$23,$F0,$DF,$0C,$F7,$38,$E1,$67,$F3,$7F,$3B,$FE,$9A,$7F
	.dc.b	$D5,$43,$8F,$E5,$CF,$FB,$B6,$74,$B5,$F8,$55,$BE,$1F,$DD,$D6,$AB
	.dc.b	$F1,$4D,$7F,$AF,$35,$FF,$62,$35,$3D,$AB,$E7,$46,$3D,$FF,$E2,$DA
	.dc.b	$3E,$7B,$F0,$D5,$AB,$E1,$8F,$FB,$3F,$6F,$F2,$2F,$CF,$FC,$BF,$4D
	.dc.b	$F4,$56,$30,$DA,$FF,$65,$57,$F9,$32,$31,$A7,$F3,$0B,$E3,$5A,$7F
	.dc.b	$CB,$4D,$FA,$2E,$DE,$BF,$76,$C6,$4E,$11,$97,$FD,$B1,$F4,$5F,$ED
	.dc.b	$8F,$8C,$FE,$64,$FF,$B6,$BE,$0A,$F8,$09,$B1,$86,$D7,$29,$D3,$67
	.dc.b	$4D,$35,$3C,$32,$FB,$51,$FE,$F5,$F9,$8F,$E2,$DB,$7E,$3B,$AB,$FA
	.dc.b	$24,$62,$7E,$2F,$4A,$73,$E1,$2E,$9F,$3F,$D3,$B7,$F1,$FE,$47,$E9
	.dc.b	$7C,$EF,$D3,$6C,$64,$F9,$91,$96,$9F,$96,$F2,$3F,$03,$D0,$F5,$3F
	.dc.b	$1F,$18,$6C,$FF,$A6,$FA,$3F,$9E,$EF,$F8,$CA,$FE,$DF,$18,$3F,$9C
	.dc.b	$C7,$E4,$A3,$DF,$F6,$5F,$C5,$A3,$F4,$11,$FB,$3F,$86,$FF,$49,$C7
	.dc.b	$F7,$27,$F6,$4D,$35,$D9,$B6,$E1,$F8,$C9,$E3,$D9,$FC,$27,$B3,$FE
	.dc.b	$00,$DF,$EA,$3D,$CF,$6F,$8B,$70,$63,$F5,$8F,$F2,$8E,$DF,$C6,$A3
	.dc.b	$F7,$35,$B9,$D5,$FE,$97,$E3,$31,$97,$EA,$BE,$8B,$F8,$7A,$F4,$BF
	.dc.b	$4F,$7D,$A7,$31,$88,$47,$F2,$37,$FA,$23,$5F,$F2,$C3,$5E,$5F,$11
	.dc.b	$FE,$1C,$8F,$E3,$B1,$CA,$3E,$6E,$9F,$E0,$09,$C8,$F8,$3F,$DE,$FD
	.dc.b	$0B,$E8,$F5,$18,$6E,$DF,$DC,$AB,$83,$F1,$B3,$FE,$69,$AB,$0F,$55
	.dc.b	$6B,$3F,$94,$7F,$42,$F6,$FB,$FD,$1F,$F8,$4B,$4F,$F1,$F6,$3C,$B0
	.dc.b	$4E,$6A,$65,$DC,$FA,$B9,$77,$FA,$CC,$E5,$F5,$FC,$0F,$F0,$07,$57
	.dc.b	$F9,$05,$B7,$FD,$B6,$51,$87,$FB,$83,$5F,$F4,$23,$9C,$FF,$A8,$36
	.dc.b	$7F,$8F,$78,$EF,$97,$FD,$9E,$1F,$C7,$D9,$3F,$C5,$FE,$1E,$E3,$FC
	.dc.b	$6E,$6E,$3F,$93,$C3,$FD,$A9,$7F,$C7,$67,$FC,$6B,$5F,$EF,$01,$CA
	.dc.b	$FA,$0F,$2F,$C0,$FF,$63,$8C,$4F,$F8,$27,$E0,$FF,$33,$18,$8F,$7C
	.dc.b	$EE,$BE,$15,$7E,$7B,$E8,$B5,$B7,$F9,$D7,$2C,$56,$B0,$56,$B0,$FE
	.dc.b	$E4,$F5,$8F,$36,$39,$BF,$0D,$F1,$8C,$4F,$AA,$3A,$56,$0E,$50,$C7
	.dc.b	$C2,$FF,$25,$C3,$BD,$70,$3E,$0E,$1A,$9F,$F5,$D5,$5C,$7D,$51,$E8
	.dc.b	$AE,$0F,$3A,$E0,$FC,$DC,$3C,$53,$E7,$3A,$C7,$9C,$F9,$85,$3E,$5C
	.dc.b	$FB,$9C,$CF,$26,$CF,$28,$F6,$DF,$97,$0F,$5D,$68,$F2,$1E,$73,$BE
	.dc.b	$3F,$20,$E2,$F8,$71,$F6,$C6,$2F,$C1,$3F,$08,$E4,$79,$8A,$7B,$F0
	.dc.b	$F2,$1F,$11,$D4,$76,$9F,$04,$FC,$39,$F2,$C7,$78,$E3,$1E,$93,$8B
	.dc.b	$CE,$BE,$8D,$6E,$7D,$AD,$76,$71,$8D,$CF,$78,$F4,$57,$8E,$3E,$9B
	.dc.b	$C6,$77,$3C,$6F,$E5,$BF,$06,$7D,$2F,$CA,$AE,$27,$5A,$F9,$2D,$AF
	.dc.b	$D2,$4F,$DE,$F7,$7F,$5E,$DB,$FD,$14,$62,$3E,$73,$7E,$21,$EF,$7F
	.dc.b	$8D,$65,$39,$8C,$37,$7F,$E9,$AD,$9D,$DF,$F5,$05,$7D,$2B,$F1,$D0
	.dc.b	$FA,$3F,$F4,$93,$86,$BF,$95,$9E,$0D,$C3,$C9,$7F,$C0,$EB,$6D,$F7
	.dc.b	$BC,$0F,$F0,$7B,$E2,$DA,$F6,$6B,$BC,$3F,$61,$87,$82,$35,$8F,$EC
	.dc.b	$F0,$E3,$A7,$AE,$BB,$B7,$2E,$EF,$BE,$77,$87,$F7,$07,$D3,$87,$78
	.dc.b	$FC,$DD,$F7,$AE,$18,$F9,$1B,$8F,$93,$1F,$95,$87,$13,$CB,$87,$E9
	.dc.b	$31,$EB,$7E,$38,$E0,$7A,$70,$E8,$F8,$DF,$93,$1B,$8F,$E0,$58,$FB
	.dc.b	$1D,$63,$F5,$AF,$AC,$7F,$82,$A7,$D0,$DC,$7C,$B7,$E8,$8F,$26,$1E
	.dc.b	$07,$EF,$4E,$9E,$07,$97,$08,$F8,$A3,$E5,$BA,$3F,$01,$1E,$39,$E9
	.dc.b	$7E,$29,$EA,$72,$C7,$E2,$9F,$1F,$0F,$73,$87,$CD,$97,$D3,$23,$83
	.dc.b	$D5,$1C,$67,$C4,$F1,$7E,$29,$F6,$B1,$F6,$3E,$23,$E0,$E3,$E3,$AF
	.dc.b	$09,$23,$8D,$7D,$AB,$F0,$C7,$BD,$C4,$F8,$8E,$11,$D6,$7C,$4F,$EB
	.dc.b	$31,$FB,$6F,$19,$FB,$95,$CA,$1E,$0B,$89,$F4,$7F,$B5,$CB,$FB,$C3
	.dc.b	$78,$BF,$C0,$6D,$BE,$34,$F9,$8C,$61,$CE,$AC,$F4,$AF,$F4,$E4,$62
	.dc.b	$B9,$71,$3E,$6D,$F4,$78,$6D,$FF,$88,$CF,$F0,$DB,$87,$FD,$51,$7B
	.dc.b	$9F,$F1,$04,$69,$B5,$D5,$B5,$C6,$7F,$89,$0F,$A2,$7D,$1F,$B6,$30
	.dc.b	$D8,$CC,$78,$C7,$D1,$3D,$CD,$57,$18,$7D,$1D,$33,$5C,$7E,$3F,$FD
	.dc.b	$1C,$DF,$43,$C1,$18,$8F,$F2,$5F,$D1,$BC,$AC,$D7,$96,$1F,$07,$4C
	.dc.b	$D7,$97,$E8,$DE,$BF,$AD,$B1,$98,$F7,$7C,$BF,$F3,$A8,$C1,$F2,$A7
	.dc.b	$6F,$F9,$2D,$AF,$C8,$6B,$FC,$ED,$DF,$47,$FB,$B2,$5E,$5F,$63,$C5
	.dc.b	$7E,$3D,$26,$1E,$77,$C7,$E2,$62,$3F,$CD,$76,$F5,$BC,$E6,$71,$FF
	.dc.b	$4F,$7D,$1F,$E4,$6B,$FD,$DA,$FF,$95,$DB,$7F,$94,$8D,$47,$FA,$7F
	.dc.b	$AD,$18,$AE,$75,$B3,$87,$57,$E0,$FE,$8D,$F1,$D4,$98,$FE,$0F,$E6
	.dc.b	$DF,$13,$88,$FF,$39,$57,$07,$F2,$AD,$73,$5F,$98,$51,$9C,$5E,$07
	.dc.b	$F6,$7C,$3F,$D3,$55,$DF,$D4,$ED,$5F,$E6,$DF,$13,$F5,$FF,$2F,$2B
	.dc.b	$F5,$1B,$FE,$F6,$7F,$66,$66,$3E,$17,$86,$AC,$C7,$EB,$FC,$2F,$DA
	.dc.b	$7C,$0F,$EC,$C9,$1D,$29,$F2,$78,$8F,$99,$F0,$5F,$09,$FF,$7D,$F0
	.dc.b	$E7,$4F,$5C,$7F,$B4,$36,$FD,$43,$F2,$7E,$43,$F6,$10,$6F,$87,$27
	.dc.b	$44,$3F,$D5,$F3,$C9,$B1,$CA,$64,$7D,$58,$FE,$27,$F4,$5F,$C5,$3C
	.dc.b	$C9,$3E,$B8,$CB,$ED,$8C,$BF,$1A,$3F,$8D,$F8,$3F,$C7,$3E,$46,$4F
	.dc.b	$C3,$8C,$BF,$16,$32,$FC,$E8,$FC,$55,$C8,$FE,$15,$CD,$E9,$CF,$E2
	.dc.b	$C3,$37,$D3,$EE,$3F,$17,$7F,$87,$BF,$81,$1F,$C6,$9A,$FF,$8B,$5E
	.dc.b	$9C,$FF,$89,$FA,$4F,$9A,$73,$C7,$E9,$FC,$4F,$FA,$C3,$AF,$F9,$E3
	.dc.b	$70,$FE,$60,$D7,$F0,$FE,$17,$F2,$5F,$A3,$70,$7E,$65,$B3,$FC,$C9
	.dc.b	$AF,$FB,$4D,$AF,$F9,$07,$AF,$FA,$87,$9F,$FD,$18,$91,$E0,$FA,$4F
	.dc.b	$CF,$AC,$FC,$F3,$F9,$08,$FE,$65,$6E,$3F,$21,$EA,$9D,$4F,$FB,$D2
	.dc.b	$74,$F9,$67,$D0,$47,$FD,$D1,$9B,$0F,$E8,$B8,$7C,$66,$EC,$A5,$6D
	.dc.b	$F0,$9F,$D1,$78,$DF,$21,$FD,$2D,$B1,$98,$F9,$FF,$48,$F8,$4D,$47
	.dc.b	$F0,$3A,$DD,$7F,$BE,$DA,$FF,$58,$6B,$FE,$6B,$77,$97,$DC,$FA,$37
	.dc.b	$C3,$FD,$C6,$D7,$84,$F3,$7C,$77,$94,$8F,$E0,$32,$7D,$26,$23,$C5
	.dc.b	$3E,$52,$3F,$DC,$31,$47,$7F,$8A,$93,$D8,$FE,$D1,$F4,$7F,$D6,$DB
	.dc.b	$E9,$7F,$6F,$8C,$47,$DF,$E0,$F9,$DC,$0F,$84,$FD,$2E,$FF,$A6,$BD
	.dc.b	$DF,$F7,$A6,$BF,$E8,$D6,$BF,$5D,$A9,$EF,$1F,$A2,$EE,$F8,$B3,$E1
	.dc.b	$3E,$EB,$88,$F0,$63,$E7,$F2,$A5,$7B,$1F,$D4,$FD,$27,$4C,$FC,$51
	.dc.b	$FC,$F7,$1F,$E8,$15,$E7,$8C,$11,$FD,$7E,$2D,$3F,$66,$D8,$EC,$7C
	.dc.b	$AF,$A4,$F9,$04,$7F,$6F,$F0,$7F,$E2,$F7,$D4,$7B,$E3,$C4,$47,$FC
	.dc.b	$4F,$82,$31,$1F,$07,$67,$A3,$FC,$EE,$30,$DD,$BE,$95,$7F,$89,$FE
	.dc.b	$83,$FD,$8B,$5F,$E2,$56,$BE,$26,$CF,$F4,$2E,$D7,$E1,$AF,$B0,$E5
	.dc.b	$E8,$7F,$63,$BE,$38,$7F,$8D,$9A,$FF,$B2,$5A,$FF,$66,$FA,$67,$CC
	.dc.b	$7F,$B3,$77,$FF,$16,$F1,$A7,$35,$FB,$99,$CC,$62,$11,$FE,$34,$CF
	.dc.b	$FC,$C9,$F0,$FC,$FA,$7F,$80,$61,$CF,$1C,$61,$E1,$6C,$66,$3B,$4F
	.dc.b	$F0,$7B,$E7,$18,$3F,$85,$7D,$2F,$F0,$1C,$31,$08,$C3,$E4,$AF,$F7
	.dc.b	$44,$63,$1F,$F2,$37,$D3,$3E,$2F,$C7,$7C,$7C,$2F,$F1,$E5,$7F,$92
	.dc.b	$EF,$86,$1C,$A7,$2F,$FB,$CF,$0E,$57,$E2,$AD,$BE,$6A,$DB,$71,$F3
	.dc.b	$D7,$08,$F3,$19,$29,$E3,$8E,$93,$CA,$13,$C7,$0F,$34,$E6,$FC,$35
	.dc.b	$C4,$E7,$5B,$C3,$D5,$87,$42,$4E,$B1,$F8,$77,$C8,$98,$78,$70,$E5
	.dc.b	$5E,$73,$FD,$07,$3F,$41,$B3,$FE,$65,$6D,$FF,$A3,$BE,$2B,$FC,$51
	.dc.b	$CD,$7D,$7F,$E8,$75,$FE,$8C,$13,$A7,$0D,$7B,$FB,$9F,$C7,$5E,$18
	.dc.b	$A3,$11,$AE,$FF,$F4,$EC,$E5,$B3,$C7,$6F,$BC,$DD,$BF,$AB,$5B,$E8
	.dc.b	$DF,$31,$FE,$32,$BD,$E1,$F7,$F0,$E1,$7F,$94,$6B,$C6,$B8,$61,$F9
	.dc.b	$29,$CE,$3E,$3A,$FB,$ED,$C3,$FD,$35,$8F,$83,$8B,$B6,$52,$F7,$87
	.dc.b	$6A,$D4,$62,$75,$7E,$68,$FC,$5C,$EA,$30,$4C,$7C,$95,$FA,$A9,$D4
	.dc.b	$62,$BA,$12,$BC,$B9,$7B,$23,$E0,$5F,$ED,$71,$7D,$33,$E4,$75,$18
	.dc.b	$78,$72,$3D,$87,$C7,$EA,$F8,$5F,$36,$E3,$FA,$68,$E4,$7F,$03,$BE
	.dc.b	$55,$C7,$0F,$25,$3A,$7F,$CD,$F7,$FC,$B7,$DA,$FF,$35,$F7,$7F,$2B
	.dc.b	$FA,$6F,$F4,$8C,$E4,$97,$FC,$C6,$B7,$1F,$C4,$E7,$73,$98,$FF,$6E
	.dc.b	$70,$B9,$FF,$CA,$73,$F8,$A8,$C4,$FF,$BB,$9B,$1F,$9F,$6B,$FE,$EC
	.dc.b	$73,$87,$F9,$71,$B3,$FD,$F2,$3F,$D5,$6F,$82,$BC,$B5,$FE,$2D,$AD
	.dc.b	$B6,$33,$CF,$FE,$43,$F1,$77,$3F,$9D,$9B,$F4,$BF,$DA,$70,$FF,$2C
	.dc.b	$F7,$FC,$8A,$FE,$C7,$39,$7F,$AF,$46,$72,$FF,$38,$CE,$5B,$E3,$78
	.dc.b	$32,$FF,$53,$61,$FF,$00,$6B,$4F,$EB,$9B,$87,$F9,$15,$B3,$FE,$46
	.dc.b	$EB,$F2,$DF,$9B,$E1,$3D,$79,$7F,$96,$71,$E9,$18,$F4,$75,$18,$3F
	.dc.b	$CE,$78,$3D,$BC,$E7,$EC,$BB,$A7,$F9,$F2,$B5,$7C,$6B,$FC,$07,$F4
	.dc.b	$BF,$EA,$D8,$FF,$3E,$FD,$17,$C1,$7F,$83,$C6,$67,$FC,$05,$39,$7F
	.dc.b	$CC,$11,$9B,$FF,$4B,$B6,$BF,$87,$FD,$2F,$FA,$11,$FF,$33,$4E,$BD
	.dc.b	$FF,$AF,$5B,$BE,$71,$F6,$AF,$84,$79,$A3,$7E,$F7,$E0,$31,$F8,$D3
	.dc.b	$EA,$8C,$6B,$AC,$BB,$9E,$A8,$E9,$ED,$E6,$B8,$4E,$9F,$5C,$EA,$31
	.dc.b	$0C,$93,$F9,$0F,$A5,$7A,$54,$F5,$FD,$0D,$C3,$8D,$A3,$31,$D3,$E2
	.dc.b	$F8,$BA,$8F,$CC,$C6,$31,$FA,$B1,$8B,$E3,$1F,$EE,$CF,$0B,$F4,$BE
	.dc.b	$07,$F0,$F7,$C9,$F4,$B8,$F0,$FF,$CC,$D9,$78,$BE,$13,$E7,$9D,$A6
	.dc.b	$C7,$8D,$C3,$F0,$EF,$F8,$D6,$18,$3F,$C9,$3F,$43,$FE,$9A,$9F,$EB
	.dc.b	$90,$8C,$3F,$D7,$DA,$FF,$50,$35,$EE,$FF,$0A,$7F,$D5,$11,$C3,$9B
	.dc.b	$FD,$BE,$DA,$FF,$56,$66,$FA,$0F,$BF,$5A,$C3,$F7,$F7,$BC,$39,$C7
	.dc.b	$F0,$8B,$E1,$C9,$A8,$F8,$FF,$0E,$FA,$2B,$F4,$B8,$6E,$9F,$06,$74
	.dc.b	$D9,$D3,$4D,$7A,$9D,$3E,$97,$AB,$63,$27,$38,$CB,$4F,$C8,$7D,$2F
	.dc.b	$FB,$23,$E3,$3F,$8F,$BF,$EC,$AF,$80,$BE,$02,$6C,$61,$B5,$CA,$74
	.dc.b	$D9,$D3,$4D,$4F,$0D,$FF,$D3,$9F,$1A,$F0,$BD,$F1,$86,$F1,$FF,$9D
	.dc.b	$76,$FE,$E6,$B0,$FF,$70,$5F,$1D,$FF,$9B,$B6,$3F,$58,$DC,$3F,$D5
	.dc.b	$5C,$AC,$39,$D7,$2C,$7B,$BF,$E6,$58,$7C,$8F,$D8,$F9,$EF,$B2,$D9
	.dc.b	$FA,$CD,$C3,$AD,$70,$9F,$B5,$5F,$86,$F1,$FF,$21,$5F,$97,$6F,$3F
	.dc.b	$F4,$18,$FC,$B3,$5F,$F1,$3B,$FC,$E9,$E1,$5E,$CF,$A5,$FE,$2B,$B7
	.dc.b	$E6,$2F,$B4,$FF,$08,$BF,$E8,$3A,$AF,$73,$D2,$BF,$73,$F1,$1F,$E8
	.dc.b	$AC,$7E,$2B,$FB,$86,$9A,$E9,$DB,$6B,$F6,$D3,$C7,$2F,$F5,$53,$C5
	.dc.b	$FD,$5F,$B7,$FA,$63,$F4,$9D,$6F,$8B,$C5,$F9,$E6,$E1,$FE,$DF,$C7
	.dc.b	$F6,$15,$8C,$3F,$01,$21,$F4,$7F,$C5,$DF,$DF,$35,$FF,$27,$B5,$E0
	.dc.b	$D7,$8F,$83,$FD,$E2,$3F,$90,$4F,$6D,$BE,$47,$8F,$9A,$99,$FF,$34
	.dc.b	$BD,$BF,$92,$DD,$5A,$C7,$FD,$BF,$4F,$92,$EC,$7F,$05,$FA,$3F,$E1
	.dc.b	$13,$EA,$8C,$E1,$D5,$D7,$81,$89,$FF,$74,$E1,$FE,$38,$D3,$E1,$C7
	.dc.b	$83,$D7,$E9,$F4,$5F,$E1,$2A,$D6,$3F,$E7,$06,$D7,$C9,$6F,$03,$35
	.dc.b	$F1,$C3,$F5,$7D,$2E,$7A,$7F,$1A,$F9,$5B,$9F,$EB,$A7,$4D,$9D,$34
	.dc.b	$FE,$39,$38,$CF,$F4,$FF,$07,$75,$FD,$F4,$DA,$FD,$94,$DA,$FE,$75
	.dc.b	$EF,$7E,$07,$E8,$5F,$43,$FE,$8C,$E6,$FF,$2F,$E1,$98,$FF,$50,$55
	.dc.b	$BF,$E0,$FE,$F7,$E2,$B0,$E3,$1D,$03,$85,$87,$47,$37,$D1,$A7,$4C
	.dc.b	$3B,$3F,$EF,$F3,$27,$2B,$F0,$4F,$1B,$ED,$7D,$5F,$05,$7E,$77,$0E
	.dc.b	$73,$9A,$E0,$7B,$EF,$9D,$7A,$6F,$9C,$62,$3C,$51,$8C,$3C,$B1,$D4
	.dc.b	$F2,$DF,$57,$95,$F5,$9D,$C3,$0F,$1E,$1D,$63,$96,$1D,$5F,$79,$8B
	.dc.b	$E4,$E0,$E4,$78,$27,$93,$8E,$8E,$47,$C2,$71,$3C,$A3,$17,$E2,$C3
	.dc.b	$CA,$DA,$EC,$DC,$37,$1D,$E7,$6F,$F1,$1B,$E8,$7D,$BB,$F0,$BF,$3A
	.dc.b	$3E,$E1,$D0,$E4,$5A,$F8,$07,$62,$C7,$18,$FC,$05,$78,$27,$31,$F3
	.dc.b	$8E,$AF,$19,$F8,$ED,$8C,$C7,$3C,$7E,$B4,$7D,$C8,$D9,$9C,$7E,$5E
	.dc.b	$3F,$06,$31,$87,$85,$DD,$78,$4F,$29,$ED,$F6,$7F,$9A,$3F,$A3,$6B
	.dc.b	$F9,$B6,$BF,$89,$73,$F0,$8F,$61,$BF,$F3,$59,$CC,$62,$BF,$D7,$5A
	.dc.b	$FF,$39,$57,$C6,$F8,$4E,$63,$13,$C0,$FE,$80,$E3,$5F,$DB,$7D,$17
	.dc.b	$FA,$46,$7F,$62,$D7,$F6,$33,$BC,$BA,$63,$E3,$BE,$3E,$CF,$8D,$B7
	.dc.b	$FB,$AB,$E0,$F8,$2B,$85,$FA,$70,$ED,$7C,$2B,$BD,$69,$E1,$97,$93
	.dc.b	$1E,$D8,$71,$6C,$E8,$E3,$3B,$61,$82,$5F,$72,$57,$B2,$7D,$17,$E1
	.dc.b	$A6,$1E,$6A,$F0,$46,$AA,$3F,$C5,$A7,$4F,$F6,$D6,$BF,$5A,$BA,$63
	.dc.b	$F5,$32,$F8,$04,$C3,$C5,$5F,$AD,$3D,$98,$F9,$9E,$99,$7C,$07,$A6
	.dc.b	$5E,$29,$D4,$7A,$27,$58,$7D,$28,$D9,$D4,$D1,$FE,$03,$C3,$C4,$FC
	.dc.b	$B7,$F9,$D4,$71,$CB,$D6,$7B,$1E,$86,$6B,$C9,$19,$CB,$DE,$7C,$B7
	.dc.b	$8C,$79,$23,$38,$FD,$6C,$7D,$37,$D2,$78,$57,$43,$85,$7B,$F2,$FA
	.dc.b	$79,$7D,$19,$CB,$F7,$A3,$38,$74,$BE,$F3,$B3,$C1,$3F,$42,$7F,$AD
	.dc.b	$F8,$BF,$E5,$DF,$A2,$FD,$76,$3F,$E6,$96,$ED,$FE,$5D,$6E,$0A,$FF
	.dc.b	$32,$7C,$B6,$1E,$51,$FE,$93,$9C,$78,$C6,$D7,$2E,$37,$D0,$BE,$8B
	.dc.b	$0D,$C3,$FC,$E6,$D7,$FC,$CF,$7A,$99,$1A,$7B,$C7,$9A,$1F,$44,$FA
	.dc.b	$2C,$36,$3F,$DC,$ED,$7F,$89,$E9,$FC,$3F,$E8,$BF,$DD,$31,$AA,$E2
	.dc.b	$47,$FD,$DF,$1A,$78,$9F,$E4,$FF,$8B,$7C,$5F,$8D,$B1,$FE,$F6,$6B
	.dc.b	$C0,$7D,$0F,$FB,$D9,$D4,$F1,$23,$FE,$FE,$8D,$3C,$7E,$87,$FD,$48
	.dc.b	$FF,$97,$F8,$BF,$C7,$F1,$88,$D7,$81,$F1,$7E,$51,$D9,$F1,$EB,$57
	.dc.b	$C2,$B0,$6E,$6D,$FF,$BA,$AB,$A3,$FE,$E7,$6B,$FC,$DD,$AF,$F9,$C5
	.dc.b	$65,$DC,$91,$E7,$D7,$F9,$F6,$2F,$03,$33,$ED,$AB,$97,$F9,$6A,$B5
	.dc.b	$3F,$F0,$07,$E8,$5C,$BF,$93,$73,$1F,$EB,$A6,$D7,$FA,$13,$9D,$1C
	.dc.b	$27,$95,$71,$C6,$46,$0F,$31,$FE,$8B,$FA,$07,$C5,$7E,$3D,$B1,$A0
	.dc.b	$E1,$FB,$DF,$30,$F8,$0E,$15,$27,$85,$6F,$3F,$F3,$BD,$EE,$73,$5F
	.dc.b	$D8,$D9,$5D,$5B,$19,$7F,$B4,$46,$5F,$F6,$B7,$77,$ED,$3E,$2A,$47
	.dc.b	$B3,$5F,$F6,$F4,$FE,$E0,$CC,$7F,$B5,$A7,$0F,$ED,$3D,$9F,$DB,$91
	.dc.b	$E5,$DC,$F8,$EC,$47,$FB,$E3,$B9,$58,$D7,$FD,$ED,$5F,$B7,$D7,$F5
	.dc.b	$51,$F0,$3E,$2B,$06,$7E,$26,$63,$F3,$69,$3A,$8C,$4E,$9F,$9F,$19
	.dc.b	$7E,$8C,$7F,$14,$F9,$FF,$E2,$BE,$2C,$9F,$4C,$65,$F5,$C6,$5F,$F7
	.dc.b	$BC,$7F,$1C,$ED,$FE,$3B,$F0,$92,$7E,$0C,$65,$F8,$71,$97,$FD,$FF
	.dc.b	$1F,$87,$B9,$1F,$80,$7E,$17,$D0,$E2,$3F,$82,$71,$FC,$4E,$BF,$B7
	.dc.b	$5A,$9E,$07,$DF,$3F,$1E,$D6,$99,$9F,$E2,$CF,$F8,$AF,$E0,$7F,$DD
	.dc.b	$9D,$3F,$CF,$1B,$1F,$9E,$6B,$FC,$73,$99,$C9,$FE,$D8,$8F,$CF,$7D
	.dc.b	$03,$6F,$E6,$0D,$8C,$C7,$5F,$33,$F5,$D1,$CE,$BF,$C8,$FC,$7F,$B3
	.dc.b	$6E,$BF,$EF,$7C,$7B,$DF,$97,$E8,$3F,$DC,$98,$79,$63,$F1,$CF,$8B
	.dc.b	$87,$C2,$77,$E5,$B1,$92,$1F,$CD,$B5,$57,$28,$C5,$FE,$B7,$36,$49
	.dc.b	$5A,$B5,$39,$8C,$4F,$80,$FE,$79,$C4,$F5,$FF,$AA,$36,$3F,$83,$35
	.dc.b	$F2,$7C,$FB,$DD,$DD,$49,$DB,$F5,$E7,$51,$83,$EC,$7D,$03,$E0,$3E
	.dc.b	$E3,$63,$47,$93,$E1,$7C,$64,$7F,$01,$4F,$A0,$C4,$6C,$F3,$CF,$86
	.dc.b	$33,$B7,$C6,$9F,$8B,$6F,$83,$F0,$9A,$DB,$FC,$F5,$F3,$FF,$D6,$2F
	.dc.b	$85,$A3,$13,$AF,$83,$FC,$A7,$BF,$AB,$DD,$6A,$FE,$A3,$5E,$55,$C2
	.dc.b	$FF,$C0,$53,$BE,$17,$81,$F0,$EF,$F0,$4D,$C3,$AE,$1C,$3C,$59,$5F
	.dc.b	$CE,$F0,$FE,$69,$8F,$DF,$FA,$0F,$04,$62,$7C,$75,$C1,$FC,$74,$EA
	.dc.b	$3F,$5D,$8F,$C5,$9F,$D7,$35,$FF,$01,$57,$37,$E2,$7D,$07,$A8,$8F
	.dc.b	$ED,$7D,$BF,$F1,$63,$E5,$8C,$BC,$9F,$DE,$46,$A3,$07,$F0,$79,$4F
	.dc.b	$F0,$76,$C6,$63,$F8,$57,$CC,$FF,$AD,$E3,$FA,$E3,$5F,$F0,$C3,$74
	.dc.b	$FE,$D9,$F4,$0F,$A0,$FF,$19,$4E,$75,$FF,$5D,$36,$FF,$58,$DE,$E5
	.dc.b	$F1,$9F,$ED,$F4,$FA,$07,$C8,$7F,$B0,$F5,$FF,$15,$CE,$63,$1D,$9F
	.dc.b	$B7,$8C,$5F,$EE,$6B,$FC,$63,$8F,$F1,$AF,$53,$73,$FB,$EC,$7A,$BB
	.dc.b	$C3,$94,$62,$17,$E0,$8C,$3D,$3C,$2F,$F1,$5C,$F1,$3F,$84,$7C,$FF
	.dc.b	$EE,$F0,$C4,$23,$0F,$4C,$3F,$DB,$D1,$8B,$FF,$22,$FD,$03,$E1,$3C
	.dc.b	$F7,$C7,$C1,$EF,$5F,$E4,$AB,$E1,$87,$29,$CB,$FE,$E9,$C3,$95,$F8
	.dc.b	$AB,$6F,$9A,$B6,$DC,$7C,$F5,$C2,$3C,$C6,$4A,$78,$E3,$A4,$F2,$84
	.dc.b	$F1,$C3,$CD,$39,$BF,$0D,$71,$39,$D6,$F0,$F5,$61,$D0,$93,$AC,$7E
	.dc.b	$1D,$F2,$26,$1E,$1C,$39,$57,$9C,$FF,$14,$4F,$D0,$6C,$FD,$86,$DF
	.dc.b	$F6,$BF,$82,$7A,$3F,$D0,$27,$FC,$59,$18,$EB,$D4,$E6,$7F,$C6,$93
	.dc.b	$97,$FC,$65,$87,$F9,$83,$B3,$FB,$B3,$63,$3E,$BA,$72,$D8,$CC,$7D
	.dc.b	$78,$C6,$DF,$E8,$8F,$9D,$7D,$07,$F3,$18,$FC,$14,$EB,$83,$F1,$6D
	.dc.b	$C3,$F2,$6D,$57,$CB,$93,$B6,$32,$7E,$EE,$5E,$7C,$7C,$19,$77,$CB
	.dc.b	$AD,$EE,$7C,$2F,$E7,$B1,$E9,$87,$2C,$BC,$18,$F3,$C3,$8C,$FE,$2A
	.dc.b	$B7,$87,$28,$E1,$87,$5A,$FE,$E5,$1E,$8C,$7C,$4F,$B2,$74,$D3,$A4
	.dc.b	$69,$FD,$86,$3F,$0F,$91,$E8,$3E,$93,$9C,$61,$B1,$FE,$D5,$D7,$F7
	.dc.b	$93,$96,$C7,$16,$CF,$0F,$82,$E2,$F2,$F1,$BF,$B3,$76,$FF,$A4,$6F
	.dc.b	$FC,$09,$7F,$E7,$15,$3A,$8F,$F1,$14,$6B,$2F,$F1,$93,$A8,$E1,$E4
	.dc.b	$EE,$7F,$DE,$93,$97,$71,$9C,$9F,$87,$F9,$E7,$FC,$78,$FB,$5B,$96
	.dc.b	$E3,$19,$7F,$A9,$DE,$0D,$CF,$53,$B6,$E5,$FE,$8B,$AE,$2F,$F8,$ED
	.dc.b	$E2,$D8,$F2,$65,$FD,$1E,$3E,$17,$AD,$8F,$03,$2D,$8C,$9F,$1E,$32
	.dc.b	$D3,$39,$7C,$68,$CB,$EE,$7D,$93,$ED,$6C,$F9,$DB,$3B,$8F,$54,$ED
	.dc.b	$A6,$CB,$1F,$53,$EC,$3D,$7E,$67,$B1,$B3,$F7,$5B,$3B,$8F,$C0,$4E
	.dc.b	$DA,$6C,$B1,$F5,$BC,$9F,$A9,$F6,$3F,$55,$B3,$F3,$DB,$3B,$8F,$A5
	.dc.b	$3B,$69,$B2,$C7,$E3,$23,$E9,$9F,$53,$E0,$2C,$FE,$55,$B3,$F9,$26
	.dc.b	$EF,$F9,$B7,$84,$FE,$6E,$ED,$7E,$63,$E9,$FE,$F9,$F8,$DF,$8E,$B5
	.dc.b	$F8,$7F,$B0,$C3,$4F,$C5,$47,$E4,$FE,$C6,$47,$E9,$E2,$DB,$2F,$F2
	.dc.b	$97,$CD,$7F,$9D,$B7,$FD,$23,$FE,$78,$E7,$BE,$0F,$FB,$4B,$1F,$F7
	.dc.b	$DC,$E5,$B1,$9C,$3A,$47,$FA,$06,$75,$18,$0F,$78,$E3,$5D,$AF,$BB
	.dc.b	$63,$27,$EF,$F3,$FD,$0E,$7F,$A2,$6C,$64,$FE,$09,$19,$69,$FC,$1F
	.dc.b	$C3,$FD,$4F,$C8,$FE,$A9,$B1,$93,$F8,$64,$65,$A7,$F0,$FF,$89,$FD
	.dc.b	$8F,$67,$EC,$9B,$19,$3F,$8A,$46,$5A,$67,$4F,$E2,$71,$96,$C7,$FA
	.dc.b	$9F,$6F,$F3,$EF,$CC,$7F,$A0,$E3,$07,$FA,$AF,$BD,$F3,$0F,$07,$F8
	.dc.b	$D6,$0B,$E0,$FC,$FA,$DC,$EA,$73,$3F,$D1,$0E,$1C,$BF,$E8,$C7,$FA
	.dc.b	$3C,$69,$B1,$97,$F9,$3E,$1F,$C4,$BD,$6F,$E2,$6D,$3F,$9A,$16,$3F
	.dc.b	$98,$B7,$2F,$C7,$F2,$39,$FF,$1B,$87,$E3,$9B,$5F,$8C,$6D,$FE,$27
	.dc.b	$97,$FD,$97,$18,$C3,$9B,$6B,$F2,$18,$75,$7F,$2B,$87,$F3,$BF,$C4
	.dc.b	$1E,$B5,$FB,$0C,$36,$BE,$7E,$6F,$0D,$7C,$27,$31,$8B,$E3,$18,$94
	.dc.b	$7D,$4F,$99,$E1,$7F,$A1,$D1,$F5,$9F,$D5,$5F,$F5,$DC,$E9,$FE,$BF
	.dc.b	$19,$C3,$FB,$CC,$F2,$AF,$B0,$D7,$39,$7D,$55,$19,$7F,$5F,$A7,$FB
	.dc.b	$D2,$BF,$6F,$BF,$EF,$1F,$DE,$B5,$C9,$FB,$C3,$25,$33,$3F,$4E,$7A
	.dc.b	$36,$7F,$27,$F6,$AF,$8B,$FE,$DD,$F2,$6F,$B6,$7C,$1E,$A7,$31,$8C
	.dc.b	$BE,$8E,$2C,$39,$47,$E5,$E0,$66,$BE,$04,$EA,$34,$D7,$F3,$DF,$86
	.dc.b	$3E,$8B,$FC,$01,$F8,$6B,$F0,$F7,$D1,$FE,$AA,$B5,$97,$E0,$AB,$53
	.dc.b	$9B,$F9,$D5,$AC,$3C,$8D,$7B,$C7,$E9,$F9,$9F,$6C,$FA,$9F,$0F,$E1
	.dc.b	$BF,$EA,$FB,$FE,$49,$5A,$9C,$C7,$F8,$46,$B8,$67,$E3,$BE,$1C,$E8
	.dc.b	$F8,$78,$71,$75,$7C,$79,$A4,$7E,$63,$0F,$CA,$B6,$33,$3E,$29,$FC
	.dc.b	$94,$F4,$77,$19,$9F,$E1,$EF,$FB,$14,$7C,$EB,$E7,$BF,$C9,$FC,$AF
	.dc.b	$A0,$F3,$56,$A7,$ED,$7B,$1F,$6A,$F8,$4F,$DC,$BE,$19,$66,$F7,$3C
	.dc.b	$4D,$61,$D2,$F8,$C7,$FB,$74,$D0,$BE,$6E,$0F,$44,$70,$8C,$BC,$A3
	.dc.b	$35,$EA,$AE,$CF,$8A,$78,$46,$27,$F9,$56,$1E,$77,$CB,$85,$7C,$95
	.dc.b	$C2,$31,$1D,$8D,$9E,$53,$FD,$CE,$7A,$B0,$F2,$E1,$E1,$3C,$93,$CA
	.dc.b	$77,$5D,$5E,$57,$E1,$BE,$31,$E6,$3A,$61,$E9,$7C,$B8,$74,$78,$9A
	.dc.b	$C3,$B5,$F3,$AE,$13,$E9,$3E,$73,$87,$C7,$5C,$F0,$F9,$91,$83,$DE
	.dc.b	$E2,$BA,$5F,$69,$F0,$1E,$29,$F6,$1B,$72,$D3,$35,$F1,$63,$AB,$5F
	.dc.b	$2C,$F5,$9D,$C7,$C3,$6B,$F0,$8E,$26,$67,$D4,$73,$C3,$E5,$C6,$CE
	.dc.b	$A7,$9E,$36,$F1,$39,$9B,$8E,$2F,$C9,$6B,$F1,$4E,$73,$ED,$9E,$71
	.dc.b	$F9,$DF,$75,$F1,$BF,$A4,$6C,$7D,$D6,$BB,$8F,$F6,$0F,$BA,$F9,$7F
	.dc.b	$F2,$04,$AB,$73,$9E,$0F,$E5,$B1,$8A,$DB,$63,$37,$D2,$79,$E3,$FB
	.dc.b	$2F,$A0,$7C,$B7,$FB,$42,$30,$DC,$F8,$61,$FA,$5C,$BC,$19,$FE,$E3
	.dc.b	$0E,$2F,$EA,$5A,$FE,$F2,$37,$87,$F0,$06,$DF,$FB,$7A,$B7,$3E,$0A
	.dc.b	$ED,$8F,$9A,$BC,$31,$DF,$1F,$23,$FD,$32,$F8,$CE,$AF,$B5,$F1,$AD
	.dc.b	$5F,$05,$7C,$1F,$EA,$B8,$79,$1B,$5C,$5D,$D7,$1A,$E1,$0A,$E0,$DA
	.dc.b	$E1,$3A,$37,$3A,$36,$E6,$BC,$91,$99,$E3,$1D,$CC,$BE,$38,$E3,$19
	.dc.b	$AF,$8F,$8C,$7E,$31,$F4,$E3,$8C,$67,$0F,$9D,$3A,$3E,$B5,$F5,$3C
	.dc.b	$2F,$13,$86,$3F,$34,$FB,$17,$E1,$8E,$37,$A7,$E7,$9C,$4C,$92,$FC
	.dc.b	$24,$AE,$91,$F4,$27,$84,$E6,$30,$71,$9F,$0B,$A7,$8B,$F4,$67,$0F
	.dc.b	$88,$E0,$72,$9E,$8F,$DC,$3D,$1C,$5F,$E0,$0F,$B2,$FF,$0B,$C7,$F6
	.dc.b	$D6,$D7,$F6,$49,$D4,$FF,$75,$F8,$DC,$F0,$7F,$71,$BB,$3F,$E5,$3E
	.dc.b	$44,$62,$F8,$CF,$F7,$9B,$C5,$3E,$D3,$8F,$5B,$D0,$7B,$F8,$6E,$1F
	.dc.b	$E0,$5A,$E2,$66,$64,$64,$FF,$1A,$FE,$1A,$F8,$5F,$F1,$A4,$7F,$8D
	.dc.b	$23,$0E,$1A,$FF,$8E,$5A,$FF,$A4,$5C,$ED,$FE,$06,$8F,$E2,$DE,$AF
	.dc.b	$C0,$69,$E0,$AC,$7A,$6F,$8A,$F0,$CA,$30,$F9,$1A,$FF,$92,$5A,$FF
	.dc.b	$A8,$5C,$DF,$87,$DA,$7D,$57,$85,$B5,$FE,$54,$AD,$F1,$FD,$CF,$C4
	.dc.b	$7E,$64,$FF,$96,$17,$B5,$E9,$6C,$67,$83,$06,$66,$46,$48,$7F,$B9
	.dc.b	$C4,$F2,$70,$DC,$BE,$43,$E2,$8E,$1F,$43,$FE,$AD,$7E,$AB,$5F,$F3
	.dc.b	$53,$5F,$F6,$13,$5F,$CB,$BB,$9F,$A9,$F4,$2F,$A8,$FA,$8D,$AF,$F3
	.dc.b	$BD,$6F,$2F,$F2,$9F,$E2,$5F,$E6,$F9,$FF,$52,$1F,$97,$F5,$7F,$DA
	.dc.b	$18,$FE,$86,$75,$3F,$E6,$26,$C7,$E9,$67,$75,$FE,$5E,$6D,$7F,$AD
	.dc.b	$6B,$73,$FA,$EC,$1F,$15,$E6,$E1,$97,$FE,$8C,$9C,$B6,$72,$FC,$F8
	.dc.b	$FB,$AF,$FA,$39,$DB,$98,$DE,$3F,$12,$BE,$1C,$7E,$C3,$EB,$7F,$63
	.dc.b	$F1,$12,$FF,$CE,$11,$97,$FC,$E5,$19,$7F,$CE,$11,$F9,$FF,$9F,$76
	.dc.b	$7E,$71,$B3,$FA,$E6,$BF,$EA,$36,$BF,$EF,$B7,$38,$FE,$AF,$E5,$DF
	.dc.b	$57,$FA,$B6,$D7,$FA,$BA,$B7,$AF,$F9,$FF,$E8,$BF,$84,$FD,$BC,$AF
	.dc.b	$F4,$44,$7F,$0C,$30,$FF,$A1,$E3,$0E,$8F,$F4,$73,$AC,$7F,$8F,$78
	.dc.b	$0A,$FE,$2B,$F7,$3F,$C5,$BE,$92,$61,$FE,$95,$8C,$BF,$E9,$78,$CB
	.dc.b	$FE,$95,$8F,$E3,$BF,$37,$FC,$7B,$49,$3F,$E9,$D8,$CB,$FE,$9F,$8C
	.dc.b	$BF,$E9,$D8,$F8,$55,$F0,$EE,$4F,$FB,$42,$93,$A8,$C4,$E9,$FD,$A4
	.dc.b	$E9,$E1,$39,$C7,$E8,$57,$19,$FC,$E7,$B3,$FC,$DF,$97,$FD,$C9,$1F
	.dc.b	$ED,$AA,$FD,$14,$FF,$38,$FB,$9D,$7D,$2E,$FE,$B2,$5E,$F8,$3F,$5B
	.dc.b	$3F,$CF,$BC,$B9,$3F,$EB,$D7,$6F,$FA,$EF,$EC,$BF,$A1,$91,$FF,$61
	.dc.b	$72,$57,$FD,$81,$F5,$3F,$D2,$0C,$69,$F6,$1C,$47,$FB,$25,$CE,$3F
	.dc.b	$BE,$F3,$BD,$9F,$4D,$2B,$CC,$7E,$D6,$3E,$01,$97,$F7,$D1,$87,$7E
	.dc.b	$6B,$83,$FC,$7F,$8F,$F0,$5A,$E0,$FF,$07,$BF,$EB,$BF,$19,$FD,$7B
	.dc.b	$D8,$98,$FF,$B7,$63,$2F,$FB,$7E,$32,$FF,$B7,$63,$2F,$FB,$7E,$33
	.dc.b	$5F,$96,$9F,$AD,$E1,$FD,$7F,$8C,$97,$E6,$CB,$F9,$03,$FC,$5E,$72
	.dc.b	$FF,$19,$9E,$2E,$E3,$37,$FC,$7E,$73,$3F,$F0,$03,$EE,$BF,$8C,$BF
	.dc.b	$DF,$4F,$EF,$27,$F7,$D7,$3A,$1B,$D7,$FC,$8A,$0D,$8D,$9A,$9E,$67
	.dc.b	$CE,$79,$F7,$FF,$80,$21,$87,$F8,$4E,$1F,$31,$FE,$0D,$6C,$FD,$67
	.dc.b	$84,$EF,$D0,$D5,$4A,$D4,$71,$30,$F6,$E1,$2C,$98,$76,$8D,$19,$7B
	.dc.b	$CF,$09,$F8,$5E,$6F,$C5,$F8,$19,$7E,$D8,$F9,$06,$1F,$9D,$18,$7E
	.dc.b	$17,$8B,$FE,$34,$F7,$1B,$F5,$DB,$F9,$FC,$62,$BC,$D1,$89,$D7,$8F
	.dc.b	$C5,$E1,$76,$B6,$DC,$BE,$C3,$6F,$FC,$27,$87,$5A,$7C,$E3,$E4,$BF
	.dc.b	$D6,$90,$9D,$4E,$7D,$DF,$93,$18,$AD,$BC,$FE,$75,$F3,$DE,$3C,$3A
	.dc.b	$D6,$B4,$FF,$29,$E3,$CE,$7B,$46,$27,$51,$F5,$AF,$9B,$E0,$9D,$63
	.dc.b	$63,$ED,$DF,$3C,$7C,$93,$99,$5F,$27,$B9,$C3,$0E,$EF,$9A,$78,$5F
	.dc.b	$3C,$3F,$CD,$37,$C7,$0E,$B7,$C8,$E0,$5A,$E0,$FA,$71,$F8,$51,$88
	.dc.b	$DC,$F2,$3E,$44,$EB,$1F,$1D,$6C,$E1,$7E,$1C,$3E,$0E,$1F,$E7,$C3
	.dc.b	$D5,$8F,$F9,$8A,$BE,$64,$65,$AF,$FB,$CB,$D5,$7C,$5F,$F9,$A3,$DE
	.dc.b	$FF,$35,$36,$33,$3D,$4F,$F6,$23,$87,$A7,$AC,$EC,$F0,$36,$38,$61
	.dc.b	$FE,$63,$9C,$92,$3E,$34,$EF,$6F,$C9,$7C,$F7,$F8,$A3,$9B,$E8,$CE
	.dc.b	$75,$FC,$3B,$6F,$EF,$B5,$57,$17,$AE,$1C,$88,$FD,$78,$D6,$1F,$9C
	.dc.b	$AE,$58,$F9,$EB,$89,$BE,$26,$5C,$AF,$86,$37,$2F,$F5,$6E,$5D,$1F
	.dc.b	$3C,$67,$0F,$3B,$F9,$39,$D4,$60,$EF,$5D,$1F,$05,$F1,$E3,$7A,$3F
	.dc.b	$EA,$88,$FD,$04,$63,$A7,$CB,$5F,$9C,$8D,$F5,$63,$0F,$E1,$51,$BB
	.dc.b	$E3,$87,$3B,$E0,$DA,$DB,$FC,$CA,$77,$87,$3C,$FF,$91,$4B,$E5,$BC
	.dc.b	$3D,$9F,$EC,$B8,$7C,$F6,$30,$DD,$5C,$BF,$98,$56,$A7,$FB,$8B,$CD
	.dc.b	$B5,$BC,$FE,$0D,$DA,$E1,$7F,$CD,$A5,$F3,$DE,$BC,$7F,$DB,$8F,$A6
	.dc.b	$74,$E7,$1F,$E7,$4E,$67,$79,$F7,$8C,$4E,$A3,$1A,$E5,$B1,$FE,$08
	.dc.b	$9E,$97,$FD,$2D,$E2,$FB,$BE,$7B,$D1,$F3,$5E,$96,$D7,$C0,$6C,$FA
	.dc.b	$F3,$46,$32,$F9,$8D,$AF,$A2,$DA,$F0,$3E,$9B,$F5,$36,$7C,$CD,$9F
	.dc.b	$86,$DA,$F8,$ED,$9F,$8A,$D9,$FA,$FF,$3D,$F4,$7E,$7B,$E9,$36,$BE
	.dc.b	$CB,$67,$D4,$DB,$FC,$23,$67,$F1,$2D,$9F,$91,$F2,$FF,$47,$E7,$BE
	.dc.b	$93,$6B,$E7,$36,$7E,$DB,$6F,$EF,$B6,$7E,$EB,$67,$F1,$F5,$F8,$93
	.dc.b	$F1,$5F,$33,$6B,$F2,$2D,$9F,$A4,$DB,$FC,$E3,$C2,$7F,$2F,$76,$BE
	.dc.b	$C7,$CF,$7E,$54,$FC,$57,$C1,$DA,$FC,$37,$C1,$61,$A7,$E2,$63,$F2
	.dc.b	$1F,$09,$23,$F4,$71,$6F,$FC,$27,$F3,$7F,$CE,$38,$3F,$54,$FF,$3A
	.dc.b	$E6,$BE,$02,$6C,$61,$B8,$7F,$99,$2B,$93,$6B,$F6,$17,$D9,$E5,$19
	.dc.b	$BE,$4D,$3F,$6F,$97,$EC,$F2,$FD,$A3,$63,$27,$EF,$23,$2D,$3F,$7F
	.dc.b	$E2,$7E,$EF,$E1,$7F,$78,$D8,$C9,$FC,$12,$32,$D3,$F8,$3F,$CF,$7F
	.dc.b	$03,$F0,$BF,$82,$36,$32,$7F,$0C,$8C,$B4,$CE,$7F,$AF,$8C,$B6,$3F
	.dc.b	$D1,$2B,$E7,$5C,$3F,$CB,$A3,$FB,$84,$62,$7F,$B3,$46,$E7,$4F,$F9
	.dc.b	$38,$C4,$6A,$78,$47,$F9,$38,$C9,$FE,$50,$F9,$07,$D8,$BE,$A7,$F9
	.dc.b	$44,$EB,$9B,$F9,$76,$1F,$C8,$FE,$73,$F9,$23,$4F,$E5,$65,$9F,$E1
	.dc.b	$8D,$9F,$E8,$FC,$7F,$E0,$2E,$7F,$C4,$E9,$F8,$A6,$D7,$E1,$5B,$7F
	.dc.b	$85,$6D,$F2,$E4,$46,$2B,$FC,$DB,$87,$57,$F1,$D8,$7F,$3B,$F8,$97
	.dc.b	$CA,$5F,$9F,$C3,$6B,$FD,$67,$3A,$6C,$E9,$A7,$D6,$F9,$ED,$69,$FE
	.dc.b	$17,$6B,$FE,$25,$6B,$F4,$DA,$FF,$A0,$9E,$3A,$7D,$D6,$CF,$FA,$1B
	.dc.b	$83,$E9,$FA,$FF,$4A,$73,$6D,$3F,$C5,$77,$FB,$2D,$FF,$66,$D3,$F5
	.dc.b	$C5,$9F,$D7,$37,$1F,$94,$D3,$E4,$EF,$FE,$10,$7F,$6D,$5F,$82,$F9
	.dc.b	$C7,$C2,$7E,$3B,$0F,$F4,$DC,$62,$BF,$26,$D8,$FF,$1D,$35,$FF,$24
	.dc.b	$39,$C3,$F3,$8D,$9F,$F5,$2F,$DE,$BE,$F5,$D7,$E1,$C7,$F2,$7F,$3A
	.dc.b	$F9,$DF,$16,$68,$C6,$3F,$A4,$6C,$7F,$92,$DA,$FF,$96,$5C,$E7,$FA
	.dc.b	$D6,$CF,$FA,$DB,$D0,$7D,$9B,$6F,$0E,$3F,$E0,$5F,$79,$F3,$DF,$E0
	.dc.b	$DF,$98,$7C,$F7,$FA,$FB,$2F,$F7,$ED,$69,$FF,$60,$E9,$FE,$28,$AD
	.dc.b	$EB,$CF,$0E,$58,$71,$8F,$F0,$76,$1C,$5F,$F6,$2D,$70,$C3,$AF,$57
	.dc.b	$8C,$F2,$74,$D8,$C9,$2B,$9C,$65,$FE,$04,$7F,$AB,$07,$CD,$BE,$2B
	.dc.b	$F8,$57,$5B,$E1,$7F,$C7,$F3,$9A,$FF,$41,$7A,$FF,$ED,$C8,$C4,$2B
	.dc.b	$86,$39,$C3,$FC,$F9,$76,$BF,$3D,$87,$49,$CC,$78,$9F,$0C,$66,$3B
	.dc.b	$99,$38,$D7,$37,$FD,$B8,$64,$F1,$47,$17,$38,$79,$B0,$F2,$C6,$CF
	.dc.b	$85,$19,$9E,$11,$8B,$F1,$DF,$78,$C1,$E2,$39,$1F,$ED,$93,$C5,$7D
	.dc.b	$AF,$E2,$C7,$03,$CB,$3C,$DE,$11,$E1,$8C,$61,$ED,$AE,$57,$F1,$AB
	.dc.b	$57,$E0,$AD,$61,$E8,$8E,$75,$E8,$74,$79,$23,$53,$C3,$0E,$F3,$96
	.dc.b	$DF,$63,$D3,$1D,$0F,$80,$E2,$79,$47,$68,$E1,$7E,$73,$C4,$EE,$3B
	.dc.b	$1B,$8C,$9B,$7A,$B5,$F8,$E6,$CC,$9F,$84,$8E,$ED,$8E,$6D,$36,$7E
	.dc.b	$09,$CB,$4C,$9F,$57,$0F,$9E,$F1,$3C,$33,$E9,$3B,$BC,$4C,$CF,$9E
	.dc.b	$78,$C7,$E5,$3D,$D7,$CF,$7E,$79,$B1,$F9,$C6,$BB,$8F,$EF,$5E,$F3
	.dc.b	$E7,$5C,$6A,$5F,$0A,$D4,$E6,$31,$5C,$2B,$77,$D2,$31,$3A,$AF,$D5
	.dc.b	$7C,$F3,$E2,$7F,$C1,$5A,$7E,$F2,$31,$8F,$EB,$F2,$EE,$FE,$BD,$B3
	.dc.b	$BA,$E3,$A7,$F8,$4E,$F8,$DE,$F2,$F2,$5F,$F0,$8C,$38,$DE,$F0,$E3
	.dc.b	$7E,$FC,$38,$DE,$F0,$E3,$7E,$BC,$7D,$38,$70,$C1,$87,$0A,$D4,$78
	.dc.b	$6B,$6C,$30,$4C,$3D,$4E,$5F,$04,$7C,$1C,$3C,$33,$E5,$C3,$C1,$87
	.dc.b	$78,$FA,$D8,$FB,$1B,$5C,$1F,$2D,$70,$3D,$13,$B8,$36,$BA,$47,$03
	.dc.b	$C5,$71,$FE,$41,$1A,$37,$3C,$CE,$55,$DE,$78,$61,$EC,$78,$C6,$5F
	.dc.b	$85,$8F,$C8,$9F,$87,$1A,$38,$C6,$4F,$B0,$FC,$77,$0F,$C0,$C3,$0F
	.dc.b	$C7,$38,$99,$3A,$12,$7C,$84,$9E,$B1,$97,$AB,$F1,$B1,$F9,$98,$FC
	.dc.b	$1C,$3E,$5C,$7C,$FC,$7E,$0D,$78,$8E,$18,$F8,$9F,$98,$75,$3C,$98
	.dc.b	$FF,$5F,$F6,$DC,$8F,$9E,$FF,$37,$5B,$87,$F6,$95,$A9,$CC,$62,$B8
	.dc.b	$61,$FE,$04,$F9,$9C,$AB,$E0,$FF,$82,$2F,$31,$88,$FF,$01,$CE,$29
	.dc.b	$F3,$DC,$B9,$DE,$83,$DD,$C3,$70,$FF,$09,$D7,$13,$33,$23,$27,$FA
	.dc.b	$4B,$E7,$9F,$0B,$E6,$84,$61,$D4,$67,$87,$C3,$18,$3F,$D2,$B3,$B7
	.dc.b	$FD,$28,$7F,$19,$70,$E1,$AF,$FA,$67,$D2,$7C,$4F,$85,$B1,$FE,$A6
	.dc.b	$6B,$FE,$9B,$F6,$5F,$0B,$BA,$FF,$52,$CA,$B6,$F8,$9A,$F7,$8D,$FC
	.dc.b	$F7,$AA,$3F,$D4,$CB,$D9,$F8,$4D,$8F,$F5,$BB,$BE,$5F,$F5,$B3,$79
	.dc.b	$7F,$D6,$4E,$E7,$9D,$72,$E0,$91,$CA,$3A,$98,$7F,$D6,$DF,$3D,$FE
	.dc.b	$9E,$7F,$D8,$AD,$73,$EA,$FD,$5F,$9E,$FE,$87,$1A,$8F,$83,$3B,$8F
	.dc.b	$A2,$D7,$FD,$8C,$D7,$F2,$5F,$21,$FE,$6D,$7F,$D4,$67,$E4,$BD,$6F
	.dc.b	$B3,$A7,$E7,$E3,$7E,$4F,$E5,$AB,$84,$7F,$56,$D7,$F4,$77,$C1,$B1
	.dc.b	$BA,$F4,$70,$7A,$77,$93,$FE,$E2,$D3,$ED,$B6,$72,$FF,$B6,$63,$F0
	.dc.b	$39,$7F,$B9,$5D,$C6,$67,$73,$9A,$EF,$3F,$E7,$4F,$90,$FD,$57,$C2
	.dc.b	$C8,$C1,$0F,$F3,$8B,$97,$FC,$E7,$1F,$A3,$F9,$E7,$65,$24,$FF,$BE
	.dc.b	$BE,$6D,$F2,$98,$8F,$D0,$8A,$E2,$FF,$BE,$9A,$E9,$AF,$FB,$EF,$BF
	.dc.b	$F8,$8F,$CD,$7F,$01,$F9,$E9,$3F,$E8,$E8,$FE,$08,$61,$FF,$44,$C6
	.dc.b	$1D,$1F,$E9,$47,$5A,$FD,$BA,$EB,$3F,$E9,$8F,$9E,$FE,$1D,$F3,$B2
	.dc.b	$30,$43,$FD,$2A,$E5,$FF,$4B,$C7,$F1,$5F,$9E,$FE,$2D,$A4,$9F,$F5
	.dc.b	$1C,$65,$FF,$52,$C6,$5F,$F4,$FC,$7F,$0F,$85,$5C,$A5,$7C,$D8,$D1
	.dc.b	$87,$F2,$07,$F8,$96,$53,$A8,$E0,$7D,$57,$86,$3F,$9B,$84,$F8,$1F
	.dc.b	$F5,$BF,$27,$F3,$CE,$6F,$F7,$0C,$60,$FD,$3F,$B1,$FA,$13,$F9,$A7
	.dc.b	$CF,$7E,$8F,$E7,$7F,$49,$F3,$D2,$B8,$6E,$C1,$5F,$CE,$FC,$E9,$5F
	.dc.b	$EC,$87,$83,$FE,$C2,$F9,$EF,$E8,$04,$7F,$D9,$BF,$3D,$5F,$F6,$3F
	.dc.b	$2F,$F4,$53,$1B,$FD,$DF,$2F,$F2,$73,$FE,$D4,$9D,$3F,$ED,$6B,$FD
	.dc.b	$D7,$9F,$FB,$BF,$9E,$93,$E8,$D3,$FC,$73,$8F,$E1,$DB,$C3,$F1,$6B
	.dc.b	$75,$F4,$B6,$F7,$61,$D2,$39,$4E,$B0,$E5,$3F,$D6,$FD,$DF,$EB,$9E
	.dc.b	$BC,$BF,$F7,$4C,$65,$FF,$75,$C6,$5F,$F7,$2C,$65,$FF,$73,$C6,$6B
	.dc.b	$E0,$4F,$F1,$5F,$86,$FE,$2D,$F2,$F2,$FD,$91,$97,$DD,$19,$7D,$5A
	.dc.b	$27,$51,$89,$D5,$FC,$39,$FE,$EF,$F3,$9F,$C8,$A3,$E8,$57,$F7,$73
	.dc.b	$3E,$24,$7F,$07,$C3,$FC,$95,$FF,$01,$37,$65,$F3,$DF,$FC,$1A,$DA
	.dc.b	$D9,$D4,$FC,$6B,$99,$78,$1F,$E0,$19,$FA,$8B,$E5,$FF,$C1,$AD,$8F
	.dc.b	$CA,$BB,$CB,$F8,$23,$70,$FF,$0E,$BB,$9E,$75,$CB,$D1,$91,$CA,$3A
	.dc.b	$98,$7F,$09,$8B,$E7,$BE,$3F,$CF,$49,$F4,$C7,$CA,$30,$F0,$8C,$3E
	.dc.b	$EF,$1F,$FD,$44,$BD,$D7,$0F,$C6,$9C,$C6,$27,$FC,$5D,$87,$F4,$68
	.dc.b	$C6,$DF,$3B,$C8,$DE,$1C,$6F,$FD,$5D,$5A,$BE,$13,$99,$F0,$3F,$52
	.dc.b	$B8,$C7,$FA,$D7,$E7,$1F,$25,$FD,$D3,$0F,$14,$A7,$3E,$EF,$6C,$3D
	.dc.b	$71,$89,$EA,$7F,$AF,$FE,$75,$F3,$DD,$A3,$96,$3D,$6F,$8B,$63,$C4
	.dc.b	$D5,$5C,$9F,$01,$AC,$2B,$D1,$B8,$78,$F1,$F4,$D7,$1A,$F2,$9B,$2C
	.dc.b	$EF,$0F,$1D,$F4,$AE,$17,$F0,$AB,$57,$E1,$38,$E1,$E7,$AD,$E1,$CA
	.dc.b	$73,$87,$92,$B5,$8F,$C3,$C3,$9D,$78,$F1,$ED,$5A,$8E,$17,$FD,$C9
	.dc.b	$F4,$1F,$E5,$4A,$FA,$0D,$9F,$F1,$5B,$FE,$58,$F7,$BF,$CC,$7F,$1B
	.dc.b	$C5,$B1,$FD,$F1,$AE,$FD,$B4,$F7,$BF,$F2,$97,$65,$F6,$0D,$25,$FF
	.dc.b	$81,$A7,$6F,$CC,$9D,$BF,$43,$E7,$BF,$CB,$FD,$3F,$5A,$73,$AF,$CD
	.dc.b	$6E,$FF,$8C,$6D,$85,$F2,$7B,$63,$CC,$8F,$7E,$4F,$F2,$06,$1D,$67
	.dc.b	$2D,$9D,$63,$8C,$7C,$15,$67,$C1,$87,$7B,$E0,$4A,$DC,$F2,$3C,$64
	.dc.b	$8D,$3C,$B0,$F1,$CF,$09,$FD,$B7,$93,$8A,$FD,$FF,$0D,$E5,$FE,$57
	.dc.b	$37,$97,$F5,$7A,$7E,$8A,$72,$D8,$CD,$FE,$CA,$72,$D8,$CE,$1F,$C1
	.dc.b	$7C,$8F,$F3,$27,$67,$F8,$AE,$51,$8C,$3F,$9C,$F9,$17,$0F,$E2,$73
	.dc.b	$71,$FE,$7E,$DB,$E3,$7F,$CF,$2F,$13,$BF,$C4,$DF,$2F,$4C,$7C,$FD
	.dc.b	$BF,$F0,$7C,$E9,$B3,$A6,$93,$2F,$EF,$13,$B6,$D7,$37,$F9,$CF,$BF
	.dc.b	$E5,$3C,$DC,$F6,$FF,$A8,$C5,$9F,$E7,$F3,$75,$F8,$93,$6B,$FA,$DC
	.dc.b	$DA,$FF,$71,$60,$F9,$CF,$71,$EF,$F9,$DB,$7E,$C8,$B5,$E8,$9B,$8F
	.dc.b	$F9,$6E,$F9,$5E,$CF,$87,$3A,$7F,$B5,$F9,$3F,$3C,$FA,$1E,$55,$8F
	.dc.b	$ED,$71,$70,$FE,$B3,$37,$0F,$BD,$36,$BF,$BA,$CD,$AF,$5F,$CE,$7D
	.dc.b	$73,$EC,$7B,$B6,$BE,$AC,$59,$FA,$33,$70,$F5,$4D,$AF,$DC,$71,$B3
	.dc.b	$5E,$70,$FE,$BB,$F4,$5F,$89,$F9,$9F,$C5,$37,$0F,$F0,$2B,$87,$FC
	.dc.b	$0F,$E6,$5A,$FE,$EA,$FE,$44,$FC,$97,$55,$9F,$C5,$FC,$07,$E5,$7E
	.dc.b	$2B,$F2,$4D,$8F,$C7,$FC,$17,$E5,$4F,$CB,$78,$36,$3F,$27,$E2,$FE
	.dc.b	$64,$FC,$F7,$8D,$63,$FC,$D9,$F2,$FF,$CA,$39,$BF,$4C,$D8,$CF,$8C
	.dc.b	$F9,$FC,$AA,$A4,$9D,$CD,$DF,$FD,$0B,$7C,$DB,$7C,$6B,$67,$6B,$FD
	.dc.b	$A7,$E2,$8F,$3E,$F9,$F8,$6D,$7F,$9F,$27,$4D,$9D,$34,$FF,$44,$60
	.dc.b	$F9,$D7,$CE,$7F,$03,$F2,$DF,$4B,$FE,$1A,$BD,$D6,$9F,$F4,$84,$EB
	.dc.b	$97,$F8,$56,$1C,$F0,$FF,$4D,$3F,$E2,$4B,$E0,$FF,$B4,$1A,$FF,$8A
	.dc.b	$0E,$17,$FE,$D4,$EB,$67,$FC,$07,$B7,$FD,$88,$BC,$3F,$87,$E2,$7C
	.dc.b	$46,$DF,$C1,$6E,$1F,$E8,$1A,$E1,$3F,$16,$B8,$34,$E0,$7F,$AA,$FE
	.dc.b	$45,$F8,$AB,$ED,$BF,$91,$1F,$C9,$E3,$2E,$27,$84,$ED,$B1,$F6,$3E
	.dc.b	$8B,$84,$E7,$1F,$27,$27,$DC,$E7,$FB,$AD,$C3,$ED,$36,$77,$1F,$76
	.dc.b	$7F,$1F,$F1,$D2,$3F,$27,$E2,$5A,$FC,$5C,$DD,$D5,$FD,$BF,$A7,$FC
	.dc.b	$07,$79,$8C,$74,$AF,$18,$2F,$75,$FA,$09,$FB,$FF,$40,$F9,$7F,$E9
	.dc.b	$59,$3E,$9F,$E7,$E3,$FE,$E2,$15,$BC,$3F,$DB,$15,$BE,$4E,$78,$7E
	.dc.b	$22,$FF,$C0,$97,$B7,$F0,$C7,$28,$D3,$6F,$FC,$55,$3F,$EE,$E7,$F2
	.dc.b	$BD,$4E,$DF,$3F,$C3,$7E,$76,$31,$3F,$A2,$AE,$17,$E7,$6E,$5F,$22
	.dc.b	$3F,$C4,$BF,$39,$FE,$97,$AE,$18,$7F,$04,$97,$CA,$53,$1F,$2D,$6B
	.dc.b	$F8,$5C,$E9,$B3,$A6,$92,$3F,$82,$9B,$2C,$F3,$7E,$67,$8F,$FB,$E5
	.dc.b	$8F,$9F,$C0,$65,$FD,$42,$77,$E1,$7E,$F6,$31,$5E,$D9,$FF,$66,$4F
	.dc.b	$EE,$AF,$A4,$F3,$A7,$D2,$66,$BF,$7F,$F8,$BF,$2C,$FF,$C7,$59,$3F
	.dc.b	$17,$F4,$63,$D6,$B5,$9F,$F9,$3A,$73,$7F,$EA,$89,$CC,$9D,$DF,$3F
	.dc.b	$93,$EF,$E9,$7F,$22,$CE,$61,$FC,$7E,$2E,$1F,$BF,$9B,$97,$F0,$89
	.dc.b	$B5,$FD,$63,$C2,$FE,$AE,$FF,$27,$F3,$FF,$96,$F8,$DF,$CB,$99,$19
	.dc.b	$3F,$96,$46,$31,$FC,$E1,$FE,$B4,$1F,$34,$F8,$CF,$F6,$1F,$8A,$F8
	.dc.b	$9F,$C5,$4E,$6B,$F9,$07,$B7,$FE,$61,$8C,$42,$B8,$63,$9C,$3F,$43
	.dc.b	$76,$BF,$CD,$15,$A7,$C3,$3A,$69,$AC,$3C,$31,$9C,$3B,$D7,$17,$C3
	.dc.b	$5E,$B7,$35,$C2,$3C,$E7,$43,$CC,$E5,$FF,$6E,$99,$9E,$95,$DC,$F8
	.dc.b	$73,$B9,$ED,$18,$C3,$D9,$1C,$AF,$C3,$18,$8D,$9E,$B7,$0F,$FB,$BC
	.dc.b	$F0,$57,$57,$13,$C9,$F4,$1C,$8F,$4C,$F2,$C3,$C3,$87,$A2,$7D,$F5
	.dc.b	$C4,$F8,$AF,$4C,$3D,$0F,$43,$07,$94,$C5,$74,$C3,$C8,$DC,$3C,$13
	.dc.b	$E8,$6D,$F7,$BF,$91,$3C,$E3,$C3,$5C,$63,$A9,$C4,$B1,$C5,$F4,$CF
	.dc.b	$9A,$38,$C6,$DA,$6D,$FA,$11,$B3,$C2,$FC,$63,$AB,$97,$A9,$E5,$77
	.dc.b	$87,$C9,$8E,$66,$4D,$9C,$E7,$C3,$3C,$E3,$66,$5E,$78,$7C,$88,$FE
	.dc.b	$75,$E8,$FE,$77,$B0,$D2,$47,$FB,$2F,$E0,$5E,$CF,$F2,$48,$E2,$65
	.dc.b	$C5,$72,$8C,$70,$FF,$B4,$E3,$10,$AE,$73,$9B,$FF,$6C,$4E,$6F,$B6
	.dc.b	$5F,$BA,$F9,$B7,$B3,$A8,$C3,$74,$E3,$97,$4C,$3F,$58,$DA,$DC,$7E
	.dc.b	$92,$F9,$BF,$A6,$9D,$57,$2C,$3C,$2F,$92,$B7,$5F,$B7,$6D,$FF,$0E
	.dc.b	$BE,$33,$F0,$0D,$61,$E5,$7F,$DE,$95,$C2,$FB,$65,$E1,$C7,$BD,$6A
	.dc.b	$F8,$2B,$E0,$F5,$C7,$C1,$5C,$23,$55,$C2,$B7,$1F,$2E,$FB,$1F,$06
	.dc.b	$74,$76,$C7,$D3,$3B,$3D,$26,$CE,$27,$CC,$BF,$05,$F6,$3B,$CF,$23
	.dc.b	$9B,$9A,$ED,$3E,$79,$CC,$78,$AB,$AE,$1E,$CC,$3C,$11,$E2,$3C,$0E
	.dc.b	$B1,$FA,$95,$C6,$72,$4A,$FB,$EF,$17,$A3,$99,$E2,$F6,$8C,$92,$33
	.dc.b	$1C,$5F,$AE,$F1,$3E,$B1,$C5,$CC,$71,$8C,$9C,$4F,$B1,$1A,$3A,$C7
	.dc.b	$98,$E1,$1F,$45,$C4,$68,$E1,$0F,$3B,$FC,$F7,$E2,$FE,$A3,$0F,$F3
	.dc.b	$DB,$78,$7F,$BA,$36,$B8,$72,$3C,$8C,$61,$CA,$7F,$D3,$D5,$8A,$FF
	.dc.b	$2C,$47,$FA,$86,$B1,$EA,$C8,$FF,$52,$3B,$FA,$1C,$47,$FA,$16,$2E
	.dc.b	$5F,$E8,$E9,$F1,$4B,$E3,$3F,$C7,$AF,$FA,$A1,$AF,$F9,$75,$B3,$FE
	.dc.b	$2B,$FC,$63,$FC,$A1,$3B,$F1,$39,$C3,$C3,$A7,$6B,$E3,$B7,$F0,$CF
	.dc.b	$C6,$5F,$2B,$E2,$6E,$3F,$EB,$86,$BF,$E6,$DF,$C6,$9D,$FB,$AF,$4C
	.dc.b	$7F,$B5,$67,$FD,$7E,$D7,$FC,$E9,$5C,$B4,$FF,$21,$7D,$CB,$E5,$BE
	.dc.b	$0B,$6B,$FD,$8E,$D7,$FC,$F8,$BC,$AD,$CC,$9D,$BC,$A3,$11,$C2,$2D
	.dc.b	$78,$27,$E7,$FC,$EB,$D8,$A4,$BF,$F6,$BF,$CF,$7E,$0D,$FF,$6C,$CE
	.dc.b	$DC,$F8,$9F,$62,$7F,$DB,$71,$AA,$EB,$E2,$FF,$A2,$A0,$2D,$19,$8C
	.dc.b	$36,$33,$1C,$5A,$FF,$A7,$7E,$77,$EE,$3F,$EE,$C6,$BF,$EA,$18,$DE
	.dc.b	$5F,$E6,$F9,$FE,$AB,$3B,$F0,$3F,$1F,$18,$8E,$9F,$0B,$89,$91,$B3
	.dc.b	$B4,$61,$F1,$45,$AE,$D3,$FA,$5F,$7F,$F4,$DF,$03,$2F,$E0,$D7,$FB
	.dc.b	$B2,$75,$18,$E5,$FF,$33,$56,$E3,$FD,$D5,$3F,$B5,$F2,$1F,$8D,$FB
	.dc.b	$2F,$FC,$A6,$DA,$FF,$2D,$37,$C1,$FF,$7D,$FC,$17,$18,$C7,$C4,$BE
	.dc.b	$F7,$F7,$AD,$BF,$F7,$94,$F1,$F0,$DF,$4D,$FB,$FF,$AC,$95,$E7,$3F
	.dc.b	$83,$C7,$F0,$33,$2F,$FC,$01,$8C,$3B,$C9,$5D,$23,$0A,$78,$EF,$FE
	.dc.b	$C1,$AD,$DF,$F0,$4F,$A0,$7D,$0F,$E1,$DB,$8F,$E3,$5A,$FF,$B6,$1A
	.dc.b	$FF,$82,$03,$AB,$FF,$59,$7C,$FB,$E7,$BF,$88,$B6,$BD,$C7,$0B,$FE
	.dc.b	$47,$F0,$8F,$A7,$FE,$45,$C5,$2B,$DD,$39,$6C,$E5,$FF,$63,$CF,$FA
	.dc.b	$8C,$E7,$1E,$0F,$A0,$93,$E0,$C1,$1E,$4F,$A1,$CC,$FF,$BF,$67,$2E
	.dc.b	$E7,$4D,$21,$B7,$2F,$88,$FE,$8D,$F8,$E3,$E5,$BF,$9F,$B6,$3E,$C3
	.dc.b	$5E,$1C,$0C,$DF,$4F,$C3,$E5,$65,$70,$79,$FC,$BF,$4F,$97,$F9,$3F
	.dc.b	$11,$2B,$AC,$7C,$CF,$86,$91,$F2,$9F,$9F,$C5,$F4,$3E,$06,$4F,$CE
	.dc.b	$76,$F9,$33,$47,$D6,$F4,$E4,$EC,$F5,$FB,$5E,$E2,$3E,$EF,$97,$AF
	.dc.b	$BB,$83,$FC,$77,$D1,$F9,$36,$F3,$FF,$1A,$E2,$42,$F8,$D6,$F0,$B7
	.dc.b	$BA,$D6,$1F,$86,$FA,$4F,$C5,$91,$FC,$4F,$D0,$57,$D9,$17,$0F,$D2
	.dc.b	$6A,$AC,$61,$E3,$85,$62,$FF,$C0,$51,$B7,$ED,$FA,$DF,$78,$8F,$DE
	.dc.b	$F9,$5A,$FC,$18,$B8,$FC,$19,$B8,$7A,$E6,$D7,$E2,$A7,$F4,$BE,$8C
	.dc.b	$8F,$D1,$BF,$A8,$F4,$71,$1F,$5E,$2E,$1E,$C9,$B8,$7A,$E6,$D7,$AA
	.dc.b	$7F,$C2,$7F,$8E,$BE,$2A,$92,$7F,$87,$43,$E5,$F1,$1F,$98,$8E,$0E
	.dc.b	$A7,$9C,$EB,$1F,$E2,$15,$AA,$E1,$5A,$7F,$7B,$DD,$FC,$08,$8F,$F0
	.dc.b	$0F,$62,$BF,$93,$8B,$7F,$62,$7F,$84,$F7,$FF,$A6,$BE,$1D,$B7,$D9
	.dc.b	$F0,$70,$DC,$7F,$B3,$CE,$9F,$B7,$5A,$69,$A9,$E0,$A7,$87,$3F,$ED
	.dc.b	$EF,$83,$FB,$9B,$E5,$4F,$9D,$79,$DF,$E2,$AC,$3C,$13,$98,$C6,$3F
	.dc.b	$EB,$2A,$DF,$C3,$F8,$6B,$FB,$EF,$CF,$7F,$07,$C3,$8D,$6A,$7C,$97
	.dc.b	$C7,$0E,$71,$FE,$04,$9D,$57,$82,$F9,$5E,$F0,$ED,$7E,$36,$DF,$09
	.dc.b	$FF,$62,$D7,$0A,$E9,$87,$67,$81,$E2,$33,$7D,$27,$84,$6B,$0F,$4D
	.dc.b	$71,$C3,$B4,$FB,$BC,$5F,$93,$18,$8E,$18,$F9,$DE,$C3,$0E,$B3,$B2
	.dc.b	$5F,$96,$32,$FF,$96,$A7,$FA,$C4,$FC,$96,$C6,$5E,$C7,$F5,$DF,$86
	.dc.b	$7A,$6F,$94,$FF,$18,$46,$1B,$19,$9F,$EC,$87,$FB,$BD,$C3,$E0,$F4
	.dc.b	$9F,$19,$FE,$39,$C5,$18,$9E,$31,$89,$D4,$E6,$79,$47,$D6,$8E,$1A
	.dc.b	$FE,$33,$E7,$9F,$3C,$EC,$D5,$CC,$3E,$4E,$3D,$5F,$97,$7C,$2B,$58
	.dc.b	$7E,$3B,$1E,$BB,$FE,$55,$B7,$FE,$50,$BE,$0F,$48,$CB,$FA,$16,$E1
	.dc.b	$DE,$FC,$D9,$79,$6B,$86,$5E,$0C,$7B,$3F,$94,$C7,$AD,$F8,$E3,$55
	.dc.b	$F9,$DC,$7B,$D6,$AF,$8C,$F4,$25,$F8,$32,$E3,$C6,$F5,$1E,$5F,$8E
	.dc.b	$BF,$7B,$1F,$DE,$6B,$F6,$55,$BA,$D6,$1F,$D5,$EB,$58,$7E,$86,$B5
	.dc.b	$7C,$27,$38,$74,$C3,$8D,$FF,$B2,$E7,$38,$73,$F3,$3E,$27,$4F,$F8
	.dc.b	$DE,$51,$8C,$7C,$1E,$5D,$C3,$94,$DC,$FD,$8D,$BE,$37,$FE,$47,$57
	.dc.b	$99,$E1,$F3,$DF,$E9,$7A,$FF,$26,$37,$0F,$F0,$AA,$F9,$5D,$FB,$D2
	.dc.b	$B6,$F2,$8C,$57,$08,$B9,$FF,$94,$6F,$95,$F0,$3C,$9E,$DF,$9C,$F4
	.dc.b	$72,$D8,$C9,$30,$FF,$13,$E9,$FB,$48,$C6,$1F,$1E,$F8,$56,$AE,$CF
	.dc.b	$87,$E1,$3D,$E7,$C0,$F1,$EE,$1F,$02,$2D,$FC,$09,$B5,$C6,$B4,$7C
	.dc.b	$68,$CB,$F8,$3F,$2F,$E8,$1F,$47,$CC,$B1,$F4,$62,$DF,$AA,$6D,$FD
	.dc.b	$F9,$B5,$F8,$D9,$B4,$E1,$6A,$F8,$9F,$B2,$7D,$AF,$80,$B8,$7D,$A8
	.dc.b	$B7,$F6,$A6,$E5,$F4,$E6,$D7,$F8,$C6,$FF,$D9,$2F,$13,$F1,$1F,$3B
	.dc.b	$F8,$BF,$9E,$FC,$63,$63,$24,$C7,$EA,$F9,$BA,$8F,$C8,$3F,$93,$3F
	.dc.b	$29,$CF,$63,$F1,$FF,$17,$F9,$7F,$90,$FC,$9B,$63,$F2,$7F,$05,$F9
	.dc.b	$63,$F2,$FD,$D6,$3F,$2F,$E2,$7E,$68,$FD,$07,$8F,$63,$FD,$59,$F3
	.dc.b	$3F,$CE,$79,$BF,$50,$D8,$CF,$8A,$F9,$AC,$BF,$E7,$38,$CD,$60,$CC
	.dc.b	$D8,$C9,$37,$FF,$41,$5F,$F4,$79,$D4,$7F,$9F,$BE,$07,$FC,$5D,$5F
	.dc.b	$EB,$6F,$47,$F7,$11,$89,$D4,$63,$D0,$CB,$6B,$81,$FE,$AA,$F9,$E7
	.dc.b	$C4,$FF,$2B,$BF,$F5,$A4,$62,$7F,$A9,$C2,$30,$FF,$56,$6B,$C2,$7F
	.dc.b	$86,$3B,$7F,$87,$7C,$33,$EB,$9D,$1A,$BE,$7D,$7F,$EC,$DD,$3F,$84
	.dc.b	$37,$C5,$F6,$F8,$BE,$E6,$E9,$FE,$D6,$6E,$9F,$15,$B5,$C6,$7F,$DB
	.dc.b	$75,$C5,$FF,$73,$66,$F9,$D7,$CF,$F9,$9B,$3F,$EE,$2F,$9F,$C4,$C8
	.dc.b	$E3,$3A,$CB,$EA,$56,$5F,$AD,$C9,$F5,$38,$FE,$AB,$6F,$FD,$DA,$D9
	.dc.b	$DC,$7F,$BC,$27,$F1,$3F,$29,$23,$F1,$BE,$0D,$AF,$C2,$F3,$AB,$1B
	.dc.b	$7F,$63,$E7,$FF,$07,$5E,$30,$60,$9F,$F0,$95,$E6,$31,$87,$F0,$3A
	.dc.b	$FC,$DC,$FD,$AF,$9D,$7C,$27,$E1,$6F,$98,$8C,$D7,$E1,$9B,$1F,$E1
	.dc.b	$A6,$BF,$EF,$F7,$39,$7E,$3D,$B3,$E6,$E0,$FC,$37,$83,$E1,$C3,$C5
	.dc.b	$EE,$4A,$F5,$E1,$FE,$01,$F4,$9F,$3B,$FE,$8D,$9C,$9E,$6C,$97,$C7
	.dc.b	$3F,$F7,$71,$E9,$7F,$B1,$5F,$2E,$CD,$63,$F2,$E9,$F0,$5F,$C0,$1B
	.dc.b	$8F,$F0,$45,$F0,$B9,$99,$39,$78,$46,$27,$F7,$A7,$22,$C6,$A7,$81
	.dc.b	$EA,$8F,$B3,$AF,$EF,$6F,$F6,$CD,$9F,$BB,$8F,$C2,$F9,$BF,$D8,$CE
	.dc.b	$6F,$8F,$81,$83,$DF,$B2,$B1,$87,$7A,$7C,$E3,$CF,$FD,$DB,$67,$C9
	.dc.b	$93,$E7,$7F,$DF,$B1,$C6,$75,$7F,$39,$BA,$7F,$0D,$6C,$FA,$2B,$8D
	.dc.b	$6F,$0F,$E3,$78,$CA,$F8,$5E,$73,$C0,$E3,$7F,$D6,$DF,$54,$E9,$B1
	.dc.b	$FC,$9E,$BF,$92,$DF,$38,$FE,$55,$E7,$7F,$28,$3F,$94,$F8,$B7,$0E
	.dc.b	$51,$67,$F9,$84,$DD,$3F,$45,$1F,$EB,$B1,$F3,$6F,$9B,$FD,$2F,$4B
	.dc.b	$E7,$7F,$CA,$D3,$9A,$FD,$49,$C5,$C2,$9E,$19,$7E,$E2,$73,$3F,$ED
	.dc.b	$03,$BD,$F6,$BE,$AF,$FA,$23,$DA,$E0,$E6,$ED,$78,$26,$BD,$EB,$8B
	.dc.b	$93,$A9,$FB,$37,$2F,$FB,$94,$C9,$C6,$79,$BE,$5B,$F5,$E1,$EA,$8D
	.dc.b	$99,$C3,$C8,$E0,$F8,$4E,$27,$DF,$18,$3C,$55,$D9,$C1,$E1,$39,$61
	.dc.b	$EC,$BF,$05,$72,$8C,$61,$EA,$AF,$11,$D0,$E2,$7B,$2B,$8E,$1F,$2E
	.dc.b	$31,$3C,$5F,$0C,$E9,$F9,$17,$F0,$27,$57,$E0,$C3,$D0,$DA,$EF,$5B
	.dc.b	$9C,$C7,$66,$C6,$63,$78,$7C,$53,$C9,$19,$3D,$D1,$C4,$D9,$ED,$72
	.dc.b	$6E,$38,$9C,$E3,$D2,$D7,$EF,$46,$DC,$9C,$E3,$8B,$4C,$9F,$4E,$3B
	.dc.b	$C7,$38,$F9,$E6,$5E,$2D,$38,$E1,$F2,$DF,$0C,$71,$7F,$5D,$E5,$3C
	.dc.b	$FF,$A8,$D8,$FC,$C3,$5F,$CB,$37,$DE,$7A,$ED,$7F,$B1,$CE,$63,$14
	.dc.b	$8C,$4E,$AF,$FA,$FD,$F2,$8C,$57,$F6,$C8,$C3,$FA,$4F,$9D,$7A,$FF
	.dc.b	$D9,$B4,$FD,$EC,$62,$7F,$63,$5C,$27,$F8,$0B,$73,$FE,$DB,$5B,$AE
	.dc.b	$97,$CE,$7C,$79,$FA,$30,$E2,$D8,$E3,$3B,$AF,$F0,$0D,$F3,$BD,$E3
	.dc.b	$E2,$C3,$85,$6A,$F8,$2B,$E1,$87,$82,$F8,$1F,$C2,$A3,$33,$C4,$C9
	.dc.b	$FC,$1B,$1F,$63,$93,$D6,$6B,$1F,$1B,$82,$38,$25,$FB,$23,$C1,$87
	.dc.b	$9A,$BA,$C1,$B7,$E4,$8F,$54,$EF,$1F,$7C,$6D,$B3,$B3,$85,$74,$39
	.dc.b	$D7,$C4,$8C,$CF,$29,$CD,$79,$0F,$65,$FC,$77,$D6,$FA,$5F,$89,$3C
	.dc.b	$4F,$0D,$F5,$3C,$D8,$F2,$9F,$91,$8F,$E2,$A7,$31,$CA,$7C,$0F,$63
	.dc.b	$F2,$95,$D6,$33,$5C,$67,$A1,$F0,$63,$37,$E5,$9F,$9F,$3C,$A7,$2F
	.dc.b	$8A,$7D,$93,$A8,$F9,$A6,$8E,$87,$F4,$AE,$5F,$EF,$1E,$17,$F9,$22
	.dc.b	$FF,$CF,$CD,$DF,$FC,$AB,$3A,$AF,$EF,$5E,$2E,$5F,$EF,$D5,$67,$51
	.dc.b	$FE,$A3,$8C,$73,$FF,$B6,$E3,$FC,$05,$E9,$E0,$8F,$F8,$17,$E7,$6B
	.dc.b	$FE,$AB,$8B,$97,$FA,$46,$7C,$12,$F9,$CF,$F2,$0B,$FE,$25,$6B,$FE
	.dc.b	$4A,$6C,$FF,$AA,$7E,$77,$F8,$54,$EF,$C2,$E7,$5A,$D7,$FC,$D2,$D9
	.dc.b	$FF,$3A,$D7,$07,$FC,$AB,$F3,$8F,$88,$F1,$B6,$3F,$C6,$CD,$7F,$CB
	.dc.b	$7F,$3B,$FE,$37,$73,$3C,$08,$FF,$8F,$63,$4F,$F9,$8B,$E7,$5F,$1B
	.dc.b	$F0,$1B,$1F,$E4,$36,$BB,$F2,$B5,$32,$34,$F1,$8C,$47,$58,$BA,$FF
	.dc.b	$B3,$6B,$FC,$E7,$3F,$E4,$FF,$9C,$7A,$BF,$41,$B1,$98,$FA,$7F,$3B
	.dc.b	$F7,$63,$FC,$F3,$3A,$6C,$F1,$37,$CD,$CF,$C3,$93,$DD,$FB,$BF,$3B
	.dc.b	$F3,$F2,$FF,$70,$60,$9F,$F6,$FD,$23,$0F,$FB,$89,$AF,$F8,$E8,$D4
	.dc.b	$70,$AF,$E9,$F1,$FE,$6B,$6B,$FE,$90,$6D,$7F,$BB,$D7,$C1,$F9,$3B
	.dc.b	$25,$71,$7C,$F1,$88,$E5,$16,$B9,$4F,$E8,$BD,$DF,$D1,$FB,$F2,$FF
	.dc.b	$CC,$91,$97,$FC,$CD,$19,$7F,$B2,$73,$7E,$CB,$A9,$5F,$B1,$F1,$1F
	.dc.b	$39,$F0,$32,$5F,$FA,$DE,$30,$DD,$7F,$67,$F0,$1C,$67,$39,$7F,$97
	.dc.b	$5B,$87,$4A,$DC,$FF,$9E,$7E,$6F,$F7,$7F,$39,$23,$04,$3C,$E7,$FA
	.dc.b	$05,$D3,$8F,$07,$FD,$0D,$3F,$C1,$0C,$E7,$E8,$9E,$07,$37,$0B,$E7
	.dc.b	$1F,$33,$FC,$09,$B1,$FE,$A5,$6B,$FE,$CA,$F9,$47,$D3,$6E,$BF,$13
	.dc.b	$5A,$AF,$F3,$FB,$5F,$F7,$4B,$56,$2F,$7D,$F3,$9F,$C7,$38,$65,$7F
	.dc.b	$A8,$A7,$2D,$9C,$BF,$07,$8F,$D7,$A7,$BF,$0E,$95,$FC,$73,$47,$CA
	.dc.b	$D2,$61,$FE,$C1,$87,$CB,$62,$3D,$D3,$70,$F8,$F5,$A9,$E1,$39,$CD
	.dc.b	$12,$7B,$BD,$6F,$F9,$D7,$CD,$3E,$5E,$92,$7D,$DF,$39,$AF,$92,$91
	.dc.b	$FE,$C2,$A4,$72,$F9,$4E,$7F,$29,$FD,$17,$85,$F0,$B4,$CE,$1F,$27
	.dc.b	$E0,$F4,$47,$E6,$72,$F1,$F9,$C9,$1F,$3D,$D3,$F8,$DB,$47,$7F,$3E
	.dc.b	$4E,$CF,$37,$AB,$EE,$23,$E9,$F9,$4A,$FE,$5F,$6F,$F1,$F7,$2F,$D7
	.dc.b	$6F,$27,$F6,$45,$5C,$27,$57,$6B,$53,$F0,$7E,$6E,$61,$F8,$77,$E2
	.dc.b	$7C,$E6,$23,$D5,$17,$0F,$C3,$68,$AC,$61,$FD,$D2,$15,$8B,$FC,$BC
	.dc.b	$6D,$FA,$9F,$0F,$F7,$08,$FD,$7F,$94,$AF,$BE,$2E,$3E,$F9,$B8,$7B
	.dc.b	$E6,$D7,$E9,$34,$57,$DD,$F3,$E4,$FE,$8D,$FC,$07,$A1,$88,$FA,$D1
	.dc.b	$71,$F6,$CD,$C7,$D9,$36,$BD,$73,$FE,$0D,$F9,$C7,$C4,$D2,$4F,$E7
	.dc.b	$30,$FD,$64,$BE,$2B,$33,$FC,$46,$38,$BB,$C7,$FC,$3B,$18,$8D,$4F
	.dc.b	$0A,$E7,$76,$F7,$3F,$9B,$ED,$97,$FC,$05,$FC,$F7,$AB,$88,$FC,$A4
	.dc.b	$5C,$3E,$DC,$FE,$8F,$BF,$F6,$74,$F8,$2F,$96,$FD,$AF,$03,$31,$8F
	.dc.b	$0B,$F8,$94,$63,$83,$F0,$57,$6F,$F0,$2D,$AD,$C7,$F0,$0B,$E5,$C7
	.dc.b	$FC,$16,$3C,$17,$C9,$D3,$E4,$F9,$BF,$E9,$FF,$07,$89,$FE,$13,$18
	.dc.b	$78,$98,$F3,$BF,$C5,$B8,$7A,$A1,$F3,$9F,$C2,$DE,$42,$B9,$4E,$DB
	.dc.b	$87,$86,$F9,$BF,$1A,$F9,$5F,$80,$E5,$3A,$7C,$58,$76,$BE,$71,$FC
	.dc.b	$82,$B8,$E1,$DE,$F8,$4F,$2B,$E1,$18,$AD,$E1,$E3,$9C,$92,$BC,$53
	.dc.b	$A7,$C2,$79,$70,$F0,$BD,$4D,$F8,$BF,$5D,$C6,$3E,$3B,$E8,$FC,$9B
	.dc.b	$E9,$3C,$4F,$C3,$94,$F2,$1F,$E1,$FB,$FA,$8E,$BE,$0B,$FC,$33,$F1
	.dc.b	$DF,$E2,$38,$FF,$59,$7A,$68,$C4,$EA,$7F,$C5,$0D,$7B,$D3,$D2,$7C
	.dc.b	$67,$38,$C3,$63,$33,$E3,$6B,$C9,$CE,$5F,$7D,$BA,$FE,$8B,$E6,$DF
	.dc.b	$39,$FE,$03,$AD,$61,$F6,$EF,$78,$7E,$49,$AA,$F8,$F4,$EE,$A4,$FD
	.dc.b	$3A,$DD,$6A,$31,$3A,$BE,$F1,$F8,$D9,$D4,$60,$98,$FE,$96,$F9,$E3
	.dc.b	$E7,$8C,$61,$CD,$B8,$73,$9D,$3E,$99,$D3,$4D,$63,$81,$F0,$1F,$C6
	.dc.b	$DF,$D9,$18,$EA,$E7,$CB,$FE,$15,$F8,$6E,$B3,$93,$75,$C6,$FF,$6F
	.dc.b	$19,$7F,$73,$5C,$63,$77,$E6,$9F,$99,$C7,$FE,$35,$EA,$FF,$1B,$71
	.dc.b	$3E,$81,$E3,$E7,$1F,$B5,$77,$0F,$C7,$56,$A7,$F9,$84,$7F,$84,$B8
	.dc.b	$95,$FE,$3E,$8E,$33,$F9,$79,$CB,$67,$53,$FE,$1A,$FC,$17,$D9,$97
	.dc.b	$93,$0F,$EC,$77,$DA,$95,$A8,$FF,$10,$E5,$E9,$F1,$7C,$D5,$AD,$3D
	.dc.b	$13,$9A,$E5,$7C,$6B,$D7,$7C,$23,$11,$C6,$7D,$B4,$C7,$FC,$27,$1C
	.dc.b	$B0,$E1,$EB,$F6,$F6,$DF,$39,$8A,$E3,$18,$7C,$33,$B9,$D6,$3F,$E6
	.dc.b	$EC,$57,$B8,$FF,$38,$9F,$02,$B7,$8F,$D8,$A7,$97,$8C,$7F,$A8,$E6
	.dc.b	$F7,$FC,$9F,$17,$ED,$6C,$FC,$26,$E5,$DA,$B8,$4F,$A2,$B8,$3B,$3E
	.dc.b	$1E,$1F,$E0,$67,$15,$E6,$FC,$0B,$FD,$BB,$1F,$23,$F0,$4F,$56,$7F
	.dc.b	$E8,$09,$CD,$7B,$E7,$35,$FD,$A2,$73,$5F,$94,$9C,$D7,$F9,$7A,$73
	.dc.b	$5F,$80,$9C,$CA,$72,$EF,$CB,$FC,$1F,$E0,$AF,$3B,$F0,$B5,$AA,$E1
	.dc.b	$E4,$FE,$1B,$06,$9A,$8C,$37,$2F,$63,$6B,$D4,$DC,$FD,$D5,$F9,$2F
	.dc.b	$C7,$BF,$17,$8A,$3F,$27,$F8,$37,$F9,$07,$1F,$C6,$B7,$1F,$95,$EF
	.dc.b	$6B,$D2,$F3,$B6,$BF,$45,$87,$FA,$89,$E3,$3F,$A0,$F7,$39,$67,$F3
	.dc.b	$EF,$13,$CA,$BF,$D3,$9F,$49,$FB,$BE,$FF,$F5,$13,$E4,$F4,$31,$E7
	.dc.b	$23,$E3,$FC,$96,$6F,$C2,$DD,$7F,$2B,$A3,$0F,$11,$CB,$E3,$6D,$FE
	.dc.b	$E7,$E8,$1F,$29,$FC,$2F,$C8,$7D,$37,$E2,$F2,$FC,$FD,$6A,$FF,$86
	.dc.b	$D6,$BC,$BF,$DE,$CE,$6B,$FD,$AB,$39,$AE,$D8,$75,$C3,$A7,$17,$FB
	.dc.b	$43,$0E,$18,$7F,$70,$6B,$FE,$38,$6B,$FE,$AC,$6F,$07,$FB,$1F,$6F
	.dc.b	$8B,$C5,$F1,$9B,$97,$FB,$1D,$B7,$FE,$99,$6E,$1F,$E8,$08,$F9,$B9
	.dc.b	$FC,$E6,$D7,$FB,$3D,$A7,$FA,$84,$B3,$FE,$EB,$F9,$CF,$EF,$DE,$0F
	.dc.b	$F3,$37,$FC,$01,$AD,$EE,$4D,$8C,$37,$0F,$E6,$F5,$C9,$B5,$F2,$EF
	.dc.b	$B3,$CA,$33,$7C,$9A,$7D,$BF,$53,$E7,$76,$FD,$46,$C6,$4F,$F0,$B6
	.dc.b	$E8,$FE,$7F,$9A,$31,$3A,$9F,$E8,$7B,$BF,$03,$7D,$4F,$E5,$D7,$D6
	.dc.b	$58,$C3,$6F,$FA,$9D,$71,$8C,$63,$FC,$E6,$7F,$DC,$F3,$97,$FA,$67
	.dc.b	$CF,$FF,$B6,$9F,$EB,$7B,$7F,$63,$FA,$4F,$E9,$73,$9B,$E4,$7F,$43
	.dc.b	$F3,$13,$AF,$84,$FE,$CD,$7C,$70,$EB,$18,$AD,$C6,$0F,$F1,$D6,$1D
	.dc.b	$70,$FF,$21,$B5,$FF,$45,$B5,$62,$FC,$5D,$F0,$BF,$BB,$5F,$0B,$63
	.dc.b	$0D,$C3,$E2,$CE,$9B,$3A,$7E,$1F,$7F,$ED,$7E,$87,$FD,$6B,$18,$3F
	.dc.b	$66,$E3,$1F,$EA,$B9,$7F,$07,$9F,$5D,$7D,$E8,$C4,$FF,$97,$83,$87
	.dc.b	$FC,$C3,$F6,$CF,$89,$FF,$07,$7C,$83,$EE,$7F,$9E,$CE,$71,$FF,$07
	.dc.b	$CE,$7D,$3F,$EF,$F3,$9C,$3E,$AC,$E6,$BF,$C1,$13,$9E,$24,$29,$5B
	.dc.b	$7F,$B7,$D7,$37,$CB,$8F,$E6,$23,$11,$FE,$BE,$6B,$CF,$A5,$F2,$7F
	.dc.b	$27,$D2,$FF,$09,$C6,$31,$F9,$31,$89,$FD,$1C,$E6,$BE,$33,$5F,$F4
	.dc.b	$2B,$5F,$F6,$6F,$82,$F8,$4F,$E5,$5F,$44,$F8,$6F,$F1,$0C,$E7,$2F
	.dc.b	$E5,$33,$97,$FC,$55,$19,$BF,$A8,$D7,$FD,$26,$D7,$BD,$FF,$8D,$B5
	.dc.b	$7E,$40,$F8,$AF,$E9,$5F,$90,$5C,$7E,$AC,$DC,$3F,$4F,$D4,$AC,$71
	.dc.b	$7D,$29,$FA,$35,$B7,$FD,$D3,$8F,$6C,$7C,$13,$99,$FB,$13,$9C,$7F
	.dc.b	$19,$8F,$F0,$7F,$AF,$FE,$0C,$7F,$07,$FB,$0B,$3F,$C1,$3C,$24,$E3
	.dc.b	$0F,$2C,$2B,$19,$FF,$95,$3E,$1B,$F0,$DF,$46,$F8,$5F,$2F,$0F,$A6
	.dc.b	$31,$8F,$FA,$EA,$B7,$5A,$CE,$CE,$4F,$F0,$6C,$65,$A6,$7D,$9E,$6F
	.dc.b	$E5,$30,$F8,$D1,$88,$FF,$80,$03,$0D,$EC,$AF,$25,$FE,$BF,$1F,$0D
	.dc.b	$F3,$C6,$BC,$B1,$E3,$7D,$AF,$84,$F9,$0E,$05,$3C,$38,$F0,$BE,$70
	.dc.b	$C3,$BE,$3E,$39,$CC,$63,$0F,$2D,$7C,$3B,$E4,$FA,$F0,$F0,$C6,$0F
	.dc.b	$81,$1D,$4D,$1F,$82,$9E,$6F,$EE,$CD,$63,$D0,$E3,$7F,$BF,$71,$7D
	.dc.b	$8F,$A5,$7A,$AE,$A7,$C0,$70,$78,$E3,$84,$78,$70,$F6,$BF,$1F,$1F
	.dc.b	$80,$FC,$73,$C7,$3F,$7E,$F9,$BE,$03,$9C,$77,$6C,$F2,$BD,$CF,$73
	.dc.b	$CA,$7D,$29,$E4,$FC,$B6,$BF,$0F,$0F,$A1,$87,$9A,$BD,$53,$B3,$A3
	.dc.b	$D9,$F1,$19,$29,$CC,$F8,$B7,$E5,$BF,$3E,$3F,$4A,$7C,$87,$48,$CD
	.dc.b	$F8,$5F,$2C,$68,$EA,$5A,$E8,$FE,$9F,$E0,$BF,$CE,$6D,$CB,$F2,$B1
	.dc.b	$88,$F2,$37,$E2,$7F,$D5,$3A,$7F,$3D,$8C,$7C,$0F,$FA,$E6,$BF,$DA
	.dc.b	$B7,$F8,$7B,$DC,$60,$FE,$88,$E2,$B8,$DF,$EA,$BE,$93,$FD,$5F,$57
	.dc.b	$E0,$BF,$21,$B7,$EE,$30,$E6,$DC,$BC,$0F,$EF,$2F,$8D,$FE,$39,$B1
	.dc.b	$BC,$3F,$6F,$7C,$B4,$FC,$E5,$69,$B8,$F8,$1F,$56,$3D,$9B,$7E,$5C
	.dc.b	$61,$F9,$CC,$3A,$3E,$EA,$E0,$FF,$63,$C3,$B3,$70,$EC,$EE,$FC,$B5
	.dc.b	$C2,$15,$C1,$B5,$C1,$5F,$81,$B5,$C6,$B8,$1C,$67,$46,$DC,$D7,$38
	.dc.b	$CC,$F1,$8E,$D3,$C6,$3C,$93,$F7,$33,$FA,$15,$FC,$15,$B7,$FC,$0F
	.dc.b	$1F,$8E,$FF,$1F,$AF,$05,$74,$3C,$17,$D0,$F0,$4F,$CC,$CB,$EC,$3F
	.dc.b	$36,$3F,$1D,$87,$68,$FE,$55,$8F,$E0,$23,$67,$0A,$D1,$31,$F1,$B9
	.dc.b	$7E,$FD,$77,$72,$47,$EB,$19,$3E,$B1,$E3,$CB,$DC,$E8,$F1,$E3,$F0
	.dc.b	$E7,$E0,$DF,$79,$D5,$F8,$DE,$6F,$A2,$7C,$71,$C8,$EA,$7D,$CB,$B9
	.dc.b	$7C,$A8,$F2,$19,$3F,$9C,$F1,$FF,$9F,$7E,$AF,$F9,$C5,$7F,$80,$E3
	.dc.b	$13,$FD,$DE,$73,$BF,$F9,$E3,$8D,$F1,$38,$AF,$F3,$1C,$2B,$17,$CA
	.dc.b	$7F,$BE,$C6,$27,$8C,$7F,$85,$8F,$F2,$A1,$FE,$12,$2F,$0F,$F8,$79
	.dc.b	$BC,$DA,$74,$7F,$A3,$BF,$21,$FE,$E0,$DE,$4F,$25,$71,$8F,$F2,$76
	.dc.b	$9F,$E1,$F5,$F9,$17,$F1,$97,$FA,$C1,$8E,$1E,$D1,$98,$FE,$B8,$D7
	.dc.b	$FC,$B0,$FF,$21,$9E,$9F,$01,$FE,$19,$9E,$FB,$FF,$63,$8C,$47,$FA
	.dc.b	$60,$EC,$68,$FF,$31,$7E,$14,$F9,$DF,$73,$63,$E2,$B5,$FF,$3D,$35
	.dc.b	$F4,$B9,$BF,$77,$E1,$4F,$8B,$FA,$D4,$8C,$3E,$11,$5C,$71,$FE,$F5
	.dc.b	$F1,$FF,$E3,$69,$FA,$AD,$7F,$D1,$4D,$7E,$4B,$5E,$3D,$DE,$0F,$A9
	.dc.b	$F8,$39,$7D,$16,$DE,$AB,$C5,$7D,$6B,$5D,$FC,$4F,$4C,$66,$BC,$BF
	.dc.b	$01,$F8,$38,$34,$97,$E8,$7E,$0F,$D2,$BE,$12,$9D,$23,$F2,$FE,$9B
	.dc.b	$E0,$2F,$34,$C3,$FD,$6F,$1A,$30,$FF,$9B,$27,$F2,$2D,$8F,$9F,$F4
	.dc.b	$6F,$81,$FC,$C3,$67,$F3,$ED,$7F,$D6,$8D,$7F,$26,$E7,$6F,$CC,$7D
	.dc.b	$1B,$CD,$FB,$6D,$AF,$B5,$5B,$DF,$FC,$64,$D8,$FF,$57,$4E,$DF,$C9
	.dc.b	$36,$FF,$49,$3B,$EF,$FF,$80,$1F,$2F,$FB,$7F,$B4,$98,$7F,$9E,$63
	.dc.b	$F7,$46,$1F,$F3,$CC,$61,$F1,$F1,$7C,$02,$3F,$01,$AF,$8A,$3D,$67
	.dc.b	$C8,$8F,$6F,$85,$EE,$E6,$95,$F9,$7C,$FF,$84,$BE,$A9,$CB,$EC,$9E
	.dc.b	$2E,$E3,$39,$FC,$A9,$CC,$FF,$A4,$3E,$7B,$F8,$A7,$A9,$23,$FD,$2D
	.dc.b	$1F,$C5,$CC,$3F,$E9,$68,$C3,$FE,$9B,$F9,$8F,$E3,$9F,$3F,$23,$04
	.dc.b	$3F,$D3,$AE,$5F,$F4,$FC,$7E,$D3,$E8,$9E,$F7,$F1,$D6,$CF,$F2,$46
	.dc.b	$BF,$EF,$96,$BF,$C6,$1C,$E5,$FC,$77,$E3,$1E,$EF,$F0,$06,$D7,$EF
	.dc.b	$EB,$7C,$9F,$E7,$76,$C7,$FB,$CE,$76,$FF,$17,$6D,$FF,$29,$9D,$FC
	.dc.b	$43,$C6,$FA,$73,$FE,$CB,$8C,$70,$7F,$45,$6E,$1F,$EC,$1F,$47,$F8
	.dc.b	$9F,$E1,$3F,$E3,$2B,$E1,$F8,$4B,$E4,$B8,$36,$FF,$8B,$56,$F5,$FE
	.dc.b	$A5,$F4,$3F,$D2,$C9,$3F,$CA,$F2,$5F,$18,$63,$E5,$D1,$C9,$E2,$FB
	.dc.b	$39,$5C,$67,$2F,$82,$3C,$7C,$32,$30,$77,$C3,$E2,$E6,$86,$5E,$2A
	.dc.b	$E1,$1F,$D0,$74,$61,$FC,$D3,$C2,$7D,$05,$25,$7F,$35,$F2,$BE,$AB
	.dc.b	$FC,$7A,$3E,$67,$E1,$0F,$8A,$A7,$09,$FE,$9D,$7D,$1F,$89,$3B,$6C
	.dc.b	$EF,$83,$FB,$25,$66,$78,$CF,$94,$FB,$5F,$84,$BE,$3B,$13,$D3,$E5
	.dc.b	$7C,$66,$1E,$26,$48,$60,$FE,$C5,$F8,$43,$E9,$FF,$AA,$B6,$3F,$AA
	.dc.b	$35,$FB,$3C,$5F,$37,$8B,$F9,$B7,$85,$C2,$FF,$1D,$D8,$FA,$BA,$67
	.dc.b	$0E,$73,$87,$F0,$D3,$B7,$27,$83,$E4,$3C,$64,$7F,$19,$E5,$FF,$82
	.dc.b	$78,$7F,$C1,$5A,$2B,$C3,$AC,$BE,$44,$3D,$27,$E1,$30,$46,$E3,$E1
	.dc.b	$CF,$C2,$99,$1F,$09,$FD,$57,$C1,$3C,$8C,$CF,$4D,$9E,$3B,$DE,$FD
	.dc.b	$83,$67,$91,$0E,$0E,$5B,$1D,$34,$F6,$FC,$3F,$F8,$BF,$53,$49,$5F
	.dc.b	$E3,$88,$CB,$FB,$1B,$7C,$B7,$88,$FB,$B5,$FE,$47,$E3,$7E,$47,$FD
	.dc.b	$BE,$FF,$51,$AF,$F7,$A5,$3A,$E1,$FF,$75,$74,$7E,$B6,$31,$3C,$23
	.dc.b	$FC,$A9,$F0,$7A,$8E,$38,$75,$58,$75,$BE,$31,$8B,$E9,$5B,$AF,$17
	.dc.b	$9C,$F8,$0F,$F6,$3C,$79,$6D,$F4,$BF,$A3,$F8,$6F,$14,$7F,$01,$8C
	.dc.b	$3D,$4C,$75,$A7,$70,$FA,$6F,$B8,$E2,$B8,$DF,$76,$E7,$FC,$36,$7C
	.dc.b	$38,$73,$7B,$1C,$F9,$BE,$13,$AC,$7A,$4E,$F0,$E8,$DB,$E9,$8F,$8A
	.dc.b	$F9,$E3,$E0,$7D,$38,$72,$3D,$33,$BC,$39,$C6,$67,$8E,$1D,$5F,$E4
	.dc.b	$95,$BC,$3D,$58,$75,$9F,$31,$C2,$FA,$E3,$EA,$9D,$13,$0E,$95,$EF
	.dc.b	$3C,$07,$F9,$EA,$3E,$8B,$63,$E3,$36,$BF,$D1,$FF,$0A,$F3,$FF,$9B
	.dc.b	$CF,$FA,$08,$FF,$73,$B8,$7F,$D0,$B1,$BE,$8E,$33,$99,$F5,$F7,$7E
	.dc.b	$3D,$BC,$7F,$E9,$CC,$93,$99,$FC,$33,$5F,$C5,$BF,$A3,$DB,$F0,$0D
	.dc.b	$D7,$F0,$1F,$49,$FE,$B0,$E3,$FC,$34,$E6,$F9,$E3,$83,$F1,$B5,$B8
	.dc.b	$FB,$73,$B7,$BC,$ED,$FC,$4C,$EF,$0F,$0C,$E5,$B3,$AF,$3B,$F6,$AF
	.dc.b	$49,$D6,$3F,$02,$35,$8E,$1F,$29,$B3,$25,$BE,$A4,$9F,$7D,$F6,$8C
	.dc.b	$4F,$F0,$4F,$A6,$7D,$2E,$6F,$F8,$24,$63,$0F,$D9,$F2,$FF,$2E,$ED
	.dc.b	$D6,$1C,$70,$B4,$BE,$13,$FB,$78,$C4,$F0,$BE,$51,$FB,$E3,$27,$CE
	.dc.b	$E2,$FF,$66,$57,$D6,$6B,$FE,$03,$D7,$F8,$04,$FF,$B7,$9B,$1F,$8F
	.dc.b	$6F,$6F,$FB,$55,$B3,$FD,$7E,$3F,$C8,$AF,$FA,$DE,$33,$3F,$E0,$FF
	.dc.b	$A5,$F4,$D7,$F8,$C7,$E6,$3F,$D8,$17,$FE,$BD,$9C,$BF,$DA,$63,$35
	.dc.b	$FE,$D3,$9F,$82,$FF,$B8,$9A,$FF,$0F,$6B,$DF,$2F,$6A,$8E,$AE,$6B
	.dc.b	$FC,$65,$6F,$A9,$FC,$3B,$FE,$F1,$AD,$D7,$1F,$73,$37,$FD,$F7,$D4
	.dc.b	$7C,$AE,$2B,$DD,$57,$A1,$5A,$BF,$74,$F9,$30,$FF,$58,$47,$EE,$EF
	.dc.b	$B3,$63,$31,$FB,$F6,$E3,$E5,$AE,$D8,$3B,$FC,$57,$C8,$F9,$1F,$29
	.dc.b	$E3,$CB,$FC,$C5,$97,$CF,$8C,$7A,$7C,$33,$47,$94,$93,$E3,$6D,$F4
	.dc.b	$8F,$F2,$D6,$3F,$ED,$C7,$A1,$FE,$F7,$FA,$27,$D3,$7C,$3F,$A5,$F8
	.dc.b	$57,$FE,$0E,$13,$AA,$FF,$32,$4E,$B6,$FD,$CD,$6A,$FE,$DD,$EE,$3F
	.dc.b	$91,$35,$FE,$A0,$6E,$FE,$64,$6E,$B9,$B6,$3F,$AB,$7D,$37,$FB,$5D
	.dc.b	$F4,$5F,$CA,$9F,$F3,$F4,$79,$34,$C7,$B5,$EB,$6C,$76,$8F,$29,$1C
	.dc.b	$1F,$0D,$FA,$BF,$4D,$F8,$CA,$EB,$97,$F8,$97,$E3,$5F,$4D,$F8,$96
	.dc.b	$E1,$F2,$FC,$BC,$4F,$95,$F8,$64,$8F,$7B,$FA,$0F,$88,$E3,$7F,$3F
	.dc.b	$0F,$F0,$FD,$7F,$A0,$76,$F9,$B5,$A7,$E2,$DF,$E6,$FB,$3F,$C8,$3E
	.dc.b	$DD,$F4,$BF,$D6,$F8,$7F,$A6,$74,$7A,$39,$AF,$91,$F1,$1F,$58,$95
	.dc.b	$F4,$9F,$24,$FE,$5B,$3E,$33,$E3,$3F,$D2,$98,$FF,$94,$3E,$9B,$55
	.dc.b	$F9,$CF,$A3,$7D,$27,$EB,$FC,$97,$D3,$7E,$47,$8B,$F1,$95,$AC,$BA
	.dc.b	$42,$B7,$E4,$73,$C3,$F8,$2D,$EE,$BF,$B6,$35,$FC,$5B,$5F,$F0,$E3
	.dc.b	$BE,$2F,$CA,$B7,$0F,$F5,$67,$CA,$3E,$E3,$FD,$53,$1E,$2B,$56,$D6
	.dc.b	$DF,$E6,$EE,$0F,$9D,$C9,$F3,$DB,$7E,$A9,$D3,$79,$3E,$DD,$FF,$9E
	.dc.b	$27,$2D,$8C,$E7,$FD,$8A,$7F,$CF,$73,$F6,$F6,$FB,$8D,$BE,$8D,$39
	.dc.b	$16,$BF,$B5,$FD,$3F,$FB,$23,$C3,$FE,$40,$FF,$B2,$B7,$BD,$FF,$C5
	.dc.b	$B0,$FF,$67,$4E,$5F,$A6,$F1,$6C,$7C,$38,$C6,$1F,$E9,$27,$83,$63
	.dc.b	$A2,$8C,$E1,$E4,$9C,$C6,$2B,$EB,$FA,$5F,$87,$E3,$FD,$F6,$1F,$86
	.dc.b	$8C,$7C,$77,$E3,$59,$5F,$88,$D3,$F1,$BD,$D8,$BF,$9D,$EB,$3E,$9F
	.dc.b	$F4,$CB,$E9,$EC,$61,$B8,$7D,$99,$D3,$67,$4F,$E1,$7E,$97,$FD,$6C
	.dc.b	$FE,$63,$77,$9B,$FE,$E9,$8D,$7B,$DF,$D3,$E7,$3F,$05,$FD,$02,$F8
	.dc.b	$E7,$FE,$45,$6B,$FC,$91,$AF,$F9,$D5,$A8,$3C,$A7,$F9,$8B,$6F,$F3
	.dc.b	$BD,$5F,$2A,$73,$87,$F0,$7F,$92,$7D,$3D,$F8,$EC,$37,$0F,$CD,$4E
	.dc.b	$9B,$3A,$69,$F0,$FC,$3F,$DF,$7D,$3B,$E2,$3D,$D1,$86,$CE,$CF,$B1
	.dc.b	$3F,$1F,$5F,$E9,$DA,$FF,$18,$7F,$A8,$7D,$33,$E4,$BF,$BC,$78,$6F
	.dc.b	$A7,$FE,$A1,$8F,$FA,$66,$B4,$FF,$0A,$9D,$7A,$BF,$DB,$A7,$38,$7F
	.dc.b	$9C,$DA,$FF,$52,$6B,$CF,$0F,$F0,$9C,$70,$E1,$FE,$BA,$DC,$3E,$97
	.dc.b	$E0,$6F,$C0,$DF,$67,$DF,$0F,$EF,$BD,$CF,$AA,$F8,$BE,$B6,$1B,$8F
	.dc.b	$F0,$F9,$D3,$63,$E5,$FC,$EE,$E3,$F8,$A7,$92,$FA,$9F,$E5,$9F,$55
	.dc.b	$FE,$EA,$9C,$C6,$3E,$33,$FC,$1B,$5B,$9C,$9F,$A7,$AC,$70,$FF,$7D
	.dc.b	$9B,$C3,$FE,$72,$E3,$7D,$2B,$EA,$7F,$CA,$31,$86,$C6,$71,$FF,$4F
	.dc.b	$4E,$5B,$19,$C1,$7D,$63,$0F,$F1,$FF,$0D,$F3,$BF,$7A,$73,$7F,$D3
	.dc.b	$A7,$2F,$ED,$67,$F7,$7F,$43,$FB,$C6,$46,$4F,$DD,$46,$30,$FF,$0D
	.dc.b	$1F,$A9,$F8,$8F,$D5,$79,$0F,$88,$FF,$71,$68,$DB,$FD,$27,$39,$BF
	.dc.b	$F0,$BD,$F0,$BD,$D6,$9F,$F5,$F4,$7E,$D7,$EA,$38,$53,$0E,$99,$66
	.dc.b	$F7,$2A,$DD,$6A,$31,$3A,$C7,$38,$F7,$9F,$18,$8C,$E1,$E5,$8C,$E1
	.dc.b	$87,$99,$8B,$F4,$CF,$8E,$30,$7F,$64,$BF,$F5,$9D,$7A,$A3,$E4,$1C
	.dc.b	$8F,$64,$6A,$BD,$73,$DC,$E1,$2C,$7B,$3C,$2E,$BD,$30,$F3,$1C,$27
	.dc.b	$D5,$1B,$3D,$D7,$D8,$F4,$BE,$33,$CB,$7E,$18,$D4,$7A,$EF,$BB,$E3
	.dc.b	$3D,$58,$FC,$FA,$F0,$47,$63,$81,$4F,$E6,$07,$1C,$3C,$F7,$DA,$3C
	.dc.b	$D3,$C8,$F1,$3F,$30,$CC,$6E,$BC,$0F,$DD,$3D,$75,$E2,$3A,$CF,$19
	.dc.b	$DB,$71,$FA,$0D,$C3,$DB,$8F,$CE,$AD,$CF,$A2,$31,$5C,$23,$07,$C2
	.dc.b	$70,$FD,$3C,$3C,$67,$95,$F8,$78,$7B,$63,$E8,$E3,$67,$E3,$BE,$98
	.dc.b	$E3,$1F,$96,$9F,$5B,$E0,$6B,$DC,$EB,$1F,$A4,$F8,$8F,$CE,$47,$E7
	.dc.b	$9A,$FA,$1A,$FE,$51,$CF,$C7,$7F,$AE,$B8,$3F,$B3,$46,$3E,$17,$F8
	.dc.b	$8D,$23,$13,$AC,$3E,$ED,$F1,$AD,$46,$2F,$8B,$FB,$CF,$A9,$FF,$5D
	.dc.b	$4F,$E5,$9A,$FE,$1E,$B9,$4F,$F0,$26,$E3,$F8,$AA,$DC,$F3,$F8,$7F
	.dc.b	$C6,$BB,$C3,$BD,$78,$71,$FD,$4D,$F6,$CF,$86,$5F,$AC,$CB,$D1,$1C
	.dc.b	$6B,$C9,$97,$4C,$3C,$0D,$BF,$F0,$55,$F9,$1B,$8F,$C6,$C3,$AC,$7E
	.dc.b	$FF,$1F,$06,$1E,$38,$FE,$05,$7E,$08,$EB,$1E,$6C,$3F,$4F,$7E,$38
	.dc.b	$DE,$1D,$6B,$71,$D6,$B7,$1D,$6B,$86,$5E,$22,$47,$85,$E9,$1E,$62
	.dc.b	$47,$12,$57,$F8,$9F,$1F,$4B,$6F,$BB,$F0,$32,$F9,$58,$FD,$08,$D1
	.dc.b	$F2,$8F,$E2,$79,$7C,$74,$FC,$17,$67,$AC,$E7,$31,$C1,$31,$F4,$E3
	.dc.b	$EB,$3E,$1E,$5E,$E9,$F2,$1F,$8B,$C3,$E4,$3D,$AB,$DE,$E9,$E6,$E6
	.dc.b	$BE,$6C,$62,$3C,$F3,$C9,$65,$F1,$1B,$97,$C4,$77,$8F,$A7,$1F,$24
	.dc.b	$68,$F8,$77,$F1,$DD,$9E,$DC,$7C,$27,$D2,$73,$3C,$B0,$FD,$0C,$FC
	.dc.b	$19,$FF,$56,$F1,$3E,$A7,$F9,$7E,$3F,$E0,$56,$F0,$FF,$80,$DB,$5C
	.dc.b	$3E,$36,$FB,$2C,$39,$4F,$F9,$EE,$B3,$18,$8F,$F4,$2D,$72,$E3,$FF
	.dc.b	$0F,$C5,$E6,$FE,$7E,$FF,$86,$31,$FF,$63,$C6,$32,$F2,$4F,$F9,$3E
	.dc.b	$36,$FF,$82,$9A,$F0,$2D,$76,$73,$F5,$1F,$C7,$29,$7F,$D3,$A3,$0D
	.dc.b	$DB,$DF,$4F,$82,$65,$8B,$DC,$FF,$A1,$65,$CF,$C1,$B1,$FE,$DD,$9D
	.dc.b	$AF,$A6,$7D,$47,$66,$C7,$FB,$8D,$AF,$C3,$3F,$D2,$5F,$50,$F8,$AE
	.dc.b	$15,$E6,$87,$CF,$53,$35,$FE,$ED,$F9,$3F,$F1,$FB,$7B,$BD,$B1,$88
	.dc.b	$EC,$3C,$3F,$7B,$B3,$E6,$C6,$DF,$AE,$D7,$FB,$BB,$5F,$F5,$43,$D7
	.dc.b	$87,$E7,$C7,$FA,$AE,$7E,$2B,$E0,$6C,$6A,$13,$C7,$87,$AF,$C2,$FA
	.dc.b	$09,$3F,$EB,$5A,$E1,$F5,$0F,$5F,$33,$EC,$87,$C4,$FD,$C6,$C7,$C5
	.dc.b	$BF,$CD,$FA,$EF,$7A,$F8,$92,$BF,$D6,$B1,$A3,$0F,$D5,$ED,$57,$F9
	.dc.b	$6E,$36,$79,$74,$FF,$0D,$FD,$43,$E1,$7F,$1A,$D9,$D1,$FE,$CF,$70
	.dc.b	$FD,$4F,$9B,$FE,$D4,$E8,$46,$9F,$D4,$B5,$EF,$3F,$A4,$76,$D8,$EB
	.dc.b	$BF,$CF,$A7,$9D,$FA,$06,$DF,$7A,$DE,$1F,$E8,$5F,$99,$FD,$EF,$D4
	.dc.b	$4A,$FF,$7D,$47,$F0,$03,$0F,$C5,$F0,$93,$98,$E9,$C9,$F0,$88,$FB
	.dc.b	$71,$FF,$69,$3A,$31,$3E,$53,$6F,$C5,$EF,$F4,$73,$C8,$E5,$1E,$B3
	.dc.b	$0F,$CB,$8C,$3A,$3C,$0E,$B5,$F8,$93,$FC,$6B,$E9,$7F,$8D,$FA,$72
	.dc.b	$FE,$BF,$17,$F2,$19,$D4,$63,$4F,$4B,$F0,$6B,$2F,$C3,$F9,$BF,$E4
	.dc.b	$BF,$51,$23,$84,$7F,$28,$30,$FC,$98,$C3,$F8,$1F,$A8,$7C,$0F,$F1
	.dc.b	$66,$C6,$63,$F8,$CF,$C8,$3E,$3F,$51,$F8,$3A,$DD,$7F,$2A,$6B,$FC
	.dc.b	$C9,$AF,$F9,$C9,$DE,$DF,$B5,$B7,$C1,$F6,$3A,$D7,$E6,$BE,$37,$FD
	.dc.b	$4D,$C1,$FD,$21,$B1,$9F,$2B,$F8,$6B,$F8,$5F,$43,$F9,$7F,$D3,$7F
	.dc.b	$8F,$23,$FA,$1B,$5F,$F4,$1B,$50,$71,$C7,$FD,$4F,$EA,$3F,$A8,$92
	.dc.b	$7C,$9F,$50,$F9,$8F,$26,$2E,$6F,$0F,$D4,$4B,$F3,$45,$7F,$A7,$E4
	.dc.b	$9F,$D6,$36,$3F,$A8,$B5,$FE,$B7,$A3,$1F,$8F,$1E,$2D,$BE,$24,$62
	.dc.b	$3F,$98,$D7,$1E,$0F,$E4,$5E,$23,$E2,$7F,$A9,$B6,$BF,$81,$35,$FE
	.dc.b	$73,$E5,$7D,$6F,$03,$FD,$53,$39,$9F,$ED,$AD,$73,$5F,$C1,$63,$11
	.dc.b	$F3,$1E,$97,$FC,$E3,$37,$D4,$3E,$37,$EE,$36,$FC,$78,$79,$A6,$47
	.dc.b	$99,$FC,$17,$D4,$3E,$4B,$33,$FD,$EB,$2E,$4D,$AF,$21,$1F,$C6,$60
	.dc.b	$F9,$6F,$09,$CB,$3F,$E9,$1F,$4C,$FA,$7F,$EF,$2D,$BF,$E5,$0D,$7F
	.dc.b	$B6,$6B,$FC,$FF,$C6,$DD,$7F,$27,$AD,$57,$F8,$51,$AF,$FB,$51,$AF
	.dc.b	$E4,$9E,$97,$F1,$FB,$1F,$51,$FA,$16,$D7,$8A,$FC,$BE,$8C,$8F,$2B
	.dc.b	$FA,$7F,$3B,$FB,$16,$1E,$C8,$F2,$12,$7F,$C3,$F8,$FF,$5F,$6C,$77
	.dc.b	$D2,$47,$43,$E0,$E3,$F1,$08,$FE,$DF,$E0,$DE,$5E,$23,$FC,$5B,$82
	.dc.b	$BE,$21,$23,$C2,$FE,$F7,$CB,$7B,$D4,$E5,$9F,$D4,$F8,$6F,$A6,$4B
	.dc.b	$F0,$79,$15,$EF,$1F,$44,$8F,$F8,$FB,$A5,$F5,$1F,$DD,$63,$F8,$0E
	.dc.b	$AE,$7F,$E0,$75,$AE,$1F,$F7,$94,$62,$F8,$5A,$F8,$3F,$E6,$2F,$82
	.dc.b	$DB,$FE,$44,$9D,$CE,$B0,$E9,$3E,$0B,$E9,$87,$0F,$17,$FC,$C5,$8F
	.dc.b	$2A,$7D,$3F,$E7,$FE,$1B,$84,$7F,$18,$8C,$3E,$3C,$3F,$8D,$BC,$3B
	.dc.b	$3F,$BF,$E7,$E6,$87,$D4,$7D,$67,$99,$BC,$2E,$1D,$2B,$83,$70,$EB
	.dc.b	$7C,$6F,$C0,$78,$F0,$EF,$3E,$2A,$E1,$39,$C3,$BD,$F1,$C3,$AE,$1C
	.dc.b	$2B,$47,$43,$C8,$59,$E1,$87,$59,$E1,$3A,$AE,$77,$DB,$97,$E5,$63
	.dc.b	$E2,$AD,$61,$C6,$B5,$1F,$03,$0E,$75,$B3,$86,$3E,$09,$F2,$9F,$E8
	.dc.b	$E8,$FA,$2D,$9F,$F1,$6B,$6B,$FD,$17,$F0,$EF,$3D,$DB,$86,$D7,$FA
	.dc.b	$4A,$75,$39,$3F,$D2,$78,$7F,$96,$B7,$BD,$BF,$92,$BF,$9C,$D7,$F0
	.dc.b	$8D,$7F,$12,$FE,$7F,$6F,$B6,$DD,$7F,$0F,$F4,$FF,$E9,$4E,$2F,$C0
	.dc.b	$CE,$6F,$9E,$18,$3F,$0F,$5B,$8F,$AD,$8F,$95,$B5,$C5,$F0,$D7,$0F
	.dc.b	$37,$FC,$F3,$97,$8A,$3A,$4F,$07,$58,$E2,$75,$39,$3F,$09,$8F,$86
	.dc.b	$32,$4C,$BE,$1D,$F2,$AE,$13,$97,$D6,$72,$C3,$C7,$5C,$23,$F3,$78
	.dc.b	$3E,$A1,$F5,$1A,$9F,$F5,$FC,$62,$7F,$67,$CB,$FD,$2F,$AB,$51,$C6
	.dc.b	$ED,$6E,$FA,$B7,$2F,$E1,$38,$71,$BF,$DE,$9C,$4F,$9F,$C5,$FE,$A0
	.dc.b	$EA,$FF,$4E,$6C,$8C,$4F,$F8,$6B,$85,$5F,$E3,$C8,$FF,$3F,$56,$DB
	.dc.b	$5C,$32,$FF,$10,$CE,$5F,$EB,$75,$C5,$FF,$5D,$C6,$63,$FD,$59,$F4
	.dc.b	$FE,$B9,$FF,$77,$B5,$FE,$22,$D7,$FC,$5A,$D4,$FC,$38,$E7,$A7,$8F
	.dc.b	$1F,$4F,$D4,$79,$6F,$73,$6F,$86,$5D,$A3,$86,$0B,$D4,$6F,$DB,$E9
	.dc.b	$E9,$3E,$5B,$18,$7F,$93,$63,$0F,$FA,$8F,$1F,$F5,$63,$6B,$C9,$9B
	.dc.b	$B3,$8E,$9F,$ED,$EB,$DB,$6B,$73,$AD,$3E,$3F,$C9,$78,$BD,$2F,$53
	.dc.b	$72,$F0,$38,$7F,$CC,$19,$7F,$53,$72,$4C,$7C,$13,$E4,$E5,$E1,$1F
	.dc.b	$EF,$0F,$A4,$FF,$7E,$57,$F9,$A7,$EA,$3D,$B3,$9F,$87,$F9,$D1,$8C
	.dc.b	$FF,$DC,$D7,$C2,$B5,$7F,$EC,$5F,$A8,$FF,$5D,$BF,$EC,$79,$D5,$70
	.dc.b	$C3,$EB,$46,$27,$FA,$8D,$70,$BE,$67,$97,$D7,$F0,$EB,$C1,$B5,$D5
	.dc.b	$C3,$FE,$D5,$FA,$8F,$BF,$5F,$E7,$6C,$3F,$CE,$13,$B2,$47,$E8,$1F
	.dc.b	$EC,$33,$C2,$13,$C1,$FF,$1B,$FC,$63,$EA,$3F,$20,$DB,$F7,$7B,$D8
	.dc.b	$9F,$6B,$23,$FD,$D1,$F1,$1B,$BF,$A1,$3F,$97,$8C,$6D,$F3,$23,$17
	.dc.b	$C6,$BF,$DF,$DE,$7F,$FA,$83,$0F,$F6,$0B,$63,$CB,$E7,$7E,$D7,$BB
	.dc.b	$FC,$AD,$ED,$7C,$EF,$87,$98,$78,$AA,$DF,$C1,$3F,$D3,$7A,$7F,$8B
	.dc.b	$E7,$47,$6F,$A8,$F3,$57,$EB,$7E,$A1,$F5,$0A,$FF,$04,$E1,$FE,$0A
	.dc.b	$9C,$C6,$23,$FD,$4F,$18,$70,$D7,$8B,$9D,$FA,$47,$0A,$FE,$21,$F2
	.dc.b	$4F,$A3,$FF,$51,$61,$FE,$B2,$BD,$DB,$0E,$53,$E3,$C3,$D0,$72,$BE
	.dc.b	$35,$F0,$4E,$11,$F2,$77,$6B,$78,$70,$DC,$3E,$8B,$6F,$EA,$B7,$2F
	.dc.b	$4F,$07,$FA,$1E,$7A,$79,$5D,$5B,$7F,$69,$A7,$DE,$2D,$7C,$2F,$A8
	.dc.b	$FF,$00,$F8,$1F,$CC,$5F,$F0,$1E,$97,$BF,$F8,$5D,$6D,$F8,$36,$AE
	.dc.b	$38,$7C,$73,$FD,$B2,$F2,$71,$1B,$9E,$71,$A7,$9C,$F2,$9F,$C4,$3C
	.dc.b	$5B,$1C,$1E,$57,$F8,$7F,$42,$5F,$6C,$9F,$50,$EC,$CE,$1F,$4F,$8B
	.dc.b	$FD,$6D,$7E,$48,$FB,$98,$7F,$51,$C3,$8D,$F0,$D3,$E7,$FC,$4B,$E9
	.dc.b	$FF,$44,$BE,$A2,$C6,$1B,$97,$C0,$E1,$56,$E3,$0F,$36,$CF,$F4,$DF
	.dc.b	$A5,$7B,$BF,$90,$D7,$FC,$D3,$18,$AF,$D0,$42,$30,$FE,$8D,$AF,$EA
	.dc.b	$1C,$E5,$F9,$A6,$CA,$5D,$1F,$A1,$F9,$0F,$16,$12,$BE,$35,$F0,$BD
	.dc.b	$F4,$FF,$2B,$C7,$A5,$FF,$05,$F9,$17,$D3,$DF,$8E,$C3,$72,$F8,$93
	.dc.b	$A6,$CE,$9A,$78,$3C,$0F,$88,$7D,$5F,$A7,$FF,$7B,$46,$49,$9F,$EB
	.dc.b	$F4,$FE,$31,$9F,$EF,$B9,$17,$C2,$B5,$87,$F7,$BF,$A6,$7C,$8F,$CF
	.dc.b	$9F,$89,$18,$CB,$E7,$B6,$3F,$88,$35,$FE,$2A,$E7,$1F,$E0,$ED,$9F
	.dc.b	$9D,$F4,$CF,$A6,$7D,$37,$87,$19,$5F,$93,$BE,$18,$B0,$E1,$7A,$8F
	.dc.b	$E1,$9C,$6F,$A6,$FF,$0A,$37,$0F,$D1,$6A,$F9,$AC,$D4,$9F,$D0,$56
	.dc.b	$EF,$F9,$B6,$FF,$1A,$F5,$1F,$C6,$7C,$37,$D3,$7F,$88,$5B,$87,$EB
	.dc.b	$FE,$9A,$D6,$EB,$44,$9D,$36,$3F,$A2,$74,$BE,$89,$F4,$DF,$CC,$9B
	.dc.b	$7F,$A6,$F9,$BC,$4C,$8F,$E1,$75,$C6,$3F,$67,$39,$BF,$EA,$1D,$FF
	.dc.b	$B0,$8C,$DF,$EC,$3E,$8E,$D6,$BC,$05,$63,$A3,$F7,$F5,$6F,$FA,$77
	.dc.b	$C0,$FE,$9F,$E9,$9F,$09,$FE,$DC,$D7,$F9,$DD,$66,$37,$9A,$B6,$FE
	.dc.b	$BE,$B6,$E6,$77,$3A,$8F,$D9,$F0,$FF,$56,$8C,$47,$ED,$F0,$EC,$F3
	.dc.b	$9D,$57,$23,$C8,$E2,$3F,$B4,$CF,$07,$AE,$15,$E8,$78,$CE,$8E,$54
	.dc.b	$71,$C3,$D3,$87,$92,$72,$71,$3A,$94,$EA,$FB,$63,$8E,$1E,$79,$EE
	.dc.b	$6C,$E7,$3B,$C3,$D5,$1F,$C5,$CC,$CF,$BB,$0F,$36,$1E,$5A,$DD,$78
	.dc.b	$2B,$4F,$59,$D5,$70,$8F,$E8,$D3,$AF,$17,$D9,$5F,$C9,$5F,$15,$59
	.dc.b	$F4,$57,$BB,$0D,$BF,$07,$1F,$6B,$F4,$0C,$63,$E4,$7D,$27,$C0,$9E
	.dc.b	$38,$76,$8F,$19,$C5,$C1,$DA,$34,$5B,$F6,$B6,$7A,$47,$9A,$FC,$71
	.dc.b	$F2,$71,$F8,$2F,$C3,$3C,$B8,$FC,$3B,$EB,$7C,$71,$F4,$5F,$09,$ED
	.dc.b	$7F,$06,$B5,$7C,$5F,$19,$88,$ED,$3D,$1F,$88,$D7,$C8,$78,$8F,$27
	.dc.b	$C1,$FF,$79,$7E,$53,$5D,$B5,$FC,$CB,$9F,$88,$7C,$07,$FB,$42,$7F
	.dc.b	$DA,$9A,$7E,$26,$73,$B7,$F6,$1B,$E1,$7B,$6C,$66,$B8,$4F,$3C,$3F
	.dc.b	$6D,$F4,$DF,$EB,$1C,$7F,$58,$D7,$F4,$58,$F6,$6D,$71,$CB,$F3,$55
	.dc.b	$BB,$E7,$F0,$1F,$9D,$7F,$81,$61,$C5,$B7,$C7,$0F,$EF,$17,$C7,$0E
	.dc.b	$78,$71,$7C,$F9,$7E,$FF,$1E,$51,$AB,$F3,$D6,$DF,$F0,$14,$E9,$B8
	.dc.b	$77,$58,$77,$7C,$35,$C5,$57,$16,$D7,$1A,$E1,$1C,$AB,$85,$6C,$E1
	.dc.b	$3A,$36,$E6,$B9,$4F,$79,$CC,$79,$67,$E7,$65,$F2,$AF,$F6,$C7,$98
	.dc.b	$A7,$93,$1B,$3E,$0C,$3A,$1E,$4B,$E8,$78,$2B,$E3,$9F,$2D,$CD,$74
	.dc.b	$C3,$F0,$18,$78,$6B,$F7,$27,$72,$E1,$DD,$E6,$7E,$12,$BA,$18,$CB
	.dc.b	$91,$F8,$07,$31,$FC,$B6,$FE,$36,$3E,$79,$CD,$70,$CB,$EF,$5F,$52
	.dc.b	$46,$5F,$88,$62,$3F,$0E,$F5,$AF,$86,$F4,$8C,$CF,$18,$FB,$46,$4F
	.dc.b	$C3,$D7,$A6,$7C,$33,$E9,$9F,$E9,$9C,$7F,$E6,$2F,$A6,$FE,$75,$7F
	.dc.b	$E6,$C6,$ED,$FE,$53,$9D,$57,$F9,$9B,$E2,$32,$A3,$2F,$F7,$9A,$CC
	.dc.b	$62,$FF,$CE,$F5,$C3,$8B,$FC,$1D,$77,$9F,$FD,$97,$5F,$EC,$29,$CC
	.dc.b	$FF,$9D,$30,$D3,$0F,$F0,$7D,$70,$C7,$CD,$CC,$AF,$2D,$3E,$9B,$F8
	.dc.b	$B6,$3F,$EC,$86,$DF,$F5,$96,$CF,$F9,$C7,$DB,$67,$CA,$73,$3C,$F8
	.dc.b	$B0,$76,$AD,$C7,$FA,$67,$E9,$9F,$4D,$E2,$6C,$66,$3C,$BF,$4C,$F8
	.dc.b	$3D,$44,$9D,$1F,$EA,$39,$D3,$FE,$A5,$F8,$DF,$F1,$2B,$7B,$7D,$F1
	.dc.b	$88,$FF,$6F,$78,$3E,$47,$FD,$C1,$19,$DF,$FD,$D2,$DD,$FF,$BD,$36
	.dc.b	$7F,$D2,$55,$C6,$B7,$7F,$E9,$E9,$57,$77,$FC,$04,$62,$35,$19,$8E
	.dc.b	$BF,$0D,$E6,$23,$FE,$BF,$FA,$6F,$F5,$6B,$ED,$8C,$BE,$87,$ED,$C6
	.dc.b	$A3,$FC,$97,$E8,$3C,$5F,$C7,$BF,$EF,$A8,$FF,$4D,$F8,$38,$99,$18
	.dc.b	$21,$C9,$FC,$3F,$D3,$3D,$DA,$49,$FF,$6A,$7C,$BB,$C3,$C4,$36,$70
	.dc.b	$7F,$88,$F0,$E5,$3F,$ED,$B6,$BB,$8F,$F1,$45,$F3,$8D,$61,$F3,$6B
	.dc.b	$71,$F3,$1A,$F5,$8F,$F6,$7F,$6F,$FB,$6B,$E5,$3F,$5F,$D8,$FA,$6A
	.dc.b	$62,$25,$FF,$B7,$A3,$2F,$FB,$82,$3E,$0F,$14,$8F,$43,$FA,$AC,$BF
	.dc.b	$D6,$71,$82,$1E,$B3,$FD,$B1,$97,$E1,$1F,$8F,$E8,$F1,$ED,$E5,$CB
	.dc.b	$27,$E5,$C6,$5F,$9B,$19,$7A,$0B,$FC,$96,$DF,$22,$B5,$BF,$D0,$AF
	.dc.b	$F7,$45,$7E,$65,$E1,$3F,$13,$E9,$3E,$2E,$32,$7F,$DF,$F1,$97,$FE
	.dc.b	$00,$C6,$5F,$F4,$94,$7C,$CF,$95,$F9,$BF,$33,$27,$FD,$ED,$8F,$87
	.dc.b	$85,$C2,$BE,$15,$AD,$3F,$8A,$E1,$C6,$3F,$D3,$D3,$FC,$5B,$E9,$9E
	.dc.b	$CD,$24,$F1,$03,$E1,$DF,$0D,$A8,$F8,$F5,$C9,$E0,$1C,$3F,$32,$78
	.dc.b	$BA,$8C,$BF,$BA,$6B,$FE,$7D,$6B,$E2,$6F,$C2,$BC,$4F,$AB,$28,$C7
	.dc.b	$0F,$F4,$66,$C7,$AF,$6F,$8F,$E7,$7F,$36,$FA,$3F,$F1,$B7,$0F,$DB
	.dc.b	$A4,$63,$83,$FD,$1C,$D8,$ED,$97,$83,$0F,$A3,$F4,$DF,$A2,$24,$FC
	.dc.b	$9C,$9F,$4A,$F9,$0F,$26,$0E,$5F,$17,$D3,$4C,$38,$4E,$5F,$97,$9A
	.dc.b	$75,$1F,$96,$8C,$3C,$2B,$E0,$60,$F8,$0A,$49,$E7,$7F,$D5,$BC,$57
	.dc.b	$D2,$D2,$4F,$DA,$FA,$6F,$B3,$C3,$F1,$6B,$73,$F7,$9A,$FE,$A1,$B1
	.dc.b	$F2,$E7,$8F,$0F,$FA,$DE,$72,$FE,$33,$77,$D2,$3E,$97,$EC,$37,$1E
	.dc.b	$D9,$79,$E6,$47,$85,$FB,$DF,$4C,$F8,$CC,$CF,$8E,$6E,$9F,$B7,$C5
	.dc.b	$7B,$AE,$47,$D7,$C5,$F2,$3E,$23,$8E,$7F,$EC,$FF,$A4,$7D,$2F,$F7
	.dc.b	$E6,$DF,$E8,$1A,$FF,$15,$C7,$FA,$37,$8D,$BA,$93,$B7,$F2,$DE,$5F
	.dc.b	$F1,$98,$FE,$23,$7C,$B4,$F4,$77,$3E,$9B,$F3,$8D,$AE,$47,$87,$E3
	.dc.b	$7C,$84,$7F,$25,$87,$E7,$6F,$F3,$D8,$7F,$5B,$6D,$6E,$BD,$15,$BD
	.dc.b	$3E,$04,$F8,$35,$95,$DC,$F8,$18,$FC,$32,$3F,$A1,$F8,$07,$97,$88
	.dc.b	$F6,$CD,$CB,$EA,$56,$A7,$99,$FD,$F0,$FE,$23,$9F,$F1,$17,$01,$C3
	.dc.b	$B6,$BF,$E0,$45,$F4,$DF,$5C,$91,$E5,$F1,$EB,$F4,$22,$DF,$92,$7F
	.dc.b	$C9,$FA,$3E,$9B,$FC,$01,$7F,$C9,$31,$74,$7F,$26,$AD,$6F,$FE,$ED
	.dc.b	$8C,$5F,$0B,$5F,$07,$FC,$B5,$E4,$FF,$B1,$E3,$18,$72,$9E,$D7,$CB
	.dc.b	$0E,$91,$FE,$00,$C3,$A5,$F3,$F0,$FF,$A1,$FC,$06,$A9,$F4,$BF,$A5
	.dc.b	$F4,$B9,$C7,$F3,$C8,$C3,$E4,$CB,$F9,$FB,$C3,$B3,$FC,$0D,$8F,$9E
	.dc.b	$1F,$4D,$F5,$DE,$07,$39,$DD,$72,$BE,$B7,$C5,$B1,$DA,$F9,$4E,$6B
	.dc.b	$8D,$F3,$C3,$CE,$F8,$DB,$5C,$2F,$B5,$70,$AD,$1E,$42,$DF,$A6,$B8
	.dc.b	$47,$53,$58,$79,$1C,$4E,$F9,$3E,$54,$E6,$79,$E1,$E0,$C7,$C7,$5A
	.dc.b	$C7,$C5,$7B,$38,$63,$E0,$C3,$CA,$7F,$96,$23,$E8,$B6,$7F,$C6,$8D
	.dc.b	$BF,$CF,$FC,$27,$F7,$DE,$BB,$E7,$7F,$5F,$AF,$D3,$4E,$5B,$19,$7B
	.dc.b	$76,$7E,$45,$FA,$8D,$7F,$04,$D7,$F0,$ED,$C7,$FC,$DF,$C3,$F4,$9B
	.dc.b	$3F,$37,$E9,$7F,$D2,$FC,$79,$BF,$D2,$F0,$7D,$F6,$F0,$78,$31,$FB
	.dc.b	$77,$C3,$0F,$A3,$5A,$7C,$73,$CC,$DF,$99,$FE,$2F,$C3,$A4,$66,$78
	.dc.b	$BA,$C7,$13,$A9,$C9,$E7,$AE,$58,$F8,$E3,$33,$B3,$F3,$D8,$F9,$5C
	.dc.b	$4E,$DB,$3F,$E2,$49,$D4,$E4,$FF,$61,$FC,$57,$69,$FD,$EB,$79,$3F
	.dc.b	$C6,$F3,$79,$93,$49,$97,$ED,$AB,$6D,$9D,$B7,$87,$FC,$A5,$D1,$FE
	.dc.b	$43,$9F,$F4,$34,$FF,$87,$DF,$F1,$33,$67,$6F,$FB,$61,$B8,$FF,$12
	.dc.b	$6D,$FF,$93,$3C,$AB,$7D,$2B,$FC,$E1,$2F,$C9,$3F,$C1,$38,$7F,$8B
	.dc.b	$5B,$3A,$5F,$4B,$D1,$CE,$5F,$E5,$96,$D7,$FB,$FD,$B5,$FE,$FD,$78
	.dc.b	$C7,$5F,$1B,$91,$1F,$05,$DA,$F0,$78,$5D,$CF,$07,$8D,$6B,$CD,$16
	.dc.b	$7C,$D3,$71,$F2,$4D,$AF,$24,$DA,$F2,$7D,$CF,$AC,$F6,$7D,$D5,$AF
	.dc.b	$85,$16,$7E,$14,$DC,$3E,$04,$DA,$F8,$13,$6B,$E0,$7D,$C7,$C7,$3E
	.dc.b	$47,$8B,$6B,$E7,$45,$9F,$9D,$37,$0F,$99,$36,$BE,$64,$DA,$F9,$9F
	.dc.b	$3B,$F4,$CF,$A9,$F7,$36,$BE,$D4,$59,$FB,$53,$70,$FB,$13,$6B,$EC
	.dc.b	$4D,$AF,$B1,$F4,$FF,$7F,$DE,$FC,$03,$6B,$F0,$CE,$1F,$C4,$7C,$15
	.dc.b	$AF,$C4,$37,$1F,$C3,$36,$7F,$0D,$F2,$BF,$8B,$3F,$19,$F7,$36,$3F
	.dc.b	$24,$E1,$FC,$A7,$CA,$DC,$3F,$29,$E2,$FE,$4E,$3F,$28,$D9,$FF,$50
	.dc.b	$7D,$0F,$FA,$2F,$5F,$D2,$B6,$33,$E3,$38,$72,$AA,$92,$77,$37,$1F
	.dc.b	$F6,$C5,$F3,$6D,$F1,$AD,$9D,$AF,$F5,$1F,$11,$23,$F5,$DE,$25,$BF
	.dc.b	$D7,$4D,$DF,$F5,$93,$6B,$F5,$87,$EC,$4F,$D9,$7D,$AD,$7F,$70,$5B
	.dc.b	$FD,$C4,$5A,$FF,$67,$7E,$27,$FE,$60,$AF,$F5,$52,$FA,$5F,$E0,$EE
	.dc.b	$6F,$F8,$2B,$99,$DF,$A9,$AA,$DB,$63,$27,$48,$CB,$4F,$F6,$97,$C9
	.dc.b	$3F,$25,$FE,$43,$87,$F1,$2F,$0B,$FD,$49,$5A,$9C,$E5,$FC,$8F,$DA
	.dc.b	$FE,$3B,$87,$1F,$1A,$D6,$AF,$FD,$BB,$E2,$35,$7A,$FA,$BF,$E4,$BA
	.dc.b	$3E,$7F,$38,$7F,$B9,$74,$7D,$8F,$FA,$66,$3F,$DC,$33,$AA,$57,$F3
	.dc.b	$6A,$D4,$62,$BF,$C5,$F5,$B8,$FF,$18,$9D,$F0,$FF,$6A,$3A,$6B,$FE
	.dc.b	$F4,$8E,$0F,$FB,$DB,$9B,$FD,$C5,$C9,$F9,$43,$F2,$BC,$97,$0F,$CC
	.dc.b	$C5,$AF,$C8,$CD,$C9,$5C,$F9,$97,$98,$C5,$5B,$6D,$F2,$7E,$87,$F0
	.dc.b	$67,$E1,$3E,$1E,$DF,$E2,$22,$D7,$E1,$A6,$E5,$FE,$F1,$FA,$07,$D2
	.dc.b	$7F,$54,$F4,$9F,$41,$FE,$A1,$C9,$8F,$C9,$2D,$F0,$8C,$E1,$EA,$95
	.dc.b	$F2,$F2,$BE,$DD,$EE,$79,$07,$15,$C6,$BC,$3E,$47,$0C,$BF,$81,$DB
	.dc.b	$E6,$E9,$8F,$9F,$B9,$FF,$0A,$9D,$36,$74,$D2,$47,$F0,$43,$65,$9E
	.dc.b	$6F,$9B,$F2,$67,$C2,$7C,$5D,$9E,$2F,$E5,$A7,$38,$7C,$23,$E3,$E9
	.dc.b	$F2,$A7,$57,$27,$58,$7C,$6C,$3F,$DD,$51,$88,$61,$D1,$F8,$FF,$93
	.dc.b	$FF,$80,$F6,$F8,$DF,$45,$FE,$A1,$9C,$FA,$BF,$66,$31,$3F,$4E,$55
	.dc.b	$C7,$0F,$F4,$7B,$63,$EB,$F0,$3E,$9B,$F6,$33,$9E,$6F,$DB,$7D,$B7
	.dc.b	$F1,$03,$F8,$8F,$D9,$5C,$7F,$8A,$C5,$C7,$F8,$9C,$DC,$BF,$8D,$FD
	.dc.b	$67,$0B,$FE,$36,$DA,$FE,$32,$D9,$DC,$FF,$20,$AF,$E0,$75,$FE,$C1
	.dc.b	$1F,$47,$F8,$37,$F2,$BE,$2B,$E6,$BF,$25,$39,$9F,$F5,$6F,$BD,$F9
	.dc.b	$A8,$C5,$6F,$0C,$74,$7F,$B8,$6A,$DF,$4C,$39,$CE,$63,$C2,$FF,$AE
	.dc.b	$2F,$94,$60,$FE,$17,$5C,$94,$6B,$0F,$24,$7A,$DF,$1D,$F6,$8E,$27
	.dc.b	$33,$8B,$97,$D6,$66,$7A,$5F,$96,$B9,$E1,$E1,$AE,$D3,$E7,$BE,$51
	.dc.b	$83,$DD,$3B,$7F,$DC,$E7,$58,$DB,$8A,$EA,$F9,$4E,$47,$9A,$79,$61
	.dc.b	$F0,$70,$F2,$CF,$B2,$B8,$9F,$09,$E9,$87,$96,$F8,$C7,$85,$B3,$C2
	.dc.b	$7D,$C6,$E7,$C4,$72,$3D,$66,$2F,$C5,$87,$A9,$B7,$D2,$B9,$5F,$8E
	.dc.b	$73,$7C,$4F,$11,$63,$C4,$FA,$E7,$89,$E4,$7C,$31,$CD,$E2,$D3,$8B
	.dc.b	$F6,$9A,$FC,$AC,$3E,$3B,$99,$EA,$F7,$3A,$BD,$DF,$A0,$79,$5D,$BC
	.dc.b	$4D,$9C,$4D,$CF,$8E,$37,$1F,$AD,$E2,$7B,$9F,$9F,$6C,$7E,$51,$AF
	.dc.b	$E4,$9B,$F0,$4F,$93,$FE,$E3,$29,$CC,$61,$BB,$7F,$4C,$6C,$EE,$FF
	.dc.b	$A7,$AB,$E9,$5F,$AB,$FA,$47,$C3,$EB,$3F,$DF,$46,$2F,$F6,$17,$C1
	.dc.b	$FD,$53,$67,$75,$C7,$3E,$F7,$C6,$F7,$9F,$AA,$FF,$85,$61,$C6,$F7
	.dc.b	$87,$1B,$FE,$CB,$87,$1B,$DE,$1C,$6F,$D7,$87,$1B,$DE,$1C,$56,$1C
	.dc.b	$6B,$77,$C6,$7F,$B8,$DF,$17,$77,$C6,$F8,$46,$AB,$83,$F2,$5B,$1A
	.dc.b	$8E,$27,$C6,$8D,$D7,$59,$D1,$B7,$C0,$66,$BC,$07,$A5,$CC,$F6,$8C
	.dc.b	$D7,$5C,$63,$FC,$6E,$FA,$CF,$1A,$F0,$9C,$67,$EF,$46,$6B,$8E,$3F
	.dc.b	$62,$72,$63,$1F,$8F,$38,$8E,$D5,$C5,$FB,$71,$9A,$E5,$18,$23,$82
	.dc.b	$57,$48,$CB,$F8,$07,$93,$98,$FB,$F3,$89,$FB,$A7,$94,$FA,$8E,$5F
	.dc.b	$A2,$74,$32,$7A,$0D,$63,$F3,$23,$86,$3F,$4F,$1F,$C2,$71,$FF,$9A
	.dc.b	$7A,$BF,$4F,$5F,$E8,$8F,$A6,$BC,$3F,$E6,$EF,$31,$18,$CF,$FC,$93
	.dc.b	$1F,$E1,$C7,$FD,$1B,$79,$C3,$FC,$E9,$56,$B8,$D6,$8F,$F4,$9F,$D0
	.dc.b	$64,$8F,$FA,$5A,$6B,$FE,$94,$8B,$8F,$96,$3F,$C3,$DF,$94,$3E,$37
	.dc.b	$C9,$08,$C3,$A8,$CF,$E5,$2B,$DD,$6A,$BF,$C9,$73,$EB,$6E,$7A,$BE
	.dc.b	$BF,$46,$EE,$A4,$AF,$F1,$87,$A1,$FC,$65,$FF,$19,$4E,$DC,$FC,$7F
	.dc.b	$9A,$4F,$18,$FF,$2D,$FB,$DF,$C7,$DF,$F1,$D3,$5C,$F4,$FB,$A3,$F9
	.dc.b	$4E,$DF,$E3,$57,$4D,$8D,$1B,$E9,$B5,$D2,$2B,$E0,$8F,$93,$E6,$F9
	.dc.b	$23,$FC,$94,$D7,$FC,$DE,$DA,$FF,$22,$7D,$13,$C9,$E3,$97,$D2,$87
	.dc.b	$C9,$53,$38,$7F,$B4,$9F,$C2,$11,$FF,$6A,$FC,$CD,$7F,$DA,$71,$70
	.dc.b	$FF,$6D,$CF,$DE,$FB,$DF,$F7,$D5,$7F,$97,$DA,$FF,$A1,$1B,$3F,$E5
	.dc.b	$95,$F0,$DC,$7E,$02,$57,$17,$A4,$62,$39,$45,$AE,$53,$F8,$EF,$A1
	.dc.b	$7C,$37,$FB,$02,$FF,$A7,$C6,$1C,$07,$0E,$DA,$FF,$B2,$1C,$D7,$F9
	.dc.b	$BF,$E8,$1E,$E7,$E7,$1B,$5F,$E7,$DA,$DF,$7F,$F9,$63,$5F,$F7,$0C
	.dc.b	$FF,$BC,$1A,$FF,$AB,$FC,$6F,$A7,$EE,$7E,$CF,$AB,$F7,$51,$89,$FF
	.dc.b	$45,$F8,$2F,$7D,$F6,$BF,$BB,$6C,$F5,$C3,$F7,$9E,$47,$EF,$79,$E4
	.dc.b	$F2,$F1,$3F,$CD,$1C,$3F,$EB,$BC,$BF,$DE,$D7,$AA,$FF,$61,$47,$FB
	.dc.b	$E6,$F9,$C7,$19,$D5,$71,$BF,$C7,$4E,$4E,$DF,$61,$FC,$34,$8F,$87
	.dc.b	$E6,$EB,$DA,$53,$8A,$E9,$5F,$C5,$7C,$D9,$7E,$97,$F8,$C7,$D0,$E2
	.dc.b	$39,$45,$AF,$24,$FD,$5F,$4E,$57,$FA,$C7,$97,$FD,$44,$FF,$AC,$F8
	.dc.b	$FF,$C2,$6F,$FB,$62,$32,$D9,$E2,$FF,$B7,$27,$77,$F8,$39,$CB,$67
	.dc.b	$A1,$FC,$B7,$D8,$FE,$5C,$C3,$E9,$3F,$CC,$FE,$93,$BC,$FD,$36,$F3
	.dc.b	$7F,$BD,$38,$51,$8B,$FE,$7A,$71,$F6,$A3,$F8,$68,$FF,$7F,$B7,$1E
	.dc.b	$DF,$57,$DF,$E5,$A4,$EE,$2B,$F8,$0A,$4E,$32,$F2,$55,$CB,$C9,$4F
	.dc.b	$A5,$F5,$19,$C1,$9F,$A2,$7D,$C4,$8F,$37,$D5,$57,$C1,$17,$1F,$85
	.dc.b	$37,$0E,$F3,$6B,$AA,$9F,$95,$F4,$D2,$7E,$13,$F3,$7E,$77,$11,$ED
	.dc.b	$8B,$87,$CF,$9B,$87,$B2,$6D,$76,$9F,$AD,$E3,$48,$F9,$CF,$D9,$FA
	.dc.b	$3C,$47,$C9,$8B,$87,$DB,$9B,$87,$C8,$9B,$5E,$A9,$FC,$17,$A9,$23
	.dc.b	$ED,$3F,$85,$F9,$DC,$47,$D5,$8B,$87,$E1,$E6,$E1,$F5,$26,$D7,$C6
	.dc.b	$9F,$C7,$7B,$32,$3F,$0C,$FE,$47,$E9,$71,$1F,$81,$8B,$87,$E4,$E6
	.dc.b	$E1,$F8,$09,$B5,$F4,$A7,$F3,$5E,$0C,$8F,$C9,$3F,$9D,$F8,$DC,$47
	.dc.b	$E3,$62,$E1,$F9,$F9,$FD,$17,$8D,$FE,$FA,$F9,$CF,$C6,$B6,$FE,$FB
	.dc.b	$78,$FF,$47,$37,$8F,$F9,$8D,$3D,$D4,$F2,$81,$97,$0D,$3F,$16,$F7
	.dc.b	$5F,$3C,$F9,$4F,$F0,$05,$74,$9E,$73,$9F,$4F,$AC,$E6,$78,$D7,$0B
	.dc.b	$FF,$4F,$FC,$FB,$9B,$B6,$3C,$6F,$C4,$70,$C7,$9D,$FF,$A9,$A7,$58
	.dc.b	$F5,$C3,$95,$E8,$F1,$E3,$DD,$B8,$65,$E8,$76,$9E,$2F,$82,$FA,$E1
	.dc.b	$DE,$F9,$4F,$0C,$3C,$18,$FA,$27,$37,$C2,$BB,$9D,$67,$58,$F8,$67
	.dc.b	$53,$CF,$0F,$87,$5A,$8F,$15,$F4,$3C,$38,$F9,$B1,$F9,$98,$7C,$7A
	.dc.b	$FE,$6E,$D7,$D1,$8F,$F7,$C8,$C4,$7F,$AB,$38,$DF,$1D,$FE,$00,$F8
	.dc.b	$04,$62,$7F,$C0,$D1,$89,$D4,$E6,$7B,$FA,$AF,$07,$B3,$63,$FC,$1D
	.dc.b	$82,$31,$1C,$27,$67,$C3,$8E,$19,$7E,$1F,$E7,$FF,$97,$F4,$66,$A5
	.dc.b	$FD,$5B,$DF,$07,$FB,$A2,$F7,$39,$AD,$E1,$F1,$AF,$73,$FA,$0A,$DC
	.dc.b	$E6,$B7,$7F,$8D,$BD,$DA,$F7,$39,$AD,$CF,$E8,$2B,$75,$A8,$C4,$E9
	.dc.b	$FD,$4C,$EB,$C0,$63,$9C,$BE,$35,$EF,$0E,$58,$70,$8F,$CE,$71,$FF
	.dc.b	$87,$F9,$38,$CF,$ED,$E3,$18,$6B,$CD,$FD,$FB,$6B,$7A,$FE,$DE,$B7
	.dc.b	$7C,$6F,$FC,$73,$7C,$75,$D9,$FD,$BE,$33,$3B,$E1,$F0,$F3,$FF,$89
	.dc.b	$E5,$18,$BE,$1E,$BD,$BF,$EC,$35,$A9,$FF,$1C,$B6,$3F,$C5,$AD,$BE
	.dc.b	$95,$C7,$1F,$F1,$F6,$17,$E7,$DF,$3F,$4C,$7C,$FD,$CF,$FC,$8B,$3A
	.dc.b	$6C,$E9,$A4,$CF,$FC,$7B,$3B,$6D,$73,$73,$F1,$FE,$13,$C5,$F3,$F6
	.dc.b	$FB,$C5,$9E,$73,$74,$FF,$26,$60,$AC,$5F,$F8,$6E,$15,$8B,$FF,$03
	.dc.b	$C3,$E7,$FE,$01,$F0,$7E,$06,$DF,$BA,$2D,$7A,$A6,$E3,$E0,$9B,$5D
	.dc.b	$E6,$D7,$6F,$6F,$E6,$1F,$37,$DB,$B5,$F2,$A2,$CF,$C6,$9B,$87,$B2
	.dc.b	$6D,$7A,$E6,$D7,$AB,$E7,$FE,$C1,$F6,$7E,$72,$D7,$D6,$8B,$3F,$4A
	.dc.b	$6E,$1F,$22,$6D,$7C,$79,$B5,$F1,$BE,$7F,$F0,$9E,$F7,$E1,$5B,$5F
	.dc.b	$82,$70,$FE,$13,$E0,$AD,$7E,$01,$B8,$FD,$66,$CF,$D5,$F9,$FF,$C8
	.dc.b	$1F,$91,$F9,$FB,$1F,$8E,$70,$FE,$43,$E5,$6E,$1F,$8C,$F1,$BF,$31
	.dc.b	$1F,$99,$6C,$FF,$72,$C7,$F7,$7F,$35,$FC,$D3,$5F,$D4,$36,$74,$7F
	.dc.b	$78,$7F,$9C,$F0,$6B,$C5,$55,$25,$70,$C9,$7C,$23,$0F,$58,$61,$E2
	.dc.b	$BE,$75,$A6,$CE,$A3,$C5,$0C,$3B,$CF,$EA,$BE,$6B,$F6,$C7,$ED,$FC
	.dc.b	$5B,$7F,$B4,$8B,$3F,$B0,$9B,$87,$E9,$AB,$F7,$C7,$EF,$FE,$7E,$BF
	.dc.b	$BC,$2D,$7E,$E2,$2C,$FF,$AC,$7E,$7F,$F8,$47,$C0,$3E,$7E,$C6,$1B
	.dc.b	$87,$FB,$05,$B8,$7F,$AC,$5B,$7F,$EB,$B6,$CF,$FA,$46,$FA,$3F,$37
	.dc.b	$E1,$5F,$3F,$FC,$9A,$FF,$9A,$C6,$3C,$7F,$1F,$C3,$7F,$23,$8C,$5F
	.dc.b	$1B,$E0,$FF,$8D,$23,$39,$7F,$B0,$EB,$85,$6A,$3F,$C7,$3F,$3E,$9F
	.dc.b	$E8,$71,$FE,$B8,$8C,$5F,$FB,$F1,$B1,$FF,$00,$5A,$EE,$7F,$9B,$28
	.dc.b	$DE,$3E,$AF,$AE,$F0,$4E,$6B,$F2,$1C,$13,$0F,$C3,$70,$5A,$FC,$31
	.dc.b	$F9,$0E,$25,$67,$1F,$C7,$55,$BF,$C5,$CD,$BE,$13,$7C,$66,$EF,$A0
	.dc.b	$FB,$A7,$DE,$FA,$0B,$8F,$FB,$EE,$2E,$1F,$EE,$B9,$B9,$7C,$6F,$A0
	.dc.b	$7C,$1F,$C0,$6E,$1F,$A7,$F3,$DF,$40,$F1,$75,$72,$BF,$59,$93,$0E
	.dc.b	$2F,$C0,$C3,$8D,$EE,$3E,$1F,$5F,$F9,$C3,$E2,$FF,$7B,$06,$FC,$66
	.dc.b	$19,$8F,$ED,$A7,$4D,$9C,$C7,$43,$25,$9E,$73,$C5,$FB,$BF,$40,$F9
	.dc.b	$5A,$7F,$75,$CB,$DF,$E0,$E2,$FD,$B5,$B6,$D6,$F6,$FF,$41,$46,$0F
	.dc.b	$87,$F3,$FF,$DE,$5F,$F4,$6E,$48,$C3,$FE,$47,$D5,$F1,$3F,$90,$9C
	.dc.b	$CE,$FD,$2C,$1E,$5A,$DD,$F6,$8F,$DE,$FC,$FC,$8F,$E3,$DC,$FF,$86
	.dc.b	$F9,$EC,$61,$F9,$28,$FC,$4D,$63,$97,$FC,$53,$1B,$BE,$3D,$0A,$FE
	.dc.b	$13,$F3,$FF,$C2,$B1,$B5,$A6,$E3,$D6,$D1,$8C,$7A,$C6,$2B,$F0,$F5
	.dc.b	$FE,$A4,$1F,$3F,$F8,$17,$F8,$07,$C8,$FF,$8F,$23,$11,$F2,$BD,$BF
	.dc.b	$F5,$BC,$63,$C4,$56,$F9,$FE,$7D,$F5,$AD,$4E,$63,$37,$CA,$BE,$AD
	.dc.b	$F2,$C3,$15,$A3,$CB,$3A,$BE,$51,$E3,$32,$7A,$27,$8B,$9C,$3C,$31
	.dc.b	$88,$F2,$57,$38,$C4,$7F,$B5,$4E,$62,$FC,$8E,$0E,$27,$73,$07,$1A
	.dc.b	$EE,$E3,$0F,$1C,$F2,$9F,$97,$3C,$8E,$D8,$7B,$1C,$4E,$F0,$F9,$71
	.dc.b	$D5,$F4,$3D,$4F,$3C,$6D,$F9,$15,$E2,$3B,$BC,$4F,$43,$8B,$F2,$DF
	.dc.b	$86,$74,$79,$30,$F8,$B3,$A6,$C7,$82,$78,$4E,$63,$C9,$5E,$1C,$3D
	.dc.b	$33,$98,$F0,$4F,$1B,$E7,$1B,$8F,$4D,$7B,$A3,$66,$63,$EE,$46,$4E
	.dc.b	$AF,$33,$C2,$E5,$E6,$7C,$07,$71,$CE,$3E,$7D,$FA,$9E,$51,$E0,$9E
	.dc.b	$51,$E0,$39,$1D,$0F,$8C,$78,$0C,$9F,$95,$DD,$EC,$F9,$9B,$1F,$9D
	.dc.b	$6B,$B8,$FE,$2F,$F0,$0F,$88,$FF,$11,$CF,$FA,$72,$73,$18,$AE,$35
	.dc.b	$BD,$97,$B2,$C6,$5C,$07,$18,$FE,$FB,$E7,$DF,$13,$FC,$A2,$30,$DC
	.dc.b	$FC,$38,$7E,$BA,$F8,$C7,$F2,$EB,$E3,$87,$EC,$9A,$FE,$B2,$76,$F9
	.dc.b	$32,$F6,$E3,$DA,$38,$63,$DD,$D6,$3D,$EB,$78,$F7,$AD,$53,$1E,$F8
	.dc.b	$7F,$18,$C7,$BC,$FB,$F0,$E2,$FC,$6A,$DE,$1C,$E3,$B6,$31,$FE,$07
	.dc.b	$3A,$78,$63,$E0,$C3,$A9,$DA,$75,$3E,$39,$D6,$1E,$08,$F1,$5F,$B9
	.dc.b	$87,$C5,$8D,$5F,$3A,$F6,$DF,$48,$EF,$3C,$E3,$B9,$97,$89,$99,$E2
	.dc.b	$78,$AF,$D9,$8F,$CF,$C3,$DB,$39,$7C,$53,$E4,$9D,$1F,$2E,$BA,$C6
	.dc.b	$71,$F9,$15,$C0,$95,$C2,$32,$F1,$9F,$0C,$60,$93,$F7,$8E,$2E,$1F
	.dc.b	$94,$70,$35,$3D,$5F,$A6,$64,$F9,$A7,$B0,$F3,$1C,$23,$F3,$2E,$49
	.dc.b	$1E,$57,$FA,$5F,$27,$F8,$FB,$A7,$FC,$13,$7F,$DA,$BE,$7E,$F0,$7F
	.dc.b	$9C,$5B,$6B,$E0,$FF,$7A,$BC,$C7,$F9,$22,$6F,$19,$B3,$8F,$9F,$97
	.dc.b	$FE,$02,$77,$38,$8F,$EF,$D1,$71,$F2,$47,$F9,$0B,$E7,$DF,$31,$E4
	.dc.b	$84,$61,$D4,$67,$E7,$D7,$BA,$D5,$7F,$9A,$0E,$35,$93,$84,$F5,$F9
	.dc.b	$FF,$F2,$A4,$FF,$94,$9A,$F1,$6E,$1F,$E8,$4F,$9F,$FE,$39,$3B,$ED
	.dc.b	$E4,$47,$FC,$B7,$1A,$7C,$5E,$D3,$9F,$D8,$D8,$FF,$30,$B5,$DF,$46
	.dc.b	$A6,$46,$9E,$31,$88,$EB,$17,$0F,$F4,$CD,$7F,$34,$7F,$D3,$7E,$5B
	.dc.b	$E4,$BE,$63,$6F,$FC,$E2,$D7,$E4,$7C,$FB,$C8,$DD,$49,$DB,$F5,$67
	.dc.b	$51,$FE,$40,$F9,$8F,$C2,$91,$FF,$21,$FC,$FD,$7F,$C7,$D1,$71,$FF
	.dc.b	$1E,$CF,$DF,$F9,$FF,$F7,$15,$7F,$A1,$1A,$FD,$B6,$CF,$FA,$D9,$7C
	.dc.b	$1F,$1F,$72,$57,$17,$A4,$62,$39,$45,$AE,$53,$F8,$FF,$9F,$7C,$1D
	.dc.b	$25,$FF,$A5,$BE,$73,$EB,$5F,$FB,$1A,$B7,$3F,$E1,$E0,$E1,$FF,$10
	.dc.b	$B5,$FF,$26,$B5,$FF,$6B,$67,$FE,$D1,$7F,$DC,$4D,$7F,$D8,$3E,$3A
	.dc.b	$31,$5F,$E6,$4E,$8F,$DC,$CE,$7D,$8B,$5A,$F0,$3A,$C7,$EE,$2B,$F7
	.dc.b	$3E,$17,$EE,$B8,$E6,$5F,$E7,$88,$CB,$FE,$7A,$8C,$BF,$E7,$58,$CB
	.dc.b	$FE,$77,$8C,$E5,$FE,$76,$9F,$E0,$BF,$3F,$2B,$FD,$06,$FF,$08,$F9
	.dc.b	$FC,$47,$F9,$FE,$2D,$7F,$9F,$A7,$F8,$6F,$93,$2B,$FD,$16,$FF,$10
	.dc.b	$F9,$FC,$47,$FA,$1E,$2D,$7F,$A1,$A7,$F8,$47,$07,$FA,$92,$BE,$C7
	.dc.b	$A1,$83,$F8,$66,$FF,$E7,$EB,$FF,$7F,$D6,$EB,$F8,$43,$63,$35,$C2
	.dc.b	$79,$1D,$08,$7F,$0F,$AE,$9B,$FF,$86,$BC,$6F,$F7,$7F,$A9,$FC,$A2
	.dc.b	$A1,$FE,$A6,$7F,$96,$FC,$BF,$7A,$E0,$D9,$FE,$68,$DA,$FF,$7C,$C7
	.dc.b	$0F,$62,$3F,$EB,$89,$FF,$79,$B7,$1E,$BF,$3F,$DB,$E5,$64,$EE,$2B
	.dc.b	$FE,$BD,$A4,$E3,$4F,$F5,$EC,$2B,$19,$79,$BE,$7F,$D2,$63,$0F,$3D
	.dc.b	$7B,$08,$F9,$7E,$7E,$BD,$E2,$E1,$DA,$6D,$FC,$29,$B5,$F0,$A7,$E3
	.dc.b	$FC,$FC,$8F,$80,$FC,$AF,$9B,$C4,$7A,$E2,$E1,$EA,$9B,$87,$CE,$9B
	.dc.b	$5F,$3A,$7E,$9F,$85,$23,$E6,$3F,$5B,$E7,$F1,$1F,$1E,$2E,$1F,$1A
	.dc.b	$6E,$1F,$6A,$6D,$7D,$A9,$FB,$FE,$94,$8F,$B0,$FE,$0B,$E6,$F1,$1F
	.dc.b	$4E,$2E,$1F,$4A,$6E,$1F,$86,$9B,$5F,$86,$9F,$C5,$FA,$92,$3F,$08
	.dc.b	$FE,$3B,$E7,$F1,$1F,$7E,$2E,$1F,$7A,$6E,$1F,$92,$9B,$5F,$92,$9F
	.dc.b	$CB,$FC,$FC,$8F,$C8,$3F,$9A,$F9,$FC,$47,$E2,$E2,$E1,$F8,$A9,$FC
	.dc.b	$FF,$C5,$FF,$BC,$FE,$73,$F3,$ED,$BF,$CF,$B7,$A1,$C6,$AC,$29,$A4
	.dc.b	$AF,$F4,$4F,$B9,$B7,$F5,$35,$D5,$7C,$F3,$E1,$3F,$B9,$56,$E7,$31
	.dc.b	$8A,$FF,$49,$C6,$3D,$1E,$B1,$8F,$9F,$73,$75,$C7,$A5,$6A,$FF,$D3
	.dc.b	$71,$85,$8F,$78,$C3,$E1,$BE,$73,$C4,$F0,$1C,$30,$F1,$D7,$06,$E3
	.dc.b	$E7,$9F,$F5,$4D,$72,$78,$1E,$28,$D3,$D2,$FC,$51,$89,$E1,$18,$C3
	.dc.b	$C5,$87,$7B,$E5,$1E,$5A,$DE,$1E,$28,$DE,$1D,$AF,$85,$AF,$6F,$23
	.dc.b	$CA,$78,$EF,$93,$88,$DB,$F0,$8F,$A6,$53,$66,$27,$E7,$3A,$F7,$DC
	.dc.b	$8F,$8B,$FE,$ED,$18,$6C,$66,$7A,$9F,$EC,$D7,$0F,$5F,$4D,$E0,$FF
	.dc.b	$7F,$C1,$18,$9E,$31,$89,$D4,$E6,$79,$47,$CC,$8E,$1A,$FE,$47,$E7
	.dc.b	$FF,$98,$F4,$66,$A5,$FC,$6C,$3A,$3F,$22,$B6,$E7,$0E,$7A,$FE,$19
	.dc.b	$B1,$F9,$26,$A9,$E3,$8F,$E4,$E7,$37,$DA,$31,$87,$E4,$EB,$78,$75
	.dc.b	$8F,$2C,$EB,$0F,$27,$80,$CB,$37,$B9,$F3,$5F,$3A,$D4,$60,$97,$CC
	.dc.b	$F8,$91,$A7,$15,$FD,$9B,$83,$0F,$F6,$8F,$5A,$DF,$F6,$89,$BC,$BF
	.dc.b	$C2,$A6,$F2,$FF,$9C,$F3,$FD,$B4,$E5,$B1,$9B,$FD,$4C,$E5,$BF,$85
	.dc.b	$BE,$7F,$D3,$D3,$EA,$A4,$BE,$7F,$38,$FF,$80,$E3,$82,$C1,$7C,$67
	.dc.b	$84,$E6,$31,$87,$2A,$FF,$5B,$56,$9E,$35,$D6,$75,$87,$8B,$0B,$6F
	.dc.b	$9F,$BF,$3F,$86,$E1,$DC,$F3,$38,$C7,$A9,$E0,$71,$97,$9D,$B3,$C0
	.dc.b	$E5,$3E,$6A,$F0,$FC,$F7,$8B,$DE,$B5,$AC,$FE,$3C,$62,$FB,$C6,$27
	.dc.b	$59,$2F,$94,$61,$B5,$C2,$75,$3C,$2F,$E0,$FB,$3F,$0B,$D9,$B3,$D1
	.dc.b	$B7,$C5,$BA,$7D,$36,$CF,$D9,$6C,$FD,$27,$E6,$FC,$F7,$CE,$F9,$AB
	.dc.b	$5E,$86,$D7,$95,$B7,$F7,$DB,$3F,$85,$6C,$FD,$E7,$EC,$FC,$F7,$DA
	.dc.b	$F7,$2D,$7C,$46,$D7,$C1,$6D,$FE,$2D,$B3,$F9,$16,$CF,$E2,$9F,$C2
	.dc.b	$FC,$04,$9F,$9F,$F3,$B6,$BE,$63,$6F,$F2,$CD,$9F,$CE,$3F,$8E,$3F
	.dc.b	$1F,$F3,$D6,$3E,$CF,$CF,$7E,$48,$FC,$9F,$8B,$63,$EE,$F8,$38,$69
	.dc.b	$F8,$08,$FC,$CF,$CF,$48,$FC,$1F,$CC,$DC,$FF,$2B,$F2,$DF,$CF,$76
	.dc.b	$FD,$33,$FC,$FF,$86,$F8,$DF,$EB,$CC,$3F,$2B,$39,$6C,$66,$F9,$AA
	.dc.b	$E3,$7C,$9B,$5C,$1B,$5C,$1F,$D8,$FE,$0E,$F9,$EF,$D8,$D6,$1A,$FE
	.dc.b	$86,$34,$FE,$78,$D1,$4D,$7C,$1F,$4F,$9F,$EA,$D8,$C9,$CE,$32,$D3
	.dc.b	$B7,$CE,$79,$3E,$73,$CA,$D8,$C9,$E3,$8C,$B4,$CE,$5F,$B4,$C1,$3A
	.dc.b	$8C,$3F,$C5,$57,$CF,$31,$FE,$31,$CC,$F8,$D7,$83,$BA,$D6,$5F,$C2
	.dc.b	$2B,$53,$9B,$FE,$8B,$87,$F8,$02,$F6,$FF,$76,$6B,$98,$DB,$FE,$B1
	.dc.b	$C9,$F2,$FF,$53,$E6,$7E,$AB,$72,$FA,$4D,$AF,$73,$72,$FE,$AB,$C7
	.dc.b	$FE,$09,$AF,$C5,$F2,$FE,$31,$A7,$E2,$8B,$7F,$8A,$6D,$72,$E4,$FE
	.dc.b	$C9,$1F,$90,$AE,$4D,$3F,$09,$F2,$BF,$8A,$F9,$9F,$C5,$B6,$32,$7E
	.dc.b	$26,$32,$D3,$F9,$8E,$0F,$9D,$7B,$FF,$D2,$7E,$39,$F7,$2F,$73,$EC
	.dc.b	$56,$B0,$FE,$73,$5A,$9C,$DF,$CC,$C3,$FA,$C5,$ED,$FF,$13,$B5,$FF
	.dc.b	$1A,$B5,$FE,$BD,$B7,$F9,$AB,$0F,$D8,$C7,$F8,$A9,$AF,$EE,$0C,$1F
	.dc.b	$BC,$3F,$B3,$3C,$3D,$4D,$46,$DB,$19,$39,$96,$3F,$AC,$8F,$C2,$DF
	.dc.b	$0D,$FD,$AF,$F0,$A7,$E1,$8F,$89,$FC,$D5,$6B,$0F,$EB,$75,$A9,$CD
	.dc.b	$FC,$3C,$3A,$36,$F8,$47,$F9,$29,$AF,$F9,$63,$F2,$AF,$C9,$E1,$FD
	.dc.b	$9B,$27,$9D,$FE,$8A,$8F,$B3,$3A,$AF,$EE,$35,$AF,$80,$FD,$74,$E6
	.dc.b	$7A,$74,$78,$E3,$11,$FE,$98,$6A,$F4,$9D,$6F,$9C,$F8,$5F,$86,$BE
	.dc.b	$43,$F8,$CC,$61,$BA,$7C,$29,$D3,$FC,$66,$74,$FF,$1B,$F9,$8F,$E4
	.dc.b	$27,$F2,$2F,$9B,$B1,$92,$65,$FC,$2E,$32,$FD,$63,$FD,$72,$3E,$7D
	.dc.b	$F2,$DF,$EB,$2E,$BF,$0C,$E6,$3F,$CE,$FF,$03,$FE,$2B,$AD,$CF,$F8
	.dc.b	$B6,$B7,$18,$8F,$F5,$DC,$EE,$BA,$69,$FE,$52,$AD,$57,$89,$E1,$7C
	.dc.b	$AB,$BC,$F8,$4E,$D5,$C6,$33,$5C,$23,$13,$FE,$98,$BF,$03,$C3,$1A
	.dc.b	$F9,$0E,$83,$0F,$03,$C4,$CB,$87,$85,$F7,$9E,$71,$E9,$C3,$DB,$18
	.dc.b	$8F,$F6,$89,$CB,$0F,$2B,$CA,$76,$F8,$8E,$47,$9C,$C1,$CA,$BC,$0E
	.dc.b	$30,$F5,$E1,$E5,$8E,$27,$AE,$FB,$B8,$8D,$3C,$EF,$C3,$7F,$4A,$30
	.dc.b	$78,$5C,$5F,$6B,$E7,$3A,$3B,$61,$EF,$9D,$36,$3E,$03,$6B,$D2,$74
	.dc.b	$8C,$9E,$B2,$B9,$37,$7F,$78,$F3,$96,$3C,$CD,$7C,$67,$38,$E2,$FD
	.dc.b	$B6,$BF,$44,$E2,$66,$3A,$B4,$C9,$B3,$AC,$79,$63,$71,$9C,$3E,$6D
	.dc.b	$FC,$28,$F6,$7C,$03,$E0,$3F,$36,$D8,$FC,$73,$5F,$CA,$37,$E1,$9F
	.dc.b	$19,$FE,$98,$94,$E6,$30,$DF,$1B,$FD,$3C,$D9,$DD,$FF,$A8,$55,$F4
	.dc.b	$AF,$D1,$FD,$03,$E3,$75,$AF,$EE,$63,$17,$FA,$8B,$E0,$FE,$A1,$B3
	.dc.b	$BA,$E3,$AF,$7B,$E3,$7B,$CF,$D5,$7F,$C1,$70,$E3,$7B,$C3,$8D,$FF
	.dc.b	$AD,$F0,$E3,$7B,$C3,$8D,$FA,$F0,$E3,$7B,$C3,$8A,$C3,$8D,$6E,$F8
	.dc.b	$CF,$FB,$0E,$F8,$BB,$BE,$37,$C2,$35,$5C,$1F,$92,$D8,$D4,$71,$3E
	.dc.b	$34,$6E,$BA,$CE,$8D,$BE,$03,$35,$E0,$3D,$2E,$67,$B4,$66,$BA,$E3
	.dc.b	$1F,$E3,$37,$D6,$78,$D7,$84,$E3,$3F,$7A,$33,$5C,$71,$FB,$13,$93
	.dc.b	$18,$FC,$79,$C4,$76,$AE,$2F,$DB,$8C,$D7,$28,$C1,$1C,$12,$BA,$46
	.dc.b	$5F,$C0,$3C,$9C,$C7,$DF,$9C,$4F,$DD,$3C,$A7,$D4,$72,$FD,$13,$A1
	.dc.b	$93,$D0,$6B,$1F,$99,$1C,$31,$FA,$78,$FE,$13,$97,$FB,$BF,$CC,$FE
	.dc.b	$9E,$9C,$1F,$E5,$D8,$C3,$71,$FF,$46,$7C,$A6,$1F,$F0,$07,$3A,$B1
	.dc.b	$87,$49,$FF,$49,$46,$25,$57,$E6,$79,$78,$2F,$69,$F0,$98,$6E,$3F
	.dc.b	$E9,$BA,$E0,$FF,$8A,$23,$2B,$4F,$4A,$BE,$97,$C2,$3F,$C6,$FE,$43
	.dc.b	$E0,$7F,$83,$44,$8C,$38,$07,$F9,$6E,$30,$E2,$36,$ED,$B7,$FE,$8C
	.dc.b	$FC,$AD,$EF,$7C,$7A,$46,$1F,$F3,$5D,$70,$E0,$FE,$31,$F8,$73,$E3
	.dc.b	$BB,$D2,$30,$F8,$9A,$F9,$DA,$FB,$1C,$E3,$FE,$9E,$E3,$F0,$4F,$F9
	.dc.b	$61,$B1,$FE,$A2,$9D,$DF,$F2,$9F,$C1,$FE,$A5,$FF,$A0,$17,$C1,$7A
	.dc.b	$DB,$19,$E0,$C1,$99,$91,$92,$1F,$E6,$89,$E2,$D8,$E2,$FF,$A3,$BE
	.dc.b	$09,$F2,$FF,$09,$B1,$F4,$9A,$FD,$76,$BF,$71,$CE,$5F,$EB,$EF,$84
	.dc.b	$7C,$C7,$D3,$6D,$77,$BF,$CB,$7C,$DF,$E5,$FB,$A4,$FF,$BA,$E3,$2F
	.dc.b	$FB,$B6,$32,$FF,$B6,$63,$EF,$7C,$F7,$C8,$8F,$F4,$2B,$EE,$C7,$FD
	.dc.b	$9E,$DD,$BF,$AB,$3B,$8C,$ED,$CB,$59,$7F,$EB,$0A,$D5,$F0,$AC,$CF
	.dc.b	$6A,$F0,$C7,$EB,$FE,$73,$F6,$1F,$3B,$2B,$F3,$78,$FE,$9B,$BB,$FC
	.dc.b	$EF,$5B,$9C,$E3,$FE,$DD,$BE,$11,$FE,$F5,$9F,$CD,$7C,$F3,$E2,$3F
	.dc.b	$2A,$D9,$FD,$2B,$5F,$D5,$35,$FD,$7B,$9C,$7F,$DE,$3F,$3C,$F9,$9F
	.dc.b	$D3,$36,$BF,$D9,$F5,$BC,$BF,$DE,$1E,$9F,$F0,$BF,$32,$57,$A0,$FE
	.dc.b	$27,$1F,$C3,$CC,$BF,$C1,$67,$FD,$F1,$39,$8E,$1A,$FF,$A6,$2B,$79
	.dc.b	$7C,$5F,$0F,$FD,$31,$5A,$53,$FC,$7B,$F0,$EF,$F5,$FC,$E4,$FE,$3C
	.dc.b	$67,$D7,$8E,$48,$78,$CF,$F5,$44,$72,$71,$82,$3F,$2B,$18,$7F,$D9
	.dc.b	$4F,$A1,$AE,$7A,$3C,$87,$97,$D8,$91,$B8,$F9,$BF,$2F,$F3,$BA,$64
	.dc.b	$FF,$BF,$38,$13,$A8,$C4,$E9,$FF,$6F,$CE,$9E,$13,$9D,$7E,$DD,$71
	.dc.b	$9F,$44,$FC,$72,$3E,$98,$F9,$5F,$3F,$F2,$FE,$5E,$4F,$B6,$3E,$77
	.dc.b	$B3,$23,$E0,$3A,$7E,$27,$CF,$7D,$32,$3F,$17,$E0,$2B,$F2,$7E,$7F
	.dc.b	$EC,$18,$CF,$F4,$5B,$A3,$33,$F3,$5D,$72,$FF,$42,$E1,$FA,$9E,$27
	.dc.b	$E0,$7C,$59,$7F,$4E,$3F,$08,$61,$FB,$11,$87,$47,$D9,$75,$8F,$8A
	.dc.b	$7F,$19,$E9,$7E,$37,$E7,$E5,$FD,$F8,$CB,$F8,$18,$CB,$F8,$58,$CB
	.dc.b	$F8,$78,$CD,$7B,$A7,$F3,$1E,$47,$E6,$7E,$7E,$5F,$E2,$E3,$2F,$E3
	.dc.b	$63,$2F,$E4,$63,$2F,$E4,$E3,$35,$F2,$A7,$F4,$9D,$BF,$A5,$F9,$C9
	.dc.b	$7F,$97,$8C,$BF,$99,$8C,$BF,$9D,$8C,$BF,$9F,$8C,$D7,$D6,$9F,$D8
	.dc.b	$7A,$FF,$B1,$F5,$E5,$FE,$8E,$32,$FE,$96,$32,$FE,$A6,$32,$FE,$AE
	.dc.b	$33,$5F,$82,$9F,$F0,$CF,$C3,$3E,$72,$FC,$E4,$C3,$F8,$3C,$E5,$B3
	.dc.b	$97,$F7,$30,$9E,$73,$D3,$D3,$FF,$4A,$7B,$6D,$3F,$19,$6F,$8F,$CC
	.dc.b	$61,$B8,$B9,$15,$FE,$20,$BE,$53,$98,$C5,$F1,$8F,$F0,$CC,$EA,$B8
	.dc.b	$DF,$47,$9E,$1D,$A9,$F3,$CF,$98,$FF,$58,$C3,$0E,$EE,$A3,$31,$89
	.dc.b	$E1,$AF,$83,$2F,$F5,$97,$CF,$BB,$7B,$56,$B1,$E9,$5F,$EB,$78,$C6
	.dc.b	$1C,$F1,$EB,$3C,$AF,$8C,$62,$FF,$C6,$B7,$C2,$B9,$4E,$65,$87,$85
	.dc.b	$F2,$1C,$23,$55,$C6,$31,$8E,$9E,$38,$78,$23,$13,$B9,$F8,$51,$BB
	.dc.b	$ED,$1E,$C3,$99,$4D,$96,$79,$47,$0C,$7E,$05,$76,$C3,$CA,$7A,$E7
	.dc.b	$58,$FB,$A7,$FC,$A7,$7F,$2F,$0F,$F2,$DB,$5D,$1F,$E5,$BF,$7B,$FC
	.dc.b	$99,$F1,$FF,$E5,$A6,$C6,$7A,$93,$96,$C7,$F9,$8E,$9E,$AF,$F6,$28
	.dc.b	$C3,$5F,$0B,$5F,$B8,$D7,$EB,$BC,$74,$FC,$43,$72,$FF,$7C,$7C,$FF
	.dc.b	$F5,$2E,$DF,$9F,$39,$BE,$3E,$56,$0F,$AD,$5B,$8F,$99,$3B,$7F,$36
	.dc.b	$DC,$7D,$35,$C3,$C6,$FF,$43,$E5,$E2,$8E,$97,$C9,$D6,$38,$9D,$4E
	.dc.b	$4F,$B9,$8F,$86,$32,$4C,$BE,$0D,$F2,$AE,$13,$97,$D6,$72,$C3,$C7
	.dc.b	$5C,$23,$F1,$5C,$6F,$15,$C1,$A9,$FE,$09,$0E,$AC,$56,$EA,$CE,$74
	.dc.b	$FD,$0C,$E5,$B1,$9B,$EE,$F6,$8C,$DF,$4F,$9E,$FE,$49,$1F,$E2,$87
	.dc.b	$F8,$F3,$7A,$7F,$D6,$3D,$7F,$E4,$D9,$CE,$1F,$E1,$C7,$8B,$7A,$F7
	.dc.b	$7F,$DB,$67,$35,$C6,$73,$5F,$C8,$67,$FD,$8B,$5E,$1E,$AC,$E6,$9C
	.dc.b	$D7,$79,$CC,$63,$0F,$24,$63,$AB,$2D,$C3,$A1,$ED,$8C,$BE,$FF,$9F
	.dc.b	$E7,$F1,$7D,$1B,$19,$3E,$24,$65,$A6,$71,$F8,$51,$96,$C7,$CB,$F9
	.dc.b	$FF,$3F,$BF,$E8,$6C,$64,$FA,$11,$96,$99,$9F,$9D,$19,$6C,$7D,$7F
	.dc.b	$9C,$F8,$7F,$3F,$F1,$1B,$19,$3E,$E4,$65,$A6,$67,$ED,$46,$5B,$1F
	.dc.b	$83,$F9,$FF,$9F,$EB,$7D,$06,$C6,$4F,$C4,$46,$5A,$66,$7F,$0D,$19
	.dc.b	$6C,$7E,$3F,$DE,$FB,$67,$DC,$F6,$EC,$7E,$4D,$B3,$F8,$F6,$F0,$7E
	.dc.b	$4D,$E1,$3F,$99,$F8,$2B,$5F,$95,$F7,$3F,$0E,$7E,$23,$C5,$B5,$F9
	.dc.b	$FF,$07,$0D,$3F,$45,$1F,$8E,$F9,$69,$1F,$A5,$F9,$FB,$87,$F9,$FF
	.dc.b	$E5,$3F,$96,$ED,$F9,$C7,$F9,$7F,$0D,$F1,$BF,$D5,$98,$7F,$A6,$E7
	.dc.b	$2D,$8C,$DF,$35,$5C,$6F,$93,$6B,$83,$6B,$83,$FA,$8F,$6D,$F3,$FF
	.dc.b	$A6,$AC,$35,$FF,$50,$36,$BF,$D2,$B3,$C2,$7F,$81,$1C,$FE,$6B,$A7
	.dc.b	$CF,$D8,$FF,$54,$36,$B8,$B7,$8F,$FD,$5C,$F8,$FE,$7F,$C9,$F3,$F6
	.dc.b	$FF,$D6,$0D,$BF,$0B,$6F,$FD,$6C,$D9,$FF,$3B,$62,$8C,$57,$F8,$A6
	.dc.b	$31,$3F,$EC,$78,$7C,$F3,$24,$A6,$12,$75,$18,$8C,$4F,$07,$07,$47
	.dc.b	$0E,$E7,$8B,$FE,$C1,$35,$3F,$ED,$9F,$92,$7D,$47,$FB,$8E,$3C,$D5
	.dc.b	$C3,$2F,$F6,$A7,$CC,$FD,$9F,$9F,$FB,$4D,$AF,$F6,$B3,$71,$FA,$ED
	.dc.b	$AF,$37,$1F,$F7,$EA,$FC,$87,$2F,$E4,$5A,$7E,$3C,$B7,$F8,$F6,$D7
	.dc.b	$2E,$4F,$8F,$1F,$95,$AE,$4D,$3F,$DD,$1F,$3F,$F8,$FF,$5B,$F2,$0D
	.dc.b	$8C,$9F,$EE,$D8,$CB,$4F,$AD,$F3,$CF,$80,$F9,$79,$7D,$48,$C4,$FD
	.dc.b	$08,$46,$1F,$A6,$D7,$EC,$39,$CB,$EA,$B6,$7F,$01,$F7,$2F,$BA,$77
	.dc.b	$FF,$52,$C3,$AE,$3F,$B9,$7F,$0D,$7F,$B7,$5E,$A6,$5D,$36,$34,$D8
	.dc.b	$FE,$08,$70,$8E,$4F,$06,$9C,$0F,$CC,$7C,$FB,$E2,$BE,$1C,$FE,$5E
	.dc.b	$31,$3F,$0D,$B1,$F9,$46,$BF,$99,$73,$A7,$E5,$9B,$3F,$A1,$B7,$D4
	.dc.b	$B8,$3E,$15,$6B,$0F,$D3,$7D,$03,$93,$C7,$08,$C6,$3F,$55,$B1,$FA
	.dc.b	$66,$BF,$AB,$73,$97,$EA,$1B,$3E,$3F,$2F,$EB,$57,$82,$FF,$8D,$7C
	.dc.b	$FF,$F8,$1B,$E6,$1F,$40,$F9,$0C,$E3,$FB,$BC,$FF,$6D,$5A,$9C,$E2
	.dc.b	$9C,$DF,$1B,$E0,$7F,$0C,$C3,$B4,$66,$BF,$91,$FD,$07,$F2,$B3,$F9
	.dc.b	$67,$D0,$5B,$E3,$3F,$C4,$67,$0F,$F2,$A9,$B8,$7D,$AF,$7B,$F8,$87
	.dc.b	$D0,$3E,$67,$F8,$8F,$85,$DE,$31,$87,$F1,$1A,$DD,$6B,$5F,$E2,$55
	.dc.b	$AA,$FF,$67,$87,$17,$CE,$BF,$8B,$DF,$2A,$FE,$41,$39,$7B,$1C,$67
	.dc.b	$B6,$7B,$8E,$B7,$D3,$0F,$0D,$74,$9E,$07,$48,$CB,$D2,$78,$57,$38
	.dc.b	$FF,$6B,$9C,$05,$F8,$DE,$F7,$E2,$74,$70,$AE,$6E,$0F,$F6,$C1,$CE
	.dc.b	$79,$3B,$9E,$53,$F1,$E3,$07,$9E,$BC,$8F,$28,$EA,$6D,$C4,$F2,$3E
	.dc.b	$04,$69,$A6,$63,$E0,$35,$F5,$9D,$CF,$46,$1F,$36,$FC,$F1,$D8,$D5
	.dc.b	$78,$67,$58,$7C,$C9,$E6,$7C,$87,$0F,$B0,$C6,$1E,$C9,$E2,$D8,$F8
	.dc.b	$D3,$C1,$F0,$1C,$0F,$29,$A8,$E1,$5E,$F8,$F6,$9E,$6C,$3E,$0B,$4C
	.dc.b	$9C,$E3,$D4,$F3,$3A,$BC,$DF,$5E,$1F,$0B,$0F,$3C,$6A,$BE,$53,$ED
	.dc.b	$9D,$3C,$EB,$AF,$BD,$FE,$A5,$8F,$CD,$B5,$FD,$2B,$5F,$C8,$B9,$F8
	.dc.b	$47,$C6,$7F,$05,$9F,$F5,$8F,$BF,$FE,$04,$9C,$EB,$FE,$07,$BE,$17
	.dc.b	$B6,$C6,$6B,$84,$F3,$C3,$F6,$DF,$40,$F8,$CF,$F0,$7C,$61,$BA,$70
	.dc.b	$C3,$F6,$19,$78,$33,$FD,$36,$1C,$5F,$DA,$35,$FD,$54,$6F,$0F,$E0
	.dc.b	$8D,$BF,$F0,$ED,$6E,$7C,$15,$DB,$1F,$35,$78,$63,$BE,$3E,$47,$FC
	.dc.b	$4D,$7C,$67,$57,$DA,$F8,$D6,$AF,$82,$BE,$0F,$F8,$BF,$0F,$23,$6B
	.dc.b	$8B,$BA,$E3,$5C,$21,$5C,$1B,$5C,$27,$46,$E7,$46,$DC,$D7,$92,$33
	.dc.b	$3C,$63,$B9,$97,$C7,$1C,$63,$35,$F1,$F1,$8F,$C6,$3E,$9C,$71,$8C
	.dc.b	$E1,$F3,$A7,$47,$D6,$BE,$A7,$85,$E2,$70,$C7,$E6,$9F,$62,$FC,$31
	.dc.b	$C6,$F4,$FC,$F3,$89,$92,$5F,$84,$95,$D2,$3E,$84,$F0,$9C,$C6,$0E
	.dc.b	$33,$E1,$74,$F1,$7E,$8C,$E1,$F1,$1C,$0E,$53,$D1,$FB,$87,$A3,$DF
	.dc.b	$79,$6F,$A0,$FF,$38,$DF,$F9,$B6,$73,$18,$E1,$FF,$3F,$B6,$FF,$CF
	.dc.b	$DF,$41,$6B,$71,$FD,$C9,$AF,$FA,$6A,$75,$82,$AF,$D0,$73,$E9,$7B
	.dc.b	$0F,$80,$C3,$71,$E9,$82,$F8,$C7,$F8,$02,$F8,$DF,$03,$31,$FE,$81
	.dc.b	$F8,$E7,$BF,$E9,$84,$61,$FF,$4B,$35,$FF,$44,$7D,$07,$91,$CC,$FF
	.dc.b	$98,$5E,$2D,$9E,$11,$FE,$97,$6B,$FE,$9C,$FA,$07,$C6,$F8,$5B,$1F
	.dc.b	$EA,$56,$BF,$E9,$7E,$5F,$03,$FE,$99,$8C,$BF,$EA,$36,$BD,$67,$C9
	.dc.b	$18,$8F,$F5,$67,$D0,$72,$E3,$FF,$52,$AF,$80,$F8,$4D,$9F,$F5,$B3
	.dc.b	$BE,$4F,$F5,$9B,$63,$9C,$F0,$38,$F0,$C8,$E3,$19,$79,$C7,$CF,$F8
	.dc.b	$07,$CC,$52,$4F,$FB,$17,$E0,$9F,$1B,$88,$FF,$5E,$D7,$07,$FD,$20
	.dc.b	$D7,$35,$FE,$93,$E8,$FC,$DF,$D0,$7E,$3B,$AE,$57,$E2,$23,$F2,$26
	.dc.b	$1F,$F7,$6C,$61,$FF,$6B,$7D,$07,$BD,$3F,$15,$FF,$65,$37,$7F,$F6
	.dc.b	$7B,$B2,$6B,$F8,$D7,$5B,$FF,$B7,$1D,$6D,$99,$97,$E0,$8F,$D2,$FC
	.dc.b	$FF,$E9,$BE,$82,$4F,$E8,$23,$2F,$E8,$A3,$2F,$F9,$AF,$0F,$F7,$9C
	.dc.b	$FE,$77,$E8,$1F,$0D,$49,$5F,$EF,$3F,$9E,$78,$D8,$8F,$F7,$5D,$70
	.dc.b	$7F,$D7,$CD,$73,$5F,$3F,$BF,$FD,$63,$1F,$B4,$F2,$7F,$7A,$7E,$FB
	.dc.b	$D4,$91,$92,$1E,$71,$C7,$FC,$4A,$37,$19,$8F,$E2,$B3,$D2,$3F,$80
	.dc.b	$CF,$F0,$FF,$9C,$FF,$5A,$47,$F0,$43,$F8,$69,$9F,$5A,$39,$21,$D6
	.dc.b	$38,$B7,$E7,$FA,$7D,$04,$8F,$E3,$52,$9F,$01,$98,$E5,$38,$7E,$67
	.dc.b	$57,$F0,$DD,$FF,$2D,$CC,$E2,$54,$9C,$C6,$21,$1F,$A2,$79,$9F,$5F
	.dc.b	$A7,$A1,$D6,$E4,$64,$87,$CD,$39,$BF,$1C,$8F,$B1,$F9,$51,$F2,$FE
	.dc.b	$83,$E6,$7C,$B4,$AF,$83,$1F,$3F,$E6,$E4,$7C,$47,$4F,$CB,$FA,$0F
	.dc.b	$A8,$47,$E4,$FB,$B5,$FA,$1F,$41,$F6,$4C,$6B,$FD,$3F,$81,$19,$9F
	.dc.b	$A2,$EB,$6F,$93,$C8,$AF,$F2,$3D,$7D,$6F,$13,$F0,$87,$E1,$7C,$59
	.dc.b	$19,$21,$F8,$17,$2F,$E0,$E3,$8B,$F8,$08,$FC,$77,$9F,$F8,$FF,$A0
	.dc.b	$93,$F8,$48,$CB,$F8,$68,$CB,$F8,$D8,$CB,$F8,$F8,$CE,$7E,$F9,$FC
	.dc.b	$D7,$91,$F9,$BF,$A0,$97,$F9,$08,$CB,$F9,$28,$CB,$F9,$98,$CB,$F9
	.dc.b	$B8,$CD,$7C,$B9,$FD,$37,$D0,$7E,$9F,$E7,$65,$FE,$72,$32,$FE,$7A
	.dc.b	$32,$FE,$96,$32,$FE,$9E,$33,$5F,$5E,$7F,$65,$E0,$7E,$CF,$D6,$97
	.dc.b	$FA,$88,$CB,$FA,$A8,$CB,$FB,$18,$CB,$FB,$38,$CD,$7E,$0E,$7F,$C1
	.dc.b	$9F,$0E,$F9,$DB,$F3,$B3,$0F,$F0,$74,$E5,$B3,$97,$F8,$54,$27,$9C
	.dc.b	$F4,$F4,$BF,$61,$ED,$B4,$FC,$6D,$BE,$73,$31,$86,$E3,$FC,$62,$7F
	.dc.b	$C2,$8A,$B9,$39,$BD,$F2,$AB,$A4,$FF,$02,$AD,$CE,$6F,$75,$D2,$79
	.dc.b	$D3,$E7,$DF,$33,$FC,$E2,$3C,$19,$7F,$09,$9C,$EB,$D6,$31,$5B,$BE
	.dc.b	$7F,$40,$ED,$F5,$E3,$D6,$B5,$7C,$30,$E5,$3F,$C6,$31,$EA,$F8,$0F
	.dc.b	$E3,$78,$73,$76,$73,$9B,$3C,$31,$F5,$B6,$FF,$90,$5F,$1B,$E8,$71
	.dc.b	$2C,$ED,$F1,$61,$E6,$7C,$38,$FA,$A7,$51,$C7,$0E,$F5,$B8,$5F,$79
	.dc.b	$CE,$3E,$6A,$D5,$F2,$AD,$63,$F0,$F0,$F1,$D6,$8E,$93,$A8,$F2,$DF
	.dc.b	$F3,$B7,$81,$6F,$E4,$B6,$BE,$D3,$BF,$77,$FC,$8B,$F1,$FC,$63,$FD
	.dc.b	$6D,$D4,$8C,$47,$FA,$CA,$30,$EE,$78,$BC,$6D,$EA,$7F,$76,$8F,$13
	.dc.b	$5F,$C4,$35,$FB,$CD,$7F,$20,$FF,$7E,$D3,$EC,$B7,$4F,$EB,$1F,$41
	.dc.b	$FE,$38,$EE,$FB,$F3,$9B,$E7,$E5,$60,$FC,$2D,$6E,$3E,$EC,$ED,$EF
	.dc.b	$3B,$7F,$07,$3B,$C3,$C3,$39,$6C,$EB,$C6,$FE,$DC,$F4,$9D,$63,$EE
	.dc.b	$7C,$33,$AA,$E2,$49,$D5,$73,$C7,$8B,$EA,$3A,$19,$2C,$FF,$00,$87
	.dc.b	$7B,$E5,$7B,$5F,$F8,$02,$73,$18,$BF,$E1,$5E,$75,$9C,$9F,$B0,$2C
	.dc.b	$7E,$BD,$B8,$7F,$08,$E5,$FF,$1E,$C6,$F1,$F0,$3F,$B2,$FC,$4B,$F8
	.dc.b	$E3,$FC,$9B,$E2,$7E,$2F,$6F,$C6,$C5,$5A,$7F,$93,$CE,$AB,$FC,$29
	.dc.b	$39,$AF,$F6,$26,$0A,$D5,$FF,$87,$2B,$57,$FC,$AA,$73,$3F,$EB,$BF
	.dc.b	$37,$7D,$7F,$C8,$FD,$9C,$F6,$FF,$8B,$70,$FE,$77,$18,$AD,$E4,$8C
	.dc.b	$4E,$8F,$F1,$5E,$7E,$BB,$E5,$8F,$F8,$85,$B1,$E8,$F8,$EE,$33,$FE
	.dc.b	$51,$F8,$57,$E0,$7F,$C8,$F0,$FE,$97,$EC,$FB,$27,$3E,$07,$8A,$31
	.dc.b	$5B,$BF,$E9,$B1,$89,$D5,$F9,$0F,$F1,$FD,$F2,$71,$48,$C3,$FD,$5B
	.dc.b	$D9,$E5,$3A,$9F,$2F,$6F,$C2,$EF,$F0,$B6,$75,$97,$FB,$31,$FE,$C7
	.dc.b	$3A,$69,$A9,$C6,$1F,$D9,$3B,$2E,$BF,$A8,$C7,$E1,$F4,$FD,$0F,$9D
	.dc.b	$F7,$B7,$1E,$ED,$C3,$E5,$67,$F8,$07,$FB,$25,$70,$AD,$9F,$DD,$BE
	.dc.b	$3F,$13,$F4,$7F,$02,$FE,$DD,$3F,$DE,$8F,$A0,$E2,$3F,$BC,$C6,$3C
	.dc.b	$5F,$C1,$B4,$E0,$7C,$79,$E1,$9F,$AD,$B1,$FE,$48,$FC,$B1,$F3,$DC
	.dc.b	$70,$FF,$47,$43,$E7,$BF,$2A,$DA,$E9,$F4,$38,$9F,$2B,$63,$27,$DC
	.dc.b	$EB,$57,$1B,$EB,$E0,$F0,$AD,$4F,$ED,$A7,$E1,$5F,$83,$E7,$FC,$3F
	.dc.b	$3F,$6F,$6F,$C4,$E9,$C5,$FC,$96,$ED,$FC,$83,$3F,$0F,$CA,$7A,$BE
	.dc.b	$77,$D6,$DC,$7F,$23,$83,$EF,$FF,$C1,$71,$F9,$4F,$21,$F8,$03,$D3
	.dc.b	$FD,$CD,$6A,$FF,$D7,$15,$A9,$CF,$6F,$F0,$28,$C6,$1D,$AF,$86,$9F
	.dc.b	$CD,$27,$38,$7F,$A5,$67,$35,$FE,$A4,$9C,$D7,$F1,$59,$CD,$27,$35
	.dc.b	$F6,$FE,$7B,$EB,$9F,$63,$C4,$B1,$F3,$A2,$D7,$DE,$9B,$87,$8F,$F1
	.dc.b	$C7,$5F,$A5,$B8,$7C,$37,$FB,$4B,$63,$44,$75,$DF,$70,$FF,$58,$7D
	.dc.b	$0F,$DD,$C3,$F9,$7F,$D5,$7E,$4A,$7F,$D8,$33,$9B,$FE,$33,$39,$94
	.dc.b	$E5,$52,$75,$39,$8C,$76,$FF,$2E,$8F,$C6,$57,$F8,$45,$AF,$FB,$35
	.dc.b	$AF,$8B,$AB,$FA,$13,$FC,$5B,$E5,$5D,$1F,$CC,$3A,$BF,$91,$46,$33
	.dc.b	$FF,$37,$C6,$27,$51,$FC,$5E,$75,$5F,$E7,$0A,$FD,$5F,$CC,$CA,$FD
	.dc.b	$0F,$CC,$DA,$FD,$8C,$DD,$3F,$35,$DA,$AC,$5F,$A7,$E4,$3F,$1D,$F3
	.dc.b	$DF,$E8,$38,$FC,$DE,$C9,$CF,$B5,$F0,$DB,$3A,$6E,$7F,$59,$B9,$7F
	.dc.b	$63,$BE,$55,$FA,$3C,$3B,$5F,$F0,$6D,$7F,$82,$9F,$C1,$BE,$62,$D7
	.dc.b	$EF,$22,$CF,$F0,$C9,$B8,$7F,$6F,$FA,$17,$C0,$7A,$BE,$9F,$FB,$2C
	.dc.b	$63,$1F,$04,$62,$75,$1F,$B8,$9D,$61,$FB,$2C,$97,$FD,$BB,$B9,$E6
	.dc.b	$FC,$6F,$A6,$C3,$71,$FE,$03,$3A,$7D,$D3,$A6,$9A,$9E,$18,$7E,$4A
	.dc.b	$0C,$AE,$4F,$8E,$BF,$BF,$FD,$0B,$E1,$3F,$82,$7D,$AF,$F2,$3D,$BF
	.dc.b	$C2,$13,$9C,$7F,$88,$4E,$5E,$B8,$27,$4F,$F1,$09,$D3,$E2,$8C,$D7
	.dc.b	$F0,$D9,$CC,$BC,$DF,$F1,$9D,$FF,$1D,$F5,$FE,$4D,$FF,$88,$67,$35
	.dc.b	$FC,$72,$73,$5E,$29,$CD,$7F,$19,$9C,$D7,$C2,$9C,$D7,$F1,$99,$CF
	.dc.b	$A2,$E3,$71,$FF,$86,$B8,$3F,$56,$DB,$FD,$43,$71,$FF,$0A,$57,$09
	.dc.b	$FF,$6E,$D7,$F5,$4F,$97,$FE,$AB,$F3,$B6,$7F,$A5,$B6,$FF,$AD,$B7
	.dc.b	$4F,$D5,$BF,$E3,$28,$FF,$68,$0F,$9E,$5F,$D0,$CD,$F2,$9F,$E6,$A9
	.dc.b	$CD,$7F,$A6,$FD,$EF,$C0,$C6,$23,$FC,$35,$3B,$BF,$F4,$C5,$69,$E9
	.dc.b	$19,$9E,$47,$86,$33,$96,$E3,$37,$D7,$0F,$1C,$E6,$7F,$DE,$E7,$13
	.dc.b	$FC,$5B,$7D,$5C,$D7,$18,$C4,$7C,$08,$C4,$F3,$8C,$5F,$89,$E8,$72
	.dc.b	$71,$3E,$F8,$D9,$E0,$9E,$4F,$FC,$01,$38,$DF,$A1,$E6,$70,$3E,$3C
	.dc.b	$70,$C3,$D9,$3B,$AE,$D5,$CD,$F3,$BC,$8C,$5F,$90,$F9,$EF,$83,$0F
	.dc.b	$7B,$A3,$CC,$71,$BE,$F1,$A6,$E1,$E3,$9E,$4D,$BF,$09,$E4,$9D,$CF
	.dc.b	$33,$65,$8D,$DF,$91,$B1,$F7,$23,$D0,$D3,$CB,$1D,$5E,$6D,$38,$BF
	.dc.b	$4C,$DC,$75,$39,$B9,$3E,$69,$E3,$72,$FD,$1C,$3E,$74,$71,$8E,$67
	.dc.b	$8C,$E3,$3E,$18,$E3,$1F,$2C,$CB,$C5,$A7,$E7,$3E,$15,$F1,$1F,$51
	.dc.b	$B1,$98,$C1,$FC,$7B,$E1,$7F,$D8,$FA,$FF,$2C,$AD,$D2,$72,$A3,$31
	.dc.b	$86,$FC,$0F,$F8,$1A,$75,$7D,$63,$18,$FE,$DF,$E7,$FF,$9A,$57,$EA
	.dc.b	$DA,$FE,$83,$0E,$F3,$FB,$D6,$E5,$FC,$D2,$B7,$3C,$F9,$FF,$9B,$BB
	.dc.b	$C3,$BD,$78,$73,$ED,$5A,$8E,$4D,$C7,$FA,$5E,$3D,$9B,$7C,$9F,$E9
	.dc.b	$F8,$F8,$2F,$8C,$7F,$03,$AE,$0F,$C7,$AE,$0A,$B8,$3C,$AF,$C0,$EF
	.dc.b	$0E,$55,$C2,$B7,$3C,$6B,$77,$D8,$EF,$3A,$36,$E6,$B9,$4F,$69,$CC
	.dc.b	$79,$27,$8C,$E4,$EB,$19,$38,$D7,$C8,$3C,$B7,$1F,$E2,$B7,$E1,$AE
	.dc.b	$33,$9C,$7E,$A5,$71,$3E,$E5,$F5,$9C,$4E,$A7,$31,$C7,$1F,$9B,$1D
	.dc.b	$AF,$AB,$A3,$07,$28,$FB,$CE,$6B,$A1,$23,$2F,$E1,$6B,$84,$FD,$33
	.dc.b	$C0,$7E,$01,$CB,$F4,$4C,$9F,$5C,$F0,$92,$3A,$3E,$19,$F4,$C7,$86
	.dc.b	$3A,$1F,$D1,$FD,$4F,$F5,$AF,$81,$FE,$0E,$BF,$F2,$7F,$CC,$DE,$1F
	.dc.b	$F6,$33,$6D,$7C,$1F,$F5,$CD,$E6,$3F,$D7,$F3,$7C,$43,$67,$97,$0B
	.dc.b	$EF,$DF,$27,$C7,$D0,$7C,$F5,$37,$3A,$CF,$FC,$5F,$AF,$FB,$D6,$31
	.dc.b	$1F,$E3,$33,$CB,$5C,$9F,$F6,$93,$5D,$61,$E1,$C3,$D9,$0F,$CB,$7F
	.dc.b	$C5,$EF,$31,$F9,$77,$FB,$05,$FF,$63,$B5,$D4,$70,$C5,$BF,$F8,$E2
	.dc.b	$30,$FF,$8E,$AB,$84,$EA,$BF,$C9,$31,$89,$E1,$1E,$4D,$7F,$C2,$2F
	.dc.b	$FB,$5B,$F2,$FF,$F3,$E3,$76,$D7,$54,$8C,$10,$F5,$57,$27,$D9,$F6
	.dc.b	$3F,$EB,$E7,$FD,$C3,$F8,$13,$D7,$F8,$6D,$8F,$F7,$2C,$ED,$E3,$EB
	.dc.b	$6A,$A4,$6A,$3E,$5C,$EF,$5F,$74,$AF,$A3,$F3,$5B,$1A,$9E,$1A,$7F
	.dc.b	$98,$3F,$0F,$7D,$0F,$D6,$F8,$BF,$E6,$B7,$D1,$B5,$B9,$FB,$75,$C3
	.dc.b	$1F,$F7,$ED,$69,$FF,$3B,$B5,$46,$DF,$C5,$74,$3A,$3E,$46,$3F,$EB
	.dc.b	$2D,$FF,$38,$7E,$77,$1B,$7F,$94,$CB,$FD,$07,$87,$1A,$CC,$F4,$C3
	.dc.b	$B6,$1F,$13,$D9,$7C,$3F,$E4,$BE,$CD,$F0,$D4,$FC,$AC,$EB,$0E,$C7
	.dc.b	$FA,$41,$C3,$F9,$79,$E2,$E8,$38,$7F,$39,$1C,$2F,$C5,$7D,$E7,$85
	.dc.b	$6E,$73,$1E,$0C,$BE,$23,$FB,$D2,$19,$C5,$EF,$62,$3E,$01,$6B,$FD
	.dc.b	$B7,$1F,$BC,$F0,$7F,$80,$7D,$7C,$AF,$6F,$67,$F0,$59,$D4,$7E,$EA
	.dc.b	$74,$FE,$EB,$B9,$3F,$BA,$FA,$37,$C8,$FE,$EA,$BF,$D5,$31,$87,$01
	.dc.b	$C3,$B6,$BF,$C0,$9C,$E5,$F9,$AF,$A0,$FD,$B4,$FF,$B0,$1A,$F3,$AF
	.dc.b	$11,$FC,$32,$32,$6A,$B9,$F1,$FF,$B2,$34,$7D,$B3,$E8,$BF,$97,$46
	.dc.b	$1B,$19,$C7,$F0,$33,$C3,$E8,$F3,$52,$73,$5B,$7F,$92,$56,$DF,$B5
	.dc.b	$F9,$8B,$E5,$7F,$9C,$36,$33,$18,$BC,$B2,$32,$72,$8F,$A1,$CD,$EF
	.dc.b	$F1,$64,$FE,$47,$44,$EA,$3F,$29,$AF,$F3,$8D,$FE,$65,$E6,$7F,$98
	.dc.b	$5F,$07,$71,$D1,$CD,$FA,$3E,$93,$D3,$F3,$12,$7D,$B1,$FD,$23,$D0
	.dc.b	$F6,$DC,$9F,$C6,$F2,$7F,$92,$5F,$87,$1F,$0B,$E9,$24,$60,$F8,$FF
	.dc.b	$2D,$FD,$63,$F3,$17,$AD,$4F,$93,$16,$7E,$6F,$D2,$7C,$C3,$0F,$F0
	.dc.b	$CE,$34,$66,$36,$7E,$82,$FA,$B7,$DE,$F0,$11,$FB,$3F,$4B,$5E,$9F
	.dc.b	$4B,$F6,$8C,$71,$FF,$4D,$EE,$60,$C3,$57,$F8,$2B,$DF,$5A,$17,$C3
	.dc.b	$93,$E0,$12,$3F,$13,$F3,$55,$F6,$77,$7E,$28,$C6,$7E,$1C,$FE,$D4
	.dc.b	$E8,$FC,$8C,$EB,$1F,$06,$2F,$B8,$7C,$CF,$F8,$19,$B7,$FC,$11,$AF
	.dc.b	$CB,$FA,$5D,$4F,$E6,$BD,$99,$3B,$8C,$BF,$A0,$8C,$BF,$E0,$2A,$E7
	.dc.b	$9F,$C7,$5F,$31,$E1,$24,$FE,$A3,$E2,$AB,$F6,$FE,$97,$F5,$26,$2F
	.dc.b	$95,$62,$7F,$5F,$39,$D3,$E7,$43,$F3,$2F,$E4,$EF,$F3,$76,$BF,$E1
	.dc.b	$16,$DF,$F8,$1F,$C9,$73,$FF,$39,$BE,$75,$A7,$FA,$8C,$7F,$16,$9D
	.dc.b	$FC,$4F,$F8,$7A,$B5,$87,$DF,$6C,$EA,$3F,$82,$CE,$FC,$5F,$F1,$2F
	.dc.b	$D0,$B9,$7F,$C6,$DF,$6B,$2F,$FA,$74,$E5,$B3,$97,$FB,$5F,$BA,$AE
	.dc.b	$73,$FE,$4B,$D3,$F4,$7F,$01,$F1,$3E,$13,$FA,$F4,$63,$0F,$CA,$46
	.dc.b	$3B,$95,$BF,$47,$DF,$E2,$DB,$FF,$18,$D6,$9E,$33,$F9,$DC,$3A,$43
	.dc.b	$0E,$98,$71,$DF,$E6,$E3,$CE,$9F,$48,$F9,$BF,$E1,$18,$7F,$70,$8C
	.dc.b	$57,$6A,$FD,$54,$61,$ED,$97,$E0,$3E,$99,$E2,$75,$AD,$61,$CF,$1E
	.dc.b	$B7,$F8,$69,$CB,$C6,$B8,$4E,$B0,$EE,$D7,$CE,$DC,$3B,$C7,$94,$CE
	.dc.b	$37,$0E,$B1,$CA,$35,$3C,$23,$D2,$64,$A7,$22,$CF,$25,$1C,$71,$F6
	.dc.b	$57,$7B,$F3,$5E,$EB,$58,$FB,$E7,$24,$AF,$1E,$3F,$01,$FE,$F3,$7F
	.dc.b	$37,$1F,$F3,$BB,$5E,$51,$FD,$FB,$E1,$BF,$31,$F4,$D7,$80,$D2,$57
	.dc.b	$F8,$1E,$FF,$AB,$F2,$76,$F0,$FF,$67,$3F,$4D,$AF,$DE,$6B,$F8,$67
	.dc.b	$F7,$7B,$7E,$15,$BA,$FF,$00,$FA,$67,$9B,$FE,$81,$AD,$61,$F4,$EF
	.dc.b	$78,$75,$8F,$B1,$7C,$38,$F5,$12,$7E,$8D,$70,$55,$B7,$33,$BB,$EF
	.dc.b	$1F,$6A,$75,$18,$26,$3E,$FA,$F0,$E3,$E7,$8C,$5F,$42,$57,$F9,$4A
	.dc.b	$F9,$CE,$A3,$9C,$FA,$CF,$05,$73,$71,$1F,$C0,$3C,$47,$CD,$67,$3F
	.dc.b	$F1,$EB,$BE,$47,$47,$F8,$FA,$7F,$60,$68,$B5,$C9,$B5,$C6,$78,$61
	.dc.b	$F9,$C9,$E0,$FE,$73,$A7,$FC,$E7,$1F,$E4,$66,$C7,$F9,$2D,$AF,$FB
	.dc.b	$19,$AF,$F3,$C7,$7D,$BF,$EA,$06,$CF,$FA,$12,$BF,$B9,$3F,$CA,$34
	.dc.b	$FF,$61,$57,$FA,$9B,$E9,$BF,$C6,$93,$FE,$D8,$F7,$B2,$FF,$48,$8C
	.dc.b	$E3,$FC,$B2,$72,$FF,$79,$8C,$BF,$ED,$A6,$BC,$7D,$4F,$F5,$9B,$6B
	.dc.b	$FC,$C7,$3D,$2F,$C5,$5D,$35,$F9,$FF,$4D,$DB,$05,$EF,$0F,$EF,$3F
	.dc.b	$4D,$FC,$71,$FF,$77,$FD,$27,$93,$CB,$4E,$6F,$FA,$1C,$E6,$31,$1F
	.dc.b	$D9,$27,$7D,$9E,$68,$FF,$09,$CE,$9E,$F1,$99,$DD,$FF,$A2,$2B,$B6
	.dc.b	$9F,$D9,$F7,$7C,$B7,$7C,$3C,$1D,$7E,$D9,$FE,$79,$87,$18,$C7,$76
	.dc.b	$A3,$FD,$15,$97,$AA,$7C,$19,$7F,$63,$9C,$B7,$3F,$EF,$A7,$A2,$79
	.dc.b	$3F,$D8,$BA,$AE,$9F,$E3,$5A,$FC,$2C,$B8,$3F,$3B,$F0,$BF,$17,$93
	.dc.b	$CC,$DC,$BB,$C7,$88,$8E,$0F,$84,$FB,$7E,$53,$85,$B1,$FF,$22,$7D
	.dc.b	$33,$E8,$3E,$16,$1F,$5E,$30,$FF,$AF,$F8,$0E,$67,$E8,$4E,$AF,$DF
	.dc.b	$BA,$31,$1F,$6D,$AF,$F8,$F9,$E1,$87,$8A,$BF,$1D,$F0,$3F,$EE,$0C
	.dc.b	$3F,$D8,$35,$A7,$FD,$D7,$1F,$E4,$47,$FD,$9D,$E8,$79,$FE,$7F,$8B
	.dc.b	$63,$EB,$D2,$7C,$E4,$8F,$F6,$26,$FF,$EF,$DE,$CE,$19,$FF,$BD,$5B
	.dc.b	$9F,$F8,$BF,$27,$D3,$79,$BE,$87,$CE,$DC,$7C,$FD,$3F,$10,$8F,$89
	.dc.b	$FC,$D6,$4B,$7D,$37,$C0,$F4,$7E,$0B,$6F,$FD,$F1,$F4,$BF,$94,$D7
	.dc.b	$FD,$73,$F4,$CF,$43,$F6,$53,$9D,$BE,$AD,$7F,$BC,$2B,$57,$CB,$B3
	.dc.b	$F5,$52,$BE,$8F,$E5,$9A,$FF,$95,$1A,$FF,$A0,$DE,$D3,$99,$FE,$1F
	.dc.b	$E4,$2D,$F3,$1E,$49,$FB,$7F,$21,$2F,$FD,$19,$97,$CB,$E9,$C6,$7F
	.dc.b	$93,$9F,$BB,$58,$AD,$47,$03,$27,$FB,$83,$E9,$9D,$BE,$26,$C6,$7B
	.dc.b	$F1,$32,$38,$3B,$8C,$61,$E1,$3F,$D4,$1F,$45,$FD,$1D,$C6,$0F,$A6
	.dc.b	$FC,$94,$6B,$A7,$F9,$7C,$E7,$B7,$F9,$25,$F1,$AF,$E0,$2D,$7F,$CF
	.dc.b	$ED,$79,$65,$FC,$19,$A7,$8C,$47,$8E,$BF,$D7,$5F,$6B,$F7,$27,$53
	.dc.b	$9A,$FF,$54,$FC,$F3,$B3,$F8,$E7,$6F,$F3,$18,$C4,$FE,$EE,$31,$32
	.dc.b	$3E,$75,$FF,$11,$EA,$7D,$8E,$63,$E3,$76,$7E,$67,$EA,$3F,$6E,$7E
	.dc.b	$E3,$EA,$2E,$1C,$A7,$4D,$C1,$5B,$8C,$3F,$23,$E6,$BF,$17,$F4,$CF
	.dc.b	$A8,$C5,$79,$3B,$53,$98,$F3,$E7,$F1,$AF,$F3,$BC,$06,$92,$7E,$74
	.dc.b	$ED,$E3,$3B,$7B,$57,$DE,$9D,$1F,$03,$27,$07,$F1,$63,$F8,$BF,$D2
	.dc.b	$5B,$F7,$45,$AE,$13,$78,$FF,$6B,$0F,$A8,$7C,$1F,$C0,$F8,$BF,$ED
	.dc.b	$91,$8C,$7C,$71,$89,$91,$E6,$FA,$2B,$7F,$87,$F1,$3F,$AE,$CF,$AB
	.dc.b	$E2,$FF,$BB,$46,$27,$F3,$DC,$3F,$2A,$73,$E9,$7F,$04,$AD,$57,$F5
	.dc.b	$86,$BF,$ED,$66,$A3,$F8,$94,$72,$BF,$93,$EB,$59,$FC,$D7,$D4,$3E
	.dc.b	$27,$F8,$3F,$D4,$7F,$39,$9C,$C6,$3D,$0F,$F0,$65,$6F,$0F,$BA,$DA
	.dc.b	$F4,$DF,$E7,$DF,$14,$67,$87,$E6,$FA,$4F,$6F,$F8,$AF,$D4,$7C,$BB
	.dc.b	$FD,$64,$E7,$1F,$C6,$4E,$6F,$8D,$FF,$87,$6B,$57,$F2,$29,$F1,$59
	.dc.b	$AF,$D6,$73,$7F,$84,$79,$CD,$26,$5F,$E0,$88,$CB,$FA,$EF,$57,$F1
	.dc.b	$73,$FD,$83,$EA,$39,$CF,$E9,$FE,$1F,$0D,$3F,$55,$1F,$D9,$A3,$39
	.dc.b	$FE,$55,$BA,$FF,$89,$FE,$03,$EB,$7A,$3F,$EE,$1A,$64,$F9,$EF,$E8
	.dc.b	$F3,$99,$FF,$76,$9F,$CE,$B0,$FF,$19,$4E,$9F,$E2,$D7,$F8,$1F,$A7
	.dc.b	$E9,$87,$3C,$F1,$5A,$C1,$5A,$C3,$FA,$AB,$DE,$3F,$DE,$07,$73,$F1
	.dc.b	$97,$CD,$C1,$F8,$F3,$78,$F3,$C3,$DB,$87,$7B,$C5,$78,$23,$07,$AA
	.dc.b	$7C,$B1,$FE,$34,$C6,$BF,$13,$1F,$2D,$70,$8F,$1C,$61,$E0,$75,$33
	.dc.b	$5C,$9F,$93,$7E,$23,$DD,$7E,$68,$D9,$EC,$3E,$7E,$1E,$83,$6F,$53
	.dc.b	$C9,$5D,$5F,$8F,$5D,$70,$F3,$46,$8E,$07,$89,$F8,$0E,$0F,$92,$E2
	.dc.b	$BB,$9E,$17,$9E,$3F,$46,$BC,$67,$7A,$F8,$2D,$7E,$A6,$1F,$2A,$78
	.dc.b	$57,$E0,$6F,$A5,$7C,$86,$BC,$0F,$74,$F0,$7A,$CF,$A2,$FC,$2D,$3C
	.dc.b	$B8,$FD,$39,$E5,$3E,$37,$E0,$1F,$39,$F6,$CF,$BC,$F2,$16,$3B,$9B
	.dc.b	$7A,$E3,$F3,$F1,$FB,$13,$F3,$71,$E6,$FD,$A6,$B9,$AE,$AE,$5F,$D1
	.dc.b	$7C,$47,$F8,$92,$3F,$12,$D7,$E1,$35,$FD,$8B,$9F,$88,$FE,$35,$A7
	.dc.b	$F2,$F8,$C7,$07,$F9,$E6,$91,$89,$D6,$1F,$A7,$BE,$35,$A8,$C5,$F1
	.dc.b	$7F,$82,$7D,$47,$F3,$59,$F7,$B5,$FD,$8D,$72,$9F,$E1,$4D,$C7,$F6
	.dc.b	$95,$B9,$E7,$D1,$FB,$77,$78,$77,$AF,$0E,$3F,$B0,$BE,$DA,$F0,$CB
	.dc.b	$F6,$79,$7A,$23,$8D,$79,$32,$E9,$87,$81,$B7,$FD,$4E,$FC,$8D,$C7
	.dc.b	$E3,$61,$D6,$3F,$59,$8F,$83,$0F,$1C,$7E,$BE,$FC,$11,$D6,$3C,$D8
	.dc.b	$7F,$11,$BF,$1C,$6F,$0E,$B5,$B8,$EB,$5B,$8E,$B5,$C3,$2F,$11,$23
	.dc.b	$C2,$F4,$8F,$31,$23,$89,$2B,$FB,$86,$3E,$96,$DF,$77,$E0,$65,$F2
	.dc.b	$B1,$FA,$11,$A3,$E5,$1F,$9C,$CB,$E3,$A7,$E0,$BB,$3D,$67,$39,$8E
	.dc.b	$09,$8F,$A7,$1F,$59,$F0,$F2,$F7,$4F,$90,$FC,$5E,$1F,$21,$ED,$5E
	.dc.b	$F7,$4F,$37,$35,$F3,$63,$11,$E7,$9E,$4B,$2F,$88,$DC,$BE,$23,$BC
	.dc.b	$7D,$38,$F9,$23,$47,$C3,$BF,$8E,$EC,$F6,$E3,$E1,$3E,$93,$99,$E5
	.dc.b	$87,$E8,$67,$E0,$CF,$F9,$A7,$C5,$FE,$B1,$87,$FB,$27,$E9,$6E,$FF
	.dc.b	$ED,$A6,$D6,$CF,$F3,$35,$63,$C9,$FF,$63,$4D,$F1,$4D,$9E,$3C,$6F
	.dc.b	$97,$7C,$BE,$1B,$8F,$F8,$3D,$AF,$F9,$6E,$F5,$3F,$E6,$4A,$E4,$DA
	.dc.b	$DD,$7F,$71,$9C,$BF,$E6,$AF,$A8,$FE,$E8,$FF,$A5,$DB,$A7,$F9,$79
	.dc.b	$BA,$7F,$79,$FA,$8F,$E2,$F5,$FE,$78,$9C,$BF,$E4,$B8,$FF,$6E,$3F
	.dc.b	$E7,$C6,$BD,$4F,$8E,$7C,$2E,$7B,$9F,$F8,$6E,$3F,$D0,$6D,$7F,$D5
	.dc.b	$5F,$50,$FA,$5D,$D7,$28,$FF,$43,$4F,$FA,$29,$AF,$FA,$C6,$B9,$65
	.dc.b	$FE,$78,$FA,$7F,$F6,$7C,$FF,$A4,$1A,$FF,$AD,$DB,$1F,$E7,$EF,$A8
	.dc.b	$7A,$5E,$89,$DD,$7C,$86,$BF,$E1,$D6,$BF,$E6,$D6,$A7,$AE,$BF,$E2
	.dc.b	$3A,$7B,$BF,$11,$B5,$E5,$F7,$3E,$3F,$D4,$7C,$18,$FF,$53,$B5,$FF
	.dc.b	$67,$37,$2F,$F4,$AF,$93,$FE,$5E,$9D,$EF,$F4,$23,$FA,$24,$EB,$B3
	.dc.b	$AC,$FE,$57,$82,$5F,$FA,$FD,$FC,$C4,$62,$3F,$D5,$51,$69,$A7,$FA
	.dc.b	$2F,$A1,$D1,$E3,$8F,$F2,$07,$BA,$F8,$8F,$C0,$B7,$1F,$F6,$0B,$5F
	.dc.b	$F7,$47,$D3,$BD,$ED,$D7,$FB,$02,$55,$B7,$F3,$2D,$7B,$C6,$E3,$A6
	.dc.b	$5F,$EB,$8B,$7D,$47,$E4,$9B,$7E,$2B,$FD,$B3,$0F,$F7,$01,$FB,$9F
	.dc.b	$4B,$0F,$FB,$2C,$B7,$F0,$9F,$F7,$84,$7E,$E7,$C2,$91,$FB,$97,$F8
	.dc.b	$27,$CD,$3E,$9B,$33,$FE,$F4,$F8,$47,$D4,$7F,$09,$6C,$71,$3F,$DC
	.dc.b	$FF,$48,$FA,$8F,$D5,$B6,$33,$1F,$AE,$FA,$6F,$DB,$C3,$4F,$F6,$CD
	.dc.b	$6E,$3F,$80,$87,$0F,$F0,$46,$BF,$E7,$96,$BF,$EC,$97,$94,$7F,$B9
	.dc.b	$27,$F7,$ED,$F6,$3C,$B7,$FC,$9B,$57,$D4,$3E,$A3,$0D,$CB,$FD,$A9
	.dc.b	$87,$CB,$AD,$D6,$B5,$DB,$2B,$7C,$BF,$ED,$BA,$DC,$60,$3E,$0A,$FA
	.dc.b	$3F,$31,$27,$5D,$0E,$14,$FD,$6C,$75,$5F,$27,$1F,$CF,$56,$A3,$8B
	.dc.b	$5D,$0B,$E9,$5F,$5B,$9A,$4F,$DA,$7E,$B7,$CF,$E2,$3E,$9F,$0F,$E7
	.dc.b	$9F,$AA,$7D,$59,$FD,$14,$66,$37,$1F,$7F,$CA,$91,$D4,$F2,$FD,$47
	.dc.b	$E4,$3E,$72,$46,$0F,$E9,$9E,$9F,$F4,$72,$3F,$84,$F3,$3F,$AA,$47
	.dc.b	$8B,$4F,$8D,$87,$E6,$BE,$A2,$5F,$C2,$8F,$CE,$FC,$D7,$F5,$CF,$A6
	.dc.b	$7C,$2D,$33,$5F,$EE,$98,$CB,$DD,$AF,$F5,$4C,$91,$88,$FE,$B4,$D7
	.dc.b	$E9,$EF,$FA,$C2,$3D,$6B,$C5,$F0,$72,$3C,$0F,$EC,$BE,$A3,$11,$DE
	.dc.b	$D3,$FD,$AD,$B1,$E5,$8C,$1F,$BA,$FA,$89,$1D,$0F,$67,$0F,$C1,$23
	.dc.b	$F8,$6F,$9D,$AF,$78,$EC,$47,$E1,$3F,$C1,$BC,$19,$1D,$0F,$CE,$68
	.dc.b	$FA,$77,$C3,$7F,$82,$9B,$3A,$6B,$F9,$4E,$2F,$CF,$7D,$46,$EB,$FC
	.dc.b	$11,$5A,$AF,$F0,$33,$5F,$F7,$DB,$55,$F8,$67,$F8,$DF,$99,$27,$C1
	.dc.b	$1E,$3F,$9B,$91,$E3,$7F,$90,$FC,$5E,$23,$EF,$60,$9F,$19,$23,$F2
	.dc.b	$51,$97,$F5,$FF,$0B,$FC,$A8,$E5,$E1,$7F,$31,$87,$D4,$7F,$00,$7F
	.dc.b	$81,$18,$F0,$FF,$C2,$31,$98,$F8,$9E,$07,$F4,$6F,$05,$12,$74,$FF
	.dc.b	$89,$9A,$E8,$E8,$7E,$D9,$CB,$C6,$BF,$9F,$F8,$D2,$3C,$73,$E1,$C3
	.dc.b	$D1,$F4,$6E,$5F,$F1,$2B,$6B,$D2,$D7,$CF,$F0,$3A,$9F,$EC,$93,$99
	.dc.b	$FF,$1D,$B5,$F3,$35,$FE,$EF,$D8,$F8,$0F,$F8,$03,$E6,$FC,$0D,$53
	.dc.b	$9A,$F8,$7F,$09,$FD,$DE,$73,$7C,$27,$F8,$F6,$FF,$D3,$AB,$57,$C6
	.dc.b	$51,$8A,$E1,$7C,$A7,$AE,$1F,$CB,$3C,$3E,$90,$FA,$57,$CD,$FF,$94
	.dc.b	$31,$FF,$03,$46,$2B,$8D,$7D,$88,$C3,$E0,$BF,$AD,$F4,$EF,$13,$AD
	.dc.b	$6B,$0E,$78,$F5,$BF,$B9,$39,$78,$D7,$09,$D6,$1D,$DA,$F9,$DB,$87
	.dc.b	$78,$F2,$99,$C6,$E1,$D6,$39,$46,$A7,$84,$7A,$4C,$94,$E4,$59,$E4
	.dc.b	$A3,$8E,$3E,$CA,$EF,$7E,$6B,$DD,$6B,$1F,$7C,$E4,$95,$E3,$C7,$E0
	.dc.b	$3F,$D4,$AF,$E6,$E3,$FE,$88,$6B,$CA,$3F,$AB,$FC,$3F,$E6,$FE,$9E
	.dc.b	$F0,$9A,$4A,$FF,$08,$DF,$F8,$F3,$9B,$B7,$83,$FA,$A8,$C4,$EA,$30
	.dc.b	$7E,$F7,$6F,$C1,$B7,$5F,$DA,$7D,$3B,$CA,$FF,$1A,$56,$B1,$FA,$37
	.dc.b	$BC,$3F,$10,$D5,$7C,$78,$77,$52,$7E,$85,$6E,$B5,$18,$9D,$5F,$78
	.dc.b	$FB,$13,$A8,$C1,$31,$FC,$7D,$F3,$C7,$CF,$18,$C3,$9B,$70,$E7,$3A
	.dc.b	$7D,$33,$A6,$9A,$C7,$03,$C3,$FF,$36,$3F,$A8,$31,$EB,$F3,$F1,$1E
	.dc.b	$A7,$F9,$36,$75,$5C,$2F,$F2,$91,$C0,$CD,$F1,$AE,$13,$AA,$E1,$A7
	.dc.b	$F9,$F7,$A7,$FD,$05,$87,$F1,$CD,$3F,$CD,$1D,$7F,$E6,$9C,$FC,$3E
	.dc.b	$3E,$2B,$F8,$CD,$5C,$BF,$9A,$E1,$C9,$FF,$17,$87,$18,$73,$CF,$FB
	.dc.b	$DD,$F7,$3F,$A1,$FD,$37,$F8,$F9,$FF,$20,$C3,$DC,$ED,$82,$B4,$FF
	.dc.b	$9D,$63,$31,$8B,$FE,$A0,$DB,$FF,$51,$5B,$E9,$F7,$E9,$C9,$F0,$BE
	.dc.b	$BD,$7E,$7C,$E7,$2F,$F2,$4F,$92,$FA,$7F,$E3,$18,$79,$BE,$97,$E1
	.dc.b	$4E,$7B,$79,$66,$9C,$DF,$0C,$39,$B7,$5F,$89,$F2,$BA,$C2,$65,$FE
	.dc.b	$0C,$AD,$3E,$A9,$CD,$F2,$EC,$EF,$DF,$D9,$B5,$C5,$C3,$F3,$3A,$AD
	.dc.b	$71,$8D,$19,$2B,$E6,$E4,$F5,$FC,$17,$A9,$B3,$E5,$70,$FD,$8F,$93
	.dc.b	$B5,$C6,$34,$64,$AF,$CC,$FA,$77,$D0,$FC,$89,$F9,$71,$87,$F0,$66
	.dc.b	$22,$3F,$E8,$6F,$82,$FF,$64,$BF,$E3,$29,$E1,$5D,$6B,$6F,$F8,$BE
	.dc.b	$B6,$E6,$B9,$7A,$1F,$81,$9F,$F1,$CC,$73,$FA,$0C,$4F,$91,$91,$E1
	.dc.b	$EF,$B7,$D5,$BC,$7F,$A9,$C3,$F6,$CF,$FB,$4F,$99,$9B,$1F,$07,$D3
	.dc.b	$F8,$7E,$86,$E7,$ED,$7E,$1F,$3C,$8F,$81,$E9,$5D,$7D,$9F,$4F,$E9
	.dc.b	$8F,$53,$67,$FC,$5F,$F4,$FF,$DE,$1F,$F5,$97,$8A,$FA,$77,$A7,$FA
	.dc.b	$EA,$D6,$3F,$EA,$8A,$D4,$E7,$AA,$61,$CB,$07,$A5,$8C,$3F,$80,$E3
	.dc.b	$CB,$8B,$FC,$84,$DB,$FF,$62,$5F,$07,$CB,$87,$D1,$F8,$D9,$1F,$77
	.dc.b	$C0,$B8,$FD,$99,$B8,$FC,$D9,$B5,$FE,$CF,$FA,$77,$67,$F1,$5D,$91
	.dc.b	$89,$F8,$2B,$B7,$4E,$72,$F8,$0D,$C7,$FD,$C0,$DA,$ED,$5C,$30,$FE
	.dc.b	$2B,$F4,$5F,$CC,$F0,$FF,$29,$07,$0F,$F9,$55,$AF,$F9,$71,$AE,$FA
	.dc.b	$B9,$9F,$C3,$3E,$9B,$EA,$E1,$FE,$68,$AF,$C5,$BC,$E7,$6F,$FA,$4A
	.dc.b	$77,$5F,$CC,$AB,$6F,$FA,$1B,$E7,$5C,$FF,$C6,$1B,$7F,$AB,$FA,$6C
	.dc.b	$4C,$8F,$F4,$EC,$EA,$BF,$55,$59,$7F,$59,$F4,$DF,$A9,$3F,$55,$F2
	.dc.b	$B6,$FF,$6D,$16,$7F,$65,$37,$7F,$F4,$36,$4F,$9B,$FB,$1F,$4A,$FA
	.dc.b	$6F,$9B,$18,$6E,$5F,$0E,$37,$E2,$63,$2F,$81,$17,$15,$7F,$A2,$B9
	.dc.b	$18,$FF,$06,$9F,$0D,$7F,$01,$DA,$5F,$F0,$DF,$A1,$B8,$FF,$09,$9B
	.dc.b	$7F,$BD,$9B,$5F,$E9,$1F,$A6,$7C,$1F,$F5,$86,$C7,$EC,$32,$7D,$36
	.dc.b	$6A,$4F,$FA,$92,$B7,$97,$EC,$6F,$73,$9A,$FE,$C7,$DE,$F3,$3E,$3F
	.dc.b	$AB,$86,$D7,$FA,$BA,$74,$D8,$F9,$BE,$A6,$E3,$04,$C7,$C1,$19,$7F
	.dc.b	$BD,$FD,$2B,$E1,$BF,$B9,$B6,$BF,$85,$FD,$37,$E3,$A9,$39,$78,$D7
	.dc.b	$03,$F6,$93,$B9,$D7,$07,$FB,$1F,$CF,$FF,$12,$61,$FE,$C6,$6E,$1F
	.dc.b	$C4,$FE,$99,$F0,$FF,$E0,$49,$CD,$F1,$F7,$30,$7F,$01,$AD,$E1,$CE
	.dc.b	$3F,$41,$C9,$23,$FA,$57,$D3,$5C,$BF,$BF,$F3,$2B,$18,$FF,$3B,$AB
	.dc.b	$7F,$ED,$4F,$65,$E9,$FF,$AE,$63,$99,$FD,$57,$E1,$7F,$A6,$57,$F4
	.dc.b	$D6,$C6,$4D,$E5,$FE,$28,$3F,$D7,$A3,$E9,$BF,$D7,$1E,$AB,$E7,$3F
	.dc.b	$37,$39,$BF,$E1,$EF,$FA,$F0,$C1,$FE,$E8,$9E,$17,$FE,$EA,$BF,$F6
	.dc.b	$14,$F6,$7A,$DF,$47,$31,$CD,$CC,$F1,$BE,$FF,$01,$6B,$B5,$F1,$9C
	.dc.b	$C7,$88,$61,$E2,$70,$78,$63,$FD,$F1,$19,$C3,$C3,$7D,$6B,$8D,$F5
	.dc.b	$7F,$96,$C6,$6B,$85,$F8,$2F,$CF,$5D,$8F,$54,$76,$BE,$50,$3C,$47
	.dc.b	$AB,$0F,$03,$83,$C2,$E1,$E6,$6C,$ED,$87,$8E,$7A,$C7,$1C,$3C,$AF
	.dc.b	$A0,$F8,$8E,$1F,$11,$D0,$EE,$74,$37,$1D,$1D,$47,$17,$B4,$66,$7C
	.dc.b	$55,$E7,$6C,$F3,$C3,$E2,$9F,$39,$EB,$3B,$69,$BC,$3E,$44,$66,$78
	.dc.b	$C7,$CF,$33,$3B,$3E,$24,$75,$36,$FC,$16,$BF,$02,$BC,$B1,$B9,$E6
	.dc.b	$D9,$E7,$1D,$DA,$73,$76,$65,$F9,$31,$C5,$B1,$9C,$3E,$44,$7E,$9F
	.dc.b	$E1,$BF,$C3,$2F,$D5,$6B,$E5,$0E,$1F,$DA,$BA,$F8,$6F,$F5,$D6,$DF
	.dc.b	$8B,$8C,$70,$A3,$1E,$9F,$DE,$AD,$DF,$48,$C5,$6E,$30,$7E,$56,$77
	.dc.b	$93,$2F,$C3,$FD,$37,$F3,$3A,$FD,$33,$5F,$D8,$57,$09,$EB,$8E,$BA
	.dc.b	$7F,$1F,$87,$3B,$FD,$16,$1C,$63,$10,$C3,$93,$E3,$C7,$CD,$3A,$AE
	.dc.b	$18,$F9,$30,$EA,$F8,$2F,$C7,$7C,$23,$51,$E3,$3F,$7D,$87,$67,$C1
	.dc.b	$3B,$7C,$B5,$D1,$A7,$08,$D3,$C3,$1E,$8F,$F1,$1C,$3C,$AF,$A6,$38
	.dc.b	$1A,$C7,$D5,$7D,$67,$91,$E8,$9D,$63,$E8,$8C,$12,$B8,$BA,$3B,$19
	.dc.b	$8E,$11,$A8,$C1,$A7,$04,$8F,$90,$FD,$2C,$7E,$26,$1E,$AA,$36,$3E
	.dc.b	$41,$AC,$3D,$51,$9C,$7E,$2C,$E6,$BC,$C7,$BA,$FB,$4F,$CA,$AE,$C7
	.dc.b	$1C,$7E,$3D,$7C,$BC,$3A,$9F,$1F,$0F,$0D,$F1,$8C,$CF,$0C,$3A,$DF
	.dc.b	$83,$0E,$93,$F0,$8F,$5D,$70,$71,$87,$9D,$F0,$1C,$F1,$FE,$E1,$C3
	.dc.b	$FE,$47,$8F,$F6,$17,$D1,$7F,$2A,$C7,$FD,$60,$DE,$0D,$F8,$76,$9C
	.dc.b	$9F,$DE,$23,$0F,$F7,$AA,$B8,$1B,$3C,$78,$5F,$4C,$F9,$4C,$37,$1F
	.dc.b	$E9,$99,$FF,$AF,$E3,$11,$FE,$D1,$C3,$75,$23,$79,$FF,$77,$AD,$DF
	.dc.b	$AA,$72,$7F,$57,$FA,$6F,$F4,$6C,$7F,$80,$A1,$18,$AF,$E9,$6D,$8F
	.dc.b	$EB,$DF,$4D,$FC,$69,$E6,$70,$3E,$21,$4C,$C7,$15,$1C,$5F,$E3,$93
	.dc.b	$FD,$63,$1F,$F0,$A4,$74,$6B,$A3,$FB,$73,$8B,$F4,$7D,$33,$E8,$FC
	.dc.b	$33,$9B,$E9,$D3,$83,$CB,$5B,$C3,$D3,$F4,$DF,$EB,$E6,$FA,$BE,$A8
	.dc.b	$C4,$78,$47,$D3,$3D,$1D,$54,$9E,$27,$C4,$9D,$38,$39,$2F,$6B,$E3
	.dc.b	$B6,$35,$AF,$D3,$FA,$6F,$34,$78,$5B,$19,$DB,$E7,$F9,$1F,$E5,$AA
	.dc.b	$FA,$75,$AD,$BF,$C0,$6D,$8F,$F0,$B3,$54,$F1,$7E,$77,$3F,$F9,$4B
	.dc.b	$7F,$CC,$92,$3F,$13,$1F,$E5,$6C,$31,$1C,$39,$FE,$83,$FA,$46,$CE
	.dc.b	$E3,$F0,$FE,$CB,$E1,$BF,$0E,$D8,$CC,$7E,$2B,$E9,$5E,$DE,$A2,$4E
	.dc.b	$8F,$C8,$CE,$9C,$1F,$5D,$7D,$37,$33,$8C,$79,$CF,$DA,$90,$F6,$E2
	.dc.b	$F5,$71,$1B,$32,$7E,$F1,$FD,$BF,$87,$23,$F5,$BD,$BF,$BC,$8F,$E0
	.dc.b	$7F,$2C,$FA,$2D,$57,$E6,$EB,$4F,$0E,$E5,$F0,$AF,$D5,$FD,$0B,$E9
	.dc.b	$29,$2F,$F6,$3F,$4B,$FA,$BE,$8F,$DD,$56,$E7,$F6,$AD,$7F,$34,$D8
	.dc.b	$FE,$01,$3C,$7A,$3F,$C9,$33,$BA,$FF,$25,$87,$0E,$AF,$FD,$0E,$FF
	.dc.b	$98,$63,$2F,$C9,$FA,$57,$D2,$FF,$27,$DD,$18,$9F,$E2,$71,$89,$D4
	.dc.b	$FC,$79,$CB,$B7,$F8,$14,$ED,$CC,$F0,$CF,$E7,$FC,$A4,$9F,$CB,$F0
	.dc.b	$23,$E9,$D6,$67,$F3,$F3,$97,$73,$A7,$E4,$F3,$7D,$22,$3F,$43,$E7
	.dc.b	$2B,$F3,$22,$EF,$F3,$26,$ED,$FE,$74,$DF,$F9,$9B,$7C,$6C,$E1,$FD
	.dc.b	$2B,$27,$D2,$FE,$23,$E6,$64,$FA,$A3,$FA,$2F,$A1,$23,$F1,$0E,$9F
	.dc.b	$5D,$FE,$43,$E9,$64,$7C,$28,$FC,$9F,$CB,$FF,$51,$FA,$37,$D2,$D3
	.dc.b	$35,$F9,$A8,$CB,$63,$4F,$CC,$9F,$CE,$ED,$23,$E7,$BA,$7F,$3D,$4F
	.dc.b	$80,$F0,$12,$3E,$AF,$D2,$D7,$71,$78,$7F,$DA,$BE,$13,$81,$79,$E1
	.dc.b	$FF,$74,$43,$E9,$7F,$5E,$67,$1F,$F6,$3E,$4E,$4F,$80,$48,$FC,$2F
	.dc.b	$CD,$57,$CF,$17,$3F,$0E,$0A,$C6,$3E,$0E,$DF,$DE,$98,$C3,$FD,$AB
	.dc.b	$F4,$8F,$9B,$FE,$FC,$D8,$FE,$0A,$D7,$F2,$3F,$4B,$A9,$91,$A7,$F8
	.dc.b	$4C,$62,$3F,$2F,$EC,$7F,$0C,$31,$9F,$FB,$EB,$C3,$60,$F9,$8F,$09
	.dc.b	$2F,$F4,$7F,$11,$5F,$E2,$53,$B7,$CB,$3F,$C6,$27,$57,$F0,$BE,$13
	.dc.b	$F8,$D9,$8D,$7F,$D9,$FF,$4B,$FC,$9F,$0F,$F0,$1B,$5F,$ED,$0D,$AF
	.dc.b	$97,$E5,$FF,$81,$A3,$15,$CB,$C4,$C1,$FC,$A2,$6D,$FE,$EB,$E2,$7F
	.dc.b	$95,$98,$7F,$BA,$7D,$0F,$D5,$7F,$C2,$6D,$73,$C7,$FC,$CB,$E0,$7F
	.dc.b	$C2,$B3,$9C,$FF,$C4,$0D,$CB,$FD,$D2,$DB,$FD,$A5,$7F,$8F,$E5,$D9
	.dc.b	$FE,$9C,$F8,$2F,$EB,$11,$8F,$7F,$FD,$A5,$A3,$04,$E5,$C5,$7F,$5F
	.dc.b	$ED,$FC,$E7,$87,$C2,$77,$8F,$5C,$39,$61,$C6,$58,$78,$2B,$51,$FC
	.dc.b	$26,$F9,$2F,$A3,$7C,$C7,$F8,$DE,$FF,$B7,$4E,$63,$18,$F9,$E3,$17
	.dc.b	$E6,$8C,$7D,$2B,$C3,$EB,$8F,$4A,$D5,$FB,$E3,$0F,$9A,$FA,$46,$31
	.dc.b	$F0,$E1,$CE,$B8,$9E,$03,$81,$CF,$1F,$2B,$70,$DC,$F1,$8D,$9C,$6B
	.dc.b	$DB,$7D,$F0,$F0,$E1,$CA,$97,$C6,$F7,$39,$BE,$15,$DC,$EB,$5B,$BE
	.dc.b	$78,$7A,$2B,$75,$E1,$9F,$8F,$8F,$C6,$C3,$A1,$2F,$D3,$19,$7F,$B4
	.dc.b	$DF,$C8,$BF,$ED,$ED,$7B,$9F,$DC,$8F,$F7,$57,$0B,$E9,$7F,$B8,$FC
	.dc.b	$27,$F1,$56,$C6,$67,$84,$E5,$E5,$5F,$C8,$F9,$7C,$3E,$0F,$ED,$E7
	.dc.b	$EB,$B5,$FB,$8D,$7F,$18,$FE,$FF,$6F,$9E,$DD,$7F,$15,$F4,$AF,$2F
	.dc.b	$F4,$37,$F5,$B1,$F0,$36,$F8,$3D,$B1,$FC,$1D,$F1,$E2,$F9,$D5,$AA
	.dc.b	$F0,$D6,$0D,$E1,$DB,$2F,$1C,$66,$7B,$3A,$BC,$4E,$A7,$27,$E2,$B0
	.dc.b	$F0,$38,$8D,$13,$2F,$6E,$1D,$DF,$8F,$87,$46,$D7,$39,$D3,$ED,$3F
	.dc.b	$3B,$C6,$FC,$55,$F2,$BA,$8C,$7D,$55,$9C,$E9,$FA,$E9,$CB,$63,$37
	.dc.b	$D1,$57,$1B,$E4,$F3,$FA,$7F,$A7,$D7,$FD,$B2,$91,$8C,$3F,$8C,$63
	.dc.b	$E5,$F9,$7C,$65,$FE,$DE,$BD,$D7,$FA,$09,$B1,$FE,$88,$6E,$1D,$6B
	.dc.b	$8D,$FB,$F0,$B5,$FE,$10,$C3,$97,$9B,$8F,$90,$CB,$23,$27,$78,$C6
	.dc.b	$5D,$E6,$E5,$EB,$9B,$5E,$AB,$7D,$57,$80,$F0,$FC,$C5,$9F,$54,$5C
	.dc.b	$7D,$53,$70,$F8,$D3,$6B,$A4,$DA,$F8,$9F,$53,$EC,$3D,$BF,$4D,$6B
	.dc.b	$E3,$45,$9F,$8D,$37,$0F,$A5,$36,$BD,$13,$6B,$E8,$7A,$5F,$20,$F9
	.dc.b	$3E,$E5,$AF,$A5,$16,$7E,$94,$DC,$3E,$F4,$DA,$F8,$93,$6B,$EE,$7D
	.dc.b	$4F,$D4,$3E,$AF,$C2,$DA,$FB,$D1,$67,$EF,$4D,$C3,$F1,$53,$6B,$E8
	.dc.b	$4D,$AF,$C4,$7C,$CF,$E0,$3E,$9B,$F0,$2D,$AF,$C5,$38,$7F,$19,$D9
	.dc.b	$6B,$F1,$8F,$E2,$0F,$C4,$F8,$76,$7F,$25,$F3,$1F,$8C,$F8,$CF,$C6
	.dc.b	$B6,$3F,$DA,$FC,$5F,$EC,$26,$FC,$2F,$E4,$9F,$CA,$C6,$71,$FF,$70
	.dc.b	$BF,$ED,$08,$E3,$F0,$FA,$F5,$B6,$FE,$93,$ED,$BF,$99,$6F,$7C,$7F
	.dc.b	$F7,$5D,$FF,$80,$78,$BF,$90,$D7,$F3,$7E,$4D,$5C,$AD,$1F,$E0,$79
	.dc.b	$E6,$E3,$34,$71,$8C,$3A,$85,$F6,$C3,$C3,$18,$AE,$91,$89,$D6,$1E
	.dc.b	$1A,$F7,$FD,$47,$C2,$FA,$2F,$86,$DC,$3E,$4B,$67,$71,$F3,$27,$F8
	.dc.b	$2C,$3E,$83,$E4,$3E,$13,$CB,$F5,$6F,$07,$73,$AC,$BF,$89,$CE,$70
	.dc.b	$F8,$17,$CF,$6F,$E9,$2F,$FB,$4A,$FA,$B6,$32,$FF,$86,$63,$2F,$F1
	.dc.b	$AE,$07,$D4,$FD,$EF,$63,$EF,$B7,$0F,$C2,$B6,$BF,$02,$DC,$BF,$D2
	.dc.b	$37,$CA,$BF,$17,$7C,$9A,$7A,$30,$7D,$27,$F2,$8F,$25,$F5,$0F,$43
	.dc.b	$E0,$CE,$63,$17,$C6,$32,$FF,$B9,$3E,$8F,$F9,$DB,$93,$F3,$7F,$2B
	.dc.b	$F9,$8F,$A8,$FC,$CB,$63,$27,$E8,$23,$2D,$3F,$13,$BF,$FA,$15,$BA
	.dc.b	$7E,$6E,$1C,$96,$24,$E8,$E3,$18,$9F,$BD,$CF,$FE,$5F,$AC,$70,$7E
	.dc.b	$82,$79,$DF,$17,$F0,$83,$E9,$7F,$CD,$71,$F9,$FF,$80,$7D,$CB,$DF
	.dc.b	$FC,$6D,$6A,$FF,$19,$5A,$9C,$FB,$5F,$82,$9F,$F3,$85,$6D,$FF,$2E
	.dc.b	$1D,$E3,$4D,$C3,$FC,$58,$BD,$EF,$ED,$79,$7F,$86,$39,$1C,$3F,$D2
	.dc.b	$AF,$F2,$B5,$A7,$FC,$FD,$3A,$E6,$FD,$1C,$E7,$0F,$F4,$13,$67,$9F
	.dc.b	$93,$AF,$4B,$FB,$6E,$0F,$CC,$FF,$DB,$AF,$F0,$13,$1F,$13,$A8,$F8
	.dc.b	$5A,$7F,$8B,$27,$26,$E3,$8E,$5F,$B1,$8C,$B6,$38,$C7,$FA,$31,$DC
	.dc.b	$F3,$6C,$66,$BC,$1A,$F8,$5F,$DF,$FE,$2B,$F0,$F0,$FD,$F6,$6F,$A8
	.dc.b	$FC,$64,$7D,$E9,$D6,$5F,$2E,$72,$FE,$F2,$33,$8F,$FA,$45,$B4,$BE
	.dc.b	$9D,$BE,$0D,$FF,$87,$FE,$6A,$F2,$3F,$5D,$F9,$AB,$E6,$F3,$18,$6E
	.dc.b	$7F,$C3,$2B,$83,$FC,$B7,$17,$E6,$DF,$C4,$73,$FC,$AD,$69,$FE,$87
	.dc.b	$F3,$FF,$D1,$8F,$E8,$FE,$05,$C3,$FA,$6C,$59,$FE,$93,$37,$7F,$E1
	.dc.b	$3F,$53,$FD,$38,$FE,$9F,$F5,$97,$0F,$EA,$B1,$6F,$FA,$8C,$DA,$FC
	.dc.b	$E7,$C0,$7F,$B2,$FE,$A5,$F1,$1F,$EB,$7E,$FF,$81,$96,$63,$FD,$0D
	.dc.b	$57,$1F,$F6,$95,$69,$5F,$27,$F4,$B3,$B5,$3B,$9D,$47,$FB,$16,$75
	.dc.b	$39,$AE,$F3,$D1,$F2,$70,$7F,$A0,$2F,$8C,$FF,$B8,$0D,$1E,$48,$D4
	.dc.b	$F2,$9E,$F5,$D6,$BC,$A7,$03,$CB,$5C,$1C,$CF,$08,$C4,$79,$CE,$13
	.dc.b	$CD,$C4,$EF,$0F,$81,$1E,$58,$C3,$DC,$C5,$FB,$27,$71,$FE,$F2,$3A
	.dc.b	$DF,$81,$D9,$ED,$8D,$E1,$F2,$27,$53,$C6,$FB,$9F,$19,$C5,$F7,$7C
	.dc.b	$46,$0E,$86,$2B,$A5,$F6,$8F,$1B,$6B,$AD,$7C,$D3,$A3,$DE,$B8,$D7
	.dc.b	$33,$89,$63,$8B,$F3,$A7,$89,$CD,$F1,$C7,$37,$8B,$4E,$2F,$D0,$8E
	.dc.b	$2F,$33,$8B,$CD,$FB,$B5,$E9,$8D,$C7,$13,$67,$13,$67,$9D,$DB,$F4
	.dc.b	$8C,$CF,$8E,$33,$1F,$2C,$CB,$F9,$AF,$1D,$F1,$34,$93,$F9,$28,$CA
	.dc.b	$1F,$0C,$F9,$0F,$F1,$9C,$7F,$7F,$AD,$3A,$8C,$D7,$28,$C6,$BC,$E3
	.dc.b	$17,$CE,$31,$7C,$A3,$15,$FA,$3F,$A9,$7C,$87,$6C,$FF,$6F,$18,$9F
	.dc.b	$DE,$57,$09,$FD,$DB,$73,$F2,$D6,$EB,$A5,$F3,$9F,$06,$7E,$CC,$38
	.dc.b	$B6,$38,$CE,$EB,$E2,$5F,$3B,$DE,$5E,$2C,$38,$56,$AF,$82,$BE,$18
	.dc.b	$78,$2F,$81,$FC,$56,$33,$3C,$4C,$9F,$C1,$71,$F6,$39,$3D,$66,$B1
	.dc.b	$F1,$B8,$23,$82,$5F,$B2,$3C,$18,$79,$AB,$AC,$7D,$83,$8C,$ED,$F1
	.dc.b	$63,$EE,$75,$7F,$52,$FD,$F8,$7A,$A7,$37,$EF,$79,$9E,$9A,$F1,$BF
	.dc.b	$8F,$C3,$C3,$5C,$4E,$4F,$90,$CD,$F6,$3D,$18,$78,$E3,$35,$EC,$3E
	.dc.b	$26,$1E,$1C,$3C,$D1,$93,$D7,$8E,$9F,$24,$FB,$E7,$CB,$1E,$A3,$24
	.dc.b	$30,$4C,$7D,$18,$F9,$2F,$A4,$FE,$06,$3D,$0E,$30,$F3,$BF,$10,$ED
	.dc.b	$F0,$9F,$EF,$CF,$AA,$FD,$6D,$FF,$9D,$9B,$B7,$FA,$05,$B4,$A7,$5B
	.dc.b	$EA,$F3,$1C,$E7,$FD,$8F,$58,$A3,$6B,$1F,$57,$2B,$FD,$9A,$EF,$6C
	.dc.b	$47,$FB,$2A,$2E,$5F,$E8,$A9,$F1,$CB,$F3,$77,$C0,$61,$B1,$FE,$A2
	.dc.b	$6B,$FE,$68,$FC,$91,$F1,$7B,$AE,$70,$F9,$EA,$66,$BF,$D5,$7F,$56
	.dc.b	$F4,$7C,$2D,$8C,$C7,$93,$EC,$FF,$D2,$D1,$83,$FD,$66,$FF,$99,$5B
	.dc.b	$E7,$72,$9F,$5F,$B5,$E4,$9F,$F5,$CB,$5F,$F3,$FB,$76,$FF,$5D,$AF
	.dc.b	$9F,$E3,$AC,$AE,$2F,$48,$C4,$72,$8B,$5C,$A7,$E5,$FD,$4B,$C1,$A4
	.dc.b	$BF,$F6,$5F,$DB,$FE,$01,$FF,$66,$CE,$DC,$FC,$7F,$D4,$93,$C6,$3F
	.dc.b	$D2,$FF,$11,$E7,$7F,$D0,$ED,$7F,$D1,$A1,$C3,$FE,$90,$6B,$FD,$09
	.dc.b	$E1,$8F,$FB,$73,$E1,$3E,$C4,$FF,$B8,$1A,$F8,$32,$FC,$3F,$D4,$BC
	.dc.b	$9A,$49,$FF,$73,$C7,$FB,$41,$B1,$97,$FD,$5F,$1C,$5D,$F9,$39,$C3
	.dc.b	$FD,$DB,$39,$9D,$D7,$82,$BF,$45,$F5,$5F,$A3,$E4,$95,$FD,$87,$0F
	.dc.b	$F7,$2C,$EA,$31,$DF,$FE,$F1,$AD,$C2,$B6,$FF,$BF,$BE,$A1,$F2,$7F
	.dc.b	$A4,$6C,$66,$3F,$4F,$F6,$EF,$88,$D4,$63,$E4,$65,$6C,$F0,$9F,$BF
	.dc.b	$F0,$FF,$D1,$4E,$3C,$4F,$E0,$3E,$6F,$EF,$4F,$DF,$7A,$52,$75,$5F
	.dc.b	$E8,$B6,$CE,$5F,$56,$DF,$E0,$1B,$FE,$17,$7C,$E7,$67,$27,$D9,$F2
	.dc.b	$3F,$EC,$E8,$6A,$8C,$36,$3F,$DA,$6D,$7F,$1A,$DD,$98,$7F,$83,$3E
	.dc.b	$9B,$F6,$95,$FE,$EC,$6C,$EA,$3F,$C2,$D5,$C6,$BD,$1F,$8C,$3E,$A7
	.dc.b	$F8,$F5,$43,$E3,$46,$5B,$19,$7E,$09,$F3,$4F,$E5,$31,$B7,$C5,$E7
	.dc.b	$FF,$1E,$C7,$FC,$49,$1F,$50,$38,$7D,$1F,$39,$B9,$91,$B9,$D6,$5E
	.dc.b	$B8,$E3,$5A,$7B,$9B,$AF,$04,$79,$3C,$59,$3E,$C7,$CF,$F5,$3E,$8F
	.dc.b	$0A,$4F,$C0,$7D,$7D,$9E,$CF,$A8,$F6,$FD,$4C,$AF,$8B,$1F,$03,$E0
	.dc.b	$64,$7C,$87,$E1,$FD,$07,$C4,$FA,$99,$3F,$31,$DB,$F2,$B9,$9C,$29
	.dc.b	$F9,$7F,$07,$2B,$81,$F5,$38,$3E,$79,$1F,$AF,$C1,$5E,$97,$F2,$39
	.dc.b	$3E,$FD,$7D,$6A,$D7,$17,$E1,$72,$F0,$CE,$2B,$F0,$B5,$F7,$3D,$59
	.dc.b	$8F,$E1,$1F,$BF,$F4,$F8,$8E,$F1,$71,$E9,$92,$B1,$87,$E4,$2A,$DF
	.dc.b	$F7,$3A,$D3,$F9,$0F,$8A,$FC,$59,$1F,$C9,$FD,$35,$7D,$B1,$71,$EF
	.dc.b	$37,$0F,$CE,$CD,$AE,$D3,$F9,$6F,$3E,$47,$E7,$9F,$CC,$FD,$0E,$23
	.dc.b	$E4,$C5,$C3,$D7,$37,$0F,$D4,$CD,$AF,$54,$FE,$8B,$DD,$91,$FA,$A7
	.dc.b	$F4,$BE,$EE,$23,$EA,$C5,$C3,$E3,$CD,$C3,$F6,$B3,$6B,$E3,$4F,$F8
	.dc.b	$5B,$F1,$87,$C5,$52,$4F,$F1,$D8,$7D,$16,$23,$F1,$D1,$C1,$D4,$F3
	.dc.b	$9E,$9E,$37,$FB,$F7,$3F,$E3,$96,$F9,$7C,$61,$FE,$29,$8C,$3F,$EA
	.dc.b	$1F,$82,$57,$0B,$DF,$32,$FF,$9D,$DA,$F6,$F3,$C3,$73,$D2,$35,$87
	.dc.b	$37,$B3,$FE,$AA,$FA,$47,$CF,$7F,$85,$B0,$E7,$49,$CE,$DC,$23,$FD
	.dc.b	$6D,$18,$7A,$9F,$EB,$9F,$A5,$76,$7B,$A3,$96,$1D,$71,$F0,$36,$39
	.dc.b	$B5,$57,$27,$C0,$72,$C2,$E1,$E1,$C7,$CE,$DC,$3F,$D8,$B7,$C6,$FC
	.dc.b	$A7,$12,$CE,$DF,$0D,$F6,$9E,$17,$D3,$0F,$25,$EE,$F9,$4F,$C6,$AD
	.dc.b	$E1,$EC,$C3,$CB,$87,$03,$D9,$3A,$BE,$E7,$8A,$35,$3E,$1C,$7E,$46
	.dc.b	$3F,$36,$BF,$DB,$E7,$A4,$AE,$5F,$01,$FE,$28,$F8,$37,$53,$E7,$3F
	.dc.b	$C5,$73,$FE,$EC,$8C,$3B,$6B,$DA,$78,$CF,$0A,$FF,$19,$79,$57,$D6
	.dc.b	$FF,$00,$DF,$89,$AF,$E2,$DA,$FE,$19,$FF,$03,$ED,$F5,$9B,$B7,$FB
	.dc.b	$EB,$E9,$7F,$CE,$3C,$FF,$72,$73,$7C,$70,$C1,$F8,$2A,$DC,$7D,$A9
	.dc.b	$DB,$F9,$76,$E3,$E8,$8F,$C0,$CF,$0C,$7C,$57,$D2,$79,$3D,$7C,$6F
	.dc.b	$EF,$D3,$A9,$C9,$1F,$D2,$D7,$3C,$BE,$05,$F3,$9E,$06,$AF,$9C,$EA
	.dc.b	$F9,$36,$75,$97,$18,$FF,$11,$7D,$27,$F1,$48,$FD,$87,$2A,$3F,$89
	.dc.b	$C6,$1D,$5F,$EC,$A7,$5C,$FE,$08,$FF,$5B,$4E,$B0,$FE,$0F,$5A,$BE
	.dc.b	$5D,$1B,$8C,$5F,$3B,$FF,$25,$D7,$F8,$FD,$B1,$FD,$BD,$FE,$39,$19
	.dc.b	$E8,$FF,$13,$75,$7F,$63,$9D,$B6,$33,$3E,$EE,$9B,$7E,$69,$B8,$7B
	.dc.b	$FD,$9B,$8F,$FA,$2B,$0F,$17,$D2,$E2,$64,$79,$5D,$C6,$30,$F3,$45
	.dc.b	$CB,$C9,$36,$BE,$7F,$A3,$D8,$EF,$F4,$B6,$3E,$04,$5C,$3E,$04,$DB
	.dc.b	$F7,$4D,$AF,$B3,$36,$BE,$8F,$A7,$EA,$3D,$7F,$2B,$6B,$E6,$45,$9F
	.dc.b	$99,$37,$0F,$95,$36,$BF,$0B,$36,$BE,$EF,$AB,$F1,$8F,$8F,$EA,$5A
	.dc.b	$FB,$11,$67,$EC,$4D,$C3,$EB,$4D,$AF,$C8,$CD,$AF,$C4,$FD,$2F,$D2
	.dc.b	$3E,$9F,$BB,$6B,$F0,$91,$67,$F0,$93,$70,$FC,$14,$DA,$FC,$EC,$DA
	.dc.b	$FC,$AF,$D2,$FD,$EF,$86,$FB,$ED,$AF,$C8,$38,$7F,$25,$DB,$6B,$F2
	.dc.b	$4F,$E1,$8F,$C3,$F8,$76,$7F,$31,$F2,$5F,$8A,$FA,$5F,$C5,$B6,$3F
	.dc.b	$D2,$BF,$4B,$F9,$0E,$33,$F9,$08,$C3,$9F,$A3,$91,$FE,$9C,$FA,$5B
	.dc.b	$9F,$F9,$EB,$E7,$FF,$98,$6F,$F9,$C6,$D7,$FA,$A9,$FE,$69,$C7,$9F
	.dc.b	$15,$54,$95,$B8,$FF,$5B,$36,$33,$7D,$E7,$8C,$EE,$B3,$C7,$FE,$53
	.dc.b	$D1,$F4,$1E,$DF,$A5,$F7,$36,$FF,$D6,$AD,$C7,$FC,$D5,$5C,$27,$FD
	.dc.b	$73,$5F,$B6,$97,$D2,$B6,$FF,$45,$40,$91,$9A,$FE,$A2,$D8,$FF,$7E
	.dc.b	$B5,$FE,$BA,$E7,$1F,$F4,$A3,$67,$A7,$CB,$BE,$D1,$E9,$78,$EB,$58
	.dc.b	$7F,$B5,$7E,$77,$ED,$7C,$87,$DB,$6E,$3F,$ED,$26,$CF,$B9,$B9,$7C
	.dc.b	$76,$CF,$FB,$4F,$54,$62,$BD,$BF,$15,$E1,$3F,$8C,$F9,$8F,$A1,$FE
	.dc.b	$6B,$3F,$EB,$C3,$E3,$CE,$F0,$FF,$5C,$66,$8E,$35,$C3,$2F,$83,$1F
	.dc.b	$93,$FA,$5F,$CA,$7D,$2D,$C3,$FD,$D6,$DC,$7E,$BB,$6F,$F0,$0F,$E4
	.dc.b	$61,$AD,$31,$C7,$6F,$F2,$93,$A6,$CE,$9A,$48,$FC,$F1,$B2,$CF,$37
	.dc.b	$E0,$7D,$23,$E0,$BF,$0B,$1F,$85,$8C,$57,$F8,$51,$B1,$FE,$1E,$0E
	.dc.b	$1F,$F1,$0B,$AC,$7F,$DB,$0D,$9F,$C8,$7E,$00,$FC,$01,$E2,$7E,$5E
	.dc.b	$B5,$87,$E5,$79,$3F,$B3,$E3,$FE,$E1,$6E,$3F,$99,$E2,$FB,$5E,$37
	.dc.b	$E7,$E7,$37,$C7,$CF,$C1,$F6,$26,$DF,$E8,$70,$3F,$A3,$1F,$4C,$F8
	.dc.b	$4F,$EB,$F1,$86,$C7,$9A,$30,$73,$39,$3C,$5C,$61,$DA,$30,$AB,$C2
	.dc.b	$E7,$7E,$B7,$2F,$89,$DA,$78,$3D,$A7,$51,$F9,$9F,$A5,$74,$FF,$01
	.dc.b	$6D,$78,$B5,$7D,$37,$E9,$A3,$CB,$6A,$DB,$CE,$F9,$1F,$A8,$AE,$11
	.dc.b	$9A,$E3,$F4,$8F,$1B,$F8,$83,$63,$C1,$F4,$BF,$E1,$0A,$FD,$74,$E5
	.dc.b	$B5,$C0,$FD,$AC,$EF,$5F,$8B,$2A,$E1,$1F,$C9,$FE,$5B,$FA,$11,$FD
	.dc.b	$0F,$AE,$E3,$F4,$3D,$4F,$87,$38,$C3,$F9,$3D,$5B,$FE,$61,$F4,$BF
	.dc.b	$D2,$CF,$E9,$9F,$37,$71,$E5,$B2,$71,$87,$F3,$0A,$B7,$FA,$8F,$67
	.dc.b	$F9,$67,$D2,$BE,$87,$F9,$75,$BC,$67,$CB,$FE,$02,$B5,$3F,$E8,$8B
	.dc.b	$E3,$7C,$2F,$7A,$62,$B4,$2F,$8B,$D7,$0E,$97,$FB,$5C,$38,$4E,$6E
	.dc.b	$BF,$ED,$93,$94,$F7,$AE,$31,$DA,$FA,$E1,$E2,$8D,$99,$AE,$17,$D2
	.dc.b	$3F,$DB,$C6,$8F,$01,$4F,$05,$74,$BE,$F1,$F1,$B0,$F3,$E1,$DE,$BD
	.dc.b	$B3,$C3,$0E,$EF,$33,$BC,$73,$C3,$C5,$7C,$B0,$EB,$5C,$4F,$EA,$AE
	.dc.b	$1E,$37,$D2,$FB,$C6,$0D,$DF,$47,$D9,$87,$96,$BE,$74,$62,$FA,$61
	.dc.b	$E1,$BE,$18,$79,$6B,$D1,$39,$69,$C6,$77,$7E,$33,$AD,$71,$3C,$B5
	.dc.b	$C5,$B5,$C4,$E6,$EE,$78,$9C,$E3,$EC,$1E,$C3,$6E,$5F,$A4,$73,$8E
	.dc.b	$A7,$37,$89,$F7,$1C,$C7,$18,$F9,$C6,$5F,$0B,$DE,$36,$D8,$CB,$B3
	.dc.b	$AE,$1F,$35,$B8,$7C,$C5,$F0,$AF,$88,$FA,$2D,$8F,$D2,$B5,$F5,$B7
	.dc.b	$E1,$1F,$1D,$FC,$C6,$79,$D7,$ED,$A7,$3A,$FF,$89,$6F,$85,$ED,$B1
	.dc.b	$9A,$E1,$3C,$F3,$FD,$9F,$D2,$BE,$3B,$F8,$0C,$61,$BA,$F0,$C3,$F6
	.dc.b	$B9,$78,$33,$FD,$46,$1C,$5F,$DC,$B5,$FD,$64,$6F,$0F,$E0,$2D,$BF
	.dc.b	$E1,$F5,$B9,$F0,$57,$6C,$7C,$D5,$E1,$8E,$F8,$F9,$1F,$F2,$25,$F1
	.dc.b	$9D,$5F,$6B,$E3,$5A,$BE,$0A,$F8,$3F,$E4,$DC,$3C,$8D,$AE,$2E,$EB
	.dc.b	$8D,$70,$85,$70,$6D,$70,$9D,$1B,$9D,$1B,$73,$5E,$48,$CC,$F1,$8E
	.dc.b	$E6,$5F,$1C,$71,$8C,$D7,$C7,$C6,$3F,$18,$FA,$71,$C6,$33,$87,$CE
	.dc.b	$9D,$1F,$5A,$FA,$9E,$17,$89,$C3,$1F,$9A,$7D,$8B,$F0,$C7,$1B,$D3
	.dc.b	$F3,$CE,$26,$49,$7E,$12,$57,$48,$FA,$13,$C2,$73,$18,$38,$CF,$85
	.dc.b	$D3,$C5,$FA,$33,$87,$C4,$70,$39,$4F,$47,$EE,$1E,$8E,$5F,$F0,$9F
	.dc.b	$D2,$FF,$62,$AF,$F4,$F3,$78,$7F,$CF,$0D,$A5,$7F,$86,$BE,$97,$0F
	.dc.b	$28,$FE,$E7,$38,$E5,$36,$B1,$F4,$B2,$BF,$BC,$BB,$CF,$11,$FE,$A8
	.dc.b	$8B,$97,$FA,$C6,$7C,$52,$FA,$57,$BD,$86,$C7,$F8,$A1,$AF,$FA,$27
	.dc.b	$E9,$5F,$15,$BA,$93,$B7,$AC,$EA,$3F,$C6,$3F,$3F,$FE,$80,$6F,$9F
	.dc.b	$E2,$8C,$47,$FA,$57,$E9,$7F,$90,$BA,$21,$FE,$99,$8D,$E7,$FE,$28
	.dc.b	$9D,$CF,$FA,$77,$E9,$7C,$0D,$CB,$DF,$18,$AF,$F1,$7A,$F9,$BC,$CC
	.dc.b	$AF,$F2,$3C,$EF,$8F,$FC,$3A,$EE,$33,$3B,$7B,$1C,$2B,$E6,$FD,$2B
	.dc.b	$BE,$92,$7F,$CA,$7F,$4B,$FE,$5A,$7F,$CA,$B3,$B6,$CE,$DF,$F5,$C4
	.dc.b	$ED,$FF,$5D,$7C,$2F,$8D,$B7,$F7,$23,$19,$FF,$93,$3E,$03,$D5,$5F
	.dc.b	$EE,$39,$E0,$FF,$8E,$03,$87,$FC,$76,$D7,$85,$7F,$A8,$DF,$F7,$67
	.dc.b	$D2,$BC,$6F,$C5,$B6,$3F,$CD,$ED,$7F,$DA,$0F,$FB,$35,$CC,$70,$23
	.dc.b	$C3,$C6,$AF,$08,$B8,$FF,$99,$AB,$94,$FF,$A1,$8F,$D5,$F5,$3E,$97
	.dc.b	$F5,$1C,$72,$7F,$CB,$B1,$FA,$B3,$0F,$1E,$CF,$F4,$5C,$E6,$3F,$D0
	.dc.b	$FF,$48,$F9,$0F,$D3,$B6,$33,$1F,$AA,$FA,$57,$C2,$6A,$24,$E8,$FF
	.dc.b	$48,$CE,$9F,$F3,$7F,$A3,$FB,$DF,$26,$46,$08,$79,$74,$FF,$35,$BF
	.dc.b	$E7,$68,$FE,$09,$32,$30,$43,$83,$E8,$F8,$C9,$1E,$C7,$F8,$0F,$D2
	.dc.b	$7E,$A3,$83,$FD,$23,$A2,$7E,$F5,$23,$0F,$E0,$5A,$FF,$B9,$5A,$FF
	.dc.b	$A0,$B6,$7C,$F3,$E9,$7F,$8D,$4C,$3F,$D3,$11,$96,$C6,$5F,$80,$7F
	.dc.b	$24,$8D,$3F,$7C,$63,$FE,$1E,$7F,$D7,$AD,$74,$DC,$7F,$D5,$AB,$E6
	.dc.b	$BE,$0F,$95,$2B,$8B,$F1,$23,$11,$CA,$2D,$72,$9F,$27,$D2,$CA,$F5
	.dc.b	$BE,$7F,$A5,$F4,$77,$C9,$F7,$BE,$BE,$AF,$67,$D2,$7B,$7E,$96,$57
	.dc.b	$C4,$8F,$81,$ED,$C8,$F8,$EF,$C3,$F9,$FF,$89,$F4,$B2,$7E,$5B,$B7
	.dc.b	$F3,$9C,$08,$F9,$5E,$F4,$9D,$9F,$4B,$83,$E7,$11,$FA,$BB,$D7,$9C
	.dc.b	$5E,$0F,$F7,$DF,$1F,$DE,$AC,$78,$EB,$FE,$BB,$97,$D1,$8C,$57,$E1
	.dc.b	$30,$FB,$97,$F7,$7D,$29,$87,$E1,$1F,$C0,$7D,$2E,$23,$D7,$17,$2F
	.dc.b	$0E,$4A,$C6,$5D,$A3,$FC,$95,$3A,$AE,$77,$F9,$0F,$A5,$FC,$61,$1F
	.dc.b	$C9,$FD,$2D,$7E,$4C,$5A,$EF,$37,$0E,$D3,$6B,$DF,$3F,$97,$F8,$59
	.dc.b	$1F,$9E,$7F,$35,$F4,$38,$8F,$AB,$17,$0F,$5C,$DC,$3D,$53,$6B,$E5
	.dc.b	$CF,$E8,$FD,$B9,$1F,$AA,$7F,$4D,$ED,$E2,$3F,$03,$17,$0F,$8F,$37
	.dc.b	$0F,$8D,$36,$BE,$BC,$FF,$84,$BE,$95,$F0,$F4,$93,$F7,$70,$FC,$0C
	.dc.b	$BE,$8B,$33,$FC,$AE,$38,$BB,$C7,$F0,$31,$88,$D4,$F7,$F1,$BF,$DC
	.dc.b	$99,$EB,$E8,$AD,$7F,$89,$AD,$18,$E5,$47,$F3,$39,$7C,$0A,$FB,$E1
	.dc.b	$BA,$E3,$5F,$95,$AD,$3D,$70,$D5,$78,$21,$F4,$8F,$9D,$FF,$09,$DF
	.dc.b	$09,$CC,$63,$1F,$D0,$46,$38,$3B,$46,$3E,$95,$D9,$D7,$1E,$95,$AB
	.dc.b	$FD,$44,$61,$F5,$DF,$48,$C6,$3E,$1C,$39,$D7,$13,$C0,$70,$39,$E3
	.dc.b	$E5,$6E,$1B,$9E,$31,$B3,$8D,$7B,$6F,$BE,$1E,$1C,$39,$52,$F8,$DE
	.dc.b	$E7,$37,$C2,$BB,$9D,$6B,$77,$CF,$0F,$45,$6E,$BC,$33,$F1,$F1,$F8
	.dc.b	$D8,$74,$25,$FA,$63,$2F,$F8,$4E,$FE,$45,$FF,$86,$DA,$F7,$3F,$C3
	.dc.b	$C7,$FA,$C3,$A5,$F3,$5F,$E1,$BF,$83,$FE,$88,$D8,$CC,$F0,$9C,$BC
	.dc.b	$AB,$FC,$4F,$E5,$78,$7D,$43,$49,$3B,$8C,$BF,$3E,$78,$BF,$47,$E9
	.dc.b	$7F,$CD,$7C,$BF,$12,$B7,$C1,$F4,$DB,$BF,$E0,$DB,$7F,$CB,$EF,$8B
	.dc.b	$D6,$32,$FE,$69,$B8,$78,$2F,$F3,$55,$C1,$ED,$8F,$84,$97,$E5,$F1
	.dc.b	$99,$62,$B7,$3D,$EB,$F2,$57,$CF,$2F,$34,$6D,$B3,$CE,$F0,$FE,$83
	.dc.b	$2F,$29,$E8,$BE,$51,$9B,$FD,$37,$2F,$F0,$0F,$A4,$F1,$CF,$E7,$E3
	.dc.b	$F8,$1F,$D2,$63,$3F,$DE,$C6,$F2,$E3,$CE,$AD,$69,$FC,$53,$1B,$8F
	.dc.b	$F0,$F9,$B8,$71,$8E,$D7,$C7,$0E,$7E,$87,$8B,$AB,$F9,$1F,$D2,$59
	.dc.b	$FE,$3F,$36,$FF,$8D,$4D,$C3,$F8,$CC,$DC,$13,$63,$0E,$7E,$93,$5F
	.dc.b	$25,$71,$FE,$59,$8F,$F2,$58,$C3,$FC,$C5,$B9,$7F,$36,$6C,$FF,$32
	.dc.b	$6C,$F1,$8E,$79,$7F,$2C,$F5,$7C,$9E,$9F,$95,$B3,$AD,$3F,$CB,$ED
	.dc.b	$9D,$34,$D4,$F0,$BF,$E8,$8D,$C7,$FA,$4F,$CA,$7C,$1F,$73,$E1,$36
	.dc.b	$FB,$36,$FF,$A6,$36,$B8,$CF,$8E,$B8,$B4,$F3,$7D,$27,$CD,$F8,$5F
	.dc.b	$9C,$D8,$C9,$EB,$8C,$B4,$CE,$1F,$D6,$E3,$2D,$8F,$85,$CF,$F6,$7B
	.dc.b	$FE,$D3,$63,$27,$C7,$8C,$B4,$CC,$FF,$67,$8C,$B6,$3E,$77,$C8,$FE
	.dc.b	$17,$E9,$3F,$0C,$D8,$FA,$46,$0A,$70,$3E,$DF,$E7,$0F,$A5,$FC,$74
	.dc.b	$7E,$46,$30,$F0,$BF,$F5,$14,$3E,$8B,$5F,$4B,$B6,$D7,$4F,$9D,$E5
	.dc.b	$F4,$9C,$DB,$1E,$0F,$A3,$ED,$F1,$BD,$DB,$1F,$98,$FA,$5F,$E6,$3C
	.dc.b	$3F,$A5,$7F,$99,$F2,$DF,$21,$36,$30,$DA,$E5,$3A,$6C,$E9,$A6,$A7
	.dc.b	$86,$DF,$E8,$9F,$A4,$F7,$4F,$EA,$B9,$DF,$1D,$FE,$0E,$8F,$D6,$D2
	.dc.b	$B6,$FF,$B3,$EB,$7E,$7F,$EF,$63,$5C,$1F,$C5,$1F,$F7,$65,$70,$6C
	.dc.b	$65,$FF,$03,$C6,$5F,$F3,$2F,$13,$E8,$FE,$B7,$D1,$FD,$76,$E5,$F7
	.dc.b	$1B,$5F,$61,$B9,$7F,$9D,$2F,$FD,$E9,$5F,$82,$F8,$CF,$C0,$7C,$6F
	.dc.b	$E0,$59,$19,$3F,$0F,$18,$C3,$F0,$73,$6F,$A4,$DA,$FF,$3F,$7C,$D3
	.dc.b	$E6,$1C,$BF,$8D,$F9,$1F,$E4,$13,$98,$C5,$7F,$18,$8C,$36,$B8,$BF
	.dc.b	$98,$F9,$BE,$31,$8C,$FF,$2A,$FE,$64,$FC,$D7,$CC,$DA,$FC,$FC,$59
	.dc.b	$FC,$DC,$DC,$FF,$A3,$F3,$FF,$A5,$78,$3F,$21,$4E,$8B,$18,$6E,$3F
	.dc.b	$97,$9D,$36,$74,$D3,$53,$C2,$57,$E0,$78,$9F,$AE,$F9,$9F,$D9,$7C
	.dc.b	$CF,$EC,$DB,$7F,$B8,$6C,$EE,$3F,$77,$3F,$8F,$F7,$9E,$57,$F4,$DF
	.dc.b	$49,$F4,$8F,$37,$53,$98,$C6,$1F,$9C,$BE,$2D,$9D,$79,$7C,$23,$FA
	.dc.b	$68,$FA,$17,$C0,$FF,$5B,$8C,$7C,$1D,$9C,$C7,$05,$19,$9D,$36,$74
	.dc.b	$DD,$7E,$3C,$7F,$B2,$4E,$A7,$39,$E2,$FE,$D3,$E4,$BF,$8C,$CF,$03
	.dc.b	$FC,$63,$F4,$0E,$BF,$DA,$4E,$B8,$BF,$C0,$FC,$0A,$D6,$1C,$FC,$9D
	.dc.b	$9F,$2E,$3F,$53,$8F,$51,$87,$3A,$FD,$0F,$43,$93,$F9,$67,$CA,$FE
	.dc.b	$61,$B3,$F9,$D7,$0F,$F3,$4F,$96,$B8,$7F,$30,$CF,$F9,$A6,$3F,$CD
	.dc.b	$5B,$5F,$CE,$9B,$3B,$8F,$E7,$B3,$FE,$0B,$F8,$17,$CA,$7F,$3D,$AC
	.dc.b	$36,$BF,$BE,$CE,$9F,$E8,$B3,$A6,$9F,$C8,$1F,$F5,$30,$F9,$D7,$4F
	.dc.b	$FA,$7B,$BF,$E1,$D7,$FB,$FE,$73,$5F,$EA,$23,$FD,$66,$71,$AE,$18
	.dc.b	$FF,$8E,$27,$73,$FE,$A8,$F7,$79,$56,$AE,$CF,$39,$AB,$0F,$0B,$D4
	.dc.b	$D1,$4D,$61,$DE,$FF,$D0,$35,$CA,$B8,$E1,$E2,$BF,$7B,$E4,$8C,$D7
	.dc.b	$09,$E9,$18,$7B,$CE,$EF,$B4,$62,$BD,$4F,$8C,$EA,$E2,$7C,$11,$F1
	.dc.b	$9E,$58,$7B,$A3,$67,$F2,$EA,$EA,$FC,$73,$C6,$E1,$D9,$F0,$23,$93
	.dc.b	$D6,$BC,$07,$AE,$BB,$C7,$94,$D1,$DE,$78,$E1,$F1,$EF,$B4,$E8,$EC
	.dc.b	$71,$8D,$4F,$08,$F2,$B6,$BB,$9F,$02,$77,$3C,$4D,$96,$37,$7E,$73
	.dc.b	$D3,$1F,$34,$F3,$1B,$72,$FC,$58,$E2,$EC,$F3,$BE,$17,$BB,$4D,$BC
	.dc.b	$63,$9B,$F3,$30,$FC,$03,$93,$C3,$3C,$67,$71,$F1,$67,$6F,$1C,$3E
	.dc.b	$4D,$FC,$0B,$F1,$7C,$1B,$E2,$3F,$1C,$DB,$FC,$DB,$5F,$C0,$37,$E0
	.dc.b	$DF,$19,$FD,$3A,$53,$98,$C3,$79,$93,$AB,$AA,$F9,$CE,$AB,$84,$FE
	.dc.b	$AF,$E7,$BF,$9D,$BF,$B2,$6B,$FA,$FA,$E2,$DC,$3C,$98,$FE,$FA,$B6
	.dc.b	$DE,$AF,$0B,$FC,$0A,$F8,$B6,$B8,$ED,$E6,$BE,$2D,$BE,$2F,$AA,$F8
	.dc.b	$B6,$B8,$E1,$F0,$2F,$8B,$6F,$8B,$F1,$2F,$8B,$6B,$8A,$AE,$2D,$AE
	.dc.b	$2F,$C9,$9D,$D7,$28,$D1,$F1,$E7,$71,$93,$55,$D0,$EA,$E6,$BC,$31
	.dc.b	$99,$E3,$1D,$8C,$BE,$28,$E3,$19,$CB,$C1,$70,$FA,$8F,$F2,$AB,$F0
	.dc.b	$DF,$13,$AC,$66,$BA,$D7,$1C,$7E,$A4,$E4,$C6,$3F,$4E,$70,$76,$3F
	.dc.b	$11,$7D,$5C,$D7,$27,$27,$23,$E8,$56,$AB,$A4,$68,$E9,$3C,$4F,$C5
	.dc.b	$3C,$4F,$A5,$3C,$5F,$BC,$67,$0F,$99,$1B,$C7,$E3,$BB,$9F,$AE,$77
	.dc.b	$AE,$D7,$FA,$8F,$75,$C2,$F9,$5F,$D4,$E0,$E2,$D4,$BE,$57,$31,$86
	.dc.b	$E4,$C3,$9E,$18,$6C,$7F,$76,$9C,$5F,$38,$FE,$F3,$58,$F8,$6C,$FC
	.dc.b	$AC,$9F,$EF,$B1,$97,$FB,$FC,$65,$FF,$13,$61,$EA,$9F,$1F,$2B,$F3
	.dc.b	$97,$C3,$E1,$B7,$FE,$77,$6B,$CB,$27,$CB,$6E,$BC,$FF,$2D,$2B,$84
	.dc.b	$7C,$05,$F9,$D3,$E1,$30,$D8,$FF,$57,$B5,$FF,$18,$79,$FF,$ED,$37
	.dc.b	$33,$CE,$3F,$DB,$91,$FE,$B6,$6B,$FE,$37,$9E,$5C,$5F,$EB,$3F,$67
	.dc.b	$F9,$2C,$FF,$AF,$1F,$E5,$91,$9E,$BF,$34,$6A,$7F,$C8,$0D,$9D,$47
	.dc.b	$19,$DF,$5C,$8C,$10,$E2,$FC,$1F,$9D,$7B,$D4,$93,$FE,$CE,$F9,$D7
	.dc.b	$C1,$62,$3F,$CA,$10,$F8,$5F,$94,$D8,$FF,$6A,$DF,$47,$FC,$7B,$F3
	.dc.b	$3F,$88,$E3,$91,$82,$1F,$65,$CB,$F6,$E3,$EB,$7C,$EB,$C7,$BE,$4C
	.dc.b	$AF,$F7,$1C,$E5,$B3,$97,$FC,$CD,$09,$E7,$32,$78,$C6,$F7,$FF,$73
	.dc.b	$57,$39,$FC,$67,$CD,$7E,$80,$8F,$F9,$63,$E6,$AB,$A3,$FC,$B7,$1F
	.dc.b	$A6,$8F,$CD,$FE,$76,$F9,$6A,$4A,$FF,$30,$FC,$AD,$7F,$35,$E4,$FE
	.dc.b	$B8,$C6,$5F,$BB,$8F,$DB,$6D,$FD,$7A,$31,$E0,$7F,$02,$F2,$7F,$6A
	.dc.b	$7E,$DB,$E6,$ED,$FF,$9B,$A2,$BF,$A7,$8B,$5C,$AF,$A7,$8D,$FC,$22
	.dc.b	$F9,$5F,$F9,$DB,$E6,$BF,$80,$FA,$B2,$30,$43,$F6,$4E,$5F,$DA,$47
	.dc.b	$EB,$FE,$75,$EC,$7F,$9C,$E7,$FB,$FC,$61,$C0,$70,$ED,$AF,$F9,$ED
	.dc.b	$CE,$5F,$16,$1F,$3A,$F6,$FF,$7C,$D8,$FF,$07,$CE,$EB,$8E,$7F,$E1
	.dc.b	$0A,$63,$C3,$29,$87,$0A,$D4,$F4,$33,$1F,$EB,$3A,$FF,$69,$BB,$3F
	.dc.b	$C3,$AF,$47,$07,$9A,$38,$E1,$E9,$9E,$33,$CE,$7F,$97,$E3,$FC,$C3
	.dc.b	$C4,$93,$FE,$A8,$8F,$E6,$7D,$FF,$CD,$3C,$09,$3F,$EA,$C8,$FE,$75
	.dc.b	$F9,$93,$E7,$6F,$CE,$CA,$F3,$72,$2B,$51,$85,$F1,$9C,$BA,$24,$E6
	.dc.b	$2B,$FC,$EA,$5F,$35,$E0,$33,$C1,$F9,$68,$CC,$F5,$F9,$AF,$29,$B7
	.dc.b	$C1,$F3,$5E,$8F,$9A,$91,$82,$1D,$DC,$BE,$18,$F7,$98,$BF,$04,$E2
	.dc.b	$7D,$9F,$0B,$F1,$3E,$12,$46,$08,$7A,$DC,$BE,$D8,$F9,$66,$2B,$D9
	.dc.b	$38,$9F,$91,$F3,$3F,$43,$AA,$46,$08,$7C,$77,$2F,$C9,$8F,$AE,$62
	.dc.b	$BE,$44,$E2,$7E,$A7,$C9,$FD,$CF,$42,$46,$08,$7D,$37,$2F,$D5,$8F
	.dc.b	$C1,$98,$AF,$A9,$38,$9F,$C0,$7C,$CF,$E2,$3E,$52,$46,$08,$7D,$F7
	.dc.b	$2F,$E0,$63,$F1,$E6,$2B,$F0,$13,$89,$FC,$67,$9B,$F9,$4F,$99,$91
	.dc.b	$82,$1F,$8B,$72,$FE,$36,$3F,$37,$E7,$FF,$BD,$7B,$3F,$C4,$5C,$6F
	.dc.b	$99,$C5,$FF,$86,$E3,$0F,$FA,$3B,$E0,$7F,$C1,$3D,$3F,$9F,$C3,$6D
	.dc.b	$9D,$61,$D3,$0C,$E1,$CA,$58,$72,$8F,$EC,$B1,$C0,$FF,$4D,$FC,$E3
	.dc.b	$3F,$EF,$B5,$FE,$16,$8C,$76,$EE,$91,$87,$AE,$3F,$EA,$2F,$9D,$FF
	.dc.b	$1E,$C6,$E7,$57,$CF,$0E,$B7,$FE,$A8,$9C,$BC,$7C,$7F,$87,$7D,$AB
	.dc.b	$6F,$99,$AE,$71,$F2,$63,$E0,$8F,$19,$B8,$F1,$4F,$0C,$3B,$CF,$FA
	.dc.b	$EA,$B8,$63,$E5,$BD,$92,$79,$89,$DE,$3E,$9C,$3A,$12,$BC,$78,$FA
	.dc.b	$B1,$EF,$87,$9C,$E6,$E3,$0E,$71,$E7,$3F,$A7,$47,$CD,$7F,$D9,$A7
	.dc.b	$12,$B9,$7F,$D7,$1E,$53,$E6,$7F,$A9,$FC,$22,$31,$3F,$D6,$63,$13
	.dc.b	$A9,$CC,$F4,$E1,$7C,$3F,$F5,$CC,$3F,$C5,$0D,$8C,$CF,$99,$AF,$18
	.dc.b	$FC,$04,$6B,$6F,$F2,$FF,$CE,$BE,$03,$FC,$09,$1F,$8D,$9D,$70,$7E
	.dc.b	$0D,$B8,$7E,$51,$AA,$F9,$79,$5D,$B1,$93,$F3,$2B,$75,$A8,$C4,$EB
	.dc.b	$0F,$03,$F9,$B9,$D4,$62,$76,$47,$D3,$8F,$A3,$1F,$35,$F3,$6D,$7E
	.dc.b	$1A,$F9,$C6,$2F,$A1,$14,$C9,$DB,$EF,$AF,$DA,$72,$3E,$77,$FA,$AE
	.dc.b	$78,$5E,$17,$8E,$FF,$D4,$51,$A7,$F4,$35,$C6,$74,$F3,$F0,$ED,$72
	.dc.b	$BE,$91,$8A,$CF,$87,$FE,$39,$EB,$FF,$1B,$DB,$C3,$C6,$1F,$EB,$B8
	.dc.b	$56,$32,$F1,$D5,$CB,$D5,$36,$B9,$5F,$F7,$A7,$8B,$6B,$5E,$1E,$FC
	.dc.b	$3B,$3B,$6E,$3F,$EB,$2B,$46,$32,$F3,$B6,$BC,$CD,$AF,$8E,$F8,$FC
	.dc.b	$3F,$27,$87,$6B,$C8,$DA,$EC,$DB,$F8,$6D,$9F,$84,$D9,$FA,$6F,$BF
	.dc.b	$C3,$F8,$1E,$1D,$AF,$80,$DA,$F5,$36,$FE,$7B,$67,$E7,$36,$7E,$FB
	.dc.b	$F2,$FC,$3F,$99,$E1,$DA,$F9,$8D,$AF,$8C,$DB,$FB,$6D,$9F,$B4,$D9
	.dc.b	$FC,$5B,$F5,$FA,$3E,$C7,$81,$6B,$EC,$36,$BE,$93,$6F,$F0,$ED,$9F
	.dc.b	$C3,$36,$7F,$2E,$FE,$0F,$C3,$FC,$27,$87,$6B,$F0,$8F,$E1,$DC,$57
	.dc.b	$DF,$6D,$FF,$A2,$BE,$61,$F1,$DB,$9F,$F2,$2C,$61,$FC,$71,$F9,$35
	.dc.b	$F3,$77,$E7,$70,$D9,$E3,$EF,$6F,$E7,$78,$36,$3A,$FC,$AF,$3F,$99
	.dc.b	$E8,$D8,$FF,$35,$7C,$C7,$F3,$3E,$2F,$D2,$3F,$CD,$79,$6F,$8E,$9B
	.dc.b	$18,$6D,$72,$9D,$36,$74,$D3,$53,$C3,$81,$C1,$FE,$8D,$F9,$DF,$7D
	.dc.b	$7F,$A5,$9B,$3F,$E9,$B6,$BF,$D3,$9A,$FF,$16,$0F,$09,$E9,$B7,$EF
	.dc.b	$FE,$29,$F5,$3F,$E9,$9C,$3F,$D0,$F7,$B7,$F8,$FF,$82,$E3,$7C,$F7
	.dc.b	$D3,$F9,$BF,$A8,$DA,$FB,$2D,$D3,$D7,$1C,$AB,$8B,$F7,$AB,$EE,$7C
	.dc.b	$F4,$8F,$C0,$FC,$D5,$BE,$23,$1F,$15,$67,$6F,$C5,$55,$BE,$D3,$6F
	.dc.b	$86,$2A,$C6,$AF,$99,$72,$FE,$37,$E7,$BE,$24,$63,$2F,$C9,$C6,$26
	.dc.b	$47,$E5,$BE,$7A,$E5,$E6,$D7,$E3,$61,$F9,$73,$F3,$1F,$1F,$6F,$F3
	.dc.b	$B1,$6B,$86,$8A,$C6,$DF,$EF,$1E,$7F,$E9,$FC,$1F,$92,$B7,$45,$8C
	.dc.b	$37,$2F,$CC,$CE,$9B,$3A,$69,$A9,$E1,$3F,$3A,$BF,$AE,$C7,$EB,$FE
	.dc.b	$6F,$F6,$1F,$2F,$6F,$F6,$AD,$BF,$4B,$71,$FD,$E3,$FD,$2E,$3F,$DC
	.dc.b	$9E,$3F,$F4,$DF,$39,$F3,$CF,$2F,$5F,$01,$FD,$0B,$2F,$F6,$D5,$6A
	.dc.b	$73,$72,$73,$7F,$EF,$C7,$FD,$BB,$7C,$70,$FC,$3C,$3E,$79,$F0,$9F
	.dc.b	$D7,$63,$1F,$0D,$67,$31,$FE,$E9,$32,$57,$35,$D1,$4F,$1A,$E4,$DA
	.dc.b	$E0,$DA,$E0,$FC,$3F,$9C,$FE,$39,$1F,$9C,$D1,$F3,$DF,$95,$BF,$F7
	.dc.b	$BD,$6B,$74,$7F,$84,$2B,$7D,$9C,$FC,$89,$3F,$23,$07,$76,$32,$FD
	.dc.b	$4F,$77,$E7,$F9,$3F,$95,$7C,$7F,$E7,$9B,$87,$F8,$73,$4E,$F6,$BE
	.dc.b	$13,$93,$FC,$42,$E2,$3F,$33,$F2,$1C,$E7,$F9,$CE,$BF,$CD,$F6,$FE
	.dc.b	$70,$D8,$C9,$FC,$FA,$32,$D3,$B5,$7E,$FF,$E0,$1F,$3D,$FD,$06,$3F
	.dc.b	$C2,$91,$89,$FE,$8F,$18,$9D,$47,$2E,$F5,$6E,$BF,$5E,$7E,$54,$FF
	.dc.b	$5B,$0F,$9D,$75,$7F,$29,$D1,$97,$C4,$9C,$D7,$F9,$F3,$DD,$FF,$21
	.dc.b	$C6,$23,$E4,$4E,$EF,$FD,$03,$1F,$EB,$A3,$B3,$D6,$35,$1C,$CC,$9F
	.dc.b	$39,$CB,$FC,$EE,$BB,$CF,$1C,$AC,$F1,$BF,$16,$1E,$89,$CC,$F1,$3A
	.dc.b	$9E,$67,$2F,$C2,$9E,$38,$78,$23,$17,$DD,$E2,$60,$F3,$57,$27,$DD
	.dc.b	$5C,$A3,$18,$78,$B0,$F2,$57,$23,$77,$D6,$FE,$54,$75,$3C,$C6,$EB
	.dc.b	$96,$1E,$58,$E2,$78,$8D,$1E,$EB,$EF,$5D,$23,$67,$47,$15,$D3,$0F
	.dc.b	$3C,$F1,$BE,$D3,$98,$EB,$5E,$D6,$E1,$E9,$3E,$FC,$EE,$7A,$CF,$3B
	.dc.b	$EB,$1C,$4C,$96,$B8,$CE,$CE,$A7,$B6,$37,$1D,$E3,$F0,$0D,$7E,$F4
	.dc.b	$73,$72,$71,$8E,$B1,$98,$DC,$71,$7E,$86,$1F,$22,$36,$66,$3E,$4C
	.dc.b	$65,$EA,$7C,$B8,$DB,$DC,$FC,$E7,$C1,$BE,$27,$DA,$D8,$CC,$60,$FF
	.dc.b	$0C,$7C,$1B,$E3,$BF,$C4,$72,$AE,$0A,$75,$C5,$FE,$B4,$9C,$C6,$27
	.dc.b	$85,$F4,$BE,$75,$A7,$AE,$5F,$B9,$F9,$EF,$E7,$B7,$FA,$86,$BF,$A8
	.dc.b	$C3,$A4,$FE,$FD,$B9,$7F,$8B,$EB,$73,$CF,$AB,$FC,$64,$EF,$0E,$F5
	.dc.b	$E3,$CF,$B5,$6A,$39,$37,$1F,$F6,$7E,$3D,$9B,$7C,$9F,$F6,$9E,$3E
	.dc.b	$0B,$E3,$1F,$C0,$2B,$83,$F1,$EB,$82,$AE,$0F,$2B,$F0,$3B,$C3,$95
	.dc.b	$70,$AD,$CF,$1A,$DD,$F6,$3B,$CE,$8D,$B9,$AE,$53,$DA,$73,$1E,$49
	.dc.b	$E3,$39,$3A,$C6,$4E,$35,$F2,$0F,$2D,$C7,$F8,$ED,$F8,$6B,$8C,$E7
	.dc.b	$1F,$A9,$5C,$4F,$B9,$7D,$67,$13,$A9,$CC,$71,$C7,$E6,$C7,$6B,$EA
	.dc.b	$E8,$C1,$CA,$3E,$F3,$9A,$E8,$48,$CB,$F8,$5A,$E1,$3F,$4C,$F0,$1F
	.dc.b	$80,$72,$FD,$13,$27,$D7,$3C,$24,$8E,$8F,$86,$7D,$31,$E1,$8E,$87
	.dc.b	$F5,$0E,$1F,$EC,$FF,$3B,$FE,$89,$BF,$F0,$1C,$7F,$6A,$F9,$EC,$70
	.dc.b	$7F,$A1,$38,$D1,$8C,$3F,$BC,$61,$6F,$9C,$7F,$73,$AD,$B6,$B5,$0A
	.dc.b	$BF,$0B,$8F,$9E,$97,$C6,$32,$F3,$8C,$BF,$E7,$1A,$4F,$8F,$8D,$F3
	.dc.b	$CF,$87,$C3,$6F,$FC,$1F,$3A,$9F,$1F,$CF,$69,$F2,$FC,$F4,$AD,$DF
	.dc.b	$17,$FD,$2A,$3E,$79,$F0,$5F,$0E,$30,$D8,$CC,$7F,$87,$8F,$F4,$07
	.dc.b	$AF,$99,$E1,$E7,$FF,$B9,$27,$4D,$7B,$D7,$27,$C3,$F3,$DF,$C9,$5E
	.dc.b	$8F,$F2,$6E,$4F,$F1,$DC,$23,$1C,$9F,$EC,$57,$84,$72,$EB,$F4,$46
	.dc.b	$A0,$71,$EC,$CC,$C8,$EF,$1C,$0C,$3F,$13,$E7,$5E,$EF,$C3,$6C,$7C
	.dc.b	$16,$BF,$2B,$E7,$9F,$05,$BA,$F9,$50,$F8,$4A,$66,$BF,$19,$F3,$DF
	.dc.b	$8A,$E8,$91,$82,$1F,$85,$72,$FE,$1E,$3E,$B7,$CF,$3C,$8B,$E5,$4A
	.dc.b	$FA,$D3,$96,$CE,$5F,$C0,$C2,$79,$CC,$9E,$31,$BE,$1F,$A7,$5C,$E7
	.dc.b	$F3,$1F,$3D,$FA,$22,$3F,$A7,$F9,$EA,$E8,$FD,$54,$7E,$A2,$3F,$3D
	.dc.b	$F3,$CF,$98,$A4,$AF,$D8,$7C,$AD,$7F,$4D,$E4,$7E,$C0,$C6,$5F,$BE
	.dc.b	$DB,$FA,$FB,$7C,$8F,$DA,$9F,$B6,$F9,$EB,$7F,$BB,$8A,$FE,$C6,$2C
	.dc.b	$F1,$8D,$1F,$C0,$63,$F7,$FF,$39,$FC,$03,$D6,$93,$FC,$16,$32,$FF
	.dc.b	$07,$8C,$BF,$BD,$8F,$D6,$7C,$F3,$E4,$29,$27,$F5,$9F,$3D,$F9,$CC
	.dc.b	$7F,$43,$5B,$9F,$F3,$98,$70,$FF,$9D,$9A,$FE,$80,$FD,$0B,$87,$12
	.dc.b	$BF,$DF,$3F,$EC,$E8,$F0,$E1,$89,$91,$82,$1F,$C2,$27,$93,$E8,$8E
	.dc.b	$73,$23,$07,$F2,$9F,$6F,$F9,$3F,$8D,$23,$07,$F2,$DE,$EF,$E5,$7D
	.dc.b	$D2,$30,$7F,$A8,$FE,$6D,$F3,$DF,$CD,$57,$CF,$53,$11,$9F,$9E,$D7
	.dc.b	$34,$BE,$51,$57,$07,$F4,$6C,$1F,$37,$D4,$D4,$65,$F0,$7C,$DC,$AF
	.dc.b	$24,$78,$FE,$6F,$C9,$F3,$72,$7D,$11,$97,$D5,$19,$7C,$3E,$47,$F1
	.dc.b	$6F,$55,$92,$F8,$56,$84,$EA,$F9,$65,$E9,$AF,$89,$F0,$DF,$17,$E1
	.dc.b	$A6,$1F,$26,$32,$FC,$B8,$CB,$D7,$0F,$0C,$EA,$31,$8A,$B7,$87,$C7
	.dc.b	$AF,$A5,$F3,$5F,$4F,$AA,$61,$F5,$E3,$2F,$D9,$8C,$BD,$23,$2F,$68
	.dc.b	$CD,$7D,$39,$FB,$DF,$29,$F7,$FD,$09,$7F,$83,$8C,$BF,$85,$8C,$BE
	.dc.b	$88,$CB,$EA,$8C,$D7,$DF,$9F,$C5,$7C,$D7,$E2,$FE,$56,$5F,$E3,$E3
	.dc.b	$2F,$E4,$63,$2F,$C4,$8C,$BF,$1A,$33,$5F,$8B,$9F,$CB,$79,$7F,$97
	.dc.b	$F9,$A9,$7F,$9B,$8C,$BF,$9D,$8C,$BF,$42,$3F,$3D,$E7,$FF,$BA,$7B
	.dc.b	$3F,$0D,$C0,$F9,$AC,$5F,$E1,$23,$0F,$F6,$0F,$82,$FF,$03,$71,$FE
	.dc.b	$33,$0D,$B6,$75,$87,$4C,$33,$87,$29,$61,$CA,$3F,$AE,$C7,$03,$FB
	.dc.b	$6F,$CE,$B3,$FE,$F5,$5F,$E0,$D8,$C7,$7E,$E9,$18,$7A,$E3,$FC,$5F
	.dc.b	$E7,$9E,$47,$B2,$77,$8F,$4A,$FE,$39,$7C,$71,$ED,$39,$AF,$EF,$F8
	.dc.b	$72,$7C,$15,$BB,$E3,$5B,$9E,$B8,$F9,$6F,$84,$79,$0F,$6C,$EE,$F9
	.dc.b	$46,$2B,$86,$3E,$79,$CC,$F4,$39,$17,$1F,$F1,$86,$1D,$9F,$4D,$F9
	.dc.b	$B0,$EC,$79,$67,$58,$FA,$27,$44,$BE,$B3,$A7,$C8,$7F,$41,$29,$70
	.dc.b	$FA,$0D,$BF,$F1,$4B,$FE,$D8,$F2,$9F,$35,$FD,$3B,$E1,$51,$89,$E8
	.dc.b	$7F,$B6,$5C,$3D,$E3,$73,$FD,$5A,$FF,$C4,$1F,$11,$78,$8D,$25,$F4
	.dc.b	$9D,$BF,$21,$BA,$FF,$95,$FE,$7B,$FC,$CB,$DF,$F7,$30,$E7,$7B,$6D
	.dc.b	$7E,$45,$BB,$FF,$63,$BE,$35,$BC,$79,$11,$FA,$F1,$AB,$E5,$39,$6C
	.dc.b	$EB,$E0,$58,$E7,$0F,$CA,$E3,$E0,$AF,$D2,$B5,$63,$D6,$77,$8F,$19
	.dc.b	$DB,$FA,$79,$D4,$60,$F4,$13,$1F,$89,$8F,$8A,$31,$7E,$FE,$1C,$57
	.dc.b	$C9,$6C,$7C,$3F,$5A,$D7,$F9,$9B,$0F,$D8,$E7,$E1,$BC,$CF,$A1,$B2
	.dc.b	$AB,$1A,$FC,$6F,$FB,$9D,$FF,$9C,$3A,$7C,$9B,$FF,$9E,$7E,$6B,$CD
	.dc.b	$1F,$E7,$CA,$DE,$5F,$D7,$6B,$57,$FE,$8D,$9C,$CF,$E6,$DA,$FF,$87
	.dc.b	$5A,$FF,$9F,$9A,$9F,$F0,$3C,$78,$CF,$F4,$1C,$FF,$A2,$FE,$6B,$C7
	.dc.b	$1C,$FC,$9B,$7F,$A2,$F7,$55,$88,$E9,$3A,$BE,$57,$E7,$8C,$5F,$08
	.dc.b	$C1,$2B,$DD,$8B,$E4,$7A,$65,$E6,$9F,$F5,$9A,$F0,$BB,$7A,$AF,$3B
	.dc.b	$71,$86,$E3,$ED,$BF,$DC,$D6,$E3,$E5,$56,$DA,$49,$FA,$1A,$7F,$B3
	.dc.b	$63,$19,$7D,$5F,$13,$DD,$F7,$2F,$37,$CF,$18,$6D,$78,$6D,$5B,$8F
	.dc.b	$B5,$5B,$69,$30,$FC,$00,$F3,$79,$BF,$ED,$4F,$C6,$7E,$53,$F8,$5F
	.dc.b	$C0,$3D,$8E,$78,$FE,$94,$63,$4F,$F6,$3D,$70,$7F,$23,$19,$C3,$FC
	.dc.b	$C9,$39,$7F,$07,$19,$BF,$F6,$3C,$E5,$B7,$F8,$1F,$C6,$BF,$2F,$87
	.dc.b	$68,$FC,$25,$3E,$77,$F0,$9F,$0B,$F0,$70,$F5,$46,$37,$DB,$73,$FF
	.dc.b	$75,$5F,$17,$E5,$5F,$1C,$7D,$75,$C2,$7F,$DE,$19,$7F,$A9,$DE,$31
	.dc.b	$F9,$3F,$B9,$E5,$97,$EA,$DB,$7F,$13,$E4,$B5,$5E,$86,$DF,$89,$B7
	.dc.b	$FE,$4F,$1F,$2F,$CA,$7E,$4C,$FE,$8E,$31,$3F,$22,$3F,$53,$F9,$8B
	.dc.b	$F3,$C7,$66,$1B,$87,$E1,$F3,$FF,$68,$3F,$99,$AE,$15,$F3,$3C,$3F
	.dc.b	$67,$C2,$DA,$E0,$F8,$48,$FE,$84,$78,$5F,$EE,$37,$F4,$99,$FA,$A7
	.dc.b	$F7,$5F,$44,$F8,$CF,$DF,$F0,$3C,$0F,$A1,$87,$FA,$A6,$B4,$FE,$B6
	.dc.b	$75,$CB,$CA,$F8,$DF,$0C,$43,$0E,$2F,$F9,$E6,$F8,$E1,$E4,$9C,$CF
	.dc.b	$F7,$76,$BE,$1B,$FE,$78,$E5,$AF,$FA,$1F,$F1,$AE,$B7,$FD,$07,$D1
	.dc.b	$43,$E9,$BF,$8D,$C6,$2E,$CE,$71,$FE,$11,$39,$6C,$67,$0F,$F4,$23
	.dc.b	$DA,$3C,$17,$CB,$E3,$39,$FD,$35,$9E,$2D,$BE,$2D,$CB,$CD,$82,$3F
	.dc.b	$97,$FD,$0F,$F8,$42,$EF,$82,$F5,$AC,$61,$AF,$F1,$F9,$FF,$0B,$C6
	.dc.b	$31,$FF,$53,$46,$27,$55,$FD,$D9,$B5,$BC,$7F,$B4,$56,$EF,$F4,$1B
	.dc.b	$3E,$73,$FA,$1F,$23,$E8,$3F,$9B,$61,$F5,$2B,$59,$7F,$31,$AD,$79
	.dc.b	$5F,$8E,$C3,$EB,$E3,$F8,$4B,$DB,$FE,$2A,$6B,$F8,$18,$DF,$0F,$DA
	.dc.b	$FA,$1F,$AF,$F4,$BF,$61,$B9,$7D,$66,$DF,$D6,$6E,$7F,$88,$FB,$0F
	.dc.b	$EE,$D5,$FD,$17,$E6,$5F,$6B,$FF,$00,$63,$F4,$B3,$AC,$3F,$D9,$D5
	.dc.b	$AF,$8B,$FE,$C1,$39,$C3,$FC,$1D,$39,$BE,$3E,$97,$F6,$88,$C6,$1F
	.dc.b	$C2,$6F,$A3,$FE,$53,$6B,$FA,$60,$E1,$FD,$4B,$74,$FC,$DF,$77,$E6
	.dc.b	$3E,$97,$F3,$2D,$C7,$F2,$ED,$BF,$CB,$B7,$2F,$D4,$FC,$5B,$E0,$7E
	.dc.b	$DF,$2F,$F6,$DC,$3F,$82,$CE,$5F,$B5,$19,$C3,$F9,$14,$E5,$FF,$30
	.dc.b	$B5,$FF,$37,$35,$F6,$35,$FE,$07,$F8,$13,$E8,$FF,$C1,$BF,$84,$BF
	.dc.b	$0D,$7D,$17,$F8,$36,$B5,$97,$F7,$BA,$D4,$E6,$F8,$DF,$0A,$FE,$3B
	.dc.b	$7C,$30,$69,$FE,$43,$C7,$8D,$6B,$0F,$C9,$D2,$F8,$3F,$E8,$76,$BF
	.dc.b	$55,$AF,$F1,$CF,$29,$F5,$5F,$A4,$FA,$87,$A1,$FE,$25,$9C,$E3,$FA
	.dc.b	$39,$CB,$FC,$CA,$33,$7F,$E8,$A6,$BF,$E9,$76,$BF,$85,$6A,$E5,$7E
	.dc.b	$7B,$FC,$F6,$3D,$F5,$FE,$E4,$3F,$25,$19,$D7,$F9,$27,$C4,$7F,$01
	.dc.b	$AF,$CE,$CF,$17,$F2,$57,$D3,$E7,$7F,$6E,$DC,$3F,$D3,$ED,$9B,$E6
	.dc.b	$7F,$90,$E3,$F7,$78,$FF,$36,$7F,$AD,$CF,$F6,$97,$35,$DA,$BE,$8C
	.dc.b	$6A,$7F,$D6,$CD,$75,$3F,$EA,$EE,$8F,$E7,$7E,$AF,$F3,$83,$F9,$CF
	.dc.b	$D4,$D9,$FE,$6F,$16,$BF,$9B,$CD,$C7,$F8,$8F,$C1,$7F,$54,$FA,$77
	.dc.b	$D4,$7C,$0E,$0F,$F4,$44,$63,$0F,$F6,$45,$6E,$B5,$95,$9C,$9F,$E5
	.dc.b	$38,$CB,$4C,$FC,$07,$37,$EC,$61,$FE,$95,$8C,$47,$F6,$6B,$E8,$EA
	.dc.b	$EB,$91,$87,$3D,$58,$7A,$31,$FF,$69,$E1,$FD,$C3,$1D,$4F,$79,$CE
	.dc.b	$1E,$53,$B1,$4E,$D8,$70,$C7,$C3,$0C,$3C,$B8,$F8,$27,$31,$8C,$3C
	.dc.b	$55,$F1,$2F,$93,$E8,$C3,$C3,$18,$3E,$0C,$75,$34,$7F,$82,$30,$EC
	.dc.b	$E0,$FE,$4D,$7E,$DA,$F9,$F8,$F2,$8C,$4F,$1C,$3E,$03,$E3,$AF,$AB
	.dc.b	$86,$E3,$A6,$1D,$A3,$07,$96,$7C,$E7,$03,$E4,$E1,$EC,$39,$15,$DD
	.dc.b	$78,$71,$F8,$AF,$6C,$3E,$2B,$EA,$31,$1A,$BF,$04,$79,$CE,$47,$C1
	.dc.b	$3E,$AD,$78,$DD,$1F,$6A,$3D,$31,$98,$ED,$1D,$1B,$19,$79,$9D,$A3
	.dc.b	$8B,$63,$D6,$F5,$6E,$1F,$1D,$B7,$EC,$9E,$66,$CF,$2C,$ED,$B1,$B7
	.dc.b	$F0,$1F,$19,$FE,$BB,$7F,$56,$D7,$E3,$35,$FD,$13,$9F,$89,$7C,$17
	.dc.b	$F3,$E9,$4E,$63,$0D,$D7,$FC,$A6,$D9,$DE,$9F,$DC,$6F,$9D,$7E,$CB
	.dc.b	$EA,$5F,$3B,$FD,$D7,$5F,$DD,$46,$30,$FD,$BD,$F0,$7F,$2C,$D9,$DD
	.dc.b	$71,$D7,$FB,$F5,$F1,$BD,$E7,$E9,$BF,$E0,$D8,$71,$BD,$E1,$C6,$FF
	.dc.b	$CD,$78,$71,$BD,$E1,$C6,$FD,$4D,$AD,$DF,$15,$7C,$6B,$78,$FA,$AF
	.dc.b	$85,$6A,$F8,$5E,$E3,$FC,$F7,$8F,$95,$B5,$C2,$75,$1C,$4E,$97,$1F
	.dc.b	$DD,$46,$8D,$D7,$38,$36,$35,$7E,$98,$D1,$97,$CA,$66,$FD,$31,$F4
	.dc.b	$F2,$FA,$37,$FC,$10,$F2,$1F,$C8,$63,$58,$FC,$28,$F2,$DF,$47,$D1
	.dc.b	$97,$CE,$CB,$F0,$93,$99,$FB,$31,$ED,$C3,$BC,$7F,$2D,$CB,$EF,$47
	.dc.b	$18,$F0,$65,$F1,$EF,$96,$19,$26,$3F,$80,$AE,$E7,$89,$F3,$47,$62
	.dc.b	$38,$3D,$57,$E2,$23,$83,$53,$F8,$43,$A9,$EA,$24,$70,$74,$7A,$31
	.dc.b	$F9,$59,$7D,$18,$E3,$7E,$85,$C5,$FE,$B0,$E8,$FF,$2E,$5F,$FA,$79
	.dc.b	$BC,$1F,$EB,$76,$D7,$0F,$49,$EE,$E3,$0E,$53,$FE,$B2,$AC,$57,$F8
	.dc.b	$BA,$3F,$D6,$B5,$C9,$FC,$EF,$91,$FE,$6B,$67,$02,$31,$1F,$B2,$6B
	.dc.b	$FA,$C6,$BC,$8F,$F7,$3C,$7F,$BA,$9B,$EC,$F7,$8E,$91,$FE,$8B,$FC
	.dc.b	$F9,$E5,$74,$E6,$46,$27,$FC,$52,$D7,$FC,$3E,$D7,$FC,$62,$E7,$6F
	.dc.b	$F6,$2F,$BD,$F0,$0F,$8F,$5D,$2F,$E0,$7E,$3B,$FC,$97,$0F,$F4,$53
	.dc.b	$5F,$F7,$8B,$53,$4F,$E5,$D7,$FE,$EC,$F7,$9F,$57,$F2,$DB,$5F,$EF
	.dc.b	$A9,$FF,$78,$D7,$0E,$2F,$F4,$C7,$E1,$8F,$90,$F9,$2D,$BF,$9A,$D7
	.dc.b	$E8,$B5,$FA,$CE,$6F,$FD,$FB,$D4,$F9,$5F,$B8,$DA,$F1,$61,$E0,$C3
	.dc.b	$F0,$18,$71,$8F,$F5,$55,$6E,$BF,$CC,$60,$FE,$8F,$7C,$2F,$F1,$1F
	.dc.b	$83,$3E,$57,$F0,$2D,$AF,$C2,$B5,$FC,$4B,$5F,$C7,$39,$C9,$8F,$FA
	.dc.b	$EF,$F0,$67,$A5,$F9,$56,$DF,$A8,$F2,$5F,$1A,$F3,$75,$3A,$7F,$DE
	.dc.b	$15,$B8,$C7,$77,$E8,$8F,$D1,$FB,$12,$75,$5F,$E4,$FC,$FC,$71,$FA
	.dc.b	$DA,$D3,$F1,$07,$C9,$BE,$BB,$F4,$2D,$8F,$D2,$B5,$FD,$4B,$5F,$D7
	.dc.b	$39,$AE,$F0,$F8,$DF,$C4,$BD,$DB,$1A,$9E,$19,$FF,$B8,$5B,$5F,$5E
	.dc.b	$B7,$3F,$EE,$56,$C7,$DB,$9D,$E7,$F1,$FC,$AF,$E0,$BE,$54,$AF,$87
	.dc.b	$1F,$C2,$0C,$3F,$84,$9F,$9D,$39,$8F,$A8,$7C,$4E,$EF,$86,$47,$E1
	.dc.b	$B5,$F9,$5F,$50,$ED,$FE,$24,$D8,$E7,$8F,$F0,$97,$E8,$7D,$A7,$FB
	.dc.b	$C5,$EE,$BE,$6B,$F8,$D3,$63,$3E,$CE,$0C,$CC,$8C,$90,$EA,$E5,$EF
	.dc.b	$1F,$CA,$3E,$07,$F9,$4F,$99,$27,$F0,$F1,$97,$F1,$31,$97,$E3,$72
	.dc.b	$27,$F9,$A7,$E1,$3F,$EF,$99,$FC,$79,$FC,$D0,$CF,$AD,$1C,$90,$FA
	.dc.b	$11,$C5,$FC,$97,$D3,$3E,$77,$F9,$8B,$63,$F9,$AB,$5F,$E7,$2D,$7F
	.dc.b	$9F,$39,$E6,$FC,$8F,$D3,$3D,$DF,$E8,$8D,$AF,$BB,$5B,$E8,$FE,$09
	.dc.b	$08,$C3,$FC,$1E,$78,$3F,$A7,$6D,$FE,$7A,$77,$F1,$0F,$CF,$FF,$A8
	.dc.b	$4F,$EC,$23,$17,$FD,$79,$AF,$F0,$1F,$90,$FF,$2C,$FD,$07,$F5,$9F
	.dc.b	$C7,$5F,$53,$BF,$C7,$5F,$13,$4D,$47,$F6,$8D,$BF,$9D,$FC,$75,$F3
	.dc.b	$4D,$38,$67,$FC,$3F,$47,$B2,$FA,$0F,$07,$B7,$2B,$53,$97,$F4,$7A
	.dc.b	$FF,$95,$F8,$3D,$9B,$3D,$FF,$31,$C6,$BF,$97,$E4,$9D,$47,$85,$AE
	.dc.b	$9B,$5F,$C2,$2F,$B6,$BF,$53,$C2,$7D,$4F,$F7,$D6,$E1,$FB,$66,$BE
	.dc.b	$0F,$C7,$7F,$88,$39,$9F,$A1,$E8,$7F,$57,$86,$7F,$D5,$F5,$E9,$52
	.dc.b	$7E,$35,$6E,$B5,$1C,$CF,$07,$C8,$78,$C8,$FF,$08,$F4,$1F,$25,$88
	.dc.b	$F8,$7F,$8E,$BE,$63,$F0,$4D,$8E,$64,$37,$1A,$39,$70,$7F,$2C,$FC
	.dc.b	$77,$F8,$64,$FF,$39,$6C,$67,$CB,$FC,$77,$88,$EC,$FF,$12,$56,$B8
	.dc.b	$FF,$B1,$6C,$FC,$C7,$FD,$4F,$97,$E1,$5A,$F5,$C7,$89,$F9,$98,$CE
	.dc.b	$5F,$9D,$6D,$FE,$AB,$0E,$D7,$FD,$BD,$7C,$C7,$88,$91,$F9,$3C,$FF
	.dc.b	$A8,$BF,$CC,$27,$6F,$E5,$FE,$09,$F5,$BF,$A9,$6C,$6E,$F9,$BF,$90
	.dc.b	$8C,$B6,$73,$97,$C9,$EA,$F8,$64,$9F,$D2,$7C,$47,$E7,$DF,$99,$1F
	.dc.b	$B7,$F1,$7F,$CA,$71,$FA,$96,$BA,$3D,$11,$BC,$24,$60,$F6,$3F,$E4
	.dc.b	$CC,$B8,$C7,$F9,$55,$AF,$C9,$8D,$1E,$1F,$5A,$47,$23,$FA,$AF,$D4
	.dc.b	$F9,$F1,$FE,$98,$DE,$CF,$E3,$6D,$9F,$F1,$03,$FD,$E6,$77,$E0,$7F
	.dc.b	$20,$6C,$FF,$89,$5A,$9E,$BE,$BE,$EF,$F9,$3F,$D4,$3E,$A7,$F1,$35
	.dc.b	$FD,$76,$31,$F1,$7C,$EB,$75,$A7,$CB,$F5,$1C,$F1,$FF,$1A,$5F,$07
	.dc.b	$C1,$3C,$2F,$94,$6A,$F9,$3E,$CB,$E4,$AF,$96,$1E,$3A,$DC,$E5,$B9
	.dc.b	$AF,$8D,$F0,$C7,$C1,$7E,$23,$C1,$8F,$89,$5F,$17,$15,$C5,$D9,$C4
	.dc.b	$F0,$4F,$19,$E1,$19,$9E,$0E,$8C,$61,$E6,$C3,$CA,$FA,$A7,$67,$0A
	.dc.b	$E7,$8F,$BF,$0F,$F4,$DC,$FC,$BC,$3F,$D4,$4D,$7D,$A7,$FA,$97,$E3
	.dc.b	$3F,$CF,$1D,$3F,$EA,$26,$C7,$FA,$99,$AE,$FC,$0F,$E5,$2F,$FA,$AA
	.dc.b	$BF,$C9,$3C,$FE,$0F,$5F,$FC,$0D,$39,$6B,$E2,$6B,$F8,$A7,$79,$7E
	.dc.b	$25,$BB,$7F,$74,$FA,$9F,$EA,$DD,$5F,$3A,$73,$7F,$88,$6B,$F8,$16
	.dc.b	$CF,$08,$F9,$D1,$AC,$7C,$F1,$8C,$7E,$E5,$6F,$1F,$14,$77,$24,$FD
	.dc.b	$B9,$D6,$3E,$99,$CD,$70,$AD,$71,$7F,$9C,$23,$04,$AE,$13,$AC,$B8
	.dc.b	$C6,$27,$53,$FE,$97,$C3,$BB,$EF,$C3,$CF,$C3,$D6,$FE,$33,$67,$87
	.dc.b	$C3,$D5,$7F,$9F,$C7,$89,$23,$D9,$3B,$9D,$47,$E6,$7E,$A3,$FB,$8D
	.dc.b	$7F,$B9,$9B,$1F,$EE,$A6,$BF,$EE,$F6,$BF,$E2,$C7,$7D,$5F,$E8,$46
	.dc.b	$CF,$F5,$28,$FF,$0E,$BF,$E7,$08,$CC,$FF,$AD,$D7,$89,$FE,$9A,$AF
	.dc.b	$F1,$67,$D0,$FF,$84,$6F,$84,$E5,$FE,$BB,$19,$BF,$F4,$F5,$F2,$9F
	.dc.b	$55,$7F,$A5,$A7,$CC,$D7,$C0,$D5,$3C,$5E,$71,$DD,$FF,$05,$DB,$EA
	.dc.b	$71,$1C,$26,$D3,$EA,$78,$72,$BE,$23,$FC,$67,$97,$C1,$9C,$C6,$30
	.dc.b	$F4,$C7,$8E,$F9,$57,$D3,$8C,$11,$E4,$7F,$8F,$ED,$BF,$8A,$7C,$7F
	.dc.b	$07,$E2,$FA,$9F,$F0,$A5,$FD,$18,$C7,$8F,$A8,$FC,$04,$78,$48,$E0
	.dc.b	$E0,$7F,$93,$72,$7D,$4F,$17,$FD,$9D,$F5,$1F,$4F,$83,$FC,$C9,$F5
	.dc.b	$3E,$D7,$FD,$8F,$19,$BF,$F0,$AC,$E5,$FF,$3A,$C7,$DB,$AF,$F3,$B8
	.dc.b	$70,$EB,$C4,$FB,$11,$88,$FF,$80,$0F,$56,$C7,$D6,$FA,$9F,$C7,$C7
	.dc.b	$38,$FB,$32,$F1,$3E,$BF,$CA,$FC,$6A,$F5,$C6,$35,$DB,$E0,$26,$3E
	.dc.b	$69,$F0,$C6,$1E,$31,$FE,$CE,$C7,$FC,$3A,$FF,$B4,$30,$F0,$3F,$E7
	.dc.b	$7F,$90,$B5,$FE,$6C,$6E,$9F,$23,$E0,$FF,$17,$18,$AF,$8F,$1F,$08
	.dc.b	$8F,$8E,$BE,$44,$7E,$77,$CD,$E1,$1F,$1A,$3F,$33,$18,$8D,$CF,$FA
	.dc.b	$5F,$EA,$5E,$97,$E8,$7D,$D4,$62,$7D,$1F,$13,$98,$C5,$FE,$0B,$C1
	.dc.b	$FA,$A4,$AF,$26,$BF,$EC,$D7,$CD,$93,$0F,$19,$FA,$F8,$FC,$DF,$C5
	.dc.b	$E1,$FF,$7B,$FD,$33,$EA,$7F,$69,$C0,$F0,$7E,$C2,$AF,$F5,$FD,$6D
	.dc.b	$68,$9F,$E1,$D5,$C2,$77,$7F,$C0,$EF,$85,$FD,$D6,$BF,$94,$6B,$C2
	.dc.b	$7C,$B3,$FC,$1F,$E5,$5F,$75,$FE,$D5,$9C,$9E,$2C,$7C,$D6,$F8,$F7
	.dc.b	$C5,$F8,$27,$31,$8F,$5F,$F8,$93,$AC,$AD,$7F,$9C,$30,$E8,$7B,$A7
	.dc.b	$4E,$30,$ED,$3B,$7F,$DF,$F5,$CA,$F9,$FD,$57,$4F,$72,$E1,$C7,$1F
	.dc.b	$2C,$63,$0F,$95,$3D,$F1,$FF,$51,$BB,$57,$CE,$3C,$5F,$55,$FE,$B4
	.dc.b	$7F,$96,$3F,$EB,$6F,$84,$BE,$37,$D9,$9D,$BF,$ED,$E6,$9A,$8D,$B4
	.dc.b	$C9,$C1,$CB,$4F,$C4,$FC,$97,$F3,$BE,$77,$D3,$7F,$3F,$E0,$FA,$7A
	.dc.b	$38,$D7,$BA,$D6,$1F,$8C,$8F,$E8,$33,$A8,$F9,$D8,$74,$F0,$33,$5F
	.dc.b	$5B,$E2,$7E,$BF,$D5,$5B,$F8,$CD,$CB,$E5,$9F,$59,$C6,$7F,$82,$E0
	.dc.b	$47,$F5,$0F,$A9,$7A,$3F,$8F,$FA,$A7,$CF,$E6,$30,$DC,$BF,$82,$57
	.dc.b	$06,$CF,$F6,$18,$C6,$5F,$C6,$DB,$97,$C4,$6D,$7E,$AB,$F0,$47,$E0
	.dc.b	$8F,$9D,$ED,$87,$AB,$A7,$F3,$7E,$2F,$E5,$BE,$B3,$F2,$ED,$C7,$F0
	.dc.b	$CD,$AE,$1C,$1F,$C2,$B4,$F8,$17,$FD,$BB,$E5,$1F,$31,$F4,$FD,$3C
	.dc.b	$37,$2F,$CF,$62,$AD,$C6,$1B,$1F,$3B,$9F,$F0,$3E,$97,$2A,$E3,$F5
	.dc.b	$2F,$99,$FF,$00,$7D,$53,$EA,$DF,$59,$FE,$00,$AD,$5F,$EB,$6B,$53
	.dc.b	$9B,$FF,$06,$56,$B2,$FE,$8C,$D7,$FA,$EB,$5F,$E2,$A1,$E1,$BF,$E5
	.dc.b	$1B,$87,$F1,$AC,$FF,$33,$E0,$77,$8F,$83,$E8,$3E,$B3,$F3,$DF,$58
	.dc.b	$F5,$BF,$C2,$F3,$9A,$FC,$BC,$E5,$E3,$5A,$38,$95,$FE,$15,$E6,$3E
	.dc.b	$AD,$E7,$FF,$14,$FA,$CF,$DC,$6F,$F9,$AA,$D4,$E6,$FF,$92,$5F,$07
	.dc.b	$9B,$6B,$F0,$3F,$03,$AA,$FF,$16,$DF,$F4,$5B,$FE,$B8,$DA,$FE,$E0
	.dc.b	$D7,$FC,$0E,$D7,$E5,$BB,$E0,$FD,$EB,$67,$F8,$C7,$B5,$DB,$D3,$FE
	.dc.b	$56,$C8,$FE,$35,$F5,$97,$0F,$CA,$CD,$C3,$F6,$93,$6B,$FC,$05,$F0
	.dc.b	$BF,$D3,$BE,$5D,$F2,$1E,$2E,$2F,$E7,$F1,$8C,$BF,$C3,$B5,$BA,$D5
	.dc.b	$D9,$C8,$FA,$CE,$0B,$56,$3F,$D2,$A7,$33,$FE,$FD,$38,$5F,$46,$9D
	.dc.b	$31,$E7,$7C,$B0,$F0,$4F,$E4,$B1,$B3,$98,$57,$65,$3F,$DB,$30,$E7
	.dc.b	$8F,$8F,$1F,$87,$5E,$03,$D1,$19,$7D,$B3,$C3,$1E,$55,$C2,$31,$3C
	.dc.b	$4F,$63,$83,$C5,$7C,$23,$FB,$EE,$3E,$79,$CC,$7A,$2B,$E1,$63,$83
	.dc.b	$A9,$A9,$E9,$1E,$13,$E4,$39,$69,$99,$EF,$8F,$AD,$F9,$07,$73,$38
	.dc.b	$FC,$C7,$EA,$5F,$82,$B9,$47,$C4,$9F,$53,$CA,$FC,$6F,$C6,$9F,$0B
	.dc.b	$4E,$A7,$C5,$3E,$99,$E1,$3A,$14,$F0,$BF,$08,$F0,$E1,$F1,$DB,$87
	.dc.b	$AE,$7E,$7C,$78,$1F,$23,$5C,$9E,$2A,$F8,$CD,$8F,$C0,$E1,$EF,$8F
	.dc.b	$C8,$47,$85,$F5,$1E,$69,$E2,$D8,$E2,$6D,$F9,$B3,$CD,$B1,$97,$DA
	.dc.b	$7C,$48,$FC,$0F,$C7,$7E,$62,$3F,$56,$D7,$EB,$B5,$FD,$43,$9F,$8E
	.dc.b	$FF,$5C,$6B,$FE,$48,$8C,$7C,$3F,$F9,$26,$91,$89,$D5,$FF,$3A,$BE
	.dc.b	$35,$A8,$C5,$F1,$7F,$77,$F5,$8F,$9D,$FE,$83,$18,$6E,$5C,$31,$FD
	.dc.b	$AE,$3D,$2B,$58,$FE,$E3,$0E,$4E,$B1,$ED,$87,$18,$FD,$94,$E9,$B8
	.dc.b	$FF,$50,$CF,$F8,$6E,$3E,$86,$E1,$CA,$BF,$CD,$37,$C7,$1F,$0C,$7E
	.dc.b	$3E,$FC,$98,$F7,$C3,$24,$70,$49,$E7,$8F,$86,$3F,$B0,$E1,$E5,$AF
	.dc.b	$0E,$5F,$A4,$C3,$C9,$5A,$BE,$18,$72,$35,$97,$D5,$C3,$8C,$7E,$8F
	.dc.b	$1F,$45,$F1,$8F,$8D,$8C,$7E,$19,$FB,$29,$F1,$4F,$9E,$18,$F8,$9F
	.dc.b	$01,$C0,$FE,$3C,$E5,$DE,$1F,$25,$57,$89,$E3,$97,$98,$F3,$E3,$F1
	.dc.b	$E3,$AB,$8C,$3C,$55,$E4,$9C,$C7,$97,$0F,$A3,$1D,$23,$DC,$FD,$5C
	.dc.b	$3C,$53,$E5,$AE,$53,$E4,$CB,$E5,$BE,$8C,$BE,$59,$F0,$E3,$33,$C2
	.dc.b	$BD,$C7,$38,$C9,$C9,$CD,$FC,$19,$F8,$17,$DC,$FA,$19,$7B,$5F,$59
	.dc.b	$0F,$11,$F3,$27,$95,$FE,$92,$3D,$C7,$FA,$F2,$FE,$0F,$91,$FE,$AA
	.dc.b	$EA,$FF,$15,$E5,$FE,$69,$6F,$07,$FA,$5D,$B7,$FE,$A4,$F8,$4B,$7C
	.dc.b	$63,$FD,$35,$38,$F2,$3F,$C5,$51,$FE,$9D,$AE,$2F,$DC,$F3,$EE,$DF
	.dc.b	$E7,$28,$FF,$29,$B5,$FF,$2E,$DE,$A7,$FD,$ED,$1C,$4F,$F7,$C4,$ED
	.dc.b	$FF,$21,$99,$3F,$C3,$BF,$59,$FE,$80,$7F,$C6,$6D,$7F,$C6,$0D,$CF
	.dc.b	$FD,$11,$DB,$EE,$E5,$4E,$75,$FF,$5C,$7C,$07,$B4,$D4,$67,$0F,$F7
	.dc.b	$3F,$CF,$FF,$27,$C7,$FD,$4A,$DB,$FF,$2B,$B6,$7F,$DD,$BF,$0A,$F8
	.dc.b	$FF,$8B,$39,$94,$BE,$B3,$E4,$B6,$3F,$CF,$33,$BF,$AC,$7C,$A7,$CE
	.dc.b	$6C,$FF,$9F,$9A,$FF,$AA,$07,$3F,$FA,$19,$D4,$FC,$E9,$DC,$7C,$C6
	.dc.b	$BF,$E7,$06,$A0,$F6,$CF,$FD,$8B,$5F,$E9,$F9,$FF,$47,$B5,$F1,$4E
	.dc.b	$F3,$FF,$3D,$FD,$53,$EB,$3F,$0E,$DA,$FF,$4B,$B5,$F5,$7D,$63,$C1
	.dc.b	$DD,$7E,$3A,$1E,$AD,$33,$5F,$EE,$FF,$7B,$F2,$DD,$F2,$30,$43,$FD
	.dc.b	$5A,$E5,$FF,$57,$C7,$E6,$BE,$5D,$F3,$54,$93,$FE,$AC,$F5,$3E,$E3
	.dc.b	$FE,$AE,$9D,$BF,$96,$3E,$43,$87,$0D,$7F,$36,$D7,$FD,$3E,$EF,$83
	.dc.b	$FD,$E3,$3F,$83,$9F,$B8,$D8,$FF,$7A,$CE,$F6,$FA,$DD,$EE,$FF,$F4
	.dc.b	$9E,$1F,$66,$FF,$DD,$1E,$47,$EE,$4F,$DD,$79,$32,$75,$87,$2B,$56
	.dc.b	$E7,$23,$0E,$D7,$F1,$FE,$4E,$4F,$C3,$7F,$81,$FD,$63,$E2,$F3,$3F
	.dc.b	$95,$87,$76,$8C,$E8,$D3,$FD,$83,$F5,$9F,$99,$BF,$CE,$2F,$AC,$FE
	.dc.b	$1E,$D8,$FF,$6F,$BB,$AE,$51,$F9,$98,$C5,$74,$AE,$27,$2F,$6E,$57
	.dc.b	$28,$CB,$D2,$3F,$90,$7A,$BF,$C8,$7C,$E9,$3F,$13,$7F,$A1,$3A,$8C
	.dc.b	$70,$FF,$A5,$1B,$59,$59,$F9,$AD,$F5,$9F,$EE,$F9,$E2,$7F,$2F,$3F
	.dc.b	$99,$B9,$78,$E7,$F8,$08,$C5,$73,$C3,$F9,$3B,$6B,$8F,$B1,$FC,$DB
	.dc.b	$EA,$DF,$59,$49,$3F,$78,$7D,$17,$F8,$F1,$C4,$F0,$7F,$3C,$1C,$3F
	.dc.b	$A1,$6B,$FC,$E1,$AF,$FB,$F1,$DD,$FD,$68,$EF,$EF,$FF,$4A,$A3,$E4
	.dc.b	$8F,$78,$F8,$D7,$D6,$7B,$B5,$FE,$2D,$18,$9F,$EB,$6D,$7F,$33,$F2
	.dc.b	$FF,$C9,$3E,$9B,$FA,$BF,$D5,$3E,$B3,$FA,$8B,$6B,$DA,$D7,$F7,$8D
	.dc.b	$7F,$92,$1F,$97,$F2,$DF,$03,$FD,$AB,$24,$62,$7F,$53,$18,$9F,$2F
	.dc.b	$D6,$3E,$43,$CF,$ED,$FF,$0C,$BF,$E7,$58,$FF,$95,$78,$7E,$33,$71
	.dc.b	$F7,$D6,$E7,$3E,$DF,$A8,$97,$EA,$7E,$BE,$EC,$9E,$FF,$88,$95,$F8
	.dc.b	$1C,$BB,$3F,$D0,$7C,$27,$D5,$FF,$7E,$6C,$7E,$1D,$AF,$D6,$C1,$F5
	.dc.b	$8F,$23,$85,$FF,$7E,$BD,$DF,$F7,$C6,$BF,$31,$AF,$86,$3E,$DF,$A7
	.dc.b	$FD,$B7,$0F,$EC,$7B,$74,$AF,$17,$D1,$F9,$48,$FE,$6F,$CE,$7C,$CE
	.dc.b	$23,$F9,$B6,$4A,$F2,$12,$3F,$3B,$19,$7F,$90,$E0,$FA,$C7,$CB,$7E
	.dc.b	$39,$B3,$E2,$3F,$55,$F5,$5F,$C7,$9B,$E5,$7F,$89,$A3,$11,$F1,$FC
	.dc.b	$77,$66,$6B,$FC,$47,$5A,$AF,$F1,$0B,$5F,$C8,$B5,$FE,$EC,$F4,$E7
	.dc.b	$FE,$43,$93,$EB,$3F,$75,$5F,$55,$B3,$FB,$1A,$E1,$D3,$D3,$E7,$7C
	.dc.b	$C4,$BF,$83,$9B,$D1,$C4,$7F,$5F,$B5,$7A,$09,$1F,$C1,$A3,$2F,$CA
	.dc.b	$F8,$47,$D6,$7E,$D9,$B1,$F2,$2B,$E3,$75,$C8,$F9,$6F,$EF,$3E,$27
	.dc.b	$F4,$19,$FC,$28,$F8,$E4,$9F,$03,$5F,$CA,$78,$CC,$3F,$82,$B6,$3A
	.dc.b	$9F,$62,$3E,$B1,$1F,$C0,$DB,$EA,$71,$1F,$06,$7E,$B9,$1F,$E4,$BF
	.dc.b	$57,$FB,$BA,$73,$7F,$08,$8C,$37,$E0,$BF,$85,$CE,$AB,$FC,$61,$5F
	.dc.b	$E0,$98,$C4,$23,$8B,$A9,$E3,$5C,$E3,$FC,$71,$5C,$2B,$94,$FF,$18
	.dc.b	$FA,$B7,$D6,$7F,$3D,$AE,$B1,$8F,$8D,$E3,$5B,$AD,$3E,$9C,$FB,$E3
	.dc.b	$FE,$45,$BE,$0B,$0E,$77,$C2,$34,$D7,$D5,$5C,$95,$72,$AE,$38,$78
	.dc.b	$A7,$8B,$5E,$25,$9E,$18,$79,$E7,$18,$77,$C7,$EC,$5E,$26,$3E,$7A
	.dc.b	$E0,$D3,$E8,$E3,$E5,$C3,$AC,$78,$8D,$1E,$8C,$3B,$4E,$DF,$74,$F0
	.dc.b	$9D,$5F,$A7,$0F,$1C,$72,$9F,$92,$FF,$A7,$9A,$F9,$0C,$57,$EE,$9F
	.dc.b	$EA,$3F,$19,$FE,$84,$E7,$FE,$E6,$D8,$FF,$29,$35,$FF,$54,$B5,$FC
	.dc.b	$17,$23,$97,$FB,$DE,$1F,$BD,$6C,$66,$7F,$CB,$93,$96,$C6,$75,$FC
	.dc.b	$4B,$75,$FE,$CB,$F5,$9F,$EA,$1E,$3F,$AD,$39,$BF,$C4,$35,$FC,$0B
	.dc.b	$67,$84,$7C,$58,$D6,$1E,$1C,$3B,$E3,$F8,$3B,$DC,$62,$F8,$4F,$80
	.dc.b	$D7,$07,$F9,$12,$74,$4C,$BE,$1C,$EB,$2E,$17,$C8,$95,$FB,$2C,$3B
	.dc.b	$DF,$09,$CB,$67,$58,$77,$8C,$1F,$13,$83,$15,$FE,$AD,$BF,$E1,$98
	.dc.b	$7A,$FE,$77,$32,$9C,$BF,$90,$6E,$3E,$3B,$DC,$2A,$C7,$2F,$AC,$FE
	.dc.b	$F3,$87,$F8,$F5,$B8,$7F,$77,$E9,$FF,$09,$4E,$7C,$4F,$F3,$4C,$7F
	.dc.b	$8D,$6B,$6D,$8C,$CF,$F8,$C2,$38,$D7,$FB,$46,$B7,$3F,$E2,$67,$FB
	.dc.b	$73,$FE,$48,$6C,$7F,$A9,$DA,$FF,$BF,$9A,$F1,$F0,$3B,$BF,$ED,$38
	.dc.b	$E3,$5F,$D7,$87,$C8,$EA,$53,$AB,$E4,$7C,$1B,$E7,$C7,$FE,$72,$BE
	.dc.b	$15,$BB,$ED,$F3,$3A,$F8,$97,$8D,$A8,$C3,$72,$F9,$91,$D3,$1F,$E7
	.dc.b	$F1,$82,$5F,$C5,$9D,$70,$F1,$8F,$91,$E4,$79,$3E,$B3,$FC,$1B,$7F
	.dc.b	$5E,$31,$E3,$79,$99,$1F,$E1,$4A,$7D,$66,$1E,$71,$C4,$D4,$67,$83
	.dc.b	$FC,$E1,$F5,$3F,$E4,$29,$FF,$0E,$FD,$67,$B9,$FF,$41,$46,$6F,$FC
	.dc.b	$79,$39,$7F,$CE,$62,$35,$5D,$3C,$2F,$A5,$F5,$9F,$82,$A1,$B8,$FA
	.dc.b	$D5,$C3,$0F,$F3,$5B,$67,$75,$C5,$E1,$3E,$4F,$0B,$EC,$FC,$B7,$C6
	.dc.b	$9F,$1D,$FF,$7D,$AF,$F1,$BC,$E7,$1F,$7D,$F8,$0F,$2F,$8D,$6B,$A1
	.dc.b	$30,$F6,$E3,$FE,$16,$8C,$5F,$F8,$B3,$E4,$3B,$57,$F9,$7D,$B1,$F3
	.dc.b	$3E,$07,$F1,$B1,$8A,$F6,$C7,$C0,$3F,$2D,$E5,$E6,$FE,$2E,$89,$CE
	.dc.b	$7B,$BF,$F5,$A5,$EE,$FE,$85,$7E,$93,$D9,$7D,$63,$D1,$C3,$70,$FF
	.dc.b	$2C,$E7,$FE,$FC,$7F,$D4,$71,$9C,$3E,$4F,$89,$F5,$0F,$2F,$C4,$DA
	.dc.b	$F0,$B8,$7F,$D3,$7D,$0F,$89,$C5,$FE,$13,$E9,$FF,$E0,$05,$7F,$B1
	.dc.b	$9B,$1F,$EA,$83,$F0,$6E,$1C,$35,$FF,$56,$BC,$38,$7F,$D8,$CD,$9F
	.dc.b	$C5,$FC,$AB,$EA,$FF,$D8,$F1,$FE,$D9,$9D,$5F,$6F,$8C,$FE,$09,$18
	.dc.b	$BE,$1F,$05,$89,$FE,$1B,$F1,$0F,$87,$E2,$DF,$57,$31,$83,$FC,$BF
	.dc.b	$97,$FB,$FA,$B3,$3A,$6C,$7E,$83,$4F,$F7,$A7,$D5,$F2,$F6,$F9,$B6
	.dc.b	$74,$7E,$9B,$1F,$14,$67,$5F,$F7,$0C,$62,$77,$08,$F4,$57,$F2,$6F
	.dc.b	$AB,$FF,$73,$55,$F1,$BF,$DA,$AF,$08,$CB,$63,$37,$FE,$EE,$9C,$B6
	.dc.b	$33,$8A,$BF,$77,$5A,$8C,$52,$3F,$DE,$DF,$07,$E1,$AF,$13,$5F,$03
	.dc.b	$F5,$3E,$42,$47,$B3,$99,$F5,$7F,$B9,$8F,$C1,$CE,$B0,$FC,$73,$FD
	.dc.b	$12,$B7,$E3,$E7,$2F,$1E,$BF,$6F,$E1,$7E,$CF,$D5,$FD,$A6,$E5,$F2
	.dc.b	$A7,$4D,$E3,$56,$E3,$0F,$F4,$DF,$AA,$FE,$DD,$1F,$9B,$D3,$F8,$EC
	.dc.b	$61,$C3,$5F,$D0,$35,$FE,$46,$1E,$18,$FC,$A8,$C7,$A9,$FD,$83,$EA
	.dc.b	$9F,$54,$FA,$AF,$14,$CC,$7F,$07,$7C,$1F,$C2,$DF,$0A,$D4,$FE,$2B
	.dc.b	$3F,$E0,$18,$FC,$65,$F5,$5F,$9E,$F0,$FF,$3E,$DC,$7E,$23,$6F,$93
	.dc.b	$FD,$BF,$AD,$F2,$74,$FA,$9C,$BF,$F0,$06,$FC,$B3,$A6,$CE,$DF,$C0
	.dc.b	$4E,$EF,$FA,$6D,$6D,$FE,$9F,$F4,$EF,$93,$F8,$D1,$F2,$A3,$17,$F4
	.dc.b	$E1,$18,$7E,$B3,$5F,$E2,$4E,$74,$FA,$0D,$9F,$EA,$FC,$DF,$92,$E8
	.dc.b	$FF,$06,$4E,$6F,$FA,$DB,$5E,$F8,$7C,$09,$DB,$FE,$13,$F2,$1F,$43
	.dc.b	$4F,$CC,$7D,$3E,$32,$FB,$33,$70,$FE,$DA,$DC,$FF,$08,$D9,$F2,$D5
	.dc.b	$AF,$ED,$B3,$7C,$27,$D3,$7F,$35,$BF,$E0,$F1,$8F,$1B,$F8,$DF,$B7
	.dc.b	$FE,$2C,$8C,$4E,$AB,$86,$0C,$BF,$7B,$5A,$8F,$C5,$5F,$1A,$DD,$7E
	.dc.b	$66,$BF,$A7,$57,$F8,$19,$B1,$F8,$FF,$4F,$FC,$6D,$18,$C3,$F4,$6D
	.dc.b	$AF,$E7,$BE,$57,$F3,$0A,$DB,$6B,$F6,$D1,$FC,$D3,$E7,$F1,$5F,$9D
	.dc.b	$8B,$87,$E7,$66,$D7,$ED,$FE,$07,$FC,$27,$F4,$CF,$9E,$F0,$EF,$FC
	.dc.b	$6A,$31,$97,$F5,$7A,$DD,$6A,$EC,$E4,$57,$05,$A7,$F7,$B9,$DC,$7F
	.dc.b	$BF,$8D,$D7,$36,$9C,$EB,$95,$71,$8C,$9C,$63,$38,$79,$EB,$D9,$3A
	.dc.b	$AE,$15,$E3,$8F,$F1,$8E,$1E,$29,$CD,$70,$AE,$78,$79,$AF,$67,$3C
	.dc.b	$3B,$3E,$5A,$F4,$57,$58,$FF,$80,$06,$EB,$93,$88,$EA,$FA,$8E,$AE
	.dc.b	$0E,$58,$79,$6F,$BC,$FB,$EF,$B9,$EE,$F7,$79,$9C,$23,$FC,$EB,$1F
	.dc.b	$2E,$31,$7D,$63,$DC,$72,$9F,$51,$FE,$0A,$AE,$EF,$92,$34,$FD,$7A
	.dc.b	$EF,$1E,$13,$B1,$D2,$39,$61,$F0,$30,$F5,$36,$3E,$21,$F0,$4F,$89
	.dc.b	$3D,$9F,$01,$C8,$D4,$F2,$69,$CB,$0F,$6C,$F1,$8E,$67,$18,$CB,$63
	.dc.b	$CC,$FE,$53,$0F,$7B,$C0,$C5,$F9,$EB,$E4,$47,$C6,$32,$78,$9C,$BF
	.dc.b	$54,$EC,$7E,$8F,$E2,$BF,$D4,$ED,$C7,$F2,$B1,$88,$FD,$13,$7E,$21
	.dc.b	$F0,$1F,$C3,$E5,$39,$8C,$37,$6F,$F4,$5B,$67,$77,$FD,$B2,$F9,$D7
	.dc.b	$EC,$3E,$95,$F2,$BF,$DC,$32,$FD,$B4,$63,$0F,$DA,$DF,$07,$F5,$ED
	.dc.b	$9D,$D7,$1D,$7F,$BC,$DF,$1B,$DE,$7E,$9B,$FE,$05,$87,$1B,$DE,$1C
	.dc.b	$6F,$FD,$47,$87,$1B,$DE,$1C,$6F,$D4,$DA,$DD,$F1,$57,$C6,$B7,$8F
	.dc.b	$AA,$F8,$56,$AF,$85,$EE,$3F,$D6,$78,$F9,$5B,$5C,$27,$51,$C4,$E9
	.dc.b	$71,$FE,$27,$1A,$37,$5C,$E0,$D8,$D5,$FA,$63,$47,$CF,$CB,$E6,$D7
	.dc.b	$FB,$19,$AF,$EF,$0F,$2D,$E7,$1F,$7C,$78,$AF,$DA,$75,$C7,$E0,$E1
	.dc.b	$E1,$8C,$CF,$A7,$2F,$9F,$87,$38,$F9,$A5,$3E,$01,$E1,$3D,$77,$F9
	.dc.b	$5C,$78,$C7,$84,$FC,$75,$F8,$6B,$B6,$3F,$80,$9C,$98,$3D,$67,$12
	.dc.b	$61,$D0,$91,$C6,$32,$FC,$08,$CC,$7A,$0C,$9D,$1C,$BE,$A7,$E5,$98
	.dc.b	$C7,$E2,$47,$90,$FC,$CD,$79,$5E,$15,$E1,$AE,$FE,$47,$FA,$97,$DD
	.dc.b	$FF,$2C,$DF,$F9,$25,$BB,$7F,$A6,$7C,$34,$63,$1F,$F4,$B4,$3D,$AC
	.dc.b	$61,$CE,$7F,$CE,$15,$8A,$E3,$1F,$E7,$3A,$E4,$FE,$83,$E9,$AF,$01
	.dc.b	$A7,$1B,$B3,$FE,$64,$9D,$36,$74,$FF,$89,$3E,$99,$E4,$F9,$F9,$91
	.dc.b	$89,$D4,$F9,$7B,$B5,$39,$9F,$F1,$36,$BF,$ED,$48,$C6,$DF,$E2,$86
	.dc.b	$C7,$C5,$AF,$5F,$CC,$7F,$23,$AF,$F4,$03,$67,$FD,$0A,$DB,$0F,$6D
	.dc.b	$F1,$7C,$B9,$A4,$FF,$A0,$AB,$83,$B9,$F9,$3F,$4A,$F9,$6A,$49,$FF
	.dc.b	$44,$F4,$FF,$9E,$9F,$F4,$5C,$ED,$B3,$B7,$DE,$79,$27,$83,$81,$3C
	.dc.b	$5F,$F5,$33,$5F,$F5,$09,$A9,$FF,$3C,$34,$FF,$3D,$7D,$2B,$E9,$7F
	.dc.b	$02,$D8,$FF,$4F,$B5,$F0,$7D,$2B,$C0,$DD,$49,$DB,$F8,$A9,$D4,$7F
	.dc.b	$AF,$FD,$7F,$C8,$76,$48,$C1,$0F,$F5,$43,$97,$FD,$55,$1F,$94,$F9
	.dc.b	$27,$D2,$D2,$4F,$FA,$D3,$E2,$FE,$BB,$FE,$B6,$9D,$B6,$76,$FC,$99
	.dc.b	$DB,$F2,$E7,$6F,$FB,$3D,$AF,$CA,$6B,$FE,$CC,$6B,$FB,$0F,$27,$F5
	.dc.b	$9E,$3C,$8F,$F6,$F4,$7E,$BC,$C3,$FE,$C5,$8C,$3E,$7F,$83,$F7,$11
	.dc.b	$FF,$67,$FD,$2B,$E0,$71,$1E,$79,$BD,$3F,$ED,$BA,$D4,$FF,$B5,$BE
	.dc.b	$97,$F0,$F3,$FE,$DF,$5F,$4B,$FB,$E6,$CE,$FD,$6C,$19,$99,$19,$21
	.dc.b	$F9,$19,$E2,$D8,$E2,$FF,$BF,$7D,$4F,$E1,$DE,$64,$8C,$10,$FF,$73
	.dc.b	$39,$7F,$DD,$11,$FC,$63,$E9,$7F,$DA,$51,$FB,$4C,$7F,$DD,$4F,$F1
	.dc.b	$93,$5E,$9C,$75,$E0,$7F,$25,$74,$60,$79,$4B,$EB,$1F,$AB,$FA,$47
	.dc.b	$D2,$FF,$22,$6C,$7F,$C0,$06,$BF,$B1,$F9,$CF,$F1,$4B,$99,$E0,$47
	.dc.b	$C0,$23,$6F,$78,$FD,$9C,$F4,$8F,$DB,$B5,$FE,$26,$1C,$2F,$85,$7D
	.dc.b	$2F,$EF,$B4,$67,$FC,$6A,$73,$5F,$D1,$5A,$FC,$3F,$8D,$FF,$21,$FD
	.dc.b	$07,$F4,$0F,$A3,$7D,$2D,$25,$7C,$9B,$E2,$FF,$14,$C1,$E4,$3D,$DF
	.dc.b	$EA,$51,$86,$C6,$67,$97,$D2,$BE,$27,$C9,$E6,$7E,$CB,$6F,$4D,$7F
	.dc.b	$32,$D3,$FC,$81,$18,$84,$63,$2F,$75,$71,$7E,$3C,$66,$79,$7B,$3E
	.dc.b	$D3,$BF,$9B,$F0,$F2,$7B,$94,$95,$F6,$2F,$92,$EF,$7C,$E7,$F6,$66
	.dc.b	$C7,$DE,$6B,$FC,$E3,$81,$F4,$AF,$1B,$85,$CA,$E0,$FD,$59,$D4,$60
	.dc.b	$FA,$1E,$77,$F5,$5B,$FE,$99,$BF,$19,$FE,$85,$8B,$E7,$7C,$84,$8F
	.dc.b	$C8,$FC,$3B,$E3,$F1,$1C,$26,$F1,$FC,$DA,$D4,$FA,$E7,$F1,$1F,$4A
	.dc.b	$F8,$FA,$70,$E5,$FC,$AF,$D2,$3E,$5F,$FC,$12,$DC,$3F,$42,$D7,$E2
	.dc.b	$78,$AE,$DD,$D4,$9D,$BF,$97,$9D,$46,$0F,$C9,$7D,$2F,$EA,$B5,$FC
	.dc.b	$3F,$0A,$31,$97,$93,$0F,$1F,$CA,$CC,$3C,$6F,$5F,$43,$11,$FD,$AF
	.dc.b	$44,$E3,$2E,$B9,$AF,$18,$FE,$B7,$E0,$9F,$37,$FA,$D6,$E5,$D8,$F8
	.dc.b	$3D,$7F,$10,$8F,$EA,$FE,$23,$F3,$0F,$F8,$43,$4F,$C9,$BC,$08,$7E
	.dc.b	$2E,$36,$A3,$6F,$F7,$E3,$13,$FD,$FA,$33,$1F,$E0,$26,$BF,$62,$BE
	.dc.b	$B9,$1F,$27,$D0,$D7,$E2,$45,$D3,$E7,$E0,$AF,$F2,$3F,$D2,$FF,$6A
	.dc.b	$8F,$D3,$5F,$F8,$AA,$31,$D1,$FC,$69,$B1,$9F,$5B,$F8,$ED,$70,$84
	.dc.b	$E9,$BE,$0E,$AF,$93,$63,$91,$FE,$38,$AE,$5F,$48,$FA,$5F,$E6,$B1
	.dc.b	$86,$DF,$F2,$C9,$D7,$C6,$71,$8F,$36,$7D,$70,$56,$96,$1C,$EF,$84
	.dc.b	$7F,$92,$27,$53,$EB,$CB,$E3,$4E,$8E,$78,$F8,$9E,$0D,$73,$7B,$9E
	.dc.b	$31,$E0,$C7,$D1,$5B,$C3,$BC,$62,$F8,$63,$E9,$9C,$DF,$69,$F1,$46
	.dc.b	$23,$E7,$CE,$EF,$A1,$2B,$9C,$F0,$8C,$3E,$33,$A1,$4E,$66,$E3,$F2
	.dc.b	$8F,$B0,$FC,$B5,$7C,$B7,$FC,$AA,$D7,$FA,$FF,$C6,$7F,$9A,$7A,$38
	.dc.b	$B6,$3F,$3C,$D7,$FC,$BA,$D7,$EE,$72,$39,$7C,$37,$FB,$96,$C6,$67
	.dc.b	$C6,$D7,$C6,$E7,$5F,$C5,$B7,$4F,$EE,$7F,$4B,$FE,$84,$E6,$F9,$73
	.dc.b	$9B,$FC,$13,$5F,$C4,$B6,$78,$47,$C7,$C7,$C8,$D9,$E1,$18,$C3,$F4
	.dc.b	$95,$C1,$F1,$E3,$E0,$25,$FD,$19,$D6,$3F,$0A,$73,$5C,$2B,$5C,$1F
	.dc.b	$DB,$A3,$04,$AE,$13,$AC,$B8,$C6,$27,$53,$FE,$A6,$C3,$BB,$F0,$30
	.dc.b	$FD,$1F,$A5,$FE,$2D,$FA,$6F,$EF,$7C,$9F,$DF,$23,$FC,$61,$F5,$18
	.dc.b	$CF,$FD,$5F,$1B,$7F,$C6,$CE,$9B,$EA,$2F,$86,$3F,$E3,$A8,$5E,$2F
	.dc.b	$94,$FF,$AA,$1E,$4F,$FA,$AB,$E2,$9E,$67,$FA,$7F,$1F,$E5,$11,$8F
	.dc.b	$82,$FF,$07,$57,$FB,$0A,$31,$3A,$AE,$13,$FE,$AC,$9C,$BD,$63,$39
	.dc.b	$FF,$54,$FB,$FF,$F0,$6D,$FF,$94,$6F,$73,$FE,$1E,$5F,$9D,$FF,$69
	.dc.b	$9F,$7F,$5F,$BA,$73,$E0,$FB,$76,$FF,$50,$D7,$3B,$E1,$9F,$8D,$AF
	.dc.b	$F9,$22,$F8,$CF,$BA,$73,$3E,$D7,$8E,$3E,$36,$CF,$F9,$27,$27,$C5
	.dc.b	$78,$E7,$36,$C3,$D3,$F0,$1E,$5E,$6D,$37,$1F,$33,$F3,$2B,$85,$FF
	.dc.b	$A6,$E7,$C9,$39,$BE,$71,$83,$89,$FE,$51,$F8,$FF,$F0,$93,$8B,$FB
	.dc.b	$97,$FE,$4B,$8F,$C3,$FD,$E7,$9F,$D4,$F9,$0D,$8F,$85,$1D,$E7,$85
	.dc.b	$7C,$98,$C1,$B7,$FC,$75,$6F,$B2,$C6,$DF,$1F,$C5,$FA,$9F,$79,$F1
	.dc.b	$5B,$97,$D5,$7B,$11,$FA,$4D,$C3,$D2,$FE,$67,$CE,$FF,$07,$4F,$13
	.dc.b	$EF,$7C,$3F,$E0,$7C,$5F,$A4,$D8,$C9,$F7,$5F,$61,$1C,$1A,$BF,$D1
	.dc.b	$7D,$87,$F1,$2C,$BF,$01,$5F,$8D,$F9,$19,$7F,$8D,$BA,$7C,$E3,$E4
	.dc.b	$11,$FC,$63,$F7,$7B,$9F,$84,$7F,$99,$E7,$91,$FE,$B1,$8C,$F8,$5F
	.dc.b	$8B,$AF,$CE,$7D,$CC,$AF,$DF,$FD,$CD,$9F,$DC,$78,$FF,$CA,$1F,$CE
	.dc.b	$35,$FD,$B3,$5F,$CA,$35,$7D,$3F,$E3,$DF,$F1,$33,$5F,$F5,$69,$AE
	.dc.b	$9F,$CB,$36,$FF,$82,$D7,$F8,$7D,$FE,$27,$EC,$7E,$D6,$31,$3F,$C3
	.dc.b	$FB,$93,$9E,$2F,$CE,$56,$AB,$D3,$18,$25,$FF,$05,$9E,$7C,$3F,$C9
	.dc.b	$FE,$41,$F5,$FC,$EB,$F9,$2F,$2F,$F0,$E7,$F9,$4E,$6F,$94,$FC,$57
	.dc.b	$8B,$FC,$1D,$B3,$FC,$A9,$BA,$79,$DF,$E1,$9F,$07,$27,$AF,$C1,$DA
	.dc.b	$F6,$4D,$D3,$F9,$D6,$5F,$EB,$FB,$FE,$6B,$3F,$C5,$7C,$79,$1F,$EE
	.dc.b	$FF,$B2,$FF,$75,$B8,$C3,$E3,$4D,$C3,$FD,$7F,$5A,$AF,$E7,$F3,$97
	.dc.b	$DA,$FF,$40,$8D,$57,$29,$FE,$8B,$F7,$1F,$C9,$4F,$D5,$34,$F8,$E5
	.dc.b	$8F,$E8,$98,$FF,$B9,$A3,$89,$FD,$3B,$E6,$7F,$97,$1F,$D2,$FF,$40
	.dc.b	$FE,$3F,$E6,$38,$13,$AA,$E0,$7F,$58,$AD,$CE,$A3,$FE,$00,$D7,$C7
	.dc.b	$C1,$A7,$FB,$6F,$C8,$FE,$37,$18,$AF,$F7,$FD,$6E,$B5,$EE,$FF,$60
	.dc.b	$A4,$F3,$7F,$32,$D7,$8F,$07,$E6,$A3,$11,$FD,$BA,$39,$47,$5B,$4C
	.dc.b	$ED,$7D,$5E,$AF,$FB,$67,$E8,$3F,$E6,$BB,$FE,$EA,$DB,$FE,$D9,$FA
	.dc.b	$17,$AB,$3F,$EF,$D3,$96,$FB,$3C,$8F,$EB,$93,$BB,$FF,$00,$D6,$DF
	.dc.b	$CD,$F8,$3F,$D7,$CF,$EF,$9F,$72,$FA,$3C,$4F,$F8,$1A,$30,$FF,$81
	.dc.b	$A7,$73,$AC,$BE,$61,$EE,$FA,$17,$B9,$F2,$FF,$43,$7D,$57,$CD,$9C
	.dc.b	$E3,$F1,$36,$3F,$1B,$05,$F0,$8D,$19,$AE,$33,$FE,$B3,$6B,$FC,$39
	.dc.b	$AE,$F0,$FE,$1F,$1C,$2F,$BC,$F9,$E7,$F0,$1F,$98,$FF,$80,$23,$D6
	.dc.b	$2B,$85,$FF,$02,$FA,$9F,$F0,$31,$F8,$7F,$AF,$FB,$71,$FE,$2A,$8C
	.dc.b	$71,$EA,$3D,$37,$F8,$2A,$DB,$FC,$6E,$BF,$C2,$1F,$19,$2B,$E8,$78
	.dc.b	$96,$BF,$C5,$DC,$28,$C6,$BF,$E3,$F8,$C5,$7F,$8E,$AF,$F3,$57,$FE
	.dc.b	$B7,$0B,$E8,$3F,$C3,$1F,$35,$F0,$A3,$13,$FC,$DB,$E0,$7F,$2F,$18
	.dc.b	$8F,$F0,$94,$EE,$BF,$DC,$55,$F9,$CF,$A7,$D7,$B6,$AF,$9B,$1C,$6F
	.dc.b	$C4,$EA,$33,$7C,$A5,$39,$C7,$A4,$E6,$31,$3A,$8E,$18,$E0,$ED,$1F
	.dc.b	$EC,$83,$DF,$1E,$EA,$F5,$BC,$0C,$46,$8F,$80,$53,$DF,$1C,$1C,$D7
	.dc.b	$88,$EF,$8F,$9A,$3F,$CB,$58,$7B,$1E,$78,$79,$70,$F0,$46,$B0,$E8
	.dc.b	$FF,$BD,$0F,$61,$92,$99,$C3,$C1,$3F,$0C,$F4,$BE,$28,$DE,$3F,$46
	.dc.b	$BC,$91,$C6,$B6,$71,$7B,$57,$1B,$F3,$47,$A2,$7E,$2C,$75,$9F,$0C
	.dc.b	$FC,$29,$E6,$D3,$9B,$D5,$F9,$27,$A8,$FC,$94,$F6,$7E,$03,$5F,$C2
	.dc.b	$C7,$47,$D0,$74,$3B,$1D,$27,$C9,$1B,$8E,$AF,$A0,$F9,$EF,$8A,$76
	.dc.b	$F3,$3A,$9C,$E3,$D1,$7F,$26,$3D,$07,$E2,$EB,$E0,$BC,$23,$51,$D2
	.dc.b	$39,$1F,$24,$FB,$55,$F3,$6B,$C9,$1E,$AA,$E3,$1E,$0C,$3F,$11,$3F
	.dc.b	$12,$36,$D3,$9B,$EC,$9F,$35,$75,$9C,$9F,$9A,$E2,$FD,$FB,$63,$F5
	.dc.b	$D1,$88,$FC,$83,$7C,$AF,$F0,$E6,$BF,$D5,$A3,$1E,$47,$F6,$AA,$FE
	.dc.b	$77,$7F,$CD,$6F,$71,$83,$FA,$EB,$8A,$E3,$7F,$9E,$FA,$7F,$F6,$1D
	.dc.b	$5F,$23,$F9,$F6,$1F,$C3,$70,$E6,$DC,$7C,$0F,$F1,$1B,$E3,$7F,$B5
	.dc.b	$6C,$6F,$0F,$D5,$DF,$93,$5E,$18,$FF,$4B,$6E,$1F,$A4,$9D,$69,$EB
	.dc.b	$CB,$BB,$FD,$47,$2E,$AD,$C3,$CE,$EF,$1E,$F5,$F1,$71,$EB,$8F,$EE
	.dc.b	$30,$F3,$C7,$96,$3F,$77,$8F,$81,$FF,$01,$CE,$B0,$ED,$1E,$5B,$E9
	.dc.b	$1D,$E3,$A6,$3E,$B8,$DC,$74,$9D,$BF,$49,$F5,$BD,$A7,$D2,$D3,$D3
	.dc.b	$8C,$7A,$9F,$C2,$2F,$C8,$F3,$8F,$21,$99,$F1,$47,$33,$C5,$1A,$7E
	.dc.b	$B4,$78,$14,$FC,$C7,$A1,$FC,$BF,$0F,$88,$EA,$7A,$B4,$E9,$8F,$CB
	.dc.b	$3E,$79,$31,$F8,$CF,$CF,$CB,$E9,$92,$3E,$53,$C5,$F9,$31,$C6,$7C
	.dc.b	$6E,$5F,$63,$95,$7E,$63,$CF,$87,$C6,$8F,$31,$F0,$89,$7E,$17,$E3
	.dc.b	$3E,$5B,$F3,$9E,$22,$47,$89,$F3,$9F,$7A,$39,$47,$8B,$2F,$BA,$DC
	.dc.b	$BF,$58,$C3,$E6,$46,$B0,$F3,$E5,$F3,$31,$F4,$47,$AE,$7F,$13,$5E
	.dc.b	$AC,$BE,$04,$E7,$0E,$98,$F8,$A3,$F6,$F9,$7F,$97,$3C,$07,$0B,$D9
	.dc.b	$FF,$1D,$61,$FE,$2E,$9F,$F3,$E6,$D8,$E1,$FF,$55,$B7,$4F,$F3,$7E
	.dc.b	$1C,$5F,$F1,$3D,$70,$6F,$89,$FE,$4C,$6D,$71,$9F,$0E,$3C,$E5,$FA
	.dc.b	$27,$F8,$67,$2F,$F5,$5F,$C9,$EF,$C9,$3B,$9D,$69,$EB,$9C,$DF,$FB
	.dc.b	$0A,$B8,$1B,$D7,$B7,$D4,$FF,$9E,$A7,$FC,$7A,$1C,$3F,$E4,$A6,$E3
	.dc.b	$FE,$50,$87,$E7,$6F,$63,$9F,$6F,$FB,$9A,$7F,$CC,$57,$C9,$FF,$74
	.dc.b	$B5,$CC,$6D,$FF,$46,$C7,$6C,$FF,$CD,$5F,$70,$F4,$3E,$23,$6B,$CF
	.dc.b	$5B,$7E,$8F,$C1,$7C,$5F,$94,$91,$FE,$81,$8A,$FF,$AF,$22,$E9,$FE
	.dc.b	$51,$9F,$A5,$CA,$F5,$E9,$2F,$FC,$C3,$DA,$FC,$29,$8E,$67,$FC,$BB
	.dc.b	$57,$1F,$F3,$AD,$EE,$BE,$E7,$E7,$6F,$A3,$A4,$C3,$E8,$C7,$E4,$9B
	.dc.b	$1F,$15,$47,$17,$E4,$35,$FA,$0D,$7F,$D2,$47,$08,$FC,$67,$E8,$AF
	.dc.b	$C0,$A9,$E1,$9F,$F6,$FF,$A7,$F8,$D2,$3F,$ED,$7C,$EB,$A3,$F3,$BD
	.dc.b	$09,$FC,$D7,$AB,$F2,$67,$FD,$BD,$CB,$F9,$7F,$95,$91,$83,$8B,$87
	.dc.b	$FD,$DE,$F2,$8F,$CF,$72,$49,$EE,$FE,$BF,$D9,$7E,$7E,$F8,$FD,$57
	.dc.b	$FB,$E2,$AF,$1F,$FA,$9A,$F7,$58,$E9,$FD,$35,$5B,$56,$9F,$6E,$FF
	.dc.b	$AF,$23,$FE,$F9,$F3,$6B,$FE,$C8,$F0,$5C,$1F,$36,$38,$9B,$3F,$E0
	.dc.b	$07,$C4,$FE,$E7,$E2,$64,$60,$87,$39,$46,$A3,$B7,$0A,$BF,$88,$74
	.dc.b	$7F,$54,$FA,$6B,$E5,$9A,$7F,$0D,$F8,$47,$BD,$8C,$7F,$6D,$7C,$5B
	.dc.b	$7C,$5B,$7F,$B5,$9D,$47,$97,$F0,$3F,$E2,$47,$37,$FD,$9F,$F4,$AF
	.dc.b	$A6,$FC,$C3,$63,$F7,$CD,$7F,$83,$35,$FE,$5F,$CF,$F6,$DF,$E1,$5C
	.dc.b	$FF,$4D,$FC,$DB,$5D,$F3,$FF,$09,$E5,$7E,$8C,$F8,$4F,$C3,$36,$BF
	.dc.b	$DB,$33,$A9,$CF,$CB,$FF,$2F,$F9,$A7,$C6,$FC,$AF,$92,$98,$7F,$24
	.dc.b	$9C,$BF,$C0,$B8,$93,$F1,$5B,$1F,$2A,$30,$7F,$1D,$E9,$FC,$2F,$8F
	.dc.b	$FC,$E4,$93,$CE,$7E,$CF,$C6,$C8,$EC,$FD,$EF,$21,$9E,$67,$F7,$BC
	.dc.b	$BF,$E1,$4A,$F1,$12,$3E,$DB,$FD,$1F,$E7,$A4,$74,$3D,$5F,$15,$F8
	.dc.b	$82,$3F,$D6,$3B,$1F,$6D,$88,$EF,$B2,$BE,$01,$23,$F1,$2F,$F5,$5E
	.dc.b	$89,$1D,$8F,$91,$EA,$FE,$58,$88,$73,$BC,$4C,$7C,$DB,$E3,$FC,$15
	.dc.b	$C8,$FC,$BD,$FD,$12,$38,$3F,$8A,$FC,$77,$F6,$73,$BB,$DA,$FC,$1C
	.dc.b	$0F,$C8,$5E,$57,$EB,$1B,$5F,$D0,$1A,$F1,$19,$79,$7D,$CE,$15,$FA
	.dc.b	$7A,$DD,$7B,$45,$F7,$9D,$12,$3F,$95,$FD,$27,$F7,$D3,$9B,$E1,$F0
	.dc.b	$BC,$84,$7E,$46,$DF,$1D,$F1,$FE,$4E,$F8,$AD,$47,$E1,$B2,$5F,$94
	.dc.b	$91,$FC,$0A,$32,$FF,$3C,$F8,$AF,$F0,$71,$CB,$0F,$EB,$72,$E7,$F8
	.dc.b	$C4,$AF,$1E,$FF,$63,$27,$76,$67,$C5,$5F,$14,$8F,$F8,$1F,$BB,$F9
	.dc.b	$4F,$D6,$FF,$87,$7D,$49,$3F,$D3,$A3,$FC,$40,$61,$FE,$4D,$18,$7F
	.dc.b	$C3,$3F,$3E,$F9,$0F,$F1,$77,$C4,$3E,$3B,$E9,$4E,$7A,$7F,$21,$39
	.dc.b	$BF,$C9,$4E,$54,$65,$7D,$03,$E5,$7E,$E4,$E6,$B8,$37,$1F,$0E,$3E
	.dc.b	$06,$E1,$F9,$FC,$39,$3E,$29,$D6,$36,$3C,$75,$C1,$61,$DA,$BF,$19
	.dc.b	$7C,$A7,$CF,$1C,$9A,$72,$BE,$8F,$2B,$E9,$5C,$70,$F4,$56,$B0,$ED
	.dc.b	$7C,$4E,$67,$C2,$C7,$E0,$E3,$DE,$F8,$CE,$AB,$8B,$B3,$A6,$3E,$B8
	.dc.b	$C9,$AC,$7C,$57,$E9,$AF,$F0,$CC,$FD,$36,$D7,$F9,$A5,$B8,$FF,$92
	.dc.b	$7E,$1F,$F8,$37,$5D,$ED,$FD,$AD,$7F,$94,$E7,$2D,$8C,$BD,$BC,$47
	.dc.b	$CA,$7F,$94,$F2,$FE,$5E,$D8,$CC,$FF,$97,$67,$2D,$8C,$EB,$F8,$36
	.dc.b	$E9,$FC,$07,$E8,$1F,$2D,$FB,$7A,$D6,$3F,$3E,$F7,$87,$E4,$9A,$AF
	.dc.b	$8F,$3E,$EA,$4F,$CF,$AD,$D6,$A3,$13,$AB,$EF,$1F,$6A,$75,$18,$26
	.dc.b	$3F,$A6,$BE,$78,$F9,$E3,$18,$73,$6E,$1C,$E7,$4F,$A6,$74,$D3,$58
	.dc.b	$E3,$91,$E3,$7F,$17,$8F,$F2,$5F,$0F,$F1,$BE,$AF,$F5,$84,$E6,$FF
	.dc.b	$C4,$93,$9A,$FF,$5C,$4E,$6B,$FD,$6D,$EB,$FF,$01,$9F,$EB,$5D,$DB
	.dc.b	$C1,$8F,$F8,$DB,$1C,$43,$EF,$DF,$47,$FE,$9D,$8F,$E7,$11,$8C,$3F
	.dc.b	$D2,$CD,$8F,$F4,$F3,$5F,$F3,$7B,$9C,$BF,$D4,$CD,$9F,$E8,$5F,$86
	.dc.b	$7F,$B3,$5F,$F5,$7C,$66,$BF,$B1,$3F,$E9,$18,$FE,$57,$1F,$D3,$BE
	.dc.b	$81,$F0,$BE,$8A,$D4,$FF,$53,$E8,$56,$AF,$AF,$C1,$7A,$EB,$AB,$5F
	.dc.b	$80,$DB,$E7,$1E,$A8,$EC,$7C,$6C,$7F,$C4,$B5,$D0,$F4,$F8,$1F,$37
	.dc.b	$1F,$F7,$47,$C1,$79,$63,$1F,$1B,$FE,$85,$AD,$9F,$2A,$3C,$04,$70
	.dc.b	$6B,$1F,$ED,$C3,$F0,$1F,$EE,$B1,$9D,$BE,$6D,$F6,$F8,$0F,$88,$D9
	.dc.b	$F9,$AF,$B4,$8F,$DA,$6D,$F6,$D7,$EA,$3E,$0C,$3F,$B3,$D7,$23,$FD
	.dc.b	$CB,$EB,$FF,$84,$DF,$F5,$0F,$57,$E2,$A3,$15,$FE,$79,$8C,$77,$7C
	.dc.b	$F9,$CC,$FF,$AD,$1A,$FF,$81,$5A,$FF,$88,$DE,$79,$FF,$BB,$DB,$7F
	.dc.b	$E1,$8F,$C0,$3F,$CA,$2F,$92,$FC,$5F,$25,$F9,$FF,$39,$F3,$FF,$C0
	.dc.b	$F1,$E5,$39,$8C,$5F,$86,$3C,$C4,$7F,$DE,$98,$A3,$19,$7D,$48,$C4
	.dc.b	$FF,$B9,$32,$7A,$DE,$28,$FC,$CF,$73,$E7,$FF,$B9,$46,$1B,$87,$DE
	.dc.b	$FC,$7F,$FC,$19,$39,$F0,$BF,$6B,$18,$9F,$F6,$D8,$70,$FF,$B7,$9A
	.dc.b	$FF,$8B,$DA,$FF,$8C,$31,$FF,$2C,$57,$EB,$23,$97,$99,$F1,$E7,$31
	.dc.b	$E3,$8F,$95,$E3,$FC,$68,$FF,$1C,$B7,$77,$E1,$1F,$9D,$9F,$F2,$27
	.dc.b	$0A,$B8,$43,$E8,$7F,$5A,$FF,$90,$DA,$FF,$98,$5A,$EF,$B7,$F7,$14
	.dc.b	$C9,$F3,$7E,$73,$F8,$E7,$D8,$3F,$33,$FC,$BE,$1E,$2A,$DE,$DE,$FF
	.dc.b	$17,$FD,$12,$FF,$0C,$5F,$3A,$F6,$BF,$91,$C6,$AF,$FD,$B9,$39,$AE
	.dc.b	$5D,$7F,$C2,$6B,$55,$FE,$5D,$3F,$96,$C6,$49,$3C,$0F,$E3,$7A,$7F
	.dc.b	$A2,$7C,$1E,$F3,$FC,$D6,$FF,$A1,$FC,$ED,$F2,$BF,$D7,$58,$7F,$A4
	.dc.b	$E3,$1D,$7B,$F2,$D1,$89,$D1,$ED,$AE,$39,$FF,$BA,$0E,$9F,$47,$C5
	.dc.b	$B3,$FE,$9F,$69,$B7,$94,$79,$7F,$47,$7D,$57,$F4,$BF,$87,$FF,$6F
	.dc.b	$56,$63,$1F,$01,$FE,$99,$8F,$9D,$5C,$1F,$F6,$8C,$67,$0F,$F4,$56
	.dc.b	$AA,$D4,$7F,$A3,$70,$EB,$7C,$B0,$EA,$F7,$39,$9F,$BF,$F6,$5F,$13
	.dc.b	$F2,$63,$0D,$7E,$8C,$69,$B1,$ED,$FA,$AD,$C7,$FB,$07,$D1,$F9,$DF
	.dc.b	$55,$F3,$DB,$19,$3F,$D8,$D1,$96,$9F,$AD,$F1,$7F,$1A,$DF,$95,$FE
	.dc.b	$DF,$18,$D3,$F9,$EA,$8C,$F4,$7F,$55,$8D,$4F,$E3,$5A,$FF,$AA,$DA
	.dc.b	$FF,$B2,$5E,$9E,$17,$FA,$8E,$3F,$C7,$73,$FD,$FB,$E9,$1F,$21,$DB
	.dc.b	$1F,$D1,$7E,$59,$F4,$30,$FF,$55,$76,$3E,$A3,$F7,$91,$FE,$AD,$E5
	.dc.b	$56,$EF,$FD,$93,$5A,$7F,$DB,$53,$AE,$3F,$E6,$53,$9D,$3F,$C2,$37
	.dc.b	$C5,$FF,$61,$C6,$71,$7C,$9B,$E8,$7F,$49,$F2,$1F,$E1,$DC,$FF,$DA
	.dc.b	$F3,$97,$F7,$51,$9C,$3F,$D9,$B3,$97,$F8,$9B,$74,$F5,$64,$E4,$FE
	.dc.b	$8D,$39,$C3,$FD,$AD,$BF,$F2,$2B,$F9,$4E,$E3,$F9,$43,$97,$75,$F3
	.dc.b	$7E,$33,$FC,$19,$3F,$CA,$B1,$FE,$08,$D7,$AC,$68,$E1,$5F,$CC,$FE
	.dc.b	$8F,$C5,$21,$5D,$DC,$7C,$CF,$C9,$5F,$33,$F8,$9B,$FE,$79,$39,$8C
	.dc.b	$4F,$F8,$2A,$38,$13,$1F,$F2,$0B,$F4,$F7,$FA,$3E,$2F,$D0,$E5,$F8
	.dc.b	$AD,$C7,$F8,$CE,$DF,$C8,$F1,$F9,$B3,$FC,$67,$77,$D0,$70,$BF,$E9
	.dc.b	$D5,$FE,$CE,$1F,$20,$FA,$2F,$74,$33,$FE,$37,$5F,$06,$B5,$3F,$CB
	.dc.b	$7E,$17,$17,$C2,$F7,$3D,$23,$FB,$2C,$FC,$9F,$A4,$E3,$3D,$7C,$66
	.dc.b	$39,$BD,$CA,$B7,$87,$47,$8C,$7A,$5C,$DF,$28,$E1,$5C,$A7,$37,$D6
	.dc.b	$7D,$06,$8F,$B9,$7D,$6B,$81,$EB,$BE,$97,$E0,$AF,$F7,$89,$ED,$38
	.dc.b	$C7,$87,$1D,$63,$F1,$27,$38,$D7,$FD,$EA,$78,$4F,$84,$F9,$5C,$DF
	.dc.b	$A1,$D1,$F5,$EB,$07,$67,$C4,$72,$29,$F1,$70,$F5,$57,$58,$C1,$E4
	.dc.b	$3A,$9E,$CA,$F0,$1D,$4B,$3F,$12,$38,$46,$63,$AC,$72,$BF,$76,$3F
	.dc.b	$82,$C3,$CC,$7A,$2B,$95,$79,$EF,$93,$D2,$33,$5F,$4E,$3E,$BC,$F8
	.dc.b	$9F,$B3,$5C,$DB,$5E,$28,$E7,$1D,$63,$9E,$3F,$69,$B8,$FC,$39,$F8
	.dc.b	$11,$EA,$BF,$3B,$CC,$FC,$9D,$F7,$9F,$AB,$1B,$39,$BF,$04,$FB,$98
	.dc.b	$7A,$5D,$9E,$D3,$D7,$1C,$5A,$64,$FC,$2C,$F7,$7F,$01,$1E,$37,$67
	.dc.b	$B4,$B3,$DD,$D9,$F7,$67,$D6,$D9,$F5,$9C,$DE,$ED,$7F,$82,$7C,$4F
	.dc.b	$F0,$47,$F4,$ED,$7F,$32,$D7,$73,$FC,$23,$D2,$FF,$3A,$73,$FF,$27
	.dc.b	$AE,$11,$8A,$46,$3D,$95,$6E,$F9,$46,$2B,$71,$89,$FD,$57,$C9,$BE
	.dc.b	$97,$F8,$BC,$FF,$0C,$8C,$57,$EB,$F1,$F0,$4F,$F1,$06,$E5,$FC,$7A
	.dc.b	$B7,$5F,$AA,$6D,$6F,$1F,$13,$71,$FD,$B4,$F0,$D3,$C1,$97,$F2,$87
	.dc.b	$58,$7E,$F7,$2E,$EF,$F4,$9A,$EB,$97,$86,$FD,$2F,$B6,$7F,$77,$5F
	.dc.b	$1E,$FA,$C7,$82,$BB,$56,$E7,$B9,$A9,$FE,$19,$7E,$59,$D1,$23,$C5
	.dc.b	$5C,$9A,$7F,$11,$BF,$04,$F8,$F1,$E8,$47,$07,$39,$5F,$77,$99,$F0
	.dc.b	$72,$FB,$58,$F9,$AB,$E0,$4F,$AC,$F0,$4E,$A3,$AC,$72,$BF,$03,$FD
	.dc.b	$C7,$0F,$96,$FF,$A6,$EB,$83,$E0,$BF,$84,$77,$78,$46,$9F,$3E,$3F
	.dc.b	$05,$A7,$8B,$28,$FE,$0E,$3E,$AB,$E3,$C7,$CA,$F8,$67,$D4,$FD,$09
	.dc.b	$E9,$87,$CA,$8F,$0C,$EB,$0F,$2D,$FC,$9A,$D5,$78,$8E,$47,$8A,$7F
	.dc.b	$17,$8F,$C2,$8D,$38,$AF,$A5,$97,$D1,$9E,$55,$F0,$5F,$31,$EB,$3F
	.dc.b	$53,$8F,$B1,$CD,$EE,$B9,$1C,$F0,$F2,$9E,$67,$35,$EE,$CB,$E6,$4E
	.dc.b	$63,$E6,$4F,$19,$C9,$FA,$AC,$7F,$03,$87,$8C,$FC,$FC,$65,$FF,$08
	.dc.b	$7E,$50,$F8,$1F,$F0,$9E,$9F,$E6,$28,$C4,$FF,$AE,$1A,$FF,$A6,$38
	.dc.b	$DE,$FF,$0E,$1F,$EF,$B5,$AB,$FF,$34,$77,$59,$FF,$32,$B7,$0F,$0E
	.dc.b	$3C,$5F,$F2,$07,$D3,$65,$34,$C7,$84,$F3,$78,$4E,$AA,$D7,$F8,$83
	.dc.b	$1F,$F4,$13,$71,$FF,$17,$BD,$7F,$48,$7D,$DF,$26,$E1,$FE,$51,$BF
	.dc.b	$F4,$54,$62,$3F,$94,$DF,$F8,$E3,$F0,$0F,$F4,$5C,$6E,$3F,$C7,$73
	.dc.b	$B8,$FF,$72,$35,$CD,$FF,$B9,$DA,$51,$1D,$EF,$FD,$35,$F5,$3F,$E4
	.dc.b	$C9,$EC,$7F,$99,$5C,$3E,$43,$FC,$D1,$3C,$6F,$FC,$5F,$F6,$FF,$EC
	.dc.b	$49,$FF,$78,$B5,$E1,$D4,$8C,$61,$FE,$86,$EF,$FA,$93,$9F,$3F,$EB
	.dc.b	$79,$4E,$27,$C6,$E1,$B9,$76,$AD,$F3,$7F,$5B,$CB,$FD,$5B,$F3,$AF
	.dc.b	$A8,$EA,$DB,$FF,$59,$F7,$70,$33,$32,$32,$7C,$21,$F5,$0F,$73,$F0
	.dc.b	$6D,$8D,$35,$F8,$C3,$8F,$F3,$2E,$A7,$E9,$3D,$1B,$3C,$63,$FC,$EE
	.dc.b	$64,$FB,$0F,$5C,$7F,$DA,$0F,$C5,$8F,$0E,$9E,$4A,$F9,$3E,$D3,$ED
	.dc.b	$2F,$C1,$4B,$FB,$BC,$2A,$D4,$61,$AF,$DF,$8D,$4F,$18,$C4,$EB,$2F
	.dc.b	$F6,$DF,$E9,$3E,$5E,$6F,$F3,$8B,$6B,$FD,$20,$DC,$1E,$DF,$FB,$4A
	.dc.b	$7F,$1F,$CA,$AD,$63,$FE,$F7,$9F,$27,$AF,$F8,$69,$E5,$6F,$83,$FD
	.dc.b	$E7,$B8,$FA,$5F,$C6,$E9,$FE,$BD,$8C,$74,$7E,$32,$F8,$65,$F0,$9A
	.dc.b	$FE,$76,$76,$FE,$89,$AF,$C3,$8D,$ED,$FE,$F7,$6D,$F7,$C3,$C3,$4B
	.dc.b	$E3,$7B,$C7,$C5,$CD,$F5,$BD,$39,$7E,$38,$CB,$F4,$E3,$2F,$86,$3F
	.dc.b	$4B,$F5,$FF,$EF,$98,$FB,$1F,$53,$F8,$03,$3E,$CC,$8C,$90,$FB,$93
	.dc.b	$C5,$FC,$BF,$43,$E5,$BF,$3A,$48,$FA,$B1,$FA,$97,$71,$F9,$82,$3F
	.dc.b	$96,$6B,$F8,$9F,$A9,$FB,$CD,$F0,$3A,$3F,$A6,$FB,$AF,$F2,$4B,$FA
	.dc.b	$F6,$B9,$F4,$BF,$9C,$6E,$FD,$07,$F2,$1C,$7F,$ED,$CF,$AB,$FE,$6D
	.dc.b	$7F,$C7,$5A,$FF,$21,$6B,$FA,$C6,$BF,$CC,$79,$3E,$BD,$FD,$6A,$C7
	.dc.b	$37,$F4,$0C,$BE,$ED,$87,$CB,$FF,$21,$71,$F9,$87,$91,$BD,$FF,$D2
	.dc.b	$27,$50,$9F,$D6,$BF,$BE,$6B,$FE,$F6,$6B,$C8,$FE,$71,$0F,$AB,$E8
	.dc.b	$48,$FE,$2D,$3E,$17,$B1,$FD,$61,$FE,$17,$97,$FB,$AE,$32,$7D,$CF
	.dc.b	$3B,$4F,$F1,$49,$FE,$2A,$E8,$FE,$94,$E0,$E7,$9F,$FB,$BB,$C6,$7C
	.dc.b	$5F,$F2,$E3,$9D,$7F,$29,$FB,$AF,$EB,$2F,$F2,$46,$E3,$E3,$6D,$FF
	.dc.b	$BE,$39,$9F,$13,$FD,$C2,$78,$DF,$F2,$46,$BF,$4C,$38,$7E,$B3,$5C
	.dc.b	$FC,$37,$F2,$4F,$AB,$91,$E5,$8E,$DC,$D2,$3D,$0F,$F8,$1F,$E7,$1F
	.dc.b	$75,$FC,$BE,$7D,$53,$FC,$EB,$0F,$CB,$C7,$96,$30,$47,$FB,$5D,$7F
	.dc.b	$86,$BE,$23,$A7,$B2,$7C,$13,$FD,$23,$E5,$FF,$69,$BF,$D7,$6E,$FE
	.dc.b	$F6,$CB,$0F,$2F,$D5,$FF,$85,$A7,$F7,$93,$FD,$29,$AF,$1F,$07,$FA
	.dc.b	$8C,$62,$3F,$38,$F4,$AF,$F1,$7F,$87,$E6,$3A,$57,$5E,$69,$7F,$D9
	.dc.b	$6F,$F5,$F9,$32,$FD,$36,$2F,$D2,$9F,$07,$C3,$0F,$F0,$A6,$1F,$B2
	.dc.b	$7C,$38,$F9,$08,$FC,$8F,$A9,$7D,$1F,$D3,$6C,$78,$67,$E0,$7C,$44
	.dc.b	$8F,$23,$F6,$7D,$1F,$C6,$C7,$BA,$3E,$09,$1F,$F1,$26,$08,$FF,$13
	.dc.b	$4B,$C8,$7C,$D7,$F6,$CF,$94,$98,$7F,$6A,$9C,$B6,$72,$FF,$66,$9F
	.dc.b	$E6,$F1,$FE,$4A,$FA,$6F,$E5,$DC,$FF,$C2,$FE,$29,$9F,$E6,$E7,$38
	.dc.b	$27,$2F,$F1,$18,$CC,$62,$1F,$4E,$FA,$8F,$D2,$4E,$6B,$77,$D7,$CE
	.dc.b	$FE,$5F,$39,$6C,$7E,$AE,$F9,$61,$E5,$8E,$18,$D8,$F1,$57,$06,$C6
	.dc.b	$5F,$1E,$3E,$2A,$E3,$1E,$B6,$E1,$E1,$AE,$18,$78,$2B,$47,$42,$8C
	.dc.b	$7D,$6E,$6B,$E1,$4F,$17,$A4,$70,$9D,$BC,$71,$F4,$3E,$63,$D1,$1A
	.dc.b	$24,$6A,$BA,$61,$EA,$AF,$F1,$AD,$7D,$66,$D7,$F0,$27,$D0,$7F,$8E
	.dc.b	$BE,$2F,$FD,$33,$CB,$FE,$71,$6C,$7F,$9D,$5A,$EF,$B3,$F7,$8F,$F9
	.dc.b	$E6,$FF,$96,$FC,$A7,$F9,$E5,$B1,$9F,$11,$39,$6C,$66,$3D,$B1,$8D
	.dc.b	$3F,$11,$F5,$0F,$A8,$FD,$C5,$FC,$7A,$D6,$BF,$93,$6C,$7E,$45,$AA
	.dc.b	$B9,$75,$F6,$C6,$4F,$37,$F0,$93,$A8,$C1,$31,$FC,$B5,$F4,$9D,$46
	.dc.b	$2B,$9B,$6B,$F0,$73,$AC,$3C,$D5,$A6,$CE,$B1,$F6,$DF,$12,$4E,$9F
	.dc.b	$87,$1A,$8C,$1F,$DF,$F9,$7F,$CC,$DF,$23,$F1,$BB,$7E,$3C,$7F,$9A
	.dc.b	$7A,$F1,$AE,$FD,$B5,$E3,$35,$EE,$BF,$CD,$D3,$98,$C4,$FF,$9B,$23
	.dc.b	$15,$FC,$82,$BF,$D2,$5F,$4F,$FE,$E9,$7F,$D7,$4D,$AF,$F5,$F3,$5F
	.dc.b	$F4,$4B,$5F,$EC,$CE,$F1,$FF,$01,$36,$7D,$1F,$6D,$E7,$A4,$E5,$FF
	.dc.b	$10,$3F,$E1,$19,$FF,$52,$7D,$3B,$E0,$38,$4E,$6B,$FD,$43,$39,$AF
	.dc.b	$4F,$9F,$FE,$11,$9F,$1B,$5E,$AD,$7C,$2D,$AE,$98,$FA,$23,$E2,$FA
	.dc.b	$1E,$07,$D1,$F4,$EF,$AE,$F5,$65,$FE,$36,$CB,$FD,$75,$39,$7D,$37
	.dc.b	$C2,$BC,$13,$9C,$BF,$DD,$0D,$7F,$D6,$6D,$7F,$C2,$0E,$F1,$FF,$16
	.dc.b	$36,$FF,$05,$F1,$DD,$5F,$0F,$D7,$79,$23,$CE,$D8,$C9,$FE,$F5,$94
	.dc.b	$79,$09,$19,$78,$BF,$EB,$7F,$8A,$B8,$7F,$BB,$BE,$85,$F5,$FF,$6F
	.dc.b	$E8,$9D,$DF,$93,$F8,$85,$6B,$1F,$F6,$F5,$69,$FF,$59,$C6,$70,$FF
	.dc.b	$69,$D7,$0C,$3E,$C3,$6B,$FD,$AF,$28,$C3,$C2,$B9,$6B,$FC,$3A,$72
	.dc.b	$FE,$AB,$95,$F2,$DE,$5A,$DE,$1F,$E6,$FB,$F3,$FD,$4F,$AF,$E5,$F4
	.dc.b	$D9,$F3,$BE,$4A,$34,$91,$E6,$AE,$6F,$03,$FD,$C1,$CB,$D1,$FF,$80
	.dc.b	$1F,$51,$FE,$00,$6F,$95,$E5,$F5,$7F,$4D,$39,$DB,$E8,$47,$84,$95
	.dc.b	$DF,$2F,$F1,$F3,$F1,$BB,$7F,$DE,$1F,$80,$73,$8F,$F7,$83,$6D,$C1
	.dc.b	$FE,$77,$FA,$8F,$D5,$47,$CE,$6B,$FE,$61,$0E,$1F,$F3,$23,$5F,$A9
	.dc.b	$87,$FB,$9E,$FC,$9F,$15,$F1,$C8,$FD,$AF,$A8,$7D,$47,$E4,$A3,$FD
	.dc.b	$41,$8F,$F4,$69,$CF,$4F,$F1,$6F,$98,$FE,$3D,$18,$AF,$F7,$DE,$CF
	.dc.b	$92,$FE,$19,$BF,$FA,$1A,$73,$18,$F2,$FF,$29,$5B,$9F,$AC,$DD,$BF
	.dc.b	$22,$D7,$FD,$01,$C9,$FE,$69,$BF,$C9,$CE,$F4,$FC,$73,$6F,$FD,$70
	.dc.b	$3F,$01,$FF,$13,$3E,$23,$DA,$7F,$39,$F2,$FF,$98,$FB,$3F,$CC,$9B
	.dc.b	$1A,$CB,$F9,$DF,$5F,$F7,$08,$C6,$A8,$FF,$2F,$C6,$1D,$4F,$DC,$8F
	.dc.b	$E8,$D7,$FD,$07,$E4,$5F,$03,$FD,$02,$B0,$DA,$EB,$3A,$7A,$4E,$9A
	.dc.b	$7D,$0E,$9E,$9F,$39,$F3,$1B,$19,$3E,$6B,$F2,$08,$E0,$D4,$F9,$CF
	.dc.b	$EB,$34,$F8,$AE,$93,$E4,$FA,$3F,$51,$EB,$F4,$2C,$7A,$1B,$7C,$F5
	.dc.b	$61,$FE,$03,$E1,$FE,$D6,$B0,$F8,$9F,$73,$F1,$7C,$3B,$97,$C2,$6E
	.dc.b	$3C,$EF,$F8,$3B,$E6,$AE,$15,$F5,$E1,$F4,$BF,$25,$F6,$1F,$E2,$8F
	.dc.b	$43,$FB,$04,$67,$2F,$F7,$37,$1F,$F7,$AB,$E5,$3C,$6F,$F6,$0D,$7F
	.dc.b	$DA,$4D,$47,$83,$87,$FA,$E4,$75,$9D,$C7,$F6,$FF,$D2,$DF,$07,$E2
	.dc.b	$8F,$F0,$27,$CE,$6F,$0F,$F0,$4A,$E4,$CC,$E9,$B8,$7E,$4A,$74,$D9
	.dc.b	$D3,$9E,$0E,$A7,$E4,$7D,$EB,$1F,$8E,$6E,$1F,$35,$BA,$FF,$91,$A1
	.dc.b	$F4,$FE,$8A,$F8,$BD,$5F,$E0,$98,$C4,$FC,$ED,$BF,$90,$BF,$42,$32
	.dc.b	$FF,$19,$6B,$FE,$FB,$6B,$F2,$9A,$F8,$35,$EC,$7F,$8A,$7F,$4C,$7D
	.dc.b	$A6,$31,$FC,$07,$D4,$FE,$2D,$BA,$FE,$BF,$EA,$7F,$C3,$D3,$9C,$7F
	.dc.b	$15,$18,$9D,$57,$03,$E6,$47,$D4,$3F,$71,$F3,$36,$3F,$6A,$DA,$FC
	.dc.b	$A3,$76,$FF,$0F,$7B,$BF,$84,$F8,$BF,$EB,$F5,$FD,$77,$BF,$FC,$E3
	.dc.b	$18,$8F,$F5,$C8,$71,$E1,$FF,$0C,$8C,$7A,$3F,$DC,$2F,$F1,$FF,$53
	.dc.b	$AC,$3F,$CF,$6F,$3A,$FF,$67,$07,$59,$F0,$BE,$53,$97,$F9,$6D,$72
	.dc.b	$73,$5C,$27,$47,$F5,$1B,$E4,$FA,$EE,$BE,$A9,$E0,$78,$DF,$F6,$E1
	.dc.b	$EA,$38,$14,$F8,$38,$F7,$38,$57,$4C,$7E,$15,$62,$BC,$67,$F8,$47
	.dc.b	$1F,$43,$E5,$3A,$4F,$C9,$BE,$07,$5C,$7D,$A7,$C0,$C7,$D3,$5F,$23
	.dc.b	$0F,$4C,$EA,$3C,$F1,$E1,$32,$F7,$6B,$F0,$63,$A3,$E6,$3E,$9D,$78
	.dc.b	$5F,$33,$5E,$07,$68,$F0,$3E,$DA,$F2,$BF,$2A,$BB,$BE,$79,$EF,$1E
	.dc.b	$E7,$C3,$56,$7E,$7B,$E9,$8D,$9D,$AF,$E6,$B7,$0F,$97,$5E,$43,$C4
	.dc.b	$7A,$CF,$B7,$5E,$97,$DB,$3E,$03,$96,$1E,$8C,$7E,$D5,$F4,$AF,$41
	.dc.b	$EF,$3E,$55,$79,$1A,$64,$F1,$61,$E6,$7E,$4C,$77,$69,$EE,$7C,$2F
	.dc.b	$9E,$BC,$37,$E3,$8F,$C2,$DF,$EC,$AB,$F4,$78,$7A,$E7,$DB,$1D,$5F
	.dc.b	$CA,$B5,$FD,$6F,$C3,$7E,$DD,$FC,$83,$5F,$DD,$B5,$DC,$FF,$46,$5F
	.dc.b	$06,$F5,$3F,$C8,$30,$AE,$53,$9D,$7F,$A6,$C6,$2B,$6D,$8C,$DF,$5A
	.dc.b	$E1,$7F,$AD,$FA,$87,$D4,$F8,$6B,$F8,$84,$62,$7F,$8B,$65,$C7,$2F
	.dc.b	$1D,$7F,$15,$C3,$9B,$C3,$0E,$55,$EB,$C3,$96,$1F,$BB,$BD,$CF,$3C
	.dc.b	$7F,$82,$D6,$B7,$E5,$7F,$C2,$2F,$CF,$1D,$F2,$F0,$E1,$D7,$1E,$F7
	.dc.b	$FD,$BB,$0E,$F9,$6F,$2F,$4C,$FF,$72,$CB,$9E,$1C,$63,$C3,$87,$D1
	.dc.b	$CB,$84,$7F,$0A,$AD,$BD,$67,$55,$C6,$35,$97,$D6,$BE,$27,$AA,$3F
	.dc.b	$89,$E5,$ED,$AF,$81,$1C,$31,$F2,$A9,$F4,$CF,$5C,$BC,$67,$47,$C1
	.dc.b	$7D,$4C,$57,$B7,$0F,$14,$E8,$91,$9C,$3E,$64,$FC,$A8,$F6,$1E,$2C
	.dc.b	$3C,$2E,$CF,$B1,$1D,$1D,$9E,$CC,$7E,$1B,$BC,$3C,$6F,$AE,$32,$D3
	.dc.b	$C3,$71,$FC,$94,$7D,$A7,$47,$82,$3E,$2E,$5F,$4A,$33,$5F,$07,$1F
	.dc.b	$96,$F8,$31,$F9,$D3,$E6,$8F,$1D,$71,$9C,$CF,$2A,$E2,$7D,$4C,$BE
	.dc.b	$C1,$1C,$1A,$8E,$8F,$CC,$7F,$6F,$97,$C5,$6E,$3F,$2F,$2F,$66,$1F
	.dc.b	$32,$72,$7B,$4F,$94,$7E,$48,$F8,$99,$7A,$E7,$E4,$CF,$A3,$2F,$84
	.dc.b	$7A,$E3,$27,$CE,$73,$87,$DD,$BF,$97,$97,$AD,$FB,$8F,$F8,$43,$C4
	.dc.b	$FF,$38,$FA,$DF,$C4,$73,$6F,$FE,$58,$97,$73,$D8,$D4,$E6,$31,$8F
	.dc.b	$FA,$53,$0C,$BF,$E5,$6A,$B7,$C9,$B5,$C2,$7F,$CF,$15,$AF,$17,$18
	.dc.b	$FF,$93,$7C,$17,$C6,$74,$BF,$F3,$03,$BC,$71,$3A,$9F,$F3,$F3,$5F
	.dc.b	$F3,$EB,$6B,$FD,$1B,$3C,$FE,$9F,$FD,$5F,$5F,$E8,$56,$BF,$E8,$46
	.dc.b	$CF,$C4,$86,$6E,$D7,$D2,$F3,$F1,$3F,$DE,$55,$FC,$EF,$0E,$6F,$FB
	.dc.b	$D5,$AE,$63,$6F,$FA,$1A,$3B,$EF,$FD,$0F,$E9,$DE,$9F,$C1,$6D,$79
	.dc.b	$EF,$E4,$71,$FC,$CF,$9F,$93,$FE,$C6,$8A,$FF,$AB,$F8,$D3,$8D,$BF
	.dc.b	$D5,$F5,$F5,$3D,$7F,$AB,$CB,$30,$FF,$45,$46,$5F,$F4,$64,$65,$E9
	.dc.b	$1F,$57,$E9,$BE,$CE,$BF,$D6,$DB,$3F,$EA,$26,$AF,$B8,$7C,$3F,$1E
	.dc.b	$4F,$87,$2A,$FC,$33,$FD,$94,$C4,$6A,$33,$1D,$E3,$F0,$DF,$34,$F5
	.dc.b	$E9,$D6,$7F,$DD,$3D,$9F,$8F,$FB,$99,$18,$21,$FE,$CB,$72,$FF,$B3
	.dc.b	$63,$F2,$DE,$6F,$E5,$F8,$64,$FF,$BD,$22,$BF,$ED,$38,$BC,$3F,$ED
	.dc.b	$C9,$FD,$2F,$8B,$ED,$6E,$86,$9F,$A3,$F8,$CB,$87,$E3,$2B,$83,$6B
	.dc.b	$83,$6B,$F1,$73,$A8,$FF,$72,$70,$FE,$B4,$E6,$FF,$B9,$FC,$4F,$D8
	.dc.b	$71,$C8,$FF,$6B,$E5,$FE,$F2,$9F,$DA,$19,$8F,$F7,$74,$E1,$DF,$1F
	.dc.b	$F0,$7F,$93,$FD,$1C,$60,$FD,$3B,$FE,$DC,$71,$DA,$8C,$43,$0F,$D7
	.dc.b	$4E,$7B,$3F,$84,$D7,$4A,$D5,$F7,$F3,$72,$6A,$BE,$6F,$26,$F1,$FD
	.dc.b	$C6,$A8,$C5,$FF,$BD,$23,$15,$FC,$62,$1F,$9E,$3E,$1B,$9C,$FC,$CB
	.dc.b	$F1,$65,$FE,$37,$C3,$8B,$F5,$63,$2F,$57,$BA,$3C,$5F,$25,$F9,$63
	.dc.b	$AB,$F1,$F2,$7D,$1B,$E8,$3F,$95,$C6,$1B,$19,$9F,$9D,$F4,$CF,$03
	.dc.b	$D1,$1F,$44,$7D,$17,$F9,$25,$FE,$1E,$D7,$F8,$BB,$5F,$D7,$35,$FF
	.dc.b	$2A,$7D,$27,$F2,$B7,$F6,$4D,$73,$E4,$7E,$E2,$31,$1F,$80,$EE,$FE
	.dc.b	$71,$F2,$1F,$C6,$23,$EC,$9F,$0D,$C6,$9F,$80,$E5,$FD,$C4,$FA,$67
	.dc.b	$83,$FC,$05,$AF,$F9,$9D,$AF,$3F,$1B,$F8,$4F,$D1,$FC,$58,$EE,$1C
	.dc.b	$5F,$F3,$B8,$C6,$FE,$6A,$FC,$4E,$72,$7F,$35,$C5,$F8,$98,$FC,$9F
	.dc.b	$C8,$E6,$7F,$8A,$4E,$5F,$C1,$4F,$E1,$C8,$FD,$3C,$7F,$8D,$63,$E1
	.dc.b	$6F,$3F,$E1,$EB,$F3,$3F,$45,$2F,$C3,$CA,$8F,$D1,$FD,$CE,$67,$C3
	.dc.b	$39,$7F,$25,$3F,$97,$23,$F8,$3F,$A1,$FE,$9A,$FE,$4D,$B9,$FE,$96
	.dc.b	$7F,$BC,$FC,$DB,$E8,$9F,$55,$86,$E1,$FB,$D9,$D4,$FE,$DB,$DD,$D3
	.dc.b	$FD,$72,$33,$3F,$D9,$1A,$FF,$3B,$6B,$FE,$A1,$73,$BF,$D5,$CB,$F2
	.dc.b	$CD,$F0,$BA,$D7,$87,$17,$D1,$F9,$49,$1F,$C3,$B9,$9F,$6B,$88,$F0
	.dc.b	$57,$07,$C8,$D7,$35,$E9,$8C,$61,$F8,$7F,$23,$F8,$BB,$7C,$3E,$8F
	.dc.b	$F8,$17,$47,$D2,$3D,$CF,$96,$DC,$BE,$43,$5F,$D1,$F8,$DF,$C8,$FE
	.dc.b	$47,$75,$F2,$6B,$55,$F1,$5A,$FF,$0B,$BE,$F3,$B3,$FC,$4D,$97,$F8
	.dc.b	$7F,$A3,$FD,$98,$DA,$57,$07,$ED,$63,$F8,$48,$FF,$0E,$F8,$3F,$82
	.dc.b	$8F,$C3,$9F,$CD,$F8,$F1,$1F,$76,$3F,$C3,$75,$D1,$B5,$CA,$3F,$C4
	.dc.b	$71,$FC,$FF,$A7,$F9,$DF,$C6,$C9,$FA,$D1,$FC,$F8,$C3,$F1,$38,$D3
	.dc.b	$98,$FE,$39,$F4,$0F,$1B,$F8,$E7,$C0,$35,$FF,$75,$4E,$79,$FF,$98
	.dc.b	$4E,$6F,$F9,$94,$E5,$46,$57,$D0,$BE,$23,$FD,$E9,$39,$AE,$0D,$C7
	.dc.b	$C3,$8F,$81,$B8,$7F,$3F,$C3,$93,$E2,$9D,$63,$63,$C7,$5C,$16,$1D
	.dc.b	$AB,$FA,$44,$79,$31,$F1,$C7,$28,$F4,$B7,$1F,$26,$1D,$67,$81,$6F
	.dc.b	$A3,$E0,$8C,$61,$E1,$C7,$CF,$3D,$0D,$14,$F2,$C7,$9E,$B9,$C7,$63
	.dc.b	$4E,$09,$1D,$71,$F4,$CF,$43,$FC,$79,$3F,$4D,$B3,$FD,$45,$EF,$7F
	.dc.b	$E6,$6F,$7F,$FD,$25,$BD,$F5,$7F,$37,$5F,$E7,$09,$CB,$63,$2F,$6F
	.dc.b	$92,$7C,$5F,$F9,$C3,$24,$62,$7F,$CF,$31,$89,$D4,$E6,$79,$47,$CB
	.dc.b	$8E,$1A,$FE,$27,$E8,$5F,$21,$FE,$87,$8F,$C4,$4E,$B8,$3F,$2A,$DC
	.dc.b	$3F,$26,$D5,$7C,$B9,$FB,$63,$27,$E6,$D6,$EB,$51,$89,$D6,$1E,$07
	.dc.b	$F0,$53,$A8,$C4,$EC,$8F,$A7,$1F,$46,$3E,$6B,$E6,$DA,$FC,$15,$F3
	.dc.b	$8C,$5F,$42,$3F,$98,$8D,$4F,$BA,$BF,$41,$E1,$BE,$6B,$FC,$5D,$9F
	.dc.b	$F3,$7F,$A3,$FF,$57,$77,$7F,$8E,$27,$38,$7F,$03,$9C,$D7,$F8,$E6
	.dc.b	$73,$3F,$E3,$36,$B9,$F4,$FF,$BF,$36,$7F,$A9,$7D,$B6,$A3,$F9,$1B
	.dc.b	$78,$CD,$25,$78,$27,$6D,$5F,$41,$FE,$9A,$BF,$EC,$95,$E8,$CB,$FA
	.dc.b	$22,$F9,$CE,$5D,$BF,$E6,$1F,$17,$A6,$7E,$09,$CF,$6F,$F5,$CC,$79
	.dc.b	$C7,$AB,$3F,$3B,$6B,$84,$FB,$23,$18,$F7,$8E,$27,$F9,$0F,$3F,$F2
	.dc.b	$63,$8B,$FF,$19,$DB,$E8,$3F,$B3,$CF,$F8,$83,$EF,$5F,$55,$E6,$9C
	.dc.b	$DF,$F8,$EB,$89,$5A,$BF,$F0,$25,$6A,$BF,$BE,$35,$FE,$B0,$D7,$FA
	.dc.b	$9B,$70,$F1,$79,$3D,$CF,$9D,$97,$96,$FA,$C7,$8F,$C2,$F8,$FF,$25
	.dc.b	$67,$FC,$00,$DC,$7F,$C1,$D8,$AB,$E4,$36,$BF,$C2,$8D,$9D,$CF,$1E
	.dc.b	$2F,$F2,$36,$3F,$6E,$FD,$33,$F8,$3D,$1F,$6F,$FB,$2B,$FF,$47,$EB
	.dc.b	$F9,$DB,$FF,$24,$CE,$5F,$F4,$84,$67,$1F,$95,$3F,$BA,$7F,$D1,$8D
	.dc.b	$7F,$D2,$CD,$57,$E0,$EB,$FC,$23,$6F,$BC,$7E,$57,$E8,$9F,$0F,$E7
	.dc.b	$C3,$C5,$87,$25,$F2,$5F,$90,$FB,$8D,$37,$1F,$4B,$DA,$77,$1F,$ED
	.dc.b	$36,$D7,$F9,$89,$AF,$AA,$79,$CF,$FA,$4F,$07,$C3,$E2,$BF,$15,$E2
	.dc.b	$79,$34,$3E,$6D,$9F,$6F,$FE,$FB,$C7,$FC,$E7,$39,$8C,$5F,$F9,$B1
	.dc.b	$AF,$F6,$8A,$E3,$EE,$7E,$A5,$B7,$FD,$B2,$7F,$B6,$FD,$1F,$EE,$5C
	.dc.b	$07,$0E,$DA,$F0,$F8,$AF,$DA,$77,$7F,$B6,$FD,$7F,$DC,$9F,$C1,$FE
	.dc.b	$85,$F9,$0B,$EB,$33,$97,$E1,$3E,$33,$E6,$11,$F9,$8F,$EF,$FF,$0C
	.dc.b	$7E,$1A,$EE,$FE,$21,$F8,$73,$E7,$3F,$8A,$56,$A7,$31,$FC,$57,$CB
	.dc.b	$7D,$A7,$F1,$7D,$DF,$77,$98,$C3,$74,$F8,$3C,$5F,$D9,$A7,$F8,$83
	.dc.b	$63,$37,$D1,$B3,$C6,$7F,$D5,$F2,$FA,$3C,$72,$FF,$BF,$3D,$B7,$D8
	.dc.b	$7F,$22,$6E,$3F,$EE,$59,$D6,$19,$A9,$39,$3F,$DD,$59,$FF,$A5,$23
	.dc.b	$94,$E8,$FE,$29,$F7,$3F,$84,$BF,$F6,$67,$B3,$FB,$78,$C6,$1C,$BC
	.dc.b	$EF,$F2,$53,$FE,$BE,$8C,$DF,$EB,$9B,$8B,$E6,$1F,$39,$CF,$B7,$A7
	.dc.b	$C3,$7F,$38,$F6,$6E,$5F,$E9,$16,$E5,$FE,$C6,$C9,$5F,$B8,$6D,$7E
	.dc.b	$43,$C2,$46,$2B,$85,$72,$D9,$7F,$0B,$37,$D4,$66,$3E,$2F,$0B,$EC
	.dc.b	$7E,$71,$FC,$5F,$B1,$F0,$D8,$AF,$99,$9B,$87,$FD,$7F,$79,$8F,$E7
	.dc.b	$D7,$C2,$B7,$5A,$7E,$47,$B3,$FD,$63,$C8,$F9,$CD,$8C,$92,$FF,$D3
	.dc.b	$FA,$7E,$FF,$75,$7F,$6B,$FA,$67,$D2,$FF,$6D,$FC,$01,$F0,$59,$8C
	.dc.b	$37,$2F,$CB,$D7,$06,$CF,$E6,$23,$19,$7E,$F1,$BC,$5F,$F0,$05,$B5
	.dc.b	$FE,$6E,$5F,$37,$D6,$BF,$17,$E8,$7F,$74,$F0,$3F,$14,$DB,$FC,$43
	.dc.b	$87,$E2,$ED,$FE,$89,$72,$7C,$EA,$7D,$B7,$EA,$56,$5F,$71,$1A,$F0
	.dc.b	$7F,$C1,$DD,$1F,$90,$BE,$57,$C6,$F8,$5E,$E7,$37,$C3,$B3,$FC,$2D
	.dc.b	$7B,$9F,$F4,$5F,$D4,$3E,$6B,$FC,$27,$E1,$3E,$37,$FC,$39,$39,$AF
	.dc.b	$BF,$39,$7D,$51,$9C,$BF,$93,$B5,$FE,$58,$D7,$BD,$FF,$0A,$8E,$19
	.dc.b	$FA,$1B,$5F,$EA,$EE,$CF,$0D,$7F,$05,$C2,$4F,$E5,$BB,$9D,$7F,$88
	.dc.b	$8F,$C9,$DF,$F0,$F6,$D7,$0C,$BF,$33,$D1,$73,$FE,$25,$F6,$7F,$C4
	.dc.b	$4F,$D2,$71,$DC,$39,$C5,$BF,$D0,$CD,$AF,$F1,$97,$C3,$7D,$6F,$9D
	.dc.b	$7D,$4F,$FA,$DB,$6F,$F8,$01,$96,$63,$FA,$5D,$5C,$7E,$ED,$69,$5F
	.dc.b	$27,$FA,$A4,$FE,$0B,$EA,$71,$1F,$E0,$0F,$90,$FC,$34,$62,$3F,$DB
	.dc.b	$A1,$DD,$F5,$75,$1F,$ED,$E3,$58,$79,$EB,$9B,$E6,$8C,$9E,$68,$CB
	.dc.b	$E4,$32,$78,$F1,$F3,$4F,$1A,$F0,$D7,$0A,$FF,$74,$9E,$E9,$F0,$57
	.dc.b	$C3,$8C,$E3,$C2,$31,$1E,$CB,$C3,$E0,$31,$5C,$9C,$1E,$DA,$D9,$F0
	.dc.b	$31,$ED,$8F,$CF,$8C,$4F,$AA,$3E,$35,$7A,$4E,$97,$A3,$C2,$57,$26
	.dc.b	$E7,$E3,$61,$F0,$9F,$8D,$3C,$27,$E6,$5F,$91,$A7,$A5,$E9,$3C,$8F
	.dc.b	$5E,$1C,$DF,$9C,$60,$F4,$63,$F3,$5F,$86,$71,$BF,$6D,$77,$8F,$B0
	.dc.b	$D7,$E9,$C7,$95,$F3,$DF,$B1,$A7,$37,$DC,$F8,$A3,$AC,$7A,$4F,$5C
	.dc.b	$7A,$67,$D6,$EE,$38,$E3,$F3,$A3,$A9,$92,$B9,$7C,$B1,$E3,$69,$E3
	.dc.b	$3D,$6F,$85,$A7,$85,$EF,$7F,$16,$3F,$41,$7E,$F7,$F1,$F3,$C5,$A7
	.dc.b	$13,$99,$F4,$2F,$DE,$E4,$FC,$DF,$6F,$EE,$1F,$D2,$35,$CC,$7E,$AE
	.dc.b	$31,$EB,$FF,$9C,$B4,$FE,$7B,$18,$F6,$FF,$A0,$DF,$F5,$EB,$FE,$82
	.dc.b	$FF,$98,$A3,$33,$BA,$E3,$5A,$BE,$97,$FA,$D3,$FA,$4F,$D4,$BE,$8F
	.dc.b	$9D,$7F,$0A,$8C,$4F,$EC,$6B,$8C,$FF,$0E,$6E,$5E,$2A,$DD,$75,$0E
	.dc.b	$2B,$86,$5E,$A6,$E3,$F9,$99,$E3,$9F,$87,$2F,$6B,$AC,$3F,$85,$E5
	.dc.b	$E0,$7D,$33,$E8,$CB,$C4,$D8,$FE,$CB,$3B,$7E,$36,$1D,$E7,$AD,$70
	.dc.b	$9F,$D5,$57,$C6,$BF,$01,$E3,$8E,$06,$B2,$FA,$71,$AA,$E5,$3F,$C8
	.dc.b	$31,$F5,$C7,$D1,$AD,$B7,$1F,$E3,$93,$F1,$63,$8E,$5E,$F3,$A5,$FB
	.dc.b	$8F,$09,$F0,$AF,$9B,$EE,$6B,$F4,$32,$F6,$BC,$4E,$18,$7A,$4F,$0C
	.dc.b	$7C,$78,$E0,$7D,$F9,$FA,$D7,$E7,$8F,$64,$F4,$76,$74,$26,$3F,$1D
	.dc.b	$D9,$DC,$D6,$19,$25,$75,$7D,$F1,$D2,$79,$18,$CB,$EA,$57,$CE,$8C
	.dc.b	$BE,$57,$8E,$5F,$42,$BE,$04,$71,$C7,$D7,$1F,$0F,$0F,$0B,$F2,$72
	.dc.b	$F8,$07,$CE,$AE,$B1,$93,$D3,$87,$D6,$23,$9A,$F0,$E5,$F1,$71,$F1
	.dc.b	$4F,$E4,$0F,$8B,$5E,$92,$3D,$6F,$8B,$E1,$8F,$61,$F1,$4E,$EF,$D0
	.dc.b	$BF,$C3,$47,$F9,$EB,$E7,$3F,$DB,$5C,$16,$BF,$C3,$F1,$85,$BB,$EE
	.dc.b	$BF,$D6,$11,$A2,$1F,$E4,$EA,$E4,$D3,$8C,$70,$71,$E3,$B9,$DF,$09
	.dc.b	$DE,$30,$D8,$CC,$FF,$8D,$6A,$BC,$A2,$EF,$FE,$C1,$AE,$07,$A7,$1F
	.dc.b	$43,$EB,$A7,$E2,$8E,$7C,$36,$3F,$CD,$4D,$7D,$43,$E9,$FF,$DB,$EE
	.dc.b	$A7,$FD,$B6,$F4,$6C,$F1,$8F,$E5,$E6,$4F,$F7,$0B,$D7,$0F,$F3,$A6
	.dc.b	$6F,$81,$7D,$2F,$C3,$6D,$F7,$3F,$CF,$DF,$77,$FE,$38,$BF,$F3,$AB
	.dc.b	$77,$FE,$7A,$D9,$FF,$3F,$BF,$1A,$33,$C3,$F0,$E3,$52,$69,$B7,$F0
	.dc.b	$3E,$A3,$EB,$3E,$63,$63,$FC,$D0,$D7,$FD,$63,$F0,$9A,$9F,$B9,$EB
	.dc.b	$7F,$A2,$B2,$F7,$B7,$2F,$F2,$BE,$1C,$67,$FD,$1F,$F5,$0F,$6B,$EC
	.dc.b	$37,$2F,$F4,$F3,$5F,$AB,$C4,$FA,$2D,$D7,$D4,$AD,$57,$D2,$6B,$FD
	.dc.b	$89,$AF,$FA,$91,$E9,$7F,$EA,$0F,$A9,$FB,$F3,$F8,$26,$C7,$9A,$BE
	.dc.b	$1E,$B2,$BE,$1B,$F9,$BF,$3B,$FA,$ED,$7F,$AC,$79,$DA,$2B,$F3,$DF
	.dc.b	$79,$30,$E2,$7F,$A4,$3E,$53,$F4,$44,$7F,$D2,$DF,$79,$FE,$F4,$74
	.dc.b	$43,$F4,$91,$FA,$18,$D4,$7E,$CF,$0F,$0F,$77,$DA,$24,$7F,$B3,$3E
	.dc.b	$62,$BF,$98,$9D,$BF,$9A,$3F,$07,$35,$A6,$63,$6F,$ED,$7E,$F7,$F7
	.dc.b	$47,$4D,$3E,$17,$DD,$FF,$B1,$67,$FD,$60,$D7,$CC,$D8,$FD,$2E,$DF
	.dc.b	$E2,$17,$8F,$9D,$FC,$17,$E2,$E5,$F2,$8F,$E1,$1D,$52,$3F,$D7,$2F
	.dc.b	$F0,$DF,$C8,$9E,$FE,$67,$F8,$54,$DC,$3F,$67,$5A,$9F,$F7,$67,$D2
	.dc.b	$FF,$17,$6F,$89,$FC,$2E,$31,$1F,$C1,$97,$C1,$E6,$7F,$82,$C6,$A7
	.dc.b	$F8,$1B,$5F,$F1,$9B,$5F,$F7,$AB,$D2,$FF,$DE,$2B,$D0,$F0,$1D,$67
	.dc.b	$FE,$00,$7C,$D3,$E4,$BF,$8F,$7C,$2B,$E0,$29,$FC,$86,$75,$5C,$3E
	.dc.b	$17,$F8,$CC,$7E,$1A,$3F,$8B,$B5,$FF,$23,$B5,$79,$1F,$CE,$7E,$A7
	.dc.b	$F9,$B1,$20,$7F,$41,$F8,$F7,$D7,$3E,$47,$F9,$97,$D5,$BE,$5F,$F9
	.dc.b	$4B,$6F,$F1,$8D,$7F,$20,$D7,$F9,$5F,$D6,$7F,$24,$7F,$28,$D7,$3E
	.dc.b	$7F,$F3,$48,$C4,$79,$38,$DF,$7B,$FA,$45,$E6,$66,$30,$DE,$0F,$E9
	.dc.b	$17,$DE,$BF,$33,$C1,$FD,$37,$B8,$D3,$FA,$77,$D6,$3E,$43,$19,$FE
	.dc.b	$96,$B6,$DA,$FD,$57,$84,$9C,$CF,$EC,$5B,$87,$F4,$6E,$AF,$E2,$3A
	.dc.b	$7C,$EE,$5F,$ED,$5F,$0F,$DB,$0F,$D4,$F4,$79,$48,$FC,$1A,$FF,$40
	.dc.b	$EB,$FD,$53,$1F,$D9,$D7,$8B,$23,$49,$7E,$97,$FB,$67,$B5,$FC,$DF
	.dc.b	$0F,$BF,$A7,$F9,$E3,$D2,$EF,$7F,$E0,$1F,$9D,$7D,$24,$8F,$10,$8C
	.dc.b	$C7,$F8,$3D,$B1,$A1,$F5,$7F,$C0,$70,$FC,$6C,$EA,$7F,$B6,$B5,$FF
	.dc.b	$4E,$B5,$DD,$78,$5D,$B7,$4F,$99,$F1,$3F,$E0,$F3,$C5,$19,$BE,$3F
	.dc.b	$1B,$82,$5F,$97,$17,$C1,$E2,$3A,$CF,$F2,$5C,$BE,$0B,$70,$34,$F9
	.dc.b	$75,$2B,$E2,$D7,$06,$FD,$EB,$E8,$7E,$33,$63,$31,$FE,$1D,$F3,$DF
	.dc.b	$53,$A8,$F1,$56,$EB,$E3,$B5,$F8,$ED,$7E,$6B,$BD,$BF,$21,$3F,$E3
	.dc.b	$33,$A4,$F9,$3E,$57,$C6,$47,$EC,$61,$FB,$9A,$FD,$6C,$7D,$FB,$57
	.dc.b	$90,$91,$FD,$37,$14,$EA,$3F,$A6,$7E,$57,$5F,$EA,$8D,$7F,$DB,$7C
	.dc.b	$48,$C4,$7E,$51,$AF,$F9,$69,$E5,$87,$EA,$FE,$73,$FC,$9C,$7F,$91
	.dc.b	$E3,$3C,$3F,$EE,$6F,$A8,$FE,$EF,$BF,$F1,$CF,$8C,$6D,$9F,$6B,$F0
	.dc.b	$39,$A7,$2F,$F2,$2B,$E2,$E6,$BF,$DE,$3F,$54,$FA,$AF,$F7,$9D,$F0
	.dc.b	$A5,$F0,$8C,$36,$FC,$18,$F8,$DE,$77,$C9,$CD,$F7,$3C,$58,$F9,$B0
	.dc.b	$E6,$D9,$E3,$3C,$CD,$DD,$7C,$73,$B7,$27,$51,$8F,$AD,$E1,$8F,$9D
	.dc.b	$DE,$1E,$0B,$E9,$5B,$9E,$55,$B7,$BC,$F2,$BF,$04,$76,$C7,$DC,$F4
	.dc.b	$C7,$DF,$8F,$58,$F2,$63,$D6,$31,$1F,$E7,$2C,$7E,$9B,$70,$FB,$2D
	.dc.b	$C7,$FC,$39,$F1,$BF,$C6,$38,$6F,$37,$F1,$7A,$FF,$11,$4E,$5B,$19
	.dc.b	$7B,$75,$FF,$41,$7F,$C5,$18,$FF,$2C,$F0,$BF,$C5,$0D,$8C,$CE,$DE
	.dc.b	$33,$F2,$A3,$96,$7F,$85,$FA,$AF,$EF,$DC,$19,$A9,$7F,$2A,$F7,$BF
	.dc.b	$F7,$28,$5F,$07,$55,$C3,$0F,$C6,$DF,$09,$FE,$F3,$5B,$9C,$D6,$EF
	.dc.b	$F4,$97,$BB,$EB,$E7,$FE,$FA,$F7,$1D,$88,$FA,$63,$55,$CA,$FA,$57
	.dc.b	$E9,$EF,$A6,$1E,$1A,$D4,$F2,$6E,$1C,$F1,$E3,$E1,$7F,$B2,$FE,$A7
	.dc.b	$F8,$7F,$37,$F9,$BF,$EA,$9E,$F6,$2F,$FB,$47,$8C,$BC,$67,$FE,$DE
	.dc.b	$BD,$D7,$F5,$6C,$7A,$3F,$D6,$3A,$93,$AA,$FF,$69,$5F,$29,$CE,$3D
	.dc.b	$25,$F5,$0F,$6F,$F9,$1C,$7F,$B1,$23,$18,$FF,$9A,$DB,$1F,$E6,$F6
	.dc.b	$BB,$9E,$91,$AD,$39,$FD,$F7,$92,$FF,$B4,$4E,$5F,$C1,$2F,$53,$DF
	.dc.b	$D0,$FA,$BF,$85,$5C,$27,$31,$8C,$BF,$DC,$37,$C6,$BC,$D5,$EA,$9E
	.dc.b	$AD,$7F,$C8,$33,$B7,$D5,$1E,$83,$E2,$60,$F9,$4C,$5F,$A3,$E3,$FC
	.dc.b	$A3,$EA,$FD,$EE,$1B,$1F,$1E,$3C,$18,$7C,$E6,$E2,$69,$C2,$3F,$DF
	.dc.b	$7F,$7C,$F9,$BC,$6D,$F5,$2F,$EE,$3F,$D6,$36,$E5,$EB,$FD,$0E,$1F
	.dc.b	$2B,$73,$FB,$11,$E0,$23,$83,$E9,$BF,$83,$C5,$F3,$1C,$73,$63,$F8
	.dc.b	$4F,$A2,$7C,$CF,$E2,$FE,$AF,$F7,$13,$98,$C7,$81,$F8,$DA,$DE,$7F
	.dc.b	$EB,$36,$E1,$FE,$B7,$6B,$FC,$0C,$38,$7F,$83,$3C,$2F,$B4,$FE,$7F
	.dc.b	$EF,$7F,$E0,$03,$FE,$33,$9D,$3C,$3E,$37,$CD,$F2,$BE,$36,$C6,$DF
	.dc.b	$AB,$3F,$ED,$F8,$F3,$46,$08,$EC,$FD,$87,$77,$FB,$ED,$C6,$3F,$19
	.dc.b	$FF,$3D,$7D,$67,$FA,$B9,$FF,$25,$E4,$8C,$4E,$CF,$F4,$9F,$D6,$7E
	.dc.b	$8D,$F9,$F1,$A9,$FD,$9C,$FE,$BA,$7F,$94,$35,$E3,$D9,$F8,$E8,$C7
	.dc.b	$6B,$57,$BF,$FB,$F9,$ED,$39,$8C,$57,$F0,$EF,$83,$7D,$63,$D1,$C3
	.dc.b	$75,$FE,$1D,$C8,$FA,$0C,$D7,$F0,$BF,$AB,$7C,$A7,$E8,$1B,$7F,$EF
	.dc.b	$36,$EF,$FE,$F7,$6B,$FD,$0F,$EA,$DF,$58,$FA,$CF,$E4,$31,$87,$F9
	.dc.b	$1F,$94,$E6,$FD,$B6,$3F,$E7,$8A,$D3,$FE,$75,$F3,$DF,$49,$FC,$7D
	.dc.b	$7C,$25,$8C,$36,$7F,$25,$9F,$FB,$1A,$B8,$C6,$23,$FD,$51,$5F,$BC
	.dc.b	$9C,$C7,$06,$F4,$7F,$2A,$6D,$7F,$5F,$FA,$97,$D7,$77,$BF,$C1,$69
	.dc.b	$FE,$F9,$AF,$C6,$FD,$6F,$BB,$5F,$F6,$BC,$60,$FC,$8D,$66,$74,$D8
	.dc.b	$FE,$89,$F2,$EF,$A9,$FE,$81,$F5,$BF,$8B,$9F,$F5,$A4,$E5,$FE,$91
	.dc.b	$19,$AF,$F7,$4C,$E5,$FC,$B7,$62,$75,$92,$B5,$B7,$F9,$AA,$B5,$F2
	.dc.b	$BF,$5A,$5F,$1D,$E2,$FA,$BC,$37,$3F,$3B,$73,$F8,$75,$C2,$7C,$B5
	.dc.b	$C1,$A7,$02,$CF,$9B,$A9,$F3,$3F,$27,$E1,$7F,$B0,$46,$2B,$F7,$31
	.dc.b	$89,$91,$F2,$3E,$56,$DF,$8B,$C9,$F9,$3E,$5F,$AD,$B5,$F3,$9B,$3B
	.dc.b	$8F,$97,$3F,$DB,$79,$BF,$D5,$71,$FE,$C7,$57,$9F,$89,$D3,$67,$F1
	.dc.b	$D3,$C2,$FF,$26,$D7,$F9,$33,$5F,$C6,$57,$17,$F8,$60,$FA,$BF,$80
	.dc.b	$E2,$7A,$71,$7F,$A8,$38,$FE,$D7,$C3,$7E,$1B,$E5,$FE,$C3,$72,$FC
	.dc.b	$5B,$70,$FA,$AD,$BF,$F0,$3F,$0F,$E0,$AB,$FC,$79,$F1,$BF,$AE,$7F
	.dc.b	$DC,$D1,$9B,$09,$D3,$E8,$8D,$26,$9F,$22,$79,$F4,$7F,$86,$7D,$FF
	.dc.b	$E0,$51,$FE,$F3,$D5,$F3,$79,$EC,$FF,$0F,$46,$27,$FD,$F1,$5C,$31
	.dc.b	$FD,$FB,$5F,$EA,$4D,$7F,$D4,$35,$FD,$EA,$38,$67,$FE,$26,$6D,$7B
	.dc.b	$36,$7E,$9A,$FA,$0F,$14,$FF,$8D,$7A,$3F,$B6,$BF,$D2,$BC,$1F,$CA
	.dc.b	$F3,$7E,$FD,$B1,$B8,$FD,$57,$8F,$8D,$7F,$3D,$C9,$FC,$17,$EA,$3F
	.dc.b	$3A,$DB,$EC,$E1,$FE,$07,$EA,$5A,$FC,$29,$FE,$C7,$1F,$40,$F9,$FF
	.dc.b	$F6,$BF,$C0,$7F,$B9,$23,$13,$F1,$0D,$CE,$AF,$ED,$CE,$A7,$FD,$8E
	.dc.b	$3E,$AF,$1E,$27,$F3,$68,$C5,$FD,$06,$9F,$84,$8D,$4F,$38,$F3,$3C
	.dc.b	$B0,$FF,$7B,$B9,$8F,$F6,$E1,$C8,$4F,$2A,$ED,$3C,$70,$FC,$8B,$C7
	.dc.b	$8E,$CF,$AE,$F9,$1F,$17,$1E,$0F,$C0,$3E,$13,$83,$93,$E1,$C3,$C0
	.dc.b	$FC,$7B,$E2,$6A,$78,$57,$C1,$6C,$EE,$3C,$53,$D0,$E2,$FA,$8C,$96
	.dc.b	$3E,$46,$3F,$12,$31,$1F,$0A,$30,$F5,$37,$8F,$C0,$7D,$73,$F0,$6F
	.dc.b	$CC,$D3,$9E,$1E,$C9,$FB,$B8,$79,$1E,$87,$7B,$E8,$EA,$BC,$EF,$63
	.dc.b	$C2,$7C,$B3,$F0,$F7,$F0,$9E,$06,$AB,$E1,$4F,$28,$C9,$DC,$FA,$E7
	.dc.b	$E0,$A3,$C8,$F4,$8F,$53,$E1,$38,$4F,$47,$B4,$7B,$B0,$F8,$F1,$DB
	.dc.b	$0F,$80,$FE,$4A,$7C,$4D,$38,$9F,$0A,$36,$FE,$52,$BD,$8F,$E2,$E7
	.dc.b	$C6,$F5,$AE,$EF,$33,$D9,$7F,$49,$F8,$F5,$5F,$4D,$7E,$24,$E6,$F5
	.dc.b	$36,$FE,$B7,$E3,$7F,$7B,$1F,$B9,$6B,$FB,$66,$BF,$C0,$DC,$FC,$6F
	.dc.b	$F9,$FB,$9B,$F9,$BC,$63,$DB,$FF,$41,$D2,$31,$3A,$C3,$FD,$F9,$7C
	.dc.b	$6B,$51,$8B,$E2,$FF,$07,$F9,$67,$D6,$7F,$44,$AF,$C2,$46,$1F,$E1
	.dc.b	$B9,$71,$C6,$C7,$F4,$AC,$7B,$61,$CF,$0E,$4D,$9D,$89,$DD,$F3,$8F
	.dc.b	$E1,$18,$77,$6D,$F8,$F3,$FE,$39,$7E,$B6,$E5,$EA,$C3,$D1,$87,$1C
	.dc.b	$7C,$18,$FE,$77,$0E,$EF,$9E,$FE,$4E,$1A,$26,$1E,$37,$AE,$5F,$AA
	.dc.b	$C7,$C9,$1A,$BE,$53,$FA,$EC,$3C,$97,$E2,$CB,$D0,$DA,$D6,$3E,$08
	.dc.b	$F8,$19,$79,$88,$E0,$F1,$1C,$67,$E1,$9C,$B0,$8E,$09,$8F,$A1,$FE
	.dc.b	$65,$8F,$C0,$59,$7C,$5B,$E9,$86,$CE,$17,$ED,$AD,$E5,$EA,$7D,$E7
	.dc.b	$4C,$BD,$4F,$F8,$2B,$0F,$8B,$87,$9E,$F9,$BE,$B9,$D5,$FE,$12,$3C
	.dc.b	$26,$63,$F0,$97,$D1,$F0,$9F,$46,$7C,$4E,$B2,$F8,$37,$E7,$C3,$C3
	.dc.b	$3E,$68,$F0,$BE,$03,$91,$C3,$0F,$04,$FC,$38,$D1,$F8,$7C,$3F,$18
	.dc.b	$F3,$9F,$61,$1C,$1F,$13,$1F,$B2,$F2,$7E,$64,$7D,$A8,$EB,$7F,$37
	.dc.b	$2F,$C3,$C7,$23,$DE,$7D,$A9,$E0,$FB,$CC,$1C,$9C,$C7,$DD,$C3,$D5
	.dc.b	$3C,$AF,$C2,$FF,$72,$FA,$BF,$F5,$8B,$FE,$40,$CB,$FD,$0D,$18,$8F
	.dc.b	$F4,$6B,$63,$FC,$A9,$E4,$3E,$63,$F8,$24,$70,$E8,$BC,$5F,$E2,$FE
	.dc.b	$D7,$D6,$63,$27,$77,$FB,$16,$31,$CA,$D3,$B6,$1F,$E0,$F5,$F0,$FA
	.dc.b	$6E,$1E,$79,$D4,$F5,$C3,$FD,$41,$F2,$5B,$8C,$65,$C2,$FF,$D7,$6D
	.dc.b	$D3,$DF,$3E,$59,$7D,$4F,$FB,$49,$EC,$7F,$99,$1C,$3F,$E7,$A6,$CF
	.dc.b	$F9,$DF,$E7,$D9,$73,$9D,$D7,$F2,$93,$AF,$AB,$FE,$D7,$69,$FE,$D9
	.dc.b	$8D,$BF,$CB,$DE,$B3,$FC,$C7,$EB,$1D,$FE,$76,$D7,$98,$E6,$7F,$A3
	.dc.b	$BE,$B3,$DD,$5F,$CE,$5B,$1F,$E7,$E6,$BF,$ED,$3F,$89,$F8,$51,$83
	.dc.b	$FC,$91,$9F,$B2,$78,$74,$EB,$5F,$A1,$CC,$F8,$6F,$B7,$F2,$FF,$85
	.dc.b	$BF,$93,$CA,$8C,$47,$FB,$2A,$78,$9C,$3A,$24,$7F,$B9,$A3,$E5,$7D
	.dc.b	$5C,$8F,$77,$0A,$3E,$C7,$DB,$7F,$BE,$27,$31,$F3,$67,$51,$F3,$63
	.dc.b	$04,$7E,$47,$D6,$3C,$7F,$C1,$36,$39,$9F,$EF,$8F,$0B,$DB,$8C,$8C
	.dc.b	$10,$FF,$79,$E7,$F0,$DF,$82,$D7,$FD,$D3,$3F,$A0,$FA,$BF,$F7,$8C
	.dc.b	$7C,$66,$BF,$EF,$C6,$F0,$FE,$C3,$CA,$E1,$EC,$FE,$73,$E3,$A5,$F1
	.dc.b	$8F,$A7,$F3,$1F,$4B,$C3,$93,$F9,$AF,$01,$CA,$AD,$CF,$9A,$B6,$FF
	.dc.b	$BF,$A3,$2F,$B2,$33,$97,$F0,$29,$FC,$0C,$7E,$EB,$E8,$BF,$D7,$B5
	.dc.b	$F8,$06,$AA,$FE,$13,$3C,$23,$5E,$97,$13,$0E,$FB,$3F,$D7,$1F,$54
	.dc.b	$F7,$3F,$77,$D9,$2B,$D9,$39,$6C,$E5,$FD,$FC,$FF,$05,$9D,$47,$FB
	.dc.b	$CB,$EA,$9E,$FF,$F0,$16,$C7,$EA,$DA,$FF,$BE,$39,$DE,$36,$EB,$F7
	.dc.b	$55,$AA,$FD,$F3,$5F,$F1,$43,$5F,$F7,$6B,$D3,$5F,$D6,$42,$1E,$37
	.dc.b	$8C,$ED,$1F,$BD,$FA,$67,$BF,$FC,$5D,$B1,$98,$F7,$6F,$FC,$6F,$E7
	.dc.b	$9F,$01,$FC,$3D,$B5,$FC,$4C,$FF,$7F,$B8,$70,$D7,$FC,$81,$F1,$AF
	.dc.b	$AA,$7D,$57,$F3,$B8,$C3,$63,$33,$E7,$EA,$FE,$59,$F5,$3F,$E5,$38
	.dc.b	$FE,$2D,$F5,$4F,$2F,$F9,$83,$63,$F8,$D4,$ED,$FC,$10,$11,$B7,$D4
	.dc.b	$D7,$F9,$73,$5F,$F2,$C3,$5E,$50,$F9,$3F,$29,$23,$CB,$5F,$D0,$FE
	.dc.b	$A9,$F1,$94,$93,$F9,$0E,$5A,$FC,$B9,$DB,$F3,$43,$87,$53,$FD,$06
	.dc.b	$31,$CB,$FC,$B7,$CC,$77,$39,$7F,$58,$76,$9F,$C3,$D7,$F4,$B6,$E1
	.dc.b	$E7,$3F,$B5,$F3,$FC,$58,$CC,$78,$E7,$FA,$83,$E4,$3C,$44,$3F,$71
	.dc.b	$C7,$FB,$B3,$A3,$FD,$37,$EA,$BF,$D2,$B5,$F9,$86,$F7,$7F,$A1,$5B
	.dc.b	$2F,$9B,$7D,$93,$EC,$FF,$B7,$57,$FA,$72,$B8,$78,$5F,$D8,$E3,$11
	.dc.b	$F7,$DA,$9E,$B7,$FB,$2E,$EF,$F0,$63,$67,$CD,$5D,$BD,$E9,$1E,$97
	.dc.b	$78,$FE,$BA,$30,$7D,$7C,$3F,$20,$FF,$77,$6B,$ED,$9F,$31,$E0,$C1
	.dc.b	$F1,$94,$EF,$3F,$98,$3F,$97,$7D,$67,$F8,$37,$C6,$34,$F4,$FD,$9B
	.dc.b	$E9,$F1,$5F,$89,$AD,$B6,$BF,$C1,$11,$88,$FD,$D3,$78,$3F,$BB,$6C
	.dc.b	$FA,$C7,$D6,$FC,$66,$DF,$F8,$1E,$F8,$71,$7F,$AF,$5F,$C4,$37,$AF
	.dc.b	$D0,$79,$BE,$16,$D7,$96,$3C,$64,$7F,$86,$79,$AF,$2B,$F0,$11,$FE
	.dc.b	$02,$FA,$D7,$C9,$FF,$90,$1B,$1F,$CE,$BA,$3F,$9B,$19,$99,$19,$3F
	.dc.b	$C3,$DF,$59,$FE,$0B,$E6,$FE,$09,$F1,$3F,$D6,$E7,$F8,$34,$E5,$46
	.dc.b	$54,$66,$31,$F5,$AF,$94,$FF,$6F,$46,$E7,$5C,$1F,$EE,$16,$DF,$6F
	.dc.b	$63,$F8,$E5,$72,$BE,$37,$B9,$CB,$71,$F3,$B6,$FC,$AF,$FB,$B7,$1F
	.dc.b	$14,$68,$ED,$35,$F4,$D7,$47,$27,$23,$F9,$34,$F2,$70,$70,$9E,$58
	.dc.b	$FA,$A3,$06,$A7,$86,$3F,$06,$FB,$4E,$A3,$17,$E3,$AD,$E1,$EB,$9C
	.dc.b	$D7,$83,$0E,$D5,$A8,$F3,$9F,$E6,$38,$FA,$CD,$C3,$F5,$EF,$90,$FF
	.dc.b	$34,$7C,$4F,$F7,$0E,$5F,$E5,$8D,$8F,$EB,$4D,$77,$D7,$FC,$45,$FE
	.dc.b	$BD,$6E,$FF,$E6,$51,$F6,$DA,$FE,$21,$AF,$E1,$9B,$1C,$7C,$5F,$93
	.dc.b	$1F,$CF,$B3,$FC,$67,$D6,$FF,$74,$E7,$CD,$4C,$3E,$15,$EF,$6F,$EB
	.dc.b	$50,$BE,$0E,$AB,$86,$1F,$91,$BE,$13,$FD,$86,$B7,$39,$AD,$DF,$AD
	.dc.b	$FC,$34,$EA,$31,$5D,$49,$5E,$8F,$47,$FC,$F8,$F3,$8C,$CF,$08,$D6
	.dc.b	$3D,$A7,$E3,$61,$5F,$55,$F2,$BE,$07,$FB,$07,$EB,$7F,$D1,$1D,$9F
	.dc.b	$E8,$9E,$8F,$F0,$EC,$63,$1F,$F6,$37,$75,$C7,$FD,$8B,$E9,$AB,$17
	.dc.b	$C2,$FF,$CB,$53,$96,$DE,$A3,$17,$FC,$C7,$EA,$1F,$47,$D6,$11,$8B
	.dc.b	$FF,$54,$7D,$5F,$F3,$17,$FD,$55,$19,$9F,$E7,$53,$97,$A5,$73,$8C
	.dc.b	$67,$FE,$3A,$AE,$0D,$FA,$CF,$04,$1E,$1E,$06,$19,$8F,$81,$CB,$4F
	.dc.b	$2C,$6B,$1F,$1B,$6F,$C1,$F0,$DD,$8F,$77,$D5,$3E,$AF,$18,$7B,$26
	.dc.b	$E1,$FD,$56,$B5,$39,$3F,$D9,$4E,$23,$E5,$4F,$4A,$E3,$7E,$FF,$7F
	.dc.b	$DA,$78,$3E,$5A,$E3,$F2,$A2,$E1,$DE,$6D,$7F,$98,$31,$46,$2F,$F4
	.dc.b	$B1,$87,$91,$FE,$DD,$FA,$B7,$CA,$FD,$B6,$C7,$FB,$3B,$EA,$9F,$19
	.dc.b	$9A,$93,$C5,$FC,$77,$AC,$AD,$CE,$4E,$55,$C3,$3F,$F2,$8D,$6E,$31
	.dc.b	$9F,$F9,$E3,$1F,$F7,$05,$F9,$7E,$23,$F2,$5F,$15,$E5,$64,$64,$FC
	.dc.b	$CC,$63,$1F,$F2,$EF,$47,$FA,$02,$31,$87,$E9,$F9,$FF,$DB,$D7,$C6
	.dc.b	$FF,$C1,$9F,$55,$F9,$57,$F8,$58,$70,$FF,$0E,$6B,$FE,$69,$6B,$BE
	.dc.b	$8F,$D2,$47,$E6,$27,$FC,$DD,$AF,$C5,$F6,$F3,$5F,$23,$7F,$F3,$45
	.dc.b	$7C,$F9,$FD,$C7,$C0,$BE,$FB,$F1,$1A,$7F,$96,$EF,$79,$FC,$AF,$AA
	.dc.b	$7C,$57,$D2,$6E,$5F,$E7,$DB,$FF,$51,$FD,$4F,$FB,$CE,$30,$79,$27
	.dc.b	$77,$FE,$2D,$9D,$37,$E3,$BF,$85,$7D,$3B,$EA,$5F,$53,$FC,$06,$F7
	.dc.b	$97,$CD,$EF,$DD,$6B,$3F,$E1,$B1,$8B,$FE,$11,$E3,$BE,$6E,$9F,$C2
	.dc.b	$FE,$99,$F5,$38,$C3,$CF,$B7,$FA,$5E,$F8,$46,$23,$FD,$17,$7C,$6B
	.dc.b	$84,$7F,$91,$EF,$9D,$6E,$F8,$CE,$6B,$79,$FE,$16,$3F,$90,$1F,$C8
	.dc.b	$66,$D7,$F2,$68,$A7,$E3,$BA,$3F,$D3,$71,$93,$CF,$D4,$F9,$27,$D2
	.dc.b	$FF,$30,$6C,$E9,$F2,$DF,$E6,$1C,$B6,$78,$1C,$72,$B5,$C7,$E3,$3A
	.dc.b	$1F,$CF,$FE,$D9,$F9,$CF,$F9,$92,$74,$4C,$3B,$56,$E7,$51,$89,$D6
	.dc.b	$BF,$D3,$72,$5E,$FE,$0B,$BD,$7A,$7E,$47,$F8,$0F,$4B,$E5,$7D,$D1
	.dc.b	$86,$CF,$DA,$F0,$3F,$2F,$18,$CF,$F8,$1D,$F1,$AF,$89,$F7,$BF,$1F
	.dc.b	$D3,$F9,$0D,$8C,$9F,$26,$32,$D3,$F0,$9C,$BF,$E8,$A6,$FC,$3F,$F5
	.dc.b	$7F,$A8,$ED,$8F,$E1,$63,$F2,$04,$9F,$C6,$4E,$B8,$11,$FE,$0E,$BE
	.dc.b	$97,$FE,$93,$AD,$4E,$6F,$B6,$7C,$EB,$EE,$7C,$07,$E0,$4F,$C1,$7D
	.dc.b	$ED,$BF,$C1,$45,$9F,$C1,$4D,$CB,$C3,$1F,$EF,$7F,$AA,$70,$7F,$71
	.dc.b	$7F,$0F,$3A,$94,$6B,$0F,$D3,$3A,$7F,$C4,$EE,$63,$FD,$58,$E5,$ED
	.dc.b	$5D,$79,$FF,$BF,$7B,$2E,$AA,$7F,$82,$BE,$A9,$F5,$58,$CB,$F6,$95
	.dc.b	$B6,$D6,$E7,$58,$FF,$36,$BE,$35,$A3,$F5,$2E,$23,$FD,$81,$7C,$DF
	.dc.b	$D3,$78,$9F,$90,$F4,$BF,$58,$DB,$FD,$63,$67,$71,$F9,$1E,$6F,$EC
	.dc.b	$51,$FA,$FF,$A9,$FC,$CF,$D5,$7E,$D5,$B5,$FB,$56,$CE,$E3,$94,$FF
	.dc.b	$88,$63,$FD,$4E,$3E,$A9,$F3,$DF,$E5,$4E,$B7,$8D,$F6,$A7,$35,$FD
	.dc.b	$3B,$53,$C6,$F8,$56,$DF,$EA,$71,$FE,$B0,$34,$7F,$4C,$8D,$3C,$CC
	.dc.b	$8B,$F0,$E3,$DB,$0F,$0C,$FF,$AD,$0D,$CE,$9E,$27,$B2,$37,$19,$78
	.dc.b	$C6,$6F,$CF,$5D,$5F,$3E,$16,$7C,$D3,$5F,$05,$74,$C3,$D3,$18,$BE
	.dc.b	$6F,$88,$F8,$13,$BA,$E5,$1E,$13,$A9,$BB,$ED,$87,$9E,$FC,$73,$EE
	.dc.b	$C3,$BB,$C4,$F8,$CE,$1E,$F1,$E2,$C3,$DF,$18,$3B,$B8,$BE,$F5,$C5
	.dc.b	$B8,$79,$AB,$81,$D9,$C4,$6A,$3D,$07,$96,$32,$D3,$35,$F0,$5B,$7E
	.dc.b	$53,$DB,$19,$3C,$91,$C5,$A6,$63,$67,$18,$FC,$44,$65,$B1,$97,$EA
	.dc.b	$1F,$01,$E2,$D3,$65,$36,$7B,$CF,$39,$93,$E8,$61,$F5,$5E,$26,$E7
	.dc.b	$CF,$3E,$18,$DE,$1F,$35,$F6,$47,$89,$E0,$DA,$E9,$1E,$26,$BC,$AF
	.dc.b	$E6,$DF,$BE,$35,$7D,$63,$11,$FB,$CF,$88,$7C,$9F,$E2,$9B,$1F,$AF
	.dc.b	$6B,$FA,$B6,$FC,$37,$F9,$B3,$9F,$F7,$B1,$8F,$61,$18,$BF,$F6,$CD
	.dc.b	$6E,$D7,$BA,$D4,$7F,$06,$9D,$61,$FB,$AF,$AA,$7D,$0F,$FB,$8A,$30
	.dc.b	$DD,$BA,$E7,$E0,$C3,$F8,$2E,$1C,$A7,$FD,$D3,$87,$29,$FD,$B3,$5D
	.dc.b	$D7,$1C,$3F,$84,$DF,$19,$FE,$33,$5B,$9F,$0C,$7E,$FE,$74,$9B,$3B
	.dc.b	$BF,$0E,$1E,$7A,$F7,$61,$DE,$7D,$38,$F9,$2B,$54,$BE,$4F,$09,$F0
	.dc.b	$5F,$18,$F2,$D7,$0A,$DC,$78,$E7,$4D,$AE,$CC,$3E,$0B,$D0,$C6,$3E
	.dc.b	$E8,$C9,$1C,$13,$0F,$4C,$65,$F7,$BC,$9C,$C7,$CC,$AF,$1D,$F9,$8F
	.dc.b	$3E,$3F,$28,$F9,$38,$FA,$CF,$86,$E6,$78,$C6,$70,$F4,$1E,$13,$E1
	.dc.b	$9E,$13,$E4,$19,$3E,$63,$C5,$C4,$F8,$1E,$07,$C7,$38,$57,$9E,$7C
	.dc.b	$75,$F3,$2B,$85,$7F,$54,$AF,$5C,$E4,$F3,$C7,$49,$F5,$63,$DA,$B6
	.dc.b	$FA,$B0,$F9,$95,$EA,$3C,$B7,$C9,$ED,$87,$8A,$B8,$63,$F4,$27,$26
	.dc.b	$0F,$83,$7D,$1F,$82,$67,$C9,$FF,$3D,$7B,$5F,$DD,$A3,$FD,$B3,$F5
	.dc.b	$3F,$E9,$97,$FC,$B8,$D7,$FC,$F5,$1B,$C3,$FC,$D4,$DE,$0F,$F6,$1D
	.dc.b	$6E,$7F,$C7,$BC,$28,$C4,$FF,$81,$30,$F0,$E1,$FE,$87,$F8,$9E,$4F
	.dc.b	$F7,$DA,$7C,$06,$30,$E4,$EA,$FF,$D2,$57,$7C,$A3,$BB,$E3,$87,$AA
	.dc.b	$F5,$3C,$7E,$A7,$FC,$99,$9F,$F9,$37,$E9,$1F,$53,$E2,$C1,$18,$7F
	.dc.b	$D4,$35,$C3,$5F,$91,$39,$9F,$E4,$31,$8A,$E5,$39,$BF,$F2,$FE,$8C
	.dc.b	$39,$61,$FE,$26,$AD,$61,$C6,$3F,$D8,$51,$85,$5C,$67,$CD,$E0,$BE
	.dc.b	$53,$D1,$F4,$FF,$02,$FC,$67,$F8,$C9,$C6,$5F,$E6,$BA,$E1,$3F,$EC
	.dc.b	$F6,$BC,$04,$F0,$BF,$C3,$36,$3F,$C7,$B3,$C2,$BF,$D5,$9F,$03,$F4
	.dc.b	$C9,$3F,$EC,$3E,$6A,$FF,$B0,$E2,$CF,$FA,$CE,$7E,$87,$D3,$FF,$B9
	.dc.b	$AB,$E0,$FC,$17,$CA,$33,$EA,$D8,$C9,$0F,$44,$F1,$7D,$5B,$2B,$84
	.dc.b	$61,$E6,$FF,$B5,$FC,$DF,$C5,$75,$48,$C1,$0F,$F6,$D3,$97,$FD,$B9
	.dc.b	$1F,$91,$F0,$FF,$2D,$F4,$FF,$EC,$C9,$FF,$62,$35,$FF,$71,$46,$DC
	.dc.b	$FA,$9C,$0C,$3F,$EE,$7F,$8D,$FC,$E1,$1F,$F3,$5E,$DF,$EA,$8F,$5D
	.dc.b	$F7,$7F,$A0,$9F,$F7,$85,$6D,$E4,$70,$E4,$FF,$39,$D7,$0E,$9D,$61
	.dc.b	$FE,$E3,$BB,$5A,$C3,$A6,$1F,$EF,$9F,$13,$F5,$C4,$7F,$DF,$DF,$07
	.dc.b	$5D,$1F,$F0,$06,$3F,$69,$1F,$B7,$F9,$8F,$F7,$54,$F8,$CF,$CD,$B8
	.dc.b	$7F,$2F,$D4,$8C,$65,$FE,$F7,$F1,$B9,$7A,$92,$F9,$5E,$E3,$F8,$14
	.dc.b	$FE,$E7,$EA,$3D,$37,$FE,$C9,$FA,$07,$D1,$FE,$F5,$B8,$F8,$C5,$70
	.dc.b	$DB,$F9,$24,$EA,$7F,$48,$DB,$E9,$7D,$F8,$3A,$32,$BA,$56,$EF,$F9
	.dc.b	$76,$1D,$67,$E0,$FD,$47,$F0,$FF,$A3,$7B,$94,$F9,$11,$FC,$4F,$E5
	.dc.b	$BF,$DC,$B5,$E7,$6B,$FB,$76,$E9,$FE,$E7,$BE,$73,$ED,$7F,$25,$19
	.dc.b	$E6,$F0,$1D,$7D,$8F,$E4,$1F,$4B,$FE,$36,$BF,$88,$D7,$F8,$23,$6B
	.dc.b	$FD,$F2,$3E,$ED,$F2,$7F,$CA,$7E,$9F,$F8,$26,$FF,$D3,$5B,$87,$FB
	.dc.b	$11,$AB,$0E,$93,$BF,$23,$FD,$8B,$3B,$6A,$AF,$E2,$EF,$F5,$93,$11
	.dc.b	$C3,$9B,$F1,$5C,$7F,$CE,$FE,$99,$F0,$D4,$FC,$A4,$5B,$FC,$14,$E9
	.dc.b	$B3,$A6,$90,$FE,$4D,$1B,$7F,$33,$D3,$F7,$DF,$F6,$B7,$37,$BA,$7C
	.dc.b	$DB,$B4,$FE,$E1,$18,$7F,$DC,$FE,$82,$B8,$5F,$A4,$87,$EA,$63,$5A
	.dc.b	$62,$3C,$65,$C3,$FB,$8D,$7D,$0F,$9C,$7D,$33,$E8,$7F,$AF,$7D,$37
	.dc.b	$F9,$FA,$FC,$98,$7F,$BE,$23,$19,$FF,$BE,$B0,$E3,$7C,$3C,$09,$7F
	.dc.b	$A9,$BD,$AA,$DB,$99,$DB,$FD,$36,$77,$3A,$8F,$EA,$EF,$9B,$E0,$BC
	.dc.b	$BD,$F2,$7F,$84,$4E,$5F,$04,$65,$E8,$FF,$82,$A3,$51,$83,$64,$3B
	.dc.b	$E6,$AE,$1F,$4C,$F9,$DF,$F0,$17,$8A,$F8,$BA,$6A,$2E,$1F,$C2,$67
	.dc.b	$4D,$9D,$34,$87,$F6,$D8,$DB,$F0,$2B,$C0,$FB,$97,$BD,$E2,$E2,$93
	.dc.b	$98,$AF,$38,$BC,$3F,$6B,$45,$F2,$AF,$E4,$FC,$BF,$81,$6F,$0F,$F3
	.dc.b	$53,$F9,$B7,$DD,$D8,$C9,$0F,$CE,$BF,$E2,$5F,$A1,$7D,$17,$F0,$7F
	.dc.b	$82,$6D,$FE,$4B,$9C,$F1,$B0,$56,$9C,$4E,$9C,$D7,$F9,$4F,$E8,$DF
	.dc.b	$19,$FE,$55,$BE,$14,$BE,$11,$86,$DF,$87,$1F,$23,$CE,$F9,$39,$C7
	.dc.b	$C9,$7C,$6B,$78,$79,$1A,$B1,$F1,$BC,$2F,$93,$E8,$38,$5E,$5A,$64
	.dc.b	$E5,$87,$89,$CE,$1E,$19,$E1,$1E,$DC,$3C,$38,$78,$23,$96,$3E,$E7
	.dc.b	$8E,$3E,$2C,$3E,$2D,$6D,$E9,$3C,$2F,$D7,$7C,$63,$C6,$78,$1F,$A4
	.dc.b	$FF,$2B,$6B,$EE,$31,$1F,$D0,$DF,$F7,$EF,$95,$FB,$F7,$F9,$91,$FE
	.dc.b	$E3,$70,$FF,$35,$F8,$4F,$F3,$AF,$07,$F3,$C6,$C7,$0B,$79,$1F,$CA
	.dc.b	$E3,$F1,$4D,$7F,$0E,$D7,$F0,$4D,$8E,$3D,$FE,$B8,$FE,$75,$97,$E2
	.dc.b	$3E,$8F,$FB,$6F,$6E,$6A,$61,$F3,$2F,$7B,$7F,$A8,$EF,$73,$9A,$DE
	.dc.b	$1F,$8F,$BD,$CF,$E7,$6B,$73,$9A,$DD,$FA,$9F,$BB,$3A,$8C,$57,$52
	.dc.b	$57,$9F,$CB,$FE,$66,$F3,$8C,$CF,$08,$D6,$3D,$A7,$E2,$E1,$56,$1D
	.dc.b	$DC,$9F,$97,$F4,$1F,$95,$7F,$AB,$AF,$FC,$69,$18,$F8,$0F,$EF,$B1
	.dc.b	$FE,$B7,$E7,$7C,$56,$30,$5E,$EB,$8E,$1C,$E7,$FD,$23,$39,$7F,$B2
	.dc.b	$D7,$18,$C7,$57,$F5,$09,$7D,$13,$E9,$7B,$72,$3F,$17,$7C,$07,$86
	.dc.b	$F7,$5A,$9C,$E9,$FE,$3B,$C3,$93,$FD,$22,$33,$5F,$E9,$D9,$CB,$FD
	.dc.b	$42,$33,$8F,$8E,$5F,$2D,$D7,$D9,$F0,$46,$30,$E9,$18,$9D,$47,$85
	.dc.b	$B8,$77,$6D,$F5,$6C,$FF,$BC,$6F,$FC,$2A,$FF,$5E,$FC,$49,$F4,$FE
	.dc.b	$0E,$4F,$87,$18,$9F,$74,$62,$75,$1C,$5B,$8F,$F6,$CA,$E0,$FC,$1A
	.dc.b	$E1,$5F,$1E,$3E,$7F,$0B,$E0,$BD,$F1,$86,$CF,$F7,$29,$D3,$C4,$C4
	.dc.b	$6A,$7F,$BD,$CE,$9F,$A0,$73,$9F,$F3,$87,$E0,$7F,$DB,$5F,$F4,$25
	.dc.b	$BF,$05,$FE,$97,$1A,$F2,$3F,$21,$39,$EA,$FC,$15,$F1,$C3,$FD,$BF
	.dc.b	$39,$BE,$07,$4C,$39,$CF,$FB,$76,$B8,$6F,$FE,$20,$BE,$13,$9C,$7B
	.dc.b	$DF,$0A,$FC,$B7,$D2,$3E,$93,$F4,$50,$FB,$8F,$F4,$6E,$7F,$E8,$BA
	.dc.b	$D3,$FE,$9F,$9D,$7A,$9F,$99,$9C,$E5,$FE,$F6,$9C,$D7,$1C,$3A,$E1
	.dc.b	$D2,$BF,$DE,$97,$C1,$FF,$30,$35,$FE,$DC,$1C,$3F,$DC,$5B,$9B,$F2
	.dc.b	$5F,$83,$87,$EF,$7E,$67,$E6,$C6,$2F,$FD,$6B,$1F,$AF,$F9,$2F,$F1
	.dc.b	$25,$7D,$6A,$D5,$27,$2F,$D2,$8C,$C6,$33,$FD,$F3,$6B,$FC,$09,$87
	.dc.b	$8F,$CE,$ED,$E8,$7F,$0D,$F6,$1F,$A7,$3E,$D7,$1E,$1B,$EE,$F5,$39
	.dc.b	$CB,$FC,$3B,$1F,$C1,$7D,$87,$C2,$FE,$23,$F4,$F7,$D9,$66,$30,$E7
	.dc.b	$EE,$F1,$87,$E4,$A9,$18,$E3,$71,$7F,$17,$F3,$1F,$4B,$FC,$8F,$F1
	.dc.b	$97,$D2,$7F,$86,$32,$FF,$66,$56,$9F,$F7,$04,$EB,$D8,$FE,$35,$39
	.dc.b	$C7,$AD,$7F,$89,$EB,$4F,$2C,$3B,$61,$D6,$B6,$7F,$86,$A3,$38,$74
	.dc.b	$6D,$7F,$90,$3E,$7B,$E4,$C7,$CA,$6C,$64,$FF,$22,$C6,$5A,$7F,$92
	.dc.b	$3D,$7F,$A3,$F6,$FF,$49,$B1,$93,$FC,$9B,$19,$69,$FE,$50,$FA,$FF
	.dc.b	$B3,$EB,$7D,$A6,$C6,$4F,$F2,$AC,$65,$A6,$72,$FF,$56,$C6,$5B,$1F
	.dc.b	$EA,$3F,$55,$F1,$7F,$86,$F1,$BF,$AB,$7C,$2F,$F6,$28,$C5,$6D,$B7
	.dc.b	$F2,$A7,$53,$FE,$66,$AF,$C6,$FD,$27,$E4,$22,$9F,$E6,$A2,$D7,$FB
	.dc.b	$01,$BC,$7F,$EA,$AE,$D7,$67,$FA,$FE,$7F,$CC,$53,$FD,$7F,$93,$F2
	.dc.b	$71,$83,$FC,$CC,$3F,$4C,$FC,$E6,$1F,$78,$4F,$07,$F0,$33,$BE,$EF
	.dc.b	$EC,$93,$9B,$FF,$5D,$CE,$4E,$D5,$F8,$DE,$AF,$EF,$51,$98,$ED,$5C
	.dc.b	$63,$C7,$87,$8A,$FD,$17,$C2,$31,$3F,$E7,$DC,$9F,$49,$23,$BC,$63
	.dc.b	$9F,$F7,$93,$9C,$BF,$33,$39,$AD,$CE,$71,$F9,$4F,$E7,$DA,$F2,$BF
	.dc.b	$BB,$F9,$43,$E0,$BF,$81,$F8,$58,$6E,$1F,$EE,$79,$D3,$E4,$F0,$BF
	.dc.b	$89,$66,$E3,$56,$A3,$CB,$3A,$7F,$DF,$1F,$01,$F2,$3D,$DF,$EB,$9A
	.dc.b	$FF,$86,$71,$4E,$7D,$2F,$F0,$B4,$60,$FA,$22,$B8,$BF,$CE,$0D,$5F
	.dc.b	$DC,$8C,$BF,$84,$AE,$93,$F6,$A3,$C6,$F4,$9D,$C3,$EC,$3F,$CD,$73
	.dc.b	$F4,$BC,$7F,$F0,$C4,$FE,$2E,$72,$FD,$28,$CE,$3F,$E3,$39,$CB,$F9
	.dc.b	$B8,$CD,$7F,$4E,$9C,$BF,$81,$8C,$D7,$CB,$CF,$F8,$E5,$FE,$8F,$3F
	.dc.b	$F0,$8A,$BF,$F6,$26,$9F,$43,$BF,$F9,$DF,$AB,$F7,$1B,$3F,$65,$B8
	.dc.b	$FE,$8C,$FF,$5C,$9F,$D4,$7E,$93,$FB,$0D,$FF,$BA,$2B,$FD,$A1,$E5
	.dc.b	$A7,$FD,$2B,$E2,$E1,$DF,$C3,$F8,$F0,$E4,$7F,$6A,$3F,$DA,$E6,$0F
	.dc.b	$C3,$3B,$AE,$91,$BA,$E8,$FB,$27,$55,$E0,$8E,$58,$79,$5E,$87,$AA
	.dc.b	$F9,$BD,$B0,$E6,$FF,$B8,$4E,$37,$D1,$F0,$18,$8D,$BC,$EF,$DB,$18
	.dc.b	$8F,$5D,$76,$BF,$13,$E0,$8E,$07,$A6,$78,$61,$E2,$AD,$DF,$AE,$F9
	.dc.b	$61,$E3,$C3,$CD,$18,$C3,$8C,$79,$4F,$79,$E2,$AF,$2B,$8C,$3B,$5F
	.dc.b	$38,$F1,$9F,$0C,$D1,$CE,$FD,$0F,$33,$D0,$69,$F1,$36,$BB,$C7,$0A
	.dc.b	$EF,$7E,$68,$D9,$9A,$ED,$5E,$F3,$89,$5E,$F1,$B3,$8C,$66,$3D,$11
	.dc.b	$98,$E3,$1F,$62,$36,$D8,$CB,$F2,$9A,$FC,$2C,$3E,$8B,$B3,$C3,$1D
	.dc.b	$E3,$31,$C4,$C9,$B3,$C7,$1B,$C3,$E7,$47,$C9,$32,$6E,$79,$BE,$18
	.dc.b	$DE,$1F,$22,$3E,$61,$D8,$F0,$1C,$8B,$3C,$9E,$CD,$8E,$57,$F3,$FE
	.dc.b	$15,$E5,$FE,$9D,$B7,$F9,$E6,$BF,$A3,$6F,$C2,$3E,$77,$FC,$6B,$29
	.dc.b	$CC,$61,$BB,$A7,$57,$FE,$68,$BE,$53,$9A,$DD,$7F,$01,$FA,$57,$CD
	.dc.b	$FF,$9B,$B2,$FD,$BC,$62,$7F,$6F,$87,$27,$F8,$43,$67,$75,$C7,$6F
	.dc.b	$F3,$BD,$F1,$BD,$E7,$EA,$BF,$E0,$78,$71,$BD,$E1,$C6,$FD,$58,$71
	.dc.b	$BD,$E1,$C6,$FD,$8D,$AD,$DF,$15,$7C,$6B,$79,$7A,$AF,$85,$6A,$F8
	.dc.b	$5E,$E3,$E5,$63,$E5,$6D,$70,$9D,$47,$13,$A5,$C7,$E6,$5F,$43,$C9
	.dc.b	$7C,$E0,$DA,$E6,$74,$74,$66,$79,$4F,$5C,$7E,$1D,$FF,$2A,$BE,$35
	.dc.b	$DB,$0F,$6E,$1C,$70,$D4,$E6,$F9,$63,$F4,$27,$26,$0F,$7C,$61,$F5
	.dc.b	$99,$3D,$37,$F0,$EB,$CB,$1F,$7A,$75,$87,$92,$38,$61,$E3,$79,$9B
	.dc.b	$26,$3F,$8A,$7C,$0F,$B9,$CE,$3E,$77,$8C,$64,$91,$97,$94,$70,$3D
	.dc.b	$67,$08,$D1,$F8,$47,$47,$DB,$34,$70,$8E,$AF,$E3,$8F,$2C,$71,$8C
	.dc.b	$C7,$17,$C2,$62,$7E,$03,$89,$E0,$7F,$81,$9F,$EC,$B7,$FE,$0B,$F4
	.dc.b	$7F,$DB,$ED,$BF,$F2,$EE,$DF,$DD,$E3,$FB,$E3,$78,$3F,$CA,$DC,$6B
	.dc.b	$E3,$18,$6D,$FF,$90,$F0,$EF,$3F,$EE,$CF,$D4,$1F,$69,$AF,$01,$FA
	.dc.b	$8F,$F1,$1C,$3F,$CB,$ED,$7F,$CF,$15,$C1,$FF,$1C,$35,$FF,$30,$C6
	.dc.b	$F4,$FF,$44,$37,$0F,$06,$1E,$4A,$7D,$9B,$BB,$1C,$4F,$26,$9B,$9D
	.dc.b	$57,$F9,$1D,$AF,$FA,$2E,$78,$3F,$E4,$B6,$BF,$E8,$F8,$DE,$5F,$E2
	.dc.b	$58,$F0,$DF,$67,$85,$F7,$8F,$93,$C0,$FC,$59,$F3,$3F,$1F,$8B,$FC
	.dc.b	$59,$8F,$F9,$0F,$1D,$D6,$9F,$B2,$D7,$FC,$E2,$71,$37,$5F,$47,$F4
	.dc.b	$37,$DC,$5F,$66,$57,$FA,$42,$72,$D9,$CB,$FE,$90,$84,$F3,$9F,$9D
	.dc.b	$F4,$58,$AF,$F3,$77,$D2,$BE,$B3,$E9,$36,$3F,$CE,$53,$B7,$FD,$6C
	.dc.b	$DC,$FE,$44,$EE,$7F,$D7,$4D,$8F,$99,$3B,$E3,$FE,$A9,$48,$C3,$FD
	.dc.b	$5E,$B8,$4E,$9B,$1F,$D6,$E7,$7C,$CD,$9F,$03,$F4,$7F,$05,$97,$FE
	.dc.b	$A1,$8F,$A8,$61,$FF,$4F,$C6,$1F,$D5,$FE,$42,$F8,$4F,$D0,$7A,$8F
	.dc.b	$26,$9F,$80,$9D,$57,$0A,$CB,$FE,$C2,$AE,$0E,$4F,$F5,$97,$E0,$5F
	.dc.b	$85,$FA,$D9,$18,$21,$FE,$B2,$72,$FF,$AD,$23,$F1,$86,$38,$3F,$D1
	.dc.b	$D3,$89,$FF,$5F,$7C,$07,$E4,$7D,$59,$18,$21,$FE,$BE,$72,$FF,$B0
	.dc.b	$23,$F3,$1F,$21,$F9,$9F,$C1,$24,$FF,$B2,$63,$2F,$FB,$2E,$32,$FF
	.dc.b	$B2,$23,$F1,$7C,$9F,$C3,$3D,$8F,$F0,$A7,$C3,$5F,$2C,$EA,$FF,$17
	.dc.b	$59,$89,$18,$20,$D3,$FD,$31,$7D,$23,$FD,$FA,$D7,$F2,$75,$FC,$27
	.dc.b	$C3,$7D,$57,$F0,$BF,$D4,$9E,$E5,$35,$16,$FF,$D9,$D4,$F6,$79,$19
	.dc.b	$8F,$F7,$2F,$99,$F8,$4E,$F9,$1F,$EE,$D8,$FE,$31,$F0,$6F,$8D,$A4
	.dc.b	$9F,$F6,$D7,$EA,$7F,$EF,$97,$FD,$B9,$3B,$73,$F5,$1F,$C8,$64,$F1
	.dc.b	$8F,$85,$B7,$EA,$27,$31,$F8,$FF,$9F,$FC,$B9,$C6,$BF,$95,$7D,$27
	.dc.b	$E4,$A7,$FD,$7C,$1C,$3F,$EC,$16,$BF,$B4,$6B,$FE,$49,$79,$71,$7F
	.dc.b	$25,$FA,$57,$D5,$7F,$37,$6D,$7F,$BD,$EB,$79,$73,$F0,$35,$72,$B5
	.dc.b	$1F,$CF,$FE,$99,$F4,$FF,$D0,$79,$7E,$95,$7F,$C0,$16,$BF,$08,$FF
	.dc.b	$2E,$4F,$07,$FC,$BD,$9F,$F3,$98,$C4,$6C,$FC,$0F,$1F,$53,$C1,$72
	.dc.b	$3B,$46,$5E,$79,$7F,$16,$AE,$77,$FD,$17,$E9,$BF,$AC,$53,$E7,$9F
	.dc.b	$8D,$3C,$6F,$ED,$B1,$87,$FA,$E5,$70,$A3,$CA,$75,$1E,$03,$A5,$73
	.dc.b	$70,$70,$F7,$39,$9D,$A3,$51,$F1,$67,$9B,$F0,$6B,$C0,$3E,$7F,$87
	.dc.b	$87,$27,$AC,$E5,$B3,$97,$A5,$AB,$D7,$32,$7D,$17,$F0,$A7,$86,$5F
	.dc.b	$1E,$BF,$BE,$74,$7F,$28,$BF,$37,$E3,$5F,$D0,$30,$3F,$D0,$A3,$F5
	.dc.b	$5E,$47,$03,$E5,$C6,$CD,$57,$3C,$3F,$BC,$37,$7F,$F5,$06,$1D,$3D
	.dc.b	$0E,$27,$3C,$7E,$5D,$7E,$CD,$F1,$FA,$DD,$8F,$3F,$C4,$48,$F6,$E5
	.dc.b	$FE,$08,$C9,$97,$68,$C3,$FD,$73,$0E,$93,$AA,$7D,$0B,$E7,$3F,$C0
	.dc.b	$B7,$FE,$B5,$8C,$4F,$F8,$D1,$AF,$E0,$FD,$1F,$EB,$DF,$83,$E6,$F8
	.dc.b	$CF,$46,$D7,$08,$FD,$35,$70,$7F,$13,$F4,$AF,$0B,$FD,$91,$0F,$85
	.dc.b	$FE,$24,$A7,$5E,$5F,$FB,$26,$73,$18,$9E,$1A,$FF,$7A,$FA,$67,$C7
	.dc.b	$7F,$7C,$BD,$E3,$D9,$B8,$F3,$BF,$F0,$0E,$3C,$E7,$C3,$87,$2C,$74
	.dc.b	$7F,$B6,$67,$55,$C3,$0E,$8D,$C3,$74,$BE,$4F,$98,$E5,$5C,$1B,$3C
	.dc.b	$2F,$AC,$7F,$B9,$EB,$8E,$1E,$88,$E4,$7C,$5B,$E4,$DB,$F0,$D7,$0C
	.dc.b	$3C,$93,$C3,$0E,$AE,$CE,$73,$B7,$C6,$6E,$FA,$CF,$09,$D4,$F1,$3F
	.dc.b	$C5,$25,$3B,$98,$8F,$AC,$EB,$E1,$5F,$1D,$FC,$5E,$3F,$A1,$C6,$39
	.dc.b	$BF,$A2,$C7,$FB,$4A,$74,$F1,$6B,$F6,$3E,$53,$FD,$18,$FF,$4C,$B4
	.dc.b	$62,$3C,$4D,$78,$7C,$3F,$69,$DB,$F4,$67,$9E,$BF,$92,$F9,$47,$D3
	.dc.b	$7E,$A6,$3F,$23,$3A,$DB,$9E,$3F,$36,$30,$85,$F2,$E5,$ED,$8C,$9E
	.dc.b	$0F,$E2,$27,$51,$82,$61,$E0,$9D,$63,$E3,$C7,$C3,$39,$AD,$C7,$F9
	.dc.b	$72,$F9,$CE,$49,$1C,$0D,$57,$3A,$FC,$F6,$3E,$A7,$F6,$4D,$CB,$E3
	.dc.b	$C7,$E4,$B0,$EB,$7C,$E1,$C4,$F6,$5F,$4D,$FC,$47,$A3,$F8,$97,$37
	.dc.b	$F1,$1F,$92,$C7,$07,$C7,$C3,$85,$E3,$1F,$E1,$95,$67,$35,$FD,$56
	.dc.b	$F9,$3F,$E5,$FB,$E1,$18,$BF,$EE,$1F,$46,$F6,$3F,$D5,$91,$FE,$80
	.dc.b	$8C,$5F,$E4,$1B,$1F,$E1,$66,$BF,$9E,$73,$97,$FA,$D5,$B3,$FE,$8A
	.dc.b	$FD,$03,$FD,$14,$FF,$66,$8C,$D7,$E0,$D7,$D2,$FB,$79,$9F,$4E,$FA
	.dc.b	$7F,$81,$6B,$DE,$1F,$D7,$2B,$4F,$5C,$3D,$75,$D5,$AF,$16,$D7,$28
	.dc.b	$F6,$47,$68,$C6,$5F,$EC,$FC,$BF,$C5,$77,$D4,$F5,$73,$F9,$FE,$9D
	.dc.b	$F2,$DC,$63,$0D,$9F,$95,$1E,$02,$38,$39,$9F,$E0,$AF,$9B,$E6,$E3
	.dc.b	$3F,$AE,$7C,$5F,$07,$FD,$E1,$1F,$2E,$31,$F1,$DE,$97,$DA,$48,$FB
	.dc.b	$4D,$39,$BF,$E1,$8F,$D5,$39,$C6,$0F,$F1,$A7,$D4,$7F,$BE,$E7,$FC
	.dc.b	$41,$8B,$EA,$3F,$CD,$51,$FE,$20,$9D,$57,$F9,$06,$B5,$D5,$F6,$27
	.dc.b	$33,$FC,$91,$AF,$FB,$39,$AF,$F0,$56,$A7,$AE,$FF,$F0,$05,$B7,$FC
	.dc.b	$2F,$E4,$5F,$25,$F8,$EB,$F4,$F9,$7F,$88,$FA,$7F,$F6,$8D,$FF,$8E
	.dc.b	$71,$7D,$47,$FB,$2E,$3F,$C7,$33,$AB,$FF,$2D,$56,$BC,$FE,$5F,$1B
	.dc.b	$F8,$AA,$D5,$7F,$3E,$6B,$FE,$E8,$6B,$FC,$7D,$E9,$5F,$E0,$F8,$46
	.dc.b	$2F,$D5,$97,$E3,$A7,$FC,$C9,$39,$5B,$3E,$A3,$E0,$C6,$1A,$BE,$67
	.dc.b	$FD,$4D,$87,$D8,$9C,$C6,$27,$FD,$1E,$D7,$FA,$BB,$5F,$F8,$02,$D7
	.dc.b	$B7,$BB,$FB,$28,$E5,$97,$F8,$A7,$0F,$24,$E7,$DB,$FE,$29,$F2,$1F
	.dc.b	$EB,$CC,$BF,$C4,$CD,$C7,$F0,$DF,$4F,$FE,$57,$9C,$FB,$FF,$BE,$F9
	.dc.b	$C7,$D4,$7E,$E1,$B3,$FD,$89,$B7,$FE,$A8,$6B,$FD,$34,$F8,$F6,$F9
	.dc.b	$0F,$E5,$1A,$23,$17,$FE,$39,$F2,$DF,$5D,$A9,$CE,$1F,$E9,$5F,$45
	.dc.b	$F4,$1F,$CC,$7E,$41,$F5,$3F,$E2,$CC,$BF,$D0,$35,$A7,$FD,$C3,$19
	.dc.b	$BE,$57,$FC,$4A,$B5,$1F,$EC,$06,$BF,$E0,$86,$BF,$E3,$78,$FE,$43
	.dc.b	$1C,$38,$FF,$DD,$F5,$F9,$F7,$F1,$91,$9B,$FE,$91,$9B,$EA,$7F,$A3
	.dc.b	$39,$C7,$FA,$9B,$99,$F8,$58,$EA,$B6,$D8,$C9,$F1,$FD,$9E,$7F,$53
	.dc.b	$D1,$B1,$93,$E6,$46,$5A,$7C,$FF,$A9,$F1,$FA,$FE,$46,$C6,$4F,$A9
	.dc.b	$19,$69,$9C,$BF,$CD,$39,$7F,$87,$23,$1A,$7F,$9E,$FD,$4F,$ED,$D7
	.dc.b	$FD,$9F,$C2,$FF,$07,$FD,$4F,$C2,$F0,$F1,$EB,$A3,$18,$FF,$AF,$6B
	.dc.b	$6D,$9D,$3F,$87,$FC,$3F,$8D,$7E,$7B,$CC,$F9,$B3,$4F,$C6,$17,$5F
	.dc.b	$3B,$74,$F3,$6B,$FE,$96,$8F,$F6,$4F,$2F,$E8,$EB,$FD,$D5,$0F,$AA
	.dc.b	$FD,$45,$FF,$09,$AD,$3F,$E9,$99,$D7,$6F,$F6,$69,$E5,$7F,$E5,$26
	.dc.b	$BF,$A5,$6B,$FE,$33,$6A,$7A,$CE,$FE,$1B,$FC,$20,$F8,$FD,$AE,$D1
	.dc.b	$FE,$11,$FA,$A1,$7E,$0B,$0D,$AF,$F5,$33,$A6,$C6,$9A,$7D,$1F,$A9
	.dc.b	$7C,$37,$F0,$AF,$17,$ED,$46,$27,$F8,$5C,$62,$64,$76,$F8,$1B,$A3
	.dc.b	$6F,$E6,$13,$FE,$6E,$3F,$DF,$43,$4F,$EC,$31,$AA,$F7,$C6,$27,$F8
	.dc.b	$61,$C6,$BF,$CF,$A7,$FB,$3A,$7F,$C1,$B1,$98,$FF,$40,$B5,$E3,$1F
	.dc.b	$E1,$A8,$E5,$5F,$67,$77,$CB,$E6,$BF,$C3,$FF,$55,$F9,$DC,$3F,$D8
	.dc.b	$FE,$4F,$F8,$53,$1F,$E3,$D3,$9C,$7F,$41,$39,$AF,$F3,$74,$E6,$B7
	.dc.b	$D5,$FE,$47,$AF,$F0,$6D,$7C,$59,$C9,$F4,$7B,$FF,$9C,$FA,$FF,$41
	.dc.b	$B1,$92,$4F,$FB,$6B,$8B,$FC,$62,$F3,$1D,$6E,$2F,$E2,$11,$FE,$D1
	.dc.b	$1E,$43,$E5,$FF,$9C,$53,$EA,$BF,$DE,$93,$9A,$FE,$48,$70,$AD,$CE
	.dc.b	$9F,$E5,$51,$FD,$85,$CE,$3F,$E8,$FF,$37,$94,$F8,$5B,$87,$79,$D5
	.dc.b	$72,$9C,$DA,$73,$7E,$29,$CC,$62,$3A,$57,$81,$F1,$1B,$71,$3E,$28
	.dc.b	$FF,$7C,$1E,$B3,$FC,$35,$5E,$B7,$DF,$87,$67,$47,$63,$77,$E1,$79
	.dc.b	$D6,$CE,$CB,$C6,$FF,$58,$57,$33,$2E,$30,$F4,$46,$27,$D2,$7D,$AC
	.dc.b	$3E,$13,$83,$BC,$FB,$0F,$90,$57,$E0,$1E,$63,$81,$E3,$3A,$E3,$F4
	.dc.b	$5C,$DF,$CB,$8C,$61,$E3,$C3,$D1,$7F,$79,$B7,$D8,$F3,$63,$F0,$DE
	.dc.b	$D5,$F0,$EB,$CB,$1E,$38,$F7,$19,$7E,$7C,$7C,$23,$B1,$5E,$D7,$E4
	.dc.b	$8E,$71,$E0,$3D,$27,$B3,$1F,$8F,$5D,$A7,$E2,$4F,$E8,$70,$F4,$B6
	.dc.b	$3B,$3F,$74,$E9,$3C,$5F,$44,$78,$2F,$C5,$1E,$4B,$EF,$39,$6C,$67
	.dc.b	$0F,$89,$8F,$DF,$8F,$A9,$8F,$CD,$BF,$0C,$7D,$C3,$B9,$C4,$F6,$C7
	.dc.b	$A1,$FC,$CB,$5F,$B8,$D9,$ED,$1F,$16,$3F,$19,$F1,$1F,$C0,$23,$F7
	.dc.b	$6D,$7F,$82,$35,$FD,$2B,$9F,$87,$7C,$7F,$F3,$B9,$4E,$63,$0D,$E4
	.dc.b	$4E,$AF,$FC,$89,$7C,$A7,$35,$BA,$FD,$6F,$D5,$3E,$A3,$FC,$95,$87
	.dc.b	$F0,$88,$C4,$FB,$30,$E4,$FE,$D1,$B3,$BA,$E3,$9F,$F9,$4E,$F8,$DE
	.dc.b	$F2,$F5,$5F,$F1,$4C,$39,$EF,$C6,$BF,$86,$5F,$08,$D5,$FC,$16,$E7
	.dc.b	$F3,$A7,$AE,$1C,$6F,$B4,$F9,$EF,$86,$3E,$3C,$BF,$CF,$AD,$AF,$E3
	.dc.b	$59,$78,$A3,$E3,$D7,$09,$F9,$78,$F8,$E7,$84,$7F,$06,$AE,$0F,$D0
	.dc.b	$6A,$9E,$0F,$2B,$F0,$BB,$BE,$CA,$BE,$5D,$F6,$C7,$D8,$7C,$09,$F6
	.dc.b	$9F,$C5,$B0,$EA,$F7,$C3,$AB,$CF,$0E,$B8,$79,$2B,$67,$07,$CE,$78
	.dc.b	$27,$CC,$7C,$DA,$FB,$6F,$9F,$1F,$98,$7D,$68,$EF,$5C,$EF,$B9,$EC
	.dc.b	$7B,$C7,$92,$FB,$CF,$E0,$B1,$F7,$13,$0E,$C7,$BB,$2F,$B8,$48,$FA
	.dc.b	$B8,$FD,$8A,$ED,$87,$88,$EC,$F8,$A3,$CD,$70,$FB,$11,$97,$F4,$33
	.dc.b	$EE,$7E,$51,$EE,$3C,$55,$F8,$0B,$F1,$47,$99,$FA,$99,$7D,$69,$F5
	.dc.b	$4F,$D1,$C3,$95,$70,$8E,$A7,$8A,$3D,$47,$96,$35,$8F,$C6,$8C,$57
	.dc.b	$84,$F8,$35,$C4,$F4,$9F,$5C,$F4,$5F,$F8,$27,$7F,$EB,$75,$FD,$69
	.dc.b	$BB,$7F,$06,$6D,$3C,$7F,$F2,$DD,$FF,$B2,$EF,$86,$5F,$E0,$A7,$FD
	.dc.b	$1D,$E9,$7F,$BD,$2F,$FD,$BB,$E8,$6F,$2E,$97,$B7,$FD,$A9,$F5,$5F
	.dc.b	$E3,$EA,$FE,$2B,$87,$F6,$C5,$19,$7F,$C5,$46,$3B,$3A,$3E,$5F,$AA
	.dc.b	$7D,$3E,$23,$47,$4C,$3F,$90,$EA,$AF,$E5,$33,$93,$97,$17,$1C,$7C
	.dc.b	$19,$78,$FD,$87,$C0,$75,$6D,$F1,$0E,$1F,$3F,$23,$CC,$E1,$5F,$E6
	.dc.b	$69,$FE,$67,$5B,$7F,$DE,$C7,$28,$D3,$4E,$11,$D6,$FE,$35,$BC,$EF
	.dc.b	$80,$DA,$F2,$56,$F5,$FB,$9F,$54,$FA,$AF,$93,$CC,$8C,$5F,$FA,$FD
	.dc.b	$AF,$F9,$67,$81,$F5,$5F,$E9,$28,$D5,$FD,$29,$73,$7A,$B5,$FC,$1D
	.dc.b	$76,$C3,$9C,$7F,$B4,$F0,$7D,$0F,$15,$F2,$B4,$C4,$4B,$FA,$D8,$F8
	.dc.b	$A7,$51,$8A,$E6,$FD,$3F,$A9,$E6,$D9,$FC,$24,$62,$3F,$17,$F5,$5F
	.dc.b	$DA,$1D,$1F,$56,$34,$FD,$46,$BF,$DA,$83,$87,$FB,$6B,$D7,$1F,$CD
	.dc.b	$4F,$E5,$27,$F0,$CD,$7F,$37,$3B,$DB,$F2,$BA,$3E,$07,$E6,$2F,$9B
	.dc.b	$A6,$22,$5F,$C4,$9D,$36,$72,$FF,$A4,$78,$1F,$54,$F8,$4F,$C0,$B6
	.dc.b	$7F,$DF,$CD,$7F,$01,$6F,$9E,$DD,$7D,$FF,$5D,$E6,$53,$37,$E3,$FA
	.dc.b	$AF,$4F,$D5,$48,$C1,$0E,$AE,$5E,$F1,$F0,$0C,$70,$FF,$0C,$9C,$4F
	.dc.b	$BF,$E1,$3E,$2F,$B1,$23,$04,$3D,$2E,$5F,$5C,$7C,$CF,$95,$F9,$BF
	.dc.b	$55,$27,$E5,$46,$5F,$99,$19,$7E,$1C,$7E,$2B,$8E,$47,$E3,$DF,$E2
	.dc.b	$7E,$DF,$F8,$CB,$E1,$F3,$3F,$EE,$4E,$6F,$E3,$8F,$E2,$EB,$F7,$B1
	.dc.b	$88,$FC,$63,$FC,$87,$4F,$F4,$6D,$FF,$8F,$3E,$AB,$AF,$9E,$7A,$CF
	.dc.b	$F2,$7F,$25,$F5,$3F,$C8,$5B,$8F,$F1,$56,$BF,$73,$E5,$7F,$93,$7D
	.dc.b	$53,$E0,$7F,$87,$B6,$7F,$8B,$35,$FF,$7E,$B5,$FD,$17,$83,$FC,$03
	.dc.b	$D1,$FC,$B9,$23,$C1,$F0,$1F,$CC,$FE,$AB,$F8,$44,$7F,$3A,$F8,$87
	.dc.b	$C8,$FF,$37,$6C,$7F,$2B,$9D,$ED,$FD,$0E,$73,$3F,$C9,$83,$87,$F9
	.dc.b	$4B,$5F,$DB,$35,$FF,$30,$CB,$EA,$7C,$84,$8F,$E7,$53,$DA,$BF,$A9
	.dc.b	$7D,$47,$E7,$A7,$F9,$F3,$5F,$EA,$6D,$8F,$EB,$1F,$53,$FC,$2E,$73
	.dc.b	$DD,$FC,$EA,$B8,$47,$F3,$E6,$BF,$C2,$DA,$FF,$9D,$9E,$9C,$FF,$C3
	.dc.b	$E9,$F5,$3F,$D7,$5B,$5E,$4B,$FE,$C1,$F5,$3F,$B7,$8F,$EA,$8D,$7F
	.dc.b	$B4,$B7,$1F,$ED,$DC,$5F,$84,$7F,$B8,$46,$E7,$FA,$9B,$5F,$E4,$0D
	.dc.b	$7F,$D1,$6D,$4F,$5C,$7F,$93,$53,$7F,$EE,$8D,$AF,$2F,$87,$FD,$DF
	.dc.b	$EA,$7F,$B9,$C7,$F7,$E6,$E3,$FE,$03,$6E,$1F,$E9,$AF,$A0,$FE,$B1
	.dc.b	$8F,$1F,$5F,$FB,$3D,$72,$C3,$FB,$5B,$5F,$E7,$0D,$7F,$D4,$6F,$45
	.dc.b	$4F,$A9,$FE,$57,$0C,$7F,$DF,$F3,$9F,$2F,$C1,$3F,$E0,$1B,$7D,$4F
	.dc.b	$AC,$91,$FE,$0D,$AF,$ED,$CE,$0F,$CD,$4F,$07,$FC,$08,$D7,$33,$EC
	.dc.b	$3F,$C3,$F3,$E5,$F1,$6E,$27,$A6,$3F,$E2,$9E,$7F,$E7,$4D,$F2,$7F
	.dc.b	$C4,$91,$8D,$BE,$57,$D6,$FE,$AB,$3F,$F0,$D4,$FF,$41,$7F,$C2,$6D
	.dc.b	$7F,$B0,$35,$FF,$65,$B5,$FD,$DD,$79,$BE,$AB,$FA,$E4,$7C,$C6,$BA
	.dc.b	$3B,$FC,$77,$90,$8F,$C7,$BF,$F1,$43,$A3,$FC,$51,$7F,$C2,$9F,$2C
	.dc.b	$FF,$89,$A3,$11,$FD,$AB,$EA,$BF,$C8,$BF,$4C,$FA,$1F,$F0,$A6,$7F
	.dc.b	$EE,$28,$C5,$7F,$6E,$A8,$FF,$90,$E3,$2D,$AE,$2F,$F9,$8F,$D0,$57
	.dc.b	$F9,$3F,$EA,$3F,$89,$F9,$1F,$EE,$EF,$88,$FE,$2F,$18,$B4,$63,$C0
	.dc.b	$FF,$08,$3F,$C7,$3E,$A9,$F2,$9F,$E0,$D8,$D6,$1C,$F6,$FF,$08,$63
	.dc.b	$CE,$73,$8F,$49,$FF,$0A,$61,$CA,$38,$1F,$F0,$02,$B8,$4F,$2C,$7C
	.dc.b	$2D,$C3,$74,$BE,$4F,$53,$95,$70,$6C,$F0,$BE,$B0,$39,$63,$EB,$C3
	.dc.b	$B4,$73,$13,$C1,$B7,$D5,$E2,$75,$71,$1B,$38,$61,$E6,$C7,$E1,$92
	.dc.b	$FB,$BF,$37,$0F,$16,$1E,$83,$FA,$55,$FD,$57,$FC,$E2,$D7,$CE,$7F
	.dc.b	$9D,$3E,$25,$F2,$5F,$C1,$A3,$FA,$94,$63,$A3,$FC,$F8,$D8,$CC,$ED
	.dc.b	$EB,$F2,$3F,$EA,$56,$FC,$57,$FA,$0F,$04,$E6,$7E,$FB,$5F,$B4,$FF
	.dc.b	$AB,$72,$FC,$6B,$73,$FF,$2A,$7C,$BB,$EA,$9C,$BF,$56,$B5,$87,$2C
	.dc.b	$B0,$7E,$02,$F8,$56,$27,$EB,$65,$E4,$AC,$65,$FD,$CA,$F7,$5A,$C3
	.dc.b	$AB,$F7,$1B,$7C,$63,$F0,$13,$C2,$BD,$37,$D3,$2F,$7B,$E3,$CB,$D3
	.dc.b	$7D,$63,$4F,$A4,$E7,$1A,$AE,$71,$B2,$3E,$78,$D5,$76,$9F,$CE,$79
	.dc.b	$8F,$69,$F5,$5E,$0E,$DF,$0F,$4F,$F0,$9F,$97,$C7,$07,$FB,$2A,$F6
	.dc.b	$FF,$05,$8C,$E3,$FE,$E7,$9C,$BF,$D2,$67,$F8,$87,$52,$B7,$97,$87
	.dc.b	$2E,$38,$F7,$BF,$F6,$67,$D4,$3D,$9E,$B9,$23,$13,$FE,$D1,$FA,$8F
	.dc.b	$F4,$9B,$C2,$B7,$3A,$AF,$F7,$4B,$75,$FF,$77,$5F,$09,$CD,$70,$F5
	.dc.b	$26,$1D,$7C,$FB,$5C,$A6,$E1,$CA,$6D,$7F,$A9,$71,$56,$2F,$F8,$26
	.dc.b	$3F,$E1,$AA,$F6,$CF,$9B,$8E,$47,$FB,$AF,$E6,$9F,$31,$8C,$3D,$35
	.dc.b	$73,$FC,$2C,$2B,$77,$F2,$2B,$4F,$8D,$F9,$31,$A9,$E7,$3F,$39,$F8
	.dc.b	$E7,$CE,$FA,$7F,$F7,$FC,$64,$91,$F2,$8D,$7C,$77,$08,$CE,$1E,$4D
	.dc.b	$3F,$91,$47,$F2,$38,$F3,$DF,$4B,$E5,$39,$3F,$D9,$9F,$4D,$FE,$FB
	.dc.b	$71,$1F,$CA,$E3,$0F,$FA,$EF,$E3,$7E,$C5,$FC,$29,$CB,$7E,$3F,$F0
	.dc.b	$F1,$8B,$FE,$69,$F0,$EB,$E3,$1E,$0A,$D9,$1F,$0C,$EF,$E0,$71,$8F
	.dc.b	$E6,$CF,$E7,$DF,$4B,$FE,$D7,$AF,$F3,$A4,$23,$0F,$F9,$DD,$AF,$FB
	.dc.b	$A4,$D7,$C4,$FF,$A8,$5B,$3F,$01,$A7,$BF,$93,$F4,$35,$D3,$4F,$C0
	.dc.b	$B6,$BF,$DE,$8D,$A6,$3E,$FB,$E3,$87,$ED,$B8,$5F,$4B,$F1,$6D,$18
	.dc.b	$D3,$F5,$FF,$2D,$FE,$59,$CF,$E9,$CE,$5F,$34,$66,$BE,$20,$71,$A7
	.dc.b	$F1,$AA,$EF,$ED,$F4,$F6,$3C,$59,$7F,$BA,$B1,$E7,$18,$9D,$47,$FB
	.dc.b	$F7,$DA,$FE,$0F,$E7,$FF,$AD,$E3,$C7,$F4,$9F,$E4,$17,$47,$C9,$F9
	.dc.b	$87,$C4,$7F,$08,$6C,$70,$F8,$AC,$4F,$C3,$A4,$53,$F8,$AF,$8A,$FA
	.dc.b	$5C,$5F,$D5,$F2,$5F,$3D,$FC,$51,$B1,$F9,$8F,$8B,$7D,$2F,$E3,$B9
	.dc.b	$7F,$C4,$97,$B2,$56,$E7,$53,$FE,$26,$9C,$BF,$32,$FA,$BF,$91,$BE
	.dc.b	$BD,$7A,$E5,$FE,$8F,$E4,$7F,$3B,$FA,$5F,$4E,$3E,$A6,$E7,$FA,$C9
	.dc.b	$D3,$75,$EB,$7C,$95,$70,$CB,$F4,$17,$FD,$1F,$D6,$F8,$BF,$4B,$F1
	.dc.b	$9B,$8F,$36,$D7,$0C,$3F,$DA,$F3,$DE,$BE,$6F,$D2,$FC,$EF,$4E,$CF
	.dc.b	$8D,$B7,$C5,$B9,$F9,$5B,$3F,$D4,$DB,$3F,$D4,$5F,$EA,$7E,$03,$D2
	.dc.b	$7D,$2E,$1B,$7F,$07,$E7,$31,$32,$3F,$80,$CE,$B0,$F5,$56,$5F,$87
	.dc.b	$D5,$FC,$C6,$BE,$51,$F9,$6F,$2B,$F1,$2D,$32,$7D,$18,$E1,$3F,$16
	.dc.b	$3F,$AE,$F2,$7E,$6E,$3F,$56,$DD,$3F,$C0,$72,$FA,$5F,$ED,$4F,$F0
	.dc.b	$79,$D3,$FE,$51,$A4,$61,$FF,$2A,$36,$39,$7A,$5E,$0F,$82,$91,$CD
	.dc.b	$FE,$F6,$E8,$C1,$F8,$9F,$A4,$1F,$DE,$8F,$F3,$CB,$97,$1E,$E6,$5B
	.dc.b	$19,$39,$47,$D6,$FA,$47,$C0,$D3,$A7,$B9,$6B,$F9,$9C,$E9,$B3,$A6
	.dc.b	$9F,$CA,$4F,$F2,$B9,$67,$D7,$A3,$E4,$BF,$C1,$73,$AD,$7F,$8C,$B7
	.dc.b	$6F,$ED,$AD,$3F,$AD,$D7,$17,$A1,$27,$8D,$F6,$BE,$AF,$F8,$63,$E9
	.dc.b	$7F,$CB,$B5,$FC,$F5,$B5,$FA,$CF,$1F,$FB,$A6,$3F,$DF,$27,$2D,$E7
	.dc.b	$E4,$7F,$8A,$27,$77,$F7,$EB,$6F,$F5,$ED,$3F,$76,$DE,$EF,$AD,$F5
	.dc.b	$5F,$3D,$91,$F3,$1C,$B4,$FE,$A5,$97,$FA,$EC,$FE,$79,$F2,$7F,$8D
	.dc.b	$EA,$7D,$37,$F9,$5B,$CE,$4F,$F5,$DA,$FF,$5F,$9C,$0F,$EE,$33,$B8
	.dc.b	$D7,$89,$FE,$A7,$AE,$35,$AC,$31,$5A,$8E,$47,$F5,$CB,$E4,$FF,$6B
	.dc.b	$BD,$BC,$4D,$9E,$9B,$EA,$F1,$BD,$BF,$EC,$F3,$AE,$1E,$C8,$CC,$71
	.dc.b	$C3,$E0,$61,$E0,$8E,$17,$D2,$F8,$47,$80,$F8,$31,$99,$E1,$1C,$CF
	.dc.b	$34,$79,$1C,$1E,$53,$96,$1E,$6A,$EB,$3B,$9E,$B7,$E4,$C3,$CF,$18
	.dc.b	$3A,$C7,$C9,$9F,$14,$EA,$FA,$61,$E4,$BE,$07,$86,$BB,$4E,$AB,$84
	.dc.b	$FC,$6A,$E1,$39,$8F,$39,$99,$DD,$78,$23,$2D,$3A,$BB,$33,$7E,$B6
	.dc.b	$9E,$A3,$EF,$1B,$8E,$27,$38,$F7,$1C,$5C,$B4,$CC,$73,$3E,$CB,$97
	.dc.b	$AC,$73,$6C,$65,$D9,$C4,$FA,$F8,$7D,$48,$EB,$1F,$25,$B1,$98,$E6
	.dc.b	$71,$3E,$7B,$97,$A9,$C4,$D9,$4C,$BB,$7C,$38,$7D,$27,$E2,$4F,$82
	.dc.b	$31,$5E,$B9,$FC,$F7,$C2,$BC,$BF,$D6,$36,$3E,$6B,$5D,$C7,$F4,$BF
	.dc.b	$83,$7C,$73,$8D,$4B,$E1,$5A,$9C,$C6,$2B,$85,$6E,$FA,$46,$27,$55
	.dc.b	$FB,$4F,$A6,$7D,$1F,$F5,$BC,$BF,$73,$18,$C7,$F7,$39,$77,$7F,$82
	.dc.b	$B6,$77,$5C,$74,$FE,$C9,$7C,$6F,$79,$79,$2F,$F8,$2E,$1C,$6F,$78
	.dc.b	$71,$BF,$7E,$1C,$6F,$78,$71,$BF,$5E,$3E,$9C,$38,$60,$C3,$85,$6A
	.dc.b	$3C,$35,$B6,$18,$26,$1E,$A7,$2F,$82,$3E,$0E,$1E,$19,$F2,$E1,$E0
	.dc.b	$C3,$BC,$7D,$6C,$7D,$8D,$AE,$0F,$96,$B8,$1E,$89,$DC,$1B,$5D,$23
	.dc.b	$81,$E2,$B8,$FF,$36,$8D,$1B,$9E,$67,$2A,$EF,$3C,$30,$F6,$3C,$63
	.dc.b	$2F,$C2,$C7,$E4,$4F,$C3,$8D,$1C,$63,$27,$D8,$7E,$3B,$87,$E0,$61
	.dc.b	$87,$E3,$9C,$4C,$9D,$09,$3E,$42,$4F,$58,$CB,$D5,$F8,$D8,$FC,$CC
	.dc.b	$7E,$0E,$1F,$2E,$3E,$7E,$3F,$06,$BC,$47,$0C,$7C,$4F,$CC,$3A,$9E
	.dc.b	$4C,$7C,$2F,$A0,$C9,$C4,$F8,$C6,$5F,$8C,$70,$74,$E3,$1F,$53,$9A
	.dc.b	$F2,$4F,$47,$F0,$FA,$FF,$70,$B6,$DF,$EE,$AB,$FE,$D7,$18,$BF,$F6
	.dc.b	$7C,$62,$7F,$82,$F9,$1F,$E4,$6D,$BF,$CA,$97,$BC,$3A,$57,$3C,$B2
	.dc.b	$DC,$7F,$C9,$D7,$87,$F8,$7F,$D2,$7F,$8E,$5F,$F7,$33,$5F,$F6,$8B
	.dc.b	$70,$FF,$51,$FD,$2B,$E9,$B9,$52,$B5,$87,$F9,$9D,$B1,$DB,$AF,$C1
	.dc.b	$A3,$0E,$53,$FE,$F5,$F6,$DE,$FF,$46,$C6,$63,$BF,$13,$C7,$D4,$49
	.dc.b	$D1,$FE,$FF,$9D,$38,$3F,$DD,$75,$D1,$B3,$D2,$BE,$27,$03,$E9,$5F
	.dc.b	$4D,$86,$DF,$F3,$6E,$0F,$A1,$18,$8F,$F5,$DF,$D2,$FF,$9C,$1D,$70
	.dc.b	$7F,$3C,$C3,$A6,$1C,$C9,$3F,$EC,$9A,$E0,$F2,$FA,$6F,$8F,$0F,$A6
	.dc.b	$A6,$22,$4F,$18,$CB,$CE,$3E,$AF,$A9,$C6,$3C,$67,$F4,$E7,$0F,$F5
	.dc.b	$1C,$7F,$D3,$D1,$8C,$FF,$25,$2F,$A6,$FE,$A8,$F2,$13,$BA,$E3,$87
	.dc.b	$80,$FC,$2D,$70,$AE,$C7,$86,$34,$F3,$AE,$4F,$F6,$29,$E1,$7F,$EF
	.dc.b	$2F,$35,$F0,$1F,$8A,$5F,$33,$4C,$44,$BF,$F6,$24,$65,$FF,$63,$47
	.dc.b	$DE,$FA,$67,$CB,$D2,$4F,$FB,$2F,$E6,$EB,$E8,$E1,$57,$08,$F9,$71
	.dc.b	$8A,$F2,$57,$97,$E9,$BC,$DF,$4D,$2F,$FD,$AD,$19,$7F,$DB,$11,$97
	.dc.b	$FD,$C3,$19,$7F,$DC,$71,$9D,$3E,$14,$FC,$1F,$7F,$E1,$7A,$72,$FF
	.dc.b	$DC,$51,$97,$FD,$C9,$19,$7F,$DD,$B1,$F2,$7E,$3F,$E5,$7D,$34,$9F
	.dc.b	$F7,$54,$65,$FF,$76,$46,$5F,$F7,$AC,$7E,$33,$92,$47,$D9,$7F,$86
	.dc.b	$7A,$FF,$E1,$9F,$83,$CC,$FE,$00,$FC,$3C,$E3,$3F,$E3,$D5,$B9,$E4
	.dc.b	$43,$87,$2F,$E4,$A7,$7E,$53,$0F,$E2,$3E,$13,$E9,$69,$30,$FC,$FF
	.dc.b	$86,$E8,$7C,$87,$F1,$8F,$A5,$7C,$3D,$25,$FE,$77,$1E,$8F,$2B,$FE
	.dc.b	$64,$B0,$E9,$E5,$7E,$4C,$F1,$FC,$47,$F2,$4F,$A5,$FD,$DE,$FF,$AA
	.dc.b	$F8,$57,$CB,$FF,$28,$6E,$3F,$AB,$AD,$F2,$7F,$90,$A1,$3A,$7E,$53
	.dc.b	$6F,$A5,$F7,$F7,$7C,$1F,$4B,$2F,$A3,$DE,$7F,$9B,$FD,$2B,$DC,$A4
	.dc.b	$AF,$DE,$7D,$2F,$EC,$33,$F1,$56,$E7,$F8,$03,$5F,$B8,$DC,$BF,$83
	.dc.b	$D7,$19,$FF,$2D,$CE,$DF,$60,$70,$F6,$BF,$F6,$F3,$C1,$AB,$E9,$5E
	.dc.b	$0F,$F5,$06,$C7,$F1,$36,$BF,$8F,$E4,$FB,$2F,$E4,$63,$38,$FF,$1C
	.dc.b	$6E,$5F,$3D,$B3,$FE,$72,$AE,$35,$BB,$F0,$12,$7F,$91,$4E,$AF,$EC
	.dc.b	$7D,$2F,$F9,$F2,$7F,$93,$35,$FD,$13,$6F,$FB,$F7,$CF,$FF,$4F,$E2
	.dc.b	$FE,$D7,$39,$BE,$7E,$CE,$0F,$EF,$15,$BC,$3A,$C6,$5A,$7F,$3E,$8D
	.dc.b	$3E,$0F,$A5,$EA,$47,$5E,$87,$D4,$CB,$F9,$FE,$2A,$D5,$71,$36,$62
	.dc.b	$77,$17,$1F,$F4,$CE,$1F,$DA,$6B,$FC,$0B,$C3,$FC,$9D,$FE,$04,$DB
	.dc.b	$FE,$9A,$DA,$E3,$F4,$7F,$A0,$AF,$E9,$D3,$97,$FD,$4A,$D7,$F5,$7E
	.dc.b	$47,$C4,$9E,$11,$DA,$3C,$24,$7E,$24,$6A,$FC,$3C,$BE,$62,$4E,$BD
	.dc.b	$2A,$FA,$22,$CF,$D8,$AF,$F0,$6C,$FF,$5C,$EA,$FE,$1D,$F4,$7F,$E2
	.dc.b	$DF,$9E,$7D,$07,$CD,$D7,$FD,$7B,$18,$BF,$EB,$55,$1F,$EE,$B1,$96
	.dc.b	$D7,$17,$F8,$87,$72,$BF,$C6,$3F,$45,$FC,$1B,$C5,$FF,$64,$FC,$17
	.dc.b	$F0,$78,$C5,$A3,$1D,$FF,$DE,$9F,$E1,$9F,$46,$F8,$BF,$E2,$11,$AC
	.dc.b	$39,$F2,$7F,$13,$C7,$9C,$E7,$1E,$93,$FC,$5F,$0E,$51,$C0,$FF,$6D
	.dc.b	$D7,$09,$E5,$8F,$85,$B8,$6E,$97,$C9,$EA,$72,$AE,$0D,$9E,$17,$D6
	.dc.b	$3F,$DD,$35,$C7,$0F,$31,$30,$EF,$3D,$0E,$F8,$FC,$08,$CC,$EE,$7B
	.dc.b	$1B,$AE,$93,$A2,$46,$A7,$E5,$D7,$38,$E2,$7F,$20,$BF,$A6,$DC,$7F
	.dc.b	$C4,$8F,$90,$FE,$4B,$F0,$4F,$8A,$73,$FF,$24,$8C,$36,$33,$3F,$CB
	.dc.b	$0F,$F6,$63,$87,$C1,$F1,$4F,$99,$FE,$5B,$8A,$31,$3C,$63,$13,$A9
	.dc.b	$CC,$F2,$8F,$9F,$1C,$33,$FE,$A9,$F2,$0F,$A3,$FF,$4C,$47,$E4,$67
	.dc.b	$5B,$7E,$25,$B8,$7E,$11,$AA,$F9,$72,$F6,$C6,$4F,$D7,$CB,$CA,$FD
	.dc.b	$9A,$DB,$9B,$E7,$3A,$C7,$C5,$7F,$8D,$BD,$C6,$2F,$84,$F6,$8F,$5E
	.dc.b	$5E,$AC,$BD,$13,$96,$D7,$86,$74,$FA,$E7,$4D,$35,$96,$39,$5F,$AA
	.dc.b	$BE,$0F,$FD,$7D,$D5,$FE,$BF,$9C,$C6,$2B,$FD,$7F,$F2,$1F,$EA,$47
	.dc.b	$F8,$2D,$F1,$6C,$EF,$05,$F2,$9F,$F1,$0D,$F1,$9F,$F3,$8C,$E5,$FF
	.dc.b	$1A,$7D,$0F,$F8,$92,$BF,$A4,$7D,$5B,$E3,$FA,$52,$B4,$FF,$9E,$B9
	.dc.b	$15,$BB,$FF,$42,$56,$B4,$FE,$BB,$39,$BE,$FE,$97,$9F,$0E,$38,$F6
	.dc.b	$C3,$F8,$E6,$3E,$2F,$8B,$C7,$66,$F2,$46,$27,$47,$82,$75,$19,$CB
	.dc.b	$CD,$5C,$2B,$47,$08,$F8,$27,$8F,$BB,$E3,$69,$FE,$A0,$C5,$39,$F8
	.dc.b	$7F,$33,$6F,$DE,$D3,$9B,$FD,$C6,$F9,$BF,$E3,$BA,$E6,$E6,$BE,$6F
	.dc.b	$CB,$6A,$FF,$7F,$6F,$BE,$CD,$79,$5B,$7D,$5B,$5F,$4E,$B8,$4F,$FA
	.dc.b	$D6,$FA,$6D,$FD,$4A,$F7,$97,$F8,$42,$F7,$87,$F9,$D7,$E8,$7F,$D5
	.dc.b	$B8,$FF,$B1,$A9,$F5,$5F,$EA,$68,$FF,$60,$4E,$AB,$FD,$77,$5A,$F2
	.dc.b	$3F,$0F,$39,$8F,$CE,$DF,$1A,$ED,$F0,$BF,$97,$9E,$18,$74,$8F,$F6
	.dc.b	$7C,$FF,$98,$6F,$B3,$FE,$48,$6B,$98,$DB,$F5,$BF,$01,$73,$7E,$8F
	.dc.b	$E1,$5F,$81,$FE,$47,$2F,$F6,$DD,$6B,$2F,$F5,$CC,$E5,$FF,$38,$46
	.dc.b	$70,$FF,$60,$4E,$5E,$D8,$75,$8C,$42,$78,$5F,$6A,$FD,$A7,$E0,$6F
	.dc.b	$C0,$DF,$2F,$FA,$CB,$DE,$5F,$8E,$EE,$7D,$1B,$E2,$3F,$79,$E7,$7F
	.dc.b	$A6,$7D,$7F,$DF,$56,$A7,$3E,$7F,$FA,$6B,$0F,$F7,$A7,$6F,$FA,$95
	.dc.b	$FF,$7F,$D6,$DC,$FC,$1F,$2C,$BB,$46,$09,$C3,$FE,$F6,$AF,$1F,$C4
	.dc.b	$7F,$02,$C3,$A4,$FF,$87,$F8,$5F,$3D,$E2,$F9,$DF,$3C,$63,$1F,$F5
	.dc.b	$44,$71,$FA,$1C,$FD,$D5,$9F,$E2,$AF,$0F,$2B,$18,$F5,$6F,$3A,$BF
	.dc.b	$D4,$9F,$15,$E8,$A4,$E5,$D7,$CF,$FC,$37,$FD,$A1,$3B,$9D,$52,$B5
	.dc.b	$1C,$69,$5B,$AF,$DB,$D6,$9F,$0F,$1F,$F2,$F7,$97,$27,$F2,$6A,$E1
	.dc.b	$7E,$28,$C5,$6E,$F9,$4F,$FB,$27,$57,$E0,$3F,$23,$C2,$F9,$2D,$CB
	.dc.b	$F9,$CB,$6B,$E8,$37,$3F,$81,$CF,$FA,$0C,$3E,$8F,$AD,$F4,$9B,$87
	.dc.b	$F4,$26,$E1,$F6,$5B,$7C,$97,$CD,$FD,$AF,$C0,$3E,$DB,$6F,$FA,$5B
	.dc.b	$6F,$F0,$4D,$BE,$4D,$9F,$EB,$2D,$97,$80,$FA,$07,$BB,$F9,$AF,$83
	.dc.b	$FD,$C4,$63,$0F,$F7,$84,$62,$75,$1F,$CD,$31,$FC,$05,$F3,$9C,$E1
	.dc.b	$F9,$5F,$80,$FC,$9F,$C1,$FE,$51,$B8,$7F,$64,$6C,$EE,$3F,$B3,$CF
	.dc.b	$F6,$07,$FC,$64,$77,$3F,$9C,$F4,$FF,$37,$9F,$CA,$7B,$7F,$83,$8C
	.dc.b	$4F,$09,$FC,$AB,$C3,$CE,$E9,$5B,$6C,$75,$9F,$CF,$57,$2A,$D4,$FF
	.dc.b	$7A,$F8,$FF,$D6,$FA,$DF,$AE,$6C,$64,$FE,$FF,$19,$69,$D6,$DF,$43
	.dc.b	$FE,$41,$8F,$07,$2B,$E2,$3F,$BC,$47,$F6,$0A,$DF,$BD,$FC,$4A,$73
	.dc.b	$1F,$D9,$2F,$93,$FE,$A1,$8C,$BC,$2B,$AC,$63,$CB,$7D,$AF,$F1,$CF
	.dc.b	$C0,$5F,$31,$FE,$0E,$9C,$DF,$92,$72,$FA,$63,$38,$7F,$AB,$E7,$2F
	.dc.b	$CE,$3F,$BB,$C6,$B9,$7F,$4B,$3A,$7F,$7D,$19,$AF,$EF,$BA,$2B,$51
	.dc.b	$F0,$67,$56,$AD,$61,$F9,$6F,$37,$F9,$8F,$D7,$5C,$3E,$5F,$9A,$8C
	.dc.b	$61,$FE,$C5,$8C,$4F,$DF,$E8,$FD,$15,$F2,$8C,$75,$3E,$13,$F9,$EF
	.dc.b	$CA,$7E,$91,$B3,$A6,$E3,$F0,$70,$FD,$64,$63,$7F,$EC,$BF,$05,$FE
	.dc.b	$1A,$FA,$27,$C9,$FD,$DB,$7D,$27,$F2,$99,$CE,$3F,$81,$3F,$D9,$01
	.dc.b	$CB,$FB,$A8,$E1,$5F,$E9,$1B,$EC,$F5,$8C,$CF,$39,$CB,$FE,$2C,$BF
	.dc.b	$0E,$58,$8F,$F6,$89,$DE,$F9,$46,$63,$FD,$BA,$64,$E5,$7D,$9E,$38
	.dc.b	$79,$AF,$AC,$6C,$CD,$F4,$70,$74,$9E,$15,$CE,$31,$3B,$9F,$83,$18
	.dc.b	$3C,$37,$E5,$7E,$3C,$72,$8C,$61,$ED,$C3,$C7,$3C,$AB,$77,$D4,$F9
	.dc.b	$4F,$9E,$77,$18,$9E,$93,$E8,$38,$9E,$23,$15,$DF,$0F,$54,$68,$FB
	.dc.b	$18,$79,$1E,$D1,$A9,$E1,$39,$AE,$E7,$9E,$33,$5D,$8E,$33,$BB,$F5
	.dc.b	$47,$DF,$32,$6E,$B9,$C7,$AC,$DC,$7D,$43,$D5,$1D,$DE,$2D,$38,$BD
	.dc.b	$5A,$F1,$3D,$A7,$78,$EA,$FD,$73,$31,$CE,$33,$1F,$17,$0F,$9E,$F5
	.dc.b	$8C,$DF,$C0,$AE,$13,$E6,$9F,$53,$4E,$05,$3C,$95,$D1,$B8,$7C,$F7
	.dc.b	$DB,$3C,$9F,$2B,$67,$84,$7A,$5A,$FE,$D3,$E2,$BF,$D4,$6F,$CF,$6B
	.dc.b	$FA,$A6,$BF,$A8,$73,$F1,$2E,$FF,$F1,$4C,$A7,$31,$86,$E9,$FE,$A2
	.dc.b	$6C,$EF,$D1,$FE,$8D,$7C,$EB,$F7,$5F,$49,$FE,$DD,$AB,$DB,$FD,$2F
	.dc.b	$5F,$E1,$37,$B9,$FD,$A3,$63,$F8,$6D,$71,$9F,$DC,$36,$38,$E5,$E8
	.dc.b	$6E,$1D,$AA,$1D,$27,$5A,$79,$9B,$1F,$EB,$CB,$E9,$3F,$D7,$DB,$1F
	.dc.b	$EC,$0C,$3C,$93,$C6,$FA,$E1,$E6,$C7,$D9,$39,$BF,$05,$6A,$3A,$4E
	.dc.b	$A7,$A6,$3E,$25,$87,$67,$C9,$3B,$58,$F9,$1D,$D7,$1A,$E1,$1A,$AE
	.dc.b	$15,$B9,$C9,$A7,$91,$9A,$E4,$78,$DC,$CF,$86,$78,$CE,$63,$A9,$97
	.dc.b	$8E,$1E,$FB,$F4,$9F,$CE,$EB,$8C,$66,$3C,$C6,$67,$8B,$F7,$AB,$AB
	.dc.b	$99,$8E,$4E,$55,$D1,$C1,$31,$FC,$1B,$F4,$8C,$63,$F4,$9E,$71,$A3
	.dc.b	$ED,$3D,$30,$F1,$C6,$23,$89,$F6,$2F,$8D,$73,$8C,$9F,$52,$32,$FC
	.dc.b	$23,$37,$E7,$AE,$84,$AF,$C4,$3E,$37,$A3,$99,$F0,$1E,$B7,$0F,$CA
	.dc.b	$31,$1F,$DA,$E1,$E6,$FF,$8E,$6F,$FD,$0B,$CB,$FD,$FA,$FF,$C2,$BB
	.dc.b	$E7,$87,$85,$D9,$FF,$3A,$E1,$C9,$D9,$0F,$F0,$73,$96,$E3,$C7,$A1
	.dc.b	$8B,$D4,$C3,$6B,$FC,$41,$7F,$E0,$AC,$FF,$CF,$2D,$CB,$FC,$67,$5B
	.dc.b	$AF,$F1,$45,$F9,$78,$BF,$BD,$61,$FE,$B7,$FD,$59,$F3,$5E,$1F,$C0
	.dc.b	$1F,$21,$4F,$66,$2A,$DB,$FE,$63,$6B,$FE,$49,$D5,$5C,$A3,$FC,$CA
	.dc.b	$D7,$FC,$9C,$73,$BE,$B5,$CB,$0F,$F5,$9D,$6D,$B8,$6A,$BC,$7F,$4C
	.dc.b	$F9,$AF,$91,$F4,$DF,$CC,$F2,$FF,$39,$B5,$EF,$3B,$7F,$CF,$4D,$7D
	.dc.b	$11,$BC,$3F,$D8,$6D,$DF,$FC,$F5,$5B,$9F,$F6,$3F,$D2,$3D,$EF,$B1
	.dc.b	$E9,$7F,$41,$C7,$FD,$06,$D7,$E2,$4E,$DF,$F4,$3B,$5F,$91,$1B,$AF
	.dc.b	$F6,$57,$67,$E0,$EB,$51,$FE,$ED,$FA,$67,$9D,$F5,$3B,$1F,$5B,$4F
	.dc.b	$AF,$3A,$CB,$FD,$30,$D7,$E8,$CF,$07,$FD,$38,$D7,$EA,$C6,$EF,$C1
	.dc.b	$7D,$67,$77,$DA,$3C,$DF,$0F,$E7,$F9,$D9,$5E,$E8,$CB,$F0,$23,$2F
	.dc.b	$C4,$8F,$CA,$7C,$CB,$E9,$FF,$3F,$30,$FF,$3F,$46,$5B,$19,$7F,$D0
	.dc.b	$87,$27,$E4,$7D,$AF,$C9,$F3,$A4,$FD,$08,$CB,$F4,$A3,$2F,$D6,$8C
	.dc.b	$BF,$62,$33,$9F,$E0,$27,$EA,$7D,$47,$D5,$FA,$89,$7F,$72,$32,$FD
	.dc.b	$E8,$CB,$F8,$28,$FB,$9A,$FD,$DF,$B8,$93,$F8,$48,$CB,$F8,$68,$CB
	.dc.b	$F8,$A8,$FD,$9F,$AA,$BF,$DA,$FD,$4B,$DC,$A6,$A2,$5F,$FB,$66,$74
	.dc.b	$D9,$CB,$FB,$4A,$E8,$E4,$86,$6B,$F2,$9F,$69,$FC,$13,$F0,$DF,$DB
	.dc.b	$75,$FF,$B7,$DB,$5F,$EB,$FE,$75,$6F,$BF,$F3,$9F,$61,$2B,$FD,$D9
	.dc.b	$7C,$3C,$36,$0F,$4F,$F9,$0F,$83,$2F,$F5,$31,$FC,$53,$EA,$3F,$95
	.dc.b	$72,$B5,$FF,$78,$46,$1B,$E5,$FF,$1A,$C1,$E9,$F1,$8C,$7C,$04,$CB
	.dc.b	$FD,$FD,$87,$14,$7F,$0F,$8F,$E6,$7E,$6C,$8C,$E5,$E1,$77,$8A,$BF
	.dc.b	$94,$FB,$9F,$99,$9F,$E4,$2D,$AE,$10,$8C,$67,$E0,$87,$03,$E2,$39
	.dc.b	$F6,$FF,$34,$87,$C5,$D3,$35,$E4,$BE,$58,$74,$8F,$5D,$70,$74,$62
	.dc.b	$B8,$BF,$CD,$BE,$99,$EF,$7F,$B6,$E3,$C9,$18,$70,$1C,$3B,$6B,$FE
	.dc.b	$E3,$73,$9F,$C0,$FA,$77,$D9,$7F,$4E,$6D,$79,$0E,$97,$FD,$4A,$F8
	.dc.b	$E7,$F1,$6B,$75,$FD,$31,$B8,$7C,$CA,$DC,$FD,$AF,$A6,$7C,$1F,$F7
	.dc.b	$7F,$99,$7E,$87,$F4,$6B,$E1,$C6,$37,$FE,$C3,$7C,$27,$C8,$D5,$8F
	.dc.b	$E0,$6B,$84,$72,$BE,$8F,$F6,$0A,$E1,$5F,$80,$6E,$7F,$8E,$BE,$35
	.dc.b	$D2,$31,$7C,$AF,$86,$3E,$4F,$6F,$D5,$F7,$0F,$83,$A6,$71,$F3,$CE
	.dc.b	$1F,$6C,$ED,$C9,$0F,$00,$C9,$1B,$8C,$C6,$2B,$B4,$7F,$7C,$FA,$47
	.dc.b	$9B,$F3,$7E,$E1,$E7,$53,$FC,$17,$3A,$AF,$BC,$D7,$AC,$F0,$7E,$9B
	.dc.b	$5E,$D5,$C8,$E8,$EA,$B0,$F4,$7C,$5E,$2C,$8C,$10,$EE,$E5,$F5,$47
	.dc.b	$F8,$EF,$E9,$1F,$99,$3E,$37,$0D,$AF,$D4,$CE,$A7,$FB,$C7,$CB,$69
	.dc.b	$93,$AA,$E0,$FF,$81,$27,$FC,$5B,$F5,$0F,$88,$FF,$05,$7C,$43,$6F
	.dc.b	$E0,$53,$9F,$67,$F8,$9C,$E7,$0F,$E2,$B3,$95,$19,$5F,$52,$F9,$2F
	.dc.b	$E1,$93,$9A,$E0,$DC,$7C,$38,$F8,$1B,$87,$F2,$2C,$39,$3E,$29,$D6
	.dc.b	$36,$3C,$75,$C1,$61,$DA,$BF,$8E,$5F,$29,$F3,$C7,$26,$9C,$AF,$A3
	.dc.b	$CA,$FA,$57,$1C,$3D,$15,$AC,$3B,$5F,$13,$99,$F0,$B1,$F8,$38,$F7
	.dc.b	$BE,$33,$AA,$E2,$EC,$E9,$8F,$AE,$32,$6B,$1F,$15,$FA,$6B,$F9,$C1
	.dc.b	$E7,$C7,$F9,$2D,$7D,$58,$CB,$5F,$F8,$03,$CD,$FA,$F7,$F9,$77,$C4
	.dc.b	$7F,$16,$E5,$FE,$5A,$D8,$CC,$ED,$E7,$D9,$FE,$7F,$7F,$0E,$D7,$F0
	.dc.b	$AD,$7E,$9B,$6F,$FA,$87,$C2,$FC,$98,$FF,$47,$6B,$F8,$9F,$A9,$FE
	.dc.b	$A1,$DB,$9A,$98,$7C,$7B,$DE,$BF,$D5,$EF,$73,$9A,$DE,$1F,$8F,$BD
	.dc.b	$CF,$F6,$CA,$DC,$E6,$B7,$7E,$A7,$F0,$33,$A8,$C5,$75,$25,$79,$FC
	.dc.b	$6F,$EE,$4F,$38,$CC,$F0,$8D,$63,$DA,$7E,$2E,$15,$61,$DD,$C9,$F9
	.dc.b	$4F,$3D,$EF,$FF,$9C,$AF,$F8,$0F,$D4,$BE,$03,$FC,$05,$E0,$FF,$80
	.dc.b	$EF,$7E,$FE,$2F,$FC,$17,$D6,$C3,$84,$7F,$A2,$27,$59,$2B,$58,$E3
	.dc.b	$0E,$77,$E0,$8D,$D3,$E9,$5F,$53,$FE,$0C,$8F,$F0,$54,$62,$BF,$BB
	.dc.b	$B6,$3F,$BE,$35,$FC,$CB,$9C,$FF,$C2,$AD,$9F,$F0,$A7,$DC,$FF,$78
	.dc.b	$7F,$BB,$46,$6B,$FC,$4E,$BE,$A7,$D7,$D0,$FA,$97,$C8,$FA,$2B,$58
	.dc.b	$7F,$70,$9C,$BD,$2F,$D5,$87,$46,$BE,$F6,$D7,$28,$F5,$C7,$53,$E2
	.dc.b	$FE,$13,$FE,$B7,$AE,$87,$A2,$7C,$BF,$53,$F6,$DF,$F6,$AD,$FF,$90
	.dc.b	$27,$FC,$07,$39,$D3,$E1,$D7,$80,$93,$F3,$1A,$72,$3E,$AF,$DD,$71
	.dc.b	$8F,$A7,$1F,$63,$EA,$F0,$FF,$6F,$B7,$CF,$E7,$93,$CA,$DB,$FF,$17
	.dc.b	$DF,$9E,$FF,$87,$B5,$FF,$24,$35,$FF,$2C,$BD,$93,$70,$EF,$3F,$E2
	.dc.b	$F8,$D7,$D5,$3E,$AB,$F2,$79,$BE,$AF,$FB,$DD,$FF,$8D,$6B,$4F,$F9
	.dc.b	$2A,$75,$E5,$7E,$52,$73,$97,$F9,$25,$AF,$F2,$66,$BF,$E5,$F6,$BF
	.dc.b	$E7,$07,$8C,$FF,$99,$9B,$5F,$E7,$4F,$95,$7C,$7F,$E4,$EF,$FC,$32
	.dc.b	$2B,$6F,$F9,$EB,$EA,$7F,$73,$7F,$E6,$0B,$7D,$5B,$EB,$3F,$57,$5A
	.dc.b	$BF,$F1,$1D,$6A,$73,$DF,$F9,$2A,$D5,$7F,$40,$6B,$FE,$79,$6B,$E1
	.dc.b	$8E,$38,$7F,$9C,$3E,$7D,$F5,$9F,$E9,$F8,$FF,$30,$5F,$87,$5F,$F4
	.dc.b	$67,$86,$FA,$C7,$D4,$FF,$06,$8C,$3F,$C0,$BB,$7F,$3D,$A7,$F8,$EA
	.dc.b	$73,$8F,$F9,$AE,$73,$3F,$D5,$9A,$FF,$A3,$DA,$FF,$A8,$1B,$3E,$1F
	.dc.b	$AC,$ED,$F3,$1E,$1C,$51,$8C,$3F,$D1,$7E,$27,$F2,$6F,$82,$FE,$31
	.dc.b	$C1,$FE,$96,$9C,$AF,$AC,$CC,$FA,$7D,$5F,$DE,$E8,$69,$2F,$FD,$25
	.dc.b	$87,$F3,$2F,$37,$A9,$DB,$E1,$6C,$FF,$A2,$7E,$B3,$E3,$BF,$E9,$E1
	.dc.b	$F5,$8F,$8C,$FE,$53,$3A,$E2,$FF,$30,$D6,$A7,$3C,$7F,$C4,$2B,$55
	.dc.b	$FD,$F9,$AF,$FA,$DD,$AF,$FB,$21,$E9,$C9,$FE,$BE,$6D,$FF,$0D,$C1
	.dc.b	$F5,$3D,$A3,$E2,$FD,$67,$C8,$F1,$3E,$4B,$63,$27,$CA,$8C,$B4,$F9
	.dc.b	$BC,$7F,$43,$DB,$FA,$2D,$8C,$9F,$4A,$32,$D3,$EA,$FC,$B7,$D8,$FA
	.dc.b	$CF,$B2,$D8,$C9,$F6,$A3,$2D,$33,$AF,$FA,$2B,$8B,$FB,$3C,$EA,$31
	.dc.b	$82,$3F,$AF,$F7,$BE,$A3,$FC,$DD,$87,$E2,$BE,$97,$FD,$D4,$FE,$DE
	.dc.b	$33,$87,$59,$CB,$FB,$68,$CF,$0F,$FB,$CE,$FA,$7C,$0F,$E4,$BD,$CE
	.dc.b	$F5,$AC,$7F,$DF,$BA,$7F,$7C,$FA,$CF,$CB,$FC,$47,$E6,$1B,$3F,$98
	.dc.b	$6E,$7E,$A6,$E7,$FE,$FA,$9F,$F6,$69,$F8,$FE,$3F,$E6,$0E,$2B,$F1
	.dc.b	$B1,$8C,$BF,$AD,$CE,$67,$FC,$92,$D4,$7F,$53,$8D,$9F,$E0,$08,$E9
	.dc.b	$47,$F1,$9F,$23,$99,$FD,$8F,$CD,$7E,$CB,$D9,$B5,$FB,$26,$E1,$F5
	.dc.b	$1B,$87,$F8,$67,$8B,$FC,$2B,$3F,$CB,$3E,$B1,$F0,$1F,$E1,$3F,$90
	.dc.b	$FF,$00,$4A,$73,$8F,$F0,$29,$CC,$62,$79,$57,$1F,$AB,$FF,$0C,$CE
	.dc.b	$71,$F5,$57,$C0,$F2,$DF,$45,$FE,$21,$FA,$CF,$F6,$F6,$7F,$12,$F7
	.dc.b	$6B,$DC,$FF,$12,$AD,$CF,$F4,$49,$CB,$FE,$76,$38,$5F,$F8,$32,$33
	.dc.b	$1F,$E7,$A6,$BF,$48,$DD,$78,$4C,$47,$4B,$FF,$20,$5F,$F5,$0A,$D9
	.dc.b	$F9,$5F,$9A,$FF,$1D,$74,$B3,$FC,$9C,$61,$B8,$FE,$7E,$3F,$94,$F9
	.dc.b	$F8,$BF,$CF,$7C,$57,$2F,$9A,$FC,$DB,$6F,$F4,$CE,$1F,$E5,$FF,$0F
	.dc.b	$6B,$FB,$29,$FE,$CA,$3F,$C6,$7F,$33,$ED,$EE,$7D,$5F,$D2,$F3,$13
	.dc.b	$F1,$FE,$27,$FA,$23,$B9,$D3,$9B,$7D,$5F,$07,$FB,$B4,$FD,$C8,$C4
	.dc.b	$7F,$B4,$8D,$CF,$3C,$BF,$3F,$3A,$9F,$33,$C3,$1F,$2C,$EA,$BC,$13
	.dc.b	$9B,$ED,$18,$3A,$5F,$63,$FD,$6B,$87,$17,$CB,$75,$F2,$63,$CE,$3D
	.dc.b	$F1,$C0,$F8,$E7,$02,$9E,$BC,$38,$3D,$CC,$5F,$8A,$3D,$91,$FE,$F6
	.dc.b	$38,$E3,$EC,$79,$9F,$56,$F9,$9C,$30,$F3,$4E,$EB,$15,$DE,$31,$7F
	.dc.b	$06,$3C,$D3,$A3,$CA,$58,$F2,$1B,$84,$F2,$7E,$05,$F8,$A7,$EF,$36
	.dc.b	$3E,$25,$74,$9E,$55,$C6,$7E,$09,$E2,$8C,$E3,$F5,$A3,$AC,$7D,$6B
	.dc.b	$F5,$9D,$63,$89,$E4,$9E,$0E,$2B,$B3,$F8,$6A,$F5,$BE,$46,$BF,$71
	.dc.b	$EC,$7A,$8A,$EC,$F3,$14,$F0,$63,$F8,$57,$B1,$A3,$C5,$5E,$47,$27
	.dc.b	$03,$F0,$46,$2B,$CE,$FE,$06,$FD,$CF,$DF,$BF,$1B,$D2,$FC,$0F,$E1
	.dc.b	$F0,$F9,$CF,$C8,$3F,$25,$5F,$22,$B8,$C7,$33,$2F,$BD,$A7,$E6,$FE
	.dc.b	$37,$FD,$66,$D9,$FD,$D4,$62,$3F,$6C,$DF,$8D,$78,$2E,$5F,$EA,$B3
	.dc.b	$98,$C5,$23,$13,$AB,$FE,$55,$7C,$A3,$15,$FD,$7A,$30,$FE,$B3,$EB
	.dc.b	$3F,$DD,$8F,$EC,$83,$87,$F6,$31,$C2,$BF,$86,$B7,$1F,$E7,$15,$B9
	.dc.b	$FE,$22,$D8,$F1,$E3,$E8,$9E,$2F,$AB,$2D,$F6,$FF,$41,$6E,$3F,$C1
	.dc.b	$F1,$F3,$C7,$0C,$BC,$D9,$75,$AF,$EF,$2D,$55,$E5,$7E,$06,$1D,$F2
	.dc.b	$F3,$E3,$DD,$B3,$DE,$74,$F0,$C3,$F8,$C6,$1F,$0F,$1F,$2C,$FF,$11
	.dc.b	$C7,$CB,$87,$BA,$F7,$8F,$5A,$F2,$B5,$F2,$E3,$D9,$B3,$EC,$9F,$F4
	.dc.b	$35,$79,$27,$E8,$D7,$C9,$8E,$D1,$FC,$87,$1F,$24,$6F,$1F,$8D,$3E
	.dc.b	$5A,$ED,$5A,$8F,$19,$EF,$C2,$3F,$83,$C3,$C2,$E8,$F0,$C7,$E1,$71
	.dc.b	$F7,$C6,$70,$F8,$67,$80,$98,$F8,$C9,$1F,$2C,$F2,$57,$BD,$87,$62
	.dc.b	$4F,$49,$E3,$1F,$18,$E2,$7C,$1B,$F9,$98,$7B,$63,$55,$E2,$7C,$D5
	.dc.b	$CB,$0F,$4C,$73,$CB,$ED,$92,$7E,$6B,$C6,$3F,$47,$87,$63,$ED,$57
	.dc.b	$23,$F1,$D1,$A3,$D7,$3F,$48,$E5,$87,$94,$F3,$4F,$1C,$BE,$AD,$F5
	.dc.b	$BF,$A7,$8F,$08,$C4,$7C,$08,$E9,$8F,$E1,$71,$F4,$4F,$CD,$7F,$C0
	.dc.b	$9A,$FF,$81,$BC,$EF,$F1,$66,$9F,$E7,$7C,$3F,$D6,$97,$FE,$F5,$6B
	.dc.b	$FE,$88,$F9,$AD,$EF,$FF,$00,$67,$35,$FE,$48,$F1,$5D,$3F,$E1,$AC
	.dc.b	$3F,$D1,$5C,$5C,$1F,$EB,$B7,$C3,$0E,$5A,$E7,$4F,$EF,$19,$78,$E3
	.dc.b	$86,$4F,$03,$F9,$24,$E7,$1F,$F4,$B7,$B1,$AD,$0F,$3C,$FF,$C3,$F5
	.dc.b	$FE,$56,$FA,$C7,$D5,$F8,$F0,$46,$1F,$F2,$61,$FE,$80,$70,$FF,$AA
	.dc.b	$FE,$B1,$F0,$9C,$67,$33,$FE,$55,$39,$D7,$F9,$75,$E4,$D7,$26,$63
	.dc.b	$AD,$FC,$9B,$7D,$1F,$C0,$6D,$79,$0E,$27,$F4,$1E,$0F,$F3,$27,$D5
	.dc.b	$BC,$AF,$63,$67,$83,$5F,$F6,$27,$D6,$3D,$1D,$D7,$CE,$8F,$E9,$B3
	.dc.b	$F0,$B8,$91,$88,$F0,$1F,$EA,$4B,$E9,$9F,$E0,$7E,$91,$F0,$1F,$63
	.dc.b	$93,$FA,$95,$F3,$6B,$FE,$D5,$9D,$BD,$DA,$FF,$B5,$BB,$F9,$1F,$82
	.dc.b	$8C,$DF,$FA,$43,$E1,$5F,$15,$F6,$DB,$7F,$1D,$AF,$A0,$77,$7C,$C7
	.dc.b	$11,$A7,$ED,$35,$FF,$48,$35,$FF,$61,$B5,$3D,$AB,$FD,$47,$2F,$A8
	.dc.b	$FC,$73,$6B,$C9,$7E,$BF,$88,$F6,$7C,$DC,$9F,$85,$19,$7E,$24,$65
	.dc.b	$F9,$11,$FA,$7F,$AC,$7C,$05,$24,$FC,$BF,$66,$BF,$EF,$73,$F4,$C6
	.dc.b	$33,$FB,$1F,$59,$F3,$7D,$29,$3F,$4E,$3E,$81,$87,$EC,$46,$1D,$1F
	.dc.b	$65,$D4,$F4,$8F,$B9,$F5,$9F,$67,$E7,$A4,$60,$87,$E0,$DC,$BF,$85
	.dc.b	$8F,$C0,$6F,$F8,$1F,$AC,$93,$F8,$78,$CB,$F8,$98,$CB,$F8,$D8,$FD
	.dc.b	$F7,$B2,$BA,$49,$FD,$67,$D6,$7E,$53,$1F,$F6,$DD,$6E,$7F,$74,$D7
	.dc.b	$FD,$B0,$D7,$F8,$03,$5F,$F6,$B3,$B8,$FF,$72,$CF,$EB,$E7,$BF,$67
	.dc.b	$85,$FD,$17,$D6,$7E,$74,$8F,$E8,$FD,$BF,$E2,$DF,$59,$F9,$78,$FE
	.dc.b	$1C,$D7,$FD,$C8,$D4,$1C,$77,$3A,$5E,$BF,$EA,$C9,$5F,$B1,$FA,$AF
	.dc.b	$E4,$9E,$0B,$D7,$FE,$4D,$F5,$9F,$CD,$EB,$AB,$5F,$E2,$11,$AF,$47
	.dc.b	$F8,$F4,$FF,$48,$9F,$E4,$8D,$7F,$DF,$4D,$7C,$F4,$FA,$CF,$09,$23
	.dc.b	$F9,$04,$F7,$F7,$FF,$3F,$1F,$C8,$DA,$FE,$F5,$BC,$4F,$1F,$E1,$CB
	.dc.b	$37,$C5,$FF,$41,$BE,$33,$FC,$E5,$AF,$06,$BF,$31,$A9,$EF,$5D,$A9
	.dc.b	$F1,$BF,$D3,$5B,$5E,$7E,$EF,$EA,$1F,$56,$E4,$A4,$9F,$E7,$DF,$59
	.dc.b	$F9,$3C,$7D,$35,$B9,$FE,$98,$D7,$FA,$80,$70,$FF,$53,$6B,$CA,$BF
	.dc.b	$B0,$53,$D7,$A7,$86,$7F,$93,$7D,$63,$E1,$7F,$B3,$36,$3F,$A9,$35
	.dc.b	$FE,$87,$F4,$0F,$15,$F3,$FC,$2F,$FB,$05,$EE,$FF,$B3,$35,$FA,$AD
	.dc.b	$7F,$32,$F5,$E2,$FB,$50,$C1,$F5,$9F,$46,$3F,$BB,$C6,$38,$BC,$11
	.dc.b	$E6,$F7,$BD,$24,$77,$9F,$E3,$DD,$1F,$DD,$F1,$FE,$BC,$F9,$A7,$D2
	.dc.b	$47,$BF,$D6,$3E,$B3,$FC,$39,$8A,$31,$3D,$0F,$EA,$1F,$08,$F3,$7F
	.dc.b	$C2,$0D,$8F,$EF,$8D,$7F,$AB,$7B,$2F,$AE,$DD,$7F,$82,$6B,$55,$FE
	.dc.b	$0F,$6B,$F9,$E6,$BF,$C0,$9E,$9B,$FC,$AC,$5F,$5A,$F4,$3D,$2D,$BF
	.dc.b	$1E,$1E,$69,$91,$E6,$7D,$FF,$5C,$F8,$8C,$CF,$E4,$B2,$E4,$DA,$F4
	.dc.b	$1F,$E2,$3F,$1B,$F8,$97,$CC,$7F,$90,$FE,$A5,$F5,$CF,$57,$0D,$C3
	.dc.b	$FC,$09,$5F,$E0,$13,$33,$23,$27,$F8,$97,$EB,$7C,$9E,$D7,$F2,$2F
	.dc.b	$8C,$69,$FC,$0A,$73,$49,$CB,$FC,$A2,$33,$18,$87,$D7,$3E,$5F,$F8
	.dc.b	$5C,$E6,$B7,$7D,$77,$FF,$0F,$CE,$5B,$1F,$C4,$AF,$96,$1E,$58,$E1
	.dc.b	$8D,$8F,$15,$70,$6C,$65,$F1,$E3,$E2,$AE,$31,$EB,$6E,$1E,$1A,$E1
	.dc.b	$87,$82,$B4,$74,$28,$C7,$D6,$E6,$BE,$14,$F1,$7A,$47,$09,$DB,$C7
	.dc.b	$1F,$43,$E6,$3D,$11,$A2,$46,$AB,$A6,$1E,$AA,$FF,$35,$57,$D6,$6D
	.dc.b	$7F,$92,$1D,$1F,$CF,$3E,$37,$F8,$B7,$27,$F9,$BD,$B1,$FC,$91,$AF
	.dc.b	$F9,$D5,$AF,$E0,$FB,$7F,$D0,$EF,$F2,$AB,$FE,$79,$F1,$1F,$E7,$96
	.dc.b	$C6,$67,$6F,$99,$E3,$9F,$E0,$DB,$97,$F5,$0F,$AE,$FE,$9F,$E3,$FC
	.dc.b	$29,$CD,$FE,$3D,$AF,$D1,$6C,$F0,$8F,$9D,$1A,$C3,$C7,$87,$7C,$7F
	.dc.b	$09,$7B,$8C,$5F,$09,$F0,$1A,$F2,$FF,$95,$CE,$89,$97,$C3,$9D,$65
	.dc.b	$C2,$F9,$12,$BF,$49,$87,$7B,$E1,$39,$6C,$EB,$0E,$F1,$83,$FC,$2B
	.dc.b	$CB,$FE,$BB,$F7,$BF,$9E,$F5,$FF,$3E,$E7,$7D,$77,$F9,$DB,$4F,$F4
	.dc.b	$45,$69,$FF,$0E,$CE,$BE,$03,$1A,$31,$E9,$5F,$EB,$AA,$D3,$C8,$C6
	.dc.b	$1D,$6B,$FD,$09,$F5,$4F,$AE,$E7,$08,$C5,$7F,$AC,$7E,$BB,$FB,$43
	.dc.b	$FE,$B3,$8C,$CF,$2A,$E3,$18,$D3,$FA,$CC,$AF,$8B,$BB,$E5,$E5,$70
	.dc.b	$3C,$3F,$5B,$FE,$AA,$8C,$92,$7C,$F3,$AF,$87,$E1,$39,$C3,$BE,$3F
	.dc.b	$DC,$A3,$FD,$91,$87,$3C,$38,$C7,$A5,$F3,$9F,$03,$EB,$3F,$D5,$71
	.dc.b	$92,$47,$C6,$35,$C1,$C2,$33,$87,$82,$73,$18,$3F,$C6,$3F,$1D,$D2
	.dc.b	$3F,$D6,$4D,$9F,$8B,$6F,$9F,$FA,$33,$9B,$E3,$86,$0F,$34,$D8,$FF
	.dc.b	$72,$4E,$F0,$FF,$30,$B5,$5F,$59,$DD,$FD,$EF,$AA,$7D,$67,$D6,$BF
	.dc.b	$F2,$3C,$62,$BF,$DF,$4D,$8F,$F1,$7B,$5F,$F8,$00,$E7,$2F,$BE,$D9
	.dc.b	$FF,$39,$FC,$AF,$FB,$1B,$D3,$F0,$61,$27,$FD,$F9,$5B,$C1,$7B,$9C
	.dc.b	$D7,$E5,$3E,$A5,$F1,$DF,$A0,$A7,$D5,$BE,$6F,$53,$98,$C6,$1F,$E8
	.dc.b	$4D,$3F,$27,$87,$28,$C1,$FE,$66,$71,$97,$FB,$FB,$E7,$9F,$57,$FE
	.dc.b	$88,$AD,$CF,$FA,$3A,$B7,$7D,$31,$F0,$D6,$B0,$E3,$F0,$B8,$C1,$7F
	.dc.b	$E3,$CF,$09,$F5,$6F,$A9,$FD,$DC,$61,$FD,$9F,$CF,$7E,$66,$B8,$56
	.dc.b	$F2,$F1,$E0,$AD,$47,$FA,$01,$B0,$C3,$BF,$D5,$F4,$EB,$EF,$82,$31
	.dc.b	$87,$BA,$B6,$F6,$C3,$9F,$83,$F4,$BE,$13,$F8,$8B,$67,$E2,$BF,$5B
	.dc.b	$EA,$F3,$C1,$FE,$84,$9E,$5E,$97,$B0,$CF,$8F,$63,$8C,$FF,$41,$DF
	.dc.b	$FC,$C3,$87,$E0,$2B,$9F,$C2,$70,$BF,$83,$F5,$7F,$0F,$87,$E6,$C7
	.dc.b	$FA,$82,$31,$87,$F9,$AA,$1F,$56,$FA,$7F,$E3,$F5,$A9,$E1,$C9,$83
	.dc.b	$F9,$6B,$C2,$70,$F8,$22,$E1,$E0,$3F,$D6,$17,$D6,$74,$7B,$3E,$AF
	.dc.b	$E1,$F8,$7F,$11,$B1,$93,$E0,$C6,$5A,$7C,$4D,$BE,$5F,$AD,$F3,$1B
	.dc.b	$19,$3E,$4C,$65,$A7,$CC,$F9,$2F,$A7,$F5,$7F,$51,$B1,$93,$E8,$C6
	.dc.b	$5A,$67,$2F,$77,$0F,$EC,$A7,$51,$8C,$11,$FB,$5F,$AB,$FF,$2E,$5F
	.dc.b	$EA,$3C,$14,$62,$F8,$7C,$46,$FD,$B9,$7B,$AD,$71,$FE,$0B,$2F,$F5
	.dc.b	$A4,$E6,$BF,$1B,$F5,$7F,$8E,$F8,$6B,$3F,$8A,$6E,$5E,$56,$E1,$CD
	.dc.b	$FE,$BF,$3D,$27,$ED,$1F,$EE,$9C,$BF,$03,$C4,$F9,$0D,$C4,$AF,$D8
	.dc.b	$9C,$70,$FC,$75,$6E,$33,$3B,$BE,$B1,$B0,$FE,$22,$F5,$7E,$FF,$97
	.dc.b	$FD,$4F,$AB,$FA,$A6,$DF,$E9,$9B,$7F,$45,$B5,$FD,$C3,$EA,$FF,$DE
	.dc.b	$D5,$FC,$19,$B5,$F6,$BE,$39,$F5,$5F,$E0,$09,$CD,$F1,$F7,$70,$7E
	.dc.b	$FA,$B7,$87,$38,$FF,$03,$F9,$0F,$9E,$A7,$F0,$4F,$AB,$FE,$79,$9A
	.dc.b	$75,$7F,$C6,$1B,$1F,$41,$B8,$FF,$00,$AE,$31,$8A,$F8,$F5,$CE,$FB
	.dc.b	$C7,$F1,$0A,$DC,$7E,$47,$CA,$93,$FC,$33,$CE,$B7,$FB,$7E,$8F,$C9
	.dc.b	$D6,$32,$FC,$75,$5B,$FE,$B3,$D1,$FE,$BF,$7F,$29,$F0,$5F,$C7,$3D
	.dc.b	$DF,$CF,$32,$32,$7F,$1A,$8C,$65,$F8,$C3,$FD,$68,$3E,$9D,$F5,$3E
	.dc.b	$CD,$1F,$55,$FF,$00,$27,$35,$F0,$0F,$F5,$C1,$C6,$B8,$4E,$DD,$46
	.dc.b	$A7,$8B,$FD,$B2,$74,$F3,$F8,$1F,$C8,$CE,$AF,$C3,$8E,$30,$E7,$56
	.dc.b	$FF,$BD,$E1,$C9,$FF,$61,$9E,$33,$35,$B7,$38,$75,$9E,$11,$F0,$A3
	.dc.b	$13,$B8,$C5,$FB,$23,$11,$EB,$8C,$3D,$CE,$AE,$27,$DF,$1C,$8F,$19
	.dc.b	$B3,$07,$C3,$9E,$B7,$E7,$C3,$E0,$CE,$CF,$01,$CA,$BA,$B8,$39,$1B
	.dc.b	$9E,$58,$7B,$A3,$55,$D2,$B8,$D6,$AF,$C3,$5E,$BA,$D6,$1F,$42,$31
	.dc.b	$87,$3B,$E1,$1D,$CC,$DF,$2C,$3E,$5C,$65,$A7,$81,$D9,$9B,$F9,$07
	.dc.b	$78,$DB,$4C,$C7,$E1,$CE,$31,$CC,$DB,$96,$BF,$50,$F9,$11,$CD,$A6
	.dc.b	$DE,$AE,$CF,$97,$1D,$63,$31,$F6,$30,$F8,$AF,$13,$99,$C5,$CB,$70
	.dc.b	$F9,$0F,$A2,$7C,$4F,$66,$CF,$08,$ED,$3C,$2B,$D7,$18,$BF,$9D,$7E
	.dc.b	$48,$FD,$9F,$C3,$FF,$8C,$5F,$8C,$D7,$F4,$AD,$7F,$80,$B9,$F8,$57
	.dc.b	$6F,$F9,$12,$53,$98,$C3,$78,$3F,$A3,$B6,$77,$E7,$FF,$24,$BE,$75
	.dc.b	$FB,$1F,$AB,$76,$7F,$28,$D3,$F8,$1C,$63,$0F,$DB,$5F,$07,$F5,$8D
	.dc.b	$9D,$D7,$1C,$FF,$98,$DF,$1B,$DE,$7E,$9B,$FE,$1F,$87,$1B,$DE,$1C
	.dc.b	$6F,$FA,$FE,$1C,$6F,$78,$71,$BF,$53,$6B,$77,$C5,$5F,$1A,$DE,$3E
	.dc.b	$AB,$E1,$5A,$BE,$17,$B8,$FE,$DF,$8F,$95,$B5,$C2,$75,$1C,$4E,$97
	.dc.b	$1F,$E2,$11,$A3,$75,$CE,$0D,$8D,$5F,$A6,$34,$65,$F2,$99,$BF,$4C
	.dc.b	$66,$BE,$D4,$E6,$FC,$58,$FD,$49,$C9,$8C,$7E,$1B,$E6,$AE,$AE,$4F
	.dc.b	$7D,$70,$74,$64,$F4,$E1,$F0,$E7,$8C,$7D,$E9,$D6,$3E,$C8,$E1,$8F
	.dc.b	$9D,$E6,$6C,$98,$F9,$DC,$BF,$80,$AE,$77,$DA,$31,$3C,$31,$FC,$4B
	.dc.b	$E7,$8C,$92,$33,$1D,$1F,$83,$19,$8F,$C4,$9C,$63,$24,$72,$71,$3E
	.dc.b	$29,$E1,$33,$1E,$93,$A3,$F1,$AB,$A3,$A9,$E9,$B7,$F7,$DF,$9A,$FF
	.dc.b	$0B,$D3,$6F,$F7,$1E,$09,$CC,$61,$B8,$3C,$5F,$F6,$FF,$B7,$8C,$7A
	.dc.b	$4F,$F9,$AF,$0E,$4D,$BE,$37,$C2,$35,$97,$1A,$FF,$01,$E5,$E3,$9F
	.dc.b	$F7,$7C,$63,$2F,$F2,$9D,$FF,$BB,$B7,$E1,$84,$BE,$67,$FA,$DA,$3C
	.dc.b	$EB,$EA,$9F,$23,$86,$C7,$FB,$59,$AF,$FA,$2B,$EA,$5F,$03,$BA,$93
	.dc.b	$B7,$D7,$3A,$8C,$10,$FF,$3B,$D7,$4D,$3F,$D8,$F3,$B9,$FF,$3D,$FD
	.dc.b	$47,$F9,$E5,$FF,$73,$B5,$CF,$8B,$F2,$3E,$A5,$E6,$FB,$2B,$58,$7C
	.dc.b	$7B,$46,$37,$FF,$29,$B6,$3E,$C5,$F2,$7F,$DC,$7F,$40,$F7,$7E,$AF
	.dc.b	$9D,$FC,$FA,$7F,$DB,$FD,$7F,$62,$B5,$1C,$2B,$73,$23,$07,$FC,$00
	.dc.b	$FA,$97,$C2,$7D,$26,$C6,$63,$EA,$F5,$3E,$AB,$51,$27,$47,$FA,$4B
	.dc.b	$6F,$75,$6E,$32,$7F,$A9,$2F,$AB,$6B,$AD,$FA,$7E,$4F,$D5,$F2,$B2
	.dc.b	$BE,$0C,$65,$F8,$71,$97,$FD,$97,$A2,$7F,$3D,$F3,$6F,$72,$92,$BE
	.dc.b	$5F,$AB,$5F,$F3,$D4,$ED,$FF,$67,$35,$EB,$1F,$37,$EA,$BE,$77,$99
	.dc.b	$27,$EA,$C6,$5F,$AF,$19,$7C,$B1,$97,$CF,$19,$DF,$F5,$5D,$CC,$D7
	.dc.b	$F6,$FE,$AB,$EE,$7D,$44,$C7,$F0,$71,$97,$F0,$B1,$97,$E1,$C7,$E0
	.dc.b	$FE,$A7,$F0,$9F,$55,$27,$F1,$71,$97,$F1,$B1,$97,$E5,$C7,$ED,$3D
	.dc.b	$75,$D2,$4F,$DF,$EB,$FF,$55,$FD,$57,$E4,$E3,$FD,$C3,$5C,$1D,$1F
	.dc.b	$ED,$67,$0E,$AB,$F8,$0C,$F1,$D7,$F4,$D5,$A6,$CF,$82,$BF,$DD,$B5
	.dc.b	$BB,$FE,$1F,$A3,$EA,$BF,$44,$4A,$FD,$4F,$AF,$FC,$3B,$EA,$BF,$29
	.dc.b	$1F,$EF,$66,$E5,$FB,$56,$E1,$D2,$FF,$59,$EA,$CB,$FD,$9B,$FA,$FF
	.dc.b	$A7,$FE,$33,$E1,$BE,$AB,$F8,$DF,$D4,$FF,$2C,$B6,$7F,$C0,$63,$0F
	.dc.b	$F0,$3A,$E1,$3F,$18,$4E,$B2,$FF,$6B,$B6,$3F,$DF,$ED,$53,$C9,$EF
	.dc.b	$3F,$ED,$E8,$7D,$57,$98,$93,$DE,$B9,$7B,$9F,$9F,$C3,$F8,$B3,$70
	.dc.b	$FF,$73,$B6,$83,$27,$C4,$FF,$3A,$9D,$49,$A6,$27,$71,$EF,$8C,$7C
	.dc.b	$67,$F4,$29,$3C,$FB,$7F,$A2,$FD,$53,$8A,$93,$0F,$91,$F5,$5F,$83
	.dc.b	$7C,$B3,$B7,$E6,$35,$F8,$A6,$A7,$FA,$8C,$F1,$84,$ED,$FF,$74,$35
	.dc.b	$F8,$26,$A7,$FD,$CE,$F3,$2F,$D4,$BE,$13,$FA,$EB,$63,$EF,$35,$F1
	.dc.b	$FD,$13,$C2,$7C,$E7,$0B,$95,$C1,$FE,$FB,$5B,$9C,$CF,$B0,$5F,$5E
	.dc.b	$1F,$EA,$0D,$E1,$FB,$37,$C6,$BC,$7E,$E7,$98,$8F,$EE,$23,$C9,$C9
	.dc.b	$88,$ED,$7F,$E9,$07,$89,$BC,$FF,$1E,$E0,$DD,$F1,$C7,$FC,$11,$87
	.dc.b	$49,$FD,$07,$BE,$F1,$FF,$C0,$6D,$8C,$C7,$F8,$1B,$D4,$7D,$3E,$A2
	.dc.b	$4E,$8F,$D5,$4E,$9C,$1F,$16,$BA,$36,$7A,$57,$83,$C9,$93,$E4,$7E
	.dc.b	$7F,$D3,$3E,$0B,$33,$CE,$6E,$7E,$CA,$D4,$FF,$78,$F9,$2F,$F1,$9F
	.dc.b	$D1,$3E,$99,$E7,$61,$B5,$B2,$1F,$96,$9E,$13,$B3,$FB,$F5,$7F,$8A
	.dc.b	$3E,$97,$C9,$EA,$7E,$D3,$E1,$9C,$1F,$96,$9C,$E2,$9C,$BF,$BA,$8C
	.dc.b	$C6,$21,$F4,$CF,$91,$FC,$FC,$E6,$B7,$7D,$78,$3F,$C2,$53,$96,$C7
	.dc.b	$E9,$AF,$96,$1E,$58,$E1,$8D,$8F,$15,$70,$6C,$65,$F1,$E3,$E2,$AE
	.dc.b	$31,$EB,$6E,$1E,$1A,$E1,$87,$82,$B4,$74,$28,$C7,$D6,$E6,$BE,$14
	.dc.b	$F1,$7A,$47,$09,$DB,$C7,$1F,$43,$E6,$3D,$11,$A2,$46,$AB,$A6,$1E
	.dc.b	$AA,$FE,$16,$73,$C3,$F8,$5D,$FD,$88,$CB,$5F,$F7,$B7,$5B,$97,$F8
	.dc.b	$8F,$C3,$7F,$0F,$6C,$66,$7A,$9F,$EC,$C7,$0F,$4F,$0D,$F3,$1E,$06
	.dc.b	$C7,$0C,$3F,$9D,$4E,$49,$1F,$0E,$77,$AF,$E2,$7E,$9B,$FA,$1F,$77
	.dc.b	$C2,$9C,$EB,$F9,$36,$DF,$E4,$5A,$AB,$8B,$D7,$0E,$44,$7E,$CC,$6B
	.dc.b	$0F,$CA,$57,$2C,$7C,$F5,$C4,$DF,$8D,$FE,$97,$AD,$E1,$CD,$B3,$AC
	.dc.b	$B1,$97,$9E,$AC,$FF,$60,$CB,$C1,$39,$8E,$D7,$C8,$F8,$58,$77,$9D
	.dc.b	$47,$EC,$3D,$5F,$F1,$F6,$3F,$EA,$9E,$8F,$F2,$0F,$9D,$FE,$6D,$8F
	.dc.b	$F3,$F9,$93,$FC,$3B,$EB,$DC,$BF,$D7,$93,$FE,$C3,$9F,$F4,$2C,$E6
	.dc.b	$FA,$C6,$0F,$F0,$67,$C1,$3D,$DE,$5C,$4F,$3D,$F3,$5D,$73,$FF,$55
	.dc.b	$D6,$A7,$38,$72,$8C,$56,$F0,$E9,$3F,$D1,$EB,$77,$FD,$4E,$5F,$29
	.dc.b	$CF,$E6,$38,$36,$FC,$AD,$AE,$0D,$DB,$FD,$51,$7C,$AB,$B5,$F2,$69
	.dc.b	$FE,$29,$FB,$A7,$B3,$F1,$F9,$5F,$77,$FC,$92,$B5,$39,$8C,$67,$FD
	.dc.b	$A6,$F8,$D7,$F9,$62,$F8,$F2,$7C,$59,$CD,$7F,$82,$A7,$35,$D0,$E1
	.dc.b	$7D,$33,$FF,$60,$5F,$07,$FD,$D1,$5C,$2F,$FB,$C7,$D1,$3E,$8B,$EF
	.dc.b	$53,$EB,$BF,$D3,$99,$7F,$91,$6B,$4F,$F9,$B2,$75,$E3,$FD,$99,$CE
	.dc.b	$1F,$E1,$E9,$CD,$71,$C3,$AE,$1D,$2B,$C3,$1F,$ED,$57,$FC,$3B,$7C
	.dc.b	$9F,$E2,$AD,$73,$1B,$7F,$DF,$FF,$71,$F9,$1F,$D5,$1F,$73,$F8,$BE
	.dc.b	$84,$E6,$30,$DC,$BF,$CC,$13,$A7,$FD,$0E,$AB,$9C,$6A,$78,$BA,$87
	.dc.b	$D7,$E3,$0F,$F1,$15,$FF,$77,$AE,$7F,$6D,$FE,$0D,$7B,$F8,$DF,$A7
	.dc.b	$8C,$57,$08,$C7,$9D,$E9,$6D,$70,$24,$FF,$92,$AD,$F5,$BF,$76,$7F
	.dc.b	$C8,$B1,$E1,$9D,$65,$F5,$38,$3F,$D7,$ED,$AF,$ED,$BE,$FF,$C9,$DF
	.dc.b	$F6,$5E,$0F,$FA,$6E,$7D,$D1,$89,$D4,$7C,$1F,$43,$0D,$3F,$D1,$55
	.dc.b	$C7,$1F,$0F,$DC,$7D,$53,$E0,$72,$DA,$DF,$BB,$86,$9F,$22,$3E,$37
	.dc.b	$9F,$23,$E8,$7D,$C5,$C3,$FC,$A1,$F8,$57,$D2,$9F,$B3,$E3,$FD,$B9
	.dc.b	$CF,$37,$F1,$58,$C7,$0A,$BF,$CD,$77,$C6,$FF,$DF,$7C,$FF,$C4,$E3
	.dc.b	$15,$FC,$9A,$31,$87,$36,$DF,$81,$FC,$97,$AB,$F9,$3F,$16,$CF,$E1
	.dc.b	$DB,$8F,$D4,$6E,$1F,$51,$FC,$D7,$37,$E6,$FA,$6D,$FE,$3D,$B7,$F7
	.dc.b	$1B,$7F,$71,$FD,$17,$CB,$FE,$8F,$DB,$B7,$F9,$76,$DF,$E1,$1B,$7F
	.dc.b	$84,$6C,$FF,$AB,$BA,$7F,$45,$18,$A4,$62,$7F,$DD,$BF,$30,$F5,$BF
	.dc.b	$88,$47,$FB,$93,$B3,$FA,$44,$E7,$C4,$FA,$51,$8A,$DE,$1F,$D3,$23
	.dc.b	$13,$AB,$FD,$77,$CB,$FE,$F7,$E1,$3F,$7C,$D8,$C9,$FB,$38,$CB,$4F
	.dc.b	$F6,$E7,$23,$E8,$7F,$AD,$43,$E8,$B0,$EA,$33,$18,$CB,$F1,$B1,$83
	.dc.b	$F0,$85,$9E,$07,$FB,$53,$7F,$C6,$4F,$E3,$6B,$94,$62,$17,$8A,$FF
	.dc.b	$6D,$9E,$3B,$FE,$09,$F1,$9F,$C1,$7E,$36,$CF,$F0,$A6,$DF,$F0,$1E
	.dc.b	$34,$63,$1F,$0F,$0A,$73,$7F,$CD,$FF,$57,$7C,$3F,$F7,$7F,$D4,$9F
	.dc.b	$4B,$FD,$02,$73,$5F,$C8,$A7,$2F,$1A,$E1,$5B,$6C,$7C,$13,$FD,$AB
	.dc.b	$EE,$FF,$1C,$CB,$E1,$C7,$CF,$FA,$0F,$EF,$D3,$98,$F8,$3C,$29,$CF
	.dc.b	$95,$F4,$9B,$3D,$77,$F9,$91,$87,$FD,$D7,$1D,$4D,$2B,$EC,$F0,$8C
	.dc.b	$FE,$75,$FC,$9E,$3F,$37,$FA,$6B,$E1,$BF,$C2,$53,$9E,$14,$E6,$F8
	.dc.b	$FA,$9F,$E1,$F8,$C6,$1F,$EE,$1B,$E1,$18,$D7,$FC,$5D,$C3,$FE,$09
	.dc.b	$AF,$E7,$1C,$2F,$98,$FA,$1E,$C7,$D3,$6E,$7F,$45,$B5,$C1,$FF,$13
	.dc.b	$3F,$EA,$F1,$F4,$0E,$B4,$4E,$FF,$99,$78,$AD,$4F,$FA,$03,$CD,$C5
	.dc.b	$F0,$85,$F0,$7A,$C7,$F9,$EA,$77,$1D,$78,$FF,$3F,$7D,$23,$FD,$6E
	.dc.b	$70,$35,$87,$07,$31,$BC,$3B,$3C,$63,$D8,$E6,$BA,$E1,$E1,$9E,$87
	.dc.b	$9A,$32,$FA,$CC,$D7,$49,$E1,$5C,$EF,$C5,$5D,$A7,$E0,$56,$E3,$95
	.dc.b	$6D,$EE,$6C,$F5,$98,$3F,$DB,$87,$82,$FC,$6F,$BA,$BA,$C6,$23,$91
	.dc.b	$E9,$BF,$1B,$CA,$BC,$07,$C3,$71,$5C,$A7,$78,$7C,$BC,$3D,$15,$F4
	.dc.b	$3E,$83,$D0,$71,$9F,$C1,$B8,$BF,$86,$73,$9F,$54,$F3,$C7,$CD,$1E
	.dc.b	$03,$DF,$8F,$37,$DD,$8F,$89,$A7,$22,$C6,$CF,$5D,$F8,$23,$9B,$5F
	.dc.b	$84,$71,$8D,$BF,$4A,$3C,$D1,$CD,$A7,$1C,$7E,$0C,$FC,$C6,$BE,$93
	.dc.b	$AC,$F1,$7C,$27,$52,$9F,$5C,$A7,$3B,$F8,$78,$7C,$5C,$3E,$05,$79
	.dc.b	$AB,$81,$DB,$0F,$25,$F0,$8F,$87,$18,$AE,$2F,$AA,$3C,$47,$12,$E1
	.dc.b	$EA,$C3,$F7,$7C,$8C,$BF,$44,$DA,$FD,$03,$5F,$D8,$B7,$E1,$3F,$8E
	.dc.b	$70,$72,$AF,$DF,$46,$2B,$F9,$34,$63,$95,$5B,$9F,$E1,$75,$BB,$E9
	.dc.b	$5A,$9F,$D9,$FD,$17,$F1,$B7,$F7,$CD,$7F,$6F,$5D,$F6,$FE,$0C,$DD
	.dc.b	$3F,$88,$56,$E7,$F8,$53,$63,$C9,$8F,$F0,$9A,$E1,$7F,$C7,$AB,$4D
	.dc.b	$DF,$F8,$D6,$1D,$B1,$F3,$37,$0F,$2C,$EB,$1F,$0B,$6F,$DF,$7C,$27
	.dc.b	$C3,$1E,$E9,$D2,$C3,$C2,$B1,$F2,$B6,$F9,$3E,$53,$C2,$7A,$DC,$BB
	.dc.b	$C3,$2F,$73,$4C,$70,$4C,$7D,$8F,$CE,$9E,$77,$E4,$3C,$B7,$E4,$3C
	.dc.b	$AE,$70,$F5,$C6,$67,$8C,$7B,$A7,$8C,$79,$A7,$8C,$E4,$E3,$19,$38
	.dc.b	$D7,$AA,$FC,$A4,$C7,$E7,$C6,$5F,$9B,$1E,$AC,$FF,$03,$7F,$D4,$DB
	.dc.b	$5F,$C6,$AB,$E9,$C7,$38,$F9,$87,$C3,$8E,$64,$73,$5C,$9F,$8D,$97
	.dc.b	$CB,$8C,$12,$78,$4F,$D3,$3A,$1C,$9C,$E3,$F0,$A3,$D4,$7E,$22,$BA
	.dc.b	$4F,$BA,$38,$9C,$EB,$D0,$FE,$27,$0F,$94,$F0,$39,$1C,$32,$FA,$37
	.dc.b	$98,$E0,$78,$0E,$97,$FE,$51,$F9,$C7,$CB,$FF,$7B,$AF,$F2,$1F,$1B
	.dc.b	$C3,$C7,$17,$F9,$F6,$31,$8F,$FB,$EF,$0E,$2D,$C3,$6F,$FA,$02,$76
	.dc.b	$FF,$86,$AF,$38,$2F,$8C,$FF,$0A,$C3,$FC,$19,$D5,$FE,$4E,$8F,$F7
	.dc.b	$6F,$CE,$BE,$13,$AF,$62,$31,$3F,$E3,$AA,$E0,$F3,$3F,$C3,$AE,$31
	.dc.b	$FF,$2F,$CF,$09,$FF,$72,$36,$3F,$CE,$13,$BC,$3F,$C3,$DF,$97,$BC
	.dc.b	$9F,$6E,$0F,$C6,$DE,$76,$1B,$1F,$E4,$CA,$E0,$DA,$D5,$48,$C1,$0F
	.dc.b	$41,$D5,$D3,$89,$E6,$FB,$71,$F9,$5E,$AF,$C4,$E0,$7D,$25,$3E,$7F
	.dc.b	$02,$69,$9B,$FF,$3C,$4E,$1F,$F3,$46,$EF,$D3,$DE,$8F,$D0,$5F,$5B
	.dc.b	$4C,$44,$AE,$33,$A6,$CE,$5F,$6F,$A6,$FA,$8F,$91,$ED,$49,$FF,$39
	.dc.b	$CE,$5B,$39,$7E,$2F,$12,$BE,$9D,$6E,$7A,$D6,$F7,$FF,$55,$5E,$E7
	.dc.b	$99,$EE,$F7,$BF,$D7,$CD,$F9,$6F,$C6,$C1,$BE,$8C,$9D,$10,$FC,$76
	.dc.b	$CA,$E7,$1F,$91,$F1,$7F,$D1,$9E,$8F,$E6,$3D,$97,$C7,$53,$F4,$F3
	.dc.b	$AA,$FF,$72,$F8,$EB,$D4,$7D,$4B,$E2,$EA,$3C,$FF,$09,$E8,$F2,$A4
	.dc.b	$FB,$63,$2F,$BE,$32,$FA,$32,$4E,$A3,$13,$AD,$FF,$D8,$B5,$A9,$FF
	.dc.b	$5F,$7D,$17,$C6,$61,$82,$1D,$4C,$9F,$E9,$A7,$F2,$DF,$37,$F9,$7F
	.dc.b	$AF,$93,$FE,$C9,$8C,$BF,$EC,$B8,$CB,$FE,$A2,$8F,$D8,$FE,$42,$F8
	.dc.b	$4F,$99,$F1,$92,$BF,$D7,$73,$96,$CE,$5F,$AF,$B2,$B9,$CC,$9F,$9D
	.dc.b	$1B,$CF,$FD,$E5,$7C,$E7,$EF,$F5,$7E,$CA,$7F,$DD,$2D,$F0,$FF,$D5
	.dc.b	$7C,$9F,$A9,$F0,$FF,$69,$E3,$7F,$08,$FC,$35,$F1,$1F,$AB,$F8,$19
	.dc.b	$9F,$FB,$4A,$72,$FF,$BA,$FE,$BF,$F7,$84,$7F,$5B,$3F,$BF,$FB,$0E
	.dc.b	$9F,$0B,$27,$FD,$E5,$15,$FF,$64,$FC,$C7,$F0,$53,$1B,$FF,$3E,$71
	.dc.b	$1F,$EF,$97,$3B,$7F,$AF,$7F,$09,$FC,$7C,$FF,$14,$6D,$27,$5D,$DF
	.dc.b	$CD,$3D,$9F,$E7,$F3,$9B,$FF,$66,$DF,$0B,$F8,$BC,$4F,$91,$FE,$18
	.dc.b	$7F,$2C,$9D,$57,$0C,$72,$FF,$0C,$9D,$46,$F9,$64,$FF,$BE,$AB,$EE
	.dc.b	$3F,$C6,$BC,$C7,$C9,$EA,$3F,$DD,$7A,$7F,$27,$31,$A7,$9E,$17,$C9
	.dc.b	$D9,$CE,$77,$A7,$FB,$87,$E8,$1F,$41,$FE,$E3,$A9,$18,$70,$7F,$BB
	.dc.b	$9C,$38,$6B,$C2,$3F,$A6,$FD,$0F,$F3,$A8,$FF,$7A,$E3,$E5,$8C,$47
	.dc.b	$F4,$6A,$E7,$3E,$96,$DF,$FB,$D6,$B7,$3C,$7C,$EF,$EB,$35,$AB,$FF
	.dc.b	$80,$3F,$81,$3E,$47,$FA,$EF,$37,$F2,$19,$F0,$E5,$FE,$FF,$8C,$47
	.dc.b	$F5,$6A,$E0,$E8,$E7,$5C,$BC,$A9,$19,$13,$C1,$E5,$3A,$7F,$B0,$D7
	.dc.b	$6C,$BE,$A5,$6D,$FA,$DF,$44,$F0,$BF,$C1,$FF,$33,$FC,$0F,$BE,$F8
	.dc.b	$5F,$B3,$C8,$2F,$83,$E3,$6B,$FE,$07,$AE,$4F,$AD,$AF,$F7,$48,$DE
	.dc.b	$1E,$2C,$3B,$E1,$2B,$A4,$7F,$81,$A9,$F1,$1E,$4D,$65,$78,$63,$46
	.dc.b	$1F,$16,$5F,$07,$6F,$4C,$7A,$67,$FB,$F7,$CE,$FF,$7F,$FA,$C9,$3F
	.dc.b	$9F,$8C,$BF,$A1,$8C,$BF,$D5,$BA,$13,$FE,$1A,$F9,$E7,$C8,$BC,$FF
	.dc.b	$8F,$F0,$FF,$E3,$89,$CC,$63,$DF,$F1,$C6,$31,$57,$F8,$FB,$E7,$DE
	.dc.b	$1F,$A2,$39,$63,$D6,$B7,$87,$28,$C4,$EB,$1E,$D5,$F6,$30,$E6,$E0
	.dc.b	$63,$E1,$AE,$11,$D0,$FB,$F1,$9C,$3E,$4D,$72,$9D,$1C,$B0,$F0,$CF
	.dc.b	$8F,$0E,$D7,$C6,$BC,$58,$FA,$EB,$51,$CA,$74,$F3,$AF,$36,$1D,$F1
	.dc.b	$F2,$92,$F8,$BF,$1F,$0E,$97,$E8,$C7,$C6,$DC,$3C,$71,$88,$61,$D7
	.dc.b	$0F,$1C,$7F,$11,$C7,$EA,$36,$BE,$83,$6F,$F8,$9B,$FE,$D1,$C9,$CB
	.dc.b	$8F,$86,$FC,$2B,$63,$33,$FC,$6A,$72,$FF,$1C,$C3,$FC,$37,$E9,$74
	.dc.b	$6C,$67,$C3,$FF,$04,$3E,$56,$BF,$8A,$7F,$C1,$9B,$7E,$3D,$BA,$7E
	.dc.b	$83,$E7,$FF,$D0,$1C,$7F,$52,$73,$7F,$79,$AF,$E1,$1B,$3C,$23,$EB
	.dc.b	$46,$B0,$F2,$E1,$DF,$1F,$C6,$5E,$E3,$17,$C2,$7C,$06,$BC,$CF,$EB
	.dc.b	$93,$A2,$65,$EB,$9D,$65,$C2,$F9,$12,$BF,$D4,$78,$77,$BE,$13,$96
	.dc.b	$CE,$B0,$EF,$18,$3F,$C1,$BE,$7F,$F7,$1C,$3F,$C8,$9F,$3F,$FD,$BF
	.dc.b	$B3,$FB,$87,$9E,$AD,$65,$FE,$20,$AD,$5F,$22,$C2,$B8,$4E,$1E,$53
	.dc.b	$69,$3C,$2B,$FC,$55,$E3,$FF,$69,$7F,$8B,$F9,$1F,$DC,$2F,$F1,$ED
	.dc.b	$8F,$EC,$6D,$7F,$B4,$39,$C7,$FB,$93,$67,$F9,$0F,$C5,$BE,$47,$C1
	.dc.b	$39,$BF,$E7,$AB,$C9,$F5,$F3,$3E,$7F,$FC,$1D,$8F,$F5,$1A,$D3,$FC
	.dc.b	$F6,$33,$7F,$EB,$09,$CB,$E6,$BF,$63,$D9,$AF,$F9,$25,$B1,$C3,$2F
	.dc.b	$4C,$7C,$5C,$9F,$2B,$E8,$8F,$91,$F3,$CF,$92,$FA,$3C,$CF,$9F,$FF
	.dc.b	$78,$6B,$FD,$9E,$B4,$FF,$6D,$9D,$72,$72,$BF,$1D,$6B,$5F,$F0,$7B
	.dc.b	$5F,$F0,$A3,$5F,$E9,$6E,$F4,$FF,$10,$37,$0F,$EC,$FE,$EB,$EA,$DF
	.dc.b	$53,$DE,$7F,$BD,$7D,$03,$E8,$3F,$13,$8B,$E8,$5F,$43,$F8,$5A,$D5
	.dc.b	$FF,$81,$6B,$53,$9B,$F9,$75,$AD,$3F,$C4,$ED,$7F,$C5,$8D,$7F,$B1
	.dc.b	$BB,$AF,$F1,$C3,$6F,$D3,$97,$FB,$89,$F3,$DF,$E5,$FA,$DF,$0F,$F9
	.dc.b	$DF,$82,$FF,$39,$D7,$F8,$56,$72,$FF,$B1,$7D,$FE,$15,$F8,$09,$CE
	.dc.b	$7F,$E3,$C6,$BF,$E4,$16,$BF,$DE,$1D,$E7,$FE,$4D,$6D,$FF,$A1,$B0
	.dc.b	$7B,$BD,$A3,$C5,$E4,$7E,$C7,$B5,$F1,$FE,$98,$C3,$67,$F9,$DE,$1F
	.dc.b	$DE,$9B,$3E,$72,$3D,$0F,$F7,$EF,$CF,$FD,$FC,$FF,$D2,$F3,$F0,$F6
	.dc.b	$95,$FE,$37,$F1,$3F,$D5,$CF,$FB,$9E,$3F,$80,$46,$2B,$FC,$BA,$D7
	.dc.b	$FC,$C4,$D7,$FC,$05,$E8,$7E,$AB,$1F,$F7,$37,$AF,$CA,$3E,$93,$70
	.dc.b	$FF,$7A,$7C,$57,$D5,$3E,$17,$2D,$8F,$27,$C6,$FE,$CB,$23,$49,$1F
	.dc.b	$57,$77,$D0,$FC,$B3,$F0,$1F,$43,$67,$FC,$A9,$F4,$1F,$49,$FF,$77
	.dc.b	$73,$7F,$1D,$8C,$6F,$FE,$C8,$8C,$72,$7F,$02,$9C,$CF,$FA,$05,$AF
	.dc.b	$F8,$E1,$AF,$F9,$49,$E7,$CD,$FE,$8F,$6D,$FE,$0F,$E3,$FA,$BF,$E6
	.dc.b	$BE,$27,$3B,$E4,$FF,$31,$F2,$3F,$99,$6D,$6E,$74,$D9,$D3,$4D,$74
	.dc.b	$7E,$87,$83,$F4,$4D,$8C,$9C,$E3,$2D,$3B,$7B,$FF,$A9,$F7,$BF,$54
	.dc.b	$D8,$C9,$E3,$8C,$B4,$CE,$CD,$BF,$AB,$E4,$AD,$CE,$6B,$F4,$D1,$FD
	.dc.b	$5B,$C1,$7D,$0F,$F2,$6A,$FF,$7D,$F6,$66,$31,$8F,$F6,$2F,$43,$F4
	.dc.b	$95,$CC,$5F,$2F,$1B,$81,$FC,$0F,$E1,$AE,$7F,$2F,$E5,$FF,$83,$79
	.dc.b	$9F,$C1,$DB,$19,$3E,$84,$65,$A7,$E3,$3E,$47,$FB,$6A,$F9,$F4,$7E
	.dc.b	$34,$CB,$88,$E0,$7F,$8D,$BA,$31,$9F,$FB,$29,$AF,$E3,$EB,$F2,$47
	.dc.b	$39,$FF,$59,$1D,$FE,$9F,$13,$F0,$7A,$3F,$B5,$7C,$9F,$F0,$DF,$A2
	.dc.b	$FE,$1C,$DB,$FC,$87,$3A,$31,$7C,$63,$13,$AA,$FE,$83,$E9,$7F,$A7
	.dc.b	$A3,$FD,$E1,$F4,$58,$F8,$4F,$F0,$57,$C1,$A3,$13,$FD,$0E,$B8,$63
	.dc.b	$D7,$C9,$CB,$F5,$CF,$F4,$A7,$97,$F1,$BE,$57,$FC,$21,$F0,$5F,$C9
	.dc.b	$30,$F7,$F1,$A7,$35,$FD,$DA,$B5,$7F,$EF,$26,$BF,$EF,$76,$BF,$EA
	.dc.b	$E7,$7D,$4C,$31,$5F,$5B,$D5,$EB,$1F,$ED,$BF,$A2,$FE,$55,$C6,$FA
	.dc.b	$27,$D1,$7F,$89,$67,$31,$8A,$E1,$EC,$64,$FF,$17,$43,$E8,$B1,$3F
	.dc.b	$17,$DB,$F9,$3F,$43,$F3,$9B,$19,$26,$FF,$57,$C4,$D4,$7F,$BB,$E3
	.dc.b	$FD,$76,$75,$EF,$FF,$65,$F8,$5F,$1E,$7F,$DC,$B8,$A7,$F9,$DD,$6C
	.dc.b	$FF,$01,$F1,$FC,$37,$07,$FA,$F4,$7D,$17,$17,$77,$FE,$F0,$8C,$46
	.dc.b	$DB,$87,$86,$38,$DF,$82,$73,$7D,$DE,$33,$EA,$73,$7D,$6B,$81,$EC
	.dc.b	$AE,$0E,$2B,$C6,$76,$3F,$AA,$9E,$AB,$F1,$61,$DA,$3F,$DB,$66,$6B
	.dc.b	$E0,$46,$7D,$2F,$04,$F4,$C7,$0F,$53,$F8,$6D,$78,$9E,$F5,$E8,$7E
	.dc.b	$49,$F5,$1C,$1C,$A7,$45,$C3,$D9,$8F,$82,$BE,$54,$61,$F0,$1F,$15
	.dc.b	$C4,$F1,$8F,$34,$7C,$37,$CB,$3C,$DF,$14,$73,$AF,$59,$E4,$78,$63
	.dc.b	$F5,$EF,$A5,$72,$3C,$27,$C8,$7D,$33,$C3,$0F,$85,$3F,$87,$BE,$8F
	.dc.b	$C4,$9F,$01,$BC,$3C,$91,$CE,$3D,$66,$5E,$F8,$FC,$3C,$7C,$93,$F9
	.dc.b	$8B,$EB,$5D,$E3,$AB,$F8,$0C,$7C,$AD,$8E,$AD,$3E,$39,$4F,$74,$7C
	.dc.b	$0A,$EF,$19,$6B,$DC,$FC,$7D,$7C,$AA,$F0,$C7,$D8,$3A,$94,$EA,$FA
	.dc.b	$4D,$BF,$50,$E6,$E5,$A7,$C9,$29,$FB,$3E,$5F,$E1,$4F,$E7,$9A,$E6
	.dc.b	$7F,$7B,$18,$F8,$7F,$E2,$FC,$CA,$4E,$78,$FF,$BB,$46,$2F,$4A,$F8
	.dc.b	$B9,$39,$38,$AE,$57,$FC,$3B,$0E,$3F,$44,$E1,$E7,$8F,$F0,$68,$C6
	.dc.b	$3B,$39,$E5,$E2,$9F,$E1,$79,$78,$1E,$17,$CB,$0E,$AD,$BE,$38,$F3
	.dc.b	$7D,$18,$7B,$34,$E3,$1F,$AD,$CB,$C2,$FE,$EF,$2F,$6C,$F8,$2B,$C5
	.dc.b	$8F,$6A,$E9,$7C,$32,$D6,$3D,$F1,$E3,$7B,$BF,$E0,$D7,$DB,$1D,$C7
	.dc.b	$7C,$7C,$95,$B9,$F2,$CE,$70,$FE,$05,$87,$48,$FB,$18,$73,$BE,$07
	.dc.b	$D6,$9D,$4F,$C0,$AF,$3E,$5F,$16,$17,$E9,$BE,$18,$FB,$A7,$E4,$9E
	.dc.b	$8A,$FC,$0E,$3E,$43,$D2,$4A,$F0,$CF,$92,$77,$1D,$30,$F4,$DF,$5A
	.dc.b	$F0,$D7,$8F,$0F,$14,$78,$32,$E8,$F3,$3E,$B4,$F4,$7C,$B8,$E9,$E6
	.dc.b	$68,$EF,$84,$7F,$AC,$DF,$67,$67,$F5,$CC,$BF,$19,$97,$9E,$3E,$01
	.dc.b	$CC,$FA,$44,$BF,$85,$1F,$02,$3E,$71,$97,$15,$F5,$63,$CF,$1D,$71
	.dc.b	$FB,$A4,$70,$78,$6B,$AC,$FC,$13,$A1,$F1,$63,$F1,$93,$C6,$30,$76
	.dc.b	$38,$4F,$27,$CA,$FC,$7A,$F2,$47,$C2,$8F,$95,$7F,$88,$9C,$9A,$7F
	.dc.b	$16,$7C,$69,$F2,$57,$85,$F8,$2F,$F8,$0F,$07,$CF,$BE,$8B,$F8,$4C
	.dc.b	$7F,$82,$CC,$B8,$E0,$FE,$25,$18,$BF,$F2,$9F,$C3,$65,$FF,$44,$5F
	.dc.b	$37,$FD,$4D,$DE,$F2,$75,$D4,$8C,$4F,$4C,$7A,$DF,$18,$FF,$27,$76
	.dc.b	$7F,$95,$5D,$1F,$C8,$DC,$3A,$F0,$3A,$FD,$13,$E7,$BC,$93,$FE,$F6
	.dc.b	$9C,$FC,$4C,$8E,$66,$4B,$F4,$5F,$E9,$F6,$FA,$1C,$A3,$11,$FE,$61
	.dc.b	$FA,$17,$A9,$9A,$FE,$63,$32,$B6,$D8,$CB,$FC,$DA,$32,$D8,$F3,$D7
	.dc.b	$FB,$2E,$3C,$AD,$7F,$CE,$13,$E2,$E2,$FA,$DA,$FF,$BB,$A7,$E2,$C7
	.dc.b	$DC,$F7,$3E,$A3,$63,$FD,$B1,$F3,$8F,$80,$CD,$F1,$9F,$F4,$97,$13
	.dc.b	$E6,$7E,$BB,$63,$AC,$EF,$0F,$F2,$87,$23,$17,$B3,$F7,$97,$B7,$4C
	.dc.b	$44,$C3,$C7,$19,$7C,$B1,$F6,$3D,$D7,$D1,$52,$4F,$C4,$AA,$FC,$C9
	.dc.b	$DB,$F3,$83,$87,$53,$97,$2F,$D0,$F8,$29,$1F,$84,$9E,$19,$7F,$B8
	.dc.b	$BD,$6F,$CC,$FB,$8F,$96,$A6,$22,$5F,$FB,$FE,$32,$FF,$C0,$18,$FD
	.dc.b	$37,$91,$FD,$EB,$D7,$BE,$4F,$FA,$EE,$BF,$3B,$39,$7F,$3F,$C0,$C5
	.dc.b	$EC,$70,$C3,$F5,$7E,$E3,$E4,$A9,$9C,$38,$DB,$D4,$F4,$F9,$F2,$73
	.dc.b	$1E,$C3,$0F,$FB,$4E,$7F,$D7,$0E,$88,$66,$77,$BF,$E0,$A7,$B4,$FF
	.dc.b	$00,$FA,$2F,$8E,$C3,$04,$3B,$19,$3F,$22,$FE,$43,$E6,$7F,$23,$E1
	.dc.b	$C9,$FE,$17,$19,$7F,$86,$C6,$5F,$CC,$C7,$F0,$0F,$A2,$7B,$34,$93
	.dc.b	$FC,$53,$DB,$AF,$ED,$27,$6F,$E2,$5A,$E6,$7F,$11,$18,$CB,$F4,$3B
	.dc.b	$3E,$43,$E9,$9D,$23,$F4,$BD,$9F,$B7,$F2,$BF,$C7,$5C,$9F,$A7,$F1
	.dc.b	$7F,$4C,$7E,$BB,$C7,$FE,$2F,$F4,$4F,$8B,$FC,$B3,$76,$FC,$E3,$5F
	.dc.b	$E1,$6D,$7F,$C8,$37,$FC,$B7,$E8,$BF,$72,$48,$FE,$03,$F0,$DF,$BC
	.dc.b	$FA,$2E,$FF,$13,$27,$F9,$BC,$57,$AC,$78,$88,$FF,$1A,$BF,$E7,$5F
	.dc.b	$37,$FC,$18,$91,$FC,$B3,$15,$7F,$32,$FA,$2F,$CB,$4F,$F3,$06,$AE
	.dc.b	$0F,$F6,$D2,$8C,$C7,$F0,$EF,$7F,$FA,$3C,$EF,$83,$FC,$BF,$5C,$A3
	.dc.b	$F9,$8B,$5C,$F6,$FE,$FE,$31,$1F,$E6,$B7,$AC,$FD,$4E,$27,$C4,$FF
	.dc.b	$19,$3D,$17,$E9,$E4,$98,$7A,$5F,$E4,$BF,$31,$88,$FE,$7B,$8F,$E0
	.dc.b	$DF,$E2,$4D,$73,$3E,$33,$F9,$6E,$52,$3A,$1F,$D5,$3E,$81,$F4,$5F
	.dc.b	$D6,$9B,$19,$8F,$EB,$BF,$45,$FE,$84,$8C,$1F,$C8,$67,$6F,$F5,$06
	.dc.b	$BF,$D5,$9A,$FF,$A2,$1D,$ED,$F4,$A7,$FA,$04,$FC,$F3,$FD,$24,$E2
	.dc.b	$3E,$BC,$F0,$DB,$FD,$23,$F4,$5F,$D3,$6F,$FB,$6B,$5E,$41,$C5,$7E
	.dc.b	$03,$97,$F8,$EC,$F0,$F4,$BF,$9B,$57,$18,$FE,$DA,$D7,$BC,$68,$FE
	.dc.b	$FF,$3F,$CD,$63,$FB,$EB,$5F,$0C,$F8,$2F,$FD,$4F,$F4,$4F,$03,$FC
	.dc.b	$0E,$DA,$FF,$02,$B5,$F8,$5F,$3A,$FA,$FC,$67,$F9,$8A,$D5,$7F,$80
	.dc.b	$9A,$F6,$BF,$EA,$11,$88,$FF,$5E,$3C,$B0,$FF,$5F,$61,$FC,$63,$BB
	.dc.b	$FC,$58,$DA,$F1,$DF,$9B,$C2,$91,$E6,$77,$93,$AB,$33,$FE,$10,$C9
	.dc.b	$5E,$22,$47,$EE,$FD,$47,$D0,$FF,$80,$3E,$8E,$4F,$F8,$EA,$3F,$C0
	.dc.b	$66,$1F,$F0,$3C,$61,$FF,$15,$FD,$13,$E2,$FF,$29,$F1,$7F,$B8,$8C
	.dc.b	$7A,$3F,$ED,$B8,$C4,$EB,$E0,$BF,$DB,$3F,$47,$E9,$C3,$51,$FE,$DC
	.dc.b	$C3,$8C,$2F,$8E,$1A,$8F,$F7,$34,$EA,$7F,$83,$DF,$39,$E1,$7C,$F0
	.dc.b	$F0,$47,$06,$AB,$E5,$3E,$4D,$BF,$DD,$B8,$78,$2B,$99,$C7,$1F,$13
	.dc.b	$C7,$0E,$78,$F8,$6B,$81,$30,$F1,$5F,$09,$D4,$62,$FA,$9C,$E7,$AE
	.dc.b	$3E,$48,$0A,$E2,$F8,$CE,$31,$B7,$91,$D2,$36,$E0,$FF,$2C,$C7,$CE
	.dc.b	$6E,$1F,$C3,$5B,$5F,$E2,$DF,$87,$FC,$3F,$15,$D4,$D2,$57,$F9,$83
	.dc.b	$0F,$F0,$B7,$A3,$FE,$35,$6C,$67,$C1,$FF,$04,$BE,$26,$BF,$81,$7F
	.dc.b	$C1,$BA,$FE,$2D,$BA,$FC,$29,$7D,$1B,$CB,$74,$7C,$8B,$DD,$7E,$19
	.dc.b	$AF,$DF,$6D,$71,$9E,$19,$4A,$F9,$75,$AB,$EB,$18,$9D,$63,$E5,$7F
	.dc.b	$07,$87,$39,$D1,$27,$4F,$DE,$8D,$46,$09,$8F,$C1,$AE,$77,$E0,$9C
	.dc.b	$DF,$AE,$7C,$35,$F0,$EF,$9B,$C6,$7F,$2F,$E4,$3C,$CF,$8B,$9F,$F9
	.dc.b	$39,$7C,$37,$FB,$06,$3F,$C2,$F1,$AA,$FE,$85,$5C,$6B,$50,$8E,$95
	.dc.b	$87,$4F,$53,$15,$D9,$FF,$58,$F8,$8F,$1B,$9D,$FF,$8A,$DB,$F2,$4F
	.dc.b	$94,$CD,$6E,$F8,$D6,$0D,$CD,$8E,$BE,$3F,$03,$9F,$CB,$58,$C9,$31
	.dc.b	$FE,$AB,$8A,$75,$1F,$D6,$31,$5E,$3C,$2F,$F0,$BD,$EF,$0E,$51,$FE
	.dc.b	$1A,$FA,$17,$C9,$FB,$1B,$3F,$E7,$EF,$9E,$7D,$0E,$6A,$4F,$FB,$52
	.dc.b	$B7,$9F,$F7,$1F,$75,$87,$0A,$D5,$F0,$AF,$ED,$B7,$C3,$1F,$CA,$E1
	.dc.b	$C3,$1D,$47,$F8,$BB,$E8,$1F,$41,$FE,$30,$C3,$F5,$51,$8C,$BF,$D5
	.dc.b	$4D,$8F,$F6,$CB,$5F,$F6,$FB,$9A,$E9,$5F,$84,$C3,$CF,$F1,$1F,$72
	.dc.b	$B5,$87,$6A,$C1,$BD,$3F,$DA,$11,$FE,$F4,$9D,$E3,$FE,$68,$F8,$3F
	.dc.b	$C6,$7B,$DF,$E6,$5A,$FF,$7C,$F3,$7F,$AF,$9F,$F4,$FF,$B5,$F8,$A8
	.dc.b	$C5,$FF,$84,$1B,$AF,$FA,$5F,$D7,$D6,$33,$0E,$6F,$68,$C5,$F1,$F1
	.dc.b	$7F,$39,$CE,$F8,$FE,$D1,$86,$E1,$E4,$8F,$D1,$7C,$6F,$D3,$8F,$F3
	.dc.b	$AC,$6A,$FA,$47,$FA,$FA,$75,$1F,$E8,$69,$F7,$F0,$4A,$FE,$7F,$9F
	.dc.b	$F9,$9B,$C0,$FF,$4E,$C7,$FA,$FA,$7F,$53,$18,$BF,$F1,$7B,$6B,$FD
	.dc.b	$6D,$E9,$EB,$DE,$F9,$6D,$C3,$FD,$8E,$DB,$FF,$61,$7C,$37,$D0,$3E
	.dc.b	$0F,$D0,$5A,$F1,$9F,$D9,$3E,$33,$C2,$7C,$6F,$A0,$B1,$FE,$DA,$E0
	.dc.b	$46,$63,$FD,$AB,$F4,$17,$1F,$F2,$2F,$CF,$7F,$A2,$27,$AC,$7F,$5E
	.dc.b	$8C,$3F,$ED,$5F,$8D,$FE,$0F,$18,$DB,$FD,$12,$DA,$FE,$88,$DA,$FE
	.dc.b	$2D,$D0,$BE,$3C,$B2,$FB,$D2,$F8,$73,$FF,$BC,$6F,$53,$FE,$E5,$FA
	.dc.b	$0F,$C9,$FC,$7F,$E5,$1B,$87,$FB,$C1,$B3,$B8,$FF,$79,$CF,$E6,$B9
	.dc.b	$FF,$37,$BD,$9F,$F7,$BB,$6B,$FD,$E4,$DD,$7F,$DD,$CF,$E8,$BD,$DF
	.dc.b	$D1,$FB,$96,$FF,$DF,$ED,$BF,$F7,$D3,$6F,$FD,$EC,$D9,$FF,$69,$EE
	.dc.b	$8C,$57,$C6,$7F,$75,$F4,$0F,$32,$9F,$B2,$F8,$3B,$87,$D6,$F0,$D5
	.dc.b	$B8,$C3,$70,$FE,$93,$8F,$FA,$BA,$73,$87,$EF,$3E,$6B,$F7,$BE,$5D
	.dc.b	$9F,$C0,$F1,$AF,$F9,$76,$1C,$AF,$75,$A9,$46,$27,$F1,$7F,$1F,$F8
	.dc.b	$63,$FA,$BD,$3E,$7D,$F2,$19,$8F,$C0,$F3,$AB,$17,$FE,$A1,$9E,$15
	.dc.b	$8E,$1F,$C7,$D5,$9F,$F1,$99,$A3,$9B,$FC,$23,$E4,$3F,$84,$FD,$05
	.dc.b	$9F,$D2,$37,$0F,$ED,$BC,$A8,$C6,$1F,$E0,$ED,$BF,$B1,$47,$F7,$08
	.dc.b	$79,$DF,$E7,$CC,$3F,$4F,$18,$F8,$1F,$E2,$FE,$1E,$EB,$59,$7F,$AE
	.dc.b	$EB,$57,$C2,$73,$1D,$F2,$FE,$11,$3F,$C4,$27,$2F,$F8,$BF,$DD,$FE
	.dc.b	$44,$FE,$A1,$B7,$F1,$F8,$1E,$BF,$F7,$C9,$CE,$3F,$EF,$E6,$DF,$3F
	.dc.b	$27,$EA,$DF,$08,$E9,$1C,$48,$FF,$04,$3E,$D5,$E3,$0F,$F6,$6F,$D0
	.dc.b	$7F,$2A,$4F,$F2,$CD,$DF,$40,$F9,$CF,$F0,$7C,$EC,$9E,$A7,$F8,$66
	.dc.b	$75,$8F,$F0,$49,$FE,$6B,$AF,$D0,$F5,$7E,$61,$FD,$13,$CE,$B5,$F2
	.dc.b	$E2,$E7,$F7,$26,$D7,$F6,$F8,$FF,$4D,$0F,$9E,$FC,$CB,$FE,$7C,$EF
	.dc.b	$7C,$AF,$F9,$3A,$73,$3F,$E9,$23,$87,$67,$F9,$86,$B7,$97,$FA,$12
	.dc.b	$B7,$5F,$EA,$F3,$67,$5B,$E9,$7F,$E6,$98,$D4,$F6,$BE,$71,$C0,$F0
	.dc.b	$46,$6B,$8C,$71,$F3,$FC,$17,$E3,$9C,$CF,$08,$C6,$1A,$8E,$07,$5C
	.dc.b	$3B,$E1,$E8,$8F,$8A,$F7,$9E,$11,$CC,$F5,$38,$7D,$35,$D9,$F4,$18
	.dc.b	$36,$60,$EA,$EE,$7C,$18,$76,$BF,$1C,$FC,$5A,$D9,$CA,$B6,$FA,$70
	.dc.b	$F7,$BD,$C5,$72,$7D,$F3,$EB,$7C,$57,$DE,$B8,$C7,$63,$8B,$88,$D6
	.dc.b	$1E,$F9,$D5,$F5,$7E,$21,$C2,$35,$3C,$23,$B9,$E7,$8C,$B4,$CD,$F9
	.dc.b	$4F,$69,$70,$F8,$55,$CE,$33,$1F,$24,$CC,$F3,$8C,$BF,$1E,$33,$1C
	.dc.b	$E3,$E1,$C7,$16,$9B,$29,$EB,$3C,$AE,$DF,$7D,$FC,$17,$E8,$35,$F8
	.dc.b	$B8,$7C,$97,$C9,$3E,$23,$93,$C2,$35,$5E,$58,$EB,$18,$9D,$C7,$4C
	.dc.b	$3E,$4D,$6E,$3F,$5D,$C9,$FC,$35,$FC,$EB,$5F,$DC,$35,$DF,$07,$FB
	.dc.b	$67,$E0,$9F,$40,$D7,$FD,$FB,$7C,$67,$31,$8A,$E1,$5B,$B5,$EE,$30
	.dc.b	$DA,$FD,$4F,$CF,$BE,$83,$FD,$D1,$97,$F0,$C8,$C6,$1F,$AE,$CB,$BB
	.dc.b	$FC,$05,$B3,$BA,$E3,$A7,$FB,$C6,$F8,$DE,$F2,$F2,$5F,$F1,$8C,$38
	.dc.b	$DE,$F0,$E3,$61,$87,$2B,$E1,$87,$2B,$F7,$E3,$EA,$C3,$8E,$1C,$1B
	.dc.b	$8F,$68,$E8,$D7,$47,$13,$DD,$7D,$5D,$B5,$D9,$FC,$87,$1F,$26,$1E
	.dc.b	$28,$F7,$61,$CD,$D6,$1E,$2B,$E5,$87,$87,$0E,$35,$EF,$6A,$8F,$31
	.dc.b	$BA,$F6,$CE,$A3,$35,$E5,$8E,$26,$5E,$E6,$67,$89,$E3,$BF,$79,$E7
	.dc.b	$C7,$ED,$61,$83,$D6,$7A,$8D,$63,$EF,$8F,$A7,$38,$9F,$55,$75,$78
	.dc.b	$CF,$47,$91,$8A,$E4,$7E,$05,$E4,$7C,$1B,$F4,$DE,$30,$F6,$3E,$5B
	.dc.b	$E8,$4B,$E7,$7F,$62,$78,$61,$E3,$AF,$3C,$F8,$9E,$0E,$CF,$21,$EA
	.dc.b	$3C,$78,$FC,$EA,$E7,$1E,$1B,$EE,$FE,$12,$B9,$BE,$99,$D3,$CC,$F0
	.dc.b	$4F,$9A,$1F,$36,$FA,$0F,$D0,$57,$F8,$CE,$30,$FF,$8E,$9A,$FF,$91
	.dc.b	$3E,$07,$74,$DF,$FD,$87,$5A,$8C,$57,$F0,$98,$C4,$EA,$F8,$BF,$AF
	.dc.b	$E8,$76,$B1,$CE,$1F,$B3,$EB,$7C,$FE,$23,$FD,$C0,$D8,$CC,$7F,$9D
	.dc.b	$BE,$81,$F3,$79,$A9,$3F,$EA,$CA,$DB,$FC,$77,$2F,$15,$F0,$8F,$F5
	.dc.b	$8F,$CF,$FF,$A4,$DB,$E7,$78,$63,$11,$C4,$FE,$11,$F4,$0F,$47,$55
	.dc.b	$27,$89,$FE,$BC,$9D,$38,$39,$0A,$EB,$C3,$F2,$A3,$15,$F0,$6B,$8C
	.dc.b	$7D,$7F,$63,$E3,$D7,$DC,$6F,$11,$F9,$7C,$2F,$99,$7B,$9A,$99,$87
	.dc.b	$DF,$77,$3A,$8F,$F5,$0A,$F5,$73,$EB,$C9,$CC,$70,$30,$FC,$5C,$3C
	.dc.b	$13,$98,$FF,$6E,$FB,$CF,$A0,$FB,$8D,$8C,$C7,$BE,$72,$C8,$C9,$F2
	.dc.b	$E3,$0F,$CA,$9E,$4D,$9E,$4F,$FA,$D7,$D0,$F8,$7E,$9B,$E8,$29,$88
	.dc.b	$93,$FE,$BB,$8C,$BC,$4F,$F6,$77,$84,$F4,$BF,$2F,$E3,$7F,$60,$6C
	.dc.b	$66,$7F,$D8,$9E,$9E,$26,$47,$27,$C7,$1E,$4F,$3B,$CB,$E4,$4A,$FF
	.dc.b	$66,$46,$5F,$F6,$74,$65,$FF,$6D,$C6,$5F,$F6,$EC,$67,$97,$FD,$71
	.dc.b	$A7,$FA,$8A,$7F,$DB,$7F,$26,$FA,$0F,$87,$10,$CB,$F8,$88,$D4,$7F
	.dc.b	$B7,$4D,$1F,$75,$CB,$7E,$5F,$87,$83,$23,$BC,$73,$30,$F1,$8C,$3F
	.dc.b	$DC,$70,$FD,$07,$D0,$3D,$5F,$E0,$0D,$9D,$78,$3F,$C6,$23,$11,$E3
	.dc.b	$F5,$B7,$52,$36,$7F,$7D,$9D,$DF,$FB,$6F,$25,$EE,$BF,$67,$5A,$C3
	.dc.b	$B6,$1F,$0E,$BF,$4B,$DF,$2F,$F4,$AF,$C7,$F8,$FF,$E2,$DF,$3C,$F8
	.dc.b	$7A,$4A,$D9,$F4,$BE,$83,$E7,$11,$FD,$7F,$C0,$FD,$1F,$A0,$E7,$F0
	.dc.b	$92,$7E,$CC,$57,$F8,$4E,$09,$C7,$07,$B6,$BE,$D7,$CC,$4B,$FC,$0B
	.dc.b	$AD,$FF,$A0,$FD,$03,$E5,$7F,$97,$36,$BE,$48,$70,$FE,$F7,$DB,$7C
	.dc.b	$27,$0A,$93,$C2,$B7,$97,$F4,$4F,$15,$87,$0A,$D5,$FF,$10,$F8,$69
	.dc.b	$18,$39,$7C,$97,$90,$8F,$E6,$BE,$4E,$BF,$CA,$B2,$4E,$32,$FB,$D5
	.dc.b	$74,$FE,$A1,$D3,$E2,$C3,$F4,$79,$E5,$FF,$21,$7C,$F3,$E8,$3F,$A8
	.dc.b	$36,$3E,$1B,$5F,$E7,$3F,$41,$FE,$60,$73,$3C,$08,$FC,$A8,$D3,$FD
	.dc.b	$02,$79,$3F,$EE,$16,$BF,$DD,$1A,$F9,$27,$FD,$84,$FF,$95,$FE,$83
	.dc.b	$F9,$B3,$FD,$29,$B9,$FF,$49,$6E,$5F,$DF,$B9,$9E,$27,$F6,$B9,$CD
	.dc.b	$F3,$F4,$30,$7F,$6F,$AD,$C7,$E0,$A7,$69,$A6,$23,$71,$FD,$73,$E7
	.dc.b	$DE,$17,$F7,$96,$C7,$E3,$1A,$FF,$6B,$F9,$C7,$D0,$63,$4F,$EC,$55
	.dc.b	$74,$FF,$0B,$B6,$FF,$CF,$0D,$BE,$38,$77,$24,$F7,$7F,$C1,$B1,$A8
	.dc.b	$F1,$78,$3E,$52,$3F,$C2,$23,$C7,$D5,$88,$F6,$4D,$C7,$FC,$3F,$5A
	.dc.b	$9F,$E5,$BF,$3D,$FC,$A7,$E8,$24,$60,$87,$F2,$97,$2F,$F2,$B8,$FF
	.dc.b	$0F,$7C,$FB,$E2,$7F,$19,$F1,$3F,$A7,$A4,$E7,$CF,$FF,$4C,$4E,$63
	.dc.b	$12,$8E,$1F,$40,$D7,$FC,$7F,$1B,$C3,$AF,$83,$FE,$42,$C3,$8C,$E7
	.dc.b	$1E,$B3,$FE,$46,$C3,$96,$38,$3F,$D5,$17,$C2,$79,$61,$D2,$72,$D5
	.dc.b	$5C,$9F,$34,$F0,$BE,$91,$EE,$AE,$17,$D6,$3F,$D6,$D7,$C7,$0F,$34
	.dc.b	$72,$3B,$DF,$2A,$E1,$1E,$0C,$7E,$0E,$3D,$E3,$81,$DE,$B8,$63,$E7
	.dc.b	$C7,$B5,$6F,$1E,$E4,$7E,$26,$1D,$30,$F0,$9F,$E6,$C2,$97,$1F,$AA
	.dc.b	$DC,$3F,$81,$BF,$EF,$1C,$9A,$7E,$FB,$E1,$91,$89,$E8,$7F,$AD,$DC
	.dc.b	$3D,$E3,$73,$FC,$1A,$FF,$9E,$7A,$17,$C2,$FF,$01,$5F,$8D,$AF,$E0
	.dc.b	$5A,$FD,$C7,$FC,$11,$A7,$E3,$5B,$A7,$FA,$23,$E8,$1E,$57,$FA,$02
	.dc.b	$B5,$87,$D1,$BD,$E1,$CE,$3E,$ED,$F0,$E6,$D4,$49,$FA,$79,$7A,$1F
	.dc.b	$AB,$5B,$73,$7D,$1F,$D1,$B6,$78,$61,$FA,$7A,$E0,$F9,$2B,$8D,$F7
	.dc.b	$8D,$63,$E9,$CB,$E1,$4E,$5B,$5E,$29,$D3,$EB,$39,$92,$B9,$E5,$C6
	.dc.b	$1E,$2F,$FA,$7B,$DC,$FF,$52,$F2,$FF,$84,$79,$2D,$FF,$A8,$FE,$13
	.dc.b	$FD,$6A,$FF,$8D,$A3,$38,$72,$9C,$BF,$E3,$18,$CE,$1D,$6B,$81,$FD
	.dc.b	$E3,$C5,$FE,$DE,$FF,$1B,$FA,$37,$91,$D3,$0F,$EC,$FE,$2F,$6E,$45
	.dc.b	$6A,$F9,$4E,$63,$15,$FE,$98,$C3,$AB,$FC,$E6,$33,$A7,$FA,$66,$5D
	.dc.b	$5E,$6F,$19,$F1,$7E,$19,$CE,$5F,$E5,$48,$C5,$6D,$B1,$93,$D9,$19
	.dc.b	$69,$9C,$7F,$D5,$7F,$38,$F9,$CF,$93,$93,$E8,$9F,$39,$F0,$6B,$58
	.dc.b	$7F,$91,$AB,$53,$9B,$E3,$7C,$2B,$FC,$47,$7C,$30,$FF,$61,$61,$C2
	.dc.b	$73,$7F,$EB,$99,$FF,$76,$D6,$DF,$E0,$8D,$73,$1B,$43,$ED,$5E,$6F
	.dc.b	$E0,$BD,$C7,$DB,$FF,$AA,$71,$FF,$1A,$D6,$9F,$F2,$FC,$EB,$E0,$3E
	.dc.b	$F4,$E7,$0F,$F6,$C4,$E6,$B8,$E1,$D7,$0E,$95,$FE,$D4,$BE,$19,$FF
	.dc.b	$B4,$DB,$7E,$6F,$B3,$7D,$BF,$E2,$23,$99,$FE,$48,$E8,$FE,$B4,$FF
	.dc.b	$9C,$9F,$F7,$FF,$C6,$7E,$45,$CE,$3F,$EC,$7E,$B4,$E7,$C6,$FC,$EC
	.dc.b	$62,$FA,$46,$27,$51,$83,$B7,$C5,$BE,$5F,$38,$7F,$BB,$B1,$FF,$2F
	.dc.b	$56,$BD,$DF,$17,$8F,$FA,$AC,$3F,$D2,$13,$E9,$8C,$4E,$A3,$81,$92
	.dc.b	$5F,$CC,$FA,$4F,$C3,$3D,$EB,$D3,$D1,$2B,$FC,$D7,$E1,$FF,$BD,$5F
	.dc.b	$89,$B7,$0E,$2F,$F8,$01,$39,$C3,$E2,$5F,$ED,$E3,$13,$FE,$6B,$AE
	.dc.b	$17,$FC,$27,$E8,$75,$F1,$9F,$21,$B1,$DB,$EC,$3E,$59,$FC,$33,$ED
	.dc.b	$9F,$83,$7F,$81,$67,$47,$9B,$C3,$D4,$67,$56,$7E,$69,$CF,$A1,$AE
	.dc.b	$09,$87,$FA,$C3,$1E,$BE,$5E,$31,$E3,$3F,$9E,$5F,$40,$F8,$CF,$E3
	.dc.b	$F2,$F9,$EC,$3F,$E9,$29,$B5,$C8,$46,$EB,$F8,$E4,$EB,$4C,$D6,$AF
	.dc.b	$17,$E1,$AB,$39,$BE,$27,$8A,$B3,$1E,$68,$C1,$F9,$4F,$97,$FC,$AF
	.dc.b	$D9,$5C,$3E,$DB,$6B,$F3,$8D,$C7,$EC,$3F,$9C,$F8,$EF,$CE,$FD,$0D
	.dc.b	$BF,$C1,$B6,$FF,$48,$DB,$FC,$03,$FA,$4F,$81,$FD,$2F,$6D,$BF,$C5
	.dc.b	$B6,$FF,$58,$DB,$FC,$43,$67,$FA,$83,$67,$FA,$73,$FC,$A3,$E8,$7F
	.dc.b	$DC,$B3,$FB,$9F,$39,$F3,$39,$F7,$3F,$9E,$36,$74,$DC,$3E,$B7,$82
	.dc.b	$BE,$35,$FA,$0E,$7F,$DE,$F5,$FE,$F9,$B1,$93,$F4,$B1,$96,$9F,$ED
	.dc.b	$4F,$53,$F8,$26,$1F,$83,$8F,$C4,$7C,$8F,$F5,$D9,$CE,$DF,$82,$9C
	.dc.b	$9C,$7C,$4F,$BD,$5B,$6C,$64,$FE,$67,$3C,$79,$95,$B9,$C9,$FC,$E0
	.dc.b	$EB,$87,$F1,$BF,$93,$FE,$33,$F5,$9F,$C6,$9B,$3F,$C0,$1B,$3B,$8F
	.dc.b	$E0,$73,$FD,$D7,$E8,$7F,$8D,$47,$F2,$9F,$A2,$FE,$D2,$FF,$BD,$23
	.dc.b	$35,$F3,$A7,$2F,$E9,$44,$6B,$2F,$F7,$94,$E9,$FD,$2C,$EB,$93,$FC
	.dc.b	$17,$39,$BF,$DB,$6A,$CB,$FC,$1F,$F6,$CF,$88,$F9,$D1,$86,$D7,$F3
	.dc.b	$F9,$D3,$F5,$E7,$4D,$35,$3C,$31,$FC,$3D,$70,$7F,$B6,$F3,$7F,$83
	.dc.b	$B3,$FD,$DF,$27,$E2,$FE,$87,$F2,$DE,$D7,$F8,$35,$B9,$7E,$5D,$B8
	.dc.b	$7E,$4B,$99,$1F,$9D,$F8,$3F,$F0,$A7,$87,$6F,$F3,$CD,$BE,$0D,$CB
	.dc.b	$C3,$82,$3F,$C5,$53,$FE,$B2,$1F,$42,$FA,$1F,$E3,$DE,$B7,$E2,$31
	.dc.b	$FE,$FB,$39,$AF,$F4,$37,$99,$8B,$DF,$47,$F1,$8A,$D5,$7F,$AE,$0E
	.dc.b	$B3,$AB,$E1,$1C,$8C,$9F,$E0,$AC,$3C,$0E,$0E,$55,$D2,$32,$F6,$8F
	.dc.b	$53,$9B,$F1,$5F,$FA,$43,$0E,$F4,$BE,$33,$98,$ED,$18,$7C,$97,$CF
	.dc.b	$0E,$31,$DA,$BD,$4F,$94,$C1,$E9,$C3,$C2,$F2,$8D,$D7,$9E,$3F,$DB
	.dc.b	$C7,$82,$76,$FB,$A7,$67,$96,$7C,$51,$DC,$E2,$E0,$E9,$87,$BB,$0F
	.dc.b	$54,$F3,$3C,$98,$7A,$9E,$37,$D9,$F3,$1A,$3B,$57,$3C,$3C,$B1,$EE
	.dc.b	$3E,$53,$87,$99,$C7,$0F,$1B,$ED,$BF,$05,$73,$6C,$64,$EE,$59,$E2
	.dc.b	$7B,$6F,$E1,$1E,$78,$F6,$1E,$29,$E7,$3B,$8F,$59,$93,$AB,$97,$8E
	.dc.b	$1F,$19,$F6,$35,$F5,$9C,$63,$C7,$5E,$18,$DE,$1F,$07,$0F,$64,$76
	.dc.b	$9D,$3C,$AF,$A4,$EA,$F9,$CF,$C7,$AF,$6B,$4F,$4D,$70,$7E,$24,$69
	.dc.b	$F2,$36,$35,$C8,$FA,$3F,$CE,$36,$BF,$5C,$D7,$F6,$8D,$F8,$6F,$F4
	.dc.b	$EE,$DF,$E7,$A8,$C7,$BD,$FE,$28,$A4,$62,$75,$7F,$E2,$3B,$E3,$5A
	.dc.b	$8C,$5F,$1C,$BF,$81,$FD,$1F,$F1,$AA,$FD,$9B,$5F,$DF,$5F,$29,$FE
	.dc.b	$12,$DC,$FF,$C6,$15,$B9,$E7,$C5,$FE,$33,$77,$87,$7A,$F0,$E5,$DA
	.dc.b	$B5,$1C,$9B,$8F,$FA,$63,$1E,$CD,$BE,$4F,$FA,$73,$1F,$05,$F1,$8F
	.dc.b	$E1,$75,$C1,$36,$77,$1C,$6F,$BB,$AB,$EA,$DC,$3C,$CC,$3E,$1C,$6A
	.dc.b	$FA,$57,$2B,$EB,$5C,$4D,$DF,$1C,$7E,$2C,$79,$B0,$E4,$78,$2B,$89
	.dc.b	$CB,$0F,$51,$DA,$79,$63,$F2,$9E,$A7,$F3,$39,$E3,$8F,$B2,$B8,$DF
	.dc.b	$58,$F3,$19,$7C,$38,$F4,$BF,$88,$79,$58,$F5,$9E,$2E,$31,$FB,$B3
	.dc.b	$C6,$18,$7A,$6F,$AC,$71,$35,$87,$92,$B8,$56,$DE,$98,$FD,$BA,$E0
	.dc.b	$7B,$B0,$EE,$E1,$F9,$6E,$63,$E2,$46,$08,$E0,$91,$EE,$39,$38,$7E
	.dc.b	$C4,$70,$74,$F8,$09,$1E,$57,$51,$C2,$39,$57,$D2,$8F,$59,$FE,$9E
	.dc.b	$F8,$37,$9F,$FE,$97,$C3,$FC,$03,$EE,$5E,$2F,$F7,$BF,$3F,$F9,$76
	.dc.b	$31,$87,$1B,$B6,$C3,$9C,$FF,$9A,$EB,$77,$C6,$3F,$C9,$67,$12,$D7
	.dc.b	$1F,$2B,$7C,$CF,$36,$9D,$B1,$4D,$33,$7F,$E7,$99,$C3,$D8,$7E,$44
	.dc.b	$F8,$2F,$1F,$DB,$3D,$7A,$7B,$2B,$73,$6B,$84,$7F,$BF,$6B,$83,$6B
	.dc.b	$55,$23,$04,$3F,$CA,$D5,$C9,$FF,$2C,$78,$FF,$E8,$E7,$E1,$EA,$FA
	.dc.b	$47,$C3,$FC,$86,$CF,$AE,$B8,$3C,$BC,$CD,$54,$8D,$47,$D0,$FA,$27
	.dc.b	$CD,$7D,$2F,$6B,$FC,$C7,$C7,$FE,$ED,$E6,$D4,$CC,$3F,$DE,$F1,$B3
	.dc.b	$0F,$F9,$BF,$E9,$1F,$39,$F1,$FE,$91,$F5,$14,$F9,$53,$AA,$FA,$4D
	.dc.b	$7F,$CF,$13,$C1,$FA,$CD,$7B,$57,$23,$A3,$AA,$C3,$F2,$33,$7C,$47
	.dc.b	$BF,$E7,$24,$E6,$3E,$11,$87,$FD,$6F,$29,$CC,$7E,$4B,$A5,$EA,$7E
	.dc.b	$4B,$EB,$24,$E6,$3F,$24,$61,$FF,$38,$46,$1F,$27,$D5,$7D,$1F,$AB
	.dc.b	$91,$82,$1E,$27,$2F,$92,$3E,$C1,$8E,$1F,$F4,$1E,$4A,$CD,$7C,$1F
	.dc.b	$A8,$FB,$DF,$51,$23,$04,$3D,$EE,$5F,$83,$1F,$85,$E4,$FC,$36,$F2
	.dc.b	$7E,$64,$65,$F9,$D1,$97,$E4,$C7,$E9,$FD,$47,$85,$F8,$3C,$65,$7E
	.dc.b	$C6,$72,$D9,$CB,$F8,$29,$E5,$12,$3F,$07,$1A,$CF,$FD,$73,$5C,$A5
	.dc.b	$0E,$54,$3E,$0B,$FC,$23,$C3,$FC,$59,$B1,$FC,$41,$AF,$29,$E3,$0F
	.dc.b	$AB,$FE,$21,$3F,$C2,$7D,$1F,$D3,$D6,$BC,$E7,$C0,$F5,$6D,$6E,$7F
	.dc.b	$61,$F1,$32,$3F,$2C,$EE,$BF,$69,$D3,$FC,$3F,$E8,$FF,$6B,$5F,$C5
	.dc.b	$DA,$FE,$5A,$35,$DB,$FB,$BF,$9D,$93,$B8,$AF,$E7,$22,$EF,$FE,$E0
	.dc.b	$BE,$6F,$FB,$8B,$E8,$9F,$1B,$FC,$7D,$B7,$FC,$9D,$AF,$C5,$3F,$D4
	.dc.b	$4B,$C3,$E1,$3F,$CF,$3D,$9F,$A9,$7F,$ED,$86,$ED,$F5,$30,$61,$D2
	.dc.b	$BC,$57,$C7,$9A,$57,$86,$3C,$9E,$4C,$8F,$23,$DF,$CB,$C4,$7E,$EB
	.dc.b	$E5,$FF,$8F,$18,$C3,$DD,$58,$9F,$E0,$D3,$9C,$BC,$92,$F8,$4F,$E8
	.dc.b	$0D,$AF,$E9,$0D,$7E,$11,$FC,$13,$C8,$DC,$FF,$47,$F0,$DF,$39,$4C
	.dc.b	$D7,$29,$C3,$FD,$4E,$76,$E4,$FE,$AD,$F4,$2F,$87,$F9,$11,$23,$0E
	.dc.b	$0F,$CF,$B8,$70,$D7,$84,$7F,$5B,$FA,$27,$8F,$4F,$EB,$D5,$C7,$6F
	.dc.b	$D5,$5F,$4B,$D3,$FD,$A6,$F8,$BE,$68,$FE,$CD,$F3,$EF,$A6,$FE,$EF
	.dc.b	$EE,$CA,$FE,$E1,$39,$6C,$E5,$FC,$67,$A2,$F5,$BF,$D1,$D9,$78,$A5
	.dc.b	$7C,$5F,$A2,$D8,$F3,$D7,$1C,$3A,$47,$19,$D7,$17,$E4,$5F,$9F,$7F
	.dc.b	$9B,$F9,$C7,$CA,$7F,$53,$BF,$83,$18,$EF,$F6,$F1,$CB,$FF,$0B,$CE
	.dc.b	$5B,$7C,$71,$ED,$5F,$DB,$FD,$BF,$EE,$1F,$57,$2F,$FA,$B4,$65,$FE
	.dc.b	$B1,$19,$7F,$A9,$C7,$F8,$47,$E8,$3F,$6B,$E2,$FF,$1C,$F8,$3F,$E2
	.dc.b	$74,$9C,$A8,$CC,$61,$DF,$D0,$BE,$53,$F6,$51,$AE,$FF,$F5,$AD,$F1
	.dc.b	$EF,$FD,$B6,$1C,$5B,$3A,$AE,$77,$CA,$73,$8F,$95,$B8,$73,$3F,$99
	.dc.b	$E1,$CD,$FE,$6B,$8F,$8D,$E0,$6B,$1E,$D5,$CB,$1F,$36,$1A,$A5,$F1
	.dc.b	$7B,$1F,$0E,$F8,$BC,$E7,$86,$1E,$18,$E3,$8F,$A5,$D1,$CA,$34,$6E
	.dc.b	$FB,$61,$E2,$9D,$61,$D9,$F0,$61,$E5,$AD,$E3,$F0,$A7,$F7,$38,$7D
	.dc.b	$66,$DF,$D0,$6D,$FE,$ED,$FF,$61,$FC,$3B,$83,$1F,$06,$8C,$4F,$F0
	.dc.b	$18,$C4,$EA,$73,$3E,$0F,$01,$F4,$1D,$72,$FF,$2F,$36,$33,$3F,$C2
	.dc.b	$E7,$2D,$8C,$EB,$F8,$C6,$E9,$FC,$EF,$E8,$7F,$CA,$BC,$7F,$3A,$73
	.dc.b	$7F,$80,$6B,$F9,$76,$CF,$08,$FA,$11,$AC,$3C,$38,$77,$C7,$F0,$17
	.dc.b	$B8,$C5,$F0,$9F,$01,$AF,$0B,$F8,$CC,$E8,$99,$7E,$73,$1F,$0C,$EA
	.dc.b	$31,$5C,$AF,$A1,$F8,$CC,$3C,$13,$A8,$CE,$1D,$1F,$F0,$7C,$E9,$F8
	.dc.b	$15,$F9,$CE,$A7,$A7,$FE,$A3,$D3,$FC,$FB,$D8,$E1,$FE,$5F,$D5,$FC
	.dc.b	$6A,$B5,$8F,$F2,$4A,$D5,$FF,$A1,$27,$35,$FC,$82,$73,$5F,$CA,$A7
	.dc.b	$35,$FE,$B6,$9C,$D7,$F2,$BF,$80,$79,$1F,$CB,$70,$FE,$E9,$18,$BF
	.dc.b	$C9,$36,$3F,$A7,$35,$FE,$A6,$E6,$BA,$57,$AA,$BF,$BF,$7C,$53,$AB
	.dc.b	$18,$7F,$8D,$30,$E1,$5E,$0F,$A1,$CE,$BF,$D0,$7E,$23,$77,$FE,$B5
	.dc.b	$9C,$E8,$9C,$E0,$9F,$55,$75,$6B,$C7,$0F,$5C,$62,$78,$47,$FB,$0D
	.dc.b	$B3,$FD,$F3,$2F,$D3,$CE,$6B,$C7,$F4,$1F,$E8,$C8,$FF,$6A,$FD,$0B
	.dc.b	$E7,$33,$F0,$3E,$EE,$04,$E6,$BF,$DA,$17,$C7,$2F,$EF,$0D,$7F,$BE
	.dc.b	$B5,$FF,$13,$35,$FF,$2D,$BC,$6B,$FB,$23,$67,$FC,$BF,$F1,$AF,$AC
	.dc.b	$FE,$FB,$1E,$0C,$7E,$5F,$D0,$3E,$8B,$E9,$DB,$E8,$BF,$CE,$F7,$FE
	.dc.b	$E5,$AD,$3F,$EF,$49,$D7,$A7,$CA,$FE,$6D,$6B,$4F,$F0,$93,$5F,$F0
	.dc.b	$BB,$5F,$F2,$03,$BD,$3F,$BE,$37,$0F,$F3,$BF,$D1,$3E,$89,$F3,$9D
	.dc.b	$E7,$FD,$F1,$CF,$FE,$56,$7F,$45,$EC,$7E,$1A,$31,$97,$E8,$63,$1E
	.dc.b	$3F,$E0,$A7,$33,$FE,$2A,$6B,$FE,$53,$6B,$FE,$88,$79,$E7,$FE,$11
	.dc.b	$6D,$E7,$DD,$7C,$CF,$68,$F2,$F8,$BF,$9A,$9F,$F0,$5C,$60,$FF,$7D
	.dc.b	$B8,$C3,$FC,$93,$6A,$F3,$12,$BF,$A1,$46,$5B,$1E,$F3,$EC,$FB,$FF
	.dc.b	$8D,$AE,$71,$FA,$AF,$03,$19,$7F,$0B,$E9,$FD,$5F,$89,$FE,$DA,$C7
	.dc.b	$F8,$54,$63,$D8,$FD,$3C,$E7,$1F,$F2,$63,$5F,$F2,$8B,$5F,$8D,$83
	.dc.b	$E8,$B9,$61,$FC,$55,$B3,$FC,$4F,$E4,$FC,$87,$F0,$2F,$83,$7D,$17
	.dc.b	$F8,$4E,$FF,$8E,$4F,$9B,$93,$C6,$48,$FE,$C9,$97,$E1,$1F,$F3,$05
	.dc.b	$F2,$DF,$A6,$06,$F9,$58,$7F,$85,$FD,$13,$E8,$BF,$89,$E0,$FA,$0C
	.dc.b	$3F,$EC,$8E,$2F,$F2,$85,$63,$6F,$F5,$04,$EE,$F9,$9F,$C3,$A7,$53
	.dc.b	$FE,$76,$6B,$FE,$9F,$6B,$FE,$D4,$7A,$57,$69,$F1,$DB,$E9,$7A,$E1
	.dc.b	$FC,$FF,$E6,$3E,$E7,$C4,$7D,$D6,$C6,$4F,$E8,$91,$96,$9F,$D1,$FE
	.dc.b	$8F,$F0,$9F,$47,$F8,$56,$C6,$4F,$E9,$91,$96,$9F,$D3,$FE,$07,$F1
	.dc.b	$9D,$9F,$8D,$6C,$64,$FE,$A9,$19,$69,$9D,$BF,$AA,$46,$5B,$1F,$3C
	.dc.b	$FA,$1E,$5F,$FA,$C5,$FE,$CB,$F2,$FF,$CA,$E3,$13,$FD,$8E,$31,$F0
	.dc.b	$D2,$BF,$B2,$7B,$F6,$BF,$B1,$74,$7E,$8F,$B3,$F4,$8D,$AF,$ED,$4D
	.dc.b	$9D,$C7,$F6,$D9,$FE,$A3,$EA,$7F,$B7,$63,$CD,$27,$F1,$1F,$3D,$AF
	.dc.b	$13,$EC,$B6,$3F,$C7,$55,$C6,$BF,$D6,$CD,$7F,$D7,$6D,$7F,$DC,$8E
	.dc.b	$F9,$BF,$D4,$0D,$94,$F8,$79,$7F,$AA,$BF,$ED,$29,$D2,$F0,$BF,$C1
	.dc.b	$FF,$1F,$FC,$0F,$E8,$FF,$82,$37,$0F,$F0,$53,$6B,$FC,$10,$DC,$7F
	.dc.b	$D6,$1F,$47,$FC,$7A,$BE,$0E,$0F,$A3,$FE,$E5,$1F,$13,$05,$6D,$FF
	.dc.b	$5D,$CE,$9F,$F7,$3C,$66,$FA,$F2,$7F,$80,$29,$F0,$FF,$E0,$0A,$D6
	.dc.b	$1F,$13,$2F,$F0,$2F,$D1,$3E,$8F,$EA,$C6,$1B,$3F,$42,$3F,$C0,$3B
	.dc.b	$1A,$48,$F4,$46,$5F,$BB,$F2,$6F,$D6,$3F,$C2,$BB,$7D,$0A,$D2,$DB
	.dc.b	$FC,$6F,$D3,$AE,$0F,$F6,$C7,$D2,$7E,$5F,$DF,$FF,$0B,$B7,$3F,$C4
	.dc.b	$DA,$F8,$79,$7F,$D3,$31,$E9,$1E,$09,$D7,$C2,$7F,$88,$3C,$6F,$D1
	.dc.b	$37,$0E,$75,$FE,$83,$6C,$7F,$41,$F3,$B7,$1F,$57,$0F,$F6,$41,$F3
	.dc.b	$BE,$27,$F1,$6B,$D7,$75,$FF,$C0,$0E,$75,$7F,$42,$AE,$07,$F8,$3A
	.dc.b	$75,$19,$FA,$4D,$CE,$70,$FC,$BC,$62,$3F,$D9,$E1,$C6,$1E,$09,$DD
	.dc.b	$73,$3C,$B1,$9A,$E1,$19,$7B,$18,$3F,$AA,$94,$F6,$5F,$ED,$DC,$1F
	.dc.b	$0A,$FC,$31,$97,$CD,$1F,$1B,$0C,$E1,$FB,$07,$31,$FE,$ED,$3E,$19
	.dc.b	$DA,$35,$7E,$58,$C1,$E0,$38,$57,$47,$1E,$77,$07,$D7,$7D,$A3,$11
	.dc.b	$CA,$30,$FB,$71,$F8,$95,$83,$A1,$6F,$E4,$E3,$ED,$7C,$93,$C1,$B1
	.dc.b	$9A,$F9,$35,$CA,$7F,$01,$7E,$73,$C2,$57,$8E,$3F,$62,$B8,$C7,$82
	.dc.b	$31,$1F,$80,$AE,$8E,$8F,$B2,$F4,$9E,$11,$F3,$0F,$A4,$E2,$BD,$91
	.dc.b	$C2,$3C,$2F,$CA,$3E,$25,$FA,$DF,$04,$74,$AE,$D1,$F3,$A3,$DA,$75
	.dc.b	$39,$C7,$68,$FB,$2D,$C3,$E4,$47,$A4,$CC,$F0,$23,$F9,$2B,$F8,$CD
	.dc.b	$8E,$CF,$19,$E6,$F9,$8F,$61,$B3,$CC,$53,$F0,$51,$E8,$72,$7E,$B2
	.dc.b	$7B,$BF,$AE,$E3,$FE,$56,$FE,$91,$AE,$63,$F7,$91,$8F,$88,$FF,$56
	.dc.b	$70,$7F,$1F,$9F,$E0,$B1,$8A,$FF,$6B,$46,$3E,$07,$F9,$14,$60,$FE
	.dc.b	$AB,$3B,$BE,$93,$AA,$FE,$05,$F4,$8E,$4F,$E2,$31,$86,$E5,$D7,$2F
	.dc.b	$D0,$E3,$E1,$AD,$65,$FA,$4C,$39,$3A,$C7,$B6,$1C,$63,$F8,$14,$E9
	.dc.b	$B8,$FF,$20,$D3,$F8,$A6,$3E,$D6,$E1,$CA,$BF,$B7,$DF,$1C,$7C,$31
	.dc.b	$FC,$02,$FC,$98,$F7,$C3,$24,$70,$49,$E7,$8F,$86,$3F,$98,$E1,$E5
	.dc.b	$AF,$0E,$5F,$A8,$C3,$C9,$5A,$BE,$18,$72,$35,$97,$D2,$C3,$8C,$7F
	.dc.b	$14,$C7,$D1,$7C,$63,$E3,$63,$1F,$86,$7E,$DA,$7C,$53,$E7,$86,$3E
	.dc.b	$27,$C0,$70,$3F,$97,$39,$77,$87,$C9,$55,$E2,$78,$E5,$E6,$3C,$F8
	.dc.b	$FC,$78,$EA,$E3,$0F,$15,$79,$27,$31,$E5,$C3,$E8,$C7,$48,$F7,$3F
	.dc.b	$57,$0F,$14,$F9,$6B,$94,$F9,$32,$F9,$6F,$A3,$2F,$96,$7C,$38,$CC
	.dc.b	$F0,$AF,$71,$CE,$32,$72,$73,$7F,$06,$7E,$05,$F7,$3E,$86,$5E,$D7
	.dc.b	$D6,$43,$C4,$7C,$C9,$E5,$7F,$A4,$8F,$71,$FA,$E8,$F6,$9F,$81,$C7
	.dc.b	$E0,$E5,$F2,$E7,$9B,$DD,$F1,$1F,$65,$C6,$3E,$B8,$F8,$07,$C2,$3C
	.dc.b	$59,$7E,$BF,$0F,$14,$72,$C7,$ED,$61,$FE,$A7,$F9,$87,$A7,$FE,$0C
	.dc.b	$C3,$FB,$80,$8C,$DF,$F8,$89,$AF,$F8,$AF,$83,$FD,$97,$F0,$7C,$5E
	.dc.b	$97,$67,$FC,$B9,$F0,$8F,$7B,$1D,$68,$C4,$F5,$AE,$19,$78,$7D,$0B
	.dc.b	$99,$A7,$2E,$B7,$D6,$E3,$0F,$F2,$FE,$2F,$57,$BB,$63,$D5,$7E,$55
	.dc.b	$F4,$8D,$F0,$D8,$FF,$33,$35,$F8,$03,$C1,$7E,$B2,$F9,$FE,$37,$FE
	.dc.b	$36,$A9,$7C,$5B,$19,$7F,$C7,$71,$96,$C7,$C6,$FA,$67,$7D,$3A,$CF
	.dc.b	$F9,$E7,$CD,$7A,$5E,$B6,$C6,$63,$E7,$FD,$33,$E5,$B5,$18,$F8,$F9
	.dc.b	$5B,$3F,$D0,$FF,$44,$F9,$FF,$9D,$D7,$27,$31,$F6,$A7,$51,$FE,$8E
	.dc.b	$E6,$FF,$1C,$3F,$1B,$A9,$18,$E9,$FF,$1E,$42,$B8,$D7,$CC,$1F,$4C
	.dc.b	$F9,$8F,$AC,$D8,$FF,$4C,$35,$E1,$F4,$CF,$86,$DD,$7C,$D9,$FF,$58
	.dc.b	$D6,$DF,$A8,$D7,$FC,$D6,$D7,$71,$F8,$5A,$7D,$1D,$3A,$CF,$C6,$F8
	.dc.b	$EF,$85,$F4,$D2,$30,$43,$E1,$64,$8D,$46,$27,$F3,$BE,$DC,$8F,$D7
	.dc.b	$3F,$37,$D5,$C4,$7C,$BE,$F7,$CF,$FE,$70,$CF,$07,$CC,$C9,$F4,$DF
	.dc.b	$57,$E9,$66,$1F,$72,$3E,$C1,$87,$CB,$18,$74,$7E,$09,$D5,$75,$9F
	.dc.b	$C0,$7D,$37,$E0,$7C,$09,$7F,$88,$8C,$BF,$8A,$8C,$BF,$0A,$3F,$11
	.dc.b	$CB,$F8,$9E,$09,$3F,$90,$8C,$BF,$92,$8C,$BF,$2A,$3F,$71,$EC,$FF
	.dc.b	$08,$8F,$C4,$37,$1F,$C8,$B7,$A3,$FD,$4F,$C7,$FB,$5A,$FF,$70,$4F
	.dc.b	$17,$F6,$CD,$7F,$D5,$AD,$7F,$DF,$2D,$7F,$DA,$9E,$4F,$E1,$75,$91
	.dc.b	$D2,$3F,$90,$7C,$5B,$D2,$FE,$55,$23,$F8,$4C,$67,$CA,$DC,$EA,$73
	.dc.b	$E3,$FF,$17,$7F,$49,$EC,$7E,$B1,$FE,$18,$D7,$FD,$86,$D7,$B3,$7B
	.dc.b	$7F,$5D,$DB,$FB,$72,$4F,$8F,$7F,$F5,$76,$5E,$7F,$87,$FD,$E7,$6F
	.dc.b	$F2,$5F,$A5,$FD,$AD,$7E,$F1,$BC,$5E,$BC,$BF,$75,$5B,$F0,$3F,$61
	.dc.b	$92,$FF,$7A,$FF,$23,$6B,$FE,$DA,$6A,$AF,$E0,$9F,$4D,$FC,$34,$91
	.dc.b	$E7,$C3,$85,$FF,$39,$FA,$67,$C8,$D2,$56,$CF,$F7,$34,$3C,$8C,$C7
	.dc.b	$F3,$18,$E2,$FF,$07,$6B,$99,$F5,$C6,$34,$FE,$33,$EF,$7E,$36,$BD
	.dc.b	$0D,$53,$BC,$7F,$DE,$EF,$F3,$A8,$F2,$1A,$F2,$BD,$47,$F6,$4A,$FC
	.dc.b	$05,$FF,$2A,$AE,$15,$E7,$AF,$19,$1F,$43,$FC,$CF,$E9,$A4,$76,$9E
	.dc.b	$35,$FC,$D2,$9F,$13,$FD,$85,$B5,$E4,$6B,$FC,$CF,$C5,$79,$3B,$AF
	.dc.b	$BB,$5A,$AF,$EB,$CD,$7F,$3C,$7F,$58,$8D,$9F,$DB,$E9,$F4,$74,$E9
	.dc.b	$3F,$13,$E9,$9F,$19,$FD,$D5,$B1,$98,$FE,$F1,$F4,$DF,$AD,$8C,$1F
	.dc.b	$CF,$E7,$6F,$F6,$D6,$BF,$DC,$DA,$F9,$9D,$E7,$FD,$62,$7F,$96,$CF
	.dc.b	$D6,$6B,$FD,$56,$77,$BF,$BB,$E9,$BF,$B6,$4F,$DC,$6B,$FD,$6D,$B6
	.dc.b	$9F,$E0,$5F,$49,$FE,$18,$F8,$3F,$F0,$95,$70,$BF,$F0,$43,$5E,$DE
	.dc.b	$97,$EB,$23,$11,$FC,$2D,$EB,$7F,$B3,$F3,$7F,$D8,$F8,$FF,$6E,$6C
	.dc.b	$72,$AF,$36,$BE,$02,$3F,$A7,$C9,$F4,$FF,$5A,$36,$7F,$6B,$C7,$0F
	.dc.b	$F8,$65,$AF,$A2,$BF,$B8,$75,$7F,$7D,$F7,$7F,$C0,$DF,$45,$2B,$FC
	.dc.b	$1B,$1F,$E0,$93,$0F,$F6,$68,$C3,$FA,$6F,$A5,$FD,$EF,$8B,$FC,$A7
	.dc.b	$E1,$FF,$92,$46,$25,$18,$EC,$FC,$23,$FC,$A7,$E9,$9F,$4F,$F9,$48
	.dc.b	$D6,$1C,$FE,$03,$F2,$F8,$F3,$9C,$E3,$D2,$7F,$37,$87,$28,$E0,$7F
	.dc.b	$3B,$AE,$13,$CB,$1F,$0B,$70,$DD,$2F,$93,$D4,$E5,$5C,$1B,$3C,$2F
	.dc.b	$AC,$7F,$49,$AE,$38,$79,$89,$87,$79,$E8,$77,$C7,$E0,$46,$67,$73
	.dc.b	$D8,$DD,$74,$9D,$12,$35,$3F,$2E,$B9,$C7,$13,$FC,$BD,$7F,$4D,$B8
	.dc.b	$FE,$D5,$B8,$FF,$8C,$FE,$15,$F1,$4E,$2C,$36,$BF,$CC,$93,$A9,$C9
	.dc.b	$FE,$3B,$C3,$FC,$C7,$F4,$D7,$B8,$D2,$5F,$F9,$B6,$76,$FC,$46,$E9
	.dc.b	$FD,$B7,$E9,$FF,$8E,$F3,$7C,$FC,$39,$DE,$DB,$5F,$90,$6E,$FF,$C7
	.dc.b	$AF,$8D,$6F,$1E,$44,$7E,$9C,$6A,$F9,$4E,$5B,$3A,$F0,$3F,$83,$3C
	.dc.b	$C9,$1A,$3D,$11,$A7,$17,$F9,$AC,$3C,$B5,$F9,$56,$AC,$3C,$53,$BC
	.dc.b	$3C,$4D,$9F,$16,$38,$7F,$BA,$7A,$7F,$EC,$38,$FF,$56,$EF,$FE,$C6
	.dc.b	$7F,$C0,$51,$99,$FF,$2B,$4E,$5E,$3E,$7D,$E6,$FF,$60,$4D,$AE,$1D
	.dc.b	$7F,$C5,$E3,$85,$FF,$AC,$FC,$5B,$F0,$27,$3F,$1A,$E9,$CC,$4D,$3F
	.dc.b	$B6,$D6,$DF,$ED,$F5,$B7,$33,$D3,$E9,$E4,$6B,$7B,$8F,$FA,$4F,$05
	.dc.b	$63,$3F,$F4,$EE,$3F,$90,$C3,$BC,$E5,$FE,$C3,$5C,$1E,$86,$8F,$F0
	.dc.b	$0F,$D3,$3D,$FF,$F1,$C3,$63,$FD,$30,$DC,$7F,$D8,$CD,$7F,$D9,$8E
	.dc.b	$71,$FF,$57,$B6,$7F,$C9,$DF,$1E,$FA,$6F,$F4,$3C,$78,$30,$F9,$B5
	.dc.b	$BC,$3F,$CA,$8D,$BE,$1F,$01,$AA,$FA,$7F,$48,$FA,$5F,$B9,$4F,$A6
	.dc.b	$FF,$7E,$65,$FE,$1C,$AD,$3F,$E3,$19,$D7,$B7,$CA,$E4,$FF,$8D,$7E
	.dc.b	$99,$F4,$CF,$9D,$CE,$DF,$EC,$3C,$B9,$B7,$0F,$C6,$5F,$0A,$DF,$07
	.dc.b	$FB,$43,$A7,$FC,$87,$97,$FB,$4F,$D5,$FC,$5C,$63,$1F,$F6,$A7,$57
	.dc.b	$F9,$0E,$BF,$C6,$73,$9B,$E5,$E2,$FE,$5F,$DC,$7C,$DF,$28,$97,$C1
	.dc.b	$D4,$63,$1F,$F6,$D6,$1F,$EF,$8A,$FF,$29,$3F,$ED,$5F,$9E,$F1,$F8
	.dc.b	$5D,$5B,$3E,$28,$F9,$BE,$FB,$E9,$B3,$87,$49,$BA,$7F,$06,$E5,$5E
	.dc.b	$F0,$E4,$7F,$BC,$F0,$FD,$71,$FA,$CF,$15,$EB,$E2,$7F,$CA,$71,$86
	.dc.b	$C7,$ED,$7E,$9B,$83,$63,$27,$FB,$FE,$32,$FF,$C0,$1F,$95,$FD,$E1
	.dc.b	$FC,$07,$E0,$9F,$4D,$FD,$CA,$74,$78,$39,$BC,$86,$E3,$D7,$19,$3F
	.dc.b	$CF,$53,$CB,$3D,$1B,$F2,$6C,$FF,$9B,$3E,$99,$F4,$DF,$C2,$9B,$1F
	.dc.b	$E8,$31,$F4,$6F,$A6,$D5,$49,$E4,$FF,$19,$77,$19,$E4,$71,$FF,$36
	.dc.b	$58,$76,$73,$5F,$E6,$28,$C4,$FF,$3C,$8E,$67,$63,$F0,$BD,$4F,$99
	.dc.b	$FE,$49,$F1,$1F,$C9,$5B,$3A,$6E,$1F,$9A,$6D,$FF,$48,$E9,$60,$FA
	.dc.b	$6F,$E5,$DF,$4D,$FC,$BD,$B8,$F8,$1B,$8F,$E9,$1B,$87,$3B,$FF,$39
	.dc.b	$D7,$9B,$B3,$F9,$C7,$C1,$FF,$39,$6C,$64,$F5,$C6,$5A,$67,$0F,$D8
	.dc.b	$C6,$5F,$8E,$7C,$CF,$37,$FD,$2B,$3F,$EB,$9C,$11,$89,$FD,$CF,$77
	.dc.b	$F4,$D9,$CF,$C2,$5A,$D4,$7F,$6A,$33,$7F,$3A,$B5,$5F,$D4,$7A,$3F
	.dc.b	$A9,$75,$DA,$FA,$6D,$C7,$F8,$3B,$70,$F8,$8F,$F5,$6F,$4B,$F8,$54
	.dc.b	$FA,$8F,$F8,$01,$E1,$FE,$27,$9F,$FA,$34,$66,$77,$7F,$89,$6E,$1F
	.dc.b	$EA,$86,$CF,$FA,$1E,$B8,$D6,$EF,$C0,$4B,$FC,$E4,$E9,$E5,$1F,$DB
	.dc.b	$7E,$37,$FB,$77,$D3,$59,$FC,$7B,$70,$FE,$46,$DC,$3F,$00,$FF,$72
	.dc.b	$FA,$5F,$E4,$35,$FE,$D8,$6D,$7D,$8E,$57,$D3,$3E,$83,$87,$24,$C3
	.dc.b	$83,$FE,$01,$9D,$46,$0F,$F7,$1E,$CD,$3E,$61,$FE,$09,$F9,$07,$D2
	.dc.b	$FF,$7B,$CF,$E5,$4E,$AF,$8B,$B9,$D6,$DF,$36,$35,$5C,$28,$FE,$42
	.dc.b	$FA,$FB,$1A,$BF,$C4,$73,$72,$BD,$C7,$E4,$A7,$FC,$31,$E5,$7E,$3D
	.dc.b	$93,$F9,$7A,$E1,$87,$F8,$A2,$B2,$FE,$7A,$FF,$37,$F0,$1F,$97,$3F
	.dc.b	$C4,$5E,$5D,$C3,$94,$5A,$FE,$AF,$36,$BF,$88,$47,$FA,$8C,$7D,$1B
	.dc.b	$E1,$3F,$DE,$3E,$6B,$E2,$BF,$C9,$73,$9A,$FF,$7A,$9F,$EA,$83,$8D
	.dc.b	$70,$9D,$BA,$8D,$4F,$15,$5D,$72,$F9,$73,$99,$F2,$3B,$9D,$47,$63
	.dc.b	$C3,$E6,$F0,$BF,$05,$D9,$C9,$B3,$BE,$1E,$47,$38,$75,$8C,$47,$FA
	.dc.b	$F4,$E1,$87,$82,$7C,$91,$EA,$BE,$EF,$53,$CE,$E2,$37,$87,$8E,$7C
	.dc.b	$38,$57,$67,$F8,$BB,$0F,$0B,$83,$67,$53,$C0,$6E,$39,$38,$37,$5E
	.dc.b	$07,$D3,$3D,$1D,$61,$EB,$C3,$EA,$61,$E5,$8C,$4F,$C4,$9F,$51,$A7
	.dc.b	$9E,$1E,$0A,$E1,$7D,$5B,$19,$9D,$9E,$98,$CB,$4C,$E1,$E7,$3E,$94
	.dc.b	$F1,$6C,$6D,$A6,$EF,$CB,$1E,$73,$E2,$1B,$72,$FD,$B8,$E3,$1E,$B6
	.dc.b	$C7,$13,$E4,$C6,$4E,$2F,$D3,$C3,$E2,$BC,$E3,$2F,$CC,$C3,$E1,$3E
	.dc.b	$C8,$E4,$F9,$2B,$D0,$D8,$CB,$E0,$3C,$55,$D9,$F0,$57,$4A,$F3,$C7
	.dc.b	$84,$E4,$70,$8D,$57,$93,$0F,$E2,$BC,$5F,$CA,$EB,$F4,$6D,$7F,$3C
	.dc.b	$D7,$F8,$13,$9F,$83,$7D,$27,$F8,$5A,$53,$98,$C3,$74,$EF,$CF,$FC
	.dc.b	$92,$97,$CA,$73,$1F,$CA,$23,$0F,$ED,$7E,$95,$F4,$BC,$75,$FE,$0D
	.dc.b	$18,$9F,$D5,$57,$09,$FE,$14,$DD,$3C,$15,$BA,$E9,$7C,$E7,$C9,$5B
	.dc.b	$9F,$16,$7E,$DC,$3B,$DE,$E7,$F9,$EE,$1C,$E7,$BD,$6F,$0E,$6D,$AD
	.dc.b	$DF,$15,$7C,$6B,$78,$FA,$AF,$85,$6A,$F8,$5E,$E3,$CB,$5B,$3F,$85
	.dc.b	$46,$63,$81,$93,$F8,$F6,$3E,$D7,$27,$C2,$35,$8F,$AA,$B9,$DF,$80
	.dc.b	$F4,$57,$13,$95,$F7,$3C,$0E,$70,$F4,$C6,$67,$8C,$7B,$27,$8C,$75
	.dc.b	$C3,$C3,$8F,$C2,$7C,$84,$C7,$C8,$FA,$AF,$AD,$71,$8C,$63,$C9,$F9
	.dc.b	$E7,$A8,$C1,$F9,$09,$E2,$E4,$E2,$4A,$FB,$EE,$5F,$09,$2F,$F1,$77
	.dc.b	$E1,$8C,$4F,$D4,$3D,$13,$A7,$C3,$87,$C0,$7A,$47,$BE,$78,$C7,$D4
	.dc.b	$7C,$2F,$8E,$34,$73,$7A,$9D,$E3,$8E,$3F,$3E,$3B,$E1,$EE,$76,$7C
	.dc.b	$1C,$7E,$1C,$EE,$3B,$1F,$E9,$0F,$82,$7A,$3F,$B7,$AF,$F2,$F4,$62
	.dc.b	$7F,$CD,$0D,$7F,$DB,$FE,$FE,$F8,$BF,$C3,$73,$9A,$E3,$79,$BE,$CD
	.dc.b	$ED,$FF,$45,$E1,$9A,$5F,$FA,$5E,$FB,$F9,$59,$8C,$47,$FB,$BB,$E9
	.dc.b	$71,$32,$36,$7F,$A8,$23,$C0,$BE,$91,$86,$1B,$1F,$E3,$2C,$FF,$8B
	.dc.b	$46,$23,$FC,$75,$F4,$AF,$9D,$E1,$72,$78,$1F,$E4,$0A,$E1,$9B,$0F
	.dc.b	$F2,$1F,$73,$1E,$57,$C0,$F2,$61,$CD,$FF,$3F,$61,$CA,$F8,$D7,$09
	.dc.b	$FF,$41,$78,$AF,$A3,$F6,$B6,$3A,$35,$FF,$35,$7D,$23,$E4,$F7,$52
	.dc.b	$76,$FD,$39,$D4,$7B,$38,$5F,$42,$F9,$9F,$B1,$E8,$FF,$B5,$78,$3F
	.dc.b	$D7,$5A,$FD,$BA,$D6,$1C,$5A,$FF,$9E,$EB,$84,$FF,$99,$B0,$FF,$65
	.dc.b	$BC,$70,$FA,$FF,$48,$F9,$DA,$4C,$3E,$47,$CF,$BE,$07,$11,$FE,$A5
	.dc.b	$AE,$0F,$CD,$6B,$FE,$A9,$9E,$0E,$6B,$8E,$9F,$EE,$FA,$DB,$FE,$F1
	.dc.b	$AD,$D7,$FA,$4F,$47,$C4,$7B,$FE,$92,$4F,$BE,$3E,$11,$87,$84,$61
	.dc.b	$FD,$4F,$A7,$F9,$B2,$3F,$1F,$E9,$2B,$FE,$7B,$EC,$4E,$32,$FF,$79
	.dc.b	$62,$BF,$A1,$F4,$9F,$47,$E8,$A6,$3F,$46,$32,$FD,$38,$CB,$EB,$8C
	.dc.b	$BE,$D8,$CD,$F6,$9F,$B9,$F4,$9F,$77,$B6,$5F,$DD,$8C,$BF,$7E,$32
	.dc.b	$FC,$78,$FC,$27,$2F,$E1,$7E,$92,$4F,$E1,$63,$2F,$E1,$E3,$2F,$CF
	.dc.b	$8F,$DC,$FA,$4F,$8D,$A4,$9F,$C1,$E1,$5F,$F7,$34,$ED,$FD,$73,$5F
	.dc.b	$C1,$46,$F2,$FD,$B4,$7F,$AE,$27,$51,$88,$FF,$64,$57,$F0,$AE,$2F
	.dc.b	$D9,$4F,$0F,$86,$FF,$10,$F0,$FF,$1D,$6D,$FF,$04,$6A,$9E,$1E,$0B
	.dc.b	$E1,$7F,$8A,$FC,$37,$E9,$2B,$FD,$92,$D4,$E6,$F9,$DF,$E0,$BA,$7F
	.dc.b	$60,$4A,$F0,$79,$CF,$83,$FD,$A7,$5F,$F2,$2F,$A4,$FD,$9E,$1E,$D6
	.dc.b	$F7,$7E,$CF,$A7,$F4,$2F,$F8,$DE,$33,$8F,$FB,$71,$A9,$CF,$7F,$F0
	.dc.b	$6C,$3F,$81,$7D,$24,$AF,$E0,$4F,$87,$C0,$FE,$67,$F4,$6F,$62,$92
	.dc.b	$BF,$95,$FA,$C9,$78,$F9,$9F,$95,$A7,$F9,$2E,$79,$47,$F2,$B6,$BA
	.dc.b	$3F,$D9,$6F,$27,$57,$FD,$47,$DA,$FB,$F5,$FC,$4D,$B3,$FE,$FA,$AE
	.dc.b	$7B,$FE,$0D,$7B,$1E,$A2,$4F,$F2,$7A,$FB,$8F,$8E,$32,$F8,$CF,$63
	.dc.b	$E1,$E6,$CD,$FE,$F3,$E9,$3F,$95,$18,$C7,$FC,$D1,$4F,$83,$FE,$B6
	.dc.b	$DC,$3F,$A5,$35,$ED,$E2,$BE,$6B,$75,$27,$6F,$F6,$49,$D4,$60,$FE
	.dc.b	$A8,$BE,$7F,$47,$18,$F5,$7D,$1B,$E2,$3F,$B6,$B6,$3F,$AE,$35,$F8
	.dc.b	$1F,$49,$FE,$86,$73,$3C,$08,$FF,$63,$8D,$3D,$A3,$FD,$10,$C9,$E8
	.dc.b	$D8,$CB,$FE,$8D,$8C,$BF,$E8,$FF,$A2,$FE,$B6,$FD,$16,$ED,$FB,$D6
	.dc.b	$EA,$97,$D0,$3D,$67,$C0,$FF,$82,$AF,$75,$B6,$A7,$35,$C6,$78,$7A
	.dc.b	$92,$BF,$C0,$15,$AB,$EB,$18,$D7,$B1,$1F,$F0,$6E,$0F,$A3,$FA,$91
	.dc.b	$B2,$1E,$D9,$FF,$17,$4E,$DF,$EE,$9E,$F7,$F7,$5F,$9F,$95,$FD,$D6
	.dc.b	$32,$FF,$77,$8C,$BF,$D7,$E3,$FC,$53,$F4,$3F,$B7,$F1,$3E,$37,$C2
	.dc.b	$7E,$7A,$93,$95,$19,$8C,$3B,$FA,$37,$D1,$FF,$B0,$E3,$5D,$DF,$85
	.dc.b	$6E,$1D,$BB,$3F,$D9,$18,$72,$7F,$D9,$58,$72,$C7,$47,$D9,$9D,$37
	.dc.b	$0E,$95,$C9,$B1,$F7,$EB,$93,$E8,$9D,$5F,$48,$F6,$D7,$06,$C6,$6B
	.dc.b	$8E,$1E,$89,$E4,$78,$67,$96,$1E,$48,$57,$0C,$3B,$C7,$03,$A9,$67
	.dc.b	$84,$EE,$3A,$19,$C3,$C9,$3A,$C7,$E6,$5F,$4A,$FD,$4C,$FD,$56,$D7
	.dc.b	$EB,$9D,$1F,$EE,$2F,$87,$FD,$0B,$FA,$CF,$85,$46,$23,$F6,$71,$8D
	.dc.b	$FF,$63,$1F,$EF,$C9,$D3,$F6,$3B,$5F,$45,$AC,$7F,$94,$B6,$33,$3F
	.dc.b	$BA,$9C,$B6,$33,$B7,$E2,$DB,$9F,$ED,$E5,$F4,$7F,$C7,$FA,$3E,$7D
	.dc.b	$6A,$7F,$08,$D7,$F2,$ED,$9E,$11,$F3,$E3,$58,$78,$B0,$F0,$63,$F8
	.dc.b	$3B,$DC,$62,$F8,$4F,$80,$D7,$7F,$F8,$32,$74,$4C,$BE,$24,$EB,$2E
	.dc.b	$17,$C8,$95,$FA,$DC,$3B,$DF,$09,$CB,$67,$58,$77,$8C,$1F,$DE,$FE
	.dc.b	$5B,$FC,$3B,$D7,$FE,$1B,$FC,$F3,$FD,$11,$97,$F9,$D6,$72,$FC,$B8
	.dc.b	$CE,$3F,$E5,$F9,$CB,$FC,$96,$33,$5F,$E5,$29,$CC,$A7,$35,$F5,$7E
	.dc.b	$C1,$F2,$1C,$B5,$7D,$97,$F8,$86,$BF,$D4,$55,$A7,$FD,$57,$3A,$CF
	.dc.b	$AC,$E7,$1F,$EF,$93,$9A,$E3,$87,$5C,$3A,$57,$F6,$FF,$A4,$FF,$3B
	.dc.b	$D7,$FA,$DB,$47,$D5,$3E,$7F,$D3,$5A,$BF,$F0,$45,$6A,$73,$F0,$9F
	.dc.b	$0A,$31,$87,$FA,$E6,$F8,$47,$CE,$C3,$A4,$E6,$BF,$C9,$58,$7E,$5E
	.dc.b	$FB,$3F,$D7,$1A,$E6,$36,$FF,$53,$F9,$17,$BB,$F4,$BE,$71,$F9,$8B
	.dc.b	$E6,$3E,$8D,$6B,$2F,$F6,$9D,$6A,$73,$7C,$6F,$85,$7F,$B4,$6F,$86
	.dc.b	$1F,$E2,$7C,$38,$4E,$6F,$8B,$FE,$BB,$F2,$2D,$7F,$86,$EF,$FC,$75
	.dc.b	$3F,$EC,$7F,$D4,$1F,$21,$E4,$F0,$3F,$13,$EC,$27,$35,$E1,$9C,$C6
	.dc.b	$30,$E9,$18,$BE,$55,$B3,$B7,$CF,$62,$BF,$C5,$9E,$F7,$FB,$D6,$7F
	.dc.b	$DE,$63,$E6,$DE,$D7,$E5,$A7,$55,$FE,$45,$AD,$4E,$7E,$17,$F3,$F1
	.dc.b	$83,$FD,$ED,$5C,$39,$BF,$D0,$F1,$9B,$F0,$D7,$18,$F0,$E0,$F5,$B1
	.dc.b	$5F,$E4,$19,$FE,$56,$79,$A1,$F6,$3F,$AA,$8C,$CF,$96,$73,$1F,$E1
	.dc.b	$5C,$3F,$D2,$8E,$9F,$1C,$6A,$7B,$78,$DE,$A8,$C4,$F0,$8E,$C7,$BB
	.dc.b	$07,$9B,$8C,$3F,$DF,$1D,$DF,$63,$DF,$FE,$73,$1E,$13,$F9,$87,$D1
	.dc.b	$FF,$90,$5D,$99,$CF,$FC,$55,$19,$7F,$DD,$15,$FC,$1A,$31,$97,$FB
	.dc.b	$6A,$3B,$4E,$7D,$6B,$7C,$E1,$7C,$23,$FC,$6D,$CA,$FD,$67,$DB,$18
	.dc.b	$6D,$27,$F2,$17,$F9,$29,$CC,$F4,$FA,$0F,$C8,$CF,$95,$B7,$F9,$B7
	.dc.b	$F2,$A7,$1F,$5E,$C7,$97,$DF,$FE,$02,$FF,$92,$73,$7D,$AF,$F9,$2E
	.dc.b	$35,$E0,$FF,$21,$8C,$65,$FE,$83,$BE,$2F,$F6,$68,$CE,$DF,$ED,$19
	.dc.b	$CE,$1D,$2B,$F9,$7C,$62,$B7,$87,$F6,$38,$C4,$FB,$FC,$BF,$81,$EE
	.dc.b	$DA,$FA,$DB,$FC,$98,$C5,$7F,$A9,$E3,$13,$AD,$FF,$B2,$E7,$FE,$E6
	.dc.b	$9C,$D6,$E5,$F6,$3F,$37,$EC,$BE,$73,$70,$E4,$DA,$E0,$DC,$3A,$5B
	.dc.b	$EA,$3E,$B7,$AD,$F5,$DB,$87,$8D,$B8,$78,$5B,$7C,$9B,$3F,$83,$6C
	.dc.b	$FF,$BC,$7A,$9E,$CF,$E7,$7E,$D7,$F9,$E4,$62,$FE,$2F,$42,$73,$5B
	.dc.b	$9F,$E9,$55,$BB,$F9,$17,$B9,$F9,$1D,$7F,$8F,$FA,$FF,$C8,$36,$32
	.dc.b	$7C,$D8,$CB,$4F,$C6,$77,$7F,$67,$FB,$57,$4F,$E0,$23,$0D,$9F,$EC
	.dc.b	$F3,$A7,$89,$88,$D6,$3F,$DA,$27,$4F,$E4,$6B,$9D,$7F,$39,$F7,$FF
	.dc.b	$99,$FC,$EF,$F3,$46,$C6,$4F,$E7,$D1,$96,$9F,$DB,$7F,$29,$7C,$87
	.dc.b	$F4,$0F,$51,$F3,$3F,$DA,$67,$35,$F9,$29,$CB,$C6,$B8,$56,$DB,$1F
	.dc.b	$AF,$F8,$CF,$EE,$5F,$AC,$FF,$33,$8F,$D0,$46,$33,$FD,$2B,$63,$81
	.dc.b	$FE,$01,$9E,$1A,$7F,$AF,$FB,$95,$B8,$FD,$A3,$79,$D7,$FE,$4F,$E8
	.dc.b	$F4,$D7,$E9,$FE,$D3,$FB,$4D,$6A,$3F,$D9,$53,$A8,$FD,$44,$EA,$B9
	.dc.b	$67,$FD,$92,$31,$7C,$23,$1E,$8F,$C7,$FA,$BF,$95,$F3,$17,$0F,$90
	.dc.b	$DB,$FD,$C7,$02,$31,$97,$ED,$E3,$15,$FC,$AE,$BF,$D7,$23,$E9,$7F
	.dc.b	$46,$FC,$3F,$53,$D5,$18,$8F,$F5,$B1,$BE,$CF,$E1,$D3,$A7,$FD,$E3
	.dc.b	$F0,$1F,$EE,$9A,$E1,$C3,$FE,$E8,$C3,$15,$AC,$39,$1D,$E3,$2F,$FB
	.dc.b	$FA,$EA,$AE,$4F,$03,$C7,$7D,$B0,$F1,$47,$13,$53,$C2,$7D,$51,$F0
	.dc.b	$1E,$17,$DA,$31,$87,$B2,$31,$1F,$ED,$A3,$9D,$F6,$7C,$33,$B7,$C8
	.dc.b	$72,$3D,$86,$0E,$55,$E0,$71,$87,$BB,$0F,$35,$7C,$5A,$D1,$E8,$AE
	.dc.b	$31,$DE,$31,$5C,$6F,$C9,$5E,$13,$E4,$38,$78,$9E,$49,$E2,$FB,$30
	.dc.b	$F7,$D7,$36,$C6,$4F,$39,$67,$89,$E6,$BF,$41,$E5,$2C,$77,$6B,$E4
	.dc.b	$9E,$B3,$C6,$3E,$C3,$5F,$9A,$71,$33,$3D,$5A,$64,$D9,$D4,$F3,$C7
	.dc.b	$37,$F3,$FE,$AF,$D1,$3E,$09,$67,$CB,$8F,$D2,$9E,$A7,$3A,$F7,$CF
	.dc.b	$86,$3B,$CE,$DB,$8F,$CF,$C7,$E5,$47,$8A,$B8,$3F,$56,$B8,$36,$B8
	.dc.b	$3F,$36,$7C,$71,$8B,$F5,$4F,$E4,$FE,$1D,$8F,$EA,$DB,$1F,$A1,$6B
	.dc.b	$FB,$86,$FC,$33,$E9,$5B,$29,$5A,$9C,$C6,$2F,$FD,$DD,$7C,$6F,$85
	.dc.b	$ED,$B3,$AB,$E3,$87,$F0,$9F,$A5,$7D,$37,$FB,$DE,$30,$DD,$38,$61
	.dc.b	$FC,$03,$2E,$D9,$FE,$FF,$0E,$2F,$F0,$46,$BF,$C0,$E3,$78,$7F,$17
	.dc.b	$6D,$FB,$65,$5C,$2B,$C3,$1D,$F1,$F3,$D7,$8A,$3C,$18,$F9,$5F,$8D
	.dc.b	$7C,$67,$57,$DE,$F8,$D6,$AF,$83,$E0,$6E,$1E,$48,$FA,$0D,$74,$7B
	.dc.b	$63,$73,$D2,$13,$F1,$32,$F4,$C7,$E0,$B0,$F7,$63,$F1,$27,$C9,$97
	.dc.b	$FB,$26,$7C,$E6,$B1,$F2,$D7,$AE,$FC,$78,$F7,$C7,$C9,$39,$9F,$29
	.dc.b	$F3,$E7,$47,$96,$34,$6E,$7D,$87,$2A,$EE,$7B,$30,$FA,$07,$BE,$E1
	.dc.b	$D5,$CB,$F4,$8C,$E5,$EC,$74,$64,$8E,$49,$8F,$BE,$38,$BF,$09,$F7
	.dc.b	$CE,$1F,$B0,$78,$27,$8B,$83,$AC,$F1,$7E,$9E,$3F,$06,$FD,$B1,$C6
	.dc.b	$75,$8F,$C4,$9F,$81,$3E,$8B,$FB,$24,$AE,$98,$FC,$79,$F2,$1D,$B1
	.dc.b	$E8,$F8,$EB,$83,$EA,$3A,$57,$BD,$F3,$99,$8F,$5B,$F6,$9E,$2F,$53
	.dc.b	$FB,$9C,$BF,$5B,$7C,$C7,$EA,$63,$FC,$15,$77,$87,$FD,$63,$C5,$FE
	.dc.b	$05,$9F,$F1,$01,$C6,$B5,$87,$F0,$EC,$39,$BF,$E3,$FC,$33,$E5,$B0
	.dc.b	$E5,$8F,$FB,$76,$F7,$7F,$C5,$A3,$F7,$5F,$AD,$FF,$BA,$E3,$FD,$33
	.dc.b	$F8,$EF,$F9,$D5,$BE,$4F,$81,$4E,$DE,$47,$F1,$F7,$1A,$7F,$A6,$EB
	.dc.b	$94,$FF,$A8,$5B,$1F,$E3,$B8,$E7,$EC,$75,$FA,$87,$CC,$77,$F9,$ED
	.dc.b	$E7,$FC,$3D,$B7,$FC,$B6,$B6,$FF,$B6,$DA,$F9,$04,$70,$6E,$7D,$EE
	.dc.b	$47,$FB,$8E,$74,$6E,$B9,$E3,$F4,$B4,$7C,$67,$CE,$FD,$71,$E7,$D3
	.dc.b	$EA,$D7,$DF,$9B,$5C,$31,$FF,$62,$DF,$06,$DE,$AB,$A7,$AB,$D7,$D5
	.dc.b	$95,$C2,$32,$F2,$8C,$BF,$EC,$DB,$4F,$CB,$F8,$1F,$95,$F0,$12,$BE
	.dc.b	$34,$65,$F8,$11,$97,$E4,$68,$9D,$47,$CB,$D5,$5B,$D7,$ED,$5E,$EB
	.dc.b	$FD,$B9,$F5,$1F,$92,$F5,$BF,$B2,$47,$FB,$7F,$BB,$13,$23,$04,$3F
	.dc.b	$DC,$0F,$CA,$FD,$75,$EA,$FE,$73,$D6,$7D,$8D,$35,$12,$FF,$D9,$B1
	.dc.b	$97,$FB,$96,$7F,$AA,$BE,$71,$FE,$F1,$FA,$3F,$37,$D1,$48,$C1,$0F
	.dc.b	$F7,$93,$97,$FD,$E9,$1E,$D3,$15,$FE,$EE,$9C,$4F,$FB,$F7,$ED,$FE
	.dc.b	$17,$C9,$48,$C1,$0F,$F7,$F3,$97,$FE,$00,$47,$C9,$F2,$3E,$57,$91
	.dc.b	$27,$D7,$DB,$FE,$0D,$F0,$9C,$5F,$E0,$E9,$C8,$8C,$DF,$18,$FD,$BC
	.dc.b	$70,$BE,$D1,$FA,$9F,$83,$FE,$1B,$F0,$2F,$80,$FD,$B6,$DF,$EB,$18
	.dc.b	$C6,$7F,$B7,$A9,$7E,$09,$CB,$6F,$8D,$7F,$81,$F9,$97,$C2,$BF,$84
	.dc.b	$DF,$06,$DF,$07,$F7,$D3,$FA,$FE,$39,$1F,$AA,$7F,$8F,$FC,$57,$ED
	.dc.b	$B1,$FF,$0A,$7E,$09,$FE,$E4,$AF,$D7,$F4,$7E,$DC,$FE,$35,$19,$8F
	.dc.b	$87,$4F,$91,$E6,$48,$E2,$7C,$CD,$D2,$FB,$5E,$7F,$0D,$2B,$53,$97
	.dc.b	$E9,$8C,$13,$E8,$FA,$F9,$3A,$9C,$BC,$A3,$2F,$99,$EC,$FF,$1D,$FB
	.dc.b	$9F,$E1,$4D,$F1,$7F,$9B,$46,$38,$FF,$97,$7A,$2F,$82,$CD,$7F,$3B
	.dc.b	$8F,$E8,$D7,$E6,$6B,$CC,$F4,$DF,$3E,$0F,$F2,$04,$EB,$0E,$BE,$77
	.dc.b	$4B,$F7,$F6,$7F,$03,$7F,$99,$36,$7F,$A0,$36,$FF,$C9,$5F,$81,$FE
	.dc.b	$66,$B7,$E2,$FF,$4F,$87,$C3,$53,$FA,$95,$F1,$AF,$F3,$73,$54,$F0
	.dc.b	$7E,$8B,$5E,$A7,$D4,$FC,$0D,$F3,$9F,$D7,$5B,$19,$8F,$E7,$D5,$97
	.dc.b	$FA,$07,$85,$27,$44,$3F,$99,$39,$7F,$A9,$57,$46,$FC,$37,$72,$3F
	.dc.b	$B8,$F8,$D7,$D4,$3E,$5B,$33,$FD,$BA,$73,$F0,$3F,$E0,$26,$C7,$AE
	.dc.b	$F7,$C0,$E0,$FF,$49,$56,$AB,$96,$2C,$37,$5F,$E7,$BF,$A3,$FD,$AD
	.dc.b	$7F,$9F,$4C,$78,$15,$A5,$11,$C1,$D3,$71,$FE,$FD,$F4,$8F,$23,$FC
	.dc.b	$14,$DA,$FD,$2D,$6F,$0E,$B1,$FD,$E6,$74,$DA,$FF,$09,$56,$E3,$FD
	.dc.b	$5D,$F7,$CF,$B0,$FF,$0B,$7A,$1F,$DF,$CF,$F0,$04,$66,$7F,$66,$D7
	.dc.b	$FC,$0D,$1B,$F6,$3F,$C3,$D3,$9C,$FE,$55,$7F,$2B,$F0,$5F,$57,$FE
	.dc.b	$23,$FA,$0F,$EB,$B3,$FE,$0F,$6D,$70,$F1,$7F,$83,$D7,$F1,$63,$87
	.dc.b	$8F,$2F,$F8,$3C,$7F,$8B,$7E,$89,$F3,$7F,$CE,$FE,$1D,$E9,$7F,$AB
	.dc.b	$E7,$3E,$8F,$FB,$36,$73,$87,$FB,$3E,$72,$A3,$2B,$E9,$1F,$1B,$FE
	.dc.b	$B9,$9C,$D7,$06,$E3,$E1,$C7,$C0,$DC,$3F,$DB,$98,$72,$7C,$53,$AC
	.dc.b	$6C,$78,$EB,$82,$C3,$B5,$7F,$B2,$6F,$94,$F9,$E3,$93,$4E,$57,$D1
	.dc.b	$E5,$7D,$2B,$8E,$1E,$8A,$D6,$1D,$AF,$89,$CC,$F8,$58,$FC,$1C,$7B
	.dc.b	$DF,$19,$D5,$71,$76,$74,$C7,$D7,$19,$35,$8F,$8A,$FD,$35,$FA,$99
	.dc.b	$FA,$6D,$AF,$E3,$CE,$8F,$F8,$03,$E7,$BE,$93,$F5,$9F,$0F,$FE,$20
	.dc.b	$6C,$66,$7F,$65,$39,$7F,$69,$7F,$C6,$BD,$3F,$DA,$36,$33,$E2,$27
	.dc.b	$2D,$8C,$C7,$C8,$8C,$69,$F8,$8F,$A4,$FF,$25,$F6,$E6,$A5,$FC,$6B
	.dc.b	$DE,$A8,$97,$C1,$D5,$70,$C3,$F1,$F7,$C2,$77,$EE,$FE,$BA,$73,$5B
	.dc.b	$C3,$F4,$17,$BB,$61,$D6,$F7,$1D,$88,$FA,$70,$E9,$7C,$B0,$E9,$5F
	.dc.b	$A7,$C3,$A6,$1C,$EB,$53,$C9,$B8,$73,$C7,$8F,$23,$DB,$FF,$41,$C7
	.dc.b	$F7,$EE,$77,$B9,$FE,$B8,$8F,$F1,$B4,$EB,$1F,$E0,$D3,$97,$FC,$6D
	.dc.b	$19,$E8,$FF,$52,$5F,$2B,$FF,$08,$4E,$5F,$9D,$49,$D3,$FE,$9F,$FA
	.dc.b	$4F,$F6,$04,$FF,$35,$6C,$7F,$39,$6B,$FC,$F5,$AF,$47,$77,$FE,$B5
	.dc.b	$6C,$FD,$DF,$8C,$7A,$2F,$33,$D5,$82,$F6,$FE,$0F,$E8,$BF,$C3,$F8
	.dc.b	$7F,$95,$31,$7D,$1F,$F8,$BE,$3F,$C4,$B3,$AA,$FF,$65,$D6,$BD,$AE
	.dc.b	$B3,$99,$FE,$A8,$D7,$E1,$B5,$F0,$7B,$3F,$EC,$D7,$2F,$23,$E8,$69
	.dc.b	$EA,$F1,$F1,$8F,$F8,$DB,$DD,$FE,$C8,$FF,$9A,$BE,$89,$F4,$7F,$DC
	.dc.b	$E3,$5E,$6F,$D9,$9C,$FA,$3F,$0A,$F8,$E7,$FD,$9D,$AF,$F6,$B6,$BF
	.dc.b	$69,$AF,$EA,$5E,$3B,$7F,$B9,$5B,$87,$E2,$BE,$43,$B6,$3D,$FC,$2F
	.dc.b	$C5,$74,$7F,$95,$3D,$7C,$5F,$F9,$49,$B9,$78,$EB,$E2,$3D,$DA,$B1
	.dc.b	$FE,$A3,$5F,$DF,$5B,$1F,$EE,$B6,$E5,$FD,$77,$C9,$7D,$1F,$68,$FF
	.dc.b	$7E,$7D,$1B,$E6,$7F,$3D,$8B,$E8,$FF,$D9,$2F,$F9,$92,$74,$FF,$A6
	.dc.b	$0D,$7A,$1F,$8C,$8C,$E7,$FE,$11,$6B,$FE,$15,$6B,$FA,$46,$BF,$C4
	.dc.b	$DE,$38,$79,$30,$F1,$E0,$F8,$EE,$D1,$F9,$AC,$7F,$D9,$E7,$8E,$3F
	.dc.b	$DF,$AF,$F9,$E8,$CE,$0B,$FF,$02,$C7,$3F,$7F,$97,$95,$F3,$23,$75
	.dc.b	$FE,$2B,$6B,$FB,$A6,$BF,$CA,$5E,$B5,$FE,$10,$8F,$F1,$D4,$FA,$3E
	.dc.b	$8F,$AD,$7F,$04,$F2,$3F,$03,$E1,$FF,$3D,$8F,$F4,$7F,$B3,$FE,$AE
	.dc.b	$74,$63,$D8,$FB,$F1,$9C,$7F,$C7,$8D,$7F,$C8,$0D,$7F,$8C,$9F,$BE
	.dc.b	$78,$65,$FE,$1D,$BF,$44,$FC,$3B,$FE,$2B,$5A,$7F,$DA,$DF,$0F,$F2
	.dc.b	$A7,$FD,$A2,$DB,$FC,$56,$1F,$9A,$8C,$57,$2F,$A3,$F3,$91,$FF,$28
	.dc.b	$60,$8F,$CD,$F4,$7F,$AD,$1F,$D1,$F2,$E6,$BF,$D7,$16,$FA,$0C,$D7
	.dc.b	$F1,$8F,$03,$F8,$86,$7F,$ED,$DF,$A3,$F5,$BF,$46,$33,$87,$F5,$DD
	.dc.b	$D5,$A8,$FE,$67,$3F,$C4,$76,$34,$9E,$17,$EB,$9E,$71,$FC,$AE,$3E
	.dc.b	$07,$D2,$F8,$DF,$85,$E8,$7C,$3F,$90,$B8,$7A,$9B,$7F,$29,$B9,$FF
	.dc.b	$93,$F6,$47,$CB,$FA,$4F,$99,$F1,$97,$0F,$86,$DC,$3E,$9B,$70,$E4
	.dc.b	$FD,$3F,$A4,$FA,$9E,$CD,$BF,$96,$DB,$FB,$6D,$BF,$13,$67,$F0,$2D
	.dc.b	$97,$23,$3F,$F2,$9D,$7F,$45,$EA,$7C,$C7,$E7,$3D,$FF,$EC,$D9,$27
	.dc.b	$35,$FD,$92,$73,$18,$C7,$FA,$4B,$6B,$E2,$7D,$4E,$B0,$FC,$77,$6F
	.dc.b	$E3,$FE,$26,$E5,$F8,$66,$E3,$F9,$66,$E1,$F3,$5F,$C4,$78,$1F,$DE
	.dc.b	$7A,$1F,$4B,$F8,$DC,$3F,$DF,$FD,$1F,$C4,$1F,$C4,$46,$71,$FD,$31
	.dc.b	$FD,$AE,$FF,$08,$49,$FF,$56,$57,$F1,$A9,$FC,$5C,$F8,$27,$FC,$5D
	.dc.b	$8F,$FA,$53,$07,$37,$F7,$0B,$DD,$68,$FF,$1A,$70,$BB,$1F,$1B,$FA
	.dc.b	$6F,$88,$FD,$3B,$6B,$85,$FF,$A7,$E3,$0F,$A2,$F9,$46,$32,$FF,$52
	.dc.b	$8F,$A5,$7D,$0F,$F5,$4F,$8E,$7D,$2F,$DA,$9C,$E5,$DB,$15,$6A,$3F
	.dc.b	$C0,$D3,$AC,$BF,$C0,$6D,$C7,$FD,$59,$5F,$AA,$7C,$18,$FE,$EF,$E9
	.dc.b	$3F,$0F,$F4,$BF,$47,$0F,$F0,$C4,$63,$87,$59,$7C,$E8,$FD,$A1,$1C
	.dc.b	$1F,$69,$FB,$DC,$7F,$AC,$8C,$1F,$1A,$BF,$D5,$78,$7F,$72,$FA,$5F
	.dc.b	$EE,$F1,$FE,$63,$EC,$F4,$57,$D2,$9C,$CF,$08,$E5,$3A,$8F,$D2,$CF
	.dc.b	$F6,$E9,$FF,$1F,$CE,$73,$F9,$DF,$4B,$F3,$7E,$97,$E6,$36,$BE,$4B
	.dc.b	$87,$E7,$6C,$FA,$DC,$65,$F9,$23,$FD,$84,$7F,$80,$7A,$FF,$8D,$FA
	.dc.b	$DF,$EE,$0F,$4F,$F6,$73,$FE,$50,$3F,$D9,$86,$CF,$F2,$24,$6E,$75
	.dc.b	$1A,$FA,$5E,$2F,$F2,$49,$FF,$75,$C6,$23,$FD,$B4,$1D,$F6,$27,$75
	.dc.b	$D3,$E1,$3F,$D1,$D5,$A9,$CD,$F9,$67,$35,$EA,$CE,$CE,$4E,$D1,$96
	.dc.b	$99,$BE,$93,$C2,$3E,$2E,$38,$7E,$01,$8A,$E5,$1E,$B8,$F8,$11,$D6
	.dc.b	$B1,$7C,$F1,$F6,$61,$D6,$FE,$0E,$1D,$4D,$BD,$A7,$6F,$80,$F8,$F7
	.dc.b	$E3,$36,$5C,$3E,$34,$7B,$0F,$A1,$7D,$DF,$29,$F6,$1C,$61,$E5,$C7
	.dc.b	$E5,$CF,$E1,$2F,$A3,$F2,$71,$F3,$61,$D9,$B1,$F2,$63,$E0,$1E,$29
	.dc.b	$E3,$1C,$E3,$2E,$CF,$11,$E7,$3E,$1B,$F3,$0E,$C5,$78,$1E,$C3,$81
	.dc.b	$EC,$31,$1F,$12,$3B,$C7,$84,$EF,$1F,$3A,$7D,$31,$F6,$CF,$1B,$C5
	.dc.b	$DC,$FA,$CE,$6F,$D2,$6B,$E1,$C7,$EB,$D7,$18,$F1,$CF,$17,$99,$F2
	.dc.b	$DF,$63,$E1,$3D,$95,$C6,$B9,$CF,$19,$E6,$71,$3C,$EF,$C6,$69,$93
	.dc.b	$9B,$96,$9F,$A3,$F8,$CF,$F4,$03,$6B,$F1,$31,$88,$FE,$0A,$DF,$8A
	.dc.b	$7C,$77,$F9,$86,$53,$98,$C3,$78,$53,$AB,$FF,$6C,$DF,$29,$CD,$6E
	.dc.b	$BF,$5B,$F4,$CF,$98,$FF,$53,$65,$FE,$E0,$8C,$4F,$FA,$9E,$39,$E5
	.dc.b	$E3,$CB,$F8,$7E,$5D,$6B,$9E,$1C,$9F,$1E,$1C,$9F,$DD,$CE,$EB,$8D
	.dc.b	$FB,$74,$FE,$33,$7B,$6E,$1F,$EF,$27,$79,$FF,$07,$C3,$96,$5E,$7A
	.dc.b	$F0,$64,$C3,$C1,$97,$08,$F0,$E5,$DE,$F8,$47,$F1,$1A,$F0,$47,$BB
	.dc.b	$1E,$D5,$BA,$F5,$E1,$E3,$AE,$31,$FC,$07,$1F,$85,$97,$73,$D3,$1D
	.dc.b	$6B,$83,$F5,$A7,$75,$DE,$3E,$49,$F3,$CE,$E5,$11,$F0,$1E,$57,$E5
	.dc.b	$7E,$05,$76,$69,$EA,$C2,$3F,$C5,$F1,$F7,$BD,$8F,$55,$73,$BE,$D1
	.dc.b	$C8,$F2,$9F,$2A,$BC,$91,$EA,$8E,$58,$FC,$07,$9C,$F8,$23,$04,$C7
	.dc.b	$E0,$3F,$57,$2E,$04,$8F,$88,$FC,$B7,$E7,$C7,$1A,$F8,$35,$D6,$72
	.dc.b	$48,$C9,$E2,$3E,$06,$5F,$17,$2F,$83,$1F,$33,$2F,$80,$7E,$76,$BE
	.dc.b	$EE,$3E,$39,$E6,$F8,$DF,$7C,$77,$79,$61,$F3,$A7,$84,$7C,$97,$DD
	.dc.b	$3F,$00,$E1,$7E,$2C,$BE,$55,$72,$3C,$E7,$E2,$4F,$71,$EC,$CB,$E1
	.dc.b	$3F,$55,$F2,$3C,$6F,$C3,$5F,$A2,$BF,$96,$7F,$97,$3E,$89,$F4,$DF
	.dc.b	$C2,$2F,$FC,$CF,$92,$73,$1F,$E5,$68,$C4,$ED,$FF,$22,$FA,$F7,$8B
	.dc.b	$A4,$DA,$FF,$48,$F2,$7F,$85,$2F,$FC,$97,$8E,$5E,$73,$FE,$3D,$FA
	.dc.b	$67,$B3,$C6,$11,$87,$51,$D7,$C1,$7D,$37,$5C,$FD,$57,$BB,$FF,$35
	.dc.b	$4B,$CD,$E0,$D8,$E6,$7F,$93,$7E,$81,$F4,$DE,$16,$C6,$63,$C9,$F4
	.dc.b	$8F,$A6,$D4,$7F,$8C,$A7,$F9,$8D,$70,$7F,$C6,$AD,$7F,$C7,$8D,$7E
	.dc.b	$83,$BD,$59,$FD,$3C,$3C,$D5,$FE,$49,$6C,$EA,$3B,$DF,$CB,$E2,$F2
	.dc.b	$47,$D7,$6B,$F7,$1B,$3F,$E7,$C8,$E1,$8F,$CB,$FA,$6F,$F6,$04,$7F
	.dc.b	$A1,$DA,$F1,$8D,$79,$BB,$F5,$E4,$EE,$32,$F1,$8C,$BF,$E9,$BA,$FF
	.dc.b	$54,$BF,$3F,$E1,$7E,$71,$1F,$F3,$F6,$6E,$6C,$47,$CA,$9F,$9C,$B6
	.dc.b	$46,$1B,$1F,$4D,$AF,$C9,$F8,$3F,$B2,$74,$D3,$FD,$F5,$F2,$5F,$8B
	.dc.b	$F5,$FF,$B7,$D7,$FA,$9D,$AF,$FA,$CD,$46,$CE,$1D,$D2,$78,$47,$E7
	.dc.b	$7E,$95,$EB,$52,$4F,$5F,$06,$BF,$EB,$FF,$83,$7D,$37,$E7,$5B,$1F
	.dc.b	$B1,$BE,$6F,$97,$E8,$BD,$1F,$27,$23,$04,$3D,$0E,$5F,$54,$7B,$CC
	.dc.b	$6B,$F8,$BE,$D6,$CB,$D5,$FC,$3F,$8F,$F8,$DF,$2D,$23,$04,$3E,$33
	.dc.b	$97,$E4,$47,$CD,$F2,$3E,$77,$8D,$27,$E7,$46,$5F,$A1,$19,$7E,$84
	.dc.b	$7E,$CF,$E2,$24,$7E,$DD,$FE,$05,$ED,$FF,$B8,$A3,$07,$E0,$63,$F0
	.dc.b	$84,$70,$7F,$BC,$3D,$C7,$C0,$FE,$AE,$1B,$7F,$BF,$E3,$0E,$7A,$17
	.dc.b	$FA,$7E,$59,$7F,$A7,$7C,$7E,$0F,$F1,$9F,$A5,$FF,$79,$CF,$F0,$56
	.dc.b	$BF,$EB,$86,$BF,$EB,$F6,$BF,$E3,$AE,$26,$8F,$93,$F7,$92,$7F,$6D
	.dc.b	$D5,$F0,$BE,$9B,$C1,$F0,$D2,$7F,$E0,$05,$A7,$11,$E1,$9F,$D5,$E2
	.dc.b	$69,$FC,$6F,$1F,$0F,$D3,$7C,$E2,$57,$FB,$CE,$B8,$3F,$EF,$5F,$98
	.dc.b	$FE,$0F,$C1,$F2,$DB,$E4,$7F,$4D,$6C,$FF,$95,$BB,$9F,$05,$FC,$FE
	.dc.b	$7F,$87,$CF,$FB,$85,$AF,$1F,$86,$FE,$03,$18,$8F,$F2,$FB,$D1,$FF
	.dc.b	$2D,$7A,$1F,$D2,$6C,$F5,$AF,$E9,$9D,$9F,$C1,$32,$FC,$03,$72,$FE
	.dc.b	$C4,$DC,$3F,$CD,$7F,$4D,$FE,$70,$7F,$CD,$B5,$C3,$C5,$FE,$27,$5F
	.dc.b	$C3,$A7,$FD,$F0,$D7,$9C,$68,$FE,$B9,$4F,$86,$A7,$92,$3F,$B0,$7D
	.dc.b	$33,$E7,$69,$2F,$F9,$A6,$15,$FF,$80,$33,$B7,$FA,$93,$5C,$CF,$F2
	.dc.b	$88,$C6,$3F,$BF,$F2,$DE,$27,$F4,$13,$A4,$7E,$8E,$5F,$11,$E4,$26
	.dc.b	$1F,$0B,$5F,$F4,$7F,$C8,$3E,$99,$F2,$FA,$AF,$2D,$F8,$48,$F9,$9F
	.dc.b	$F0,$35,$71,$F9,$7B,$67,$96,$5F,$B5,$FA,$5F,$D8,$E1,$E1,$3F,$D3
	.dc.b	$EE,$1F,$DC,$E5,$FE,$1B,$8C,$4F,$EF,$7E,$9B,$ED,$56,$8E,$3E,$1F
	.dc.b	$F4,$CA,$E5,$1F,$E2,$30,$E1,$E9,$1B,$3F,$C2,$D2,$F2,$69,$FE,$19
	.dc.b	$AF,$0E,$1F,$E2,$DF,$A5,$7D,$37,$F8,$75,$B8,$FA,$C3,$87,$CC,$3D
	.dc.b	$2F,$EF,$35,$CE,$BF,$B1,$D6,$EB,$D0,$D7,$47,$AC,$78,$1F,$C3,$3E
	.dc.b	$9B,$FC,$5D,$F4,$5F,$D7,$B6,$FE,$44,$DC,$7F,$89,$78,$AA,$FE,$5E
	.dc.b	$70,$F1,$A6,$3F,$C3,$A3,$FC,$71,$F4,$8F,$9E,$FD,$47,$C4,$3D,$2F
	.dc.b	$F2,$74,$E7,$D1,$FF,$6B,$CE,$70,$FF,$6C,$CE,$54,$65,$7D,$33,$E3
	.dc.b	$7F,$CB,$13,$9A,$E0,$DC,$7C,$38,$F8,$1B,$87,$FB,$9B,$0E,$4F,$8A
	.dc.b	$75,$8D,$8F,$1D,$70,$58,$76,$AF,$F3,$45,$F2,$9F,$3C,$72,$69,$CA
	.dc.b	$FA,$3C,$AF,$A5,$71,$C3,$D1,$5A,$C3,$B5,$F1,$39,$9F,$0B,$1F,$83
	.dc.b	$8F,$7B,$E3,$3A,$AE,$2E,$CE,$98,$FA,$E3,$26,$B1,$F1,$5F,$A6,$BF
	.dc.b	$45,$3F,$4D,$B5,$FA,$A7,$47,$FA,$27,$E2,$3F,$C3,$FF,$4D,$FE,$63
	.dc.b	$6C,$7E,$45,$AF,$EB,$5A,$FE,$13,$C2,$7A,$BF,$E6,$BC,$3F,$91,$36
	.dc.b	$33,$3F,$9B,$9C,$B6,$33,$9F,$E0,$DB,$97,$FA,$77,$E9,$9F,$01,$FE
	.dc.b	$69,$AD,$63,$F3,$EF,$78,$7E,$49,$AA,$F8,$F1,$EE,$A4,$FC,$FA,$DD
	.dc.b	$6A,$31,$3A,$BE,$F1,$F6,$A7,$51,$82,$63,$F0,$EB,$F0,$F3,$A8,$C5
	.dc.b	$74,$25,$7E,$96,$75,$1E,$9A,$F0,$57,$0A,$D6,$3F,$06,$30,$7F,$83
	.dc.b	$FD,$BF,$F4,$E6,$5F,$C4,$F9,$7F,$CB,$EF,$F2,$FE,$74,$EB,$2F,$E2
	.dc.b	$B5,$A7,$FC,$B5,$F0,$3F,$E9,$5E,$8F,$E1,$7E,$F5,$C7,$B4,$DA,$0B
	.dc.b	$E5,$7F,$E3,$3F,$A6,$FE,$F3,$7F,$C9,$5B,$5F,$CE,$7E,$31,$E9,$7F
	.dc.b	$95,$EB,$5E,$8F,$5A,$5F,$0C,$3A,$4F,$87,$0E,$3E,$6E,$EB,$58,$7F
	.dc.b	$B8,$FE,$97,$FC,$48,$FF,$AF,$5B,$8F,$F2,$B6,$BF,$D1,$DA,$F1,$F5
	.dc.b	$FC,$51,$C3,$2F,$F5,$C6,$3E,$8F,$17,$35,$30,$FE,$85,$7B,$7E,$2D
	.dc.b	$6D,$CC,$FB,$7D,$EF,$EC,$D1,$FD,$8F,$E9,$1F,$4B,$FD,$CA,$3F,$A6
	.dc.b	$4E,$B3,$FE,$C1,$5A,$F3,$F9,$79,$92,$FE,$F7,$2B,$E4,$31,$87,$F5
	.dc.b	$2C,$79,$6F,$FE,$67,$6D,$FE,$26,$B7,$39,$F0,$7E,$F7,$47,$F9,$6B
	.dc.b	$D4,$F4,$46,$3C,$8F,$F5,$CE,$1E,$68,$CB,$FE,$58,$FA,$5B,$9F,$86
	.dc.b	$6E,$3F,$E9,$7B,$56,$2F,$F1,$76,$FA,$57,$CD,$7E,$3E,$5F,$4B,$FD
	.dc.b	$AD,$FE,$E1,$3A,$7F,$C1,$86,$BC,$FF,$C5,$46,$6F,$FB,$9B,$72,$FF
	.dc.b	$79,$57,$4F,$90,$DE,$32,$3A,$BE,$1C,$53,$9C,$3F,$BE,$D7,$FB,$18
	.dc.b	$F0,$C3,$DC,$FC,$D4,$66,$7F,$C3,$0F,$28,$CF,$C2,$48,$F4,$DF,$FB
	.dc.b	$A5,$FF,$06,$1C,$71,$FF,$4F,$34,$F8,$E5,$B3,$F1,$BE,$2B,$A4,$7F
	.dc.b	$B5,$7C,$7F,$B7,$CB,$FC,$DA,$BF,$C3,$1E,$C7,$FA,$91,$FC,$4C,$FE
	.dc.b	$C6,$31,$5F,$E1,$A6,$E1,$FE,$CF,$BF,$E0,$37,$CB,$D7,$EB,$7C,$5B
	.dc.b	$19,$3F,$DE,$9A,$7E,$33,$E2,$3E,$34,$7F,$A7,$A7,$13,$E1,$FA,$49
	.dc.b	$3F,$EA,$26,$DF,$F9,$07,$9F,$FD,$5B,$1F,$93,$3A,$7D,$2D,$AF,$F1
	.dc.b	$A7,$81,$FB,$CB,$FD,$9F,$81,$FE,$C7,$7F,$1F,$19,$4D,$3F,$74,$F0
	.dc.b	$9F,$E5,$59,$BE,$93,$F8,$B5,$6A,$B8,$DE,$0F,$74,$DB,$E3,$E6,$7C
	.dc.b	$1F,$8F,$F8,$4D,$8C,$9D,$A3,$2D,$3C,$3F,$49,$F2,$7E,$37,$E5,$36
	.dc.b	$32,$79,$A3,$2D,$3D,$3F,$49,$F4,$7D,$8F,$A4,$D8,$C9,$EE,$8C,$B4
	.dc.b	$CE,$7F,$47,$5F,$F1,$D4,$63,$93,$F2,$33,$FE,$56,$9C,$BF,$E7,$78
	.dc.b	$FD,$27,$CB,$7F,$49,$6D,$7C,$CF,$A4,$C7,$22,$A9,$27,$47,$FA,$26
	.dc.b	$B8,$DF,$8F,$AF,$F1,$9F,$0F,$F8,$D6,$D7,$D8,$6C,$EE,$3E,$DC,$FD
	.dc.b	$AE,$EF,$C2,$4F,$F3,$86,$BF,$84,$6F,$27,$F0,$B8,$FF,$47,$3F,$C2
	.dc.b	$EB,$F1,$73,$FD,$97,$AF,$FC,$2F,$E7,$F5,$9F,$C9,$CD,$BF,$F4,$D5
	.dc.b	$6A,$72,$7F,$C0,$0F,$8A,$FE,$5D,$F0,$BF,$CB,$DB,$19,$3F,$BE,$78
	.dc.b	$BF,$EB,$E7,$17,$FA,$58,$C3,$F4,$7E,$8D,$F4,$1F,$45,$B5,$FE,$05
	.dc.b	$CF,$F3,$FB,$A7,$37,$C6,$F8,$1F,$D1,$2B,$73,$AA,$FC,$0D,$7E,$AB
	.dc.b	$E8,$BE,$39,$30,$FF,$4D,$46,$5B,$B2,$7F,$56,$7C,$9F,$A2,$75,$62
	.dc.b	$BF,$77,$36,$FE,$4D,$6A,$72,$7F,$AC,$F9,$FC,$8F,$F4,$C6,$CF,$FA
	.dc.b	$AB,$E8,$BF,$D3,$D3,$BD,$7F,$C2,$15,$AA,$FF,$4B,$47,$F6,$2A,$DC
	.dc.b	$7F,$AF,$27,$9E,$7F,$1F,$E8,$7E,$61,$F1,$7E,$86,$E3,$F9,$7D,$93
	.dc.b	$8C,$3F,$8D,$6E,$BC,$61,$F9,$2F,$7B,$FC,$5B,$F4,$3F,$A3,$BF,$F0
	.dc.b	$5F,$A3,$FA,$E8,$C4,$7F,$AA,$83,$8E,$2F,$F0,$16,$9C,$BA,$BF,$D9
	.dc.b	$58,$5B,$E7,$35,$E5,$7C,$70,$ED,$7C,$27,$AB,$FC,$A2,$33,$1F,$EE
	.dc.b	$FB,$E5,$87,$82,$36,$64,$A7,$F2,$E9,$E2,$F8,$2B,$A6,$1E,$38,$FF
	.dc.b	$66,$9C,$2B,$A6,$1F,$02,$32,$F9,$E7,$84,$7F,$B4,$0E,$17,$D9,$DD
	.dc.b	$F6,$7C,$66,$0F,$86,$60,$E5,$5D,$5C,$61,$EE,$AE,$91,$AA,$E9,$5E
	.dc.b	$58,$C1,$D1,$C5,$71,$BF,$3B,$67,$A9,$E8,$8C,$B4,$CC,$F8,$CE,$6E
	.dc.b	$5A,$66,$BE,$C7,$A5,$E9,$36,$57,$EE,$46,$E7,$EB,$35,$FA,$87,$A6
	.dc.b	$36,$7B,$8E,$A7,$CB,$9E,$AD,$37,$1C,$4E,$AF,$16,$99,$3B,$9F,$85
	.dc.b	$AF,$83,$5E,$38,$FA,$C6,$4F,$87,$3C,$DE,$38,$FE,$1A,$3C,$31,$B9
	.dc.b	$F8,$B1,$B8,$E7,$3B,$7C,$66,$E3,$8C,$66,$3A,$E3,$F8,$0C,$7E,$C5
	.dc.b	$F4,$9F,$51,$F9,$8F,$86,$65,$FA,$96,$C7,$E8,$DA,$FE,$E5,$BF,$08
	.dc.b	$F8,$86,$BF,$EA,$E9,$CC,$62,$91,$89,$D5,$FF,$BE,$EB,$53,$9A,$FD
	.dc.b	$8F,$D0,$BE,$4F,$FC,$DF,$97,$F0,$A8,$C4,$FF,$03,$CB,$C5,$95,$C7
	.dc.b	$C1,$7B,$C3,$8E,$DF,$E7,$8C,$78,$E3,$FB,$EC,$78,$B7,$0E,$35,$BB
	.dc.b	$E3,$3F,$EC,$AB,$E3,$87,$F1,$CC,$38,$B7,$0E,$35,$BB,$E2,$AF,$8E
	.dc.b	$1F,$07,$1F,$41,$FC,$76,$33,$3C,$4C,$9E,$58,$D5,$74,$9D,$CB,$1F
	.dc.b	$1D,$F5,$8F,$97,$5B,$3C,$73,$A8,$FF,$72,$B5,$F3,$46,$DF,$0E,$3E
	.dc.b	$DC,$3F,$DD,$79,$FC,$D9,$B8,$7C,$76,$9D,$32,$FC,$14,$F1,$8F,$19
	.dc.b	$EA,$3A,$CF,$9E,$FA,$46,$8F,$80,$49,$C9,$27,$DC,$F5,$9E,$24,$8E
	.dc.b	$84,$9F,$98,$75,$8E,$11,$98,$F8,$27,$94,$F0,$C6,$9E,$2F,$57,$DD
	.dc.b	$7E,$A7,$33,$F0,$DE,$2F,$5C,$BE,$B9,$F6,$32,$FA,$07,$D1,$73,$97
	.dc.b	$D6,$CB,$F1,$77,$F7,$5B,$3A,$8F,$01,$9C,$3D,$18,$F9,$63,$07,$88
	.dc.b	$F2,$90,$CB,$ED,$7E,$09,$F1,$71,$F8,$98,$F9,$71,$FC,$24,$F3,$B7
	.dc.b	$CE,$BE,$87,$FD,$CB,$1F,$E6,$E3,$2E,$38,$55,$FC,$22,$B8,$3F,$E3
	.dc.b	$DF,$4B,$4E,$6A,$DF,$28,$FF,$36,$57,$07,$FC,$2B,$5C,$17,$D0,$FF
	.dc.b	$B8,$9E,$F5,$FE,$40,$8C,$47,$F1,$8C,$3F,$DF,$71,$89,$FF,$42,$FA
	.dc.b	$0F,$A0,$FF,$0C,$C7,$1F,$5B,$B5,$6E,$F8,$C7,$A2,$F8,$6D,$E3,$F9
	.dc.b	$D7,$CE,$D2,$67,$E4,$C3,$E1,$AF,$A0,$7C,$4E,$A3,$CC,$7C,$8B,$E6
	.dc.b	$F1,$31,$7F,$06,$B8,$C6,$44,$F0,$7B,$9C,$0F,$F2,$FB,$97,$2D,$7E
	.dc.b	$4E,$FD,$1F,$F5,$33,$75,$FB,$CD,$C7,$EE,$E6,$EC,$EF,$5A,$CF,$ED
	.dc.b	$DE,$F0,$FC,$14,$7C,$9A,$DE,$1C,$BE,$65,$E9,$FC,$D7,$FC,$DF,$7F
	.dc.b	$07,$24,$62,$3F,$DE,$0F,$07,$4D,$73,$E3,$73,$3A,$57,$D7,$F8,$09
	.dc.b	$3F,$27,$54,$7C,$DE,$4C,$CF,$D4,$3E,$6C,$63,$6F,$F4,$6E,$EA,$DD
	.dc.b	$7C,$CA,$FF,$46,$BC,$0F,$A2,$FF,$A4,$33,$7A,$2F,$8B,$FC,$63,$FD
	.dc.b	$BE,$BF,$DF,$BD,$98,$99,$18,$21,$FE,$E1,$7F,$53,$F3,$EF,$46,$92
	.dc.b	$7E,$55,$FF,$64,$D3,$F7,$15,$B9,$FF,$6D,$35,$FF,$73,$36,$3F,$4F
	.dc.b	$3C,$78,$7F,$02,$BE,$6B,$D3,$F1,$73,$0E,$F1,$EC,$30,$FF,$BD,$E3
	.dc.b	$0E,$8F,$2B,$AB,$EF,$3F,$13,$E2,$BE,$2F,$C7,$4B,$F5,$C6,$5F,$6C
	.dc.b	$65,$FD,$5E,$DF,$3A,$7E,$6F,$89,$F3,$BC,$49,$5F,$12,$32,$FC,$68
	.dc.b	$CB,$CB,$3F,$F0,$EB,$F8,$7F,$7F,$F8,$01,$1F,$F7,$0F,$9D,$5F,$C4
	.dc.b	$73,$27,$1A,$FF,$10,$F4,$7F,$DC,$35,$CC,$FD,$AC,$FE,$22,$B8,$46
	.dc.b	$0F,$DD,$8D,$3F,$47,$C5,$FA,$72,$47,$FB,$CB,$BB,$F9,$37,$CE,$3E
	.dc.b	$92,$92,$7F,$4F,$5F,$C6,$F2,$7D,$05,$24,$FE,$B3,$A3,$5F,$41,$D3
	.dc.b	$E0,$64,$FF,$C0,$08,$AF,$E9,$F2,$4E,$37,$F8,$15,$DF,$E8,$25,$FE
	.dc.b	$F7,$24,$6F,$A7,$F9,$3F,$C9,$3D,$EF,$E8,$0D,$BF,$E4,$0D,$7F,$6D
	.dc.b	$E6,$3D,$AD,$D4,$9D,$BF,$D2,$67,$51,$83,$F9,$53,$FC,$6D,$BE,$37
	.dc.b	$17,$F8,$17,$77,$F0,$67,$F9,$83,$5C,$F8,$3F,$D5,$7E,$83,$F0,$F1
	.dc.b	$83,$71,$FD,$66,$78,$6E,$D3,$FA,$2D,$F0,$8C,$5F,$F6,$AA,$DD,$6B
	.dc.b	$DF,$EC,$73,$C3,$FB,$27,$D0,$3E,$5E,$93,$0F,$E4,$B8,$57,$F8,$E4
	.dc.b	$ED,$FE,$94,$D7,$F9,$6C,$6F,$1F,$EC,$FE,$57,$EF,$E7,$F9,$7F,$89
	.dc.b	$9C,$FF,$AE,$E0,$F7,$FC,$64,$AF,$EB,$5F,$14,$FA,$07,$C9,$66,$7C
	.dc.b	$75,$73,$F8,$37,$BA,$C6,$3F,$E0,$C8,$E5,$F2,$5F,$E0,$A6,$E1,$EA
	.dc.b	$C3,$FC,$0D,$F3,$FF,$AA,$8F,$E7,$4D,$C7,$E6,$C2,$31,$87,$F7,$6F
	.dc.b	$A0,$FA,$4F,$F7,$88,$CC,$FF,$87,$1A,$F1,$F1,$7F,$C2,$95,$C2,$35
	.dc.b	$1F,$D0,$A7,$C5,$1F,$DD,$1A,$F5,$8F,$E8,$FF,$3E,$FA,$0F,$F0,$F3
	.dc.b	$67,$FB,$F3,$5F,$EA,$9F,$29,$FD,$B5,$46,$72,$F6,$B7,$2F,$F0,$8B
	.dc.b	$67,$FC,$6B,$5F,$DD,$7C,$77,$7B,$E8,$3F,$C5,$DF,$39,$FD,$4E,$FF
	.dc.b	$B1,$37,$2F,$E3,$53,$FC,$24,$DF,$8B,$2B,$F7,$91,$FE,$37,$F9,$E7
	.dc.b	$CC,$FF,$2A,$F8,$37,$A7,$FE,$24,$9C,$FA,$5F,$E4,$89,$CE,$1F,$E4
	.dc.b	$A9,$CA,$8C,$AF,$A0,$7C,$57,$F8,$BA,$73,$5C,$1B,$8F,$87,$1F,$03
	.dc.b	$70,$FF,$2B,$E1,$C9,$F1,$4E,$B1,$B1,$E3,$AE,$0B,$0E,$D5,$FE,$3E
	.dc.b	$BE,$53,$E7,$8E,$4D,$39,$5F,$47,$95,$F4,$AE,$38,$7A,$2B,$58,$76
	.dc.b	$BE,$27,$33,$E1,$63,$F0,$71,$EF,$7C,$67,$55,$C5,$D9,$D3,$1F,$5C
	.dc.b	$64,$D6,$3E,$2B,$F4,$D7,$F9,$32,$7E,$9B,$6B,$F8,$FB,$A3,$FC,$BF
	.dc.b	$F0,$7F,$E1,$3F,$A0,$FE,$B3,$1F,$ED,$9F,$41,$18,$9D,$4F,$F5,$E6
	.dc.b	$BD,$FC,$37,$C7,$F2,$C7,$F8,$F3,$63,$33,$FD,$9A,$72,$D8,$CE,$9F
	.dc.b	$84,$6E,$7F,$E7,$4F,$A0,$FF,$28,$F4,$7C,$E9,$CD,$FE,$45,$AF,$DF
	.dc.b	$6C,$F0,$8F,$9B,$1A,$C3,$C3,$87,$7C,$7E,$F5,$EE,$31,$7C,$27,$C0
	.dc.b	$6B,$DE,$FE,$83,$3A,$26,$5F,$A7,$C7,$C3,$3A,$8C,$57,$2B,$E8,$7E
	.dc.b	$17,$0F,$04,$EA,$33,$87,$47,$FD,$6F,$3A,$7E,$05,$7E,$83,$47,$27
	.dc.b	$F1,$5D,$3F,$D2,$16,$E2,$FF,$61,$47,$FB,$02,$B7,$D5,$8C,$7A,$D5
	.dc.b	$9C,$D7,$28,$C6,$1D,$6B,$FC,$E1,$87,$5A,$FF,$4F,$7C,$FF,$F8,$31
	.dc.b	$FF,$52,$53,$E9,$BF,$C3,$31,$AF,$4F,$C1,$39,$F5,$7C,$58,$B0,$E4
	.dc.b	$FF,$8B,$1A,$FF,$AA,$1A,$FF,$AE,$1A,$FF,$B7,$3E,$E9,$F2,$DE,$DF
	.dc.b	$8E,$7D,$E7,$F9,$8F,$3F,$F5,$A5,$69,$FF,$5D,$4E,$BC,$CF,$1C,$E7
	.dc.b	$0F,$F6,$0C,$E6,$B8,$E1,$D7,$0E,$95,$FE,$93,$BE,$19,$FF,$5A,$6D
	.dc.b	$FF,$65,$FB,$C7,$DD,$F6,$C7,$E0,$FC,$0F,$F3,$2A,$FF,$18,$7A,$58
	.dc.b	$EE,$FA,$DE,$2F,$F4,$28,$C4,$E8,$FF,$67,$57,$1D,$3F,$CD,$BE,$2F
	.dc.b	$FA,$A2,$73,$87,$09,$CF,$D1,$5D,$BF,$B7,$E1,$FE,$15,$BF,$F2,$07
	.dc.b	$CA,$79,$3B,$FC,$39,$23,$13,$A3,$C3,$3A,$8C,$E3,$FE,$B5,$BE,$8F
	.dc.b	$93,$D7,$CE,$3F,$DF,$B0,$FF,$52,$1F,$EE,$B1,$F1,$BF,$E4,$D7,$FD
	.dc.b	$D4,$6A,$BF,$DE,$6E,$BD,$AF,$C4,$B9,$AF,$F1,$AB,$93,$8F,$8B,$F9
	.dc.b	$F7,$15,$FE,$C9,$78,$06,$38,$7F,$C8,$5F,$43,$FE,$CB,$8E,$67,$FB
	.dc.b	$F7,$D2,$FF,$32,$BF,$EE,$3F,$BE,$FF,$3E,$39,$F7,$FE,$CC,$62,$FF
	.dc.b	$DB,$15,$C1,$FF,$51,$C6,$71,$FF,$10,$CE,$5F,$F2,$7F,$22,$75,$87
	.dc.b	$9E,$D7,$B8,$C3,$63,$FD,$29,$3F,$AC,$BE,$B8,$7D,$CF,$46,$61,$EB
	.dc.b	$FA,$1B,$87,$99,$B7,$E4,$6C,$FC,$7C,$BF,$C2,$B1,$FB,$AC,$8F,$E2
	.dc.b	$0C,$1F,$8A,$E8,$95,$FE,$8D,$FA,$FB,$8F,$C4,$7F,$1E,$73,$2C,$7F
	.dc.b	$A4,$FD,$7F,$C9,$9F,$94,$F6,$2C,$7D,$9C,$FF,$C9,$3D,$EF,$9D,$FE
	.dc.b	$3D,$E5,$3E,$71,$F4,$5F,$C4,$2B,$58,$71,$3F,$05,$5B,$9D,$5F,$2B
	.dc.b	$E3,$5B,$D3,$F1,$18,$72,$7F,$99,$C7,$29,$DC,$60,$DD,$7F,$97,$9F
	.dc.b	$B5,$5C,$5F,$B8,$1C,$3A,$8E,$0F,$FB,$3F,$B3,$F5,$B9,$7F,$B6,$7B
	.dc.b	$BF,$6F,$DB,$FB,$86,$E5,$FE,$BB,$6D,$FF,$06,$7F,$7B,$F0,$7F,$BE
	.dc.b	$F7,$2D,$7F,$B0,$5B,$5F,$C3,$1B,$87,$FB,$9D,$FE,$2B,$DA,$FB,$1F
	.dc.b	$E0,$F7,$86,$BF,$1F,$9D,$3B,$8F,$F6,$71,$B2,$9B,$75,$87,$F1,$C8
	.dc.b	$D3,$5F,$E9,$FF,$3A,$F8,$CF,$1F,$C5,$7F,$4E,$8C,$4F,$38,$C4,$FC
	.dc.b	$AC,$D5,$AB,$FA,$31,$89,$D4,$E6,$7F,$61,$8F,$AB,$ED,$7E,$0F,$A7
	.dc.b	$F0,$9B,$97,$A5,$B5,$F8,$1A,$E1,$3E,$DA,$FC,$97,$D0,$7F,$64,$FA
	.dc.b	$DF,$C5,$67,$F9,$A8,$C7,$26,$AF,$FB,$6D,$69,$FC,$4C,$E9,$FC,$A6
	.dc.b	$3F,$89,$8F,$C3,$B6,$3F,$C7,$FC,$A9,$CB,$63,$F2,$32,$CB,$EB,$E1
	.dc.b	$F7,$BE,$D7,$EF,$B6,$BE,$B3,$6B,$DD,$C1,$E5,$79,$2A,$FE,$FF,$EE
	.dc.b	$FC,$FD,$3F,$8E,$FD,$13,$B7,$3C,$5F,$D8,$23,$19,$FB,$2F,$E9,$65
	.dc.b	$FC,$85,$AF,$F3,$96,$BF,$E9,$0A,$F8,$3F,$8F,$BB,$FE,$77,$D1,$FF
	.dc.b	$84,$23,$18,$7C,$C8,$C4,$EA,$3E,$5C,$EA,$FF,$75,$9B,$89,$87,$0B
	.dc.b	$FF,$35,$DE,$EF,$F7,$6D,$7F,$9C,$35,$FF,$1C,$F2,$FF,$6D,$9F,$D0
	.dc.b	$7D,$7F,$E6,$7C,$DF,$D6,$B6,$3E,$AE,$5F,$98,$9F,$CA,$92,$33,$5F
	.dc.b	$CC,$9F,$D5,$7D,$0F,$E8,$FC,$5F,$DB,$B6,$37,$1F,$9E,$23,$FE,$84
	.dc.b	$C7,$FD,$80,$3E,$89,$F4,$1F,$C6,$3E,$91,$5F,$DA,$67,$35,$FC,$6B
	.dc.b	$E0,$31,$7B,$AB,$39,$1E,$97,$07,$F9,$06,$1D,$1C,$D7,$E5,$2F,$FB
	.dc.b	$B5,$6F,$AD,$7B,$8E,$57,$E8,$C7,$53,$CD,$CC,$7F,$B5,$CF,$66,$1E
	.dc.b	$56,$9E,$58,$E3,$8F,$A2,$3F,$DB,$46,$8F,$3B,$97,$C1,$3C,$E7,$BE
	.dc.b	$3C,$71,$F3,$DE,$EF,$C5,$8F,$67,$DE,$7C,$57,$07,$27,$C1,$3C,$9F
	.dc.b	$11,$E0,$3F,$A9,$E3,$EF,$C7,$E4,$47,$43,$CE,$53,$EE,$D7,$68,$DE
	.dc.b	$3E,$33,$C0,$57,$27,$B4,$B3,$E0,$C3,$D9,$8F,$BF,$0F,$14,$FC,$BA
	.dc.b	$E0,$FB,$5F,$A1,$87,$B4,$F1,$46,$CE,$87,$B8,$FB,$18,$78,$DD,$1E
	.dc.b	$C7,$CF,$3D,$A3,$EB,$CF,$46,$C7,$80,$F4,$BC,$5B,$8F,$CD,$AE,$73
	.dc.b	$E2,$6B,$B3,$C3,$1B,$9E,$37,$F0,$27,$F0,$F5,$E7,$6C,$F9,$5F,$B0
	.dc.b	$73,$32,$77,$79,$E3,$F3,$23,$F0,$35,$D5,$E3,$7E,$17,$C7,$19,$7E
	.dc.b	$3D,$77,$69,$D4,$DD,$75,$BF,$34,$7F,$00,$F7,$FF,$6A,$FE,$DD,$AF
	.dc.b	$E8,$9A,$EF,$1F,$F2,$A7,$C2,$3E,$1F,$FD,$AB,$3F,$ED,$39,$CC,$62
	.dc.b	$B9,$D6,$F8,$7F,$CE,$51,$89,$D5,$70,$9F,$D9,$7D,$23,$E5,$3F,$CC
	.dc.b	$59,$7F,$09,$8C,$65,$FB,$7C,$BA,$3F,$B7,$6C,$EE,$B8,$E9,$FE,$6B
	.dc.b	$BE,$37,$BC,$BD,$37,$FC,$57,$0E,$37,$BC,$38,$DF,$FA,$27,$0E,$37
	.dc.b	$BC,$38,$DF,$A6,$BF,$84,$56,$A3,$94,$79,$E7,$6F,$FA,$4F,$0E,$EF
	.dc.b	$5C,$3B,$AB,$E4,$EE,$B8,$AA,$E2,$EE,$B8,$D7,$08,$D5,$70,$AD,$CE
	.dc.b	$4D,$3C,$8C,$D7,$23,$C6,$E6,$7C,$33,$C6,$73,$1D,$4F,$C1,$E5,$F4
	.dc.b	$6F,$F9,$DE,$1F,$09,$FE,$2D,$87,$D8,$9F,$95,$97,$E0,$30,$F0,$57
	.dc.b	$C7,$88,$FC,$C8,$CB,$F6,$63,$33,$DE,$3F,$91,$63,$F7,$A3,$AC,$78
	.dc.b	$31,$F4,$DF,$2C,$B2,$4C,$BC,$CF,$33,$15,$F4,$71,$F3,$30,$F1,$13
	.dc.b	$0F,$11,$2B,$F0,$8F,$28,$FA,$C7,$1A,$C6,$5F,$0E,$BE,$35,$71,$AF
	.dc.b	$01,$E9,$BE,$13,$E6,$3F,$21,$7D,$72,$FA,$95,$F2,$23,$18,$FB,$EB
	.dc.b	$67,$90,$EE,$49,$EF,$87,$07,$B9,$E2,$AF,$F0,$0E,$0F,$A2,$7D,$B7
	.dc.b	$F9,$32,$93,$98,$C3,$63,$FC,$3B,$3C,$1B,$3F,$ED,$F9,$FF,$71,$F8
	.dc.b	$0F,$67,$1C,$1F,$ED,$38,$C5,$FF,$AD,$6B,$A3,$FE,$38,$C1,$CB,$E0
	.dc.b	$6C,$78,$23,$04,$7F,$D2,$5A,$76,$7F,$CB,$6D,$AF,$74,$E6,$5F,$92
	.dc.b	$BE,$13,$FC,$54,$D8,$F8,$4D,$7F,$CD,$ED,$7F,$D8,$2E,$6B,$FC,$44
	.dc.b	$3F,$5E,$F3,$CF,$F9,$49,$B1,$FE,$27,$8F,$0D,$78,$EB,$96,$7F,$EE
	.dc.b	$5B,$DB,$70,$D5,$7A,$7E,$9B,$E4,$E1,$FE,$82,$6B,$FE,$D4,$6A,$23
	.dc.b	$86,$EB,$F1,$FC,$77,$D3,$BE,$CB,$0D,$AF,$F4,$BD,$F8,$35,$FB,$BF
	.dc.b	$51,$C7,$85,$F6,$54,$C4,$4B,$FF,$63,$46,$5F,$F6,$44,$7D,$3F,$A2
	.dc.b	$7D,$17,$F9,$6D,$B3,$F9,$36,$BF,$EA,$06,$BF,$EF,$27,$38,$7C,$DF
	.dc.b	$A8,$F4,$4F,$FB,$EC,$38,$8F,$A5,$3C,$38,$7E,$35,$23,$0F,$C9,$AE
	.dc.b	$13,$A6,$C7,$CD,$9D,$ED,$FE,$E8,$F6,$FE,$27,$7C,$AF,$F7,$34,$7C
	.dc.b	$73,$0F,$FA,$CA,$30,$FE,$8F,$9D,$A7,$E8,$49,$1F,$B0,$6B,$FE,$D7
	.dc.b	$F9,$D7,$C2,$FD,$B6,$C7,$39,$C7,$D6,$E7,$DE,$93,$FE,$FF,$A4,$EA
	.dc.b	$31,$3A,$7B,$46,$5F,$04,$67,$7F,$F4,$16,$8B,$E7,$5D,$7E,$53,$C9
	.dc.b	$F5,$52,$30,$43,$D4,$E5,$F6,$47,$AF,$EB,$7D,$9F,$5B,$27,$D1,$19
	.dc.b	$7D,$51,$97,$E2,$47,$E3,$7E,$33,$F1,$84,$8F,$C6,$35,$FC,$37,$1B
	.dc.b	$E7,$3F,$83,$36,$3F,$DE,$FC,$8A,$E9,$1F,$EF,$A9,$D3,$CA,$B2,$FD
	.dc.b	$5F,$8E,$FA,$84,$7E,$87,$AD,$FC,$4F,$E9,$DF,$01,$F6,$7E,$B6,$57
	.dc.b	$EE,$78,$BF,$84,$56,$A3,$EC,$7D,$2F,$DF,$23,$F6,$EB,$F0,$5F,$21
	.dc.b	$D7,$E0,$E4,$FD,$F8,$AF,$E2,$7D,$0F,$C4,$18,$C7,$ED,$F2,$FF,$91
	.dc.b	$67,$F0,$D3,$A7,$C9,$DD,$FA,$8C,$D8,$FA,$FD,$79,$5A,$3F,$89,$4A
	.dc.b	$3F,$88,$7B,$AF,$84,$DD,$7F,$50,$C3,$F8,$C7,$52,$9F,$0B,$C0,$DE
	.dc.b	$BF,$13,$1C,$61,$DE,$F8,$56,$B0,$FD,$EF,$D3,$FE,$92,$BF,$82,$37
	.dc.b	$0F,$E6,$8D,$9F,$E0,$7E,$9F,$82,$BD,$42,$74,$FF,$2D,$87,$8F,$FD
	.dc.b	$35,$B1,$C2,$76,$F1,$F0,$B5,$52,$35,$1F,$D5,$7E,$95,$F1,$14,$93
	.dc.b	$FB,$7F,$AC,$AF,$F0,$1C,$7F,$80,$18,$CF,$FA,$75,$F2,$9F,$DF,$4E
	.dc.b	$9F,$5C,$65,$FE,$4C,$BE,$2B,$C4,$48,$FE,$0D,$F4,$F5,$FE,$21,$F3
	.dc.b	$3F,$C3,$CC,$61,$E0,$AC,$4F,$F0,$D9,$CE,$DF,$CB,$A5,$E2,$FF,$73
	.dc.b	$6D,$7C,$E6,$BF,$CE,$7E,$15,$E3,$6E,$BF,$BD,$47,$F0,$E9,$D7,$91
	.dc.b	$97,$94,$EA,$24,$60,$FB,$7F,$4A,$F9,$3F,$F1,$04,$23,$0F,$F8,$8D
	.dc.b	$AF,$F4,$63,$FB,$63,$87,$0E,$B4,$FE,$FD,$F4,$CF,$AB,$FF,$08,$36
	.dc.b	$BC,$37,$FE,$17,$FA,$17,$C2,$3C,$4F,$F0,$B7,$E6,$5F,$DB,$F8,$7F
	.dc.b	$AA,$B7,$05,$F0,$AE,$57,$C6,$7F,$2F,$5B,$4F,$1A,$E1,$1F,$9C,$FA
	.dc.b	$4F,$C7,$FC,$2F,$F8,$1E,$DF,$49,$FD,$92,$73,$5E,$68,$C7,$52,$F8
	.dc.b	$47,$F6,$AF,$A5,$79,$7F,$DB,$31,$E7,$87,$F8,$4F,$1E,$71,$88,$46
	.dc.b	$2F,$AE,$1D,$27,$FC,$35,$87,$47,$07,$0C,$7C,$4F,$94,$EC,$E3,$0E
	.dc.b	$D8,$75,$9E,$78,$F9,$71,$EF,$1C,$4E,$98,$F6,$AE,$10,$C7,$C4,$EE
	.dc.b	$32,$4C,$7C,$17,$F0,$F0,$ED,$5E,$4C,$3A,$56,$A3,$07,$87,$0E,$B8
	.dc.b	$FA,$27,$53,$E4,$3C,$0E,$27,$78,$7A,$A7,$F2,$D8,$FD,$86,$D7,$C6
	.dc.b	$6E,$1F,$E7,$FF,$87,$79,$5F,$8C,$9F,$CF,$1F,$ED,$57,$0F,$E8,$63
	.dc.b	$7A,$71,$9C,$CF,$E1,$FA,$5E,$C7,$FA,$2F,$2F,$E1,$8D,$8C,$CF,$69
	.dc.b	$CB,$63,$3A,$FD,$86,$E9,$FE,$36,$FA,$5F,$E0,$7C,$5F,$56,$73,$7F
	.dc.b	$8D,$6B,$F8,$C6,$CF,$08,$FB,$11,$AC,$3C,$38,$77,$C7,$F1,$B7,$B8
	.dc.b	$C5,$F0,$9F,$01,$AE,$B6,$38,$9D,$12,$BF,$3D,$7C,$F2,$F8,$D8,$6E
	.dc.b	$7A,$9A,$C3,$9D,$F1,$AD,$B4,$FF,$3E,$5F,$37,$E0,$5F,$E9,$F4,$FF
	.dc.b	$5C,$FC,$AF,$FA,$53,$2F,$91,$BD,$BF,$E1,$3D,$0A,$C4,$31,$FD,$E5
	.dc.b	$FF,$79,$C3,$9B,$FD,$8A,$33,$3F,$DD,$CE,$75,$CA,$33,$7F,$E9,$FF
	.dc.b	$A3,$FF,$57,$5F,$FB,$35,$B7,$F8,$E6,$BF,$CB,$5A,$FF,$9D,$1D,$CF
	.dc.b	$3F,$82,$F5,$7C,$5F,$F7,$3E,$BF,$4D,$6B,$0F,$F3,$DF,$D1,$3D,$4D
	.dc.b	$FD,$1B,$C3,$F7,$CE,$6B,$FC,$2D,$8A,$B5,$1F,$E0,$09,$D6,$9F,$A4
	.dc.b	$6B,$FD,$0D,$AF,$83,$0F,$F5,$F3,$67,$94,$7D,$1F,$9E,$7D,$1B,$E7
	.dc.b	$FC,$73,$F4,$BE,$3F,$FC,$03,$7F,$E1,$BF,$9D,$FF,$28,$3F,$E1,$28
	.dc.b	$CE,$3F,$47,$D1,$56,$A3,$FD,$41,$3A,$7F,$AC,$35,$FF,$4F,$B5,$FF
	.dc.b	$63,$35,$3D,$B8,$BF,$DB,$0D,$BF,$F1,$AD,$3E,$8F,$D1,$E0,$7A,$7B
	.dc.b	$BF,$1D,$97,$93,$E5,$71,$87,$AD,$B9,$FF,$0E,$CB,$FC,$21,$3F,$C4
	.dc.b	$23,$C3,$7F,$29,$F2,$B5,$CC,$7C,$38,$EF,$97,$F8,$CA,$30,$7F,$8E
	.dc.b	$0F,$E3,$A7,$F9,$4C,$7C,$67,$FA,$A1,$D7,$83,$FA,$57,$3E,$8F,$E9
	.dc.b	$1E,$38,$7F,$24,$6B,$FE,$03,$6B,$FE,$CA,$6B,$FE,$F1,$78,$ED,$FF
	.dc.b	$00,$5A,$67,$E8,$DF,$5D,$DF,$0F,$D9,$F2,$3E,$8F,$E5,$FD,$27,$F9
	.dc.b	$BB,$0F,$F2,$9C,$E5,$FF,$30,$46,$72,$FF,$70,$4E,$5F,$E7,$6D,$7F
	.dc.b	$C3,$4D,$7F,$DC,$2D,$41,$E5,$97,$F1,$48,$C5,$F9,$3B,$7C,$EF,$D2
	.dc.b	$F3,$26,$3E,$6F,$98,$B7,$E4,$7E,$D6,$26,$92,$7F,$A8,$F2,$7E,$72
	.dc.b	$3E,$19,$F7,$8B,$1F,$A5,$E8,$FC,$09,$F8,$2F,$4A,$C7,$C6,$7B,$91
	.dc.b	$FC,$75,$A3,$F1,$D7,$FA,$AF,$9E,$FC,$59,$F8,$CF,$A4,$B7,$FE,$82
	.dc.b	$EF,$7D,$0F,$F2,$BF,$21,$F4,$8F,$8D,$FE,$57,$5A,$BF,$F5,$15,$6A
	.dc.b	$73,$7F,$9B,$AD,$6F,$FD,$C1,$AF,$F9,$3D,$AF,$D4,$0F,$0E,$3F,$A8
	.dc.b	$DC,$3E,$CE,$1F,$E6,$8F,$73,$BC,$7F,$0B,$EC,$FD,$3F,$7F,$EA,$1B
	.dc.b	$19,$3F,$88,$46,$5A,$7F,$13,$F7,$BF,$5F,$F0,$3F,$B0,$6C,$64,$FE
	.dc.b	$31,$19,$69,$FE,$B5,$BF,$F6,$FB,$FC,$74,$FE,$0E,$E3,$3F,$E3,$9D
	.dc.b	$FF,$DB,$27,$87,$C3,$FE,$F9,$DB,$5C,$9C,$CB,$3C,$9F,$EC,$1A,$BE
	.dc.b	$75,$F4,$3E,$0F,$86,$CC,$62,$BC,$D1,$89,$91,$E9,$ED,$B8,$7F,$31
	.dc.b	$CD,$F4,$9E,$AF,$43,$D6,$DB,$F8,$0D,$AE,$0D,$CF,$EC,$FD,$27,$9B
	.dc.b	$E8,$FE,$D5,$7E,$CB,$95,$39,$E6,$FC,$6B,$F8,$92,$63,$FC,$A6,$76
	.dc.b	$D8,$DB,$F9,$2E,$8F,$F5,$8B,$F9,$B8,$FA,$35,$F4,$BE,$92,$CF,$D7
	.dc.b	$6E,$1E,$F6,$E5,$FE,$16,$EC,$FF,$70,$C7,$F8,$2B,$DC,$F9,$55,$FD
	.dc.b	$57,$E2,$5D,$D9,$E3,$FE,$8B,$18,$CB,$FD,$DB,$7C,$32,$FE,$1C,$D8
	.dc.b	$FE,$27,$92,$BF,$A9,$63,$FC,$5A,$B4,$FF,$18,$FA,$3F,$DE,$5F,$F1
	.dc.b	$AC,$3F,$C5,$63,$81,$F4,$9F,$E2,$0A,$DD,$FF,$86,$EB,$53,$9F,$07
	.dc.b	$FA,$85,$6A,$BF,$D8,$4D,$7F,$68,$D7,$C5,$1C,$70,$FD,$07,$D1,$59
	.dc.b	$ED,$D2,$F3,$FF,$43,$F1,$9F,$92,$6D,$FA,$4F,$54,$ED,$C5,$33,$FC
	.dc.b	$E5,$7E,$AB,$3E,$55,$A9,$FD,$07,$8F,$FA,$EF,$9F,$FC,$FB,$67,$58
	.dc.b	$FF,$33,$6C,$7E,$D3,$9B,$71,$E0,$CB,$FD,$78,$7F,$81,$BE,$8B,$F8
	.dc.b	$A7,$4B,$E6,$FF,$CB,$3E,$9A,$7F,$D1,$7F,$03,$FE,$0E,$77,$3A,$73
	.dc.b	$6F,$A3,$E0,$FE,$86,$7F,$CC,$71,$88,$F0,$B7,$0F,$06,$3F,$D6,$23
	.dc.b	$53,$FE,$DB,$3A,$5F,$82,$74,$F8,$27,$95,$F6,$8C,$1B,$AE,$27,$FC
	.dc.b	$01,$C3,$8B,$E5,$BA,$FC,$4C,$79,$C7,$BE,$38,$1F,$1C,$E0,$53,$D7
	.dc.b	$5D,$F0,$F4,$C2,$FA,$E3,$EE,$8F,$19,$DE,$BC,$D8,$70,$AF,$04,$FF
	.dc.b	$46,$C7,$E7,$3E,$2C,$3C,$AF,$23,$E3,$38,$39,$3D,$B0,$F0,$46,$0D
	.dc.b	$61,$F1,$A7,$D7,$19,$8F,$51,$F7,$30,$EE,$74,$2B,$C6,$7C,$D8,$FC
	.dc.b	$A7,$C7,$1C,$9F,$43,$5F,$8E,$78,$1E,$C7,$C2,$2B,$F3,$A7,$C8,$EC
	.dc.b	$E6,$53,$99,$E2,$3E,$21,$F4,$4F,$85,$87,$E1,$5F,$B7,$1E,$A6,$CF
	.dc.b	$85,$EA,$73,$BF,$0B,$63,$E0,$61,$F3,$1B,$7D,$67,$C7,$39,$79,$B5
	.dc.b	$FC,$01,$ED,$9F,$7B,$E9,$BE,$F3,$B8,$F9,$67,$BC,$FD,$24,$7A,$5E
	.dc.b	$67,$19,$E7,$3F,$8F,$9F,$B1,$1E,$D9,$FE,$03,$F0,$1F,$E8,$67,$F3
	.dc.b	$AD,$73,$3F,$BC,$8C,$7C,$23,$E4,$BF,$8C,$52,$73,$18,$6F,$0F,$F9
	.dc.b	$49,$B3,$BB,$FF,$11,$5F,$3A,$FC,$FF,$D2,$3E,$43,$FD,$3B,$97,$F8
	.dc.b	$9E,$31,$87,$E9,$2F,$83,$FB,$16,$E1,$FC,$36,$77,$3F,$B5,$6C,$7F
	.dc.b	$11,$9D,$D7,$8B,$2F,$5E,$BC,$71,$FF,$1B,$37,$1F,$E0,$B7,$E0,$8E
	.dc.b	$55,$F0,$B0,$D5,$77,$BF,$4B,$FE,$72,$BE,$ED,$AE,$F8,$7B,$6B,$BB
	.dc.b	$67,$BC,$E9,$E1,$97,$F1,$6C,$BD,$B9,$78,$0F,$1C,$70,$35,$87,$4A
	.dc.b	$FF,$41,$D7,$0C,$3E,$05,$6E,$7D,$07,$D2,$AF,$4A,$C7,$D2,$DC,$7C
	.dc.b	$8B,$0E,$8F,$6A,$F0,$3B,$BE,$D8,$F9,$F1,$F7,$4F,$F2,$2C,$7E,$64
	.dc.b	$79,$23,$C6,$4C,$BD,$AF,$DF,$CB,$C3,$97,$82,$3C,$59,$78,$30,$E7
	.dc.b	$1F,$20,$FE,$5F,$3B,$38,$46,$A3,$CC,$77,$C3,$CB,$87,$80,$F0,$E3
	.dc.b	$E9,$9F,$71,$EC,$9F,$44,$66,$3C,$66,$4F,$13,$9B,$F0,$E3,$F2,$B0
	.dc.b	$F9,$B1,$F9,$6B,$F0,$9F,$0F,$0F,$0B,$F2,$2F,$B5,$7A,$4E,$38,$FD
	.dc.b	$8B,$E5,$8F,$C1,$C3,$8B,$E8,$3E,$1C,$FC,$7C,$7C,$47,$CC,$7E,$CD
	.dc.b	$79,$B2,$F5,$57,$B0,$E2,$63,$0F,$0B,$F8,$93,$F6,$D7,$FA,$5C,$BF
	.dc.b	$C7,$5F,$3A,$E8,$FF,$51,$61,$FE,$3C,$8C,$5F,$F8,$7D,$AF,$F9,$D7
	.dc.b	$E5,$F7,$C3,$CA,$EC,$BB,$DE,$3B,$DC,$CC,$61,$DE,$4C,$BA,$BE,$1E
	.dc.b	$9C,$C1,$A7,$2B,$FF,$7E,$E1,$E5,$8C,$69,$23,$07,$F8,$B7,$E9,$1E
	.dc.b	$57,$9A,$11,$87,$51,$D7,$E8,$F5,$39,$9F,$F6,$31,$FE,$FA,$70,$E1
	.dc.b	$AF,$FB,$2D,$AE,$E7,$DF,$F4,$9E,$B8,$FF,$25,$35,$FF,$28,$6F,$C6
	.dc.b	$7C,$DF,$47,$F2,$67,$FC,$8C,$D7,$60,$39,$D7,$FD,$15,$F4,$9F,$E3
	.dc.b	$D5,$7F,$ED,$F9,$D3,$FE,$82,$96,$5F,$59,$B7,$E0,$DF,$E8,$7D,$17
	.dc.b	$D1,$FA,$49,$3F,$ED,$18,$CB,$FE,$D3,$8C,$BF,$EB,$B8,$FA,$1F,$42
	.dc.b	$FA,$4A,$49,$FF,$31,$7D,$0A,$B1,$1F,$40,$57,$17,$06,$2A,$B4,$CC
	.dc.b	$6D,$F8,$E2,$BD,$AE,$1B,$AF,$C6,$8F,$99,$3C,$63,$FD,$69,$C6,$F6
	.dc.b	$FF,$25,$E0,$CB,$FF,$79,$47,$E5,$4C,$3F,$ED,$B8,$C3,$FA,$BC,$BF
	.dc.b	$3A,$47,$FD,$F4,$53,$EE,$F3,$BE,$81,$F0,$FF,$AA,$6C,$F3,$3F,$D3
	.dc.b	$33,$BF,$9A,$D7,$BF,$27,$34,$9E,$46,$63,$BC,$27,$31,$C2,$75,$A7
	.dc.b	$FB,$56,$B5,$39,$3A,$FC,$A7,$97,$E6,$E4,$60,$87,$67,$2F,$82,$3D
	.dc.b	$9F,$47,$ED,$FA,$49,$3E,$88,$CB,$EA,$8C,$BE,$88,$FC,$6F,$D2,$48
	.dc.b	$FF,$56,$71,$23,$F8,$07,$1F,$F0,$89,$FC,$1C,$E5,$FC,$3C,$FE,$2A
	.dc.b	$31,$1F,$91,$6B,$FC,$09,$CE,$7F,$87,$8F,$C6,$78,$5D,$8F,$13,$F3
	.dc.b	$3E,$3F,$EC,$11,$FA,$7E,$B7,$F1,$3F,$A3,$7C,$0F,$E6,$9B,$3F,$9E
	.dc.b	$6B,$FA,$36,$BF,$C0,$4F,$E6,$1C,$0F,$9D,$FC,$31,$23,$F0,$3F,$49
	.dc.b	$F8,$9F,$8F,$F0,$7C,$24,$9F,$C2,$C5,$7B,$47,$8C,$8F,$EC,$F9,$FF
	.dc.b	$29,$E5,$FE,$4C,$91,$FE,$DB,$C9,$5E,$DF,$66,$4E,$5F,$E8,$7E,$EB
	.dc.b	$E1,$B3,$3E,$D9,$F8,$44,$7E,$13,$C3,$E1,$E4,$73,$3F,$DD,$5F,$49
	.dc.b	$FA,$47,$F7,$ED,$E4,$FA,$CD,$F2,$3E,$47,$A6,$BF,$E8,$93,$C6,$FF
	.dc.b	$9F,$B5,$F9,$CD,$7F,$98,$35,$3D,$5F,$E6,$93,$FD,$89,$FF,$7F,$35
	.dc.b	$F2,$F8,$1F,$D4,$FE,$8D,$F1,$54,$93,$FB,$8F,$A4,$AF,$F2,$39,$DB
	.dc.b	$FC,$89,$AE,$67,$F9,$1C,$63,$3F,$DC,$61,$FC,$24,$E6,$FE,$66,$1F
	.dc.b	$19,$E3,$25,$FF,$08,$FA,$4A,$F8,$63,$A9,$1F,$13,$FC,$57,$E6,$A4
	.dc.b	$7F,$55,$BE,$D3,$FD,$C4,$78,$94,$93,$BD,$FF,$39,$F0,$8F,$17,$33
	.dc.b	$FD,$0E,$F8,$BF,$D0,$9A,$E6,$FF,$A1,$C6,$30,$FD,$B5,$FF,$21,$AE
	.dc.b	$C7,$F6,$97,$0F,$F4,$69,$E1,$87,$EB,$7E,$91,$F2,$7F,$E0,$96,$D7
	.dc.b	$20,$E1,$FE,$9F,$F4,$9F,$EB,$37,$51,$97,$FC,$16,$D7,$B4,$FF,$5A
	.dc.b	$8C,$43,$E8,$1F,$07,$FB,$D8,$F0,$7D,$27,$F7,$A7,$FC,$04,$D9,$DF
	.dc.b	$83,$FE,$34,$6C,$F0,$F5,$3F,$C4,$55,$BB,$E9,$3F,$D9,$BE,$8F,$F2
	.dc.b	$7F,$09,$FE,$0D,$97,$D1,$FF,$0F,$9C,$D7,$92,$31,$D1,$EF,$AD,$CF
	.dc.b	$F8,$57,$E9,$1E,$4F,$F1,$79,$CC,$B0,$E7,$18,$6E,$1D,$31,$F1,$3F
	.dc.b	$C7,$F0,$E7,$78,$3F,$C4,$55,$C1,$B8,$F8,$EB,$94,$79,$CF,$26,$3E
	.dc.b	$38,$CE,$3E,$1A,$E1,$5C,$B1,$EC,$B1,$F0,$BC,$CF,$54,$F0,$6C,$F0
	.dc.b	$C7,$C9,$1C,$71,$F0,$3D,$4E,$51,$A3,$77,$DB,$0F,$25,$F3,$C3,$B5
	.dc.b	$6B,$0E,$78,$78,$0F,$F1,$44,$7D,$67,$FC,$B2,$D7,$E1,$9F,$E5,$CF
	.dc.b	$83,$78,$FF,$8F,$8F,$F1,$84,$63,$6F,$F1,$7B,$63,$33,$B7,$AF,$27
	.dc.b	$F8,$95,$F0,$B5,$CF,$AD,$FE,$38,$C3,$F8,$7B,$67,$86,$9F,$81,$6E
	.dc.b	$DF,$D7,$7E,$93,$F8,$17,$37,$CB,$9C,$E3,$F8,$06,$BF,$8B,$6C,$F0
	.dc.b	$8F,$9B,$1A,$C7,$C3,$18,$C7,$EE,$D6,$F1,$F1,$47,$72,$4F,$DF,$9D
	.dc.b	$63,$FA,$3C,$3A,$12,$B9,$F5,$7F,$92,$F0,$E7,$87,$72,$3F,$A2,$C3
	.dc.b	$9C,$7A,$EB,$BC,$F0,$35,$5C,$F0,$F0,$4F,$E5,$3E,$8F,$FD,$3B,$9F
	.dc.b	$F1,$CE,$A7,$D3,$3E,$9B,$FC,$B9,$1F,$E1,$7A,$DE,$7F,$E8,$A8,$DC
	.dc.b	$EA,$FF,$C1,$95,$AA,$FF,$50,$B5,$CF,$C0,$FF,$88,$5B,$3F,$E7,$7B
	.dc.b	$7C,$93,$E5,$B5,$49,$CC,$61,$B5,$C6,$B8,$36,$7C,$95,$C6,$72,$FF
	.dc.b	$9F,$3E,$35,$F1,$DA,$C3,$9C,$FB,$69,$F2,$3C,$BD,$2F,$F3,$67,$C7
	.dc.b	$6E,$73,$9F,$F9,$9A,$31,$5E,$59,$CD,$FF,$B0,$B0,$F6,$CE,$49,$7F
	.dc.b	$46,$B7,$3F,$E4,$5F,$93,$78,$BD,$2B,$59,$FB,$3E,$E5,$F8,$AF,$FB
	.dc.b	$9E,$72,$7F,$56,$8C,$F8,$7E,$7A,$F2,$12,$7E,$43,$74,$F8,$0F,$5B
	.dc.b	$7B,$7C,$6B,$CD,$F0,$FE,$1F,$3D,$F4,$5F,$0E,$30,$D9,$FA,$91,$F0
	.dc.b	$C8,$E0,$E6,$79,$E7,$D5,$8F,$D1,$9F,$A9,$F8,$4F,$FA,$36,$BF,$D2
	.dc.b	$7E,$36,$3E,$17,$EA,$7B,$A8,$C4,$FE,$2E,$B8,$61,$F4,$5A,$FF,$98
	.dc.b	$78,$91,$88,$FF,$08,$3C,$23,$F4,$2D,$9F,$F7,$47,$03,$D3,$FD,$1D
	.dc.b	$6B,$2F,$3F,$91,$F8,$7F,$D7,$DF,$33,$8C,$7F,$D5,$F1,$87,$F6,$19
	.dc.b	$7F,$5E,$7D,$15,$F9,$D8,$C6,$7E,$98,$C1,$CD,$FF,$5F,$D7,$FB,$96
	.dc.b	$B9,$CF,$87,$BE,$47,$6E,$0B,$7F,$B9,$C1,$1D,$C9,$3F,$25,$F1,$7A
	.dc.b	$92,$3E,$63,$6B,$FD,$14,$3B,$19,$3E,$AF,$E1,$1F,$0F,$E9,$2E,$1F
	.dc.b	$03,$BB,$DC,$47,$EC,$3F,$C0,$2F,$9F,$C2,$FE,$7F,$F4,$AF,$F1,$5D
	.dc.b	$7C,$A8,$C7,$AB,$FB,$EC,$DE,$13,$E4,$30,$BE,$52,$C6,$1B,$87,$AA
	.dc.b	$87,$29,$FE,$22,$D8,$D3,$4E,$15,$DE,$3F,$C7,$8D,$8F,$E1,$DD,$7D
	.dc.b	$9F,$AF,$F3,$9F,$85,$3F,$0D,$F5,$16,$FF,$0B,$16,$7E,$EC,$DE,$9F
	.dc.b	$E6,$7F,$B0,$7F,$B5,$6F,$FD,$EB,$E6,$FF,$19,$8C,$5F,$FB,$D2,$31
	.dc.b	$B4,$AF,$C8,$FC,$95,$AF,$E7,$56,$8F,$CC,$6F,$CD,$B5,$FD,$16,$74
	.dc.b	$FF,$37,$6D,$7F,$1D,$6B,$FA,$D6,$BF,$E7,$B7,$78,$FA,$AB,$F4,$DF
	.dc.b	$3B,$DE,$FF,$4B,$F4,$FF,$A9,$3F,$55,$E4,$58,$C9,$31,$FD,$0C,$65
	.dc.b	$F5,$FD,$37,$EC,$4F,$88,$59,$F5,$38,$7F,$67,$F4,$36,$7F,$A2,$7C
	.dc.b	$B7,$E2,$3E,$5B,$F1,$2D,$AF,$BA,$FE,$13,$90,$D2,$47,$E1,$A3,$2F
	.dc.b	$03,$FD,$01,$5F,$E6,$51,$F4,$1F,$95,$7F,$23,$3C,$1F,$EC,$13,$F9
	.dc.b	$5F,$5F,$F2,$B5,$B9,$FC,$FD,$AB,$3D,$1B,$9E,$AF,$FB,$FF,$0E,$97
	.dc.b	$C6,$7C,$4F,$38,$CC,$F1,$8C,$5F,$5A,$E6,$7F,$00,$C3,$BB,$E8,$8D
	.dc.b	$4F,$58,$FF,$48,$9C,$0F,$3E,$1C,$1F,$F3,$E1,$9C,$3C,$2F,$53,$15
	.dc.b	$D3,$0F,$21,$EB,$71,$1B,$BF,$3D,$F5,$8F,$9D,$D7,$84,$F9,$71,$FD
	.dc.b	$FB,$93,$11,$E0,$9E,$4F,$AA,$B9,$1A,$2E,$1E,$48,$F8,$53,$C2,$FD
	.dc.b	$D8,$FB,$27,$DF,$7C,$1F,$45,$FA,$EF,$84,$78,$4F,$90,$7D,$68,$D3
	.dc.b	$83,$C8,$E2,$7D,$38,$7A,$E3,$F0,$4D,$9E,$E7,$98,$E5,$87,$9B,$1F
	.dc.b	$97,$1E,$6A,$E8,$7B,$5F,$B2,$70,$3C,$D7,$F0,$E3,$71,$F6,$6F,$C0
	.dc.b	$FD,$E6,$BE,$B3,$C5,$3C,$5B,$3E,$83,$DF,$3E,$C6,$9E,$3C,$7E,$64
	.dc.b	$F3,$3C,$B5,$CD,$B3,$BB,$F6,$C6,$EB,$8B,$F2,$8F,$09,$E9,$9E,$ED
	.dc.b	$3E,$45,$F5,$9E,$38,$7C,$9C,$7E,$CE,$3F,$76,$FE,$51,$D0,$F8,$58
	.dc.b	$74,$69,$D1,$F3,$4F,$27,$E5,$C7,$23,$D6,$7D,$A2,$9C,$8F,$AB,$87
	.dc.b	$C7,$7A,$1B,$C3,$CE,$FD,$89,$EE,$FD,$BB,$F4,$3F,$17,$1F,$8E,$FC
	.dc.b	$53,$B9,$FA,$C9,$F5,$34,$EE,$79,$EB,$BD,$74,$8F,$E0,$77,$F8,$26
	.dc.b	$D7,$26,$9C,$8E,$87,$B1,$FE,$39,$E4,$7E,$EA,$3F,$83,$35,$FE,$16
	.dc.b	$D7,$F9,$73,$9F,$71,$F1,$FF,$E8,$B9,$FF,$24,$5F,$F8,$01,$4E,$B9
	.dc.b	$3F,$C0,$77,$C6,$F8,$36,$33,$7C,$6B,$85,$FF,$0E,$C3,$A7,$D0,$7F
	.dc.b	$91,$E3,$F8,$7C,$63,$CA,$F0,$5F,$F3,$BB,$E2,$DC,$BC,$CF,$F3,$FC
	.dc.b	$39,$61,$D4,$FF,$0B,$B8,$8E,$18,$F6,$9F,$F0,$DD,$F9,$33,$E5,$8F
	.dc.b	$BD,$B3,$FC,$CE,$75,$97,$AF,$2F,$0B,$E7,$9E,$83,$2F,$04,$6F,$1C
	.dc.b	$DF,$2A,$F2,$E3,$DE,$B9,$47,$F4,$39,$D3,$F4,$30,$FA,$98,$F9,$23
	.dc.b	$FA,$46,$3F,$02,$3D,$B1,$C6,$77,$7D,$27,$81,$A3,$79,$73,$24,$79
	.dc.b	$EB,$E5,$1E,$1C,$BF,$AA,$61,$DE,$7A,$1B,$BE,$8E,$2B,$DA,$7C,$9C
	.dc.b	$3C,$6F,$D0,$7E,$4B,$A3,$BD,$7C,$87,$57,$F2,$27,$8C,$7C,$DC,$BE
	.dc.b	$43,$C7,$0F,$0B,$C0,$F8,$B1,$A7,$81,$E3,$A8,$F8,$23,$F1,$F8,$7D
	.dc.b	$2C,$7E,$2B,$ED,$3D,$93,$C8,$F9,$D5,$CA,$FE,$6C,$EB,$2F,$B2,$7D
	.dc.b	$AC,$BE,$45,$F9,$E7,$A3,$E7,$C3,$E8,$BF,$0D,$F0,$BE,$BB,$EA,$F1
	.dc.b	$31,$3F,$5E,$78,$C7,$95,$E2,$FD,$93,$E1,$9F,$6B,$1F,$89,$87,$9B
	.dc.b	$2F,$9B,$8F,$9A,$BD,$D8,$FC,$C7,$93,$E2,$73,$8F,$A9,$ED,$8F,$8A
	.dc.b	$BE,$D4,$62,$75,$1D,$4E,$D8,$78,$72,$F2,$54,$3E,$69,$30,$F0,$57
	.dc.b	$0C,$7F,$07,$7F,$37,$2F,$BD,$39,$C3,$EA,$63,$F6,$23,$18,$F4,$CB
	.dc.b	$CD,$7F,$80,$8F,$A5,$5E,$68,$E6,$7C,$6C,$7B,$9F,$19,$CC,$F6,$3E
	.dc.b	$99,$F5,$4F,$8A,$60,$F9,$33,$E3,$8E,$98,$71,$9D,$93,$1F,$63,$FE
	.dc.b	$B0,$8F,$F2,$73,$F9,$16,$E3,$FC,$81,$CC,$F0,$9F,$F5,$C4,$FF,$24
	.dc.b	$7F,$DA,$6D,$7F,$87,$98,$AF,$F8,$01,$B7,$F3,$4F,$12,$CF,$F2,$A6
	.dc.b	$BE,$18,$F2,$7E,$46,$F6,$F7,$92,$31,$5E,$9A,$DD,$74,$F0,$FF,$DC
	.dc.b	$7F,$03,$C2,$F7,$7F,$EF,$8F,$7B,$5E,$41,$D6,$18,$F8,$C9,$5D,$4F
	.dc.b	$F7,$75,$23,$8F,$C7,$6A,$7F,$D0,$E6,$A3,$1A,$7F,$9E,$73,$FF,$7A
	.dc.b	$D7,$8F,$E4,$A5,$F5,$AF,$64,$79,$35,$CC,$FF,$BF,$A7,$2F,$FB,$9D
	.dc.b	$AE,$74,$57,$FB,$9A,$72,$78,$7F,$55,$7C,$A7,$C5,$6D,$F8,$6F,$8B
	.dc.b	$7E,$17,$E4,$11,$F8,$59,$A9,$F2,$79,$9F,$83,$2F,$5D,$F5,$FF,$4D
	.dc.b	$B3,$D4,$E7,$6F,$A8,$F2,$12,$33,$F0,$75,$EB,$3B,$78,$CF,$AE,$7F
	.dc.b	$13,$D0,$F3,$29,$2B,$ED,$1F,$DA,$7E,$43,$0F,$E6,$A3,$FC,$24,$FE
	.dc.b	$69,$AF,$F8,$29,$AF,$F8,$01,$EF,$94,$8E,$B5,$E1,$7D,$BF,$51,$F9
	.dc.b	$EF,$71,$EE,$BE,$6F,$0D,$AD,$9F,$4F,$C2,$FB,$5C,$7F,$E2,$B8,$FD
	.dc.b	$1C,$60,$F1,$F0,$FD,$3F,$86,$FB,$84,$9F,$9B,$F4,$FF,$A8,$FD,$85
	.dc.b	$EF,$7E,$15,$B3,$F8,$B6,$BF,$8F,$6B,$FE,$35,$E9,$43,$DB,$E4,$49
	.dc.b	$E4,$F9,$39,$9F,$31,$F8,$F2,$4F,$E7,$EB,$93,$7D,$6F,$31,$1F,$C2
	.dc.b	$FC,$4D,$7C,$91,$E9,$23,$E9,$7F,$39,$F4,$B2,$3A,$1D,$FC,$6F,$84
	.dc.b	$47,$F7,$9C,$8F,$D8,$BF,$37,$1B,$3E,$24,$FC,$72,$38,$3F,$33,$F1
	.dc.b	$1F,$AB,$3A,$3F,$C0,$B6,$7D,$2F,$48,$C3,$63,$FD,$F8,$6B,$C0,$FC
	.dc.b	$37,$E3,$AE,$3F,$E3,$F5,$C6,$FE,$93,$5F,$E7,$0D,$57,$CD,$FE,$0D
	.dc.b	$1D,$FC,$54,$34,$60,$F0,$3C,$27,$83,$2F,$0F,$34,$BF,$E2,$D8,$A3
	.dc.b	$F3,$9F,$16,$F8,$5D,$57,$F1,$7A,$D3,$E2,$AF,$E2,$91,$88,$D4,$72
	.dc.b	$F4,$64,$76,$8F,$83,$C5,$23,$C2,$FF,$46,$FA,$7F,$44,$60,$F7,$C7
	.dc.b	$62,$38,$3F,$87,$FC,$5F,$F2,$13,$A3,$FC,$E0,$7D,$13,$E1,$BF,$A9
	.dc.b	$B6,$3F,$84,$B5,$E1,$F4,$4F,$A5,$DD,$7F,$06,$9E,$6D,$AE,$11,$FC
	.dc.b	$39,$AF,$FA,$87,$E6,$DD,$5C,$5B,$1F,$C4,$1A,$FF,$14,$6B,$FC,$61
	.dc.b	$CF,$37,$F7,$AF,$9C,$FE,$5F,$3E,$D6,$C7,$F7,$E9,$DF,$3F,$FA,$C7
	.dc.b	$E6,$1D,$6F,$73,$F0,$FE,$FF,$F1,$BC,$95,$A9,$CF,$9B,$C6,$F8,$52
	.dc.b	$F8,$7C,$D3,$E4,$3F,$D7,$D1,$CF,$1E,$55,$BC,$39,$63,$D2,$3F,$2F
	.dc.b	$3A,$AE,$77,$CA,$F7,$5C,$4D,$E3,$E2,$AE,$CD,$9F,$F6,$76,$3E,$3A
	.dc.b	$EC,$7B,$27,$4F,$2C,$73,$8F,$86,$F8,$E3,$E1,$AD,$1E,$13,$95,$70
	.dc.b	$BE,$70,$C3,$C0,$EF,$0E,$8F,$2C,$7C,$F7,$E4,$AD,$C6,$27,$85,$F4
	.dc.b	$BE,$31,$D7,$0F,$23,$E2,$3F,$AC,$C7,$CE,$7F,$AE,$B5,$F9,$27,$F6
	.dc.b	$0F,$75,$EA,$3E,$6B,$FA,$FC,$61,$B1,$99,$FE,$CB,$39,$7B,$FB,$8F
	.dc.b	$99,$FE,$CF,$87,$F8,$B5,$B1,$99,$DC,$60,$F9,$F1,$9C,$FF,$21,$F3
	.dc.b	$5F,$E4,$1E,$0C,$D4,$BF,$95,$7B,$E0,$FE,$3B,$0B,$E0,$EA,$B8,$61
	.dc.b	$F9,$1B,$E1,$3F,$C9,$2B,$73,$9A,$DD,$FE,$96,$F7,$7D,$7D,$E6,$3A
	.dc.b	$8F,$CB,$CE,$B2,$F8,$11,$F8,$5C,$BA,$E1,$CF,$1E,$13,$FA,$CB,$EB
	.dc.b	$5A,$8E,$73,$AB,$E7,$97,$C4,$CB,$C3,$5E,$D3,$F2,$5F,$33,$FC,$EF
	.dc.b	$3F,$F0,$BB,$63,$FC,$36,$D7,$FC,$3E,$D7,$E4,$BB,$E4,$FF,$41,$B6
	.dc.b	$7F,$A4,$FC,$23,$E8,$3F,$C7,$B1,$FE,$2C,$9D,$5F,$F5,$2F,$9A,$FE
	.dc.b	$8D,$3F,$E5,$D6,$C7,$F2,$FB,$79,$7F,$D8,$AB,$FD,$25,$5A,$C7,$FB
	.dc.b	$2B,$63,$3E,$16,$F0,$F1,$DD,$B5,$EE,$7C,$5E,$6E,$FD,$3F,$F2,$A7
	.dc.b	$A9,$BA,$4E,$7D,$D4,$E6,$26,$7F,$E4,$8F,$9A,$7C,$66,$32,$FE,$FD
	.dc.b	$8F,$2A,$F7,$5F,$96,$57,$C2,$72,$7C,$3D,$3D,$5E,$83,$E6,$9E,$77
	.dc.b	$96,$73,$18,$AE,$04,$C3,$FD,$1F,$7C,$7C,$9E,$13,$9E,$0F,$93,$87
	.dc.b	$29,$CD,$FA,$3E,$57,$C3,$E7,$FF,$95,$30,$FA,$D1,$8F,$6F,$E3,$32
	.dc.b	$3F,$C4,$F3,$E3,$AE,$51,$C0,$8F,$92,$35,$8F,$DA,$AF,$AD,$F3,$0F
	.dc.b	$9A,$FA,$1E,$3E,$1B,$8F,$FA,$8E,$74,$D8,$F8,$FF,$15,$F8,$EC,$15
	.dc.b	$AC,$BF,$C6,$0D,$8F,$F1,$6B,$6F,$FC,$43,$B2,$F9,$4E,$6F,$FC,$52
	.dc.b	$FF,$57,$AF,$F4,$FF,$C4,$BE,$6B,$D3,$E4,$7E,$2E,$31,$3E,$78,$CB
	.dc.b	$C3,$D0,$B8,$79,$26,$E7,$FE,$C3,$BF,$F0,$45,$F1,$3B,$FB,$FD,$4F
	.dc.b	$8F,$C1,$63,$E2,$38,$7E,$37,$C5,$DC,$7E,$0B,$F3,$8F,$37,$DB,$D9
	.dc.b	$F5,$FC,$DF,$A4,$F5,$7C,$15,$8F,$A1,$DD,$F5,$8F,$CA,$FC,$DF,$F8
	.dc.b	$52,$32,$7C,$4F,$99,$CF,$AB,$F0,$DB,$1F,$EA,$EF,$17,$FC,$54,$DF
	.dc.b	$90,$FD,$24,$63,$5F,$E0,$51,$8F,$62,$57,$1E,$BB,$5F,$C1,$E6,$F1
	.dc.b	$FF,$20,$86,$0A,$CD,$FF,$97,$BC,$67,$23,$E6,$FE,$C1,$F6,$7E,$6E
	.dc.b	$D7,$DE,$8B,$8F,$FB,$6A,$95,$8C,$BF,$CE,$FF,$36,$DB,$F8,$83,$71
	.dc.b	$FC,$3F,$CD,$E2,$64,$7F,$BF,$67,$58,$F5,$AC,$BF,$E8,$78,$47,$E3
	.dc.b	$78,$3A,$B6,$B9,$38,$7F,$2B,$F3,$76,$F9,$4D,$C7,$F1,$D3,$6B,$E2
	.dc.b	$66,$F9,$BF,$CD,$1F,$9B,$F2,$EC,$FE,$82,$2E,$3F,$0E,$6E,$1F,$0B
	.dc.b	$E6,$FF,$44,$7C,$22,$DF,$BD,$C3,$FA,$AF,$9B,$B7,$FD,$07,$E6,$BF
	.dc.b	$07,$F3,$7F,$84,$6D,$7E,$25,$C3,$FE,$F8,$9F,$F2,$98,$F9,$B7,$47
	.dc.b	$F8,$03,$9D,$7C,$2B,$B5,$F1,$7F,$CA,$87,$1A,$E1,$3F,$CD,$E3,$AD
	.dc.b	$F4,$7F,$AF,$C7,$F6,$E8,$C4,$7F,$96,$8E,$46,$AB,$FD,$5B,$1A,$8E
	.dc.b	$75,$DA,$72,$FF,$85,$A3,$27,$13,$99,$D6,$7B,$DF,$B2,$B8,$CF,$1F
	.dc.b	$5A,$BF,$A2,$C3,$CD,$87,$58,$F7,$E1,$CA,$73,$1C,$4E,$F8,$75,$7C
	.dc.b	$27,$90,$E9,$87,$67,$11,$BA,$F7,$47,$6B,$F7,$3C,$8C,$1E,$D8,$DE
	.dc.b	$1E,$8A,$E5,$5E,$78,$C6,$1E,$6B,$F8,$53,$D2,$F8,$C7,$A2,$74,$F9
	.dc.b	$1A,$F8,$0F,$49,$E3,$AE,$B8,$7A,$E3,$58,$78,$9F,$89,$7F,$1E,$75
	.dc.b	$5C,$27,$91,$E1,$8C,$B4,$CC,$F1,$6E,$1E,$93,$DC,$EE,$3D,$66,$CB
	.dc.b	$1B,$8F,$1B,$5F,$86,$7C,$B7,$9C,$71,$69,$B8,$F8,$93,$E6,$9D,$47
	.dc.b	$43,$91,$4F,$8E,$F8,$9C,$9E,$03,$33,$D9,$C9,$D8,$F7,$98,$3E,$6E
	.dc.b	$1F,$52,$FE,$55,$75,$9E,$67,$87,$0F,$A4,$F7,$6B,$D6,$38,$46,$23
	.dc.b	$AE,$1E,$DC,$3E,$04,$60,$F1,$9D,$8E,$98,$78,$DF,$A5,$1E,$B7,$47
	.dc.b	$77,$15,$DF,$0F,$4C,$62,$B9,$61,$F3,$2F,$AE,$1E,$37,$E1,$5F,$07
	.dc.b	$DD,$5C,$2F,$CF,$39,$9E,$27,$A3,$0F,$03,$E6,$3F,$71,$E4,$FF,$95
	.dc.b	$9B,$D9,$F8,$C8,$C4,$7F,$15,$6F,$B8,$F8,$77,$02,$BF,$DF,$D7,$FE
	.dc.b	$A8,$AD,$4E,$6F,$8D,$F0,$BD,$9C,$6B,$6E,$03,$8D,$7F,$93,$FC,$A7
	.dc.b	$F8,$BE,$7F,$8B,$B5,$FE,$3F,$3B,$D3,$C5,$7D,$EF,$8F,$95,$FE,$BC
	.dc.b	$6D,$FF,$20,$BE,$0F,$8E,$B8,$5F,$FB,$0B,$0E,$D7,$C2,$BB,$D6,$9E
	.dc.b	$19,$79,$31,$ED,$87,$16,$CE,$8E,$33,$B6,$18,$25,$F7,$25,$7B,$27
	.dc.b	$D1,$7E,$1A,$61,$E6,$AF,$04,$6A,$A3,$FD,$12,$74,$FD,$96,$BF,$42
	.dc.b	$BA,$63,$F1,$23,$AE,$3E,$A6,$9D,$31,$FE,$B7,$3B,$BE,$67,$47,$91
	.dc.b	$9A,$E5,$19,$BE,$E7,$AD,$E3,$1D,$E7,$8C,$E6,$38,$99,$78,$E1,$F2
	.dc.b	$AF,$E3,$E1,$F3,$DF,$2C,$F1,$8D,$63,$F2,$DF,$9E,$FC,$87,$E3,$5F
	.dc.b	$5B,$F7,$4E,$6F,$AD,$7C,$79,$CB,$F2,$EB,$CA,$E0,$D1,$D0,$F0,$9E
	.dc.b	$5A,$F1,$57,$4C,$7E,$25,$F1,$7E,$1D,$73,$C7,$E5,$3C,$8E,$B5,$EE
	.dc.b	$AC,$57,$29,$F6,$9E,$53,$85,$7C,$C9,$DE,$1E,$F9,$D9,$F7,$8F,$41
	.dc.b	$89,$F0,$B8,$BF,$03,$C2,$FC,$EE,$F1,$FA,$0E,$F1,$F7,$91,$F2,$19
	.dc.b	$3E,$21,$0F,$15,$7D,$0B,$F4,$46,$CC,$CF,$37,$E1,$BF,$5C,$8F,$38
	.dc.b	$F4,$CF,$3A,$FC,$4C,$F3,$BF,$91,$3F,$BF,$C7,$FD,$0D,$1F,$C3,$DB
	.dc.b	$8F,$F2,$1A,$FE,$47,$BF,$86,$B4,$FE,$2E,$B7,$3F,$C9,$DF,$F4,$86
	.dc.b	$3F,$29,$EB,$E2,$73,$6C,$EE,$B9,$1C,$7E,$49,$ED,$6E,$30,$D9,$FE
	.dc.b	$6B,$3A,$7F,$DA,$3F,$05,$9B,$4E,$67,$FD,$2D,$EF,$F1,$F0,$BF,$D3
	.dc.b	$EE,$E7,$55,$FE,$CF,$F9,$BD,$76,$1A,$63,$E6,$EE,$3F,$EE,$08,$AF
	.dc.b	$FA,$8E,$3A,$7C,$DC,$8F,$71,$E8,$7A,$EB,$98,$F9,$A6,$4D,$E1,$EA
	.dc.b	$9D,$C6,$31,$47,$AF,$E5,$5F,$37,$4E,$96,$C3,$E6,$7C,$37,$BC,$98
	.dc.b	$F9,$36,$60,$F9,$0C,$CF,$93,$0F,$97,$78,$7A,$42,$F8,$D7,$8F,$E6
	.dc.b	$E4,$F5,$C1,$1E,$7F,$92,$CC,$FC,$0B,$F0,$CE,$30,$EB,$1E,$4A,$FC
	.dc.b	$05,$7F,$42,$F8,$9B,$CE,$69,$8F,$9B,$B3,$F4,$0E,$05,$9E,$0D,$8D
	.dc.b	$BF,$48,$FE,$9F,$F3,$7D,$8C,$E3,$E4,$CD,$F3,$7F,$93,$F7,$1E,$F3
	.dc.b	$E6,$F0,$DC,$7E,$15,$7E,$3F,$C3,$91,$F6,$DD,$D7,$DB,$F8,$89,$3F
	.dc.b	$25,$FB,$DF,$31,$F9,$DF,$9A,$7B,$B4,$95,$F3,$FE,$AB,$5E,$DC,$8F
	.dc.b	$C5,$D2,$37,$1F,$47,$67,$CE,$7E,$2C,$91,$C2,$7C,$1E,$AC,$8E,$6F
	.dc.b	$E4,$7E,$2B,$11,$F8,$7C,$13,$8E,$2F,$91,$57,$B3,$FC,$59,$D3,$F9
	.dc.b	$0F,$98,$FC,$D1,$9C,$7F,$C6,$38,$BC,$5F,$80,$48,$FD,$5F,$33,$E7
	.dc.b	$3F,$31,$1E,$E9,$D3,$E3,$7E,$B4,$6A,$30,$7E,$57,$E2,$7F,$54,$70
	.dc.b	$D3,$F3,$A3,$E6,$F8,$38,$6E,$9F,$C7,$DB,$4C,$5F,$38,$E5,$FE,$29
	.dc.b	$5A,$C3,$A7,$83,$83,$F1,$55,$BC,$3F,$80,$D7,$4C,$39,$F8,$1A,$38
	.dc.b	$4E,$BA,$3A,$91,$F8,$BF,$1A,$F8,$6C,$47,$EE,$F0,$56,$33,$FD,$DE
	.dc.b	$2C,$38,$5F,$53,$9F,$A3,$F9,$F8,$FF,$33,$46,$4E,$9C,$7E,$E2,$3F
	.dc.b	$B1,$F9,$C6,$98,$8D,$9E,$E9,$FD,$EC,$EA,$30,$7F,$09,$F8,$DF,$E3
	.dc.b	$C7,$1C,$BF,$88,$C3,$E6,$DF,$0D,$FC,$19,$B1,$FC,$05,$AF,$EF,$7E
	.dc.b	$71,$F1,$9B,$A9,$3B,$7F,$87,$CE,$A3,$F8,$57,$CD,$FE,$71,$BD,$BF
	.dc.b	$C5,$23,$11,$FC,$27,$E7,$1E,$DE,$6B,$FB,$54,$CA,$DB,$63,$2F,$F6
	.dc.b	$F8,$CB,$FD,$C7,$E6,$3F,$45,$5F,$E9,$3F,$7B,$F8,$CC,$A7,$2A,$33
	.dc.b	$18,$77,$F3,$8F,$90,$FE,$F9,$1A,$F0,$3C,$F7,$C7,$D0,$F0,$46,$27
	.dc.b	$58,$F8,$67,$FC,$09,$87,$3B,$C1,$FE,$A7,$C7,$C1,$3C,$AB,$33,$E7
	.dc.b	$C3,$A4,$F2,$C7,$C6,$EE,$7C,$47,$39,$DB,$6B,$9E,$1E,$18,$D9,$D6
	.dc.b	$3B,$E1,$E2,$8E,$06,$A7,$86,$3E,$C8,$E3,$3B,$79,$1B,$C7,$CA,$F2
	.dc.b	$9E,$04,$C7,$D5,$1E,$0C,$7E,$5B,$F8,$0B,$FA,$AD,$8C,$BE,$F3,$F0
	.dc.b	$BE,$E3,$D3,$7C,$E7,$E1,$27,$FD,$8F,$18,$7F,$0F,$3B,$9D,$36,$3F
	.dc.b	$17,$6F,$6D,$F3,$7C,$E3,$0D,$8C,$CF,$06,$BE,$97,$39,$FE,$31,$B9
	.dc.b	$FF,$91,$7E,$73,$FC,$6F,$E2,$FD,$09,$CD,$FE,$0D,$AF,$E5,$DB,$3C
	.dc.b	$23,$E8,$C6,$B1,$F2,$C6,$31,$FC,$4D,$6F,$1F,$14,$77,$24,$FE,$0A
	.dc.b	$75,$8F,$EA,$70,$E8,$4A,$E7,$EF,$7F,$86,$B0,$E7,$87,$E2,$BE,$6F
	.dc.b	$F8,$BE,$7F,$E3,$2F,$9D,$FE,$DE,$FF,$82,$E3,$33,$FE,$25,$9C,$BF
	.dc.b	$E0,$98,$CB,$FC,$01,$AF,$F9,$AD,$A8,$38,$7F,$6A,$3E,$0D,$F0,$9F
	.dc.b	$E7,$3F,$85,$7C,$37,$F8,$CA,$73,$5F,$C3,$B2,$E9,$18,$85,$70,$CB
	.dc.b	$38,$7F,$10,$BB,$5E,$3A,$D3,$E1,$9D,$34,$D6,$1E,$18,$CE,$1D,$EB
	.dc.b	$8B,$E1,$AF,$E3,$07,$13,$C7,$5C,$5F,$41,$99,$E3,$5D,$0F,$81,$3B
	.dc.b	$9F,$8B,$5D,$E3,$11,$DA,$31,$87,$B2,$37,$5C,$A3,$11,$B8,$C3,$E5
	.dc.b	$30,$7B,$A3,$6F,$F2,$D3,$CD,$5E,$07,$15,$E0,$79,$18,$AE,$B1,$F0
	.dc.b	$E3,$07,$53,$78,$79,$F0,$F8,$31,$F1,$3D,$8E,$87,$19,$FC,$43,$8B
	.dc.b	$F8,$D3,$CE,$F4,$F9,$B1,$F4,$DF,$5C,$7D,$0D,$8C,$CE,$CF,$19,$6B
	.dc.b	$99,$ED,$2D,$FB,$0F,$0C,$7D,$C9,$E7,$3B,$9C,$E3,$EC,$AF,$A6,$79
	.dc.b	$E3,$38,$79,$23,$C1,$18,$BE,$6F,$78,$FC,$3D,$F5,$6C,$65,$EA,$78
	.dc.b	$CF,$5C,$6D,$CD,$F8,$DF,$BE,$6E,$3D,$AF,$94,$EA,$53,$8D,$F9,$E3
	.dc.b	$A4,$72,$9E,$8F,$DE,$3A,$57,$AD,$F1,$BE,$77,$85,$73,$8C,$47,$19
	.dc.b	$E7,$87,$D7,$8E,$27,$8C,$E3,$87,$C6,$70,$71,$BE,$8E,$27,$96,$1E
	.dc.b	$E8,$F2,$9E,$E2,$9E,$EA,$E8,$E0,$EE,$78,$8C,$1D,$DD,$DF,$A2,$3E
	.dc.b	$D9,$D5,$C4,$FC,$88,$E0,$73,$8C,$E1,$F0,$AF,$CE,$FC,$9A,$F4,$4E
	.dc.b	$AB,$D8,$76,$3D,$07,$3C,$3E,$0B,$E6,$BE,$6F,$8E,$33,$3C,$E3,$47
	.dc.b	$33,$DF,$5E,$09,$D9,$30,$F7,$B6,$7C,$11,$82,$38,$24,$F3,$8C,$B6
	.dc.b	$33,$87,$BC,$8F,$9E,$77,$18,$8F,$7C,$F1,$24,$66,$3D,$2D,$7E,$7C
	.dc.b	$69,$E8,$78,$0A,$E5,$EB,$1C,$70,$F9,$31,$83,$71,$97,$E4,$1D,$9C
	.dc.b	$4F,$DF,$9E,$33,$E4,$38,$90,$E8,$43,$45,$3A,$1B,$32,$43,$26,$DE
	.dc.b	$EE,$0D,$38,$23,$E8,$6C,$79,$E3,$04,$71,$D0,$CB,$FB,$2B,$67,$FA
	.dc.b	$DB,$5F,$EF,$6D,$E3,$78,$CF,$07,$FC,$05,$39,$8C,$57,$FA,$2A,$FF
	.dc.b	$C0,$CA,$F8,$DF,$09,$CC,$62,$78,$1F,$E0,$D7,$18,$7F,$80,$FD,$D7
	.dc.b	$C3,$74,$8C,$37,$4E,$79,$78,$A7,$FC,$15,$7C,$A7,$C7,$7C,$B0,$FE
	.dc.b	$DC,$D7,$FB,$F4,$ED,$F1,$5F,$3A,$E3,$87,$79,$FF,$11,$5F,$1C,$3B
	.dc.b	$DE,$E7,$8E,$5D,$AF,$84,$2F,$86,$1E,$8A,$FF,$00,$63,$E1,$AD,$5F
	.dc.b	$0B,$DC,$74,$3C,$F8,$77,$76,$77,$39,$B9,$AE,$93,$CA,$76,$72,$BD
	.dc.b	$9C,$5F,$F1,$04,$6E,$7C,$87,$2B,$F2,$1E,$C7,$37,$E3,$9E,$C7,$17
	.dc.b	$C4,$66,$78,$D7,$80,$F1,$1E,$07,$2E,$31,$F9,$B8,$79,$5F,$C0,$C7
	.dc.b	$17,$2F,$13,$58,$FC,$D7,$93,$F6,$9C,$9F,$2E,$38,$9D,$1E,$47,$03
	.dc.b	$51,$C5,$CC,$70,$8E,$CF,$12,$38,$24,$7E,$22,$BD,$B3,$C6,$3A,$99
	.dc.b	$3C,$A7,$EA,$6F,$E4,$DE,$30,$F9,$8F,$C1,$C3,$C9,$7E,$0B,$E6,$71
	.dc.b	$9E,$18,$7D,$7C,$39,$CF,$AA,$F8,$9E,$67,$81,$93,$C8,$7C,$33,$37
	.dc.b	$DD,$FC,$5B,$EB,$73,$1D,$1E,$31,$9C,$7F,$15,$1E,$E7,$F1,$38,$79
	.dc.b	$A3,$D8,$7A,$B1,$FC,$5E,$3E,$99,$CC,$F8,$30,$EC,$F8,$0E,$13,$F3
	.dc.b	$A3,$D9,$7E,$09,$F6,$1B,$25,$F6,$7A,$C6,$F0,$F7,$3E,$68,$E3,$3F
	.dc.b	$49,$F8,$8F,$17,$E3,$47,$1A,$F2,$BC,$6B,$E4,$57,$17,$E4,$3C,$5F
	.dc.b	$8D,$8F,$E9,$EF,$E3,$3F,$62,$38,$E1,$F0,$E7,$C2,$FB,$DF,$C0,$DF
	.dc.b	$85,$FA,$38,$7C,$2B,$E9,$1F,$0A,$7C,$93,$F2,$B0,$F2,$3F,$83,$7C
	.dc.b	$8F,$1C,$7E,$B5,$7B,$AF,$9E,$3F,$19,$F3,$57,$3C,$7C,$D3,$EE,$7C
	.dc.b	$D5,$F4,$71,$F3,$3F,$6C,$F4,$DF,$4B,$F9,$2F,$48,$F5,$90,$E4,$ED
	.dc.b	$C4,$FB,$1D,$63,$EC,$7F,$0A,$7C,$CC,$7F,$17,$1E,$E8,$F0,$C7,$BB
	.dc.b	$1F,$A6,$DC,$3E,$14,$FE,$0B,$1F,$6C,$7C,$77,$DB,$87,$86,$7D,$84
	.dc.b	$71,$87,$78,$F4,$DF,$57,$C2,$75,$3F,$27,$8F,$B5,$E4,$FE,$EB,$2F
	.dc.b	$E3,$04,$C7,$F9,$03,$F3,$E7,$DC,$7E,$18,$B9,$65,$FC,$15,$7C,$D7
	.dc.b	$CF,$87,$C7,$7F,$4B,$97,$B2,$75,$3E,$9C,$BD,$73,$9C,$BE,$5B,$6F
	.dc.b	$C0,$76,$8F,$F6,$F7,$6B,$BB,$FC,$0F,$32,$30,$E0,$07,$2E,$5A,$F1
	.dc.b	$8F,$F0,$8C,$B7,$FF,$09,$F7,$BD,$DC,$C6,$1B,$7D,$B0,$EB,$7C,$63
	.dc.b	$C3,$D6,$EE,$71,$79,$38,$DD,$7F,$E2,$0A,$D4,$E6,$31,$5C,$2D,$7C
	.dc.b	$1F,$6F,$80,$F0,$3F,$C5,$33,$99,$63,$D1,$E3,$87,$4C,$7B,$47,$9B
	.dc.b	$1E,$D7,$C7,$1E,$93,$A3,$9C,$EA,$F9,$47,$43,$95,$EE,$B8,$C6,$4E
	.dc.b	$31,$CF,$0F,$26,$1E,$76,$DF,$F9,$0A,$F8,$CE,$70,$F0,$61,$D8,$E9
	.dc.b	$7C,$1F,$87,$8F,$C2,$C3,$CB,$1F,$86,$9F,$84,$DA,$F9,$8D,$C7,$F1
	.dc.b	$0F,$F9,$47,$C0,$78,$18,$E3,$46,$27,$F1,$D1,$89,$D4,$E6,$7C,$1D
	.dc.b	$EE,$FE,$B9,$23,$13,$F9,$78,$C4,$EA,$73,$3C,$A3,$E4,$47,$0D,$7E
	.dc.b	$F7,$80,$F0,$3F,$21,$1F,$2E,$75,$B7,$E3,$5B,$87,$E5,$1A,$AF,$97
	.dc.b	$27,$6C,$64,$FC,$EA,$DD,$6A,$31,$3A,$C3,$C0,$FE,$1E,$75,$18,$9D
	.dc.b	$91,$F4,$E3,$E8,$C7,$CD,$7F,$8F,$EF,$79,$1F,$AB,$C3,$F7,$F1,$8C
	.dc.b	$FE,$BB,$63,$F8,$03,$5D,$CF,$EC,$F9,$39,$4E,$DF,$DB,$78,$0F,$9D
	.dc.b	$FD,$F6,$1F,$C2,$27,$77,$C6,$3F,$D7,$BE,$07,$FB,$DF,$1F,$F5,$33
	.dc.b	$7B,$FF,$D5,$44,$70,$48,$FF,$7C,$C6,$5B,$19,$7F,$DE,$71,$96,$C6
	.dc.b	$71,$FF,$7C,$13,$C1,$FF,$79,$57,$FB,$66,$BF,$D7,$E7,$90,$EA,$43
	.dc.b	$99,$67,$95,$FF,$BB,$A3,$FD,$9C,$7F,$B2,$49,$BF,$FB,$1B,$0F,$16
	.dc.b	$1E,$07,$E1,$CF,$38,$C3,$CA,$78,$39,$AE,$CF,$FB,$5C,$F3,$1E,$42
	.dc.b	$5F,$91,$FF,$7B,$1B,$C3,$E2,$46,$5F,$81,$1E,$63,$9C,$67,$0F,$2C
	.dc.b	$74,$C3,$CA,$F3,$8F,$5D,$F9,$23,$E4,$DF,$6A,$DC,$62,$B6,$F9,$0C
	.dc.b	$1E,$D3,$07,$8E,$79,$57,$58,$C4,$F2,$9F,$4E,$1F,$1A,$7C,$D5,$B8
	.dc.b	$C6,$1E,$B9,$F9,$87,$BB,$0F,$2B,$D0,$F1,$5F,$47,$57,$CD,$A7,$9A
	.dc.b	$FA,$D7,$26,$CE,$EB,$D3,$5B,$38,$BD,$4C,$97,$0F,$61,$F4,$30,$F4
	.dc.b	$4F,$CF,$3C,$B1,$9C,$3C,$11,$D2,$31,$7E,$17,$9C,$7C,$0B,$EF,$1A
	.dc.b	$8F,$24,$61,$D4,$7E,$D7,$99,$E0,$FF,$80,$F2,$FF,$01,$C6,$2B,$FC
	.dc.b	$13,$18,$9D,$47,$F7,$67,$FD,$85,$3B,$8F,$F4,$EB,$79,$DE,$0B,$6E
	.dc.b	$55,$FE,$EF,$AD,$4F,$F8,$C7,$C1,$FF,$12,$5F,$F9,$61,$B1,$9F,$07
	.dc.b	$FC,$87,$7F,$E4,$E3,$FC,$40,$7F,$88,$B4,$FF,$33,$63,$FE,$26,$8F
	.dc.b	$F1,$9E,$9F,$E4,$96,$9C,$67,$6F,$18,$FF,$31,$CF,$18,$EB,$3F,$E3
	.dc.b	$B3,$65,$9D,$C7,$13,$FC,$CB,$7F,$E3,$AB,$FF,$1B,$56,$E7,$37,$D6
	.dc.b	$39,$DF,$96,$7B,$B6,$33,$1E,$78,$EF,$1C,$E3,$FC,$AF,$3C,$E7,$27
	.dc.b	$12,$C7,$13,$C2,$79,$CE,$25,$C3,$E7,$CF,$F9,$9E,$3F,$CD,$27,$F9
	.dc.b	$9C,$FF,$4A,$39,$C7,$FC,$AB,$1F,$30,$FF,$2D,$95,$FB,$58,$7D,$88
	.dc.b	$C9,$EE,$3C,$87,$8B,$0F,$65,$F8,$A3,$C1,$1F,$E8,$66,$E1,$FE,$86
	.dc.b	$8F,$F5,$BC,$7F,$9E,$B0,$F2,$63,$F4,$63,$C4,$FE,$25,$D1,$C8,$FF
	.dc.b	$41,$CF,$27,$DA,$64,$F8,$2E,$4F,$11,$6B,$C4,$7F,$AC,$E3,$AC,$65
	.dc.b	$EB,$19,$7A,$C6,$5E,$B1,$98,$EA,$D3,$AC,$7F,$AB,$B0,$F9,$58,$FC
	.dc.b	$C8,$EF,$3B,$7B,$9B,$32,$53,$C2,$ED,$F0,$99,$AF,$0B,$63,$C2,$FF
	.dc.b	$A6,$2B,$C3,$7F,$2E,$39,$9B,$3C,$26,$CF,$09,$4D,$96,$BB,$B6,$BB
	.dc.b	$E3,$FE,$A1,$AF,$B5,$3F,$87,$3F,$13,$5F,$A4,$6D,$7E,$79,$A7,$E7
	.dc.b	$A3,$F1,$77,$E8,$C7,$F0,$A7,$E9,$E7,$55,$FA,$76,$9F,$94,$C3,$F2
	.dc.b	$57,$F5,$F1,$FE,$37,$7F,$EA,$EF,$07,$FD,$3B,$7F,$EA,$96,$F7,$FF
	.dc.b	$A7,$B7,$FF,$52,$EB,$89,$FF,$34,$D6,$CF,$F3,$5D,$7F,$AB,$9F,$F5
	.dc.b	$99,$93,$FC,$C1,$DB,$FE,$7C,$7F,$D0,$A7,$F9,$F1,$C7,$7F,$FA,$0B
	.dc.b	$A3,$FD,$1F,$C5,$FE,$83,$C7,$FD,$15,$5A,$EF,$FF,$40,$61,$FE,$85
	.dc.b	$AD,$3C,$EB,$FD,$29,$3A,$8E,$F3,$A2,$5F,$FA,$46,$75,$DB,$FE,$8D
	.dc.b	$AD,$CF,$88,$D3,$E2,$BF,$F3,$EE,$1F,$E8,$C8,$FF,$48,$90,$FF,$69
	.dc.b	$ED,$FE,$87,$7F,$D1,$3D,$DF,$E8,$77,$E2,$C6,$5F,$F6,$EB,$5F,$F3
	.dc.b	$29,$FE,$87,$EE,$FF,$42,$B8,$3F,$D0,$C7,$FB,$87,$EF,$9F,$1F,$FE
	.dc.b	$59,$8F,$F4,$97,$E0,$1F,$ED,$0E,$2F,$F2,$F1,$F5,$2B,$6E,$27,$7A
	.dc.b	$71,$53,$A7,$B5,$FF,$9D,$5E,$4D,$7F,$DE,$31,$B7,$FC,$E7,$5D,$78
	.dc.b	$DE,$2B,$DE,$ED,$E0,$7F,$9D,$6B,$FC,$E5,$39,$7C,$96,$9D,$3F,$DC
	.dc.b	$67,$4F,$F7,$39,$FC,$7C,$EA,$FB,$DF,$6E,$FF,$35,$F3,$6D,$7F,$A2
	.dc.b	$63,$8C,$69,$FF,$1F,$9F,$E3,$63,$FC,$23,$D5,$FE,$16,$7F,$C3,$C6
	.dc.b	$7A,$7F,$C4,$8F,$F8,$80,$FF,$0A,$94,$FF,$31,$76,$A7,$F4,$BD,$EE
	.dc.b	$F6,$8B,$FF,$0D,$D7,$F8,$5A,$F8,$46,$6F,$85,$6E,$7C,$33,$E3,$AF
	.dc.b	$F1,$46,$18,$BF,$F1,$AC,$FB,$1A,$FF,$92,$CE,$07,$A0,$A7,$B9,$DF
	.dc.b	$4F,$F9,$22,$3E,$19,$ED,$3A,$63,$FE,$3E,$AF,$F1,$B4,$71,$8C,$5F
	.dc.b	$C9,$8D,$E3,$FE,$41,$79,$1E,$3C,$3B,$BF,$E4,$73,$C8,$E2,$FA,$C7
	.dc.b	$23,$C6,$7A,$0C,$3F,$06,$FC,$31,$DB,$0F,$54,$E6,$FC,$53,$C4,$F1
	.dc.b	$57,$07,$15,$C3,$0F,$9A,$48,$F2,$CE,$F0,$F3,$4F,$BE,$79,$DF,$69
	.dc.b	$CC,$78,$CF,$71,$F1,$E3,$55,$CC,$8E,$09,$1F,$02,$79,$BE,$63,$99
	.dc.b	$F0,$AB,$CC,$F5,$38,$57,$47,$FD,$01,$8F,$E3,$95,$7E,$39,$FC,$99
	.dc.b	$FE,$6A,$29,$A9,$FC,$CB,$83,$C0,$7C,$4B,$F3,$E3,$F3,$E3,$C3,$18
	.dc.b	$C3,$FC,$E7,$CF,$FE,$71,$9F,$D1,$65,$FE,$6E,$7F,$A5,$B5,$C9,$FE
	.dc.b	$79,$3F,$CE,$59,$7F,$9E,$88,$E0,$FE,$C5,$DD,$FD,$A5,$FF,$12,$73
	.dc.b	$7F,$90,$1F,$EE,$8F,$F8,$4D,$FE,$FB,$DD,$FD,$E7,$87,$FB,$74,$7F
	.dc.b	$8F,$72,$FE,$73,$DF,$FE,$83,$8F,$87,$4C,$3F,$B0,$E3,$F2,$E3,$13
	.dc.b	$FB,$7B,$E3,$85,$8F,$29,$FE,$0E,$9F,$03,$FE,$5B,$BF,$46,$BF,$DE
	.dc.b	$71,$FF,$05,$BF,$E0,$9C,$BD,$0F,$F8,$4A,$F8,$BF,$E0,$8B,$E9,$7F
	.dc.b	$E7,$AC,$7F,$C0,$D1,$EB,$3C,$27,$C0,$8D,$38,$9E,$91,$88,$FF,$0A
	.dc.b	$9C,$4B,$5F,$E0,$EA,$FF,$1A,$5F,$CD,$BF,$36,$1F,$E2,$F7,$9D,$FF
	.dc.b	$97,$31,$FF,$0D,$BE,$F3,$CE,$64,$FA,$F3,$F0,$2F,$FC,$4D,$7F,$01
	.dc.b	$FF,$17,$E3,$E5,$8C,$57,$F8,$EB,$1F,$87,$5F,$01,$F3,$47,$F8,$FC
	.dc.b	$FF,$24,$9F,$E3,$93,$85,$73,$BF,$45,$76,$9F,$26,$3F,$4F,$1F,$C2
	.dc.b	$E1,$D2,$79,$1D,$1F,$0B,$F2,$EB,$93,$F9,$0B,$EA,$F8,$0E,$B1,$F2
	.dc.b	$A3,$E0,$47,$F8,$F0,$F9,$F8,$7E,$12,$75,$7E,$28,$C4,$FB,$71,$FA
	.dc.b	$55,$F2,$2F,$BC,$F9,$6B,$DD,$7F,$4B,$1F,$C4,$4F,$E0,$CF,$19,$CE
	.dc.b	$BC,$6F,$E0,$70,$F9,$EF,$33,$67,$C2,$7C,$2F,$A5,$D9,$87,$94,$7E
	.dc.b	$AF,$2F,$D4,$61,$FA,$AC,$7F,$C9,$79,$7E,$AA,$FF,$88,$E1,$FE,$5A
	.dc.b	$6E,$1E,$C8,$E9,$3F,$4C,$FA,$33,$F2,$0F,$C6,$9F,$47,$2F,$81,$7F
	.dc.b	$85,$C3,$E4,$CF,$A7,$0F,$DF,$65,$FC,$97,$BB,$FC,$BD,$3F,$DD,$DA
	.dc.b	$FF,$97,$CF,$F3,$09,$4F,$F3,$1B,$FE,$6B,$7F,$BF,$78,$CF,$1B,$FC
	.dc.b	$E2,$D8,$FF,$10,$CE,$DF,$F3,$56,$5F,$CB,$94,$ED,$FF,$11,$1D,$2B
	.dc.b	$FC,$EC,$FF,$8A,$BE,$DD,$F0,$9F,$E7,$86,$C7,$98,$FF,$08,$38,$70
	.dc.b	$D7,$D6,$E6,$FC,$5E,$6B,$CB,$FF,$44,$B6,$BC,$17,$FE,$8C,$F0,$9E
	.dc.b	$17,$F9,$9E,$7F,$C3,$D1,$8E,$5F,$F3,$59,$23,$FC,$CF,$E5,$79,$57
	.dc.b	$27,$F9,$95,$FF,$36,$F8,$7F,$E6,$B6,$CE,$ED,$E1,$FF,$9A,$27,$BB
	.dc.b	$54,$6B,$C8,$E1,$39,$F0,$7F,$CB,$55,$FE,$5E,$EB,$F1,$46,$23,$FC
	.dc.b	$BF,$E1,$7F,$93,$23,$FC,$CD,$C3,$F5,$5B,$5F,$E3,$7F,$21,$F1,$7F
	.dc.b	$E6,$D9,$CE,$1F,$E3,$FF,$0B,$FC,$B1,$E2,$7F,$97,$31,$FB,$D1,$8F
	.dc.b	$13,$FC,$B5,$97,$C3,$8F,$F2,$C5,$F1,$F1,$3F,$CB,$64,$9F,$F3,$4C
	.dc.b	$FF,$9A,$3C,$46,$9F,$8F,$9C,$D7,$F9,$FE,$FF,$BB,$9F,$E0,$5E,$DF
	.dc.b	$EE,$C4,$F0,$3F,$C1,$31,$FD,$C3,$C2,$FF,$03,$63,$FE,$05,$7F,$C1
	.dc.b	$11,$99,$FF,$2E,$F8,$9F,$E0,$49,$E5,$3F,$E0,$33,$FC,$02,$4B,$CB
	.dc.b	$E1,$AE,$EF,$F8,$A0,$D3,$8F,$13,$FC,$65,$1C,$E7,$8B,$EA,$21,$C5
	.dc.b	$CB,$67,$93,$FE,$26,$30,$73,$73,$5E,$EB,$F5,$47,$F9,$CC,$F6,$E5
	.dc.b	$FD,$E6,$BD,$07,$12,$18,$21,$E4,$C3,$D2,$E0,$3C,$F1,$F9,$95,$DE
	.dc.b	$7F,$D0,$17,$E2,$7A,$9C,$CF,$55,$74,$7E,$91,$C2,$FC,$CE,$0F,$F1
	.dc.b	$C3,$87,$E0,$9E,$38,$E4,$F9,$0F,$F1,$FD,$F9,$1F,$F2,$16,$1F,$16
	.dc.b	$B9,$57,$02,$4F,$0B,$F4,$3F,$E3,$A3,$13,$C1,$F4,$98,$3F,$C9,$47
	.dc.b	$86,$BB,$38,$AE,$B1,$83,$F2,$31,$BC,$3E,$AB,$B3,$AC,$6E,$3E,$64
	.dc.b	$7C,$23,$58,$FF,$A3,$1F,$86,$7C,$D9,$E0,$FC,$09,$EB,$1F,$8C,$AF
	.dc.b	$B1,$87,$BA,$BE,$76,$3F,$1D,$F9,$87,$BB,$0E,$D5,$C3,$1F,$85,$87
	.dc.b	$6A,$DD,$69,$F8,$91,$EE,$AE,$11,$E8,$C7,$FD,$1D,$1F,$42,$B8,$12
	.dc.b	$33,$3E,$16,$C6,$63,$FD,$25,$D3,$FE,$71,$7F,$00,$DA,$FF,$38,$46
	.dc.b	$2F,$FC,$DA,$FF,$34,$F0,$BF,$CC,$54,$9C,$F5,$7F,$75,$8C,$57,$F8
	.dc.b	$84,$FF,$0F,$1F,$DC,$4F,$F1,$EF,$0F,$F8,$65,$FE,$EB,$8F,$F9,$22
	.dc.b	$3F,$BA,$C7,$13,$FC,$1C,$71,$8F,$F1,$92,$7F,$D3,$55,$FE,$05,$BF
	.dc.b	$F0,$5C,$FE,$DF,$C4,$FF,$05,$46,$B1,$FF,$08,$67,$FE,$01,$AF,$0B
	.dc.b	$FE,$27,$C7,$87,$89,$FE,$13,$9F,$F7,$B6,$1C,$AB,$FC,$F7,$1F,$E1
	.dc.b	$2A,$F3,$BF,$E0,$F3,$FC,$54,$7B,$F0,$F1,$BF,$E8,$8B,$E8,$FC,$B9
	.dc.b	$FE,$19,$8F,$F8,$9B,$1F,$F0,$8E,$1D,$B0,$E2,$47,$07,$A6,$FB,$E5
	.dc.b	$F2,$5F,$9A,$7C,$58,$DB,$EA,$C7,$C3,$97,$AE,$07,$8C,$F6,$1C,$AB
	.dc.b	$E4,$65,$F1,$22,$1A,$C3,$FC,$93,$3A,$8F,$01,$D3,$2F,$A9,$8F,$F8
	.dc.b	$F3,$1F,$09,$F0,$A3,$33,$F7,$DF,$F1,$F5,$73,$BF,$81,$3E,$93,$E3
	.dc.b	$E5,$F6,$9E,$0F,$F9,$5B,$2F,$F2,$45,$78,$CE,$58,$F8,$6B,$D3,$78
	.dc.b	$78,$E5,$F9,$37,$E5,$CF,$E0,$DF,$B0,$70,$7E,$5B,$F7,$EB,$FD,$3A
	.dc.b	$FE,$56,$3F,$CE,$AF,$FA,$74,$F8,$67,$CD,$8F,$02,$BF,$A4,$DC,$BE
	.dc.b	$26,$3E,$D8,$ED,$7F,$E8,$AB,$FB,$C7,$E8,$8F,$A2,$E4,$F4,$39,$7F
	.dc.b	$0C,$F9,$8F,$C3,$63,$F3,$27,$C0,$7C,$CC,$63,$F8,$1A,$E2,$75,$7F
	.dc.b	$CF,$97,$FE,$71,$3E,$86,$1F,$E6,$AC,$7F,$CE,$39,$7E,$6F,$0F,$99
	.dc.b	$97,$87,$1F,$29,$F1,$9F,$4E,$3F,$C1,$9F,$6C,$64,$F0,$C6,$5E,$2F
	.dc.b	$FA,$56,$7A,$C7,$96,$3F,$D1,$B1,$FB,$4A,$F1,$47,$D3,$AF,$13,$EA
	.dc.b	$9D,$63,$F3,$F0,$FA,$67,$D4,$BF,$CA,$D7,$D2,$3C,$47,$13,$E3,$1F
	.dc.b	$8A,$3B,$61,$FC,$1B,$0F,$F4,$57,$93,$FE,$76,$9F,$F3,$C9,$FE,$7E
	.dc.b	$DB,$F6,$2F,$F9,$FB,$5D,$3F,$E8,$CF,$17,$FD,$1E,$4A,$FF,$33,$47
	.dc.b	$13,$FD,$04,$7E,$EB,$4F,$F4,$47,$8B,$FE,$87,$7F,$D1,$55,$FE,$8D
	.dc.b	$F1,$3F,$CF,$98,$F8,$1B,$7F,$E8,$4F,$0B,$FC,$FB,$F1,$9E,$4C,$3F
	.dc.b	$D2,$55,$FE,$85,$AE,$15,$FE,$8D,$23,$FE,$7B,$3F,$D4,$7E,$07,$F9
	.dc.b	$FD,$FF,$48,$35,$CF,$87,$FE,$85,$7F,$CF,$FE,$1F,$F9,$F5,$BE,$1F
	.dc.b	$FA,$0F,$8B,$FD,$05,$18,$39,$F8,$5F,$E7,$DC,$3F,$CF,$91,$8F,$0F
	.dc.b	$FC,$FA,$7F,$9E,$BC,$3F,$F3,$90,$DF,$E3,$1F,$E7,$7F,$0F,$2E,$2B
	.dc.b	$FC,$FE,$D5,$E1,$FF,$9E,$BB,$7F,$CD,$B5,$C2,$7E,$DF,$83,$FE,$76
	.dc.b	$4D,$27,$81,$FE,$77,$8D,$BE,$37,$FD,$77,$5F,$E7,$8F,$1B,$FD,$03
	.dc.b	$1A,$C7,$FC,$F2,$79,$67,$4E,$23,$F1,$67,$F9,$17,$AB,$FB,$F3,$FE
	.dc.b	$1F,$7F,$C3,$0F,$F1,$29,$78,$3E,$19,$FF,$07,$4F,$F8,$43,$B3,$FC
	.dc.b	$4F,$5F,$E1,$B8,$E9,$1F,$E0,$CF,$03,$FC,$31,$1F,$E2,$0B,$FF,$11
	.dc.b	$3F,$E2,$A2,$78,$3F,$E1,$89,$EA,$70,$8C,$9F,$E3,$6A,$FF,$16,$9E
	.dc.b	$0A,$E7,$5F,$E1,$D8,$F0,$1A,$9F,$F1,$A1,$FE,$5F,$3F,$C4,$85,$C3
	.dc.b	$DE,$7F,$92,$8F,$0C,$71,$8F,$F2,$A5,$F8,$EB,$B4,$71,$9F,$F7,$34
	.dc.b	$7F,$64,$AE,$51,$E5,$C7,$FC,$72,$F3,$AF,$8A,$71,$3F,$76,$7F,$9A
	.dc.b	$E7,$8B,$E8,$3F,$CA,$57,$E4,$7D,$4F,$A2,$7D,$F1,$87,$C0,$FB,$8F
	.dc.b	$B8,$FF,$A7,$27,$EF,$E3,$F7,$2F,$D1,$39,$3F,$C8,$24,$AF,$31,$FE
	.dc.b	$50,$70,$F4,$3F,$D1,$E7,$D6,$AE,$B8,$FE,$1E,$7E,$2C,$FE,$2E,$30
	.dc.b	$FB,$6F,$F1,$0F,$F9,$D4,$F5,$1D,$31,$F9,$EF,$73,$C4,$79,$0F,$96
	.dc.b	$64,$FF,$0A,$BE,$C7,$47,$53,$E5,$D7,$C6,$C7,$E5,$46,$A7,$BD,$7C
	.dc.b	$58,$FF,$66,$9E,$33,$F4,$07,$A6,$FF,$22,$FF,$9A,$4E,$45,$3C,$87
	.dc.b	$86,$3D,$F3,$F9,$FB,$FF,$65,$9F,$85,$9D,$57,$4C,$3E,$5E,$3E,$B9
	.dc.b	$F8,$27,$4B,$E4,$F8,$E3,$FD,$51,$3C,$23,$FC,$16,$7C,$43,$E9,$47
	.dc.b	$37,$F6,$8F,$E5,$9F,$A2,$76,$C3,$E1,$4F,$F0,$03,$F2,$B1,$FA,$4C
	.dc.b	$3D,$17,$E2,$8E,$18,$7D,$E6,$E1,$DD,$F6,$BD,$DF,$A6,$7C,$DC,$7F
	.dc.b	$5C,$FC,$E3,$B1,$E4,$3C,$E7,$FA,$E7,$BB,$FD,$24,$9A,$48,$FF,$47
	.dc.b	$F8,$1F,$E9,$26,$F5,$FF,$96,$A7,$FD,$23,$5F,$E5,$B9,$FF,$0F,$78
	.dc.b	$1F,$DE,$EF,$FC,$59,$CF,$FD,$F2,$3F,$BF,$F8,$3F,$E3,$0C,$7F,$C4
	.dc.b	$3B,$7F,$8A,$71,$FF,$2D,$61,$FE,$05,$AF,$F2,$CC,$7F,$89,$B1,$FF
	.dc.b	$0C,$BF,$E9,$6A,$FF,$0C,$BF,$E1,$53,$FC,$83,$8F,$7A,$E3,$1F,$E6
	.dc.b	$9D,$BF,$C3,$AF,$17,$FC,$A1,$7E,$89,$F2,$4F,$F9,$D2,$FC,$8F,$F8
	.dc.b	$4C,$FF,$26,$57,$F9,$53,$2F,$5C,$79,$63,$AE,$1F,$E6,$C6,$E3,$FE
	.dc.b	$38,$9F,$C9,$4E,$5F,$7E,$5F,$E5,$F7,$FC,$E0,$79,$CF,$F0,$F6,$1F
	.dc.b	$E5,$AC,$3B,$C6,$4F,$25,$73,$6D,$F7,$7E,$1B,$FE,$9E,$7B,$9F,$E6
	.dc.b	$E2,$65,$FE,$66,$9F,$F1,$CD,$FC,$E7,$A1,$8C,$7D,$78,$FF,$90,$AF
	.dc.b	$9D,$79,$72,$F9,$F8,$73,$C7,$FC,$B1,$7B,$BF,$A0,$7F,$8E,$EF,$E6
	.dc.b	$E3,$D6,$3D,$F8,$FC,$DC,$79,$BF,$E8,$29,$D4,$FB,$AF,$A4,$79,$B0
	.dc.b	$EF,$3F,$E4,$43,$FC,$92,$7F,$95,$8F,$F2,$49,$C4,$D1,$E4,$35,$87
	.dc.b	$92,$39,$BF,$0D,$F9,$51,$F9,$FB,$FF,$4F,$3F,$29,$F3,$47,$07,$E3
	.dc.b	$C7,$D2,$BF,$BC,$FF,$A0,$8F,$B5,$7E,$C2,$57,$E8,$63,$71,$FE,$64
	.dc.b	$33,$3F,$E6,$A3,$9E,$5F,$AF,$7F,$57,$7F,$E7,$BA,$F9,$67,$E0,$32
	.dc.b	$FC,$F3,$F9,$73,$27,$C3,$C7,$ED,$3E,$0C,$7F,$72,$7A,$DD,$E3,$FA
	.dc.b	$D7,$58,$78,$1F,$D9,$57,$C7,$8F,$CF,$D7,$AE,$7E,$31,$E3,$C3,$E2
	.dc.b	$D6,$E3,$E3,$1A,$3E,$36,$1F,$8B,$BF,$81,$1F,$6B,$0F,$86,$FD,$CE
	.dc.b	$F7,$83,$FE,$81,$AF,$F3,$9B,$FE,$6D,$DB,$FD,$03,$E0,$FF,$1F,$BF
	.dc.b	$F2,$85,$F0,$6D,$71,$AD,$EF,$FE,$86,$9F,$F4,$4D,$7F,$A1,$5F,$F0
	.dc.b	$89,$FE,$FC,$79,$38,$F0,$3F,$CF,$E9,$A6,$3C,$57,$85,$E5,$9F,$F4
	.dc.b	$21,$FE,$63,$9D,$F8,$5F,$E8,$68,$FF,$33,$77,$7F,$9F,$5F,$F4,$DB
	.dc.b	$50,$70,$DF,$0B,$FD,$05,$BF,$F9,$F8,$9E,$17,$FA,$0D,$FF,$40,$D7
	.dc.b	$F9,$FA,$BF,$D1,$BE,$0F,$FA,$04,$9E,$0F,$C0,$8F,$F3,$8F,$6F,$F9
	.dc.b	$B7,$8E,$78,$AF,$17,$FC,$DB,$18,$7F,$CD,$F5,$FE,$38,$F0,$3F,$CE
	.dc.b	$FC,$9C,$DF,$F3,$E7,$DA,$BE,$A7,$A4,$FB,$3C,$17,$DB,$3E,$27,$38
	.dc.b	$FC,$0C,$38,$5F,$F9,$E6,$B4,$FF,$A3,$63,$33,$FE,$82,$9C,$BF,$EC
	.dc.b	$68,$CE,$7F,$E8,$7C,$3B,$E1,$DA,$95,$C9,$FF,$28,$B5,$FF,$66,$B5
	.dc.b	$F2,$63,$FE,$91,$7E,$47,$87,$FE,$26,$F0,$FF,$BF,$47,$F8,$CB,$0F
	.dc.b	$F0,$3F,$81,$FE,$4B,$8F,$F1,$1F,$81,$FE,$13,$3F,$C5,$EF,$F8,$64
	.dc.b	$FF,$09,$9F,$E1,$63,$FC,$0A,$3C,$0F,$F3,$03,$FE,$06,$3F,$C1,$67
	.dc.b	$F8,$D0,$B1,$D2,$BC,$31,$FE,$37,$8F,$05,$78,$E3,$C1,$18,$75,$1F
	.dc.b	$E5,$C9,$E3,$1D,$23,$84,$62,$3D,$D7,$F0,$1F,$F3,$04,$F1,$8C,$61
	.dc.b	$F1,$76,$FF,$10,$36,$BD,$07,$28,$CB,$FE,$47,$BE,$CF,$43,$E4,$DF
	.dc.b	$59,$DE,$1E,$99,$F3,$1E,$A3,$64,$3D,$F3,$C5,$F0,$9D,$4C,$9B,$AF
	.dc.b	$0D,$F9,$EB,$E6,$61,$EC,$AF,$33,$63,$C5,$1E,$B3,$AC,$F1,$BF,$93
	.dc.b	$5B,$7C,$55,$D3,$0F,$5C,$60,$DB,$8B,$F2,$57,$16,$C6,$5F,$49,$30
	.dc.b	$F4,$BF,$EC,$73,$FC,$90,$7E,$33,$1F,$03,$E6,$3D,$73,$CE,$78,$E1
	.dc.b	$EC,$96,$3F,$E4,$7B,$FF,$4D,$F7,$7F,$93,$67,$FD,$73,$39,$7F,$C2
	.dc.b	$B1,$FE,$4C,$E7,$78,$3F,$E5,$19,$CD,$FF,$94,$9F,$E8,$51,$FD,$55
	.dc.b	$FE,$B5,$A7,$F4,$C2,$3F,$D6,$23,$FA,$94,$A7,$FB,$CB,$FD,$36,$79
	.dc.b	$F0,$7F,$4E,$CB,$FC,$04,$DA,$FE,$D9,$3D,$8E,$91,$97,$FB,$0C,$78
	.dc.b	$1D,$13,$0F,$EA,$F5,$D2,$3B,$91,$C1,$27,$BC,$F0,$6E,$1E,$F8,$C5
	.dc.b	$78,$0F,$19,$CE,$77,$7E,$58,$FF,$08,$5F,$F9,$A1,$B3,$FE,$1A,$8F
	.dc.b	$89,$1F,$E2,$8F,$09,$E1,$FF,$4C,$8F,$EE,$B3,$FD,$DC,$95,$EF,$C3
	.dc.b	$C7,$3D,$0F,$F1,$F4,$78,$9F,$21,$E7,$3E,$26,$3E,$B2,$63,$E7,$C7
	.dc.b	$C3,$3F,$EB,$96,$BE,$2A,$F0,$57,$08,$EA,$78,$4F,$F1,$DC,$78,$70
	.dc.b	$F9,$91,$E1,$C3,$CF,$7D,$27,$E2,$36,$BE,$21,$F0,$67,$FC,$9D,$87
	.dc.b	$97,$1F,$89,$7C,$1B,$5C,$30,$F9,$51,$86,$9F,$E2,$4A,$FF,$2C,$E3
	.dc.b	$FE,$77,$BF,$A3,$3F,$E5,$3C,$7C,$4F,$03,$E0,$1F,$E2,$93,$FD,$0F
	.dc.b	$97,$08,$FF,$21,$47,$FA,$67,$1F,$80,$47,$E0,$5F,$C7,$BF,$87,$1F
	.dc.b	$1E,$BB,$BE,$1C,$7D,$0D,$3D,$02,$7C,$FD,$2F,$A6,$FF,$02,$4A,$30
	.dc.b	$FF,$81,$DA,$FF,$8A,$1A,$FE,$34,$E1,$1F,$E2,$35,$F2,$6F,$AC,$C3
	.dc.b	$63,$FC,$3F,$AF,$F6,$DB,$EA,$DB,$DD,$CA,$DF,$17,$E6,$EF,$6F,$F8
	.dc.b	$F3,$C5,$F4,$F8,$BF,$E3,$DA,$FF,$1E,$B7,$1F,$EF,$5E,$33,$C7,$F3
	.dc.b	$36,$BF,$C9,$27,$F8,$DB,$C3,$FF,$27,$3E,$DF,$0B,$FC,$66,$48,$FF
	.dc.b	$19,$F8,$7F,$E3,$12,$79,$7F,$E2,$DE,$7F,$F1,$9E,$DF,$E3,$0C,$3F
	.dc.b	$C9,$F8,$7F,$8D,$5F,$F2,$0C,$3C,$4F,$F1,$57,$89,$FE,$0F,$9F,$DC
	.dc.b	$78,$5F,$E2,$E7,$FC,$AE,$DD,$BF,$C2,$8A,$7F,$C5,$7D,$AF,$33,$DD
	.dc.b	$C7,$F2,$30,$FF,$26,$DE,$E3,$FC,$59,$E0,$70,$F0,$FF,$C5,$79,$7F
	.dc.b	$87,$FC,$2F,$F1,$5D,$7F,$8E,$67,$2F,$F8,$9A,$BC,$D8,$7F,$89,$A3
	.dc.b	$FC,$99,$5F,$E2,$98,$FF,$17,$CE,$9B,$3E,$03,$C9,$F6,$6F,$B6,$7C
	.dc.b	$BF,$F8,$DF,$0F,$0D,$FF,$7D,$3F,$AF,$78,$BF,$D8,$DB,$87,$F5,$98
	.dc.b	$C6,$DF,$D6,$BC,$2F,$E4,$0D,$AF,$F2,$94,$FF,$7B,$F0,$FF,$6F,$28
	.dc.b	$5F,$F9,$CA,$7F,$C6,$33,$98,$C5,$FF,$76,$BF,$ED,$5E,$07,$F6,$7F
	.dc.b	$17,$FC,$01,$1E,$7B,$F0,$3E,$63,$75,$FD,$FD,$D6,$3E,$82,$61,$EB
	.dc.b	$24,$70,$3F,$C0,$25,$3A,$61,$FD,$F0,$95,$E5,$7F,$89,$C7,$FA,$33
	.dc.b	$0F,$F4,$6C,$7C,$0B,$F8,$55,$F1,$9F,$83,$5F,$E5,$18,$FF,$52,$D7
	.dc.b	$C0,$8F,$F0,$5B,$EE,$7F,$C6,$24,$3F,$C4,$55,$E2,$7F,$C0,$A7,$F9
	.dc.b	$79,$F9,$75,$DB,$1B,$87,$5A,$E4,$7F,$8D,$71,$F9,$24,$AF,$3C,$78
	.dc.b	$8F,$31,$6F,$E9,$3F,$E1,$C3,$FC,$D9,$87,$5A,$FB,$51,$C0,$FE,$C9
	.dc.b	$19,$AF,$26,$3F,$24,$F8,$33,$E9,$6E,$3F,$27,$1F,$44,$FC,$D3,$A6
	.dc.b	$1D,$A7,$E8,$C7,$F8,$A8,$FF,$1A,$1F,$E7,$B3,$83,$CD,$F2,$1C,$F1
	.dc.b	$C1,$CB,$1F,$B3,$87,$D7,$C3,$FC,$D3,$7C,$4F,$97,$7C,$8F,$05,$F2
	.dc.b	$AF,$5B,$63,$EA,$CF,$19,$F6,$9D,$63,$95,$FD,$F7,$EE,$C6,$9F,$A5
	.dc.b	$3C,$5F,$BF,$5D,$5F,$9A,$7C,$93,$BC,$F5,$9F,$39,$F8,$78,$FC,$0D
	.dc.b	$7D,$3B,$F9,$B1,$FE,$59,$AF,$A8,$F6,$7C,$D7,$C9,$F2,$47,$CD,$3F
	.dc.b	$15,$8F,$EB,$1F,$D4,$63,$F8,$BB,$FD,$6D,$F8,$67,$DA,$7E,$D2,$3F
	.dc.b	$D2,$BE,$2F,$F9,$7E,$3D,$6D,$E8,$FF,$2E,$F1,$7F,$98,$65,$18,$53
	.dc.b	$FE,$63,$F1,$7F,$CC,$51,$A6,$CF,$F8,$6B,$C4,$FE,$EF,$E2,$FF,$7A
	.dc.b	$8F,$F0,$BD,$7F,$89,$23,$FB,$E9,$9A,$FF,$08,$6B,$FE,$1D,$6C,$FF
	.dc.b	$83,$F3,$FF,$09,$DF,$27,$FC,$0F,$7C,$B0,$FF,$02,$CF,$6D,$FF,$C2
	.dc.b	$51,$FE,$47,$6D,$FE,$D6,$7D,$91,$FE,$48,$8F,$F2,$BB,$A7,$F5,$95
	.dc.b	$FE,$1A,$8D,$47,$F9,$1B,$3F,$F2,$15,$FB,$8F,$0E,$28,$FF,$0A,$D7
	.dc.b	$32,$65,$ED,$8F,$14,$FF,$91,$E7,$37,$FE,$2A,$AD,$3F,$E2,$93,$FC
	.dc.b	$4F,$5F,$E3,$F7,$C2,$7F,$88,$B2,$F8,$B7,$FE,$8C,$AF,$F4,$59,$CA
	.dc.b	$FD,$39,$7C,$08,$CD,$F7,$3F,$C7,$47,$B4,$F8,$38,$FC,$98,$DC,$7D
	.dc.b	$FC,$7D,$27,$89,$FF,$19,$47,$F9,$3D,$D1,$83,$AE,$5F,$E6,$8C,$3D
	.dc.b	$F1,$A7,$CB,$8F,$C3,$AE,$F5,$F4,$5F,$F3,$6D,$FC,$1B,$E8,$7F,$9B
	.dc.b	$71,$FC,$04,$F0,$C7,$D1,$3E,$C9,$FC,$A3,$E7,$CB,$E3,$DF,$F9,$80
	.dc.b	$91,$D2,$7E,$33,$F1,$46,$1E,$37,$FD,$3B,$5F,$92,$BF,$C1,$47,$D3
	.dc.b	$C3,$CE,$FE,$7E,$32,$FF,$9A,$0F,$9D,$3F,$EB,$1C,$BE,$1D,$F5,$9C
	.dc.b	$DF,$F9,$9D,$FF,$32,$5F,$E3,$B1,$F3,$E1,$F2,$1F,$88,$FF,$A0,$67
	.dc.b	$DD,$5F,$8A,$8D,$47,$8C,$C9,$FA,$D8,$D1,$E8,$8F,$64,$68,$FF,$38
	.dc.b	$1E,$9C,$7E,$5B,$F4,$B0,$FD,$D5,$F9,$2F,$91,$E9,$8F,$DB,$4F,$C5
	.dc.b	$3F,$5F,$1F,$5D,$FC,$59,$F9,$F3,$DE,$6C,$F8,$67,$E8,$8F,$51,$E1
	.dc.b	$C3,$E2,$3F,$84,$EA,$FF,$35,$A3,$FC,$F6,$7F,$92,$4A,$7E,$DF,$5F
	.dc.b	$F6,$67,$8A,$F5,$7F,$D0,$15,$FE,$84,$AF,$F3,$B7,$8D,$FE,$8C,$8C
	.dc.b	$47,$56,$BD,$33,$FF,$08,$BF,$E8,$8F,$1B,$FD,$00,$F7,$38,$57,$FA
	.dc.b	$37,$C5,$78,$D4,$93,$E7,$F1,$7F,$CF,$38,$74,$AF,$F4,$24,$FF,$A5
	.dc.b	$DA,$FF,$B9,$7C,$2F,$F3,$E7,$2F,$FA,$6E,$3E,$14,$61,$C7,$13,$C4
	.dc.b	$FF,$3F,$C6,$1A,$FF,$A8,$E3,$5E,$27,$CB,$F9,$3F,$3C,$FC,$D9,$CA
	.dc.b	$8C,$BF,$D2,$A9,$5C,$74,$FF,$6D,$46,$27,$84,$7F,$85,$9F,$41,$FE
	.dc.b	$0F,$2F,$5F,$FA,$F1,$C1,$B2,$1F,$E3,$63,$67,$FB,$09,$E0,$47,$04
	.dc.b	$36,$43,$99,$FE,$00,$30,$75,$32,$7F,$89,$CC,$9B,$32,$79,$8E,$04
	.dc.b	$34,$53,$91,$C1,$F3,$3D,$0C,$9B,$3A,$1C,$88,$68,$A6,$8C,$10,$C1
	.dc.b	$1E,$EF,$84,$F8,$85,$3C,$46,$CC,$9E,$37,$8B,$4C,$9E,$27,$2D,$32
	.dc.b	$78,$DC,$B4,$C9,$DC,$F0,$BA,$70,$79,$48,$78,$4D,$9C,$9E,$04,$70
	.dc.b	$43,$64,$39,$9E,$73,$07,$53,$27,$9C,$C9,$B3,$27,$98,$E0,$43,$45
	.dc.b	$39,$1C,$1F,$33,$D0,$C9,$B3,$A1,$C8,$86,$8A,$68,$C1,$0C,$11,$EE
	.dc.b	$F8,$4F,$88,$53,$C4,$6C,$C9,$E3,$78,$B4,$C9,$E2,$72,$D3,$27,$8D
	.dc.b	$CB,$4C,$9D,$CF,$0B,$A7,$07,$94,$87,$84,$D9,$C9,$E0,$47,$04,$36
	.dc.b	$43,$99,$E7,$30,$75,$32,$79,$CC,$9B,$32,$79,$8E,$04,$34,$53,$91
	.dc.b	$C1,$F3,$3D,$0C,$9B,$3A,$1C,$88,$68,$A6,$8C,$1F,$C1,$70,$FF,$3E
	.dc.b	$3F,$CF,$7A,$7F,$CB,$0F,$F4,$46,$B9,$D7,$F9,$9C,$F0,$8F,$F2,$A9
	.dc.b	$93,$84,$6A,$78,$C6,$88,$7F,$47,$72,$D8,$E0,$47,$99,$0E,$C4,$39
	.dc.b	$9F,$E4,$E3,$07,$33,$81,$A9,$F4,$39,$36,$71,$35,$3E,$47,$06,$CC
	.dc.b	$CF,$95,$E3,$3D,$1D,$91,$C1,$27,$A3,$67,$94,$F4,$3C,$E7,$72,$1C
	.dc.b	$E7,$89,$23,$8C,$6A,$7A,$3C,$8E,$11,$A6,$99,$8D,$BE,$A9,$E8,$E5
	.dc.b	$AE,$67,$94,$F0,$8F,$41,$93,$84,$6A,$78,$C6,$88,$79,$DC,$B6,$38
	.dc.b	$11,$E6,$43,$B1,$0E,$67,$B0,$C1,$CC,$E0,$6A,$7D,$0E,$4D,$9C,$4D
	.dc.b	$4F,$91,$C1,$B3,$33,$E5,$78,$CF,$47,$64,$70,$49,$E8,$D9,$E5,$3D
	.dc.b	$0F,$39,$DC,$87,$39,$E2,$48,$E3,$1A,$9E,$8F,$23,$84,$69,$A6,$63
	.dc.b	$6F,$AA,$7A,$39,$6B,$99,$E5,$3C,$23,$D0,$64,$E1,$1A,$9E,$31,$A2
	.dc.b	$1E,$77,$2D,$8E,$04,$79,$90,$EC,$43,$99,$EC,$30,$73,$38,$1A,$9F
	.dc.b	$43,$93,$67,$13,$53,$E4,$70,$6C,$CC,$F9,$5E,$33,$D1,$D9,$1C,$12
	.dc.b	$7A,$36,$79,$4F,$F9,$8B,$1F,$F3,$23,$FE,$6B,$C7,$FC,$C9,$97,$F9
	.dc.b	$AA,$FF,$CD,$AD,$8C,$E5,$C6,$BF,$CF,$17,$FC,$DA,$7F,$CF,$31,$93
	.dc.b	$FC,$F4,$E1,$ED,$1B,$23,$C8,$87,$07,$2D,$32,$75,$21,$D0,$C9,$4E
	.dc.b	$04,$36,$7F,$A1,$5E,$AE,$2F,$FD,$16,$EC,$E4,$66,$7D,$4F,$13,$B9
	.dc.b	$4E,$05,$38,$4F,$67,$99,$0D,$16,$7C,$2F,$03,$AB,$87,$53,$E2,$8C
	.dc.b	$47,$22,$38,$24,$6C,$93,$E3,$9D,$C6,$1E,$ED,$7D,$73,$CA,$78,$36
	.dc.b	$37,$18,$9E,$11,$E9,$38,$B8,$7D,$31,$B2,$3C,$88,$70,$72,$D3,$27
	.dc.b	$52,$1D,$0C,$94,$E0,$43,$67,$C0,$7A,$B8,$9F,$63,$B3,$91,$99,$F5
	.dc.b	$3C,$4E,$E5,$38,$14,$E1,$3D,$9E,$64,$34,$59,$F0,$BC,$0E,$AE,$1D
	.dc.b	$4F,$8A,$31,$1C,$88,$E0,$91,$B2,$4F,$8E,$77,$18,$7B,$B5,$F5,$CF
	.dc.b	$29,$E0,$D8,$DC,$62,$78,$47,$A4,$E2,$E1,$F4,$C6,$C8,$F2,$21,$C1
	.dc.b	$CB,$4C,$9D,$48,$74,$32,$53,$81,$0D,$9F,$01,$EA,$E2,$7D,$8E,$CE
	.dc.b	$46,$67,$D4,$F1,$3B,$94,$E0,$53,$84,$F6,$79,$90,$D1,$67,$C2,$F0
	.dc.b	$3A,$B8,$75,$3E,$28,$C4,$72,$23,$82,$46,$C9,$3E,$39,$DC,$61,$EE
	.dc.b	$D7,$D7,$3C,$A7,$83,$63,$71,$89,$E1,$1E,$93,$8B,$87,$D3,$1B,$23
	.dc.b	$C8,$87,$07,$2D,$32,$75,$21,$D0,$C9,$4E,$04,$36,$7C,$07,$AB,$89
	.dc.b	$F6,$3B,$39,$19,$9F,$53,$C4,$EE,$53,$81,$4E,$13,$D9,$E6,$43,$45
	.dc.b	$9F,$0B,$C0,$EA,$E1,$D4,$F8,$A3,$11,$C8,$8E,$09,$1B,$24,$F8,$E7
	.dc.b	$71,$87,$BB,$5F,$5C,$F2,$9E,$0D,$8F,$D9,$FE,$52,$20

;--------------------------------------------------------------------------------
;	FSGLDIV.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68060
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsgldiv060_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#48,-(sp)
	peamsg	'FSGLDIV.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSGLDIV.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsgldiv060_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsgldiv.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSGLDIV.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsgldiv.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSGLDIV.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsgldiv060_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3D,$61,$F3,$7F,$DA,$BD,$E7,$99,$C6,$30,$DA,$FE
	.dc.b	$50,$D9,$DC,$7E,$72,$3F,$B8,$BF,$98,$29,$F9,$E6,$B4,$CF,$9D,$C9
	.dc.b	$E2,$D7,$FB,$DB,$5E,$86,$A3,$9C,$66,$3C,$47,$48,$C9,$1C,$10,$E4
	.dc.b	$53,$BB,$B7,$24,$32,$6C,$F1,$98,$34,$60,$87,$90,$C9,$4E,$04,$36
	.dc.b	$7C,$53,$65,$34,$71,$37,$1F,$05,$D1,$88,$F5,$3C,$0C,$9C,$48,$60
	.dc.b	$91,$CC,$91,$CA,$3A,$1F,$04,$E3,$1A,$70,$72,$8E,$0E,$9A,$E4,$DC
	.dc.b	$71,$7D,$ED,$7D,$0D,$7C,$26,$A3,$9C,$66,$3B,$1D,$23,$24,$70,$43
	.dc.b	$91,$4E,$EE,$DC,$90,$C9,$B3,$C6,$60,$D1,$82,$1E,$43,$25,$38,$10
	.dc.b	$D9,$F1,$4D,$94,$D1,$C4,$DC,$7C,$17,$46,$23,$D4,$F0,$32,$71,$21
	.dc.b	$82,$47,$32,$47,$28,$E8,$7C,$13,$8C,$69,$C1,$CA,$38,$3A,$6B,$93
	.dc.b	$71,$C5,$F7,$B5,$F4,$35,$F0,$9A,$8E,$71,$98,$EC,$74,$8C,$91,$C1
	.dc.b	$0E,$45,$3B,$BB,$72,$43,$26,$CF,$19,$83,$46,$08,$79,$0C,$94,$E0
	.dc.b	$43,$67,$C5,$36,$53,$47,$13,$71,$F0,$5D,$18,$8F,$53,$F8,$1C,$BF
	.dc.b	$D1,$17,$FC,$F2,$3F,$A1,$35,$FE,$97,$87,$F4,$47,$F9,$F9,$A2,$9A
	.dc.b	$8E,$13,$D2,$32,$DC,$3F,$9E,$36,$FC,$93,$F2,$4F,$51,$0F,$49,$70
	.dc.b	$F6,$12,$3D,$71,$C8,$F8,$B1,$97,$A1,$C8,$8E,$5D,$C7,$27,$B3,$5D
	.dc.b	$B6,$38,$47,$D0,$32,$76,$72,$F8,$CD,$38,$7A,$9B,$8E,$84,$70,$43
	.dc.b	$CA,$53,$9B,$B7,$24,$32,$6C,$F1,$10,$F0,$9B,$30,$7A,$8C,$1A,$39
	.dc.b	$98,$34,$60,$87,$28,$F7,$3A,$31,$1E,$77,$67,$53,$27,$9C,$C1,$0E
	.dc.b	$45,$8E,$A4,$8E,$91,$C8,$F3,$46,$5E,$87,$22,$39,$77,$1C,$9E,$CD
	.dc.b	$76,$D9,$FC,$C4,$7E,$86,$BF,$37,$87,$E9,$5B,$87,$E8,$A3,$F5,$12
	.dc.b	$3F,$C5,$05,$3F,$54,$53,$F5,$47,$F8,$A2,$BB,$B8,$AE,$AF,$EB,$27
	.dc.b	$C1,$7F,$E4,$2C,$9F,$37,$FE,$2E,$C3,$FC,$CA,$D8,$CD,$F5,$6B,$E1
	.dc.b	$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7C,$51,$B7,$2D,$72,$71,$29,$A3
	.dc.b	$07,$F9,$61,$E0,$D3,$27,$F9,$7A,$3C,$0E,$4E,$44,$3A,$90,$D1,$4E
	.dc.b	$51,$D9,$A7,$69,$F8,$6F,$33,$89,$0D,$99,$29,$D5,$E2,$D9,$F3,$BA
	.dc.b	$39,$46,$9B,$1A,$8C,$46,$C8,$E0,$91,$B2,$4F,$AE,$32,$D8,$CC,$FB
	.dc.b	$5A,$FA,$67,$6E,$67,$81,$1C,$10,$E1,$3C,$9F,$74,$6D,$CB,$5C,$9C
	.dc.b	$4A,$68,$C1,$F0,$9E,$0D,$32,$7C,$28,$F0,$39,$39,$10,$EA,$43,$45
	.dc.b	$39,$47,$66,$9D,$A7,$E1,$BC,$CE,$24,$36,$64,$A7,$57,$8B,$67,$CE
	.dc.b	$E8,$E5,$1A,$6C,$6A,$31,$1B,$23,$82,$46,$C9,$3E,$B8,$CB,$63,$33
	.dc.b	$ED,$6B,$E9,$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7D,$D1,$B7,$2D,$72
	.dc.b	$71,$29,$A3,$07,$C2,$78,$34,$C9,$F0,$A3,$C0,$E4,$E4,$43,$A9,$0D
	.dc.b	$14,$E5,$1D,$9A,$76,$9F,$86,$F3,$38,$90,$D9,$92,$9D,$5E,$2D,$9F
	.dc.b	$3B,$A3,$94,$7E,$F9,$FE,$8D,$F2,$DF,$EE,$44,$D2,$3F,$EE,$86,$BF
	.dc.b	$EE,$C6,$BF,$EF,$06,$BF,$D3,$5A,$FF,$51,$6B,$E0,$3A,$10,$E8,$53
	.dc.b	$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10
	.dc.b	$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$A6,$7B,$4A,$79
	.dc.b	$8A,$78,$4A,$76,$29,$D4,$A7,$57,$E9,$B8,$23,$82,$3E,$16,$BE,$16
	.dc.b	$BE,$16,$BE,$16,$BE,$16,$BE,$13,$A1,$0E,$85,$3A,$14,$E8,$53,$A1
	.dc.b	$4E,$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$10,$FA,$67,$B4,$A7,$98,$A7,$84,$A7,$62
	.dc.b	$9D,$4A,$75,$7E,$9B,$82,$38,$23,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$6B
	.dc.b	$E1,$6B,$E1,$3A,$10,$E8,$53,$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8
	.dc.b	$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$FE,$09,$1F,$E7,$22,$61,$FE,$70,$8F,$F4,$16,$1F,$E7,$C6,$1F
	.dc.b	$E7,$C2,$8F,$83,$7C,$0B,$E3,$FF,$D0,$13,$FE,$85,$9C,$C6,$23,$84
	.dc.b	$ED,$DC,$E9,$FF,$4C,$9F,$E8,$DC,$DF,$09,$FE,$8D,$DD,$8F,$FA,$3E
	.dc.b	$73,$5F,$E9,$39,$CB,$C6,$BF,$D2,$9F,$88,$F7,$8F,$F4,$5C,$7F,$A5
	.dc.b	$FD,$27,$9F,$CE,$3F,$D3,$5F,$06,$F4,$31,$7F,$E9,$F8,$C3,$D4,$FF
	.dc.b	$51,$5F,$27,$01,$C3,$F1,$1A,$FF,$A9,$5A,$F8,$EF,$E3,$3C,$5F,$0B
	.dc.b	$4E,$B1,$D3,$2F,$83,$0A,$78,$7E,$8D,$DF,$BD,$7F,$AC,$EF,$E3,$C6
	.dc.b	$2F,$53,$9C,$FF,$D7,$9F,$2C,$FD,$45,$F0,$BA,$B8,$7C,$1C,$39,$3F
	.dc.b	$0E,$B8,$4F,$C8,$79,$78,$1E,$3F,$D8,$FC,$F9,$F8,$9B,$97,$83,$25
	.dc.b	$71,$6E,$3D,$EB,$6F,$E3,$DB,$87,$68,$FA,$3E,$B3,$D5,$E5,$E1,$C9
	.dc.b	$CC,$7C,$C9,$D4,$7E,$55,$AF,$E6,$9A,$FF,$B6,$DA,$FF,$B8,$7E,$7D
	.dc.b	$FA,$13,$CF,$F4,$7B,$72,$BF,$0D,$5A,$9F,$C5,$B6,$3E,$46,$8A,$E1
	.dc.b	$1F,$96,$13,$B7,$0D,$DF,$F4,$2D,$DF,$C9,$8F,$8E,$BE,$A7,$EA,$7F
	.dc.b	$A1,$BD,$F9,$4F,$3F,$EB,$4B,$E9,$29,$E6,$9D,$5F,$A9,$AF,$97,$85
	.dc.b	$5C,$63,$BD,$71,$B9,$1C,$5A,$F1,$97,$BB,$E2,$F7,$7C,$7E,$9C,$AF
	.dc.b	$4C,$65,$F5,$C6,$5E,$51,$96,$9C,$0F,$83,$E5,$FD,$0F,$D2,$9F,$11
	.dc.b	$F0,$3E,$2A,$57,$D6,$9C,$B6,$72,$F9,$E7,$94,$48,$D1,$0F,$67,$A1
	.dc.b	$F3,$3C,$CF,$9B,$E5,$49,$FA,$71,$97,$EA,$C6,$5F,$89,$19,$69,$C0
	.dc.b	$FB,$7F,$31,$F6,$3E,$72,$47,$DE,$7D,$DF,$0D,$23,$04,$3F,$13,$9A
	.dc.b	$3F,$03,$F0,$3B,$9C,$C6,$1F,$C4,$F6,$FE,$18,$8F,$E3,$3F,$47,$FC
	.dc.b	$4F,$EA,$1F,$8A,$64,$7E,$42,$3F,$1B,$E3,$C8,$FC,$9B,$F9,$0F,$2B
	.dc.b	$F8,$EF,$A2,$E3,$FC,$9F,$99,$2F,$F9,$5C,$E5,$B3,$97,$F1,$53,$CA
	.dc.b	$24,$7F,$4F,$9E,$0F,$F5,$46,$A2,$07,$2F,$F2,$2E,$8F,$CF,$F6,$FF
	.dc.b	$4E,$EC,$FC,$FF,$C0,$7E,$A0,$FD,$4F,$64,$BF,$D7,$C6,$5F,$D8,$C7
	.dc.b	$F2,$3F,$D5,$3F,$97,$42,$90,$FA,$1D,$1B,$98,$7F,$2D,$8D,$36,$32
	.dc.b	$FE,$CE,$1F,$3B,$DF,$E7,$64,$F1,$8F,$11,$87,$84,$61,$D1,$DD,$D6
	.dc.b	$5F,$06,$BA,$CF,$F1,$CF,$D2,$5F,$39,$ED,$F9,$E9,$39,$8F,$E3,$F3
	.dc.b	$A8,$FF,$02,$F9,$08,$C4,$7D,$23,$F8,$FB,$87,$EA,$C6,$E7,$FC,$15
	.dc.b	$A7,$C8,$F9,$EF,$99,$EF,$CB,$F8,$D1,$F3,$CC,$3F,$16,$30,$FE,$33
	.dc.b	$BD,$A7,$E0,$FE,$AD,$5F,$D8,$B5,$7E,$15,$F6,$4F,$CB,$DF,$0C,$17
	.dc.b	$C1,$FC,$67,$2A,$75,$39,$CF,$EF,$DF,$3B,$F1,$65,$E1,$CB,$F2,$78
	.dc.b	$F1,$8F,$CB,$79,$3D,$6B,$F2,$DE,$27,$E6,$23,$14,$8C,$5F,$2C,$FC
	.dc.b	$D5,$A7,$C7,$1D,$A7,$F4,$39,$F2,$6C,$71,$C7,$C9,$8F,$B2,$F8,$BE
	.dc.b	$EA,$DB,$4D,$9F,$AB,$8D,$B6,$73,$8F,$9E,$2D,$72,$74,$7C,$CB,$F1
	.dc.b	$3F,$D5,$5F,$43,$F1,$1A,$E4,$F3,$56,$5A,$73,$3C,$C5,$3E,$59,$E4
	.dc.b	$BF,$3D,$FD,$18,$D3,$E7,$6D,$F9,$E3,$58,$7C,$DB,$F0,$57,$43,$C8
	.dc.b	$5C,$74,$E2,$B8,$E3,$EA,$BF,$A5,$3E,$3B,$8E,$F0,$FB,$AF,$AB,$0E
	.dc.b	$67,$4A,$E3,$1E,$63,$8B,$AB,$F8,$F3,$A3,$13,$F1,$A7,$89,$27,$E8
	.dc.b	$BA,$8F,$54,$72,$3E,$AC,$F8,$5F,$51,$23,$A3,$C8,$C3,$F1,$67,$83
	.dc.b	$A3,$EA,$BE,$17,$07,$E1,$27,$D1,$3E,$18,$FB,$B1,$C9,$E9,$1D,$B0
	.dc.b	$F9,$B1,$A8,$E0,$FA,$8C,$1D,$A3,$EF,$57,$86,$3F,$26,$F5,$8C,$CF
	.dc.b	$5C,$7F,$0E,$7E,$23,$0F,$C2,$1F,$85,$73,$8F,$E1,$2B,$F1,$58,$FA
	.dc.b	$B1,$F0,$9D,$B0,$F4,$61,$C6,$FE,$CC,$62,$69,$C2,$F9,$C7,$B1,$FC
	.dc.b	$C5,$F9,$EF,$C7,$5C,$0E,$C7,$CE,$AF,$55,$E6,$FD,$B3,$93,$E2,$61
	.dc.b	$ED,$7F,$6E,$6D,$C3,$F8,$F3,$F4,$EF,$C0,$71,$8F,$CB,$9F,$A7,$1E
	.dc.b	$23,$EB,$9D,$CC,$3A,$3C,$6E,$8C,$CF,$82,$7A,$3D,$FD,$5F,$F2,$94
	.dc.b	$FF,$0A,$6D,$7E,$ED,$B3,$FC,$E5,$DC,$FF,$80,$FD,$EF,$F0,$24,$62
	.dc.b	$11,$8C,$FF,$DB,$AD,$8F,$EB,$0D,$7F,$C1,$6F,$F2,$27,$FB,$E7,$9D
	.dc.b	$5E,$67,$F1,$13,$25,$33,$8F,$F3,$F7,$2D,$73,$9F,$F6,$7E,$7F,$F7
	.dc.b	$0C,$E6,$FC,$11,$89,$F3,$D7,$78,$C3,$E8,$31,$7E,$5A,$E2,$D8,$C9
	.dc.b	$E5,$8C,$B4,$CD,$F8,$5B,$7E,$08,$F5,$5F,$82,$7E,$29,$A3,$C5,$1A
	.dc.b	$BF,$03,$63,$B3,$4E,$C5,$AF,$6E,$1B,$3C,$45,$7C,$47,$C7,$AE,$AD
	.dc.b	$AE,$AD,$AE,$AD,$3A,$9B,$AE,$AD,$3A,$CF,$F9,$2A,$FD,$CF,$E2,$CF
	.dc.b	$71,$F0,$2B,$D0,$FC,$09,$EC,$D3,$95,$F9,$9A,$7C,$FB,$F6,$3F,$65
	.dc.b	$AF,$59,$E6,$D3,$8D,$7A,$1E,$73,$5F,$F3,$15,$AF,$FA,$CC,$7A,$6B
	.dc.b	$CE,$F3,$37,$1E,$56,$9F,$28,$EF,$7F,$11,$CC,$FC,$C7,$27,$A6,$3E
	.dc.b	$53,$4F,$2B,$C5,$FD,$0B,$6B,$E6,$C7,$EA,$AF,$E0,$BF,$3D,$AF,$E9
	.dc.b	$8E,$E7,$A4,$E6,$77,$3D,$2F,$C3,$7F,$66,$F9,$DF,$4C,$F8,$6F,$F4
	.dc.b	$93,$DB,$0F,$8F,$7F,$26,$FF,$BF,$9F,$26,$FE,$9B,$4F,$19,$4E,$A7
	.dc.b	$E0,$4A,$7E,$9A,$BB,$3D,$27,$B4,$79,$63,$D6,$7B,$63,$E2,$D7,$ED
	.dc.b	$B0,$D4,$62,$7F,$26,$78,$4E,$27,$C9,$BF,$A6,$F8,$5A,$FA,$77,$FF
	.dc.b	$4B,$F9,$1E,$FD,$BE,$04,$FE,$3E,$73,$9F,$FA,$B6,$73,$3C,$78,$7F
	.dc.b	$DE,$33,$93,$F5,$35,$C2,$7F,$D3,$5F,$2D,$D6,$7F,$CE,$5F,$0F,$67
	.dc.b	$FD,$51,$6A,$C5,$FF,$AA,$AA,$CE,$6F,$FD,$51,$18,$AC,$BC,$F9,$BF
	.dc.b	$D2,$5F,$3D,$E6,$C7,$FD,$22,$D8,$CD,$FF,$A6,$3A,$13,$96,$C7,$FA
	.dc.b	$D7,$A3,$A4,$62,$FF,$D6,$91,$E4,$FA,$0F,$2F,$CC,$5B,$FF,$5E,$36
	.dc.b	$BF,$D7,$4D,$DB,$FD,$60,$D9,$FF,$4A,$36,$7E,$1B,$F0,$3E,$63,$E0
	.dc.b	$FC,$8D,$AF,$F6,$43,$6B,$FD,$8C,$DB,$FF,$5E,$3F,$23,$E8,$3E,$4F
	.dc.b	$CC,$DB,$FF,$68,$36,$FF,$D9,$CD,$BF,$F6,$23,$F5,$63,$FD,$D9,$E0
	.dc.b	$7C,$B6,$D7,$FA,$F5,$FF,$77,$C6,$53,$4F,$B3,$E6,$73,$8C,$CF,$FB
	.dc.b	$86,$31,$3A,$8F,$C5,$4F,$47,$47,$FB,$9F,$E6,$7F,$06,$7E,$13,$82
	.dc.b	$C7,$FB,$AF,$E8,$3F,$13,$9F,$F2,$18,$C1,$D3,$99,$19,$C3,$FD,$8F
	.dc.b	$27,$FD,$F5,$0F,$A0,$E6,$74,$9A,$73,$87,$C5,$FE,$47,$7F,$F0,$0E
	.dc.b	$FF,$C2,$71,$EF,$39,$EB,$98,$71,$F8,$6B,$9F,$88,$F0,$BE,$A2,$47
	.dc.b	$C4,$E7,$65,$DA,$3D,$18,$FB,$2B,$E0,$FC,$5C,$9F,$37,$C9,$DC,$7B
	.dc.b	$5F,$C2,$AF,$8F,$F0,$59,$D7,$D1,$B2,$31,$8F,$68,$C3,$F0,$6B,$8B
	.dc.b	$E5,$CD,$D3,$F0,$CF,$A7,$F4,$16,$75,$1F,$5B,$E2,$71,$A7,$93,$14
	.dc.b	$63,$3F,$1B,$6F,$C4,$FD,$EF,$9C,$91,$F4,$FE,$36,$E1,$CE,$7F,$07
	.dc.b	$F3,$98,$C3,$DA,$DB,$F6,$36,$BF,$02,$FD,$BE,$69,$3F,$7F,$E5,$2D
	.dc.b	$7D,$7E,$4F,$A7,$5F,$90,$F8,$8C,$E3,$F2,$9B,$87,$C9,$6D,$FE,$39
	.dc.b	$FE,$31,$FA,$53,$D7,$A7,$E6,$BE,$1A,$DF,$E0,$67,$4F,$E4,$7E,$57
	.dc.b	$38,$7C,$96,$E3,$F6,$9B,$3F,$65,$FD,$E7,$BD,$F9,$88,$FD,$D5,$23
	.dc.b	$0F,$EF,$5A,$FE,$94,$FD,$33,$87,$0F,$0A,$FC,$4F,$D0,$F6,$ED,$FD
	.dc.b	$9B,$FB,$4F,$8F,$FE,$81,$E8,$BF,$64,$7C,$4E,$73,$FE,$03,$39,$8C
	.dc.b	$67,$FA,$58,$C4,$FE,$C6,$F9,$DF,$28,$E3,$FB,$DB,$AF,$BB,$63,$F7
	.dc.b	$1F,$23,$89,$91,$C1,$FE,$13,$18,$C7,$F6,$83,$E9,$DF,$0D,$FC,$95
	.dc.b	$B7,$FC,$63,$EA,$9F,$85,$F7,$BE,$35,$A6,$DF,$F2,$B8,$C4,$7C,$D3
	.dc.b	$B4,$64,$97,$FC,$07,$F0,$6F,$74,$F2,$8C,$6F,$FC,$DF,$EA,$9E,$65
	.dc.b	$3F,$B2,$7D,$13,$E4,$B1,$7F,$CB,$B0,$5F,$08,$FC,$1B,$63,$F9,$F3
	.dc.b	$5F,$33,$5F,$C2,$7C,$DB,$CC,$79,$BF,$8C,$8C,$36,$33,$3F,$D2,$67
	.dc.b	$2F,$F4,$BF,$80,$7C,$8B,$E7,$3F,$2B,$18,$6C,$66,$78,$35,$F9,$54
	.dc.b	$F3,$5D,$5E,$36,$CF,$85,$AF,$06,$BF,$9F,$F9,$C7,$95,$FA,$1E,$FF
	.dc.b	$F1,$2F,$CE,$7C,$BA,$F9,$4D,$7F,$0F,$5C,$1F,$9C,$D7,$FB,$43,$5F
	.dc.b	$C0,$35,$F0,$57,$ED,$1F,$99,$F1,$DE,$1E,$1F,$D9,$79,$BF,$C1,$E7
	.dc.b	$ED,$35,$FC,$BF,$02,$B8,$47,$F5,$D8,$C3,$C3,$C7,$CB,$FC,$C1,$B1
	.dc.b	$A8,$E7,$3E,$38,$C4,$7E,$5D,$AF,$F0,$56,$BF,$43,$E7,$3F,$39,$A7
	.dc.b	$F9,$AF,$64,$63,$2F,$F7,$AB,$69,$E9,$BE,$69,$E2,$7A,$B0,$E3,$56
	.dc.b	$F8,$46,$23,$FB,$F5,$70,$6F,$83,$C7,$E7,$3F,$CB,$B9,$F8,$DB,$9F
	.dc.b	$F6,$AB,$57,$C8,$CF,$C4,$F1,$99,$87,$43,$F5,$5E,$47,$F6,$77,$C3
	.dc.b	$F3,$9C,$70,$3F,$3A,$75,$3F,$C3,$1B,$97,$F8,$1E,$B7,$3C,$8E,$BE
	.dc.b	$B7,$DA,$F4,$1F,$36,$F4,$3B,$7C,$E4,$BF,$F0,$34,$E5,$B3,$97,$FC
	.dc.b	$03,$82,$BA,$56,$E7,$FA,$E3,$5F,$E8,$17,$E0,$BF,$E8,$ED,$8F,$F2
	.dc.b	$13,$5F,$E3,$4B,$D1,$7C,$8F,$CB,$F9,$37,$C6,$53,$D9,$3A,$BF,$F0
	.dc.b	$F3,$5F,$F0,$BC,$F0,$7F,$C7,$4D,$7F,$B9,$46,$F5,$FF,$5C,$61,$E1
	.dc.b	$C3,$A9,$2B,$8F,$8B,$E3,$F9,$CF,$27,$93,$27,$AC,$65,$EF,$19,$7C
	.dc.b	$71,$96,$9C,$0F,$2F,$CA,$FD,$7F,$99,$7C,$67,$C1,$F9,$C9,$5F,$E3
	.dc.b	$B9,$CB,$67,$2F,$C0,$9E,$51,$23,$44,$3E,$2F,$C9,$7C,$DF,$92,$F9
	.dc.b	$DE,$4C,$9F,$8F,$19,$7E,$4C,$65,$F9,$B1,$96,$9C,$0F,$9F,$F3,$9F
	.dc.b	$67,$E7,$24,$7D,$27,$DD,$F1,$32,$30,$43,$FC,$4D,$9A,$3F,$05,$EA
	.dc.b	$6E,$73,$18,$7E,$EF,$67,$E1,$C8,$FE,$03,$E7,$3F,$15,$F3,$9F,$8B
	.dc.b	$64,$7E,$12,$3F,$1D,$F3,$92,$3F,$0E,$FE,$47,$C9,$FE,$3F,$E9,$38
	.dc.b	$FF,$29,$E5,$CB,$FE,$C9,$39,$6C,$E5,$FC,$9C,$F2,$89,$1F,$94,$8D
	.dc.b	$75,$7F,$BA,$AB,$94,$FE,$6B,$A7,$F4,$5F,$39,$23,$F3,$71,$FA,$53
	.dc.b	$0F,$E8,$23,$0F,$FA,$67,$E5,$7F,$93,$79,$07,$FB,$5C,$7F,$9F,$1B
	.dc.b	$87,$FA,$45,$D9,$3D,$5F,$E3,$8E,$BC,$8F,$F4,$4B,$AF,$9C,$CF,$C7
	.dc.b	$CC,$3C,$11,$C0,$C3,$DE,$30,$E8,$F3,$3A,$C3,$C5,$3F,$C6,$7E,$53
	.dc.b	$FD,$0B,$5E,$BF,$9C,$F1,$99,$F8,$D9,$19,$21,$FE,$6F,$9E,$2F,$BF
	.dc.b	$E7,$3D,$7F,$33,$23,$E3,$47,$B8,$C3,$F1,$63,$0F,$DE,$EE,$65,$FE
	.dc.b	$C1,$F9,$A5,$7C,$4F,$9C,$F8,$E7,$E2,$2B,$76,$AD,$BF,$EC,$9E,$44
	.dc.b	$EA,$73,$C5,$F5,$EF,$9D,$F8,$71,$F0,$63,$FE,$6D,$C7,$8C,$62,$7F
	.dc.b	$B1,$F0,$2F,$8C,$7F,$95,$2B,$6F,$8E,$F9,$65,$E1,$7A,$78,$5F,$E7
	.dc.b	$9C,$FC,$F4,$B4,$E6,$B8,$E5,$E1,$AD,$61,$CD,$F7,$56,$DA,$6E,$FC
	.dc.b	$78,$76,$6E,$1E,$29,$D5,$71,$7A,$E1,$C7,$0E,$D1,$E5,$9D,$5F,$9A
	.dc.b	$35,$51,$F8,$07,$88,$A7,$13,$58,$78,$EF,$AD,$73,$9D,$B6,$75,$5C
	.dc.b	$30,$F4,$56,$B0,$F1,$E1,$C8,$E6,$F9,$8C,$D7,$19,$F2,$5F,$C3,$9C
	.dc.b	$C7,$82,$B8,$BF,$10,$E0,$7C,$43,$C0,$64,$E1,$5A,$C3,$D1,$87,$9B
	.dc.b	$0F,$4D,$68,$E2,$47,$24,$8E,$04,$8F,$CD,$5F,$B6,$FC,$15,$E9,$9F
	.dc.b	$04,$FA,$1E,$95,$F8,$87,$C1,$3B,$72,$74,$AF,$46,$1F,$1E,$39,$C7
	.dc.b	$6A,$F1,$47,$C1,$AF,$33,$F2,$A3,$38,$7E,$32,$3A,$1D,$8F,$C0,$1E
	.dc.b	$28,$F2,$4F,$8C,$8E,$09,$1D,$48,$78,$5C,$BD,$23,$8B,$E0,$7C,$AE
	.dc.b	$63,$CB,$87,$C9,$8F,$25,$7D,$67,$84,$F4,$8C,$47,$13,$24,$30,$47
	.dc.b	$E0,$3F,$85,$F2,$FE,$71,$27,$E9,$3E,$49,$EA,$4A,$F2,$5F,$AA,$FA
	.dc.b	$46,$2F,$E6,$E3,$F9,$6A,$E4,$FD,$7A,$F9,$87,$96,$FA,$C6,$EB,$D8
	.dc.b	$EE,$BC,$57,$FC,$A3,$D4,$FF,$1C,$5F,$F1,$86,$C6,$71,$FC,$54,$7F
	.dc.b	$8F,$71,$0F,$9C,$B1,$98,$C1,$FE,$BE,$9F,$F7,$53,$88,$D4,$F0,$9C
	.dc.b	$BF,$C7,$67,$FD,$AF,$CB,$5E,$78,$FF,$26,$72,$7F,$0C,$3C,$73,$C7
	.dc.b	$99,$5C,$63,$0D,$9E,$13,$A7,$91,$DB,$4F,$F8,$01,$19,$BF,$04,$E4
	.dc.b	$F8,$27,$89,$CD,$F6,$8E,$6D,$8C,$C7,$82,$FC,$51,$EB,$AE,$D7,$F0
	.dc.b	$AB,$47,$23,$FC,$EA,$F9,$2B,$E8,$E1,$D0,$E4,$E2,$39,$5E,$2B,$93
	.dc.b	$4F,$19,$63,$67,$58,$DB,$63,$6D,$37,$1F,$0A,$FD,$2F,$C4,$3E,$31
	.dc.b	$F8,$67,$2F,$C8,$C3,$E2,$4F,$28,$D5,$FC,$48,$ED,$1C,$E7,$6D,$3E
	.dc.b	$2D,$F9,$1F,$15,$71,$69,$EA,$2C,$6D,$EE,$64,$AE,$5F,$B9,$5E,$56
	.dc.b	$9E,$18,$F9,$0D,$8F,$3B,$ED,$3C,$AF,$1B,$F8,$31,$F9,$63,$27,$13
	.dc.b	$CB,$1B,$AF,$54,$79,$23,$0F,$9C,$F1,$9D,$EB,$C6,$D8,$F1,$BD,$E3
	.dc.b	$2D,$8C,$BF,$58,$E6,$64,$F1,$9E,$53,$C4,$53,$67,$1A,$F6,$B6,$BD
	.dc.b	$2E,$AF,$EC,$BE,$53,$A9,$4E,$A5,$38,$1F,$18,$C4,$F6,$8E,$71,$E7
	.dc.b	$38,$94,$E3,$7F,$01,$F2,$1F,$26,$79,$3D,$4E,$0E,$23,$84,$FF,$8F
	.dc.b	$F8,$3F,$CF,$9E,$37,$F0,$7D,$FF,$84,$61,$EB,$9C,$E3,$FC,$92,$73
	.dc.b	$3F,$2E,$72,$FF,$97,$27,$F5,$31,$89,$DC,$FF,$0D,$8C,$4E,$B8,$7F
	.dc.b	$DA,$EF,$FA,$3F,$E7,$3C,$4F,$FA,$1B,$E7,$2D,$FF,$95,$26,$DF,$FA
	.dc.b	$4E,$6D,$A6,$C6,$1B,$7E,$93,$F9,$EF,$9A,$E6,$7C,$C7,$2E,$8E,$F3
	.dc.b	$98,$C5,$71,$CD,$39,$8F,$F5,$35,$70,$BE,$71,$83,$FD,$55,$F3,$9E
	.dc.b	$6F,$9C,$F3,$B6,$32,$7F,$AB,$E3,$2D,$33,$87,$FA,$BE,$32,$D8,$FF
	.dc.b	$5B,$7C,$E7,$C2,$F9,$9F,$86,$D8,$C9,$FE,$BB,$8C,$B4,$FF,$5F,$7C
	.dc.b	$CF,$CA,$F9,$CF,$96,$D8,$C9,$FE,$C3,$8C,$B4,$F8,$97,$FD,$A3,$C1
	.dc.b	$FA,$7E,$5F,$D8,$8C,$4F,$FB,$2E,$31,$09,$DC,$F2,$8C,$3B,$31,$1A
	.dc.b	$8F,$C7,$1F,$D5,$9C,$72,$7F,$B4,$7E,$2F,$F0,$67,$E1,$38,$6D,$FF
	.dc.b	$B6,$FE,$5F,$F1,$39,$FF,$6A,$8C,$1F,$77,$99,$19,$CB,$BC,$63,$E7
	.dc.b	$38,$9C,$AA,$9C,$69,$F2,$DF,$8A,$DF,$FB,$DE,$9F,$DC,$F0,$EB,$39
	.dc.b	$EC,$98,$71,$F8,$5B,$9F,$83,$24,$7A,$09,$3F,$93,$E7,$67,$DA,$3F
	.dc.b	$C3,$39,$7A,$AB,$DF,$F1,$52,$7C,$DF,$39,$72,$ED,$3F,$13,$E0,$B1
	.dc.b	$B7,$9F,$44,$63,$0E,$B1,$87,$E0,$57,$17,$C9,$9B,$AB,$E1,$1F,$43
	.dc.b	$E5,$2C,$EA,$3E,$9F,$C4,$E3,$3F,$80,$DC,$FC,$4D,$AF,$0B,$F6,$BE
	.dc.b	$72,$47,$D1,$F8,$BB,$7F,$26,$7E,$FF,$CE,$63,$0F,$63,$6F,$D6,$DA
	.dc.b	$70,$FE,$0A,$3E,$D7,$34,$AF,$BD,$F3,$96,$FF,$27,$C9,$F4,$AB,$F1
	.dc.b	$BF,$0B,$9C,$BE,$53,$70,$F9,$2D,$BE,$FC,$48,$FE,$23,$F3,$8F,$57
	.dc.b	$F9,$4E,$48,$C6,$1F,$E5,$08,$C4,$EA,$73,$87,$F8,$5E,$73,$5F,$95
	.dc.b	$F9,$CE,$B7,$FA,$26,$C6,$4F,$C6,$5F,$2C,$BF,$05,$3F,$BB,$F7,$3F
	.dc.b	$3B,$5F,$9F,$F9,$CC,$76,$7E,$B1,$B1,$93,$F8,$2C,$F0,$CF,$F3,$93
	.dc.b	$FB,$2F,$90,$FE,$7B,$E8,$3E,$71,$F0,$F9,$C7,$F9,$84,$E6,$31,$87
	.dc.b	$F0,$28,$C4,$FE,$C6,$F9,$DF,$28,$E3,$F3,$8E,$BE,$ED,$8F,$DC,$7C
	.dc.b	$86,$26,$47,$07,$F8,$3C,$63,$1F,$E2,$23,$E6,$9F,$09,$FD,$7A,$30
	.dc.b	$D8,$CD,$FF,$A1,$A7,$2D,$8C,$C7,$E2,$BE,$6B,$D0,$F6,$8C,$F1,$7F
	.dc.b	$9E,$F9,$55,$A8,$FE,$EF,$3A,$BF,$EB,$B3,$97,$FC,$F7,$19,$BF,$EB
	.dc.b	$33,$97,$FD,$3F,$F3,$2F,$27,$D4,$DB,$F4,$36,$7F,$D3,$ED,$7F,$CC
	.dc.b	$8E,$70,$FF,$05,$7E,$34,$D3,$FA,$14,$E6,$FF,$C1,$53,$97,$FA,$4F
	.dc.b	$CC,$FF,$1E,$F2,$7E,$33,$71,$FF,$56,$B5,$FF,$39,$35,$FF,$4B,$FB
	.dc.b	$8F,$8C,$7C,$D7,$CF,$8C,$36,$33,$3F,$EB,$89,$CB,$DB,$C9,$73,$F8
	.dc.b	$9B,$19,$8C,$1F,$6F,$E9,$DE,$8F,$FA,$A3,$83,$FC,$53,$18,$AF,$EC
	.dc.b	$AD,$7F,$13,$EB,$FF,$61,$73,$5C,$76,$FF,$58,$D6,$EF,$9C,$E6,$BF
	.dc.b	$DA,$B7,$C5,$FC,$9F,$CD,$3C,$AF,$C7,$42,$30,$FE,$45,$AF,$FB,$61
	.dc.b	$AF,$FA,$75,$CD,$FF,$90,$9B,$8F,$FB,$11,$B3,$FE,$E6,$F9,$BF,$8A
	.dc.b	$F8,$AF,$CC,$F8,$5B,$3C,$21,$3C,$2F,$FA,$77,$CC,$FB,$7C,$BF,$EE
	.dc.b	$2D,$F2,$0D,$F9,$8C,$33,$0F,$EE,$0E,$5F,$EF,$5F,$AB,$F5,$D3,$FB
	.dc.b	$1A,$F8,$3E,$CD,$9D,$F0,$E0,$CC,$C8,$C9,$FC,$6B,$6F,$9B,$E3,$7C
	.dc.b	$88,$FF,$73,$B5,$07,$0D,$8F,$8B,$1F,$0C,$F9,$7A,$30,$7C,$B6,$BD
	.dc.b	$4F,$89,$F3,$5F,$0B,$1F,$F0,$B3,$5F,$E2,$78,$7F,$76,$7F,$97,$CF
	.dc.b	$98,$8F,$F1,$18,$D3,$FE,$DB,$3C,$D1,$DD,$E2,$7C,$AF,$97,$FC,$87
	.dc.b	$86,$F9,$87,$CA,$7C,$AF,$2E,$5F,$F8,$4E,$72,$D9,$FE,$41,$5C,$5B
	.dc.b	$3A,$9F,$EA,$CD,$7E,$49,$FE,$6E,$75,$CB,$FC,$85,$BC,$5F,$E7,$A6
	.dc.b	$CF,$F7,$8D,$9F,$23,$E8,$97,$C8,$FE,$37,$E6,$7F,$01,$7F,$E3,$C6
	.dc.b	$E3,$D6,$78,$1E,$62,$4F,$F6,$FA,$E6,$D8,$DB,$F0,$5E,$DC,$3F,$23
	.dc.b	$E5,$BF,$6D,$7F,$E4,$1C,$9F,$33,$F9,$23,$5F,$05,$31,$FE,$F9,$7F
	.dc.b	$BD,$7E,$1E,$E8,$C4,$7F,$A6,$9A,$FF,$A5,$1E,$FB,$66,$BA,$4F,$F9
	.dc.b	$6B,$07,$C7,$F9,$7E,$2F,$E2,$65,$2B,$FC,$99,$8A,$75,$1E,$47,$ED
	.dc.b	$C6,$63,$04,$3F,$C4,$6E,$5B,$1F,$46,$BF,$63,$F1,$F2,$7F,$6E,$FC
	.dc.b	$7F,$88,$91,$F0,$9E,$E4,$7F,$31,$D3,$FA,$7F,$1F,$71,$E0,$9F,$DE
	.dc.b	$C7,$EF,$BE,$66,$4F,$F0,$47,$F0,$5F,$33,$F5,$7C,$6F,$59,$27,$E8
	.dc.b	$BF,$DD,$F0,$73,$7F,$7E,$9F,$F4,$33,$5F,$C5,$7A,$7F,$C3,$2F,$F4
	.dc.b	$FF,$A8,$3E,$23,$F8,$7C,$CA,$FF,$3C,$4E,$5B,$3F,$A6,$8C,$47,$F8
	.dc.b	$E5,$AF,$E0,$9A,$FF,$B6,$9D,$F3,$7F,$81,$5B,$CB,$9A,$FE,$3F,$DF
	.dc.b	$FC,$83,$B3,$F9,$0F,$CD,$4C,$7F,$93,$C6,$5F,$E5,$31,$97,$F1,$11
	.dc.b	$FC,$A3,$E6,$BF,$94,$FC,$D4,$9F,$E5,$F1,$97,$F9,$8C,$65,$FC,$84
	.dc.b	$7F,$30,$F8,$1F,$E6,$3D,$D2,$7F,$9B,$C6,$5F,$E7,$31,$97,$F3,$11
	.dc.b	$97,$F3,$51,$9D,$7F,$59,$3F,$43,$E1,$3F,$9E,$FC,$D4,$BF,$9F,$19
	.dc.b	$69,$0F,$F5,$EF,$EB,$1F,$C4,$1B,$EB,$7F,$53,$87,$CD,$D3,$11,$27
	.dc.b	$FD,$91,$3A,$6C,$E5,$FF,$23,$FA,$BF,$C5,$1F,$C4,$E3,$FC,$4C,$FE
	.dc.b	$CF,$19,$8D,$F8,$DF,$DE,$A3,$11,$E5,$6B,$FE,$20,$8D,$1C,$88,$F9
	.dc.b	$3C,$3F,$ED,$9F,$2B,$ED,$D7,$FD,$A4,$B3,$C9,$AA,$87,$FB,$8E,$34
	.dc.b	$D8,$CB,$FE,$A0,$F9,$4F,$F5,$F3,$BF,$9B,$EE,$7F,$48,$8C,$C6,$FE
	.dc.b	$3B,$04,$7E,$4C,$FC,$2F,$09,$8F,$F6,$AF,$98,$7C,$CF,$FB,$73,$37
	.dc.b	$CD,$7B,$0E,$15,$B9,$FF,$21,$DF,$2C,$7F,$C0,$11,$FE,$46,$F1,$BF
	.dc.b	$A6,$FA,$7D,$27,$FB,$BF,$0F,$F2,$47,$FB,$CC,$65,$42,$FA,$4E,$63
	.dc.b	$18,$78,$6B,$9D,$FF,$5A,$CF,$4D,$C7,$CB,$5B,$AE,$B8,$F9,$EF,$71
	.dc.b	$9C,$3C,$8A,$3C,$13,$BB,$E7,$39,$7D,$CF,$1C,$3C,$0F,$BA,$78,$37
	.dc.b	$0D,$30,$E4,$64,$A7,$99,$D5,$F9,$A3,$38,$7C,$26,$D7,$C6,$C3,$C7
	.dc.b	$1F,$48,$F8,$A7,$CE,$7D,$38,$FB,$30,$EB,$1C,$30,$C5,$74,$9C,$BC
	.dc.b	$8C,$12,$3C,$84,$9F,$8A,$F2,$AF,$41,$23,$C2,$68,$F0,$12,$7E,$14
	.dc.b	$F1,$8F,$5C,$F1,$9F,$8C,$F9,$A7,$C1,$1E,$57,$13,$D6,$30,$F5,$7E
	.dc.b	$61,$E9,$71,$87,$E0,$27,$8C,$F6,$9F,$09,$1C,$1E,$67,$8E,$3F,$80
	.dc.b	$8E,$B3,$C9,$FC,$DC,$66,$BB,$57,$96,$7C,$77,$ED,$76,$7D,$73,$B9
	.dc.b	$D2,$3B,$BF,$97,$C7,$F1,$AE,$4F,$79,$0C,$1D,$A3,$D4,$F6,$3F,$0C
	.dc.b	$64,$F6,$1C,$6B,$CE,$E0,$E6,$E1,$E5,$87,$BE,$3F,$28,$FD,$48,$CC
	.dc.b	$FB,$E7,$F3,$17,$EF,$7F,$5E,$FD,$F9,$ED,$1D,$49,$7E,$E7,$8D,$7A
	.dc.b	$9E,$2F,$E3,$27,$E3,$C6,$1F,$CC,$57,$57,$F0,$3E,$A7,$F9,$52,$7F
	.dc.b	$54,$DB,$FE,$34,$DC,$BF,$D1,$AF,$F9,$6A,$BF,$D1,$FF,$35,$8C,$3F
	.dc.b	$CF,$8D,$8F,$EB,$8D,$7F,$CD,$AD,$78,$9F,$40,$FE,$89,$E0,$7D,$F7
	.dc.b	$66,$78,$91,$98,$E3,$5F,$C9,$DE,$27,$31,$5B,$AE,$31,$98,$FE,$44
	.dc.b	$F0,$D3,$93,$C0,$FE,$22,$67,$4F,$6B,$E3,$C3,$C4,$EF,$0F,$16,$1D
	.dc.b	$9E,$78,$76,$8F,$0E,$18,$AE,$B1,$8C,$3A,$CF,$2C,$3C,$B5,$E3,$8C
	.dc.b	$3E,$43,$13,$FE,$09,$AE,$51,$F4,$0E,$B5,$CA,$FE,$13,$E8,$6C,$7C
	.dc.b	$33,$67,$57,$D1,$1C,$63,$37,$E9,$3C,$67,$AF,$0F,$29,$70,$F2,$57
	.dc.b	$7A,$E3,$1E,$E3,$2F,$73,$27,$A6,$BA,$C7,$D0,$AE,$AD,$38,$BE,$37
	.dc.b	$9C,$59,$FF,$5D,$E1,$F1,$1F,$C5,$C7,$47,$B4,$F0,$3C,$8F,$0B,$F8
	.dc.b	$13,$B8,$FC,$9D,$F8,$9F,$9C,$D7,$AD,$73,$7A,$9C,$CF,$74,$6D,$FC
	.dc.b	$CC,$7A,$8F,$0B,$FE,$FA,$BF,$43,$FA,$6A,$F1,$36,$7B,$3E,$03,$B4
	.dc.b	$7C,$17,$95,$FA,$A7,$71,$E0,$3C,$97,$F5,$DB,$1E,$27,$F4,$07,$A8
	.dc.b	$E6,$7D,$7A,$F1,$BF,$92,$3C,$31,$C6,$FD,$31,$93,$8C,$F9,$5E,$A7
	.dc.b	$A4,$CC,$F1,$BF,$44,$74,$3E,$7C,$7C,$76,$9E,$83,$E7,$F8,$AE,$0F
	.dc.b	$F6,$55,$27,$FC,$7A,$FD,$E9,$D7,$95,$FC,$13,$5F,$E0,$BE,$73,$95
	.dc.b	$F2,$D9,$C3,$94,$E6,$31,$8B,$17,$5B,$E6,$F3,$7B,$C3,$C5,$97,$6C
	.dc.b	$BA,$C7,$F9,$7F,$A3,$71,$8C,$BC,$E7,$5D,$53,$E0,$CF,$DD,$F3,$36
	.dc.b	$5F,$2D,$63,$0D,$C7,$FC,$D8,$DC,$3A,$DF,$C7,$8F,$5C,$7C,$7E,$BA
	.dc.b	$9F,$06,$DF,$E6,$4B,$F5,$9F,$E9,$98,$D9,$FC,$BA,$7F,$D3,$87,$18
	.dc.b	$AF,$56,$B9,$33,$1E,$BF,$33,$9C,$FF,$A0,$DF,$0F,$5F,$8B,$8A,$CF
	.dc.b	$FA,$89,$B7,$FE,$AC,$6E,$FF,$D0,$33,$FA,$B1,$F5,$3F,$52,$77,$53
	.dc.b	$DB,$F0,$17,$1F,$6E,$2E,$35,$F0,$9F,$F5,$8F,$CE,$6E,$BF,$D7,$6F
	.dc.b	$C6,$E3,$F8,$FF,$39,$6B,$FD,$6E,$DA,$FF,$60,$B7,$5F,$37,$0A,$31
	.dc.b	$5F,$99,$DB,$F1,$B1,$8B,$FC,$85,$6E,$BF,$D8,$7E,$9F,$D6,$E4,$FA
	.dc.b	$ED,$8C,$9F,$EC,$88,$CB,$4F,$F6,$5F,$27,$DE,$F8,$0F,$BE,$D8,$C9
	.dc.b	$FE,$D0,$8C,$B4,$F8,$FB,$FC,$6F,$2B,$E7,$1F,$88,$F9,$CB,$3A,$E1
	.dc.b	$4F,$E1,$F4,$FD,$6C,$62,$7F,$13,$7C,$DB,$19,$9F,$F7,$44,$E5,$FF
	.dc.b	$75,$F3,$76,$6D,$7D,$AD,$D9,$3E,$4F,$F5,$4D,$9F,$F7,$5F,$C6,$62
	.dc.b	$7F,$73,$AF,$F8,$33,$0F,$E1,$47,$E0,$DC,$3F,$C1,$1A,$E7,$C0,$F6
	.dc.b	$3F,$5B,$BB,$F5,$27,$83,$E7,$2E,$3F,$EF,$AF,$98,$EF,$E0,$FE,$B9
	.dc.b	$B1,$F3,$31,$FD,$86,$1F,$01,$AF,$93,$7F,$E0,$37,$E0,$6D,$7E,$CB
	.dc.b	$67,$EB,$3F,$E2,$A8,$F8,$8D,$BF,$E1,$F5,$C1,$FE,$0B,$19,$E8,$FE
	.dc.b	$4B,$8F,$F1,$2B,$F2,$74,$FC,$AF,$92,$F8,$EC,$8F,$B3,$3B,$C3,$D1
	.dc.b	$3E,$58,$C4,$7A,$5D,$5F,$ED,$F8,$9F,$39,$CE,$BF,$8B,$C6,$30,$F8
	.dc.b	$8F,$D6,$3F,$92,$64,$F5,$B1,$3A,$3C,$73,$E1,$23,$83,$E1,$F1,$F0
	.dc.b	$BF,$82,$FE,$0B,$82,$4F,$E2,$E9,$E8,$62,$FE,$4E,$8B,$F7,$12,$FE
	.dc.b	$1B,$67,$D7,$F3,$5C,$75,$F8,$0F,$F4,$8F,$D5,$1F,$1D,$4F,$6F,$2D
	.dc.b	$C7,$F5,$F3,$A7,$F4,$ED,$C3,$F4,$AD,$7F,$0C,$D7,$FA,$DB,$BC,$3F
	.dc.b	$93,$6F,$6B,$38,$7C,$C3,$FA,$2F,$B3,$FA,$7C,$3F,$AA,$87,$17,$CB
	.dc.b	$E7,$B7,$C1,$30,$FE,$83,$5C,$6F,$F5,$67,$19,$FD,$3B,$67,$71,$FB
	.dc.b	$6F,$9A,$E5,$3F,$82,$7C,$7C,$72,$BF,$37,$F1,$36,$BC,$B8,$7E,$1D
	.dc.b	$FE,$A6,$D7,$C2,$D7,$F7,$AD,$7F,$BE,$68,$E6,$FC,$6E,$1F,$8E,$6E
	.dc.b	$1F,$AA,$6D,$7E,$A1,$BA,$7C,$CF,$9E,$FC,$A9,$F9,$6F,$8C,$B7,$FB
	.dc.b	$28,$B7,$FB,$09,$B5,$FD,$AF,$E5,$5D,$FF,$DD,$9B,$87,$F3,$0F,$A2
	.dc.b	$7C,$56,$6A,$4F,$F5,$AC,$57,$C1,$FD,$FD,$F0,$9D,$4E,$70,$F0,$5F
	.dc.b	$D4,$8E,$17,$F6,$1A,$FF,$24,$6B,$FE,$2F,$FB,$17,$D0,$7F,$83,$1B
	.dc.b	$87,$9F,$CE,$C4,$C8,$F3,$4E,$AB,$CB,$59,$7F,$1F,$F2,$EE,$97,$8B
	.dc.b	$FE,$0C,$8C,$36,$33,$3F,$E1,$29,$CB,$FE,$14,$F6,$DF,$14,$F9,$7F
	.dc.b	$F0,$AC,$61,$B1,$99,$E0,$D7,$F3,$35,$FB,$6F,$73,$C5,$DA,$69,$27
	.dc.b	$FC,$47,$F2,$EF,$2F,$FC,$42,$D8,$FF,$12,$B5,$DC,$68,$FE,$A3,$C7
	.dc.b	$FE,$2D,$9D,$36,$33,$C1,$FD,$5B,$D4,$FF,$1C,$5F,$E0,$7E,$5D,$D1
	.dc.b	$A8,$FF,$32,$7C,$BF,$E2,$1F,$C3,$35,$FD,$EC,$F0,$79,$4F,$F8,$E2
	.dc.b	$31,$1F,$E3,$E6,$BF,$BC,$6B,$FD,$A1,$DE,$3F,$E4,$56,$E1,$9A,$F2
	.dc.b	$52,$FE,$53,$FC,$BF,$E5,$BD,$1E,$6F,$F7,$A6,$F9,$46,$FC,$BE,$19
	.dc.b	$5C,$4F,$E7,$11,$CB,$E4,$FA,$C7,$F9,$89,$7C,$0E,$4D,$70,$C9,$FF
	.dc.b	$05,$4E,$9B,$39,$7F,$87,$E7,$E6,$F0,$FF,$43,$1F,$DB,$BB,$3D,$BE
	.dc.b	$27,$C0,$8F,$F4,$A3,$F1,$9B,$5F,$E6,$48,$C4,$7F,$2D,$8C,$78,$BF
	.dc.b	$0B,$E5,$FE,$04,$FF,$87,$5A,$FF,$20,$8E,$E6,$F5,$91,$FC,$0E,$D3
	.dc.b	$A8,$C3,$5F,$89,$F2,$FF,$87,$F4,$9F,$2C,$F8,$EF,$95,$F2,$F2,$FF
	.dc.b	$C8,$33,$96,$CE,$5F,$9D,$3C,$E7,$51,$FE,$10,$6F,$47,$FA,$C9,$BC
	.dc.b	$FF,$EB,$06,$CF,$FA,$BF,$8B,$FC,$F2,$F8,$3E,$3B,$F0,$BF,$29,$F5
	.dc.b	$63,$FD,$3A,$DC,$3A,$4F,$84,$DF,$24,$AE,$44,$3C,$91,$B7,$FC,$95
	.dc.b	$5F,$A5,$AF,$F6,$3F,$CB,$7E,$BA,$7F,$15,$F2,$FF,$8D,$3E,$44,$66
	.dc.b	$77,$F0,$D9,$AF,$C5,$D6,$A2,$47,$F9,$AA,$B8,$69,$FE,$D0,$6C,$7C
	.dc.b	$1F,$8D,$EB,$F0,$DF,$03,$09,$3C,$63,$2F,$A2,$32,$FF,$8F,$A9,$3A
	.dc.b	$8C,$37,$1F,$F5,$46,$4C,$3F,$51,$F1,$B2,$FF,$46,$FB,$BE,$12,$46
	.dc.b	$08,$77,$7F,$61,$E3,$EA,$30,$7E,$B1,$FD,$A7,$CB,$C8,$FD,$7B,$F7
	.dc.b	$3E,$5F,$E5,$F8,$D6,$7E,$9C,$61,$FA,$BF,$2F,$F8,$22,$3F,$B8,$8F
	.dc.b	$CD,$7C,$BF,$F9,$16,$3F,$77,$3F,$C0,$CC,$FC,$0C,$8C,$90,$FF,$32
	.dc.b	$CF,$16,$C7,$FB,$73,$A9,$39,$9F,$F3,$A3,$51,$FE,$CB,$70,$FE,$6F
	.dc.b	$AB,$F8,$A7,$CB,$C8,$FE,$1F,$1F,$C5,$CC,$3F,$C5,$E3,$0F,$F1,$4F
	.dc.b	$97,$FE,$39,$F1,$12,$30,$43,$F8,$F3,$97,$F9,$04,$7F,$23,$F9,$5F
	.dc.b	$E4,$9D,$52,$7F,$8F,$C6,$5F,$E4,$31,$97,$F9,$44,$65,$FE,$55,$19
	.dc.b	$E5,$FF,$78,$76,$2B,$E8,$FC,$2F,$F3,$1F,$97,$98,$7F,$B2,$35,$FF
	.dc.b	$35,$B6,$7E,$64,$62,$3F,$D7,$ED,$7E,$6C,$7F,$92,$9F,$F7,$37,$C6
	.dc.b	$FF,$80,$16,$3F,$EE,$CF,$97,$FE,$93,$26,$FB,$12,$BF,$D8,$D3,$96
	.dc.b	$CE,$5F,$F2,$FF,$AF,$FC,$41,$FE,$1B,$9F,$F1,$03,$FA,$CC,$67,$07
	.dc.b	$81,$FE,$C8,$8C,$3F,$EF,$BE,$DF,$89,$7F,$C3,$2F,$8F,$B3,$23,$FC
	.dc.b	$26,$FB,$5F,$C4,$DF,$85,$FF,$34,$B5,$FC,$6F,$A2,$F9,$2F,$F3,$6E
	.dc.b	$5E,$29,$4E,$75,$FE,$D7,$5F,$1E,$B5,$3C,$6B,$26,$AA,$3F,$EE,$BA
	.dc.b	$EB,$9F,$E7,$39,$DF,$26,$F8,$BE,$BF,$2F,$30,$FD,$0E,$7E,$3B,$DD
	.dc.b	$7F,$4D,$BE,$15,$FF,$00,$1A,$FF,$BF,$9A,$FF,$B7,$DA,$FC,$DF,$0B
	.dc.b	$EF,$AF,$96,$7C,$77,$C4,$D1,$F2,$BF,$00,$FA,$F7,$C2,$B7,$7F,$3A
	.dc.b	$31,$7C,$6B,$84,$FF,$9A,$F8,$15,$A8,$E1,$A7,$EF,$BC,$BE,$70,$AF
	.dc.b	$E5,$4E,$70,$FC,$27,$8F,$FC,$12,$3F,$C2,$31,$EB,$BD,$47,$43,$D9
	.dc.b	$1C,$1F,$E1,$79,$76,$BE,$B1,$A3,$84,$67,$1F,$8B,$8E,$F1,$F0,$36
	.dc.b	$BB,$9E,$79,$DE,$1E,$38,$E8,$7A,$EB,$D1,$7C,$63,$38,$7C,$4C,$7C
	.dc.b	$B7,$C7,$1D,$11,$C1,$D6,$F9,$11,$C1,$AB,$FD,$0E,$1E,$88,$F8,$51
	.dc.b	$F2,$0F,$8D,$5E,$FB,$F0,$C7,$02,$9E,$92,$E1,$E7,$7D,$38,$FD,$B7
	.dc.b	$47,$B3,$0F,$7E,$1E,$48,$F9,$AF,$8B,$1F,$44,$60,$F1,$C7,$9B,$1F
	.dc.b	$81,$18,$34,$76,$9E,$2F,$D1,$C7,$CB,$87,$D2,$8F,$51,$EF,$3C,$B5
	.dc.b	$EF,$71,$8F,$AE,$F9,$BF,$84,$74,$4C,$3D,$04,$9F,$1C,$F1,$8C,$47
	.dc.b	$95,$F8,$0F,$94,$EF,$8F,$D3,$7C,$B7,$E3,$3B,$3E,$8B,$F1,$E1,$F3
	.dc.b	$70,$ED,$39,$C3,$BC,$62,$3B,$5F,$C8,$7D,$F3,$F8,$37,$C5,$7E,$EC
	.dc.b	$3E,$AC,$7A,$B0,$ED,$18,$8F,$51,$9C,$7F,$03,$19,$C7,$E3,$61,$EE
	.dc.b	$7C,$46,$2B,$CC,$71,$3F,$08,$72,$3E,$BD,$47,$24,$C3,$E1,$BE,$8A
	.dc.b	$F7,$DF,$87,$0F,$88,$7A,$27,$A1,$F9,$4B,$E9,$7E,$F8,$B7,$FB,$5F
	.dc.b	$5B,$FC,$C1,$8F,$E6,$DB,$87,$F2,$46,$E1,$FE,$12,$7F,$CC,$D7,$FE
	.dc.b	$14,$F9,$7C,$67,$FD,$AD,$B1,$FC,$54,$38,$7F,$8C,$35,$E4,$7F,$06
	.dc.b	$3F,$DE,$BC,$5F,$B0,$73,$9F,$F2,$47,$26,$CE,$A7,$1E,$3F,$EB,$11
	.dc.b	$B8,$CB,$C6,$33,$A7,$9E,$78,$57,$77,$71,$E9,$9D,$C7,$AE,$FC,$91
	.dc.b	$88,$F5,$9C,$8F,$24,$6A,$BA,$36,$32,$6E,$FC,$D1,$96,$C6,$6F,$CD
	.dc.b	$1E,$9A,$E7,$39,$3C,$85,$3C,$71,$B7,$BB,$4E,$B8,$64,$B5,$EF,$AF
	.dc.b	$E7,$75,$F4,$23,$0F,$26,$D7,$98,$E8,$F9,$8C,$9F,$36,$BB,$34,$E6
	.dc.b	$F1,$77,$3E,$A9,$E6,$D8,$CB,$EC,$6B,$E5,$3F,$B1,$61,$F4,$DC,$1F
	.dc.b	$E6,$87,$0F,$E1,$AB,$C6,$FC,$86,$BD,$4E,$77,$F3,$AF,$C7,$1F,$7A
	.dc.b	$76,$FE,$61,$AF,$E2,$0E,$27,$C7,$2C,$7B,$DB,$7F,$1D,$A7,$BC,$A7
	.dc.b	$8C,$F5,$57,$E0,$2B,$BE,$1F,$5B,$0F,$F1,$B4,$75,$AD,$9E,$12,$E1
	.dc.b	$F7,$9F,$2C,$57,$6F,$17,$F5,$B1,$67,$E4,$D7,$7B,$FC,$2D,$7C,$BC
	.dc.b	$3E,$D9,$EB,$7B,$B5,$E6,$FE,$05,$F0,$B5,$F0,$BF,$0F,$93,$FC,$A5
	.dc.b	$E5,$7F,$94,$74,$FF,$29,$57,$D5,$9C,$DF,$F9,$10,$E1,$C3,$FE,$D4
	.dc.b	$71,$3F,$2E,$76,$FF,$9F,$23,$FC,$A1,$C4,$F9,$1E,$B3,$86,$AE,$5F
	.dc.b	$F2,$AC,$61,$B5,$FE,$4F,$8F,$33,$FE,$47,$E5,$B5,$C6,$FB,$4F,$F9
	.dc.b	$86,$73,$87,$F9,$BF,$E5,$2E,$4F,$92,$B1,$86,$DF,$F9,$A5,$B8,$FA
	.dc.b	$CF,$F3,$AD,$71,$9F,$83,$1F,$EC,$8F,$43,$D6,$D8,$FF,$42,$28,$CB
	.dc.b	$63,$31,$E4,$F2,$FD,$73,$C4,$F0,$63,$FE,$87,$8C,$C7,$13,$36,$8C
	.dc.b	$D7,$3B,$FF,$53,$F4,$FC,$2E,$FF,$B1,$87,$47,$FC,$DF,$3F,$EA,$CE
	.dc.b	$1D,$56,$DB,$19,$3E,$FF,$03,$E4,$DD,$BF,$5D,$B3,$FE,$BF,$9D,$4E
	.dc.b	$49,$1F,$EB,$AF,$96,$FB,$33,$A8,$C1,$FE,$BE,$E1,$F4,$7C,$B7,$A5
	.dc.b	$B1,$93,$FD,$87,$19,$69,$9D,$3D,$D1,$96,$C7,$FB,$2B,$D1,$F8,$9C
	.dc.b	$7F,$15,$B1,$93,$FD,$9F,$19,$69,$FE,$D2,$E3,$F9,$9E,$FF,$CD,$6C
	.dc.b	$64,$FF,$6B,$C6,$5A,$7C,$AD,$5E,$5F,$C9,$24,$7C,$98,$CB,$63,$ED
	.dc.b	$71,$FC,$89,$CE,$5F,$6E,$31,$3A,$AE,$CD,$9F,$C5,$DB,$E4,$7F,$4A
	.dc.b	$DA,$FF,$76,$FC,$5E,$27,$EA,$E5,$FE,$02,$EA,$FF,$68,$DB,$E2,$73
	.dc.b	$DD,$C9,$FC,$77,$64,$CB,$F8,$2F,$5A,$F8,$47,$FB,$EA,$7F,$29,$F2
	.dc.b	$BD,$BB,$AC,$FE,$3B,$3F,$F6,$2B,$63,$F7,$7F,$05,$FE,$F8,$8C,$1D
	.dc.b	$9C,$6F,$FA,$86,$EB,$FE,$E8,$6C,$FF,$B8,$EF,$AD,$6E,$3D,$FC,$D1
	.dc.b	$F4,$E6,$F9,$0C,$57,$F0,$3E,$8F,$E1,$57,$F2,$BE,$47,$39,$FC,$66
	.dc.b	$E3,$FF,$00,$31,$46,$30,$FF,$7F,$C6,$27,$E0,$3F,$40,$F7,$FA,$B6
	.dc.b	$32,$7B,$F8,$F1,$5E,$F6,$E1,$EF,$6C,$A5,$C2,$DB,$F0,$11,$EB,$D9
	.dc.b	$C1,$F8,$03,$F0,$BE,$75,$C3,$EA,$47,$E2,$3E,$57,$17,$F8,$C6,$E7
	.dc.b	$F8,$66,$D7,$C1,$7F,$7F,$F2,$AF,$8C,$A7,$DC,$E6,$B5,$FC,$3A,$74
	.dc.b	$FD,$D9,$CE,$1F,$75,$B8,$7D,$D6,$CF,$BF,$5F,$74,$7F,$2C,$F8,$3F
	.dc.b	$7D,$7F,$5E,$6D,$6F,$67,$CA,$EF,$82,$4F,$E0,$BE,$56,$E7,$FA,$A6
	.dc.b	$E3,$FA,$46,$CF,$DF,$7C,$1C,$9F,$D1,$32,$FB,$D7,$FE,$77,$0E,$1F
	.dc.b	$F3,$D3,$5F,$B0,$6A,$3F,$43,$C3,$D2,$FF,$40,$FE,$93,$9F,$F4,$B3
	.dc.b	$67,$F4,$8D,$AF,$D2,$37,$1F,$D2,$3F,$AC,$F9,$59,$5F,$AB,$F8,$9B
	.dc.b	$7F,$AB,$9B,$7F,$AB,$9B,$5E,$6F,$93,$77,$7F,$5A,$6C,$7F,$48,$F9
	.dc.b	$47,$CA,$E6,$A4,$FC,$3C,$97,$C1,$F8,$CD,$7F,$D3,$CD,$7C,$13,$FE
	.dc.b	$02,$C5,$F3,$0F,$99,$EB,$5F,$DE,$67,$31,$FD,$BB,$E5,$DF,$21,$4F
	.dc.b	$EA,$7A,$7F,$5A,$BF,$F0,$00,$9D,$E7,$FE,$00,$6C,$7D,$B6,$BF,$EB
	.dc.b	$36,$BF,$69,$AF,$EC,$FE,$5B,$F9,$57,$8B,$F0,$5B,$1F,$85,$6B,$FE
	.dc.b	$BD,$6B,$F8,$4F,$65,$F1,$2F,$97,$F9,$51,$86,$C6,$67,$F2,$13,$97
	.dc.b	$B7,$63,$DA,$F1,$36,$33,$18,$3E,$B7,$CB,$FF,$83,$E3,$FC,$43,$F2
	.dc.b	$EF,$99,$FC,$BF,$17,$F9,$2A,$31,$3F,$E2,$6A,$E0,$FC,$C6,$BF,$ED
	.dc.b	$96,$BF,$A0,$6B,$FC,$65,$AB,$E6,$1D,$1E,$0E,$1F,$D3,$FC,$C7,$E3
	.dc.b	$AB,$FC,$6A,$D7,$ED,$D7,$07,$FC,$72,$D7,$EF,$C6,$F1,$FF,$31,$B6
	.dc.b	$BF,$67,$5B,$9F,$F3,$3B,$C6,$BF,$DE,$4D,$77,$7F,$E4,$28,$C4,$7F
	.dc.b	$40,$F9,$6F,$43,$4F,$E9,$EE,$3C,$4F,$ED,$1F,$31,$7C,$C4,$D2,$57
	.dc.b	$F5,$69,$E1,$F1,$5F,$66,$3F,$C9,$2B,$E0,$B2,$6B,$86,$4F,$F7,$69
	.dc.b	$D3,$67,$2F,$E8,$BC,$4F,$81,$B7,$F2,$7E,$BF,$37,$91,$FA,$AA,$FF
	.dc.b	$31,$D6,$9F,$EA,$0D,$D7,$FA,$85,$70,$3E,$0C,$FF,$22,$DE,$D3,$E5
	.dc.b	$FE,$17,$CC,$7C,$0A,$FF,$05,$B5,$F2,$47,$03,$7B,$C8,$F1,$C6,$5F
	.dc.b	$E1,$D5,$D5,$B1,$FE,$76,$8C,$47,$F3,$03,$FD,$29,$F2,$CF,$91,$F9
	.dc.b	$BF,$2F,$27,$FC,$C7,$39,$6C,$E5,$FA,$33,$F7,$BD,$55,$6E,$7F,$C1
	.dc.b	$4D,$E6,$70,$7F,$77,$AD,$E1,$C6,$B5,$3F,$D7,$78,$BF,$CE,$38,$78
	.dc.b	$BE,$43,$F1,$1F,$1B,$F6,$A3,$FD,$12,$DC,$BA,$C7,$B2,$7A,$1C,$39
	.dc.b	$24,$73,$21,$E6,$8D,$BF,$E3,$CA,$FD,$23,$FE,$B9,$F9,$6F,$F4,$F3
	.dc.b	$F8,$BF,$97,$FC,$99,$F0,$63,$31,$FE,$29,$6E,$1E,$4E,$56,$4B,$E3
	.dc.b	$7F,$E9,$0B,$DE,$7F,$15,$B1,$E1,$F8,$7E,$C7,$E8,$BE,$37,$C1,$F0
	.dc.b	$DF,$13,$09,$5C,$E3,$2F,$AE,$32,$F9,$63,$2D,$38,$1F,$06,$7F,$5F
	.dc.b	$F1,$B2,$3F,$54,$FB,$FE,$12,$46,$08,$7C,$17,$F6,$FE,$36,$A3,$07
	.dc.b	$EC,$DF,$DD,$FC,$BC,$8F,$DB,$3F,$7B,$E5,$FE,$67,$C2,$59,$FA,$91
	.dc.b	$87,$EB,$7C,$BF,$E1,$08,$FE,$FE,$3F,$29,$F2,$8F,$82,$FE,$13,$52
	.dc.b	$BF,$D7,$F3,$96,$CE,$5F,$E1,$B3,$CA,$24,$7F,$0B,$8D,$72,$7E,$43
	.dc.b	$A9,$7C,$EB,$F8,$AF,$57,$F1,$8F,$97,$91,$FC,$4E,$3F,$8D,$98,$7F
	.dc.b	$8D,$C6,$1F,$E3,$1F,$2F,$FC,$83,$E2,$24,$60,$87,$F2,$27,$2F,$F2
	.dc.b	$48,$FE,$4F,$F2,$BF,$CA,$3B,$24,$FF,$23,$8C,$BF,$C9,$63,$2F,$F2
	.dc.b	$C8,$CB,$FC,$BA,$33,$9F,$E6,$FA,$95,$F3,$BD,$EF,$E6,$BF,$2F,$30
	.dc.b	$F9,$B1,$97,$F9,$94,$65,$FF,$47,$6A,$9F,$E8,$9F,$22,$F5,$AF,$CB
	.dc.b	$CB,$FF,$57,$CE,$5B,$39,$7F,$E0,$04,$F2,$8F,$E6,$3E,$AF,$F0,$A9
	.dc.b	$FE,$0F,$97,$F0,$A3,$3F,$23,$23,$F8,$04,$65,$FF,$46,$6B,$FC,$31
	.dc.b	$57,$38,$C3,$7C,$1F,$F3,$A7,$87,$F1,$2B,$99,$CA,$3F,$C3,$98,$7F
	.dc.b	$61,$F9,$3F,$E6,$58,$7F,$05,$7D,$94,$8C,$65,$B8,$87,$FB,$42,$77
	.dc.b	$3A,$8C,$1F,$CC,$FE,$51,$F1,$BD,$FE,$5E,$4F,$8C,$74,$AB,$71,$98
	.dc.b	$C5,$70,$8F,$26,$7F,$E9,$1C,$58,$77,$AF,$9F,$E0,$FA,$6D,$F2,$EF
	.dc.b	$90,$E1,$F2,$DF,$04,$DE,$BF,$EB,$99,$D5,$7F,$AE,$F8,$58,$73,$9C
	.dc.b	$F9,$4D,$18,$F3,$C3,$D8,$7D,$6B,$E1,$97,$F9,$2D,$B1,$FE,$C6,$AE
	.dc.b	$2F,$CB,$AF,$1B,$E8,$AE,$19,$79,$5F,$F2,$A7,$8D,$FE,$98,$CB,$CF
	.dc.b	$5F,$84,$8C,$4F,$08,$E4,$7A,$F1,$F0,$BB,$3A,$1E,$F9,$E2,$EE,$F2
	.dc.b	$FA,$DB,$9A,$77,$3C,$CE,$E6,$CB,$1F,$2A,$78,$CF,$08,$E5,$3D,$2B
	.dc.b	$CB,$1B,$8E,$33,$B3,$84,$E8,$DB,$98,$E5,$51,$E2,$78,$0C,$61,$F2
	.dc.b	$2F,$B3,$6B,$94,$FA,$2F,$B5,$F1,$8E,$66,$5E,$78,$7D,$48,$D5,$F5
	.dc.b	$AE,$31,$C2,$3B,$3C,$0F,$9C,$66,$79,$3D,$49,$1C,$1D,$1C,$1E,$D5
	.dc.b	$E0,$9E,$31,$89,$E3,$18,$23,$C5,$FC,$34,$75,$23,$82,$47,$E1,$4E
	.dc.b	$C7,$12,$1F,$81,$8E,$2E,$30,$F9,$B3,$E0,$8D,$3D,$1F,$C3,$4F,$07
	.dc.b	$27,$1B,$FB,$AF,$80,$DC,$FE,$77,$CB,$F8,$84,$70,$75,$BF,$C6,$CF
	.dc.b	$C8,$38,$9E,$5C,$7E,$FE,$1F,$1C,$95,$E7,$AE,$31,$8A,$E0,$78,$E3
	.dc.b	$DA,$E0,$EC,$E1,$EE,$7B,$08,$E4,$FC,$F7,$AB,$FE,$4B,$6E,$1F,$B9
	.dc.b	$8C,$63,$FC,$6D,$FF,$28,$50,$7C,$BD,$8C,$C6,$0F,$F6,$85,$7F,$9B
	.dc.b	$DC,$46,$A7,$84,$E5,$FD,$FC,$FF,$61,$E1,$AF,$3C,$7F,$87,$B9,$3F
	.dc.b	$7E,$78,$E7,$8F,$27,$F9,$8A,$77,$3C,$73,$FF,$40,$4E,$5F,$04,$72
	.dc.b	$8C,$3F,$E8,$5A,$E7,$7E,$26,$C7,$80,$E8,$7C,$33,$C9,$7E,$08,$DB
	.dc.b	$63,$B4,$F3,$69,$CE,$77,$7E,$4C,$3D,$F8,$7F,$9F,$1E,$46,$2F,$E9
	.dc.b	$3D,$70,$C6,$1C,$CE,$AE,$6F,$E0,$D7,$86,$36,$66,$7D,$57,$F0,$9F
	.dc.b	$66,$1E,$47,$E2,$1D,$2F,$E4,$D7,$26,$9D,$A3,$C5,$19,$78,$B5,$EA
	.dc.b	$73,$33,$1C,$63,$C9,$1F,$8F,$AF,$23,$F7,$CE,$07,$28,$E8,$E5,$AF
	.dc.b	$D6,$39,$5F,$C1,$AE,$C7,$A2,$3E,$0C,$65,$A7,$AA,$36,$F9,$8C,$CF
	.dc.b	$99,$FC,$0C,$73,$7B,$9C,$4D,$B9,$76,$73,$37,$1C,$EB,$DF,$1E,$93
	.dc.b	$67,$BE,$7C,$6D,$3A,$9C,$CD,$9F,$0D,$DB,$4D,$9C,$4F,$0B,$E5,$8F
	.dc.b	$0D,$FE,$06,$74,$74,$3C,$91,$A7,$E9,$9D,$4E,$67,$C7,$E0,$FF,$0C
	.dc.b	$79,$5F,$ED,$6D,$FF,$C9,$93,$E3,$9C,$E5,$FE,$1A,$8C,$70,$AF,$84
	.dc.b	$FD,$7A,$DD,$7F,$8D,$0E,$D5,$B9,$E5,$1F,$B6,$9D,$DF,$FA,$5F,$95
	.dc.b	$F2,$7D,$2B,$FC,$49,$C9,$79,$BF,$C8,$93,$9C,$3F,$C6,$13,$9B,$E3
	.dc.b	$8F,$F9,$4E,$31,$5B,$CB,$FC,$A3,$3F,$E5,$7F,$95,$C7,$CA,$65,$B1
	.dc.b	$93,$FC,$B9,$19,$69,$F0,$FC,$FE,$9E,$5F,$9F,$5F,$E6,$51,$89,$D1
	.dc.b	$FE,$65,$9D,$4A,$74,$FC,$73,$8C,$EC,$E2,$FC,$BD,$3F,$C9,$D5,$F0
	.dc.b	$7B,$FF,$C9,$F5,$FE,$92,$E0,$C7,$56,$5B,$19,$3F,$D2,$F1,$97,$FC
	.dc.b	$D9,$F2,$AE,$DE,$9F,$2F,$F3,$E3,$13,$FE,$9E,$8C,$4E,$A3,$8B,$73
	.dc.b	$F3,$63,$F3,$9B,$F2,$FE,$7E,$2F,$43,$6F,$FD,$52,$D9,$DC,$7F,$AB
	.dc.b	$27,$6D,$36,$7F,$AC,$3D,$1F,$85,$C3,$F0,$DB,$19,$3F,$D6,$B1,$96
	.dc.b	$9F,$EB,$8E,$1F,$95,$EF,$7C,$B6,$C6,$4F,$F5,$EC,$65,$A7,$C8,$F2
	.dc.b	$FE,$97,$0F,$C7,$6C,$64,$FF,$62,$C6,$5A,$67,$2F,$8F,$19,$6C,$7F
	.dc.b	$A8,$F8,$96,$F9,$2F,$D0,$36,$7F,$DA,$3F,$19,$89,$FD,$36,$5F,$DC
	.dc.b	$FA,$95,$FE,$05,$C3,$9F,$CA,$EB,$B3,$9B,$F8,$BE,$A9,$5E,$1C,$D3
	.dc.b	$8D,$3E,$EF,$CB,$F6,$EF,$FC,$93,$6F,$FD,$6D,$A7,$FA,$B5,$F5,$7C
	.dc.b	$0F,$89,$CD,$7E,$E6,$31,$7F,$9D,$6E,$BE,$88,$6A,$9C,$DF,$F8,$7F
	.dc.b	$BD,$8F,$7B,$E5,$3F,$0B,$96,$3E,$BF,$93,$B3,$FC,$56,$F8,$D7,$CC
	.dc.b	$F9,$3C,$E7,$E1,$CF,$E8,$C6,$31,$F9,$5B,$A3,$18,$F9,$A3,$13,$F0
	.dc.b	$9F,$A6,$7B,$FD,$6B,$19,$3E,$2F,$1E,$2B,$E1,$37,$1F,$16,$DF,$7E
	.dc.b	$7D,$1B,$7E,$12,$F8,$9F,$43,$83,$F0,$27,$E1,$BC,$EB,$19,$3E,$E7
	.dc.b	$CB,$E2,$FF,$86,$68,$8C,$65,$F0,$63,$15,$C6,$B7,$5F,$CC,$FE,$5D
	.dc.b	$F1,$94,$FB,$9C,$B6,$FF,$DF,$13,$A7,$F0,$73,$9B,$FC,$03,$71,$F7
	.dc.b	$B6,$7F,$31,$AF,$1A,$FE,$4D,$F0,$3F,$0D,$FF,$1F,$07,$0F,$F9,$05
	.dc.b	$AF,$F9,$95,$AE,$F7,$FD,$5F,$CB,$7F,$39,$9F,$9D,$8F,$C1,$8C,$4E
	.dc.b	$B0,$FF,$7D,$D6,$9F,$C1,$CF,$1E,$2F,$CE,$6F,$67,$F4,$2D,$C7,$F3
	.dc.b	$CD,$C7,$EE,$BF,$A3,$E6,$FD,$25,$DB,$FD,$4B,$6F,$F4,$CD,$BF,$C2
	.dc.b	$BF,$AB,$F9,$69,$5F,$B0,$F8,$7B,$7F,$AD,$9B,$7F,$8C,$9B,$5F,$E6
	.dc.b	$7F,$95,$76,$FF,$80,$B0,$FF,$46,$46,$27,$FD,$11,$08,$C3,$FE,$8B
	.dc.b	$6B,$F5,$1C,$DF,$F8,$37,$E5,$5F,$2B,$FD,$9E,$73,$7F,$E9,$9A,$56
	.dc.b	$A3,$07,$97,$0F,$F0,$24,$EA,$F9,$46,$2B,$86,$1F,$DA,$FE,$51,$F1
	.dc.b	$9E,$B8,$46,$31,$F5,$B6,$3F,$D4,$8D,$7F,$1A,$E7,$1F,$E2,$9E,$0B
	.dc.b	$F5,$F7,$CC,$FF,$80,$6B,$58,$7F,$99,$67,$2F,$AB,$0F,$F5,$B7,$CB
	.dc.b	$7F,$2C,$C3,$E5,$B6,$3F,$D6,$8D,$7F,$3A,$D7,$FD,$39,$EC,$3E,$21
	.dc.b	$F2,$DF,$52,$30,$D8,$CC,$FF,$AF,$E7,$2F,$6E,$B7,$B3,$E2,$6C,$66
	.dc.b	$30,$7F,$B1,$FE,$D9,$F8,$03,$E0,$3F,$D5,$DC,$3F,$C5,$BC,$2B,$BF
	.dc.b	$EE,$1B,$8B,$0E,$37,$86,$C7,$F4,$69,$DF,$8F,$FC,$FD,$D6,$3D,$E7
	.dc.b	$FA,$5F,$C7,$6B,$C2,$7D,$15,$31,$12,$7F,$A8,$C6,$5E,$24,$3F,$A8
	.dc.b	$39,$7F,$A9,$47,$A1,$7C,$AB,$B7,$F4,$31,$86,$C6,$63,$F4,$AD,$7F
	.dc.b	$DD,$CD,$DB,$FA,$E4,$71,$3F,$08,$53,$FC,$85,$F2,$DF,$49,$FE,$C9
	.dc.b	$E1,$FC,$AF,$35,$FA,$03,$E2,$30,$DA,$FE,$DF,$5F,$14,$CC,$C8,$C9
	.dc.b	$FC,$6A,$3E,$57,$D1,$C8,$C1,$0F,$EE,$8E,$5F,$E6,$31,$F7,$31,$FD
	.dc.b	$DB,$F4,$FC,$29,$3F,$9F,$9C,$B6,$13,$A8,$FF,$00,$47,$33,$2E,$E3
	.dc.b	$53,$FD,$8A,$75,$1B,$F9,$6F,$C1,$12,$BB,$39,$7F,$C2,$2E,$5F,$F4
	.dc.b	$A5,$7F,$12,$8D,$46,$29,$18,$9F,$EE,$3F,$31,$EF,$FD,$B7,$F7,$F7
	.dc.b	$FE,$35,$F8,$FF,$AD,$5F,$E0,$76,$F4,$FF,$1B,$AD,$D7,$81,$B5,$BE
	.dc.b	$AF,$CF,$DF,$0A,$4A,$B7,$7C,$FC,$3F,$11,$2B,$C5,$F0,$FE,$5F,$88
	.dc.b	$E7,$F3,$D2,$7E,$96,$09,$D4,$7A,$67,$4F,$96,$32,$D3,$8B,$E6,$BF
	.dc.b	$85,$F2,$5E,$8F,$98,$95,$DA,$32,$FC,$58,$CB,$F0,$63,$2D,$38,$1F
	.dc.b	$2F,$E0,$BE,$1F,$CB,$C8,$F9,$CF,$D1,$E5,$91,$82,$1F,$35,$F9,$7F
	.dc.b	$1B,$A8,$C1,$F5,$DF,$9F,$94,$8F,$B4,$FD,$2F,$1B,$F3,$9F,$B7,$3E
	.dc.b	$7B,$15,$F7,$E7,$31,$F6,$3E,$66,$4F,$E0,$DF,$B7,$F3,$3F,$C2,$7E
	.dc.b	$07,$FC,$2B,$CB,$F8,$5F,$99,$FB,$E6,$6E,$1F,$8B,$72,$FF,$04,$9E
	.dc.b	$3C,$C8,$91,$87,$FC,$CB,$CE,$AE,$31,$FE,$CB,$9D,$79,$0A,$FF,$13
	.dc.b	$4F,$E6,$78,$FF,$22,$7E,$4B,$DA,$91,$92,$1F,$ED,$19,$E2,$DF,$99
	.dc.b	$FC,$CC,$48,$FD,$24,$7E,$70,$C3,$C6,$30,$FE,$A7,$E1,$BF,$43,$E6
	.dc.b	$C8,$C1,$0F,$03,$97,$C5,$1F,$A8,$31,$AF,$EC,$67,$13,$FC,$47,$8B
	.dc.b	$FD,$0A,$FE,$DF,$E6,$7F,$5E,$7D,$EF,$83,$91,$92,$1F,$C1,$A7,$8B
	.dc.b	$F0,$79,$FF,$AA,$FE,$BB,$F3,$0F,$9B,$9F,$F3,$19,$D1,$FE,$B9,$8D
	.dc.b	$3F,$BE,$9E,$3E,$AF,$2F,$98,$93,$B8,$CB,$C6,$32,$FF,$C0,$1A,$4E
	.dc.b	$A3,$13,$AD,$3F,$D3,$15,$A9,$F1,$7C,$97,$A3,$E5,$E4,$60,$87,$57
	.dc.b	$2F,$78,$F7,$F4,$FC,$0E,$59,$3E,$D8,$CB,$EF,$8C,$BE,$78,$F8,$FF
	.dc.b	$25,$F2,$3E,$4E,$4F,$C5,$8C,$BF,$1E,$32,$FB,$E3,$2F,$C1,$8C,$E5
	.dc.b	$EE,$9F,$A7,$F3,$FF,$53,$E4,$E5,$FD,$18,$CB,$F4,$E3,$2F,$CB,$8F
	.dc.b	$B7,$F1,$BF,$73,$E4,$A4,$FD,$98,$CB,$F6,$E3,$2F,$D3,$8F,$C1,$FC
	.dc.b	$97,$F7,$5F,$80,$78,$34,$CD,$7E,$0E,$70,$FF,$90,$FD,$3F,$F2,$54
	.dc.b	$EA,$3E,$F7,$C9,$7E,$38,$8F,$E2,$FA,$6B,$A3,$F1,$D1,$F9,$48,$FC
	.dc.b	$AF,$C7,$7E,$5B,$E8,$A5,$7E,$4E,$32,$FE,$56,$32,$FE,$36,$3F,$3B
	.dc.b	$F2,$7F,$9E,$F8,$89,$3F,$9B,$8C,$BF,$9D,$8C,$BF,$95,$8C,$B4,$E0
	.dc.b	$7E,$8F,$E4,$F3,$C7,$F9,$69,$BE,$3F,$F1,$DF,$19,$D6,$BF,$D7,$FC
	.dc.b	$97,$F9,$13,$F5,$EF,$F1,$4E,$3F,$EF,$56,$E1,$C6,$D5,$C3,$2F,$F3
	.dc.b	$86,$2B,$DC,$23,$17,$B8,$FE,$E2,$FF,$9B,$BC,$CF,$EE,$70,$F9,$A7
	.dc.b	$6F,$F9,$CB,$CE,$FE,$69,$39,$8C,$3F,$CD,$E7,$6F,$F3,$9F,$87,$FE
	.dc.b	$E5,$AE,$6B,$B3,$67,$57,$D2,$F8,$C2,$D3,$9B,$E3,$3E,$13,$B6,$1E
	.dc.b	$8B,$E3,$7C,$27,$B1,$CE,$F9,$3C,$30,$CC,$F9,$A3,$A1,$B3,$33,$C1
	.dc.b	$F2,$5F,$7C,$3C,$78,$78,$2B,$57,$E1,$BE,$6F,$59,$F2,$DF,$19,$F8
	.dc.b	$27,$63,$C6,$E5,$EF,$86,$5E,$77,$DA,$75,$7D,$70,$FA,$75,$A3,$9D
	.dc.b	$C7,$04,$C3,$C7,$19,$7D,$51,$DD,$CB,$EA,$BE,$F8,$7B,$8F,$15,$71
	.dc.b	$BF,$0D,$F2,$BE,$13,$E1,$F5,$DE,$BF,$DA,$6C,$FD,$33,$FD,$6B,$19
	.dc.b	$3F,$1E,$FF,$80,$FE,$6B,$FC,$09,$3C,$23,$F2,$71,$89,$DC,$23,$0F
	.dc.b	$26,$BC,$0F,$BE,$7F,$4E,$D3,$F0,$EF,$73,$3E,$32,$33,$1D,$4C,$B8
	.dc.b	$9E,$33,$B9,$EA,$6F,$1F,$CE,$BD,$CC,$CF,$0A,$E5,$1D,$0C,$96,$78
	.dc.b	$BC,$F2,$F8,$B7,$DE,$73,$7C,$EB,$8C,$78,$27,$55,$F1,$AB,$53,$C2
	.dc.b	$BB,$4E,$5A,$7C,$92,$9F,$26,$3A,$5F,$A1,$B1,$E0,$9F,$2C,$E5,$B1
	.dc.b	$93,$F8,$5D,$F9,$1E,$D5,$C8,$F2,$16,$7C,$8D,$8D,$C7,$48,$FE,$23
	.dc.b	$B7,$F2,$1F,$96,$FA,$5A,$7F,$29,$F9,$4B,$97,$F3,$1E,$15,$63,$2F
	.dc.b	$E4,$D0,$AC,$61,$CE,$FF,$96,$CE,$5B,$7A,$8C,$5F,$F2,$C3,$9F,$CB
	.dc.b	$64,$D7,$C5,$DB,$E7,$16,$78,$61,$FC,$E6,$B1,$97,$F4,$1A,$ED,$F4
	.dc.b	$78,$C3,$D4,$DC,$3D,$2D,$C3,$FA,$13,$E5,$FA,$0F,$37,$CB,$5B,$F5
	.dc.b	$B6,$FC,$39,$A3,$19,$7F,$51,$7D,$DF,$2D,$2F,$E1,$FC,$B5,$C3,$8C
	.dc.b	$DB,$FE,$B1,$36,$BE,$87,$C5,$7C,$FF,$11,$E3,$3D,$7F,$95,$39,$8C
	.dc.b	$57,$D0,$8C,$4E,$A3,$FC,$19,$3A,$CB,$D3,$F2,$9F,$4F,$E5,$FE,$A3
	.dc.b	$6B,$EC,$B6,$77,$1F,$72,$7E,$D7,$C8,$FD,$BF,$8E,$B3,$F8,$06,$D7
	.dc.b	$C6,$6E,$BF,$DD,$5F,$C1,$7C,$7F,$E0,$FE,$2E,$DF,$E2,$1B,$7F,$39
	.dc.b	$B7,$FD,$F5,$B3,$F9,$56,$CF,$E5,$1F,$C7,$7C,$67,$E3,$FE,$32,$D7
	.dc.b	$E5,$1B,$5F,$69,$B7,$FE,$09,$7F,$2D,$F1,$9F,$97,$F8,$CB,$7F,$9C
	.dc.b	$6D,$FE,$09,$B7,$FE,$11,$7F,$47,$FA,$FB,$C4,$BE,$26,$1B,$87,$F8
	.dc.b	$4A,$74,$D9,$D3,$4D,$4F,$0C,$3F,$43,$5C,$1B,$5E,$0F,$8C,$F0,$FC
	.dc.b	$65,$C3,$B3,$6B,$8B,$72,$FE,$65,$DC,$C5,$3F,$BF,$97,$C0,$FF,$8A
	.dc.b	$B0,$F4,$7C,$66,$63,$18,$F9,$23,$1E,$67,$C2,$6D,$70,$3E,$07,$C6
	.dc.b	$7C,$6F,$8C,$F8,$ED,$8C,$9F,$16,$32,$D3,$37,$F0,$A3,$2D,$8F,$99
	.dc.b	$E6,$7D,$2F,$37,$E9,$B6,$32,$7D,$18,$CB,$4F,$A9,$F1,$9F,$6B,$D8
	.dc.b	$FB,$6D,$8C,$9F,$66,$32,$D3,$FD,$E9,$AF,$F9,$52,$BF,$07,$F1,$9F
	.dc.b	$CA,$71,$4A,$73,$5C,$0F,$F2,$4C,$65,$FC,$35,$71,$8C,$57,$E0,$A3
	.dc.b	$15,$CA,$3F,$96,$1D,$9C,$4F,$DE,$E0,$7C,$67,$E5,$8F,$CB,$FC,$5D
	.dc.b	$C7,$C5,$87,$9B,$F6,$F7,$8D,$F9,$CF,$8D,$FC,$24,$62,$BF,$35,$1F
	.dc.b	$A0,$F1,$64,$60,$FC,$F3,$FA,$4F,$8A,$E3,$E0,$D9,$EF,$AB,$04,$EA
	.dc.b	$7F,$4D,$F1,$B7,$3F,$D0,$BF,$AE,$24,$72,$AE,$2F,$03,$AF,$B7,$FA
	.dc.b	$7C,$3F,$AF,$B6,$3F,$D1,$DB,$BE,$37,$F4,$13,$9C,$7E,$BB,$7B,$3F
	.dc.b	$4D,$F1,$BA,$F0,$3F,$7E,$DC,$7F,$78,$DB,$FD,$BB,$D3,$A6,$4F,$F0
	.dc.b	$1F,$8D,$B5,$FD,$8F,$46,$4B,$FE,$DB,$E3,$2B,$75,$FB,$88,$C6,$1E
	.dc.b	$EA,$DD,$6A,$3F,$D5,$53,$A8,$ED,$7F,$ED,$49,$F8,$BD,$3D,$BD,$9F
	.dc.b	$82,$C8,$FF,$53,$DB,$E3,$71,$87,$F7,$28,$C4,$F2,$37,$3A,$CB,$CC
	.dc.b	$FC,$CE,$69,$3D,$3E,$32,$E1,$E4,$9F,$7F,$C6,$E3,$0F,$73,$6F,$D6
	.dc.b	$DA,$FF,$5C,$62,$8F,$B3,$BF,$C8,$F8,$DB,$7E,$A6,$DF,$A5,$B8,$7C
	.dc.b	$B6,$CF,$C8,$6C,$F6,$7F,$75,$D6,$F3,$7E,$AF,$BB,$86,$DF,$EA,$AD
	.dc.b	$5B,$8F,$97,$5B,$69,$F3,$BE,$33,$FC,$01,$F1,$4F,$8D,$FC,$0C,$61
	.dc.b	$B3,$FB,$6F,$59,$5B,$8E,$46,$5C,$63,$FC,$13,$E6,$7F,$1B,$3E,$56
	.dc.b	$C7,$F8,$CE,$9F,$4A,$F3,$BF,$8C,$56,$B7,$FF,$3C,$5A,$38,$BF,$E9
	.dc.b	$26,$BF,$EE,$26,$BF,$99,$6F,$63,$8F,$CB,$5F,$9B,$F8,$EF,$CF,$EB
	.dc.b	$FA,$06,$E9,$F9,$26,$E3,$F9,$16,$D7,$19,$FC,$BD,$71,$7F,$8D,$FC
	.dc.b	$B7,$AF,$97,$F5,$7F,$21,$FE,$5B,$8C,$4F,$E7,$A3,$1A,$FE,$2E,$30
	.dc.b	$72,$71,$97,$E8,$67,$F9,$E7,$CC,$3E,$09,$DB,$FE,$AE,$A4,$E6,$30
	.dc.b	$D8,$FF,$59,$B5,$FF,$5D,$FC,$43,$CE,$7C,$C7,$38,$C3,$63,$33,$FE
	.dc.b	$BC,$9C,$BD,$3E,$17,$FC,$24,$DF,$17,$C3,$18,$8F,$F6,$4E,$08,$C4
	.dc.b	$7F,$0F,$D9,$F3,$1F,$C9,$63,$FA,$2F,$2B,$E6,$3F,$45,$18,$6B,$FE
	.dc.b	$7D,$AE,$3E,$47,$F5,$CE,$2E,$98,$73,$AF,$F3,$DD,$F0,$E5,$FF,$09
	.dc.b	$B6,$B7,$7C,$69,$8F,$87,$DE,$C3,$E3,$F3,$6E,$1F,$D2,$AB,$C9,$58
	.dc.b	$33,$32,$32,$43,$C6,$E5,$F2,$C7,$85,$7C,$C3,$E0,$F1,$F3,$1F,$A7
	.dc.b	$9F,$E0,$0D,$7F,$B9,$CE,$DF,$E0,$8D,$7F,$DE,$6D,$7F,$DF,$6D,$7F
	.dc.b	$A0,$1E,$09,$FC,$66,$3F,$47,$B3,$E2,$77,$BE,$61,$E7,$61,$B8,$7D
	.dc.b	$5A,$F8,$A6,$66,$46,$4F,$E3,$1E,$2F,$ED,$5F,$B9,$F0,$5F,$30,$F8
	.dc.b	$51,$98,$FF,$07,$B7,$8B,$EC,$CE,$E7,$9D,$FD,$99,$7C,$16,$F8,$FF
	.dc.b	$AE,$69,$F8,$B8,$FB,$26,$CD,$5F,$F8,$A1,$B1,$B8,$FC,$84,$6E,$E4
	.dc.b	$60,$87,$F6,$D8,$E7,$7E,$2F,$82,$FC,$39,$27,$FC,$27,$28,$F1,$19
	.dc.b	$8F,$CB,$B8,$76,$FE,$FE,$7A,$47,$D8,$3F,$C8,$73,$CD,$C1,$EB,$9E
	.dc.b	$1F,$31,$FE,$70,$8F,$D6,$FA,$DF,$B5,$BF,$F3,$AB,$72,$FF,$26,$D6
	.dc.b	$EB,$FC,$10,$DA,$DF,$5F,$E9,$2F,$85,$7F,$94,$2B,$57,$CB,$E0,$7E
	.dc.b	$A4,$62,$75,$1C,$7E,$0B,$3E,$E7,$3F,$16,$57,$F8,$E2,$FF,$CB,$4D
	.dc.b	$9C,$BF,$EF,$6C,$7F,$45,$9B,$2F,$4D,$74,$AE,$04,$38,$65,$E4,$F4
	.dc.b	$7D,$5F,$01,$27,$C1,$19,$7E,$3C,$65,$F8,$71,$96,$9C,$0F,$9B,$F0
	.dc.b	$1F,$17,$D8,$91,$F4,$1F,$87,$CD,$23,$04,$3E,$7B,$F3,$7E,$03,$51
	.dc.b	$83,$EC,$BF,$47,$29,$1F,$71,$FA,$9E,$37,$E8,$FE,$09,$EA,$E2,$BF
	.dc.b	$03,$39,$8F,$B5,$EC,$C9,$FC,$2B,$F7,$7E,$07,$F8,$A7,$AD,$F8,$FE
	.dc.b	$2F,$C3,$FC,$0F,$E0,$4C,$DC,$3F,$1A,$E5,$FF,$0B,$4F,$1E,$3F,$E1
	.dc.b	$ED,$8F,$E5,$FC,$9F,$E8,$87,$89,$C0,$FC,$77,$77,$E3,$F8,$E4,$7E
	.dc.b	$6A,$32,$FE,$72,$32,$FE,$16,$3F,$2D,$F0,$1F,$97,$A9,$3F,$A2,$8C
	.dc.b	$BF,$A4,$8C,$BF,$8D,$8F,$CF,$7C,$07,$E7,$FC,$99,$3F,$AA,$8C,$BF
	.dc.b	$AC,$8C,$BF,$95,$8C,$BF,$97,$8C,$E9,$FB,$09,$FA,$DF,$2B,$FA,$BF
	.dc.b	$72,$5F,$FA,$63,$8D,$3A,$8F,$DC,$CE,$9F,$F3,$E4,$E9,$FD,$EF,$37
	.dc.b	$F5,$8F,$91,$F9,$DE,$19,$BA,$FF,$38,$AF,$D5,$56,$D8,$72,$3F,$B9
	.dc.b	$CF,$F8,$41,$BE,$FF,$53,$B7,$A9,$23,$24,$3F,$DA,$7C,$7F,$CC,$AB
	.dc.b	$94,$60,$FE,$DE,$E3,$6F,$DD,$46,$2F,$B3,$E2,$BF,$67,$BF,$ED,$F6
	.dc.b	$24,$F9,$A3,$2F,$A2,$32,$F7,$D3,$E1,$4F,$C5,$E9,$F8,$DC,$D2,$BD
	.dc.b	$F1,$97,$E0,$C6,$5E,$59,$FE,$B1,$F8,$BE,$F7,$CF,$F6,$A4,$60,$87
	.dc.b	$07,$2F,$28,$FA,$C6,$35,$F8,$39,$7D,$7A,$CD,$7D,$2F,$6F,$EE,$7B
	.dc.b	$72,$30,$43,$CA,$E5,$F3,$C7,$E0,$FE,$63,$F0,$9E,$FC,$9F,$BB,$19
	.dc.b	$7E,$FC,$65,$F7,$C7,$E2,$FE,$63,$FC,$07,$F1,$4F,$06,$99,$AF,$C3
	.dc.b	$CE,$1F,$F4,$54,$ED,$C9,$F8,$CF,$98,$FC,$B1,$1F,$C7,$F5,$D7,$E7
	.dc.b	$C5,$DB,$E2,$F0,$7E,$7E,$7E,$9F,$CB,$FE,$7F,$E6,$24,$7E,$62,$3F
	.dc.b	$44,$61,$FB,$31,$87,$F3,$BF,$31,$FA,$7F,$87,$91,$82,$1F,$80,$72
	.dc.b	$FE,$0A,$3F,$5C,$47,$78,$79,$BE,$63,$11,$B3,$3D,$9F,$A3,$8F,$DB
	.dc.b	$7C,$67,$F9,$3B,$E4,$BF,$D4,$37,$F8,$26,$D7,$EA,$67,$77,$FE,$66
	.dc.b	$9C,$BF,$E5,$86,$F3,$FE,$FE,$7F,$C9,$9F,$0A,$F3,$18,$FF,$BF,$FE
	.dc.b	$5D,$D9,$FE,$42,$A7,$9D,$FE,$44,$9C,$E6,$13,$A7,$95,$71,$C3,$AC
	.dc.b	$2B,$8B,$C2,$33,$18,$AE,$53,$BA,$F1,$5F,$85,$E9,$87,$69,$F1,$57
	.dc.b	$AE,$31,$1E,$4A,$E0,$E9,$AF,$86,$37,$3C,$63,$C0,$6C,$CC,$F0,$8D
	.dc.b	$5F,$9F,$0F,$4E,$1E,$6B,$D9,$D2,$32,$F9,$9A,$F9,$4D,$9E,$03,$C8
	.dc.b	$5A,$EF,$3D,$63,$17,$E3,$AE,$75,$A9,$F4,$1A,$3C,$58,$47,$04,$C3
	.dc.b	$D5,$19,$7E,$95,$FB,$CE,$D1,$9C,$3C,$53,$AA,$C1,$ED,$9F,$5C,$66
	.dc.b	$3D,$58,$79,$8F,$C1,$FA,$CF,$5B,$ED,$CA,$30,$F3,$31,$1A,$8C,$C7
	.dc.b	$FA,$5B,$E6,$3F,$25,$18,$8F,$CA,$46,$27,$FC,$C2,$D8,$FF,$33,$35
	.dc.b	$E4,$7D,$13,$FD,$D7,$BF,$E2,$9E,$66,$7C,$4F,$EA,$78,$FE,$6A,$32
	.dc.b	$D8,$CE,$7E,$09,$E1,$5C,$D5,$F5,$AD,$C7,$90,$D9,$E3,$36,$62,$FC
	.dc.b	$0E,$1F,$05,$75,$AE,$8D,$9E,$17,$E3,$AD,$C7,$5A,$DB,$4D,$9E,$38
	.dc.b	$D9,$E0,$8D,$B6,$36,$7C,$A7,$73,$DC,$FF,$4A,$5F,$BA,$75,$18,$BF
	.dc.b	$35,$70,$9F,$04,$E5,$F2,$46,$5E,$27,$FB,$5A,$BA,$3F,$4A,$78,$4F
	.dc.b	$F1,$6D,$DC,$0F,$97,$FF,$05,$ED,$FC,$AF,$E6,$2E,$7F,$EC,$4A,$D3
	.dc.b	$FC,$AE,$33,$97,$F2,$BB,$B8,$74,$9B,$7C,$2B,$FC,$18,$A3,$3F,$17
	.dc.b	$30,$FE,$6D,$F1,$36,$F7,$37,$2F,$E7,$D3,$6B,$C9,$3D,$BE,$63,$17
	.dc.b	$FC,$F5,$B9,$7B,$9B,$91,$F2,$FC,$C7,$9B,$E6,$2E,$3F,$D1,$DB,$87
	.dc.b	$A1,$B9,$FF,$4F,$7D,$BF,$31,$2B,$FA,$77,$CA,$DB,$F7,$CD,$BF,$EA
	.dc.b	$D3,$6B,$E6,$7C,$3F,$CA,$F8,$77,$88,$F5,$FE,$4C,$E6,$31,$5F,$D6
	.dc.b	$E3,$13,$A8,$FF,$04,$BF,$43,$E6,$3E,$8F,$CC,$59,$FE,$C4,$DD,$3E
	.dc.b	$9B,$73,$FE,$D4,$FD,$8F,$95,$FB,$3F,$1B,$6F,$FB,$53,$6F,$ED,$B6
	.dc.b	$FF,$B9,$3F,$80,$F8,$EF,$C0,$FC,$55,$BF,$EE,$4D,$BF,$C1,$B6,$FF
	.dc.b	$BD,$36,$7F,$20,$D9,$FE,$F0,$FE,$33,$E2,$FF,$1B,$F1,$76,$BF,$BF
	.dc.b	$36,$BF,$1E,$DB,$FF,$03,$3F,$94,$F8,$BF,$CA,$FC,$C5,$BF,$F0,$33
	.dc.b	$6F,$F2,$ED,$BF,$F0,$73,$F9,$EF,$97,$FE,$69,$5F,$DF,$E7,$FC,$A7
	.dc.b	$1F,$CE,$63,$07,$F8,$0F,$A9,$1C,$2F,$FC,$DC,$ED,$BE,$27,$5F,$13
	.dc.b	$B3,$6F,$9C,$AF,$9C,$E6,$38,$CE,$A7,$BD,$79,$FE,$37,$D1,$F1,$96
	.dc.b	$B5,$8A,$31,$87,$91,$B9,$7B,$1F,$DF,$AF,$81,$79,$7F,$0F,$E3,$33
	.dc.b	$18,$C7,$AC,$62,$75,$1F,$EB,$76,$E1,$F0,$BE,$33,$E6,$7C,$67,$CD
	.dc.b	$6D,$79,$9B,$3B,$8F,$4C,$ED,$A6,$CB,$1F,$3B,$CD,$FB,$1E,$6F,$D9
	.dc.b	$6C,$FC,$26,$CE,$E3,$E2,$CF,$DF,$F8,$CF,$C0,$7B,$36,$7E,$4B,$6B
	.dc.b	$ED,$B7,$5F,$C1,$BF,$C9,$A3,$FC,$97,$7F,$89,$F8,$CC,$57,$F9,$3E
	.dc.b	$FE,$9C,$62,$75,$1F,$80,$AE,$2E,$8C,$57,$18,$FC,$5D,$71,$9D,$47
	.dc.b	$E3,$BE,$33,$F3,$07,$E6,$7E,$2E,$CF,$E3,$2D,$F3,$0F,$1B,$F3,$DF
	.dc.b	$1B,$F8,$08,$C5,$7E,$22,$3F,$45,$E2,$48,$C1,$F8,$C7,$F4,$DF,$15
	.dc.b	$C7,$C1,$B3,$DE,$D1,$98,$FC,$9F,$C6,$5D,$7F,$4A,$FE,$BC,$91,$C6
	.dc.b	$6C,$FF,$B9,$73,$7B,$5F,$AD,$7F,$AD,$B6,$BF,$65,$D4,$F8,$DF,$CD
	.dc.b	$CE,$6F,$FC,$FE,$DE,$EF,$DB,$FC,$6E,$BC,$0F,$E0,$0D,$C7,$F5,$2D
	.dc.b	$BF,$DD,$BD,$3A,$64,$FE,$BB,$E3,$6D,$79,$F1,$57,$F6,$8F,$1B,$FD
	.dc.b	$11,$7F,$4E,$72,$FE,$EE,$33,$5C,$6B,$83,$D3,$5F,$F7,$1C,$E9,$F8
	.dc.b	$3D,$3E,$E3,$B7,$B3,$70,$ED,$1E,$1F,$8D,$C6,$1F,$F0,$07,$04,$63
	.dc.b	$0E,$AD,$BF,$E3,$93,$A7,$D1,$CC,$7F,$D4,$57,$2F,$EA,$CE,$63,$27
	.dc.b	$D0,$70,$EF,$34,$62,$3F,$97,$46,$2B,$E2,$7C,$6F,$DA,$AF,$94,$D8
	.dc.b	$C9,$E4,$2D,$7D,$1D,$BE,$87,$C6,$7E,$09,$B3,$F4,$5B,$3B,$8F,$A5
	.dc.b	$3B,$69,$B2,$C7,$FB,$21,$F8,$3D,$4F,$33,$EF,$7B,$9F,$C0,$E3,$13
	.dc.b	$F7,$63,$13,$A8,$FE,$01,$3A,$AF,$A9,$5F,$87,$F8,$CF,$F0,$47,$C4
	.dc.b	$BE,$33,$18,$7F,$46,$F5,$53,$98,$FC,$55,$70,$AD,$E1,$E7,$17,$C2
	.dc.b	$3F,$8C,$7C,$BF,$EB,$2F,$EB,$D7,$FB,$76,$30,$E1,$AF,$DC,$6B,$F8
	.dc.b	$27,$7C,$3F,$E4,$EF,$9A,$7A,$3D,$B8,$FF,$A6,$C6,$30,$FD,$06,$3F
	.dc.b	$A5,$F8,$DF,$D3,$36,$32,$7E,$96,$32,$D3,$39,$7E,$6E,$32,$D8,$FE
	.dc.b	$D7,$F2,$CF,$41,$CB,$FB,$3F,$8F,$FF,$2E,$CE,$63,$15,$C2,$B6,$FE
	.dc.b	$D6,$B7,$18,$BF,$D5,$47,$F4,$1F,$97,$7C,$0B,$B7,$E5,$D2,$73,$18
	.dc.b	$6C,$7D,$06,$BF,$A8,$F8,$77,$9C,$F9,$8E,$71,$86,$C6,$67,$EC,$CE
	.dc.b	$5E,$9F,$09,$FE,$CD,$6F,$89,$E1,$8C,$47,$ED,$B0,$46,$20,$7E,$23
	.dc.b	$F6,$27,$C0,$FE,$FF,$7C,$F4,$A7,$35,$F9,$56,$BF,$91,$E2,$7C,$CF
	.dc.b	$F9,$86,$35,$87,$37,$B3,$6F,$A3,$9A,$E3,$E7,$E7,$0F,$EB,$57,$DF
	.dc.b	$D0,$C1,$CF,$CA,$78,$CF,$93,$CD,$6C,$87,$F2,$37,$2F,$E9,$F3,$F6
	.dc.b	$5E,$3B,$BF,$98,$D7,$9D,$7E,$E1,$FB,$5B,$FE,$DC,$D7,$FA,$C3,$63
	.dc.b	$F9,$25,$BF,$57,$F2,$D7,$0E,$DE,$97,$BE,$1F,$F2,$83,$67,$F7,$35
	.dc.b	$C1,$FD,$E0,$F9,$97,$C2,$FE,$42,$30,$D8,$CC,$7E,$51,$AF,$F0,$06
	.dc.b	$E5,$FC,$D9,$B8,$7F,$3E,$6C,$FE,$13,$E6,$BF,$13,$5F,$CF,$3F,$1A
	.dc.b	$79,$7F,$6E,$91,$87,$FC,$3C,$D7,$D3,$F6,$CF,$84,$E1,$39,$BE,$97
	.dc.b	$83,$A5,$6E,$FE,$7F,$07,$DF,$FD,$15,$E2,$D3,$E5,$C4,$9D,$90,$F9
	.dc.b	$3F,$AA,$FF,$83,$1C,$63,$FE,$E6,$E1,$FC,$44,$63,$7F,$EB,$F5,$B8
	.dc.b	$FB,$BF,$35,$F9,$5E,$37,$BF,$4F,$B7,$12,$7F,$0A,$D7,$F4,$92,$F8
	.dc.b	$0C,$65,$E1,$AB,$8F,$EC,$BB,$3F,$A1,$57,$E1,$BE,$57,$F4,$9F,$91
	.dc.b	$78,$F0,$FE,$E1,$48,$C5,$FE,$2A,$B8,$65,$F9,$66,$BF,$B5,$97,$CC
	.dc.b	$E3,$0F,$85,$56,$FC,$55,$AB,$FF,$26,$7E,$E3,$FC,$21,$FF,$4A,$71
	.dc.b	$3C,$BF,$A8,$7E,$9E,$75,$3F,$E3,$B6,$E3,$FE,$07,$0E,$30,$FF,$58
	.dc.b	$36,$59,$BF,$1E,$67,$D7,$C5,$95,$E4,$7F,$84,$4E,$A3,$CF,$D9,$F8
	.dc.b	$7E,$59,$18,$21,$E5,$8F,$41,$F8,$DF,$52,$C6,$B4,$F4,$67,$FA,$4F
	.dc.b	$09,$BF,$F7,$4A,$F8,$7F,$3D,$F9,$52,$3F,$23,$C3,$F8,$84,$70,$7C
	.dc.b	$CF,$4F,$E4,$11,$C1,$F3,$BE,$7B,$F4,$1C,$F2,$3E,$8B,$FA,$3E,$6F
	.dc.b	$E2,$DF,$B9,$3E,$3A,$92,$BF,$B2,$FB,$1F,$E3,$27,$EC,$FC,$F7,$F2
	.dc.b	$F8,$C4,$7F,$39,$6F,$07,$F8,$F7,$7F,$7D,$7E,$FF,$EF,$3F,$D9,$D7
	.dc.b	$FE,$9B,$7F,$B8,$37,$D6,$D7,$C8,$48,$D3,$5F,$E3,$72,$F8,$FC,$63
	.dc.b	$FE,$48,$8E,$B3,$AA,$F0,$57,$F0,$1F,$9F,$93,$F9,$27,$F8,$27,$B9
	.dc.b	$88,$FF,$11,$67,$F5,$A7,$18,$FA,$F9,$7F,$4F,$48,$6F,$FD,$CA,$3F
	.dc.b	$D5,$4D,$7F,$D2,$8D,$8E,$5D,$9F,$43,$D7,$95,$C1,$FE,$2D,$1F,$53
	.dc.b	$E3,$BF,$8C,$7C,$84,$AE,$B1,$96,$90,$F4,$FE,$93,$FE,$6A,$C3,$F5
	.dc.b	$18,$AB,$85,$79,$1F,$EC,$55,$BF,$0B,$FB,$DE,$A8,$7C,$F6,$89,$7F
	.dc.b	$E0,$0C,$7E,$F4,$BE,$7B,$33,$E2,$AB,$5A,$9F,$C7,$FC,$8F,$83,$E0
	.dc.b	$25,$FC,$88,$CB,$48,$7B,$3C,$5F,$C5,$7C,$5F,$E2,$FE,$4E,$4F,$E2
	.dc.b	$E3,$2D,$21,$FE,$0D,$FC,$07,$F1,$0E,$32,$FC,$44,$F0,$8D,$73,$FE
	.dc.b	$93,$E2,$A4,$7D,$C7,$FC,$11,$E1,$FE,$E6,$3E,$EC,$69,$A7,$FA,$DF
	.dc.b	$C6,$DF,$8B,$23,$FD,$74,$F2,$F1,$31,$1F,$EA,$F8,$B9,$7F,$AE,$A7
	.dc.b	$FB,$37,$A1,$FC,$56,$BF,$D6,$9E,$2F,$88,$CF,$54,$8F,$A9,$CF,$FD
	.dc.b	$4A,$31,$1F,$EB,$BF,$16,$E3,$FE,$AE,$C3,$A3,$6B,$DB,$3E,$EF,$0E
	.dc.b	$5F,$FB,$45,$F8,$3D,$F8,$8F,$F6,$34,$5B,$FF,$68,$CF,$C6,$F0,$25
	.dc.b	$7F,$B6,$5F,$93,$86,$23,$FD,$9D,$16,$BF,$DB,$33,$FE,$29,$DD,$E2
	.dc.b	$52,$5F,$FB,$8F,$BE,$BF,$F0,$06,$76,$FC,$23,$F6,$CE,$1F,$F6,$E4
	.dc.b	$7F,$90,$7F,$65,$FD,$04,$3B,$BF,$C8,$73,$F7,$8D,$6F,$27,$C0,$43
	.dc.b	$F8,$3D,$72,$C9,$87,$F8,$12,$7C,$9E,$57,$AC,$F0,$77,$58,$F3,$57
	.dc.b	$09,$C4,$74,$9F,$F2,$F7,$CE,$3E,$3E,$92,$7D,$DA,$3E,$8D,$CF,$99
	.dc.b	$FF,$2B,$5F,$17,$FD,$88,$D7,$F8,$FC,$F0,$73,$7C,$72,$FF,$29,$D6
	.dc.b	$DF,$E7,$95,$BA,$EB,$5B,$D7,$FC,$FF,$7B,$7F,$85,$D6,$EF,$9F,$CB
	.dc.b	$FB,$09,$87,$AB,$6F,$E1,$4F,$9A,$32,$FA,$9F,$B1,$E1,$7F,$92,$7E
	.dc.b	$1F,$FC,$57,$7F,$E9,$2F,$A2,$7C,$77,$FA,$53,$05,$7F,$8A,$5D,$46
	.dc.b	$63,$10,$9E,$2F,$FA,$72,$3A,$B9,$9E,$1E,$C7,$F4,$18,$C1,$BF,$5F
	.dc.b	$C1,$7C,$A3,$C3,$18,$AE,$38,$78,$A7,$B5,$F3,$BE,$18,$79,$1B,$87
	.dc.b	$1A,$DD,$F1,$7D,$B7,$C7,$0F,$65,$6E,$7C,$26,$F0,$F5,$E1,$DA,$F8
	.dc.b	$4F,$A4,$D5,$71,$AD,$DF,$A2,$7D,$E7,$42,$C7,$9E,$35,$7D,$70,$E7
	.dc.b	$50,$DE,$1E,$07,$D3,$19,$9F,$7C,$79,$E7,$57,$E2,$8F,$31,$E0,$C3
	.dc.b	$CD,$3A,$36,$FA,$4C,$DF,$33,$B6,$1F,$1A,$B0,$79,$23,$2F,$6C,$3D
	.dc.b	$4E,$47,$C4,$FF,$89,$DF,$A6,$63,$D1,$F8,$D1,$98,$F6,$99,$3F,$28
	.dc.b	$3E,$8F,$FD,$D3,$1B,$8F,$F7,$5C,$62,$77,$1B,$39,$1F,$9D,$8D,$3D
	.dc.b	$1F,$C2,$2F,$8C,$70,$E2,$39,$61,$F9,$57,$51,$9A,$E3,$5C,$27,$EF
	.dc.b	$9E,$42,$B9,$AE,$51,$D3,$2F,$6C,$E5,$F4,$D7,$4B,$ED,$5C,$9B,$7E
	.dc.b	$3B,$E1,$3E,$B3,$47,$88,$EC,$7E,$F5,$DB,$F0,$CD,$C7,$92,$BC,$11
	.dc.b	$C5,$B1,$98,$DD,$F9,$9A,$78,$63,$C9,$7E,$69,$F6,$B6,$7C,$4D,$47
	.dc.b	$CC,$3C,$07,$43,$CD,$3D,$1C,$B4,$F2,$16,$7C,$8D,$87,$C2,$B8,$DF
	.dc.b	$21,$FE,$A5,$C3,$F9,$8B,$6B,$FA,$E7,$1B,$CD,$C6,$1C,$6A,$E1,$CB
	.dc.b	$0F,$E8,$B1,$C7,$A1,$87,$19,$C9,$FD,$2B,$E2,$78,$9B,$FA,$37,$B5
	.dc.b	$89,$D1,$34,$FF,$36,$CE,$A3,$1C,$7F,$D3,$AF,$85,$6A,$EC,$E6,$B6
	.dc.b	$7F,$95,$AF,$95,$7F,$5E,$BF,$81,$5F,$DC,$BA,$EF,$77,$F6,$4C,$3F
	.dc.b	$A7,$79,$F8,$99,$1F,$DC,$A7,$58,$7F,$7C,$AC,$BF,$DF,$BE,$8F,$E2
	.dc.b	$68,$FA,$3B,$18,$6E,$1F,$D6,$DB,$5F,$E0,$67,$C7,$E3,$C9,$D1,$E1
	.dc.b	$71,$3F,$E1,$B7,$A4,$64,$FE,$CD,$F4,$7E,$23,$C7,$F1,$D6,$32,$4A
	.dc.b	$FF,$0A,$C6,$5F,$F0,$BF,$C6,$FA,$8F,$5F,$0D,$AF,$EE,$51,$67,$FC
	.dc.b	$3B,$37,$7F,$E0,$90,$FD,$5D,$8F,$E3,$E3,$0F,$C6,$FA,$4C,$4E,$B1
	.dc.b	$FF,$1A,$56,$9B,$19,$8F,$F1,$5C,$F0,$A9,$5F,$E0,$5F,$A4,$B5,$FE
	.dc.b	$30,$7E,$9F,$CC,$49,$FF,$05,$7D,$25,$AF,$F1,$AC,$DC,$BF,$DA,$9C
	.dc.b	$2A,$C5,$FF,$A6,$FF,$73,$71,$FE,$29,$B1,$FE,$15,$FA,$0C,$4C,$8F
	.dc.b	$F1,$CC,$EB,$2F,$F2,$1D,$65,$FF,$63,$4F,$F8,$4F,$B1,$F3,$BF,$1D
	.dc.b	$B1,$FE,$57,$F8,$97,$E4,$8E,$8D,$5C,$AF,$93,$7C,$30,$EA,$30,$E3
	.dc.b	$7B,$8C,$9F,$C2,$2F,$C1,$F4,$F9,$DF,$ED,$E0,$8F,$81,$D3,$32,$FF
	.dc.b	$26,$FC,$0F,$8D,$F7,$C6,$71,$FF,$67,$B7,$0F,$F6,$1B,$FC,$26,$BE
	.dc.b	$E4,$7F,$B1,$B9,$31,$D3,$96,$61,$D9,$CB,$4F,$9B,$CD,$F4,$CE,$9D
	.dc.b	$36,$32,$4C,$BF,$DC,$31,$97,$E2,$E0,$E2,$5F,$97,$A7,$53,$FE,$DA
	.dc.b	$7F,$01,$CB,$31,$F8,$35,$70,$E7,$87,$E2,$67,$F0,$35,$EF,$AD,$61
	.dc.b	$FE,$E4,$E8,$F8,$47,$C3,$E8,$B7,$F3,$22,$E3,$FE,$F8,$9B,$5F,$53
	.dc.b	$E4,$BF,$0C,$7E,$1F,$E7,$2E,$1F,$82,$8B,$7F,$7B,$8D,$58,$BE,$57
	.dc.b	$F8,$8F,$1B,$0F,$E4,$DF,$AD,$A7,$F4,$7E,$77,$E5,$7F,$CB,$2F,$FD
	.dc.b	$FB,$39,$8C,$5F,$FC,$01,$FD,$D1,$E5,$FF,$30,$6C,$7E,$6E,$77,$A3
	.dc.b	$05,$6A,$3F,$33,$E2,$B9,$DD,$BF,$89,$3F,$15,$F0,$17,$2F,$F5,$3E
	.dc.b	$0F,$84,$F0,$1F,$8F,$FA,$9B,$3A,$F4,$3C,$86,$33,$FC,$1D,$A3,$F2
	.dc.b	$DF,$1D,$2B,$81,$FA,$07,$D9,$E5,$C8,$FD,$1F,$CC,$5C,$39,$CD,$DF
	.dc.b	$FC,$51,$BF,$AE,$B1,$7F,$1E,$BF,$49,$F5,$38,$C3,$FD,$A1,$C0,$8F
	.dc.b	$95,$E5,$4B,$FE,$09,$F0,$EF,$80,$C6,$1E,$AA,$BA,$7F,$8D,$F9,$D7
	.dc.b	$BC,$3E,$3D,$69,$FF,$0A,$EA,$FD,$3B,$FC,$B7,$1A,$3F,$D9,$6E,$31
	.dc.b	$FD,$4F,$EE,$BF,$8C,$9F,$F6,$BB,$63,$F8,$8E,$3F,$2F,$E5,$FF,$B8
	.dc.b	$CE,$73,$FF,$25,$07,$17,$FE,$26,$8C,$76,$7E,$47,$8D,$7C,$A2,$57
	.dc.b	$F2,$59,$7C,$16,$32,$F0,$C6,$27,$CB,$F2,$DE,$23,$F3,$9F,$9A,$3E
	.dc.b	$6B,$13,$A2,$5F,$AA,$75,$18,$CF,$D3,$A3,$0E,$7A,$72,$C3,$83,$F0
	.dc.b	$3E,$32,$61,$FD,$73,$CB,$B5,$F2,$A6,$E3,$FD,$AB,$D3,$56,$2F,$E2
	.dc.b	$64,$BF,$F1,$97,$0F,$E9,$F0,$F1,$C6,$3C,$1F,$9F,$F1,$73,$0F,$E8
	.dc.b	$BA,$DD,$3E,$CC,$DC,$3F,$9F,$CD,$AF,$F7,$37,$D2,$FE,$1E,$FF,$A3
	.dc.b	$B6,$BF,$D0,$5E,$33,$EC,$1F,$31,$FD,$FE,$B5,$87,$2E,$BF,$F0,$2E
	.dc.b	$36,$3F,$D3,$2A,$7B,$3F,$EA,$26,$BF,$02,$7F,$12,$7F,$95,$BC,$AF
	.dc.b	$DB,$57,$E4,$23,$2D,$3F,$14,$7F,$9C,$31,$4F,$0A,$FC,$CD,$7F,$98
	.dc.b	$3D,$87,$C6,$D3,$FC,$77,$E3,$DC,$79,$DF,$F9,$BA,$B7,$18,$C5,$18
	.dc.b	$9F,$E3,$3E,$CA,$F9,$47,$F7,$4A,$DE,$18,$DB,$FC,$23,$57,$1F,$46
	.dc.b	$3D,$1E,$CD,$C3,$F5,$BA,$FE,$75,$E9,$19,$BF,$AA,$DA,$E6,$FF,$3B
	.dc.b	$AE,$13,$A8,$FF,$1B,$7D,$27,$FA,$0F,$E0,$BE,$AB,$70,$FF,$67,$B5
	.dc.b	$FF,$6B,$B5,$FF,$4E,$8F,$85,$7C,$B3,$E9,$7E,$FC,$61,$B1,$99,$FF
	.dc.b	$6D,$CE,$5E,$FE,$F7,$FB,$B1,$BF,$0D,$E4,$8C,$47,$89,$AF,$DF,$FA
	.dc.b	$4F,$D4,$BF,$55,$AF,$D1,$6E,$9F,$C9,$7E,$95,$F0,$7E,$A8,$FC,$EC
	.dc.b	$EB,$83,$A7,$7E,$5B,$87,$F8,$D7,$0E,$2F,$7F,$73,$F3,$BC,$AC,$11
	.dc.b	$86,$CF,$FB,$EA,$75,$39,$3B,$47,$FA,$F2,$76,$FA,$63,$3E,$4F,$E3
	.dc.b	$7E,$3F,$77,$CA,$BC,$44,$3F,$B5,$F3,$27,$AE,$8F,$A0,$FE,$5B,$5F
	.dc.b	$9E,$8C,$78,$BA,$9C,$CF,$E6,$7E,$97,$F3,$8E,$E3,$35,$FC,$F1,$BB
	.dc.b	$7F,$3B,$6C,$FF,$B5,$AB,$8B,$FE,$52,$FA,$5F,$F6,$2B,$FE,$82,$FA
	.dc.b	$5F,$E1,$6D,$EB,$F8,$6A,$76,$F1,$F8,$1F,$E1,$6E,$5B,$1A,$8F,$CA
	.dc.b	$99,$3F,$88,$FD,$2B,$DF,$E2,$D8,$CC,$60,$8F,$FA,$5A,$34,$FF,$9A
	.dc.b	$9B,$A7,$FA,$8B,$E9,$5E,$C7,$C9,$E0,$FA,$FF,$41,$FD,$3F,$0F,$E1
	.dc.b	$7C,$88,$C4,$7F,$95,$38,$3E,$B3,$F3,$6B,$F9,$3E,$3F,$5D,$B3,$A9
	.dc.b	$F6,$CE,$5F,$79,$FE,$2D,$8D,$47,$D5,$9F,$24,$FF,$AC,$3E,$6B,$FC
	.dc.b	$14,$FD,$BF,$2B,$F2,$3E,$AF,$F7,$0F,$92,$BC,$C7,$F0,$93,$2B,$BE
	.dc.b	$3F,$8B,$8F,$29,$23,$39,$23,$84,$7B,$DE,$BE,$A7,$E3,$FE,$67,$C3
	.dc.b	$F4,$BE,$4A,$FF,$05,$B7,$0F,$C7,$4E,$E7,$AB,$F1,$08,$E0,$FC,$B3
	.dc.b	$B7,$9F,$93,$FA,$2F,$A4,$FF,$5E,$C7,$D4,$E0,$FE,$F8,$DF,$43,$5E
	.dc.b	$24,$8F,$7D,$70,$7E,$8B,$5F,$E1,$4D,$7F,$C3,$8E,$F1,$FF,$10,$B7
	.dc.b	$B7,$35,$F4,$3E,$95,$9F,$97,$AF,$F1,$3C,$B3,$2F,$DB,$63,$FA,$2C
	.dc.b	$B9,$D7,$E1,$A3,$F4,$13,$E0,$8C,$11,$E2,$F5,$7E,$07,$A7,$F9,$6F
	.dc.b	$91,$91,$F0,$DF,$8D,$E0,$C8,$F8,$2F,$72,$3F,$13,$E0,$7C,$44,$70
	.dc.b	$7E,$53,$AB,$F4,$1C,$BF,$A4,$24,$7D,$1E,$3F,$E2,$BF,$47,$FE,$8A
	.dc.b	$8F,$C6,$B5,$FC,$FB,$5F,$E7,$0D,$7F,$CB,$7C,$9F,$DD,$1F,$F7,$57
	.dc.b	$B8,$FA,$7F,$CE,$B6,$3F,$01,$E9,$7F,$0E,$FA,$67,$D3,$5F,$1A,$57
	.dc.b	$F8,$08,$4E,$9C,$35,$FD,$8B,$5F,$E2,$2D,$7F,$B3,$B5,$1D,$27,$C4
	.dc.b	$47,$AF,$D1,$FF,$0A,$FA,$29,$1F,$97,$7F,$86,$FB,$18,$8F,$F5,$4E
	.dc.b	$89,$C7,$3F,$FA,$4A,$3F,$55,$D9,$FE,$6F,$DF,$FB,$BD,$FF,$9F,$B4
	.dc.b	$FE,$9B,$EE,$FF,$32,$8F,$F7,$3C,$E9,$BE,$AF,$D6,$23,$FA,$9E,$CF
	.dc.b	$B3,$F4,$DF,$C8,$3E,$8E,$4F,$F4,$DB,$4E,$A3,$EE,$BF,$C7,$A3,$31
	.dc.b	$82,$1E,$28,$E0,$FF,$B1,$BE,$21,$9F,$F6,$45,$F4,$74,$93,$FC,$C2
	.dc.b	$1F,$41,$88,$E5,$37,$7F,$F5,$BD,$6A,$7D,$1F,$4D,$D7,$E7,$E4,$60
	.dc.b	$87,$C6,$8F,$CC,$9E,$3F,$17,$F1,$9F,$17,$F8,$DF,$90,$95,$F8,$88
	.dc.b	$CB,$FE,$64,$E9,$FF,$5B,$B6,$3F,$AE,$C6,$23,$F1,$B1,$C1,$FF,$75
	.dc.b	$6B,$F8,$C8,$FE,$73,$F4,$CE,$9F,$D5,$FC,$6F,$F8,$23,$E1,$1F,$23
	.dc.b	$4C,$E3,$F5,$66,$BF,$80,$7F,$C1,$FE,$2F,$FA,$46,$3F,$AC,$AF,$1B
	.dc.b	$24,$8C,$FD,$35,$7F,$C2,$31,$78,$FF,$DC,$74,$AF,$EF,$BE,$A7,$F2
	.dc.b	$0B,$EB,$E3,$78,$0C,$F5,$C8,$FF,$44,$CE,$DF,$EB,$ED,$7F,$BF,$B5
	.dc.b	$FF,$2A,$BB,$C3,$F0,$F1,$FB,$08,$C7,$06,$AB,$DF,$E3,$7C,$0F,$12
	.dc.b	$5F,$C0,$7E,$1F,$85,$88,$FF,$16,$C5,$C7,$DB,$3F,$23,$C3,$95,$F2
	.dc.b	$1F,$97,$86,$23,$FC,$6B,$16,$BE,$2C,$FF,$8D,$B7,$FF,$6E,$D7,$F6
	.dc.b	$E6,$BF,$E0,$06,$BF,$E5,$96,$BF,$EE,$B7,$73,$FE,$F0,$6E,$59,$F1
	.dc.b	$7E,$EF,$81,$F8,$58,$B7,$F7,$A3,$0F,$F8,$63,$8F,$E4,$BF,$E4,$C8
	.dc.b	$DB,$FE,$A0,$9F,$47,$99,$FE,$18,$C3,$FC,$2D,$3E,$58,$C1,$1F,$F2
	.dc.b	$84,$7F,$85,$A7,$D1,$3F,$97,$EE,$FF,$2E,$FD,$0B,$E9,$BF,$CD,$AD
	.dc.b	$C3,$FC,$D4,$D7,$FC,$7F,$EF,$23,$11,$FE,$96,$8C,$4F,$42,$47,$F8
	.dc.b	$D3,$E9,$BF,$38,$FF,$AF,$6B,$A6,$DF,$EE,$06,$E5,$FE,$CD,$6C,$BC
	.dc.b	$C5,$F2,$8F,$F6,$04,$EB,$1F,$0C,$75,$C7,$FC,$F7,$5B,$C7,$B7,$D3
	.dc.b	$7B,$89,$5E,$E6,$BD,$1F,$8A,$7E,$F7,$C5,$FF,$33,$FC,$3F,$F8,$DE
	.dc.b	$BF,$A6,$FD,$2B,$E6,$FF,$D6,$F2,$F6,$3F,$D7,$33,$9C,$7F,$A9,$CE
	.dc.b	$5E,$35,$C3,$DA,$61,$89,$DC,$61,$7C,$87,$37,$FD,$89,$8F,$FB,$1A
	.dc.b	$31,$87,$59,$DE,$1D,$70,$FE,$C7,$5A,$C3,$A6,$3C,$31,$F3,$DF,$A9
	.dc.b	$B8,$78,$B0,$F2,$3F,$ED,$3B,$E3,$5B,$8C,$E3,$E1,$C3,$B5,$E2,$FD
	.dc.b	$58,$F7,$BF,$1E,$1C,$5F,$39,$DC,$A7,$7A,$EB,$8F,$92,$3E,$09,$D6
	.dc.b	$22,$C3,$B3,$6B,$B4,$79,$23,$07,$77,$64,$8F,$91,$1A,$3C,$47,$93
	.dc.b	$1F,$9D,$87,$63,$A1,$D6,$B9,$4F,$E0,$30,$F4,$57,$E2,$63,$A3,$6B
	.dc.b	$95,$FC,$37,$9B,$E2,$73,$3C,$CF,$92,$7C,$77,$D9,$F1,$5F,$E3,$BA
	.dc.b	$EC,$DC,$BF,$14,$DC,$BF,$C4,$AF,$F8,$F7,$E9,$AD,$FF,$A7,$9B,$87
	.dc.b	$FA,$F9,$AF,$F8,$B0,$38,$7F,$C5,$EF,$91,$FF,$18,$FC,$77,$F4,$49
	.dc.b	$FC,$C4,$68,$FB,$A2,$35,$97,$86,$33,$18,$7C,$46,$39,$3B,$3C,$F5
	.dc.b	$EB,$3F,$D4,$9E,$F1,$E6,$3E,$D1,$DE,$7F,$36,$F8,$EF,$8C,$74,$8E
	.dc.b	$11,$87,$CE,$73,$9C,$1D,$B0,$E7,$7E,$1C,$3D,$53,$A6,$9F,$EE,$88
	.dc.b	$D5,$7B,$27,$4D,$9D,$61,$E6,$8F,$1C,$61,$F1,$DF,$C2,$BF,$8D,$1F
	.dc.b	$56,$3C,$75,$C5,$B1,$F0,$8F,$55,$73,$BF,$54,$7C,$6B,$F2,$BC,$CF
	.dc.b	$A0,$78,$0B,$87,$09,$F4,$C6,$DF,$5F,$17,$F4,$7F,$98,$FF,$02,$E9
	.dc.b	$FC,$FD,$B9,$FF,$9D,$B9,$5D,$4F,$3B,$39,$31,$E5,$7F,$D2,$70,$FE
	.dc.b	$CF,$3F,$D1,$DA,$EF,$9D,$18,$AE,$33,$FD,$6A,$5F,$1B,$DA,$7F,$AE
	.dc.b	$3E,$0C,$9F,$4C,$F7,$35,$7C,$32,$FE,$B5,$78,$38,$56,$E3,$24,$C7
	.dc.b	$FA,$BC,$EA,$F8,$4E,$67,$FB,$55,$F5,$7C,$D5,$E9,$8C,$47,$F8,$4F
	.dc.b	$C9,$F8,$BB,$3B,$BF,$C9,$18,$7F,$82,$FD,$0C,$C6,$2F,$FC,$B6,$7F
	.dc.b	$4E,$70,$F3,$9F,$F1,$15,$FC,$AB,$E8,$FC,$CD,$3E,$AF,$D2,$FC,$5E
	.dc.b	$17,$D1,$D8,$C3,$70,$FF,$0E,$37,$1F,$0D,$FD,$E8,$F2,$F8,$F2,$B6
	.dc.b	$7F,$86,$1C,$57,$F8,$CD,$F2,$66,$69,$E5,$C5,$F4,$7E,$53,$CD,$F2
	.dc.b	$37,$0F,$F1,$74,$5C,$3C,$74,$AC,$6B,$FE,$5A,$CF,$F2,$7F,$1F,$EF
	.dc.b	$3E,$07,$15,$CB,$FC,$71,$17,$1E,$73,$6F,$FC,$77,$F3,$1E,$4B,$FF
	.dc.b	$6F,$36,$BC,$3F,$45,$F9,$E5,$19,$AE,$3A,$7E,$72,$72,$D9,$E5,$1F
	.dc.b	$CC,$CC,$DD,$7F,$DC,$31,$CE,$3F,$5D,$A3,$E8,$BC,$13,$9A,$FF,$17
	.dc.b	$4E,$6B,$FC,$A9,$F2,$FF,$7C,$FC,$07,$D1,$58,$C9,$2B,$E5,$C6,$5F
	.dc.b	$C7,$E4,$FA,$37,$37,$E6,$1B,$7F,$E6,$1F,$A2,$C4,$C8,$FF,$32,$7C
	.dc.b	$3D,$C3,$EB,$BF,$1F,$B1,$F4,$B4,$FE,$0F,$F2,$0F,$A6,$75,$E7,$16
	.dc.b	$3F,$A2,$3F,$86,$5F,$CD,$C3,$9C,$FC,$99,$CB,$63,$31,$F3,$5B,$AF
	.dc.b	$F9,$E1,$7D,$37,$79,$ED,$18,$AF,$0D,$7F,$22,$DB,$F5,$31,$83,$2E
	.dc.b	$2B,$FD,$E5,$8F,$C7,$EC,$95,$FE,$EF,$6C,$FF,$BD,$DF,$E3,$1F,$07
	.dc.b	$F6,$FA,$8D,$3F,$3D,$5F,$EF,$7E,$7D,$91,$FF,$7D,$75,$DC,$3E,$CE
	.dc.b	$4E,$9F,$B0,$76,$EB,$B7,$E9,$AF,$F2,$CD,$6E,$3F,$25,$AF,$E0,$6F
	.dc.b	$F8,$64,$E6,$BF,$31,$39,$AD,$CE,$63,$F6,$59,$7F,$2B,$AF,$1D,$FB
	.dc.b	$BA,$FF,$15,$CF,$31,$F1,$6B,$F0,$2B,$17,$EA,$AF,$E7,$9F,$0E,$F8
	.dc.b	$AC,$E7,$FB,$F8,$C6,$3C,$B3,$FD,$16,$1F,$C6,$1B,$5E,$4E,$AF,$9F
	.dc.b	$DB,$2F,$83,$F4,$A3,$17,$F8,$2F,$96,$FC,$C1,$F9,$9F,$A3,$B7,$F2
	.dc.b	$67,$FA,$F4,$FE,$4C,$91,$F9,$16,$C7,$E3,$3C,$BE,$2D,$9F,$46,$BF
	.dc.b	$A2,$FA,$6F,$91,$9F,$CA,$8C,$70,$FF,$48,$FA,$6F,$F4,$5D,$FD,$66
	.dc.b	$E1,$F5,$1A,$FF,$8F,$ED,$18,$8F,$EF,$1B,$3C,$0F,$C8,$9F,$92,$F8
	.dc.b	$2B,$87,$E1,$70,$E3,$F1,$1F,$D3,$71,$34,$93,$E3,$F5,$34,$47,$07
	.dc.b	$F1,$DE,$6F,$CE,$F8,$5F,$A0,$3D,$93,$67,$C1,$7F,$E0,$26,$F9,$9F
	.dc.b	$A6,$F9,$7F,$84,$C8,$F8,$4F,$94,$8E,$0F,$C2,$7D,$36,$F7,$FC,$23
	.dc.b	$F2,$FC,$A9,$3F,$A3,$F1,$9F,$12,$F8,$1C,$E1,$FC,$93,$3F,$EF,$17
	.dc.b	$F4,$70,$F8,$84,$70,$7E,$73,$B1,$EB,$72,$E2,$FC,$C3,$FE,$02,$FA
	.dc.b	$47,$65,$3F,$91,$F0,$BE,$67,$18,$FD,$3A,$DB,$FB,$66,$E9,$FB,$46
	.dc.b	$BF,$81,$6B,$FC,$69,$DE,$29,$FE,$5B,$BF,$FB,$87,$E0,$BB,$DF,$F8
	.dc.b	$42,$31,$3E,$1F,$97,$F4,$1F,$CE,$3E,$95,$F4,$D8,$9D,$1E,$59,$FF
	.dc.b	$0A,$64,$69,$23,$E8,$57,$BD,$E3,$87,$86,$30,$FF,$54,$A7,$CB,$FC
	.dc.b	$93,$FB,$1F,$93,$6F,$E3,$47,$80,$8F,$C7,$6D,$7F,$87,$38,$7F,$53
	.dc.b	$AF,$F9,$7B,$0E,$B7,$F6,$7D,$BF,$AD,$F1,$72,$3F,$80,$79,$FE,$B7
	.dc.b	$EA,$46,$71,$FA,$4F,$C0,$24,$7F,$55,$FA,$5F,$C6,$37,$A3,$FC,$0D
	.dc.b	$18,$AF,$F0,$13,$5F,$F0,$33,$9C,$BF,$CE,$7F,$4E,$EE,$FF,$09,$4E
	.dc.b	$70,$FE,$AF,$39,$7F,$11,$87,$E3,$CF,$F3,$47,$92,$E7,$C4,$E8,$99
	.dc.b	$7F,$94,$34,$56,$E7,$39,$FE,$26,$BF,$CF,$DE,$C3,$E3,$3F,$CE,$8D
	.dc.b	$AF,$E7,$BE,$C3,$C5,$CD,$49,$FE,$A5,$97,$F9,$16,$3F,$CA,$D3,$9C
	.dc.b	$3C,$58,$73,$7F,$C5,$15,$BB,$FE,$F9,$39,$9F,$F6,$F4,$E5,$DC,$7F
	.dc.b	$94,$67,$BB,$63,$F3,$DF,$2E,$EB,$F4,$CC,$AD,$BF,$E1,$69,$D4,$60
	.dc.b	$FE,$85,$F4,$DF,$EA,$1F,$82,$FF,$09,$37,$0F,$F1,$E3,$5F,$F2,$03
	.dc.b	$5F,$F7,$A7,$C2,$BE,$61,$F4,$DF,$E1,$B8,$C3,$63,$33,$FE,$47,$9C
	.dc.b	$BD,$BD,$EF,$E1,$ED,$F1,$3C,$71,$88,$F0,$B5,$FF,$16,$ED,$FE,$75
	.dc.b	$E5,$FE,$91,$A6,$36,$FE,$45,$5F,$22,$72,$FF,$94,$3C,$AB,$7C,$63
	.dc.b	$44,$3F,$CE,$F1,$AA,$7C,$77,$F8,$D5,$B3,$F6,$A3,$04,$7E,$BB,$C4
	.dc.b	$F0,$F9,$9F,$ED,$0F,$2F,$13,$A2,$1F,$E8,$67,$CF,$CE,$C3,$F2,$53
	.dc.b	$FD,$75,$AF,$F9,$81,$B1,$FE,$8B,$FA,$5F,$90,$FF,$B8,$A3,$3A,$7F
	.dc.b	$91,$1B,$49,$ED,$CD,$E6,$BE,$37,$FE,$5A,$E2,$7C,$43,$E3,$7E,$5E
	.dc.b	$2F,$C3,$7F,$A0,$3F,$E0,$D6,$BF,$EC,$2A,$E3,$3F,$49,$B1,$BF,$2F
	.dc.b	$9D,$70,$CF,$F6,$D5,$AA,$FF,$35,$7D,$27,$FB,$95,$FF,$34,$35,$FF
	.dc.b	$49,$35,$FF,$4B,$35,$FF,$39,$87,$84,$FF,$03,$8C,$7A,$BE,$C9,$CE
	.dc.b	$7C,$3E,$97,$E1,$D7,$F9,$F3,$E2,$DF,$59,$F0,$BC,$C7,$D5,$D3,$E2
	.dc.b	$CE,$AB,$FC,$60,$D7,$FD,$50,$D7,$FD,$58,$D7,$CE,$D7,$C1,$5D,$08
	.dc.b	$FF,$8E,$E9,$F3,$BF,$53,$F2,$95,$FE,$43,$CE,$7C,$E5,$35,$13,$0F
	.dc.b	$F4,$95,$70,$6C,$7F,$A5,$54,$65,$FF,$60,$37,$7F,$94,$DD,$7E,$7B
	.dc.b	$65,$D2,$C7,$FC,$67,$1F,$E5,$39,$FF,$5B,$BF,$E1,$1F,$63,$D0,$7A
	.dc.b	$FB,$E4,$79,$9F,$24,$F0,$8F,$1C,$60,$DB,$FE,$B1,$8D,$63,$FE,$71
	.dc.b	$F2,$BF,$33,$F5,$6F,$AC,$FC,$DE,$3F,$E9,$1C,$3F,$CC,$CD,$7F,$D7
	.dc.b	$53,$B7,$FC,$D4,$D7,$FD,$C4,$D7,$FD,$CC,$D7,$F2,$2D,$7F,$D8,$AF
	.dc.b	$1B,$FC,$93,$74,$DD,$F9,$3E,$63,$F5,$FF,$86,$3E,$47,$F8,$5F,$E9
	.dc.b	$09,$A6,$A2,$65,$FE,$CD,$AE,$0D,$8F,$B9,$18,$8F,$F7,$D3,$72,$FD
	.dc.b	$23,$71,$FD,$4B,$67,$ED,$E0,$F9,$8F,$C7,$7C,$C4,$9F,$F2,$53,$E8
	.dc.b	$F9,$C9,$1E,$37,$C2,$47,$FC,$DD,$C5,$FD,$2D,$FF,$4B,$60,$9E,$31
	.dc.b	$FE,$21,$F3,$3F,$39,$E0,$C8,$FF,$2C,$3F,$9F,$FA,$3F,$8F,$E4,$7B
	.dc.b	$89,$3F,$0D,$FF,$78,$73,$FE,$4F,$E6,$FF,$4C,$49,$FF,$31,$7C,$27
	.dc.b	$F1,$EF,$AB,$7C,$4F,$F2,$0F,$AC,$FF,$24,$D7,$FA,$AD,$AF,$FB,$FE
	.dc.b	$76,$FF,$AB,$DA,$FF,$12,$01,$CB,$96,$BF,$C6,$1A,$F0,$79,$F2,$7F
	.dc.b	$19,$6F,$26,$EF,$CD,$EB,$7F,$01,$FA,$8F,$E0,$5F,$39,$31,$FF,$3F
	.dc.b	$46,$5F,$F4,$04,$65,$FF,$27,$47,$F0,$9F,$0B,$F8,$57,$89,$27,$FD
	.dc.b	$0D,$19,$7F,$D1,$11,$97,$FC,$AD,$1F,$C4,$7C,$5F,$E2,$5F,$11,$27
	.dc.b	$FD,$1D,$19,$7F,$D2,$11,$97,$FC,$BD,$19,$7F,$CC,$11,$9D,$7F,$DD
	.dc.b	$53,$FD,$9B,$F7,$77,$CC,$FF,$1F,$F0,$26,$1F,$EC,$39,$CB,$67,$81
	.dc.b	$1F,$F7,$DF,$CA,$7D,$5C,$3F,$C8,$DC,$A8,$7B,$7F,$DE,$97,$CD,$53
	.dc.b	$11,$2F,$9C,$65,$EB,$1F,$8B,$F5,$7F,$15,$F2,$72,$7F,$0D,$19,$7F
	.dc.b	$0C,$7D,$F7,$F9,$97,$37,$F8,$0E,$32,$7E,$26,$32,$71,$7F,$24,$FE
	.dc.b	$5F,$3F,$41,$E9,$F8,$D9,$3F,$02,$32,$FC,$28,$CB,$FE,$05,$E7,$FC
	.dc.b	$BE,$9F,$0E,$BF,$C0,$3F,$BB,$FF,$77,$9C,$FE,$9C,$CF,$FC,$3E,$DC
	.dc.b	$3F,$3F,$BF,$F6,$FB,$E1,$1F,$A2,$6A,$69,$AC,$54,$F9,$5F,$F0,$8F
	.dc.b	$D2,$FF,$8F,$AF,$CC,$7F,$86,$FE,$BD,$F4,$BF,$E2,$A6,$C7,$8B,$4F
	.dc.b	$F0,$1D,$71,$CF,$FC,$04,$F3,$9F,$DB,$35,$DF,$C0,$FF,$63,$8C,$47
	.dc.b	$EE,$9E,$55,$A8,$EC,$66,$7F,$11,$AF,$E5,$F1,$FC,$56,$52,$BF,$2D
	.dc.b	$19,$69,$0F,$6E,$9F,$93,$ED,$FD,$07,$D3,$FE,$4F,$DC,$95,$F9,$F8
	.dc.b	$CB,$48,$72,$E7,$FE,$2F,$B7,$FA,$8C,$74,$3E,$07,$F1,$B1,$86,$C6
	.dc.b	$6B,$F2,$B3,$96,$C6,$5F,$A7,$19,$6C,$66,$3F,$CA,$DF,$1D,$FE,$4F
	.dc.b	$24,$FC,$8C,$DF,$00,$9F,$F2,$A7,$C3,$EA,$BE,$1C,$E5,$E0,$D7,$31
	.dc.b	$FE,$6A,$FD,$E3,$FC,$7D,$1F,$E4,$2C,$11,$88,$FC,$AB,$5F,$F3,$23
	.dc.b	$63,$7C,$7C,$C9,$3E,$0F,$9C,$FF,$45,$FE,$DD,$F9,$F8,$FF,$5F,$B5
	.dc.b	$FF,$5E,$B5,$FF,$4C,$B5,$7C,$3A,$39,$5F,$FB,$5E,$31,$CB,$D8,$FF
	.dc.b	$3B,$7B,$D6,$FF,$DA,$77,$C3,$1E,$B5,$A9,$CB,$B7,$FD,$CD,$3B,$7F
	.dc.b	$C6,$BA,$7F,$91,$9E,$7F,$2B,$FC,$3B,$1F,$23,$68,$D2,$4F,$02,$1F
	.dc.b	$5B,$E6,$7F,$D0,$FF,$16,$F5,$FF,$D2,$12,$FA,$7F,$DE,$3F,$D9,$27
	.dc.b	$FD,$FC,$FF,$66,$8C,$A8,$CB,$C7,$AF,$FD,$A7,$5D,$2B,$75,$DF,$7F
	.dc.b	$ED,$FC,$78,$F7,$FF,$D0,$6D,$C7,$38,$F2,$AB,$1E,$3C,$7A,$4F,$FB
	.dc.b	$82,$FF,$D1,$78,$F1,$AF,$11,$BB,$F4,$57,$38,$C9,$E9,$AE,$31,$CE
	.dc.b	$36,$71,$37,$8F,$A5,$FD,$14,$6D,$A7,$C7,$9F,$5C,$77,$C7,$E0,$E3
	.dc.b	$E8,$AD,$1D,$8F,$11,$6F,$B6,$3C,$31,$C6,$3E,$6A,$E2,$79,$2F,$A3
	.dc.b	$F4,$DB,$7F,$52,$FE,$21,$CE,$78,$5F,$CC,$7C,$27,$A8,$B8,$6C,$E1
	.dc.b	$3C,$B1,$F8,$0E,$1F,$71,$F0,$E3,$E9,$1C,$5C,$CF,$A2,$33,$18,$8F
	.dc.b	$45,$7F,$8C,$BE,$33,$F8,$46,$1F,$3D,$B8,$FE,$59,$B3,$BC,$3F,$BA
	.dc.b	$FC,$55,$C3,$FC,$78,$D7,$FD,$9E,$D7,$FC,$7E,$D7,$FD,$D6,$7C,$13
	.dc.b	$FC,$1B,$C5,$F1,$5D,$99,$F2,$11,$98,$E3,$5F,$A4,$78,$9C,$CC,$E7
	.dc.b	$FE,$4F,$AE,$31,$9B,$FC,$13,$C3,$4E,$4F,$91,$E9,$3C,$1C,$CF,$96
	.dc.b	$39,$DE,$2B,$CD,$87,$96,$BA,$D7,$1C,$3E,$3D,$7E,$88,$F0,$38,$C3
	.dc.b	$BD,$F9,$27,$C5,$3E,$98,$C5,$F4,$6C,$F3,$BE,$CF,$D2,$3C,$67,$BC
	.dc.b	$B8,$79,$F0,$EF,$1E,$33,$B9,$FE,$74,$3C,$A7,$37,$67,$1C,$3C,$91
	.dc.b	$C7,$0F,$89,$87,$C1,$8E,$F3,$ED,$37,$5C,$63,$B6,$1E,$8E,$1F,$E8
	.dc.b	$1E,$F7,$F7,$3C,$3F,$A2,$B6,$BF,$9A,$FC,$0D,$AD,$D7,$F4,$0A,$FE
	.dc.b	$B5,$18,$AF,$EA,$31,$8E,$3F,$F6,$45,$7F,$93,$2B,$FA,$77,$D4,$F3
	.dc.b	$93,$DF,$AF,$FB,$8F,$1E,$62,$5F,$F4,$E7,$5C,$C9,$CE,$5F,$D3,$26
	.dc.b	$F0,$FF,$9A,$AB,$57,$C4,$95,$FD,$9A,$BC,$C4,$8F,$F0,$6F,$57,$FB
	.dc.b	$35,$FF,$15,$DF,$F7,$8F,$7F,$C3,$CB,$F1,$63,$13,$AA,$E1,$87,$F9
	.dc.b	$ED,$AF,$F8,$05,$AF,$F8,$02,$FF,$C0,$59,$FF,$84,$3D,$CF,$69,$EE
	.dc.b	$F7,$2E,$1F,$E0,$E8,$B7,$FE,$12,$9B,$8F,$F8,$AF,$17,$D1,$79,$5B
	.dc.b	$87,$13,$CD,$19,$3F,$C2,$53,$FE,$14,$6C,$F0,$BF,$54,$63,$E8,$BA
	.dc.b	$1F,$5F,$C3,$FE,$E7,$87,$C1,$8C,$BF,$E9,$CF,$1E,$DF,$F9,$05,$F2
	.dc.b	$7D,$0C,$9F,$F5,$0F,$CD,$5A,$FF,$1C,$E4,$AE,$DC,$D9,$D3,$FD,$4B
	.dc.b	$9F,$E3,$2F,$FC,$E2,$3E,$99,$F1,$7F,$99,$8C,$3E,$7F,$27,$13,$AD
	.dc.b	$3F,$CC,$F5,$A7,$FD,$3B,$19,$8F,$F5,$1B,$5F,$F7,$53,$5F,$DC,$3B
	.dc.b	$C3,$FC,$B9,$F3,$6D,$7B,$E3,$F8,$78,$C6,$1E,$0F,$80,$7C,$9F,$9D
	.dc.b	$B1,$FE,$53,$F8,$87,$C6,$66,$A4,$E7,$E9,$6E,$BF,$E7,$6B,$47,$D5
	.dc.b	$F9,$CF,$AD,$F2,$F6,$FF,$D9,$ED,$BF,$F3,$A3,$76,$F9,$AF,$C9,$F0
	.dc.b	$7E,$EC,$FC,$E6,$CF,$F3,$CF,$0D,$FA,$BB,$E0,$F8,$79,$B3,$2F,$F1
	.dc.b	$D3,$FE,$5D,$C3,$9E,$3F,$C7,$CF,$F2,$8C,$64,$F8,$D9,$F1,$8F,$F3
	.dc.b	$DF,$0A,$3F,$21,$E9,$FE,$47,$C3,$B8,$FF,$BA,$9B,$7F,$E8,$F6,$EB
	.dc.b	$F8,$87,$F4,$9C,$32,$BF,$DE,$3E,$B5,$BF,$C8,$CF,$E7,$7E,$2F,$18
	.dc.b	$7E,$41,$FD,$07,$C5,$7F,$17,$FD,$57,$FB,$2E,$3F,$EF,$A8,$C3,$A3
	.dc.b	$FD,$19,$1A,$C3,$FD,$45,$3A,$9E,$06,$8E,$11,$FE,$B0,$9E,$17,$FD
	.dc.b	$AB,$C5,$43,$E3,$FC,$5E,$3F,$8D,$B8,$78,$1B,$87,$36,$E1,$CE,$DF
	.dc.b	$25,$EB,$EA,$F6,$37,$0F,$4B,$70,$F2,$36,$FF,$BE,$C7,$FA,$2B,$87
	.dc.b	$F1,$07,$F4,$9E,$77,$C4,$3C,$1C,$E3,$FB,$0E,$7F,$B8,$FF,$9F,$EF
	.dc.b	$F0,$C4,$70,$7F,$B7,$2F,$F1,$FF,$11,$CB,$1F,$F7,$8C,$62,$7E,$7F
	.dc.b	$67,$D5,$3E,$B7,$D3,$D8,$FD,$FF,$E2,$3F,$B9,$B0,$70,$DF,$07,$FD
	.dc.b	$69,$1B,$E4,$FA,$3E,$57,$E0,$4F,$C1,$4D,$BF,$BF,$E5,$7F,$11,$B3
	.dc.b	$F8,$73,$04,$8F,$D1,$61,$E2,$9D,$C6,$23,$C9,$3E,$23,$FA,$3F,$23
	.dc.b	$E5,$FF,$24,$73,$9F,$C4,$3E,$1F,$9A,$97,$FE,$DD,$FD,$3D,$DB,$8B
	.dc.b	$FE,$53,$92,$FD,$E4,$BF,$14,$65,$B1,$FD,$DE,$DD,$9C,$F9,$FF,$2D
	.dc.b	$18,$AF,$E5,$A3,$CF,$FE,$68,$D3,$E8,$CF,$F9,$19,$7E,$FC,$F8,$3D
	.dc.b	$D4,$9F,$88,$71,$CF,$E4,$B5,$FE,$18,$D7,$FC,$58,$EE,$FB,$7C,$1B
	.dc.b	$E3,$7A,$52,$73,$1F,$D2,$7D,$CF,$D2,$DF,$F4,$66,$D7,$ED,$7B,$35
	.dc.b	$DD,$2B,$EC,$56,$AB,$F8,$F3,$5F,$F6,$0C,$FF,$BF,$E3,$9E,$1F,$B3
	.dc.b	$F4,$79,$57,$F0,$17,$C5,$E0,$4C,$3F,$85,$FC,$FD,$AF,$EC,$38,$7F
	.dc.b	$46,$9F,$0D,$75,$23,$83,$F8,$9F,$C9,$71,$C7,$F8,$13,$FD,$9F,$F0
	.dc.b	$C7,$CA,$D3,$FC,$85,$CC,$FC,$4F,$FC,$09,$7F,$C0,$2B,$78,$7F,$6A
	.dc.b	$6E,$7F,$9C,$6B,$FC,$FD,$AF,$F9,$7D,$AF,$EC,$9E,$33,$FE,$88,$F1
	.dc.b	$1F,$86,$BE,$1F,$C9,$87,$E2,$0F,$DC,$CE,$A3,$FC,$01,$F4,$9F,$9E
	.dc.b	$9F,$EF,$AD,$C7,$EF,$47,$0F,$9C,$7C,$5F,$0E,$E9,$3F,$B1,$BE,$17
	.dc.b	$FD,$5D,$AF,$F9,$DD,$AB,$E8,$3F,$15,$E8,$7F,$84,$7F,$1A,$7C,$4F
	.dc.b	$F9,$AE,$73,$87,$F8,$42,$72,$FA,$EF,$96,$FF,$E6,$48,$D5,$F1,$8C
	.dc.b	$4F,$F8,$82,$B8,$BE,$C3,$F3,$5F,$BD,$3E,$A7,$FC,$AC,$DC,$3B,$79
	.dc.b	$4F,$85,$7C,$F6,$AE,$57,$F9,$82,$F8,$61,$D7,$0E,$17,$A8,$F0,$71
	.dc.b	$BF,$72,$7C,$DF,$FA,$1D,$B8,$F8,$33,$FB,$AF,$D8,$8C,$DF,$26,$EB
	.dc.b	$F8,$16,$CF,$F3,$6C,$BD,$1D,$5D,$71,$93,$FE,$8D,$AE,$78,$7F,$AA
	.dc.b	$F2,$F2,$57,$FA,$5A,$73,$3E,$1F,$91,$F3,$1F,$E8,$CF,$81,$B8,$7D
	.dc.b	$3C,$3D,$93,$8A,$F8,$5A,$2F,$1A,$7F,$62,$FA,$6F,$ED,$7E,$A7,$F3
	.dc.b	$96,$DF,$F3,$E6,$BF,$E2,$16,$BF,$EC,$1F,$91,$FE,$4D,$CB,$FD,$05
	.dc.b	$B1,$FD,$19,$AF,$F8,$A5,$AF,$FB,$26,$7F,$C8,$FC,$BE,$1F,$07,$F2
	.dc.b	$33,$E1,$6B,$E1,$6B,$FA,$AF,$77,$FA,$2B,$FE,$B7,$6B,$FE,$37,$D9
	.dc.b	$18,$8F,$9C,$D7,$97,$57,$FB,$6F,$4F,$DD,$4F,$26,$C6,$78,$3E,$E7
	.dc.b	$D3,$BE,$73,$18,$31,$EF,$EF,$63,$9F,$F3,$1F,$3F,$79,$0F,$0A,$DD
	.dc.b	$FD,$9D,$58,$7F,$6B,$9C,$CF,$AE,$72,$FB,$E7,$51,$88,$FC,$EB,$75
	.dc.b	$F5,$CF,$8F,$3F,$F3,$CD,$7F,$B1,$B3,$FE,$D4,$DF,$4F,$D7,$7E,$D7
	.dc.b	$D1,$C1,$8A,$DD,$7E,$E6,$B4,$F1,$3F,$51,$3F,$13,$8D,$F2,$8F,$8E
	.dc.b	$F8,$F4,$FA,$77,$A3,$86,$C7,$F0,$DB,$E3,$5C,$EF,$94,$FF,$A8,$AB
	.dc.b	$6F,$FA,$B1,$B8,$B8,$FF,$C4,$F8,$72,$AF,$F5,$4F,$D3,$FF,$84,$DF
	.dc.b	$F5,$5F,$D3,$FF,$2C,$6F,$8D,$E6,$9D,$47,$FA,$E1,$AF,$FA,$0D,$AF
	.dc.b	$F8,$D0,$38,$7F,$C6,$ED,$7F,$CB,$3F,$11,$F0,$BD,$D7,$D1,$53,$E2
	.dc.b	$FA,$9F,$E9,$5A,$FF,$5F,$B5,$FF,$64,$CF,$07,$FD,$86,$D7,$FD,$9B
	.dc.b	$1B,$D3,$FC,$BE,$DD,$3C,$CD,$9F,$F2,$5B,$5F,$F6,$9F,$C2,$7D,$6F
	.dc.b	$A7,$FF,$37,$C7,$F3,$2F,$A7,$FE,$B8,$DF,$3F,$57,$23,$04,$3F,$DB
	.dc.b	$F5,$C9,$B1,$B6,$9D,$7E,$6F,$C1,$E0,$FE,$A7,$07,$CC,$D3,$35,$FE
	.dc.b	$E2,$C3,$FD,$9E,$F9,$63,$4E,$6B,$6F,$89,$BA,$7F,$A5,$6D,$F3,$1F
	.dc.b	$94,$F4,$BF,$C5,$D7,$FE,$EE,$C3,$FD,$04,$DF,$6F,$5E,$E4,$8C,$10
	.dc.b	$FF,$7E,$D7,$26,$C6,$DA,$75,$F8,$AF,$D1,$FD,$3B,$E5,$BF,$84,$7C
	.dc.b	$9F,$FA,$9E,$BF,$DD,$95,$A9,$91,$82,$1F,$DA,$C5,$73,$7F,$BB,$18
	.dc.b	$8E,$2E,$E3,$4F,$FB,$CF,$E3,$BE,$87,$CC,$C8,$FC,$03,$E6,$F8,$79
	.dc.b	$18,$21,$CD,$FB,$1F,$09,$A8,$C1,$F8,$97,$EE,$78,$52,$3F,$18,$FD
	.dc.b	$FF,$5F,$E1,$F9,$36,$7F,$21,$18,$7F,$25,$EB,$FE,$14,$8F,$E5,$7D
	.dc.b	$4F,$E1,$BF,$4A,$F8,$2F,$E1,$FF,$4F,$FE,$0D,$AF,$86,$D7,$E4,$4E
	.dc.b	$DF,$8E,$D7,$E6,$46,$FA,$7F,$D9,$2D,$F0,$3F,$80,$56,$E7,$FD,$9A
	.dc.b	$F1,$AF,$E0,$81,$C3,$C2,$FF,$D5,$0E,$5F,$F5,$57,$A6,$EB,$FE,$8F
	.dc.b	$3F,$B1,$FA,$7F,$D2,$1F,$A5,$F4,$E4,$64,$87,$F8,$86,$78,$B7,$C3
	.dc.b	$FD,$67,$8D,$23,$F7,$91,$FA,$F3,$0F,$18,$C3,$FC,$07,$C6,$FD,$A7
	.dc.b	$C8,$C8,$C1,$0F,$03,$97,$C5,$1F,$BB,$31,$AF,$EC,$67,$13,$FE,$4B
	.dc.b	$87,$D0,$BE,$9B,$F8,$17,$7C,$AD,$4E,$5B,$3F,$ED,$1A,$E2,$D9,$D4
	.dc.b	$FD,$29,$CB,$FE,$DA,$DB,$FD,$FD,$ED,$FF,$77,$F0,$BF,$C8,$73,$FE
	.dc.b	$50,$6B,$FE,$4B,$51,$B3,$87,$BD,$27,$84,$7E,$4B,$D4,$FC,$8F,$BB
	.dc.b	$27,$F1,$B1,$97,$F2,$71,$97,$F2,$71,$97,$FD,$DF,$CB,$FE,$0F,$7F
	.dc.b	$DD,$CD,$73,$1F,$88,$C3,$DF,$97,$F6,$19,$64,$97,$D3,$FB,$CF,$81
	.dc.b	$AC,$AF,$81,$19,$7E,$14,$7F,$78,$FA,$7F,$ED,$BD,$1F,$E5,$DF,$A7
	.dc.b	$6B,$FE,$1F,$6D,$FF,$10,$AD,$A0,$FF,$78,$7F,$4A,$D7,$FB,$53,$5E
	.dc.b	$0D,$5F,$4D,$FE,$35,$BF,$F7,$9F,$D3,$BE,$8F,$FC,$32,$DA,$FE,$43
	.dc.b	$5B,$7C,$4D,$D9,$3B,$9F,$F0,$E3,$63,$E5,$47,$83,$E0,$7F,$C3,$35
	.dc.b	$FE,$49,$8E,$71,$FB,$A8,$C4,$6C,$FE,$02,$7F,$BD,$F9,$3A,$63,$FE
	.dc.b	$85,$DD,$F4,$DF,$91,$3F,$C6,$33,$AA,$E1,$9F,$E6,$C8,$F4,$CB,$F4
	.dc.b	$F1,$F9,$FF,$A3,$FC,$D7,$BB,$2B,$F4,$F1,$D0,$D3,$C6,$D1,$88,$FD
	.dc.b	$34,$7F,$9A,$1F,$F5,$17,$7F,$F9,$64,$F2,$FC,$0C,$8E,$4F,$EB,$C8
	.dc.b	$FE,$B7,$E0,$DF,$21,$FE,$61,$24,$7F,$AB,$7D,$CA,$FF,$A5,$A2,$E9
	.dc.b	$F9,$9E,$8F,$F6,$05,$7F,$95,$3E,$93,$E9,$CE,$70,$78,$5F,$E6,$D3
	.dc.b	$5C,$53,$2F,$8D,$7F,$69,$FF,$3F,$35,$FC,$A5,$A3,$11,$FE,$7B,$FA
	.dc.b	$5F,$CD,$BF,$EA,$AF,$0F,$FC,$F4,$7F,$9E,$63,$35,$FF,$00,$5D,$CF
	.dc.b	$FA,$ED,$AF,$FC,$00,$6B,$D7,$AB,$FC,$FD,$1C,$B0,$D4,$75,$3F,$D1
	.dc.b	$1E,$E5,$AF,$F5,$F5,$6D,$E4,$7F,$B5,$E3,$C9,$F4,$3F,$E8,$FF,$16
	.dc.b	$4F,$08,$CB,$F9,$4F,$87,$5A,$7C,$64,$9F,$F5,$E7,$A4,$CF,$F7,$5F
	.dc.b	$4D,$F9,$89,$FD,$E7,$C5,$7F,$8A,$9C,$F5,$FF,$B1,$7D,$CF,$F3,$25
	.dc.b	$6E,$54,$C3,$A4,$27,$53,$9C,$39,$CF,$6B,$E7,$87,$58,$C9,$D3,$0E
	.dc.b	$78,$72,$7A,$F3,$74,$BF,$F3,$B3,$6B,$F8,$A3,$BC,$2E,$1F,$ED,$DB
	.dc.b	$E6,$F7,$AF,$33,$70,$E1,$18,$3E,$1C,$F0,$7D,$07,$9B,$1F,$8C,$EC
	.dc.b	$F4,$5E,$67,$84,$7C,$6C,$7D,$EF,$23,$85,$7C,$56,$CF,$08,$EB,$7E
	.dc.b	$D9,$CD,$F5,$3C,$57,$D2,$3B,$19,$36,$76,$75,$5D,$30,$F8,$95,$BB
	.dc.b	$EC,$48,$CD,$F2,$8C,$3E,$97,$E3,$92,$3E,$51,$23,$E4,$CF,$8F,$E1
	.dc.b	$FF,$84,$C6,$03,$87,$E4,$98,$CB,$C3,$19,$8F,$EA,$1F,$0F,$FD,$46
	.dc.b	$31,$1F,$ED,$43,$81,$FE,$36,$8D,$4F,$F3,$86,$C6,$4F,$B0,$7F,$80
	.dc.b	$B8,$BF,$22,$F6,$33,$E4,$FF,$71,$CB,$F4,$11,$96,$C6,$72,$ED,$3C
	.dc.b	$2B,$9B,$E5,$AD,$C6,$2F,$BD,$F1,$AF,$25,$69,$F2,$9C,$6F,$FC,$A3
	.dc.b	$3B,$8F,$F6,$B5,$72,$BE,$91,$E1,$38,$C6,$5A,$66,$FD,$6F,$53,$FD
	.dc.b	$C1,$5D,$5F,$75,$F8,$2F,$AD,$79,$4E,$37,$D5,$B5,$D6,$BE,$1C,$6D
	.dc.b	$F8,$B8,$7B,$4D,$95,$D9,$DE,$7C,$F3,$C5,$F8,$B1,$98,$EE,$67,$83
	.dc.b	$F9,$0F,$B5,$FD,$0D,$B8,$FF,$25,$E1,$FF,$5A,$36,$75,$18,$C3,$F9
	.dc.b	$24,$FF,$87,$7A,$9D,$2F,$37,$FC,$65,$C1,$FD,$17,$1B,$7B,$AD,$3F
	.dc.b	$CF,$3E,$3F,$38,$FF,$51,$6E,$3D,$38,$F3,$A7,$F3,$89,$BA,$1A,$70
	.dc.b	$C7,$FD,$BB,$8F,$52,$63,$FE,$94,$BE,$51,$EF,$AF,$69,$1F,$EC,$A7
	.dc.b	$9F,$CE,$EA,$E2,$BF,$C9,$B6,$9F,$4E,$9F,$DA,$FD,$AF,$ED,$CD,$BF
	.dc.b	$EE,$71,$8E,$C3,$4F,$7F,$D9,$7F,$91,$E7,$FD,$09,$3A,$7C,$1B,$27
	.dc.b	$59,$7D,$23,$C7,$ED,$7C,$13,$E1,$7B,$56,$F8,$E8,$9C,$57,$1C,$3E
	.dc.b	$CD,$E3,$3F,$B3,$93,$E7,$BD,$04,$CB,$FC,$1A,$DC,$39,$67,$FC,$15
	.dc.b	$F2,$5F,$F8,$36,$7D,$BD,$1F,$1B,$07,$CF,$7C,$32,$63,$FE,$18,$6E
	.dc.b	$1F,$E9,$1F,$27,$84,$7E,$13,$CA,$FF,$30,$57,$C6,$F9,$9F,$98,$47
	.dc.b	$FC,$42,$DB,$FF,$4D,$73,$F8,$63,$93,$F9,$08,$7D,$3B,$E2,$7F,$C4
	.dc.b	$6D,$BF,$F2,$DC,$EA,$72,$48,$FF,$5D,$7D,$23,$E9,$FA,$5F,$1B,$F8
	.dc.b	$73,$9A,$FC,$54,$73,$78,$D7,$0B,$FC,$5D,$6E,$BE,$F0,$F7,$9F,$4B
	.dc.b	$FE,$38,$6E,$1E,$66,$CF,$F9,$C9,$AF,$FB,$D1,$CF,$17,$F9,$3F,$E1
	.dc.b	$9F,$49,$F0,$E7,$37,$FB,$D9,$CB,$FE,$C5,$FA,$5F,$A5,$F4,$FF,$4D
	.dc.b	$B1,$93,$FD,$95,$19,$69,$F0,$3B,$FF,$C6,$0F,$CC,$6E,$5F,$E3,$5F
	.dc.b	$2B,$FC,$A7,$18,$8F,$DC,$47,$C4,$BF,$B1,$3C,$33,$FE,$1F,$AA,$73
	.dc.b	$5B,$C3,$C0,$DC,$3F,$DB,$9E,$8F,$E2,$FC,$1F,$C6,$36,$BF,$DC,$4D
	.dc.b	$9D,$C7,$FB,$96,$7F,$35,$BF,$EE,$C9,$1F,$E7,$06,$CF,$FB,$A7,$E1
	.dc.b	$F7,$3F,$EE,$D7,$F3,$5F,$0F,$FC,$7B,$E9,$53,$83,$F7,$EE,$1F,$F7
	.dc.b	$A4,$6D,$D5,$7E,$E6,$74,$FF,$BD,$E7,$53,$C0,$E6,$77,$F8,$AF,$07
	.dc.b	$89,$63,$BB,$6B,$A3,$77,$FF,$4A,$61,$FC,$4E,$7C,$1F,$21,$E9,$F8
	.dc.b	$EF,$53,$70,$F4,$B6,$77,$1E,$C9,$FB,$3C,$1F,$78,$91,$F8,$36,$CF
	.dc.b	$F1,$DF,$01,$F4,$FF,$D2,$23,$F4,$7C,$9F,$7E,$FF,$9A,$56,$BC,$BE
	.dc.b	$B1,$8B,$E1,$A7,$FB,$72,$F8,$61,$D6,$73,$5F,$D6,$F4,$FC,$FC,$FC
	.dc.b	$DE,$9F,$AD,$F4,$FF,$5D,$FE,$11,$F4,$5F,$BD,$CB,$E2,$F6,$60,$FB
	.dc.b	$87,$DD,$F1,$66,$5F,$72,$2E,$3F,$67,$C3,$E6,$7E,$0C,$B3,$FC,$1B
	.dc.b	$0F,$E4,$EF,$E1,$3E,$9F,$35,$F8,$17,$F1,$44,$8E,$15,$67,$F8,$4D
	.dc.b	$3E,$67,$D2,$7F,$6E,$FA,$77,$C1,$E3,$0F,$04,$E7,$AB,$18,$7F,$47
	.dc.b	$B4,$63,$5F,$F6,$C4,$62,$F8,$9F,$D6,$1C,$57,$F1,$A8,$FE,$A4,$7F
	.dc.b	$22,$2B,$FD,$6D,$7D,$3B,$E9,$FF,$9B,$46,$AF,$F4,$F5,$FD,$9F,$CF
	.dc.b	$FF,$04,$CB,$E1,$3F,$BA,$3D,$63,$31,$89,$F1,$46,$3D,$2E,$AD,$AF
	.dc.b	$E6,$BE,$BB,$B7,$FA,$33,$63,$F6,$BE,$6E,$26,$47,$EE,$3C,$DB,$8F
	.dc.b	$ED,$5B,$9F,$EA,$9B,$3F,$8F,$7A,$F7,$C8,$FE,$03,$F4,$F6,$FF,$A0
	.dc.b	$E4,$AF,$E1,$1F,$23,$9C,$7F,$6A,$DC,$3F,$34,$DB,$7A,$8E,$D4,$FF
	.dc.b	$61,$FA,$67,$CA,$53,$FC,$67,$D2,$FA,$7F,$F0,$16,$5F,$D6,$2B,$6D
	.dc.b	$AF,$F1,$AF,$D3,$BE,$0F,$94,$FF,$89,$FC,$44,$E6,$38,$19,$71,$AF
	.dc.b	$F5,$EC,$3A,$BF,$E1,$DA,$E3,$7F,$DD,$FE,$9F,$F2,$F8,$7F,$74,$6D
	.dc.b	$08,$57,$2E,$03,$87,$6F,$2D,$FF,$CB,$3F,$52,$FA,$AF,$07,$81,$FE
	.dc.b	$3E,$8C,$61,$FE,$43,$FA,$25,$FF,$94,$E3,$FB,$EF,$D4,$FF,$AE,$1D
	.dc.b	$46,$72,$FE,$F7,$39,$7F,$C7,$55,$CB,$D1,$FF,$0A,$4E,$6B,$F3,$F3
	.dc.b	$9B,$EB,$F0,$BF,$E5,$0F,$A8,$7D,$4D,$3F,$CC,$5F,$3D,$6B,$FC,$B1
	.dc.b	$82,$B7,$18,$6E,$7F,$E4,$3F,$83,$5F,$18,$FA,$5F,$52,$F9,$CF,$F3
	.dc.b	$9B,$70,$FF,$26,$F3,$E2,$64,$7C,$FA,$72,$AF,$84,$7F,$9C,$DA,$FD
	.dc.b	$88,$D5,$7F,$A5,$67,$79,$FF,$00,$EC,$FF,$31,$4F,$6B,$FF,$28,$BE
	.dc.b	$4F,$8E,$98,$FF,$98,$3D,$FB,$5E,$A9,$E8,$49,$F3,$65,$E5,$8F,$E1
	.dc.b	$B3,$F8,$7F,$A9,$FE,$05,$EB,$7F,$90,$9B,$1E,$D6,$BB,$F1,$53,$FE
	.dc.b	$48,$F8,$DF,$F3,$87,$2D,$F2,$0D,$25,$F3,$F1,$3F,$DF,$ED,$E7,$F0
	.dc.b	$46,$23,$C8,$D7,$F4,$1A,$3E,$3F,$D3,$C5,$FC,$9A,$31,$BF,$E9,$23
	.dc.b	$13,$A8,$FF,$5D,$7C,$43,$E6,$B5,$1D,$27,$58,$7D,$A6,$E1,$D2,$3F
	.dc.b	$D2,$B7,$C5,$FE,$01,$F0,$1F,$D3,$1B,$DF,$D1,$FC,$1C,$65,$FF,$49
	.dc.b	$7D,$4B,$E7,$75,$82,$B5,$A7,$AD,$B9,$FC,$06,$E1,$FC,$7B,$0E,$4F
	.dc.b	$80,$87,$F0,$B8,$DF,$42,$77,$3F,$E9,$8F,$DE,$FF,$D4,$0F,$EA,$1A
	.dc.b	$E7,$DA,$F5,$57,$FA,$CF,$E7,$BD,$99,$2F,$6F,$FB,$89,$B9,$7F,$15
	.dc.b	$8D,$5F,$2D,$97,$D3,$CD,$7D,$7F,$C7,$CD,$FB,$AB,$E5,$B0,$D8,$FF
	.dc.b	$23,$CF,$13,$33,$23,$27,$FA,$43,$F1,$2F,$F7,$43,$7C,$8E,$D1,$88
	.dc.b	$FF,$47,$F8,$FF,$6D,$FF,$48,$C6,$75,$FD,$2B,$76,$FE,$5E,$D9,$FE
	.dc.b	$65,$5C,$6B,$F5,$7F,$56,$F7,$FF,$D5,$74,$8C,$3F,$EA,$E6,$BF,$C2
	.dc.b	$5A,$FF,$39,$73,$3D,$27,$E3,$46,$30,$FE,$83,$F5,$9F,$A7,$AD,$5F
	.dc.b	$FA,$C2,$B4,$FF,$47,$6A,$9E,$15,$FD,$2B,$EA,$FF,$DE,$F3,$FF,$00
	.dc.b	$7D,$47,$DC,$FF,$AA,$23,$0D,$7F,$D6,$F3,$C1,$BF,$13,$F6,$E3,$19
	.dc.b	$7F,$0E,$6E,$1F,$D6,$70,$EB,$F0,$DB,$B9,$87,$FB,$22,$F6,$AB,$6E
	.dc.b	$7C,$6F,$CA,$FD,$3F,$F7,$38,$FC,$86,$3F,$D9,$7A,$93,$99,$F3,$D7
	.dc.b	$94,$8F,$33,$89,$FE,$0A,$87,$D6,$EA,$38,$9B,$3F,$B1,$DF,$79,$DC
	.dc.b	$7F,$B8,$27,$FC,$22,$FF,$B8,$BE,$D9,$F3,$3F,$9F,$6C,$6F,$DC,$FD
	.dc.b	$15,$70,$8F,$F7,$1B,$77,$FE,$6E,$DC,$7F,$C0,$37,$F6,$AB,$FC,$09
	.dc.b	$39,$6D,$F4,$3C,$1E,$3F,$EB,$FE,$B5,$F0,$5F,$B1,$FA,$E7,$C1,$D3
	.dc.b	$51,$30,$FF,$7D,$D7,$06,$C7,$1F,$90,$D4,$7F,$BE,$A7,$4D,$9D,$3F
	.dc.b	$EC,$0F,$8A,$FE,$04,$47,$E8,$8F,$AE,$FE,$3E,$DF,$2F,$F8,$34,$60
	.dc.b	$FE,$3B,$E1,$BE,$E7,$F8,$61,$A9,$95,$FE,$CE,$8F,$E2,$1E,$FC,$8F
	.dc.b	$F6,$8B,$F8,$AF,$0F,$AF,$C6,$49,$F1,$E3,$FE,$82,$7F,$0B,$D3,$F8
	.dc.b	$CF,$4F,$F2,$24,$8C,$F4,$A3,$F8,$F7,$9F,$FC,$77,$EB,$7F,$7B,$5F
	.dc.b	$E3,$86,$BF,$E3,$B6,$A6,$9F,$B8,$27,$89,$FC,$A7,$E5,$7A,$32,$7E
	.dc.b	$4C,$FF,$00,$C7,$FC,$C5,$5F,$CE,$BF,$5B,$FC,$34,$17,$C4,$E4,$D7
	.dc.b	$4C,$C3,$DF,$3A,$6C,$E5,$FD,$09,$FE,$70,$87,$D3,$F7,$FB,$29,$3C
	.dc.b	$A3,$C4,$61,$E3,$87,$8E,$73,$1E,$0F,$88,$F4,$77,$48,$C1,$0E,$38
	.dc.b	$7E,$C1,$B1,$EF,$99,$1E,$A8,$E4,$F0,$8F,$EB,$7F,$BE,$7D,$78,$FF
	.dc.b	$24,$B5,$F6,$F8,$48,$C3,$FE,$35,$B7,$7F,$2E,$29,$7C,$9F,$A1,$1F
	.dc.b	$47,$E9,$3E,$97,$3C,$AF,$9B,$19,$7E,$7C,$65,$F6,$C7,$E2,$7C,$1F
	.dc.b	$E9,$7C,$7F,$C1,$67,$FC,$FC,$D7,$F1,$91,$A3,$83,$F8,$47,$FB,$77
	.dc.b	$D9,$7F,$B3,$E3,$04,$3E,$D9,$F9,$18,$DF,$C9,$7F,$2E,$FA,$AF,$E5
	.dc.b	$FF,$55,$2B,$F3,$F1,$97,$F4,$31,$97,$EE,$F5,$27,$FC,$13,$F5,$9F
	.dc.b	$D9,$FC,$8F,$F0,$33,$6B,$FD,$1A,$DE,$6F,$E2,$71,$C3,$DF,$8F,$F9
	.dc.b	$FA,$75,$E4,$7F,$A7,$F9,$3F,$CB,$7B,$FF,$87,$AB,$E2,$F7,$7C,$82
	.dc.b	$3F,$E7,$73,$F5,$CF,$ED,$48,$FF,$12,$E0,$F2,$3A,$3F,$5B,$1A,$7E
	.dc.b	$3C,$6B,$EC,$3F,$DE,$91,$FC,$69,$AF,$CC,$6D,$FF,$9F,$7E,$C5,$F5
	.dc.b	$9F,$E2,$19,$CE,$DF,$EB,$C6,$DF,$FA,$FD,$BB,$2F,$8C,$FC,$3F,$A9
	.dc.b	$F9,$A4,$7F,$8B,$EB,$F7,$DF,$CD,$EC,$9D,$46,$1F,$F1,$FC,$EF,$36
	.dc.b	$5F,$E3,$BE,$A7,$C5,$7E,$A0,$98,$7F,$99,$3E,$2A,$BF,$A8,$9D,$BF
	.dc.b	$AC,$9F,$F2,$14,$C8,$E0,$43,$F8,$1F,$BF,$EC,$6C,$6D,$A7,$F9,$CB
	.dc.b	$7F,$DC,$F3,$48,$FD,$CB,$FE,$4F,$E7,$C4,$78,$E2,$E7,$FE,$AD,$C3
	.dc.b	$DD,$9F,$FA,$33,$F7,$AF,$ED,$5F,$0A,$69,$FC,$A2,$7E,$9C,$67,$C5
	.dc.b	$FF,$3C,$CE,$9F,$F2,$F7,$CD,$61,$99,$7F,$83,$2B,$83,$FE,$8A,$9C
	.dc.b	$BF,$F0,$06,$B9,$39,$38,$3F,$CB,$1B,$E2,$F8,$0F,$F3,$67,$BB,$63
	.dc.b	$24,$3E,$5B,$FE,$6D,$F8,$AF,$F3,$66,$89,$FE,$83,$BD,$52,$F0,$31
	.dc.b	$E4,$7D,$38,$C4,$F1,$8C,$78,$4C,$31,$87,$3B,$EF,$6F,$AD,$E5,$8F
	.dc.b	$76,$C7,$8D,$E7,$76,$F9,$61,$D6,$B9,$47,$5B,$E4,$D8,$E0,$7B,$70
	.dc.b	$F0,$E1,$D2,$3D,$E6,$6B,$DB,$5C,$63,$27,$C2,$BF,$06,$3C,$6F,$CF
	.dc.b	$3E,$23,$91,$E0,$38,$63,$71,$F0,$E1,$D6,$B8,$9C,$DF,$6C,$EF,$1F
	.dc.b	$5D,$F4,$3B,$1D,$0B,$7D,$1D,$5F,$4B,$E7,$1A,$C7,$E4,$4E,$63,$73
	.dc.b	$A8,$F1,$46,$08,$77,$72,$FE,$12,$FC,$AD,$3D,$D5,$DA,$BA,$1E,$F2
	.dc.b	$47,$D3,$8D,$37,$0F,$C9,$6F,$FE,$0D,$8F,$73,$71,$FC,$4B,$6B,$F0
	.dc.b	$4E,$E7,$F3,$5F,$5D,$CB,$3F,$CD,$46,$04,$ED,$4E,$DD,$B5,$FD,$13
	.dc.b	$F6,$D7,$C6,$FF,$22,$7E,$99,$D7,$4F,$E2,$CE,$B8,$F8,$3D,$8D,$CF
	.dc.b	$19,$DC,$65,$E3,$19,$D3,$D3,$39,$7D,$71,$98,$F3,$9D,$2B,$BD,$F9
	.dc.b	$E7,$53,$EA,$34,$7A,$0E,$87,$EF,$DD,$BE,$BA,$EF,$5D,$23,$8D,$FA
	.dc.b	$2F,$B4,$E6,$3D,$2D,$AE,$35,$DE,$B6,$D3,$9C,$7B,$A3,$F8,$3C,$78
	.dc.b	$E6,$BE,$9C,$7A,$BF,$B4,$3C,$C6,$CE,$97,$A3,$CC,$F6,$3E,$3B,$9B
	.dc.b	$F0,$47,$3C,$3E,$E5,$70,$3B,$9F,$DB,$38,$BF,$A1,$79,$7F,$81,$DB
	.dc.b	$F0,$51,$8C,$BF,$D0,$7C,$8F,$A6,$D4,$7F,$4B,$BB,$92,$F8,$3F,$CF
	.dc.b	$E3,$3D,$EC,$39,$5F,$F4,$09,$E3,$7F,$D4,$FC,$CB,$CD,$FE,$D4,$AF
	.dc.b	$0F,$76,$26,$47,$F4,$87,$AE,$1F,$DF,$E3,$04,$CB,$FA,$84,$EA,$B8
	.dc.b	$C6,$27,$5A,$7F,$67,$9C,$BF,$0F,$B5,$F4,$9E,$E8,$FF,$33,$3E,$0F
	.dc.b	$AA,$C4,$EA,$FF,$CD,$B3,$83,$4E,$0F,$97,$F5,$9E,$83,$D3,$F6,$16
	.dc.b	$32,$4C,$BF,$C0,$F1,$97,$FC,$EF,$ED,$3E,$53,$C8,$DA,$E1,$9F,$DB
	.dc.b	$8C,$57,$0A,$D9,$CE,$74,$FB,$7E,$B9,$F4,$54,$FB,$3E,$33,$6C,$65
	.dc.b	$FD,$B7,$0F,$C4,$5F,$08,$C1,$97,$18,$FF,$A8,$AF,$94,$EA,$3F,$B6
	.dc.b	$4E,$AB,$F8,$94,$7F,$8C,$BD,$5F,$C6,$B6,$BF,$DA,$BF,$5A,$F9,$5C
	.dc.b	$F9,$2B,$A4,$AD,$CD,$BF,$F4,$0F,$3F,$E4,$9B,$D5,$F9,$AE,$C7,$D4
	.dc.b	$BC,$8F,$0D,$EF,$5F,$2B,$72,$FF,$06,$37,$1F,$E5,$77,$C6,$7F,$DF
	.dc.b	$F3,$99,$FD,$36,$3E,$E7,$8B,$5F,$F0,$C4,$79,$5F,$E7,$ED,$7E,$66
	.dc.b	$7F,$5C,$FA,$BE,$F5,$BF,$99,$8B,$7F,$89,$18,$6D,$FD,$3A,$E0,$FF
	.dc.b	$B5,$BD,$F7,$C1,$7E,$E1,$7C,$45,$8C,$37,$0F,$B5,$E0,$AB,$71,$87
	.dc.b	$9B,$67,$E5,$FD,$54,$8F,$C1,$70,$FF,$AA,$79,$71,$97,$D6,$AB,$AF
	.dc.b	$F5,$5A,$D5,$FC,$C9,$DD,$7C,$67,$E5,$3F,$C4,$DB,$B9,$F9,$D9,$BD
	.dc.b	$C7,$B9,$FC,$59,$7C,$35,$8C,$37,$1F,$7C,$E9,$B3,$A7,$E8,$F6,$7F
	.dc.b	$A8,$1B,$E3,$F5,$F9,$CF,$DB,$B2,$32,$74,$8C,$69,$FF,$00,$3E,$73
	.dc.b	$C9,$D7,$E5,$6D,$78,$9B,$3B,$8F,$2C,$ED,$A6,$CB,$1F,$E5,$B7,$FC
	.dc.b	$CD,$A7,$C3,$FB,$3F,$85,$F6,$3F,$0D,$B3,$F0,$1B,$5C,$45,$71,$9F
	.dc.b	$8B,$5F,$2F,$DE,$F9,$9D,$B6,$7E,$4B,$6F,$95,$FD,$38,$C7,$27,$F8
	.dc.b	$4E,$31,$5F,$AA,$E9,$FD,$99,$F5,$FE,$6E,$C6,$49,$87,$4E,$24,$EA
	.dc.b	$3F,$80,$D7,$E1,$09,$5F,$EE,$7E,$54,$7E,$1B,$A5,$F5,$5F,$A3,$3F
	.dc.b	$0D,$E3,$DB,$FE,$35,$83,$E3,$B7,$18,$6C,$7E,$B6,$BF,$8D,$C6,$35
	.dc.b	$F7,$68,$E6,$FC,$91,$FB,$1F,$86,$B8,$FE,$BB,$F0,$9F,$F4,$E3,$D6
	.dc.b	$3F,$B6,$C6,$32,$FA,$87,$8E,$3B,$92,$79,$45,$3E,$8C,$FF,$8F,$5B
	.dc.b	$D9,$9C,$7E,$94,$E5,$FE,$6B,$5F,$A8,$3E,$09,$6B,$F0,$8B,$EA,$9E
	.dc.b	$E7,$F3,$A8,$D5,$72,$EC,$C1,$FE,$0A,$8E,$13,$87,$AC,$67,$DC,$F8
	.dc.b	$D1,$83,$FA,$50,$73,$87,$C2,$FC,$8F,$F8,$04,$FD,$66,$C7,$9B,$1E
	.dc.b	$13,$9F,$03,$FA,$7D,$6B,$2F,$EA,$6D,$9F,$A3,$5C,$2F,$B5,$F0,$3D
	.dc.b	$50,$F7,$31,$5E,$68,$C3,$F8,$8F,$5B,$FA,$D3,$7C,$AF,$E6,$1F,$57
	.dc.b	$FD,$82,$73,$8F,$DC,$6E,$7F,$D9,$9B,$7F,$BA,$E7,$5F,$1C,$24,$FF
	.dc.b	$33,$F8,$8B,$8F,$F3,$F7,$B7,$D7,$49,$FC,$77,$A3,$6B,$C1,$37,$3F
	.dc.b	$C1,$CD,$AF,$F7,$4C,$FF,$90,$BF,$24,$7D,$33,$E5,$F5,$8A,$73,$1F
	.dc.b	$DF,$3E,$AB,$F5,$F8,$7E,$A9,$B5,$FC,$37,$C0,$7C,$AB,$E1,$78,$78
	.dc.b	$53,$0E,$0F,$F9,$2A,$75,$18,$25,$FF,$A4,$30,$EA,$FF,$AA,$9A,$88
	.dc.b	$E1,$C7,$E1,$AF,$3F,$FC,$26,$DA,$FF,$0D,$65,$E6,$F7,$73,$52,$7F
	.dc.b	$82,$D6,$EF,$D9,$8F,$F9,$75,$FD,$BC,$67,$87,$F5,$FF,$1D,$FE,$20
	.dc.b	$FA,$07,$C4,$7E,$FA,$30,$DC,$7F,$88,$CE,$9F,$E0,$D3,$A7,$FC,$07
	.dc.b	$F8,$EF,$F0,$F7,$FC,$70,$DC,$7F,$9C,$FC,$17,$E7,$56,$29,$D3,$7C
	.dc.b	$9F,$F3,$14,$71,$F3,$3F,$CF,$9E,$06,$27,$FC,$F7,$F5,$2F,$8B,$A7
	.dc.b	$FA,$0B,$A5,$E9,$E3,$5F,$93,$5B,$6D,$6E,$75,$A7,$E0,$EB,$4D,$9D
	.dc.b	$57,$03,$E9,$FE,$F7,$AD,$7F,$D3,$6D,$8F,$EB,$FB,$3E,$D7,$FD,$23
	.dc.b	$39,$C3,$FD,$34,$DE,$73,$9D,$FF,$D3,$55,$C3,$2F,$F2,$AF,$D5,$BE
	.dc.b	$7B,$FD,$D7,$8F,$FB,$7A,$31,$E6,$FF,$B8,$21,$18,$75,$3C,$1F,$F7
	.dc.b	$EF,$C7,$FF,$85,$3E,$AF,$83,$63,$FD,$D2,$D7,$7E,$6A,$7B,$5F,$F9
	.dc.b	$97,$E3,$EF,$8E,$69,$2F,$FC,$9B,$EB,$7F,$9C,$38,$FF,$D8,$2D,$E3
	.dc.b	$FF,$29,$B6,$7F,$0F,$D5,$FA,$77,$F6,$D1,$9E,$3F,$ED,$F5,$9B,$E6
	.dc.b	$D3,$E8,$DD,$BF,$F6,$47,$D5,$FF,$70,$7A,$1F,$E6,$27,$0F,$F9,$89
	.dc.b	$AF,$F9,$C1,$AE,$FC,$3E,$98,$27,$8E,$1F,$ED,$C6,$E1,$FE,$20,$FA
	.dc.b	$B7,$CB,$F0,$9C,$E3,$E4,$C7,$07,$9F,$99,$7C,$27,$FC,$DF,$F5,$6F
	.dc.b	$AB,$F6,$66,$8C,$4E,$A7,$97,$53,$EA,$F7,$5A,$AF,$F1,$65,$F1,$BC
	.dc.b	$9C,$2B,$C5,$E8,$3E,$AF,$E5,$5B,$C4,$7C,$BE,$1B,$1F,$E7,$39,$E2
	.dc.b	$66,$64,$64,$FF,$0D,$FD,$53,$EA,$FF,$31,$08,$C3,$F9,$B6,$BF,$60
	.dc.b	$D4,$70,$8C,$E7,$FE,$BC,$6F,$1F,$E2,$5B,$5F,$CC,$7D,$77,$37,$B3
	.dc.b	$EA,$DE,$EF,$BB,$EA,$FF,$A1,$5F,$F8,$C1,$AF,$F8,$82,$76,$FF,$8D
	.dc.b	$1A,$F9,$5A,$FE,$39,$AF,$E6,$DA,$FF,$43,$78,$E5,$FE,$D4,$6E,$5B
	.dc.b	$BF,$27,$5F,$D2,$FA,$AF,$F5,$3E,$1F,$ED,$DF,$82,$7D,$5B,$D8,$FD
	.dc.b	$04,$61,$E1,$F1,$B2,$BF,$83,$5F,$27,$E6,$35,$FD,$2B,$5F,$D9,$3B
	.dc.b	$D3,$FD,$D2,$DC,$F3,$5E,$0C,$7F,$BA,$5F,$F9,$62,$BF,$8C,$3F,$EB
	.dc.b	$AF,$A3,$7D,$2F,$E7,$DB,$1B,$F2,$3D,$C7,$E1,$2B,$8C,$79,$A3,$04
	.dc.b	$78,$10,$FF,$3F,$CB,$D7,$FC,$C8,$F7,$7F,$37,$F5,$7F,$EB,$2A,$FF
	.dc.b	$36,$B5,$FF,$2E,$CE,$DF,$F3,$8B,$5F,$C9,$35,$FE,$16,$D7,$F8,$C3
	.dc.b	$5F,$F0,$03,$C7,$56,$BA,$BF,$29,$E4,$E0,$FD,$7F,$B1,$FE,$77,$C3
	.dc.b	$FD,$03,$F4,$DF,$E0,$56,$FB,$5A,$F8,$79,$1F,$D0,$6B,$83,$FA,$76
	.dc.b	$BF,$C8,$1A,$FF,$2D,$77,$97,$92,$3F,$C2,$D1,$8C,$B5,$5F,$C0,$B0
	.dc.b	$FF,$7C,$D7,$FB,$C7,$EA,$FF,$83,$91,$FF,$4B,$FD,$47,$F1,$E4,$FF
	.dc.b	$2A,$7F,$C4,$4D,$73,$E4,$7F,$02,$8D,$79,$1F,$EF,$FF,$AB,$FE,$23
	.dc.b	$1C,$CF,$E5,$31,$9A,$FF,$16,$69,$FE,$BB,$F1,$1F,$57,$FC,$68,$93
	.dc.b	$F9,$1F,$0F,$9F,$D5,$F7,$24,$7F,$69,$C5,$1F,$DA,$F1,$73,$7F,$BE
	.dc.b	$AB,$F3,$5E,$9F,$E6,$4F,$E4,$FE,$77,$F2,$6F,$AB,$FD,$CC,$BD,$67
	.dc.b	$C9,$7F,$5D,$6F,$2E,$FC,$39,$5F,$DE,$6F,$F8,$2A,$69,$23,$71,$FE
	.dc.b	$B2,$8E,$F9,$B8,$1F,$35,$FD,$2F,$E4,$3F,$BA,$5F,$C6,$C7,$FD,$E7
	.dc.b	$0F,$8B,$D1,$BE,$C9,$5B,$3F,$DE,$33,$B7,$18,$A3,$11,$FE,$E3,$9F
	.dc.b	$07,$C3,$F8,$7E,$96,$4F,$0A,$FE,$57,$CF,$FE,$D4,$AC,$CF,$FB,$AE
	.dc.b	$B8,$BB,$F1,$3D,$5E,$0C,$8C,$10,$FF,$79,$39,$7F,$DE,$91,$F0,$4C
	.dc.b	$5F,$FB,$8A,$71,$3E,$D8,$7D,$1F,$DD,$8F,$F6,$BF,$A5,$FD,$CF,$E4
	.dc.b	$BF,$B5,$46,$E7,$55,$FA,$73,$8F,$3F,$F7,$48,$CF,$07,$E5,$1B,$5F
	.dc.b	$2B,$C0,$FA,$9F,$11,$F5,$7B,$64,$FC,$58,$CB,$F1,$E3,$2F,$F9,$A2
	.dc.b	$FE,$CC,$FE,$3F,$C5,$FE,$EB,$F5,$7F,$EC,$3A,$FB,$01,$C3,$F8,$78
	.dc.b	$E0,$FE,$3D,$FE,$E3,$D1,$FC,$5E,$3F,$21,$18,$36,$F1,$3A,$CF,$FB
	.dc.b	$C1,$FE,$79,$F2,$DF,$CE,$FE,$5A,$46,$08,$7F,$34,$72,$FF,$36,$8F
	.dc.b	$F0,$DF,$D4,$FF,$74,$8F,$E8,$DE,$3F,$F4,$53,$3E,$9C,$8F,$B1,$3B
	.dc.b	$7F,$C3,$6D,$E1,$FE,$35,$D5,$FC,$9E,$7E,$1E,$9F,$E2,$B7,$E5,$1F
	.dc.b	$EB,$3C,$FF,$D1,$4F,$B3,$E1,$3F,$75,$E1,$48,$FD,$A3,$F1,$88,$E0
	.dc.b	$EA,$ED,$EA,$7D,$1C,$BF,$C5,$3F,$56,$F9,$6F,$F1,$F7,$D5,$F8,$2B
	.dc.b	$C7,$9A,$31,$A7,$F3,$39,$5F,$17,$EE,$B5,$EE,$D8,$E0,$DE,$3F,$F6
	.dc.b	$85,$F5,$F9,$DF,$9F,$F5,$1F,$E4,$FF,$41,$14,$CE,$3F,$3A,$6B,$F3
	.dc.b	$9E,$1F,$35,$F9,$FF,$8E,$FF,$48,$67,$FA,$A6,$F5,$FF,$AC,$AF,$FC
	.dc.b	$71,$39,$6C,$65,$FF,$29,$B6,$3F,$58,$DC,$FF,$D5,$7C,$3F,$E5,$C2
	.dc.b	$57,$F1,$2F,$83,$F8,$F3,$FE,$B0,$CF,$33,$F3,$27,$2F,$FB,$4A,$BF
	.dc.b	$CC,$B3,$FB,$7E,$8F,$39,$27,$CA,$D7,$CA,$ED,$F0,$67,$FE,$82,$FA
	.dc.b	$9F,$E5,$B1,$FE,$8C,$F8,$74,$62,$3F,$D3,$51,$8F,$1F,$FD,$6C,$D8
	.dc.b	$FF,$2E,$35,$FC,$AB,$75,$FE,$E8,$DD,$3F,$D9,$BC,$9E,$1B,$97,$FC
	.dc.b	$DE,$B4,$FF,$BC,$27,$2F,$4C,$3C,$F8,$74,$6E,$1D,$27,$94,$7F,$98
	.dc.b	$1F,$4B,$50,$BF,$7F,$8D,$EC,$9D,$4E,$5F,$E7,$F5,$C2,$B7,$3D,$AB
	.dc.b	$FD,$2F,$EF,$FF,$A3,$BE,$33,$FD,$1B,$8F,$6F,$AA,$FC,$AD,$F8,$25
	.dc.b	$E0,$FF,$B1,$67,$32,$9C,$BC,$78,$3F,$D9,$51,$D2,$13,$C2,$31,$E5
	.dc.b	$DB,$FF,$68,$36,$BD,$EE,$FC,$36,$38,$C7,$AC,$62,$BA,$C7,$82,$FC
	.dc.b	$CD,$8F,$21,$EB,$9E,$18,$F5,$BC,$CF,$AE,$F8,$BF,$05,$AF,$BE,$B8
	.dc.b	$DF,$8A,$BC,$93,$B2,$9C,$CB,$8F,$C0,$9E,$11,$E8,$38,$14,$E0,$58
	.dc.b	$F9,$58,$79,$E3,$18,$79,$27,$E8,$C6,$F0,$F2,$BE,$FC,$3D,$B3,$AB
	.dc.b	$F9,$D3,$B9,$CD,$7C,$BC,$72,$4C,$7C,$EF,$86,$F8,$92,$3D,$6F,$96
	.dc.b	$38,$5F,$B1,$C4,$F2,$3E,$A4,$7F,$88,$DB,$1A,$8F,$40,$E0,$FF,$0F
	.dc.b	$E3,$F8,$B6,$E1,$F8,$16,$E3,$FA,$67,$FC,$4B,$5F,$A8,$FA,$BC,$61
	.dc.b	$FE,$DB,$6C,$7E,$B0,$D1,$DE,$34,$D3,$42,$3E,$EB,$FE,$4F,$DE,$B9
	.dc.b	$3D,$87,$5F,$43,$FC,$FD,$5F,$A8,$8E,$ED,$8C,$E7,$E0,$8C,$57,$4B
	.dc.b	$FC,$AD,$F7,$3F,$0C,$7A,$34,$DB,$83,$C5,$58,$3C,$0E,$2F,$D7,$87
	.dc.b	$82,$3D,$B7,$DA,$3F,$D2,$97,$C2,$31,$87,$8A,$BD,$D3,$E0,$34,$F9
	.dc.b	$2B,$8D,$FC,$A9,$D3,$4F,$0D,$73,$BF,$97,$5E,$3B,$FA,$91,$89,$F8
	.dc.b	$51,$D5,$E0,$D3,$E0,$47,$99,$A6,$6F,$CB,$87,$C3,$C3,$E7,$1E,$29
	.dc.b	$FA,$E7,$80,$E0,$F9,$DF,$B5,$7F,$C2,$B8,$3F,$A7,$F9,$9F,$E5,$38
	.dc.b	$FE,$5D,$C4,$8C,$53,$B1,$D0,$FA,$5C,$E3,$CB,$1F,$E8,$38,$7F,$84
	.dc.b	$70,$E6,$D7,$F9,$E3,$67,$84,$7F,$61,$C7,$FD,$E1,$5F,$D2,$BC,$CF
	.dc.b	$0E,$5E,$0E,$DE,$F5,$AB,$FE,$A9,$58,$37,$9A,$B1,$87,$32,$57,$F7
	.dc.b	$D8,$CB,$CA,$3F,$B5,$DF,$58,$FE,$B3,$7E,$78,$C4,$7F,$9D,$7E,$7F
	.dc.b	$E2,$F1,$FF,$A0,$9F,$0F,$D4,$63,$B9,$14,$F8,$53,$BC,$7F,$C3,$71
	.dc.b	$89,$DC,$7E,$DA,$FE,$86,$6F,$AB,$F6,$9E,$EF,$AB,$B5,$FE,$1B,$8B
	.dc.b	$5F,$E1,$99,$B8,$FF,$A3,$FC,$A7,$CF,$FA,$1B,$87,$13,$FD,$38,$E2
	.dc.b	$FF,$CB,$F4,$8C,$47,$F0,$F8,$C5,$7B,$27,$33,$FE,$EC,$6B,$FD,$83
	.dc.b	$EA,$1F,$57,$F2,$1B,$87,$FB,$13,$C6,$6F,$9A,$93,$9D,$2E,$1F,$E6
	.dc.b	$0B,$F8,$93,$FE,$98,$F1,$BF,$DB,$2D,$F5,$BF,$13,$F5,$6F,$95,$FC
	.dc.b	$DE,$0B,$DB,$FE,$70,$6E,$FF,$E8,$2A,$7C,$4F,$09,$CE,$3F,$E6,$49
	.dc.b	$CB,$FE,$A8,$E9,$FD,$23,$7A,$FF,$53,$E0,$3E,$AD,$E5,$78,$AF,$7A
	.dc.b	$7F,$A0,$1A,$FF,$90,$EF,$F5,$0F,$F8,$9D,$B1,$A8,$FF,$22,$C7,$E4
	.dc.b	$BE,$D5,$F3,$FD,$63,$0F,$CF,$E0,$99,$FC,$FF,$85,$B8,$7D,$EE,$6F
	.dc.b	$6B,$FB,$FE,$37,$C1,$6B,$8F,$FD,$4B,$18,$C3,$FD,$F7,$EF,$BE,$B3
	.dc.b	$F6,$EB,$E3,$6C,$61,$B3,$C1,$B8,$FF,$A6,$AB,$84,$FC,$BD,$7E,$31
	.dc.b	$F2,$BD,$A6,$38,$9D,$1F,$E6,$49,$F8,$A4,$70,$7F,$94,$32,$3F,$E5
	.dc.b	$2E,$4F,$A9,$E0,$F3,$CB,$FD,$7B,$18,$9E,$07,$FA,$8B,$E0,$5E,$F7
	.dc.b	$F1,$45,$F5,$96,$30,$D9,$E3,$7F,$CA,$6B,$71,$8F,$21,$1D,$23,$0E
	.dc.b	$A7,$F6,$3E,$0F,$FA,$BA,$7F,$6B,$E5,$78,$7E,$B3,$F7,$4C,$8E,$9F
	.dc.b	$02,$F6,$31,$87,$FB,$1A,$31,$3D,$BE,$B3,$D5,$E0,$FA,$DB,$19,$3C
	.dc.b	$71,$96,$99,$AF,$F6,$8E,$7F,$AB,$8C,$70,$FF,$BA,$A3,$13,$AF,$AC
	.dc.b	$F9,$1F,$0F,$F2,$5B,$7F,$01,$B3,$B8,$F8,$73,$F3,$F9,$BE,$87,$83
	.dc.b	$67,$E3,$B6,$BC,$6D,$C7,$C4,$FE,$BF,$B6,$57,$CE,$F9,$DB,$7F,$B1
	.dc.b	$9F,$D6,$92,$7F,$D5,$ED,$8F,$C2,$FE,$C1,$FE,$9A,$9F,$E0,$13,$FB
	.dc.b	$E8,$C4,$FE,$6F,$E9,$7F,$46,$7E,$1F,$A6,$DF,$E1,$25,$F2,$AA,$FD
	.dc.b	$FB,$67,$FA,$14,$EA,$72,$7F,$6B,$E7,$FC,$4F,$3F,$E4,$CF,$D8,$FC
	.dc.b	$5D,$BF,$C1,$DF,$F2,$0C,$13,$FD,$B3,$97,$CF,$BF,$E9,$2B,$FD,$6B
	.dc.b	$5A,$BF,$33,$63,$FA,$A3,$5F,$F2,$5B,$5F,$F2,$93,$5F,$F0,$45,$7F
	.dc.b	$8F,$EB,$B6,$28,$C5,$F0,$C3,$C9,$D7,$FC,$DA,$3F,$49,$87,$FA,$A1
	.dc.b	$FF,$2F,$18,$C3,$D0,$BE,$B1,$F0,$3C,$7B,$64,$FC,$A3,$9D,$FF,$98
	.dc.b	$9A,$F4,$8F,$E5,$F1,$D4,$FB,$DF,$03,$FD,$85,$FB,$06,$67,$F8,$5C
	.dc.b	$65,$F3,$1F,$ED,$EF,$AD,$FE,$67,$3F,$7D,$B1,$A8,$F4,$4F,$F5,$49
	.dc.b	$CE,$3F,$4E,$91,$C1,$FF,$3C,$B5,$FF,$40,$35,$F0,$78,$1F,$D2,$23
	.dc.b	$9D,$FE,$5F,$27,$C0,$F7,$AF,$F1,$EC,$FC,$98,$FE,$CB,$ED,$FF,$67
	.dc.b	$8F,$CB,$37,$0F,$E6,$DF,$5B,$AF,$46,$57,$E5,$AB,$55,$FE,$8E,$6B
	.dc.b	$FE,$E4,$9F,$EB,$71,$CF,$3F,$C1,$F8,$CF,$8B,$E7,$7F,$D0,$DF,$1F
	.dc.b	$A5,$31,$FE,$3B,$EB,$5A,$F9,$E3,$1F,$EE,$6F,$F8,$A5,$AF,$8E,$BC
	.dc.b	$24,$AF,$E6,$BF,$5A,$FA,$97,$D6,$F5,$D3,$F4,$93,$98,$FF,$04,$FD
	.dc.b	$6B,$C0,$A7,$F9,$8B,$E6,$1F,$13,$8C,$BF,$7D,$5B,$7F,$76,$DD,$3F
	.dc.b	$C7,$CD,$7F,$D6,$AD,$7F,$D7,$8E,$F2,$FD,$B0,$FB,$07,$CB,$F7,$AF
	.dc.b	$F1,$BC,$62,$FF,$87,$7D,$77,$F6,$86,$FA,$5E,$7F,$84,$FF,$11,$CE
	.dc.b	$6F,$E5,$B5,$FF,$63,$B5,$FF,$67,$35,$07,$1E,$17,$EF,$7E,$3F,$FC
	.dc.b	$5F,$F4,$8F,$8E,$FE,$21,$18,$6E,$5F,$C8,$E7,$4F,$F5,$A9,$D3,$FD
	.dc.b	$AB,$57,$D7,$7F,$25,$CB,$FC,$AA,$D8,$FE,$1B,$F0,$DF,$A0,$94,$70
	.dc.b	$F2,$BF,$CE,$F5,$BB,$FF,$02,$36,$BF,$CD,$FE,$06,$AB,$FD,$3B,$F5
	.dc.b	$DF,$DA,$27,$F7,$BF,$02,$8C,$63,$FC,$73,$AB,$FA,$FB,$FC,$AE,$75
	.dc.b	$EB,$F2,$26,$7F,$E0,$1A,$DB,$F2,$A7,$6F,$F8,$21,$B8,$7F,$BE,$1A
	.dc.b	$FF,$BF,$DA,$F1,$C3,$CC,$E7,$EB,$BC,$B5,$D5,$B1,$FE,$8C,$FA,$EF
	.dc.b	$F5,$CC,$62,$CD,$37,$C5,$FE,$C0,$9D,$E5,$FE,$6A,$FA,$D7,$D1,$3C
	.dc.b	$DF,$EA,$F1,$86,$C6,$72,$FF,$4E,$4E,$5F,$F4,$F7,$CA,$7F,$8B,$3E
	.dc.b	$BB,$FA,$FB,$63,$3E,$72,$72,$D8,$E7,$E4,$BE,$53,$C2,$D8,$CC,$60
	.dc.b	$FF,$56,$6E,$F5,$FF,$97,$72,$7E,$23,$EB,$BF,$D3,$9C,$7F,$9C,$E2
	.dc.b	$FF,$6F,$C6,$31,$61,$C6,$B0,$D8,$EB,$87,$F9,$D8,$93,$FD,$EA,$55
	.dc.b	$FC,$26,$BF,$C8,$5D,$2F,$AA,$ED,$8F,$FB,$B9,$EB,$5B,$AF,$C8,$B7
	.dc.b	$2F,$CF,$CF,$0B,$E5,$5F,$89,$AD,$63,$F8,$D7,$94,$7E,$25,$AE,$7C
	.dc.b	$6F,$34,$62,$3F,$53,$F5,$AF,$AD,$F8,$59,$23,$13,$FB,$56,$BF,$D8
	.dc.b	$B3,$64,$FA,$1E,$35,$AB,$FE,$36,$DC,$FF,$CE,$CD,$C3,$FD,$C3,$5F
	.dc.b	$13,$D0,$7D,$6F,$D3,$97,$84,$F9,$BC,$36,$3F,$D2,$B3,$C4,$CC,$C8
	.dc.b	$C9,$FE,$33,$FA,$DF,$92,$FF,$4C,$F5,$7F,$CD,$4F,$F8,$DD,$AF,$F5
	.dc.b	$08,$DE,$9F,$DD,$DB,$A7,$F2,$49,$FE,$DF,$DB,$E7,$BE,$37,$BE,$3F
	.dc.b	$57,$D6,$7F,$8D,$2F,$FC,$89,$F5,$BF,$D2,$9B,$E7,$EB,$DC,$91,$82
	.dc.b	$1F,$D8,$AB,$93,$63,$FC,$92,$A3,$2F,$94,$C4,$6A,$33,$1E,$9F,$AC
	.dc.b	$FF,$5D,$3F,$E9,$AF,$7D,$F5,$BF,$EE,$48,$C3,$5F,$EE,$73,$C1,$BF
	.dc.b	$1D,$FA,$9A,$E3,$EC,$5B,$E1,$AF,$F8,$7F,$0E,$0D,$BF,$BD,$18,$E0
	.dc.b	$F2,$47,$1F,$5F,$3E,$37,$EA,$3E,$A5,$F5,$34,$CE,$BF,$E6,$FC,$7F
	.dc.b	$8B,$BF,$EB,$19,$DB,$67,$6F,$AA,$AE,$5E,$57,$B7,$A9,$F9,$13,$FC
	.dc.b	$31,$3F,$E7,$BF,$AD,$FF,$61,$B7,$E4,$75,$EF,$C8,$C1,$0F,$F0,$CD
	.dc.b	$72,$6C,$6D,$A7,$5F,$1F,$F4,$1E,$E3,$E1,$BF,$45,$F5,$5F,$EE,$4A
	.dc.b	$FF,$46,$35,$FF,$12,$CE,$DF,$F4,$83,$5F,$F1,$4C,$6F,$4F,$F2,$83
	.dc.b	$78,$BF,$35,$5B,$9F,$F2,$A3,$63,$37,$E4,$3C,$7F,$27,$FB,$7F,$AD
	.dc.b	$98,$FF,$B7,$FC,$34,$7E,$E7,$EB,$3F,$8A,$CF,$F8,$29,$BE,$42,$B7
	.dc.b	$3C,$6B,$87,$97,$FC,$0F,$EB,$6C,$FF,$AA,$F0,$E3,$69,$FE,$11,$F0
	.dc.b	$32,$BC,$4F,$E1,$FC,$6F,$17,$CB,$D9,$FC,$24,$61,$FC,$37,$A5,$F8
	.dc.b	$D2,$3E,$BF,$AA,$FE,$25,$F5,$9F,$B0,$E3,$FF,$55,$7B,$0F,$93,$FE
	.dc.b	$9B,$C9,$FE,$26,$8C,$3C,$3C,$69,$8F,$F8,$AA,$F8,$DF,$08,$FE,$D8
	.dc.b	$D5,$87,$7B,$FE,$1C,$D8,$DC,$7F,$57,$C3,$BC,$7F,$B3,$E7,$5C,$6F
	.dc.b	$95,$FE,$CD,$3F,$EF,$AF,$89,$FE,$A3,$1F,$E6,$B8,$F2,$61,$FD,$8E
	.dc.b	$37,$97,$FB,$F7,$AB,$77,$2F,$63,$0E,$D3,$D3,$C9,$EB,$F4,$F2,$B7
	.dc.b	$5E,$39,$D4,$63,$15,$6E,$3F,$A6,$47,$F2,$56,$F8,$5E,$8E,$F9,$1C
	.dc.b	$63,$D6,$61,$E3,$18,$74,$77,$75,$9F,$FB,$5A,$BF,$82,$CF,$FA,$F2
	.dc.b	$1F,$4A,$E3,$FE,$EF,$F2,$5F,$D7,$6B,$D6,$D7,$84,$ED,$FF,$2D,$35
	.dc.b	$FE,$CD,$1B,$AF,$F7,$2B,$77,$FC,$25,$6E,$7C,$9F,$0F,$F5,$B9,$A6
	.dc.b	$1F,$1E,$3E,$C9,$87,$E3,$C6,$1F,$C1,$78,$5F,$DA,$7E,$B3,$FD,$87
	.dc.b	$1F,$60,$38,$7F,$D5,$BC,$9F,$C5,$70,$FF,$6A,$37,$2F,$EB,$D5,$B9
	.dc.b	$FF,$6C,$36,$3F,$2B,$7F,$A6,$3C,$1E,$8C,$C3,$C6,$7F,$33,$F9,$7F
	.dc.b	$E7,$5F,$2F,$23,$04,$3F,$C1,$9E,$52,$35,$18,$9F,$F0,$AF,$D6,$BE
	.dc.b	$2F,$FA,$27,$8F,$2B,$FD,$15,$39,$6C,$E5,$EB,$3C,$A3,$F7,$BD,$FF
	.dc.b	$1F,$D6,$95,$FA,$B8,$CB,$F2,$76,$FE,$FB,$B7,$C4,$AC,$CF,$CA,$AE
	.dc.b	$2F,$FB,$62,$B6,$FF,$AA,$B8,$13,$FE,$1D,$FA,$C7,$CB,$7F,$91,$3E
	.dc.b	$B7,$7D,$C7,$8E,$0A,$DB,$F6,$DA,$FF,$8C,$E3,$89,$C2,$33,$C9,$FE
	.dc.b	$CC,$DD,$F5,$7F,$E5,$2F,$91,$7C,$FD,$3E,$6C,$4C,$7F,$73,$3A,$6C
	.dc.b	$E5,$FD,$0F,$CF,$7F,$95,$7C,$D9,$1F,$A2,$8F,$F2,$5F,$61,$CC,$7F
	.dc.b	$92,$7E,$0F,$FA,$2E,$7F,$B0,$9F,$F2,$F7,$1C,$AF,$F3,$4B,$FE,$61
	.dc.b	$C7,$11,$C6,$2D,$7F,$B1,$27,$BF,$47,$F9,$9F,$CA,$97,$FB,$48,$CB
	.dc.b	$FE,$74,$8C,$BF,$BA,$8F,$F3,$FF,$D5,$FF,$22,$E7,$FF,$43,$66,$FB
	.dc.b	$6F,$F3,$91,$AF,$22,$61,$FE,$7E,$8C,$BF,$EB,$DB,$E4,$FF,$A3,$1B
	.dc.b	$F1,$06,$9A,$E6,$FF,$35,$F3,$F4,$AF,$F6,$9C,$EB,$0F,$F4,$14,$F9
	.dc.b	$BC,$9F,$F4,$5F,$C4,$7F,$4B,$C3,$FD,$23,$A6,$71,$F1,$56,$A2,$46
	.dc.b	$0F,$E0,$9F,$19,$FE,$6D,$E1,$FF,$7F,$FD,$6F,$DF,$94,$FF,$BF,$5C
	.dc.b	$F1,$7F,$98,$63,$10,$9E,$11,$CE,$31,$3B,$F1,$7F,$DE,$F1,$80,$76
	.dc.b	$F1,$BF,$CD,$97,$C9,$B5,$C7,$0F,$0D,$70,$C3,$B5,$EE,$32,$73,$2C
	.dc.b	$6A,$B9,$47,$84,$D1,$63,$D7,$3E,$AC,$3C,$ED,$BE,$78,$79,$2F,$C4
	.dc.b	$D7,$E3,$D7,$0C,$3B,$F9,$BE,$59,$34,$E8,$E2,$3C,$B8,$6C,$8E,$09
	.dc.b	$1E,$03,$C6,$7B,$AB,$C0,$FC,$56,$BE,$E9,$E6,$E7,$0E,$07,$27,$E1
	.dc.b	$61,$EB,$C7,$E4,$46,$CC,$9B,$AF,$49,$30,$FC,$25,$71,$39,$61,$E2
	.dc.b	$BF,$C2,$6F,$FE,$11,$C3,$E3,$E3,$F8,$28,$C4,$EB,$2F,$AF,$1B,$AF
	.dc.b	$C3,$FD,$76,$E7,$FD,$3D,$08,$C3,$FE,$A1,$6B,$FE,$B7,$6B,$C8,$FC
	.dc.b	$10,$F3,$DA,$FE,$2A,$35,$AF,$E8,$DD,$46,$5F,$01,$CA,$78,$BD,$0C
	.dc.b	$CF,$18,$CB,$CE,$33,$AF,$FA,$DE,$72,$FB,$2F,$C5,$18,$9F,$61,$DE
	.dc.b	$FC,$53,$A6,$C6,$4F,$14,$65,$A6,$6F,$C5,$19,$6C,$7E,$4C,$FF,$6F
	.dc.b	$57,$B3,$0C,$9E,$BB,$FD,$CB,$C2,$B1,$5D,$9A,$67,$0F,$75,$79,$A3
	.dc.b	$31,$EF,$9F,$23,$63,$31,$F0,$1A,$FD,$B3,$66,$63,$AB,$E1,$6C,$7F
	.dc.b	$59,$E3,$FE,$61,$E7,$FD,$6D,$7E,$BF,$0B,$89,$EB,$E7,$3F,$E2,$B8
	.dc.b	$70,$7F,$CB,$73,$A7,$F9,$F6,$1B,$CF,$FC,$C9,$39,$7F,$9C,$C6,$70
	.dc.b	$ED,$86,$F0,$FE,$85,$E7,$DE,$63,$BE,$8F,$EE,$F9,$FF,$4B,$8D,$11
	.dc.b	$FF,$36,$C6,$B2,$FE,$8F,$39,$5A,$7B,$B4,$7B,$F9,$8C,$37,$0F,$E6
	.dc.b	$7A,$7C,$38,$C5,$FB,$1F,$75,$7F,$99,$BE,$83,$D9,$83,$DE,$CF,$80
	.dc.b	$9C,$C6,$1B,$97,$F9,$99,$B8,$FF,$80,$FE,$AF,$E0,$1F,$07,$EB,$AD
	.dc.b	$7B,$A2,$CF,$F8,$26,$6E,$5F,$6F,$CD,$7C,$F7,$93,$E4,$7C,$F1,$8C
	.dc.b	$7E,$CC,$62,$75,$1F,$E9,$CA,$FF,$78,$46,$30,$FF,$42,$7D,$3B,$EB
	.dc.b	$7E,$D7,$92,$D3,$E1,$CE,$71,$FF,$13,$C6,$2B,$6D,$8C,$9F,$E7,$3F
	.dc.b	$21,$F5,$3F,$85,$FA,$D7,$C5,$7E,$2F,$3F,$F1,$44,$E6,$31,$5C,$2F
	.dc.b	$FD,$5B,$7C,$1E,$95,$B7,$FD,$01,$D4,$FA,$BF,$B3,$F5,$8F,$9F,$ED
	.dc.b	$39,$AF,$85,$39,$78,$D7,$0A,$DB,$63,$27,$08,$CB,$4F,$99,$B7,$9F
	.dc.b	$4F,$C9,$C7,$E3,$B7,$79,$BF,$2B,$3F,$9A,$D9,$FF,$28,$B6,$B8,$9F
	.dc.b	$ED,$1A,$E3,$3F,$EC,$EF,$84,$FD,$87,$1F,$EF,$1F,$CF,$C6,$7C,$0F
	.dc.b	$D9,$46,$2F,$9F,$6A,$38,$D7,$E8,$9D,$5F,$3B,$E3,$DB,$F2,$35,$FE
	.dc.b	$01,$E2,$BE,$B7,$F4,$18,$FD,$09,$CC,$63,$3F,$48,$E2,$F9,$CF,$B2
	.dc.b	$F9,$76,$75,$C3,$FD,$DF,$18,$7A,$18,$BE,$99,$FF,$50,$BE,$98,$F8
	.dc.b	$30,$FE,$21,$F0,$1F,$ED,$BC,$3F,$DE,$1F,$15,$8F,$81,$CB,$63,$27
	.dc.b	$FB,$D6,$32,$FE,$97,$BB,$FD,$0E,$FF,$B8,$BD,$3F,$D6,$F9,$3F,$C7
	.dc.b	$7E,$05,$18,$9D,$1C,$67,$51,$E4,$FA,$CF,$2F,$6D,$AF,$F2,$1E,$FF
	.dc.b	$EF,$58,$C6,$FF,$5B,$5F,$F4,$55,$68,$FE,$2E,$AF,$9C,$66,$5C,$CA
	.dc.b	$D4,$E6,$31,$7D,$B0,$F0,$4E,$69,$F5,$9F,$2B,$E1,$7E,$5B,$67,$4D
	.dc.b	$C7,$CC,$DB,$FE,$9B,$6F,$81,$FA,$7E,$C7,$D4,$6E,$1D,$DB,$87,$BD
	.dc.b	$B7,$F7,$79,$DE,$CF,$E4,$3E,$B3,$FC,$D3,$87,$EF,$A3,$18,$7E,$F1
	.dc.b	$B1,$93,$F1,$73,$C2,$7F,$27,$5D,$27,$31,$C5,$A7,$EB,$7E,$8F,$F5
	.dc.b	$27,$E3,$39,$EC,$7C,$F9,$7C,$9A,$BF,$5F,$F1,$3F,$C6,$E3,$15,$FB
	.dc.b	$78,$FC,$AF,$3C,$8C,$1F,$45,$FC,$65,$F1,$9B,$3D,$E9,$19,$8F,$B9
	.dc.b	$CF,$FE,$36,$AF,$D7,$D6,$A7,$3D,$9F,$97,$23,$CE,$ED,$7E,$DA,$3F
	.dc.b	$50,$DE,$9F,$EF,$D0,$FA,$A7,$D6,$7F,$37,$8D,$F0,$FE,$CD,$A7,$D6
	.dc.b	$F7,$9A,$EF,$0E,$71,$89,$F9,$47,$0C,$7F,$CA,$71,$97,$C5,$1F,$9D
	.dc.b	$11,$9B,$FE,$05,$F5,$9F,$C3,$1B,$DF,$E8,$CF,$77,$F5,$EB,$4F,$FA
	.dc.b	$CB,$04,$61,$FF,$31,$CB,$E2,$BF,$AA,$4E,$6B,$E1,$46,$2F,$B3,$63
	.dc.b	$CD,$5C,$5F,$C3,$7B,$4F,$47,$FA,$F3,$63,$F9,$D7,$D6,$62,$64,$7F
	.dc.b	$3D,$F8,$CB,$9F,$F3,$8F,$21,$18,$D7,$FA,$03,$6B,$F9,$D3,$DB,$D0
	.dc.b	$91,$F8,$6F,$4E,$E1,$FC,$D3,$3F,$ED,$4F,$F4,$EF,$AC,$7D,$43,$E6
	.dc.b	$37,$9F,$F9,$FA,$73,$18,$C7,$FD,$B7,$18,$7D,$F9,$7F,$91,$63,$18
	.dc.b	$7F,$42,$FA,$CF,$D8,$3F,$E9,$A6,$BF,$4C,$38,$7E,$B3,$5F,$F7,$4B
	.dc.b	$CB,$0F,$F2,$E7,$D6,$3E,$4F,$FC,$15,$39,$C7,$F9,$6C,$E5,$FD,$C4
	.dc.b	$66,$FF,$C8,$F3,$97,$CD,$E1,$7F,$90,$23,$18,$7F,$4F,$FA,$C7,$A5
	.dc.b	$FD,$8B,$E0,$1F,$05,$FE,$19,$9C,$E7,$FD,$6A,$72,$FF,$07,$8C,$E5
	.dc.b	$FD,$86,$72,$FF,$66,$8C,$E1,$FE,$AC,$F0,$15,$A8,$FF,$3C,$F8,$4F
	.dc.b	$92,$FF,$14,$7D,$13,$E2,$BF,$AA,$46,$1B,$9F,$F1,$79,$D3,$FC,$22
	.dc.b	$74,$FF,$90,$FE,$B3,$F8,$BE,$9F,$E1,$0C,$51,$87,$FD,$13,$97,$E9
	.dc.b	$DC,$F8,$FF,$E6,$69,$D5,$FF,$5C,$6D,$FF,$54,$BE,$11,$8F,$25,$F5
	.dc.b	$9F,$DA,$AF,$FD,$BB,$18,$F5,$3F,$BF,$61,$FE,$94,$AD,$3F,$87,$9D
	.dc.b	$57,$09,$CE,$5F,$E8,$1E,$8F,$F4,$84,$62,$7F,$CF,$FF,$0B,$FA,$0A
	.dc.b	$FF,$6C,$DF,$2C,$3F,$C4,$B3,$9A,$F0,$63,$FE,$80,$C7,$FD,$35,$F5
	.dc.b	$9E,$2B,$FE,$15,$82,$31,$1F,$EA,$A6,$BF,$E2,$16,$C6,$EB,$FD,$17
	.dc.b	$D7,$D1,$B8,$7F,$4D,$FA,$BF,$F4,$77,$27,$FC,$00,$C5,$18,$7F,$50
	.dc.b	$63,$E1,$BF,$DF,$8E,$23,$4F,$EB,$7E,$35,$E5,$BE,$AF,$A4,$61,$B1
	.dc.b	$99,$EB,$39,$7A,$78,$CF,$8E,$F2,$61,$FC,$75,$B1,$99,$DC,$7D,$CF
	.dc.b	$BE,$FF,$3A,$47,$F8,$5F,$74,$62,$3F,$CB,$F1,$8F,$85,$FE,$6D,$C9
	.dc.b	$FE,$BB,$FA,$BB,$C1,$FE,$D6,$C3,$84,$FD,$A6,$C6,$6B,$78,$7F,$B5
	.dc.b	$AE,$E1,$E0,$FA,$FB,$81,$A7,$0F,$7B,$FD,$CD,$F6,$F8,$C3,$D1,$5B
	.dc.b	$6D,$6F,$A1,$5B,$9F,$5D,$6D,$B5,$B9,$F2,$E8,$FC,$6F,$FC,$51,$5F
	.dc.b	$EB,$56,$BF,$B3,$6C,$7F,$B7,$7B,$78,$C6,$36,$FF,$25,$B6,$BF,$DC
	.dc.b	$6D,$CF,$FE,$00,$57,$B7,$C1,$7C,$15,$F5,$E6,$1F,$E0,$C8,$D1,$87
	.dc.b	$FD,$13,$3F,$E0,$27,$47,$F8,$8F,$F3,$27,$D8,$7D,$08,$46,$1D,$47
	.dc.b	$D4,$E9,$7E,$57,$99,$FF,$43,$D5,$DB,$FC,$FC,$DD,$BF,$BC,$36,$7F
	.dc.b	$35,$55,$EF,$EE,$F8,$67,$FD,$E9,$1E,$58,$F0,$7C,$4F,$FB,$A2,$30
	.dc.b	$43,$C3,$1A,$F4,$F9,$32,$35,$8B,$14,$7E,$AE,$FA,$3F,$EA,$0F,$AB
	.dc.b	$FF,$1B,$BF,$B9,$6B,$FB,$86,$BF,$D8,$5A,$FF,$85,$5D,$D7,$FA,$7D
	.dc.b	$B5,$9F,$7B,$E7,$7D,$67,$EB,$EF,$C1,$1F,$46,$B7,$1E,$66,$A9,$E1
	.dc.b	$7F,$EA,$BF,$05,$F0,$7F,$67,$9F,$E6,$1F,$36,$33,$7F,$E6,$16,$BF
	.dc.b	$EB,$E6,$D7,$3A,$E5,$F0,$7F,$83,$9C,$E7,$FE,$31,$6D,$FC,$A6,$BF
	.dc.b	$AA,$9F,$C5,$7E,$4C,$9F,$CC,$7C,$0C,$AF,$F3,$94,$E5,$B3,$C0,$8F
	.dc.b	$FA,$87,$ED,$3F,$BB,$BA,$21,$FE,$D6,$9E,$4F,$FB,$9A,$36,$E7,$C3
	.dc.b	$FE,$09,$0F,$A6,$A7,$9A,$24,$F9,$1A,$FF,$1C,$A7,$D1,$E3,$5F,$F6
	.dc.b	$D4,$7F,$91,$67,$81,$FC,$95,$DB,$FE,$7E,$F8,$2F,$DC,$11,$FF,$49
	.dc.b	$7D,$73,$E8,$3F,$60,$D8,$CC,$60,$FE,$01,$E3,$C8,$C9,$0F,$DE,$BF
	.dc.b	$C1,$7E,$86,$47,$FA,$39,$FE,$57,$F0,$FF,$D0,$7E,$C7,$F7,$DF,$3B
	.dc.b	$FE,$F3,$D3,$FD,$E8,$DE,$4F,$F3,$73,$70,$FF,$37,$F9,$8F,$96,$E6
	.dc.b	$D8,$FF,$6D,$63,$FD,$36,$72,$78,$3F,$02,$FA,$78,$FF,$C0,$1F,$5B
	.dc.b	$E9,$5F,$FA,$CD,$AA,$35,$C9,$FC,$4F,$EA,$1F,$19,$FC,$56,$F7,$5A
	.dc.b	$BE,$57,$BA,$D3,$23,$27,$FB,$06,$F9,$BF,$EB,$2F,$37,$F9,$67,$CE
	.dc.b	$48,$C1,$0F,$E5,$2E,$5F,$E5,$71,$FC,$CF,$E7,$BF,$9A,$61,$27,$FD
	.dc.b	$7B,$19,$7F,$D7,$F1,$97,$F9,$9C,$7F,$3B,$FB,$5F,$E7,$9E,$94,$9F
	.dc.b	$F6,$2C,$65,$FF,$63,$C6,$5F,$E7,$71,$97,$F9,$EC,$67,$3F,$F2,$7C
	.dc.b	$FF,$56,$DB,$FA,$ED,$7F,$20,$9F,$E4,$4F,$CF,$8C,$BF,$E4,$96,$C7
	.dc.b	$F9,$63,$EC,$B9,$78,$52,$B9,$72,$73,$BD,$BF,$CB,$3D,$17,$9C,$C3
	.dc.b	$FA,$C7,$C3,$FF,$59,$F9,$C9,$8F,$FB,$8A,$32,$FF,$B9,$23,$2F,$F5
	.dc.b	$88,$FE,$A9,$F9,$27,$F8,$1A,$3F,$9C,$B5,$37,$8D,$D7,$C3,$B2,$57
	.dc.b	$FC,$01,$AD,$4E,$23,$F2,$1A,$7F,$B3,$67,$A1,$A8,$FF,$7B,$F6,$FF
	.dc.b	$83,$79,$7F,$49,$B7,$F8,$46,$FF,$BA,$4C,$AF,$8A,$27,$4E,$1A,$FF
	.dc.b	$74,$8F,$EE,$1E,$0F,$FA,$0E,$38,$9F,$C4,$3E,$17,$27,$F8,$77,$F5
	.dc.b	$CF,$EC,$B1,$E8,$3F,$AD,$CF,$17,$07,$13,$9B,$9E,$4F,$C0,$E0,$AE
	.dc.b	$71,$F8,$4A,$7D,$63,$E7,$FF,$C0,$8D,$8F,$2B,$5F,$17,$73,$F5,$4F
	.dc.b	$D6,$C7,$8A,$75,$7D,$5B,$AF,$FB,$ED,$B2,$BE,$5D,$9E,$0C,$64,$F8
	.dc.b	$1F,$F2,$2C,$EA,$3E,$7E,$1F,$C5,$7E,$4F,$F4,$04,$8F,$A1,$C0,$F4
	.dc.b	$71,$1F,$9F,$9B,$5F,$12,$B5,$3D,$3A,$9F,$63,$FA,$B2,$47,$4C,$AB
	.dc.b	$FA,$98,$B7,$CE,$7F,$65,$F6,$12,$BA,$3F,$E5,$1F,$4F,$11,$FB,$18
	.dc.b	$B5,$F5,$BC,$0F,$F0,$1F,$19,$A4,$BF,$F2,$D7,$C7,$F3,$C3,$F1,$D3
	.dc.b	$96,$C7,$F9,$63,$F7,$EF,$F0,$4F,$1F,$F0,$36,$CF,$F3,$16,$E1,$DF
	.dc.b	$E9,$F8,$79,$32,$BE,$7D,$6B,$15,$7F,$9D,$BE,$BD,$F1,$34,$97,$E4
	.dc.b	$F9,$AA,$FF,$A2,$27,$6F,$FA,$35,$AF,$FA,$F2,$37,$B7,$FB,$02,$72
	.dc.b	$DA,$FF,$3D,$4E,$B4,$FF,$1E,$FC,$2A,$F7,$18,$6E,$1F,$B5,$C9,$7E
	.dc.b	$9F,$06,$4F,$8D,$FF,$43,$73,$62,$3D,$11,$73,$FA,$D3,$FE,$69,$F8
	.dc.b	$87,$D6,$FF,$5B,$B7,$D6,$7F,$9E,$1F,$F2,$8C,$FE,$DE,$31,$08,$C7
	.dc.b	$C2,$7F,$95,$A3,$15,$CE,$B7,$5D,$B0,$FF,$2D,$C6,$3C,$2E,$0D,$BC
	.dc.b	$E1,$CE,$F8,$47,$92,$31,$87,$57,$57,$C9,$B3,$C8,$F4,$CE,$DB,$86
	.dc.b	$5F,$55,$71,$7D,$ED,$7D,$71,$BB,$F1,$E1,$E7,$9F,$21,$C0,$D5,$F7
	.dc.b	$78,$9A,$3B,$9C,$EF,$B5,$74,$C3,$DB,$86,$6A,$1C,$67,$4F,$A8,$C5
	.dc.b	$72,$39,$DF,$86,$77,$3E,$03,$67,$9A,$A3,$82,$61,$F6,$2B,$91,$CE
	.dc.b	$FB,$4F,$09,$D5,$79,$9F,$88,$62,$3D,$15,$F8,$DE,$37,$CD,$FB,$A7
	.dc.b	$FD,$FC,$D8,$CC,$7B,$5A,$FE,$37,$EB,$7F,$D4,$71,$88,$FF,$52,$C6
	.dc.b	$27,$F2,$51,$83,$A9,$88,$D3,$4F,$A8,$3E,$3B,$FC,$1F,$87,$E7,$A3
	.dc.b	$93,$63,$3A,$76,$95,$7E,$2D,$E5,$C5,$5C,$9D,$AB,$8B,$E1,$BE,$93
	.dc.b	$BA,$E5,$5A,$AF,$80,$71,$9D,$34,$F4,$9D,$AB,$9B,$8A,$E7,$3C,$27
	.dc.b	$27,$B6,$78,$5F,$C8,$BF,$04,$7C,$08,$C5,$71,$6C,$75,$3C,$55,$CE
	.dc.b	$FB,$C6,$63,$BC,$FC,$79,$CC,$6E,$78,$CE,$5A,$7C,$57,$BC,$6C,$E7
	.dc.b	$19,$3F,$88,$5F,$A1,$F5,$98,$BF,$E0,$5C,$7F,$CB,$7D,$1F,$F2,$5E
	.dc.b	$BF,$CD,$5B,$3F,$E4,$3E,$37,$AD,$8C,$18,$71,$AF,$E8,$17,$98,$C4
	.dc.b	$FF,$87,$7A,$7F,$B0,$5E,$79,$97,$7C,$E9,$8F,$F6,$0E,$FB,$8F,$F4
	.dc.b	$D9,$BA,$7F,$6B,$9B,$5F,$D0,$E6,$D7,$F4,$39,$B9,$FF,$49,$9C,$E7
	.dc.b	$FD,$C6,$77,$83,$AD,$F2,$FE,$D8,$C3,$EF,$AF,$EF,$1D,$F9,$AD,$E7
	.dc.b	$FD,$AA,$B0,$6F,$75,$62,$7F,$BD,$FC,$D6,$4D,$79,$56,$32,$4D,$BE
	.dc.b	$45,$FF,$B6,$23,$19,$FD,$6C,$9D,$3E,$33,$95,$FF,$92,$E7,$D9,$19
	.dc.b	$AF,$EF,$EE,$A3,$93,$EB,$F9,$09,$3F,$E2,$EF,$7E,$DF,$3C,$15,$8C
	.dc.b	$3F,$C5,$D5,$6F,$FC,$25,$EF,$FC,$53,$E4,$78,$EF,$7F,$13,$FE,$10
	.dc.b	$9B,$97,$E1,$6B,$53,$93,$FC,$D3,$DA,$F6,$BF,$21,$48,$C4,$FF,$9D
	.dc.b	$BC,$B7,$EA,$CC,$77,$5A,$BF,$F3,$5B,$74,$FF,$33,$37,$04,$FF,$96
	.dc.b	$6F,$94,$78,$B1,$E6,$47,$C1,$1A,$9E,$B3,$E6,$FD,$B1,$E4,$D3,$F1
	.dc.b	$5F,$AE,$FF,$3E,$C3,$FD,$0B,$3A,$6C,$ED,$FF,$47,$4E,$EF,$FD,$07
	.dc.b	$5B,$7E,$45,$FC,$EF,$94,$F9,$ED,$BF,$F3,$3B,$67,$71,$FE,$6A,$9F
	.dc.b	$8B,$CB,$23,$FD,$3E,$3E,$87,$F9,$BD,$7C,$89,$D3,$67,$73,$AD,$7E
	.dc.b	$5D,$69,$B3,$AF,$07,$0F,$FB,$37,$85,$F2,$9F,$86,$F7,$7F,$0E,$D9
	.dc.b	$FF,$3F,$B7,$2F,$C3,$37,$0F,$D1,$7C,$47,$EC,$CF,$C7,$FC,$95,$BF
	.dc.b	$F2,$EC,$5B,$FC,$74,$DA,$FF,$43,$7A,$3F,$96,$E2,$FC,$BB,$70,$FF
	.dc.b	$47,$B6,$FF,$2C,$DA,$E3,$3F,$E9,$4A,$E2,$D3,$FD,$2F,$E8,$FE,$8B
	.dc.b	$D0,$FD,$1B,$63,$27,$FA,$72,$32,$D3,$FD,$3F,$E7,$FE,$AB,$C9,$FD
	.dc.b	$5B,$63,$27,$FA,$92,$32,$D3,$FD,$A3,$F3,$EF,$BB,$FD,$4B,$63,$EC
	.dc.b	$FC,$E6,$26,$47,$07,$F9,$F3,$F8,$38,$CE,$BF,$ED,$4A,$E4,$DA,$DF
	.dc.b	$2A,$B7,$AB,$0F,$D6,$FC,$33,$F2,$E7,$C4,$7F,$4B,$8C,$36,$3F,$96
	.dc.b	$E5,$FE,$E5,$9C,$D7,$FB,$B3,$D2,$FC,$4B,$79,$78,$61,$FC,$1A,$73
	.dc.b	$7F,$C3,$7C,$CF,$E2,$67,$5F,$AA,$B8,$7F,$B3,$3C,$BE,$1D,$3F,$C6
	.dc.b	$1B,$1D,$E2,$DF,$F1,$79,$BA,$FF,$B3,$FC,$BF,$E3,$E6,$2F,$F3,$CF
	.dc.b	$F1,$8F,$35,$E4,$53,$FB,$36,$A9,$C5,$7F,$12,$BF,$F7,$D5,$F0,$8C
	.dc.b	$1F,$E8,$B7,$18,$A7,$E7,$3F,$DD,$2F,$A5,$70,$8F,$E3,$B5,$C2,$BA
	.dc.b	$E1,$FD,$77,$E0,$9D,$9E,$2B,$FE,$F1,$18,$C3,$94,$30,$FE,$6E,$F8
	.dc.b	$63,$C9,$E4,$F3,$EC,$F8,$0C,$8E,$35,$C3,$2F,$3A,$C3,$F8,$E4,$7C
	.dc.b	$0A,$E1,$87,$F4,$5F,$06,$CF,$FB,$F5,$FE,$87,$F8,$0B,$EB,$29,$FE
	.dc.b	$15,$FA,$87,$E0,$7F,$82,$BF,$8B,$82,$F8,$69,$EC,$6C,$7A,$CF,$E8
	.dc.b	$6E,$27,$E6,$DF,$4E,$7E,$F7,$C7,$1F,$F1,$07,$E0,$6F,$8E,$CE,$5C
	.dc.b	$FD,$64,$E6,$FE,$5F,$D3,$BE,$2F,$FA,$EB,$63,$EC,$53,$EE,$DF,$6F
	.dc.b	$AB,$95,$C9,$FF,$18,$4E,$A3,$04,$C7,$FC,$E9,$87,$56,$D7,$C4,$7F
	.dc.b	$DA,$D1,$97,$FD,$0F,$F0,$8F,$B1,$FE,$E2,$DB,$FC,$1F,$B1,$FD,$05
	.dc.b	$FF,$6F,$46,$6B,$E5,$07,$1C,$3F,$49,$B3,$FE,$E5,$AE,$55,$FD,$CA
	.dc.b	$72,$DB,$F0,$9F,$E0,$29,$DC,$75,$8F,$2F,$AD,$2F,$CB,$F4,$16,$BC
	.dc.b	$78,$7F,$77,$C7,$C5,$7C,$EB,$51,$F9,$4F,$41,$F0,$FF,$E4,$1F,$D8
	.dc.b	$BF,$C2,$19,$7E,$AF,$B1,$5A,$8C,$4E,$B3,$FF,$0E,$5F,$1F,$67,$FC
	.dc.b	$46,$DA,$FD,$A6,$2F,$DF,$DE,$9F,$F9,$45,$B3,$DF,$DE,$ED,$8F,$F8
	.dc.b	$F6,$72,$DA,$E0,$7F,$94,$70,$B7,$F8,$AF,$92,$FF,$37,$B7,$E0,$BF
	.dc.b	$CB,$9F,$BB,$F9,$F3,$FE,$13,$D2,$BD,$E5,$F9,$96,$CE,$9B,$87,$F0
	.dc.b	$CB,$E4,$FF,$8A,$AF,$96,$1F,$E8,$3A,$DE,$1F,$E8,$86,$F0,$30,$E3
	.dc.b	$39,$AE,$1F,$2B,$31,$FA,$15,$FE,$23,$CB,$C5,$39,$BF,$33,$EC,$F8
	.dc.b	$8C,$4F,$F0,$0F,$A6,$FE,$63,$EE,$7F,$B5,$6D,$18,$7F,$DB,$0D,$7F
	.dc.b	$DC,$ED,$7F,$C8,$A3,$E2,$FF,$CC,$7F,$47,$CD,$B1,$FE,$E1,$6B,$FE
	.dc.b	$F0,$6B,$CA,$7F,$CC,$9C,$9E,$36,$C7,$8B,$C8,$FF,$2F,$36,$7F,$DD
	.dc.b	$9C,$2F,$A6,$7C,$77,$93,$4F,$B3,$18,$BF,$E1,$2D,$7F,$D1,$7E,$36
	.dc.b	$F8,$7E,$B4,$E6,$B8,$DE,$6F,$FD,$7B,$39,$AF,$F6,$15,$E7,$0F,$17
	.dc.b	$57,$CB,$6D,$F9,$AF,$F4,$AD,$9D,$46,$23,$F9,$23,$71,$F4,$CF,$87
	.dc.b	$91,$D2,$F4,$BC,$17,$BC,$FD,$37,$83,$85,$6F,$1F,$D6,$FD,$33,$E7
	.dc.b	$FD,$99,$A3,$13,$A9,$E3,$E8,$6A,$73,$3F,$EA,$B6,$D7,$FB,$99,$B8
	.dc.b	$FF,$B9,$6B,$D1,$F3,$4F,$90,$BE,$0C,$C7,$F8,$44,$E5,$B3,$97,$FD
	.dc.b	$CD,$09,$E7,$3F,$4E,$5F,$48,$FA,$6F,$CD,$46,$1B,$19,$8F,$CF,$35
	.dc.b	$FC,$C9,$FE,$9E,$71,$9F,$F6,$46,$E5,$F9,$E6,$CB,$A5,$F4,$D8,$8F
	.dc.b	$1F,$D2,$FD,$7F,$15,$DD,$4F,$46,$09,$A7,$98,$95,$E6,$6B,$FD,$22
	.dc.b	$38,$3F,$E1,$06,$BF,$DD,$70,$7C,$BB,$E9,$BE,$37,$D3,$7F,$2A,$AF
	.dc.b	$F0,$BF,$D3,$7F,$68,$6F,$A3,$AF,$26,$47,$F0,$3A,$E0,$FC,$60,$E1
	.dc.b	$F9,$2D,$7F,$81,$BC,$38,$3F,$C1,$AD,$E0,$CD,$7D,$9C,$3F,$77,$7F
	.dc.b	$D9,$3D,$B7,$C6,$7D,$DE,$DF,$98,$7C,$D8,$CC,$FF,$8C,$1A,$F9,$5A
	.dc.b	$FD,$C6,$BF,$C5,$9A,$FF,$B7,$1E,$39,$7F,$88,$9B,$8E,$EF,$C5,$D5
	.dc.b	$F8,$FF,$A5,$FF,$2B,$E1,$F9,$8F,$A6,$FC,$E9,$9C,$24,$7F,$2B,$9D
	.dc.b	$BF,$25,$AF,$E4,$1A,$FF,$2A,$77,$8F,$F8,$D1,$B8,$E6,$BF,$81,$79
	.dc.b	$EF,$97,$F2,$78,$F8,$F6,$26,$5F,$E5,$4C,$53,$A8,$F2,$3F,$A6,$8C
	.dc.b	$C6,$08,$7F,$85,$9C,$B6,$3F,$8F,$43,$DD,$FD,$EF,$D3,$4A,$FD,$DB
	.dc.b	$FB,$5F,$A6,$7D,$1D,$3F,$6A,$E2,$7F,$A8,$35,$FC,$73,$5F,$F3,$E7
	.dc.b	$85,$FC,$BD,$08,$E1,$BF,$F9,$69,$BB,$FF,$9F,$2B,$87,$90,$7F,$CD
	.dc.b	$97,$FC,$43,$E0,$7F,$8A,$9F,$CF,$7E,$25,$E7,$3E,$3F,$F8,$77,$C2
	.dc.b	$FF,$90,$EF,$FA,$FB,$5F,$1B,$5F,$F5,$CE,$5F,$DA,$EF,$9F,$B1,$D7
	.dc.b	$0F,$F1,$4D,$7F,$A4,$67,$2D,$F9,$8F,$B8,$FF,$A3,$3E,$57,$FC,$D0
	.dc.b	$FF,$A3,$BE,$97,$FA,$0B,$7C,$FF,$E4,$53,$A8,$FF,$26,$B5,$FE,$3A
	.dc.b	$D7,$FB,$BB,$5F,$96,$D4,$F3,$E1,$FF,$42,$B7,$3E,$3C,$FF,$CC,$70
	.dc.b	$7D,$0F,$F3,$EF,$47,$F9,$FF,$B7,$32,$FE,$7B,$19,$7F,$9F,$C6,$5F
	.dc.b	$E7,$B1,$FD,$1B,$E9,$BF,$A3,$D4,$9F,$E8,$B1,$97,$FA,$3C,$65,$FE
	.dc.b	$8B,$1F,$D3,$BD,$BF,$E9,$FA,$C9,$FE,$9B,$19,$7F,$A7,$C6,$5F,$E9
	.dc.b	$B1,$97,$FA,$7C,$67,$5F,$F5,$9F,$2A,$BF,$B3,$F1,$7F,$6B,$BF,$F5
	.dc.b	$27,$33,$CE,$FE,$59,$18,$6E,$1F,$ED,$23,$FC,$D3,$E2,$72,$24,$FF
	.dc.b	$98,$2F,$A4,$F8,$3E,$7B,$FB,$4F,$C2,$7F,$6A,$F7,$E5,$FF,$68,$8C
	.dc.b	$BF,$DA,$A3,$2F,$F6,$88,$FE,$CD,$F4,$DF,$A9,$87,$1B,$8F,$ED,$1A
	.dc.b	$F7,$24,$FF,$90,$EB,$FC,$4C,$FF,$66,$6B,$FE,$40,$6B,$FB,$47,$B1
	.dc.b	$23,$F0,$DD,$BF,$E0,$2E,$FF,$F0,$D7,$37,$FB,$43,$5F,$F4,$67,$46
	.dc.b	$71,$FF,$01,$D6,$A2,$46,$0F,$F7,$7F,$82,$F9,$FF,$F1,$27,$D3,$7F
	.dc.b	$84,$27,$FD,$E4,$D7,$F8,$B4,$7F,$81,$E7,$84,$72,$8C,$1C,$BD,$4F
	.dc.b	$F0,$67,$D2,$FF,$AD,$67,$FD,$39,$F7,$1F,$CC,$9F,$F7,$F3,$54,$6D
	.dc.b	$FF,$5E,$B5,$FF,$06,$B5,$FF,$33,$B5,$FE,$36,$D7,$9F,$C3,$78,$88
	.dc.b	$FF,$20,$D1,$9F,$F2,$2F,$05,$F4,$FF,$B2,$AF,$C5,$7C,$E7,$F8,$47
	.dc.b	$CC,$91,$82,$1F,$EF,$BC,$3F,$D7,$73,$F9,$C3,$EA,$FA,$FF,$9E,$F9
	.dc.b	$6F,$F0,$D7,$8B,$2F,$F2,$D1,$96,$90,$FD,$1F,$27,$C3,$C3,$F4,$1D
	.dc.b	$AF,$95,$FF,$47,$47,$FB,$29,$B1,$9A,$FF,$07,$4E,$5B,$19,$7D,$0D
	.dc.b	$8E,$B3,$FE,$5E,$F5,$A5,$7F,$B1,$5A,$FE,$DB,$C5,$7C,$A6,$2F,$E5
	.dc.b	$CB,$C4,$FF,$36,$DE,$78,$BF,$A7,$C7,$F9,$C2,$FF,$CC,$FF,$1B,$FE
	.dc.b	$18,$BF,$F4,$73,$5F,$AC,$D5,$82,$30,$DF,$33,$99,$27,$FD,$B9,$F3
	.dc.b	$DF,$E8,$BF,$A4,$7D,$3F,$FA,$2D,$B3,$FE,$A0,$6B,$FE,$5A,$F8,$94
	.dc.b	$62,$3F,$D6,$B1,$89,$E8,$48,$FF,$75,$7D,$05,$7C,$FA,$FF,$B5,$EF
	.dc.b	$AE,$5F,$EE,$26,$C7,$C4,$DD,$7C,$6B,$58,$5A,$D4,$FA,$FC,$3F,$51
	.dc.b	$31,$F5,$35,$F0,$8C,$BD,$71,$C6,$3F,$E0,$07,$3F,$FA,$07,$E3,$5F
	.dc.b	$4F,$F3,$A5,$F4,$FF,$E8,$B7,$FD,$A9,$E1,$7F,$B5,$A3,$15,$FC,$0A
	.dc.b	$77,$F1,$3F,$ED,$A8,$E7,$B7,$F8,$3A,$B7,$87,$5F,$87,$E3,$8F,$FB
	.dc.b	$81,$B8,$FF,$85,$1D,$DD,$C7,$A5,$F9,$70,$DD,$F8,$0F,$F0,$DE,$3E
	.dc.b	$38,$E3,$19,$3F,$DD,$93,$C6,$BC,$D8,$FA,$DB,$3E,$77,$BE,$16,$3A
	.dc.b	$D7,$A6,$77,$1F,$E9,$46,$E3,$E0,$3C,$38,$79,$89,$1D,$27,$C9,$8F
	.dc.b	$C6,$6E,$3E,$37,$D6,$D7,$D3,$8F,$8D,$A7,$03,$13,$F0,$6B,$B5,$F8
	.dc.b	$DF,$29,$27,$D7,$5F,$35,$A7,$32,$4F,$AA,$7D,$18,$7D,$02,$38,$24
	.dc.b	$FA,$0E,$87,$E2,$0B,$7E,$F9,$FA,$BC,$7F,$EB,$C9,$F0,$B5,$F6,$B6
	.dc.b	$7E,$33,$98,$FF,$2C,$FC,$EF,$F9,$6A,$31,$08,$C5,$7F,$AB,$DB,$1F
	.dc.b	$ED,$66,$BF,$E6,$23,$81,$F6,$CC,$7C,$8F,$F1,$C9,$FD,$14,$78,$0D
	.dc.b	$1F,$E6,$A8,$D6,$1D,$E3,$22,$32,$F8,$4C,$B8,$7C,$55,$F7,$5E,$1C
	.dc.b	$35,$F2,$E5,$E2,$AF,$E5,$EF,$92,$3D,$38,$66,$BE,$8C,$79,$E3,$0F
	.dc.b	$90,$E7,$87,$FB,$E2,$F9,$BE,$9B,$D9,$D7,$0F,$24,$EA,$BD,$53,$AA
	.dc.b	$E0,$73,$C3,$D1,$7C,$1B,$5F,$01,$F2,$BE,$89,$E0,$F9,$67,$83,$FB
	.dc.b	$93,$E8,$57,$77,$C3,$3C,$23,$CD,$1B,$BF,$33,$F3,$0F,$01,$D4,$B8
	.dc.b	$70,$9E,$91,$B7,$B7,$07,$F4,$2F,$33,$FC,$A1,$97,$F3,$B6,$D7,$F5
	.dc.b	$FE,$4F,$F4,$F4,$FF,$83,$4D,$16,$3F,$99,$72,$FF,$91,$E3,$FA,$85
	.dc.b	$7F,$A8,$EF,$B4,$60,$FE,$A9,$E8,$76,$8F,$EA,$6F,$83,$A9,$8B,$BF
	.dc.b	$58,$AA,$93,$0F,$15,$F0,$AC,$65,$FD,$1B,$1E,$B5,$E7,$6C,$7F,$7F
	.dc.b	$9D,$F1,$7F,$59,$C3,$D1,$8F,$7C,$3C,$E4,$BF,$D6,$1F,$DE,$3A,$BF
	.dc.b	$DD,$D1,$F0,$EF,$FB,$D7,$85,$EB,$D3,$FC,$39,$39,$AD,$DF,$F8,$B5
	.dc.b	$AF,$F8,$1A,$BF,$D9,$57,$CA,$BF,$C3,$F5,$AA,$FE,$D9,$F3,$9D,$4E
	.dc.b	$DE,$55,$AF,$87,$8A,$71,$5F,$E1,$DA,$B8,$7F,$A3,$F0,$FF,$33,$61
	.dc.b	$E5,$E7,$CB,$FE,$16,$7F,$07,$3F,$E1,$57,$CB,$3C,$2F,$E1,$CE,$6B
	.dc.b	$CF,$5A,$FA,$7E,$36,$7E,$F7,$86,$F8,$2C,$C4,$BF,$B7,$EA,$FF,$B9
	.dc.b	$B2,$F8,$D3,$F8,$58,$C6,$5E,$9F,$82,$EA,$74,$F8,$2B,$87,$CF,$8E
	.dc.b	$1F,$7F,$8C,$7F,$D6,$0F,$E7,$9F,$F7,$DF,$BB,$7B,$BF,$DB,$B7,$F9
	.dc.b	$DF,$35,$F5,$0C,$75,$72,$BF,$07,$87,$F3,$78,$FF,$80,$18,$72,$8C
	.dc.b	$B5,$FC,$4B,$9C,$FF,$C9,$BD,$4F,$C1,$FF,$21,$5F,$E1,$E3,$CB,$87
	.dc.b	$E7,$A3,$35,$FE,$54,$97,$D4,$7F,$96,$2B,$F6,$91,$8F,$2B,$F5,$9F
	.dc.b	$52,$F0,$7E,$05,$EE,$BB,$37,$0F,$1D,$7F,$8A,$30,$E7,$39,$7F,$CB
	.dc.b	$F1,$9A,$FF,$3D,$E9,$F7,$3E,$A7,$EE,$B6,$BF,$6B,$CD,$EF,$7F,$CC
	.dc.b	$EA,$77,$1A,$7E,$97,$4C,$8F,$E1,$C3,$E9,$FF,$A4,$E1,$F1,$78,$15
	.dc.b	$B8,$F9,$EF,$D2,$25,$7D,$68,$4F,$08,$EB,$4F,$90,$E5,$AF,$FA,$02
	.dc.b	$7C,$11,$F9,$8F,$99,$FC,$CF,$C0,$D9,$F6,$B7,$1F,$F4,$EB,$6F,$EE
	.dc.b	$E8,$8F,$D1,$FC,$74,$BF,$8B,$F2,$F7,$0F,$E0,$93,$F9,$FE,$FC,$63
	.dc.b	$F8,$6E,$CF,$D8,$3D,$3D,$3F,$D5,$F3,$7E,$B1,$B9,$7D,$06,$E3,$FE
	.dc.b	$B0,$6D,$71,$9F,$AB,$5C,$5A,$7D,$8F,$4F,$F6,$FE,$97,$EE,$1B,$19
	.dc.b	$3E,$EC,$65,$A7,$E0,$3D,$1F,$DF,$FC,$57,$F0,$06,$C6,$4F,$C2,$C6
	.dc.b	$5A,$7E,$6B,$EA,$5F,$53,$FB,$76,$C7,$D8,$FA,$9C,$4C,$8F,$BD,$3A
	.dc.b	$D7,$F6,$AD,$7F,$9C,$B5,$FF,$13,$3B,$D3,$F0,$1C,$D6,$B3,$7F,$D9
	.dc.b	$1F,$C9,$3F,$D9,$BE,$31,$F5,$3F,$C5,$25,$8A,$73,$F1,$DB,$AF,$E8
	.dc.b	$38,$FF,$B3,$27,$37,$FB,$8F,$59,$D3,$FB,$DC,$7F,$C5,$2D,$7F,$C8
	.dc.b	$AD,$7F,$C8,$8E,$7C,$37,$57,$F6,$9B,$FD,$8F,$A3,$E1,$FA,$9F,$E6
	.dc.b	$2F,$F3,$2F,$73,$F9,$16,$5F,$95,$AF,$DB,$C6,$3B,$64,$7E,$A6,$75
	.dc.b	$9F,$F9,$46,$7C,$51,$88,$FF,$28,$65,$E2,$F3,$FF,$9F,$9C,$A3,$FE
	.dc.b	$00,$BF,$CC,$63,$FA,$15,$FF,$18,$6C,$FF,$79,$E0,$FE,$B5,$39,$F3
	.dc.b	$B8,$1F,$06,$B7,$5F,$DD,$1A,$F8,$A3,$47,$F8,$37,$E1,$DE,$07,$2B
	.dc.b	$FF,$04,$C6,$1F,$17,$9D,$E5,$38,$76,$D8,$C9,$FC,$2F,$14,$F8,$C9
	.dc.b	$19,$7F,$AF,$F8,$9C,$72,$FF,$6C,$D7,$D6,$8F,$EA,$5F,$50,$FA,$7A
	.dc.b	$7F,$8C,$BE,$A1,$F5,$3F,$84,$CB,$F9,$ED,$6D,$B5,$F8,$38,$C6,$3F
	.dc.b	$D5,$1A,$FF,$86,$1A,$FF,$A1,$5D,$F4,$AB,$F3,$3E,$CB,$EA,$5F,$09
	.dc.b	$E4,$9F,$F1,$54,$61,$FE,$9F,$F5,$0F,$90,$FE,$D8,$D8,$FF,$18,$7D
	.dc.b	$4B,$E7,$B3,$52,$7F,$9A,$56,$F4,$FE,$D6,$D7,$FC,$5E,$D7,$FD,$38
	.dc.b	$EF,$3F,$44,$17,$D8,$7E,$22,$3F,$B5,$4E,$EB,$F6,$B5,$AE,$9F,$F0
	.dc.b	$17,$C6,$3E,$AA,$9F,$E0,$5F,$7B,$FA,$7D,$7F,$6F,$9D,$5F,$F6,$E6
	.dc.b	$E3,$FE,$02,$6B,$FE,$46,$6B,$FE,$B3,$6B,$FC,$19,$E3,$3F,$CF,$7E
	.dc.b	$23,$C3,$5F,$E2,$27,$CB,$ED,$C8,$FE,$FB,$83,$E9,$71,$8F,$F8,$1F
	.dc.b	$1F,$F0,$B5,$7F,$5F,$6C,$79,$6F,$F8,$44,$62,$3F,$6F,$E2,$BE,$33
	.dc.b	$FC,$5D,$EB,$FF,$8D,$9B,$87,$F7,$A6,$D7,$E6,$AB,$84,$FF,$8E,$2B
	.dc.b	$FC,$8F,$F5,$4F,$52,$9F,$E6,$3C,$2D,$7F,$99,$E7,$4D,$9D,$34,$FF
	.dc.b	$3C,$7C,$BF,$FA,$37,$1F,$F3,$E3,$77,$FF,$14,$FD,$57,$F7,$CA,$DF
	.dc.b	$C4,$FF,$A3,$6B,$51,$F9,$C6,$DB,$C1,$5F,$18,$EB,$7C,$4F,$F4,$D5
	.dc.b	$78,$63,$FD,$D9,$D8,$8C,$47,$59,$DF,$CC,$F8,$4F,$F4,$E5,$FF,$8C
	.dc.b	$71,$F5,$C6,$5A,$7F,$A6,$A3,$55,$EE,$9F,$F2,$B7,$D5,$7F,$99,$FE
	.dc.b	$37,$FC,$34,$D9,$FF,$28,$B5,$FF,$4A,$B5,$FE,$A6,$3E,$3F,$FC,$E5
	.dc.b	$F4,$BF,$E2,$36,$C7,$F9,$69,$AF,$FA,$79,$AF,$29,$FE,$3F,$C5,$E2
	.dc.b	$F1,$FF,$C5,$73,$E2,$6B,$E2,$6B,$FC,$33,$EA,$9F,$25,$E4,$E2,$FB
	.dc.b	$3E,$0D,$DF,$FB,$4F,$1F,$D4,$8C,$60,$E3,$63,$CA,$3F,$D7,$8D,$7F
	.dc.b	$96,$F0,$FF,$0F,$7E,$65,$72,$AE,$39,$5B,$FE,$CF,$5F,$47,$E3,$BB
	.dc.b	$E3,$F8,$47,$5D,$12,$74,$43,$FD,$9D,$5C,$9B,$1C,$B2,$55,$39,$4F
	.dc.b	$F1,$7C,$FF,$89,$FD,$53,$E9,$78,$78,$F2,$BF,$D9,$93,$96,$CF,$F9
	.dc.b	$6F,$74,$E6,$7F,$D6,$6D,$7E,$73,$5F,$9B,$F3,$EF,$95,$F7,$BF,$C2
	.dc.b	$E3,$FD,$D6,$D7,$FA,$BB,$B7,$31,$C3,$C0,$91,$83,$FD,$23,$F5,$0F
	.dc.b	$AA,$FB,$3D,$3F,$DC,$38,$EF,$A7,$F5,$F4,$FF,$4D,$5E,$F0,$E3,$86
	.dc.b	$AE,$46,$0E,$BF,$3B,$F5,$BE,$1F,$FD,$39,$1F,$EC,$8F,$8F,$FC,$3A
	.dc.b	$C1,$58,$6C,$6C,$87,$FA,$BD,$CB,$63,$B4,$C8,$F0,$9F,$EA,$EF,$A9
	.dc.b	$FE,$3C,$FF,$AC,$3E,$A7,$F2,$0F,$FA,$E5,$B9,$FF,$AD,$E7,$73,$FE
	.dc.b	$BD,$6C,$7F,$AE,$E7,$7C,$9F,$B8,$6D,$7F,$93,$2B,$73,$FB,$C7,$9C
	.dc.b	$7E,$F8,$38,$75,$E2,$7F,$A8,$5C,$B7,$D9,$7C,$67,$C8,$F7,$7E,$95
	.dc.b	$7F,$B4,$5B,$97,$FB,$3E,$B7,$3F,$EC,$E6,$BB,$E0,$E4,$7D,$28,$CC
	.dc.b	$F1,$BC,$12,$BF,$DA,$F7,$F9,$99,$FF,$6E,$7D,$42,$FF,$2B,$F5,$32
	.dc.b	$FF,$DB,$B3,$96,$CE,$5F,$CB,$CF,$E0,$A7,$51,$FE,$EB,$6D,$76,$AD
	.dc.b	$1F,$C0,$3C,$DF,$DB,$D7,$6F,$0F,$1D,$92,$30,$43,$FC,$77,$3F,$E2
	.dc.b	$67,$FC,$1B,$BF,$F1,$A8,$C4,$78,$6B,$C8,$47,$96,$1F,$E1,$AF,$09
	.dc.b	$97,$FB,$72,$7F,$67,$F5,$3F,$B6,$23,$FE,$FE,$FA,$97,$D2,$FE,$81
	.dc.b	$B1,$C0,$87,$F0,$F3,$FD,$2A,$E9,$FD,$EF,$4F,$19,$91,$FC,$0A,$3F
	.dc.b	$82,$FC,$04,$8F,$E0,$AF,$F3,$4F,$87,$FF,$75,$FC,$6F,$EC,$3E,$0F
	.dc.b	$FC,$67,$5D,$E7,$2F,$C4,$3F,$D4,$9E,$5A,$B8,$70,$F3,$8F,$F0,$94
	.dc.b	$EE,$BF,$DE,$FE,$9E,$4F,$07,$CC,$FD,$69,$F6,$0F,$95,$F9,$FC,$5E
	.dc.b	$2F,$A7,$7C,$EF,$F0,$46,$D7,$8C,$E1,$CF,$FC,$22,$F8,$E9,$FD,$A2
	.dc.b	$B7,$5D,$9B,$87,$BA,$B7,$3F,$EC,$06,$A9,$DD,$73,$C3,$B6,$1F,$CC
	.dc.b	$3C,$DF,$E6,$3E,$AC,$C3,$F9,$8C,$65,$FE,$67,$19,$7F,$99,$C7,$F3
	.dc.b	$8F,$A8,$FE,$73,$F5,$32,$7F,$9C,$C6,$5F,$E7,$71,$97,$F9,$DC,$7F
	.dc.b	$40,$F5,$FF,$A0,$E3,$27,$FA,$0C,$65,$FE,$87,$19,$7F,$A1,$C6,$5F
	.dc.b	$E8,$B1,$9D,$7F,$CE,$7E,$9B,$CE,$5F,$F7,$A9,$FF,$27,$78,$3F,$CE
	.dc.b	$7D,$49,$8F,$E1,$67,$2D,$E3,$56,$A3,$FD,$53,$E5,$FF,$24,$AD,$C7
	.dc.b	$F5,$6F,$81,$FE,$B7,$EC,$C8,$FE,$B7,$1F,$D7,$8C,$3F,$D7,$23,$0F
	.dc.b	$E0,$FE,$A7,$FA,$A7,$27,$E0,$78,$FF,$10,$7F,$4C,$8C,$CF,$FB,$F9
	.dc.b	$AF,$DF,$8D,$BC,$FB,$78,$18,$7F,$15,$18,$7F,$B7,$F7,$7F,$84,$39
	.dc.b	$7F,$53,$9F,$CA,$6E,$3F,$DE,$27,$FB,$B9,$B6,$57,$23,$F3,$9F,$12
	.dc.b	$F9,$8F,$F0,$DF,$D4,$FF,$2D,$AF,$F1,$6B,$68,$2B,$83,$CA,$B8,$FA
	.dc.b	$F2,$30,$7E,$A7,$EA,$1F,$01,$FE,$28,$FA,$9F,$E5,$B9,$7F,$8E,$9B
	.dc.b	$5D,$AB,$6F,$1A,$E1,$E6,$48,$C1,$0F,$F7,$3D,$72,$EA,$3D,$21,$7D
	.dc.b	$1F,$E3,$DB,$FF,$A6,$FE,$67,$FC,$0D,$EE,$49,$FC,$64,$7E,$4F,$A4
	.dc.b	$E6,$3F,$1B,$39,$8D,$7C,$97,$F8,$3F,$C1,$95,$F9,$58,$CB,$FC,$77
	.dc.b	$37,$57,$F0,$46,$D6,$A7,$CB,$8B,$E4,$BF,$C8,$DE,$4C,$9F,$D0,$47
	.dc.b	$94,$C3,$FE,$5C,$8C,$3D,$FC,$1F,$F0,$1B,$FE,$9B,$6E,$1F,$C8,$3D
	.dc.b	$4F,$F2,$DC,$3E,$46,$C4,$AE,$4E,$A3,$F6,$DF,$0D,$FE,$7A,$E0,$FF
	.dc.b	$56,$FD,$17,$F9,$6C,$CF,$99,$30,$FF,$4C,$46,$5F,$DD,$57,$1E,$7F
	.dc.b	$C6,$FA,$4F,$A4,$7D,$3F,$F9,$97,$B2,$61,$FD,$16,$B5,$3F,$E7,$7A
	.dc.b	$DC,$FF,$A9,$6B,$6F,$F5,$BA,$DB,$C6,$76,$E5,$B9,$FF,$6D,$6E,$1F
	.dc.b	$E8,$ED,$9F,$3B,$DF,$29,$87,$37,$E4,$C7,$8F,$C3,$FF,$3E,$7C,$5C
	.dc.b	$AF,$04,$65,$FF,$5D,$46,$5F,$14,$7F,$95,$3E,$25,$2C,$FF,$63,$F4
	.dc.b	$FF,$E6,$DB,$FF,$5B,$6D,$FE,$B8,$8C,$5F,$ED,$25,$18,$BE,$51,$8F
	.dc.b	$86,$FF,$5F,$D7,$6C,$3A,$C6,$30,$E8,$DA,$C9,$FC,$0A,$B8,$3C,$CC
	.dc.b	$5F,$6C,$3C,$77,$CA,$78,$35,$FF,$67,$4E,$EF,$C6,$D8,$D9,$E0,$3D
	.dc.b	$E5,$BF,$0D,$FB,$70,$C6,$1D,$67,$E1,$4E,$6B,$CC,$78,$2B,$6C,$33
	.dc.b	$3D,$30,$F1,$CE,$6B,$D7,$5E,$1C,$3C,$CF,$33,$8E,$1E,$13,$C5,$1A
	.dc.b	$C3,$A4,$FB,$0D,$9C,$EE,$38,$25,$FA,$49,$87,$D3,$8C,$BE,$67,$D3
	.dc.b	$7D,$F0,$F1,$47,$4A,$F8,$5C,$5F,$EA,$28,$F4,$37,$1F,$B6,$DB,$FC
	.dc.b	$3B,$B9,$FE,$83,$F2,$FF,$D0,$A3,$10,$8C,$7C,$17,$FA,$36,$30,$27
	.dc.b	$73,$A8,$FF,$49,$3F,$75,$7C,$63,$87,$11,$CB,$2F,$D0,$3A,$8C,$D7
	.dc.b	$1A,$E1,$3F,$78,$F3,$95,$CD,$72,$8E,$9A,$7F,$A8,$E7,$2F,$A6,$BA
	.dc.b	$5F,$6A,$E4,$DB,$F1,$DF,$09,$F5,$9A,$3C,$47,$63,$FB,$49,$E9,$33
	.dc.b	$5D,$9A,$71,$9F,$34,$65,$B1,$9B,$F3,$34,$F0,$C7,$92,$FC,$D3,$ED
	.dc.b	$6C,$F8,$AF,$D0,$F9,$2F,$B5,$74,$BE,$53,$E4,$69,$E2,$2C,$6C,$ED
	.dc.b	$1B,$6C,$1F,$E7,$BC,$4F,$2B,$EB,$EF,$FE,$42,$F8,$A7,$D3,$63,$4D
	.dc.b	$D5,$C9,$D6,$BD,$D7,$F3,$BA,$D4,$E6,$F8,$73,$FF,$4B,$9C,$DF,$0C
	.dc.b	$3F,$A0,$7D,$3E,$B3,$3C,$BA,$1E,$16,$62,$69,$FD,$25,$D4,$62,$F9
	.dc.b	$CF,$F7,$DC,$1F,$21,$9D,$BD,$CD,$9F,$F1,$5E,$DE,$B8,$C5,$FB,$AF
	.dc.b	$8D,$7F,$66,$F1,$7D,$3F,$05,$F0,$F1,$76,$A7,$35,$B3,$C5,$5B,$9D
	.dc.b	$46,$0F,$F0,$27,$C8,$FB,$CF,$81,$E6,$D8,$C9,$32,$FF,$62,$69,$FE
	.dc.b	$A6,$9D,$47,$F8,$AA,$FF,$C9,$55,$F0,$7E,$5F,$C1,$DB,$F4,$1B,$19
	.dc.b	$3F,$C0,$31,$97,$FC,$07,$CB,$F6,$BC,$87,$8B,$E9,$8C,$36,$7F,$C5
	.dc.b	$53,$A7,$89,$88,$F8,$1A,$F3,$F0,$EC,$F0,$6E,$7F,$E8,$66,$F0,$7F
	.dc.b	$A5,$30,$47,$E4,$67,$F8,$D7,$B9,$F8,$EF,$A6,$FC,$3E,$1F,$EC,$5E
	.dc.b	$FF,$F6,$63,$FE,$3D,$8C,$D7,$FB,$42,$72,$FC,$A8,$CE,$3F,$EE,$29
	.dc.b	$CB,$FE,$A4,$8C,$D7,$F0,$89,$CB,$F8,$88,$CD,$7F,$33,$BF,$27,$3F
	.dc.b	$9F,$E9,$BF,$C7,$57,$FE,$4F,$FA,$6F,$F0,$5B,$FE,$DB,$8C,$DF,$DC
	.dc.b	$9C,$BF,$EE,$68,$CE,$5F,$EB,$69,$CB,$FE,$F8,$8C,$D7,$E6,$A7,$2F
	.dc.b	$E7,$3D,$5F,$3D,$FE,$E1,$7D,$37,$ED,$9C,$F3,$1C,$D7,$F9,$95,$C9
	.dc.b	$8A,$FC,$8B,$8A,$FA,$9A,$FF,$A4,$EB,$FD,$2B,$97,$F4,$25,$F4,$4F
	.dc.b	$A6,$DF,$47,$F0,$B9,$CC,$61,$E2,$7F,$A7,$8C,$9F,$76,$38,$4F,$38
	.dc.b	$C1,$C3,$2F,$EB,$11,$97,$F0,$13,$C2,$BE,$07,$D0,$FC,$1F,$76,$E1
	.dc.b	$EF,$6D,$FB,$9B,$8F,$FA,$F5,$FD,$D7,$C6,$FE,$97,$A6,$DF,$E8,$9B
	.dc.b	$7F,$EE,$06,$DF,$E6,$DF,$A1,$E8,$7D,$1F,$42,$DF,$CF,$6D,$FC,$E6
	.dc.b	$DF,$FB,$35,$B3,$F7,$5B,$3F,$75,$FB,$9F,$4D,$F7,$7D,$0B,$5F,$6D
	.dc.b	$B5,$F6,$9B,$7F,$ED,$97,$F0,$9E,$AF,$E1,$7E,$22,$DF,$E0,$DB,$7F
	.dc.b	$82,$6D,$FF,$B8,$5F,$DB,$7D,$2B,$E9,$7E,$B9,$82,$DF,$D7,$B4,$EE
	.dc.b	$3E,$D1,$F8,$89,$E0,$E3,$8F,$55,$FE,$73,$8F,$EF,$75,$8C,$3F,$21
	.dc.b	$56,$73,$7E,$0B,$E0,$3E,$29,$F4,$BF,$C6,$63,$31,$8F,$0B,$FB,$03
	.dc.b	$AF,$1E,$D6,$F2,$FF,$0F,$DF,$F8,$C7,$97,$3F,$4B,$FC,$5F,$D2,$46
	.dc.b	$34,$FD,$67,$99,$FB,$13,$17,$FA,$C7,$F6,$9E,$CF,$1E,$9B,$5F,$1F
	.dc.b	$04,$66,$3F,$6D,$E5,$DC,$3F,$6A,$FE,$F4,$91,$C6,$6C,$FC,$23,$FB
	.dc.b	$DF,$BD,$FC,$B7,$C1,$3F,$DE,$38,$58,$66,$26,$1F,$EE,$9F,$8D,$7C
	.dc.b	$C6,$31,$FE,$AB,$18,$9F,$BD,$8F,$47,$85,$6F,$83,$FC,$8F,$2A,$E1
	.dc.b	$7E,$58,$7C,$97,$94,$FF,$06,$CF,$F2,$09,$FE,$F3,$19,$C3,$F8,$C7
	.dc.b	$83,$67,$F8,$BB,$70,$EF,$E2,$7F,$17,$9E,$D7,$FC,$8E,$BF,$A7,$7D
	.dc.b	$1B,$CF,$A7,$F8,$87,$E9,$5F,$4B,$F7,$33,$FC,$D5,$6D,$B5,$FE,$20
	.dc.b	$FA,$57,$C1,$F2,$9F,$F7,$44,$61,$D9,$88,$D6,$7F,$EE,$79,$D3,$F5
	.dc.b	$FE,$8D,$F1,$BF,$90,$8F,$F2,$DC,$62,$FF,$D0,$10,$8C,$3F,$E8,$36
	.dc.b	$BF,$E6,$23,$53,$FE,$1B,$FA,$4F,$BF,$5F,$B5,$AE,$3C,$1F,$E4,$69
	.dc.b	$CD,$F8,$EF,$87,$2F,$F7,$5B,$DE,$1F,$D0,$AF,$78,$7E,$9B,$E2,$7F
	.dc.b	$C5,$78,$FF,$86,$BD,$5F,$EA,$E8,$46,$A7,$FC,$05,$F0,$7F,$E0,$18
	.dc.b	$C6,$1F,$E3,$8B,$E3,$7C,$3D,$BF,$F0,$C5,$ED,$B5,$FC,$3E,$F8,$D7
	.dc.b	$F9,$6A,$F8,$C3,$D8,$FF,$2F,$7C,$FF,$EE,$3E,$6B,$C1,$18,$F0,$91
	.dc.b	$8C,$7F,$A8,$56,$DE,$D5,$B9,$FE,$DD,$F1,$3F,$E2,$BF,$52,$D7,$F8
	.dc.b	$A1,$B9,$FC,$36,$ED,$FE,$21,$7F,$C9,$DF,$48,$F4,$BF,$01,$4A,$4E
	.dc.b	$6F,$FD,$81,$18,$AD,$B6,$33,$3B,$AF,$F2,$5F,$BE,$F2,$3B,$E1,$FE
	.dc.b	$76,$7F,$CF,$3F,$23,$FE,$87,$C3,$FC,$D9,$F6,$D8,$F8,$4F,$F3,$DB
	.dc.b	$63,$27,$F7,$3F,$85,$57,$1C,$BB,$D6,$DB,$5B,$9D,$69,$FE,$93,$AF
	.dc.b	$0F,$CA,$7F,$A2,$3E,$DE,$DF,$F9,$F5,$B8,$7A,$9B,$7F,$E9,$B7,$FC
	.dc.b	$91,$F4,$DF,$C7,$3E,$2F,$F4,$AD,$AF,$F7,$33,$5F,$D2,$35,$FF,$63
	.dc.b	$FC,$6F,$FA,$93,$9B,$F5,$AD,$8F,$F7,$73,$5F,$D6,$35,$FF,$67,$CF
	.dc.b	$F1,$AE,$3F,$0F,$8F,$FE,$F3,$9F,$0B,$5F,$0B,$5F,$D2,$6E,$FC,$BB
	.dc.b	$FD,$EB,$1F,$35,$AF,$F5,$C8,$D6,$5F,$ED,$ED,$3F,$D7,$9F,$27,$76
	.dc.b	$E7,$87,$09,$FF,$7B,$DF,$0C,$3A,$61,$B7,$FD,$45,$85,$97,$D3,$BE
	.dc.b	$47,$0D,$AF,$C0,$35,$FF,$4B,$FC,$6F,$F6,$27,$33,$C0,$8F,$F8,$E7
	.dc.b	$CF,$4E,$E3,$D9,$5C,$F2,$F3,$B7,$4F,$8A,$D9,$FF,$6A,$D7,$1A,$F6
	.dc.b	$E2,$E3,$F5,$36,$BF,$2C,$D7,$F4,$7F,$07,$A9,$91,$A7,$E2,$65,$FE
	.dc.b	$E1,$9B,$E0,$FF,$BA,$2B,$FD,$47,$E1,$60,$CF,$C6,$48,$C9,$0F,$E4
	.dc.b	$4F,$CF,$5F,$4A,$BC,$36,$3F,$C5,$0D,$7F,$C3,$7C,$EF,$AD,$FD,$D4
	.dc.b	$6A,$B8,$91,$FE,$2F,$1A,$7F,$C0,$54,$FA,$3E,$E4,$8D,$B5,$DB,$76
	.dc.b	$70,$FF,$39,$C7,$C5,$39,$8F,$E7,$95,$EC,$8C,$29,$DB,$FE,$43,$FC
	.dc.b	$13,$FE,$00,$B7,$E3,$7E,$24,$62,$3F,$C3,$1F,$46,$FA,$BC,$D4,$9E
	.dc.b	$07,$F9,$36,$74,$E0,$FF,$10,$57,$3C,$BF,$C7,$2D,$CB,$CF,$5C,$2B
	.dc.b	$FC,$4D,$F4,$DF,$CC,$5F,$ED,$AD,$7F,$B5,$07,$0F,$F6,$D6,$BF,$09
	.dc.b	$E1,$7F,$E4,$86,$D6,$7C,$4F,$C1,$7D,$3F,$F0,$3A,$D4,$7E,$16,$B7
	.dc.b	$7F,$D9,$DA,$AB,$85,$FD,$0F,$A7,$FE,$D7,$3F,$EF,$6F,$2F,$F2,$E7
	.dc.b	$E3,$E3,$31,$FE,$43,$6E,$3F,$E5,$C6,$D7,$3F,$37,$F3,$FE,$AF,$16
	.dc.b	$5F,$F9,$EE,$B6,$FF,$6D,$9E,$DE,$EF,$EA,$3E,$8B,$FB,$4C,$FE,$FD
	.dc.b	$AF,$86,$36,$F3,$6B,$FE,$78,$EA,$FE,$ED,$F0,$B8,$CF,$CD,$57,$3F
	.dc.b	$F0,$55,$6A,$FF,$86,$75,$FE,$7E,$FF,$C9,$ED,$AF,$F0,$57,$99,$FB
	.dc.b	$1F,$A7,$FE,$AF,$7F,$E9,$56,$BF,$E6,$76,$DF,$FA,$01,$B5,$FE,$13
	.dc.b	$F2,$FF,$81,$9C,$FC,$C9,$7E,$28,$FE,$0F,$E6,$48,$F2,$BF,$17,$9B
	.dc.b	$C5,$F4,$32,$7C,$4F,$C9,$F1,$A4,$7C,$96,$BF,$E2,$BF,$1F,$F8,$57
	.dc.b	$E5,$FF,$E7,$48,$FF,$57,$35,$FF,$40,$37,$5F,$87,$E1,$70,$F8,$29
	.dc.b	$5C,$1F,$E3,$B1,$88,$E3,$16,$78,$C7,$CE,$F0,$7F,$91,$FD,$4C,$8C
	.dc.b	$10,$FF,$20,$39,$7F,$C8,$51,$FC,$AB,$EE,$3F,$96,$FD,$DF,$F9,$56
	.dc.b	$FF,$D0,$AD,$E2,$FF,$20,$4E,$EF,$FA,$44,$E5,$73,$23,$0B,$99,$5D
	.dc.b	$BC,$DE,$FF,$2D,$30,$E7,$19,$7A,$C6,$5F,$C9,$61,$FD,$A7,$1F,$C7
	.dc.b	$B6,$3C,$2A,$78,$65,$EA,$BF,$ED,$5E,$8B,$E5,$29,$30,$F5,$F8,$F5
	.dc.b	$FE,$F9,$3B,$7F,$C5,$8D,$7C,$71,$F2,$3E,$3F,$E4,$FD,$14,$9F,$8B
	.dc.b	$19,$7E,$3C,$65,$F5,$C7,$F5,$1F,$9B,$FF,$4B,$C7,$FB,$D5,$AF,$FB
	.dc.b	$1D,$B9,$7F,$B8,$3E,$A9,$F5,$5C,$7E,$1A,$5F,$17,$FA,$E4,$EA,$30
	.dc.b	$43,$A4,$6A,$7F,$83,$7D,$57,$F0,$76,$4F,$F0,$78,$CB,$FC,$26,$32
	.dc.b	$FF,$05,$F1,$53,$FE,$19,$C1,$EF,$FF,$BA,$2B,$FD,$1D,$1F,$D6,$1F
	.dc.b	$F3,$54,$65,$FF,$36,$71,$EA,$31,$3F,$ED,$6B,$FE,$8D,$5C,$3E,$3F
	.dc.b	$B5,$6A,$F8,$47,$F1,$FF,$43,$FB,$E7,$D4,$FF,$85,$6C,$D3,$1A,$FF
	.dc.b	$B8,$63,$CB,$7F,$DA,$27,$5E,$07,$23,$E5,$FC,$7D,$AD,$9F,$EA,$99
	.dc.b	$DB,$81,$7D,$5F,$F8,$01,$F1,$CF,$9C,$FF,$08,$36,$38,$35,$FE,$61
	.dc.b	$F5,$2F,$03,$75,$27,$6F,$F8,$92,$75,$1E,$4E,$0F,$A3,$F3,$7F,$18
	.dc.b	$91,$FE,$3F,$F9,$27,$D8,$62,$3E,$89,$F5,$27,$15,$FE,$F8,$9F,$D4
	.dc.b	$FC,$BC,$AF,$D1,$BF,$E3,$4F,$88,$FF,$0D,$46,$08,$7E,$AA,$35,$E2
	.dc.b	$C8,$FF,$28,$4F,$07,$FD,$47,$1C,$F4,$D4,$60,$DC,$65,$FD,$9F,$C7
	.dc.b	$FE,$F0,$8F,$4F,$12,$BF,$BA,$8B,$97,$E4,$7C,$66,$3F,$C5,$2B,$ED
	.dc.b	$F8,$3F,$E2,$5F,$4F,$FC,$E4,$D8,$CF,$B8,$9F,$F2,$9F,$EF,$1F,$C8
	.dc.b	$6B,$55,$F2,$9B,$8F,$E7,$DB,$D3,$FD,$9A,$B8,$61,$CE,$FA,$E1,$E0
	.dc.b	$8C,$47,$F7,$19,$DC,$E4,$91,$FE,$79,$8E,$7F,$49,$C4,$91,$9F,$1E
	.dc.b	$BE,$5B,$4E,$32,$F8,$97,$FC,$AE,$9F,$14,$F8,$FF,$F7,$04,$3E,$A1
	.dc.b	$F5,$1F,$EB,$6E,$AF,$F2,$7C,$62,$47,$28,$CC,$FF,$AF,$2B,$9D,$6E
	.dc.b	$BB,$C6,$27,$A4,$72,$9D,$5F,$E9,$5B,$5B,$9E,$F3,$E7,$C7,$85,$F2
	.dc.b	$9F,$43,$6F,$8C,$EA,$78,$DF,$37,$D0,$64,$A7,$23,$71,$9B,$F0,$46
	.dc.b	$4D,$D7,$A3,$0F,$83,$1C,$0D,$4F,$09,$D4,$73,$9D,$BF,$12,$BB,$5F
	.dc.b	$96,$FE,$16,$1E,$B3,$E6,$4E,$8E,$55,$0E,$F5,$C4,$E5,$7D,$AF,$9C
	.dc.b	$F9,$70,$F8,$AF,$9A,$74,$6E,$B9,$5F,$5A,$F4,$D7,$68,$F1,$3C,$4D
	.dc.b	$61,$E6,$74,$37,$7D,$4E,$5B,$1F,$1D,$AF,$C3,$38,$47,$FA,$0B,$E6
	.dc.b	$FF,$1F,$3A,$9F,$C8,$C6,$27,$51,$C0,$FF,$52,$88,$D3,$87,$F0,$0B
	.dc.b	$7F,$EB,$18,$FE,$5E,$3A,$9F,$7C,$D7,$0F,$07,$A1,$99,$E3,$8F,$8A
	.dc.b	$B9,$46,$27,$C0,$6E,$B9,$5F,$8A,$75,$3F,$EB,$C8,$C4,$EA,$38,$1E
	.dc.b	$33,$C3,$5D,$A3,$4D,$8D,$4F,$26,$9C,$A3,$DC,$F2,$BF,$8F,$7E,$28
	.dc.b	$F8,$33,$CC,$DD,$73,$6C,$73,$AF,$80,$72,$BF,$1B,$63,$C3,$3C,$E3
	.dc.b	$71,$FE,$DA,$BF,$03,$E3,$31,$5C,$9A,$72,$9E,$95,$E2,$3A,$47,$E3
	.dc.b	$F8,$1F,$0F,$FE,$13,$D3,$F9,$B4,$63,$0F,$E5,$9E,$7E,$5F,$C6,$C6
	.dc.b	$6F,$F9,$55,$E5,$E8,$DE,$57,$17,$F9,$4E,$FF,$A4,$CE,$F7,$FF,$79
	.dc.b	$7B,$D7,$13,$C6,$FF,$DD,$D7,$FD,$42,$34,$47,$FA,$07,$9F,$72,$FE
	.dc.b	$79,$37,$4F,$E7,$73,$6B,$D3,$36,$BF,$CE,$F7,$D9,$E3,$3F,$E5,$EF
	.dc.b	$07,$D9,$F1,$5F,$B5,$7C,$9E,$7E,$34,$45,$25,$7F,$53,$9B,$7F,$D6
	.dc.b	$7D,$0F,$61,$ED,$E2,$B5,$FD,$66,$2C,$FF,$5F,$9B,$AF,$F9,$1F,$27
	.dc.b	$D6,$FF,$07,$C3,$FC,$63,$18,$CF,$71,$FD,$42,$30,$EA,$78,$61,$FE
	.dc.b	$1E,$9C,$BE,$59,$D5,$7D,$BB,$7D,$4B,$E1,$30,$DA,$FF,$15,$7C,$86
	.dc.b	$26,$47,$F8,$C6,$75,$87,$FB,$A7,$C7,$5E,$A3,$FC,$5E,$3E,$A1,$F4
	.dc.b	$DF,$8B,$8C,$3F,$8B,$FA,$8F,$E4,$B8,$FF,$93,$6B,$6D,$9E,$11,$8A
	.dc.b	$FF,$7C,$57,$07,$BC,$67,$3F,$F3,$94,$7D,$DB,$FF,$2C,$69,$F7,$1F
	.dc.b	$F2,$E8,$FA,$4F,$E3,$B3,$E8,$9D,$37,$8F,$D1,$18,$BF,$F3,$F5,$F1
	.dc.b	$EC,$CD,$48,$EE,$FE,$A5,$D1,$83,$E5,$7B,$BF,$40,$FD,$66,$CF,$77
	.dc.b	$13,$A2,$65,$F3,$E3,$2F,$D1,$F6,$F8,$FD,$3F,$26,$D7,$F9,$8D,$B3
	.dc.b	$B8,$FF,$33,$4F,$D3,$FA,$67,$4D,$3D,$1B,$FE,$A7,$6F,$97,$3A,$BF
	.dc.b	$F4,$13,$70,$FB,$ED,$9F,$8B,$5C,$6B,$FC,$4F,$4A,$D4,$61,$B1,$FE
	.dc.b	$2C,$6C,$7D,$5E,$CF,$23,$FE,$F4,$E9,$FE,$25,$EF,$3E,$BA,$C6,$1B
	.dc.b	$3F,$EF,$A6,$E1,$FE,$EE,$7F,$80,$F5,$FF,$97,$E7,$F5,$5E,$57,$4F
	.dc.b	$9B,$CB,$23,$27,$FA,$93,$C3,$64,$AD,$63,$FE,$FA,$F8,$5F,$2F,$B3
	.dc.b	$E6,$6E,$1C,$6D,$18,$AE,$13,$9A,$E0,$D3,$84,$EF,$83,$FC,$FD,$6F
	.dc.b	$67,$E1,$F3,$7C,$46,$DF,$76,$E3,$FE,$A2,$CF,$D3,$18,$C3,$9F,$CD
	.dc.b	$7C,$CF,$7F,$E6,$B7,$0F,$43,$70,$E5,$92,$31,$7F,$86,$F8,$9F,$C6
	.dc.b	$1F,$5B,$E0,$6E,$1F,$88,$8B,$87,$29,$B5,$C6,$B4,$7C,$78,$CB,$FE
	.dc.b	$70,$CD,$A3,$E9,$7F,$52,$7E,$13,$E5,$ED,$7F,$B5,$FC,$F7,$C0,$FF
	.dc.b	$11,$6C,$7F,$B1,$EB,$13,$F8,$CF,$98,$93,$FE,$CB,$F6,$2E,$BF,$03
	.dc.b	$E0,$B9,$78,$3F,$92,$6D,$77,$8B,$3B,$8C,$1F,$81,$FA,$0F,$CC,$9A
	.dc.b	$E0,$FB,$6F,$F2,$5D,$FF,$95,$CF,$FB,$BD,$B5,$F6,$79,$7F,$8F,$4F
	.dc.b	$EB,$27,$37,$FC,$81,$B1,$9F,$2B,$FB,$C5,$70,$BF,$95,$1D,$CF,$E2
	.dc.b	$3E,$2D,$AF,$F7,$53,$E3,$F8,$79,$1F,$99,$E5,$B7,$FC,$FB,$15,$63
	.dc.b	$2F,$CA,$D5,$BF,$34,$FE,$E3,$E2,$F1,$83,$0F,$D5,$C7,$F4,$5F,$A0
	.dc.b	$6B,$4F,$EE,$FF,$60,$FA,$FC,$65,$FE,$B2,$D3,$F9,$FD,$F0,$8C,$1F
	.dc.b	$C4,$1C,$65,$FD,$96,$F9,$4E,$A3,$FD,$77,$3A,$AE,$78,$7F,$4A,$9E
	.dc.b	$4D,$76,$D5,$3D,$9F,$F3,$8F,$DC,$7C,$F4,$FD,$1F,$4D,$8B,$EA,$3F
	.dc.b	$B2,$4E,$C9,$E9,$7C,$08,$CB,$33,$F4,$61,$D2,$35,$18,$AF,$07,$CB
	.dc.b	$FF,$76,$D3,$FB,$1B,$67,$F9,$9F,$A9,$FC,$A5,$E7,$19,$C3,$FD,$BA
	.dc.b	$D4,$D9,$E1,$1F,$DF,$A3,$44,$F5,$7F,$BF,$4E,$AF,$C3,$1F,$E0,$19
	.dc.b	$FF,$21,$DF,$83,$D8,$DF,$81,$96,$46,$4F,$7E,$1F,$C0,$1C,$63,$FC
	.dc.b	$6F,$EE,$D1,$FE,$55,$5F,$11,$63,$0D,$7F,$9D,$46,$9A,$68,$FF,$18
	.dc.b	$F9,$FF,$E0,$4D,$3F,$C6,$CD,$8F,$E2,$5E,$D7,$B5,$FF,$1A,$46,$70
	.dc.b	$71,$7F,$5F,$F5,$BF,$C9,$F3,$9E,$8F,$F3,$4F,$9F,$89,$FF,$29,$F5
	.dc.b	$BC,$6A,$7F,$9B,$FE,$13,$FC,$3D,$AF,$F9,$1E,$75,$A7,$F9,$19,$BA
	.dc.b	$7F,$90,$5B,$3F,$12,$3F,$5D,$5C,$AB,$FC,$95,$5A,$6D,$F8,$1F,$EB
	.dc.b	$D3,$BA,$F5,$D6,$DF,$F6,$E6,$8F,$AC,$FF,$40,$36,$FF,$B3,$6D,$FD
	.dc.b	$95,$FD,$77,$B3,$F1,$2B,$64,$8D,$FA,$17,$0F,$F2,$4F,$D1,$BE,$67
	.dc.b	$FD,$AD,$1F,$ED,$C1,$19,$DF,$11,$89,$DB,$FE,$E2,$9D,$3C,$7E,$1B
	.dc.b	$FD,$4B,$F0,$1C,$1B,$19,$F8,$14,$E5,$B1,$E1,$AF,$F2,$FE,$9E,$06
	.dc.b	$C6,$7C,$7F,$F8,$00,$D9,$F4,$6E,$FA,$3F,$2F,$27,$FA,$05,$BB,$FF
	.dc.b	$9F,$5B,$7F,$63,$E5,$BF,$C5,$35,$BD,$ED,$ED,$B5,$F9,$0B,$E3,$CE
	.dc.b	$E5,$63,$AC,$9F,$46,$F6,$3E,$4C,$61,$B1,$9C,$7D,$2D,$7F,$C7,$67
	.dc.b	$E7,$9C,$69,$FE,$9B,$6E,$9F,$E9,$76,$CF,$F3,$FF,$41,$F4,$BE,$39
	.dc.b	$CE,$3E,$1A,$FF,$50,$C7,$0D,$17,$C2,$7F,$D4,$B5,$FA,$76,$F5,$7B
	.dc.b	$63,$11,$FE,$1F,$D5,$F0,$59,$A9,$3C,$0F,$CF,$C7,$CA,$C1,$F2,$FF
	.dc.b	$EE,$A9,$FF,$20,$2F,$A4,$D3,$63,$3C,$D8,$33,$32,$32,$7F,$38,$FA
	.dc.b	$3F,$1B,$F9,$86,$BF,$9D,$6B,$F3,$9A,$FF,$9A,$9D,$F0,$7F,$A7,$9B
	.dc.b	$C1,$9C,$BF,$07,$D0,$FA,$47,$CF,$FE,$1B,$16,$0F,$92,$DF,$AD,$F5
	.dc.b	$B0,$FF,$49,$73,$6E,$B5,$7E,$9B,$FD,$7B,$C9,$B1,$A8,$F5,$46,$0E
	.dc.b	$44,$7F,$B8,$57,$96,$3C,$9F,$48,$F6,$3C,$DF,$02,$F8,$CA,$6A,$26
	.dc.b	$1F,$C0,$2B,$83,$63,$CB,$18,$8F,$E1,$0D,$E0,$FF,$62,$B7,$83,$FD
	.dc.b	$2A,$D9,$1F,$E7,$6F,$86,$F9,$1F,$46,$FA,$4F,$95,$F4,$9F,$E1,$3A
	.dc.b	$FE,$E6,$D7,$FD,$C3,$3B,$7F,$BB,$B5,$FE,$34,$D7,$F5,$6D,$7F,$D4
	.dc.b	$ED,$7C,$6F,$1C,$7F,$DB,$AD,$C7,$77,$E4,$F0,$BE,$F7,$D2,$7F,$78
	.dc.b	$C3,$F3,$3E,$A7,$E6,$0C,$FA,$92,$3F,$94,$CE,$DF,$E5,$CD,$7F,$7E
	.dc.b	$D7,$FD,$76,$EF,$2F,$F7,$7B,$71,$CD,$7E,$9B,$D3,$F1,$F9,$5F,$AD
	.dc.b	$F9,$FF,$F5,$46,$3F,$EA,$B6,$AC,$79,$CF,$08,$F2,$BF,$90,$C7,$FA
	.dc.b	$24,$60,$87,$1C,$7C,$B3,$C6,$3F,$D4,$19,$3E,$27,$F7,$3D,$52,$7F
	.dc.b	$7A,$FE,$97,$E9,$1E,$8D,$3F,$3E,$E2,$7F,$A7,$B5,$FE,$A8,$D7,$FC
	.dc.b	$E9,$E0,$BE,$93,$94,$7F,$8E,$27,$86,$BF,$DD,$7C,$3F,$E1,$F5,$FC
	.dc.b	$33,$E9,$25,$FF,$10,$7E,$47,$5F,$97,$E0,$7C,$64,$9F,$1B,$FD,$8A
	.dc.b	$DC,$5F,$CA,$49,$E0,$7F,$18,$F3,$3F,$81,$7D,$23,$C3,$FE,$41,$EC
	.dc.b	$7D,$9C,$BF,$CE,$8D,$53,$B7,$FC,$A2,$D7,$FB,$A3,$5F,$E8,$AD,$7F
	.dc.b	$E0,$0B,$5F,$D6,$BC,$76,$FF,$0B,$47,$6A,$F2,$9F,$CB,$BC,$6F,$E5
	.dc.b	$FE,$1F,$F3,$0F,$9D,$95,$FC,$D2,$32,$FF,$36,$8C,$BF,$ED,$FE,$3F
	.dc.b	$E2,$98,$7C,$F5,$F3,$BF,$CF,$FE,$3B,$FA,$DC,$FF,$BB,$2B,$F9,$EC
	.dc.b	$C7,$FD,$F0,$6C,$C1,$FE,$E2,$F4,$F9,$7C,$C4,$8F,$F7,$F4,$75,$30
	.dc.b	$FF,$BA,$23,$0E,$8F,$F4,$97,$13,$04,$F0,$C3,$FD,$FF,$7F,$DE,$7C
	.dc.b	$17,$CB,$7D,$8E,$E9,$97,$FA,$6F,$D1,$56,$A3,$0D,$7F,$4C,$7F,$95
	.dc.b	$9C,$61,$FE,$47,$9E,$37,$F1,$7C,$7F,$8D,$F2,$BF,$1F,$CB,$98,$FA
	.dc.b	$35,$4E,$A3,$13,$A7,$BE,$DE,$D9,$FE,$A5,$F1,$2F,$A4,$FE,$A9,$F4
	.dc.b	$5F,$DF,$AF,$FD,$3A,$D7,$CE,$7F,$B3,$A3,$89,$FD,$4A,$79,$47,$F9
	.dc.b	$09,$AF,$F9,$1D,$AF,$F8,$61,$CE,$BF,$90,$6F,$8D,$FD,$02,$7B,$1E
	.dc.b	$2F,$92,$FE,$31,$F2,$9F,$C6,$59,$1F,$C6,$63,$2F,$F1,$B8,$CB,$F0
	.dc.b	$3B,$D3,$FE,$36,$C1,$F4,$8F,$92,$FF,$05,$FC,$7C,$C3,$FD,$A9,$5A
	.dc.b	$9E,$15,$B9,$CB,$5F,$B7,$3B,$7F,$96,$7A,$7F,$DF,$FE,$93,$FC,$63
	.dc.b	$0D,$DF,$2D,$F4,$0D,$78,$B2,$7F,$CD,$75,$FE,$3F,$7F,$C7,$ED,$7E
	.dc.b	$33,$5F,$EC,$EF,$62,$47,$DC,$F9,$3F,$F0,$87,$C9,$BE,$83,$FC,$5D
	.dc.b	$F4,$9F,$E5,$4C,$3F,$42,$D7,$EF,$CE,$DF,$F6,$FB,$5C,$CF,$43,$9F
	.dc.b	$87,$F1,$BE,$93,$FC,$6B,$E9,$4A,$F8,$F1,$96,$90,$FC,$AF,$C8,$7F
	.dc.b	$B2,$1F,$D3,$FD,$27,$F9,$93,$C2,$34,$91,$FE,$7C,$8C,$BC,$08,$7F
	.dc.b	$94,$7C,$8F,$F2,$67,$C1,$FF,$12,$EB,$FF,$06,$F7,$FF,$9D,$67,$26
	.dc.b	$B4,$B7,$FE,$B2,$8A,$FF,$96,$E3,$F8,$17,$CC,$79,$49,$1E,$36,$BF
	.dc.b	$BE,$F1,$7F,$81,$C7,$08,$FF,$5E,$F9,$3F,$E8,$8C,$FF,$D3,$FB,$FF
	.dc.b	$AA,$9B,$3F,$EA,$F6,$BF,$E8,$06,$BF,$A7,$73,$D3,$FE,$CF,$6E,$9F
	.dc.b	$E2,$A6,$CB,$47,$C7,$7F,$30,$AD,$7C,$07,$F7,$5A,$FF,$63,$B5,$FE
	.dc.b	$2F,$7C,$AF,$8C,$7A,$67,$58,$78,$5B,$7D,$31,$FF,$3E,$63,$DE,$FF
	.dc.b	$92,$D3,$E9,$3E,$03,$63,$E0,$35,$F8,$0D,$7E,$03,$9C,$3C,$13,$E3
	.dc.b	$3B,$F9,$7F,$E6,$FF,$8D,$7D,$27,$F8,$77,$07,$D1,$BE,$7F,$F8,$85
	.dc.b	$7E,$76,$31,$3F,$E5,$28,$C4,$EA,$B9,$56,$A9,$5D,$B1,$FF,$2C,$FD
	.dc.b	$26,$63,$F5,$0D,$EA,$FE,$3C,$E7,$0F,$23,$6B,$FC,$C3,$86,$F0,$B8
	.dc.b	$F9,$B0,$FE,$4D,$3B,$9F,$5C,$F9,$6B,$71,$E9,$AD,$CF,$18,$C9,$E8
	.dc.b	$C3,$D7,$1C,$61,$87,$85,$B1,$C5,$A7,$C5,$9F,$4C,$71,$C3,$81,$1C
	.dc.b	$12,$7C,$87,$8B,$1F,$9A,$F3,$C7,$17,$E4,$8D,$9E,$E2,$DF,$C1,$3E
	.dc.b	$16,$1E,$72,$61,$EB,$74,$FA,$63,$C9,$3D,$CE,$87,$C3,$BF,$B3,$3E
	.dc.b	$FA,$E7,$3B,$7B,$C6,$49,$19,$7E,$57,$0F,$CA,$8E,$CD,$7B,$1F,$E0
	.dc.b	$69,$D9,$FE,$FC,$5F,$3F,$FB,$98,$C3,$63,$FE,$00,$B5,$FE,$D0,$D7
	.dc.b	$FD,$52,$FA,$9F,$ED,$9A,$38,$71,$1C,$0F,$B6,$7F,$AC,$23,$4E,$0E
	.dc.b	$65,$73,$5F,$A1,$8C,$B6,$33,$87,$A2,$7E,$68,$8E,$B9,$71,$8F,$79
	.dc.b	$A8,$CC,$F1,$AE,$97,$E8,$AE,$F3,$FE,$07,$7A,$46,$5E,$77,$C1,$F0
	.dc.b	$98,$9E,$17,$DA,$3E,$59,$E1,$AF,$D2,$BF,$02,$B9,$1E,$7C,$33,$5C
	.dc.b	$67,$E5,$4E,$AF,$D0,$76,$BF,$25,$FA,$6F,$E2,$DF,$82,$F8,$36,$F8
	.dc.b	$3E,$8A,$EF,$7E,$48,$F5,$1F,$71,$C4,$75,$7B,$C7,$D7,$6D,$76,$8F
	.dc.b	$3D,$76,$8F,$E3,$F8,$B7,$7C,$5F,$E0,$B4,$FE,$78,$DC,$7F,$B4,$F0
	.dc.b	$FF,$99,$9F,$E8,$F1,$9C,$BF,$9E,$BF,$E8,$C7,$FA,$6B,$5C,$C7,$F3
	.dc.b	$E8,$C6,$1E,$1F,$83,$E4,$D9,$FF,$3A,$DB,$D1,$CE,$4B,$A4,$AD,$CD
	.dc.b	$C7,$FA,$CD,$6E,$7F,$AE,$D7,$9F,$97,$FB,$1D,$EF,$3F,$EB,$53,$E4
	.dc.b	$BE,$18,$7F,$52,$8C,$1E,$BE,$8F,$F3,$FF,$7F,$C1,$6E,$5E,$8F,$3F
	.dc.b	$1A,$AA,$9C,$5C,$4F,$F8,$6E,$30,$7C,$7D,$3F,$BB,$FA,$1E,$7D,$8D
	.dc.b	$E1,$C3,$2F,$E3,$B9,$69,$FE,$95,$FA,$0F,$F4,$AB,$71,$E1,$AF,$F8
	.dc.b	$5E,$72,$D8,$F7,$46,$23,$FC,$ED,$A2,$31,$01,$19,$9F,$C0,$6E,$F9
	.dc.b	$97,$C2,$7E,$1E,$30,$FD,$3F,$0D,$F2,$79,$AD,$E7,$FE,$74,$6E,$5F
	.dc.b	$EC,$DE,$DD,$61,$F6,$FD,$9F,$F0,$65,$F8,$F2,$FF,$6E,$FD,$1B,$E7
	.dc.b	$3F,$2D,$8A,$31,$5F,$EC,$2F,$A4,$FE,$C2,$F1,$3F,$D2,$91,$A9,$FF
	.dc.b	$51,$B6,$FF,$93,$F2,$2F,$FC,$1F,$7C,$6F,$E2,$5E,$E3,$FC,$A3,$3A
	.dc.b	$9F,$F5,$6E,$FF,$82,$6F,$2F,$EA,$E9,$F4,$7F,$C7,$A7,$31,$D5,$FC
	.dc.b	$79,$FE,$47,$9F,$D7,$38,$3F,$C9,$0E,$2B,$B6,$5F,$CF,$AB,$AF,$BB
	.dc.b	$F1,$27,$31,$F6,$CC,$9F,$47,$DE,$FA,$87,$D7,$F7,$EC,$64,$FF,$75
	.dc.b	$EF,$FE,$8F,$71,$97,$FB,$4A,$30,$FF,$A0,$3D,$EF,$C0,$7D,$27,$E0
	.dc.b	$5B,$5F,$EF,$4E,$6F,$F7,$7B,$67,$4D,$3E,$17,$CE,$FF,$B8,$9B,$DB
	.dc.b	$F2,$78,$3F,$84,$46,$09,$8F,$F6,$FA,$E0,$DA,$F9,$31,$88,$FE,$9C
	.dc.b	$EB,$7F,$91,$76,$FF,$DB,$13,$6B,$FB,$74,$DE,$37,$C0,$7F,$93,$EF
	.dc.b	$FD,$F1,$F3,$BF,$C4,$63,$17,$FE,$F6,$51,$9E,$ED,$36,$B8,$9F,$BB
	.dc.b	$F2,$FF,$88,$77,$1E,$9F,$39,$86,$4F,$E4,$F2,$7C,$2F,$EB,$A3,$91
	.dc.b	$97,$19,$F8,$FE,$23,$CD,$EC,$79,$DB,$5C,$F5,$46,$2B,$84,$E6,$B8
	.dc.b	$34,$E1,$3B,$E0,$FD,$4D,$BD,$8F,$89,$C9,$F1,$5B,$7D,$DB,$8F,$C7
	.dc.b	$6D,$F2,$F6,$BE,$67,$C0,$7C,$D6,$DF,$9D,$B7,$F3,$DB,$5F,$E2,$08
	.dc.b	$72,$3E,$33,$F8,$3B,$6F,$E0,$7C,$1E,$26,$47,$F8,$E6,$B8,$C7,$E2
	.dc.b	$89,$19,$AE,$52,$BF,$E1,$3B,$BE,$13,$9C,$E6,$FF,$DA,$F3,$97,$F4
	.dc.b	$9F,$49,$FC,$9C,$FC,$57,$CC,$58,$FE,$09,$E9,$3E,$13,$F8,$1B,$63
	.dc.b	$CC,$DC,$3F,$C9,$B3,$FD,$3F,$89,$F3,$3F,$95,$3F,$9A,$7B,$77,$0F
	.dc.b	$C0,$7C,$27,$F2,$AC,$5E,$1F,$E6,$E3,$0C,$9E,$EF,$88,$8F,$F9,$7F
	.dc.b	$A7,$F4,$BF,$49,$FA,$33,$86,$9F,$A3,$7F,$9B,$C7,$E5,$38,$7E,$B3
	.dc.b	$6B,$FD,$F7,$CF,$FC,$B6,$FF,$DF,$BE,$7F,$03,$F9,$1D,$7E,$72,$3F
	.dc.b	$C9,$0D,$7F,$AC,$1F,$EE,$58,$D9,$FD,$4F,$C9,$E5,$7F,$C4,$DF,$1F
	.dc.b	$24,$8F,$D1,$73,$5B,$FB,$07,$F6,$3C,$3E,$C3,$EF,$AF,$19,$2B,$F1
	.dc.b	$13,$AC,$BF,$47,$F1,$BC,$DB,$3F,$CF,$FE,$91,$AF,$F5,$A6,$C7,$F4
	.dc.b	$FF,$49,$F4,$8F,$8A,$D5,$CA,$FC,$B5,$F0,$D3,$FD,$20,$D7,$FC,$D8
	.dc.b	$D7,$E2,$BB,$C7,$FC,$89,$F4,$CE,$5E,$95,$F0,$25,$39,$BF,$EA,$DF
	.dc.b	$49,$F5,$68,$FD,$7D,$3C,$FF,$4B,$FD,$D2,$34,$7F,$6F,$9F,$F3,$F3
	.dc.b	$5F,$AE,$D7,$FC,$20,$D7,$AF,$A9,$FE,$53,$F7,$DF,$71,$FC,$97,$1F
	.dc.b	$1C,$E6,$31,$5F,$60,$7C,$CF,$F8,$3B,$4F,$E0,$8D,$BF,$C3,$FB,$5F
	.dc.b	$CD,$9F,$EE,$51,$9A,$FF,$11,$3F,$E1,$67,$FD,$2C,$D7,$F1,$ED,$77
	.dc.b	$EC,$FF,$78,$8E,$5E,$F3,$E8,$9E,$77,$F8,$62,$73,$18,$C3,$C3,$EF
	.dc.b	$E4,$FF,$0E,$E4,$F3,$31,$3A,$3F,$8B,$5F,$F9,$73,$3F,$2D,$F8,$4F
	.dc.b	$F2,$87,$91,$FE,$C7,$AF,$E8,$9F,$4D,$FE,$67,$8C,$57,$E7,$BB,$3F
	.dc.b	$A5,$CE,$6B,$4D,$78,$9F,$95,$F2,$BF,$C3,$5C,$1F,$EC,$EE,$6F,$E5
	.dc.b	$FC,$9F,$E5,$DF,$4B,$FC,$D5,$5A,$BF,$F2,$65,$69,$E5,$5B,$69,$FA
	.dc.b	$1F,$01,$E5,$7F,$9C,$DB,$1F,$D8,$7E,$2F,$FC,$7B,$AF,$ED,$B0,$FE
	.dc.b	$7D,$7B,$25,$6F,$E1,$6D,$6F,$0E,$6F,$FB,$46,$F8,$D7,$FB,$E5,$B1
	.dc.b	$99,$DB,$FC,$3B,$DC,$FF,$32,$FD,$37,$FA,$7A,$8F,$F9,$F7,$9F,$0C
	.dc.b	$C3,$FC,$95,$5C,$1F,$F5,$1C,$E6,$FF,$31,$F4,$CF,$96,$FF,$77,$47
	.dc.b	$FB,$94,$46,$76,$FE,$AC,$D8,$D3,$5F,$F7,$67,$C5,$7F,$B0,$BA,$FF
	.dc.b	$AE,$36,$3A,$35,$DF,$C4,$7F,$B2,$5E,$7E,$5B,$4E,$ED,$8C,$C6,$0F
	.dc.b	$91,$F4,$DE,$88,$FC,$67,$C4,$7F,$83,$1F,$F6,$0F,$B3,$FC,$DB,$A3
	.dc.b	$F1,$15,$B6,$EB,$9E,$0F,$F6,$AD,$5B,$75,$31,$E9,$1F,$81,$C7,$95
	.dc.b	$FE,$3B,$E9,$9E,$DF,$1C,$1D,$3F,$CA,$F0,$FC,$57,$99,$D7,$9D,$87
	.dc.b	$09,$FE,$64,$D7,$F4,$D5,$C5,$B5,$C6,$75,$B7,$E8,$DB,$97,$F4,$9B
	.dc.b	$E1,$3F,$E8,$E6,$C7,$9F,$2F,$16,$3F,$A0,$C7,$FD,$CF,$1E,$9D,$39
	.dc.b	$C7,$FA,$D2,$74,$FE,$85,$B8,$7F,$A6,$A7,$79,$FC,$7B,$EB,$D5,$F0
	.dc.b	$FE,$0D,$C7,$7E,$9A,$67,$FE,$80,$8D,$18,$7F,$D0,$11,$87,$FD,$31
	.dc.b	$F4,$AF,$86,$FA,$DD,$1F,$E0,$AF,$8B,$BE,$BF,$FB,$7A,$F7,$5F,$6D
	.dc.b	$5F,$17,$FD,$5D,$7C,$5B,$5C,$5F,$E7,$93,$B6,$FC,$97,$6D,$64,$79
	.dc.b	$23,$B4,$9C,$C7,$F7,$CB,$FF,$0F,$D6,$A6,$4F,$82,$37,$C1,$FE,$4B
	.dc.b	$BE,$53,$FE,$B7,$FA,$57,$B5,$EB,$E9,$F3,$5F,$F9,$45,$BC,$BF,$E1
	.dc.b	$4A,$DD,$F1,$BE,$1B,$48,$C1,$0F,$EB,$57,$C9,$FF,$7A,$35,$57,$47
	.dc.b	$FB,$07,$D2,$3E,$93,$E7,$7D,$2F,$E3,$A7,$FD,$F8,$D5,$3B,$7F,$DA
	.dc.b	$4D,$7F,$B5,$46,$CE,$11,$9C,$FF,$CD,$CE,$E7,$FC,$62,$74,$AF,$F3
	.dc.b	$53,$4F,$F3,$64,$6D,$FF,$1B,$FD,$2F,$F7,$67,$FD,$9F,$E8,$FE,$54
	.dc.b	$FC,$34,$66,$3F,$BE,$36,$FF,$DE,$FC,$CA,$E1,$5F,$EE,$BA,$D6,$DD
	.dc.b	$BC,$EE,$64,$9F,$F6,$EF,$9D,$DB,$C6,$FE,$0D,$F3,$DF,$DD,$A7,$FD
	.dc.b	$5C,$D7,$C3,$1B,$73,$E4,$70,$30,$E8,$F1,$46,$34,$63,$FE,$4B,$CF
	.dc.b	$FD,$7D,$5F,$EE,$EF,$A5,$FD,$A1,$11,$FE,$1B,$FA,$5F,$F4,$D3,$8B
	.dc.b	$FF,$4F,$46,$23,$FD,$E9,$3C,$5E,$1E,$67,$EF,$FE,$97,$F6,$15,$BB
	.dc.b	$E3,$39,$AF,$F7,$FF,$D2,$FF,$05,$23,$F0,$B9,$B8,$FC,$07,$C6,$64
	.dc.b	$78,$F4,$FF,$38,$E2,$AF,$97,$D7,$2B,$E5,$BF,$C4,$7E,$97,$F8,$17
	.dc.b	$D2,$FF,$A4,$27,$FD,$B1,$C9,$FC,$DF,$D6,$FB,$31,$F0,$A7,$4F,$F8
	.dc.b	$F5,$BB,$7F,$AC,$2B,$81,$D3,$C1,$95,$D0,$87,$17,$F9,$17,$81,$FC
	.dc.b	$8F,$E7,$A4,$FF,$BB,$7A,$BF,$7E,$FF,$89,$5B,$87,$E4,$27,$73,$9A
	.dc.b	$FE,$59,$F3,$DF,$D3,$FE,$2B,$F4,$77,$FE,$E4,$AC,$9D,$67,$4C,$30
	.dc.b	$7F,$BC,$FD,$0E,$5F,$2F,$23,$04,$3F,$C7,$16,$8D,$46,$27,$C4,$62
	.dc.b	$FF,$DE,$B3,$89,$FF,$47,$7A,$4F,$97,$FA,$1D,$92,$73,$1F,$D9,$67
	.dc.b	$51,$F7,$CF,$EE,$7C,$3F,$E1,$77,$1E,$6A,$31,$1F,$E3,$59,$E2,$FA
	.dc.b	$3E,$5B,$E3,$78,$F2,$38,$EA,$9F,$94,$66,$3D,$F3,$87,$C4,$7F,$BB
	.dc.b	$FE,$97,$FA,$67,$D2,$7F,$B8,$23,$2D,$7F,$D0,$9F,$16,$9F,$EA,$4D
	.dc.b	$8D,$FC,$7F,$23,$33,$23,$24,$39,$BF,$C2,$3E,$53,$F8,$4F,$C9,$C9
	.dc.b	$FE,$13,$19,$7F,$85,$C6,$5F,$F3,$87,$8E,$9F,$F0,$F7,$1B,$E9,$7F
	.dc.b	$E0,$05,$7E,$4F,$DA,$FE,$F0,$6A,$33,$1F,$EB,$76,$CF,$F2,$49,$DC
	.dc.b	$FD,$78,$E9,$F2,$7F,$DC,$7E,$97,$FC,$17,$2D,$9F,$2D,$F2,$8D,$78
	.dc.b	$52,$7E,$5C,$65,$FA,$35,$C8,$E7,$5C,$A3,$66,$AB,$FC,$01,$F2,$6F
	.dc.b	$A2,$FF,$1A,$7D,$2B,$C5,$A6,$A2,$5F,$E3,$27,$4D,$9C,$BF,$07,$E9
	.dc.b	$7F,$C3,$9F,$35,$23,$E0,$70,$27,$E4,$19,$8F,$DE,$4E,$1F,$D1,$FD
	.dc.b	$2F,$F8,$BF,$E6,$64,$60,$87,$F9,$71,$CB,$F9,$D8,$FF,$27,$7C,$1F
	.dc.b	$F0,$6F,$37,$FC,$81,$E3,$60,$CF,$8F,$6F,$FD,$35,$15,$FF,$26,$C7
	.dc.b	$EB,$FE,$5B,$FC,$81,$F1,$F2,$7F,$6F,$19,$7F,$5F,$F0,$9F,$EC,$56
	.dc.b	$CE,$57,$47,$F5,$07,$FC,$F5,$F0,$1F,$D4,$1F,$F3,$AF,$47,$F9,$34
	.dc.b	$FF,$37,$78,$89,$D5,$71,$8C,$70,$6E,$FC,$37,$C2,$F7,$AF,$F5,$8B
	.dc.b	$DD,$68,$87,$92,$35,$3C,$23,$CB,$7F,$E9,$F6,$9F,$EA,$1A,$E6,$FF
	.dc.b	$83,$9A,$FF,$AC,$E5,$F4,$5F,$E6,$7E,$D9,$3F,$EB,$09,$CB,$67,$2F
	.dc.b	$FA,$5A,$7D,$51,$FE,$55,$F8,$A7,$D2,$FC,$AA,$7D,$2B,$E8,$3F,$BA
	.dc.b	$C2,$BF,$38,$FF,$78,$9D,$3F,$DE,$A3,$2F,$29,$E3,$1D,$B6,$E1,$18
	.dc.b	$BF,$05,$F2,$BE,$DD,$7E,$08,$C6,$1D,$1B,$5C,$2F,$96,$1D,$30,$F1
	.dc.b	$61,$C6,$7C,$2D,$8F,$35,$71,$6D,$71,$7C,$25,$C3,$C1,$7D,$A3,$D8
	.dc.b	$70,$29,$C0,$B1,$C1,$D6,$1E,$8A,$E9,$87,$8E,$78,$4E,$5E,$38,$79
	.dc.b	$5C,$E1,$EA,$7C,$91,$9C,$3E,$64,$78,$2F,$A4,$78,$63,$EC,$4E,$CE
	.dc.b	$13,$A3,$6E,$70,$F2,$CC,$3C,$6F,$80,$C6,$1F,$52,$B8,$C6,$67,$C4
	.dc.b	$6B,$0F,$91,$5E,$DD,$1F,$21,$F2,$DB,$5F,$2D,$AF,$A2,$FF,$5D,$3F
	.dc.b	$E7,$08,$7C,$F6,$1F,$F3,$9B,$5F,$CB,$B5,$FF,$26,$B5,$E2,$7D,$81
	.dc.b	$E8,$FF,$09,$C3,$F2,$51,$C0,$D4,$67,$1F,$15,$7E,$1E,$78,$F0,$D7
	.dc.b	$AD,$F2,$BE,$11,$E6,$39,$CE,$EB,$94,$EA,$7D,$26,$8F,$31,$D2,$FF
	.dc.b	$D2,$31,$B8,$F3,$5F,$7B,$E9,$1C,$63,$D6,$F1,$3F,$77,$3C,$5C,$1F
	.dc.b	$E7,$27,$17,$E4,$AE,$75,$D6,$BC,$73,$96,$9D,$EF,$E2,$B4,$F8,$13
	.dc.b	$C4,$F1,$CF,$16,$C6,$CF,$82,$EE,$79,$B6,$39,$C7,$C2,$AF,$2B,$63
	.dc.b	$F6,$B9,$7F,$3D,$F8,$AF,$E8,$99,$FF,$45,$F3,$DC,$CF,$23,$39,$7F
	.dc.b	$36,$BB,$8A,$F7,$58,$C7,$A5,$FF,$54,$8C,$72,$30,$E3,$7C,$2B,$F9
	.dc.b	$A7,$C0,$FF,$55,$C7,$F9,$D3,$5F,$E8,$AD,$7F,$AC,$B7,$1F,$EA,$CF
	.dc.b	$4C,$9F,$17,$E0,$8C,$36,$FF,$B3,$B6,$FF,$A9,$1E,$BA,$E3,$3E,$43
	.dc.b	$65,$8D,$E1,$F1,$AB,$E1,$FC,$63,$17,$8B,$E3,$F3,$73,$18,$C7,$FB
	.dc.b	$BC,$62,$75,$1F,$E7,$26,$E3,$DB,$27,$9F,$ED,$3D,$DC,$36,$FF,$C0
	.dc.b	$31,$6B,$84,$DE,$1F,$F0,$06,$6F,$A0,$F3,$61,$FE,$08,$F8,$AF,$4C
	.dc.b	$63,$0F,$04,$63,$5F,$98,$DA,$E0,$7F,$92,$FE,$5B,$FD,$E0,$FF,$A9
	.dc.b	$FB,$3F,$0B,$92,$73,$87,$FA,$A6,$73,$18,$9F,$F7,$CC,$63,$E1,$7F
	.dc.b	$0B,$ED,$FF,$7A,$C3,$FC,$FB,$F4,$8F,$9A,$FC,$4F,$D2,$7F,$74,$9C
	.dc.b	$C6,$3C,$0F,$CA,$E8,$BE,$15,$FE,$94,$BE,$13,$F8,$03,$57,$C6,$7F
	.dc.b	$D1,$7B,$7D,$79,$FC,$2C,$3E,$8F,$FD,$5C,$EB,$97,$F4,$6E,$7D,$EF
	.dc.b	$0D,$F1,$DF,$FC,$61,$39,$38,$1F,$EA,$93,$18,$78,$E7,$84,$7F,$C0
	.dc.b	$03,$33,$FA,$58,$CC,$A7,$21,$FF,$1F,$F0,$3E,$47,$E5,$F8,$7F,$41
	.dc.b	$B5,$F6,$5B,$5F,$A2,$C3,$EC,$46,$1F,$E0,$9E,$47,$FB,$16,$BF,$D7
	.dc.b	$AB,$E9,$3F,$76,$E6,$FF,$4A,$E6,$77,$EF,$EA,$B6,$D8,$FF,$5C,$FC
	.dc.b	$CF,$FA,$FD,$F8,$9B,$DC,$3E,$B4,$DC,$7E,$5C,$DA,$FF,$18,$CD,$AF
	.dc.b	$89,$35,$46,$67,$ED,$7B,$DF,$86,$8C,$56,$DB,$1F,$C5,$BD,$FF,$F1
	.dc.b	$F5,$7F,$B5,$3E,$6B,$1D,$B9,$6C,$64,$FF,$6C,$C6,$5F,$F7,$37,$6F
	.dc.b	$EC,$7C,$DF,$E4,$3F,$02,$F9,$8F,$D8,$C6,$1B,$5F,$EE,$A9,$D3,$CB
	.dc.b	$C0,$56,$E3,$FD,$CF,$F1,$5E,$7F,$73,$D0,$D8,$C9,$FE,$EC,$8C,$B4
	.dc.b	$CF,$07,$F1,$CD,$53,$A8,$C3,$F0,$FD,$8F,$89,$CB,$6F,$FD,$EE,$DA
	.dc.b	$E4,$DE,$3F,$E9,$98,$23,$E6,$7A,$7F,$37,$D3,$B8,$7F,$C0,$06,$E1
	.dc.b	$E3,$6E,$1C,$9F,$B1,$F1,$9F,$2B,$E1,$2D,$FC,$76,$DF,$F3,$6E,$04
	.dc.b	$63,$0F,$EB,$D1,$8A,$FC,$8D,$F3,$69,$CC,$FC,$FF,$D2,$7E,$78,$FC
	.dc.b	$27,$99,$63,$F6,$17,$FD,$67,$D3,$7C,$17,$E9,$3D,$8F,$E2,$71,$8A
	.dc.b	$FD,$4C,$7E,$37,$CF,$91,$83,$99,$FB,$BF,$41,$C4,$9F,$CA,$79,$DD
	.dc.b	$3E,$22,$DF,$0E,$1F,$F1,$DC,$66,$31,$87,$2F,$A3,$C5,$7D,$F7,$F3
	.dc.b	$C4,$9E,$51,$FE,$4A,$AC,$3F,$C9,$EB,$F2,$6D,$EB,$FE,$F7,$CF,$FC
	.dc.b	$BA,$31,$7C,$AB,$7B,$FF,$0A,$F2,$7F,$9A,$C6,$27,$DD,$D1,$FC,$72
	.dc.b	$FF,$C1,$98,$FF,$0C,$BD,$B6,$BA,$73,$72,$3F,$6D,$CB,$63,$27,$C0
	.dc.b	$E0,$7C,$66,$71,$FC,$F3,$74,$FC,$AB,$6B,$E8,$57,$07,$E1,$FD,$1B
	.dc.b	$83,$FC,$25,$82,$31,$3F,$ED,$76,$C7,$F9,$CD,$AF,$FB,$18,$E1,$3F
	.dc.b	$D9,$BE,$8D,$C9,$FD,$62,$B8,$E5,$FE,$34,$9C,$E1,$E2,$BE,$17,$B6
	.dc.b	$C7,$D8,$FA,$2F,$A5,$7F,$EF,$FF,$71,$F4,$5F,$C2,$2B,$5E,$87,$F8
	.dc.b	$EA,$73,$E1,$7F,$68,$C3,$8D,$7F,$97,$EF,$86,$BF,$D9,$E7,$2F,$FA
	.dc.b	$12,$33,$8F,$F8,$E6,$73,$87,$4B,$EE,$7E,$55,$C5,$FF,$71,$F8,$FF
	.dc.b	$F0,$14,$FF,$59,$F7,$1F,$44,$F9,$6F,$EF,$75,$AB,$56,$A7,$3E,$CF
	.dc.b	$F9,$52,$31,$87,$E6,$AF,$84,$62,$FC,$15,$BC,$3B,$5E,$E3,$18,$78
	.dc.b	$3D,$DD,$78,$B6,$BF,$5D,$8F,$9A,$31,$8F,$F9,$2A,$31,$3E,$1A,$E2
	.dc.b	$FF,$3B,$3F,$A0,$FD,$17,$F4,$26,$99,$3F,$A0,$39,$69,$FD,$2B,$CF
	.dc.b	$FF,$55,$DF,$F9,$17,$8F,$FC,$A3,$18,$E0,$FF,$3D,$7B,$DF,$8A,$BF
	.dc.b	$F0,$27,$9B,$FD,$8B,$87,$FC,$CD,$87,$06,$CF,$F5,$6E,$BF,$F3,$0E
	.dc.b	$5F,$E5,$EF,$86,$7C,$AF,$F9,$7A,$73,$9F,$FA,$1E,$73,$7C,$7C,$9F
	.dc.b	$F3,$EC,$63,$0F,$F3,$6E,$6C,$38,$DF,$F1,$DA,$D6,$1F,$E5,$69,$CD
	.dc.b	$FF,$A3,$67,$35,$EA,$9C,$F8,$69,$CD,$7F,$9C,$70,$7D,$17,$FA,$0E
	.dc.b	$FF,$D2,$91,$8C,$3F,$91,$ED,$E3,$7F,$94,$7B,$49,$D6,$1F,$E7,$1B
	.dc.b	$E3,$87,$3B,$FE,$4B,$F4,$4F,$94,$FF,$72,$DF,$FB,$C6,$31,$C1,$FE
	.dc.b	$F3,$84,$61,$D4,$F0,$7F,$82,$7C,$47,$FA,$97,$9B,$83,$63,$FD,$F8
	.dc.b	$D7,$7F,$0B,$FA,$27,$C1,$E3,$FF,$A4,$5B,$B7,$87,$1F,$F0,$73,$67
	.dc.b	$53,$FE,$47,$3F,$D3,$1F,$9A,$BD,$5C,$72,$7F,$93,$67,$31,$8D,$BF
	.dc.b	$C9,$D8,$2F,$85,$7C,$CF,$23,$4F,$03,$E6,$C6,$9F,$F5,$7C,$F1,$BD
	.dc.b	$F2,$FF,$B4,$2F,$8C,$F5,$F3,$2E,$06,$9C,$3F,$0D,$FF,$64,$67,$E7
	.dc.b	$AE,$19,$78,$2B,$6F,$F6,$2A,$DE,$7E,$AC,$3A,$4E,$C8,$7F,$AF,$A7
	.dc.b	$6F,$E1,$7E,$29,$F6,$1E,$EE,$74,$62,$7F,$C7,$ED,$79,$77,$6E,$31
	.dc.b	$A7,$FA,$D1,$B9,$7E,$D9,$B5,$FE,$ED,$C5,$7F,$1B,$D6,$CF,$B2,$FD
	.dc.b	$2D,$EC,$7D,$08,$C3,$BD,$66,$3F,$E0,$59,$CB,$6B,$8B,$FE,$4F,$9F
	.dc.b	$B2,$FF,$93,$BC,$DB,$E6,$7F,$56,$8F,$F0,$2B,$5F,$E4,$1F,$9B,$7F
	.dc.b	$96,$DC,$CF,$02,$3F,$CA,$23,$4F,$FA,$C6,$7F,$DC,$EF,$4F,$4F,$B9
	.dc.b	$E2,$F6,$9E,$BE,$27,$44,$3F,$D7,$51,$A7,$1C,$5F,$99,$7F,$CC,$F5
	.dc.b	$C5,$DD,$7B,$BE,$91,$F1,$34,$93,$FC,$FB,$9B,$FD,$E2,$FF,$89,$27
	.dc.b	$6D,$9D,$BF,$D0,$67,$6F,$FB,$36,$76,$D9,$DE,$1F,$AD,$F0,$3F,$55
	.dc.b	$F1,$9E,$3F,$02,$5F,$77,$F5,$73,$FA,$FF,$7B,$33,$D6,$6D,$F3,$CB
	.dc.b	$FC,$F3,$3C,$48,$6C,$F6,$57,$DE,$FA,$17,$D1,$7F,$17,$94,$61,$FE
	.dc.b	$34,$D7,$FC,$34,$6A,$38,$46,$6F,$E8,$37,$AF,$F8,$73,$6B,$FD,$AD
	.dc.b	$F4,$7F,$C4,$AB,$77,$E3,$24,$FF,$94,$A7,$57,$FC,$5F,$C6,$FF,$6E
	.dc.b	$4F,$F9,$5B,$E0,$5F,$27,$F9,$86,$C7,$F9,$6A,$76,$FF,$BF,$DB,$8F
	.dc.b	$E1,$5B,$7D,$7E,$27,$97,$CB,$FE,$8A,$73,$7F,$EE,$D6,$BF,$D0,$67
	.dc.b	$6F,$F9,$A3,$E1,$BF,$03,$87,$E4,$5B,$87,$F2,$F6,$CF,$FB,$DF,$F4
	.dc.b	$7F,$E4,$95,$F9,$76,$D7,$F3,$56,$CE,$FC,$9F,$DA,$DB,$EA,$E9,$FB
	.dc.b	$8A,$DD,$1B,$3B,$8F,$F6,$67,$B7,$FC,$2C,$8F,$FB,$3F,$5F,$E0,$3F
	.dc.b	$85,$FF,$10,$8F,$F4,$1B,$56,$7F,$AD,$8C,$3F,$EA,$CF,$86,$FE,$2C
	.dc.b	$49,$FF,$6D,$3F,$C6,$78,$BF,$86,$FD,$5B,$E6,$A9,$2B,$FD,$23,$F8
	.dc.b	$5F,$E0,$10,$9E,$0F,$FA,$59,$AF,$93,$83,$FA,$E4,$63,$1F,$E0,$55
	.dc.b	$CE,$B9,$5F,$77,$81,$DA,$B8,$5F,$F6,$5A,$F1,$7A,$0F,$25,$F0,$9F
	.dc.b	$CC,$FE,$3E,$63,$FE,$F7,$8F,$E6,$C6,$1F,$F6,$94,$61,$FF,$7E,$7C
	.dc.b	$87,$F3,$BC,$E4,$60,$87,$FB,$65,$CB,$FE,$DB,$8F,$E8,$5E,$C7,$F4
	.dc.b	$3F,$9B,$93,$FE,$BE,$C7,$FC,$42,$D9,$CB,$FE,$E3,$9E,$2E,$4E,$0E
	.dc.b	$B6,$FF,$72,$CF,$F5,$DE,$FF,$D0,$FA,$5F,$DF,$FA,$B0,$7F,$53,$F3
	.dc.b	$EE,$3F,$ED,$19,$D3,$6B,$F5,$3B,$FC,$9A,$FE,$B5,$E6,$7F,$5B,$F2
	.dc.b	$65,$FA,$63,$2F,$AE,$32,$FF,$BE,$E3,$FA,$AF,$D6,$BE,$92,$92,$7F
	.dc.b	$DB,$3F,$35,$F9,$4E,$6F,$F0,$F5,$6E,$7F,$DB,$AD,$7F,$02,$1C,$6F
	.dc.b	$FD,$26,$B9,$4F,$E4,$A7,$35,$D2,$FC,$13,$CA,$B7,$9F,$E1,$FE,$7B
	.dc.b	$FC,$3B,$F0,$BF,$8A,$3F,$17,$C6,$8F,$EF,$86,$BD,$99,$87,$DB,$8F
	.dc.b	$EE,$7F,$58,$F0,$69,$27,$FD,$FB,$F4,$0F,$AC,$C4,$7B,$2B,$83,$FE
	.dc.b	$FC,$6B,$FA,$0C,$D5,$C6,$3F,$E0,$0F,$E1,$6F,$53,$FC,$06,$D8,$CC
	.dc.b	$7F,$8E,$3E,$E9,$F1,$DA,$89,$3A,$3F,$73,$C9,$FE,$4B,$C5,$7C,$27
	.dc.b	$47,$D2,$C3,$B3,$6F,$B6,$1F,$E2,$FF,$A2,$7C,$CD,$25,$76,$E5,$FE
	.dc.b	$0F,$93,$8F,$FC,$6B,$3F,$95,$BE,$2D,$BE,$2F,$C3,$9D,$BF,$E9,$89
	.dc.b	$FE,$63,$D6,$69,$FC,$D2,$FB,$7D,$37,$97,$CD,$97,$DD,$FC,$0C,$7A
	.dc.b	$7E,$62,$4F,$8D,$FF,$2D,$69,$88,$F0,$45,$E0,$F9,$1F,$03,$F1,$E7
	.dc.b	$FC,$A9,$5C,$1F,$F2,$E3,$FE,$63,$F4,$31,$1A,$32,$7F,$98,$5F,$95
	.dc.b	$F3,$32,$3E,$23,$FE,$6B,$F8,$6C,$47,$C1,$8B,$87,$8A,$7F,$CE,$BF
	.dc.b	$07,$FE,$89,$AF,$F3,$DB,$67,$FC,$E4,$DE,$2F,$A7,$93,$E6,$78,$DC
	.dc.b	$3F,$D6,$35,$B7,$FC,$C1,$19,$7A,$96,$FE,$1C,$5B,$FF,$49,$D7,$1E
	.dc.b	$CF,$EA,$D7,$FE,$79,$AF,$3F,$B9,$23,$BB,$FE,$89,$F8,$2C,$47,$AA
	.dc.b	$2E,$3F,$42,$7F,$CC,$7F,$13,$FE,$46,$8F,$EA,$7C,$9F,$8A,$B4,$BE
	.dc.b	$77,$FD,$59,$39,$8C,$4F,$0E,$8F,$F5,$84,$73,$8F,$C6,$46,$3E,$03
	.dc.b	$FB,$6C,$60,$FE,$C3,$3C,$9C,$1E,$0B,$EF,$5C,$70,$ED,$3C,$2F,$9E
	.dc.b	$1E,$3B,$DC,$F4,$6B,$EA,$9D,$DF,$38,$EF,$87,$86,$F9,$47,$63,$86
	.dc.b	$16,$7D,$93,$A9,$E3,$87,$69,$CD,$F5,$AD,$11,$C1,$2B,$CC,$75,$AE
	.dc.b	$58,$79,$A3,$DD,$86,$58,$7A,$23,$4F,$98,$F1,$9A,$BF,$AD,$1A,$C3
	.dc.b	$CB,$3E,$03,$67,$3A,$F7,$4E,$EF,$D8,$79,$23,$27,$5C,$3E,$FC,$66
	.dc.b	$B9,$0D,$3F,$87,$E1,$F8,$36,$E1,$F8,$26,$E1,$FE,$F0,$F8,$7F,$F3
	.dc.b	$8F,$CD,$DC,$3F,$64,$D7,$FD,$F8,$D7,$FC,$52,$D4,$FD,$95,$F1,$9F
	.dc.b	$EF,$FD,$3F,$39,$1C,$0F,$C7,$9A,$E1,$E0,$F2,$33,$3C,$75,$F2,$C6
	.dc.b	$27,$77,$DE,$B5,$3E,$83,$47,$63,$A5,$F8,$A3,$71,$EC,$37,$1E,$67
	.dc.b	$77,$FC,$4E,$77,$1E,$99,$E5,$7E,$28,$E3,$5E,$69,$CC,$78,$6F,$CD
	.dc.b	$5B,$9C,$9D,$E3,$33,$B3,$99,$EB,$8D,$D7,$53,$C9,$5C,$DB,$3C,$67
	.dc.b	$6F,$18,$CC,$77,$8E,$A7,$C1,$76,$D8,$D9,$FC,$8B,$E5,$55,$FC,$B1
	.dc.b	$B3,$FD,$FF,$8D,$ED,$E3,$4F,$E6,$F5,$6F,$91,$C4,$FE,$D9,$6E,$C6
	.dc.b	$1C,$A7,$53,$FD,$0B,$BF,$93,$F9,$DE,$6F,$F0,$54,$F2,$BE,$3E,$DE
	.dc.b	$B8,$55,$AB,$E5,$7F,$D1,$30,$E7,$87,$08,$C1,$27,$67,$B7,$0E,$97
	.dc.b	$CF,$3F,$EA,$78,$60,$E7,$92,$B1,$87,$F8,$17,$E4,$FD,$3E,$07,$F7
	.dc.b	$F7,$7E,$DE,$38,$51,$49,$5E,$DC,$15,$8D,$3E,$87,$B7,$E0,$3C,$3E
	.dc.b	$CD,$BE,$11,$67,$EB,$4D,$D7,$EC,$E6,$FB,$0F,$F1,$7E,$1F,$E6,$F6
	.dc.b	$DF,$68,$C7,$83,$EA,$9C,$DF,$26,$CF,$F7,$4C,$75,$5B,$EF,$93,$F5
	.dc.b	$FC,$57,$C4,$E3,$2F,$F5,$3E,$7E,$3B,$C6,$7F,$E3,$DB,$DE,$1E,$59
	.dc.b	$F1,$73,$C8,$E9,$F5,$37,$1E,$7A,$FE,$2D,$B5,$F8,$0C,$15,$BC,$7F
	.dc.b	$37,$4F,$B5,$7B,$5F,$90,$8C,$3F,$88,$FB,$8F,$AF,$9F,$E6,$A7,$2D
	.dc.b	$AE,$2D,$AE,$1C,$9C,$E7,$35,$FE,$86,$6E,$1E,$26,$DF,$F9,$43,$0E
	.dc.b	$F5,$BC,$3C,$9F,$96,$BC,$8A,$7E,$73,$8D,$F4,$38,$C7,$CD,$5B,$6D
	.dc.b	$6E,$75,$87,$F9,$5F,$E0,$BE,$37,$AD,$F1,$DB,$7F,$65,$B3,$B8,$FB
	.dc.b	$93,$F1,$D7,$C5,$D3,$F7,$2B,$E8,$BF,$D2,$0F,$C9,$36,$58,$E0,$FF
	.dc.b	$06,$38,$7A,$D8,$BF,$F5,$6C,$70,$6C,$F6,$AE,$B1,$89,$FC,$87,$95
	.dc.b	$F8,$1F,$27,$F0,$4D,$8C,$9F,$95,$8C,$B4,$FD,$27,$BB,$FA,$13,$F1
	.dc.b	$58,$58,$C9,$2F,$FD,$01,$19,$7F,$D0,$5E,$BF,$E4,$7E,$77,$F2,$4D
	.dc.b	$AF,$D2,$36,$77,$1F,$A7,$9D,$B4,$D9,$63,$FD,$1D,$D1,$F9,$DF,$B0
	.dc.b	$FC,$F3,$67,$F6,$0D,$9D,$C7,$EC,$E7,$F4,$9F,$61,$FA,$5F,$9D,$B3
	.dc.b	$FB,$76,$D7,$FA,$6D,$BD,$3F,$B5,$7F,$4D,$F7,$2F,$AB,$A7,$D3,$F3
	.dc.b	$6E,$1F,$A7,$9D,$36,$74,$D2,$61,$FE,$EA,$9D,$AC,$3F,$A8,$CF,$E2
	.dc.b	$EF,$AC,$FF,$AB,$FE,$3B,$F7,$C7,$EF,$FE,$1A,$D7,$ED,$FD,$37,$AB
	.dc.b	$FB,$D6,$C7,$F1,$5A,$C4,$FF,$07,$F8,$F9,$3F,$C6,$7E,$B2,$E3,$E3
	.dc.b	$83,$FE,$20,$FC,$65,$F4,$7D,$7E,$13,$FA,$CF,$B2,$9C,$DF,$8E,$74
	.dc.b	$D3,$FB,$2C,$EE,$B3,$7D,$CE,$DC,$3D,$30,$F8,$BC,$1F,$C6,$B6,$65
	.dc.b	$9A,$FE,$35,$F7,$1F,$ED,$9C,$FF,$8C,$37,$1E,$9E,$3B,$EE,$7F,$94
	.dc.b	$65,$FC,$92,$B4,$DE,$8E,$67,$F7,$4A,$E1,$E5,$E3,$1E,$B3,$FD,$83
	.dc.b	$EE,$7F,$97,$C6,$0F,$F0,$0F,$9E,$8C,$E7,$ED,$1F,$55,$FC,$83,$2F
	.dc.b	$E6,$AD,$BF,$9F,$6F,$8B,$7C,$27,$F3,$7A,$D6,$1F,$51,$AF,$4F,$3F
	.dc.b	$FB,$9D,$70,$C3,$F9,$D5,$68,$FE,$5F,$5F,$ED,$36,$BF,$78,$E5,$1F
	.dc.b	$ED,$77,$B5,$7F,$4A,$75,$EA,$3E,$2F,$FA,$53,$6B,$FA,$47,$DC,$BC
	.dc.b	$0C,$D4,$9F,$E9,$95,$BB,$F8,$97,$B9,$CD,$6E,$7E,$45,$6F,$5F,$EB
	.dc.b	$77,$B9,$CD,$7F,$5B,$FC,$F7,$F2,$F8,$7F,$59,$6C,$F6,$9D,$32,$74
	.dc.b	$FF,$63,$AE,$D1,$F0,$EB,$8D,$FF,$6A,$6C,$66,$B8,$5F,$37,$EF,$4B
	.dc.b	$E6,$FC,$87,$F6,$FF,$E0,$0F,$CC,$DF,$96,$32,$D3,$FB,$84,$6A,$B9
	.dc.b	$CE,$9B,$3A,$E8,$C6,$3E,$5F,$A3,$F8,$47,$F7,$DF,$E0,$1F,$C1,$5F
	.dc.b	$92,$33,$7F,$15,$D7,$16,$27,$F6,$5F,$06,$E6,$F8,$70,$FA,$47,$E9
	.dc.b	$CF,$4F,$FC,$21,$C3,$FE,$5A,$9C,$C6,$33,$FF,$31,$C6,$1F,$F1,$34
	.dc.b	$EE,$B9,$47,$EA,$3F,$3C,$7C,$7F,$F9,$31,$B1,$FE,$47,$F9,$BC,$4C
	.dc.b	$8E,$0F,$F9,$5A,$31,$97,$F0,$BF,$89,$7E,$7C,$FB,$4F,$F3,$53,$70
	.dc.b	$FF,$29,$7C,$E6,$26,$47,$F9,$EA,$B8,$C7,$F9,$8A,$73,$7D,$6F,$93
	.dc.b	$67,$57,$C6,$73,$C9,$FE,$8C,$C3,$C7,$F2,$1E,$23,$FC,$DB,$F3,$B6
	.dc.b	$F8,$45,$C7,$FD,$25,$36,$FF,$C7,$7F,$55,$FE,$25,$F3,$3F,$6D,$48
	.dc.b	$C3,$FB,$96,$BF,$BD,$35,$F0,$1F,$ED,$A7,$F7,$DF,$25,$FE,$1C,$FA
	.dc.b	$5E,$6D,$8C,$FB,$69,$CB,$63,$B7,$C4,$FF,$BD,$5B,$B7,$93,$14,$E6
	.dc.b	$7C,$51,$89,$FE,$1C,$3E,$7B,$F0,$33,$F4,$5A,$FD,$B6,$E9,$FE,$B6
	.dc.b	$EB,$7D,$63,$EA,$FF,$8D,$4F,$CF,$AD,$F1,$7F,$5A,$C7,$B7,$3E,$A3
	.dc.b	$13,$AC,$BF,$D7,$18,$F3,$7C,$5D,$FF,$3B,$0F,$33,$68,$D2,$4F,$02
	.dc.b	$1F,$EC,$BB,$F0,$2A,$E6,$EB,$1F,$45,$6B,$3F,$F6,$4D,$F0,$F2,$DE
	.dc.b	$97,$F7,$EC,$3F,$A7,$B5,$FF,$47,$B7,$0F,$8D,$F2,$1F,$EC,$89,$DF
	.dc.b	$B1,$CE,$B5,$87,$F5,$06,$CF,$FB,$66,$B8,$3F,$C1,$3E,$7B,$3E,$0B
	.dc.b	$EA,$DF,$07,$EC,$8C,$3E,$99,$E7,$A9,$E3,$5C,$2F,$F8,$2B,$5E,$D9
	.dc.b	$FC,$67,$FD,$57,$F4,$1F,$9E,$69,$FA,$0F,$91,$46,$5B,$19,$8E,$51
	.dc.b	$88,$FE,$E4,$DD,$BF,$DF,$FD,$2F,$AB,$FC,$BD,$7F,$A8,$69,$F4,$5F
	.dc.b	$62,$9F,$AE,$71,$E2,$FF,$12,$C6,$32,$EF,$53,$0F,$33,$E5,$23,$83
	.dc.b	$B3,$B7,$FC,$0D,$F3,$2F,$AC,$F8,$5F,$58,$F8,$CF,$17,$EB,$AF,$AE
	.dc.b	$A6,$A2,$61,$FC,$12,$B8,$36,$3E,$54,$62,$3F,$BF,$37,$83,$FD,$8E
	.dc.b	$DD,$7F,$C0,$8D,$9F,$E0,$5D,$8F,$1F,$CB,$E1,$FC,$5F,$D7,$BF,$D7
	.dc.b	$15,$FE,$48,$C7,$E6,$C6,$23,$FC,$45,$87,$FA,$A9,$FF,$67,$35,$F1
	.dc.b	$D7,$98,$8F,$AD,$EE,$F0,$7E,$A6,$1F,$DB,$E3,$FD,$B1,$F5,$EF,$97
	.dc.b	$FC,$57,$D7,$FF,$8A,$A7,$FD,$B8,$D7,$FC,$69,$3B,$78,$CF,$E3,$E3
	.dc.b	$11,$FE,$23,$6B,$FC,$AD,$AF,$F8,$A5,$DE,$5F,$EF,$06,$E9,$9A,$F1
	.dc.b	$F8,$5F,$9D,$F2,$5F,$19,$F9,$FF,$A0,$7C,$65,$35,$13,$3F,$E7,$F5
	.dc.b	$C1,$B1,$FA,$58,$C4,$7F,$8E,$5B,$9B,$2F,$F1,$F4,$61,$B5,$E1,$F8
	.dc.b	$DF,$D8,$7C,$B7,$DD,$C3,$ED,$57,$F9,$16,$31,$1F,$E5,$B6,$E1,$D6
	.dc.b	$3F,$D0,$FF,$5D,$FC,$AD,$3F,$CE,$9F,$EA,$4D,$7F,$CA,$0D,$78,$F9
	.dc.b	$7F,$BF,$A7,$D7,$FF,$03,$AE,$87,$58,$CD,$79,$87,$8B,$FC,$3F,$E2
	.dc.b	$7F,$89,$12,$35,$1F,$C4,$F7,$FE,$1D,$F5,$FF,$C7,$A7,$FB,$33,$5F
	.dc.b	$F3,$23,$5F,$F5,$1B,$67,$F8,$E7,$D1,$7F,$1F,$24,$F6,$F2,$3F,$02
	.dc.b	$FF,$AA,$BE,$6D,$EB,$7F,$1D,$FA,$F7,$CD,$53,$F9,$04,$EA,$BD,$81
	.dc.b	$C3,$FE,$75,$6B,$FD,$F5,$AF,$F9,$ED,$AF,$86,$BA,$91,$EB,$5F,$DD
	.dc.b	$78,$7F,$B6,$60,$FA,$3F,$E7,$5E,$EC,$BF,$64,$7F,$3D,$30,$FF,$45
	.dc.b	$8C,$3F,$07,$DD,$FE,$85,$8C,$8C,$10,$FE,$98,$E5,$FE,$9D,$1F,$D2
	.dc.b	$7E,$BF,$FA,$57,$D0,$C9,$F9,$91,$97,$E7,$46,$5F,$EA,$D1,$97,$FA
	.dc.b	$C4,$67,$2F,$6C,$FF,$6B,$F0,$3F,$45,$5F,$88,$FD,$A3,$FC,$4D,$5F
	.dc.b	$E0,$D6,$F4,$FF,$AD,$2B,$7E,$BF,$F8,$5F,$B3,$91,$30,$FC,$07,$A9
	.dc.b	$FD,$87,$CE,$FE,$C5,$E7,$49,$FC,$24,$65,$FC,$34,$65,$FE,$DD,$1F
	.dc.b	$D8,$7E,$47,$FD,$5D,$1F,$82,$F8,$5F,$A8,$DE,$3D,$7C,$2C,$8F,$F2
	.dc.b	$1D,$70,$7F,$D7,$4D,$7F,$C8,$4D,$7F,$D8,$0E,$F7,$FD,$D4,$7F,$B4
	.dc.b	$67,$A3,$FE,$D9,$BF,$F0,$1C,$FE,$1B,$E2,$7F,$C2,$DF,$0F,$F8,$9E
	.dc.b	$05,$FE,$CA,$30,$FE,$D3,$9F,$FD,$F2,$FE,$96,$B8,$F8,$F2,$30,$7E
	.dc.b	$83,$EC,$1E,$17,$F8,$8F,$EC,$3F,$A9,$78,$07,$FA,$A5,$2B,$6F,$1A
	.dc.b	$FF,$05,$46,$23,$FD,$E4,$DD,$3F,$C9,$55,$FB,$29,$FD,$7F,$D7,$BD
	.dc.b	$6F,$F1,$97,$D8,$7C,$0D,$FD,$B9,$A3,$17,$F0,$2F,$83,$CA,$FF,$C3
	.dc.b	$11,$88,$FF,$76,$35,$FF,$3A,$35,$FF,$79,$3B,$C7,$E9,$B7,$0C,$D7
	.dc.b	$93,$E4,$BF,$C8,$9F,$54,$FA,$CF,$F2,$47,$6F,$F9,$F7,$2F,$F5,$DB
	.dc.b	$72,$FE,$6D,$D8,$BE,$15,$FC,$A6,$31,$7D,$09,$1F,$CA,$39,$3F,$CA
	.dc.b	$2E,$EB,$FC,$D1,$B3,$1F,$F5,$74,$FF,$2E,$6A,$C7,$C3,$5F,$E6,$FC
	.dc.b	$7C,$5E,$67,$A8,$95,$DF,$E3,$FD,$BF,$61,$FE,$9E,$AF,$43,$6C,$D2
	.dc.b	$57,$95,$E2,$FF,$40,$F8,$46,$DF,$E6,$D6,$C6,$63,$06,$7D,$1B,$19
	.dc.b	$21,$FE,$E6,$9F,$F0,$E6,$7E,$CF,$B0,$F2,$12,$7F,$D6,$03,$2F,$A1
	.dc.b	$19,$8F,$89,$C5,$FA,$89,$CC,$7C,$D9,$F3,$7C,$5F,$FA,$23,$A7,$FA
	.dc.b	$94,$FF,$A2,$7E,$BB,$FC,$C9,$18,$23,$FC,$62,$34,$71,$71,$C4,$CD
	.dc.b	$F6,$19,$A8,$7F,$B8,$6B,$73,$A8,$C1,$A3,$FD,$DD,$1A,$9E,$DE,$EF
	.dc.b	$FA,$9E,$3F,$C0,$F3,$FC,$10,$F4,$46,$48,$F7,$78,$BF,$5F,$E1,$BF
	.dc.b	$D0,$5F,$1D,$FE,$61,$DF,$F8,$6F,$1F,$E6,$29,$2F,$A9,$ED,$39,$8C
	.dc.b	$4F,$0E,$9F,$14,$73,$8F,$CF,$9F,$C6,$1C,$7C,$5F,$1B,$F4,$36,$7F
	.dc.b	$8E,$BD,$7E,$0B,$F3,$91,$FE,$B4,$BE,$B1,$8C,$38,$4F,$63,$CF,$7E
	.dc.b	$6A,$DE,$3E,$59,$F1,$5F,$19,$F3,$D6,$E7,$CC,$D7,$E0,$CE,$DE,$35
	.dc.b	$E6,$AE,$31,$4E,$25,$38,$C5,$C7,$E2,$61,$E8,$25,$79,$AB,$D5,$3A
	.dc.b	$8E,$B1,$8C,$3C,$93,$98,$F1,$99,$3D,$C7,$C2,$2B,$B7,$C6,$7C,$B3
	.dc.b	$CE,$53,$81,$5F,$AF,$3E,$5A,$F3,$CF,$07,$EB,$5F,$6C,$3E,$BE,$3F
	.dc.b	$50,$97,$ED,$77,$7D,$71,$FC,$0C,$FB,$A7,$26,$B8,$3F,$91,$D7,$A1
	.dc.b	$B9,$7E,$35,$B8,$7E,$19,$DC,$FF,$66,$FB,$0F,$EC,$F1,$88,$46,$2F
	.dc.b	$F8,$4B,$63,$F5,$46,$84,$F1,$8F,$B9,$0F,$91,$FF,$3A,$57,$E9,$E3
	.dc.b	$C0,$7C,$B3,$5C,$DC,$1E,$91,$B9,$E3,$97,$8A,$78,$46,$1F,$1D,$FD
	.dc.b	$C7,$99,$F9,$F3,$BE,$BC,$F0,$D4,$76,$BF,$5D,$F3,$8F,$E0,$D7,$EA
	.dc.b	$8C,$47,$29,$DD,$72,$AF,$05,$7F,$A4,$6F,$DD,$7F,$15,$F0,$3E,$33
	.dc.b	$17,$C9,$C5,$72,$BE,$91,$E9,$BE,$95,$F0,$70,$E6,$76,$BF,$35,$FC
	.dc.b	$97,$95,$F4,$39,$47,$4C,$3C,$67,$18,$C9,$BA,$EB,$7D,$CF,$33,$B9
	.dc.b	$E3,$1F,$4D,$B3,$FD,$0F,$E9,$BF,$D6,$5B,$7F,$AC,$F3,$70,$A7,$39
	.dc.b	$7F,$86,$B8,$DE,$DE,$31,$E5,$3F,$D0,$6B,$FC,$87,$1F,$96,$BE,$53
	.dc.b	$FD,$73,$0E,$B5,$AA,$E2,$7E,$5A,$3C,$3E,$07,$56,$D7,$F7,$3F,$6F
	.dc.b	$18,$7F,$8B,$E3,$FA,$85,$60,$F2,$73,$FF,$57,$AC,$4E,$89,$87,$F7
	.dc.b	$38,$CB,$FD,$52,$BC,$98,$78,$23,$C9,$87,$76,$CF,$A7,$AB,$FC,$49
	.dc.b	$F3,$BF,$0D,$B8,$FF,$82,$7D,$BC,$70,$2A,$9C,$5C,$4F,$F8,$02,$B8
	.dc.b	$E1,$FB,$86,$BF,$D8,$F4,$FE,$E5,$F0,$3E,$EE,$03,$7C,$FC,$33,$0F
	.dc.b	$F0,$D3,$96,$9F,$E3,$9F,$B0,$FF,$0E,$B7,$C1,$E1,$8F,$F8,$38,$FF
	.dc.b	$09,$4E,$9C,$64,$8C,$47,$F7,$78,$C4,$FC,$59,$CC,$FF,$85,$DA,$F4
	.dc.b	$EF,$F9,$10,$7E,$CF,$8C,$F8,$DC,$C4,$C7,$FC,$8D,$F0,$37,$0F,$F2
	.dc.b	$BB,$CB,$9E,$4F,$F9,$47,$EC,$2D,$7A,$32,$FC,$CC,$FF,$93,$FE,$BB
	.dc.b	$3D,$5F,$EB,$97,$F3,$94,$FA,$37,$B9,$86,$DF,$FA,$62,$CF,$F9,$FB
	.dc.b	$EC,$3E,$DD,$FC,$93,$95,$F1,$3F,$40,$E2,$7F,$C7,$ED,$3F,$C8,$37
	.dc.b	$D9,$B3,$F9,$CA,$7C,$6F,$3B,$F1,$1E,$17,$D3,$F6,$0F,$2E,$9F,$A4
	.dc.b	$E5,$69,$8B,$E3,$97,$FA,$CA,$F8,$46,$3A,$3F,$1B,$F2,$DB,$AF,$C2
	.dc.b	$46,$27,$FC,$FD,$F0,$BF,$63,$D9,$FB,$2D,$8C,$9F,$E8,$38,$CB,$4F
	.dc.b	$0D,$7C,$AF,$8D,$FA,$30,$7F,$87,$2F,$B0,$FF,$5E,$BC,$0D,$16,$3E
	.dc.b	$6B,$89,$FF,$35,$B5,$FE,$C6,$D7,$FC,$9C,$EF,$8B,$ED,$C7,$DC,$F1
	.dc.b	$5E,$E7,$92,$FC,$71,$8B,$FF,$4F,$79,$BF,$98,$F3,$3F,$32,$D8,$C9
	.dc.b	$FE,$A3,$8C,$B4,$FD,$97,$C0,$7E,$88,$FD,$0D,$D8,$C9,$FE,$A5,$F4
	.dc.b	$31,$97,$FA,$D9,$FD,$3F,$C4,$7E,$A3,$EC,$2D,$7F,$AC,$5B,$5F,$63
	.dc.b	$C1,$46,$36,$F7,$F1,$27,$37,$FE,$B9,$C3,$A3,$4E,$87,$FA,$F6,$3F
	.dc.b	$6F,$F6,$1F,$B8,$6C,$64,$FF,$61,$46,$5A,$7F,$B1,$7E,$C3,$F7,$FF
	.dc.b	$4F,$FC,$01,$B1,$93,$FD,$95,$19,$69,$F8,$8F,$B0,$7D,$6F,$EE,$1B
	.dc.b	$1F,$9C,$3F,$11,$C7,$99,$91,$FB,$09,$DE,$DF,$EB,$06,$BF,$E2,$86
	.dc.b	$BF,$E9,$37,$7A,$FE,$1A,$7F,$29,$EE,$74,$AF,$E5,$11,$83,$FD,$C5
	.dc.b	$F2,$3F,$C7,$0F,$E3,$BF,$13,$63,$F4,$3E,$B3,$DA,$FE,$0A,$D8,$FE
	.dc.b	$0A,$DB,$FF,$62,$B5,$FE,$89,$C0,$F9,$2F,$E5,$27,$F2,$AF,$B0,$B7
	.dc.b	$FB,$EF,$6D,$F5,$EF,$B0,$FE,$3F,$93,$E2,$7F,$BE,$56,$A7,$31,$2F
	.dc.b	$C6,$F9,$88,$FF,$B6,$27,$AF,$17,$2C,$3E,$3C,$FF,$A9,$2B,$13,$FC
	.dc.b	$BB,$EB,$FF,$21,$3F,$C0,$9B,$AF,$F5,$1F,$31,$F6,$1F,$CE,$23,$F9
	.dc.b	$64,$EB,$93,$F6,$F0,$8E,$2F,$F9,$81,$AF,$FB,$2D,$A8,$E1,$C1,$FC
	.dc.b	$D2,$3A,$61,$FE,$ED,$FB,$07,$A1,$E7,$AF,$EF,$D1,$89,$FE,$33,$F6
	.dc.b	$1F,$D0,$31,$F1,$B6,$3F,$DE,$DF,$1E,$F8,$8F,$E9,$B3,$9C,$FC,$4F
	.dc.b	$F4,$87,$FC,$EE,$D7,$FD,$C4,$D7,$7E,$EF,$F4,$A8,$E5,$83,$ED,$9F
	.dc.b	$23,$DA,$BF,$C6,$D3,$98,$FE,$BB,$ED,$FF,$19,$C3,$FA,$B3,$67,$E5
	.dc.b	$FD,$8B,$C1,$DF,$99,$36,$FE,$AF,$7C,$70,$FF,$46,$B5,$FF,$7C,$35
	.dc.b	$F9,$CF,$58,$FA,$1F,$44,$FD,$5B,$B4,$7F,$42,$9D,$52,$B5,$87,$F7
	.dc.b	$4F,$B2,$FB,$17,$FD,$21,$B5,$FD,$BF,$2E,$2A,$33,$9F,$E1,$9D,$C7
	.dc.b	$FA,$91,$AF,$E0,$CE,$07,$F7,$67,$2F,$E1,$DE,$75,$EA,$CB,$BC,$62
	.dc.b	$7D,$11,$8F,$B1,$F4,$1F,$E0,$CF,$B2,$FD,$75,$FA,$63,$2F,$F8,$0A
	.dc.b	$D1,$E1,$24,$65,$FF,$38,$66,$FC,$85,$E0,$72,$CB,$FC,$71,$39,$8F
	.dc.b	$91,$EB,$4C,$3F,$B7,$FD,$9F,$E4,$DF,$95,$19,$BF,$F6,$39,$F2,$5F
	.dc.b	$84,$48,$CF,$9D,$FD,$1F,$8B,$FD,$05,$F5,$8F,$B3,$7B,$1F,$E3,$A9
	.dc.b	$CC,$62,$FF,$C1,$9D,$BF,$7A,$F8,$5E,$E7,$FC,$27,$18,$7F,$C2,$BF
	.dc.b	$0C,$FB,$37,$9D,$FE,$6F,$6C,$FF,$85,$FE,$25,$F4,$59,$A9,$3C,$AB
	.dc.b	$7B,$FF,$86,$AF,$73,$9A,$FF,$48,$7D,$93,$EC,$E9,$FE,$6F,$FA,$3B
	.dc.b	$5F,$E2,$8F,$92,$56,$E3,$0D,$C7,$B5,$70,$7F,$CD,$95,$C2,$FF,$D1
	.dc.b	$F5,$B9,$FC,$94,$E6,$FF,$D5,$13,$99,$FF,$80,$2B,$DB,$F3,$9F,$E8
	.dc.b	$3F,$A4,$B5,$E2,$8B,$5E,$F9,$BD,$DF,$CA,$3E,$CF,$FC,$8D,$E8,$FF
	.dc.b	$94,$9B,$5F,$82,$6B,$FA,$50,$E1,$FD,$47,$CC,$7F,$8D,$FB,$3F,$CB
	.dc.b	$6D,$8F,$C5,$B5,$E4,$D7,$F6,$13,$FE,$81,$D3,$C5,$F1,$E6,$92,$7F
	.dc.b	$D8,$DF,$49,$F8,$27,$F0,$1F,$4B,$FB,$9C,$FF,$C3,$7C,$68,$C6,$FF
	.dc.b	$CF,$F0,$61,$C7,$0F,$A6,$DC,$38,$DF,$09,$76,$31,$E9,$1F,$81,$BE
	.dc.b	$18,$F5,$5D,$BC,$6F,$87,$C7,$5C,$3F,$D1,$78,$71,$AE,$D9,$F9,$1D
	.dc.b	$54,$9D,$10,$FC,$C3,$97,$D3,$1E,$7E,$9F,$E0,$7E,$D7,$7F,$84,$95
	.dc.b	$FA,$AF,$0B,$FC,$1C,$D9,$CB,$FA,$39,$E5,$3A,$8F,$F6,$17,$6B,$EC
	.dc.b	$55,$F1,$63,$0D,$8C,$CF,$FA,$27,$E9,$79,$71,$7F,$9B,$2F,$73,$FB
	.dc.b	$EA,$DB,$D9,$B3,$FE,$54,$AE,$3C,$3F,$CD,$7E,$53,$FA,$ED,$FF,$A6
	.dc.b	$39,$DF,$60,$EA,$FB,$11,$87,$ED,$4F,$FB,$2F,$CF,$B7,$CA,$3F,$93
	.dc.b	$DF,$17,$A3,$5F,$F5,$D4,$6F,$2F,$F0,$7A,$FA,$7E,$F5,$9C,$BF,$AE
	.dc.b	$B6,$FF,$C1,$D5,$B7,$FD,$3E,$D7,$C3,$87,$23,$CB,$72,$3C,$71,$97
	.dc.b	$8C,$7B,$3E,$B5,$F1,$7F,$03,$EB,$FF,$D8,$B5,$F1,$5A,$FF,$B4,$E7
	.dc.b	$6F,$FA,$C1,$AF,$F4,$88,$DE,$7F,$EA,$2D,$51,$8D,$3F,$B3,$5F,$0A
	.dc.b	$FF,$18,$B6,$3F,$BE,$CE,$EF,$FB,$D3,$54,$ED,$FF,$19,$3F,$EB,$F7
	.dc.b	$EE,$7D,$7F,$FB,$65,$BF,$29,$EF,$F8,$E9,$18,$21,$ED,$72,$FF,$62
	.dc.b	$9E,$4D,$9C,$E3,$FE,$48,$FA,$E7,$CA,$FE,$1B,$EB,$FF,$C6,$D7,$FE
	.dc.b	$D4,$6B,$FE,$FB,$9D,$BF,$ED,$86,$BF,$EF,$F8,$DD,$7F,$95,$1B,$A7
	.dc.b	$F8,$36,$B7,$3E,$1B,$FF,$80,$36,$9E,$6F,$F2,$F6,$BC,$63,$78,$3E
	.dc.b	$2B,$F3,$BF,$43,$FE,$81,$AF,$F7,$5B,$56,$1F,$E6,$6A,$E1,$1F,$EF
	.dc.b	$06,$BA,$C7,$FC,$45,$5C,$6B,$FC,$DA,$D8,$FF,$13,$57,$09,$E9,$7F
	.dc.b	$AE,$8C,$4E,$A3,$CB,$F1,$9F,$B2,$FA,$CF,$EB,$18,$FF,$9C,$70,$FF
	.dc.b	$68,$FD,$43,$E9,$3F,$94,$36,$75,$8F,$F8,$D6,$F8,$C7,$9C,$FF,$3D
	.dc.b	$7D,$7F,$F3,$27,$FC,$F9,$1A,$C3,$FD,$0A,$DA,$FF,$1F,$56,$E8,$D2
	.dc.b	$7A,$3E,$7F,$A0,$FE,$75,$DA,$FA,$0F,$2E,$33,$0F,$E3,$D1,$FC,$3B
	.dc.b	$EB,$DF,$4D,$C3,$8A,$57,$A8,$FF,$48,$4E,$9F,$E4,$DF,$4B,$C6,$A4
	.dc.b	$60,$E0,$3E,$7B,$F8,$A6,$FF,$EA,$0F,$AF,$7C,$E7,$F1,$E6,$D7,$0A
	.dc.b	$DF,$0F,$F9,$75,$B1,$FE,$A1,$9D,$BF,$2C,$38,$BF,$F5,$6C,$F0,$9F
	.dc.b	$17,$89,$DF,$09,$1D,$E3,$E3,$FD,$2F,$C8,$F7,$24,$FF,$43,$8C,$BF
	.dc.b	$D1,$63,$2F,$DB,$F4,$5C,$4A,$FA,$3E,$DF,$D2,$C2,$5F,$F4,$D8,$CB
	.dc.b	$FD,$3E,$32,$F3,$A4,$FD,$AF,$53,$ED,$FA,$32,$BF,$AB,$46,$5F,$EB
	.dc.b	$11,$97,$94,$65,$E9,$19,$D7,$FA,$B4,$FF,$71,$F5,$3F,$43,$5F,$E0
	.dc.b	$8E,$AF,$F1,$45,$7E,$7A,$35,$DB,$1F,$C2,$6A,$AD,$F1,$7F,$B5,$63
	.dc.b	$13,$B3,$D9,$E4,$FE,$53,$C9,$91,$FD,$BE,$3F,$2E,$61,$F8,$51,$87
	.dc.b	$F1,$7F,$1C,$F8,$1F,$EC,$BE,$FF,$CF,$9F,$C7,$B5,$FC,$24,$ED,$E3
	.dc.b	$3C,$38,$E4,$60,$87,$E2,$EB,$96,$BF,$EF,$66,$E7,$FE,$AF,$C3,$A5
	.dc.b	$7D,$7F,$95,$FF,$14,$7C,$0F,$F9,$06,$3F,$D0,$FE,$C7,$EC,$A6,$46
	.dc.b	$08,$71,$3F,$4F,$F5,$EF,$07,$FC,$65,$F5,$FF,$D3,$F8,$4F,$F5,$1A
	.dc.b	$56,$DE,$35,$C3,$D1,$91,$83,$F6,$7F,$5C,$FA,$AF,$F1,$F7,$D7,$FE
	.dc.b	$EE,$7F,$34,$DE,$17,$07,$FB,$36,$F8,$57,$2B,$E3,$E1,$48,$C1,$0E
	.dc.b	$6E,$5F,$DB,$4F,$27,$F9,$17,$B8,$FA,$BF,$E5,$11,$FA,$BE,$97,$C1
	.dc.b	$FF,$2C,$3F,$C8,$D3,$AC,$FC,$B4,$BE,$5E,$6E,$E3,$F2,$B8,$72,$7F
	.dc.b	$D6,$77,$C7,$44,$71,$8F,$E6,$5C,$5F,$9B,$F8,$6F,$E7,$67,$83,$CD
	.dc.b	$95,$D2,$32,$F4,$8F,$2F,$D7,$FF,$99,$FB,$A4,$F8,$23,$2F,$FA,$FE
	.dc.b	$32,$FC,$2C,$13,$FE,$6F,$F9,$2D,$CE,$63,$0F,$FB,$1E,$33,$B7,$C2
	.dc.b	$DB,$E5,$56,$AB,$A7,$D7,$FF,$97,$F8,$7F,$B1,$43,$DF,$46,$2F,$E9
	.dc.b	$57,$CE,$38,$7D,$0C,$BF,$93,$1F,$E8,$0E,$4F,$DF,$C7,$FA,$3B,$E1
	.dc.b	$7F,$C8,$99,$FF,$A6,$BC,$3D,$4C,$3F,$DB,$CF,$FA,$2C,$C3,$FE,$CC
	.dc.b	$AE,$7C,$5F,$E9,$79,$D4,$FF,$B5,$E7,$BF,$77,$E5,$E3,$FD,$C5,$3A
	.dc.b	$79,$1E,$5F,$89,$FF,$4C,$77,$C8,$E5,$19,$7F,$DE,$11,$97,$C5,$1F
	.dc.b	$E7,$0F,$89,$FF,$2A,$E5,$F8,$6E,$3F,$CA,$C3,$E9,$FF,$13,$7F,$90
	.dc.b	$3F,$17,$19,$78,$75,$7F,$B3,$BB,$BF,$C3,$71,$B1,$3D,$9C,$3C,$5B
	.dc.b	$5D,$AF,$C3,$5C,$9B,$87,$6A,$F0,$D6,$F0,$FF,$6E,$B6,$3C,$B7,$C6
	.dc.b	$7C,$95,$C6,$7C,$75,$B6,$99,$C3,$CD,$87,$82,$3C,$A6,$67,$84,$EA
	.dc.b	$FA,$CE,$89,$5C,$49,$1E,$68,$E4,$77,$C3,$CC,$FA,$30,$CB,$CA,$7A
	.dc.b	$D7,$28,$F8,$26,$9C,$3F,$3A,$B9,$BE,$C3,$91,$D4,$FA,$2E,$AF,$E9
	.dc.b	$4F,$E0,$30,$F6,$CE,$4D,$3F,$00,$CE,$1E,$93,$FD,$F1,$F0,$DF,$0B
	.dc.b	$53,$49,$8F,$E2,$87,$D7,$70,$70,$D7,$F1,$ED,$7F,$8C,$B5,$E2,$7E
	.dc.b	$34,$FE,$53,$D9,$F7,$5D,$99,$EE,$FE,$53,$9F,$F3,$08,$CB,$63,$3A
	.dc.b	$F9,$67,$AC,$E6,$BA,$46,$27,$D5,$5C,$23,$D2,$77,$9D,$DF,$8E,$75
	.dc.b	$3E,$C3,$47,$A4,$E8,$7E,$B9,$DB,$E6,$AE,$F5,$D2,$38,$DF,$AD,$E2
	.dc.b	$7F,$3E,$AE,$2E,$0F,$8A,$77,$BF,$25,$73,$AF,$74,$E4,$EB,$7F,$0E
	.dc.b	$B7,$19,$9D,$9E,$18,$CB,$63,$89,$F1,$1D,$CF,$38,$C9,$F0,$E3,$2E
	.dc.b	$E3,$DE,$FF,$0F,$C7,$FC,$EF,$F4,$BF,$D8,$B3,$FE,$C7,$C7,$FE,$AD
	.dc.b	$9C,$65,$FD,$79,$DB,$FE,$38,$F2,$91,$AC,$2F,$23,$1C,$61,$CF,$0F
	.dc.b	$E6,$51,$8C,$38,$C7,$F3,$5F,$07,$1B,$F3,$CB,$FD,$17,$18,$9D,$1E
	.dc.b	$19,$D6,$3F,$E1,$38,$C1,$AC,$BF,$CD,$98,$7E,$A2,$31,$9F,$BA,$31
	.dc.b	$3F,$D2,$70,$EC,$DA,$E6,$FF,$6D,$F9,$8F,$77,$67,$FA,$DF,$E6,$FE
	.dc.b	$07,$67,$F8,$46,$73,$C0,$9C,$F0,$A7,$31,$86,$E3,$FE,$62,$EA,$FF
	.dc.b	$79,$BF,$EA,$7C,$1F,$5C,$F0,$3E,$5D,$6B,$4F,$F1,$0C,$E6,$F9,$7C
	.dc.b	$07,$92,$31,$3F,$E8,$DA,$E1,$C1,$FD,$FE,$72,$DC,$3E,$A6,$7C,$5B
	.dc.b	$8F,$8E,$B6,$FF,$9D,$78,$3D,$CE,$67,$78,$F9,$3E,$3B,$F0,$3E,$33
	.dc.b	$DE,$F2,$46,$1B,$7F,$E2,$E6,$D7,$F8,$E5,$F5,$7B,$DC,$BE,$B2,$D7
	.dc.b	$F8,$D1,$B5,$FE,$3F,$6E,$BF,$EE,$C7,$F2,$90,$FA,$27,$AD,$F9,$4F
	.dc.b	$AE,$7D,$76,$A7,$38,$FF,$A5,$27,$2F,$FB,$4A,$33,$7F,$EB,$79,$CB
	.dc.b	$FE,$C9,$8C,$E3,$FA,$A9,$CB,$FE,$4E,$9F,$16,$1E,$1C,$3C,$0D,$C3
	.dc.b	$FD,$BF,$F5,$CF,$86,$FD,$4F,$2B,$E4,$FD,$73,$9A,$FF,$54,$CE,$5E
	.dc.b	$35,$C2,$B6,$D8,$FF,$57,$FA,$BF,$EE,$B6,$FA,$FF,$B6,$8C,$67,$FE
	.dc.b	$D3,$51,$9F,$81,$D3,$6B,$89,$FE,$F6,$1F,$17,$FC,$0D,$FE,$05,$ED
	.dc.b	$FD,$27,$27,$F9,$A3,$7F,$F3,$8B,$A3,$94,$F1,$F1,$7F,$D6,$D1,$83
	.dc.b	$81,$FE,$70,$38,$65,$FE,$F9,$7A,$AA,$EB,$3C,$A3,$55,$D6,$FE,$07
	.dc.b	$93,$F0,$7C,$AB,$8F,$C6,$6D,$FC,$36,$E3,$FC,$9B,$44,$7E,$BF,$E4
	.dc.b	$BF,$3F,$E6,$DC,$3F,$D1,$2D,$C3,$FD,$28,$DC,$3F,$97,$F5,$A3,$E9
	.dc.b	$7C,$1F,$D3,$FA,$DB,$87,$D8,$6E,$1F,$55,$B8,$78,$DB,$3F,$EE,$9C
	.dc.b	$D1,$8A,$79,$89,$CD,$7E,$11,$FC,$17,$D6,$FE,$0D,$B1,$93,$F1,$31
	.dc.b	$96,$9F,$8C,$FA,$DF,$C5,$7D,$17,$E2,$DB,$19,$3F,$23,$19,$69,$F7
	.dc.b	$30,$FF,$55,$FD,$6B,$EA,$BF,$15,$F5,$CF,$77,$F7,$73,$9B,$E0,$27
	.dc.b	$4F,$3B,$4E,$DF,$C2,$47,$E2,$67,$71,$83,$C3,$7E,$0B,$EE,$DA,$E9
	.dc.b	$7C,$F6,$FC,$CE,$1C,$9F,$C2,$7C,$87,$F1,$0F,$4D,$F3,$5F,$AD,$8C
	.dc.b	$33,$0F,$D9,$FD,$75,$D3,$F6,$0F,$EC,$FD,$DF,$E5,$3D,$96,$BF,$DA
	.dc.b	$6F,$5F,$6E,$47,$C9,$F6,$9F,$5A,$FA,$EC,$E5,$ED,$3F,$C9,$37,$FD
	.dc.b	$DE,$B7,$C3,$AD,$FF,$DC,$71,$8C,$7A,$63,$EF,$7F,$DB,$FC,$EE,$BF
	.dc.b	$81,$1F,$E2,$59,$D6,$5F,$8E,$9C,$BF,$EE,$6F,$AD,$FC,$85,$7D,$56
	.dc.b	$C7,$D1,$F2,$DF,$5C,$F3,$BF,$9C,$D6,$B4,$FE,$54,$DC,$3F,$DE,$3E
	.dc.b	$B3,$EB,$5F,$5D,$C3,$BF,$FB,$AC,$62,$BC,$11,$89,$D4,$7F,$6F,$9D
	.dc.b	$69,$F1,$FE,$BB,$FB,$C4,$FF,$96,$61,$18,$7F,$CB,$8D,$7F,$CE,$4D
	.dc.b	$7F,$D7,$AE,$F2,$FF,$17,$7D,$73,$E2,$3F,$AA,$CE,$6F,$F9,$24,$E5
	.dc.b	$FC,$D4,$66,$FF,$8E,$4E,$5F,$37,$B9,$FE,$31,$C1,$39,$7F,$A8,$7B
	.dc.b	$1F,$C8,$56,$0B,$7D,$03,$EB,$5E,$27,$F6,$8C,$38,$5E,$EB,$53,$9F
	.dc.b	$37,$FC,$81,$18,$9F,$D8,$57,$0C,$3A,$E1,$D2,$73,$7C,$23,$13,$F9
	.dc.b	$5A,$E1,$7C,$E3,$F3,$1F,$59,$FC,$FD,$FD,$B4,$79,$EF,$FB,$EC,$E7
	.dc.b	$2F,$6D,$78,$27,$2F,$76,$E3,$FC,$3E,$F9,$3F,$C1,$EF,$93,$EF,$C7
	.dc.b	$D5,$69,$CB,$D6,$78,$7D,$67,$C2,$FA,$CF,$F2,$5E,$5F,$DF,$ED,$39
	.dc.b	$F1,$7C,$ED,$BE,$27,$8E,$F8,$BF,$57,$0F,$04,$ED,$F8,$6D,$C3,$B7
	.dc.b	$AB,$E1,$E4,$F8,$6D,$BF,$63,$6F,$E5,$B6,$BE,$3F,$C2,$3E,$AF,$FB
	.dc.b	$FD,$FF,$33,$8C,$7A,$BF,$E0,$5E,$0F,$F3,$0D,$6B,$1F,$E5,$15,$AB
	.dc.b	$E1,$39,$C3,$FB,$35,$F0,$C3,$FC,$CF,$F5,$7F,$C1,$3C,$17,$C0,$BE
	.dc.b	$9F,$3E,$7F,$FA,$15,$B9,$FF,$A3,$DB,$AF,$F9,$F1,$FF,$43,$7D,$5B
	.dc.b	$E4,$BF,$D1,$9C,$38,$6D,$FF,$4D,$F8,$C5,$6E,$3F,$D3,$55,$B6,$9B
	.dc.b	$8D,$61,$FE,$8A,$9D,$36,$75,$EB,$FF,$A6,$FD,$CF,$1B,$63,$27,$FA
	.dc.b	$9E,$32,$D3,$FA,$B7,$D5,$FF,$8F,$FC,$BF,$E1,$CD,$BF,$F7,$43,$5F
	.dc.b	$F7,$F3,$5F,$C3,$7C,$7F,$F8,$C3,$B7,$F8,$B3,$63,$FD,$E0,$D7,$F1
	.dc.b	$81,$C3,$F8,$F9,$7C,$3F,$FA,$F2,$3C,$2D,$7C,$2D,$78,$35,$FF,$61
	.dc.b	$FC,$3B,$D0,$FC,$D7,$47,$ED,$38,$3F,$D0,$51,$8D,$3F,$D9,$95,$BB
	.dc.b	$FF,$66,$73,$E5,$E7,$56,$F8,$C7,$DB,$AE,$0F,$FA,$FA,$B5,$F3,$5F
	.dc.b	$E0,$78,$FC,$03,$5F,$EF,$D4,$8C,$63,$FE,$E0,$FB,$27,$D9,$F2,$E7
	.dc.b	$98,$72,$7C,$9D,$2A,$DC,$E4,$87,$6C,$3A,$EB,$FD,$A2,$77,$3F,$E0
	.dc.b	$FF,$33,$F8,$67,$A9,$EF,$6C,$7F,$B0,$7C,$4F,$EF,$4F,$FB,$BF,$E0
	.dc.b	$B5,$18,$8E,$4D,$D7,$F9,$0D,$72,$9F,$F7,$0D,$7A,$3D,$3F,$E2,$71
	.dc.b	$FB,$DF,$1F,$F0,$77,$FC,$09,$B9,$FF,$7E,$AD,$D7,$F8,$97,$5F,$F0
	.dc.b	$07,$8B,$F5,$7C,$4F,$E9,$13,$70,$E2,$FF,$80,$67,$ED,$3F,$BD,$F8
	.dc.b	$1B,$E0,$9A,$7D,$FE,$B7,$D9,$BE,$4B,$37,$FE,$3B,$BE,$0D,$BE,$15
	.dc.b	$B8,$D7,$A1,$D8,$8F,$F3,$F8,$FC,$EB,$A2,$1F,$E2,$5A,$FE,$E8,$EF
	.dc.b	$B3,$8B,$9A,$BC,$1F,$D3,$B2,$F3,$D7,$BB,$EA,$DC,$54,$98,$7F,$4F
	.dc.b	$E7,$F8,$AA,$76,$FF,$95,$DA,$FF,$8D,$DB,$87,$79,$E3,$18,$E3,$FE
	.dc.b	$8C,$DA,$FE,$F8,$DC,$7F,$3D,$1E,$5F,$85,$8F,$77,$E8,$5F,$FA,$46
	.dc.b	$30,$43,$C5,$1A,$F0,$F8,$C6,$0D,$63,$E5,$8F,$BF,$F8,$23,$ED,$29
	.dc.b	$2B,$FC,$E9,$D8,$F0,$31,$1F,$E5,$5A,$E0,$FF,$9E,$5A,$FF,$96,$67
	.dc.b	$83,$9A,$E3,$87,$FB,$B6,$B6,$FF,$BB,$EB,$75,$FE,$61,$FA,$B7,$C5
	.dc.b	$FF,$8E,$E1,$18,$7F,$C7,$CD,$7F,$93,$9A,$8E,$11,$9A,$FF,$7E,$37
	.dc.b	$7F,$EF,$0D,$AF,$F1,$77,$D6,$7F,$29,$AD,$DF,$E9,$2B,$55,$D4,$FF
	.dc.b	$4C,$56,$EF,$FC,$05,$F9,$5B,$E6,$BF,$56,$DA,$FF,$4F,$35,$FF,$3F
	.dc.b	$72,$7F,$6B,$7F,$D4,$4D,$73,$CD,$FB,$48,$C6,$7F,$E2,$89,$57,$EB
	.dc.b	$9F,$F4,$43,$63,$53,$C3,$0F,$F0,$B3,$6B,$FC,$99,$5B,$9F,$E1,$5F
	.dc.b	$31,$FC,$44,$93,$FE,$B5,$F9,$FF,$C8,$3F,$E9,$5F,$AD,$FE,$4E,$FF
	.dc.b	$95,$5A,$9B,$1F,$73,$1F,$43,$6F,$FC,$50,$DA,$FF,$2F,$37,$C0,$FF
	.dc.b	$4C,$7A,$BF,$C5,$3C,$EF,$81,$87,$C1,$6E,$1F,$E3,$16,$E1,$FE,$CA
	.dc.b	$F2,$7F,$A7,$D6,$EB,$91,$27,$FD,$B5,$F1,$6A,$DC,$F3,$35,$3F,$EB
	.dc.b	$3B,$EB,$5B,$F3,$3F,$9B,$7C,$FF,$F3,$7F,$9C,$97,$FE,$E5,$D3,$FD
	.dc.b	$61,$3A,$8C,$79,$0A,$DC,$7F,$37,$84,$FF,$41,$F2,$FF,$A1,$7C,$3C
	.dc.b	$AE,$31,$97,$9C,$65,$E5,$1F,$D2,$7E,$DF,$FA,$57,$CF,$C9,$F0,$C6
	.dc.b	$5F,$1C,$65,$F1,$46,$5F,$24,$67,$83,$FC,$EF,$9F,$FB,$2E,$BF,$B1
	.dc.b	$7E,$8D,$FC,$CA,$FF,$B3,$36,$BF,$AE,$B7,$83,$FA,$B7,$7F,$0F,$1E
	.dc.b	$5F,$FA,$76,$F7,$1F,$D8,$67,$FB,$17,$C1,$7F,$63,$F9,$99,$7F,$2A
	.dc.b	$32,$FC,$C8,$CB,$F2,$E3,$F1,$1F,$55,$23,$B6,$3F,$87,$8F,$EB,$FF
	.dc.b	$59,$FB,$99,$FF,$57,$73,$27,$84,$7E,$05,$DF,$3E,$34,$FC,$94,$5C
	.dc.b	$3F,$D0,$58,$73,$DD,$F5,$6F,$AB,$FF,$09,$BD,$72,$FC,$33,$5F,$EF
	.dc.b	$0E,$DC,$C7,$0C,$24,$60,$F5,$8F,$53,$FC,$4D,$EF,$FF,$85,$7C,$CF
	.dc.b	$F3,$BC,$EA,$72,$F8,$3D,$1D,$54,$8D,$57,$07,$C7,$3F,$E0,$4F,$A9
	.dc.b	$7C,$15,$24,$F8,$87,$B7,$FD,$F5,$70,$7F,$17,$AE,$31,$B6,$BA,$E0
	.dc.b	$FF,$55,$46,$27,$F8,$F5,$73,$79,$5F,$82,$F8,$1F,$EE,$5A,$E1,$7F
	.dc.b	$EB,$3B,$E0,$FF,$AD,$7E,$A1,$E1,$FF,$87,$9B,$7E,$96,$BF,$E2,$9F
	.dc.b	$A9,$7C,$E6,$EA,$4E,$DF,$F2,$6C,$EA,$3E,$77,$D3,$3C,$8F,$F2,$33
	.dc.b	$63,$31,$FE,$54,$F8,$F7,$C5,$6A,$24,$E8,$FE,$15,$3A,$7E,$D7,$E3
	.dc.b	$EF,$A6,$FF,$29,$36,$33,$1F,$E6,$0E,$47,$B1,$A8,$93,$A3,$F8,$B4
	.dc.b	$E9,$B3,$C4,$8E,$BE,$82,$BF,$E4,$D9,$E0,$D9,$F0,$4F,$93,$DC,$93
	.dc.b	$E8,$7C,$9B,$E2,$3F,$CD,$98,$FF,$88,$FA,$0F,$F8,$97,$B3,$FD,$9F
	.dc.b	$5F,$E8,$BF,$C8,$FE,$DD,$FF,$37,$6F,$79,$CD,$7F,$B5,$BE,$87,$87
	.dc.b	$C0,$4B,$FF,$66,$56,$B6,$56,$67,$FD,$1F,$3D,$0D,$46,$7A,$FF,$D4
	.dc.b	$91,$9B,$F9,$D3,$E5,$F9,$D9,$7D,$9F,$A5,$C5,$88,$EB,$17,$0F,$E5
	.dc.b	$78,$AB,$FC,$E3,$F1,$3F,$62,$56,$E6,$FC,$53,$FE,$C3,$E1,$CB,$9E
	.dc.b	$5F,$F6,$6C,$62,$77,$7C,$EB,$47,$47,$11,$A9,$E1,$5C,$AB,$C3,$3C
	.dc.b	$B9,$3F,$77,$1C,$8F,$F6,$A3,$88,$E5,$1C,$27,$93,$BB,$EF,$1F,$EE
	.dc.b	$2A,$E0,$EA,$B8,$D6,$E7,$2F,$9D,$AF,$C1,$3B,$1E,$BB,$F4,$BB,$C3
	.dc.b	$2F,$99,$B8,$7F,$9E,$2F,$B9,$ED,$8C,$DF,$69,$D6,$1E,$53,$D1,$3A
	.dc.b	$39,$E1,$E0,$C3,$C8,$7A,$63,$4F,$C1,$8D,$57,$48,$F3,$1B,$34,$74
	.dc.b	$38,$E1,$E5,$72,$47,$04,$BF,$34,$7C,$33,$8B,$E8,$7F,$C3,$DC,$D7
	.dc.b	$87,$F8,$64,$7C,$53,$27,$91,$CB,$F9,$2F,$A9,$FC,$9C,$62,$11,$89
	.dc.b	$FF,$38,$B6,$3F,$2A,$68,$EA,$7E,$61,$FA,$AB,$E2,$DD,$98,$8E,$18
	.dc.b	$FE,$81,$D4,$66,$B8,$D7,$09,$FC,$A1,$E7,$2B,$9A,$E5,$1D,$31,$F2
	.dc.b	$CE,$5F,$4C,$70,$BE,$CD,$3A,$9C,$4F,$D8,$CF,$17,$DF,$7E,$49,$E3
	.dc.b	$7E,$9A,$D5,$7C,$13,$A4,$79,$2B,$84,$76,$6C,$66,$37,$7E,$68,$E9
	.dc.b	$7D,$A3,$D6,$D9,$F2,$B5,$FE,$17,$7E,$27,$E3,$57,$27,$C0,$76,$3D
	.dc.b	$65,$3D,$0E,$CF,$21,$6B,$C8,$78,$A3,$F8,$7E,$3F,$CD,$7E,$BF,$EB
	.dc.b	$6B,$FC,$A9,$B8,$FF,$1E,$E5,$70,$BE,$8F,$38,$FF,$3F,$BB,$8F,$3C
	.dc.b	$3F,$9D,$C6,$B0,$E3,$7A,$E6,$4D,$F1,$25,$FF,$60,$FA,$3B,$97,$F4
	.dc.b	$09,$BA,$7F,$34,$9B,$5F,$A0,$C3,$FC,$88,$FF,$4A,$AF,$06,$8A,$DE
	.dc.b	$1F,$D5,$66,$E7,$EC,$9C,$F5,$FF,$5E,$EC,$F6,$B7,$1F,$EE,$7F,$2D
	.dc.b	$8E,$D5,$52,$4E,$8F,$F0,$1B,$E2,$FA,$29,$1F,$DF,$34,$BA,$7F,$77
	.dc.b	$9B,$B7,$F5,$09,$B5,$FD,$3B,$2F,$EE,$39,$F8,$4C,$E1,$FE,$4D,$CB
	.dc.b	$DD,$5F,$E1,$29,$D3,$FE,$0A,$C6,$D7,$F8,$D2,$DF,$52,$F9,$DF,$7B
	.dc.b	$6F,$FC,$3B,$F2,$78,$99,$1F,$E5,$79,$D5,$FF,$8F,$B7,$FE,$CD,$7A
	.dc.b	$8F,$F2,$08,$FA,$67,$D3,$7E,$22,$30,$FE,$17,$E1,$DF,$53,$9A,$DE
	.dc.b	$5F,$E5,$96,$DF,$46,$F0,$FF,$A8,$2F,$8D,$6F,$0E,$CF,$D6,$3F,$CC
	.dc.b	$45,$BF,$F2,$AA,$FA,$7F,$F2,$8B,$AE,$3E,$24,$7F,$CF,$66,$FC,$9C
	.dc.b	$57,$BA,$3A,$63,$FF,$00,$5A,$7F,$81,$8E,$07,$53,$E5,$79,$DF,$40
	.dc.b	$FD,$37,$89,$C2,$BF,$D1,$1C,$09,$D5,$70,$3F,$CA,$38,$E2,$BE,$95
	.dc.b	$FF,$85,$E7,$FC,$AB,$D7,$D2,$3A,$B6,$32,$7F,$96,$A3,$2F,$FA,$1F
	.dc.b	$E0,$5E,$EF,$D7,$6C,$7C,$2D,$5F,$4B,$9A,$93,$C5,$FE,$0A,$FC,$D8
	.dc.b	$CE,$3F,$84,$BE,$6D,$AD,$DF,$4B,$E4,$7F,$BD,$3E,$5F,$F8,$57,$ED
	.dc.b	$3F,$D4,$A3,$07,$FB,$ED,$C5,$FF,$BD,$E7,$73,$FC,$47,$E0,$91,$88
	.dc.b	$FF,$3B,$FD,$3F,$EE,$CE,$BE,$75,$8C,$92,$FF,$3D,$82,$75,$18,$FA
	.dc.b	$0F,$2F,$2F,$99,$B7,$E5,$6C,$EE,$3D,$13,$B6,$9B,$2C,$78,$3E,$2B
	.dc.b	$E0,$FD,$D7,$C2,$6C,$FC,$16,$CE,$E3,$E2,$4F,$C8,$F4,$3E,$4F,$75
	.dc.b	$9F,$90,$DA,$F6,$B7,$A3,$FB,$C7,$73,$5F,$DF,$C7,$F7,$BA,$FF,$06
	.dc.b	$D7,$EE,$FC,$CC,$E1,$FE,$64,$E4,$7E,$16,$F7,$B5,$72,$BF,$DA,$97
	.dc.b	$C3,$06,$1C,$2F,$51,$FC,$2A,$74,$F8,$2B,$FD,$DB,$5B,$79,$DF,$F5
	.dc.b	$4F,$39,$F5,$2F,$87,$C3,$6B,$F1,$1F,$69,$89,$FE,$29,$E9,$3E,$2E
	.dc.b	$99,$AF,$F7,$54,$65,$FF,$7E,$F0,$BA,$9F,$27,$F8,$F3,$BF,$6D,$AF
	.dc.b	$F4,$3F,$1F,$F8,$8F,$57,$E6,$0F,$1B,$F9,$27,$AA,$F9,$7F,$54,$E7
	.dc.b	$0F,$F7,$D4,$E5,$B3,$C0,$DD,$58,$EE,$7F,$83,$29,$F6,$5D,$0D,$72
	.dc.b	$FF,$BD,$B1,$F1,$C7,$F3,$5F,$A8,$7C,$45,$3F,$B1,$FC,$1F,$FB,$77
	.dc.b	$1E,$B8,$7F,$A8,$AB,$71,$88,$46,$1F,$E8,$6D,$D7,$FC,$89,$E8,$2F
	.dc.b	$A5,$6B,$0F,$06,$1F,$E5,$68,$ED,$1F,$91,$7F,$D5,$B8,$75,$7F,$CB
	.dc.b	$6D,$7F,$D6,$3F,$7F,$F0,$2B,$C8,$D8,$FE,$C5,$7F,$95,$9C,$E5,$FD
	.dc.b	$56,$B5,$97,$E7,$DB,$3F,$DC,$EB,$85,$FF,$5E,$9C,$BF,$EC,$46,$A7
	.dc.b	$39,$7B,$FB,$3B,$63,$FD,$8F,$E8,$7F,$A3,$5F,$F6,$86,$E5,$FA,$FF
	.dc.b	$BD,$7C,$3E,$EA,$69,$F3,$EF,$6F,$FB,$42,$B6,$E6,$76,$75,$3F,$B9
	.dc.b	$46,$9C,$47,$3D,$FC,$87,$F7,$7F,$27,$FB,$6C,$F9,$63,$33,$B3,$F0
	.dc.b	$13,$AC,$7F,$83,$61,$FE,$84,$BE,$15,$B7,$FC,$0B,$5B,$8F,$F7,$3C
	.dc.b	$EE,$B8,$FC,$A7,$C5,$3F,$C1,$1F,$79,$F8,$68,$CD,$7C,$19,$D3,$F9
	.dc.b	$19,$D3,$D2,$33,$87,$CD,$79,$FC,$36,$2F,$CF,$F2,$9F,$4C,$EF,$BD
	.dc.b	$C3,$E7,$45,$C3,$FC,$3B,$8F,$E0,$AF,$8D,$7F,$53,$F8,$0F,$F3,$2F
	.dc.b	$D1,$BE,$8F,$FA,$E6,$1F,$D2,$A7,$31,$8B,$FE,$5F,$E4,$FF,$99,$30
	.dc.b	$FD,$6E,$1C,$7D,$2F,$EC,$38,$70,$7F,$CA,$BF,$4B,$FE,$22,$AF,$F2
	.dc.b	$33,$72,$78,$5F,$E7,$8F,$4B,$7E,$84,$BF,$DC,$F4,$B0,$E1,$7B,$9C
	.dc.b	$C7,$3F,$84,$7D,$37,$F1,$1C,$9B,$FF,$AF,$67,$3E,$17,$F9,$FF,$EC
	.dc.b	$FF,$CE,$75,$AC,$7F,$DA,$B8,$71,$C3,$83,$63,$37,$C7,$E1,$7F,$D1
	.dc.b	$73,$9B,$63,$DA,$FF,$88,$B6,$3F,$CA,$58,$F8,$31,$EB,$8F,$FA,$53
	.dc.b	$E9,$1F,$2D,$4F,$F4,$2F,$D8,$DC,$B9,$CE,$9B,$3A,$69,$FC,$77,$5F
	.dc.b	$F2,$DF,$77,$F1,$5C,$FE,$57,$0F,$FB,$9D,$FE,$31,$3A,$6C,$65,$FF
	.dc.b	$78,$FC,$5F,$FA,$2B,$E9,$79,$36,$3F,$90,$35,$DF,$BE,$9F,$05,$BD
	.dc.b	$AF,$8B,$5E,$06,$BF,$EF,$E8,$D7,$0F,$09,$CC,$8F,$9B,$F4,$BF,$D8
	.dc.b	$5F,$9A,$D7,$ED,$1A,$F4,$B8,$3F,$E8,$1F,$8A,$7C,$A6,$39,$3F,$09
	.dc.b	$7C,$5B,$79,$E0,$F0,$64,$AD,$47,$FB,$3F,$0E,$78,$FD,$EF,$95,$FE
	.dc.b	$DA,$FF,$43,$B4,$62,$3C,$67,$E1,$9C,$3E,$68,$DF,$A7,$F5,$DB,$1D
	.dc.b	$63,$C8,$F1,$6F,$53,$EA,$3F,$C9,$33,$B2,$1F,$12,$B9,$BF,$EA,$38
	.dc.b	$DB,$FE,$AF,$F2,$3F,$C2,$4F,$EC,$1A,$E7,$D7,$EF,$93,$E4,$7C,$7D
	.dc.b	$EB,$DF,$07,$F6,$86,$ED,$FE,$C5,$6E,$3F,$C0,$30,$E4,$FF,$85,$BD
	.dc.b	$77,$D4,$7B,$BC,$5F,$E2,$93,$7C,$7F,$93,$97,$F9,$02,$F7,$3C,$6F
	.dc.b	$4C,$30,$7F,$7F,$FA,$3F,$F4,$B3,$FA,$36,$BF,$DE,$5A,$FD,$66,$BF
	.dc.b	$8A,$77,$5F,$E4,$D6,$E7,$9F,$88,$FA,$BD,$4F,$A8,$FF,$39,$61,$E1
	.dc.b	$F9,$EC,$FD,$DB,$C8,$A7,$DC,$96,$4A,$C3,$63,$D1,$1F,$BB,$76,$D8
	.dc.b	$D4,$7A,$E3,$89,$1F,$F1,$1E,$1E,$5E,$DF,$07,$89,$F4,$A7,$F5,$32
	.dc.b	$FB,$BF,$F5,$6C,$70,$6B,$FE,$E8,$9E,$1B,$7F,$9B,$DE,$51,$F9,$E6
	.dc.b	$BF,$B2,$6B,$FB,$B6,$A7,$9E,$DB,$E6,$F2,$9F,$33,$B3,$F0,$31,$CF
	.dc.b	$E6,$3E,$7E,$2F,$A0,$A7,$B2,$26,$3F,$C8,$71,$47,$C9,$6C,$66,$30
	.dc.b	$70,$7D,$CF,$87,$C9,$FA,$DF,$55,$FC,$DA,$7F,$C9,$5E,$07,$F9,$8D
	.dc.b	$B3,$AF,$26,$47,$F3,$1A,$E0,$FF,$91,$DA,$FF,$04,$6B,$FC,$61,$DE
	.dc.b	$DF,$E9,$B6,$E9,$9A,$FC,$46,$1F,$C5,$6F,$FC,$B9,$F5,$5F,$E7,$77
	.dc.b	$FC,$BD,$F5,$4F,$5B,$F1,$CD,$8D,$FC,$0F,$E4,$6B,$84,$7F,$96,$DA
	.dc.b	$FF,$18,$6B,$FC,$B1,$AF,$F3,$57,$8E,$7F,$EB,$16,$E5,$BA,$F0,$F8
	.dc.b	$9F,$9F,$FA,$97,$D5,$52,$63,$FE,$75,$E7,$FD,$1B,$7A,$7F,$41,$18
	.dc.b	$8F,$F3,$83,$5F,$E5,$AD,$7F,$A0,$D7,$ED,$9F,$E9,$51,$C7,$3F,$F6
	.dc.b	$33,$72,$DF,$6F,$F2,$CE,$F7,$CF,$52,$61,$FE,$89,$F9,$6F,$C0,$3F
	.dc.b	$D4,$07,$D5,$7F,$3C,$7F,$A8,$B5,$FE,$B4,$D7,$7C,$7F,$BB,$9F,$B5
	.dc.b	$1C,$DB,$9F,$FB,$6D,$B9,$7F,$63,$F2,$BF,$D7,$9E,$EF,$F0,$9F,$77
	.dc.b	$FD,$C1,$6A,$FF,$6F,$71,$3E,$E3,$FA,$4F,$C7,$7F,$85,$9E,$13,$33
	.dc.b	$FF,$03,$61,$FE,$99,$7F,$B4,$B5,$FE,$E4,$D7,$FD,$1A,$7C,$17,$C3
	.dc.b	$1C,$0E,$F7,$FD,$A2,$7F,$C7,$FB,$BD,$FF,$9D,$F5,$72,$7F,$A1,$47
	.dc.b	$D1,$30,$FF,$7C,$D7,$FD,$B5,$39,$8F,$E9,$1E,$7F,$D7,$F8,$B9,$18
	.dc.b	$21,$C7,$0F,$E3,$3B,$7F,$49,$9F,$BD,$5F,$7F,$EA,$BF,$01,$F5,$73
	.dc.b	$0F,$EA,$91,$97,$FA,$B4,$65,$E9,$19,$7B,$46,$6F,$FA,$B4,$FF,$66
	.dc.b	$FA,$BF,$E6,$B1,$FE,$6A,$EC,$FE,$F3,$F0,$7F,$A6,$C1,$EA,$2F,$7C
	.dc.b	$3F,$E2,$18,$C7,$99,$CC,$95,$FD,$A7,$C3,$FC,$C7,$C6,$7E,$67,$C9
	.dc.b	$93,$FD,$B6,$32,$FF,$6F,$8C,$BF,$16,$3F,$15,$E7,$7F,$53,$FA,$B7
	.dc.b	$D3,$D3,$35,$FE,$E9,$E9,$FE,$B4,$FE,$22,$BF,$AA,$46,$23,$FD,$C8
	.dc.b	$D7,$FC,$42,$E7,$83,$F2,$D1,$F8,$C9,$EA,$FE,$AE,$FF,$C8,$B3,$AE
	.dc.b	$77,$D5,$BE,$AF,$FC,$42,$F3,$E0,$FD,$23,$70,$FF,$46,$BB,$7F,$C9
	.dc.b	$57,$CE,$31,$1F,$8D,$78,$5F,$6C,$7A,$E1,$FE,$1E,$F4,$1E,$C7,$F0
	.dc.b	$5C,$21,$FE,$96,$8C,$B6,$38,$11,$F0,$E5,$E6,$F1,$7F,$C0,$1F,$1F
	.dc.b	$FE,$E8,$9C,$9F,$E5,$FF,$79,$F1,$DF,$E0,$56,$C6,$84,$F1,$7F,$89
	.dc.b	$35,$FF,$27,$B5,$FF,$31,$B5,$FF,$36,$35,$E7,$F1,$DE,$42,$3D,$E7
	.dc.b	$FC,$36,$FF,$1F,$FA,$B7,$8F,$FE,$52,$FA,$BF,$ED,$73,$FB,$76,$F1
	.dc.b	$7D,$E8,$57,$07,$D4,$D7,$A5,$7F,$88,$E3,$13,$F8,$1A,$FC,$9D,$7F
	.dc.b	$B1,$3E,$53,$F9,$AB,$FE,$49,$F8,$8F,$E7,$07,$F9,$52,$33,$1F,$E8
	.dc.b	$4E,$C4,$62,$FF,$23,$18,$F3,$78,$1F,$E6,$CF,$A0,$7C,$B7,$F9,$7B
	.dc.b	$EA,$3F,$B4,$63,$FE,$7A,$B4,$62,$3F,$D1,$F1,$8F,$5F,$FD,$60,$D8
	.dc.b	$FF,$40,$9F,$BE,$9E,$78,$F8,$9B,$C9,$FB,$F6,$CF,$F9,$9F,$DA,$FF
	.dc.b	$38,$91,$FC,$3F,$57,$E5,$23,$FC,$EB,$F1,$39,$9F,$F5,$04,$5C,$3F
	.dc.b	$55,$D0,$AF,$67,$BD,$E6,$C7,$FC,$6B,$9F,$B2,$30,$4C,$7F,$DE,$73
	.dc.b	$FE,$36,$7D,$51,$FE,$84,$F0,$3F,$D2,$BE,$7B,$EA,$BF,$0F,$5F,$E9
	.dc.b	$D6,$BF,$EA,$C6,$BF,$EE,$86,$C7,$1F,$A1,$E7,$1F,$90,$D3,$FD,$F4
	.dc.b	$DD,$FF,$D6,$5F,$57,$AA,$98,$7F,$BC,$2B,$4F,$FA,$3E,$72,$FF,$A8
	.dc.b	$E3,$38,$77,$8F,$97,$B2,$BC,$3F,$41,$FE,$B3,$B3,$49,$5F,$EB,$F8
	.dc.b	$CB,$E1,$8E,$2F,$4E,$4F,$F3,$F7,$C7,$7F,$94,$B4,$FF,$6E,$73,$FE
	.dc.b	$5A,$3F,$CE,$4D,$DF,$FD,$B2,$E7,$8D,$5C,$39,$FF,$DB,$B7,$CF,$1F
	.dc.b	$CB,$C6,$30,$FF,$70,$F3,$F3,$9F,$F7,$53,$CE,$ED,$72,$C3,$AD,$7E
	.dc.b	$8C,$FF,$78,$61,$E4,$9E,$91,$E1,$33,$3C,$31,$E9,$1C,$45,$71,$79
	.dc.b	$E3,$EC,$BF,$06,$1C,$E7,$78,$78,$B1,$F4,$47,$B0,$E2,$7A,$0F,$0C
	.dc.b	$D7,$8B,$CF,$0F,$04,$78,$B0,$F0,$3C,$0F,$34,$7C,$9C,$7B,$93,$1F
	.dc.b	$2C,$EB,$1F,$74,$7C,$28,$E3,$7D,$1F,$09,$BC,$3C,$98,$7F,$AB,$71
	.dc.b	$F5,$9F,$18,$F8,$F8,$F7,$AF,$69,$C4,$91,$C3,$0F,$03,$C5,$F2,$B9
	.dc.b	$8E,$2E,$5F,$76,$FF,$C6,$9B,$1E,$C8,$C4,$7A,$1A,$87,$CF,$D8,$CC
	.dc.b	$60,$38,$8D,$4F,$F1,$E6,$CF,$33,$A3,$F5,$D7,$4F,$F8,$0A,$7F,$9B
	.dc.b	$C6,$8F,$C0,$1E,$38,$D5,$F5,$8C,$9C,$B8,$FA,$3E,$5A,$E7,$5C,$23
	.dc.b	$C4,$7A,$71,$CD,$7E,$46,$3C,$11,$87,$F3,$47,$8E,$B0,$79,$9E,$D7
	.dc.b	$D2,$FB,$47,$BF,$0E,$91,$FE,$1E,$BE,$11,$8C,$3C,$15,$E3,$BE,$51
	.dc.b	$89,$D4,$F8,$8D,$1F,$2E,$FA,$57,$C3,$BF,$41,$DA,$7E,$2C,$F9,$1F
	.dc.b	$54,$F6,$7E,$49,$D0,$F9,$D7,$E1,$7E,$0C,$F6,$7E,$4D,$F9,$2F,$B1
	.dc.b	$E0,$BF,$A9,$87,$C6,$69,$DD,$DB,$E1,$8B,$5E,$BC,$3F,$86,$63,$FD
	.dc.b	$27,$E5,$BF,$C8,$BA,$7F,$40,$6C,$FF,$A8,$78,$DF,$49,$8C,$3F,$A5
	.dc.b	$55,$C3,$95,$FF,$35,$8F,$F4,$E4,$62,$7F,$06,$D7,$FA,$B7,$8D,$C7
	.dc.b	$90,$F5,$EB,$FF,$20,$FD,$56,$62,$63,$FD,$3D,$D6,$69,$CE,$1F,$D3
	.dc.b	$E6,$E9,$FD,$CA,$6D,$FF,$63,$9F,$2E,$3E,$08,$FE,$99,$18,$3F,$D6
	.dc.b	$9D,$7F,$E1,$0E,$CF,$80,$D9,$FE,$B3,$F5,$78,$ED,$FE,$FD,$1F,$DE
	.dc.b	$2B,$8C,$63,$3F,$F0,$D3,$5F,$F0,$CB,$5F,$9D,$F5,$7E,$C3,$DB,$AD
	.dc.b	$BF,$EC,$F1,$67,$FD,$6D,$3F,$E6,$69,$C9,$FE,$08,$D3,$C8,$6F,$5F
	.dc.b	$F1,$03,$88,$FF,$12,$E7,$FE,$2C,$8F,$39,$23,$3F,$50,$FA,$4F,$82
	.dc.b	$D8,$FF,$11,$F6,$3E,$AF,$35,$27,$FC,$4F,$ED,$BE,$AF,$1A,$FF,$B1
	.dc.b	$5F,$C5,$2F,$AB,$7C,$D6,$1B,$7F,$E3,$0F,$6B,$93,$FE,$73,$8C,$D7
	.dc.b	$F7,$68,$FE,$41,$18,$DF,$FD,$35,$4F,$83,$E9,$87,$E3,$63,$17,$FE
	.dc.b	$7D,$3E,$B5,$7F,$9A,$5A,$7F,$35,$1F,$57,$FE,$72,$78,$72,$FA,$9F
	.dc.b	$B5,$96,$E3,$FD,$F5,$18,$83,$4E,$AE,$3E,$81,$E7,$F3,$AF,$D7,$52
	.dc.b	$72,$FF,$A6,$39,$1C,$6F,$4B,$EA,$9F,$A9,$E4,$EB,$3A,$8C,$D7,$D4
	.dc.b	$F8,$FB,$9F,$FB,$DD,$E1,$7C,$7C,$0B,$5F,$79,$B5,$F3,$C7,$72,$BF
	.dc.b	$82,$D7,$E1,$1E,$4F,$FA,$A7,$E1,$9F,$07,$F8,$06,$C7,$FA,$C8,$F9
	.dc.b	$BA,$E6,$64,$7F,$9E,$B0,$57,$07,$FC,$FA,$D7,$FA,$7B,$5F,$EC,$4E
	.dc.b	$F8,$7F,$DC,$33,$F6,$BE,$1F,$AD,$7A,$E3,$07,$E6,$3E,$8B,$F8,$97
	.dc.b	$C1,$7D,$88,$C1,$FE,$82,$71,$3F,$0F,$47,$52,$F8,$DF,$F1,$B8,$C3
	.dc.b	$C4,$FE,$11,$F4,$31,$E9,$E8,$DC,$3F,$D5,$5E,$12,$BC,$7E,$CE,$75
	.dc.b	$FF,$7B,$69,$FE,$D8,$C7,$C9,$F5,$7E,$8E,$CF,$4B,$63,$27,$A6,$32
	.dc.b	$D3,$39,$7A,$EB,$F5,$51,$FE,$15,$D9,$0B,$E9,$5B,$F8,$CF,$8F,$F5
	.dc.b	$7F,$21,$B8,$7C,$76,$CE,$E3,$E5,$4F,$CE,$F5,$3E,$7F,$87,$67,$E7
	.dc.b	$36,$BC,$8D,$D3,$E0,$BF,$C2,$BD,$0F,$E2,$93,$F9,$53,$FC,$57,$1F
	.dc.b	$D5,$F9,$5F,$56,$F8,$2E,$13,$31,$FC,$76,$1C,$B0,$FE,$E3,$3F,$EE
	.dc.b	$B7,$FC,$1C,$D7,$FC,$5A,$F5,$BF,$D0,$C7,$86,$7C,$78,$FF,$00,$8F
	.dc.b	$CC,$CE,$A3,$1E,$9B,$EA,$DF,$17,$FD,$B6,$30,$FE,$4B,$EA,$F1,$3A
	.dc.b	$CF,$FC,$91,$7F,$EF,$FF,$59,$F4,$5E,$5C,$FF,$DD,$6D,$8C,$D7,$EA
	.dc.b	$DB,$3F,$E3,$6E,$5F,$CC,$F6,$FE,$5C,$EF,$DD,$6F,$B9,$FE,$45,$FA
	.dc.b	$B7,$95,$FC,$EB,$D7,$78,$BE,$99,$CD,$70,$CB,$FD,$63,$3B,$3C,$D5
	.dc.b	$C2,$7F,$7C,$D9,$FF,$7C,$B5,$FF,$19,$57,$E7,$DF,$E2,$5C,$28,$C6
	.dc.b	$DF,$3E,$55,$CE,$3E,$7C,$EA,$FF,$CC,$DF,$56,$F8,$9F,$EA,$6D,$8F
	.dc.b	$E5,$DF,$0A,$F8,$1C,$D4,$9F,$E2,$95,$BC,$BF,$CB,$E1,$C3,$FE,$62
	.dc.b	$6B,$E0,$9F,$F1,$56,$2F,$B0,$7C,$6F,$5A,$FE,$5D,$39,$8F,$EC,$FF
	.dc.b	$59,$E9,$9F,$F4,$43,$7C,$E6,$9F,$9D,$CF,$E8,$52,$B8,$37,$C5,$FC
	.dc.b	$34,$7C,$B8,$FF,$3B,$B5,$FF,$45,$35,$EF,$1F,$95,$FC,$01,$F3,$1E
	.dc.b	$48,$C3,$FD,$F3,$E9,$7F,$B4,$57,$EC,$DB,$A7,$D5,$FA,$DF,$EB,$73
	.dc.b	$93,$81,$F9,$99,$E2,$FF,$A2,$9A,$FF,$A7,$5A,$FF,$AA,$5E,$B9,$7F
	.dc.b	$AA,$FE,$85,$F4,$BD,$B0,$FC,$25,$6A,$D5,$AF,$5F,$D1,$CF,$31,$F9
	.dc.b	$3E,$8F,$F8,$42,$FD,$53,$97,$EF,$E1,$FE,$A3,$6C,$7A,$49,$5F,$A1
	.dc.b	$C1,$FA,$C7,$3C,$BC,$75,$FE,$6B,$9C,$FB,$9F,$27,$DB,$98,$7D,$AF
	.dc.b	$AE,$FC,$93,$F7,$33,$FF,$0C,$E7,$F3,$6B,$E4,$12,$7F,$B4,$4E,$A7
	.dc.b	$E5,$4F,$7F,$B2,$FF,$1C,$57,$38,$C7,$BF,$F6,$7E,$72,$4F,$87,$CD
	.dc.b	$B8,$7D,$B9,$FA,$C4,$9E,$EF,$F2,$5F,$17,$FA,$3F,$C4,$7F,$A1,$FE
	.dc.b	$A7,$E9,$FB,$7F,$DB,$23,$1C,$DF,$E1,$DC,$3F,$B6,$46,$27,$51,$FC
	.dc.b	$26,$B8,$BF,$E2,$1F,$AC,$FE,$AB,$93,$B9,$18,$C9,$F0,$BF,$DA,$6F
	.dc.b	$87,$B3,$FE,$7A,$AE,$37,$C2,$3F,$D1,$7E,$D6,$5F,$F4,$E7,$D7,$FF
	.dc.b	$09,$BF,$F7,$A3,$72,$FF,$10,$FD,$7F,$FA,$25,$FF,$12,$7C,$62,$74
	.dc.b	$DF,$33,$99,$FE,$9D,$AE,$1D,$9F,$D1,$67,$58,$7F,$27,$6E,$1F,$EA
	.dc.b	$8C,$3F,$91,$3E,$2C,$3F,$D5,$B1,$88,$4B,$EA,$1F,$3B,$FE,$B5,$6E
	.dc.b	$1F,$CC,$F8,$3E,$2F,$D7,$E6,$A4,$F8,$6B,$78,$7F,$A3,$AB,$2F,$F6
	.dc.b	$DE,$0F,$F3,$BF,$7F,$F8,$D5,$B1,$FC,$89,$A8,$87,$CF,$8F,$F1,$E7
	.dc.b	$C8,$BE,$27,$F6,$73,$FC,$AA,$31,$E0,$EA,$73,$3C,$84,$E9,$F0,$7C
	.dc.b	$07,$C2,$6E,$DE,$58,$C4,$71,$C5,$18,$8F,$34,$BE,$BF,$FB,$5C,$7C
	.dc.b	$7E,$7F,$F3,$AB,$7C,$6F,$5F,$27,$D9,$AD,$CF,$E0,$9B,$1F,$E0,$1E
	.dc.b	$2F,$C9,$D7,$0E,$1F,$F7,$63,$6F,$EC,$DF,$09,$F1,$FC,$DB,$D7,$D6
	.dc.b	$0E,$77,$D2,$61,$B8,$FE,$26,$F8,$36,$F5,$52,$30,$43,$FC,$B5,$5C
	.dc.b	$9B,$1E,$5D,$57,$D2,$7F,$D0,$5D,$6F,$17,$FC,$33,$1F,$8D,$FD,$EF
	.dc.b	$EB,$6F,$F0,$5B,$AF,$E0,$EF,$87,$BD,$F0,$5B,$7C,$4F,$7D,$70,$7F
	.dc.b	$7D,$87,$67,$FD,$25,$DF,$FC,$F9,$F9,$4D,$7F,$84,$78,$5C,$78,$3F
	.dc.b	$CE,$BF,$5F,$A8,$E9,$5B,$AE,$71,$BF,$8A,$57,$D5,$9F,$EB,$FF,$1B
	.dc.b	$F5,$FA,$DF,$5C,$F4,$B0,$DA,$FE,$E3,$E9,$7F,$AD,$5B,$7A,$A9,$18
	.dc.b	$37,$EB,$72,$F9,$59,$18,$21,$FE,$B1,$C7,$C3,$1A,$8F,$F3,$06,$1F
	.dc.b	$E0,$A7,$7C,$9F,$E5,$1A,$DC,$7F,$53,$C9,$7E,$DB,$E1,$3F,$E6,$9F
	.dc.b	$0D,$F3,$1E,$FF,$AE,$FF,$5C,$CF,$F5,$B6,$BB,$8F,$8D,$3C,$23,$5F
	.dc.b	$03,$C4,$CC,$C8,$C9,$0E,$47,$F8,$62,$5F,$35,$E1,$FA,$49,$39,$8E
	.dc.b	$E6,$1F,$89,$82,$73,$1B,$3D,$0E,$F5,$FF,$23,$5F,$69,$FF,$13,$FD
	.dc.b	$77,$F2,$87,$FC,$53,$E0,$7F,$97,$1B,$E6,$7E,$0E,$75,$1F,$E8,$F6
	.dc.b	$DF,$F9,$46,$B8,$4F,$FA,$4D,$B1,$FE,$55,$9D,$F6,$FF,$24,$6D,$7F
	.dc.b	$27,$AD,$CF,$F2,$86,$C7,$F2,$B9,$DA,$FA,$DF,$F3,$9B,$FE,$40,$FA
	.dc.b	$EF,$F2,$C3,$7E,$0F,$F3,$53,$A8,$FF,$53,$37,$1F,$F3,$45,$70,$9F
	.dc.b	$F5,$53,$63,$FC,$D9,$3B,$C7,$F9,$DB,$CE,$BF,$9A,$35,$CF,$C3,$7F
	.dc.b	$BE,$5F,$33,$7E,$B5,$F5,$DF,$C7,$3E,$51,$F7,$94,$FD,$84,$EA,$FF
	.dc.b	$CB,$7E,$6F,$F9,$41,$FF,$2E,$C6,$5F,$F6,$03,$73,$FF,$41,$4E,$E7
	.dc.b	$FD,$88,$D8,$CF,$3F,$EF,$63,$1A,$7F,$56,$BF,$DC,$5F,$FB,$F7,$BD
	.dc.b	$F4,$3E,$7F,$A8,$93,$FD,$AE,$72,$FF,$9C,$3E,$BF,$F9,$36,$7F,$E9
	.dc.b	$3F,$A8,$FB,$32,$6C,$EB,$2F,$EC,$B1,$FD,$46,$79,$72,$FF,$11,$AD
	.dc.b	$47,$F6,$B9,$D3,$FE,$DF,$F7,$BF,$D4,$18,$7F,$B3,$FE,$5B,$FC,$07
	.dc.b	$E7,$FF,$4D,$6D,$7F,$A1,$2B,$79,$7F,$74,$6C,$27,$6A,$FA,$79,$FF
	.dc.b	$D1,$5F,$F5,$75,$70,$9F,$F4,$81,$FE,$BF,$AE,$33,$C3,$0F,$E5,$55
	.dc.b	$DB,$CA,$6A,$F7,$FF,$AE,$AF,$AF,$A6,$22,$65,$F3,$E3,$2F,$D1,$8E
	.dc.b	$3E,$87,$2F,$12,$4F,$D5,$8C,$BF,$5E,$32,$FD,$D8,$F0,$FD,$77,$8B
	.dc.b	$EB,$E4,$FD,$D8,$CB,$F7,$E3,$2F,$E1,$63,$2F,$E1,$E3,$3A,$FF,$80
	.dc.b	$72,$57,$F5,$FF,$AF,$FE,$5F,$3F,$E0,$5E,$BF,$D6,$60,$D5,$18,$F4
	.dc.b	$D7,$C3,$8F,$FB,$53,$6F,$85,$72,$BE,$3D,$DD,$FC,$CF,$9B,$F2,$1F
	.dc.b	$3B,$C7,$98,$7E,$66,$32,$FE,$6E,$32,$FE,$86,$3E,$CF,$97,$FD,$8F
	.dc.b	$EB,$DF,$57,$4C,$D7,$E3,$67,$0F,$FB,$4A,$76,$E4,$87,$E3,$75,$FF
	.dc.b	$22,$C6,$A3,$07,$F9,$25,$C6,$FF,$90,$AF,$F1,$1F,$D7,$BB,$EC,$4B
	.dc.b	$FF,$2F,$4E,$5B,$39,$7A,$EC,$9F,$EC,$71,$A3,$FC,$B1,$6C,$7F,$C3
	.dc.b	$1E,$C7,$FA,$32,$7F,$80,$1F,$98,$3F,$C3,$4E,$5F,$F1,$2B,$67,$F0
	.dc.b	$51,$B3,$AC,$F4,$F7,$3F,$C1,$DF,$25,$FE,$4E,$9F,$82,$3E,$BF,$FC
	.dc.b	$18,$E1,$B7,$E5,$8F,$F3,$74,$F1,$8C,$42,$31,$7F,$E7,$2A,$E2,$DF
	.dc.b	$57,$C3,$EF,$4B,$F2,$9F,$E7,$9A,$DB,$67,$4F,$9B,$EB,$FF,$8E,$3F
	.dc.b	$E8,$DF,$AF,$FE,$D8,$FC,$E6,$E5,$E4,$9D,$CF,$1F,$17,$FC,$BB,$5A
	.dc.b	$CB,$FD,$11,$BA,$31,$5F,$E8,$BB,$E3,$87,$C1,$EA,$7C,$CF,$F3,$2A
	.dc.b	$FE,$71,$F0,$BF,$C3,$0F,$F2,$74,$66,$B8,$37,$2E,$FD,$CA,$E1,$7F
	.dc.b	$0B,$0E,$BE,$57,$F9,$5F,$E9,$3F,$B2,$61,$E6,$F8,$1E,$17,$FE,$61
	.dc.b	$C3,$5E,$A4,$8C,$10,$FF,$34,$DF,$27,$F8,$07,$2F,$BB,$24,$E5,$B5
	.dc.b	$FE,$82,$9F,$87,$C1,$FE,$DE,$F6,$7B,$61,$D2,$B5,$1E,$7F,$6B,$FC
	.dc.b	$CB,$F8,$5F,$9A,$BF,$CD,$2D,$E1,$F9,$59,$FF,$8E,$67,$CB,$5C,$34
	.dc.b	$91,$F0,$E7,$55,$FE,$D9,$FA,$F7,$D3,$FF,$9E,$3E,$CE,$5F,$FB,$6E
	.dc.b	$72,$D9,$CB,$FB,$2E,$05,$7F,$A3,$EB,$73,$FE,$EC,$6D,$FF,$A2,$E7
	.dc.b	$8B,$E1,$BE,$06,$38,$7E,$A4,$E2,$FC,$5F,$4D,$FE,$8C,$F8,$89,$18
	.dc.b	$21,$E3,$72,$F8,$63,$FD,$05,$F1,$5F,$E5,$4C,$7F,$DB,$9C,$FF,$8F
	.dc.b	$8F,$90,$DD,$FF,$DB,$EE,$78,$7F,$B5,$61,$F8,$A3,$FD,$C9,$5C,$DC
	.dc.b	$1F,$ED,$CA,$E3,$CD,$FE,$E8,$9E,$AD,$AE,$77,$D7,$0F,$1C,$EE,$F8
	.dc.b	$E1,$D9,$AF,$FB,$A2,$F8,$CF,$3A,$E3,$3B,$29,$E3,$2E,$1E,$7C,$3C
	.dc.b	$11,$E8,$8F,$2B,$C4,$C5,$F6,$7D,$ED,$7A,$1C,$70,$F2,$CF,$5C,$3C
	.dc.b	$93,$93,$C3,$3A,$AE,$55,$B9,$F0,$5F,$D1,$7C,$F3,$A3,$75,$D8,$E5
	.dc.b	$86,$8E,$B3,$0F,$4B,$97,$C4,$4A,$F8,$24,$C3,$E8,$47,$18,$C3,$F0
	.dc.b	$24,$BD,$47,$D0,$7B,$9B,$1B,$35,$19,$38,$B3,$E9,$79,$CF,$08,$C1
	.dc.b	$C0,$D8,$8D,$CE,$23,$06,$CE,$37,$FC,$A5,$FE,$45,$BF,$E0,$9F,$4D
	.dc.b	$5E,$4F,$EA,$15,$B9,$E6,$F1,$8C,$E5,$DE,$B8,$5F,$3B,$DC,$F9,$AB
	.dc.b	$71,$E8,$AD,$BF,$A4,$3C,$25,$8D,$BE,$89,$E5,$5D,$6B,$8D,$F9,$5B
	.dc.b	$3E,$3A,$E5,$5B,$69,$DC,$F2,$9F,$42,$FB,$BE,$2B,$F7,$4F,$16,$C7
	.dc.b	$1B,$EE,$FA,$CE,$B1,$E3,$6D,$F8,$E3,$89,$92,$B9,$3D,$B1,$D4,$E2
	.dc.b	$75,$8D,$B6,$37,$1F,$E9,$4C,$5C,$4F,$94,$FF,$04,$69,$FE,$09,$E8
	.dc.b	$72,$BE,$67,$3A,$FE,$3B,$0E,$0F,$F8,$12,$B7,$3A,$D3,$F3,$95,$A7
	.dc.b	$F9,$E4,$67,$2F,$0E,$5B,$BE,$38,$7F,$45,$F1,$F1,$93,$E0,$F8,$66
	.dc.b	$9C,$C6,$1B,$5C,$4F,$25,$71,$71,$9F,$F4,$DA,$E1,$1A,$6C,$66,$37
	.dc.b	$A7,$C3,$AE,$53,$FD,$DB,$E3,$FD,$DF,$18,$FC,$B1,$E8,$7C,$1F,$07
	.dc.b	$FB,$BD,$6A,$73,$97,$F6,$78,$C5,$6F,$B9,$18,$9F,$99,$E0,$BE,$C3
	.dc.b	$5F,$61,$F1,$E3,$13,$FD,$C6,$31,$3A,$8E,$2D,$D3,$FC,$1B,$3F,$33
	.dc.b	$9F,$1B,$FB,$5B,$7F,$E1,$96,$D7,$FA,$47,$7F,$F6,$FC,$FF,$88,$7E
	.dc.b	$B9,$F5,$5F,$67,$C6,$FC,$34,$E6,$31,$D3,$F0,$EB,$73,$FE,$E5,$8C
	.dc.b	$4E,$A3,$FC,$9B,$83,$EC,$1F,$33,$F8,$5F,$5D,$F3,$BF,$8D,$AD,$4E
	.dc.b	$63,$15,$C2,$FF,$29,$7C,$1E,$B1,$9B,$FF,$3F,$CE,$5F,$F3,$6C,$FD
	.dc.b	$23,$FC,$BE,$3E,$71,$E8,$F8,$63,$53,$FE,$94,$75,$19,$E4,$FD,$14
	.dc.b	$60,$F6,$9C,$38,$7F,$CE,$CE,$6B,$C3,$3C,$2F,$BC,$F0,$8C,$5F,$CB
	.dc.b	$F9,$3F,$A1,$D7,$6B,$F6,$5C,$3F,$96,$CB,$FC,$C5,$C2,$BD,$D7,$D9
	.dc.b	$6E,$7F,$EA,$0F,$AE,$1F,$B3,$1F,$5F,$5C,$36,$BF,$80,$46,$9A,$E9
	.dc.b	$FB,$7F,$0E,$F8,$3F,$87,$F6,$0F,$8A,$FB,$53,$9A,$F9,$F3,$97,$8D
	.dc.b	$70,$AD,$B6,$32,$70,$8C,$B4,$F8,$7F,$49,$FC,$37,$E0,$BF,$D7,$11
	.dc.b	$83,$F8,$C3,$8C,$FF,$D8,$3B,$AB,$85,$7E,$75,$AF,$F9,$7F,$E8,$3A
	.dc.b	$7B,$5F,$B9,$6C,$64,$EB,$19,$69,$E0,$FB,$0F,$27,$57,$95,$B1,$93
	.dc.b	$CB,$19,$69,$9C,$7F,$DA,$F9,$7E,$6F,$8F,$F9,$C2,$BE,$73,$AF,$8F
	.dc.b	$F8,$5F,$61,$F0,$DB,$87,$C2,$6C,$EE,$3E,$2C,$FC,$9F,$57,$E5,$77
	.dc.b	$59,$F9,$2D,$AF,$1B,$72,$F1,$BF,$B7,$F4,$3F,$80,$CF,$FB,$EB,$91
	.dc.b	$F6,$1F,$89,$8F,$F6,$EC,$EA,$FF,$DB,$55,$AE,$7E,$58,$7F,$4D,$8C
	.dc.b	$63,$FC,$22,$F8,$3F,$E2,$C8,$CE,$1D,$2B,$86,$3D,$F1,$FF,$71,$E1
	.dc.b	$C2,$97,$C2,$7F,$DD,$5E,$93,$EB,$DF,$19,$FC,$47,$D5,$7D,$7F,$F5
	.dc.b	$FA,$D4,$E6,$26,$BF,$EE,$89,$E5,$D5,$8C,$D9,$76,$CB,$F4,$51,$F9
	.dc.b	$2E,$AF,$CC,$7A,$3F,$CB,$5B,$8F,$F2,$6F,$5B,$F7,$CF,$FA,$33,$C1
	.dc.b	$F1,$7D,$7B,$C6,$FE,$4D,$5A,$C7,$FD,$2B,$5A,$EC,$CB,$70,$F3,$60
	.dc.b	$9F,$D4,$7C,$96,$6F,$F1,$EF,$F3,$1F,$AF,$7C,$63,$0F,$F2,$55,$27
	.dc.b	$31,$86,$C7,$F9,$3D,$AF,$F9,$51,$CE,$FF,$E1,$2F,$AF,$FE,$CF,$3F
	.dc.b	$1A,$BF,$CB,$F5,$A7,$FC,$D3,$4A,$DB,$E3,$AF,$F3,$6D,$6D,$F9,$F3
	.dc.b	$AA,$E7,$7F,$E1,$7B,$E5,$EA,$7F,$4D,$FA,$FF,$85,$5F,$67,$EB,$DF
	.dc.b	$33,$9D,$7F,$01,$18,$9F,$EB,$35,$C3,$3F,$40,$E6,$56,$E3,$22,$78
	.dc.b	$65,$F9,$09,$D3,$EB,$9D,$61,$E0,$C7,$C5,$8F,$95,$FE,$C9,$F4,$BF
	.dc.b	$E8,$C8,$F8,$A7,$FA,$3B,$E8,$DF,$4B,$FD,$33,$0F,$0D,$6D,$F3,$CE
	.dc.b	$BC,$FE,$71,$FE,$3B,$8C,$5F,$F7,$6A,$E1,$97,$CF,$9C,$D7,$43,$86
	.dc.b	$1D,$E7,$D5,$5C,$2F,$C9,$39,$7F,$D3,$71,$FE,$8E,$AF,$1F,$D7,$7F
	.dc.b	$83,$1F,$3E,$DE,$38,$C6,$1D,$23,$1E,$AF,$C2,$6E,$1E,$76,$DF,$28
	.dc.b	$F8,$AB,$26,$1D,$27,$C7,$EA,$FC,$1F,$AD,$7D,$6F,$CE,$8C,$36,$79
	.dc.b	$DA,$B7,$1F,$1E,$B6,$D3,$71,$AC,$3E,$9C,$E9,$FA,$BF,$37,$E1,$F8
	.dc.b	$6F,$BA,$DB,$FA,$4D,$9D,$C7,$49,$FF,$2A,$EF,$FE,$19,$8F,$7F,$C1
	.dc.b	$7E,$0B,$D9,$FE,$C9,$F0,$BA,$64,$EA,$B8,$61,$FD,$3A,$B2,$FE,$73
	.dc.b	$E0,$DF,$55,$FD,$46,$FF,$DB,$7F,$0D,$FE,$88,$F9,$1F,$F3,$14,$E6
	.dc.b	$FF,$C9,$51,$8F,$5B,$FC,$FA,$DB,$FF,$3B,$BF,$E8,$5F,$AD,$FE,$31
	.dc.b	$83,$E1,$5F,$3F,$9F,$1B,$FC,$C4,$DC,$3F,$CC,$ED,$CB,$F6,$37,$C6
	.dc.b	$31,$D5,$FE,$9D,$C3,$9D,$FE,$0B,$E2,$99,$78,$27,$38,$FE,$EA,$5F
	.dc.b	$4E,$FA,$DF,$F4,$AF,$1F,$FA,$6E,$31,$8F,$FA,$42,$31,$3A,$8E,$D3
	.dc.b	$AB,$F2,$FC,$37,$CE,$CF,$F7,$DC,$3F,$E6,$0E,$DF,$E5,$6D,$AF,$F7
	.dc.b	$B3,$5F,$F7,$DB,$51,$3E,$3B,$FD,$17,$D9,$FC,$D1,$B1,$FE,$FF,$6B
	.dc.b	$C0,$E1,$F0,$DF,$A3,$7B,$FB,$AD,$FC,$6D,$8C,$C6,$0E,$4F,$89,$7B
	.dc.b	$BF,$EC,$98,$F9,$DF,$A9,$3E,$CB,$C1,$AF,$F3,$7B,$4E,$6B,$EA,$5F
	.dc.b	$17,$E9,$36,$B8,$4E,$E7,$EB,$36,$3A,$4E,$F8,$3C,$93,$F8,$BA,$F1
	.dc.b	$F1,$F8,$A7,$2D,$87,$33,$F4,$EF,$E8,$31,$FE,$DF,$6B,$E0,$9E,$19
	.dc.b	$7A,$9E,$55,$C7,$C8,$B7,$C3,$3F,$4D,$70,$CB,$C9,$8F,$E6,$2B,$59
	.dc.b	$7E,$B3,$8F,$FC,$D1,$3F,$C6,$3B,$3C,$9E,$9F,$9D,$B1,$FE,$E0,$F5
	.dc.b	$3C,$35,$C2,$3F,$71,$87,$F8,$15,$E4,$D7,$F2,$73,$E8,$F8,$27,$DE
	.dc.b	$D2,$4F,$F9,$A3,$E2,$6B,$FE,$52,$8B,$B7,$F9,$4A,$7E,$54,$FF,$68
	.dc.b	$CE,$EF,$DE,$BF,$AE,$37,$5F,$EA,$DF,$A4,$3E,$57,$E9,$74,$A3,$18
	.dc.b	$FF,$B1,$AF,$83,$FA,$06,$CE,$EB,$BF,$A1,$2B,$83,$D2,$DE,$16,$67
	.dc.b	$FB,$9E,$1F,$E2,$17,$67,$F3,$C9,$DB,$67,$86,$5E,$AB,$E9,$1F,$03
	.dc.b	$BB,$F4,$D5,$FD,$F9,$B5,$FD,$ED,$B9,$FF,$75,$FB,$77,$D5,$71,$E9
	.dc.b	$98,$7F,$B9,$2F,$6F,$F5,$7A,$DB,$FD,$41,$AF,$79,$E2,$FF,$7C,$8C
	.dc.b	$E1,$E1,$F8,$1F,$31,$27,$97,$15,$7D,$11,$67,$FB,$F5,$7E,$12,$76
	.dc.b	$79,$67,$59,$7A,$EF,$F0,$3F,$5D,$FC,$13,$0F,$F0,$D3,$67,$F8,$83
	.dc.b	$6D,$D0,$FC,$47,$EE,$C2,$74,$D3,$A9,$FA,$18,$FC,$1C,$F1,$E4,$CC
	.dc.b	$FF,$7E,$38,$1F,$A9,$9E,$B5,$FE,$35,$8D,$35,$FF,$1A,$B7,$0F,$F5
	.dc.b	$6F,$B1,$EC,$F1,$25,$70,$7F,$51,$12,$30,$43,$FD,$6E,$FE,$93,$EB
	.dc.b	$BF,$61,$1F,$CA,$1A,$FF,$33,$6B,$FC,$F1,$A9,$F8,$A4,$C7,$F5,$FF
	.dc.b	$5F,$F8,$69,$34,$9C,$1F,$C7,$71,$7E,$F7,$FD,$6E,$BE,$37,$DC,$3E
	.dc.b	$3F,$F7,$2D,$8C,$C7,$EF,$3C,$FF,$E2,$F7,$FE,$55,$6F,$0F,$F4,$B6
	.dc.b	$D7,$F5,$46,$AB,$7C,$C7,$F0,$FF,$B1,$EE,$C9,$FF,$7A,$C7,$F0,$AE
	.dc.b	$27,$05,$24,$FD,$DF,$D2,$1E,$5E,$23,$FC,$DB,$5C,$1F,$C1,$35,$FF
	.dc.b	$38,$CF,$07,$94,$F1,$D3,$FC,$E1,$5B,$7F,$CE,$55,$BA,$FF,$78,$FC
	.dc.b	$57,$F2,$72,$3F,$E4,$CF,$25,$FB,$67,$EE,$17,$8B,$FE,$C8,$9D,$35
	.dc.b	$E6,$71,$9E,$03,$AF,$FA,$5B,$63,$51,$8F,$A3,$CB,$23,$27,$38,$E3
	.dc.b	$E1,$C8,$EC,$FF,$5F,$F6,$1F,$B4,$FB,$75,$FF,$55,$67,$EA,$AE,$17
	.dc.b	$FE,$2C,$CB,$FA,$AA,$C3,$8E,$FA,$C3,$8D,$6B,$0E,$57,$FE,$52,$C7
	.dc.b	$B7,$2F,$F9,$22,$BF,$CD,$B3,$FD,$77,$F5,$2F,$C8,$3F,$E4,$F6,$E1
	.dc.b	$D3,$DF,$F8,$67,$F6,$18,$CD,$FF,$88,$BC,$04,$62,$FF,$BD,$E1,$D9
	.dc.b	$F5,$7B,$1F,$33,$EA,$24,$7D,$18,$F9,$E6,$1F,$81,$18,$7F,$86,$7E
	.dc.b	$2C,$FA,$5F,$EB,$ED,$8C,$C7,$F6,$2F,$AE,$7C,$3E,$A2,$4E,$8F,$E2
	.dc.b	$D3,$A6,$CE,$9F,$F6,$3D,$75,$7F,$D9,$33,$B7,$EA,$7C,$0F,$E1,$88
	.dc.b	$FE,$2F,$E9,$AB,$A3,$F1,$D1,$F8,$C8,$FE,$EF,$F7,$7F,$CC,$3F,$35
	.dc.b	$7C,$07,$F7,$A6,$DF,$F2,$B9,$FE,$6A,$6E,$74,$72,$F2,$3E,$1F,$DA
	.dc.b	$CA,$F8,$EF,$F8,$93,$F0,$9F,$F6,$E7,$37,$FB,$77,$54,$EE,$BD,$EE
	.dc.b	$F9,$64,$72,$21,$FE,$7E,$BE,$71,$F8,$BE,$9F,$F1,$27,$EA,$7E,$8A
	.dc.b	$FF,$4C,$37,$1F,$CA,$CF,$EB,$0D,$E7,$2B,$F6,$51,$FB,$0F,$B4,$E7
	.dc.b	$F1,$52,$7F,$6D,$19,$7F,$C6,$98,$27,$51,$FA,$79,$F1,$FC,$D4,$8D
	.dc.b	$BF,$E3,$6F,$BD,$91,$82,$1F,$E8,$67,$FC,$77,$5A,$8C,$1F,$C3,$FD
	.dc.b	$FF,$03,$63,$59,$FE,$EA,$55,$CA,$BC,$FF,$27,$F1,$88,$F3,$F8,$DF
	.dc.b	$F2,$49,$87,$E4,$C6,$1F,$F5,$47,$E1,$6C,$FF,$CC,$2D,$8C,$C7,$F9
	.dc.b	$4B,$EE,$A4,$60,$87,$FA,$89,$CB,$FE,$AF,$9E,$4D,$9C,$E5,$EB,$AE
	.dc.b	$F1,$27,$E1,$47,$8B,$E1,$E4,$79,$9F,$F3,$1E,$B2,$30,$43,$FD,$66
	.dc.b	$FF,$96,$3E,$25,$D2,$C7,$FD,$5B,$F3,$4F,$AB,$FF,$54,$53,$B3,$FD
	.dc.b	$55,$39,$C7,$FD,$65,$18,$AD,$E1,$CE,$55,$B5,$7E,$0C,$3B,$DF,$6C
	.dc.b	$39,$57,$43,$27,$FA,$FF,$0E,$98,$73,$C3,$94,$F6,$BE,$4F,$03,$BD
	.dc.b	$F4,$7C,$37,$CA,$7C,$F8,$6E,$FD,$D5,$6F,$84,$76,$C3,$C1,$87,$1A
	.dc.b	$F3,$1E,$12,$C6,$AF,$A4,$EE,$7C,$67,$53,$31,$1E,$46,$8B,$5E,$46
	.dc.b	$CF,$38,$F2,$61,$F0,$27,$78,$7A,$2B,$46,$DF,$59,$9B,$F0,$61,$F2
	.dc.b	$E3,$35,$8B,$F8,$AF,$43,$D0,$78,$AB,$4F,$C0,$8F,$15,$7C,$18,$D1
	.dc.b	$FE,$ED,$F2,$F9,$FA,$FF,$DF,$63,$E9,$DF,$FB,$AE,$72,$7F,$BB,$A1
	.dc.b	$F4,$19,$7F,$DE,$2D,$7F,$DF,$8D,$47,$03,$C0,$70,$3E,$68,$F6,$DD
	.dc.b	$5F,$87,$8E,$33,$67,$31,$C0,$E9,$8F,$E6,$DE,$55,$C5,$CE,$5E,$27
	.dc.b	$C7,$5C,$AB,$84,$68,$D4,$66,$7E,$05,$F8,$6F,$A3,$6B,$C4,$DB,$E6
	.dc.b	$DA,$E3,$3E,$EB,$E8,$FC,$1A,$F4,$46,$23,$E4,$9B,$BE,$ED,$3B,$96
	.dc.b	$7B,$B6,$BB,$C7,$49,$E2,$D9,$E2,$D3,$89,$67,$89,$F1,$E3,$E5,$9C
	.dc.b	$1C,$57,$9E,$3E,$A9,$E1,$3D,$91,$AA,$F0,$B6,$3F,$69,$E5,$B9,$DF
	.dc.b	$3F,$F7,$34,$FB,$BA,$B9,$7F,$D0,$99,$7F,$4F,$AD,$3F,$8C,$9D,$7B
	.dc.b	$38,$C9,$18,$9F,$E8,$F3,$9C,$BC,$17,$BA,$D5,$FF,$2F,$BF,$CB,$D7
	.dc.b	$97,$EE,$B5,$F8,$33,$E1,$F9,$6A,$9C,$C6,$1B,$7F,$D9,$5B,$87,$51
	.dc.b	$87,$59,$FD,$2E,$BF,$03,$1E,$D1,$C4,$C9,$EF,$73,$5C,$F2,$46,$2B
	.dc.b	$96,$3E,$3B,$E3,$7F,$12,$B5,$1E,$EC,$BF,$C4,$2F,$F8,$2B,$EA,$3D
	.dc.b	$BE,$1F,$CB,$F3,$DE,$0A,$70,$DA,$E0,$7F,$80,$EB,$33,$F1,$BE,$4B
	.dc.b	$E3,$F7,$DA,$FF,$04,$37,$3F,$89,$BF,$E0,$E3,$1B,$7F,$82,$5E,$35
	.dc.b	$EA,$CD,$E8,$BE,$47,$38,$FC,$D9,$CC,$7A,$67,$55,$C5,$B5,$E0,$D9
	.dc.b	$86,$EF,$F0,$DE,$8B,$F0,$3E,$5F,$51,$86,$E5,$FE,$A7,$6E,$3F,$E9
	.dc.b	$56,$E9,$FE,$92,$FA,$CF,$17,$CC,$74,$6D,$FF,$AB,$DB,$7F,$E9,$D6
	.dc.b	$D7,$FA,$8B,$F0,$AF,$9D,$3F,$89,$3F,$98,$FC,$67,$E6,$E3,$33,$FC
	.dc.b	$92,$73,$7F,$80,$69,$C7,$C0,$5F,$1E,$4F,$D4,$4E,$AF,$95,$6A,$F9
	.dc.b	$36,$7E,$BE,$1E,$09,$FE,$7F,$87,$2F,$A3,$7C,$B5,$F4,$F0,$DA,$F2
	.dc.b	$4E,$9B,$3A,$69,$FE,$D3,$FA,$AF,$A6,$7D,$4F,$B0,$B1,$92,$65,$FE
	.dc.b	$C5,$8C,$BF,$1F,$F3,$07,$C1,$FB,$BE,$AD,$EF,$7C,$09,$CE,$1F,$1E
	.dc.b	$72,$D9,$E0,$6E,$AC,$64,$97,$DE,$32,$FF,$B5,$FC,$FF,$C5,$FC,$4F
	.dc.b	$E3,$1B,$5F,$EE,$F6,$CE,$E3,$FD,$E5,$3F,$92,$DB,$F3,$57,$67,$F3
	.dc.b	$4D,$AF,$D0,$76,$7A,$9F,$F7,$35,$70,$AF,$F7,$E7,$8F,$F9,$C3,$F9
	.dc.b	$37,$C1,$3F,$2C,$7C,$AE,$6B,$64,$C3,$F8,$C5,$6E,$75,$1F,$EF,$3B
	.dc.b	$BC,$FF,$DD,$2B,$57,$C6,$3F,$DE,$7C,$7F,$AC,$3F,$5B,$F0,$16,$7F
	.dc.b	$8A,$F8,$6E,$F5,$67,$2F,$F8,$01,$77,$0F,$F0,$0D,$FE,$6A,$7F,$9E
	.dc.b	$74,$7F,$83,$B1,$E6,$FF,$58,$AD,$B9,$9F,$E1,$ED,$8C,$C7,$2F,$3B
	.dc.b	$9B,$F8,$7C,$3F,$4B,$1F,$EA,$DB,$FC,$34,$63,$D9,$91,$FA,$F9,$D5
	.dc.b	$7E,$DF,$9B,$FD,$E3,$39,$AD,$CF,$EE,$EB,$78,$7E,$E6,$F7,$39,$AF
	.dc.b	$6F,$95,$FD,$77,$F0,$C7,$E1,$8F,$AF,$CE,$7F,$CF,$A7,$31,$FB,$CF
	.dc.b	$C3,$1E,$1D,$39,$5F,$1B,$FE,$0D,$3A,$7F,$97,$79,$DF,$CB,$9B,$D3
	.dc.b	$C5,$A7,$F3,$3F,$BD,$7C,$5F,$F4,$76,$C7,$F5,$0F,$C4,$9F,$6E,$FC
	.dc.b	$07,$57,$FC,$43,$EC,$3F,$98,$DF,$F1,$76,$F1,$7F,$3B,$C7,$9B,$FC
	.dc.b	$86,$33,$C9,$FC,$7D,$DC,$7F,$85,$DA,$FF,$8F,$DA,$FF,$94,$9B,$D1
	.dc.b	$FC,$8B,$E2,$FF,$90,$FD,$AF,$FA,$AF,$1F,$C4,$C6,$3D,$E9,$1F,$CA
	.dc.b	$E7,$58,$7E,$2A,$B2,$FF,$35,$8C,$E3,$FC,$CD,$B5,$FE,$0C,$EB,$5F
	.dc.b	$0A,$FF,$2D,$FA,$8D,$3F,$DB,$97,$DF,$DF,$F9,$9F,$07,$E6,$64,$64
	.dc.b	$FA,$51,$8C,$FC,$53,$70,$EB,$7D,$27,$37,$FD,$17,$F1,$17,$A5,$FD
	.dc.b	$0B,$EB,$DF,$57,$FD,$12,$73,$5F,$D2,$E7,$2D,$9E,$06,$EA,$C6,$49
	.dc.b	$7F,$D6,$6F,$9B,$FD,$6E,$F9,$B9,$E4,$F1,$7C,$57,$85,$B7,$E1,$8B
	.dc.b	$3B,$8C,$11,$E9,$1C,$63,$19,$FF,$65,$FC,$6F,$FC,$43,$5F,$E9,$4C
	.dc.b	$1E,$E7,$F6,$29,$CC,$A7,$37,$F0,$6B,$E3,$1B,$F0,$6C,$64,$95,$FD
	.dc.b	$DE,$F9,$DF,$28,$C7,$D1,$FC,$FF,$73,$C0,$DB,$F9,$D1,$67,$71,$83
	.dc.b	$F8,$EF,$D8,$7F,$20,$3F,$90,$D5,$8C,$92,$7F,$8F,$C6,$5F,$F1,$07
	.dc.b	$E0,$9F,$E3,$DF,$D5,$5F,$55,$FE,$5C,$8C,$36,$FF,$C6,$13,$A7,$F9
	.dc.b	$BF,$C0,$AB,$71,$FC,$EB,$E5,$FF,$98,$3F,$E7,$5F,$9C,$7A,$FF,$E7
	.dc.b	$19,$CE,$5F,$E7,$69,$CD,$71,$F3,$BF,$CE,$91,$89,$FF,$22,$D7,$0F
	.dc.b	$86,$4E,$79,$3F,$D0,$ED,$C3,$C1,$8F,$7C,$79,$BF,$E7,$FF,$8D,$FE
	.dc.b	$5F,$3E,$5F,$4B,$1E,$87,$81,$B1,$FE,$68,$CB,$BC,$63,$4F,$F3,$86
	.dc.b	$5F,$D5,$AB,$FC,$5D,$F5,$CF,$A1,$FF,$6D,$57,$FB,$9E,$31,$A7,$FB
	.dc.b	$A6,$11,$87,$53,$C1,$FF,$58,$66,$F9,$47,$D5,$71,$8C,$36,$33,$3F
	.dc.b	$EF,$59,$CB,$E1,$F8,$9F,$F5,$1B,$75,$F1,$63,$FE,$69,$6C,$EA,$7F
	.dc.b	$DC,$D0,$FA,$D7,$C0,$7F,$BA,$78,$BF,$D6,$B1,$8A,$F8,$CD,$7F,$D2
	.dc.b	$9F,$5C,$EA,$E1,$C1,$FE,$B7,$AD,$5F,$2C,$33,$87,$FA,$E2,$73,$5F
	.dc.b	$EB,$DC,$39,$4F,$6F,$80,$F9,$59,$79,$5B,$7F,$E5,$09,$FF,$6C,$1C
	.dc.b	$E3,$27,$69,$E1,$1C,$CF,$F0,$8F,$D7,$3E,$83,$57,$CE,$7F,$D5,$CD
	.dc.b	$CF,$E0,$E1,$E8,$F4,$18,$71,$BF,$F4,$B2,$F9,$CF,$6F,$9B,$A6,$C6
	.dc.b	$67,$F5,$7D,$F8,$99,$1C,$9E,$DD,$DF,$B6,$8F,$F1,$3F,$D0,$BE,$AB
	.dc.b	$AF,$82,$F9,$BA,$79,$67,$55,$C2,$B2,$FE,$F2,$75,$1F,$4A,$3E,$16
	.dc.b	$1F,$97,$69,$F9,$9E,$0F,$43,$F3,$9A,$88,$70,$3F,$D3,$0F,$0D,$FF
	.dc.b	$D2,$CD,$E3,$E1,$D4,$FA,$DF,$F7,$75,$FF,$A6,$6B,$FC,$AA,$FF,$B6
	.dc.b	$A5,$FB,$23,$E5,$FF,$13,$18,$7D,$55,$C2,$BC,$3F,$11,$EB,$8C,$47
	.dc.b	$98,$FC,$04,$64,$8F,$85,$E6,$76,$C1,$E3,$FC,$7F,$AE,$FD,$46,$8D
	.dc.b	$3D,$7E,$07,$FA,$3B,$1F,$43,$5F,$F5,$D5,$70,$9F,$8E,$FE,$C6,$3C
	.dc.b	$8D,$73,$D5,$F4,$E3,$11,$F3,$DE,$4D,$8F,$21,$F1,$3E,$BB,$FC,$CD
	.dc.b	$1E,$0F,$9D,$F8,$F8,$7F,$0D,$8F,$2B,$EB,$89,$1F,$C5,$ED,$1E,$12
	.dc.b	$46,$5D,$BD,$4F,$AD,$5F,$90,$7F,$DB,$1F,$5D,$F3,$F2,$FF,$0F,$F6
	.dc.b	$FF,$27,$9F,$F7,$E3,$5D,$9F,$93,$9F,$C9,$26,$91,$FC,$B3,$D7,$3F
	.dc.b	$C8,$B6,$3F,$28,$DE,$BF,$E9,$3C,$9D,$7C,$5F,$DA,$E0,$FA,$BA,$67
	.dc.b	$0F,$F7,$56,$1F,$CD,$5F,$C4,$35,$EF,$5E,$22,$3C,$1E,$4B,$2F,$F4
	.dc.b	$CF,$4B,$EA,$BF,$61,$F5,$BF,$E6,$AA,$FF,$7C,$35,$FF,$80,$0D,$AF
	.dc.b	$F7,$F7,$D7,$7E,$A6,$7F,$60,$FE,$C9,$FD,$8B,$5C,$E9,$FB,$B8,$C4
	.dc.b	$2B,$F1,$2A,$3A,$36,$7C,$BC,$7F,$CD,$BE,$11,$FA,$13,$C5,$C3,$6F
	.dc.b	$FD,$00,$D7,$E0,$71,$7F,$04,$E0,$7C,$8F,$F0,$FF,$33,$F9,$75,$7E
	.dc.b	$CD,$AF,$F0,$A6,$BF,$C3,$5A,$FD,$7F,$43,$F8,$C5,$7F,$83,$A3,$F8
	.dc.b	$96,$FF,$B3,$7F,$8A,$46,$5F,$18,$F1,$DC,$4F,$AE,$FE,$45,$E7,$FD
	.dc.b	$5A,$E7,$A7,$F1,$C8,$FD,$5E,$3D,$67,$FC,$27,$F0,$5C,$39,$64,$7F
	.dc.b	$7F,$AE,$4F,$F5,$46,$BF,$C4,$5A,$F5,$8E,$2F,$98,$FF,$0F,$BF,$CD
	.dc.b	$7E,$BB,$F7,$11,$EA,$CF,$D3,$F2,$BF,$D4,$F0,$7D,$25,$3F,$A1,$44
	.dc.b	$AF,$F7,$2C,$65,$FE,$97,$1F,$D6,$FE,$53,$F7,$F1,$FB,$4D,$7F,$4A
	.dc.b	$67,$BE,$46,$48,$78,$E7,$8B,$7C,$5F,$EC,$3E,$96,$37,$FF,$4F,$B7
	.dc.b	$83,$FA,$46,$BF,$F0,$06,$BC,$47,$0C,$66,$1F,$48,$FB,$18,$FF,$6F
	.dc.b	$7B,$19,$F0,$18,$F5,$84,$62,$BA,$C6,$FD,$7F,$C9,$BD,$71,$6F,$FE
	.dc.b	$A2,$8C,$35,$FC,$4D,$AB,$8C,$61,$BA,$7F,$AF,$1B,$A7,$F6,$5A,$E1
	.dc.b	$CF,$F3,$FE,$33,$E8,$7D,$6F,$C6,$F8,$A9,$97,$8F,$74,$EA,$3D,$DB
	.dc.b	$AB,$71,$F4,$67,$FB,$D7,$D5,$FF,$B3,$A3,$E4,$0E,$57,$C9,$7E,$2C
	.dc.b	$DF,$C5,$C9,$FF,$2B,$57,$17,$FC,$56,$D7,$FB,$43,$5F,$EF,$EE,$F8
	.dc.b	$3F,$48,$DD,$F3,$5F,$E0,$DF,$AD,$7C,$BF,$E3,$F4,$FC,$2F,$C6,$CC
	.dc.b	$BE,$04,$65,$F7,$C6,$5F,$C7,$47,$F8,$5F,$EB,$5F,$11,$7B,$25,$7C
	.dc.b	$09,$CB,$67,$81,$1F,$8B,$F0,$3F,$C8,$35,$7D,$2F,$F8,$8F,$F3,$47
	.dc.b	$D6,$D3,$E6,$C4,$BF,$8E,$FF,$89,$EB,$8C,$60,$87,$EE,$DC,$B6,$3F
	.dc.b	$C9,$5F,$5D,$FC,$2F,$8B,$F8,$C7,$53,$EA,$7D,$37,$FB,$FE,$DB,$5F
	.dc.b	$C7,$7B,$BF,$2F,$B7,$F9,$3A,$F3,$CD,$F8,$F6,$E1,$FA,$26,$E3,$D6
	.dc.b	$BF,$80,$7A,$BC,$7E,$1F,$F8,$57,$C0,$CC,$3F,$59,$82,$75,$18,$6B
	.dc.b	$FC,$07,$9B,$C5,$F5,$BF,$E9,$FC,$BD,$4D,$B3,$49,$5D,$9E,$2F,$FB
	.dc.b	$3B,$8B,$FC,$C2,$47,$07,$FB,$4B,$0C,$3F,$C8,$FE,$0B,$EB,$BA,$31
	.dc.b	$1F,$17,$E2,$BC,$44,$8F,$F4,$3E,$3F,$3E,$31,$1F,$BE,$C5,$1C,$63
	.dc.b	$FD,$C3,$F0,$5F,$E8,$BF,$AC,$69,$FE,$94,$6C,$FF,$A4,$BD,$44,$62
	.dc.b	$3F,$CA,$11,$89,$E0,$73,$24,$7F,$BB,$BE,$BA,$BF,$D6,$27,$83,$FE
	.dc.b	$D5,$AE,$9C,$9F,$83,$D5,$7C,$23,$FB,$15,$F0,$BE,$57,$C3,$8F,$F9
	.dc.b	$8E,$1C,$27,$A7,$C4,$7A,$C8,$FA,$EB,$FD,$7C,$FF,$B1,$9F,$F5,$BA
	.dc.b	$BF,$7C,$FC,$53,$FD,$33,$C3,$FE,$8D,$F8,$B7,$CB,$7F,$88,$ED,$F5
	.dc.b	$DF,$D6,$1F,$F4,$54,$FF,$1D,$7F,$D1,$91,$95,$19,$78,$F8,$1F,$E9
	.dc.b	$08,$EB,$AF,$08,$C5,$F7,$F3,$79,$57,$16,$F8,$5F,$E9,$87,$33,$BA
	.dc.b	$E3,$7C,$F0,$F2,$DF,$FA,$7B,$1E,$31,$93,$CB,$7E,$D8,$E7,$59,$C7
	.dc.b	$D1,$1C,$2B,$D2,$7C,$0A,$E1,$87,$82,$3F,$CE,$2F,$B7,$0F,$33,$63
	.dc.b	$FD,$61,$8F,$67,$67,$3C,$7C,$15,$E7,$AE,$13,$E0,$3C,$F3,$E5,$BF
	.dc.b	$0C,$7B,$8E,$26,$62,$3C,$71,$F5,$D7,$4C,$7C,$47,$8F,$1F,$71,$23
	.dc.b	$D8,$E9,$E0,$4A,$F4,$1D,$70,$FA,$4F,$8C,$F6,$9F,$82,$C7,$E1,$BC
	.dc.b	$2B,$C1,$5E,$E7,$C1,$19,$8F,$99,$A7,$F8,$5B,$1F,$C3,$CF,$E6,$6B
	.dc.b	$24,$7F,$D8,$DF,$55,$FE,$DD,$AF,$F0,$EC,$7F,$B7,$E3,$13,$FE,$CB
	.dc.b	$84,$61,$E2,$75,$2C,$7E,$6D,$7B,$DF,$E5,$78,$C4,$72,$3D,$E6,$B8
	.dc.b	$7E,$FE,$3D,$EE,$CF,$FB,$C2,$F9,$4E,$63,$8C,$61,$F1,$E9,$E1,$8F
	.dc.b	$F3,$33,$E0,$AF,$CD,$4F,$73,$F0,$C7,$BC,$E7,$7E,$07,$FD,$FF,$15
	.dc.b	$E1,$87,$37,$8D,$F6,$C3,$C9,$87,$87,$0E,$F5,$EF,$8C,$3E,$03,$13
	.dc.b	$FE,$F7,$AE,$51,$EA,$8C,$57,$2C,$3D,$35,$AA,$F2,$9E,$83,$E5,$9F
	.dc.b	$37,$0E,$E7,$9F,$0F,$09,$70,$EB,$3E,$3B,$EB,$87,$78,$F6,$B6,$F9
	.dc.b	$D7,$D1,$6C,$F9,$CF,$AD,$3C,$6F,$BD,$7E,$37,$83,$FD,$4B,$3F,$CF
	.dc.b	$7C,$EF,$F4,$D6,$FF,$CF,$7E,$1A,$CE,$E6,$E3,$FC,$C6,$F8,$BF,$D3
	.dc.b	$9A,$FF,$37,$8F,$F4,$C3,$C6,$7F,$AB,$7C,$EB,$E4,$F9,$EE,$8C,$4F
	.dc.b	$F9,$97,$95,$F5,$DA,$9C,$E1,$FD,$55,$BA,$7F,$50,$6D,$FF,$4E,$AE
	.dc.b	$18,$C8,$FF,$3E,$63,$DE,$3D,$7E,$13,$2E,$F5,$AC,$39,$E5,$FD,$3D
	.dc.b	$F0,$A3,$F6,$CE,$1B,$7F,$01,$E0,$DE,$DF,$F1,$77,$D6,$7C,$A6,$DF
	.dc.b	$BF,$C6,$C7,$7A,$A9,$C5,$C4,$FF,$82,$9B,$87,$8C,$FF,$19,$4F,$87
	.dc.b	$3F,$67,$7F,$BB,$4F,$F0,$6E,$1F,$37,$E5,$B1,$32,$3E,$83,$B8,$C6
	.dc.b	$1F,$E9,$2F,$53,$FD,$3F,$5F,$EA,$16,$C7,$F8,$8E,$BD,$91,$8E,$1F
	.dc.b	$F2,$6B,$A9,$F3,$6A,$8C,$47,$FA,$5D,$AF,$FA,$A9,$B1,$BF,$AE,$EA
	.dc.b	$7D,$8F,$55,$F5,$D8,$CB,$C9,$3F,$73,$CA,$FF,$35,$C6,$0F,$F5,$6F
	.dc.b	$8F,$E6,$F9,$B9,$1F,$EA,$1F,$52,$E9,$EF,$9E,$5E,$4E,$36,$FF,$31
	.dc.b	$63,$FE,$55,$C3,$FD,$3F,$E0,$7F,$AE,$9B,$1F,$8C,$FA,$EF,$AB,$5F
	.dc.b	$EF,$9A,$D6,$3F,$EE,$F6,$BF,$D0,$2B,$F1,$0F,$F3,$C6,$C6,$A3,$F9
	.dc.b	$0C,$7E,$C9,$7D,$5B,$DC,$C5,$76,$B6,$3F,$85,$AD,$3F,$EB,$6F,$AE
	.dc.b	$7A,$9C,$DB,$19,$F9,$7C,$4C,$8F,$F5,$EF,$81,$FE,$98,$79,$47,$8E
	.dc.b	$31,$9F,$FB,$D7,$EB,$BF,$3D,$E1,$B9,$3E,$C4,$61,$B8,$7F,$BA,$E7
	.dc.b	$4F,$FB,$4E,$74,$FF,$B3,$3E,$A9,$F5,$BF,$05,$B1,$FE,$D0,$D5,$F0
	.dc.b	$59,$A9,$3F,$D9,$F2,$5F,$07,$FB,$53,$5F,$D5,$1A,$9F,$A7,$1B,$DF
	.dc.b	$E5,$CF,$F0,$18,$6D,$DA,$13,$A7,$BC,$7E,$5F,$E1,$FE,$1F,$35,$C7
	.dc.b	$E0,$36,$BF,$DF,$FA,$A3,$18,$7F,$AD,$A3,$15,$FB,$1C,$FF,$5E,$7E
	.dc.b	$8F,$1B,$19,$3F,$8B,$57,$E6,$3D,$4C,$E3,$FE,$BD,$7F,$55,$E4,$7D
	.dc.b	$6E,$3B,$7F,$4D,$B5,$E4,$6D,$F8,$DB,$3F,$C8,$FB,$7F,$D2,$31,$8A
	.dc.b	$46,$27,$C6,$7F,$29,$F0,$9D,$DF,$90,$AF,$83,$F0,$9F,$BD,$E7,$FC
	.dc.b	$4B,$6F,$F0,$CD,$9D,$C7,$45,$6E,$AF,$E6,$D3,$98,$C7,$A9,$FC,$16
	.dc.b	$BF,$83,$36,$75,$5F,$E9,$C6,$CE,$9E,$CF,$F8,$AB,$B7,$F6,$5E,$7F
	.dc.b	$F0,$38,$FC,$0C,$FF,$33,$E6,$7B,$BA,$B9,$3F,$93,$BE,$17,$FE,$0F
	.dc.b	$9F,$E0,$31,$88,$FF,$17,$47,$6F,$27,$8C,$76,$3F,$A9,$4F,$EC,$38
	.dc.b	$7F,$2B,$F5,$6F,$AA,$7D,$5F,$F7,$88,$C3,$FC,$97,$CD,$C4,$EB,$3F
	.dc.b	$6F,$D5,$BC,$6F,$E2,$6D,$8F,$D5,$37,$3F,$F2,$93,$5F,$EA,$5F,$1E
	.dc.b	$FD,$CF,$FD,$E3,$5F,$56,$31,$ED,$7F,$61,$F4,$BF,$A1,$4E,$6B,$77
	.dc.b	$ED,$EA,$7D,$43,$EA,$FF,$90,$7D,$5F,$F5,$AA,$D4,$E7,$0F,$AB,$39
	.dc.b	$E2,$FD,$93,$BA,$FD,$EB,$5F,$ED,$0D,$7F,$BE,$36,$1F,$72,$EC,$FF
	.dc.b	$10,$C6,$1F,$1F,$A5,$89,$D6,$1F,$86,$FA,$C7,$BD,$FC,$DD,$B1,$F8
	.dc.b	$EF,$8A,$C4,$C8,$FE,$2D,$3A,$CF,$FC,$FC,$D7,$F8,$AB,$5F,$F0,$03
	.dc.b	$BC,$BF,$8A,$ED,$FE,$2B,$F5,$3A,$57,$CB,$E1,$FE,$CF,$87,$97,$E0
	.dc.b	$3C,$87,$F8,$4F,$81,$F0,$58,$9D,$1E,$49,$FE,$AB,$9F,$FB,$B2,$30
	.dc.b	$4C,$3B,$79,$BC,$5B,$1F,$A8,$FA,$87,$D6,$7F,$59,$6C,$7E,$B3,$EB
	.dc.b	$3F,$8E,$D7,$E7,$44,$E9,$C3,$5F,$E6,$ED,$7F,$C4,$4D,$7F,$C6,$CD
	.dc.b	$79,$D7,$F9,$6B,$E8,$DF,$12,$F9,$AF,$24,$E6,$31,$8F,$A5,$F4,$1F
	.dc.b	$DC,$71,$7D,$66,$27,$47,$F4,$FC,$BA,$B6,$BF,$BE,$C6,$32,$FF,$5D
	.dc.b	$37,$8F,$FA,$6E,$7E,$4F,$BB,$E9,$5F,$BF,$8C,$61,$F2,$FE,$03,$E5
	.dc.b	$1F,$E0,$3F,$AD,$FF,$2B,$D7,$C8,$8C,$CF,$8D,$F0,$91,$C1,$FC,$A6
	.dc.b	$9F,$AF,$BE,$B3,$8D,$F8,$63,$15,$F6,$FB,$3E,$D1,$E1,$FA,$6B,$19
	.dc.b	$3E,$14,$7C,$02,$3F,$ED,$9F,$47,$F9,$B6,$1F,$E6,$3A,$FB,$DA,$FF
	.dc.b	$8B,$A3,$1C,$5F,$E3,$B6,$DF,$F3,$66,$E5,$FC,$1F,$EA,$DF,$21,$FE
	.dc.b	$6D,$6D,$7F,$A2,$FE,$BB,$13,$23,$FA,$47,$72,$B7,$7F,$CB,$74,$FF
	.dc.b	$42,$61,$B8,$FE,$61,$F1,$6F,$9D,$75,$7F,$A6,$9B,$97,$F4,$FF,$9C
	.dc.b	$FF,$6E,$6F,$FE,$08,$BE,$17,$B2,$56,$FD,$1B,$8F,$6C,$79,$BF,$E1
	.dc.b	$4B,$E3,$5F,$F0,$05,$B1,$99,$DB,$FE,$83,$8F,$37,$B1,$E5,$3F,$D5
	.dc.b	$FF,$25,$FD,$5F,$2F,$25,$FF,$71,$6C,$EF,$0F,$F5,$3D,$6D,$FF,$50
	.dc.b	$5F,$47,$FB,$1F,$D7,$3E,$85,$9F,$F8,$BE,$30,$D8,$CE,$7F,$E2,$E1
	.dc.b	$3A,$71,$83,$E5,$9D,$3F,$E3,$98,$C3,$63,$33,$C1,$AF,$F8,$EA,$9F
	.dc.b	$3C,$CF,$C8,$D9,$F1,$4E,$A7,$27,$CD,$F9,$DF,$F2,$1E,$DF,$65,$BA
	.dc.b	$FC,$26,$DF,$5F,$A7,$7C,$CF,$FA,$CE,$F8,$4D,$BE,$0D,$3C,$97,$CB
	.dc.b	$06,$1C,$A7,$FD,$AF,$3F,$EE,$77,$C9,$EC,$7F,$96,$E9,$39,$9F,$C3
	.dc.b	$35,$FC,$13,$9C,$3D,$55,$E1,$AE,$2D,$BF,$F2,$DD,$F8,$23,$57,$C7
	.dc.b	$0F,$81,$1B,$F8,$5F,$57,$4F,$EA,$7D,$DB,$D4,$69,$AF,$AE,$FE,$5D
	.dc.b	$97,$EC,$27,$4D,$9D,$BF,$B4,$9D,$BF,$BD,$F8,$6F,$91,$D9,$FE,$3D
	.dc.b	$F6,$2F,$C1,$1F,$0C,$CA,$FC,$76,$5F,$BD,$8F,$F4,$13,$63,$39,$27
	.dc.b	$FD,$0D,$CC,$FA,$DC,$57,$E5,$E3,$13,$A9,$CC,$FF,$A2,$27,$2E,$E3
	.dc.b	$3C,$7F,$CF,$1B,$C0,$9F,$CE,$C6,$1C,$7D,$17,$C0,$AE,$1F,$33,$EE
	.dc.b	$24,$FF,$37,$A4,$7E,$2F,$0F,$43,$E5,$24,$F8,$DA,$FF,$4B,$9E,$2F
	.dc.b	$38,$FC,$A2,$B7,$D6,$F5,$EA,$FF,$58,$CF,$F2,$8F,$C0,$FF,$54,$FF
	.dc.b	$8D,$DB,$97,$F5,$A9,$DC,$FF,$59,$6B,$F0,$4F,$DE,$B8,$7E,$26,$5F
	.dc.b	$C0,$7E,$3F,$E9,$CE,$67,$71,$E3,$23,$FE,$74,$FA,$DF,$0E,$BC,$AB
	.dc.b	$F8,$54,$FF,$9F,$3E,$53,$F0,$E7,$AE,$24,$7D,$D7,$6F,$DB,$6B,$E7
	.dc.b	$C3,$F8,$A4,$79,$0E,$31,$ED,$7E,$4D,$BE,$7F,$F4,$BD,$7F,$C9,$6B
	.dc.b	$E4,$7B,$1F,$A8,$D7,$FC,$5F,$1F,$D4,$A7,$4F,$F9,$85,$AF,$C7,$E4
	.dc.b	$FE,$0B,$E1,$7E,$6A,$73,$3F,$EB,$16,$BF,$EB,$07,$9E,$FF,$E4,$46
	.dc.b	$ED,$BA,$FB,$FF,$59,$FC,$6B,$0F,$E2,$3F,$51,$FA,$CF,$8B,$99,$79
	.dc.b	$63,$D6,$78,$48,$FE,$BB,$14,$62,$3D,$4F,$07,$FD,$53,$7F,$D0,$9F
	.dc.b	$F4,$D6,$4F,$A2,$7C,$97,$F2,$58,$C3,$63,$33,$FE,$B2,$FA,$CF,$F4
	.dc.b	$23,$7E,$1F,$FD,$FC,$A3,$2F,$F4,$F6,$BF,$EF,$E6,$BF,$93,$FA,$CF
	.dc.b	$E1,$EF,$E5,$43,$87,$5E,$F7,$68,$C4,$7F,$B4,$3C,$07,$B1,$FC,$50
	.dc.b	$91,$FE,$A1,$FA,$B7,$D6,$7E,$BB,$7F,$5C,$FF,$63,$6B,$E1,$6B,$FA
	.dc.b	$76,$BF,$C2,$5F,$F0,$8B,$FE,$D4,$E3,$FE,$72,$FF,$9A,$5B,$BB,$93
	.dc.b	$97,$A5,$FB,$EF,$4B,$FD,$CD,$7F,$ED,$C5,$F3,$99,$35,$AC,$9F,$EF
	.dc.b	$07,$F8,$85,$F9,$F3,$CA,$3F,$BE,$35,$F9,$4E,$70,$F1,$47,$F2,$18
	.dc.b	$C6,$1F,$C0,$DB,$19,$F8,$AF,$E7,$BF,$4B,$23,$FD,$EA,$FF,$42,$FA
	.dc.b	$C7,$CD,$66,$7F,$C7,$19,$2B,$19,$78,$3C,$B6,$1C,$2F,$F6,$1F,$41
	.dc.b	$FD,$2B,$DF,$91,$82,$1F,$B3,$8F,$DD,$9F,$D4,$BE,$6B,$F7,$D1,$FE
	.dc.b	$4A,$F5,$EB,$E2,$F7,$3F,$AA,$9F,$D9,$A3,$39,$7F,$AD,$9D,$CF,$EF
	.dc.b	$5A,$FF,$45,$F0,$7F,$C3,$3E,$27,$F5,$89,$CC,$A7,$2F,$86,$7B,$1E
	.dc.b	$5F,$2F,$FB,$2F,$BF,$27,$FA,$86,$1F,$0F,$83,$FD,$67,$1F,$E6,$2F
	.dc.b	$77,$F2,$73,$9A,$E1,$3F,$31,$FE,$FF,$3A,$E5,$F3,$67,$FE,$DE,$AE
	.dc.b	$B1,$FD,$77,$85,$8B,$E9,$7F,$BB,$9F,$0F,$E3,$A5,$FC,$96,$BF,$E7
	.dc.b	$1E,$07,$D6,$7F,$54,$F6,$9F,$2D,$FD,$5B,$F5,$97,$97,$4D,$44,$C3
	.dc.b	$F2,$F5,$C1,$B1,$FC,$1A,$31,$1F,$E5,$36,$F0,$FE,$5A,$3F,$C5,$B3
	.dc.b	$CD,$FF,$6A,$7D,$53,$E4,$FE,$E7,$9F,$FE,$14,$FD,$BB,$86,$1F,$8F
	.dc.b	$6F,$17,$C6,$9D,$E1,$F8,$76,$BE,$FD,$FF,$B8,$FD,$6F,$E6,$63,$E5
	.dc.b	$9F,$E0,$77,$0F,$FB,$95,$B8,$79,$DF,$F6,$DC,$70,$0E,$AF,$FB,$83
	.dc.b	$63,$A7,$77,$83,$E3,$BF,$C0,$3F,$5A,$FA,$9F,$F1,$7F,$CB,$FF,$2A
	.dc.b	$CB,$F5,$4D,$CB,$FB,$D5,$6F,$1F,$03,$5F,$F3,$F3,$5F,$E5,$8D,$78
	.dc.b	$DF,$F3,$88,$E1,$7F,$45,$B5,$AA,$F6,$6F,$F8,$07,$E2,$F4,$7E,$3B
	.dc.b	$2F,$F2,$19,$23,$FD,$7B,$F5,$9F,$11,$FE,$0B,$19,$69,$A7,$FC,$97
	.dc.b	$F1,$92,$3A,$F5,$7F,$B3,$1E,$7F,$09,$97,$F8,$7C,$DD,$3F,$8E,$F8
	.dc.b	$2C,$7F,$CB,$B4,$9C,$FE,$1D,$F8,$98,$FF,$2A,$B5,$FE,$A1,$48,$C4
	.dc.b	$7F,$80,$5A,$FF,$05,$78,$4F,$4E,$6F,$F3,$9C,$63,$6F,$E1,$58,$7F
	.dc.b	$99,$6F,$E3,$FD,$6F,$CC,$7F,$85,$46,$5F,$34,$F0,$F8,$5F,$F3,$7D
	.dc.b	$6F,$73,$C7,$3F,$74,$FF,$A0,$B0,$E5,$9F,$F2,$FA,$D6,$5F,$47,$2F
	.dc.b	$56,$FF,$E7,$28,$C4,$7A,$E3,$07,$A8,$8F,$DA,$F1,$FF,$D3,$1F,$00
	.dc.b	$FB,$6F,$F5,$13,$67,$FD,$C8,$D7,$FD,$D0,$D7,$FD,$C8,$E7,$1F,$F7
	.dc.b	$A3,$76,$FF,$6E,$B6,$7F,$D9,$1F,$51,$FE,$36,$7F,$E0,$0C,$67,$E2
	.dc.b	$CD,$31,$3B,$BE,$B8,$7A,$09,$7F,$BC,$B5,$6F,$0F,$0F,$C8,$7B,$09
	.dc.b	$5E,$C6,$BE,$27,$E1,$9F,$EB,$B9,$FF,$64,$47,$DA,$3F,$D1,$DF,$1E
	.dc.b	$F9,$AF,$C9,$43,$EB,$FF,$85,$CF,$F3,$98,$C7,$6F,$F9,$D6,$31,$3A
	.dc.b	$AE,$2D,$D3,$FD,$35,$E6,$F2,$F0,$3F,$CF,$8D,$C7,$38,$F3,$C3,$C3
	.dc.b	$3C,$67,$B9,$E6,$AF,$F4,$26,$1C,$63,$23,$1F,$56,$3F,$EA,$99,$E4
	.dc.b	$6C,$98,$F7,$8D,$36,$BF,$CE,$8F,$B6,$FC,$ED,$8F,$39,$F2,$31,$F3
	.dc.b	$C7,$1B,$EB,$3E,$73,$A9,$BB,$AF,$9E,$BE,$3D,$75,$8F,$5E,$1F,$32
	.dc.b	$3A,$12,$3A,$1A,$C3,$E1,$C6,$63,$A5,$F9,$9F,$89,$3E,$A7,$8E,$3F
	.dc.b	$34,$F8,$B5,$E1,$C7,$D3,$3C,$8F,$1E,$3E,$E7,$8C,$F9,$EB,$93,$F2
	.dc.b	$5F,$A8,$47,$07,$CD,$D7,$FC,$43,$97,$D6,$A4,$62,$FF,$22,$DC,$BF
	.dc.b	$DC,$4F,$F8,$83,$EB,$6E,$3F,$E7,$1B,$46,$1F,$F3,$A3,$5F,$F6,$53
	.dc.b	$5F,$F7,$63,$F1,$57,$C9,$FF,$94,$27,$FB,$4C,$68,$FC,$19,$FE,$D3
	.dc.b	$8D,$65,$E1,$AE,$51,$89,$F0,$9C,$0F,$F7,$D3,$97,$C6,$7E,$60,$F0
	.dc.b	$EF,$F8,$47,$C1,$A7,$8A,$7F,$CC,$6F,$92,$3D,$E7,$53,$89,$4C,$E1
	.dc.b	$C6,$BD,$D1,$C2,$70,$F8,$4C,$61,$FE,$9B,$BE,$EF,$B6,$3C,$D1,$87
	.dc.b	$DB,$3C,$8F,$26,$1E,$99,$D4,$FB,$A7,$55,$CF,$0F,$09,$C2,$7D,$55
	.dc.b	$F1,$5F,$83,$3E,$3C,$3E,$24,$79,$8F,$A4,$E3,$0F,$03,$63,$CE,$DB
	.dc.b	$EF,$5E,$13,$A4,$FC,$7A,$E2,$FC,$36,$E1,$FE,$2D,$C5,$F3,$BF,$3B
	.dc.b	$3F,$F4,$AC,$63,$3F,$F4,$BC,$63,$7F,$F6,$AE,$9F,$CE,$6F,$83,$F9
	.dc.b	$A7,$EB,$E4,$31,$7F,$E4,$B6,$CF,$E7,$70,$EC,$DA,$FF,$4C,$3D,$73
	.dc.b	$F0,$BF,$A6,$F9,$E7,$C9,$E3,$0F,$D6,$DF,$07,$FA,$06,$1E,$CC,$34
	.dc.b	$DC,$3F,$A8,$61,$C9,$EC,$4A,$E3,$7C,$F3,$F8,$75,$B9,$CE,$1D,$2B
	.dc.b	$C1,$1E,$F3,$C7,$1E,$F6,$DF,$B6,$31,$D7,$FE,$1D,$F0,$FD,$4D,$9F
	.dc.b	$F1,$BF,$9D,$8E,$AF,$F2,$94,$7F,$98,$A7,$58,$7F,$89,$A7,$2D,$BE
	.dc.b	$B1,$88,$F2,$77,$7F,$86,$3A,$BB,$FD,$7B,$D0,$F9,$53,$9C,$BB,$4E
	.dc.b	$5F,$F1,$3D,$70,$AD,$F7,$23,$13,$E5,$C5,$F1,$FE,$93,$39,$7F,$73
	.dc.b	$D3,$FC,$79,$19,$AF,$F2,$0B,$E8,$23,$FC,$87,$3F,$F5,$4E,$9E,$6F
	.dc.b	$17,$DF,$C5,$FE,$4E,$DB,$FD,$EF,$7F,$E5,$6A,$D1,$1F,$5F,$B7,$73
	.dc.b	$FF,$72,$F8,$DC,$CF,$87,$F5,$96,$BF,$CB,$F1,$67,$D1,$82,$B1,$8F
	.dc.b	$FA,$97,$EB,$5E,$CF,$E4,$ED,$18,$C7,$FD,$25,$DB,$FE,$C2,$7F,$D2
	.dc.b	$91,$99,$FF,$0B,$1C,$2B,$F1,$0D,$9F,$E8,$4D,$3F,$A2,$4F,$17,$F2
	.dc.b	$06,$6F,$FD,$FD,$93,$E7,$FC,$95,$DE,$73,$1E,$7F,$91,$FF,$64,$E1
	.dc.b	$FE,$A7,$6E,$1E,$4F,$AD,$78,$BB,$F5,$65,$ED,$F8,$13,$A8,$C1,$F9
	.dc.b	$7F,$07,$EA,$DF,$E0,$5B,$92,$7F,$DC,$DE,$C3,$EB,$BF,$47,$5A,$9F
	.dc.b	$E1,$5D,$1F,$E4,$6E,$4F,$CB,$FA,$AF,$C2,$37,$18,$7E,$97,$D7,$3E
	.dc.b	$6E,$98,$F3,$AE,$3F,$EC,$59,$D3,$F3,$9B,$B7,$F7,$16,$BF,$39,$AF
	.dc.b	$EF,$DD,$ED,$FE,$E7,$FD,$C7,$FB,$FD,$75,$C3,$F0,$91,$8B,$F8,$5D
	.dc.b	$3E,$E3,$F6,$3C,$0F,$86,$C6,$1F,$26,$7D,$FF,$4D,$8B,$FF,$2E,$77
	.dc.b	$B8,$13,$FA,$5F,$9B,$98,$7E,$EB,$E1,$AE,$1C,$3C,$15,$63,$7F,$D5
	.dc.b	$55,$BF,$9F,$CA,$E7,$FA,$67,$9B,$E4,$EC,$FD,$D8,$FE,$37,$3B,$8F
	.dc.b	$F0,$C7,$3F,$F9,$F2,$72,$48,$FF,$6B,$EA,$F8,$AE,$7A,$FB,$23,$13
	.dc.b	$F8,$9F,$89,$FC,$19,$F1,$BB,$2C,$64,$E7,$E4,$E2,$F9,$3F,$C1,$FE
	.dc.b	$C2,$4F,$FB,$6A,$69,$F0,$E2,$EF,$F9,$38,$71,$B8,$BF,$AD,$5F,$FB
	.dc.b	$8A,$FF,$7D,$E8,$FE,$9B,$0E,$18,$7F,$8B,$BB,$F3,$E1,$E9,$98,$7E
	.dc.b	$EA,$FF,$5B,$5F,$E1,$39,$FD,$FC,$62,$3F,$C6,$51,$FC,$2F,$2F,$CF
	.dc.b	$57,$F6,$BC,$3B,$DF,$86,$3F,$9E,$F7,$7F,$63,$FA,$F7,$D7,$3E,$BF
	.dc.b	$F8,$A7,$D7,$BE,$1B,$53,$9C,$3A,$1F,$D6,$67,$4F,$F4,$FF,$51,$E6
	.dc.b	$7F,$58,$6C,$7D,$E6,$EF,$FE,$F8,$6B,$FD,$43,$99,$F2,$3F,$B8,$3F
	.dc.b	$BC,$7D,$7B,$EB,$DF,$05,$9C,$7C,$78,$7F,$82,$7E,$BF,$FA,$74,$7F
	.dc.b	$22,$6C,$7F,$99,$5A,$FF,$22,$6B,$F1,$FE,$BD,$D9,$FD,$CF,$95,$18
	.dc.b	$BF,$F3,$74,$62,$75,$39,$AF,$E5,$BD,$FF,$EB,$1F,$1B,$FB,$BF,$C0
	.dc.b	$7E,$DF,$1F,$3E,$3C,$5F,$EE,$D1,$9B,$FE,$C1,$C0,$F8,$9F,$EF,$ED
	.dc.b	$AF,$7C,$FE,$EB,$E0,$B3,$52,$7E,$73,$F4,$48,$E0,$FD,$5F,$9F,$BA
	.dc.b	$FD,$53,$DB,$29,$3D,$BC,$9F,$F7,$13,$F7,$63,$35,$F6,$5F,$BC,$48
	.dc.b	$FB,$CF,$F8,$5B,$F2,$27,$D5,$F0,$D7,$F5,$1B,$27,$37,$F6,$BE,$91
	.dc.b	$F0,$EF,$9C,$FF,$1C,$46,$1F,$ED,$1F,$5E,$F9,$5C,$D6,$EF,$93,$75
	.dc.b	$FE,$BF,$87,$19,$FF,$53,$36,$33,$3B,$7F,$C6,$91,$9C,$FF,$7D,$E2
	.dc.b	$7F,$29,$F0,$7B,$E7,$FD,$31,$F0,$FD,$7C,$CB,$EA,$E9,$EA,$7C,$71
	.dc.b	$9C,$BB,$BF,$E3,$FC,$BD,$ED,$CC,$D3,$E5,$7B,$FC,$74,$F9,$3B,$7F
	.dc.b	$8A,$31,$F9,$1F,$5C,$FA,$EF,$8E,$D8,$F1,$FC,$FE,$26,$47,$C2,$9D
	.dc.b	$69,$F1,$9A,$FC,$26,$BF,$E7,$5D,$11,$C3,$1F,$8B,$E1,$3E,$7B,$AD
	.dc.b	$FF,$35,$7C,$7C,$73,$1F,$8B,$F0,$B6,$98,$7F,$2B,$CB,$AD,$FC,$5C
	.dc.b	$BE,$F6,$1E,$78,$C1,$FE,$64,$E6,$CE,$1F,$B0,$6B,$FE,$6D,$E6,$FC
	.dc.b	$8C,$EC,$95,$FD,$33,$D8,$FF,$35,$DF,$2C,$BD,$71,$8F,$84,$95,$FE
	.dc.b	$6B,$FA,$B7,$C9,$E3,$3F,$F2,$B6,$3F,$E4,$EA,$FF,$08,$B6,$19,$AF
	.dc.b	$8D,$7F,$9D,$6B,$5F,$18,$FC,$2F,$FD,$4F,$87,$FA,$51,$BA,$7F,$AB
	.dc.b	$BE,$83,$F8,$33,$FE,$B0,$8C,$E1,$FE,$A9,$BE,$5E,$E7,$FA,$6A,$73
	.dc.b	$82,$58,$72,$BE,$15,$A9,$E3,$1F,$EA,$F8,$F1,$D7,$FA,$5E,$72,$FF
	.dc.b	$92,$9A,$E7,$CC,$FE,$C2,$FF,$AB,$1B,$8F,$A7,$D9,$FF,$56,$C6,$30
	.dc.b	$F4,$36,$BC,$CD,$BF,$3D,$70,$9F,$F3,$47,$D7,$FF,$80,$FD,$5F,$F2
	.dc.b	$FB,$63,$FC,$BA,$1C,$3F,$E6,$06,$BF,$EB,$1F,$25,$F5,$6E,$7F,$F3
	.dc.b	$4C,$61,$B1,$99,$EB,$39,$7B,$D7,$FA,$97,$1F,$1F,$B7,$F3,$E7,$2D
	.dc.b	$7C,$4D,$7E,$0F,$C8,$FF,$B0,$F6,$FF,$75,$EB,$FE,$D8,$9F,$EB,$B7
	.dc.b	$FD,$92,$B5,$B7,$FB,$9F,$C6,$7C,$73,$E3,$F3,$87,$CA,$BB,$87,$18
	.dc.b	$E7,$8F,$57,$FC,$8B,$7C,$5C,$F8,$1F,$E8,$AA,$FE,$E6,$DA,$FE,$1D
	.dc.b	$3F,$EF,$36,$BF,$83,$6E,$7E,$A9,$F0,$C6,$33,$F4,$7E,$57,$FE,$22
	.dc.b	$AD,$5F,$80,$87,$F9,$62,$35,$3F,$AC,$F7,$7D,$BD,$5F,$D3,$1B,$D2
	.dc.b	$7E,$0E,$EB,$FF,$40,$61,$FA,$F9,$D9,$1F,$F0,$4C,$EF,$DC,$E3,$E1
	.dc.b	$FF,$51,$9F,$F2,$8F,$D6,$E0,$CF,$99,$23,$24,$3F,$CA,$8F,$D2,$E8
	.dc.b	$FF,$22,$FD,$9F,$FA,$D1,$B3,$FD,$4B,$14,$E5,$B1,$98,$E5,$18,$8F
	.dc.b	$E4,$FB,$7F,$33,$8C,$6F,$FC,$CB,$E0,$7C,$1F,$6F,$F7,$63,$FC,$9F
	.dc.b	$F4,$9E,$BC,$3F,$D9,$2D,$8F,$E6,$13,$BC,$3F,$9F,$3F,$85,$AF,$EC
	.dc.b	$6D,$73,$CF,$EB,$8C,$47,$EA,$5E,$4D,$8F,$19,$F3,$3E,$43,$E2,$7D
	.dc.b	$BF,$F4,$79,$FF,$0D,$E2,$F8,$9F,$CC,$C2,$30,$FF,$81,$67,$83,$F0
	.dc.b	$8F,$DC,$38,$7F,$BC,$B5,$FE,$B0,$D7,$7E,$67,$D3,$8E,$79,$FF,$51
	.dc.b	$6E,$7C,$27,$F0,$7E,$87,$A3,$5F,$2F,$D8,$7E,$EF,$0F,$F3,$DB,$6F
	.dc.b	$BC,$EF,$1F,$F0,$7E,$3F,$E8,$56,$C7,$84,$8F,$B5,$E0,$FF,$A7,$7E
	.dc.b	$27,$F3,$FE,$3F,$F9,$92,$3F,$C8,$38,$7E,$03,$04,$48,$C3,$FD,$42
	.dc.b	$78,$3F,$1F,$AF,$FA,$0F,$67,$F2,$68,$D3,$63,$31,$FE,$B4,$9E,$2F
	.dc.b	$F6,$B7,$9E,$7F,$E3,$A6,$E1,$FE,$F0,$6C,$FF,$A9,$3A,$1F,$71,$EA
	.dc.b	$E8,$FB,$BF,$5F,$E8,$AF,$F2,$DB,$5F,$35,$FE,$57,$0E,$95,$E5,$8F
	.dc.b	$CD,$4F,$8E,$30,$47,$FD,$07,$97,$F5,$E8,$F0,$47,$FB,$2B,$EF,$BC
	.dc.b	$3E,$9C,$8F,$0B,$F9,$0F,$39,$F4,$34,$F8,$D3,$FE,$00,$6C,$6A,$3E
	.dc.b	$53,$5F,$57,$5F,$FB,$BF,$E0,$79,$47,$F9,$E2,$75,$C1,$FE,$28,$9F
	.dc.b	$CF,$FD,$6F,$B2,$BE,$23,$5F,$BA,$D4,$DF,$AE,$FE,$95,$1F,$2D,$AF
	.dc.b	$E1,$1A,$9C,$D3,$EF,$99,$FF,$48,$8C,$36,$33,$5F,$EA,$EF,$BC,$FE
	.dc.b	$2F,$EB,$FE,$53,$E7,$3F,$6F,$C4,$F3,$A9,$9B,$FA,$70,$9F,$B4,$48
	.dc.b	$FC,$73,$5F,$C0,$BC,$5F,$DF,$72,$3E,$C3,$FD,$01,$1F,$CE,$3A,$3A
	.dc.b	$37,$C1,$FE,$72,$7F,$3A,$B9,$19,$21,$FC,$27,$8D,$5C,$A3,$1F,$B8
	.dc.b	$7E,$ED,$FF,$7E,$AF,$A4,$FE,$9A,$D8,$FE,$92,$EE,$8D,$35,$1D,$27
	.dc.b	$8F,$9F,$FD,$6B,$EB,$1F,$3B,$D3,$EC,$A5,$FE,$6A,$72,$D9,$FD,$B4
	.dc.b	$62,$3F,$D9,$ED,$7F,$CC,$0D,$7F,$82,$F6,$A3,$87,$37,$FA,$C9,$BC
	.dc.b	$1A,$6D,$7C,$0F,$1F,$DF,$F2,$32,$3D,$EF,$C2,$C6,$46,$08,$7C,$A7
	.dc.b	$E3,$FD,$7F,$C8,$F8,$C9,$3F,$22,$32,$FC,$A8,$CB,$F4,$A3,$FA,$1E
	.dc.b	$1F,$E3,$68,$FF,$70,$F2,$FF,$BD,$9B,$DB,$AF,$6A,$47,$F1,$DA,$E0
	.dc.b	$FF,$91,$1A,$FF,$0F,$6B,$FC,$4D,$DE,$AD,$71,$5D,$2B,$FA,$80,$FA
	.dc.b	$F7,$C2,$7F,$84,$3E,$75,$C5,$79,$25,$FF,$83,$A7,$2D,$9C,$BF,$93
	.dc.b	$8F,$C3,$E5,$FE,$D0,$77,$19,$E1,$EC,$47,$F2,$3F,$53,$F9,$43,$0E
	.dc.b	$CC,$9F,$D1,$AB,$D8,$FF,$48,$FB,$CF,$E8,$EB,$EB,$7F,$8A,$ED,$FC
	.dc.b	$D5,$B1,$BF,$57,$FB,$AD,$7F,$84,$E3,$F9,$C3,$5F,$E7,$8D,$78,$1F
	.dc.b	$D3,$DE,$E4,$8F,$F1,$BF,$DF,$FE,$C7,$0F,$2D,$3E,$0F,$FA,$13,$FD
	.dc.b	$79,$AE,$7C,$BF,$D9,$F1,$48,$F7,$67,$FD,$36,$B6,$72,$ED,$FD,$AD
	.dc.b	$7D,$72,$57,$D5,$7F,$7B,$F6,$32,$30,$43,$F8,$2B,$FE,$52,$D7,$FA
	.dc.b	$04,$7F,$B3,$DA,$FE,$C1,$AF,$FB,$09,$BD,$9F,$3F,$CB,$E4,$49,$FF
	.dc.b	$29,$FB,$AF,$91,$FF,$2D,$B6,$33,$18,$33,$F4,$16,$32,$43,$FC,$D4
	.dc.b	$F9,$FE,$3F,$FB,$F6,$26,$92,$7F,$D5,$38,$FE,$0E,$38,$47,$F6,$C9
	.dc.b	$E0,$F9,$BC,$8F,$F4,$6F,$DB,$BE,$23,$FD,$10,$D9,$FF,$59,$F8,$A8
	.dc.b	$C4,$7F,$B2,$5A,$FF,$B3,$5C,$F2,$7E,$5A,$78,$57,$63,$E1,$FA,$B6
	.dc.b	$FF,$D4,$B7,$C3,$93,$FD,$C4,$D8,$71,$FF,$A0,$EB,$84,$7F,$92,$70
	.dc.b	$EA,$F2,$C3,$AB,$D3,$EB,$3D,$64,$7D,$78,$23,$11,$CE,$30,$F1,$3E
	.dc.b	$DF,$CA,$FF,$97,$3E,$29,$D8,$D7,$FC,$73,$F5,$DC,$2D,$4F,$23,$FC
	.dc.b	$4D,$39,$8C,$4F,$0C,$BF,$DB,$58,$73,$85,$6E,$31,$E1,$71,$C3,$FC
	.dc.b	$60,$DA,$FF,$24,$BB,$C6,$E1,$D1,$ED,$3F,$E4,$EC,$7C,$2F,$69,$FF
	.dc.b	$75,$5F,$99,$E1,$19,$38,$61,$E5,$C7,$C1,$19,$3D,$F5,$C5,$DE,$1E
	.dc.b	$07,$E1,$D7,$AA,$FB,$C7,$98,$F2,$57,$B2,$3C,$38,$FA,$E7,$B6,$3E
	.dc.b	$8C,$7C,$26,$49,$5D,$27,$77,$E3,$7C,$38,$F8,$EF,$EC,$B6,$75,$1D
	.dc.b	$6B,$E0,$DF,$87,$0F,$79,$23,$CA,$68,$F0,$47,$91,$C6,$1F,$08,$FF
	.dc.b	$08,$3E,$59,$E5,$19,$7B,$46,$49,$19,$7E,$9E,$BF,$E1,$2B,$F4,$CA
	.dc.b	$31,$97,$E2,$5B,$97,$FA,$31,$FF,$0C,$CF,$FA,$3B,$EB,$31,$8F,$F9
	.dc.b	$DE,$11,$87,$FC,$F4,$D7,$FC,$FC,$D7,$91,$EC,$3F,$D0,$5B,$7F,$54
	.dc.b	$9F,$D1,$46,$9B,$19,$CB,$C1,$5F,$81,$AF,$F5,$0C,$E5,$F0,$9D,$4C
	.dc.b	$F0,$ED,$F0,$E9,$D2,$3F,$1D,$18,$BF,$F4,$95,$7F,$66,$AF,$1C,$FB
	.dc.b	$8E,$87,$F1,$FA,$E8,$E7,$0E,$8E,$8E,$13,$83,$9E,$1E,$0B,$ED,$87
	.dc.b	$B2,$BF,$48,$75,$29,$C9,$F2,$E1,$E8,$8C,$1F,$29,$C5,$75,$6C,$7A
	.dc.b	$2B,$E3,$3F,$50,$C4,$7A,$A7,$84,$78,$1B,$7E,$88,$F2,$D7,$A1,$F9
	.dc.b	$F1,$ED,$BF,$56,$1F,$08,$E0,$57,$DD,$1B,$9E,$31,$EC,$7F,$84,$63
	.dc.b	$FD,$1F,$D7,$FF,$0B,$E7,$FD,$0F,$81,$18,$D3,$FD,$C3,$C4,$E9,$7C
	.dc.b	$B6,$72,$FC,$F6,$1C,$2F,$FA,$3E,$1F,$E0,$E9,$FE,$8E,$D7,$6D,$CB
	.dc.b	$C5,$3F,$D0,$3E,$73,$A4,$7F,$5B,$7B,$65,$E1,$F8,$2C,$E2,$9A,$4B
	.dc.b	$FE,$FB,$37,$1F,$E9,$13,$74,$FF,$1A,$E1,$CE,$FF,$AE,$CE,$EB,$FB
	.dc.b	$33,$E5,$C3,$B4,$F9,$4F,$EF,$3D,$0F,$92,$F8,$5D,$7F,$E3,$D7,$CF
	.dc.b	$BB,$E4,$F3,$5B,$BF,$EF,$D5,$F0,$EF,$84,$74,$8C,$4E,$F2,$FF,$6F
	.dc.b	$E0,$F0,$FA,$9D,$BE,$26,$E1,$FE,$1C,$8B,$5F,$EB,$1A,$56,$36,$FF
	.dc.b	$2A,$E5,$FE,$53,$F9,$6F,$29,$9C,$7F,$C5,$8E,$1F,$7B,$E4,$23,$FC
	.dc.b	$57,$7F,$F5,$9F,$CC,$FD,$6F,$2D,$F4,$3F,$63,$D6,$4E,$62,$5F,$F9
	.dc.b	$DD,$D4,$63,$5F,$F6,$2C,$7E,$35,$7C,$2B,$EA,$3A,$4E,$63,$17,$FE
	.dc.b	$54,$A1,$C2,$B4,$70,$8D,$46,$5B,$F5,$8F,$73,$C0,$D8,$E9,$A3,$CE
	.dc.b	$FD,$0C,$70,$AD,$9F,$9D,$B5,$F1,$7F,$9D,$35,$F2,$9E,$67,$0E,$1E
	.dc.b	$1C,$1F,$E7,$6F,$AA,$FF,$6C,$DF,$73,$D2,$E2,$BF,$DD,$7F,$57,$F4
	.dc.b	$73,$FA,$0D,$AF,$F5,$24,$59,$DC,$60,$FF,$54,$7A,$3F,$67,$D1,$FB
	.dc.b	$4D,$8C,$9F,$EA,$D8,$CB,$4F,$1C,$7D,$7C,$0F,$D9,$FD,$37,$D7,$3F
	.dc.b	$02,$72,$F0,$3F,$AF,$5A,$32,$FF,$61,$6B,$F2,$0D,$57,$CF,$8D,$E5
	.dc.b	$FE,$9C,$F1,$5F,$27,$E3,$AF,$C5,$46,$30,$FB,$7C,$8F,$B0,$FC,$B7
	.dc.b	$DD,$FE,$5D,$B3,$A6,$E7,$FE,$4B,$8F,$E4,$75,$C5,$E7,$5F,$9F,$FA
	.dc.b	$1F,$D0,$6F,$67,$BB,$86,$9E,$28,$C5,$F5,$1D,$4F,$80,$FD,$CD,$7E
	.dc.b	$76,$31,$D9,$FA,$DF,$9D,$FE,$23,$5A,$BE,$55,$AB,$FF,$47,$CE,$67
	.dc.b	$F2,$33,$97,$E6,$35,$CC,$73,$E0,$FE,$4F,$A7,$FB,$1E,$FC,$53,$97
	.dc.b	$E1,$F1,$FE,$FF,$83,$F9,$BF,$EE,$3F,$B2,$9F,$9F,$18,$C7,$51,$FA
	.dc.b	$C9,$D4,$7C,$1E,$96,$3F,$72,$BF,$A8,$D6,$B2,$FE,$72,$FF,$79,$6F
	.dc.b	$CA,$F8,$38,$FC,$2D,$AE,$CD,$C7,$B5,$A3,$E4,$FC,$0C,$AF,$E1,$43
	.dc.b	$F6,$57,$4E,$70,$F9,$BB,$AF,$F9,$D3,$6F,$E7,$C6,$08,$FF,$10,$D5
	.dc.b	$EC,$73,$D7,$FD,$E7,$18,$9F,$E3,$5F,$50,$FC,$05,$F3,$1F,$DB,$E3
	.dc.b	$0F,$CE,$F7,$71,$3A,$AF,$C3,$7D,$A3,$E0,$FF,$A7,$36,$3F,$DA,$2D
	.dc.b	$DB,$F0,$2D,$7F,$5D,$F5,$2F,$85,$FE,$38,$DA,$FD,$BB,$67,$FC,$AC
	.dc.b	$D7,$F8,$EE,$8E,$47,$CF,$FF,$66,$C7,$F8,$C6,$FF,$6A,$31,$C5,$32
	.dc.b	$F2,$BE,$82,$3F,$C1,$BF,$08,$FC,$86,$5F,$CF,$9B,$D6,$C7,$AD,$7F
	.dc.b	$2E,$9C,$C3,$E1,$3C,$47,$E5,$BE,$82,$D7,$C8,$8F,$8E,$47,$E3,$B7
	.dc.b	$0F,$DB,$0F,$85,$FF,$67,$CF,$13,$FA,$6B,$88,$FD,$9D,$BC,$3F,$84
	.dc.b	$7E,$9B,$DB,$B7,$F6,$23,$EB,$91,$FA,$ED,$AF,$0D,$BE,$1B,$8C,$61
	.dc.b	$FE,$D1,$F8,$0F,$E0,$E3,$CE,$DE,$0F,$D4,$FC,$96,$BA,$66,$5F,$CC
	.dc.b	$AB,$85,$FF,$2F,$6B,$FD,$61,$AF,$F6,$E7,$A4,$7F,$5C,$FA,$97,$C4
	.dc.b	$F4,$AF,$8F,$18,$9F,$EE,$34,$F9,$9F,$39,$FC,$22,$EC,$EA,$3F,$B6
	.dc.b	$4E,$E3,$CD,$18,$23,$FD,$F2,$BF,$CA,$57,$F0,$27,$FC,$15,$18,$C9
	.dc.b	$5F,$DC,$BF,$01,$71,$7F,$83,$DB,$1F,$C5,$7E,$07,$13,$23,$FA,$BC
	.dc.b	$EB,$1F,$F5,$B3,$5F,$EA,$CD,$7B,$4F,$8B,$AF,$94,$7C,$47,$C1,$A4
	.dc.b	$BF,$E4,$9E,$E5,$AF,$C7,$E1,$F2,$5E,$F5,$FE,$0F,$8C,$47,$82,$3F
	.dc.b	$CD,$DD,$7F,$B1,$F3,$FF,$99,$FD,$A7,$F3,$46,$E3,$FC,$C5,$B7,$FC
	.dc.b	$91,$BA,$FF,$91,$BF,$1F,$FF,$29,$37,$3F,$F2,$EF,$D0,$7F,$98,$27
	.dc.b	$38,$7F,$8C,$3B,$11,$8C,$7F,$AD,$B7,$3F,$F4,$55,$F1,$C3,$FC,$FF
	.dc.b	$F1,$5F,$DD,$6F,$FB,$17,$C2,$A3,$19,$FF,$9E,$3D,$FF,$C7,$D7,$FB
	.dc.b	$B6,$73,$A7,$FB,$AD,$B1,$9D,$7F,$CE,$7C,$FD,$99,$5F,$EE,$DC,$7B
	.dc.b	$47,$FA,$4E,$FC,$57,$FE,$36,$9C,$D6,$E7,$FD,$07,$3E,$66,$E3,$E5
	.dc.b	$FA,$2F,$5F,$C2,$CB,$FF,$2A,$61,$EB,$8F,$F4,$2C,$E7,$0F,$E1,$7F
	.dc.b	$51,$FD,$13,$D0,$FF,$11,$B6,$3F,$C4,$01,$C3,$FE,$23,$6B,$FE,$4C
	.dc.b	$CB,$F8,$27,$D2,$FF,$8B,$1B,$1C,$9A,$FF,$8A,$DA,$FF,$95,$25,$F2
	.dc.b	$2C,$BC,$6D,$9F,$1C,$EA,$72,$7C,$5F,$B7,$FF,$2E,$EB,$FA,$EA,$FA
	.dc.b	$D1,$8D,$7E,$B3,$63,$B7,$C9,$7F,$73,$7E,$3E,$49,$D6,$1F,$83,$6C
	.dc.b	$66,$7E,$6D,$71,$7F,$13,$F8,$07,$F2,$F7,$EF,$35,$FF,$25,$5A,$31
	.dc.b	$1F,$7D,$AF,$FA,$2A,$35,$E2,$78,$A7,$96,$5E,$D9,$CD,$78,$BC,$AD
	.dc.b	$5C,$8F,$C0,$DF,$0C,$FA,$63,$FD,$92,$7F,$C7,$9D,$1F,$E5,$86,$F1
	.dc.b	$FC,$5E,$9F,$D5,$B7,$BB,$53,$FE,$3E,$AB,$AF,$ED,$EB,$84,$63,$1F
	.dc.b	$DA,$5D,$B7,$53,$E6,$FF,$D8,$B1,$FE,$33,$FB,$6F,$A9,$3F,$DE,$5B
	.dc.b	$3A,$9F,$F1,$B3,$72,$FF,$19,$CF,$0F,$34,$F0,$F7,$3E,$F5,$BE,$73
	.dc.b	$FD,$C8,$FF,$A0,$1A,$FE,$79,$AF,$F9,$FD,$AF,$FA,$A1,$DD,$FE,$36
	.dc.b	$3A,$7C,$6F,$69,$CE,$9F,$EC,$2E,$9F,$7F,$B9,$9F,$8C,$FC,$84,$3F
	.dc.b	$01,$78,$98,$6E,$5F,$EC,$6C,$7F,$DE,$0D,$BF,$F4,$BB,$EB,$6C,$6A
	.dc.b	$3D,$51,$83,$C0,$F1,$3F,$DA,$D6,$FA,$9F,$CC,$69,$EB,$BF,$F6,$95
	.dc.b	$6B,$CE,$7C,$A7,$A1,$B1,$E5,$F8,$9F,$80,$78,$88,$FC,$D6,$BF,$E1
	.dc.b	$8A,$EA,$D8,$F8,$7F,$52,$F9,$3F,$AF,$F5,$3F,$BF,$AF,$F4,$A6,$9F
	.dc.b	$E0,$FF,$67,$C4,$6B,$C1,$93,$FE,$BA,$AE,$0F,$FA,$DD,$AF,$FB,$81
	.dc.b	$AF,$FB,$D9,$ED,$AF,$DB,$6F,$1E,$EB,$CB,$E6,$79,$39,$3F,$1D,$E4
	.dc.b	$CC,$7F,$DF,$78,$7F,$B4,$1B,$3F,$5A,$31,$1F,$CA,$9A,$F9,$63,$47
	.dc.b	$57,$89,$FE,$AF,$F9,$F7,$83,$F9,$AF,$A8,$FF,$69,$69,$FC,$77,$C2
	.dc.b	$FE,$15,$F0,$3F,$B7,$F1,$3F,$CF,$B8,$8F,$F2,$3D,$F0,$DF,$FE,$00
	.dc.b	$BF,$8A,$83,$48,$FE,$39,$E0,$F3,$6B,$C0,$FE,$0D,$97,$E4,$71,$FD
	.dc.b	$27,$A7,$FE,$1C,$9F,$14,$63,$C3,$FD,$9C,$E6,$7C,$BF,$4D,$ED,$3C
	.dc.b	$04,$7F,$DE,$B8,$23,$F4,$32,$F8,$1E,$15,$DB,$2F,$6D,$76,$9F,$E7
	.dc.b	$D8,$BF,$01,$7C,$56,$1B,$87,$11,$3B,$9F,$E1,$3F,$51,$FE,$9B,$9D
	.dc.b	$0F,$A9,$FD,$83,$F3,$CF,$A0,$E1,$C3,$5C,$F8,$3F,$BA,$9F,$C7,$C7
	.dc.b	$D7,$6A,$73,$6F,$A9,$F8,$53,$E2,$E0,$FE,$CE,$FA,$7C,$A7,$CB,$BE
	.dc.b	$8F,$F9,$5B,$67,$87,$A9,$FC,$92,$F8,$C7,$F7,$E6,$BE,$B0,$E1,$F7
	.dc.b	$35,$FC,$73,$CB,$1F,$E4,$8D,$E1,$DD,$7E,$1B,$E9,$FF,$9D,$F8,$7F
	.dc.b	$CD,$FE,$6E,$63,$E9,$F3,$FF,$82,$E5,$FD,$E7,$5F,$E8,$D7,$AA,$FF
	.dc.b	$74,$DF,$27,$E0,$7C,$87,$F1,$47,$FD,$DB,$EB,$7F,$43,$3F,$A5,$C6
	.dc.b	$70,$E8,$DD,$3F,$DF,$B3,$BA,$E7,$E5,$70,$F8,$AF,$F0,$37,$D4,$FE
	.dc.b	$0E,$7F,$0C,$D7,$F0,$F7,$FA,$FD,$18,$FE,$1E,$72,$7F,$60,$71,$97
	.dc.b	$F8,$56,$F9,$37,$A3,$F2,$75,$C2,$F9,$E3,$E3,$8F,$0C,$EA,$B8,$74
	.dc.b	$66,$7B,$F8,$0F,$AE,$7C,$7F,$F6,$5F,$B8,$95,$F9,$FA,$D4,$FF,$7F
	.dc.b	$AD,$CE,$5A,$F4,$6E,$1F,$2B,$A3,$FB,$87,$77,$F7,$1F,$84,$97,$F9
	.dc.b	$3F,$5F,$F2,$EF,$AD,$B7,$F7,$5F,$F7,$F4,$6E,$79,$BF,$9D,$C7,$F2
	.dc.b	$DE,$54,$8C,$10,$FC,$C4,$7F,$64,$EB,$F3,$C7,$E6,$5A,$FE,$62,$3F
	.dc.b	$40,$47,$F7,$7F,$53,$FE,$DB,$7F,$55,$F3,$FF,$85,$5B,$7F,$8D,$FB
	.dc.b	$3F,$97,$1F,$E1,$69,$D7,$21,$E1,$3B,$AF,$9D,$E3,$AF,$87,$4F,$F8
	.dc.b	$B7,$EF,$FE,$D7,$57,$E2,$76,$7C,$5F,$F8,$89,$B8,$FD,$7A,$DB,$E4
	.dc.b	$6E,$46,$9A,$D3,$C9,$C9,$FE,$22,$FA,$1F,$E5,$17,$F0,$3E,$87,$FC
	.dc.b	$4B,$C0,$75,$E4,$49,$FF,$40,$8A,$E2,$FF,$A1,$1A,$FF,$07,$6B,$FC
	.dc.b	$8D,$DF,$37,$F8,$9D,$BB,$66,$BF,$C6,$BF,$53,$FD,$9A,$BF,$C5,$9F
	.dc.b	$7F,$FE,$FA,$C3,$F9,$7A,$F9,$1F,$F1,$DB,$63,$3F,$19,$82,$38,$3F
	.dc.b	$9B,$FC,$DF,$F5,$5C,$BF,$D8,$DF,$7D,$D6,$7F,$5D,$C7,$72,$FE,$20
	.dc.b	$DA,$DE,$1F,$E6,$3B,$C5,$FE,$11,$BC,$67,$3C,$5F,$83,$F7,$7F,$CD
	.dc.b	$5E,$A7,$F0,$3B,$FE,$8D,$7F,$DF,$3C,$4F,$9A,$D8,$FF,$57,$B5,$FE
	.dc.b	$64,$DD,$3F,$C0,$15,$FC,$2B,$3F,$F7,$03,$6B,$F2,$FC,$5F,$39,$BC
	.dc.b	$5F,$E7,$B8,$C6,$1F,$6E,$3F,$CD,$FF,$1A,$F9,$3D,$44,$BF,$E0,$D7
	.dc.b	$CF,$55,$EE,$7F,$85,$62,$FA,$AF,$5B,$63,$D6,$D7,$D6,$D7,$E1,$B9
	.dc.b	$D3,$DB,$3F,$20,$F4,$70,$7F,$AC,$BE,$B5,$C7,$F9,$4F,$71,$18,$C7
	.dc.b	$FD,$D8,$D7,$FD,$5E,$D7,$FD,$D8,$E7,$0F,$F5,$53,$78,$3F,$D6,$0D
	.dc.b	$9F,$F6,$97,$D3,$BE,$3F,$C5,$8C,$9F,$F7,$6D,$6F,$15,$EE,$73,$5B
	.dc.b	$7F,$A1,$CF,$A2,$3F,$D9,$B3,$B9,$CF,$4F,$EA,$9F,$F7,$13,$68,$D2
	.dc.b	$70,$FF,$B5,$9F,$C0,$46,$A3,$AF,$4F,$FA,$8B,$E3,$9F,$57,$FE,$1E
	.dc.b	$C1,$F5,$5F,$AB,$53,$F5,$E3,$0D,$AF,$F7,$F5,$71,$9C,$BC,$FC,$2D
	.dc.b	$F1,$7F,$88,$DB,$87,$F9,$4D,$DF,$0D,$C3,$9B,$FE,$57,$C3,$75,$E4
	.dc.b	$9F,$F1,$66,$3E,$18,$FF,$2E,$3E,$AA,$E1,$3C,$A7,$81,$AC,$3F,$C6
	.dc.b	$97,$D0,$F0,$C6,$8F,$7E,$3E,$9A,$EF,$5E,$96,$CF,$9C,$F0,$CE,$DB
	.dc.b	$7E,$27,$C0,$53,$E5,$D7,$8A,$BC,$D7,$E2,$3C,$8F,$69,$F2,$61,$ED
	.dc.b	$6E,$1F,$3E,$75,$5C,$27,$C2,$6B,$0F,$33,$E0,$2D,$FE,$C7,$1F,$82
	.dc.b	$78,$6F,$99,$27,$9D,$EA,$3C,$93,$F5,$0C,$4F,$81,$C4,$68,$FA,$4F
	.dc.b	$CE,$CF,$FC,$69,$7E,$B6,$E3,$F9,$16,$E3,$FE,$97,$7F,$C7,$58,$7F
	.dc.b	$A6,$7E,$AF,$18,$7F,$94,$E1,$18,$7F,$CA,$CD,$7F,$CE,$4D,$79,$1F
	.dc.b	$ED,$AC,$7F,$CB,$9C,$2E,$2F,$8F,$3A,$BF,$D1,$3F,$BF,$8C,$BE,$03
	.dc.b	$91,$FE,$7F,$9E,$4E,$0F,$09,$4E,$45,$39,$69,$DA,$38,$BE,$08,$CD
	.dc.b	$F9,$AB,$F1,$51,$E5,$8C,$3C,$F0,$F4,$DF,$8B,$0E,$51,$C2,$3A,$D7
	.dc.b	$C9,$8C,$5F,$4B,$EE,$7A,$DC,$46,$A3,$F4,$C7,$CA,$C3,$D1,$3A,$8C
	.dc.b	$DF,$67,$E9,$9F,$05,$F2,$5F,$86,$31,$3F,$32,$FB,$5F,$A2,$79,$CF
	.dc.b	$58,$EF,$1E,$23,$CE,$7F,$B7,$6F,$C0,$FD,$0C,$36,$73,$2E,$1E,$1A
	.dc.b	$E6,$77,$2B,$F9,$1D,$FF,$C1,$D8,$7F,$B0,$F2,$FE,$5F,$3F,$7B,$3F
	.dc.b	$BF,$1F,$CF,$BD,$CC,$57,$1A,$B9,$7F,$38,$7F,$0A,$FF,$B1,$6B,$9C
	.dc.b	$FE,$6D,$B1,$FE,$AF,$6B,$BC,$1F,$1F,$D2,$7F,$A0,$B7,$89,$F0,$79
	.dc.b	$AD,$E1,$E9,$6E,$1F,$D3,$2F,$85,$6D,$95,$B7,$2D,$24,$F9,$6B,$9B
	.dc.b	$E5,$9F,$6B,$F2,$5A,$FF,$9A,$39,$DD,$DF,$13,$B5,$18,$C3,$FC,$1B
	.dc.b	$F2,$B9,$8C,$5F,$FA,$69,$AF,$F8,$1A,$FA,$DF,$26,$CE,$A4,$7F,$88
	.dc.b	$BD,$7F,$07,$71,$FA,$BD,$DF,$E1,$AF,$7B,$31,$33,$FF,$0F,$BA,$EF
	.dc.b	$4E,$70,$F2,$7D,$2F,$9C,$C7,$2F,$F8,$D1,$C6,$1F,$E2,$07,$CA,$47
	.dc.b	$FC,$7B,$E4,$3E,$3B,$E2,$36,$3F,$D2,$3E,$46,$26,$47,$FB,$F7,$C7
	.dc.b	$FB,$78,$7F,$10,$E1,$FF,$7F,$57,$B2,$73,$5F,$0B,$27,$D5,$7F,$8A
	.dc.b	$23,$E2,$E1,$F9,$48,$C4,$FF,$B5,$8D,$FD,$37,$E6,$9E,$3E,$25,$9D
	.dc.b	$D7,$E7,$3D,$C7,$B1,$9A,$E1,$18,$C7,$B9,$E3,$FA,$77,$C5,$75,$6C
	.dc.b	$6B,$2F,$B3,$F5,$2F,$80,$D5,$CA,$FF,$49,$DF,$0B,$FE,$88,$D7,$E0
	.dc.b	$07,$0F,$86,$7C,$1A,$F3,$8F,$81,$28,$FA,$77,$2F,$FD,$87,$16,$BF
	.dc.b	$DD,$BC,$5E,$59,$FA,$9F,$25,$9D,$BF,$D1,$2F,$E3,$BD,$7F,$8F,$3F
	.dc.b	$94,$E3,$7D,$E7,$FB,$D2,$3F,$D9,$FC,$1F,$A5,$8C,$65,$FF,$00,$6F
	.dc.b	$8B,$FE,$D2,$8C,$E3,$FD,$43,$AB,$FD,$A1,$7C,$7D,$9F,$B1,$E4,$FE
	.dc.b	$0B,$0F,$E4,$DE,$0F,$CF,$F8,$CC,$B3,$4F,$F5,$ED,$7E,$F2,$BF,$AE
	.dc.b	$4F,$EE,$9F,$F3,$03,$63,$A4,$7F,$98,$F9,$B9,$3F,$89,$8F,$C0,$3F
	.dc.b	$EF,$FF,$C0,$3F,$5F,$EF,$3E,$D2,$C6,$1B,$8F,$E0,$75,$56,$E3,$0A
	.dc.b	$B8,$47,$FA,$43,$E0,$BF,$84,$1F,$BC,$F8,$5F,$B5,$19,$26,$2E,$0F
	.dc.b	$91,$F2,$FF,$E0,$19,$CC,$61,$E5,$C3,$FC,$EB,$E1,$BF,$80,$C6,$0F
	.dc.b	$CB,$38,$7D,$18,$7D,$68,$C6,$1D,$E7,$F8,$9E,$8A,$D7,$87,$E1,$AD
	.dc.b	$09,$D6,$1E,$19,$D6,$3F,$7E,$F8,$E1,$F3,$30,$FE,$37,$93,$EC,$F1
	.dc.b	$3F,$2E,$30,$FC,$A9,$DC,$E9,$F0,$E4,$AD,$C6,$36,$7D,$5F,$6F,$6F
	.dc.b	$ED,$B7,$1F,$AA,$DA,$E0,$DC,$FF,$B0,$E2,$F9,$1F,$E1,$8D,$BF,$E0
	.dc.b	$1D,$5F,$C1,$3A,$33,$52,$7F,$C8,$18,$FC,$58,$FC,$A5,$F1,$8F,$F2
	.dc.b	$23,$5F,$E0,$2E,$70,$FD,$37,$3F,$F2,$D9,$FE,$3D,$39,$BF,$AB,$39
	.dc.b	$AF,$E5,$9F,$48,$F7,$BF,$AC,$36,$BF,$8B,$7D,$4B,$EA,$5F,$45,$AB
	.dc.b	$FE,$51,$F5,$2F,$8B,$FD,$53,$70,$F5,$B7,$5F,$F2,$E3,$5F,$E7,$1C
	.dc.b	$1F,$EB,$2F,$AC,$FE,$D1,$F0,$07,$FC,$1B,$B3,$EE,$7F,$81,$E3,$FC
	.dc.b	$36,$B4,$FC,$CD,$FF,$93,$57,$F8,$4B,$EA,$DF,$43,$EC,$F3,$3F,$9F
	.dc.b	$46,$2F,$F9,$96,$88,$C3,$FE,$71,$6B,$FC,$B9,$AF,$F4,$F6,$BF,$EE
	.dc.b	$7F,$C2,$5F,$41,$FE,$24,$F1,$51,$8B,$EF,$7D,$B2,$FF,$35,$D6,$AF
	.dc.b	$FA,$97,$5F,$FB,$62,$FC,$AD,$C7,$F5,$3C,$0D,$5F,$59,$C3,$E1,$A6
	.dc.b	$1F,$8A,$C3,$86,$1F,$D1,$9A,$FF,$67,$6B,$FE,$0A,$7B,$67,$FD,$BF
	.dc.b	$EE,$BE,$6C,$FF,$67,$9C,$D7,$F3,$D9,$CD,$7F,$90,$27,$FC,$36,$DF
	.dc.b	$77,$F1,$7F,$4D,$EC,$65,$F9,$5F,$E6,$15,$FE,$1B,$C3,$C5,$18,$3F
	.dc.b	$7F,$DD,$C2,$3F,$73,$E1,$7E,$6D,$FF,$10,$FD,$5B,$E3,$7F,$BD,$37
	.dc.b	$1F,$BB,$F5,$BF,$98,$7E,$FC,$66,$BF,$C0,$4D,$7F,$B5,$B5,$FF,$06
	.dc.b	$35,$EF,$5F,$D1,$E3,$96,$9F,$EA,$4F,$A3,$F1,$7D,$5F,$8A,$BF,$CC
	.dc.b	$B1,$8C,$3C,$3F,$15,$E6,$3F,$C2,$19,$3E,$B7,$13,$A3,$F8,$1E,$4A
	.dc.b	$F0,$12,$32,$F6,$C3,$F4,$B9,$FC,$48,$C7,$8F,$F1,$3C,$19,$7F,$A0
	.dc.b	$FA,$D7,$D2,$E3,$0F,$87,$5E,$D2,$57,$89,$B1,$FC,$EB,$57,$DC,$BE
	.dc.b	$67,$97,$0F,$C3,$8C,$61,$F4,$3D,$4F,$AA,$7F,$8B,$F3,$F9,$31,$F3
	.dc.b	$A3,$2F,$C0,$7E,$71,$1F,$E4,$3E,$73,$DA,$FE,$87,$F4,$3F,$DC,$1B
	.dc.b	$3F,$DA,$9B,$A7,$F3,$D6,$E5,$FE,$4B,$FC,$AD,$F6,$BF,$E7,$76,$DF
	.dc.b	$F3,$6F,$AE,$7C,$76,$6A,$4F,$F3,$AA,$DE,$3F,$B4,$BD,$CE,$6B,$FD
	.dc.b	$35,$F4,$1F,$E1,$CF,$93,$37,$CC,$FF,$0C,$46,$3D,$AD,$92,$FF,$D6
	.dc.b	$93,$AA,$E4,$7F,$96,$B1,$FF,$57,$35,$FF,$41,$DF,$3A,$FF,$58,$4E
	.dc.b	$65,$2F,$D7,$DE,$1F,$FA,$BD,$B7,$E0,$F9,$3C,$4C,$8F,$F6,$34,$EA
	.dc.b	$FF,$BD,$DB,$1F,$F5,$AD,$6A,$FF,$CD,$9F,$63,$FD,$8F,$CC,$FE,$BA
	.dc.b	$DC,$3F,$CB,$21,$C3,$FE,$5C,$6B,$FE,$7D,$BF,$F5,$6F,$4F,$F6,$66
	.dc.b	$C7,$26,$BF,$E6,$46,$BF,$E8,$59,$72,$7C,$28,$F0,$C6,$32,$F2,$CE
	.dc.b	$67,$C4,$D7,$E6,$E7,$FE,$69,$7E,$1F,$C3,$7F,$A4,$79,$3F,$DC,$F3
	.dc.b	$99,$FF,$07,$FC,$87,$FA,$51,$D9,$0F,$93,$1B,$9F,$F2,$F4,$EE,$73
	.dc.b	$C3,$FE,$04,$C7,$EC,$36,$CD,$24,$FD,$67,$C9,$BF,$A1,$B1,$F8,$2F
	.dc.b	$9F,$FF,$44,$E3,$E8,$9C,$E1,$DA,$B9,$1F,$E9,$75,$5C,$2F,$FC,$FD
	.dc.b	$CA,$FA,$DE,$BE,$7A,$31,$87,$E9,$E7,$C3,$EE,$B5,$7D,$CF,$0B,$DD
	.dc.b	$FF,$5E,$6E,$BF,$B7,$C3,$8D,$F0,$AF,$F2,$07,$CB,$74,$F2,$FF,$77
	.dc.b	$1F,$E4,$3E,$AC,$4C,$8C,$10,$FE,$96,$F8,$7D,$7F,$F1,$8E,$BE,$96
	.dc.b	$E7,$FE,$0A,$AF,$CB,$CE,$5B,$19,$7D,$13,$CD,$FF,$5F,$73,$BE,$C7
	.dc.b	$F9,$D4,$7F,$09,$F7,$7D,$1F,$6F,$23,$04,$3F,$D7,$B8,$FC,$B9,$F1
	.dc.b	$92,$3C,$2F,$C9,$23,$E0,$AF,$E4,$CF,$F1,$7F,$37,$E3,$6F,$23,$04
	.dc.b	$3B,$E7,$FD,$4D,$F6,$CF,$C2,$23,$BE,$0F,$E5,$51,$F8,$CC,$DE,$8B
	.dc.b	$E7,$7F,$07,$F3,$7F,$AB,$BF,$F4,$9F,$CB,$7D,$73,$3D,$33,$2F,$9B
	.dc.b	$5F,$AF,$7F,$D9,$4D,$5B,$A3,$0D,$9E,$53,$FE,$26,$9F,$07,$42,$19
	.dc.b	$F9,$BD,$DF,$29,$F8,$9F,$88,$93,$FE,$76,$BF,$F7,$CB,$67,$F0,$B1
	.dc.b	$88,$FA,$AD,$7F,$CB,$6D,$7D,$31,$D9,$E4,$7F,$AF,$BC,$4F,$8B,$ED
	.dc.b	$7E,$67,$9A,$4F,$86,$32,$FC,$47,$C2,$47,$06,$C8,$7F,$91,$BE,$9B
	.dc.b	$FD,$24,$FF,$A9,$7E,$5F,$F4,$C4,$7D,$DE,$3F,$D6,$3D,$A4,$7E,$9B
	.dc.b	$FE,$94,$7F,$C2,$5E,$96,$DE,$27,$F9,$AB,$D4,$FF,$05,$BF,$EC,$8F
	.dc.b	$CD,$9E,$DF,$F1,$0F,$8C,$46,$27,$53,$98,$F3,$C7,$FB,$1B,$99,$F5
	.dc.b	$4F,$8E,$FF,$52,$C6,$1B,$19,$9F,$E9,$73,$97,$FD,$57,$F5,$EE,$4F
	.dc.b	$53,$63,$31,$83,$EE,$FC,$97,$F0,$BF,$06,$4F,$E0,$E3,$F3,$66,$9F
	.dc.b	$CB,$B5,$FC,$1B,$A7,$81,$FE,$FF,$FA,$D7,$CB,$7F,$22,$F5,$DF,$3D
	.dc.b	$FC,$67,$E0,$A5,$FE,$5F,$47,$6F,$FB,$B2,$F7,$39,$8C,$5F,$08,$FF
	.dc.b	$32,$DF,$58,$FE,$49,$F3,$92,$7F,$DC,$CF,$87,$E7,$31,$1F,$E6,$C8
	.dc.b	$B8,$F8,$23,$FD,$ED,$8A,$BF,$97,$79,$D2,$7F,$DE,$0F,$DD,$E1,$FA
	.dc.b	$51,$FE,$76,$8D,$34,$F4,$0E,$5F,$AC,$E4,$98,$78,$7C,$FF,$F7,$B5
	.dc.b	$75,$E4,$FE,$7D,$E3,$CB,$FF,$80,$0F,$CB,$F9,$79,$18,$21,$FE,$FD
	.dc.b	$7F,$A3,$7C,$14,$8F,$1E,$0F,$41,$3F,$57,$E0,$B5,$5F,$E9,$08,$B9
	.dc.b	$F9,$77,$FF,$70,$4F,$FB,$2B,$DB,$FE,$A2,$A7,$E5,$3F,$E2,$FE,$1F
	.dc.b	$EE,$2D,$38,$7A,$58,$AE,$31,$FC,$5C,$FE,$30,$E1,$C3,$5C,$C7,$ED
	.dc.b	$63,$C3,$5E,$06,$F2,$F0,$9E,$DE,$0F,$F7,$23,$D3,$7E,$79,$FE,$4D
	.dc.b	$AA,$73,$87,$D3,$6C,$FF,$2B,$F2,$9E,$1B,$EA,$BA,$6F,$32,$F9,$B8
	.dc.b	$70,$7F,$9C,$D6,$DF,$A3,$C1,$5F,$F5,$ED,$FF,$7B,$3C,$DF,$07,$23
	.dc.b	$71,$97,$AC,$7C,$CF,$B9,$71,$FC,$EE,$67,$D3,$53,$51,$2F,$FC,$A2
	.dc.b	$2B,$8B,$86,$BF,$E5,$66,$BF,$54,$E1,$1C,$DF,$94,$7F,$4F,$8E,$3B
	.dc.b	$FE,$0B,$D8,$7D,$07,$F8,$5F,$E7,$A5,$FC,$27,$F0,$BF,$54,$FA,$5A
	.dc.b	$6A,$7F,$DC,$33,$87,$F3,$3F,$45,$FD,$79,$73,$23,$0D,$F5,$3F,$01
	.dc.b	$F1,$72,$B8,$C7,$F8,$D2,$B7,$CB,$FE,$EB,$9F,$DB,$F3,$BF,$46,$FE
	.dc.b	$71,$1F,$E1,$46,$BF,$97,$8D,$79,$BF,$E2,$A9,$CC,$FE,$1D,$E3,$C1
	.dc.b	$FC,$36,$B7,$3D,$A7,$3D,$5F,$92,$F7,$FF,$C2,$53,$F3,$29,$EF,$BE
	.dc.b	$67,$FC,$87,$E1,$CA,$F3,$CE,$5B,$3F,$E3,$C8,$C4,$7F,$8B,$1A,$FA
	.dc.b	$5A,$F9,$DF,$51,$1F,$F2,$0E,$3E,$26,$C7,$F9,$82,$31,$1E,$76,$E5
	.dc.b	$FE,$7C,$F0,$FF,$CD,$BF,$B0,$7F,$AC,$2B,$FD,$43,$C0,$8C,$47,$E3
	.dc.b	$63,$1E,$A7,$FA,$06,$73,$3F,$C1,$67,$2F,$F9,$D6,$38,$DF,$F3,$99
	.dc.b	$CC,$FF,$A4,$BB,$BC,$37,$23,$83,$FE,$6C,$8C,$47,$FA,$F2,$F9,$EF
	.dc.b	$FE,$CC,$6C,$7E,$3B,$7F,$DC,$63,$DE,$5E,$77,$9C,$95,$EE,$8C,$B6
	.dc.b	$3E,$19,$1F,$E6,$74,$E6,$FE,$4F,$F0,$EF,$93,$FE,$59,$4F,$A8,$FD
	.dc.b	$33,$FE,$3D,$9F,$DF,$BF,$E3,$F8,$CB,$FE,$D5,$F0,$3F,$96,$46,$27
	.dc.b	$A6,$7B,$EB,$FF,$67,$61,$DB,$DF,$D6,$1F,$CC,$DB,$8F,$F9,$2D,$E5
	.dc.b	$3B,$AE,$37,$CE,$BC,$97,$E0,$9F,$F6,$D3,$63,$D1,$5E,$CC,$79,$DF
	.dc.b	$19,$F4,$17,$1F,$6B,$F0,$70,$EE,$D3,$9C,$F9,$63,$BC,$53,$C6,$53
	.dc.b	$FD,$F1,$7E,$97,$89,$CC,$F0,$96,$BC,$37,$E6,$8C,$4F,$A6,$BB,$C6
	.dc.b	$F0,$F7,$3D,$B0,$F5,$56,$E3,$C6,$6C,$CC,$47,$FD,$15,$5D,$31,$FA
	.dc.b	$31,$83,$CE,$F8,$F0,$C9,$27,$E2,$BE,$78,$E2,$48,$F9,$2E,$5E,$C4
	.dc.b	$9E,$AF,$CE,$E1,$FF,$4E,$47,$C6,$6B,$ED,$0E,$32,$FC,$6C,$62,$BF
	.dc.b	$A4,$2F,$A9,$FF,$4A,$57,$28,$FF,$4B,$46,$27,$6D,$8F,$EA,$0D,$7F
	.dc.b	$C0,$CF,$DB,$5F,$17,$FE,$39,$9F,$D0,$C7,$43,$E4,$1F,$EA,$38,$D6
	.dc.b	$1D,$A3,$27,$2E,$1E,$8F,$9C,$CC,$F0,$7C,$67,$E1,$CE,$05,$38,$65
	.dc.b	$E2,$8F,$F2,$63,$E4,$8E,$97,$9A,$EF,$87,$AE,$FB,$57,$B8,$EC,$F2
	.dc.b	$C3,$D1,$87,$19,$F0,$57,$EA,$5F,$1C,$77,$8C,$57,$19,$F4,$CF,$13
	.dc.b	$4F,$6B,$E4,$F9,$AF,$FC,$01,$5E,$48,$F9,$98,$78,$4E,$37,$CE,$BC
	.dc.b	$A7,$0B,$E7,$87,$89,$B7,$DC,$F4,$9F,$39,$F8,$6D,$3E,$A9,$F3,$27
	.dc.b	$A3,$C2,$BC,$0F,$E4,$BF,$61,$7D,$0F,$F9,$EF,$5F,$E8,$91,$8C,$3F
	.dc.b	$CC,$3F,$BE,$7E,$61,$FF,$3A,$EA,$9D,$61,$FE,$1C,$8C,$61,$FE,$5C
	.dc.b	$8F,$CC,$CE,$9F,$EA,$8D,$73,$1B,$7C,$9E,$47,$8E,$F8,$B6,$3F,$C3
	.dc.b	$38,$35,$C6,$3E,$3A,$B8,$9A,$71,$F0,$97,$C7,$1F,$5D,$6D,$B5,$C1
	.dc.b	$F8,$38,$5C,$7D,$D5,$B7,$FB,$93,$5F,$6D,$7A,$F1,$F2,$5F,$F8,$13
	.dc.b	$E2,$7D,$5D,$FF,$E2,$77,$C9,$97,$F8,$5E,$7F,$C3,$D1,$AE,$C3,$F1
	.dc.b	$6B,$84,$F3,$26,$3F,$E1,$7F,$A9,$D7,$71,$A4,$9F,$E1,$B5,$FE,$26
	.dc.b	$8F,$0F,$CD,$F9,$2B,$85,$7E,$02,$3C,$5E,$E4,$8C,$1F,$E3,$37,$18
	.dc.b	$7F,$8F,$DF,$36,$5F,$EE,$B9,$FF,$11,$C7,$DD,$E0,$FF,$44,$7D,$1F
	.dc.b	$D3,$F1,$BE,$96,$1F,$E6,$49,$CD,$FF,$96,$38,$B4,$47,$FD,$0F,$76
	.dc.b	$BF,$CD,$67,$FB,$47,$CC,$E9,$E2,$7E,$1E,$31,$5F,$E6,$28,$CB,$4F
	.dc.b	$F6,$7C,$68,$F1,$63,$F4,$DE,$BF,$83,$7E,$85,$BC,$7D,$7E,$A9,$F2
	.dc.b	$3D,$EB,$5D,$3F,$EB,$C7,$85,$FF,$A0,$DA,$FF,$AE,$5A,$83,$86,$DF
	.dc.b	$AF,$F5,$4F,$C1,$E5,$E8,$AE,$EF,$FB,$0E,$33,$5F,$2F,$CB,$7C,$BF
	.dc.b	$CF,$E4,$FD,$14,$62,$FF,$D0,$B1,$96,$9F,$EB,$49,$C5,$7D,$5F,$B4
	.dc.b	$7C,$AF,$EA,$DB,$5F,$E8,$2F,$AB,$FF,$72,$BF,$EE,$88,$CD,$71,$AE
	.dc.b	$07,$E8,$AB,$71,$9A,$FA,$7C,$AF,$A9,$FB,$4D,$8F,$AF,$E2,$7F,$AE
	.dc.b	$9D,$CE,$8F,$E1,$55,$FE,$9E,$6D,$71,$8F,$ED,$AD,$7F,$D6,$CE,$78
	.dc.b	$3F,$DE,$7F,$4D,$E4,$7F,$DE,$B1,$9B,$FF,$78,$4E,$5F,$F7,$1F,$23
	.dc.b	$E1,$7F,$38,$7F,$0F,$F7,$DF,$57,$8A,$FF,$59,$CD,$CB,$F4,$15,$A9
	.dc.b	$C9,$FE,$E0,$3F,$4D,$F2,$1F,$C6,$7D,$C4,$60,$FF,$5E,$C5,$AF,$D5
	.dc.b	$4E,$A7,$27,$FB,$0B,$EC,$3F,$60,$7F,$2B,$F8,$4F,$DB,$46,$49,$5F
	.dc.b	$A2,$9D,$78,$1C,$E3,$13,$AD,$FC,$71,$FE,$CC,$D5,$7C,$2B,$B9,$FE
	.dc.b	$D2,$F3,$FF,$80,$1E,$BE,$1B,$19,$25,$FA,$A3,$2F,$B3,$D8,$FE,$11
	.dc.b	$E8,$7C,$36,$D7,$EA,$72,$F1,$F1,$B6,$56,$B0,$E5,$7F,$EE,$4C,$3F
	.dc.b	$99,$FA,$0F,$BC,$A7,$F4,$7F,$11,$F2,$B8,$BF,$F6,$FD,$6D,$FE,$56
	.dc.b	$DC,$3F,$E0,$07,$94,$8C,$47,$FB,$85,$AF,$EF,$E3,$58,$7F,$23,$F9
	.dc.b	$6E,$D3,$FC,$49,$FE,$35,$F5,$4F,$84,$FC,$EB,$71,$FF,$23,$D2,$30
	.dc.b	$FF,$92,$9A,$F8,$7A,$FF,$A5,$64,$69,$2B,$F4,$BC,$AF,$93,$FE,$6C
	.dc.b	$7F,$37,$F8,$EB,$3F,$3C,$FF,$23,$7D,$4B,$C4,$FE,$BB,$F8,$1F,$F1
	.dc.b	$B9,$D6,$7F,$E5,$0A,$D4,$63,$81,$5B,$AF,$F7,$D3,$5F,$E4,$82,$36
	.dc.b	$FF,$63,$8D,$3F,$2D,$AF,$E4,$FE,$13,$FA,$71,$E4,$6D,$78,$65,$F2
	.dc.b	$BE,$83,$FB,$AF,$D4,$7C,$59,$D4,$65,$FE,$D3,$3C,$1B,$1E,$32,$3F
	.dc.b	$03,$A5,$B7,$E1,$7A,$BF,$4F,$9F,$C0,$BF,$EC,$5C,$BF,$23,$E8,$A6
	.dc.b	$3C,$7A,$DF,$51,$8D,$BE,$2D,$7C,$12,$57,$91,$B1,$FD,$6B,$D6,$E1
	.dc.b	$9F,$E9,$34,$FE,$FB,$3F,$3B,$E4,$FE,$B9,$FE,$2D,$F9,$C7,$C4,$BE
	.dc.b	$57,$39,$FC,$7A,$FA,$04,$7E,$1B,$6F,$FB,$4F,$C4,$70,$6C,$FC,$FE
	.dc.b	$DF,$C2,$9D,$F0,$FE,$55,$3F,$77,$93,$F3,$3A,$7C,$67,$F3,$31,$F7
	.dc.b	$6B,$F0,$04,$BF,$A6,$FF,$82,$FA,$FA,$C7,$D1,$D7,$ED,$57,$E3,$3C
	.dc.b	$BF,$CA,$9E,$FF,$A6,$FF,$1D,$65,$DE,$33,$7F,$79,$FC,$71,$1C,$1D
	.dc.b	$7D,$6D,$DF,$F8,$21,$FF,$1E,$F9,$1F,$C4,$27,$F3,$CD,$95,$FF,$28
	.dc.b	$E1,$62,$E3,$E3,$E3,$4C,$7F,$C6,$F8,$F1,$C7,$F4,$7D,$3F,$CB,$AB
	.dc.b	$FC,$89,$39,$9F,$EA,$7F,$29,$FE,$2D,$FA,$1F,$F1,$73,$70,$F8,$8D
	.dc.b	$9F,$F0,$FB,$74,$FC,$AD,$FF,$82,$9B,$EF,$FF,$95,$BD,$0F,$6B,$30
	.dc.b	$F6,$B9,$69,$FE,$86,$FC,$A5,$F6,$7F,$E7,$66,$C7,$FA,$2B,$E0,$7F
	.dc.b	$80,$61,$FD,$2B,$DC,$56,$B4,$FF,$46,$36,$74,$DB,$FF,$72,$DF,$27
	.dc.b	$FD,$CF,$19,$AF,$F7,$7B,$CD,$FE,$17,$7D,$B0,$FF,$3B,$4E,$67,$8C
	.dc.b	$63,$E1,$FD,$07,$F9,$DB,$D4,$B8,$7B,$62,$D7,$F9,$67,$E0,$D5,$8C
	.dc.b	$FF,$B2,$CB,$E9,$7F,$77,$EC,$7F,$38,$6C,$7F,$C0,$06,$BF,$C8,$0E
	.dc.b	$1E,$1F,$FB,$E1,$C1,$FD,$37,$D0,$74,$FF,$42,$F3,$D1,$89,$DC,$62
	.dc.b	$75,$39,$9E,$D7,$FE,$3C,$E0,$F2,$7B,$46,$92,$7C,$DE,$EF,$F3,$4D
	.dc.b	$3D,$B6,$8C,$70,$7F,$38,$8C,$57,$FB,$2B,$9B,$2F,$F9,$7E,$32,$FF
	.dc.b	$9E,$9B,$7D,$2F,$7E,$5F,$FA,$E2,$F8,$DF,$CB,$F9,$8B,$81,$A7,$0E
	.dc.b	$27,$B1,$8A,$FF,$3F,$D6,$DB,$5B,$9D,$47,$8E,$2E,$5F,$E0,$CB,$E7
	.dc.b	$3F,$E8,$8F,$D6,$FE,$33,$75,$F2,$C6,$23,$F0,$9E,$C6,$26,$46,$CF
	.dc.b	$F7,$6C,$79,$F8,$38,$F8,$B2,$7F,$DE,$31,$97,$C1,$19,$7F,$21,$1F
	.dc.b	$13,$F1,$B7,$D0,$52,$4F,$F7,$2E,$47,$C6,$FF,$87,$E3,$FD,$43,$5C
	.dc.b	$1B,$5C,$1F,$E2,$FC,$9F,$E9,$5B,$E3,$E8,$F7,$F1,$7F,$C2,$34,$C3
	.dc.b	$F8,$D3,$FE,$14,$6B,$F9,$86,$BF,$9D,$73,$C1,$FA,$66,$E1,$FE,$00
	.dc.b	$6C,$FE,$7B,$2F,$F0,$C3,$E8,$BF,$13,$4F,$74,$69,$FF,$62,$CB,$EA
	.dc.b	$7F,$23,$97,$C6,$F9,$5B,$A1,$F9,$53,$2F,$E2,$B5,$F8,$67,$EC,$3B
	.dc.b	$3F,$BA,$BA,$21,$E0,$8E,$F1,$A8,$F1,$F9,$5E,$2F,$DA,$BF,$0D,$B7
	.dc.b	$F4,$36,$ED,$F0,$67,$EE,$12,$BE,$BB,$D3,$E8,$F1,$1D,$0C,$9F,$7F
	.dc.b	$04,$7E,$A3,$DA,$7C,$B7,$C2,$FC,$91,$F3,$14,$D4,$4C,$3F,$DC,$B5
	.dc.b	$C1,$B1,$FB,$2F,$65,$F7,$5B,$9F,$F7,$9D,$6D,$B5,$B9,$F1,$7C,$B7
	.dc.b	$D2,$E8,$7C,$8D,$33,$7E,$09,$AF,$47,$EC,$FE,$00,$FA,$AC,$F1,$FF
	.dc.b	$2E,$CD,$58,$E2,$FC,$C5,$EF,$0F,$D1,$7E,$49,$F8,$CE,$83,$4F,$C7
	.dc.b	$C3,$F0,$1F,$F4,$DB,$FE,$5A,$31,$E3,$7F,$12,$8C,$C7,$E1,$0C,$9F
	.dc.b	$AF,$D7,$F2,$5A,$7F,$7F,$6E,$5F,$E5,$CA,$E3,$3F,$D1,$9B,$1F,$AF
	.dc.b	$9D,$D7,$FB,$03,$EF,$7F,$DC,$8F,$F9,$B2,$51,$88,$FC,$AC,$63,$E2
	.dc.b	$FF,$3F,$93,$E0,$3D,$1E,$44,$BF,$66,$5F,$C7,$1F,$CE,$EA,$FC,$87
	.dc.b	$FA,$96,$BF,$E1,$76,$E7,$FC,$0D,$B5,$FD,$49,$BF,$72,$F1,$BF,$23
	.dc.b	$3F,$27,$EA,$DF,$1F,$FC,$21,$B1,$FE,$9A,$6B,$FD,$79,$AF,$F6,$37
	.dc.b	$38,$FF,$6C,$6E,$3F,$E8,$66,$C8,$7E,$03,$FA,$C7,$11,$A4,$D3,$11
	.dc.b	$BF,$3F,$FC,$F3,$F4,$3F,$D4,$BE,$63,$DF,$06,$FC,$1A,$7F,$AB,$B7
	.dc.b	$0F,$F4,$14,$F0,$37,$3C,$FC,$3E,$9F,$3F,$2B,$71,$97,$8C,$65,$FE
	.dc.b	$23,$87,$86,$7F,$B1,$FE,$D4,$F7,$6F,$D4,$CB,$FC,$C4,$E5,$B3,$97
	.dc.b	$F3,$10,$9E,$73,$27,$8C,$6F,$7F,$F0,$DD,$73,$9F,$D0,$7B,$5F,$EC
	.dc.b	$17,$F4,$CD,$73,$E0,$7F,$78,$FD,$AB,$FD,$B5,$1A,$8F,$EF,$73,$BB
	.dc.b	$FF,$13,$B7,$8B,$FD,$83,$7C,$2F,$E0,$EF,$FE,$4C,$7E,$6F,$E0,$DF
	.dc.b	$54,$FA,$DE,$FC,$8C,$90,$E2,$F8,$3E,$23,$FA,$CE,$26,$92,$7F,$DD
	.dc.b	$B1,$97,$DB,$1C,$5F,$E0,$9E,$2F,$ED,$9F,$F5,$BC,$65,$F8,$AF,$F8
	.dc.b	$62,$3F,$77,$C3,$F8,$9F,$8F,$FC,$67,$C0,$CB,$FC,$3C,$7E,$3C,$C3
	.dc.b	$F0,$23,$0F,$F1,$3F,$D8,$DF,$11,$FE,$24,$6C,$66,$3F,$C4,$FF,$84
	.dc.b	$3E,$67,$51,$27,$47,$F1,$D9,$D3,$FC,$7F,$F6,$4F,$F2,$5B,$FE,$50
	.dc.b	$0E,$1F,$F2,$9B,$5F,$F2,$C3,$5F,$F7,$CB,$BE,$3F,$F3,$1B,$78,$F3
	.dc.b	$E6,$FF,$8F,$3C,$0F,$C6,$56,$A3,$FC,$7F,$5B,$BF,$DE,$35,$FE,$1D
	.dc.b	$D3,$FE,$44,$F2,$7F,$A5,$C7,$F3,$76,$BF,$CD,$5B,$7F,$E7,$4F,$C7
	.dc.b	$3E,$C4,$63,$2F,$E1,$15,$FC,$B3,$CB,$E8,$7F,$94,$2B,$F9,$9F,$C6
	.dc.b	$BF,$05,$7C,$BF,$F9,$72,$30,$D8,$CD,$FF,$02,$FA,$7C,$4C,$8F,$E2
	.dc.b	$D9,$B2,$FE,$31,$87,$29,$E0,$7F,$51,$FB,$77,$C7,$FF,$98,$9B,$19
	.dc.b	$8F,$E2,$DF,$2B,$96,$46,$4F,$E2,$71,$A6,$90,$FD,$6F,$C0,$65,$FD
	.dc.b	$7C,$5C,$7C,$D0,$EA,$FE,$45,$5F,$07,$F0,$67,$DC,$52,$57,$BB,$4F
	.dc.b	$E6,$96,$FC,$3B,$FC,$7B,$AB,$0E,$37,$B1,$19,$C3,$9C,$7F,$22,$AF
	.dc.b	$EE,$FD,$9F,$D4,$A0,$D3,$F9,$BF,$C5,$BE,$7F,$FD,$13,$19,$9A,$7F
	.dc.b	$A1,$E2,$BF,$DF,$E3,$4F,$8F,$25,$73,$8D,$1C,$0F,$F0,$14,$F8,$2B
	.dc.b	$F0,$9F,$9C,$3C,$0A,$4A,$E9,$F5,$6F,$B9,$C4,$7F,$93,$B5,$5F,$18
	.dc.b	$F8,$91,$8B,$FF,$52,$DF,$19,$FF,$0B,$7C,$A7,$FB,$49,$FF,$62,$7C
	.dc.b	$AF,$F9,$DA,$31,$1F,$EC,$48,$C1,$FE,$AC,$8C,$F1,$FF,$B6,$1B,$C5
	.dc.b	$F6,$3F,$57,$FE,$61,$7F,$E1,$F9,$CB,$FD,$BD,$A9,$CF,$02,$78,$DF
	.dc.b	$5F,$04,$FF,$77,$ED,$FB,$38,$73,$F6,$FF,$67,$18,$8E,$91,$8E,$2F
	.dc.b	$DB,$E5,$FE,$0F,$BE,$51,$8C,$3A,$DB,$EC,$39,$5F,$F8,$7D,$B8,$7F
	.dc.b	$00,$73,$5C,$1B,$87,$F8,$2D,$DD,$DC,$3C,$37,$C5,$F0,$56,$67,$E2
	.dc.b	$5F,$39,$DC,$64,$F5,$97,$0F,$53,$E7,$8E,$6D,$3E,$21,$BB,$F0,$3D
	.dc.b	$AF,$95,$F7,$8D,$9C,$4E,$E5,$BE,$F8,$7B,$A3,$17,$D1,$DC,$64,$98
	.dc.b	$7A,$A3,$67,$83,$08,$E0,$93,$C1,$F3,$0F,$92,$FE,$6B,$87,$C0,$6E
	.dc.b	$3F,$61,$B7,$FE,$3B,$7F,$9D,$D7,$F8,$F7,$E3,$F1,$9F,$F3,$88,$C1
	.dc.b	$D5,$C3,$FE,$E0,$30,$6E,$38,$B4,$F8,$C3,$D4,$FE,$3F,$97,$E4,$63
	.dc.b	$AB,$63,$39,$F6,$AF,$B9,$5C,$77,$AF,$86,$F8,$46,$27,$77,$CA,$BC
	.dc.b	$55,$A7,$C6,$71,$BF,$74,$EE,$3F,$AB,$D7,$2B,$E9,$1E,$A9,$D3,$E8
	.dc.b	$AE,$33,$AB,$EC,$7B,$67,$B1,$FD,$8A,$B8,$B8,$3F,$CD,$8E,$2F,$C1
	.dc.b	$1E,$66,$CF,$36,$DF,$C5,$9F,$F5,$FB,$E2,$69,$F4,$70,$F8,$07,$CF
	.dc.b	$F3,$BF,$A7,$E7,$FD,$42,$7F,$CF,$91,$8C,$BF,$8A,$5D,$9F,$EF,$33
	.dc.b	$5D,$CF,$FA,$E5,$46,$F8,$3F,$8A,$56,$E7,$35,$BA,$F0,$47,$E3,$5A
	.dc.b	$F6,$8F,$F1,$23,$C0,$FC,$97,$27,$F3,$8F,$85,$F2,$36,$7F,$9F,$FD
	.dc.b	$B6,$39,$FF,$DD,$31,$FC,$F1,$B7,$FD,$05,$B9,$FF,$70,$CD,$CC,$C3
	.dc.b	$96,$5E,$FA,$FF,$58,$63,$FE,$D7,$6E,$3F,$E8,$8E,$A7,$D2,$EF,$92
	.dc.b	$5E,$DF,$89,$3A,$8C,$12,$BF,$DB,$D7,$D0,$F2,$4E,$E3,$35,$D1,$B5
	.dc.b	$FD,$A6,$75,$5F,$6C,$E5,$1F,$E2,$0F,$05,$F3,$3F,$21,$B1,$FD,$C7
	.dc.b	$B9,$F3,$F9,$A9,$3F,$D9,$2B,$78,$FF,$81,$6B,$2F,$FA,$23,$E5,$3F
	.dc.b	$01,$EB,$1B,$E8,$E1,$98,$7F,$AE,$1C,$B4,$FF,$0A,$57,$F8,$53,$D6
	.dc.b	$FC,$77,$29,$E3,$1F,$D7,$27,$FC,$97,$39,$78,$B6,$34,$7F,$90,$27
	.dc.b	$8E,$BF,$9F,$9D,$3F,$E3,$05,$F6,$DF,$8F,$8C,$36,$FF,$C5,$B9,$3E
	.dc.b	$B7,$F2,$33,$9C,$7F,$C5,$6D,$C3,$9F,$47,$C4,$AD,$BF,$A9,$F7,$5E
	.dc.b	$6D,$3F,$5D,$DE,$FD,$27,$3A,$7C,$4A,$B9,$FB,$03,$8D,$7F,$74,$D9
	.dc.b	$FE,$17,$56,$BF,$DF,$13,$57,$EC,$7F,$EE,$DA,$FA,$2D,$8F,$93,$83
	.dc.b	$F4,$27,$C6,$7F,$01,$AD,$65,$FC,$25,$B8,$7F,$9F,$9B,$87,$F3,$FB
	.dc.b	$E3,$3D,$BE,$0B,$91,$FC,$02,$7C,$6D,$AF,$14,$7F,$51,$6B,$FE,$8A
	.dc.b	$FB,$3F,$F7,$A5,$7E,$21,$B1,$E8,$B7,$BB,$FC,$5E,$73,$8F,$F1,$F6
	.dc.b	$DF,$3F,$63,$E8,$DF,$08,$E9,$1F,$9C,$F9,$2F,$CE,$FA,$36,$7F,$CB
	.dc.b	$1C,$DF,$E7,$46,$CE,$9A,$6A,$7F,$94,$7E,$2D,$FE,$9C,$8F,$CF,$B7
	.dc.b	$4F,$CB,$FE,$30,$FC,$61,$F4,$BC,$3D,$A9,$87,$07,$F9,$A5,$6E,$73
	.dc.b	$7F,$85,$C3,$A3,$FD,$E5,$AF,$FA,$C9,$A8,$38,$F2,$5A,$3E,$47,$F7
	.dc.b	$BF,$80,$7E,$F9,$B9,$FC,$26,$E1,$FB,$0B,$E3,$18,$E2,$FE,$9D,$F6
	.dc.b	$1F,$E7,$7B,$FD,$FB,$6F,$EC,$7C,$CF,$F4,$D8,$C5,$70,$F6,$30,$7D
	.dc.b	$69,$B7,$FE,$9B,$FC,$91,$F2,$9F,$EB,$F8,$C3,$F9,$0C,$7F,$81,$CE
	.dc.b	$A7,$FA,$E3,$C3,$AE,$D7,$1A,$DB,$C7,$1F,$F5,$AD,$71,$BE,$B7,$C1
	.dc.b	$E2,$D7,$27,$E1,$BD,$0F,$E4,$BE,$7F,$F2,$66,$C6,$4F,$C5,$C6,$5A
	.dc.b	$7E,$3B,$C9,$FE,$5B,$F3,$FF,$CB,$9B,$19,$3F,$27,$19,$69,$F3,$A3
	.dc.b	$E6,$1F,$DA,$3F,$0A,$76,$E2,$74,$4D,$7F,$D4,$D3,$A8,$C6,$9F,$6A
	.dc.b	$FF,$71,$C1,$FD,$22,$7F,$C8,$76,$8C,$3F,$E4,$66,$BF,$E4,$BF,$9F
	.dc.b	$FE,$63,$3F,$91,$84,$61,$FF,$7C,$7C,$7F,$F2,$D7,$3E,$5F,$F8,$47
	.dc.b	$F0,$9F,$F2,$14,$7F,$23,$6E,$5F,$E1,$BB,$E5,$29,$CE,$1F,$9F,$F9
	.dc.b	$77,$EC,$CF,$07,$FC,$3D,$18,$7F,$C4,$1C,$AF,$89,$F2,$4E,$6F,$FA
	.dc.b	$BE,$4F,$86,$F2,$79,$5F,$E2,$58,$C4,$FF,$5C,$CF,$F9,$BB,$67,$6F
	.dc.b	$2A,$BA,$7F,$58,$CB,$C1,$1E,$2C,$B9,$CF,$BE,$F1,$3F,$0C,$97,$FD
	.dc.b	$AB,$F0,$B7,$D2,$62,$FE,$35,$5D,$FF,$C8,$15,$AB,$F3,$4E,$5F,$E0
	.dc.b	$99,$3E,$25,$D9,$9C,$7F,$5F,$18,$AF,$77,$A9,$F5,$4F,$F2,$36,$2F
	.dc.b	$A0,$C4,$E8,$97,$FD,$6E,$75,$18,$CF,$FC,$A9,$BF,$F8,$2E,$3F,$8B
	.dc.b	$41,$F2,$FE,$46,$F8,$1E,$B8,$7C,$98,$C4,$FF,$83,$3F,$2D,$FE,$79
	.dc.b	$3F,$DA,$5B,$7C,$E3,$F8,$49,$FE,$41,$9E,$1E,$7F,$F9,$7E,$B7,$1F
	.dc.b	$97,$6C,$F0,$F8,$0D,$54,$AF,$39,$F2,$BF,$69,$F8,$D5,$E1,$9F,$C6
	.dc.b	$C6,$30,$FF,$39,$60,$FB,$9F,$01,$F9,$2E,$27,$66,$32,$F6,$4D,$CB
	.dc.b	$F1,$15,$A9,$C9,$23,$F1,$9F,$03,$F1,$1F,$85,$5F,$9E,$8C,$63,$E2
	.dc.b	$FA,$5F,$80,$7E,$89,$FF,$10,$5F,$CB,$8C,$B4,$F7,$DF,$F3,$29,$DB
	.dc.b	$F3,$E7,$6F,$FB,$23,$24,$7F,$80,$5B,$5F,$C2,$9B,$3B,$E2,$FA,$47
	.dc.b	$8F,$C9,$B5,$F7,$62,$CF,$EA,$70,$FE,$B1,$39,$3F,$CC,$DE,$87,$FA
	.dc.b	$67,$F2,$B7,$D5,$FF,$9D,$23,$0D,$9F,$F2,$9C,$E9,$E2,$62,$3F,$D4
	.dc.b	$5E,$E7,$E9,$63,$FC,$B2,$DC,$3F,$D0,$93,$AE,$99,$5A,$7F,$D0,$71
	.dc.b	$89,$FF,$1E,$57,$F8,$2D,$BE,$AF,$FA,$23,$EE,$3B,$B3,$0E,$EE,$5A
	.dc.b	$49,$E2,$FF,$A1,$67,$84,$76,$F3,$3D,$27,$FA,$53,$ED,$ED,$7A,$A2
	.dc.b	$CF,$FA,$76,$6E,$BF,$EC,$7F,$B1,$7C,$F7,$FB,$EE,$BF,$9D,$1F,$E0
	.dc.b	$B8,$CF,$BD,$FD,$11,$B1,$A0,$E1,$E3,$F1,$EF,$9A,$FD,$9C,$61,$AF
	.dc.b	$F4,$78,$D7,$4F,$F5,$06,$C7,$5A,$FF,$4F,$FC,$87,$87,$E2,$4D,$24
	.dc.b	$EF,$EC,$BF,$A9,$69,$ED,$6E,$FF,$D5,$70,$46,$29,$EC,$BE,$31,$F5
	.dc.b	$59,$9F,$6D,$5B,$E3,$1F,$E9,$E6,$BF,$EC,$99,$D7,$D9,$7F,$B8,$24
	.dc.b	$D2,$4F,$CB,$76,$F3,$7C,$98,$78,$FE,$53,$33,$FE,$87,$AE,$6E,$CF
	.dc.b	$F6,$44,$79,$FE,$CB,$E2,$C7,$FA,$C9,$AF,$FA,$3D,$BA,$7F,$B5,$3C
	.dc.b	$1E,$1B,$4A,$E0,$F8,$B5,$4D,$E2,$ED,$F6,$52,$FF,$DB,$11,$97,$C1
	.dc.b	$19,$7F,$DD,$91,$F2,$FE,$CB,$FB,$74,$7F,$83,$9A,$FE,$2D,$AF,$E4
	.dc.b	$9A,$FF,$83,$5D,$E3,$FC,$FE,$11,$8E,$1D,$79,$FF,$77,$95,$F2,$DF
	.dc.b	$C1,$AF,$FD,$AB,$7F,$ED,$D7,$F8,$FC,$BE,$A7,$94,$7F,$94,$9A,$F6
	.dc.b	$F4,$3D,$47,$F8,$99,$C1,$EB,$71,$5D,$63,$D9,$1C,$1F,$0B,$E5,$B6
	.dc.b	$9F,$17,$DE,$F4,$65,$FE,$00,$BF,$2B,$5F,$F6,$46,$09,$E1,$1F,$2D
	.dc.b	$F0,$91,$C1,$F0,$1D,$BF,$E1,$5E,$AF,$E0,$2F,$C4,$F5,$BE,$47,$D9
	.dc.b	$7E,$1B,$6F,$F0,$4B,$67,$FD,$1D,$87,$F8,$F9,$E0,$7D,$BA,$F8,$24
	.dc.b	$70,$68,$F4,$BF,$56,$DF,$65,$FA,$FF,$37,$FA,$5D,$7D,$4F,$B2,$FF
	.dc.b	$0B,$B7,$D6,$D7,$9D,$23,$FC,$AF,$5C,$1F,$E2,$8D,$7F,$8D,$B5,$FF
	.dc.b	$2B,$3B,$E4,$FC,$53,$75,$CD,$7F,$01,$F6,$DF,$05,$E5,$F3,$BF,$19
	.dc.b	$D1,$FE,$17,$CB,$F1,$2D,$7F,$DF,$98,$FF,$89,$DF,$ED,$67,$F8,$02
	.dc.b	$7F,$11,$19,$8F,$3C,$60,$8F,$08,$D6,$5E,$69,$FC,$E7,$D9,$7F,$B9
	.dc.b	$2B,$F3,$3F,$65,$F8,$FA,$FF,$11,$75,$23,$1A,$FF,$90,$9B,$5F,$EC
	.dc.b	$7F,$B2,$FE,$3E,$F9,$9A,$FF,$3D,$6B,$FD,$11,$AF,$F9,$FD,$AF,$FA
	.dc.b	$E3,$D9,$FC,$DB,$CF,$C5,$FE,$55,$3D,$31,$F2,$D6,$A9,$1F,$E3,$9C
	.dc.b	$1F,$4C,$F8,$7F,$E6,$D1,$86,$C6,$6B,$E0,$FA,$BD,$BC,$5F,$DA,$64
	.dc.b	$69,$27,$FA,$C6,$AF,$B2,$FE,$BF,$9F,$C8,$6E,$7F,$E6,$86,$D3,$E0
	.dc.b	$BF,$91,$F4,$FF,$45,$FB,$17,$D9,$7F,$47,$F1,$7F,$D6,$77,$F8,$06
	.dc.b	$BE,$EC,$7F,$D4,$D5,$C2,$3F,$0A,$D7,$FB,$63,$5F,$EE,$6D,$78,$D7
	.dc.b	$F1,$18,$E1,$97,$F2,$36,$D6,$AB,$C5,$E4,$7F,$4C,$FB,$27,$CB,$FF
	.dc.b	$4E,$F9,$0F,$6F,$9C,$6C,$EA,$FF,$C8,$72,$8C,$47,$AC,$FE,$D3,$3C
	.dc.b	$8E,$35,$D7,$C2,$ED,$F0,$D2,$B9,$D7,$F8,$3D,$E5,$87,$86,$B5,$18
	.dc.b	$95,$73,$8F,$D5,$FD,$8B,$E5,$BF,$BB,$2F,$B2,$A6,$22,$4F,$08,$CB
	.dc.b	$CA,$32,$FA,$C4,$F3,$78,$1B,$3F,$77,$EB,$FF,$9E,$9F,$DE,$FD,$97
	.dc.b	$F2,$37,$93,$78,$7F,$AB,$4E,$E7,$B3,$63,$3E,$17,$F8,$3E,$31,$C3
	.dc.b	$FE,$EB,$B7,$C7,$FD,$7F,$7B,$EC,$7C,$74,$CB,$E5,$C6,$5F,$9B,$19
	.dc.b	$7E,$54,$75,$FB,$2F,$F0,$E7,$C4,$3E,$D2,$99,$AF,$6C,$FF,$5D,$CC
	.dc.b	$D2,$47,$F9,$04,$FF,$24,$BC,$5F,$6B,$DB,$D4,$FC,$77,$CA,$FE,$3F
	.dc.b	$E9,$E5,$7E,$0A,$32,$FE,$12,$32,$FE,$06,$3F,$C6,$5F,$66,$FB,$3F
	.dc.b	$F1,$A7,$D9,$FF,$BF,$EB,$F9,$6B,$5F,$B5,$3B,$7F,$98,$B5,$CC,$F4
	.dc.b	$39,$FD,$37,$F8,$FF,$EC,$FF,$CB,$13,$FC,$E3,$DE,$7C,$F7,$F9,$0D
	.dc.b	$B1,$FC,$F2,$76,$FE,$25,$AF,$F9,$A9,$A8,$8E,$1B,$E9,$FF,$82,$E3
	.dc.b	$9F,$47,$F8,$75,$BC,$FC,$69,$18,$7A,$5F,$E8,$DF,$E9,$FE,$6F,$F5
	.dc.b	$3B,$FF,$13,$7C,$BF,$DF,$9F,$EA,$6D,$7F,$41,$3B,$C3,$B3,$63,$D0
	.dc.b	$D7,$F9,$AF,$A1,$FE,$65,$AF,$E7,$57,$FE,$F1,$FA,$67,$D9,$BE,$47
	.dc.b	$F8,$77,$BB,$31,$FF,$66,$D6,$A7,$FC,$E4,$D8,$CD,$71,$6C,$FF,$17
	.dc.b	$FA,$6F,$F3,$A7,$D9,$BE,$AB,$FC,$64,$DA,$FC,$A3,$5F,$F4,$B3,$5F
	.dc.b	$F4,$EB,$9C,$BF,$C9,$4D,$DB,$E7,$36,$7F,$8E,$FC,$9F,$F2,$72,$3F
	.dc.b	$C6,$9F,$E5,$5F,$0B,$88,$FE,$31,$16,$D7,$F8,$1F,$2F,$F9,$75,$FC
	.dc.b	$0F,$B3,$7C,$9F,$EA,$F7,$46,$32,$FF,$57,$46,$27,$53,$99,$FF,$5B
	.dc.b	$CE,$5F,$9B,$19,$C7,$EA,$37,$1F,$17,$D6,$7F,$80,$A9,$DD,$FE,$5B
	.dc.b	$69,$FE,$5D,$70,$47,$FD,$48,$FF,$64,$76,$7F,$46,$F8,$B7,$D9,$EA
	.dc.b	$21,$FE,$22,$9F,$E9,$F3,$A8,$C1,$A3,$F9,$94,$6A,$7F,$0B,$F5,$EF
	.dc.b	$07,$E3,$36,$BE,$2B,$5F,$86,$D7,$F5,$1A,$23,$5A,$7E,$5D,$B9,$FE
	.dc.b	$D6,$75,$7F,$EB,$2F,$95,$7D,$9F,$FB,$0F,$E5,$5F,$67,$FE,$B6,$FC
	.dc.b	$6B,$F5,$59,$FD,$ED,$3F,$D9,$71,$8C,$FF,$DB,$D5,$C4,$FF,$44,$5F
	.dc.b	$2A,$FF,$70,$46,$08,$FF,$92,$23,$4F,$F5,$56,$DF,$7E,$47,$DA,$3C
	.dc.b	$DF,$37,$83,$FE,$2C,$C1,$F6,$9F,$72,$DF,$AE,$3B,$7F,$D9,$ED,$AF
	.dc.b	$05,$F1,$BF,$BB,$9F,$FB,$0B,$3F,$E3,$5E,$16,$3E,$13,$F8,$A3,$70
	.dc.b	$CF,$3F,$FB,$F7,$1F,$14,$F2,$7A,$47,$FC,$01,$AF,$E3,$98,$F1,$BF
	.dc.b	$3C,$F1,$C3,$F9,$44,$F1,$6C,$71,$9A,$7F,$2D,$C7,$FC,$80,$E0,$3E
	.dc.b	$97,$A9,$BC,$3C,$6F,$1C,$3C,$75,$D6,$36,$72,$9D,$B6,$FD,$EF,$12
	.dc.b	$9E,$0A,$F8,$B5,$EB,$86,$3E,$B7,$AE,$1E,$86,$D7,$89,$F3,$5F,$AE
	.dc.b	$7D,$55,$F1,$C9,$7D,$5E,$86,$2B,$C8,$78,$23,$47,$8F,$DD,$FF,$4C
	.dc.b	$57,$C5,$6E,$3F,$89,$6E,$1F,$D6,$5F,$F4,$F5,$FF,$5B,$F9,$4C,$63
	.dc.b	$FD,$56,$30,$27,$6F,$FB,$C8,$C4,$6D,$C4,$6A,$7E,$4C,$3D,$CF,$D6
	.dc.b	$57,$F0,$E8,$EA,$D8,$CE,$5C,$EB,$8C,$62,$BC,$91,$8E,$3E,$B1,$E3
	.dc.b	$D7,$94,$7E,$1E,$3D,$31,$87,$F0,$27,$82,$FD,$D7,$D6,$3F,$BA,$5F
	.dc.b	$57,$E7,$1E,$03,$8C,$7F,$05,$AF,$54,$78,$6B,$9D,$F8,$63,$07,$F7
	.dc.b	$FA,$E1,$3E,$D8,$E7,$3B,$BF,$45,$6A,$7C,$17,$E9,$3B,$CF,$97,$0B
	.dc.b	$3E,$63,$FC,$29,$5E,$B7,$E0,$E1,$FD,$5F,$67,$81,$FC,$77,$4F,$E3
	.dc.b	$D5,$F0,$0F,$D5,$F0,$FF,$0C,$9D,$69,$F8,$3E,$2B,$5F,$E1,$96,$D7
	.dc.b	$F1,$8C,$73,$3F,$E2,$59,$AF,$F1,$BA,$FF,$27,$34,$FF,$28,$47,$89
	.dc.b	$ED,$1A,$8E,$B8,$FF,$BE,$23,$FD,$E3,$3C,$71,$FE,$51,$39,$8F,$F4
	.dc.b	$D7,$4F,$F4,$5F,$8D,$F3,$B6,$FF,$D4,$3D,$2F,$A2,$CF,$67,$1C,$BF
	.dc.b	$A5,$61,$E3,$C5,$8F,$17,$FA,$9C,$66,$31,$1F,$ED,$0E,$4F,$83,$47
	.dc.b	$E2,$F7,$3E,$D7,$31,$34,$FF,$25,$57,$D2,$38,$3F,$DC,$67,$57,$FE
	.dc.b	$9B,$9C,$B6,$BA,$CE,$A7,$94,$7A,$67,$FD,$E6,$D8,$E3,$1E,$7A,$FF
	.dc.b	$09,$F8,$AF,$9C,$F8,$ED,$8F,$F6,$3F,$85,$FE,$98,$7F,$C7,$F1,$9A
	.dc.b	$F9,$ED,$7F,$BA,$B5,$FE,$E4,$D7,$E5,$B7,$8F,$FD,$AD,$F3,$1F,$8A
	.dc.b	$F8,$13,$7D,$1C,$33,$0F,$F6,$7B,$96,$9F,$28,$FF,$02,$FB,$5F,$92
	.dc.b	$E0,$3C,$72,$FF,$62,$46,$27,$84,$7F,$81,$E7,$E4,$4E,$63,$FC,$6F
	.dc.b	$18,$3F,$01,$F6,$BF,$91,$AF,$F0,$73,$67,$FD,$D3,$F6,$BF,$90,$8C
	.dc.b	$59,$D6,$6F,$C8,$7F,$C9,$78,$7F,$9A,$A7,$31,$8C,$3F,$DB,$5F,$09
	.dc.b	$FD,$F6,$BC,$F1,$FE,$EA,$F1,$3F,$67,$DC,$FB,$6D,$4F,$09,$FD,$9D
	.dc.b	$E3,$93,$FC,$39,$7B,$EF,$CD,$72,$C3,$F8,$2E,$3D,$BE,$D9,$F0,$FF
	.dc.b	$BC,$6C,$7C,$8C,$1F,$6C,$F9,$3D,$5C,$AF,$F6,$65,$F0,$D7,$FA,$33
	.dc.b	$5F,$E9,$CD,$7F,$D4,$8E,$F4,$FF,$19,$7D,$F3,$F3,$E7,$DD,$75,$B5
	.dc.b	$6A,$73,$87,$C6,$FB,$47,$C6,$7F,$14,$6C,$7A,$75,$7C,$46,$6A,$4E
	.dc.b	$7E,$B2,$E5,$DA,$3F,$96,$DA,$7F,$31,$F3,$BF,$99,$F6,$6D,$7E,$33
	.dc.b	$A5,$DF,$FE,$84,$8C,$2A,$E1,$18,$6E,$3D,$CF,$F6,$77,$DB,$FF,$AC
	.dc.b	$67,$F0,$2D,$8D,$7D,$C3,$EE,$7F,$DB,$B5,$AF,$7F,$F9,$9D,$EE,$FF
	.dc.b	$BE,$35,$FF,$69,$35,$EF,$1F,$C2,$63,$B7,$17,$FB,$0F,$EF,$BF,$D6
	.dc.b	$D7,$D4,$FF,$60,$B8,$BF,$46,$9F,$03,$E7,$7F,$79,$F7,$5F,$BD,$6D
	.dc.b	$FC,$26,$D7,$0D,$91,$8C,$7F,$AA,$FD,$C7,$FA,$1E,$FF,$42,$DB,$F8
	.dc.b	$DD,$9F,$92,$FA,$5F,$EA,$33,$9B,$E3,$F0,$1F,$A6,$E5,$7D,$2E,$A3
	.dc.b	$0F,$F5,$CF,$BA,$7D,$0F,$EC,$6F,$FD,$05,$18,$F0,$75,$39,$D5,$AA
	.dc.b	$B5,$1F,$C3,$1E,$6F,$0A,$E2,$D3,$55,$CE,$B9,$3F,$AB,$8C,$9B,$8F
	.dc.b	$8D,$EB,$7F,$27,$F5,$3F,$94,$36,$7F,$1A,$D9,$DC,$7E,$42,$7F,$97
	.dc.b	$7C,$B7,$F2,$FF,$AA,$B3,$F9,$46,$D7,$D3,$6E,$7F,$E9,$5E,$34,$7C
	.dc.b	$EB,$97,$FC,$AB,$CC,$7D,$D3,$C2,$CE,$5F,$9B,$1C,$7F,$D2,$DF,$F2
	.dc.b	$43,$5F,$F2,$83,$5F,$9A,$70,$C3,$FA,$E4,$71,$CF,$FC,$35,$F8,$AB
	.dc.b	$97,$BC,$61,$FF,$09,$FC,$6F,$F3,$B8,$FF,$0F,$B7,$4F,$B2,$3E,$6F
	.dc.b	$F9,$D6,$49,$D3,$F7,$BE,$E3,$FC,$B5,$5F,$E5,$F8,$46,$1F,$F3,$13
	.dc.b	$5F,$F3,$53,$5F,$C3,$FD,$23,$EE,$FF,$81,$57,$F5,$28,$C7,$89,$FE
	.dc.b	$2E,$E6,$7C,$97,$8A,$73,$5F,$D0,$B2,$7C,$87,$F8,$C3,$37,$A7,$F0
	.dc.b	$23,$0C,$AF,$44,$E9,$B1,$FE,$07,$8C,$69,$FE,$81,$CB,$C3,$7E,$3C
	.dc.b	$BB,$E1,$FD,$4A,$F7,$B7,$FB,$9E,$F7,$39,$BF,$21,$FE,$44,$FB,$A7
	.dc.b	$DD,$E2,$74,$79,$27,$C4,$47,$07,$F7,$6A,$FF,$2C,$F0,$F1,$CF,$F9
	.dc.b	$34,$62,$7C,$9E,$D7,$C3,$3C,$1F,$5B,$63,$27,$C2,$8F,$80,$47,$07
	.dc.b	$7F,$BB,$FF,$5E,$BC,$2F,$BC,$61,$FC,$E7,$DD,$7F,$4A,$9F,$EE,$AD
	.dc.b	$8F,$29,$FE,$4F,$8C,$FA,$3F,$E6,$69,$D4,$FF,$AB,$DA,$FE,$81,$AF
	.dc.b	$F0,$E6,$A7,$B6,$DF,$47,$EE,$FF,$45,$7D,$8F,$90,$E2,$BC,$FE,$8F
	.dc.b	$84,$FC,$76,$9F,$A1,$AF,$4C,$65,$FC,$F6,$1F,$D7,$DB,$1F,$E4,$D8
	.dc.b	$C5,$7F,$B0,$5B,$87,$E0,$FE,$0F,$E3,$3F,$E8,$E8,$CD,$FF,$28,$C3
	.dc.b	$FC,$CD,$F1,$7F,$0F,$C7,$93,$F7,$ED,$EF,$E3,$2F,$9F,$5E,$12,$57
	.dc.b	$99,$B1,$F8,$4E,$DF,$F2,$5C,$FD,$49,$FD,$64,$62,$7E,$CF,$0F,$D3
	.dc.b	$3F,$61,$87,$F5,$2B,$FB,$71,$97,$E1,$BF,$04,$8F,$FB,$87,$CE,$7C
	.dc.b	$37,$F6,$88,$FE,$33,$F7,$6F,$BB,$FF,$41,$FA,$9F,$EA,$68,$C4,$FF
	.dc.b	$81,$EB,$86,$5F,$D3,$27,$2F,$F6,$C8,$CA,$8C,$E5,$FE,$BA,$F8,$CF
	.dc.b	$F1,$CC,$FF,$2F,$6C,$2F,$FA,$BE,$5E,$6C,$BF,$D6,$F5,$A9,$FF,$5E
	.dc.b	$7C,$0E,$0F,$F4,$CC,$D8,$FF,$28,$CF,$F8,$6F,$0E,$DE,$CF,$FA,$7B
	.dc.b	$EE,$F9,$B2,$B6,$E5,$A4,$9F,$E7,$A7,$FB,$1E,$3F,$D3,$15,$D6,$3F
	.dc.b	$DA,$11,$8F,$3F,$CE,$7F,$AB,$3E,$EE,$DF,$92,$2D,$7C,$79,$BA,$7F
	.dc.b	$B7,$BE,$E5,$F4,$EF,$5F,$F9,$D4,$61,$B1,$9B,$F2,$89,$D3,$8F,$95
	.dc.b	$7D,$0F,$F8,$22,$30,$D7,$DB,$1A,$EA,$FE,$90,$D8,$EB,$5F,$EB,$1F
	.dc.b	$96,$F1,$FC,$79,$A4,$9D,$FD,$DF,$FB,$AA,$8D,$24,$7B,$1F,$ED,$3A
	.dc.b	$7C,$16,$EF,$FD,$B2,$B8,$D7,$E4,$7E,$D3,$2F,$F8,$72,$1F,$77,$8D
	.dc.b	$DF,$9A,$BE,$4B,$0D,$BF,$F4,$AD,$F0,$6D,$F9,$C8,$FD,$06,$BF,$E0
	.dc.b	$58,$E0,$D8,$F1,$F4,$FE,$B3,$E6,$FC,$B2,$FB,$CF,$8C,$FF,$A9,$1A
	.dc.b	$FF,$83,$DB,$B7,$E8,$61,$5F,$B6,$F1,$79,$D6,$B6,$FF,$07,$B6,$BF
	.dc.b	$C7,$AD,$AE,$11,$FB,$7E,$4E,$3F,$57,$23,$04,$3C,$4E,$5F,$DD,$47
	.dc.b	$AB,$EE,$FF,$BC,$C7,$F5,$DF,$B5,$7C,$EF,$D2,$E8,$46,$27,$FD,$83
	.dc.b	$5C,$1F,$F1,$63,$5F,$D5,$B5,$FF,$14,$B5,$FF,$2B,$B5,$3F,$EF,$07
	.dc.b	$C5,$E8,$78,$67,$B6,$DF,$26,$77,$E7,$FA,$4D,$54,$8F,$2C,$65,$A4
	.dc.b	$3F,$BB,$FC,$F3,$4F,$85,$F0,$3E,$FB,$FF,$02,$D7,$F7,$F6,$BE,$E8
	.dc.b	$DB,$9F,$4F,$81,$87,$47,$C6,$8C,$63,$FE,$15,$E8,$FE,$03,$3F,$2B
	.dc.b	$D8,$F9,$DF,$79,$F7,$F7,$3F,$81,$9F,$8B,$3F,$5C,$DF,$6C,$8E,$44
	.dc.b	$3E,$9C,$6D,$FF,$0C,$5B,$ED,$FF,$33,$EF,$7F,$A4,$A7,$FC,$39,$F7
	.dc.b	$9F,$E5,$96,$FB,$1A,$F4,$64,$60,$87,$F9,$AE,$B9,$36,$36,$D3,$FD
	.dc.b	$2B,$AF,$D9,$F7,$9F,$0D,$E2,$F4,$BF,$1D,$D4,$F5,$69,$9C,$38,$F7
	.dc.b	$7F,$9C,$B3,$5E,$A7,$FC,$69,$7C,$5D,$9F,$EA,$49,$DB,$83,$FC,$DD
	.dc.b	$3D,$5F,$53,$F9,$FF,$BC,$FF,$79,$C7,$F8,$FF,$EF,$3F,$28,$FF,$8F
	.dc.b	$9B,$A7,$F9,$3D,$B8,$FF,$A1,$BE,$F1,$E4,$7F,$06,$6D,$F3,$ED,$FD
	.dc.b	$65,$F0,$8F,$F5,$CC,$E9,$FF,$25,$EB,$F9,$57,$D1,$19,$C7,$F4,$4D
	.dc.b	$C7,$FC,$35,$5B,$AF,$F7,$C4,$3E,$B9,$F1,$7F,$CB,$23,$0D,$8C,$CF
	.dc.b	$BF,$E3,$FD,$1E,$3F,$99,$91,$E2,$8F,$E0,$9F,$79,$FD,$A3,$3F,$F3
	.dc.b	$83,$71,$FF,$16,$36,$9F,$0C,$F2,$7F,$25,$1F,$4F,$EE,$DF,$79,$FC
	.dc.b	$B3,$C7,$FF,$5C,$4F,$D7,$6B,$FE,$B9,$9D,$BC,$67,$87,$C6,$48,$C1
	.dc.b	$0F,$ED,$33,$C1,$B3,$FE,$A9,$51,$97,$FD,$7F,$E5,$BE,$F1,$F3,$DF
	.dc.b	$D0,$BE,$47,$D9,$26,$CE,$AB,$F2,$33,$C6,$B8,$44,$33,$3B,$5B,$FF
	.dc.b	$B7,$7C,$3E,$DF,$19,$27,$9E,$3F,$E0,$37,$7C,$3F,$E2,$A8,$C5,$FF
	.dc.b	$B9,$6B,$85,$68,$FE,$9F,$3A,$AF,$F6,$DF,$DD,$BE,$6F,$FA,$DA,$FB
	.dc.b	$CA,$62,$25,$FE,$A6,$74,$D9,$CB,$FE,$ED,$8E,$8E,$48,$66,$7F,$B4
	.dc.b	$FB,$7F,$ED,$F8,$FD,$A7,$DE,$7F,$B7,$DB,$E5,$6B,$C3,$91,$FE,$DF
	.dc.b	$AE,$0F,$F8,$91,$AF,$F8,$E9,$AF,$C5,$F8,$0F,$B1,$F4,$D2,$3E,$5C
	.dc.b	$7D,$B3,$0F,$C7,$E3,$FB,$93,$98,$FE,$11,$4F,$BB,$FF,$00,$FC,$73
	.dc.b	$EF,$29,$EC,$89,$5F,$C3,$E7,$4D,$9C,$BF,$54,$FF,$21,$CF,$E4,$EB
	.dc.b	$B4,$7B,$E3,$07,$FB,$FF,$0F,$C8,$7C,$D7,$E4,$FD,$A9,$5F,$86,$8F
	.dc.b	$CB,$18,$7D,$9C,$7F,$DE,$5B,$1F,$E1,$AF,$BB,$7D,$DF,$F8,$73,$EF
	.dc.b	$1E,$85,$35,$12,$FF,$CA,$62,$B8,$B8,$6B,$FE,$58,$6B,$FE,$6E,$FB
	.dc.b	$BF,$F2,$BB,$FC,$CF,$C2,$7D,$E7,$F8,$B1,$B1,$FC,$DE,$76,$FC,$76
	.dc.b	$F6,$7F,$80,$E7,$73,$F2,$DB,$1F,$E0,$79,$DF,$5A,$A4,$74,$7C,$9E
	.dc.b	$8F,$F0,$38,$C4,$6C,$38,$79,$79,$3F,$D3,$DF,$B5,$F3,$3F,$79,$BC
	.dc.b	$BF,$E4,$69,$D6,$3F,$E1,$97,$84,$FF,$A4,$1A,$F7,$6B,$DB,$CC,$F9
	.dc.b	$FF,$3C,$FB,$C7,$C9,$FF,$13,$EE,$97,$FD,$82,$B5,$3F,$E6,$A6,$C7
	.dc.b	$FA,$29,$AF,$FA,$81,$AF,$BD,$AF,$C1,$FB,$CF,$F3,$63,$F9,$77,$F6
	.dc.b	$39,$7F,$8C,$DB,$BE,$FE,$27,$FC,$CB,$1C,$2B,$FD,$5A,$D7,$47,$E2
	.dc.b	$9C,$3F,$C5,$BE,$4B,$F9,$71,$1F,$E4,$3F,$07,$5F,$CD,$C5,$C3,$FC
	.dc.b	$D9,$D9,$FC,$CE,$BE,$07,$DE,$3E,$47,$E1,$7E,$46,$F1,$29,$A8,$99
	.dc.b	$7C,$5A,$E0,$D8,$FA,$31,$88,$FF,$59,$F8,$1F,$D0,$BE,$1F,$E9,$C7
	.dc.b	$F4,$06,$BF,$BD,$8F,$EC,$4F,$EC,$4E,$27,$F5,$17,$FA,$93,$93,$FD
	.dc.b	$A6,$F0,$23,$FE,$92,$F9,$67,$DE,$FF,$A4,$A7,$33,$FE,$1D,$7F,$AA
	.dc.b	$CE,$A3,$F0,$9F,$6C,$F0,$FF,$D3,$7F,$9C,$BD,$9A,$7E,$1E,$75,$5C
	.dc.b	$2B,$F2,$31,$88,$FC,$43,$5F,$C4,$35,$FD,$47,$CA,$3E,$F3,$FD,$53
	.dc.b	$F7,$BF,$A5,$E4,$FA,$ED,$E8,$DC,$FF,$AB,$2B,$85,$FF,$6A,$AD,$3F
	.dc.b	$EA,$86,$A8,$DB,$FC,$AD,$AF,$FB,$82,$7F,$BF,$46,$23,$B9,$FE,$E2
	.dc.b	$F0,$FF,$B8,$AF,$BC,$FF,$4F,$73,$BE,$F1,$F6,$DF,$EC,$26,$D7,$F9
	.dc.b	$2F,$0F,$F0,$E6,$9F,$C3,$70,$FE,$5D,$18,$9E,$9C,$5F,$C4,$71,$E9
	.dc.b	$8F,$33,$E0,$65,$E2,$C3,$8B,$6B,$E5,$F0,$7F,$13,$9E,$AD,$8F,$9F
	.dc.b	$3B,$C3,$F8,$F6,$3E,$1B,$DC,$79,$6B,$9B,$4D,$D7,$17,$E6,$36,$BF
	.dc.b	$0D,$1B,$3E,$85,$F8,$5E,$39,$66,$7C,$17,$DA,$7E,$09,$E4,$AE,$0F
	.dc.b	$6A,$E1,$7D,$E7,$37,$E0,$AE,$26,$CA,$F6,$7A,$1E,$18,$E3,$3E,$93
	.dc.b	$A4,$64,$EF,$51,$7C,$6F,$F3,$BB,$FB,$0D,$AF,$C2,$B6,$BF,$A5,$3F
	.dc.b	$CE,$3E,$4A,$F8,$7F,$D1,$F0,$FE,$9D,$18,$70,$D7,$FA,$63,$5D,$9E
	.dc.b	$E1,$EC,$B8,$BE,$84,$6B,$5F,$E1,$53,$98,$E0,$2B,$71,$99,$FE,$19
	.dc.b	$18,$AE,$75,$B3,$C1,$5B,$D7,$D5,$19,$BE,$F5,$C2,$F9,$4F,$81,$B3
	.dc.b	$DA,$39,$CE,$DA,$6C,$B1,$B3,$AB,$96,$99,$AF,$99,$7E,$59,$F9,$67
	.dc.b	$E3,$37,$FD,$B7,$77,$C3,$D3,$E2,$4F,$CE,$35,$C2,$9D,$65,$EE,$CF
	.dc.b	$F4,$58,$72,$8F,$F6,$35,$FC,$FC,$BF,$71,$87,$19,$CF,$67,$6A,$FC
	.dc.b	$03,$5F,$F0,$43,$5F,$F0,$83,$5F,$EA,$51,$FE,$86,$9F,$F7,$1B,$D0
	.dc.b	$C7,$C5,$79,$A7,$9C,$60,$D3,$FE,$57,$9E,$2F,$F9,$FF,$C2,$73,$E2
	.dc.b	$3F,$9A,$DF,$96,$72,$D9,$D4,$FF,$89,$67,$37,$FE,$9A,$E9,$FE,$9E
	.dc.b	$FF,$4A,$BF,$26,$69,$CD,$FF,$96,$F6,$FF,$5F,$5F,$FB,$BD,$B1,$FE
	.dc.b	$6D,$EA,$7D,$DF,$B6,$3F,$D4,$3C,$B6,$B6,$7D,$A2,$DF,$F2,$9B,$E2
	.dc.b	$F5,$9E,$07,$CF,$9E,$F5,$FE,$D5,$8C,$12,$7F,$D6,$BF,$95,$F2,$9F
	.dc.b	$44,$FB,$23,$15,$FA,$DD,$1E,$23,$E6,$FE,$53,$71,$FF,$16,$37,$1F
	.dc.b	$F1,$63,$5F,$F1,$63,$9B,$E9,$E1,$BE,$1B,$EF,$63,$DB,$1B,$39,$C7
	.dc.b	$FD,$83,$F1,$5F,$43,$D5,$37,$EE,$F0,$C9,$D9,$FE,$C8,$8C,$69,$FE
	.dc.b	$8F,$C5,$ED,$7E,$55,$B5,$FE,$4E,$8C,$11,$FF,$34,$69,$FA,$C9,$FF
	.dc.b	$05,$46,$0F,$F2,$44,$FD,$B8,$CD,$7E,$4F,$EE,$FF,$3E,$FF,$93,$DA
	.dc.b	$FF,$93,$DA,$FE,$45,$AA,$11,$CA,$FF,$C7,$FF,$77,$FC,$97,$0F,$F3
	.dc.b	$04,$E5,$FE,$DD,$19,$BF,$F5,$77,$C1,$7F,$7A,$9F,$F4,$CB,$63,$F8
	.dc.b	$FB,$5F,$E7,$4D,$7F,$D0,$4E,$EF,$F9,$87,$73,$EE,$FF,$82,$57,$6F
	.dc.b	$17,$F7,$13,$9F,$03,$F3,$DF,$76,$F8,$6F,$E0,$8D,$CF,$FD,$56,$DC
	.dc.b	$3F,$98,$B5,$FE,$98,$E7,$2F,$E8,$1F,$76,$FB,$A7,$DD,$F6,$C7,$F8
	.dc.b	$44,$E6,$31,$8F,$D7,$C5,$39,$8C,$43,$1E,$D1,$FE,$A7,$9D,$4F,$FB
	.dc.b	$5F,$E3,$1F,$0F,$FC,$75,$B5,$FE,$C4,$6C,$FF,$4A,$6B,$FD,$95,$CE
	.dc.b	$7F,$2B,$ED,$9F,$5B,$FC,$0E,$73,$7F,$EB,$7D,$D5,$A8,$F5,$78,$6E
	.dc.b	$AF,$F6,$BC,$61,$FF,$40,$37,$E0,$FF,$51,$EB,$7E,$A5,$B1,$93,$84
	.dc.b	$7F,$13,$FB,$9F,$F5,$6C,$7F,$B7,$9B,$AF,$F6,$46,$BF,$DF,$DA,$FF
	.dc.b	$B1,$DD,$EB,$FC,$27,$EE,$7F,$84,$D7,$CE,$9C,$BF,$CB,$E3,$34,$AF
	.dc.b	$EB,$35,$A8,$FD,$14,$EA,$73,$ED,$FF,$08,$8C,$5F,$F9,$37,$2F,$E1
	.dc.b	$17,$FF,$00,$3E,$E7,$F8,$47,$87,$FC,$25,$B5,$F1,$DB,$3B,$8F,$C1
	.dc.b	$FD,$B7,$EE,$A3,$FC,$38,$D7,$FD,$C8,$D7,$FD,$D8,$D4,$FF,$02,$CB
	.dc.b	$EF,$7C,$E7,$F3,$F8,$C5,$7D,$DE,$04,$FF,$5F,$FB,$97,$D1,$7E,$E7
	.dc.b	$0F,$F0,$F4,$63,$E6,$75,$39,$C3,$FD,$F9,$39,$7F,$3C,$F2,$77,$3C
	.dc.b	$1A,$6A,$79,$64,$AE,$4F,$D3,$9E,$B3,$A9,$CD,$FF,$B4,$35,$7D,$67
	.dc.b	$F3,$5F,$43,$F9,$B3,$70,$FC,$A3,$71,$E8,$39,$7F,$35,$19,$C3,$A7
	.dc.b	$DC,$7F,$3F,$FA,$5F,$E8,$0D,$C3,$F3,$CD,$C3,$9E,$1F,$A3,$8C,$5F
	.dc.b	$F8,$8B,$81,$1F,$E1,$36,$DF,$F8,$FA,$FE,$BC,$E4,$91,$FC,$1E,$9F
	.dc.b	$62,$FA,$3F,$B5,$5F,$E1,$59,$CC,$62,$B8,$56,$DF,$C5,$D6,$E3,$17
	.dc.b	$F0,$87,$C3,$FF,$52,$70,$DC,$3F,$CB,$D6,$F9,$5F,$E7,$D4,$38,$F9
	.dc.b	$9F,$E0,$BF,$B7,$FF,$34,$D7,$F1,$FE,$37,$C6,$F3,$8E,$16,$AD,$BF
	.dc.b	$52,$B7,$7F,$E3,$1F,$9E,$7D,$C7,$F0,$EA,$FB,$31,$8F,$03,$FC,$6D
	.dc.b	$7C,$31,$FE,$5D,$39,$7F,$96,$EA,$F8,$BF,$37,$9D,$FE,$21,$8C,$4F
	.dc.b	$F3,$2D,$13,$A8,$F5,$6B,$AC,$FE,$45,$6A,$73,$87,$8A,$B5,$BF,$FA
	.dc.b	$17,$3F,$83,$3C,$F0,$F7,$56,$AF,$E1,$E1,$67,$E6,$35,$7D,$C3,$EA
	.dc.b	$BE,$76,$12,$FE,$77,$9D,$F7,$33,$FF,$38,$E3,$CF,$3F,$F3,$87,$2E
	.dc.b	$F3,$F9,$D7,$BC,$3C,$38,$FC,$A8,$C4,$F8,$BD,$5F,$B8,$7F,$91,$F4
	.dc.b	$7D,$3E,$30,$F9,$13,$F4,$BE,$E1,$F0,$D9,$C3,$EE,$46,$2B,$67,$E1
	.dc.b	$5C,$47,$CC,$AE,$55,$FE,$55,$FB,$77,$CE,$D3,$EF,$D3,$EF,$71,$8F
	.dc.b	$F6,$AA,$DB,$F9,$1F,$B9,$F5,$D2,$72,$FE,$3E,$33,$18,$E2,$FB,$11
	.dc.b	$8C,$3F,$1A,$EB,$D0,$F2,$1F,$90,$E2,$FD,$26,$3E,$18,$CB,$FA,$1F
	.dc.b	$81,$F6,$E1,$FE,$A4,$C7,$FB,$56,$5F,$9F,$9C,$D7,$E9,$27,$35,$F8
	.dc.b	$99,$CD,$7E,$BE,$73,$5F,$B1,$C1,$5A,$AF,$77,$91,$F2,$4F,$D2,$E0
	.dc.b	$FB,$9C,$61,$F1,$E7,$F5,$3E,$07,$F9,$72,$30,$7E,$E5,$C5,$FF,$07
	.dc.b	$CB,$FA,$3C,$E7,$83,$9C,$FE,$E6,$B7,$8F,$BF,$8B,$FD,$65,$87,$E2
	.dc.b	$30,$FE,$40,$DB,$FF,$6F,$B5,$FF,$7D,$39,$AF,$D5,$5F,$F0,$17,$A5
	.dc.b	$79,$7B,$7F,$D4,$7F,$07,$FD,$DF,$07,$DC,$BE,$E7,$3E,$8F,$FA,$7E
	.dc.b	$31,$87,$F8,$6A,$F8,$46,$2F,$FD,$4D,$5B,$BF,$F5,$37,$C5,$7F,$93
	.dc.b	$B2,$FF,$6E,$B7,$1F,$F5,$24,$EB,$AA,$47,$F8,$F2,$75,$87,$F4,$FC
	.dc.b	$BB,$61,$CA,$FA,$FA,$93,$0F,$27,$DC,$DC,$3F,$9B,$63,$FE,$2C,$7F
	.dc.b	$94,$9F,$ED,$CA,$FF,$6F,$46,$63,$FD,$C7,$18,$3E,$2F,$CB,$73,$C5
	.dc.b	$87,$FA,$EE,$3D,$BE,$6C,$8F,$89,$F7,$37,$0F,$0C,$F6,$24,$FF,$59
	.dc.b	$C7,$FB,$8B,$FD,$6F,$ED,$DF,$4A,$F6,$B0,$DC,$3D,$22,$77,$3A,$3F
	.dc.b	$B2,$7C,$A3,$E8,$3F,$C2,$75,$C2,$31,$DB,$A9,$CC,$FC,$49,$CB,$DA
	.dc.b	$FF,$D9,$1F,$2B,$E5,$F8,$C3,$49,$3F,$33,$D4,$72,$BE,$97,$FA,$BF
	.dc.b	$CA,$FF,$61,$AF,$B5,$18,$F5,$3F,$DC,$8D,$8F,$BD,$F4,$17,$0E,$73
	.dc.b	$70,$E9,$5F,$DA,$70,$E9,$5A,$8C,$61,$FE,$EF,$C3,$87,$C3,$7F,$68
	.dc.b	$E7,$FB,$35,$F8,$BF,$87,$C7,$27,$FB,$D5,$B1,$F8,$FA,$B8,$FB,$1B
	.dc.b	$9F,$FA,$8E,$7F,$45,$58,$99,$18,$21,$FE,$1E,$9F,$F4,$FB,$FE,$82
	.dc.b	$F4,$7F,$27,$8F,$93,$0F,$EA,$DC,$9E,$09,$FC,$05,$7F,$89,$9D,$E5
	.dc.b	$FE,$09,$AF,$EB,$93,$BA,$8F,$53,$D3,$F1,$6F,$7F,$A7,$AB,$27,$FD
	.dc.b	$27,$39,$6C,$E5,$F0,$4F,$28,$E7,$D1,$D3,$92,$57,$28,$CB,$D2,$32
	.dc.b	$FF,$8F,$AF,$F3,$AF,$83,$E1,$BC,$BF,$51,$23,$04,$38,$39,$79,$47
	.dc.b	$B0,$C6,$DF,$EE,$C9,$C4,$FB,$3E,$8F,$E0,$FA,$12,$30,$43,$C8,$E5
	.dc.b	$F3,$47,$C8,$31,$5F,$EF,$89,$C4,$FC,$8F,$3F,$E6,$FD,$5C,$8C,$10
	.dc.b	$F8,$0E,$5F,$85,$1F,$50,$C5,$7E,$07,$25,$66,$BE,$AF,$D9,$FD,$AF
	.dc.b	$B9,$91,$82,$1F,$35,$CB,$F3,$E3,$F0,$26,$2F,$B4,$E2,$7F,$03,$F7
	.dc.b	$FF,$86,$FC,$0E,$46,$08,$7D,$97,$2F,$DB,$8F,$C6,$98,$AF,$54,$E2
	.dc.b	$7F,$1B,$F7,$5F,$92,$F5,$E4,$60,$FC,$8F,$47,$E5,$A6,$46,$0F,$CA
	.dc.b	$FB,$9F,$9A,$99,$18,$3F,$33,$EA,$FE,$7B,$D4,$91,$83,$FD,$DF,$E7
	.dc.b	$FF,$18,$6F,$B3,$FB,$C8,$FF,$17,$C6,$BC,$5C,$7C,$76,$7C,$49,$3C
	.dc.b	$A3,$F8,$2A,$FB,$17,$CC,$E1,$7D,$15,$31,$12,$7F,$D0,$1C,$5F,$C4
	.dc.b	$9E,$53,$B7,$3E,$1F,$F9,$FA,$FA,$C6,$0F,$F4,$43,$87,$FD,$24,$F8
	.dc.b	$1D,$1D,$0A,$EE,$BC,$1F,$63,$F8,$E9,$FF,$55,$35,$FE,$9A,$D7,$FB
	.dc.b	$3B,$5F,$F4,$8B,$B8,$E7,$D1,$F1,$23,$17,$E9,$D3,$E7,$65,$F3,$F2
	.dc.b	$98,$FD,$48,$CB,$F5,$A3,$2F,$FB,$13,$5F,$45,$FF,$1F,$5F,$6B,$F6
	.dc.b	$49,$3F,$7B,$E9,$2B,$C6,$2E,$9F,$CB,$69,$58,$BE,$95,$6F,$FD,$B9
	.dc.b	$5A,$7E,$87,$CD,$BF,$0D,$FE,$5F,$1E,$5F,$81,$FC,$51,$FD,$22,$75
	.dc.b	$CD,$6B,$79,$FF,$50,$9C,$DF,$49,$CE,$5E,$DF,$4B,$F2,$9F,$51,$2B
	.dc.b	$F3,$2F,$E5,$FE,$87,$11,$F1,$62,$E3,$ED,$9F,$CA,$FD,$8B,$D7,$BF
	.dc.b	$1D,$30,$FC,$2C,$E5,$B3,$97,$FD,$B9,$09,$E7,$32,$78,$C6,$EF,$F3
	.dc.b	$55,$CE,$7E,$C7,$BF,$FA,$DE,$09,$1F,$B2,$8F,$D8,$18,$7E,$F4,$61
	.dc.b	$FC,$CF,$EC,$CF,$6F,$F5,$FF,$B5,$7F,$48,$9F,$F7,$83,$71,$FD,$55
	.dc.b	$6E,$78,$D7,$ED,$A3,$11,$FE,$F7,$6C,$FE,$C6,$BF,$B0,$4F,$FB,$EF
	.dc.b	$E1,$3F,$83,$37,$BF,$F7,$F1,$FC,$1A,$35,$D5,$83,$33,$23,$F6,$33
	.dc.b	$CA,$FF,$7C,$DE,$2F,$F7,$CB,$CE,$33,$5C,$EB,$F7,$3F,$94,$FF,$AE
	.dc.b	$1F,$DD,$FC,$1B,$E3,$FF,$88,$B6,$37,$E7,$7F,$14,$AE,$10,$C3,$F8
	.dc.b	$74,$61,$B3,$CB,$E4,$BF,$31,$F8,$7F,$F8,$D6,$FF,$61,$F0,$DF,$C9
	.dc.b	$4C,$F8,$D2,$3F,$47,$3B,$C3,$F3,$8D,$DB,$F6,$DF,$09,$FC,$B7,$E1
	.dc.b	$7F,$97,$7C,$14,$C3,$F9,$9C,$65,$FE,$6B,$19,$7F,$80,$47,$F3,$5F
	.dc.b	$81,$FE,$6D,$F0,$12,$7F,$9D,$C6,$5F,$E7,$B1,$97,$F8,$44,$7F,$8A
	.dc.b	$7B,$3F,$C5,$5D,$92,$7F,$C6,$11,$97,$FC,$65,$19,$7F,$C5,$D1,$FE
	.dc.b	$31,$FA,$DF,$F1,$97,$D5,$C9,$FF,$1C,$46,$5F,$F1,$D4,$65,$FF,$1B
	.dc.b	$46,$5F,$F1,$C4,$67,$6F,$F1,$F7,$88,$AD,$46,$27,$5B,$7F,$8F,$6B
	.dc.b	$57,$FE,$47,$FA,$3F,$F2,$27,$34,$8C,$10,$FF,$23,$39,$7F,$C9,$11
	.dc.b	$FE,$4B,$F9,$2F,$F2,$67,$C7,$49,$FF,$29,$46,$5F,$F2,$A4,$65,$FF
	.dc.b	$28,$46,$5F,$F2,$94,$67,$2F,$5C,$FF,$95,$FE,$9B,$FC,$B1,$F2,$12
	.dc.b	$FF,$CB,$D1,$97,$FC,$C1,$19,$7F,$CB,$91,$97,$FC,$BD,$19,$AF,$8F
	.dc.b	$39,$7E,$4C,$67,$3F,$8D,$3F,$DD,$3E,$2F,$FD,$F7,$4B,$7D,$63,$6C
	.dc.b	$42,$BF,$9F,$3A,$8C,$C6,$27,$84,$F3,$8E,$AE,$CC,$79,$9F,$AF,$70
	.dc.b	$75,$F3,$BB,$D7,$28,$C1,$E5,$71,$5D,$1B,$5C,$DC,$D7,$2B,$F2,$B6
	.dc.b	$FC,$38,$73,$9F,$4D,$F1,$77,$39,$3D,$55,$C5,$F8,$53,$B7,$99,$EA
	.dc.b	$C3,$E1,$57,$28,$C4,$ED,$E3,$7F,$3F,$C2,$FE,$4D,$7F,$19,$B5,$F1
	.dc.b	$5B,$3B,$F0,$7F,$DB,$7F,$17,$70,$FF,$7A,$B5,$FF,$64,$9F,$ED,$08
	.dc.b	$DB,$4D,$9F,$40,$FF,$74,$6B,$EF,$78,$E9,$F4,$5C,$9B,$15,$B7,$31
	.dc.b	$B7,$27,$5F,$88,$FF,$64,$C6,$67,$84,$63,$4F,$2C,$F7,$8C,$57,$9A
	.dc.b	$31,$7D,$E3,$AC,$61,$F2,$5F,$49,$D3,$63,$F2,$1F,$12,$E5,$7C,$87
	.dc.b	$FB,$73,$4F,$E1,$BF,$1F,$71,$FE,$15,$37,$06,$1C,$6B,$0F,$39,$B8
	.dc.b	$26,$C6,$1B,$7F,$ED,$96,$DF,$FB,$55,$B8,$78,$A7,$FD,$BD,$5C,$5F
	.dc.b	$F7,$07,$D0,$76,$FA,$0E,$ED,$8C,$9F,$EE,$58,$CB,$4C,$E1,$FE,$E3
	.dc.b	$8C,$B6,$3F,$DD,$9E,$67,$AB,$E7,$7D,$6D,$8C,$9F,$EF,$18,$CB,$4F
	.dc.b	$F7,$A7,$8F,$F0,$BE,$77,$E1,$B6,$32,$7F,$BE,$63,$2D,$33,$7F,$EF
	.dc.b	$78,$CB,$63,$39,$FF,$42,$9C,$B6,$3F,$CF,$DB,$7D,$DF,$2F,$EA,$FE
	.dc.b	$CA,$9F,$AB,$18,$69,$C2,$34,$F2,$E0,$FF,$04,$39,$3E,$FF,$A0,$F9
	.dc.b	$2F,$C0,$FE,$E7,$FC,$9E,$7F,$D4,$F1,$8D,$36,$D8,$C9,$E6,$9E,$13
	.dc.b	$FE,$0F,$8F,$E6,$19,$27,$F5,$F9,$FA,$6B,$F2,$30,$F9,$2F,$0F,$3F
	.dc.b	$89,$B8,$75,$6D,$75,$38,$BF,$9B,$F6,$9E,$1F,$F6,$BA,$FF,$3B,$B6
	.dc.b	$B8,$65,$E3,$F1,$B8,$4E,$73,$FF,$3A,$36,$FD,$39,$FF,$98,$5E,$7E
	.dc.b	$37,$C6,$F1,$FF,$55,$97,$A7,$8A,$EB,$F3,$A3,$CD,$B3,$EA,$1E,$26
	.dc.b	$AE,$65,$FE,$BB,$AD,$78,$B8,$C3,$8E,$08,$FD,$C5,$3F,$1D,$FF,$66
	.dc.b	$BF,$EC,$33,$1E,$EE,$A3,$37,$F1,$E7,$26,$E3,$3C,$1F,$4E,$33,$3B
	.dc.b	$9C,$C7,$E0,$DF,$F5,$7B,$C7,$DE,$ED,$7D,$DB,$1F,$ED,$0F,$83,$7C
	.dc.b	$5F,$C6,$6C,$7B,$FE,$91,$F0,$99,$A9,$39,$F8,$DB,$9F,$CA,$B4,$7E
	.dc.b	$5B,$93,$F2,$FE,$45,$BF,$C9,$36,$FE,$FF,$37,$F7,$F9,$FC,$B5,$70
	.dc.b	$7F,$DD,$5F,$49,$FE,$05,$F1,$FF,$92,$3F,$E0,$7F,$1A,$E1,$FC,$86
	.dc.b	$B6,$A3,$35,$A6,$CE,$9B,$3F,$97,$AE,$93,$CA,$B8,$3F,$A9,$EA,$FD
	.dc.b	$89,$FB,$2F,$46,$C6,$49,$7E,$C8,$CB,$FA,$3E,$EF,$D0,$F8,$FF,$9C
	.dc.b	$3F,$77,$E8,$5B,$FD,$1C,$7E,$FB,$D1,$C5,$FC,$67,$F8,$0F,$6C,$AF
	.dc.b	$DE,$75,$5A,$F9,$73,$7A,$BF,$75,$E0,$FE,$F6,$B1,$7F,$A8,$AB,$7F
	.dc.b	$BE,$9B,$5F,$C0,$FC,$AF,$E1,$FB,$FF,$10,$6E,$1F,$C2,$DB,$7F,$69
	.dc.b	$B8,$FF,$09,$87,$D1,$BC,$9F,$D9,$B6,$FF,$DE,$FF,$82,$3B,$33,$52
	.dc.b	$7F,$C8,$37,$FE,$42,$BE,$0E,$0E,$B3,$DD,$D3,$57,$23,$E6,$DF,$2B
	.dc.b	$F8,$A5,$F2,$B6,$30,$DC,$7F,$23,$3A,$6C,$E9,$A6,$A7,$85,$FE,$2E
	.dc.b	$B8,$3F,$9E,$E8,$FE,$73,$C9,$FC,$E9,$B7,$FC,$D7,$CF,$FE,$7B,$E7
	.dc.b	$FF,$3E,$6C,$7F,$39,$D3,$FA,$0F,$9F,$FD,$09,$B1,$FC,$F7,$C8,$FF
	.dc.b	$02,$FE,$E9,$FD,$33,$3F,$D6,$C6,$26,$1F,$D6,$25,$F3,$58,$AF,$E6
	.dc.b	$31,$89,$D4,$F4,$FB,$9E,$AD,$8E,$5F,$37,$E0,$F9,$BF,$0B,$63,$27
	.dc.b	$78,$CB,$4C,$E1,$E0,$8C,$B6,$3C,$DF,$37,$EC,$E3,$F6,$B6,$32,$7A
	.dc.b	$E3,$2D,$33,$3E,$C8,$CB,$63,$E1,$71,$FC,$8F,$85,$F9,$2D,$8C,$9F
	.dc.b	$1E,$32,$D3,$33,$F2,$23,$2D,$8F,$9D,$F3,$7F,$53,$CB,$FA,$AD,$8C
	.dc.b	$9F,$4E,$32,$D3,$33,$F5,$23,$2D,$8F,$B5,$E4,$FE,$03,$CB,$FC,$0B
	.dc.b	$63,$27,$DF,$8C,$B4,$CC,$FE,$02,$32,$D8,$FC,$37,$CD,$7E,$33,$E6
	.dc.b	$BF,$1A,$D8,$C9,$F8,$B8,$CB,$4F,$C7,$7C,$CF,$F8,$93,$87,$F2,$AD
	.dc.b	$8C,$9F,$E2,$38,$CB,$4F,$F1,$37,$17,$F8,$B3,$8B,$FC,$5A,$D8,$C9
	.dc.b	$FE,$2B,$8C,$B4,$FF,$17,$71,$7F,$8D,$38,$BF,$C6,$AD,$8C,$9F,$E3
	.dc.b	$38,$CB,$4C,$E7,$FE,$14,$F6,$93,$A8,$C3,$FE,$40,$E3,$FF,$14,$7B
	.dc.b	$6F,$88,$FF,$16,$DF,$F1,$0C,$3F,$DB,$D3,$9C,$3F,$CF,$B3,$FC,$1A
	.dc.b	$31,$E3,$23,$84,$FF,$A8,$2F,$9E,$1D,$9B,$87,$47,$BF,$36,$6B,$AD
	.dc.b	$55,$F3,$AD,$5E,$BF,$F2,$B9,$CC,$61,$FE,$5D,$3B,$9D,$36,$3F,$99
	.dc.b	$7C,$13,$E4,$FF,$97,$46,$1A,$FF,$3A,$8D,$78,$7D,$DB,$1D,$B8,$5F
	.dc.b	$05,$D9,$B1,$98,$C1,$FD,$07,$81,$F3,$AF,$89,$57,$C3,$3F,$D5,$31
	.dc.b	$97,$13,$C2,$77,$1F,$34,$DF,$C3,$E3,$93,$E9,$DF,$19,$D5,$F1,$79
	.dc.b	$C6,$67,$75,$C7,$8B,$FC,$DB,$2F,$88,$EF,$E6,$FF,$74,$8F,$F3,$E6
	.dc.b	$1F,$E8,$A6,$BF,$E6,$27,$83,$99,$E3,$CB,$D3,$E0,$AC,$FB,$9B,$9F
	.dc.b	$C4,$AB,$7C,$AB,$A5,$60,$CC,$F0,$F8,$4F,$ED,$F8,$FF,$82,$1F,$C1
	.dc.b	$57,$F9,$D1,$B1,$98,$E0,$71,$E4,$8F,$FA,$6C,$F1,$FC,$EF,$F4,$E7
	.dc.b	$FD,$3D,$F0,$5D,$CC,$F6,$C8,$C9,$0F,$F4,$1C,$F1,$6F,$47,$4E,$49
	.dc.b	$1F,$EB,$29,$FC,$E3,$67,$2F,$29,$F1,$C7,$93,$B3,$CB,$D5,$2B,$94
	.dc.b	$65,$E9,$19,$7C,$71,$97,$CB,$19,$CF,$FC,$5F,$87,$F0,$69,$F2,$75
	.dc.b	$FC,$2F,$16,$46,$08,$7B,$9C,$BF,$02,$3E,$49,$8A,$EB,$38,$9F,$81
	.dc.b	$E2,$FC,$EF,$16,$46,$08,$7C,$A7,$2F,$CC,$8F,$AA,$62,$BD,$33,$89
	.dc.b	$F9,$9E,$2F,$DA,$F1,$64,$60,$87,$D6,$72,$FD,$88,$FC,$09,$8A,$F8
	.dc.b	$B3,$89,$FB,$1E,$2F,$E1,$BC,$59,$18,$21,$F8,$27,$2F,$E1,$23,$F1
	.dc.b	$A6,$2B,$E8,$CE,$27,$F0,$9E,$2F,$E4,$BE,$76,$46,$0F,$C4,$73,$7E
	.dc.b	$5A,$64,$60,$FC,$67,$CE,$FE,$6A,$64,$60,$FC,$87,$CE,$FE,$7B,$E7
	.dc.b	$64,$60,$FE,$27,$97,$F9,$CB,$E2,$7F,$8D,$C7,$FB,$79,$7C,$5F,$EE
	.dc.b	$1B,$1C,$3C,$0F,$F7,$AC,$70,$AE,$5E,$B6,$FC,$29,$7B,$1E,$12,$FF
	.dc.b	$CD,$91,$91,$F3,$AF,$9C,$FE,$13,$18,$F8,$7B,$26,$CC,$9E,$53,$96
	.dc.b	$CE,$5F,$F8,$00,$76,$9E,$71,$23,$8C,$78,$3E,$77,$FD,$9B,$1D,$32
	.dc.b	$F0,$99,$E8,$91,$FD,$56,$76,$FF,$A0,$5A,$FF,$A1,$9A,$FC,$8A,$F9
	.dc.b	$97,$23,$E3,$65,$F1,$67,$E8,$19,$8F,$A9,$38,$77,$93,$E2,$7E,$B1
	.dc.b	$23,$3F,$21,$5F,$B9,$17,$6F,$C0,$4D,$DF,$ED,$CD,$AF,$AF,$6A,$FE
	.dc.b	$77,$F0,$8F,$9D,$72,$FE,$17,$E4,$66,$1F,$7E,$B5,$3C,$2B,$73,$96
	.dc.b	$BF,$ED,$16,$E7,$F9,$3F,$4B,$F2,$1F,$3B,$2B,$DE,$FE,$4F,$E7,$71
	.dc.b	$1F,$97,$8B,$5F,$9B,$9F,$C4,$FC,$EB,$E4,$EF,$C9,$CC,$3F,$21,$39
	.dc.b	$6C,$E5,$FC,$3C,$27,$9C,$C9,$E3,$1B,$BF,$F6,$65,$73,$9F,$D5,$7C
	.dc.b	$EF,$EA,$78,$24,$7D,$88,$FD,$61,$87,$F6,$11,$87,$F4,$9F,$3A,$F8
	.dc.b	$1F,$D1,$FC,$EF,$F4,$A9,$FC,$DB,$71,$FF,$6E,$56,$E7,$8D,$7E,$AA
	.dc.b	$31,$1F,$A2,$6C,$FF,$B8,$EB,$FA,$D3,$FE,$E5,$F9,$DF,$DE,$73,$7F
	.dc.b	$84,$9F,$E0,$37,$FA,$A6,$CE,$FD,$AF,$DB,$C7,$09,$FF,$0F,$BA,$AE
	.dc.b	$5E,$06,$AA,$57,$F0,$1A,$D3,$C2,$72,$FE,$E3,$E7,$5E,$67,$F0,$4F
	.dc.b	$91,$7C,$1D,$3F,$83,$CE,$AB,$F8,$63,$5E,$55,$A3,$97,$C2,$7E,$73
	.dc.b	$E7,$3F,$CC,$13,$FC,$87,$E7,$7F,$90,$19,$F1,$E4,$7E,$76,$77,$BB
	.dc.b	$7F,$F1,$EF,$6A,$73,$1F,$C9,$BE,$77,$F9,$57,$CE,$FF,$2B,$F8,$29
	.dc.b	$8F,$EE,$A3,$2F,$EF,$23,$2F,$F3,$38,$FE,$65,$F0,$3F,$CC,$FD,$F9
	.dc.b	$3F,$C0,$A3,$2F,$F0,$48,$CB,$FC,$EE,$3F,$C5,$5D,$BF,$E2,$BE,$D9
	.dc.b	$3F,$C4,$3C,$2F,$CC,$6D,$FE,$2D,$AC,$CF,$F8,$B6,$B8,$BB,$F9,$7F
	.dc.b	$F1,$A7,$CC,$C8,$C1,$0F,$F1,$A3,$97,$FC,$6D,$1F,$E3,$B3,$17,$FE
	.dc.b	$3B,$9C,$4F,$18,$C3,$E5,$8C,$BF,$E3,$FF,$98,$FF,$22,$74,$48,$F5
	.dc.b	$47,$F9,$1C,$C3,$FE,$45,$8C,$3E,$FF,$98,$FF,$26,$7C,$C4,$8C,$10
	.dc.b	$FF,$26,$39,$7F,$C9,$D1,$FE,$53,$31,$87,$F9,$4E,$71,$3F,$2F,$E6
	.dc.b	$3F,$CB,$1F,$31,$23,$04,$3F,$CB,$0E,$5F,$F2,$D4,$7F,$97,$CC,$57
	.dc.b	$F9,$7E,$71,$3C,$63,$0F,$D9,$8C,$BF,$D7,$BE,$29,$E5,$AB,$FD,$23
	.dc.b	$F3,$AD,$7F,$D1,$B5,$FC,$F6,$F3,$18,$F3,$7F,$D2,$35,$B9,$FF,$4C
	.dc.b	$D7,$4B,$DC,$63,$0E,$CA,$B8,$2A,$EC,$E7,$0E,$CF,$23,$55,$CA,$B8
	.dc.b	$D7,$79,$E0,$78,$B0,$F2,$DF,$28,$C9,$CB,$0E,$95,$C6,$F9,$DF,$0A
	.dc.b	$E4,$6B,$0F,$06,$1F,$11,$B1,$A3,$AD,$FA,$AF,$CB,$3A,$9F,$01,$FE
	.dc.b	$C6,$F0,$79,$63,$FE,$C0,$F0,$FF,$94,$E1,$EB,$6D,$FC,$87,$F9,$6D
	.dc.b	$FF,$BC,$8E,$47,$FB,$08,$7C,$5F,$3C,$31,$18,$36,$7F,$B1,$4A,$73
	.dc.b	$C7,$E9,$BF,$EE,$CE,$BF,$A0,$E7,$3E,$DF,$13,$FC,$6E,$3E,$59,$93
	.dc.b	$FD,$A6,$E6,$B9,$C7,$85,$EB,$19,$9F,$2E,$1D,$A3,$15,$EB,$8C,$5F
	.dc.b	$8A,$39,$57,$8A,$31,$3A,$9F,$19,$A3,$C7,$1A,$7C,$F5,$DB,$D3,$FF
	.dc.b	$0A,$FC,$C7,$AF,$6F,$F6,$47,$CC,$DC,$3F,$DE,$75,$A7,$F9,$14,$67
	.dc.b	$0F,$F2,$BF,$53,$0C,$5F,$F8,$82,$AC,$E6,$F8,$9F,$CA,$EB,$33,$A6
	.dc.b	$C6,$4D,$DF,$FB,$66,$32,$FF,$B6,$FE,$57,$AF,$CB,$76,$6D,$7F,$37
	.dc.b	$6C,$EE,$3F,$9D,$4E,$DA,$6C,$B1,$FE,$E7,$F3,$7D,$3F,$2D,$EA,$6C
	.dc.b	$FF,$43,$6C,$EE,$3F,$A3,$4F,$C0,$F2,$3E,$0F,$CB,$59,$FE,$94,$DA
	.dc.b	$FF,$7A,$37,$6F,$F7,$C3,$67,$E8,$EA,$8C,$57,$D0,$6E,$DF,$E4,$E6
	.dc.b	$CE,$EB,$FA,$C7,$99,$F5,$3E,$53,$FB,$94,$60,$FF,$35,$B8,$C3,$F9
	.dc.b	$14,$78,$27,$83,$F6,$9A,$FF,$9B,$3E,$47,$EF,$FC,$D2,$BF,$03,$E9
	.dc.b	$A7,$31,$86,$D7,$13,$FC,$F7,$5C,$5B,$C9,$D3,$E4,$BA,$B6,$32,$70
	.dc.b	$8C,$B4,$FF,$36,$78,$6F,$13,$F2,$DE,$6A,$31,$3E,$0F,$1F,$31,$8C
	.dc.b	$FE,$08,$AE,$17,$EE,$F1,$F6,$DC,$3F,$D6,$98,$FF,$8B,$9B,$B7,$E9
	.dc.b	$73,$7C,$EB,$C5,$FD,$46,$6C,$38,$3F,$EB,$06,$F2,$FF,$0D,$BE,$3E
	.dc.b	$2F,$09,$CE,$5F,$3E,$72,$FF,$B2,$47,$CE,$FF,$B0,$5C,$3F,$EA,$16
	.dc.b	$FB,$DF,$B9,$77,$EF,$59,$E1,$97,$C2,$9C,$9C,$0E,$73,$BA,$E0,$D8
	.dc.b	$C9,$FE,$D7,$F8,$27,$C5,$FC,$06,$C7,$FB,$7F,$17,$CE,$BE,$13,$57
	.dc.b	$2B,$52,$F8,$DC,$67,$F5,$B0,$E3,$1F,$92,$E3,$FC,$9F,$8F,$70,$FC
	.dc.b	$6B,$6F,$EE,$F7,$BA,$BF,$BF,$57,$E4,$AF,$8B,$FE,$F3,$F9,$BF,$CE
	.dc.b	$78,$FF,$C7,$DF,$CF,$78,$D7,$14,$D8,$C3,$6B,$94,$E9,$B3,$A6,$9A
	.dc.b	$9E,$13,$E6,$C3,$F3,$13,$FA,$6E,$CF,$D7,$FA,$3F,$B0,$64,$64,$FD
	.dc.b	$5C,$62,$BF,$37,$DB,$FA,$1F,$1F,$F3,$E7,$EE,$3D,$1B,$7F,$A8,$8F
	.dc.b	$DD,$FA,$38,$BF,$13,$FB,$EE,$D9,$5F,$B8,$EC,$B5,$E7,$9B,$DB,$FD
	.dc.b	$97,$8B,$E7,$56,$2F,$FA,$5E,$DF,$A3,$7F,$80,$DF,$F0,$A9,$D6,$0B
	.dc.b	$FA,$DE,$57,$F0,$EE,$0F,$E1,$ED,$C3,$F8,$4B,$6B,$E4,$37,$2F,$EE
	.dc.b	$D0,$F9,$C7,$93,$88,$FE,$30,$65,$C3,$FC,$6E,$37,$7F,$E0,$A8,$D3
	.dc.b	$63,$31,$FC,$22,$9E,$EB,$E2,$FF,$7F,$5A,$9F,$ED,$D3,$9B,$FC,$77
	.dc.b	$CC,$BE,$53,$F1,$CB,$E4,$EC,$61,$B3,$C1,$B9,$7F,$18,$E9,$5F,$1A
	.dc.b	$EB,$7C,$5A,$71,$3F,$98,$F4,$FF,$3A,$E5,$FE,$76,$D8,$FE,$6B,$E8
	.dc.b	$7F,$3E,$F4,$3F,$9F,$B6,$3F,$9C,$F0,$7F,$42,$F4,$3F,$A1,$B6,$3F
	.dc.b	$9E,$F9,$5F,$E0,$4F,$9C,$FE,$9D,$C3,$F8,$5E,$9F,$E3,$7C,$1F,$7E
	.dc.b	$73,$7F,$E1,$06,$C6,$7E,$67,$94,$F3,$63,$FD,$46,$FF,$21,$4F,$9C
	.dc.b	$CD,$FF,$00,$8C,$56,$CE,$DF,$33,$E5,$F9,$AF,$33,$63,$55,$E5,$8C
	.dc.b	$3C,$DB,$87,$81,$B3,$D0,$E3,$D6,$C1,$87,$C3,$F9,$AF,$89,$C9,$6F
	.dc.b	$75,$F8,$36,$CE,$9A,$6A,$78,$65,$ED,$6E,$5F,$09,$FA,$1C,$FF,$47
	.dc.b	$E1,$AD,$F6,$6D,$FC,$56,$DF,$CA,$6C,$FD,$36,$CF,$D7,$7E,$E7,$CC
	.dc.b	$FD,$DF,$32,$D7,$A9,$B5,$F4,$5B,$7F,$59,$B3,$F7,$DB,$3F,$83,$7F
	.dc.b	$11,$E4,$FE,$27,$CC,$B5,$F1,$9B,$5F,$75,$B7,$F8,$26,$CF,$E2,$DB
	.dc.b	$3F,$8F,$7F,$29,$F3,$3F,$95,$F9,$9B,$5F,$49,$B5,$F8,$96,$DF,$E3
	.dc.b	$9F,$CE,$7C,$C7,$F8,$B7,$96,$DF,$F8,$B1,$B7,$FD,$43,$DA,$46,$30
	.dc.b	$FE,$4D,$18,$AF,$F1,$9F,$37,$F8,$DF,$9B,$FC,$70,$D8,$C9,$FE,$38
	.dc.b	$8C,$B4,$FF,$1D,$F2,$FF,$8F,$F9,$3F,$C8,$0D,$8C,$9F,$E4,$08,$CB
	.dc.b	$4C,$E5,$E7,$8C,$B6,$3F,$C8,$FC,$9F,$EA,$1F,$71,$F1,$3F,$E8,$6C
	.dc.b	$FF,$31,$7F,$EB,$28,$C7,$36,$DD,$5D,$9F,$E7,$D5,$C2,$F9,$C6,$27
	.dc.b	$5D,$0A,$FE,$39,$3D,$4D,$63,$FE,$51,$BE,$30,$F9,$C6,$AF,$4B,$F9
	.dc.b	$D4,$61,$B1,$99,$FF,$63,$CE,$5F,$F6,$4F,$C1,$BE,$4F,$F9,$6C,$61
	.dc.b	$AF,$FB,$46,$35,$E2,$FF,$47,$6C,$74,$E1,$7C,$1F,$89,$B1,$98,$C1
	.dc.b	$FE,$D8,$DB,$E0,$76,$FF,$56,$D7,$E2,$C7,$C3,$ED,$C7,$17,$D1,$6E
	.dc.b	$1F,$32,$ED,$72,$8F,$8D,$5B,$69,$F4,$AB,$7C,$2B,$BD,$BF,$D9,$B9
	.dc.b	$FF,$2E,$DE,$3F,$C0,$37,$0F,$B7,$DB,$73,$FB,$13,$71,$FF,$1F,$57
	.dc.b	$FA,$42,$31,$59,$64,$64,$FF,$4A,$C7,$1E,$D9,$1F,$E9,$87,$A7,$6E
	.dc.b	$23,$F5,$D1,$72,$FD,$6C,$F8,$7B,$65,$7F,$A8,$1F,$27,$6E,$23,$F6
	.dc.b	$D1,$6B,$F6,$B3,$EB,$EE,$7C,$CD,$25,$FF,$38,$F9,$9A,$FF,$4F,$8B
	.dc.b	$7F,$D5,$27,$E2,$7E,$B7,$F8,$0A,$FF,$29,$B6,$7F,$C5,$0D,$E0,$FF
	.dc.b	$78,$AF,$86,$E3,$CD,$2F,$FA,$6C,$ED,$FF,$14,$3B,$73,$1B,$7B,$BE
	.dc.b	$29,$F1,$FB,$72,$79,$BE,$6E,$6C,$47,$9A,$2E,$3E,$69,$B8,$F9,$A6
	.dc.b	$D7,$8A,$7D,$FF,$31,$23,$CE,$FC,$2F,$7B,$11,$F0,$A2,$E1,$F0,$A6
	.dc.b	$E1,$F0,$A6,$D7,$BA,$7E,$5F,$BD,$23,$E1,$BF,$3B,$E2,$71,$1F,$3A
	.dc.b	$2E,$1F,$3A,$6E,$1F,$3A,$6D,$7C,$A9,$FA,$FF,$13,$23,$E7,$BF,$6B
	.dc.b	$C0,$C4,$7D,$A8,$B8,$7D,$A9,$B8,$7D,$A9,$B5,$F5,$A7,$F0,$7F,$2F
	.dc.b	$F8,$4F,$97,$93,$F7,$23,$2D,$21,$F8,$8F,$96,$FC,$59,$1F,$C1,$BF
	.dc.b	$8E,$F8,$AF,$C7,$FC,$9F,$E4,$3E,$4A,$57,$E2,$23,$F2,$7F,$23,$23
	.dc.b	$F1,$6F,$E5,$BE,$4B,$F2,$FF,$23,$27,$F1,$EF,$E6,$BE,$4B,$F8,$04
	.dc.b	$3E,$69,$BE,$1B,$1F,$E2,$A3,$51,$9D,$F4,$C8,$D4,$F0,$7F,$DE,$67
	.dc.b	$59,$D3,$88,$D7,$0F,$F9,$F9,$BC,$3F,$E7,$C7,$F9,$C4,$FE,$9B,$E4
	.dc.b	$7F,$8A,$78,$3F,$E8,$08,$FF,$61,$B6,$FF,$63,$3F,$B6,$37,$30,$E4
	.dc.b	$7A,$FE,$6B,$F8,$DB,$CC,$38,$7B,$35,$FF,$47,$B5,$1C,$3C,$5F,$97
	.dc.b	$1D,$2B,$FD,$32,$DC,$7D,$9D,$BE,$AF,$43,$D7,$F1,$F3,$1F,$F2,$7E
	.dc.b	$7F,$D6,$DF,$6C,$E5,$F7,$CF,$17,$71,$9C,$7F,$D7,$14,$AD,$57,$D2
	.dc.b	$C5,$F2,$7E,$62,$47,$DD,$FB,$C7,$9B,$88,$E1,$37,$1F,$F5,$9D,$6A
	.dc.b	$7C,$FF,$29,$F5,$3E,$52,$46,$08,$7D,$67,$2F,$D8,$8F,$C8,$FC,$D3
	.dc.b	$E2,$E9,$27,$F2,$10,$F8,$BC,$47,$E3,$23,$83,$A9,$E7,$3A,$C7,$FD
	.dc.b	$93,$5A,$AE,$15,$A7,$F3,$9E,$C3,$AB,$FD,$8D,$1F,$E1,$06,$C6,$63
	.dc.b	$FD,$94,$D7,$F1,$0D,$BF,$F6,$BB,$78,$FF,$DA,$AD,$9F,$C3,$FA,$6E
	.dc.b	$CF,$D8,$FC,$D3,$E3,$69,$2B,$F6,$1F,$34,$E6,$C4,$7E,$B2,$B8,$3F
	.dc.b	$B5,$6B,$FB,$09,$E0,$FE,$E7,$E2,$BC,$0D,$C3,$F8,$0C,$62,$3F,$6E
	.dc.b	$F9,$DF,$F7,$1C,$68,$8F,$3D,$6B,$C2,$2E,$BF,$EF,$8A,$E5,$C4,$F8
	.dc.b	$9F,$E1,$ED,$BF,$E1,$8D,$7F,$83,$FE,$19,$FD,$8D,$46,$6D,$C1,$FF
	.dc.b	$00,$23,$0D,$BF,$E3,$77,$C7,$F1,$8F,$F7,$A6,$1F,$C6,$5A,$FF,$14
	.dc.b	$6E,$1D,$23,$F7,$7F,$86,$FC,$38,$C4,$27,$78,$7F,$00,$8C,$7C,$0F
	.dc.b	$F2,$A9,$CC,$72,$D3,$F8,$BF,$CB,$48,$FD,$EB,$FC,$6B,$E5,$F1,$1F
	.dc.b	$C6,$A2,$E3,$FB,$A9,$FE,$3F,$F2,$F2,$BF,$82,$BF,$C8,$BE,$5F,$11
	.dc.b	$FC,$8A,$2D,$7F,$02,$9F,$F1,$C7,$CF,$CA,$FF,$6D,$FC,$02,$3F,$C7
	.dc.b	$9D,$F9,$9F,$F4,$F4,$5B,$FF,$4E,$CF,$F9,$07,$E5,$25,$72,$7F,$C8
	.dc.b	$9E,$E6,$23,$FD,$4D,$16,$BF,$D4,$B3,$78,$3F,$D1,$D3,$6B,$FD,$13
	.dc.b	$37,$E0,$BF,$D0,$D3,$72,$FF,$1B,$D7,$A3,$DD,$FF,$29,$11,$FF,$7A
	.dc.b	$FE,$CB,$FE,$29,$BF,$F4,$74,$E6,$7F,$D7,$0F,$F9,$6E,$74,$7F,$8B
	.dc.b	$61,$97,$E0,$69,$39,$8C,$37,$2F,$E7,$F3,$AF,$73,$A7,$BB,$30,$F1
	.dc.b	$BD,$E3,$11,$DA,$B8,$DF,$38,$5F,$82,$F8,$47,$B3,$DD,$F4,$11,$F7
	.dc.b	$FB,$B5,$E8,$72,$BF,$F5,$4E,$5E,$8A,$F8,$15,$B8,$C5,$75,$39,$94
	.dc.b	$E7,$87,$C2,$AE,$18,$FA,$2F,$FC,$0F,$F0,$DF,$E0,$6C,$3F,$B0,$FC
	.dc.b	$EB,$97,$F4,$92,$F7,$7F,$4D,$39,$CF,$FB,$3C,$E5,$E3,$5C,$3D,$5F
	.dc.b	$A7,$0A,$E0,$AB,$C0,$E6,$BB,$DF,$66,$C7,$43,$BD,$F4,$BE,$77,$C9
	.dc.b	$B7,$C6,$BC,$AD,$7C,$15,$C6,$FC,$D8,$75,$9F,$25,$6F,$0F,$64,$79
	.dc.b	$4F,$0C,$EE,$3C,$15,$E6,$9D,$1C,$8D,$5F,$77,$89,$FE,$D0,$DB,$C2
	.dc.b	$D3,$3E,$8F,$E6,$EF,$E2,$BF,$9D,$8F,$F7,$61,$B3,$FD,$A4,$53,$2E
	.dc.b	$D7,$C3,$7F,$87,$23,$0F,$BE,$3F,$DE,$5A,$D7,$67,$A8,$C9,$4C,$E9
	.dc.b	$F6,$9C,$B5,$CE,$5F,$E3,$BF,$4B,$F9,$94,$E6,$FC,$11,$89,$FF,$1B
	.dc.b	$46,$2F,$C4,$DB,$F0,$36,$F9,$47,$96,$F9,$3F,$E4,$99,$D5,$FE,$D3
	.dc.b	$E0,$3F,$A1,$6B,$FC,$01,$B8,$FE,$E3,$E0,$2E,$1F,$AF,$9B,$C5,$FC
	.dc.b	$21,$BC,$66,$D6,$19,$7F,$C6,$1C,$B4,$97,$FB,$2A,$E5,$1F,$C7,$AB
	.dc.b	$8C,$E7,$87,$B9,$E0,$E1,$B5,$FC,$92,$2C,$F3,$9B,$97,$F1,$C9,$B5
	.dc.b	$FC,$52,$6D,$7F,$0B,$F6,$3D,$67,$B3,$DF,$B5,$FC,$C2,$2C,$F9,$E6
	.dc.b	$E1,$FC,$A3,$E0,$3E,$19,$F1,$3E,$02,$E1,$FC,$E2,$2D,$FB,$E6,$D7
	.dc.b	$1A,$D1,$FC,$FA,$32,$FF,$2A,$AB,$5F,$C8,$66,$ED,$FC,$EB,$E0,$3E
	.dc.b	$99,$F5,$3D,$FB,$87,$F4,$88,$B7,$F3,$E6,$E3,$FE,$6A,$F9,$9F,$BB
	.dc.b	$E8,$1B,$E8,$E1,$98,$FF,$9E,$1C,$B4,$FF,$3D,$FA,$9C,$8E,$7E,$B5
	.dc.b	$8C,$93,$2F,$F4,$DC,$65,$FF,$54,$FE,$6C,$F8,$FF,$C8,$C2,$31,$5F
	.dc.b	$E9,$2C,$FF,$C3,$5D,$4F,$B6,$78,$BC,$2F,$78,$74,$C7,$07,$E7,$2F
	.dc.b	$85,$63,$1F,$F6,$07,$ED,$3F,$E5,$16,$EF,$BF,$A0,$7C,$CF,$92,$B5
	.dc.b	$97,$FB,$01,$BA,$FF,$8A,$9B,$87,$FA,$A3,$26,$1C,$A3,$A9,$EB,$3F
	.dc.b	$D5,$A3,$E8,$1C,$FF,$B2,$8C,$36,$1C,$F6,$24,$F3,$CB,$FC,$93,$5C
	.dc.b	$A3,$07,$5A,$DC,$FF,$6C,$8E,$A6,$A7,$84,$65,$EF,$19,$AF,$AB,$E3
	.dc.b	$BC,$AA,$7C,$DF,$A0,$7C,$76,$2F,$FD,$D9,$5B,$6D,$6E,$75,$8F,$FB
	.dc.b	$AB,$EC,$BB,$B7,$3F,$E2,$7F,$87,$3F,$0E,$7C,$B7,$E1,$AF,$7C,$5F
	.dc.b	$E5,$A7,$8E,$1F,$EE,$96,$BF,$E5,$B6,$BF,$ED,$C6,$FC,$FF,$FB,$7B
	.dc.b	$DC,$FE,$42,$DD,$7E,$63,$FE,$E2,$F9,$EC,$DF,$F9,$FE,$72,$D3,$FC
	.dc.b	$2F,$3C,$5A,$27,$8D,$4A,$E2,$F4,$8C,$4F,$FC,$00,$FA,$1F,$DA,$FD
	.dc.b	$17,$ED,$9B,$7F,$EE,$A6,$CE,$E3,$FD,$DB,$3F,$AA,$FA,$2F,$F6,$64
	.dc.b	$7E,$16,$51,$87,$FC,$5B,$CE,$E4,$7A,$BC,$7C,$09,$87,$1C,$38,$6F
	.dc.b	$FB,$2B,$DC,$E6,$BF,$85,$F5,$FF,$0C,$DA,$CF,$FC,$00,$6E,$7F,$E3
	.dc.b	$2E,$CF,$91,$8F,$FB,$F2,$F8,$5F,$E6,$FC,$6F,$F6,$24,$62,$F8,$36
	.dc.b	$BF,$D3,$39,$BE,$83,$E0,$1F,$D6,$BE,$8D,$F5,$2E,$CC,$E7,$EB,$AF
	.dc.b	$F5,$0E,$5F,$E9,$FC,$39,$4F,$FA,$86,$31,$3A,$47,$CA,$71,$9F,$FB
	.dc.b	$67,$93,$F9,$5F,$D2,$7F,$2C,$6D,$FA,$F6,$FC,$13,$67,$4D,$3F,$1F
	.dc.b	$F3,$CF,$8B,$FC,$9B,$63,$3F,$21,$89,$91,$C1,$DC,$63,$3F,$C9,$D7
	.dc.b	$16,$CE,$AB,$81,$E5,$CF,$FA,$19,$FD,$13,$E9,$2C,$7A,$7E,$93,$FA
	.dc.b	$39,$FE,$04,$FC,$BF,$FD,$AF,$19,$3E,$C3,$FC,$CB,$8F,$FD,$20,$BE
	.dc.b	$93,$87,$95,$C5,$E5,$F2,$32,$F8,$FD,$15,$D7,$FC,$EF,$83,$E9,$3C
	.dc.b	$06,$72,$EA,$E2,$FC,$3F,$19,$6F,$8B,$E7,$F1,$24,$F9,$B8,$2D,$76
	.dc.b	$9B,$87,$A2,$6D,$7F,$A9,$31,$FE,$22,$FB,$AB,$E1,$4E,$AF,$FD,$3F
	.dc.b	$4E,$2F,$8C,$7C,$7F,$99,$B8,$7C,$68,$B5,$EE,$9B,$8F,$C8,$9B,$5E
	.dc.b	$19,$B5,$DB,$E8,$FE,$91,$F4,$FC,$AB,$5F,$4A,$2C,$FC,$A9,$B8,$7D
	.dc.b	$49,$B5,$ED,$9B,$5E,$AE,$7F,$BC,$7D,$FF,$A3,$B5,$F7,$A2,$CF,$D6
	.dc.b	$9B,$87,$E0,$26,$D7,$C9,$9B,$5F,$1B,$E8,$FF,$14,$7E,$2F,$E2,$ED
	.dc.b	$7E,$2A,$2C,$FE,$0A,$6E,$1F,$8C,$9B,$5F,$56,$6D,$7D,$2F,$8B,$FC
	.dc.b	$B1,$F9,$7F,$52,$D7,$E5,$A2,$CF,$E3,$A6,$E1,$FE,$E4,$FA,$1F,$F2
	.dc.b	$F9,$FE,$60,$F9,$0B,$87,$FB,$D2,$2D,$FF,$96,$E6,$D7,$FB,$B3,$E5
	.dc.b	$BF,$CC,$E7,$F9,$A3,$E8,$6E,$1F,$EF,$C8,$B7,$FE,$63,$9B,$5F,$EF
	.dc.b	$4F,$A1,$FF,$37,$9F,$E7,$0F,$A0,$B8,$7F,$25,$F8,$74,$E3,$0F,$F3
	.dc.b	$65,$5B,$E5,$7B,$AE,$4D,$3F,$97,$7E,$9B,$FE,$12,$BF,$F5,$2F,$C8
	.dc.b	$FF,$1E,$F1,$D3,$FE,$A0,$F8,$C4,$E6,$38,$46,$2B,$FD,$FF,$1B,$9D
	.dc.b	$42,$FF,$94,$D6,$EF,$AF,$A5,$9A,$F0,$9C,$8C,$4E,$F0,$E9,$1F,$CC
	.dc.b	$BE,$95,$E1,$2F,$F9,$C4,$E6,$30,$E2,$77,$3A,$84,$61,$E3,$F2,$0F
	.dc.b	$89,$79,$5B,$8C,$36,$33,$3C,$1A,$FF,$49,$A7,$8A,$F9,$0F,$04,$61
	.dc.b	$B1,$99,$F3,$35,$FE,$9E,$35,$7C,$9F,$F4,$58,$F9,$7C,$1F,$5E,$3F
	.dc.b	$88,$46,$0E,$0E,$1F,$9B,$3C,$BC,$4E,$BC,$1F,$4E,$BA,$56,$9E,$11
	.dc.b	$9F,$0F,$78,$77,$BB,$7D,$BC,$3F,$3F,$2F,$DB,$CB,$CF,$18,$E4,$F3
	.dc.b	$36,$BC,$BE,$1D,$AD,$C7,$F8,$07,$45,$65,$B1,$9A,$C4,$C8,$E0,$EE
	.dc.b	$38,$78,$72,$7A,$3C,$FC,$3C,$43,$2F,$46,$1F,$E5,$A8,$D9,$E3,$9F
	.dc.b	$17,$87,$23,$CC,$F9,$7C,$3C,$47,$3D,$3E,$A4,$E3,$0E,$B5,$EF,$FA
	.dc.b	$5F,$F3,$55,$FF,$87,$DA,$F1,$3F,$00,$E2,$7E,$63,$FE,$36,$8E,$1F
	.dc.b	$01,$CC,$8F,$3F,$67,$E6,$7D,$2F,$E2,$EF,$EA,$37,$2F,$DF,$36,$BF
	.dc.b	$1C,$BE,$53,$8F,$3C,$9F,$F1,$A4,$EF,$88,$EE,$33,$3B,$7D,$53,$A8
	.dc.b	$F1,$D7,$93,$D5,$95,$D1,$F3,$F2,$E2,$38,$E1,$FE,$39,$74,$78,$A3
	.dc.b	$59,$7B,$2A,$E5,$EC,$9F,$83,$F2,$B2,$FD,$2F,$C4,$F9,$3C,$47,$68
	.dc.b	$B8,$7C,$69,$B7,$F1,$A6,$D7,$C6,$9F,$9B,$F2,$72,$3E,$2B,$F4,$3E
	.dc.b	$4F,$11,$EA,$8B,$87,$D2,$9B,$87,$D2,$9B,$5F,$4A,$7E,$CF,$C9,$C8
	.dc.b	$FA,$2F,$DC,$EF,$C4,$7C,$68,$B8,$7D,$E9,$B8,$7D,$E9,$B5,$F7,$A7
	.dc.b	$F0,$BF,$27,$F8,$6F,$89,$93,$F7,$A3,$2D,$21,$F5,$3E,$27,$F1,$A4
	.dc.b	$7F,$0A,$FE,$43,$E4,$FF,$23,$F2,$7F,$92,$F4,$25,$7E,$2A,$3F,$2B
	.dc.b	$E8,$48,$FC,$6B,$F9,$8F,$43,$F3,$3E,$84,$9F,$C8,$BF,$9C,$F3,$BF
	.dc.b	$85,$43,$E9,$5B,$E1,$B1,$FE,$A1,$6B,$FA,$D3,$86,$FB,$99,$1B,$79
	.dc.b	$46,$23,$84,$F2,$E1,$FE,$B1,$1E,$1A,$DB,$FA,$6F,$43,$F8,$E7,$85
	.dc.b	$FE,$8B,$AF,$E4,$8D,$E3,$FD,$8C,$FE,$D8,$DD,$43,$91,$FE,$B9,$FA
	.dc.b	$4F,$E4,$2F,$8B,$AB,$DC,$7C,$28,$CE,$1E,$47,$73,$FE,$9B,$6B,$FD
	.dc.b	$31,$AF,$FA,$71,$BD,$CC,$7F,$B4,$CF,$FA,$DF,$E9,$7D,$67,$B3,$D0
	.dc.b	$91,$FE,$9D,$8C,$BF,$E9,$F8,$CB,$C4,$D1,$93,$D9,$EA,$FD,$AF,$A5
	.dc.b	$79,$54,$F6,$44,$9D,$90,$FB,$53,$FD,$77,$3F,$9B,$F4,$BF,$43,$E8
	.dc.b	$65,$7F,$AC,$63,$E9,$98,$7F,$DA,$51,$87,$FD,$CD,$F4,$AF,$98,$FC
	.dc.b	$6B,$63,$31,$8F,$8C,$CB,$23,$27,$28,$C3,$F8,$F9,$E4,$D9,$E4,$FF
	.dc.b	$BD,$7D,$87,$57,$E7,$3D,$27,$5D,$3F,$3F,$3A,$AE,$15,$FE,$1B,$7F
	.dc.b	$C4,$6D,$8D,$47,$F8,$91,$AF,$FB,$77,$E9,$5F,$37,$FA,$AF,$A5,$7A
	.dc.b	$34,$FD,$6C,$EA,$BF,$DC,$3C,$1F,$A0,$8C,$47,$EA,$DB,$19,$BE,$9C
	.dc.b	$DF,$B6,$FA,$57,$83,$49,$87,$4E,$DF,$F0,$FC,$FA,$30,$FD,$6B,$C2
	.dc.b	$74,$7F,$BC,$9C,$3F,$BB,$6C,$70,$D7,$AC,$E6,$EC,$E6,$39,$8E,$37
	.dc.b	$C8,$7F,$0C,$FA,$5F,$EA,$13,$FC,$1D,$B5,$FC,$02,$B7,$3C,$9A,$FF
	.dc.b	$0D,$6C,$6E,$BA,$7A,$3F,$C6,$3E,$91,$F0,$DF,$C6,$BE,$97,$EF,$E3
	.dc.b	$E4,$6B,$FC,$5E,$76,$FA,$1A,$E6,$7A,$1C,$F3,$FE,$23,$F4,$B2,$7E
	.dc.b	$4F,$83,$FB,$D8,$FE,$2B,$F4,$59,$9F,$EE,$FC,$49,$C6,$BF,$C6,$2B
	.dc.b	$F8,$DF,$D2,$CB,$E7,$68,$FE,$3F,$F4,$B9,$9E,$71,$6F,$F9,$0C,$FF
	.dc.b	$8E,$FE,$26,$57,$F8,$FD,$FF,$1F,$78,$58,$8F,$F0,$4F,$BC,$9C,$5F
	.dc.b	$F8,$DA,$BF,$C8,$9F,$13,$2F,$FC,$90,$FF,$91,$FD,$DC,$47,$28,$B5
	.dc.b	$FE,$3E,$9B,$D3,$FE,$69,$9B,$5F,$E6,$89,$BB,$FF,$9C,$26,$E5,$FE
	.dc.b	$33,$BF,$F3,$87,$BB,$FE,$55,$26,$1F,$EB,$5E,$5A,$FA,$A2,$E5,$FE
	.dc.b	$65,$F8,$9F,$F1,$51,$FE,$29,$9F,$F1,$3B,$63,$5E,$F7,$F9,$96,$3F
	.dc.b	$23,$87,$FA,$D5,$D3,$63,$46,$FC,$DB,$7F,$EC,$78,$AF,$68,$E9,$EE
	.dc.b	$C8,$FF,$66,$3D,$FD,$DC,$47,$92,$2E,$3E,$99,$B8,$78,$E6,$D7,$FA
	.dc.b	$5F,$15,$63,$64,$7F,$B1,$E7,$A1,$FE,$DB,$9F,$EF,$7F,$11,$FD,$EB
	.dc.b	$14,$BE,$95,$CF,$FE,$A4,$A7,$BB,$FE,$A6,$9C,$D7,$09,$CB,$CA,$32
	.dc.b	$A3,$B3,$D6,$BA,$7A,$9F,$E8,$38,$C1,$CE,$FB,$4F,$85,$B1,$DE,$78
	.dc.b	$5F,$6B,$EA,$DA,$DD,$78,$5B,$3A,$BF,$0E,$1E,$77,$B5,$53,$C4,$53
	.dc.b	$D5,$87,$82,$38,$C7,$A0,$EC,$6E,$BE,$1B,$97,$C0,$FA,$6B,$E7,$F0
	.dc.b	$FE,$4E,$7E,$2B,$6F,$E2,$36,$77,$C1,$FE,$F0,$F8,$6B,$87,$FB,$60
	.dc.b	$38,$7F,$D9,$E7,$FB,$8E,$38,$34,$E1,$1E,$D7,$FD,$F9,$C1,$5E,$67
	.dc.b	$B4,$C9,$4C,$E9,$F7,$9C,$B5,$CE,$7E,$5F,$49,$5D,$6F,$C1,$18,$AE
	.dc.b	$91,$89,$D4,$78,$67,$57,$D6,$B5,$3C,$0D,$57,$09,$F2,$57,$0B,$FD
	.dc.b	$D7,$C0,$77,$CB,$F8,$23,$6F,$F7,$DF,$01,$71,$FD,$64,$DE,$2F,$D5
	.dc.b	$B7,$88,$DA,$C3,$2F,$FA,$0C,$E9,$B3,$9B,$FE,$23,$57,$0F,$E2,$D3
	.dc.b	$6B,$F8,$87,$0F,$73,$C1,$C1,$63,$AC,$5C,$3F,$8D,$CD,$BF,$E4,$13
	.dc.b	$6B,$F8,$D4,$DA,$FE,$11,$EC,$FA,$CF,$67,$BF,$6B,$D3,$16,$7F,$93
	.dc.b	$CD,$C3,$F9,$67,$C0,$7C,$33,$E2,$7C,$05,$C3,$E0,$C5,$BF,$E5,$F3
	.dc.b	$6B,$8D,$68,$F8,$F1,$97,$F9,$85,$5A,$FE,$3D,$37,$6F,$E3,$BF,$01
	.dc.b	$F4,$CF,$A9,$EF,$DC,$3E,$8C,$5B,$FE,$7F,$37,$1F,$F4,$2F,$D2,$FD
	.dc.b	$DF,$10,$DF,$3F,$0C,$C7,$FB,$8B,$96,$9F,$DD,$3D,$6E,$47,$3F,$62
	.dc.b	$C6,$49,$97,$F8,$02,$32,$FF,$A9,$7E,$91,$F2,$3F,$91,$84,$62,$BF
	.dc.b	$C6,$7D,$2F,$A5,$78,$BB,$AD,$5F,$3C,$3F,$35,$6F,$CC,$5F,$01,$B9
	.dc.b	$CC,$63,$2F,$F1,$AF,$D3,$7F,$A4,$DB,$C1,$CB,$E9,$9F,$3D,$E5,$AD
	.dc.b	$67,$FE,$BD,$6F,$1F,$EE,$F4,$5F,$F8,$FB,$1E,$67,$A9,$C5,$AB,$F8
	.dc.b	$14,$BE,$99,$D3,$FB,$38,$C3,$61,$D3,$62,$4F,$F9,$60,$D9,$67,$6D
	.dc.b	$8D,$DF,$86,$3F,$B0,$9C,$1C,$1F,$C6,$27,$8B,$88,$D5,$FF,$B5,$3C
	.dc.b	$97,$99,$F6,$5B,$3F,$63,$E8,$1F,$1D,$9A,$93,$9E,$0B,$8F,$83,$66
	.dc.b	$1F,$32,$BF,$BA,$CB,$E9,$3C,$6F,$F4,$16,$BA,$6B,$FD,$F1,$AF,$1D
	.dc.b	$BF,$8F,$7D,$0B,$ED,$FF,$C1,$37,$FE,$76,$BF,$7E,$DF,$CD,$6B,$5E
	.dc.b	$77,$F2,$5F,$A5,$FF,$03,$FB,$97,$3F,$B7,$7F,$CE,$E7,$FC,$17,$F4
	.dc.b	$39,$A9,$97,$FA,$32,$78,$B6,$D5,$A7,$E8,$CE,$9A,$76,$9F,$F0,$BC
	.dc.b	$76,$79,$57,$EE,$FE,$97,$F7,$9F,$4B,$67,$FC,$D2,$DB,$FF,$38,$B7
	.dc.b	$1F,$EA,$5E,$37,$FB,$CE,$3F,$65,$F4,$8F,$0A,$9F,$AC,$E2,$7A,$D8
	.dc.b	$CB,$FD,$3B,$5B,$7F,$6F,$F8,$AB,$D5,$D6,$9F,$76,$31,$3F,$E7,$CE
	.dc.b	$FF,$E2,$5B,$FF,$13,$6C,$64,$FF,$40,$C6,$5A,$66,$BF,$D1,$71,$96
	.dc.b	$C7,$93,$E9,$3F,$B7,$47,$F8,$1F,$14,$62,$3F,$1B,$18,$9F,$F0,$64
	.dc.b	$7F,$91,$23,$0E,$1B,$BF,$F6,$BF,$86,$FE,$27,$5D,$CF,$F6,$4C,$23
	.dc.b	$35,$FE,$A0,$E3,$FE,$61,$F4,$BF,$CC,$5B,$7F,$E9,$E6,$CE,$E3,$FD
	.dc.b	$43,$3F,$91,$F9,$E7,$C7,$53,$1F,$39,$6B,$F1,$D3,$A6,$CE,$9A,$4B
	.dc.b	$FC,$7C,$ED,$B5,$CD,$FF,$5B,$E9,$FD,$1C,$FE,$91,$F4,$B6,$FF,$D6
	.dc.b	$BF,$4B,$FD,$2C,$FF,$04,$7D,$17,$F8,$82,$32,$7F,$B1,$1F,$E6,$FB
	.dc.b	$7E,$41,$7D,$2F,$0F,$2F,$8B,$CB,$E4,$A5,$F1,$FA,$1B,$A7,$F1,$6E
	.dc.b	$C6,$2F,$A5,$F0,$9A,$CF,$B3,$8C,$3C,$5F,$19,$70,$E5,$7F,$EE,$69
	.dc.b	$F3,$7D,$2F,$A8,$F5,$F1,$D8,$C9,$2B,$8C,$65,$E7,$37,$1F,$74,$DA
	.dc.b	$FF,$47,$E5,$E8,$E2,$F8,$E7,$C8,$FA,$3B,$3F,$1E,$2E,$3E,$69,$B8
	.dc.b	$7C,$99,$B5,$D3,$25,$62,$D8,$7C,$6F,$9B,$FA,$A7,$D6,$F3,$2E,$1F
	.dc.b	$56,$2D,$7C,$49,$B9,$7D,$79,$B5,$DE,$6D,$76,$C7,$F9,$C3,$7C,$BF
	.dc.b	$C1,$FD,$1F,$E1,$19,$7F,$83,$72,$D2,$4F,$D4,$8C,$B6,$3F,$15,$F2
	.dc.b	$FF,$8E,$3F,$1F,$F1,$B6,$32,$49,$FB,$B1,$97,$EF,$CD,$CF,$F2,$53
	.dc.b	$6B,$E5,$7C,$6F,$E6,$8F,$CD,$FB,$36,$3F,$35,$17,$0F,$C4,$CD,$BF
	.dc.b	$F7,$87,$D0,$FF,$99,$CF,$F3,$47,$C9,$5C,$3F,$CC,$B1,$6F,$FD,$E7
	.dc.b	$36,$BF,$DF,$1F,$2F,$FE,$6F,$3F,$CE,$1F,$35,$70,$FF,$36,$C5,$BF
	.dc.b	$F7,$DC,$DA,$FF,$80,$1F,$43,$FE,$77,$3F,$CF,$1F,$41,$70,$FF,$3A
	.dc.b	$C5,$BF,$D5,$7C,$62,$B1,$7C,$AF,$67,$F9,$F2,$32,$FE,$6B,$E9,$7F
	.dc.b	$C1,$37,$FE,$B2,$F9,$3C,$7B,$3F,$E9,$DF,$91,$46,$E7,$0E,$AF,$FD
	.dc.b	$E3,$3A,$9F,$CB,$DF,$48,$C7,$4A,$3F,$D6,$71,$87,$57,$CE,$BC,$47
	.dc.b	$FB,$F3,$E9,$5E,$13,$2F,$E7,$B2,$9C,$C6,$1B,$1F,$D0,$9A,$FF,$B3
	.dc.b	$FE,$51,$F1,$AF,$2B,$9C,$61,$B1,$99,$FE,$97,$39,$7A,$78,$BF,$D9
	.dc.b	$5B,$E9,$78,$63,$11,$FE,$DD,$C1,$18,$8F,$F7,$4F,$ED,$CF,$8B,$FF
	.dc.b	$12,$EF,$F2,$70,$4E,$6B,$E6,$ED,$C2,$F6,$FD,$9F,$87,$B3,$FD,$DE
	.dc.b	$B8,$3F,$2A,$F9,$DE,$9B,$1D,$0F,$AF,$5C,$97,$EA,$6F,$77,$FC,$73
	.dc.b	$CD,$FD,$CA,$31,$23,$FC,$9E,$70,$9D,$F2,$FE,$37,$E8,$AC,$EE,$B8
	.dc.b	$BC,$1B,$7D,$6F,$7C,$DF,$87,$BE,$37,$D4,$F0,$56,$DB,$87,$F9,$3B
	.dc.b	$0E,$6F,$86,$32,$3D,$F7,$ED,$AF,$9A,$F2,$C6,$1B,$19,$9E,$39,$7E
	.dc.b	$67,$E7,$F3,$52,$7A,$DF,$17,$7D,$4A,$F5,$AF,$9D,$F8,$79,$7F,$A0
	.dc.b	$BE,$66,$FC,$39,$DC,$CC,$3F,$D0,$93,$96,$CE,$5F,$A3,$7F,$9C,$7D
	.dc.b	$E7,$EB,$FA,$BE,$4C,$66,$3F,$B8,$7E,$92,$F9,$CC,$D4,$9E,$35,$BE
	.dc.b	$1F,$F2,$6D,$EE,$7A,$FC,$EF,$EC,$DF,$8B,$0F,$8C,$FB,$2D,$8D,$F8
	.dc.b	$1F,$6A,$B8,$47,$E2,$1B,$7F,$E3,$E6,$ED,$FD,$BA,$FF,$86,$56,$EF
	.dc.b	$91,$1F,$0F,$5F,$D8,$F8,$1F,$0F,$44,$9F,$AB,$CC,$9D,$46,$1A,$FC
	.dc.b	$4F,$9C,$BA,$7F,$A2,$6B,$84,$EC,$F7,$CF,$E1,$BD,$DF,$7F,$CE,$49
	.dc.b	$EF,$19,$69,$0F,$A3,$F3,$99,$7F,$D6,$FD,$1F,$94,$9C,$63,$F4,$2B
	.dc.b	$F2,$BF,$9C,$FF,$22,$BF,$9F,$F3,$BF,$38,$CF,$91,$23,$F2,$73,$B7
	.dc.b	$AB,$6F,$ED,$F3,$6E,$73,$56,$73,$1C,$BD,$AF,$B7,$F3,$B2,$30,$43
	.dc.b	$AC,$79,$8F,$C0,$FC,$A5,$8D,$57,$FA,$CF,$15,$6A,$7F,$13,$D9,$F8
	.dc.b	$82,$3E,$AF,$5B,$E0,$91,$C1,$F9,$0F,$9D,$FC,$79,$97,$DF,$18,$7F
	.dc.b	$2B,$F3,$BF,$94,$F5,$64,$60,$FD,$CF,$DD,$7F,$38,$CF,$FD,$90,$DE
	.dc.b	$CF,$E3,$DC,$88,$C5,$7F,$8D,$3B,$1E,$B7,$E8,$3D,$6F,$D0,$B2,$7F
	.dc.b	$4B,$1F,$C6,$BE,$A3,$E9,$63,$FE,$E4,$FA,$96,$BF,$C7,$9B,$7F,$BA
	.dc.b	$AD,$BF,$CB,$DA,$9A,$62,$1C,$1D,$F9,$65,$73,$7B,$47,$ED,$BD,$89
	.dc.b	$3F,$BA,$7F,$73,$EC,$62,$3F,$83,$C5,$E0,$FE,$07,$3F,$CF,$3C,$27
	.dc.b	$ED,$28,$C3,$6F,$F9,$8C,$EA,$73,$EC,$69,$93,$AA,$E0,$AD,$B7,$F4
	.dc.b	$FC,$3B,$E1,$F2,$FE,$0B,$FC,$07,$7F,$0D,$BB,$7F,$9A,$A3,$FC,$0E
	.dc.b	$D7,$FD,$8A,$DC,$3F,$D9,$AD,$C3,$FA,$E4,$F3,$3F,$95,$ED,$F8,$07
	.dc.b	$07,$F6,$09,$E0,$FD,$5B,$FB,$5E,$97,$F4,$CE,$47,$75,$25,$7F,$28
	.dc.b	$CE,$BF,$C0,$A7,$6F,$E5,$27,$F9,$5C,$FF,$64,$FD,$D7,$CB,$3F,$EA
	.dc.b	$96,$E9,$F1,$A3,$FD,$0F,$18,$D3,$FD,$2E,$FF,$2C,$8C,$9C,$C8,$7F
	.dc.b	$5E,$74,$FF,$8C,$3C,$7F,$EE,$3F,$54,$F8,$BA,$7F,$74,$E5,$FE,$6B
	.dc.b	$8F,$FB,$D9,$B8,$FC,$A6,$E1,$FE,$07,$6A,$BF,$F7,$DB,$FD,$62,$32
	.dc.b	$4C,$3C,$3F,$19,$F0,$7D,$69,$5E,$77,$E2,$7A,$D8,$8F,$89,$17,$2F
	.dc.b	$89,$EA,$B5,$5F,$F8,$33,$EA,$6E,$1F,$DA,$1A,$FF,$87,$1B,$7F,$36
	.dc.b	$3F,$DF,$3E,$77,$1A,$91,$C5,$FF,$0B,$FC,$47,$F8,$4A,$3F,$C4,$DF
	.dc.b	$87,$7F,$B9,$9F,$EE,$BF,$87,$BE,$6B,$FC,$3A,$D8,$FE,$F3,$3B,$7F
	.dc.b	$C5,$AD,$C7,$C4,$DE,$2F,$CB,$B6,$7F,$C6,$94,$FA,$5F,$F1,$4C,$E6
	.dc.b	$4D,$31,$3B,$8F,$F0,$4F,$B4,$E0,$FF,$18,$B6,$33,$1F,$E3,$4F,$BB
	.dc.b	$F8,$B8,$FE,$A2,$B5,$5F,$E4,$46,$E5,$F2,$5B,$3F,$AE,$AE,$35,$FE
	.dc.b	$4B,$FC,$E3,$F3,$0D,$3F,$C8,$2E,$17,$EB,$BF,$B3,$8C,$12,$3F,$C8
	.dc.b	$F1,$FC,$DB,$C9,$FE,$71,$E3,$7E,$CC,$93,$F9,$FC,$6B,$FB,$18,$B9
	.dc.b	$7E,$23,$93,$FC,$75,$3F,$A5,$F7,$7F,$65,$72,$30,$43,$F6,$8E,$5F
	.dc.b	$DB,$47,$F0,$3F,$1A,$47,$F5,$07,$FC,$8B,$EE,$E2,$3F,$81,$45,$C3
	.dc.b	$F3,$3E,$E7,$F9,$06,$7F,$83,$77,$FF,$0D,$23,$FE,$50,$FA,$BA,$E8
	.dc.b	$FE,$BB,$1F,$E5,$28,$CD,$7F,$5E,$8B,$C3,$FE,$C3,$9B,$5F,$E6,$7F
	.dc.b	$0B,$FD,$98,$F2,$3B,$1F,$E6,$CF,$0D,$FA,$63,$D1,$FF,$3F,$C6,$1B
	.dc.b	$19,$9F,$F3,$2F,$C0,$62,$64,$73,$9D,$3F,$D0,$71,$FF,$6B,$57,$F9
	.dc.b	$DB,$F3,$17,$C8,$FF,$A1,$BF,$24,$F2,$DE,$C7,$86,$8C,$EF,$FC,$97
	.dc.b	$9D,$87,$2B,$FF,$3C,$B5,$FF,$19,$B6,$78,$37,$2F,$F5,$06,$0F,$C8
	.dc.b	$3F,$2F,$87,$4C,$BC,$C4,$C7,$FD,$13,$3A,$FD,$A3,$F5,$D1,$E7,$B7
	.dc.b	$E2,$1F,$8E,$8C,$37,$3F,$45,$70,$6F,$83,$FE,$98,$8C,$67,$FE,$16
	.dc.b	$6D,$FC,$2C,$79,$4E,$6B,$81,$27,$CB,$D3,$5F,$F2,$9D,$F2,$D0,$F4
	.dc.b	$CF,$E0,$57,$FA,$D3,$E3,$7F,$AE,$E7,$FE,$C3,$FB,$1F,$CB,$43,$E0
	.dc.b	$3F,$B5,$C6,$1B,$D7,$FD,$BA,$78,$E6,$AE,$3E,$7F,$F8,$06,$31,$5B
	.dc.b	$BE,$31,$FE,$D1,$BE,$CF,$86,$7A,$57,$1B,$E7,$86,$6B,$86,$1C,$96
	.dc.b	$1C,$9E,$E7,$5B,$E5,$87,$4B,$E5,$3C,$2F,$95,$F1,$C3,$BC,$F2,$3A
	.dc.b	$E1,$C9,$F0,$9C,$A7,$86,$1D,$63,$84,$7C,$4E,$7F,$F5,$84,$72,$C7
	.dc.b	$FD,$E1,$18,$9D,$7C,$47,$18,$DD,$7F,$BD,$7E,$33,$73,$FE,$08,$84
	.dc.b	$61,$FF,$05,$B5,$FF,$10,$B5,$E4,$7A,$C7,$AB,$FE,$0F,$CF,$F8,$A4
	.dc.b	$69,$B1,$9C,$BC,$15,$F2,$2B,$8F,$15,$7C,$67,$5A,$E0,$F9,$0C,$46
	.dc.b	$E7,$94,$EA,$7C,$E6,$8F,$31,$D2,$FC,$75,$C6,$35,$1E,$87,$57,$FE
	.dc.b	$2C,$AF,$0F,$0F,$E8,$3E,$0F,$FC,$59,$AF,$EC,$9B,$8F,$F0,$2E,$4F
	.dc.b	$F7,$E2,$C7,$F5,$33,$A5,$7C,$E7,$57,$CD,$FE,$0B,$E0,$EE,$75,$5C
	.dc.b	$A7,$67,$EF,$6F,$A7,$1F,$F1,$1B,$E1,$C8,$9A,$6B,$9B,$FC,$05,$7C
	.dc.b	$B0,$F1,$D6,$E3,$18,$7F,$12,$E6,$F1,$7E,$D0,$F8,$6A,$78,$36,$7D
	.dc.b	$96,$30,$E9,$B2,$F8,$46,$1B,$9F,$F1,$FB,$E3,$39,$C3,$94,$E7,$E3
	.dc.b	$6F,$07,$F8,$E2,$D1,$F2,$F9,$BF,$A7,$F8,$F7,$C0,$FF,$54,$E5,$EA
	.dc.b	$F1,$7A,$E5,$FD,$0E,$F7,$3F,$D3,$6F,$81,$F2,$AB,$79,$7F,$9A,$6B
	.dc.b	$77,$FE,$F4,$F0,$5D,$DF,$55,$B8,$F7,$CF,$13,$23,$E1,$CE,$AB,$F9
	.dc.b	$F6,$3E,$A9,$CD,$6E,$7B,$3E,$3B,$C6,$3F,$D5,$F0,$E4,$D9,$D5,$F2
	.dc.b	$3B,$F3,$79,$4F,$67,$15,$8C,$93,$1F,$1C,$65,$FF,$45,$74,$7E,$33
	.dc.b	$C7,$78,$B6,$30,$DB,$FF,$43,$36,$BF,$D2,$4F,$C5,$E5,$93,$A3,$E1
	.dc.b	$B8,$9F,$98,$F4,$8C,$9F,$95,$87,$CF,$7F,$AA,$23,$0D,$8F,$EF,$DD
	.dc.b	$6F,$96,$7A,$7F,$AF,$9E,$3E,$15,$BE,$1C,$1F,$E7,$96,$E3,$FE,$CB
	.dc.b	$C3,$8B,$DA,$B9,$7E,$18,$F8,$BA,$7E,$EF,$37,$D7,$BD,$8C,$E1,$E9
	.dc.b	$BE,$0D,$BE,$15,$B8,$C1,$EF,$F9,$07,$CF,$FA,$5B,$1F,$EE,$8F,$A8
	.dc.b	$7B,$39,$A9,$3C,$5F,$24,$EA,$30,$4E,$0F,$8F,$0F,$8B,$E3,$18,$7E
	.dc.b	$47,$D6,$7E,$DB,$8B,$FD,$3C,$DB,$FF,$78,$5B,$E7,$77,$E1,$C9,$FB
	.dc.b	$15,$BC,$7F,$A2,$DE,$E7,$35,$F8,$DE,$D9,$1F,$EC,$FE,$7B,$7F,$ED
	.dc.b	$B9,$B9,$7F,$9C,$66,$D7,$FB,$D3,$E9,$BF,$97,$DF,$F8,$FC,$38,$AF
	.dc.b	$F7,$17,$23,$27,$CE,$7E,$16,$B5,$87,$3E,$EC,$1F,$2A,$6C,$7E,$4E
	.dc.b	$77,$8F,$B3,$EB,$B2,$FC,$C7,$F5,$1E,$0C,$AF,$F7,$57,$81,$6B,$FD
	.dc.b	$ED,$37,$4F,$F4,$64,$DA,$FE,$91,$E4,$37,$7D,$2F,$F0,$F3,$F6,$9F
	.dc.b	$57,$6B,$F8,$6C,$5C,$BF,$C9,$5E,$1A,$B1,$97,$FB,$5F,$5F,$C8,$57
	.dc.b	$C7,$6C,$7F,$98,$FE,$ED,$D5,$F1,$7B,$A6,$7F,$05,$FE,$A9,$3A,$8C
	.dc.b	$13,$0F,$F0,$66,$1D,$9F,$EC,$47,$F6,$37,$0F,$D9,$AF,$F7,$53,$FD
	.dc.b	$9F,$91,$F8,$C3,$EA,$3F,$5C,$DC,$3F,$AC,$4E,$BE,$23,$35,$27,$35
	.dc.b	$BB,$FE,$0A,$AB,$51,$F1,$BE,$17,$F8,$F1,$FC,$7F,$EB,$AD,$FF,$62
	.dc.b	$E0,$FF,$7A,$36,$74,$D3,$F8,$B7,$E6,$6F,$A1,$FE,$20,$D8,$F3,$FD
	.dc.b	$73,$DA,$CD,$49,$E2,$FF,$0B,$9D,$46,$09,$97,$F1,$E8,$FA,$0F,$E7
	.dc.b	$EF,$B3,$9E,$2F,$E7,$1F,$5D,$FC,$E7,$E8,$2E,$1E,$B7,$FC,$19,$F8
	.dc.b	$43,$EB,$69,$B9,$FE,$F3,$19,$7E,$17,$D7,$7F,$43,$EA,$FE,$88,$D8
	.dc.b	$F8,$DC,$DD,$3E,$0B,$FA,$43,$63,$9C,$5D,$3F,$2B,$37,$AB,$FD,$D9
	.dc.b	$EF,$7F,$4F,$31,$7F,$EA,$CE,$4F,$F0,$A5,$FF,$AE,$FB,$9A,$7E,$EA
	.dc.b	$BC,$FE,$17,$F5,$7F,$93,$F6,$B2,$32,$7D,$B8,$C5,$F3,$AE,$4F,$FA
	.dc.b	$C7,$A1,$5A,$CE,$CF,$FB,$2A,$73,$3F,$E1,$EF,$BA,$7C,$9F,$F8,$DD
	.dc.b	$B1,$FE,$2A,$9F,$EB,$2E,$E7,$44,$75,$EA,$5B,$F7,$5F,$27,$C9,$DD
	.dc.b	$9C,$7C,$7E,$FB,$2F,$CF,$47,$FB,$6E,$DF,$47,$E3,$3F,$C8,$B3,$FA
	.dc.b	$1C,$BD,$11,$96,$9E,$7E,$FC,$65,$E2,$9E,$3D,$B8,$C7,$C3,$87,$E4
	.dc.b	$1F,$07,$C4,$FC,$13,$BF,$A3,$72,$F8,$D1,$71,$F9,$D3,$6B,$8D,$7C
	.dc.b	$7E,$AC,$3F,$EE,$CC,$3F,$25,$7F,$CF,$7E,$AD,$F1,$BF,$DF,$5B,$5F
	.dc.b	$3F,$EB,$F1,$32,$3F,$C1,$53,$AA,$FB,$D5,$97,$A5,$7C,$6F,$57,$19
	.dc.b	$7B,$9F,$F0,$57,$BF,$F4,$E3,$F0,$74,$8C,$63,$E7,$F5,$B7,$CD,$23
	.dc.b	$FC,$35,$3A,$AF,$C5,$56,$5F,$F2,$1F,$99,$FE,$85,$A7,$B5,$FC,$D1
	.dc.b	$FF,$45,$73,$77,$9D,$4E,$69,$39,$AF,$F3,$E1,$CB,$BF,$FC,$B9,$7C
	.dc.b	$1F,$CB,$7C,$74,$9F,$F3,$37,$95,$70,$FC,$8C,$DD,$7F,$1F,$A7,$F2
	.dc.b	$29,$CD,$FF,$8D,$77,$FF,$0A,$C7,$F7,$16,$D7,$0E,$2F,$F1,$94,$62
	.dc.b	$BF,$A1,$B6,$BF,$D4,$CD,$D3,$E6,$57,$0E,$49,$7D,$DF,$F3,$B4,$62
	.dc.b	$FF,$CF,$9F,$97,$7F,$83,$5F,$F6,$DF,$0F,$FB,$92,$31,$F1,$5F,$B6
	.dc.b	$9F,$8F,$A3,$C8,$FD,$EC,$E7,$E0,$D7,$C3,$0E,$5E,$DE,$9B,$A7,$FB
	.dc.b	$AD,$FA,$B8,$F3,$8C,$64,$C3,$9D,$F1,$8F,$F6,$27,$D5,$FF,$53,$F4
	.dc.b	$7F,$DE,$50,$8C,$3F,$EF,$56,$BF,$DA,$8D,$77,$FF,$BC,$1F,$F7,$D7
	.dc.b	$C6,$7F,$00,$FA,$CE,$6D,$8C,$F8,$29,$CB,$63,$B7,$83,$E7,$6E,$7E
	.dc.b	$4C,$7F,$AE,$36,$75,$3F,$DF,$21,$F5,$9F,$EB,$48,$F8,$31,$8F,$67
	.dc.b	$FC,$03,$C5,$F3,$6B,$55,$F5,$8F,$A7,$F2,$DA,$E2,$FF,$74,$61,$C6
	.dc.b	$79,$D7,$0B,$E5,$86,$E7,$89,$F4,$07,$D6,$7F,$A0,$9F,$9A,$D7,$E5
	.dc.b	$1A,$F8,$6F,$A2,$D3,$6F,$F8,$63,$AF,$F7,$F3,$C5,$B1,$9E,$4F,$C0
	.dc.b	$FC,$D5,$BF,$F8,$01,$7B,$9F,$F1,$0B,$63,$35,$BE,$BF,$E1,$57,$C7
	.dc.b	$CE,$7D,$5F,$F9,$06,$FF,$DD,$6D,$7F,$B1,$37,$2F,$DA,$FC,$5F,$18
	.dc.b	$C6,$7F,$E4,$A6,$D2,$FA,$FB,$DE,$1E,$BF,$F6,$DC,$FF,$B9,$FE,$07
	.dc.b	$D5,$83,$E8,$DF,$19,$86,$D6,$C8,$34,$FE,$CF,$5C,$6B,$81,$DE,$FE
	.dc.b	$35,$7F,$29,$E9,$FA,$38,$FF,$58,$8F,$D1,$B5,$FA,$ED,$7F,$44,$E7
	.dc.b	$0F,$E6,$6D,$C3,$FC,$9E,$D9,$FE,$4B,$F4,$DF,$E3,$17,$FD,$99,$F5
	.dc.b	$9F,$A9,$7D,$15,$FC,$BA,$7F,$CC,$30,$32,$72,$F7,$7D,$35,$C7,$0F
	.dc.b	$E8,$6F,$39,$FD,$B3,$5F,$F3,$0B,$5E,$8D,$4F,$3B,$DF,$81,$D8,$FC
	.dc.b	$36,$1F,$C6,$6B,$F0,$3E,$EE,$39,$64,$7E,$1A,$3F,$04,$7F,$0C,$E1
	.dc.b	$FC,$5B,$78,$8D,$24,$F1,$21,$F0,$39,$58,$3E,$23,$F1,$DE,$DF,$93
	.dc.b	$E2,$E5,$FE,$5A,$32,$FE,$BB,$2F,$CD,$C7,$84,$91,$9C,$91,$C2,$3F
	.dc.b	$3A,$FC,$0F,$96,$FD,$27,$D5,$7F,$9E,$67,$EA,$70,$FF,$88,$B3,$FE
	.dc.b	$CF,$18,$E0,$FB,$5E,$8F,$0F,$1E,$4F,$8F,$0E,$4F,$F2,$BD,$11,$88
	.dc.b	$FF,$4B,$C6,$27,$E5,$4F,$0C,$7F,$C8,$73,$9A,$F0,$1F,$AD,$9D,$61
	.dc.b	$F8,$BE,$6F,$4F,$A5,$F6,$BE,$B2,$5F,$0C,$3C,$33,$A8,$F3,$3E,$02
	.dc.b	$38,$38,$61,$FE,$F9,$6C,$7E,$1B,$EB,$3F,$0F,$D9,$27,$CB,$C9,$F6
	.dc.b	$A3,$E5,$FD,$64,$9F,$88,$FB,$88,$FF,$26,$9F,$C0,$FB,$5B,$8F,$F3
	.dc.b	$96,$2A,$FC,$AC,$FE,$5B,$E5,$3F,$2F,$F5,$92,$FC,$3C,$69,$FE,$3D
	.dc.b	$F3,$DF,$D2,$A7,$E7,$B5,$FF,$70,$B5,$FE,$30,$DC,$BF,$27,$F5,$7F
	.dc.b	$A4,$FA,$C9,$5D,$A3,$F8,$6F,$D4,$FD,$28,$FD,$D7,$D5,$78,$5F,$AE
	.dc.b	$1C,$3F,$BE,$11,$C5,$BC,$FF,$D1,$A3,$F0,$D1,$FE,$DC,$6B,$FC,$1D
	.dc.b	$AF,$03,$BB,$CC,$E9,$3F,$CB,$63,$E2,$FB,$3F,$BE,$23,$F2,$3E,$B2
	.dc.b	$BF,$B3,$D7,$FC,$C2,$F0,$34,$7F,$B9,$3E,$B1,$3F,$D4,$FC,$34,$62
	.dc.b	$75,$39,$87,$B9,$62,$4E,$8E,$91,$F3,$3E,$B3,$F9,$CE,$5F,$C9,$A7
	.dc.b	$E4,$37,$B7,$FC,$CB,$A2,$B8,$57,$F8,$7A,$30,$D8,$F9,$84,$7E,$3B
	.dc.b	$C1,$EB,$E9,$7E,$23,$BF,$FB,$47,$B8,$FB,$0A,$66,$BF,$59,$CB,$FC
	.dc.b	$11,$FF,$64,$37,$2F,$DC,$B7,$C1,$FE,$81,$AF,$F7,$3F,$8C,$FF,$17
	.dc.b	$57,$FA,$B9,$B3,$FE,$D3,$BD,$70,$ED,$87,$EB,$A7,$6F,$F9,$D1,$FE
	.dc.b	$78,$FF,$4B,$E4,$FE,$75,$7F,$E8,$AF,$AD,$7C,$67,$F6,$46,$D7,$E5
	.dc.b	$AB,$78,$FF,$53,$84,$7F,$43,$7F,$C8,$8D,$74,$D7,$C1,$EC,$7F,$87
	.dc.b	$A3,$9C,$60,$ED,$1E,$1F,$83,$F8,$5E,$D4,$9F,$63,$F1,$7D,$6C,$47
	.dc.b	$F4,$5D,$13,$8D,$3E,$45,$7F,$7E,$FA,$D3,$FE,$00,$F8,$67,$9D,$4D
	.dc.b	$44,$C7,$F6,$B5,$C1,$B1,$C7,$D1,$FF,$04,$FD,$6B,$E3,$3F,$C1,$7F
	.dc.b	$5A,$F8,$8A,$6A,$26,$1F,$E6,$EA,$E0,$D8,$FF,$20,$C6,$23,$FC,$8A
	.dc.b	$DD,$3F,$D7,$ED,$E2,$FA,$4D,$9F,$F0,$36,$0F,$9A,$FF,$0E,$FB,$BF
	.dc.b	$12,$7F,$87,$FD,$6B,$DC,$FF,$10,$B6,$3F,$89,$CE,$DF,$E2,$4D,$CB
	.dc.b	$FD,$9E,$DC,$BF,$04,$D9,$FF,$0B,$57,$C7,$7F,$8D,$7D,$6F,$E4,$97
	.dc.b	$D6,$FE,$D6,$FF,$14,$D8,$FF,$0F,$4E,$C8,$FF,$8D,$9E,$51,$FE,$80
	.dc.b	$C3,$F5,$3E,$67,$F8,$37,$BB,$F9,$4D,$FF,$3A,$8C,$11,$FD,$9B,$FD
	.dc.b	$00,$FD,$76,$5F,$AF,$CE,$CF,$EA,$63,$0F,$EA,$E3,$2F,$E3,$E3,$2F
	.dc.b	$EF,$3C,$9F,$11,$1F,$F0,$85,$FF,$26,$7F,$BF,$37,$0F,$E0,$33,$FD
	.dc.b	$48,$FF,$31,$7B,$FF,$D5,$3C,$0F,$61,$2F,$9B,$5F,$E1,$6E,$DF,$ED
	.dc.b	$BE,$FB,$E9,$71,$1A,$3F,$89,$CF,$F9,$BE,$33,$87,$F1,$28,$BB,$FF
	.dc.b	$B2,$A7,$FD,$15,$E3,$3E,$99,$E9,$7F,$B1,$A3,$0D,$8C,$E3,$FE,$9C
	.dc.b	$9C,$B6,$32,$FF,$96,$A3,$2D,$8C,$E3,$FE,$FE,$F8,$2F,$F4,$97,$D6
	.dc.b	$3E,$63,$FD,$2D,$F5,$BF,$D1,$B0,$FF,$05,$B5,$FF,$32,$FB,$8A,$E1
	.dc.b	$1F,$E1,$D6,$BF,$E2,$F6,$BF,$E2,$E6,$BF,$E6,$F8,$E5,$5F,$E5,$99
	.dc.b	$E3,$87,$F8,$C1,$BA,$ED,$B6,$F5,$9F,$53,$FC,$CB,$17,$D5,$FE,$D6
	.dc.b	$BF,$C4,$7F,$59,$FE,$A0,$7F,$CB,$CD,$CB,$FC,$7B,$3B,$9F,$6B,$5F
	.dc.b	$F2,$2B,$5F,$43,$5F,$F5,$73,$6B,$C3,$DB,$FE,$BD,$8E,$95,$DA,$3F
	.dc.b	$D8,$58,$78,$73,$E7,$1B,$26,$1F,$E3,$A9,$D3,$FE,$95,$F9,$07,$A5
	.dc.b	$FE,$B1,$87,$D6,$FF,$AD,$94,$BE,$13,$FA,$54,$E6,$31,$3F,$E9,$F8
	.dc.b	$C5,$73,$6E,$5F,$EA,$3F,$A2,$E1,$87,$F5,$36,$DE,$7A,$BF,$AB,$63
	.dc.b	$E4,$BC,$4F,$9F,$1E,$95,$AC,$3C,$6D,$9F,$13,$CA,$77,$1C,$6B,$9B
	.dc.b	$EC,$AF,$15,$71,$86,$1E,$86,$C7,$3A,$E3,$87,$C0,$9C,$9F,$2E,$F9
	.dc.b	$BD,$CC,$9F,$2B,$0F,$9D,$CD,$FE,$CB,$8C,$5F,$FB,$1E,$31,$7F,$06
	.dc.b	$51,$8A,$E3,$8F,$F7,$B9,$FF,$67,$7C,$77,$18,$C4,$23,$0E,$DA,$FF
	.dc.b	$80,$DA,$F2,$3C,$E3,$D4,$6F,$F2,$23,$55,$F8,$57,$51,$97,$C0,$72
	.dc.b	$9E,$2F,$43,$33,$C6,$32,$F3,$8E,$3A,$E6,$BE,$DC,$7A,$4F,$CA,$EB
	.dc.b	$F0,$6F,$AD,$71,$C3,$CD,$1D,$27,$E1,$C6,$2F,$CC,$F4,$BF,$F7,$B5
	.dc.b	$72,$AD,$C7,$63,$FC,$51,$19,$7F,$6B,$F0,$FE,$CD,$7F,$7E,$DC,$3F
	.dc.b	$E0,$07,$27,$F0,$AF,$07,$17,$C2,$7F,$D7,$1C,$7F,$ED,$86,$DE,$AF
	.dc.b	$84,$60,$95,$FE,$CF,$AE,$0D,$AF,$9A,$FF,$0C,$6B,$FC,$31,$E9,$CB
	.dc.b	$FE,$0D,$AE,$78,$FD,$4C,$13,$9B,$FE,$67,$F5,$BF,$EF,$7B,$FF,$67
	.dc.b	$36,$3F,$DF,$9C,$8F,$A9,$F6,$F3,$AB,$59,$7F,$B7,$9E,$17,$FE,$E4
	.dc.b	$6B,$FC,$DD,$AA,$79,$7C,$5F,$AA,$31,$3D,$31,$F9,$B8,$BE,$43,$C3
	.dc.b	$39,$AF,$E7,$53,$9A,$FF,$76,$F3,$7F,$63,$F1,$BE,$9B,$63,$FB,$8E
	.dc.b	$98,$EF,$FF,$67,$47,$06,$E1,$FA,$FA,$E0,$D8,$CB,$FE,$6D,$8C,$BF
	.dc.b	$E6,$FD,$1D,$EF,$1B,$E3,$B7,$0F,$F0,$0E,$B8,$99,$1F,$EF,$D9,$D5
	.dc.b	$B7,$FB,$1E,$27,$B6,$3F,$B4,$CE,$A7,$F5,$F8,$F8,$30,$E7,$5C,$70
	.dc.b	$FC,$AE,$1C,$6F,$78,$71,$9F,$ED,$4F,$82,$BD,$4D,$73,$7F,$57,$0E
	.dc.b	$D5,$FC,$41,$F7,$F2,$FB,$8F,$CE,$63,$F8,$DC,$7B,$C6,$6B,$86,$3F
	.dc.b	$E5,$D6,$C7,$58,$C6,$3F,$C7,$F8,$DE,$A3,$A3,$F5,$0F,$F8,$A7,$0F
	.dc.b	$F5,$54,$63,$C4,$CB,$63,$27,$FA,$BE,$3E,$9F,$0C,$8C,$1F,$45,$C6
	.dc.b	$1F,$CB,$F5,$FB,$71,$E5,$24,$FF,$30,$97,$CF,$7F,$6C,$84,$7D,$12
	.dc.b	$C6,$5C,$75,$3E,$B5,$E8,$FF,$08,$9E,$7E,$05,$BE,$1C,$3F,$CB,$30
	.dc.b	$E1,$D5,$9A,$FE,$1F,$18,$CB,$C3,$5F,$C1,$BE,$B7,$FD,$53,$18,$3F
	.dc.b	$D8,$8E,$31,$F1,$FD,$6B,$D9,$F1,$4E,$6F,$97,$AF,$DA,$BF,$89,$C6
	.dc.b	$30,$E9,$87,$F4,$99,$CC,$7F,$B4,$7E,$4D,$F3,$9E,$D6,$C7,$0C,$DF
	.dc.b	$58,$F6,$F5,$72,$BE,$75,$F0,$C7,$FA,$FB,$5F,$EC,$8D,$7C,$13,$F1
	.dc.b	$2D,$F2,$1C,$E7,$E8,$C6,$2F,$D3,$0F,$AD,$FE,$18,$DE,$1E,$5F,$39
	.dc.b	$F5,$67,$27,$F1,$1D,$3D,$3F,$1F,$9A,$F0,$3F,$96,$ED,$98,$FF,$BF
	.dc.b	$FA,$2E,$5F,$EE,$39,$BC,$3F,$CD,$7B,$3F,$97,$5F,$CC,$ED,$FD,$0D
	.dc.b	$7E,$17,$EA,$FF,$4B,$7F,$D0,$9B,$5F,$D7,$37,$7D,$36,$FC,$09,$3F
	.dc.b	$97,$AF,$D6,$E7,$FE,$0B,$6B,$F9,$80,$E1,$FC,$DB,$D9,$FF,$03,$5B
	.dc.b	$EB,$7A,$C7,$CC,$7F,$6B,$E2,$4B,$F5,$6B,$FC,$0D,$FE,$CE,$FF,$13
	.dc.b	$6C,$6E,$BF,$51,$AF,$F0,$0F,$4B,$1A,$FF,$18,$FA,$9F,$E5,$67,$F0
	.dc.b	$2F,$A3,$B8,$FF,$29,$8F,$ED,$9F,$17,$FE,$93,$8C,$1F,$2D,$C6,$3F
	.dc.b	$B5,$3D,$FD,$FF,$34,$FE,$BB,$F5,$AE,$CC,$5F,$E1,$B8,$7F,$4C,$FC
	.dc.b	$BA,$FE,$B1,$7C,$A3,$04,$7F,$81,$66,$FB,$1F,$AB,$3C,$EB,$FC,$0F
	.dc.b	$39,$E4,$FD,$DF,$D7,$3E,$16,$98,$F8,$BB,$87,$EA,$67,$4D,$9D,$34
	.dc.b	$FA,$3F,$0F,$FC,$A0,$FE,$53,$BD,$8C,$93,$4F,$A3,$CA,$9D,$47,$C7
	.dc.b	$FA,$D7,$D1,$FF,$17,$6D,$FE,$EF,$EB,$BF,$CD,$6F,$EB,$23,$35,$C6
	.dc.b	$BE,$3C,$62,$3F,$CB,$8D,$7F,$24,$D7,$F9,$5B,$CB,$6F,$E4,$57,$F5
	.dc.b	$6F,$B5,$FF,$22,$8C,$5F,$AB,$EA,$3F,$A3,$FD,$17,$F4,$86,$DF,$F1
	.dc.b	$8F,$3D,$F5,$AF,$AC,$FE,$BC,$D9,$FE,$BC,$DC,$3F,$A2,$35,$FD,$47
	.dc.b	$67,$F5,$6F,$AE,$FE,$AB,$D7,$FD,$59,$B7,$FA,$2E,$7F,$F1,$4E,$1F
	.dc.b	$92,$DF,$F8,$04,$63,$E1,$65,$78,$1F,$19,$1F,$F3,$FE,$3F,$D9,$3D
	.dc.b	$EF,$EC,$67,$0C,$FE,$CB,$DB,$A6,$57,$EC,$BE,$56,$D7,$FB,$CB,$3F
	.dc.b	$C9,$BE,$CA,$F8,$24,$AF,$82,$FE,$B1,$E3,$9F,$F9,$2F,$97,$FD,$0F
	.dc.b	$1F,$2F,$EB,$9F,$2B,$FE,$46,$6C,$7E,$C7,$D5,$C4,$C8,$FF,$05,$4E
	.dc.b	$B1,$FE,$E0,$D7,$FC,$58,$D7,$FE,$00,$3B,$C3,$C5,$D7,$CE,$7F,$84
	.dc.b	$E3,$FA,$47,$FB,$97,$22,$DF,$4B,$E5,$3F,$83,$F7,$5C,$BD,$51,$FE
	.dc.b	$1F,$9D,$C7,$9A,$30,$47,$F8,$67,$57,$1C,$3C,$2F,$C1,$F8,$B9,$3F
	.dc.b	$C4,$7D,$2B,$5E,$4C,$15,$DC,$95,$F1,$1B,$3E,$CE,$9E,$3B,$7D,$17
	.dc.b	$FC,$11,$F5,$CF,$8E,$A7,$F2,$2F,$AE,$B8,$FF,$5E,$9D,$3F,$E1,$96
	.dc.b	$E1,$FE,$CA,$6B,$FE,$CF,$6B,$FD,$15,$DE,$1F,$81,$F5,$BA,$CF,$D0
	.dc.b	$7F,$C3,$5E,$FF,$D5,$BF,$E9,$8D,$97,$57,$F0,$DF,$5F,$7C,$D2,$FF
	.dc.b	$C5,$97,$F7,$EF,$FC,$56,$D4,$E6,$EF,$97,$FE,$B0,$A7,$B7,$FC,$DA
	.dc.b	$FF,$A5,$DA,$B5,$39,$CF,$FC,$DB,$18,$EF,$FF,$2C,$B7,$0F,$CE,$7C
	.dc.b	$7F,$E6,$CF,$F2,$EF,$91,$63,$24,$CB,$BC,$65,$FE,$9D,$C3,$FE,$19
	.dc.b	$C7,$FD,$72,$DA,$FF,$3A,$F5,$7F,$90,$63,$15,$FE,$4A,$6B,$F7,$79
	.dc.b	$BF,$CD,$B3,$B6,$57,$F9,$4E,$BF,$C4,$55,$C6,$7F,$AA,$C6,$3E,$09
	.dc.b	$F5,$CF,$6F,$FD,$EF,$F1,$5F,$B8,$C1,$5A,$9C,$F1,$FF,$A2,$BC,$8B
	.dc.b	$5A,$C3,$96,$7F,$E6,$0C,$79,$63,$C6,$31,$87,$F9,$92,$31,$6A,$F0
	.dc.b	$72,$7D,$CC,$7A,$D6,$A7,$C9,$4F,$AD,$77,$BD,$CF,$F3,$94,$61,$B1
	.dc.b	$99,$FF,$0E,$1F,$E7,$49,$D3,$81,$F1,$DF,$C1,$3E,$BB,$FC,$F8,$D8
	.dc.b	$D3,$5E,$6D,$7F,$C5,$13,$FD,$93,$1F,$1F,$83,$F2,$A7,$B4,$E5,$B1
	.dc.b	$97,$FD,$95,$F5,$DF,$ED,$D7,$E3,$9F,$A6,$EB,$79,$DF,$E3,$5D,$7E
	.dc.b	$3C,$62,$77,$8A,$F9,$47,$CD,$E0,$FA,$11,$8B,$FA,$AD,$8C,$FB,$1D
	.dc.b	$5F,$F6,$3F,$CE,$5B,$FF,$4D,$FD,$77,$FA,$59,$FF,$2A,$64,$F4,$3F
	.dc.b	$C9,$1D,$7F,$ED,$66,$F3,$7F,$92,$EF,$85,$FF,$8B,$DB,$5B,$8E,$75
	.dc.b	$C1,$EF,$F0,$D5,$57,$0B,$E9,$39,$8C,$4F,$C7,$F9,$CE,$5E,$5B,$EB
	.dc.b	$7F,$DE,$73,$FD,$FD,$AF,$F4,$46,$E5,$FE,$EE,$F8,$CE,$31,$8C,$BF
	.dc.b	$D7,$77,$CF,$E0,$B2,$79,$FB,$3F,$89,$DD,$D9,$F0,$5E,$BC,$1F,$57
	.dc.b	$FE,$D9,$D3,$FD,$4F,$18,$F8,$DE,$15,$27,$35,$B7,$FD,$95,$5B,$7E
	.dc.b	$8C,$FC,$5B,$FE,$B7,$1F,$E5,$DF,$B0,$F2,$78,$A7,$5E,$94,$9D,$10
	.dc.b	$FF,$31,$57,$27,$FC,$C9,$F5,$FF,$A9,$7F,$CE,$7F,$53,$FE,$63,$7F
	.dc.b	$D2,$8D,$77,$E1,$FD,$79,$E1,$C6,$D3,$FD,$C7,$C0,$C3,$84,$E7,$0F
	.dc.b	$E8,$71,$8A,$FF,$49,$5F,$26,$E1,$D4,$95,$E2,$E9,$FA,$FD,$DF,$4F
	.dc.b	$0F,$C1,$E5,$DF,$A3,$37,$F8,$1A,$D4,$48,$C1,$0F,$C2,$47,$07,$E0
	.dc.b	$FC,$57,$E1,$3D,$BE,$DF,$1B,$27,$F1,$91,$97,$E2,$C6,$5F,$C7,$46
	.dc.b	$5A,$70,$3E,$5F,$CC,$FE,$63,$EB,$7F,$8E,$C7,$CE,$F4,$BD,$86,$7C
	.dc.b	$99,$19,$21,$FE,$1E,$9E,$2D,$8D,$B4,$FB,$3C,$9D,$FD,$3F,$91,$F5
	.dc.b	$F2,$7F,$CC,$97,$FE,$E8,$6C,$E5,$FC,$1C,$F1,$72,$43,$C5,$1F,$4F
	.dc.b	$EB,$FE,$A7,$5C,$9F,$C4,$BE,$FF,$8F,$91,$82,$1E,$B7,$EE,$7B,$7A
	.dc.b	$8C,$1F,$92,$7F,$01,$F2,$DF,$81,$FA,$F9,$3F,$96,$8F,$E0,$7F,$41
	.dc.b	$FC,$BF,$B3,$F2,$1C,$FF,$93,$9E,$35,$FE,$83,$AD,$75,$FE,$63,$A5
	.dc.b	$F5,$DF,$8D,$FA,$F9,$7F,$A2,$8F,$E2,$DF,$59,$F4,$33,$FC,$E4,$BE
	.dc.b	$B5,$9F,$F1,$B6,$C0,$FF,$69,$57,$1C,$BF,$DE,$4D,$8D,$4F,$07,$FD
	.dc.b	$B0,$DB,$EF,$3B,$93,$67,$55,$FE,$FD,$6C,$F8,$67,$B6,$9F,$B5,$F6
	.dc.b	$7F,$54,$49,$FD,$FF,$D7,$57,$F8,$1C,$5C,$7E,$5E,$AA,$FE,$75,$E0
	.dc.b	$BE,$BF,$18,$7B,$EF,$F8,$04,$E6,$B7,$5A,$A5,$6A,$1E,$F6,$AA,$57
	.dc.b	$3D,$7E,$56,$1C,$E3,$07,$F8,$16,$F8,$B8,$3E,$2F,$D7,$FD,$3F,$94
	.dc.b	$91,$9D,$93,$F7,$0C,$C7,$D0,$9C,$3B,$8F,$F0,$53,$F2,$DA,$F6,$33
	.dc.b	$3D,$1F,$F0,$87,$9C,$E4,$FF,$2B,$C7,$A3,$83,$FD,$E1,$F5,$FF,$CE
	.dc.b	$8F,$CC,$F8,$12,$75,$E5,$A7,$51,$F9,$99,$FE,$A9,$F2,$5F,$E2,$0E
	.dc.b	$2F,$CF,$B7,$83,$F9,$9B,$6B,$FC,$E6,$FF,$2F,$69,$C8,$87,$95,$D3
	.dc.b	$FE,$32,$F2,$7F,$BB,$7D,$7B,$E3,$E9,$FD,$E7,$C5,$FF,$1C,$E3,$FA
	.dc.b	$46,$E3,$FA,$1C,$17,$C5,$FA,$CD,$7F,$07,$1B,$9E,$31,$FA,$96,$BC
	.dc.b	$AF,$C2,$FD,$F8,$DE,$AC,$9E,$BF,$C6,$25,$7B,$7D,$5A,$F8,$0F,$8B
	.dc.b	$18,$CF,$E5,$CF,$F8,$57,$EB,$CF,$F8,$5F,$E2,$1E,$65,$35,$13,$0F
	.dc.b	$C9,$4E,$9B,$39,$7C,$03,$EB,$DF,$1D,$FE,$23,$FA,$F7,$C4,$D3,$FC
	.dc.b	$4D,$3A,$AF,$CF,$35,$E7,$3C,$1F,$D1,$35,$EF,$1B,$CF,$F8,$CB,$78
	.dc.b	$7E,$B5,$6E,$7F,$8E,$B6,$3E,$D4,$EE,$BF,$8E,$7B,$3F,$39,$FD,$87
	.dc.b	$D7,$FF,$89,$1B,$EC,$FF,$8F,$67,$51,$F0,$1B,$8F,$F2,$AA,$E1,$3F
	.dc.b	$11,$B1,$FC,$BA,$77,$87,$E2,$FE,$BF,$F3,$96,$FA,$FF,$F0,$63,$FC
	.dc.b	$BD,$AA,$36,$70,$8C,$9A,$7F,$4D,$DB,$FB,$7F,$23,$FC,$5B,$F1,$72
	.dc.b	$FF,$A4,$47,$F4,$E3,$0F,$E2,$79,$DD,$DF,$EA,$F8,$D9,$FA,$BD,$BF
	.dc.b	$6F,$A4,$8C,$10,$E5,$D0,$8D,$46,$27,$FA,$EF,$91,$FE,$51,$F9,$79
	.dc.b	$3F,$D6,$E3,$2F,$F9,$E2,$32,$FE,$77,$4F,$E2,$53,$FD,$9B,$C8,$FF
	.dc.b	$2D,$7C,$7C,$AF,$EC,$91,$97,$FD,$03,$19,$79,$78,$7F,$CB,$9B,$3B
	.dc.b	$69,$FE,$88,$FA,$6C,$72,$7F,$90,$71,$FE,$45,$58,$E4,$FF,$23,$FB
	.dc.b	$CB,$FF,$42,$7C,$43,$E9,$DE,$9F,$F9,$BF,$DC,$9A,$7F,$40,$AD,$4F
	.dc.b	$0A,$DC,$FF,$A6,$2B,$6F,$EF,$6B,$6F,$F4,$8C,$DF,$5C,$F9,$AF,$F3
	.dc.b	$F7,$D7,$FB,$77,$FF,$04,$B7,$6F,$F4,$2D,$6F,$0E,$4D,$7F,$05,$3B
	.dc.b	$5D,$08,$C6,$3E,$0C,$7F,$D3,$37,$C7,$1E,$FC,$1E,$23,$33,$FE,$9E
	.dc.b	$FA,$D7,$D1,$FF,$A4,$FE,$B7,$F0,$DA,$7F,$4B,$6B,$F7,$E7,$6F,$BD
	.dc.b	$AF,$E5,$E3,$78,$7F,$89,$9B,$AF,$FA,$8A,$B7,$3F,$65,$B1,$9B,$F2
	.dc.b	$1F,$EA,$5F,$AC,$E3,$83,$45,$F0,$8C,$36,$33,$5C,$7A,$BF,$D8,$5F
	.dc.b	$26,$F4,$3F,$C0,$B8,$3E,$BB,$FD,$4D,$2F,$85,$FE,$99,$18,$6F,$C5
	.dc.b	$FF,$4F,$9E,$5B,$7F,$8C,$6B,$8E,$1D,$1E,$18,$66,$F8,$61,$CA,$3F
	.dc.b	$C7,$18,$72,$7C,$53,$D6,$F9,$3C,$23,$91,$C2,$B9,$2A,$E5,$7E,$2C
	.dc.b	$37,$39,$AD,$D7,$B6,$32,$F0,$BF,$05,$62,$B6,$70,$C3,$C7,$7E,$87
	.dc.b	$E4,$F2,$BE,$BF,$8C,$A3,$0E,$EF,$FC,$AD,$18,$8F,$45,$72,$AF,$EF
	.dc.b	$51,$FE,$5B,$F9,$3D,$C6,$21,$18,$76,$D7,$FC,$04,$D7,$91,$F3,$0F
	.dc.b	$F0,$37,$B7,$FE,$C7,$D3,$FD,$C9,$1A,$6C,$67,$3F,$05,$7C,$AA,$4E
	.dc.b	$5F,$09,$D4,$FF,$24,$F1,$6D,$F3,$99,$38,$B9,$9E,$93,$CE,$7D,$26
	.dc.b	$DC,$3E,$5A,$F8,$B1,$88,$EB,$7E,$8B,$E9,$1C,$5F,$C3,$F0,$7E,$9B
	.dc.b	$A7,$FD,$CD,$AF,$EB,$DB,$3F,$EE,$CE,$37,$85,$8C,$3F,$57,$57,$0E
	.dc.b	$57,$FB,$F8,$D7,$53,$0E,$33,$92,$5F,$F0,$3E,$4E,$5C,$AC,$38,$F3
	.dc.b	$7F,$10,$BD,$E9,$FC,$1E,$F7,$5A,$7C,$5F,$5E,$F8,$AF,$0E,$6F,$93
	.dc.b	$CC,$61,$B8,$FF,$98,$E7,$4F,$F3,$5A,$E4,$E6,$77,$A7,$F6,$FA,$DB
	.dc.b	$FD,$76,$B6,$E6,$7E,$37,$37,$F4,$BF,$23,$E1,$F8,$5F,$E5,$9C,$7F
	.dc.b	$A5,$78,$1F,$EB,$87,$FC,$E9,$19,$C3,$FA,$44,$E5,$FE,$FD,$19,$D9
	.dc.b	$5F,$A8,$AD,$47,$FB,$42,$75,$3F,$EC,$1E,$27,$C8,$7F,$7D,$8F,$EA
	.dc.b	$4D,$C3,$AE,$7F,$6E,$31,$87,$F6,$96,$D7,$36,$D7,$F4,$EC,$3E,$1D
	.dc.b	$6F,$B7,$91,$F8,$2F,$1F,$CF,$87,$82,$30,$E8,$FF,$4A,$46,$AF,$FC
	.dc.b	$11,$87,$6E,$54,$E7,$0F,$27,$37,$5F,$A0,$F4,$37,$1F,$6B,$71,$E6
	.dc.b	$DA,$FF,$4B,$F5,$7E,$5B,$C9,$FF,$01,$5F,$FA,$46,$31,$E3,$65,$B1
	.dc.b	$93,$FD,$2D,$1F,$1F,$EA,$FF,$3F,$E5,$BE,$23,$15,$FE,$AA,$6E,$7F
	.dc.b	$E3,$79,$D4,$FD,$2A,$E3,$5F,$AD,$A6,$9F,$D4,$EB,$FD,$9F,$DE,$FA
	.dc.b	$FF,$C7,$EB,$F8,$2B,$FF,$7F,$5E,$DF,$F4,$65,$6F,$AF,$A4,$E6,$BF
	.dc.b	$DC,$B7,$CB,$0F,$F5,$34,$E5,$FF,$6F,$C6,$70,$0F,$AF,$FF,$48,$CF
	.dc.b	$93,$EB,$FF,$70,$FF,$B8,$63,$33,$E2,$9C,$BF,$E9,$08,$CE,$1F,$EF
	.dc.b	$69,$CB,$EB,$9E,$DE,$C7,$F0,$FC,$D5,$AC,$3F,$D7,$9F,$08,$F9,$FD
	.dc.b	$7D,$73,$E4,$3E,$54,$E6,$BE,$4C,$E5,$FF,$7B,$46,$72,$F8,$F3,$97
	.dc.b	$FD,$4B,$19,$CF,$E7,$64,$AD,$47,$9E,$75,$7F,$1A,$B5,$87,$8B,$1F
	.dc.b	$2B,$F7,$7E,$BF,$E6,$4F,$3E,$1C,$70,$FA,$9B,$87,$A1,$BA,$79,$9F
	.dc.b	$CC,$75,$49,$FF,$74,$71,$DA,$FF,$3D,$CD,$D7,$FD,$EB,$36,$BF,$DC
	.dc.b	$9F,$5B,$FC,$D2,$FE,$0F,$D7,$FE,$46,$31,$7F,$03,$81,$39,$EA,$F9
	.dc.b	$ED,$C3,$E6,$B7,$0E,$71,$F8,$8B,$FB,$98,$74,$8F,$8D,$87,$47,$FE
	.dc.b	$00,$77,$FE,$CC,$FD,$A7,$8D,$63,$24,$C3,$FD,$2F,$19,$7F,$83,$F6
	.dc.b	$FE,$CB,$D0,$FE,$34,$DA,$FE,$3E,$D9,$DC,$79,$4F,$CE,$F7,$FF,$AA
	.dc.b	$AB,$E1,$7A,$DF,$93,$AF,$8E,$D7,$F4,$79,$A3,$10,$8C,$5F,$F8,$56
	.dc.b	$31,$3F,$DA,$74,$7D,$73,$EA,$BB,$DF,$8E,$D5,$A9,$CC,$62,$7E,$65
	.dc.b	$71,$78,$47,$67,$95,$71,$7F,$84,$FD,$77,$FB,$F9,$BF,$05,$FA,$D8
	.dc.b	$7C,$3D,$89,$3A,$38,$C6,$2F,$FD,$8F,$F0,$7F,$CA,$4F,$E5,$5E,$1D
	.dc.b	$AF,$56,$EF,$19,$59,$BF,$F6,$7D,$DC,$FF,$80,$7D,$73,$E8,$FF,$80
	.dc.b	$7D,$73,$C2,$FE,$49,$39,$D3,$F9,$1C,$E5,$F8,$31,$9C,$7F,$90,$CE
	.dc.b	$5F,$90,$7F,$1A,$8D,$61,$FB,$C9,$D3,$E2,$C3,$C3,$87,$81,$B8,$7B
	.dc.b	$33,$7D,$3F,$F4,$9F,$9E,$FE,$94,$D9,$FF,$5F,$7D,$73,$EB,$7F,$A7
	.dc.b	$7D,$37,$F8,$9E,$31,$3C,$E3,$2F,$5E,$AB,$BF,$FB,$D1,$FE,$AD,$D1
	.dc.b	$C7,$AA,$D7,$67,$0F,$97,$D7,$B7,$FE,$FC,$7F,$AF,$12,$38,$1B,$3C
	.dc.b	$FD,$5E,$53,$FB,$27,$C8,$58,$C9,$F7,$39,$7F,$B9,$3F,$E0,$DD,$3F
	.dc.b	$25,$87,$F5,$D8,$E0,$D7,$8B,$4F,$A7,$E9,$BE,$B9,$F2,$7F,$E4,$B6
	.dc.b	$D7,$7F,$47,$13,$23,$C5,$DB,$70,$FE,$D8,$DE,$8F,$EC,$9D,$8D,$3F
	.dc.b	$41,$5E,$0C,$BA,$BC,$30,$FF,$7A,$43,$E8,$BC,$47,$F7,$CE,$EB,$3A
	.dc.b	$8F,$EF,$FD,$58,$BF,$EF,$8D,$CB,$C3,$8F,$E3,$AF,$FC,$65,$8F,$57
	.dc.b	$E5,$57,$BF,$E1,$E4,$7C,$DF,$3E,$E3,$E7,$9F,$F0,$6F,$4E,$32,$EA
	.dc.b	$DB,$FE,$33,$97,$E3,$2F,$84,$7D,$79,$FF,$06,$7D,$6B,$E2,$A9,$FE
	.dc.b	$18,$FA,$EB,$5F,$C2,$A7,$4F,$DF,$F9,$2C,$E1,$FE,$19,$6E,$5E,$E6
	.dc.b	$CF,$89,$FF,$0D,$7B,$7F,$3A,$7E,$1F,$B1,$8E,$5F,$F1,$3B,$6F,$F1
	.dc.b	$8D,$C3,$FC,$4C,$FF,$9A,$3E,$BB,$FD,$21,$3F,$EF,$5E,$EC,$3F,$C9
	.dc.b	$CC,$1F,$E6,$A8,$DF,$C3,$FF,$32,$BF,$F3,$4C,$62,$BF,$1D,$F1,$52
	.dc.b	$FF,$CE,$3E,$4D,$BF,$C3,$4D,$CF,$FC,$D9,$36,$BF,$C4,$75,$F2,$F9
	.dc.b	$3F,$C7,$1C,$3F,$E0,$28,$FF,$32,$71,$E2,$65,$7F,$97,$BA,$AD,$7F
	.dc.b	$14,$EA,$FF,$2E,$C6,$32,$FC,$CF,$85,$F9,$87,$FC,$49,$97,$5A,$DE
	.dc.b	$1F,$D1,$FE,$05,$F5,$CF,$8F,$FF,$22,$7C,$4F,$EB,$34,$FF,$78,$4E
	.dc.b	$6F,$FD,$D5,$CB,$FD,$83,$9B,$F8,$DC,$62,$B7,$87,$2A,$FF,$7A,$DF
	.dc.b	$19,$FF,$77,$C7,$F6,$8D,$7F,$A4,$DF,$1C,$BC,$78,$71,$B6,$7F,$E7
	.dc.b	$AE,$FC,$43,$EB,$7F,$A8,$FB,$7F,$D9,$DB,$1F,$EF,$D6,$BF,$40,$E1
	.dc.b	$3F,$EF,$07,$03,$E3,$7F,$75,$F5,$DF,$DC,$5B,$1A,$6B,$CD,$AF,$DC
	.dc.b	$AF,$EA,$59,$78,$FC,$13,$49,$3E,$3C,$BF,$D2,$B3,$F8,$4E,$1F,$A1
	.dc.b	$18,$9F,$98,$D7,$E4,$9F,$13,$6F,$F2,$D3,$A7,$FB,$13,$6B,$A6,$FF
	.dc.b	$3D,$B1,$C7,$96,$DF,$FA,$9B,$C0,$FC,$AF,$2F,$E2,$BC,$0B,$C7,$F5
	.dc.b	$DB,$87,$86,$AD,$F1,$8F,$F5,$9C,$E3,$33,$6A,$4E,$88,$7E,$6E,$7F
	.dc.b	$B0,$3F,$EB,$CE,$FE,$5E,$FC,$8C,$10,$E0,$E5,$E5,$1E,$1F,$9D,$91
	.dc.b	$FE,$C6,$7C,$FD,$6F,$0B,$33,$E1,$9B,$AF,$F3,$29,$DF,$B5,$C3,$EC
	.dc.b	$A5,$FF,$5B,$8C,$BF,$D7,$63,$2F,$FA,$DE,$3E,$2F,$ED,$7F,$82,$8F
	.dc.b	$F0,$EA,$F9,$0C,$9A,$E6,$93,$A8,$CB,$C2,$32,$FF,$01,$9E,$4E,$CD
	.dc.b	$1D,$BD,$5F,$0F,$34,$8C,$10,$E4,$E5,$E9,$1E,$83,$19,$FF,$64,$AF
	.dc.b	$E0,$93,$A3,$D7,$F5,$7E,$EF,$83,$91,$82,$1E,$77,$2F,$A6,$3E,$29
	.dc.b	$8A,$ED,$38,$9F,$8F,$F0,$DF,$2B,$E4,$A4,$60,$87,$C3,$72,$FC,$58
	.dc.b	$FA,$26,$2B,$D5,$38,$9F,$A7,$E5,$7D,$6F,$3A,$46,$08,$7C,$F7,$2F
	.dc.b	$D1,$8F,$BA,$62,$BE,$34,$E2,$7E,$FF,$D5,$7E,$08,$8F,$E0,$BE,$AB
	.dc.b	$F0,$A6,$1F,$B9,$18,$74,$43,$F0,$11,$F8,$CF,$AA,$FC,$6F,$A1,$27
	.dc.b	$F1,$AF,$E4,$3E,$43,$F2,$3F,$55,$67,$F2,$31,$87,$F2,$7F,$55,$F9
	.dc.b	$62,$3F,$96,$FA,$AF,$DF,$AF,$AA,$4E,$14,$D3,$11,$27,$FD,$1B,$3A
	.dc.b	$6C,$E5,$FF,$36,$9D,$67,$93,$A3,$8B,$FA,$AF,$3F,$F5,$7F,$53,$27
	.dc.b	$F5,$71,$97,$F5,$B1,$97,$F4,$D1,$FB,$2F,$4F,$F8,$BF,$EE,$DF,$D0
	.dc.b	$F8,$7F,$C7,$ED,$F1,$FF,$D7,$7D,$0A,$E1,$87,$FA,$52,$B4,$FF,$54
	.dc.b	$6F,$8E,$AE,$13,$BF,$3F,$BF,$D3,$48,$C1,$0F,$EF,$56,$8D,$46,$27
	.dc.b	$CE,$63,$2F,$3C,$E2,$7F,$D7,$3F,$52,$F1,$3F,$D9,$10,$8C,$3F,$EC
	.dc.b	$B6,$BF,$D8,$1A,$FF,$65,$73,$4D,$3F,$C2,$F1,$86,$D7,$E0,$7C,$0F
	.dc.b	$97,$F4,$DF,$33,$E6,$26,$3F,$1E,$32,$FC,$98,$CB,$F0,$23,$F1,$BF
	.dc.b	$7F,$FD,$6A,$3F,$06,$FF,$42,$6F,$99,$AF,$3A,$47,$FB,$62,$75,$A7
	.dc.b	$9E,$02,$33,$9F,$F8,$C1,$ED,$3B,$9D,$57,$8A,$E4,$75,$8F,$C3,$FC
	.dc.b	$8F,$E2,$3E,$66,$4F,$E0,$E3,$2F,$E1,$63,$2F,$DC,$8F,$D4,$FA,$DF
	.dc.b	$A1,$8F,$F0,$7B,$5F,$F0,$A3,$5F,$F2,$2B,$5F,$5B,$C7,$0F,$43,$72
	.dc.b	$DF,$4F,$ED,$BC,$EF,$DA,$DF,$FB,$F3,$EF,$FD,$F5,$FE,$50,$7E,$1C
	.dc.b	$7C,$66,$A7,$6E,$73,$FE,$03,$1A,$23,$83,$3A,$49,$F0,$47,$F0,$8F
	.dc.b	$C7,$3F,$D4,$91,$FE,$2E,$0E,$1F,$F1,$8B,$5F,$F1,$B3,$5F,$F3,$3B
	.dc.b	$BC,$FE,$C3,$70,$CF,$89,$FC,$5B,$C5,$F4,$DB,$C9,$FF,$2E,$3E,$16
	.dc.b	$BF,$E4,$06,$BF,$E4,$56,$BF,$E7,$67,$78,$7E,$1D,$B7,$9F,$67,$F9
	.dc.b	$2F,$ED,$FF,$23,$B5,$D8,$F9,$CF,$F1,$3F,$E2,$6D,$FF,$C2,$1A,$7E
	.dc.b	$5B,$4F,$CC,$4E,$7D,$3F,$E2,$B7,$2B,$5A,$7E,$72,$B5,$18,$0E,$1F
	.dc.b	$E1,$73,$FC,$7B,$E6,$64,$7F,$19,$7F,$90,$FC,$CE,$23,$F8,$8C,$5C
	.dc.b	$7F,$91,$CF,$F8,$E7,$EB,$BF,$C7,$5F,$1D,$2F,$C1,$B3,$DA,$56,$E7
	.dc.b	$30,$AE,$11,$FE,$3D,$9F,$F2,$27,$D9,$FF,$91,$78,$A4,$F0,$8C,$BC
	.dc.b	$A3,$2F,$F9,$16,$32,$FF,$91,$E3,$39,$7F,$A8,$27,$2F,$FA,$8A,$33
	.dc.b	$B7,$FA,$3E,$7F,$CA,$9F,$39,$FE,$55,$F8,$99,$87,$A2,$32,$FA,$A3
	.dc.b	$2F,$F9,$56,$3F,$CB,$FF,$4F,$FE,$E3,$8F,$86,$BC,$CC,$9A,$F3,$24
	.dc.b	$EA,$32,$F0,$8C,$BF,$EB,$B9,$E4,$EC,$D1,$F4,$3E,$93,$FA,$33,$7C
	.dc.b	$5F,$F3,$8C,$1B,$E5,$49,$D1,$0F,$F6,$44,$F2,$6C,$72,$99,$1D,$E3
	.dc.b	$93,$C2,$32,$F0,$3F,$24,$7F,$AA,$7B,$1B,$FF,$32,$FA,$67,$8D,$FE
	.dc.b	$E0,$F8,$57,$B1,$FE,$E2,$9C,$E1,$FC,$E2,$72,$F1,$AE,$18,$74,$85
	.dc.b	$70,$7C,$18,$6D,$E5,$87,$5B,$F0,$36,$3B,$5F,$3B,$E5,$82,$31,$7C
	.dc.b	$2B,$A1,$FE,$C5,$BE,$4F,$80,$F1,$C7,$9B,$0D,$DF,$AA,$72,$79,$0C
	.dc.b	$DF,$67,$C7,$3C,$2F,$AD,$71,$C3,$CB,$7B,$BF,$07,$97,$FE,$80,$6F
	.dc.b	$BB,$E1,$7F,$D9,$ED,$8D,$47,$F6,$4F,$84,$FE,$CB,$18,$84,$62,$7F
	.dc.b	$DC,$6D,$8F,$F7,$43,$5F,$F6,$B1,$EC,$3F,$B7,$FA,$1F,$DC,$70,$FA
	.dc.b	$91,$DC,$D0,$8D,$69,$DA,$33,$18,$7B,$98,$E6,$FA,$2F,$3A,$EB,$3B
	.dc.b	$8F,$1C,$62,$77,$1E,$C3,$91,$E2,$31,$1B,$71,$7D,$AB,$A4,$77,$BF
	.dc.b	$F0,$5B,$EE,$BF,$C7,$75,$B8,$5F,$19,$FE,$11,$E0,$FF,$09,$4E,$7E
	.dc.b	$2F,$17,$FE,$14,$8E,$15,$8C,$11,$B9,$FD,$F4,$E6,$31,$3F,$B8,$8C
	.dc.b	$60,$BB,$18,$6D,$FF,$07,$6D,$7F,$00,$6E,$1E,$6B,$E5,$5F,$CF,$3E
	.dc.b	$95,$F3,$DE,$17,$DB,$19,$8C,$63,$96,$C6,$4E,$51,$96,$99,$C7,$F8
	.dc.b	$2B,$F0,$EB,$AC,$7F,$C0,$1F,$13,$E0,$78,$0F,$94,$B1,$86,$DF,$F4
	.dc.b	$E6,$D7,$43,$E5,$D7,$19,$FE,$A7,$F0,$9D,$3E,$17,$AB,$63,$27,$F5
	.dc.b	$88,$CB,$4C,$D7,$5D,$FF,$B0,$4E,$A3,$16,$8C,$6D,$F8,$28,$C5,$F2
	.dc.b	$8C,$4F,$DD,$FD,$F5,$CD,$79,$B0,$DB,$FE,$DF,$3A,$6C,$E9,$A7,$FA
	.dc.b	$57,$E8,$3F,$11,$E5,$9B,$C5,$86,$4E,$CF,$F4,$E4,$63,$4F,$F4,$9F
	.dc.b	$17,$E5,$7C,$A7,$E6,$9F,$E9,$5B,$FF,$51,$4E,$63,$15,$C2,$7F,$C2
	.dc.b	$95,$C3,$BF,$C7,$7B,$7F,$41,$4F,$D3,$9F,$37,$86,$E1,$FE,$1D,$EB
	.dc.b	$78,$CF,$84,$D5,$CA,$FF,$26,$5F,$0C,$3F,$6E,$7F,$A0,$9C,$38,$6B
	.dc.b	$FB,$CF,$A6,$FF,$5F,$4F,$F5,$A6,$C7,$E2,$F1,$7D,$23,$E6,$3D,$35
	.dc.b	$AE,$7F,$F4,$53,$C2,$FC,$2D,$7B,$7C,$17,$FA,$3D,$B8,$F9,$97,$DE
	.dc.b	$FF,$0E,$8F,$F1,$E3,$FC,$3E,$3F,$38,$E5,$A7,$F2,$B8,$DB,$FE,$A0
	.dc.b	$6E,$5F,$C3,$67,$86,$1F,$EA,$57,$A3,$6B,$81,$D2,$78,$38,$8D,$47
	.dc.b	$D8,$FA,$6F,$A5,$7F,$E0,$06,$D7,$EC,$7F,$09,$7E,$12,$F9,$4E,$1C
	.dc.b	$33,$0F,$F4,$16,$1C,$30,$FA,$ED,$7F,$97,$B5,$FF,$77,$78,$BF,$9C
	.dc.b	$F9,$3F,$CE,$B6,$BF,$DE,$8D,$9F,$F7,$5B,$74,$F8,$1F,$2F,$FE,$DB
	.dc.b	$AF,$9A,$FD,$B6,$9F,$1D,$C5,$7D,$07,$83,$78,$75,$3B,$6B,$D1,$B1
	.dc.b	$B8,$FF,$33,$F6,$7E,$56,$FF,$A2,$AF,$A2,$FD,$BF,$37,$EE,$1B,$5C
	.dc.b	$1B,$7F,$C4,$6B,$84,$FF,$8C,$3E,$A1,$F3,$BF,$AC,$6C,$7F,$A1,$7E
	.dc.b	$7D,$F7,$AF,$98,$D5,$CA,$FF,$5A,$5F,$0C,$7F,$28,$D7,$F5,$6D,$7F
	.dc.b	$AD,$62,$FA,$4F,$E1,$FE,$47,$F1,$06,$DF,$C0,$6D,$7A,$5B,$B7,$C2
	.dc.b	$6C,$F4,$6C,$FF,$64,$BF,$25,$7C,$8F,$A4,$FE,$3B,$97,$F1,$E6,$E1
	.dc.b	$F3,$9B,$7F,$1D,$FE,$47,$F4,$9F,$DA,$FE,$B7,$ED,$E1,$F4,$A3,$17
	.dc.b	$A3,$FD,$3B,$3A,$7D,$F1,$A7,$F1,$3F,$23,$F8,$7F,$80,$FC,$43,$63
	.dc.b	$27,$E3,$A3,$2D,$33,$8F,$E3,$63,$2D,$8F,$0F,$27,$9C,$F4,$7D,$1D
	.dc.b	$8F,$2F,$C7,$FA,$FE,$83,$D8,$D8,$F4,$FC,$47,$BC,$C6,$1F,$D6,$74
	.dc.b	$FF,$69,$4F,$BB,$E3,$7E,$21,$F1,$7B,$EC,$64,$F8,$7F,$1F,$8B,$E4
	.dc.b	$FC,$B2,$46,$E3,$2F,$CA,$FA,$0F,$A1,$BF,$D1,$6C,$64,$F9,$F1,$96
	.dc.b	$99,$9F,$44,$65,$B1,$F5,$B7,$FB,$9F,$11,$F7,$5B,$19,$3E,$DC,$65
	.dc.b	$A6,$67,$E2,$46,$5B,$1F,$82,$F8,$7F,$C4,$7C,$47,$E2,$5B,$19,$3F
	.dc.b	$0F,$19,$69,$99,$FA,$11,$96,$C7,$E3,$BE,$23,$F2,$9E,$77,$E5,$5B
	.dc.b	$19,$3F,$27,$19,$69,$99,$FB,$91,$96,$C7,$E6,$BD,$CF,$D0,$7B,$9F
	.dc.b	$A1,$6C,$64,$FC,$FC,$65,$A6,$67,$F1,$11,$96,$C7,$E9,$BD,$CF,$D6
	.dc.b	$7B,$7F,$AD,$6C,$64,$FD,$5C,$65,$A7,$F8,$73,$C1,$FF,$11,$1F,$E5
	.dc.b	$FF,$99,$FF,$0F,$5F,$F8,$BF,$DB,$FE,$31,$5B,$24,$7F,$8C,$E7,$7D
	.dc.b	$BF,$E7,$0A,$DC,$F1,$FA,$0F,$F1,$77,$C7,$FF,$8B,$DB,$87,$F8,$B1
	.dc.b	$B3,$B8,$FF,$17,$CF,$F8,$D7,$E8,$3F,$C6,$DF,$41,$67,$FC,$66,$DA
	.dc.b	$F3,$37,$1F,$13,$67,$FD,$01,$8A,$31,$5F,$E8,$08,$C4,$FF,$8F,$BE
	.dc.b	$24,$FF,$AD,$F2,$E3,$18,$AF,$F5,$BC,$74,$F8,$8C,$7C,$3A,$37,$7C
	.dc.b	$6B,$B5,$EA,$73,$1C,$8F,$15,$E5,$DD,$71,$5F,$42,$F0,$7F,$D7,$51
	.dc.b	$FE,$19,$8C,$7A,$FF,$E1,$B8,$DC,$E9,$FF,$19,$CE,$9F,$F1,$07,$C5
	.dc.b	$3E,$4D,$E9,$F1,$8C,$36,$33,$3C,$1A,$F4,$F1,$BF,$CD,$2F,$8D,$AE
	.dc.b	$7E,$1B,$C9,$82,$73,$3F,$EB,$3F,$3F,$0F,$C7,$F8,$74,$62,$3E,$34
	.dc.b	$62,$7F,$D8,$7C,$3F,$3E,$7F,$D1,$FE,$4E,$38,$7A,$55,$9F,$F4,$9B
	.dc.b	$6F,$94,$77,$C1,$79,$5E,$47,$FB,$43,$9B,$FD,$15,$E4,$5E,$3F,$43
	.dc.b	$70,$FF,$4F,$D5,$C7,$D2,$DC,$3F,$C6,$52,$3F,$D2,$F5,$99,$91,$9C
	.dc.b	$3B,$B7,$0E,$ED,$3B,$97,$0F,$8F,$E4,$77,$F2,$24,$E7,$24,$F8,$CC
	.dc.b	$C7,$FA,$A2,$70,$EF,$C8,$F4,$11,$ED,$EB,$BC,$3C,$47,$FA,$C2,$6E
	.dc.b	$7F,$EF,$0A,$D4,$F0,$F0,$BE,$24,$7F,$AD,$63,$9F,$85,$89,$91,$82
	.dc.b	$1F,$EB,$F9,$47,$CD,$F9,$BF,$C6,$DF,$FB,$F9,$7C,$46,$4D,$73,$CC
	.dc.b	$3E,$34,$E9,$B3,$97,$FD,$99,$1D,$38,$3E,$B7,$02,$38,$4E,$AB,$86
	.dc.b	$1E,$1F,$13,$C9,$CD,$2B,$BC,$79,$CC,$3D,$A3,$0E,$8F,$2B,$A9,$E9
	.dc.b	$1E,$8F,$0F,$E0,$78,$72,$30,$43,$D2,$E5,$F5,$C7,$C7,$31,$3F,$1E
	.dc.b	$71,$3F,$13,$C3,$F9,$9E,$14,$8C,$10,$F8,$AE,$5F,$8F,$1F,$4C,$C5
	.dc.b	$7D,$39,$C4,$FD,$0F,$0B,$EC,$77,$48,$C1,$0F,$A2,$E5,$FA,$71,$F7
	.dc.b	$CC,$57,$DF,$9C,$4F,$DC,$FA,$1F,$C2,$11,$FB,$FF,$43,$F8,$73,$0F
	.dc.b	$DE,$8C,$3A,$21,$F8,$28,$FC,$77,$BB,$F8,$FF,$A1,$93,$F8,$A7,$F2
	.dc.b	$5E,$47,$E4,$FE,$86,$CF,$E3,$A3,$0F,$E4,$3E,$83,$F3,$04,$7F,$27
	.dc.b	$F3,$9F,$C2,$57,$D0,$A7,$0A,$69,$88,$93,$FC,$0C,$E3,$1A,$71,$3C
	.dc.b	$DB,$39,$5A,$FF,$5B,$78,$9A,$3F,$4F,$F2,$DF,$AF,$F9,$79,$18,$21
	.dc.b	$FA,$87,$2F,$EA,$A3,$F6,$DF,$11,$FC,$8F,$DD,$FE,$41,$3F,$EB,$23
	.dc.b	$27,$ED,$23,$4C,$30,$7F,$AE,$3C,$FE,$5F,$19,$23,$04,$3F,$D6,$8E
	.dc.b	$5F,$F5,$B4,$78,$4C,$63,$FE,$C0,$9C,$4F,$C9,$FA,$17,$89,$F4,$3B
	.dc.b	$24,$E6,$3E,$B4,$EA,$3F,$B7,$35,$F6,$9F,$CF,$5C,$38,$6B,$FE,$B8
	.dc.b	$F9,$7F,$89,$F4,$32,$3F,$DA,$B1,$F1,$CC,$3F,$EC,$F8,$C3,$F7,$FE
	.dc.b	$87,$FA,$BB,$7C,$CF,$C5,$47,$F5,$88,$D7,$9D,$83,$33,$23,$24,$3F
	.dc.b	$07,$E2,$FD,$6A,$E5,$18,$3F,$AC,$38,$D7,$F1,$0F,$8A,$3F,$DC,$9F
	.dc.b	$43,$F7,$FE,$66,$47,$FB,$C2,$3F,$04,$61,$FF,$75,$46,$1F,$CD,$7A
	.dc.b	$DF,$9F,$7F,$39,$E7,$B9,$7F,$4C,$D8,$DF,$4F,$EA,$6B,$84,$7D,$23
	.dc.b	$FC,$4D,$1F,$E2,$36,$9F,$E2,$58,$DB,$F6,$3E,$87,$D5,$9F,$E8,$DF
	.dc.b	$76,$9F,$98,$6C,$FF,$79,$77,$19,$DB,$F7,$51,$A9,$34,$DB,$C6,$37
	.dc.b	$B4,$8F,$DD,$4E,$97,$D0,$FF,$8D,$5F,$DF,$78,$7F,$E0,$A7,$EA,$07
	.dc.b	$18,$FE,$EE,$78,$4E,$9A,$FE,$59,$AF,$ED,$1A,$FF,$85,$5A,$BC,$77
	.dc.b	$BB,$FC,$5F,$E8,$7B,$E9,$FB,$C6,$D7,$F8,$76,$B6,$FF,$15,$6B,$F5
	.dc.b	$5A,$FE,$A1,$AF,$F0,$66,$AF,$35,$DA,$FA,$1F,$F1,$87,$CC,$7F,$8E
	.dc.b	$3D,$4B,$C3,$FE,$10,$C3,$F3,$18,$70,$BF,$E2,$46,$AA,$4F,$F8,$A7
	.dc.b	$0E,$8D,$FA,$1F,$E3,$04,$75,$F4,$35,$EB,$29,$C7,$07,$FA,$17,$1F
	.dc.b	$3D,$FF,$90,$7E,$17,$FC,$85,$F0,$B3,$1F,$F1,$A4,$65,$FF,$1B,$46
	.dc.b	$5F,$4F,$B4,$9F,$F2,$57,$C2,$FF,$92,$F8,$E5,$7F,$8F,$63,$2F,$F8
	.dc.b	$FE,$32,$F2,$8C,$BD,$23,$39,$79,$ED,$5A,$8C,$4E,$B8,$FF,$CE,$D3
	.dc.b	$FE,$5B,$F8,$5F,$F2,$E7,$C2,$CC,$7F,$C9,$D1,$97,$FC,$A1,$19,$7D
	.dc.b	$71,$FE,$65,$FA,$1F,$DD,$47,$D6,$5E,$56,$4D,$77,$C9,$FF,$2F,$4E
	.dc.b	$9B,$39,$79,$4F,$17,$27,$07,$FC,$E9,$F4,$0F,$26,$F8,$D2,$BF,$CD
	.dc.b	$53,$96,$CE,$5F,$C5,$C2,$79,$CC,$9E,$31,$BD,$BF,$05,$5C,$E7,$8D
	.dc.b	$67,$5F,$F2,$6C,$F3,$9E,$E7,$F7,$BF,$85,$FD,$E5,$FF,$A1,$FE,$85
	.dc.b	$E2,$FF,$6A,$97,$AF,$FD,$B2,$73,$7F,$E8,$D9,$CB,$C6,$B8,$78,$3F
	.dc.b	$E0,$08,$57,$05,$5E,$07,$35,$DE,$FB,$36,$3A,$1D,$EF,$A5,$F3,$BE
	.dc.b	$4D,$BE,$35,$E5,$6B,$E0,$AE,$37,$E6,$C3,$AC,$F9,$2B,$78,$7B,$23
	.dc.b	$CA,$78,$67,$71,$E0,$AF,$34,$E8,$E4,$6A,$FB,$BC,$47,$BB,$FE,$7E
	.dc.b	$AF,$84,$D9,$F9,$0D,$BF,$F5,$D3,$FE,$84,$AF,$F5,$E7,$C2,$63,$DD
	.dc.b	$FE,$A4,$D8,$FF,$02,$B5,$FF,$6C,$9E,$08,$DB,$E7,$1E,$83,$6F,$87
	.dc.b	$1D,$8D,$46,$5E,$02,$B7,$19,$D7,$EF,$4F,$16,$C6,$74,$F0,$C6,$2F
	.dc.b	$BD,$73,$3B,$B8,$BE,$AD,$8C,$9E,$28,$CB,$4C,$CF,$E5,$BB,$1B,$BE
	.dc.b	$77,$D5,$AF,$EA,$3E,$5A,$E5,$FA,$69,$B8,$FF,$43,$B5,$63,$0E,$75
	.dc.b	$67,$37,$FB,$C8,$C5,$65,$B1,$93,$F8,$0C,$65,$A6,$6F,$F8,$0C,$77
	.dc.b	$FA,$17,$D0,$5B,$C3,$6F,$FD,$E9,$3A,$6C,$E9,$A6,$A7,$85,$FF,$58
	.dc.b	$AE,$0D,$AF,$77,$6F,$FB,$C6,$FF,$DF,$9F,$19,$8F,$13,$2D,$8C,$9F
	.dc.b	$F0,$06,$32,$AF,$FA,$6F,$C2,$74,$F8,$6E,$AD,$9D,$6F,$FD,$16,$31
	.dc.b	$5C,$27,$27,$FB,$9B,$15,$F2,$8C,$CF,$F7,$99,$CC,$63,$74,$7F,$BF
	.dc.b	$23,$0E,$A7,$B1,$FD,$8A,$BA,$36,$79,$7C,$67,$EE,$1B,$CB,$F8,$38
	.dc.b	$C4,$FF,$80,$94,$67,$9B,$4D,$AE,$27,$F8,$0F,$E8,$7F,$19,$E4,$1B
	.dc.b	$C5,$86,$4E,$CF,$F0,$54,$63,$1F,$F0,$2F,$0F,$E6,$3C,$B7,$D0,$FF
	.dc.b	$51,$BF,$F0,$84,$E6,$31,$5C,$27,$E4,$D7,$0E,$EF,$1D,$ED,$FD,$1D
	.dc.b	$3E,$85,$F3,$7F,$EB,$B8,$FF,$57,$C6,$30,$FF,$32,$46,$E7,$4F,$E0
	.dc.b	$E7,$4F,$F9,$A5,$BC,$9F,$C3,$7B,$5E,$33,$E0,$BC,$35,$AC,$7F,$19
	.dc.b	$5A,$9F,$2F,$D0,$3D,$FE,$F9,$23,$17,$E4,$6C,$7F,$9E,$5A,$FF,$AB
	.dc.b	$1C,$E5,$FC,$7F,$67,$E2,$FF,$E0,$8B,$FF,$71,$56,$9F,$CF,$4E,$BC
	.dc.b	$8F,$E1,$3F,$43,$FE,$D2,$9F,$CA,$BE,$86,$E3,$FA,$56,$E9,$C2,$30
	.dc.b	$7F,$A8,$E7,$83,$4C,$E5,$FC,$CE,$32,$D8,$CD,$7E,$C6,$72,$D8,$C8
	.dc.b	$8C,$BF,$B7,$F9,$F7,$C3,$FC,$A6,$D7,$DC,$6C,$FF,$A8,$5A,$FF,$B5
	.dc.b	$1C,$DF,$F4,$7F,$9E,$7E,$03,$FE,$7D,$BF,$DB,$DA,$F7,$1F,$C1,$EB
	.dc.b	$7E,$C7,$E9,$3C,$5F,$D2,$FC,$8D,$AF,$F3,$53,$70,$FF,$32,$B7,$2F
	.dc.b	$F3,$27,$62,$3E,$9F,$CB,$7F,$5D,$BF,$A6,$BE,$67,$31,$A6,$BA,$6E
	.dc.b	$1F,$4E,$78,$4F,$23,$81,$4E,$07,$F9,$FF,$A7,$F7,$BF,$41,$FB,$E6
	.dc.b	$C6,$4F,$F4,$24,$65,$A7,$C6,$F9,$F7,$CF,$7E,$71,$B1,$FA,$E6,$E1
	.dc.b	$FE,$D3,$6B,$FC,$C4,$FE,$64,$EB,$4F,$EC,$3F,$58,$F2,$FF,$7B,$39
	.dc.b	$BF,$E6,$73,$97,$FD,$31,$F4,$1F,$C5,$BC,$9F,$E2,$ED,$8C,$9F,$E9
	.dc.b	$D8,$CB,$4C,$E3,$FE,$9A,$8C,$B6,$3F,$D4,$99,$7F,$22,$FA,$0F,$E4
	.dc.b	$6D,$8C,$9F,$EA,$98,$CB,$4F,$F5,$67,$D0,$7F,$75,$F9,$FF,$BD,$7F
	.dc.b	$EA,$88,$C4,$FF,$2E,$6B,$C0,$FE,$8D,$3C,$35,$FD,$67,$03,$E3,$FF
	.dc.b	$13,$EF,$FE,$29,$B7,$F8,$A6,$E3,$F8,$56,$DF,$26,$CF,$EC,$EF,$A3
	.dc.b	$F9,$2B,$F4,$71,$C8,$F1,$FD,$05,$C7,$FD,$9B,$93,$E2,$BD,$DF,$3F
	.dc.b	$EF,$6D,$FA,$BE,$1F,$E1,$18,$BE,$2F,$C5,$F8,$B9,$3F,$03,$E7,$ED
	.dc.b	$76,$9F,$95,$F1,$D8,$E0,$FF,$67,$66,$8C,$E1,$F2,$1D,$3F,$DF,$FD
	.dc.b	$64,$EE,$3E,$6F,$CF,$FD,$5E,$2F,$AC,$D8,$C9,$F4,$A3,$2D,$33,$97
	.dc.b	$C2,$8C,$B6,$3E,$CF,$0F,$E0,$7E,$23,$F0,$4D,$8C,$9F,$7A,$32,$D3
	.dc.b	$33,$F3,$A3,$2D,$8F,$C2,$FC,$3F,$E3,$7E,$23,$F1,$CD,$8C,$9F,$8A
	.dc.b	$8C,$B4,$CC,$FD,$A8,$CB,$63,$F2,$3F,$11,$F9,$9F,$4B,$F3,$4D,$8C
	.dc.b	$9F,$96,$8C,$B4,$CC,$FE,$1A,$32,$D8,$FC,$EF,$BB,$FA,$5F,$03,$F4
	.dc.b	$CD,$8C,$9F,$A2,$8C,$B4,$CC,$FE,$4A,$32,$D8,$FD,$4F,$BB,$FB,$1F
	.dc.b	$77,$F6,$4D,$8C,$9F,$AE,$8C,$B4,$FF,$16,$FC,$A7,$F8,$AC,$FF,$37
	.dc.b	$7C,$BF,$F8,$B3,$97,$FD,$21,$C8,$F7,$D5,$BB,$E2,$7F,$8F,$EF,$83
	.dc.b	$F9,$39,$DB,$FE,$3A,$F9,$EF,$F1,$C7,$99,$FE,$39,$6C,$64,$FF,$1F
	.dc.b	$C6,$5A,$7F,$90,$BE,$7B,$FC,$81,$F3,$DF,$E4,$16,$C6,$4F,$F2,$3C
	.dc.b	$65,$A6,$76,$FF,$23,$46,$5B,$1F,$E4,$EF,$8A,$FF,$57,$7C,$3B,$1F
	.dc.b	$54,$61,$F6,$7C,$4D,$B6,$DF,$16,$B0,$DC,$3F,$D8,$F7,$C9,$B5,$C3
	.dc.b	$46,$1C,$1F,$F7,$FC,$EA,$5F,$41,$FC,$47,$D8,$FF,$1E,$B6,$3F,$A7
	.dc.b	$1F,$E7,$F7,$0E,$1A,$FF,$53,$F8,$B7,$CF,$BC,$9F,$F2,$3C,$61,$B1
	.dc.b	$99,$EA,$27,$4E,$29,$E3,$FF,$BB,$23,$0D,$7C,$4D,$7F,$B2,$47,$0F
	.dc.b	$88,$EB,$B7,$E3,$70,$FE,$D1,$C1,$F4,$6F,$E3,$7C,$FE,$38,$3F,$18
	.dc.b	$D8,$F9,$55,$6F,$94,$7C,$E7,$F3,$13,$9D,$CD,$AC,$7C,$FF,$F7,$7E
	.dc.b	$8F,$CF,$D2,$31,$C3,$F9,$B6,$E3,$F6,$7D,$CB,$97,$E9,$66,$DF,$1A
	.dc.b	$FF,$4C,$C6,$2B,$2C,$8C,$9F,$E9,$E8,$E3,$F0,$72,$3F,$D4,$2F,$4E
	.dc.b	$9C,$47,$FA,$7E,$2E,$5E,$89,$F0,$FC,$FF,$8B,$E7,$A5,$FF,$AA,$A3
	.dc.b	$2F,$FA,$B2,$32,$FF,$AA,$63,$D7,$2E,$BA,$49,$FD,$F7,$5D,$58,$7F
	.dc.b	$06,$AE,$11,$FC,$0E,$B9,$D7,$4E,$B9,$3F,$C1,$DE,$FD,$78,$8E,$7C
	.dc.b	$1F,$D6,$F0,$56,$71,$ED,$76,$B5,$36,$FF,$67,$6A,$F5,$F5,$C9,$FE
	.dc.b	$2C,$FB,$B8,$B1,$1E,$03,$8C,$63,$1F,$64,$DC,$7C,$93,$6B,$BC,$FC
	.dc.b	$8E,$39,$1F,$C8,$DF,$97,$C3,$88,$ED,$C6,$CF,$F2,$71,$B3,$F9,$54
	.dc.b	$EF,$2F,$99,$77,$2F,$89,$3F,$5B,$8A,$5F,$F3,$17,$EC,$F1,$62,$3B
	.dc.b	$E3,$FE,$8C,$74,$7F,$37,$8D,$63,$F6,$6A,$E1,$F4,$67,$F0,$77,$2F
	.dc.b	$F9,$F3,$F8,$6E,$1C,$47,$68,$B8,$7E,$0E,$6D,$FD,$B9,$B5,$F5,$A7
	.dc.b	$F1,$FC,$5F,$90,$F9,$A9,$3F,$D2,$E3,$2D,$21,$EC,$F9,$3F,$CB,$9A
	.dc.b	$7D,$F1,$87,$FA,$A7,$C1,$FE,$70,$8F,$F5,$7F,$87,$FC,$F9,$1C,$03
	.dc.b	$EB,$FD,$CF,$F2,$ED,$3F,$B1,$AF,$03,$F8,$0B,$9D,$7F,$C6,$4D,$AF
	.dc.b	$E8,$AE,$DF,$F5,$8F,$D0,$7F,$B9,$DB,$87,$F0,$C5,$19,$7F,$28,$3E
	.dc.b	$5F,$A3,$A9,$E1,$3B,$C5,$87,$FA,$46,$30,$FF,$A4,$EF,$8C,$62,$FD
	.dc.b	$CD,$7A,$1F,$EC,$68,$F5,$72,$31,$7C,$67,$C0,$25,$7C,$7C,$7E,$2F
	.dc.b	$CB,$62,$3E,$4C,$D9,$CC,$6F,$2F,$8F,$F3,$FF,$28,$91,$93,$E4,$FC
	.dc.b	$F6,$1F,$A7,$16,$BE,$9C,$FD,$2F,$98,$FA,$7F,$37,$2F,$96,$9E,$99
	.dc.b	$D4,$63,$45,$6E,$3E,$FC,$E9,$FC,$0C,$67,$0F,$C0,$CF,$F4,$2F,$8A
	.dc.b	$FA,$9F,$03,$77,$FF,$63,$E1,$E9,$9C,$BD,$FE,$6F,$55,$23,$55,$C1
	.dc.b	$FC,$7C,$FE,$0F,$E8,$3F,$D4,$51,$FD,$C8,$38,$7F,$BA,$B5,$FF,$01
	.dc.b	$B5,$FF,$69,$3B,$C7,$F0,$B1,$F9,$D8,$C6,$FA,$F6,$3F,$17,$E4,$FE
	.dc.b	$4F,$E8,$1E,$BD,$26,$1F,$54,$43,$D7,$CC,$7E,$62,$38,$BB,$6B,$98
	.dc.b	$DE,$5E,$48,$E7,$58,$C7,$FD,$D5,$7D,$67,$EC,$69,$FB,$02,$3F,$6F
	.dc.b	$E6,$2B,$FB,$88,$B8,$7E,$E2,$7F,$5D,$F0,$3F,$CC,$2B,$F5,$AD,$9F
	.dc.b	$13,$70,$FF,$7D,$7D,$CF,$E1,$AB,$E0,$4E,$5F,$E0,$CD,$7F,$D8,$DE
	.dc.b	$37,$EE,$E7,$84,$7E,$EB,$DF,$FF,$12,$61,$EF,$6E,$1F,$C3,$E7,$FD
	.dc.b	$7D,$B7,$F0,$39,$7C,$D5,$89,$3F,$46,$78,$BF,$C5,$DD,$B9,$8D,$BE
	.dc.b	$17,$A4,$FF,$0F,$FB,$77,$B9,$49,$5F,$77,$E2,$FF,$CA,$AF,$DF,$9D
	.dc.b	$B6,$76,$FF,$16,$9D,$BF,$C6,$3F,$1D,$73,$7F,$1D,$6C,$7E,$25,$AF
	.dc.b	$F1,$EF,$BE,$7C,$EE,$EA,$4E,$DF,$E4,$D3,$A8,$FD,$D4,$FF,$32,$23
	.dc.b	$FB,$DF,$9C,$AF,$F3,$A8,$BB,$FF,$3A,$9F,$E7,$3F,$2B,$FC,$E9,$95
	.dc.b	$FC,$16,$32,$F3,$21,$FD,$11,$FF,$1F,$F9,$72,$3F,$87,$FA,$E8,$FF
	.dc.b	$21,$F9,$99,$9F,$F0,$0C,$E5,$DC,$E9,$E3,$E5,$FF,$92,$88,$F4,$F2
	.dc.b	$AB,$CA,$2E,$5F,$E5,$89,$BC,$BF,$EB,$19,$B5,$FE,$8F,$9B,$E2,$FF
	.dc.b	$AE,$E6,$E3,$FD,$0E,$1F,$BF,$7F,$40,$25,$FF,$97,$9F,$EB,$4F,$F8
	.dc.b	$69,$CB,$FC,$B9,$E1,$E6,$6F,$CA,$93,$97,$97,$97,$88,$E5,$17,$4F
	.dc.b	$F6,$7D,$AB,$1A,$7F,$73,$C7,$CB,$78,$C3,$FD,$63,$7B,$7C,$5E,$5F
	.dc.b	$9C,$8F,$97,$CB,$AF,$A6,$2E,$5E,$09,$B8,$F7,$B5,$62,$FB,$D5,$C7
	.dc.b	$E3,$CD,$A4,$FF,$7B,$F7,$7F,$D7,$B8,$FF,$82,$FE,$85,$E5,$FF,$AC
	.dc.b	$FD,$7F,$E0,$11,$8C,$FF,$C2,$17,$FE,$59,$78,$57,$FA,$E7,$D7,$FF
	.dc.b	$37,$C2,$B8,$3F,$EB,$58,$CC,$62,$B9,$4E,$EB,$C0,$F4,$3B,$D7,$08
	.dc.b	$ED,$86,$67,$FD,$81,$87,$9A,$78,$C7,$33,$8C,$EE,$FD,$F1,$93,$89
	.dc.b	$9A,$E4,$F0,$35,$1C,$30,$F4,$4E,$A7,$89,$AC,$3C,$CF,$1F,$03,$F8
	.dc.b	$FB,$F0,$70,$FF,$1C,$46,$30,$F1,$B6,$BE,$4B,$C2,$BF,$C7,$DE,$E7
	.dc.b	$0F,$07,$FD,$BE,$D8,$FF,$76,$B5,$FF,$6A,$89,$E4,$A3,$E4,$3F,$E4
	.dc.b	$5E,$0A,$E4,$F8,$86,$4A,$67,$3F,$B6,$E5,$AE,$74,$FF,$7F,$F1,$FF
	.dc.b	$7A,$9C,$DF,$82,$31,$3E,$AA,$EF,$18,$7D,$66,$2F,$CB,$5C,$5B,$19
	.dc.b	$3C,$B1,$97,$F0,$1C,$3F,$C1,$AF,$FC,$17,$AF,$EE,$1B,$7F,$A9,$F3
	.dc.b	$ED,$FE,$DA,$6D,$F1,$AF,$E1,$11,$8E,$1F,$F1,$9C,$60,$95,$FC,$3E
	.dc.b	$75,$7B,$64,$EB,$0F,$E1,$33,$8A,$FE,$2D,$DD,$D4,$ED,$E4,$5B,$FE
	.dc.b	$31,$16,$7F,$8E,$4D,$BF,$E2,$73,$6B,$F8,$24,$DA,$FE,$23,$E4,$7A
	.dc.b	$4F,$57,$7D,$AF,$E4,$91,$67,$F9,$44,$DC,$3F,$8F,$F7,$FC,$13,$E1
	.dc.b	$77,$DC,$3F,$96,$45,$BF,$E6,$13,$6B,$8D,$68,$FE,$65,$19,$7F,$8D
	.dc.b	$55,$AF,$E4,$F3,$76,$FE,$7F,$F2,$DF,$44,$FA,$5F,$2D,$70,$FE,$79
	.dc.b	$16,$FF,$A0,$4D,$C7,$FD,$47,$3F,$73,$D2,$4D,$8C,$36,$FF,$CE,$0D
	.dc.b	$BF,$F5,$6B,$C7,$E5,$A4,$E8,$FF,$3B,$38,$9F,$F5,$9B,$D2,$32,$7F
	.dc.b	$94,$FF,$45,$FD,$A3,$79,$FF,$23,$CD,$FE,$1E,$F0,$BF,$27,$82,$F6
	.dc.b	$FF,$AB,$9B,$A7,$F9,$4D,$B8,$7F,$B0,$EF,$93,$D4,$FF,$31,$7E,$D2
	.dc.b	$DF,$AB,$63,$FD,$35,$2F,$AC,$7C,$26,$AE,$57,$27,$CF,$3A,$8C,$1F
	.dc.b	$EA,$3F,$9D,$FF,$6A,$74,$FE,$B1,$B1,$9D,$3F,$52,$34,$FF,$3F,$CE
	.dc.b	$A3,$13,$FE,$DD,$8C,$4E,$A7,$FA,$AC,$F3,$E9,$ED,$18,$38,$8A,$E3
	.dc.b	$3C,$E3,$55,$CE,$B8,$DF,$D0,$F9,$C7,$8F,$4F,$7F,$9B,$71,$FF,$46
	.dc.b	$4E,$9B,$3A,$69,$FE,$B9,$F9,$AF,$C8,$1F,$91,$F9,$AB,$19,$26,$7F
	.dc.b	$EF,$A8,$CB,$EA,$D1,$F1,$2F,$97,$F9,$8D,$BF,$A0,$BD,$6C,$CC,$9C
	.dc.b	$BC,$23,$17,$D6,$B9,$36,$75,$5C,$0F,$F5,$57,$CD,$BD,$DF,$CA,$B6
	.dc.b	$3F,$D9,$5F,$54,$F8,$DC,$D4,$9F,$F6,$25,$6F,$0F,$F4,$50,$0E,$5C
	.dc.b	$B5,$D7,$C0,$3F,$39,$ED,$5F,$ED,$96,$CF,$E0,$3C,$27,$85,$DF,$9F
	.dc.b	$FD,$97,$87,$1C,$FF,$9F,$B7,$1E,$51,$FC,$17,$3C,$E5,$D2,$FF,$DA
	.dc.b	$78,$F4,$F9,$DF,$E1,$5F,$3B,$FC,$2D,$B8,$FF,$BD,$DB,$A7,$8F,$44
	.dc.b	$63,$0E,$71,$8A,$FF,$7F,$5F,$36,$9F,$16,$7C,$3F,$BC,$A3,$FA,$AB
	.dc.b	$5D,$1F,$E5,$15,$F3,$DA,$88,$6B,$16,$4A,$E5,$1F,$25,$E3,$3B,$9D
	.dc.b	$3E,$0E,$18,$E6,$7F,$20,$D8,$CD,$73,$7C,$FC,$78,$C7,$F7,$FC,$6F
	.dc.b	$98,$FC,$D9,$F9,$CF,$83,$B8,$FE,$72,$2E,$3F,$97,$9B,$7C,$A6,$D7
	.dc.b	$F0,$7B,$56,$2F,$DF,$97,$F1,$77,$DD,$F3,$DF,$2B,$C1,$F9,$6F,$CC
	.dc.b	$F9,$D9,$8F,$B6,$2D,$7C,$5F,$9E,$FA,$26,$34,$FA,$AF,$D4,$F8,$3F
	.dc.b	$AB,$F3,$B7,$0F,$8C,$E1,$A7,$C9,$8C,$5F,$D1,$9F,$BB,$F1,$38,$C3
	.dc.b	$F0,$2F,$E0,$3E,$76,$61,$F4,$78,$2E,$1F,$66,$6D,$FE,$22,$6D,$7E
	.dc.b	$16,$6D,$7C,$49,$BC,$7F,$83,$E0,$FC,$69,$F8,$EF,$8A,$B8,$7D,$F8
	.dc.b	$B7,$F8,$79,$B8,$FE,$5A,$6D,$7E,$4E,$6D,$7D,$2F,$9D,$FC,$C9,$F9
	.dc.b	$AF,$16,$D7,$E2,$E2,$CF,$E4,$E6,$E1,$FA,$29,$B5,$F9,$F9,$B5,$F7
	.dc.b	$BE,$77,$F4,$A7,$E9,$BE,$76,$D7,$E5,$E2,$CF,$E7,$E6,$E1,$FA,$E9
	.dc.b	$B5,$FA,$B9,$B5,$F8,$AF,$9D,$FD,$89,$FB,$2F,$9D,$B5,$FA,$38,$B3
	.dc.b	$FA,$B9,$B8,$7E,$EA,$6D,$7E,$DE,$6D,$7E,$D3,$F7,$AF,$EF,$6D,$F4
	.dc.b	$FF,$CA,$70,$F7,$EC,$4B,$FD,$B4,$F0,$E0,$3B,$8C,$CF,$F9,$67,$D5
	.dc.b	$FF,$2D,$79,$76,$7F,$DC,$8D,$C7,$FD,$D0,$DD,$3F,$DB,$EF,$F9,$87
	.dc.b	$CB,$95,$FE,$EB,$F1,$AD,$FF,$BB,$E6,$DF,$FB,$9A,$6D,$7F,$B5,$FC
	.dc.b	$5F,$F3,$51,$FE,$6B,$F1,$2C,$7F,$BD,$62,$E1,$FE,$F9,$9B,$7F,$EF
	.dc.b	$09,$B5,$FE,$E1,$9B,$5F,$C7,$3F,$7F,$FE,$37,$EF,$BE,$7B,$FD,$E9
	.dc.b	$1F,$C7,$27,$FD,$45,$18,$BF,$E5,$91,$83,$97,$8F,$8B,$E5,$1D,$67
	.dc.b	$15,$D2,$AC,$E6,$7F,$D6,$F3,$96,$C7,$F5,$FF,$9F,$FE,$ED,$EB,$7F
	.dc.b	$62,$8C,$79,$08,$C4,$27,$6E,$DA,$FF,$6A,$F8,$47,$CE,$7F,$81,$A3
	.dc.b	$A4,$63,$CA,$E6,$D8,$CC,$ED,$FE,$E7,$5F,$E4,$0E,$5F,$0B,$63,$C1
	.dc.b	$E4,$FF,$B2,$DB,$3E,$9E,$87,$C5,$F8,$A7,$DF,$18,$F9,$3F,$EF,$D1
	.dc.b	$F3,$63,$0E,$A7,$83,$96,$F0,$FD,$09,$DC,$E6,$B9,$CE,$6B,$7C,$1F
	.dc.b	$E0,$BB,$F1,$FC,$E7,$83,$9B,$FB,$CD,$23,$1C,$5B,$8C,$65,$FD,$7D
	.dc.b	$7C,$E6,$A3,$FB,$3C,$69,$FF,$07,$36,$FA,$DE,$F0,$F8,$77,$6B,$58
	.dc.b	$77,$AC,$19,$9D,$7C,$E4,$C7,$C4,$F1,$F9,$CC,$47,$82,$2E,$5E,$C9
	.dc.b	$ED,$F3,$9D,$FE,$72,$5F,$AE,$32,$FB,$63,$2F,$AA,$3D,$12,$F9,$FA
	.dc.b	$49,$FF,$01,$43,$E7,$F1,$1F,$E2,$F8,$E0,$EA,$79,$CF,$4F,$9F,$93
	.dc.b	$C9,$EF,$F2,$38,$8E,$51,$72,$FF,$17,$E0,$AC,$6D,$F7,$61,$58,$BF
	.dc.b	$F0,$DC,$6D,$F3,$FC,$8F,$B0,$8F,$AB,$A2,$BE,$98,$B8,$F8,$26,$E1
	.dc.b	$DE,$6D,$76,$9F,$8F,$CB,$23,$E2,$BF,$2B,$87,$11,$F1,$62,$E1,$EC
	.dc.b	$9B,$87,$AE,$6D,$7A,$A7,$E9,$F0,$C8,$FA,$2F,$D6,$E1,$C4,$7D,$18
	.dc.b	$B8,$7C,$89,$B8,$7C,$79,$B5,$F1,$A7,$EF,$E7,$23,$EE,$BF,$82,$E0
	.dc.b	$C4,$7D,$D8,$B8,$7D,$49,$B8,$7D,$39,$B5,$F4,$A7,$F1,$7C,$3F,$8C
	.dc.b	$F9,$69,$3F,$8A,$8C,$B4,$87,$E2,$FE,$5F,$F2,$66,$9F,$C7,$46,$1F
	.dc.b	$C9,$FC,$17,$E6,$08,$FE,$5B,$E2,$BF,$36,$47,$07,$F9,$6F,$DD,$FF
	.dc.b	$48,$E9,$FE,$72,$F0,$3F,$71,$18,$D7,$F8,$53,$6B,$FC,$CE,$FF,$35
	.dc.b	$7F,$CD,$43,$E7,$D8,$7F,$07,$8C,$36,$33,$1C,$3E,$47,$A3,$A3,$9C
	.dc.b	$69,$FF,$6C,$99,$2C,$70,$8D,$63,$FD,$62,$74,$D8,$CE,$1F,$CC,$2B
	.dc.b	$D5,$D7,$FE,$A4,$F9,$7F,$79,$30,$F8,$9F,$31,$5F,$F4,$C4,$5C,$7F
	.dc.b	$D2,$F3,$F1,$BE,$62,$57,$CB,$7E,$4F,$CC,$62,$3F,$D4,$11,$6B,$FD
	.dc.b	$3F,$3F,$3B,$E2,$FE,$7F,$C6,$4B,$FA,$91,$97,$EB,$46,$5F,$F5,$54
	.dc.b	$65,$FF,$56,$46,$6B,$FD,$57,$3F,$CD,$FE,$21,$A7,$FB,$5E,$FF,$C4
	.dc.b	$BE,$0F,$F3,$5F,$4F,$F8,$BC,$30,$FF,$62,$D6,$E3,$1B,$FF,$B4,$23
	.dc.b	$13,$AF,$43,$B4,$62,$7F,$0F,$F2,$DF,$81,$F9,$F7,$C0,$FE,$07,$E2
	.dc.b	$A6,$5F,$88,$9C,$B6,$7F,$BF,$A8,$CB,$FE,$05,$6B,$FE,$E8,$6B,$FD
	.dc.b	$B3,$E7,$FF,$D3,$EF,$E5,$1A,$FE,$29,$B9,$FE,$1D,$7C,$07,$E5,$DC
	.dc.b	$E5,$CF,$E0,$70,$66,$6D,$F5,$8A,$F5,$8F,$D4,$E9,$23,$F5,$CF,$EB
	.dc.b	$3E,$63,$11,$FE,$E3,$8B,$8F,$FB,$8A,$7F,$51,$EA,$3C,$7F,$D5,$7C
	.dc.b	$FF,$DA,$C3,$F5,$ED,$7F,$4B,$3B,$7F,$66,$D7,$33,$D0,$E7,$CD,$FB
	.dc.b	$5F,$87,$FF,$0D,$4F,$EB,$9B,$8F,$EA,$DB,$AF,$F8,$61,$FF,$38,$4F
	.dc.b	$F0,$07,$28,$38,$6F,$A5,$DF,$DE,$F0,$32,$39,$39,$7F,$85,$FC,$FB
	.dc.b	$E1,$3F,$85,$7C,$FF,$AF,$2F,$DF,$37,$2F,$F1,$5D,$6E,$F8,$DF,$F1
	.dc.b	$18,$C4,$7F,$0A,$6D,$F2,$F4,$3F,$8B,$7C,$FB,$A7,$F8,$C7,$CF,$BE
	.dc.b	$5E,$9A,$89,$8F,$F2,$0A,$E0,$D8,$E3,$F3,$BF,$CA,$89,$1F,$CB,$DF
	.dc.b	$E5,$BF,$3B,$88,$FD,$69,$FC,$B2,$31,$97,$F9,$CB,$C7,$79,$4B,$FE
	.dc.b	$6B,$97,$2C,$3F,$9D,$46,$1F,$E7,$91,$97,$AD,$72,$9D,$47,$F8,$E3
	.dc.b	$CF,$FF,$1A,$11,$FF,$20,$79,$F5,$D1,$FE,$43,$8F,$F1,$D4,$7F,$8E
	.dc.b	$FC,$E9,$3F,$E4,$77,$FC,$7D,$E6,$E2,$3F,$C8,$B1,$73,$FF,$28,$CD
	.dc.b	$E3,$FF,$28,$4D,$AF,$F2,$6C,$DC,$BF,$CD,$73,$7D,$9F,$E7,$74,$97
	.dc.b	$CF,$7F,$3C,$24,$7F,$94,$5F,$EA,$CF,$C2,$72,$FE,$5F,$1F,$9D,$1C
	.dc.b	$7D,$0E,$1E,$7C,$AD,$67,$FE,$6B,$A4,$F5,$F3,$F5,$5D,$62,$E1,$D2
	.dc.b	$6D,$F1,$9B,$5E,$9B,$57,$9F,$CF,$93,$E0,$34,$FA,$FC,$FC,$C7,$AE
	.dc.b	$2E,$3E,$A9,$B8,$F9,$E6,$D7,$79,$B9,$FC,$89,$B8,$A7,$FB,$C7,$BE
	.dc.b	$E9,$FE,$89,$83,$E7,$DE,$8F,$FA,$6A,$BF,$C9,$D3,$98,$C7,$B7,$FE
	.dc.b	$9E,$AD,$CF,$F0,$5A,$E9,$AF,$F0,$79,$CC,$A7,$2F,$33,$85,$73,$BE
	.dc.b	$B1,$C5,$F1,$61,$E6,$BE,$77,$C9,$B7,$C6,$F9,$9C,$4F,$25,$71,$BE
	.dc.b	$78,$66,$7C,$95,$C5,$A6,$70,$ED,$5C,$23,$D0,$68,$B1,$F1,$B0,$F2
	.dc.b	$E1,$DE,$7C,$38,$7F,$29,$F0,$3F,$96,$C7,$C3,$9F,$A3,$18,$BF,$8A
	.dc.b	$D8,$CF,$83,$FE,$E9,$F7,$B9,$36,$D5,$D1,$FE,$6D,$3A,$7F,$DA,$61
	.dc.b	$C3,$FE,$CC,$3D,$67,$FB,$CF,$83,$E2,$3C,$8C,$F1,$A3,$31,$CC,$CB
	.dc.b	$89,$E3,$3B,$9E,$A6,$F3,$FC,$1B,$DC,$CC,$F0,$AE,$51,$D0,$C9,$67
	.dc.b	$8B,$CF,$4F,$6D,$F7,$9C,$DF,$3A,$FC,$77,$0F,$F0,$AB,$F2,$6B,$FC
	.dc.b	$11,$B8,$FF,$01,$F1,$2D,$FE,$E6,$6D,$F1,$AF,$DC,$36,$7F,$88,$56
	.dc.b	$B8,$7F,$53,$5A,$6C,$FF,$15,$BD,$56,$D9,$5B,$72,$D3,$97,$77,$53
	.dc.b	$B7,$93,$63,$24,$BF,$E4,$51,$97,$F9,$24,$DC,$BF,$91,$CD,$AF,$E4
	.dc.b	$1E,$4F,$A4,$F5,$78,$16,$3D,$51,$70,$FE,$5D,$36,$FF,$96,$F7,$7C
	.dc.b	$13,$E1,$77,$DC,$3E,$14,$5B,$FE,$6D,$36,$B8,$D6,$8F,$91,$19,$7F
	.dc.b	$9A,$55,$AF,$E5,$93,$76,$FE,$8B,$F3,$FF,$44,$FA,$5F,$3F,$70,$FA
	.dc.b	$51,$6F,$FA,$34,$DC,$7F,$B9,$78,$4A,$FB,$EA,$6C,$61,$B7,$FE,$AD
	.dc.b	$6D,$FF,$72,$AE,$13,$FE,$B1,$F8,$AE,$67,$17,$04,$9E,$11,$96,$C7
	.dc.b	$FA,$C3,$E7,$FF,$7A,$DE,$7F,$CA,$F3,$FF,$8A,$7C,$2F,$CB,$DA,$F6
	.dc.b	$FF,$AB,$1B,$A7,$EC,$E9,$F7,$3F,$E3,$BC,$7A,$1F,$D4,$E3,$37,$D3
	.dc.b	$E8,$3F,$C4,$D3,$FE,$C2,$6C,$7F,$87,$25,$F4,$0F,$86,$E1,$C5,$30
	.dc.b	$E9,$2F,$98,$C6,$1C,$63,$13,$FD,$8F,$E7,$FF,$C9,$DD,$5F,$B2,$6C
	.dc.b	$66,$9D,$79,$7F,$5B,$B7,$F0,$FA,$DC,$7F,$8C,$1B,$19,$8E,$35,$CE
	.dc.b	$F9,$11,$ED,$76,$7F,$DA,$BF,$40,$F2,$7E,$7B,$6B,$DB,$E7,$E2,$64
	.dc.b	$7C,$0F,$95,$B7,$EF,$7F,$1B,$F2,$52,$7F,$DE,$DF,$25,$6B,$F1,$53
	.dc.b	$74,$FC,$AC,$DA,$FF,$39,$6A,$F8,$E7,$CE,$FC,$76,$CF,$CB,$5E,$C6
	.dc.b	$66,$4E,$5E,$11,$8C,$BE,$45,$72,$6C,$EA,$B8,$1F,$E1,$98,$7C,$FB
	.dc.b	$E0,$7F,$DE,$D1,$FE,$87,$32,$E1,$FF,$45,$46,$E7,$FA,$AC,$62,$74
	.dc.b	$F0,$8F,$F1,$36,$6F,$61,$F2,$FF,$A0,$C3,$95,$FF,$9B,$AB,$53,$FC
	.dc.b	$EB,$E1,$5F,$3E,$F8,$4C,$37,$2F,$E7,$7E,$1F,$2C,$3F,$AC,$5F,$0B
	.dc.b	$D9,$FD,$0B,$E1,$5F,$21,$8C,$3F,$A1,$C6,$1E,$7F,$2D,$FC,$2B,$E5
	.dc.b	$BF,$85,$B6,$32,$78,$23,$2D,$33,$AF,$F0,$B8,$CB,$63,$FA,$8F,$CF
	.dc.b	$FF,$73,$69,$FD,$59,$7C,$B5,$88,$68,$E2,$E2,$7F,$D1,$33,$F7,$F4
	.dc.b	$FF,$4A,$36,$BF,$D8,$1E,$62,$3F,$17,$1E,$2E,$4F,$01,$1F,$2F,$35
	.dc.b	$C3,$F1,$7F,$37,$F9,$53,$F2,$DF,$11,$6B,$F2,$D1,$67,$F2,$33,$72
	.dc.b	$FD,$E7,$42,$B1,$7F,$A2,$85,$62,$FF,$D1,$19,$7F,$13,$AF,$8F,$E1
	.dc.b	$FC,$5F,$17,$E3,$3F,$1F,$E7,$E4,$E8,$F9,$AF,$CB,$F9,$F9,$18,$3E
	.dc.b	$8B,$F3,$FE,$1F,$E8,$7C,$FD,$9F,$AA,$E1,$A7,$D8,$8C,$65,$F4,$27
	.dc.b	$EC,$7C,$76,$37,$FA,$2F,$DB,$F9,$F9,$87,$E0,$38,$2E,$1F,$62,$6D
	.dc.b	$FD,$89,$B5,$F6,$36,$56,$2F,$E2,$D5,$E3,$FB,$9C,$1F,$89,$3F,$15
	.dc.b	$F1,$B7,$1F,$C7,$C5,$BF,$C3,$CD,$CB,$F0,$F3,$6B,$C1,$36,$BE,$9F
	.dc.b	$CF,$FE,$54,$FC,$B7,$91,$6B,$F3,$71,$67,$F2,$73,$70,$FC,$9C,$DA
	.dc.b	$F6,$4D,$AF,$BF,$E3,$FE,$84,$FD,$17,$83,$6B,$F4,$F1,$67,$F3,$F3
	.dc.b	$70,$FC,$FC,$DA,$F9,$13,$6B,$F1,$7F,$3F,$FA,$D3,$F5,$DF,$3F,$6B
	.dc.b	$F6,$71,$67,$F5,$73,$70,$FD,$5C,$DA,$FA,$93,$6B,$FD,$BD,$DD,$FD
	.dc.b	$DA,$7F,$5E,$D8,$CF,$C0,$FF,$11,$F5,$A5,$7F,$10,$7F,$8A,$46,$27
	.dc.b	$FD,$C5,$EB,$FF,$94,$FC,$CF,$F2,$A3,$6F,$FC,$A8,$D9,$DC,$7F,$95
	.dc.b	$E7,$FC,$B5,$E6,$48,$FF,$2D,$79,$76,$BF,$DD,$D3,$75,$FF,$76,$CD
	.dc.b	$AF,$F7,$57,$89,$FE,$63,$3F,$CC,$9E,$25,$8F,$F3,$24,$5C,$3F,$DF
	.dc.b	$13,$6F,$FD,$EF,$36,$BF,$DE,$93,$69,$CC,$F9,$FF,$E4,$51,$FE,$9D
	.dc.b	$F9,$EF,$E2,$F1,$FC,$8A,$FF,$D4,$71,$87,$FD,$29,$F0,$48,$DF,$8F
	.dc.b	$87,$53,$C2,$FB,$DF,$37,$B1,$88,$E0,$EA,$AA,$F9,$FF,$EE,$3E,$BF
	.dc.b	$F5,$E8,$C7,$90,$8C,$4E,$A7,$FB,$1B,$5F,$F5,$CF,$C3,$3E,$7B,$FC
	.dc.b	$01,$1F,$DA,$23,$1E,$57,$56,$C6,$67,$6F,$FB,$0A,$BF,$C6,$DC,$FE
	.dc.b	$0F,$2B,$FD,$93,$3E,$06,$BB,$6B,$FD,$EE,$5E,$3F,$F7,$DE,$0F,$95
	.dc.b	$23,$E0,$46,$78,$38,$36,$3E,$5F,$97,$5D,$9F,$3C,$A7,$CE,$2D,$76
	.dc.b	$E2,$FC,$04,$F1,$C3,$FA,$25,$7F,$84,$DA,$B9,$7F,$DA,$4F,$D3,$F1
	.dc.b	$3F,$C2,$91,$FE,$D5,$B7,$E3,$0F,$A0,$CF,$2F,$BE,$EE,$1E,$DC,$7A
	.dc.b	$E3,$D1,$B5,$FD,$9B,$0E,$55,$E3,$BE,$2F,$84,$C1,$FE,$CA,$8D,$BE
	.dc.b	$79,$F9,$5E,$87,$DF,$F0,$BC,$2D,$9F,$9D,$8F,$F8,$5B,$8B,$FC,$47
	.dc.b	$1F,$ED,$4E,$8F,$F1,$FB,$AC,$39,$CE,$A3,$B0,$BE,$8E,$1B,$7F,$16
	.dc.b	$F9,$CF,$0D,$7F,$61,$C5,$F1,$31,$7E,$81,$F9,$BA,$F0,$08,$CF,$3F
	.dc.b	$CA,$BE,$33,$E7,$BE,$57,$C6,$3C,$B7,$C1,$D1,$FE,$4F,$FC,$BD,$F1
	.dc.b	$1F,$31,$B1,$98,$FA,$FD,$2E,$57,$CB,$EE,$65,$6C,$F3,$D6,$E7,$A7
	.dc.b	$C4,$FE,$DE,$58,$23,$0D,$8F,$DE,$35,$FF,$2D,$B7,$7F,$C8,$B7,$2F
	.dc.b	$F3,$23,$67,$FC,$F5,$93,$E8,$3C,$31,$EF,$C2,$BF,$63,$2F,$64,$FC
	.dc.b	$1F,$07,$DF,$4F,$32,$99,$BF,$91,$35,$FC,$09,$EC,$7E,$3F,$0F,$E9
	.dc.b	$B4,$F3,$C6,$DB,$3F,$1A,$75,$A7,$AB,$C2,$F0,$F9,$52,$BE,$B4,$7D
	.dc.b	$52,$38,$39,$78,$38,$7C,$7B,$A7,$18,$FE,$26,$7F,$D3,$D8,$FE,$4A
	.dc.b	$3F,$77,$DF,$F9,$B7,$5B,$BF,$2E,$F8,$57,$F7,$1A,$FC,$0A,$E1,$A7
	.dc.b	$FA,$55,$B1,$FC,$65,$AF,$F4,$5E,$AF,$C5,$FE,$5D,$FE,$DD,$8F,$F5
	.dc.b	$04,$E8,$93,$AE,$7F,$DF,$C3,$CC,$FC,$9E,$3D,$B2,$F3,$5E,$A6,$61
	.dc.b	$D1,$AF,$D6,$F8,$37,$91,$8D,$7E,$0D,$5D,$7F,$D5,$D5,$AB,$FC,$DF
	.dc.b	$E5,$CF,$92,$FD,$1B,$63,$31,$FC,$53,$F2,$EF,$F4,$DC,$7F,$AE,$DA
	.dc.b	$FE,$19,$B8,$FF,$2A,$6E,$1F,$CF,$F4,$79,$CF,$A6,$FD,$8B,$70,$E9
	.dc.b	$EE,$77,$33,$3F,$C0,$FD,$17,$CA,$52,$57,$AB,$E4,$7C,$DE,$5C,$8F
	.dc.b	$33,$DB,$E8,$E4,$7E,$EB,$24,$62,$3F,$8C,$60,$F9,$5F,$E6,$7F,$A2
	.dc.b	$7C,$79,$FF,$68,$B5,$FF,$6C,$07,$0F,$FB,$6D,$AF,$F5,$F7,$97,$07
	.dc.b	$F4,$E6,$EF,$BE,$5F,$E8,$3E,$13,$E6,$3C,$31,$30,$FF,$75,$5F,$07
	.dc.b	$FB,$55,$6D,$FF,$66,$F9,$9F,$1F,$C7,$91,$82,$1E,$E7,$2F,$C0,$8F
	.dc.b	$9D,$C5,$F3,$FE,$9E,$4F,$FB,$62,$32,$FF,$B6,$A3,$2F,$C8,$8F,$EB
	.dc.b	$3F,$4B,$F3,$23,$FE,$00,$2F,$9E,$C9,$AE,$19,$3A,$8C,$BC,$23,$2F
	.dc.b	$F1,$29,$E4,$EC,$D1,$FD,$28,$7D,$27,$F0,$1D,$FF,$C1,$D1,$8F,$1B
	.dc.b	$FB,$6C,$E6,$3F,$C2,$54,$CB,$35,$27,$A5,$6D,$56,$DF,$F7,$DF,$95
	.dc.b	$F9,$5F,$A7,$91,$82,$1F,$8C,$72,$FE,$3A,$3C,$7F,$07,$FD,$26,$CD
	.dc.b	$24,$FF,$90,$E3,$2F,$92,$38,$BF,$3B,$D8,$AF,$F3,$89,$DB,$67,$EA
	.dc.b	$CF,$F6,$DF,$C9,$BF,$B5,$73,$7F,$87,$57,$CF,$E4,$D7,$95,$30,$FE
	.dc.b	$EF,$E9,$AB,$71,$86,$BF,$E0,$17,$AF,$07,$D2,$DF,$FC,$83,$3F,$D3
	.dc.b	$1B,$8F,$F8,$C1,$B3,$F6,$3E,$95,$E8,$7F,$89,$27,$37,$FE,$49,$6A
	.dc.b	$6C,$F0,$8F,$C2,$47,$E2,$1F,$DB,$9B,$E4,$FE,$F1,$19,$8F,$DD,$B5
	.dc.b	$F4,$47,$84,$8F,$E3,$FD,$BF,$CF,$3E,$B8,$CB,$4D,$72,$FF,$8F,$3E
	.dc.b	$95,$EE,$52,$4F,$E6,$7F,$0A,$7C,$FE,$23,$FB,$8D,$70,$7F,$3C,$D7
	.dc.b	$FB,$AC,$F0,$7F,$45,$F9,$77,$F2,$E6,$F8,$9F,$E4,$C8,$C4,$7F,$7C
	.dc.b	$FC,$23,$D3,$C3,$FE,$60,$9C,$B6,$B8,$1F,$E5,$49,$DF,$0F,$F9,$43
	.dc.b	$F5,$5F,$E4,$35,$FD,$C7,$DE,$46,$27,$58,$FF,$39,$FA,$CF,$F2,$AC
	.dc.b	$E6,$3F,$C9,$8D,$AE,$31,$B3,$F1,$FE,$42,$B1,$1F,$CD,$9F,$E8,$9C
	.dc.b	$18,$C7,$FC,$ED,$5B,$AE,$35,$A7,$FC,$9F,$6F,$91,$7C,$97,$F9,$A1
	.dc.b	$B3,$A6,$BF,$E5,$FF,$D0,$FF,$97,$3F,$E6,$18,$CE,$3F,$05,$BA,$7F
	.dc.b	$9D,$1B,$3F,$E6,$9A,$E3,$5F,$1F,$47,$D3,$FC,$42,$57,$F9,$D6,$EB
	.dc.b	$FE,$6F,$9D,$BE,$A6,$BF,$E6,$FF,$83,$F6,$39,$AE,$37,$DD,$FF,$36
	.dc.b	$4E,$A7,$A4,$6B,$1F,$B8,$EB,$0F,$F4,$DF,$7B,$D2,$FD,$8B,$7D,$AC
	.dc.b	$C7,$ED,$3B,$DF,$9A,$7D,$DA,$D6,$5F,$E7,$E6,$BF,$E8,$C6,$E9,$FE
	.dc.b	$AD,$9F,$C1,$7E,$16,$F8,$AF,$F7,$17,$12,$31,$87,$EE,$A3,$13,$A9
	.dc.b	$CC,$FF,$74,$9C,$BF,$80,$8C,$DF,$E5,$DB,$B7,$FB,$4F,$27,$DD,$FF
	.dc.b	$9B,$6F,$86,$3F,$EA,$CB,$DC,$7F,$A8,$DB,$5C,$3C,$8F,$F5,$CF,$C5
	.dc.b	$FF,$89,$2B,$FD,$E7,$EC,$7F,$A1,$63,$11,$FD,$51,$AF,$FB,$8D,$FF
	.dc.b	$13,$B6,$39,$9F,$EA,$3F,$BA,$FC,$4E,$7C,$63,$33,$FD,$86,$B8,$E0
	.dc.b	$BD,$CF,$33,$55,$C1,$FF,$5B,$46,$95,$7F,$88,$74,$FF,$7B,$7D,$5B
	.dc.b	$E8,$BE,$37,$DD,$BE,$8F,$FD,$6C,$DC,$3C,$5E,$D7,$FB,$AD,$5F,$17
	.dc.b	$DD,$19,$8C,$47,$83,$DD,$FE,$EF,$18,$CB,$FC,$7F,$3E,$2C,$7E,$45
	.dc.b	$F1,$6E,$59,$9F,$95,$87,$29,$F9,$CD,$8F,$0C,$ED,$F7,$D7,$37,$89
	.dc.b	$9B,$F0,$BE,$AB,$EE,$D3,$99,$4F,$51,$C6,$FC,$0F,$6B,$E9,$7F,$17
	.dc.b	$97,$F9,$9E,$1F,$3E,$91,$89,$D7,$85,$E0,$8F,$E7,$32,$3F,$DB,$07
	.dc.b	$08,$FE,$BB,$F1,$79,$9D,$CE,$9B,$19,$9D,$CF,$A4,$79,$FF,$EA,$BD
	.dc.b	$3E,$AC,$77,$AF,$90,$EB,$8A,$BE,$29,$DD,$71,$D7,$AC,$62,$B7,$7C
	.dc.b	$AB,$57,$FE,$6A,$8C,$4E,$A3,$C9,$3A,$7F,$C7,$73,$A7,$D0,$78,$6F
	.dc.b	$F2,$BF,$19,$FE,$C3,$D7,$FD,$89,$3C,$0F,$B1,$C6,$E0,$56,$F2,$FF
	.dc.b	$3D,$FB,$B6,$FF,$C9,$2D,$BF,$D6,$63,$99,$FA,$53,$5F,$F3,$C5,$7F
	.dc.b	$94,$DA,$7E,$BA,$75,$32,$74,$F5,$E3,$FD,$74,$60,$95,$FE,$18,$39
	.dc.b	$9F,$E1,$A9,$F6,$BE,$88,$FE,$49,$F8,$5F,$F9,$BD,$FE,$BE,$D8,$FD
	.dc.b	$FF,$C9,$FB,$7A,$D3,$9D,$7F,$BD,$36,$FF,$1D,$3A,$F3,$BC,$D7,$C6
	.dc.b	$B5,$82,$B5,$7E,$2B,$F2,$CE,$6B,$C4,$FF,$9D,$2B,$6F,$FA,$AB,$8F
	.dc.b	$FA,$77,$B3,$F4,$5B,$87,$F9,$43,$BB,$E5,$E7,$FE,$FF,$9C,$DF,$F4
	.dc.b	$BE,$FF,$F3,$D5,$FF,$4F,$6C,$2B,$FA,$76,$BD,$67,$31,$FD,$CF,$DD
	.dc.b	$F8,$9E,$29,$FC,$07,$8A,$F9,$DC,$C4,$CB,$FB,$93,$A8,$C6,$5F,$DD
	.dc.b	$E2,$E0,$AF,$E9,$D7,$CD,$E2,$4A,$E0,$FF,$9B,$A3,$51,$83,$FC,$21
	.dc.b	$ED,$7D,$9F,$18,$FD,$BF,$21,$F5,$59,$89,$87,$F8,$6D,$D4,$62,$FF
	.dc.b	$DB,$FC,$9F,$9D,$F3,$3E,$DF,$49,$7D,$7E,$66,$63,$FE,$34,$78,$46
	.dc.b	$2F,$FC,$2D,$F5,$BF,$28,$C6,$DF,$E1,$F7,$17,$FE,$19,$72,$FF,$92
	.dc.b	$DF,$D6,$7B,$7F,$D3,$63,$FD,$78,$DA,$FB,$51,$FC,$87,$C0,$7D,$56
	.dc.b	$70,$FD,$FD,$DA,$FF,$60,$B6,$FF,$94,$36,$F8,$F8,$FE,$03,$F7,$9E
	.dc.b	$7F,$F9,$16,$FF,$D9,$8D,$BF,$09,$FC,$9B,$EA,$5E,$CF,$0E,$49,$7C
	.dc.b	$1F,$E1,$58,$AB,$73,$93,$FD,$B1,$F5,$3F,$92,$7F,$DA,$ED,$C1,$F8
	.dc.b	$13,$F0,$27,$D2,$FF,$11,$AD,$51,$A7,$1A,$E1,$3B,$F0,$66,$1F,$C2
	.dc.b	$EB,$F4,$B7,$FE,$E6,$39,$4F,$EE,$1B,$3F,$C4,$7E,$A1,$F1,$DF,$C8
	.dc.b	$5B,$1F,$D4,$F8,$BF,$3D,$5F,$17,$E4,$B5,$72,$BF,$DE,$D7,$C2,$BE
	.dc.b	$5E,$DF,$16,$F5,$1F,$EC,$9F,$1F,$F9,$87,$C2,$3E,$32,$C6,$1B,$8F
	.dc.b	$FB,$45,$B5,$FE,$D1,$7C,$DF,$0B,$27,$FD,$03,$F6,$DF,$E8,$07,$C7
	.dc.b	$19,$AF,$BA,$33,$FF,$68,$CF,$0E,$6D,$6D,$FD,$32,$73,$5F,$EC,$C6
	.dc.b	$BF,$2D,$AF,$F7,$73,$A5,$FF,$24,$EE,$FD,$77,$D6,$7E,$C7,$EB,$FF
	.dc.b	$64,$DC,$BF,$D4,$2D,$BF,$C8,$D7,$09,$FF,$52,$D7,$F5,$3F,$C2,$5F
	.dc.b	$31,$4F,$C8,$FE,$12,$F5,$F1,$7F,$5C,$F2,$57,$07,$18,$FF,$83,$63
	.dc.b	$0F,$D6,$AE,$53,$A8,$4E,$B5,$FC,$0F,$D7,$FF,$A1,$FE,$37,$FB,$53
	.dc.b	$FE,$89,$FA,$AB,$E7,$A6,$C6,$1B,$8F,$8A,$B9,$36,$95,$B8,$C3,$71
	.dc.b	$FD,$AD,$77,$9F,$5E,$CF,$A0,$FF,$14,$46,$1F,$F0,$D5,$7C,$C7,$2D
	.dc.b	$E9,$FF,$00,$46,$33,$FC,$ED,$F3,$9C,$D6,$9D,$F8,$9C,$3D,$1B,$8F
	.dc.b	$76,$D7,$76,$E1,$D3,$1F,$EB,$B5,$F5,$33,$FB,$87,$F7,$7E,$77,$BF
	.dc.b	$89,$FA,$5C,$1F,$11,$B5,$B9,$D6,$5F,$9C,$AD,$3F,$BF,$3E,$CC,$FE
	.dc.b	$FE,$33,$1F,$C0,$DA,$FE,$67,$E9,$7F,$B2,$5F,$E0,$5B,$1F,$27,$E1
	.dc.b	$BF,$C4,$91,$8C,$FF,$C6,$3F,$94,$36,$FF,$14,$37,$0E,$5E,$9E,$27
	.dc.b	$E7,$F8,$32,$7F,$27,$C8,$F8,$AE,$91,$FE,$B7,$E1,$FE,$15,$97,$F8
	.dc.b	$FB,$37,$D9,$F6,$E3,$F1,$B6,$FF,$31,$AF,$CE,$DF,$FC,$D4,$D5,$7C
	.dc.b	$A7,$77,$FE,$43,$CF,$C5,$85,$CF,$58,$7F,$08,$C7,$A5,$6B,$0E,$FF
	.dc.b	$35,$F0,$0F,$F2,$47,$E5,$0F,$AA,$C4,$E8,$98,$7F,$12,$9D,$46,$35
	.dc.b	$F9,$77,$E3,$FC,$2B,$F6,$77,$E1,$8C,$76,$FC,$EF,$93,$98,$7E,$DF
	.dc.b	$F5,$3F,$E4,$17,$E7,$9C,$B6,$75,$E2,$E3,$2F,$B3,$5D,$FE,$F3,$15
	.dc.b	$FE,$91,$D9,$A2,$7F,$CC,$9F,$94,$BE,$CE,$9F,$26,$9F,$25,$8C,$3F
	.dc.b	$85,$56,$DB,$5B,$9D,$69,$FC,$D2,$B4,$FF,$72,$B7,$C6,$7C,$2D,$3F
	.dc.b	$CF,$F5,$A7,$FA,$9C,$EB,$7F,$2F,$89,$33,$FC,$6C,$BE,$87,$19,$77
	.dc.b	$AB,$87,$2A,$D5,$FA,$27,$2F,$C3,$CB,$E1,$39,$3F,$32,$E2,$FD,$5C
	.dc.b	$9F,$18,$FC,$66,$1F,$E3,$8A,$F8,$71,$96,$9F,$A2,$8D,$4F,$28,$CB
	.dc.b	$FB,$BC,$FF,$0E,$DA,$FE,$F5,$18,$8F,$EF,$DF,$94,$FF,$B3,$AB,$F9
	.dc.b	$D3,$63,$E6,$5F,$F6,$F9,$CF,$C2,$FF,$A7,$EB,$53,$FE,$AC,$6C,$FE
	.dc.b	$0E,$38,$57,$F6,$A9,$CA,$F1,$9F,$07,$FE,$1A,$BF,$E9,$E3,$F0,$DF
	.dc.b	$4E,$1F,$CB,$A7,$57,$FE,$30,$AD,$67,$FE,$74,$F5,$BF,$D6,$D1,$83
	.dc.b	$FC,$D0,$E3,$0E,$71,$FE,$9A,$FD,$73,$FA,$13,$FD,$7B,$E2,$51,$8C
	.dc.b	$3F,$B6,$E3,$FE,$36,$7F,$C7,$F3,$AF,$0B,$FD,$51,$5B,$CF,$FC,$3F
	.dc.b	$39,$C3,$FC,$F1,$EE,$EA,$FF,$D8,$DE,$13,$D5,$A7,$FB,$2F,$EB,$DF
	.dc.b	$31,$8C,$3F,$B6,$56,$DB,$5B,$9D,$69,$FE,$16,$AD,$36,$75,$5C,$0E
	.dc.b	$7E,$37,$F8,$D3,$E4,$7F,$63,$5F,$ED,$88,$C7,$67,$FA,$53,$B6,$CE
	.dc.b	$6F,$FE,$00,$4F,$FB,$8A,$B3,$2C,$7F,$93,$D6,$A3,$EF,$70,$FE,$DB
	.dc.b	$1E,$F1,$BB,$E3,$F6,$2F,$9F,$7A,$5F,$ED,$D8,$C3,$63,$35,$FD,$5E
	.dc.b	$72,$FF,$59,$F9,$57,$D2,$3C,$0F,$F7,$34,$61,$B1,$99,$E0,$D7,$FC
	.dc.b	$C1,$5F,$E2,$8E,$DF,$17,$AE,$69,$27,$C5,$CD,$FE,$3B,$D7,$DB,$7F
	.dc.b	$E8,$29,$F8,$41,$C2,$8F,$A3,$DB,$FA,$2F,$A0,$E5,$C7,$FE,$45,$BD
	.dc.b	$D7,$4C,$33,$80,$E8,$E3,$0E,$51,$E0,$A7,$D8,$FF,$7D,$8E,$11,$87
	.dc.b	$E7,$46,$5F,$33,$63,$51,$FE,$62,$6F,$5F,$F9,$F2,$36,$7F,$8B,$3A
	.dc.b	$7F,$94,$3F,$E4,$39,$DF,$2F,$AE,$FF,$CC,$17,$C6,$73,$5F,$C0,$EF
	.dc.b	$8B,$FA,$ED,$DF,$01,$C3,$1F,$81,$D9,$F7,$A3,$14,$8C,$3F,$EE,$7E
	.dc.b	$4C,$C6,$23,$A4,$E2,$FA,$C6,$CF,$F7,$77,$17,$4C,$1F,$63,$F9,$09
	.dc.b	$FE,$20,$DD,$3F,$8A,$57,$8B,$6F,$E3,$6F,$19,$D5,$FF,$93,$9B,$7F
	.dc.b	$A4,$9F,$4A,$FB,$07,$CC,$7D,$36,$C7,$D9,$6B,$FE,$37,$6B,$FE,$B7
	.dc.b	$73,$8F,$D5,$6E,$1F,$E8,$F6,$CF,$E9,$78,$9F,$51,$E2,$97,$AF,$F3
	.dc.b	$3A,$1F,$4E,$F9,$AC,$37,$1F,$F4,$05,$F0,$6D,$FA,$48,$FA,$5A,$FF
	.dc.b	$AB,$23,$83,$63,$C1,$8B,$E7,$BC,$9F,$09,$E3,$9F,$F3,$C5,$7F,$90
	.dc.b	$9A,$F3,$8D,$BF,$1D,$F9,$44,$70,$7A,$1D,$BF,$E0,$9E,$EF,$CE,$6D
	.dc.b	$F2,$7E,$C3,$C1,$F6,$32,$BD,$F1,$F1,$CF,$AC,$47,$EB,$35,$F7,$3B
	.dc.b	$7F,$CB,$35,$FC,$7D,$FF,$54,$76,$FE,$49,$FF,$0D,$7D,$87,$B3,$0F
	.dc.b	$F0,$BB,$78,$BF,$D9,$15,$B9,$FD,$93,$C6,$BF,$DF,$0D,$77,$D1,$FB
	.dc.b	$D8,$C4,$7F,$09,$79,$56,$A3,$C7,$3F,$81,$FB,$1F,$F6,$4C,$79,$BE
	.dc.b	$C7,$F0,$98,$F0,$8F,$E5,$CD,$7F,$C7,$71,$F9,$CC,$0F,$84,$91,$FC
	.dc.b	$4E,$32,$D8,$E0,$47,$FD,$99,$F6,$3F,$EF,$F7,$FD,$03,$F6,$2F,$81
	.dc.b	$FC,$B3,$63,$31,$FA,$CF,$B1,$FF,$52,$C7,$FB,$55,$AF,$FA,$7D,$AF
	.dc.b	$FB,$CC,$38,$7F,$DE,$CD,$7F,$A6,$7B,$0F,$B0,$ED,$F1,$12,$7F,$D2
	.dc.b	$B3,$97,$FD,$9F,$F3,$8F,$A7,$FE,$0C,$D8,$CC,$7A,$BE,$0B,$D1,$EE
	.dc.b	$FF,$33,$CB,$F9,$1D,$79,$5A,$F9,$23,$F9,$E7,$7F,$FB,$EF,$EA,$FF
	.dc.b	$71,$91,$A7,$83,$DD,$FE,$1F,$F6,$3F,$2F,$07,$47,$F3,$BE,$DF,$95
	.dc.b	$18,$E9,$FE,$9C,$FF,$04,$E2,$E3,$F2,$72,$3F,$A6,$61,$CD,$F9,$4D
	.dc.b	$78,$B5,$FE,$B0,$F8,$70,$FE,$94,$DB,$DD,$7F,$24,$C3,$FC,$19,$7F
	.dc.b	$2B,$C9,$F9,$3E,$2C,$8C,$10,$F4,$1F,$E4,$DE,$3F,$E3,$AF,$D2,$8F
	.dc.b	$A3,$3F,$4B,$97,$E9,$FC,$AC,$AF,$AF,$19,$7E,$CC,$65,$EB,$E5,$30
	.dc.b	$FB,$D1,$E0,$3F,$93,$F0,$7F,$3C,$F1,$9F,$3F,$4C,$44,$9F,$EF,$5C
	.dc.b	$1F,$DA,$EB,$71,$8D,$FE,$8C,$62,$3B,$3F,$75,$D1,$FE,$03,$70,$EA
	.dc.b	$FC,$0E,$8F,$29,$7E,$C7,$F8,$93,$D3,$E8,$BD,$07,$F5,$48,$CE,$1F
	.dc.b	$DF,$DB,$87,$E1,$1B,$5F,$E1,$36,$BF,$D2,$34,$FC,$FF,$8F,$FA,$0C
	.dc.b	$24,$FE,$92,$32,$FE,$9A,$32,$FC,$ED,$3F,$A1,$3E,$7F,$84,$F9,$9F
	.dc.b	$11,$23,$04,$3F,$74,$31,$4F,$F8,$75,$B1,$FE,$23,$6B,$C1,$B9,$FB
	.dc.b	$E7,$E1,$1F,$6F,$D7,$FE,$FD,$F3,$1F,$DE,$B8,$4B,$E3,$F2,$6B,$C9
	.dc.b	$99,$7D,$6A,$DC,$EA,$31,$5F,$E0,$47,$FC,$C5,$E3,$7F,$81,$3E,$C5
	.dc.b	$EE,$53,$FC,$0F,$EB,$7F,$9A,$B5,$FF,$19,$37,$93,$F6,$ED,$C3,$FC
	.dc.b	$8C,$D5,$6F,$3F,$9F,$37,$F8,$4B,$DA,$7D,$15,$33,$7F,$89,$C3,$F2
	.dc.b	$4F,$F2,$76,$BE,$0A,$FF,$25,$46,$23,$C3,$1C,$9F,$D3,$47,$E6,$F2
	.dc.b	$FF,$25,$5F,$E6,$1F,$F2,$67,$D8,$BD,$DF,$F1,$2F,$D8,$BC,$EA,$7F
	.dc.b	$8A,$27,$58,$FE,$75,$AF,$EC,$DB,$7D,$2F,$9F,$C7,$7F,$8C,$3E,$C5
	.dc.b	$E4,$FF,$8C,$BE,$4F,$E0,$E3,$FA,$76,$BF,$BD,$9D,$BF,$AB,$6B,$99
	.dc.b	$E8,$73,$F0,$BF,$CA,$3F,$53,$FE,$6D,$F7,$CD,$25,$7F,$9D,$63,$2F
	.dc.b	$02,$1F,$D2,$BE,$23,$FA,$49,$1C,$1F,$D3,$AB,$0F,$EA,$62,$E8,$E6
	.dc.b	$FF,$44,$75,$FF,$19,$BF,$F3,$2F,$D1,$BC,$1F,$13,$71,$FD,$5E,$48
	.dc.b	$C4,$7F,$67,$8C,$4F,$03,$99,$23,$FD,$D9,$F5,$1F,$CE,$9F,$F7,$4E
	.dc.b	$8F,$98,$F8,$0D,$8E,$ED,$7B,$35,$FE,$A7,$7F,$2A,$7E,$69,$23,$F5
	.dc.b	$DF,$11,$E7,$7F,$DF,$95,$CF,$5F,$F7,$AB,$75,$FF,$45,$B6,$7F,$DA
	.dc.b	$F9,$F3,$7C,$F8,$5B,$F9,$51,$A3,$F9,$8F,$85,$FE,$CF,$EF,$FF,$57
	.dc.b	$36,$3F,$89,$B5,$FF,$07,$35,$FE,$DA,$E6,$7A,$B7,$1F,$F5,$F3,$65
	.dc.b	$BB,$EA,$BF,$C3,$17,$E4,$D5,$F5,$EF,$B0,$FF,$4A,$7D,$8B,$CC,$A7
	.dc.b	$FA,$92,$75,$97,$F4,$AB,$E3,$7F,$93,$9C,$CF,$DF,$6B,$F4,$9A,$FF
	.dc.b	$AD,$1D,$E1,$E3,$6E,$99,$AF,$1F,$99,$FE,$A0,$F8,$CF,$F1,$9D,$FC
	.dc.b	$CD,$5F,$98,$FE,$C2,$30,$DC,$3F,$C2,$75,$C1,$BE,$EF,$FB,$66,$3F
	.dc.b	$C6,$D9,$9A,$48,$F0,$12,$3F,$9A,$7D,$87,$E5,$70,$FF,$6E,$57,$59
	.dc.b	$FF,$77,$35,$E9,$5F,$EB,$FA,$E1,$7E,$4F,$67,$FC,$57,$4F,$B2,$FF
	.dc.b	$53,$66,$FB,$27,$D6,$FF,$B1,$DB,$5F,$DD,$73,$61,$CB,$0F,$F5,$74
	.dc.b	$FA,$E3,$AF,$7E,$2B,$FC,$95,$3F,$C6,$BE,$2F,$E2,$FC,$17,$F8,$E2
	.dc.b	$B5,$7C,$27,$CD,$BA,$FC,$D1,$C1,$C6,$3D,$70,$FF,$20,$BC,$AB,$35
	.dc.b	$CA,$BC,$07,$C6,$AF,$04,$73,$8C,$9E,$99,$F0,$61,$CE,$33,$8F,$AA
	.dc.b	$3C,$11,$AA,$ED,$1B,$E8,$FE,$5F,$5F,$1A,$FF,$4B,$3E,$39,$CE,$5F
	.dc.b	$49,$FE,$6B,$87,$BF,$E2,$B6,$FF,$9F,$DA,$FF,$97,$9A,$EC,$47,$25
	.dc.b	$1E,$57,$FC,$CB,$E9,$FF,$04,$BF,$AF,$1C,$5B,$19,$CB,$FC,$2B,$5C
	.dc.b	$6B,$E9,$4E,$5F,$21,$F2,$03,$97,$CC,$7D,$83,$C3,$C1,$F5,$1E,$3A
	.dc.b	$F8,$23,$11,$C3,$0C,$3D,$4F,$59,$DE,$BB,$3F,$06,$BF,$33,$B3,$E3
	.dc.b	$3F,$90,$EB,$FC,$8A,$BC,$75,$A9,$CE,$3F,$EA,$BF,$92,$B5,$FE,$8A
	.dc.b	$6F,$13,$8D,$B6,$AB,$F2,$95,$53,$4E,$33,$C2,$FF,$CB,$93,$A6,$CF
	.dc.b	$96,$3C,$51,$FE,$44,$7A,$B5,$E0,$66,$3C,$B9,$BD,$FF,$05,$7F,$8E
	.dc.b	$63,$0F,$F2,$7F,$B2,$FF,$3E,$B7,$D1,$F2,$FC,$AF,$C0,$ED,$56,$B2
	.dc.b	$FF,$4A,$C6,$2B,$CD,$08,$C5,$FF,$70,$AE,$13,$FD,$81,$AF,$EE,$67
	.dc.b	$6F,$FA,$53,$3F,$2B,$DA,$7C,$1C,$1F,$CB,$67,$2F,$F4,$EF,$84,$FE
	.dc.b	$89,$87,$F5,$A6,$DF,$FA,$7B,$C1,$F8,$B5,$BF,$03,$ED,$F3,$AF,$77
	.dc.b	$FD,$69,$AF,$F5,$86,$BF,$AA,$6E,$7F,$E2,$2F,$82,$F8,$DE,$31,$FC
	.dc.b	$2F,$8E,$FB,$2C,$C4,$D7,$FB,$DB,$A8,$C6,$3F,$E0,$C8,$BA,$7F,$4B
	.dc.b	$BE,$57,$C0,$95,$FE,$0C,$AF,$A3,$8F,$8E,$7F,$C2,$31,$8F,$21,$EF
	.dc.b	$FD,$B6,$CF,$F8,$73,$EC,$B1,$32,$3F,$C4,$F3,$AC,$3F,$C4,$CD,$7F
	.dc.b	$C3,$2D,$7F,$C6,$7F,$01,$F5,$A6,$F2,$7E,$9A,$3F,$CE,$71,$FE,$89
	.dc.b	$75,$3F,$ED,$D9,$D1,$BF,$66,$D7,$F8,$83,$EC,$BE,$69,$8A,$FF,$18
	.dc.b	$B8,$8F,$F1,$7B,$F0,$B8,$0D,$3F,$2B,$3F,$E9,$DF,$17,$F6,$DE,$FF
	.dc.b	$F5,$AB,$FF,$68,$F8,$2F,$B2,$C5,$FF,$B8,$EA,$F1,$FF,$07,$6D,$FF
	.dc.b	$20,$AF,$E1,$8F,$FB,$56,$33,$5F,$C8,$27,$2F,$FB,$B7,$D1,$FF,$29
	.dc.b	$57,$FB,$6D,$B1,$FF,$00,$30,$7D,$93,$DC,$FE,$19,$5A,$CB,$F8,$C3
	.dc.b	$70,$FE,$63,$4F,$9A,$DF,$17,$F1,$8C,$79,$C7,$F1,$5F,$B0,$FD,$2E
	.dc.b	$3F,$C8,$5B,$87,$F6,$BF,$B1,$7D,$93,$E8,$78,$77,$CA,$FE,$25,$7F
	.dc.b	$A4,$DC,$D2,$4F,$F0,$9F,$C1,$1F,$4D,$C6,$BF,$84,$C6,$2F,$FD,$F9
	.dc.b	$F6,$6F,$91,$FE,$54,$D8,$FA,$BC,$2F,$99,$CD,$49,$CF,$CB,$DC,$7E
	.dc.b	$55,$A3,$F9,$BF,$C6,$3C,$AB,$E6,$61,$B8,$7F,$B6,$9B,$83,$8F,$FC
	.dc.b	$79,$7C,$6B,$FD,$BF,$7E,$6F,$8A,$91,$FE,$51,$FA,$FF,$F4,$93,$EC
	.dc.b	$8C,$E1,$FE,$FC,$DB,$FD,$B1,$3F,$CE,$DB,$19,$9F,$CB,$C6,$31,$FE
	.dc.b	$AF,$F7,$7F,$06,$FA,$1F,$E2,$18,$CE,$FF,$AF,$FB,$4F,$D8,$7D,$A5
	.dc.b	$AF,$D8,$37,$2F,$64,$7E,$E2,$31,$7F,$E5,$9F,$15,$1F,$D6,$FE,$C5
	.dc.b	$F4,$54,$FC,$67,$DA,$7F,$B4,$30,$F1,$1F,$E3,$3C,$55,$C2,$33,$1F
	.dc.b	$E3,$78,$C3,$F9,$F9,$F8,$EF,$90,$D5,$76,$F0,$3F,$B8,$FD,$A7,$F3
	.dc.b	$7F,$95,$BE,$8B,$ED,$2C,$61,$B9,$F8,$2B,$F9,$F4,$63,$1E,$D1,$89
	.dc.b	$D5,$AF,$94,$61,$B5,$E5,$F6,$9F,$55,$FE,$51,$AF,$6C,$63,$C5,$FF
	.dc.b	$02,$CE,$6B,$83,$FE,$22,$AD,$B9,$E1,$FF,$44,$39,$9D,$9C,$AB,$9F
	.dc.b	$93,$D3,$DA,$B8,$72,$6D,$76,$6E,$1D,$1F,$B5,$AC,$AF,$CB,$7C,$1D
	.dc.b	$BF,$A0,$7D,$89,$FE,$4F,$C5,$F3,$6B,$FC,$20,$D9,$FB,$C7,$DB,$9D
	.dc.b	$9F,$84,$F8,$9E,$98,$FF,$B3,$B3,$71,$A7,$FC,$4D,$F2,$FF,$D9,$F2
	.dc.b	$FC,$2E,$28,$C3,$F2,$3E,$2F,$7E,$B7,$F8,$B7,$EC,$DC,$1F,$CF,$1B
	.dc.b	$87,$F8,$95,$B7,$E6,$0E,$1F,$CB,$71,$7D,$AF,$17,$EB,$9F,$E3,$EE
	.dc.b	$57,$C8,$F8,$B8,$3E,$AC,$FA,$3E,$D3,$C5,$CB,$73,$F4,$E7,$FE,$0E
	.dc.b	$7F,$44,$FF,$8D,$33,$F8,$ED,$73,$3E,$93,$FC,$53,$97,$82,$FC,$59
	.dc.b	$1E,$F8,$7E,$AA,$74,$F8,$3E,$8B,$E2,$1F,$E5,$1F,$B4,$7D,$86,$27
	.dc.b	$47,$C5,$9F,$90,$47,$07,$5F,$B3,$FD,$E6,$7F,$84,$D1,$5F,$16,$73
	.dc.b	$B7,$F9,$A6,$5F,$2F,$F5,$4F,$F3,$17,$DA,$7E,$F2,$75,$19,$C7,$EB
	.dc.b	$3F,$64,$8E,$0F,$17,$9F,$BC,$BF,$82,$6A,$CB,$FA,$44,$7E,$63,$ED
	.dc.b	$3F,$B0,$5F,$E2,$9B,$1C,$2B,$F2,$13,$9F,$5B,$FC,$ED,$5A,$AF,$CD
	.dc.b	$B5,$FE,$34,$D7,$FD,$9C,$D4,$F6,$C9,$5F,$ED,$3B,$FF,$70,$FC,$67
	.dc.b	$6C,$3F,$19,$F1,$DF,$22,$BA,$F8,$5E,$23,$F2,$16,$FA,$6C,$4F,$AE
	.dc.b	$7F,$C3,$39,$7F,$15,$8F,$F4,$13,$73,$FC,$F3,$5F,$5C,$F8,$35,$E7
	.dc.b	$1F,$2B,$1F,$93,$A7,$FA,$86,$FF,$0F,$87,$C9,$E2,$FD,$14,$BE,$0F
	.dc.b	$E3,$C4,$9F,$0B,$E4,$23,$83,$F2,$DE,$63,$E7,$78,$65,$F6,$A3,$13
	.dc.b	$FC,$03,$ED,$3F,$DA,$AD,$F8,$9F,$0F,$A7,$FE,$AB,$9C,$DF,$F1,$66
	.dc.b	$BF,$C4,$9A,$FF,$51,$6A,$0E,$3B,$9F,$11,$FE,$1D,$C3,$FA,$A8,$FB
	.dc.b	$4F,$87,$7F,$CF,$E7,$4F,$F5,$59,$D6,$9F,$E7,$EF,$77,$FD,$8B,$18
	.dc.b	$3F,$A1,$D7,$0D,$FF,$0B,$7F,$86,$C3,$AC,$63,$E2,$5F,$69,$FD,$33
	.dc.b	$0F,$EC,$F1,$8F,$3B,$FD,$17,$F0,$7F,$EB,$3A,$D6,$3F,$EB,$26,$E1
	.dc.b	$CD,$B8,$7C,$1D,$7D,$F7,$C7,$0F,$F6,$B7,$89,$FE,$1B,$C7,$FA,$0B
	.dc.b	$71,$F3,$DB,$EC,$9F,$3D,$C3,$DD,$99,$FF,$38,$C3,$83,$DE,$B6,$E6
	.dc.b	$77,$8F,$F3,$3A,$F0,$3F,$EE,$D8,$FF,$3A,$C6,$23,$FD,$DD,$E6,$7C
	.dc.b	$8F,$95,$FE,$91,$1F,$EE,$C8,$C7,$9D,$ED,$E2,$FD,$FD,$6A,$BF,$DD
	.dc.b	$F5,$A7,$86,$3F,$83,$9D,$61,$FD,$3E,$31,$87,$3F,$07,$85,$F2,$9C
	.dc.b	$BF,$8B,$FB,$37,$D2,$3D,$2F,$F7,$A4,$61,$B1,$99,$FE,$A5,$39,$7F
	.dc.b	$AA,$7C,$C3,$EA,$7F,$DA,$51,$86,$BF,$DB,$A3,$5E,$37,$FC,$01,$6C
	.dc.b	$74,$F6,$9D,$FE,$26,$C6,$63,$07,$9B,$A3,$FC,$6D,$A7,$F7,$AC,$BF
	.dc.b	$D1,$D5,$FE,$17,$3D,$D3,$A7,$FC,$79,$1A,$7F,$36,$68,$F8,$A1,$D7
	.dc.b	$07,$D3,$6E,$FF,$5F,$E8,$FF,$C8,$D3,$9F,$B4,$FF,$24,$74,$FF,$B7
	.dc.b	$FC,$8C,$DF,$DD,$6E,$3F,$46,$B7,$9F,$91,$FE,$FD,$19,$3E,$9B,$87
	.dc.b	$C8,$DC,$8F,$28,$E0,$7F,$88,$A3,$2F,$F8,$F3,$A3,$F9,$A3,$FB,$4E
	.dc.b	$1F,$F4,$93,$7E,$03,$97,$77,$DB,$7A,$76,$7F,$0E,$E5,$CD,$F0,$BF
	.dc.b	$F2,$F5,$ED,$E4,$7F,$08,$9E,$3D,$5F,$E7,$08,$FE,$2B,$F6,$9F,$8B
	.dc.b	$7F,$89,$37,$83,$F6,$F3,$B9,$FE,$31,$CB,$ED,$DD,$7B,$7F,$CB,$0D
	.dc.b	$96,$7F,$C9,$2B,$E2,$43,$ED,$1F,$39,$F1,$FA,$5F,$56,$ED,$C3,$6F
	.dc.b	$FC,$77,$7C,$1B,$7F,$4E,$31,$1F,$E3,$56,$BF,$E4,$C6,$BF,$E8,$B5
	.dc.b	$ED,$76,$F9,$D9,$3D,$A3,$CF,$3A,$8F,$F1,$35,$AB,$71,$FE,$5E,$9D
	.dc.b	$36,$B8,$91,$FF,$17,$74,$BE,$87,$AF,$C4,$7C,$1A,$FF,$40,$D7,$F8
	.dc.b	$CD,$AF,$58,$DB,$9F,$17,$81,$87,$47,$C6,$8C,$71,$FF,$46,$DB,$FD
	.dc.b	$05,$3E,$7F,$B3,$F0,$7D,$A4,$8F,$9B,$1F,$00,$C3,$FE,$26,$C5,$39
	.dc.b	$8D,$90,$E8,$7F,$92,$FC,$3F,$E3,$6F,$F9,$CB,$BF,$F2,$D8,$F8,$FE
	.dc.b	$CF,$DA,$DD,$3F,$0B,$3A,$E7,$FF,$60,$36,$DA,$30,$E3,$5F,$E2,$56
	.dc.b	$CF,$F7,$5A,$E0,$FF,$A1,$7B,$3A,$91,$FF,$30,$7D,$A7,$E7,$72,$FE
	.dc.b	$23,$5F,$EF,$D9,$FF,$32,$5F,$87,$1F,$06,$1E,$6F,$8F,$95,$A8,$FC
	.dc.b	$D5,$78,$2F,$94,$FF,$A8,$9B,$87,$6B,$E3,$3F,$EA,$6F,$49,$E8,$FE
	.dc.b	$8F,$ED,$1F,$05,$49,$5F,$E9,$CF,$B4,$7C,$4F,$EB,$1B,$1F,$E7,$BA
	.dc.b	$E2,$DA,$D5,$7E,$BF,$DC,$7D,$96,$2B,$A4,$E6,$3F,$87,$FC,$73,$E9
	.dc.b	$E9,$2B,$B7,$C2,$78,$7E,$03,$D0,$C8,$F3,$47,$7F,$AD,$91,$FE,$8F
	.dc.b	$C9,$1D,$BE,$03,$F8,$2F,$DA,$7C,$2A,$FF,$5C,$F6,$FA,$9F,$F6,$5B
	.dc.b	$79,$3F,$A3,$CE,$E7,$FD,$9E,$D8,$FE,$97,$3B,$EA,$FF,$1C,$57,$F0
	.dc.b	$0E,$0E,$FC,$72,$3F,$9F,$E1,$CD,$FF,$69,$B5,$E2,$D7,$F8,$EE,$FF
	.dc.b	$DB,$DF,$F7,$BA,$FB,$4F,$92,$7C,$AF,$3A,$4E,$A3,$2F,$08,$F9,$FC
	.dc.b	$5F,$43,$E6,$64,$F4,$8C,$BD,$A3,$2B,$5F,$F1,$6E,$1F,$72,$3F,$DC
	.dc.b	$3E,$3F,$CE,$6F,$D1,$FF,$32,$83,$78,$64,$E8,$87,$FB,$EA,$79,$36
	.dc.b	$39,$4C,$8E,$F1,$FC,$FB,$ED,$3F,$80,$C7,$1F,$A5,$F0,$19,$F4,$64
	.dc.b	$7F,$6E,$9D,$BF,$58,$38,$7E,$CB,$5F,$A3,$E4,$7E,$5A,$A4,$7D,$28
	.dc.b	$FC,$C9,$87,$E3,$71,$FF,$96,$73,$56,$A7,$FC,$F9,$A7,$F8,$4B,$07
	.dc.b	$AD,$F1,$FE,$56,$56,$B1,$F1,$CF,$BC,$CC,$7C,$79,$C3,$B3,$FD,$11
	.dc.b	$5D,$1F,$81,$3F,$E0,$B8,$C4,$7C,$13,$E0,$C7,$F8,$4B,$D1,$7C,$D7
	.dc.b	$F6,$D6,$FC,$8F,$F8,$12,$0D,$F2,$A5,$F8,$6B,$6D,$AD,$B6,$7F,$BA
	.dc.b	$EF,$FE,$5F,$D3,$FC,$67,$F6,$8F,$80,$FE,$F2,$DB,$F6,$1F,$E5,$8F
	.dc.b	$1F,$89,$FE,$25,$BF,$F3,$44,$7D,$B6,$BF,$B2,$6B,$FE,$AB,$7F,$26
	.dc.b	$F4,$F7,$5F,$4D,$FE,$08,$6C,$78,$27,$6F,$DC,$39,$73,$6E,$A4,$6E
	.dc.b	$3D,$A7,$94,$8F,$F8,$FA,$79,$BF,$91,$FB,$37,$BF,$FE,$18,$FB,$4F
	.dc.b	$F6,$84,$FC,$06,$BE,$11,$3C,$1F,$8A,$D7,$EE,$35,$ED,$3F,$E4,$B8
	.dc.b	$C4,$7F,$16,$FB,$47,$8B,$FE,$28,$F9,$6F,$83,$3D,$DB,$C1,$FE,$68
	.dc.b	$AD,$CF,$D4,$6B,$F8,$B6,$BF,$AD,$6B,$DE,$BF,$CB,$FE,$FB,$EA,$7F
	.dc.b	$65,$7F,$EB,$8F,$DA,$FF,$0F,$7F,$3F,$99,$18,$9F,$F6,$85,$70,$EB
	.dc.b	$FF,$73,$61,$CF,$0E,$B3,$FE,$69,$8C,$47,$E7,$DA,$FF,$01,$C1,$DB
	.dc.b	$FE,$C6,$8F,$E5,$BF,$13,$FD,$50,$FE,$AB,$E0,$58,$C9,$0F,$34,$FF
	.dc.b	$18,$CB,$FD,$C9,$F4,$EE,$FF,$F3,$5F,$E7,$FF,$E1,$CB,$FD,$33,$5F
	.dc.b	$F3,$5C,$ED,$FF,$74,$B5,$FE,$2D,$9A,$31,$1F,$D3,$A3,$15,$F3,$27
	.dc.b	$33,$F1,$34,$7C,$DF,$F9,$D3,$E7,$64,$FF,$BD,$A7,$2D,$9F,$92,$47
	.dc.b	$E4,$46,$9F,$1B,$72,$FA,$73,$FE,$8A,$8C,$6F,$FE,$DD,$C3,$13,$2F
	.dc.b	$B1,$FC,$1F,$E1,$BD,$8F,$F1,$9F,$8A,$FF,$18,$F6,$FF,$A4,$BE,$B7
	.dc.b	$F4,$39,$FF,$BF,$1B,$92,$FF,$D9,$3E,$E7,$F0,$B8,$C6,$1C,$A7,$F8
	.dc.b	$F3,$5F,$E8,$0D,$7F,$BB,$B5,$FC,$5F,$DA,$3E,$D3,$FD,$47,$F6,$AF
	.dc.b	$36,$9F,$EA,$89,$D5,$7F,$31,$E3,$46,$23,$F3,$55,$C5,$FE,$4C,$D7
	.dc.b	$FD,$57,$1B,$D1,$97,$8B,$1F,$D3,$CF,$3A,$FC,$03,$5D,$1F,$EE,$DF
	.dc.b	$67,$FD,$73,$F6,$BF,$B8,$BF,$8D,$D4,$8C,$57,$C9,$F9,$0F,$F1,$DB
	.dc.b	$FC,$35,$B9,$78,$A3,$FD,$5B,$3C,$2B,$97,$8B,$FE,$AE,$BD,$E1,$FE
	.dc.b	$20,$6D,$F5,$C7,$47,$FB,$33,$EC,$BF,$D5,$D1,$DE,$32,$FF,$3E,$9E
	.dc.b	$38,$2B,$79,$7F,$BE,$25,$EC,$FF,$84,$ED,$F6,$BF,$EA,$77,$FD,$87
	.dc.b	$F6,$AF,$B2,$FF,$6D,$B6,$3B,$57,$1A,$E1,$0B,$EB,$87,$E3,$67,$A6
	.dc.b	$4C,$39,$5F,$F1,$56,$D7,$97,$0E,$9F,$1B,$F0,$DB,$5F,$3F,$E0,$3F
	.dc.b	$C8,$B8,$F8,$32,$FF,$30,$4F,$7B,$E1,$85,$8F,$AD,$87,$27,$CD,$87
	.dc.b	$2A,$FA,$0D,$8D,$D7,$17,$D9,$5C,$6F,$BC,$7C,$8B,$E2,$EE,$3E,$0B
	.dc.b	$93,$98,$E6,$FE,$7F,$1A,$6D,$7D,$36,$DF,$EC,$BC,$5F,$F0,$77,$C6
	.dc.b	$5C,$3F,$C6,$0D,$7F,$CC,$ED,$7F,$78,$D4,$FA,$D7,$A3,$F9,$5D,$3E
	.dc.b	$CC,$70,$3D,$07,$F0,$58,$D6,$7E,$28,$C9,$CB,$8B,$A3,$D8,$CC,$F0
	.dc.b	$7C,$E6,$23,$77,$E3,$9D,$4F,$A8,$F9,$1B,$FE,$1F,$E3,$BF,$55,$9F
	.dc.b	$EA,$E7,$59,$7E,$33,$81,$5A,$CB,$CD,$39,$8C,$57,$8A,$31,$3F,$E2
	.dc.b	$48,$7C,$7E,$9F,$E4,$8F,$0C,$7F,$55,$A6,$67,$75,$C6,$B0,$6F,$8D
	.dc.b	$58,$9D,$1F,$E3,$DC,$BF,$D9,$2E,$2F,$F4,$5E,$FE,$DB,$19,$D3,$F8
	.dc.b	$56,$5F,$A9,$AF,$5F,$5B,$ED,$7F,$D0,$B7,$FD,$D5,$B8,$7F,$60,$6B
	.dc.b	$FD,$7D,$AF,$F8,$01,$DE,$1D,$B3,$F2,$CF,$F8,$DE,$FB,$7A,$5E,$DC
	.dc.b	$7C,$38,$70,$C7,$F9,$2E,$1C,$31,$FE,$41,$87,$09,$FF,$2E,$4E,$5F
	.dc.b	$E9,$BD,$FF,$D0,$DF,$E9,$AD,$CB,$FA,$6B,$5F,$E9,$AD,$78,$F7,$FD
	.dc.b	$5F,$01,$F6,$FF,$7B,$9D,$7C,$F3,$FE,$C1,$3F,$23,$C6,$7C,$1D,$3E
	.dc.b	$F7,$8A,$FB,$6C,$6B,$FD,$F6,$6E,$BF,$E2,$BA,$D4,$E4,$91,$FD,$F2
	.dc.b	$7F,$2C,$FF,$82,$5A,$B7,$FF,$1C,$46,$1B,$3F,$DE,$FE,$5F,$BE,$7F
	.dc.b	$80,$C1,$1F,$80,$F2,$7F,$D4,$75,$FE,$33,$6E,$3F,$E1,$E6,$BF,$E1
	.dc.b	$D6,$BF,$E3,$E7,$77,$F8,$CF,$96,$79,$3D,$7D,$FF,$D0,$46,$30,$FF
	.dc.b	$73,$F3,$FE,$8F,$CF,$FC,$06,$5F,$EE,$CF,$1F,$13,$2B,$67,$FB,$BA
	.dc.b	$31,$3F,$8B,$E1,$7D,$B7,$C1,$6D,$7F,$89,$A3,$04,$7F,$D4,$9F,$A6
	.dc.b	$EA,$B9,$4F,$ED,$63,$1A,$FF,$05,$F7,$FF,$AC,$5F,$FA,$D6,$FF,$D5
	.dc.b	$FE,$4F,$D2,$6C,$66,$77,$1F,$EB,$B6,$D7,$F7,$9F,$11,$F6,$FF,$C3
	.dc.b	$AF,$FD,$57,$5A,$7F,$93,$46,$6F,$FD,$1D,$E9,$7F,$93,$E7,$C4,$D8
	.dc.b	$FF,$64,$35,$FF,$75,$35,$FF,$1A,$1F,$E6,$1E,$3F,$E3,$38,$3E,$CD
	.dc.b	$F3,$9F,$C5,$EF,$A6,$7E,$79,$CE,$3F,$EC,$0F,$B5,$FD,$46,$5F,$E3
	.dc.b	$F6,$A6,$93,$C0,$FE,$1B,$F6,$AF,$A8,$FE,$49,$3F,$AB,$BF,$F4,$3C
	.dc.b	$E5,$ED,$5C,$2B,$6D,$8F,$F6,$8F,$DA,$3E,$3F,$E9,$36,$3E,$B3,$71
	.dc.b	$FF,$76,$35,$F8,$61,$D6,$7E,$FF,$95,$FE,$71,$18,$AF,$EB,$1A,$27
	.dc.b	$F9,$D7,$C4,$3E,$86,$F9,$58,$6D,$FF,$BB,$DB,$97,$FA,$A6,$B8,$4F
	.dc.b	$FB,$CE,$BB,$FC,$EF,$A8,$91,$EA,$6D,$7F,$B7,$F3,$7D,$3B,$E6,$B8
	.dc.b	$DF,$F9,$C6,$7F,$AE,$CE,$63,$E3,$DF,$F5,$89,$E3,$5C,$2F,$A5,$ED
	.dc.b	$FF,$54,$FD,$A7,$ED,$7E,$D3,$F6,$CD,$C7,$B1,$FE,$5A,$F0,$51,$9A
	.dc.b	$E3,$3A,$5E,$87,$FA,$F2,$FF,$B2,$7D,$9B,$E8,$FF,$54,$DB,$FD,$73
	.dc.b	$6F,$EB,$87,$0F,$DA,$75,$A7,$EB,$7E,$CF,$FD,$7D,$5E,$1A,$E1,$7F
	.dc.b	$A0,$9C,$C2,$BF,$CD,$9F,$69,$FC,$FB,$E4,$FF,$BE,$BF,$D0,$3E,$1A
	.dc.b	$FA,$49,$B1,$86,$E5,$EF,$AE,$4D,$A5,$6E,$30,$DC,$3C,$51,$FE,$A9
	.dc.b	$AF,$03,$F2,$75,$7D,$17,$FA,$B6,$30,$FF,$89,$EB,$FD,$E8,$E5,$BE
	.dc.b	$0F,$F8,$3E,$31,$9F,$F8,$01,$E0,$D9,$D4,$EF,$C4,$E1,$ED,$5C,$3B
	.dc.b	$B6,$BB,$B6,$FA,$3F,$73,$3F,$B6,$4A,$FB,$6D,$AF,$79,$FE,$48,$DB
	.dc.b	$FA,$23,$83,$EF,$47,$D7,$F3,$BA,$57,$E2,$F2,$5F,$0C,$7F,$C7,$5C
	.dc.b	$3F,$3A,$B5,$39,$7F,$90,$FC,$A7,$F7,$67,$F0,$8D,$CB,$E4,$FC,$47
	.dc.b	$F8,$DE,$31,$87,$F9,$17,$EC,$9B,$7F,$90,$38,$9F,$1F,$AC,$BF,$96
	.dc.b	$D6,$9F,$D0,$4E,$BD,$6F,$A5,$E1,$C9,$CD,$F1,$D3,$F4,$BC,$7F,$C3
	.dc.b	$DF,$F7,$56,$8F,$B4,$FF,$25,$67,$FE,$46,$9F,$8D,$39,$C3,$F3,$D3
	.dc.b	$9D,$7D,$1C,$9E,$7C,$B5,$87,$CD,$AD,$5F,$FB,$DF,$5F,$5E,$5B,$24
	.dc.b	$74,$C2,$DF,$1F,$9F,$F1,$9F,$E5,$DF,$B4,$7D,$76,$2F,$C7,$3F,$1F
	.dc.b	$EC,$9F,$0F,$9C,$3F,$03,$9A,$73,$7C,$0F,$92,$E2,$7C,$11,$8A,$F8
	.dc.b	$7F,$67,$23,$F7,$DF,$69,$FB,$CE,$0F,$B1,$39,$BE,$F8,$7F,$4D,$7E
	.dc.b	$14,$79,$7C,$DD,$57,$FB,$B7,$2F,$24,$63,$1F,$8D,$87,$1A,$FF,$44
	.dc.b	$7D,$A7,$F6,$4C,$7F,$30,$D8,$E5,$87,$E4,$64,$27,$57,$FD,$D1,$B1
	.dc.b	$FE,$4E,$F8,$BF,$81,$7D,$BD,$3F,$C8,$46,$2F,$84,$63,$5F,$F4,$76
	.dc.b	$1C,$B1,$FF,$79,$C6,$3C,$29,$7F,$94,$97,$D2,$E3,$3F,$4D,$76,$D7
	.dc.b	$1A,$7C,$3D,$3E,$06,$3E,$7A,$D3,$FC,$B7,$3F,$E4,$8F,$E8,$27,$78
	.dc.b	$7C,$09,$FC,$2C,$EF,$87,$F5,$58,$3E,$D3,$D3,$12,$7F,$53,$E4,$BE
	.dc.b	$A7,$18,$FE,$22,$31,$3C,$CD,$CE,$B0,$FB,$53,$AA,$FE,$19,$F6,$9F
	.dc.b	$EE,$7B,$FF,$1D,$36,$3F,$A6,$B5,$FF,$02,$B5,$FF,$14,$BB,$AE,$BF
	.dc.b	$0D,$FE,$BD,$8C,$57,$EA,$EF,$FD,$12,$FF,$39,$F8,$47,$DA,$7F,$B2
	.dc.b	$21,$F6,$9F,$13,$2F,$F4,$14,$E6,$31,$97,$F9,$DE,$31,$A7,$FA,$31
	.dc.b	$B7,$FD,$B7,$E1,$DF,$69,$FD,$1A,$30,$DC,$7F,$D1,$DF,$47,$FE,$BE
	.dc.b	$8C,$63,$FE,$AC,$6D,$7E,$5A,$F9,$F8,$7F,$E9,$AB,$E0,$FF,$B6,$FB
	.dc.b	$FF,$C3,$38,$FF,$37,$6E,$3F,$E8,$1F,$9D,$D7,$B1,$32,$FF,$34,$57
	.dc.b	$0B,$FF,$18,$B5,$FF,$2A,$B5,$FF,$73,$CF,$FA,$AB,$EE,$D5,$FA,$AF
	.dc.b	$95,$FF,$46,$D6,$A3,$E7,$4E,$A1,$3A,$7F,$DE,$7E,$3E,$63,$18,$23
	.dc.b	$1E,$7F,$ED,$BD,$17,$27,$EF,$6F,$75,$AC,$3F,$DF,$6F,$7C,$BD,$6D
	.dc.b	$8F,$5B,$5D,$C7,$F7,$57,$F8,$1F,$DA,$BE,$95,$E9,$7F,$A3,$63,$0D
	.dc.b	$8C,$CF,$F8,$8A,$72,$FF,$89,$3E,$61,$F5,$7F,$ED,$68,$C3,$5F,$F2
	.dc.b	$FC,$6B,$C4,$FF,$4D,$36,$3A,$7A,$EF,$92,$F1,$36,$33,$18,$3F,$C6
	.dc.b	$7B,$FC,$6E,$6F,$F4,$6E,$FF,$52,$31,$5F,$5F,$87,$E6,$70,$FF,$86
	.dc.b	$A3,$07,$F9,$11,$7D,$DF,$29,$DE,$16,$B4,$DB,$EB,$1F,$7A,$7A,$D6
	.dc.b	$7F,$1C,$74,$DF,$2A,$5F,$83,$05,$6A,$30,$D7,$A1,$FE,$58,$7F,$D4
	.dc.b	$AE,$A7,$B5,$48,$C1,$0E,$35,$DA,$3D,$19,$3E,$F3,$FD,$67,$D7,$FE
	.dc.b	$F3,$6F,$37,$C6,$6E,$1F,$17,$ED,$6E,$3F,$94,$9B,$87,$3A,$FE,$1D
	.dc.b	$18,$AC,$FC,$04,$9F,$E2,$4F,$0F,$84,$C4,$7F,$8C,$B0,$FC,$FC,$E3
	.dc.b	$2F,$D5,$47,$E8,$AB,$C3,$C0,$F6,$29,$30,$FF,$29,$FA,$75,$FF,$29
	.dc.b	$C5,$C3,$FC,$A1,$3E,$AF,$CA,$7F,$D5,$55,$FE,$5D,$6B,$E3,$6C,$F8
	.dc.b	$9F,$F0,$6C,$6F,$0E,$44,$79,$78,$B5,$E5,$16,$B9,$4F,$AB,$D3,$95
	.dc.b	$FE,$6A,$7D,$BF,$21,$88,$FF,$35,$C5,$AF,$F3,$44,$DE,$1F,$E0,$DA
	.dc.b	$2B,$17,$FE,$3A,$AF,$AF,$F9,$4B,$E3,$69,$27,$FC,$F9,$1F,$C5,$71
	.dc.b	$F9,$EE,$E3,$33,$B7,$EB,$46,$F1,$FE,$2C,$EE,$33,$3B,$AE,$7D,$1F
	.dc.b	$63,$CE,$95,$FE,$88,$7E,$DF,$B3,$88,$FF,$44,$C5,$AF,$F4,$34,$DC
	.dc.b	$7D,$F3,$6B,$FC,$AF,$3E,$EF,$56,$47,$B1,$F8,$3F,$41,$88,$F4,$C5
	.dc.b	$C3,$DF,$37,$0F,$5C,$DA,$FF,$63,$E6,$AF,$C8,$FA,$5F,$92,$F4,$A5
	.dc.b	$7F,$A8,$A3,$2D,$21,$FE,$A4,$F2,$FF,$32,$47,$FD,$52,$FE,$73,$E2
	.dc.b	$3E,$BF,$D8,$FE,$7B,$E2,$65,$7D,$68,$FB,$7F,$65,$23,$EC,$BF,$7B
	.dc.b	$E3,$BE,$FF,$D8,$C9,$FB,$AF,$EA,$3E,$E3,$F7,$71,$FA,$9F,$95,$BD
	.dc.b	$C6,$98,$FA,$4B,$3F,$ED,$99,$D3,$67,$4D,$21,$FE,$A9,$8D,$B6,$7F
	.dc.b	$73,$E9,$7F,$0A,$FD,$C7,$FB,$34,$FF,$01,$6F,$37,$FB,$16,$77,$3F
	.dc.b	$ED,$19,$CF,$37,$F6,$56,$C7,$FA,$62,$7D,$BF,$6A,$E1,$FF,$7B,$CF
	.dc.b	$F9,$82,$30,$E1,$AF,$FB,$F1,$AF,$F9,$98,$E1,$1D,$08,$FC,$3F,$57
	.dc.b	$D5,$F5,$5E,$BF,$B1,$93,$E7,$8C,$BE,$98,$CB,$EF,$8C,$BF,$06,$33
	.dc.b	$9F,$F2,$0C,$95,$FC,$CB,$DA,$7E,$EB,$FF,$00,$70,$FF,$18,$35,$FF
	.dc.b	$6D,$C6,$BE,$0F,$E7,$7B,$3C,$6B,$58,$FF,$85,$6B,$57,$CA,$F6,$FC
	.dc.b	$FF,$57,$EB,$7C,$D4,$8C,$10,$FB,$0E,$5F,$B5,$1F,$8E,$FB,$67,$C2
	.dc.b	$D2,$4F,$FA,$4B,$1F,$C0,$CF,$F8,$22,$31,$C7,$F4,$63,$8C,$60,$FF
	.dc.b	$4C,$B8,$7F,$DE,$CD,$9E,$7E,$DF,$69,$CD,$D9,$CC,$73,$1F,$6D,$F2
	.dc.b	$5F,$C0,$18,$CF,$F1,$2D,$3F,$17,$1B,$7F,$0C,$FC,$CC,$6C,$FE,$BA
	.dc.b	$31,$3E,$8F,$6B,$1E,$9C,$8F,$14,$A3,$F6,$BF,$85,$7F,$7E,$96,$3F
	.dc.b	$A6,$AD,$47,$0A,$E4,$EB,$1F,$CC,$5F,$1B,$FD,$6F,$DB,$FF,$45,$AF
	.dc.b	$80,$D7,$81,$FE,$C4,$71,$1C,$3C,$DF,$E0,$F1,$8B,$FD,$23,$6B,$87
	.dc.b	$9D,$AA,$DB,$23,$BB,$97,$E8,$7E,$3C,$F5,$3F,$89,$B6,$33,$1F,$C5
	.dc.b	$BF,$15,$7C,$1E,$A2,$4E,$8F,$B5,$3A,$7E,$E7,$CA,$3C,$CF,$E3,$ED
	.dc.b	$8C,$C7,$F2,$2F,$88,$FF,$26,$47,$C2,$13,$B7,$0D,$DB,$F8,$BB,$76
	.dc.b	$FE,$0D,$5C,$A7,$F7,$1F,$53,$FC,$74,$97,$FC,$67,$E2,$6B,$FC,$7E
	.dc.b	$2C,$FE,$FE,$7F,$92,$7B,$F2,$BF,$90,$3F,$C9,$FE,$03,$11,$FC,$9E
	.dc.b	$2D,$7F,$07,$9F,$F1,$9F,$B5,$2B,$FD,$80,$FF,$8E,$3E,$97,$CD,$9F
	.dc.b	$F9,$8A,$35,$3F,$E3,$1F,$81,$FF,$0E,$BF,$E3,$3A,$DA,$F7,$D0,$F8
	.dc.b	$FE,$84,$9F,$DB,$4E,$BD,$DC,$47,$39,$B9,$FF,$8F,$2F,$FC,$10,$D8
	.dc.b	$CE,$5F,$EA,$19,$ED,$5D,$1D,$C6,$67,$73,$E5,$9D,$BE,$9B,$F8,$1E
	.dc.b	$EC,$8E,$4F,$C3,$F8,$FC,$47,$BE,$2E,$3E,$59,$F9,$1F,$61,$2B,$C4
	.dc.b	$FC,$BF,$77,$11,$F1,$E2,$D7,$B6,$6E,$7E,$B9,$B5,$E7,$9F,$F3,$5F
	.dc.b	$E4,$7E,$8B,$FF,$2F,$B6,$FF,$D8,$4D,$AF,$A4,$FC,$26,$E9,$C8,$8F
	.dc.b	$D2,$8D,$6A,$79,$36,$77,$3F,$E7,$19,$D9,$A2,$47,$9B,$4F,$3E,$92
	.dc.b	$BB,$3E,$AD,$31,$1C,$E2,$E5,$EC,$9B,$97,$B2,$6D,$7B,$27,$FC,$2B
	.dc.b	$F1,$3F,$D2,$2D,$2F,$B1,$7D,$8F,$F7,$5A,$FF,$1A,$4E,$A7,$2F,$08
	.dc.b	$CB,$C6,$78,$78,$DF,$CB,$23,$06,$DC,$57,$28,$57,$28,$C3,$6B,$9D
	.dc.b	$F8,$5B,$5F,$ED,$9D,$2F,$8B,$F9,$4B,$F6,$4E,$4F,$05,$FB,$A7,$F3
	.dc.b	$31,$F0,$3E,$2F,$17,$FD,$52,$11,$87,$FA,$BB,$5F,$F7,$09,$D4,$B8
	.dc.b	$FA,$DF,$F7,$EE,$BE,$77,$89,$F0,$8F,$9F,$F1,$C8,$D4,$72,$CF,$FB
	.dc.b	$2C,$E6,$38,$1E,$72,$B9,$AF,$A7,$C3,$F9,$1E,$3E,$1A,$7E,$79,$B8
	.dc.b	$FE,$7B,$8E,$E3,$F9,$E9,$B7,$C6,$BF,$39,$18,$E3,$5D,$89,$3A,$3F
	.dc.b	$45,$18,$BF,$CF,$7C,$17,$13,$97,$0D,$AF,$D4,$45,$9F,$D2,$4D,$CB
	.dc.b	$F6,$B3,$6B,$F6,$B3,$6B,$F6,$B3,$72,$FD,$84,$DE,$5F,$F4,$55,$FF
	.dc.b	$1B,$AF,$6C,$7F,$89,$A3,$04,$BF,$E6,$53,$AA,$DB,$27,$58,$7F,$2B
	.dc.b	$9C,$57,$F3,$9F,$2B,$A9,$DB,$DE,$B7,$FC,$AE,$2C,$FF,$3F,$9B,$7F
	.dc.b	$CD,$7D,$FF,$29,$E6,$F7,$EE,$1F,$CC,$E2,$DF,$F4,$79,$B5,$C6,$B4
	.dc.b	$7F,$39,$8C,$BF,$CF,$6A,$D7,$F3,$99,$BA,$7F,$4E,$F7,$FE,$29,$F1
	.dc.b	$BE,$B2,$E1,$FD,$0E,$2D,$FF,$57,$9B,$8F,$F9,$A7,$CF,$FC,$4F,$AE
	.dc.b	$6F,$CF,$61,$98,$FF,$A3,$5C,$B4,$FF,$48,$7D,$6F,$23,$9F,$97,$63
	.dc.b	$24,$CB,$FD,$35,$19,$7F,$C1,$9F,$93,$7F,$86,$A7,$FC,$A0,$D8,$FF
	.dc.b	$18,$F4,$BE,$A9,$F0,$FF,$A2,$95,$6F,$5F,$F5,$3B,$70,$FF,$4B,$36
	.dc.b	$58,$73,$BE,$53,$99,$78,$6F,$8C,$EC,$DC,$3F,$CD,$D8,$3E,$9D,$AE
	.dc.b	$AE,$57,$F9,$A6,$F8,$61,$EA,$BD,$CE,$6B,$F7,$7F,$47,$FC,$EB,$CF
	.dc.b	$B7,$FE,$35,$AF,$F3,$E4,$FF,$3E,$F8,$AC,$D4,$9C,$D6,$CF,$F2,$BD
	.dc.b	$6E,$33,$3B,$9C,$9B,$DB,$F8,$B5,$6E,$33,$3F,$37,$EA,$1E,$B5,$3F
	.dc.b	$88,$ED,$F0,$7C,$CC,$65,$FE,$87,$AD,$B6,$B7,$AA,$B7,$8F,$F9,$2E
	.dc.b	$DF,$17,$DC,$FE,$3B,$EF,$BF,$02,$FA,$79,$75,$9D,$3F,$ED,$8A,$E2
	.dc.b	$D8,$FF,$73,$DF,$FB,$56,$72,$FF,$A6,$35,$7C,$B3,$D2,$FA,$2D,$C3
	.dc.b	$E7,$AF,$4B,$33,$27,$2F,$08,$C5,$FD,$0A,$E4,$D9,$D5,$70,$3F,$DE
	.dc.b	$DF,$23,$FB,$0F,$96,$FD,$8B,$63,$27,$FB,$EE,$32,$D3,$FD,$7D,$AB
	.dc.b	$F2,$47,$CC,$FC,$86,$CF,$F9,$3B,$C0,$7C,$BE,$6A,$4F,$2A,$DE,$7F
	.dc.b	$E9,$EF,$11,$87,$0A,$D5,$FF,$06,$F8,$2F,$E0,$FF,$2F,$67,$B3,$70
	.dc.b	$E5,$68,$C7,$57,$FB,$DD,$B5,$FE,$18,$C7,$BC,$62,$FF,$CD,$57,$C6
	.dc.b	$BF,$DA,$5F,$4C,$F1,$7F,$80,$36,$3E,$1F,$3B,$D0,$CD,$49,$FF,$02
	.dc.b	$60,$BE,$0F,$F8,$1D,$AF,$FB,$71,$AF,$97,$0F,$DC,$E7,$FD,$EF,$3F
	.dc.b	$8B,$D3,$FC,$A3,$BF,$F9,$4B,$71,$FA,$0D,$C3,$E1,$BF,$8E,$E4,$93
	.dc.b	$F8,$EE,$4B,$5F,$89,$9B,$8F,$ED,$39,$95,$8B,$FF,$1F,$F4,$7E,$6A
	.dc.b	$F1,$82,$3F,$3B,$58,$CD,$F9,$2B,$E2,$3F,$C0,$ED,$C7,$FD,$27,$F6
	.dc.b	$EF,$67,$35,$AF,$8A,$95,$FE,$98,$B7,$C6,$63,$4F,$C4,$43,$E0,$7F
	.dc.b	$C2,$79,$7F,$A8,$6F,$84,$63,$DC,$CF,$91,$27,$C3,$EF,$DC,$7F,$25
	.dc.b	$83,$DF,$E6,$67,$93,$E9,$46,$6F,$D3,$3E,$0E,$AD,$C7,$B1,$F7,$F8
	.dc.b	$72,$3D,$5E,$85,$C7,$D3,$37,$1F,$BD,$B7,$F6,$3B,$F8,$13,$97,$F3
	.dc.b	$11,$FC,$15,$E6,$68,$F9,$7E,$D7,$C2,$3E,$1F,$B3,$63,$24,$C3,$DF
	.dc.b	$19,$7E,$0C,$DC,$3D,$F3,$6B,$C3,$E9,$7D,$A3,$ED,$F4,$D8,$FA,$D1
	.dc.b	$70,$FA,$B3,$6F,$DB,$36,$BD,$73,$6B,$FD,$B5,$B7,$F8,$BD,$FA,$7E
	.dc.b	$CF,$DC,$F6,$7E,$EB,$2F,$F1,$2E,$5A,$49,$E2,$FE,$3A,$78,$47,$E2
	.dc.b	$3C,$4F,$C9,$9F,$94,$F1,$2D,$7E,$3E,$2C,$FE,$3A,$6E,$7F,$5A,$6D
	.dc.b	$7D,$49,$B5,$EA,$F6,$7F,$26,$7E,$53,$D9,$B5,$FA,$08,$B3,$F9,$09
	.dc.b	$B8,$7F,$BC,$7C,$7F,$F2,$89,$FE,$52,$FB,$AB,$87,$FB,$BE,$2D,$FF
	.dc.b	$29,$F6,$D5,$8B,$FF,$7D,$7D,$97,$F9,$68,$FF,$2D,$CD,$C3,$FD,$F1
	.dc.b	$16,$F9,$4D,$AF,$E2,$1F,$8E,$7F,$AF,$AF,$FD,$41,$D5,$FC,$A6,$3F
	.dc.b	$D2,$F7,$FE,$A0,$8C,$3F,$E9,$1F,$81,$46,$FD,$5C,$3A,$BE,$74,$F1
	.dc.b	$18,$74,$C3,$8C,$E6,$3C,$4F,$73,$18,$74,$74,$5A,$E7,$16,$78,$D7
	.dc.b	$49,$CD,$F8,$96,$4F,$AD,$E4,$D9,$D2,$FA,$DC,$CC,$9C,$BC,$23,$19
	.dc.b	$79,$C7,$D0,$FF,$1A,$CF,$FA,$2B,$63,$FD,$94,$D7,$FD,$DA,$7F,$BB
	.dc.b	$9C,$3F,$EE,$DE,$5F,$F0,$5B,$FE,$D1,$F8,$54,$62,$38,$9F,$EE,$67
	.dc.b	$0F,$1F,$2F,$C3,$E4,$7F,$31,$8F,$0B,$5E,$37,$FC,$E7,$C9,$D9,$FE
	.dc.b	$DF,$D9,$F0,$CE,$DF,$3F,$0F,$CD,$8C,$61,$F3,$5A,$FD,$9A,$E3,$E5
	.dc.b	$F0,$E4,$FE,$E9,$5A,$BE,$58,$67,$0F,$C3,$4E,$6B,$77,$9C,$1F,$43
	.dc.b	$FE,$2D,$BF,$B8,$FF,$8A,$DB,$E2,$6E,$35,$1E,$36,$E5,$F8,$57,$84
	.dc.b	$66,$78,$7A,$F2,$39,$90,$E4,$F8,$F2,$FF,$10,$C0,$FF,$0F,$38,$6F
	.dc.b	$AD,$F7,$BA,$7F,$BF,$4B,$C0,$C7,$47,$C4,$AB,$7F,$09,$BA,$7D,$26
	.dc.b	$DF,$19,$F0,$61,$D3,$C0,$C9,$23,$38,$7D,$0F,$47,$11,$E6,$9B,$5E
	.dc.b	$49,$EF,$C2,$F9,$BA,$4B,$E6,$7E,$EE,$1F,$1B,$98,$FF,$1F,$47,$17
	.dc.b	$7A,$7E,$26,$BA,$D7,$B3,$E8,$7F,$D4,$55,$A6,$BF,$E4,$76,$E3,$FE
	.dc.b	$25,$7F,$CF,$D1,$BC,$79,$11,$E5,$EA,$57,$94,$5A,$E5,$3E,$EF,$90
	.dc.b	$95,$E6,$CD,$1F,$0B,$DF,$CC,$FB,$22,$DF,$F8,$DE,$BF,$D3,$33,$C6
	.dc.b	$73,$B7,$F9,$46,$79,$CF,$5B,$FB,$7F,$3F,$FE,$C2,$8F,$8A,$D7,$FC
	.dc.b	$CC,$DC,$3F,$C8,$8F,$F2,$68,$CF,$A5,$C8,$8F,$2E,$9A,$E8,$EB,$1E
	.dc.b	$08,$FB,$BF,$43,$27,$E0,$3F,$80,$FA,$1C,$47,$DA,$8B,$87,$BE,$6E
	.dc.b	$7E,$D9,$B5,$F8,$39,$F6,$FA,$12,$3D,$AF,$C0,$F4,$31,$1E,$E8,$B8
	.dc.b	$7B,$A6,$E1,$FE,$79,$CD,$58,$BF,$E9,$50,$AF,$CB,$FC,$3F,$E6,$3E
	.dc.b	$12,$57,$DB,$8C,$B4,$87,$E5,$BE,$17,$F3,$E4,$7F,$04,$FE,$8B,$E7
	.dc.b	$3E,$BF,$A1,$FA,$4F,$85,$95,$F6,$23,$ED,$FA,$52,$3E,$DB,$F7,$BE
	.dc.b	$33,$EF,$FC,$64,$9F,$BE,$FE,$BB,$E8,$7F,$7F,$1F,$AF,$FA,$1B,$DC
	.dc.b	$69,$8F,$A1,$B3,$FB,$23,$81,$67,$83,$63,$6B,$91,$C9,$FD,$42,$B7
	.dc.b	$39,$8C,$5F,$78,$FD,$27,$D0,$FF,$12,$F8,$5E,$31,$FE,$C1,$AF,$E0
	.dc.b	$F3,$23,$04,$3F,$81,$BE,$EF,$A1,$70,$FB,$BE,$AE,$57,$AE,$72,$D9
	.dc.b	$FF,$56,$A8,$CB,$FD,$80,$C4,$6A,$33,$1F,$DB,$3C,$4F,$F6,$7F,$D0
	.dc.b	$7A,$7E,$86,$4F,$FB,$66,$3D,$86,$1F,$F6,$6C,$61,$D1,$FE,$DF,$75
	.dc.b	$B7,$FA,$DF,$45,$7F,$38,$F8,$E7,$D0,$BE,$13,$F9,$CF,$C1,$3E,$56
	.dc.b	$9B,$9D,$67,$FE,$F0,$6B,$F4,$C5,$F4,$BE,$07,$D3,$F8,$AF,$A9,$EC
	.dc.b	$7D,$5F,$8B,$97,$FE,$F5,$8C,$BF,$EF,$78,$CB,$FE,$EA,$8F,$C8,$FD
	.dc.b	$0B,$E1,$A9,$27,$F1,$D1,$FE,$10,$6C,$65,$F3,$08,$E4,$F0,$F7,$75
	.dc.b	$C1,$F8,$88,$D5,$63,$9F,$FC,$43,$7D,$67,$9F,$D0,$FC,$2B,$E1,$3F
	.dc.b	$C8,$EB,$7E,$5F,$EA,$2B,$51,$FE,$2A,$7E,$4D,$FF,$89,$1A,$E7,$5F
	.dc.b	$C5,$C6,$23,$FB,$C3,$FD,$AD,$F5,$FD,$0B,$E4,$7F,$6A,$D8,$FD,$84
	.dc.b	$EF,$2F,$63,$70,$F1,$7B,$1E,$53,$8F,$B5,$FB,$EF,$A1,$FE,$87,$87
	.dc.b	$C8,$6E,$1F,$15,$B7,$FE,$44,$7F,$A6,$CE,$FD,$3E,$44,$7F,$81,$46
	.dc.b	$AB,$94,$E6,$78,$1A,$F4,$BF,$87,$FD,$0F,$F7,$88,$FE,$15,$F4,$2F
	.dc.b	$80,$FE,$26,$D8,$FE,$1D,$3B,$7E,$FB,$70,$FB,$8D,$CB,$FC,$B6,$D9
	.dc.b	$FF,$12,$53,$E3,$7F,$8F,$FD,$0B,$D1,$FE,$43,$F4,$3F,$76,$FF,$C5
	.dc.b	$4D,$53,$B7,$F9,$03,$5F,$C8,$B7,$2F,$C7,$37,$2F,$F3,$6D,$7F,$97
	.dc.b	$E7,$F7,$1F,$29,$FC,$6C,$97,$FE,$16,$F0,$9F,$43,$88,$FD,$97,$8A
	.dc.b	$95,$66,$FF,$81,$5F,$F2,$5F,$83,$98,$73,$7F,$94,$72,$62,$39,$E1
	.dc.b	$FE,$54,$BF,$E1,$55,$A9,$FF,$13,$FC,$17,$F8,$79,$FF,$25,$FB,$DF
	.dc.b	$E4,$03,$FC,$87,$F1,$BE,$BD,$BF,$CD,$71,$99,$E2,$EB,$E1,$F7,$F2
	.dc.b	$B2,$7F,$C6,$76,$8E,$7D,$F9,$9F,$F6,$74,$5C,$FF,$9F,$F0,$7F,$AA
	.dc.b	$2B,$19,$7F,$97,$E3,$FD,$57,$58,$BF,$F5,$F6,$1F,$6E,$73,$1E,$09
	.dc.b	$D4,$73,$8F,$1F,$C0,$7B,$08,$F9,$BE,$56,$BF,$EE,$18,$B9,$79,$30
	.dc.b	$57,$C4,$F9,$59,$7E,$E7,$E3,$FC,$06,$23,$FD,$D5,$16,$F9,$4D,$CF
	.dc.b	$D9,$82,$B1,$7E,$CA,$FF,$39,$FC,$FF,$C9,$8F,$F2,$EB,$5F,$C3,$B7
	.dc.b	$0F,$92,$FC,$33,$18,$7C,$96,$CF,$FB,$95,$B9,$9E,$1A,$78,$09,$3E
	.dc.b	$0D,$6B,$E0,$8B,$87,$82,$7C,$FA,$CA,$F3,$BE,$AD,$71,$1E,$88,$B5
	.dc.b	$E8,$9B,$97,$9E,$6D,$79,$E7,$FC,$33,$F1,$DF,$D2,$6F,$FD,$45,$F4
	.dc.b	$0F,$A1,$FF,$6E,$CF,$F8,$DE,$F3,$18,$F8,$3F,$F7,$0D,$6E,$7F,$D5
	.dc.b	$55,$CD,$46,$54,$66,$30,$F5,$8D,$D7,$57,$B9,$D3,$CA,$FF,$11,$3A
	.dc.b	$31,$87,$FA,$E7,$8F,$ED,$5F,$85,$B5,$C1,$B3,$BD,$BF,$C2,$5F,$1D
	.dc.b	$70,$FF,$79,$B5,$FF,$73,$8A,$FC,$33,$FE,$C7,$8D,$34,$F2,$1F,$EC
	.dc.b	$BF,$91,$FF,$17,$E7,$F0,$63,$4D,$8C,$E7,$E2,$9E,$22,$32,$F9,$0C
	.dc.b	$B8,$E2,$FC,$77,$27,$F8,$F3,$3F,$CB,$B7,$87,$F2,$DC,$97,$0F,$CB
	.dc.b	$4D,$BE,$35,$FA,$28,$C7,$27,$F9,$E2,$30,$4A,$FD,$2C,$EA,$F6,$F0
	.dc.b	$E1,$95,$FA,$CE,$0B,$7F,$AC,$9B,$97,$EB,$26,$D7,$EA,$E6,$D7,$EA
	.dc.b	$E6,$E5,$FA,$B9,$B9,$7F,$87,$EF,$F7,$31,$EB,$E5,$FF,$46,$47,$F2
	.dc.b	$68,$C5,$E5,$91,$93,$F9,$5C,$63,$0F,$E8,$F5,$C9,$B5,$CA,$B8,$46
	.dc.b	$BC,$DF,$01,$E1,$F8,$1B,$3F,$CD,$A2,$CF,$49,$B9,$FF,$21,$F8,$2F
	.dc.b	$41,$E9,$F8,$2B,$87,$F3,$E8,$B7,$E4,$9B,$5C,$6B,$47,$F4,$48,$CB
	.dc.b	$FC,$8A,$AD,$7F,$3C,$9B,$AB,$37,$C1,$7C,$93,$E5,$7A,$F7,$1F,$E9
	.dc.b	$F1,$70,$F8,$B3,$72,$FF,$35,$7A,$3F,$8B,$F0,$3F,$CD,$18,$7F,$73
	.dc.b	$F4,$71,$32,$3F,$BB,$3B,$8C,$61,$FD,$B3,$E7,$F9,$1C,$FC,$9B,$7F
	.dc.b	$DF,$62,$CF,$F7,$29,$B8,$FF,$95,$7E,$83,$FC,$2D,$7F,$EA,$D6,$D7
	.dc.b	$F9,$CF,$A1,$F4,$2F,$88,$FD,$24,$AB,$78,$73,$C7,$F5,$18,$70,$F3
	.dc.b	$37,$5A,$CB,$F7,$D9,$7F,$88,$DF,$F5,$E7,$88,$F8,$DE,$AD,$86,$7E
	.dc.b	$3F,$A1,$6D,$AB,$95,$AD,$7F,$D6,$FA,$2F,$C1,$87,$2C,$7A,$3C,$F5
	.dc.b	$FD,$F7,$CE,$F6,$DA,$CF,$94,$FF,$4A,$47,$8B,$E2,$F3,$33,$3F,$01
	.dc.b	$E6,$9E,$4E,$29,$1C,$63,$0E,$A3,$A9,$DF,$3F,$F6,$7C,$65,$B0,$BF
	.dc.b	$03,$FE,$92,$FA,$17,$AB,$F5,$1B,$87,$87,$5F,$85,$E5,$E6,$A4,$E7
	.dc.b	$EB,$7F,$9B,$3F,$EE,$DB,$E3,$87,$E0,$67,$2F,$7F,$8D,$FF,$4E,$3F
	.dc.b	$50,$4A,$33,$1F,$62,$31,$5F,$ED,$98,$DC,$E9,$B1,$F1,$F3,$7D,$2B
	.dc.b	$E5,$7D,$97,$FC,$CE,$73,$1F,$63,$E6,$5E,$95,$31,$E9,$DC,$7E,$7C
	.dc.b	$E9,$B3,$A6,$93,$2F,$BF,$3B,$6D,$73,$7F,$CB,$7F,$29,$FB,$9F,$6F
	.dc.b	$F7,$4D,$BF,$F3,$63,$67,$71,$FE,$6F,$9F,$E9,$9F,$0A,$FA,$37,$D0
	.dc.b	$61,$B5,$FE,$75,$F9,$9C,$4C,$8F,$C6,$78,$EF,$8A,$DE,$5F,$EE,$9C
	.dc.b	$7A,$5F,$F5,$79,$CF,$07,$FB,$E7,$E0,$FF,$88,$70,$FF,$11,$6E,$5F
	.dc.b	$E8,$36,$E3,$FE,$78,$6D,$71,$9F,$F4,$45,$71,$69,$F9,$FA,$7D,$17
	.dc.b	$FB,$2E,$3F,$00,$D8,$FC,$41,$FB,$C7,$0E,$1A,$FE,$35,$E1,$5F,$BC
	.dc.b	$E6,$7B,$7F,$C4,$A7,$37,$FE,$09,$9C,$BF,$E9,$DF,$07,$F9,$77,$57
	.dc.b	$F2,$F6,$C6,$4F,$F5,$14,$65,$A7,$E2,$38,$FF,$1C,$7E,$3F,$8E,$C6
	.dc.b	$49,$9F,$E2,$23,$2F,$E2,$A6,$E7,$FB,$CE,$5F,$F6,$85,$F0,$8F,$F4
	.dc.b	$7F,$D1,$3E,$27,$FC,$1C,$D9,$FB,$FC,$3F,$B2,$AD,$D6,$BE,$2A,$57
	.dc.b	$E1,$2D,$F1,$78,$CF,$FD,$77,$1F,$EB,$A9,$E7,$EE,$F6,$F8,$2F,$C3
	.dc.b	$3F,$87,$F1,$A4,$F8,$39,$2E,$1F,$EC,$4D,$7F,$D2,$33,$E4,$F8,$0C
	.dc.b	$ED,$FE,$A8,$9C,$D6,$E7,$F2,$5E,$D7,$E4,$DB,$7E,$7E,$EF,$79,$F0
	.dc.b	$3C,$FB,$19,$25,$FA,$A3,$2F,$B2,$6E,$3F,$2A,$6D,$7F,$AD,$F1,$FF
	.dc.b	$1A,$5F,$A7,$DB,$FD,$27,$B3,$FA,$56,$57,$CE,$72,$D2,$4F,$17,$E9
	.dc.b	$CF,$08,$F9,$BE,$5F,$D8,$3E,$CF,$35,$AF,$A9,$16,$7E,$9C,$DC,$BF
	.dc.b	$03,$36,$BF,$D9,$DC,$4A,$C5,$BC,$0F,$E0,$1E,$D7,$EF,$4F,$DF,$7B
	.dc.b	$57,$0F,$C3,$C5,$AF,$B1,$37,$2F,$AD,$36,$BE,$34,$DA,$ED,$8B,$C0
	.dc.b	$FC,$99,$F9,$4F,$02,$DF,$E3,$E2,$D7,$E3,$A6,$E3,$F9,$C9,$B5,$EC
	.dc.b	$9B,$5E,$BD,$DE,$D7,$F1,$73,$F8,$C7,$B5,$6F,$F4,$31,$6B,$F2,$53
	.dc.b	$71,$FF,$72,$78,$7F,$E5,$B3,$FC,$B9,$F4,$57,$0F,$F8,$03,$16,$FF
	.dc.b	$CA,$F3,$6B,$FD,$D9,$F4,$5F,$E6,$33,$FC,$C9,$ED,$DC,$3F,$97,$A8
	.dc.b	$C5,$FF,$98,$22,$D2,$BF,$C2,$FC,$2F,$A2,$FF,$0C,$C7,$FA,$87,$AF
	.dc.b	$31,$FE,$97,$8C,$61,$F7,$63,$6B,$E0,$7F,$8F,$B9,$F4,$7B,$39,$BE
	.dc.b	$6E,$27,$78,$78,$6B,$15,$AC,$3B,$CE,$5B,$87,$82,$F8,$BE,$5C,$3B
	.dc.b	$1E,$53,$99,$E3,$9E,$79,$3E,$8B,$38,$7F,$99,$DB,$0F,$A2,$B1,$2B
	.dc.b	$8B,$A8,$C5,$F8,$07,$D1,$7F,$1C,$D3,$FC,$14,$D8,$FE,$96,$6B,$C4
	.dc.b	$FF,$61,$38,$8D,$3F,$D4,$3E,$19,$D2,$F2,$3F,$C2,$91,$86,$C6,$67
	.dc.b	$AC,$E5,$E9,$E4,$7F,$B7,$5B,$F0,$DE,$48,$C4,$7F,$5E,$C3,$F9,$E4
	.dc.b	$F9,$B7,$7E,$4B,$FE,$E8,$E3,$FC,$3B,$E8,$BF,$9F,$3A,$D7,$FC,$5D
	.dc.b	$C7,$F7,$2B,$85,$7F,$03,$6D,$6E,$3E,$EC,$73,$9E,$11,$CA,$FF,$2B
	.dc.b	$F2,$FF,$D3,$E3,$F0,$6D,$7C,$2D,$9F,$C5,$AF,$2F,$94,$FF,$13,$8D
	.dc.b	$57,$4D,$33,$8A,$73,$5B,$BC,$C6,$27,$55,$83,$33,$CB,$17,$D7,$7F
	.dc.b	$93,$F9,$FF,$43,$7F,$2B,$E9,$B1,$CD,$F2,$9B,$9F,$C5,$BB,$5F,$41
	.dc.b	$B8,$2A,$98,$8C,$FC,$74,$BF,$F1,$7D,$23,$8F,$CD,$66,$59,$79,$2B
	.dc.b	$51,$C2,$B6,$FF,$7A,$E9,$F1,$F8,$1F,$A7,$8F,$EF,$BD,$58,$99,$18
	.dc.b	$21,$FD,$C5,$F5,$FD,$0F,$FA,$9E,$3F,$C8,$6B,$D2,$C9,$AF,$4E,$4E
	.dc.b	$A3,$2F,$08,$CB,$EA,$E0,$FE,$33,$5C,$E3,$81,$B3,$FC,$27,$DB,$F3
	.dc.b	$3E,$E3,$FC,$77,$B7,$D4,$6D,$FF,$71,$9F,$79,$BA,$87,$22,$1E,$13
	.dc.b	$FA,$F4,$F2,$69,$C8,$FF,$30,$FE,$38,$F4,$BE,$EB,$FC,$9A,$39,$72
	.dc.b	$23,$18,$FE,$21,$E1,$3C,$AB,$8F,$44,$8C,$10,$E7,$3C,$B2,$FE,$7A
	.dc.b	$D8,$FE,$E8,$F4,$7C,$38,$FE,$EF,$0F,$89,$C9,$F1,$CF,$91,$D5,$2B
	.dc.b	$FC,$F7,$5F,$9E,$6C,$E5,$E7,$3C,$5C,$9C,$1F,$83,$E7,$7C,$2F,$3A
	.dc.b	$4F,$C6,$8C,$BF,$22,$32,$FB,$23,$2F,$BA,$33,$87,$C7,$9F,$BD,$D1
	.dc.b	$2B,$CE,$FE,$07,$9A,$46,$08,$7C,$27,$2D,$3D,$EF,$E2,$7A,$3F,$15
	.dc.b	$F1,$D2,$7E,$1B,$F7,$3C,$9F,$BB,$E4,$D9,$FC,$0C,$61,$FC,$1F,$C6
	.dc.b	$FE,$08,$8F,$CC,$F9,$FF,$E0,$6B,$E8,$1F,$2F,$87,$FD,$D3,$1F,$E9
	.dc.b	$96,$BF,$93,$38,$46,$CE,$1F,$09,$23,$84,$65,$E5,$19,$79,$4F,$81
	.dc.b	$FB,$FF,$43,$FA,$98,$91,$82,$1F,$8A,$72,$FE,$32,$3F,$61,$F4,$DF
	.dc.b	$B1,$B9,$3F,$8A,$8C,$BF,$8C,$8C,$BF,$94,$8F,$DC,$5F,$EE,$7E,$96
	.dc.b	$4F,$E4,$A3,$2F,$E5,$23,$2F,$E7,$23,$2F,$E7,$A3,$3C,$5F,$C0,$67
	.dc.b	$E7,$FD,$17,$F4,$1A,$FF,$80,$0D,$7F,$BB,$B5,$FF,$61,$B5,$11,$E1
	.dc.b	$3F,$E2,$08,$46,$3A,$77,$18,$F1,$FE,$E7,$27,$F1,$5E,$FF,$E2,$DF
	.dc.b	$21,$2B,$F5,$F1,$97,$F6,$31,$97,$F7,$31,$F8,$8F,$9F,$7A,$77,$D3
	.dc.b	$95,$FE,$DF,$9C,$B6,$72,$FE,$26,$3F,$C9,$0E,$48,$66,$77,$A7,$B6
	.dc.b	$79,$CF,$F1,$0F,$0F,$FA,$CF,$E2,$BF,$C8,$FC,$43,$7C,$2F,$F1,$65
	.dc.b	$7F,$24,$AD,$B0,$C1,$FE,$F5,$F9,$F7,$BB,$FE,$28,$8F,$9F,$18,$70
	.dc.b	$D7,$FC,$5A,$D7,$FD,$F2,$E7,$2F,$A8,$DE,$1F,$A6,$D9,$FB,$79,$FF
	.dc.b	$89,$2B,$ED,$7D,$07,$F8,$2E,$7E,$FC,$63,$E0,$FF,$6F,$26,$F9,$F2
	.dc.b	$B7,$8F,$E0,$2B,$51,$83,$F1,$2E,$1F,$DD,$3D,$DB,$3D,$AA,$47,$11
	.dc.b	$3B,$BF,$C8,$7A,$FF,$6A,$7F,$C9,$ED,$7F,$22,$7F,$BC,$38,$3E,$63
	.dc.b	$4F,$9D,$1C,$67,$9C,$FF,$14,$8C,$63,$F9,$AF,$8A,$FC,$0D,$7E,$7F
	.dc.b	$F1,$CF,$F2,$BE,$1F,$5D,$AF,$F9,$81,$AF,$89,$AF,$E9,$1D,$C7,$3F
	.dc.b	$63,$F9,$1C,$62,$FF,$53,$EF,$FB,$DF,$F4,$CF,$98,$F9,$1F,$F1,$5F
	.dc.b	$E2,$FF,$E4,$1F,$46,$F0,$FF,$84,$74,$FE,$BF,$5F,$E0,$33,$33,$23
	.dc.b	$FA,$EC,$7F,$81,$FE,$46,$47,$F6,$07,$FC,$15,$DD,$88,$FE,$D3,$17
	.dc.b	$6F,$F0,$64,$FF,$8E,$3C,$AF,$F1,$CF,$8F,$2F,$FC,$6B,$19,$7F,$C6
	.dc.b	$F1,$97,$FC,$07,$EB,$A7,$FC,$83,$E3,$FF,$90,$BC,$89,$5F,$E3,$E8
	.dc.b	$CB,$FE,$40,$8C,$BC,$A3,$2F,$48,$CE,$3F,$E4,$19,$CB,$FE,$43,$8C
	.dc.b	$F0,$7F,$9F,$A7,$FC,$A3,$E3,$FF,$94,$BC,$89,$87,$F9,$3A,$32,$FF
	.dc.b	$94,$23,$2F,$AA,$3F,$CB,$3E,$3F,$F9,$6B,$C8,$93,$FE,$56,$8C,$BF
	.dc.b	$E5,$88,$CB,$F0,$A3,$2F,$C4,$8C,$E5,$FE,$59,$9F,$F3,$3F,$DF,$BD
	.dc.b	$6B,$EB,$CC,$3F,$5D,$E3,$2B,$51,$86,$BF,$E3,$27,$FC,$BC,$D9,$EB
	.dc.b	$52,$38,$47,$57,$84,$65,$E0,$68,$EC,$FF,$9F,$3E,$FB,$FC,$51,$1F
	.dc.b	$E4,$07,$FC,$F8,$F9,$9A,$E7,$D5,$E0,$71,$F3,$24,$6E,$32,$F1,$8C
	.dc.b	$BC,$4F,$E0,$DE,$F3,$B7,$FD,$03,$0F,$9E,$FC,$73,$FD,$6A,$5E,$B7
	.dc.b	$F5,$69,$CC,$A7,$2F,$F6,$08,$CC,$75,$DF,$1C,$DF,$D7,$A3,$15,$DE
	.dc.b	$31,$7D,$5B,$1C,$0F,$16,$1E,$6C,$3A,$BE,$8F,$79,$F1,$7D,$63,$0D
	.dc.b	$CB,$FB,$9C,$EA,$F9,$C6,$27,$FD,$E3,$EF,$D9,$FE,$46,$D7,$FD,$B2
	.dc.b	$2B,$93,$86,$BF,$E0,$08,$F4,$3F,$E0,$2F,$69,$D3,$88,$D1,$F0,$4F
	.dc.b	$F0,$0C,$69,$C1,$E0,$CB,$E7,$A8,$D4,$62,$15,$D6,$73,$9F,$F8,$4B
	.dc.b	$91,$C2,$E5,$ED,$AF,$FB,$4A,$1C,$D8,$D3,$84,$6D,$FC,$94,$67,$36
	.dc.b	$1D,$30,$C3,$DA,$6D,$F1,$AF,$CE,$46,$2B,$2D,$BE,$A7,$ED,$3E,$FB
	.dc.b	$FC,$BA,$D9,$EB,$18,$AF,$F5,$2B,$D7,$1F,$F7,$5B,$96,$CF,$0E,$9E
	.dc.b	$2D,$BE,$87,$78,$CB,$4C,$DF,$68,$CB,$FC,$CB,$D3,$F7,$F6,$BD,$4B
	.dc.b	$18,$6D,$FF,$BC,$1B,$5F,$CE,$9B,$8F,$FB,$BD,$B3,$F5,$EF,$A3,$FE
	.dc.b	$F8,$F0,$BB,$F8,$5E,$06,$C6,$4F,$F7,$EC,$65,$A7,$FC,$00,$F0,$BC
	.dc.b	$FF,$27,$E8,$6C,$64,$FE,$47,$69,$D4,$61,$B9,$7F,$52,$6E,$5F,$97
	.dc.b	$E1,$5F,$1B,$E9,$7B,$6D,$6E,$75,$3F,$D7,$3E,$AB,$E5,$FD,$57,$CC
	.dc.b	$6D,$F9,$1B,$3B,$8F,$3C,$FE,$43,$C9,$E7,$87,$FB,$E7,$2C,$61,$96
	.dc.b	$5F,$FB,$59,$CB,$4F,$F6,$CD,$FE,$77,$C9,$7E,$37,$FE,$7A,$9C,$9F
	.dc.b	$E3,$58,$CE,$5E,$1A,$E1,$5B,$8F,$F7,$4C,$61,$FD,$40,$FC,$65,$EB
	.dc.b	$61,$B8,$7F,$90,$7A,$9F,$36,$FA,$BD,$5C,$AF,$F3,$05,$F0,$C3,$FC
	.dc.b	$72,$75,$9F,$F4,$73,$FE,$8C,$8D,$3F,$E3,$EF,$BC,$7B,$DE,$46,$DF
	.dc.b	$F9,$47,$F0,$E7,$E1,$CF,$A9,$D5,$CA,$FF,$38,$5F,$0C,$3C,$2D,$7B
	.dc.b	$4F,$89,$7D,$03,$E3,$BF,$91,$DF,$F9,$7E,$31,$AF,$F0,$D7,$51,$67
	.dc.b	$E1,$3B,$78,$D3,$97,$F9,$45,$72,$AD,$1C,$6F,$93,$4C,$9C,$DF,$AB
	.dc.b	$F4,$1F,$8A,$8F,$F5,$9B,$6D,$F3,$AF,$B3,$7C,$6F,$0E,$89,$87,$FA
	.dc.b	$4F,$0F,$CE,$E1,$FE,$6C,$6B,$D2,$7E,$27,$67,$FA,$89,$FE,$53,$F0
	.dc.b	$AE,$BB,$18,$6E,$1F,$E1,$89,$D3,$67,$4F,$C5,$F8,$4F,$91,$F4,$BF
	.dc.b	$25,$B1,$93,$E5,$46,$5A,$67,$1F,$98,$FF,$B6,$B7,$FB,$13,$B9,$C9
	.dc.b	$F7,$38,$51,$FD,$0B,$C3,$FA,$BC,$5F,$C1,$5F,$B3,$5C,$35,$FE,$3B
	.dc.b	$8F,$F8,$CA,$73,$5B,$8E,$5C,$5C,$DA,$7F,$A7,$7E,$7F,$BB,$71,$FE
	.dc.b	$A9,$C7,$F5,$7E,$77,$F5,$F5,$AC,$FC,$E1,$C3,$FE,$A0,$6B,$FE,$F2
	.dc.b	$6B,$FE,$FA,$6E,$7F,$CF,$FE,$C3,$DD,$F0,$7E,$F6,$E5,$EB,$6D,$FB
	.dc.b	$F5,$46,$35,$F8,$8D,$AF,$EE,$D3,$A7,$E2,$5F,$CB,$F9,$4F,$99,$F0
	.dc.b	$36,$7E,$43,$71,$EC,$DC,$7E,$83,$F4,$FE,$1F,$EA,$76,$5B,$FA,$0D
	.dc.b	$BF,$33,$6F,$EC,$3F,$B1,$FA,$1F,$D9,$7C,$D5,$BF,$D9,$36,$FF,$07
	.dc.b	$CB,$FB,$D8,$C6,$1F,$B4,$E5,$46,$2F,$F1,$91,$87,$F7,$98,$76,$7F
	.dc.b	$CE,$BF,$8C,$FF,$11,$6E,$5F,$8D,$F8,$8F,$F0,$7C,$E7,$3F,$C6,$3F
	.dc.b	$EE,$FA,$E3,$3A,$D3,$6F,$E5,$7E,$7E,$4E,$8F,$C9,$3F,$99,$E9,$91
	.dc.b	$F9,$5E,$8B,$8F,$DB,$9F,$CF,$7D,$2E,$37,$FC,$F3,$8B,$FC,$E3,$97
	.dc.b	$EE,$BE,$8B,$F4,$B6,$3D,$ED,$BF,$5B,$6B,$FB,$76,$68,$C5,$7F,$8F
	.dc.b	$63,$0F,$C8,$BE,$AF,$EB,$7E,$83,$F6,$9F,$41,$FB,$56,$C6,$4F,$D9
	.dc.b	$46,$5A,$67,$1F,$CD,$46,$5B,$1F,$5B,$E8,$BE,$9F,$07,$D4,$6C,$64
	.dc.b	$FB,$71,$96,$99,$9F,$B3,$19,$6C,$7F,$05,$9F,$E1,$93,$FC,$35,$B1
	.dc.b	$93,$F8,$54,$65,$A6,$67,$F6,$51,$96,$C7,$E3,$BE,$2F,$F1,$7F,$37
	.dc.b	$F8,$C6,$C6,$4F,$C9,$C6,$5A,$66,$7F,$23,$19,$6C,$7F,$1D,$E4,$FE
	.dc.b	$49,$F4,$1F,$C9,$5B,$19,$3F,$91,$46,$5A,$7F,$B5,$BF,$60,$FE,$56
	.dc.b	$FF,$9E,$DB,$5F,$E7,$27,$F9,$4C,$64,$FF,$39,$BA,$ED,$34,$DC,$F1
	.dc.b	$9D,$F6,$F0,$EC,$B1,$E1,$6C,$FF,$39,$A4,$63,$8F,$FC,$85,$18,$AF
	.dc.b	$F7,$6F,$E3,$1F,$EC,$87,$FD,$49,$F3,$DF,$E9,$38,$C6,$1F,$E9,$88
	.dc.b	$C7,$C2,$FF,$A8,$7E,$09,$1B,$F8,$6C,$37,$0E,$77,$FE,$92,$C3,$9D
	.dc.b	$F0,$8C,$61,$C1,$D6,$17,$55,$F0,$C7,$A5,$78,$CE,$38,$F6,$AD,$DF
	.dc.b	$1C,$9E,$0D,$F0,$70,$DB,$E1,$3A,$6C,$E9,$A7,$AB,$07,$D0,$7C,$CF
	.dc.b	$57,$FC,$77,$18,$E0,$FF,$60,$3F,$E3,$E9,$D3,$63,$28,$FF,$77,$7C
	.dc.b	$9F,$F8,$BD,$FF,$22,$FC,$4A,$31,$1B,$3F,$DF,$CE,$1E,$3E,$27,$8A
	.dc.b	$FD,$FF,$11,$75,$FF,$71,$57,$8D,$AF,$33,$D6,$E1,$FF,$4E,$1D,$F8
	.dc.b	$FE,$B7,$07,$FA,$23,$4C,$70,$FF,$A8,$5B,$5F,$EC,$EF,$77,$FD,$E7
	.dc.b	$3C,$A3,$EA,$47,$C0,$BE,$6D,$9D,$4F,$DB,$B7,$BD,$F6,$23,$EE,$35
	.dc.b	$D3,$6F,$CA,$B8,$7C,$75,$CB,$2B,$3B,$C3,$FD,$DC,$DC,$3F,$DC,$AF
	.dc.b	$17,$BC,$F0,$23,$E2,$BE,$B8,$BE,$2F,$EA,$73,$FF,$A9,$FE,$2E,$F2
	.dc.b	$7F,$B1,$5B,$87,$C5,$AB,$A7,$FA,$C9,$B8,$2A,$98,$8C,$FC,$F4,$BF
	.dc.b	$F6,$0B,$C3,$E7,$B1,$1F,$E3,$7B,$4E,$32,$FF,$5D,$57,$83,$E8,$1F
	.dc.b	$1F,$7E,$42,$63,$F9,$3B,$FF,$63,$56,$A3,$1A,$FE,$5F,$54,$E6,$7F
	.dc.b	$C8,$CA,$F9,$C6,$30,$F7,$FC,$43,$C3,$BE,$1C,$AE,$A3,$15,$6E,$33
	.dc.b	$18,$AE,$11,$FE,$55,$95,$76,$99,$3C,$27,$2D,$9C,$BE,$8F,$81,$FA
	.dc.b	$7F,$0D,$FA,$3D,$7F,$DC,$CD,$D7,$E0,$6B,$FE,$C2,$BF,$1B,$5F,$F3
	.dc.b	$34,$ED,$F5,$37,$3F,$8F,$87,$D1,$3B,$CC,$39,$BF,$84,$F9,$BF,$F4
	.dc.b	$BC,$7C,$07,$F9,$33,$FE,$73,$6E,$3F,$EF,$B7,$71,$9F,$77,$97,$44
	.dc.b	$8E,$51,$D4,$C3,$CE,$30,$FC,$EC,$7F,$8A,$DF,$C2,$E4,$F8,$A7,$C6
	.dc.b	$EA,$95,$B8,$CB,$C6,$32,$FC,$48,$CB,$F1,$A3,$DD,$F4,$1E,$FF,$0E
	.dc.b	$4F,$BE,$32,$FC,$18,$CB,$F0,$63,$2F,$C3,$8C,$DF,$C3,$9F,$B7,$D1
	.dc.b	$2B,$D8,$FD,$EE,$79,$18,$21,$F6,$9C,$B4,$F8,$8F,$E1,$B9,$FF,$0F
	.dc.b	$F2,$12,$7E,$43,$F6,$7E,$7F,$ED,$7D,$05,$9F,$B5,$18,$7E,$E7,$C1
	.dc.b	$FD,$F2,$3F,$47,$E6,$7F,$83,$AF,$99,$7C,$86,$17,$C8,$D3,$11,$27
	.dc.b	$F5,$67,$18,$D3,$89,$E6,$D9,$CB,$FE,$B0,$8E,$AE,$4F,$C2,$7C,$CF
	.dc.b	$E9,$A2,$46,$08,$7E,$89,$CB,$FA,$48,$FD,$6F,$CC,$FE,$BA,$A4,$FE
	.dc.b	$3E,$32,$FE,$46,$32,$FE,$B2,$3F,$6B,$7F,$B6,$F9,$59,$3F,$97,$8C
	.dc.b	$BF,$99,$8C,$BF,$B4,$8C,$BF,$B6,$8C,$F1,$7E,$FA,$7E,$7F,$D0,$7F
	.dc.b	$50,$AF,$8D,$C1,$F3,$CC,$F8,$F2,$3F,$BF,$CE,$DF,$ED,$41,$C3,$FD
	.dc.b	$B5,$AF,$F2,$9E,$8F,$E3,$AF,$EB,$FE,$0B,$F8,$89,$FC,$4B,$C2,$91
	.dc.b	$92,$1F,$C3,$9F,$C2,$FC,$F3,$E8,$2F,$CF,$4A,$FC,$0C,$E5,$B3,$97
	.dc.b	$FD,$DB,$09,$E7,$32,$78,$C6,$F5,$78,$9F,$E1,$1B,$E9,$5F,$C4,$FC
	.dc.b	$2F,$EB,$7F,$17,$FC,$7C,$FE,$41,$E1,$FF,$25,$C3,$F8,$B4,$EA,$A1
	.dc.b	$FC,$75,$FC,$FF,$D0,$3E,$0B,$DB,$E2,$CA,$FC,$DC,$E5,$B3,$FE,$30
	.dc.b	$51,$97,$E9,$B5,$FF,$0F,$35,$FF,$15,$7D,$07,$F7,$F5,$87,$B9,$FF
	.dc.b	$1A,$43,$E1,$BF,$66,$D8,$FD,$6B,$C6,$78,$42,$31,$AF,$6A,$E2,$72
	.dc.b	$F3,$E5,$72,$8C,$BD,$23,$F8,$0F,$AF,$F5,$63,$F7,$7F,$0D,$FD,$BF
	.dc.b	$1E,$ED,$C3,$FC,$89,$5B,$AF,$5C,$63,$87,$E3,$4E,$67,$FC,$80,$D7
	.dc.b	$47,$F9,$6B,$E8,$3F,$CA,$0F,$F0,$AF,$71,$F1,$1F,$C5,$1B,$1F,$C3
	.dc.b	$A7,$6F,$80,$FF,$2E,$38,$7F,$C9,$CD,$7F,$CB,$4D,$77,$ED,$7F,$9A
	.dc.b	$BC,$F7,$53,$E6,$FF,$C5,$3F,$35,$FE,$3D,$F5,$2F,$0F,$F8,$17,$1F
	.dc.b	$EE,$35,$FD,$E8,$CC,$C8,$FE,$B9,$1F,$DF,$BE,$6E,$47,$F5,$F7,$FC
	.dc.b	$03,$E0,$62,$3F,$BD,$C5,$E3,$FF,$01,$CF,$F8,$DB,$CC,$FF,$1B,$F9
	.dc.b	$32,$FF,$C2,$BE,$C2,$75,$18,$9D,$3F,$E3,$58,$FF,$1F,$F9,$3F,$E4
	.dc.b	$0F,$2A,$4F,$28,$CB,$D2,$32,$FF,$8F,$63,$2F,$F8,$FE,$33,$8F,$E6
	.dc.b	$78,$18,$AF,$73,$98,$FE,$99,$7D,$4F,$01,$EA,$F2,$7F,$CA,$7E,$54
	.dc.b	$8C,$10,$FF,$27,$39,$7F,$CA,$11,$FE,$5A,$F2,$7F,$CB,$7E,$54,$9F
	.dc.b	$8B,$19,$7E,$3C,$65,$FF,$2C,$46,$5F,$F2,$D4,$67,$0F,$44,$FF,$9A
	.dc.b	$3E,$83,$F4,$15,$FE,$47,$7F,$CA,$0F,$F9,$19,$AF,$E4,$F3,$47,$06
	.dc.b	$FB,$1C,$FD,$99,$5C,$63,$B1,$87,$95,$74,$C3,$E7,$3C,$C9,$5B,$8C
	.dc.b	$D7,$F9,$FB,$E7,$BF,$D0,$D7,$FE,$5A,$7F,$CF,$CD,$F4,$F5,$EB,$C8
	.dc.b	$C1,$0F,$54,$F2,$6C,$75,$99,$1D,$63,$FC,$13,$F0,$0E,$EF,$E3,$B2
	.dc.b	$F9,$FF,$CB,$3F,$ED,$19,$7B,$1F,$EC,$D9,$CC,$A7,$2F,$FB,$5E,$33
	.dc.b	$1D,$73,$C7,$37,$FB,$56,$31,$5D,$E3,$17,$D5,$B1,$C0,$F1,$61,$E6
	.dc.b	$C3,$AB,$E8,$F7,$DF,$41,$D6,$30,$DC,$BF,$DD,$53,$AB,$E7,$18,$9F
	.dc.b	$EE,$BF,$01,$67,$FD,$C0,$D7,$FD,$BC,$2B,$93,$86,$BF,$EF,$78,$F4
	.dc.b	$3F,$EF,$9F,$73,$FA,$46,$3F,$22,$34,$7C,$13,$FD,$F1,$1A,$CB,$C7
	.dc.b	$19,$39,$79,$1D,$1E,$35,$CE,$B8,$47,$FA,$97,$83,$FC,$83,$5F,$8B
	.dc.b	$E6,$FE,$B1,$A7,$F0,$19,$CD,$7E,$22,$73,$5F,$EB,$99,$CD,$6E,$73
	.dc.b	$5D,$79,$AF,$32,$D5,$87,$8E,$1C,$E3,$15,$97,$8F,$D0,$7F,$96,$27
	.dc.b	$FD,$72,$F0,$C3,$FD,$7C,$E5,$FD,$B3,$9C,$0E,$71,$FE,$38,$E4,$C5
	.dc.b	$7F,$B4,$DC,$74,$F8,$31,$F0,$B6,$7C,$2D,$C3,$BB,$6F,$F9,$27,$AB
	.dc.b	$F0,$3A,$DE,$B5,$8C,$37,$0F,$E7,$ED,$C3,$FD,$E0,$DB,$FE,$5C,$D9
	.dc.b	$FF,$74,$5F,$47,$FA,$3F,$85,$DF,$C2,$F0,$36,$32,$7F,$4C,$8C,$B4
	.dc.b	$FE,$9F,$E1,$79,$FE,$5B,$D0,$D8,$C9,$FD,$52,$32,$D3,$39,$7F,$25
	.dc.b	$C1,$3A,$8C,$37,$2F,$E9,$0D,$AF,$5B,$F1,$FE,$5F,$E4,$7C,$FD,$9F
	.dc.b	$EC,$6D,$BE,$CD,$BF,$EA,$4F,$E3,$BC,$9F,$0D,$7F,$B4,$70,$C6,$19
	.dc.b	$64,$64,$FF,$6B,$46,$30,$FF,$78,$DF,$E6,$BC,$97,$D0,$27,$B4,$E6
	.dc.b	$31,$7F,$ED,$F8,$C4,$EA,$38,$65,$FE,$43,$BE,$73,$FA,$61,$F3,$EF
	.dc.b	$5F,$FC,$83,$1D,$CC,$B8,$7C,$51,$BB,$FF,$43,$C6,$9B,$19,$8F,$F1
	.dc.b	$55,$3F,$1B,$78,$DE,$1A,$D6,$3E,$59,$CE,$1F,$EC,$0F,$9E,$7C,$07
	.dc.b	$87,$24,$62,$7C,$2D,$8F,$F4,$CB,$5F,$51,$C2,$7F,$82,$FC,$DB,$F0
	.dc.b	$27,$D3,$7C,$3B,$E5,$97,$F9,$A6,$B5,$3F,$C3,$3E,$6D,$F1,$DF,$E8
	.dc.b	$AC,$FE,$69,$F5,$63,$3F,$3D,$B9,$D6,$9F,$E4,$B9,$CD,$70,$84,$74
	.dc.b	$8E,$15,$C5,$A6,$E3,$4D,$A4,$EE,$30,$FD,$DF,$9F,$FC,$9D,$FF,$AF
	.dc.b	$1A,$FC,$46,$BF,$7D,$AF,$E3,$DE,$39,$7F,$34,$FA,$97,$CA,$F5,$E8
	.dc.b	$FE,$5F,$29,$CE,$3E,$DF,$0D,$F0,$BC,$A3,$F3,$51,$8E,$AF,$E6,$B3
	.dc.b	$9C,$D3,$9A,$E5,$39,$8C,$4F,$F8,$8E,$30,$2B,$E8,$7C,$1F,$D1,$F9
	.dc.b	$BB,$87,$C7,$6D,$FD,$3D,$D1,$8C,$FF,$D5,$1C,$1F,$7A,$31,$7F,$76
	.dc.b	$B6,$FD,$1C,$3B,$3E,$FF,$0B,$EF,$E1,$FE,$59,$7E,$AC,$64,$FE,$A6
	.dc.b	$E3,$1F,$F1,$E3,$C2,$3E,$2B,$99,$E5,$1C,$38,$7F,$71,$F3,$FE,$2B
	.dc.b	$FF,$7B,$B5,$FC,$FB,$5F,$DD,$35,$FE,$14,$F1,$AF,$F6,$37,$94,$FA
	.dc.b	$87,$C8,$F6,$D7,$FB,$5C,$E6,$3D,$3F,$31,$EA,$F8,$3B,$A7,$C1,$6E
	.dc.b	$7E,$C6,$E5,$F0,$DB,$3E,$E6,$CF,$E0,$35,$47,$C6,$F9,$5F,$8F,$F0
	.dc.b	$36,$FE,$73,$6F,$E4,$B7,$0F,$A0,$FC,$EF,$87,$F9,$FF,$0B,$6F,$EB
	.dc.b	$36,$FE,$8B,$6F,$EC,$3F,$B9,$F9,$AF,$DD,$7C,$D5,$BF,$D8,$36,$FF
	.dc.b	$14,$DB,$FC,$4F,$2A,$31,$5F,$89,$6D,$7F,$00,$9D,$4F,$E9,$FE,$7B
	.dc.b	$FD,$AA,$DC,$7F,$0F,$F1,$5F,$E1,$A9,$CD,$FF,$A0,$DF,$F7,$E5,$71
	.dc.b	$9D,$67,$B7,$F1,$FF,$41,$27,$47,$E5,$DF,$C9,$F4,$48,$FC,$D7,$2D
	.dc.b	$C7,$F2,$B3,$F9,$8F,$A0,$C6,$DF,$9D,$71,$7F,$A1,$72,$FE,$69,$F4
	.dc.b	$5F,$A5,$B1,$EF,$6D,$FA,$DB,$5F,$EC,$1C,$D1,$8A,$FE,$13,$18,$7E
	.dc.b	$45,$F5,$7F,$65,$F4,$1F,$AE,$FA,$0F,$D7,$B6,$32,$7E,$DE,$32,$D3
	.dc.b	$38,$FE,$D2,$32,$D8,$FA,$DF,$41,$F4,$F8,$BE,$A3,$63,$27,$DB,$8C
	.dc.b	$B4,$CC,$FD,$98,$CB,$63,$F8,$54,$FF,$06,$9F,$E0,$ED,$8C,$9F,$C3
	.dc.b	$E3,$2D,$33,$3F,$C3,$23,$2D,$8F,$C7,$7C,$67,$E2,$FE,$77,$F1,$8D
	.dc.b	$8C,$9F,$93,$8C,$B4,$CC,$FE,$46,$32,$D8,$FE,$45,$C9,$FC,$7B,$C5
	.dc.b	$FE,$3E,$D8,$C9,$FC,$9E,$32,$D3,$FA,$47,$CF,$FF,$23,$AF,$E8,$6D
	.dc.b	$F1,$0F,$F2,$59,$FE,$96,$E4,$8F,$FA,$01,$D4,$F2,$8C,$76,$67,$AE
	.dc.b	$C7,$81,$B3,$FE,$EF,$AF,$E2,$73,$E4,$AE,$0F,$F5,$DF,$A0,$FF,$67
	.dc.b	$BF,$EA,$BE,$EF,$EB,$B1,$8B,$7C,$17,$F2,$C7,$FA,$EC,$7F,$B6,$EB
	.dc.b	$84,$63,$E1,$91,$8C,$D5,$CE,$F9,$46,$2B,$77,$DB,$0F,$0F,$C3,$62
	.dc.b	$F7,$87,$19,$F2,$DE,$5B,$8F,$79,$F2,$61,$CC,$F2,$47,$8F,$37,$81
	.dc.b	$C6,$7E,$AF,$81,$74,$FA,$B3,$69,$36,$30,$DB,$E8,$F3,$B7,$D0,$7F
	.dc.b	$42,$F0,$FF,$C9,$AD,$8F,$F1,$D9,$AE,$04,$E9,$B1,$AF,$88,$7C,$8F
	.dc.b	$F8,$E2,$30,$D7,$FC,$8B,$1A,$F1,$3F,$CB,$6D,$8E,$F8,$7C,$0F,$88
	.dc.b	$F1,$E9,$FE,$ED,$9C,$B5,$FF,$27,$CE,$DF,$F2,$8C,$BF,$4F,$7C,$6F
	.dc.b	$F9,$B7,$93,$F0,$93,$98,$C7,$1F,$F9,$FE,$31,$7F,$47,$C6,$CB,$FE
	.dc.b	$FD,$C3,$9B,$F8,$A9,$DB,$F6,$5B,$87,$E2,$38,$97,$CA,$BE,$7F,$8D
	.dc.b	$73,$34,$EF,$C8,$FB,$17,$4E,$6F,$F3,$27,$22,$DF,$27,$F3,$53,$C1
	.dc.b	$CD,$F1,$7F,$1C,$71,$3F,$21,$7D,$EF,$DF,$EA,$3F,$1F,$7C,$27,$FA
	.dc.b	$57,$C8,$FF,$3B,$46,$30,$FF,$36,$56,$EB,$5D,$5F,$C1,$BC,$BB,$4A
	.dc.b	$D3,$C2,$F9,$2C,$3A,$E1,$C3,$2F,$E2,$38,$5D,$FD,$D7,$BC,$BF,$D6
	.dc.b	$7F,$AB,$3E,$5F,$BC,$23,$13,$FC,$7B,$6E,$46,$A3,$31,$CC,$FE,$06
	.dc.b	$E3,$3F,$DF,$E5,$FC,$8A,$BD,$74,$FA,$A7,$83,$86,$C7,$FA,$3C,$D4
	.dc.b	$7A,$FF,$62,$79,$1B,$89,$3B,$AE,$2F,$8F,$05,$F0,$8E,$9F,$8A,$BE
	.dc.b	$5F,$FD,$87,$08,$C3,$FE,$C6,$6B,$FB,$A6,$BF,$EB,$67,$39,$FF,$1C
	.dc.b	$8E,$53,$EE,$9C,$EB,$FE,$E3,$EF,$FE,$5B,$F4,$79,$A9,$7E,$47,$EF
	.dc.b	$CE,$A3,$CF,$F1,$58,$F9,$9F,$F4,$0C,$7F,$AC,$E5,$18,$8F,$A4,$7A
	.dc.b	$E7,$C4,$47,$99,$C7,$F4,$57,$BD,$98,$E3,$9A,$B1,$A7,$F3,$8C,$58
	.dc.b	$70,$BF,$97,$87,$97,$D0,$91,$82,$1D,$23,$E2,$1E,$EF,$23,$E4,$61
	.dc.b	$F2,$72,$E9,$5B,$38,$C7,$86,$77,$B7,$FB,$FF,$C3,$FF,$2C,$57,$C3
	.dc.b	$FC,$57,$E1,$5F,$9D,$B5,$BF,$3B,$F2,$37,$C2,$FE,$3B,$6F,$ED,$36
	.dc.b	$BE,$C7,$E2,$DF,$ED,$8A,$FC,$A4,$E5,$1F,$E9,$2B,$EA,$7D,$A9,$E1
	.dc.b	$1D,$BC,$FF,$BB,$8B,$5A,$79,$22,$4F,$66,$BF,$AD,$DD,$7F,$D7,$23
	.dc.b	$44,$3B,$1F,$85,$74,$F1,$FC,$57,$CE,$FE,$25,$AE,$4F,$DA,$FE,$2C
	.dc.b	$F8,$4A,$48,$FD,$A4,$65,$FC,$97,$CD,$3D,$3F,$CC,$B6,$37,$5C,$5F
	.dc.b	$54,$65,$FD,$B7,$98,$8F,$E3,$0D,$8F,$F7,$A3,$5F,$57,$D4,$79,$BE
	.dc.b	$B6,$47,$91,$F0,$FD,$54,$8F,$2B,$5E,$FF,$4D,$FC,$4B,$EA,$9F,$33
	.dc.b	$FC,$09,$B3,$FD,$E0,$38,$7F,$BD,$B5,$F9,$2E,$79,$FF,$81,$37,$9F
	.dc.b	$E7,$B6,$7F,$0F,$DA,$F8,$3F,$16,$12,$7F,$90,$56,$DF,$DE,$CF,$C4
	.dc.b	$E9,$F8,$BC,$12,$7E,$2C,$65,$F8,$F1,$97,$E6,$C7,$CC,$D7,$E6,$FD
	.dc.b	$64,$9F,$9B,$19,$7E,$7C,$65,$FA,$B1,$FC,$E3,$2F,$20,$F9,$BC,$19
	.dc.b	$CE,$57,$F3,$99,$D3,$67,$2F,$F1,$68,$E7,$1C,$A3,$84,$FF,$60,$F3
	.dc.b	$DF,$92,$BE,$03,$0D,$AF,$ED,$2D,$7F,$17,$CD,$A9,$91,$A7,$F1,$B1
	.dc.b	$F8,$EF,$27,$F1,$FE,$DC,$AF,$C7,$C6,$5F,$C8,$C6,$5F,$CB,$C7,$E2
	.dc.b	$7C,$5F,$C4,$12,$3F,$11,$C5,$FC,$A5,$B1,$B7,$EE,$BF,$2F,$E8,$F1
	.dc.b	$1F,$CE,$27,$83,$67,$4F,$F0,$EE,$5F,$ED,$7F,$86,$3A,$3F,$BF,$2F
	.dc.b	$93,$A6,$22,$57,$1E,$8F,$E2,$FE,$62,$F8,$4E,$44,$66,$F8,$C7,$F4
	.dc.b	$CF,$AA,$7C,$07,$F7,$A6,$C7,$F7,$76,$BF,$D4,$BE,$B3,$AB,$C6,$78
	.dc.b	$11,$FE,$FD,$1A,$CF,$FB,$55,$3E,$2B,$FB,$AB,$6B,$F6,$8D,$7F,$CD
	.dc.b	$AD,$7F,$8D,$B9,$9E,$73,$E7,$3F,$C2,$35,$FE,$61,$8C,$D7,$A6,$74
	.dc.b	$F9,$3D,$FF,$F0,$BE,$1F,$EA,$38,$FE,$DA,$DA,$FF,$00,$37,$4F,$E5
	.dc.b	$BF,$80,$7F,$5A,$9C,$F6,$F2,$25,$FF,$87,$67,$55,$FC,$D3,$F1,$6F
	.dc.b	$F2,$84,$FF,$7A,$6D,$FF,$83,$5B,$3F,$CE,$BF,$17,$FE,$0D,$18,$AF
	.dc.b	$EA,$6D,$4D,$8F,$E2,$5D,$5F,$AC,$F8,$8F,$D5,$92,$BF,$97,$FA,$95
	.dc.b	$FF,$1E,$4E,$DF,$D7,$CF,$F3,$69,$91,$C0,$87,$F3,$CE,$4C,$BF,$CF
	.dc.b	$A2,$E5,$FE,$44,$F1,$BF,$35,$3F,$E6,$2E,$0F,$94,$FF,$97,$DB,$8F
	.dc.b	$E0,$F6,$FF,$2D,$7C,$77,$F0,$59,$C9,$23,$35,$CF,$C2,$D1,$1F,$F7
	.dc.b	$BC,$D7,$DA,$A3,$18,$FF,$96,$3D,$CF,$C5,$70,$7F,$7E,$CB,$F2,$95
	.dc.b	$A8,$C5,$F1,$C3,$A9,$E1,$21,$DA,$7F,$CE,$51,$FE,$7C,$E3,$FF,$2D
	.dc.b	$F4,$DF,$67,$F8,$0E,$7F,$EE,$69,$CE,$1F,$B5,$E9,$69,$FE,$1C,$84
	.dc.b	$E9,$FF,$39,$36,$B9,$5F,$4C,$FF,$80,$DF,$1A,$7D,$BB,$EF,$7F,$B0
	.dc.b	$46,$1B,$19,$C7,$FD,$B7,$39,$6C,$65,$F0,$B7,$1F,$53,$70,$F0,$FC
	.dc.b	$87,$C1,$8C,$57,$A9,$B7,$D6,$3F,$CF,$58,$AB,$FA,$27,$DA,$7F,$8B
	.dc.b	$27,$FC,$37,$C0,$8C,$47,$D7,$6B,$FE,$A9,$6C,$6F,$E5,$39,$92,$7E
	.dc.b	$77,$D6,$BE,$FB,$11,$FA,$2B,$FF,$1D,$EA,$D7,$B6,$1F,$EA,$FC,$38
	.dc.b	$5F,$FA,$F2,$F7,$5A,$8C,$56,$F0,$FF,$51,$FC,$77,$E2,$67,$FD,$31
	.dc.b	$F8,$A7,$FB,$99,$BF,$1D,$AE,$69,$1F,$EA,$5A,$E1,$A7,$FA,$45,$B9
	.dc.b	$7F,$3E,$FA,$A7,$8F,$C2,$FF,$D0,$B5,$A7,$FC,$B2,$DE,$65,$70,$9E
	.dc.b	$DE,$47,$FB,$35,$7D,$6B,$EB,$BF,$C9,$78,$7F,$8E,$E3,$13,$D3,$CC
	.dc.b	$FF,$27,$C6,$21,$3F,$EC,$0D,$3F,$CD,$33,$98,$C4,$F6,$BE,$B7,$C9
	.dc.b	$B1,$C1,$F0,$E1,$E5,$BE,$4F,$9F,$9F,$FC,$91,$87,$46,$C6,$7D,$7E
	.dc.b	$39,$7F,$9E,$A7,$FC,$9D,$8F,$F9,$F3,$E4,$71,$5F,$E6,$68,$46,$1F
	.dc.b	$F3,$4B,$5F,$F3,$6B,$5E,$47,$AC,$FF,$38,$FC,$0F,$E2,$5C,$1C,$0A
	.dc.b	$1D,$65,$F2,$63,$9B,$63,$39,$F9,$6B,$E2,$57,$6E,$0A,$FF,$A5,$B8
	.dc.b	$3C,$19,$FF,$B5,$E3,$F0,$B3,$FC,$37,$7F,$B8,$FF,$BC,$63,$27,$FB
	.dc.b	$30,$B3,$C6,$B9,$CF,$E3,$F0,$B5,$FE,$8E,$9A,$F8,$67,$F8,$D7,$D7
	.dc.b	$7F,$B3,$5E,$D3,$C3,$85,$39,$AF,$EA,$3F,$54,$F8,$CE,$BD,$68,$C6
	.dc.b	$3E,$9F,$AC,$CC,$62,$F8,$DE,$0E,$DA,$AB,$17,$FE,$21,$C3,$A5,$62
	.dc.b	$DA,$FD,$7C,$FF,$4B,$86,$63,$ED,$5F,$0C,$3F,$0C,$DB,$FF,$56,$B5
	.dc.b	$FE,$2F,$97,$95,$FF,$59,$F1,$3F,$7E,$7D,$87,$CA,$8C,$3F,$17,$C0
	.dc.b	$7C,$DE,$6B,$79,$7F,$8E,$6F,$6D,$8C,$E1,$FC,$AA,$B8,$61,$2B,$F9
	.dc.b	$DD,$6A,$BC,$6D,$7F,$48,$D7,$FA,$7F,$EE,$8E,$FF,$33,$6B,$FA,$B6
	.dc.b	$5F,$ED,$D7,$C9,$19,$AF,$EA,$AD,$D7,$FA,$F3,$67,$FA,$AF,$C3,$F4
	.dc.b	$C3,$EF,$F9,$0F,$32,$9F,$0F,$C8,$B7,$FE,$09,$8B,$97,$F4,$89,$B9
	.dc.b	$7D,$6B,$E5,$7B,$3C,$F3,$A7,$FB,$DD,$DB,$FF,$31,$E6,$AC,$70,$FF
	.dc.b	$95,$7C,$DF,$29,$F5,$7D,$9B,$8F,$F8,$6E,$2E,$1F,$D6,$A6,$E7,$FE
	.dc.b	$21,$F5,$BF,$41,$E7,$FF,$9E,$70,$FF,$6F,$F8,$F8,$99,$1F,$EE,$27
	.dc.b	$71,$8C,$3F,$C3,$3E,$5F,$C6,$31,$87,$BD,$C5,$FF,$8B,$5C,$BF,$E1
	.dc.b	$87,$F6,$BF,$02,$F9,$AA,$7E,$EB,$C2,$7D,$E3,$E4,$B3,$87,$FA,$1F
	.dc.b	$26,$1C,$63,$0D,$8F,$D4,$C1,$FE,$39,$5C,$E3,$78,$7F,$06,$FA,$67
	.dc.b	$C3,$53,$C5,$4F,$8B,$C5,$F8,$6B,$6D,$AD,$CE,$B0,$F0,$9E,$1A,$7C
	.dc.b	$7F,$73,$F8,$5F,$DD,$7A,$B0,$E9,$19,$69,$FC,$16,$35,$5D,$05,$74
	.dc.b	$9D,$C7,$F4,$78,$C3,$6B,$F9,$83,$E3,$6A,$FA,$37,$CF,$7F,$28,$6D
	.dc.b	$FF,$AD,$77,$7D,$1B,$CE,$D5,$CA,$FA,$97,$C3,$0F,$83,$87,$0B,$D4
	.dc.b	$7C,$4F,$1D,$E0,$7D,$5C,$3F,$C9,$CF,$F5,$16,$CF,$F3,$58,$C3,$C5
	.dc.b	$E0,$DB,$D3,$B7,$D5,$F4,$52,$BE,$57,$A7,$6B,$E5,$64,$AC,$73,$FF
	.dc.b	$2C,$E0,$FF,$3A,$61,$EC,$9E,$17,$F8,$B3,$F1,$97,$C6,$7F,$BB,$47
	.dc.b	$CB,$9F,$9B,$39,$8E,$07,$FB,$7B,$AB,$FC,$C1,$5F,$E6,$D7,$FC,$C6
	.dc.b	$DF,$67,$F6,$BD,$3F,$B6,$6C,$64,$E1,$1F,$C5,$BE,$B2,$47,$3F,$B9
	.dc.b	$B8,$7F,$A3,$A6,$E9,$FE,$A1,$9B,$5F,$A5,$FA,$67,$CA,$3C,$0F,$ED
	.dc.b	$71,$8F,$96,$B3,$9A,$FF,$7C,$4E,$5B,$19,$CB,$F4,$33,$96,$C6,$71
	.dc.b	$FF,$09,$4E,$5B,$19,$AF,$22,$BF,$0E,$1E,$27,$F8,$CF,$A1,$FC,$6B
	.dc.b	$DA,$B7,$F2,$DB,$97,$E3,$F7,$46,$32,$FF,$61,$3F,$C8,$B5,$FE,$47
	.dc.b	$F5,$F7,$0F,$A8,$DC,$39,$36,$FF,$D9,$4F,$E9,$FB,$65,$7E,$9F,$AD
	.dc.b	$F2,$58,$C3,$F1,$F5,$70,$FF,$19,$74,$7E,$B2,$B1,$87,$E3,$79,$7F
	.dc.b	$88,$3F,$8C,$6E,$1F,$E4,$36,$BF,$8C,$6B,$F9,$5F,$B0,$7D,$9F,$F8
	.dc.b	$89,$B5,$FE,$BB,$FD,$F7,$F5,$39,$7F,$90,$ED,$5A,$7F,$87,$7E,$23
	.dc.b	$FE,$30,$BF,$D0,$36,$3F,$96,$4E,$EB,$94,$E7,$B9,$87,$EE,$27,$FC
	.dc.b	$34,$DC,$3F,$9A,$73,$3F,$12,$E5,$1F,$CD,$EF,$8E,$DE,$9F,$0B,$FC
	.dc.b	$77,$93,$EF,$71,$7F,$D0,$E3,$0D,$8D,$9A,$D7,$D6,$FF,$91,$B2,$F0
	.dc.b	$5D,$CB,$F9,$3E,$9E,$AA,$C4,$F8,$3E,$56,$63,$FD,$47,$F1,$1F,$E6
	.dc.b	$77,$F0,$27,$39,$7C,$87,$FC,$A3,$3A,$AF,$93,$1A,$AF,$F3,$96,$4F
	.dc.b	$C4,$7F,$0F,$5E,$BA,$F3,$46,$30,$FA,$7F,$31,$F0,$8E,$BE,$25,$BF
	.dc.b	$85,$16,$7F,$A3,$65,$FC,$69,$E3,$3A,$3F,$B2,$65,$E2,$C1,$F0,$DC
	.dc.b	$63,$0F,$F9,$AB,$F1,$17,$AF,$4F,$95,$AB,$E6,$F1,$8F,$F7,$5A,$DB
	.dc.b	$6B,$73,$AC,$BF,$AD,$56,$9F,$EC,$F6,$F8,$7F,$81,$97,$E5,$6B,$4F
	.dc.b	$C8,$8C,$E3,$FC,$CE,$7F,$97,$4E,$FC,$8F,$2F,$CC,$FA,$59,$5E,$57
	.dc.b	$2D,$26,$3E,$2C,$3E,$1C,$62,$FC,$1A,$7F,$A3,$3F,$11,$7C,$0D,$3F
	.dc.b	$19,$3F,$E8,$93,$F5,$91,$AA,$FF,$0C,$37,$0F,$EF,$4D,$9F,$C1,$9C
	.dc.b	$67,$FA,$74,$65,$B5,$FD,$E2,$39,$57,$F4,$C6,$BA,$9F,$EA,$7F,$4D
	.dc.b	$EA,$8F,$F0,$43,$6F,$FD,$19,$FB,$17,$EE,$E5,$39,$AF,$EB,$CD,$8F
	.dc.b	$F1,$53,$7A,$D5,$C6,$31,$EE,$FF,$37,$BF,$C6,$9F,$DF,$7F,$34,$FF
	.dc.b	$1A,$63,$FE,$9B,$9D,$63,$FE,$26,$A5,$69,$FE,$05,$3A,$AF,$F6,$C4
	.dc.b	$6B,$D7,$FF,$4A,$79,$7F,$EC,$08,$C1,$FE,$AB,$FD,$57,$F8,$DB,$7C
	.dc.b	$2F,$F4,$D7,$8D,$FE,$BA,$9C,$E7,$FE,$36,$F7,$91,$8C,$BF,$D5,$AD
	.dc.b	$C3,$C7,$7C,$70,$FF,$66,$FC,$57,$2F,$8C,$7C,$A7,$FB,$46,$7F,$DD
	.dc.b	$D1,$8C,$DC,$4F,$AF,$7C,$B6,$B0,$E1,$9F,$FC,$01,$6D,$FF,$C0,$06
	.dc.b	$E1,$CF,$0E,$57,$C6,$BF,$DD,$D8,$D5,$97,$97,$0E,$27,$F2,$DC,$BC
	.dc.b	$39,$77,$AE,$2D,$E6,$7E,$A8,$F4,$3B,$46,$1F,$06,$8E,$47,$DA,$6A
	.dc.b	$F8,$61,$CD,$B8,$FA,$CF,$D4,$B8,$DF,$D9,$7C,$AB,$86,$3E,$BF,$B2
	.dc.b	$7C,$F3,$CD,$FF,$17,$C6,$1B,$19,$CB,$FC,$5E,$27,$4E,$3E,$55,$E0
	.dc.b	$FF,$88,$23,$0D,$7F,$C6,$F1,$AF,$0F,$FC,$7A,$D8,$EB,$5F,$E6,$8E
	.dc.b	$0F,$1F,$C0,$7F,$04,$9C,$B5,$FF,$20,$CE,$DF,$F2,$1C,$BD,$7F,$F0
	.dc.b	$AC,$7D,$16,$BF,$1E,$35,$EB,$FF,$94,$78,$3E,$3F,$D9,$DE,$6F,$F4
	.dc.b	$F5,$6A,$F9,$5E,$5F,$E2,$F8,$74,$C3,$8C,$7E,$04,$F8,$DF,$0D,$FE
	.dc.b	$5D,$E2,$F3,$B6,$3F,$CC,$19,$FF,$AC,$DF,$F3,$89,$FE,$AD,$70,$FF
	.dc.b	$82,$CE,$67,$57,$35,$FA,$38,$C9,$F8,$3E,$6F,$D2,$7D,$9F,$9E,$33
	.dc.b	$3B,$AE,$3A,$78,$B0,$F6,$61,$C2,$FF,$35,$DE,$FB,$37,$D9,$FF,$9D
	.dc.b	$FC,$16,$BF,$EC,$3A,$D4,$E6,$31,$5C,$2F,$F8,$C9,$FE,$F2,$FB,$3D
	.dc.b	$2B,$E2,$E9,$AF,$F0,$B9,$E4,$FF,$A7,$FA,$95,$CB,$2F,$D4,$67,$D3
	.dc.b	$17,$D9,$BD,$FC,$37,$1F,$1D,$F0,$6E,$3F,$43,$5F,$E4,$B5,$B9,$FE
	.dc.b	$58,$D8,$FE,$61,$8F,$AA,$7F,$9F,$F9,$9F,$E9,$86,$9F,$D2,$1C,$79
	.dc.b	$3F,$DC,$E3,$07,$FB,$C5,$C3,$FD,$32,$3D,$9F,$66,$F9,$0E,$DF,$62
	.dc.b	$FB,$3F,$93,$DC,$FA,$8F,$E2,$B8,$FC,$E8,$C7,$8F,$C2,$A4,$FF,$AB
	.dc.b	$2F,$93,$63,$D1,$18,$8F,$DF,$B7,$5F,$F2,$E3,$78,$BF,$D7,$0D,$9F
	.dc.b	$F4,$FF,$4B,$E5,$3C,$FF,$49,$E6,$FA,$DF,$F6,$25,$79,$5A,$FF,$BC
	.dc.b	$6D,$3C,$23,$CC,$F9,$08,$E0,$FC,$58,$8E,$2E,$1A,$FF,$88,$BE,$CD
	.dc.b	$EF,$FC,$A3,$C9,$F1,$F2,$78,$46,$5F,$8C,$FC,$42,$38,$36,$43,$FC
	.dc.b	$5F,$E4,$FF,$A0,$75,$FF,$41,$77,$FF,$AB,$E7,$EA,$7D,$9F,$C0,$9F
	.dc.b	$F2,$BB,$77,$FF,$71,$56,$E9,$5C,$BD,$3F,$C6,$4E,$67,$FD,$C2,$D7
	.dc.b	$8C,$68,$FC,$A4,$73,$9E,$53,$F8,$5F,$B3,$7C,$D6,$BD,$4F,$C5,$E4
	.dc.b	$D6,$99,$BF,$9F,$8A,$BC,$64,$8F,$F7,$BC,$65,$B1,$C5,$F5,$E7,$EE
	.dc.b	$8F,$CC,$E2,$F2,$FF,$37,$F6,$7F,$EF,$CD,$3F,$CF,$6D,$BF,$F4,$EB
	.dc.b	$70,$0F,$B3,$FF,$62,$BF,$84,$6B,$FD,$61,$AF,$03,$8B,$87,$0D,$7F
	.dc.b	$DF,$3F,$2B,$FE,$7A,$7A,$7C,$6F,$EF,$E7,$AE,$1F,$E9,$0A,$D5,$7F
	.dc.b	$B9,$5F,$E2,$9E,$61,$A7,$BB,$E5,$F9,$7C,$3F,$F0,$7C,$0D,$25,$FF
	.dc.b	$70,$EE,$E5,$F3,$9F,$C0,$30,$34,$EF,$F3,$5F,$C4,$FE,$C5,$F1,$7F
	.dc.b	$C5,$7B,$9F,$47,$4D,$44,$CB,$FB,$E8,$AE,$2E,$1A,$FF,$80,$9A,$FC
	.dc.b	$B3,$84,$73,$7E,$F9,$F5,$A3,$8E,$FF,$EE,$5B,$7D,$9F,$C6,$E7,$F8
	.dc.b	$FC,$53,$0F,$E6,$BC,$5F,$D7,$70,$FE,$9F,$1F,$E4,$08,$C7,$27,$FB
	.dc.b	$75,$FA,$6D,$AE,$93,$CB,$87,$E9,$FC,$4C,$8C,$10,$F3,$F9,$DF,$E0
	.dc.b	$98,$D4,$63,$0F,$AA,$FF,$10,$D3,$D7,$8F,$E2,$D7,$C6,$FC,$F3,$E8
	.dc.b	$6F,$2B,$84,$65,$E5,$19,$7E,$38,$9E,$6F,$03,$67,$F6,$5F,$B3,$7C
	.dc.b	$BF,$F6,$1F,$93,$FE,$83,$3E,$A6,$EF,$F2,$3D,$65,$F0,$AE,$57,$FD
	.dc.b	$CA,$72,$D9,$D4,$FC,$1F,$8B,$FC,$DF,$A1,$F9,$CF,$7E,$4F,$D6,$8C
	.dc.b	$BF,$62,$32,$FC,$98,$FC,$A7,$95,$F9,$3E,$E9,$3F,$CE,$BA,$3F,$9B
	.dc.b	$E9,$F7,$A3,$F2,$55,$FD,$AE,$7F,$9D,$C7,$E3,$DA,$FE,$ED,$AF,$E3
	.dc.b	$E3,$BB,$D5,$FD,$17,$9D,$FE,$07,$FB,$3F,$D4,$CF,$F8,$09,$7D,$0E
	.dc.b	$4D,$77,$49,$D4,$65,$E1,$1F,$E0,$1F,$B3,$FF,$35,$C7,$F5,$4F,$49
	.dc.b	$F3,$DF,$E0,$56,$C7,$F5,$89,$DB,$F8,$06,$E9,$F0,$5B,$D1,$FC,$5D
	.dc.b	$B3,$FC,$82,$BC,$EF,$3F,$92,$ED,$E1,$48,$C1,$0F,$F3,$CD,$27,$FC
	.dc.b	$2C,$D8,$FE,$02,$D7,$71,$ED,$23,$ED,$F8,$2F,$F1,$07,$C2,$FE,$CA
	.dc.b	$3F,$99,$FD,$9B,$E2,$FF,$C4,$8D,$8F,$EE,$F3,$B7,$FD,$23,$E2,$72
	.dc.b	$23,$FD,$F2,$7F,$BB,$BF,$DF,$BE,$CD,$F0,$BF,$E2,$FF,$B3,$7E,$01
	.dc.b	$4F,$F1,$94,$EA,$B8,$56,$08,$FF,$82,$2B,$F7,$2F,$F9,$47,$B5,$F6
	.dc.b	$9F,$C8,$BC,$0F,$D3,$B6,$33,$3F,$EA,$29,$CB,$63,$2F,$FB,$85,$BC
	.dc.b	$9F,$B1,$6F,$27,$F1,$7F,$6F,$FA,$29,$2B,$F8,$FB,$FD,$23,$9F,$11
	.dc.b	$FC,$46,$2D,$FF,$22,$F2,$3F,$D3,$D5,$FB,$9E,$4F,$F4,$15,$FF,$5D
	.dc.b	$CB,$EF,$70,$FA,$E7,$33,$FE,$77,$6B,$FE,$71,$7E,$D4,$F0,$8F,$F3
	.dc.b	$94,$7D,$DB,$FF,$6A,$7C,$87,$7F,$2B,$A1,$2B,$E2,$57,$F2,$67,$FA
	.dc.b	$25,$FE,$3E,$78,$47,$47,$FC,$F1,$F0,$9F,$8F,$CF,$F8,$B5,$F1,$6C
	.dc.b	$F0,$AF,$F4,$04,$D9,$CC,$73,$3F,$CF,$31,$FE,$8C,$E7,$FF,$34,$77
	.dc.b	$FF,$5C,$6C,$FF,$B6,$9A,$FF,$B2,$3E,$03,$F8,$4C,$62,$FF,$A7,$B7
	.dc.b	$83,$FD,$87,$3F,$C5,$9F,$F0,$8E,$BF,$EF,$7F,$96,$F0,$7D,$1F,$EF
	.dc.b	$B1,$FF,$4B,$72,$7B,$8C,$F7,$C8,$F1,$4E,$DF,$F0,$A6,$48,$C4,$7F
	.dc.b	$B4,$9A,$A3,$53,$CE,$3F,$B7,$C6,$1F,$F4,$DF,$DA,$7F,$3D,$FB,$3F
	.dc.b	$F1,$FC,$7F,$A8,$3E,$D3,$F8,$4B,$FD,$81,$BB,$7F,$47,$9D,$CF,$F7
	.dc.b	$AE,$34,$62,$3F,$1C,$D7,$FD,$84,$D7,$FD,$78,$D8,$E3,$EF,$78,$C9
	.dc.b	$3E,$3F,$AA,$AF,$7A,$F0,$DF,$8E,$BF,$D7,$D8,$2F,$FD,$61,$F2,$5F
	.dc.b	$87,$E1,$FA,$7F,$69,$FE,$FE,$C3,$F4,$CD,$8F,$F2,$04,$EF,$2F,$F2
	.dc.b	$4B,$F9,$77,$FC,$34,$D7,$7C,$DF,$EB,$B8,$C4,$70,$8F,$09,$1F,$F6
	.dc.b	$5F,$D8,$FF,$70,$AE,$FE,$6F,$F9,$EA,$7F,$C8,$5F,$67,$E2,$F4,$3F
	.dc.b	$CA,$5E,$DA,$31,$3F,$E5,$C8,$C3,$E6,$AE,$51,$88,$64,$9C,$D7,$29
	.dc.b	$EE,$7B,$AF,$BC,$79,$34,$EC,$D9,$FC,$66,$17,$8F,$FC,$B1,$1E,$1A
	.dc.b	$FF,$37,$C6,$23,$C3,$87,$B6,$31,$1F,$E7,$4F,$89,$DD,$FF,$91,$61
	.dc.b	$18,$7F,$C9,$0D,$7F,$CE,$8D,$79,$1E,$A3,$FC,$FF,$F0,$DF,$E4,$29
	.dc.b	$F8,$51,$A6,$C6,$71,$F0,$57,$BA,$BF,$D0,$13,$97,$C2,$75,$0E,$5F
	.dc.b	$F4,$6F,$A1,$FE,$50,$CB,$FC,$29,$08,$C6,$BF,$C0,$B6,$FB,$4F,$FB
	.dc.b	$52,$74,$D9,$E3,$1F,$E9,$16,$BD,$A3,$FC,$E2,$F0,$3F,$CE,$71,$D9
	.dc.b	$75,$BE,$CF,$FD,$CF,$5D,$4F,$F5,$1E,$1F,$D1,$63,$35,$FB,$9F,$B3
	.dc.b	$7C,$8F,$3E,$94,$63,$0F,$CC,$FC,$06,$63,$17,$C6,$F0,$78,$F2,$69
	.dc.b	$FE,$8C,$C3,$C2,$7F,$AD,$31,$EB,$5C,$23,$FB,$46,$1D,$30,$FE,$AC
	.dc.b	$D7,$83,$55,$72,$9F,$D6,$F8,$3C,$E7,$FC,$EF,$87,$BB,$4F,$E6,$FF
	.dc.b	$62,$FB,$2F,$91,$C8,$8C,$63,$E6,$C3,$E5,$F8,$AF,$B3,$DD,$6A,$7F
	.dc.b	$95,$B5,$FE,$6A,$D7,$F4,$6D,$7F,$9A,$36,$FC,$18,$7C,$7F,$E0,$0F
	.dc.b	$E7,$18,$F8,$67,$2F,$8F,$81,$3A,$C3,$F9,$DF,$DA,$7F,$BD,$AB,$83
	.dc.b	$63,$FD,$2F,$9F,$FB,$DE,$73,$E0,$7C,$0A,$D5,$7F,$4C,$6B,$FB,$A6
	.dc.b	$BF,$D2,$DB,$C1,$FE,$45,$F3,$3F,$0D,$E4,$9F,$C0,$62,$F2,$73,$13
	.dc.b	$2F,$F0,$3B,$A8,$C6,$3F,$D2,$6F,$95,$68,$F8,$17,$CD,$C4,$EE,$FF
	.dc.b	$4F,$87,$8E,$7F,$C1,$B1,$8B,$E7,$5F,$E0,$E7,$FD,$C7,$3F,$E1,$8F
	.dc.b	$33,$CF,$EE,$FA,$99,$1F,$E1,$67,$91,$1F,$E2,$DB,$3E,$0B,$FD,$07
	.dc.b	$1F,$ED,$2F,$6B,$F4,$11,$FD,$46,$30,$4C,$7F,$D8,$D3,$AF,$1B,$6F
	.dc.b	$CA,$F2,$E5,$6F,$9F,$FC,$3F,$18,$BF,$53,$ED,$24,$FF,$BD,$3D,$97
	.dc.b	$D1,$7F,$03,$A4,$62,$7F,$39,$E0,$BE,$D1,$F2,$FB,$AD,$5F,$FB,$39
	.dc.b	$B9,$7F,$BB,$2F,$87,$A5,$BC,$BF,$88,$46,$32,$FD,$46,$0F,$B4,$7C
	.dc.b	$0F,$8D,$B3,$E2,$D1,$F1,$F9,$A9,$39,$FA,$2B,$83,$4F,$16,$09,$E9
	.dc.b	$F2,$92,$FF,$71,$D2,$FB,$4F,$83,$1E,$4D,$15,$BC,$BF,$A8,$46,$27
	.dc.b	$FA,$94,$7A,$67,$F9,$8C,$E6,$37,$1F,$CB,$F2,$7D,$D7,$AA,$FB,$4E
	.dc.b	$63,$15,$FE,$F8,$FB,$17,$C9,$FD,$56,$DF,$D4,$FB,$27,$A3,$9A,$93
	.dc.b	$9E,$5B,$87,$C3,$8F,$EB,$F7,$EE,$8C,$1F,$D0,$BE,$58,$FF,$33,$7F
	.dc.b	$AC,$CF,$F0,$B7,$2D,$E4,$D4,$ED,$B1,$93,$EE,$E4,$F7,$7D,$47,$F4
	.dc.b	$7E,$07,$AF,$8A,$DC,$FF,$5A,$AE,$11,$F3,$9B,$87,$F8,$05,$AF,$A0
	.dc.b	$F4,$BA,$E0,$FC,$96,$6F,$7B,$B5,$7C,$0C,$D3,$9C,$3F,$94,$77,$7F
	.dc.b	$BB,$9B,$D9,$FB,$6F,$77,$F6,$ED,$8C,$9C,$23,$2F,$FA,$7B,$E9,$7F
	.dc.b	$8F,$9F,$BE,$FB,$5B,$19,$3F,$9B,$7C,$73,$EC,$33,$A7,$D4,$F1,$13
	.dc.b	$9A,$FE,$9D,$F6,$AF,$97,$79,$1F,$DE,$E3,$1F,$39,$67,$35,$C1,$46
	.dc.b	$6B,$4D,$9D,$37,$1F,$C2,$DF,$49,$E5,$5C,$1F,$8F,$EC,$7F,$1A,$FA
	.dc.b	$7F,$E3,$6D,$8C,$9F,$32,$32,$D3,$E7,$ED,$FC,$8B,$ED,$7F,$91,$B6
	.dc.b	$32,$7D,$48,$CB,$4F,$C6,$F8,$DF,$A5,$3F,$86,$F2,$3E,$67,$13,$A3
	.dc.b	$F1,$53,$F8,$D2,$38,$3F,$BD,$0F,$B1,$7C,$C7,$2D,$BF,$4F,$39,$8F
	.dc.b	$E5,$1F,$24,$FB,$5A,$7F,$86,$F8,$FF,$4F,$7F,$AE,$13,$BB,$FE,$43
	.dc.b	$F6,$BF,$E3,$59,$FC,$D0,$70,$FE,$75,$AF,$ED,$9B,$CB,$F8,$6E,$BF
	.dc.b	$01,$FE,$3D,$E6,$7D,$AF,$7B,$FC,$66,$5F,$BC,$7C,$99,$7F,$17,$EE
	.dc.b	$67,$DF,$A7,$DB,$E4,$7F,$45,$D7,$F2,$58,$7A,$DA,$FA,$1A,$F1,$AF
	.dc.b	$E4,$B1,$88,$FC,$C6,$5D,$E7,$F8,$7E,$7E,$6B,$ED,$87,$A6,$B5,$7E
	.dc.b	$FD,$FC,$47,$F5,$7F,$B5,$FE,$35,$9F,$A2,$75,$97,$CC,$AF,$F3,$1E
	.dc.b	$7F,$6A,$3C,$71,$82,$3F,$E3,$9C,$FF,$01,$C7,$DA,$31,$C5,$F8,$B7
	.dc.b	$EB,$FC,$EF,$C3,$3C,$5E,$25,$8C,$9F,$02,$3C,$24,$70,$74,$F8,$AE
	.dc.b	$19,$FF,$32,$C7,$F2,$11,$FD,$CF,$ED,$5E,$C5,$3E,$E4,$BE,$7F,$18
	.dc.b	$7F,$6A,$AD,$BF,$D5,$1B,$87,$FA,$9D,$AF,$E3,$9A,$FF,$29,$77,$BF
	.dc.b	$CB,$BF,$F6,$16,$BD,$AB,$CD,$F1,$9F,$0D,$FC,$E5,$FF,$2F,$C3,$97
	.dc.b	$93,$E6,$F9,$EF,$53,$23,$C8,$FF,$A5,$A7,$F9,$CD,$70,$9C,$9F,$4E
	.dc.b	$DA,$F1,$9F,$0B,$FE,$17,$FB,$5F,$F2,$E6,$9F,$81,$94,$62,$FF,$1B
	.dc.b	$3F,$E7,$B7,$FA,$34,$66,$7F,$AF,$BF,$E6,$57,$F6,$2D,$77,$F0,$7F
	.dc.b	$EA,$C8,$C4,$70,$8F,$D7,$7A,$4F,$AE,$F2,$46,$1F,$F1,$67,$D5,$FA
	.dc.b	$B0,$FF,$0F,$B6,$FF,$9F,$7D,$AF,$FA,$D7,$2F,$54,$E5,$B9,$FF,$AF
	.dc.b	$23,$85,$7F,$06,$6B,$FD,$85,$AF,$FA,$4B,$CA,$FF,$5F,$60,$FB,$5F
	.dc.b	$F1,$AB,$FD,$FB,$C2,$7C,$1F,$F4,$48,$FE,$43,$3A,$BE,$51,$8F,$77
	.dc.b	$FC,$EF,$87,$4A,$D7,$B4,$FB,$5F,$E3,$B7,$FE,$2C,$6E,$1F,$E8,$CF
	.dc.b	$23,$FD,$8D,$18,$DB,$F7,$17,$C7,$C5,$C9,$FE,$84,$AB,$87,$F9,$1B
	.dc.b	$E4,$79,$FC,$97,$F7,$BC,$BF,$DD,$77,$FC,$06,$31,$BB,$ED,$5F,$05
	.dc.b	$9F,$63,$FD,$CD,$77,$0E,$39,$7F,$88,$31,$E9,$1F,$EF,$39,$D6,$3C
	.dc.b	$F0,$E1,$39,$BE,$18,$74,$7F,$9F,$C7,$43,$FD,$FB,$97,$C2,$8F,$E7
	.dc.b	$AD,$8F,$F1,$7F,$3B,$ED,$5E,$8F,$96,$30,$F8,$B9,$7F,$CB,$58,$78
	.dc.b	$2D,$5A,$8C,$36,$1A,$FC,$78,$C3,$F1,$AB,$9C,$EB,$2F,$1E,$3E,$2C
	.dc.b	$3B,$4B,$ED,$1F,$46,$F4,$3F,$CC,$31,$86,$C6,$67,$FA,$D9,$C5,$C3
	.dc.b	$FD,$7B,$E6,$1E,$1F,$F8,$6E,$37,$18,$F1,$7F,$CD,$AD,$8C,$CE,$DF
	.dc.b	$EC,$F8,$7F,$99,$B6,$F1,$FC,$0F,$C4,$9F,$0C,$E5,$B1,$97,$FC,$91
	.dc.b	$3F,$EB,$5F,$5B,$FC,$95,$A6,$3D,$D4,$62,$BE,$BC,$61,$FB,$F5,$C5
	.dc.b	$B1,$97,$E6,$1F,$59,$C7,$0F,$D3,$6F,$07,$EC,$F8,$5F,$6B,$FE,$5B
	.dc.b	$AF,$C9,$E9,$FE,$B0,$C3,$F0,$1C,$DF,$EF,$CF,$83,$D6,$3F,$BB,$F9
	.dc.b	$0F,$F4,$1D,$7D,$8B,$E0,$F8,$8F,$B5,$F1,$3E,$B7,$FC,$FF,$1A,$DF
	.dc.b	$E5,$CE,$67,$67,$E7,$0A,$73,$9E,$B9,$79,$A7,$C7,$2C,$BF,$65,$F4
	.dc.b	$38,$EE,$7D,$AF,$F2,$7A,$FF,$58,$B5,$FF,$49,$C6,$BC,$8F,$F3,$E7
	.dc.b	$83,$FD,$F7,$0F,$E2,$FF,$6B,$8D,$7F,$D2,$F7,$BC,$39,$E1,$95,$19
	.dc.b	$7F,$CD,$9C,$4B,$D6,$1F,$C8,$B4,$E1,$8B,$ED,$5F,$01,$86,$E3,$FE
	.dc.b	$A6,$D3,$C9,$97,$CE,$6E,$1B,$B9,$5B,$C3,$EF,$4F,$96,$1E,$4B,$CE
	.dc.b	$F1,$FB,$BF,$EE,$29,$FF,$42,$FC,$6E,$EB,$51,$FD,$2A,$BF,$DA,$2F
	.dc.b	$26,$C6,$A3,$F6,$2D,$7F,$C9,$BF,$66,$FB,$4F,$8D,$E0,$BE,$AD,$E3
	.dc.b	$E1,$B7,$FE,$F6,$BE,$0D,$BD,$54,$8C,$10,$FF,$30,$4F,$06,$CF,$83
	.dc.b	$A1,$F2,$9F,$0E,$9F,$39,$E4,$FA,$EF,$F6,$46,$1F,$E3,$56,$BD,$23
	.dc.b	$81,$BF,$62,$47,$9E,$32,$D2,$1F,$E2,$FF,$B4,$FF,$7E,$BE,$DF,$85
	.dc.b	$F0,$7C,$84,$8C,$10,$FF,$1B,$60,$8D,$47,$BA,$7D,$44,$78,$73,$FF
	.dc.b	$8A,$FC,$CF,$F0,$B3,$FE,$17,$F0,$FF,$D6,$73,$FE,$51,$FB,$4F,$6C
	.dc.b	$FF,$8C,$5B,$D3,$FE,$DD,$AD,$D7,$FA,$2D,$B5,$BD,$7E,$DD,$F0,$AF
	.dc.b	$F7,$DD,$6A,$F9,$7A,$DF,$91,$8C,$4E,$A3,$F0,$7F,$68,$F9,$CD,$7A
	.dc.b	$DD,$36,$97,$F9,$B3,$AC,$69,$F5,$46,$9F,$34,$65,$A7,$13,$FD,$E8
	.dc.b	$F8,$3C,$EF,$C9,$FD,$A7,$FA,$8B,$6F,$F2,$93,$74,$FF,$43,$B6,$C3
	.dc.b	$ED,$3F,$D8,$0F,$56,$BF,$E6,$D8,$E1,$D6,$79,$47,$1F,$81,$FD,$1F
	.dc.b	$9B,$FE,$50,$8E,$51,$FB,$7B,$E7,$8F,$F9,$96,$B5,$5F,$EF,$7F,$3F
	.dc.b	$F9,$8B,$63,$F8,$6C,$62,$7D,$7E,$0F,$3F,$85,$F7,$B2,$FD,$11,$E2
	.dc.b	$F9,$C9,$1F,$E9,$5C,$11,$D3,$E5,$FF,$85,$7D,$9F,$FB,$5A,$BE,$67
	.dc.b	$5B,$E9,$7F,$87,$36,$3E,$84,$ED,$FF,$6A,$37,$3F,$F3,$54,$EE,$7F
	.dc.b	$DB,$0D,$8F,$F3,$74,$EF,$87,$E8,$D2,$3A,$3F,$ED,$A6,$BA,$6B,$F6
	.dc.b	$1B,$E6,$36,$FF,$18,$DF,$F3,$F7,$E7,$DF,$F9,$5D,$B7,$FD,$35,$DC
	.dc.b	$67,$A7,$E7,$F0,$CA,$FC,$B6,$1D,$67,$E9,$EF,$F5,$3E,$1E,$56,$B1
	.dc.b	$4E,$A3,$13,$A7,$EF,$47,$CC,$F8,$BF,$9B,$F1,$52,$7E,$8C,$65,$FA
	.dc.b	$71,$97,$E5,$C6,$5F,$9B,$19,$D3,$E8,$CF,$F5,$9F,$45,$F5,$AF,$94
	.dc.b	$FE,$8B,$E5,$4C,$7F,$65,$5A,$9E,$15,$B9,$FF,$18,$CE,$5B,$19,$7E
	.dc.b	$C9,$FD,$99,$C3,$F7,$36,$FA,$9E,$6F,$E6,$BD,$C9,$3F,$1E,$3F,$3A
	.dc.b	$61,$FD,$14,$61,$FC,$7F,$91,$F9,$2E,$A9,$18,$21,$F8,$E7,$2F,$E5
	.dc.b	$23,$F3,$A7,$F5,$EE,$5F,$E8,$B1,$FE,$4C,$3F,$21,$19,$31,$83,$EC
	.dc.b	$FF,$4F,$3F,$82,$5F,$3F,$FE,$00,$6C,$67,$AF,$06,$66,$46,$4F,$E3
	.dc.b	$1F,$67,$FE,$62,$7F,$8D,$7A,$0F,$9A,$FE,$F8,$D8,$DF,$C1,$7F,$7F
	.dc.b	$AE,$11,$FE,$64,$6B,$F4,$9A,$FE,$B1,$AF,$F1,$E5,$F2,$1D,$BC,$09
	.dc.b	$3E,$48,$EC,$61,$73,$FF,$B9,$70,$56,$A7,$81,$F2,$EF,$AB,$6F,$FC
	.dc.b	$1B,$18,$8F,$F3,$DE,$1F,$B0,$7F,$98,$FD,$9B,$E2,$3F,$C3,$8D,$8F
	.dc.b	$E7,$D3,$B7,$FD,$13,$DD,$FE,$18,$8F,$F5,$7C,$7E,$49,$AF,$F0,$C6
	.dc.b	$BF,$CF,$BE,$C9,$F0,$1F,$E2,$9F,$B2,$7D,$9D,$3F,$C5,$93,$AA,$FE
	.dc.b	$9C,$D7,$FD,$26,$D7,$F4,$2D,$7F,$8D,$B5,$FD,$27,$D9,$FF,$8D,$3E
	.dc.b	$BE,$47,$EC,$3B,$13,$FB,$03,$F2,$9D,$BF,$B8,$6C,$6A,$7F,$63,$1F
	.dc.b	$96,$F1,$3F,$C4,$31,$FC,$47,$D9,$FE,$8E,$7F,$48,$E8,$B1,$92,$1F
	.dc.b	$D3,$DF,$E0,$DC,$7F,$4F,$A9,$F0,$DF,$7A,$BF,$90,$B5,$FF,$36,$CF
	.dc.b	$D9,$CF,$F1,$7C,$5E,$88,$F9,$F1,$89,$FE,$13,$18,$3F,$92,$BD,$27
	.dc.b	$FC,$D3,$5C,$E7,$33,$DB,$C5,$E8,$47,$9B,$5D,$1F,$C6,$A3,$8F,$4F
	.dc.b	$F9,$B1,$F6,$CC,$9D,$3F,$E7,$8F,$81,$FC,$A5,$3C,$9F,$F7,$5F,$D4
	.dc.b	$7F,$06,$BF,$D0,$C7,$FA,$97,$CA,$F3,$9F,$E8,$5F,$89,$56,$EF,$95
	.dc.b	$FF,$BB,$23,$93,$FE,$C6,$6B,$DB,$C2,$FE,$25,$5C,$9B,$F4,$5F,$C0
	.dc.b	$67,$FD,$1F,$C9,$E7,$3C,$11,$9C,$3B,$DA,$31,$97,$FB,$0E,$B8,$5F
	.dc.b	$F8,$BE,$B5,$E0,$F2,$3B,$92,$3F,$DD,$F1,$A7,$FA,$0F,$D9,$7F,$8E
	.dc.b	$DF,$EC,$7F,$65,$FC,$81,$F8,$A7,$FA,$DD,$C6,$5F,$0B,$8D,$5C,$6B
	.dc.b	$FD,$6A,$D9,$FF,$5A,$57,$0C,$BF,$C6,$7C,$88,$C5,$FF,$AC,$AF,$85
	.dc.b	$7F,$AE,$1B,$19,$F6,$3F,$B0,$46,$2F,$97,$D5,$62,$65,$7F,$60,$AD
	.dc.b	$60,$AF,$F5,$67,$C7,$FE,$2B,$1F,$1F,$D9,$7F,$BC,$F3,$FF,$65,$36
	.dc.b	$3F,$C3,$33,$BC,$7F,$C9,$6D,$CD,$5B,$9F,$F6,$E3,$54,$F1,$6F,$2F
	.dc.b	$88,$8F,$F8,$4B,$EC,$3F,$D5,$6F,$D0,$38,$F9,$DF,$CD,$97,$DB,$3C
	.dc.b	$0F,$2F,$CD,$7F,$8B,$A3,$13,$E6,$8C,$4F,$AA,$B9,$56,$AB,$AE,$BF
	.dc.b	$E6,$68,$C6,$5F,$E5,$86,$E1,$CA,$BE,$0D,$F2,$C3,$A6,$1F,$E6,$EE
	.dc.b	$3F,$F2,$EE,$3F,$ED,$78,$C7,$B1,$AC,$15,$AC,$3D,$8F,$F9,$92,$3F
	.dc.b	$CE,$FF,$0F,$C1,$FF,$08,$B5,$FF,$25,$B5,$D9,$D2,$BF,$C2,$EA,$3D
	.dc.b	$98,$FF,$86,$BE,$13,$F7,$3A,$7C,$1A,$E2,$D8,$CC,$F2,$AE,$35,$FE
	.dc.b	$5A,$9C,$BE,$43,$E9,$CF,$5E,$1D,$C7,$FA,$43,$D5,$FE,$0F,$A7,$F0
	.dc.b	$88,$C4,$FD,$FA,$CB,$FE,$2E,$AB,$3F,$E6,$86,$F1,$3E,$CF,$FD,$C3
	.dc.b	$5F,$EA,$5A,$FF,$79,$BF,$E5,$A8,$CC,$7E,$24,$C9,$FB,$A9,$E5,$D9
	.dc.b	$DA,$3C,$66,$67,$8E,$3F,$BD,$FB,$37,$C7,$79,$3A,$D1,$8C,$BF,$31
	.dc.b	$EE,$E6,$31,$8F,$F0,$8A,$DD,$E4,$F5,$F3,$7F,$A4,$70,$FF,$2A,$DF
	.dc.b	$0C,$7C,$15,$AC,$F8,$57,$E9,$5B,$8F,$AF,$0F,$3C,$FF,$AF,$B1,$EE
	.dc.b	$EC,$FE,$49,$F6,$4F,$B2,$FA,$57,$F9,$78,$C4,$FF,$24,$6C,$7F,$98
	.dc.b	$39,$51,$88,$FF,$4D,$C6,$2B,$EA,$F8,$DF,$9C,$BE,$F1,$89,$FB,$66
	.dc.b	$E3,$FB,$7B,$8B,$EF,$A7,$F4,$CE,$0F,$EE,$B3,$C3,$EC,$BF,$DD,$D7
	.dc.b	$FC,$E1,$B9,$79,$5A,$F9,$1A,$FE,$CD,$DD,$F5,$CB,$FD,$33,$5F,$D1
	.dc.b	$E7,$2F,$F3,$48,$CD,$7F,$80,$BC,$AF,$C0,$79,$07,$F2,$7E,$47,$F6
	.dc.b	$4B,$FE,$E9,$1A,$23,$FE,$AC,$8D,$57,$38,$D9,$FE,$08,$9F,$A7,$0D
	.dc.b	$BF,$CB,$31,$86,$CE,$CF,$F0,$BF,$BF,$DA,$BF,$39,$8A,$3C,$FE,$5F
	.dc.b	$1C,$BF,$DC,$77,$FE,$16,$8C,$11,$FF,$10,$F7,$F0,$69,$FE,$BF,$F6
	.dc.b	$3F,$63,$E6,$FF,$A1,$23,$FD,$85,$E2,$62,$64,$6C,$C9,$73,$FF,$15
	.dc.b	$F9,$5E,$32,$47,$F8,$A1,$B1,$FB,$5F,$90,$7E,$07,$F4,$27,$F8,$0C
	.dc.b	$62,$95,$FD,$17,$D9,$7D,$07,$CA,$8F,$F3,$CB,$63,$33,$FE,$A3,$6C
	.dc.b	$FF,$26,$6F,$2F,$F8,$73,$C2,$7D,$9B,$E7,$BF,$83,$56,$B0,$F0,$D6
	.dc.b	$A7,$F8,$57,$D9,$3E,$07,$C4,$DA,$FF,$59,$37,$2F,$E6,$AD,$7F,$D9
	.dc.b	$AE,$72,$F1,$64,$FB,$3E,$B8,$7F,$24,$CB,$CB,$1E,$0F,$92,$EE,$4C
	.dc.b	$BB,$B7,$2F,$F0,$CE,$2F,$A5,$7C,$77,$1B,$F9,$73,$FC,$DA,$73,$18
	.dc.b	$3E,$7D,$34,$61,$FC,$DA,$FA,$61,$CE,$BB,$E1,$C9,$FE,$F5,$F6,$4F
	.dc.b	$8F,$FA,$CD,$CB,$FD,$AC,$D9,$FE,$B2,$D7,$FD,$E2,$E7,$0F,$AD,$F5
	.dc.b	$FF,$EB,$6A,$F3,$57,$E1,$2B,$51,$F5,$A7,$53,$FD,$07,$E4,$7F,$13
	.dc.b	$3C,$3C,$44,$37,$FE,$9E,$EA,$B8,$46,$1C,$B6,$B9,$4F,$17,$C7,$2C
	.dc.b	$1F,$2B,$F0,$09,$3F,$01,$B3,$FE,$4E,$B6,$2F,$7B,$FA,$B6,$1C,$B0
	.dc.b	$F7,$D6,$A7,$38,$7E,$16,$F8,$57,$8A,$F8,$61,$D6,$72,$FF,$63,$EC
	.dc.b	$FF,$7A,$57,$C3,$7F,$CD,$CD,$F6,$FF,$80,$F5,$7F,$02,$6C,$64,$E1
	.dc.b	$1F,$C9,$BE,$8F,$F8,$81,$23,$FD,$4A,$DB,$FD,$F7,$D7,$6E,$7E,$97
	.dc.b	$62,$3F,$C0,$3F,$62,$F9,$57,$C8,$61,$7C,$BD,$8C,$37,$0F,$C1,$D7
	.dc.b	$18,$C6,$3F,$E7,$A8,$C4,$EA,$FA,$B6,$B7,$CB,$FE,$9E,$AD,$DF,$C7
	.dc.b	$F4,$BF,$91,$FD,$27,$F2,$46,$E1,$F2,$5B,$3B,$8F,$99,$3F,$CA,$B5
	.dc.b	$FE,$57,$F6,$36,$7E,$83,$6B,$EA,$37,$1F,$EC,$9C,$28,$FC,$1F,$5F
	.dc.b	$EA,$89,$7F,$AA,$6D,$FF,$A6,$21,$F5,$AF,$95,$E3,$1F,$A8,$32,$E2
	.dc.b	$78,$4E,$E7,$F1,$55,$FE,$34,$76,$7E,$7B,$E4,$1F,$63,$FE,$2B,$6C
	.dc.b	$7D,$CB,$C4,$FF,$8B,$FE,$BF,$FC,$67,$A7,$FB,$2B,$89,$F6,$3C,$2F
	.dc.b	$FD,$01,$5A,$C3,$F9,$FD,$6B,$0F,$E8,$7D,$B2,$7F,$80,$CE,$5F,$C5
	.dc.b	$C7,$8F,$C2,$F0,$73,$D9,$FD,$AE,$1F,$8F,$6C,$7F,$94,$2F,$C7,$5C
	.dc.b	$27,$AE,$5F,$C8,$1B,$1B,$8C,$65,$E8,$BC,$4F,$BF,$79,$7F,$C1,$3E
	.dc.b	$C7,$F8,$9E,$BF,$E6,$3C,$FE,$76,$3F,$02,$FC,$5C,$0F,$8D,$E3,$3F
	.dc.b	$ED,$2D,$7E,$DC,$63,$0F,$95,$5B,$CB,$DB,$87,$2A,$FA,$1F,$35,$27
	.dc.b	$C5,$F1,$17,$2F,$9F,$3E,$0F,$47,$19,$74,$6D,$FD,$2C,$91,$8B,$FD
	.dc.b	$B4,$62,$7F,$90,$FD,$8B,$D4,$FF,$37,$36,$3E,$BE,$BF,$E1,$77,$73
	.dc.b	$A3,$E9,$7C,$47,$AF,$0F,$F6,$15,$F1,$7E,$14,$67,$2F,$F4,$64,$E5
	.dc.b	$FC,$64,$66,$FF,$AC,$CE,$5F,$C5,$C6,$6B,$FC,$17,$3F,$C9,$AB,$9F
	.dc.b	$8B,$F8,$3C,$5F,$37,$EA,$89,$8F,$E4,$31,$F8,$18,$F9,$A7,$D3,$A6
	.dc.b	$A3,$CE,$DC,$7C,$AF,$FA,$47,$EC,$7F,$CD,$13,$FE,$DF,$6C,$FE,$94
	.dc.b	$FF,$3E,$C7,$F8,$82,$33,$87,$F7,$46,$C7,$F2,$7F,$31,$F5,$3D,$7D
	.dc.b	$FF,$CE,$46,$2F,$FA,$AD,$F0,$D7,$FC,$09,$7B,$9F,$D5,$D6,$F0,$FF
	.dc.b	$79,$7D,$37,$B9,$FF,$01,$87,$0F,$F8,$19,$AF,$F8,$D9,$AE,$F3,$FF
	.dc.b	$1D,$FD,$37,$FA,$C6,$7E,$4E,$9F,$E8,$1F,$21,$3F,$E6,$EF,$1F,$F8
	.dc.b	$87,$D8,$7F,$8C,$71,$7A,$0F,$81,$FE,$87,$AF,$F1,$BA,$D4,$E6,$F8
	.dc.b	$FB,$3C,$EB,$4F,$FA,$EF,$EC,$1E,$2D,$3F,$CF,$BE,$35,$AF,$F0,$6F
	.dc.b	$C0,$AB,$71,$86,$E9,$FE,$CD,$D3,$FC,$41,$BA,$BF,$D6,$95,$A8,$FF
	.dc.b	$71,$7C,$5F,$0F,$8D,$FE,$FD,$87,$FB,$9F,$3E,$0F,$FB,$E7,$D5,$FF
	.dc.b	$2F,$EE,$FB,$0F,$F2,$A6,$7F,$E6,$2C,$38,$DB,$1E,$98,$7F,$BA,$71
	.dc.b	$E7,$7F,$EE,$D6,$D6,$63,$FC,$3C,$F8,$5A,$F3,$8F,$1D,$F9,$B1,$7C
	.dc.b	$23,$B7,$38,$F3,$C3,$D0,$7D,$69,$D6,$3E,$79,$FF,$3C,$F6,$3E,$C1
	.dc.b	$E8,$72,$C1,$39,$8C,$65,$F7,$FD,$CD,$4E,$71,$E9,$7C,$30,$E7,$1E
	.dc.b	$C9,$D4,$FC,$C3,$58,$7C,$26,$CF,$C0,$FA,$F7,$CF,$FF,$90,$E7,$F3
	.dc.b	$E2,$33,$E8,$62,$31,$3B,$7F,$4B,$3A,$78,$FC,$A7,$F0,$AF,$0B,$83
	.dc.b	$63,$3E,$2F,$F8,$7D,$B3,$D6,$BF,$CD,$19,$DF,$7B,$E2,$D7,$92,$72
	.dc.b	$D8,$CB,$FE,$78,$DF,$EB,$E1,$FE,$7A,$E3,$FB,$5C,$5F,$C7,$2B,$E4
	.dc.b	$D6,$9B,$1F,$ED,$F6,$D7,$1E,$2F,$F7,$33,$63,$EF,$36,$BF,$0B,$F1
	.dc.b	$37,$F0,$2F,$E0,$F8,$FE,$25,$F8,$CD,$E0,$D6,$10,$F1,$CE,$B1,$FD
	.dc.b	$0B,$63,$F8,$0B,$D2,$33,$1D,$E6,$3C,$8F,$26,$4F,$89,$FF,$4D,$F5
	.dc.b	$7E,$BB,$3F,$CD,$FC,$46,$39,$3F,$5F,$68,$C6,$1F,$9B,$C2,$DF,$E8
	.dc.b	$9B,$5B,$9F,$F4,$63,$5F,$F7,$AC,$6B,$E2,$64,$69,$E3,$F0,$18,$8F
	.dc.b	$F7,$DC,$5C,$BC,$D6,$AF,$0F,$8D,$FE,$3B,$BF,$F3,$4F,$A1,$83,$3E
	.dc.b	$9C,$8C,$90,$FF,$77,$3E,$AF,$98,$7B,$B7,$B2,$57,$FA,$96,$72,$D9
	.dc.b	$CA,$CF,$E5,$4F,$EE,$A3,$44,$35,$C1,$FC,$6A,$35,$18,$3F,$9D,$B8
	.dc.b	$9E,$75,$EE,$F9,$EF,$7F,$B9,$2F,$FD,$0F,$19,$7F,$D1,$31,$97,$A7
	.dc.b	$0B,$55,$6E,$73,$1F,$E5,$87,$89,$D0,$F8,$1E,$EF,$DE,$F9,$2F,$EB
	.dc.b	$F9,$7E,$1D,$BA,$7A,$E7,$E5,$1B,$A8,$72,$21,$E4,$9E,$2F,$F5,$A6
	.dc.b	$E2,$D3,$FC,$FD,$3F,$EA,$4F,$B7,$FB,$FE,$9C,$8C,$10,$F5,$B9,$7D
	.dc.b	$B1,$F8,$63,$19,$7F,$6D,$9C,$4F,$FB,$87,$0F,$E3,$13,$ED,$F9,$1F
	.dc.b	$80,$7C,$1E,$89,$5B,$8C,$BC,$63,$2F,$B6,$32,$FB,$E3,$F2,$FF,$29
	.dc.b	$23,$FD,$7E,$FE,$6B,$E5,$24,$60,$87,$D6,$72,$D3,$FD,$8E,$FE,$8B
	.dc.b	$E0,$3F,$47,$F7,$72,$7F,$D9,$6F,$D9,$F3,$3E,$D7,$C1,$59,$F8,$11
	.dc.b	$87,$E1,$7C,$6F,$DF,$23,$FE,$D4,$F3,$BF,$80,$AF,$A8,$63,$85,$8D
	.dc.b	$31,$12,$47,$E7,$CE,$51,$B7,$3B,$7F,$55,$AE,$B1,$8E,$5E,$2F,$86
	.dc.b	$F8,$3C,$E7,$C2,$EC,$EE,$F2,$7F,$47,$FB,$FF,$F6,$1B,$FE,$40,$BE
	.dc.b	$7B,$F8,$DB,$67,$97,$32,$38,$63,$FE,$12,$6C,$FF,$8C,$9E,$0F,$6A
	.dc.b	$E7,$3D,$3E,$93,$AF,$9F,$2B,$AD,$A7,$51,$89,$D3,$E8,$8F,$2F,$97
	.dc.b	$E6,$F6,$24,$F2,$8C,$BD,$23,$2F,$C0,$8C,$BF,$0A,$33,$9F,$C2,$9F
	.dc.b	$BB,$F5,$3F,$CA,$EB,$FB,$FB,$5F,$F2,$FB,$5F,$A6,$D7,$EE,$BC,$67
	.dc.b	$FC,$1C,$DC,$B7,$E1,$7E,$1F,$AF,$E7,$FD,$A7,$D0,$FB,$39,$8F,$C4
	.dc.b	$8C,$BF,$1A,$32,$FD,$A8,$FC,$85,$7E,$47,$C5,$93,$F8,$13,$F2,$D1
	.dc.b	$A7,$1C,$49,$DC,$7E,$26,$74,$FE,$2E,$33,$9F,$E4,$E7,$F9,$17,$E0
	.dc.b	$9F,$C8,$D9,$5F,$C2,$63,$2F,$F0,$B8,$CB,$FC,$7E,$3E,$6F,$57,$F8
	.dc.b	$9E,$1A,$A3,$0D,$8F,$F1,$63,$5F,$A7,$1C,$9F,$74,$66,$FF,$51,$D3
	.dc.b	$FB,$6F,$AB,$7D,$BD,$F9,$09,$87,$C5,$9C,$B6,$72,$FE,$D6,$13,$CE
	.dc.b	$64,$F1,$8D,$DF,$F9,$2E,$B9,$CF,$EF,$7E,$B3,$FD,$C0,$FF,$91,$5A
	.dc.b	$FF,$92,$9A,$FF,$AA,$9A,$FE,$F9,$DC,$FF,$95,$DB,$C5,$9F,$8A,$F7
	.dc.b	$7A,$FF,$C5,$3A,$3C,$F8,$7F,$94,$DA,$FF,$96,$1A,$FF,$AE,$1A,$FF
	.dc.b	$0C,$77,$87,$F9,$95,$B7,$9C,$BF,$90,$FD,$A7,$BA,$BF,$84,$BF,$C2
	.dc.b	$3E,$6E,$61,$FB,$C8,$FE,$16,$61,$FE,$23,$18,$7F,$80,$E5,$FC,$40
	.dc.b	$8F,$F0,$4F,$8B,$AF,$F1,$88,$B7,$FC,$62,$7F,$C5,$FF,$2D,$FE,$30
	.dc.b	$DE,$5F,$F8,$C2,$32,$FF,$8C,$A3,$2F,$F8,$13,$D2,$4F,$F8,$E3,$3F
	.dc.b	$F1,$CF,$A1,$2B,$FC,$73,$19,$7F,$C7,$71,$97,$94,$65,$E9,$19,$C7
	.dc.b	$FC,$6F,$39,$7F,$C7,$31,$9D,$7F,$C9,$B3,$FE,$48,$F4,$7F,$C9,$3E
	.dc.b	$74,$C3,$FC,$93,$19,$7F,$C9,$71,$97,$D5,$1F,$E5,$2F,$85,$FF,$04
	.dc.b	$9F,$DE,$AF,$FD,$8E,$D9,$FF,$73,$BA,$78,$79,$71,$FE,$7F,$19,$6C
	.dc.b	$65,$FF,$15,$C7,$F8,$5B,$D8,$FE,$15,$1F,$D1,$9F,$F1,$6B,$7D,$2F
	.dc.b	$F3,$2C,$6A,$3F,$C6,$6D,$D3,$FC,$E0,$F0,$8C,$CF,$0E,$D9,$1C,$C8
	.dc.b	$72,$72,$F0,$33,$5D,$23,$FC,$EB,$F4,$FF,$E5,$89,$FF,$21,$BF,$E4
	.dc.b	$DC,$BF,$DD,$6D,$87,$90,$8E,$17,$FE,$85,$6D,$70,$D3,$A9,$DB,$79
	.dc.b	$87,$F6,$8A,$E9,$87,$3A,$CC,$FF,$80,$7D,$FF,$F7,$46,$3F,$E0,$4F
	.dc.b	$A5,$7D,$2F,$F7,$E9,$7C,$6F,$F8,$02,$73,$7F,$E0,$A9,$CB,$C6,$B8
	.dc.b	$7C,$2F,$F9,$CE,$15,$C1,$57,$81,$CD,$77,$BE,$CD,$8E,$87,$7B,$E9
	.dc.b	$7C,$EF,$97,$6B,$E9,$BC,$39,$7F,$BF,$1B,$87,$FB,$E9,$B3,$B8,$FF
	.dc.b	$61,$99,$2B,$97,$FD,$AC,$64,$AE,$5F,$F1,$5F,$C0,$79,$9F,$F1,$6F
	.dc.b	$2B,$87,$11,$A3,$DE,$7F,$8A,$A3,$4E,$0E,$F9,$7C,$75,$1A,$8C,$42
	.dc.b	$BA,$CE,$74,$FF,$1E,$EE,$FA,$57,$8F,$DB,$5F,$F6,$74,$7F,$71,$CF
	.dc.b	$FC,$5A,$E9,$BC,$9B,$79,$C6,$73,$FF,$4D,$BC,$5B,$7B,$7F,$33,$19
	.dc.b	$BE,$F3,$9C,$3C,$98,$5A,$E9,$5E,$48,$D7,$33,$DC,$BE,$66,$1B,$9F
	.dc.b	$EF,$DB,$9F,$EF,$1B,$87,$EF,$1B,$3F,$EF,$86,$CF,$FB,$E1,$B9,$30
	.dc.b	$F0,$57,$F0,$BC,$3D,$1D,$2F,$62,$F8,$F8,$6D,$7F,$18,$6E,$5F,$C5
	.dc.b	$9B,$9F,$FB,$F6,$F9,$57,$F1,$EB,$E4,$D3,$F9,$0F,$AF,$E0,$F5,$FC
	.dc.b	$2D,$8C,$9F,$C9,$A3,$2D,$3F,$94,$F0,$FA,$3E,$8B,$D2,$D8,$C9,$FC
	.dc.b	$BA,$32,$D3,$39,$FF,$2E,$8C,$B6,$33,$97,$AE,$72,$D8,$FE,$73,$F1
	.dc.b	$FF,$23,$CF,$F9,$2D,$8C,$9F,$CF,$A3,$2D,$3F,$CE,$7F,$75,$F8,$7F
	.dc.b	$58,$DF,$BA,$C3,$27,$67,$F9,$EA,$31,$97,$FA,$3B,$E8,$79,$7D,$5F
	.dc.b	$36,$D7,$FA,$01,$B3,$B8,$FF,$41,$CF,$E6,$E1,$F7,$EF,$87,$C3,$6B
	.dc.b	$FD,$3D,$F8,$4B,$F0,$97,$B5,$AB,$95,$FE,$A8,$C3,$FD,$53,$87,$17
	.dc.b	$07,$69,$FE,$06,$EE,$34,$F1,$FA,$6F,$F5,$FD,$7F,$9B,$5B,$0E,$1F
	.dc.b	$E1,$7F,$2A,$FB,$D7,$D4,$FA,$EF,$75,$D9,$E3,$7D,$7E,$0F,$C2,$FF
	.dc.b	$AF,$A3,$2D,$8C,$AF,$A5,$7C,$9B,$D8,$FE,$05,$18,$F6,$2C,$E7,$5F
	.dc.b	$F6,$24,$E5,$B1,$9B,$F5,$47,$FB,$5A,$75,$18,$3C,$13,$BC,$BF,$94
	.dc.b	$CE,$9B,$19,$AE,$4A,$FC,$78,$79,$1F,$BD,$F5,$19,$AF,$F4,$C3,$71
	.dc.b	$F0,$6D,$F2,$FE,$87,$77,$2F,$EC,$D2,$F8,$3F,$7D,$AF,$FB,$A1,$AF
	.dc.b	$F7,$EF,$0D,$F0,$9F,$C9,$57,$C2,$D8,$C3,$6F,$FD,$C7,$3A,$6C,$E9
	.dc.b	$F9,$7F,$4F,$F4,$3E,$73,$E8,$B6,$32,$7D,$08,$CB,$4C,$E1,$F3,$F8
	.dc.b	$3E,$EC,$EA,$31,$BA,$3F,$6D,$7F,$B7,$6D,$FF,$BB,$1B,$5F,$F0,$01
	.dc.b	$BC,$1F,$F0,$01,$FA,$3E,$34,$A5,$7D,$FF,$39,$F5,$0F,$A4,$D4,$FE
	.dc.b	$0C,$FA,$17,$F9,$46,$DF,$D4,$E0,$4E,$5B,$1F,$B0,$F3,$1E,$1E,$2F
	.dc.b	$F1,$D1,$8F,$9B,$FE,$25,$87,$F1,$36,$DF,$9B,$5F,$EF,$35,$B8,$C1
	.dc.b	$FD,$CB,$0E,$93,$A5,$5C,$B0,$EF,$18,$AE,$31,$FE,$42,$8E,$31,$CE
	.dc.b	$7F,$92,$7D,$7F,$F6,$BF,$80,$7A,$D8,$AE,$39,$7F,$A3,$23,$18,$FB
	.dc.b	$2B,$75,$AC,$58,$73,$BE,$9E,$BF,$5E,$EB,$7D,$30,$FE,$16,$D9,$D3
	.dc.b	$4D,$4F,$F0,$0D,$7F,$80,$E9,$67,$F7,$ED,$CF,$F1,$6D,$D7,$F7,$2D
	.dc.b	$9F,$DC,$B6,$7F,$80,$BF,$A7,$E9,$93,$E5,$F9,$8B,$5F,$6B,$A7,$F8
	.dc.b	$13,$FA,$FF,$27,$F5,$DD,$DF,$AF,$7E,$57,$24,$C3,$F5,$B1,$6B,$E4
	.dc.b	$76,$FC,$F3,$E8,$7D,$5D,$AF,$89,$1F,$4F,$DC,$C5,$7C,$F7,$13,$F2
	.dc.b	$5C,$ED,$F8,$8E,$2F,$F3,$6C,$FE,$F3,$CB,$FE,$05,$E5,$FF,$03,$6C
	.dc.b	$64,$FE,$03,$19,$69,$9C,$3E,$04,$65,$B1,$F6,$7D,$CF,$C4,$7C,$DF
	.dc.b	$E2,$5B,$19,$3E,$F4,$65,$A6,$67,$F1,$51,$96,$C7,$F1,$4F,$37,$F8
	.dc.b	$D7,$81,$FC,$6D,$B1,$93,$F8,$CC,$65,$A6,$67,$EC,$46,$5B,$1F,$91
	.dc.b	$F9,$BF,$D0,$7C,$DF,$E8,$5B,$19,$3F,$2D,$19,$69,$99,$FD,$14,$65
	.dc.b	$B1,$FC,$B3,$C2,$FE,$65,$D3,$FC,$CD,$B1,$93,$F9,$8C,$65,$A6,$67
	.dc.b	$F2,$11,$96,$C7,$EA,$7C,$CF,$DC,$7C,$DF,$EE,$5B,$19,$3F,$5D,$19
	.dc.b	$69,$FE,$E4,$F9,$0F,$F2,$99,$FE,$54,$F7,$AC,$64,$97,$FE,$E6,$8C
	.dc.b	$BF,$EE,$8F,$77,$FC,$B7,$F0,$BF,$E5,$C6,$D7,$FB,$C1,$B3,$B8,$FF
	.dc.b	$79,$CF,$F9,$E7,$E8,$3F,$AB,$C7,$FA,$83,$BB,$FD,$1F,$1F,$D4,$EB
	.dc.b	$FD,$29,$18,$F3,$3F,$D3,$BF,$02,$8D,$F9,$D8,$75,$87,$3B,$FF,$70
	.dc.b	$E1,$CE,$F8,$46,$30,$E0,$EB,$0B,$E4,$AF,$CB,$3E,$23,$58,$75,$9D
	.dc.b	$5F,$0A,$ED,$8F,$FA,$A3,$83,$FD,$03,$87,$1B,$FE,$93,$F0,$1D,$A5
	.dc.b	$E1,$D8,$99,$75,$9E,$3A,$7A,$2F,$FA,$8D,$EB,$0F,$20,$F9,$D7,$C7
	.dc.b	$7F,$AF,$A3,$FC,$87,$18,$F3,$FF,$C1,$AD,$8C,$CE,$DF,$F2,$67,$C2
	.dc.b	$BB,$3F,$C6,$71,$FE,$4C,$8C,$7A,$3F,$E1,$66,$C6,$67,$6F,$F9,$56
	.dc.b	$FF,$C2,$7E,$6F,$83,$8B,$FA,$6C,$F9,$1A,$ED,$AF,$FB,$43,$E2,$FC
	.dc.b	$4F,$CF,$6B,$F2,$DB,$B7,$C5,$5D,$BF,$E0,$09,$D6,$36,$B4,$DE,$3E
	.dc.b	$CD,$97,$03,$0E,$71,$F7,$FE,$13,$FC,$E1,$87,$D3,$7E,$43,$78,$39
	.dc.b	$5F,$FB,$06,$AD,$70,$C7,$FD,$D6,$F0,$6D,$6A,$61,$DD,$E2,$61,$F8
	.dc.b	$1D,$DE,$DF,$17,$FC,$0F,$97,$FA,$2E,$7F,$BB,$B5,$FF,$08,$A9,$E1
	.dc.b	$CB,$F3,$5B,$3F,$02,$78,$57,$4E,$EC,$E7,$F0,$EE,$D3,$BA,$93,$0F
	.dc.b	$0F,$CE,$D7,$C3,$17,$1F,$3C,$F4,$F9,$D7,$7D,$EF,$98,$7F,$84,$27
	.dc.b	$2D,$9C,$BF,$E2,$C8,$4F,$39,$F3,$FC,$EB,$E3,$2F,$C5,$CC,$3F,$C3
	.dc.b	$73,$96,$CE,$5F,$F1,$BC,$27,$9C,$C9,$E3,$1B,$CB,$FC,$6F,$5C,$E7
	.dc.b	$C9,$E1,$7B,$7C,$09,$1E,$B8,$F8,$06,$1F,$54,$61,$D1,$F0,$5D,$57
	.dc.b	$F9,$13,$35,$7B,$BE,$37,$EA,$FC,$E7,$FB,$0E,$BF,$C9,$2D,$E4,$FF
	.dc.b	$30,$F1,$B3,$5F,$1C,$FF,$C7,$B7,$B7,$F4,$2D,$AE,$13,$B9,$FF,$2D
	.dc.b	$CE,$4D,$57,$48,$FB,$1E,$0F,$DD,$F9,$D9,$18,$21,$F6,$5C,$BF,$6E
	.dc.b	$3F,$08,$63,$2F,$81,$38,$9F,$3F,$3F,$C0,$F8,$C9,$18,$21,$E6,$72
	.dc.b	$FA,$23,$E3,$98,$AF,$8F,$38,$9F,$C8,$7C,$EF,$E5,$48,$FE,$4F,$E6
	.dc.b	$3F,$30,$61,$FC,$94,$61,$D1,$0F,$CB,$47,$E7,$FE,$07,$F4,$1F,$39
	.dc.b	$27,$F3,$CF,$DA,$F8,$FF,$B7,$F1,$F6,$7E,$AC,$61,$FA,$FF,$1F,$F8
	.dc.b	$02,$3F,$A7,$F9,$DF,$DD,$AF,$8D,$63,$85,$8D,$31,$12,$7F,$D0,$E7
	.dc.b	$18,$D3,$89,$E6,$D9,$CB,$FE,$A9,$8E,$AE,$4F,$F4,$27,$2F,$F0,$AF
	.dc.b	$83,$7C,$ED,$31,$12,$7F,$D1,$51,$97,$89,$FE,$8D,$F9,$DE,$5E,$7C
	.dc.b	$8C,$10,$FF,$50,$B9,$7F,$D4,$71,$E1,$F2,$FC,$5E,$94,$9F,$F4,$C4
	.dc.b	$65,$FF,$4D,$46,$5F,$F5,$5C,$65,$FF,$56,$C6,$79,$3F,$D6,$53,$F4
	.dc.b	$3E,$73,$F8,$BD,$7C,$3E,$4F,$A0,$67,$C2,$91,$FE,$FE,$9D,$BF,$D5
	.dc.b	$C3,$87,$FA,$D3,$5F,$F5,$D7,$CE,$FC,$BF,$8F,$91,$FE,$B1,$8F,$9C
	.dc.b	$61,$FF,$61,$C6,$1F,$B5,$5F,$83,$F1,$64,$60,$87,$DB,$72,$FD,$D8
	.dc.b	$FC,$51,$8C,$9B,$FF,$6F,$E1,$FE,$67,$5C,$A7,$F8,$9F,$AD,$FC,$5A
	.dc.b	$E4,$FF,$10,$8F,$E3,$26,$1F,$E1,$B1,$87,$F2,$79,$BA,$7F,$35,$CF
	.dc.b	$27,$31,$F4,$4F,$94,$FC,$86,$A3,$E6,$B8,$69,$AF,$9D,$7C,$D7,$EA
	.dc.b	$97,$CD,$53,$11,$27,$F4,$F3,$A6,$CE,$56,$7F,$19,$F1,$4F,$67,$64
	.dc.b	$37,$AF,$F8,$66,$35,$18,$3D,$5F,$3A,$F8,$7F,$37,$C5,$49,$FD,$CC
	.dc.b	$E5,$B3,$EF,$8C,$47,$C6,$6B,$F9,$16,$BF,$E2,$9E,$8F,$63,$E1,$1F
	.dc.b	$3B,$FD,$7F,$2F,$64,$63,$CD,$FE,$23,$5C,$27,$2F,$29,$FC,$F3,$5E
	.dc.b	$DA,$7D,$38,$C4,$7F,$8C,$BB,$3B,$BF,$66,$BE,$57,$92,$97,$CE,$FF
	.dc.b	$0D,$3F,$87,$65,$2F,$86,$FF,$D8,$AB,$71,$83,$B5,$6D,$F9,$D7,$D7
	.dc.b	$5F,$E3,$04,$8E,$11,$E1,$F8,$BC,$47,$0B,$F9,$B9,$79,$9B,$8F,$86
	.dc.b	$BF,$C5,$3F,$0D,$FE,$2A,$E2,$98,$7F,$8E,$23,$2F,$F8,$EA,$32,$FF
	.dc.b	$8E,$A3,$FC,$63,$B7,$F8,$CB,$79,$3F,$E4,$08,$CB,$FE,$42,$8C,$BF
	.dc.b	$E4,$28,$CB,$FE,$44,$8C,$E5,$FE,$4A,$9C,$BF,$E4,$C8,$CE,$3F,$E7
	.dc.b	$39,$FF,$20,$FA,$1F,$E4,$2F,$36,$61,$FE,$50,$8C,$BF,$E5,$28,$CB
	.dc.b	$FE,$52,$8F,$F2,$5F,$BF,$FE,$0D,$3F,$D2,$18,$FD,$C6,$E1,$FE,$A9
	.dc.b	$AD,$9C,$3C,$A8,$FD,$29,$E4,$E0,$FB,$2F,$27,$FC,$F6,$FF,$86,$FD
	.dc.b	$8F,$E2,$B1,$F7,$9F,$F1,$5B,$F7,$1A,$E7,$BB,$FC,$C3,$1C,$30,$FC
	.dc.b	$8B,$A6,$FA,$1A,$99,$5C,$E3,$89,$87,$A1,$92,$18,$3F,$23,$F3,$BF
	.dc.b	$AD,$CB,$FC,$72,$FF,$94,$E3,$F2,$2D,$77,$B7,$F9,$DA,$38,$47,$FA
	.dc.b	$69,$D6,$1F,$A8,$6D,$74,$AE,$07,$1D,$25,$F6,$8F,$F0,$5F,$BE,$F8
	.dc.b	$5F,$E3,$72,$F9,$D7,$CE,$7F,$AB,$E9,$F1,$5F,$EB,$19,$CD,$7F,$31
	.dc.b	$9C,$BF,$CC,$E3,$2A,$3B,$69,$9B,$E9,$87,$82,$F7,$5C,$AF,$C5,$5E
	.dc.b	$1C,$3C,$0F,$0C,$39,$4F,$87,$0F,$4F,$C1,$3E,$43,$52,$8C,$3B,$AF
	.dc.b	$E9,$11,$88,$F3,$57,$2A,$FE,$9F,$F0,$3A,$6C,$F3,$6D,$71,$8F,$EA
	.dc.b	$53,$B7,$FD,$96,$7A,$8F,$F6,$C7,$27,$F4,$7D,$3E,$2C,$70,$34,$7F
	.dc.b	$5B,$8D,$65,$E0,$9D,$C6,$27,$AC,$63,$7F,$87,$D6,$F9,$DF,$F5,$8C
	.dc.b	$7F,$1C,$7F,$D5,$6D,$F1,$FF,$AC,$6D,$FC,$C2,$73,$87,$F2,$67,$8B
	.dc.b	$79,$77,$7F,$E9,$09,$CD,$71,$9C,$D7,$FB,$BA,$73,$5B,$9C,$E1,$E5
	.dc.b	$C2,$D7,$83,$0F,$D8,$FB,$9A,$E7,$7B,$36,$30,$DC,$BF,$78,$DC,$FF
	.dc.b	$80,$36,$FF,$76,$D9,$FF,$7A,$B6,$59,$30,$EF,$5F,$C1,$F0,$EE,$FF
	.dc.b	$13,$F6,$3D,$3D,$4F,$5E,$C6,$1B,$3F,$C6,$5B,$8F,$F1,$86,$E7,$FC
	.dc.b	$8D,$B3,$FE,$BA,$BE,$8F,$F2,$1F,$63,$BF,$97,$E0,$6C,$64,$FE,$4D
	.dc.b	$19,$69,$FC,$A7,$8F,$CF,$F3,$BE,$86,$C6,$4F,$E5,$D1,$96,$99,$CB
	.dc.b	$F9,$7C,$65,$B1,$9C,$BF,$9E,$4E,$5B,$1F,$CE,$7E,$77,$E3,$FA,$1F
	.dc.b	$21,$B1,$93,$F9,$F4,$65,$A7,$F7,$2F,$19,$F2,$3F,$88,$6C,$67,$E3
	.dc.b	$F1,$32,$38,$3B,$8C,$65,$FD,$E7,$E3,$F9,$7C,$7F,$36,$DF,$76,$CE
	.dc.b	$E3,$C5,$3F,$9C,$87,$CD,$BE,$23,$FC,$4B,$1F,$E5,$13,$2E,$1F,$F2
	.dc.b	$A4,$6E,$BF,$CA,$ED,$8F,$F2,$DB,$75,$FF,$25,$7C,$CB,$F0,$27,$CA
	.dc.b	$F8,$6B,$58,$7F,$97,$EB,$53,$E5,$F9,$BF,$F6,$1C,$FF,$99,$1B,$8F
	.dc.b	$F9,$A1,$AF,$FA,$4D,$AF,$A5,$DE,$3F,$C1,$3E,$91,$CF,$D7,$E0,$FF
	.dc.b	$7B,$29,$CE,$1F,$E2,$FF,$9C,$7C,$9F,$F9,$66,$30,$BD,$4F,$E0,$B3
	.dc.b	$AF,$56,$CE,$6B,$71,$8A,$CB,$63,$27,$48,$CB,$4F,$F6,$EF,$CD,$F3
	.dc.b	$D7,$FD,$0E,$D7,$FD,$64,$D7,$EC,$35,$FF,$45,$3C,$72,$FF,$3E,$79
	.dc.b	$0F,$9C,$7C,$97,$E0,$EB,$58,$7F,$37,$9C,$BF,$E4,$AF,$19,$F0,$BF
	.dc.b	$C8,$17,$C3,$58,$C3,$67,$83,$73,$FF,$2B,$D7,$09,$F9,$7F,$31,$F2
	.dc.b	$3E,$63,$E4,$B6,$32,$7D,$08,$CB,$4C,$D7,$FA,$3F,$87,$EB,$CE,$A3
	.dc.b	$1C,$28,$FD,$85,$FE,$C5,$B7,$FB,$56,$D7,$F9,$95,$BB,$FF,$99,$5F
	.dc.b	$8D,$C9,$FA,$39,$6D,$F8,$87,$EA,$C6,$63,$11,$F5,$A3,$FA,$5D,$BE
	.dc.b	$9F,$FA,$5C,$FD,$69,$CC,$63,$CB,$FF,$44,$47,$F0,$7F,$9C,$FE,$10
	.dc.b	$DB,$FE,$18,$DB,$FF,$3E,$37,$2F,$F3,$E3,$67,$FD,$18,$D9,$FF,$4C
	.dc.b	$7C,$EF,$F5,$AF,$7B,$FC,$19,$5F,$C5,$E3,$1E,$A7,$F1,$76,$C6,$4F
	.dc.b	$F4,$6C,$F0,$AF,$F4,$9B,$FE,$17,$8C,$FC,$27,$5F,$5B,$B3,$63,$27
	.dc.b	$28,$CB,$FA,$EC,$7F,$57,$9F,$EB,$1B,$19,$3F,$67,$19,$69,$9C,$3F
	.dc.b	$03,$19,$6C,$7E,$6F,$A7,$F3,$47,$C1,$E8,$B1,$F9,$FE,$AF,$CF,$78
	.dc.b	$DF,$9F,$6C,$7C,$0E,$FF,$90,$63,$0F,$82,$FC,$BF,$9C,$93,$F1,$BE
	.dc.b	$76,$D7,$C6,$9F,$A1,$ED,$63,$8F,$FD,$8F,$C3,$FE,$54,$C3,$E6,$CE
	.dc.b	$9F,$9F,$56,$FF,$57,$E4,$FE,$CF,$C9,$FD,$A3,$6B,$F7,$0D,$9D,$C7
	.dc.b	$EE,$E7,$6D,$36,$58,$FB,$7E,$D7,$E1,$7E,$6B,$F0,$CD,$9F,$C0,$B6
	.dc.b	$77,$1F,$84,$9D,$B4,$D9,$63,$F8,$1F,$97,$FC,$2F,$B7,$F8,$63,$67
	.dc.b	$F8,$83,$67,$71,$FC,$4E,$76,$D3,$65,$8F,$C9,$FC,$DF,$E7,$7E,$6F
	.dc.b	$F3,$CD,$9F,$CC,$B6,$77,$1F,$9C,$9D,$B4,$D9,$63,$F8,$DF,$7F,$F2
	.dc.b	$3E,$8F,$E4,$8D,$9F,$E5,$0D,$9D,$C7,$F2,$B9,$DB,$4D,$96,$3F,$57
	.dc.b	$E6,$FE,$D7,$E6,$BF,$6C,$D9,$FD,$8B,$67,$71,$FB,$49,$FF,$1E,$FC
	.dc.b	$74,$8F,$F6,$97,$B9,$6B,$FD,$A3,$37,$97,$FD,$B9,$36,$BF,$C7,$5E
	.dc.b	$EF,$F9,$1B,$E1,$3F,$C8,$ED,$8F,$F6,$E3,$70,$FF,$6D,$B6,$D2,$9F
	.dc.b	$F2,$AF,$CE,$FF,$4A,$8F,$F4,$6F,$6E,$7C,$44,$FF,$45,$A7,$C0,$FF
	.dc.b	$56,$78,$F9,$9D,$1C,$DF,$1A,$E1,$5B,$BE,$77,$B8,$C4,$EA,$B8,$4E
	.dc.b	$5E,$5E,$65,$C3,$BE,$3E,$67,$D3,$19,$C7,$AC,$E5,$F4,$C6,$78,$5E
	.dc.b	$F5,$31,$E6,$DB,$E1,$3A,$6C,$E9,$A7,$87,$07,$CE,$BE,$31,$E8,$FF
	.dc.b	$8D,$E7,$FC,$71,$39,$8C,$47,$D1,$8C,$38,$6B,$C7,$E1,$1D,$BF,$E2
	.dc.b	$08,$E3,$18,$F4,$39,$C7,$FB,$86,$74,$F1,$6B,$E2,$E0,$FF,$6B,$3D
	.dc.b	$DA,$F9,$EF,$FC,$17,$E7,$74,$6C,$7A,$F8,$3F,$C6,$FF,$31,$FE,$53
	.dc.b	$E4,$FB,$9C,$0F,$94,$C7,$27,$FA,$3D,$B8,$7F,$A8,$A5,$78,$C3,$9C
	.dc.b	$FD,$0A,$C5,$73,$8F,$A7,$5B,$F9,$DF,$87,$1F,$75,$AF,$E4,$0F,$F5
	.dc.b	$7B,$8C,$7F,$20,$BC,$0E,$5B,$4A,$E4,$F5,$8C,$47,$3C,$95,$CF,$3F
	.dc.b	$D2,$36,$7B,$D7,$A3,$E4,$BF,$CE,$DD,$3F,$A3,$C3,$FC,$BB,$EC,$63
	.dc.b	$9B,$FD,$64,$DC,$3E,$25,$DA,$FF,$31,$B7,$65,$53,$11,$9F,$98,$97
	.dc.b	$FB,$BC,$7C,$16,$9E,$5F,$33,$AA,$56,$A7,$0F,$09,$F1,$79,$8F,$8B
	.dc.b	$BF,$19,$30,$FF,$06,$C6,$8C,$3F,$E0,$E8,$C3,$E3,$D7,$F8,$57,$CE
	.dc.b	$3C,$3F,$72,$F0,$A9,$88,$95,$C4,$FF,$74,$CE,$DC,$70,$23,$11,$EC
	.dc.b	$BE,$F5,$CA,$30,$72,$BE,$EF,$2A,$F2,$3F,$E2,$8F,$8F,$FA,$5F,$5D
	.dc.b	$FB,$9D,$BE,$D3,$71,$FE,$49,$AA,$BE,$39,$C2,$E1,$CC,$87,$8B,$35
	.dc.b	$F3,$84,$61,$E6,$79,$7C,$5F,$8F,$E1,$48,$C1,$0E,$99,$FF,$1C,$7D
	.dc.b	$91,$F3,$A6,$46,$08,$70,$7D,$FE,$0F,$C0,$F8,$C9,$3F,$3B,$1F,$E3
	.dc.b	$6D,$9C,$BE,$B8,$4F,$28,$D1,$C5,$DE,$BF,$2A,$BF,$03,$E2,$7E,$0B
	.dc.b	$B6,$5F,$D3,$8C,$BF,$56,$32,$FB,$A3,$2F,$C0,$8C,$D7,$EB,$F3,$57
	.dc.b	$DA,$F8,$C9,$7E,$B7,$EE,$FC,$64,$8C,$10,$F6,$39,$69,$F0,$DF,$C2
	.dc.b	$FC,$67,$E1,$BE,$32,$4F,$C7,$7F,$3B,$E1,$7E,$7B,$B2,$CF,$E3,$E3
	.dc.b	$0F,$E4,$7B,$7F,$46,$47,$E8,$7C,$67,$F0,$55,$FA,$3B,$E7,$30,$BE
	.dc.b	$72,$98,$89,$3F,$A9,$38,$C6,$9C,$4F,$36,$CE,$5F,$D3,$9C,$0C,$9F
	.dc.b	$83,$F8,$FF,$D2,$FC,$7C,$8C,$10,$FC,$23,$97,$F0,$D1,$FC,$7F,$91
	.dc.b	$F9,$2B,$DC,$C2,$F7,$A9,$88,$97,$FC,$0E,$B8,$CE,$A3,$15,$CD,$FC
	.dc.b	$C7,$A9,$FB,$AF,$0A,$46,$08,$7E,$65,$CB,$F9,$B8,$FE,$02,$63,$A3
	.dc.b	$F8,$0C,$E2,$7E,$67,$CD,$B8,$3F,$B8,$C2,$30,$FF,$74,$6B,$FE,$00
	.dc.b	$6B,$FE,$05,$73,$5F,$E2,$98,$E5,$3F,$81,$9C,$F8,$5F,$E3,$3E,$DF
	.dc.b	$E3,$F7,$FC,$53,$BB,$F8,$BF,$7C,$C7,$F7,$11,$FC,$68,$C3,$FB,$78
	.dc.b	$C3,$F8,$5E,$1F,$E9,$AD,$F1,$FF,$21,$1F,$D2,$63,$5C,$D8,$33,$32
	.dc.b	$32,$43,$F1,$33,$C5,$B1,$C5,$FE,$73,$F2,$3F,$99,$6F,$A9,$FD,$86
	.dc.b	$3F,$C5,$51,$AE,$8C,$19,$99,$19,$3E,$9F,$07,$D2,$DF,$F9,$DC,$7C
	.dc.b	$F8,$C1,$1D,$90,$F3,$F7,$FC,$08,$F8,$3E,$34,$9F,$8D,$19,$7E,$44
	.dc.b	$65,$F7,$C6,$5F,$83,$19,$DF,$EE,$69,$E4,$7E,$F6,$0E,$CF,$41,$2F
	.dc.b	$F1,$0D,$7F,$67,$1B,$7F,$23,$1F,$5B,$C0,$FE,$0D,$F3,$4F,$83,$FC
	.dc.b	$9B,$6B,$FC,$8A,$D7,$FC,$A2,$D7,$FC,$AC,$E7,$0F,$C9,$37,$97,$FD
	.dc.b	$0D,$28,$C3,$FE,$89,$F9,$BF,$F2,$4D,$7F,$66,$DB,$FB,$4F,$8B,$2B
	.dc.b	$FA,$54,$7F,$6C,$30,$FF,$49,$8C,$3F,$D3,$FC,$1F,$EE,$24,$7F,$A9
	.dc.b	$69,$5F,$EA,$11,$6F,$FB,$AC,$FF,$8A,$7F,$76,$FD,$54,$7F,$8C,$5F
	.dc.b	$E3,$D5,$FD,$C5,$B1,$9C,$F8,$1C,$7C,$19,$5F,$E0,$EA,$E7,$3F,$EA
	.dc.b	$37,$F3,$11,$CF,$8F,$C0,$78,$79,$24,$71,$8C,$BC,$E3,$2F,$13,$46
	.dc.b	$48,$64,$98,$7A,$E3,$DD,$EB,$7B,$FD,$99,$5E,$78,$CB,$E9,$8C,$BF
	.dc.b	$06,$3E,$37,$CF,$7C,$7F,$9E,$93,$EF,$8C,$BF,$06,$32,$FC,$98,$CB
	.dc.b	$F2,$E3,$38,$7C,$F9,$FF,$2C,$4F,$F9,$67,$C3,$8F,$F8,$E3,$D2,$FF
	.dc.b	$10,$79,$A6,$9F,$E0,$E8,$FF,$13,$36,$FF,$CB,$71,$D0,$E5,$5D,$1D
	.dc.b	$99,$38,$B9,$8E,$31,$FE,$48,$F9,$7F,$E5,$3B,$FF,$B7,$57,$83,$FE
	.dc.b	$71,$6C,$EF,$B3,$06,$66,$46,$48,$7F,$8F,$67,$8B,$63,$8B,$FD,$CF
	.dc.b	$DC,$7C,$27,$F9,$CA,$1F,$32,$F3,$7F,$5F,$4F,$88,$FD,$8C,$E6,$BF
	.dc.b	$DF,$13,$97,$FD,$F5,$19,$51,$DB,$4C,$DF,$4C,$3C,$17,$BA,$E5,$7E
	.dc.b	$2A,$F0,$E1,$E0,$78,$61,$CA,$7C,$38,$7A,$7D,$E7,$C6,$EA,$51,$87
	.dc.b	$74,$73,$3B,$9D,$47,$FB,$2C,$C9,$FE,$E9,$F7,$B8,$46,$A3,$94,$62
	.dc.b	$7F,$DD,$6D,$8E,$07,$AC,$FE,$33,$C0,$E0,$F5,$47,$7D,$7E,$43,$FE
	.dc.b	$F8,$8C,$BD,$8E,$47,$A6,$79,$38,$35,$3C,$A3,$2F,$D7,$E2,$70,$3C
	.dc.b	$9E,$3A,$7F,$B1,$63,$FC,$AF,$8F,$E5,$67,$2F,$E5,$7B,$EE,$5F,$E8
	.dc.b	$6A,$E5,$3F,$9B,$9C,$E1,$DA,$ED,$74,$BF,$CA,$46,$2B,$2D,$BF,$DA
	.dc.b	$7A,$BC,$B9,$5E,$25,$8C,$36,$7F,$74,$DC,$3F,$6E,$DC,$FF,$DA,$2D
	.dc.b	$9F,$C0,$67,$FB,$F8,$C5,$7F,$A8,$30,$EF,$5E,$23,$A9,$FC,$C7,$E7
	.dc.b	$BD,$BD,$CF,$9E,$B1,$86,$CF,$F3,$76,$DF,$F3,$F6,$E5,$FD,$0D,$B3
	.dc.b	$FC,$ED,$EB,$E3,$76,$F8,$EB,$3F,$D0,$5B,$7F,$D2,$5B,$7F,$D3,$5F
	.dc.b	$2F,$C7,$79,$BE,$36,$DF,$F4,$96,$DF,$F5,$16,$DF,$F5,$56,$CF,$F4
	.dc.b	$D6,$CF,$FB,$13,$34,$63,$64,$62,$BF,$AB,$61,$E1,$7F,$AC,$7C,$6F
	.dc.b	$C8,$F8,$DF,$92,$D8,$C9,$FD,$76,$32,$D3,$FD,$81,$E8,$FE,$33,$D2
	.dc.b	$37,$D2,$C3,$27,$67,$FB,$1E,$31,$A7,$FB,$17,$E0,$7F,$2B,$FC,$01
	.dc.b	$FB,$1A,$FF,$66,$46,$1E,$CD,$72,$7F,$BF,$A3,$84,$FF,$A5,$3E,$63
	.dc.b	$C7,$3F,$E9,$D6,$C7,$FB,$FB,$E6,$5F,$46,$F9,$4F,$D4,$CA,$B7,$C7
	.dc.b	$FE,$6F,$78,$C7,$F8,$78,$D6,$9F,$E7,$37,$FC,$43,$1A,$6C,$65,$FE
	.dc.b	$39,$F2,$EE,$5E,$ED,$CF,$F9,$06,$1F,$C2,$33,$F2,$FB,$DA,$B9,$5E
	.dc.b	$1C,$39,$C7,$55,$87,$37,$4D,$7F,$C6,$BF,$30,$F8,$E7,$CA,$7F,$08
	.dc.b	$8C,$7C,$9D,$9C,$E3,$E4,$3F,$9A,$4E,$9C,$5A,$3A,$5F,$1B,$E4,$DA
	.dc.b	$DC,$E9,$B6,$AD,$C6,$1F,$8B,$DD,$23,$FD,$79,$F3,$2F,$96,$C6,$5F
	.dc.b	$ED,$2C,$FF,$C0,$D5,$F1,$63,$E1,$DF,$C5,$9C,$C7,$C4,$8F,$A9,$D8
	.dc.b	$F8,$07,$CC,$7F,$26,$F8,$CF,$EA,$B8,$FF,$A4,$A3,$1E,$6F,$F2,$AA
	.dc.b	$DC,$FC,$2B,$5F,$0C,$5C,$9F,$D0,$A7,$B6,$3C,$AF,$E0,$E5,$F3,$FE
	.dc.b	$63,$E8,$36,$32,$7C,$68,$CB,$4C,$E1,$D9,$FA,$BC,$09,$DC,$E4,$FA
	.dc.b	$D1,$99,$FD,$47,$B7,$FB,$0F,$6B,$F6,$2D,$8C,$9F,$AD,$8C,$B4,$FC
	.dc.b	$0D,$23,$FB,$EC,$61,$B1,$E1,$C9,$D8,$F6,$FF,$5F,$5A,$D7,$FD,$B8
	.dc.b	$F0,$BF,$D7,$35,$FF,$46,$B5,$FF,$03,$36,$FA,$FB,$5F,$C2,$BD,$BF
	.dc.b	$E1,$6D,$CB,$F8,$33,$6F,$C9,$92,$31,$8F,$D7,$8C,$57,$F0,$DB,$E6
	.dc.b	$D3,$F4,$3E,$8B,$F3,$1F,$D3,$E3,$F3,$E3,$1D,$1F,$DD,$1D,$75,$5A
	.dc.b	$FD,$05,$F2,$78,$E5,$FE,$08,$AE,$33,$9A,$E1,$7F,$C9,$BD,$BF,$E4
	.dc.b	$FE,$E5,$AF,$E4,$4D,$CB,$DA,$DC,$7D,$8F,$EF,$79,$7F,$80,$47,$E5
	.dc.b	$CF,$F0,$BF,$67,$F0,$58,$CF,$4E,$F0,$FD,$CD,$6D,$FC,$7C,$E9,$FD
	.dc.b	$DB,$70,$FC,$85,$F3,$9E,$37,$FA,$DF,$88,$91,$FC,$E7,$CF,$B7,$F8
	.dc.b	$7E,$A7,$81,$FB,$3F,$6B,$F6,$8D,$BF,$E8,$1E,$87,$F4,$83,$17,$F7
	.dc.b	$1F,$E9,$7E,$84,$9F,$E8,$DE,$1D,$AF,$C0,$CF,$F5,$0F,$3B,$1C,$1F
	.dc.b	$81,$71,$7F,$D3,$5C,$BF,$E1,$6E,$84,$7F,$07,$F6,$7F,$84,$7C,$A5
	.dc.b	$AF,$E0,$9B,$23,$19,$7B,$5B,$8F,$F0,$59,$D5,$70,$BE,$6D,$39,$D7
	.dc.b	$F6,$6F,$4B,$FB,$3F,$A3,$67,$FB,$13,$71,$FC,$EB,$71,$FC,$E3,$67
	.dc.b	$D4,$D9,$77,$7E,$DA,$3F,$91,$7C,$97,$F2,$3F,$99,$B7,$EC,$6D,$FD
	.dc.b	$76,$E1,$EB,$6C,$FF,$8F,$34,$46,$2B,$BC,$62,$7F,$BD,$7A,$7F,$E0
	.dc.b	$2E,$DF,$F0,$1B,$63,$27,$F7,$F8,$CB,$4C,$D7,$EF,$23,$2D,$8F,$A7
	.dc.b	$F3,$3F,$CF,$7E,$67,$F9,$F3,$63,$27,$D8,$8C,$B4,$CC,$FE,$56,$32
	.dc.b	$D8,$FF,$09,$77,$FF,$86,$BC,$9F,$F0,$DB,$63,$27,$F8,$5E,$32,$D3
	.dc.b	$FA,$57,$AD,$FC,$09,$FF,$31,$B6,$BF,$C7,$D5,$FE,$5E,$C6,$57,$07
	.dc.b	$FC,$C3,$18,$9F,$E6,$15,$FD,$83,$5F,$EA,$4F,$F4,$18,$CE,$1F,$E6
	.dc.b	$8F,$2B,$FC,$D2,$DC,$7C,$6D,$9F,$F7,$7F,$B3,$FE,$9C,$F7,$5F,$2F
	.dc.b	$FE,$78,$AF,$F4,$7D,$27,$2F,$F9,$E7,$DC,$E1,$5B,$BB,$3F,$E7,$D6
	.dc.b	$F2,$FF,$BF,$DB,$87,$FA,$57,$0E,$73,$9C,$7F,$D7,$B8,$73,$6E,$1C
	.dc.b	$DB,$2B,$E5,$5B,$AF,$33,$F3,$BC,$4F,$0F,$EE,$FF,$DE,$71,$EF,$AF
	.dc.b	$F7,$EB,$D9,$FA,$69,$D5,$F1,$CB,$BC,$62,$7F,$CE,$98,$74,$F1,$3D
	.dc.b	$F0,$F9,$9F,$EC,$5E,$4F,$F8,$5D,$B1,$FE,$3D,$6B,$FE,$4F,$6B,$FF
	.dc.b	$00,$3B,$3F,$C7,$4F,$F9,$0F,$E0,$FF,$87,$7A,$3F,$E2,$16,$C6,$67
	.dc.b	$6F,$3F,$3B,$FD,$E2,$F7,$6B,$9C,$7C,$91,$88,$FB,$DF,$11,$FE,$F5
	.dc.b	$7E,$1F,$17,$CD,$F8,$8C,$71,$7D,$6E,$34,$63,$8B,$8D,$FD,$49,$61
	.dc.b	$98,$FF,$5F,$55,$C3,$A4,$6E,$71,$F3,$3F,$CA,$E3,$EE,$B5,$F1,$B7
	.dc.b	$1F,$C3,$AF,$03,$8E,$B2,$B8,$BD,$23,$11,$CA,$2D,$72,$9F,$26,$4F
	.dc.b	$89,$F8,$5D,$5F,$E3,$D8,$FF,$50,$E1,$FE,$43,$6B,$FE,$21,$78,$39
	.dc.b	$9E,$3C,$DF,$A9,$F6,$EC,$FF,$B7,$DB,$AF,$E6,$6A,$DB,$E0,$29,$2F
	.dc.b	$C1,$F0,$15,$F0,$45,$CB,$CD,$3D,$3C,$37,$CC,$DF,$99,$98,$7F,$58
	.dc.b	$8D,$18,$7F,$BA,$46,$1F,$F3,$07,$CC,$3E,$47,$CE,$BE,$36,$98,$89
	.dc.b	$3C,$23,$2F,$28,$CB,$FE,$47,$9E,$4E,$CD,$1F,$DB,$3E,$67,$E2,$FC
	.dc.b	$3F,$FA,$CF,$4F,$F3,$73,$78,$BF,$A8,$4F,$BC,$DD,$43,$91,$0F,$F2
	.dc.b	$B4,$F1,$7F,$B1,$B7,$87,$FC,$FB,$3E,$EF,$8E,$F7,$F8,$92,$FE,$1C
	.dc.b	$65,$F8,$B1,$97,$D9,$19,$7D,$D1,$9C,$7F,$D0,$D9,$2B,$DB,$E0,$FB
	.dc.b	$BE,$2A,$61,$FE,$8F,$CD,$3A,$8C,$4E,$9F,$5C,$65,$F6,$C6,$70,$F8
	.dc.b	$F3,$F6,$3C,$6F,$B3,$D9,$2F,$EE,$C6,$5F,$BF,$19,$7E,$B4,$65,$FB
	.dc.b	$11,$9A,$F5,$CF,$D6,$F8,$A9,$5E,$97,$EC,$FC,$CC,$8C,$10,$FA,$6E
	.dc.b	$5A,$7C,$17,$F0,$3F,$33,$F8,$2F,$90,$93,$F1,$5F,$C9,$F8,$3F,$94
	.dc.b	$EC,$B3,$F9,$88,$C3,$F9,$AE,$DF,$CC,$91,$F9,$DF,$2B,$FB,$45,$F3
	.dc.b	$2F,$87,$C3,$FE,$A1,$8F,$F5,$1B,$5F,$F5,$81,$C2,$36,$70,$F8,$69
	.dc.b	$1C,$23,$2F,$28,$CB,$CA,$7C,$0F,$E0,$7E,$67,$F4,$3F,$2F,$23,$04
	.dc.b	$3F,$36,$E5,$FC,$EC,$7F,$0E,$E3,$7C,$CF,$F9,$BA,$7F,$D3,$6B,$E1
	.dc.b	$34,$D8,$CF,$C1,$60,$CC,$C8,$C9,$F9,$5F,$99,$FD,$AF,$81,$23,$04
	.dc.b	$3F,$5E,$E5,$FD,$8C,$7E,$F0,$C7,$3F,$EF,$27,$13,$FE,$D0,$F9,$97
	.dc.b	$07,$F2,$6F,$A8,$FE,$B3,$7F,$EE,$F6,$F8,$5F,$D5,$EB,$77,$C6,$FF
	.dc.b	$AD,$32,$32,$E0,$38,$75,$19,$7F,$82,$F7,$7F,$12,$F0,$24,$7E,$D6
	.dc.b	$3F,$8A,$98,$7F,$86,$C6,$1F,$F7,$4F,$0F,$F4,$6C,$98,$FF,$6D,$8F
	.dc.b	$E7,$D3,$AF,$23,$F2,$51,$B9,$D7,$36,$26,$57,$48,$D9,$87,$F2,$75
	.dc.b	$D1,$B1,$FD,$3B,$E3,$BF,$31,$87,$8A,$7F,$D0,$71,$FE,$F7,$47,$FB
	.dc.b	$ED,$C3,$79,$B9,$1C,$FE,$3A,$47,$0A,$E9,$E2,$AF,$BE,$7F,$57,$87
	.dc.b	$DB,$D5,$30,$FB,$D1,$FB,$1E,$1F,$E9,$D1,$F4,$5A,$FD,$76,$BF,$6E
	.dc.b	$3E,$B2,$D7,$79,$7B,$63,$1E,$37,$CC,$3E,$6D,$48,$FA,$B1,$97,$EB
	.dc.b	$C6,$5F,$97,$19,$7E,$6C,$79,$FB,$3E,$BF,$74,$9F,$CC,$46,$5F,$53
	.dc.b	$E6,$C7,$FC,$0B,$1E,$B8,$F4,$FC,$D3,$E1,$BF,$89,$7C,$D3,$E5,$A9
	.dc.b	$FC,$56,$75,$7F,$3F,$7F,$F1,$B4,$62,$3F,$38,$D7,$F4,$2D,$77,$7F
	.dc.b	$9D,$E1,$79,$0F,$1B,$FB,$7F,$8F,$2F,$FA,$94,$7F,$72,$30,$FF,$6A
	.dc.b	$8C,$3F,$D5,$F8,$FF,$BB,$91,$FE,$B5,$AD,$7F,$B8,$C5,$E2,$FE,$F7
	.dc.b	$3F,$E2,$2F,$73,$EC,$47,$FA,$6B,$C5,$C1,$9D,$A6,$1F,$EA,$A8,$CB
	.dc.b	$FE,$AC,$8E,$5D,$1C,$FA,$24,$FF,$AC,$23,$2F,$FA,$CA,$32,$FF,$AA
	.dc.b	$A3,$2F,$FA,$B2,$33,$7F,$EA,$C9,$CB,$FE,$AE,$8C,$D7,$FA,$EA,$7D
	.dc.b	$5E,$B7,$AF,$E4,$66,$1F,$EC,$08,$CB,$FE,$C2,$8C,$BF,$EB,$A8,$F8
	.dc.b	$5F,$21,$F0,$FE,$42,$4F,$FB,$22,$32,$FF,$B2,$A3,$2F,$FB,$0A,$32
	.dc.b	$FF,$B1,$23,$39,$7F,$B1,$27,$FC,$9B,$3F,$E4,$EF,$1E,$3F,$E5,$B3
	.dc.b	$27,$F9,$75,$CB,$FE,$5D,$32,$7F,$97,$DC,$CF,$F9,$96,$32,$FF,$99
	.dc.b	$E3,$39,$7F,$B2,$67,$FC,$BF,$F3,$1F,$E4,$05,$5F,$C2,$23,$FC,$49
	.dc.b	$1F,$9D,$1C,$88,$CC,$70,$AF,$E2,$F3,$9E,$CE,$A7,$6F,$0A,$57,$39
	.dc.b	$D3,$6F,$C5,$87,$86,$B7,$3D,$4F,$0C,$EB,$0F,$F5,$27,$B8,$F8,$AF
	.dc.b	$4D,$3E,$65,$E7,$6B,$E4,$3F,$D5,$51,$89,$FE,$5B,$69,$CD,$74,$BE
	.dc.b	$77,$BB,$5F,$7C,$FF,$9A,$CE,$63,$94,$E2,$FC,$0D,$3B,$DF,$48,$ED
	.dc.b	$7F,$CF,$BD,$F7,$CD,$62,$D1,$87,$8C,$EE,$75,$39,$9F,$F6,$19,$C2
	.dc.b	$3F,$B6,$7B,$BB,$8F,$EE,$2D,$77,$3C,$5F,$E9,$F1,$E7,$7F,$A8,$F0
	.dc.b	$FF,$55,$CF,$E1,$46,$8F,$9D,$8F,$F7,$C9,$D3,$E0,$8C,$9F,$ED,$98
	.dc.b	$CB,$E2,$32,$76,$D7,$E3,$70,$7E,$37,$CA,$FE,$C9,$A7,$E4,$FB,$AD
	.dc.b	$7E,$3B,$85,$58,$C3,$F0,$B5,$70,$E5,$7F,$8D,$8C,$5F,$08,$C1,$F9
	.dc.b	$1B,$CC,$E9,$B1,$FE,$CD,$F5,$7A,$7B,$8F,$5A,$C6,$1B,$3F,$ED,$36
	.dc.b	$E5,$FE,$D3,$6E,$7F,$ED,$96,$CF,$EA,$2F,$A3,$DC,$E4,$5A,$E4,$FE
	.dc.b	$A7,$B9,$F2,$5E,$F5,$F2,$56,$30,$D9,$E0,$DA,$FD,$66,$CB,$E3,$5D
	.dc.b	$6F,$8B,$4E,$27,$87,$C5,$F0,$7C,$97,$85,$B1,$93,$CD,$19,$69,$E9
	.dc.b	$F9,$1F,$47,$C8,$FA,$5B,$19,$3D,$D1,$96,$99,$CB,$D3,$8A,$75,$18
	.dc.b	$6E,$DF,$E2,$29,$D5,$71,$8C,$4F,$C9,$F9,$1F,$95,$F2,$16,$BE,$73
	.dc.b	$70,$EE,$DB,$E8,$FE,$3F,$CA,$7A,$74,$C7,$A9,$70,$FF,$65,$45,$C3
	.dc.b	$FD,$C3,$37,$0F,$F7,$5F,$C1,$FE,$67,$E4,$BF,$CA,$B8,$7F,$B4,$63
	.dc.b	$0F,$76,$B9,$3F,$96,$47,$0B,$FF,$51,$FC,$CF,$92,$7F,$CD,$4D,$8F
	.dc.b	$F3,$83,$5F,$F3,$AB,$5F,$F2,$88,$78,$5F,$ED,$3E,$75,$CD,$FB,$1A
	.dc.b	$D6,$7F,$B3,$A5,$6B,$0F,$E9,$54,$F9,$77,$1F,$F9,$16,$3B,$C6,$27
	.dc.b	$FD,$00,$D8,$FF,$43,$35,$FF,$45,$B9,$BF,$F3,$57,$6B,$E6,$5F,$19
	.dc.b	$E5,$AD,$61,$FC,$72,$72,$FF,$5D,$F9,$77,$C7,$3E,$53,$F8,$64,$63
	.dc.b	$E5,$2C,$E6,$B8,$28,$CD,$69,$B3,$A6,$E9,$FD,$6E,$FA,$4F,$2A,$E0
	.dc.b	$FC,$6C,$7E,$45,$7C,$0E,$0F,$8C,$FF,$A1,$23,$2F,$C2,$8C,$BF,$35
	.dc.b	$FE,$49,$2F,$9D,$7B,$DF,$0E,$73,$18,$9F,$EE,$9F,$00,$F9,$3F,$E5
	.dc.b	$5F,$1F,$FD,$4A,$73,$18,$F3,$3F,$97,$56,$E7,$FB,$E5,$6F,$2F,$F4
	.dc.b	$C5,$EE,$7E,$3E,$3F,$2B,$E5,$7E,$5B,$63,$27,$CC,$8C,$B4,$CD,$7C
	.dc.b	$9E,$04,$EA,$30,$FE,$A7,$DA,$FD,$57,$B3,$6F,$F4,$AD,$AF,$E4,$3D
	.dc.b	$68,$C7,$3F,$FA,$A2,$31,$5F,$98,$9F,$EE,$AF,$FB,$4D,$B5,$FE,$D9
	.dc.b	$6B,$FE,$DE,$6B,$FA,$A7,$75,$FE,$E7,$F4,$3F,$51,$18,$AF,$F1,$4F
	.dc.b	$C2,$DB,$FF,$17,$E0,$F6,$7F,$82,$7B,$5F,$C1,$5B,$7F,$C0,$1B,$5E
	.dc.b	$C6,$E5,$EB,$6C,$FF,$8D,$AF,$A3,$FC,$22,$FF,$B5,$75,$BE,$67,$F3
	.dc.b	$D8,$FF,$8C,$63,$1D,$5A,$9C,$E5,$FB,$07,$F4,$58,$7F,$91,$EB,$97
	.dc.b	$4E,$EF,$83,$67,$F8,$BF,$B5,$FC,$7F,$DB,$FE,$40,$D8,$C9,$FC,$72
	.dc.b	$32,$D3,$F6,$3C,$BF,$B6,$CB,$F2,$07,$FB,$AF,$B3,$F7,$71,$9E,$9F
	.dc.b	$DC,$B6,$B8,$9F,$C0,$6B,$8B,$FE,$F0,$AE,$2E,$6B,$8B,$4F,$CF,$FC
	.dc.b	$47,$E9,$8F,$E7,$1E,$6D,$8F,$D1,$FC,$4F,$EA,$BD,$BF,$D5,$B6,$3F
	.dc.b	$9B,$FA,$3F,$CF,$CC,$71,$7E,$51,$FE,$85,$E8,$49,$FE,$7B,$E2,$DA
	.dc.b	$FC,$CC,$FF,$47,$F4,$31,$B7,$E6,$5C,$5F,$F4,$47,$2F,$E0,$5F,$DE
	.dc.b	$FB,$5F,$BE,$F9,$3B,$3F,$BA,$6E,$1F,$E0,$6F,$09,$18,$C1,$5F,$E8
	.dc.b	$09,$CB,$FC,$0A,$FA,$34,$E8,$7F,$56,$F4,$7F,$AF,$7A,$3F,$D7,$DB
	.dc.b	$19,$3F,$AD,$C6,$5A,$67,$1F,$DC,$C6,$5B,$1F,$C4,$BE,$4F,$F8,$CF
	.dc.b	$CC,$FF,$1A,$6C,$64,$FE,$2F,$19,$69,$99,$F8,$11,$96,$C7,$F6,$EF
	.dc.b	$4F,$FB,$B7,$57,$F7,$76,$C6,$4F,$EE,$71,$96,$99,$9F,$E2,$31,$96
	.dc.b	$C7,$F2,$AF,$99,$FE,$63,$F3,$3F,$CC,$9B,$19,$3F,$97,$C6,$5A,$66
	.dc.b	$7E,$C4,$65,$B1,$FE,$06,$ED,$FF,$07,$79,$3F,$E0,$F6,$C6,$4F,$F0
	.dc.b	$5C,$65,$A7,$F9,$47,$0F,$F2,$81,$FE,$51,$9B,$19,$25,$7F,$8D,$A3
	.dc.b	$2F,$F8,$E2,$BF,$CB,$1E,$47,$F9,$65,$B5,$FE,$5D,$6C,$EE,$3F,$CC
	.dc.b	$13,$FE,$72,$F6,$BF,$23,$2F,$7F,$FB,$FB,$FE,$D6,$BF,$F3,$F3,$9F
	.dc.b	$47,$FC,$DD,$EE,$70,$AD,$FB,$5F,$E1,$AC,$31,$5F,$E7,$36,$DF,$36
	.dc.b	$CF,$F8,$82,$EE,$1D,$B0,$E6,$DC,$39,$65,$FE,$AC,$7C,$2D,$C7,$A3
	.dc.b	$70,$FF,$39,$78,$5D,$BE,$67,$FB,$56,$3E,$6F,$6F,$FB,$76,$1E,$B3
	.dc.b	$FD,$E7,$3A,$C7,$FD,$E3,$3A,$AF,$6C,$E6,$BF,$DF,$33,$9A,$FF,$7B
	.dc.b	$CE,$65,$F3,$2F,$85,$79,$3F,$E4,$29,$FF,$21,$CE,$63,$10,$38,$CF
	.dc.b	$07,$33,$B7,$7F,$07,$FC,$2B,$D3,$E3,$1F,$EE,$CE,$FF,$F1,$8B,$A6
	.dc.b	$BE,$19,$E0,$F4,$F3,$FF,$DC,$AD,$F2,$79,$E1,$FD,$9D,$B3,$A9,$F5
	.dc.b	$4E,$5F,$94,$7C,$08,$E7,$B7,$E2,$A3,$15,$F5,$FF,$0B,$7E,$60,$F9
	.dc.b	$AF,$F3,$2D,$EF,$1E,$8D,$C7,$9F,$07,$C8,$6E,$4C,$79,$E1,$CA,$B5
	.dc.b	$97,$E1,$6B,$C2,$79,$B1,$EB,$3F,$3E,$F8,$CE,$E7,$2F,$D3,$6B,$CB
	.dc.b	$1F,$D2,$CF,$08,$E9,$8F,$91,$D5,$F8,$9F,$C7,$4E,$EB,$F1,$7E,$AF
	.dc.b	$FA,$13,$C5,$FD,$8F,$2B,$8D,$39,$E6,$FC,$6B,$71,$FC,$BF,$C6,$DB
	.dc.b	$E1,$36,$F8,$D7,$E2,$F0,$EB,$18,$8F,$F6,$BA,$BE,$19,$F1,$F9,$C9
	.dc.b	$5C,$72,$FC,$EC,$EA,$30,$7E,$A2,$76,$F4,$8F,$32,$FC,$21,$E1,$7E
	.dc.b	$5A,$30,$D8,$CC,$7E,$69,$AF,$E9,$1B,$97,$F1,$68,$E2,$7F,$A8,$A3
	.dc.b	$38,$7E,$CB,$CD,$F8,$DF,$3C,$FC,$21,$F2,$FF,$26,$9F,$1D,$FD,$CB
	.dc.b	$1F,$5C,$63,$AB,$8D,$F0,$9F,$7E,$1C,$F0,$FD,$7B,$63,$F7,$2D,$7F
	.dc.b	$7C,$D7,$FD,$82,$E7,$3F,$E5,$ED,$C3,$F8,$03,$67,$F8,$9F,$A9,$E2
	.dc.b	$F0,$FE,$E7,$A7,$27,$C7,$92,$75,$1E,$28,$C3,$63,$CF,$18,$8F,$2B
	.dc.b	$71,$F1,$B7,$1F,$1B,$67,$F9,$2F,$9D,$F1,$3E,$3F,$F1,$DA,$C9,$F1
	.dc.b	$C6,$5F,$13,$E0,$23,$83,$64,$3F,$27,$E6,$FD,$8F,$DA,$DE,$FF,$E6
	.dc.b	$FC,$F9,$5F,$4A,$72,$D9,$FA,$71,$88,$FA,$2D,$7E,$8B,$5F,$9E,$EF
	.dc.b	$4F,$A6,$DD,$33,$5F,$AB,$F3,$FC,$9E,$D7,$EB,$7D,$09,$8F,$8E,$32
	.dc.b	$FE,$0B,$04,$79,$09,$19,$76,$B1,$F0,$C7,$EE,$7E,$AF,$F7,$5E,$A4
	.dc.b	$AF,$D7,$F8,$1F,$C1,$63,$E3,$7A,$F2,$7C,$0F,$C2,$23,$FD,$32,$7F
	.dc.b	$86,$7D,$4E,$E3,$C1,$B7,$C2,$AF,$E1,$13,$FA,$8F,$D7,$9F,$55,$49
	.dc.b	$7E,$2F,$0F,$2F,$F0,$21,$FA,$6B,$83,$F7,$CD,$8F,$EA,$AD,$7F,$AF
	.dc.b	$B5,$FE,$C0,$7F,$72,$F2,$FF,$8D,$11,$F0,$7D,$27,$F0,$C9,$65,$FC
	.dc.b	$7F,$E9,$25,$78,$E7,$2D,$9F,$19,$1F,$1B,$5F,$86,$7F,$71,$8E,$2E
	.dc.b	$0F,$6C,$7F,$2F,$FC,$9B,$F8,$44,$70,$DD,$FB,$BF,$F7,$DE,$4F,$E7
	.dc.b	$7F,$0F,$FC,$B6,$BE,$CB,$5E,$13,$BE,$2F,$F0,$05,$F2,$9F,$E7,$95
	.dc.b	$BB,$E7,$E5,$7F,$3F,$C3,$85,$2D,$6F,$2B,$5E,$AC,$AF,$0C,$74,$30
	.dc.b	$F8,$A3,$0F,$E3,$AD,$F2,$BE,$2F,$16,$4E,$63,$C4,$61,$F8,$58,$27
	.dc.b	$31,$B3,$E0,$BB,$CF,$FA,$FD,$F6,$9F,$E5,$BF,$86,$FF,$81,$E7,$3F
	.dc.b	$2D,$EF,$3F,$9D,$C6,$76,$FE,$4E,$DA,$79,$08,$C3,$7C,$EF,$91,$F4
	.dc.b	$DF,$27,$E8,$7E,$6F,$8B,$30,$F8,$11,$97,$EB,$C6,$5F,$AF,$1F,$D7
	.dc.b	$3F,$0B,$67,$FD,$CF,$85,$F9,$43,$CA,$C3,$70,$FD,$35,$F0,$6D,$FF
	.dc.b	$63,$8C,$47,$F9,$09,$AF,$F9,$1D,$AF,$E9,$9D,$EF,$FD,$8D,$BB,$E6
	.dc.b	$BC,$DA,$7E,$2B,$C6,$FC,$5F,$74,$C7,$F2,$71,$97,$F2,$B1,$97,$ED
	.dc.b	$C7,$D0,$DB,$F2,$7E,$24,$9F,$AF,$19,$69,$0F,$C1,$F8,$9F,$1F,$E7
	.dc.b	$3E,$47,$0C,$9F,$87,$19,$7F,$CB,$3A,$A3,$E2,$12,$32,$ED,$FD,$A4
	.dc.b	$6D,$F9,$BF,$4B,$F8,$6C,$FF,$53,$DD,$27,$F1,$B1,$96,$90,$FC,$CF
	.dc.b	$89,$F8,$E9,$FD,$8F,$87,$27,$F2,$B1,$96,$90,$FD,$0F,$87,$FE,$20
	.dc.b	$FA,$27,$C5,$7F,$87,$DB,$3F,$E7,$66,$BF,$E8,$06,$BF,$E8,$57,$3C
	.dc.b	$BF,$E1,$F6,$F2,$FF,$26,$6C,$FF,$A1,$FE,$8F,$FC,$E9,$5F,$E2,$56
	.dc.b	$D7,$F2,$C6,$CE,$FD,$DF,$F2,$55,$7F,$08,$F3,$FF,$C6,$F5,$AA,$34
	.dc.b	$E3,$3B,$8F,$E8,$1F,$AB,$BE,$17,$FC,$78,$D8,$FF,$49,$B5,$FF,$4E
	.dc.b	$B5,$FF,$51,$39,$9F,$F1,$E3,$71,$FE,$86,$D9,$FF,$52,$7E,$AF,$FC
	.dc.b	$C2,$BF,$C5,$3E,$0B,$CB,$FF,$19,$B6,$BF,$C6,$6D,$7F,$C6,$6D,$7F
	.dc.b	$C6,$6E,$67,$9C,$FF,$0E,$3F,$CD,$9F,$BC,$3E,$2F,$17,$FC,$5A,$6B
	.dc.b	$FD,$06,$74,$E4,$87,$F4,$CF,$13,$F8,$A4,$F4,$AF,$F2,$AF,$B5,$FC
	.dc.b	$86,$BF,$C8,$ED,$7F,$C8,$ED,$7F,$C8,$AD,$7F,$8F,$BC,$70,$FF,$25
	.dc.b	$B7,$4D,$F9,$3F,$D1,$FA,$FF,$D1,$1C,$4F,$C9,$BF,$CD,$58,$7F,$32
	.dc.b	$EB,$FE,$98,$6A,$33,$87,$FB,$4E,$B8,$3F,$EC,$E6,$E5,$FE,$D8,$6B
	.dc.b	$FE,$DD,$6B,$FD,$F1,$CD,$A7,$A5,$76,$A5,$6A,$7F,$B1,$FE,$09,$F8
	.dc.b	$18,$F4,$7C,$1F,$F6,$1F,$AC,$97,$FE,$92,$9C,$B7,$05,$79,$DB,$1E
	.dc.b	$76,$BE,$76,$BE,$77,$38,$F9,$E3,$CB,$18,$C7,$D3,$5F,$DC,$FE,$AF
	.dc.b	$E4,$32,$BC,$4E,$5F,$27,$57,$F7,$9F,$7E,$46,$08,$78,$E3,$D0,$47
	.dc.b	$D0,$FD,$52,$3F,$DD,$FE,$77,$FD,$5F,$E0,$FF,$0A,$8F,$F1,$5B,$51
	.dc.b	$FE,$35,$9F,$F5,$94,$66,$3F,$C7,$1E,$CE,$08,$FF,$8D,$0E,$6F,$FA
	.dc.b	$E7,$E2,$7F,$AF,$6B,$F8,$5F,$85,$FF,$1C,$C6,$2B,$ED,$C6,$3C,$EF
	.dc.b	$F2,$F4,$60,$F1,$CE,$F8,$3F,$DA,$91,$E6,$F8,$6E,$8F,$85,$AF,$F9
	.dc.b	$19,$AE,$CE,$87,$F9,$47,$E6,$FF,$D8,$CF,$C2,$6B,$F0,$B3,$46,$23
	.dc.b	$87,$0F,$BD,$D7,$CB,$66,$8D,$24,$FE,$2F,$E2,$3F,$0D,$AF,$E1,$F2
	.dc.b	$FD,$74,$62,$7F,$C7,$BE,$AF,$FB,$41,$FC,$B4,$67,$1F,$C6,$BE,$6A
	.dc.b	$E3,$3D,$49,$3F,$97,$9E,$1E,$CF,$F9,$43,$EA,$39,$5F,$0A,$FF,$26
	.dc.b	$C6,$3F,$20,$FD,$6E,$1F,$A7,$6C,$FF,$94,$BE,$1F,$57,$2B,$4F,$6C
	.dc.b	$BF,$D9,$D1,$82,$57,$EF,$EF,$9B,$FB,$F6,$BF,$C2,$5A,$FE,$F9,$F0
	.dc.b	$E7,$FE,$42,$F9,$2F,$04,$FF,$99,$1F,$47,$BB,$2F,$FD,$43,$F5,$6F
	.dc.b	$95,$C5,$FF,$9E,$33,$65,$FE,$61,$BF,$E2,$8D,$9F,$36,$1F,$36,$31
	.dc.b	$1E,$78,$F7,$61,$D3,$2F,$F2,$AB,$F5,$7C,$67,$88,$F6,$70,$DC,$BF
	.dc.b	$CD,$1F,$57,$89,$91,$F8,$BA,$E3,$1F,$92,$6C,$66,$B9,$65,$FD,$0B
	.dc.b	$C2,$EA,$7D,$EF,$1D,$E2,$E2,$FF,$77,$A7,$F4,$0A,$C5,$FF,$A4,$EF
	.dc.b	$75,$A3,$FA,$4C,$EA,$B8,$47,$FB,$32,$77,$5F,$D5,$F0,$5F,$48,$FE
	.dc.b	$A6,$FA,$63,$35,$E9,$8F,$D5,$7B,$AF,$07,$F3,$96,$8C,$57,$FB,$1F
	.dc.b	$F1,$1F,$EF,$0F,$FB,$26,$33,$3F,$EA,$F6,$BF,$C4,$5A,$FF,$B1,$9A
	.dc.b	$FF,$1C,$6E,$8F,$B2,$7C,$1F,$D9,$6D,$FC,$3B,$7E,$92,$FB,$97,$CE
	.dc.b	$6F,$09,$7B,$C7,$A4,$7E,$DE,$F8,$56,$B0,$FF,$63,$7D,$53,$EA,$BE
	.dc.b	$0B,$71,$E5,$9F,$8A,$B7,$5A,$3F,$89,$56,$E3,$F8,$93,$5F,$E6,$4D
	.dc.b	$7F,$C0,$7E,$57,$E4,$8F,$E0,$7F,$11,$F3,$23,$ED,$46,$70,$F8,$1E
	.dc.b	$0B,$C2,$FE,$87,$1E,$DE,$45,$7F,$0C,$9C,$DF,$E0,$FC,$2F,$CF,$99
	.dc.b	$E2,$B8,$71,$8E,$1E,$8E,$2B,$FD,$C9,$7F,$EC,$AA,$FA,$10,$FC,$07
	.dc.b	$E1,$74,$FF,$6D,$8C,$71,$FE,$22,$5E,$5D,$89,$7F,$CD,$6B,$6D,$AF
	.dc.b	$99,$1F,$D5,$DA,$FF,$60,$36,$7F,$BE,$D7,$C4,$74,$8C,$3F,$BC,$EF
	.dc.b	$95,$F2,$32,$FF,$5F,$BF,$1E,$33,$5E,$EE,$EB,$9F,$F8,$1F,$87,$E7
	.dc.b	$47,$CB,$F8,$2F,$99,$EA,$5C,$7A,$37,$0F,$8E,$DC,$39,$65,$FE,$A3
	.dc.b	$AF,$93,$A3,$E5,$7E,$31,$F1,$FC,$3B,$3A,$8F,$EB,$93,$B8,$F8,$D1
	.dc.b	$82,$3F,$EF,$FF,$77,$8E,$4C,$7F,$DE,$B1,$FD,$07,$A7,$F0,$9E,$1D
	.dc.b	$BF,$88,$DB,$FB,$ED,$BF,$4B,$FC,$A7,$F6,$A7,$C9,$53,$FA,$67,$D7
	.dc.b	$5C,$3F,$5D,$3A,$7F,$2F,$E6,$3E,$BB,$59,$79,$67,$FB,$5C,$7F,$62
	.dc.b	$CF,$FC,$99,$1F,$5F,$C6,$C3,$F9,$B6,$B9,$3F,$47,$1C,$2B,$F2,$F3
	.dc.b	$A6,$CE,$8E,$47,$E0,$FA,$FC,$19,$FE,$9D,$B1,$93,$AC,$65,$A7,$83
	.dc.b	$D1,$FD,$77,$9D,$E9,$6C,$64,$FC,$84,$65,$A7,$E0,$BD,$1F,$D3,$1F
	.dc.b	$E0,$2F,$8F,$7C,$66,$27,$47,$E9,$67,$F0,$64,$70,$7C,$2F,$9E,$E1
	.dc.b	$9F,$CD,$1D,$9F,$C6,$E3,$E7,$7B,$52,$BE,$4F,$D4,$5B,$FD,$67,$C4
	.dc.b	$7E,$40,$93,$FB,$57,$FB,$07,$D2,$32,$FE,$D8,$D9,$FA,$7E,$E6,$27
	.dc.b	$EE,$7D,$3F,$E2,$7D,$4B,$5F,$86,$70,$FF,$6D,$E8,$FF,$5B,$36,$3F
	.dc.b	$B7,$46,$32,$FF,$2D,$6D,$FC,$D7,$DA,$FC,$41,$CA,$3E,$04,$66,$FF
	.dc.b	$A0,$7E,$D3,$FE,$16,$7F,$CD,$3E,$9A,$31,$5F,$1B,$A5,$F4,$FF,$E0
	.dc.b	$79,$CD,$6F,$DA,$FF,$00,$47,$F8,$8A,$31,$87,$F8,$16,$30,$74,$C7
	.dc.b	$FA,$57,$C0,$73,$BF,$CF,$3D,$7C,$39,$7F,$9A,$F9,$9B,$A7,$FB,$3F
	.dc.b	$1F,$E0,$0F,$F9,$E7,$2F,$45,$F8,$89,$7E,$17,$F5,$5C,$7C,$BC,$3F
	.dc.b	$E3,$5B,$23,$D9,$F2,$52,$FF,$53,$EB,$DC,$BC,$13,$EE,$24,$FB,$5B
	.dc.b	$1F,$3A,$DF,$4F,$C7,$1E,$EF,$C9,$F9,$49,$87,$ED,$BC,$BB,$87,$B6
	.dc.b	$7E,$59,$27,$E5,$36,$3B,$73,$70,$C3,$D6,$FD,$5E,$A9,$7F,$C0,$BE
	.dc.b	$02,$DF,$C9,$9F,$AE,$49,$FA,$CD,$8F,$D3,$43,$F7,$1F,$F3,$2C,$71
	.dc.b	$8F,$F0,$71,$9F,$82,$FF,$11,$FD,$0F,$E1,$A3,$FB,$1B,$67,$E2,$7C
	.dc.b	$7E,$BA,$A5,$7F,$5F,$AD,$57,$FA,$CD,$AF,$F5,$D6,$BF,$93,$FA,$8F
	.dc.b	$E4,$9F,$51,$FC,$95,$B5,$FC,$81,$B3,$F4,$DB,$B7,$F5,$9F,$DD,$3F
	.dc.b	$24,$4F,$6B,$FD,$1E,$E3,$B6,$98,$8B,$87,$F7,$EF,$A6,$78,$FF,$E4
	.dc.b	$1C,$9F,$13,$FD,$C1,$FF,$69,$CE,$BD,$EF,$F6,$07,$B9,$FE,$D6,$FA
	.dc.b	$67,$E3,$8F,$2F,$FD,$E5,$8F,$1C,$58,$F1,$7A,$E1,$FD,$FE,$31,$3B
	.dc.b	$6C,$75,$BE,$99,$F8,$2F,$C7,$9F,$9F,$0E,$0D,$BE,$8D,$96,$5E,$6C
	.dc.b	$BC,$B5,$AC,$FA,$57,$9A,$F8,$36,$3C,$17,$C7,$1D,$E7,$E6,$AF,$1D
	.dc.b	$62,$FC,$17,$B9,$F7,$9A,$29,$D4,$FF,$6D,$57,$C5,$7F,$DF,$4D,$7C
	.dc.b	$C7,$FB,$7F,$E3,$1F,$4A,$FA,$9F,$F1,$74,$61,$B1,$99,$E0,$D7,$AF
	.dc.b	$9E,$E1,$FF,$1B,$60,$8C,$4F,$91,$AF,$09,$DC,$6B,$7F,$B0,$DD,$BE
	.dc.b	$E3,$6B,$E8,$FD,$4B,$E7,$3F,$C7,$D7,$C3,$0F,$97,$7B,$BF,$C3,$BC
	.dc.b	$75,$61,$C6,$77,$51,$F9,$79,$7A,$63,$EB,$CE,$E7,$37,$D6,$B5,$18
	.dc.b	$AD,$E3,$E4,$CB,$D3,$1D,$72,$F3,$57,$DA,$CB,$C7,$7B,$8F,$EE,$3F
	.dc.b	$49,$FE,$62,$7F,$CF,$2D,$7F,$AB,$B7,$9B,$FC,$A0,$FF,$A8,$23,$7C
	.dc.b	$BF,$E6,$19,$CB,$78,$D3,$DA,$30,$F8,$3E,$16,$F9,$CC,$3F,$61,$87
	.dc.b	$6F,$81,$CC,$1B,$B4,$C3,$CF,$1A,$30,$FF,$83,$47,$67,$FA,$4D,$FF
	.dc.b	$4A,$F5,$7A,$5F,$EE,$6D,$C3,$FD,$0D,$3B,$9E,$4D,$7F,$92,$78,$BF
	.dc.b	$E8,$57,$AB,$63,$51,$FD,$EB,$EA,$3F,$D3,$EF,$F3,$6F,$89,$FD,$0B
	.dc.b	$79,$FE,$1F,$9C,$AD,$CF,$F8,$96,$31,$F1,$3F,$23,$89,$87,$2A,$FE
	.dc.b	$8A,$DC,$D3,$E1,$AE,$B8,$74,$9C,$DB,$0E,$AF,$FB,$46,$75,$7F,$63
	.dc.b	$83,$EA,$7D,$27,$2F,$17,$E9,$FD,$4C,$C3,$84,$65,$FC,$4C,$65,$F5
	.dc.b	$60,$9D,$46,$1B,$97,$E3,$3E,$57,$CF,$F0,$DF,$80,$DE,$5F,$96,$32
	.dc.b	$FE,$5A,$32,$F6,$8C,$B4,$E0,$7E,$73,$D0,$F9,$FF,$51,$FE,$86,$8F
	.dc.b	$CF,$FD,$4F,$E3,$8C,$FC,$14,$8C,$90,$FA,$93,$C5,$B1,$B6,$9F,$AA
	.dc.b	$F7,$FB,$FB,$FF,$9A,$F7,$E4,$FF,$BF,$ED,$3A,$8F,$D8,$CE,$9E,$F1
	.dc.b	$96,$9C,$5F,$D2,$FA,$9F,$A6,$FA,$59,$3F,$BA,$7D,$FE,$E4,$8C,$10
	.dc.b	$F5,$BF,$AE,$F5,$35,$18,$3F,$81,$BF,$96,$FA,$97,$D2,$D2,$4F,$F0
	.dc.b	$87,$F7,$AB,$EA,$7F,$CE,$4E,$48,$7E,$D5,$CB,$FC,$3B,$C6,$FD,$F9
	.dc.b	$1F,$E2,$7F,$4D,$D2,$58,$7F,$04,$FA,$09,$5F,$F0,$02,$72,$D9,$CB
	.dc.b	$FC,$6A,$79,$44,$8F,$F7,$F4,$F0,$7F,$DF,$47,$F1,$87,$0E,$1A,$FE
	.dc.b	$97,$E3,$37,$D1,$FE,$F3,$F9,$5F,$E1,$5D,$FF,$A9,$6D,$7F,$A7,$AB
	.dc.b	$73,$C6,$3F,$90,$CE,$BB,$CD,$AD,$CF,$FA,$9A,$3F,$91,$3F,$ED,$8F
	.dc.b	$53,$F9,$4C,$FF,$88,$63,$FD,$1B,$3A,$6F,$8F,$AA,$91,$FC,$D3,$1F
	.dc.b	$44,$F2,$F9,$5F,$1F,$85,$2B,$FD,$CB,$19,$78,$67,$ED,$9D,$47,$FA
	.dc.b	$73,$25,$6E,$3C,$73,$AC,$3F,$88,$E6,$BD,$D7,$43,$F8,$A7,$D1,$7F
	.dc.b	$42,$C7,$E7,$7D,$47,$B8,$FE,$4F,$19,$9D,$F9,$7F,$61,$FF,$54,$C7
	.dc.b	$F8,$E5,$FE,$A2,$68,$D7,$8C,$8D,$47,$DB,$F9,$5F,$A7,$E0,$C8,$F2
	.dc.b	$67,$E7,$9F,$9E,$66,$3E,$7C,$E1,$F5,$0F,$A8,$65,$FD,$83,$85,$F5
	.dc.b	$2F,$1B,$0D,$BE,$37,$C1,$B7,$AA,$91,$82,$1F,$06,$78,$36,$7F,$AB
	.dc.b	$46,$23,$E1,$E5,$F8,$EF,$0B,$F2,$3E,$0C,$9F,$7F,$37,$FB,$F9,$B3
	.dc.b	$97,$F1,$B3,$F9,$88,$FA,$7A,$7E,$6B,$C0,$95,$F6,$F8,$3E,$74,$EA
	.dc.b	$3E,$8F,$02,$3E,$C4,$7E,$01,$B3,$C2,$33,$9B,$E5,$FF,$86,$AE,$1A
	.dc.b	$62,$24,$FC,$A8,$CB,$F3,$23,$2F,$C7,$7E,$29,$1F,$2E,$7F,$B4,$EA
	.dc.b	$91,$EB,$CB,$F8,$BF,$37,$F6,$AC,$BF,$15,$7E,$89,$FE,$37,$5F,$90
	.dc.b	$9C,$BD,$0F,$83,$7F,$C0,$3B,$64,$75,$8F,$CE,$9C,$C8,$F0,$3E,$3F
	.dc.b	$D4,$BE,$33,$F9,$7F,$D4,$FF,$0E,$93,$67,$51,$F9,$46,$BF,$A7,$6B
	.dc.b	$FC,$81,$AF,$EA,$1A,$FF,$8E,$A3,$FC,$D7,$9F,$F8,$2A,$3B,$FB,$1E
	.dc.b	$23,$FC,$7B,$1F,$C6,$BC,$AC,$4F,$E9,$2B,$8F,$5F,$ED,$1B,$0A,$E0
	.dc.b	$FE,$AB,$EA,$5F,$0B,$FD,$31,$7C,$65,$31,$12,$7F,$17,$3C,$1B,$1F
	.dc.b	$E3,$78,$C4,$7F,$3E,$DF,$F8,$17,$89,$FE,$52,$F5,$BF,$86,$75,$9F
	.dc.b	$E1,$D3,$AA,$FC,$15,$7D,$F2,$3F,$9C,$6B,$FE,$39,$E3,$47,$F1,$83
	.dc.b	$FC,$BF,$F5,$0F,$8A,$C5,$74,$21,$F0,$F7,$FF,$19,$4E,$E3,$F5,$9E
	.dc.b	$B7,$F2,$87,$FC,$B3,$E3,$FA,$4F,$F2,$5C,$67,$83,$F2,$CE,$E7,$F2
	.dc.b	$8D,$7F,$CE,$B9,$FF,$93,$3C,$7F,$D8,$CE,$67,$FD,$5D,$F4,$FF,$7E
	.dc.b	$FF,$9C,$F8,$EF,$9B,$FE,$A2,$7F,$9F,$39,$D5,$BD,$7F,$D0,$31,$87
	.dc.b	$78,$7F,$2C,$BE,$4F,$F4,$19,$D3,$6B,$F1,$CF,$F3,$A8,$CB,$FD,$D9
	.dc.b	$AE,$63,$6F,$81,$FF,$62,$3D,$FE,$93,$FB,$39,$E7,$8C,$C7,$CF,$3D
	.dc.b	$FF,$05,$A7,$81,$E3,$78,$11,$C1,$EB,$FA,$7C,$3C,$EF,$E1,$CE,$33
	.dc.b	$F9,$35,$E1,$EC,$FE,$F3,$ED,$CC,$3E,$04,$65,$F8,$D1,$97,$A6,$3F
	.dc.b	$BA,$BF,$F3,$DE,$3F,$42,$F8,$56,$9F,$F0,$2F,$CD,$FF,$8D,$3C,$1E
	.dc.b	$1B,$7E,$33,$D3,$FC,$99,$9A,$D7,$AB,$FE,$B4,$FA,$87,$C6,$53,$39
	.dc.b	$7F,$5E,$F8,$9F,$EB,$5A,$FD,$EF,$84,$FE,$DF,$18,$9F,$F7,$B4,$63
	.dc.b	$D0,$FD,$F4,$60,$FF,$62,$CE,$F7,$FF,$68,$47,$FB,$23,$E1,$BA,$3F
	.dc.b	$E3,$46,$BF,$EF,$23,$91,$63,$FD,$B1,$BB,$E6,$FF,$D3,$F9,$7F,$84
	.dc.b	$3E,$87,$F6,$1C,$3F,$DD,$AB,$4F,$F7,$E8,$CE,$1F,$EC,$99,$7C,$C7
	.dc.b	$2A,$FD,$9F,$8F,$6B,$FD,$BB,$5A,$9F,$F0,$3F,$81,$FC,$59,$FC,$D7
	.dc.b	$C3,$6F,$4F,$D2,$BC,$7C,$6F,$F1,$17,$7E,$7A,$53,$4D,$77,$3E,$77
	.dc.b	$18,$FE,$A5,$BA,$7F,$8C,$5B,$C1,$FB,$67,$C1,$F5,$0D,$29,$D3,$E1
	.dc.b	$6E,$1F,$B1,$9D,$36,$74,$D2,$61,$FE,$3F,$F9,$EE,$2D,$8C,$E3,$FC
	.dc.b	$1A,$72,$D8,$ED,$E3,$7E,$DA,$D8,$7F,$BF,$67,$FC,$75,$18,$9D,$4E
	.dc.b	$70,$FE,$00,$DB,$FE,$27,$C9,$EE,$AF,$8E,$DC,$3F,$87,$36,$FF,$CA
	.dc.b	$CF,$D9,$F1,$BF,$DD,$7F,$45,$FC,$21,$B5,$FE,$46,$E1,$4E,$5B,$19
	.dc.b	$8F,$B5,$E5,$BE,$6F,$A5,$6B,$1F,$E1,$9C,$9F,$D2,$E7,$3E,$3C,$C7
	.dc.b	$FC,$9B,$E5,$7F,$2B,$7F,$9E,$C6,$6F,$F8,$A6,$7F,$F0,$02,$3F,$11
	.dc.b	$5B,$9C,$E2,$9E,$11,$FC,$F9,$F4,$12,$FD,$8F,$A3,$DF,$ED,$BF,$F9
	.dc.b	$D3,$44,$7E,$76,$DE,$F3,$C7,$FE,$5B,$18,$6C,$67,$1F,$F1,$DB,$67
	.dc.b	$FC,$E8,$DC,$3F,$C1,$0B,$EA,$3B,$39,$AF,$F3,$BB,$97,$FC,$95,$F4
	.dc.b	$AF,$8F,$FB,$0D,$8F,$F4,$47,$D3,$3E,$9D,$F3,$FA,$B9,$5C,$9F,$DC
	.dc.b	$56,$E7,$39,$FF,$41,$1F,$4E,$FA,$0F,$7B,$70,$FF,$2B,$B6,$7F,$D2
	.dc.b	$0D,$7F,$D9,$2E,$75,$FE,$23,$F6,$5F,$E1,$AA,$FF,$42,$CE,$5F,$F5
	.dc.b	$94,$66,$BE,$DF,$A3,$F8,$F2,$4F,$E4,$DB,$1F,$8C,$F0,$8F,$E4,$23
	.dc.b	$F8,$64,$62,$7F,$C0,$FE,$0F,$E6,$48,$FE,$75,$B5,$CF,$D5,$FC,$CC
	.dc.b	$72,$7F,$05,$F5,$1F,$3E,$7F,$D5,$CD,$7F,$DC,$ED,$7F,$D5,$ED,$7C
	.dc.b	$AF,$1C,$FF,$C9,$0B,$E2,$7A,$F0,$76,$6E,$3F,$EA,$A9,$D3,$67,$4F
	.dc.b	$F8,$8B,$BF,$F6,$84,$7F,$70,$DC,$3E,$37,$7F,$ED,$23,$E2,$BF,$1B
	.dc.b	$E0,$BE,$3F,$C1,$5B,$FF,$5F,$ED,$FE,$8F,$6C,$E9,$A6,$A7,$E2,$F8
	.dc.b	$1F,$A4,$CF,$F8,$14,$FF,$A7,$1B,$97,$C8,$9D,$1F,$CF,$3D,$FE,$35
	.dc.b	$FE,$23,$6C,$FF,$BA,$DB,$B7,$F8,$C1,$FE,$87,$C9,$F7,$BD,$AB,$7E
	.dc.b	$86,$DF,$E1,$5B,$7F,$EF,$0D,$3F,$84,$3F,$E9,$9F,$A4,$7C,$8F,$F2
	.dc.b	$16,$E1,$FE,$E6,$6E,$1F,$E6,$76,$BF,$E9,$D7,$35,$FE,$44,$FA,$7F
	.dc.b	$E5,$91,$8A,$FC,$AF,$8D,$FE,$FF,$9F,$EC,$19,$A7,$F2,$FE,$06,$1A
	.dc.b	$7F,$BC,$23,$4F,$D4,$8D,$34,$D1,$C3,$1F,$E5,$F3,$C1,$FD,$34,$FE
	.dc.b	$93,$9F,$C3,$9D,$BE,$CD,$AE,$2D,$CF,$CD,$0C,$BF,$82,$4F,$AB,$D2
	.dc.b	$FD,$7F,$97,$70,$FC,$4B,$6F,$F6,$CD,$C3,$E9,$3F,$AC,$F4,$3F,$10
	.dc.b	$4A,$FC,$43,$6B,$D7,$F3,$BB,$9F,$88,$D8,$CC,$7E,$9E,$7E,$6C,$FC
	.dc.b	$EF,$7A,$4F,$CB,$F9,$EB,$7F,$C7,$F9,$9F,$0D,$F5,$3D,$DF,$DF,$B6
	.dc.b	$FF,$DF,$5E,$B3,$3F,$ED,$0D,$8F,$AD,$58,$9F,$BB,$F4,$DF,$82,$F3
	.dc.b	$ED,$7F,$16,$DF,$F2,$11,$98,$FB,$DE,$C5,$D3,$EC,$BF,$86,$24,$71
	.dc.b	$3F,$D2,$FD,$1F,$E3,$6B,$FA,$BC,$9F,$CD,$EB,$FB,$E7,$D3,$7F,$86
	.dc.b	$B0,$FE,$D2,$DA,$DF,$43,$E8,$F7,$EB,$4C,$FF,$21,$F0,$37,$1F,$C5
	.dc.b	$B7,$0E,$2D,$9F,$CE,$E1,$FE,$F3,$7C,$D4,$F5,$59,$7F,$02,$C3,$F1
	.dc.b	$71,$8F,$07,$F4,$7F,$05,$E8,$66,$1F,$A1,$E4,$B8,$FE,$71,$B8,$7A
	.dc.b	$9B,$3E,$27,$E0,$FC,$64,$8F,$D5,$79,$D6,$FE,$1C,$FE,$CB,$E9,$B1
	.dc.b	$87,$C4,$6D,$FB,$1B,$5F,$A1,$D5,$1F,$3B,$E3,$A5,$7E,$DF,$C4,$B7
	.dc.b	$F4,$27,$F7,$FC,$D8,$C7,$E8,$B6,$FE,$4B,$6B,$B3,$F6,$BA,$A4,$FF
	.dc.b	$03,$FA,$6B,$5F,$72,$7F,$CE,$52,$FA,$67,$BD,$AC,$7F,$73,$39,$8C
	.dc.b	$3F,$D9,$67,$6F,$D8,$E3,$FA,$F7,$D2,$73,$87,$F8,$17,$E7,$FF,$11
	.dc.b	$7F,$DC,$22,$08,$CB,$83,$1D,$1F,$E2,$88,$CC,$70,$77,$97,$F8,$C7
	.dc.b	$E7,$FF,$C6,$11,$8B,$FE,$53,$EE,$27,$F9,$4F,$C4,$7F,$2A,$FA,$6B
	.dc.b	$5F,$C9,$DB,$8F,$F2,$26,$E3,$FA,$07,$FC,$AF,$D9,$2B,$F2,$FD,$CC
	.dc.b	$9F,$13,$9C,$7F,$CB,$31,$FA,$9A,$F1,$DF,$F9,$82,$B5,$8F,$F2,$6B
	.dc.b	$7D,$1B,$C8,$FF,$24,$C3,$E2,$3F,$C0,$0F,$F8,$62,$75,$EF,$FF,$AF
	.dc.b	$FD,$8F,$F0,$DF,$D1,$BE,$99,$E6,$7F,$32,$C7,$8E,$3D,$B2,$EB,$18
	.dc.b	$CB,$9C,$F7,$C7,$EB,$46,$27,$C3,$8F,$5C,$3A,$56,$B0,$E6,$DC,$BC
	.dc.b	$98,$F9,$E7,$2A,$F9,$6D,$A8,$F2,$55,$CB,$C1,$87,$39,$F2,$1C,$73
	.dc.b	$F6,$DE,$F1,$EB,$3C,$8D,$97,$3F,$4D,$70,$8C,$E1,$D9,$F3,$63,$F9
	.dc.b	$17,$91,$F9,$68,$F8,$AD,$BF,$AA,$FE,$6B,$E2,$DF,$44,$FA,$6E,$51
	.dc.b	$86,$C6,$67,$F3,$D3,$97,$F4,$16,$F4,$3F,$DA,$11,$86,$BE,$36,$BF
	.dc.b	$59,$AF,$E1,$5E,$39,$FD,$D6,$EB,$BE,$2F,$AB,$BB,$E9,$9F,$31,$FB
	.dc.b	$5C,$38,$57,$D3,$C3,$8E,$3D,$63,$8D,$F0,$8D,$44,$7E,$6C,$F0,$C3
	.dc.b	$B6,$1D,$71,$F1,$56,$AF,$86,$3F,$8E,$C7,$A4,$E4,$91,$E4,$7F,$07
	.dc.b	$87,$47,$F1,$FC,$DF,$B3,$F9,$0F,$E1,$DC,$BF,$EB,$56,$F2,$7F,$04
	.dc.b	$6D,$7F,$85,$79,$FE,$CB,$FB,$C5,$1A,$BE,$93,$98,$C4,$F0,$8F,$DF
	.dc.b	$F8,$F8,$FA,$49,$1C,$E3,$66,$1F,$E0,$B1,$87,$F9,$57,$CD,$FF,$01
	.dc.b	$6F,$37,$9B,$C2,$4E,$67,$F8,$6F,$C1,$66,$31,$AF,$F6,$66,$EB,$FE
	.dc.b	$9E,$BE,$38,$7A,$2F,$87,$5F,$AA,$E6,$3E,$28,$CB,$48,$31,$FE,$8B
	.dc.b	$F1,$BF,$05,$F0,$4F,$09,$FF,$78,$D7,$F1,$27,$FB,$EE,$4B,$E0,$F9
	.dc.b	$67,$FC,$19,$3F,$47,$D5,$F9,$A7,$B7,$C3,$91,$F1,$9A,$F8,$EB,$FC
	.dc.b	$CF,$F4,$DF,$CF,$E3,$44,$3C,$73,$F6,$E3,$51,$E6,$F3,$3E,$DF,$97
	.dc.b	$23,$E9,$C7,$CD,$23,$83,$FA,$0C,$FA,$9F,$F0,$C6,$3F,$DF,$9D,$1D
	.dc.b	$4D,$F7,$7D,$7F,$0F,$F1,$3F,$49,$27,$F0,$31,$96,$90,$F7,$1F,$E6
	.dc.b	$BB,$71,$EE,$3E,$7C,$DE,$0F,$F1,$ED,$6A,$7F,$BB,$7D,$23,$D2,$FC
	.dc.b	$CF,$97,$27,$31,$F8,$F9,$D4,$60,$87,$8B,$A9,$7F,$E5,$D8,$CE,$1F
	.dc.b	$E3,$6A,$E3,$1E,$8E,$74,$62,$7C,$BC,$7F,$AA,$FA,$49,$7F,$B1,$7C
	.dc.b	$5A,$C8,$F4,$B5,$F1,$62,$DB,$EE,$19,$C3,$C5,$58,$AF,$DE,$78,$3F
	.dc.b	$B8,$23,$FB,$FF,$A4,$F8,$18,$C8,$F8,$EF,$E1,$CF,$E0,$EB,$E9,$3E
	.dc.b	$03,$F5,$0C,$67,$A8,$CE,$3F,$4B,$0F,$99,$87,$F0,$8F,$9B,$98,$7D
	.dc.b	$48,$CB,$48,$7B,$A3,$2F,$CB,$D5,$38,$BF,$BF,$5F,$C5,$FE,$89,$D9
	.dc.b	$FC,$67,$F7,$8F,$F1,$7D,$7F,$A0,$1B,$D7,$FE,$CE,$AD,$E5,$FE,$1C
	.dc.b	$6B,$F7,$E7,$6F,$FA,$ED,$BC,$69,$DD,$FF,$A3,$1F,$BD,$7D,$79,$3C
	.dc.b	$87,$F2,$BF,$99,$FE,$D3,$5B,$8C,$E3,$E1,$75,$E6,$7A,$49,$1F,$E3
	.dc.b	$0C,$B8,$E1,$E9,$F9,$6C,$CF,$9E,$AE,$1E,$39,$B5,$FE,$EC,$BE,$37
	.dc.b	$EB,$BF,$91,$5E,$5F,$A4,$F9,$5E,$2C,$AF,$F1,$FE,$69,$D4,$7C,$79
	.dc.b	$D3,$FA,$59,$D3,$E7,$F9,$0F,$A4,$47,$E7,$78,$B5,$F9,$D1,$74,$F9
	.dc.b	$53,$F8,$BE,$2F,$B5,$8C,$BF,$F5,$39,$F9,$38,$D3,$8E,$5F,$C2,$B6
	.dc.b	$3F,$AF,$E8,$FA,$5F,$EB,$B3,$92,$1F,$9B,$AE,$2F,$F1,$09,$E1,$A7
	.dc.b	$F6,$3F,$69,$F5,$34,$95,$F8,$9E,$9A,$FF,$07,$9D,$BF,$E6,$86,$BF
	.dc.b	$C2,$F9,$3F,$4F,$C7,$F9,$3F,$13,$B7,$C6,$4A,$FC,$A4,$7E,$74,$C3
	.dc.b	$D6,$30,$FC,$4F,$D3,$9F,$1D,$FD,$E7,$E8,$3F,$CD,$DC,$9F,$EE,$06
	.dc.b	$E3,$FE,$93,$AD,$DF,$1B,$FD,$FC,$62,$3F,$7C,$DB,$FF,$3C,$7D,$3B
	.dc.b	$E4,$F7,$72,$F9,$1F,$5A,$74,$FE,$B7,$C6,$F2,$7C,$7C,$8C,$10,$FE
	.dc.b	$5D,$9A,$35,$1F,$1E,$7F,$7D,$E1,$7B,$7E,$3E,$4F,$EE,$63,$2F,$FA
	.dc.b	$5A,$D3,$A8,$FA,$13,$FE,$22,$CD,$F8,$DF,$F8,$C1,$6C,$AF,$F0,$04
	.dc.b	$E6,$31,$3C,$2B,$F9,$2F,$D3,$FF,$02,$9C,$D6,$EF,$8D,$77,$8C,$42
	.dc.b	$B8,$3D,$FE,$A1,$F1,$9F,$E7,$38,$46,$1F,$F3,$B3,$5F,$F6,$43,$5F
	.dc.b	$F6,$6B,$9C,$7F,$D5,$CD,$C7,$FD,$A4,$D9,$F7,$5F,$F9,$BA,$BF,$D6
	.dc.b	$9F,$4E,$F9,$1F,$E8,$2D,$AF,$D8,$B5,$FE,$84,$BE,$33,$73,$27,$6F
	.dc.b	$28,$FF,$1E,$7E,$8E,$E6,$A4,$AF,$F2,$97,$BD,$5F,$ED,$53,$B7,$FD
	.dc.b	$E0,$D7,$FD,$9F,$1B,$C7,$EC,$C3,$D1,$56,$E7,$31,$8B,$FF,$2D,$DF
	.dc.b	$1D,$7F,$C9,$17,$B6,$CE,$B0,$FF,$80,$1F,$9B,$FF,$93,$9E,$CD,$C0
	.dc.b	$E7,$CB,$FF,$73,$7D,$27,$F9,$6E,$34,$4F,$6F,$FC,$D7,$1F,$E7,$4F
	.dc.b	$80,$7D,$FF,$F9,$83,$1F,$E8,$1F,$39,$FD,$6C,$D4,$66,$3F,$C4,$B3
	.dc.b	$C3,$DA,$C9,$23,$F9,$CF,$75,$77,$5C,$78,$7E,$3D,$F1,$9D,$9F,$9E
	.dc.b	$DF,$F4,$51,$E4,$EC,$FE,$D9,$DF,$2B,$42,$74,$E1,$AF,$82,$7D,$27
	.dc.b	$F6,$3F,$A5,$B1,$BC,$53,$88,$FE,$99,$37,$1F,$3D,$6A,$78,$57,$37
	.dc.b	$C1,$F0,$0F,$5F,$FC,$05,$F3,$B2,$73,$1F,$3E,$75,$18,$21,$FD,$6F
	.dc.b	$E7,$F5,$32,$33,$8F,$E3,$B9,$1E,$9A,$FA,$7A,$12,$FC,$6F,$FA,$AF
	.dc.b	$EA,$1F,$2D,$E1,$F6,$A5,$7A,$1F,$E7,$1F,$1D,$FE,$30,$F8,$BF,$F5
	.dc.b	$9E,$9F,$4B,$E3,$BF,$8A,$5A,$72,$FD,$18,$CC,$63,$91,$18,$F4,$FF
	.dc.b	$D3,$55,$B3,$98,$F8,$9E,$CE,$1A,$FF,$B9,$4E,$05,$AE,$AF,$E0,$7D
	.dc.b	$CF,$E5,$1D,$3F,$81,$6C,$7F,$9A,$7E,$81,$F2,$BF,$B6,$D3,$F2,$35
	.dc.b	$A6,$FB,$19,$C3,$F2,$32,$8C,$60,$E5,$71,$7F,$11,$BF,$05,$6B,$36
	.dc.b	$1F,$E1,$99,$F2,$7B,$7C,$B0,$FB,$3F,$3D,$BA,$D6,$5F,$AC,$6E,$7E
	.dc.b	$6B,$E1,$CE,$BE,$11,$31,$FF,$1B,$77,$DC,$7F,$C9,$4F,$8F,$EA,$FF
	.dc.b	$90,$4F,$E9,$5B,$3F,$E2,$9F,$9F,$D7,$91,$2B,$4F,$82,$31,$3F,$C0
	.dc.b	$2B,$8E,$7F,$BA,$6D,$FF,$59,$D7,$F2,$B7,$C2,$B5,$87,$C1,$D5,$E0
	.dc.b	$E7,$0F,$F2,$33,$F2,$7E,$C5,$E3,$D3,$E0,$F2,$DA,$FD,$F4,$E9,$B3
	.dc.b	$A6,$93,$5F,$F3,$67,$D5,$F1,$6C,$7F,$2C,$F8,$FF,$BB,$E5,$FF,$2A
	.dc.b	$7E,$B7,$87,$FE,$B9,$9D,$12,$BF,$D5,$DE,$5A,$B7,$F0,$BC,$A3,$15
	.dc.b	$FC,$E3,$D4,$C1,$D3,$E1,$6C,$64,$98,$7F,$AF,$F1,$4E,$A3,$BD,$5E
	.dc.b	$2F,$E0,$55,$AB,$E3,$1F,$D3,$B0,$FE,$F7,$18,$3E,$A4,$EE,$F9,$C7
	.dc.b	$F9,$73,$93,$FC,$40,$DF,$6B,$F3,$F3,$FA,$1F,$A9,$B5,$A8,$FF,$59
	.dc.b	$37,$1F,$EC,$31,$FA,$5A,$E3,$DA,$BE,$31,$D3,$E9,$9F,$3B,$4F,$A3
	.dc.b	$F5,$3F,$EA,$AC,$FF,$CD,$F3,$A6,$CF,$6C,$7F,$89,$CE,$5F,$E1,$61
	.dc.b	$C3,$FC,$39,$AF,$F8,$B7,$8B,$F8,$D7,$9D,$F8,$0F,$1F,$FA,$0E,$1F
	.dc.b	$D3,$35,$7D,$3B,$BB,$57,$2B,$FA,$8E,$1C,$9E,$15,$B7,$38,$74,$C7
	.dc.b	$FC,$E8,$7F,$01,$F7,$FF,$30,$E2,$7F,$C9,$1D,$3F,$C1,$7E,$13,$F8
	.dc.b	$A4,$7E,$AE,$B7,$A7,$F4,$19,$CD,$FF,$13,$9C,$D7,$EC,$65,$F1,$5D
	.dc.b	$FE,$5F,$A3,$67,$FD,$08,$DB,$FF,$44,$37,$0F,$B7,$F9,$93,$DB,$FC
	.dc.b	$1B,$6F,$FD,$47,$F4,$D7,$0F,$F7,$A4,$DB,$E3,$7C,$0F,$C5,$D6,$6F
	.dc.b	$FA,$5D,$5B,$FF,$30,$37,$A1,$A7,$EA,$A7,$AE,$1F,$27,$BB,$E3,$7B
	.dc.b	$5F,$2F,$DE,$B9,$7F,$9C,$5B,$7F,$E7,$C6,$E9,$FE,$93,$7E,$97,$D3
	.dc.b	$4A,$FF,$53,$7B,$36,$FF,$D0,$73,$6F,$E8,$4D,$AF,$EC,$7B,$3E,$3B
	.dc.b	$DA,$7F,$36,$FA,$8B,$3F,$1A,$2E,$3F,$ED,$79,$B8,$7E,$A2,$F9,$5F
	.dc.b	$FB,$BF,$DA,$CB,$FE,$E9,$7F,$94,$7D,$44,$8F,$F5,$E7,$9D,$71,$61
	.dc.b	$FE,$05,$F3,$17,$98,$FC,$45,$5B,$FE,$6D,$E8,$FF,$38,$C3,$FA,$E7
	.dc.b	$47,$F8,$2E,$7F,$DE,$3E,$77,$E4,$F3,$FE,$A2,$D8,$C9,$FE,$CD,$8F
	.dc.b	$E7,$9F,$35,$23,$FD,$A1,$9D,$C3,$CB,$1F,$D1,$74,$56,$74,$FC,$E5
	.dc.b	$DA,$D4,$DB,$F2,$CD,$E7,$66,$FA,$3F,$D4,$1F,$DB,$7C,$3F,$E2,$98
	.dc.b	$FE,$8E,$32,$FF,$B8,$DD,$78,$B8,$C7,$FD,$D1,$E8,$FF,$55,$F4,$FF
	.dc.b	$64,$DB,$FF,$75,$36,$77,$1F,$E5,$2C,$BF,$BF,$D7,$F4,$AF,$84,$FB
	.dc.b	$75,$F0,$0F,$F4,$23,$8D,$BF,$C7,$59,$FE,$E2,$73,$CD,$C4,$98,$F5
	.dc.b	$9D,$78,$9F,$E9,$F9,$CD,$75,$CB,$FC,$1F,$AE,$31,$FC,$13,$FB,$0F
	.dc.b	$16,$65,$CF,$C2,$B8,$79,$72,$FF,$55,$7C,$7F,$A4,$F9,$5D,$B6,$FD
	.dc.b	$DC,$AE,$1F,$EF,$47,$F7,$BF,$4E,$CF,$9F,$E9,$FE,$17,$C0,$7D,$36
	.dc.b	$C7,$F2,$8E,$34,$E3,$4F,$57,$1A,$F1,$C3,$F9,$BF,$37,$FC,$10,$63
	.dc.b	$1F,$C5,$3F,$DE,$3F,$5B,$78,$74,$FC,$17,$A5,$6B,$F9,$BC,$E9,$B3
	.dc.b	$A6,$93,$2F,$99,$BB,$C2,$E5,$18,$7C,$1D,$D2,$BE,$B6,$F6,$BE,$66
	.dc.b	$3F,$AB,$6D,$7E,$56,$75,$A7,$FB,$12,$B4,$FE,$E7,$6C,$E1,$FA,$77
	.dc.b	$DD,$C1,$23,$F0,$9F,$39,$72,$EF,$36,$FF,$1B,$36,$BF,$B0,$60,$AF
	.dc.b	$F1,$97,$C9,$E7,$0F,$F2,$E6,$A8,$FF,$11,$FE,$C2,$C2,$9F,$AD,$F7
	.dc.b	$AB,$FD,$8A,$34,$D3,$45,$25,$FC,$CF,$0F,$8B,$63,$97,$C8,$78,$0F
	.dc.b	$F3,$7A,$FA,$6F,$DF,$3A,$24,$FE,$66,$D5,$C2,$B7,$39,$8E,$95,$FB
	.dc.b	$7F,$D7,$BC,$F3,$98,$C1,$F0,$3C,$C8,$FE,$F7,$D8,$7D,$36,$2B,$C1
	.dc.b	$57,$8B,$FC,$67,$36,$FF,$CC,$58,$7F,$12,$BF,$F4,$43,$E6,$8F,$17
	.dc.b	$8D,$FE,$83,$8C,$4F,$EE,$DE,$1E,$21,$CE,$1F,$EC,$17,$26,$2F,$BF
	.dc.b	$A7,$CC,$E9,$E1,$D8,$F8,$11,$70,$ED,$36,$FF,$C7,$3F,$49,$FE,$70
	.dc.b	$F6,$95,$FE,$B5,$F4,$BF,$D8,$33,$98,$C0,$FA,$37,$D3,$7F,$77,$F8
	.dc.b	$6F,$F0,$44,$F2,$AE,$11,$88,$46,$2F,$AE,$7F,$EA,$F9,$EB,$87,$4E
	.dc.b	$1F,$C9,$47,$80,$E3,$3B,$C3,$A6,$1C,$E7,$C1,$39,$C3,$D7,$87,$3D
	.dc.b	$7D,$B8,$70,$AD,$5F,$0C,$F3,$87,$91,$B1,$EF,$CF,$DD,$39,$8F,$26
	.dc.b	$5E,$17,$CB,$5C,$32,$F0,$46,$B0,$EE,$E8,$FF,$2C,$D7,$CD,$7F,$DC
	.dc.b	$AD,$7B,$1F,$EE,$8F,$65,$F4,$6F,$A6,$FF,$30,$CF,$FB,$D6,$30,$FF
	.dc.b	$99,$27,$73,$A6,$C7,$83,$81,$F0,$FC,$B0,$46,$27,$53,$98,$C4,$79
	.dc.b	$9D,$6D,$F4,$9B,$AF,$E2,$1B,$3F,$87,$FA,$67,$CC,$FE,$3A,$3E,$4C
	.dc.b	$E9,$FC,$2B,$71,$F1,$E3,$F2,$EF,$8D,$7D,$43,$71,$67,$71,$D2,$38
	.dc.b	$5F,$7C,$7C,$D1,$BC,$3B,$1E,$6D,$98,$F6,$C3,$91,$2F,$8B,$F9,$68
	.dc.b	$D4,$7E,$83,$95,$D0,$F8,$BF,$77,$37,$FA,$E1,$BC,$BF,$EB,$46,$E2
	.dc.b	$BE,$3C,$58,$79,$CD,$C3,$A5,$7F,$15,$8C,$56,$7D,$3D,$7D,$14,$AF
	.dc.b	$E3,$51,$97,$F8,$E4,$65,$FF,$68,$47,$8B,$D2,$7C,$A7,$FB,$CA,$7F
	.dc.b	$D5,$B1,$8E,$7F,$C7,$B6,$33,$3B,$7E,$E0,$8D,$3F,$E0,$F8,$D6,$BF
	.dc.b	$BC,$9C,$BF,$ED,$CF,$A1,$F1,$4F,$FB,$B6,$BF,$99,$3F,$04,$CF,$5C
	.dc.b	$8F,$0B,$E4,$8E,$11,$E2,$BF,$5C,$E4,$E1,$1E,$37,$E8,$EF,$FE,$54
	.dc.b	$D3,$C3,$F4,$DF,$3F,$D5,$97,$FE,$C0,$C5,$3A,$8F,$82,$F8,$48,$E0
	.dc.b	$E0,$ED,$FF,$1C,$7D,$2F,$F4,$C7,$A7,$97,$F7,$7E,$82,$46,$08,$7A
	.dc.b	$ED,$3E,$F2,$47,$D2,$7E,$49,$1F,$F2,$B5,$7F,$08,$7D,$5F,$4D,$F8
	.dc.b	$BF,$A6,$91,$82,$1D,$E3,$E6,$11,$FB,$EF,$D6,$23,$DB,$8F,$F1,$5F
	.dc.b	$4C,$F8,$FF,$CC,$7B,$92,$BF,$DD,$33,$96,$CF,$E3,$23,$11,$F6,$5A
	.dc.b	$FF,$07,$E2,$46,$23,$FD,$F8,$F0,$E4,$FC,$63,$78,$B3,$7F,$AB,$D1
	.dc.b	$C3,$FA,$BF,$99,$98,$FF,$7F,$7D,$3A,$48,$FD,$0E,$28,$F3,$12,$3F
	.dc.b	$3E,$7E,$E6,$DB,$71,$8F,$F3,$75,$71,$7E,$9C,$7E,$EF,$C0,$91,$FE
	.dc.b	$0F,$7F,$7D,$ED,$FC,$AC,$7E,$21,$27,$C6,$FF,$40,$5F,$4B,$EC,$7F
	.dc.b	$A1,$B5,$F2,$B5,$FD,$AB,$5F,$93,$3F,$C3,$7E,$66,$47,$D6,$8F,$B6
	.dc.b	$7D,$52,$3F,$11,$AF,$B9,$DB,$F9,$08,$FE,$35,$F4,$AE,$AF,$E3,$9E
	.dc.b	$A7,$F0,$6E,$2F,$F6,$C5,$EA,$63,$FE,$90,$36,$60,$FF,$50,$57,$DD
	.dc.b	$7F,$94,$37,$C0,$37,$E1,$BF,$C1,$93,$C3,$8F,$C6,$47,$FD,$75,$1A
	.dc.b	$C3,$FD,$D5,$E5,$F9,$89,$3F,$EB,$FB,$7C,$9E,$23,$FD,$0F,$36,$BF
	.dc.b	$C2,$F8,$F8,$AB,$1C,$DF,$EE,$1A,$B7,$E8,$9F,$8B,$E1,$48,$EC,$FF
	.dc.b	$41,$C9,$F2,$F4,$D4,$FC,$29,$C3,$FE,$B6,$F8,$FF,$9A,$47,$D1,$E2
	.dc.b	$57,$FD,$3B,$17,$4F,$3E,$AA,$FE,$A7,$5F,$8C,$E2,$FA,$F8,$C8,$FF
	.dc.b	$22,$DF,$F8,$D1,$D1,$8D,$7F,$BE,$71,$FF,$AF,$A3,$F2,$77,$F9,$78
	.dc.b	$C4,$7E,$52,$30,$78,$23,$35,$D5,$B5,$FE,$54,$D9,$F4,$DF,$D8,$B0
	.dc.b	$FF,$71,$7C,$47,$FA,$65,$FF,$7A,$35,$CF,$B5,$F9,$2E,$DF,$ED,$93
	.dc.b	$9C,$BE,$BB,$71,$FF,$30,$F1,$A3,$18,$7E,$3A,$F8,$FB,$32,$3B,$BF
	.dc.b	$9E,$F1,$24,$60,$87,$E6,$1F,$9B,$F4,$CF,$8C,$FE,$FD,$F2,$1F,$2F
	.dc.b	$4F,$C4,$B7,$4F,$F1,$35,$6F,$0F,$9C,$7F,$9D,$9C,$3F,$C4,$9A,$FF
	.dc.b	$72,$6B,$CA,$FF,$84,$47,$0B,$FB,$4D,$AD,$57,$61,$F4,$CF,$A0,$F3
	.dc.b	$FB,$52,$7C,$7B,$B3,$4F,$F0,$0F,$1E,$56,$A7,$2D,$9C,$AC,$FE,$0F
	.dc.b	$B9,$F1,$09,$1F,$B1,$F0,$FF,$85,$98,$78,$69,$F6,$30,$FF,$61,$C6
	.dc.b	$CF,$8F,$6F,$A6,$FF,$1E,$47,$F8,$D1,$AF,$E7,$1A,$F8,$DB,$1C,$6F
	.dc.b	$FC,$6E,$D9,$EB,$3F,$E4,$78,$FD,$27,$D2,$FF,$1F,$9F,$95,$F4,$CF
	.dc.b	$8A,$FD,$6C,$E6,$7D,$27,$FA,$C1,$C5,$7E,$B6,$B8,$4F,$2A,$FE,$81
	.dc.b	$18,$8F,$EA,$2D,$7F,$C6,$2D,$7F,$CD,$DF,$4C,$F8,$FF,$E9,$0D,$8F
	.dc.b	$F4,$4B,$5F,$EB,$6D,$7F,$C7,$6E,$6F,$F8,$03,$74,$EB,$5F,$C2,$21
	.dc.b	$F1,$3F,$E4,$BF,$A6,$74,$7F,$92,$1B,$8F,$F9,$31,$AF,$F8,$DD,$AF
	.dc.b	$F7,$F7,$38,$FF,$91,$DB,$8F,$FB,$5B,$D0,$46,$2B,$FC,$D5,$EC,$7F
	.dc.b	$B3,$EF,$86,$1E,$32,$4F,$FB,$F7,$DF,$AF,$D2,$C3,$9C,$E2,$BF,$9D
	.dc.b	$DF,$3C,$7C,$1F,$3B,$FE,$5F,$C3,$FC,$28,$D7,$3E,$D7,$F9,$9A,$31
	.dc.b	$1F,$E6,$37,$FC,$C3,$3A,$8F,$DD,$E2,$9C,$C7,$AB,$E9,$BF,$D0,$BF
	.dc.b	$00,$FA,$6F,$F3,$AC,$FE,$C3,$D1,$7C,$87,$F6,$C3,$73,$A9,$FC,$6D
	.dc.b	$71,$F6,$FF,$D8,$ED,$8F,$F0,$CB,$5F,$F3,$CB,$5F,$F5,$B3,$9D,$BF
	.dc.b	$DE,$6D,$D1,$3F,$EC,$AB,$F0,$57,$F6,$2F,$43,$F8,$64,$E7,$BB,$FC
	.dc.b	$03,$E0,$48,$FE,$37,$8F,$B6,$7D,$27,$9B,$03,$49,$3E,$88,$CB,$C9
	.dc.b	$AF,$C0,$6E,$1E,$3F,$A2,$EB,$3B,$AA,$FE,$CB,$C1,$64,$AF,$A3,$C6
	.dc.b	$F5,$BF,$C3,$3F,$35,$30,$F4,$CE,$5B,$3F,$3A,$31,$1F,$31,$AF,$CC
	.dc.b	$6B,$F2,$B4,$FF,$64,$CF,$80,$91,$FE,$16,$F9,$1F,$1F,$9F,$D4,$91
	.dc.b	$F6,$AF,$FB,$6B,$D7,$D8,$F1,$11,$D3,$FE,$49,$F8,$7F,$F6,$B6,$FE
	.dc.b	$1F,$A4,$FF,$7A,$46,$29,$9A,$73,$EB,$FF,$37,$9C,$C6,$2F,$94,$7F
	.dc.b	$B0,$BE,$12,$BF,$EF,$13,$66,$4A,$66,$FB,$3E,$0E,$8F,$C6,$F3,$7E
	.dc.b	$59,$B1,$F9,$66,$BF,$96,$6B,$F2,$D7,$A7,$FE,$17,$8C,$35,$EC,$DA
	.dc.b	$4F,$0A,$F0,$7A,$A7,$F2,$F1,$FE,$34,$7F,$C9,$3F,$31,$9C,$3F,$5F
	.dc.b	$39,$6D,$75,$8C,$47,$9E,$30,$4A,$FD,$E7,$A1,$CF,$3F,$D7,$DF,$37
	.dc.b	$FC,$9A,$FF,$82,$67,$F5,$3F,$3F,$FC,$B9,$BE,$17,$17,$FC,$37,$D0
	.dc.b	$F9,$8C,$67,$F7,$2F,$83,$FC,$19,$B8,$FE,$F9,$B7,$CA,$B8,$EA,$9C
	.dc.b	$DF,$3C,$7E,$54,$74,$8F,$E1,$2D,$7F,$3C,$D7,$85,$FF,$1B,$E9,$EB
	.dc.b	$3D,$3A,$BE,$9C,$78,$3E,$8D,$E2,$53,$D7,$3F,$B8,$F9,$4C,$61,$FC
	.dc.b	$32,$B6,$FF,$1E,$6E,$3F,$C7,$9A,$FF,$1E,$6B,$FB,$37,$7A,$FD,$0F
	.dc.b	$A3,$EB,$3F,$C1,$F8,$11,$F3,$3C,$6F,$E7,$5F,$15,$7C,$73,$E6,$ED
	.dc.b	$FF,$72,$67,$FD,$02,$72,$D9,$FC,$A5,$72,$8F,$F6,$6B,$71,$FE,$5E
	.dc.b	$7F,$52,$9D,$D7,$F8,$36,$BA,$79,$DD,$FE,$03,$C4,$C8,$C9,$FE,$38
	.dc.b	$C7,$FD,$36,$E3,$0F,$EC,$57,$CD,$B5,$CE,$75,$A7,$E7,$F2,$FE,$B1
	.dc.b	$3E,$22,$4F,$A2,$F9,$5F,$A0,$6F,$FE,$46,$9F,$F2,$73,$63,$FC,$46
	.dc.b	$7E,$E3,$E6,$7F,$2E,$F0,$F5,$AC,$EE,$3A,$35,$F8,$AD,$7F,$CF,$71
	.dc.b	$FB,$F5,$F4,$4F,$99,$C3,$70,$FF,$7D,$CE,$A7,$24,$8F,$F2,$5F,$D1
	.dc.b	$7E,$85,$FF,$33,$57,$1A,$FF,$1D,$EA,$AD,$47,$F8,$37,$C3,$FE,$B7
	.dc.b	$7F,$E0,$06,$C7,$09,$DD,$7C,$D9,$CB,$7C,$9F,$C2,$C7,$0A,$FC,$1E
	.dc.b	$E8,$C4,$7F,$89,$5A,$FD,$0F,$7B,$F9,$4D,$FF,$A3,$BA,$3F,$53,$18
	.dc.b	$CB,$FC,$D5,$1F,$AC,$F6,$7F,$43,$18,$3F,$D5,$B7,$99,$E3,$F0,$9B
	.dc.b	$F9,$1B,$5F,$E7,$F7,$0D,$3F,$D6,$B1,$8A,$FB,$FF,$44,$F5,$3E,$FB
	.dc.b	$6B,$FD,$25,$F4,$36,$B7,$5A,$3E,$E5,$66,$3F,$1C,$D7,$F4,$ED,$7F
	.dc.b	$C2,$EF,$3D,$3F,$D5,$FD,$5C,$F0,$FF,$4B,$3F,$13,$D4,$F8,$5E,$CD
	.dc.b	$9F,$F4,$D3,$86,$9F,$ED,$08,$C6,$1F,$E0,$BD,$9F,$45,$F2,$CF,$9F
	.dc.b	$E9,$DB,$FF,$52,$45,$AF,$F6,$DC,$DC,$BF,$DD,$DA,$FF,$BD,$6F,$E1
	.dc.b	$FC,$3F,$F2,$0F,$25,$F3,$DF,$A6,$AD,$63,$FA,$A6,$BF,$B1,$6B,$FC
	.dc.b	$29,$AF,$F8,$E9,$A9,$E7,$87,$F5,$0F,$53,$B6,$1F,$CC,$3C,$44,$5F
	.dc.b	$A1,$99,$FF,$BC,$BC,$DB,$97,$F2,$79,$B8,$FF,$BC,$26,$D7,$F5,$6F
	.dc.b	$2F,$F2,$78,$7B,$5B,$1F,$EF,$C8,$B8,$7F,$8D,$F5,$FE,$8D,$D3,$E2
	.dc.b	$AF,$F3,$27,$D1,$79,$4F,$CE,$E7,$63,$24,$BF,$F2,$05,$27,$51,$FC
	.dc.b	$F6,$AE,$7F,$DF,$7C,$25,$EF,$0E,$53,$98,$7D,$17,$EB,$0F,$EA,$BE
	.dc.b	$17,$F1,$D9,$D4,$66,$BF,$DD,$35,$C1,$FF,$75,$D6,$DF,$F3,$6D,$F8
	.dc.b	$2B,$F6,$FE,$7F,$D1,$F4,$6D,$7F,$A0,$DB,$9F,$F5,$56,$E5,$FE,$08
	.dc.b	$7F,$AB,$7B,$DF,$72,$BF,$D5,$4D,$AF,$E6,$B9,$7F,$00,$5E,$17,$FA
	.dc.b	$F3,$93,$89,$FD,$5A,$BF,$13,$19,$BF,$EA,$F1,$83,$A6,$9F,$E2,$CD
	.dc.b	$38,$CF,$FA,$65,$FC,$C7,$87,$FC,$D7,$9B,$F8,$BE,$5F,$91,$8C,$1D
	.dc.b	$FE,$22,$5F,$66,$CF,$F4,$2E,$7F,$F0,$F4,$FF,$A9,$78,$3F,$2D,$E7
	.dc.b	$FE,$5D,$FA,$7F,$45,$F0,$BD,$AB,$5F,$EB,$2D,$91,$98,$F5,$BE,$F2
	.dc.b	$3F,$D6,$F1,$FE,$B7,$8B,$CB,$FD,$21,$C6,$3F,$C0,$CF,$F8,$17,$E8
	.dc.b	$BF,$D6,$F5,$FE,$C0,$6C,$FF,$A4,$78,$FF,$B8,$72,$FF,$6D,$9C,$B7
	.dc.b	$C1,$FF,$02,$C7,$0A,$FE,$82,$D7,$FD,$C8,$D7,$EC,$BC,$AF,$F1,$78
	.dc.b	$FF,$08,$7C,$55,$FB,$BE,$EF,$37,$64,$9F,$C9,$EB,$71,$FF,$73,$63
	.dc.b	$FA,$F7,$FB,$E7,$32,$31,$1E,$4B,$F3,$92,$FF,$DB,$1A,$73,$D3,$FC
	.dc.b	$B6,$FC,$5E,$19,$3F,$A2,$F9,$6B,$87,$82,$7E,$09,$27,$FC,$4B,$8A
	.dc.b	$31,$3F,$E3,$5F,$8C,$E3,$8F,$F9,$A1,$FF,$1B,$7D,$0F,$F9,$EF,$0F
	.dc.b	$F7,$8B,$6B,$E9,$7D,$16,$A6,$1F,$E3,$29,$D4,$FF,$70,$6B,$FE,$E2
	.dc.b	$D5,$18,$8F,$D4,$BD,$70,$FF,$3C,$F8,$3D,$27,$FC,$94,$F9,$3E,$2E
	.dc.b	$61,$FE,$EC,$AF,$F4,$62,$FA,$2F,$E0,$D6,$9F,$F1,$96,$0B,$E3,$3F
	.dc.b	$52,$31,$7E,$99,$FF,$26,$C6,$27,$51,$FE,$6A,$FA,$1F,$35,$7A,$EB
	.dc.b	$57,$FC,$86,$73,$7E,$CF,$2B,$E5,$9F,$E5,$DF,$A2,$B8,$7C,$C8,$EE
	.dc.b	$47,$E5,$3F,$EB,$1F,$01,$7F,$C7,$EF,$FC,$8A,$F8,$2B,$F9,$0B,$99
	.dc.b	$DE,$5E,$7F,$53,$D0,$DA,$F9,$07,$07,$D9,$E8,$C8,$CE,$1D,$BD,$1C
	.dc.b	$3F,$F0,$07,$1F,$E4,$18,$F9,$2F,$E2,$56,$B4,$FF,$5F,$5B,$E8,$3F
	.dc.b	$D2,$7F,$00,$87,$C0,$FF,$86,$DC,$4E,$9C,$FA,$7B,$FA,$07,$D1,$7F
	.dc.b	$A0,$A7,$27,$0C,$FF,$DB,$2E,$0F,$03,$8C,$3B,$46,$31,$F1,$47,$86
	.dc.b	$FA,$F0,$FF,$B5,$63,$79,$78,$DB,$87,$3C,$BF,$D1,$B3,$97,$CD,$97
	.dc.b	$96,$B7,$8F,$69,$DD,$F1,$CB,$D1,$39,$59,$78,$B7,$C4,$79,$CF,$66
	.dc.b	$5D,$DE,$27,$C2,$9E,$0E,$AF,$B3,$C4,$FF,$46,$4F,$CD,$6C,$7C,$86
	.dc.b	$DF,$F9,$F3,$D7,$7D,$03,$E8,$B0,$DA,$FE,$79,$3A,$9C,$9F,$E8,$4C
	.dc.b	$1C,$0F,$84,$C3,$6F,$FA,$24,$EA,$72,$48,$F7,$CE,$F6,$FA,$6D,$DB
	.dc.b	$F0,$9F,$44,$F9,$4F,$F5,$FD,$6B,$0F,$93,$8F,$5C,$79,$C7,$CF,$DD
	.dc.b	$87,$19,$DE,$5D,$5F,$C9,$B5,$FC,$D3,$99,$E3,$8F,$83,$2F,$C1,$CE
	.dc.b	$F1,$F0,$65,$E3,$C3,$EB,$63,$DD,$F1,$92,$34,$7E,$4E,$34,$FE,$97
	.dc.b	$93,$F8,$0F,$1F,$FB,$73,$93,$F8,$73,$78,$FF,$80,$B7,$0F,$E0,$BE
	.dc.b	$D5,$BE,$31,$FB,$C6,$BF,$BB,$9D,$73,$2A,$FB,$38,$F3,$25,$F1,$AF
	.dc.b	$F3,$BB,$67,$2F,$F0,$59,$E2,$FF,$0A,$1F,$A0,$3E,$73,$FD,$93,$18
	.dc.b	$6C,$66,$3F,$D9,$AD,$7F,$C1,$4D,$D3,$F8,$6B,$74,$FF,$7F,$B6,$7F
	.dc.b	$92,$FD,$1F,$C3,$F0,$9F,$AE,$3C,$0F,$7F,$C2,$4B,$FE,$3F,$5A,$9E
	.dc.b	$15,$B9,$FF,$0D,$4E,$5B,$19,$7F,$BE,$B5,$FF,$67,$B9,$C9,$97,$F6
	.dc.b	$18,$C3,$6F,$E8,$FD,$47,$83,$E9,$3E,$9F,$9D,$31,$F2,$E0,$9D,$47
	.dc.b	$89,$FF,$79,$E5,$EE,$6C,$EE,$74,$72,$8E,$2F,$F3,$BF,$AB,$F7,$FD
	.dc.b	$1F,$E0,$BC,$E9,$3E,$08,$CB,$F0,$1E,$A4,$70,$6C,$87,$F4,$7F,$77
	.dc.b	$E5,$FC,$97,$E3,$BC,$D9,$3E,$C8,$CB,$F3,$1F,$49,$1C,$1B,$21,$FD
	.dc.b	$4F,$EA,$FE,$F7,$BE,$F5,$BF,$37,$F2,$12,$BE,$EC,$E5,$B3,$F2,$23
	.dc.b	$11,$FE,$36,$6B,$FD,$5C,$FF,$3E,$B8,$7F,$C3,$4F,$0E,$4F,$BB,$1A
	.dc.b	$9F,$19,$FA,$CF,$AA,$FD,$6F,$D4,$49,$FD,$53,$E8,$F9,$09,$1F,$8E
	.dc.b	$C1,$1E,$52,$47,$F9,$AF,$C1,$FF,$38,$F9,$3C,$23,$FC,$3D,$D3,$E4
	.dc.b	$AF,$E0,$1E,$EF,$F2,$F9,$C8,$BF,$F3,$CF,$CA,$68,$91,$FC,$16,$5F
	.dc.b	$4D,$F3,$3E,$13,$E3,$12,$7C,$CF,$F9,$F5,$E9,$3F,$DB,$B7,$FF,$61
	.dc.b	$37,$2F,$F5,$AB,$67,$7F,$37,$FB,$5F,$B3,$77,$7F,$17,$FA,$97,$4D
	.dc.b	$35,$12,$FF,$67,$5C,$1B,$1F,$A2,$8C,$47,$F9,$C1,$B9,$7D,$6E,$E4
	.dc.b	$63,$3F,$F2,$A3,$6B,$E3,$62,$F8,$6F,$97,$E2,$FC,$CF,$95,$95,$F1
	.dc.b	$E3,$2F,$C9,$8C,$BF,$43,$8B,$FD,$4F,$3F,$53,$C3,$FA,$BF,$33,$2B
	.dc.b	$E8,$46,$5F,$A5,$19,$79,$70,$FF,$08,$7E,$7F,$AD,$F7,$BE,$2A,$47
	.dc.b	$D9,$8F,$C0,$98,$78,$C6,$1D,$1F,$7D,$D7,$07,$D8,$9F,$E5,$1F,$73
	.dc.b	$FE,$67,$AF,$C2,$78,$BF,$8C,$33,$EE,$48,$FE,$53,$3B,$7F,$1F,$C6
	.dc.b	$8C,$61,$F8,$4F,$9E,$FC,$AF,$C5,$FE,$5B,$E9,$A6,$1F,$91,$8C,$BF
	.dc.b	$93,$8C,$BF,$0A,$3F,$AF,$2F,$A8,$47,$E2,$7E,$7A,$4F,$F5,$19,$CB
	.dc.b	$67,$FA,$8C,$62,$3F,$DA,$CD,$7F,$80,$1D,$63,$67,$82,$38,$61,$FE
	.dc.b	$5E,$6F,$47,$F9,$77,$91,$CF,$FD,$FE,$BF,$47,$D5,$FB,$4E,$89,$97
	.dc.b	$EB,$E3,$F6,$E6,$1F,$BF,$18,$7E,$6F,$9F,$FB,$CD,$24,$60,$87,$D0
	.dc.b	$DD,$3F,$68,$FE,$07,$ED,$FA,$77,$FE,$0B,$CD,$2F,$E3,$46,$5F,$DE
	.dc.b	$6E,$8F,$8E,$48,$CB,$B7,$F8,$69,$F4,$DF,$E2,$3E,$4F,$F8,$4F,$DD
	.dc.b	$7D,$CF,$E0,$B8,$BE,$37,$EE,$FF,$3B,$87,$F9,$4F,$1A,$31,$3F,$E8
	.dc.b	$7B,$E3,$CF,$CF,$0E,$59,$FF,$95,$5B,$3A,$8F,$F0,$F5,$FF,$97,$71
	.dc.b	$FE,$01,$C0,$F3,$FF,$C4,$92,$F9,$9A,$7E,$82,$25,$FE,$71,$AF,$EE
	.dc.b	$E5,$F4,$1F,$92,$FC,$81,$E1,$7F,$87,$1B,$5F,$C9,$5A,$FE,$30,$38
	.dc.b	$7F,$1E,$EB,$2F,$F5,$4B,$76,$FE,$A4,$D9,$FF,$3A,$79,$1F,$C5,$5F
	.dc.b	$F5,$BE,$A8,$C4,$7F,$40,$8C,$74,$7D,$39,$CE,$1F,$43,$F2,$07,$85
	.dc.b	$9A,$98,$74,$BD,$E5,$F9,$DB,$DC,$FF,$8F,$7E,$9D,$EC,$7F,$8F,$1B
	.dc.b	$1F,$D1,$5A,$FC,$16,$BF,$AE,$73,$97,$FB,$21,$B7,$FD,$C5,$B3,$FE
	.dc.b	$C9,$FA,$8F,$EB,$95,$F7,$3D,$4F,$F1,$0F,$21,$A4,$AF,$F1,$CC,$65
	.dc.b	$FE,$2F,$1F,$C7,$8F,$C1,$FA,$76,$32,$43,$F0,$0E,$5F,$D7,$45,$AF
	.dc.b	$CC,$F1,$FF,$80,$F2,$FF,$6F,$78,$9F,$C8,$7D,$2F,$29,$30,$F1,$B5
	.dc.b	$FD,$C3,$FC,$C0,$FC,$AF,$D3,$7F,$9E,$FD,$27,$ED,$1F,$A0,$ED,$FD
	.dc.b	$37,$CF,$7E,$68,$D4,$7E,$92,$77,$DD,$FD,$D5,$B8,$FE,$45,$AF,$F1
	.dc.b	$E6,$BF,$E1,$F7,$3A,$FF,$87,$1B,$A7,$FB,$E5,$B3,$FE,$0B,$D5,$F3
	.dc.b	$DE,$2F,$82,$FD,$6F,$D3,$4A,$FF,$80,$18,$F3,$CF,$FD,$1D,$18,$CB
	.dc.b	$A4,$62,$3C,$8D,$7C,$8D,$77,$7E,$D8,$E1,$5D,$63,$C5,$3F,$C0,$BD
	.dc.b	$7A,$FF,$88,$B3,$F9,$73,$8A,$FB,$35,$F2,$BD,$FF,$E0,$BF,$2F,$30
	.dc.b	$F3,$1F,$ED,$BC,$53,$B8,$F9,$91,$86,$C7,$94,$8F,$2C,$7E,$DC,$71
	.dc.b	$8F,$E1,$9F,$3D,$F6,$BD,$FF,$61,$23,$AB,$EF,$F7,$E4,$78,$1A,$FF
	.dc.b	$BD,$3C,$06,$9F,$5F,$BF,$FC,$B5,$F1,$BF,$BE,$8C,$50,$46,$5F,$F7
	.dc.b	$C4,$F0,$AE,$71,$83,$FB,$2F,$C4,$72,$7F,$DE,$26,$4B,$A7,$F1,$37
	.dc.b	$67,$17,$F3,$7D,$7F,$E6,$88,$FB,$AD,$7F,$06,$D7,$E8,$36,$3F,$47
	.dc.b	$F0,$37,$73,$49,$C7,$F9,$FF,$DC,$95,$FE,$76,$D7,$F3,$91,$89,$FE
	.dc.b	$8F,$F3,$F9,$7F,$3D,$19,$CB,$F1,$0D,$7F,$4C,$D7,$77,$E5,$8C,$47
	.dc.b	$81,$E5,$7F,$C1,$BE,$0B,$B4,$E6,$AF,$67,$F6,$19,$CC,$2F,$FA,$6E
	.dc.b	$1F,$89,$6D,$FF,$53,$F8,$1D,$C6,$31,$F6,$F9,$58,$3A,$F6,$2B,$18
	.dc.b	$73,$9F,$41,$1F,$90,$DC,$3F,$B0,$7C,$F7,$06,$CF,$8F,$DD,$F0,$9D
	.dc.b	$BE,$66,$C6,$4F,$A5,$1F,$14,$8E,$0F,$EE,$FF,$3D,$C2,$BF,$B7,$3F
	.dc.b	$1B,$C5,$7D,$17,$F5,$DA,$FF,$7B,$B6,$BF,$BA,$7C,$F6,$E3,$1B,$7F
	.dc.b	$9F,$DD,$4F,$FA,$EE,$F9,$F4,$FE,$35,$B8,$7F,$BF,$AB,$75,$FC,$E9
	.dc.b	$7C,$1F,$53,$E5,$FC,$F5,$AF,$1C,$5A,$FF,$03,$4D,$C7,$C5,$36,$BF
	.dc.b	$C4,$33,$69,$87,$BA,$BC,$98,$78,$2F,$C9,$18,$3F,$49,$4F,$A4,$61
	.dc.b	$F8,$C8,$C3,$F2,$7E,$C7,$FD,$77,$7F,$C6,$A7,$2D,$AE,$2F,$FA,$ED
	.dc.b	$B1,$FE,$17,$6B,$FE,$64,$6B,$FE,$E5,$F9,$EF,$F6,$BC,$F5,$6C,$7F
	.dc.b	$A8,$7D,$7F,$F7,$A4,$E7,$C6,$F3,$56,$B5,$FE,$04,$D9,$FF,$35,$57
	.dc.b	$0B,$FF,$26,$CE,$5F,$E3,$1F,$8C,$7F,$8E,$6B,$D8,$D8,$DE,$1F,$0E
	.dc.b	$31,$87,$F0,$F7,$53,$FE,$7E,$6B,$FE,$FF,$6B,$DB,$D9,$FA,$7C,$DF
	.dc.b	$8D,$3F,$57,$F4,$17,$1F,$C9,$C7,$EB,$B2,$C5,$FF,$91,$DF,$DF,$FB
	.dc.b	$1F,$C4,$BE,$32,$F6,$9F,$E0,$3E,$13,$1C,$C4,$7F,$A8,$F8,$1F,$E0
	.dc.b	$2F,$9E,$C6,$5F,$EC,$DB,$47,$E0,$7E,$8F,$FD,$D3,$7F,$7D,$B5,$BF
	.dc.b	$47,$5F,$0D,$2B,$FD,$F5,$5A,$AF,$F5,$0B,$5F,$E8,$C1,$C3,$FD,$25
	.dc.b	$EB,$C5,$FE,$7A,$C5,$F3,$DD,$6B,$CC,$FE,$D7,$E2,$E6,$1E,$CA,$7C
	.dc.b	$DE,$2F,$FD,$93,$D8,$BF,$8B,$F3,$39,$CB,$B3,$F2,$3D,$4F,$93,$F3
	.dc.b	$17,$3E,$8D,$C7,$E3,$36,$FC,$CF,$C7,$F8,$19,$5F,$3B,$BF,$C6,$FC
	.dc.b	$98,$CD,$FF,$9E,$77,$47,$EC,$9B,$3F,$2A,$30,$7F,$3C,$A6,$3C,$B1
	.dc.b	$FF,$4F,$C6,$2B,$E1,$7C,$C7,$E0,$BE,$2F,$F0,$6D,$8C,$9F,$1E,$32
	.dc.b	$D3,$F9,$1F,$C5,$FF,$21,$3F,$17,$F0,$76,$32,$7C,$FF,$93,$C6,$BF
	.dc.b	$89,$7F,$23,$ED,$7E,$4B,$E5,$AD,$7D,$56,$D7,$E3,$DB,$A7,$DF,$6C
	.dc.b	$FF,$B8,$79,$51,$8A,$75,$7F,$96,$A3,$F3,$FE,$27,$E8,$3D,$AB,$87
	.dc.b	$E0,$DB,$7F,$9D,$6E,$3F,$8D,$7F,$4F,$ED,$7E,$A3,$D7,$B7,$F8,$B6
	.dc.b	$DF,$E9,$5B,$7F,$95,$7F,$A6,$7D,$B2,$E9,$F5,$F9,$EE,$1F,$E4,$0F
	.dc.b	$31,$5B,$8F,$F5,$D4,$62,$BF,$A5,$4E,$71,$FD,$39,$FC,$F1,$C3,$FD
	.dc.b	$45,$D6,$3F,$82,$E4,$EF,$3F,$E3,$8C,$3F,$C6,$9B,$7F,$3C,$9F,$E0
	.dc.b	$BE,$C4,$9F,$D4,$7C,$BD,$BE,$07,$F4,$AF,$B0,$FE,$62,$FF,$5F,$6B
	.dc.b	$FE,$94,$6B,$BF,$33,$FC,$5D,$8F,$FA,$3B,$E1,$3F,$89,$9F,$E1,$4F
	.dc.b	$67,$FC,$33,$8F,$DD,$8C,$7A,$5F,$DD,$71,$FF,$06,$4F,$F1,$6A,$FE
	.dc.b	$E2,$F7,$F6,$34,$7F,$6F,$9D,$E5,$F1,$B8,$51,$88,$F4,$4F,$92,$FF
	.dc.b	$AC,$C7,$BB,$2F,$F1,$27,$CC,$74,$C7,$C0,$F8,$FB,$A4,$78,$FE,$7A
	.dc.b	$DF,$C6,$CF,$BD,$78,$C9,$5F,$73,$3F,$EB,$B5,$F2,$E7,$E1,$79,$7C
	.dc.b	$B8,$FF,$9B,$4F,$E7,$34,$FE,$75,$87,$E1,$FE,$55,$E5,$FF,$86,$5B
	.dc.b	$1F,$17,$D6,$C4,$C8,$FC,$5C,$EA,$FE,$23,$5F,$F0,$C3,$5F,$F6,$6B
	.dc.b	$BC,$3C,$7B,$FF,$91,$FD,$1E,$95,$F1,$F2,$FF,$22,$61,$C3,$AB,$C8
	.dc.b	$7D,$DF,$98,$B1,$E9,$8F,$F1,$26,$46,$92,$3C,$EF,$E1,$7E,$07,$8E
	.dc.b	$1D,$DF,$81,$E1,$CA,$FC,$27,$CC,$5A,$FA,$58,$2B,$B9,$2B,$E1,$B6
	.dc.b	$7D,$5E,$E7,$1D,$BD,$FA,$23,$E7,$78,$73,$0F,$C9,$79,$B7,$0E,$F3
	.dc.b	$EC,$24,$FD,$06,$C7,$C7,$F9,$8E,$18,$F6,$7F,$C5,$9F,$37,$F9,$C9
	.dc.b	$FF,$54,$37,$8B,$FC,$ED,$AE,$BB,$26,$5F,$C0,$AF,$71,$FE,$38,$BF
	.dc.b	$F2,$6C,$62,$75,$7F,$E7,$FF,$95,$FD,$87,$A9,$FB,$16,$CF,$EB,$5B
	.dc.b	$5F,$75,$B9,$FC,$4F,$3F,$F7,$07,$EE,$7E,$62,$DF,$ED,$62,$DF,$E1
	.dc.b	$66,$D7,$FA,$CF,$E7,$BF,$BC,$FD,$07,$FB,$17,$E2,$1E,$0F,$F8,$0A
	.dc.b	$73,$82,$72,$BE,$81,$F4,$2F,$5F,$F4,$F5,$AA,$5F,$28,$C3,$70,$ED
	.dc.b	$ED,$32,$F1,$47,$FB,$5F,$2E,$B8,$F4,$8D,$CE,$B1,$FF,$6E,$61,$CD
	.dc.b	$B8,$73,$7D,$15,$CF,$2F,$36,$DC,$67,$CB,$5B,$CF,$CA,$DB,$EF,$3C
	.dc.b	$8F,$81,$7C,$9C,$1C,$B1,$E9,$8F,$8E,$30,$71,$CF,$D7,$96,$B1,$ED
	.dc.b	$3E,$8B,$FC,$53,$C0,$FE,$9B,$7F,$49,$B3,$ED,$7F,$A8,$FC,$5B,$B1
	.dc.b	$EF,$72,$8C,$36,$33,$3F,$97,$9C,$BF,$99,$BF,$F1,$AF,$BD,$E2,$6C
	.dc.b	$67,$DD,$FF,$76,$B6,$75,$18,$DF,$EE,$B7,$7F,$C5,$36,$BF,$11,$F4
	.dc.b	$2F,$6B,$FC,$9F,$7C,$30,$F9,$37,$BB,$FC,$4B,$C7,$81,$87,$19,$DD
	.dc.b	$47,$E9,$E5,$E8,$C3,$F2,$F7,$C3,$1E,$F9,$79,$30,$E1,$5E,$1C,$3E
	.dc.b	$B6,$3D,$5E,$79,$78,$F2,$E9,$87,$19,$DE,$3D,$1F,$F1,$0F,$D0,$FE
	.dc.b	$8D,$FF,$3D,$B5,$FE,$B8,$DE,$4F,$E7,$6F,$F9,$22,$37,$C9,$FE,$72
	.dc.b	$9C,$B7,$8F,$FC,$0F,$5C,$5B,$EE,$E6,$FC,$3E,$2D,$EB,$5B,$D4,$C3
	.dc.b	$6B,$CB,$58,$33,$32,$32,$7F,$B2,$FB,$7F,$DC,$AF,$F8,$07,$4E,$DE
	.dc.b	$27,$C7,$D7,$FB,$3C,$AB,$6F,$F8,$60,$C7,$95,$F9,$E6,$9F,$A0,$8D
	.dc.b	$BF,$E2,$C6,$A3,$FD,$D3,$F4,$2F,$27,$D7,$EE,$C9,$FF,$1D,$4E,$5B
	.dc.b	$39,$7E,$8C,$FC,$7F,$41,$5B,$9F,$F0,$43,$71,$FF,$00,$CE,$AF,$84
	.dc.b	$E7,$87,$E9,$F9,$58,$F2,$7E,$6F,$BB,$32,$F1,$C6,$5F,$C1,$C6,$5E
	.dc.b	$F1,$96,$9C,$0F,$C4,$F9,$5E,$0F,$1F,$EC,$FB,$92,$7D,$F1,$97,$F1
	.dc.b	$F1,$97,$D7,$19,$69,$C0,$FC,$AF,$B9,$EC,$F9,$DF,$C2,$FB,$92,$7E
	.dc.b	$5C,$65,$FC,$DC,$65,$F8,$F1,$96,$9C,$0F,$D0,$FB,$BF,$3F,$E3,$9F
	.dc.b	$2D,$F9,$2F,$7A,$57,$F9,$C6,$72,$D9,$CB,$FA,$89,$E5,$12,$34,$43
	.dc.b	$F5,$BE,$F7,$E7,$BD,$D9,$1F,$B2,$7C,$5E,$F4,$8C,$10,$FF,$24,$60
	.dc.b	$8F,$D5,$7C,$FF,$F2,$09,$CC,$71,$7F,$CA,$D1,$9E,$8F,$F1,$AA,$F5
	.dc.b	$38,$92,$7F,$DF,$B0,$F7,$BE,$27,$BF,$D9,$93,$F0,$E3,$C3,$E3,$C8
	.dc.b	$E2,$F9,$3D,$DF,$CD,$7D,$0F,$F8,$B6,$7F,$4D,$97,$F9,$1B,$E8,$5C
	.dc.b	$DF,$C0,$DB,$3F,$E9,$FA,$E0,$FF,$99,$DB,$5C,$6B,$81,$DB,$AE,$57
	.dc.b	$FA,$92,$B8,$BF,$B5,$1D,$E8,$CC,$7E,$E6,$31,$3F,$D7,$BC,$86,$DF
	.dc.b	$C1,$B1,$F9,$3E,$F7,$D0,$3E,$8F,$89,$31,$F9,$31,$97,$E5,$C7,$D6
	.dc.b	$F0,$BE,$BF,$BD,$27,$E7,$C6,$5F,$A3,$19,$7E,$CC,$7D,$EF,$53,$EF
	.dc.b	$FB,$D2,$7E,$BC,$65,$FB,$31,$97,$F0,$31,$97,$F0,$71,$9C,$7F,$07
	.dc.b	$3F,$C3,$FE,$6B,$FD,$0F,$5F,$82,$F0,$FF,$1A,$67,$E0,$24,$7F,$B4
	.dc.b	$27,$6F,$FA,$A5,$AF,$CF,$E0,$46,$23,$FD,$97,$B3,$DF,$FC,$CF,$BF
	.dc.b	$27,$F2,$31,$F9,$C3,$0F,$E6,$A3,$0F,$E2,$BC,$8F,$B9,$F4,$3C,$E3
	.dc.b	$07,$FB,$FB,$9D,$EF,$FE,$34,$FE,$A5,$09,$E0,$FF,$41,$6E,$FF,$95
	.dc.b	$9E,$13,$D0,$E1,$8F,$2A,$E3,$1F,$EE,$E6,$BE,$48,$FE,$95,$18,$87
	.dc.b	$2F,$E3,$3A,$3F,$73,$F0,$12,$7F,$63,$1F,$BC,$30,$FE,$EA,$30,$FB
	.dc.b	$B7,$F9,$BF,$0D,$FC,$0B,$29,$39,$E0,$4F,$DD,$3F,$C7,$35,$C2,$7E
	.dc.b	$D4,$62,$77,$1F,$E3,$EC,$9B,$FF,$0E,$E0,$93,$F0,$E7,$2D,$9C,$BF
	.dc.b	$C4,$27,$E3,$44,$8E,$8F,$C4,$23,$FE,$95,$F7,$9F,$43,$F8,$CE,$0F
	.dc.b	$F6,$5F,$CA,$FE,$E6,$BF,$DE,$0D,$8F,$F7,$0D,$70,$E4,$FF,$07,$36
	.dc.b	$BA,$5F,$E4,$5F,$F2,$8B,$63,$51,$F9,$9F,$37,$FC,$29,$83,$E3,$A9
	.dc.b	$FA,$58,$93,$FA,$17,$E4,$11,$FF,$32,$FD,$03,$BB,$C7,$F9,$1B,$E6
	.dc.b	$A9,$F9,$29,$D5,$7E,$35,$AF,$FA,$DD,$AF,$C1,$6B,$FC,$6D,$AF,$F9
	.dc.b	$B3,$C7,$73,$7C,$26,$C7,$E9,$34,$FE,$81,$3F,$E3,$1F,$A1,$78,$3B
	.dc.b	$89,$5B,$BE,$2B,$87,$FD,$FB,$7C,$23,$F4,$9F,$42,$F5,$FF,$5B,$F4
	.dc.b	$3F,$D2,$67,$F5,$0D,$7F,$C5,$D3,$B7,$8C,$FF,$8B,$E3,$11,$FE,$D7
	.dc.b	$6B,$F8,$16,$BF,$CD,$FA,$1E,$97,$F8,$FB,$A2,$4F,$E8,$63,$F8,$E1
	.dc.b	$F9,$D2,$3F,$C6,$5A,$FB,$7C,$1E,$91,$C2,$3F,$85,$F7,$B9,$FF,$C0
	.dc.b	$31,$B3,$FA,$5E,$1E,$27,$C0,$43,$F2,$FE,$8F,$F2,$DF,$43,$FC,$A7
	.dc.b	$E9,$C9,$FD,$B4,$65,$FE,$58,$FE,$CC,$8F,$FB,$E7,$E1,$5F,$1F,$FA
	.dc.b	$5D,$3F,$CA,$19,$F9,$7E,$83,$F3,$27,$F9,$BE,$B7,$DF,$6F,$86,$1F
	.dc.b	$C4,$B5,$5F,$08,$FE,$53,$7C,$1B,$5F,$A9,$8C,$47,$EC,$DA,$FF,$6D
	.dc.b	$6B,$FE,$63,$7C,$FF,$05,$FA,$CF,$9F,$93,$D2,$3C,$86,$1F,$DE,$C6
	.dc.b	$1D,$10,$FA,$13,$D1,$FA,$B8,$7F,$88,$DD,$73,$34,$F1,$63,$FE,$99
	.dc.b	$9E,$2D,$9F,$E1,$BE,$AF,$C5,$F7,$3F,$80,$F7,$4B,$F8,$31,$97,$BC
	.dc.b	$65,$E7,$19,$69,$C0,$F2,$FC,$FF,$D1,$F7,$3F,$D4,$1E,$D7,$F1,$9C
	.dc.b	$FE,$CB,$69,$AF,$DE,$AF,$B1,$EE,$70,$F8,$D9,$5F,$52,$3F,$AD,$FC
	.dc.b	$67,$EE,$B6,$FA,$1E,$07,$F8,$7E,$31,$5D,$23,$1E,$17,$EE,$E3,$07
	.dc.b	$FA,$66,$77,$97,$F9,$16,$3F,$98,$7C,$5F,$46,$9C,$0F,$F7,$29,$4F
	.dc.b	$F5,$13,$B8,$FC,$BF,$6F,$D3,$9F,$D4,$74,$B6,$FE,$5D,$BF,$E3,$AB
	.dc.b	$5A,$6E,$B5,$E4,$E2,$BE,$DF,$BF,$71,$FE,$D7,$8B,$E8,$B1,$9F,$E9
	.dc.b	$AB,$58,$F7,$9C,$FC,$27,$CE,$C7,$F3,$ED,$C7,$FA,$6F,$D0,$6B,$19
	.dc.b	$5F,$8E,$F8,$2B,$5F,$D4,$9B,$AF,$F5,$66,$CF,$ED,$3A,$D1,$E0,$F3
	.dc.b	$64,$FE,$63,$E0,$6E,$1F,$9C,$CF,$FA,$44,$F9,$2B,$D3,$82,$F7,$8F
	.dc.b	$CB,$C9,$87,$0C,$7F,$B2,$FD,$06,$B2,$F1,$36,$3F,$80,$37,$6F,$E3
	.dc.b	$8F,$C4,$F7,$3F,$DC,$B3,$FB,$9B,$46,$23,$F2,$B1,$8B,$FE,$28,$D8
	.dc.b	$FE,$08,$DC,$7F,$BC,$FC,$FF,$5A,$FD,$BB,$6B,$FB,$DB,$67,$73,$FE
	.dc.b	$1D,$F9,$8F,$F4,$B3,$FE,$01,$AF,$F3,$CC,$61,$FF,$02,$7B,$08,$C4
	.dc.b	$7F,$94,$5A,$FF,$9F,$5E,$15,$F8,$DE,$4F,$F5,$FD,$F7,$EA,$FC,$3C
	.dc.b	$62,$FF,$9F,$61,$D7,$E0,$3F,$20,$DB,$FE,$23,$F2,$B8,$C9,$54,$93
	.dc.b	$A3,$D8,$E2,$FD,$33,$C2,$7D,$D7,$D6,$73,$5F,$46,$31,$97,$A7,$E8
	.dc.b	$B9,$36,$FF,$CA,$DF,$3C,$CB,$EC,$62,$8C,$4F,$CB,$8C,$4E,$A7,$31
	.dc.b	$F3,$DB,$5F,$B4,$FA,$97,$57,$EC,$6B,$8E,$1F,$E7,$0A,$56,$A3,$FC
	.dc.b	$D1,$F3,$BF,$ED,$76,$F8,$FF,$B4,$F6,$BE,$75,$FF,$05,$C1,$7B,$7F
	.dc.b	$B7,$18,$AF,$04,$66,$7F,$83,$57,$2B,$FF,$5F,$4E,$5F,$F6,$07,$CF
	.dc.b	$FF,$0E,$C3,$F1,$4D,$7F,$C9,$8D,$77,$DF,$F2,$23,$86,$1F,$E4,$EF
	.dc.b	$C6,$FA,$50,$4E,$AF,$1E,$D2,$B7,$C3,$F9,$5E,$EF,$B5,$C9,$FB,$7C
	.dc.b	$7F,$2D,$18,$26,$1F,$D7,$34,$FC,$64,$7F,$02,$7F,$90,$63,$FD,$1B
	.dc.b	$E2,$BF,$68,$DA,$FF,$35,$76,$A7,$2D,$8C,$C7,$F2,$4F,$35,$F3,$FD
	.dc.b	$2B,$59,$7F,$36,$9C,$DF,$FA,$87,$E7,$FF,$DE,$EF,$E2,$5B,$3F,$B5
	.dc.b	$6B,$FE,$77,$6B,$FE,$F1,$77,$C1,$FC,$F6,$9F,$37,$F8,$B9,$CD,$FF
	.dc.b	$87,$27,$2F,$87,$E1,$BD,$15,$B6,$C7,$F0,$7F,$89,$F9,$A4,$7F,$57
	.dc.b	$C0,$8C,$65,$ED,$F9,$7F,$DC,$47,$C3,$7E,$27,$AB,$F1,$7E,$5A,$E1
	.dc.b	$FE,$42,$CD,$18,$C3,$E3,$37,$5F,$63,$EB,$F8,$2F,$D8,$ED,$FE,$74
	.dc.b	$C3,$E4,$46,$08,$FF,$93,$BB,$99,$71,$8C,$37,$1F,$ED,$33,$A9,$FA
	.dc.b	$95,$FD,$7A,$31,$3F,$67,$E5,$BE,$F7,$C4,$FD,$F6,$DF,$AD,$B3,$B8
	.dc.b	$F7,$4F,$F1,$BF,$96,$FC,$B9,$23,$E5,$B6,$7F,$0D,$F2,$1B,$9F,$8C
	.dc.b	$FE,$37,$D9,$FC,$77,$CA,$59,$F9,$4D,$AF,$C7,$37,$7F,$C2,$36,$7F
	.dc.b	$04,$D9,$FF,$74,$F0,$23,$F3,$5E,$07,$E6,$FD,$9B,$7F,$5D,$B7,$F9
	.dc.b	$B6,$E1,$F9,$17,$F4,$5E,$C7,$E8,$FD,$5B,$7F,$7D,$B7,$FA,$36,$DF
	.dc.b	$E6,$5F,$E8,$7F,$3E,$BF,$E6,$CD,$BF,$E7,$6D,$BF,$EA,$ED,$7F,$85
	.dc.b	$0F,$05,$1B,$C3,$FC,$45,$C4,$E2,$EF,$8F,$DF,$6E,$3F,$89,$6E,$3F
	.dc.b	$86,$7F,$7D,$E3,$C9,$FC,$D7,$97,$6B,$FA,$EF,$CF,$FF,$1C,$F4,$BF
	.dc.b	$C1,$7E,$07,$F5,$E8,$C7,$A1,$FE,$25,$D7,$F7,$1F,$03,$FC,$24,$FF
	.dc.b	$16,$C6,$72,$FD,$C4,$63,$D0,$E5,$EA,$59,$FE,$51,$7F,$E4,$3E,$26
	.dc.b	$3F,$E1,$D9,$C9,$FC,$B6,$33,$5D,$BE,$5B,$9C,$FF,$AD,$32,$46,$63
	.dc.b	$0E,$B5,$FF,$27,$CE,$A7,$AC,$7C,$FC,$3F,$AE,$C7,$76,$C3,$5F,$F1
	.dc.b	$6F,$CB,$7C,$3E,$39,$7E,$1F,$22,$DF,$66,$EB,$FE,$59,$CB,$F9,$5B
	.dc.b	$EF,$CB,$AE,$5F,$E3,$8A,$FF,$05,$FC,$93,$E1,$29,$F3,$3D,$5B,$9F
	.dc.b	$A3,$81,$5B,$8F,$85,$E7,$6B,$5F,$7B,$74,$F1,$36,$7F,$D8,$FA,$FF
	.dc.b	$91,$E7,$E9,$5F,$F1,$C9,$EC,$DB,$F9,$3F,$2B,$E6,$F1,$E5,$7D,$1F
	.dc.b	$80,$B5,$F3,$9B,$87,$C1,$6C,$F8,$1F,$7F,$87,$23,$ED,$7C,$AD,$BF
	.dc.b	$F1,$7E,$4A,$FC,$6F,$B5,$9C,$7E,$63,$70,$F5,$B6,$FE,$C6,$9F,$E5
	.dc.b	$19,$FC,$56,$1F,$C1,$DF,$6F,$87,$F9,$8F,$33,$E8,$B2,$3F,$19,$F3
	.dc.b	$F7,$0F,$C4,$B7,$2F,$B8,$D9,$F9,$6F,$F8,$8F,$E6,$7F,$C9,$F5,$FE
	.dc.b	$05,$6A,$69,$3B,$3F,$C9,$FF,$11,$C5,$B7,$F9,$DF,$94,$FD,$6F,$A5
	.dc.b	$FA,$E6,$C6,$4F,$D1,$46,$5A,$7E,$97,$CE,$FD,$A9,$FB,$6F,$96,$B1
	.dc.b	$92,$65,$FA,$68,$CB,$FE,$30,$F9,$BF,$EE,$BF,$11,$FE,$BC,$F0,$7F
	.dc.b	$8B,$C6,$25,$F0,$E9,$FF,$12,$7C,$F3,$E7,$DE,$BF,$FB,$06,$B5,$7F
	.dc.b	$E2,$9C,$39,$46,$21,$18,$C3,$BF,$B2,$CB,$C5,$4C,$BB,$63,$D6,$F7
	.dc.b	$5A,$AE,$77,$C5,$F3,$CE,$9A,$73,$CB,$CF,$A7,$1C,$BC,$53,$E7,$BE
	.dc.b	$59,$F9,$E7,$39,$F8,$AF,$85,$6B,$1E,$D7,$B8,$C6,$1D,$F3,$F4,$3D
	.dc.b	$73,$E2,$7F,$9D,$72,$F9,$D1,$FD,$18,$F5,$C6,$6B,$D9,$19,$F8,$A7
	.dc.b	$63,$DC,$FF,$3F,$46,$1B,$19,$9F,$F5,$64,$E5,$FF,$57,$60,$F9,$E7
	.dc.b	$9B,$E3,$6C,$E9,$AF,$96,$78,$4E,$CF,$95,$3C,$34,$F9,$0D,$D7,$F0
	.dc.b	$FF,$3E,$F6,$7F,$B1,$D6,$AF,$89,$31,$FA,$FC,$2B,$E1,$19,$AE,$0F
	.dc.b	$DC,$3F,$17,$8F,$77,$F1,$13,$C1,$D6,$5E,$86,$CE,$DE,$38,$7E,$43
	.dc.b	$0F,$06,$3E,$6C,$BF,$47,$7B,$C3,$96,$5E,$23,$F3,$F7,$C3,$1F,$CC
	.dc.b	$F9,$4F,$41,$F4,$9F,$E0,$2E,$EF,$F6,$A3,$79,$7F,$D1,$D5,$B8,$C5
	.dc.b	$FF,$55,$B7,$C8,$62,$B5,$3F,$D7,$9A,$B0,$EA,$E6,$B8,$BB,$F9,$DF
	.dc.b	$ED,$6D,$ED,$E9,$06,$FA,$32,$74,$43,$B6,$7E,$8A,$E7,$1F,$EB,$AF
	.dc.b	$03,$E8,$3F,$9B,$6B,$FE,$22,$6B,$FC,$A7,$8B,$FD,$AC,$FF,$29,$7E
	.dc.b	$0B,$4F,$EF,$65,$3C,$27,$C5,$F8,$6F,$8F,$C8,$FD,$D5,$F2,$9F,$2B
	.dc.b	$27,$C8,$7F,$6C,$C3,$FC,$0B,$18,$F2,$38,$DF,$09,$FF,$04,$E1,$CF
	.dc.b	$0F,$F7,$53,$63,$F6,$C7,$EA,$27,$83,$FB,$98,$D3,$63,$3C,$0C,$3F
	.dc.b	$D9,$95,$A8,$C5,$73,$BF,$BD,$C0,$F9,$3F,$2E,$98,$F9,$39,$7E,$6C
	.dc.b	$D3,$A8,$F2,$E5,$F0,$E3,$D3,$3A,$9C,$B5,$F3,$47,$17,$E6,$CF,$17
	.dc.b	$F1,$5A,$FD,$4E,$87,$B5,$E6,$F9,$29,$5F,$5A,$72,$D9,$FB,$B1,$88
	.dc.b	$FE,$23,$18,$23,$B2,$1F,$7B,$9B,$F5,$5E,$D7,$E9,$3D,$57,$E7,$3F
	.dc.b	$C9,$F8,$BE,$4D,$7E,$BF,$E4,$BE,$49,$B9,$D7,$27,$F9,$3D,$FC,$7D
	.dc.b	$F2,$C7,$ED,$9F,$E7,$58,$C9,$1E,$44,$3F,$59,$F4,$DD,$7E,$57,$EB
	.dc.b	$FC,$A4,$9F,$F1,$D5,$27,$51,$F8,$8B,$F4,$C7,$81,$B1,$99,$E0,$D9
	.dc.b	$E2,$FE,$82,$76,$FC,$5E,$67,$D3,$7B,$FD,$19,$3E,$27,$E1,$78,$12
	.dc.b	$3B,$3E,$1C,$7F,$82,$C7,$F5,$1C,$3E,$6C,$FC,$4D,$BF,$8D,$7C,$07
	.dc.b	$1A,$F1,$5F,$0E,$0F,$A3,$9A,$BE,$87,$4C,$BF,$86,$FE,$73,$E9,$7E
	.dc.b	$A7,$37,$84,$93,$F0,$5F,$27,$27,$D9,$24,$7F,$1B,$F9,$0F,$E0,$BF
	.dc.b	$45,$FD,$53,$3F,$C0,$FE,$B7,$FA,$AA,$FC,$43,$63,$13,$BE,$6F,$F5
	.dc.b	$9B,$FD,$72,$BF,$B8,$35,$CF,$5F,$F1,$48,$C4,$03,$CD,$C3,$5C,$C7
	.dc.b	$F3,$7F,$C0,$9E,$F7,$F3,$9C,$5F,$27,$7E,$12,$61,$FC,$EE,$72,$D9
	.dc.b	$CB,$EC,$9E,$51,$CF,$D3,$E9,$E7,$4A,$E5,$19,$7A,$46,$5F,$8B,$1E
	.dc.b	$3F,$57,$C9,$F2,$32,$7C,$51,$97,$C9,$19,$7E,$6C,$65,$F9,$F1,$9D
	.dc.b	$FE,$9C,$FE,$DF,$F3,$AF,$A5,$C1,$FD,$7B,$47,$C1,$FC,$33,$51,$31
	.dc.b	$FF,$51,$57,$07,$FD,$D0,$DE,$A5,$FF,$B9,$E7,$E4,$7C,$97,$CD,$FA
	.dc.b	$09,$18,$21,$F8,$47,$2F,$E1,$A3,$F1,$7F,$0F,$FD,$C3,$DB,$7D,$C3
	.dc.b	$E2,$30,$DB,$FE,$A7,$D0,$BE,$31,$FC,$2A,$30,$D8,$FC,$51,$1E,$2F
	.dc.b	$07,$F2,$FF,$11,$F8,$5F,$94,$FC,$37,$C3,$49,$FC,$24,$65,$FC,$34
	.dc.b	$65,$FC,$FC,$7C,$FE,$1F,$C7,$7C,$9C,$9F,$97,$19,$69,$0F,$D3,$FC
	.dc.b	$9F,$EB,$3E,$07,$F2,$DF,$23,$27,$DF,$19,$7D,$CF,$E9,$48,$E0,$D9
	.dc.b	$0F,$D9,$FC,$9F,$E0,$F8,$BF,$45,$E8,$C9,$FB,$F1,$96,$90,$FD,$DF
	.dc.b	$91,$F8,$BD,$7F,$55,$F2,$32,$7F,$0F,$19,$69,$0F,$E9,$5F,$BA,$FE
	.dc.b	$DF,$CD,$75,$35,$FE,$64,$DA,$FE,$6C,$D7,$F9,$23,$5F,$E0,$47,$E5
	.dc.b	$23,$F3,$C4,$7F,$C6,$95,$FE,$2B,$6F,$BF,$FE,$39,$8C,$47,$92,$31
	.dc.b	$EF,$E5,$91,$C0,$E7,$1F,$D8,$71,$FE,$27,$89,$A4,$F1,$FF,$A6,$3E
	.dc.b	$F2,$3E,$1F,$83,$FF,$16,$FD,$13,$7F,$F1,$8B,$71,$FE,$C6,$D7,$FC
	.dc.b	$3E,$D7,$FD,$2C,$E7,$C2,$FE,$60,$DF,$07,$E2,$B6,$7F,$94,$7D,$1F
	.dc.b	$F2,$58,$C4,$1A,$62,$35,$7E,$62,$4F,$F9,$0A,$3F,$97,$78,$6F,$93
	.dc.b	$FF,$24,$B6,$7F,$BD,$B5,$FF,$1B,$B5,$FF,$57,$39,$C7,$FA,$43,$71
	.dc.b	$FB,$AD,$9F,$E8,$1F,$B0,$BD,$AF,$F2,$7F,$EE,$9F,$E0,$C9,$78,$EE
	.dc.b	$4E,$0D,$CF,$73,$0F,$F1,$AD,$7F,$98,$38,$0D,$24,$6C,$FF,$19,$BD
	.dc.b	$E9,$F6,$6E,$BF,$1F,$C1,$7E,$37,$E5,$A6,$1F,$E5,$AC,$3F,$CA,$3B
	.dc.b	$FF,$94,$1F,$F2,$BC,$7E,$6B,$2F,$0D,$FF,$97,$A7,$33,$B8,$C1,$D9
	.dc.b	$E8,$FE,$6F,$C4,$C4,$79,$A2,$F0,$FF,$8D,$36,$FE,$47,$3E,$AF,$83
	.dc.b	$FD,$17,$91,$23,$CD,$1F,$30,$F1,$11,$F8,$ED,$79,$1C,$0F,$D7,$FA
	.dc.b	$3F,$53,$D5,$F5,$12,$3E,$7B,$ED,$F5,$64,$7D,$26,$BF,$EE,$69,$FF
	.dc.b	$55,$3F,$EE,$4E,$EF,$DF,$FC,$67,$F7,$AC,$FF,$CF,$31,$8A,$DE,$FF
	.dc.b	$BC,$AD,$5F,$13,$FD,$F9,$F1,$7C,$1A,$64,$FF,$71,$94,$FF,$07,$3B
	.dc.b	$9F,$C2,$F3,$FF,$87,$67,$ED,$35,$FF,$7C,$F0,$FE,$57,$8B,$F1,$B3
	.dc.b	$9A,$FC,$6F,$C0,$DA,$34,$95,$F9,$4F,$6F,$FD,$27,$AF,$FA,$52,$FF
	.dc.b	$2B,$18,$9F,$F6,$BF,$13,$C8,$C6,$7E,$2A,$B7,$C2,$FA,$12,$7F,$D4
	.dc.b	$D9,$7F,$91,$E3,$18,$FE,$1F,$E5,$B8,$CE,$6B,$F9,$CC,$74,$FD,$CD
	.dc.b	$EB,$53,$C3,$D9,$FB,$1F,$96,$C6,$5F,$EB,$AA,$DB,$6B,$7B,$2B,$78
	.dc.b	$FF,$A8,$2F,$6F,$FA,$ED,$AF,$E9,$DA,$FF,$AE,$9A,$B4,$47,$2D,$3F
	.dc.b	$45,$F2,$DD,$9B,$97,$97,$CF,$F6,$1E,$AE,$1F,$F5,$6C,$64,$98,$F8
	.dc.b	$E7,$EE,$61,$FE,$2A,$9F,$F5,$FB,$67,$51,$FD,$4B,$E2,$BA,$B4,$FF
	.dc.b	$75,$FE,$05,$FE,$E6,$6F,$6F,$DA,$5F,$B6,$FF,$A6,$5C,$92,$FF,$DC
	.dc.b	$35,$C5,$B5,$F9,$35,$19,$7F,$B3,$8F,$A4,$FB,$9A,$7F,$9D,$DB,$7F
	.dc.b	$E6,$DF,$31,$F6,$2F,$3B,$F1,$B8,$71,$E8,$B8,$70,$8F,$F3,$A3,$6F
	.dc.b	$8D,$6A,$A6,$1F,$E7,$9C,$7C,$11,$F9,$8A,$DC,$E4,$FE,$FB,$7D,$70
	.dc.b	$F0,$47,$FB,$1A,$3F,$29,$1E,$C6,$BC,$2B,$F6,$95,$FD,$D7,$0E,$DE
	.dc.b	$77,$AF,$0F,$F0,$2F,$CD,$62,$64,$7F,$B1,$67,$58,$7F,$AC,$DA,$FF
	.dc.b	$84,$2F,$A4,$FC,$C8,$FE,$A3,$CC,$F9,$AF,$CE,$9F,$9E,$F9,$AB,$5F
	.dc.b	$E1,$48,$B5,$FE,$24,$9B,$AA,$BE,$96,$3F,$E6,$17,$E2,$53,$1F,$F0
	.dc.b	$87,$8D,$C2,$76,$FB,$8F,$83,$5C,$6B,$D5,$18,$33,$DD,$F1,$0F,$D9
	.dc.b	$FC,$D5,$BE,$F8,$7F,$87,$67,$F6,$FF,$27,$9C,$3F,$C7,$6F,$CC,$EA
	.dc.b	$97,$FE,$30,$FA,$AB,$5F,$EE,$7C,$D5,$FC,$03,$EA,$73,$AF,$F5,$57
	.dc.b	$F1,$BF,$44,$69,$FC,$33,$F6,$7F,$F6,$06,$5F,$0B,$15,$6E,$30,$D8
	.dc.b	$FA,$8D,$7F,$42,$D7,$FC,$92,$D7,$FD,$98,$F1,$BF,$A5,$1F,$E0,$CA
	.dc.b	$7C,$1F,$92,$FE,$9F,$86,$9C,$C7,$C8,$CF,$F0,$4D,$E3,$F5,$7D,$3F
	.dc.b	$E4,$99,$1C,$AB,$F7,$D8,$FC,$79,$F8,$1F,$27,$F0,$7B,$7E,$17,$D4
	.dc.b	$DC,$38,$36,$DA,$FA,$34,$F2,$4E,$63,$19,$23,$F7,$78,$B3,$A7,$CF
	.dc.b	$F9,$3B,$97,$CF,$9D,$3F,$9F,$C9,$18,$7F,$44,$D7,$F8,$7B,$5F,$F3
	.dc.b	$BB,$BC,$7C,$7F,$39,$DA,$7D,$35,$FE,$FD,$CF,$E6,$4F,$E0,$7E,$53
	.dc.b	$F0,$5F,$33,$6B,$E4,$36,$F8,$9F,$5D,$C6,$1F,$01,$FE,$1D,$F9,$C7
	.dc.b	$D6,$9D,$F4,$A3,$15,$F1,$BD,$5D,$F2,$4C,$7E,$B7,$CA,$DC,$7B,$EB
	.dc.b	$F8,$18,$FC,$B7,$CD,$7E,$5F,$D7,$B7,$F7,$5B,$7C,$39,$11,$8C,$FE
	.dc.b	$AB,$6B,$E1,$4E,$9F,$C3,$5F,$57,$F1,$1E,$BF,$E9,$7D,$3F,$D3,$36
	.dc.b	$32,$7E,$36,$32,$D3,$F2,$1E,$77,$EB,$7E,$67,$F5,$CD,$8C,$9F,$95
	.dc.b	$8C,$B4,$FC,$27,$1F,$E2,$8F,$F0,$0F,$CA,$BC,$EC,$4E,$8F,$C4,$73
	.dc.b	$AB,$FA,$B3,$63,$31,$83,$F8,$07,$CC,$F2,$E0,$F9,$0F,$DA,$E4,$95
	.dc.b	$FA,$9F,$2E,$D7,$DC,$E6,$70,$7F,$42,$6D,$7F,$37,$E7,$FC,$04,$E4
	.dc.b	$D7,$1F,$E1,$8F,$E2,$1F,$31,$6B,$F6,$DF,$31,$F8,$AE,$1F,$E2,$8D
	.dc.b	$8F,$E4,$F8,$A7,$C0,$48,$F0,$46,$5B,$AF,$E5,$0E,$59,$FA,$5F,$CB
	.dc.b	$E1,$27,$F8,$7B,$FD,$15,$EE,$E6,$7F,$BE,$E7,$F9,$B8,$F8,$04,$9F
	.dc.b	$0B,$FE,$36,$1F,$2F,$C7,$9F,$F8,$44,$62,$7F,$A0,$FE,$A9,$70,$FD
	.dc.b	$3B,$6B,$7F,$33,$FD,$E6,$31,$5F,$E0,$97,$85,$7F,$67,$6B,$9C,$3F
	.dc.b	$BA,$C6,$23,$FA,$E3,$C9,$FF,$5D,$E1,$FE,$39,$AE,$07,$C2,$F6,$7F
	.dc.b	$8C,$F5,$C8,$F8,$BF,$33,$6F,$F8,$19,$FC,$5A,$7F,$1E,$49,$FC,$9B
	.dc.b	$63,$F7,$76,$F8,$7E,$5A,$7C,$47,$FC,$D7,$F1,$C9,$A7,$CB,$F9,$9B
	.dc.b	$8F,$F7,$6F,$71,$5B,$8F,$F0,$34,$62,$BF,$D0,$93,$9C,$7F,$C6,$B3
	.dc.b	$9A,$DC,$E7,$1F,$AB,$EE,$76,$BF,$3B,$FE,$7E,$FD,$CF,$E3,$47,$F9
	.dc.b	$85,$BE,$1F,$9F,$E9,$B5,$C1,$32,$F1,$D7,$F9,$A3,$33,$49,$3F,$E5
	.dc.b	$3F,$B7,$73,$71,$CB,$F9,$7E,$09,$CB,$FD,$D7,$E8,$3F,$CF,$8D,$F8
	.dc.b	$1F,$C8,$7C,$E7,$F9,$2A,$73,$7F,$59,$AF,$F9,$59,$A8,$8E,$1B,$3F
	.dc.b	$EB,$1D,$5F,$78,$F8,$9F,$F4,$9B,$FE,$83,$C7,$FB,$BB,$96,$ED,$A9
	.dc.b	$DB,$63,$27,$EC,$3E,$47,$F5,$07,$EA,$7E,$6E,$C6,$49,$97,$E2,$A3
	.dc.b	$2F,$E4,$BE,$83,$FD,$4F,$E7,$7F,$B3,$3E,$57,$D9,$29,$CA,$F8,$B4
	.dc.b	$FC,$1F,$A3,$7D,$37,$E5,$72,$E5,$7F,$17,$0E,$38,$70,$8C,$72,$FF
	.dc.b	$86,$1E,$98,$F8,$2B,$6F,$5C,$3A,$4E,$8E,$39,$76,$7C,$D7,$CF,$0E
	.dc.b	$D1,$89,$F1,$E5,$E6,$ED,$FF,$14,$D6,$B2,$EF,$7D,$0E,$C7,$13,$C1
	.dc.b	$5C,$73,$F0,$4F,$09,$D5,$73,$C7,$C5,$9F,$A7,$1E,$EF,$B6,$33,$5D
	.dc.b	$CF,$F5,$0D,$FC,$E6,$C7,$C5,$6E,$5F,$E1,$DF,$8E,$FF,$76,$7D,$35
	.dc.b	$F8,$23,$49,$5F,$E8,$1C,$1E,$17,$FB,$46,$30,$D7,$C2,$D7,$E9,$35
	.dc.b	$FB,$4F,$1D,$3E,$43,$73,$DF,$17,$CE,$FA,$67,$C1,$FE,$2A,$F7,$3F
	.dc.b	$22,$F8,$61,$D2,$38,$56,$F0,$E7,$3F,$4B,$1F,$4D,$71,$37,$14,$D5
	.dc.b	$F4,$7F,$30,$F6,$E1,$65,$E4,$9E,$19,$79,$1E,$37,$D1,$FC,$77,$3B
	.dc.b	$E7,$1E,$D7,$E9,$3C,$2F,$F2,$D4,$62,$BF,$CD,$ED,$7F,$C6,$2F,$F8
	.dc.b	$2D,$CC,$F1,$E5,$FF,$25,$4E,$EB,$9C,$E6,$97,$E1,$F4,$B0,$BE,$9B
	.dc.b	$FD,$C1,$3F,$E7,$E7,$F8,$63,$7A,$F7,$1A,$8F,$F2,$1B,$70,$FF,$39
	.dc.b	$3C,$23,$33,$C3,$E2,$BC,$50,$EF,$F9,$0F,$F5,$2F,$A1,$F1,$F9,$07
	.dc.b	$73,$AC,$7F,$D2,$B5,$C7,$9B,$FC,$E8,$D3,$FC,$ED,$3D,$1F,$E5,$91
	.dc.b	$FE,$79,$E0,$7D,$33,$E0,$BE,$27,$17,$F8,$A3,$D9,$BE,$B7,$F8,$EF
	.dc.b	$1F,$EB,$97,$B6,$DE,$AA,$46,$08,$7F,$9D,$30,$EF,$3F,$45,$B1,$F7
	.dc.b	$3D,$8C,$5B,$DF,$FA,$7C,$CF,$42,$9A,$89,$7F,$84,$8C,$BE,$68,$CB
	.dc.b	$E1,$AE,$8E,$4F,$C5,$E7,$F4,$BB,$1F,$41,$E0,$F6,$65,$7F,$B3,$E7
	.dc.b	$2D,$9C,$BF,$92,$9E,$51,$23,$44,$3F,$2F,$EF,$FE,$47,$D7,$7D,$37
	.dc.b	$F4,$29,$FC,$DF,$C0,$7C,$03,$51,$9E,$6F,$F6,$9D,$70,$AD,$1F,$A4
	.dc.b	$AD,$B0,$FF,$7C,$D7,$1E,$4F,$ED,$55,$C6,$7F,$57,$F0,$1D,$BD,$FF
	.dc.b	$9F,$EE,$C9,$E3,$19,$7F,$63,$19,$7F,$D5,$38,$7F,$B0,$9D,$90,$CC
	.dc.b	$FD,$AF,$53,$D7,$F0,$52,$7C,$AF,$BB,$C2,$91,$82,$1D,$9F,$C1,$FC
	.dc.b	$06,$A3,$07,$F0,$47,$E3,$F3,$48,$F8,$2F,$E2,$BB,$FE,$5F,$1D,$9F
	.dc.b	$E1,$71,$87,$E3,$F1,$7D,$02,$3F,$C4,$3E,$03,$F6,$FF,$49,$FC,$EF
	.dc.b	$93,$FD,$DD,$F4,$DF,$B2,$6F,$26,$BB,$66,$1F,$E1,$99,$CB,$67,$FD
	.dc.b	$BD,$5C,$5F,$F0,$1D,$F5,$6D,$FE,$F2,$3F,$A3,$57,$F8,$8D,$A9,$CF
	.dc.b	$8C,$BF,$F3,$EC,$F7,$3F,$8F,$AF,$81,$FE,$3E,$D8,$D4,$F1,$F8,$1C
	.dc.b	$9A,$A9,$3A,$AE,$AF,$FB,$89,$7A,$7D,$3C,$E9,$3B,$8E,$E6,$1E,$11
	.dc.b	$87,$A7,$A9,$E4,$F8,$09,$18,$21,$DD,$CB,$E1,$8F,$59,$8D,$BE,$AC
	.dc.b	$E2,$7E,$17,$63,$E8,$FE,$84,$7C,$4F,$80,$F8,$67,$ED,$63,$33,$BE
	.dc.b	$1F,$DB,$57,$EA,$A7,$F9,$8B,$FE,$3A,$9D,$3F,$EC,$D8,$CB,$F7,$BD
	.dc.b	$DF,$9F,$ED,$C8,$F9,$51,$F4,$8C,$3F,$2E,$30,$FD,$93,$EF,$7C,$DF
	.dc.b	$F5,$9F,$69,$F4,$CF,$91,$C3,$6B,$87,$32,$FF,$32,$6A,$A4,$EA,$32
	.dc.b	$FB,$8E,$E5,$9E,$4D,$8D,$73,$7E,$3F,$D1,$FC,$5F,$C6,$4A,$FC,$34
	.dc.b	$7E,$38,$C3,$F8,$78,$C3,$F5,$35,$70,$7E,$53,$E9,$A4,$E6,$3E,$A1
	.dc.b	$E5,$E0,$FB,$4D,$AE,$33,$F4,$63,$F5,$1E,$EF,$E7,$BE,$96,$4F,$C4
	.dc.b	$8F,$DD,$98,$7F,$35,$18,$74,$7C,$18,$E6,$D3,$EF,$70,$FE,$AB,$CB
	.dc.b	$91,$EB,$8F,$C5,$1E,$52,$3C,$0F,$C3,$67,$FB,$2F,$87,$91,$F0,$E3
	.dc.b	$F2,$47,$B4,$8F,$03,$F1,$D8,$7F,$86,$F9,$FF,$D4,$3E,$2D,$F8,$3F
	.dc.b	$F1,$25,$27,$FA,$4E,$AB,$FC,$F1,$F2,$49,$3D,$0F,$CF,$F0,$BE,$57
	.dc.b	$FC,$35,$6E,$9F,$F5,$A4,$E7,$DA,$FF,$14,$D6,$A7,$2F,$F8,$8D,$B1
	.dc.b	$A2,$5F,$F1,$09,$D2,$E2,$6B,$FC,$8E,$3F,$C6,$FE,$2E,$1F,$E2,$ED
	.dc.b	$7D,$D8,$7F,$5E,$9E,$15,$CB,$C3,$FC,$44,$7F,$2A,$8C,$57,$EF,$1A
	.dc.b	$FE,$FB,$E9,$5C,$5F,$E3,$6F,$A5,$F2,$C9,$B3,$A8,$FD,$B3,$5E,$53
	.dc.b	$FE,$34,$8C,$47,$F7,$F6,$BF,$EC,$27,$86,$DF,$E3,$C6,$E5,$9A,$F1
	.dc.b	$65,$FE,$48,$9C,$DF,$FB,$95,$A9,$B8,$7C,$98,$78,$0F,$A4,$FF,$28
	.dc.b	$FD,$2B,$D1,$A7,$F9,$52,$75,$5F,$C4,$5A,$FF,$22,$6B,$FE,$19,$6B
	.dc.b	$FE,$DC,$6A,$C5,$F4,$3F,$E1,$38,$FF,$38,$F0,$F4,$DB,$F1,$91,$8C
	.dc.b	$BF,$C8,$D5,$BB,$E3,$3F,$8E,$98,$FF,$5D,$8C,$B6,$7F,$CC,$71,$88
	.dc.b	$FF,$17,$35,$FF,$78,$B5,$1E,$78,$FF,$2F,$3F,$8C,$3C,$FF,$47,$F9
	.dc.b	$E3,$CB,$EE,$48,$FC,$D4,$65,$F1,$46,$5F,$F3,$14,$F4,$C3,$DF,$7F
	.dc.b	$12,$BF,$53,$E0,$6E,$73,$5B,$E1,$F2,$DE,$F0,$FD,$45,$FC,$6F,$73
	.dc.b	$F6,$3E,$1C,$9F,$85,$19,$7F,$5D,$19,$7C,$71,$96,$9C,$0F,$DA,$7D
	.dc.b	$1F,$D2,$F4,$3F,$D2,$7E,$1C,$9F,$9D,$1F,$5B,$D0,$FF,$4B,$FD,$34
	.dc.b	$9F,$A5,$1F,$E9,$EF,$8A,$61,$FC,$4E,$7F,$09,$DB,$F4,$14,$EB,$83
	.dc.b	$DF,$39,$8C,$4F,$03,$EB,$B8,$DB,$FD,$5D,$1F,$6B,$E1,$B8,$BF,$E6
	.dc.b	$D6,$BF,$EE,$73,$91,$63,$FD,$7A,$7E,$5B,$DB,$FC,$9E,$1F,$94,$F0
	.dc.b	$1E,$07,$F2,$9E,$3F,$C7,$56,$9F,$E4,$71,$9C,$3E,$FF,$B7,$97,$F8
	.dc.b	$CF,$D2,$DC,$BF,$47,$39,$C7,$AC,$E7,$99,$EC,$53,$5D,$6F,$2B,$1A
	.dc.b	$7E,$7E,$73,$F4,$B8,$D3,$F3,$19,$7F,$39,$8C,$6F,$EB,$C5,$18,$C3
	.dc.b	$F9,$BB,$E5,$FA,$5F,$F1,$DC,$7E,$B5,$B8,$74,$A7,$C8,$6F,$D6,$97
	.dc.b	$B7,$CF,$18,$AF,$F6,$15,$F2,$C7,$C5,$F3,$B9,$6E,$1C,$4F,$E8,$91
	.dc.b	$97,$FA,$5B,$E5,$F4,$3F,$C5,$BC,$4B,$FF,$04,$D7,$F8,$DA,$31,$3A
	.dc.b	$9C,$E1,$FC,$09,$B8,$7C,$3F,$86,$E9,$87,$F5,$36,$D7,$BD,$B3,$B9
	.dc.b	$FF,$1A,$7D,$27,$FB,$A5,$FF,$14,$36,$BF,$C8,$ED,$7F,$D7,$CD,$7F
	.dc.b	$40,$EE,$7F,$90,$CB,$E9,$70,$EB,$C3,$FC,$53,$97,$F1,$96,$FA,$4F
	.dc.b	$C3,$CF,$F9,$1D,$B1,$FC,$67,$CD,$7D,$2B,$E3,$78,$D6,$B2,$FE,$63
	.dc.b	$7C,$F9,$B2,$7E,$67,$0E,$37,$8C,$7B,$4F,$F9,$7A,$B8,$47,$F9,$27
	.dc.b	$0F,$0C,$7F,$B7,$5A,$FE,$FD,$F0,$47,$FA,$97,$EB,$BF,$DE,$2F,$43
	.dc.b	$FD,$31,$8A,$33,$5E,$BF,$47,$D5,$5F,$DE,$5B,$1F,$EE,$BC,$BF,$C1
	.dc.b	$39,$7F,$9C,$9B,$5B,$8F,$D8,$76,$F4,$C7,$FD,$DB,$87,$27,$F8,$04
	.dc.b	$66,$BF,$C3,$7F,$41,$FB,$23,$F6,$7F,$4B,$63,$FC,$35,$17,$2F,$D3
	.dc.b	$4D,$BE,$BE,$57,$8F,$1F,$F3,$C6,$1E,$98,$C1,$1F,$83,$F5,$1F,$1B
	.dc.b	$0F,$F1,$25,$70,$7F,$AA,$5A,$75,$A7,$F8,$A3,$BB,$B1,$30,$F8,$CD
	.dc.b	$BF,$DD,$7C,$87,$DD,$C7,$F8,$65,$7F,$8C,$E3,$1D,$5E,$72,$47,$CF
	.dc.b	$6D,$7F,$03,$FA,$4D,$CF,$F8,$ED,$FC,$AF,$D2,$1F,$C3,$36,$3F,$D1
	.dc.b	$CD,$9F,$F5,$6B,$5F,$E8,$81,$D6,$9F,$CC,$FE,$89,$F0,$5D,$B0,$FE
	.dc.b	$55,$18,$C3,$FD,$51,$7C,$1F,$F2,$BC,$67,$0F,$EF,$87,$5A,$E3,$7F
	.dc.b	$0F,$87,$EB,$12,$FF,$1C,$DA,$FE,$5B,$6F,$8D,$E1,$5F,$EE,$A8,$C3
	.dc.b	$7A,$7C,$BF,$49,$E6,$6C,$64,$E1,$19,$69,$FE,$C1,$A7,$D2,$7E,$BD
	.dc.b	$B1,$FE,$B7,$6E,$1F,$ED,$66,$BF,$29,$CE,$BF,$27,$E6,$3F,$84,$C6
	.dc.b	$2B,$E2,$C6,$2F,$E5,$B6,$3D,$8F,$CE,$F8,$1F,$9F,$F2,$B6,$7E,$03
	.dc.b	$71,$F8,$AD,$C7,$EC,$3F,$C2,$FE,$8F,$E9,$4F,$FB,$A5,$AF,$F8,$CB
	.dc.b	$99,$18,$8F,$F4,$23,$63,$3C,$1F,$80,$F9,$0E,$8D,$C7,$E9,$7C,$AF
	.dc.b	$E1,$3D,$7F,$C2,$B6,$32,$7D,$78,$CB,$4C,$E3,$F7,$A3,$2D,$8F,$BD
	.dc.b	$EB,$7E,$43,$D2,$FC,$8B,$63,$27,$E0,$E3,$2D,$3F,$0D,$E9,$7E,$63
	.dc.b	$DE,$FC,$CB,$63,$27,$E2,$E3,$2D,$3F,$09,$C1,$F8,$32,$3F,$83,$6E
	.dc.b	$1F,$96,$F9,$3F,$BF,$1F,$A6,$6C,$FE,$5D,$B3,$B8,$FF,$31,$F0,$FF
	.dc.b	$9D,$37,$FA,$7E,$DF,$D9,$3F,$5F,$E3,$DA,$D9,$F6,$78,$9C,$1F,$B4
	.dc.b	$F1,$3F,$00,$D9,$DE,$5F,$73,$87,$F0,$86,$5F,$EB,$16,$8F,$DE,$FC
	.dc.b	$97,$E2,$77,$B7,$CB,$4F,$C3,$C6,$63,$16,$DB,$18,$F4,$7F,$24,$49
	.dc.b	$DD,$59,$F0,$E7,$F9,$73,$FA,$3C,$D3,$AB,$FD,$2B,$E4,$B1,$3F,$E3
	.dc.b	$08,$46,$37,$FE,$96,$DA,$FF,$23,$BF,$DB,$7E,$8F,$71,$EB,$6D,$F3
	.dc.b	$F9,$3D,$54,$AF,$F7,$0F,$A9,$6B,$C0,$DC,$3F,$AC,$B6,$7F,$CA,$AF
	.dc.b	$F0,$0E,$89,$1F,$EE,$AF,$92,$B7,$FC,$93,$B5,$5F,$D9,$BE,$13,$38
	.dc.b	$FF,$64,$6E,$1F,$E6,$06,$DF,$FB,$D9,$FF,$2F,$7C,$67,$F9,$A2,$7F
	.dc.b	$D6,$CD,$9F,$8D,$F2,$5A,$DE,$57,$FB,$F3,$D6,$B5,$F2,$DB,$8F,$F7
	.dc.b	$46,$CF,$F9,$C1,$FF,$36,$7D,$17,$F9,$36,$3F,$D4,$ED,$A7,$B9,$FE
	.dc.b	$7A,$FA,$3D,$F0,$4B,$FC,$98,$FA,$3F,$C4,$3F,$EC,$DC,$39,$39,$E6
	.dc.b	$E1,$8F,$FB,$16,$30,$6E,$FF,$C6,$F7,$CA,$7F,$A7,$FC,$E7,$F9,$A2
	.dc.b	$BF,$AE,$C7,$FB,$42,$30,$E1,$AF,$F6,$26,$BF,$EA,$77,$7A,$7F,$85
	.dc.b	$BE,$73,$FC,$87,$18,$AF,$F4,$9C,$7F,$A3,$7E,$49,$F0,$D7,$DF,$C3
	.dc.b	$6B,$F6,$7B,$AB,$71,$87,$FD,$F4,$D9,$E1,$97,$E7,$FE,$43,$FC,$56
	.dc.b	$7F,$8B,$3E,$8E,$D7,$4E,$34,$E2,$BF,$4D,$57,$5F,$F2,$8F,$CE,$38
	.dc.b	$3F,$BE,$D3,$E2,$1F,$43,$FE,$C7,$AD,$79,$9F,$EB,$F8,$FF,$7C,$FD
	.dc.b	$0B,$E8,$DD,$8D,$F5,$97,$FB,$FB,$1E,$2F,$5C,$3A,$67,$E0,$CF,$BC
	.dc.b	$EE,$31,$3D,$B1,$EB,$87,$4A,$61,$D3,$0E,$31,$8C,$38,$BC,$F3,$F1
	.dc.b	$E7,$E7,$CF,$CD,$3E,$33,$45,$39,$4F,$7C,$3B,$67,$E3,$AE,$38,$F8
	.dc.b	$27,$37,$CA,$31,$8F,$8B,$1F,$06,$5C,$E3,$CB,$5D,$8F,$CA,$DF,$C6
	.dc.b	$7F,$C6,$6D,$7E,$3E,$5F,$9A,$7F,$DF,$3E,$DB,$E8,$F1,$F1,$BF,$ED
	.dc.b	$76,$C6,$67,$F4,$33,$97,$BF,$84,$E7,$E9,$8A,$31,$3A,$9C,$C6,$23
	.dc.b	$CA,$EB,$5F,$84,$DD,$FF,$10,$D9,$6C,$FA,$37,$BA,$9F,$9D,$5B,$7E
	.dc.b	$EB,$72,$F2,$65,$F6,$AF,$8D,$70,$88,$FD,$99,$DB,$F7,$70,$F0,$61
	.dc.b	$DE,$35,$87,$52,$5F,$59,$D6,$5E,$C8,$EA,$DA,$E9,$8F,$89,$FC,$57
	.dc.b	$4B,$A9,$EA,$F7,$E6,$FF,$4E,$F8,$77,$97,$E0,$5B,$A9,$39,$C7,$F9
	.dc.b	$4E,$3C,$9F,$E0,$ED,$7F,$B0,$46,$EB,$93,$67,$58,$7E,$7A,$5F,$47
	.dc.b	$FE,$D3,$BF,$E2,$2F,$F1,$66,$F6,$73,$8D,$42,$B9,$46,$1B,$3B,$38
	.dc.b	$7C,$1F,$A6,$9D,$FF,$16,$BF,$D7,$BF,$39,$E7,$D0,$EE,$75,$AF,$F2
	.dc.b	$FE,$2D,$4C,$C3,$FC,$32,$74,$8F,$8D,$C4,$FA,$37,$B9,$F2,$70,$FF
	.dc.b	$11,$FC,$5D,$F5,$0F,$0A,$DE,$3F,$E2,$4C,$39,$61,$BA,$93,$9A,$DB
	.dc.b	$F0,$F0,$FC,$C5,$F0,$8F,$EB,$78,$76,$6D,$F7,$C3,$C1,$ED,$FD,$5F
	.dc.b	$32,$57,$58,$CB,$F7,$A3,$2F,$73,$F6,$D9,$27,$71,$9E,$3F,$C8,$57
	.dc.b	$0E,$0F,$E3,$B9,$FC,$5B,$FA,$9D,$6F,$9D,$FC,$37,$AD,$32,$F9,$B3
	.dc.b	$96,$CE,$5E,$DC,$FF,$E4,$57,$24,$3F,$D7,$B3,$C5,$B3,$F9,$3F,$73
	.dc.b	$F3,$9E,$B3,$E8,$FF,$D9,$FA,$7E,$57,$17,$BB,$F9,$93,$73,$AE,$4F
	.dc.b	$F4,$B3,$72,$FD,$2C,$ED,$FF,$2A,$46,$70,$ED,$1C,$88,$7E,$AB,$DD
	.dc.b	$EB,$EA,$7E,$9F,$D8,$93,$CA,$32,$FE,$C2,$32,$FD,$3A,$F5,$4E,$A3
	.dc.b	$17,$FE,$D5,$79,$1F,$B8,$F4,$7D,$7F,$47,$23,$CE,$FB,$BD,$E9,$18
	.dc.b	$21,$D1,$FD,$E7,$BB,$A8,$C1,$FC,$11,$F8,$FC,$D2,$3E,$1B,$FC,$0F
	.dc.b	$BF,$E5,$F2,$D9,$FE,$17,$18,$7E,$4F,$1F,$D0,$23,$FC,$43,$DD,$FD
	.dc.b	$E7,$D0,$7F,$41,$E5,$FE,$19,$F4,$7F,$DA,$AB,$F2,$4D,$8C,$4E,$F9
	.dc.b	$7F,$98,$36,$BF,$C8,$B5,$B9,$FE,$46,$D8,$FE,$89,$3B,$4F,$1E,$B4
	.dc.b	$ED,$B5,$E6,$EB,$FE,$49,$3F,$E8,$4B,$FF,$17,$AF,$77,$26,$BD,$D9
	.dc.b	$3A,$8C,$BC,$23,$9F,$A7,$D3,$CD,$93,$D2,$32,$F6,$8C,$BF,$12,$3C
	.dc.b	$7E,$A7,$93,$DD,$93,$E4,$8C,$BE,$68,$CB,$F3,$23,$2F,$CE,$8C,$ED
	.dc.b	$F4,$A7,$F5,$BF,$43,$F0,$F3,$7B,$DF,$08,$CE,$D3,$1F,$87,$19,$7F
	.dc.b	$D7,$95,$C7,$E8,$FE,$4F,$B7,$27,$E4,$C7,$CC,$30,$FE,$02,$30,$FF
	.dc.b	$36,$E2,$7C,$E7,$F5,$BF,$69,$F4,$6F,$8E,$C3,$6F,$85,$7E,$38,$CC
	.dc.b	$C8,$C9,$0F,$E4,$95,$D2,$35,$1F,$C9,$CF,$C3,$CF,$E0,$BC,$FF,$C1
	.dc.b	$FC,$6C,$9F,$C1,$C6,$5F,$C2,$C6,$5F,$CD,$C7,$CC,$D7,$F1,$7F,$3B
	.dc.b	$27,$E3,$C6,$5F,$C6,$C6,$5F,$E0,$DB,$7D,$27,$E2,$7B,$BF,$94,$FA
	.dc.b	$39,$18,$21,$FA,$A7,$2F,$AF,$64,$FE,$C8,$CF,$27,$EC,$AB,$15,$F6
	.dc.b	$B8,$3F,$43,$E6,$48,$C1,$0F,$DB,$B9,$7C,$F1,$F8,$AD,$7F,$53,$EC
	.dc.b	$49,$FC,$2C,$65,$FD,$5C,$65,$F7,$C7,$E4,$AF,$FC,$35,$87,$FA,$77
	.dc.b	$C4,$BF,$04,$75,$7F,$90,$AC,$C4,$8C,$1F,$EA,$6F,$99,$FF,$0B,$3F
	.dc.b	$D7,$DB,$7F,$E1,$D6,$F3,$FF,$8B,$1F,$F0,$B3,$7D,$9F,$F1,$34,$6A
	.dc.b	$BF,$DA,$AD,$F0,$4F,$08,$79,$8D,$3F,$88,$78,$B6,$7F,$D5,$D1,$FD
	.dc.b	$70,$FE,$39,$3B,$F0,$FF,$C3,$D1,$27,$FC,$BD,$9F,$FA,$42,$B5,$1F
	.dc.b	$E1,$CF,$A3,$FF,$5B,$BF,$E1,$EF,$A3,$F1,$26,$C6,$A3,$F3,$CD,$AF
	.dc.b	$E6,$53,$C0,$DF,$17,$F8,$CE,$78,$57,$F9,$84,$38,$7F,$CC,$8D,$7F
	.dc.b	$CE,$8D,$77,$1C,$C8,$F9,$F1,$FF,$20,$78,$0F,$A2,$FF,$21,$7D,$1B
	.dc.b	$D1,$A6,$A2,$5F,$F8,$F2,$74,$D9,$CB,$FE,$21,$B7,$CF,$FF,$80,$78
	.dc.b	$7E,$BE,$1F,$70,$FF,$21,$C7,$F8,$16,$77,$59,$3F,$15,$3A,$61,$82
	.dc.b	$1D,$E7,$91,$FE,$40,$D7,$F1,$DF,$47,$F8,$A3,$BF,$B5,$2B,$F2,$D1
	.dc.b	$97,$FC,$99,$5C,$79,$7F,$C8,$ED,$8F,$C8,$CE,$DF,$05,$73,$7F,$3D
	.dc.b	$19,$BF,$F1,$DD,$71,$C7,$E1,$FC,$7F,$E7,$3C,$39,$87,$82,$3E,$19
	.dc.b	$87,$F5,$11,$87,$47,$EB,$5D,$5F,$69,$F9,$7E,$7F,$F9,$DF,$C1,$97
	.dc.b	$EC,$8F,$9F,$E7,$FF,$9E,$FE,$8E,$4F,$C0,$8F,$F4,$17,$C4,$3E,$8F
	.dc.b	$FD,$15,$3F,$EC,$FE,$DF,$F5,$94,$E7,$3F,$F3,$5C,$62,$B7,$AF,$F0
	.dc.b	$0A,$D5,$F1,$3F,$35,$F0,$FC,$1A,$64,$FF,$5C,$94,$FF,$67,$3B,$9F
	.dc.b	$C5,$FB,$6E,$3F,$DA,$9F,$EB,$4E,$E7,$83,$9C,$3F,$09,$57,$4F,$F5
	.dc.b	$65,$6A,$FE,$FF,$B7,$97,$97,$D1,$DC,$3F,$DD,$DC,$AA,$C7,$4F,$FB
	.dc.b	$06,$B5,$A7,$E7,$BD,$8C,$F8,$6F,$A3,$CE,$29,$CC,$61,$B9,$FE,$41
	.dc.b	$B9,$FE,$6D,$BA,$FE,$29,$B3,$FE,$9A,$FA,$27,$AB,$E0,$F9,$BF,$1C
	.dc.b	$62,$7F,$99,$46,$27,$51,$FE,$A9,$9D,$61,$FC,$D6,$B5,$3C,$0D,$57
	.dc.b	$09,$FD,$E5,$70,$C3,$F9,$D5,$F0,$9E,$57,$D7,$CF,$F8,$7F,$09,$E4
	.dc.b	$6C,$64,$FE,$8D,$19,$69,$99,$FE,$91,$19,$6C,$7F,$B3,$3E,$8B,$FD
	.dc.b	$92,$FE,$0A,$3F,$09,$F4,$58,$7F,$C1,$93,$71,$FC,$25,$2B,$0A,$A5
	.dc.b	$61,$E7,$E1,$7E,$0D,$46,$67,$94,$E7,$E8,$7E,$F4,$7F,$B8,$1B,$5F
	.dc.b	$E9,$1F,$A2,$7C,$66,$F8,$25,$FF,$3D,$AF,$15,$F0,$7F,$A0,$35,$F3
	.dc.b	$07,$0F,$09,$FC,$C6,$6F,$A2,$FF,$64,$5F,$8E,$73,$18,$AF,$F1,$85
	.dc.b	$F9,$AF,$F4,$D3,$97,$FD,$0D,$19,$BF,$EC,$73,$97,$E4,$79,$DF,$DB
	.dc.b	$F1,$FD,$2F,$56,$71,$FE,$E7,$E7,$FC,$0B,$FA,$DD,$4A,$D6,$1C,$67
	.dc.b	$31,$8A,$E1,$7F,$EE,$78,$E9,$7F,$E0,$0F,$9E,$FD,$79,$FB,$0F,$A2
	.dc.b	$B8,$7F,$84,$62,$DF,$F8,$46,$6E,$BF,$DE,$F1,$FF,$2E,$5F,$E2,$61
	.dc.b	$F4,$5E,$FB,$F4,$F9,$5F,$32,$31,$A7,$F8,$36,$31,$C1,$FC,$1E,$31
	.dc.b	$87,$3A,$DE,$3F,$E4,$28,$FF,$12,$7C,$7E,$E3,$18,$FF,$89,$63,$1E
	.dc.b	$C7,$F0,$A6,$D7,$03,$F3,$D8,$7F,$87,$7A,$7F,$87,$FD,$0F,$36,$CE
	.dc.b	$B1,$46,$31,$FF,$71,$46,$27,$FC,$6D,$5F,$9D,$FA,$23,$FC,$7B,$E8
	.dc.b	$9F,$01,$98,$C3,$6B,$ED,$CE,$9F,$D9,$60,$BE,$71,$AA,$E1,$2E,$0F
	.dc.b	$E6,$DC,$6C,$39,$4F,$FA,$0F,$0E,$2E,$F0,$FD,$B5,$6F,$0F,$95,$C7
	.dc.b	$E7,$F8,$DB,$5E,$46,$DF,$11,$7D,$E3,$39,$78,$B2,$47,$C1,$EA,$F8
	.dc.b	$5F,$45,$71,$F7,$B7,$1E,$98,$FA,$63,$18,$F3,$7F,$84,$60,$FA,$2F
	.dc.b	$99,$F2,$F8,$6E,$3F,$22,$74,$FC,$C9,$D3,$4D,$4F,$0C,$7F,$D5,$15
	.dc.b	$C1,$F1,$E6,$BF,$B5,$EE,$FD,$BF,$93,$B8,$FD,$86,$DF,$06,$E7,$F0
	.dc.b	$DF,$CE,$7D,$17,$D4,$8F,$8D,$F1,$F8,$E1,$FC,$33,$70,$FC,$1B,$70
	.dc.b	$F1,$BF,$8B,$F9,$6F,$C6,$7A,$F6,$BF,$12,$DA,$F4,$B7,$0F,$A8,$D9
	.dc.b	$FC,$0E,$A8,$C5,$7D,$C8,$C4,$FE,$53,$D6,$FC,$D7,$A3,$F9,$B6,$C6
	.dc.b	$4F,$CC,$C6,$5A,$7E,$73,$D1,$FD,$17,$B9,$FA,$36,$C6,$4F,$D0,$C6
	.dc.b	$5A,$7D,$BE,$0F,$D5,$7C,$97,$E0,$9B,$19,$3F,$53,$19,$69,$9C,$BE
	.dc.b	$DC,$65,$B1,$FC,$0B,$81,$ED,$FD,$63,$F7,$1E,$15,$9D,$70,$7E,$EF
	.dc.b	$C5,$FB,$6D,$8E,$5C,$1F,$78,$C6,$3F,$A7,$F0,$19,$7C,$C8,$FC,$6F
	.dc.b	$C8,$FE,$13,$6F,$E0,$CD,$C3,$C9,$1F,$88,$D7,$15,$CE,$D1,$9A,$F5
	.dc.b	$3A,$E3,$FF,$57,$4E,$A7,$DD,$A7,$E4,$8F,$E9,$AD,$32,$7F,$48,$F9
	.dc.b	$0C,$61,$FD,$1D,$B7,$FE,$40,$6C,$FF,$91,$DF,$E7,$1F,$44,$69,$FD
	.dc.b	$6B,$E4,$AC,$FF,$30,$9D,$3F,$D5,$BD,$3C,$E1,$FD,$55,$B8,$7F,$93
	.dc.b	$9B,$3F,$E5,$57,$F8,$1F,$44,$8F,$EB,$FF,$25,$6F,$F7,$73,$FD,$AB
	.dc.b	$E1,$31,$87,$F9,$61,$B7,$FE,$5F,$6D,$7F,$1C,$D1,$1F,$E5,$4F,$89
	.dc.b	$FE,$DD,$5F,$DB,$3D,$5C,$57,$F7,$66,$C6,$4F,$F1,$57,$B4,$AE,$38
	.dc.b	$7F,$71,$AD,$B6,$B7,$3A,$C7,$FC,$DB,$5F,$E6,$4F,$A1,$FF,$23,$57
	.dc.b	$F8,$B4,$7D,$0F,$DD,$C3,$FD,$8D,$3A,$7F,$CF,$91,$9B,$FF,$3D,$62
	.dc.b	$AD,$5F,$2E,$3E,$DB,$FF,$9D,$E3,$07,$AA,$77,$87,$38,$FF,$1F,$5F
	.dc.b	$3A,$E1,$39,$C7,$FC,$85,$F3,$4F,$A3,$FF,$0D,$FC,$CF,$F9,$06,$31
	.dc.b	$8F,$F8,$5A,$31,$3A,$8F,$EC,$33,$AA,$FF,$0C,$57,$FA,$23,$E3,$9F
	.dc.b	$0D,$73,$C3,$70,$FF,$12,$CE,$9B,$3A,$69,$FE,$26,$F8,$CF,$F4,$9F
	.dc.b	$EE,$C9,$FF,$18,$C6,$1B,$3F,$E3,$09,$D3,$C4,$C4,$7F,$97,$7E,$6F
	.dc.b	$FD,$17,$E4,$7F,$AC,$3E,$83,$FC,$9F,$18,$D1,$F1,$09,$FF,$34,$7D
	.dc.b	$03,$E8,$BF,$23,$97,$1B,$FF,$36,$61,$C7,$0E,$11,$8E,$3F,$F0,$6B
	.dc.b	$D3,$1E,$F1,$89,$DE,$3D,$7A,$BF,$D7,$91,$CA,$D8,$75,$C3,$8E,$7E
	.dc.b	$19,$CE,$3C,$E7,$CF,$9F,$9A,$7C,$07,$8E,$B8,$5F,$38,$C1,$AC,$BC
	.dc.b	$59,$EB,$1E,$93,$AC,$BC,$97,$C6,$F7,$3D,$8F,$1E,$1C,$4F,$F4,$03
	.dc.b	$83,$FD,$05,$3F,$3D,$FF,$3F,$99,$2F,$C5,$BD,$B7,$D1,$72,$8C,$36
	.dc.b	$33,$3F,$E8,$D9,$CB,$DB,$BF,$F1,$0D,$E0,$F1,$E0,$9C,$CF,$8A,$31
	.dc.b	$3A,$9E,$39,$FC,$06,$EF,$F8,$86,$DB,$47,$D1,$3D,$DF,$F4,$DE,$1C
	.dc.b	$71,$FA,$38,$70,$C3,$EE,$BC,$B0,$E3,$7C,$31,$ED,$7C,$09,$7F,$5F
	.dc.b	$1E,$CE,$31,$EC,$FD,$3C,$BD,$6E,$F0,$E8,$FE,$63,$2F,$14,$EA,$B8
	.dc.b	$9F,$94,$BE,$B5,$F8,$AF,$5D,$D1,$FE,$31,$8F,$F4,$DF,$81,$FE,$B8
	.dc.b	$EF,$FF,$2F,$F8,$3F,$CB,$67,$FD,$47,$5B,$7F,$C9,$6D,$9D,$D7,$1E
	.dc.b	$5F,$F3,$0D,$F1,$BD,$FA,$99,$7F,$CB,$BF,$45,$FE,$CE,$7F,$D5,$EF
	.dc.b	$F1,$06,$F6,$EF,$34,$EE,$7F,$CB,$6D,$CB,$FC,$EA,$F1,$8C,$D7,$18
	.dc.b	$D6,$3F,$ED,$4E,$EF,$87,$7E,$EF,$9A,$ED,$99,$F3,$62,$AD,$DF,$1E
	.dc.b	$2D,$D4,$AF,$E5,$71,$ED,$E3,$7D,$13,$DB,$F8,$18,$3D,$07,$C5,$61
	.dc.b	$B8,$FF,$5C,$BE,$0D,$BD,$54,$8C,$10,$FF,$44,$5F,$58,$D4,$7F,$A2
	.dc.b	$F0,$F1,$D7,$96,$DE,$CF,$CD,$E4,$79,$54,$D4,$4C,$3F,$03,$19,$7C
	.dc.b	$D1,$97,$F0,$71,$97,$C1,$5D,$9F,$F5,$7F,$43,$E6,$7C,$1E,$9C,$9C
	.dc.b	$C7,$D0,$9D,$47,$E4,$27,$4F,$FA,$94,$FE,$57,$3B,$3F,$8A,$B8,$3F
	.dc.b	$82,$B8,$9E,$71,$FC,$5B,$D3,$7D,$17,$F1,$8E,$5F,$F5,$EE,$8F,$67
	.dc.b	$E2,$1F,$97,$AD,$DD,$BE,$1C,$3F,$EB,$EB,$E1,$3F,$A6,$BD,$30,$FF
	.dc.b	$71,$D7,$05,$7D,$1F,$D6,$FB,$3D,$7D,$CF,$A9,$EB,$49,$E3,$19,$7F
	.dc.b	$67,$19,$7F,$CC,$57,$FE,$C4,$76,$43,$33,$F7,$FD,$1F,$57,$D1,$49
	.dc.b	$F2,$BE,$DF,$6E,$46,$08,$76,$7F,$11,$ED,$6A,$30,$7F,$06,$7E,$37
	.dc.b	$24,$8F,$82,$FE,$3F,$B7,$E5,$72,$59,$FE,$1B,$18,$7E,$3F,$B9,$F3
	.dc.b	$C8,$FF,$12,$F6,$FF,$6B,$F3,$9F,$D0,$39,$7F,$DE,$BF,$42,$E6,$FE
	.dc.b	$10,$DB,$FF,$7F,$D6,$E5,$12,$30,$FF,$BC,$27,$83,$FE,$E8,$6D,$62
	.dc.b	$77,$27,$8F,$88,$AD,$DF,$68,$C3,$E8,$ED,$F2,$FB,$9E,$6F,$72,$63
	.dc.b	$E4,$CB,$D7,$3A,$8C,$5F,$3C,$57,$C2,$7C,$D5,$F0,$3D,$2F,$83,$E6
	.dc.b	$49,$E2,$7B,$63,$4E,$33,$F4,$4E,$E3,$E0,$4F,$CA,$F5,$3E,$5F,$D1
	.dc.b	$49,$E1,$97,$F9,$E9,$B3,$97,$E4,$4F,$17,$27,$07,$5A,$7D,$89,$FD
	.dc.b	$77,$CF,$FC,$8A,$FA,$5E,$EF,$DA,$33,$B4,$8F,$0C,$65,$F6,$F7,$2B
	.dc.b	$96,$BF,$4F,$E8,$BF,$07,$ED,$4B,$F4,$C7,$E1,$8C,$3F,$7E,$30,$FE
	.dc.b	$23,$E8,$7F,$AC,$7B,$2F,$A2,$7C,$6E,$1B,$5B,$20,$DF,$F2,$55,$C6
	.dc.b	$B8,$1F,$9A,$B9,$5F,$5A,$74,$DB,$EA,$FF,$21,$E7,$57,$E7,$BC,$FF
	.dc.b	$CF,$FC,$5C,$AF,$AB,$19,$7E,$BC,$65,$FC,$E4,$7D,$0E,$0F,$D3,$FC
	.dc.b	$EC,$9F,$97,$C0,$9D,$47,$E0,$27,$4F,$FB,$C6,$74,$FE,$BB,$DC,$FD
	.dc.b	$8F,$CD,$C8,$C1,$0F,$D6,$39,$7D,$91,$FB,$C3,$1B,$FE,$F2,$71,$3E
	.dc.b	$4E,$0F,$DE,$F9,$B2,$30,$43,$F7,$0E,$5F,$34,$7E,$3B,$4F,$E0,$BE
	.dc.b	$C4,$9F,$7C,$65,$FC,$EC,$65,$F7,$47,$E5,$AF,$FC,$35,$83,$E0,$5E
	.dc.b	$36,$1B,$7F,$2A,$B8,$3F,$B2,$AE,$15,$B3,$31,$F8,$7E,$8F,$E1,$1F
	.dc.b	$35,$FE,$1B,$AF,$E3,$5D,$08,$C4,$EB,$D9,$FF,$13,$47,$F8,$7A,$D5
	.dc.b	$B9,$C9,$23,$F8,$04,$EF,$A9,$A7,$F1,$AF,$16,$D7,$F0,$56,$BE,$53
	.dc.b	$FA,$CC,$7F,$28,$9E,$3E,$1F,$F8,$AA,$24,$FE,$D1,$FE,$4F,$F4,$2E
	.dc.b	$1F,$F1,$6F,$D1,$79,$E8,$D9,$D4,$7F,$18,$0E,$1F,$F0,$14,$71,$38
	.dc.b	$4F,$F8,$E2,$75,$B7,$F3,$A9,$D7,$CA,$66,$A5,$FF,$A7,$2F,$A3,$FE
	.dc.b	$9A,$6B,$FD,$18,$F0,$C7,$F8,$3B,$C0,$7D,$0F,$F9,$1B,$E8,$9E,$95
	.dc.b	$3F,$C9,$33,$AB,$F5,$35,$FF,$0D,$4F,$07,$DC,$D7,$FC,$3D,$1B,$CF
	.dc.b	$FA,$D6,$AF,$9E,$FF,$05,$71,$FE,$0F,$2F,$C0,$9F,$E5,$18,$FF,$07
	.dc.b	$4E,$F0,$C9,$F8,$09,$D3,$0C,$10,$FF,$24,$4F,$23,$FC,$A6,$FE,$06
	.dc.b	$BF,$25,$F4,$5F,$90,$3C,$1E,$DC,$9F,$C6,$C6,$5F,$EA,$38,$7A,$67
	.dc.b	$51,$CF,$91,$5B,$8F,$0C,$EB,$2F,$D0,$D6,$A7,$A9,$D3,$1F,$2C,$FC
	.dc.b	$65,$F1,$FF,$A4,$F1,$A5,$6E,$3E,$31,$87,$C5,$18,$74,$7E,$99,$D5
	.dc.b	$75,$9F,$9D,$E8,$7F,$A2,$3C,$19,$7E,$38,$FA,$5E,$87,$FA,$33,$E8
	.dc.b	$A4,$F9,$E3,$FD,$1B,$F1,$0F,$A2,$FE,$8D,$2E,$0F,$F4,$D4,$E7,$3F
	.dc.b	$F0,$BC,$62,$B6,$D8,$CF,$6F,$FB,$CA,$DF,$0F,$87,$FD,$EA,$D7,$FD
	.dc.b	$B6,$77,$8E,$11,$BA,$FC,$77,$B9,$FE,$29,$BF,$F6,$1F,$85,$7C,$1F
	.dc.b	$1F,$17,$F9,$9D,$FE,$0D,$19,$CB,$D5,$2F,$73,$1C,$EF,$09,$F4,$5B
	.dc.b	$D3,$F3,$95,$A9,$CE,$5F,$77,$D9,$E3,$D0,$FA,$2E,$19,$A7,$31,$86
	.dc.b	$D7,$13,$F8,$FC,$65,$A6,$74,$FC,$94,$78,$FE,$89,$EB,$78,$3E,$6F
	.dc.b	$0D,$BF,$CD,$CE,$9F,$E5,$73,$A6,$9A,$9E,$17,$FE,$B3,$AE,$0F,$E8
	.dc.b	$6B,$83,$6B,$81,$C9,$E1,$E7,$FC,$2F,$85,$F3,$36,$7C,$6D,$9D,$C7
	.dc.b	$F3,$E9,$DB,$4D,$96,$3F,$D5,$9F,$43,$FE,$D7,$8F,$F2,$38,$FA,$2C
	.dc.b	$4E,$BC,$1F,$BD,$39,$7F,$01,$0A,$DF,$37,$F9,$C6,$B4,$7F,$3B,$8C
	.dc.b	$BF,$E7,$23,$33,$FE,$BD,$FA,$1F,$B9,$1F,$9B,$84,$61,$FC,$F3,$5F
	.dc.b	$E6,$C6,$B7,$FF,$3C,$BF,$CE,$63,$85,$FE,$6F,$E8,$7F,$DA,$57,$FD
	.dc.b	$36,$72,$FF,$9C,$63,$35,$E3,$BF,$CE,$5A,$72,$FF,$A1,$AF,$8D,$7F
	.dc.b	$8F,$E7,$37,$E1,$BE,$31,$89,$FF,$48,$D7,$0B,$E9,$E9,$7E,$A7,$CF
	.dc.b	$FE,$E7,$E6,$FE,$8E,$31,$E7,$A3,$19,$FF,$6D,$6E,$3F,$DD,$6F,$8C
	.dc.b	$FF,$79,$BF,$D4,$FC,$EC,$8F,$F0,$0F,$D0,$DB,$FE,$65,$37,$2F,$E5
	.dc.b	$F3,$6B,$FD,$4F,$87,$F8,$1B,$C9,$F8,$7F,$43,$F0,$23,$13,$FE,$1D
	.dc.b	$8C,$69,$FE,$66,$8C,$6F,$EA,$BE,$2F,$B3,$D6,$FD,$EF,$C7,$5A,$FF
	.dc.b	$0A,$B7,$5F,$E8,$CD,$D3,$FA,$1B,$F1,$39,$FF,$82,$FD,$0D,$BF,$F0
	.dc.b	$EB,$6F,$FA,$73,$6F,$FA,$5B,$F9,$8F,$A1,$3F,$C5,$BE,$7D,$F0,$19
	.dc.b	$8C,$37,$0F,$F5,$44,$E9,$FF,$0B,$57,$27,$33,$BC,$3F,$B4,$56,$DF
	.dc.b	$F1,$50,$AE,$0E,$A7,$EF,$F1,$78,$FE,$2E,$E1,$E0,$6D,$74,$6E,$5F
	.dc.b	$E4,$6C,$51,$EB,$E9,$F6,$7D,$0D,$C3,$D2,$DC,$3C,$AD,$C3,$93,$FB
	.dc.b	$CB,$7D,$0F,$C5,$F9,$6C,$37,$0F,$F1,$B4,$E9,$F8,$B3,$A6,$9A,$9E
	.dc.b	$18,$7C,$2A,$E0,$FC,$6A,$FA,$5E,$EF,$D3,$F9,$3B,$87,$D0,$6D,$7C
	.dc.b	$C6,$E5,$F0,$5F,$D5,$7D,$0F,$CF,$8F,$89,$F1,$D8,$E0,$FB,$CD,$C3
	.dc.b	$ED,$B7,$0F,$AE,$FE,$0F,$E5,$7F,$09,$EB,$DA,$FC,$0B,$6B,$EE,$B7
	.dc.b	$0F,$A4,$D9,$4F,$F9,$D3,$85,$39,$8C,$3F,$91,$F5,$BF,$25,$E9,$5B
	.dc.b	$FC,$7B,$6B,$F1,$6D,$C3,$F0,$0F,$E6,$7C,$EF,$CD,$7B,$D6,$FF,$2E
	.dc.b	$DB,$FC,$9B,$6F,$F1,$0F,$DF,$E0,$FD,$17,$C8,$DB,$FC,$FB,$6F,$F3
	.dc.b	$6D,$BF,$C8,$36,$7D,$FE,$9B,$54,$E6,$FE,$13,$F5,$BD,$B9,$7F,$AC
	.dc.b	$EF,$B5,$FC,$6F,$8D,$C3,$FB,$5F,$4B,$EE,$B6,$FF,$E0,$0C,$FC,$C6
	.dc.b	$F1,$66,$F9,$3F,$BD,$F9,$0F,$C2,$F0,$5A,$E2,$E1,$E7,$A5,$BF,$13
	.dc.b	$F8,$C2,$47,$03,$67,$8B,$2F,$C8,$9F,$E1,$B6,$99,$3F,$D5,$9F,$1F
	.dc.b	$8A,$FF,$36,$4A,$31,$C1,$FE,$C0,$8C,$57,$1A,$DD,$7F,$47,$FA,$1C
	.dc.b	$C7,$F4,$16,$DF,$1F,$91,$D5,$4B,$FF,$5E,$7A,$77,$0F,$0B,$6F,$C0
	.dc.b	$D9,$FB,$78,$FF,$86,$25,$1F,$EA,$5E,$C7,$4F,$F7,$0C,$3F,$3F,$18
	.dc.b	$9F,$EF,$4D,$8F,$F2,$B3,$5F,$EF,$2E,$6B,$E0,$7C,$17,$F1,$1B,$FF
	.dc.b	$22,$36,$FF,$DA,$8D,$BF,$92,$FF,$92,$FE,$23,$FC,$9D,$3F,$ED,$9F
	.dc.b	$53,$19,$7F,$93,$5B,$19,$3F,$D2,$B3,$C2,$FE,$9C,$E9,$B3,$AA,$E4
	.dc.b	$FF,$A7,$3E,$87,$E2,$3F,$E1,$BF,$A1,$FC,$FB,$9D,$7F,$C4,$F1,$8B
	.dc.b	$FF,$75,$7B,$6B,$E3,$1F,$32,$75,$AF,$FB,$02,$73,$82,$72,$FF,$A1
	.dc.b	$23,$31,$8A,$FF,$45,$46,$38,$BF,$C6,$BF,$32,$FA,$1F,$F3,$77,$CC
	.dc.b	$61,$B8,$FF,$BB,$E7,$4F,$FC,$01,$9D,$34,$FE,$37,$97,$E0,$B1,$77
	.dc.b	$3E,$35,$F0,$7F,$E8,$07,$FD,$39,$7C,$0F,$E0,$AE,$9C,$57,$28,$DB
	.dc.b	$FE,$CF,$74,$DD,$BA,$57,$57,$FD,$11,$F1,$7F,$E8,$EF,$A1,$E3,$97
	.dc.b	$EA,$63,$13,$AC,$7B,$E4,$BD,$C7,$E7,$AB,$6F,$93,$17,$CD,$7F,$A1
	.dc.b	$3C,$8F,$E7,$7F,$0C,$F9,$EF,$E7,$B5,$AE,$5F,$F0,$4B,$FE,$B2,$8F
	.dc.b	$E8,$3F,$3E,$FA,$17,$5F,$C9,$C3,$95,$FF,$46,$CB,$B3,$C4,$C6,$3D
	.dc.b	$27,$71,$DF,$2F,$1E,$5D,$B1,$EB,$9E,$2B,$8D,$57,$81,$D4,$B7,$E2
	.dc.b	$CF,$CD,$9F,$97,$1F,$6E,$7E,$4A,$F0,$B6,$33,$9F,$95,$F1,$57,$1C
	.dc.b	$3C,$27,$7A,$E0,$E2,$FC,$19,$FA,$23,$58,$73,$78,$9F,$E7,$EA,$F8
	.dc.b	$AD,$8F,$9E,$DC,$BF,$D1,$1F,$16,$F6,$1F,$43,$86,$D7,$FB,$9E,$75
	.dc.b	$39,$3F,$D2,$18,$3E,$73,$FD,$E9,$5F,$EE,$D8,$C7,$1E,$A7,$33,$E5
	.dc.b	$6B,$F2,$1E,$3A,$7C,$86,$ED,$9A,$FA,$DB,$BE,$85,$ED,$FE,$33,$0E
	.dc.b	$15,$F6,$72,$F0,$E5,$F2,$F2,$E6,$B1,$E5,$87,$1C,$BB,$4E,$F0,$E5
	.dc.b	$7C,$2F,$73,$89,$8E,$CF,$C9,$46,$9C,$63,$E4,$6D,$FE,$7B,$2F,$4C
	.dc.b	$7E,$47,$97,$F5,$DC,$FF,$A6,$E7,$FE,$11,$18,$9F,$D9,$F3,$FF,$08
	.dc.b	$6F,$C1,$6F,$9F,$FC,$17,$39,$95,$F2,$8C,$37,$0E,$13,$FC,$52,$7F
	.dc.b	$D5,$0F,$6C,$B3,$8F,$F0,$1F,$92,$B4,$6C,$CA,$F2,$C6,$8C,$3F,$E7
	.dc.b	$AF,$C9,$1E,$97,$FB,$5E,$11,$87,$FD,$B4,$D7,$FD,$AE,$6A,$38,$46
	.dc.b	$72,$FE,$59,$39,$3F,$89,$46,$73,$F8,$7F,$27,$EE,$E0,$7D,$B3,$E0
	.dc.b	$7D,$7F,$35,$31,$FE,$5B,$5A,$9E,$15,$B9,$F1,$CE,$5B,$19,$7F,$DD
	.dc.b	$B1,$96,$C6,$72,$FE,$85,$39,$6C,$67,$1F,$B3,$7E,$0F,$82,$F9,$FF
	.dc.b	$27,$31,$F2,$DA,$75,$1E,$27,$E3,$BF,$9D,$8D,$39,$21,$F8,$4B,$FE
	.dc.b	$5F,$1A,$8C,$1C,$67,$EF,$F6,$7C,$2F,$A9,$FC,$0F,$C8,$4A,$F1,$C6
	.dc.b	$5F,$86,$F8,$B4,$F9,$ED,$C0,$D3,$84,$F8,$27,$6F,$E7,$BC,$7F,$B1
	.dc.b	$D3,$FC,$63,$6F,$CD,$67,$F9,$23,$3E,$04,$BF,$AB,$5F,$C7,$9F,$E9
	.dc.b	$F8,$23,$11,$FD,$39,$AF,$F0,$37,$B6,$B9,$AE,$93,$F1,$39,$FF,$B1
	.dc.b	$7C,$7F,$93,$E8,$7F,$47,$F5,$52,$BF,$13,$8A,$75,$1E,$47,$E2,$46
	.dc.b	$63,$04,$3E,$13,$96,$C7,$EC,$3E,$1B,$F6,$3F,$19,$27,$F8,$1B,$F1
	.dc.b	$BD,$39,$1F,$09,$EE,$47,$F5,$79,$7F,$7B,$E1,$DC,$78,$27,$F7,$B1
	.dc.b	$FB,$EF,$0E,$4F,$F1,$07,$F8,$0F,$D5,$7D,$4F,$81,$F5,$92,$7E,$8B
	.dc.b	$F9,$2C,$3F,$C1,$1F,$15,$FC,$20,$93,$FE,$0C,$FA,$5F,$F0,$1B,$FE
	.dc.b	$3E,$5F,$0D,$FC,$43,$D1,$FF,$33,$CF,$F8,$49,$AE,$E3,$F8,$CF,$90
	.dc.b	$AE,$33,$FC,$22,$31,$3F,$A4,$6C,$66,$76,$F8,$A3,$2F,$53,$FC,$5E
	.dc.b	$E3,$A3,$F9,$9D,$FF,$AB,$A7,$D5,$CB,$FC,$9A,$FF,$E0,$0C,$7B,$7E
	.dc.b	$23,$13,$23,$04,$3F,$19,$3F,$90,$7F,$C8,$AB,$DA,$E9,$F1,$52,$7E
	.dc.b	$44,$77,$30,$F2,$8C,$3F,$37,$E2,$BC,$9C,$D2,$30,$43,$C2,$E5,$F1
	.dc.b	$C7,$AC,$C6,$7F,$D4,$B0,$FC,$A3,$63,$F9,$67,$E1,$EF,$AA,$F8,$5F
	.dc.b	$19,$30,$FE,$57,$39,$6C,$F0,$23,$F8,$3F,$27,$E5,$70,$FC,$BF,$89
	.dc.b	$93,$F8,$78,$CB,$F8,$98,$CB,$F2,$A3,$FA,$DB,$FE,$52,$F9,$1F,$C5
	.dc.b	$7C,$3C,$8F,$F2,$B7,$A8,$9D,$B6,$3F,$C3,$71,$FD,$56,$31,$1F,$E1
	.dc.b	$B8,$C1,$C0,$8F,$E5,$7E,$37,$F0,$5D,$BF,$83,$F8,$A9,$3F,$9F,$8C
	.dc.b	$BF,$A1,$8C,$BF,$82,$8F,$97,$F3,$3F,$8B,$C6,$4F,$CA,$8C,$B4,$87
	.dc.b	$E2,$AF,$D7,$D1,$F9,$3F,$8E,$93,$F0,$23,$2F,$F8,$E1,$F7,$11,$C1
	.dc.b	$B2,$1F,$96,$ED,$FF,$03,$FB,$2F,$C9,$1E,$37,$E8,$7E,$2E,$5F,$FA
	.dc.b	$5A,$B5,$3C,$2B,$73,$96,$BF,$D4,$27,$6F,$F0,$5F,$83,$FF,$0A,$FC
	.dc.b	$9B,$F5,$3F,$F1,$3C,$FE,$A3,$E4,$3F,$60,$6A,$33,$D9,$FE,$AF,$AE
	.dc.b	$1E,$3E,$49,$7F,$C4,$7E,$27,$AF,$0F,$F8,$4E,$8D,$24,$FF,$6C,$8C
	.dc.b	$BC,$08,$7F,$1B,$F8,$FF,$E0,$44,$70,$7F,$AA,$BF,$76,$FF,$0B,$E3
	.dc.b	$FA,$B9,$D4,$FF,$02,$7F,$C5,$F3,$AF,$41,$2B,$F2,$F8,$3E,$3C,$4D
	.dc.b	$24,$FF,$8A,$27,$6F,$FB,$21,$AF,$F7,$E6,$E3,$FE,$4E,$C1,$F1,$DF
	.dc.b	$AA,$FC,$21,$A7,$F8,$DD,$B8,$7F,$83,$9A,$FF,$8D,$9A,$FF,$3C,$73
	.dc.b	$7F,$E3,$66,$E7,$FE,$C2,$6C,$FF,$A4,$3C,$87,$C3,$78,$70,$93,$FE
	.dc.b	$CD,$AD,$F1,$AF,$73,$9A,$DB,$FE,$1A,$15,$C2,$77,$19,$FA,$27,$C5
	.dc.b	$7F,$94,$DB,$1F,$E2,$A6,$BF,$E5,$26,$BF,$D6,$1C,$E7,$FE,$52,$6E
	.dc.b	$7F,$ED,$E6,$CF,$DA,$FA,$3F,$F5,$0D,$7F,$90,$B0,$7B,$7F,$80,$F9
	.dc.b	$29,$7F,$E5,$29,$CB,$67,$FC,$59,$18,$8F,$F2,$7B,$5F,$F0,$FB,$5F
	.dc.b	$F7,$5B,$BA,$F0,$4F,$F9,$8E,$31,$8F,$E3,$BD,$27,$0F,$C6,$BF,$37
	.dc.b	$5F,$E5,$3E,$26,$65,$E5,$8E,$47,$90,$8F,$F9,$9B,$1F,$DE,$BE,$48
	.dc.b	$DB,$C4,$FD,$27,$C6,$61,$FF,$21,$4F,$F0,$4B,$F0,$C6,$8F,$89,$EA
	.dc.b	$7E,$9B,$EA,$64,$60,$87,$C4,$8F,$8E,$47,$C8,$FC,$12,$3F,$B5,$F9
	.dc.b	$2F,$F4,$36,$9F,$DE,$A3,$FC,$F6,$D7,$FC,$96,$D7,$F6,$BE,$EA,$31
	.dc.b	$1F,$EC,$D7,$FB,$D3,$FE,$88,$6B,$D3,$D5,$FF,$33,$46,$23,$FC,$47
	.dc.b	$F1,$AF,$83,$FF,$5C,$C3,$D9,$FF,$5E,$4E,$74,$FE,$0B,$18,$AD,$B6
	.dc.b	$33,$F1,$7F,$EA,$AF,$7E,$E1,$FE,$30,$6B,$FE,$E9,$3B,$46,$DA,$7A
	.dc.b	$FD,$AF,$F5,$7B,$F0,$32,$46,$23,$5C,$BE,$F8,$C6,$BF,$17,$C1,$B7
	.dc.b	$FC,$62,$BE,$24,$60,$CF,$AA,$8F,$F5,$9C,$FF,$B2,$1F,$E5,$5E,$B6
	.dc.b	$71,$FF,$44,$4E,$5B,$5C,$AB,$E8,$36,$3E,$DB,$C7,$4F,$E4,$1F,$53
	.dc.b	$6B,$D9,$36,$BE,$E7,$4A,$B1,$9F,$DF,$F1,$6E,$5F,$D2,$31,$FE,$7D
	.dc.b	$DF,$89,$91,$EA,$7A,$4F,$EB,$A7,$38,$FA,$A7,$E2,$92,$BF,$10,$D8
	.dc.b	$FF,$49,$62,$FA,$EE,$3A,$7F,$39,$7C,$7E,$5C,$C3,$FA,$87,$D4,$5C
	.dc.b	$3E,$3C,$FD,$22,$4F,$E5,$5B,$1D,$BE,$6F,$86,$1F,$D1,$5F,$89,$0F
	.dc.b	$9D,$79,$D8,$6E,$1F,$D3,$FE,$B9,$F2,$59,$A9,$3F,$ED,$6A,$DD,$FF
	.dc.b	$BF,$9A,$FF,$A1,$1A,$FF,$62,$1F,$95,$7D,$39,$FE,$CB,$97,$E0,$E7
	.dc.b	$3C,$7F,$67,$CB,$0F,$92,$F4,$E1,$C6,$BF,$A3,$36,$F9,$37,$0F,$EE
	.dc.b	$37,$CA,$F8,$E1,$27,$C2,$79,$6B,$86,$1E,$0B,$F0,$4E,$A3,$FB,$DC
	.dc.b	$E9,$F3,$C7,$FB,$E2,$3C,$91,$FE,$A4,$6B,$D6,$BF,$5D,$E7,$FF,$85
	.dc.b	$BE,$07,$F2,$CD,$BF,$F0,$77,$D4,$63,$95,$54,$E2,$E2,$7F,$C0,$CD
	.dc.b	$C3,$FD,$67,$7C,$DF,$F5,$AF,$3F,$F4,$4F,$A8,$FC,$CF,$39,$BF,$51
	.dc.b	$86,$63,$FE,$1F,$72,$D3,$FC,$25,$8B,$DF,$F3,$1C,$31,$FF,$0A,$46
	.dc.b	$67,$73,$E8,$23,$E8,$7D,$71,$EF,$3C,$BE,$ED,$9E,$F1,$E8,$F8,$0C
	.dc.b	$5F,$EA,$9F,$C3,$76,$BD,$EA,$7B,$BD,$EB,$87,$F9,$6E,$74,$FF,$8E
	.dc.b	$FE,$77,$38,$7E,$D1,$FD,$2F,$94,$F8,$4A,$7F,$00,$F3,$7F,$9D,$5F
	.dc.b	$E0,$EF,$FD,$7D,$3A,$8F,$F7,$1C,$7F,$90,$2B,$94,$62,$36,$D8,$CC
	.dc.b	$70,$6F,$57,$F2,$5E,$36,$0F,$23,$B5,$AB,$53,$9C,$3F,$2D,$9B,$F6
	.dc.b	$EF,$CD,$B7,$A7,$D5,$DF,$FC,$8A,$72,$7E,$83,$3E,$FD,$B9,$AF,$80
	.dc.b	$FC,$1F,$97,$F8,$5F,$2D,$72,$FD,$25,$F9,$A3,$1A,$F9,$67,$FC,$29
	.dc.b	$A2,$73,$AF,$A1,$F9,$FF,$0B,$2B,$9F,$B8,$F5,$F1,$8F,$F9,$0F,$89
	.dc.b	$7F,$D5,$E3,$FD,$DC,$D8,$CD,$FC,$D8,$C1,$FC,$E2,$DE,$B7,$4C,$BE
	.dc.b	$74,$62,$BD,$79,$BE,$5B,$F0,$1F,$2D,$F8,$16,$CE,$9B,$9F,$BC,$FF
	.dc.b	$05,$57,$19,$E7,$5F,$3F,$AA,$47,$6F,$52,$DF,$1E,$6F,$B1,$5F,$8F
	.dc.b	$F9,$6C,$E9,$F7,$9F,$C9,$7B,$7F,$93,$F9,$5B,$97,$A1,$B8,$F8,$F9
	.dc.b	$3F,$13,$18,$C3,$F0,$ED,$AF,$AA,$DA,$F4,$BF,$9F,$F9,$5F,$D0,$7C
	.dc.b	$AD,$AF,$8A,$DA,$ED,$BA,$31,$87,$E5,$1F,$D4,$7B,$DF,$A9,$F9,$5B
	.dc.b	$87,$CE,$6E,$1C,$9B,$7F,$9C,$7F,$BD,$7B,$6F,$07,$FD,$37,$7F,$CC
	.dc.b	$E3,$1C,$BF,$77,$DE,$DD,$6B,$1F,$A9,$5A,$BF,$DA,$B6,$3F,$AA,$1F
	.dc.b	$E3,$59,$E0,$FF,$57,$8D,$36,$3A,$63,$F8,$8F,$7B,$BE,$3F,$8D,$7F
	.dc.b	$83,$7C,$AC,$C7,$F1,$7F,$2B,$71,$FB,$DF,$9A,$B8,$7F,$C5,$AD,$AF
	.dc.b	$E1,$3C,$E8,$C5,$79,$A3,$15,$FE,$37,$E4,$FE,$0B,$F1,$5F,$C5,$4F
	.dc.b	$E2,$DF,$2D,$71,$FC,$DF,$2F,$F8,$83,$26,$7F,$C6,$E3,$0C,$9F,$0C
	.dc.b	$E9,$B1,$E6,$3C,$1F,$53,$FE,$DC,$8F,$E0,$2D,$D3,$DB,$3F,$E2,$68
	.dc.b	$C7,$07,$F8,$FF,$5F,$E3,$5F,$0B,$FE,$32,$F2,$7F,$D8,$58,$F2,$8C
	.dc.b	$11,$F8,$71,$AC,$7E,$13,$5F,$F0,$65,$74,$9D,$D7,$87,$E5,$F9,$57
	.dc.b	$E8,$F1,$47,$87,$E5,$E5,$FA,$BC,$4B,$7F,$B3,$C9,$5F,$E6,$26,$D7
	.dc.b	$92,$30,$47,$E1,$FC,$2F,$2D,$3F,$9D,$65,$FE,$6E,$9F,$E6,$B7,$FC
	.dc.b	$5B,$0E,$DE,$2F,$C7,$F8,$5F,$88,$CA,$F8,$2F,$84,$8E,$0F,$A3,$96
	.dc.b	$EB,$DC,$FC,$DF,$92,$93,$F5,$BD,$9B,$5E,$D9,$F6,$12,$7E,$7B,$63
	.dc.b	$E4,$7C,$2F,$0D,$3D,$6F,$D9,$F8,$F9,$7F,$82,$F9,$5B,$7F,$26,$7E
	.dc.b	$41,$27,$ED,$B6,$3E,$A7,$A7,$C3,$0F,$8E,$FF,$95,$BE,$67,$E4,$4F
	.dc.b	$F2,$D0,$E2,$7E,$6E,$BB,$E1,$98,$FF,$6F,$AE,$37,$FB,$E6,$BF,$E5
	.dc.b	$36,$BF,$EC,$1E,$1F,$F0,$C3,$F8,$7F,$95,$FC,$D7,$B5,$F9,$B6,$DF
	.dc.b	$E8,$1B,$3B,$8F,$D1,$FB,$3F,$A1,$3F,$45,$EC,$D8,$C9,$27,$F2,$91
	.dc.b	$97,$DD,$F2,$CF,$92,$75,$25,$F1,$BF,$E6,$37,$15,$B9,$D4,$E7,$B1
	.dc.b	$F3,$9F,$D7,$6F,$8F,$B3,$FE,$73,$C3,$84,$60,$E7,$8F,$81,$C0,$8D
	.dc.b	$E1,$D6,$FA,$3E,$39,$D5,$71,$8C,$57,$0B,$F0,$4E,$5F,$F7,$8D,$F2
	.dc.b	$73,$8F,$A9,$B8,$78,$AF,$92,$C7,$DB,$EB,$7F,$B2,$F2,$EF,$97,$5A
	.dc.b	$F5,$E3,$CD,$A7,$BE,$F8,$E9,$98,$F4,$CD,$C7,$E2,$C7,$F4,$66,$BE
	.dc.b	$43,$17,$F1,$DF,$F8,$03,$E9,$7F,$A6,$1F,$F2,$1F,$C7,$A3,$11,$AF
	.dc.b	$86,$FF,$22,$4E,$5F,$F2,$35,$FF,$88,$7C,$EB,$EC,$9A,$4B,$FF,$27
	.dc.b	$CE,$DF,$8D,$3B,$6C,$FC,$1D,$1F,$38,$F2,$BF,$D6,$9B,$FD,$6C,$38
	.dc.b	$36,$FE,$14,$63,$87,$F2,$58,$72,$BE,$31,$F8,$B9,$D6,$3F,$94,$6B
	.dc.b	$F8,$97,$35,$C6,$3A,$DC,$7C,$E7,$D5,$C7,$CC,$E7,$1F,$23,$A7,$F3
	.dc.b	$D1,$A8,$FC,$D7,$96,$F9,$CF,$F3,$44,$7F,$95,$9A,$FF,$89,$5B,$DB
	.dc.b	$FE,$6F,$7F,$DB,$F1,$BF,$1B,$FD,$B1,$39,$6F,$27,$EC,$EF,$B4,$62
	.dc.b	$91,$E1,$F8,$0C,$3E,$3D,$AE,$2F,$7A,$98,$89,$7E,$78,$CB,$E9,$8E
	.dc.b	$CF,$F2,$2F,$13,$AF,$73,$C5,$A6,$BC,$04,$D3,$FD,$25,$5C,$4F,$E4
	.dc.b	$CE,$23,$71,$FE,$12,$F9,$C7,$BB,$E8,$F7,$A4,$FF,$A3,$A7,$2D,$9C
	.dc.b	$BF,$36,$7E,$27,$2A,$B7,$3F,$E5,$F6,$F1,$7F,$81,$5B,$C5,$F8,$E6
	.dc.b	$CF,$CC,$DE,$DF,$3F,$4B,$E5,$FB,$D2,$BC,$91,$E0,$BF,$F1,$74,$7E
	.dc.b	$06,$74,$F8,$67,$44,$78,$9F,$87,$ED,$E5,$E8,$7D,$8F,$76,$4F,$C1
	.dc.b	$8C,$BF,$8D,$8C,$BE,$D8,$CB,$4E,$07,$E4,$FD,$DF,$8B,$D5,$FC,$27
	.dc.b	$6F,$C9,$63,$F8,$63,$3E,$14,$BF,$CD,$46,$5F,$F5,$FD,$71,$DF,$F6
	.dc.b	$AD,$8F,$E4,$CD,$DB,$FA,$4D,$FE,$4B,$DD,$F0,$FA,$1F,$94,$F7,$66
	.dc.b	$1D,$23,$2F,$EA,$63,$2F,$FA,$5F,$0F,$F1,$2B,$B2,$19,$9F,$CF,$FB
	.dc.b	$BF,$A0,$F6,$64,$FE,$CD,$F8,$3E,$7C,$8C,$10,$EC,$FE,$A3,$83,$51
	.dc.b	$83,$F7,$8F,$EB,$3C,$49,$1F,$BF,$7F,$5F,$EC,$FC,$EE,$CB,$3F,$C0
	.dc.b	$E3,$0F,$F0,$5F,$67,$F6,$A4,$7F,$DB,$BF,$39,$FE,$86,$E4,$71,$FF
	.dc.b	$73,$F1,$92,$F5,$7F,$7F,$F0,$7F,$ED,$6C,$3F,$DC,$AD,$7F,$D2,$73
	.dc.b	$B7,$FD,$34,$D7,$FD,$71,$1B,$C3,$BC,$7F,$0C,$8C,$61,$E1,$AE,$15
	.dc.b	$AC,$3C,$78,$7F,$A7,$30,$D3,$5F,$E9,$55,$DD,$FF,$4F,$35,$FE,$38
	.dc.b	$FE,$27,$C5,$F6,$FB,$5F,$C6,$A3,$FC,$D3,$3F,$13,$9F,$06,$66,$46
	.dc.b	$4F,$93,$ED,$71,$F5,$A4,$60,$87,$FB,$5E,$D1,$A8,$C4,$F8,$7D,$AF
	.dc.b	$17,$3C,$9F,$AB,$19,$7E,$BC,$65,$E9,$19,$7B,$46,$75,$F9,$B9,$FD
	.dc.b	$AA,$FE,$2F,$F3,$8F,$94,$F8,$1E,$E4,$C7,$FC,$65,$87,$FA,$6E,$B5
	.dc.b	$18,$E1,$FE,$CB,$18,$8F,$F5,$C3,$5F,$F4,$BB,$5D,$DF,$D5,$C7,$EC
	.dc.b	$C7,$E4,$3D,$BF,$A0,$7D,$1E,$19,$19,$21,$F2,$5F,$E8,$BE,$93,$E0
	.dc.b	$BF,$27,$ED,$CA,$F3,$1F,$E4,$88,$D9,$97,$F2,$31,$87,$67,$F9,$06
	.dc.b	$3F,$A4,$CF,$58,$FF,$0B,$C6,$0F,$57,$17,$E5,$A3,$FB,$1A,$F7,$3F
	.dc.b	$14,$7E,$2F,$B6,$4F,$EB,$A3,$2F,$EC,$23,$E9,$FC,$37,$E4,$71,$93
	.dc.b	$EA,$F1,$1B,$AB,$73,$FE,$EA,$8C,$36,$3F,$45,$3C,$1F,$5F,$57,$E7
	.dc.b	$7D,$C9,$5F,$03,$54,$FC,$C3,$31,$F9,$79,$C3,$B8,$E6,$FF,$8E,$9A
	.dc.b	$FF,$9E,$4F,$F3,$D7,$B0,$F9,$B7,$8D,$FA,$DF,$9C,$95,$F5,$AB,$53
	.dc.b	$FE,$0D,$6C,$7E,$F5,$AF,$50,$E1,$F0,$35,$FF,$45,$7C,$63,$E7,$3F
	.dc.b	$C5,$1D,$3F,$A2,$E3,$7C,$8F,$EF,$CF,$F0,$E5,$6F,$C6,$B7,$FE,$2F
	.dc.b	$C3,$D9,$39,$7F,$88,$35,$F2,$35,$DF,$C1,$7F,$2B,$EC,$FF,$15,$7B
	.dc.b	$0D,$BF,$C3,$67,$82,$75,$4B,$47,$F8,$69,$B1,$FC,$75,$AF,$8A,$3F
	.dc.b	$C3,$FD,$9F,$CD,$BE,$0F,$F8,$D1,$27,$F4,$70,$F8,$EF,$E5,$F1,$FC
	.dc.b	$4E,$74,$D3,$71,$AD,$FF,$9A,$E1,$E4,$F0,$7F,$C6,$7F,$29,$2F,$F5
	.dc.b	$F3,$96,$E4,$AF,$43,$63,$D1,$18,$3F,$C6,$FF,$36,$F8,$9F,$C1,$78
	.dc.b	$3F,$9C,$A6,$5F,$D1,$B9,$7F,$A9,$75,$FF,$91,$B0,$E1,$5F,$BA,$6B
	.dc.b	$FD,$15,$AF,$E2,$5A,$FE,$5D,$AF,$FB,$45,$E3,$5F,$19,$BC,$5B,$C7
	.dc.b	$F2,$DE,$4F,$FB,$66,$FC,$11,$E3,$23,$F9,$D8,$D3,$FE,$FD,$F9,$B7
	.dc.b	$CC,$FF,$97,$BE,$6F,$FC,$E7,$3F,$F0,$05,$B9,$7D,$6A,$DC,$FF,$16
	.dc.b	$6B,$FD,$85,$AF,$E9,$5A,$FE,$CD,$AF,$FB,$2E,$DE,$C7,$F9,$E3,$8D
	.dc.b	$F3,$74,$FF,$2D,$CE,$AB,$F8,$93,$5F,$A7,$A7,$FA,$16,$B8,$C7,$E2
	.dc.b	$2B,$8D,$C8,$FE,$ED,$E1,$7F,$BF,$A3,$F8,$AE,$1C,$E7,$F8,$44,$E5
	.dc.b	$DC,$7F,$08,$7F,$27,$F0,$FE,$73,$CB,$DF,$23,$24,$3C,$A7,$BD,$D3
	.dc.b	$8A,$57,$58,$C3,$74,$FE,$CF,$93,$E2,$B8,$C7,$F8,$5A,$F8,$E1,$8D
	.dc.b	$3E,$15,$7C,$8F,$4B,$E6,$7C,$7C,$CF,$E4,$C6,$5F,$D6,$C6,$5F,$0C
	.dc.b	$65,$A7,$03,$F6,$7F,$1D,$FE,$8E,$ED,$FF,$0A,$C7,$B9,$77,$FF,$1C
	.dc.b	$8F,$DF,$B5,$F8,$11,$C3,$4F,$52,$F4,$7F,$D5,$BC,$D2,$BF,$A8,$4E
	.dc.b	$5F,$EC,$BF,$16,$F8,$0F,$D9,$C7,$FB,$DF,$E2,$BF,$A0,$4E,$71,$FF
	.dc.b	$24,$46,$2B,$7E,$BF,$FB,$52,$B5,$7C,$4F,$E7,$9E,$EF,$06,$99,$3F
	.dc.b	$DF,$25,$01,$E1,$C7,$FE,$D4,$F5,$FF,$C9,$79,$7F,$88,$36,$78,$B8
	.dc.b	$E4,$FF,$7E,$ED,$FE,$32,$7E,$3C,$67,$26,$3F,$ED,$EF,$0B,$16,$FA
	.dc.b	$9F,$F3,$2E,$1F,$1A,$73,$18,$BF,$9F,$F1,$1C,$F0,$FF,$73,$BC,$BA
	.dc.b	$7F,$C6,$18,$7C,$E8,$D6,$3F,$ED,$A8,$FE,$63,$F2,$58,$9E,$0D,$8C
	.dc.b	$9C,$4B,$87,$DE,$F1,$B9,$9D,$FC,$2B,$3F,$72,$2C,$FE,$27,$1F,$E2
	.dc.b	$73,$92,$46,$7E,$73,$7C,$1F,$A4,$6C,$FF,$01,$6C,$B3,$FE,$9D,$1E
	.dc.b	$CF,$27,$FA,$7C,$FD,$06,$BF,$3A,$51,$8D,$79,$36,$33,$7F,$D5,$3E
	.dc.b	$6F,$A3,$4F,$CB,$CF,$CD,$8C,$DF,$B5,$FB,$8B,$E6,$BF,$C0,$35,$FE
	.dc.b	$87,$6C,$7F,$A2,$DA,$FF,$83,$DA,$FF,$87,$5D,$E5,$F8,$EF,$A7,$FF
	.dc.b	$29,$57,$F9,$2A,$72,$FF,$9A,$A3,$35,$FC,$50,$7C,$DB,$83,$CA,$DA
	.dc.b	$FF,$12,$79,$6F,$8F,$CD,$49,$FF,$2B,$D6,$EF,$F8,$FD,$FF,$89,$2F
	.dc.b	$51,$DE,$74,$FF,$8A,$5A,$FF,$8D,$DA,$FE,$B1,$AF,$9A,$FF,$49,$5F
	.dc.b	$C0,$7B,$7A,$E7,$FD,$CF,$B5,$1F,$A9,$F3,$BF,$C3,$3E,$FF,$DE,$6E
	.dc.b	$9F,$E1,$78,$C7,$29,$A7,$EA,$BE,$7B,$94,$66,$7F,$C3,$3D,$29,$FD
	.dc.b	$57,$CD,$49,$FF,$0B,$FC,$DD,$DF,$F3,$93,$75,$FE,$77,$36,$BF,$C8
	.dc.b	$B8,$FF,$9D,$DF,$F1,$57,$BF,$E4,$23,$83,$C3,$E7,$79,$23,$F7,$EF
	.dc.b	$9A,$7D,$64,$9F,$F1,$A3,$71,$FD,$77,$C0,$6E,$7F,$C5,$AF,$CC,$EA
	.dc.b	$7B,$9F,$A4,$E3,$FE,$22,$D8,$CD,$7F,$98,$DB,$7F,$E6,$3A,$DC,$FF
	.dc.b	$8E,$3E,$67,$A5,$FF,$0F,$7F,$43,$E4,$BE,$0B,$F6,$38,$7F,$41,$6C
	.dc.b	$66,$FF,$DD,$4D,$9F,$F7,$73,$77,$FF,$65,$F9,$9F,$80,$AF,$F5,$4C
	.dc.b	$E5,$FE,$6B,$DE,$9D,$5F,$E8,$6B,$4F,$F0,$78,$CD,$FF,$95,$27,$35
	.dc.b	$C7,$9F,$F9,$57,$CD,$FE,$5B,$63,$49,$D5,$FC,$25,$FD,$26,$1E,$6E
	.dc.b	$CE,$33,$FC,$EB,$2F,$D6,$69,$FE,$64,$5F,$2B,$9F,$9B,$D3,$67,$4D
	.dc.b	$DB,$FB,$85,$70,$9E,$55,$F3,$BE,$17,$E5,$92,$3E,$5B,$6B,$FB,$C7
	.dc.b	$9F,$FE,$0D,$9F,$93,$3F,$D3,$A3,$13,$F1,$E3,$1E,$97,$5A,$FF,$3D
	.dc.b	$D6,$B4,$FA,$7F,$29,$ED,$F9,$5F,$73,$6B,$DC,$DC,$7D,$2D,$AF,$F0
	.dc.b	$E6,$FF,$C5,$A7,$FC,$66,$D7,$FB,$7F,$07,$D0,$E8,$F9,$C4,$C3,$E5
	.dc.b	$7C,$9F,$D6,$8F,$9C,$FC,$FF,$67,$E8,$7C,$9D,$AF,$A0,$DC,$3E,$5B
	.dc.b	$72,$FC,$6B,$67,$F1,$7C,$28,$C5,$7E,$25,$FB,$9F,$27,$F7,$7E,$4E
	.dc.b	$DF,$DD,$6D,$7D,$86,$E1,$F9,$A7,$F0,$9E,$D7,$E1,$7E,$4E,$DF,$E1
	.dc.b	$5B,$7F,$80,$6D,$FE,$89,$FF,$04,$7B,$2E,$EF,$F4,$D5,$FF,$B5,$ED
	.dc.b	$39,$E4,$FE,$C2,$D9,$D3,$67,$FD,$BC,$DC,$13,$FA,$CA,$7B,$5E,$0C
	.dc.b	$3B,$E1,$F8,$C8,$C4,$FE,$2D,$B5,$F8,$87,$F4,$5F,$27,$23,$F4,$5F
	.dc.b	$27,$6F,$FC,$CD,$F3,$6D,$FF,$51,$F2,$7F,$E1,$B8,$C5,$FE,$A6,$3F
	.dc.b	$59,$F1,$32,$30,$7E,$B5,$FD,$87,$1F,$1C,$6C,$FF,$8E,$B6,$FF,$36
	.dc.b	$60,$9D,$4F,$F9,$6B,$2A,$7F,$97,$FD,$CF,$F2,$0C,$64,$91,$E8,$9F
	.dc.b	$F1,$BC,$EA,$3F,$D0,$1E,$17,$FA,$02,$BC,$07,$E7,$1C,$47,$5F,$81
	.dc.b	$FF,$1C,$46,$31,$FF,$40,$36,$BF,$DF,$7F,$27,$AA,$EE,$D8,$FF,$38
	.dc.b	$B6,$FD,$0F,$4F,$93,$93,$FE,$76,$F1,$ED,$79,$A7,$D3,$F0,$18,$CB
	.dc.b	$FC,$B4,$DB,$FE,$5B,$9F,$F9,$66,$F8,$47,$FA,$02,$7D,$3E,$1C,$8F
	.dc.b	$F4,$1F,$BF,$6B,$E1,$CF,$C6,$C7,$18,$FF,$99,$DB,$7D,$9B,$5F,$E3
	.dc.b	$DC,$51,$F1,$BE,$52,$57,$FA,$43,$D5,$B7,$F4,$27,$E9,$FC,$06,$31
	.dc.b	$FF,$3A,$36,$FD,$6D,$AE,$CF,$D2,$F8,$19,$3F,$E9,$CF,$8A,$B5,$F7
	.dc.b	$27,$EF,$F9,$F8,$C3,$FD,$00,$DB,$F8,$ED,$AF,$53,$FE,$44,$F9,$7F
	.dc.b	$95,$1F,$E8,$C3,$FD,$1A,$E1,$C3,$5C,$F0,$7F,$6A,$8F,$A5,$7F,$B6
	.dc.b	$9F,$F3,$17,$B2,$F9,$AE,$D7,$FE,$86,$7F,$2B,$EB,$FE,$5B,$E5,$6E
	.dc.b	$7F,$EB,$86,$E1,$F9,$F6,$E3,$FE,$9B,$7F,$3B,$EB,$CA,$FF,$5F,$FA
	.dc.b	$D6,$FF,$4D,$36,$FF,$D4,$53,$6B,$DD,$F2,$9E,$5E,$AF,$F5,$5F,$C3
	.dc.b	$FF,$93,$F1,$FF,$56,$CE,$5F,$F5,$77,$AF,$FD,$0B,$E6,$FF,$AA,$D6
	.dc.b	$E5,$5B,$8C,$36,$BB,$ED,$FE,$F0,$AE,$67,$FA,$BA,$F9,$CE,$B0,$F1
	.dc.b	$B6,$76,$76,$AE,$58,$79,$FD,$16,$1C,$AB,$D3,$5A,$BE,$57,$E4,$8C
	.dc.b	$4F,$87,$D5,$FB,$13,$9C,$7C,$33,$C9,$AF,$85,$CE,$3E,$0A,$E3,$87
	.dc.b	$63,$D1,$5F,$8B,$38,$1F,$D0,$F2,$F9,$8E,$5A,$FF,$BA,$BD,$17,$B5
	.dc.b	$FD,$26,$1F,$12,$9C,$C7,$F4,$D8,$C4,$ED,$B1,$DF,$E2,$FF,$A8,$B7
	.dc.b	$CF,$E9,$82,$73,$3E,$36,$BF,$86,$78,$EB,$F7,$1B,$96,$6F,$E1,$FC
	.dc.b	$DB,$CA,$FD,$55,$EE,$7E,$4E,$3D,$DB,$87,$2C,$7F,$DD,$F8,$71,$73
	.dc.b	$31,$FC,$0D,$71,$7F,$13,$B2,$B8,$46,$AF,$A3,$F9,$46,$D7,$1C,$3F
	.dc.b	$11,$5C,$1F,$35,$71,$C3,$C1,$1A,$C7,$D7,$97,$BE,$B4,$49,$D3,$C2
	.dc.b	$35,$1D,$0C,$C6,$E3,$24,$7E,$0C,$F6,$7F,$63,$8F,$A0,$97,$F3,$A7
	.dc.b	$53,$ED,$8F,$03,$C1,$D3,$83,$47,$84,$EE,$67,$1F,$89,$7E,$3C,$BE
	.dc.b	$65,$F6,$9C,$D7,$53,$94,$E9,$E5,$87,$99,$FA,$B9,$7C,$DC,$3D,$D1
	.dc.b	$E2,$CB,$E2,$47,$F9,$32,$75,$3D,$67,$2D,$9D,$63,$E3,$BE,$D5,$AC
	.dc.b	$3C,$57,$ED,$3B,$5F,$AF,$0E,$B1,$83,$A1,$27,$8D,$78,$23,$46,$E3
	.dc.b	$35,$EA,$C7,$E1,$CE,$6F,$AC,$60,$98,$7E,$76,$F8,$D7,$80,$EB,$3F
	.dc.b	$32,$33,$18,$9F,$6D,$FC,$08,$F9,$B9,$7D,$EA,$DD,$F6,$8C,$61,$F2
	.dc.b	$A7,$38,$72,$BF,$BE,$7D,$4C,$BF,$1D,$3E,$42,$7E,$F2,$F1,$B3,$87
	.dc.b	$32,$57,$09,$D5,$FF,$22,$6B,$D4,$FC,$89,$FB,$67,$3A,$7F,$37,$6E
	.dc.b	$9F,$E8,$C6,$CB,$14,$63,$DE,$76,$7F,$17,$AF,$06,$7E,$32,$3F,$CE
	.dc.b	$27,$7A,$7E,$67,$E7,$FF,$95,$4F,$F5,$66,$EE,$69,$23,$FC,$8A,$7F
	.dc.b	$62,$7F,$C0,$9D,$3F,$DF,$63,$91,$AB,$7D,$9B,$DC,$DF,$57,$F6,$EC
	.dc.b	$2D,$F3,$3A,$9C,$86,$1F,$E0,$9C,$39,$DF,$07,$E0,$F3,$DD,$5F,$1F
	.dc.b	$4C,$44,$C3,$C9,$3A,$6C,$E5,$EF,$97,$FB,$DA,$7F,$D1,$71,$C3,$E6
	.dc.b	$FB,$1D,$FD,$59,$19,$21,$C8,$FF,$0F,$BE,$7F,$8D,$F0,$46,$72,$47
	.dc.b	$68,$C3,$74,$F8,$0F,$06,$E5,$CE,$E5,$7C,$FA,$D6,$48,$61,$DA,$F7
	.dc.b	$1D,$AF,$C9,$1F,$4A,$1F,$8F,$7F,$C0,$07,$E9,$B5,$F6,$1A,$F8,$3F
	.dc.b	$A5,$F9,$6F,$E7,$B3,$FC,$15,$6D,$B3,$D6,$31,$97,$97,$25,$F1,$8F
	.dc.b	$A1,$F0,$BF,$6F,$E4,$A6,$5F,$62,$3E,$F1,$87,$E5,$6E,$9C,$C7,$E2
	.dc.b	$B9,$1F,$13,$F9,$C2,$47,$F7,$E6,$BE,$C8,$DB,$C6,$3F,$1D,$CA,$F8
	.dc.b	$6F,$8D,$5A,$AE,$15,$83,$73,$F9,$3F,$91,$FC,$A7,$3C,$AF,$C8,$C6
	.dc.b	$5F,$C9,$C6,$5F,$5C,$7D,$3F,$C9,$5E,$85,$24,$FE,$6B,$E4,$AB,$F5
	.dc.b	$27,$6F,$EA,$9A,$FC,$68,$FB,$DF,$71,$FC,$02,$0D,$23,$FC,$11,$FF
	.dc.b	$1B,$72,$FE,$F5,$BC,$3F,$E8,$36,$CF,$F9,$F3,$E8,$FA,$B0,$FD,$DC
	.dc.b	$6B,$93,$F8,$BC,$FF,$00,$F0,$1F,$A9,$3E,$27,$0D,$C3,$F8,$7C,$EA
	.dc.b	$7F,$7D,$C3,$A6,$4E,$AB,$83,$F8,$49,$E7,$F1,$32,$34,$FF,$07,$F8
	.dc.b	$FC,$47,$ED,$F0,$F1,$CE,$35,$FE,$21,$CE,$BF,$2F,$C4,$CC,$3C,$0F
	.dc.b	$F1,$1F,$9A,$C4,$73,$CD,$38,$C3,$9D,$6E,$BF,$89,$B5,$FC,$DF,$C6
	.dc.b	$D9,$DC,$68,$FE,$33,$1F,$1B,$E1,$24,$7C,$17,$F9,$17,$D0,$62,$3C
	.dc.b	$51,$6B,$D7,$3F,$3B,$CC,$95,$F2,$5F,$E5,$5F,$19,$88,$F5,$45,$AF
	.dc.b	$87,$37,$97,$F8,$77,$07,$DA,$AC,$5F,$FA,$FA,$AC,$E6,$3B,$1F,$CC
	.dc.b	$A3,$EF,$FC,$44,$8F,$B4,$FF,$3D,$FE,$01,$FF,$38,$E3,$FE,$C9,$BF
	.dc.b	$C3,$CE,$EB,$E7,$3B,$F4,$7F,$8D,$7C,$D7,$F0,$39,$FF,$43,$35,$FF
	.dc.b	$48,$07,$0F,$FA,$4D,$AF,$FB,$7D,$E1,$7F,$D7,$5B,$9E,$7C,$2F,$E5
	.dc.b	$1F,$37,$E6,$8C,$5F,$F5,$AA,$DB,$FC,$CD,$AA,$78,$3F,$CD,$FD,$87
	.dc.b	$07,$F7,$46,$C7,$FA,$7D,$AF,$A9,$AF,$FA,$AD,$CE,$3F,$DC,$DB,$87
	.dc.b	$FB,$E5,$B3,$FD,$17,$CC,$7C,$77,$F4,$19,$CC,$FF,$BD,$70,$C1,$FD
	.dc.b	$16,$B7,$1F,$DD,$A7,$64,$9D,$57,$F4,$BF,$CF,$FF,$D6,$95,$FD,$F3
	.dc.b	$E6,$3F,$AE,$63,$FD,$56,$11,$87,$FA,$C4,$F0,$7F,$A7,$B7,$2F,$F0
	.dc.b	$73,$6B,$93,$67,$FB,$05,$3D,$0F,$17,$B5,$27,$FC,$1B,$5B,$7C,$33
	.dc.b	$FD,$87,$F3,$F7,$83,$49,$3F,$E1,$4F,$B5,$7C,$4E,$FA,$A4,$FF,$86
	.dc.b	$2D,$1F,$DA,$EB,$FA,$E4,$FF,$73,$F7,$3D,$FB,$7F,$75,$E6,$7C,$7F
	.dc.b	$C3,$9C,$D2,$72,$F1,$DF,$FC,$46,$E6,$B8,$46,$27,$73,$FE,$CE,$34
	.dc.b	$7F,$BB,$FD,$CF,$05,$7E,$89,$B1,$93,$63,$E6,$9F,$10,$E3,$FF,$1A
	.dc.b	$C6,$1F,$F2,$1F,$C0,$BF,$18,$FC,$CC,$6A,$B8,$5F,$F8,$22,$DF,$09
	.dc.b	$8D,$BF,$C9,$11,$8B,$E7,$CE,$F4,$75,$1F,$87,$9D,$61,$FE,$12,$F8
	.dc.b	$47,$81,$A8,$FF,$10,$46,$2B,$CB,$F9,$A3,$E7,$35,$4A,$D3,$CE,$F3
	.dc.b	$18,$CF,$D3,$BB,$0E,$58,$FF,$95,$AB,$4F,$E6,$DB,$87,$F9,$DF,$F3
	.dc.b	$F7,$CB,$74,$CB,$CD,$E2,$A7,$38,$7F,$98,$3F,$2C,$7C,$C7,$99,$B1
	.dc.b	$FE,$70,$FA,$87,$E6,$2F,$84,$D5,$CA,$E4,$FC,$3A,$DC,$E7,$4F,$F1
	.dc.b	$CE,$1D,$1B,$5B,$C3,$A6,$1C,$89,$3F,$CA,$2B,$83,$FB,$90,$E1,$FD
	.dc.b	$E3,$5F,$F2,$97,$D1,$BE,$A3,$E8,$B6,$BF,$D1,$7C,$8F,$AE,$7D,$76
	.dc.b	$AE,$57,$FA,$12,$F8,$61,$FE,$6D,$C3,$85,$EA,$3F,$CA,$B3,$A7,$CC
	.dc.b	$D7,$F8,$83,$5F,$2D,$FF,$A6,$5F,$EB,$72,$F9,$DF,$AE,$DC,$3F,$D3
	.dc.b	$9E,$87,$EE,$1F,$F0,$AC,$66,$BF,$C1,$4D,$E7,$FE,$B0,$D9,$FF,$30
	.dc.b	$57,$1A,$DD,$F6,$F9,$59,$5D,$AE,$D7,$49,$B8,$7F,$A8,$F0,$F8,$F7
	.dc.b	$E2,$9C,$BE,$A7,$F4,$9E,$EA,$77,$7C,$EB,$81,$1F,$F4,$A4,$6A,$FD
	.dc.b	$37,$F0,$BE,$76,$5F,$C2,$F9,$1B,$87,$C0,$9B,$7E,$59,$B5,$F0,$3E
	.dc.b	$2F,$C7,$89,$BF,$47,$86,$4F,$CE,$72,$D3,$E7,$F9,$9F,$50,$DE,$5F
	.dc.b	$41,$C4,$68,$E1,$19,$7E,$BF,$CD,$7D,$C3,$C7,$F2,$16,$32,$49,$F8
	.dc.b	$71,$97,$EE,$F8,$BF,$84,$3D,$7F,$13,$6B,$F0,$71,$67,$E5,$CD,$D7
	.dc.b	$FD,$CD,$F8,$23,$E8,$FF,$89,$B7,$0F,$C6,$F9,$B8,$99,$1F,$EE,$49
	.dc.b	$D6,$1F,$5A,$B2,$FF,$AD,$23,$2F,$ED,$5A,$FF,$6F,$6B,$FE,$C5,$6A
	.dc.b	$F9,$B7,$CD,$FF,$26,$6D,$FE,$83,$D7,$C4,$C8,$FF,$7B,$CE,$AB,$F1
	.dc.b	$35,$97,$F1,$78,$FE,$03,$E2,$3F,$02,$DB,$FD,$5B,$67,$71,$FA,$E9
	.dc.b	$FE,$6F,$F2,$CF,$9A,$A7,$E7,$3E,$45,$F4,$78,$BF,$F7,$06,$FF,$D4
	.dc.b	$31,$61,$C6,$72,$7F,$84,$5C,$47,$FB,$CE,$F9,$CE,$A3,$FD,$D5,$3A
	.dc.b	$AE,$8F,$08,$FE,$44,$D7,$A6,$1F,$C2,$DE,$2D,$EB,$FD,$97,$89,$FA
	.dc.b	$2F,$8B,$FD,$1B,$71,$FE,$1A,$DB,$FD,$9B,$71,$F6,$7C,$2F,$F8,$76
	.dc.b	$BC,$D4,$8C,$29,$FE,$E3,$87,$FB,$BB,$1F,$EC,$35,$A9,$E7,$EE,$E0
	.dc.b	$FE,$0F,$F2,$9E,$FC,$7F,$C6,$D5,$B7,$FD,$F1,$F8,$E3,$B7,$F8,$6A
	.dc.b	$EA,$B1,$86,$CF,$CB,$CD,$5B,$8C,$3C,$DB,$3C,$34,$F5,$47,$82,$31
	.dc.b	$9F,$B2,$B9,$57,$F1,$8E,$1F,$E1,$BF,$15,$FC,$39,$B8,$7F,$38,$6C
	.dc.b	$EE,$3F,$9D,$CF,$F1,$4F,$5B,$F8,$AF,$C1,$59,$FE,$7E,$DA,$FE,$4A
	.dc.b	$DC,$7F,$99,$3F,$96,$FA,$B9,$5E,$FE,$5F,$F6,$0B,$F5,$B9,$53,$AC
	.dc.b	$3D,$A7,$E3,$9F,$CE,$65,$9C,$7F,$1F,$1F,$8B,$9F,$CD,$B6,$B9,$DF
	.dc.b	$FA,$72,$72,$FF,$A7,$BD,$17,$E8,$8F,$A0,$FF,$19,$FB,$AF,$1F,$FD
	.dc.b	$73,$39,$B6,$0A,$D2,$E3,$FE,$95,$CB,$CB,$0E,$38,$DA,$65,$E0,$8C
	.dc.b	$37,$1E,$B8,$F4,$C7,$9C,$63,$0E,$6D,$C3,$9E,$5E,$57,$C9,$3F,$ED
	.dc.b	$0F,$5F,$F3,$F8,$79,$B1,$ED,$5A,$3C,$73,$AA,$E1,$D1,$E4,$8C,$4F
	.dc.b	$6C,$F1,$33,$2F,$16,$1F,$E4,$78,$F5,$E5,$E3,$CB,$C1,$8F,$2A,$D5
	.dc.b	$FA,$71,$7C,$BB,$E1,$DE,$DF,$4C,$3C,$35,$A9,$CB,$E9,$31,$3C,$EB
	.dc.b	$93,$C7,$3F,$A1,$3A,$38,$4E,$9B,$5D,$32,$F3,$D7,$0A,$D4,$7D,$EA
	.dc.b	$E4,$D8,$CE,$BF,$23,$2F,$16,$1C,$E3,$FD,$75,$3A,$BF,$C4,$B7,$3F
	.dc.b	$87,$1E,$2C,$B8,$91,$FA,$D1,$AC,$7C,$18,$75,$9E,$6D,$7C,$B8,$78
	.dc.b	$DE,$84,$9E,$37,$F1,$71,$ED,$18,$9D,$57,$18,$C1,$1F,$5D,$F8,$70
	.dc.b	$F2,$46,$30,$E6,$DB,$F4,$E7,$F6,$B2,$F8,$51,$E6,$3D,$35,$E3,$7C
	.dc.b	$26,$0B,$7D,$9C,$63,$E3,$6E,$1E,$73,$C1,$19,$8E,$D3,$C5,$F0,$12
	.dc.b	$FE,$96,$1E,$5C,$7E,$54,$79,$8C,$E7,$F8,$38,$C9,$32,$F4,$63,$DE
	.dc.b	$B8,$4F,$C5,$BE,$95,$F1,$30,$F3,$9F,$B0,$C3,$A4,$66,$3C,$F3,$C1
	.dc.b	$F8,$66,$27,$6E,$2F,$CB,$9F,$17,$C3,$7E,$4C,$FE,$19,$2B,$C3,$9F
	.dc.b	$C3,$AD,$C7,$CE,$C7,$C1,$9F,$AF,$3F,$56,$7E,$08,$FC,$35,$7F,$A7
	.dc.b	$6F,$F5,$4D,$AF,$DA,$B7,$2F,$E6,$1F,$04,$F0,$9F,$3F,$86,$D7,$FA
	.dc.b	$9C,$4E,$E7,$47,$F3,$9C,$1F,$3B,$FF,$00,$6B,$84,$63,$E1,$35,$39
	.dc.b	$9F,$33,$5F,$D9,$BC,$73,$FD,$D3,$77,$CD,$7F,$1C,$F9,$FF,$F0,$45
	.dc.b	$FF,$0D,$AD,$3F,$BA,$6B,$FC,$69,$CF,$02,$78,$CF,$F1,$09,$DE,$3E
	.dc.b	$78,$C5,$F7,$E5,$FE,$9B,$18,$26,$5E,$69,$D4,$78,$D8,$EB,$2F,$EA
	.dc.b	$B9,$78,$6F,$9D,$F1,$AD,$1E,$9B,$E6,$E0,$F5,$3D,$23,$FA,$46,$5E
	.dc.b	$A7,$91,$30,$F8,$F5,$FC,$5A,$75,$18,$9D,$4F,$73,$57,$E4,$9D,$12
	.dc.b	$35,$3E,$09,$D4,$62,$74,$F9,$8E,$67,$9D,$CE,$3F,$36,$B9,$4E,$4E
	.dc.b	$44,$C7,$E3,$39,$3A,$47,$A2,$33,$1F,$16,$7A,$57,$9E,$73,$5E,$EA
	.dc.b	$E3,$39,$C3,$E1,$C7,$80,$C9,$E1,$72,$E0,$97,$D9,$FF,$03,$C6,$A3
	.dc.b	$11,$AC,$BF,$1F,$39,$7F,$1A,$4C,$BE,$EE,$1E,$5B,$E2,$75,$C7,$D3
	.dc.b	$8F,$C8,$BE,$18,$FC,$C3,$E0,$38,$BF,$97,$18,$C3,$E6,$5F,$28,$F7
	.dc.b	$61,$DA,$FC,$B9,$7D,$A9,$F4,$5F,$0C,$BE,$2E,$1C,$67,$27,$87,$1E
	.dc.b	$FC,$5F,$A4,$9F,$A6,$DA,$FD,$6B,$70,$FD,$4B,$6F,$E0,$E7,$FA,$4A
	.dc.b	$EF,$F3,$FF,$EA,$09,$FF,$0C,$B5,$F0,$4E,$EF,$F2,$38,$2B,$58,$FF
	.dc.b	$8A,$DE,$18,$F9,$5B,$87,$EC,$9A,$FF,$90,$1B,$5C,$F9,$FA,$C7,$4C
	.dc.b	$78,$47,$E8,$39,$DF,$2B,$FC,$83,$0F,$F0,$F7,$CD,$FF,$06,$EB,$FF
	.dc.b	$4C,$F9,$4F,$91,$E0,$DA,$D9,$FE,$A2,$E3,$5F,$F9,$8E,$3F,$C6,$2D
	.dc.b	$8C,$C7,$F2,$F8,$C1,$DF,$6F,$F0,$DB,$63,$FC,$36,$DC,$9D,$FF,$E9
	.dc.b	$F9,$FF,$4E,$66,$EA,$D2,$F9,$EA,$62,$24,$F0,$8C,$BC,$A3,$8F,$C0
	.dc.b	$72,$F9,$09,$3D,$A3,$2F,$82,$32,$FF,$A2,$B1,$F9,$73,$E7,$F9,$FF
	.dc.b	$14,$FF,$AC,$54,$64,$D6,$12,$75,$19,$78,$46,$5E,$D3,$C9,$D9,$A3
	.dc.b	$FD,$A7,$F3,$EF,$63,$E4,$79,$9F,$EA,$09,$FF,$61,$35,$FF,$50,$89
	.dc.b	$E0,$F2,$9F,$A4,$E5,$B1,$A8,$F8,$13,$F9,$DE,$3F,$F6,$FE,$AF,$9F
	.dc.b	$FA,$7C,$92,$FE,$BC,$7D,$63,$0F,$C9,$9F,$C6,$36,$3C,$3C,$6F,$9B
	.dc.b	$FC,$27,$84,$FC,$03,$C5,$93,$9B,$FD,$53,$39,$DB,$8D,$CC,$BF,$BF
	.dc.b	$3F,$93,$23,$83,$F3,$4E,$DF,$B9,$CA,$F6,$3F,$1B,$F1,$BF,$8E,$E8
	.dc.b	$95,$F9,$28,$CB,$F9,$48,$CB,$ED,$8F,$97,$F3,$EF,$A0,$FC,$E7,$BE
	.dc.b	$F8,$DA,$6A,$25,$FF,$B1,$37,$F7,$3F,$EC,$6B,$E4,$E7,$7F,$EF,$B1
	.dc.b	$88,$F9,$9F,$41,$F6,$BE,$83,$FC,$03,$1F,$97,$C9,$F1,$FE,$23,$51
	.dc.b	$0F,$F6,$E4,$FE,$ED,$34,$8E,$CF,$C5,$BD,$B1,$FC,$1F,$CC,$7E,$EF
	.dc.b	$C5,$7C,$FB,$DE,$FD,$E7,$CE,$EF,$6F,$F7,$0F,$84,$8C,$67,$FE,$DC
	.dc.b	$BE,$0F,$BC,$C5,$F5,$8C,$C7,$FB,$F3,$0E,$F3,$FC,$13,$9B,$5E,$FF
	.dc.b	$83,$A6,$5F,$3B,$F2,$CE,$A3,$18,$23,$11,$F9,$7F,$A0,$F0,$7B,$1F
	.dc.b	$F0,$06,$FD,$AD,$B3,$49,$5D,$1E,$2F,$F1,$AF,$A0,$FE,$32,$47,$07
	.dc.b	$F1,$EF,$A0,$C3,$FA,$78,$BC,$3F,$B0,$9F,$85,$ED,$F8,$49,$5F,$E6
	.dc.b	$1C,$7E,$8C,$62,$3E,$36,$28,$E3,$1F,$CB,$3E,$73,$E6,$7A,$3E,$B2
	.dc.b	$47,$3A,$FF,$28,$BF,$DA,$5F,$F2,$9E,$3F,$96,$C3,$EF,$D7,$2A,$FE
	.dc.b	$6F,$C3,$5F,$0C,$5A,$FE,$13,$37,$8F,$FC,$9F,$87,$46,$D7,$F3,$E9
	.dc.b	$FC,$07,$37,$CF,$26,$1E,$56,$BE,$E7,$6F,$E2,$79,$1C,$FF,$E7,$BC
	.dc.b	$BF,$5B,$2F,$A0,$FE,$27,$18,$FA,$07,$C7,$FF,$24,$6D,$7E,$4C,$E1
	.dc.b	$E4,$7F,$28,$BF,$E3,$71,$FB,$06,$BF,$E7,$26,$A3,$89,$FE,$97,$29
	.dc.b	$FC,$D9,$F1,$EB,$C6,$7C,$5E,$B7,$F3,$6F,$6F,$ED,$5F,$8F,$D0,$F9
	.dc.b	$0F,$56,$A8,$DB,$FD,$09,$AF,$F0,$66,$BF,$E8,$86,$BE,$76,$BC,$F9
	.dc.b	$3C,$44,$7C,$5F,$01,$5E,$D1,$75,$F9,$F6,$AF,$E9,$FF,$41,$FE,$BB
	.dc.b	$BF,$99,$F1,$7F,$D7,$DF,$EA,$ED,$BF,$AF,$3B,$9F,$EB,$6D,$7F,$8F
	.dc.b	$B5,$FF,$50,$B5,$FA,$0D,$4F,$38,$EB,$EF,$79,$0F,$EC,$75,$FE,$0F
	.dc.b	$AF,$C0,$7D,$03,$C2,$FF,$10,$36,$3E,$D3,$5F,$F6,$93,$5F,$E6,$6B
	.dc.b	$AB,$89,$23,$8B,$FE,$2B,$F7,$D6,$C7,$F1,$1B,$BE,$7F,$FD,$7B,$5A
	.dc.b	$9C,$C6,$2B,$86,$0B,$E0,$FE,$46,$F8,$39,$AF,$F7,$47,$BB,$CA,$3F
	.dc.b	$A4,$B5,$FF,$60,$9B,$32,$51,$F2,$CF,$8F,$FD,$6C,$7F,$B4,$A3,$1C
	.dc.b	$BF,$9F,$6C,$7F,$BD,$DA,$FF,$58,$5F,$3D,$CA,$3F,$51,$18,$9E,$53
	.dc.b	$99,$F0,$B7,$E1,$3F,$AF,$CF,$ED,$DB,$3F,$2E,$3F,$C1,$3F,$3A,$F9
	.dc.b	$EF,$ED,$FB,$FF,$04,$BD,$B7,$BF,$A1,$2F,$F7,$13,$D3,$0F,$ED,$EF
	.dc.b	$08,$FF,$74,$FC,$F7,$F7,$F7,$C5,$5D,$9C,$FA,$36,$FF,$82,$7C,$E3
	.dc.b	$E4,$71,$08,$C5,$7F,$63,$F9,$E7,$CF,$7F,$93,$A3,$F8,$A4,$EA,$BF
	.dc.b	$C7,$ED,$D3,$F8,$FB,$67,$F8,$95,$F2,$F8,$3E,$F3,$9B,$F1,$E3,$F3
	.dc.b	$63,$A4,$7F,$83,$1A,$FF,$1B,$3C,$D1,$B3,$D9,$C2,$FC,$BF,$FC,$AD
	.dc.b	$8F,$9A,$73,$18,$AF,$F2,$AF,$CF,$3E,$7B,$E0,$B6,$FF,$C0,$58,$3E
	.dc.b	$7F,$F9,$AC,$7F,$33,$9D,$5F,$2B,$FC,$24,$62,$3F,$C4,$6D,$7D,$0D
	.dc.b	$7F,$9A,$3C,$B5,$FC,$1F,$D6,$3D,$0F,$1D,$FD,$79,$4E,$70,$FF,$06
	.dc.b	$79,$4F,$A0,$FB,$9E,$8F,$F4,$A7,$EF,$79,$DF,$D4,$BB,$B3,$5B,$BE
	.dc.b	$35,$F9,$FB,$E1,$1F,$DE,$E7,$5A,$FF,$8F,$27,$2D,$87,$D0,$74,$3B
	.dc.b	$F9,$FF,$1A,$7F,$C4,$D3,$AB,$F3,$D6,$DF,$EC,$D5,$BD,$3F,$BE,$56
	.dc.b	$EF,$8D,$ED,$FF,$26,$45,$9D,$CF,$A6,$38,$4F,$FB,$BD,$AF,$F6,$1C
	.dc.b	$3F,$B6,$CE,$FE,$83,$E1,$9E,$8F,$8F,$B9,$7C,$C8,$D3,$F0,$DB,$7F
	.dc.b	$E5,$ED,$7F,$BE,$FC,$27,$C0,$3E,$0F,$CD,$5B,$F6,$C5,$AF,$F3,$24
	.dc.b	$DD,$3D,$FF,$41,$F5,$0C,$5F,$9F,$54,$66,$3E,$A3,$EF,$23,$F5,$FE
	.dc.b	$83,$EE,$9F,$4B,$E7,$2C,$7E,$1A,$3F,$07,$EB,$FD,$CC,$3E,$7C,$63
	.dc.b	$8B,$F0,$BE,$3C,$BF,$AB,$F0,$F7,$5F,$B7,$3C,$BE,$4F,$18,$FF,$A2
	.dc.b	$38,$3F,$CD,$91,$FC,$2B,$E8,$1F,$3F,$4F,$C2,$F9,$F7,$3F,$F0,$E4
	.dc.b	$E9,$FF,$0A,$37,$2F,$F4,$58,$70,$FF,$A3,$9A,$FF,$03,$78,$63,$E3
	.dc.b	$DB,$F1,$4F,$F8,$C2,$33,$5F,$E8,$29,$CB,$FC,$2F,$E5,$5F,$41,$FC
	.dc.b	$89,$B1,$F9,$6F,$57,$13,$23,$FC,$7D,$4E,$7C,$65,$FE,$F7,$8C,$4F
	.dc.b	$E7,$B4,$FD,$2F,$C3,$7E,$99,$B1,$93,$F4,$71,$96,$9F,$E4,$1D,$9F
	.dc.b	$41,$FC,$D9,$B1,$FE,$4F,$F8,$E7,$CF,$E6,$A4,$FF,$8F,$EB,$79,$7C
	.dc.b	$C6,$BF,$EB,$86,$BF,$CB,$5D,$F0,$7D,$DC,$3F,$C5,$55,$E3,$BF,$E1
	.dc.b	$B1,$8B,$FF,$60,$79,$1F,$C0,$BE,$1F,$F8,$1B,$6F,$F7,$ED,$9D,$C7
	.dc.b	$F0,$29,$FE,$A3,$F0,$6F,$98,$A7,$F1,$AF,$95,$F1,$D7,$F4,$19,$D3
	.dc.b	$FE,$65,$6D,$FF,$6F,$1C,$EF,$97,$7D,$5F,$4B,$FF,$3B,$46,$1F,$F3
	.dc.b	$C7,$D0,$BC,$2F,$E5,$2B,$C0,$B1,$86,$CF,$06,$E3,$E1,$C5,$7C,$6B
	.dc.b	$AD,$F1,$69,$C4,$FE,$47,$C7,$FC,$AB,$E2,$FF,$95,$B6,$32,$7F,$28
	.dc.b	$8C,$B4,$FE,$57,$F0,$7F,$CC,$BE,$17,$F9,$9B,$63,$27,$F3,$08,$CB
	.dc.b	$4F,$C7,$7D,$0F,$E3,$4F,$BB,$56,$32,$7E,$2E,$3F,$C4,$5C,$E6,$92
	.dc.b	$3F,$D4,$5D,$CF,$91,$47,$FA,$4F,$E0,$1E,$57,$E3,$EF,$FA,$BD,$EE
	.dc.b	$B5,$3F,$CE,$7E,$81,$F4,$3C,$B8,$98,$F2,$5E,$72,$B7,$87,$29,$FC
	.dc.b	$1F,$9E,$E8,$E5,$87,$1C,$38,$63,$8C,$BC,$0F,$7C,$7F,$C2,$13,$A7
	.dc.b	$FD,$58,$63,$2E,$F9,$7F,$AF,$A7,$4D,$8C,$E1,$DF,$DF,$FC,$BB,$E1
	.dc.b	$8C,$E3,$FD,$5E,$72,$DE,$BF,$25,$FB,$A7,$2F,$C0,$AE,$2E,$67,$78
	.dc.b	$2B,$F4,$B8,$FA,$9B,$5D,$63,$D6,$D3,$D5,$1D,$97,$CC,$FF,$AF,$7E
	.dc.b	$4B,$E6,$5F,$5C,$3C,$B5,$AF,$8C,$57,$D1,$8C,$36,$F9,$47,$DA,$CF
	.dc.b	$C1,$A7,$DB,$8D,$99,$3F,$DB,$53,$C2,$34,$D9,$E2,$7F,$91,$5C,$4F
	.dc.b	$0B,$F1,$BC,$CE,$B5,$F4,$F0,$F4,$D7,$09,$CD,$F3,$BE,$37,$BC,$FD
	.dc.b	$51,$89,$D6,$38,$3E,$2E,$7F,$27,$1E,$84,$AE,$18,$FD,$7C,$38,$3E
	.dc.b	$59,$D5,$F2,$CB,$CB,$7C,$2B,$CA,$7B,$67,$57,$CB,$2F,$8B,$5B,$C3
	.dc.b	$E1,$E7,$F5,$EB,$59,$FA,$31,$F0,$5F,$92,$B5,$18,$C7,$6F,$73,$EA
	.dc.b	$46,$B1,$EF,$3A,$9F,$2C,$E6,$FB,$56,$A7,$A1,$CB,$3F,$8D,$87,$9C
	.dc.b	$98,$7B,$27,$75,$C4,$97,$F8,$08,$F2,$46,$1B,$19,$7E,$39,$32,$F8
	.dc.b	$78,$78,$2F,$D1,$18,$CF,$E7,$C6,$31,$F8,$95,$BB,$F1,$E1,$EC,$3B
	.dc.b	$61,$F6,$23,$27,$DD,$29,$E1,$AE,$78,$76,$8C,$CF,$87,$3F,$06,$3F
	.dc.b	$43,$3F,$95,$5E,$C2,$1F,$3B,$2F,$26,$7F,$06,$31,$8F,$58,$FA,$11
	.dc.b	$FE,$A9,$3F,$5A,$5C,$7F,$66,$DC,$BF,$4E,$EF,$E1,$9E,$53,$E8,$7F
	.dc.b	$C6,$B2,$9C,$C6,$1B,$1F,$E3,$A6,$BD,$FE,$85,$F1,$3C,$B1,$46,$27
	.dc.b	$FC,$83,$18,$9D,$4E,$67,$94,$7F,$0B,$8E,$19,$7F,$04,$6F,$07,$FA
	.dc.b	$07,$E8,$7F,$A0,$5E,$EE,$3F,$AF,$AD,$BF,$D2,$A7,$6E,$67,$77,$D5
	.dc.b	$FE,$43,$3A,$8C,$4E,$C8,$FF,$4D,$8D,$5F,$F1,$5B,$EF,$5F,$CB,$1B
	.dc.b	$87,$BE,$3F,$5B,$5C,$1F,$E4,$33,$A2,$61,$E2,$E4,$FF,$72,$63,$E2
	.dc.b	$26,$3D,$E7,$57,$EA,$8C,$3E,$67,$E3,$57,$27,$35,$CA,$7C,$35,$C9
	.dc.b	$E0,$60,$BC,$0C,$7D,$11,$ED,$9F,$03,$BA,$F8,$98,$77,$8C,$61,$E0
	.dc.b	$23,$F4,$E3,$4F,$33,$C2,$59,$E6,$F6,$31,$3C,$B0,$F2,$57,$2C,$BE
	.dc.b	$75,$FB,$6F,$67,$73,$94,$6A,$F9,$63,$EC,$8C,$4F,$A0,$98,$76,$78
	.dc.b	$E5,$F5,$9F,$B7,$8F,$86,$31,$8F,$CD,$AE,$15,$EC,$BE,$AF,$92,$33
	.dc.b	$3E,$79,$C9,$E2,$8C,$E3,$F2,$2B,$D6,$71,$C7,$D1,$87,$82,$7D,$78
	.dc.b	$F8,$F2,$F3,$9F,$77,$2E,$D5,$BC,$3E,$21,$E5,$71,$8F,$82,$33,$3C
	.dc.b	$2B,$96,$1F,$8C,$9E,$A7,$C2,$32,$70,$74,$FF,$4C,$FA,$17,$57,$E1
	.dc.b	$70,$46,$27,$8C,$78,$4C,$CE,$E7,$53,$DC,$98,$7F,$8F,$E3,$2D,$9E
	.dc.b	$04,$9F,$F2,$47,$82,$F9,$FF,$E2,$B1,$FE,$19,$F9,$D3,$FE,$5E,$F9
	.dc.b	$3F,$CD,$53,$B1,$1F,$8F,$AD,$D7,$FA,$54,$FF,$15,$B8,$F3,$7F,$DD
	.dc.b	$55,$C2,$7C,$2D,$7F,$C0,$93,$BE,$AF,$F0,$F3,$FC,$CA,$39,$9F,$CA
	.dc.b	$DC,$71,$9F,$E6,$F8,$FA,$FA,$39,$57,$FA,$9B,$37,$D0,$E2,$BE,$03
	.dc.b	$63,$FD,$69,$3F,$03,$3F,$F8,$01,$18,$8F,$F5,$03,$C9,$D6,$5F,$1A
	.dc.b	$F9,$C6,$0F,$06,$1D,$7E,$1F,$B7,$CA,$49,$D5,$7D,$19,$D4,$63,$05
	.dc.b	$6E,$3C,$33,$E9,$F2,$BC,$51,$FE,$B3,$51,$93,$59,$C9,$FF,$61,$C6
	.dc.b	$5F,$F6,$2C,$65,$E5,$19,$7A,$47,$CA,$5F,$42,$F7,$FE,$57,$9D,$FE
	.dc.b	$A5,$9F,$5B,$5F,$C1,$4E,$DF,$88,$72,$9F,$B9,$19,$8F,$EF,$06,$4F
	.dc.b	$EE,$2E,$5F,$C6,$E7,$F9,$7F,$A1,$FB,$1C,$72,$7E,$44,$7D,$B3,$0F
	.dc.b	$D8,$8C,$3E,$2E,$27,$CE,$7E,$1B,$C3,$7D,$0F,$8F,$93,$FD,$65,$18
	.dc.b	$D3,$E1,$5F,$0C,$39,$61,$C7,$0E,$1C,$9F,$A8,$C3,$85,$68,$87,$F8
	.dc.b	$02,$BF,$32,$47,$E0,$E9,$FB,$3F,$A1,$FC,$A7,$34,$9F,$C2,$47,$E5
	.dc.b	$CC,$3F,$94,$8C,$3F,$3B,$E8,$1F,$43,$F9,$FF,$29,$F2,$54,$FB,$93
	.dc.b	$AA,$E1,$5F,$AF,$8C,$47,$D4,$6B,$FE,$2D,$6B,$FC,$03,$E8,$7F,$C6
	.dc.b	$1C,$3F,$ED,$9C,$D8,$BE,$53,$C4,$7E,$2E,$B7,$BF,$1C,$32,$7E,$92
	.dc.b	$B8,$3F,$88,$9D,$BF,$C5,$1F,$E0,$89,$A4,$7F,$C7,$3E,$0B,$E7,$DE
	.dc.b	$EF,$F0,$1F,$7F,$9E,$67,$AD,$6E,$7F,$84,$B6,$FF,$2A,$D7,$FC,$9C
	.dc.b	$D7,$FD,$D3,$7D,$9D,$AF,$6F,$C1,$C9,$27,$F8,$35,$79,$67,$BD,$6E
	.dc.b	$78,$C7,$AA,$31,$1E,$06,$BE,$0F,$53,$F8,$FF,$C8,$BE,$63,$F8,$4C
	.dc.b	$FF,$13,$8C,$78,$5D,$AA,$4F,$99,$EC,$47,$F2,$7E,$1B,$E8,$3F,$93
	.dc.b	$12,$3F,$8B,$7C,$8D,$7F,$92,$C5,$E5,$FE,$53,$3E,$CF,$67,$E8,$7D
	.dc.b	$34,$BF,$83,$AA,$75,$1F,$21,$F8,$04,$7E,$5F,$A1,$FC,$E7,$F1,$E7
	.dc.b	$B7,$4F,$93,$12,$76,$43,$E4,$3F,$2C,$D2,$C9,$D3,$FD,$83,$8F,$F1
	.dc.b	$15,$FD,$17,$9B,$75,$AB,$56,$9F,$5D,$6D,$F6,$F5,$7E,$3F,$E3,$A4
	.dc.b	$60,$87,$DB,$8F,$BC,$47,$FD,$23,$F4,$3F,$D9,$70,$F0,$43,$E8,$9F
	.dc.b	$23,$FC,$D5,$B1,$FC,$BA,$B8,$69,$FD,$D1,$B1,$FD,$F2,$76,$FF,$77
	.dc.b	$6D,$F6,$9D,$C9,$FE,$EC,$FF,$AB,$5A,$F5,$F2,$3F,$BB,$46,$23,$FC
	.dc.b	$13,$ED,$FD,$D7,$FA,$0F,$47,$FA,$42,$FD,$6D,$C7,$FC,$13,$5B,$AF
	.dc.b	$F0,$53,$6B,$7A,$FF,$53,$BE,$15,$DB,$8B,$FA,$B5,$EF,$0E,$71,$83
	.dc.b	$87,$C0,$5A,$FE,$A7,$5B,$79,$1E,$68,$FE,$C1,$F4,$5F,$EC,$48,$FE
	.dc.b	$B3,$F1,$DF,$DB,$9B,$E8,$EB,$DE,$91,$82,$1F,$E2,$3A,$E4,$D8,$DB
	.dc.b	$4E,$BE,$A7,$F6,$FF,$A1,$7D,$17,$F7,$1F,$A5,$FE,$67,$C3,$F8,$26
	.dc.b	$EF,$FE,$30,$AD,$E1,$FD,$B9,$7D,$1F,$1F,$90,$93,$98,$FE,$F9,$F0
	.dc.b	$4D,$FF,$45,$3F,$ED,$8F,$98,$FF,$1D,$CE,$73,$FF,$21,$FC,$0A,$73
	.dc.b	$7C,$23,$17,$FE,$48,$BE,$0F,$33,$FD,$E9,$EF,$F2,$BF,$EF,$ED,$8C
	.dc.b	$9B,$3F,$D7,$E5,$F2,$1F,$29,$FB,$1C,$A7,$DE,$BE,$93,$1C,$1F,$E5
	.dc.b	$2B,$DB,$FE,$33,$A5,$6E,$3F,$C2,$73,$70,$FE,$05,$1F,$DE,$7E,$9B
	.dc.b	$FB,$3B,$AC,$3A,$CB,$CF,$56,$A3,$FC,$C1,$F1,$3F,$D7,$EF,$FC,$3E
	.dc.b	$D8,$EA,$71,$F3,$DF,$4B,$D2,$13,$BD,$3F,$7E,$DB,$FE,$29,$F4,$BD
	.dc.b	$E7,$3E,$37,$93,$C8,$4E,$6B,$FC,$E7,$7C,$70,$FE,$2F,$39,$7F,$CF
	.dc.b	$B1,$9A,$E3,$EC,$F8,$FE,$8D,$F2,$B4,$F8,$3F,$48,$FA,$5F,$F2,$7E
	.dc.b	$7F,$E3,$CA,$DB,$6B,$83,$FE,$63,$AE,$1E,$1E,$33,$FF,$43,$5F,$07
	.dc.b	$E2,$F2,$3E,$97,$FC,$9D,$87,$33,$FA,$D3,$89,$E3,$8F,$D4,$AE,$1F
	.dc.b	$09,$D7,$87,$F9,$B4,$63,$0F,$F2,$8F,$D2,$3E,$93,$F9,$5E,$3F,$E9
	.dc.b	$88,$C4,$FF,$94,$A1,$18,$7F,$CA,$AD,$7F,$D3,$E6,$A7,$F0,$B6,$FA
	.dc.b	$47,$B1,$F7,$6F,$94,$FE,$66,$73,$87,$92,$F8,$C6,$30,$FF,$55,$5F
	.dc.b	$0A,$FF,$4A,$4E,$6F,$FA,$97,$17,$F4,$2F,$A2,$F5,$E1,$FD,$53,$D3
	.dc.b	$46,$27,$58,$7F,$86,$A7,$2D,$9F,$CE,$FB,$FF,$E2,$3F,$A3,$FC,$65
	.dc.b	$6B,$0F,$CC,$CE,$6F,$FC,$33,$2F,$A4,$D9,$CB,$E0,$FF,$AA,$CE,$A3
	.dc.b	$2D,$3F,$C7,$D3,$8B,$FD,$2E,$5F,$D7,$2B,$D1,$3D,$6B,$D1,$18,$23
	.dc.b	$FE,$AA,$FA,$6F,$DF,$DF,$2A,$E0,$FA,$23,$3A,$7C,$89,$79,$5D,$89
	.dc.b	$7F,$15,$B7,$F0,$3D,$AF,$F3,$5C,$FC,$09,$F8,$F1,$89,$F8,$B9,$FE
	.dc.b	$57,$0F,$F3,$13,$6F,$FD,$8C,$BE,$73,$E3,$72,$C9,$FF,$30,$3F,$26
	.dc.b	$31,$7E,$BF,$6F,$E1,$92,$BE,$D3,$63,$F5,$3F,$36,$3E,$CC,$7E,$95
	.dc.b	$C4,$FD,$8E,$1F,$9A,$47,$F0,$8D,$AF,$81,$C5,$F8,$28,$F7,$BF,$84
	.dc.b	$F2,$3E,$A9,$2B,$F1,$8D,$AF,$91,$F2,$9B,$9F,$F6,$FE,$C8,$FE,$41
	.dc.b	$F4,$8F,$A3,$FC,$CD,$A3,$17,$FB,$36,$C7,$FB,$19,$AF,$EC,$9C,$F0
	.dc.b	$7F,$A4,$70,$FC,$0B,$FE,$28,$8C,$D7,$FB,$DE,$72,$AA,$4E,$97,$37
	.dc.b	$FA,$17,$97,$CF,$8F,$E0,$1F,$E7,$1F,$2C,$FA,$4F,$4B,$71,$FE,$14
	.dc.b	$DC,$3F,$DB,$AD,$7F,$84,$B9,$CF,$D1,$CB,$D2,$FF,$1A,$FE,$AF,$1F
	.dc.b	$D6,$7C,$55,$C7,$FD,$4C,$DC,$7F,$22,$DC,$3F,$22,$FF,$4D,$D9,$F4
	.dc.b	$9F,$39,$B7,$FC,$7D,$B7,$FE,$F2,$6B,$FC,$79,$CD,$FF,$AF,$7E,$91
	.dc.b	$EE,$F5,$C3,$F8,$4C,$63,$0F,$F7,$FD,$F0,$7E,$76,$E9,$D6,$3F,$D9
	.dc.b	$0F,$C2,$5F,$2C,$3F,$85,$78,$5F,$C2,$FE,$32,$E3,$FE,$C5,$6D,$FE
	.dc.b	$C1,$B9,$FE,$C1,$FE,$D9,$F0,$4F,$77,$F1,$AD,$BF,$E7,$CD,$BF,$E7
	.dc.b	$61,$C3,$FC,$F9,$D5,$FF,$B7,$7E,$65,$F4,$9D,$B0,$FE,$53,$18,$C3
	.dc.b	$F1,$95,$FE,$E3,$FA,$3F,$E0,$11,$83,$E1,$AF,$A4,$FE,$62,$EB,$2F
	.dc.b	$E5,$EE,$67,$84,$7E,$52,$35,$5D,$6B,$8F,$7F,$26,$DF,$0E,$15,$F3
	.dc.b	$C3,$AD,$EE,$BF,$DE,$1F,$49,$FC,$E7,$E4,$7F,$9D,$36,$32,$7F,$BD
	.dc.b	$63,$2D,$3F,$DF,$1C,$3F,$D0,$7E,$1F,$FA,$13,$63,$27,$FB,$F6,$32
	.dc.b	$D3,$F0,$FE,$0F,$E5,$48,$FE,$55,$B8,$FF,$52,$EB,$FF,$60,$CF,$EE
	.dc.b	$7D,$C7,$CF,$FE,$EE,$BF,$87,$5F,$F5,$CB,$DC,$2B,$6F,$F0,$1E,$9F
	.dc.b	$E4,$AF,$5F,$8D,$FF,$5F,$4E,$A3,$9E,$3F,$2E,$F9,$CF,$2F,$29,$F4
	.dc.b	$6F,$A4,$E3,$A7,$FB,$2E,$31,$87,$9E,$3F,$A1,$F4,$EB,$2F,$F6,$5D
	.dc.b	$69,$F3,$4F,$1B,$D6,$1F,$E4,$6E,$8B,$7B,$CB,$C5,$EC,$7E,$6D,$F2
	.dc.b	$46,$70,$FF,$6E,$4E,$5E,$99,$F8,$F2,$FF,$6D,$CE,$5B,$87,$82,$31
	.dc.b	$8F,$BB,$2C,$4C,$BE,$56,$F9,$C7,$C6,$7F,$99,$E3,$D6,$74,$9D,$CE
	.dc.b	$BE,$03,$C5,$8F,$B6,$73,$7C,$75,$F6,$E3,$F1,$F1,$ED,$39,$D3,$F0
	.dc.b	$19,$73,$9E,$E6,$B2,$EB,$A7,$B6,$30,$6A,$78,$63,$F7,$DB,$7F,$23
	.dc.b	$1F,$04,$62,$BF,$A4,$35,$E7,$5E,$6C,$BD,$38,$C9,$ED,$97,$9D,$F8
	.dc.b	$46,$23,$55,$E1,$9D,$67,$F0,$EB,$44,$9D,$57,$B3,$1F,$2D,$F5,$AF
	.dc.b	$1E,$5F,$12,$39,$E7,$F0,$73,$F4,$12,$33,$7C,$B0,$F1,$DF,$17,$B6
	.dc.b	$78,$C7,$E0,$DF,$7C,$3E,$26,$7E,$7C,$FC,$B5,$E7,$C7,$93,$F8,$29
	.dc.b	$D4,$62,$76,$47,$E8,$61,$DE,$BD,$67,$83,$1F,$1B,$D4,$C1,$6F,$AB
	.dc.b	$8B,$E8,$DB,$EA,$F1,$3A,$9D,$A7,$8B,$8C,$3E,$35,$F8,$2F,$9C,$E7
	.dc.b	$1F,$93,$8F,$9E,$F7,$3E,$FC,$FE,$85,$7C,$72,$61,$CF,$2F,$84,$4A
	.dc.b	$F3,$1F,$86,$C7,$BE,$1D,$1F,$C7,$47,$A6,$31,$3E,$C9,$DD,$F6,$AF
	.dc.b	$16,$7F,$75,$E1,$1E,$86,$E1,$F7,$73,$F8,$38,$FA,$A7,$FD,$3E,$FE
	.dc.b	$B5,$B5,$FB,$D6,$D7,$FA,$A7,$E1,$1F,$16,$FA,$4C,$36,$BF,$B7,$CE
	.dc.b	$A7,$27,$FA,$C7,$07,$D1,$BE,$23,$0D,$BF,$EE,$F3,$A9,$C9,$23,$F7
	.dc.b	$93,$BD,$7F,$78,$DD,$BF,$C8,$1A,$BE,$93,$FC,$01,$7F,$C5,$2B,$4F
	.dc.b	$EF,$9A,$FF,$1D,$73,$3C,$79,$3F,$A8,$61,$DB,$1E,$44,$BF,$D0,$61
	.dc.b	$D9,$F2,$C6,$67,$84,$6A,$B9,$E3,$FC,$73,$1F,$0E,$5E,$FA,$E2,$6C
	.dc.b	$E5,$3B,$C2,$C3,$2F,$33,$83,$D5,$1A,$8C,$C7,$4C,$7C,$64,$AF,$8F
	.dc.b	$97,$A6,$31,$3F,$16,$F9,$4E,$6B,$C5,$3C,$DB,$5F,$D9,$E7,$53,$C4
	.dc.b	$D1,$63,$59,$7C,$28,$C5,$72,$23,$ED,$AE,$6F,$B0,$E6,$5A,$E6,$FC
	.dc.b	$1C,$7E,$54,$7A,$B1,$F8,$2F,$B2,$78,$5F,$69,$CB,$E0,$8C,$9E,$43
	.dc.b	$C5,$87,$9D,$C9,$2F,$BC,$FC,$A9,$D4,$62,$79,$1D,$2B,$BB,$88,$D3
	.dc.b	$9C,$3D,$F1,$82,$65,$F9,$48,$CB,$F8,$82,$65,$FD,$BA,$32,$FC,$38
	.dc.b	$E1,$87,$91,$DD,$FD,$4C,$3C,$6F,$AE,$33,$97,$CC,$8F,$9F,$8F,$8B
	.dc.b	$0E,$D8,$73,$7B,$18,$9A,$75,$C7,$F0,$78,$F6,$BE,$53,$F9,$D2,$61
	.dc.b	$F5,$63,$E4,$4F,$C6,$38,$63,$E4,$CB,$E7,$5F,$73,$8B,$87,$FA,$97
	.dc.b	$D2,$7F,$B4,$DD,$61,$EB,$24,$FF,$06,$9D,$4F,$F8,$1F,$E4,$7F,$C9
	.dc.b	$DC,$9F,$C9,$1B,$1F,$E1,$26,$BF,$E0,$66,$BF,$E2,$57,$F0,$24,$7F
	.dc.b	$C6,$5E,$0B,$A3,$FD,$1B,$1F,$E2,$69,$D3,$E1,$37,$1A,$7F,$C5,$B2
	.dc.b	$F4,$1E,$8F,$EA,$63,$0D,$8C,$CF,$81,$7D,$56,$66,$4E,$5E,$11,$C7
	.dc.b	$D0,$E5,$F5,$12,$B9,$46,$5E,$91,$97,$CB,$7F,$EE,$77,$DF,$E4,$FF
	.dc.b	$3E,$6F,$07,$9E,$31,$1E,$C7,$FD,$74,$E8,$86,$63,$87,$5F,$16,$46
	.dc.b	$B8,$7F,$CF,$FA,$2F,$9C,$FF,$95,$39,$5F,$A1,$FC,$48,$C3,$6F,$FD
	.dc.b	$0F,$3A,$DB,$E6,$7D,$33,$D6,$FE,$2D,$39,$C3,$FC,$D8,$D7,$9D,$70
	.dc.b	$9F,$A5,$E8,$49,$FA,$4F,$D5,$F8,$DC,$47,$C8,$8B,$8F,$AB,$75,$7F
	.dc.b	$0A,$FA,$1F,$F6,$CD,$FF,$95,$1A,$FF,$9C,$DB,$D4,$69,$FE,$A7,$F3
	.dc.b	$DD,$1F,$E6,$A8,$C3,$6B,$FD,$95,$E9,$FF,$B5,$DF,$F6,$99,$D4,$B7
	.dc.b	$D5,$B5,$C9,$71,$7F,$C0,$1F,$43,$0B,$E7,$9F,$0F,$85,$D9,$4C,$44
	.dc.b	$BF,$F6,$E4,$E9,$B3,$95,$A7,$B1,$0F,$A6,$7B,$DE,$58,$C3,$63,$31
	.dc.b	$FE,$E9,$E6,$FE,$03,$5C,$E3,$67,$C2,$9D,$BE,$0D,$FF,$DA,$51,$89
	.dc.b	$F1,$66,$BE,$75,$FE,$98,$FA,$4F,$AA,$4C,$3F,$C4,$4D,$7A,$1F,$E9
	.dc.b	$07,$D1,$B9,$F9,$7C,$0F,$A8,$CC,$4A,$FA,$AE,$A3,$E0,$7D,$1B,$E9
	.dc.b	$29,$2B,$E6,$FA,$15,$F0,$89,$E0,$FE,$09,$AF,$9B,$4F,$B9,$3F,$81
	.dc.b	$F1,$9F,$42,$F7,$70,$DB,$FA,$BC,$1F,$5E,$B7,$5F,$83,$D3,$6C,$AD
	.dc.b	$EF,$FE,$3A,$C3,$94,$78,$79,$3E,$23,$A3,$FC,$45,$F5,$FF,$41,$BE
	.dc.b	$A7,$96,$31,$AF,$F2,$9F,$AB,$C4,$CA,$E5,$A2,$F7,$D8,$BF,$CD,$B6
	.dc.b	$BC,$18,$73,$6C,$65,$F8,$CB,$D6,$E3,$52,$76,$F4,$8F,$37,$C0,$C9
	.dc.b	$F1,$E0,$8F,$0F,$8B,$99,$FC,$3E,$5E,$79,$C7,$0F,$E4,$B2,$FE,$45
	.dc.b	$5C,$BE,$93,$E1,$11,$F3,$FC,$65,$79,$E4,$9C,$5F,$3A,$BC,$3F,$E6
	.dc.b	$CA,$D5,$F5,$39,$94,$E7,$7C,$1F,$E0,$B8,$2F,$EA,$7C,$34,$9F,$7B
	.dc.b	$F3,$BE,$BB,$11,$E4,$8B,$97,$B6,$7F,$79,$F9,$6F,$BA,$BF,$80,$91
	.dc.b	$C3,$80,$E1,$D6,$DF,$CA,$A3,$95,$7F,$01,$6E,$9A,$AF,$E4,$7D,$3F
	.dc.b	$8D,$79,$DF,$88,$8E,$B8,$FF,$88,$7D,$1B,$E9,$3F,$9B,$B6,$7F,$9D
	.dc.b	$B5,$F3,$07,$0F,$A5,$D6,$5F,$C5,$5B,$8F,$C0,$6C,$FF,$22,$FA,$5F
	.dc.b	$F5,$53,$FB,$C8,$CB,$D6,$FC,$A7,$F2,$59,$DD,$F5,$8C,$CF,$F4,$2A
	.dc.b	$E3,$5A,$BF,$9F,$EE,$3E,$03,$F9,$77,$8B,$FE,$A1,$CB,$F9,$53,$6B
	.dc.b	$73,$FC,$CE,$B8,$63,$FD,$39,$AF,$F1,$26,$DF,$2F,$81,$E1,$52,$B8
	.dc.b	$3E,$08,$FE,$7D,$F4,$5E,$E2,$69,$FD,$4B,$F2,$5F,$F1,$45,$FF,$5C
	.dc.b	$97,$D2,$F0,$24,$7F,$60,$FA,$57,$C8,$7F,$49,$F8,$17,$D2,$7F,$50
	.dc.b	$AF,$F6,$C7,$D2,$7E,$42,$73,$B7,$F5,$2F,$5D,$39,$BE,$11,$8B,$FE
	.dc.b	$AF,$7C,$1E,$67,$E5,$FD,$4E,$57,$F8,$46,$C6,$4D,$9F,$ED,$42,$FC
	.dc.b	$CF,$F8,$23,$8F,$FB,$CB,$63,$FA,$6F,$C3,$38,$9F,$55,$FD,$A7,$5F
	.dc.b	$EC,$17,$B6,$F7,$E7,$1F,$D6,$CF,$EE,$5F,$E4,$3F,$89,$BE,$1B,$59
	.dc.b	$FF,$72,$AD,$3F,$D3,$3E,$75,$F2,$7A,$84,$63,$0F,$E7,$5F,$3C,$F9
	.dc.b	$FF,$F0,$3C,$7F,$84,$A7,$55,$CE,$F9,$46,$34,$FF,$01,$61,$C7,$DA
	.dc.b	$CD,$4C,$3F,$A3,$64,$F9,$FF,$F0,$2C,$75,$9F,$4C,$E7,$DB,$F3,$7C
	.dc.b	$F7,$AB,$2F,$F0,$23,$6F,$DD,$F5,$EF,$9D,$DF,$14,$9F,$F1,$4D,$6F
	.dc.b	$76,$4C,$38,$56,$AF,$85,$7F,$72,$BE,$17,$BC,$33,$8F,$8A,$FF,$C2
	.dc.b	$D7,$FC,$82,$FE,$7B,$FC,$7D,$AE,$63,$6F,$F8,$9B,$F2,$AF,$C0,$E3
	.dc.b	$FE,$3E,$6C,$7F,$84,$B1,$F3,$7C,$BF,$D7,$9C,$E3,$FE,$25,$6D,$FF
	.dc.b	$30,$6D,$FB,$B3,$5F,$1E,$59,$7F,$E5,$AB,$EF,$7F,$CD,$5A,$F3,$9F
	.dc.b	$3F,$A5,$FD,$CF,$E3,$3F,$1A,$DB,$FE,$DF,$F2,$98,$F4,$FF,$D3,$B1
	.dc.b	$FD,$A2,$B0,$7E,$53,$D7,$FF,$54,$56,$27,$FB,$6F,$C6,$71,$35,$F7
	.dc.b	$BF,$E6,$68,$C9,$33,$FE,$CB,$3A,$8E,$06,$77,$FF,$0E,$64,$9D,$4E
	.dc.b	$4F,$ED,$0F,$0A,$FF,$15,$5F,$77,$AC,$FF,$80,$3E,$23,$D8,$79,$3E
	.dc.b	$9A,$C6,$49,$3C,$A3,$2F,$4F,$9B,$F5,$9E,$CF,$3A,$D7,$F8,$63,$04
	.dc.b	$E2,$BC,$75,$73,$F5,$FB,$9F,$34,$C6,$1F,$E0,$17,$0F,$F8,$59,$CD
	.dc.b	$7F,$8C,$9F,$AF,$D1,$FE,$D6,$9F,$F1,$6B,$67,$FC,$95,$F4,$1F,$E5
	.dc.b	$B9,$CF,$C6,$F0,$23,$FE,$2E,$8D,$4F,$F8,$82,$B9,$61,$FE,$77,$F9
	.dc.b	$27,$C9,$7F,$0D,$6E,$1C,$7E,$5F,$13,$23,$FD,$3B,$3A,$C3,$C7,$59
	.dc.b	$7F,$CC,$7E,$C3,$87,$F8,$C3,$6F,$FD,$4F,$F3,$78,$99,$1F,$EA,$89
	.dc.b	$D5,$7F,$99,$FE,$29,$7A,$8E,$B7,$FB,$45,$5C,$1F,$F0,$C3,$5F,$F0
	.dc.b	$EB,$5F,$F4,$FF,$90,$F9,$4F,$E5,$8D,$C7,$FD,$39,$78,$99,$1F,$EB
	.dc.b	$C9,$D5,$7F,$9E,$FE,$37,$CF,$B7,$FA,$86,$72,$FF,$A7,$38,$3D,$37
	.dc.b	$F4,$FE,$6F,$F5,$91,$8C,$3E,$DC,$63,$4C,$B6,$B8,$1F,$17,$E1,$BE
	.dc.b	$59,$FB,$5D,$93,$89,$D1,$2F,$E5,$46,$5F,$AB,$5F,$27,$6F,$F3,$75
	.dc.b	$7F,$B9,$9B,$3F,$E2,$C6,$BF,$E4,$6E,$C7,$43,$E5,$7F,$D9,$5E,$8B
	.dc.b	$E5,$FF,$D9,$75,$A9,$CC,$62,$BE,$4D,$6E,$FF,$DC,$31,$CF,$1F,$F6
	.dc.b	$9C,$E7,$0E,$51,$8B,$E5,$87,$4C,$38,$F1,$5B,$DC,$30,$ED,$EA,$BF
	.dc.b	$20,$FF,$0C,$E5,$CE,$93,$9F,$6B,$97,$2F,$F8,$07,$0E,$38,$70,$BE
	.dc.b	$9A,$7F,$85,$E3,$13,$AB,$E3,$F2,$BE,$5C,$7A,$65,$DF,$37,$CE,$FE
	.dc.b	$D3,$0F,$73,$72,$F9,$2F,$F8,$A2,$B8,$39,$CD,$7C,$BE,$4B,$E0,$4E
	.dc.b	$63,$AF,$B7,$89,$97,$C9,$DC,$62,$FD,$67,$D2,$FB,$2F,$95,$C5,$FE
	.dc.b	$0A,$8C,$E1,$F8,$09,$CF,$A9,$A6,$E3,$ED,$B4,$62,$F8,$C3,$E0,$FA
	.dc.b	$9D,$BE,$CA,$DF,$18,$B5,$C2,$6E,$DF,$87,$8F,$DF,$F9,$5F,$81,$6D
	.dc.b	$FC,$DF,$05,$79,$E8,$5D,$25,$6E,$6D,$FC,$23,$99,$FE,$06,$AE,$32
	.dc.b	$8C,$52,$30,$7F,$10,$71,$10,$79,$8F,$9E,$FD,$0F,$1F,$38,$C7,$4F
	.dc.b	$E6,$D7,$CF,$F8,$2F,$EA,$C6,$B1,$FC,$C3,$63,$38,$74,$7B,$D7,$1D
	.dc.b	$7F,$5D,$5B,$6D,$6E,$9F,$2E,$F2,$9E,$FF,$6A,$D4,$E6,$31,$9F,$D1
	.dc.b	$8C,$5F,$8A,$31,$7C,$A3,$19,$FE,$BD,$EF,$19,$C7,$C7,$5C,$30,$D6
	.dc.b	$1D,$27,$F0,$0D,$D3,$E4,$65,$E4,$C3,$94,$7F,$B6,$A7,$58,$FD,$76
	.dc.b	$E7,$F0,$67,$55,$C2,$B4,$D7,$CB,$8F,$BB,$3F,$1C,$E7,$0E,$B5,$A3
	.dc.b	$DA,$7E,$25,$C9,$23,$B6,$1E,$19,$EF,$7D,$08,$F2,$CB,$DF,$5D,$C9
	.dc.b	$3E,$39,$ED,$18,$6C,$66,$78,$BD,$1B,$1D,$71,$F8,$33,$96,$C6,$4F
	.dc.b	$29,$D0,$93,$F5,$B1,$E1,$1A,$8E,$A6,$5D,$CF,$48,$CB,$F6,$33,$FB
	.dc.b	$D3,$A3,$94,$66,$BA,$C7,$CD,$AE,$2F,$EA,$B3,$FC,$06,$1A,$CF,$EE
	.dc.b	$C6,$27,$A1,$C7,$0E,$8E,$27,$EA,$57,$C7,$9F,$31,$27,$C6,$77,$C7
	.dc.b	$CF,$87,$8A,$3A,$19,$C7,$C5,$97,$D3,$9F,$84,$6F,$1E,$F5,$F3,$4E
	.dc.b	$24,$7D,$37,$DB,$0F,$14,$F6,$CF,$F1,$0E,$F0,$ED,$3E,$CC,$3B,$46
	.dc.b	$27,$C8,$7B,$DC,$57,$2C,$3C,$73,$FC,$6C,$FB,$E5,$3F,$38,$57,$F7
	.dc.b	$AE,$7E,$05,$E4,$3E,$7F,$84,$FF,$BD,$23,$0E,$27,$73,$A8,$46,$1E
	.dc.b	$35,$FE,$0B,$EC,$D3,$63,$3E,$0F,$F7,$66,$CE,$A3,$19,$FF,$06,$6E
	.dc.b	$9F,$C0,$DB,$5F,$CC,$BE,$7F,$FD,$E0,$FF,$0E,$8C,$BF,$C8,$9B,$87
	.dc.b	$86,$FF,$94,$D7,$0A,$FD,$F5,$68,$98,$7F,$34,$9D,$57,$8E,$3B,$12
	.dc.b	$BF,$9D,$CE,$B0,$F3,$4E,$6B,$85,$6B,$C9,$FA,$D1,$82,$57,$09,$D1
	.dc.b	$C5,$C4,$6A,$3E,$04,$F2,$9C,$92,$3F,$93,$61,$F1,$B5,$63,$E5,$9C
	.dc.b	$E3,$D5,$A7,$80,$F8,$6E,$5F,$94,$FC,$3C,$3C,$11,$C3,$1F,$23,$A3
	.dc.b	$93,$8C,$BE,$86,$5E,$5C,$7C,$77,$CA,$F8,$5E,$F2,$F4,$65,$E7,$9C
	.dc.b	$E1,$ED,$69,$F0,$8F,$36,$3D,$9F,$24,$7F,$73,$9E,$53,$99,$E2,$72
	.dc.b	$C7,$C6,$F2,$BF,$C0,$BE,$AB,$F0,$57,$13,$B5,$70,$9F,$8E,$4C,$3E
	.dc.b	$B4,$65,$F6,$65,$F3,$E3,$17,$D3,$0E,$F5,$F9,$4C,$BE,$26,$3D,$E7
	.dc.b	$35,$EB,$9C,$C6,$31,$F7,$9F,$0C,$F1,$4F,$5C,$7E,$C4,$7B,$5F,$16
	.dc.b	$5F,$12,$B9,$4F,$C3,$8C,$4E,$E3,$E1,$1D,$8F,$93,$87,$C2,$9F,$C6
	.dc.b	$65,$EC,$9D,$4E,$71,$FC,$3C,$E6,$78,$E1,$F7,$A7,$37,$FB,$3E,$0F
	.dc.b	$E3,$57,$F3,$5B,$19,$BF,$84,$7F,$8D,$8A,$79,$A7,$C4,$3E,$7F,$FD
	.dc.b	$97,$1F,$E1,$2E,$9F,$CC,$F6,$BF,$1E,$7C,$2E,$1B,$5B,$21,$FE,$41
	.dc.b	$3F,$C5,$B5,$FC,$BF,$2F,$E5,$8D,$8F,$F0,$C3,$5F,$F0,$C3,$F7,$08
	.dc.b	$FF,$97,$7D,$2F,$1F,$C9,$BD,$2A,$49,$FF,$47,$43,$D7,$C4,$7F,$9E
	.dc.b	$A3,$83,$A9,$E7,$3D,$3D,$2E,$BE,$B4,$AE,$71,$97,$AC,$65,$E9,$1F
	.dc.b	$1B,$E3,$FF,$D8,$91,$FE,$9F,$6B,$FE,$8A,$6F,$3F,$FA,$F5,$FF,$79
	.dc.b	$46,$F9,$39,$11,$E5,$D9,$5E,$51,$6B,$94,$FF,$0E,$E7,$7D,$07,$BA
	.dc.b	$FF,$D3,$3E,$73,$D7,$FB,$6D,$8F,$F6,$15,$70,$DB,$95,$F1,$C7,$FD
	.dc.b	$92,$DE,$BF,$F4,$15,$6E,$FF,$D7,$98,$3E,$83,$EA,$FA,$F2,$7E,$9B
	.dc.b	$F6,$3C,$FC,$47,$D3,$8B,$97,$FB,$7F,$8D,$5F,$BC,$FA,$0F,$EE,$DE
	.dc.b	$C7,$FA,$33,$AD,$E5,$BE,$3F,$F2,$AD,$CF,$FC,$D3,$7C,$3A,$FF,$CC
	.dc.b	$8F,$F7,$18,$FF,$3F,$B7,$C2,$34,$91,$FE,$9F,$C7,$FD,$35,$3A,$8C
	.dc.b	$57,$46,$9F,$ED,$4F,$5D,$F4,$1E,$18,$FF,$6B,$FC,$D3,$D2,$C6,$A8
	.dc.b	$FF,$62,$4F,$28,$C7,$81,$A9,$E1,$39,$6A,$CF,$E2,$47,$28,$C6,$5F
	.dc.b	$EB,$7B,$E7,$1C,$70,$F2,$2F,$A0,$7C,$3F,$93,$99,$ED,$53,$73,$AA
	.dc.b	$F7,$9C,$6B,$24,$7D,$55,$FC,$29,$F1,$9F,$E9,$3F,$9E,$F8,$9E,$7B
	.dc.b	$EC,$FA,$4E,$A3,$17,$E6,$E2,$5D,$3E,$C1,$2B,$FC,$4E,$D7,$B9,$FE
	.dc.b	$BC,$78,$3F,$37,$89,$EC,$FC,$2F,$A1,$7C,$CF,$D4,$F8,$17,$D4,$D3
	.dc.b	$51,$30,$F8,$35,$C1,$B1,$F5,$FE,$8B,$F0,$1E,$83,$E8,$5F,$45,$F8
	.dc.b	$0F,$03,$FC,$9D,$D5,$F9,$B6,$F4,$FF,$B8,$EB,$79,$7E,$33,$0E,$58
	.dc.b	$74,$9C,$CF,$CC,$9C,$BF,$85,$E0,$F0,$FC,$47,$D1,$8F,$E8,$AD,$7F
	.dc.b	$DE,$0D,$7F,$DE,$AD,$93,$F6,$72,$FF,$7D,$C6,$3D,$DE,$A7,$AF,$CA
	.dc.b	$97,$F9,$9D,$3F,$86,$3F,$80,$EC,$46,$30,$F8,$1D,$7C,$2B,$53,$9D
	.dc.b	$BF,$30,$D8,$5F,$2A,$F6,$FD,$17,$F9,$AF,$3F,$3B,$74,$34,$97,$F2
	.dc.b	$E3,$2D,$8F,$A9,$E7,$7C,$8F,$A2,$E4,$49,$FF,$3E,$DF,$C4,$8C,$47
	.dc.b	$D1,$B4,$71,$8F,$85,$F2,$15,$D4,$5D,$7F,$01,$37,$5F,$A7,$36,$BD
	.dc.b	$FA,$AB,$ED,$FD,$17,$B4,$93,$E4,$C5,$18,$8F,$23,$C1,$FA,$9F,$45
	.dc.b	$FC,$03,$E8,$9A,$FF,$02,$E7,$F8,$75,$FA,$86,$EA,$D7,$FD,$4D,$7C
	.dc.b	$2B,$F8,$0B,$5F,$F5,$0B,$5F,$F5,$53,$5E,$37,$FC,$C6,$38,$57,$59
	.dc.b	$08,$CD,$70,$8F,$E1,$F7,$F9,$38,$FF,$5C,$7D,$17,$F3,$F9,$FD,$FF
	.dc.b	$D1,$3E,$3F,$F8,$EB,$63,$F8,$BC,$EF,$1F,$04,$23,$93,$FE,$C1,$6B
	.dc.b	$DB,$CC,$FE,$91,$18,$8F,$3B,$CF,$0D,$47,$69,$F0,$FA,$3F,$CB,$AB
	.dc.b	$53,$C2,$70,$69,$FE,$57,$F0,$6F,$A2,$FE,$6B,$E6,$7F,$AB,$F2,$F3
	.dc.b	$B7,$2F,$F6,$C5,$6E,$F8,$DF,$F5,$D8,$C4,$7F,$B5,$1A,$FF,$B7,$5A
	.dc.b	$F2,$8F,$E8,$71,$C2,$FE,$2B,$6B,$FD,$CD,$6F,$9E,$FE,$93,$F4,$3F
	.dc.b	$0F,$13,$49,$97,$F6,$F5,$F1,$F4,$93,$FE,$E3,$8C,$BF,$EE,$5F,$85
	.dc.b	$7C,$77,$F5,$D8,$6F,$FE,$4E,$7F,$DD,$31,$98,$C4,$F0,$D7,$FC,$2B
	.dc.b	$F0,$7F,$EF,$D8,$C5,$F2,$8F,$F7,$BF,$AF,$A8,$FF,$80,$0D,$7F,$D9
	.dc.b	$E6,$CC,$94,$7C,$C7,$F8,$5B,$7F,$F7,$E3,$63,$F3,$64,$70,$E0,$E3
	.dc.b	$EC,$7F,$99,$5F,$D0,$C7,$EF,$E7,$9F,$66,$67,$BB,$67,$51,$CC,$7C
	.dc.b	$EB,$E6,$FF,$59,$A7,$E7,$A3,$18,$7F,$85,$BE,$75,$F3,$DF,$E1,$38
	.dc.b	$FE,$F7,$3A,$C3,$FB,$4B,$5F,$43,$5F,$6B,$5F,$DC,$B5,$F1,$E1,$E2
	.dc.b	$9F,$6F,$CF,$7F,$83,$9E,$F1,$99,$F9,$73,$97,$FB,$CF,$CF,$7B,$B2
	.dc.b	$FD,$BB,$6F,$E3,$FC,$F7,$EF,$DF,$F0,$9C,$66,$BF,$85,$B5,$39,$D1
	.dc.b	$5C,$67,$BE,$BE,$37,$6F,$D1,$AF,$9F,$1F,$65,$AF,$DE,$6B,$E0,$8F
	.dc.b	$66,$0F,$A8,$7D,$4F,$5B,$F3,$4E,$63,$C1,$F3,$EF,$9F,$A7,$C5,$CF
	.dc.b	$CF,$F4,$18,$D7,$FC,$63,$5B,$7F,$05,$A2,$30,$FE,$15,$AF,$E4,$5A
	.dc.b	$FE,$59,$DE,$3F,$CC,$FD,$C7,$E5,$6F,$88,$F0,$5F,$F8,$D6,$31,$87
	.dc.b	$F6,$BF,$8F,$FB,$5E,$7F,$F8,$21,$FB,$9E,$97,$F8,$03,$E6,$73,$E7
	.dc.b	$FF,$AC,$EF,$81,$FD,$66,$BF,$45,$7C,$63,$07,$F5,$DF,$90,$E2,$74
	.dc.b	$F9,$BF,$EA,$91,$FD,$EE,$32,$D3,$FC,$D5,$1A,$CF,$FB,$B4,$E9,$B3
	.dc.b	$A3,$1B,$7F,$87,$B1,$57,$91,$B3,$E3,$25,$7E,$EF,$0F,$F1,$3F,$C5
	.dc.b	$FB,$8F,$37,$D0,$DA,$FF,$08,$47,$32,$3F,$DE,$5F,$E0,$5A,$B3,$7C
	.dc.b	$9F,$BC,$F8,$1E,$5D,$BF,$6C,$5C,$BF,$C5,$D4,$AC,$6B,$EB,$F6,$7E
	.dc.b	$89,$8C,$3F,$C8,$2E,$1F,$F0,$53,$EF,$23,$FE,$20,$F4,$3F,$DD,$CF
	.dc.b	$F9,$4D,$B8,$7F,$8F,$7E,$07,$FC,$D7,$39,$F9,$0F,$C1,$D6,$B0,$FF
	.dc.b	$2F,$7B,$98,$9F,$C2,$FB,$28,$FD,$87,$C7,$3E,$4E,$9F,$8A,$C9,$F3
	.dc.b	$38,$CF,$FD,$23,$5B,$7F,$CA,$1E,$F6,$75,$FF,$25,$BF,$BB,$C5,$C7
	.dc.b	$4F,$CA,$FC,$3B,$E8,$71,$87,$FA,$72,$B6,$FE,$B1,$B8,$7F,$2F,$6B
	.dc.b	$FC,$ED,$AF,$F4,$07,$7B,$FF,$B5,$EF,$FB,$45,$75,$C3,$C3,$EB,$7F
	.dc.b	$0F,$F9,$97,$CA,$D3,$F5,$3F,$1D,$CA,$FF,$A5,$CE,$9F,$F5,$8B,$73
	.dc.b	$FF,$37,$7C,$07,$E2,$27,$37,$FB,$28,$C4,$FD,$7D,$7E,$67,$D0,$FC
	.dc.b	$D6,$DF,$FB,$47,$8B,$E2,$36,$74,$D3,$E3,$7C,$2F,$CD,$3E,$5C,$58
	.dc.b	$C9,$F2,$63,$FA,$D7,$07,$FB,$32,$30,$4C,$3F,$D0,$DC,$4F,$89,$FF
	.dc.b	$22,$7A,$DF,$C4,$67,$FC,$8B,$7C,$5D,$CE,$BC,$6F,$F6,$F7,$6B,$E5
	.dc.b	$F1,$5F,$E8,$88,$C6,$5F,$E4,$EC,$7A,$78,$1F,$AF,$E1,$DB,$FB,$1C
	.dc.b	$2D,$F0,$C1,$7C,$23,$0D,$BD,$D7,$83,$DA,$7D,$0F,$F8,$83,$2F,$06
	.dc.b	$7F,$E5,$E9,$CF,$BD,$E0,$E6,$FF,$04,$65,$EF,$9C,$DF,$0D,$3F,$C4
	.dc.b	$38,$F3,$7C,$78,$FF,$BF,$79,$7F,$80,$56,$EF,$8D,$68,$FF,$54,$CE
	.dc.b	$9C,$6C,$FA,$1F,$DC,$C6,$1B,$97,$C6,$7F,$C6,$78,$78,$EB,$87,$C7
	.dc.b	$FC,$89,$DE,$5F,$E7,$A9,$CD,$6E,$78,$E1,$D6,$F8,$12,$75,$EF,$D9
	.dc.b	$F2,$F0,$BE,$87,$E9,$FC,$AF,$D4,$6D,$F8,$E7,$4D,$9D,$34,$D7,$C2
	.dc.b	$7D,$C3,$EE,$FD,$0D,$8C,$93,$0F,$0C,$65,$F9,$47,$F1,$7F,$43,$F1
	.dc.b	$AD,$9F,$89,$D2,$AC,$F4,$AE,$9F,$E9,$FA,$FE,$10,$DB,$FE,$10,$D7
	.dc.b	$F3,$ED,$7B,$1F,$63,$E7,$3E,$5F,$17,$E9,$63,$2F,$E7,$23,$3C,$5F
	.dc.b	$EA,$CF,$39,$F4,$1F,$C1,$EB,$F5,$91,$8E,$BF,$D3,$AF,$A1,$F4,$67
	.dc.b	$F9,$18,$D3,$6F,$F8,$83,$FE,$5E,$8C,$E5,$FE,$AC,$6C,$66,$77,$3E
	.dc.b	$37,$C5,$8F,$F9,$9B,$E5,$BF,$AC,$7A,$DD,$72,$F9,$B1,$8B,$FC,$D6
	.dc.b	$3D,$B3,$F8,$13,$97,$D7,$8F,$4A,$FF,$65,$CE,$71,$ED,$E8,$32,$FD
	.dc.b	$65,$2B,$4B,$1E,$D7,$C2,$B7,$5A,$AE,$17,$CD,$F2,$E1,$E1,$73,$5C
	.dc.b	$27,$58,$FA,$1B,$87,$C3,$CB,$DF,$18,$BF,$4B,$5E,$67,$68,$DD,$F5
	.dc.b	$24,$67,$3B,$87,$79,$B8,$77,$9B,$5D,$A6,$E7,$F0,$23,$07,$E8,$33
	.dc.b	$F8,$07,$C6,$8D,$47,$93,$0E,$CE,$F2,$F5,$4E,$AB,$8C,$7A,$DE,$59
	.dc.b	$FC,$77,$38,$78,$A3,$13,$B2,$47,$E1,$71,$F4,$E1,$E1,$CB,$EB,$3F
	.dc.b	$04,$F1,$4E,$B0,$EF,$9F,$C8,$9C,$93,$3F,$C4,$C6,$5F,$59,$2F,$E4
	.dc.b	$4E,$DE,$F1,$98,$FC,$46,$5E,$8C,$3C,$19,$7C,$28,$C4,$F1,$3E,$6D
	.dc.b	$78,$32,$DD,$FC,$1C,$7C,$58,$F6,$25,$F6,$8C,$BC,$A3,$39,$FC,$FC
	.dc.b	$BE,$3C,$F7,$8C,$5F,$68,$F8,$D8,$FA,$5F,$94,$73,$3B,$63,$B8,$F7
	.dc.b	$57,$29,$F5,$1C,$1C,$47,$3C,$FE,$7D,$77,$AF,$F7,$04,$7E,$ED,$B8
	.dc.b	$7E,$A5,$B8,$7F,$C0,$1F,$2D,$F4,$34,$93,$FE,$DF,$F8,$8F,$F2,$1C
	.dc.b	$7F,$A8,$30,$FF,$00,$77,$5F,$08,$D2,$5F,$FB,$9E,$76,$FE,$76,$76
	.dc.b	$D9,$FE,$41,$F4,$2F,$27,$FD,$F7,$B7,$F1,$0B,$DB,$6B,$F8,$2C,$63
	.dc.b	$83,$F9,$85,$F1,$C3,$35,$1E,$C7,$F1,$C8,$D3,$8B,$E8,$DA,$E7,$87
	.dc.b	$89,$FE,$51,$87,$79,$FE,$3E,$D7,$F8,$E3,$9E,$06,$1D,$B0,$EA,$4B
	.dc.b	$E7,$8F,$C0,$C7,$B6,$3D,$23,$D1,$5B,$8F,$E4,$D7,$CE,$F8,$C7,$9C
	.dc.b	$E0,$7B,$6F,$9B,$E4,$BE,$73,$C2,$FF,$A9,$DF,$3A,$E3,$1D,$DE,$07
	.dc.b	$33,$89,$A9,$E4,$7C,$67,$27,$08,$F4,$61,$F0,$B0,$F1,$E5,$F8,$29
	.dc.b	$EC,$DB,$E5,$7E,$B3,$93,$D9,$F8,$35,$CB,$2F,$A1,$5C,$67,$C8,$68
	.dc.b	$F1,$B9,$79,$99,$3B,$39,$39,$61,$E7,$8E,$33,$D8,$FB,$98,$FC,$78
	.dc.b	$E6,$4A,$E3,$19,$7A,$4F,$1C,$BE,$F5,$74,$7E,$5D,$F7,$8F,$1C,$ED
	.dc.b	$FA,$06,$27,$77,$D6,$75,$97,$CE,$8C,$4F,$5A,$E3,$3F,$3A,$B7,$97
	.dc.b	$CE,$8C,$57,$4C,$BE,$63,$C6,$3F,$0B,$1F,$94,$C7,$CD,$1C,$30,$EB
	.dc.b	$7C,$CF,$AC,$FC,$E9,$DC,$66,$3E,$A9,$9C,$7F,$21,$3A,$8F,$A9,$5E
	.dc.b	$07,$FC,$45,$F4,$3F,$ED,$77,$FC,$07,$D5,$FA,$6F,$05,$CE,$F8,$AC
	.dc.b	$36,$B6,$43,$FC,$75,$1A,$3F,$27,$52,$3F,$CD,$D3,$C5,$D9,$C7,$0F
	.dc.b	$F4,$CF,$A4,$F4,$FC,$4D,$9F,$F3,$B3,$5F,$F3,$FA,$F6,$B7,$32,$76
	.dc.b	$F2,$8E,$7E,$A7,$4F,$6A,$57,$48,$CB,$DA,$32,$F6,$8F,$89,$E4,$7F
	.dc.b	$AC,$E3,$FD,$10,$D7,$FD,$22,$DC,$7F,$D6,$6F,$FB,$8A,$37,$C9,$C8
	.dc.b	$8F,$2E,$CA,$F2,$8B,$5C,$A7,$F8,$CF,$2B,$E8,$7E,$14,$FF,$B0,$3C
	.dc.b	$B7,$B1,$F6,$5B,$8F,$FA,$A6,$B7,$CF,$FE,$B7,$6C,$39,$15,$C2,$3F
	.dc.b	$D5,$6D,$C3,$FD,$75,$5C,$29,$C2,$F5,$3E,$B9,$30,$FA,$FF,$17,$5F
	.dc.b	$AF,$17,$0F,$BF,$3F,$C6,$3E,$81,$53,$D8,$78,$7F,$EB,$1F,$25,$F1
	.dc.b	$DF,$95,$6D,$7F,$96,$AF,$8A,$BE,$2F,$F9,$A2,$33,$18,$D1,$5C,$AB
	.dc.b	$FD,$92,$DC,$BF,$CE,$77,$C3,$1F,$F3,$E3,$6B,$E4,$56,$F7,$F3,$65
	.dc.b	$EB,$9F,$55,$FF,$B2,$7E,$7D,$E6,$71,$CD,$0E,$CC,$36,$F8,$9F,$55
	.dc.b	$E2,$E2,$78,$D7,$0A,$93,$9A,$DB,$F0,$EF,$E2,$FC,$FB,$E7,$BC,$5C
	.dc.b	$8F,$5E,$9E,$DB,$56,$DF,$19,$8A,$DB,$E3,$AE,$4E,$A7,$83,$E1,$3F
	.dc.b	$D3,$3F,$35,$F2,$7C,$A7,$CF,$EB,$93,$EB,$C6,$31,$F1,$79,$0B,$A7
	.dc.b	$1B,$E1,$A7,$E2,$2F,$86,$3E,$6B,$E0,$DB,$E1,$84,$8C,$1E,$CF,$9E
	.dc.b	$7C,$C7,$D5,$F7,$5F,$3F,$4F,$89,$3A,$AE,$15,$97,$EA,$57,$07,$27
	.dc.b	$DA,$F2,$9F,$3C,$F9,$FF,$C1,$77,$FE,$62,$BE,$37,$47,$EB,$1F,$C7
	.dc.b	$5F,$0A,$B7,$C6,$7F,$05,$7C,$78,$BF,$38,$DE,$1F,$F7,$35,$6E,$BE
	.dc.b	$96,$0F,$88,$FE,$81,$1E,$0F,$67,$F4,$58,$FE,$95,$B8,$7F,$BB,$EB
	.dc.b	$73,$FA,$CB,$E5,$EC,$7E,$6E,$F7,$87,$F0,$B6,$CE,$A3,$CF,$E2,$7C
	.dc.b	$02,$3E,$AE,$9A,$FC,$2D,$3F,$89,$63,$EC,$AF,$8F,$5F,$13,$27,$C2
	.dc.b	$F8,$7E,$36,$63,$F3,$E7,$2D,$9C,$BE,$F9,$E5,$1C,$FE,$7F,$D5,$E1
	.dc.b	$CA,$E5,$19,$7F,$61,$AA,$75,$1F,$33,$05,$6E,$30,$DC,$FE,$AF,$C9
	.dc.b	$5E,$0F,$99,$36,$BF,$99,$6C,$AF,$77,$CF,$FC,$DF,$85,$95,$ED,$8C
	.dc.b	$BE,$48,$CB,$E3,$8F,$DD,$7C,$F3,$CF,$FD,$E7,$4F,$C4,$AF,$DA,$1F
	.dc.b	$A0,$E0,$DC,$C8,$FD,$FC,$EF,$5F,$E5,$71,$B3,$F8,$1D,$F5,$6D,$76
	.dc.b	$B9,$18,$3F,$84,$FC,$EF,$E1,$97,$CF,$3E,$37,$F8,$83,$63,$7E,$4F
	.dc.b	$F1,$3A,$E1,$39,$9F,$4B,$72,$F2,$CF,$53,$F9,$E4,$FE,$B5,$AF,$F4
	.dc.b	$66,$BF,$D4,$1A,$9E,$55,$E1,$3C,$BE,$87,$F2,$5F,$7D,$F3,$DF,$C9
	.dc.b	$FC,$BF,$F4,$FE,$3F,$19,$BC,$BF,$D5,$2B,$79,$7F,$2A,$6B,$F4,$A7
	.dc.b	$6F,$F6,$26,$A8,$DD,$7D,$26,$D7,$5B,$F2,$1E,$3F,$8D,$FE,$75,$F3
	.dc.b	$DF,$B4,$F2,$5F,$1B,$F6,$B1,$FE,$5E,$DC,$3F,$B4,$D7,$D5,$1F,$1B
	.dc.b	$FD,$0F,$E9,$3F,$90,$69,$F8,$86,$E1,$F8,$4A,$DC,$FF,$46,$F8,$47
	.dc.b	$C5,$FF,$73,$8F,$F6,$CF,$C0,$FF,$76,$9C,$DF,$F8,$03,$87,$FC,$89
	.dc.b	$18,$AD,$9F,$E0,$47,$11,$FE,$49,$8F,$F0,$3F,$AF,$C6,$FF,$95,$42
	.dc.b	$30,$FF,$2D,$6B,$FE,$E0,$3A,$8F,$96,$7B,$3F,$E0,$8C,$7F,$A9,$FD
	.dc.b	$77,$F4,$2D,$FF,$B0,$D6,$9F,$EE,$BC,$29,$D6,$1F,$AC,$AF,$DD,$56
	.dc.b	$BF,$3F,$FF,$0E,$BC,$BC,$1E,$CF,$F5,$98,$CD,$7E,$D3,$E7,$DF,$3F
	.dc.b	$A8,$FF,$12,$C6,$2B,$FC,$0B,$F4,$2F,$A1,$FF,$06,$C7,$F7,$69,$D5
	.dc.b	$73,$BE,$59,$3E,$87,$FC,$05,$7B,$9C,$C6,$2B,$84,$FF,$84,$AB,$87
	.dc.b	$AF,$E3,$96,$1D,$B7,$FF,$17,$D6,$AF,$FC,$0F,$F3,$FF,$0A,$FF,$87
	.dc.b	$B6,$B7,$F4,$3F,$C0,$6B,$7B,$F5,$AE,$17,$FD,$95,$AF,$F8,$F9,$AF
	.dc.b	$F9,$21,$A9,$ED,$3F,$E4,$BF,$A1,$7D,$0F,$5B,$F9,$98,$A7,$35,$E2
	.dc.b	$3F,$C6,$31,$9A,$E1,$18,$9D,$E1,$FC,$AB,$E8,$1E,$97,$D7,$E5,$7D
	.dc.b	$03,$E1,$7E,$45,$6B,$1F,$C4,$CE,$5F,$F1,$6C,$67,$0F,$F1,$45,$F2
	.dc.b	$BE,$38,$75,$C3,$85,$78,$63,$13,$AC,$3A,$BF,$89,$F5,$FF,$C0,$1F
	.dc.b	$1B,$F8,$BF,$71,$EB,$FE,$FA,$FF,$B4,$4E,$63,$18,$FF,$5D,$8C,$78
	.dc.b	$9C,$FD,$6F,$F5,$85,$FF,$64,$8C,$7C,$6C,$8F,$EC,$E5,$BF,$EC,$D1
	.dc.b	$78,$BF,$C1,$D3,$4F,$F0,$74,$78,$7E,$07,$15,$FE,$0B,$6E,$7E,$86
	.dc.b	$E7,$FE,$45,$BF,$F0,$07,$C4,$FA,$7E,$B3,$CC,$D8,$C9,$FD,$EA,$32
	.dc.b	$D3,$FB,$EF,$D0,$7C,$13,$E1,$79,$D6,$32,$4C,$BF,$C5,$11,$97,$FC
	.dc.b	$05,$EF,$7C,$9F,$B4,$F8,$CD,$AF,$F0,$A3,$67,$71,$FE,$0C,$9F,$AD
	.dc.b	$EA,$FF,$B8,$63,$FC,$7B,$F0,$18,$F9,$1F,$B1,$EF,$23,$13,$A3,$FC
	.dc.b	$9D,$5C,$70,$FF,$15,$4F,$ED,$FE,$35,$F2,$3C,$BD,$BC,$37,$0F,$F3
	.dc.b	$EC,$E9,$FF,$26,$CE,$9A,$7F,$A0,$AB,$FD,$80,$FF,$A4,$FE,$85,$EB
	.dc.b	$FF,$05,$9C,$E3,$FE,$5C,$9C,$D7,$1E,$2F,$C9,$46,$27,$FC,$27,$5C
	.dc.b	$39,$FF,$D6,$D3,$97,$FC,$23,$19,$BF,$F6,$3F,$32,$B5,$1F,$EC,$29
	.dc.b	$D5,$FF,$B7,$67,$2F,$FB,$3B,$E5,$9F,$25,$FA,$3F,$7D,$CD,$FC,$76
	.dc.b	$73,$8F,$FA,$BE,$72,$FF,$A8,$A3,$37,$FE,$78,$9C,$BF,$E2,$C8,$CE
	.dc.b	$7F,$EC,$F9,$CB,$FE,$D1,$F8,$97,$D0,$FE,$CD,$FE,$EB,$87,$FA,$2A
	.dc.b	$31,$EE,$65,$B1,$93,$FD,$1F,$1F,$3F,$E2,$3F,$79,$C6,$F8,$6F,$F7
	.dc.b	$0F,$B1,$FC,$47,$5F,$F7,$6F,$0F,$FA,$FF,$1F,$F6,$55,$ED,$FE,$D7
	.dc.b	$E2,$FC,$9C,$79,$E3,$CA,$31,$3C,$23,$18,$74,$AD,$CF,$28,$C3,$B3
	.dc.b	$1C,$7F,$C4,$3E,$53,$8B,$63,$3E,$07,$CC,$D3,$D1,$39,$BF,$EF,$99
	.dc.b	$7E,$02,$98,$72,$6E,$1C,$B9,$31,$3E,$6F,$6D,$F4,$3F,$C7,$33,$EB
	.dc.b	$C9,$FE,$2D,$C3,$D1,$5A,$F1,$7C,$B3,$9D,$3C,$22,$75,$5E,$C9,$D5
	.dc.b	$FC,$F9,$CD,$70,$D7,$FC,$E5,$E1,$7C,$0F,$A1,$7D,$0D,$31,$F0,$17
	.dc.b	$3F,$8F,$3A,$6C,$E9,$A4,$C3,$B4,$ED,$B5,$CD,$F1,$70,$3E,$87,$E4
	.dc.b	$7C,$BF,$C9,$6E,$1F,$49,$B5,$C1,$BA,$FD,$4B,$7C,$37,$D1,$3E,$97
	.dc.b	$D0,$DC,$3E,$DC,$5C,$3C,$33,$6F,$F2,$91,$FC,$6F,$E8,$7F,$07,$F3
	.dc.b	$7F,$E4,$8C,$3F,$15,$39,$7F,$13,$19,$8F,$A6,$64,$F9,$CE,$6B,$AF
	.dc.b	$5A,$FA,$EB,$F9,$8B,$E1,$5F,$8E,$BE,$1D,$1F,$C2,$6F,$78,$FE,$6F
	.dc.b	$D1,$7D,$07,$E5,$69,$8A,$73,$D5,$FA,$45,$F4,$3E,$1C,$39,$61,$C6
	.dc.b	$F8,$35,$7C,$C7,$F9,$FF,$D7,$DE,$3E,$78,$C4,$FE,$4B,$1E,$F8,$7C
	.dc.b	$69,$CE,$BF,$BF,$9C,$D7,$C6,$9C,$E7,$E8,$9C,$D7,$D2,$9C,$D7,$93
	.dc.b	$0F,$07,$3B,$2E,$CD,$9E,$8D,$BF,$06,$7E,$BC,$39,$D7,$B2,$B5,$97
	.dc.b	$C8,$8C,$61,$EE,$CF,$AB,$23,$AB,$96,$93,$1F,$45,$F4,$8F,$45,$71
	.dc.b	$9C,$D5,$BF,$81,$8F,$97,$0E,$18,$78,$8E,$93,$AC,$BE,$0D,$F1,$9F
	.dc.b	$BF,$97,$C6,$CF,$E4,$DF,$AE,$3D,$E4,$C3,$C9,$3A,$AE,$31,$E0,$9D
	.dc.b	$5F,$B7,$2F,$3D,$78,$8D,$C6,$70,$ED,$7C,$49,$19,$9D,$9C,$4F,$65
	.dc.b	$7D,$4C,$7D,$71,$9C,$3D,$59,$7C,$3B,$F7,$4F,$0C,$7E,$16,$78,$BF
	.dc.b	$3C,$FE,$46,$F8,$3C,$F2,$F7,$C6,$32,$F5,$93,$0E,$75,$BC,$7E,$66
	.dc.b	$7F,$16,$3C,$07,$AA,$B8,$3F,$70,$E0,$7C,$58,$E1,$8F,$92,$77,$18
	.dc.b	$C7,$E2,$D6,$F1,$F2,$63,$F2,$B1,$F0,$C6,$27,$C1,$39,$7D,$D9,$FC
	.dc.b	$8C,$7C,$71,$E6,$C7,$B6,$1F,$17,$1F,$83,$7B,$C3,$96,$1E,$4C,$7B
	.dc.b	$63,$EE,$39,$38,$CB,$F7,$B1,$FA,$53,$F2,$85,$72,$F9,$4D,$7C,$4B
	.dc.b	$E8,$5F,$43,$FE,$E2,$94,$E6,$30,$D8,$FF,$73,$B5,$ED,$E0,$7F,$5D
	.dc.b	$7F,$DD,$77,$FD,$BB,$B7,$A3,$63,$33,$B7,$8C,$FF,$13,$8E,$59,$7F
	.dc.b	$0D,$6E,$7F,$C4,$7E,$85,$E4,$FF,$03,$AD,$5F,$12,$61,$FC,$7E,$B6
	.dc.b	$E2,$76,$FF,$1D,$C3,$AB,$FC,$CE,$76,$E6,$B8,$61,$FB,$7A,$DF,$0A
	.dc.b	$F7,$87,$42,$56,$EF,$FA,$14,$EA,$31,$87,$52,$57,$F3,$19,$D4,$7F
	.dc.b	$08,$C3,$CA,$FF,$3E,$6E,$5F,$2E,$B8,$65,$F0,$F0,$EB,$1E,$2B,$E5
	.dc.b	$8F,$8A,$B5,$18,$3C,$35,$C0,$F9,$D1,$A8,$F0,$4F,$C2,$78,$57,$81
	.dc.b	$C1,$A3,$99,$F2,$E7,$95,$F9,$EF,$B4,$F2,$7B,$BC,$A3,$4E,$4E,$46
	.dc.b	$70,$F8,$27,$D2,$C7,$E3,$C7,$87,$0F,$80,$DB,$EA,$F9,$0C,$1D,$72
	.dc.b	$FB,$58,$F9,$B0,$EB,$97,$BE,$73,$5E,$18,$C6,$3E,$0C,$BC,$F3,$D0
	.dc.b	$EB,$5C,$1F,$04,$68,$F6,$C6,$72,$F3,$C6,$70,$F6,$63,$E3,$C7,$C1
	.dc.b	$87,$E6,$E3,$0F,$49,$DB,$AB,$FC,$06,$1D,$72,$F9,$F3,$E5,$C7,$C3
	.dc.b	$97,$A7,$0F,$7C,$FD,$8B,$E0,$FD,$13,$84,$76,$7C,$19,$7C,$2C,$BF
	.dc.b	$17,$5C,$27,$73,$F8,$F6,$1C,$0F,$51,$8C,$BE,$0D,$F5,$7F,$19,$1A
	.dc.b	$23,$FE,$15,$FA,$1F,$F6,$D4,$7F,$86,$7A,$BF,$37,$E0,$3A,$1F,$1B
	.dc.b	$86,$D6,$C8,$7F,$90,$23,$47,$E5,$AA,$47,$F9,$22,$32,$F0,$C3,$FD
	.dc.b	$1B,$E9,$BD,$4F,$0B,$67,$FD,$2E,$D7,$FC,$F4,$BD,$BD,$CC,$9D,$BC
	.dc.b	$A3,$9F,$A9,$D3,$DA,$95,$D2,$32,$F6,$8C,$BD,$A3,$E1,$F8,$FF,$E9
	.dc.b	$E8,$FF,$47,$6F,$83,$3D,$56,$32,$43,$FD,$21,$3F,$ED,$E7,$19,$FF
	.dc.b	$A8,$2B,$84,$6A,$BF,$85,$F0,$BE,$87,$E0,$5F,$FA,$CB,$E8,$1E,$D7
	.dc.b	$D6,$6E,$1F,$EA,$8A,$DF,$4F,$F9,$93,$0E,$5B,$EF,$09,$7B,$8F,$9F
	.dc.b	$EB,$48,$F9,$EF,$D2,$F8,$FC,$47,$D0,$8B,$9F,$D8,$9F,$E2,$7F,$35
	.dc.b	$FE,$4B,$F5,$9E,$27,$F9,$D3,$A5,$E9,$BE,$53,$F1,$AD,$CF,$FC,$D1
	.dc.b	$87,$1B,$FF,$1D,$36,$17,$CA,$F8,$9E,$08,$FF,$40,$E1,$E0,$7F,$C8
	.dc.b	$B5,$C6,$B8,$47,$FB,$83,$0F,$03,$FE,$73,$9D,$7D,$0B,$CF,$BD,$92
	.dc.b	$BA,$E7,$EB,$AD,$46,$0F,$2B,$85,$97,$8F,$24,$FF,$AD,$A7,$84,$62
	.dc.b	$FC,$38,$7B,$E5,$F4,$2F,$8D,$F5,$73,$3D,$6A,$6E,$75,$5E,$33,$8D
	.dc.b	$69,$EF,$5C,$5D,$4F,$07,$CF,$A7,$F9,$A7,$E0,$BE,$87,$98,$FA,$1D
	.dc.b	$6F,$F3,$63,$19,$79,$FC,$85,$D3,$8D,$F0,$C3,$FD,$6B,$7C,$31,$F3
	.dc.b	$5F,$06,$DF,$0C,$24,$60,$F6,$7D,$03,$E7,$3E,$E7,$BC,$FA,$1A,$7C
	.dc.b	$99,$D5,$70,$AC,$BF,$42,$B8,$39,$3E,$D7,$94,$FA,$07,$D0,$FE,$27
	.dc.b	$C9,$FF,$21,$70,$FD,$B6,$F0,$7F,$AE,$6B,$78,$72,$6B,$F3,$0F,$CC
	.dc.b	$CF,$06,$9F,$9D,$C3,$C1,$85,$F8,$9F,$E6,$73,$C7,$A1,$ED,$7E,$63
	.dc.b	$3F,$C6,$37,$3F,$F6,$55,$F0,$9F,$D1,$62,$8C,$4F,$EC,$EB,$86,$1F
	.dc.b	$AC,$6D,$6F,$D9,$F8,$1E,$1C,$AF,$CB,$E9,$FC,$39,$F2,$3F,$8B,$6C
	.dc.b	$6D,$A7,$0E,$9F,$67,$C3,$7E,$CF,$17,$C5,$53,$35,$F4,$B4,$56,$63
	.dc.b	$F4,$F5,$C1,$D4,$7C,$7F,$A1,$E9,$E0,$49,$E5,$19,$78,$C6,$5F,$A9
	.dc.b	$AF,$C5,$9D,$46,$30,$46,$37,$FC,$07,$C8,$5E,$0F,$9F,$36,$FE,$EC
	.dc.b	$FD,$9F,$A1,$F7,$7C,$2C,$9F,$64,$65,$F5,$C6,$5F,$04,$7F,$01,$FA
	.dc.b	$17,$A1,$FC,$0F,$A7,$E1,$D7,$EF,$5A,$FE,$34,$4F,$07,$F8,$13,$5D
	.dc.b	$47,$0C,$FF,$93,$47,$EA,$F7,$F0,$5C,$8E,$D1,$E1,$30,$F7,$FA,$0F
	.dc.b	$C4,$AF,$A1,$7C,$77,$F1,$96,$C7,$F1,$39,$D9,$C2,$33,$1E,$96,$E1
	.dc.b	$E4,$9D,$F9,$5C,$EE,$5F,$38,$FE,$4D,$3F,$CE,$A3,$FD,$6C,$D4,$47
	.dc.b	$0D,$8F,$F5,$A2,$F8,$07,$D0,$FF,$31,$F2,$BF,$D2,$F8,$FC,$66,$F1
	.dc.b	$F4,$BE,$15,$FC,$DE,$F8,$C7,$F3,$1A,$DB,$E2,$6B,$C0,$FE,$6D,$1C
	.dc.b	$1F,$F6,$71,$92,$B9,$7F,$DA,$4D,$7C,$5F,$41,$FB,$AF,$1D,$F1,$9F
	.dc.b	$73,$2F,$E6,$CD,$CB,$79,$FF,$4D,$B7,$C5,$FF,$4F,$FA,$1F,$E4,$F9
	.dc.b	$7E,$2D,$B8,$FE,$1A,$B7,$3F,$D3,$7E,$09,$F1,$1F,$D7,$A1,$EB,$FF
	.dc.b	$91,$1F,$F6,$FC,$66,$31,$3C,$2F,$FC,$29,$EF,$FC,$B8,$C5,$F2,$8F
	.dc.b	$87,$EB,$6A,$3F,$DE,$4D,$7F,$DA,$C6,$CC,$94,$7C,$D3,$D7,$FF,$02
	.dc.b	$61,$FD,$67,$E8,$7F,$A4,$70,$7F,$67,$AD,$3F,$DE,$F8,$93,$AC,$3F
	.dc.b	$5D,$4E,$FE,$77,$FD,$73,$E8,$7F,$1E,$F3,$8C,$D3,$05,$6A,$3F,$71
	.dc.b	$F3,$AF,$9F,$FD,$9C,$7E,$36,$31,$87,$E1,$21,$18,$7F,$0E,$D7,$F2
	.dc.b	$86,$A7,$C3,$8B,$E7,$FF,$C1,$57,$FE,$05,$BE,$5A,$7F,$89,$EB,$5E
	.dc.b	$A7,$A3,$1E,$57,$FE,$1D,$9C,$BC,$71,$EB,$8F,$4A,$FE,$25,$F3,$DF
	.dc.b	$16,$BD,$0D,$8F,$CE,$B5,$F3,$B5,$FD,$63,$BC,$7E,$BF,$CE,$BE,$7B
	.dc.b	$C9,$39,$BF,$F1,$FC,$E5,$FF,$10,$46,$6F,$FC,$63,$39,$7C,$BB,$FD
	.dc.b	$4C,$13,$97,$F9,$1D,$71,$73,$87,$82,$B8,$56,$F0,$FE,$05,$F3,$DF
	.dc.b	$D6,$9F,$C4,$7C,$F3,$E1,$3E,$54,$E7,$3F,$F0,$17,$12,$B5,$7F,$E3
	.dc.b	$7A,$D6,$7F,$E2,$D9,$CD,$FF,$3D,$9C,$D7,$F8,$AA,$73,$5F,$CE,$A7
	.dc.b	$35,$F8,$F9,$CD,$79,$BD,$0F,$C9,$E1,$E1,$C7,$F2,$BE,$77,$F8,$17
	.dc.b	$E2,$BF,$2F,$EB,$7F,$83,$BC,$0C,$FA,$1F,$EA,$DC,$BF,$B8,$C6,$2B
	.dc.b	$7E,$97,$F0,$A9,$FE,$F1,$18,$9E,$3F,$15,$27,$FB,$C9,$76,$FE,$DD
	.dc.b	$17,$83,$FC,$37,$34,$FF,$0D,$C7,$6F,$77,$15,$FE,$17,$6E,$7F,$E2
	.dc.b	$46,$E5,$FE,$52,$BF,$F0,$57,$C2,$F9,$BE,$7B,$D6,$D8,$C9,$FE,$0C
	.dc.b	$8C,$B4,$FF,$07,$FC,$DF,$B8,$F7,$F9,$56,$32,$4C,$BF,$C6,$D8,$7F
	.dc.b	$4B,$8C,$69,$FE,$32,$DF,$FD,$6F,$1F,$E1,$AF,$9E,$F9,$BE,$D7,$CA
	.dc.b	$6C,$64,$FF,$11,$46,$5F,$F2,$AF,$A1,$FE,$E5,$7F,$C6,$7E,$E7,$DA
	.dc.b	$8C,$65,$FE,$34,$8C,$7C,$4F,$DC,$F7,$51,$8B,$E2,$7F,$A1,$FE,$25
	.dc.b	$F1,$BB,$F5,$FF,$81,$46,$27,$FC,$83,$18,$9D,$47,$16,$E3,$FE,$13
	.dc.b	$B7,$1F,$E3,$3E,$75,$F3,$DF,$C3,$67,$38,$7F,$A8,$A7,$2F,$FA,$66
	.dc.b	$33,$7F,$E6,$19,$CB,$D6,$FA,$5F,$36,$DF,$16,$CF,$F8,$97,$E4,$9F
	.dc.b	$1F,$F9,$EF,$71,$CF,$FC,$7E,$73,$87,$FA,$D2,$72,$FF,$AA,$E3,$37
	.dc.b	$FE,$73,$9C,$BD,$6F,$A5,$FE,$AB,$E1,$9F,$3D,$7D,$AC,$37,$2F,$F4
	.dc.b	$93,$70,$FF,$3F,$57,$09,$FF,$4B,$57,$C7,$F8,$5F,$DA,$F1,$3E,$0F
	.dc.b	$FD,$BD,$EA,$FF,$0F,$C7,$FC,$3F,$C5,$FE,$C5,$C7,$FD,$9B,$7B,$7F
	.dc.b	$C4,$75,$BA,$FC,$1E,$1D,$30,$E7,$E2,$FF,$87,$71,$E5,$5F,$E9,$1C
	.dc.b	$79,$2C,$39,$4E,$BB,$3F,$DB,$D5,$A9,$CE,$3D,$72,$F0,$E3,$FE,$91
	.dc.b	$F9,$1E,$F9,$7F,$A4,$5B,$E1,$7F,$A7,$B8,$6F,$7A,$69,$30,$E7,$7F
	.dc.b	$E9,$9A,$C3,$E5,$C3,$FC,$83,$EC,$BE,$7B,$F8,$B4,$E4,$65,$FE,$6E
	.dc.b	$F0,$BC,$75,$BF,$1B,$C7,$CF,$FB,$BB,$E1,$39,$CF,$C3,$87,$17,$C7
	.dc.b	$19,$9E,$D3,$96,$FC,$EB,$E7,$BD,$ED,$BE,$6F,$F8,$B1,$B1,$A2,$3A
	.dc.b	$F7,$6D,$74,$9B,$C5,$F8,$3C,$18,$73,$AD,$61,$E8,$E1,$7C,$F7,$CE
	.dc.b	$F9,$3F,$9E,$D9,$D3,$6B,$D7,$5C,$27,$95,$7D,$6F,$83,$91,$D7,$E7
	.dc.b	$AD,$F0,$9B,$9F,$97,$05,$63,$8F,$F2,$1F,$3D,$F8,$2F,$98,$F4,$5F
	.dc.b	$E3,$A3,$1C,$1F,$88,$CB,$F8,$DC,$63,$3F,$C1,$C6,$1F,$AB,$7C,$B0
	.dc.b	$EB,$18,$BE,$5D,$6C,$39,$61,$F8,$5F,$45,$F3,$BF,$C0,$33,$FD,$74
	.dc.b	$63,$A7,$F3,$2B,$E7,$B9,$67,$F9,$48,$D3,$6F,$96,$9F,$E8,$9F,$95
	.dc.b	$FF,$48,$7A,$B9,$C7,$D9,$18,$CF,$C5,$7C,$1F,$F4,$DC,$67,$3F,$1C
	.dc.b	$E6,$53,$9C,$3F,$D4,$39,$78,$A7,$F4,$06,$BA,$FF,$CA,$8F,$6C,$7C
	.dc.b	$31,$8C,$7F,$83,$61,$CE,$BC,$33,$9C,$3A,$E3,$E1,$AD,$57,$5C,$BE
	.dc.b	$9C,$71,$CB,$C5,$8F,$4C,$79,$5F,$4B,$DC,$62,$B7,$8F,$C7,$8D,$D7
	.dc.b	$53,$D7,$87,$5C,$7C,$53,$A8,$F1,$99,$CB,$DB,$8F,$62,$5F,$C1,$CE
	.dc.b	$E7,$E5,$9B,$97,$BA,$F8,$DF,$83,$1E,$C7,$9F,$2F,$0D,$F0,$9F,$61
	.dc.b	$D4,$E0,$E5,$F5,$61,$F6,$AF,$AB,$E1,$8D,$9C,$5C,$13,$0F,$89,$5F
	.dc.b	$89,$C7,$A6,$5F,$03,$2F,$26,$3D,$2F,$78,$F4,$C7,$C1,$87,$0B,$F5
	.dc.b	$E3,$CE,$B5,$7C,$2F,$C1,$0C,$FE,$AB,$EE,$8C,$BE,$C9,$F9,$18,$75
	.dc.b	$BF,$89,$7C,$2B,$47,$1A,$EC,$70,$7C,$26,$49,$1C,$72,$F7,$57,$4C
	.dc.b	$FE,$8D,$70,$3A,$D6,$F2,$F6,$C7,$48,$C5,$79,$63,$E7,$61,$E2,$7C
	.dc.b	$07,$2B,$EB,$97,$BB,$1F,$C6,$65,$C6,$3E,$2E,$3D,$2F,$C5,$5A,$6E
	.dc.b	$5F,$B0,$7F,$D6,$F1,$FB,$F6,$C6,$5F,$AE,$7F,$AF,$3E,$19,$F3,$CF
	.dc.b	$9E,$FF,$62,$46,$1B,$19,$9F,$F6,$18,$9D,$38,$BF,$ED,$9D,$9F,$EC
	.dc.b	$D6,$C6,$7C,$0F,$F7,$DB,$67,$51,$8C,$FF,$82,$B7,$6F,$E3,$8D,$AF
	.dc.b	$C3,$7C,$F3,$CA,$47,$F0,$49,$D3,$FC,$61,$B8,$F8,$B0,$FE,$5F,$7C
	.dc.b	$6B,$86,$31,$FE,$03,$7D,$F6,$56,$E7,$35,$BB,$F3,$BF,$CD,$67,$51
	.dc.b	$89,$D9,$1F,$0E,$3E,$88,$C6,$5E,$EB,$E8,$FF,$2E,$BE,$6E,$67,$67
	.dc.b	$F3,$8C,$7C,$D3,$A8,$C5,$72,$69,$FD,$46,$34,$FC,$58,$D3,$4D,$5F
	.dc.b	$97,$0F,$51,$30,$F1,$3E,$A9,$E3,$97,$CD,$9C,$9E,$9A,$E4,$E2,$7C
	.dc.b	$75,$CD,$FB,$73,$A7,$33,$B7,$C9,$3D,$DF,$82,$62,$3A,$65,$F5,$AB
	.dc.b	$94,$E4,$E6,$48,$D3,$E8,$CB,$F0,$A7,$CA,$9E,$38,$7A,$E7,$89,$CE
	.dc.b	$B8,$BE,$28,$D3,$9C,$BE,$6C,$60,$97,$D6,$7C,$19,$7D,$5B,$E7,$97
	.dc.b	$C3,$9F,$41,$CF,$1F,$1D,$F9,$AF,$86,$3D,$A3,$15,$F8,$4C,$39,$E5
	.dc.b	$F2,$72,$F0,$E1,$F9,$5C,$79,$BE,$2B,$FB,$E5,$CB,$EA,$61,$CA,$3E
	.dc.b	$0D,$F2,$76,$7C,$1C,$7C,$35,$F5,$4F,$54,$ED,$C6,$3E,$39,$DB,$F8
	.dc.b	$5A,$F2,$1E,$17,$0F,$69,$E7,$3F,$26,$3F,$CC,$FF,$9C,$7E,$7F,$13
	.dc.b	$4F,$D1,$77,$BE,$63,$FC,$A7,$5F,$E6,$A9,$D6,$9F,$8F,$BE,$2F,$E4
	.dc.b	$6F,$87,$7E,$6B,$77,$C6,$B5,$87,$48,$FF,$2C,$FC,$0F,$1E,$E3,$4C
	.dc.b	$7D,$9D,$9F,$F3,$34,$57,$FC,$C3,$1C,$7D,$49,$1F,$E6,$B7,$C9,$F8
	.dc.b	$4F,$F2,$88,$FF,$33,$C6,$9A,$7F,$A2,$23,$5B,$FF,$9D,$FE,$77,$F8
	.dc.b	$74,$FF,$A2,$1F,$E8,$D1,$9E,$BF,$81,$1A,$8E,$13,$89,$8F,$FA,$52
	.dc.b	$75,$C5,$FC,$C3,$81,$18,$9F,$F5,$0B,$DA,$74,$D8,$C9,$FB,$FE,$44
	.dc.b	$BB,$F0,$D9,$FF,$30,$35,$FF,$50,$E4,$F8,$ED,$D4,$9D,$BF,$22,$3E
	.dc.b	$4F,$B3,$F2,$AE,$57,$C7,$8C,$BF,$26,$32,$F6,$8F,$E0,$7E,$2C,$8F
	.dc.b	$F2,$8B,$F8,$7F,$37,$FD,$FF,$18,$21,$FE,$4A,$8D,$7A,$7C,$63,$06
	.dc.b	$B7,$FF,$56,$57,$28,$FF,$5E,$F0,$3D,$7F,$CA,$2F,$7E,$98,$89,$3F
	.dc.b	$EC,$89,$D3,$67,$2F,$FB,$3B,$E8,$1E,$DF,$E8,$5B,$1F,$EC,$E6,$BF
	.dc.b	$EC,$EF,$9F,$7D,$4E,$EA,$4E,$DF,$D4,$4E,$A3,$FC,$EB,$F3,$9F,$D7
	.dc.b	$7D,$67,$91,$FE,$95,$8C,$36,$7F,$BC,$CE,$9F,$F7,$4F,$BA,$FA,$9D
	.dc.b	$C7,$FA,$5A,$EB,$FE,$94,$6F,$8A,$77,$D6,$BE,$55,$DB,$1F,$02,$C7
	.dc.b	$C0,$F3,$34,$77,$2D,$73,$C7,$C3,$D6,$F8,$FA,$4A,$DE,$9F,$97,$C3
	.dc.b	$AD,$6E,$7A,$9F,$A8,$AE,$0F,$8F,$1F,$D6,$5F,$17,$FC,$0D,$E0,$BE
	.dc.b	$6B,$E0,$B6,$B9,$B5,$E4,$33,$7D,$13,$DC,$E3,$72,$B8,$BF,$22,$75
	.dc.b	$1F,$0B,$C0,$7C,$D7,$E4,$E3,$0D,$7B,$C6,$BA,$3E,$8F,$83,$EF,$9C
	.dc.b	$D6,$F8,$3E,$F5,$EE,$79,$DF,$0D,$E4,$7D,$0C,$39,$5F,$E1,$74,$46
	.dc.b	$36,$FE,$24,$BE,$A3,$8D,$FC,$2A,$DE,$18,$C3,$95,$EE,$BD,$7F,$0B
	.dc.b	$F8,$EC,$FD,$ED,$D0,$FB,$38,$F0,$CA,$E2,$79,$E3,$87,$B9,$23,$E3
	.dc.b	$BF,$13,$DC,$C4,$7C,$48,$B9,$79,$26,$ED,$FE,$EE,$C5,$58,$BF,$B9
	.dc.b	$9A,$F1,$9F,$D6,$AD,$6B,$F1,$A7,$D9,$E9,$FE,$7B,$3F,$A2,$DC,$FF
	.dc.b	$DF,$6D,$CB,$AF,$B1,$C2,$64,$7E,$0E,$75,$87,$D2,$AF,$D9,$FC,$EB
	.dc.b	$E2,$E9,$2F,$F3,$FC,$AF,$98,$C4,$7E,$58,$57,$17,$F4,$AD,$75,$3C
	.dc.b	$5C,$D7,$2C,$FF,$73,$5B,$7F,$77,$5B,$AE,$D3,$F2,$9E,$BA,$FA,$0F
	.dc.b	$E1,$73,$FA,$F8,$FF,$45,$CE,$DE,$33,$B7,$D7,$E5,$57,$27,$43,$25
	.dc.b	$32,$4B,$F2,$C7,$F1,$AE,$17,$CD,$FF,$AB,$2B,$FD,$37,$18,$70,$D7
	.dc.b	$FD,$62,$D7,$FB,$EB,$9A,$FE,$36,$DD,$FF,$D1,$CD,$9F,$F5,$3F,$CE
	.dc.b	$FF,$82,$EB,$77,$E3,$24,$FF,$16,$9D,$5F,$FA,$5F,$E6,$DF,$01,$FC
	.dc.b	$C9,$B5,$FC,$71,$AF,$F1,$2F,$3D,$F9,$63,$EA,$B8,$65,$2B,$83,$FC
	.dc.b	$E3,$EB,$5E,$5F,$F3,$EF,$71,$CD,$FC,$DE,$DE,$D7,$F8,$05,$FF,$60
	.dc.b	$CE,$B8,$3F,$D8,$73,$99,$FF,$6B,$57,$18,$C5,$6E,$79,$FA,$9F,$ED
	.dc.b	$CF,$43,$8B,$63,$27,$F8,$94,$A7,$FA,$94,$BF,$36,$F9,$8F,$EB,$59
	.dc.b	$7F,$4C,$8C,$71,$FF,$4A,$FA,$CF,$F6,$C3,$FD,$0A,$33,$7F,$A3,$7E
	.dc.b	$34,$67,$83,$9B,$FE,$DE,$E2,$7C,$F7,$F7,$48,$D1,$FD,$33,$E6,$F9
	.dc.b	$E5,$FD,$71,$B1,$FE,$E6,$C1,$F5,$2F,$67,$AE,$7F,$EE,$AB,$DB,$75
	.dc.b	$CA,$8C,$DF,$2A,$CC,$70,$9D,$57,$F6,$C7,$FD,$BD,$1E,$E7,$FC,$A6
	.dc.b	$D7,$31,$B7,$FD,$F7,$F3,$6F,$9A,$F0,$B7,$2F,$EE,$B8,$FA,$BC,$BC
	.dc.b	$D4,$9F,$ED,$B5,$BB,$FE,$C7,$7B,$9C,$D6,$E7,$F8,$95,$6F,$4F,$F0
	.dc.b	$95,$EE,$73,$5B,$9F,$F0,$3D,$6E,$FE,$66,$9F,$D4,$5E,$A7,$8B,$DE
	.dc.b	$F4,$9F,$3B,$77,$CD,$FF,$82,$67,$44,$CB,$C7,$5B,$9D,$46,$27,$58
	.dc.b	$F8,$AB,$5E,$6F,$69,$FE,$6D,$E4,$7D,$CF,$28,$C8,$FE,$D9,$F0,$D9
	.dc.b	$8C,$63,$FD,$E9,$B8,$FF,$46,$7F,$0B,$E5,$7F,$70,$AE,$11,$C7,$E3
	.dc.b	$66,$1F,$DD,$FE,$BE,$E3,$CE,$6E,$3F,$DC,$26,$D7,$F5,$09,$B5,$FD
	.dc.b	$72,$6E,$5E,$89,$CE,$5F,$E0,$09,$CE,$3F,$C4,$B0,$FE,$A9,$E2,$79
	.dc.b	$BC,$E7,$D3,$7B,$62,$4E,$8F,$F0,$64,$63,$3F,$F0,$35,$3E,$83,$91
	.dc.b	$CF,$EB,$ED,$FF,$84,$E2,$D7,$09,$BB,$7F,$6C,$F5,$BE,$61,$8B,$FE
	.dc.b	$DA,$E1,$FE,$D4,$E6,$9B,$FD,$68,$FD,$7F,$C0,$FF,$A0,$6B,$FC,$8E
	.dc.b	$DA,$FF,$0E,$7C,$BB,$8B,$7E,$BC,$BD,$BF,$62,$75,$18,$3F,$CA,$3F
	.dc.b	$10,$F8,$FF,$DE,$B6,$3F,$CB,$3F,$33,$89,$91,$C1,$FC,$1F,$BA,$9C
	.dc.b	$E5,$FD,$EF,$D3,$7C,$BF,$F0,$86,$E1,$FE,$6A,$F8,$07,$CC,$66,$A4
	.dc.b	$FF,$99,$2B,$77,$FE,$39,$BD,$CE,$6B,$73,$DD,$FC,$AC,$EA,$31,$3F
	.dc.b	$E7,$AF,$39,$F3,$0F,$98,$FF,$47,$79,$FF,$ED,$0B,$FE,$97,$39,$B4
	.dc.b	$7F,$A8,$7D,$DD,$FC,$67,$F4,$CA,$D4,$AF,$96,$3D,$B1,$DC,$F4,$6F
	.dc.b	$80,$E0,$FF,$11,$67,$DB,$0E,$38,$F4,$7C,$94,$F8,$E7,$C8,$72,$5E
	.dc.b	$1D,$8C,$37,$0E,$9A,$FF,$BD,$6B,$71,$8D,$3F,$D0,$91,$D2,$31,$87
	.dc.b	$7B,$F4,$D7,$FB,$CF,$E5,$AF,$11,$F5,$F5,$7E,$C6,$7E,$16,$5F,$21
	.dc.b	$E1,$8F,$76,$9A,$3E,$6D,$F3,$8F,$0D,$73,$D3,$EA,$57,$06,$D6,$E3
	.dc.b	$91,$9B,$FA,$F9,$2F,$94,$78,$EB,$85,$FC,$CF,$01,$F3,$3F,$51,$FC
	.dc.b	$44,$66,$31,$CD,$96,$E7,$F1,$AB,$8B,$67,$EF,$F7,$7F,$96,$BE,$0E
	.dc.b	$FC,$0F,$F9,$D6,$B8,$7C,$D6,$26,$47,$D3,$AE,$0E,$8C,$57,$1D,$FE
	.dc.b	$46,$EF,$88,$7C,$A7,$E3,$1F,$DC,$4E,$9C,$DF,$DC,$73,$B7,$E3,$1C
	.dc.b	$CE,$FB,$B9,$5F,$36,$DF,$16,$CE,$E3,$E1,$7D,$2F,$C9,$3E,$57,$81
	.dc.b	$63,$24,$9F,$4F,$07,$F8,$81,$B5,$F4,$7E,$A6,$57,$CD,$FB,$3B,$5C
	.dc.b	$26,$E9,$F6,$66,$F2,$FF,$8F,$BF,$25,$57,$EC,$3B,$D1,$8C,$7E,$17
	.dc.b	$CB,$E6,$31,$97,$E4,$EB,$77,$93,$E6,$55,$BF,$D4,$4E,$9B,$7C,$B2
	.dc.b	$F9,$95,$BC,$3F,$3D,$C6,$F9,$BF,$A3,$9F,$E9,$7E,$4B,$FB,$9F,$AB
	.dc.b	$F8,$19,$CD,$FF,$10,$9C,$BC,$A3,$2A,$33,$8F,$F1,$9C,$3B,$71,$7E
	.dc.b	$E2,$73,$8F,$7F,$45,$97,$EC,$E7,$38,$FF,$50,$CB,$AC,$62,$3B,$57
	.dc.b	$07,$F5,$F1,$C3,$1E,$FA,$F3,$AD,$67,$FB,$8A,$E4,$6F,$0E,$B3,$AC
	.dc.b	$BC,$99,$71,$3E,$16,$7E,$99,$D6,$7F,$0D,$AF,$92,$77,$97,$18,$F2
	.dc.b	$CE,$B1,$F0,$5F,$3B,$D6,$5E,$5A,$EB,$9F,$AB,$0E,$35,$AC,$BC,$B1
	.dc.b	$F4,$70,$E4,$E8,$98,$7D,$1C,$FD,$B7,$C6,$7D,$0F,$99,$DD,$F4,$AF
	.dc.b	$14,$E5,$B3,$A3,$89,$E1,$CF,$A3,$F0,$4D,$63,$E9,$9E,$24,$BE,$C7
	.dc.b	$F5,$A8,$D3,$89,$EB,$3B,$7F,$05,$8F,$BA,$31,$1D,$DF,$74,$66,$79
	.dc.b	$4E,$5E,$D8,$FC,$28,$F5,$19,$3A,$E1,$EA,$CF,$F0,$B9,$FD,$2A,$E1
	.dc.b	$8F,$82,$7E,$CD,$6E,$BE,$E1,$BC,$7C,$58,$7A,$B1,$F6,$65,$EE,$8C
	.dc.b	$10,$F1,$39,$7C,$84,$C3,$BB,$E8,$7C,$B9,$FD,$5A,$F7,$C6,$27,$EA
	.dc.b	$1B,$C3,$CB,$5B,$8F,$04,$EF,$2F,$93,$97,$AE,$BF,$0D,$8F,$8E,$B5
	.dc.b	$8F,$AB,$3F,$86,$FE,$4B,$1F,$ED,$87,$E0,$4B,$3F,$C1,$9B,$97,$ED
	.dc.b	$DD,$FC,$17,$F9,$17,$E6,$F8,$C7,$FB,$9B,$CD,$46,$21,$3B,$76,$D7
	.dc.b	$BF,$3F,$FB,$C1,$E7,$87,$F5,$2E,$2E,$4D,$8C,$CE,$DE,$33,$FC,$46
	.dc.b	$39,$6F,$FC,$1D,$B9,$FE,$5F,$E6,$DD,$1F,$E0,$BA,$D5,$F1,$26,$1F
	.dc.b	$CB,$AB,$6E,$27,$6F,$F1,$8C,$3A,$BF,$E6,$E9,$DB,$9A,$E1,$87,$F3
	.dc.b	$5A,$DE,$EB,$DE,$1D,$09,$5B,$BF,$EF,$53,$A8,$C6,$1D,$49,$5F,$CD
	.dc.b	$E7,$51,$FC,$FF,$0F,$2B,$FC,$F5,$B9,$7C,$BA,$E1,$97,$C3,$C3,$AC
	.dc.b	$78,$AF,$96,$3E,$2A,$D4,$60,$F0,$D7,$03,$E7,$46,$A3,$C1,$3F,$09
	.dc.b	$E1,$5E,$07,$06,$8E,$67,$CB,$9E,$57,$E7,$BE,$D3,$C9,$EE,$F2,$8D
	.dc.b	$39,$39,$19,$C3,$E0,$9F,$4B,$1F,$8F,$1E,$1C,$3E,$03,$6F,$AB,$E4
	.dc.b	$30,$75,$CB,$ED,$63,$E6,$C3,$AE,$5E,$F9,$CD,$78,$63,$18,$F8,$32
	.dc.b	$F3,$CF,$43,$AD,$70,$7C,$11,$A3,$DB,$19,$CB,$CF,$19,$C3,$D9,$8F
	.dc.b	$8F,$1F,$06,$1F,$9B,$8C,$3D,$27,$6E,$AF,$F0,$18,$75,$CB,$E7,$CF
	.dc.b	$97,$1F,$0E,$5E,$9C,$3D,$F3,$F6,$2F,$83,$F4,$4E,$11,$D9,$F0,$65
	.dc.b	$F0,$B2,$F1,$E5,$EA,$AD,$36,$75,$5E,$5C,$BF,$03,$87,$63,$BC,$76
	.dc.b	$9F,$CB,$3F,$E2,$CF,$9B,$FD,$1F,$5B,$F7,$6F,$E3,$F8,$FF,$A8,$9B
	.dc.b	$8F,$FA,$D3,$05,$FE,$A9,$B9,$FF,$A4,$79,$3F,$D2,$51,$8C,$BF,$D7
	.dc.b	$71,$8A,$FE,$02,$DA,$E1,$3D,$67,$A1,$23,$F5,$79,$7F,$9C,$7C,$7B
	.dc.b	$E3,$BD,$9A,$4C,$3F,$D5,$70,$E9,$C4,$7F,$9E,$A3,$83,$A9,$E7,$3D
	.dc.b	$3E,$22,$4F,$27,$C3,$DF,$FC,$AF,$83,$FD,$8B,$82,$77,$5D,$5D,$ED
	.dc.b	$E8,$F9,$5F,$93,$2D,$11,$FD,$1F,$0F,$F6,$7B,$78,$3F,$D5,$6F,$09
	.dc.b	$DD,$73,$EB,$8F,$FB,$22,$74,$48,$CC,$7F,$A6,$23,$C3,$C6,$B7,$8F
	.dc.b	$F3,$39,$3C,$17,$E5,$6D,$6E,$7F,$8D,$B6,$7F,$D5,$15,$C9,$B3,$F1
	.dc.b	$BE,$23,$E3,$FC,$2F,$C8,$A9,$5F,$EB,$98,$CB,$FE,$BB,$8C,$BF,$E9
	.dc.b	$68,$FE,$37,$F3,$9F,$80,$F9,$E5,$D3,$35,$FE,$6B,$9F,$C4,$BF,$E7
	.dc.b	$96,$BF,$E7,$76,$F5,$9A,$72,$8F,$F3,$6B,$F8,$EF,$31,$F3,$BF,$90
	.dc.b	$E3,$7B,$17,$E7,$66,$1F,$EF,$89,$CB,$67,$2F,$FB,$22,$13,$CE,$7F
	.dc.b	$A2,$F2,$BE,$75,$F3,$9F,$9E,$F5,$5F,$35,$4D,$CE,$B2,$FF,$51,$B5
	.dc.b	$FF,$55,$37,$A7,$EF,$43,$0E,$D8,$75,$F9,$DF,$E8,$FF,$3A,$A9,$F0
	.dc.b	$0F,$13,$FD,$F7,$C2,$F9,$4F,$88,$70,$AD,$CF,$FA,$AE,$FF,$C3,$51
	.dc.b	$FE,$A5,$6F,$21,$A4,$8E,$27,$FA,$BD,$F1,$2C,$3C,$4E,$4D,$F7,$3D
	.dc.b	$D7,$CE,$70,$AD,$D7,$EB,$3E,$23,$19,$7B,$70,$71,$BE,$03,$37,$FB
	.dc.b	$38,$C3,$E4,$39,$17,$C1,$61,$C2,$32,$79,$F0,$F4,$CE,$A3,$CF,$5F
	.dc.b	$A5,$74,$7C,$FF,$59,$F3,$7E,$2C,$9F,$07,$FE,$EA,$AF,$7B,$5E,$22
	.dc.b	$B8,$BE,$76,$BA,$6E,$9F,$11,$B9,$7C,$3F,$9D,$7C,$2F,$C2,$F1,$5F
	.dc.b	$37,$F9,$AC,$7F,$CB,$F1,$8E,$DF,$F2,$B3,$63,$FC,$B6,$D7,$73,$C7
	.dc.b	$2F,$B3,$39,$6C,$67,$1F,$A9,$E5,$7C,$7B,$F8,$BA,$A3,$17,$F8,$28
	.dc.b	$C7,$0F,$C7,$9C,$DF,$2F,$91,$C4,$CC,$3F,$0F,$5A,$7F,$98,$29,$F8
	.dc.b	$18,$FE,$BA,$7D,$6D,$7D,$7D,$FF,$02,$31,$1F,$96,$CF,$E3,$CF,$22
	.dc.b	$47,$83,$97,$AF,$CE,$71,$24,$76,$6B,$F8,$4A,$FA,$31,$C2,$3D,$3E
	.dc.b	$FD,$75,$82,$71,$B7,$C5,$CD,$78,$D7,$FD,$A1,$7B,$C3,$F5,$F8,$F6
	.dc.b	$7A,$98,$F6,$3F,$52,$FB,$5A,$FF,$B6,$5A,$A1,$18,$6D,$7E,$2E,$3F
	.dc.b	$67,$8F,$E0,$9B,$1F,$80,$F9,$8F,$D5,$7C,$03,$E7,$7F,$59,$D1,$FB
	.dc.b	$2C,$BF,$3E,$DC,$7F,$7B,$5B,$9F,$DB,$35,$FF,$56,$B5,$FF,$5B,$B5
	.dc.b	$FE,$1A,$D7,$F9,$03,$C6,$3A,$D7,$90,$F1,$FC,$4F,$9B,$7F,$2F,$D0
	.dc.b	$4C,$3C,$31,$97,$FD,$75,$87,$CF,$7C,$71,$FC,$92,$31,$1E,$48,$E0
	.dc.b	$FE,$F2,$38,$61,$FC,$6B,$1F,$E1,$0F,$C0,$AD,$B6,$BF,$8B,$4E,$B3
	.dc.b	$FE,$53,$B7,$F3,$79,$FE,$35,$F3,$CF,$3F,$F8,$C3,$63,$F8,$F4,$ED
	.dc.b	$FE,$1A,$D7,$FD,$BE,$D7,$AF,$87,$FC,$FA,$31,$1F,$D2,$5E,$58,$FF
	.dc.b	$32,$6E,$1C,$27,$C7,$E3,$FF,$29,$F9,$E7,$C8,$FF,$48,$6C,$FF,$B8
	.dc.b	$DA,$FF,$BA,$1A,$FF,$BC,$9F,$0A,$69,$31,$FE,$A7,$EF,$38,$FF,$A6
	.dc.b	$57,$F6,$3F,$47,$FA,$7C,$E7,$4F,$F2,$E4,$62,$B7,$97,$F8,$42,$FF
	.dc.b	$B3,$5E,$DE,$47,$F9,$98,$A7,$FB,$9C,$A7,$F6,$DF,$33,$AB,$B9,$E8
	.dc.b	$BE,$5D,$F3,$1F,$AE,$6E,$DF,$D9,$27,$FB,$2C,$61,$C1,$C6,$78,$7C
	.dc.b	$DC,$AC,$CF,$58,$C3,$C3,$E6,$B9,$6B,$FD,$A1,$B5,$FE,$03,$C1,$F3
	.dc.b	$1F,$89,$8F,$EC,$93,$AB,$FF,$04,$3E,$37,$CC,$D7,$FB,$DB,$5D,$E3
	.dc.b	$FC,$2A,$39,$60,$FA,$5F,$F3,$D5,$F6,$AF,$F0,$0C,$C9,$CD,$FF,$62
	.dc.b	$F9,$A7,$CD,$EB,$0F,$F0,$4C,$61,$FF,$06,$E7,$EA,$F1,$35,$7F,$E1
	.dc.b	$FA,$D3,$CE,$F2,$A3,$38,$FF,$84,$AF,$96,$1F,$E0,$59,$CB,$E0,$BF
	.dc.b	$E4,$B1,$D2,$FE,$C3,$5D,$4F,$D4,$8E,$30,$F7,$9F,$37,$FE,$75,$AE
	.dc.b	$D7,$E8,$9C,$FB,$BE,$6C,$3E,$1E,$5F,$E2,$06,$DF,$F8,$EE,$B9,$F9
	.dc.b	$1B,$F8,$29,$3E,$17,$CE,$47,$07,$F8,$D0,$7E,$5D,$FC,$D5,$F8,$11
	.dc.b	$F2,$9C,$F8,$9F,$33,$C6,$FE,$87,$E3,$53,$FA,$1E,$FF,$DF,$3C,$87
	.dc.b	$C5,$66,$66,$1F,$E0,$A7,$75,$FD,$45,$FE,$A3,$7D,$27,$5F,$2B,$D4
	.dc.b	$FB,$7E,$6F,$F8,$1F,$E7,$31,$3F,$D6,$E6,$E1,$E1,$9B,$C3,$FD,$4E
	.dc.b	$6D,$7F,$84,$A6,$D7,$F5,$DC,$55,$E2,$6D,$78,$89,$3F,$E2,$68,$FE
	.dc.b	$37,$1F,$DA,$FC,$3F,$2E,$8F,$9C,$F8,$91,$27,$E0,$3D,$48,$FE,$F3
	.dc.b	$3F,$F1,$17,$CE,$72,$39,$FC,$E5,$9F,$F0,$EC,$5C,$BE,$74,$DD,$7F
	.dc.b	$B8,$7A,$7F,$4C,$C5,$FF,$7B,$70,$FD,$17,$DC,$47,$F8,$4F,$BB,$FE
	.dc.b	$A0,$7F,$CC,$8D,$BF,$F1,$BF,$AF,$F7,$27,$3E,$AF,$EA,$AB,$58,$7F
	.dc.b	$8E,$7E,$09,$E9,$66,$BE,$CC,$62,$FF,$CE,$1F,$16,$F9,$1F,$D9,$B6
	.dc.b	$3F,$CD,$7F,$39,$89,$91,$FE,$84,$F5,$3F,$D6,$EF,$28,$FC,$2C,$63
	.dc.b	$5F,$F3,$4F,$C4,$FF,$A0,$6B,$FC,$FC,$D8,$FF,$28,$FC,$03,$E6,$3F
	.dc.b	$81,$CE,$76,$FF,$46,$BB,$AF,$E6,$8D,$7F,$D3,$2D,$7F,$D5,$8D,$7F
	.dc.b	$DD,$FF,$03,$F9,$8E,$0F,$F3,$8E,$3F,$D1,$7C,$4E,$33,$9E,$3F,$F5
	.dc.b	$0F,$9F,$FE,$BA,$F3,$DF,$36,$F9,$CC,$E1,$FE,$B9,$9C,$C6,$33,$FF
	.dc.b	$4E,$7B,$96,$73,$9F,$79,$FF,$62,$61,$CE,$31,$D7,$FE,$4D,$EF,$E1
	.dc.b	$39,$A6,$7F,$EC,$CA,$D5,$F0,$F8,$E7,$C8,$78,$65,$1F,$A8,$BE,$3A
	.dc.b	$FF,$AC,$E3,$77,$C7,$C2,$E1,$18,$AD,$8B,$E7,$3A,$C3,$6F,$BE,$5C
	.dc.b	$4F,$9C,$C3,$71,$F6,$F1,$FE,$DA,$BA,$9C,$2B,$67,$FB,$42,$7E,$EB
	.dc.b	$63,$EE,$B5,$FB,$AD,$79,$47,$CA,$F8,$FF,$F0,$83,$F3,$E3,$27,$B3
	.dc.b	$24,$EA,$31,$E2,$3E,$73,$E5,$3F,$93,$BF,$AA,$E5,$BD,$1A,$9D,$B6
	.dc.b	$32,$7D,$FF,$1D,$F0,$3F,$E8,$68,$C3,$70,$F0,$FC,$E7,$E2,$6B,$7F
	.dc.b	$07,$6B,$5C,$9F,$81,$F9,$27,$C6,$E6,$BF,$01,$18,$CB,$F1,$7F,$12
	.dc.b	$F9,$5F,$C2,$3F,$BB,$D3,$E1,$39,$9D,$F8,$7A,$AD,$B6,$32,$74,$8C
	.dc.b	$B4,$CD,$7C,$AF,$9C,$F9,$27,$CA,$F1,$AC,$FC,$A8,$B3,$F4,$26,$E9
	.dc.b	$F9,$DE,$27,$CD,$FD,$23,$E9,$FC,$E5,$C7,$E9,$C5,$C3,$EC,$CD,$BF
	.dc.b	$DC,$7C,$E7,$2B,$FD,$87,$82,$8C,$5F,$C2,$F9,$8F,$DB,$46,$2F,$8E
	.dc.b	$3F,$8F,$BE,$0D,$9E,$38,$30,$FE,$05,$18,$6C,$FD,$4E,$CE,$AD,$AF
	.dc.b	$C2,$7C,$E7,$D3,$C5,$E0,$F3,$C1,$F2,$9F,$EB,$EF,$6F,$7C,$7F,$AF
	.dc.b	$8C,$4F,$EA,$9B,$7C,$A3,$C9,$18,$BF,$37,$4B,$1F,$04,$77,$9D,$57
	.dc.b	$0F,$17,$FC,$48,$F5,$6E,$1D,$70,$E5,$8F,$8E,$73,$87,$36,$DF,$3C
	.dc.b	$BD,$39,$FE,$CF,$2E,$F3,$EE,$AD,$B6,$B6,$F8,$5E,$11,$A9,$F3,$E3
	.dc.b	$E2,$BC,$D4,$BE,$07,$F1,$A9,$D3,$8C,$BE,$33,$DC,$C6,$3F,$1B,$3F
	.dc.b	$8B,$1E,$93,$25,$9F,$85,$1A,$8E,$51,$D3,$1F,$69,$E6,$9D,$38,$CB
	.dc.b	$D4,$7C,$6C,$3C,$B5,$C6,$33,$8F,$AE,$3B,$E7,$F3,$E3,$19,$7C,$F7
	.dc.b	$E3,$CF,$17,$39,$79,$E7,$D9,$5C,$63,$D6,$7A,$23,$39,$F0,$8C,$1C
	.dc.b	$46,$3E,$69,$F0,$47,$2C,$BD,$D7,$D7,$1F,$14,$FC,$8B,$E0,$F3,$3A
	.dc.b	$9D,$9C,$B8,$AF,$8D,$8F,$B3,$0E,$F7,$D0,$8F,$A2,$BC,$59,$7C,$87
	.dc.b	$F0,$19,$FD,$08,$C5,$70,$9F,$5C,$E6,$FE,$06,$1E,$AC,$EE,$1C,$31
	.dc.b	$ED,$8F,$1C,$7C,$D3,$F6,$71,$E5,$1E,$2C,$7D,$0F,$BF,$1F,$5B,$E4
	.dc.b	$CF,$E3,$63,$EC,$24,$79,$23,$2F,$1C,$3B,$46,$23,$C3,$7D,$9A,$7C
	.dc.b	$FC,$3C,$78,$FD,$28,$FF,$6D,$63,$FC,$21,$B8,$7E,$F9,$B7,$FE,$D8
	.dc.b	$F8,$6F,$F4,$3F,$CE,$5F,$64,$D2,$57,$FB,$8B,$0F,$E9,$DC,$5F,$EE
	.dc.b	$06,$C6,$7A,$13,$96,$C6,$63,$F8,$24,$63,$3F,$D5,$37,$6F,$E3,$3F
	.dc.b	$38,$E8,$FE,$4D,$5A,$BE,$24,$C3,$F8,$05,$6D,$C4,$ED,$FE,$51,$87
	.dc.b	$57,$F3,$B3,$B7,$35,$C3,$0F,$E4,$D5,$BE,$25,$EF,$0E,$84,$AD,$DF
	.dc.b	$F8,$06,$75,$18,$C3,$A9,$2B,$F9,$64,$EA,$3F,$9C,$E1,$E5,$7F,$A1
	.dc.b	$B7,$2F,$9B,$5C,$32,$F8,$78,$75,$8F,$15,$F2,$C7,$C5,$5A,$8C,$1E
	.dc.b	$1A,$E0,$7C,$E8,$D4,$78,$27,$E1,$3C,$2B,$C0,$E0,$D1,$CC,$F9,$73
	.dc.b	$CA,$FC,$F7,$DA,$79,$3D,$DE,$51,$A7,$27,$23,$38,$7C,$13,$E9,$63
	.dc.b	$F1,$E3,$C3,$87,$C0,$6D,$F5,$7C,$86,$0E,$B9,$7D,$AC,$7C,$D8,$75
	.dc.b	$CB,$DF,$39,$AF,$0C,$63,$1F,$06,$5E,$79,$E8,$75,$AE,$0F,$82,$34
	.dc.b	$7B,$63,$39,$79,$E3,$38,$7B,$31,$F1,$E3,$E0,$C3,$F3,$71,$87,$A4
	.dc.b	$ED,$D5,$FE,$03,$0E,$B9,$7C,$F9,$F2,$E3,$E1,$CB,$D3,$87,$BE,$7E
	.dc.b	$C5,$F0,$7E,$89,$C2,$3B,$3E,$0C,$BE,$16,$5E,$3C,$BD,$55,$A6,$CE
	.dc.b	$AB,$CB,$97,$E0,$70,$EC,$77,$8E,$D3,$F9,$67,$FC,$6F,$F3,$9F,$A3
	.dc.b	$F1,$1D,$BF,$E2,$59,$FF,$3D,$37,$1F,$F5,$B5,$7E,$A6,$CF,$19,$D6
	.dc.b	$3F,$E3,$7C,$39,$E1,$FA,$3B,$DE,$77,$0E,$13,$FA,$7A,$DB,$E4,$9C
	.dc.b	$79,$2F,$76,$92,$7F,$D6,$9D,$75,$FF,$48,$4E,$DF,$F5,$C4,$F2,$9E
	.dc.b	$7E,$A4,$9F,$F5,$E3,$E1,$EE,$FE,$4F,$1C,$A3,$4D,$3F,$D2,$B9,$F9
	.dc.b	$FE,$5B,$DF,$1F,$EB,$97,$F9,$F6,$FF,$EC,$96,$ED,$F0,$DD,$EA,$69
	.dc.b	$AA,$E5,$CF,$D5,$87,$FB,$62,$35,$3C,$67,$33,$FE,$31,$AF,$F6,$2B
	.dc.b	$FE,$DE,$E0,$D9,$D7,$87,$FC,$C3,$DB,$FC,$55,$B3,$BB,$FA,$55,$BB
	.dc.b	$E7,$3F,$EB,$8F,$91,$F8,$D1,$23,$FD,$6F,$1F,$24,$C3,$FE,$BE,$8C
	.dc.b	$3F,$E5,$3F,$9C,$FB,$DE,$53,$E3,$A9,$FC,$66,$24,$FF,$B3,$63,$2F
	.dc.b	$12,$1F,$EC,$D9,$E4,$EC,$CE,$DF,$EF,$9E,$07,$B3,$F8,$D5,$F3,$74
	.dc.b	$C4,$4A,$FF,$47,$57,$19,$D4,$63,$2F,$CE,$5F,$48,$FF,$4F,$FC,$DB
	.dc.b	$E6,$FF,$35,$F3,$4F,$99,$A7,$E8,$67,$55,$C2,$B2,$F5,$1C,$8B,$E5
	.dc.b	$1B,$3F,$CF,$9F,$37,$FE,$99,$F7,$1E,$47,$EB,$FD,$57,$CA,$FE,$AD
	.dc.b	$B7,$C0,$FF,$55,$57,$07,$FD,$33,$7C,$EF,$95,$F1,$3F,$A6,$46,$67
	.dc.b	$67,$29,$DB,$E9,$C5,$87,$82,$3F,$D5,$15,$C5,$FD,$07,$C5,$73,$AF
	.dc.b	$77,$BD,$73,$3B,$EC,$55,$35,$12,$FC,$47,$FB,$16,$76,$E0,$7A,$2F
	.dc.b	$99,$F0,$7C,$27,$FB,$82,$78,$E4,$8C,$43,$7F,$8C,$F2,$AF,$55,$F3
	.dc.b	$8E,$35,$C2,$BE,$65,$F3,$F4,$37,$87,$C9,$EF,$7C,$C7,$E2,$71,$ED
	.dc.b	$C3,$FC,$9F,$BD,$F6,$9F,$15,$B3,$EE,$BE,$3D,$9C,$F0,$E5,$A7,$82
	.dc.b	$F7,$7D,$0F,$C5,$B8,$8F,$C0,$5F,$29,$F8,$3E,$27,$C8,$78,$6A,$8C
	.dc.b	$57,$B1,$B9,$FF,$3D,$5F,$25,$F6,$DC,$DF,$E6,$9B,$7C,$F8,$F5,$5E
	.dc.b	$9E,$AF,$77,$C3,$CC,$3D,$DA,$FE,$81,$F5,$69,$F2,$2B,$51,$E6,$D5
	.dc.b	$7C,$E3,$8F,$C3,$7C,$9F,$AD,$91,$82,$1C,$5C,$BD,$B1,$4F,$CD,$24
	.dc.b	$70,$DF,$E1,$7B,$58,$8E,$13,$67,$31,$F5,$FD,$5F,$C4,$C7,$D5,$FB
	.dc.b	$4F,$B8,$7E,$0E,$33,$C4,$AF,$F6,$0C,$61,$FB,$33,$C1,$E9,$3F,$AB
	.dc.b	$F7,$5F,$1F,$FA,$DE,$7F,$D4,$D7,$E8,$5A,$FE,$0C,$4F,$07,$F4,$CD
	.dc.b	$75,$1C,$31,$FD,$0B,$6F,$CD,$3C,$AF,$C4,$78,$7B,$24,$74,$8F,$36
	.dc.b	$7E,$5D,$E4,$F8,$23,$2F,$9E,$32,$F9,$F2,$FA,$6E,$CD,$10,$D4,$F2
	.dc.b	$7F,$C3,$07,$B9,$AF,$CD,$8E,$6E,$9A,$FF,$87,$F6,$FE,$31,$9F,$F0
	.dc.b	$0F,$9C,$79,$7F,$C6,$5B,$5F,$C3,$6B,$7A,$7C,$46,$C7,$DB,$9D,$BF
	.dc.b	$45,$B7,$FC,$32,$77,$3F,$C1,$3C,$2F,$3B,$4F,$4C,$7F,$2B,$7C,$D1
	.dc.b	$AF,$9C,$7C,$7F,$F2,$BF,$C5,$1F,$55,$4F,$E6,$13,$AB,$FE,$48,$F9
	.dc.b	$BD,$07,$C2,$FE,$3D,$BA,$F0,$CB,$F9,$C7,$BC,$F9,$EF,$E4,$98,$FF
	.dc.b	$B3,$BD,$2F,$C1,$CE,$77,$FC,$D7,$AB,$FD,$D2,$93,$97,$81,$F9,$F7
	.dc.b	$11,$B8,$FE,$E2,$68,$FF,$68,$94,$CB,$B5,$E9,$78,$2F,$F2,$BF,$32
	.dc.b	$F9,$BF,$EB,$D8,$FF,$5F,$D5,$F4,$5F,$AF,$DB,$FA,$35,$ED,$FE,$97
	.dc.b	$3A,$CB,$8C,$74,$F9,$FB,$1F,$DD,$B8,$DF,$55,$88,$60,$AD,$F9,$0B
	.dc.b	$DE,$1F,$D5,$7E,$7F,$C1,$87,$E6,$1A,$FF,$9C,$5A,$FE,$C1,$AF,$F0
	.dc.b	$77,$8E,$9E,$EF,$9F,$FF,$33,$57,$5D,$BC,$F2,$9C,$DF,$F0,$3B,$E3
	.dc.b	$7F,$E0,$49,$CD,$7F,$60,$9C,$D7,$F6,$39,$CD,$7F,$02,$F9,$D7,$CF
	.dc.b	$7F,$66,$AC,$35,$FE,$39,$1A,$9F,$D9,$B6,$3F,$6E,$DC,$3E,$B7,$CF
	.dc.b	$BD,$BF,$99,$39,$BF,$E4,$DC,$2A,$D4,$7F,$7F,$9D,$60,$AD,$5F,$2C
	.dc.b	$7E,$CC,$62,$F8,$46,$30,$E7,$7F,$C9,$F0,$E7,$7B,$8C,$7C,$0F,$E0
	.dc.b	$B0,$7C,$FF,$F3,$27,$C5,$19,$8C,$4F,$09,$FF,$0B,$9C,$3C,$5F,$6D
	.dc.b	$6D,$B1,$AA,$E3,$A7,$F8,$4B,$1F,$A3,$87,$F9,$02,$73,$5F,$D2,$7C
	.dc.b	$E7,$95,$23,$FC,$3B,$E6,$BE,$BF,$F3,$35,$AA,$FC,$AC,$E6,$BF,$AC
	.dc.b	$46,$2B,$6F,$3C,$3F,$B8,$46,$3E,$56,$57,$F8,$2B,$E8,$2D,$7F,$78
	.dc.b	$B5,$63,$4F,$F0,$55,$5B,$F4,$56,$9F,$3F,$9B,$9C,$3F,$C3,$2D,$E8
	.dc.b	$FF,$6A,$B7,$3F,$F1,$E5,$7F,$6F,$F1,$BD,$7E,$F7,$A1,$B1,$93,$E1
	.dc.b	$C6,$5A,$7F,$88,$BE,$7F,$F6,$FE,$B1,$BF,$3F,$86,$4E,$CF,$F1,$4C
	.dc.b	$63,$3F,$81,$F3,$BF,$33,$97,$E4,$B6,$BF,$C0,$8D,$9D,$C7,$D6,$9F
	.dc.b	$E1,$5F,$03,$FE,$98,$8F,$F2,$AF,$C1,$FF,$87,$9F,$F2,$E4,$66,$BF
	.dc.b	$C6,$F3,$97,$FB,$F4,$67,$8B,$FC,$A7,$39,$7F,$CB,$71,$9A,$FF,$4D
	.dc.b	$FA,$7F,$C6,$7E,$29,$F2,$5F,$C6,$3D,$57,$D2,$66,$30,$DB,$FF,$34
	.dc.b	$CE,$9F,$F3,$85,$72,$73,$3F,$C9,$7E,$2B,$FC,$FB,$1F,$E7,$0C,$5B
	.dc.b	$7F,$A4,$FC,$CF,$E0,$1B,$7F,$88,$EF,$6F,$F9,$52,$B7,$7F,$EA,$78
	.dc.b	$4E,$B5,$FF,$38,$D6,$A7,$FA,$2F,$87,$FE,$18,$AD,$5F,$FA,$9F,$1F
	.dc.b	$04,$E7,$D8,$FE,$6F,$87,$1C,$BC,$78,$F6,$9C,$DF,$08,$79,$CF,$9C
	.dc.b	$7C,$F6,$7B,$3F,$C9,$7D,$BB,$AD,$4E,$78,$BF,$59,$9F,$C0,$9C,$ED
	.dc.b	$FE,$AB,$9C,$ED,$FD,$B3,$2E,$B3,$9C,$7B,$7C,$73,$E7,$6F,$83,$86
	.dc.b	$D7,$FA,$D6,$74,$D9,$D3,$4F,$99,$C4,$F9,$DF,$57,$C7,$7F,$80,$67
	.dc.b	$31,$8F,$5B,$E1,$71,$7D,$28,$C4,$E8,$FF,$62,$57,$19,$FA,$67,$DD
	.dc.b	$9D,$B8,$39,$4F,$0A,$F9,$5D,$EF,$9D,$BE,$CE,$1B,$5F,$ED,$19,$D3
	.dc.b	$67,$4D,$3F,$DD,$DE,$47,$F9,$8B,$E0,$7E,$B7,$8C,$F9,$CF,$B9,$39
	.dc.b	$AF,$F6,$E4,$E5,$E3,$F0,$2B,$E3,$7C,$23,$13,$C2,$3F,$05,$F1,$6F
	.dc.b	$94,$BE,$1E,$1B,$5F,$EE,$99,$D3,$67,$4D,$35,$3C,$37,$F9,$7D,$BF
	.dc.b	$ED,$E8,$C7,$67,$EA,$A3,$FD,$75,$F3,$9F,$28,$F9,$7E,$25,$C3,$E4
	.dc.b	$45,$C3,$9F,$0A,$B1,$87,$38,$7C,$67,$D4,$3E,$AF,$CE,$5C,$7E,$94
	.dc.b	$5C,$39,$CD,$BF,$CF,$7C,$E7,$2B,$F9,$BD,$3F,$A5,$8C,$5F,$B3,$8F
	.dc.b	$F7,$B5,$FE,$A5,$BE,$38,$FD,$8C,$38,$36,$75,$DF,$FE,$52,$C7,$9D
	.dc.b	$FC,$3C,$39,$E3,$F9,$2F,$05,$5A,$BF,$F7,$96,$5F,$82,$F9,$CF,$E3
	.dc.b	$57,$F9,$5F,$9A,$FE,$DB,$EB,$FE,$2A,$73,$9F,$F0,$69,$CD,$6F,$C2
	.dc.b	$FF,$09,$36,$F8,$72,$FE,$A2,$B7,$7C,$6F,$6D,$AD,$ED,$FC,$7E,$B7
	.dc.b	$9F,$FA,$C9,$B8,$FF,$B2,$5B,$D4,$CF,$CB,$5B,$CB,$AB,$71,$E9,$18
	.dc.b	$C1,$7C,$73,$F3,$4E,$65,$8F,$81,$B1,$C6,$B8,$65,$D3,$09,$5F,$04
	.dc.b	$FE,$03,$3A,$71,$9F,$BD,$B9,$7F,$86,$F3,$F4,$C7,$B7,$1F,$16,$5E
	.dc.b	$13,$DF,$3A,$9F,$1E,$3E,$0C,$3A,$1F,$D3,$31,$F0,$38,$C7,$E1,$57
	.dc.b	$80,$F8,$37,$CF,$0E,$D8,$F8,$0F,$74,$65,$C6,$3D,$EF,$D1,$9F,$C8
	.dc.b	$8C,$61,$E9,$7E,$DB,$DC,$F3,$9E,$1C,$BC,$B5,$E4,$CB,$CD,$3C,$4F
	.dc.b	$93,$8F,$86,$FA,$C7,$29,$F7,$4E,$63,$B6,$5F,$1B,$1F,$24,$63,$2F
	.dc.b	$5C,$62,$FB,$CF,$3C,$7D,$79,$F6,$C3,$E1,$E5,$EA,$BD,$C6,$2F,$CB
	.dc.b	$87,$3C,$3D,$44,$8F,$9B,$19,$7D,$95,$EE,$CF,$E0,$C6,$2B,$83,$E3
	.dc.b	$38,$E3,$E3,$CB,$E8,$C7,$82,$31,$97,$D0,$BE,$73,$C4,$99,$7C,$3A
	.dc.b	$DC,$7D,$1C,$FE,$46,$5E,$CC,$7E,$2D,$6A,$3C,$D3,$AC,$BC,$B8,$73
	.dc.b	$C3,$E6,$64,$61,$F8,$E2,$9F,$80,$2C,$7F,$08,$73,$F0,$6F,$9C,$7C
	.dc.b	$E7,$09,$FF,$74,$46,$1C,$4E,$E7,$51,$FE,$F0,$9D,$3D,$BA,$7F,$C2
	.dc.b	$EF,$3B,$FE,$A9,$C5,$C1,$B1,$99,$DB,$C6,$7F,$89,$47,$2D,$3F,$90
	.dc.b	$37,$3F,$F4,$97,$CE,$7F,$49,$BD,$DC,$7F,$95,$56,$DF,$E1,$F3,$B7
	.dc.b	$33,$BB,$EA,$FF,$78,$9D,$46,$27,$64,$7F,$7D,$1A,$BF,$E4,$D7,$DE
	.dc.b	$BF,$8C,$B7,$0F,$7D,$F0,$E8,$FA,$EF,$52,$5F,$2C,$7C,$53,$A8,$C6
	.dc.b	$1E,$16,$D7,$38,$F5,$46,$08,$FC,$09,$E4,$FF,$39,$C3,$C8,$4B,$EF
	.dc.b	$1C,$8D,$37,$85,$97,$BE,$72,$78,$CB,$8F,$8D,$C1,$F0,$AB,$94,$F1
	.dc.b	$8C,$1A,$CB,$E2,$E5,$F7,$32,$F3,$CE,$B0,$F8,$AF,$69,$D3,$E7,$34
	.dc.b	$FC,$0C,$7C,$F5,$CA,$FA,$D7,$12,$4E,$9E,$51,$A7,$99,$F2,$4C,$E5
	.dc.b	$F1,$DF,$E9,$B8,$FA,$E7,$81,$27,$F2,$15,$D6,$BD,$D1,$C7,$1F,$8C
	.dc.b	$7B,$23,$B9,$E0,$9E,$11,$9A,$E9,$97,$D6,$AE,$6F,$CC,$C7,$DD,$87
	.dc.b	$5C,$BE,$16,$5E,$58,$C5,$FC,$08,$C5,$FB,$0E,$98,$F7,$AD,$E5,$E2
	.dc.b	$C3,$F1,$78,$79,$B2,$F9,$73,$98,$F9,$86,$6B,$E0,$E3,$DA,$3C,$E7
	.dc.b	$C8,$CB,$CF,$97,$86,$30,$5C,$BE,$FE,$1D,$63,$A4,$FB,$DD,$65,$EE
	.dc.b	$BE,$71,$DA,$FE,$6C,$FA,$32,$F9,$33,$D8,$D4,$71,$79,$9F,$16,$B9
	.dc.b	$D7,$C3,$BF,$8B,$5C,$F2,$F8,$B9,$7B,$6B,$58,$75,$69,$EA,$8D,$57
	.dc.b	$1B,$E7,$5F,$23,$1F,$2D,$FA,$23,$18,$76,$C7,$C7,$87,$19,$EC,$DA
	.dc.b	$ED,$3A,$9E,$38,$76,$7E,$49,$FB,$7C,$BE,$66,$3E,$0C,$3E,$14,$FC
	.dc.b	$43,$E1,$1F,$10,$FC,$A6,$5E,$A7,$A9,$E2,$2E,$3E,$27,$85,$7E,$1A
	.dc.b	$78,$47,$8A,$7D,$CD,$AF,$0C,$7C,$86,$9E,$D2,$C7,$66,$FC,$1B,$CC
	.dc.b	$FF,$65,$4F,$F5,$98,$C7,$89,$FD,$59,$B1,$99,$DB,$FD,$9A,$32,$BC
	.dc.b	$E7,$27,$F6,$D9,$CC,$62,$51,$89,$FE,$E5,$DC,$FD,$15,$E9,$7F,$74
	.dc.b	$C3,$84,$E6,$31,$5C,$2B,$78,$7F,$7B,$BD,$C6,$23,$C9,$3D,$BB,$1F
	.dc.b	$27,$CA,$58,$F5,$79,$57,$5C,$75,$39,$9E,$67,$0B,$E6,$F0,$3F,$CE
	.dc.b	$BF,$77,$FE,$FF,$7F,$C0,$9C,$48,$C4,$7F,$83,$A3,$1E,$6F,$F8,$11
	.dc.b	$B1,$C2,$77,$A7,$C9,$9C,$CF,$89,$B1,$F1,$AD,$F3,$7E,$D9,$F0,$7C
	.dc.b	$E7,$5C,$0D,$26,$1D,$DE,$11,$A8,$F6,$4F,$27,$DD,$93,$07,$D1,$70
	.dc.b	$C2,$57,$0C,$39,$3F,$22,$F8,$56,$A3,$D5,$3A,$CB,$DC,$DD,$7E,$C5
	.dc.b	$79,$3D,$4F,$51,$2F,$E3,$35,$E6,$ED,$F3,$E4,$FA,$1F,$C0,$43,$F0
	.dc.b	$27,$BF,$F8,$3E,$F7,$C9,$BE,$5B,$0D,$BF,$BC,$71,$AE,$77,$CA,$3F
	.dc.b	$C7,$AD,$7F,$C8,$8D,$7F,$C9,$AD,$7F,$CA,$2E,$F6,$FD,$34,$75,$3B
	.dc.b	$C6,$75,$FF,$2C,$62,$EB,$7B,$DE,$2F,$80,$F3,$E5,$31,$E3,$19,$7C
	.dc.b	$AF,$F9,$49,$F8,$4D,$C4,$D3,$84,$7E,$80,$D9,$F9,$FC,$3C,$5F,$45
	.dc.b	$AC,$FD,$37,$CA,$BC,$26,$F2,$57,$C8,$F8,$0F,$9B,$D5,$2B,$DF,$19
	.dc.b	$7E,$5B,$E5,$23,$83,$64,$3F,$6B,$8F,$C5,$F3,$7E,$FF,$CA,$BE,$4F
	.dc.b	$E0,$36,$BF,$CE,$6D,$7F,$CF,$4D,$7F,$D0,$7E,$63,$D2,$F9,$0D,$8F
	.dc.b	$F3,$DB,$5F,$F4,$13,$5F,$F4,$3F,$B3,$FC,$95,$79,$0F,$93,$FF,$48
	.dc.b	$F9,$AA,$D4,$E6,$BF,$D2,$D3,$9E,$2F,$F4,$C4,$E6,$53,$97,$AD,$FF
	.dc.b	$A4,$F8,$76,$D8,$C9,$FE,$9E,$8C,$B4,$C8,$FB,$07,$C2,$FE,$52,$3F
	.dc.b	$D4,$F1,$8E,$8F,$F4,$DB,$63,$FD,$44,$D7,$FD,$57,$F1,$EF,$90,$E5
	.dc.b	$1F,$EB,$18,$C5,$7F,$A9,$1B,$1F,$EA,$B6,$BF,$EB,$4F,$AF,$7C,$77
	.dc.b	$F4,$BD,$BF,$4F,$18,$BF,$CA,$47,$EF,$1F,$DB,$C6,$6B,$F3,$4D,$7A
	.dc.b	$57,$91,$46,$5F,$F6,$2B,$5F,$F6,$43,$CE,$BF,$AE,$72,$BE,$45,$F3
	.dc.b	$3E,$5C,$78,$E3,$71,$FF,$63,$D6,$A7,$CB,$F4,$79,$9F,$0B,$BC,$3F
	.dc.b	$B6,$75,$67,$34,$D2,$5F,$C0,$9B,$7E,$E9,$F6,$57,$0A,$FF,$68,$B6
	.dc.b	$7F,$DB,$0D,$7E,$74,$F8,$70,$FE,$07,$5E,$63,$F8,$03,$89,$F9,$7E
	.dc.b	$67,$94,$F4,$69,$FC,$BE,$BE,$8C,$65,$F0,$BE,$42,$38,$3A,$F9,$7B
	.dc.b	$AF,$DD,$DF,$F5,$C9,$FE,$A7,$F3,$CF,$4B,$E6,$79,$A8,$C4,$FF,$40
	.dc.b	$F9,$CC,$C6,$30,$FE,$D0,$D7,$FE,$00,$D7,$1B,$FC,$AB,$6F,$FB,$CD
	.dc.b	$6E,$82,$7C,$1D,$5D,$8F,$BD,$E6,$FF,$65,$F4,$B1,$87,$A2,$6D,$FF
	.dc.b	$63,$F2,$FF,$C6,$53,$92,$47,$9A,$79,$CE,$A3,$13,$A7,$D3,$8F,$F0
	.dc.b	$98,$F3,$D7,$AA,$31,$3F,$E0,$F5,$F2,$1F,$10,$EB,$F1,$76,$FE,$6C
	.dc.b	$71,$23,$F0,$9F,$DC,$E4,$FB,$BF,$E0,$1C,$5F,$E3,$66,$D7,$F8,$BF
	.dc.b	$AB,$E1,$C6,$32,$FF,$17,$B6,$F9,$76,$7B,$AB,$6F,$E9,$FD,$57,$E9
	.dc.b	$EF,$85,$C3,$72,$FF,$09,$E3,$D6,$B7,$5A,$3F,$87,$B8,$9F,$F2,$F5
	.dc.b	$71,$CB,$FC,$B4,$D7,$EE,$9F,$7B,$DA,$78,$AF,$7B,$F0,$3E,$1F,$E4
	.dc.b	$6B,$53,$97,$FB,$A7,$3F,$E1,$B9,$DE,$FB,$E0,$3F,$CC,$58,$71,$B6
	.dc.b	$5D,$B2,$EB,$8F,$8A,$31,$3D,$0E,$F3,$A7,$F2,$58,$75,$D3,$11,$D3
	.dc.b	$3F,$26,$3D,$F0,$E6,$D8,$CE,$3E,$7C,$39,$5F,$73,$96,$3D,$30,$E3
	.dc.b	$5E,$5B,$E3,$5E,$2C,$BD,$B3,$9B,$E1,$5A,$AE,$67,$AF,$3F,$8B,$7C
	.dc.b	$23,$DA,$78,$70,$EE,$E3,$2F,$A1,$5C,$33,$EB,$8C,$AE,$18,$F9,$2B
	.dc.b	$51,$89,$D5,$FC,$17,$E5,$4E,$A3,$13,$B2,$3E,$0C,$FE,$2B,$A2,$67
	.dc.b	$EF,$9F,$64,$EA,$31,$3A,$9F,$54,$7C,$FC,$FE,$04,$7C,$49,$D3,$67
	.dc.b	$59,$7A,$1E,$64,$8F,$A1,$19,$7A,$93,$0F,$64,$EE,$3A,$19,$C3,$CC
	.dc.b	$F0,$32,$6A,$78,$B9,$CF,$E9,$B9,$9E,$2F,$08,$C9,$F6,$27,$8C,$67
	.dc.b	$3F,$C1,$4E,$6B,$8B,$C2,$32,$7D,$98,$CB,$8C,$FD,$D7,$D2,$B8,$46
	.dc.b	$2B,$6F,$C3,$BF,$36,$7F,$3A,$B7,$1F,$73,$1F,$9A,$D3,$DD,$8F,$86
	.dc.b	$B8,$3E,$18,$D6,$3E,$18,$C6,$7F,$76,$31,$8F,$87,$1F,$76,$1C,$A3
	.dc.b	$17,$D7,$3B,$1E,$99,$D6,$3E,$33,$D3,$3A,$7C,$39,$FB,$9D,$61,$DA
	.dc.b	$75,$97,$8E,$B4,$FE,$32,$75,$5C,$31,$F4,$E7,$F0,$AB,$51,$89,$FC
	.dc.b	$69,$BC,$BC,$D8,$F6,$BF,$C9,$67,$F0,$2B,$47,$E8,$B3,$F5,$E3,$D3
	.dc.b	$0F,$D7,$E7,$E4,$C3,$F2,$99,$F9,$2B,$58,$7B,$B0,$E3,$5A,$CF,$E4
	.dc.b	$4F,$6A,$F7,$56,$A3,$EC,$E3,$DA,$31,$1F,$91,$8C,$67,$E9,$8F,$D1
	.dc.b	$4E,$B1,$FC,$2D,$FE,$FE,$31,$87,$C3,$CF,$DD,$3E,$63,$58,$73,$C3
	.dc.b	$EB,$9E,$5C,$BE,$2E,$7E,$F8,$F2,$9F,$16,$78,$65,$F0,$6B,$DD,$9F
	.dc.b	$CE,$AE,$47,$9A,$B7,$97,$C2,$C7,$C5,$7D,$B3,$F3,$E1,$C6,$7E,$5E
	.dc.b	$1F,$D7,$0F,$E6,$25,$C7,$FA,$6B,$70,$FE,$B4,$EF,$C7,$7C,$03,$C3
	.dc.b	$E5,$3F,$E9,$48,$C3,$89,$DC,$EA,$3F,$D2,$B3,$A7,$C1,$F0,$1F,$EA
	.dc.b	$27,$9E,$08,$C4,$6B,$CA,$E5,$39,$6C,$F1,$CB,$FA,$E3,$75,$FE,$D0
	.dc.b	$DB,$FF,$1B,$7B,$EF,$80,$FE,$FB,$7C,$30,$FE,$B9,$7B,$BF,$EE,$6F
	.dc.b	$1E,$26,$1C,$67,$75,$1F,$EA,$D5,$B9,$D4,$62,$75,$7D,$E3,$FB,$04
	.dc.b	$EA,$30,$4C,$7D,$35,$FD,$52,$75,$18,$AE,$84,$AF,$EC,$73,$A8,$FF
	.dc.b	$0C,$5F,$AF,$2F,$91,$1B,$BF,$F0,$A5,$6E,$3D,$F8,$79,$A7,$B6,$1D
	.dc.b	$32,$F8,$F1,$82,$57,$6A,$F3,$4E,$A3,$07,$72,$D7,$37,$B1,$8A,$E7
	.dc.b	$97,$CE,$AE,$53,$E4,$39,$16,$B9,$38,$C7,$DB,$1F,$26,$3E,$FC,$F2
	.dc.b	$9C,$DF,$A2,$79,$3F,$0A,$75,$97,$C7,$26,$3E,$CA,$F3,$9C,$0F,$41
	.dc.b	$D8,$CD,$F6,$73,$8F,$AD,$CD,$74,$7F,$03,$7E,$0B,$C9,$DE,$3D,$93
	.dc.b	$BC,$7D,$E7,$B6,$3C,$47,$8E,$FB,$3C,$2B,$E6,$C6,$27,$86,$1E,$5A
	.dc.b	$E1,$97,$E0,$63,$04,$C3,$F3,$58,$F9,$6B,$A1,$E8,$3C,$E7,$DD,$C3
	.dc.b	$B6,$5F,$36,$F9,$65,$EB,$CB,$C3,$8F,$AE,$BE,$86,$1C,$A7,$EC,$61
	.dc.b	$CA,$3E,$69,$EE,$C3,$E6,$46,$63,$17,$CD,$FA,$47,$9A,$B9,$E5,$ED
	.dc.b	$9D,$4F,$C0,$CB,$D9,$5E,$4C,$BC,$D9,$7A,$6F,$A6,$3D,$F2,$F5,$E3
	.dc.b	$EF,$CB,$A6,$1E,$9C,$3D,$73,$E1,$7F,$37,$3A,$8C,$1E,$6C,$7C,$0F
	.dc.b	$3C,$BC,$75,$E1,$BF,$1C,$FC,$2C,$3C,$37,$F8,$2A,$D4,$F1,$3C,$47
	.dc.b	$E4,$30,$F0,$4F,$63,$84,$73,$CB,$DB,$3A,$AE,$75,$AA,$F9,$B8,$7E
	.dc.b	$2A,$B5,$1C,$63,$CA,$E8,$F1,$1E,$E3,$E7,$1E,$42,$99,$C3,$ED,$CF
	.dc.b	$48,$E5,$5E,$19,$F7,$DF,$CA,$C7,$E2,$D7,$84,$FC,$91,$CF,$1F,$CB
	.dc.b	$BD,$EB,$AC,$F0,$8E,$B5,$F5,$AF,$E2,$3D,$63,$FD,$55,$C8,$F0,$1D
	.dc.b	$3F,$D6,$63,$0D,$8C,$CF,$F6,$29,$CB,$63,$2B,$AD,$CA,$85,$70,$75
	.dc.b	$19,$8C,$4F,$09,$75,$BD,$F7,$8D,$FE,$D6,$AD,$4E,$63,$18,$77,$8C
	.dc.b	$5F,$86,$B7,$7C,$A3,$15,$BF,$09,$F3,$3C,$A7,$C3,$5B,$7C,$E6,$31
	.dc.b	$D6,$3D,$A3,$2A,$F9,$C7,$61,$C8,$FD,$8F,$FD,$F9,$1F,$E0,$C8,$C7
	.dc.b	$77,$F8,$25,$B1,$99,$DB,$FE,$1B,$8C,$BE,$46,$E9,$F2,$5B,$A7,$B7
	.dc.b	$E0,$BD,$91,$E1,$F1,$3B,$60,$69,$30,$F0,$3C,$23,$51,$EB,$9E,$4F
	.dc.b	$B7,$35,$FD,$98,$F2,$57,$06,$DF,$33,$E4,$CE,$F1,$F4,$56,$DB,$3B
	.dc.b	$AE,$D5,$BC,$7F,$19,$8F,$CB,$F6,$7D,$64,$AF,$90,$D7,$C6,$FA,$CF
	.dc.b	$0F,$DD,$FD,$F8,$7B,$2F,$96,$FC,$17,$73,$DD,$7C,$2E,$1B,$7F,$74
	.dc.b	$E3,$5C,$EF,$94,$7F,$90,$DA,$FF,$92,$9A,$FF,$95,$5A,$FE,$9C,$E6
	.dc.b	$75,$78,$ED,$F3,$5B,$A7,$EB,$71,$75,$3C,$9F,$16,$BE,$7F,$7E,$65
	.dc.b	$CE,$32,$F9,$5F,$F2,$C3,$F0,$9B,$89,$A7,$08,$FD,$19,$B3,$F3,$D8
	.dc.b	$78,$BD,$FD,$67,$E7,$AB,$9F,$9F,$35,$7C,$7F,$7F,$E6,$54,$C7,$E0
	.dc.b	$46,$5F,$94,$F9,$08,$E0,$D9,$0F,$DB,$E3,$F1,$3D,$FF,$BD,$F1,$4F
	.dc.b	$3F,$DE,$DA,$FF,$3C,$35,$FF,$3F,$B5,$FF,$45,$7B,$AF,$7B,$E3,$B6
	.dc.b	$3F,$D0,$0D,$7F,$D0,$ED,$7F,$D1,$9E,$97,$F2,$55,$E4,$BD,$DF,$C0
	.dc.b	$79,$8A,$D4,$E6,$BF,$95,$4E,$78,$BF,$96,$4E,$65,$39,$7A,$DF,$FA
	.dc.b	$63,$87,$6D,$8C,$9F,$EA,$58,$CB,$4C,$8F,$73,$FA,$2F,$1F,$FA,$75
	.dc.b	$B1,$FE,$A4,$6B,$FE,$AF,$6B,$F9,$BF,$98,$7C,$D7,$F0,$58,$C3,$5F
	.dc.b	$E8,$D1,$AA,$FF,$55,$36,$3F,$D6,$3E,$DF,$F4,$29,$FD,$83,$63,$EF
	.dc.b	$E8,$F7,$3F,$A5,$3F,$C4,$E7,$4D,$F1,$BC,$14,$9E,$4F,$FB,$1D,$AF
	.dc.b	$EC,$DA,$FF,$5B,$79,$DF,$EC,$7D,$A7,$3F,$82,$38,$5D,$BF,$F6,$55
	.dc.b	$6A,$7D,$3E,$E6,$67,$C4,$EF,$A3,$FB,$57,$C1,$E7,$24,$D2,$5F,$EC
	.dc.b	$33,$56,$30,$FD,$AD,$F0,$AF,$F6,$A3,$63,$E5,$DF,$C2,$8C,$47,$C1
	.dc.b	$8F,$16,$2F,$77,$A6,$3F,$DD,$72,$4E,$71,$FE,$45,$F8,$87,$E7,$1B
	.dc.b	$CB,$EB,$BF,$D8,$46,$09,$97,$FB,$B2,$B8,$36,$BE,$AC,$62,$3F,$DF
	.dc.b	$4E,$AB,$C5,$6F,$7B,$E5,$4E,$70,$FA,$91,$F4,$7C,$6F,$EE,$1F,$0D
	.dc.b	$7C,$A3,$F5,$7C,$BF,$EB,$3F,$2F,$98,$98,$7F,$71,$7F,$2D,$7C,$A3
	.dc.b	$07,$F7,$5F,$4F,$A1,$CF,$CC,$7B,$58,$9F,$ED,$19,$7E,$A3,$6F,$EB
	.dc.b	$F7,$BA,$D6,$FF,$DA,$22,$CF,$F8,$7B,$0E,$78,$72,$9D,$D7,$F4,$7C
	.dc.b	$7F,$C1,$91,$E6,$C7,$F6,$33,$FE,$10,$6C,$FF,$84,$6F,$E1,$1E,$7F
	.dc.b	$62,$E1,$E6,$8E,$44,$7F,$C2,$4F,$ED,$F5,$7B,$FF,$CE,$A7,$FC,$04
	.dc.b	$DC,$7F,$C4,$74,$F7,$7E,$74,$E7,$4F,$F1,$CB,$6F,$FC,$70,$DB,$FF
	.dc.b	$20,$DF,$19,$EB,$E4,$FE,$AF,$D6,$7B,$EF,$98,$C3,$6B,$67,$F8,$C6
	.dc.b	$77,$3A,$3F,$88,$38,$8F,$F2,$E4,$F1,$C3,$FB,$91,$FE,$58,$70,$E3
	.dc.b	$DC,$FB,$BE,$FF,$F9,$77,$C5,$7B,$9F,$DE,$67,$35,$FD,$D2,$3F,$C2
	.dc.b	$FF,$01,$EB,$8C,$1F,$E6,$59,$DD,$72,$8C,$73,$30,$ED,$87,$58,$C4
	.dc.b	$F2,$E4,$63,$D7,$0E,$52,$C7,$AE,$1C,$71,$E7,$3D,$9B,$87,$2E,$2F
	.dc.b	$F3,$96,$1C,$B0,$E3,$7B,$3D,$27,$5C,$BC,$73,$E9,$BE,$5A,$76,$AD
	.dc.b	$63,$70,$E7,$18,$3D,$59,$FA,$73,$F3,$47,$69,$D9,$E2,$C3,$AC,$67
	.dc.b	$2F,$0D,$70,$CB,$D1,$18,$9F,$F3,$8E,$1D,$5F,$54,$EB,$1F,$25,$6F
	.dc.b	$0E,$B3,$9A,$ED,$5A,$3E,$3D,$F3,$8F,$54,$EC,$98,$7E,$03,$3F,$96
	.dc.b	$F9,$49,$3E,$27,$59,$F5,$8E,$E4,$AF,$99,$7D,$F3,$F8,$79,$FB,$6B
	.dc.b	$C3,$87,$27,$83,$D9,$C9,$F0,$B1,$F1,$BE,$12,$1A,$CF,$E0,$CF,$1B
	.dc.b	$EF,$9F,$C0,$BE,$37,$D0,$93,$F5,$71,$F3,$63,$E5,$8E,$33,$F7,$AB
	.dc.b	$8C,$63,$2F,$8F,$3E,$5C,$FE,$14,$62,$3C,$98,$74,$C7,$39,$7C,$AC
	.dc.b	$FC,$F5,$EA,$C7,$A4,$FC,$3C,$39,$46,$27,$B1,$F1,$8E,$19,$FE,$07
	.dc.b	$1E,$D1,$93,$C3,$9F,$A2,$7D,$84,$CB,$E2,$67,$E4,$AE,$71,$87,$96
	.dc.b	$1E,$1C,$FC,$F1,$F2,$2F,$93,$F1,$6F,$93,$9B,$E5,$97,$AA,$B6,$FE
	.dc.b	$13,$1F,$24,$F0,$C7,$E9,$46,$2F,$C3,$7E,$D8,$C5,$F3,$BE,$B9,$FD
	.dc.b	$FA,$EA,$FC,$93,$11,$ED,$CB,$D3,$8F,$CB,$C7,$E2,$E5,$E1,$C3,$8C
	.dc.b	$7C,$EC,$7B,$3F,$3E,$F8,$CE,$A3,$89,$E6,$C3,$B6,$7F,$4A,$F8,$9E
	.dc.b	$FC,$3B,$4E,$B3,$F5,$E7,$E6,$C3,$E5,$1C,$CF,$06,$1C,$E3,$58,$73
	.dc.b	$9D,$3F,$92,$CB,$CB,$87,$69,$E7,$3E,$2A,$FC,$85,$F3,$9F,$BE,$6A
	.dc.b	$3A,$9A,$C3,$CB,$1A,$AF,$01,$CD,$FD,$96,$1E,$07,$C7,$5F,$45,$E0
	.dc.b	$F3,$C7,$C9,$5E,$7C,$3C,$55,$F8,$73,$C4,$70,$3A,$BA,$CF,$F2,$4D
	.dc.b	$CF,$EB,$E3,$E1,$BF,$E4,$17,$FD,$7E,$7F,$A7,$B7,$3F,$EC,$AF,$F4
	.dc.b	$33,$FB,$3F,$92,$F8,$07,$7E,$A7,$FD,$2B,$18,$78,$CE,$E7,$4D,$8E
	.dc.b	$B9,$B9,$FF,$D4,$F1,$86,$BF,$DC,$DA,$FF,$71,$6B,$FD,$B5,$AF,$F6
	.dc.b	$36,$C7,$5E,$FF,$33,$67,$B4,$62,$76,$47,$04,$8D,$90,$F2,$39,$69
	.dc.b	$93,$CC,$E5,$A6,$4F,$43,$96,$99,$3D,$4E,$5A,$66,$7D,$4F,$80,$87
	.dc.b	$62,$9D,$0A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA
	.dc.b	$70,$47,$04,$9F,$A1,$1E,$D6,$BE,$66,$BE,$16,$BD,$9A,$F5,$6C,$75
	.dc.b	$9F,$94,$D9,$ED,$18,$9D,$91,$C1,$23,$64,$3C,$8E,$5A,$64,$F3,$39
	.dc.b	$69,$93,$D0,$E5,$A6,$4F,$53,$96,$99,$9F,$53,$E0,$21,$D8,$A7,$42
	.dc.b	$9C,$8A,$F2,$78,$10,$E0,$53,$81,$5E,$0E,$88,$68,$AE,$9C,$11,$C1
	.dc.b	$27,$E8,$47,$B5,$AF,$99,$AF,$85,$AF,$66,$BD,$5B,$1D,$67,$E5,$36
	.dc.b	$7B,$46,$27,$64,$70,$48,$D9,$0F,$23,$96,$99,$3C,$CE,$5A,$64,$F4
	.dc.b	$39,$69,$93,$D4,$E5,$A6,$67,$D4,$F8,$08,$76,$29,$D0,$A7,$22,$BC
	.dc.b	$9E,$04,$38,$14,$E0,$57,$83,$FC,$02,$BF,$9E,$76,$3C,$D4,$C9,$CC
	.dc.b	$62,$3F,$A3,$E3,$AA,$E1,$1B,$CB,$FA,$54,$62,$36,$1C,$5F,$48,$E1
	.dc.b	$5C,$EF,$B4,$E6,$7F,$D2,$47,$F5,$D3,$FA,$8E,$9F,$D8,$1F,$F4,$99
	.dc.b	$92,$18,$21,$FD,$8C,$A7,$42,$E3,$FE,$C4,$78,$9B,$21,$B2,$BB,$78
	.dc.b	$C7,$8D,$CB,$5E,$04,$8E,$0E,$B4,$FE,$BB,$3B,$9C,$CE,$E3,$04,$9D
	.dc.b	$92,$3E,$1B,$5F,$49,$C5,$C3,$E6,$6B,$E1,$38,$90,$E3,$3E,$22,$4F
	.dc.b	$86,$34,$73,$29,$CC,$E2,$ED,$B3,$E0,$7C,$87,$02,$9A,$31,$3E,$08
	.dc.b	$E0,$D8,$CC,$F5,$7C,$47,$12,$18,$21,$E5,$29,$E0,$2C,$75,$78,$9B
	.dc.b	$21,$B2,$BB,$78,$C7,$8D,$CB,$5E,$04,$8E,$0E,$A7,$DB,$3B,$9C,$CE
	.dc.b	$E3,$04,$9D,$92,$3E,$1B,$5F,$49,$C5,$C3,$E6,$6B,$E1,$38,$90,$E3
	.dc.b	$3E,$22,$4F,$86,$34,$73,$29,$CC,$E2,$ED,$B3,$E0,$7C,$87,$02,$9A
	.dc.b	$31,$3E,$08,$E0,$D8,$CC,$F5,$7C,$47,$12,$18,$21,$E5,$29,$E0,$2C
	.dc.b	$75,$78,$9B,$21,$B2,$BB,$78,$C7,$8D,$CB,$5E,$04,$8E,$0E,$A7,$DB
	.dc.b	$3B,$9C,$CE,$E3,$04,$9D,$92,$3E,$1B,$5F,$49,$C5,$C3,$FA,$0A,$FF
	.dc.b	$3D,$3F,$E8,$4C,$BF,$D0,$0D,$C7,$FC,$FF,$18,$8F,$6D,$72,$7D,$07
	.dc.b	$80,$87,$83,$0F,$49,$30,$F4,$46,$8E,$65,$39,$9C,$5D,$B6,$7C,$0F
	.dc.b	$AC,$E0,$53,$46,$27,$C1,$1C,$1B,$19,$9E,$AF,$88,$E2,$43,$04,$3C
	.dc.b	$A5,$3C,$05,$8E,$AF,$13,$64,$36,$57,$6F,$18,$F1,$B9,$6B,$C0,$91
	.dc.b	$C1,$D4,$FB,$67,$73,$99,$DC,$60,$93,$B2,$47,$C3,$6B,$E9,$38,$B8
	.dc.b	$7C,$CD,$7C,$27,$12,$1C,$67,$C4,$49,$F0,$C6,$8E,$65,$39,$9C,$5D
	.dc.b	$B6,$7C,$0F,$90,$E0,$53,$47,$F0,$BC,$3F,$C1,$04,$BF,$F0,$9E,$3F
	.dc.b	$E0,$C6,$CF,$F8,$40,$FF,$0A,$46,$71,$FF,$09,$46,$72,$FF,$0B,$D7
	.dc.b	$08,$FF,$15,$9F,$E2,$E2,$18,$26,$5F,$E2,$E6,$E9,$FE,$55,$7F,$C8
	.dc.b	$4D,$7F,$C7,$6D,$7F,$C6,$CD,$7A,$C7,$32,$46,$63,$AE,$9F,$E4,$F9
	.dc.b	$CB,$67,$51,$89,$F0,$CE,$48,$E0,$87,$91,$CB,$4C,$9E,$67,$2D,$32
	.dc.b	$7A,$27,$C2,$48,$DC,$65,$AF,$A8,$E2,$53,$C0,$43,$B1,$4E,$47,$09
	.dc.b	$F7,$39,$39,$1C,$08,$70,$34,$F0,$72,$70,$34,$61,$D3,$82,$1B,$23
	.dc.b	$A6,$CF,$C1,$7C,$CD,$7C,$2D,$7B,$35,$EB,$1C,$C9,$19,$8E,$B3,$F2
	.dc.b	$A7,$2D,$9F,$C3,$65,$FD,$02,$BF,$3C,$7F,$9E,$CF,$D1,$1F,$D0,$4F
	.dc.b	$D6,$D3,$46,$9F,$E8,$4D,$5C,$3F,$E8,$3A,$D4,$E6,$3F,$D1,$3B,$B8
	.dc.b	$1C,$4E,$2D,$67,$FE,$8E,$C7,$8A,$CB,$AB,$9C,$FC,$2E,$72,$F0,$46
	.dc.b	$27,$79,$77,$C7,$B5,$EC,$E5,$9F,$81,$63,$E0,$BE,$F1,$8C,$3A,$63
	.dc.b	$EA,$CF,$D3,$9F,$1C,$38,$4F,$8B,$1C,$5F,$0C,$19,$F8,$F0,$CD,$49
	.dc.b	$F2,$1D,$DF,$87,$3B,$8C,$E1,$D9,$F4,$98,$BE,$D3,$BC,$BD,$35,$BA
	.dc.b	$E8,$6E,$FA,$E5,$EA,$8C,$12,$B8,$E1,$D2,$75,$18,$AF,$43,$6B,$E6
	.dc.b	$61,$E1,$CB,$E1,$47,$36,$CF,$D4,$BE,$71,$82,$4F,$B2,$75,$9F,$D8
	.dc.b	$7B,$12,$77,$7D,$EB,$84,$72,$7D,$F3,$CB,$3F,$3C,$E7,$3F,$95,$7C
	.dc.b	$6F,$73,$9C,$FE,$06,$3E,$18,$C6,$7C,$AB,$ED,$E7,$EA,$7C,$24,$39
	.dc.b	$9A,$AE,$4E,$71,$FA,$87,$8A,$79,$47,$02,$46,$67,$8C,$63,$0F,$C6
	.dc.b	$3D,$B1,$F5,$57,$49,$CB,$C7,$2F,$45,$71,$C7,$F0,$38,$74,$AE,$11
	.dc.b	$ED,$C3,$9E,$1C,$89,$1C,$30,$E7,$5C,$30,$FC,$0E,$3E,$A8,$C6,$7F
	.dc.b	$5E,$7E,$15,$70,$C3,$9C,$70,$CB,$C5,$18,$AE,$19,$78,$B2,$E0,$78
	.dc.b	$6F,$F1,$87,$CB,$AF,$AF,$86,$A3,$35,$C5,$D9,$F1,$F2,$F3,$57,$7C
	.dc.b	$FF,$17,$18,$C3,$96,$3F,$89,$BE,$39,$FB,$2F,$F8,$16,$7E,$8B,$DC
	.dc.b	$63,$3B,$7F,$84,$CF,$C5,$8F,$C3,$8F,$16,$1C,$5E,$B9,$FA,$9B,$8F
	.dc.b	$D3,$CF,$C9,$18,$3E,$46,$3D,$9E,$6D,$CF,$F0,$11,$BC,$7E,$5D,$72
	.dc.b	$7C,$53,$A6,$E1,$F0,$EB,$EB,$9C,$73,$F8,$B9,$79,$EB,$79,$7D,$36
	.dc.b	$D7,$D0,$C7,$B5,$FE,$A2,$3C,$E7,$33,$E2,$46,$9F,$B5,$5C,$E7,$85
	.dc.b	$7A,$2F,$C1,$18,$3E,$75,$73,$3E,$09,$71,$F6,$46,$E3,$BC,$F2,$7B
	.dc.b	$9E,$4A,$F4,$9D,$F2,$F9,$59,$7D,$38,$FC,$D7,$23,$91,$CB,$FD,$33
	.dc.b	$01,$27,$51,$98,$C4,$F0,$C3,$FB,$74,$2B,$71,$86,$C3,$A7,$FA,$CC
	.dc.b	$61,$FE,$B7,$D4,$EA,$74,$FF,$5E,$9D,$E3,$DF,$0E,$58,$F9,$B0,$E1
	.dc.b	$7B,$8C,$61,$C9,$F5,$E1,$DB,$AB,$FA,$33,$E0,$9D,$61,$D0,$61,$D2
	.dc.b	$F8,$9D,$30,$ED,$3A,$C3,$9C,$E4,$DC,$7F,$66,$EB,$4F,$C9,$8C,$37
	.dc.b	$91,$3A,$AF,$97,$18,$9D,$4E,$6B,$CD,$39,$AE,$F8,$76,$E6,$F2,$CE
	.dc.b	$6B,$76,$EC,$E6,$4C,$3B,$C6,$5B,$1C,$A3,$13,$F1,$70,$F5,$F0,$E6
	.dc.b	$A5,$FC,$0C,$3D,$71,$F1,$AB,$73,$9C,$39,$EB,$FE,$1A,$BD,$CE,$6F
	.dc.b	$86,$7F,$E3,$58,$C1,$C7,$B7,$E1,$91,$F4,$B7,$3F,$7B,$CB,$AF,$D4
	.dc.b	$FD,$88,$77,$BA,$FE,$DF,$6B,$C0,$78,$1B,$AD,$65,$F7,$2B,$53,$9B
	.dc.b	$FB,$98,$74,$6D,$F0,$8F,$D2,$B5,$FF,$2A,$39,$DB,$F6,$0D,$C3,$F6
	.dc.b	$38,$7E,$9F,$CF,$ED,$1F,$8C,$B7,$88,$F1,$3D,$3A,$3C,$4F,$74,$E6
	.dc.b	$FD,$13,$97,$94,$66,$FF,$72,$D7,$FC,$D0,$D7,$F8,$19,$E0,$BF,$7C
	.dc.b	$71,$CF,$F8,$43,$6B,$D3,$0F,$17,$C7,$52,$7C,$2F,$D0,$9D,$46,$0F
	.dc.b	$2F,$8D,$EC,$F1,$FD,$6D,$8C,$9F,$C4,$31,$F8,$71,$E4,$24,$67,$14
	.dc.b	$72,$8F,$B5,$6F,$23,$17,$FC,$04,$79,$2F,$2B,$EE,$B6,$FC,$0D,$9F
	.dc.b	$E4,$2D,$7F,$D2,$1E,$5F,$EA,$1B,$D1,$F8,$68,$C4,$7C,$C6,$BF,$E9
	.dc.b	$7F,$0F,$F9,$8F,$2B,$CC,$FE,$5B,$18,$6C,$FE,$65,$B3,$B8,$FE,$73
	.dc.b	$E5,$3C,$DF,$E6,$11,$B3,$2E,$27,$84,$27,$86,$DD,$E3,$6D,$5E,$73
	.dc.b	$CE,$FA,$51,$86,$C6,$63,$EB,$35,$FE,$8F,$E7,$BA,$78,$B6,$3E,$E3
	.dc.b	$5D,$C7,$6F,$49,$CD,$F9,$4C,$7F,$AC,$46,$2B,$F4,$CD,$7F,$9E,$75
	.dc.b	$FF,$AC,$7D,$1E,$1C,$9F,$9C,$AD,$57,$FB,$1D,$AE,$70,$F0,$11,$FE
	.dc.b	$05,$EB,$3D,$3E,$70,$C5,$39,$8F,$1D,$69,$5E,$5F,$C8,$37,$2F,$5E
	.dc.b	$1D,$30,$FF,$64,$79,$3C,$6B,$59,$7A,$6B,$06,$F9,$3F,$B4,$E3,$F0
	.dc.b	$A3,$E4,$11,$FD,$45,$78,$5F,$81,$1D,$F3,$7A,$94,$EB,$3F,$33,$D6
	.dc.b	$F1,$11,$FE,$49,$83,$D6,$C4,$7F,$81,$31,$57,$D9,$24,$78,$9F,$07
	.dc.b	$AF,$23,$B1,$FD,$1F,$2F,$9B,$E4,$FF,$66,$6F,$95,$F0,$A7,$FB,$97
	.dc.b	$B1,$89,$FE,$09,$5B,$9F,$EB,$B1,$8A,$F8,$11,$FE,$FA,$5E,$CF,$42
	.dc.b	$47,$F6,$CF,$33,$FC,$49,$C7,$FD,$5E,$B7,$3F,$D6,$EB,$6F,$F6,$5A
	.dc.b	$DF,$0F,$59,$CD,$F1,$7F,$0F,$E6,$FF,$7B,$BF,$EE,$B1,$E5,$23,$FD
	.dc.b	$D1,$F4,$7B,$5F,$00,$91,$FE,$00,$F6,$FF,$C1,$AF,$F6,$D8,$DB,$FD
	.dc.b	$F1,$AF,$F3,$A3,$FC,$3B,$EE,$7F,$24,$7F,$C1,$2D,$7F,$BF,$37,$0F
	.dc.b	$F1,$AD,$3D,$DF,$4C,$E7,$2F,$F1,$93,$6F,$AF,$91,$EE,$BE,$11,$FE
	.dc.b	$24,$F5,$9E,$F7,$F8,$4E,$30,$D7,$FC,$65,$1A,$F1,$FB,$DF,$EB,$A3
	.dc.b	$13,$FE,$4D,$AE,$0F,$F5,$68,$CB,$EF,$6B,$F0,$7D,$B7,$AD,$FE,$01
	.dc.b	$87,$53,$DB,$FE,$BF,$5A,$F1,$FF,$CC,$73,$9A,$FF,$32,$CE,$5E,$35
	.dc.b	$FE,$68,$E7,$7C,$03,$E0,$3F,$B4,$E1,$C6,$D8,$71,$8C,$37,$0F,$0E
	.dc.b	$5E,$59,$F0,$63,$D3,$0E,$75,$AC,$BC,$AD,$C7,$C6,$79,$B0,$E5,$1E
	.dc.b	$5C,$BC,$93,$C0,$D5,$57,$CB,$5C,$1C,$9D,$30,$F1,$3B,$CB,$D0,$EB
	.dc.b	$2F,$5D,$71,$AD,$CF,$90,$EB,$7C,$67,$B4,$60,$ED,$87,$99,$E1,$87
	.dc.b	$67,$58,$76,$BF,$56,$1D,$A3,$13,$C3,$0E,$B7,$E6,$AD,$63,$F1,$EF
	.dc.b	$64,$7E,$36,$5F,$3B,$0F,$64,$6C,$8A,$B9,$BA,$21,$E6,$8E,$6F,$BF
	.dc.b	$18,$F9,$5F,$4E,$5E,$FC,$3C,$53,$CB,$1F,$44,$E6,$FC,$33,$D4,$F0
	.dc.b	$61,$E2,$70,$4A,$E3,$8F,$AE,$31,$97,$C4,$3E,$B6,$1D,$63,$C1,$19
	.dc.b	$CB,$CB,$5C,$6F,$E9,$12,$78,$C6,$5E,$15,$ED,$3E,$DC,$66,$BE,$0E
	.dc.b	$5E,$EC,$7D,$18,$73,$9C,$E5,$F2,$30,$E5,$7F,$11,$F8,$17,$C3,$1E
	.dc.b	$F3,$CF,$2F,$55,$7B,$B0,$E6,$FD,$78,$CE,$3D,$E3,$AE,$3D,$EF,$B4
	.dc.b	$E5,$FD,$8D,$70,$C3,$BD,$7E,$0B,$1F,$16,$3D,$B1,$F2,$DE,$E3,$18
	.dc.b	$FD,$29,$B8,$72,$7C,$99,$7C,$57,$C6,$7D,$CB,$E4,$E0,$EB,$1B,$CB
	.dc.b	$CF,$5F,$4A,$7A,$E1,$F0,$31,$F2,$C7,$1C,$7C,$4D,$AF,$14,$EA,$F9
	.dc.b	$DF,$E0,$2B,$99,$E4,$9D,$65,$F9,$2C,$39,$DF,$0A,$D6,$5E,$88,$C1
	.dc.b	$F2,$B2,$F4,$5F,$1C,$BD,$17,$E3,$CB,$D1,$87,$1A,$D4,$78,$E7,$47
	.dc.b	$6B,$E6,$FE,$62,$35,$87,$82,$BE,$B5,$F3,$AD,$1F,$94,$3D,$B3,$F8
	.dc.b	$9B,$E7,$1F,$89,$34,$78,$23,$4F,$0A,$EB,$3C,$31,$FC,$44,$7E,$2C
	.dc.b	$F4,$CF,$9C,$E0,$7D,$D9,$E1,$8F,$D2,$C3,$CD,$3E,$1C,$3E,$43,$C0
	.dc.b	$FC,$09,$70,$FA,$38,$76,$9F,$2C,$FE,$C4,$FE,$C1,$7F,$D7,$DB,$8F
	.dc.b	$FA,$A6,$FF,$AD,$47,$F8,$17,$C9,$7C,$07,$F9,$6E,$3F,$C1,$07,$FA
	.dc.b	$15,$C3,$B6,$BF,$EC,$68,$E3,$E2,$75,$F6,$9E,$FD,$25,$74,$C1,$3A
	.dc.b	$8C,$37,$3F,$EC,$AD,$D3,$FB,$77,$C0,$3E,$01,$D1,$FD,$4E,$B5,$87
	.dc.b	$F7,$66,$BF,$DC,$1C,$F1,$2F,$95,$71,$CE,$5F,$F5,$0B,$DD,$6A,$31
	.dc.b	$3A,$CB,$CA,$FF,$52,$9D,$46,$27,$64,$7F,$BD,$C6,$B2,$FF,$07,$DF
	.dc.b	$9B,$3F,$87,$87,$59,$F7,$E7,$EE,$C7,$94,$70,$9D,$47,$F5,$3C,$3A
	.dc.b	$3F,$06,$32,$FC,$17,$CD,$5C,$F1,$F3,$12,$FA,$57,$F7,$FB,$E6,$F1
	.dc.b	$3C,$27,$53,$55,$CE,$78,$D7,$39,$D3,$F0,$27,$51,$89,$D9,$1F,$54
	.dc.b	$69,$F0,$4F,$36,$CF,$C4,$CB,$DB,$8F,$C8,$9F,$2E,$7F,$77,$0E,$53
	.dc.b	$9C,$39,$37,$0E,$57,$EC,$3A,$9E,$53,$C7,$87,$6C,$FE,$B1,$F8,$4C
	.dc.b	$3A,$B8,$CF,$E9,$E5,$E1,$AE,$11,$8C,$FE,$5E,$5E,$1C,$7B,$1E,$0B
	.dc.b	$E0,$F8,$E3,$4E,$67,$8E,$1D,$E3,$DF,$18,$BE,$F1,$F5,$30,$B9,$FD
	.dc.b	$3C,$7A,$BF,$0B,$0E,$4F,$D5,$9D,$E7,$F4,$EB,$6F,$D5,$BF,$8A,$78
	.dc.b	$F3,$F8,$38,$F0,$8F,$8E,$7A,$71,$EB,$5F,$3B,$1F,$05,$7C,$3B,$E3
	.dc.b	$7B,$8F,$CA,$E3,$CE,$3B,$37,$3F,$91,$97,$92,$7E,$4C,$7C,$63,$39
	.dc.b	$FC,$38,$E8,$7C,$32,$E3,$F0,$E3,$4F,$92,$34,$D3,$58,$F9,$9C,$65
	.dc.b	$EB,$3E,$25,$F8,$73,$FC,$26,$5E,$76,$E1,$E0,$AF,$99,$8F,$81,$E3
	.dc.b	$3A,$C3,$AD,$7E,$47,$2F,$73,$F4,$31,$ED,$9F,$C6,$8E,$86,$F1,$E8
	.dc.b	$FC,$4C,$3A,$63,$DA,$BE,$86,$5E,$57,$96,$5E,$0B,$F9,$E7,$42,$B9
	.dc.b	$C7,$E6,$E7,$F0,$F0,$E9,$5F,$21,$B5,$F2,$30,$ED,$39,$3C,$98,$78
	.dc.b	$9E,$D8,$74,$9D,$D7,$E4,$E7,$38,$FA,$9E,$27,$E8,$31,$F1,$2E,$37
	.dc.b	$BE,$EC,$FE,$51,$3F,$EC,$58,$C6,$3F,$D7,$2B,$75,$AA,$1C,$EB,$6F
	.dc.b	$FB,$2D,$BD,$9F,$CC,$EB,$8D,$F0,$8F,$E6,$DD,$CF,$7D,$F0,$3F,$CE
	.dc.b	$67,$86,$1C,$AF,$BD,$EE,$B5,$18,$BE,$2F,$FB,$FF,$B3,$39,$2F,$84
	.dc.b	$F8,$4E,$17,$A8,$CC,$B1,$FF,$7D,$DE,$70,$EB,$39,$C3,$FD,$B5,$E0
	.dc.b	$3C,$8F,$89,$3F,$E9,$F8,$C4,$FF,$81,$39,$51,$89,$F9,$27,$2A,$E3
	.dc.b	$3A,$7D,$2D,$AF,$44,$FA,$FB,$DE,$76,$B0,$E9,$9F,$F6,$1C,$3B,$FA
	.dc.b	$5D,$24,$D2,$63,$E5,$8C,$BD,$A7,$8B,$E0,$D5,$8B,$C0,$F0,$61,$2B
	.dc.b	$86,$1C,$B3,$F8,$C7,$9E,$B4,$E3,$1E,$AD,$9E,$19,$7E,$13,$1E,$D8
	.dc.b	$74,$C7,$DD,$D1,$EB,$24,$7A,$5B,$87,$37,$EE,$53,$DE,$E3,$5F,$2A
	.dc.b	$9E,$CB,$DA,$FC,$3F,$6B,$DC,$7B,$BF,$8A,$BD,$D6,$A7,$38,$73,$C3
	.dc.b	$93,$FE,$42,$AD,$B6,$33,$1F,$E4,$66,$BF,$82,$36,$7C,$82,$EB,$FE
	.dc.b	$1B,$C7,$F2,$9E,$EF,$8E,$B9,$CF,$EE,$17,$8F,$E5,$F5,$34,$DC,$7C
	.dc.b	$CF,$E5,$B1,$F8,$2D,$AE,$33,$E7,$8C,$1C,$9F,$C8,$E3,$FA,$5F,$77
	.dc.b	$17,$E7,$BE,$2D,$BF,$97,$8A,$B6,$F9,$FE,$0B,$E5,$7C,$1F,$AD,$B9
	.dc.b	$7C,$B7,$C8,$47,$E5,$37,$0F,$03,$DB,$D0,$E5,$3F,$77,$DD,$76,$FB
	.dc.b	$DB,$1F,$55,$B7,$FE,$78,$6B,$F4,$FE,$13,$F7,$ED,$EF,$F9,$11,$88
	.dc.b	$FD,$CB,$5F,$B5,$E2,$FF,$24,$F2,$DE,$F7,$FA,$12,$30,$D9,$4E,$9E
	.dc.b	$26,$23,$FD,$0D,$D7,$FC,$4E,$30,$7F,$30,$F2,$FF,$77,$C1,$D2,$31
	.dc.b	$3B,$BE,$90,$AD,$BC,$8F,$F4,$5F,$C2,$FE,$5D,$A7,$E6,$7C,$94,$65
	.dc.b	$B1,$F9,$DF,$79,$EF,$F1,$6C,$66,$30,$75,$F6,$9F,$0F,$FC,$83,$4F
	.dc.b	$E9,$F1,$8A,$FD,$A3,$5F,$E1,$3D,$4D,$DF,$0F,$C7,$93,$F9,$4D,$EE
	.dc.b	$FF,$54,$D7,$38,$F8,$48,$FE,$6B,$DE,$FC,$A3,$7A,$BA,$3F,$D7,$2B
	.dc.b	$C5,$9F,$9B,$B7,$32,$94,$61,$B1,$FA,$6C,$DF,$01,$9A,$93,$EE,$7B
	.dc.b	$EB,$ED,$9C,$CF,$FB,$2A,$F9,$B6,$BE,$51,$1F,$57,$57,$2C,$CF,$2C
	.dc.b	$BE,$6B,$E5,$24,$F9,$7E,$02,$BD,$E3,$EB,$11,$FE,$35,$87,$47,$E4
	.dc.b	$7C,$07,$73,$8D,$FF,$4F,$F8,$97,$C4,$FC,$AF,$2D,$18,$BF,$E9,$8D
	.dc.b	$7F,$A2,$F9,$BF,$D3,$3C,$5E,$13,$99,$FE,$CB,$5C,$AF,$F9,$63,$63
	.dc.b	$FB,$64,$ED,$FF,$09,$C7,$83,$E2,$B9,$92,$3F,$AD,$79,$6F,$88,$C4
	.dc.b	$7F,$53,$B5,$63,$3F,$EA,$55,$79,$FA,$D6,$AF,$D7,$7F,$88,$D3,$FB
	.dc.b	$2D,$F4,$9F,$4D,$6D,$FE,$7D,$97,$95,$F4,$DF,$B2,$5F,$17,$F0,$89
	.dc.b	$3E,$8F,$32,$BC,$A7,$6F,$C2,$9E,$47,$F5,$1C,$5E,$FF,$F5,$D9,$FE
	.dc.b	$F4,$D7,$E8,$B7,$4F,$EE,$1D,$3E,$B8,$C6,$BF,$E3,$06,$D7,$F8,$AD
	.dc.b	$B7,$FE,$04,$AF,$D4,$FA,$DF,$E0,$5F,$81,$BE,$99,$A7,$5F,$42,$E3
	.dc.b	$FE,$4D,$9D,$36,$7F,$B5,$57,$38,$FE,$D9,$ED,$3D,$EF,$E5,$31,$FC
	.dc.b	$17,$D4,$FF,$15,$CE,$7C,$8F,$F2,$AF,$82,$9C,$DF,$08,$C5,$B9,$5E
	.dc.b	$FB,$E0,$3F,$C6,$B5,$AA,$65,$D5,$F0,$1F,$E6,$27,$11,$B6,$E3,$F9
	.dc.b	$4B,$E9,$5E,$48,$F1,$E5,$63,$96,$5E,$59,$CC,$76,$CB,$CB,$97,$92
	.dc.b	$73,$97,$7B,$E1,$7C,$B0,$E9,$87,$08,$EC,$7A,$A7,$6D,$9C,$E5,$E8
	.dc.b	$8F,$89,$97,$A1,$F7,$9B,$2C,$6D,$C6,$1F,$33,$0F,$04,$6B,$0E,$F5
	.dc.b	$B3,$B4,$EA,$BD,$58,$78,$AB,$67,$B2,$75,$8F,$C0,$26,$1E,$F3,$E3
	.dc.b	$63,$E7,$C3,$9C,$78,$70,$E8,$7C,$49,$D5,$72,$9F,$4B,$D2,$FB,$C7
	.dc.b	$5B,$E6,$49,$FB,$99,$7D,$39,$D4,$73,$CB,$DB,$87,$69,$C9,$0E,$E7
	.dc.b	$CE,$AE,$4E,$09,$3C,$A3,$2F,$C9,$CB,$E2,$E5,$EB,$9C,$C6,$2B,$E0
	.dc.b	$1F,$12,$33,$3F,$0A,$33,$8F,$D7,$C3,$CD,$1E,$0B,$E5,$39,$BE,$33
	.dc.b	$C3,$1F,$8D,$7C,$31,$F2,$63,$D7,$0E,$17,$6F,$E2,$63,$F6,$F0,$E0
	.dc.b	$F6,$8E,$78,$F6,$BF,$9D,$7F,$23,$2F,$A6,$F9,$71,$E8,$FC,$92,$99
	.dc.b	$3E,$01,$8C,$BC,$AE,$0F,$51,$F4,$DF,$9C,$EF,$1F,$23,$D4,$D9,$D6
	.dc.b	$FD,$AE,$27,$AC,$60,$F4,$C6,$F1,$FC,$3B,$D6,$FD,$38,$7B,$2F,$8D
	.dc.b	$7C,$2A,$D1,$E6,$9D,$61,$E4,$8F,$C4,$C6,$2F,$78,$78,$6F,$E7,$C7
	.dc.b	$6C,$BE,$7B,$A8,$F1,$DF,$A7,$0F,$7E,$1C,$E3,$15,$F8,$D7,$81,$F4
	.dc.b	$EF,$9B,$A3,$E6,$61,$F0,$31,$F5,$C7,$C9,$8C,$61,$DA,$B8,$C6,$2F
	.dc.b	$B5,$73,$8F,$C0,$E1,$E3,$75,$1D,$B1,$F6,$BF,$16,$F9,$C6,$27,$E1
	.dc.b	$4F,$DD,$C3,$C1,$97,$E3,$6B,$47,$C2,$AE,$A7,$82,$F9,$DF,$1C,$7C
	.dc.b	$B3,$AA,$FA,$D3,$E7,$6C,$66,$3E,$35,$F7,$79,$1E,$92,$C7,$78,$D4
	.dc.b	$7E,$1C,$F9,$13,$FD,$09,$FF,$63,$65,$FD,$45,$B8,$FF,$62,$6E,$1F
	.dc.b	$ED,$1F,$29,$F0,$0F,$17,$0D,$AF,$F6,$81,$FE,$85,$70,$FF,$B4,$E7
	.dc.b	$8C,$FF,$70,$F2,$9E,$FD,$25,$72,$C9,$3A,$8C,$37,$5F,$EC,$8D,$D3
	.dc.b	$FC,$6F,$F0,$0F,$80,$74,$FF,$52,$AD,$61,$FD,$B5,$AF,$F7,$27,$3C
	.dc.b	$4B,$E5,$5C,$73,$97,$FD,$5A,$F7,$5A,$8C,$4E,$B2,$F2,$BF,$DA,$27
	.dc.b	$51,$89,$D9,$1F,$EB,$D1,$AC,$BF,$C2,$37,$E6,$CF,$E1,$E1,$D6,$7D
	.dc.b	$F9,$FB,$B1,$E5,$1C,$27,$51,$FE,$13,$C3,$A3,$F0,$63,$2F,$C1,$7C
	.dc.b	$D5,$CF,$1F,$31,$2F,$A5,$7F,$82,$6F,$9B,$C4,$F0,$9D,$4D,$57,$39
	.dc.b	$E3,$5C,$E7,$4F,$C0,$9D,$46,$27,$64,$7D,$51,$A7,$C1,$3C,$DB,$3F
	.dc.b	$13,$2F,$6E,$3F,$22,$7C,$B9,$FD,$DC,$39,$4E,$70,$E4,$DC,$39,$5F
	.dc.b	$B0,$EA,$79,$4F,$1E,$1D,$B3,$FA,$C7,$E1,$30,$EA,$E3,$3F,$A7,$97
	.dc.b	$86,$B8,$46,$33,$F9,$79,$78,$71,$EC,$78,$2F,$83,$E3,$8D,$39,$9E
	.dc.b	$38,$77,$8F,$7C,$62,$FB,$C7,$D4,$C2,$E7,$F4,$F1,$EA,$FC,$2C,$39
	.dc.b	$3F,$56,$77,$9F,$D3,$AD,$BF,$56,$FE,$29,$E3,$CF,$E0,$E3,$C2,$3E
	.dc.b	$39,$E9,$C7,$AD,$7C,$EC,$7C,$15,$F0,$EF,$8D,$EE,$3F,$2B,$8E,$49
	.dc.b	$1F,$90,$DD,$F1,$9F,$EF,$C9,$FF,$31,$74,$BC,$17,$C2,$FE,$3A,$F7
	.dc.b	$E6,$FE,$42,$B5,$39,$C3,$96,$3F,$E7,$49,$CE,$3F,$E6,$DC,$38,$BD
	.dc.b	$B1,$EA,$E6,$F8,$E9,$FE,$84,$BE,$38,$7F,$9F,$B0,$E2,$F5,$C3,$8B
	.dc.b	$DF,$E3,$B8,$EC,$85,$A7,$31,$85,$F1,$DE,$B7,$33,$BD,$70,$66,$6E
	.dc.b	$9F,$1A,$9E,$97,$13,$97,$4D,$9D,$45,$C7,$CD,$37,$5F,$77,$C7,$FF
	.dc.b	$AA,$EF,$E8,$BF,$C4,$33,$39,$BF,$98,$E4,$C5,$72,$E2,$FA,$91,$8D
	.dc.b	$7E,$15,$F2,$9C,$F5,$69,$B8,$F8,$1F,$E1,$7C,$AF,$8A,$7C,$87,$DF
	.dc.b	$E2,$7C,$3E,$A7,$38,$7D,$B8,$C5,$6D,$B1,$93,$E4,$7C,$97,$CF,$F9
	.dc.b	$2F,$A0,$D8,$C9,$F3,$63,$2D,$3F,$D5,$3D,$1F,$ED,$B7,$F2,$FF,$26
	.dc.b	$F6,$BF,$8F,$4E,$78,$FF,$25,$39,$AE,$3E,$27,$E6,$63,$13,$FE,$AC
	.dc.b	$AE,$1C,$9F,$91,$9C,$BF,$ED,$18,$CF,$1F,$CC,$9C,$BF,$8F,$AF,$D3
	.dc.b	$DF,$5C,$7F,$4D,$F2,$AF,$2E,$FC,$AE,$1B,$A7,$EC,$27,$4D,$9D,$34
	.dc.b	$FD,$57,$CB,$3E,$5F,$F6,$5F,$17,$FE,$00,$9C,$C6,$3C,$3F,$DC,$56
	.dc.b	$E7,$F7,$B5,$BD,$3F,$59,$39,$7F,$71,$19,$E0,$F8,$1E,$5F,$F6,$7F
	.dc.b	$3F,$FB,$47,$63,$E6,$1E,$AF,$D0,$AD,$67,$F9,$09,$CB,$F6,$23,$39
	.dc.b	$7E,$D2,$72,$FD,$18,$CD,$F5,$BE,$97,$FB,$BC,$3C,$38,$70,$C7,$CB
	.dc.b	$39,$3C,$15,$FD,$7F,$C0,$FF,$07,$5F,$A3,$37,$CB,$F8,$B4,$56,$B0
	.dc.b	$FD,$3D,$69,$E9,$3F,$C2,$E7,$4F,$F0,$C9,$D2,$BF,$F0,$63,$D3,$BF
	.dc.b	$D7,$3A,$9C,$E6,$F4,$7D,$FE,$8B,$E6,$3C,$93,$9C,$FC,$3B,$AB,$51
	.dc.b	$E4,$CF,$E9,$61,$CB,$0E,$2D,$9F,$4F,$85,$FE,$44,$DF,$E6,$F8,$4F
	.dc.b	$80,$FA,$73,$9A,$EF,$39,$78,$F2,$FD,$6D,$BF,$03,$39,$CF,$E0,$57
	.dc.b	$F2,$FB,$E5,$18,$3F,$98,$CE,$EB,$F1,$3E,$3B,$E1,$7E,$FB,$70,$F7
	.dc.b	$AF,$7B,$33,$27,$2F,$08,$C6,$1D,$F1,$FF,$30,$FC,$CF,$BB,$E0,$7D
	.dc.b	$BA,$23,$18,$FB,$1B,$7C,$67,$53,$F0,$AB,$8F,$67,$C8,$F8,$0B,$3F
	.dc.b	$19,$B7,$EF,$E4,$FE,$D1,$18,$DB,$E3,$B6,$BE,$36,$2C,$3E,$0F,$C8
	.dc.b	$67,$16,$9F,$E7,$A9,$C6,$5F,$D6,$6B,$53,$FE,$7E,$BF,$B7,$F2,$DE
	.dc.b	$E6,$E3,$F6,$1B,$8F,$B5,$B9,$FE,$5B,$E5,$DF,$19,$FB,$4B,$7C,$D7
	.dc.b	$D7,$7F,$8E,$4E,$9F,$D1,$9A,$E5,$FD,$54,$67,$1F,$C0,$4E,$4F,$CB
	.dc.b	$46,$70,$FF,$38,$46,$67,$F0,$C2,$75,$1F,$DE,$4D,$4F,$9F,$0F,$36
	.dc.b	$1E,$5B,$FE,$FD,$6F,$99,$FE,$1B,$EB,$BE,$6D,$E5,$79,$A7,$31,$8C
	.dc.b	$FC,$B1,$89,$D4,$7F,$00,$AE,$51,$8C,$3F,$39,$18,$9E,$26,$EF,$A4
	.dc.b	$FB,$30,$F2,$4F,$FA,$FA,$3D,$D8,$7F,$06,$7E,$6B,$5F,$5B,$57,$C7
	.dc.b	$3E,$3F,$F9,$4E,$AF,$93,$F8,$F5,$FE,$1F,$AD,$61,$FC,$62,$B5,$B7
	.dc.b	$F1,$F9,$CD,$7E,$D6,$73,$5F,$C2,$E7,$31,$FC,$FE,$FA,$D7,$8B,$0F
	.dc.b	$25,$72,$8F,$EC,$06,$6F,$DD,$19,$8F,$EC,$9F,$04,$F9,$8F,$E7,$BF
	.dc.b	$38,$F9,$DF,$E6,$53,$9C,$3F,$8F,$CE,$5F,$E3,$D1,$E4,$7F,$C7,$9C
	.dc.b	$0F,$9E,$FC,$36,$3F,$8A,$AD,$3F,$EE,$68,$CE,$BF,$37,$CB,$FF,$22
	.dc.b	$7B,$EF,$93,$FF,$78,$C6,$1B,$69,$D3,$EE,$9D,$3F,$E4,$DF,$85,$FF
	.dc.b	$08,$5E,$21,$18,$79,$5F,$F8,$16,$31,$7D,$27,$7E,$A7,$F4,$97,$67
	.dc.b	$F9,$67,$E7,$DF,$27,$FE,$0D,$C7,$FD,$B1,$AB,$DA,$FE,$9F,$18,$6E
	.dc.b	$FF,$E1,$3B,$E3,$87,$F4,$77,$F9,$13,$F3,$BE,$83,$A4,$FF,$91,$B8
	.dc.b	$51,$89,$F1,$78,$D5,$7C,$AF,$47,$F0,$26,$34,$FC,$F3,$4F,$D0,$46
	.dc.b	$DF,$C3,$BC,$73,$FE,$CF,$F2,$8F,$A1,$EC,$DA,$FE,$BD,$5B,$CF,$FC
	.dc.b	$5C,$D8,$FF,$3F,$4E,$DF,$E1,$6D,$BF,$EE,$73,$B9,$FF,$2E,$7A,$7E
	.dc.b	$6F,$82,$34,$91,$FB,$78,$CB,$FD,$BF,$CA,$FE,$EC,$DF,$4B,$E2,$4E
	.dc.b	$A3,$F8,$DB,$70,$FF,$01,$D7,$09,$FE,$3E,$D8,$FF,$03,$CE,$F2,$FF
	.dc.b	$25,$36,$BF,$D3,$55,$B9,$FF,$27,$36,$3F,$D3,$D3,$B7,$FC,$F1,$E2
	.dc.b	$FD,$79,$7B,$14,$F3,$C4,$9F,$64,$65,$FF,$06,$57,$F8,$15,$B3,$97
	.dc.b	$D3,$3C,$5F,$ED,$17,$F8,$7F,$2F,$FD,$91,$F3,$17,$D2,$3F,$8B,$F3
	.dc.b	$3F,$CA,$17,$C0,$FF,$07,$C6,$CD,$79,$7F,$9A,$8F,$EC,$7C,$1F,$0E
	.dc.b	$BF,$23,$C1,$CE,$3E,$36,$3F,$EB,$DF,$83,$C4,$CB,$FF,$10,$D6,$BD
	.dc.b	$3F,$F4,$DC,$62,$7C,$DF,$2B,$23,$04,$3F,$BE,$68,$8D,$47,$F8,$3E
	.dc.b	$FC,$83,$E6,$FE,$07,$24,$9F,$0C,$79,$CC,$3D,$23,$0F,$F8,$83,$E8
	.dc.b	$FE,$56,$46,$98,$FA,$3B,$3B,$21,$FE,$11,$94,$7C,$5F,$67,$E9,$7C
	.dc.b	$F4,$AF,$87,$19,$7F,$C7,$F1,$97,$D9,$1F,$BF,$F8,$A7,$47,$DD,$FA
	.dc.b	$47,$C7,$53,$51,$2F,$FC,$AF,$3A,$6C,$E5,$FE,$FF,$F4,$BE,$D7,$FC
	.dc.b	$BB,$F4,$DF,$96,$6F,$57,$F0,$F9,$D4,$70,$DB,$F1,$D5,$B8,$FF,$18
	.dc.b	$37,$B7,$FC,$E3,$5C,$1F,$F3,$9F,$C9,$BE,$9F,$F2,$DF,$07,$FC,$67
	.dc.b	$CE,$FF,$02,$E3,$FE,$93,$F1,$3F,$D7,$51,$88,$5F,$17,$FC,$E0,$D7
	.dc.b	$FC,$81,$5C,$FD,$AF,$F5,$05,$F2,$7F,$D8,$95,$BC,$7C,$78,$F8,$AB
	.dc.b	$58,$78,$40,$C7,$C5,$7C,$7C,$1F,$EB,$3F,$50,$F8,$0E,$3E,$03,$7F
	.dc.b	$C9,$63,$CB,$1F,$6C,$E7,$AF,$BF,$9E,$C7,$95,$EF,$7E,$B9,$74,$CB
	.dc.b	$9C,$F8,$E7,$CB,$8F,$38,$E0,$7F,$67,$BE,$6E,$C4,$7F,$A6,$7A,$3C
	.dc.b	$FF,$08,$F9,$BA,$62,$25,$74,$AF,$F5,$04,$EA,$31,$5F,$EA,$97,$C2
	.dc.b	$7D,$6F,$9C,$F8,$7F,$1F,$FD,$B7,$7F,$F7,$D7,$AB,$F3,$AB,$47,$0E
	.dc.b	$25,$F0,$99,$39,$8E,$13,$A8,$F2,$5F,$5D,$7E,$1B,$63,$EC,$CE,$DF
	.dc.b	$85,$9B,$C0,$FE,$AD,$97,$D4,$7F,$1C,$DC,$3E,$DC,$69,$F9,$ED,$AF
	.dc.b	$96,$F0,$8C,$9C,$3C,$DF,$C5,$F9,$1F,$E6,$0F,$90,$FC,$2F,$8E,$F5
	.dc.b	$29,$AB,$4D,$33,$8F,$D1,$AE,$0F,$DE,$BE,$8F,$D5,$ED,$FF,$12,$5F
	.dc.b	$E1,$1D,$C7,$C0,$6E,$7F,$92,$77,$1F,$84,$6E,$87,$75,$CF,$2F,$CE
	.dc.b	$C6,$88,$E0,$CD,$7E,$8B,$E3,$BF,$DC,$CF,$E2,$5A,$FD,$26,$E1,$F4
	.dc.b	$7A,$38,$71,$CA,$E0,$FD,$28,$E9,$EF,$C9,$E8,$F7,$FA,$2C,$47,$58
	.dc.b	$B7,$FB,$0C,$15,$FB,$0F,$A9,$7C,$6F,$D4,$C2,$1F,$AA,$C5,$3A,$8C
	.dc.b	$35,$FC,$5F,$07,$C8,$3E,$5C,$F0,$AE,$C7,$49,$D9,$83,$ED,$7C,$E7
	.dc.b	$3C,$BF,$86,$FC,$B7,$F9,$7E,$7F,$70,$7F,$20,$71,$1F,$C9,$B8,$95
	.dc.b	$C7,$87,$F8,$ED,$69,$B8,$FE,$B6,$31,$87,$FB,$72,$B8,$3F,$A6,$8C
	.dc.b	$BD,$E7,$81,$E6,$C7,$F6,$33,$AE,$63,$4E,$B8,$F9,$27,$75,$CA,$FD
	.dc.b	$77,$C2,$3A,$46,$72,$F8,$31,$72,$FE,$63,$87,$3A,$C1,$E7,$9B,$19
	.dc.b	$8E,$B8,$F8,$E7,$39,$7B,$A7,$D8,$6A,$33,$97,$9A,$30,$4C,$7E,$75
	.dc.b	$71,$CB,$D1,$39,$CB,$E6,$BF,$CB,$72,$F1,$5F,$AA,$34,$79,$CD,$5F
	.dc.b	$3C,$BC,$D1,$AA,$E3,$7D,$2F,$91,$1F,$87,$7C,$72,$F8,$77,$EC,$AF
	.dc.b	$3E,$5E,$B7,$84,$78,$5C,$BC,$32,$F5,$C6,$B1,$F0,$CE,$F1,$F1,$11
	.dc.b	$FC,$45,$F5,$8C,$47,$85,$F2,$C6,$63,$91,$27,$AB,$B3,$25,$73,$7E
	.dc.b	$4A,$EA,$D9,$F0,$9E,$9C,$BE,$4E,$1E,$C8,$F9,$E7,$53,$A6,$5F,$1D
	.dc.b	$CD,$73,$C3,$BD,$F2,$C3,$DD,$18,$CB,$B1,$F9,$6B,$E3,$5B,$7C,$66
	.dc.b	$23,$F1,$6F,$03,$EE,$61,$E1,$C7,$E9,$C7,$93,$2F,$8C,$F9,$F2,$F2
	.dc.b	$57,$9E,$3F,$DF,$71,$F9,$06,$E1,$FC,$15,$B8,$7F,$B0,$1F,$F6,$FF
	.dc.b	$CE,$3E,$73,$1E,$FA,$31,$3F,$EC,$68,$C4,$EA,$73,$3E,$0F,$57,$FA
	.dc.b	$4B,$FE,$28,$6B,$FD,$B1,$AF,$F8,$D1,$B8,$9F,$EA,$FA,$7F,$8A,$1B
	.dc.b	$B7,$EF,$3E,$71,$F2,$4F,$1B,$FC,$05,$5A,$C7,$B7,$46,$0F,$EC,$17
	.dc.b	$C2,$B1,$3F,$C3,$2B,$83,$E3,$AD,$DF,$58,$D4,$70,$C7,$C8,$EA,$F1
	.dc.b	$8F,$BE,$2D,$38,$99,$78,$63,$99,$30,$E6,$FF,$03,$8D,$46,$09,$7D
	.dc.b	$67,$58,$F9,$67,$31,$F0,$0E,$93,$F1,$E7,$4F,$19,$DC,$FB,$CE,$78
	.dc.b	$FA,$1C,$47,$4C,$3D,$37,$CA,$7B,$4F,$2C,$FE,$C6,$5F,$0E,$7C,$26
	.dc.b	$B1,$F0,$1E,$3C,$7B,$57,$09,$DE,$3D,$A7,$39,$FD,$4A,$D6,$3F,$D6
	.dc.b	$73,$F8,$B8,$F4,$C7,$B1,$31,$E4,$F8,$8F,$5C,$6E,$33,$5E,$9C,$3B
	.dc.b	$1F,$89,$8C,$CF,$5A,$E2,$75,$BF,$0B,$E0,$32,$6A,$78,$B9,$C3,$D3
	.dc.b	$5E,$1A,$E3,$3C,$27,$2F,$5C,$FE,$73,$9C,$3C,$B9,$FC,$DC,$BD,$0F
	.dc.b	$E3,$B2,$EF,$5B,$8C,$63,$DB,$1E,$B3,$93,$EF,$9F,$15,$C9,$E2,$AE
	.dc.b	$6E,$30,$F0,$CF,$0C,$3C,$2F,$D8,$BE,$D3,$A9,$FC,$44,$63,$0F,$B6
	.dc.b	$7A,$72,$F9,$98,$75,$8C,$61,$F0,$73,$F4,$E7,$F0,$AB,$E2,$EE,$F9
	.dc.b	$CF,$F6,$CD,$7F,$97,$7B,$3F,$29,$39,$F3,$BF,$2D,$DA,$E7,$FF,$32
	.dc.b	$DF,$F8,$D7,$0E,$39,$FE,$F4,$61,$C6,$7F,$D7,$27,$FA,$EA,$33,$3C
	.dc.b	$23,$13,$B9,$E4,$DD,$FF,$DA,$6D,$DB,$FD,$5F,$E8,$FF,$92,$AF,$C5
	.dc.b	$9F,$9B,$C0,$F1,$E1,$FE,$8B,$8C,$7C,$E6,$59,$5F,$EA,$BA,$DE,$1F
	.dc.b	$E8,$CA,$CE,$5F,$E9,$69,$7C,$E7,$53,$B7,$5D,$C3,$94,$5A,$E1,$37
	.dc.b	$2F,$F6,$7E,$7F,$E3,$6B,$FF,$5D,$3F,$A0,$DC,$E6,$FF,$D4,$AE,$4C
	.dc.b	$57,$2E,$2F,$A7,$18,$D3,$E1,$C6,$3A,$B2,$DC,$7B,$BF,$C3,$F9,$DE
	.dc.b	$EB,$DB,$FB,$BC,$8F,$9C,$D4,$E7,$0F,$F5,$04,$62,$A4,$7C,$9F,$A2
	.dc.b	$BB,$7C,$7F,$9D,$FA,$5F,$0D,$F4,$DB,$5F,$3D,$B3,$B8,$FA,$53,$FB
	.dc.b	$6E,$97,$CE,$FE,$43,$CA,$7C,$46,$63,$0D,$BF,$C6,$E1,$F8,$DB,$E3
	.dc.b	$18,$81,$FE,$BC,$AF,$EA,$51,$F9,$38,$C3,$B6,$9F,$E8,$09,$E6,$FF
	.dc.b	$B6,$3D,$77,$C2,$F8,$63,$FD,$55,$F3,$BF,$A6,$F2,$9F,$2F,$63,$0D
	.dc.b	$C7,$FD,$64,$DA,$FF,$59,$3F,$B0,$F9,$C7,$CE,$FE,$C3,$E1,$BF,$BD
	.dc.b	$5F,$FA,$8E,$72,$FB,$3B,$1D,$7E,$EE,$D5,$F0,$C9,$5F,$AA,$C3,$8C
	.dc.b	$62,$FE,$2F,$86,$9C,$D7,$4A,$60,$F6,$1F,$43,$E7,$D3,$EC,$7C,$07
	.dc.b	$F7,$DF,$7F,$FB,$F5,$B8,$DF,$3B,$AE,$0F,$A7,$5A,$9C,$C7,$AF,$7F
	.dc.b	$B5,$87,$F0,$BD,$2C,$FF,$8B,$79,$53,$99,$FB,$B1,$89,$D4,$F8,$DE
	.dc.b	$32,$C3,$CB,$F5,$4E,$AD,$4F,$2B,$AF,$FB,$53,$C2,$CE,$BF,$33,$E7
	.dc.b	$F3,$E3,$7F,$84,$72,$F7,$F3,$27,$37,$34,$F7,$9F,$57,$8D,$F4,$59
	.dc.b	$CF,$C9,$3F,$C9,$78,$5E,$F7,$83,$D9,$7C,$4F,$92,$73,$5D,$2F,$EA
	.dc.b	$D6,$A3,$11,$F4,$6F,$81,$EA,$F8,$6F,$F2,$94,$EA,$F9,$D7,$88,$FF
	.dc.b	$17,$7A,$9F,$51,$FD,$37,$95,$FE,$4B,$E8,$FA,$0D,$BF,$D5,$4F,$09
	.dc.b	$CD,$71,$E1,$FF,$20,$5E,$EB,$B7,$1B,$CE,$7D,$07,$D2,$AD,$5F,$E2
	.dc.b	$A7,$39,$7B,$EF,$85,$FD,$49,$7D,$05,$89,$5B,$9E,$38,$F8,$1D,$C6
	.dc.b	$67,$C1,$97,$EF,$23,$D7,$97,$F9,$A7,$C4,$F5,$92,$33,$3B,$C3,$E4
	.dc.b	$56,$E7,$C1,$F5,$7D,$2F,$AF,$87,$F1,$09,$3C,$9B,$7C,$B0,$FE,$13
	.dc.b	$3E,$9F,$A1,$E2,$FD,$FC,$7F,$CF,$97,$8B,$F6,$55,$DB,$E7,$E1,$CA
	.dc.b	$7A,$57,$C7,$AD,$61,$FE,$FC,$AF,$E0,$D5,$FD,$D7,$0F,$F4,$44,$7D
	.dc.b	$A2,$4F,$C0,$7E,$39,$F4,$7D,$9B,$1F,$90,$F9,$E7,$C0,$7E,$B2,$1F
	.dc.b	$40,$F9,$3F,$D6,$D6,$A7,$F3,$46,$A3,$35,$FB,$DC,$95,$B7,$F1,$ED
	.dc.b	$7F,$9D,$35,$FF,$07,$3B,$D7,$FC,$28,$DB,$FE,$87,$EE,$3E,$81,$F1
	.dc.b	$DE,$09,$F5,$7D,$07,$F0,$4A,$D7,$91,$E8,$8C,$5F,$F3,$0A,$D9,$FE
	.dc.b	$C1,$C1,$3E,$96,$C6,$63,$07,$37,$F7,$5F,$36,$FA,$C7,$CA,$67,$5F
	.dc.b	$DA,$1F,$0A,$F1,$AF,$FA,$AE,$F8,$56,$30,$FF,$17,$7C,$FB,$E8,$3F
	.dc.b	$8E,$DB,$E8,$7E,$16,$5F,$D9,$2B,$57,$FC,$2E,$B5,$C3,$FC,$8A,$73
	.dc.b	$5F,$BE,$6B,$FD,$A5,$AF,$83,$0F,$B1,$1C,$2B,$FC,$88,$DB,$FE,$E7
	.dc.b	$97,$BF,$E3,$7B,$5F,$F3,$08,$7D,$0F,$E0,$5B,$5F,$EE,$D6,$E3,$FC
	.dc.b	$41,$AF,$F7,$B7,$C6,$F6,$AE,$44,$9F,$F7,$B4,$7F,$69,$3F,$CA,$FF
	.dc.b	$21,$FC,$6B,$D4,$FE,$C6,$D8,$C9,$FE,$07,$8C,$BF,$E0,$CF,$8B,$FF
	.dc.b	$0A,$E9,$88,$46,$1E,$26,$3D,$BF,$EA,$4E,$23,$5E,$C3,$0F,$F3,$37
	.dc.b	$CF,$BE,$5B,$CE,$DB,$FE,$5E,$1C,$3F,$CC,$9A,$FF,$87,$5F,$F0,$BB
	.dc.b	$C1,$AE,$FC,$2E,$A7,$C4,$F9,$AE,$79,$FF,$9E,$DB,$AF,$F9,$E5,$B5
	.dc.b	$FE,$2C,$F9,$DF,$F3,$FB,$FD,$42,$3C,$2E,$F1,$E8,$47,$D3,$1C,$94
	.dc.b	$B8,$BF,$D0,$D8,$7F,$A2,$E3,$13,$D6,$F3,$87,$FA,$2F,$E1,$7F,$C3
	.dc.b	$F3,$FD,$81,$AF,$EA,$1B,$7F,$E9,$5F,$9A,$7B,$7C,$AB,$FC,$8B,$5A
	.dc.b	$BE,$9E,$BE,$67,$FC,$CE,$DB,$E3,$1D,$E3,$F1,$75,$E0,$8F,$F2,$63
	.dc.b	$5E,$67,$D8,$7B,$3F,$E5,$49,$FE,$FA,$FF,$78,$6B,$E9,$9F,$14,$FF
	.dc.b	$29,$8F,$77,$85,$83,$C9,$EE,$D8,$ED,$1E,$42,$38,$3B,$7B,$5E,$B3
	.dc.b	$83,$FD,$47,$1F,$89,$E4,$7F,$44,$D3,$FA,$5C,$7C,$7F,$3F,$FB,$C7
	.dc.b	$3E,$BC,$9F,$E6,$D7,$FD,$2B,$0F,$E9,$51,$83,$3C,$7F,$02,$4D,$3E
	.dc.b	$1D,$FF,$80,$FE,$17,$C3,$19,$DE,$9E,$1F,$3F,$FD,$31,$8F,$F6,$39
	.dc.b	$D3,$67,$55,$C2,$3F,$BD,$CF,$D3,$58,$78,$AB,$F5,$6D,$8D,$FC,$CF
	.dc.b	$88,$8F,$F8,$2F,$EB,$6B,$FE,$24,$B4,$F2,$6C,$72,$3F,$C4,$B9,$7F
	.dc.b	$87,$FC,$BF,$51,$23,$FC,$2D,$E2,$D7,$FC,$2D,$17,$83,$FC,$45,$3F
	.dc.b	$13,$D6,$95,$F0,$5D,$70,$FF,$A7,$B4,$F9,$B1,$E3,$24,$FC,$27,$F9
	.dc.b	$0F,$C5,$B9,$FF,$3C,$D9,$FF,$24,$B5,$FF,$1B,$FD,$77,$F9,$E9,$CC
	.dc.b	$FE,$0F,$D8,$7C,$7D,$39,$C7,$EA,$7E,$5F,$FB,$6D,$7F,$95,$1A,$FF
	.dc.b	$5C,$6E,$BF,$E8,$8F,$7B,$F5,$B3,$9D,$FF,$21,$1F,$E8,$08,$FF,$21
	.dc.b	$35,$D1,$FE,$60,$F9,$27,$CE,$7E,$E5,$B1,$98,$FC,$D7,$C2,$B9,$B5
	.dc.b	$1F,$D1,$2B,$75,$FE,$5C,$6B,$FE,$C7,$6B,$FE,$CC,$0F,$0C,$7F,$D3
	.dc.b	$9E,$0E,$7C,$9F,$F5,$57,$37,$FA,$8E,$31,$EA,$71,$BE,$1C,$DF,$EB
	.dc.b	$2F,$9B,$7C,$E6,$B3,$79,$69,$CC,$7F,$A9,$CC,$B8,$7F,$D7,$87,$FA
	.dc.b	$E3,$83,$FD,$F8,$EE,$79,$57,$5C,$FF,$D7,$9D,$96,$FF,$33,$7C,$AF
	.dc.b	$6F,$7A,$7C,$0F,$FA,$5A,$30,$BE,$73,$A3,$63,$3D,$78,$33,$32,$32
	.dc.b	$7A,$CF,$97,$F3,$9F,$05,$A6,$76,$47,$A7,$E0,$FF,$D6,$13,$98,$FA
	.dc.b	$9B,$FD,$76,$C7,$CF,$6A,$3A,$9F,$EC,$FA,$EA,$FB,$FE,$3F,$FB,$A3
	.dc.b	$7D,$0F,$24,$EA,$3E,$7E,$9F,$E9,$BF,$3B,$E6,$78,$DF,$EE,$99,$CB
	.dc.b	$79,$BF,$06,$FF,$5C,$8C,$CF,$0E,$59,$1F,$3E,$75,$5F,$EF,$3F,$0B
	.dc.b	$FD,$FF,$E6,$FE,$15,$B5,$F8,$36,$BE,$0C,$F5,$87,$EB,$2F,$F2,$7F
	.dc.b	$26,$F7,$7F,$04,$DB,$E5,$5B,$3E,$36,$8F,$03,$FC,$45,$18,$78,$4F
	.dc.b	$C2,$6E,$1F,$A0,$77,$0A,$E4,$FE,$46,$37,$5C,$F0,$C4,$C3,$F2,$13
	.dc.b	$A7,$93,$6B,$9C,$7E,$57,$E3,$FF,$2A,$DE,$6F,$CF,$46,$23,$E9,$F3
	.dc.b	$E2,$64,$6C,$FA,$11,$C3,$D9,$91,$D5,$E7,$F3,$98,$8F,$C0,$DF,$FA
	.dc.b	$81,$D1,$E3,$8F,$DD,$53,$E7,$1F,$1B,$86,$D7,$ED,$5A,$FE,$33,$CA
	.dc.b	$7C,$8E,$EB,$F3,$37,$CD,$B7,$C3,$AB,$F8,$A4,$E7,$0F,$CF,$46,$27
	.dc.b	$AF,$0F,$E0,$A7,$F0,$27,$53,$F1,$5F,$39,$AE,$2F,$E2,$DF,$2F,$FE
	.dc.b	$3E,$AF,$CE,$75,$23,$13,$FD,$46,$B8,$5F,$EB,$6B,$59,$7E,$E6,$F7
	.dc.b	$39,$6F,$3A,$F8,$CF,$F2,$C9,$DB,$6B,$91,$E6,$D3,$F6,$73,$AC,$3B
	.dc.b	$D7,$F2,$23,$81,$FC,$93,$0E,$D5,$C2,$F9,$CE,$EB,$8E,$5F,$AF,$C7
	.dc.b	$C7,$18,$CB,$E1,$DF,$0A,$D6,$1C,$F0,$E5,$97,$8C,$F6,$E3,$C9,$C5
	.dc.b	$F1,$C3,$E4,$5F,$1C,$3E,$6D,$6B,$0F,$46,$1C,$70,$E1,$1F,$25,$F8
	.dc.b	$4E,$0F,$7D,$72,$8C,$C7,$91,$D6,$3E,$B8,$FF,$11,$D7,$57,$44,$AF
	.dc.b	$6B,$C2,$FC,$91,$88,$D6,$3F,$11,$A7,$4C,$7D,$91,$99,$F2,$4F,$B2
	.dc.b	$72,$FA,$CE,$39,$7D,$67,$C1,$97,$C0,$8C,$47,$63,$DF,$7D,$4F,$96
	.dc.b	$F0,$39,$E3,$ED,$C3,$C2,$49,$DC,$F3,$CB,$E2,$3D,$0F,$51,$CE,$78
	.dc.b	$38,$38,$65,$F1,$2B,$9C,$EA,$79,$1C,$F0,$F4,$C6,$6F,$B9,$F4,$E3
	.dc.b	$35,$EB,$AF,$89,$5E,$09,$CC,$62,$74,$F5,$C3,$DD,$18,$BE,$91,$89
	.dc.b	$E0,$4A,$E6,$F9,$A7,$C7,$87,$E6,$32,$FA,$B3,$98,$C3,$FA,$83,$FC
	.dc.b	$48,$FF,$B0,$9B,$97,$FB,$5A,$BF,$67,$1F,$E2,$73,$FD,$CB,$F0,$CF
	.dc.b	$9C,$FF,$6D,$FB,$A8,$C4,$EC,$DC,$EA,$33,$1F,$E3,$3F,$57,$FA,$8B
	.dc.b	$FE,$2B,$6B,$FC,$45,$AF,$F8,$D9,$B6,$7F,$B0,$6D,$FC,$71,$BB,$7F
	.dc.b	$9A,$FE,$71,$F2,$2E,$2F,$F0,$95,$6B,$1E,$DD,$78,$3F,$B9,$5F,$0A
	.dc.b	$C4,$FF,$0D,$AE,$0F,$8E,$B7,$7D,$63,$51,$C3,$1F,$23,$AB,$C6,$3E
	.dc.b	$F8,$B5,$FE,$9C,$C7,$C0,$F2,$25,$F2,$7F,$91,$46,$A3,$04,$AE,$93
	.dc.b	$AC,$3C,$93,$98,$F7,$9D,$27,$E3,$4E,$9E,$33,$B9,$F7,$1C,$F0,$F3
	.dc.b	$B8,$8E,$9C,$2C,$3C,$97,$C8,$95,$E0,$AF,$F1,$14,$EA,$31,$3A,$7A
	.dc.b	$3E,$4B,$EF,$9F,$DF,$CB,$E3,$E3,$DF,$2F,$46,$1C,$88,$72,$3E,$63
	.dc.b	$A7,$04,$AE,$8F,$B4,$F8,$87,$27,$39,$7A,$27,$94,$63,$1F,$43,$6B
	.dc.b	$BE,$3E,$8C,$FF,$0F,$39,$C7,$D1,$39,$CB,$D7,$9F,$C3,$C3,$C9,$87
	.dc.b	$03,$F2,$D7,$C2,$3C,$A6,$AB,$F1,$8E,$30,$F1,$3F,$06,$78,$67,$F6
	.dc.b	$E7,$27,$8F,$0F,$03,$8C,$3C,$18,$74,$AE,$11,$8C,$3E,$34,$62,$7D
	.dc.b	$39,$FC,$4A,$D4,$FE,$13,$0E,$B1,$8C,$FF,$07,$97,$91,$F9,$31,$9C
	.dc.b	$7E,$49,$EB,$71,$3B,$8F,$BC,$D7,$DD,$5D,$B1,$F4,$5F,$68,$C4,$FF
	.dc.b	$A8,$FE,$73,$FD,$BF,$A7,$FA,$AD,$B1,$FE,$37,$E7,$FC,$5F,$0A,$73
	.dc.b	$A3,$1F,$F4,$84,$E7,$D2,$DD,$6A,$7F,$CD,$FD,$0C,$33,$53,$2F,$03
	.dc.b	$F9,$8E,$C5,$6E,$72,$7F,$A2,$67,$5E,$0F,$5D,$3F,$CE,$4B,$E7,$33
	.dc.b	$32,$B8,$1F,$E9,$18,$C5,$7F,$A6,$BE,$73,$99,$D3,$A6,$D7,$FA,$5E
	.dc.b	$2C,$FF,$A8,$26,$EF,$FE,$E2,$CB,$FC,$91,$7F,$EC,$B7,$F7,$59,$FA
	.dc.b	$BA,$7D,$91,$83,$78,$7F,$98,$F9,$73,$3A,$6D,$F3,$6C,$EE,$3F,$D4
	.dc.b	$9E,$FB,$DB,$FD,$FF,$32,$31,$3F,$EA,$FF,$9B,$CC,$62,$FF,$C9,$0D
	.dc.b	$D3,$FD,$64,$DE,$3F,$F5,$2F,$12,$F8,$E5,$F4,$3E,$6F,$E8,$FC,$C5
	.dc.b	$CF,$FD,$78,$DC,$FF,$D8,$AD,$CF,$FD,$8A,$FF,$0F,$F2,$1F,$37,$4F
	.dc.b	$C0,$79,$AF,$85,$C6,$3F,$EE,$9A,$DB,$6B,$73,$AC,$7F,$CC,$95,$A6
	.dc.b	$CE,$AB,$81,$FE,$AB,$F1,$7F,$1D,$E3,$3E,$4A,$C6,$1B,$3F,$EB,$16
	.dc.b	$DF,$FA,$91,$FD,$1F,$CD,$BE,$6E,$9F,$99,$F5,$1F,$0F,$8C,$3F,$DE
	.dc.b	$9C,$1F,$A7,$C5,$E1,$31,$E5,$5A,$E6,$FD,$5C,$62,$BF,$DE,$78,$F4
	.dc.b	$AD,$47,$29,$D6,$1F,$EF,$5D,$DF,$0D,$FD,$C7,$DB,$FE,$E5,$69,$CE
	.dc.b	$08,$4F,$F7,$7E,$E7,$BC,$E6,$E5,$7B,$BE,$58,$E1,$E5,$8F,$47,$9C
	.dc.b	$EE,$31,$8F,$5C,$3A,$4F,$86,$3C,$B3,$A8,$E1,$7D,$6F,$BC,$6C,$FF
	.dc.b	$7C,$DF,$28,$EE,$78,$A7,$87,$81,$FE,$08,$C3,$E0,$B6,$FF,$E0,$0E
	.dc.b	$AF,$9A,$DF,$7C,$9E,$D8,$FC,$5B,$E0,$FF,$BF,$EF,$84,$EA,$3F,$DC
	.dc.b	$35,$E7,$E9,$7B,$34,$F8,$DB,$BE,$0B,$FC,$89,$87,$1A,$DB,$6B,$83
	.dc.b	$E2,$AE,$1E,$6E,$31,$F7,$F8,$5F,$E3,$FF,$99,$FA,$4D,$BE,$1E,$13
	.dc.b	$E6,$33,$E7,$2E,$92,$B7,$37,$0F,$97,$93,$0E,$55,$A3,$AF,$CC,$FB
	.dc.b	$23,$35,$C1,$BC,$7F,$8B,$6D,$7F,$0C,$AE,$18,$79,$3C,$29,$3E,$8F
	.dc.b	$99,$B8,$FB,$B4,$FE,$09,$3F,$53,$5F,$F2,$F5,$F8,$6F,$79,$FF,$93
	.dc.b	$E7,$FC,$C5,$E0,$7C,$03,$CD,$F3,$37,$0E,$58,$FF,$01,$6C,$E9,$A4
	.dc.b	$CB,$C5,$19,$6C,$67,$0F,$05,$73,$9F,$85,$7D,$DF,$54,$7D,$23,$E4
	.dc.b	$70,$5B,$F1,$C5,$9F,$95,$36,$FF,$73,$F2,$EF,$63,$F4,$8D,$C3,$F2
	.dc.b	$FE,$9B,$E6,$1F,$11,$AB,$95,$F9,$7B,$E1,$97,$EA,$30,$E1,$7A,$8F
	.dc.b	$D5,$89,$DE,$3E,$2C,$3F,$80,$5F,$0C,$3A,$47,$93,$E5,$E4,$7E,$AB
	.dc.b	$E5,$1F,$30,$F8,$EC,$E7,$DB,$36,$18,$CF,$FD,$11,$87,$0B,$C6,$3F
	.dc.b	$C4,$6F,$6F,$AE,$FF,$89,$56,$E3,$04,$C3,$B4,$EB,$1F,$14,$E7,$1F
	.dc.b	$E4,$BF,$2E,$F9,$6F,$E4,$F8,$7F,$1F,$8C,$57,$FA,$E6,$11,$87,$FD
	.dc.b	$78,$D7,$FD,$88,$E6,$BA,$D7,$87,$E1,$7F,$8E,$4E,$6F,$B4,$E0,$D3
	.dc.b	$FC,$36,$35,$7F,$CD,$F2,$7C,$BF,$F2,$56,$D7,$E0,$BE,$89,$F3,$B9
	.dc.b	$AD,$63,$2B,$F8,$67,$D7,$BC,$6C,$65,$F8,$03,$FA,$B9,$FD,$5B,$E2
	.dc.b	$7F,$DB,$1E,$6F,$F8,$F9,$B1,$93,$FD,$C5,$19,$7F,$A6,$7C,$23,$E5
	.dc.b	$FF,$A6,$C6,$C4,$67,$87,$84,$62,$77,$E9,$30,$FE,$A1,$7B,$AE,$5F
	.dc.b	$34,$F9,$2F,$EA,$98,$FF,$5E,$8C,$4F,$F3,$56,$BF,$C7,$3E,$C1,$F1
	.dc.b	$7C,$37,$FE,$B9,$5F,$E1,$1A,$E1,$E2,$E5,$FE,$CD,$19,$7F,$51,$F6
	.dc.b	$3F,$49,$FE,$85,$F3,$3D,$B2,$FE,$DE,$DC,$3F,$B4,$36,$7F,$B8,$7C
	.dc.b	$E7,$F8,$09,$E3,$A7,$75,$3B,$6C,$66,$78,$47,$ED,$A7,$DA,$FF,$75
	.dc.b	$36,$7E,$5D,$CB,$E2,$9F,$CD,$3F,$E4,$7F,$9B,$F0,$7C,$19,$A7,$8F
	.dc.b	$E6,$BF,$C1,$13,$FD,$7E,$74,$D9,$DB,$FD,$8E,$76,$FF,$0A,$9D,$B6
	.dc.b	$76,$FF,$0E,$9D,$E5,$FE,$16,$AD,$B6,$B6,$FF,$86,$A7,$79,$7F,$84
	.dc.b	$E9,$F3,$9E,$A2,$61,$E3,$F7,$AB,$E3,$8B,$87,$39,$B7,$CE,$6D,$7F
	.dc.b	$89,$70,$FF,$1D,$BF,$D2,$27,$F0,$9E,$63,$CA,$FF,$16,$E3,$FD,$0A
	.dc.b	$7F,$11,$E8,$FF,$69,$E8,$DD,$6A,$7F,$C6,$17,$C5,$C9,$C3,$D1,$FD
	.dc.b	$46,$3F,$D7,$F8,$3E,$1C,$F0,$6C,$2B,$F0,$FB,$F1,$F8,$4F,$2D,$6A
	.dc.b	$73,$5C,$EB,$4D,$21,$FE,$0C,$9E,$B9,$FF,$80,$6F,$9D,$4A,$DE,$9F
	.dc.b	$DF,$67,$BF,$CB,$C9,$F4,$3E,$BF,$9B,$C4,$75,$8B,$5F,$E0,$59,$F8
	.dc.b	$1E,$6C,$AF,$F0,$CB,$F0,$FC,$BC,$47,$F8,$36,$2D,$7F,$83,$67,$DF
	.dc.b	$EB,$4A,$F9,$0E,$B6,$F9,$0E,$63,$DC,$E5,$FF,$31,$7C,$4B,$8F,$F6
	.dc.b	$4D,$8F,$F1,$F3,$5F,$F1,$2F,$CE,$7F,$8A,$DC,$CF,$02,$3F,$E3,$98
	.dc.b	$D3,$FD,$B3,$E5,$7F,$BA,$BF,$E4,$96,$B9,$CF,$F1,$3E,$EF,$E7,$A3
	.dc.b	$FC,$ED,$3A,$D3,$FC,$F6,$DC,$FF,$CE,$8D,$9F,$F2,$B7,$C0,$FF,$0B
	.dc.b	$F9,$27,$CE,$52,$4F,$F9,$37,$E0,$DC,$38,$8F,$F0,$0D,$70,$7F,$CB
	.dc.b	$CD,$7F,$C8,$B3,$C1,$CD,$71,$CB,$FD,$1F,$5B,$7F,$D0,$17,$FC,$AF
	.dc.b	$C7,$7C,$E3,$E7,$3F,$D1,$9E,$57,$FA,$96,$73,$18,$F8,$8F,$F4,$B4
	.dc.b	$63,$3F,$F5,$0F,$53,$87,$FA,$FC,$7F,$A9,$27,$4B,$0E,$98,$F7,$8D
	.dc.b	$35,$FF,$56,$FA,$1D,$D6,$1F,$D1,$B9,$F0,$FF,$48,$F9,$CB,$E5,$1B
	.dc.b	$32,$BC,$B3,$96,$CE,$5F,$26,$9F,$EA,$BE,$07,$CE,$78,$65,$18,$7F
	.dc.b	$D7,$E6,$23,$C7,$F0,$8F,$A0,$E1,$5B,$AE,$A7,$FA,$F5,$C6,$DF,$ED
	.dc.b	$18,$E4,$7C,$AA,$E4,$F6,$9D,$E1,$E5,$C3,$A9,$E0,$C3,$E1,$F8,$6F
	.dc.b	$62,$D4,$C3,$EA,$CE,$5B,$39,$7C,$BB,$7F,$A3,$6B,$9C,$FD,$1F,$11
	.dc.b	$1F,$D2,$B4,$FF,$74,$46,$3A,$75,$39,$3F,$DB,$57,$FE,$97,$F4,$73
	.dc.b	$33,$1F,$F7,$3D,$F1,$7C,$B3,$F7,$3A,$FF,$C3,$38,$72,$5E,$EE,$4D
	.dc.b	$54,$3F,$23,$1A,$6C,$65,$F6,$C7,$47,$F1,$6E,$DC,$C6,$F2,$F7,$C7
	.dc.b	$E4,$3E,$3B,$FC,$2A,$FE,$79,$AF,$C8,$6C,$70,$CF,$FD,$6D,$CF,$F9
	.dc.b	$69,$CE,$3F,$11,$B7,$D7,$93,$EC,$D6,$FD,$E9,$3D,$DE,$5F,$2D,$88
	.dc.b	$F0,$45,$CB,$C1,$82,$BF,$51,$6F,$9B,$FF,$26,$61,$FB,$36,$BF,$94
	.dc.b	$8D,$7A,$FF,$AC,$F1,$9F,$21,$CA,$F7,$97,$F0,$36,$EB,$FC,$09,$B8
	.dc.b	$FE,$DA,$F8,$BD,$F0,$E4,$47,$F8,$3C,$69,$FD,$2F,$13,$E6,$F9,$C7
	.dc.b	$EC,$3E,$57,$FC,$69,$97,$E9,$FA,$91,$8C,$FF,$BC,$DF,$8B,$0E,$59
	.dc.b	$7E,$CA,$B5,$8F,$EC,$5B,$3A,$8F,$D7,$46,$9F,$06,$B8,$3F,$2F,$7E
	.dc.b	$4C,$7C,$31,$99,$FD,$E9,$C0,$F5,$CE,$EB,$A5,$70,$9D,$B7,$2F,$D5
	.dc.b	$D6,$EF,$A4,$7B,$EF,$CF,$39,$BE,$31,$D0,$CE,$3E,$FB,$E3,$97,$F1
	.dc.b	$B9,$CD,$60,$F8,$53,$72,$F8,$51,$D2,$34,$FA,$F1,$F2,$E1,$D6,$34
	.dc.b	$48,$CB,$C3,$2F,$66,$3E,$1B,$F9,$37,$C6,$FD,$91,$E1,$CB,$D9,$18
	.dc.b	$C3,$A5,$FD,$CB,$E0,$F3,$8C,$9F,$E2,$FA,$E3,$19,$AE,$47,$92,$76
	.dc.b	$7A,$E3,$47,$57,$D1,$97,$CB,$C3,$CF,$18,$7C,$D8,$78,$A3,$E4,$9E
	.dc.b	$2C,$BE,$4E,$1E,$0C,$3A,$CE,$A3,$C4,$65,$C4,$6A,$78,$BC,$23,$CC
	.dc.b	$68,$F7,$57,$85,$C9,$A9,$E2,$EC,$CE,$5F,$49,$E2,$7A,$CF,$46,$5F
	.dc.b	$21,$E1,$1B,$9E,$71,$CB,$2F,$88,$FE,$72,$FC,$55,$D6,$3C,$57,$CD
	.dc.b	$E1,$97,$CA,$7E,$8C,$EC,$ED,$3B,$71,$7F,$96,$9F,$C6,$9F,$EA,$98
	.dc.b	$FD,$43,$63,$FD,$5E,$FE,$0C,$FF,$57,$7B,$4F,$85,$7C,$DF,$FA,$C6
	.dc.b	$7F,$DC,$31,$87,$13,$B9,$D4,$7F,$B8,$A7,$4F,$87,$D2,$FE,$8C,$FF
	.dc.b	$81,$5A,$FF,$7B,$6B,$FE,$06,$6E,$1F,$E9,$3E,$7F,$EF,$51,$FD,$57
	.dc.b	$7F,$ED,$3F,$36,$F8,$F7,$99,$AB,$98,$FF,$06,$C3,$87,$17,$FA,$5B
	.dc.b	$0E,$17,$AC,$7F,$92,$56,$AF,$C9,$87,$F6,$EB,$DC,$73,$8E,$18,$7A
	.dc.b	$DD,$63,$E2,$9D,$13,$0E,$B3,$AC,$78,$CF,$63,$57,$E5,$BE,$57,$C2
	.dc.b	$72,$6D,$F8,$31,$A9,$F1,$4E,$8F,$95,$7D,$DC,$65,$F2,$DE,$7C,$4C
	.dc.b	$BD,$B8,$F7,$26,$1D,$1E,$F9,$7A,$2F,$A9,$CA,$B8,$D6,$C9,$1E,$77
	.dc.b	$D9,$3C,$70,$F1,$5F,$6A,$FF,$60,$5F,$27,$81,$0E,$F1,$E2,$7D,$79
	.dc.b	$FD,$77,$38,$76,$73,$97,$CE,$CB,$DD,$9F,$C7,$9E,$99,$78,$AF,$CF
	.dc.b	$5A,$CF,$CD,$8F,$5A,$D5,$F9,$B0,$E4,$F8,$0E,$73,$F8,$9B,$E5,$1E
	.dc.b	$23,$E3,$4F,$0A,$E7,$1C,$4E,$53,$C3,$2F,$89,$19,$9F,$0C,$E7,$1F
	.dc.b	$9B,$39,$C3,$C1,$9F,$C9,$8C,$67,$F0,$63,$18,$78,$09,$5C,$32,$F2
	.dc.b	$3E,$F2,$57,$E2,$72,$F1,$D7,$08,$C6,$3E,$28,$C6,$3F,$E1,$2C,$7B
	.dc.b	$E1,$D7,$3F,$16,$5D,$B1,$E3,$7F,$7B,$3F,$63,$70,$E3,$3A,$8F,$8B
	.dc.b	$1F,$85,$9D,$8F,$B2,$7C,$B7,$DD,$8E,$74,$EF,$CE,$5F,$EA,$EF,$9E
	.dc.b	$7B,$1F,$6A,$73,$5F,$E7,$A9,$CD,$7F,$A5,$A7,$3C,$1F,$E5,$59,$CD
	.dc.b	$7F,$86,$27,$3C,$1F,$E7,$89,$CD,$7F,$A5,$67,$35,$FE,$B2,$9C,$D2
	.dc.b	$24,$EB,$BD,$5A,$9F,$F3,$57,$8B,$F9,$99,$77,$7F,$AB,$B2,$FF,$36
	.dc.b	$7C,$06,$63,$17,$FE,$6D,$BC,$19,$9B,$5F,$E7,$3E,$EE,$1F,$1B,$C5
	.dc.b	$B1,$93,$FC,$F1,$19,$69,$FE,$D8,$FB,$37,$B3,$E5,$6C,$67,$83,$13
	.dc.b	$23,$83,$B8,$C6,$5F,$EA,$AA,$E2,$D9,$D5,$70,$3F,$DC,$9F,$2E,$F9
	.dc.b	$8F,$89,$EE,$3A,$BF,$7D,$E0,$AB,$57,$FE,$EE,$AD,$3C,$AF,$8D,$F0
	.dc.b	$AD,$E3,$FE,$BC,$BD,$BF,$E9,$7F,$03,$E7,$1F,$3F,$B2,$E3,$FE,$9D
	.dc.b	$8B,$3F,$E9,$D9,$BC,$FF,$83,$F9,$7F,$E8,$DE,$B7,$F4,$7E,$67,$9E
	.dc.b	$F9,$5F,$CA,$56,$B3,$FF,$61,$CE,$5F,$CA,$46,$72,$FF,$7E,$CE,$5F
	.dc.b	$CA,$46,$6F,$AD,$F4,$BF,$C8,$61,$E1,$C3,$86,$3E,$59,$C9,$E0,$AF
	.dc.b	$C8,$7C,$B3,$E6,$7A,$2F,$99,$B1,$86,$D7,$2D,$19,$AF,$84,$E6,$31
	.dc.b	$7C,$63,$85,$3C,$9F,$EE,$3A,$7C,$7E,$07,$A7,$F0,$6B,$5C,$1E,$F9
	.dc.b	$CE,$1C,$B0,$E2,$FF,$BB,$23,$34,$C3,$FB,$C3,$CA,$7C,$93,$A9,$CF
	.dc.b	$7B,$E6,$7E,$67,$BA,$F9,$AF,$14,$E7,$0F,$64,$E5,$F5,$C6,$71,$F0
	.dc.b	$47,$FB,$FA,$75,$18,$3F,$81,$DF,$6B,$EA,$DA,$E6,$D9,$E0,$7E,$2B
	.dc.b	$D8,$FC,$07,$A6,$FA,$2F,$8D,$CC,$9C,$C6,$1B,$87,$D0,$9D,$3F,$56
	.dc.b	$B9,$3F,$6F,$C7,$7A,$EF,$77,$BE,$DF,$93,$9C,$C6,$2B,$85,$6D,$E9
	.dc.b	$5B,$8C,$67,$F0,$79,$58,$FF,$39,$BF,$07,$D7,$3E,$17,$F3,$9E,$4B
	.dc.b	$ED,$3C,$13,$9A,$FC,$64,$E5,$E3,$5C,$2B,$6D,$8D,$F4,$74,$3A,$F8
	.dc.b	$36,$32,$4C,$FC,$D8,$7F,$04,$E6,$FD,$D3,$F8,$79,$FE,$79,$C0,$CF
	.dc.b	$F0,$15,$AF,$AA,$F3,$F2,$FA,$9B,$7E,$86,$E3,$D9,$F6,$FB,$DF,$03
	.dc.b	$EA,$6D,$7B,$9B,$5E,$46,$E1,$E4,$6C,$F9,$31,$5F,$CA,$8C,$DB,$2F
	.dc.b	$44,$E3,$87,$FA,$7D,$6A,$7D,$9F,$15,$F4,$BC,$6F,$A2,$D9,$FA,$ED
	.dc.b	$C3,$F3,$95,$C2,$7E,$AD,$7E,$E7,$E6,$9F,$35,$FC,$6F,$D6,$7C,$DB
	.dc.b	$D4,$D4,$E6,$31,$87,$F0,$6B,$E2,$FF,$0F,$BE,$37,$C2,$31,$5C,$32
	.dc.b	$FD,$3D,$F0,$7F,$61,$7C,$2B,$A6,$1D,$F0,$FE,$1D,$F3,$4F,$97,$FE
	.dc.b	$61,$C4,$FA,$57,$C1,$EA,$73,$18,$C7,$F8,$DD,$F1,$7F,$92,$5F,$1B
	.dc.b	$E1,$18,$AE,$19,$FE,$F2,$F8,$3F,$C1,$6F,$85,$74,$C3,$BE,$1C,$AB
	.dc.b	$FD,$71,$7C,$1F,$F0,$71,$FE,$0F,$70,$E1,$AF,$F8,$4F,$EB,$DF,$41
	.dc.b	$FD,$37,$E7,$5F,$6A,$F9,$0F,$E6,$15,$AC,$FF,$9F,$56,$A7,$37,$C6
	.dc.b	$F8,$57,$F3,$CB,$E1,$87,$F1,$FC,$38,$4E,$6F,$F9,$9C,$FA,$2B,$6F
	.dc.b	$A8,$38,$75,$1C,$1F,$F1,$65,$7C,$8D,$3F,$D6,$9C,$0F,$9E,$FF,$18
	.dc.b	$63,$FD,$3A,$B4,$FF,$8C,$E7,$57,$FD,$36,$5F,$2D,$F9,$5F,$B6,$7D
	.dc.b	$36,$63,$0F,$F5,$3A,$FF,$1E,$7A,$BF,$E3,$DF,$65,$F1,$DF,$EE,$B8
	.dc.b	$C3,$65,$3A,$7D,$D3,$A7,$FC,$8D,$F1,$3F,$E5,$8D,$B1,$08,$C3,$CA
	.dc.b	$FE,$FC,$62,$FA,$4E,$FD,$EF,$EC,$0E,$CF,$F2,$97,$CE,$FF,$8A,$DF
	.dc.b	$EE,$3F,$6E,$F9,$8F,$C4,$E5,$FD,$9A,$31,$3F,$DD,$AB,$83,$FD,$49
	.dc.b	$7C,$E7,$0E,$2F,$99,$EE,$7F,$6E,$C3,$E9,$B6,$3F,$2B,$5C,$2F,$FC
	.dc.b	$03,$E5,$BE,$9B,$F0,$11,$F9,$A8,$C3,$83,$FC,$81,$15,$A6,$63,$6F
	.dc.b	$F6,$6F,$87,$FF,$06,$3F,$E0,$E6,$BF,$DA,$A3,$78,$FF,$81,$1B,$87
	.dc.b	$EA,$AB,$73,$FE,$08,$6C,$79,$6F,$E0,$5F,$F9,$F7,$E7,$3D,$8F,$A9
	.dc.b	$AF,$E6,$DA,$FE,$D9,$AF,$1C,$F9,$AF,$9E,$FE,$E2,$FF,$89,$9A,$FF
	.dc.b	$80,$A7,$BE,$3F,$E1,$C6,$E1,$FC,$0A,$B7,$3F,$E2,$06,$C7,$86,$FB
	.dc.b	$D7,$0B,$FF,$4B,$56,$9B,$EA,$6A,$BC,$BF,$3C,$F7,$7C,$FC,$3F,$E1
	.dc.b	$1B,$FF,$1D,$B5,$FF,$0B,$CE,$DF,$F1,$FB,$5F,$F0,$DC,$6F,$3F,$F1
	.dc.b	$AB,$73,$FE,$37,$5B,$9F,$F1,$CB,$63,$F8,$FC,$EE,$BF,$D5,$CD,$53
	.dc.b	$B7,$FB,$07,$9C,$FA,$FF,$B3,$E7,$7F,$AB,$BE,$0E,$FA,$9F,$E7,$5B
	.dc.b	$FF,$09,$D7,$DE,$F3,$FF,$BF,$3A,$A9,$3A,$3F,$C3,$3F,$5B,$BC,$4D
	.dc.b	$3B,$FB,$3F,$D3,$67,$64,$3F,$B4,$CF,$07,$FC,$69,$E5,$F1,$7F,$C3
	.dc.b	$FF,$3D,$E6,$30,$F0,$BC,$3A,$3A,$70,$38,$3F,$C0,$9E,$FF,$BB,$D0
	.dc.b	$95,$FE,$30,$8F,$41,$87,$FC,$43,$18,$7F,$C8,$DE,$D7,$A8,$8F,$F9
	.dc.b	$27,$E1,$AB,$FE,$3D,$D7,$FC,$74,$F0,$34,$7F,$8F,$BD,$DF,$9F,$CF
	.dc.b	$23,$04,$3F,$C6,$2E,$5F,$8B,$1F,$4F,$EE,$1F,$51,$F3,$26,$1F,$E5
	.dc.b	$98,$CB,$63,$2F,$94,$E4,$FE,$C3,$E6,$3F,$8F,$C7,$F9,$57,$E2,$7F
	.dc.b	$A0,$BF,$E9,$96,$BF,$E7,$28,$DE,$FF,$E8,$C6,$EF,$FE,$FF,$E8,$FF
	.dc.b	$51,$79,$5F,$8E,$C3,$8D,$EF,$AF,$F7,$FF,$2A,$F7,$BF,$85,$FA,$0F
	.dc.b	$B9,$FE,$E5,$AF,$F8,$7D,$AF,$FA,$0A,$B8,$3F,$EA,$A6,$BF,$E1,$E8
	.dc.b	$DE,$1F,$D9,$1B,$A7,$F8,$1A,$B7,$3F,$EA,$46,$C7,$F6,$59,$DD,$2F
	.dc.b	$C9,$EB,$7A,$31,$FE,$2F,$F3,$9A,$D9,$F4,$0E,$6F,$D6,$B7,$1F,$F5
	.dc.b	$FD,$74,$4F,$18,$D4,$7F,$A1,$5B,$AF,$FA,$E2,$7B,$7D,$13,$E4,$FF
	.dc.b	$72,$BC,$AA,$62,$26,$1F,$ED,$19,$D3,$67,$2F,$CB,$EE,$FE,$07,$F1
	.dc.b	$3F,$CE,$BD,$4F,$F2,$A4,$FF,$BC,$3A,$FF,$C5,$B1,$88,$72,$2F,$ED
	.dc.b	$9C,$AB,$7E,$FF,$FB,$F1,$B4,$38,$CF,$07,$FD,$CD,$87,$86,$B8,$47
	.dc.b	$F8,$39,$FB,$13,$FC,$2B,$E1,$B9,$F0,$FF,$AA,$9B,$DD,$FE,$D9,$BE
	.dc.b	$07,$FB,$D1,$FF,$0E,$B8,$9E,$0D,$8E,$0B,$3F,$83,$3A,$8C,$61,$F9
	.dc.b	$DA,$E7,$18,$3F,$C3,$BF,$43,$FE,$C7,$7B,$71,$7F,$02,$75,$7F,$86
	.dc.b	$F0,$3E,$1F,$AA,$BE,$15,$CE,$31,$F1,$5E,$EC,$39,$67,$E4,$6E,$7F
	.dc.b	$E3,$1C,$38,$CF,$A3,$CD,$7C,$8F,$C6,$6E,$5F,$05,$AF,$91,$7C,$86
	.dc.b	$E6,$4E,$DE,$51,$88,$E1,$16,$BC,$13,$F4,$BE,$37,$E9,$FC,$54,$BF
	.dc.b	$9D,$19,$7E,$84,$65,$F8,$BC,$1F,$EF,$97,$EA,$6B,$F7,$08,$FD,$7F
	.dc.b	$8A,$AF,$08,$B8,$FD,$8A,$57,$E4,$2B,$F2,$FC,$FF,$88,$F9,$7F,$F2
	.dc.b	$3D,$FD,$E6,$EF,$FC,$16,$78,$74,$7E,$A2,$FE,$CD,$69,$72,$23,$11
	.dc.b	$D8,$FC,$89,$4E,$3E,$F7,$84,$EB,$7F,$A4,$F4,$9F,$3B,$FE,$36,$CF
	.dc.b	$F6,$31,$8F,$92,$D4,$E7,$4F,$F7,$8F,$DD,$3C,$EC,$D7,$A3,$97,$FC
	.dc.b	$DF,$A7,$E5,$4F,$DB,$B8,$CB,$AE,$3C,$E7,$FD,$C8,$D7,$7D,$3E,$AE
	.dc.b	$0F,$C5,$BF,$AC,$D3,$C9,$36,$B8,$56,$A3,$C0,$DB,$EF,$8F,$97,$0E
	.dc.b	$98,$F8,$F4,$FE,$68,$DC,$3C,$75,$B7,$9C,$65,$F2,$9F,$B1,$CB,$C6
	.dc.b	$EA,$FA,$91,$D6,$5E,$6A,$E5,$97,$97,$1F,$0D,$FA,$F1,$F0,$65,$FC
	.dc.b	$7E,$B5,$8F,$F1,$96,$E5,$E0,$9D,$63,$D0,$98,$73,$BF,$14,$EA,$2E
	.dc.b	$3F,$C8,$EF,$93,$E2,$AF,$51,$2B,$E4,$BC,$0F,$15,$F6,$7D,$B8,$FA
	.dc.b	$2B,$95,$77,$9C,$9C,$4F,$92,$E4,$EA,$F7,$8E,$24,$7E,$99,$C7,$1F
	.dc.b	$86,$F2,$8F,$56,$5F,$37,$2D,$63,$F6,$F2,$F6,$DF,$A2,$72,$7C,$57
	.dc.b	$C0,$64,$B1,$C1,$EC,$67,$1F,$46,$1E,$0C,$BE,$04,$E6,$3D,$15,$C7
	.dc.b	$2F,$39,$31,$F2,$65,$E7,$BF,$09,$DA,$B8,$4F,$CD,$C7,$BC,$FC,$23
	.dc.b	$77,$CF,$2F,$54,$EA,$7C,$E7,$3C,$3B,$46,$6B,$F1,$31,$8C,$7E,$75
	.dc.b	$6F,$2F,$5B,$CA,$B7,$18,$7C,$11,$F7,$0F,$1E,$3E,$7A,$F8,$91,$FA
	.dc.b	$13,$FB,$81,$F3,$4A,$7E,$A8,$B3,$FB,$97,$3F,$02,$F9,$D7,$CF,$7F
	.dc.b	$B4,$E3,$0D,$8C,$CF,$FB,$6A,$72,$FF,$B7,$29,$EF,$FF,$8A,$E3,$0D
	.dc.b	$7C,$2D,$7F,$C6,$CD,$7F,$60,$F1,$D3,$F6,$8D,$C7,$F5,$5F,$3D,$FD
	.dc.b	$6B,$CD,$FE,$31,$39,$BE,$3E,$66,$0F,$E3,$55,$B8,$FE,$2F,$3B,$7F
	.dc.b	$C3,$AD,$C7,$D3,$5C,$3B,$7F,$D1,$95,$C2,$3A,$34,$FE,$29,$7D,$72
	.dc.b	$F3,$CE,$9B,$3A,$AE,$71,$82,$65,$F0,$AB,$C1,$97,$18,$C5,$F7,$AF
	.dc.b	$E8,$F8,$77,$9C,$C7,$29,$F8,$27,$7A,$E4,$E2,$35,$C4,$BE,$53,$9B
	.dc.b	$E6,$45,$3B,$C7,$D3,$39,$9F,$93,$8F,$6A,$D6,$3E,$09,$C9,$E9,$AF
	.dc.b	$B3,$7C,$A3,$C0,$43,$78,$FB,$A3,$51,$89,$F0,$65,$F5,$A3,$DC,$75
	.dc.b	$39,$39,$AE,$57,$E3,$8C,$57,$B7,$0F,$3D,$F1,$3B,$1E,$47,$37,$F3
	.dc.b	$63,$26,$A3,$33,$E4,$CF,$F1,$47,$E4,$E3,$33,$E4,$8C,$E5,$F7,$A3
	.dc.b	$C1,$87,$91,$F0,$67,$F5,$1E,$F1,$97,$C1,$19,$8C,$12,$BC,$11,$97
	.dc.b	$BC,$FC,$5B,$F1,$BD,$23,$38,$FC,$68,$F1,$C6,$31,$FA,$F9,$7C,$98
	.dc.b	$C4,$FE,$13,$0E,$D9,$64,$F6,$61,$F8,$E8,$CC,$FD,$68,$C6,$7F,$73
	.dc.b	$3F,$81,$9F,$91,$B7,$C2,$BC,$87,$0C,$BA,$E7,$FD,$AF,$E7,$9F,$3D
	.dc.b	$F8,$BC,$7D,$BC,$B8,$BF,$F4,$17,$4B,$E1,$BF,$17,$39,$DB,$FC,$7F
	.dc.b	$18,$C3,$F2,$9D,$7F,$E3,$F8,$C6,$1C,$9B,$5C,$23,$FC,$83,$7C,$F4
	.dc.b	$FF,$27,$57,$2F,$7B,$FD,$71,$5F,$E9,$36,$DF,$FA,$19,$7C,$F6,$FB
	.dc.b	$25,$6D,$E5,$18,$AE,$1F,$3D,$D7,$CD,$EC,$DB,$EC,$D9,$DC,$78,$67
	.dc.b	$DB,$F3,$CF,$9E,$A6,$33,$B5,$FE,$A9,$9D,$36,$74,$D2,$5F,$FA,$82
	.dc.b	$76,$DA,$E6,$FF,$A7,$FE,$79,$F3,$1F,$3B,$C9,$FE,$8F,$87,$FA,$EB
	.dc.b	$C2,$56,$A3,$FD,$35,$D9,$FC,$1E,$73,$7C,$63,$15,$BC,$3F,$D3,$D1
	.dc.b	$F4,$7C,$09,$87,$D1,$EC,$B5,$F4,$66,$F1,$FF,$B0,$B8,$FF,$91,$F2
	.dc.b	$9F,$F5,$6E,$1F,$7B,$D4,$FE,$87,$EA,$7F,$44,$A7,$A2,$F8,$3F,$C6
	.dc.b	$D6,$AF,$FD,$C5,$5A,$9C,$C7,$FB,$89,$B9,$B0,$E9,$87,$3A,$F0,$1B
	.dc.b	$2C,$ED,$EB,$1B,$8F,$F6,$8F,$B4,$F9,$EF,$14,$7F,$A6,$A1,$F1,$98
	.dc.b	$C9,$1B,$C3,$FD,$F1,$8F,$B6,$37,$18,$3D,$6E,$EF,$C4,$75,$BE,$91
	.dc.b	$9A,$E5,$1E,$29,$79,$1F,$DC,$E7,$E5,$72,$BE,$6F,$53,$9C,$7E,$4C
	.dc.b	$E5,$F1,$58,$E7,$C5,$EB,$9D,$47,$90,$CC,$F0,$CD,$3A,$C7,$FB,$ED
	.dc.b	$7F,$AB,$7D,$5F,$89,$F3,$CF,$9E,$F4,$78,$29,$CC,$61,$B9,$79,$2F
	.dc.b	$93,$DE,$F9,$39,$BE,$4E,$6F,$93,$4E,$5C,$2F,$53,$EB,$7A,$4F,$85
	.dc.b	$F9,$53,$9A,$F6,$4E,$5E,$35,$C2,$B6,$D8,$F9,$3E,$3B,$E2,$DE,$AF
	.dc.b	$6C,$FF,$0F,$39,$8C,$57,$0A,$DB,$E0,$AD,$C6,$35,$FC,$1F,$0B,$3F
	.dc.b	$F6,$C5,$F8,$3E,$25,$ED,$7E,$47,$B1,$F3,$DE,$09,$CD,$7D,$79,$CB
	.dc.b	$C6,$B8,$56,$DB,$1B,$E8,$E8,$75,$F0,$2C,$64,$99,$F9,$B0,$FE,$22
	.dc.b	$DA,$F1,$72,$F9,$7E,$7A,$DF,$8D,$B5,$C5,$BC,$FF,$C0,$F0,$47,$AF
	.dc.b	$DA,$F7,$7C,$0D,$C3,$D8,$DC,$3C,$4D,$C3,$93,$67,$FD,$91,$B2,$FE
	.dc.b	$44,$66,$FF,$C6,$B3,$98,$F9,$D3,$A6,$9F,$47,$DB,$FA,$1E,$37,$CF
	.dc.b	$6C,$64,$FA,$11,$96,$9F,$9D,$F9,$D7,$C5,$7F,$0C,$F5,$DF,$3C,$F5
	.dc.b	$BF,$7D,$5A,$BF,$DB,$D6,$A7,$37,$C6,$F8,$57,$58,$C4,$EA,$38,$CE
	.dc.b	$B4,$FD,$E6,$1D,$5F,$E2,$38,$75,$AF,$D9,$7C,$E3,$E3,$BF,$90,$F0
	.dc.b	$3E,$79,$F3,$DF,$C4,$EB,$57,$FC,$32,$B5,$39,$BE,$37,$C2,$BA,$C6
	.dc.b	$27,$51,$FE,$AE,$C7,$F8,$A6,$1D,$1C,$35,$FF,$59,$B5,$FF,$5B,$BB
	.dc.b	$C3,$F8,$9B,$67,$F8,$F7,$CD,$3E,$A7,$F7,$11,$E2,$3F,$88,$7C,$F3
	.dc.b	$E7,$BF,$9F,$D3,$E7,$DF,$1D,$FC,$C6,$B5,$7F,$CA,$AB,$53,$9B,$E3
	.dc.b	$7C,$2B,$AC,$62,$75,$18,$3F,$99,$FD,$E3,$EF,$1E,$E6,$B6,$FE,$65
	.dc.b	$97,$AD,$71,$3E,$87,$F9,$D6,$3F,$CA,$6B,$53,$9A,$FE,$91,$0F,$A0
	.dc.b	$FC,$87,$D0,$BE,$5B,$31,$87,$FA,$75,$7F,$45,$F6,$BF,$A3,$7B,$AF
	.dc.b	$8E,$FA,$D1,$86,$CA,$74,$FF,$BA,$27,$4F,$F4,$FF,$8A,$FF,$25,$73
	.dc.b	$E2,$11,$87,$95,$FF,$BC,$A3,$17,$D2,$77,$F0,$1F,$DF,$DD,$9F,$D6
	.dc.b	$FE,$83,$FC,$4A,$FF,$4B,$FA,$17,$CC,$7F,$BF,$32,$FE,$AF,$18,$9F
	.dc.b	$EA,$15,$C1,$FE,$C0,$BE,$57,$87,$17,$F6,$7E,$2F,$EF,$75,$FB,$4F
	.dc.b	$7B,$FB,$0E,$3F,$D8,$B0,$E1,$3F,$D6,$FC,$E7,$D0,$F2,$8F,$EE,$11
	.dc.b	$87,$51,$95,$F1,$5F,$DA,$5C,$9F,$D9,$5F,$F0,$0B,$7E,$03,$85,$73
	.dc.b	$8F,$EF,$0D,$C3,$FB,$AD,$70,$9F,$EF,$8D,$8F,$EF,$33,$BC,$FF,$C0
	.dc.b	$94,$FA,$1F,$5B,$FE,$03,$6D,$7F,$80,$1B,$3C,$3E,$87,$F0,$55,$FD
	.dc.b	$EE,$78,$BD,$0E,$9B,$7E,$FD,$FF,$09,$46,$9F,$0B,$5C,$C7,$C5,$9F
	.dc.b	$EE,$11,$FE,$07,$6B,$E5,$8F,$95,$3E,$48,$FF,$0F,$4E,$AF,$FC,$2E
	.dc.b	$DB,$EF,$7F,$47,$E8,$7F,$C3,$B7,$FE,$13,$E1,$FF,$04,$37,$DE,$D7
	.dc.b	$C1,$48,$C1,$0F,$F1,$45,$72,$6C,$7D,$DA,$E7,$1F,$E2,$D6,$E5,$FE
	.dc.b	$2A,$AE,$0B,$1F,$EB,$3F,$43,$F8,$9F,$6F,$FC,$D9,$F0,$F7,$D7,$FF
	.dc.b	$34,$65,$FE,$05,$AF,$C7,$7C,$03,$D6,$46,$EE,$56,$CF,$F0,$67,$D0
	.dc.b	$F0,$A3,$4F,$07,$BD,$FD,$0E,$76,$43,$FC,$B3,$87,$79,$DF,$A1,$C6
	.dc.b	$7F,$C2,$5F,$37,$E7,$30,$F0,$C3,$BB,$A2,$19,$AF,$27,$17,$BF,$D6
	.dc.b	$93,$FE,$23,$8C,$BF,$E1,$F8,$CB,$D6,$3D,$3E,$F4,$8F,$F1,$FB,$EC
	.dc.b	$F8,$AC,$47,$F8,$C3,$4F,$F1,$6B,$A3,$FC,$8B,$1F,$17,$E0,$BE,$87
	.dc.b	$3C,$9F,$F1,$A4,$65,$FF,$18,$46,$5F,$74,$7D,$4F,$A1,$7D,$0F,$CC
	.dc.b	$98,$7F,$93,$23,$2D,$8C,$BE,$53,$93,$FC,$1F,$E8,$7F,$8D,$47,$F9
	.dc.b	$23,$E3,$3F,$9E,$3F,$E7,$76,$BF,$E7,$B8,$DF,$47,$F9,$E1,$BD,$1F
	.dc.b	$E9,$4F,$2F,$F1,$D7,$C2,$B5,$D7,$FC,$57,$E6,$1F,$05,$FC,$6F,$C7
	.dc.b	$7D,$0F,$F6,$6D,$3F,$C2,$ED,$7F,$D1,$55,$C1,$FF,$3F,$35,$FF,$08
	.dc.b	$C6,$F0,$FD,$04,$7F,$5E,$8C,$67,$FC,$52,$B8,$57,$FA,$41,$B1,$FC
	.dc.b	$D6,$77,$5C,$7D,$6F,$46,$3F,$CB,$7E,$87,$3E,$03,$E8,$1C,$BF,$AD
	.dc.b	$6E,$1F,$E9,$3A,$E6,$9E,$31,$A8,$FF,$3B,$B7,$4F,$F5,$1C,$F6,$FA
	.dc.b	$1F,$EF,$75,$FE,$A9,$7F,$9E,$B7,$C8,$FD,$E4,$6A,$3F,$D0,$2D,$AF
	.dc.b	$F7,$33,$C2,$33,$3C,$3D,$7F,$A9,$DF,$FC,$17,$E3,$BF,$9B,$FA,$9F
	.dc.b	$E4,$BB,$6B,$FE,$EE,$EB,$FE,$7B,$39,$97,$1B,$0E,$71,$FE,$BF,$C3
	.dc.b	$9F,$C2,$7F,$B9,$B0,$E2,$FD,$3A,$DD,$7F,$A6,$2B,$6F,$81,$AF,$68
	.dc.b	$FE,$17,$F1,$3C,$A3,$CB,$DD,$FE,$C5,$7F,$C2,$17,$FF,$00,$23,$10
	.dc.b	$CB,$D9,$9F,$FA,$97,$D7,$E4,$FF,$A9,$F0,$B7,$C2,$B6,$DC,$3A,$5F
	.dc.b	$87,$D5,$7D,$0F,$FA,$EB,$2F,$09,$F2,$39,$3F,$81,$BE,$1B,$FF,$74
	.dc.b	$67,$F8,$9F,$0B,$95,$F1,$AD,$1F,$E2,$37,$19,$FA,$30,$E6,$DA,$E3
	.dc.b	$F1,$DE,$AF,$39,$F2,$94,$98,$F9,$32,$FF,$78,$FC,$9E,$23,$BC,$7F
	.dc.b	$BD,$27,$18,$FC,$6B,$5E,$34,$F4,$C7,$6C,$3A,$35,$E6,$F0,$7C,$1B
	.dc.b	$FF,$BA,$DB,$F1,$9F,$38,$F9,$FF,$23,$23,$24,$3C,$B3,$C5,$BA,$7D
	.dc.b	$62,$3D,$BE,$36,$BA,$3C,$31,$F7,$23,$F1,$11,$F9,$8E,$9F,$C0,$FC
	.dc.b	$DF,$F9,$17,$5F,$C2,$B7,$4F,$E1,$13,$C3,$A7,$F5,$55,$AF,$7F,$89
	.dc.b	$99,$98,$7C,$FC,$39,$3F,$AA,$6B,$E1,$3F,$0F,$E8,$BE,$83,$FC,$5D
	.dc.b	$87,$EC,$63,$1E,$D7,$E7,$DB,$5C,$3E,$85,$E7,$E7,$0E,$3A,$FF,$1F
	.dc.b	$E5,$FF,$15,$E5,$FB,$7D,$BF,$0B,$D5,$E0,$BD,$E4,$CF,$AB,$F9,$23
	.dc.b	$F7,$B8,$72,$C7,$A5,$F1,$D3,$C0,$7C,$59,$D5,$0E,$33,$B7,$F4,$A6
	.dc.b	$0E,$D8,$F9,$1B,$3C,$B0,$F1,$E5,$E9,$8E,$B3,$C6,$31,$7D,$63,$8B
	.dc.b	$DA,$FA,$4E,$A7,$F5,$B7,$D3,$1F,$5C,$70,$24,$FD,$1C,$BE,$2C,$77
	.dc.b	$9D,$CF,$33,$59,$7C,$1C,$7C,$19,$79,$B2,$F0,$DF,$C3,$CB,$C1,$8F
	.dc.b	$C5,$9C,$B6,$B8,$1F,$07,$2F,$0C,$EA,$7C,$46,$F0,$F8,$19,$78,$F1
	.dc.b	$F2,$92,$7E,$E5,$F2,$C3,$AD,$F9,$8E,$35,$BA,$EA,$76,$C7,$CE,$E3
	.dc.b	$1F,$11,$E8,$8C,$B8,$BE,$B3,$F3,$B0,$F0,$61,$DE,$3E,$6E,$1F,$0E
	.dc.b	$73,$18,$3D,$87,$1C,$BE,$63,$93,$C9,$1C,$32,$F8,$11,$C1,$FA,$6F
	.dc.b	$77,$2F,$09,$E7,$18,$C7,$D5,$5C,$EF,$E1,$46,$2B,$71,$8C,$7E,$A4
	.dc.b	$63,$0F,$04,$FC,$FC,$BE,$1D,$7C,$2C,$7C,$B1,$8C,$BC,$95,$BC,$BC
	.dc.b	$18,$72,$25,$6D,$F8,$19,$78,$DE,$26,$CF,$61,$83,$EB,$63,$E6,$7E
	.dc.b	$81,$FE,$A9,$8F,$D7,$36,$3F,$D5,$ED,$C3,$FD,$59,$F0,$2F,$A0,$7D
	.dc.b	$0E,$1B,$5F,$EB,$33,$FD,$E2,$E1,$FF,$5B,$4F,$19,$FF,$5B,$60,$F7
	.dc.b	$9E,$6E,$1B,$7C,$E7,$53,$92,$47,$F0,$69,$DE,$DF,$C9,$7E,$87,$FA
	.dc.b	$AF,$9D,$FB,$F9,$CF,$07,$F8,$FD,$B7,$FA,$F6,$EE,$BE,$33,$D7,$1E
	.dc.b	$44,$7F,$83,$C6,$B1,$FE,$55,$7D,$27,$FC,$42,$DC,$BD,$31,$C3,$B7
	.dc.b	$FC,$A7,$5C,$30,$E8,$D9,$D6,$78,$CF,$D1,$56,$7F,$9A,$65,$E1,$9C
	.dc.b	$C7,$9C,$97,$F0,$F0,$EF,$3A,$8E,$C4,$7F,$89,$57,$3A,$F8,$97,$DD
	.dc.b	$F5,$BF,$D5,$23,$51,$83,$BC,$F5,$39,$DF,$82,$39,$9A,$2D,$F9,$70
	.dc.b	$ED,$8F,$C6,$8F,$24,$FB,$AB,$93,$DE,$32,$DB,$EF,$7C,$B2,$FA,$07
	.dc.b	$DC,$C7,$CF,$8F,$5A,$F3,$5F,$89,$E2,$F9,$4E,$86,$4F,$15,$F5,$CB
	.dc.b	$E9,$67,$F8,$7C,$FC,$B8,$75,$BE,$07,$29,$D5,$F1,$8F,$85,$18,$7C
	.dc.b	$13,$C5,$E1,$8F,$A3,$2F,$5E,$3E,$08,$C3,$CF,$3F,$95,$5C,$AB,$6F
	.dc.b	$E2,$8C,$4F,$1C,$BC,$F3,$AA,$F9,$F8,$F8,$DF,$9B,$8F,$8C,$F6,$46
	.dc.b	$73,$F9,$38,$FC,$63,$CF,$87,$4C,$7C,$93,$F3,$30,$EB,$97,$A2,$FC
	.dc.b	$91,$8B,$E1,$18,$AF,$AB,$87,$C5,$CF,$FC,$11,$3F,$48,$FF,$5D,$95
	.dc.b	$CC,$6E,$3C,$13,$1F,$F0,$D7,$D0,$BE,$87,$F2,$95,$EB,$9F,$F0,$A7
	.dc.b	$2E,$70,$FF,$1B,$7C,$37,$E4,$23,$13,$FE,$64,$E8,$4E,$70,$FF,$66
	.dc.b	$DF,$9A,$BF,$C6,$31,$87,$D3,$7C,$A7,$4D,$8C,$CF,$39,$FF,$3A,$79
	.dc.b	$0F,$A1,$C5,$7F,$9D,$3E,$01,$D9,$FA,$DF,$11,$18,$C3,$FC,$F8,$F1
	.dc.b	$6C,$EE,$35,$5C,$E3,$17,$97,$5F,$43,$BF,$36,$E1,$D1,$B7,$E3,$6F
	.dc.b	$1F,$FA,$AB,$14,$79,$7E,$81,$F4,$34,$C7,$05,$C7,$FD,$25,$3A,$6C
	.dc.b	$E9,$A4,$C7,$FD,$2B,$3B,$6D,$73,$7F,$D4,$FD,$AF,$A1,$7C,$D7,$C7
	.dc.b	$F2,$BF,$A8,$63,$FE,$B4,$AD,$4E,$7B,$7F,$8C,$46,$27,$57,$C9,$FF
	.dc.b	$54,$79,$8F,$3B,$73,$9C,$BE,$B7,$87,$F3,$CF,$A1,$DB,$70,$FA,$B1
	.dc.b	$67,$EE,$4D,$E2,$FF,$75,$74,$BD,$56,$1F,$EE,$9F,$84,$FC,$A4,$63
	.dc.b	$0F,$F7,$34,$62,$7F,$B9,$47,$FA,$0A,$7F,$4E,$FE,$49,$AF,$E8,$1A
	.dc.b	$EE,$3F,$CE,$DE,$AF,$EB,$AF,$C5,$18,$3F,$DA,$FF,$40,$F8,$CF,$15
	.dc.b	$6F,$0B,$5F,$ED,$CF,$75,$F4,$38,$C9,$DC,$9E,$55,$D6,$73,$1C,$23
	.dc.b	$13,$BC,$79,$3C,$6E,$D7,$2B,$63,$F1,$6B,$58,$72,$AD,$4F,$F0,$AF
	.dc.b	$33,$FB,$FB,$BE,$77,$CF,$79,$F1,$56,$B0,$FF,$6F,$CE,$63,$13,$FE
	.dc.b	$9D,$6E,$2E,$57,$AF,$BE,$3E,$9C,$38,$CB,$DB,$C2,$FA,$1C,$DF,$BA
	.dc.b	$31,$3F,$17,$3F,$F5,$EC,$FD,$8B,$E5,$1F,$E3,$16,$BF,$C3,$A3,$47
	.dc.b	$49,$EF,$C1,$67,$E3,$C7,$F2,$EF,$0D,$EA,$FC,$EF,$55,$F0,$BF,$0E
	.dc.b	$73,$5F,$52,$72,$F9,$63,$37,$D0,$FF,$60,$38,$E3,$FA,$9C,$9F,$E4
	.dc.b	$BA,$E1,$8F,$E1,$7C,$97,$D0,$BD,$7F,$16,$F9,$8C,$5F,$86,$31,$32
	.dc.b	$3E,$F7,$89,$74,$F1,$60,$F8,$97,$B5,$FA,$28,$C3,$63,$37,$FE,$5B
	.dc.b	$9C,$B6,$33,$5D,$1B,$AF,$E3,$FC,$8F,$C4,$D7,$E6,$EF,$8D,$FC,$3E
	.dc.b	$8F,$E6,$FD,$5F,$9A,$BD,$CE,$5F,$E2,$AD,$8D,$12,$FF,$47,$3A,$C3
	.dc.b	$C1,$3F,$27,$1F,$C0,$DF,$FB,$35,$AF,$FB,$ED,$B8,$FC,$3E,$5F,$80
	.dc.b	$71,$3C,$9F,$77,$72,$F5,$53,$DB,$F8,$C4,$8F,$88,$DC,$3E,$23,$ED
	.dc.b	$8F,$23,$1E,$6E,$5A,$48,$E2,$FB,$8E,$0F,$CF,$F6,$FE,$B1,$23,$EA
	.dc.b	$36,$35,$1F,$95,$F3,$FE,$24,$FE,$EF,$E7,$DF,$15,$FC,$4F,$D6,$7D
	.dc.b	$03,$D6,$D4,$E6,$31,$8F,$EC,$6F,$8B,$FC,$B9,$B9,$7F,$A8,$5A,$FF
	.dc.b	$80,$DA,$FF,$86,$9D,$F0,$FF,$86,$9B,$29,$7C,$B3,$E7,$5F,$51,$E4
	.dc.b	$AF,$E0,$5F,$3E,$F9,$1F,$E5,$99,$3E,$85,$F4,$3F,$C6,$6B,$57,$FC
	.dc.b	$B6,$B5,$39,$BF,$E8,$35,$AC,$FF,$D7,$4D,$7F,$C3,$8D,$7F,$C6,$6E
	.dc.b	$EB,$FC,$66,$DB,$F5,$67,$E9,$EF,$ED,$1E,$8C,$2E,$1F,$C1,$6B,$53
	.dc.b	$93,$F8,$F7,$D0,$BE,$83,$FA,$44,$BE,$85,$F1,$FF,$CD,$6B,$57,$FD
	.dc.b	$26,$B5,$39,$BF,$EB,$B5,$AC,$3F,$DA,$CD,$7F,$C7,$0D,$7F,$C9,$EE
	.dc.b	$F4,$FF,$27,$B7,$0F,$AD,$F4,$2F,$A1,$FF,$0F,$C7,$A3,$0F,$E9,$19
	.dc.b	$7C,$7C,$7E,$1E,$9F,$DB,$A3,$11,$F7,$A3,$1E,$C4,$BF,$EF,$35,$FE
	.dc.b	$0D,$6C,$7F,$76,$AF,$C3,$7A,$DF,$87,$F6,$5F,$17,$FE,$F1,$8C,$36
	.dc.b	$53,$A7,$FC,$AD,$3A,$7F,$21,$F1,$1F,$E5,$CD,$F1,$08,$C3,$CA,$FF
	.dc.b	$CE,$71,$8B,$E9,$3B,$F7,$BF,$B8,$BB,$3F,$39,$F3,$DF,$E3,$27,$FC
	.dc.b	$2A,$D7,$FC,$D0,$1C,$3F,$E6,$B6,$BF,$E8,$25,$F1,$FF,$E1,$E8,$E0
	.dc.b	$D8,$CC,$76,$8D,$7B,$EF,$9E,$FF,$45,$E7,$FE,$8A,$8C,$57,$F6,$D6
	.dc.b	$BF,$D6,$4F,$F3,$7F,$AF,$C3,$7F,$F4,$9C,$EA,$7F,$D0,$0D,$73,$7E
	.dc.b	$02,$3F,$DE,$7E,$17,$FC,$16,$DF,$6F,$A3,$FC,$06,$BC,$38,$FF,$A7
	.dc.b	$BE,$7B,$D7,$3E,$36,$C6,$73,$E3,$2F,$9E,$FB,$EF,$F8,$1E,$B7,$9F
	.dc.b	$FA,$A9,$B1,$FC,$3D,$AF,$1D,$F9,$C7,$82,$3F,$D3,$8D,$79,$5F,$D1
	.dc.b	$8C,$47,$38,$F1,$DF,$F7,$A8,$FF,$0C,$35,$F5,$47,$A3,$E7,$9E,$FD
	.dc.b	$24,$FF,$87,$FE,$61,$EB,$E2,$3F,$0F,$5C,$1F,$F2,$FB,$5C,$D7,$E1
	.dc.b	$63,$1A,$7F,$B2,$2F,$FC,$37,$5D,$7D,$CE,$93,$FE,$6B,$F4,$BF,$C0
	.dc.b	$DF,$3D,$F5,$1B,$5F,$D8,$A3,$EB,$FA,$CF,$75,$F5,$9B,$F5,$15,$FE
	.dc.b	$72,$C3,$8E,$1F,$EC,$56,$BF,$ED,$89,$DB,$FE,$0D,$F9,$FE,$44,$7F
	.dc.b	$C5,$DE,$0D,$7F,$C2,$11,$73,$FF,$2F,$65,$FD,$0E,$7F,$BB,$CE,$79
	.dc.b	$7F,$C0,$4F,$A2,$75,$3C,$8F,$01,$0F,$36,$09,$F5,$3F,$22,$3F,$C4
	.dc.b	$1A,$FB,$CE,$9C,$16,$3F,$C4,$D1,$C0,$8F,$EF,$B6,$F9,$1E,$D7,$8C
	.dc.b	$91,$FE,$4B,$F8,$5A,$FF,$89,$F2,$FF,$1B,$BF,$E4,$F9,$F5,$4F,$CF
	.dc.b	$E3,$93,$FE,$16,$78,$EB,$F6,$F3,$FC,$4C,$FF,$8E,$67,$D6,$47,$FC
	.dc.b	$7D,$F3,$EF,$9F,$FC,$03,$63,$FC,$74,$D7,$FC,$75,$F3,$4F,$9C,$DD
	.dc.b	$7C,$D9,$87,$07,$F6,$BF,$3E,$E7,$A4,$9F,$F4,$ED,$BE,$2F,$11,$FE
	.dc.b	$47,$F8,$57,$CF,$FE,$41,$B1,$FE,$83,$C3,$A3,$FE,$85,$F9,$7F,$23
	.dc.b	$73,$FE,$0B,$18,$8F,$F2,$97,$C5,$7F,$BD,$1D,$1F,$EC,$88,$D3,$FE
	.dc.b	$13,$3F,$C2,$8E,$1C,$35,$CC,$FF,$1A,$F7,$9F,$3F,$FD,$DA,$7F,$D7
	.dc.b	$91,$8F,$53,$C7,$5F,$C6,$3E,$7F,$55,$FC,$BD,$BC,$9F,$ED,$D8,$46
	.dc.b	$3B,$3F,$C4,$DD,$3F,$EA,$DE,$5E,$58,$FE,$D3,$E3,$BF,$DD,$18,$7F
	.dc.b	$B2,$1A,$F5,$8F,$DC,$F1,$BE,$52,$F9,$52,$BE,$9C,$68,$C3,$F8,$08
	.dc.b	$C3,$F3,$FE,$7F,$E6,$91,$FF,$5D,$FD,$4D,$7F,$AE,$F1,$27,$FA,$C3
	.dc.b	$63,$E8,$46,$0F,$E1,$BF,$1F,$FD,$03,$D4,$FF,$2A,$4F,$FB,$A2,$31
	.dc.b	$F0,$BF,$4A,$BF,$79,$F1,$5F,$5E,$F7,$C1,$FC,$56,$31,$5F,$23,$C2
	.dc.b	$FF,$69,$E3,$FE,$0A,$8C,$7A,$FA,$8F,$F7,$AC,$E9,$61,$D2,$30,$DB
	.dc.b	$E1,$87,$F2,$3B,$E5,$E7,$BE,$83,$FD,$75,$8F,$73,$FC,$26,$E1,$EB
	.dc.b	$97,$F8,$6E,$31,$8F,$AB,$8D,$F5,$3D,$27,$35,$F8,$29,$FE,$0B,$7C
	.dc.b	$86,$5F,$EF,$1F,$1B,$87,$C8,$79,$BC,$97,$CA,$D2,$5F,$96,$FF,$A9
	.dc.b	$FC,$A6,$23,$A4,$7F,$56,$9C,$6B,$F3,$AD,$78,$CF,$F2,$B1,$DB,$0E
	.dc.b	$8D,$79,$BC,$1F,$07,$0F,$FB,$C1,$BF,$1D,$F3,$23,$FD,$61,$1A,$F9
	.dc.b	$3C,$19,$99,$19,$39,$E9,$C4,$8F,$6F,$8E,$AF,$B2,$FE,$5B,$F8,$09
	.dc.b	$F0,$4F,$E0,$E1,$CE,$E9,$C3,$63,$EE,$1A,$8F,$C1,$F4,$3E,$47,$71
	.dc.b	$FC,$2F,$BF,$F8,$65,$6C,$FF,$50,$38,$BE,$91,$FC,$04,$D1,$5C,$F0
	.dc.b	$7E,$7B,$E6,$FF,$CA,$2D,$F8,$1E,$F5,$F8,$FF,$51,$F4,$1F,$E3,$08
	.dc.b	$FC,$34,$63,$DD,$FD,$03,$63,$87,$D0,$FF,$48,$77,$5C,$73,$FE,$49
	.dc.b	$5B,$9F,$E2,$CD,$8C,$CE,$E7,$FD,$79,$3B,$C7,$F7,$71,$F3,$39,$FF
	.dc.b	$C6,$F1,$8D,$3F,$2B,$DB,$E9,$C3,$D1,$3F,$C0,$B0,$E5,$CF,$FC,$CB
	.dc.b	$0E,$2D,$BE,$1A,$7D,$69,$CE,$5F,$CE,$9A,$F8,$67,$6F,$DA,$CB,$C2
	.dc.b	$FE,$AC,$D1,$5C,$E3,$E3,$CB,$CB,$8F,$86,$B5,$87,$6B,$DC,$FC,$2A
	.dc.b	$DD,$78,$F0,$F0,$5E,$CF,$3E,$5E,$E9,$D1,$D2,$76,$7A,$1C,$13,$1F
	.dc.b	$95,$5C,$AF,$BC,$FA,$A7,$35,$CA,$3E,$EE,$5F,$1E,$31,$3A,$9C,$98
	.dc.b	$C7,$DD,$19,$3E,$15,$71,$9F,$14,$E8,$F8,$11,$99,$E1,$19,$96,$3E
	.dc.b	$F7,$CD,$19,$69,$AC,$3D,$79,$7C,$C8,$F8,$B7,$D1,$D6,$1D,$EF,$95
	.dc.b	$6F,$0E,$CF,$1A,$EA,$76,$8C,$E5,$F4,$F2,$F8,$13,$D0,$A7,$0A,$E9
	.dc.b	$97,$92,$B8,$4F,$C2,$38,$14,$E0,$FD,$5C,$7C,$D1,$EA,$C7,$E2,$63
	.dc.b	$E2,$C7,$BE,$3D,$AF,$77,$CA,$7F,$05,$97,$C1,$8C,$1F,$60,$F0,$57
	.dc.b	$BC,$E4,$F4,$C7,$CF,$3F,$6A,$31,$5F,$03,$0F,$14,$72,$AF,$43,$A3
	.dc.b	$DD,$1A,$7F,$8D,$47,$FA,$C9,$B3,$FE,$2D,$8C,$47,$FB,$AB,$E8,$5F
	.dc.b	$43,$FE,$E2,$8F,$F7,$84,$61,$FF,$72,$FC,$2A,$31,$1F,$CA,$A3,$13
	.dc.b	$FE,$39,$F8,$17,$9B,$49,$5F,$E1,$28,$CB,$63,$F8,$04,$63,$3F,$E1
	.dc.b	$DF,$43,$FD,$73,$87,$35,$2F,$F8,$15,$EF,$83,$F8,$ED,$EE,$73,$5B
	.dc.b	$C3,$F8,$E5,$EE,$7F,$3F,$5B,$9C,$D6,$EF,$FC,$9B,$7B,$D9,$7B,$9C
	.dc.b	$D6,$F0,$FF,$2A,$DE,$EB,$51,$89,$D3,$FD,$6A,$75,$DC,$CB,$3C,$09
	.dc.b	$F3,$56,$EF,$FA,$56,$1D,$23,$07,$B8,$99,$FC,$8C,$7A,$46,$32,$F3
	.dc.b	$B4,$F8,$31,$A9,$F8,$31,$93,$AB,$CF,$2F,$2E,$3E,$4C,$BC,$D9,$79
	.dc.b	$30,$E5,$8F,$6C,$39,$91,$FA,$B1,$AC,$7E,$21,$DE,$F9,$BF,$D9,$32
	.dc.b	$F7,$46,$2F,$94,$F0,$7C,$8F,$1C,$7C,$87,$CC,$CB,$CF,$97,$D5,$CF
	.dc.b	$EC,$E5,$E1,$BF,$6E,$7E,$72,$3F,$12,$F8,$BF,$1E,$B8,$C6,$31,$F1
	.dc.b	$C6,$27,$C5,$87,$7C,$FD,$39,$79,$23,$19,$78,$73,$F6,$67,$E0,$C3
	.dc.b	$C6,$78,$4E,$67,$87,$0E,$77,$E1,$9F,$97,$9D,$AF,$89,$18,$78,$E7
	.dc.b	$E6,$BE,$31,$89,$D4,$E7,$1F,$26,$5F,$26,$BC,$11,$8C,$7E,$F6,$3E
	.dc.b	$A2,$4E,$9C,$BF,$E1,$CF,$A1,$7C,$BF,$DF,$D7,$E3,$65,$F6,$09,$9F
	.dc.b	$F8,$52,$B7,$87,$29,$FF,$0A,$F9,$B8,$C7,$FC,$64,$D9,$FF,$1C,$B6
	.dc.b	$3F,$D8,$CD,$41,$E5,$3D,$31,$FC,$97,$84,$EC,$7D,$0F,$E1,$6F,$79
	.dc.b	$F4,$BC,$1C,$2B,$79,$7F,$94,$2B,$79,$7F,$A0,$3E,$6B,$FC,$AD,$3F
	.dc.b	$E7,$B6,$C7,$F9,$E9,$E2,$DE,$DF,$1C,$6B,$4F,$F5,$4B,$72,$FF,$67
	.dc.b	$BC,$23,$33,$C3,$E8,$78,$F9,$D7,$1F,$03,$71,$F0,$36,$FA,$BF,$0B
	.dc.b	$E8,$5F,$3B,$4C,$7C,$6D,$C3,$FD,$29,$3A,$6C,$E9,$A4,$C3,$FD,$2F
	.dc.b	$3B,$6D,$73,$7F,$D4,$DD,$AF,$A0,$7D,$0F,$F0,$58,$C3,$F5,$BC,$D7
	.dc.b	$A1,$9A,$DE,$5F,$EB,$BB,$DB,$63,$38,$7F,$AA,$2F,$8C,$E7,$0F,$9F
	.dc.b	$E0,$C8,$FA,$DD,$76,$FE,$B4,$DD,$FE,$9C,$DE,$3F,$C7,$FC,$1F,$FA
	.dc.b	$16,$FF,$1F,$CA,$8C,$47,$E4,$5B,$19,$C7,$FB,$73,$65,$EC,$7F,$38
	.dc.b	$78,$4F,$E6,$A3,$F9,$DF,$53,$DB,$7D,$0F,$F3,$E9,$E1,$87,$36,$E3
	.dc.b	$9C,$3C,$E7,$33,$F4,$B8,$73,$BD,$63,$EA,$AB,$2F,$79,$F4,$38,$7C
	.dc.b	$38,$73,$BF,$F6,$FC,$6A,$31,$5C,$70,$E1,$7C,$5B,$3A,$AE,$77,$F0
	.dc.b	$A3,$F7,$2B,$C8,$FE,$F9,$97,$C5,$6C,$FF,$BC,$7D,$7F,$4D,$6F,$7B
	.dc.b	$5A,$8F,$E0,$6D,$CD,$C6,$BD,$DF,$2D,$B1,$8F,$D6,$9F,$B2,$47,$07
	.dc.b	$C1,$C7,$F8,$74,$7C,$8D,$FE,$14,$7A,$CF,$C2,$C7,$52,$4F,$C6,$9D
	.dc.b	$E7,$F6,$A7,$D7,$DC,$F4,$E9,$F4,$7C,$AB,$87,$86,$74,$D9,$D3,$4F
	.dc.b	$AD,$E2,$3E,$61,$E9,$7E,$02,$30,$FD,$DE,$FC,$4E,$AB,$8B,$75,$F6
	.dc.b	$5F,$09,$FB,$55,$C2,$FA,$7C,$33,$D7,$A7,$E3,$3C,$5F,$F7,$A5,$F9
	.dc.b	$27,$4D,$9D,$BF,$8A,$9D,$E3,$E0,$AD,$BF,$37,$91,$3A,$E0,$FD,$3B
	.dc.b	$6F,$F9,$4F,$95,$E1,$F4,$E4,$75,$7C,$91,$8C,$3B,$71,$7A,$8F,$47
	.dc.b	$D0,$DB,$F4,$45,$9F,$16,$0A,$C6,$3F,$E3,$2D,$DA,$3D,$7F,$8A,$7C
	.dc.b	$3C,$A9,$F0,$DB,$8F,$48,$B7,$CE,$B9,$46,$CF,$8C,$68,$FF,$7F,$BA
	.dc.b	$9F,$91,$5D,$E7,$84,$7C,$BF,$5B,$EB,$9F,$C0,$3D,$7F,$EE,$18,$77
	.dc.b	$8C,$D7,$D7,$75,$E6,$62,$FF,$3C,$3E,$7D,$F4,$3F,$C8,$9B,$7F,$E8
	.dc.b	$D6,$CF,$EF,$8F,$EE,$EE,$1C,$3A,$C3,$F9,$23,$67,$FB,$EF,$CB,$3E
	.dc.b	$81,$F4,$1F,$C0,$AB,$58,$74,$C7,$07,$F1,$EB,$E1,$58,$C3,$F7,$5F
	.dc.b	$3C,$F9,$0F,$E7,$2D,$8F,$F5,$23,$70,$F4,$07,$0F,$AD,$D6,$7F,$CE
	.dc.b	$DB,$3E,$8C,$FF,$64,$F8,$2F,$F8,$CC,$EA,$FF,$C3,$2D,$C7,$DD,$D7
	.dc.b	$AA,$97,$C4,$FE,$65,$3A,$71,$7F,$53,$E7,$5F,$37,$F0,$7E,$79,$F1
	.dc.b	$1F,$CA,$27,$37,$CA,$72,$FF,$1D,$8C,$DF,$3B,$FE,$8F,$F3,$CF,$9E
	.dc.b	$7A,$FC,$E6,$61,$C5,$DC,$EA,$30,$7F,$29,$CF,$E0,$60,$7E,$17,$7B
	.dc.b	$EB,$3F,$A1,$56,$F6,$FE,$7F,$5F,$DF,$2F,$F2,$2D,$E2,$FF,$50,$FB
	.dc.b	$AF,$9E,$FC,$1D,$75,$F6,$3F,$A3,$BF,$E4,$6F,$53,$FC,$8F,$EC,$3E
	.dc.b	$17,$EE,$C6,$1B,$29,$D3,$F9,$99,$D3,$FE,$50,$F8,$37,$B7,$FD,$F2
	.dc.b	$38,$5F,$F5,$C9,$CE,$7C,$25,$18,$76,$73,$AE,$58,$78,$21,$F3,$FF
	.dc.b	$E2,$87,$FA,$93,$5F,$D9,$B6,$FF,$C1,$9F,$46,$FB,$4F,$ED,$B1,$FE
	.dc.b	$0F,$97,$CF,$63,$8B,$FB,$8C,$63,$0E,$B3,$95,$C8,$FA,$D7,$8B,$99
	.dc.b	$FF,$6D,$7D,$15,$F3,$4D,$35,$F0,$9F,$56,$7F,$B5,$4E,$9B,$3B,$7F
	.dc.b	$B7,$4E,$DF,$E1,$F3,$B7,$F8,$8C,$7F,$81,$DD,$FB,$5C,$5F,$E2,$BF
	.dc.b	$45,$EB,$6E,$7E,$78,$C6,$DF,$E3,$1F,$A2,$FB,$D5,$FD,$F6,$73,$BF
	.dc.b	$F8,$51,$B1,$FE,$8E,$6A,$9E,$2F,$63,$95,$F7,$3E,$34,$6A,$FF,$C8
	.dc.b	$15,$E5,$AF,$E4,$8D,$7C,$15,$FC,$41,$F0,$35,$7D,$13,$E4,$3E,$93
	.dc.b	$63,$FC,$2C,$D7,$F9,$BF,$CE,$7F,$A9,$5C,$CF,$02,$3F,$E1,$D8,$D3
	.dc.b	$FC,$FA,$79,$3F,$CE,$5A,$FF,$8E,$1A,$FF,$AC,$9B,$E8,$FF,$62,$FA
	.dc.b	$1F,$C5,$36,$7F,$C0,$51,$F8,$DE,$3F,$C5,$FD,$16,$BD,$34,$D2,$57
	.dc.b	$F9,$3B,$91,$5B,$FA,$29,$3F,$E0,$D7,$A4,$FF,$3D,$8C,$10,$FF,$06
	.dc.b	$46,$BC,$3E,$31,$83,$5C,$1F,$E7,$2A,$E5,$19,$24,$7F,$75,$AE,$51
	.dc.b	$FE,$1D,$FA,$2F,$51,$1F,$46,$F5,$FF,$0E,$45,$AE,$D3,$E8,$F6,$A5
	.dc.b	$7B,$DF,$5F,$C3,$E2,$3F,$C4,$58,$FF,$8B,$DD,$1F,$16,$3E,$57,$C1
	.dc.b	$48,$FF,$18,$3C,$70,$FF,$06,$39,$8F,$F1,$73,$9E,$6F,$F7,$5F,$0B
	.dc.b	$E8,$9F,$45,$F6,$9B,$7F,$E4,$E6,$BF,$E3,$5F,$9C,$7C,$F6,$EA,$4E
	.dc.b	$DF,$C0,$4E,$A3,$FC,$D9,$F4,$0E,$8F,$E1,$8D,$8C,$C7,$0F,$A2,$CB
	.dc.b	$23,$27,$FA,$4A,$93,$FC,$57,$E5,$FB,$4F,$F8,$61,$BB,$7F,$AC,$5B
	.dc.b	$AF,$FA,$9B,$E2,$7F,$DC,$B5,$FE,$7F,$9C,$A0,$E1,$BA,$7E,$6E,$78
	.dc.b	$C7,$78,$FD,$C4,$F8,$A3,$FC,$44,$D7,$FD,$42,$D7,$B7,$D0,$7F,$78
	.dc.b	$7F,$D2,$0D,$7F,$C6,$2D,$C3,$FD,$8F,$F1,$9F,$B8,$7F,$D9,$91,$9C
	.dc.b	$7F,$D5,$AD,$AE,$BE,$CF,$EC,$6F,$85,$7F,$B1,$B4,$FE,$FB,$5F,$2F
	.dc.b	$C3,$FA,$E7,$D8,$F5,$A4,$64,$87,$E0,$5F,$A1,$F3,$F2,$3E,$53,$F1
	.dc.b	$7E,$83,$11,$F4,$E2,$E9,$FE,$E0,$DD,$5A,$AF,$8F,$5F,$28,$8F,$F8
	.dc.b	$D3,$C1,$FE,$A7,$F0,$FF,$EC,$AF,$49,$F2,$5F,$EE,$0E,$CF,$F6,$05
	.dc.b	$FF,$80,$63,$15,$B7,$5E,$A5,$9F,$EB,$F5,$C1,$5F,$47,$37,$BC,$3C
	.dc.b	$31,$D4,$E3,$7C,$FD,$17,$CF,$FF,$AC,$A7,$51,$8F,$8B,$D4,$E6,$7B
	.dc.b	$E7,$FE,$F8,$E4,$7D,$06,$AE,$57,$2B,$E1,$A7,$97,$3F,$F6,$FD,$F9
	.dc.b	$7C,$37,$CA,$D2,$61,$E7,$AF,$E9,$5F,$25,$88,$F0,$9F,$E0,$0A,$E2
	.dc.b	$ED,$AE,$A7,$8B,$9A,$E5,$87,$8E,$B6,$DB,$F9,$1C,$1F,$ED,$FB,$F1
	.dc.b	$3F,$EA,$76,$FC,$5E,$BE,$42,$46,$08,$78,$67,$93,$75,$EA,$47,$AF
	.dc.b	$C6,$57,$47,$82,$3C,$71,$F8,$18,$FD,$EF,$4D,$EE,$34,$FC,$07,$C8
	.dc.b	$BE,$43,$15,$F8,$13,$89,$6B,$8B,$F3,$E7,$83,$9A,$FE,$18,$BB,$BF
	.dc.b	$82,$C6,$1B,$1F,$C4,$5A,$EC,$FE,$6F,$F3,$DF,$E2,$37,$EC,$F9,$08
	.dc.b	$C7,$07,$E6,$63,$1E,$B7,$E7,$D7,$CF,$FF,$80,$27,$55,$C3,$A7,$AD
	.dc.b	$EE,$31,$83,$0E,$3E,$5E,$6B,$E4,$73,$7F,$8B,$31,$FD,$45,$FF,$28
	.dc.b	$8C,$63,$D2,$3F,$51,$87,$91,$F9,$13,$A7,$F9,$86,$DF,$9A,$EB,$DE
	.dc.b	$7E,$2B,$E1,$5F,$98,$3B,$4F,$23,$79,$79,$1C,$46,$EF,$A3,$F2,$2F
	.dc.b	$DA,$F2,$35,$87,$8D,$C6,$5F,$1B,$0E,$91,$AC,$3A,$CE,$62,$E5,$FC
	.dc.b	$FA,$F9,$63,$E0,$AD,$57,$12,$3F,$22,$35,$5F,$1E,$72,$F3,$BE,$B1
	.dc.b	$F4,$F1,$F4,$D7,$08,$F9,$38,$78,$0F,$99,$97,$C8,$C7,$B4,$67,$0E
	.dc.b	$73,$9C,$7B,$5F,$80,$E7,$3F,$31,$F0,$9F,$12,$37,$7E,$C3,$AC,$EB
	.dc.b	$0F,$25,$F5,$BE,$67,$53,$91,$EB,$CB,$E6,$DF,$28,$F1,$4F,$17,$64
	.dc.b	$AE,$B9,$7C,$2C,$BD,$93,$C2,$31,$8F,$8E,$7E,$1C,$62,$7C,$55,$CB
	.dc.b	$1F,$0E,$5E,$C2,$4F,$94,$E5,$87,$C8,$BE,$59,$79,$C9,$5F,$65,$B5
	.dc.b	$F0,$1B,$97,$AF,$1F,$B1,$3E,$BC,$BC,$B9,$78,$A3,$86,$1E,$2B,$E1
	.dc.b	$5E,$EC,$39,$E3,$E2,$AF,$3D,$FC,$CB,$DE,$5E,$88,$FD,$06,$3D,$5F
	.dc.b	$04,$7D,$D7,$F2,$87,$F8,$63,$1F,$F4,$FB,$F8,$43,$11,$FE,$97,$F7
	.dc.b	$DF,$28,$F9,$FC,$36,$BF,$8F,$9F,$EE,$97,$0E,$27,$8C,$FF,$A8,$75
	.dc.b	$7B,$6F,$84,$C3,$6F,$99,$FE,$34,$70,$FF,$8B,$9A,$FF,$8B,$9E,$1B
	.dc.b	$FE,$B5,$BB,$7F,$98,$3E,$7D,$D9,$FC,$B6,$B5,$87,$F3,$BB,$DE,$1F
	.dc.b	$C5,$1B,$C0,$C3,$8F,$26,$EE,$5F,$F0,$FB,$DD,$6A,$31,$3A,$C3,$BC
	.dc.b	$7F,$3A,$9D,$46,$09,$97,$AE,$BF,$94,$CE,$A3,$15,$D0,$95,$FD,$06
	.dc.b	$75,$1E,$9A,$F0,$57,$0A,$D6,$5F,$06,$30,$4A,$E0,$79,$63,$4E,$27
	.dc.b	$CA,$D9,$F9,$93,$A8,$FA,$D5,$E0,$74,$7F,$4B,$BF,$2B,$A3,$D3,$5E
	.dc.b	$1C,$BE,$75,$F3,$CF,$F0,$B8,$73,$BE,$18,$FA,$B3,$C5,$6A,$79,$9D
	.dc.b	$EF,$96,$7F,$72,$38,$E5,$DF,$3F,$1D,$F1,$C7,$DB,$5A,$39,$63,$DB
	.dc.b	$0E,$55,$BC,$3E,$0C,$E7,$2F,$81,$9F,$AA,$31,$87,$92,$30,$F9,$8E
	.dc.b	$C7,$77,$38,$F8,$B0,$ED,$3F,$2A,$73,$87,$D7,$8E,$33,$A9,$CC,$7D
	.dc.b	$03,$38,$FE,$53,$3F,$87,$97,$93,$1F,$95,$5E,$3C,$3B,$BE,$3C,$FE
	.dc.b	$2E,$1F,$00,$97,$F9,$4C,$39,$D7,$19,$ED,$9F,$C2,$BE,$37,$B8,$C5
	.dc.b	$F0,$C7,$C2,$7E,$33,$2F,$49,$3E,$7D,$F2,$9F,$92,$8C,$37,$1F,$CF
	.dc.b	$C2,$77,$87,$7C,$FF,$C4,$F1,$87,$F3,$91,$FE,$14,$8C,$61,$E2,$D7
	.dc.b	$FD,$CB,$18,$6E,$3F,$A2,$F0,$1F,$30,$E5,$F2,$5E,$F2,$FF,$65,$73
	.dc.b	$BD,$5F,$F1,$D7,$97,$FE,$3B,$C5,$D9,$FE,$DA,$DB,$FC,$99,$7B,$7F
	.dc.b	$DA,$55,$B9,$D6,$1F,$E2,$1A,$D2,$9D,$7A,$3F,$B1,$9C,$C6,$2F,$AE
	.dc.b	$3F,$E4,$BB,$FE,$6B,$5C,$9B,$19,$7F,$CA,$F1,$97,$FD,$C9,$8F,$C0
	.dc.b	$F8,$8F,$5D,$E3,$57,$C1,$62,$D7,$BA,$F5,$9B,$9D,$46,$67,$D5,$39
	.dc.b	$36,$F2,$A5,$F0,$7B,$CE,$AB,$E9,$FC,$D6,$7E,$6B,$4D,$9D,$52,$31
	.dc.b	$87,$80,$EC,$E3,$4F,$AD,$F3,$AF,$96,$F3,$BF,$ED,$8C,$3C,$0E,$5B
	.dc.b	$E8,$EA,$76,$D8,$C9,$C2,$32,$D3,$38,$78,$CF,$C6,$74,$FC,$6F,$2F
	.dc.b	$FD,$25,$3F,$12,$72,$FC,$18,$CC,$71,$E5,$FE,$A0,$FF,$72,$6B,$FE
	.dc.b	$92,$6A,$39,$9F,$DE,$36,$FC,$EF,$C8,$7C,$FF,$AF,$FA,$0D,$C3,$E5
	.dc.b	$B6,$BE,$55,$70,$9F,$9F,$5F,$83,$97,$1B,$D2,$FC,$0F,$A9,$F7,$2F
	.dc.b	$D9,$39,$7E,$E4,$66,$BE,$24,$E5,$EF,$19,$E0,$FC,$E4,$E5,$FD,$74
	.dc.b	$66,$BF,$37,$7F,$9D,$BF,$DA,$B5,$F2,$B6,$FB,$47,$FA,$E3,$1F,$DF
	.dc.b	$FD,$7F,$E4,$7D,$DF,$C9,$37,$0F,$C6,$B6,$BE,$1F,$22,$31,$9F,$EC
	.dc.b	$7A,$9E,$87,$E8,$BD,$07,$C0,$7C,$79,$CE,$5F,$9D,$9C,$BF,$81,$8C
	.dc.b	$E5,$FE,$24,$6B,$FE,$30,$6B,$FE,$D0,$6B,$F7,$FE,$F7,$C9,$3D,$7E
	.dc.b	$D7,$EB,$46,$27,$ED,$DA,$73,$9F,$9D,$B7,$C4,$FC,$A5,$F1,$BF,$F2
	.dc.b	$0B,$5F,$F7,$0B,$5F,$15,$7D,$3F,$11,$F3,$0F,$53,$E2,$71,$7F,$0B
	.dc.b	$9C,$C6,$31,$FE,$0B,$18,$7F,$C8,$AD,$C3,$FC,$A2,$D7,$FD,$E2,$D7
	.dc.b	$F4,$0F,$83,$E1,$FE,$97,$AF,$DF,$14,$62,$FB,$C6,$1E,$ED,$BE,$ED
	.dc.b	$7B,$B5,$EE,$BE,$3F,$F9,$37,$C9,$FF,$45,$F1,$7C,$11,$87,$F9,$CD
	.dc.b	$F0,$8F,$8D,$5C,$63,$C3,$18,$3F,$39,$CF,$FD,$53,$E0,$3F,$93,$1F
	.dc.b	$C9,$FC,$0B,$5F,$C8,$A2,$E5,$F9,$69,$BC,$EA,$F3,$F3,$7E,$57,$E0
	.dc.b	$FE,$23,$63,$E1,$BE,$42,$3F,$11,$F3,$F2,$FF,$3E,$FA,$CF,$B9,$3F
	.dc.b	$2A,$31,$C7,$F3,$DD,$C7,$CC,$6E,$5F,$31,$D4,$7F,$3A,$DA,$CF,$F3
	.dc.b	$9E,$3F,$C4,$D6,$33,$FE,$DD,$D5,$F8,$CC,$3C,$07,$DA,$E4,$FD,$87
	.dc.b	$27,$DF,$6C,$64,$FB,$AF,$D0,$23,$FE,$92,$C5,$F3,$4F,$8A,$FE,$C1
	.dc.b	$8B,$EA,$5F,$53,$FC,$86,$B5,$87,$F5,$CA,$D4,$E6,$F8,$DF,$0A,$FE
	.dc.b	$7D,$7C,$30,$FE,$AD,$87,$09,$CD,$FF,$19,$AD,$57,$FA,$C1,$AF,$F3
	.dc.b	$D3,$CF,$1F,$C7,$FE,$6D,$E2,$7E,$4B,$E7,$1C,$FF,$CD,$E7,$38,$FF
	.dc.b	$72,$EB,$56,$A3,$FA,$94,$EB,$1F,$ED,$55,$AB,$E5,$87,$56,$DF,$08
	.dc.b	$FF,$03,$7C,$E3,$E7,$1F,$13,$AC,$7F,$B1,$65,$D7,$6F,$F0,$C3,$71
	.dc.b	$FE,$B1,$87,$1B,$E1,$5F,$E1,$DF,$5D,$F6,$5F,$C2,$7E,$F9,$F5,$7F
	.dc.b	$D4,$E7,$38,$FF,$0A,$9C,$BF,$DB,$63,$39,$FE,$36,$72,$FF,$75,$8C
	.dc.b	$E1,$FA,$78,$74,$AB,$73,$E3,$C7,$C0,$7F,$89,$7D,$B7,$DD,$BE,$93
	.dc.b	$39,$FF,$7D,$E3,$7C,$6E,$FE,$BB,$9C,$FD,$B9,$CB,$FE,$26,$C1,$3A
	.dc.b	$D7,$FB,$E5,$7E,$C1,$FC,$4F,$DF,$F0,$7F,$C1,$D1,$9D,$3F,$CA,$B3
	.dc.b	$97,$FB,$37,$C0,$7F,$81,$7E,$19,$F4,$5F,$EF,$28,$C3,$70,$4E,$9F
	.dc.b	$F8,$03,$3A,$7F,$C1,$DF,$0C,$FA,$2F,$F2,$C4,$70,$BF,$CB,$4E,$70
	.dc.b	$E1,$28,$C3,$B3,$9D,$72,$C3,$C1,$0F,$A1,$7D,$9F,$E8,$70,$FF,$30
	.dc.b	$7D,$1F,$F6,$EA,$FF,$15,$35,$FF,$18,$57,$07,$FC,$5C,$BE,$AB,$97
	.dc.b	$0F,$EA,$E9,$E6,$E2,$FF,$C5,$51,$87,$8D,$65,$7D,$DB,$E3,$70,$D8
	.dc.b	$FF,$1E,$35,$FF,$21,$6A,$F8,$0D,$D7,$38,$FD,$3C,$FF,$90,$DA,$FF
	.dc.b	$92,$6B,$96,$9F,$E7,$F6,$E7,$FC,$0E,$B7,$3F,$E8,$36,$C7,$F0,$79
	.dc.b	$DD,$7F,$91,$7E,$2F,$FC,$FA,$FF,$94,$5A,$FF,$95,$DB,$7F,$E8,$DF
	.dc.b	$91,$FF,$3B,$CE,$FE,$23,$E1,$56,$B0,$FE,$28,$D9,$FF,$29,$57,$07
	.dc.b	$A5,$7C,$98,$FF,$57,$47,$F9,$85,$AE,$8E,$C6,$7E,$0B,$81,$E0,$9F
	.dc.b	$A5,$F4,$6F,$A3,$FE,$75,$48,$C3,$FC,$F5,$AF,$F2,$93,$FC,$CC,$E1
	.dc.b	$C3,$AC,$BF,$D5,$9F,$49,$FE,$1A,$9F,$F3,$B3,$5F,$F3,$E4,$EF,$3F
	.dc.b	$E6,$8D,$8F,$F3,$94,$EE,$B9,$FA,$9E,$8C,$3F,$1D,$E1,$BE,$7F,$F8
	.dc.b	$EE,$1F,$E2,$06,$BF,$E2,$06,$BF,$E2,$06,$BF,$E2,$07,$77,$FE,$D3
	.dc.b	$F8,$0F,$F0,$A4,$FF,$93,$FD,$CE,$F5,$AD,$38,$6D,$82,$57,$FA,$C7
	.dc.b	$E7,$18,$62,$3F,$BA,$5F,$D6,$C7,$F9,$36,$5E,$9C,$3F,$51,$EE,$B1
	.dc.b	$E3,$87,$F8,$F3,$0F,$BA,$FA,$6B,$FB,$85,$F1,$8F,$81,$18,$9E,$D9
	.dc.b	$7F,$B5,$BF,$00,$7C,$7F,$66,$DF,$5B,$F6,$F1,$71,$24,$7C,$B6,$BF
	.dc.b	$E6,$2F,$A2,$F3,$EC,$78,$C7,$C1,$FB,$4F,$5E,$A6,$98,$F6,$6C,$EC
	.dc.b	$87,$FA,$A2,$7F,$C0,$EF,$C1,$F9,$0F,$61,$1F,$BC,$D7,$FB,$7F,$98
	.dc.b	$E8,$FE,$25,$E4,$7F,$4C,$7E,$CD,$7D,$AF,$8C,$F8,$C4,$BF,$C4,$35
	.dc.b	$FF,$3F,$F9,$7E,$E3,$83,$F4,$FE,$2B,$E7,$11,$FC,$83,$5F,$F4,$39
	.dc.b	$F1,$8F,$C4,$BF,$C8,$7F,$01,$FD,$7C,$1A,$47,$F6,$6F,$EB,$FA,$3F
	.dc.b	$C7,$3E,$87,$E1,$BE,$03,$F2,$55,$FE,$D7,$AD,$73,$FF,$87,$5B,$0E
	.dc.b	$A5,$70,$6D,$F8,$E7,$F9,$A7,$D1,$FF,$37,$7F,$7C,$1C,$3F,$C0,$5A
	.dc.b	$FF,$8A,$DA,$FF,$BA,$5E,$58,$FF,$94,$3D,$5F,$41,$DB,$C9,$FC,$2F
	.dc.b	$D5,$FE,$1B,$E8,$A6,$1F,$A9,$8C,$BF,$AB,$8C,$BF,$85,$8F,$E9,$5F
	.dc.b	$32,$FA,$4F,$E3,$2D,$9F,$F2,$A3,$5F,$5B,$5F,$E3,$4E,$71,$FF,$0C
	.dc.b	$70,$BE,$CF,$FC,$83,$5F,$E0,$B6,$A9,$FF,$0F,$2A,$E5,$C3,$FE,$4D
	.dc.b	$6C,$2F,$A7,$C0,$7F,$65,$6C,$0E,$95,$C7,$1F,$D4,$FD,$97,$EA,$BC
	.dc.b	$69,$5F,$C3,$63,$2F,$F0,$F8,$CB,$FA,$98,$FE,$DD,$F4,$DE,$88,$FE
	.dc.b	$70,$D7,$FC,$C4,$D7,$C4,$D4,$FC,$0C,$FF,$C7,$9C,$6F,$A6,$7C,$7F
	.dc.b	$F1,$86,$DF,$AE,$BE,$2D,$F1,$CF,$DD,$83,$E5,$DF,$65,$FC,$81,$FE
	.dc.b	$DF,$87,$E5,$DB,$3B,$9D,$1B,$F8,$29,$5D,$EF,$AD,$F4,$8C,$1C,$EB
	.dc.b	$6F,$CE,$3B,$57,$47,$F9,$C3,$FC,$63,$BA,$46,$08,$7F,$15,$72,$FF
	.dc.b	$17,$8F,$E3,$FE,$A7,$F2,$0A,$93,$FD,$0E,$32,$FF,$45,$8C,$BF,$C7
	.dc.b	$E3,$F3,$7C,$DF,$9A,$24,$7E,$69,$AF,$A7,$27,$CA,$FF,$74,$F7,$5F
	.dc.b	$3F,$F3,$A3,$0F,$E4,$EF,$AD,$FC,$4B,$E3,$C7,$FE,$A3,$C3,$8C,$FF
	.dc.b	$07,$C3,$8B,$F9,$5E,$76,$1F,$DC,$3E,$8D,$F2,$BC,$73,$FC,$D7,$A2
	.dc.b	$9C,$E1,$F9,$DC,$BA,$CE,$6F,$84,$FF,$15,$F2,$1F,$45,$8A,$D4,$F1
	.dc.b	$C7,$AF,$D1,$FF,$87,$B1,$E6,$D8,$F0,$E5,$E0,$A5,$78,$23,$0F,$0F
	.dc.b	$43,$B5,$F1,$AF,$F4,$BD,$6A,$FB,$D6,$89,$3E,$33,$79,$E1,$EF,$3F
	.dc.b	$C0,$72,$FF,$40,$C6,$0F,$94,$E3,$1F,$14,$73,$C9,$97,$8E,$3F,$86
	.dc.b	$56,$EF,$A7,$C0,$6E,$B5,$B7,$C8,$8F,$25,$F0,$6C,$66,$3C,$58,$76
	.dc.b	$CF,$E8,$C6,$CF,$31,$72,$FC,$1B,$71,$F8,$6F,$4C,$BC,$8F,$B3,$1C
	.dc.b	$1E,$5B,$EB,$8F,$87,$0E,$6D,$C3,$D7,$7E,$4C,$BD,$78,$F4,$8F,$01
	.dc.b	$D4,$D6,$1D,$23,$57,$CE,$75,$87,$CC,$7D,$71,$97,$99,$D3,$2F,$84
	.dc.b	$FA,$0D,$D7,$43,$DB,$1A,$8C,$E5,$BB,$F5,$D6,$B1,$F4,$CE,$72,$F6
	.dc.b	$D7,$8C,$DC,$66,$FC,$F3,$F3,$1E,$A4,$9D,$BF,$00,$E0,$7A,$B1,$F8
	.dc.b	$18,$F9,$67,$64,$AF,$09,$D5,$C5,$78,$09,$3C,$67,$71,$93,$E3,$61
	.dc.b	$EC,$7C,$91,$A9,$E1,$1B,$BF,$9C,$79,$2B,$85,$F8,$27,$9E,$5F,$67
	.dc.b	$1F,$5D,$EE,$31,$3D,$8F,$2D,$F2,$C7,$CB,$1C,$8E,$39,$7D,$7C,$3F
	.dc.b	$21,$18,$7E,$76,$5E,$68,$F9,$13,$B7,$E6,$1D,$CC,$CF,$27,$E9,$9F
	.dc.b	$E0,$68,$FD,$DB,$63,$FC,$1B,$18,$8F,$F7,$BF,$B8,$FA,$3F,$F1,$4C
	.dc.b	$7F,$BD,$E3,$0F,$F8,$B3,$E0,$BF,$DC,$13,$A6,$C7,$F8,$5B,$DC,$FF
	.dc.b	$04,$37,$C2,$FF,$0D,$C6,$23,$FC,$32,$D7,$FC,$20,$F0,$DF,$F5,$ED
	.dc.b	$D3,$F9,$AF,$C4,$FF,$BA,$7B,$7F,$82,$CE,$6F,$F9,$F3,$5F,$F0,$E3
	.dc.b	$9D,$95,$C6,$7F,$5B,$3B,$E1,$34,$E1,$3B,$C7,$C7,$97,$80,$97,$FC
	.dc.b	$52,$75,$97,$AA,$73,$5C,$2B,$5D,$2C,$F3,$7B,$9F,$1D,$FF,$2B,$C3
	.dc.b	$9C,$7B,$23,$6D,$39,$D6,$1F,$7C,$6B,$3F,$7E,$1D,$E7,$A9,$31,$F2
	.dc.b	$1F,$D7,$30,$EF,$1C,$8D,$63,$E5,$9D,$57,$37,$8C,$64,$92,$BF,$0E
	.dc.b	$3E,$5C,$FE,$6E,$5E,$5C,$BB,$C6,$2F,$B9,$ED,$C7,$BD,$F1,$CF,$EA
	.dc.b	$61,$F1,$73,$F1,$67,$E0,$C3,$D1,$87,$91,$F1,$65,$DB,$0E,$31,$DB
	.dc.b	$3F,$2B,$E0,$38,$DF,$AA,$33,$7C,$E7,$27,$96,$7F,$27,$87,$8A,$30
	.dc.b	$43,$75,$CA,$7C,$D9,$FB,$EF,$CD,$39,$AE,$13,$9A,$F1,$63,$E9,$CB
	.dc.b	$E1,$1F,$07,$3B,$7E,$B9,$D3,$EB,$BE,$4F,$C2,$CB,$CC,$FA,$89,$1A
	.dc.b	$C7,$C5,$18,$CF,$E3,$E7,$F4,$A3,$B5,$F2,$BE,$18,$78,$2B,$58,$FF
	.dc.b	$A0,$EF,$C0,$FF,$A1,$4C,$63,$F4,$1B,$5F,$B3,$6E,$1F,$2E,$7E,$20
	.dc.b	$FA,$27,$D1,$E6,$FF,$4B,$39,$8C,$57,$1A,$DD,$FC,$7C,$3F,$C1,$31
	.dc.b	$87,$FD,$43,$7C,$67,$CD,$1F,$E4,$29,$D4,$7A,$1B,$19,$BE,$C7,$E0
	.dc.b	$7A,$FF,$DA,$BC,$B8,$BF,$F0,$DF,$37,$FA,$39,$FF,$2C,$75,$3D,$AF
	.dc.b	$A7,$39,$C7,$FC,$C9,$39,$AE,$3E,$27,$FA,$DA,$7E,$65,$6D,$F9,$C7
	.dc.b	$F9,$6A,$34,$FF,$AC,$E3,$4F,$08,$FE,$05,$E1,$FF,$B0,$27,$C0,$FF
	.dc.b	$A9,$3E,$7D,$E9,$EB,$0E,$07,$F9,$BF,$D0,$7D,$1E,$72,$FF,$7A,$43
	.dc.b	$DB,$C6,$FF,$EF,$6A,$B3,$9B,$E5,$18,$AC,$B6,$3B,$7D,$1F,$0C,$78
	.dc.b	$B6,$32,$78,$E3,$2D,$3F,$D0,$9F,$46,$F9,$EF,$1A,$FA,$3B,$18,$6C
	.dc.b	$F0,$6E,$9F,$E9,$5A,$E1,$3D,$2B,$83,$4E,$07,$FA,$DA,$5C,$7F,$D3
	.dc.b	$77,$FF,$67,$F9,$6F,$A4,$F9,$15,$AE,$9F,$8F,$39,$C3,$96,$1C,$5F
	.dc.b	$D5,$36,$9C,$3F,$ED,$9F,$A4,$E7,$39,$87,$13,$E5,$FE,$8F,$B9,$F4
	.dc.b	$9B,$8F,$DB,$6E,$3F,$EB,$3A,$E1,$3F,$7E,$BC,$FC,$CC,$BE,$EF,$37
	.dc.b	$E0,$AB,$E0,$4E,$5F,$F5,$A4,$67,$0F,$F6,$B1,$FE,$AE,$AF,$CA,$B6
	.dc.b	$7F,$C1,$B5,$C6,$3F,$36,$64,$AE,$77,$FC,$C5,$78,$B9,$3C,$AF,$E3
	.dc.b	$BE,$8B,$F1,$FE,$ED,$C7,$F3,$0D,$C7,$DC,$DC,$3F,$B2,$6D,$F2,$23
	.dc.b	$EA,$F5,$3E,$93,$F3,$7F,$08,$FA,$3C,$C6,$1B,$97,$E0,$6B,$83,$FA
	.dc.b	$7F,$2B,$39,$FC,$2F,$A4,$6B,$D3,$E8,$BF,$07,$18,$BF,$F6,$E6,$69
	.dc.b	$CD,$7C,$8A,$D5,$FC,$66,$BF,$C0,$9A,$FF,$31,$77,$B7,$F8,$06,$11
	.dc.b	$8C,$11,$FB,$AF,$8F,$EF,$18,$AF,$F0,$47,$8A,$F8,$C7,$D1,$FC,$4E
	.dc.b	$5F,$E0,$53,$98,$C5,$F1,$3F,$DD,$75,$C1,$C6,$3F,$E0,$58,$C3,$F9
	.dc.b	$1F,$B1,$E7,$18,$C7,$C2,$BE,$93,$E9,$FD,$26,$9B,$87,$4C,$BC,$D1
	.dc.b	$98,$FE,$31,$E1,$DC,$98,$F4,$F9,$0F,$E4,$F8,$7B,$5B,$91,$A4,$C3
	.dc.b	$B4,$65,$FE,$49,$F0,$5E,$89,$FE,$3E,$C8,$FE,$51,$E2,$5C,$7F,$5D
	.dc.b	$37,$1F,$C8,$F7,$FF,$17,$BE,$67,$6E,$8F,$19,$1F,$D2,$E7,$F4,$E3
	.dc.b	$18,$7B,$27,$E7,$C7,$F3,$0F,$4B,$E5,$BF,$33,$9B,$DA,$49,$E7,$AF
	.dc.b	$E3,$DB,$1B,$7F,$A0,$45,$BE,$C7,$D9,$9C,$65,$EA,$9B,$5F,$80,$9B
	.dc.b	$5F,$D1,$A6,$E3,$F8,$BE,$0F,$9E,$4B,$F1,$B6,$BB,$CE,$E7,$FA,$97
	.dc.b	$CC,$7E,$1A,$BF,$95,$FD,$1B,$E8,$FF,$AA,$66,$FA,$47,$C8,$6A,$73
	.dc.b	$18,$C7,$F9,$C5,$F1,$7F,$D3,$ED,$CF,$F8,$9B,$5F,$ED,$ED,$7F,$C3
	.dc.b	$CE,$F0,$FD,$73,$67,$F5,$BF,$1E,$FA,$FF,$D5,$47,$83,$1F,$E8,$7F
	.dc.b	$47,$FC,$C3,$1F,$9B,$6F,$A5,$72,$EB,$CA,$FE,$E1,$39,$8C,$63,$FE
	.dc.b	$B6,$C3,$96,$1F,$E0,$46,$BF,$E3,$36,$BF,$C0,$1F,$06,$9F,$C1,$5B
	.dc.b	$3F,$C1,$3E,$91,$F4,$AF,$88,$EF,$87,$F3,$CB,$DB,$CF,$3F,$F0,$25
	.dc.b	$F0,$AD,$E3,$FC,$47,$DA,$7C,$9F,$F0,$8F,$A5,$7C,$9F,$F6,$49,$CE
	.dc.b	$BF,$BB,$9C,$BF,$DB,$63,$39,$FF,$46,$6B,$FE,$22,$6B,$FE,$51,$6B
	.dc.b	$FC,$79,$E3,$87,$8F,$0F,$F0,$D7,$C0,$BE,$59,$F4,$BE,$39,$FE,$EF
	.dc.b	$C4,$FA,$5F,$EB,$2D,$CF,$FD,$CA,$DC,$FF,$AC,$B5,$FF,$1A,$BF,$C7
	.dc.b	$9B,$E3,$F2,$24,$FF,$BB,$A3,$F9,$9F,$C1,$7F,$34,$F8,$87,$77,$F6
	.dc.b	$48,$C3,$65,$3A,$7F,$C8,$73,$A7,$F9,$EF,$C3,$3C,$CF,$E4,$31,$C2
	.dc.b	$FF,$CE,$D3,$9E,$0E,$12,$8C,$3B,$39,$D7,$2C,$3C,$10,$FA,$27,$D1
	.dc.b	$FA,$DB,$1F,$DF,$0D,$46,$4D,$7D,$14,$8D,$4A,$31,$1F,$E0,$4F,$A2
	.dc.b	$7A,$9F,$E8,$DC,$FF,$AE,$46,$27,$FC,$64,$D7,$FC,$23,$93,$DF,$E1
	.dc.b	$BF,$FA,$56,$B5,$5F,$E6,$A6,$B9,$C3,$EA,$46,$23,$FB,$3B,$D7,$0F
	.dc.b	$ED,$98,$7E,$96,$7F,$C8,$4D,$7F,$C4,$33,$BB,$FF,$52,$7C,$57,$F9
	.dc.b	$9E,$7F,$C9,$0D,$7F,$C4,$ED,$BF,$EF,$1F,$27,$FE,$70,$9D,$FC,$37
	.dc.b	$A9,$FE,$F3,$5B,$BE,$B3,$99,$57,$17,$AC,$7E,$3E,$BC,$11,$FE,$8A
	.dc.b	$6B,$C8,$FB,$0F,$57,$FC,$0F,$3F,$E5,$E7,$FC,$C0,$D7,$A9,$FE,$63
	.dc.b	$FA,$27,$C9,$F8,$DB,$19,$8F,$37,$D1,$7E,$6A,$30,$7E,$1E,$76,$FF
	.dc.b	$91,$1A,$FF,$A8,$1A,$FF,$B3,$DD,$E5,$FE,$1F,$9F,$F0,$C4,$FF,$9D
	.dc.b	$5A,$FF,$95,$27,$79,$7F,$88,$BC,$D7,$D1,$7F,$16,$AF,$F0,$63,$5F
	.dc.b	$EA,$6D,$C3,$FB,$CF,$27,$D3,$AE,$3D,$7F,$9F,$6D,$FF,$78,$BE,$4F
	.dc.b	$FB,$89,$AF,$85,$F9,$57,$FE,$30,$6E,$19,$F8,$5F,$C0,$7C,$0B,$E5
	.dc.b	$1E,$FF,$86,$F7,$8B,$2F,$5E,$19,$38,$DF,$09,$F6,$C6,$5A,$7F,$8D
	.dc.b	$B0,$ED,$86,$63,$FD,$9F,$7D,$1F,$65,$7B,$F0,$FC,$14,$7B,$78,$BC
	.dc.b	$84,$7A,$63,$FC,$11,$F9,$71,$C4,$8F,$C7,$7F,$84,$63,$EF,$FA,$0F
	.dc.b	$49,$27,$9F,$A5,$5F,$F0,$AC,$ED,$EF,$3F,$02,$7E,$37,$C5,$C9,$FB
	.dc.b	$AF,$1D,$BE,$83,$EB,$24,$7D,$97,$DD,$F1,$1F,$3C,$91,$EC,$FA,$6A
	.dc.b	$F8,$B7,$78,$FF,$61,$B1,$F8,$CA,$FB,$1F,$0D,$23,$F1,$EF,$D6,$F9
	.dc.b	$CC,$47,$33,$91,$8C,$BF,$CF,$9F,$43,$FE,$7A,$8F,$F6,$93,$5F,$F5
	.dc.b	$8B,$63,$FD,$3F,$C1,$FC,$B6,$76,$7F,$51,$9F,$EA,$49,$A4,$7F,$23
	.dc.b	$EF,$7E,$87,$97,$FD,$13,$C3,$FE,$D0,$E6,$FE,$21,$7D,$E7,$C9,$F4
	.dc.b	$2F,$A1,$FE,$28,$DC,$7F,$DD,$0D,$7E,$CE,$4F,$93,$DD,$78,$3D,$19
	.dc.b	$5C,$23,$F3,$7F,$43,$F9,$CF,$9C,$93,$F8,$E8,$CB,$F9,$08,$CB,$DA
	.dc.b	$3F,$93,$7C,$C3,$86,$92,$7F,$DB,$7F,$13,$5F,$F6,$AC,$ED,$FE,$B8
	.dc.b	$D7,$33,$FD,$EE,$31,$C4,$DB,$FD,$5F,$CC,$FA,$1F,$F1,$EE,$1F,$EA
	.dc.b	$48,$C7,$B9,$DE,$FE,$BF,$D0,$FE,$E3,$C5,$97,$FA,$A8,$FD,$D9,$87
	.dc.b	$E4,$C6,$1F,$F5,$78,$FA,$0E,$EF,$FB,$D1,$B5,$FE,$C5,$6E,$5F,$73
	.dc.b	$E8,$5F,$1D,$FD,$2E,$73,$6E,$95,$3E,$6B,$F8,$AB,$67,$C3,$5C,$29
	.dc.b	$F2,$EF,$A1,$FE,$36,$FF,$67,$CF,$FD,$E6,$DC,$FF,$1A,$EF,$0F,$E2
	.dc.b	$8D,$7F,$7E,$ED,$F2,$36,$FB,$DF,$63,$BF,$BF,$2B,$9C,$7F,$24,$8F
	.dc.b	$E4,$BE,$6C,$9F,$E2,$F1,$97,$F8,$CC,$65,$FC,$BC,$7F,$2C,$F4,$BF
	.dc.b	$96,$D4,$9F,$E3,$F1,$97,$F9,$0C,$65,$FC,$FC,$7E,$4B,$A2,$47,$D8
	.dc.b	$E5,$47,$F4,$0F,$98,$7C,$8E,$AB,$F2,$73,$F8,$F2,$3F,$94,$7F,$A3
	.dc.b	$FC,$AF,$F6,$AF,$6D,$F2,$1F,$B8,$8C,$3F,$C2,$71,$EF,$7F,$EE,$BF
	.dc.b	$A1,$FC,$0B,$FE,$54,$9F,$EC,$D1,$88,$5F,$47,$FA,$84,$66,$31,$BF
	.dc.b	$F5,$7F,$A0,$FE,$2D,$1B,$F4,$95,$CF,$8F,$F8,$D4,$63,$A3,$F8,$E4
	.dc.b	$62,$75,$8F,$8A,$72,$D3,$CC,$7F,$61,$FA,$1F,$F0,$A6,$1D,$DB,$9E
	.dc.b	$E7,$D2,$FF,$80,$63,$31,$C7,$D3,$F3,$CE,$5B,$1E,$EB,$B9,$2F,$FC
	.dc.b	$45,$97,$A9,$B5,$C4,$1F,$C0,$CC,$14,$F0,$1E,$A1,$AB,$4F,$F1,$36
	.dc.b	$7F,$3E,$73,$18,$CF,$9C,$E6,$B8,$7A,$9C,$DB,$7F,$00,$F1,$B8,$9F
	.dc.b	$8D,$A7,$BA,$7A,$D7,$1C,$3B,$47,$06,$E7,$F0,$6B,$86,$1C,$F1,$FC
	.dc.b	$65,$70,$BE,$67,$86,$F8,$3D,$70,$8F,$94,$DE,$3E,$47,$2F,$B6,$BA
	.dc.b	$61,$DE,$7E,$49,$C5,$C3,$CB,$0F,$1D,$FE,$1B,$0F,$0C,$EE,$F9,$92
	.dc.b	$75,$97,$E1,$27,$91,$F0,$63,$C7,$7F,$40,$EB,$1C,$E3,$2F,$93,$1F
	.dc.b	$44,$62,$39,$E1,$E2,$9C,$C7,$12,$3F,$05,$B5,$DA,$36,$64,$AE,$49
	.dc.b	$1E,$16,$BE,$A3,$C3,$7E,$DC,$7E,$39,$31,$FB,$B1,$97,$E2,$4F,$88
	.dc.b	$E8,$7C,$33,$07,$4C,$BF,$07,$1C,$30,$F2,$CF,$03,$C5,$7D,$F2,$F9
	.dc.b	$F7,$C1,$F8,$D1,$F7,$72,$F8,$D9,$7A,$A3,$85,$FE,$2F,$2F,$35,$F0
	.dc.b	$CB,$D3,$8F,$3A,$F2,$DF,$3B,$DC,$62,$B6,$F1,$BF,$B3,$8F,$82,$3D
	.dc.b	$D9,$7A,$9F,$71,$F9,$83,$FC,$61,$87,$EB,$9B,$7F,$E9,$D6,$DF,$FB
	.dc.b	$8F,$DE,$FF,$6D,$FD,$0D,$F7,$0D,$25,$7F,$84,$3D,$AF,$F2,$4B,$D9
	.dc.b	$AF,$F3,$96,$BF,$BC,$6B,$FC,$29,$FF,$28,$E5,$FC,$15,$BB,$7F,$A8
	.dc.b	$7E,$1B,$FD,$CD,$C1,$FC,$1E,$73,$7C,$FC,$4C,$1F,$A9,$AD,$C7,$EB
	.dc.b	$E7,$6F,$79,$DB,$FB,$F9,$DE,$1E,$19,$CB,$67,$5D,$3F,$C9,$63,$04
	.dc.b	$C7,$DD,$5E,$19,$D4,$62,$BC,$53,$D2,$3D,$93,$A9,$F5,$1A,$25,$7F
	.dc.b	$53,$BF,$2D,$73,$8F,$04,$F0,$75,$5C,$E3,$DC,$72,$9E,$07,$83,$1F
	.dc.b	$14,$71,$8C,$BF,$EB,$5B,$EE,$E4,$87,$4B,$EF,$5C,$A3,$15,$CB,$81
	.dc.b	$7C,$A7,$27,$2E,$06,$7F,$6A,$B7,$97,$A6,$73,$9F,$56,$DF,$D2,$CB
	.dc.b	$18,$7D,$C9,$CE,$7E,$4A,$F8,$07,$79,$ED,$97,$8E,$B7,$9F,$C0,$AF
	.dc.b	$2E,$7E,$2C,$7A,$E5,$F1,$A3,$13,$D0,$F5,$9C,$30,$E9,$8F,$8C,$95
	.dc.b	$F8,$3B,$E7,$5E,$4B,$E7,$97,$DA,$8E,$79,$79,$1F,$A5,$9F,$C3,$C7
	.dc.b	$9C,$71,$CB,$BE,$5D,$30,$F8,$07,$3C,$39,$57,$97,$0E,$55,$BC,$FF
	.dc.b	$05,$3C,$E7,$2F,$CD,$31,$9F,$A5,$FE,$C9,$F4,$2F,$A1,$FC,$C6,$BE
	.dc.b	$AC,$39,$E7,$E2,$7F,$C1,$70,$AE,$18,$73,$AF,$F1,$74,$63,$0F,$04
	.dc.b	$63,$0E,$BA,$7E,$4B,$0E,$51,$E0,$BE,$15,$FE,$3A,$F4,$9C,$B9,$A9
	.dc.b	$97,$F6,$BB,$DE,$7F,$EC,$4B,$DC,$E6,$BF,$51,$CD,$FE,$87,$8F,$F6
	.dc.b	$B6,$9F,$E2,$48,$C3,$86,$BF,$ED,$B6,$BF,$ED,$F7,$7E,$0F,$F8,$A5
	.dc.b	$B3,$FE,$D4,$ED,$7A,$FF,$E2,$FA,$D7,$8B,$6B,$93,$70,$F1,$37,$B1
	.dc.b	$87,$2A,$FF,$55,$E3,$DA,$FF,$D4,$59,$BE,$87,$82,$F6,$EC,$61,$B7
	.dc.b	$FE,$42,$CF,$FC,$85,$92,$F8,$4E,$4F,$95,$7C,$5C,$61,$E0,$F9,$FE
	.dc.b	$FA,$F8,$1B,$3A,$C3,$A4,$62,$98,$F8,$31,$4E,$63,$C7,$F4,$2F,$9A
	.dc.b	$F8,$4F,$FA,$43,$1E,$8E,$5B,$E2,$6A,$76,$D8,$C9,$C2,$32,$D3,$38
	.dc.b	$F9,$1F,$F3,$2B,$F8,$EE,$8F,$A8,$D8,$F6,$79,$4F,$A0,$CF,$3F,$F5
	.dc.b	$2C,$FF,$CD,$4D,$AE,$6D,$BF,$F6,$C5,$F1,$9F,$F7,$2C,$E7,$6F,$A5
	.dc.b	$F2,$1F,$6B,$DC,$FB,$6D,$AF,$99,$C0,$8C,$63,$E9,$3F,$06,$E3,$2F
	.dc.b	$CA,$67,$FE,$7D,$9F,$CC,$35,$FD,$03,$5F,$F7,$7B,$79,$3C,$D1,$D4
	.dc.b	$CE,$1F,$77,$97,$EC,$BF,$94,$AF,$8F,$5C,$2F,$C3,$3B,$E2,$CD,$7E
	.dc.b	$5B,$E7,$FF,$2F,$EE,$DA,$F6,$B7,$2F,$66,$DF,$95,$8C,$63,$F8,$97
	.dc.b	$EA,$F9,$2F,$A0,$A7,$E8,$BC,$5B,$8F,$E6,$31,$56,$E3,$0D,$8E,$FC
	.dc.b	$0A,$E3,$1F,$07,$E8,$1A,$7C,$76,$E1,$FC,$2A,$51,$87,$F8,$6B,$5F
	.dc.b	$C6,$9A,$9F,$C6,$C6,$F3,$FD,$27,$C4,$62,$65,$FA,$6B,$4F,$F1,$D9
	.dc.b	$CD,$FE,$07,$C2,$7C,$E3,$E3,$BF,$86,$46,$1F,$E0,$BE,$D6,$27,$58
	.dc.b	$7E,$F2,$B4,$D8,$CD,$FC,$4A,$E1,$7C,$7E,$82,$5F,$E5,$3B,$6D,$F8
	.dc.b	$F0,$56,$35,$FE,$01,$56,$FF,$A4,$D3,$E8,$3C,$87,$F1,$AF,$6A,$CF
	.dc.b	$8E,$2E,$3F,$5A,$6E,$1F,$C2,$DE,$15,$FA,$6E,$FB,$5B,$8F,$D4,$CF
	.dc.b	$F2,$69,$CD,$7F,$B8,$F5,$F9,$3F,$41,$F1,$4F,$E5,$7E,$3D,$C3,$E2
	.dc.b	$45,$C7,$F0,$F3,$6F,$96,$6F,$A0,$F9,$C7,$F3,$46,$E1,$F3,$62,$DF
	.dc.b	$E4,$26,$D7,$29,$B5,$FD,$6E,$6D,$31,$FB,$F5,$8E,$9F,$E7,$55,$A9
	.dc.b	$FC,$1F,$D0,$7D,$F3,$FA,$2F,$C9,$DC,$7E,$F4,$5C,$7F,$43,$37,$25
	.dc.b	$7F,$24,$F9,$F7,$CF,$7F,$07,$6C,$7E,$85,$B7,$FD,$B1,$AF,$E8,$1C
	.dc.b	$D7,$48,$08,$CC,$E3,$E5,$9F,$3F,$F9,$C8,$FD,$D4,$EA,$B8,$B5,$E5
	.dc.b	$87,$F6,$BB,$E3,$A7,$F3,$AA,$D7,$CE,$FF,$29,$C3,$F7,$0D,$C3,$FC
	.dc.b	$06,$D7,$F6,$ED,$7F,$A2,$BB,$CB,$D5,$08,$C4,$FB,$3E,$75,$F3,$CF
	.dc.b	$80,$FE,$97,$5A,$BF,$82,$D7,$97,$91,$FE,$03,$BE,$35,$C2,$26,$1F
	.dc.b	$C6,$EF,$95,$6A,$31,$3F,$D6,$3D,$57,$C7,$FF,$02,$F9,$E7,$C6,$66
	.dc.b	$30,$DB,$FE,$4D,$5C,$1F,$EF,$FE,$CB,$E7,$9F,$3D,$BD,$BF,$A2,$5E
	.dc.b	$DB,$8F,$F8,$6E,$F8,$56,$F8,$3E,$3F,$1B,$E7,$BF,$B6,$B7,$3E,$FE
	.dc.b	$BB,$E3,$B3,$5A,$F1,$65,$79,$3E,$9B,$FC,$2F,$AF,$F5,$5C,$7F,$8B
	.dc.b	$E5,$FD,$77,$DE,$FC,$6F,$C1,$BB,$7F,$C7,$F1,$86,$DA,$74,$FF,$36
	.dc.b	$9D,$3F,$96,$F8,$37,$97,$FE,$1A,$8E,$17,$FE,$08,$9C,$E1,$C2,$51
	.dc.b	$87,$67,$3A,$E5,$87,$82,$1F,$3C,$F9,$FF,$F0,$9E,$1F,$8E,$8C,$4F
	.dc.b	$F5,$B6,$BF,$DD,$7E,$95,$EC,$70,$DF,$FC,$37,$4F,$A0,$C5,$FE,$C2
	.dc.b	$30,$F3,$8C,$AB,$FE,$68,$FF,$69,$FA,$0F,$EC,$8D,$F5,$37,$18,$8F
	.dc.b	$F0,$36,$AF,$7F,$35,$27,$81,$FD,$CE,$74,$E0,$E3,$E1,$78,$A3,$F5
	.dc.b	$11,$9D,$7F,$5F,$F0,$FF,$E5,$D8,$FE,$F8,$D7,$FC,$2C,$D8,$FD,$AF
	.dc.b	$C7,$7F,$99,$67,$7F,$09,$EE,$AD,$6F,$FE,$40,$6C,$FF,$14,$AE,$0F
	.dc.b	$4A,$E0,$47,$FC,$19,$1A,$BF,$2C,$FF,$80,$E7,$FD,$C2,$DA,$F3,$47
	.dc.b	$F8,$09,$AF,$F9,$43,$E7,$DF,$25,$F4,$9B,$1F,$E1,$96,$BF,$E3,$3F
	.dc.b	$A0,$FC,$C3,$99,$E0,$47,$FC,$41,$1A,$7F,$C7,$53,$C9,$FF,$09,$B5
	.dc.b	$F1,$4F,$A6,$31,$1F,$D3,$FE,$83,$F8,$5B,$FE,$07,$6B,$FE,$11,$6E
	.dc.b	$3F,$E7,$DE,$6F,$C7,$FA,$09,$CE,$BF,$EC,$86,$DF,$5F,$0F,$F2,$BE
	.dc.b	$8F,$FA,$4F,$0E,$35,$FE,$C0,$AF,$07,$C2,$E0,$DF,$CD,$3E,$77,$13
	.dc.b	$A2,$1F,$E4,$A8,$D3,$8D,$3F,$C0,$17,$D6,$3D,$57,$D2,$F9,$C6,$0E
	.dc.b	$19,$FF,$58,$C9,$8F,$49,$F8,$5C,$3E,$82,$3F,$EA,$38,$FD,$BE,$5E
	.dc.b	$09,$CC,$F0,$35,$D9,$EC,$FA,$09,$1C,$5F,$7F,$D0,$62,$3D,$91,$6F
	.dc.b	$FD,$77,$82,$BE,$4F,$C6,$4B,$FB,$0E,$F3,$FA,$4E,$63,$ED,$39,$7F
	.dc.b	$01,$F1,$5F,$48,$8F,$AF,$E8,$2B,$F8,$18,$B8,$F8,$A7,$EC,$FC,$44
	.dc.b	$AF,$C4,$3F,$63,$E7,$31,$1F,$89,$8B,$5E,$A9,$FE,$27,$F4,$1F,$E6
	.dc.b	$88,$FE,$DE,$DC,$BF,$D6,$ED,$B7,$07,$D4,$F0,$B8,$C4,$3F,$D0,$53
	.dc.b	$B9,$D4,$60,$FD,$C4,$7E,$E5,$34,$C7,$D0,$3E,$7F,$F9,$3B,$67,$FD
	.dc.b	$1C,$D7,$F2,$1F,$41,$A9,$91,$A7,$F3,$31,$F9,$AF,$3F,$F3,$7F,$37
	.dc.b	$2B,$F1,$11,$97,$F1,$51,$97,$F3,$91,$FC,$E3,$E5,$DB,$52,$4F,$FA
	.dc.b	$DF,$E2,$AB,$FE,$BB,$9D,$BF,$EA,$56,$BF,$EB,$0F,$92,$4F,$08,$FF
	.dc.b	$4C,$CE,$9B,$5F,$C7,$E3,$11,$FA,$5E,$67,$CF,$FE,$DB,$C2,$93,$FA
	.dc.b	$38,$FD,$C9,$87,$F6,$D1,$87,$FD,$71,$F3,$EF,$8A,$FE,$AA,$D8,$CC
	.dc.b	$7F,$08,$F9,$5F,$EC,$31,$82,$1F,$ED,$99,$E0,$FF,$B3,$23,$6F,$FB
	.dc.b	$3B,$E7,$DF,$3F,$FC,$49,$FE,$A3,$19,$F3,$31,$32,$3F,$CB,$3D,$3F
	.dc.b	$E5,$BE,$75,$F0,$8C,$37,$A7,$F8,$43,$DA,$B8,$C6,$2F,$C1,$84,$9C
	.dc.b	$9F,$9C,$AE,$0F,$F1,$B8,$FE,$43,$E5,$C8,$FE,$1D,$1F,$C9,$0C,$3F
	.dc.b	$C8,$63,$0F,$F1,$3F,$3B,$F9,$4D,$C8,$C1,$0F,$E5,$0E,$5F,$E5,$51
	.dc.b	$F8,$FE,$69,$1F,$89,$7F,$9C,$F2,$BE,$3F,$33,$F8,$F9,$BA,$7C,$7A
	.dc.b	$D4,$F1,$3F,$DD,$5C,$8A,$FC,$61,$23,$FD,$D2,$7F,$A8,$7E,$7F,$EC
	.dc.b	$FA,$EF,$8D,$FF,$45,$F9,$A9,$CC,$61,$B5,$FC,$1B,$A1,$F3,$DF,$E9
	.dc.b	$28,$61,$CD,$ED,$EB,$E6,$31,$8F,$58,$C6,$3D,$5B,$5D,$6E,$D6,$E3
	.dc.b	$FD,$3F,$7C,$BE,$7F,$FA,$83,$7C,$DD,$E7,$E6,$B5,$6B,$2F,$6B,$70
	.dc.b	$FF,$61,$36,$7B,$47,$AA,$77,$7E,$2A,$D1,$BB,$EC,$F3,$A6,$BF,$13
	.dc.b	$CF,$F3,$E1,$F1,$63,$13,$D4,$F0,$E1,$F0,$1E,$39,$FF,$85,$33,$F9
	.dc.b	$51,$8C,$39,$BF,$EC,$0B,$E7,$3F,$06,$32,$6C,$EC,$57,$9C,$70,$C3
	.dc.b	$DB,$3C,$23,$0F,$5B,$F1,$4F,$A5,$B1,$E8,$38,$38,$7B,$46,$EF,$CF
	.dc.b	$9F,$D2,$C7,$DD,$87,$6B,$D9,$DE,$B7,$97,$CB,$9D,$13,$1F,$0E,$1C
	.dc.b	$EF,$C3,$71,$F7,$D7,$AA,$75,$18,$AF,$34,$F3,$AE,$F8,$79,$27,$24
	.dc.b	$8C,$D7,$5C,$BE,$4C,$E6,$3E,$D6,$1D,$A7,$35,$BC,$BE,$15,$6B,$2F
	.dc.b	$66,$1C,$6F,$73,$9C,$7C,$58,$71,$8F,$88,$64,$C1,$F2,$0F,$93,$8F
	.dc.b	$87,$1E,$F7,$D0,$8F,$DE,$9F,$85,$5C,$B2,$F8,$D9,$79,$23,$19,$7C
	.dc.b	$E8,$C5,$70,$26,$1F,$8F,$6E,$1E,$02,$3F,$85,$32,$7A,$A3,$C6,$E5
	.dc.b	$AF,$AF,$0F,$1E,$5F,$0A,$38,$9E,$EB,$F1,$D7,$0A,$DC,$FA,$31,$F9
	.dc.b	$19,$7C,$38,$F0,$63,$E4,$79,$63,$D7,$0E,$59,$7B,$E3,$C1,$3F,$DD
	.dc.b	$A3,$FD,$00,$FD,$A3,$18,$7F,$2A,$7F,$DA,$BE,$AB,$E7,$F1,$ED,$23
	.dc.b	$13,$FC,$C2,$31,$3A,$9C,$CF,$F8,$13,$D6,$7C,$17,$F3,$A6,$C7,$F3
	.dc.b	$8D,$91,$88,$E1,$3B,$3F,$75,$1C,$34,$FE,$5F,$F0,$7F,$EC,$BE,$DC
	.dc.b	$D4,$BF,$D3,$DE,$F7,$FE,$83,$7B,$9C,$D6,$F0,$FE,$13,$7B,$97,$02
	.dc.b	$F8,$56,$AF,$86,$1F,$C8,$70,$E1,$7B,$E9,$65,$AC,$BF,$8E,$57,$57
	.dc.b	$F9,$5B,$C3,$2E,$B8,$73,$C7,$F9,$1F,$73,$F0,$27,$CD,$F4,$C5,$D7
	.dc.b	$FE,$0D,$9C,$E7,$FE,$04,$BE,$1F,$07,$BA,$D6,$16,$BF,$D2,$D7,$C3
	.dc.b	$0E,$71,$89,$D4,$7F,$9A,$25,$78,$C3,$C9,$56,$5F,$37,$7E,$A7,$0D
	.dc.b	$C7,$C9,$3A,$6C,$E9,$A7,$F8,$73,$17,$CE,$3E,$57,$C1,$DD,$F8,$6A
	.dc.b	$D4,$E5,$FF,$55,$70,$FA,$2F,$8D,$FC,$77,$FB,$D1,$89,$FC,$F5,$71
	.dc.b	$8F,$F0,$4B,$5F,$F0,$07,$1B,$E1,$1F,$41,$EF,$F8,$8F,$8B,$39,$8C
	.dc.b	$61,$FE,$34,$8C,$3F,$02,$77,$5C,$A3,$06,$DC,$5F,$D2,$3F,$22,$E3
	.dc.b	$8B,$EF,$7A,$5C,$B1,$E7,$F5,$7F,$47,$E6,$EE,$3F,$39,$B8,$F3,$E4
	.dc.b	$FC,$54,$63,$1F,$C4,$36,$BF,$C4,$B9,$30,$FB,$9F,$37,$9C,$7C,$7F
	.dc.b	$55,$F7,$FE,$A3,$D6,$DC,$3D,$4D,$9F,$56,$FF,$26,$31,$9F,$E6,$63
	.dc.b	$15,$C0,$FC,$2F,$CE,$74,$7E,$26,$1F,$8D,$E2,$F9,$2D,$8C,$9F,$8C
	.dc.b	$8C,$B4,$CE,$3E,$2D,$15,$F9,$7F,$23,$73,$9E,$BF,$CD,$7C,$E7,$D6
	.dc.b	$6E,$1F,$55,$B3,$B8,$FC,$DC,$ED,$D1,$F9,$E9,$C4,$FD,$1E,$93,$FA
	.dc.b	$73,$F1,$3F,$81,$BE,$17,$17,$F8,$98,$C3,$FA,$89,$DC,$E9,$FB,$9F
	.dc.b	$10,$33,$3F,$C0,$5C,$3C,$E7,$F7,$10,$F2,$A9,$8F,$94,$B5,$E1,$F9
	.dc.b	$6C,$34,$E9,$1C,$BE,$4E,$47,$7C,$2E,$3F,$99,$F3,$1F,$14,$FA,$DF
	.dc.b	$E1,$03,$D7,$FE,$1B,$3A,$8C,$C6,$27,$86,$1F,$C6,$E7,$83,$FE,$A0
	.dc.b	$AE,$0E,$6B,$84,$FE,$FE,$FE,$7D,$69,$FE,$02,$D7,$FC,$F8,$D7,$EA
	.dc.b	$F6,$FC,$BF,$A3,$F8,$FF,$13,$F2,$1B,$8F,$CD,$6D,$7D,$26,$F5,$FF
	.dc.b	$AF,$F8,$5C,$1F,$43,$E2,$7E,$8B,$70,$FA,$CD,$C3,$ED,$B6,$F9,$7D
	.dc.b	$17,$D8,$F8,$9F,$B2,$DB,$FB,$CD,$BF,$C1,$B6,$B8,$CF,$E0,$EB,$8B
	.dc.b	$4F,$E2,$DF,$82,$3E,$33,$F0,$BE,$A7,$F1,$C8,$C4,$FE,$2E,$31,$3F
	.dc.b	$37,$1F,$EC,$B5,$A7,$F9,$6F,$82,$8C,$47,$F2,$A6,$BF,$3B,$5F,$F6
	.dc.b	$16,$1F,$97,$F8,$6F,$CA,$FB,$BF,$96,$6C,$64,$FC,$E4,$65,$FE,$8D
	.dc.b	$DE,$F1,$3F,$1B,$EA,$7F,$41,$8C,$4F,$E2,$63,$13,$A8,$E2,$DC,$7F
	.dc.b	$98,$0A,$E2,$FF,$B3,$1A,$FF,$4A,$6B,$FE,$C5,$6A,$E5,$FD,$A7,$C1
	.dc.b	$FE,$C3,$E8,$BF,$62,$DC,$3F,$6C,$DA,$FD,$DB,$78,$BF,$93,$6C,$F8
	.dc.b	$6F,$DC,$FA,$DF,$55,$B8,$7D,$A7,$EA,$70,$FD,$16,$D7,$19,$FF,$72
	.dc.b	$CE,$5F,$F1,$F6,$8F,$3F,$F8,$2F,$91,$F2,$5B,$3F,$39,$BA,$7C,$6A
	.dc.b	$E1,$3F,$C3,$EB,$83,$D2,$BF,$0D,$5F,$2D,$B5,$C2,$3F,$B9,$FB,$5F
	.dc.b	$E2,$EF,$5F,$FC,$19,$F8,$2F,$F0,$8C,$3F,$93,$ED,$FE,$1F,$AD,$47
	.dc.b	$F8,$76,$75,$7F,$D9,$E7,$2F,$4C,$FF,$C2,$F5,$FE,$00,$8F,$F2,$D4
	.dc.b	$61,$D6,$1F,$9B,$F9,$8F,$F1,$6F,$B3,$E2,$E9,$FF,$10,$C6,$30,$F0
	.dc.b	$46,$27,$FC,$51,$7F,$CF,$05,$6D,$FF,$07,$B5,$D3,$5F,$E8,$6F,$F8
	.dc.b	$F3,$D9,$FF,$0E,$FD,$67,$56,$E1,$F0,$8D,$DF,$F8,$A6,$FA,$B5,$EA
	.dc.b	$D7,$73,$FB,$7E,$66,$AF,$8A,$F9,$E7,$D0,$F9,$0B,$8F,$CD,$8B,$87
	.dc.b	$F8,$26,$95,$8E,$0F,$F2,$96,$9F,$E5,$4F,$53,$FC,$6B,$F3,$7F,$15
	.dc.b	$B9,$7A,$71,$F8,$BA,$A7,$E8,$36,$7E,$B5,$F3,$6C,$66,$BE,$D7,$D8
	.dc.b	$7E,$6D,$FC,$07,$CC,$7F,$8D,$6B,$FC,$91,$39,$CB,$F1,$13,$9A,$FD
	.dc.b	$EC,$F1,$7F,$C8,$6D,$7E,$D3,$5F,$EE,$ED,$4F,$5A,$DF,$4F,$F9,$9B
	.dc.b	$97,$FD,$4D,$D5,$FE,$51,$AD,$5F,$2A,$C1,$BA,$56,$2F,$FC,$C1,$F4
	.dc.b	$4F,$A4,$FF,$38,$E6,$FA,$DF,$E5,$CF,$F9,$6E,$75,$5F,$36,$35,$CF
	.dc.b	$FE,$61,$9C,$E7,$FA,$16,$BF,$E6,$26,$BF,$99,$6B,$FE,$A0,$E5,$7C
	.dc.b	$77,$ED,$3E,$4F,$FC,$BF,$18,$C7,$F6,$51,$89,$D4,$7F,$98,$67,$55
	.dc.b	$F8,$9A,$D4,$FF,$12,$9C,$CF,$F9,$B9,$AF,$F8,$9B,$AF,$F9,$55,$FF
	.dc.b	$1C,$9F,$F4,$B6,$5F,$EA,$17,$F3,$7F,$27,$E4,$E3,$FE,$12,$DC,$7F
	.dc.b	$83,$37,$1F,$CD,$36,$FF,$CA,$57,$D2,$BF,$C0,$3D,$0F,$93,$D4,$78
	.dc.b	$E7,$58,$7C,$5C,$7F,$8B,$D6,$A3,$E4,$7C,$07,$F1,$AE,$4F,$8E,$DC
	.dc.b	$3E,$23,$67,$FC,$CB,$82,$31,$97,$91,$E3,$5F,$C7,$FE,$46,$DF,$0F
	.dc.b	$80,$FF,$51,$CF,$D4,$6B,$F4,$7E,$03,$FD,$69,$F0,$5F,$61,$B1,$FE
	.dc.b	$AD,$AF,$F2,$9F,$C4,$FF,$89,$BC,$6F,$E4,$CD,$8F,$F5,$3B,$5F,$E3
	.dc.b	$CD,$F8,$B7,$6F,$FB,$4A,$5E,$BF,$FA,$9D,$FF,$2F,$CE,$B6,$E1,$5B
	.dc.b	$AD,$52,$B4,$F2,$C3,$E8,$78,$FF,$CB,$70,$FF,$6F,$7B,$17,$3F,$F6
	.dc.b	$CB,$6F,$F3,$15,$7A,$3F,$B9,$7C,$16,$78,$FF,$DD,$5D,$D7,$E0,$7F
	.dc.b	$5D,$87,$E5,$EB,$87,$C1,$7F,$9A,$E3,$FD,$74,$D7,$FD,$36,$DD,$3F
	.dc.b	$25,$D5,$FE,$5B,$9D,$FA,$9D,$E1,$F2,$34,$EB,$12,$7F,$AD,$46,$5E
	.dc.b	$27,$83,$DB,$F1,$91,$FE,$C1,$EA,$57,$C7,$17,$0F,$1C,$DE,$1F,$CD
	.dc.b	$FC,$6F,$8C,$C5,$B8,$1A,$31,$D4,$FF,$6D,$AD,$D6,$BD,$AF,$ED,$53
	.dc.b	$C6,$D1,$89,$D4,$FF,$83,$A7,$2D,$9F,$89,$E8,$48,$F9,$B1,$F4,$0C
	.dc.b	$3F,$D3,$E3,$0F,$F4,$91,$F3,$5F,$B5,$7C,$6D,$7F,$BF,$35,$FF,$7F
	.dc.b	$B5,$FE,$EC,$EF,$46,$7F,$DE,$7E,$6D,$F4,$BF,$8B,$B4,$61,$F1,$5F
	.dc.b	$1A,$FF,$00,$7E,$05,$FE,$9D,$9F,$33,$5F,$F2,$5B,$70,$FF,$0D,$66
	.dc.b	$F8,$DE,$D3,$9B,$E7,$D3,$83,$F3,$9E,$77,$D6,$AC,$63,$F4,$AF,$B3
	.dc.b	$FE,$11,$0E,$1E,$66,$EB,$9F,$C6,$3E,$6E,$9D,$25,$9F,$F7,$CF,$25
	.dc.b	$E2,$FE,$03,$D2,$46,$2B,$FB,$AC,$E6,$31,$1F,$E1,$48,$FD,$77,$99
	.dc.b	$23,$FC,$05,$5C,$A7,$F6,$9F,$83,$BC,$8A,$4A,$FF,$1F,$70,$57,$CA
	.dc.b	$FF,$AD,$A7,$84,$7F,$83,$7C,$14,$62,$3F,$C3,$13,$FC,$17,$0F,$4C
	.dc.b	$79,$C8,$7F,$83,$8A,$7F,$BD,$FE,$B3,$F8,$31,$C7,$3F,$9D,$A7,$F8
	.dc.b	$83,$F0,$7F,$F7,$55,$7F,$86,$1A,$FF,$90,$1A,$FF,$90,$1A,$FF,$90
	.dc.b	$1D,$C7,$F5,$BC,$7F,$89,$CF,$B8,$FF,$45,$38,$8F,$07,$C6,$FE,$13
	.dc.b	$E9,$FF,$71,$5F,$E5,$DE,$54,$62,$3F,$CD,$AD,$7F,$CA,$D1,$AE,$EF
	.dc.b	$DD,$B6,$3E,$56,$0C,$FE,$A9,$DE,$BF,$49,$F2,$CF,$A0,$FE,$07,$68
	.dc.b	$C3,$FC,$19,$AF,$F9,$C5,$AF,$E5,$DC,$E7,$F0,$7D,$FF,$C1,$9D,$AA
	.dc.b	$57,$92,$32,$FF,$03,$AE,$3C,$1F,$E0,$7F,$C2,$3F,$39,$3F,$E9,$E6
	.dc.b	$BF,$ED,$46,$BF,$E7,$7F,$05,$18,$8E,$5C,$FF,$94,$F9,$79,$1F,$E5
	.dc.b	$9A,$DC,$FE,$63,$E1,$BF,$33,$F0,$52,$BE,$4C,$65,$F9,$71,$97,$EE
	.dc.b	$73,$FE,$5A,$7E,$5F,$B7,$F2,$B4,$FF,$79,$57,$C9,$8C,$11,$FC,$F6
	.dc.b	$CC,$FF,$AE,$36,$3E,$24,$F4,$77,$3A,$30,$F8,$B8,$9D,$9F,$B2,$F8
	.dc.b	$39,$3F,$82,$8F,$DA,$98,$78,$46,$1F,$89,$BB,$D6,$FE,$84,$D8,$FA
	.dc.b	$51,$82,$3F,$8E,$F8,$0F,$E0,$27,$F7,$79,$E1,$1E,$5E,$EF,$E0,$BF
	.dc.b	$29,$23,$F2,$B1,$FC,$20,$C3,$EB,$8C,$3A,$3F,$36,$EB,$9B,$F5,$D3
	.dc.b	$FC,$43,$D2,$FE,$23,$EC,$4B,$FD,$0C,$65,$FD,$1C,$65,$F9,$11,$FC
	.dc.b	$63,$C6,$FE,$33,$E8,$C9,$FD,$4C,$65,$FD,$5C,$65,$FA,$11,$FE,$17
	.dc.b	$F5,$BF,$8B,$EF,$FD,$B1,$B3,$FE,$D7,$6B,$FC,$F5,$AF,$FB,$3D,$CE
	.dc.b	$5F,$E6,$3E,$37,$2F,$E1,$0E,$DC,$5F,$7A,$FF,$D6,$57,$C3,$A9,$0E
	.dc.b	$7F,$DE,$5F,$18,$DD,$F1,$7F,$33,$F8,$57,$F8,$86,$F8,$DA,$31,$3C
	.dc.b	$DB,$1F,$CD,$23,$FC,$69,$EC,$BE,$5B,$FC,$69,$4F,$95,$A6,$A3,$3E
	.dc.b	$24,$BF,$E7,$D5,$FC,$A3,$1F,$0B,$C6,$33,$F2,$9D,$B4,$91,$83,$F7
	.dc.b	$59,$7C,$67,$9F,$CA,$F8,$8D,$BE,$E7,$0E,$8E,$33,$EA,$E7,$91,$D1
	.dc.b	$F5,$FC,$9E,$23,$E3,$F8,$9E,$E3,$1C,$7F,$E1,$08,$FF,$19,$62,$BD
	.dc.b	$47,$BA,$B7,$A7,$FA,$8E,$BF,$C6,$F3,$EA,$E5,$F9,$67,$CA,$F9,$3B
	.dc.b	$1F,$1B,$E4,$FE,$69,$87,$FA,$67,$8A,$BE,$71,$F7,$27,$6F,$53,$F9
	.dc.b	$07,$E1,$4E,$0F,$F3,$63,$63,$31,$F6,$BE,$53,$2F,$DC,$F6,$24,$E8
	.dc.b	$87,$E2,$9C,$BE,$4A,$E8,$DB,$EE,$47,$F0,$3F,$29,$5F,$C9,$FC,$1F
	.dc.b	$E2,$8C,$70,$FC,$DD,$18,$30,$D5,$FE,$2A,$F7,$BF,$FB,$3E,$9F,$29
	.dc.b	$EE,$25,$7E,$41,$AF,$E8,$BE,$53,$F3,$26,$32,$F0,$57,$29,$FC,$BC
	.dc.b	$E9,$F9,$75,$FE,$4D,$F9,$8F,$F6,$04,$7F,$9B,$1A,$FF,$8E,$C3,$87
	.dc.b	$FC,$7C,$D7,$F7,$26,$BE,$1F,$FD,$39,$D7,$23,$FC,$D1,$3D,$31,$FF
	.dc.b	$74,$75,$3E,$3B,$F6,$3F,$29,$2F,$F5,$B1,$FB,$43,$0F,$F0,$08,$C3
	.dc.b	$FB,$2F,$88,$FD,$C9,$1F,$DA,$FC,$05,$7F,$83,$C5,$C3,$C3,$3F,$E0
	.dc.b	$AF,$82,$FF,$63,$D7,$E4,$F0,$F5,$FA,$29,$CF,$4F,$9E,$BF,$8B,$CE
	.dc.b	$7C,$5F,$F6,$CD,$FF,$1A,$E6,$7C,$DF,$F8,$CB,$0E,$18,$7F,$85,$71
	.dc.b	$E7,$7D,$9B,$1D,$1B,$7C,$BC,$3F,$F6,$FD,$6F,$1E,$F7,$B6,$BD,$5C
	.dc.b	$D7,$18,$CE,$5E,$77,$DC,$71,$29,$C4,$CC,$71,$73,$1C,$71,$F5,$B6
	.dc.b	$FC,$93,$C3,$2F,$81,$18,$8F,$F1,$76,$5F,$39,$B8,$FF,$8D,$63,$11
	.dc.b	$F8,$AF,$8A,$75,$3D,$8F,$F5,$FC,$61,$B1,$99,$FC,$B4,$E5,$FF,$1F
	.dc.b	$7C,$0F,$F0,$C7,$ED,$B5,$FB,$2D,$7F,$C9,$0D,$CF,$FD,$49,$E0,$FC
	.dc.b	$88,$FE,$3F,$AF,$E0,$76,$7C,$DB,$E1,$BF,$95,$4F,$C4,$AD,$EF,$E0
	.dc.b	$C3,$91,$F5,$2F,$87,$6E,$33,$F9,$B7,$C1,$F2,$D6,$F1,$EB,$87,$39
	.dc.b	$DC,$F1,$C7,$C9,$8F,$49,$C9,$23,$C9,$3C,$F1,$E1,$7F,$DA,$3E,$2B
	.dc.b	$FD,$57,$1F,$E9,$96,$DF,$F9,$19,$7B,$7F,$D1,$F9,$BF,$A4,$79,$89
	.dc.b	$DC,$61,$B7,$D7,$DF,$E0,$F7,$BB,$87,$F7,$FA,$DB,$7D,$BC,$E3,$D7
	.dc.b	$DA,$B4,$9B,$18,$6E,$1E,$96,$DF,$FA,$64,$FE,$33,$EA,$FF,$5F,$7F
	.dc.b	$8C,$B5,$FE,$33,$DE,$8C,$47,$F8,$52,$31,$5F,$C9,$67,$D3,$3D,$23
	.dc.b	$E4,$79,$D8,$AD,$CF,$F5,$EE,$EF,$F6,$65,$7C,$DF,$6B,$E0,$F3,$27
	.dc.b	$35,$F3,$E7,$31,$88,$FF,$43,$35,$FE,$1C,$D5,$3C,$FE,$07,$D2,$DB
	.dc.b	$FF,$44,$57,$6E,$1F,$AD,$EA,$F6,$BE,$FE,$27,$CE,$F1,$6E,$1F,$83
	.dc.b	$6E,$5F,$EF,$DC,$BD,$15,$DB,$2F,$F1,$55,$FF,$47,$9C,$B6,$33,$5F
	.dc.b	$0E,$BE,$DF,$B5,$D3,$3F,$EC,$D5,$FE,$48,$BF,$C9,$7A,$3F,$0B,$7F
	.dc.b	$C0,$B6,$3A,$47,$84,$8E,$0F,$7B,$F8,$8F,$67,$86,$5F,$4E,$BF,$17
	.dc.b	$C3,$F3,$9B,$7F,$19,$F6,$11,$F5,$36,$79,$47,$E9,$FD,$9E,$4F,$CE
	.dc.b	$EA,$FC,$BF,$B3,$F6,$9B,$19,$3E,$23,$F2,$48,$E0,$D5,$7E,$C6,$39
	.dc.b	$47,$ED,$63,$F4,$67,$E2,$BE,$6B,$FC,$49,$19,$3F,$D8,$CE,$23,$EC
	.dc.b	$35,$FF,$68,$35,$FE,$B0,$D7,$FC,$E5,$F1,$43,$B7,$A7,$FB,$47,$15
	.dc.b	$DA,$31,$E5,$48,$FF,$3D,$F0,$7F,$06,$AF,$F3,$64,$60,$FF,$3B,$A9
	.dc.b	$F5,$7A,$5F,$98,$C7,$FC,$4D,$5F,$8B,$6E,$3E,$A9,$FD,$AF,$6F,$97
	.dc.b	$D8,$F2,$1E,$5C,$EE,$1F,$E8,$EF,$9A,$7C,$37,$F0,$C1,$F3,$5F,$C4
	.dc.b	$BC,$E4,$EA,$BC,$D9,$FF,$A4,$EF,$71,$CD,$FE,$0B,$5C,$1B,$5C,$6B
	.dc.b	$FD,$4C,$D7,$FD,$4C,$E7,$7F,$F3,$85,$71,$BE,$F9,$F9,$BE,$33,$C9
	.dc.b	$7B,$9F,$8F,$ED,$FD,$1F,$73,$E9,$36,$38,$F1,$23,$19,$FF,$BF,$A3
	.dc.b	$13,$AA,$FB,$5C,$1F,$6F,$DC,$B3,$C9,$B7,$C5,$BA,$FD,$B7,$F0,$5E
	.dc.b	$DF,$E0,$FD,$CB,$7E,$26,$DF,$85,$B7,$F8,$36,$CF,$E0,$9B,$3F,$90
	.dc.b	$7F,$92,$FC,$CB,$E0,$FF,$21,$EA,$61,$B7,$F5,$BC,$35,$6E,$30,$DC
	.dc.b	$3E,$17,$BD,$C6,$7E,$3F,$BB,$F9,$BF,$77,$F3,$8D,$8C,$9F,$32,$32
	.dc.b	$D3,$F0,$B4,$EF,$FE,$30,$B5,$FE,$CB,$D1,$FD,$CD,$FE,$A1,$3A,$73
	.dc.b	$EA,$7F,$4A,$8E,$35,$FE,$6A,$3A,$78,$9F,$D4,$9B,$87,$F9,$CD,$A7
	.dc.b	$F9,$D6,$7A,$47,$F9,$C9,$B1,$F9,$9F,$67,$B5,$7E,$E7,$DC,$FD,$D7
	.dc.b	$CD,$5C,$3F,$12,$DB,$FC,$43,$72,$FD,$D3,$F5,$3E,$47,$F8,$17,$AD
	.dc.b	$6F,$F2,$2D,$BF,$F5,$EF,$0A,$BF,$84,$7B,$D9,$C7,$E1,$FA,$3F,$C2
	.dc.b	$FC,$CF,$84,$DC,$BE,$6E,$E8,$F8,$4D,$8F,$84,$DC,$BE,$14,$E8,$E7
	.dc.b	$3F,$A4,$F0,$AC,$FD,$58,$3F,$E4,$DE,$6F,$F0,$EC,$7E,$CB,$E4,$3F
	.dc.b	$81,$BF,$C0,$BD,$EF,$ED,$BE,$C7,$F8,$52,$73,$7C,$CE,$51,$FE,$1C
	.dc.b	$9D,$D7,$3F,$84,$C6,$1D,$AF,$87,$C5,$BD,$BF,$F1,$0F,$90,$CF,$F3
	.dc.b	$1C,$1F,$EF,$5A,$D5,$78,$7B,$BF,$C5,$31,$8C,$FF,$8B,$B7,$1F,$26
	.dc.b	$3C,$FE,$23,$78,$7F,$8B,$AB,$C1,$ED,$7F,$87,$9B,$87,$FA,$FD,$B8
	.dc.b	$7E,$4D,$AF,$FA,$F9,$CE,$5E,$5B,$FF,$1A,$57,$C9,$F9,$AE,$F8,$7C
	.dc.b	$76,$47,$CD,$F9,$AB,$8F,$9B,$55,$63,$1F,$F2,$6C,$2B,$17,$F6,$32
	.dc.b	$F9,$1E,$BF,$A7,$33,$49,$3E,$B8,$CB,$F1,$9B,$97,$87,$E8,$FE,$2D
	.dc.b	$7F,$8F,$3E,$6B,$F3,$17,$FA,$F6,$D7,$FB,$8D,$AF,$ED,$DA,$FF,$B8
	.dc.b	$9D,$E7,$FE,$4F,$6C,$FE,$43,$8F,$FD,$39,$D9,$DF,$AE,$4F,$F0,$CA
	.dc.b	$DE,$1F,$EA,$1B,$FF,$73,$4E,$6B,$73,$FE,$5D,$AD,$D7,$4A,$D1,$FC
	.dc.b	$5B,$E6,$5F,$2D,$FC,$2D,$B1,$FC,$2D,$B7,$FE,$FF,$6B,$FC,$4D,$CE
	.dc.b	$5F,$E6,$96,$CF,$D3,$FA,$D7,$77,$F9,$92,$73,$7F,$E5,$4E,$67,$C9
	.dc.b	$FE,$A7,$E6,$BF,$CC,$31,$8C,$7F,$61,$18,$9F,$F2,$CC,$E6,$BF,$CA
	.dc.b	$0D,$7F,$72,$75,$8D,$9C,$6B,$F7,$5F,$3F,$67,$F6,$DD,$6B,$FD,$7F
	.dc.b	$3B,$E6,$FC,$7C,$DF,$BD,$6E,$3F,$E5,$BA,$FF,$51,$E1,$E0,$C3,$95
	.dc.b	$F9,$89,$5F,$D4,$EB,$AF,$CD,$F2,$BF,$0C,$7F,$A7,$AF,$8D,$7F,$0F
	.dc.b	$AD,$70,$23,$F8,$AC,$F9,$FE,$0F,$F8,$87,$2F,$C8,$6C,$F7,$7E,$39
	.dc.b	$1F,$43,$71,$F0,$BF,$C6,$7E,$53,$94,$F9,$FE,$0F,$E7,$91,$F8,$0F
	.dc.b	$D2,$F8,$39,$1F,$0D,$B7,$FD,$4C,$7C,$57,$F2,$CF,$17,$FD,$39,$1F
	.dc.b	$D6,$E3,$1E,$47,$F5,$D8,$D8,$9D,$BF,$72,$76,$FF,$36,$F8,$E7,$A7
	.dc.b	$FD,$B6,$38,$5F,$E0,$A7,$3B,$F0,$94,$61,$D9,$CE,$B9,$61,$E6,$DF
	.dc.b	$F0,$9E,$57,$F6,$D6,$D7,$E0,$63,$1C,$5F,$DF,$E3,$F0,$9F,$09,$82
	.dc.b	$38,$3F,$D4,$5F,$09,$FE,$69,$7F,$C2,$CD,$7F,$26,$DC,$FF,$8F,$FC
	.dc.b	$07,$E7,$1F,$C1,$4F,$FA,$F2,$78,$3F,$91,$6B,$F8,$66,$BB,$F7,$7D
	.dc.b	$51,$D7,$87,$FA,$FE,$3F,$D3,$7E,$03,$B4,$63,$A9,$18,$C3,$91,$E8
	.dc.b	$F7,$F1,$B3,$1F,$F2,$74,$E4,$FE,$D1,$EB,$6A,$33,$1E,$EA,$C1,$D7
	.dc.b	$2F,$EB,$D5,$89,$F8,$F3,$E8,$23,$83,$FB,$A7,$AD,$C7,$23,$C6,$3E
	.dc.b	$5F,$B9,$E1,$23,$FD,$EF,$E3,$6B,$FE,$D9,$C9,$3F,$18,$91,$F6,$1E
	.dc.b	$DF,$37,$23,$EC,$C7,$C5,$5F,$34,$EF,$C3,$63,$FC,$8A,$D7,$FC,$0C
	.dc.b	$3E,$6F,$F4,$0E,$A7,$F5,$53,$B8,$FF,$21,$35,$FE,$3E,$D7,$FC,$33
	.dc.b	$F3,$7F,$E9,$C7,$FC,$A6,$D7,$BB,$79,$3F,$17,$E5,$FD,$9D,$3D,$53
	.dc.b	$9E,$4F,$E5,$6D,$BE,$BE,$EF,$E0,$EF,$84,$76,$8E,$24,$7F,$C4,$DB
	.dc.b	$30,$C4,$7E,$FE,$FC,$8F,$F9,$75,$AF,$82,$BF,$CB,$9E,$8F,$F8,$8F
	.dc.b	$C1,$FC,$93,$67,$FB,$C4,$7E,$07,$E7,$73,$E3,$23,$8B,$67,$FC,$01
	.dc.b	$18,$3F,$CD,$1C,$EF,$98,$7C,$77,$E4,$DB,$5F,$EB,$26,$BF,$E1,$5D
	.dc.b	$F5,$32,$34,$FE,$4E,$3C,$71,$0F,$F0,$71,$B3,$FC,$7F,$87,$F2,$7C
	.dc.b	$BF,$C8,$11,$E8,$21,$FE,$31,$F9,$CF,$F7,$23,$FE,$3C,$6B,$FE,$48
	.dc.b	$6F,$0F,$F0,$6C,$BD,$9A,$FC,$89,$CE,$3F,$C0,$33,$FF,$00,$E1,$ED
	.dc.b	$F4,$75,$F1,$BF,$77,$E7,$3F,$6F,$9F,$E1,$9B,$87,$F8,$5D,$B7,$FD
	.dc.b	$A7,$0F,$EF,$6F,$F9,$96,$33,$9F,$CD,$F9,$CF,$98,$75,$ED,$FD,$47
	.dc.b	$CB,$BD,$FA,$4C,$3E,$87,$C9,$57,$FA,$34,$ED,$FF,$50,$35,$CC,$FF
	.dc.b	$51,$8C,$69,$F0,$F3,$7C,$0F,$DD,$3A,$47,$F2,$CF,$0D,$F3,$8F,$81
	.dc.b	$FE,$3A,$DC,$BF,$D9,$8D,$7F,$CD,$78,$7D,$87,$FC,$17,$19,$BF,$07
	.dc.b	$CB,$7E,$2C,$E7,$7F,$8D,$F8,$8F,$C7,$7C,$1C,$C3,$F1,$D1,$97,$F2
	.dc.b	$11,$97,$BC,$7C,$8F,$84,$91,$FE,$D0,$D7,$FD,$C8,$FF,$A6,$B1,$FC
	.dc.b	$16,$DE,$57,$FA,$4C,$27,$6F,$CC,$AF,$90,$47,$E6,$BF,$CD,$69,$95
	.dc.b	$3C,$33,$F0,$BE,$57,$F5,$1F,$0D,$23,$F5,$11,$FA,$B3,$0F,$C8,$8C
	.dc.b	$3F,$33,$D7,$F9,$E4,$7F,$61,$EC,$57,$E7,$FC,$1F,$ED,$4C,$69,$F7
	.dc.b	$7C,$1F,$DC,$7C,$AC,$AF,$DC,$47,$EE,$CC,$3F,$62,$30,$E8,$FD,$F3
	.dc.b	$A9,$E9,$1F,$C0,$7D,$4F,$E0,$9E,$D4,$8C,$10,$FC,$1B,$97,$F0,$B1
	.dc.b	$FC,$2F,$C8,$FE,$19,$EA,$C9,$FE,$19,$19,$7F,$87,$46,$5F,$C6,$C7
	.dc.b	$F7,$1F,$5B,$F8,$B6,$F7,$A0,$D3,$EB,$73,$5B,$FF,$16,$57,$06,$D7
	.dc.b	$06,$D7,$F8,$C5,$FF,$18,$46,$5F,$F6,$3F,$1B,$9B,$F0,$47,$58,$FF
	.dc.b	$65,$FC,$D7,$F8,$2D,$FF,$35,$37,$BC,$DB,$FF,$3A,$CF,$F9,$C3,$DB
	.dc.b	$77,$7F,$8B,$5B,$1F,$E4,$3E,$DF,$F2,$4C,$E4,$FE,$6F,$7F,$C6,$A7
	.dc.b	$BF,$87,$FE,$09,$8C,$63,$E0,$F9,$CC,$7C,$AF,$F8,$5B,$2F,$56,$1F
	.dc.b	$A5,$F1,$7F,$CD,$71,$88,$F5,$CE,$F5,$F2,$12,$3C,$93,$B7,$FD,$1F
	.dc.b	$87,$BB,$E5,$BC,$44,$8E,$C3,$5F,$F3,$1D,$F8,$FA,$24,$EA,$7F,$CC
	.dc.b	$F1,$EB,$F9,$5C,$CF,$C8,$9C,$BE,$29,$F9,$04,$7B,$E1,$F4,$3C,$5F
	.dc.b	$88,$72,$EA,$FE,$3D,$9F,$F8,$DE,$BC,$0D,$7E,$A7,$3F,$CB,$F9,$4F
	.dc.b	$90,$FD,$52,$47,$C2,$6B,$F8,$2F,$95,$FA,$47,$07,$D9,$F3,$8E,$0F
	.dc.b	$F2,$3B,$63,$31,$F6,$7E,$5B,$2F,$F9,$BE,$75,$3F,$05,$AF,$FA,$BD
	.dc.b	$AF,$FA,$91,$DE,$5F,$D8,$79,$3F,$C7,$73,$F1,$1A,$FF,$B0,$27,$7C
	.dc.b	$7C,$71,$F2,$92,$78,$7C,$B5,$7C,$51,$E5,$23,$F9,$7A,$47,$E4,$FE
	.dc.b	$0A,$4F,$C2,$8F,$83,$F2,$B2,$3E,$2B,$F9,$93,$0F,$BD,$F8,$44,$7B
	.dc.b	$3F,$9F,$F9,$69,$1C,$8F,$C7,$8F,$99,$78,$DF,$E5,$56,$C6,$9A,$FE
	.dc.b	$EB,$E5,$B5,$3F,$EB,$99,$CC,$FE,$E1,$AF,$FB,$B9,$AF,$F0,$0F,$92
	.dc.b	$FD,$97,$CB,$C8,$FA,$11,$FB,$53,$0F,$F0,$88,$C3,$F5,$7E,$2B,$F7
	.dc.b	$44,$7E,$C7,$C8,$D7,$F8,$7C,$5E,$2F,$BD,$3F,$E0,$AF,$85,$FF,$34
	.dc.b	$F4,$31,$61,$3D,$1F,$F4,$2F,$8F,$8A,$D4,$E6,$F8,$DF,$FB,$BB,$E7
	.dc.b	$3F,$C5,$B5,$AA,$FE,$D1,$5A,$78,$5F,$3B,$EF,$1A,$8C,$3F,$C0,$6F
	.dc.b	$9C,$EB,$0F,$1C,$FF,$B9,$A3,$17,$D6,$39,$19,$39,$78,$9F,$EE,$39
	.dc.b	$DE,$1D,$A7,$78,$F9,$F0,$EA,$FB,$BC,$06,$5E,$2A,$F4,$9D,$4E,$38
	.dc.b	$74,$7C,$13,$C3,$1F,$04,$70,$3F,$9D,$E1,$EA,$6E,$5F,$E9,$A6,$E5
	.dc.b	$FD,$37,$E3,$1D,$0F,$82,$C3,$6B,$FA,$29,$FE,$F6,$70,$FF,$48,$9E
	.dc.b	$33,$FE,$40,$F8,$2F,$E3,$13,$F2,$DA,$FD,$F6,$BF,$E1,$F6,$E3,$FE
	.dc.b	$93,$EF,$F9,$0D,$9F,$7F,$CE,$3E,$2B,$FD,$DF,$C3,$F2,$F0,$FE,$50
	.dc.b	$DC,$38,$C6,$38,$3F,$1F,$C0,$C3,$97,$76,$AE,$65,$F3,$F3,$F0,$65
	.dc.b	$F5,$6F,$73,$9C,$BA,$3E,$BB,$E3,$5B,$CF,$CB,$9F,$C1,$9C,$B6,$BD
	.dc.b	$73,$A7,$91,$F8,$BF,$11,$EF,$FF,$B5,$F4,$FE,$4D,$C2,$7F,$A5,$C6
	.dc.b	$6B,$FB,$FC,$E5,$73,$7F,$A3,$E7,$4F,$F3,$19,$D3,$9F,$86,$B7,$FB
	.dc.b	$4C,$3F,$C0,$B5,$8C,$10,$FC,$2D,$D3,$98,$C5,$59,$CD,$FF,$60,$51
	.dc.b	$9A,$D3,$67,$4F,$7B,$7E,$18,$F1,$BB,$69,$FE,$93,$BF,$E5,$73,$97
	.dc.b	$E1,$FA,$1C,$63,$17,$F0,$A7,$D9,$5C,$63,$FD,$FB,$3A,$D7,$E3,$5F
	.dc.b	$BF,$17,$DE,$FF,$B9,$71,$FF,$03,$8F,$C3,$5F,$2B,$B9,$D6,$5D,$CF
	.dc.b	$AD,$3A,$AF,$F0,$7D,$F4,$E5,$F9,$91,$CF,$0E,$51,$8A,$E1,$7C,$EF
	.dc.b	$85,$F8,$A7,$39,$7D,$F9,$CD,$78,$6B,$A1,$E0,$F9,$EF,$C5,$BF,$88
	.dc.b	$F9,$3F,$74,$62,$7D,$1C,$5F,$90,$9C,$EF,$F4,$1B,$7C,$4E,$91,$96
	.dc.b	$99,$AF,$84,$3D,$9F,$F2,$B4,$7E,$63,$97,$EF,$46,$27,$C3,$5F,$AE
	.dc.b	$CF,$FD,$47,$39,$7F,$35,$5C,$3D,$2F,$C1,$B7,$0F,$F3,$35,$73,$C3
	.dc.b	$FC,$B6,$D7,$FA,$0B,$57,$C4,$3F,$0E,$7C,$67,$E3,$B8,$7F,$41,$39
	.dc.b	$8C,$61,$FE,$48,$8C,$3F,$B2,$9D,$D7,$28,$C1,$B7,$17,$F8,$5A,$E2
	.dc.b	$FC,$0F,$91,$E3,$87,$83,$F0,$FF,$A9,$AF,$EE,$63,$1B,$7E,$89,$B9
	.dc.b	$7F,$02,$6E,$1F,$29,$B5,$FC,$0F,$16,$1D,$2F,$F4,$66,$8F,$EE,$71
	.dc.b	$AE,$4F,$11,$E3,$F8,$9B,$3F,$06,$91,$8C,$7D,$91,$89,$D5,$7F,$05
	.dc.b	$FC,$41,$F1,$FC,$E7,$FC,$C7,$18,$7A,$61,$F0,$7E,$1E,$46,$B5,$B9
	.dc.b	$76,$F8,$1E,$91,$F1,$DB,$3E,$5F,$87,$F9,$47,$87,$E3,$6C,$7F,$8C
	.dc.b	$B9,$3F,$C6,$2F,$8B,$5F,$C9,$FA,$5E,$76,$C6,$4E,$11,$96,$99,$CB
	.dc.b	$F3,$F5,$D1,$FD,$5E,$BF,$B6,$9F,$E5,$8F,$FA,$B1,$F8,$3A,$FC,$5F
	.dc.b	$81,$F8,$CD,$8C,$9F,$1A,$32,$FC,$8F,$8F,$F9,$9F,$1B,$F3,$5B,$19
	.dc.b	$3E,$6C,$65,$A7,$D0,$DB,$EA,$70,$7D,$56,$C6,$4F,$AB,$19,$69,$9C
	.dc.b	$BD,$B1,$96,$C7,$E1,$7C,$8F,$C0,$75,$FE,$6D,$B1,$93,$F0,$31,$96
	.dc.b	$9F,$84,$DB,$F1,$1F,$0B,$F8,$96,$C6,$4F,$C4,$C6,$5A,$7D,$DF,$86
	.dc.b	$FC,$87,$C4,$FF,$0E,$6C,$64,$FC,$8C,$65,$A6,$71,$FB,$31,$96,$C7
	.dc.b	$E6,$3E,$23,$F3,$9F,$13,$F9,$D6,$C6,$4F,$CE,$C6,$5A,$7F,$09,$F9
	.dc.b	$A7,$C3,$FE,$97,$97,$F8,$7C,$62,$7F,$4B,$18,$9D,$47,$16,$E1,$FC
	.dc.b	$BB,$F1,$07,$CC,$7E,$BF,$83,$FB,$54,$62,$FF,$5F,$18,$9D,$47,$FA
	.dc.b	$52,$75,$5F,$98,$AD,$4F,$03,$55,$C2,$7F,$8D,$63,$FE,$AC,$9F,$F6
	.dc.b	$EF,$E2,$5C,$BC,$0F,$E2,$13,$9A,$FE,$9B,$39,$AF,$E1,$F3,$9A,$FF
	.dc.b	$0C,$78,$EA,$D5,$FF,$36,$AD,$5F,$F8,$56,$73,$E0,$A7,$FC,$B3,$3E
	.dc.b	$2F,$6F,$E5,$FB,$5F,$E1,$5B,$46,$27,$47,$D3,$9D,$47,$D2,$F8,$2F
	.dc.b	$F1,$06,$6F,$BE,$78,$39,$C3,$FD,$33,$3F,$E2,$A9,$CC,$7D,$DA,$E3
	.dc.b	$18,$D3,$E3,$C6,$2F,$C9,$C3,$2F,$E4,$FC,$65,$BF,$8F,$37,$1F,$8B
	.dc.b	$A2,$B1,$7F,$2B,$E0,$3F,$C4,$BF,$21,$F8,$A6,$C7,$E3,$DB,$8F,$D8
	.dc.b	$6E,$1F,$E4,$78,$7C,$F7,$9E,$3F,$C8,$BF,$4E,$C3,$FC,$69,$39,$C3
	.dc.b	$E0,$CE,$5F,$F6,$EC,$66,$FF,$CA,$33,$97,$FC,$07,$19,$B4,$E5,$F1
	.dc.b	$46,$63,$15,$FE,$B8,$8C,$5F,$8B,$4F,$46,$19,$6B,$FE,$CC,$9F,$A0
	.dc.b	$FF,$B0,$0F,$6C,$69,$7E,$24,$EA,$FF,$19,$FE,$15,$71,$FF,$7B,$CD
	.dc.b	$AF,$F1,$64,$DC,$FE,$44,$DA,$FF,$6D,$89,$C5,$7F,$B7,$A6,$CE,$6F
	.dc.b	$FC,$9D,$F5,$4F,$4F,$F2,$DF,$21,$86,$D7,$F9,$22,$74,$FE,$BA,$74
	.dc.b	$D3,$53,$C3,$0F,$F2,$45,$70,$7F,$63,$5E,$0F,$4F,$F4,$9F,$19,$70
	.dc.b	$FD,$CB,$6B,$F6,$ED,$D3,$F2,$2D,$9F,$D9,$36,$7F,$D0,$BC,$7F,$B6
	.dc.b	$7F,$CF,$F8,$FF,$99,$2F,$8D,$FA,$F0,$F0,$BF,$E6,$4F,$AA,$FC,$47
	.dc.b	$67,$F0,$1F,$9E,$FF,$39,$46,$32,$46,$38,$7F,$78,$DB,$FE,$22,$DC
	.dc.b	$FF,$87,$B7,$2F,$D4,$BD,$3C,$CF,$F4,$5F,$CC,$5B,$FF,$55,$BE,$0F
	.dc.b	$33,$FD,$1F,$F4,$F6,$7F,$D5,$EF,$FA,$57,$10,$F8,$5B,$EE,$7F,$A9
	.dc.b	$23,$F2,$11,$8F,$82,$FF,$74,$B6,$32,$7F,$A9,$23,$9A,$F8,$A7,$C8
	.dc.b	$FF,$AE,$A3,$47,$82,$33,$B7,$38,$E1,$3B,$C3,$C5,$18,$9F,$04,$38
	.dc.b	$5E,$9F,$82,$7F,$D8,$DF,$51,$75,$FF,$53,$B7,$0F,$F5,$CD,$F0,$7E
	.dc.b	$C7,$A7,$9E,$5F,$F6,$BF,$D4,$5F,$68,$F1,$D3,$C1,$57,$07,$B5,$C7
	.dc.b	$A1,$E3,$D3,$51,$33,$FE,$9B,$19,$7F,$11,$19,$7F,$A6,$C6,$5F,$F5
	.dc.b	$45,$76,$7F,$1D,$DF,$DF,$E4,$A4,$60,$87,$F5,$67,$2F,$28,$F4,$18
	.dc.b	$E0,$FE,$67,$38,$9E,$31,$87,$FB,$1C,$65,$F1,$9E,$FF,$91,$8F,$E8
	.dc.b	$4F,$88,$60,$FE,$CE,$E1,$D1,$FD,$B9,$D4,$FF,$B0,$6F,$F8,$F3,$FC
	.dc.b	$EF,$E0,$93,$F6,$69,$F6,$7F,$E7,$87,$FA,$CF,$1F,$E4,$23,$11,$FE
	.dc.b	$B8,$AE,$2E,$8F,$CF,$B8,$7F,$D7,$B1,$C2,$FF,$A5,$36,$BF,$37,$5B
	.dc.b	$9F,$E9,$DF,$42,$F8,$CF,$E6,$98,$7E,$82,$30,$E0,$38,$76,$D7,$F9
	.dc.b	$E3,$99,$FE,$B5,$F4,$4F,$33,$E0,$36,$BD,$47,$4B,$F8,$97,$C7,$1F
	.dc.b	$D8,$D6,$EB,$FD,$B2,$DC,$3D,$37,$C0,$EB,$84,$AE,$B8,$73,$7F,$C3
	.dc.b	$1C,$7F,$A3,$F4,$A4,$7F,$90,$63,$F5,$9E,$9F,$F8,$65,$CC,$7F,$B8
	.dc.b	$A7,$22,$3F,$4D,$DB,$27,$FC,$5E,$FE,$A7,$0C,$47,$F8,$BA,$2E,$5F
	.dc.b	$E0,$89,$EF,$F8,$9B,$C0,$FD,$E4,$BE,$39,$EF,$61,$B8,$FF,$B7,$EF
	.dc.b	$83,$6F,$55,$23,$06,$F3,$FE,$09,$F8,$AB,$E2,$A9,$E2,$89,$3F,$8D
	.dc.b	$EF,$56,$E3,$0D,$7F,$DD,$18,$FF,$8F,$DF,$27,$0F,$6F,$4A,$46,$08
	.dc.b	$35,$FF,$26,$DF,$F8,$6E,$3A,$CF,$C4,$AF,$D3,$7E,$2A,$F6,$7D,$3B
	.dc.b	$4B,$FC,$4C,$E5,$B3,$97,$B7,$37,$F4,$4D,$3D,$77,$D2,$A5,$7B,$23
	.dc.b	$86,$7F,$E3,$1C,$3A,$4F,$BB,$D1,$F9,$3E,$14,$8F,$A5,$1F,$68,$C3
	.dc.b	$E0,$18,$FC,$A9,$D4,$7C,$AF,$AA,$FA,$7C,$12,$30,$43,$95,$7D,$28
	.dc.b	$D4,$63,$75,$7C,$6F,$9A,$FE,$89,$EE,$BF,$01,$FE,$73,$7F,$C1,$E3
	.dc.b	$1E,$1E,$E3,$E4,$D7,$0A,$DC,$7A,$6B,$6E,$88,$66,$37,$B3,$87,$FB
	.dc.b	$8D,$70,$AF,$84,$37,$61,$F9,$2F,$AA,$FC,$9F,$1C,$AF,$C6,$C6,$5F
	.dc.b	$C7,$C6,$5F,$81,$1F,$A8,$C3,$F3,$7C,$32,$7F,$49,$19,$7F,$2F,$19
	.dc.b	$7F,$2F,$1F,$A2,$EF,$FD,$1F,$D5,$49,$FC,$EC,$65,$FC,$FC,$65,$FA
	.dc.b	$11,$97,$E9,$46,$74,$FC,$DC,$FE,$BB,$D4,$FD,$7F,$CD,$CB,$FD,$4C
	.dc.b	$65,$FD,$5C,$65,$FB,$91,$FB,$6F,$8F,$FD,$BF,$C8,$49,$FD,$8C,$65
	.dc.b	$FD,$9C,$65,$FC,$24,$7E,$A3,$C1,$FD,$FE,$72,$7F,$4F,$19,$7F,$77
	.dc.b	$19,$7E,$BF,$07,$F1,$D7,$66,$88,$6A,$7F,$93,$FE,$2D,$FE,$11,$8F
	.dc.b	$D0,$78,$68,$C6,$BF,$85,$94,$E7,$D0,$FF,$0B,$7C,$AB,$E4,$79,$FA
	.dc.b	$18,$F4,$66,$1F,$D7,$F0,$E9,$3E,$0F,$8D,$93,$9C,$D1,$C2,$3F,$53
	.dc.b	$8F,$FB,$19,$D7,$8D,$E8,$3C,$3E,$DD,$8E,$5F,$47,$EC,$30,$FB,$30
	.dc.b	$46,$63,$67,$43,$C5,$F1,$5F,$10,$8F,$F2,$BC,$1F,$25,$88,$F2,$FC
	.dc.b	$8F,$C2,$31,$C3,$F1,$F6,$F9,$58,$66,$B9,$56,$B9,$3F,$CB,$F5,$F3
	.dc.b	$E7,$E9,$90,$F0,$7B,$D5,$F8,$7E,$EF,$D0,$31,$8F,$7D,$7F,$C2,$F8
	.dc.b	$7F,$27,$D1,$5B,$BF,$9C,$7F,$86,$7E,$6F,$F9,$C3,$FE,$08,$6B,$FE
	.dc.b	$BD,$6B,$FE,$6A,$6B,$FE,$76,$78,$E1,$F8,$5F,$21,$6E,$3F,$C2,$9D
	.dc.b	$FC,$C9,$87,$CB,$8C,$BF,$E0,$AC,$39,$E5,$FE,$80,$FA,$57,$C9,$7E
	.dc.b	$3B,$9A,$5F,$F8,$A7,$A9,$5A,$8C,$35,$FC,$4C,$F0,$7F,$9E,$4F,$0A
	.dc.b	$FF,$35,$56,$F5,$F5,$69,$F9,$2B,$FC,$06,$1A,$BD,$FB,$5F,$E7,$9C
	.dc.b	$9F,$26,$F8,$6C,$37,$0F,$1D,$70,$7F,$C8,$15,$C2,$B6,$66,$3C,$B9
	.dc.b	$FF,$BA,$5D,$DF,$37,$FA,$A5,$71,$7A,$9F,$E3,$2F,$99,$7C,$7F,$FA
	.dc.b	$1F,$CE,$93,$98,$FF,$2C,$4E,$A3,$FC,$0D,$DD,$FD,$49,$B5,$AA,$FF
	.dc.b	$49,$7E,$2F,$FE,$37,$C3,$FB,$93,$FE,$A2,$E0,$FC,$03,$63,$3E,$C7
	.dc.b	$03,$8F,$B3,$31,$FF,$0A,$57,$37,$FB,$6F,$13,$E5,$3F,$D3,$DE,$CC
	.dc.b	$9F,$D8,$47,$F0,$E3,$0F,$F6,$CD,$7F,$C3,$59,$FD,$89,$FC,$E7,$95
	.dc.b	$FC,$87,$E1,$9F,$39,$F9,$98,$C3,$7A,$3F,$D9,$F3,$AC,$5F,$3C,$FA
	.dc.b	$0F,$8D,$1F,$ED,$3B,$E0,$D3,$F4,$58,$74,$6E,$1D,$27,$FD,$B7,$7C
	.dc.b	$F8,$BF,$DA,$78,$F7,$B7,$5B,$1E,$51,$86,$C6,$71,$E9,$3E,$6C,$BC
	.dc.b	$B7,$CE,$B5,$3E,$53,$95,$70,$69,$E9,$C7,$C8,$E6,$78,$D7,$3C,$3C
	.dc.b	$27,$B6,$B7,$5D,$6B,$BB,$FC,$1E,$7F,$A8,$36,$FF,$D2,$13,$A8,$FF
	.dc.b	$79,$FC,$4F,$F9,$0F,$E8,$3F,$D2,$8D,$8F,$F4,$CB,$5F,$F7,$C9,$FE
	.dc.b	$F9,$F7,$93,$D3,$D1,$7B,$14,$95,$FC,$6E,$32,$D8,$F7,$46,$34,$FC
	.dc.b	$06,$8F,$A0,$7A,$9F,$D7,$E7,$E6,$56,$F8,$7F,$91,$37,$1F,$E4,$4D
	.dc.b	$C3,$95,$F1,$AD,$E3,$E0,$25,$FD,$1C,$7B,$F5,$63,$2E,$F5,$6F,$F1
	.dc.b	$35,$AB,$FA,$19,$78,$5F,$C4,$B7,$63,$E3,$9E,$18,$74,$AF,$CC,$78
	.dc.b	$AE,$6F,$F6,$46,$7F,$E8,$7F,$A3,$FF,$52,$3F,$C1,$3D,$9F,$F0,$07
	.dc.b	$47,$F3,$D9,$CE,$1F,$B1,$9C,$D7,$F9,$D6,$73,$5F,$D0,$EF,$AD,$FF
	.dc.b	$98,$A7,$35,$FE,$7B,$F8,$27,$6F,$2A,$FF,$37,$C6,$9B,$19,$CB,$FD
	.dc.b	$59,$CD,$6B,$8D,$FF,$A7,$DE,$13,$FE,$B1,$AD,$57,$5B,$78,$0F,$A1
	.dc.b	$E9,$9F,$F5,$8C,$BF,$95,$46,$3B,$B8,$56,$E7,$F9,$3C,$62,$75,$1F
	.dc.b	$D4,$FE,$87,$E8,$BF,$CF,$BE,$87,$FC,$0E,$E7,$8F,$DC,$A3,$39,$7F
	.dc.b	$54,$AE,$2F,$F3,$E8,$CF,$07,$F8,$C7,$45,$6A,$3F,$C6,$97,$CF,$D2
	.dc.b	$F9,$55,$AC,$7B,$E1,$CA,$98,$FF,$B8,$71,$E5,$18,$39,$1F,$DB,$7C
	.dc.b	$1F,$C3,$3F,$D1,$BD,$CE,$B1,$8A,$FE,$65,$18,$DF,$E5,$B6,$B8,$1C
	.dc.b	$E3,$2D,$33,$3F,$E8,$EF,$4F,$FC,$83,$1F,$D5,$3A,$FE,$C4,$62,$7F
	.dc.b	$A0,$46,$3C,$CF,$B2,$DA,$E0,$7B,$B7,$FF,$2C,$52,$F9,$46,$1B,$3F
	.dc.b	$EC,$DC,$3F,$99,$3C,$4F,$8F,$5C,$70,$FC,$B7,$A4,$FA,$17,$B7,$F8
	.dc.b	$AE,$0C,$C6,$30,$FE,$DB,$18,$9D,$47,$F9,$8E,$75,$87,$F5,$BA,$D4
	.dc.b	$E6,$BF,$2B,$EE,$62,$BF,$CE,$BF,$43,$F4,$75,$FD,$75,$79,$23,$17
	.dc.b	$EB,$8C,$69,$F9,$E6,$C7,$F8,$01,$B7,$D2,$7F,$47,$38,$BF,$F2,$D5
	.dc.b	$FE,$26,$BF,$D0,$2D,$CC,$BD,$7F,$17,$2F,$8D,$B8,$7F,$A0,$67,$4D
	.dc.b	$9F,$DE,$7D,$0F,$FA,$26,$35,$58,$F7,$39,$47,$1F,$57,$DC,$70,$D2
	.dc.b	$C7,$5F,$47,$0D,$3E,$D6,$9F,$EF,$AC,$7E,$2B,$EA,$F8,$FF,$90,$78
	.dc.b	$BD,$CB,$5C,$27,$F1,$1E,$EF,$9B,$5F,$C9,$B6,$75,$CC,$8C,$57,$E6
	.dc.b	$DB,$C3,$F8,$F6,$D7,$E3,$2B,$83,$E6,$F0,$59,$FC,$F6,$FB,$9F,$1B
	.dc.b	$87,$E3,$B6,$B8,$52,$31,$93,$2F,$F3,$A4,$E6,$B7,$7F,$43,$DE,$FA
	.dc.b	$3E,$E5,$9E,$6D,$C3,$94,$7F,$08,$CB,$CB,$39,$C7,$AB,$F6,$B8,$3E
	.dc.b	$DF,$15,$BF,$2B,$70,$E6,$7D,$37,$17,$E6,$6C,$F7,$CF,$FD,$14,$FB
	.dc.b	$AF,$AB,$F8,$BF,$33,$F1,$3D,$7F,$A1,$6C,$64,$F8,$B1,$96,$9F,$23
	.dc.b	$83,$F2,$3E,$DF,$E4,$9B,$19,$3E,$6C,$65,$A7,$E0,$3E,$87,$F3,$3E
	.dc.b	$B7,$F1,$66,$C6,$4F,$AB,$19,$69,$9D,$7E,$1E,$FF,$80,$9D,$46,$34
	.dc.b	$FF,$32,$57,$DE,$F7,$3F,$4F,$F4,$3F,$A8,$6C,$64,$FC,$1C,$65,$A7
	.dc.b	$FB,$C7,$E2,$1F,$3D,$FB,$0E,$7F,$E3,$71,$89,$FC,$5C,$62,$75,$1C
	.dc.b	$5B,$8F,$DA,$1F,$42,$FA,$1F,$DD,$F1,$7F,$6F,$8C,$5F,$E5,$A3,$13
	.dc.b	$A8,$FD,$74,$EA,$BE,$05,$6A,$78,$1A,$AE,$13,$F8,$AF,$01,$93,$E6
	.dc.b	$FF,$C0,$DF,$2D,$CF,$A3,$FD,$E5,$39,$C7,$F5,$D3,$97,$FD,$E3,$19
	.dc.b	$C3,$E9,$CE,$56,$1D,$B0,$EB,$8F,$C8,$F6,$BB,$FB,$D7,$2E,$AD,$C3
	.dc.b	$9B,$73,$FF,$73,$65,$FE,$B6,$AF,$A9,$F1,$9F,$E2,$19,$7D,$0B,$C2
	.dc.b	$FA,$B3,$98,$C5,$70,$3F,$8E,$D7,$0A,$DC,$62,$77,$8F,$B3,$87,$E2
	.dc.b	$9F,$1B,$E8,$6D,$7C,$48,$B3,$F0,$A6,$E1,$F2,$BE,$0F,$FC,$4F,$F0
	.dc.b	$FF,$88,$6E,$1F,$8D,$6D,$FD,$5E,$24,$62,$FF,$B9,$FD,$07,$91,$FF
	.dc.b	$26,$FD,$0B,$1F,$F1,$B4,$E7,$1F,$EF,$53,$97,$F9,$A4,$67,$1F,$E3
	.dc.b	$D3,$97,$F2,$91,$9B,$4E,$5F,$0E,$BF,$E0,$4A,$FE,$FB,$5F,$19,$FF
	.dc.b	$47,$B5,$CC,$6D,$FF,$14,$C7,$85,$3E,$3E,$47,$D7,$7C,$0C,$7C,$F7
	.dc.b	$FE,$51,$FA,$17,$5F,$F9,$0B,$E8,$AD,$7F,$4F,$9B,$97,$FB,$FA,$6E
	.dc.b	$7F,$E1,$A9,$B5,$FE,$37,$9B,$5F,$E3,$C9,$B2,$9C,$FB,$0F,$A2,$7A
	.dc.b	$DF,$A6,$F8,$DF,$F2,$3C,$62,$FF,$4B,$18,$9D,$47,$F6,$AA,$E5,$18
	.dc.b	$C7,$F9,$DC,$62,$78,$9B,$BE,$93,$FA,$CB,$EF,$EA,$FE,$DB,$B6,$DF
	.dc.b	$ED,$1B,$5F,$B2,$6E,$9F,$A8,$6C,$FF,$53,$EA,$FD,$C4,$FF,$99,$71
	.dc.b	$FF,$2D,$DB,$C4,$56,$AF,$FC,$C3,$9F,$C3,$CB,$C5,$1E,$76,$DF,$B3
	.dc.b	$E7,$7F,$1D,$87,$2F,$A2,$FF,$36,$C6,$31,$46,$38,$BF,$88,$B7,$1F
	.dc.b	$E1,$ED,$C3,$F8,$6B,$74,$FE,$06,$F4,$F4,$3F,$D2,$7F,$45,$6F,$FC
	.dc.b	$09,$EA,$23,$C3,$E8,$7F,$A6,$3E,$6A,$D6,$9F,$F4,$F6,$5F,$E7,$EF
	.dc.b	$86,$F7,$FC,$1F,$FB,$AA,$11,$87,$FD,$DA,$D7,$FC,$BE,$DF,$8C,$7D
	.dc.b	$17,$F8,$2E,$5C,$9F,$EB,$87,$FD,$17,$3A,$DF,$85,$6E,$B5,$4A,$D3
	.dc.b	$CB,$0F,$A7,$C4,$F5,$BF,$CD,$F8,$FF,$B2,$7E,$8A,$EB,$FE,$75,$6E
	.dc.b	$1E,$2B,$E0,$FE,$55,$7A,$FA,$E6,$FF,$6F,$7C,$E5,$E3,$3C,$B6,$5F
	.dc.b	$0C,$39,$7B,$7C,$7A,$1E,$5D,$38,$52,$69,$9A,$FE,$9B,$38,$79,$98
	.dc.b	$21,$F8,$D8,$E7,$E0,$F7,$FA,$29,$3E,$B8,$CB,$FD,$62,$32,$FE,$56
	.dc.b	$32,$FF,$58,$8C,$F0,$7F,$B1,$F0,$FE,$1F,$3C,$CE,$44,$39,$1F,$24
	.dc.b	$F8,$1F,$45,$1C,$10,$FE,$CA,$E5,$FD,$31,$F1,$0C,$4F,$F5,$E8,$C4
	.dc.b	$7F,$9A,$FE,$0D,$1F,$6A,$1F,$45,$FE,$89,$7F,$CE,$0D,$7F,$C1,$D3
	.dc.b	$C1,$E5,$3C,$67,$55,$FE,$BE,$6F,$1F,$FA,$E2,$F8,$5F,$F8,$5F,$E8
	.dc.b	$5F,$1D,$E1,$6C,$7F,$9F,$DA,$FF,$87,$7E,$8B,$F3,$EE,$67,$F5,$4F
	.dc.b	$FB,$AA,$76,$FF,$B1,$83,$87,$FD,$92,$D7,$84,$7C,$19,$FE,$CB,$1F
	.dc.b	$E8,$F6,$BE,$78,$F3,$1B,$F3,$A4,$78,$CF,$F1,$E7,$17,$E7,$7C,$F9
	.dc.b	$18,$21,$FE,$4E,$E5,$FF,$13,$BF,$C3,$23,$F5,$7E,$67,$F8,$AA,$BF
	.dc.b	$41,$D7,$2B,$FC,$4E,$FE,$8F,$0C,$47,$F8,$9A,$2E,$FF,$E1,$C9,$E7
	.dc.b	$F4,$4F,$03,$F6,$F2,$F8,$FF,$F5,$36,$1F,$E1,$86,$BF,$E3,$BA,$E0
	.dc.b	$FF,$AE,$32,$46,$23,$FD,$13,$5C,$FE,$03,$AD,$F2,$87,$12,$BF,$C7
	.dc.b	$37,$C6,$73,$18,$D7,$F8,$37,$D1,$3E,$2E,$9E,$A8,$93,$FE,$B5,$9D
	.dc.b	$36,$72,$FA,$78,$79,$FA,$92,$3D,$58,$FE,$33,$1F,$0E,$BE,$CB,$D5
	.dc.b	$7F,$A0,$B1,$FF,$11,$3C,$23,$F4,$DF,$3C,$F6,$BC,$FC,$12,$BF,$DA
	.dc.b	$D3,$96,$CE,$5E,$93,$CA,$24,$7A,$E3,$58,$FC,$48,$57,$37,$EA,$7A
	.dc.b	$3F,$1B,$C1,$91,$EC,$8F,$BE,$61,$F5,$46,$1F,$91,$F3,$7F,$3B,$82
	.dc.b	$46,$08,$7C,$17,$2F,$C3,$8F,$83,$F4,$3F,$D0,$3D,$E7,$D1,$7F,$34
	.dc.b	$AF,$F7,$EC,$63,$C4,$DC,$7C,$4A,$E1,$5B,$8F,$F6,$8D,$6D,$D1,$0C
	.dc.b	$C6,$F4,$FC,$3C,$EE,$7F,$D4,$1A,$2F,$F1,$3F,$45,$F8,$AE,$09,$5F
	.dc.b	$87,$8C,$BF,$89,$8C,$BF,$7E,$3F,$4F,$CD,$F9,$2D,$E4,$FE,$22,$32
	.dc.b	$FE,$46,$32,$FE,$9E,$3F,$33,$DF,$F9,$AF,$A2,$93,$F9,$78,$CB,$F9
	.dc.b	$98,$CB,$F8,$F8,$CB,$F9,$18,$CE,$9F,$A6,$F2,$1C,$AB,$FD,$3F,$A9
	.dc.b	$FA,$8F,$9D,$98,$FE,$96,$32,$FE,$9E,$32,$FE,$76,$3F,$5F,$F2,$1F
	.dc.b	$B0,$F9,$19,$3F,$AD,$8C,$BF,$AF,$8C,$BF,$A5,$8F,$D1,$F8,$5F,$B8
	.dc.b	$D6,$4F,$E8,$A3,$2F,$ED,$E3,$2F,$D8,$8C,$BF,$AF,$8C,$E9,$FA,$1E
	.dc.b	$2F,$E7,$F3,$FE,$CD,$FA,$1F,$F0,$6B,$7D,$0F,$E4,$F1,$8A,$F8,$DF
	.dc.b	$2E,$F9,$2C,$FA,$3F,$E1,$18,$79,$14,$CE,$3F,$C7,$70,$7C,$7F,$42
	.dc.b	$47,$F1,$9C,$7A,$D7,$3F,$6F,$C3,$E4,$4A,$D6,$DF,$C7,$F9,$99,$FF
	.dc.b	$B2,$67,$B5,$7A,$3E,$8A,$46,$0F,$13,$87,$E0,$3D,$5F,$85,$F1,$7F
	.dc.b	$00,$8F,$80,$FF,$37,$53,$E4,$F3,$0D,$3E,$2F,$C9,$FC,$33,$2F,$F2
	.dc.b	$2D,$D8,$66,$BA,$67,$EA,$BD,$E9,$FC,$EF,$C8,$61,$FD,$BF,$1E,$58
	.dc.b	$7D,$43,$E8,$7B,$D5,$E7,$A3,$DD,$FA,$86,$63,$EF,$18,$36,$E6,$BF
	.dc.b	$92,$7C,$E7,$F3,$49,$F6,$65,$FD,$35,$46,$67,$7E,$6F,$E0,$79,$FF
	.dc.b	$C3,$D3,$9B,$FA,$D8,$FF,$93,$F0,$E6,$7C,$AE,$6D,$E1,$23,$D3,$1F
	.dc.b	$8B,$30,$FF,$AC,$3E,$89,$F2,$9F,$8F,$E8,$93,$FE,$B7,$9C,$B6,$72
	.dc.b	$F8,$38,$BF,$0F,$5F,$E6,$3A,$DC,$F5,$AD,$EB,$F9,$EA,$D4,$6F,$DA
	.dc.b	$FF,$39,$E8,$F9,$57,$C3,$61,$B8,$F8,$EB,$83,$F8,$9A,$E1,$5B,$33
	.dc.b	$1E,$D9,$C3,$BA,$E6,$E4,$FC,$90,$F9,$B7,$C8,$7F,$9F,$FC,$F9,$3A
	.dc.b	$9C,$B6,$72,$FF,$89,$67,$FC,$A9,$3A,$8F,$8D,$F4,$5F,$E3,$4D,$7F
	.dc.b	$92,$3F,$E9,$C6,$FB,$3F,$E8,$F8,$D6,$9F,$B3,$6D,$7F,$B4,$DE,$11
	.dc.b	$99,$E1,$EC,$FF,$07,$F9,$6F,$F4,$B7,$B3,$30,$FC,$CC,$65,$FF,$61
	.dc.b	$C6,$5F,$B5,$CC,$9F,$C7,$F9,$5F,$E5,$8F,$86,$7C,$FF,$E5,$FA,$93
	.dc.b	$98,$C3,$70,$7D,$03,$E8,$BE,$16,$1C,$A7,$FC,$59,$87,$2F,$2F,$FD
	.dc.b	$9D,$8F,$26,$CE,$B1,$E9,$8F,$3E,$4F,$F6,$6D,$EE,$BF,$DA,$6D,$C7
	.dc.b	$B4,$7E,$B2,$F8,$E5,$E2,$AD,$12,$7B,$63,$E0,$C7,$B4,$78,$8E,$51
	.dc.b	$72,$F4,$47,$80,$E2,$59,$E2,$E4,$E2,$7F,$BA,$63,$8B,$FE,$7F,$6B
	.dc.b	$EB,$3F,$D0,$5F,$11,$FE,$37,$FA,$2F,$F4,$3B,$63,$FD,$E4,$D7,$7E
	.dc.b	$F2,$7F,$D1,$B4,$F4,$5E,$CE,$1B,$7E,$36,$BE,$36,$BF,$7D,$DE,$DF
	.dc.b	$32,$BA,$57,$C7,$FA,$2F,$EF,$9C,$DF,$16,$F7,$B7,$1A,$C1,$BF,$53
	.dc.b	$FC,$D3,$8F,$3A,$DE,$5F,$84,$6B,$FC,$91,$CE,$CC,$BB,$CF,$CD,$CF
	.dc.b	$C6,$DA,$E3,$18,$BF,$C8,$5F,$1C,$3C,$39,$79,$70,$E1,$97,$99,$FB
	.dc.b	$59,$76,$C3,$CD,$3F,$94,$F0,$BF,$B9,$FB,$3F,$E8,$2E,$BF,$F0,$F7
	.dc.b	$A3,$FE,$C4,$DF,$FA,$8A,$FA,$ED,$47,$F6,$D8,$D6,$7F,$EA,$BB,$CA
	.dc.b	$C3,$AD,$72,$BC,$C6,$27,$F9,$87,$C1,$3B,$72,$B9,$AC,$61,$B8,$7F
	.dc.b	$64,$9D,$36,$74,$FF,$1D,$FA,$2E,$F9,$FF,$80,$A3,$07,$B3,$B3,$CF
	.dc.b	$8F,$F2,$7A,$FF,$40,$3F,$D6,$FC,$04,$62,$3F,$D5,$51,$8A,$FF,$70
	.dc.b	$B6,$3F,$DC,$19,$3E,$87,$F3,$15,$FC,$F7,$E8,$BF,$C0,$8F,$F8,$1E
	.dc.b	$75,$D1,$F2,$21,$3A,$BF,$EC,$75,$A9,$E5,$E9,$FC,$E8,$C4,$FF,$68
	.dc.b	$9E,$33,$9C,$3C,$33,$C9,$BC,$5F,$7B,$EA,$75,$97,$3F,$0B,$FC,$6B
	.dc.b	$8F,$E0,$3D,$EC,$70,$7C,$A6,$C6,$4F,$54,$F0,$CB,$FD,$D9,$3A,$7D
	.dc.b	$31,$9C,$7F,$1B,$EF,$65,$FF,$57,$7A,$DF,$E4,$0C,$3F,$1F,$D9,$F5
	.dc.b	$23,$18,$7F,$BD,$E3,$1E,$6F,$DC,$6D,$70,$3F,$B0,$D7,$0B,$EB,$5F
	.dc.b	$82,$F1,$37,$87,$F8,$3B,$D5,$7D,$1B,$DD,$FC,$3F,$07,$E6,$A7,$31
	.dc.b	$8C,$FF,$DE,$55,$B7,$F4,$B5,$B8,$C5,$FF,$82,$F8,$3E,$7D,$7F,$8D
	.dc.b	$E7,$39,$FF,$B8,$1B,$1F,$AB,$F7,$B9,$63,$E1,$FA,$3F,$91,$3F,$AF
	.dc.b	$8C,$61,$F9,$C6,$C7,$ED,$9B,$8F,$85,$B8,$FF,$BB,$AF,$9C,$FE,$D6
	.dc.b	$D7,$A3,$FC,$99,$3F,$91,$AE,$1C,$BE,$3F,$6B,$C8,$DA,$FF,$30,$6A
	.dc.b	$BE,$11,$CA,$B6,$FF,$96,$BE,$8B,$FD,$08,$F0,$94,$FF,$04,$9C,$E5
	.dc.b	$F5,$67,$35,$BF,$5B,$E2,$1B,$CE,$C7,$3F,$4B,$0D,$3C,$11,$F2,$FE
	.dc.b	$42,$47,$8F,$DF,$B8,$FF,$94,$0F,$F3,$8F,$BF,$E3,$DB,$F2,$4D,$AF
	.dc.b	$C6,$B6,$78,$73,$23,$1C,$1F,$9C,$7F,$84,$4F,$4A,$FF,$33,$3F,$E6
	.dc.b	$7A,$E1,$18,$81,$B3,$7E,$23,$77,$BB,$F1,$F8,$7E,$43,$6F,$E7,$36
	.dc.b	$FF,$CB,$95,$C2,$7E,$95,$7C,$FF,$73,$E8,$7B,$B6,$7E,$B3,$6F,$8B
	.dc.b	$74,$FF,$71,$61,$FE,$89,$AF,$B9,$C3,$F6,$78,$BE,$D3,$63,$27,$E0
	.dc.b	$63,$2D,$33,$87,$9A,$32,$D8,$FC,$37,$95,$F8,$5E,$BF,$D2,$36,$32
	.dc.b	$7C,$9E,$6F,$C9,$47,$F3,$76,$C6,$79,$BF,$C1,$2F,$CF,$E0,$F9,$9C
	.dc.b	$5F,$91,$F6,$FF,$24,$D9,$FC,$C3,$72,$F9,$6D,$DB,$F8,$CF,$D1,$FE
	.dc.b	$67,$D7,$FE,$30,$DB,$FE,$26,$DB,$FC,$3B,$6B,$8C,$FE,$06,$7F,$D8
	.dc.b	$33,$F8,$8E,$53,$4F,$EA,$59,$7E,$0F,$DB,$FD,$47,$D1,$FE,$A5,$B8
	.dc.b	$7E,$BD,$B8,$7E,$09,$B8,$FF,$00,$86,$CF,$89,$7D,$07,$EC,$F4,$FE
	.dc.b	$45,$18,$C7,$F7,$51,$89,$F9,$D3,$9B,$FC,$ED,$F3,$9C,$9F,$A2,$FA
	.dc.b	$37,$D1,$FE,$FF,$8F,$FB,$8C,$63,$2F,$E0,$D1,$89,$D4,$7F,$A7,$27
	.dc.b	$58,$7E,$72,$B5,$3F,$26,$73,$3F,$AE,$6B,$FE,$D5,$6B,$F0,$31,$7C
	.dc.b	$E7,$F8,$53,$E5,$FA,$EB,$F1,$74,$F8,$93,$9C,$FE,$1B,$F0,$09,$97
	.dc.b	$E4,$70,$E6,$DF,$73,$A7,$BF,$F2,$9B,$1F,$CA,$25,$1F,$28,$91,$C4
	.dc.b	$E9,$F1,$DF,$E2,$CC,$9F,$46,$F0,$FE,$AC,$E6,$31,$5D,$63,$EA,$91
	.dc.b	$E6,$7E,$87,$5F,$B1,$7F,$83,$E5,$46,$27,$F8,$9C,$63,$E8,$FE,$37
	.dc.b	$1C,$9F,$E2,$F7,$C6,$F8,$70,$FF,$90,$55,$EE,$3F,$B8,$6A,$F8,$3F
	.dc.b	$F1,$AF,$C4,$7E,$39,$B8,$FC,$57,$F1,$A4,$7E,$73,$FC,$B3,$E8,$BD
	.dc.b	$F6,$CB,$FC,$B5,$F4,$6C,$3F,$C7,$B5,$A9,$CC,$63,$83,$91,$31,$FD
	.dc.b	$95,$74,$C3,$F9,$BB,$C6,$3F,$9A,$C2,$30,$FE,$AB,$E8,$FD,$93,$FE
	.dc.b	$F2,$9C,$CE,$EB,$C1,$3F,$DA,$2B,$73,$FD,$46,$B7,$F4,$6E,$BF,$F2
	.dc.b	$27,$D1,$DC,$3F,$DF,$B3,$6F,$FC,$81,$37,$3F,$DE,$42,$74,$FF,$50
	.dc.b	$6B,$EA,$6B,$FD,$41,$BE,$CB,$E8,$DE,$C7,$E9,$7E,$37,$FC,$A1,$18
	.dc.b	$CB,$F0,$27,$E1,$B9,$13,$AA,$E1,$3F,$E5,$FA,$E1,$8F,$D7,$8E,$15
	.dc.b	$B7,$FB,$16,$1E,$1B,$FF,$38,$36,$33,$7C,$AF,$E9,$CF,$9F,$D5,$FD
	.dc.b	$D7,$67,$99,$B7,$E1,$6D,$7F,$9B,$2F,$FC,$6A,$FE,$5E,$B8,$56,$CF
	.dc.b	$6C,$7E,$6F,$DE,$CB,$C6,$EE,$BF,$A1,$9B,$5F,$DE,$B1,$7C,$F7,$E1
	.dc.b	$BA,$BF,$5B,$F4,$7F,$E7,$98,$C6,$E8,$C7,$2F,$F1,$36,$E1,$E2,$BE
	.dc.b	$2F,$FA,$26,$FF,$23,$87,$6B,$FE,$35,$DD,$C3,$5F,$F4,$4D,$7F,$80
	.dc.b	$5F,$F2,$A7,$D1,$F8,$7D,$3F,$F5,$2B,$E3,$F4,$7F,$D4,$FF,$37,$73
	.dc.b	$FF,$60,$BF,$EA,$6B,$FF,$0A,$FC,$47,$B7,$E1,$7F,$C3,$0D,$8F,$F0
	.dc.b	$EB,$5F,$D3,$B7,$E3,$9F,$47,$FB,$79,$FF,$5F,$57,$EB,$27,$3C,$9B
	.dc.b	$AD,$70,$7F,$B1,$6F,$FC,$4B,$5A,$78,$9F,$EC,$7D,$9E,$BF,$ED,$B0
	.dc.b	$FF,$6D,$FD,$1D,$D7,$FD,$B0,$DB,$E1,$57,$1F,$AF,$EB,$67,$91,$EE
	.dc.b	$3E,$77,$0D,$C3,$E9,$57,$1D,$1E,$ED,$24,$FD,$9E,$9F,$82,$FD,$98
	.dc.b	$ED,$0F,$2E,$9C,$27,$EC,$CE,$1E,$51,$9D,$FF,$BD,$69,$F8,$EF,$0F
	.dc.b	$C4,$4A,$F0,$7C,$E5,$7F,$07,$16,$BF,$9C,$4D,$E1,$FD,$D7,$D1,$F9
	.dc.b	$0C,$5F,$CD,$BC,$3E,$B8,$CE,$1F,$9E,$8F,$F0,$D3,$5F,$DD,$E9,$FE
	.dc.b	$1D,$7E,$B1,$A7,$FC,$C7,$8F,$F4,$9F,$A3,$F8,$67,$CB,$98,$7C,$97
	.dc.b	$2F,$E8,$4F,$B4,$3E,$15,$1D,$DB,$1F,$E2,$D6,$BF,$E6,$F6,$BF,$E7
	.dc.b	$47,$33,$FC,$AB,$E8,$FF,$D2,$B3,$FC,$08,$38,$7F,$D1,$13,$C3,$3F
	.dc.b	$E8,$BF,$46,$F9,$1F,$2B,$6B,$93,$5F,$F4,$87,$D1,$B4,$DD,$49,$DB
	.dc.b	$EE,$9D,$46,$0F,$F2,$45,$73,$7F,$D1,$AD,$7C,$73,$F4,$63,$C1,$97
	.dc.b	$F4,$6A,$FE,$E5,$5E,$D6,$BF,$EA,$39,$DE,$1F,$E9,$4F,$5F,$FC,$8D
	.dc.b	$8F,$F9,$5D,$B5,$FE,$21,$8C,$7A,$5E,$68,$FF,$21,$B7,$1F,$F0,$77
	.dc.b	$37,$F9,$3E,$7F,$8E,$35,$FF,$02,$D7,$E9,$3B,$64,$7E,$A5,$FD,$3F
	.dc.b	$A5,$88,$FE,$94,$7F,$83,$E3,$17,$FE,$13,$9F,$DF,$62,$FA,$37,$83
	.dc.b	$E9,$6E,$1F,$E8,$96,$BF,$E5,$A6,$BF,$E0,$09,$E5,$3F,$C2,$6D,$F2
	.dc.b	$5F,$EC,$1A,$FE,$8F,$18,$F8,$0F,$0E,$1D,$23,$FC,$98,$FB,$E7,$A3
	.dc.b	$5F,$9C,$D7,$FC,$EA,$DC,$3D,$5A,$FF,$15,$FA,$17,$C7,$53,$BC,$78
	.dc.b	$38,$FD,$C4,$C3,$DC,$D7,$E4,$F8,$0B,$FF,$14,$47,$F4,$F6,$BF,$3B
	.dc.b	$0F,$95,$5F,$B0,$FA,$1F,$F2,$CC,$FF,$A6,$5A,$FF,$B6,$DA,$FF,$B8
	.dc.b	$1A,$FF,$8A,$9D,$E7,$FE,$51,$BF,$8D,$C7,$FE,$50,$6D,$7F,$B9,$2B
	.dc.b	$73,$C7,$CA,$F8,$5C,$12,$BF,$6D,$5A,$7E,$4F,$A3,$F7,$C8,$FD,$F6
	.dc.b	$BF,$0B,$0F,$E7,$8D,$F0,$FE,$CC,$FC,$FF,$9D,$FA,$1B,$CA,$F8,$D1
	.dc.b	$97,$E4,$46,$5E,$91,$F1,$FC,$0F,$E4,$58,$9A,$49,$F9,$4F,$F5,$5F
	.dc.b	$80,$7D,$1F,$F4,$59,$FF,$09,$C6,$3C,$3E,$91,$F0,$2B,$85,$6E,$3F
	.dc.b	$4B,$5B,$74,$7D,$DF,$A3,$FC,$57,$14,$8C,$10,$F8,$8E,$5F,$8D,$1F
	.dc.b	$9C,$E1,$FD,$51,$23,$F5,$4D,$7E,$66,$BF,$97,$38,$3F,$8D,$EF,$FC
	.dc.b	$D7,$C9,$48,$C1,$0F,$A8,$E5,$FA,$D1,$FA,$13,$1C,$9F,$89,$9C,$4F
	.dc.b	$E6,$7D,$4F,$D3,$7C,$94,$8C,$10,$FC,$03,$97,$F0,$51,$FA,$DF,$90
	.dc.b	$FD,$77,$C8,$C9,$FD,$24,$65,$FD,$34,$65,$FC,$54,$7E,$97,$C1,$FB
	.dc.b	$E4,$8F,$D3,$35,$FC,$86,$9F,$BB,$38,$3F,$9A,$8C,$3F,$B3,$8C,$BF
	.dc.b	$B3,$FA,$1F,$F0,$FB,$FE,$56,$EB,$46,$23,$5E,$77,$F8,$6F,$E6,$1F
	.dc.b	$47,$FC,$92,$9E,$86,$7C,$79,$87,$39,$CB,$7E,$43,$81,$1F,$F2,$A6
	.dc.b	$1F,$E2,$66,$FB,$3D,$7C,$69,$18,$3A,$3E,$1F,$A3,$91,$83,$C0,$E1
	.dc.b	$F7,$3C,$9F,$81,$F1,$9E,$92,$3E,$2F,$95,$AF,$F1,$CC,$3F,$BB,$73
	.dc.b	$FF,$B7,$5B,$C3,$FD,$6F,$46,$1C,$2B,$5F,$2B,$F1,$CC,$56,$F0,$F9
	.dc.b	$04,$3E,$0F,$BB,$5F,$07,$B9,$F3,$CC,$5A,$B9,$4F,$C8,$9D,$3F,$3C
	.dc.b	$FF,$0D,$FC,$FF,$F3,$0A,$FF,$5D,$B6,$BF,$DA,$2D,$97,$96,$D3,$F0
	.dc.b	$17,$C3,$E2,$3F,$01,$C9,$2F,$FC,$35,$87,$14,$F3,$C3,$94,$EE,$7F
	.dc.b	$B6,$9A,$AF,$F3,$37,$D1,$BE,$52,$92,$7F,$07,$F2,$D5,$FA,$5C,$DF
	.dc.b	$91,$31,$9F,$F9,$62,$F9,$4F,$E2,$A7,$4F,$FA,$C3,$AD,$3A,$8F,$EF
	.dc.b	$F1,$8F,$6F,$C6,$48,$FC,$9F,$B9,$5F,$F0,$2F,$03,$2F,$F6,$AC,$7F
	.dc.b	$AF,$23,$1F,$0F,$BC,$FC,$98,$7F,$A1,$67,$35,$BB,$E2,$F8,$B2,$FF
	.dc.b	$7B,$3F,$E4,$5F,$9C,$7C,$8F,$F9,$7D,$B0,$21,$1C,$B8,$34,$E1,$C3
	.dc.b	$AC,$7F,$C5,$3E,$7F,$FA,$48,$FF,$30,$C6,$71,$ED,$C4,$FA,$3F,$F1
	.dc.b	$DD,$FE,$47,$0F,$C3,$C7,$FA,$AD,$FF,$00,$B6,$35,$3C,$7D,$9F,$F5
	.dc.b	$14,$F2,$BE,$AE,$DB,$ED,$6A,$7F,$87,$7C,$BC,$C7,$F8,$2B,$FC,$47
	.dc.b	$C0,$C4,$7E,$D6,$2E,$5F,$B5,$9F,$D0,$F9,$3F,$E5,$EF,$86,$7D,$07
	.dc.b	$E8,$23,$0D,$E6,$FF,$69,$BF,$37,$E8,$5F,$47,$F3,$2F,$8E,$7F,$47
	.dc.b	$B5,$8F,$28,$C3,$5F,$D1,$B9,$BE,$7C,$9F,$ED,$AC,$7C,$14,$C3,$AB
	.dc.b	$C3,$2F,$1B,$D1,$B7,$CA,$32,$6C,$E5,$5D,$2B,$9E,$3E,$6B,$E3,$97
	.dc.b	$AE,$B4,$4B,$EC,$72,$18,$77,$70,$77,$3F,$1D,$7D,$5F,$C8,$19,$3D
	.dc.b	$A7,$E4,$FE,$21,$EE,$BE,$8F,$72,$9C,$C6,$1B,$1D,$1A,$F8,$3D,$07
	.dc.b	$AD,$CB,$04,$62,$7C,$71,$89,$D4,$E6,$79,$47,$CA,$9D,$47,$C1,$C5
	.dc.b	$F4,$7F,$EC,$3E,$6D,$61,$D7,$7F,$9A,$DD,$7F,$95,$B7,$0F,$93,$7C
	.dc.b	$5F,$93,$8F,$6B,$E3,$97,$43,$EE,$D6,$E3,$35,$C2,$FA,$3F,$CB,$9E
	.dc.b	$D9,$79,$BD,$4F,$DA,$57,$02,$65,$E4,$C7,$AC,$EA,$31,$87,$DF,$E9
	.dc.b	$79,$DF,$EC,$FD,$7F,$D5,$7E,$BF,$EF,$F9,$FF,$D5,$FF,$47,$67,$F8
	.dc.b	$2D,$6F,$0F,$F4,$0B,$6B,$FD,$6B,$7B,$9F,$E9,$B2,$BC,$61,$E0,$AB
	.dc.b	$F0,$6E,$CB,$CF,$86,$E7,$FC,$87,$3F,$E1,$B5,$B8,$C1,$FC,$61,$FF
	.dc.b	$71,$38,$CF,$FA,$EC,$6D,$FF,$4B,$BA,$7F,$AA,$F7,$3E,$3F,$F8,$F4
	.dc.b	$7F,$28,$6B,$FC,$BD,$AF,$F2,$E6,$A8,$E5,$5F,$CE,$E7,$DF,$AB,$DA
	.dc.b	$F8,$97,$B9,$F8,$D9,$BE,$8B,$F3,$78,$7F,$4D,$FA,$37,$A7,$A9,$CE
	.dc.b	$1F,$D5,$27,$35,$FD,$7E,$73,$3F,$D1,$AF,$A7,$3F,$A1,$B5,$FB,$86
	.dc.b	$CF,$F4,$96,$ED,$FC,$FD,$B3,$F8,$0F,$A3,$F1,$5F,$A3,$E1,$BF,$13
	.dc.b	$3F,$EF,$2F,$7B,$1B,$FC,$06,$C7,$79,$DE,$5D,$DA,$F9,$DA,$FF,$5F
	.dc.b	$C7,$FD,$AB,$7C,$A3,$F1,$FE,$F5,$CF,$EB,$FA,$CF,$3F,$E7,$F5,$7C
	.dc.b	$E8,$C6,$7F,$83,$E8,$FC,$FC,$E6,$BE,$CD,$6B,$1F,$E3,$6D,$7F,$B6
	.dc.b	$35,$FE,$E2,$EF,$0F,$0C,$FD,$EF,$17,$A6,$3F,$E6,$BF,$A3,$7B,$DF
	.dc.b	$82,$E1,$FD,$27,$BC,$9C,$D7,$97,$7F,$BB,$97,$F6,$FA,$D6,$1F,$90
	.dc.b	$6D,$7F,$30,$6E,$9F,$E0,$16,$BF,$E0,$67,$35,$CF,$0E,$F3,$FB,$8F
	.dc.b	$7F,$C7,$18,$9F,$07,$D1,$7D,$67,$FB,$DE,$59,$C7,$FC,$25,$8A,$B5
	.dc.b	$1F,$C1,$32,$F6,$C7,$DA,$C9,$5B,$BF,$54,$E6,$31,$7E,$B8,$E7,$5C
	.dc.b	$1B,$C1,$F8,$E9,$FF,$4B,$46,$AB,$C5,$D1,$E0,$3C,$3E,$DD,$AD,$BE
	.dc.b	$02,$3D,$DF,$DD,$DB,$E8,$FF,$D3,$B3,$C0,$FC,$43,$8C,$5A,$7C,$6B
	.dc.b	$F7,$FB,$32,$F7,$A5,$BD,$C7,$C8,$D3,$FC,$94,$FB,$FD,$4F,$86,$47
	.dc.b	$E1,$7A,$FF,$28,$FE,$4F,$9B,$DF,$C4,$FF,$1D,$C3,$FD,$23,$E8,$EB
	.dc.b	$7F,$D0,$36,$75,$CE,$8F,$CF,$92,$3F,$B9,$C6,$5B,$1C,$9F,$D4,$7D
	.dc.b	$9B,$D9,$C6,$1E,$C3,$FD,$13,$93,$DE,$F3,$F0,$FA,$1B,$87,$CA,$6D
	.dc.b	$71,$E3,$FA,$11,$8E,$3F,$F2,$4E,$0F,$82,$F8,$3E,$FF,$C2,$6E,$3F
	.dc.b	$51,$B8,$F3,$AE,$13,$F6,$2B,$E4,$F1,$7C,$AE,$5B,$3F,$71,$B7,$C5
	.dc.b	$B8,$F8,$9B,$3F,$E6,$4C,$3F,$DB,$EF,$E1,$2F,$A3,$F9,$4F,$37,$EB
	.dc.b	$F6,$FE,$89,$B1,$92,$63,$E7,$E5,$D4,$7E,$3B,$8B,$EF,$FB,$BF,$80
	.dc.b	$6C,$64,$FC,$9C,$65,$A7,$DE,$FA,$4F,$C3,$FB,$3F,$C8,$5B,$19,$3E
	.dc.b	$FB,$FC,$80,$8E,$0D,$61,$F2,$FE,$0B,$94,$7E,$8B,$DF,$FC,$9F,$D2
	.dc.b	$7E,$51,$B1,$93,$F4,$F1,$96,$9F,$DC,$FB,$9F,$12,$FA,$1F,$CE,$F6
	.dc.b	$7F,$24,$8C,$57,$EC,$23,$13,$23,$F6,$79,$DC,$FF,$D9,$BF,$48,$FA
	.dc.b	$4F,$D3,$F1,$7F,$6E,$8C,$5F,$EE,$A3,$13,$FD,$E2,$73,$5F,$E5,$A6
	.dc.b	$BF,$EA,$B6,$BF,$EB,$07,$7C,$FF,$E0,$66,$DF,$E1,$67,$AB,$EF,$CD
	.dc.b	$F3,$9F,$E2,$0F,$98,$DD,$7F,$24,$9C,$D7,$F5,$59,$CE,$DF,$1A,$FE
	.dc.b	$21,$C7,$EA,$AD,$75,$F8,$1E,$9F,$07,$F2,$5B,$19,$26,$5F,$AF,$C5
	.dc.b	$F5,$3B,$9F,$37,$C7,$FF,$4D,$D0,$D2,$47,$F4,$F8,$CB,$C0,$F4,$71
	.dc.b	$7D,$99,$F7,$32,$3D,$9F,$4B,$71,$F1,$4D,$E0,$FF,$63,$E4,$AC,$5F
	.dc.b	$F4,$EC,$5F,$0F,$E5,$24,$78,$DB,$8F,$F1,$3E,$0F,$C6,$C7,$CF,$FA
	.dc.b	$AF,$C6,$3F,$E3,$0F,$A4,$FE,$55,$7F,$E9,$26,$DF,$FA,$69,$AF,$FB
	.dc.b	$AD,$AF,$FB,$C1,$DE,$7F,$E4,$66,$CF,$F6,$4F,$A5,$F3,$BF,$CD,$63
	.dc.b	$35,$FE,$F9,$6A,$73,$8A,$B8,$CF,$F2,$D3,$5E,$67,$A4,$97,$FA,$B9
	.dc.b	$DE,$1E,$39,$FF,$26,$FD,$23,$B3,$FC,$AA,$DB,$FF,$58,$B7,$0F,$F5
	.dc.b	$CB,$51,$03,$BB,$FF,$32,$B6,$76,$3E,$95,$D9,$E2,$C3,$FC,$BB,$F4
	.dc.b	$AF,$6F,$F4,$5E,$76,$1B,$87,$F9,$DF,$DA,$56,$E3,$FC,$AC,$DD,$3F
	.dc.b	$CA,$CD,$7F,$CB,$CD,$7F,$C9,$0E,$F4,$FF,$2D,$36,$7D,$BE,$77,$F5
	.dc.b	$7F,$91,$F1,$47,$86,$7F,$C6,$2E,$B9,$FF,$6B,$87,$F5,$FC,$7C,$BE
	.dc.b	$C7,$EE,$FB,$FD,$AF,$F9,$B7,$0F,$F3,$8D,$78,$A3,$04,$C3,$B3,$FC
	.dc.b	$0F,$E0,$38,$C6,$09,$3F,$B7,$8C,$BE,$1F,$A0,$FC,$5E,$8C,$DF,$4B
	.dc.b	$FE,$8A,$9C,$C6,$30,$F9,$91,$F9,$AB,$68,$8C,$37,$97,$FD,$0D,$18
	.dc.b	$3F,$33,$19,$AF,$E2,$DE,$07,$7A,$F0,$7A,$7C,$09,$3F,$E8,$DB,$FF
	.dc.b	$02,$37,$D2,$EA,$47,$07,$E3,$F1,$83,$E2,$BF,$D0,$5F,$0F,$86,$C6
	.dc.b	$DA,$FE,$5D,$BF,$1C,$FA,$4F,$F1,$BC,$FC,$0A,$FC,$E4,$E7,$9B,$75
	.dc.b	$AD,$BF,$DA,$57,$FA,$2A,$D3,$C4,$F9,$39,$FF,$85,$5F,$77,$AF,$E8
	.dc.b	$72,$DA,$F9,$CD,$97,$AF,$49,$5F,$1D,$D7,$A9,$FB,$1C,$BF,$D6,$54
	.dc.b	$8C,$6B,$FE,$A5,$8C,$5F,$F4,$9C,$F2,$F9,$AB,$9B,$F4,$DA,$FF,$AF
	.dc.b	$E7,$8B,$FD,$17,$C8,$E3,$CB,$F8,$D7,$9E,$9F,$87,$BB,$7F,$C3,$6F
	.dc.b	$F1,$BD,$B8,$38,$F2,$BE,$67,$13,$A2,$1F,$D4,$DC,$BE,$39,$F8,$C4
	.dc.b	$7F,$14,$F4,$FA,$39,$1C,$B6,$F8,$AF,$90,$87,$FA,$FF,$F1,$47,$D3
	.dc.b	$62,$3E,$01,$F4,$C8,$7E,$5C,$F4,$FD,$2C,$7F,$52,$7C,$C1,$F0,$C8
	.dc.b	$C3,$63,$FC,$E8,$D7,$F9,$D7,$D2,$BA,$77,$5F,$87,$9E,$6D,$AE,$11
	.dc.b	$FA,$36,$BF,$CC,$FE,$91,$F2,$3E,$36,$C7,$FA,$0D,$AF,$F4,$9F,$A5
	.dc.b	$FD,$FB,$99,$E0,$47,$FD,$15,$1A,$7F,$A8,$4F,$2D,$1C,$FE,$1B,$E1
	.dc.b	$5D,$79,$2B,$FB,$9B,$F0,$B6,$7D,$D1,$88,$FE,$F9,$3F,$96,$F3,$3F
	.dc.b	$C8,$6D,$F4,$3C,$58,$FF,$81,$3E,$53,$13,$FA,$DA,$DD,$FF,$84,$E3
	.dc.b	$17,$E9,$8F,$EB,$1D,$3F,$C9,$7B,$BF,$26,$48,$FF,$17,$7A,$15,$FF
	.dc.b	$12,$45,$CB,$FC,$27,$3E,$5F,$A5,$7D,$2D,$25,$FF,$8D,$EE,$BF,$EF
	.dc.b	$19,$DB,$FE,$F5,$6B,$99,$F5,$47,$F9,$8A,$3F,$CC,$8D,$7F,$C3,$4D
	.dc.b	$BF,$F1,$2E,$5E,$3C,$BF,$6B,$F4,$AF,$90,$A6,$70,$E7,$C1,$E9,$23
	.dc.b	$FD,$F3,$0F,$B6,$B5,$5B,$E8,$37,$3F,$EB,$58,$79,$DE,$03,$37,$CA
	.dc.b	$5F,$48,$F2,$BF,$4C,$D8,$FF,$6B,$35,$E9,$D7,$A9,$FD,$14,$E6,$7F
	.dc.b	$A1,$35,$FF,$09,$35,$FF,$0A,$BC,$EB,$D7,$97,$AB,$83,$E2,$9D,$6B
	.dc.b	$E6,$FA,$1F,$6C,$8F,$C3,$E7,$FF,$01,$E8,$FA,$6C,$CF,$C1,$FA,$6F
	.dc.b	$9E,$63,$2E,$F4,$FA,$2F,$A7,$C7,$2F,$A1,$F3,$67,$4E,$34,$F9,$F5
	.dc.b	$B8,$F9,$D5,$F7,$23,$E4,$F8,$72,$7E,$4B,$F8,$0F,$9C,$C4,$7F,$66
	.dc.b	$DF,$F9,$73,$FD,$45,$AE,$67,$E6,$9F,$D1,$3E,$0D,$F4,$DF,$E7,$E9
	.dc.b	$F3,$C6,$3C,$AE,$BA,$FB,$BE,$6F,$F1,$FD,$12,$FD,$5B,$A7,$F2,$66
	.dc.b	$63,$F1,$13,$87,$F3,$BC,$5F,$AB,$23,$CB,$E6,$6B,$F8,$FE,$4F,$CE
	.dc.b	$98,$C7,$E7,$F8,$7F,$A0,$F9,$99,$5E,$28,$FD,$19,$87,$F3,$11,$87
	.dc.b	$47,$A1,$D4,$F4,$8F,$5F,$AF,$FA,$CF,$99,$91,$82,$1F,$A3,$72,$FE
	.dc.b	$96,$3F,$67,$F2,$FF,$B4,$F9,$99,$3F,$1A,$32,$FC,$88,$CB,$FA,$D8
	.dc.b	$FD,$3F,$83,$23,$F4,$EF,$F7,$2F,$95,$FF,$0D,$46,$0F,$BD,$1F,$A8
	.dc.b	$23,$83,$EA,$ED,$FC,$18,$E8,$FF,$A9,$7E,$97,$FC,$30,$DF,$5B,$F9
	.dc.b	$94,$63,$4F,$D5,$F8,$6F,$A6,$CF,$AF,$FC,$C6,$BF,$C3,$38,$7F,$8D
	.dc.b	$BC,$A4,$62,$3F,$C7,$51,$89,$E1,$F2,$9D,$89,$3F,$A3,$C1,$F3,$5F
	.dc.b	$E1,$05,$EE,$F9,$3E,$56,$56,$E2,$BD,$B2,$4F,$94,$91,$E5,$87,$C9
	.dc.b	$FB,$89,$1F,$B6,$CF,$D3,$5F,$0B,$E1,$64,$E6,$FF,$D3,$6F,$9F,$E6
	.dc.b	$72,$F8,$67,$E2,$11,$F0,$3E,$BF,$99,$91,$CE,$3E,$7B,$0E,$C7,$C1
	.dc.b	$F8,$2C,$3E,$D3,$E8,$10,$F6,$1F,$1F,$D8,$8F,$23,$F8,$88,$FA,$2F
	.dc.b	$E7,$2E,$1A,$FF,$B3,$5B,$C7,$F8,$0E,$4F,$F0,$F4,$6A,$7F,$CE,$8D
	.dc.b	$7F,$D8,$8D,$7F,$D8,$ED,$7C,$1E,$8F,$DB,$E4,$91,$E1,$8F,$F2,$5F
	.dc.b	$D3,$3E,$66,$92,$7F,$09,$F3,$55,$FF,$42,$75,$2B,$84,$7F,$A1,$A3
	.dc.b	$15,$FE,$6F,$9C,$CF,$06,$E7,$F9,$CE,$8F,$C7,$1D,$5F,$EC,$77,$D7
	.dc.b	$E0,$3C,$A4,$BF,$CB,$7C,$1D,$7A,$47,$72,$3E,$3C,$3F,$CC,$8A,$FF
	.dc.b	$B7,$F6,$33,$7C,$5F,$F9,$DA,$30,$F5,$9F,$ED,$9F,$40,$F9,$5F,$F2
	.dc.b	$83,$70,$FB,$AD,$7B,$5F,$FB,$73,$B7,$DE,$FF,$BE,$67,$53,$FE,$AD
	.dc.b	$6B,$9A,$FF,$29,$46,$23,$FC,$09,$F4,$CF,$7B,$FD,$06,$FF,$A9,$23
	.dc.b	$B9,$AF,$7B,$33,$23,$FA,$F4,$EF,$2F,$EE,$AD,$C7,$FD,$E2,$F5,$9F
	.dc.b	$E0,$5F,$39,$FC,$04,$8F,$F0,$5E,$EA,$F8,$3D,$8F,$F0,$B4,$E3,$0F
	.dc.b	$E2,$F5,$F9,$8F,$27,$FC,$BF,$F1,$6F,$81,$FC,$BC,$61,$BC,$CC,$BF
	.dc.b	$D9,$D1,$7E,$8D,$F4,$DF,$37,$0E,$55,$F9,$EC,$39,$37,$0E,$53,$A8
	.dc.b	$FD,$84,$EA,$F9,$F1,$7F,$B5,$71,$F0,$D7,$80,$E4,$53,$8E,$1D,$23
	.dc.b	$C9,$8F,$8D,$E3,$8F,$7E,$D6,$5E,$4C,$38,$9D,$72,$F0,$BC,$CE,$F5
	.dc.b	$C6,$78,$57,$2C,$75,$7C,$2F,$C9,$39,$7F,$DE,$AD,$7F,$82,$D7,$F9
	.dc.b	$C5,$F5,$9F,$EF,$BF,$8D,$FF,$22,$7D,$37,$FB,$95,$B1,$FE,$FE,$6B
	.dc.b	$BF,$82,$4F,$FB,$AE,$DE,$83,$D9,$C3,$6F,$C6,$D7,$F8,$BB,$5F,$E2
	.dc.b	$EE,$F4,$F9,$0D,$D7,$F0,$1F,$4D,$FE,$BC,$E6,$F8,$D7,$FC,$A6,$B4
	.dc.b	$FF,$1E,$6B,$9C,$3E,$3D,$71,$8F,$8F,$3A,$C7,$D3,$5D,$1B,$3F,$56
	.dc.b	$B8,$63,$E4,$AD,$5F,$13,$C3,$B2,$B7,$97,$B2,$B6,$4C,$3A,$7A,$9F
	.dc.b	$E7,$EC,$7A,$47,$E3,$F9,$9F,$50,$F9,$AF,$F4,$1F,$3F,$F9,$AA,$31
	.dc.b	$BF,$F9,$5D,$7C,$E6,$A3,$F5,$17,$CA,$F1,$87,$FA,$2E,$F8,$BF,$B7
	.dc.b	$6E,$1D,$70,$DD,$F1,$C3,$B7,$CD,$E2,$DF,$39,$63,$0D,$9F,$23,$71
	.dc.b	$EC,$F8,$97,$7B,$BF,$BF,$81,$FE,$02,$BF,$E5,$31,$8F,$4B,$85,$6E
	.dc.b	$7F,$82,$5F,$F2,$7A,$FF,$4F,$3F,$CA,$DA,$FF,$09,$9E,$0F,$F8,$83
	.dc.b	$C2,$7C,$2F,$AF,$E6,$3E,$2F,$A9,$FD,$02,$31,$8F,$F8,$46,$B6,$FC
	.dc.b	$FA,$DC,$62,$FE,$54,$62,$A5,$7D,$5F,$97,$B5,$FC,$D3,$0F,$E9,$BF
	.dc.b	$0D,$F2,$FE,$5F,$9B,$6F,$C0,$DA,$E1,$92,$31,$C9,$FE,$C6,$C3,$9D
	.dc.b	$F0,$3F,$09,$DF,$D1,$F5,$7B,$FF,$67,$E5,$FD,$2D,$8C,$9F,$8B,$8C
	.dc.b	$B4,$CE,$5E,$29,$FC,$17,$C2,$EA,$3C,$BF,$01,$F8,$5F,$87,$F8,$AD
	.dc.b	$BF,$94,$D9,$DC,$7E,$66,$76,$E8,$FC,$E7,$CA,$E2,$7E,$47,$DC,$7F
	.dc.b	$94,$27,$F4,$1F,$29,$F6,$23,$13,$F9,$B8,$C7,$C0,$7E,$5D,$B5,$C0
	.dc.b	$FF,$26,$57,$0D,$FE,$53,$63,$F5,$3D,$9F,$D5,$5D,$DF,$E8,$7D,$19
	.dc.b	$1C,$BD,$AB,$97,$99,$B8,$F4,$6C,$F4,$EA,$F4,$47,$EA,$DB,$1F,$ED
	.dc.b	$1F,$C6,$5F,$25,$FB,$26,$C7,$EF,$FC,$FC,$4F,$C1,$F6,$BF,$69,$16
	.dc.b	$74,$73,$F6,$F9,$1C,$FC,$1B,$1F,$E6,$DF,$C6,$5F,$25,$FC,$03,$81
	.dc.b	$F1,$5F,$A4,$9C,$D7,$FB,$96,$72,$F1,$AE,$15,$B6,$C6,$4F,$F4,$95
	.dc.b	$72,$E0,$FF,$48,$B5,$FD,$DF,$62,$31,$1F,$C6,$B6,$E1,$C2,$BC,$7C
	.dc.b	$9F,$C5,$C7,$E3,$36,$3E,$53,$6F,$FC,$29,$68,$F9,$9F,$25,$F3,$7E
	.dc.b	$4E,$DF,$D1,$6D,$F0,$6E,$3E,$18,$79,$4C,$55,$F5,$FE,$4F,$EC,$78
	.dc.b	$D7,$1F,$B8,$DC,$3C,$6D,$C7,$A6,$3C,$E3,$15,$F0,$5B,$5E,$0C,$7F
	.dc.b	$88,$5F,$F0,$1F,$C0,$FF,$CA,$57,$F8,$9F,$13,$F8,$DC,$62,$FE,$0C
	.dc.b	$63,$DB,$FE,$4B,$C2,$AD,$57,$D2,$6B,$FE,$31,$8D,$4F,$F2,$CC,$FE
	.dc.b	$07,$17,$E5,$7E,$47,$F2,$CD,$CB,$F3,$4D,$C3,$E8,$B7,$3F,$95,$0F
	.dc.b	$A2,$FE,$A3,$5F,$D6,$39,$5F,$23,$FD,$1E,$73,$5F,$42,$73,$5F,$EA
	.dc.b	$A9,$CE,$5F,$ED,$4B,$E1,$E9,$7E,$6D,$B5,$FD,$23,$0F,$E9,$37,$FD
	.dc.b	$89,$C3,$5C,$D7,$BF,$83,$BC,$7E,$2B,$C7,$FD,$A6,$1F,$B5,$6E,$1F
	.dc.b	$B9,$6C,$EE,$3F,$DB,$5C,$0F,$33,$F7,$9E,$A7,$D6,$6C,$E9,$FA,$B3
	.dc.b	$B8,$FF,$7D,$36,$BF,$DF,$2D,$7D,$3E,$8F,$F0,$7F,$67,$E4,$36,$BE
	.dc.b	$33,$67,$FD,$0A,$7D,$17,$1C,$1F,$53,$6F,$EF,$B5,$F1,$EB,$FC,$E5
	.dc.b	$18,$70,$D7,$C3,$1B,$F6,$DE,$03,$C8,$FE,$35,$F2,$3F,$E1,$29,$CC
	.dc.b	$62,$BF,$CF,$D8,$FE,$B6,$73,$1C,$EF,$93,$6B,$FB,$EC,$EB,$0F,$87
	.dc.b	$E5,$FF,$86,$3F,$1C,$7C,$8F,$E8,$A3,$0D,$C3,$E5,$C7,$CE,$23,$EF
	.dc.b	$A7,$E3,$8F,$A6,$F4,$7E,$39,$FE,$E3,$AF,$F7,$CC,$E5,$F4,$9F,$E9
	.dc.b	$7C,$93,$BB,$F9,$F5,$B7,$E6,$35,$FC,$A3,$5F,$F0,$EB,$5F,$F5,$16
	.dc.b	$3C,$F5,$FD,$4B,$70,$FF,$1F,$F8,$6F,$8A,$FF,$17,$4E,$6F,$8F,$95
	.dc.b	$83,$FC,$7B,$5B,$BF,$4C,$E9,$AB,$E9,$7F,$B2,$3F,$EE,$7F,$AF,$79
	.dc.b	$9F,$E3,$B9,$CE,$69,$CD,$F2,$9F,$EE,$B7,$FB,$16,$BF,$E3,$A6,$BF
	.dc.b	$EB,$F6,$F8,$9F,$C5,$7D,$3F,$F2,$2F,$D4,$FF,$1C,$6D,$FE,$B1,$FD
	.dc.b	$39,$1F,$D3,$B6,$FF,$C6,$1D,$68,$FC,$37,$87,$CE,$B9,$12,$BF,$DB
	.dc.b	$5C,$DF,$03,$1F,$14,$FE,$EB,$D0,$FF,$2B,$F1,$7F,$2E,$6C,$64,$FD
	.dc.b	$BB,$FB,$A2,$3F,$AB,$E9,$FF,$3E,$FB,$5F,$97,$F9,$4F,$F3,$AC,$62
	.dc.b	$7E,$9F,$8C,$C9,$5A,$F5,$58,$71,$BF,$B3,$87,$1D,$3D,$37,$BA,$FF
	.dc.b	$31,$CE,$67,$FC,$D2,$D7,$FD,$DC,$D7,$F7,$36,$EC,$FD,$67,$C9,$78
	.dc.b	$5B,$8F,$82,$74,$FA,$32,$56,$E3,$0D,$B7,$2F,$F6,$BC,$7F,$D3,$F8
	.dc.b	$FE,$E7,$D5,$D4,$E7,$2E,$B7,$C3,$1F,$3C,$62,$7F,$B4,$E4,$BE,$35
	.dc.b	$AC,$7F,$82,$7C,$97,$C9,$6C,$7A,$5B,$8F,$53,$E0,$B8,$C3,$C4,$F1
	.dc.b	$9F,$77,$C7,$DA,$E1,$3F,$E0,$3E,$9F,$F4,$DF,$CA,$7D,$CF,$2F,$F4
	.dc.b	$75,$D7,$89,$EB,$AB,$5E,$AF,$FA,$8B,$C4,$BA,$7E,$0A,$BF,$84,$6B
	.dc.b	$FE,$9C,$F8,$5F,$AF,$CF,$FC,$D5,$B9,$FF,$5D,$6B,$FE,$A6,$6F,$C5
	.dc.b	$BD,$FF,$A3,$3F,$CD,$AB,$FD,$57,$39,$BD,$D6,$B8,$3F,$D6,$D6,$15
	.dc.b	$B7,$91,$FD,$03,$7F,$F2,$5B,$E9,$F8,$7F,$F5,$BE,$1F,$E1,$EC,$D1
	.dc.b	$8D,$3F,$23,$7C,$2F,$E5,$BF,$E2,$C7,$E1,$1B,$3F,$C6,$35,$FD,$EE
	.dc.b	$FF,$15,$3B,$9E,$3E,$C6,$F2,$FF,$1B,$36,$4F,$3E,$1F,$D7,$7C,$9E
	.dc.b	$63,$1C,$EF,$93,$A6,$71,$FD,$7E,$1C,$9F,$D0,$46,$9C,$D7,$7F,$52
	.dc.b	$47,$07,$AF,$91,$88,$F7,$C5,$D3,$F3,$73,$76,$FF,$7D,$E4,$F9,$0F
	.dc.b	$29,$9C,$3E,$86,$18,$7F,$65,$19,$DB,$FC,$AD,$F0,$5F,$D7,$5B,$3F
	.dc.b	$C4,$F5,$4E,$5B,$19,$8F,$87,$F1,$12,$3E,$0C,$7C,$B3,$0F,$F2,$88
	.dc.b	$C3,$FD,$33,$C4,$7E,$3B,$FE,$9C,$8F,$F0,$64,$63,$E0,$BF,$B8,$36
	.dc.b	$33,$3B,$7F,$C7,$11,$97,$F5,$5E,$0B,$EA,$BE,$D3,$63,$C1,$5C,$78
	.dc.b	$7F,$DC,$CF,$FA,$06,$33,$DB,$F8,$BC,$3C,$AF,$E0,$3C,$7C,$56,$EB
	.dc.b	$FD,$F0,$DB,$E7,$5B,$9E,$67,$A1,$F5,$CF,$18,$FF,$4E,$B5,$FF,$36
	.dc.b	$3C,$F8,$BF,$D9,$FF,$55,$EA,$3C,$35,$F9,$FA,$7D,$0B,$E7,$7F,$C2
	.dc.b	$D1,$86,$C6,$63,$FC,$34,$D7,$FC,$74,$DC,$7F,$CE,$3F,$44,$F8,$8E
	.dc.b	$CD,$AF,$F0,$0D,$6F,$2F,$F0,$E7,$82,$F8,$0F,$D9,$7B,$08,$C5,$FF
	.dc.b	$82,$A3,$13,$FA,$DF,$C7,$BF,$88,$4E,$6B,$6F,$F2,$FA,$DB,$9A,$E1
	.dc.b	$E2,$7C,$1F,$C7,$FF,$A9,$4F,$F8,$D5,$AF,$F9,$55,$AF,$F9,$F5,$AF
	.dc.b	$F1,$70,$F0,$D7,$FC,$9D,$F9,$07,$EF,$67,$D0,$D7,$FC,$CB,$3B,$CF
	.dc.b	$EB,$FB,$1F,$70,$93,$F9,$66,$BF,$E0,$8F,$27,$F8,$49,$F8,$A8,$F8
	.dc.b	$FF,$6D,$FA,$C9,$FF,$49,$68,$8C,$47,$9E,$31,$D9,$FE,$53,$6C,$7F
	.dc.b	$9B,$67,$96,$7F,$D8,$23,$F8,$B9,$DA,$BF,$11,$F4,$AF,$94,$FC,$3B
	.dc.b	$67,$FC,$BC,$D7,$FC,$FC,$D7,$FD,$4C,$E6,$BF,$DA,$BC,$E3,$F9,$41
	.dc.b	$67,$FC,$8F,$3B,$D3,$E8,$36,$3F,$DC,$F3,$B7,$FB,$13,$6B,$FD,$87
	.dc.b	$3B,$EA,$FE,$0F,$F1,$1F,$AA,$25,$FF,$04,$6B,$CB,$E5,$7C,$67,$EE
	.dc.b	$23,$CB,$F0,$9E,$6E,$49,$5F,$E2,$CC,$93,$A8,$C4,$E9,$F2,$C7,$CA
	.dc.b	$F9,$3F,$92,$48,$F9,$2D,$7F,$BE,$7C,$EB,$1F,$E9,$8D,$8F,$8B,$3C
	.dc.b	$DF,$F1,$EC,$65,$E3,$39,$7E,$2C,$FC,$ED,$A4,$7A,$63,$E8,$98,$7E
	.dc.b	$5C,$61,$FE,$43,$BF,$F1,$22,$3F,$CB,$1A,$FD,$2F,$88,$FB,$67,$F2
	.dc.b	$08,$FB,$BF,$29,$F7,$BD,$B9,$5F,$26,$32,$FC,$B8,$CB,$F7,$23,$2F
	.dc.b	$DE,$8C,$F0,$FE,$2A,$7F,$13,$F2,$9F,$8A,$EA,$97,$F5,$63,$2F,$D7
	.dc.b	$8C,$BF,$88,$8F,$C8,$FB,$FF,$92,$F7,$24,$FD,$D8,$CB,$F7,$E3,$2F
	.dc.b	$E4,$23,$FA,$07,$E2,$6F,$90,$FE,$A2,$D9,$FE,$AA,$D7,$FA,$8B,$5F
	.dc.b	$E1,$9C,$C8,$D6,$6E,$8F,$F1,$37,$AA,$E4,$FC,$7E,$1F,$80,$E5,$F0
	.dc.b	$E5,$23,$04,$3F,$CA,$59,$75,$8F,$D7,$FC,$C4,$9F,$CD,$3F,$B2,$F8
	.dc.b	$AF,$EB,$5E,$FF,$F9,$86,$7F,$3D,$E3,$FE,$D8,$CF,$C0,$7E,$E3,$C5
	.dc.b	$93,$FA,$37,$4F,$E9,$7D,$2E,$7E,$B4,$8C,$1F,$DE,$F0,$47,$F8,$A7
	.dc.b	$F2,$2F,$ED,$DC,$3F,$01,$B4,$D9,$7C,$23,$FD,$3B,$3A,$E1,$34,$E3
	.dc.b	$5C,$30,$F6,$F9,$F2,$FC,$4F,$87,$DB,$C4,$7F,$9B,$6F,$FD,$AA,$F6
	.dc.b	$F8,$0C,$F2,$FD,$58,$F2,$67,$FE,$E7,$74,$43,$E8,$B9,$7E,$9F,$E4
	.dc.b	$5F,$E0,$D6,$FA,$9F,$E0,$98,$C5,$75,$D7,$D7,$E9,$66,$BE,$37,$CB
	.dc.b	$4C,$3F,$C4,$98,$70,$E0,$FF,$36,$70,$23,$12,$C7,$E2,$DF,$18,$F0
	.dc.b	$9C,$BA,$3C,$44,$7E,$6F,$CB,$57,$E2,$79,$BF,$5C,$C6,$BF,$0F,$2F
	.dc.b	$05,$E6,$BE,$AD,$6B,$3F,$2E,$6F,$13,$E0,$12,$BE,$DF,$CB,$57,$E9
	.dc.b	$7B,$DF,$85,$31,$9F,$7B,$E5,$3F,$82,$9D,$3F,$12,$FE,$67,$85,$23
	.dc.b	$E3,$BF,$8F,$EE,$C4,$7D,$FF,$5B,$F2,$46,$31,$ED,$58,$9F,$C7,$CE
	.dc.b	$73,$FF,$02,$7C,$D3,$C3,$FF,$15,$B6,$BF,$C1,$06,$A3,$24,$7F,$DE
	.dc.b	$BE,$CB,$E5,$FF,$CA,$27,$F8,$DE,$B9,$4F,$E4,$FE,$5B,$F4,$DF,$2F
	.dc.b	$23,$F4,$31,$FA,$93,$0F,$E6,$63,$0F,$E9,$BE,$5F,$F5,$C4,$7F,$53
	.dc.b	$F0,$55,$FD,$04,$5D,$BF,$25,$DC,$AF,$F2,$57,$CE,$BE,$79,$E2,$7F
	.dc.b	$AE,$FE,$29,$E5,$7F,$89,$EB,$53,$98,$C5,$70,$CD,$7C,$17,$CF,$3E
	.dc.b	$7F,$E2,$47,$13,$FD,$91,$8F,$47,$8E,$1D,$31,$D4,$27,$58,$75,$C3
	.dc.b	$A7,$0F,$FB,$6F,$0E,$17,$C8,$EB,$97,$96,$FC,$91,$CE,$31,$3F,$ED
	.dc.b	$9B,$E7,$3E,$0C,$3B,$CF,$4B,$E3,$18,$9F,$6E,$5E,$C7,$C0,$75,$CB
	.dc.b	$96,$3D,$A3,$C7,$1F,$07,$1E,$F1,$C2,$FE,$13,$F8,$E6,$BE,$C3,$15
	.dc.b	$F3,$DF,$C9,$FC,$63,$D2,$7B,$BF,$96,$9F,$F7,$A4,$E6,$31,$1F,$EF
	.dc.b	$69,$DB,$B9,$D3,$C6,$DF,$37,$FE,$F4,$8C,$35,$F3,$35,$FC,$33,$5F
	.dc.b	$BA,$F1,$D3,$E4,$B7,$5F,$A1,$F3,$EE,$3F,$E6,$55,$AF,$8E,$FA,$57
	.dc.b	$BC,$39,$C7,$C9,$BE,$1D,$5A,$89,$3F,$2E,$B8,$6E,$BD,$CE,$6B,$78
	.dc.b	$77,$8F,$C4,$56,$F1,$F1,$E5,$EC,$BD,$B6,$B7,$87,$4C,$39,$1F,$A1
	.dc.b	$F1,$DD,$5F,$EA,$5D,$3F,$82,$FB,$9F,$D3,$FA,$3F,$A2,$4F,$F9,$16
	.dc.b	$72,$4C,$7F,$DB,$75,$B8,$FF,$04,$CE,$E7,$31,$FC,$43,$DA,$D4,$7F
	.dc.b	$45,$9F,$F2,$DC,$E4,$FE,$2B,$E3,$6B,$1F,$2D,$DA,$DD,$FF,$BA,$A3
	.dc.b	$15,$D2,$1D,$BF,$EB,$DC,$3F,$16,$D8,$FC,$83,$5F,$E4,$ED,$7F,$96
	.dc.b	$FA,$3C,$E3,$86,$7F,$0B,$DB,$C4,$FA,$BC,$94,$3C,$8F,$37,$CF,$E6
	.dc.b	$31,$97,$F3,$08,$C4,$FF,$7B,$CB,$FC,$67,$5A,$7F,$9C,$DF,$49,$D4
	.dc.b	$7F,$6E,$9D,$3C,$23,$9D,$FF,$3C,$F1,$EC,$FF,$8C,$AB,$FC,$8D,$3F
	.dc.b	$D0,$3D,$BF,$27,$91,$F2,$DB,$1F,$B3,$C7,$D7,$3E,$0B,$FF,$11,$5F
	.dc.b	$27,$CA,$D7,$D3,$1C,$96,$9F,$88,$EC,$C3,$FD,$DE,$9E,$2F,$DB,$F6
	.dc.b	$FE,$33,$72,$61,$E9,$8F,$16,$5F,$39,$B7,$CA,$BD,$0D,$AF,$2C,$F4
	.dc.b	$9F,$EB,$FE,$4D,$9F,$89,$EE,$7E,$2F,$DC,$FA,$CD,$AF,$0E,$7E,$78
	.dc.b	$EE,$48,$EE,$DA,$E7,$1F,$DB,$3C,$AE,$6F,$23,$FC,$4B,$F3,$EF,$95
	.dc.b	$FC,$D7,$B9,$F8,$98,$C4,$FF,$74,$8C,$4F,$C0,$9C,$DF,$E1,$1A,$FF
	.dc.b	$9E,$9A,$F6,$8E,$35,$FD,$FB,$BE,$CF,$D4,$8F,$DA,$FB,$9F,$AB,$3C
	.dc.b	$5E,$E5,$8F,$44,$FF,$98,$70,$4F,$8C,$93,$F8,$78,$CB,$FC,$EF,$BB
	.dc.b	$F8,$5F,$57,$48,$FF,$07,$B6,$FC,$47,$F9,$17,$E3,$5F,$19,$FB,$F6
	.dc.b	$D7,$F8,$5B,$CC,$C4,$FC,$7F,$7B,$E2,$12,$7C,$4F,$E5,$F1,$5B,$DC
	.dc.b	$E8,$75,$F7,$6D,$FB,$BA,$5F,$3E,$F6,$FF,$88,$71,$BD,$6F,$DF,$4E
	.dc.b	$6F,$E2,$E7,$F6,$E7,$37,$F9,$6A,$D6,$FF,$BA,$6B,$FE,$B0,$6B,$FE
	.dc.b	$68,$77,$D1,$FE,$5B,$6E,$1F,$EB,$2C,$7F,$39,$39,$BF,$8C,$7C,$7F
	.dc.b	$76,$D7,$93,$64,$63,$1F,$9F,$9F,$C4,$8C,$61,$FE,$7F,$AD,$DF,$1F
	.dc.b	$6F,$E9,$7B,$9F,$4D,$B1,$93,$B4,$65,$A7,$87,$DB,$FB,$5E,$37,$DB
	.dc.b	$6C,$64,$F3,$46,$5A,$67,$3F,$3E,$1F,$C8,$F1,$F3,$5F,$29,$FF,$66
	.dc.b	$FC,$BF,$F9,$66,$FE,$07,$8B,$FC,$CE,$31,$3C,$A3,$1E,$CC,$AF,$8D
	.dc.b	$EE,$5A,$EF,$C7,$F9,$1F,$85,$FC,$93,$6B,$E6,$36,$77,$1F,$3E,$7F
	.dc.b	$A8,$7C,$FF,$F6,$99,$F9,$FC,$9F,$E5,$9A,$FD,$5C,$27,$5A,$FF,$AB
	.dc.b	$EB,$53,$FA,$43,$5E,$8F,$F3,$16,$DF,$F8,$92,$FB,$47,$F9,$F9,$AF
	.dc.b	$F2,$83,$67,$37,$BC,$7E,$2F,$DB,$F3,$D7,$EB,$FD,$DF,$D8,$79,$36
	.dc.b	$BF,$14,$DC,$3E,$9B,$72,$FC,$43,$F3,$FC,$EF,$DC,$7A,$F6,$FF,$24
	.dc.b	$DB,$FA,$F3,$FB,$DC,$B1,$87,$F6,$1C,$FF,$BD,$BF,$93,$F6,$7E,$5F
	.dc.b	$A7,$FC,$0D,$B9,$7C,$18,$F8,$95,$FE,$02,$C7,$E1,$46,$0D,$BF,$8A
	.dc.b	$F2,$EE,$1F,$EE,$9F,$73,$FD,$13,$E5,$FF,$0F,$F9,$FF,$F0,$9C,$63
	.dc.b	$2F,$D5,$46,$3C,$75,$53,$F8,$B4,$D3,$F8,$57,$4B,$CE,$FF,$0C,$7C
	.dc.b	$9B,$E7,$7F,$29,$18,$6D,$70,$8D,$FD,$2E,$36,$FC,$17,$CF,$BE,$3B
	.dc.b	$F0,$4D,$BF,$F3,$2B,$67,$FC,$D4,$D7,$FD,$82,$E7,$7F,$F0,$73,$67
	.dc.b	$F0,$D8,$BC,$CF,$F1,$44,$E6,$FF,$B8,$CE,$6B,$F3,$2D,$8F,$E1,$7E
	.dc.b	$2B,$E2,$3F,$C6,$13,$9C,$7F,$C3,$8D,$BE,$F8,$FF,$8E,$EF,$84,$74
	.dc.b	$8F,$F1,$C7,$CF,$7F,$5A,$C7,$FC,$02,$D4,$D2,$77,$70,$3F,$C6,$5F
	.dc.b	$62,$F5,$31,$87,$D4,$F5,$BF,$C1,$3C,$7F,$EE,$78,$FF,$0F,$37,$1F
	.dc.b	$F1,$7B,$5F,$D5,$79,$E8,$D6,$CA,$FF,$0A,$CE,$6F,$EF,$E5,$FC,$45
	.dc.b	$93,$E1,$3F,$A3,$F2,$7F,$7A,$BE,$58,$B0,$E5,$5B,$8F,$E3,$B7,$D2
	.dc.b	$31,$5C,$2B,$C9,$E9,$FB,$09,$3E,$C6,$DF,$F2,$67,$F2,$7F,$4B,$FC
	.dc.b	$95,$FF,$38,$F4,$7F,$A0,$3D,$CF,$CB,$FC,$FE,$3E,$05,$18,$C7,$FC
	.dc.b	$07,$5B,$7F,$D7,$0D,$C7,$F8,$81,$FE,$F4,$AE,$33,$A6,$C6,$A7,$86
	.dc.b	$7F,$09,$B5,$FE,$40,$F4,$9B,$3E,$97,$C9,$97,$97,$B3,$F5,$BF,$41
	.dc.b	$70,$F9,$AD,$CF,$B6,$9F,$B3,$9F,$15,$7F,$96,$F3,$FF,$4F,$B6,$75
	.dc.b	$5F,$E1,$49,$CD,$7D,$8F,$6F,$A6,$BE,$7B,$DB,$70,$FD,$FC,$EA,$BF
	.dc.b	$CB,$37,$FE,$AC,$3F,$81,$FC,$77,$CD,$6C,$7C,$CC,$11,$E3,$24,$7C
	.dc.b	$96,$E3,$E5,$9D,$CF,$E3,$3E,$83,$A3,$F5,$BD,$4F,$F1,$03,$F5,$9B
	.dc.b	$5F,$45,$AF,$F4,$7B,$FF,$13,$FB,$7F,$16,$7F,$C0,$73,$BF,$63,$04
	.dc.b	$BF,$F7,$AC,$7F,$9D,$BE,$1B,$DF,$CB,$FE,$FB,$6C,$7E,$11,$AF,$F5
	.dc.b	$56,$FC,$5B,$E0,$FF,$D0,$F3,$F0,$AB,$FA,$D4,$E7,$87,$75,$AD,$15
	.dc.b	$A8,$C3,$F4,$33,$FF,$1D,$61,$F0,$7D,$CF,$CA,$B9,$6B,$F4,$23,$6F
	.dc.b	$C5,$F7,$33,$87,$F9,$29,$B2,$69,$8F,$6F,$FB,$86,$5F,$9B,$94,$63
	.dc.b	$5F,$23,$70,$74,$FE,$7B,$E7,$F1,$97,$F9,$62,$BA,$57,$D9,$6B,$F6
	.dc.b	$5E,$55,$FA,$6F,$9F,$BD,$67,$9E,$5E,$2B,$FC,$27,$AF,$82,$3F,$C5
	.dc.b	$3C,$6A,$FE,$86,$53,$8D,$7F,$8B,$E3,$FB,$4B,$DD,$FE,$26,$F6,$FE
	.dc.b	$3E,$FE,$39,$1F,$E4,$3F,$3F,$E6,$3A,$6D,$F5,$FD,$1F,$39,$27,$F9
	.dc.b	$3F,$D2,$57,$F3,$11,$F4,$88,$FE,$75,$F8,$3F,$41,$23,$F4,$51,$F3
	.dc.b	$FC,$47,$D0,$7F,$A3,$A7,$FD,$00,$D7,$FD,$27,$1A,$F8,$5F,$97,$F4
	.dc.b	$5C,$6B,$55,$FD,$11,$AF,$F4,$26,$BF,$CE,$DB,$5E,$0E,$CE,$44,$9F
	.dc.b	$E8,$7E,$85,$7C,$35,$FE,$78,$7C,$B3,$F7,$27,$5C,$5F,$D0,$6B,$55
	.dc.b	$C2,$B4,$F0,$3F,$DF,$51,$A7,$D1,$1F,$A0,$BF,$24,$7E,$8B,$E8,$1F
	.dc.b	$05,$F9,$58,$46,$1F,$F2,$CB,$5F,$F2,$B7,$D0,$FF,$18,$73,$3F,$C1
	.dc.b	$27,$51,$FD,$11,$AF,$F9,$55,$AB,$D9,$FD,$DF,$C0,$FE,$95,$B1,$FE
	.dc.b	$D2,$6B,$FE,$9A,$FA,$1F,$E1,$0E,$7C,$05,$71,$6C,$66,$3F,$4F,$3C
	.dc.b	$23,$FC,$DD,$F4,$1F,$E3,$77,$FC,$EA,$D7,$3E,$57,$3A,$F8,$93,$99
	.dc.b	$FE,$BE,$D7,$FC,$E0,$D4,$1C,$69,$F9,$2F,$5B,$F1,$84,$9F,$F0,$2F
	.dc.b	$D1,$57,$FD,$6F,$DF,$FE,$41,$7F,$C1,$3E,$D6,$78,$FF,$CB,$5F,$42
	.dc.b	$E8,$F9,$0D,$AF,$90,$D7,$9F,$CB,$6A,$7F,$83,$C4,$36,$FE,$07,$E8
	.dc.b	$5F,$2D,$49,$3F,$EE,$0F,$A1,$7C,$F6,$23,$F9,$6D,$70,$7F,$96,$B5
	.dc.b	$CD,$7F,$48,$8C,$65,$FC,$32,$C7,$F1,$C2,$D7,$FA,$A2,$77,$E0,$7E
	.dc.b	$BF,$E2,$FF,$68,$49,$F2,$7C,$6D,$7F,$DB,$F6,$FA,$1E,$86,$71,$F0
	.dc.b	$7C,$3F,$7E,$69,$7E,$D8,$F1,$18,$79,$C6,$1F,$27,$CB,$7C,$72,$3F
	.dc.b	$80,$AF,$E4,$FF,$40,$CB,$33,$F1,$AB,$E1,$11,$F8,$EF,$F3,$C9,$F4
	.dc.b	$47,$3D,$3F,$C4,$91,$88,$FF,$63,$6A,$AF,$90,$7C,$9D,$A4,$FD,$78
	.dc.b	$CB,$F6,$63,$2F,$C6,$8F,$E3,$DC,$12,$3F,$92,$BF,$4F,$E6,$71,$1F
	.dc.b	$33,$E3,$BE,$B1,$8D,$3E,$D7,$CC,$FD,$9F,$7A,$57,$E1,$E3,$EE,$18
	.dc.b	$7E,$AC,$61,$D1,$F8,$D7,$53,$D2,$3F,$21,$F3,$3F,$85,$F5,$E4,$60
	.dc.b	$87,$E0,$1C,$BF,$82,$8F,$C6,$7C,$1F,$E3,$7C,$09,$3F,$9B,$8C,$BF
	.dc.b	$9D,$8C,$BF,$8A,$8F,$E9,$5F,$42,$F9,$5A,$49,$FA,$C7,$F7,$CE,$2F
	.dc.b	$E2,$2F,$F4,$C9,$E0,$FF,$84,$BC,$14,$62,$3F,$CD,$31,$8A,$FE,$6D
	.dc.b	$39,$9F,$F4,$9B,$BC,$FF,$1B,$CF,$F8,$73,$C5,$3F,$A0,$F9,$AF,$D5
	.dc.b	$11,$FD,$C7,$C7,$FF,$63,$F8,$27,$C1,$FF,$43,$6C,$F9,$5A,$F9,$1A
	.dc.b	$FF,$1C,$3F,$C3,$3D,$BF,$E3,$27,$F8,$27,$91,$FB,$93,$F7,$5F,$15
	.dc.b	$FB,$BF,$73,$C1,$EC,$CA,$FE,$15,$15,$ED,$1E,$32,$3F,$E3,$8F,$69
	.dc.b	$F4,$1F,$DD,$F7,$FF,$09,$B7,$C0,$FF,$44,$36,$BF,$D5,$9A,$7F,$B7
	.dc.b	$5F,$AB,$C6,$8F,$F6,$C4,$61,$AF,$F8,$09,$B7,$DB,$E6,$BE,$47,$AB
	.dc.b	$F1,$C9,$5F,$1D,$AA,$7C,$BF,$09,$FE,$36,$C4,$F2,$9F,$1F,$D0,$FF
	.dc.b	$2A,$6F,$93,$FE,$17,$8C,$47,$3D,$3D,$BE,$B6,$6B,$FB,$D5,$6A,$BF
	.dc.b	$DA,$ED,$7F,$DA,$AD,$7F,$D9,$CF,$4C,$7E,$4F,$C1,$F9,$4E,$93,$E4
	.dc.b	$E8,$F1,$11,$FC,$EF,$07,$D2,$F8,$3C,$47,$E6,$30,$57,$90,$91,$F9
	.dc.b	$F8,$CB,$CB,$57,$A5,$F1,$4E,$79,$7F,$B7,$69,$E2,$7C,$32,$5F,$1F
	.dc.b	$9A,$AF,$86,$3E,$19,$1F,$03,$F4,$7E,$0E,$47,$D0,$8F,$9D,$E1,$C8
	.dc.b	$FA,$6F,$D7,$F0,$31,$1E,$D8,$F9,$E4,$7D,$8F,$DD,$F6,$64,$72,$3F
	.dc.b	$81,$8F,$9D,$78,$9F,$E2,$D6,$C6,$9A,$FD,$F3,$FD,$0B,$B3,$D1,$E1
	.dc.b	$5F,$EB,$0A,$DD,$7F,$BC,$1A,$FE,$D5,$AF,$2F,$9A,$FC,$97,$CD,$C8
	.dc.b	$FA,$51,$F9,$53,$0F,$87,$93,$F5,$F3,$98,$FB,$1F,$37,$F9,$B2,$3F
	.dc.b	$6F,$E1,$6B,$C7,$34,$E3,$8F,$F6,$F5,$FE,$4D,$FA,$07,$CE,$7F,$92
	.dc.b	$7E,$31,$E5,$7F,$0D,$9C,$F8,$BF,$EC,$D8,$C5,$6D,$FF,$16,$FD,$03
	.dc.b	$E8,$7F,$B7,$6B,$B3,$FC,$63,$87,$25,$8F,$67,$38,$73,$6E,$3C,$70
	.dc.b	$E9,$EC,$7F,$B5,$F2,$F0,$CF,$5A,$63,$D7,$1E,$91,$8C,$7A,$63,$CA
	.dc.b	$38,$56,$F0,$E7,$1E,$46,$DF,$96,$B8,$4E,$4F,$29,$4E,$D8,$74,$C3
	.dc.b	$BE,$5F,$0A,$75,$8F,$C1,$7F,$9C,$1C,$0F,$F7,$1B,$97,$C6,$6B,$E3
	.dc.b	$9E,$A7,$F9,$D6,$3F,$DD,$31,$8F,$83,$FF,$74,$B6,$33,$3B,$7B,$D3
	.dc.b	$E7,$9E,$F7,$FB,$C2,$30,$D8,$CC,$F9,$1A,$F2,$73,$AF,$CE,$AE,$33
	.dc.b	$F1,$75,$7D,$0F,$E4,$3C,$8F,$9F,$87,$0E,$0F,$C6,$36,$FF,$0A,$DB
	.dc.b	$E3,$C8,$BE,$39,$CB,$F9,$77,$BC,$7B,$46,$27,$59,$F9,$6B,$58,$73
	.dc.b	$AD,$12,$75,$5F,$8F,$BE,$71,$89,$FC,$07,$3F,$F6,$0C,$39,$72,$FF
	.dc.b	$60,$6E,$5F,$D8,$38,$AD,$FF,$08,$9B,$7C,$6B,$F9,$84,$63,$9B,$F6
	.dc.b	$31,$83,$F9,$A7,$CF,$67,$DB,$D3,$63,$27,$F3,$98,$CB,$4F,$ED,$BF
	.dc.b	$91,$FC,$4C,$7F,$C0,$DE,$82,$31,$1F,$EA,$48,$C7,$07,$C5,$8D,$CE
	.dc.b	$94,$EA,$BA,$D7,$B3,$EE,$1E,$17,$C0,$AD,$5F,$FB,$EB,$0F,$F5,$26
	.dc.b	$1F,$D3,$CD,$38,$8F,$E2,$D1,$8C,$7C,$51,$E5,$9D,$3E,$48,$CB,$E4
	.dc.b	$8C,$BB,$8E,$53,$FE,$42,$1F,$71,$FE,$2E,$7E,$2C,$78,$9F,$AB,$F9
	.dc.b	$2B,$E8,$31,$1F,$DB,$A3,$0F,$CC,$3C,$38,$FC,$B6,$E3,$F9,$77,$FC
	.dc.b	$98,$DC,$7F,$C9,$6D,$E2,$FC,$95,$71,$AF,$EF,$7F,$3A,$F3,$FC,$38
	.dc.b	$7F,$51,$A7,$E4,$AF,$9C,$F3,$FD,$1F,$CF,$8C,$63,$E7,$8C,$4F,$9A
	.dc.b	$73,$7F,$E5,$C6,$BF,$84,$6B,$F8,$37,$79,$79,$9B,$7F,$EE,$AE,$66
	.dc.b	$7F,$E0,$88,$FF,$1D,$CE,$AF,$F0,$EB,$E6,$7E,$6D,$79,$DB,$5C,$1F
	.dc.b	$31,$1F,$F3,$A4,$F1,$6C,$71,$7E,$3F,$9D,$6F,$E3,$53,$D7,$FB,$7F
	.dc.b	$3F,$F1,$5B,$7E,$67,$E2,$11,$EE,$DA,$E6,$FE,$E3,$17,$9D,$CE,$BF
	.dc.b	$09,$F3,$FF,$88,$3F,$41,$97,$FB,$A2,$32,$7F,$97,$9C,$61,$FE,$8C
	.dc.b	$6B,$FB,$46,$BF,$B3,$6A,$D3,$E7,$F9,$7D,$BD,$FF,$23,$6F,$F2,$DF
	.dc.b	$3F,$F9,$A3,$F3,$7C,$36,$3C,$5F,$13,$F9,$8C,$BF,$DA,$D1,$FE,$A1
	.dc.b	$6E,$5E,$59,$FD,$6A,$FA,$47,$C6,$E1,$B7,$FD,$6A,$46,$E7,$47,$7F
	.dc.b	$8A,$FD,$6F,$C2,$FC,$66,$D7,$C5,$74,$F8,$DB,$5F,$1D,$E1,$1F,$A6
	.dc.b	$F1,$2C,$FF,$89,$7D,$D6,$BF,$BA,$F8,$2F,$E5,$F1,$8B,$FD,$D4,$62
	.dc.b	$75,$1F,$E6,$A9,$D5,$7F,$8B,$FB,$BF,$D2,$93,$9C,$B9,$9F,$83,$F1
	.dc.b	$6C,$64,$95,$FE,$37,$E6,$4E,$A3,$A7,$CF,$FF,$0E,$3C,$BF,$3F,$6F
	.dc.b	$F8,$74,$59,$F6,$4D,$ED,$FE,$B1,$CE,$CD,$79,$C3,$F6,$9A,$30,$C6
	.dc.b	$3F,$89,$8F,$D5,$3F,$1F,$B7,$FA,$9F,$E4,$DF,$C7,$23,$17,$FE,$8C
	.dc.b	$8C,$67,$E0,$BE,$2F,$8A,$33,$3E,$0B,$FE,$53,$84,$8F,$99,$F3,$D7
	.dc.b	$1F,$99,$37,$0E,$93,$6B,$FD,$B5,$CA,$E4,$FE,$67,$8F,$F6,$AC,$7F
	.dc.b	$25,$18,$AF,$F3,$DD,$6D,$FC,$94,$E9,$B1,$98,$FC,$B3,$5F,$F1,$8B
	.dc.b	$9C,$3F,$B0,$36,$7F,$8F,$64,$EE,$F1,$4E,$6F,$F6,$9F,$64,$F6,$BF
	.dc.b	$BE,$B7,$0F,$C6,$7C,$FE,$26,$47,$FA,$E2,$75,$5F,$99,$AC,$BF,$E7
	.dc.b	$F8,$7B,$7D,$0F,$CD,$F8,$56,$FF,$30,$DA,$FD,$13,$74,$FE,$F5,$F4
	.dc.b	$7F,$E7,$AA,$F0,$79,$DF,$2E,$31,$7D,$E3,$1E,$AF,$F0,$09,$CC,$FF
	.dc.b	$95,$9A,$FE,$F8,$38,$7F,$80,$BD,$2F,$C3,$3F,$20,$EB,$7F,$07,$D1
	.dc.b	$F8,$E7,$ED,$BE,$5E,$C7,$EC,$DB,$87,$EF,$5B,$7F,$A2,$D3,$FC,$B5
	.dc.b	$5F,$2F,$BF,$F7,$F7,$F4,$9B,$19,$3F,$7D,$19,$7F,$8B,$D7,$F7,$2F
	.dc.b	$67,$F8,$DC,$FF,$9D,$3C,$B4,$62,$3F,$C5,$2D,$7F,$C5,$0D,$7F,$C4
	.dc.b	$AF,$29,$E9,$E1,$7F,$90,$97,$D5,$3E,$2B,$FC,$4D,$5B,$C7,$E6,$7D
	.dc.b	$37,$F6,$C7,$CA,$79,$AB,$83,$FE,$DA,$8D,$63,$FE,$D5,$C3,$AE,$1F
	.dc.b	$A7,$9C,$CF,$FA,$39,$AF,$6F,$0B,$F9,$EC,$62,$3F,$BE,$78,$5F,$D0
	.dc.b	$3C,$2F,$E3,$38,$FE,$B1,$B3,$FC,$FD,$AF,$FA,$69,$CE,$3F,$E4,$96
	.dc.b	$CF,$E1,$FD,$97,$99,$FB,$B9,$CD,$FF,$93,$3F,$27,$FE,$DB,$7F,$D2
	.dc.b	$F8,$11,$89,$FA,$8D,$7F,$CB,$6D,$7F,$CB,$4E,$F1,$FF,$16,$B6,$BF
	.dc.b	$C3,$1E,$BB,$E5,$3F,$CA,$D3,$9B,$F1,$7A,$18,$3F,$CB,$93,$6F,$E9
	.dc.b	$CE,$9B,$8F,$9A,$B7,$18,$BE,$93,$EA,$EA,$FF,$23,$C7,$F9,$AE,$31
	.dc.b	$3E,$F8,$C1,$FE,$9A,$CB,$D3,$5F,$E6,$DF,$84,$F3,$BF,$E8,$CF,$82
	.dc.b	$7E,$51,$FA,$2C,$FF,$AC,$7C,$86,$63,$18,$7F,$48,$F3,$13,$9E,$AF
	.dc.b	$F3,$4E,$1C,$9B,$7C,$F0,$FD,$F4,$E7,$4F,$EF,$CD,$86,$3F,$EA,$B8
	.dc.b	$C3,$FD,$17,$D2,$ED,$5F,$22,$7F,$41,$EC,$FF,$9B,$BE,$97,$D5,$D5
	.dc.b	$FD,$2B,$1F,$F0,$F6,$A9,$CD,$79,$AB,$FA,$B5,$79,$1A,$FF,$84,$2F
	.dc.b	$AE,$1E,$69,$E9,$18,$9F,$3C,$FF,$3C,$F4,$7A,$BE,$5A,$D9,$E3,$E6
	.dc.b	$79,$5F,$E0,$4D,$7D,$D3,$D3,$47,$CF,$7F,$A2,$3C,$EF,$A9,$18,$C7
	.dc.b	$FD,$11,$18,$9F,$9F,$39,$CF,$FE,$00,$E9,$FD,$EE,$72,$4B,$F9,$15
	.dc.b	$BC,$7F,$B8,$7A,$7C,$E7,$3D,$9D,$08,$FD,$6A,$FF,$7E,$B7,$1F,$09
	.dc.b	$1C,$1F,$DA,$1F,$F1,$CF,$C0,$7C,$BF,$94,$FF,$1E,$36,$3F,$A0,$07
	.dc.b	$0F,$F4,$36,$F8,$6F,$98,$FC,$2C,$B5,$5F,$F4,$9A,$D6,$9C,$2F,$75
	.dc.b	$AB,$FF,$25,$D6,$9E,$57,$F7,$FB,$DF,$25,$FE,$00,$B4,$62,$BF,$2E
	.dc.b	$D7,$E7,$B5,$FC,$EA,$F6,$F9,$36,$38,$C6,$0C,$FC,$E7,$F6,$1C,$FF
	.dc.b	$41,$48,$C6,$BE,$56,$DF,$F9,$E7,$93,$FD,$44,$FE,$8E,$B8,$3E,$69
	.dc.b	$EA,$D2,$1F,$E5,$3F,$43,$9F,$37,$F9,$05,$EB,$9F,$C0,$AB,$F8,$F7
	.dc.b	$F9,$DA,$FF,$60,$D7,$F4,$27,$D3,$71,$9F,$F8,$EF,$AB,$8F,$95,$2B
	.dc.b	$8B,$E0,$D3,$F6,$91,$83,$F8,$1D,$F5,$7F,$83,$B5,$FD,$A3,$5F,$04
	.dc.b	$FA,$B9,$1F,$4B,$4F,$04,$F9,$B3,$5B,$E7,$3D,$A4,$9F,$F2,$B7,$55
	.dc.b	$7F,$57,$3B,$7C,$B3,$F8,$58,$C4,$6A,$3C,$5E,$94,$8F,$8F,$1F,$63
	.dc.b	$C6,$FE,$A3,$E1,$5F,$18,$D3,$E9,$F2,$3E,$6B,$18,$7F,$3B,$9F,$F2
	.dc.b	$9D,$F2,$8C,$1F,$D7,$7F,$29,$78,$7D,$DB,$1F,$D3,$1A,$FD,$8E,$7F
	.dc.b	$CE,$FA,$7B,$A9,$3B,$7C,$71,$88,$FB,$F6,$BE,$7C,$CE,$9F,$F5,$BC
	.dc.b	$E7,$0E,$37,$CC,$F3,$3E,$AB,$F1,$C7,$F9,$06,$78,$3F,$E7,$0F,$CA
	.dc.b	$BF,$CA,$EF,$F6,$60,$E1,$FC,$89,$8F,$5B,$F4,$AE,$BE,$06,$CE,$E3
	.dc.b	$F8,$7B,$5F,$F2,$83,$5F,$F6,$7B,$57,$D2,$FF,$BD,$9F,$F4,$1B,$5F
	.dc.b	$F6,$B3,$5F,$E3,$6D,$7F,$CB,$8E,$F3,$F8,$D1,$FE,$DC,$FA,$67,$CF
	.dc.b	$7A,$1B,$1E,$FA,$E1,$AF,$F9,$33,$F2,$07,$CA,$FF,$02,$C5,$18,$BF
	.dc.b	$F0,$5C,$62,$7A,$63,$BA,$D5,$7F,$30,$AD,$3F,$EE,$F6,$B9,$8D,$BF
	.dc.b	$E7,$5E,$17,$E4,$BC,$AB,$FC,$D1,$F1,$BF,$C2,$BD,$B3,$FC,$32,$75
	.dc.b	$9F,$FB,$EA,$D5,$AC,$3F,$D3,$17,$C4,$E1,$9F,$1F,$03,$A7,$B1,$2B
	.dc.b	$FC,$4F,$87,$8A,$71,$1F,$E9,$D9,$B9,$FF,$B7,$33,$FE,$27,$9F,$FA
	.dc.b	$37,$EC,$1D,$FF,$86,$6E,$1F,$8D,$6B,$F8,$CF,$C2,$3F,$85,$B9,$9F
	.dc.b	$EB,$93,$A8,$FE,$B2,$D7,$B4,$68,$F8,$95,$23,$91,$EA,$B7,$D2,$FF
	.dc.b	$32,$F8,$8F,$D5,$D7,$FB,$C3,$AB,$E5,$4E,$59,$7F,$0A,$B6,$B6,$FA
	.dc.b	$B5,$CF,$E3,$A4,$ED,$FA,$9F,$19,$88,$FE,$05,$69,$C6,$DF,$26,$BF
	.dc.b	$57,$F4,$DF,$D6,$6F,$F5,$ED,$9F,$CD,$C2,$31,$9A,$1E,$9B,$EE,$3F
	.dc.b	$BD,$47,$EE,$EB,$A3,$FE,$02,$6B,$FE,$C3,$6B,$CD,$AF,$7F,$3F,$FA
	.dc.b	$07,$AD,$23,$A4,$7F,$43,$F8,$49,$1F,$88,$F3,$51,$F9,$5F,$8A,$CC
	.dc.b	$FF,$23,$9C,$BB,$9D,$3F,$90,$F3,$FF,$C6,$AF,$D9,$F5,$BF,$3C,$7F
	.dc.b	$5B,$C6,$C6,$48,$7F,$2F,$9E,$2F,$7D,$3F,$AC,$11,$FC,$27,$C0,$D7
	.dc.b	$47,$A6,$3F,$59,$19,$DF,$FC,$E5,$17,$7F,$D7,$73,$7F,$A4,$EF,$F1
	.dc.b	$DF,$01,$FB,$32,$3E,$BA,$AF,$F4,$39,$4E,$2F,$F8,$BF,$37,$F9,$62
	.dc.b	$FF,$79,$E0,$FF,$6F,$C2,$1F,$9B,$72,$FC,$F3,$27,$57,$FB,$17,$E5
	.dc.b	$7F,$86,$87,$90,$8C,$37,$8F,$FC,$2B,$F5,$1F,$E2,$FC,$BF,$94,$CE
	.dc.b	$6F,$FD,$79,$7F,$86,$8C,$47,$FB,$BE,$30,$77,$CB,$FC,$77,$EC,$7F
	.dc.b	$6E,$3C,$78,$FE,$7B,$A3,$FC,$B1,$CE,$69,$30,$EF,$EB,$FF,$72,$C7
	.dc.b	$FC,$5F,$1D,$E3,$07,$FB,$77,$D0,$FE,$43,$1F,$B5,$6E,$5F,$E4,$CF
	.dc.b	$13,$FB,$FE,$5E,$EF,$A5,$F8,$BF,$45,$30,$FE,$3F,$C0,$9C,$47,$B6
	.dc.b	$7C,$44,$7F,$DD,$58,$F9,$38,$3E,$CB,$79,$7F,$DA,$FA,$72,$9F,$97
	.dc.b	$AF,$E0,$2F,$FD,$A2,$D7,$FB,$FB,$63,$F0,$79,$7C,$59,$DF,$27,$C3
	.dc.b	$23,$F5,$1A,$FF,$83,$F1,$FE,$66,$FC,$3F,$1B,$F0,$73,$FE,$25,$F5
	.dc.b	$1F,$5D,$49,$5F,$0B,$5A,$FF,$A2,$39,$7F,$C7,$53,$FE,$24,$AE,$33
	.dc.b	$FE,$FF,$6B,$B6,$E7,$F1,$37,$FF,$17,$9D,$1F,$C6,$C3,$E9,$BC,$84
	.dc.b	$AF,$A7,$F3,$F5,$E9,$1E,$12,$3E,$17,$FC,$77,$E2,$C8,$FF,$7E,$D7
	.dc.b	$56,$FD,$1F,$C1,$23,$F8,$0F,$9C,$AF,$A6,$3D,$C4,$7D,$CF,$F9,$23
	.dc.b	$A2,$47,$43,$BF,$C4,$FC,$D2,$3F,$CD,$73,$FF,$09,$F3,$3E,$59,$AE
	.dc.b	$AB,$E4,$5F,$CD,$23,$F3,$5F,$F2,$BE,$72,$3B,$9E,$EF,$67,$EE,$11
	.dc.b	$FE,$F3,$CB,$FE,$1E,$7F,$C9,$73,$FE,$62,$7E,$C4,$FD,$B2,$3F,$9C
	.dc.b	$F1,$FF,$CC,$BC,$92,$39,$C7,$F9,$A0,$C3,$F8,$F8,$C3,$E0,$F1,$7F
	.dc.b	$CD,$A4,$7C,$79,$57,$F2,$D1,$75,$FD,$4C,$FF,$93,$3E,$81,$F2,$3F
	.dc.b	$ED,$1B,$FF,$6C,$7B,$7F,$D7,$D4,$EB,$E0,$3F,$DC,$33,$98,$C4,$F0
	.dc.b	$1F,$40,$FA,$2F,$F1,$AF,$C4,$2F,$86,$3D,$E3,$17,$FD,$B7,$0E,$71
	.dc.b	$88,$F0,$CE,$EB,$FB,$96,$1D,$7C,$3F,$F7,$64,$74,$CB,$CB,$7C,$EF
	.dc.b	$83,$63,$37,$C6,$36,$76,$8B,$8F,$8A,$FA,$4F,$B6,$B8,$3E,$F3,$81
	.dc.b	$67,$83,$DC,$E1,$97,$C2,$AE,$71,$F3,$4E,$25,$9F,$96,$DC,$3E,$A3
	.dc.b	$BF,$88,$FC,$DF,$D1,$71,$8F,$F8,$03,$AA,$31,$09,$DB,$B6,$BD,$F8
	.dc.b	$BF,$C3,$6F,$3C,$3F,$D7,$3E,$57,$26,$C6,$67,$6F,$19,$F8,$F1,$CB
	.dc.b	$7F,$B7,$AB,$E8,$9C,$FF,$E8,$69,$FB,$98,$F4,$6D,$F1,$DB,$F1,$8D
	.dc.b	$CB,$F1,$78,$72,$E9,$EF,$9C,$9F,$A1,$5B,$CF,$C1,$9F,$7C,$79,$DE
	.dc.b	$F2,$EE,$FE,$0B,$2E,$98,$F2,$8D,$57,$08,$FC,$04,$EA,$B8,$C7,$E7
	.dc.b	$FC,$87,$D1,$7F,$B8,$34,$FF,$3D,$B6,$7F,$DA,$8F,$EB,$A3,$3A,$7F
	.dc.b	$AA,$39,$7F,$C2,$73,$B6,$C6,$67,$FD,$1D,$3C,$6B,$9C,$E5,$FF,$33
	.dc.b	$43,$DA,$C6,$1E,$4F,$6E,$CE,$70,$F2,$C6,$2B,$2D,$8C,$9F,$CA,$CF
	.dc.b	$F7,$0F,$D0,$FC,$6C,$BF,$C2,$ED,$8F,$F5,$17,$6B,$C1,$F2,$78,$CA
	.dc.b	$D6,$9F,$E2,$46,$C6,$78,$BC,$F7,$C7,$0F,$05,$EE,$79,$DF,$FA,$CA
	.dc.b	$7E,$69,$8C,$BD,$AD,$38,$E1,$E1,$8F,$8D,$0F,$A2,$7C,$FE,$F1,$FF
	.dc.b	$80,$38,$75,$6D,$F0,$AD,$DF,$4A,$D6,$9F,$0A,$3E,$7F,$AE,$F6,$FC
	.dc.b	$F1,$86,$BC,$4F,$AB,$F3,$0F,$A2,$CE,$7E,$29,$B8,$7F,$95,$AB,$51
	.dc.b	$C4,$F6,$FD,$13,$E5,$7C,$3E,$F7,$D0,$8C,$4F,$FA,$16,$31,$32,$3F
	.dc.b	$D9,$F5,$FE,$03,$79,$C6,$70,$61,$FE,$54,$C3,$9C,$62,$BF,$DC,$17
	.dc.b	$C6,$30,$7F,$98,$27,$75,$C7,$E8,$3E,$2D,$79,$5B,$7E,$A7,$0F,$93
	.dc.b	$E2,$6E,$1D,$38,$98,$AF,$38,$FC,$5B,$B8,$7F,$AD,$6F,$7F,$1D,$F5
	.dc.b	$3E,$83,$E2,$36,$7E,$4B,$87,$E1,$FC,$D5,$CF,$C1,$9A,$B1,$97,$D3
	.dc.b	$AB,$7F,$F0,$07,$85,$F3,$7F,$82,$3F,$43,$A3,$DB,$C5,$7E,$52,$30
	.dc.b	$D8,$FC,$1F,$D1,$70,$6C,$64,$FA,$F1,$97,$FD,$81,$F3,$5F,$90,$3F
	.dc.b	$23,$C5,$63,$B7,$A7,$E2,$3F,$1F,$F2,$16,$3F,$5B,$C0,$FA,$17,$B9
	.dc.b	$FC,$2E,$30,$FE,$5B,$CE,$C4,$EA,$BD,$5E,$A7,$E8,$3B,$FD,$ED,$8C
	.dc.b	$93,$0F,$C2,$F8,$AE,$AF,$E3,$D3,$F5,$F8,$57,$CA,$38,$72,$FE,$87
	.dc.b	$C6,$C6,$3F,$EF,$CE,$1F,$D7,$FC,$D7,$F1,$EF,$8A,$46,$30,$FE,$13
	.dc.b	$A2,$73,$1F,$EE,$7A,$E1,$5B,$BF,$2D,$7E,$F9,$BA,$FE,$DB,$C7,$EC
	.dc.b	$CC,$3A,$D6,$F3,$FD,$B5,$66,$FF,$DE,$1E,$C7,$F0,$43,$CD,$F4,$57
	.dc.b	$0F,$F7,$C4,$59,$F0,$CD,$BF,$F6,$AF,$57,$F9,$66,$F6,$78,$27,$4A
	.dc.b	$9A,$BE,$23,$F8,$94,$F9,$BC,$1F,$E8,$71,$8A,$F7,$C6,$3C,$4F,$94
	.dc.b	$DA,$E0,$7C,$AC,$58,$FF,$1E,$3E,$87,$C6,$59,$D4,$5C,$3E,$5C,$DC
	.dc.b	$D5,$FE,$4B,$E6,$FB,$EF,$5C,$7F,$1B,$C9,$FC,$A9,$B1,$F9,$69,$DE
	.dc.b	$1F,$8A,$D1,$E1,$6A,$3F,$88,$9F,$9B,$C3,$18,$FF,$BF,$AB,$78,$73
	.dc.b	$9F,$EC,$BF,$1E,$FA,$2A,$7E,$2F,$E6,$ED,$7E,$2A,$74,$D9,$D3,$4F
	.dc.b	$C1,$F8,$9D,$3B,$BF,$28,$D8,$C9,$CE,$32,$D3,$F6,$7F,$45,$FE,$65
	.dc.b	$7C,$1E,$6F,$EC,$E3,$1A,$FE,$AB,$1F,$ED,$6A,$33,$AF,$CE,$8F,$C2
	.dc.b	$46,$2F,$E8,$7A,$7C,$C8,$F8,$1F,$DE,$98,$9F,$D1,$FC,$7F,$C6,$F0
	.dc.b	$7F,$5C,$DA,$FD,$0B,$67,$71,$F1,$E7,$F6,$91,$F3,$BB,$EC,$FC,$B6
	.dc.b	$C7,$AD,$B8,$FA,$7A,$59,$FF,$06,$A7,$95,$FE,$5B,$F0,$FF,$C1,$BE
	.dc.b	$C7,$F1,$39,$FE,$A3,$83,$E8,$9F,$3F,$AB,$9A,$7F,$81,$30,$E1,$7F
	.dc.b	$E3,$1B,$DC,$E6,$BF,$91,$7D,$17,$F1,$78,$FF,$35,$B7,$1F,$F3,$83
	.dc.b	$5F,$F3,$D3,$5F,$F5,$43,$BB,$FA,$90,$8C,$4F,$F5,$7F,$B1,$7C,$87
	.dc.b	$83,$C9,$FE,$65,$BF,$F8,$DF,$0F,$CE,$B7,$1E,$7E,$DF,$F5,$0C,$7F
	.dc.b	$9D,$CE,$5F,$F5,$C3,$6B,$F8,$33,$54,$F1,$7A,$F8,$BF,$E3,$8F,$9D
	.dc.b	$7A,$34,$FF,$1A,$FC,$9D,$C7,$FC,$5B,$3A,$6C,$E9,$A4,$BF,$EF,$53
	.dc.b	$B7,$FA,$1F,$B2,$BE,$91,$AB,$B9,$FE,$0E,$B5,$87,$38,$FC,$F7,$23
	.dc.b	$1F,$F2,$59,$FE,$4E,$D1,$6F,$8D,$F6,$76,$4C,$7D,$2F,$F9,$42,$31
	.dc.b	$9F,$F8,$6B,$E1,$1F,$45,$F9,$1C,$5E,$53,$E8,$FF,$CE,$15,$AE,$CF
	.dc.b	$F2,$0C,$62,$FF,$31,$39,$CF,$FA,$9F,$8D,$FE,$24,$7F,$9C,$46,$72
	.dc.b	$F0,$57,$03,$C1,$5B,$84,$EF,$83,$F7,$CF,$E7,$BC,$CF,$F2,$8F,$D1
	.dc.b	$79,$3B,$BF,$9F,$57,$F4,$A8,$C4,$CC,$BC,$7E,$8D,$C7,$E0,$61,$C2
	.dc.b	$7C,$AD,$8F,$E9,$B9,$7F,$6A,$77,$7C,$71,$F1,$D6,$F7,$FE,$A5,$8E
	.dc.b	$71,$F7,$EE,$F9,$7F,$F3,$3F,$9D,$F2,$A3,$13,$FD,$3E,$31,$32,$3E
	.dc.b	$4F,$A7,$74,$FA,$D7,$C1,$B5,$FD,$7A,$75,$9F,$C6,$AF,$CF,$DF,$FA
	.dc.b	$33,$7F,$B7,$D7,$FE,$A6,$FC,$79,$EA,$F8,$E3,$C9,$9C,$BF,$F0,$5F
	.dc.b	$AF,$FE,$72,$8F,$09,$FE,$27,$3F,$B9,$7C,$1F,$DC,$FA,$4F,$F7,$A3
	.dc.b	$63,$FD,$FC,$D7,$FB,$C0,$71,$E3,$BE,$33,$FC,$E3,$3F,$80,$AD,$CE
	.dc.b	$76,$DD,$6B,$55,$6A,$30,$FE,$3B,$B5,$F3,$1E,$0B,$FC,$4C,$63,$1F
	.dc.b	$88,$D7,$FA,$77,$D5,$6F,$5E,$5E,$F5,$9F,$C9,$D7,$F9,$75,$E3,$19
	.dc.b	$E7,$71,$3E,$9F,$86,$DF,$B4,$F9,$9B,$89,$A7,$1E,$CF,$F3,$F6,$5E
	.dc.b	$48,$CB,$4F,$D5,$FB,$58,$7F,$C9,$BC,$EF,$5F,$C4,$D8,$E0,$7E,$F7
	.dc.b	$E9,$7F,$CB,$EF,$E0,$DA,$E7,$CF,$F1,$78,$3E,$89,$CF,$1F,$EC,$1B
	.dc.b	$C5,$FB,$A6,$DF,$E9,$AB,$85,$C8,$EC,$F4,$69,$FE,$2D,$FA,$6F,$39
	.dc.b	$B7,$CD,$F2,$9E,$12,$3F,$C2,$BC,$0A,$E8,$F5,$47,$C9,$8C,$E5,$FE
	.dc.b	$4A,$F6,$7E,$19,$8C,$FF,$D4,$DE,$53,$C8,$FB,$9E,$5F,$F2,$B6,$C6
	.dc.b	$72,$FF,$35,$F2,$3E,$9B,$35,$27,$FA,$1B,$F8,$19,$D4,$7F,$BC,$BE
	.dc.b	$9B,$FD,$C8,$DF,$27,$BC,$62,$3F,$A1,$FB,$38,$99,$1B,$3C,$31,$87
	.dc.b	$FD,$89,$3C,$75,$FE,$86,$DD,$3F,$AF,$E0,$BE,$15,$AC,$2D,$6A,$7F
	.dc.b	$39,$83,$E9,$9E,$E6,$1B,$1F,$EA,$86,$BF,$E5,$07,$F8,$4B,$98,$E0
	.dc.b	$47,$87,$91,$DB,$E9,$7F,$B6,$C7,$FA,$C9,$AF,$F9,$69,$B9,$7F,$82
	.dc.b	$3E,$9B,$FA,$0C,$EF,$D3,$FE,$12,$CA,$E2,$D8,$CB,$FC,$3A,$32,$FF
	.dc.b	$A1,$E9,$F4,$CF,$99,$FD,$E4,$61,$B1,$99,$FF,$63,$E1,$FE,$16,$79
	.dc.b	$6B,$CE,$75,$1F,$E3,$26,$DF,$F9,$4C,$E9,$1C,$BD,$9F,$E0,$DF,$47
	.dc.b	$AB,$FF,$0A,$FB,$CF,$92,$FE,$14,$DB,$FF,$5D,$D6,$DF,$F4,$FB,$7C
	.dc.b	$8F,$F2,$3D,$70,$3A,$E9,$DB,$BB,$9F,$B7,$30,$E7,$19,$7F,$CF,$D1
	.dc.b	$97,$FD,$AF,$87,$EF,$17,$D2,$BB,$BF,$0E,$D8,$FF,$75,$B5,$FF,$49
	.dc.b	$FD,$33,$C1,$DD,$49,$DB,$F9,$09,$D4,$60,$FF,$48,$6D,$F0,$8E,$0F
	.dc.b	$93,$E7,$FF,$97,$7C,$67,$E7,$E3,$FD,$82,$DD,$D9,$AB,$E7,$9C,$3A
	.dc.b	$E5,$FC,$C8,$E7,$F2,$52,$3F,$DB,$AF,$D6,$F8,$FC,$47,$FB,$9A,$2F
	.dc.b	$2F,$F9,$DF,$05,$7E,$AB,$E9,$7F,$A2,$5F,$FA,$D1,$B5,$FE,$62,$6E
	.dc.b	$9F,$ED,$9F,$65,$F4,$DF,$D8,$6F,$F1,$A7,$FB,$76,$B6,$DF,$4F,$B1
	.dc.b	$FD,$4E,$B8,$FA,$F8,$9F,$F6,$C7,$B9,$FD,$08,$C3,$FD,$1B,$E2,$7F
	.dc.b	$A2,$91,$FC,$6F,$C6,$57,$47,$FB,$86,$3F,$2F,$1F,$D5,$67,$F3,$35
	.dc.b	$2B,$FC,$7B,$19,$7F,$2F,$19,$7F,$DB,$11,$FD,$52,$E4,$7F,$BB,$5F
	.dc.b	$D1,$FC,$2E,$23,$FA,$DC,$5D,$7E,$74,$DD,$7F,$CF,$7E,$8F,$F9,$EA
	.dc.b	$F9,$1F,$A6,$9F,$D6,$FC,$24,$8F,$D4,$BF,$DB,$70,$C4,$7E,$7A,$2E
	.dc.b	$1F,$EE,$79,$FD,$AF,$81,$FD,$AA,$A1,$FB,$27,$2F,$D6,$E6,$FF,$32
	.dc.b	$BA,$3F,$53,$1F,$D3,$3E,$9B,$EC,$F9,$E6,$9F,$72,$9F,$4D,$FE,$11
	.dc.b	$C7,$FC,$FB,$3A,$E2,$E4,$7F,$83,$67,$5B,$32,$FF,$62,$D7,$3F,$33
	.dc.b	$30,$C7,$FD,$07,$EE,$F2,$8F,$C6,$8E,$5F,$CC,$B8,$6B,$F9,$9E,$7F
	.dc.b	$07,$87,$EB,$F6,$24,$FE,$5B,$7F,$64,$79,$4F,$EC,$53,$27,$FB,$4B
	.dc.b	$FD,$97,$6F,$57,$D2,$FC,$5F,$A3,$95,$F2,$F8,$13,$88,$70,$FC,$4A
	.dc.b	$C3,$FE,$E7,$AD,$CF,$F7,$6C,$7F,$B9,$11,$F3,$70,$2F,$EA,$72,$49
	.dc.b	$CB,$F3,$75,$C4,$78,$38,$BE,$76,$DF,$66,$B3,$9F,$86,$ED,$7F,$70
	.dc.b	$F1,$BF,$C0,$46,$30,$FF,$7C,$FA,$CF,$A6,$FF,$07,$B6,$3E,$53,$5F
	.dc.b	$01,$F6,$76,$F4,$3A,$9E,$24,$7E,$46,$95,$FF,$66,$EF,$FE,$13,$31
	.dc.b	$9F,$FB,$23,$07,$D3,$78,$89,$5F,$47,$E9,$2B,$E2,$DD,$5C,$23,$CF
	.dc.b	$5F,$93,$AD,$DF,$FB,$5F,$C3,$FF,$11,$18,$DB,$E4,$E0,$FA,$6F,$82
	.dc.b	$4C,$7F,$01,$F4,$15,$F0,$4E,$DF,$8B,$3F,$A1,$9D,$61,$FE,$E5,$E1
	.dc.b	$FF,$18,$18,$C3,$BF,$C7,$7C,$E2,$61,$FC,$B2,$1F,$32,$D3,$33,$E0
	.dc.b	$AB,$82,$FF,$5C,$70,$AC,$BF,$E2,$EC,$FF,$C8,$06,$2F,$EF,$FB,$5F
	.dc.b	$78,$8F,$F9,$2B,$BA,$BE,$E8,$B8,$7F,$BF,$E7,$FC,$95,$E2,$7F,$92
	.dc.b	$F8,$E5,$FF,$94,$A3,$2F,$F9,$52,$32,$FE,$23,$BB,$FE,$00,$CF,$F9
	.dc.b	$4F,$C4,$93,$FE,$5A,$7F,$CA,$DA,$62,$38,$78,$1F,$D4,$5B,$3A,$7F
	.dc.b	$83,$FD,$23,$E5,$7F,$C6,$7F,$17,$FE,$4D,$9F,$E2,$53,$9F,$86,$FF
	.dc.b	$61,$FB,$BF,$EB,$E8,$C4,$FF,$AF,$E3,$F8,$C7,$D2,$3E,$9B,$FC,$53
	.dc.b	$87,$3A,$5F,$08,$C3,$6F,$A6,$3E,$27,$9D,$F2,$73,$8F,$8B,$0F,$2D
	.dc.b	$6D,$B1,$9C,$7C,$5E,$2F,$FB,$53,$2E,$F7,$BB,$F1,$B7,$2F,$14,$2F
	.dc.b	$8B,$DC,$DE,$16,$77,$18,$3E,$11,$C7,$2F,$43,$E4,$3F,$CB,$31,$F2
	.dc.b	$1F,$F2,$FB,$5F,$69,$FE,$5E,$F8,$DF,$CD,$7D,$37,$F9,$82,$11,$87
	.dc.b	$5B,$A7,$53,$99,$EB,$CB,$FD,$AD,$BE,$57,$3C,$3F,$D4,$CD,$9D,$4F
	.dc.b	$DD,$7F,$B9,$6B,$F5,$9B,$BF,$DE,$FA,$67,$57,$EA,$EF,$E0,$9F,$7A
	.dc.b	$B6,$FE,$29,$AE,$70,$FA,$D5,$CB,$AB,$E6,$56,$B2,$F2,$DF,$46,$E5
	.dc.b	$F9,$5E,$06,$1C,$6B,$57,$C4,$98,$F6,$CB,$CD,$87,$AE,$3B,$92,$BE
	.dc.b	$CE,$1D,$2B,$79,$FC,$3A,$FC,$87,$33,$95,$C5,$FE,$83,$DB,$FD,$6A
	.dc.b	$DC,$BF,$BB,$4B,$E3,$F1,$7C,$6A,$CF,$2B,$FF,$69,$C6,$2B,$2F,$0F
	.dc.b	$96,$7C,$85,$F1,$B0,$DB,$FF,$0C,$4E,$9B,$3A,$69,$FE,$06,$FC,$73
	.dc.b	$F9,$C3,$FE,$E2,$F0,$BD,$1A,$27,$2F,$F0,$A8,$CC,$62,$3F,$D6,$ED
	.dc.b	$7F,$C3,$07,$1F,$0F,$F0,$8E,$9A,$FF,$70,$FC,$5D,$E5,$7C,$7C,$1F
	.dc.b	$4E,$FA,$2F,$95,$E5,$2F,$75,$FC,$FE,$B4,$FF,$85,$E3,$38,$7F,$6E
	.dc.b	$8E,$77,$CA,$73,$E2,$FE,$5E,$72,$71,$C7,$47,$CF,$FA,$CB,$1D,$7C
	.dc.b	$AF,$A5,$F4,$FC,$DB,$19,$3E,$8C,$65,$A6,$72,$FE,$E1,$8F,$FA,$66
	.dc.b	$3E,$BF,$BB,$B9,$CF,$8F,$F7,$FC,$DF,$43,$6B,$DC,$D9,$DC,$7E,$02
	.dc.b	$76,$E8,$FC,$1F,$D5,$E2,$7C,$3E,$8F,$E2,$FE,$2B,$E2,$36,$BE,$53
	.dc.b	$67,$71,$F8,$C9,$DB,$A3,$F1,$FF,$2D,$89,$F6,$F5,$7E,$5F,$E2,$FE
	.dc.b	$83,$6B,$EB,$36,$77,$1F,$98,$9D,$BA,$3F,$37,$CD,$89,$FE,$C5,$D4
	.dc.b	$FA,$6F,$D2,$1F,$86,$FB,$67,$16,$30,$FF,$2D,$46,$1F,$D2,$CE,$E7
	.dc.b	$57,$D3,$E9,$31,$5F,$AA,$6C,$7F,$89,$F5,$FF,$61,$CF,$F9,$95,$B1
	.dc.b	$E4,$F8,$FF,$CB,$C6,$3A,$3F,$43,$F4,$DF,$90,$BF,$F3,$E7,$86,$D5
	.dc.b	$87,$EC,$2B,$58,$FF,$9F,$67,$37,$FB,$CF,$43,$F8,$11,$FC,$0F,$C7
	.dc.b	$B1,$FE,$80,$BF,$03,$DA,$9E,$57,$5F,$5F,$F8,$53,$67,$F8,$3B,$71
	.dc.b	$F9,$AD,$7F,$76,$2F,$DE,$E1,$AF,$F0,$06,$BE,$9F,$8C,$FE,$3F,$D6
	.dc.b	$F9,$EF,$E3,$51,$86,$E1,$FE,$AF,$9D,$3F,$C6,$A7,$4E,$7E,$7B,$F9
	.dc.b	$17,$C7,$74,$6C,$64,$FE,$43,$19,$69,$FC,$90,$F1,$FD,$3F,$F2,$B6
	.dc.b	$99,$3C,$4E,$5A,$67,$1F,$F1,$96,$09,$D4,$61,$FC,$95,$7F,$3E,$F9
	.dc.b	$D7,$C6,$63,$0F,$F6,$64,$61,$FC,$AC,$FF,$24,$B7,$90,$EA,$4E,$63
	.dc.b	$0D,$8C,$D7,$D6,$AF,$F6,$25,$FE,$36,$3E,$67,$C7,$78,$F8,$BE,$9B
	.dc.b	$63,$27,$92,$32,$FE,$2F,$C0,$F5,$7D,$3F,$E5,$5B,$19,$3D,$71,$96
	.dc.b	$99,$C3,$C9,$C4,$AF,$74,$7E,$2D,$B1,$99,$F5,$74,$38,$1F,$21,$F8
	.dc.b	$AF,$9D,$FC,$5B,$6B,$F1,$0D,$C7,$EA,$B7,$AF,$FC,$29,$C3,$FE,$84
	.dc.b	$FA,$0F,$C9,$FC,$E7,$F4,$CF,$84,$46,$31,$FF,$70,$46,$1F,$99,$5C
	.dc.b	$5B,$19,$C7,$F4,$B3,$97,$DF,$F0,$7F,$BB,$F6,$3F,$78,$DB,$FD,$B3
	.dc.b	$67,$FD,$86,$7E,$B5,$C6,$5F,$C2,$1F,$8A,$49,$FF,$21,$6E,$8F,$C2
	.dc.b	$62,$E9,$FD,$7F,$C7,$FB,$5B,$7E,$C7,$64,$7D,$6F,$2F,$9F,$FD,$BE
	.dc.b	$5F,$15,$B3,$F1,$1D,$3F,$09,$B1,$F9,$1F,$A5,$FC,$83,$FE,$DE,$CD
	.dc.b	$F4,$CF,$A3,$FF,$06,$D6,$A7,$FC,$90,$6A,$33,$D1,$F8,$69,$E7,$7F
	.dc.b	$C4,$9A,$FF,$83,$9A,$FE,$A3,$E9,$BE,$26,$48,$FF,$6F,$46,$1F,$F7
	.dc.b	$8F,$E0,$6F,$CB,$1E,$8F,$F8,$3A,$B5,$9F,$E2,$1B,$87,$3E,$44,$63
	.dc.b	$16,$3C,$F0,$E5,$C5,$FE,$2A,$F8,$0F,$E0,$D8,$F8,$F2,$FF,$11,$C3
	.dc.b	$EA,$B5,$E2,$6D,$B5,$FE,$7F,$9E,$39,$7E,$F5,$AF,$F3,$46,$BE,$5E
	.dc.b	$57,$D2,$FF,$93,$3E,$5B,$F3,$CD,$BF,$CE,$BF,$88,$23,$F9,$C6,$D3
	.dc.b	$2F,$D4,$DF,$4A,$F5,$1F,$D6,$BD,$07,$87,$DA,$3A,$92,$FF,$DB,$13
	.dc.b	$A8,$C5,$F4,$ED,$FF,$30,$F8,$5F,$B3,$6C,$FE,$C9,$FC,$E9,$1F,$D8
	.dc.b	$BF,$C9,$FD,$47,$E5,$AF,$A5,$FE,$61,$E0,$FF,$9C,$67,$31,$8B,$F4
	.dc.b	$E7,$FE,$78,$BE,$11,$FC,$D2,$F8,$46,$27,$FD,$E1,$E3,$FE,$FF,$3F
	.dc.b	$F4,$F6,$3D,$30,$FE,$60,$DA,$FE,$CE,$D7,$FC,$C6,$D7,$F2,$8F,$A0
	.dc.b	$FE,$8D,$E6,$78,$1B,$3E,$67,$4F,$F4,$88,$D3,$4D,$61,$FE,$89,$AF
	.dc.b	$B1,$1D,$A3,$17,$FE,$B5,$F1,$D8,$FF,$9A,$2F,$67,$C3,$C7,$AB,$87
	.dc.b	$C2,$FC,$2F,$7B,$37,$F0,$BF,$2E,$FE,$6D,$97,$F5,$CF,$0F,$E8,$46
	.dc.b	$2F,$E4,$C6,$3C,$9F,$EC,$8D,$AE,$07,$BE,$B8,$5F,$5A,$FE,$D5,$C9
	.dc.b	$BC,$3F,$93,$FA,$1F,$E4,$1C,$3E,$DE,$AF,$A9,$7A,$BF,$E9,$AA,$D5
	.dc.b	$FB,$AB,$53,$9F,$47,$FD,$67,$C7,$FE,$45,$AE,$B1,$8E,$7F,$F5,$1C
	.dc.b	$7F,$0B,$F8,$9F,$B3,$F0,$FF,$DB,$DB,$1F,$E2,$66,$BF,$EA,$26,$FC
	.dc.b	$63,$E4,$7F,$59,$3F,$EC,$8A,$FF,$53,$CE,$78,$77,$5A,$E5,$FF,$64
	.dc.b	$5F,$F1,$BA,$D3,$C4,$FE,$3D,$B7,$F9,$29,$F8,$7F,$45,$FE,$C6,$6F
	.dc.b	$C5,$7F,$AD,$31,$FF,$59,$D6,$E7,$E5,$FB,$F9,$E3,$FF,$2A,$F1,$BE
	.dc.b	$03,$14,$E4,$F9,$D1,$88,$60,$C3,$7A,$B9,$F1,$5F,$76,$73,$18,$8F
	.dc.b	$D6,$F8,$B8,$9E,$F8,$FE,$5B,$D0,$BD,$87,$57,$CC,$FF,$15,$BA,$32
	.dc.b	$FE,$67,$07,$B9,$E0,$24,$7F,$8C,$BE,$AE,$BB,$8B,$C1,$FA,$7C,$15
	.dc.b	$8E,$0F,$F3,$CC,$7F,$2E,$BF,$F1,$DF,$8B,$95,$B7,$C1,$C7,$D3,$32
	.dc.b	$7F,$3D,$1F,$0A,$FF,$C7,$11,$C3,$2F,$F1,$2D,$FC,$99,$CB,$67,$F8
	.dc.b	$7C,$FF,$BB,$32,$F3,$78,$BF,$18,$E9,$3F,$AB,$BC,$3F,$E5,$08,$CB
	.dc.b	$FC,$03,$CD,$77,$3D,$3F,$1C,$61,$B1,$99,$FF,$10,$C6,$0F,$AF,$F1
	.dc.b	$9F,$CC,$63,$F6,$6D,$7B,$4F,$17,$C1,$F9,$7B,$E7,$7F,$1B,$82,$31
	.dc.b	$3F,$CD,$A3,$13,$F8,$4E,$8F,$F5,$EC,$E6,$B6,$F4,$9E,$71,$8D,$3F
	.dc.b	$90,$57,$0A,$D1,$FE,$3C,$9E,$2F,$F8,$1C,$FF,$38,$B8,$7F,$C0,$0D
	.dc.b	$73,$5F,$03,$D8,$91,$E3,$8F,$D2,$7D,$47,$F8,$7E,$3E,$EB,$5F,$F3
	.dc.b	$B3,$5F,$F5,$6B,$5F,$13,$BC,$1B,$63,$C4,$EA,$76,$F5,$FF,$65,$F4
	.dc.b	$FF,$90,$C3,$FA,$E8,$70,$FF,$93,$9B,$7F,$E3,$0F,$A3,$7B,$DC,$B9
	.dc.b	$65,$F2,$BD,$E5,$FE,$19,$BD,$CF,$6F,$A8,$FD,$D3,$FE,$C6,$6B,$F2
	.dc.b	$DA,$FF,$8C,$DA,$9F,$E0,$58,$FB,$A3,$FC,$71,$7F,$E5,$D7,$C5,$EA
	.dc.b	$FF,$0F,$AD,$E1,$F5,$3E,$3B,$F2,$FC,$26,$92,$FE,$F3,$FC,$33,$27
	.dc.b	$CC,$D3,$F0,$F1,$F8,$8F,$0B,$99,$2B,$F0,$CD,$7F,$A5,$57,$77,$FA
	.dc.b	$8B,$5F,$CE,$C7,$F2,$5F,$A6,$7C,$17,$EF,$A5,$18,$7F,$80,$B5,$FF
	.dc.b	$5D,$35,$FF,$78,$B9,$DB,$E3,$47,$2F,$82,$EC,$77,$A9,$1F,$82,$AF
	.dc.b	$F5,$4B,$67,$2F,$EC,$BE,$37,$E4,$91,$FD,$23,$5F,$7F,$C7,$FC,$93
	.dc.b	$F5,$B1,$F2,$FE,$73,$E6,$65,$2B,$E1,$C6,$5F,$8B,$19,$7E,$4C,$7C
	.dc.b	$5F,$A4,$F8,$98,$9A,$49,$F8,$AF,$F5,$2F,$75,$F4,$BF,$DA,$E7,$FC
	.dc.b	$77,$18,$F4,$3A,$47,$C1,$AE,$15,$B8,$FF,$0F,$D6,$DD,$1F,$EF,$4D
	.dc.b	$3F,$0D,$9C,$8C,$10,$FC,$0B,$97,$FC,$8B,$1F,$C4,$3B,$7F,$26,$48
	.dc.b	$FE,$1C,$D7,$FC,$97,$87,$F5,$C3,$83,$FE,$1A,$F1,$FF,$93,$70,$7F
	.dc.b	$8D,$F8,$5E,$F7,$E6,$4F,$EC,$7E,$24,$BF,$C8,$C6,$5E,$51,$97,$F2
	.dc.b	$F1,$97,$FC,$B7,$1F,$A3,$F7,$3F,$B6,$EB,$27,$F3,$71,$97,$C9,$19
	.dc.b	$7F,$43,$1F,$DD,$1F,$D6,$75,$47,$D4,$64,$FD,$2B,$97,$FC,$D6,$7F
	.dc.b	$82,$7E,$93,$F5,$0F,$F3,$10,$E1,$FE,$68,$D7,$FD,$78,$D7,$7C,$FF
	.dc.b	$E0,$8F,$A1,$FE,$81,$D7,$FE,$C1,$F3,$91,$8A,$E3,$5E,$13,$87,$89
	.dc.b	$23,$F9,$4E,$1F,$BB,$DF,$FC,$D1,$7F,$C2,$27,$FC,$9E,$D7,$FC,$98
	.dc.b	$D8,$DF,$D2,$7B,$7A,$BC,$9F,$49,$2F,$AC,$7F,$60,$FA,$44,$D2,$4F
	.dc.b	$67,$D9,$9F,$F6,$6F,$B3,$FF,$1A,$4F,$F6,$56,$BE,$28,$D7,$AB,$F0
	.dc.b	$FE,$96,$4E,$E2,$BE,$88,$B9,$FF,$31,$F3,$18,$74,$8C,$3F,$E4,$76
	.dc.b	$FC,$A7,$8C,$FA,$1E,$15,$C3,$E2,$EB,$F4,$CC,$3F,$E0,$FB,$E5,$C7
	.dc.b	$F5,$78,$3F,$C9,$33,$C0,$F6,$61,$EF,$F9,$19,$1E,$F7,$EF,$78,$78
	.dc.b	$8F,$A3,$F4,$BF,$81,$31,$A7,$A2,$B1,$3F,$62,$73,$8F,$FB,$EF,$F2
	.dc.b	$5F,$EB,$73,$FE,$23,$6D,$FC,$36,$CF,$9F,$C2,$FF,$0B,$C6,$2B,$FC
	.dc.b	$C9,$9B,$E9,$7F,$24,$6B,$CD,$93,$F8,$88,$CB,$FE,$1D,$AE,$5C,$25
	.dc.b	$F0,$DD,$89,$3F,$90,$F1,$EB,$F9,$0F,$A0,$FD,$01,$8D,$7D,$15,$89
	.dc.b	$FC,$BC,$E7,$5F,$81,$ED,$FA,$89,$3F,$9C,$F0,$EB,$FE,$C4,$9F,$F2
	.dc.b	$2E,$9F,$E8,$26,$C7,$A6,$B7,$8F,$F7,$99,$CC,$F1,$F5,$7E,$23,$B3
	.dc.b	$FC,$2B,$F4,$0F,$87,$FE,$C8,$D8,$FF,$31,$B5,$FE,$08,$1C,$3F,$C1
	.dc.b	$DD,$5F,$FB,$77,$D8,$FF,$04,$4F,$FA,$FD,$B1,$FD,$A2,$77,$87,$E4
	.dc.b	$7C,$BF,$E1,$1E,$34,$C3,$F7,$B1,$FC,$2C,$C3,$FB,$E8,$C3,$FC,$0B
	.dc.b	$E0,$FF,$88,$11,$FE,$0B,$E3,$D7,$F8,$2C,$5C,$7F,$85,$CF,$FA,$A7
	.dc.b	$E9,$3F,$D2,$BE,$B7,$FB,$37,$1F,$F2,$94,$63,$07,$C2,$A7,$3E,$F7
	.dc.b	$F8,$FE,$31,$2F,$A3,$7D,$37,$F1,$59,$CC,$B0,$E9,$18,$6E,$1D,$31
	.dc.b	$F1,$57,$3B,$E5,$39,$8F,$1B,$6F,$96,$3E,$38,$4F,$1F,$5F,$FD,$C5
	.dc.b	$1B,$24,$6B,$1F,$0B,$B3,$C2,$79,$EB,$A3,$B6,$B9,$C7,$CF,$97,$C0
	.dc.b	$C3,$BC,$EE,$7F,$D5,$B7,$F2,$E3,$FA,$73,$5F,$61,$FD,$47,$E1,$FF
	.dc.b	$75,$F4,$DF,$EB,$46,$C7,$E4,$9A,$EF,$9D,$3D,$BE,$71,$E9,$52,$57
	.dc.b	$E6,$B0,$4E,$A3,$0D,$DB,$EA,$37,$6F,$97,$AB,$E9,$9F,$3D,$FE,$67
	.dc.b	$BD,$E5,$F2,$32,$EA,$DC,$79,$65,$F8,$5C,$38,$F3,$66,$A4,$FD,$C3
	.dc.b	$EA,$57,$0C,$BC,$19,$77,$CF,$F2,$B3,$D2,$BF,$18,$DC,$BC,$31,$CE
	.dc.b	$F8,$E3,$DB,$3F,$3C,$F7,$38,$E5,$F6,$B3,$F2,$C6,$70,$E7,$5F,$9E
	.dc.b	$F5,$1D,$3F,$EE,$EC,$7F,$D9,$FC,$3F,$E8,$9E,$6F,$E0,$5C,$F6,$7F
	.dc.b	$C0,$55,$BC,$BF,$82,$5E,$56,$1D,$23,$0F,$6B,$B5,$C7,$0F,$F4,$B4
	.dc.b	$BE,$69,$F1,$7C,$17,$D3,$58,$C3,$73,$FD,$2D,$F2,$AD,$C6,$0F,$F0
	.dc.b	$C3,$D1,$C6,$7E,$68,$DA,$9F,$E3,$9F,$45,$FE,$84,$8F,$F6,$D7,$6A
	.dc.b	$31,$1F,$B4,$8C,$5F,$F2,$E9,$CC,$FF,$2D,$7E,$23,$63,$C3,$97,$BF
	.dc.b	$E1,$33,$7C,$AB,$C5,$5F,$C0,$3C,$77,$D2,$7F,$AC,$E3,$D2,$3E,$9B
	.dc.b	$FA,$EB,$FE,$FD,$8D,$57,$F4,$F9,$CB,$D0,$57,$4F,$5B,$D6,$DB,$FE
	.dc.b	$8A,$D9,$FF,$25,$35,$FF,$7E,$39,$D3,$FC,$3D,$87,$81,$FA,$5F,$4D
	.dc.b	$F8,$39,$D5,$FA,$E9,$EA,$FD,$DF,$92,$D4,$62,$7E,$2F,$02,$73,$5E
	.dc.b	$6A,$D6,$5E,$46,$BE,$46,$BD,$E3,$8D,$F4,$F1,$AC,$FC,$3C,$7E,$6F
	.dc.b	$A1,$F8,$FF,$A6,$F1,$B6,$FE,$63,$E1,$23,$F1,$B1,$46,$2F,$A4,$62
	.dc.b	$7D,$7F,$39,$D1,$FA,$DF,$37,$F9,$DF,$8C,$F8,$2D,$8C,$9D,$5F,$80
	.dc.b	$47,$06,$AF,$EE,$72,$FE,$A7,$DC,$E7,$3F,$83,$EC,$FD,$57,$C9,$FC
	.dc.b	$E6,$C6,$4F,$53,$F3,$48,$E0,$D5,$F6,$E2,$FC,$B7,$27,$E0,$2B,$F6
	.dc.b	$DF,$4D,$2F,$F5,$BC,$2F,$A4,$FB,$51,$F7,$27,$F0,$91,$8C,$7F,$1A
	.dc.b	$D7,$F1,$AD,$7F,$44,$D7,$F7,$1F,$4D,$CB,$9B,$F8,$33,$6F,$C3,$C0
	.dc.b	$F7,$FF,$24,$D8,$FC,$6F,$BD,$89,$FD,$AF,$D2,$FE,$A3,$1F,$F5,$43
	.dc.b	$5F,$EE,$21,$C3,$FD,$D1,$BC,$5F,$C6,$7D,$0F,$E3,$07,$F1,$9F,$32
	.dc.b	$E3,$FC,$42,$F9,$61,$FC,$77,$CB,$F0,$B6,$75,$8F,$9E,$BF,$C6,$ED
	.dc.b	$7C,$B3,$FC,$3A,$31,$1E,$67,$87,$12,$B1,$5F,$CC,$BF,$01,$7C,$07
	.dc.b	$82,$7F,$5F,$F3,$7F,$CD,$BB,$9F,$3F,$FC,$C6,$30,$D9,$FE,$33,$3A
	.dc.b	$78,$98,$8E,$5F,$3D,$FC,$E7,$E0,$EC,$FE,$F1,$BA,$FE,$2D,$BA,$7F
	.dc.b	$39,$7F,$A0,$4F,$8F,$E7,$2B,$E0,$6D,$7E,$83,$14,$62,$FF,$A5,$C6
	.dc.b	$2B,$F8,$59,$CC,$A7,$33,$F2,$FC,$EC,$3F,$AB,$FC,$73,$E9,$FF,$2F
	.dc.b	$18,$6D,$70,$8D,$FD,$4E,$33,$FD,$9D,$FF,$5B,$AF,$8F,$F3,$7D,$3B
	.dc.b	$FE,$73,$63,$27,$F4,$49,$E1,$5E,$D9,$FC,$AF,$85,$E4,$F8,$6B,$7B
	.dc.b	$6D,$7E,$1A,$7F,$08,$49,$FF,$3F,$C7,$E6,$63,$7B,$7E,$16,$3F,$35
	.dc.b	$EF,$E2,$BF,$DF,$DB,$BE,$4F,$F0,$9F,$2D,$F8,$56,$CF,$E0,$9B,$7F
	.dc.b	$4D,$B8,$7F,$B8,$7E,$09,$F4,$1F,$8D,$FA,$2F,$E8,$91,$8C,$3F,$17
	.dc.b	$18,$99,$1F,$8E,$F9,$6B,$8F,$F5,$93,$F6,$5F,$0B,$FB,$8F,$80,$FD
	.dc.b	$CB,$6B,$F5,$2F,$E9,$48,$FC,$2E,$F4,$62,$BA,$46,$27,$F6,$3E,$DF
	.dc.b	$47,$F8,$DE,$6E,$BF,$D3,$7C,$AF,$A9,$B3,$A7,$0F,$E9,$BE,$02,$E3
	.dc.b	$FD,$3E,$DF,$0F,$FA,$F8,$F8,$6D,$AE,$EE,$1F,$F0,$3F,$CB,$3E,$5B
	.dc.b	$15,$F0,$FE,$9F,$F0,$EF,$F1,$60,$E1,$FE,$32,$D7,$FD,$16,$D7,$7D
	.dc.b	$9F,$E1,$68,$FC,$8F,$27,$E1,$7D,$07,$D3,$7C,$5B,$F0,$5F,$F7,$D8
	.dc.b	$C5,$6D,$B9,$7F,$A6,$3E,$99,$F4,$EF,$5F,$87,$2C,$AF,$E8,$F7,$C3
	.dc.b	$26,$1C,$2F,$51,$F8,$1E,$1F,$DE,$3F,$ED,$A6,$BF,$05,$AF,$F3,$86
	.dc.b	$BF,$EA,$A7,$8E,$5F,$A1,$8F,$F4,$6F,$8D,$FE,$36,$FC,$3F,$E7,$65
	.dc.b	$FC,$DE,$F6,$FF,$8C,$F8,$DF,$49,$FE,$28,$6C,$7F,$8A,$1B,$87,$F8
	.dc.b	$A1,$AF,$F8,$A1,$CE,$3F,$A5,$AF,$F3,$2F,$AA,$F1,$3F,$4B,$13,$0F
	.dc.b	$C8,$BA,$8C,$61,$F8,$B8,$B7,$FE,$6A,$BE,$57,$C0,$FC,$CF,$77,$9C
	.dc.b	$8F,$9D,$B7,$F9,$57,$F9,$0F,$D4,$7E,$D9,$FF,$3E,$7B,$0F,$A8,$78
	.dc.b	$BF,$C9,$BC,$2C,$C6,$31,$FE,$35,$18,$9F,$F3,$3D,$71,$6D,$6F,$1F
	.dc.b	$CC,$4E,$6F,$F3,$50,$8C,$3F,$E5,$96,$E7,$FE,$88,$C1,$F0,$1D,$EF
	.dc.b	$FC,$FB,$5E,$73,$F9,$DF,$9D,$E6,$6C,$6B,$1F,$F4,$D6,$3F,$DF,$A7
	.dc.b	$84,$F8,$9B,$1E,$69,$DB,$FE,$F4,$8C,$E3,$FC,$AD,$B5,$FE,$F8,$F5
	.dc.b	$9E,$5F,$F9,$CA,$B5,$87,$FA,$86,$B9,$37,$0F,$80,$DB,$E3,$F2,$9E
	.dc.b	$9A,$F6,$FD,$33,$CB,$FE,$AF,$E2,$7D,$68,$C4,$FF,$4B,$8C,$4F,$D2
	.dc.b	$9C,$E3,$F4,$1A,$FC,$16,$BE,$D7,$79,$7B,$5B,$7E,$DC,$DC,$DD,$A3
	.dc.b	$FC,$3D,$EA,$BD,$CF,$F5,$1B,$63,$FA,$9B,$71,$FB,$ED,$7F,$8D,$2F
	.dc.b	$53,$89,$27,$8B,$FE,$AE,$90,$F8,$BF,$A9,$E5,$7E,$29,$B1,$FC,$95
	.dc.b	$AF,$F8,$F1,$BF,$22,$F9,$6F,$F6,$14,$FE,$46,$BF,$C8,$33,$9D,$B7
	.dc.b	$5A,$E4,$56,$A3,$0F,$F9,$23,$37,$BB,$FD,$CB,$1F,$F2,$A3,$63,$31
	.dc.b	$F1,$23,$1F,$01,$23,$E3,$B5,$FB,$0F,$F9,$53,$1F,$A8,$DD,$7E,$C3
	.dc.b	$67,$FD,$E7,$C1,$FE,$26,$7F,$13,$1F,$E0,$4C,$DC,$28,$C3,$67,$F4
	.dc.b	$4D,$7C,$27,$F8,$5F,$CD,$ED,$CD,$F9,$E7,$C1,$7D,$2E,$DB,$DE,$A4
	.dc.b	$C3,$FC,$41,$F4,$F5,$FD,$F4,$ED,$FC,$FC,$F6,$C9,$5B,$9F,$CC,$4F
	.dc.b	$A0,$8F,$A1,$E1,$E0,$C8,$E6,$7E,$AB,$4F,$29,$1F,$F7,$AF,$D4,$3E
	.dc.b	$2F,$11,$FA,$79,$F8,$A4,$7E,$2B,$E5,$F1,$24,$75,$3F,$C1,$3E,$97
	.dc.b	$D3,$F3,$FF,$4E,$DE,$B3,$ED,$F5,$F0,$FF,$8E,$27,$75,$FD,$36,$31
	.dc.b	$5C,$23,$FD,$EF,$F4,$9F,$E0,$57,$FD,$62,$D7,$FD,$BC,$DE,$5F,$F2
	.dc.b	$07,$2B,$EA,$3F,$DD,$D5,$C3,$D4,$F3,$5E,$F5,$FD,$6E,$68,$C5,$30
	.dc.b	$E5,$E5,$70,$C2,$4F,$33,$F6,$B9,$78,$32,$EF,$1E,$49,$F7,$61,$E7
	.dc.b	$8F,$F5,$7E,$3D,$E7,$27,$EA,$7E,$A3,$EA,$4F,$FA,$AD,$AF,$F9,$15
	.dc.b	$B8,$7F,$7D,$FC,$C1,$F3,$1F,$A3,$A5,$70,$7F,$C5,$0D,$7F,$D7,$ED
	.dc.b	$7F,$86,$B5,$7D,$4F,$F5,$57,$FD,$78,$D7,$FC,$AC,$DB,$FF,$37,$0F
	.dc.b	$65,$EF,$F8,$E7,$5C,$1F,$EC,$E6,$DF,$F1,$C6,$DF,$1A,$F5,$7D,$4F
	.dc.b	$EC,$E3,$FC,$34,$DB,$FF,$6B,$B6,$F9,$77,$70,$EE,$FE,$25,$3F,$BB
	.dc.b	$9F,$F2,$43,$5E,$31,$F5,$3E,$43,$F0,$64,$8F,$F2,$9F,$D3,$55,$8F
	.dc.b	$E3,$34,$7C,$E7,$F0,$46,$CF,$E5,$2F,$F1,$BE,$24,$8F,$CB,$BB,$FA
	.dc.b	$5C,$47,$4D,$7F,$B5,$5A,$FA,$1A,$C7,$FC,$75,$F5,$2E,$CF,$BC,$DC
	.dc.b	$3F,$DE,$0D,$7F,$D1,$4F,$F8,$BD,$CC,$7F,$35,$8D,$47,$F9,$AD,$AF
	.dc.b	$F4,$E0,$E1,$FE,$A2,$F5,$9F,$1F,$C2,$78,$8F,$1D,$7E,$B3,$E4,$A5
	.dc.b	$FE,$D9,$F8,$DF,$4D,$88,$F8,$07,$EB,$72,$7C,$AF,$CB,$35,$8F,$CF
	.dc.b	$F8,$DF,$9D,$BC,$BF,$F4,$3C,$7D,$13,$0F,$4E,$FF,$C9,$4E,$63,$E2
	.dc.b	$FC,$47,$C8,$23,$F2,$3D,$BA,$FF,$8D,$37,$4F,$C7,$24,$7F,$B3,$A3
	.dc.b	$2F,$F0,$0E,$1F,$F3,$87,$D4,$FF,$6D,$8C,$0F,$79,$19,$8F,$AB,$7F
	.dc.b	$CF,$F8,$BF,$15,$C3,$27,$FD,$CD,$19,$7F,$D4,$B1,$97,$FD,$E3,$1F
	.dc.b	$C6,$BE,$26,$47,$F2,$47,$FA,$56,$78,$8F,$F7,$BD,$FF,$4D,$31,$BF
	.dc.b	$F2,$6F,$37,$F9,$74,$FC,$9E,$27,$D1,$7E,$74,$FE,$A9,$E4,$49,$FF
	.dc.b	$5B,$C6,$5F,$F7,$E4,$65,$E5,$19,$7E,$A5,$76,$BF,$D5,$7B,$FF,$D7
	.dc.b	$F7,$95,$FE,$C2,$8F,$D6,$18,$78,$C6,$1F,$9B,$C7,$FE,$50,$7F,$65
	.dc.b	$F0,$D1,$CE,$1F,$EC,$7C,$3D,$D1,$93,$DD,$3F,$DB,$DF,$E5,$1F,$53
	.dc.b	$FB,$17,$EC,$87,$0F,$FB,$D5,$B3,$FD,$3B,$E4,$7F,$98,$CF,$09,$DB
	.dc.b	$FE,$D7,$6B,$FA,$D6,$BF,$E3,$C6,$BD,$FA,$FF,$C1,$FF,$4D,$FD,$4B
	.dc.b	$23,$FD,$52,$74,$7C,$27,$F8,$8B,$FA,$16,$E1,$FD,$89,$BC,$DF,$E7
	.dc.b	$75,$DF,$F0,$3A,$7F,$C7,$19,$1A,$4C,$3F,$D2,$FE,$4B,$EA,$5E,$07
	.dc.b	$F7,$F6,$CF,$E8,$27,$53,$EA,$3D,$7A,$FF,$81,$7E,$A5,$F3,$34,$CD
	.dc.b	$F1,$8C,$BE,$06,$BF,$F0,$06,$D1,$88,$FE,$D6,$D7,$DD,$E0,$FC,$A2
	.dc.b	$3D,$6B,$C5,$EC,$C8,$F0,$3F,$43,$C2,$C4,$77,$BF,$F2,$CB,$67,$FC
	.dc.b	$31,$18,$8F,$0C,$7D,$8D,$24,$74,$3D,$9F,$27,$F0,$48,$FD,$9F,$0E
	.dc.b	$BD,$E3,$A9,$1F,$84,$FE,$13,$EA,$64,$74,$3E,$87,$D4,$FF,$70,$6F
	.dc.b	$97,$FE,$33,$8C,$47,$E1,$7C,$4C,$4F,$F8,$C6,$73,$3F,$E1,$46,$BF
	.dc.b	$E2,$46,$AD,$D1,$D3,$9F,$F1,$DE,$77,$E6,$0E,$B5,$E4,$F8,$5F,$11
	.dc.b	$1F,$CC,$7B,$F5,$FB,$96,$9F,$21,$23,$F2,$31,$97,$F4,$1F,$53,$FA
	.dc.b	$63,$96,$0E,$BF,$47,$BF,$F0,$49,$5F,$A7,$F0,$EB,$E0,$8F,$84,$47
	.dc.b	$BB,$FE,$8D,$9F,$F6,$0E,$77,$D8,$FD,$FD,$7F,$5A,$D7,$E1,$DF,$F8
	.dc.b	$C7,$E8,$3F,$DB,$13,$FC,$29,$AF,$F6,$00,$E3,$2F,$DB,$7C,$B7,$1F
	.dc.b	$87,$FF,$55,$4F,$0A,$FE,$E8,$70,$8C,$98,$7F,$79,$E6,$7F,$08,$F2
	.dc.b	$24,$7F,$05,$8F,$E1,$66,$1F,$E0,$D1,$87,$F8,$57,$C1,$FF,$10,$23
	.dc.b	$FC,$37,$C7,$AF,$F0,$D8,$B9,$7F,$0D,$9F,$F2,$87,$D2,$FF,$A8,$7D
	.dc.b	$6F,$F6,$BF,$C3,$32,$FF,$06,$4E,$71,$FF,$6D,$46,$3E,$0F,$FC,$9D
	.dc.b	$5B,$97,$D3,$BE,$A7,$F8,$EC,$E6,$58,$73,$8C,$37,$0E,$98,$F8,$9F
	.dc.b	$F7,$4E,$1C,$FD,$8F,$F7,$24,$6B,$0F,$19,$FE,$EE,$BE,$4D,$CB,$CB
	.dc.b	$3C,$B2,$F1,$BC,$72,$F0,$CE,$AB,$86,$5E,$17,$D7,$7C,$1C,$9E,$03
	.dc.b	$D7,$3C,$5B,$3C,$72,$F4,$47,$33,$27,$E0,$71,$F9,$8F,$E1,$9A,$FB
	.dc.b	$4F,$C4,$FC,$47,$FA,$1B,$EA,$7F,$11,$08,$C3,$F8,$B6,$BC,$39,$D3
	.dc.b	$C2,$9F,$3D,$FE,$1D,$AF,$13,$5F,$13,$5F,$C0,$B5,$FF,$56,$7D,$8B
	.dc.b	$E9,$FF,$71,$1F,$4B,$D2,$FE,$BF,$A6,$37,$FC,$24,$2F,$87,$67,$CD
	.dc.b	$8D,$B6,$33,$7C,$FD,$26,$1C,$F8,$93,$E2,$CB,$0F,$FA,$FE,$B7,$87
	.dc.b	$5F,$88,$FE,$28,$DF,$0B,$79,$FF,$BA,$7E,$46,$D6,$B0,$F5,$37,$5F
	.dc.b	$4B,$73,$F0,$DF,$1C,$7A,$12,$F8,$E3,$D3,$1E,$AF,$83,$5F,$F7,$C5
	.dc.b	$7F,$88,$97,$D4,$7F,$AC,$A3,$AC,$6A,$FF,$CC,$8E,$A3,$86,$5B,$9F
	.dc.b	$A7,$57,$1F,$9D,$36,$F8,$57,$2F,$89,$C1,$9F,$93,$B8,$F8,$E2,$CF
	.dc.b	$82,$6F,$5F,$FB,$D3,$F3,$14,$77,$7C,$D1,$98,$C6,$39,$6E,$1F,$E4
	.dc.b	$0A,$E2,$D9,$F4,$7A,$6F,$3A,$F9,$F8,$6D,$7F,$92,$27,$4D,$9D,$34
	.dc.b	$D4,$F0,$CB,$FD,$19,$5C,$1B,$5E,$9F,$0B,$D5,$A5,$C3,$CC,$DA,$FF
	.dc.b	$4A,$64,$8C,$69,$F0,$DB,$5F,$84,$9D,$3E,$FB,$EA,$FA,$7C,$DF,$7F
	.dc.b	$9D,$F0,$1B,$19,$3D,$D1,$96,$99,$C7,$D7,$19,$6C,$7C,$DF,$99,$FA
	.dc.b	$9F,$4F,$F5,$5B,$19,$3E,$94,$65,$A6,$67,$DF,$19,$6C,$7D,$1F,$3F
	.dc.b	$EB,$F7,$FD,$86,$C6,$4F,$AD,$19,$69,$99,$FA,$71,$96,$C7,$E1,$7E
	.dc.b	$97,$F1,$87,$E3,$7E,$1E,$C7,$E2,$5B,$3F,$51,$BB,$FE,$55,$E1,$3F
	.dc.b	$8F,$F5,$ED,$7D,$AF,$A9,$FC,$C1,$F9,$0F,$4A,$D7,$E3,$7C,$EC,$34
	.dc.b	$FC,$84,$7E,$5B,$D2,$91,$F9,$BF,$96,$B8,$7F,$B9,$3E,$E7,$F9,$7F
	.dc.b	$89,$FB,$55,$AF,$F1,$49,$D7,$91,$FC,$7A,$3F,$94,$CE,$BC,$6E,$13
	.dc.b	$9C,$7F,$8F,$CE,$5F,$F6,$AB,$D9,$B5,$D4,$E7,$3C,$6D,$5C,$5F,$0E
	.dc.b	$1E,$5A,$E1,$7F,$99,$F6,$3F,$79,$F5,$1F,$BD,$6C,$7E,$E5,$B7,$FA
	.dc.b	$46,$D7,$F4,$2F,$83,$7D,$EF,$E8,$AB,$E0,$DF,$F3,$89,$CF,$85,$FC
	.dc.b	$65,$DF,$75,$BF,$E0,$B7,$C0,$FF,$79,$3C,$F0,$EC,$E9,$B3,$C6,$FF
	.dc.b	$8A,$79,$FF,$C5,$7E,$A6,$E5,$FC,$45,$B8,$FE,$F1,$B8,$7F,$1D,$6C
	.dc.b	$FE,$ED,$B2,$F1,$9E,$0A,$7F,$99,$4B,$E6,$9E,$AF,$F3,$3F,$88,$B8
	.dc.b	$FF,$B5,$26,$D7,$FB,$DE,$B8,$C6,$23,$FC,$8E,$D7,$FA,$F9,$D6,$3F
	.dc.b	$B7,$BA,$6B,$FC,$47,$DE,$FE,$69,$3F,$CD,$5B,$5F,$CC,$5B,$3F,$C7
	.dc.b	$1B,$9F,$E2,$3C,$9F,$C8,$F9,$7F,$92,$6D,$FE,$43,$9F,$F9,$FB,$F9
	.dc.b	$47,$FA,$6D,$71,$8D,$57,$16,$9C,$6B,$9E,$1F,$77,$67,$E5,$9F,$97
	.dc.b	$5E,$57,$E5,$27,$E8,$4E,$63,$13,$C3,$A3,$FA,$4F,$57,$F6,$1A,$D3
	.dc.b	$F5,$3F,$29,$F3,$E1,$FE,$8D,$FC,$B9,$F4,$79,$CF,$FB,$56,$89,$CB
	.dc.b	$FE,$0E,$DB,$E7,$E1,$FD,$CF,$0F,$F3,$DB,$5F,$AA,$D7,$FC,$2A,$DE
	.dc.b	$DF,$F1,$8F,$4F,$EF,$3A,$7F,$7A,$DC,$3E,$63,$6F,$F8,$2B,$70,$FD
	.dc.b	$DB,$67,$E8,$36,$7B,$F8,$7F,$AB,$FC,$AF,$FC,$F1,$5E,$93,$CB,$19
	.dc.b	$F4,$FE,$23,$F8,$8A,$E3,$3F,$E4,$CC,$FF,$DF,$AD,$9D,$4B,$EA,$5F
	.dc.b	$53,$FE,$11,$CD,$F6,$5F,$3A,$B5,$39,$8C,$57,$0C,$3F,$CB,$2D,$AF
	.dc.b	$F5,$13,$5F,$F5,$3B,$5F,$F1,$7F,$E6,$4F,$AE,$FF,$08,$57,$F8,$7E
	.dc.b	$30,$FF,$83,$4C,$46,$9F,$85,$3F,$B9,$9D,$BF,$DF,$23,$96,$11,$DF
	.dc.b	$3A,$9F,$5D,$9B,$FC,$C4,$7A,$73,$FF,$30,$D6,$F0,$F2,$5E,$E7,$3E
	.dc.b	$BB,$D1,$FF,$1A,$7A,$BF,$E3,$56,$E7,$FE,$43,$6E,$7F,$E4,$06,$DF
	.dc.b	$2F,$43,$FC,$79,$E9,$FF,$8F,$5B,$7F,$E4,$B6,$DF,$F9,$21,$B5,$C6
	.dc.b	$7F,$CA,$15,$C5,$A7,$12,$D7,$16,$CF,$AB,$D1,$FF,$26,$71,$7F,$93
	.dc.b	$5B,$3F,$E5,$B6,$D7,$F9,$61,$B5,$F0,$BC,$DF,$F2,$A7,$A3,$FE,$55
	.dc.b	$6D,$FF,$98,$DB,$7F,$E6,$06,$D7,$19,$FF,$34,$57,$16,$9F,$63,$B1
	.dc.b	$F2,$7F,$E6,$35,$F2,$96,$30,$D9,$E0,$DC,$7F,$8A,$57,$09,$E9,$5C
	.dc.b	$1A,$70,$2C,$F0,$AE,$58,$7E,$12,$F9,$4F,$F1,$FF,$C7,$FF,$D1,$D3
	.dc.b	$F8,$85,$EE,$7F,$9E,$5C,$E1,$FE,$8E,$72,$FE,$51,$CD,$7F,$53,$72
	.dc.b	$F5,$F7,$39,$57,$36,$D7,$16,$CF,$F6,$D3,$FA,$6F,$BB,$FE,$AC,$FC
	.dc.b	$CD,$F3,$7F,$D4,$25,$EA,$7F,$AF,$A7,$31,$9A,$E1,$5B,$AD,$3C,$65
	.dc.b	$3A,$7F,$AD,$CE,$9E,$42,$BA,$BF,$E0,$7C,$38,$3E,$63,$FD,$87,$E7
	.dc.b	$F5,$BF,$F6,$1F,$91,$FD,$C2,$79,$B5,$E0,$DF,$83,$7B,$9F,$E3,$49
	.dc.b	$E6,$7F,$B3,$23,$3B,$F4,$C3,$FB,$8C,$62,$76,$7F,$BC,$6B,$95,$F5
	.dc.b	$7A,$65,$EA,$A7,$B5,$FD,$EA,$30,$DD,$7A,$D7,$06,$E5,$F0,$2F,$8E
	.dc.b	$36,$F8,$3E,$7B,$E0,$FE,$0F,$E3,$3F,$C0,$F7,$F8,$DF,$A8,$BB,$7E
	.dc.b	$2D,$B5,$B3,$F1,$95,$8F,$05,$C6,$AC,$36,$B6,$43,$F3,$91,$A3,$75
	.dc.b	$CF,$93,$F1,$58,$FE,$8F,$8B,$06,$6E,$1F,$B9,$72,$FE,$EC,$C9,$FB
	.dc.b	$47,$2F,$ED,$8C,$C7,$57,$91,$C4,$FF,$77,$7B,$5E,$DF,$01,$ED,$D3
	.dc.b	$11,$27,$FD,$E9,$19,$78,$90,$FF,$79,$B9,$7F,$DE,$B1,$D7,$DA,$ED
	.dc.b	$EC,$C9,$FF,$7F,$46,$5F,$F8,$01,$19,$7F,$DF,$B1,$E5,$F6,$3C,$DF
	.dc.b	$09,$27,$FA,$9E,$3F,$E4,$D6,$CE,$5F,$F1,$54,$27,$94,$68,$E2,$EF
	.dc.b	$2F,$EA,$75,$A9,$E3,$39,$E1,$47,$F9,$7E,$FC,$0F,$68,$F9,$5E,$CC
	.dc.b	$8F,$2C,$7C,$D3,$0F,$92,$30,$FF,$6C,$F2,$5F,$57,$F8,$F6,$C6,$63
	.dc.b	$1F,$21,$96,$46,$4E,$51,$C7,$BF,$97,$D5,$C9,$ED,$19,$7C,$11,$97
	.dc.b	$FB,$CC,$7E,$89,$7C,$FB,$C6,$FE,$31,$1F,$0A,$30,$E1,$AF,$F1,$D6
	.dc.b	$BF,$C9,$DC,$EF,$FE,$76,$FA,$07,$C7,$77,$6D,$7F,$85,$0F,$0D,$F8
	.dc.b	$FF,$2D,$67,$E2,$94,$61,$F2,$B5,$FE,$5C,$D7,$F9,$C3,$9C,$3F,$D0
	.dc.b	$FF,$54,$FA,$0F,$80,$DA,$F1,$56,$F8,$7F,$D1,$F3,$FE,$66,$9F,$32
	.dc.b	$EB,$E2,$72,$F7,$64,$EA,$32,$F0,$8C,$BF,$D3,$E7,$93,$B3,$47,$8A
	.dc.b	$3E,$05,$F8,$63,$DB,$18,$23,$E4,$E1,$F5,$9F,$26,$3F,$8D,$7B,$DF
	.dc.b	$98,$CF,$FC,$6A,$FE,$96,$BF,$C7,$AD,$7F,$D2,$4E,$F2,$FD,$0F,$81
	.dc.b	$FE,$94,$E8,$E7,$D1,$2F,$8C,$7C,$AF,$53,$E5,$FA,$92,$7E,$3C,$65
	.dc.b	$F9,$31,$97,$E1,$46,$5F,$89,$19,$CB,$EB,$CF,$E9,$7C,$1F,$D3,$79
	.dc.b	$72,$FF,$55,$19,$7F,$59,$19,$7F,$CA,$B1,$F8,$BF,$A8,$F9,$64,$8F
	.dc.b	$C4,$B5,$FF,$59,$7C,$23,$CB,$FE,$84,$D8,$FC,$2C,$FE,$DF,$E5,$7F
	.dc.b	$71,$C3,$27,$F7,$91,$97,$F7,$D1,$97,$FC,$CF,$19,$7F,$CD,$31,$9C
	.dc.b	$FF,$CD,$3D,$CA,$FE,$AB,$E2,$BF,$34,$7C,$56,$17,$C5,$53,$11,$32
	.dc.b	$FF,$3F,$D7,$19,$D4,$62,$B9,$BF,$C4,$FE,$AF,$F8,$97,$54,$8C,$10
	.dc.b	$FF,$3F,$B9,$7F,$D0,$31,$FC,$27,$F3,$57,$8B,$79,$A5,$7E,$F6,$72
	.dc.b	$D9,$CB,$FE,$31,$8F,$C8,$3F,$CC,$1D,$B9,$8D,$ED,$FF,$00,$27,$72
	.dc.b	$77,$1F,$E9,$2F,$CD,$BF,$D6,$B3,$FE,$9A,$6B,$FE,$EF,$6C,$7D,$0E
	.dc.b	$1F,$F0,$04,$7F,$01,$97,$4A,$3F,$9A,$C7,$1E,$CD,$F8,$1F,$CE,$7D
	.dc.b	$19,$3F,$CF,$5E,$73,$FC,$F7,$E9,$64,$FF,$41,$7F,$A0,$78,$58,$8F
	.dc.b	$F5,$5F,$AD,$FD,$0C,$C6,$9F,$84,$71,$1F,$D2,$5C,$F8,$1F,$B6,$CF
	.dc.b	$FC,$D3,$3E,$FF,$27,$D6,$7B,$3C,$99,$19,$21,$F0,$DC,$BF,$16,$38
	.dc.b	$BF,$26,$3F,$AB,$73,$FF,$57,$F5,$E4,$FF,$5B,$8C,$BF,$D7,$63,$2F
	.dc.b	$FB,$1A,$32,$FF,$B2,$23,$39,$7E,$82,$7E,$8F,$7F,$D2,$F0,$25,$FD
	.dc.b	$98,$CB,$F6,$E3,$2F,$DB,$8C,$BF,$76,$33,$5F,$83,$9F,$F1,$E7,$C5
	.dc.b	$3E,$AB,$FB,$85,$C3,$FD,$BF,$19,$6C,$65,$FF,$6F,$9C,$98,$FF,$7B
	.dc.b	$35,$5F,$EE,$29,$E5,$1F,$8F,$F0,$3F,$1B,$E4,$48,$FC,$A4,$7E,$40
	.dc.b	$C3,$F9,$58,$C3,$A3,$F3,$4E,$AB,$F3,$D3,$EC,$F9,$9F,$6F,$D2,$4B
	.dc.b	$F5,$C6,$5F,$6C,$65,$57,$9E,$7F,$CA,$DE,$3C,$9F,$F6,$53,$FE,$59
	.dc.b	$F8,$3C,$47,$FB,$12,$2D,$7F,$AF,$67,$FC,$BD,$D3,$FE,$5F,$F2,$25
	.dc.b	$FF,$B4,$E3,$2F,$FB,$56,$32,$FF,$B3,$A3,$FC,$E1,$F7,$EF,$93,$FF
	.dc.b	$59,$FC,$4F,$F5,$28,$C5,$7F,$8D,$3D,$F4,$E6,$B7,$3F,$B8,$A7,$CA
	.dc.b	$E3,$C0,$47,$29,$FF,$59,$DF,$17,$A3,$5C,$9C,$EB,$94,$F1,$C3,$C3
	.dc.b	$3D,$F0,$C9,$AA,$F0,$7B,$4E,$AB,$E0,$4B,$EE,$7F,$0E,$9D,$38,$CD
	.dc.b	$18,$8F,$1C,$62,$79,$46,$27,$58,$F9,$21,$F3,$2F,$63,$FA,$D4,$7F
	.dc.b	$B6,$E3,$1E,$87,$F8,$F9,$B1,$FE,$D7,$6B,$FE,$E1,$F7,$FE,$A7,$C9
	.dc.b	$FF,$91,$1B,$1F,$ED,$D6,$BB,$D9,$3D,$7D,$3F,$F7,$4C,$F5,$6B,$E3
	.dc.b	$6B,$FE,$70,$3D,$2E,$1F,$F3,$97,$C9,$3C,$4F,$8D,$F3,$3F,$DD,$DF
	.dc.b	$91,$EE,$27,$55,$F2,$2B,$4F,$CA,$9D,$57,$2D,$37,$B7,$DC,$E1,$56
	.dc.b	$AF,$95,$6B,$97,$FC,$4B,$C2,$F0,$B1,$5F,$6D,$B8,$7D,$AF,$16,$D6
	.dc.b	$F3,$FB,$CD,$BF,$B5,$7A,$AD,$B2,$B6,$E5,$A7,$C0,$AF,$7F,$5F,$C0
	.dc.b	$C3,$D3,$F2,$98,$AF,$F1,$DE,$9F,$8C,$7F,$D5,$71,$9C,$BF,$D3,$D8
	.dc.b	$5B,$E3,$5F,$EB,$08,$C5,$AE,$FC,$8C,$AF,$F5,$AF,$C9,$5C,$7F,$D6
	.dc.b	$73,$72,$F2,$60,$AC,$5F,$F9,$3B,$E5,$70,$FF,$AC,$2B,$D5,$1E,$7C
	.dc.b	$3F,$8B,$CE,$EB,$CD,$39,$AD,$C6,$32,$CB,$6B,$F9,$EF,$CC,$FB,$7C
	.dc.b	$67,$CC,$D8,$C3,$67,$FA,$1B,$72,$F0,$9F,$E0,$4A,$E3,$3D,$0D,$96
	.dc.b	$37,$B2,$77,$18,$CD,$F2,$3E,$AD,$BD,$6D,$C3,$E1,$1F,$15,$C5,$71
	.dc.b	$6E,$BF,$19,$B3,$B3,$8C,$FF,$86,$AD,$7F,$07,$BF,$E1,$7C,$CD,$BF
	.dc.b	$EC,$2D,$BF,$4E,$48,$C6,$5F,$D7,$5B,$5F,$E8,$69,$D3,$FD,$AA,$FA
	.dc.b	$BE,$2C,$D9,$3E,$43,$EC,$FC,$97,$DA,$6D,$6E,$74,$D9,$D3,$4D,$4F
	.dc.b	$0C,$FE,$63,$73,$FE,$E5,$F1,$7F,$7F,$BB,$F0,$0D,$BF,$EF,$ED,$BF
	.dc.b	$7B,$6B,$8C,$FF,$81,$2B,$8B,$4F,$37,$C9,$7E,$48,$FC,$9F,$C2,$D8
	.dc.b	$F5,$37,$0F,$C0,$36,$FF,$3A,$F0,$9F,$81,$F2,$56,$BF,$0D,$F2,$3F
	.dc.b	$9E,$3F,$31,$F1,$F6,$BF,$C2,$FF,$07,$86,$9F,$E1,$A8,$FC,$F7,$C1
	.dc.b	$C8,$F9,$9F,$25,$70,$FC,$DF,$CB,$FF,$36,$F1,$7F,$81,$3F,$E0,$4F
	.dc.b	$0E,$F8,$FF,$C8,$77,$FE,$57,$39,$C7,$F1,$AF,$16,$C7,$F2,$E3,$73
	.dc.b	$C6,$18,$2B,$93,$E2,$9D,$39,$AE,$55,$C2,$BB,$4F,$A2,$72,$FE,$77
	.dc.b	$E3,$FF,$81,$FC,$87,$F0,$46,$DF,$E2,$DB,$3F,$B1,$6D,$7F,$46,$F6
	.dc.b	$9F,$33,$FA,$6B,$FF,$54,$BF,$C5,$2F,$F8,$0C,$EB,$BB,$F9,$43,$C3
	.dc.b	$C2,$B5,$C6,$B6,$F1,$C3,$B3,$96,$CF,$F1,$AF,$3B,$F8,$DF,$C8,$DC
	.dc.b	$BF,$44,$DC,$3F,$84,$36,$FF,$92,$B6,$7F,$95,$36,$7F,$80,$3F,$CF
	.dc.b	$57,$C3,$3E,$53,$FB,$2C,$7F,$6F,$32,$E1,$FE,$E5,$1B,$AD,$47,$DA
	.dc.b	$8C,$38,$6D,$F4,$6C,$FF,$3D,$F3,$FC,$51,$8A,$FE,$29,$F0,$FF,$CE
	.dc.b	$A3,$F9,$DB,$6B,$F8,$13,$67,$F9,$43,$6F,$F1,$BE,$4F,$E5,$BC,$BF
	.dc.b	$CB,$B6,$FF,$2B,$C9,$FD,$15,$FC,$53,$FD,$65,$CC,$6F,$0F,$C9,$36
	.dc.b	$FA,$CF,$2B,$FE,$C9,$F1,$DF,$96,$F1,$7F,$9E,$73,$E3,$A3,$FA,$63
	.dc.b	$6F,$FD,$16,$DC,$BF,$95,$BF,$4F,$E5,$FF,$AE,$47,$D1,$6B,$FE,$21
	.dc.b	$6B,$FA,$86,$A7,$D8,$E7,$3F,$C9,$FC,$C3,$E7,$3B,$47,$EE,$BA,$3F
	.dc.b	$77,$CD,$72,$F8,$AD,$C3,$F6,$4D,$DF,$F7,$2D,$9F,$92,$D9,$FD,$D7
	.dc.b	$07,$E7,$67,$D1,$E0,$FE,$26,$7D,$0D,$7D,$0D,$D0,$D3,$E2,$1F,$E1
	.dc.b	$3F,$9E,$7D,$9B,$C2,$CE,$1F,$D7,$30,$7C,$D3,$DE,$EC,$D9,$FC,$BC
	.dc.b	$EA,$72,$48,$E5,$3B,$9F,$EF,$B8,$78,$3E,$5F,$F5,$13,$9D,$BF,$C5
	.dc.b	$9F,$28,$F8,$1F,$F0,$8E,$3F,$E2,$78,$C3,$FD,$64,$C4,$69,$FE,$C5
	.dc.b	$1A,$C7,$FC,$05,$1C,$72,$8E,$BE,$0A,$CF,$F8,$2A,$BC,$53,$FC,$AB
	.dc.b	$A1,$18,$94,$62,$B8,$E3,$FE,$30,$F3,$FF,$C7,$3E,$A7,$F8,$E9,$B8
	.dc.b	$FF,$75,$F5,$D1,$8C,$7F,$C3,$31,$89,$D5,$7F,$90,$BC,$EF,$F2,$1F
	.dc.b	$A3,$67,$93,$6F,$8B,$71,$FF,$20,$B6,$7F,$CA,$6D,$9F,$4E,$1F,$E0
	.dc.b	$07,$C3,$7D,$AB,$CC,$D8,$F2,$F9,$FF,$E5,$4E,$5F,$F2,$AB,$6B,$D8
	.dc.b	$D9,$DC,$7B,$E7,$FC,$B7,$E6,$7F,$97,$3C,$FB,$3F,$0D,$B5,$F0,$9B
	.dc.b	$9F,$F9,$69,$B3,$FE,$6C,$6C,$FA,$9F,$F3,$57,$63,$E2,$6F,$C5,$E1
	.dc.b	$B7,$FE,$25,$9D,$36,$74,$D3,$53,$C2,$FF,$C9,$B5,$C1,$B5,$C1,$B5
	.dc.b	$C0,$E4,$FF,$95,$3E,$67,$FD,$21,$8F,$FA,$1D,$7B,$9F,$E8,$17,$35
	.dc.b	$F8,$F7,$2F,$F8,$ED,$CD,$7E,$41,$CB,$D7,$DC,$E5,$5C,$DB,$5C,$5B
	.dc.b	$3F,$E6,$C3,$FD,$81,$EF,$7F,$81,$BE,$69,$F2,$1F,$D5,$7D,$2F,$F5
	.dc.b	$DC,$62,$BF,$AC,$56,$EB,$58,$59,$C8,$AE,$0F,$F8,$4A,$F9,$D7,$F5
	.dc.b	$CB,$E1,$39,$BA,$FF,$86,$23,$FC,$33,$E4,$7F,$69,$8E,$AD,$7D,$21
	.dc.b	$C3,$C4,$C7,$9F,$FE,$DD,$F8,$27,$B4,$DF,$AC,$E6,$31,$5C,$0F,$F1
	.dc.b	$4D,$70,$AD,$C6,$27,$77,$E3,$AD,$4F,$C1,$CB,$C1,$EC,$FF,$BB,$75
	.dc.b	$F8,$74,$8C,$4F,$87,$1F,$87,$3C,$AB,$75,$AB,$F1,$B5,$4E,$EF,$F1
	.dc.b	$1F,$17,$FE,$FB,$C3,$F1,$DF,$1B,$76,$FC,$63,$70,$FC,$55,$5F,$01
	.dc.b	$C6,$8C,$2A,$A6,$22,$57,$E5,$63,$2F,$E5,$E3,$A7,$2F,$F8,$7F,$1F
	.dc.b	$F7,$32,$E2,$C9,$AA,$87,$ED,$5C,$BF,$B7,$32,$7E,$C9,$CB,$FB,$43
	.dc.b	$33,$DA,$32,$F8,$23,$DD,$D8,$F6,$2F,$B3,$27,$F9,$A4,$68,$C3,$FC
	.dc.b	$D2,$30,$E8,$FE,$74,$EA,$BF,$DD,$F3,$DB,$D8,$EF,$EC,$4B,$FE,$81
	.dc.b	$19,$7F,$A1,$46,$5F,$E8,$31,$E6,$F5,$FC,$FE,$C4,$9F,$E9,$11,$97
	.dc.b	$FA,$54,$65,$FE,$93,$19,$7F,$A5,$C6,$70,$70,$FF,$5B,$CF,$FC,$AB
	.dc.b	$5D,$0E,$64,$39,$9F,$D5,$E3,$E4,$FB,$12,$30,$43,$FA,$CB,$97,$FA
	.dc.b	$DC,$7E,$37,$C8,$7C,$35,$24,$FF,$76,$F8,$9A,$FF,$71,$8B,$8F,$F8
	.dc.b	$0A,$79,$77,$F3,$F8,$89,$7F,$DF,$A3,$2F,$F8,$02,$32,$FF,$7B,$8F
	.dc.b	$D0,$2F,$9A,$78,$F8,$6C,$7F,$87,$DA,$FF,$27,$F9,$AF,$E1,$CE,$67
	.dc.b	$FC,$3E,$F3,$6C,$F0,$8F,$F3,$F3,$5F,$F5,$98,$F9,$A6,$3E,$46,$C7
	.dc.b	$F8,$BD,$AF,$F3,$7F,$97,$78,$DB,$AF,$0D,$6A,$BF,$D1,$8D,$7B,$B5
	.dc.b	$FF,$61,$D3,$E4,$3E,$1B,$FB,$F8,$FF,$1F,$35,$F0,$46,$8D,$F5,$C8
	.dc.b	$DC,$65,$E3,$19,$78,$9E,$B8,$F7,$11,$EF,$A5,$7F,$D3,$F7,$E5,$7F
	.dc.b	$C9,$6D,$73,$3F,$E4,$EE,$E7,$BF,$F9,$A9,$FE,$79,$C7,$FE,$37,$87
	.dc.b	$37,$F1,$D6,$CF,$FB,$93,$8B,$F4,$6F,$F8,$EA,$39,$46,$2F,$AF,$3F
	.dc.b	$1B,$97,$C7,$0E,$D3,$F2,$BD,$2F,$97,$E9,$4A,$F9,$71,$97,$E6,$C6
	.dc.b	$5F,$B9,$19,$7E,$F4,$67,$1F,$F6,$3E,$FF,$E6,$F9,$FF,$2D,$F8,$1F
	.dc.b	$AB,$F2,$A4,$60,$87,$F9,$65,$CB,$FE,$5B,$8F,$C7,$7C,$AC,$8F,$C7
	.dc.b	$3F,$B5,$E2,$C4,$7D,$28,$FC,$69,$1F,$F6,$9E,$DF,$E6,$2F,$99,$FD
	.dc.b	$E7,$0C,$9F,$F3,$7C,$7E,$FC,$C3,$FE,$6B,$8C,$3A,$3F,$CE,$AE,$B3
	.dc.b	$FF,$3F,$CF,$F5,$5F,$09,$F3,$3F,$E9,$3B,$FA,$07,$F4,$95,$F0,$DB
	.dc.b	$6C,$6A,$78,$FB,$79,$35,$52,$75,$5D,$5F,$F4,$07,$CC,$FF,$17,$E0
	.dc.b	$91,$FE,$8A,$8F,$E3,$46,$1F,$F4,$34,$61,$FF,$53,$7C,$CB,$C2,$FE
	.dc.b	$1E,$FE,$33,$9B,$F8,$0B,$64,$1E,$0F,$29,$E3,$CD,$23,$04,$3F,$C7
	.dc.b	$53,$CD,$E5,$1E,$09,$FE,$F9,$F2,$FF,$89,$9F,$D2,$9F,$E9,$87,$1B
	.dc.b	$FE,$7D,$B8,$7F,$A7,$FC,$37,$1F,$2E,$DE,$7E,$1C,$C7,$E9,$F2,$B1
	.dc.b	$FE,$2F,$97,$FA,$E3,$CB,$FE,$80,$4A,$DC,$FF,$42,$7F,$D6,$BF,$33
	.dc.b	$FD,$14,$8F,$FA,$E3,$E6,$6B,$FE,$B3,$F0,$BF,$A5,$18,$C3,$FD,$8A
	.dc.b	$E2,$3F,$D7,$EE,$76,$FF,$5C,$61,$FE,$6E,$9F,$0F,$8D,$F1,$0F,$8B
	.dc.b	$E1,$C8,$C9,$0F,$8F,$A7,$B6,$35,$18,$A5,$73,$8F,$6C,$FF,$5F,$E9
	.dc.b	$FE,$C1,$EA,$C9,$FF,$69,$C6,$5F,$F6,$AC,$65,$FF,$68,$46,$5F,$F6
	.dc.b	$94,$67,$3F,$F7,$1C,$FD,$CE,$EF,$BB,$DF,$2F,$E6,$C6,$5F,$9F,$19
	.dc.b	$7D,$11,$97,$D5,$19,$AF,$C1,$4F,$F9,$0F,$E3,$DF,$33,$FD,$E6,$E1
	.dc.b	$F8,$08,$CB,$63,$2F,$FB,$A4,$E4,$C7,$FD,$E6,$6A,$BF,$97,$4F,$28
	.dc.b	$F9,$9E,$07,$E6,$3C,$19,$1F,$88,$8F,$CD,$98,$7E,$94,$61,$D1,$F8
	.dc.b	$E7,$55,$F9,$D9,$F6,$7C,$C7,$B7,$E6,$65,$FA,$E3,$2F,$B6,$32,$FE
	.dc.b	$0B,$04,$FF,$97,$3C,$49,$3F,$EC,$B7,$FC,$BF,$F0,$38,$8F,$F6,$3C
	.dc.b	$5C,$7F,$CC,$33,$FE,$64,$EA,$FF,$32,$F8,$D2,$FF,$DA,$91,$97,$FD
	.dc.b	$AD,$19,$7F,$DA,$31,$FE,$77,$F9,$97,$C8,$FF,$AE,$BE,$5F,$F8,$5F
	.dc.b	$42,$73,$7F,$EA,$B9,$CC,$62,$1E,$EA,$73,$F2,$9C,$E5,$7C,$F0,$C3
	.dc.b	$63,$A1,$AC,$3A,$61,$CE,$31,$7C,$27,$38,$78,$B0,$C1,$99,$EB,$EB
	.dc.b	$FE,$B2,$FA,$E2,$8F,$E1,$11,$E8,$BF,$7C,$63,$1F,$05,$6F,$AB,$99
	.dc.b	$D3,$C1,$98,$74,$BE,$CF,$33,$F9,$9F,$CC,$BD,$5F,$F1,$FC,$7F,$8E
	.dc.b	$04,$67,$DD,$C4,$70,$9D,$BF,$E3,$E9,$D3,$AF,$6F,$EB,$FC,$67,$08
	.dc.b	$FF,$5C,$EE,$8C,$4E,$A7,$84,$E5,$EF,$5F,$E1,$CE,$EE,$4D,$8C,$FA
	.dc.b	$5F,$EA,$D6,$CF,$F7,$8C,$DD,$FF,$E2,$0E,$0F,$8F,$18,$F0,$FE,$27
	.dc.b	$85,$9E,$3F,$9B,$77,$1F,$C1,$DE,$DF,$91,$C7,$F5,$67,$4F,$D7,$BB
	.dc.b	$5B,$C7,$EA,$FE,$4B,$F3,$72,$FB,$AB,$DA,$E2,$E7,$2E,$AE,$67,$7C
	.dc.b	$5A,$AD,$B6,$32,$74,$8C,$B4,$FC,$07,$83,$FE,$40,$EC,$FF,$69,$F8
	.dc.b	$76,$7F,$17,$37,$1F,$7C,$DD,$7F,$25,$37,$4F,$F7,$25,$FF,$62,$7F
	.dc.b	$DD,$0D,$AD,$7A,$92,$35,$E1,$DC,$7F,$A6,$DA,$B1,$97,$FB,$9A,$AD
	.dc.b	$FE,$F3,$C2,$7C,$CF,$8D,$7C,$CD,$8C,$36,$7F,$9B,$37,$2F,$E3,$4F
	.dc.b	$AF,$89,$F2,$17,$4C,$36,$FF,$BE,$CE,$9B,$3A,$69,$A9,$E1,$7F,$DA
	.dc.b	$2B,$83,$6B,$D1,$DF,$E9,$F9,$9B,$87,$F4,$B6,$D7,$F2,$F6,$EB,$FD
	.dc.b	$19,$B3,$F0,$DB,$3F,$C5,$B4,$47,$B7,$E2,$FD,$DF,$25,$6F,$D6,$DB
	.dc.b	$F3,$37,$0F,$F4,$06,$68,$C5,$7F,$2C,$8C,$3F,$12,$FA,$BF,$D8,$FE
	.dc.b	$1B,$EA,$77,$7D,$56,$C6,$4F,$ED,$11,$96,$99,$C7,$FA,$74,$65,$B1
	.dc.b	$F4,$BE,$33,$EC,$7C,$6F,$D9,$6C,$64,$FA,$F1,$96,$99,$9F,$A9,$19
	.dc.b	$6C,$7F,$7B,$F8,$7F,$C6,$1F,$8D,$F7,$AC,$7F,$7F,$6C,$FF,$63,$6E
	.dc.b	$BF,$DE,$9E,$13,$FE,$05,$F7,$AD,$7F,$69,$F9,$3F,$CC,$1F,$91,$F8
	.dc.b	$EB,$5F,$8E,$F8,$4C,$34,$FC,$8C,$7E,$5F,$E3,$A4,$7F,$84,$7C,$4B
	.dc.b	$87,$F9,$73,$E6,$3F,$95,$74,$7E,$E5,$FE,$59,$F2,$F7,$CB,$FE,$95
	.dc.b	$87,$FA,$2A,$72,$D8,$CD,$F3,$55,$C6,$F9,$36,$B8,$36,$B8,$3F,$B5
	.dc.b	$F8,$EF,$DB,$7C,$95,$AF,$F1,$43,$6B,$FC,$18,$DE,$1F,$F1,$1B,$FC
	.dc.b	$3F,$E1,$DF,$2F,$FB,$FF,$43,$0D,$C3,$FC,$E3,$3A,$7F,$C6,$93,$A6
	.dc.b	$9A,$CF,$A6,$7D,$5B,$19,$39,$C6,$5A,$67,$2F,$F3,$1D,$FC,$48,$C6
	.dc.b	$5F,$EE,$78,$C4,$FF,$A3,$87,$CB,$7C,$C9,$FF,$77,$D3,$E7,$5E,$2F
	.dc.b	$F2,$FB,$DE,$5F,$CC,$6B,$53,$9C,$3E,$AD,$7F,$73,$6B,$FE,$AA,$38
	.dc.b	$CF,$F6,$D8,$CC,$7F,$BF,$7B,$FE,$67,$87,$F3,$5B,$5F,$29,$B3,$EE
	.dc.b	$6E,$49,$FF,$5B,$F1,$7F,$77,$9F,$C6,$5F,$E3,$5B,$3F,$8B,$6D,$FE
	.dc.b	$2D,$B7,$CE,$17,$CE,$BF,$25,$7C,$DA,$7F,$32,$F2,$FF,$9C,$79,$7F
	.dc.b	$CE,$5B,$19,$3F,$9B,$C6,$5A,$7F,$BA,$2F,$E9,$F3,$7F,$5D,$F3,$DF
	.dc.b	$31,$FE,$41,$AD,$6A,$AD,$4E,$6F,$EE,$4E,$6B,$E7,$57,$AE,$7E,$A3
	.dc.b	$5F,$A9,$3B,$7F,$C5,$59,$FF,$87,$FA,$BF,$69,$5F,$B5,$6D,$FE,$CD
	.dc.b	$B5,$FB,$36,$EB,$F2,$6F,$95,$7E,$E6,$F9,$34,$F8,$1F,$2C,$F6,$7F
	.dc.b	$BE,$5B,$E7,$7F,$29,$5F,$89,$AD,$3F,$EF,$49,$FF,$03,$4E,$B2,$FF
	.dc.b	$7B,$D6,$9F,$F2,$0B,$5F,$F2,$33,$5E,$97,$F0,$DF,$F7,$FF,$E5,$7F
	.dc.b	$DE,$B0,$EC,$21,$F3,$2F,$CB,$DE,$0F,$E9,$6F,$85,$7F,$BE,$AB,$53
	.dc.b	$9F,$3F,$B4,$AB,$6F,$F9,$65,$AF,$E9,$DA,$FF,$9F,$BA,$9E,$37,$F1
	.dc.b	$1F,$2B,$F8,$93,$73,$FE,$1E,$DA,$FE,$06,$DD,$FF,$81,$78,$9F,$C6
	.dc.b	$4F,$E3,$5E,$45,$BF,$E2,$F1,$6F,$F8,$5C,$DA,$FD,$27,$A7,$FE,$7F
	.dc.b	$F8,$FF,$F1,$77,$C7,$FF,$8B,$DB,$8F,$F8,$BD,$B8,$7F,$8A,$9B,$7C
	.dc.b	$BC,$07,$C5,$7F,$8D,$FE,$2F,$FC,$70,$DC,$3F,$C7,$0D,$C3,$FC,$66
	.dc.b	$DB,$E4,$D9,$FF,$47,$B6,$7F,$CC,$6D,$CB,$FD,$2C,$DC,$1F,$17,$FE
	.dc.b	$47,$F8,$BF,$F2,$43,$70,$FF,$24,$37,$0F,$F2,$1B,$71,$F5,$7C,$4F
	.dc.b	$F9,$3F,$E2,$FF,$CA,$0D,$BF,$F2,$83,$6F,$FC,$96,$DA,$E3,$3F,$E5
	.dc.b	$5A,$E2,$D3,$F2,$35,$FE,$BE,$F9,$7F,$F3,$E5,$7D,$25,$EA,$FF,$98
	.dc.b	$5C,$D7,$FB,$39,$CB,$CB,$54,$6A,$FA,$57,$1F,$7F,$93,$6F,$85,$72
	.dc.b	$A6,$1E,$1B,$DD,$69,$B3,$A8,$F1,$46,$27,$F1,$3F,$9B,$FF,$4C,$9F
	.dc.b	$23,$FE,$74,$AE,$07,$FB,$71,$D3,$8C,$51,$BA,$E7,$E5,$7F,$A0,$63
	.dc.b	$17,$FE,$EE,$78,$CF,$83,$DE,$D5,$6D,$B5,$E2,$7F,$AF,$61,$FB,$FF
	.dc.b	$80,$FE,$01,$F3,$3F,$96,$8F,$EB,$5E,$4F,$F4,$58,$C4,$7F,$5D,$9D
	.dc.b	$F2,$7F,$B4,$EB,$4F,$FB,$4A,$33,$3C,$B0,$AF,$11,$7D,$15,$F2,$7F
	.dc.b	$DB,$D5,$FC,$4F,$D0,$7A,$7D,$65,$18,$76,$62,$35,$19,$87,$C2,$3D
	.dc.b	$BF,$BB,$1B,$3F,$90,$46,$76,$F0,$5F,$F7,$B8,$C4,$EC,$E7,$5C,$AF
	.dc.b	$B4,$EB,$2F,$8E,$BD,$AF,$F0,$24,$61,$BA,$F5,$BF,$1C,$F0,$6C,$78
	.dc.b	$67,$78,$79,$9A,$F8,$B2,$F9,$FC,$2F,$CE,$1F,$19,$FE,$13,$9F,$E3
	.dc.b	$70,$F8,$DC,$70,$FF,$3F,$BE,$11,$8C,$11,$CF,$0C,$36,$3F,$5D,$7C
	.dc.b	$DB,$5A,$8F,$0C,$6A,$B1,$32,$30,$7E,$C3,$8B,$AF,$33,$E1,$29,$88
	.dc.b	$93,$FB,$68,$CB,$C4,$87,$EA,$67,$93,$B3,$39,$FF,$0B,$AE,$31,$BA
	.dc.b	$EA,$FF,$B8,$8F,$7F,$6F,$E5,$1F,$F7,$67,$2E,$18,$60,$87,$13,$FD
	.dc.b	$E4,$72,$63,$83,$F8,$B7,$C5,$7C,$97,$F8,$02,$F8,$EF,$9E,$D8,$DE
	.dc.b	$78,$33,$32,$32,$74,$FA,$9E,$3A,$C8,$C1,$0E,$EE,$5F,$0C,$78,$0C
	.dc.b	$69,$FD,$8B,$0F,$B7,$3A,$39,$46,$6B,$7C,$1F,$8B,$A5,$FA,$BE,$13
	.dc.b	$D7,$F5,$33,$1F,$83,$19,$7E,$1C,$65,$F8,$71,$F9,$0F,$29,$EA,$D2
	.dc.b	$4F,$FA,$43,$D4,$AF,$FA,$12,$2E,$9F,$E8,$39,$E5,$F3,$DC,$FE,$7A
	.dc.b	$5F,$FA,$62,$32,$FF,$A6,$A3,$2F,$FA,$3A,$3F,$48,$BE,$33,$DA,$FF
	.dc.b	$9F,$CF,$F4,$03,$87,$06,$F4,$AF,$F8,$B6,$34,$F1,$78,$61,$FE,$59
	.dc.b	$AF,$9E,$DD,$FC,$33,$AD,$FF,$D1,$DF,$35,$F1,$30,$EA,$D7,$C2,$D7
	.dc.b	$CC,$D7,$8E,$9E,$CF,$9B,$7C,$25,$3D,$F5,$CF,$2F,$F3,$64,$B8,$78
	.dc.b	$6B,$31,$FF,$3E,$4E,$5B,$39,$7F,$D6,$51,$E9,$7F,$C9,$CE,$DC,$C6
	.dc.b	$EB,$BD,$79,$63,$E8,$FC,$DB,$E4,$FE,$5D,$A3,$0F,$CE,$6B,$F4,$9A
	.dc.b	$FD,$77,$37,$FE,$87,$F4,$1E,$77,$DE,$6D,$7F,$A5,$EB,$7B,$7F,$27
	.dc.b	$FC,$E5,$F2,$1F,$C9,$1F,$F6,$4D,$7F,$84,$5A,$FF,$7B,$76,$E6,$38
	.dc.b	$79,$52,$30,$7F,$A9,$70,$E3,$97,$97,$BF,$C8,$79,$6A,$56,$E3,$2F
	.dc.b	$18,$CB,$FE,$E1,$A4,$EA,$31,$3F,$AD,$F9,$EF,$D7,$7C,$FC,$9F,$F7
	.dc.b	$DC,$65,$FF,$7E,$C6,$5F,$F7,$8C,$7D,$49,$FA,$7A,$7C,$09,$F9,$D1
	.dc.b	$82,$3E,$DF,$95,$F9,$C7,$17,$FC,$71,$D6,$F8,$6F,$E0,$1F,$0D,$27
	.dc.b	$31,$FC,$0C,$C3,$FE,$27,$94,$E6,$36,$74,$77,$AF,$F9,$F2,$7B,$4F
	.dc.b	$83,$E7,$3F,$87,$75,$48,$C1,$0E,$CE,$5F,$04,$7F,$15,$E2,$FE,$2D
	.dc.b	$F3,$92,$7D,$71,$97,$DB,$19,$7D,$11,$F8,$FF,$83,$FC,$87,$CD,$49
	.dc.b	$FC,$14,$65,$FC,$24,$65,$FB,$91,$97,$EF,$46,$72,$FC,$E4,$FE,$57
	.dc.b	$9A,$57,$E5,$5F,$EF,$7F,$36,$F9,$BF,$CD,$7B,$7F,$9B,$F6,$A5,$FE
	.dc.b	$6E,$37,$94,$8F,$F3,$75,$A3,$71,$FE,$59,$8F,$D3,$BF,$E0,$A6,$F9
	.dc.b	$E6,$9F,$EE,$38,$E7,$D9,$E4,$E4,$91,$83,$BB,$87,$B3,$EA,$8F,$D9
	.dc.b	$79,$7F,$B3,$F6,$E5,$7E,$CE,$32,$FE,$D6,$32,$FE,$D6,$32,$FE,$DE
	.dc.b	$33,$BF,$FB,$0B,$BB,$FC,$05,$3E,$DF,$6F,$E4,$FC,$DC,$8F,$87,$1F
	.dc.b	$30,$C3,$F0,$E3,$0E,$8F,$92,$EA,$BF,$D9,$F9,$FF,$A7,$A7,$FD,$35
	.dc.b	$F7,$5F,$E0,$17,$F8,$51,$FC,$3C,$FF,$19,$B9,$9F,$F6,$BB,$79,$3F
	.dc.b	$8B,$4E,$CE,$7E,$77,$03,$0F,$43,$27,$DC,$F4,$7F,$0B,$EE,$C8,$C1
	.dc.b	$0F,$BC,$E5,$FC,$04,$7E,$30,$C6,$1E,$C9,$C4,$FB,$3C,$BF,$4F,$A3
	.dc.b	$23,$04,$3F,$D4,$59,$7F,$89,$6F,$DB,$3F,$02,$A5,$7C,$68,$E4,$F0
	.dc.b	$8F,$CE,$F8,$1F,$9E,$F1,$24,$FE,$62,$32,$FE,$6A,$32,$FE,$66,$3F
	.dc.b	$C6,$3F,$35,$23,$FC,$6E,$FF,$8D,$3E,$AF,$11,$FE,$38,$8B,$8F,$F8
	.dc.b	$9E,$7F,$C7,$3F,$35,$FE,$3A,$F8,$89,$7F,$E4,$08,$CB,$FE,$42,$8C
	.dc.b	$BF,$E4,$28,$FF,$31,$BF,$EA,$FF,$97,$FF,$6F,$73,$D8,$FF,$73,$37
	.dc.b	$AB,$FD,$BF,$17,$D8,$FF,$54,$D7,$08,$FF,$5F,$35,$7C,$0A,$3F,$DD
	.dc.b	$51,$C4,$E5,$71,$E2,$7F,$BD,$4F,$07,$B0,$F9,$AA,$48,$DE,$29,$7C
	.dc.b	$DE,$67,$C3,$87,$F9,$1E,$F1,$8F,$8E,$7F,$D2,$37,$C1,$F0,$8F,$9D
	.dc.b	$FE,$8D,$F0,$7F,$E2,$C6,$C7,$F8,$C1,$AF,$F8,$D5,$AF,$FA,$EE,$DF
	.dc.b	$35,$FE,$34,$F8,$4F,$53,$63,$33,$FE,$3A,$9C,$BD,$7C,$CF,$E0,$68
	.dc.b	$97,$FE,$45,$DB,$C9,$3A,$9C,$B7,$67,$BE,$EB,$77,$7C,$88,$C3,$7C
	.dc.b	$1C,$E1,$F2,$0F,$F5,$AC,$E9,$C5,$F4,$BD,$F0,$7C,$B8,$C6,$1C,$CE
	.dc.b	$D3,$AE,$1F,$F6,$17,$CE,$FF,$71,$9F,$B2,$FF,$6C,$6F,$A3,$BC,$7F
	.dc.b	$D8,$B3,$97,$C2,$F1,$6F,$A3,$BC,$53,$9C,$BC,$8F,$16,$DE,$E3,$18
	.dc.b	$F9,$23,$17,$CF,$3F,$F0,$DE,$5F,$47,$AB,$E2,$E5,$F3,$3C,$1F,$8F
	.dc.b	$18,$AD,$E8,$8C,$4E,$AF,$FD,$C5,$5A,$9C,$C7,$53,$3D,$DB,$BF,$A3
	.dc.b	$77,$0F,$B7,$EB,$5C,$7F,$C8,$B8,$71,$BF,$AF,$DF,$9A,$93,$9B,$E2
	.dc.b	$DB,$56,$AF,$D1,$3F,$EF,$7F,$80,$77,$78,$57,$81,$63,$0D,$9E,$D3
	.dc.b	$A6,$CE,$9F,$17,$23,$E5,$FD,$8F,$E5,$A3,$31,$8C,$B2,$D8,$C9,$CA
	.dc.b	$32,$D3,$38,$FA,$86,$6B,$ED,$19,$86,$3F,$C2,$2B,$D3,$E0,$7B,$FB
	.dc.b	$FE,$03,$63,$27,$BA,$32,$D3,$37,$D6,$BA,$36,$BD,$FF,$1D,$F0,$3E
	.dc.b	$62,$DF,$C4,$6D,$79,$72,$46,$38,$FF,$23,$18,$AF,$97,$7C,$DA,$73
	.dc.b	$3E,$9F,$C3,$FD,$9E,$FF,$B4,$D8,$C9,$F6,$23,$2D,$33,$8F,$AE,$32
	.dc.b	$D8,$FB,$7F,$1B,$F6,$BE,$3F,$ED,$B6,$32,$7E,$02,$32,$D3,$33,$EF
	.dc.b	$8C,$B6,$3F,$17,$F1,$1F,$91,$3F,$25,$F3,$B6,$3F,$1E,$D9,$FA,$0D
	.dc.b	$D3,$F6,$FB,$A3,$8D,$7E,$5B,$E7,$6D,$FD,$7F,$93,$FC,$F1,$F9,$4F
	.dc.b	$92,$B5,$F9,$8F,$88,$C3,$4F,$CD,$C7,$E6,$BE,$2E,$47,$E8,$BC,$1B
	.dc.b	$87,$E4,$FE,$63,$FA,$1F,$4F,$EF,$5F,$E8,$BF,$39,$7C,$BF,$CA,$F0
	.dc.b	$7F,$32,$9C,$E3,$FB,$D7,$8B,$65,$5C,$6F,$93,$67,$4D,$C3,$C5,$5C
	.dc.b	$27,$F6,$FF,$23,$FB,$DF,$8C,$FD,$F3,$63,$27,$EF,$23,$2D,$3F,$D5
	.dc.b	$7F,$3B,$FC,$17,$CF,$FE,$2D,$F0,$A8,$C4,$E8,$FE,$0B,$3A,$8C,$E3
	.dc.b	$FE,$AD,$CF,$8E,$7C,$9B,$7F,$EB,$86,$CE,$E3,$FD,$77,$3B,$69,$B2
	.dc.b	$C7,$F0,$31,$F3,$7F,$C0,$9F,$ED,$87,$F6,$D7,$0E,$1A,$E6,$BF,$B2
	.dc.b	$BC,$0E,$6F,$F3,$76,$CF,$F3,$3F,$3D,$F3,$8E,$5D,$61,$F7,$AB,$FB
	.dc.b	$E7,$86,$D7,$F7,$98,$FF,$B3,$FC,$4F,$97,$E1,$7C,$C6,$DF,$FB,$69
	.dc.b	$B3,$B8,$FC,$24,$FE,$2E,$7F,$18,$D8,$C9,$F8,$98,$CB,$4C,$CF,$E1
	.dc.b	$A3,$2D,$8F,$E7,$3E,$67,$F3,$8F,$33,$F9,$CB,$63,$27,$F3,$E8,$CB
	.dc.b	$4F,$EA,$D3,$F4,$BA,$3E,$5B,$F3,$DA,$FD,$36,$BB,$8F,$A0,$E5,$E7
	.dc.b	$9A,$7F,$B4,$7C,$CB,$E8,$BF,$80,$56,$A7,$F5,$DD,$9F,$AF,$6C,$7E
	.dc.b	$AD,$B8,$7E,$99,$BD,$5F,$B3,$E9,$ED,$5F,$B1,$BE,$4D,$39,$1E,$EF
	.dc.b	$9C,$7C,$47,$A9,$B1,$EA,$C5,$18,$7D,$AD,$7F,$1C,$E7,$0F,$C8,$B6
	.dc.b	$7F,$C8,$99,$FB,$FA,$BF,$25,$39,$BF,$F2,$47,$CD,$FF,$7A,$BF,$03
	.dc.b	$63,$E5,$B5,$FC,$E3,$5F,$D2,$BB,$C3,$F4,$2D,$9F,$F2,$A7,$E4,$6F
	.dc.b	$CC,$1F,$21,$FA,$3A,$D6,$1F,$DC,$B9,$9E,$3F,$F0,$6F,$2F,$F8,$3B
	.dc.b	$72,$FF,$0B,$4E,$9B,$3A,$69,$AF,$13,$F8,$79,$FC,$43,$C9,$B1,$92
	.dc.b	$61,$FD,$FE,$D3,$A8,$FF,$01,$77,$BE,$77,$FC,$59,$F2,$5F,$E2,$D6
	.dc.b	$E1,$FE,$27,$6D,$70,$6F,$07,$F9,$07,$E6,$BF,$C6,$9F,$3B,$FE,$35
	.dc.b	$6D,$FF,$8B,$DB,$7E,$06,$D7,$19,$FF,$1A,$57,$16,$9C,$4B,$5C,$5B
	.dc.b	$3F,$E5,$1F,$8D,$FF,$22,$7C,$87,$F9,$15,$B3,$FE,$3F,$6D,$7C,$06
	.dc.b	$D7,$F9,$67,$E2,$FF,$C9,$9F,$1F,$FE,$4D,$6D,$FF,$91,$DB,$7F,$21
	.dc.b	$B5,$C6,$7F,$C9,$95,$C5,$A7,$F9,$47,$E1,$3F,$D1,$53,$FE,$6A,$7F
	.dc.b	$C6,$91,$9F,$80,$FF,$2F,$46,$30,$FF,$1B,$3C,$27,$F9,$3B,$9A,$E5
	.dc.b	$5C,$3D,$6E,$2D,$AE,$87,$86,$32,$D3,$37,$DA,$3F,$CD,$DF,$3B,$FE
	.dc.b	$64,$9F,$F3,$43,$FE,$77,$7F,$9B,$B9,$9F,$F2,$53,$97,$FC,$9E,$E6
	.dc.b	$FF,$D0,$EE,$7C,$83,$9A,$E7,$EF,$F8,$3D,$DF,$0B,$63,$C2,$FF,$4D
	.dc.b	$C3,$FD,$85,$F0,$3F,$EC,$3F,$9D,$71,$FF,$53,$F3,$FF,$D7,$31,$8A
	.dc.b	$FE,$AD,$5B,$AD,$61,$67,$22,$B8,$3F,$EC,$FB,$E7,$5F,$CC,$2F,$84
	.dc.b	$E6,$EB,$E4,$1E,$93,$D3,$E1,$18,$6C,$66,$3C,$2D,$5F,$0A,$F7,$3F
	.dc.b	$D5,$51,$C4,$FF,$71,$46,$78,$3F,$DC,$95,$C6,$31,$3B,$39,$D7,$29
	.dc.b	$D3,$D3,$4F,$7D,$3D,$BF,$F0,$1C,$61,$BA,$7F,$81,$A7,$AC,$F0,$AE
	.dc.b	$53,$9A,$DE,$1E,$66,$BE,$2C,$7E,$87,$0B,$E7,$5F,$13,$FE,$12,$9F
	.dc.b	$F6,$6C,$3E,$2B,$1B,$FF,$81,$A3,$18,$7F,$AE,$DE,$2D,$BD,$BF,$EA
	.dc.b	$08,$DB,$FE,$9F,$6D,$F6,$9D,$E1,$E3,$6C,$EA,$FC,$18,$62,$77,$CE
	.dc.b	$E3,$BF,$0B,$33,$FD,$F4,$68,$C3,$FB,$A8,$C3,$A3,$F8,$23,$AC,$FF
	.dc.b	$82,$4E,$A7,$81,$A3,$81,$FC,$E0,$F4,$F8,$0E,$5A,$62,$21,$FE,$DF
	.dc.b	$72,$FF,$B8,$4C,$9F,$D0,$9C,$BD,$4F,$F7,$47,$C4,$FC,$37,$FD,$96
	.dc.b	$BE,$33,$E4,$B6,$37,$A6,$0C,$CC,$8C,$9D,$3E,$67,$8E,$B2,$30,$43
	.dc.b	$FD,$AD,$7F,$79,$B1,$E0,$99,$1E,$68,$E4,$F0,$8C,$BC,$0D,$1D,$73
	.dc.b	$6D,$FE,$52,$BF,$F7,$1C,$FC,$0F,$99,$F6,$9E,$EF,$83,$91,$92,$1E
	.dc.b	$67,$F1,$7E,$3B,$D3,$A4,$9F,$EF,$BE,$AD,$7F,$BA,$45,$CB,$C7,$1F
	.dc.b	$E0,$79,$E7,$F3,$9D,$3E,$76,$4F,$F8,$1A,$32,$FF,$82,$23,$2F,$F7
	.dc.b	$F8,$FD,$1A,$F8,$D6,$D8,$6C,$7F,$95,$1A,$FF,$A1,$E5,$A6,$EB,$FC
	.dc.b	$71,$3C,$DB,$5C,$23,$FD,$08,$D7,$FD,$1B,$F3,$BF,$0D,$FF,$30,$B5
	.dc.b	$FF,$4A,$37,$2F,$F4,$6F,$CE,$FF,$2A,$9D,$E9,$E0,$AE,$11,$FE,$91
	.dc.b	$6B,$DA,$3C,$B3,$E8,$8F,$F3,$7B,$EC,$AE,$B8,$7C,$48,$D4,$FF,$A9
	.dc.b	$5B,$3A,$8E,$33,$BC,$64,$60,$87,$17,$E7,$FC,$E3,$E5,$E9,$27,$FD
	.dc.b	$03,$E7,$BC,$AC,$47,$F9,$F2,$B8,$3F,$E6,$56,$B9,$AF,$F3,$47,$73
	.dc.b	$E5,$DF,$23,$FC,$69,$79,$34,$C4,$4B,$0D,$7F,$1F,$7C,$EB,$84,$EA
	.dc.b	$73,$7D,$A3,$D3,$3F,$8C,$CF,$F9,$14,$7B,$3B,$B1,$32,$30,$43,$FD
	.dc.b	$A5,$3E,$67,$99,$9B,$FF,$66,$C7,$48,$DC,$FE,$B3,$E7,$7F,$5B,$EC
	.dc.b	$49,$FF,$33,$C6,$5F,$F3,$4C,$65,$FF,$30,$47,$D4,$99,$1F,$EB,$2D
	.dc.b	$7C,$4F,$93,$CD,$CB,$F4,$E7,$F3,$F3,$A8,$FF,$70,$35,$FF,$6C,$F4
	.dc.b	$BE,$0F,$F8,$0F,$C1,$C9,$FF,$3D,$C7,$F0,$43,$0F,$F9,$CE,$30,$E8
	.dc.b	$FF,$40,$BA,$D3,$EA,$D7,$59,$FF,$42,$FC,$EF,$F0,$EE,$99,$18,$21
	.dc.b	$FE,$7F,$72,$FF,$A0,$63,$F8,$AF,$7F,$F1,$6F,$9D,$93,$FE,$90,$8C
	.dc.b	$BF,$E9,$28,$CB,$FE,$89,$8F,$C0,$7C,$17,$E0,$7E,$72,$4F,$E7,$23
	.dc.b	$2F,$E7,$A3,$2F,$E0,$23,$2F,$E0,$A3,$39,$7E,$02,$7F,$35,$EB,$CA
	.dc.b	$FC,$95,$FE,$63,$9B,$F4,$13,$FD,$E3,$E6,$9F,$3B,$FA,$1F,$67,$F4
	.dc.b	$5D,$93,$1E,$23,$0F,$D4,$4F,$1D,$24,$EA,$58,$27,$94,$FE,$9F,$2F
	.dc.b	$14,$FE,$D3,$4F,$D4,$B6,$33,$D9,$E5,$D7,$CC,$F9,$FE,$76,$61,$DE
	.dc.b	$F8,$3E,$55,$1A,$8F,$01,$97,$9B,$E7,$EC,$FF,$00,$37,$DA,$FE,$04
	.dc.b	$7F,$03,$F3,$24,$64,$87,$4D,$7F,$85,$46,$A3,$14,$AE,$71,$DE,$7E
	.dc.b	$87,$57,$D1,$F9,$D9,$3E,$B8,$CB,$ED,$8C,$BD,$F2,$4E,$A3,$13,$AD
	.dc.b	$BE,$5C,$FF,$8D,$3E,$77,$FC,$19,$5E,$43,$F8,$F1,$9F,$36,$3F,$0C
	.dc.b	$C9,$F6,$E3,$8D,$7F,$1E,$6F,$07,$E2,$F4,$FE,$47,$5C,$4E,$17,$F8
	.dc.b	$DF,$43,$F1,$DE,$E4,$9F,$B1,$19,$7E,$D4,$65,$F7,$46,$5F,$81,$19
	.dc.b	$C7,$F0,$F3,$ED,$F3,$FD,$DE,$84,$BF,$44,$65,$F5,$46,$5F,$B1,$9F
	.dc.b	$F8,$8E,$FD,$95,$97,$DD,$5C,$6B,$E8,$78,$3F,$A7,$F1,$64,$7E,$46
	.dc.b	$3F,$54,$61,$FA,$D1,$87,$FC,$83,$F3,$5F,$E3,$E2,$3F,$E4,$4F,$9D
	.dc.b	$AF,$F8,$DA,$2E,$1F,$D3,$3D,$25,$7F,$91,$BE,$67,$FC,$8F,$F3,$73
	.dc.b	$0F,$F2,$6C,$65,$FF,$27,$C6,$5F,$F2,$0C,$7F,$9C,$1F,$AF,$EF,$BE
	.dc.b	$5F,$FD,$C7,$D1,$86,$CF,$FB,$19,$FF,$73,$B9,$3F,$81,$7B,$1F,$EC
	.dc.b	$48,$DC,$2B,$A3,$AA,$E1,$D6,$BE,$13,$C8,$CB,$FB,$27,$95,$F4,$BE
	.dc.b	$75,$C6,$79,$E6,$F9,$AA,$48,$F4,$1F,$E6,$EF,$9B,$C3,$D0,$FF,$38
	.dc.b	$C6,$2F,$AD,$2B,$DD,$4F,$9D,$FE,$A9,$F0,$5F,$9E,$8F,$CF,$C6,$1C
	.dc.b	$1B,$F3,$7F,$C6,$0E,$E3,$4F,$1F,$82,$F4,$7C,$8F,$08,$FF,$6B,$E6
	.dc.b	$8C,$4E,$A7,$AB,$5E,$9E,$6F,$C7,$7C,$6D,$7F,$C2,$B7,$FE,$57,$E0
	.dc.b	$EC,$D8,$F5,$64,$FA,$8F,$F6,$37,$81,$FE,$1C,$3F,$77,$19,$EA,$FF
	.dc.b	$35,$C6,$27,$75,$FE,$6E,$3F,$DE,$EE,$23,$53,$FE,$72,$E0,$FC,$1D
	.dc.b	$F3,$9F,$AB,$C2,$F9,$E7,$CD,$6F,$93,$FD,$E7,$97,$6C,$7F,$17,$9F
	.dc.b	$F8,$FF,$CD,$D3,$FE,$AA,$F8,$5A,$E7,$3E,$BC,$DF,$F0,$05,$FC,$1C
	.dc.b	$64,$F4,$15,$FB,$9E,$77,$F6,$27,$E1,$47,$F1,$C8,$C3,$C1,$B8,$79
	.dc.b	$1B,$97,$F6,$DF,$C9,$1F,$01,$E4,$9C,$DB,$E5,$5F,$9D,$74,$F0,$3F
	.dc.b	$D2,$F1,$8D,$BF,$D3,$31,$FE,$4C,$E1,$7C,$2E,$36,$FF,$28,$DE,$DF
	.dc.b	$EE,$AD,$C7,$FA,$FB,$6F,$95,$71,$9F,$16,$3E,$1B,$FF,$02,$60,$F9
	.dc.b	$9D,$43,$F3,$B7,$B5,$C7,$9B,$EC,$56,$A7,$35,$FE,$01,$BE,$4F,$F8
	.dc.b	$20,$E5,$7C,$63,$38,$FF,$74,$9C,$9F,$E0,$E5,$F9,$6B,$E2,$BC,$1E
	.dc.b	$93,$E4,$3D,$55,$A9,$CC,$62,$B8,$5F,$EB,$5B,$5F,$B6,$6B,$FB,$B6
	.dc.b	$BF,$D6,$BF,$01,$76,$7C,$7F,$09,$F4,$AF,$57,$C3,$5A,$C7,$FC,$4B
	.dc.b	$5A,$9C,$DF,$1B,$E1,$5E,$DB,$E1,$87,$F8,$53,$0E,$13,$9B,$E2,$F4
	.dc.b	$E4,$FF,$51,$39,$3E,$37,$E5,$4F,$8C,$F3,$77,$FB,$AB,$F6,$93,$9A
	.dc.b	$F2,$CE,$5F,$EE,$F1,$9C,$7C,$53,$97,$FD,$3B,$19,$AF,$F7,$5E,$2A
	.dc.b	$D5,$2B,$57,$D3,$0F,$91,$7F,$CD,$63,$CF,$1F,$62,$7F,$DF,$91,$E4
	.dc.b	$FA,$0C,$7C,$5F,$A9,$B1,$93,$DF,$19,$69,$30,$F9,$58,$3E,$93,$8C
	.dc.b	$7F,$90,$07,$E3,$0F,$4F,$EE,$7D,$47,$AA,$31,$87,$DA,$8C,$4E,$B9
	.dc.b	$7F,$1F,$5A,$BF,$B1,$5A,$BF,$CC,$4E,$77,$FE,$49,$F9,$E3,$EA,$38
	.dc.b	$E5,$CB,$D7,$F0,$EF,$71,$F0,$37,$3E,$AD,$C7,$E8,$E6,$8C,$57,$F8
	.dc.b	$C3,$3F,$F6,$BB,$FA,$9C,$DF,$45,$AC,$7F,$A6,$7E,$36,$F8,$4F,$CE
	.dc.b	$36,$FD,$3F,$0D,$89,$F7,$B6,$32,$7E,$7A,$78,$57,$CD,$F1,$AE,$BF
	.dc.b	$D0,$F3,$61,$FB,$69,$CE,$1E,$78,$FD,$DF,$A3,$23,$A7,$BF,$71,$FB
	.dc.b	$18,$7F,$BF,$67,$27,$CE,$F2,$7D,$35,$D4,$91,$FE,$83,$AA,$7E,$07
	.dc.b	$E9,$7F,$94,$C7,$F1,$3E,$1F,$8F,$39,$F2,$7F,$8F,$C6,$35,$FE,$B9
	.dc.b	$7C,$2B,$F8,$B4,$E6,$FF,$CB,$F3,$9A,$FE,$CB,$39,$A0,$9E,$4F,$F8
	.dc.b	$25,$AF,$2B,$FE,$A6,$F0,$8C,$BF,$EA,$FF,$CE,$3F,$B7,$DB,$0F,$F5
	.dc.b	$D4,$3E,$0B,$F9,$A3,$A8,$CC,$63,$D4,$CB,$63,$FD,$6D,$3C,$5B,$F2
	.dc.b	$7C,$BE,$07,$9B,$63,$27,$08,$CB,$4F,$F6,$2F,$D5,$7F,$86,$7E,$A3
	.dc.b	$FA,$73,$FE,$1B,$F3,$AF,$80,$9B,$18,$6D,$72,$9D,$36,$74,$D3,$53
	.dc.b	$C3,$5F,$EB,$FE,$0F,$C1,$7F,$D8,$5C,$DF,$D7,$FD,$14,$E7,$D0,$FF
	.dc.b	$6A,$CE,$63,$FC,$37,$5F,$E9,$AA,$DF,$07,$D4,$6E,$3F,$E3,$E6,$CF
	.dc.b	$F9,$0F,$06,$FF,$E7,$6A,$F9,$BF,$33,$F5,$3E,$0F,$EA,$B6,$32,$7D
	.dc.b	$28,$CB,$FD,$87,$93,$F9,$5E,$0E,$A7,$D4,$FF,$9D,$EB,$5E,$47,$F7
	.dc.b	$B9,$CF,$AF,$FD,$93,$0E,$31,$8A,$D9,$FE,$FA,$71,$87,$E9,$6F,$F2
	.dc.b	$2F,$E5,$DB,$5F,$D1,$63,$F3,$11,$F9,$6D,$1F,$49,$FE,$26,$E9,$FC
	.dc.b	$A5,$27,$38,$7F,$81,$EB,$58,$FE,$92,$72,$FE,$9A,$F8,$C7,$4B,$E1
	.dc.b	$7C,$5B,$5F,$17,$E0,$7E,$06,$9F,$C7,$A3,$07,$B5,$C4,$FF,$01,$9F
	.dc.b	$59,$27,$FC,$E1,$9A,$3E,$0F,$D4,$7F,$16,$A6,$5F,$E0,$0E,$CF,$F0
	.dc.b	$7C,$E6,$31,$87,$F8,$5E,$31,$E8,$7F,$03,$8F,$F7,$1C,$62,$B8,$F8
	.dc.b	$9D,$89,$3F,$E8,$19,$FE,$57,$87,$FA,$35,$7C,$87,$DD,$FC,$CF,$FC
	.dc.b	$CB,$5D,$E3,$1F,$01,$E4,$7D,$24,$8E,$8F,$DC,$F2,$3F,$C6,$B8,$7B
	.dc.b	$27,$F8,$AF,$03,$0F,$F2,$BC,$7F,$8C,$6B,$7A,$FD,$CA,$D7,$B9,$FE
	.dc.b	$30,$9E,$78,$7F,$A8,$DA,$FF,$20,$3F,$A8,$FB,$5E,$13,$F5,$5D,$8E
	.dc.b	$9F,$F1,$DF,$D1,$FE,$C5,$B9,$FF,$96,$A7,$4F,$EB,$DB,$7F,$EB,$1E
	.dc.b	$CF,$DD,$D7,$43,$F2,$9D,$8F,$89,$FE,$46,$7F,$24,$F8,$AB,$8F,$F2
	.dc.b	$08,$B8,$7C,$33,$F9,$74,$E3,$7F,$F3,$9E,$4F,$8A,$FF,$30,$7C,$1F
	.dc.b	$F8,$97,$4F,$E1,$71,$8F,$57,$F8,$53,$EE,$26,$7E,$97,$E0,$FC,$4F
	.dc.b	$2F,$5F,$F8,$83,$67,$E0,$3A,$7E,$03,$74,$F8,$19,$7F,$B5,$67,$67
	.dc.b	$F9,$BF,$E9,$B9,$B8,$9F,$E4,$75,$FE,$6F,$9C,$C7,$BF,$DB,$7E,$06
	.dc.b	$F9,$5F,$6F,$B7,$FE,$85,$9C,$C6,$31,$FD,$B1,$FD,$72,$32,$FF,$B8
	.dc.b	$1B,$7F,$EC,$CB,$E9,$39,$6C,$7F,$B4,$3E,$B9,$F2,$7F,$32,$FC,$11
	.dc.b	$8B,$EF,$5F,$C2,$F8,$7F,$CC,$FC,$5F,$C3,$AD,$7B,$AF,$B5,$E3,$2F
	.dc.b	$78,$79,$EB,$58,$7F,$0E,$9C,$DF,$F9,$FA,$79,$3F,$EE,$06,$BC,$71
	.dc.b	$FE,$EB,$8F,$81,$FE,$89,$F2,$4F,$4F,$FD,$49,$F4,$DF,$E7,$6B,$FF
	.dc.b	$58,$CE,$73,$F5,$E8,$AD,$69,$FE,$C0,$AD,$5F,$AA,$73,$7F,$EB,$E9
	.dc.b	$CD,$7F,$26,$9C,$D7,$FB,$2E,$73,$5B,$C3,$C7,$F2,$BF,$D8,$67,$35
	.dc.b	$B9,$C1,$AA,$FF,$64,$CE,$A7,$FC,$33,$EC,$FF,$AF,$FF,$3D,$EF,$18
	.dc.b	$AF,$5C,$62,$3F,$C6,$C7,$D2,$71,$3C,$7E,$5F,$FD,$6F,$3A,$CF,$FD
	.dc.b	$D5,$7F,$13,$37,$D2,$EB,$0D,$DB,$E3,$FF,$C5,$7F,$4D,$A8,$FE,$1D
	.dc.b	$E3,$7F,$50,$8C,$47,$F1,$19,$DD,$FF,$13,$8F,$ED,$58,$E1,$C4,$D7
	.dc.b	$F9,$1C,$57,$FC,$73,$16,$FB,$57,$39,$F2,$1F,$EC,$6F,$A7,$C5,$7E
	.dc.b	$95,$B1,$93,$B9,$47,$CA,$FF,$AA,$F4,$E1,$1F,$E4,$88,$C3,$B6,$F2
	.dc.b	$2B,$F9,$74,$61,$BF,$13,$FD,$62,$9A,$7C,$4A,$FF,$75,$47,$BD,$AE
	.dc.b	$B2,$E3,$8E,$6F,$95,$F1,$F0,$3F,$9E,$36,$F5,$5C,$FA,$FF,$A0,$3F
	.dc.b	$E0,$28,$F6,$C6,$36,$F3,$B6,$3F,$C0,$DC,$0E,$0C,$47,$03,$F4,$FF
	.dc.b	$9F,$3D,$FE,$1D,$08,$C4,$FF,$06,$38,$1F,$EB,$FD,$FF,$D6,$0F,$2A
	.dc.b	$E3,$3C,$23,$58,$F8,$FC,$37,$CC,$FF,$23,$9F,$E2,$31,$8E,$FD,$4E
	.dc.b	$67,$F8,$7F,$D4,$3C,$6C,$D4,$9E,$35,$BE,$0F,$F0,$ED,$EE,$73,$5E
	.dc.b	$C8,$FE,$B1,$1F,$C1,$9A,$FF,$55,$C3,$C4,$F0,$35,$E3,$F1,$3C,$0F
	.dc.b	$F3,$DF,$25,$ED,$FE,$5E,$1D,$3F,$DC,$63,$0D,$8F,$E7,$53,$C1,$B3
	.dc.b	$F4,$3E,$49,$F5,$F4,$FB,$57,$C6,$BC,$15,$97,$F8,$FD,$70,$7F,$9D
	.dc.b	$CE,$F4,$FE,$75,$E3,$BB,$FC,$6D,$AF,$E9,$8D,$7F,$75,$48,$C4,$7F
	.dc.b	$83,$E3,$13,$F8,$6F,$1B,$FD,$CD,$3F,$CC,$1A,$FF,$61,$AE,$97,$FD
	.dc.b	$FB,$8B,$F0,$5E,$4F,$E2,$2B,$FD,$17,$1F,$06,$32,$C8,$FF,$0B,$39
	.dc.b	$69,$FE,$17,$7E,$86,$FF,$E9,$06,$CF,$59,$FF,$47,$3D,$EB,$FA,$A7
	.dc.b	$BB,$E9,$3B,$4F,$FA,$71,$E2,$47,$FC,$48,$D7,$FB,$0E,$3F,$F0,$01
	.dc.b	$F8,$1E,$87,$B2,$7F,$53,$E7,$3C,$CA,$4A,$FF,$61,$F9,$D5,$FE,$D7
	.dc.b	$3B,$7F,$D8,$F3,$CA,$7F,$51,$F9,$FF,$DD,$9F,$F1,$A3,$5F,$F5,$53
	.dc.b	$5F,$F7,$1E,$A8,$C4,$23,$1C,$7F,$BE,$F9,$4F,$E7,$7A,$FF,$A1,$1B
	.dc.b	$E0,$F2,$C3,$C1,$3F,$C3,$FF,$40,$7D,$0F,$E7,$F2,$46,$1F,$D1,$35
	.dc.b	$FE,$74,$D7,$FA,$33,$99,$FE,$35,$FA,$0B,$E8,$3F,$8A,$B6,$BF,$88
	.dc.b	$56,$F1,$FF,$30,$7C,$F3,$87,$C4,$DA,$FD,$93,$5F,$E9,$4D,$7F,$AB
	.dc.b	$39,$C7,$F9,$57,$E8,$4F,$90,$FE,$4A,$DA,$FE,$41,$5B,$C3,$FC,$E1
	.dc.b	$F5,$9F,$ED,$E9,$FE,$00,$D7,$FA,$C3,$5F,$EC,$8D,$78,$F3,$FF,$32
	.dc.b	$FA,$CF,$C5,$47,$F9,$94,$38,$CB,$F9,$F5,$75,$6D,$71,$E7,$97,$FC
	.dc.b	$D6,$B5,$87,$9B,$1F,$AD,$06,$92,$7D,$2F,$8F,$BA,$47,$38,$FE,$87
	.dc.b	$F9,$07,$F9,$D6,$3F,$C5,$AF,$F2,$8C,$BF,$D9,$6D,$C3,$FD,$A0,$EE
	.dc.b	$7F,$C8,$ED,$9D,$F8,$3D,$0E,$BD,$1F,$D5,$3E,$A3,$F8,$7D,$FF,$A7
	.dc.b	$5A,$83,$86,$F7,$FE,$4E,$3F,$87,$E3,$FE,$A5,$FA,$BF,$F5,$ED,$7F
	.dc.b	$59,$6B,$F3,$E7,$6F,$56,$A8,$DE,$3F,$DB,$DB,$5E,$5B,$FD,$FF,$A1
	.dc.b	$FC,$03,$E9,$66,$1F,$E5,$2E,$F4,$EA,$31,$3A,$7F,$80,$C7,$EF,$BE
	.dc.b	$3F,$F8,$DF,$61,$A4,$9F,$F1,$27,$7F,$EF,$63,$F7,$5E,$7C,$9F,$ED
	.dc.b	$75,$C5,$F8,$1F,$35,$FE,$00,$D5,$F3,$74,$C4,$49,$F8,$D1,$97,$89
	.dc.b	$0F,$CB,$4F,$27,$66,$77,$6F,$F5,$FD,$16,$FF,$CC,$63,$0B,$E0,$3F
	.dc.b	$93,$9F,$CA,$3C,$79,$87,$C9,$8C,$BF,$2E,$3F,$96,$F7,$FF,$2E,$F9
	.dc.b	$B9,$3F,$3E,$32,$FD,$18,$CB,$FC,$BA,$3F,$1D,$AF,$E1,$C9,$1F,$DE
	.dc.b	$79,$51,$88,$F7,$FB,$DF,$24,$E2,$FF,$88,$E7,$6F,$E3,$63,$4E,$67
	.dc.b	$FC,$65,$F2,$9F,$E4,$BE,$8B,$ED,$7E,$8F,$6F,$CE,$4E,$6F,$F1,$1D
	.dc.b	$FF,$E3,$6F,$C8,$3E,$44,$25,$F9,$1F,$E4,$A3,$0D,$53,$C3,$8B,$E8
	.dc.b	$E3,$FD,$50,$93,$FE,$C0,$BE,$7F,$13,$E4,$F7,$3C,$1C,$32,$B5,$7F
	.dc.b	$E5,$AE,$5F,$F2,$86,$5F,$DF,$9F,$BF,$3E,$23,$DF,$9F,$FA,$0E,$37
	.dc.b	$78,$98,$78,$67,$5A,$FE,$F6,$3D,$33,$FE,$56,$E2,$7C,$6D,$24,$FA
	.dc.b	$7F,$35,$7C,$86,$21,$97,$93,$77,$E8,$6F,$94,$FA,$6D,$AF,$F2,$E6
	.dc.b	$5D,$9B,$96,$AA,$46,$0D,$FC,$C7,$94,$8F,$8F,$DB,$AF,$8B,$6F,$BB
	.dc.b	$8B,$E5,$BC,$26,$B4,$FF,$3D,$66,$C7,$38,$7D,$5B,$E0,$F8,$0F,$F3
	.dc.b	$97,$E1,$4F,$94,$FF,$3D,$B6,$33,$1F,$90,$F8,$8F,$F4,$36,$3F,$E8
	.dc.b	$7A,$D5,$79,$4F,$C6,$78,$3F,$0D,$C7,$07,$F2,$1D,$FA,$E3,$27,$F1
	.dc.b	$15,$BC,$3E,$73,$76,$FB,$58,$F6,$9F,$07,$93,$E6,$25,$FE,$47,$F1
	.dc.b	$1D,$5F,$09,$FD,$3E,$1F,$02,$BC,$6D,$BE,$71,$F0,$A7,$7C,$12,$30
	.dc.b	$7C,$6F,$A6,$F8,$A4,$7F,$41,$F1,$5F,$E2,$F7,$FB,$86,$69,$FD,$BF
	.dc.b	$A7,$FE,$A7,$9F,$D3,$35,$DC,$73,$7E,$3C,$67,$1F,$C8,$37,$3F,$87
	.dc.b	$5F,$EB,$0F,$A6,$FE,$87,$7F,$E8,$76,$BF,$ED,$F6,$BF,$EE,$B6,$BF
	.dc.b	$A5,$77,$7F,$E3,$7E,$EF,$F1,$D1,$D7,$CD,$FF,$45,$7D,$A7,$FB,$07
	.dc.b	$17,$E7,$EF,$90,$C3,$72,$FF,$44,$D7,$07,$FB,$76,$3F,$BC,$BE,$37
	.dc.b	$C0,$D4,$FF,$A9,$FF,$1B,$7D,$4F,$FA,$F5,$B1,$FE,$BC,$6B,$F8,$FF
	.dc.b	$15,$F0,$CF,$B4,$E1,$72,$B8,$3F,$EA,$38,$FF,$6B,$FD,$7B,$4F,$E0
	.dc.b	$DC,$9F,$E3,$D8,$C5,$FF,$81,$63,$13,$AE,$FF,$EA,$5F,$5D,$FE,$53
	.dc.b	$9C,$D7,$F8,$2F,$6F,$E2,$0B,$0E,$6F,$FB,$B4,$C6,$1F,$C5,$67,$73
	.dc.b	$A6,$C7,$79,$68,$F8,$77,$93,$98,$C3,$B9,$EA,$7A,$73,$FF,$0F,$7C
	.dc.b	$4F,$F6,$8F,$8F,$E8,$D9,$D1,$FE,$CE,$F5,$3F,$DD,$DA,$FF,$98,$71
	.dc.b	$F9,$BF,$69,$C2,$7F,$DE,$5A,$FF,$AE,$5B,$7D,$9A,$F3,$C7,$DC,$FF
	.dc.b	$33,$DB,$F9,$6E,$5F,$ED,$D6,$C7,$C8,$3E,$47,$93,$FE,$E0,$71,$1A
	.dc.b	$7E,$6F,$C8,$BD,$CF,$EB,$51,$C6,$31,$E3,$7F,$BA,$5B,$19,$9D,$BF
	.dc.b	$43,$DF,$79,$1E,$36,$C7,$81,$AE,$E3,$CD,$93,$EC,$1F,$49,$FD,$63
	.dc.b	$8B,$FD,$11,$39,$8C,$71,$FF,$86,$7A,$9F,$41,$8C,$BF,$DF,$B8,$F2
	.dc.b	$E3,$FA,$78,$F1,$6C,$66,$59,$7D,$5C,$39,$D6,$B2,$E9,$4F,$7B,$FD
	.dc.b	$25,$3F,$95,$8F,$EC,$91,$87,$F2,$8D,$8F,$CC,$B6,$37,$F0,$FD,$23
	.dc.b	$86,$1F,$85,$EB,$7C,$FE,$6A,$61,$F8,$CC,$3C,$51,$EF,$E9,$5F,$0A
	.dc.b	$D6,$3D,$37,$FF,$67,$FD,$87,$97,$C2,$FF,$01,$FC,$1A,$31,$7F,$2F
	.dc.b	$D4,$FF,$03,$D6,$A7,$3A,$7F,$B5,$5B,$7C,$74,$DE,$1C,$4F,$E2,$A7
	.dc.b	$81,$CD,$7E,$C3,$83,$FB,$54,$EA,$3F,$CA,$7F,$60,$FA,$1D,$EF,$FC
	.dc.b	$4B,$D5,$E3,$6A,$D3,$C6,$74,$F3,$35,$5C,$B0,$EF,$87,$F6,$B9,$C9
	.dc.b	$E1,$5E,$5B,$DB,$FE,$19,$1F,$E9,$B8,$C4,$FF,$43,$9C,$CF,$06,$CE
	.dc.b	$A3,$FD,$8C,$DD,$BF,$C6,$38,$BF,$10,$FC,$56,$1E,$1A,$D3,$FE,$2D
	.dc.b	$FB,$17,$4F,$C8,$EF,$7D,$37,$F3,$AA,$F4,$56,$9F,$F2,$34,$67,$3F
	.dc.b	$F6,$B4,$E5,$FE,$04,$D7,$F9,$5B,$5F,$05,$FB,$23,$86,$7F,$E4,$89
	.dc.b	$46,$2F,$FC,$C7,$2F,$53,$EE,$D6,$AE,$FD,$8F,$F8,$AE,$FF,$CB,$5E
	.dc.b	$87,$BA,$BF,$D8,$D3,$97,$FC,$37,$19,$51,$97,$FA,$0B,$5F,$F2,$93
	.dc.b	$5E,$7E,$AF,$D4,$8E,$98,$FA,$5B,$7F,$E7,$AD,$3F,$8F,$3F,$81,$8C
	.dc.b	$DF,$93,$E3,$BE,$86,$2F,$7B,$D7,$18,$6D,$FF,$2D,$AF,$3C,$F8,$49
	.dc.b	$19,$B4,$70,$9F,$F1,$CE,$1F,$EA,$DA,$F1,$4F,$CA,$FA,$FF,$F6,$0C
	.dc.b	$7F,$A0,$7E,$BB,$E2,$46,$2F,$FC,$83,$18,$F8,$AF,$C9,$CE,$67,$FB
	.dc.b	$3B,$5F,$F3,$C3,$5F,$F7,$B3,$CF,$4F,$BC,$DB,$FF,$3F,$7D,$8B,$E2
	.dc.b	$BC,$98,$79,$7E,$23,$F2,$99,$7C,$5B,$FD,$56,$1F,$EB,$47,$D1,$B6
	.dc.b	$DF,$8D,$19,$CF,$FA,$A3,$5F,$3D,$FA,$8E,$0F,$F9,$84,$64,$F8,$7C
	.dc.b	$E9,$FE,$89,$FB,$1F,$F3,$76,$3F,$EA,$76,$DF,$1C,$5F,$5D,$FA,$B9
	.dc.b	$CD,$FC,$08,$F9,$24,$BF,$3B,$FB,$EF,$03,$71,$F2,$A3,$F3,$F1,$8C
	.dc.b	$3E,$5C,$7F,$18,$F9,$9F,$EC,$35,$FC,$55,$B3,$FE,$78,$C1,$F6,$3B
	.dc.b	$F3,$BF,$01,$E4,$7D,$A2,$61,$F2,$1F,$14,$F8,$0E,$D7,$67,$F8,$D7
	.dc.b	$D7,$7F,$AC,$9F,$E4,$5C,$5F,$3F,$5F,$DD,$4E,$6B,$F9,$14,$E6,$BF
	.dc.b	$D7,$93,$99,$FF,$17,$B5,$CF,$91,$FC,$82,$31,$00,$F5,$BE,$F3,$FD
	.dc.b	$03,$27,$D7,$FE,$1E,$FE,$E5,$EF,$0E,$5E,$13,$D7,$FE,$E9,$97,$F2
	.dc.b	$85,$F6,$3F,$D1,$1C,$D7,$07,$2F,$F4,$37,$38,$7F,$8F,$9C,$BF,$E5
	.dc.b	$D7,$37,$DE,$BA,$7C,$07,$5F,$66,$E1,$CD,$B8,$71,$6E,$3E,$57,$FA
	.dc.b	$EF,$D8,$FF,$C0,$1F,$95,$BD,$FF,$D4,$2B,$FD,$E3,$39,$C3,$FB,$02
	.dc.b	$9D,$47,$CB,$8D,$61,$FE,$C1,$75,$E6,$F4,$C3,$AB,$6B,$9D,$F2,$B5
	.dc.b	$F2,$C3,$C3,$39,$F3,$BF,$89,$FA,$9F,$97,$E9,$FE,$EB,$18,$C7,$F3
	.dc.b	$11,$8F,$0B,$E5,$CE,$72,$F9,$5F,$6F,$9C,$3E,$A7,$CC,$FD,$5F,$A3
	.dc.b	$B9,$7D,$36,$DF,$D1,$6E,$9F,$75,$FE,$93,$CB,$FC,$CB,$9F,$F0,$FE
	.dc.b	$A6,$3C,$9F,$C5,$74,$A3,$19,$7F,$31,$BF,$E8,$59,$7E,$4A,$FF,$2C
	.dc.b	$49,$FD,$2B,$EE,$FB,$DF,$E3,$91,$83,$FC,$15,$F1,$FF,$E3,$BD,$3F
	.dc.b	$BB,$BF,$B8,$6D,$FF,$85,$23,$13,$F9,$87,$83,$73,$D4,$ED,$F6,$FC
	.dc.b	$07,$AB,$C5,$B5,$FE,$21,$9F,$D2,$C6,$2B,$F4,$B5,$F0,$3E,$0B,$18
	.dc.b	$FF,$73,$FB,$3F,$E2,$0F,$F8,$13,$C0,$FE,$E3,$18,$F4,$BF,$59,$87
	.dc.b	$ED,$63,$F9,$1B,$FE,$AA,$6B,$FB,$90,$E1,$E1,$F2,$1F,$E1,$E8,$FE
	.dc.b	$3D,$1A,$8F,$95,$97,$81,$F0,$FC,$5F,$EF,$FE,$CF,$FC,$D3,$5F,$CE
	.dc.b	$6D,$39,$F8,$1D,$EF,$FE,$31,$9C,$BF,$0E,$FD,$37,$E4,$BF,$F0,$6E
	.dc.b	$3F,$E4,$08,$68,$BF,$E2,$5B,$B0,$FF,$31,$57,$F4,$9B,$E1,$7B,$8F
	.dc.b	$F4,$AC,$EA,$3F,$C3,$D3,$AB,$F8,$8D,$AD,$FC,$47,$8B,$DC,$FD,$27
	.dc.b	$47,$F9,$1B,$B3,$FB,$EB,$7B,$4D,$26,$3F,$E5,$F9,$DB,$FC,$5F,$E7
	.dc.b	$39,$B4,$98,$FF,$17,$FA,$9F,$E4,$13,$27,$3A,$7F,$25,$9C,$57,$C4
	.dc.b	$9F,$E6,$97,$FE,$10,$F9,$9F,$21,$23,$FD,$89,$19,$6E,$49,$FE,$00
	.dc.b	$7F,$9A,$7C,$5B,$3E,$99,$7D,$37,$BC,$91,$C9,$B8,$78,$C6,$1F,$E2
	.dc.b	$28,$FF,$3A,$7D,$3F,$F0,$D6,$57,$57,$2D,$25,$73,$AE,$51,$FE,$81
	.dc.b	$F8,$15,$F2,$AF,$07,$B8,$FB,$27,$CC,$FF,$9F,$23,$0D,$8C,$E1,$FE
	.dc.b	$D2,$6D,$7E,$94,$FF,$39,$B8,$DF,$F1,$57,$CA,$BC,$5D,$FF,$E9,$E8
	.dc.b	$C5,$F8,$BE,$C9,$F6,$5F,$E9,$CF,$91,$FF,$35,$4F,$F9,$FA,$73,$87
	.dc.b	$FA,$6E,$73,$69,$CD,$78,$E3,$17,$EB,$9C,$CF,$EE,$E7,$37,$FE,$EC
	.dc.b	$9E,$38,$FA,$1B,$1F,$EE,$9F,$B9,$F4,$46,$2F,$EB,$FD,$9B,$D2,$FF
	.dc.b	$59,$FD,$87,$F9,$FF,$0F,$CD,$4E,$5F,$F5,$94,$66,$BE,$B4,$E5,$F9
	.dc.b	$5C,$48,$C4,$7F,$11,$8C,$5F,$EA,$E7,$33,$F9,$A7,$96,$5F,$ED,$86
	.dc.b	$E1,$FE,$08,$C1,$F4,$5F,$D6,$A7,$37,$EA,$C3,$07,$AA,$6D,$FF,$B0
	.dc.b	$67,$FD,$66,$7E,$5F,$D6,$7D,$9F,$AF,$2F,$F7,$DD,$7F,$B3,$E3,$0F
	.dc.b	$FB,$F0,$FF,$41,$57,$23,$F1,$05,$34,$7F,$A1,$FD,$BF,$F5,$3B,$E0
	.dc.b	$AF,$F6,$2E,$6F,$B3,$CD,$6A,$9F,$23,$FE,$2F,$F2,$F5,$1F,$CE,$7C
	.dc.b	$5F,$E3,$91,$88,$FE,$79,$3B,$BF,$CD,$56,$9F,$2C,$66,$79,$65,$5E
	.dc.b	$22,$FA,$3F,$E5,$A6,$BE,$43,$FD,$AE,$F4,$69,$FA,$99,$FF,$23,$71
	.dc.b	$F3,$8C,$1A,$1F,$2A,$F8,$9F,$E7,$B1,$98,$C6,$BC,$B0,$FE,$B1,$18
	.dc.b	$9D,$57,$0B,$ED,$1D,$23,$9D,$7C,$5A,$FF,$7F,$4F,$BD,$AF,$BF,$2E
	.dc.b	$B5,$C1,$B8,$6F,$E3,$BF,$B4,$61,$CD,$B3,$AB,$E4,$7F,$99,$70,$FF
	.dc.b	$55,$3F,$E6,$7B,$FF,$76,$46,$23,$FD,$3C,$D8,$FC,$1E,$BF,$DD,$5D
	.dc.b	$FC,$4F,$E3,$A7,$85,$7F,$15,$6B,$F1,$23,$47,$1D,$7F,$CD,$BF,$64
	.dc.b	$F7,$B1,$4E,$34,$E6,$3C,$D5,$CD,$FE,$B1,$F6,$5F,$CA,$9B,$DF,$CB
	.dc.b	$A5,$39,$9F,$D5,$7D,$93,$C7,$D4,$E7,$6F,$EE,$57,$C3,$0C,$9E,$4C
	.dc.b	$D8,$71,$C7,$B9,$FD,$63,$0E,$CF,$F0,$96,$BF,$D8,$1A,$FF,$50,$7C
	.dc.b	$19,$7F,$A3,$A5,$E4,$77,$6D,$76,$3F,$A4,$F9,$2F,$6B,$E3,$F3,$FF
	.dc.b	$7C,$6C,$66,$7F,$A9,$7D,$D6,$27,$F8,$0C,$2B,$83,$FC,$11,$AF,$F3
	.dc.b	$BA,$E4,$D9,$F8,$5E,$3B,$EC,$E9,$D6,$7E,$A9,$D7,$C8,$7C,$87,$99
	.dc.b	$B1,$AC,$7F,$B2,$CF,$4F,$6B,$4F,$9A,$33,$3E,$5A,$E2,$FF,$4F,$6B
	.dc.b	$FC,$DF,$CC,$FC,$B6,$FF,$E0,$E6,$C7,$F7,$E6,$F0,$B3,$47,$E1,$CD
	.dc.b	$F1,$E3,$A3,$07,$C0,$8B,$7F,$E0,$F8,$AF,$F3,$5A,$E6,$E4,$FF,$0C
	.dc.b	$C7,$D0,$AE,$91,$83,$FC,$3B,$EE,$78,$48,$FF,$82,$7E,$5E,$BF,$D5
	.dc.b	$B0,$FF,$47,$3C,$08,$7F,$80,$F1,$FF,$6B,$79,$DF,$97,$F8,$6F,$F1
	.dc.b	$1C,$FF,$B4,$63,$1E,$8E,$59,$19,$3F,$DA,$D1,$FA,$9F,$AC,$FB,$D1
	.dc.b	$FE,$45,$6B,$FE,$1F,$6F,$5F,$FB,$8E,$B8,$72,$7B,$E7,$75,$FD,$AD
	.dc.b	$AF,$F9,$09,$AF,$FB,$DF,$EC,$1F,$67,$FC,$0A,$11,$87,$F1,$A6,$23
	.dc.b	$5F,$64,$F9,$2E,$15,$BA,$FF,$1F,$09,$7D,$9F,$03,$77,$DF,$E7,$BE
	.dc.b	$6D,$70,$6E,$9F,$E7,$58,$46,$39,$BF,$D2,$DF,$66,$F8,$1E,$5C,$BE
	.dc.b	$98,$7C,$9D,$33,$87,$AB,$EC,$FF,$0E,$F9,$9B,$97,$76,$E3,$FE,$A3
	.dc.b	$FA,$9F,$DC,$57,$1E,$7F,$EF,$B5,$CA,$3F,$C8,$8D,$7F,$D5,$4D,$41
	.dc.b	$EA,$E2,$5F,$01,$F4,$5B,$1D,$CF,$46,$BF,$56,$BE,$1C,$60,$FA,$B3
	.dc.b	$FC,$61,$F6,$4F,$3F,$06,$46,$CF,$F0,$E7,$D9,$FF,$9F,$5F,$E1,$3D
	.dc.b	$48,$E5,$A7,$FA,$96,$31,$E3,$7F,$33,$9E,$15,$CB,$AB,$13,$23,$9C
	.dc.b	$EA,$FE,$D7,$D9,$7F,$0A,$6F,$87,$F8,$F8,$C7,$07,$FB,$47,$EB,$5E
	.dc.b	$7E,$6B,$FC,$57,$5F,$C4,$31,$FF,$6B,$35,$F9,$AD,$4F,$47,$E8,$D7
	.dc.b	$FB,$41,$FC,$DB,$5F,$CB,$C6,$F4,$77,$FF,$84,$67,$E2,$AF,$A5,$FD
	.dc.b	$F1,$FB,$FF,$52,$4E,$A3,$2F,$08,$FD,$E7,$C1,$48,$FD,$E3,$FC,$1F
	.dc.b	$D4,$C4,$7F,$93,$FE,$67,$E0,$BF,$51,$AF,$F0,$19,$FD,$B6,$7F,$B1
	.dc.b	$8F,$5F,$CE,$E0,$CF,$85,$23,$24,$3F,$6B,$3C,$5B,$1C,$5D,$70,$FE
	.dc.b	$C6,$74,$FE,$0F,$4F,$E7,$63,$EC,$FF,$90,$1F,$C8,$7C,$99,$7F,$22
	.dc.b	$32,$FC,$A8,$FE,$4D,$DF,$FC,$9F,$E7,$A4,$FC,$E8,$CB,$F4,$23,$2F
	.dc.b	$F9,$CE,$FF,$81,$3F,$80,$CB,$F1,$44,$7E,$EF,$DB,$D7,$FC,$E3,$C7
	.dc.b	$EF,$7F,$D7,$4D,$7F,$17,$3F,$8E,$3E,$6F,$BF,$23,$F9,$0C,$FC,$2E
	.dc.b	$7F,$DD,$FC,$BF,$F9,$8F,$AF,$FC,$52,$DA,$FE,$E6,$D7,$D1,$E0,$7F
	.dc.b	$2D,$FB,$4F,$F1,$AC,$7F,$04,$6A,$69,$AE,$9F,$C9,$FB,$BF,$D3,$49
	.dc.b	$3D,$BE,$D3,$F3,$5C,$8C,$FC,$B5,$FE,$5A,$8F,$E5,$58,$38,$7C,$47
	.dc.b	$A2,$75,$5F,$C5,$B0,$F1,$4F,$FA,$6A,$B8,$47,$F8,$69,$AF,$FA,$71
	.dc.b	$AF,$1C,$D8,$FF,$B9,$F1,$F2,$4A,$72,$FC,$89,$D6,$1E,$1B,$EB,$7F
	.dc.b	$D0,$78,$9F,$25,$FD,$55,$B8,$7F,$54,$6A,$9E,$F8,$BE,$D1,$F2,$9C
	.dc.b	$70,$FF,$5A,$E1,$C3,$0F,$F7,$E3,$5F,$84,$D7,$F9,$CB,$DB,$0E,$BF
	.dc.b	$68,$F8,$DF,$C0,$B6,$BC,$97,$E7,$F9,$D9,$1E,$77,$F0,$F9,$7F,$94
	.dc.b	$A3,$AC,$79,$08,$FF,$92,$31,$46,$23,$DD,$AF,$E4,$B0,$7B,$5E,$23
	.dc.b	$BC,$FF,$72,$FB,$3F,$EC,$CF,$F7,$A6,$B9,$F3,$FF,$33,$F0,$FF,$DB
	.dc.b	$F6,$FE,$E1,$5A,$AF,$C2,$07,$0F,$E1,$DA,$FF,$5E,$6B,$BA,$FD,$2F
	.dc.b	$7F,$C5,$8F,$49,$F8,$37,$15,$E0,$9F,$F0,$95,$7F,$39,$AF,$07,$8D
	.dc.b	$83,$F5,$FF,$69,$EA,$BF,$3C,$7F,$BC,$2B,$FA,$11,$1C,$1E,$2C,$A5
	.dc.b	$FC,$38,$EF,$F6,$92,$3E,$33,$FB,$CF,$89,$FF,$1D,$C7,$CB,$8E,$C4
	.dc.b	$7F,$CF,$B6,$8C,$47,$BB,$95,$E7,$FF,$3A,$CB,$F2,$AD,$E0,$F9,$57
	.dc.b	$FE,$31,$FA,$9F,$EA,$18,$7F,$8A,$9B,$7F,$E2,$36,$D7,$F8,$33,$B1
	.dc.b	$F8,$43,$E0,$39,$79,$9F,$E9,$0C,$58,$74,$7E,$DB,$5F,$E0,$0D,$7F
	.dc.b	$D8,$5F,$00,$F9,$6F,$F6,$2E,$4F,$B5,$7C,$86,$1B,$7F,$E0,$AB,$E0
	.dc.b	$DB,$D5,$48,$C1,$FE,$98,$F8,$37,$DA,$BE,$D7,$FD,$94,$D9,$D3,$5F
	.dc.b	$F4,$07,$DA,$EA,$64,$69,$FF,$53,$C7,$F9,$6F,$ED,$3C,$72,$E7,$FF
	.dc.b	$27,$C6,$1B,$97,$F2,$17,$FC,$FF,$C4,$FB,$3F,$F2,$E4,$6A,$FF,$C7
	.dc.b	$6D,$7A,$78,$3F,$CA,$70,$E8,$FD,$5A,$E2,$E7,$0F,$E5,$DC,$1F,$AC
	.dc.b	$AF,$2B,$63,$37,$B3,$FC,$93,$1F,$CD,$7E,$27,$FC,$2B,$F2,$3E,$7C
	.dc.b	$3F,$C9,$B1,$8A,$F2,$C6,$3D,$2F,$F7,$D4,$79,$8F,$BB,$F6,$BC,$5F
	.dc.b	$F6,$A4,$6B,$4F,$F4,$FD,$F3,$9E,$0D,$8F,$8B,$C6,$BE,$54,$D3,$F9
	.dc.b	$7D,$FF,$B8,$1B,$1F,$C2,$8D,$79,$3F,$E6,$C7,$E1,$35,$FE,$A5,$F2
	.dc.b	$7F,$E2,$FE,$EF,$F7,$4B,$63,$93,$5F,$E2,$AD,$7E,$97,$C0,$3C,$6F
	.dc.b	$13,$63,$C5,$87,$FA,$32,$72,$7F,$BC,$FE,$D5,$F4,$FF,$D6,$36,$F8
	.dc.b	$71,$8E,$2F,$9B,$87,$E9,$BE,$8F,$5C,$5F,$4A,$B5,$3F,$21,$B7,$CB
	.dc.b	$0D,$CF,$1C,$39,$EE,$BD,$E1,$FD,$CF,$E0,$3F,$D3,$75,$E0,$BF,$F3
	.dc.b	$2F,$C3,$72,$E8,$FC,$B7,$C7,$F4,$9C,$D6,$EF,$8C,$7E,$04,$C9,$F8
	.dc.b	$3A,$E3,$3B,$BE,$55,$C3,$4F,$F5,$AB,$71,$FC,$E6,$3A,$AF,$5F,$6B
	.dc.b	$ED,$7B,$F8,$BF,$E0,$08,$C4,$FF,$80,$A3,$FD,$19,$D1,$FD,$F3,$3F
	.dc.b	$E0,$57,$C1,$FF,$53,$DF,$07,$37,$C5,$FE,$41,$5C,$75,$FE,$97,$5B
	.dc.b	$7F,$A6,$EA,$9F,$ED,$35,$FE,$B2,$C7,$FC,$26,$BE,$D5,$F4,$5D,$70
	.dc.b	$FE,$43,$E8,$DA,$D4,$E7,$1E,$93,$9B,$E3,$5F,$C8,$63,$15,$CE,$FF
	.dc.b	$98,$FC,$25,$F1,$8F,$3F,$B5,$FF,$4F,$E1,$FE,$EE,$9C,$B6,$78,$1B
	.dc.b	$E2,$B1,$FE,$F4,$FB,$57,$37,$FB,$9F,$1F,$F1,$D4,$62,$7F,$61,$08
	.dc.b	$C3,$FB,$36,$BF,$DF,$9C,$ED,$FE,$F1,$6C,$FF,$98,$21,$E7,$FB,$67
	.dc.b	$35,$FE,$DE,$6B,$DE,$FE,$67,$77,$F8,$D6,$B8,$47,$C5,$FB,$5F,$F0
	.dc.b	$CE,$3F,$E6,$8F,$2F,$E0,$E1,$FE,$06,$13,$A7,$FB,$64,$6B,$2F,$F2
	.dc.b	$6C,$6B,$D1,$E8,$78,$32,$FF,$5C,$61,$F8,$13,$FC,$BD,$87,$3B,$E2
	.dc.b	$D8,$CD,$71,$F8,$8E,$71,$C7,$0F,$F7,$DD,$3E,$37,$DB,$97,$F9,$B6
	.dc.b	$B5,$5E,$89,$CF,$BB,$EF,$64,$FF,$A1,$B3,$FF,$55,$3C,$30,$E8,$D9
	.dc.b	$DD,$FF,$82,$EB,$6F,$C5,$AD,$BF,$1F,$ED,$1F,$11,$F4,$3E,$CB,$E3
	.dc.b	$63,$FE,$7D,$9C,$BF,$E9,$68,$CD,$04,$E9,$E9,$5F,$9B,$FB,$57,$C4
	.dc.b	$FE,$76,$B5,$C1,$F4,$9B,$8F,$7B,$E1,$5B,$78,$7C,$47,$97,$6B,$5E
	.dc.b	$36,$E5,$EB,$9F,$07,$C2,$E3,$2E,$D1,$6F,$FD,$31,$9A,$B1,$7F,$E8
	.dc.b	$FF,$B5,$7B,$DF,$92,$6D,$FD,$6D,$5F,$65,$9A,$F6,$32,$7E,$DF,$87
	.dc.b	$C3,$0F,$9D,$3F,$BD,$8C,$69,$F0,$63,$17,$C6,$BF,$DF,$9E,$5F,$F0
	.dc.b	$9F,$8E,$FE,$A9,$8F,$FB,$1B,$37,$DA,$E6,$BE,$67,$6C,$AF,$0C,$DC
	.dc.b	$3F,$49,$A7,$09,$C9,$F9,$5F,$11,$97,$F8,$22,$B7,$7F,$C4,$FE,$C7
	.dc.b	$F9,$44,$7F,$9F,$1A,$FF,$B0,$5A,$FE,$0D,$A9,$FE,$61,$B7,$EC,$1B
	.dc.b	$87,$E2,$73,$7D,$97,$E8,$EF,$F7,$55,$A7,$A7,$C3,$F8,$CF,$E6,$15
	.dc.b	$C3,$97,$18,$7A,$FD,$6F,$ED,$79,$7F,$B2,$B5,$43,$ED,$7F,$9A,$3A
	.dc.b	$AF,$F3,$23,$97,$97,$A3,$A9,$DB,$63,$B7,$BD,$CB,$D9,$E6,$D8,$C9
	.dc.b	$E3,$8C,$B4,$FE,$09,$F6,$BF,$ED,$DF,$93,$FE,$94,$FF,$B7,$FC,$DB
	.dc.b	$E1,$26,$C6,$1B,$5C,$A7,$4D,$9D,$34,$D4,$F0,$E1,$FE,$C3,$E1,$FC
	.dc.b	$15,$E8,$7F,$15,$D3,$E2,$FD,$AF,$F5,$D8,$C6,$1F,$E7,$1B,$E1,$18
	.dc.b	$BF,$ED,$3B,$AF,$85,$75,$E0,$FA,$3F,$2F,$F4,$3E,$87,$E8,$B7,$2F
	.dc.b	$B0,$DC,$79,$B7,$8F,$F1,$1E,$57,$F5,$79,$FC,$73,$5F,$C7,$37,$A3
	.dc.b	$FA,$04,$7E,$5E,$3F,$34,$7A,$FE,$D7,$F8,$9E,$1F,$CC,$2B,$F9,$D5
	.dc.b	$74,$8C,$56,$FC,$7E,$31,$FA,$7F,$8F,$FD,$36,$3F,$8D,$7F,$6B,$7F
	.dc.b	$E0,$A7,$33,$FE,$AF,$73,$3B,$9F,$F0,$24,$62,$75,$18,$EA,$FD,$87
	.dc.b	$BF,$27,$FA,$54,$7E,$9B,$E4,$B1,$5F,$A6,$AB,$82,$FF,$59,$DC,$BC
	.dc.b	$C7,$F3,$7B,$7D,$A7,$F1,$57,$FD,$EE,$D7,$4D,$7F,$9F,$35,$E3,$E2
	.dc.b	$FF,$85,$E3,$F8,$DC,$FF,$8C,$B7,$FE,$58,$FF,$5C,$8C,$D7,$2A,$E0
	.dc.b	$7D,$3F,$22,$D6,$FE,$D3,$A1,$FB,$EF,$B4,$FE,$05,$7E,$88,$CC,$EC
	.dc.b	$F8,$B3,$AC,$13,$FE,$2B,$BF,$F1,$7D,$FF,$81,$5B,$A7,$C7,$8F,$F4
	.dc.b	$05,$FF,$B6,$70,$FE,$E7,$5A,$7E,$43,$67,$FB,$03,$76,$57,$19,$ED
	.dc.b	$EC,$7E,$8B,$96,$61,$FC,$57,$E7,$2D,$7E,$CA,$6E,$5F,$C5,$FA,$95
	.dc.b	$8B,$FE,$2A,$FF,$18,$3F,$8C,$FD,$55,$8E,$84,$C3,$F6,$71,$97,$F7
	.dc.b	$FF,$31,$FB,$43,$FC,$A9,$EE,$FF,$08,$8C,$D7,$EF,$67,$4F,$FB,$D7
	.dc.b	$C0,$B3,$FC,$23,$E9,$BF,$7A,$72,$F8,$FB,$5F,$C1,$62,$CF,$F9,$B3
	.dc.b	$05,$63,$4F,$F3,$F7,$7A,$F1,$87,$F9,$DA,$AE,$1F,$C5,$2B,$53,$D4
	.dc.b	$FE,$B1,$87,$47,$12,$C3,$EE,$7B,$2F,$B4,$7D,$5F,$F9,$DE,$30,$FC
	.dc.b	$2E,$DC,$4E,$B0,$FF,$1C,$56,$9B,$19,$C3,$D9,$5C,$2F,$A7,$D8,$BE
	.dc.b	$CF,$F8,$B6,$28,$C6,$3F,$ED,$B8,$C4,$EA,$73,$87,$ED,$4F,$F3,$AB
	.dc.b	$8C,$BD,$6D,$BF,$F4,$DF,$D9,$F8,$2E,$47,$A2,$75,$6F,$17,$FD,$39
	.dc.b	$7A,$8F,$F5,$0F,$D9,$BC,$EF,$E5,$9E,$4A,$31,$8F,$FB,$C6,$31,$3A
	.dc.b	$9C,$D7,$9E,$73,$5F,$E5,$F6,$C7,$FA,$B3,$EC,$9F,$45,$FD,$2E,$F8
	.dc.b	$EB,$FE,$BF,$6E,$1F,$EB,$A6,$DF,$A6,$F8,$CF,$59,$CE,$1D,$AF,$96
	.dc.b	$3F,$EA,$FB,$DE,$1E,$E8,$FF,$56,$3F,$EA,$83,$FD,$7F,$F6,$7F,$E9
	.dc.b	$4C,$7F,$CE,$31,$9B,$FF,$04,$4E,$63,$FD,$18,$66,$7E,$B7,$B1,$DF
	.dc.b	$EC,$FA,$B2,$39,$B9,$69,$DA,$5F,$1B,$FE,$69,$FB,$65,$FE,$07,$C3
	.dc.b	$FE,$29,$18,$AF,$C2,$56,$EB,$57,$67,$22,$B8,$3F,$E2,$D8,$D3,$FD
	.dc.b	$06,$AB,$FD,$62,$76,$FF,$9E,$E3,$FC,$F9,$C9,$FD,$26,$3A,$B5,$E4
	.dc.b	$7F,$B2,$BC,$9F,$CC,$3E,$C3,$07,$FA,$17,$E4,$DF,$0B,$FB,$58,$C8
	.dc.b	$8C,$ED,$E0,$AF,$EA,$D1,$89,$DB,$6F,$C4,$E7,$1F,$7B,$FE,$E6,$9F
	.dc.b	$7B,$5F,$89,$9F,$4C,$7C,$55,$C1,$B1,$E0,$F8,$9F,$EC,$D3,$D1,$B1
	.dc.b	$C7,$2E,$BE,$7F,$F6,$B7,$FD,$97,$8F,$FA,$7B,$1F,$F5,$DB,$5F,$D1
	.dc.b	$6F,$BD,$BD,$B7,$67,$E3,$35,$F8,$8E,$79,$DC,$6B,$D6,$1D,$71,$FF
	.dc.b	$2F,$7D,$A3,$D9,$E0,$DC,$BC,$79,$77,$E6,$7D,$A7,$F2,$49,$FD,$C3
	.dc.b	$5F,$E5,$5A,$FF,$AF,$67,$7D,$BE,$3F,$B4,$78,$FC,$F0,$E1,$A7,$F6
	.dc.b	$BB,$DE,$3D,$30,$E1,$7B,$64,$64,$FF,$49,$61,$D1,$C1,$FC,$37,$0F
	.dc.b	$23,$FE,$19,$BE,$AE,$6B,$F2,$2B,$95,$E5,$3D,$6F,$9D,$18,$6C,$66
	.dc.b	$7F,$A5,$FD,$A6,$26,$47,$F4,$29,$D3,$FD,$6E,$72,$FF,$37,$AE,$8F
	.dc.b	$F3,$D9,$DB,$FE,$2F,$8C,$BB,$F1,$DF,$19,$E4,$6C,$7F,$50,$6B,$FD
	.dc.b	$A7,$D7,$D4,$C8,$D3,$F8,$5F,$21,$3F,$77,$C9,$FC,$27,$17,$D9,$6D
	.dc.b	$FF,$6B,$E3,$E1,$1B,$64,$7F,$9F,$5C,$B4,$87,$FA,$21,$CB,$4C,$F1
	.dc.b	$FF,$7C,$8C,$BF,$DF,$BE,$4F,$C8,$49,$FE,$F3,$F2,$75,$FF,$49,$C5
	.dc.b	$9F,$F5,$C6,$4A,$FD,$17,$A1,$FE,$07,$BF,$F5,$F7,$9B,$83,$3E,$7D
	.dc.b	$8C,$90,$FF,$4A,$BF,$A1,$FB,$37,$DA,$52,$4E,$F2,$FC,$D4,$EA,$3F
	.dc.b	$AE,$35,$FD,$29,$FE,$A8,$71,$1F,$E3,$9B,$FD,$D7,$DA,$3E,$CA,$92
	.dc.b	$C7,$21,$F6,$8F,$8F,$D4,$66,$B9,$3C,$30,$FF,$2C,$28,$CB,$CE,$B9
	.dc.b	$3B,$E2,$FF,$40,$7C,$E3,$87,$F8,$73,$67,$47,$91,$C3,$E5,$FB,$37
	.dc.b	$BD,$C2,$A4,$F0,$15,$C3,$2F,$2D,$F0,$9D,$7D,$9B,$9B,$F8,$F3,$63
	.dc.b	$31,$FC,$87,$EA,$3F,$75,$18,$21,$F1,$E7,$83,$F2,$63,$6E,$67,$F8
	.dc.b	$1A,$9D,$BF,$E2,$D6,$BF,$EA,$36,$BF,$51,$F5,$BE,$8C,$3C,$E7,$97
	.dc.b	$DB,$B1,$92,$1E,$37,$F9,$C7,$D9,$B9,$6C,$49,$F9,$FE,$5A,$B5,$18
	.dc.b	$6B,$F8,$E8,$D4,$F1,$8C,$4E,$B4,$FC,$1F,$D8,$FF,$08,$6F,$6F,$F4
	.dc.b	$B8,$C7,$0F,$FB,$07,$EB,$7F,$A5,$61,$F6,$67,$2F,$E5,$1A,$9B,$1C
	.dc.b	$23,$F3,$11,$FC,$59,$FF,$2C,$35,$FF,$67,$35,$E9,$5F,$CD,$B2,$75
	.dc.b	$7F,$81,$27,$E1,$AF,$9A,$FD,$F9,$FC,$03,$D2,$93,$A8,$CB,$C2,$3F
	.dc.b	$79,$EF,$C8,$FD,$BB,$FC,$23,$D2,$C4,$7E,$E3,$C9,$77,$2B,$3B,$FF
	.dc.b	$AC,$6F,$FB,$96,$DF,$AA,$C3,$F6,$CB,$E6,$32,$6B,$B6,$4F,$E8,$E7
	.dc.b	$4D,$9C,$BC,$A7,$8B,$93,$83,$FC,$7B,$C3,$FE,$3F,$F6,$72,$7E,$2C
	.dc.b	$65,$F8,$F1,$97,$F0,$DB,$FF,$3E,$E0,$F9,$BD,$9F,$CA,$3E,$46,$57
	.dc.b	$CE,$8F,$E5,$66,$1E,$31,$87,$F0,$98,$FE,$24,$8F,$B7,$EC,$EB,$F8
	.dc.b	$28,$BB,$7F,$14,$D5,$5A,$AF,$C2,$D7,$ED,$39,$7E,$64,$60,$FF,$24
	.dc.b	$FC,$83,$D9,$FC,$E7,$47,$F2,$68,$C7,$5F,$F8,$9D,$B5,$F8,$8F,$87
	.dc.b	$FF,$14,$7D,$97,$F5,$E8,$34,$FE,$C3,$3C,$23,$E3,$E1,$FD,$30,$8F
	.dc.b	$F2,$28,$7D,$9B,$E2,$BC,$15,$FE,$4B,$DF,$F9,$8E,$0E,$0F,$09,$AE
	.dc.b	$7B,$97,$F2,$59,$D3,$67,$4F,$7C,$7F,$D8,$16,$D7,$E0,$46,$1B,$3D
	.dc.b	$0F,$E9,$1C,$4F,$95,$FF,$24,$36,$3F,$9A,$35,$FE,$AF,$AB,$EC,$DF
	.dc.b	$21,$C2,$E5,$70,$7E,$74,$EA,$30,$7C,$5F,$B3,$7C,$5F,$D3,$6C,$68
	.dc.b	$F2,$FC,$AF,$90,$8F,$CD,$AF,$F2,$3E,$7F,$D8,$A3,$FA,$A4,$E9,$E2
	.dc.b	$6F,$D9,$C6,$DF,$16,$1A,$31,$FF,$10,$C7,$97,$D4,$DE,$3D,$7E,$CB
	.dc.b	$FA,$94,$FF,$6F,$6E,$1F,$DA,$9B,$3F,$8F,$F8,$6F,$ED,$B3,$9F,$3F
	.dc.b	$FC,$E9,$5A,$BF,$EF,$2D,$C9,$7C,$30,$F0,$5F,$E5,$FB,$7E,$2E,$1F
	.dc.b	$49,$B8,$F5,$8F,$ED,$58,$FF,$30,$27,$91,$F9,$FF,$B3,$F0,$4A,$72
	.dc.b	$F0,$9D,$61,$E6,$7C,$1A,$49,$D1,$DB,$EC,$FD,$64,$7F,$41,$F1,$1F
	.dc.b	$E2,$D7,$E3,$46,$5A,$6B,$D6,$C3,$F6,$4F,$24,$7F,$70,$C4,$D3,$FB
	.dc.b	$9F,$D2,$FF,$3F,$9F,$F1,$03,$74,$FF,$0D,$B7,$6F,$F1,$2F,$88,$FB
	.dc.b	$37,$B9,$FE,$A5,$AD,$61,$FE,$2A,$6A,$73,$5C,$67,$FC,$CD,$F5,$DF
	.dc.b	$EA,$FC,$1F,$66,$F8,$EC,$37,$1F,$EE,$BE,$FA,$F8,$C7,$F8,$9A,$F8
	.dc.b	$DF,$03,$31,$FD,$37,$EC,$DF,$67,$FE,$AF,$6C,$7F,$8B,$BD,$34,$62
	.dc.b	$3F,$BF,$62,$FB,$3E,$17,$27,$85,$6F,$83,$FC,$CB,$5F,$EC,$7F,$B1
	.dc.b	$F3,$57,$FB,$E7,$C4,$FD,$64,$62,$1C,$DF,$E3,$C8,$C3,$FE,$41,$E1
	.dc.b	$7D,$9F,$F9,$2A,$35,$C1,$FB,$56,$DF,$53,$FC,$B3,$7C,$B0,$FA,$F7
	.dc.b	$C5,$F0,$DF,$16,$D7,$11,$5C,$56,$1E,$2F,$82,$E1,$1F,$EB,$BF,$4B
	.dc.b	$FD,$B4,$FA,$AF,$FC,$D3,$F0,$BF,$E0,$AF,$8D,$DB,$67,$CD,$83,$85
	.dc.b	$F6,$7A,$C7,$9E,$DF,$E6,$1C,$38,$61,$D4,$F0,$63,$E2,$C7,$96,$1E
	.dc.b	$49,$F4,$F9,$5F,$E5,$07,$E7,$B5,$F8,$AD,$77,$7F,$ED,$6F,$8F,$FF
	.dc.b	$0F,$78,$B7,$DD,$FE,$37,$5F,$AB,$13,$A7,$0D,$7F,$61,$C1,$FE,$70
	.dc.b	$F1,$BF,$DC,$2D,$8E,$4D,$77,$F0,$1F,$C5,$9F,$DC,$69,$F1,$73,$FE
	.dc.b	$59,$B7,$E0,$F8,$1F,$37,$8E,$2F,$D9,$C7,$F4,$8E,$CF,$9A,$E9,$B7
	.dc.b	$BC,$7E,$DC,$E5,$F9,$98,$F7,$AE,$38,$6E,$1F,$84,$3D,$3C,$36,$BE
	.dc.b	$DA,$F8,$0C,$CC,$9C,$BC,$23,$18,$78,$2B,$93,$67,$55,$C0,$F5,$F3
	.dc.b	$FF,$4D,$EB,$FF,$68,$EF,$67,$E0,$CD,$BF,$6C,$DC,$BF,$1D,$37,$4F
	.dc.b	$F5,$1D,$FF,$0A,$8C,$56,$62,$1F,$C3,$BC,$DB,$87,$F0,$28,$B8,$FF
	.dc.b	$B8,$A6,$C7,$F2,$4F,$6D,$E7,$F8,$5F,$F0,$14,$7F,$A3,$1C,$B7,$D2
	.dc.b	$D4,$ED,$B1,$93,$FD,$27,$FA,$23,$ED,$7D,$8F,$FB,$9A,$33,$18,$F8
	.dc.b	$BC,$B6,$32,$72,$8C,$B4,$CE,$7F,$E9,$BA,$E8,$FF,$A7,$71,$FE,$7B
	.dc.b	$3E,$9F,$82,$F6,$79,$3E,$D6,$C6,$4F,$74,$65,$A6,$6B,$A4,$65,$F0
	.dc.b	$3E,$C9,$F6,$B6,$7E,$0B,$67,$71,$F1,$27,$6D,$36,$58,$F5,$7C,$DF
	.dc.b	$D3,$F5,$FE,$A3,$67,$EA,$36,$77,$1F,$5E,$76,$D3,$65,$8F,$91,$E8
	.dc.b	$FD,$4F,$8F,$FA,$AD,$9F,$B2,$D9,$DC,$7D,$C9,$DB,$4D,$96,$3E,$94
	.dc.b	$7E,$2C,$FC,$67,$C4,$D9,$FC,$63,$67,$EC,$37,$8B,$FC,$69,$CD,$FE
	.dc.b	$80,$8E,$35,$F9,$3C,$2D,$FD,$FF,$8E,$FC,$C1,$F8,$EF,$22,$D7,$E4
	.dc.b	$BC,$6C,$34,$FC,$AC,$7E,$53,$CA,$91,$F9,$F8,$B6,$CF,$FC,$8D,$F4
	.dc.b	$DF,$E4,$6E,$0F,$DC,$3F,$E4,$8E,$7B,$E3,$FF,$B0,$71,$61,$C5,$B5
	.dc.b	$B7,$38,$74,$AE,$15,$B1,$5B,$70,$F8,$6B,$89,$D2,$7F,$6B,$3A,$79
	.dc.b	$D7,$5B,$F2,$E1,$E6,$7F,$7D,$9F,$EF,$F3,$B8,$FE,$FD,$B8,$FE,$AD
	.dc.b	$B8,$7D,$4E,$4F,$F0,$65,$7F,$05,$F0,$7F,$84,$78,$3F,$C2,$5B,$19
	.dc.b	$3F,$85,$46,$5A,$7F,$0D,$ED,$FE,$21,$DB,$FC,$45,$B1,$93,$F8,$94
	.dc.b	$65,$A6,$72,$FE,$01,$19,$6C,$7C,$67,$E9,$FD,$23,$8F,$F9,$87,$8C
	.dc.b	$FA,$47,$8B,$FC,$CE,$D7,$BC,$39,$5E,$EB,$53,$9A,$FE,$DC,$D7,$FC
	.dc.b	$B4,$D7,$FD,$28,$D7,$FC,$2B,$AF,$FA,$D6,$FF,$97,$78,$3F,$CC,$5F
	.dc.b	$E6,$4D,$72,$7F,$33,$32,$7B,$79,$7F,$D3,$3D,$1F,$92,$D3,$F2,$6D
	.dc.b	$9D,$73,$23,$18,$FE,$1D,$BC,$5F,$93,$6D,$7E,$35,$B3,$FE,$2C,$FD
	.dc.b	$15,$ED,$7F,$57,$5F,$79,$63,$0D,$BF,$A5,$D8,$AD,$C6,$1E,$6D,$9F
	.dc.b	$A5,$F4,$B8,$C3,$F4,$99,$3E,$EB,$FA,$AC,$7E,$9C,$4E,$E7,$FB,$84
	.dc.b	$E9,$FF,$5F,$4E,$B2,$AF,$FB,$1E,$BB,$3F,$E9,$16,$B9,$8D,$BF,$B4
	.dc.b	$D3,$EC,$3F,$AF,$DF,$F8,$0E,$5F,$C0,$9A,$FD,$06,$C6,$E3,$E9,$9B
	.dc.b	$29,$B3,$F3,$0B,$EF,$DF,$23,$FE,$07,$F5,$DF,$81,$3E,$1B,$F3,$15
	.dc.b	$AC,$3F,$D0,$93,$97,$9C,$67,$0F,$F2,$23,$5F,$F5,$83,$5F,$F7,$53
	.dc.b	$5F,$E3,$7F,$8D,$BE,$B7,$FC,$31,$F8,$53,$F3,$E7,$D7,$FE,$A6,$B5
	.dc.b	$8F,$EC,$6B,$53,$9B,$F9,$78,$2F,$6F,$F9,$5D,$AF,$FB,$0D,$AF,$FA
	.dc.b	$EB,$EA,$1F,$65,$E2,$FC,$FB,$F9,$7E,$5F,$8E,$9C,$BE,$18,$CD,$7F
	.dc.b	$2D,$E9,$56,$A3,$F9,$86,$6A,$DF,$37,$F5,$2A,$D2,$C3,$F9,$ED,$6E
	.dc.b	$FF,$C7,$15,$F9,$19,$F3,$72,$E3,$A3,$2C,$8F,$33,$96,$9F,$A6,$FC
	.dc.b	$0B,$F4,$B7,$E7,$7C,$31,$FE,$DA,$73,$8F,$F6,$67,$33,$FE,$A0,$73
	.dc.b	$3B,$D3,$FC,$99,$18,$9D,$46,$3E,$2B,$38,$5B,$F1,$36,$BC,$4D,$C3
	.dc.b	$C0,$D9,$F2,$36,$7C,$4D,$D5,$8F,$C5,$9F,$36,$1E,$0B,$F8,$1F,$07
	.dc.b	$E6,$FA,$CF,$3B,$63,$27,$C5,$8C,$B4,$F9,$1F,$67,$EE,$F0,$BD,$ED
	.dc.b	$8C,$9F,$36,$32,$D3,$38,$7C,$F8,$CB,$63,$F8,$CF,$D5,$3E,$67,$FC
	.dc.b	$EC,$BE,$72,$C6,$1B,$3C,$1B,$5F,$91,$AE,$13,$D2,$B8,$34,$E0,$59
	.dc.b	$E1,$5C,$B7,$FF,$12,$DF,$29,$FC,$E7,$D6,$D9,$FF,$14,$AF,$C0,$3F
	.dc.b	$D1,$8E,$70,$FF,$0A,$F0,$23,$51,$FC,$D9,$D5,$FF,$BA,$9C,$BD,$B1
	.dc.b	$E7,$7D,$1B,$5C,$5B,$3F,$CD,$4F,$F1,$1F,$C1,$7F,$89,$3E,$B9,$F5
	.dc.b	$DF,$E2,$69,$FF,$13,$F9,$BF,$AD,$53,$A7,$42,$B9,$56,$2B,$85,$6E
	.dc.b	$BA,$56,$9E,$D3,$AA,$FF,$72,$4E,$67,$C3,$85,$7C,$E3,$E2,$1F,$35
	.dc.b	$C2,$30,$D8,$CC,$79,$AB,$FC,$83,$18,$8F,$F7,$8F,$C3,$7F,$92,$DE
	.dc.b	$27,$F9,$1B,$E2,$3F,$DA,$5C,$1C,$30,$F8,$91,$89,$D9,$CA,$74,$F1
	.dc.b	$AE,$9A,$7C,$1A,$FF,$70,$4F,$B5,$AF,$CC,$CF,$AD,$70,$6D,$EF,$E1
	.dc.b	$BE,$A6,$1C,$DB,$3A,$BE,$47,$E1,$7E,$C3,$ED,$5F,$E4,$3E,$BE,$ED
	.dc.b	$F8,$EE,$24,$62,$F8,$1F,$8F,$BC,$71,$FE,$4E,$6D,$70,$AD,$10,$FC
	.dc.b	$FC,$68,$DD,$72,$FA,$D7,$C6,$DF,$1A,$5F,$F8,$4A,$72,$D9,$CB,$FD
	.dc.b	$16,$3F,$AD,$B9,$21,$99,$DE,$7F,$E9,$69,$E7,$3C,$6B,$39,$FE,$BA
	.dc.b	$79,$CE,$4F,$E7,$DC,$BF,$03,$BB,$F6,$D1,$FD,$07,$9F,$13,$23,$04
	.dc.b	$3F,$A2,$39,$7F,$A3,$47,$33,$18,$FF,$45,$FC,$13,$B7,$54,$AF,$E9
	.dc.b	$D1,$E1,$30,$FF,$4F,$8C,$3F,$D4,$FE,$5B,$CD,$F3,$52,$30,$43,$FA
	.dc.b	$B3,$97,$FA,$C4,$7B,$4C,$5F,$F5,$79,$C4,$F1,$8C,$3F,$D8,$63,$2F
	.dc.b	$F6,$1E,$3F,$8D,$F8,$1C,$8F,$EC,$D1,$F2,$4C,$3F,$D9,$E3,$0F,$F8
	.dc.b	$0F,$C0,$FC,$6F,$8C,$69,$8F,$0E,$CE,$C8,$7F,$81,$E5,$1C,$BC,$3F
	.dc.b	$CD,$78,$EF,$79,$F1,$98,$6E,$1F,$E1,$3A,$E0,$FF,$88,$E3,$2F,$F8
	.dc.b	$41,$BA,$FF,$83,$E7,$F4,$D2,$FC,$F9,$F2,$5F,$C8,$A3,$F9,$44,$61
	.dc.b	$C1,$B8,$AD,$33,$1B,$7F,$CD,$1E,$03,$E0,$FB,$B6,$3F,$CA,$33,$BC
	.dc.b	$3F,$CD,$3F,$02,$F9,$6E,$AD,$AF,$0B,$5F,$33,$5F,$E7,$2E,$67,$FC
	.dc.b	$E7,$F1,$1F,$A8,$6F,$6F,$C0,$9D,$76,$FF,$45,$FD,$1B,$E7,$61,$FE
	.dc.b	$64,$5F,$0F,$F1,$9B,$19,$E0,$C1,$99,$91,$92,$1F,$E8,$99,$E2,$D8
	.dc.b	$E2,$FF,$9C,$BE,$AB,$FD,$50,$FC,$E6,$BF,$4D,$AF,$F5,$66,$BF,$D7
	.dc.b	$5D,$E5,$FE,$92,$FA,$B7,$D2,$FD,$E6,$D7,$7F,$83,$FE,$47,$FA,$3F
	.dc.b	$F9,$BE,$3F,$D4,$C7,$F2,$6F,$84,$FE,$C7,$87,$F6,$7F,$29,$5B,$AF
	.dc.b	$F2,$F3,$67,$55,$C3,$6F,$F0,$0B,$69,$7C,$BD,$AF,$27,$DC,$F9,$79
	.dc.b	$25,$7F,$A5,$34,$4E,$A3,$13,$A7,$FD,$69,$3A,$78,$4E,$72,$F8,$55
	.dc.b	$C6,$7F,$5D,$EB,$7E,$DB,$D6,$91,$FE,$7B,$8F,$DC,$98,$7F,$69,$18
	.dc.b	$7F,$D4,$FF,$53,$FB,$67,$F4,$6D,$7F,$50,$D7,$FC,$16,$D7,$FC,$24
	.dc.b	$EE,$FF,$D8,$FE,$23,$EA,$3F,$66,$DA,$EF,$E0,$FF,$0C,$F2,$BF,$86
	.dc.b	$F9,$52,$7F,$D2,$51,$97,$FD,$29,$19,$7F,$86,$46,$5F,$E1,$D1,$9C
	.dc.b	$3F,$D2,$13,$FC,$63,$CA,$FE,$33,$E5,$4B,$FF,$4F,$46,$5F,$F5,$04
	.dc.b	$65,$FE,$31,$1F,$C8,$3C,$AF,$E4,$39,$49,$FF,$53,$46,$5F,$F5,$44
	.dc.b	$65,$FE,$41,$1F,$02,$BE,$0D,$C9,$FF,$6A,$D7,$F9,$8D,$B3,$97,$FD
	.dc.b	$E9,$1F,$E6,$57,$24,$33,$3B,$C3,$E7,$D7,$E8,$3E,$0E,$50,$FC,$DB
	.dc.b	$FC,$EB,$CA,$FD,$2F,$D4,$FE,$9B,$E9,$25,$6F,$0F,$D2,$75,$2B,$FA
	.dc.b	$0F,$9B,$2B,$FD,$8C,$F0,$7F,$9F,$7D,$E7,$F4,$62,$3F,$EC,$BE,$2A
	.dc.b	$FF,$43,$F4,$FF,$A5,$98,$C7,$FD,$8C,$E2,$3F,$DA,$4E,$7B,$7F,$CD
	.dc.b	$F9,$3E,$AF,$F7,$BE,$B4,$BF,$45,$6A,$7F,$80,$99,$8F,$DD,$CE,$1D
	.dc.b	$9F,$B8,$AE,$8E,$3B,$BF,$81,$CF,$1C,$7E,$1D,$FF,$62,$F7,$3F,$B1
	.dc.b	$FA,$D3,$0F,$F7,$44,$65,$FF,$75,$46,$5F,$EC,$51,$97,$FB,$24,$66
	.dc.b	$BF,$DC,$D3,$FC,$63,$C3,$FE,$33,$F1,$F2,$FC,$99,$FF,$1E,$7F,$8A
	.dc.b	$CE,$5F,$E2,$F3,$C5,$DC,$66,$FE,$DC,$E6,$7F,$DF,$5F,$69,$FC,$77
	.dc.b	$5F,$F4,$DF,$A9,$FC,$6F,$6F,$F7,$ED,$7F,$93,$2B,$4F,$F1,$E6,$C6
	.dc.b	$A3,$E4,$4F,$19,$C9,$A7,$A1,$9F,$13,$FC,$AF,$E0,$FF,$78,$C3,$FD
	.dc.b	$56,$BE,$73,$26,$BD,$49,$8F,$D5,$9D,$36,$72,$B4,$FE,$79,$3D,$63
	.dc.b	$F9,$AB,$C1,$D5,$F8,$2B,$8D,$7F,$43,$CB,$FB,$64,$FF,$AF,$9F,$F2
	.dc.b	$8C,$67,$D9,$F8,$9E,$7C,$C3,$F1,$15,$AB,$E1,$59,$9E,$E7,$F4,$6F
	.dc.b	$61,$D1,$FE,$59,$24,$7E,$46,$3F,$93,$3C,$37,$FF,$80,$33,$A8,$C4
	.dc.b	$F2,$78,$74,$7F,$98,$F9,$7F,$CC,$9D,$12,$7A,$46,$5E,$D1,$97,$9C
	.dc.b	$7F,$A2,$BC,$47,$CA,$7F,$B0,$B9,$7F,$DB,$F1,$8A,$FF,$69,$C6,$27
	.dc.b	$51,$FD,$8A,$7F,$C1,$EF,$FB,$3E,$74,$F0,$35,$3C,$33,$FF,$22,$B6
	.dc.b	$B8,$FC,$02,$F8,$DF,$33,$89,$CE,$F8,$B6,$B8,$CF,$FB,$DB,$3F,$F6
	.dc.b	$63,$7A,$7C,$10,$6D,$43,$A4,$65,$B1,$97,$C4,$3E,$83,$E3,$FC,$0F
	.dc.b	$F9,$29,$B1,$FE,$4F,$6B,$FE,$F2,$6B,$FE,$D6,$F6,$5F,$1D,$FE,$52
	.dc.b	$F8,$1F,$F2,$BB,$63,$33,$FE,$59,$9C,$BD,$7D,$7F,$33,$D5,$AF,$F4
	.dc.b	$13,$CC,$E1,$FE,$89,$68,$C6,$AF,$9F,$FC,$0F,$37,$CD,$7F,$CE,$CD
	.dc.b	$E1,$39,$F6,$2E,$3F,$56,$F9,$56,$38,$BE,$D4,$FF,$89,$B0,$E6,$ED
	.dc.b	$EE,$DB,$E1,$C3,$FD,$67,$07,$D0,$3C,$4C,$DF,$F8,$C2,$31,$9F,$E0
	.dc.b	$D7,$B9,$AB,$E3,$76,$77,$7E,$76,$E7,$E6,$AE,$F7,$BB,$98,$75,$79
	.dc.b	$46,$30,$F5,$7C,$BF,$C2,$EE,$FF,$26,$6B,$F3,$70,$FE,$98,$A7,$51
	.dc.b	$ED,$8D,$57,$EF,$DD,$47,$0D,$F7,$3F,$26,$AE,$5F,$5E,$6D,$FD,$4B
	.dc.b	$EE,$63,$E4,$72,$C3,$27,$91,$C6,$1F,$C3,$FD,$47,$CC,$76,$5F,$41
	.dc.b	$63,$0D,$BF,$ED,$13,$A6,$CE,$9F,$ED,$9F,$41,$FE,$BE,$6F,$89,$EA
	.dc.b	$8C,$61,$FC,$7D,$FE,$85,$3C,$3C,$6D,$56,$DB,$19,$39,$96,$B9,$4F
	.dc.b	$A7,$E8,$3D,$5F,$39,$6B,$CE,$DA,$F8,$0D,$E1,$FF,$42,$66,$8C,$57
	.dc.b	$E4,$E3,$0F,$C0,$BE,$AF,$B9,$F6,$4F,$B5,$B1,$93,$E1,$C6,$5A,$66
	.dc.b	$FE,$14,$65,$B1,$F3,$BE,$83,$EA,$7D,$07,$D5,$6C,$64,$FA,$71,$96
	.dc.b	$99,$9F,$BD,$19,$6C,$7D,$6F,$9B,$FA,$9F,$41,$F5,$5B,$19,$3E,$DC
	.dc.b	$65,$A6,$67,$ED,$46,$5B,$1F,$86,$8F,$C6,$1F,$8D,$F8,$2B,$1F,$8A
	.dc.b	$6C,$FE,$4D,$BA,$7D,$47,$84,$FE,$42,$ED,$7E,$67,$E8,$3F,$30,$7E
	.dc.b	$37,$E1,$ED,$7E,$4B,$E1,$B0,$D3,$F2,$B1,$F9,$3F,$89,$91,$F9,$C8
	.dc.b	$B7,$FD,$EB,$E7,$BF,$9B,$EF,$FB,$47,$F9,$CF,$2D,$F1,$BF,$0B,$5B
	.dc.b	$7F,$57,$87,$F9,$42,$B8,$B6,$33,$3B,$BE,$57,$CD,$B5,$B6,$CF,$ED
	.dc.b	$F2,$FD,$B6,$5F,$B7,$6D,$7E,$D1,$B3,$B8,$FD,$BC,$FE,$F7,$C3,$FD
	.dc.b	$F7,$87,$67,$F7,$6D,$AF,$E0,$8D,$E0,$FD,$2B,$FC,$17,$BB,$F8,$37
	.dc.b	$6D,$BF,$E0,$6D,$BF,$E1,$8D,$BF,$D6,$B6,$7F,$58,$D9,$FF,$75,$F3
	.dc.b	$FF,$24,$AF,$E9,$B0,$F9,$F7,$1E,$23,$FA,$F4,$63,$0F,$EC,$11,$89
	.dc.b	$D3,$FE,$A3,$9D,$3A,$3A,$47,$F2,$E6,$CF,$FA,$1B,$E7,$5F,$74,$FA
	.dc.b	$ED,$61,$CB,$1F,$E0,$F7,$FC,$9F,$C0,$FE,$50,$D3,$F9,$29,$71,$FE
	.dc.b	$5C,$D9,$FE,$D3,$CB,$FE,$02,$E8,$FC,$56,$9F,$8B,$6D,$7E,$1D,$B7
	.dc.b	$F8,$D6,$DF,$2E,$64,$62,$BF,$D8,$58,$75,$7F,$21,$87,$F4,$AF,$8B
	.dc.b	$7C,$BD,$F9,$9C,$36,$BF,$D4,$93,$A6,$CE,$9A,$7F,$AB,$33,$7D,$06
	.dc.b	$DF,$9C,$D7,$EA,$B5,$FF,$0B,$35,$FF,$6C,$3C,$74,$FC,$0B,$67,$FD
	.dc.b	$B9,$C3,$F5,$3D,$DF,$B1,$39,$B6,$9F,$E2,$CB,$FD,$A6,$FF,$B5,$69
	.dc.b	$FB,$12,$CF,$ED,$DB,$8F,$CB,$69,$F2,$B7,$FE,$DD,$3F,$BD,$AF,$C4
	.dc.b	$7C,$F3,$E1,$7F,$1F,$28,$C5,$FE,$3E,$31,$3A,$9C,$C7,$E4,$DB,$87
	.dc.b	$E8,$1B,$5F,$EF,$BF,$C2,$5F,$84,$BB,$3C,$38,$FE,$67,$E7,$DF,$03
	.dc.b	$E2,$6E,$1E,$16,$E7,$FA,$16,$BF,$A8,$73,$9F,$EC,$1B,$29,$FF,$34
	.dc.b	$7D,$E3,$5F,$06,$1F,$E1,$4F,$83,$7C,$FF,$F8,$63,$E6,$5F,$41,$C7
	.dc.b	$2F,$F6,$3C,$E6,$31,$97,$FA,$7E,$F8,$D7,$FB,$DA,$F8,$CF,$5B,$E0
	.dc.b	$F7,$E9,$95,$FE,$D2,$E8,$B3,$CE,$6E,$DF,$E5,$6F,$49,$92,$F3,$87
	.dc.b	$FA,$1D,$FF,$16,$FB,$BE,$2C,$7F,$D5,$F1,$8C,$B2,$DA,$FF,$5C,$36
	.dc.b	$7F,$D6,$BE,$F7,$2C,$F9,$B6,$BF,$D7,$CD,$9D,$C7,$FB,$06,$76,$D3
	.dc.b	$65,$8D,$9F,$EC,$28,$DB,$63,$FD,$95,$F3,$5E,$AF,$9B,$F5,$B6,$32
	.dc.b	$7F,$B3,$E3,$2D,$3F,$DA,$5F,$11,$F0,$BC,$5F,$86,$D8,$C9,$FE,$D7
	.dc.b	$8C,$B4,$CD,$FF,$B5,$A3,$2D,$8F,$C5,$6F,$F4,$FE,$81,$F2,$1F,$E5
	.dc.b	$C5,$F2,$36,$30,$DA,$E2,$7F,$A1,$67,$6E,$2D,$C4,$9E,$B3,$98,$C5
	.dc.b	$73,$AF,$F7,$75,$72,$6D,$F8,$2F,$84,$F2,$9D,$D7,$1C,$3D,$4F,$F9
	.dc.b	$D3,$E7,$B9,$47,$E3,$D7,$CB,$66,$74,$DC,$3C,$93,$A6,$CE,$9A,$6A
	.dc.b	$78,$5A,$FB,$47,$F5,$98,$FD,$B7,$BF,$FB,$7F,$9F,$7C,$8F,$F6,$4F
	.dc.b	$88,$FF,$5B,$C6,$2B,$FB,$3D,$6E,$B5,$8D,$9C,$8A,$E0,$FF,$03,$BE
	.dc.b	$75,$FD,$AE,$F8,$4E,$6E,$BE,$41,$F1,$0F,$A0,$E1,$18,$6C,$66,$3C
	.dc.b	$2D,$5F,$11,$FE,$BE,$D3,$F8,$A4,$E6,$31,$3C,$CF,$E3,$0E,$23,$5F
	.dc.b	$09,$F2,$67,$47,$47,$1A,$7B,$AB,$FD,$A1,$3E,$C6,$BF,$0F,$3F,$F7
	.dc.b	$74,$78,$27,$85,$72,$9C,$FC,$2F,$FB,$CE,$7A,$36,$38,$CE,$EF,$F3
	.dc.b	$3C,$0F,$8A,$FF,$7D,$67,$F8,$BF,$8B,$BB,$7F,$B9,$69,$18,$BE,$37
	.dc.b	$79,$55,$CE,$7F,$1F,$59,$8E,$31,$F9,$2A,$C1,$99,$91,$93,$F5,$5F
	.dc.b	$40,$F9,$1E,$CB,$E4,$29,$88,$93,$C2,$32,$F2,$8C,$BF,$9B,$CB,$FE
	.dc.b	$00,$57,$38,$E0,$6C,$86,$CF,$DF,$BB,$71,$92,$31,$9F,$F4,$17,$D1
	.dc.b	$3F,$C0,$FC,$1F,$E1,$ED,$E6,$F9,$11,$FB,$88,$D7,$3E,$0C,$CC,$8C
	.dc.b	$90,$F0,$CF,$16,$C7,$17,$C1,$F3,$3D,$BA,$A4,$60,$87,$27,$2F,$48
	.dc.b	$F2,$FC,$CF,$9B,$E6,$64,$FA,$E3,$2F,$B6,$32,$F9,$23,$2F,$9A,$33
	.dc.b	$9F,$F6,$5C,$55,$A8,$C4,$EB,$5F,$F0,$16,$AB,$DE,$1F,$2F,$93,$E4
	.dc.b	$FC,$CC,$8C,$10,$F8,$AE,$5F,$8F,$1F,$97,$F1,$9E,$0D,$24,$FF,$39
	.dc.b	$87,$89,$88,$FF,$04,$4D,$D3,$FB,$F5,$6A,$78,$78,$7F,$A5,$F1,$DE
	.dc.b	$FB,$E2,$F0,$DA,$D9,$0F,$C7,$4F,$09,$D9,$E7,$AF,$D8,$4B,$E8,$1F
	.dc.b	$27,$86,$C7,$F9,$95,$AF,$E6,$47,$77,$91,$D4,$FF,$91,$DE,$8D,$9E
	.dc.b	$31,$FE,$9E,$6B,$FE,$35,$1F,$02,$F9,$CF,$2B,$63,$FC,$EC,$D7,$C5
	.dc.b	$F1,$0F,$9D,$DD,$78,$6B,$55,$FE,$AC,$6B,$DD,$AF,$F9,$17,$E8,$3D
	.dc.b	$4F,$B5,$7C,$47,$C6,$6C,$7F,$A1,$DD,$F0,$FF,$AF,$9B,$D1,$EB,$77
	.dc.b	$3C,$EB,$97,$0C,$8E,$51,$D4,$C3,$F3,$7E,$83,$FD,$46,$FF,$A5,$5A
	.dc.b	$E7,$E1,$3E,$CF,$D0,$7F,$54,$8D,$47,$D1,$9D,$C7,$FB,$31,$AF,$CF
	.dc.b	$6B,$FE,$10,$F9,$CF,$F3,$8B,$FE,$A8,$3F,$99,$7C,$2F,$FB,$53,$4F
	.dc.b	$F2,$94,$EA,$7F,$8B,$B6,$3F,$CA,$D3,$BD,$3F,$5E,$DA,$ED,$EC,$7A
	.dc.b	$38,$3D,$3C,$12,$7F,$D6,$5A,$7D,$B6,$CE,$5F,$C8,$47,$E0,$71,$FF
	.dc.b	$76,$BB,$8C,$CE,$E7,$35,$DE,$7F,$65,$EA,$7E,$DF,$D4,$91,$82,$1F
	.dc.b	$E7,$B7,$2F,$F9,$F6,$3F,$4F,$F4,$1F,$B5,$8F,$D1,$36,$FF,$DF,$ED
	.dc.b	$D3,$F3,$9E,$17,$F5,$6A,$DF,$77,$E9,$6B,$94,$7F,$C0,$06,$BF,$A3
	.dc.b	$6B,$FE,$44,$ED,$FE,$7E,$FF,$A3,$FC,$6F,$E2,$07,$F1,$1F,$1E,$47
	.dc.b	$F0,$F8,$CB,$FC,$46,$32,$F1,$34,$64,$FE,$2D,$E3,$7F,$1B,$F1,$E4
	.dc.b	$60,$87,$FA,$7D,$CB,$FE,$A1,$8F,$E4,$5E,$3F,$F2,$3C,$E4,$FF,$21
	.dc.b	$8C,$BF,$C8,$E3,$2F,$FA,$A6,$3F,$89,$C7,$F9,$B2,$3F,$DE,$8F,$C0
	.dc.b	$D5,$AF,$CC,$C7,$78,$CB,$DC,$72,$30,$FF,$17,$BE,$8D,$8D,$3D,$A7
	.dc.b	$FC,$21,$CB,$FE,$E3,$8F,$F3,$83,$5F,$CF,$FB,$BF,$CF,$FC,$6D,$FC
	.dc.b	$FF,$0F,$A0,$95,$C2,$3F,$A2,$79,$12,$3F,$A1,$3A,$7F,$D9,$1F,$3D
	.dc.b	$FD,$24,$8F,$F4,$7F,$9E,$AF,$FB,$37,$D0,$FE,$9C,$63,$83,$FA,$3B
	.dc.b	$88,$FE,$9C,$E7,$83,$FA,$FE,$AF,$A0,$F6,$FA,$92,$FD,$B1,$F0,$0C
	.dc.b	$3F,$BC,$8C,$3A,$3E,$1B,$AC,$3F,$81,$CF,$F6,$1F,$6B,$FB,$17,$A9
	.dc.b	$2F,$FB,$04,$65,$FE,$C5,$19,$7F,$DD,$11,$97,$FD,$D5,$19,$AF,$EC
	.dc.b	$73,$F5,$3E,$2B,$EA,$FC,$F4,$BF,$AB,$19,$7E,$BC,$65,$FE,$2B,$19
	.dc.b	$7F,$8B,$C6,$6B,$F8,$DC,$FF,$93,$FE,$83,$F8,$D6,$9F,$E0,$6E,$BF
	.dc.b	$E3,$1E,$A7,$FB,$C3,$1D,$78,$91,$FE,$F8,$7F,$80,$4C,$1F,$E3,$AB
	.dc.b	$E6,$FE,$E4,$D1,$FE,$BB,$73,$2F,$03,$FB,$D4,$FF,$90,$97,$CD,$FF
	.dc.b	$98,$DB,$1F,$EC,$37,$79,$FE,$92,$3F,$D9,$31,$8C,$BF,$C3,$EF,$08
	.dc.b	$E9,$5C,$FD,$29,$1C,$E3,$B1,$87,$D5,$EB,$FC,$5F,$62,$47,$A6,$3E
	.dc.b	$41,$87,$E0,$46,$1F,$F2,$B7,$3F,$F9,$50,$8F,$F9,$67,$9E,$BF,$E5
	.dc.b	$08,$B9,$FF,$8E,$E7,$FC,$B7,$C7,$FE,$5C,$E7,$97,$FE,$61,$8C,$BF
	.dc.b	$E6,$38,$CB,$FE,$5A,$8F,$F4,$57,$88,$F9,$4F,$F0,$77,$BE,$E6,$FF
	.dc.b	$21,$CE,$6F,$FD,$A3,$18,$AD,$BA,$F3,$EC,$FF,$BD,$AB,$82,$BE,$8E
	.dc.b	$6F,$7C,$7F,$E4,$06,$D7,$F8,$3A,$B7,$39,$AF,$09,$C4,$AE,$5E,$D8
	.dc.b	$78,$DC,$FB,$3F,$EE,$28,$FF,$2E,$AE,$9F,$0B,$67,$66,$0C,$C4,$72
	.dc.b	$78,$87,$D0,$7C,$9F,$7F,$FC,$6D,$1F,$E3,$78,$C3,$83,$79,$FF,$84
	.dc.b	$DD,$C6,$9E,$3E,$FF,$F9,$3B,$E3,$78,$47,$FB,$93,$D9,$46,$27,$53
	.dc.b	$D5,$AF,$4F,$5F,$D8,$DC,$B9,$E1,$FE,$4A,$6C,$EA,$7F,$D0,$3C,$FF
	.dc.b	$EB,$FE,$BF,$9A,$D8,$F9,$A3,$7F,$F6,$37,$6F,$EB,$78,$7E,$BD,$6D
	.dc.b	$BD,$39,$E2,$EA,$DC,$39,$52,$F1,$87,$FB,$46,$AF,$D5,$2E,$E9,$86
	.dc.b	$E7,$F8,$29,$D3,$67,$4D,$35,$3C,$0F,$5D,$AB,$8C,$61,$FC,$34,$FC
	.dc.b	$2F,$47,$F8,$57,$57,$F0,$BB,$FF,$73,$46,$27,$E4,$E7,$70,$F6,$CD
	.dc.b	$AD,$D7,$AE,$F9,$5F,$F9,$EE,$30,$72,$2D,$F2,$6D,$7F,$BD,$7E,$DF
	.dc.b	$A5,$1B,$E9,$E1,$95,$C2,$74,$D9,$CD,$FF,$33,$F2,$7C,$DE,$11,$BE
	.dc.b	$4E,$19,$7F,$CD,$5C,$B4,$F2,$FE,$70,$F9,$AF,$80,$D8,$CF,$D0,$E2
	.dc.b	$64,$70,$77,$18,$C7,$FD,$FF,$5C,$5B,$3A,$AE,$07,$F4,$7F,$7F,$D2
	.dc.b	$7A,$BA,$EC,$64,$95,$FD,$46,$32,$FF,$53,$9B,$C1,$F7,$74,$FF,$5F
	.dc.b	$DF,$08,$FC,$06,$2F,$9F,$F6,$9E,$EF,$9F,$B5,$F0,$A2,$E3,$EA,$9B
	.dc.b	$7C,$A6,$D7,$FB,$C2,$D5,$8B,$FD,$76,$1F,$E1,$5A,$FE,$C9,$E6,$7D
	.dc.b	$6F,$37,$EB,$B2,$32,$7F,$69,$8C,$57,$F6,$F9,$B9,$7B,$66,$D7,$87
	.dc.b	$E7,$BE,$B9,$F6,$3D,$EB,$1F,$72,$2E,$1F,$52,$6D,$FC,$49,$B5,$EB
	.dc.b	$9B,$5E,$AF,$A6,$FC,$6F,$CE,$7E,$39,$B5,$FD,$FD,$C3,$FE,$01,$F4
	.dc.b	$6D,$7F,$81,$DF,$CA,$9F,$96,$EF,$B3,$F9,$0F,$81,$FC,$CF,$C5,$FE
	.dc.b	$41,$B1,$F9,$4F,$2F,$F2,$67,$E5,$3C,$4B,$1F,$98,$F2,$3F,$2E,$7E
	.dc.b	$8B,$C9,$B1,$FB,$2F,$98,$FE,$63,$CB,$FB,$D6,$C6,$7C,$A7,$D0,$E5
	.dc.b	$FB,$71,$9A,$C1,$99,$B1,$92,$6E,$E4,$FF,$0C,$61,$FD,$02,$B7,$39
	.dc.b	$AF,$E4,$9E,$1B,$F4,$87,$A9,$FC,$3D,$FF,$12,$CF,$F0,$87,$2D,$F5
	.dc.b	$35,$3B,$6C,$64,$FE,$23,$6F,$9E,$7C,$F7,$F1,$DF,$2D,$F5,$3F,$E0
	.dc.b	$7B,$DD,$69,$FE,$33,$3A,$E7,$FE,$49,$87,$3C,$3F,$C6,$8F,$F6,$8B
	.dc.b	$E0,$FF,$67,$6B,$98,$DB,$F0,$33,$FE,$09,$BF,$F3,$35,$E2,$FC,$1F
	.dc.b	$17,$E1,$37,$3F,$84,$DB,$F9,$47,$F5,$BA,$E3,$3F,$1E,$B8,$B4,$E2
	.dc.b	$7F,$34,$F9,$F7,$E8,$9F,$CB,$2F,$FC,$B9,$E6,$7F,$41,$9C,$BF,$CB
	.dc.b	$0C,$4E,$F2,$FF,$2A,$4E,$AF,$EB,$7D,$1E,$AF,$ED,$73,$7D,$BE,$5B
	.dc.b	$5F,$6D,$B9,$7A,$B9,$7F,$05,$18,$CB,$E2,$3F,$93,$F8,$D9,$7F,$97
	.dc.b	$F0,$EE,$1F,$0B,$A3,$FC,$1B,$58,$C3,$F2,$3D,$2B,$C6,$1F,$DC,$EF
	.dc.b	$6F,$E8,$6F,$F4,$13,$AC,$3F,$BC,$7D,$0B,$E5,$FF,$B2,$64,$FA,$AF
	.dc.b	$9F,$5F,$48,$56,$F1,$FE,$C3,$5B,$E5,$FC,$4C,$E6,$BF,$C5,$8D,$7F
	.dc.b	$C6,$2D,$7F,$C1,$07,$4C,$3F,$12,$F9,$3E,$1B,$F3,$5D,$6E,$EE,$B1
	.dc.b	$86,$E5,$FA,$2C,$13,$FA,$1F,$9D,$FD,$6E,$1E,$99,$CD,$7F,$07,$87
	.dc.b	$C9,$FF,$07,$6C,$67,$E5,$71,$32,$38,$3B,$8C,$4F,$D5,$38,$96,$35
	.dc.b	$3C,$0F,$DB,$F8,$FF,$BE,$9F,$CE,$F8,$0C,$7F,$AC,$56,$BC,$2F,$57
	.dc.b	$9F,$FB,$4C,$38,$70,$FE,$D1,$B5,$FE,$31,$C3,$97,$D2,$EF,$0F,$DD
	.dc.b	$FE,$89,$D3,$3E,$FF,$A3,$79,$B2,$FD,$EC,$E5,$FF,$59,$60,$9D,$63
	.dc.b	$F5,$AB,$4D,$CF,$F8,$74,$BE,$4B,$BC,$7F,$1D,$CE,$57,$F1,$D8,$B9
	.dc.b	$7E,$8F,$C3,$56,$33,$FE,$0D,$56,$FF,$85,$79,$FF,$C9,$CF,$E5,$1E
	.dc.b	$2D,$8F,$E5,$11,$71,$E9,$36,$FF,$C7,$3F,$7D,$FE,$32,$3F,$C6,$7F
	.dc.b	$7D,$70,$FF,$18,$45,$BF,$DB,$7B,$6A,$C5,$FF,$90,$BD,$AF,$F1,$D9
	.dc.b	$FE,$3C,$F8,$EB,$8F,$F8,$E6,$2E,$1C,$A6,$D7,$1A,$D1,$FE,$3F,$8C
	.dc.b	$BF,$E7,$CA,$B5,$FE,$70,$9B,$E2,$31,$FF,$0F,$CF,$9B,$DA,$FF,$26
	.dc.b	$FB,$7F,$E4,$E6,$63,$FE,$4B,$72,$D3,$FC,$9D,$EB,$FF,$95,$0F,$F2
	.dc.b	$AF,$B7,$63,$24,$BF,$87,$19,$7E,$2C,$DD,$3F,$CC,$B3,$6B,$FA,$FF
	.dc.b	$E4,$0A,$FF,$36,$36,$3F,$CC,$AA,$73,$32,$72,$F0,$8C,$5F,$F2,$9A
	.dc.b	$E4,$D9,$D5,$70,$24,$65,$EF,$5C,$AF,$FD,$09,$6F,$67,$CA,$79,$BD
	.dc.b	$AB,$7E,$68,$B8,$78,$66,$DF,$29,$B5,$F0,$26,$D7,$FB,$3B,$E0,$BF
	.dc.b	$D4,$9F,$3C,$F8,$FF,$F6,$8F,$6B,$0C,$C7,$FB,$46,$AD,$FF,$24,$AD
	.dc.b	$2B,$E4,$F5,$31,$5D,$2F,$8C,$7F,$2C,$BA,$FF,$A9,$8D,$9C,$7C,$57
	.dc.b	$B9,$E1,$8F,$E5,$CD,$8C,$C6,$E3,$F9,$BF,$C3,$7D,$9D,$7F,$D8,$15
	.dc.b	$AA,$4E,$5E,$67,$FB,$15,$C4,$6D,$B5,$D3,$7F,$F1,$65,$79,$8E,$7C
	.dc.b	$1F,$D1,$B4,$F8,$31,$83,$DF,$7D,$E3,$2D,$D3,$CF,$5C,$5B,$5C,$2F
	.dc.b	$BD,$FC,$CE,$47,$0B,$D8,$FE,$AB,$AF,$E2,$1B,$BF,$E7,$9B,$8F,$FB
	.dc.b	$BE,$9E,$F6,$30,$E7,$56,$7A,$57,$E8,$A3,$15,$96,$46,$4F,$D3,$C7
	.dc.b	$1D,$A4,$7E,$A9,$E9,$E2,$62,$3F,$4D,$17,$2F,$CF,$CD,$CB,$F6,$B3
	.dc.b	$6B,$D1,$37,$2F,$83,$37,$24,$FA,$F9,$FF,$51,$7F,$ED,$7E,$5C,$19
	.dc.b	$E5,$B1,$92,$1F,$CB,$1C,$BF,$CB,$A2,$E7,$FE,$EE,$9E,$DA,$CA,$FF
	.dc.b	$72,$3E,$1D,$71,$1F,$CD,$A2,$D7,$FB,$DA,$7C,$D9,$CA,$FF,$76,$3E
	.dc.b	$9C,$B1,$1F,$CF,$A2,$D7,$FB,$FA,$6E,$BF,$E7,$8C,$FF,$39,$3F,$B2
	.dc.b	$D7,$F9,$95,$F3,$C3,$A9,$0E,$87,$FB,$EF,$0F,$C1,$7D,$8B,$DB,$A4
	.dc.b	$9F,$F8,$01,$CB,$5F,$F2,$34,$ED,$18,$9F,$A5,$3F,$8D,$F0,$9E,$8D
	.dc.b	$25,$7F,$A2,$FD,$1A,$FF,$6F,$8B,$A7,$F9,$CE,$79,$7B,$F2,$BF,$D2
	.dc.b	$4F,$5F,$A6,$C4,$7F,$77,$8B,$5F,$E7,$B9,$FD,$0C,$BE,$3B,$FC,$56
	.dc.b	$FF,$A5,$1A,$FF,$92,$5B,$3F,$E8,$CF,$D2,$7F,$C7,$93,$BF,$3B,$9C
	.dc.b	$7E,$42,$3F,$D3,$AD,$79,$47,$83,$F4,$B7,$CC,$52,$4F,$FA,$93,$F4
	.dc.b	$B7,$97,$88,$FF,$2D,$47,$FB,$09,$D7,$B7,$9C,$BF,$D3,$FF,$80,$3E
	.dc.b	$63,$DA,$DA,$FF,$57,$B5,$FF,$32,$3F,$E5,$07,$31,$C0,$8F,$0F,$5E
	.dc.b	$BC,$22,$D7,$82,$7E,$5F,$CF,$BE,$12,$92,$FF,$D7,$FF,$41,$FE,$7F
	.dc.b	$7F,$D8,$33,$B7,$3E,$8F,$D4,$9F,$F6,$1C,$6A,$BA,$FA,$5F,$E7,$08
	.dc.b	$96,$8C,$47,$F9,$D2,$30,$E0,$38,$76,$DC,$7F,$D8,$FF,$3F,$F5,$DB
	.dc.b	$DB,$E0,$AF,$F6,$BC,$FF,$80,$5B,$E4,$7E,$2A,$31,$1F,$E8,$F5,$E7
	.dc.b	$66,$64,$F0,$39,$C6,$1E,$11,$6F,$C7,$3F,$A7,$F8,$8F,$D4,$7C,$54
	.dc.b	$BF,$F7,$84,$65,$FF,$79,$46,$5F,$F2,$B4,$7F,$3E,$F8,$77,$E9,$9E
	.dc.b	$FF,$1F,$F6,$B1,$8D,$B5,$39,$BF,$F6,$C7,$C5,$E2,$7F,$A2,$7C,$67
	.dc.b	$F9,$36,$7F,$7B,$A7,$F0,$D8,$C4,$7F,$B8,$30,$E8,$EB,$1F,$F7,$67
	.dc.b	$D5,$7F,$01,$FB,$09,$39,$3F,$DC,$D1,$FC,$14,$CB,$FE,$6F,$8C,$3B
	.dc.b	$F2,$9E,$32,$B7,$97,$FB,$DE,$F7,$7F,$EC,$1F,$A2,$7C,$C7,$E0,$E1
	.dc.b	$18,$7F,$0C,$D7,$F1,$6D,$7F,$20,$E6,$BF,$C3,$B0,$FA,$07,$CE,$7F
	.dc.b	$0C,$6C,$7B,$4E,$F7,$FD,$B7,$E3,$9F,$55,$FC,$8F,$69,$5E,$D9,$CB
	.dc.b	$67,$2F,$FA,$46,$79,$47,$7F,$A1,$93,$DF,$04,$78,$FE,$8B,$33,$FE
	.dc.b	$E3,$9C,$BB,$9D,$34,$86,$DC,$AC,$FF,$A6,$7E,$74,$F8,$4F,$F0,$43
	.dc.b	$67,$EB,$35,$E1,$97,$F4,$4E,$26,$8F,$AB,$E3,$F0,$52,$F8,$BD,$3E
	.dc.b	$0F,$AF,$C2,$7C,$9F,$8F,$95,$DA,$3E,$67,$C6,$C8,$F9,$4F,$CF,$E3
	.dc.b	$FA,$1F,$13,$27,$E7,$3B,$7C,$A3,$4F,$2C,$7D,$7F,$46,$4F,$03,$DB
	.dc.b	$EF,$FC,$02,$3F,$03,$E1,$2B,$D2,$DC,$2A,$CF,$1F,$CA,$BB,$C5,$F1
	.dc.b	$6B,$58,$76,$3A,$1F,$85,$FA,$5F,$C5,$11,$FC,$47,$D1,$D7,$C1,$C8
	.dc.b	$9C,$63,$FE,$03,$85,$63,$0F,$A9,$1B,$BE,$C7,$43,$ED,$F9,$3F,$78
	.dc.b	$8F,$DE,$F8,$2A,$FB,$F2,$FE,$55,$7E,$BA,$FC,$2D,$6B,$1F,$C4,$C5
	.dc.b	$9F,$81,$3F,$A5,$F4,$26,$1F,$A3,$7F,$51,$E8,$62,$3D,$B1,$70,$F6
	.dc.b	$4D,$BF,$5C,$DA,$FB,$99,$7F,$5F,$7E,$A7,$E7,$2F,$8F,$FF,$30,$36
	.dc.b	$3F,$87,$35,$DF,$C0,$6A,$64,$69,$E3,$18,$8E,$B1,$72,$FF,$19,$57
	.dc.b	$F8,$96,$7E,$AF,$77,$F0,$32,$3F,$C0,$7D,$8A,$E8,$FE,$09,$1F,$C2
	.dc.b	$63,$FC,$6B,$F2,$72,$7F,$92,$7C,$12,$3F,$C7,$38,$E6,$7F,$C6,$B1
	.dc.b	$71,$FF,$22,$57,$B5,$BF,$31,$FE,$40,$25,$74,$F5,$6B,$FE,$3E,$8B
	.dc.b	$3C,$A7,$FD,$11,$F9,$37,$F8,$F2,$BF,$C8,$3F,$1B,$FE,$88,$EA,$4F
	.dc.b	$F8,$06,$53,$98,$C3,$7E,$13,$FD,$85,$CB,$FC,$1E,$7A,$57,$08,$C7
	.dc.b	$C2,$6A,$3F,$71,$3C,$A7,$85,$67,$0F,$0B,$67,$55,$C2,$F1,$3B,$D3
	.dc.b	$F8,$AC,$FA,$B0,$FE,$31,$18,$8F,$11,$FC,$1D,$C6,$3E,$F7,$FC,$6D
	.dc.b	$1A,$C5,$1A,$AE,$7C,$7D,$CF,$07,$BB,$EB,$B7,$D1,$3D,$6F,$F2,$7C
	.dc.b	$7F,$88,$63,$1F,$0D,$FE,$1D,$6C,$7F,$88,$9A,$FF,$89,$FE,$3B,$FC
	.dc.b	$27,$F3,$3F,$E2,$36,$C7,$F8,$A1,$AE,$FD,$B4,$F5,$F6,$5F,$21,$49
	.dc.b	$5D,$E3,$2F,$F8,$D7,$77,$C5,$BE,$85,$E0,$7C,$BA,$FE,$5D,$39,$8C
	.dc.b	$57,$0F,$A2,$DD,$7C,$9A,$D3,$6F,$3C,$9F,$8D,$9C,$B6,$B3,$C7,$FE
	.dc.b	$A3,$AB,$7F,$CF,$7E,$8B,$FD,$2B,$97,$81,$F6,$B6,$37,$92,$73,$1F
	.dc.b	$6D,$E2,$DD,$37,$18,$CB,$F4,$0F,$06,$CE,$A7,$6D,$BF,$83,$9B,$E8
	.dc.b	$BE,$0F,$5F,$FA,$E6,$31,$3F,$E2,$17,$E6,$B6,$76,$FF,$B0,$5B,$5F
	.dc.b	$39,$B9,$7D,$4A,$A7,$49,$FE,$AC,$BE,$7B,$26,$BE,$8A,$E1,$E1,$8B
	.dc.b	$5C,$26,$E3,$FE,$43,$F5,$3B,$FC,$01,$BE,$9E,$19,$8F,$F8,$EE,$74
	.dc.b	$D9,$CE,$1F,$DA,$7E,$8B,$FD,$19,$3F,$E4,$16,$C6,$7C,$3F,$6C,$63
	.dc.b	$1F,$F7,$03,$6B,$FC,$20,$DA,$FE,$FC,$FF,$27,$8E,$9E,$17,$22,$3D
	.dc.b	$EE,$D7,$F7,$BC,$5F,$0B,$EE,$3D,$FD,$36,$FD,$91,$6B,$84,$DC,$BF
	.dc.b	$CD,$13,$6B,$FC,$49,$7F,$E8,$87,$89,$EF,$F8,$CF,$69,$EE,$F0,$2C
	.dc.b	$64,$98,$FC,$78,$CB,$F2,$66,$E1,$F1,$26,$D7,$83,$D2,$FA,$E7,$D8
	.dc.b	$F4,$AC,$7D,$58,$B8,$7C,$19,$B7,$FE,$81,$9B,$5E,$A9,$B5,$F3,$BC
	.dc.b	$5F,$AA,$7D,$6F,$84,$B5,$F6,$E2,$CF,$DF,$9B,$87,$DA,$9B,$5F,$6A
	.dc.b	$6D,$7F,$8D,$F3,$7C,$9F,$E4,$3E,$53,$F2,$2D,$BF,$C7,$38,$7F,$21
	.dc.b	$E9,$5B,$FB,$8F,$E6,$0F,$CC,$F7,$D9,$FC,$87,$C9,$FE,$73,$E5,$3F
	.dc.b	$1E,$D8,$FC,$A7,$C5,$7E,$48,$FC,$9F,$89,$63,$F3,$1E,$4F,$E5,$8F
	.dc.b	$D2,$F9,$36,$3F,$D4,$5F,$41,$FC,$C3,$97,$F7,$2D,$8C,$F9,$0F,$2F
	.dc.b	$2F,$F9,$AA,$33,$58,$33,$36,$32,$4D,$FF,$CB,$F7,$FC,$FA,$75,$1F
	.dc.b	$E5,$7F,$89,$FF,$12,$57,$FA,$BD,$7A,$BF,$C2,$9C,$CE,$E3,$1E,$96
	.dc.b	$5B,$5C,$5F,$E2,$72,$FA,$27,$C8,$7F,$AA,$A0,$48,$CD,$7F,$BE,$5B
	.dc.b	$1F,$EF,$E6,$BC,$0E,$13,$DA,$37,$D5,$FD,$2B,$E3,$1F,$66,$E9,$D5
	.dc.b	$FA,$FC,$3F,$67,$87,$73,$F8,$0D,$CF,$E2,$37,$2F,$EA,$BE,$57,$FB
	.dc.b	$4A,$31,$5F,$E9,$DD,$FF,$86,$5F,$CB,$C3,$AC,$FF,$B7,$F4,$7D,$1F
	.dc.b	$F2,$F8,$C3,$6B,$E9,$7A,$FF,$CE,$AB,$7E,$5D,$AD,$69,$ED,$6E,$1E
	.dc.b	$1B,$D5,$7D,$7E,$5F,$B1,$C9,$6B,$EE,$37,$2F,$C0,$37,$1F,$6E,$9F
	.dc.b	$8F,$AF,$C8,$FA,$DF,$90,$3F,$23,$E1,$D8,$C9,$30,$FC,$6F,$42,$75
	.dc.b	$18,$AB,$87,$F7,$EE,$8F,$F0,$7D,$63,$0F,$B9,$F4,$2F,$85,$FC,$36
	.dc.b	$3F,$86,$8C,$4F,$E0,$63,$73,$A5,$3A,$74,$62,$3F,$27,$19,$C3,$F0
	.dc.b	$DC,$EF,$0B,$35,$AF,$7A,$63,$F9,$6B,$FF,$0A,$FA,$4F,$A1,$FD,$2C
	.dc.b	$FF,$BF,$F2,$5F,$0C,$FF,$D9,$A7,$91,$FE,$EB,$7C,$7B,$75,$8F,$C3
	.dc.b	$A7,$C1,$FF,$02,$6E,$3F,$C1,$97,$C3,$66,$64,$E5,$E1,$18,$9F,$DF
	.dc.b	$9C,$8B,$1A,$9E,$07,$B4,$FD,$66,$BF,$BE,$BF,$F1,$9D,$7C,$5F,$03
	.dc.b	$D3,$AB,$E7,$FF,$7B,$7B,$C3,$8D,$60,$DC,$D8,$ED,$7F,$C0,$3E,$85
	.dc.b	$E7,$FF,$06,$6C,$EB,$67,$D0,$BE,$23,$57,$2B,$9E,$0C,$38,$BF,$EF
	.dc.b	$CC,$38,$DE,$E3,$F8,$F6,$3F,$C6,$CF,$E3,$98,$5C,$3F,$90,$45,$9F
	.dc.b	$E4,$93,$75,$FD,$2F,$9F,$FC,$8C,$FE,$49,$E2,$5C,$3F,$94,$45,$BF
	.dc.b	$E5,$93,$6B,$FC,$79,$F1,$9F,$E3,$23,$FC,$67,$F4,$37,$0F,$F1,$F4
	.dc.b	$5B,$FF,$1A,$CD,$AF,$F2,$27,$AF,$FE,$3A,$3F,$C7,$7F,$19,$70,$FF
	.dc.b	$23,$45,$BF,$F1,$EC,$DA,$E3,$5A,$3F,$C9,$91,$97,$F9,$9F,$91,$FE
	.dc.b	$7A,$BC,$5F,$F9,$E2,$AC,$FF,$9F,$BD,$7F,$F2,$59,$FE,$4C,$F6,$6E
	.dc.b	$1F,$E5,$78,$B7,$FE,$4E,$9B,$8F,$F9,$77,$D6,$FF,$29,$9F,$E5,$4F
	.dc.b	$62,$E1,$FE,$5F,$8B,$7F,$E5,$69,$B5,$C6,$B4,$7F,$99,$63,$2F,$AB
	.dc.b	$E7,$D7,$FE,$62,$6C,$7E,$4B,$24,$7F,$8F,$23,$0E,$B0,$61,$FD,$02
	.dc.b	$30,$DB,$E9,$87,$56,$5F,$27,$2D,$25,$F4,$7C,$D1,$A8,$C0,$C3,$CB
	.dc.b	$1E,$9F,$3A,$57,$9B,$D9,$B7,$C2,$FD,$75,$8C,$7D,$15,$6F,$E2,$4D
	.dc.b	$BE,$D9,$2B,$11,$FD,$B7,$DE,$FE,$EB,$F4,$0F,$A1,$FE,$61,$D3,$FE
	.dc.b	$B6,$8C,$57,$F3,$3A,$DD,$6B,$2B,$39,$15,$C1,$FF,$00,$DF,$3A,$FE
	.dc.b	$6F,$7C,$27,$37,$5F,$F0,$3C,$7F,$82,$3D,$47,$B3,$C2,$51,$87,$C0
	.dc.b	$D7,$D8,$6A,$3F,$C1,$5F,$09,$FE,$7C,$DB,$A4,$63,$87,$F0,$35,$B3
	.dc.b	$FC,$2F,$5B,$70,$1C,$5F,$77,$5A,$7B,$DF,$C9,$CF,$B5,$AF,$C4,$CF
	.dc.b	$C1,$8E,$70,$F0,$DF,$1E,$1F,$F1,$03,$6F,$55,$F3,$B8,$7F,$37,$ED
	.dc.b	$FF,$89,$63,$F3,$0D,$DB,$F4,$8D,$AF,$C0,$FB,$96,$F8,$C7,$E8,$A7
	.dc.b	$1C,$46,$D6,$19,$3B,$3F,$53,$1C,$37,$91,$FA,$C7,$9F,$85,$88,$FC
	.dc.b	$C5,$FD,$5C,$BF,$5D,$5A,$9E,$04,$36,$E5,$A6,$63,$8B,$FB,$AA,$F6
	.dc.b	$7C,$07,$F9,$06,$3F,$DD,$5C,$B8,$33,$CB,$63,$24,$3F,$DC,$AE,$5F
	.dc.b	$F7,$3C,$5B,$FF,$7B,$4F,$6C,$E5,$7F,$BD,$DF,$0E,$78,$8F,$F7,$7C
	.dc.b	$5A,$FF,$7F,$4F,$9B,$29,$5F,$EF,$F7,$D3,$96,$23,$FD,$EF,$16,$BF
	.dc.b	$CA,$79,$7F,$86,$67,$81,$0D,$9F,$C7,$DD,$38,$24,$72,$9E,$F1,$FC
	.dc.b	$97,$E1,$BF,$D0,$2D,$F2,$3F,$0D,$18,$8F,$E2,$17,$FB,$07,$44,$3F
	.dc.b	$8A,$46,$EF,$CF,$CB,$F8,$FF,$45,$E8,$52,$5F,$FA,$47,$CF,$AF,$FA
	.dc.b	$56,$2D,$FF,$95,$A7,$97,$BD,$2B,$FD,$30,$F5,$FA,$1C,$47,$FA,$76
	.dc.b	$2D,$7F,$97,$A7,$F4,$52,$FA,$1F,$F1,$8B,$FE,$0C,$6B,$FE,$3A,$6C
	.dc.b	$FF,$89,$3E,$87,$E0,$4E,$7C,$EE,$75,$AE,$8F,$F4,$B3,$67,$FD,$3F
	.dc.b	$5C,$2F,$FC,$5B,$F4,$2F,$8C,$F1,$B6,$BF,$C3,$AD,$7F,$C9,$7F,$43
	.dc.b	$FE,$28,$73,$3C,$08,$FF,$89,$23,$4F,$F9,$4B,$E8,$5F,$23,$F0,$1B
	.dc.b	$1F,$E2,$A6,$BB,$F5,$75,$32,$34,$F1,$8C,$47,$58,$B8,$FF,$92,$2B
	.dc.b	$FC,$C7,$3F,$E3,$6F,$A0,$7C,$17,$D0,$6C,$66,$3E,$9F,$D0,$FF,$A2
	.dc.b	$63,$04,$3F,$CD,$73,$C1,$FF,$36,$C6,$DC,$FA,$1D,$48,$FF,$91,$BE
	.dc.b	$87,$E8,$70,$FF,$B2,$6D,$5F,$E8,$6A,$FF,$69,$C6,$1C,$35,$FF,$45
	.dc.b	$B5,$FF,$49,$53,$C6,$FC,$5B,$63,$C1,$18,$F3,$32,$C8,$C9,$E5,$8C
	.dc.b	$3F,$E8,$89,$E4,$D9,$E4,$FF,$BD,$7E,$1F,$F5,$5F,$13,$23,$04,$3F
	.dc.b	$DF,$8E,$5F,$F7,$F4,$7F,$41,$F8,$67,$D0,$FB,$78,$3F,$DD,$0D,$D7
	.dc.b	$FD,$EB,$3A,$D7,$F7,$1F,$15,$C6,$B5,$8A,$B4,$FF,$B7,$E3,$3E,$53
	.dc.b	$BD,$7F,$C0,$7E,$87,$F8,$17,$D0,$CC,$3F,$C6,$94,$9D,$46,$27,$4F
	.dc.b	$F8,$7E,$74,$F0,$9C,$E9,$FE,$DA,$AE,$33,$FE,$B2,$FA,$17,$CE,$7F
	.dc.b	$00,$6C,$7F,$A2,$9A,$FF,$AE,$3E,$61,$F4,$DB,$AF,$C3,$4A,$B6,$FF
	.dc.b	$B4,$DA,$88,$E1,$FF,$1E,$7C,$5B,$E8,$7F,$91,$71,$49,$FF,$2A,$4E
	.dc.b	$5B,$39,$7E,$04,$FF,$77,$9D,$47,$9F,$E8,$3F,$8D,$E4,$69,$8F,$A1
	.dc.b	$B3,$FE,$A8,$9D,$36,$74,$D2,$1D,$4E,$13,$FE,$07,$F9,$E7,$C0,$52
	.dc.b	$4F,$AB,$E8,$75,$EF,$48,$FF,$38,$52,$39,$7B,$FC,$FE,$03,$D9,$F1
	.dc.b	$92,$F7,$1E,$FF,$8B,$91,$F1,$9F,$85,$C5,$F0,$FE,$1E,$4F,$CA,$76
	.dc.b	$FF,$99,$78,$3F,$D9,$4F,$CE,$F3,$BE,$51,$C1,$FF,$35,$69,$FE,$63
	.dc.b	$7C,$FE,$DF,$BC,$F8,$1E,$FD,$8C,$90,$E2,$E5,$E7,$17,$A5,$CB,$FE
	.dc.b	$AE,$BF,$F2,$0C,$62,$BF,$0B,$F4,$3F,$74,$8F,$E1,$3E,$86,$BE,$B8
	.dc.b	$B3,$C7,$BD,$A2,$F3,$8F,$FA,$F2,$17,$8C,$3F,$27,$5F,$6F,$DC,$91
	.dc.b	$F5,$9F,$BD,$EF,$62,$3E,$2C,$5C,$BC,$19,$FF,$91,$67,$81,$0D,$9F
	.dc.b	$80,$AF,$CC,$79,$D2,$3F,$3E,$FE,$6F,$CF,$C4,$7D,$78,$B7,$ED,$CF
	.dc.b	$FA,$5C,$E4,$E2,$43,$EF,$57,$F9,$6F,$E8,$5F,$1B,$49,$3F,$EC,$18
	.dc.b	$FF,$0A,$6D,$FE,$E0,$77,$19,$9D,$BF,$EA,$D8,$DB,$FE,$B5,$76,$E6
	.dc.b	$37,$87,$D3,$9D,$EF,$F5,$3B,$BF,$70,$4B,$FD,$FF,$A7,$5F,$CB,$45
	.dc.b	$BF,$81,$3F,$E3,$0F,$97,$95,$FE,$35,$7F,$C6,$78,$E2,$3F,$C5,$F1
	.dc.b	$6B,$F0,$9E,$EF,$99,$BF,$2D,$FE,$3A,$25,$7F,$8F,$FD,$4A,$FF,$8E
	.dc.b	$22,$CF,$29,$FF,$43,$7D,$0F,$F8,$E2,$BF,$D6,$7C,$5F,$C7,$F5,$4F
	.dc.b	$F7,$3F,$89,$FC,$24,$63,$DE,$FF,$66,$56,$EB,$95,$AF,$71,$86,$D7
	.dc.b	$09,$FF,$72,$4F,$5A,$E5,$EF,$E7,$0F,$0B,$67,$55,$C2,$F1,$3B,$D3
	.dc.b	$F4,$F3,$EA,$3F,$DE,$0E,$1F,$26,$1E,$D8,$C6,$3E,$C5,$F2,$DC,$F9
	.dc.b	$65,$F3,$CD,$5C,$30,$E5,$7E,$BA,$7D,$0B,$D4,$FF,$25,$C7,$F4,$93
	.dc.b	$2E,$1F,$E9,$91,$BF,$82,$FF,$0C,$36,$38,$1F,$ED,$6F,$65,$F2,$FF
	.dc.b	$E1,$DF,$8C,$46,$27,$67,$FB,$65,$C3,$CA,$37,$3D,$BD,$57,$C6,$D2
	.dc.b	$57,$2C,$3F,$BE,$C7,$F8,$AF,$D3,$FE,$BD,$DB,$FD,$8C,$F2,$B8,$AF
	.dc.b	$E4,$07,$F6,$77,$11,$AF,$4B,$F9,$23,$4F,$AD,$3F,$DD,$23,$F9,$47
	.dc.b	$1F,$D3,$E5,$BC,$3F,$27,$0E,$D3,$FC,$56,$FF,$BD,$F9,$5F,$CB,$FA
	.dc.b	$7F,$4B,$F2,$1C,$27,$C5,$39,$CF,$F0,$13,$99,$E3,$1E,$D9,$F4,$C7
	.dc.b	$76,$BC,$A3,$47,$F8,$0E,$BD,$F8,$2B,$B5,$39,$9F,$90,$BE,$67,$33
	.dc.b	$E0,$18,$7F,$47,$F6,$DF,$A6,$FF,$40,$CF,$B1,$FE,$38,$71,$5F,$EB
	.dc.b	$2E,$CE,$F5,$C3,$B2,$D6,$B0,$E8,$DC,$3F,$41,$7A,$AF,$3E,$7F,$9C
	.dc.b	$F8,$8F,$4D,$3E,$1D,$F1,$7A,$EF,$56,$A7,$35,$FE,$CD,$8C,$56,$DB
	.dc.b	$1E,$AD,$7F,$C5,$5F,$89,$3B,$BE,$3F,$63,$E9,$DE,$97,$C6,$AD,$61
	.dc.b	$C6,$B5,$39,$BF,$AB,$82,$F6,$FE,$89,$AF,$EA,$9A,$FE,$CF,$F1,$2F
	.dc.b	$4B,$F4,$B0,$FF,$1F,$61,$E3,$F4,$7E,$BD,$6B,$8F,$FB,$E4,$E6,$BD
	.dc.b	$96,$AD,$47,$2C,$FF,$BA,$BA,$8C,$CF,$EE,$A7,$2F,$84,$FF,$27,$46
	.dc.b	$9E,$9A,$FC,$6C,$1E,$A7,$92,$3C,$BF,$07,$F2,$BE,$15,$F2,$18,$CF
	.dc.b	$E4,$46,$1F,$14,$EE,$75,$8F,$87,$4F,$F5,$04,$63,$2F,$DD,$E7,$FE
	.dc.b	$AC,$C3,$E9,$53,$D6,$D4,$7C,$BF,$07,$C5,$F2,$1F,$21,$B1,$F0,$5B
	.dc.b	$97,$BC,$F9,$0E,$31,$EE,$F1,$9E,$FE,$15,$AE,$1F,$88,$7F,$B8,$27
	.dc.b	$F0,$9F,$89,$7B,$67,$3E,$B7,$E3,$23,$18,$FF,$14,$BE,$15,$EF,$9C
	.dc.b	$DF,$F3,$59,$CD,$7D,$E9,$CD,$7F,$A0,$67,$35,$BB,$FD,$DF,$E2,$4F
	.dc.b	$67,$F7,$B5,$AB,$E1,$EA,$F8,$F5,$CB,$71,$FC,$4D,$23,$15,$EE,$6E
	.dc.b	$7E,$97,$8D,$7E,$43,$F4,$EF,$C3,$5F,$18,$FD,$37,$E2,$7F,$E4,$76
	.dc.b	$FB,$9E,$4F,$27,$F5,$F3,$9C,$BE,$5C,$62,$75,$5C,$0F,$CE,$F9,$9B
	.dc.b	$CF,$E5,$57,$F1,$A8,$C4,$FC,$98,$FE,$17,$E2,$3E,$B2,$9E,$0F,$6A
	.dc.b	$D7,$C0,$FA,$FC,$34,$F4,$C7,$52,$3F,$EA,$AF,$95,$BA,$7E,$27,$E9
	.dc.b	$9F,$3D,$FC,$C3,$D3,$7D,$73,$AB,$F9,$7D,$6B,$1F,$E2,$55,$A9,$CD
	.dc.b	$F1,$BE,$15,$FB,$CB,$E1,$87,$ED,$70,$E1,$39,$BF,$E2,$35,$AA,$FE
	.dc.b	$EE,$D7,$FC,$00,$D7,$FD,$8F,$9F,$C6,$DB,$E3,$B6,$FF,$D8,$ED,$9F
	.dc.b	$F4,$F3,$78,$3F,$CD,$11,$F3,$B1,$F9,$ED,$AF,$F6,$7B,$4F,$F5,$31
	.dc.b	$67,$FD,$29,$F4,$DF,$E0,$AF,$07,$FA,$BB,$FE,$0C,$CE,$F7,$FE,$FE
	.dc.b	$FF,$8A,$4E,$70,$FE,$74,$F1,$6C,$FC,$6C,$3A,$BC,$5B,$7D,$30,$EA
	.dc.b	$DA,$E2,$FD,$6F,$6B,$E9,$75,$D9,$FA,$FB,$FE,$B5,$FA,$93,$F4,$08
	.dc.b	$FF,$88,$33,$6F,$FE,$31,$8F,$F1,$1F,$9C,$FC,$35,$F5,$DF,$D8,$17
	.dc.b	$D7,$D8,$C3,$6B,$FA,$C5,$71,$8C,$61,$FC,$DE,$31,$3F,$D0,$BE,$A1
	.dc.b	$F1,$7F,$E0,$2E,$0F,$EB,$9F,$09,$FE,$00,$9C,$D7,$F6,$39,$CB,$C6
	.dc.b	$B8,$79,$AB,$E1,$18,$6C,$FB,$EF,$9E,$3F,$E4,$06,$BF,$E4,$86,$BF
	.dc.b	$E7,$BF,$CC,$BF,$C2,$D3,$EC,$5F,$17,$FE,$02,$73,$3E,$D7,$33,$BF
	.dc.b	$73,$55,$B6,$C7,$C1,$EE,$FD,$97,$D4,$FF,$AD,$A3,$07,$EC,$1C,$61
	.dc.b	$FD,$07,$1F,$E0,$D3,$EB,$AF,$EF,$71,$89,$F6,$47,$F7,$EF,$BE,$7C
	.dc.b	$7F,$F8,$83,$F0,$3F,$E6,$B4,$9C,$E1,$FE,$0F,$F7,$95,$A9,$CD,$7F
	.dc.b	$6F,$9C,$FA,$5F,$E3,$49,$CF,$17,$F8,$CE,$F8,$63,$E1,$26,$3F,$8F
	.dc.b	$AE,$78,$FF,$9D,$BA,$9F,$37,$F0,$BD,$1F,$F1,$FC,$62,$FE,$14,$62
	.dc.b	$7F,$C3,$33,$9A,$FF,$3C,$B5,$E9,$1F,$E1,$EF,$09,$F2,$DF,$C7,$FE
	.dc.b	$AD,$F2,$FF,$E4,$C9,$CE,$5F,$C7,$A7,$2F,$F8,$5A,$33,$87,$FA,$19
	.dc.b	$AF,$FA,$39,$AF,$FA,$5D,$AB,$81,$FA,$83,$E4,$BF,$C9,$3F,$A8,$5C
	.dc.b	$3E,$84,$DB,$FD,$34,$DE,$2F,$F2,$F4,$E6,$7F,$D4,$98,$74,$C3,$FC
	.dc.b	$DE,$DC,$3F,$D3,$6D,$9F,$F5,$1F,$6F,$EE,$FE,$CF,$F7,$47,$EE,$FE
	.dc.b	$9E,$DF,$EE,$3C,$24,$E2,$FF,$71,$0A,$C6,$7F,$E7,$1E,$17,$D6,$7F
	.dc.b	$99,$3E,$2B,$FA,$1E,$1F,$87,$9C,$E5,$FC,$12,$73,$7C,$6F,$D3,$5A
	.dc.b	$BF,$F5,$CE,$9F,$84,$9C,$9E,$ED,$9E,$87,$F9,$B7,$EB,$7F,$CF,$71
	.dc.b	$8C,$FF,$88,$FC,$42,$73,$5C,$49,$3F,$1A,$B8,$63,$E0,$6E,$7F,$42
	.dc.b	$BF,$CF,$5F,$53,$CA,$FE,$BC,$5B,$FE,$39,$8A,$B1,$6B,$57,$F9,$E7
	.dc.b	$CE,$FD,$3C,$63,$89,$F9,$47,$F9,$EA,$3F,$D0,$1F,$0F,$FE,$D8,$EE
	.dc.b	$FF,$47,$CE,$70,$FF,$30,$4E,$6B,$FD,$B2,$D8,$FF,$70,$B5,$FF,$75
	.dc.b	$35,$F1,$F0,$3E,$DF,$EF,$D7,$F3,$8F,$BE,$FE,$33,$5F,$BF,$9C,$BF
	.dc.b	$CE,$63,$3A,$7F,$AC,$67,$2F,$BE,$7F,$D7,$53,$AA,$EF,$5A,$7E,$04
	.dc.b	$66,$7F,$4D,$3D,$9F,$CE,$B5,$E3,$CF,$F9,$E8,$C4,$7F,$B7,$6B,$FA
	.dc.b	$7F,$DF,$73,$8F,$F7,$54,$63,$C8,$FF,$64,$C3,$D6,$FF,$5C,$D6,$E3
	.dc.b	$15,$B3,$FD,$2A,$E0,$FF,$38,$17,$2F,$76,$EC,$DF,$63,$DB,$19,$87
	.dc.b	$27,$FD,$8F,$3A,$8C,$1D,$B0,$7C,$9F,$E1,$BE,$C7,$9C,$7F,$B8,$F3
	.dc.b	$FE,$26,$F8,$8D,$88,$DE,$49,$DE,$7F,$EF,$2B,$E9,$9F,$E3,$E7,$38
	.dc.b	$FE,$53,$19,$5A,$7C,$1E,$17,$C8,$7B,$E5,$E4,$7F,$37,$EC,$FD,$F8
	.dc.b	$C4,$7E,$7A,$B9,$3A,$BA,$E4,$F4,$63,$E1,$57,$CE,$31,$1E,$07,$51
	.dc.b	$98,$E1,$5D,$23,$11,$EA,$3A,$9E,$D7,$2F,$87,$89,$F0,$7D,$AB,$F6
	.dc.b	$B1,$89,$EA,$7F,$BB,$E7,$4D,$3F,$77,$F2,$EF,$8A,$6D,$CA,$BE,$64
	.dc.b	$E6,$31,$5C,$2B,$6F,$31,$3B,$70,$DD,$3E,$82,$FA,$0F,$B1,$18,$6E
	.dc.b	$9E,$0B,$EB,$3C,$1B,$1E,$49,$DD,$FA,$9A,$F8,$A7,$6F,$9F,$0F,$B7
	.dc.b	$87,$86,$2D,$70,$8F,$35,$F8,$6B,$F1,$5C,$1F,$93,$F2,$3F,$1D,$1F
	.dc.b	$A8,$E1,$FE,$E7,$8F,$F2,$48,$FE,$10,$F8,$5E,$F9,$78,$AF,$94,$FE
	.dc.b	$FA,$F9,$F0,$77,$8F,$BE,$7C,$AF,$B0,$CB,$F5,$9A,$E4,$FE,$3D,$C9
	.dc.b	$8D,$13,$49,$3A,$3C,$D5,$F7,$23,$AE,$0F,$C5,$3F,$B2,$57,$F9,$81
	.dc.b	$AF,$FB,$DE,$FA,$BC,$8F,$21,$BF,$17,$9C,$F0,$D3,$E1,$D7,$0A,$FF
	.dc.b	$18,$7D,$B3,$EA,$BF,$CF,$1D,$88,$C4,$FF,$80,$1A,$FF,$6A,$6B,$FE
	.dc.b	$08,$73,$7F,$C6,$BE,$FF,$FC,$8B,$3F,$E7,$86,$BE,$0B,$F9,$77,$DD
	.dc.b	$F8,$33,$AA,$E5,$DF,$AC,$3A,$ED,$FA,$CF,$03,$FD,$99,$97,$F5,$66
	.dc.b	$BF,$D5,$9A,$FF,$56,$6B,$C7,$E0,$FE,$7F,$82,$F9,$3A,$7D,$EE,$B5
	.dc.b	$F4,$CF,$FC,$4F,$F0,$BE,$8F,$1C,$D3,$EC,$F8,$DF,$E5,$0C,$FF,$C6
	.dc.b	$31,$96,$9F,$E0,$AF,$2B,$0F,$F8,$16,$3E,$8D,$74,$8F,$F0,$54,$60
	.dc.b	$EB,$F0,$72,$39,$9F,$E2,$3F,$4B,$EC,$D9,$A4,$8F,$F1,$CC,$65,$FF
	.dc.b	$55,$FE,$9E,$FA,$2F,$C8,$B6,$3F,$CF,$13,$B7,$FC,$6F,$F3,$9F,$96
	.dc.b	$F4,$4D,$31,$F7,$56,$7F,$C8,$11,$5F,$F1,$FC,$7C,$97,$DE,$43,$FC
	.dc.b	$50,$53,$FD,$79,$FA,$8B,$D6,$FD,$2B,$63,$FC,$BA,$7E,$CB,$F5,$2F
	.dc.b	$E5,$CF,$F3,$E6,$BF,$D0,$DA,$FF,$55,$6B,$FE,$2A,$78,$F0,$FF,$08
	.dc.b	$F8,$FF,$D4,$74,$7F,$5B,$6D,$6F,$CA,$FE,$03,$F8,$BB,$EC,$3F,$DC
	.dc.b	$12,$8C,$3F,$EE,$36,$BF,$D4,$9A,$FF,$57,$73,$87,$F1,$8F,$C5,$FF
	.dc.b	$29,$3F,$BD,$0E,$1F,$E8,$B8,$FF,$6F,$AE,$2F,$F7,$1F,$A5,$FF,$37
	.dc.b	$CF,$FB,$CD,$AF,$F6,$26,$BF,$D9,$DA,$F1,$F5,$7F,$8E,$FD,$17,$C0
	.dc.b	$83,$67,$9E,$3F,$CB,$9B,$7F,$E0,$19,$E0,$6E,$FF,$93,$CF,$EC,$AF
	.dc.b	$FC,$B2,$D4,$D2,$77,$FF,$2E,$FB,$77,$CA,$53,$9C,$7F,$28,$F8,$9F
	.dc.b	$EF,$F7,$FE,$EB,$E3,$C1,$9F,$56,$46,$48,$7F,$BB,$5F,$E7,$7F,$6C
	.dc.b	$F9,$3F,$B6,$D9,$FF,$04,$87,$0F,$F8,$31,$AF,$F8,$45,$CE,$FF,$D4
	.dc.b	$BE,$DB,$F5,$4D,$EE,$EF,$72,$FE,$F5,$6B,$87,$FC,$54,$D8,$FF,$3F
	.dc.b	$E1,$D2,$7F,$CF,$DD,$6F,$B3,$FD,$A7,$C0,$CB,$FF,$58,$47,$ED,$CC
	.dc.b	$3F,$EB,$08,$C3,$F8,$DB,$FA,$2F,$B9,$AF,$B7,$8B,$E1,$91,$F3,$CF
	.dc.b	$F6,$2C,$7F,$D6,$F1,$CE,$71,$D9,$9E,$49,$3E,$E8,$CB,$F0,$23,$2F
	.dc.b	$E7,$0F,$F4,$ED,$27,$71,$91,$1F,$90,$77,$3A,$DB,$FC,$D7,$9B,$EC
	.dc.b	$7F,$8C,$79,$72,$FF,$DA,$91,$FC,$6C,$C3,$FE,$D4,$8C,$3F,$ED,$9F
	.dc.b	$AB,$FE,$41,$F6,$52,$30,$43,$FD,$B6,$E5,$FF,$6E,$C7,$F8,$1B,$ED
	.dc.b	$1F,$03,$FC,$19,$B3,$F8,$C6,$BF,$E5,$66,$BF,$E5,$B7,$39,$7F,$87
	.dc.b	$3E,$11,$F1,$DF,$E1,$26,$D7,$F0,$7A,$DF,$27,$F9,$CD,$B1,$FE,$C8
	.dc.b	$9D,$BF,$2D,$B7,$FE,$28,$9D,$F9,$BF,$E5,$CE,$1F,$F1,$0B,$6F,$FC
	.dc.b	$20,$D7,$FD,$F7,$EE,$7F,$89,$3F,$16,$7D,$27,$0F,$C5,$9F,$45,$4D
	.dc.b	$47,$F4,$BF,$83,$98,$7F,$86,$FD,$E7,$37,$F9,$16,$3F,$09,$9B,$1E
	.dc.b	$71,$E2,$F2,$7E,$F3,$FC,$9F,$3F,$F0,$F6,$FF,$E4,$26,$F0,$FF,$BA
	.dc.b	$30,$FF,$3B,$F3,$F9,$FD,$FF,$37,$24,$C7,$94,$57,$C3,$9F,$C6,$C3
	.dc.b	$93,$63,$FD,$E7,$8F,$67,$FD,$EB,$E2,$3E,$CF,$FC,$8E,$D8,$FE,$78
	.dc.b	$D7,$F3,$9E,$C7,$D6,$51,$9D,$7F,$DC,$5A,$BF,$15,$7C,$C7,$D4,$6D
	.dc.b	$F4,$C7,$B9,$99,$91,$93,$C2,$F9,$08,$FD,$CF,$C5,$9F,$3F,$88,$FA
	.dc.b	$36,$FA,$2F,$01,$9C,$BF,$CD,$F9,$F8,$AB,$73,$A7,$B9,$FD,$6F,$F1
	.dc.b	$57,$DC,$FF,$9A,$9B,$19,$8F,$C7,$F9,$7F,$C4,$B0,$FF,$3D,$D6,$AB
	.dc.b	$F0,$7D,$FF,$E1,$76,$FB,$5C,$F0,$91,$F9,$69,$D5,$FC,$A6,$F2,$7F
	.dc.b	$B5,$F0,$E9,$3D,$BE,$7B,$C8,$4B,$FC,$FE,$5F,$E8,$C9,$D4,$75,$CF
	.dc.b	$F4,$B8,$FC,$58,$F7,$5F,$17,$CF,$3A,$31,$5E,$CA,$F8,$44,$9F,$D6
	.dc.b	$FC,$77,$F3,$D7,$E4,$47,$EC,$BC,$8F,$F7,$45,$7E,$11,$B9,$FA,$E7
	.dc.b	$95,$7E,$A2,$72,$FF,$97,$1A,$B3,$4F,$D3,$F8,$FF,$F0,$DC,$1A,$49
	.dc.b	$FE,$C7,$19,$7F,$81,$D3,$E2,$FC,$C7,$38,$FD,$E7,$C1,$7E,$31,$FF
	.dc.b	$4C,$62,$FC,$DF,$FB,$CC,$7F,$41,$8C,$6F,$FE,$6C,$E4,$93,$C2,$B7
	.dc.b	$AA,$D3,$67,$31,$8F,$05,$8F,$86,$BF,$A7,$7C,$4B,$CA,$FC,$C6,$3C
	.dc.b	$E5,$39,$E9,$FF,$4B,$D6,$B1,$E9,$87,$E5,$5F,$0D,$6E,$FF,$D6,$8D
	.dc.b	$7B,$D7,$15,$8F,$93,$CE,$FF,$65,$FD,$AB,$E5,$69,$2B,$F9,$27,$CB
	.dc.b	$7F,$2C,$7F,$C8,$93,$B7,$FC,$2A,$D7,$E1,$91,$C3,$80,$E1,$D4,$FF
	.dc.b	$A6,$BD,$5E,$3C,$FF,$EA,$EC,$7E,$4C,$63,$B3,$77,$E2,$F6,$F0,$FD
	.dc.b	$2C,$7F,$C8,$EF,$D4,$C3,$FD,$6B,$18,$84,$75,$DF,$FC,$CA,$EA,$BD
	.dc.b	$3B,$F9,$EF,$83,$63,$37,$ED,$6B,$E1,$E3,$7C,$47,$C4,$6C,$7C,$13
	.dc.b	$E0,$B8,$7B,$35,$F9,$3F,$1B,$F8,$5F,$0F,$FC,$82,$D8,$E4,$D7,$E6
	.dc.b	$35,$FA,$1C,$3F,$9C,$6F,$3F,$D6,$8C,$47,$C7,$6B,$FA,$9F,$73,$F2
	.dc.b	$5A,$FF,$B7,$3A,$7F,$CC,$3A,$FC,$BE,$0F,$C8,$FC,$D6,$9F,$BF,$3A
	.dc.b	$C3,$E6,$C6,$30,$D4,$7E,$9A,$75,$87,$D1,$BC,$CF,$87,$0E,$CE,$FE
	.dc.b	$83,$F4,$8F,$E2,$6D,$18,$8F,$0B,$5E,$13,$FE,$68,$9D,$BF,$E9,$B6
	.dc.b	$C7,$F8,$FF,$91,$CD,$FE,$CC,$AD,$5F,$EC,$FA,$BC,$38,$70,$6C,$66
	.dc.b	$7F,$7D,$5C,$70,$FF,$23,$5E,$E7,$F6,$BF,$69,$FD,$65,$FD,$AB,$67
	.dc.b	$C5,$CD,$FE,$9E,$9C,$F8,$3D,$67,$F8,$1D,$6F,$2F,$E0,$DD,$68,$C5
	.dc.b	$FF,$04,$E7,$E1,$39,$C3,$96,$BF,$C4,$BE,$CF,$A5,$BE,$D1,$F6,$3A
	.dc.b	$D5,$5A,$9C,$D7,$F2,$A8,$C5,$6D,$B1,$FE,$C7,$FB,$3F,$D0,$BF,$E9
	.dc.b	$86,$ED,$FA,$16,$BF,$C0,$DA,$FF,$08,$77,$B7,$E0,$1B,$3F,$E2,$98
	.dc.b	$FF,$2D,$BD,$FA,$FE,$35,$3E,$CB,$CE,$FF,$8E,$3E,$9D,$E7,$6A,$73
	.dc.b	$87,$F8,$72,$73,$87,$FB,$7E,$78,$DF,$F1,$76,$BF,$C7,$1A,$FF,$40
	.dc.b	$6B,$E1,$CF,$B1,$F6,$7B,$58,$3E,$CF,$38,$77,$F0,$75,$ED,$3E,$9B
	.dc.b	$CD,$18,$6C,$FF,$B0,$B3,$FF,$63,$36,$7F,$41,$9F,$C1,$6C,$EA,$3F
	.dc.b	$D3,$0D,$C7,$FC,$77,$97,$F9,$9A,$FC,$47,$C1,$EB,$F1,$FA,$FF,$1D
	.dc.b	$B1,$93,$C4,$F6,$23,$83,$58,$7F,$9A,$2B,$FC,$97,$18,$BF,$DD,$FD
	.dc.b	$7B,$EC,$BE,$37,$D9,$FC,$1A,$FE,$D7,$5A,$9C,$EB,$F7,$AF,$85,$7F
	.dc.b	$B9,$1B,$8F,$F2,$D6,$BF,$D3,$DA,$FF,$54,$77,$7D,$67,$B9,$FD,$63
	.dc.b	$EC,$DF,$13,$DF,$1F,$07,$C0,$79,$B6,$B7,$FE,$92,$D1,$18,$CF,$C5
	.dc.b	$5F,$53,$3F,$F4,$8D,$F9,$E3,$06,$F3,$FE,$3B,$F6,$7F,$94,$9F,$34
	.dc.b	$FE,$27,$E4,$5F,$67,$4F,$85,$E1,$DA,$F4,$E5,$FE,$12,$CB,$FC,$45
	.dc.b	$1F,$16,$7C,$44,$7F,$A9,$4B,$C7,$EB,$1C,$A1,$5C,$9F,$F7,$04,$E9
	.dc.b	$FF,$71,$46,$73,$FF,$0C,$78,$1F,$E7,$5F,$7F,$F8,$7F,$85,$FE,$C6
	.dc.b	$C7,$FD,$6F,$A7,$F8,$F9,$FF,$57,$46,$71,$E7,$F1,$BF,$1C,$93,$F7
	.dc.b	$5F,$CE,$E7,$C6,$7C,$7E,$BF,$6F,$99,$FE,$03,$39,$D3,$FA,$3F,$D8
	.dc.b	$BE,$6B,$F9,$57,$A0,$FB,$2F,$C9,$E5,$FE,$FB,$AD,$69,$F2,$AB,$5E
	.dc.b	$87,$F0,$C9,$CD,$7F,$79,$6B,$FE,$1B,$6B,$FE,$20,$6B,$FE,$44,$78
	.dc.b	$E3,$E4,$C7,$F9,$AF,$E0,$4F,$85,$ED,$1F,$23,$7F,$93,$CB,$74,$F8
	.dc.b	$ED,$CB,$FA,$C7,$0F,$D1,$9F,$F2,$72,$AE,$31,$F3,$A7,$E9,$EB,$F5
	.dc.b	$1B,$19,$3E,$9C,$65,$A7,$E3,$3E,$D1,$E0,$7F,$09,$8F,$F2,$FC,$63
	.dc.b	$2F,$EA,$CE,$BC,$3B,$5F,$02,$B6,$F1,$C1,$5C,$67,$B4,$E6,$31,$3F
	.dc.b	$CA,$27,$8D,$74,$BE,$AD,$8F,$CC,$6F,$FE,$C8,$7E,$D7,$0F,$EE,$9C
	.dc.b	$FA,$DF,$82,$8C,$61,$F8,$18,$FB,$C6,$7B,$6D,$7E,$B3,$A5,$F1,$0F
	.dc.b	$B4,$FE,$D6,$BE,$D2,$C6,$1B,$5C,$67,$4D,$9D,$3F,$AF,$FB,$37,$C1
	.dc.b	$FF,$7F,$E8,$79,$BF,$EE,$EA,$D4,$E6,$31,$AF,$C5,$3F,$D2,$37,$CA
	.dc.b	$BF,$7A,$D8,$FF,$28,$35,$FF,$39,$35,$FF,$3B,$BB,$BF,$F1,$7B,$67
	.dc.b	$FD,$1B,$D3,$F2,$27,$38,$7F,$83,$3E,$5F,$FC,$3D,$87,$FA,$31,$FE
	.dc.b	$EA,$FF,$8B,$5C,$EA,$73,$5F,$C2,$1C,$98,$C3,$FD,$50,$E3,$E4,$7F
	.dc.b	$BE,$E5,$FC,$1E,$BF,$D4,$DC,$EF,$B4,$7C,$5E,$BC,$3F,$DF,$C6,$32
	.dc.b	$F9,$D1,$97,$F7,$D1,$F2,$AF,$DF,$3F,$E0,$AF,$B3,$7C,$AF,$F8,$AB
	.dc.b	$ED,$3F,$AA,$6F,$F3,$C4,$EB,$1F,$F5,$7C,$EB,$0F,$F2,$BC,$E6,$BF
	.dc.b	$D0,$0D,$7F,$D4,$4D,$7F,$D4,$ED,$7F,$D9,$6F,$18,$E7,$7F,$E4,$3F
	.dc.b	$C3,$9F,$87,$7F,$8D,$B1,$FE,$71,$7B,$7F,$D6,$3E,$7B,$E9,$3E,$1F
	.dc.b	$A7,$FE,$26,$8C,$4F,$C2,$8C,$4C,$8F,$8B,$EF,$DD,$3E,$07,$88,$FA
	.dc.b	$DF,$E5,$7F,$6D,$FE,$F5,$BF,$E9,$13,$97,$F9,$54,$66,$B8,$D7,$F8
	.dc.b	$DA,$B5,$1F,$99,$E4,$56,$38,$3F,$CF,$F5,$70,$FF,$6E,$FD,$43,$ED
	.dc.b	$7F,$CF,$DF,$06,$8C,$4F,$FA,$B6,$31,$3A,$9C,$E5,$FE,$D1,$9C,$D7
	.dc.b	$FB,$9A,$73,$87,$F5,$AF,$01,$F3,$DF,$E6,$A9,$CD,$7F,$BE,$E7,$2F
	.dc.b	$EF,$67,$F8,$1F,$D7,$FF,$04,$64,$64,$FE,$05,$18,$C7,$FE,$00,$F0
	.dc.b	$3E,$83,$D6,$DA,$F5,$36,$7D,$0D,$7F,$DD,$B9,$FD,$DA,$FF,$3F,$CE
	.dc.b	$F3,$FF,$42,$FE,$A7,$FE,$4F,$C3,$FC,$0B,$7B,$7E,$07,$A9,$F1,$48
	.dc.b	$FC,$9D,$7F,$B7,$65,$D9,$FE,$D9,$1F,$4A,$7F,$D0,$DF,$6D,$F1,$59
	.dc.b	$1D,$0F,$E5,$1F,$10,$9C,$DF,$F1,$7C,$3E,$C5,$E3,$2F,$F0,$4D,$EF
	.dc.b	$0E,$57,$FC,$77,$E1,$7F,$A8,$F6,$7F,$AF,$5B,$3F,$EE,$B6,$C8,$87
	.dc.b	$FB,$CB,$D1,$FD,$64,$6B,$05,$6E,$30,$F1,$F1,$5F,$5B,$FD,$B7,$0E
	.dc.b	$58,$FF,$AB,$3E,$DB,$F0,$31,$FE,$C9,$FB,$6F,$E3,$0B,$64,$EB,$3F
	.dc.b	$F1,$2D,$6A,$73,$CB,$F9,$AA,$E7,$7F,$71,$AF,$AE,$77,$9F,$E7,$67
	.dc.b	$A3,$96,$CB,$36,$1F,$CD,$2F,$8C,$FF,$19,$3D,$92,$FB,$4F,$4C,$7F
	.dc.b	$BD,$70,$FE,$C1,$18,$8F,$F2,$11,$F3,$0D,$1F,$EE,$D8,$FF,$25,$BA
	.dc.b	$3F,$DE,$CF,$F9,$2D,$EF,$6F,$B5,$FF,$5B,$C7,$CD,$39,$E7,$FE,$FB
	.dc.b	$F6,$3F,$DA,$F5,$BB,$E5,$1C,$29,$F2,$3F,$7F,$ED,$BA,$47,$C1,$CB
	.dc.b	$F8,$86,$3F,$7A,$3D,$15,$BC,$BE,$F5,$6B,$1F,$65,$74,$3F,$CF,$73
	.dc.b	$EF,$EF,$E5,$E9,$F6,$9F,$C6,$E1,$F8,$F8,$C4,$27,$75,$D2,$BC,$13
	.dc.b	$F9,$49,$F2,$65,$C2,$3C,$0F,$28,$CD,$71,$9F,$19,$C7,$0F,$04,$67
	.dc.b	$0E,$11,$89,$F7,$9C,$0F,$D2,$57,$91,$CC,$F0,$74,$7E,$A7,$0F,$D2
	.dc.b	$47,$87,$8F,$FD,$3B,$3D,$1A,$FC,$20,$E3,$E5,$1F,$15,$FE,$E0,$9C
	.dc.b	$D7,$ED,$27,$3A,$F6,$9F,$DC,$46,$1D,$98,$AE,$57,$D5,$E9,$7F,$4E
	.dc.b	$9F,$49,$FC,$02,$30,$DD,$3C,$15,$C1,$B8,$FC,$7C,$58,$72,$70,$D7
	.dc.b	$E5,$3E,$5C,$3C,$17,$ED,$D7,$BD,$7D,$E3,$9E,$3F,$C4,$A3,$78,$67
	.dc.b	$0E,$58,$FC,$3A,$F3,$5F,$C2,$6B,$E4,$C3,$D0,$FC,$3B,$FC,$5F,$0F
	.dc.b	$EE,$7C,$4F,$E3,$F1,$FE,$B2,$E0,$FE,$ED,$5F,$EB,$BC,$BE,$66,$7F
	.dc.b	$1F,$DE,$DF,$0F,$D3,$6D,$7D,$26,$E5,$FC,$2B,$8B,$33,$F1,$A6,$E1
	.dc.b	$FE,$99,$8E,$D1,$C3,$8D,$5A,$8F,$E0,$B5,$BA,$FD,$7B,$5F,$E1,$8D
	.dc.b	$7C,$87,$F9,$C7,$ED,$9E,$17,$91,$B1,$E0,$3F,$AE,$FD,$A3,$ED,$BD
	.dc.b	$7D,$08,$C4,$EA,$7D,$BF,$6C,$FA,$ED,$D6,$AB,$FC,$0C,$D7,$FB,$7B
	.dc.b	$5F,$E6,$8D,$7F,$A0,$3C,$78,$BF,$DC,$75,$FE,$4A,$9F,$F4,$53,$5F
	.dc.b	$EC,$53,$BE,$0F,$EA,$5D,$EF,$84,$FA,$7D,$5F,$EB,$06,$C6,$6F,$FD
	.dc.b	$29,$E0,$7F,$33,$CB,$F8,$F5,$EE,$7F,$B2,$56,$DF,$E7,$11,$96,$C6
	.dc.b	$5F,$EF,$DF,$0D,$F2,$3C,$53,$4F,$C1,$F8,$7F,$E4,$B9,$FF,$14,$46
	.dc.b	$5A,$7F,$80,$3C,$7C,$3F,$DF,$EB,$9B,$A3,$FD,$31,$D0,$BE,$91,$BB
	.dc.b	$FE,$66,$FF,$6B,$AE,$2F,$F8,$E5,$AF,$A0,$FF,$13,$F9,$FF,$80,$C0
	.dc.b	$D3,$F0,$7E,$47,$E5,$27,$FC,$47,$3A,$78,$90,$FF,$17,$CF,$8B,$1F
	.dc.b	$F6,$0F,$DA,$FE,$93,$CB,$34,$C7,$D6,$5B,$FF,$60,$45,$7D,$F2,$9F
	.dc.b	$9D,$10,$FF,$29,$1B,$DF,$FA,$C7,$3F,$F9,$76,$3E,$1B,$5F,$F1,$B1
	.dc.b	$8D,$3F,$B8,$AF,$B4,$FE,$6B,$5F,$D8,$63,$1E,$2F,$EE,$A7,$3C,$BF
	.dc.b	$E2,$6F,$8A,$F2,$57,$F8,$A6,$F8,$56,$A3,$FC,$59,$3F,$B0,$7F,$D4
	.dc.b	$4D,$9D,$DF,$F0,$6F,$B2,$7D,$A5,$25,$F5,$C7,$FC,$6B,$F6,$8F,$93
	.dc.b	$CC,$FF,$AB,$61,$F3,$3F,$C4,$1B,$1C,$4F,$F1,$F6,$1D,$9C,$0F,$A0
	.dc.b	$FF,$3C,$B8,$6B,$FE,$B9,$6E,$7F,$17,$D8,$F8,$33,$9F,$47,$FC,$37
	.dc.b	$5C,$23,$FC,$40,$D7,$FC,$9A,$D7,$FC,$A0,$F4,$CF,$FC,$A3,$5F,$15
	.dc.b	$BE,$27,$F2,$A9,$D5,$70,$EF,$C1,$1F,$5F,$DA,$3E,$17,$11,$FE,$D1
	.dc.b	$87,$C8,$FF,$34,$6C,$7B,$EB,$F9,$CF,$C3,$7F,$84,$E3,$E4,$75,$FF
	.dc.b	$BC,$7D,$BC,$9A,$F5,$64,$EA,$93,$A8,$C4,$FF,$42,$FB,$47,$C7,$D2
	.dc.b	$4F,$D2,$FA,$8F,$F0,$1B,$F5,$27,$6F,$EF,$1A,$FF,$8C,$1A,$FF,$93
	.dc.b	$5A,$FF,$9E,$9D,$F1,$FE,$6E,$7F,$71,$3F,$EF,$36,$C7,$E7,$E7,$7C
	.dc.b	$3F,$EB,$7F,$2F,$F7,$1E,$EC,$C3,$E1,$F8,$3F,$D0,$5B,$39,$7F,$6F
	.dc.b	$3F,$BF,$8F,$81,$C7,$27,$D8,$FF,$64,$CB,$F9,$E6,$5E,$18,$C5,$CA
	.dc.b	$D9,$F1,$7C,$1E,$1C,$F2,$30,$43,$D8,$E5,$F7,$47,$73,$17,$F0,$E7
	.dc.b	$13,$EC,$FB,$1F,$E2,$DE,$E4,$8C,$10,$FE,$26,$E5,$FE,$2B,$1F,$C7
	.dc.b	$7E,$9F,$F8,$F7,$DA,$49,$F8,$F1,$97,$E4,$C6,$5F,$E3,$B1,$FE,$06
	.dc.b	$FB,$47,$C6,$D2,$4F,$ED,$BE,$07,$F9,$92,$CF,$F8,$B5,$70,$8F,$E8
	.dc.b	$2D,$7F,$CE,$AD,$7F,$D2,$4D,$7F,$D7,$EE,$E3,$F9,$04,$BE,$2B,$FC
	.dc.b	$32,$DA,$FE,$0F,$5B,$E3,$FF,$64,$7A,$0F,$1B,$F8,$1E,$3F,$E8,$98
	.dc.b	$C7,$07,$F8,$99,$B8,$7E,$2B,$DE,$FF,$12,$FD,$93,$ED,$3F,$D4,$6D
	.dc.b	$9F,$EB,$6D,$7F,$D1,$ED,$7F,$DA,$FD,$CE,$9F,$F2,$86,$5F,$E9,$98
	.dc.b	$C7,$C0,$FF,$51,$3C,$5F,$68,$F9,$DF,$26,$CE,$6F,$1F,$BD,$33,$FE
	.dc.b	$51,$BF,$B6,$71,$1E,$29,$F3,$91,$FF,$51,$CB,$0F,$91,$1F,$5F,$87
	.dc.b	$E3,$C6,$34,$F8,$CD,$8F,$F7,$8E,$DF,$E3,$F9,$FE,$A5,$E1,$BE,$43
	.dc.b	$FC,$96,$DC,$3F,$9C,$B5,$EC,$7F,$BE,$FC,$CF,$AC,$FF,$BF,$63,$4F
	.dc.b	$FB,$4D,$AF,$FC,$00,$6B,$B9,$FF,$6A,$C7,$5A,$FB,$9F,$66,$FA,$EF
	.dc.b	$B8,$DB,$F1,$DF,$9A,$64,$79,$9F,$C1,$7D,$A3,$E6,$73,$3F,$EC,$AC
	.dc.b	$79,$36,$BC,$24,$7E,$9E,$0F,$9E,$F0,$9C,$B2,$FE,$F1,$F6,$2F,$B3
	.dc.b	$FF,$38,$B6,$FF,$AF,$B5,$F8,$FC,$3F,$ED,$1F,$17,$75,$FE,$F4,$AD
	.dc.b	$57,$FB,$E5,$A8,$8E,$1B,$1F,$9F,$EE,$7D,$A5,$3A,$CF,$8B,$D0,$F2
	.dc.b	$91,$FC,$7F,$99,$FE,$11,$CF,$D7,$3E,$32,$4F,$F8,$FB,$1F,$EB,$CD
	.dc.b	$8F,$06,$92,$3B,$1F,$03,$1F,$86,$47,$F3,$BF,$1D,$FC,$ED,$FC,$48
	.dc.b	$B4,$FC,$32,$47,$85,$AF,$CA,$F2,$BF,$DF,$D8,$FF,$01,$6F,$3F,$C7
	.dc.b	$9E,$95,$F5,$3E,$D2,$47,$D4,$7F,$83,$62,$F2,$33,$3C,$4F,$9F,$3F
	.dc.b	$54,$8E,$47,$81,$FE,$65,$E6,$FF,$5A,$E2,$FB,$4F,$EF,$95,$FA,$68
	.dc.b	$C7,$07,$F9,$DA,$A4,$FF,$89,$E7,$2D,$AE,$3A,$FF,$0D,$F8,$3F,$CD
	.dc.b	$4F,$F8,$B7,$DE,$79,$7F,$9A,$AF,$F2,$9B,$70,$FC,$44,$EB,$A7,$FD
	.dc.b	$0B,$CF,$FE,$B0,$AD,$60,$AD,$3D,$63,$3A,$FF,$A0,$6B,$FD,$9B,$F6
	.dc.b	$6F,$B4,$A4,$BF,$F0,$CF,$D9,$FF,$29,$D3,$FC,$B9,$5B,$9F,$F2,$0B
	.dc.b	$5F,$C0,$B6,$3F,$95,$4F,$1D,$3E,$D4,$EE,$BE,$A3,$5C,$DF,$FA,$49
	.dc.b	$46,$5F,$D4,$2F,$6B,$F0,$9C,$9F,$EA,$F8,$C4,$FE,$B6,$31,$3A,$EE
	.dc.b	$FD,$86,$9F,$EA,$1F,$7B,$17,$D2,$AB,$E0,$6E,$1D,$21,$87,$FB,$0E
	.dc.b	$F9,$47,$7C,$3D,$BC,$7F,$E3,$28,$F6,$B5,$CF,$C3,$FF,$AB,$B0,$FE
	.dc.b	$34,$FF,$AC,$3E,$37,$F0,$3E,$05,$E2,$34,$97,$F9,$7B,$FF,$38,$73
	.dc.b	$F3,$E3,$FC,$AC,$FC,$E6,$BF,$21,$AB,$E3,$BF,$CB,$7C,$5E,$98,$FE
	.dc.b	$29,$18,$7D,$FA,$FC,$8E,$77,$CC,$2F,$85,$73,$F7,$B0,$F0,$39,$B8
	.dc.b	$E2,$FF,$72,$63,$D5,$F9,$58,$74,$BE,$73,$9C,$7F,$80,$2F,$B4,$EB
	.dc.b	$D1,$FE,$AD,$8C,$57,$E5,$BE,$1F,$2F,$0A,$B9,$FB,$1A,$FC,$17,$3C
	.dc.b	$AE,$9F,$F6,$6D,$7F,$A7,$30,$EB,$7F,$C7,$1B,$8F,$87,$0E,$37,$C2
	.dc.b	$B5,$1E,$0E,$A7,$DA,$7F,$4F,$C7,$F7,$CD,$AF,$F7,$9E,$BF,$9B,$9C
	.dc.b	$F8,$7C,$6B,$59,$7F,$B5,$1B,$3F,$4F,$AB,$55,$EA,$B7,$D9,$FC,$0A
	.dc.b	$7D,$A3,$EC,$F5,$39,$8C,$4F,$FA,$5A,$B8,$BF,$C6,$63,$39,$F9,$AB
	.dc.b	$85,$FC,$AF,$B4,$76,$D3,$E7,$F5,$FF,$93,$7D,$CC,$65,$FE,$C6,$AD
	.dc.b	$B6,$B7,$BF,$F9,$62,$30,$7F,$B3,$3E,$CF,$C3,$9F,$F3,$96,$BF,$D0
	.dc.b	$DA,$FF,$6B,$6A,$7D,$33,$FE,$F3,$6C,$FF,$94,$21,$F6,$9F,$84,$9C
	.dc.b	$D7,$3C,$F0,$79,$A6,$DF,$97,$E1,$3D,$FE,$E3,$ED,$39,$46,$1B,$3B
	.dc.b	$3F,$C8,$7B,$7F,$95,$DC,$63,$FD,$A3,$2F,$F5,$EC,$63,$3F,$F3,$15
	.dc.b	$6E,$B5,$87,$9F,$BB,$C3,$EC,$DC,$3F,$C9,$F9,$FF,$9A,$DF,$95,$39
	.dc.b	$BF,$F0,$BE,$0E,$DC,$65,$F1,$6A,$DF,$F9,$D2,$B4,$F2,$3F,$DF,$FF
	.dc.b	$60,$F8,$DF,$BF,$F6,$7F,$06,$73,$18,$DF,$F5,$F5,$B9,$C9,$FE,$21
	.dc.b	$C7,$F8,$3E,$6F,$B4,$7C,$9F,$0C,$65,$F0,$AF,$F0,$6D,$6A,$31,$7C
	.dc.b	$B0,$FF,$3A,$47,$F9,$C3,$2F,$85,$7F,$E1,$49,$CB,$FE,$38,$8C,$D7
	.dc.b	$1F,$7F,$CE,$7F,$06,$FB,$4F,$C1,$D2,$F8,$D6,$F0,$F8,$77,$BE,$5E
	.dc.b	$31,$FE,$80,$F9,$27,$DA,$7E,$91,$B1,$F2,$BC,$4C,$4F,$C6,$64,$7C
	.dc.b	$53,$FC,$41,$E4,$70,$CF,$D5,$3F,$C6,$A3,$19,$7C,$18,$C5,$F2,$AF
	.dc.b	$F3,$DF,$8D,$FE,$5F,$F8,$8F,$E3,$FE,$17,$FA,$D7,$2F,$37,$AD,$89
	.dc.b	$FA,$FF,$1D,$27,$C5,$87,$F9,$5F,$4F,$F1,$DC,$7E,$0A,$7C,$57,$23
	.dc.b	$78,$FF,$9B,$7E,$63,$18,$7F,$91,$3E,$C9,$F6,$7F,$80,$8F,$F1,$E4
	.dc.b	$62,$FF,$C6,$50,$8C,$3F,$E3,$56,$BF,$E3,$B7,$3B,$FE,$29,$B3,$FE
	.dc.b	$26,$FA,$F7,$C3,$FF,$40,$9C,$DF,$F9,$C5,$AF,$7C,$3F,$9C,$57,$0B
	.dc.b	$F8,$33,$97,$DB,$BB,$5F,$97,$C3,$F3,$1B,$3A,$6D,$7D,$2A,$E1,$3C
	.dc.b	$AB,$E9,$CF,$D4,$F4,$AC,$F6,$6C,$71,$6E,$9F,$ED,$5B,$47,$F5,$EF
	.dc.b	$B4,$FF,$03,$F8,$77,$BF,$DF,$18,$C7,$ED,$BF,$E0,$CD,$75,$5B,$6C
	.dc.b	$EA,$B8,$9F,$C5,$32,$FF,$62,$DF,$19,$D5,$F1,$7E,$3E,$DF,$EB,$56
	.dc.b	$FA,$FF,$DC,$B2,$7D,$A7,$EC,$27,$F0,$33,$F5,$48,$FE,$42,$77,$E2
	.dc.b	$E3,$2F,$CD,$FC,$63,$EC,$FF,$BC,$AF,$B4,$B1,$86,$E1,$F9,$49,$D3
	.dc.b	$67,$4F,$DE,$FB,$37,$C2,$FE,$57,$0F,$F5,$B4,$62,$7F,$2A,$D8,$FF
	.dc.b	$44,$35,$FF,$47,$39,$E3,$FE,$3A,$D9,$FC,$A7,$8D,$FE,$12,$8C,$57
	.dc.b	$6B,$C1,$F9,$29,$B5,$FA,$CF,$98,$FF,$08,$BF,$E5,$A7,$FB,$45,$7E
	.dc.b	$F9,$CB,$FE,$9A,$73,$87,$F8,$D3,$AD,$1A,$6F,$A1,$FE,$22,$8C,$1C
	.dc.b	$2F,$FD,$61,$97,$F0,$7C,$3F,$D7,$AD,$C3,$FC,$7F,$BB,$EC,$FF,$DF
	.dc.b	$97,$F2,$FC,$0F,$71,$1F,$B9,$1A,$C3,$9C,$FF,$8D,$3E,$C1,$F2,$DF
	.dc.b	$46,$51,$8B,$FE,$0A,$D8,$FF,$5C,$35,$FF,$5F,$39,$C7,$FA,$6B,$67
	.dc.b	$FD,$E1,$F6,$4F,$B2,$FF,$0F,$DF,$FB,$62,$B4,$FF,$BD,$5A,$F8,$B0
	.dc.b	$FF,$25,$D7,$1F,$5B,$D1,$1F,$E4,$CE,$17,$D1,$D3,$F4,$1A,$FF,$8F
	.dc.b	$70,$FF,$72,$CE,$B3,$FE,$1A,$D8,$CD,$70,$F6,$B9,$57,$F1,$2F,$09
	.dc.b	$F5,$BF,$E5,$A6,$C7,$E9,$FE,$CA,$E1,$FE,$F2,$9B,$C1,$FD,$B9,$B5
	.dc.b	$FE,$EB,$6C,$FF,$0A,$AA,$FF,$1D,$BF,$F2,$D7,$D2,$BE,$CA,$9F,$B0
	.dc.b	$F9,$EB,$8F,$FB,$66,$74,$D9,$D3,$4F,$D9,$4E,$2F,$F7,$7F,$47,$73
	.dc.b	$FC,$B7,$6A,$B1,$7F,$EB,$88,$56,$2F,$FD,$BD,$B3,$E8,$3F,$CE,$0D
	.dc.b	$BF,$F7,$67,$D3,$FF,$8F,$9F,$F7,$74,$66,$99,$7F,$A6,$3E,$13,$EC
	.dc.b	$E1,$C8,$FF,$3D,$61,$C6,$B7,$87,$FB,$CB,$0E,$58,$71,$D7,$FB,$2D
	.dc.b	$68,$FF,$79,$38,$8F,$F8,$01,$F6,$5F,$33,$D0,$98,$7C,$97,$FC,$F5
	.dc.b	$18,$BE,$59,$B0,$E5,$69,$EB,$7C,$23,$91,$23,$8C,$F0,$5E,$03,$E1
	.dc.b	$18,$F7,$CB,$FD,$73,$F0,$5F,$ED,$0E,$9F,$F5,$4B,$72,$FE,$FB,$F5
	.dc.b	$3F,$EC,$97,$FC,$0B,$19,$CB,$E0,$36,$F9,$F9,$BF,$AC,$CD,$87,$1B
	.dc.b	$DC,$FE,$C3,$67,$D9,$7E,$07,$2F,$75,$49,$CF,$2F,$F3,$56,$CE,$A7
	.dc.b	$E2,$46,$1C,$37,$2F,$BD,$7D,$67,$FC,$4B,$F5,$FF,$CA,$2B,$57,$CA
	.dc.b	$B5,$7C,$30,$FF,$63,$DE,$E7,$35,$E1,$F4,$BF,$8C,$63,$ED,$3C,$85
	.dc.b	$35,$1F,$E1,$E7,$E1,$61,$FE,$23,$9E,$BF,$65,$FE,$D6,$9C,$D7,$07
	.dc.b	$F1,$87,$FB,$D9,$FF,$17,$E0,$9E,$15,$27,$FC,$33,$5C,$8F,$C0,$D6
	.dc.b	$E3,$33,$DE,$5F,$1B,$F8,$DF,$B1,$E9,$1F,$88,$FB,$27,$CC,$6E,$B5
	.dc.b	$97,$E4,$F8,$55,$AC,$BC,$38,$74,$18,$F4,$7F,$35,$E4,$A7,$9C,$7D
	.dc.b	$BC,$6B,$A8,$AF,$E8,$62,$CF,$18,$EB,$8F,$D0,$9C,$D7,$09,$F1,$5F
	.dc.b	$85,$FF,$14,$C6,$67,$71,$C8,$F1,$14,$F1,$14,$F1,$16,$FC,$55,$D6
	.dc.b	$3E,$07,$1B,$D0,$F2,$47,$FB,$A1,$B1,$98,$EE,$D5,$F2,$8F,$89,$FE
	.dc.b	$2F,$1C,$4F,$E0,$51,$9E,$1F,$E0,$95,$C6,$31,$3B,$39,$D7,$29,$D3
	.dc.b	$D3,$0F,$A9,$4F,$A5,$FD,$8C,$61,$BA,$FE,$D2,$7A,$CF,$0A,$E5,$39
	.dc.b	$AD,$E1,$E6,$6B,$E2,$C7,$EE,$E1,$CA,$B8,$4E,$2B,$73,$98,$F7,$DF
	.dc.b	$67,$A1,$B3,$9B,$97,$B1,$F6,$F7,$FC,$AF,$87,$FC,$33,$1F,$F1,$27
	.dc.b	$07,$F6,$CA,$FF,$3F,$E1,$DD,$FA,$3E,$E6,$F8,$7E,$03,$63,$E8,$B7
	.dc.b	$17,$23,$93,$55,$C8,$EB,$58,$C3,$FD,$2B,$1F,$BA,$9C,$4C,$8D,$99
	.dc.b	$29,$FC,$32,$78,$B6,$3F,$86,$46,$23,$F9,$47,$D7,$BE,$C7,$D1,$D2
	.dc.b	$8C,$4F,$FA,$FA,$31,$3E,$BF,$B1,$7D,$67,$0B,$DD,$FF,$54,$6E,$FF
	.dc.b	$DA,$5B,$B7,$F0,$3B,$E3,$5D,$67,$8B,$E4,$C3,$FD,$7F,$8F,$82,$3F
	.dc.b	$93,$B5,$FE,$79,$CA,$EC,$7C,$77,$D1,$8C,$36,$33,$3F,$D5,$8C,$44
	.dc.b	$7F,$AB,$C6,$9F,$F7,$2F,$77,$F4,$9D,$3F,$BD,$7C,$17,$C6,$F0,$CD
	.dc.b	$3E,$F7,$87,$FE,$3E,$BF,$F1,$04,$65,$A7,$FB,$8A,$34,$FF,$72,$8D
	.dc.b	$34,$D7,$8B,$8B,$F0,$C7,$23,$FD,$F3,$5C,$A3,$FC,$2D,$2F,$3B,$ED
	.dc.b	$B6,$BF,$C3,$B1,$F7,$BC,$7F,$C9,$46,$08,$7F,$83,$27,$83,$FE,$0E
	.dc.b	$8D,$BF,$EA,$EF,$96,$FC,$E7,$96,$69,$8F,$AB,$B3,$FE,$AD,$8A,$FF
	.dc.b	$AD,$A3,$E5,$30,$FF,$23,$1A,$C7,$FC,$50,$E4,$FF,$0B,$99,$E5,$FF
	.dc.b	$00,$FD,$8F,$F2,$39,$FE,$7D,$28,$C4,$EB,$C3,$FD,$7F,$C4,$30,$FD
	.dc.b	$8D,$6A,$FF,$C5,$0D,$79,$57,$09,$FD,$BF,$D8,$3E,$76,$92,$BF,$C6
	.dc.b	$19,$78,$BE,$C5,$F2,$59,$9F,$EE,$B7,$C5,$DE,$9F,$E1,$A8,$C4,$7F
	.dc.b	$7A,$AE,$2E,$8F,$F1,$E7,$CE,$7F,$98,$5C,$07,$0F,$F8,$09,$B8,$7C
	.dc.b	$4F,$5F,$E0,$4F,$09,$CE,$1F,$E2,$A6,$C7,$6F,$3F,$F8,$AD,$70,$8E
	.dc.b	$91,$FE,$15,$9E,$F1,$FE,$2D,$6B,$CE,$35,$D5,$FC,$7A,$7F,$59,$87
	.dc.b	$D6,$6E,$1F,$E3,$AF,$B1,$FF,$76,$E1,$E5,$BE,$04,$7D,$F1,$A7,$FD
	.dc.b	$B9,$E6,$FF,$2D,$F8,$07,$B3,$4C,$44,$9F,$F7,$34,$65,$FF,$74,$47
	.dc.b	$F3,$2F,$B1,$7C,$4D,$24,$FC,$EF,$A5,$78,$58,$8F,$F1,$7D,$70,$7E
	.dc.b	$9B,$5F,$F1,$9C,$F0,$73,$5F,$E3,$D9,$D4,$7F,$95,$5A,$FF,$97,$1A
	.dc.b	$FF,$9D,$3A,$BF,$93,$BF,$E8,$9F,$31,$1F,$E9,$D8,$FC,$34,$63,$D9
	.dc.b	$FD,$AF,$93,$27,$35,$B7,$93,$78,$BF,$D5,$B3,$F0,$F8,$A5,$7B,$1F
	.dc.b	$EA,$F9,$FF,$BC,$A3,$04,$3E,$1C,$6B,$D7,$DF,$57,$0E,$59,$5F,$2E
	.dc.b	$32,$FC,$D8,$CB,$F2,$23,$2F,$CA,$8C,$E3,$FB,$EC,$3F,$28,$FF,$B1
	.dc.b	$3E,$BF,$F8,$9E,$12,$FD,$F1,$FC,$58,$C3,$F0,$23,$0F,$C4,$FA,$2F
	.dc.b	$E3,$7F,$4D,$23,$04,$3E,$3B,$97,$E4,$C7,$F7,$AF,$B1,$7C,$45,$24
	.dc.b	$FE,$DB,$DE,$FE,$53,$97,$F1,$6A,$DC,$FE,$E9,$AF,$F9,$E9,$B1,$FC
	.dc.b	$72,$78,$F5,$FF,$A8,$27,$75,$FE,$9D,$6B,$9B,$FF,$45,$46,$23,$FB
	.dc.b	$37,$83,$F2,$2B,$FD,$8D,$E6,$7F,$94,$78,$3F,$C3,$0D,$AF,$C3,$FA
	.dc.b	$FF,$E1,$6F,$AF,$7D,$8D,$24,$FF,$15,$AE,$0F,$E3,$3D,$AF,$E8,$A4
	.dc.b	$75,$F6,$2F,$99,$EF,$A3,$93,$B7,$B1,$2F,$A4,$E5,$FE,$3F,$9F,$F8
	.dc.b	$73,$8B,$F9,$FB,$77,$FF,$61,$B6,$7D,$F8,$7B,$32,$7C,$5D,$24,$F4
	.dc.b	$CB,$FC,$67,$E2,$BE,$C2,$92,$7F,$97,$FD,$8F,$D4,$CF,$FD,$A3,$5B
	.dc.b	$9F,$E6,$6D,$7F,$D6,$6D,$D3,$FA,$AD,$71,$9F,$F7,$B4,$E6,$BE,$9F
	.dc.b	$D8,$3E,$B3,$E9,$B6,$BA,$E3,$E6,$99,$1E,$07,$EC,$FD,$8B,$E5,$73
	.dc.b	$3E,$29,$B9,$FF,$A8,$F9,$17,$BA,$E4,$7D,$BC,$5F,$35,$E2,$38,$E7
	.dc.b	$FE,$F0,$FA,$F7,$D8,$FF,$96,$9B,$7F,$D5,$1A,$B2,$F0,$79,$7F,$D5
	.dc.b	$E3,$55,$C4,$8F,$E2,$E3,$4F,$FB,$72,$B9,$D7,$A7,$B9,$F6,$3F,$98
	.dc.b	$6D,$74,$3C,$5F,$2F,$E5,$23,$D7,$2A,$F5,$8B,$7E,$69,$A7,$E9,$AB
	.dc.b	$E1,$D7,$E7,$DA,$F1,$3E,$03,$F0,$C8,$FE,$97,$E2,$FF,$9C,$D7,$F8
	.dc.b	$5B,$15,$6A,$BD,$CF,$EC,$AB,$77,$F3,$E1,$30,$82,$B9,$B8,$36,$E1
	.dc.b	$C3,$5E,$2F,$FB,$E1,$FB,$1F,$19,$F6,$CF,$17,$89,$63,$24,$3E,$E5
	.dc.b	$FF,$0A,$C1,$3F,$E6,$2E,$3F,$F5,$14,$BE,$C5,$B6,$1B,$87,$F9,$8E
	.dc.b	$B8,$3F,$E1,$FA,$E1,$5B,$33,$1E,$01,$F0,$6F,$1B,$F4,$D1,$98,$C7
	.dc.b	$37,$FA,$72,$73,$BB,$0F,$97,$E7,$3E,$03,$F1,$F3,$9C,$7A,$DF,$0B
	.dc.b	$FA,$38,$F4,$9F,$F5,$24,$E5,$61,$CF,$CC,$FF,$5E,$FD,$83,$EC,$69
	.dc.b	$2B,$FB,$BF,$D8,$FF,$2C,$7F,$C7,$73,$B6,$CE,$DF,$F0,$04,$ED,$F8
	.dc.b	$42,$78,$38,$6E,$7F,$E6,$46,$E7,$F0,$2B,$94,$E7,$D9,$FC,$67,$1F
	.dc.b	$FA,$86,$D1,$8E,$FF,$C8,$46,$31,$FC,$9D,$7F,$AB,$34,$F9,$5E,$EE
	.dc.b	$E7,$F2,$F5,$B5,$5D,$70,$D4,$7A,$5E,$17,$CA,$3A,$E1,$E0,$E3,$FF
	.dc.b	$15,$47,$E8,$30,$FE,$3D,$EB,$FE,$7E,$3E,$14,$E9,$E1,$19,$F8,$AF
	.dc.b	$C3,$78,$37,$8B,$FD,$91,$5F,$EE,$76,$BF,$AC,$9D,$BF,$AB,$C3,$FC
	.dc.b	$DD,$CD,$DD,$B1,$9E,$3F,$CD,$BE,$56,$BF,$8E,$E5,$FC,$87,$1F,$B2
	.dc.b	$3F,$0F,$1F,$EF,$18,$C6,$DF,$93,$7F,$25,$C1,$F3,$67,$55,$F2,$B7
	.dc.b	$E9,$5A,$7B,$7C,$45,$9E,$B3,$70,$FC,$F4,$DF,$CF,$3F,$9C,$CF,$F9
	.dc.b	$21,$B5,$FE,$4D,$5F,$49,$FA,$3E,$5F,$D2,$60,$9D,$C6,$1B,$7F,$AC
	.dc.b	$6D,$FE,$91,$E7,$1D,$FC,$1E,$8E,$DB,$3B,$6D,$F7,$C2,$DF,$48,$F1
	.dc.b	$7A,$78,$33,$E0,$DC,$7D,$91,$71,$F6,$4D,$CB,$D1,$F7,$3D,$3E,$CF
	.dc.b	$AB,$70,$F8,$8D,$BF,$88,$DA,$F8,$7E,$D7,$FA,$12,$BF,$CD,$BD,$1E
	.dc.b	$88,$7A,$16,$24,$E8,$FF,$39,$C6,$30,$FF,$3E,$7E,$A9,$FE,$11,$9F
	.dc.b	$F3,$DB,$63,$3F,$53,$F1,$23,$18,$FD,$46,$D7,$F1,$C6,$D7,$F1,$45
	.dc.b	$5C,$7C,$C9,$5D,$9E,$91,$8A,$F1,$F9,$3D,$CF,$07,$89,$6F,$CB,$16
	.dc.b	$7D,$33,$71,$F2,$4D,$AF,$24,$DA,$F2,$7D,$E7,$AC,$F6,$7D,$F5,$AF
	.dc.b	$83,$16,$7E,$2C,$DC,$3E,$04,$DA,$F8,$13,$6B,$E0,$7D,$EF,$C7,$3E
	.dc.b	$47,$9B,$6B,$E6,$C5,$9F,$A3,$37,$0F,$99,$36,$BE,$64,$DA,$F9,$9F
	.dc.b	$55,$F4,$CF,$A9,$E4,$5A,$FB,$31,$67,$EE,$CD,$C3,$EC,$4D,$AF,$B1
	.dc.b	$36,$BE,$C7,$D9,$7D,$FF,$84,$FC,$03,$6B,$F0,$AE,$1F,$C3,$F9,$16
	.dc.b	$BF,$16,$DC,$7F,$0C,$D9,$FC,$37,$D1,$7E,$2C,$FC,$67,$CF,$58,$FC
	.dc.b	$8B,$87,$F2,$7F,$23,$70,$FC,$BF,$97,$F9,$38,$FC,$A3,$67,$FC,$2F
	.dc.b	$8F,$F6,$BF,$AB,$79,$9F,$D8,$E7,$27,$F8,$72,$33,$B7,$F0,$27,$7C
	.dc.b	$36,$B8,$65,$FE,$01,$9D,$DF,$28,$46,$1D,$4A,$B9,$5C,$BE,$17,$BC
	.dc.b	$3C,$D7,$B9,$CB,$E0,$8F,$D7,$79,$BF,$B1,$3F,$65,$E4,$DA,$FD,$BC
	.dc.b	$5C,$BF,$77,$36,$BF,$6D,$5F,$B9,$3F,$75,$E3,$57,$F7,$E5,$AF,$E0
	.dc.b	$71,$67,$F9,$8F,$E4,$DF,$C8,$4F,$E5,$73,$FE,$23,$F1,$11,$9F,$1B
	.dc.b	$F9,$4B,$70,$F8,$D3,$AF,$1F,$35,$A6,$CE,$AB,$85,$FF,$8A,$FF,$55
	.dc.b	$7C,$AF,$F1,$DF,$6F,$F9,$BF,$A6,$9C,$E1,$DB,$C6,$56,$A7,$38,$73
	.dc.b	$C3,$CB,$E4,$E5,$FE,$83,$0F,$A9,$4F,$F4,$4D,$DF,$79,$FE,$A5,$CF
	.dc.b	$F0,$17,$BA,$D4,$E6,$3F,$C4,$B7,$CA,$BF,$C0,$CD,$8F,$F3,$33,$5F
	.dc.b	$F3,$5B,$5F,$CA,$72,$39,$3F,$24,$7E,$4F,$92,$DF,$E1,$A2,$D7,$E4
	.dc.b	$26,$E5,$F8,$C9,$B5,$D2,$6D,$7F,$92,$38,$1E,$5F,$D8,$3E,$CF,$99
	.dc.b	$6F,$EA,$45,$AF,$C0,$52,$B1,$8F,$F9,$66,$5F,$52,$FA,$BF,$ED,$BE
	.dc.b	$A3,$EA,$7F,$D5,$39,$FC,$AA,$D3,$FA,$B9,$D7,$95,$F5,$27,$38,$7F
	.dc.b	$94,$1B,$8F,$F9,$9E,$BA,$79,$1B,$CB,$F7,$56,$FA,$4A,$63,$E9,$2E
	.dc.b	$7F,$B2,$9D,$36,$74,$D2,$47,$F0,$13,$65,$9E,$6F,$F9,$93,$F5,$27
	.dc.b	$C1,$FC,$0D,$9E,$2F,$E3,$A7,$38,$7B,$E7,$2F,$E3,$23,$37,$CE,$F9
	.dc.b	$78,$BC,$5F,$CB,$FE,$AD,$FE,$01,$9F,$F4,$A6,$CF,$AB,$FF,$5B,$E3
	.dc.b	$F9,$6A,$D7,$07,$C7,$9C,$DF,$F9,$EA,$73,$5F,$EB,$D6,$C7,$FA,$A3
	.dc.b	$81,$F5,$DE,$0F,$5B,$F7,$BE,$BC,$C3,$F7,$1F,$73,$70,$FC,$2C,$DC
	.dc.b	$FF,$84,$CD,$AF,$F5,$0F,$43,$EC,$FF,$83,$CF,$F0,$86,$CF,$F0,$36
	.dc.b	$E1,$F9,$06,$DF,$F9,$05,$FF,$1A,$9F,$E3,$6A,$B5,$FE,$4D,$8B,$5F
	.dc.b	$E4,$D9,$B5,$FE,$49,$FA,$EF,$F1,$E9,$FE,$3E,$E0,$B8,$7F,$95,$62
	.dc.b	$DF,$F9,$56,$6D,$71,$AD,$1F,$E5,$B8,$CB,$FE,$53,$AB,$5F,$E5,$39
	.dc.b	$BE,$4F,$F9,$6F,$E3,$BF,$C9,$A7,$F9,$3B,$E7,$EE,$1F,$E6,$58,$B7
	.dc.b	$FE,$65,$9B,$8F,$F9,$87,$E1,$FF,$CA,$A7,$F9,$5B,$E2,$6E,$1F,$E6
	.dc.b	$D8,$B7,$FE,$6D,$9B,$5C,$6B,$47,$F9,$CE,$32,$FF,$9A,$FA,$FF,$CC
	.dc.b	$27,$F9,$8B,$AE,$C7,$F9,$E6,$2D,$FF,$9E,$66,$DF,$F9,$D6,$6D,$7F
	.dc.b	$9D,$66,$D7,$F9,$D6,$6E,$7F,$E7,$D9,$CC,$FF,$B1,$07,$D6,$7E,$ED
	.dc.b	$13,$A3,$FD,$D1,$19,$8E,$0D,$7F,$E0,$0E,$1C,$23,$11,$FE,$C4,$3A
	.dc.b	$E1,$D1,$E9,$3B,$8C,$47,$FB,$40,$D9,$BA,$F0,$C6,$A3,$A5,$78,$67
	.dc.b	$31,$CB,$E0,$FA,$3C,$6E,$BE,$8B,$EF,$5C,$23,$06,$CF,$09,$83,$CE
	.dc.b	$60,$F3,$5F,$86,$BA,$C7,$A7,$97,$FD,$0A,$F8,$C3,$87,$E8,$35,$E1
	.dc.b	$F0,$1F,$17,$E1,$5F,$26,$DB,$E9,$DF,$19,$CC,$62,$B8,$56,$EF,$BC
	.dc.b	$62,$75,$1F,$72,$B8,$BE,$E5,$F2,$5F,$57,$2F,$BD,$18,$9F,$92,$F0
	.dc.b	$8C,$E1,$DD,$DD,$71,$D3,$D4,$DB,$F0,$56,$E7,$C9,$1E,$19,$D2,$BF
	.dc.b	$03,$E5,$AE,$0F,$2C,$3C,$75,$C2,$B7,$87,$92,$7C,$F7,$C2,$FA,$5E
	.dc.b	$CE,$11,$D3,$7F,$F7,$07,$D2,$FE,$56,$FF,$DC,$0D,$D7,$FD,$C0,$DC
	.dc.b	$3F,$DC,$1F,$21,$6F,$8C,$7F,$BB,$E7,$1F,$0C,$6D,$61,$93,$B3,$FD
	.dc.b	$ED,$18,$7F,$DE,$D1,$70,$FF,$79,$4F,$4F,$9D,$95,$FE,$FD,$7B,$FD
	.dc.b	$56,$23,$FD,$FD,$16,$BF,$DF,$53,$72,$FF,$7D,$4D,$AF,$F7,$D4,$FA
	.dc.b	$18,$7F,$93,$31,$FD,$AD,$AF,$F8,$2E,$49,$71,$70,$BF,$F2,$C4,$70
	.dc.b	$AD,$B6,$15,$C6,$BE,$57,$E3,$1F,$E1,$B9,$FF,$2D,$B5,$FF,$31,$37
	.dc.b	$0F,$14,$72,$8F,$ED,$2F,$0F,$23,$99,$1E,$7E,$0D,$79,$C5,$AE,$78
	.dc.b	$2F,$9D,$68,$E8,$43,$C7,$7E,$8F,$9C,$91,$E8,$7D,$7F,$5F,$88,$F4
	.dc.b	$C5,$C3,$BD,$AB,$F1,$5E,$17,$F8,$72,$FF,$CE,$DE,$06,$0C,$F8,$12
	.dc.b	$32,$43,$FC,$8E,$F2,$EF,$91,$FE,$7F,$79,$34,$87,$F9,$39,$CB,$FC
	.dc.b	$3F,$83,$F9,$B7,$EA,$EF,$03,$F4,$12,$8C,$3F,$E7,$66,$BC,$4F,$E9
	.dc.b	$3F,$AC,$2F,$8C,$EB,$5F,$F2,$D4,$FB,$EB,$80,$71,$1F,$E5,$F9,$E1
	.dc.b	$87,$F4,$9F,$A6,$78,$FE,$46,$E1,$FE,$88,$6B,$D3,$D6,$FE,$56,$E6
	.dc.b	$7D,$70,$F9,$5A,$73,$8F,$7F,$C3,$7F,$38,$AF,$F4,$9B,$5F,$3B,$72
	.dc.b	$FF,$3A,$3F,$CF,$5B,$E1,$F2,$24,$F2,$F0,$EB,$CA,$2C,$F2,$9F,$9F
	.dc.b	$EA,$F9,$6B,$FD,$46,$D7,$E3,$B6,$7F,$D0,$9F,$4A,$F1,$F8,$E7,$F6
	.dc.b	$21,$E4,$53,$38,$7F,$B0,$DF,$C8,$91,$FF,$63,$79,$35,$FF,$49,$C5
	.dc.b	$C3,$FD,$AF,$3F,$84,$FC,$2B,$FD,$E5,$5F,$EB,$86,$BF,$71,$B3,$FE
	.dc.b	$99,$5F,$19,$C7,$E1,$E5,$71,$7A,$46,$23,$94,$5A,$E5,$3F,$94,$FA
	.dc.b	$57,$C6,$7F,$81,$6F,$FC,$33,$18,$70,$D7,$FC,$14,$D7,$FC,$22,$E6
	.dc.b	$BF,$D5,$DF,$47,$FE,$E4,$9F,$F6,$9B,$5F,$CB,$CE,$F9,$BF,$C4,$FF
	.dc.b	$0B,$F3,$BE,$07,$FA,$33,$6B,$FC,$57,$87,$F1,$88,$C4,$7F,$AF,$BB
	.dc.b	$37,$BF,$FB,$92,$3F,$DD,$DF,$79,$FD,$37,$E1,$8E,$B1,$FE,$0F,$E9
	.dc.b	$7F,$08,$E5,$97,$FE,$FD,$8C,$BF,$EF,$F8,$CB,$FE,$CC,$8C,$BF,$EC
	.dc.b	$E8,$CE,$7F,$C8,$BC,$2F,$F1,$F3,$7C,$EF,$E2,$A4,$AF,$5E,$DF,$E8
	.dc.b	$EF,$3B,$11,$FE,$D8,$9B,$5C,$E7,$F8,$EF,$9D,$2B,$89,$FE,$E5,$7F
	.dc.b	$90,$F9,$D9,$8F,$F7,$0C,$5A,$F2,$4F,$F1,$2E,$3F,$E6,$B1,$F5,$BC
	.dc.b	$AC,$1F,$C5,$B8,$7F,$A7,$60,$AD,$E7,$FE,$51,$6C,$7F,$97,$9A,$FE
	.dc.b	$98,$E6,$7E,$02,$AB,$FC,$5A,$BF,$AD,$79,$6D,$7F,$B0,$CF,$B2,$30
	.dc.b	$FF,$3C,$3F,$9E,$F9,$BF,$E0,$7F,$A6,$F0,$C9,$A6,$2B,$AB,$FD,$7F
	.dc.b	$2F,$ED,$ED,$BF,$ED,$2D,$AE,$58,$F5,$F2,$FB,$7D,$1C,$BF,$25,$FF
	.dc.b	$4F,$6C,$65,$FE,$71,$7E,$38,$E5,$1A,$38,$F9,$DE,$93,$86,$DF,$DA
	.dc.b	$2F,$D5,$3D,$9F,$71,$EF,$FA,$BB,$19,$21,$E0,$84,$6A,$31,$16,$7B
	.dc.b	$4F,$C9,$FA,$69,$5E,$C7,$E6,$7D,$3E,$23,$A4,$5A,$F3,$4D,$D7,$E5
	.dc.b	$CD,$AF,$97,$3F,$57,$C4,$91,$F2,$1F,$B1,$F4,$98,$8F,$44,$5C,$3E
	.dc.b	$14,$DC,$3E,$BC,$DA,$FA,$F3,$F8,$1F,$56,$47,$D4,$7F,$09,$F3,$D8
	.dc.b	$8F,$89,$17,$0F,$9D,$37,$0F,$C1,$CD,$AF,$C1,$CF,$E3,$7E,$9A,$47
	.dc.b	$E0,$1F,$C8,$7B,$18,$8F,$A1,$17,$0F,$B5,$37,$0F,$C7,$CD,$AF,$C7
	.dc.b	$CF,$E6,$7B,$E4,$7E,$31,$FC,$E7,$C7,$62,$3E,$E4,$5C,$3F,$0D,$3F
	.dc.b	$E3,$ED,$FF,$C7,$FC,$52,$FF,$C7,$F1,$97,$FC,$83,$19,$7F,$C8,$31
	.dc.b	$FE,$46,$DE,$47,$F9,$19,$FF,$24,$F0,$E2,$3F,$C8,$F1,$70,$FF,$17
	.dc.b	$CF,$F9,$3B,$79,$5F,$E4,$E7,$FC,$A3,$C3,$88,$FF,$27,$C5,$AF,$F1
	.dc.b	$BC,$DD,$BF,$CC,$3E,$BA,$B1,$7F,$E5,$3A,$FF,$2D,$F0,$7F,$B7,$FD
	.dc.b	$C7,$CE,$F0,$E4,$C3,$6F,$FD,$DF,$3A,$7F,$CB,$95,$A6,$9A,$9E,$0A
	.dc.b	$78,$73,$7F,$96,$6F,$83,$FE,$5B,$BE,$54,$F9,$F7,$C2,$7F,$86,$E3
	.dc.b	$FD,$E9,$18,$F4,$BF,$B5,$36,$3F,$DE,$0D,$7F,$DE,$FE,$FA,$DF,$1B
	.dc.b	$FD,$C6,$30,$D8,$CC,$FF,$BE,$67,$2F,$4F,$4F,$F9,$4A,$3C,$6E,$1B
	.dc.b	$E0,$79,$23,$11,$FE,$8A,$F9,$FF,$F4,$A7,$1F,$C5,$8C,$7C,$A7,$FA
	.dc.b	$1B,$83,$E2,$D6,$A7,$E4,$70,$FF,$A6,$A1,$5B,$8C,$5D,$AE,$6F,$5E
	.dc.b	$05,$F3,$C3,$AE,$1B,$9F,$F5,$27,$C9,$FF,$A5,$B9,$BF,$D1,$BF,$04
	.dc.b	$8C,$63,$FE,$A4,$F9,$4B,$9F,$FA,$92,$B5,$3F,$EB,$1A,$CE,$1D,$23
	.dc.b	$FD,$69,$78,$AD,$DF,$FA,$C2,$F0,$66,$6D,$7F,$AE,$FE,$3F,$81,$C7
	.dc.b	$CF,$B1,$92,$5F,$FA,$EA,$32,$FF,$AF,$3E,$23,$C1,$F2,$1E,$16,$D7
	.dc.b	$FB,$1D,$B3,$B8,$FF,$65,$4F,$AB,$D2,$74,$53,$1E,$E5,$AF,$ED,$53
	.dc.b	$A6,$CE,$9A,$7F,$6F,$F9,$F7,$CF,$FC,$66,$C6,$7E,$7F,$13,$23,$83
	.dc.b	$B8,$C6,$5F,$D8,$AB,$8B,$67,$55,$C0,$F0,$7C,$9F,$73,$C1,$F3,$F6
	.dc.b	$32,$4A,$E9,$19,$7B,$4D,$E4,$FF,$68,$E0,$AC,$5F,$FB,$BA,$1F,$3F
	.dc.b	$ED,$3D,$DF,$09,$67,$E1,$45,$C7,$D5,$37,$0E,$B3,$6B,$A4,$DA,$FF
	.dc.b	$79,$DB,$E0,$3E,$51,$F2,$FE,$02,$DF,$CF,$8B,$5F,$1E,$6E,$3E,$A9
	.dc.b	$B5,$E9,$9B,$5D,$BE,$7F,$EB,$1F,$5F,$E7,$2D,$7D,$B8,$B3,$F4,$E6
	.dc.b	$E1,$F1,$A6,$D7,$C5,$9B,$5E,$AF,$9F,$FC,$17,$C1,$FE,$0D,$B5,$F8
	.dc.b	$77,$0F,$E2,$7E,$1E,$D7,$E0,$5B,$8F,$D4,$6C,$FD,$3F,$9F,$FC,$71
	.dc.b	$F8,$FF,$9F,$B1,$F9,$37,$0F,$E5,$7E,$66,$E1,$F8,$DF,$2F,$F2,$D1
	.dc.b	$F9,$76,$CF,$F9,$93,$E7,$FF,$95,$EB,$FC,$11,$B1,$9F,$2F,$EB,$36
	.dc.b	$75,$06,$98,$E1,$B3,$B2,$63,$FE,$F7,$BF,$E7,$33,$A8,$FD,$17,$CE
	.dc.b	$7E,$AC,$FD,$67,$93,$6F,$F6,$11,$67,$F5,$13,$77,$FC,$DD,$7E,$CC
	.dc.b	$FD,$A7,$8D,$5F,$DC,$16,$BF,$61,$16,$7F,$D2,$5E,$27,$F2,$0F,$79
	.dc.b	$E2,$D8,$C3,$70,$FF,$38,$4E,$9B,$3A,$69,$A9,$E1,$87,$F9,$82,$B8
	.dc.b	$2B,$FF,$5B,$7C,$FB,$E7,$FF,$9A,$FC,$52,$31,$8F,$83,$C7,$CC,$63
	.dc.b	$0E,$53,$FE,$C4,$8C,$61,$D3,$0F,$65,$F0,$7F,$A0,$DB,$E7,$D2,$C3
	.dc.b	$FD,$B9,$18,$6C,$66,$7F,$B4,$12,$74,$E0,$FE,$D8,$E2,$F8,$47,$79
	.dc.b	$F6,$FD,$C7,$9A,$F9,$47,$E3,$38,$24,$7E,$47,$82,$DF,$E2,$79,$15
	.dc.b	$8C,$7C,$9C,$7F,$E0,$4B,$C6,$1F,$DF,$63,$FC,$0F,$58,$C1,$1B,$9F
	.dc.b	$CF,$4F,$E0,$7C,$89,$1F,$7B,$CA,$B7,$F6,$66,$E3,$FE,$D9,$9B,$5F
	.dc.b	$13,$E8,$1F,$33,$F3,$9B,$1F,$ED,$2F,$49,$F4,$0F,$17,$57,$2B,$FD
	.dc.b	$AF,$9B,$0E,$2F,$DE,$18,$72,$BE,$11,$FA,$D1,$F1,$DF,$BB,$6D,$FE
	.dc.b	$FD,$7C,$76,$66,$4E,$5E,$11,$89,$E9,$17,$4F,$9F,$1C,$AB,$81,$E8
	.dc.b	$FA,$07,$CD,$7E,$39,$B1,$F0,$75,$FF,$21,$BB,$9D,$11,$D7,$81,$6F
	.dc.b	$E0,$0F,$9F,$FE,$EC,$E1,$B8,$FE,$57,$5F,$99,$F1,$7F,$9E,$9C,$E1
	.dc.b	$F9,$86,$DF,$F9,$05,$B5,$C2,$7E,$8C,$6B,$D6,$FF,$78,$7A,$BF,$C1
	.dc.b	$4F,$E0,$DF,$3D,$6B,$F8,$1C,$5C,$3F,$7D,$37,$2F,$D4,$5F,$FB,$E7
	.dc.b	$E7,$F8,$E1,$FC,$29,$B7,$FC,$15,$B5,$C3,$C1,$46,$2B,$FC,$B1,$6F
	.dc.b	$63,$31,$FE,$5B,$BF,$F2,$0D,$7F,$8D,$F1,$FF,$1C,$32,$74,$72,$AE
	.dc.b	$38,$FF,$BD,$15,$6E,$31,$F3,$DF,$E4,$13,$FC,$85,$C1,$6F,$9C,$FF
	.dc.b	$92,$E7,$0F,$39,$B8,$FF,$95,$26,$D7,$86,$6D,$71,$8E,$79,$7B,$6B
	.dc.b	$BC,$63,$14,$7F,$94,$FE,$62,$4E,$7E,$7A,$E1,$EB,$9B,$97,$F9,$92
	.dc.b	$6D,$7C,$1F,$97,$FF,$2E,$1F,$E5,$DF,$92,$B1,$DE,$2E,$1F,$16,$6D
	.dc.b	$FF,$9B,$A6,$D7,$CB,$9B,$5F,$E6,$DE,$7F,$F3,$41,$FE,$69,$E8,$B5
	.dc.b	$EB,$8B,$3F,$46,$6E,$1F,$E7,$A9,$B5,$F5,$E6,$D7,$58,$E4,$7C,$58
	.dc.b	$CC,$7F,$B0,$C7,$CF,$FF,$04,$7F,$9C,$F2,$FF,$3B,$8C,$46,$DA,$FF
	.dc.b	$0E,$AF,$F5,$34,$62,$3F,$D8,$47,$4B,$FE,$81,$5C,$DF,$E8,$51,$A9
	.dc.b	$E7,$39,$8F,$E8,$DE,$FF,$4A,$E3,$76,$72,$75,$3C,$EE,$5F,$34,$70
	.dc.b	$74,$60,$D9,$83,$FD,$BE,$5A,$F1,$3B,$3F,$D8,$4E,$1D,$FB,$AF,$7F
	.dc.b	$D7,$87,$FB,$1E,$31,$3E,$66,$BF,$3D,$BF,$04,$F9,$1F,$EF,$D3,$FE
	.dc.b	$92,$BF,$ED,$33,$9D,$7A,$46,$2B,$77,$C9,$B1,$99,$F8,$54,$F9,$0F
	.dc.b	$EE,$7A,$FD,$A8,$C6,$3A,$C3,$C3,$39,$6E,$9E,$2B,$E2,$DA,$E1,$84
	.dc.b	$7E,$6E,$1D,$1B,$5C,$1F,$F7,$54,$7A,$DD,$DF,$67,$33,$B5,$3C,$F0
	.dc.b	$F2,$5F,$77,$D5,$5B,$9D,$61,$F0,$6F,$F0,$9C,$BF,$C8,$7E,$63,$FC
	.dc.b	$27,$8F,$FB,$81,$BB,$FF,$1D,$6D,$7F,$B7,$FE,$62,$DF,$18,$FE,$19
	.dc.b	$38,$E7,$36,$B0,$C9,$D9,$FC,$4E,$30,$FF,$29,$8B,$87,$FB,$CA,$7A
	.dc.b	$7C,$C4,$AF,$E3,$4F,$7F,$98,$C4,$7F,$31,$8B,$5F,$EF,$A9,$B9,$7F
	.dc.b	$37,$9B,$5F,$EF,$99,$F4,$30,$FE,$4C,$7A,$F8,$F0,$FF,$3D,$2D,$FF
	.dc.b	$54,$CB,$FC,$17,$1F,$CC,$3E,$3B,$FC,$32,$DF,$1F,$E3,$C6,$23,$FC
	.dc.b	$25,$AA,$1E,$16,$AA,$57,$13,$A4,$61,$E3,$3C,$F0,$FE,$CA,$DD,$3F
	.dc.b	$AE,$BD,$27,$91,$C2,$F9,$57,$0B,$F2,$FC,$B4,$AF,$53,$E8,$F9,$7C
	.dc.b	$47,$82,$2D,$78,$27,$F0,$7E,$17,$F8,$62,$BF,$CF,$BD,$F8,$33,$E0
	.dc.b	$48,$C9,$0F,$F3,$E3,$CB,$BE,$47,$FA,$19,$E4,$D2,$1F,$E8,$37,$2F
	.dc.b	$F8,$7B,$E7,$9E,$07,$E6,$21,$18,$7F,$C8,$AD,$7F,$C4,$BF,$3E,$AE
	.dc.b	$13,$9C,$FF,$CE,$2D,$E0,$FF,$3C,$B6,$C2,$F9,$3F,$E5,$2F,$9F,$FF
	.dc.b	$2B,$37,$CB,$F2,$46,$23,$FC,$6B,$F3,$FF,$C7,$DD,$10,$FF,$1C,$C6
	.dc.b	$F1,$FF,$41,$CE,$E7,$FC,$7B,$F0,$4F,$0B,$E0,$36,$3F,$CC,$6D,$77
	.dc.b	$E1,$6A,$64,$69,$E3,$18,$8E,$B1,$6F,$FD,$27,$5F,$CE,$1F,$F4,$AF
	.dc.b	$A6,$F9,$2F,$A0,$DB,$FF,$39,$B5,$FF,$29,$FC,$FF,$F9,$A9,$CC,$F0
	.dc.b	$23,$FE,$7A,$8D,$3F,$EC,$1F,$23,$F1,$E4,$7F,$D9,$7E,$4D,$74,$7F
	.dc.b	$B3,$A3,$F2,$B1,$F8,$3F,$9F,$7C,$0D,$25,$7F,$A2,$21,$F1,$38,$8F
	.dc.b	$F3,$4C,$70,$75,$3C,$E7,$59,$FF,$AD,$AB,$55,$C2,$B4,$FF,$A5,$3E
	.dc.b	$7D,$F1,$3F,$95,$6C,$66,$3F,$31,$F3,$FF,$5F,$0F,$F6,$0D,$6A,$BF
	.dc.b	$C0,$AD,$7F,$D9,$0D,$7F,$D9,$AD,$5F,$04,$D3,$FD,$AF,$1F,$E5,$08
	.dc.b	$C7,$BD,$FC,$F9,$B1,$FE,$F6,$EB,$C5,$70,$8F,$F5,$54,$E9,$B1,$9C
	.dc.b	$3F,$CE,$15,$FC,$13,$CB,$FE,$0B,$E8,$CB,$FF,$80,$31,$97,$FC,$AF
	.dc.b	$E0,$7D,$EB,$FF,$3C,$56,$67,$FE,$00,$57,$17,$F9,$4D,$6F,$B7,$FC
	.dc.b	$F6,$FF,$8C,$AF,$FD,$01,$E6,$7F,$14,$25,$76,$F3,$2B,$CA,$2C,$FF
	.dc.b	$A1,$E7,$F8,$DF,$99,$2B,$CA,$FF,$1E,$F3,$31,$1E,$28,$B5,$FE,$8F
	.dc.b	$9F,$B3,$E3,$CA,$FF,$69,$F1,$7F,$33,$7F,$D6,$7B,$FF,$BF,$EF,$FD
	.dc.b	$FB,$39,$6D,$71,$AD,$96,$32,$F4,$8F,$E3,$D5,$CF,$4F,$87,$E6,$7F
	.dc.b	$34,$A8,$7C,$E7,$FB,$CF,$CC,$F7,$9F,$F5,$1B,$65,$E6,$23,$0B,$7E
	.dc.b	$D3,$A9,$FE,$A0,$DF,$2F,$A9,$DB,$2E,$FE,$57,$83,$E7,$25,$7F,$A5
	.dc.b	$AD,$38,$8F,$EC,$B0,$9C,$6B,$FE,$C2,$8D,$CF,$19,$F5,$F9,$92,$30
	.dc.b	$75,$9F,$79,$1F,$07,$CE,$57,$BC,$5B,$ED,$37,$5F,$8B,$36,$BE,$04
	.dc.b	$FC,$AF,$9A,$91,$EB,$7E,$6F,$CE,$62,$3D,$71,$70,$F5,$4D,$C3,$E8
	.dc.b	$CD,$AF,$99,$3F,$5B,$C3,$91,$F1,$DF,$B3,$F3,$98,$8F,$8F,$17,$0F
	.dc.b	$8D,$37,$0F,$BB,$36,$BE,$C4,$FE,$0B,$D6,$91,$F4,$DF,$C2,$FC,$DE
	.dc.b	$23,$E9,$C5,$C3,$E9,$4D,$C3,$F1,$33,$6B,$F0,$93,$F8,$EF,$9F,$91
	.dc.b	$F7,$DF,$C8,$FA,$F8,$8F,$BF,$17,$0F,$BD,$37,$0F,$CA,$CD,$AF,$C8
	.dc.b	$4F,$E6,$BE,$7E,$47,$E2,$DF,$CE,$FC,$FE,$23,$F1,$71,$70,$FC,$54
	.dc.b	$FF,$8D,$37,$FF,$1A,$F1,$4B,$FF,$1D,$46,$5F,$F1,$E4,$65,$FF,$18
	.dc.b	$C7,$F8,$F3,$79,$1F,$E4,$17,$FC,$7F,$C3,$88,$FF,$1B,$C5,$C3,$FC
	.dc.b	$6F,$3F,$E4,$4D,$E5,$7F,$92,$5F,$F2,$3F,$0E,$23,$FC,$7F,$16,$BF
	.dc.b	$C7,$F3,$76,$FF,$1F,$CD,$AF,$DC,$FA,$EA,$FF,$28,$F0,$7F,$85,$A5
	.dc.b	$ED,$BE,$7B,$FC,$8F,$C9,$98,$C4,$F2,$8C,$4F,$F9,$32,$AC,$F6,$6F
	.dc.b	$57,$F9,$6E,$AE,$3F,$DF,$F0,$E7,$3F,$E5,$FA,$E3,$6F,$9F,$7C,$17
	.dc.b	$F8,$56,$3C,$D1,$8F,$4B,$FD,$C7,$1B,$9D,$3F,$E3,$09,$D3,$FE,$E8
	.dc.b	$F7,$56,$F8,$AE,$31,$86,$C6,$67,$83,$5F,$8C,$79,$7D,$3F,$F3,$EB
	.dc.b	$8B,$46,$23,$CE,$D7,$C8,$DC,$BF,$CF,$DD,$2E,$2F,$F2,$3E,$BF,$13
	.dc.b	$6F,$F6,$94,$63,$83,$FC,$F3,$D7,$BC,$BE,$BC,$7F,$9F,$30,$C6,$1C
	.dc.b	$5B,$19,$BE,$55,$F8,$2B,$CB,$F8,$4A,$B8,$FF,$A4,$7A,$7B,$72,$FF
	.dc.b	$A2,$BA,$EE,$3F,$E8,$A9,$B8,$7E,$32,$6D,$F1,$AF,$F5,$04,$63,$45
	.dc.b	$D8,$C3,$67,$FD,$4C,$DB,$FF,$4D,$E7,$C7,$AF,$93,$6B,$FD,$58,$D9
	.dc.b	$DC,$7F,$AB,$E7,$C1,$F1,$B2,$3F,$A6,$72,$DA,$FF,$53,$CD,$D3,$FA
	.dc.b	$54,$DA,$FF,$3C,$E8,$FC,$65,$F3,$BE,$E5,$F3,$D6,$30,$DA,$FF,$40
	.dc.b	$E1,$F5,$EF,$94,$E7,$2F,$F5,$FE,$1C,$AF,$67,$81,$7C,$F7,$7F,$9E
	.dc.b	$F0,$36,$77,$5F,$DD,$2B,$71,$8C,$3D,$17,$C2,$31,$28,$C5,$78,$9B
	.dc.b	$8F,$9E,$75,$5C,$3E,$7B,$E0,$7B,$FF,$05,$B1,$93,$AC,$65,$A6,$70
	.dc.b	$F7,$C6,$5B,$1E,$4F,$5F,$E6,$7B,$3F,$35,$B1,$93,$D3,$19,$69,$99
	.dc.b	$F9,$71,$96,$C7,$C0,$F9,$DF,$B1,$F3,$3F,$65,$B1,$93,$E2,$C6,$5A
	.dc.b	$66,$7E,$BC,$65,$B1,$F3,$3E,$77,$F0,$9E,$DF,$E1,$5B,$19,$3E,$8C
	.dc.b	$65,$A6,$67,$F0,$71,$96,$C7,$D8,$F7,$FF,$20,$7E,$47,$E7,$2C,$7D
	.dc.b	$C6,$CF,$E3,$1B,$BF,$DB,$78,$4F,$E0,$BE,$76,$D7,$E4,$BE,$77,$F3
	.dc.b	$87,$E7,$7C,$6B,$5F,$88,$F0,$B0,$D3,$F1,$71,$FA,$3F,$9D,$91,$F8
	.dc.b	$EF,$97,$B8,$33,$FE,$31,$F3,$1F,$CC,$76,$FE,$12,$FF,$33,$E1,$BE
	.dc.b	$47,$FA,$8B,$1F,$CE,$CE,$5B,$19,$C3,$A4,$7F,$BB,$A7,$51,$80,$F7
	.dc.b	$8E,$35,$DA,$FB,$B6,$32,$7E,$A7,$F1,$57,$CE,$FE,$FE,$3F,$8E,$C6
	.dc.b	$1F,$D5,$98,$8D,$3C,$5B,$87,$8F,$AD,$4F,$83,$E7,$F3,$DD,$1B,$87
	.dc.b	$85,$B7,$C6,$B8,$4F,$96,$BC,$7F,$37,$E4,$F9,$BB,$3E,$96,$DF,$16
	.dc.b	$EF,$FC,$4F,$04,$62,$BF,$8F,$C6,$1F,$85,$7D,$5F,$E1,$F6,$F9,$DF
	.dc.b	$1B,$F3,$F0,$7C,$6F,$F0,$88,$FA,$46,$F3,$FE,$8F,$3A,$F0,$7F,$A4
	.dc.b	$CE,$70,$FE,$F7,$87,$48,$FE,$DE,$D7,$FC,$2C,$D7,$FD,$27,$93,$E5
	.dc.b	$BE,$AF,$CC,$7D,$66,$DF,$DB,$6D,$7C,$B6,$E3,$F9,$4E,$4F,$EE,$55
	.dc.b	$F8,$EE,$6F,$C9,$BF,$94,$6B,$93,$F2,$66,$4A,$66,$FF,$25,$CB,$FE
	.dc.b	$0F,$8D,$47,$DB,$F9,$4F,$C5,$FC,$C7,$E3,$1B,$3F,$91,$6C,$EE,$3F
	.dc.b	$29,$3F,$23,$E7,$3F,$C7,$51,$FE,$F7,$FC,$11,$E6,$E7,$D9,$FE,$EB
	.dc.b	$F1,$C9,$CD,$7D,$AB,$E3,$8F,$F8,$70,$38,$7F,$C3,$ED,$7F,$C8,$CD
	.dc.b	$7F,$5A,$0F,$DD,$8F,$85,$AE,$1B,$3F,$BC,$8D,$35,$D3,$4D,$47,$08
	.dc.b	$F9,$67,$03,$F1,$BF,$8C,$BE,$13,$FC,$27,$F8,$C3,$F1,$B7,$C3,$FE
	.dc.b	$3A,$B5,$8F,$F6,$8A,$D4,$E6,$FF,$C5,$98,$7F,$8F,$AF,$6F,$C5,$6B
	.dc.b	$98,$DB,$FE,$69,$FC,$C3,$F2,$58,$7F,$74,$C9,$E7,$BE,$13,$F3,$F5
	.dc.b	$AB,$FE,$2D,$5A,$9C,$F4,$74,$9F,$99,$5B,$7F,$CA,$2D,$7C,$11,$BF
	.dc.b	$0B,$F8,$6F,$CE,$7C,$2F,$C6,$DF,$1F,$FC,$62,$30,$DD,$3E,$14,$E9
	.dc.b	$FE,$41,$3A,$7F,$91,$7C,$BF,$F2,$03,$F9,$0F,$CE,$D8,$C9,$32,$FE
	.dc.b	$2D,$19,$7F,$90,$7C,$03,$DC,$FF,$21,$7C,$D7,$F9,$0D,$B7,$FE,$17
	.dc.b	$9D,$36,$74,$D3,$FC,$39,$E4,$FF,$8C,$9B,$EE,$7F,$94,$23,$18,$7F
	.dc.b	$2F,$7F,$C5,$8D,$F7,$75,$3B,$6D,$74,$6C,$EE,$3A,$99,$2B,$93,$B4
	.dc.b	$F8,$7D,$DF,$17,$93,$6B,$CE,$DA,$F2,$37,$C6,$FF,$22,$63,$FC,$9D
	.dc.b	$FF,$23,$7D,$D3,$6F,$F3,$5A,$E0,$B1,$86,$E3,$FE,$3F,$9D,$36,$74
	.dc.b	$D3,$53,$C3,$1F,$EA,$95,$C1,$FE,$AD,$F7,$77,$0F,$F2,$C2,$F1,$7F
	.dc.b	$CF,$AE,$70,$FF,$32,$B9,$7F,$CA,$0E,$69,$CA,$8D,$47,$6C,$79,$DF
	.dc.b	$46,$DF,$26,$CE,$E3,$AC,$ED,$A7,$F5,$48,$FF,$54,$8F,$A1,$7C,$C7
	.dc.b	$FA,$E3,$AF,$FD,$51,$18,$9F,$F5,$DF,$C4,$7F,$19,$8C,$47,$F1,$B9
	.dc.b	$DD,$7F,$25,$AD,$3C,$EE,$D7,$23,$C1,$19,$69,$9B,$EA,$FF,$19,$6B
	.dc.b	$FE,$F0,$36,$78,$67,$AB,$FE,$F1,$32,$71,$9F,$23,$8B,$F3,$BD,$E3
	.dc.b	$C5,$5D,$AF,$9C,$F1,$3D,$1C,$CE,$5E,$6D,$8C,$C6,$0F,$E7,$CB,$E2
	.dc.b	$DF,$13,$FE,$3D,$8F,$DF,$4E,$A7,$3C,$1F,$D1,$6F,$A5,$71,$BE,$11
	.dc.b	$8A,$DC,$F2,$3F,$A6,$38,$C7,$EA,$AF,$89,$F1,$C6,$1B,$9F,$87,$0F
	.dc.b	$24,$F0,$6C,$79,$67,$77,$D9,$AF,$97,$1F,$A3,$87,$86,$B8,$4E,$2B
	.dc.b	$73,$98,$F3,$DF,$67,$A1,$B3,$9B,$97,$B1,$F8,$08,$F6,$3B,$31,$1C
	.dc.b	$F8,$7F,$55,$D9,$F4,$B1,$FD,$AF,$CB,$5E,$0F,$D9,$B7,$0F,$D8,$F1
	.dc.b	$FF,$42,$7F,$6B,$79,$C3,$F6,$95,$68,$D9,$95,$FB,$C8,$D1,$87,$F8
	.dc.b	$3C,$61,$D1,$D8,$FF,$69,$7D,$DB,$E5,$3C,$0F,$F7,$48,$CF,$8D,$89
	.dc.b	$91,$82,$1F,$E6,$99,$E4,$F4,$8E,$73,$23,$07,$F4,$6F,$87,$F7,$75
	.dc.b	$7F,$72,$8F,$E9,$3E,$2E,$26,$46,$08,$7F,$49,$78,$FC,$3F,$2F,$87
	.dc.b	$93,$FD,$4A,$32,$FF,$54,$8C,$BF,$D4,$63,$2F,$F5,$38,$CF,$07,$F4
	.dc.b	$09,$CB,$FD,$0A,$33,$C1,$FE,$2B,$E0,$FF,$3A,$D7,$AB,$DD,$F5,$FC
	.dc.b	$3C,$C7,$FB,$1C,$65,$FE,$CB,$19,$7F,$B1,$47,$DF,$F6,$1F,$37,$49
	.dc.b	$3F,$E9,$FF,$97,$AF,$FA,$3E,$2E,$9F,$E8,$79,$E5,$F3,$7F,$8D,$F4
	.dc.b	$1F,$0C,$F1,$F0,$DC,$7F,$D5,$55,$C1,$FF,$18,$46,$5F,$F5,$B5,$72
	.dc.b	$AD,$47,$FA,$23,$D7,$70,$FF,$95,$63,$FC,$9D,$18,$70,$D7,$FC,$B4
	.dc.b	$D7,$F8,$B9,$FE,$64,$8F,$D2,$2F,$D6,$5F,$11,$86,$C7,$FA,$7A,$BA
	.dc.b	$69,$FE,$44,$FC,$71,$F2,$7E,$3A,$46,$1F,$33,$5F,$F3,$63,$5F,$1B
	.dc.b	$99,$FF,$27,$F7,$F8,$27,$FD,$4E,$D8,$F0,$57,$A3,$F1,$4F,$9F,$5F
	.dc.b	$EB,$25,$F1,$99,$35,$BC,$9D,$46,$5E,$11,$97,$FD,$6B,$3C,$9D,$9A
	.dc.b	$3F,$D7,$FF,$18,$F9,$AF,$80,$D8,$FA,$8D,$7F,$D1,$4D,$7E,$8B,$9C
	.dc.b	$7F,$CD,$BF,$1D,$FB,$A9,$FF,$62,$B6,$3F,$CE,$33,$BE,$BF,$F6,$9F
	.dc.b	$CE,$FE,$67,$C8,$98,$7F,$BD,$63,$F3,$86,$1F,$F7,$34,$61,$FF,$6B
	.dc.b	$FD,$07,$C8,$7F,$DA,$6F,$BD,$BA,$FE,$12,$35,$8F,$FA,$09,$B8,$7F
	.dc.b	$60,$78,$46,$67,$86,$B2,$FB,$C7,$33,$0F,$FB,$E7,$C6,$FD,$97,$CF
	.dc.b	$C8,$FF,$54,$5F,$FB,$CE,$7F,$6C,$66,$3F,$E0,$04,$E1,$FF,$78,$FD
	.dc.b	$03,$E4,$3F,$24,$D8,$FD,$33,$5F,$F5,$E3,$5F,$D1,$B9,$BF,$F4,$FF
	.dc.b	$D0,$3E,$5B,$F4,$AD,$AE,$F7,$FC,$27,$D3,$FE,$15,$E6,$49,$F4,$EF
	.dc.b	$FC,$5F,$C2,$FF,$60,$46,$0F,$F7,$AD,$6D,$64,$AE,$17,$D6,$78,$78
	.dc.b	$CC,$39,$C7,$0A,$DB,$A8,$FE,$B8,$E4,$EE,$72,$FC,$73,$FD,$88,$EC
	.dc.b	$FE,$44,$7F,$61,$72,$F1,$A4,$63,$0F,$34,$FF,$96,$AB,$F8,$5C,$62
	.dc.b	$BC,$D3,$CF,$E0,$BA,$79,$BD,$7C,$D9,$3E,$38,$CB,$E5,$8C,$BE,$1C
	.dc.b	$7F,$45,$97,$F8,$BB,$67,$D0,$7D,$4E,$B9,$7F,$34,$4E,$9C,$35,$F8
	.dc.b	$86,$A3,$8B,$88,$D6,$9F,$82,$AC,$BF,$84,$AF,$C1,$C7,$C7,$25,$7C
	.dc.b	$D7,$E5,$47,$CB,$FA,$0F,$99,$F4,$12,$BE,$94,$7C,$FF,$85,$91,$F5
	.dc.b	$5D,$3F,$07,$E8,$3E,$A1,$1F,$B5,$F1,$95,$F8,$DF,$41,$F6,$4C,$67
	.dc.b	$F4,$F6,$46,$67,$F0,$4E,$B6,$FE,$D9,$3F,$E5,$89,$FC,$37,$87,$F8
	.dc.b	$33,$F0,$9E,$24,$8C,$90,$FA,$CE,$5F,$B1,$1C,$5F,$C8,$47,$E3,$7E
	.dc.b	$7F,$F1,$DE,$CC,$9F,$CA,$46,$5F,$CB,$46,$5F,$C1,$46,$5F,$C2,$46
	.dc.b	$72,$F7,$4F,$E6,$7C,$8F,$CD,$7C,$FC,$BF,$D0,$46,$5F,$D1,$46,$5F
	.dc.b	$C7,$46,$5F,$C8,$46,$6B,$E5,$4F,$E9,$7B,$7F,$4D,$F3,$F2,$FF,$59
	.dc.b	$19,$7F,$5D,$19,$7F,$35,$19,$7F,$39,$19,$AF,$AD,$3F,$B1,$F7,$7F
	.dc.b	$65,$EF,$4B,$FD,$C4,$65,$FD,$D4,$65,$FD,$34,$65,$FD,$44,$66,$BF
	.dc.b	$05,$3F,$E5,$DF,$91,$7C,$3D,$F8,$79,$87,$F0,$E9,$CB,$67,$2F,$F0
	.dc.b	$38,$4F,$39,$FF,$25,$F8,$72,$BF,$8B,$FB,$BF,$CA,$F9,$3F,$C7,$18
	.dc.b	$7F,$87,$72,$FF,$7E,$DE,$EF,$F8,$75,$ED,$FF,$72,$D6,$EF,$FC,$81
	.dc.b	$2F,$3B,$FC,$B3,$F3,$D2,$FB,$C7,$F9,$70,$C3,$D2,$30,$F9,$3C,$AF
	.dc.b	$F3,$0F,$A7,$23,$04,$3C,$6E,$5F,$2C,$7F,$9A,$0C,$67,$FC,$7F,$0F
	.dc.b	$F0,$A6,$3E,$69,$FF,$36,$FC,$F7,$FB,$13,$E1,$1C,$7E,$A8,$C3,$71
	.dc.b	$09,$DA,$F1,$D8,$FF,$77,$9C,$C6,$2F,$9D,$6E,$91,$89,$D4,$F6,$AE
	.dc.b	$35,$CE,$FF,$DB,$3F,$41,$C3,$8F,$FC,$98,$DB,$FF,$23,$87,$0F,$F9
	.dc.b	$29,$AF,$FB,$77,$AD,$F2,$FF,$E5,$5F,$85,$FF,$22,$B6,$33,$3D,$27
	.dc.b	$2F,$6A,$7A,$BF,$D7,$A3,$CA,$D7,$E3,$C6,$BC,$EF,$3C,$E7,$57,$C6
	.dc.b	$7F,$B8,$F8,$7E,$06,$BF,$E8,$98,$C6,$3F,$3E,$31,$E0,$7F,$5C,$E0
	.dc.b	$F9,$B5,$A6,$F8,$19,$AE,$4D,$BF,$AD,$5F,$F0,$03,$0E,$F1,$89,$D6
	.dc.b	$3E,$4E,$FF,$ED,$3C,$BF,$D9,$F0,$78,$18,$C7,$CD,$57,$1E,$B3,$70
	.dc.b	$E5,$5F,$E0,$4B,$E2,$DB,$D4,$62,$FF,$BD,$28,$CD,$E9,$B5,$CF,$2E
	.dc.b	$3E,$07,$26,$C6,$4F,$04,$65,$A7,$8F,$C0,$F0,$9E,$7F,$1D,$CD,$89
	.dc.b	$D1,$33,$FF,$07,$4E,$A3,$1A,$FF,$82,$FE,$83,$F8,$BC,$7F,$88,$9F
	.dc.b	$E2,$2D,$F0,$3E,$24,$62,$FF,$C3,$0F,$06,$F8,$3A,$9D,$BC,$3D,$CE
	.dc.b	$3F,$1B,$70,$EC,$DB,$E6,$DC,$7C,$8D,$9F,$F6,$A6,$28,$C5,$7E,$E6
	.dc.b	$31,$3E,$9F,$85,$F3,$FC,$37,$A1,$B1,$93,$DD,$19,$69,$9A,$F6,$C6
	.dc.b	$5B,$1F,$17,$E1,$BE,$1F,$C0,$7C,$46,$C6,$4F,$95,$19,$69,$99,$F9
	.dc.b	$31,$96,$C7,$D1,$F8,$5F,$9F,$F0,$DF,$41,$B1,$93,$EB,$46,$5A,$66
	.dc.b	$7E,$AC,$65,$B1,$F7,$7C,$6F,$B7,$E4,$FD,$C6,$C6,$4F,$C1,$46,$5A
	.dc.b	$66,$7F,$03,$19,$6C,$7E,$27,$C8,$FC,$39,$F8,$8F,$1E,$C7,$E3,$5B
	.dc.b	$3F,$8A,$6F,$07,$E4,$DE,$13,$F9,$3F,$22,$D7,$E4,$3D,$CF,$C9,$9F
	.dc.b	$94,$F2,$6D,$7E,$67,$C3,$C3,$4F,$CE,$47,$E6,$BE,$62,$47,$E7,$FE
	.dc.b	$82,$E1,$FE,$7B,$F9,$9F,$E5,$BB,$7F,$05,$7F,$97,$F0,$DF,$2F,$FD
	.dc.b	$45,$87,$FA,$3E,$72,$D8,$CD,$F3,$55,$C6,$F9,$36,$B8,$36,$B8,$3F
	.dc.b	$C4,$FD,$B7,$D0,$7E,$CA,$B0,$D7,$FD,$28,$DA,$FF,$46,$4F,$09,$FD
	.dc.b	$E1,$CF,$E7,$7A,$7D,$05,$8F,$F4,$E3,$6B,$8B,$78,$FF,$D2,$16,$8F
	.dc.b	$27,$D0,$79,$7D,$1B,$87,$FA,$95,$B8,$78,$9B,$87,$26,$CF,$FA,$E5
	.dc.b	$B3,$FE,$9F,$7F,$9A,$2F,$9F,$65,$FE,$A4,$8F,$EF,$67,$F7,$C8,$CD
	.dc.b	$7F,$6A,$8E,$13,$B7,$73,$A7,$FB,$7B,$6F,$A3,$6B,$FC,$0F,$F3,$CF
	.dc.b	$AA,$FF,$6B,$47,$FA,$86,$75,$7F,$EB,$5F,$9A,$FA,$DF,$41,$F5,$DB
	.dc.b	$5F,$EC,$76,$CF,$D4,$6E,$3F,$EB,$EE,$3F,$EF,$75,$F8,$DE,$5F,$C7
	.dc.b	$34,$FC,$61,$6F,$F1,$8D,$AE,$5C,$9F,$ED,$D8,$FC,$95,$72,$69,$FE
	.dc.b	$D9,$F9,$6F,$C6,$7B,$1F,$8D,$6C,$64,$FF,$6F,$46,$5A,$7F,$B0,$7E
	.dc.b	$7D,$EE,$FC,$DC,$BF,$C5,$B1,$89,$FA,$4D,$8F,$F0,$EB,$5F,$B2,$1D
	.dc.b	$65,$F7,$5B,$3F,$E3,$8F,$C0,$5F,$81,$3B,$FF,$0B,$5A,$C1,$B7,$E1
	.dc.b	$30,$FD,$A4,$3D,$9F,$DD,$3F,$E4,$28,$CC,$62,$32,$D8,$C9,$CA,$32
	.dc.b	$D3,$2F,$FB,$7B,$E8,$1F,$13,$F0,$E1,$18,$9F,$CA,$36,$3F,$C7,$ED
	.dc.b	$7E,$1B,$9D,$3F,$3C,$D9,$FF,$2C,$DB,$EA,$DC,$3F,$A6,$AD,$61,$FE
	.dc.b	$5E,$FA,$17,$27,$8D,$B8,$FE,$95,$B3,$FE,$50,$6B,$F5,$1C,$E5,$FA
	.dc.b	$F6,$CF,$8B,$CB,$FA,$14,$9C,$BF,$E6,$BE,$E7,$CE,$FF,$8B,$7E,$5D
	.dc.b	$F4,$2F,$8F,$FE,$39,$39,$8C,$65,$FE,$76,$CF,$F8,$7C,$E6,$B7,$39
	.dc.b	$3F,$CF,$77,$D2,$3F,$93,$7D,$0C,$9D,$65,$E3,$FA,$1C,$3F,$C9,$26
	.dc.b	$E3,$FC,$72,$6D,$7F,$96,$7D,$9F,$F1,$EF,$C3,$FF,$8F,$9B,$1F,$E4
	.dc.b	$36,$E1,$FE,$40,$6D,$FF,$A5,$FD,$9F,$F2,$57,$8B,$FE,$27,$8C,$5F
	.dc.b	$FA,$E3,$D9,$CD,$69,$B1,$93,$FD,$7B,$19,$69,$9C,$3F,$D7,$71,$96
	.dc.b	$C6,$6B,$FD,$51,$EC,$F8,$BD,$9F,$1B,$67,$FD,$92,$DA,$FF,$62,$B6
	.dc.b	$93,$FE,$A8,$F8,$17,$D0,$B6,$FF,$33,$3F,$7A,$7E,$7C,$EA,$73,$C5
	.dc.b	$A6,$CE,$AB,$93,$67,$4D,$35,$3C,$31,$61,$E4,$7F,$D7,$1F,$41,$E1
	.dc.b	$73,$5E,$78,$78,$BF,$E7,$B7,$55,$C7,$24,$6A,$BD,$2E,$9E,$55,$C7
	.dc.b	$1E,$4D,$8D,$B7,$0F,$09,$E4,$71,$87,$7B,$ED,$3A,$C3,$C3,$87,$4C
	.dc.b	$3F,$2F,$3F,$ED,$B1,$F4,$2F,$99,$FE,$4F,$DF,$F0,$63,$13,$FC,$AB
	.dc.b	$E1,$FF,$DB,$B1,$88,$FE,$0F,$78,$BF,$F6,$6D,$59,$FE,$D3,$7C,$E7
	.dc.b	$31,$E0,$7B,$D6,$E3,$C9,$7D,$5F,$E5,$91,$97,$FD,$C0,$71,$3A,$B9
	.dc.b	$7F,$DE,$06,$4F,$29,$4E,$31,$D6,$BC,$91,$ED,$34,$E1,$F1,$D7,$99
	.dc.b	$E4,$76,$E4,$72,$FB,$1B,$1C,$DA,$FC,$F6,$FC,$53,$E2,$7F,$1F,$3F
	.dc.b	$EA,$CA,$FF,$07,$CE,$77,$4E,$4B,$19,$70,$1C,$5F,$6B,$EB,$97,$D7
	.dc.b	$97,$C4,$FF,$87,$23,$0D,$CB,$C3,$5C,$1B,$8F,$D5,$BE,$38,$DB,$E0
	.dc.b	$F9,$2F,$83,$6B,$4C,$3E,$A5,$7A,$70,$F2,$57,$09,$F5,$5F,$46,$9E
	.dc.b	$18,$D5,$71,$C3,$E3,$D6,$E3,$CB,$5D,$AB,$C7,$BF,$EC,$BA,$BF,$C7
	.dc.b	$79,$7E,$9F,$E6,$2E,$FF,$BB,$6E,$1F,$C0,$2A,$DF,$18,$FD,$CC,$E3
	.dc.b	$88,$DA,$93,$A2,$1F,$B3,$9F,$EA,$8F,$F0,$28,$E5,$32,$30,$7F,$1E
	.dc.b	$FA,$17,$CB,$F8,$97,$CB,$D3,$11,$27,$84,$65,$E5,$19,$7F,$82,$4F
	.dc.b	$27,$66,$8F,$E5,$BD,$AF,$85,$F8,$2B,$C4,$A6,$22,$4F,$08,$CB,$CA
	.dc.b	$38,$FC,$2F,$2F,$85,$93,$DA,$32,$F8,$23,$2F,$F1,$9B,$4E,$A3,$13
	.dc.b	$AE,$0F,$E5,$D5,$A9,$E8,$73,$21,$CE,$BD,$7E,$E7,$B3,$E1,$64,$FC
	.dc.b	$08,$CB,$F0,$A3,$2F,$8E,$3F,$07,$E3,$BE,$86,$92,$7F,$AA,$C3,$E8
	.dc.b	$71,$1F,$EA,$69,$B9,$7F,$A7,$A7,$9F,$D0,$FE,$43,$C9,$7C,$2B,$C6
	.dc.b	$C3,$71,$EB,$5C,$1F,$F0,$8D,$70,$AD,$99,$8F,$E8,$E3,$E3,$1C,$1F
	.dc.b	$9E,$8C,$36,$33,$1F,$D1,$BE,$87,$C6,$FF,$48,$8C,$BF,$E6,$C3,$89
	.dc.b	$FE,$A3,$72,$FF,$9C,$1A,$FF,$AA,$BE,$85,$F2,$9E,$16,$C7,$F5,$46
	.dc.b	$BE,$8E,$EF,$0B,$EA,$8C,$BF,$E7,$C6,$BD,$67,$FD,$68,$A3,$2F,$F2
	.dc.b	$AF,$A1,$E7,$BF,$FA,$F5,$7C,$5F,$C3,$6C,$FF,$66,$77,$E0,$7F,$B0
	.dc.b	$9B,$1C,$E7,$81,$C7,$82,$47,$18,$CB,$CE,$3E,$87,$C5,$BE,$6E,$92
	.dc.b	$7F,$BA,$FC,$6B,$E7,$31,$1F,$3E,$B8,$3F,$ED,$16,$B9,$AF,$F6,$A7
	.dc.b	$88,$E8,$7D,$0F,$E6,$3C,$89,$7F,$96,$8F,$CD,$98,$7F,$DF,$71,$87
	.dc.b	$FC,$0F,$F4,$3F,$01,$3F,$19,$FC,$6B,$78,$3F,$DD,$0E,$C9,$B7,$E3
	.dc.b	$DD,$70,$7F,$BB,$1D,$6F,$99,$97,$E0,$8F,$D7,$F8,$BF,$B0,$FA,$19
	.dc.b	$3F,$AE,$8C,$BF,$B0,$8C,$BF,$E2,$1D,$3F,$9A,$BF,$E2,$0F,$A1,$7C
	.dc.b	$87,$E8,$5B,$19,$8F,$D2,$7D,$0B,$CF,$D4,$7E,$32,$B7,$5F,$A1,$6B
	.dc.b	$FE,$CE,$6B,$FE,$5B,$BF,$F6,$23,$FC,$13,$CB,$FE,$12,$7F,$0A,$F4
	.dc.b	$64,$64,$87,$A5,$CB,$EB,$8E,$2F,$F0,$E8,$FE,$AB,$F3,$BF,$EC,$48
	.dc.b	$FE,$22,$7F,$16,$33,$F0,$51,$C9,$0F,$F2,$34,$71,$6F,$9B,$D3,$CE
	.dc.b	$91,$FE,$70,$9F,$EA,$6D,$9C,$BC,$A7,$C7,$1F,$2B,$AB,$E5,$F4,$CA
	.dc.b	$FF,$28,$EF,$F9,$D6,$CE,$5F,$E2,$F1,$F9,$CE,$6F,$F3,$94,$23,$83
	.dc.b	$A9,$E1,$3A,$9E,$1C,$DF,$E2,$8A,$E1,$5F,$18,$8F,$AD,$F9,$37,$F2
	.dc.b	$BE,$87,$E5,$FC,$C4,$AF,$81,$1F,$3B,$E8,$64,$7C,$37,$4F,$CA,$FA
	.dc.b	$1F,$A6,$47,$E4,$7C,$65,$7E,$7F,$D0,$FD,$83,$1A,$FF,$A1,$F8,$11
	.dc.b	$99,$FA,$0E,$B5,$FF,$53,$4F,$F9,$26,$7E,$A7,$89,$F8,$23,$F0,$7E
	.dc.b	$2C,$8C,$90,$FB,$EE,$5F,$C0,$C7,$17,$EF,$47,$E3,$3E,$7F,$F1,$BE
	.dc.b	$C4,$9F,$C1,$46,$5F,$C2,$46,$5F,$C5,$C6,$5F,$C6,$C6,$72,$F7,$4F
	.dc.b	$E6,$3C,$8F,$CC,$FC,$FC,$BF,$C7,$46,$5F,$C8,$46,$5F,$CB,$C6,$5F
	.dc.b	$CC,$C6,$6B,$E5,$4F,$E9,$3E,$7B,$F4,$BF,$3F,$2F,$F3,$51,$97,$F3
	.dc.b	$91,$97,$F4,$71,$97,$F4,$B1,$9A,$FA,$D3,$FB,$0E,$FF,$D8,$FB,$B2
	.dc.b	$FF,$4D,$19,$7F,$51,$19,$7F,$5F,$19,$7F,$63,$19,$AF,$C1,$4F,$F9
	.dc.b	$43,$E4,$9F,$0B,$7E,$1A,$61,$FE,$E2,$9C,$B6,$72,$FF,$B0,$61,$3C
	.dc.b	$E7,$FC,$77,$E1,$CA,$FF,$1F,$BD,$FD,$2C,$47,$F8,$C6,$2E,$3F,$ED
	.dc.b	$BC,$3F,$C2,$53,$FE,$35,$F3,$BF,$C8,$BF,$3D,$23,$FC,$93,$1F,$E4
	.dc.b	$83,$0F,$F8,$EE,$30,$FF,$93,$BC,$AF,$F2,$6F,$C8,$C8,$C1,$0F,$F2
	.dc.b	$13,$97,$FC,$89,$1F,$E5,$43,$18,$7F,$90,$A7,$13,$FE,$5A,$F9,$EF
	.dc.b	$E0,$5F,$08,$E2,$FF,$55,$FA,$2F,$79,$5A,$9C,$C6,$2F,$FD,$63,$7C
	.dc.b	$55,$F1,$F3,$F8,$46,$2B,$82,$AE,$15,$DF,$0D,$B6,$7A,$E1,$E0,$C3
	.dc.b	$9C,$3E,$87,$CB,$C9,$FE,$2F,$8C,$61,$FD,$55,$FF,$19,$4E,$9B,$19
	.dc.b	$7F,$91,$1F,$EC,$5E,$C7,$CB,$73,$F8,$5F,$F1,$9B,$63,$33,$FE,$3D
	.dc.b	$9C,$BD,$BD,$6F,$EB,$6F,$4C,$11,$88,$FE,$61,$18,$F3,$BD,$0D,$8F
	.dc.b	$1E,$4F,$C7,$FF,$CA,$7E,$1F,$C2,$9F,$95,$82,$31,$5B,$F6,$F1,$5F
	.dc.b	$37,$7F,$AF,$78,$E1,$FA,$D7,$BC,$38,$E3,$DB,$0E,$07,$DA,$FC,$71
	.dc.b	$F2,$3C,$72,$FF,$77,$E1,$C6,$B6,$7F,$AC,$B8,$BC,$3C,$DF,$E9,$0F
	.dc.b	$A1,$B9,$FF,$4F,$F4,$7F,$CF,$2F,$FA,$AA,$33,$7F,$EF,$A9,$CB,$FE
	.dc.b	$9D,$8C,$BD,$67,$81,$FF,$00,$2B,$FB,$0D,$70,$6D,$F4,$E6,$68,$EC
	.dc.b	$E7,$7B,$C1,$8B,$E7,$5F,$86,$BD,$5F,$2C,$61,$E9,$8B,$E3,$B3,$5B
	.dc.b	$C3,$C8,$74,$BE,$0E,$30,$ED,$3B,$5A,$FD,$8D,$3C,$F8,$74,$CB,$BF
	.dc.b	$C7,$7C,$6C,$5E,$63,$E7,$F3,$1D,$4C,$B8,$CB,$AC,$7B,$8C,$9E,$3C
	.dc.b	$7E,$2D,$F5,$BE,$93,$E7,$AF,$93,$C7,$F2,$F8,$3F,$DB,$D3,$9D,$FF
	.dc.b	$84,$3F,$CB,$FE,$33,$EB,$C6,$2F,$F9,$74,$62,$F9,$3F,$73,$C6,$FF
	.dc.b	$7C,$DF,$F3,$7F,$17,$1C,$7D,$DB,$19,$3C,$93,$C3,$0F,$E7,$53,$A7
	.dc.b	$FD,$77,$19,$E1,$FE,$69,$E1,$E5,$FD,$2F,$27,$DE,$F4,$FD,$9E,$2F
	.dc.b	$DA,$6E,$5F,$5E,$74,$FF,$4B,$9D,$34,$D7,$32,$B3,$97,$F4,$DE,$7B
	.dc.b	$8F,$82,$BF,$D7,$B7,$FD,$57,$C4,$FC,$4F,$87,$F3,$5B,$19,$3A,$C6
	.dc.b	$5A,$4C,$3F,$AD,$77,$F0,$7E,$F9,$F8,$EE,$2F,$B1,$D7,$F9,$66,$D7
	.dc.b	$A1,$B3,$C3,$6F,$D0,$C6,$33,$F4,$3C,$6B,$FB,$67,$CB,$DB,$E1,$F8
	.dc.b	$53,$E5,$3F,$0D,$5F,$A8,$8C,$3F,$DC,$CC,$46,$9F,$9D,$1A,$9F,$EE
	.dc.b	$91,$A3,$33,$FB,$BF,$D4,$DF,$49,$9A,$F7,$F6,$A5,$F4,$9B,$3C,$5F
	.dc.b	$21,$71,$EE,$DB,$F4,$B7,$5F,$D9,$5F,$E8,$A7,$35,$FA,$78,$C4,$FE
	.dc.b	$2B,$35,$F9,$7E,$93,$CD,$E6,$DC,$38,$3F,$0F,$A3,$19,$7C,$86,$E3
	.dc.b	$E0,$F1,$BD,$DF,$45,$EF,$6E,$1F,$B3,$FC,$27,$FB,$9B,$FC,$2F,$CD
	.dc.b	$7E,$12,$F5,$BF,$93,$D6,$B3,$FC,$1D,$6A,$73,$D7,$FC,$C2,$31,$87
	.dc.b	$58,$C6,$1C,$E3,$07,$77,$17,$F8,$E1,$87,$7C,$FF,$A8,$E1,$D2,$7E
	.dc.b	$AF,$0F,$F2,$7C,$3E,$57,$95,$F6,$FC,$6F,$B8,$C8,$C9,$F3,$E3,$15
	.dc.b	$EA,$FC,$C1,$D9,$E9,$6D,$7C,$37,$FB,$7B,$63,$44,$75,$DB,$6B,$D5
	.dc.b	$F4,$DF,$72,$7F,$A6,$7D,$67,$E4,$F5,$FF,$56,$CE,$6B,$F9,$BC,$E6
	.dc.b	$BF,$D5,$53,$9A,$4E,$6B,$FA,$6C,$E6,$31,$08,$FC,$73,$FE,$12,$6B
	.dc.b	$D3,$CC,$FF,$5E,$B6,$FF,$C2,$EF,$F5,$7F,$96,$FF,$1B,$47,$F9,$7F
	.dc.b	$A7,$FA,$BC,$62,$7F,$CB,$D1,$8E,$8F,$E7,$4D,$AE,$07,$E7,$7E,$7F
	.dc.b	$F5,$C7,$EB,$FE,$7A,$C6,$49,$7F,$72,$32,$FC,$AF,$92,$FC,$67,$D1
	.dc.b	$FF,$9D,$2B,$FA,$56,$89,$CF,$B9,$FD,$F2,$31,$3F,$9B,$AE,$17,$CE
	.dc.b	$31,$5B,$BE,$31,$D1,$B3,$F8,$ED,$3F,$83,$1F,$C1,$FE,$66,$D7,$EE
	.dc.b	$62,$CF,$E5,$26,$F0,$7F,$BA,$BE,$95,$F0,$5E,$AF,$9D,$FF,$08,$C6
	.dc.b	$31,$F3,$46,$27,$51,$FD,$9E,$75,$87,$EC,$34,$5F,$F8,$6B,$BB,$FC
	.dc.b	$21,$2E,$B5,$3D,$FD,$79,$DF,$29,$B9,$7C,$56,$DF,$56,$E9,$FB,$0B
	.dc.b	$E5,$5F,$DE,$71,$EF,$8F,$6B,$FF,$15,$FD,$23,$E1,$7F,$81,$7D,$CF
	.dc.b	$F2,$8C,$3F,$C0,$13,$97,$F8,$74,$67,$3F,$E0,$13,$97,$F8,$6C,$67
	.dc.b	$3F,$F2,$8C,$27,$51,$FC,$12,$75,$5F,$E3,$0A,$79,$BF,$E1,$C8,$FE
	.dc.b	$39,$EB,$FC,$AB,$E7,$39,$AF,$E3,$73,$9A,$E5,$39,$AF,$E2,$33,$9A
	.dc.b	$FF,$1E,$CE,$6B,$FC,$7F,$9A,$B5,$E8,$39,$1C,$9F,$E1,$FE,$2F,$D5
	.dc.b	$B7,$0F,$F0,$24,$E9,$FF,$67,$4E,$9A,$7F,$47,$F9,$8F,$EB,$1F,$3B
	.dc.b	$FD,$65,$B1,$93,$FA,$64,$65,$A7,$6E,$7F,$F2,$D7,$6F,$C6,$8C,$1F
	.dc.b	$E3,$97,$1B,$7F,$79,$C7,$F5,$6F,$F6,$48,$CE,$FF,$E2,$09,$CB,$FE
	.dc.b	$4F,$8C,$D7,$F9,$5E,$5F,$2F,$E0,$8F,$E1,$6D,$7E,$0B,$6F,$C0,$DB
	.dc.b	$EC,$F0,$9F,$F2,$77,$B6,$EE,$C5,$F0,$9E,$66,$A3,$37,$FE,$9A,$F7
	.dc.b	$DF,$92,$BE,$47,$FC,$E3,$EA,$FF,$A2,$27,$31,$8C,$3F,$CD,$D1,$87
	.dc.b	$C3,$3B,$AE,$51,$83,$FC,$7D,$F8,$5F,$E1,$FD,$67,$D2,$F9,$E7,$3E
	.dc.b	$17,$FA,$2E,$31,$AF,$F8,$9E,$F8,$46,$34,$FF,$26,$C6,$35,$FF,$44
	.dc.b	$B7,$0E,$67,$F9,$C6,$B8,$3F,$E8,$7F,$AC,$FD,$2E,$1F,$EE,$AF,$A5
	.dc.b	$78,$FF,$E9,$29,$CE,$89,$CD,$FF,$95,$E7,$31,$8B,$FF,$1E,$46,$2F
	.dc.b	$97,$B3,$2F,$C5,$CD,$6F,$73,$75,$FF,$4A,$67,$F8,$8B,$F3,$4F,$FA
	.dc.b	$D2,$BF,$E0,$08,$FA,$67,$4F,$FB,$B3,$BB,$FD,$A7,$18,$9F,$F7,$7F
	.dc.b	$C4,$7F,$AC,$23,$11,$FB,$4C,$31,$7F,$D7,$AA,$CF,$F0,$7B,$E7,$39
	.dc.b	$8F,$03,$DE,$B7,$1F,$C3,$6A,$BF,$CB,$E2,$A8,$CD,$74,$8F,$29,$D0
	.dc.b	$F0,$4F,$09,$DB,$63,$47,$A4,$A7,$C4,$BF,$6B,$E1,$32,$78,$47,$2B
	.dc.b	$E3,$FD,$8D,$8F,$1B,$5F,$61,$C2,$3F,$95,$7C,$67,$FB,$13,$5F,$F7
	.dc.b	$AC,$63,$E1,$FF,$D9,$B5,$B3,$FD,$BF,$5B,$70,$1C,$5F,$77,$58,$7D
	.dc.b	$65,$F2,$3F,$CE,$63,$0D,$D7,$C1,$97,$D2,$AE,$31,$E6,$9D,$61,$E2
	.dc.b	$6B,$E1,$C7,$CF,$87,$76,$D7,$19,$E9,$5C,$2F,$AC,$7B,$2B,$6F,$53
	.dc.b	$67,$8D,$CB,$DF,$0C,$BF,$6A,$7D,$AE,$E3,$F2,$3E,$13,$8F,$F6,$73
	.dc.b	$FA,$FE,$3F,$EB,$55,$FE,$CC,$9C,$CF,$FA,$53,$8B,$F9,$5B,$B2,$1F
	.dc.b	$B4,$AE,$3B,$A1,$CB,$FC,$4B,$B7,$30,$CB,$C5,$78,$AF,$E0,$55,$79
	.dc.b	$BB,$DF,$53,$BE,$6C,$7B,$5F,$09,$F0,$4E,$5F,$E2,$17,$E0,$C7,$BC
	.dc.b	$7F,$B8,$FF,$0B,$FF,$4F,$BC,$30,$E7,$F4,$EF,$8C,$F2,$36,$78,$7A
	.dc.b	$5C,$EF,$8C,$7F,$AE,$27,$4D,$AD,$54,$8E,$C7,$C4,$9D,$E5,$E7,$CD
	.dc.b	$18,$C7,$E5,$64,$CB,$BD,$FF,$4B,$EE,$FF,$7B,$63,$FD,$53,$AB,$E5
	.dc.b	$78,$3F,$EC,$88,$FE,$93,$3A,$7F,$DD,$7D,$1F,$D4,$BE,$1F,$77,$27
	.dc.b	$71,$F1,$FF,$08,$F1,$47,$D7,$6E,$67,$3B,$32,$78,$DC,$78,$66,$1F
	.dc.b	$03,$0E,$11,$E6,$9C,$CF,$FB,$82,$9E,$AF,$A8,$CE,$DF,$6F,$0E,$35
	.dc.b	$FD,$CA,$74,$F5,$CB,$CA,$3E,$33,$E1,$F9,$12,$75,$1E,$D3,$0F,$86
	.dc.b	$30,$FF,$B9,$7C,$2F,$CA,$7B,$BF,$92,$CB,$FD,$D7,$CF,$89,$95,$FE
	.dc.b	$EB,$8D,$98,$53,$E9,$C7,$FC,$01,$1F,$82,$DD,$EE,$3E,$3F,$F4,$4D
	.dc.b	$9F,$67,$95,$F4,$CC,$CC,$8C,$9F,$8C,$1F,$AD,$7F,$94,$DF,$F6,$E7
	.dc.b	$EB,$7F,$EE,$B7,$4D,$7F,$DC,$11,$BC,$BF,$CB,$ED,$EA,$FF,$2E,$56
	.dc.b	$E7,$FC,$C6,$D8,$FF,$30,$4E,$EB,$F8,$06,$8F,$A0,$FD,$FC,$3F,$5C
	.dc.b	$FF,$54,$3E,$76,$BF,$EF,$19,$E0,$FA,$DA,$FF,$BD,$63,$78,$FF,$9D
	.dc.b	$1B,$87,$F9,$C6,$B7,$3F,$E7,$8F,$B7,$FF,$76,$DF,$C5,$FC,$29,$F0
	.dc.b	$9E,$46,$C7,$C9,$9D,$BF,$63,$0F,$F3,$E4,$62,$FF,$D0,$75,$C2,$B4
	.dc.b	$7F,$1F,$71,$1E,$1F,$83,$F8,$35,$C6,$FF,$D0,$D5,$A6,$E1,$AA,$F8
	.dc.b	$FF,$8F,$3E,$9F,$E4,$FC,$AF,$FA,$77,$3F,$BC,$D7,$C2,$27,$83,$F8
	.dc.b	$36,$BA,$8E,$19,$FF,$A7,$1B,$8F,$FA,$66,$B7,$3F,$C4,$F9,$FE,$5C
	.dc.b	$FF,$2E,$6D,$FF,$3C,$6F,$29,$E3,$3B,$E0,$FD,$49,$2B,$F4,$13,$A7
	.dc.b	$FA,$37,$C3,$3E,$47,$F0,$FF,$72,$F6,$29,$F8,$A9,$D5,$7E,$7D,$AF
	.dc.b	$C2,$9E,$0F,$E8,$DA,$FC,$68,$DE,$7F,$EB,$C6,$E5,$FE,$B9,$AD,$CF
	.dc.b	$FB,$01,$B1,$FE,$BD,$9D,$D7,$C1,$3F,$82,$90,$F6,$FA,$B5,$FC,$2E
	.dc.b	$9F,$15,$C6,$1F,$36,$7F,$97,$7A,$1F,$C0,$3E,$BA,$57,$F1,$B8,$CB
	.dc.b	$FC,$1F,$B1,$3A,$8F,$C8,$F8,$5F,$CC,$EB,$F5,$FF,$52,$F9,$DF,$E0
	.dc.b	$55,$FE,$D5,$8C,$38,$0E,$1D,$B5,$FE,$14,$E7,$2F,$F7,$1F,$D3,$FF
	.dc.b	$01,$9F,$EF,$CD,$79,$D7,$88,$FE,$03,$19,$35,$5C,$F6,$FF,$6C,$FD
	.dc.b	$CB,$EA,$7F,$C0,$BB,$23,$18,$FF,$68,$8C,$4E,$A7,$FA,$AC,$62,$75
	.dc.b	$1C,$1F,$DA,$C6,$5B,$1F,$D6,$3F,$54,$FE,$8D,$1F,$CC,$DA,$FD,$86
	.dc.b	$E7,$FD,$D5,$E8,$DF,$1F,$FC,$31,$1A,$C7,$FC,$26,$DA,$FF,$10,$BC
	.dc.b	$23,$E9,$71,$7C,$4F,$12,$57,$F3,$9D,$13,$F6,$8C,$C7,$EF,$B5,$FE
	.dc.b	$92,$FF,$00,$9D,$BC,$67,$79,$7D,$6B,$E5,$3E,$DF,$A9,$F3,$FD,$54
	.dc.b	$8C,$1F,$DA,$7D,$0F,$5D,$48,$C1,$F1,$5F,$7F,$D5,$48,$C1,$F2,$7D
	.dc.b	$0F,$F1,$77,$EB,$AF,$85,$A7,$C6,$8B,$3F,$3F,$EA,$BE,$49,$87,$E8
	.dc.b	$38,$74,$7F,$82,$6B,$A3,$7E,$37,$B9,$1F,$B5,$F5,$75,$FB,$5F,$57
	.dc.b	$F5,$8C,$6F,$FE,$0B,$B5,$E6,$BF,$05,$5A,$E3,$FF,$18,$D7,$CE,$9F
	.dc.b	$D7,$B6,$F9,$70,$FE,$0B,$E8,$7E,$0B,$F8,$4E,$B9,$7E,$98,$CB,$F8
	.dc.b	$F8,$CB,$D2,$3F,$CA,$DF,$79,$FE,$39,$8F,$DE,$B6,$FF,$C4,$4D,$C7
	.dc.b	$F1,$FF,$07,$C3,$DA,$FC,$9F,$C8,$CC,$3F,$C6,$17,$B7,$FC,$E0,$DF
	.dc.b	$0D,$D7,$FD,$0A,$7F,$2F,$5C,$DE,$3F,$3D,$E4,$23,$FA,$7F,$8C,$AF
	.dc.b	$E9,$FE,$AF,$F4,$66,$35,$F3,$D6,$27,$F5,$D3,$9D,$7F,$D0,$D0,$FC
	.dc.b	$71,$E1,$7F,$82,$1B,$1F,$E1,$D6,$BF,$B5,$F2,$B5,$3F,$B2,$F9,$59
	.dc.b	$3B,$8C,$BF,$BD,$8C,$BF,$DA,$2B,$9E,$5F,$E8,$78,$FF,$2D,$72,$7F
	.dc.b	$8C,$BE,$D2,$47,$F8,$B7,$DF,$4F,$F8,$71,$B1,$FC,$D6,$76,$FF,$74
	.dc.b	$8C,$BF,$E5,$DF,$77,$FD,$43,$A3,$B1,$F1,$78,$6D,$7F,$91,$6F,$FC
	.dc.b	$3A,$FF,$AA,$9B,$5C,$6B,$81,$FE,$80,$BF,$C4,$FC,$8C,$BF,$C0,$BF
	.dc.b	$E8,$4F,$6B,$11,$F8,$C8,$BA,$7F,$A4,$71,$FF,$2F,$63,$F5,$3F,$25
	.dc.b	$71,$7F,$8E,$9B,$87,$F9,$3D,$AF,$F9,$E7,$D7,$D4,$FF,$A4,$BE,$92
	.dc.b	$4E,$E3,$2F,$B6,$32,$FF,$86,$AB,$9E,$3F,$EB,$9F,$1D,$F9,$23,$E5
	.dc.b	$3F,$0F,$F2,$7F,$E0,$29,$CC,$63,$A3,$FD,$A1,$18,$C7,$FD,$A5,$5F
	.dc.b	$E0,$8E,$DF,$F4,$B7,$B3,$B9,$56,$D5,$6F,$81,$D7,$BC,$3F,$D3,$CF
	.dc.b	$1B,$E1,$5F,$E1,$2F,$AD,$7D,$27,$4C,$7F,$B8,$C6,$2B,$FB,$83,$63
	.dc.b	$33,$B7,$FC,$F1,$F2,$7F,$8E,$FA,$EF,$EE,$CD,$8F,$EF,$6D,$77,$D8
	.dc.b	$9E,$BE,$37,$E9,$A7,$F5,$B8,$7F,$4C,$F0,$32,$48,$FF,$43,$64,$FA
	.dc.b	$EF,$F4,$97,$4F,$C8,$E2,$F9,$71,$8A,$F8,$B1,$FE,$F0,$E1,$7D,$3F
	.dc.b	$FB,$BF,$83,$EA,$5E,$DB,$F1,$39,$BE,$6D,$C3,$F3,$97,$72,$F1,$4E
	.dc.b	$67,$FD,$75,$F4,$1F,$B5,$7F,$83,$FB,$08,$C4,$0D,$E9,$FE,$1D,$9D
	.dc.b	$36,$4F,$FA,$9E,$33,$3F,$ED,$1F,$7B,$2F,$F9,$B2,$33,$D5,$FC,$15
	.dc.b	$B5,$E2,$BD,$CF,$EA,$6E,$CE,$4F,$F7,$07,$D5,$BE,$43,$38,$BD,$F4
	.dc.b	$E6,$BF,$80,$7C,$3E,$A7,$37,$6B,$57,$FE,$40,$6E,$3F,$B9,$BD,$57
	.dc.b	$6E,$BF,$F7,$A7,$D7,$78,$32,$7A,$5F,$EE,$E7,$F8,$74,$66,$31,$97
	.dc.b	$FB,$CE,$31,$F2,$1C,$F9,$91,$8B,$FE,$49,$C9,$F4,$AD,$5F,$E7,$B9
	.dc.b	$FF,$27,$36,$3F,$CA,$6D,$7F,$DA,$CE,$74,$FE,$64,$D9,$FB,$99,$7F
	.dc.b	$B9,$BD,$6F,$93,$7D,$B0,$BF,$5D,$FE,$FF,$BF,$F3,$37,$C4,$E3,$93
	.dc.b	$E1,$71,$A3,$13,$FE,$61,$AE,$18,$7F,$99,$9A,$FF,$BA,$5A,$FF,$56
	.dc.b	$77,$A7,$F4,$76,$CF,$9A,$DF,$5D,$DA,$FE,$B7,$B7,$E2,$FA,$CF,$B4
	.dc.b	$DB,$FC,$2B,$F9,$AE,$6F,$C1,$B7,$98,$D2,$4F,$03,$FD,$09,$CF,$D1
	.dc.b	$FC,$7F,$83,$F0,$3C,$3F,$C4,$36,$32,$7F,$A1,$30,$FC,$8C,$77,$24
	.dc.b	$67,$0F,$EB,$EF,$0C,$BF,$D3,$7A,$7F,$95,$A3,$1B,$7E,$A3,$0F,$CD
	.dc.b	$F0,$FD,$0F,$8C,$FC,$C3,$6B,$F4,$6F,$B4,$8F,$76,$E1,$F9,$D7,$EC
	.dc.b	$FD,$17,$29,$FD,$F6,$1D,$FE,$B5,$F5,$DF,$7A,$7F,$51,$18,$7F,$D5
	.dc.b	$06,$23,$E7,$46,$70,$FF,$08,$9F,$B3,$70,$FF,$85,$32,$46,$23,$F8
	.dc.b	$74,$62,$7A,$57,$FB,$73,$E3,$6D,$7F,$B0,$3E,$39,$EF,$7E,$CD,$B5
	.dc.b	$EA,$39,$E2,$F9,$BD,$57,$A9,$F3,$92,$7E,$C4,$ED,$FF,$5E,$F5,$FF
	.dc.b	$BC,$7B,$B8,$47,$FB,$B9,$B8,$F8,$FD,$1F,$17,$CB,$F8,$DB,$5E,$39
	.dc.b	$F8,$5D,$3F,$18,$91,$E0,$7F,$97,$E1,$FE,$FF,$F1,$BD,$9E,$77,$B5
	.dc.b	$B5,$CE,$DF,$5B,$FD,$92,$3F,$D7,$5F,$5C,$F7,$BF,$77,$39,$AF,$F5
	.dc.b	$BC,$E6,$B9,$4E,$67,$EB,$B5,$FF,$24,$B5,$FE,$84,$D7,$C3,$D9,$D8
	.dc.b	$FE,$63,$E6,$AD,$F4,$79,$C3,$E1,$68,$E4,$79,$5F,$64,$FB,$5D,$B6
	.dc.b	$B7,$3A,$6C,$E9,$A7,$FB,$67,$EB,$9D,$DE,$26,$C6,$7D,$1C,$4C,$8E
	.dc.b	$0E,$E3,$1B,$3B,$7F,$CB,$1F,$5B,$FC,$E1,$FE,$A7,$F5,$DF,$9D,$73
	.dc.b	$D3,$FD,$41,$46,$71,$FF,$38,$D7,$F4,$0B,$FD,$1B,$5C,$DF,$F3,$D8
	.dc.b	$C4,$7F,$6F,$7C,$13,$A3,$F1,$5D,$EF,$BC,$FB,$55,$E7,$9D,$34,$FE
	.dc.b	$C5,$EE,$BB,$3F,$9B,$F7,$7F,$61,$8C,$57,$D4,$E2,$FD,$05,$F1,$C5
	.dc.b	$5A,$8C,$37,$5F,$EF,$7C,$7F,$6F,$CE,$E9,$9F,$EC,$FD,$99,$9F,$DA
	.dc.b	$FA,$4B,$97,$EE,$66,$E3,$CB,$99,$58,$BF,$93,$F0,$1F,$8C,$FA,$D7
	.dc.b	$D7,$FF,$56,$E3,$4E,$63,$0F,$CB,$25,$7F,$12,$CF,$E6,$46,$1F,$8C
	.dc.b	$2B,$9C,$FE,$DF,$B1,$87,$DB,$AE,$33,$F9,$4E,$0F,$E2,$07,$F1,$1F
	.dc.b	$AB,$B1,$92,$4F,$F1,$78,$CB,$E0,$FA,$F7,$C4,$FB,$7E,$47,$FB,$8C
	.dc.b	$62,$FC,$58,$FF,$BF,$DF,$E0,$11,$9B,$E5,$7F,$17,$DA,$E2,$FF,$7A
	.dc.b	$F1,$3F,$C5,$73,$FE,$C0,$C3,$FD,$E7,$87,$F0,$99,$CE,$9E,$EF,$5F
	.dc.b	$87,$A5,$FE,$0F,$9F,$F1,$CE,$1F,$CE,$9A,$FF,$B3,$1A,$FF,$91,$5E
	.dc.b	$EF,$F1,$18,$FB,$1E,$B7,$4C,$3F,$8C,$7D,$7B,$E3,$3F,$83,$FD,$7F
	.dc.b	$F3,$8C,$3E,$9C,$E6,$BE,$94,$E7,$A3,$FD,$95,$39,$AD,$DF,$2F,$07
	.dc.b	$FC,$33,$39,$AF,$F6,$87,$A7,$FE,$34,$C7,$EE,$7D,$7F,$C9,$9C,$FB
	.dc.b	$BF,$C6,$23,$13,$FE,$E8,$AE,$19,$7E,$19,$B7,$FE,$5A,$E0,$5F,$FA
	.dc.b	$EA,$FF,$9A,$56,$9F,$EB,$7C,$EE,$7F,$F0,$93,$63,$FC,$0C,$DC,$7F
	.dc.b	$B3,$F8,$3F,$B0,$AD,$9F,$B7,$3F,$AF,$7D,$3D,$BF,$E4,$BF,$19,$86
	.dc.b	$9F,$CA,$23,$FB,$24,$66,$FF,$B4,$37,$4F,$F0,$66,$3F,$A0,$9F,$0B
	.dc.b	$5F,$EE,$CD,$7D,$7D,$9E,$A3,$F8,$51,$FE,$59,$EE,$B8,$FF,$16,$A7
	.dc.b	$C6,$7C,$02,$47,$C0,$6D,$FF,$0C,$7F,$CC,$5E,$0F,$F1,$56,$4F,$37
	.dc.b	$2D,$24,$F1,$7F,$DF,$93,$C2,$3F,$91,$7C,$13,$EA,$1F,$31,$FE,$9C
	.dc.b	$8C,$36,$33,$7F,$C1,$E7,$2D,$8C,$D7,$46,$E9,$FE,$FD,$DF,$FD,$23
	.dc.b	$7C,$2B,$F9,$57,$D7,$FE,$31,$79,$0F,$AD,$FF,$78,$65,$FC,$F2,$B5
	.dc.b	$39,$D3,$FD,$2D,$1F,$BF,$BF,$F1,$43,$5F,$F1,$3B,$5F,$F5,$03,$78
	.dc.b	$CF,$85,$BE,$0F,$FA,$67,$1F,$C8,$69,$FE,$4D,$AD,$7D,$77,$EB,$F3
	.dc.b	$FF,$42,$FD,$77,$F3,$47,$FA,$8C,$67,$2F,$5C,$E5,$FE,$91,$1F,$BB
	.dc.b	$7F,$C7,$CD,$7F,$C7,$AD,$7F,$D6,$CD,$C8,$F5,$6F,$BB,$E7,$E5,$FF
	.dc.b	$5C,$4E,$49,$19,$C3,$CA,$FF,$A6,$AF,$FD,$B1,$F1,$3F,$ED,$9F,$AF
	.dc.b	$74,$7F,$95,$3B,$3F,$DA,$71,$8A,$FF,$2B,$D6,$EB,$58,$59,$C9,$FE
	.dc.b	$1C,$8C,$B4,$CF,$AF,$CD,$FF,$74,$61,$FE,$E0,$8C,$47,$90,$61,$89
	.dc.b	$F1,$D7,$F5,$7C,$3C,$15,$CB,$0A,$F1,$C3,$84,$7F,$BD,$5E,$B0,$9C
	.dc.b	$57,$9A,$73,$18,$9F,$F7,$E0,$BF,$35,$78,$63,$D8,$78,$4A,$7A,$79
	.dc.b	$3F,$C7,$AF,$C4,$6B,$D2,$8A,$32,$F1,$BF,$0F,$C7,$7F,$BA,$38,$77
	.dc.b	$5C,$A3,$15,$C6,$B7,$DF,$FE,$35,$8C,$1D,$A7,$75,$C6,$75,$5F,$4E
	.dc.b	$5F,$33,$F0,$23,$0D,$CB,$AE,$3F,$17,$15,$F8,$AF,$71,$98,$E6,$DC
	.dc.b	$3B,$DF,$28,$FB,$AF,$BB,$0F,$9F,$AF,$29,$C9,$E9,$AF,$64,$79,$DE
	.dc.b	$71,$ED,$9F,$33,$4F,$06,$31,$E6,$7A,$23,$BD,$F8,$63,$F1,$78,$7C
	.dc.b	$CC,$3F,$41,$C3,$FB,$0E,$DF,$C1,$E3,$FE,$CF,$E1,$FF,$19,$E1,$FC
	.dc.b	$02,$73,$C1,$F7,$3A,$3F,$8D,$76,$62,$BF,$77,$57,$1E,$91,$C4,$FF
	.dc.b	$40,$F3,$FF,$44,$70,$7F,$60,$C7,$BB,$6B,$FC,$70,$F4,$3F,$DD,$91
	.dc.b	$AC,$BF,$CA,$3C,$DC,$E7,$FC,$C1,$92,$31,$3E,$BB,$F2,$7D,$77,$FB
	.dc.b	$0E,$FF,$CD,$6D,$7F,$CC,$0D,$9F,$F2,$EE,$3F,$ED,$A7,$FD,$E5,$1B
	.dc.b	$9F,$F6,$D3,$5F,$F6,$CB,$56,$48,$C3,$7D,$4F,$91,$93,$EB,$DF,$1D
	.dc.b	$86,$E1,$E6,$AF,$9D,$CD,$FD,$6F,$E2,$AF,$80,$69,$D3,$E3,$BF,$CF
	.dc.b	$37,$FE,$D1,$AE,$0F,$F5,$B9,$F9,$B1,$88,$FF,$7B,$B5,$FF,$7B,$7D
	.dc.b	$7F,$89,$FF,$42,$37,$67,$27,$FB,$E2,$30,$FE,$17,$07,$8B,$F8,$39
	.dc.b	$CD,$FC,$46,$EE,$73,$BA,$F8,$D7,$FA,$82,$76,$4D,$FD,$D8,$75,$7F
	.dc.b	$0F,$87,$36,$CF,$A6,$BD,$71,$8F,$8F,$F2,$12,$3F,$DA,$B8,$7F,$39
	.dc.b	$CF,$DB,$1E,$79,$E0,$FF,$A2,$F0,$F4,$46,$23,$F9,$59,$FD,$FB,$D7
	.dc.b	$FF,$4A,$BF,$ED,$AE,$7F,$BF,$1F,$94,$D7,$F5,$4E,$A7,$F3,$2D,$C3
	.dc.b	$FD,$49,$5C,$0E,$3E,$0F,$CB,$F1,$A5,$FC,$57,$4D,$3C,$DA,$FF,$88
	.dc.b	$1B,$1F,$10,$FD,$6A,$FA,$EF,$CF,$65,$FD,$11,$B1,$FC,$E9,$AB,$EB
	.dc.b	$FF,$9E,$3F,$CF,$B2,$F8,$EF,$33,$F8,$21,$BF,$4A,$C9,$CD,$74,$CD
	.dc.b	$F2,$BC,$1B,$7D,$71,$FE,$AF,$18,$8F,$F5,$CF,$D7,$BE,$6B,$85,$F8
	.dc.b	$23,$E5,$5F,$F5,$E6,$D7,$F5,$4B,$E5,$9F,$A3,$EB,$DF,$09,$E8,$6D
	.dc.b	$78,$4F,$EB,$6E,$1F,$F6,$4F,$D7,$FF,$13,$C7,$FB,$5C,$E6,$7F,$06
	.dc.b	$D7,$35,$F4,$23,$11,$D2,$3E,$3D,$3E,$1E,$9E,$38,$F9,$7F,$5F,$EF
	.dc.b	$AF,$20,$70,$FF,$B7,$1B,$8F,$D4,$F6,$5F,$3B,$CB,$7F,$AB,$1F,$EB
	.dc.b	$3B,$E6,$D7,$94,$7E,$63,$79,$1F,$0F,$B1,$CF,$F2,$B7,$FC,$DD,$EA
	.dc.b	$BF,$37,$77,$2F,$F7,$F4,$FE,$13,$E2,$1F,$29,$49,$7E,$CF,$AF,$7B
	.dc.b	$78,$8F,$F7,$D5,$70,$7F,$A2,$B5,$CD,$7E,$72,$31,$87,$E5,$2D,$F2
	.dc.b	$9F,$94,$6D,$FC,$AA,$DE,$7F,$B4,$8F,$DB,$90,$F8,$1A,$7F,$89,$DA
	.dc.b	$64,$FD,$C9,$FB,$EF,$4F,$F9,$07,$A3,$23,$F8,$D3,$FD,$03,$E8,$BF
	.dc.b	$8D,$BF,$D0,$BE,$29,$F5,$F4,$E1,$4D,$7F,$5B,$F5,$6F,$AF,$FD,$6B
	.dc.b	$67,$F0,$6D,$78,$69,$FE,$3E,$FA,$DF,$DB,$C6,$0F,$F1,$AC,$ED,$FE
	.dc.b	$32,$D7,$FC,$8A,$D7,$75,$FC,$22,$3B,$5F,$F0,$0A,$FD,$E5,$7E,$51
	.dc.b	$AF,$71,$3C,$2F,$FC,$CB,$F5,$CF,$AF,$FF,$05,$6C,$8C,$5F,$EC,$63
	.dc.b	$13,$A9,$FE,$97,$18,$9D,$47,$07,$DD,$19,$6C,$7F,$4E,$D9,$E3,$FD
	.dc.b	$32,$4F,$E7,$63,$96,$7F,$D0,$BA,$11,$C2,$75,$5C,$1F,$F0,$8C,$70
	.dc.b	$CB,$E4,$F1,$7D,$62,$4F,$C5,$FA,$6A,$FF,$48,$3E,$2C,$7C,$D2,$3F
	.dc.b	$E2,$2D,$51,$FC,$B3,$6F,$E7,$93,$F3,$5B,$9F,$E1,$A7,$F1,$DF,$5F
	.dc.b	$F8,$FF,$AF,$97,$ED,$8F,$EE,$9E,$7F,$F6,$62,$47,$F4,$87,$FC,$63
	.dc.b	$B7,$C0,$F2,$7F,$31,$F5,$F2,$BE,$34,$7E,$6F,$D1,$FF,$20,$7D,$6B
	.dc.b	$E4,$69,$9A,$FE,$55,$E8,$27,$51,$FD,$AD,$AF,$F6,$A8,$D4,$FF,$7C
	.dc.b	$9C,$CF,$FA,$63,$1F,$A7,$C3,$FA,$B2,$47,$69,$E7,$BF,$F7,$AC,$3C
	.dc.b	$1F,$1D,$E6,$25,$7D,$FF,$AF,$AF,$86,$D3,$E1,$24,$79,$DF,$DD,$7D
	.dc.b	$7C,$8F,$04,$7C,$2E,$29,$1E,$37,$F8,$07,$D2,$E2,$3B,$47,$B4,$8F
	.dc.b	$C3,$7F,$83,$77,$C8,$E4,$7C,$AF,$AF,$7C,$3F,$F9,$BD,$B1,$98,$FE
	.dc.b	$1D,$F0,$B9,$7F,$C4,$B3,$A9,$F9,$CD,$7F,$D8,$7A,$A3,$11,$FE,$BA
	.dc.b	$8C,$4F,$F1,$9F,$90,$93,$DC,$F2,$7D,$0F,$8C,$8F,$EA,$FE,$2E,$BE
	.dc.b	$08,$F1,$91,$FF,$16,$60,$8F,$E4,$DE,$74,$9E,$A7,$F6,$0F,$AF,$FE
	.dc.b	$64,$FF,$0D,$6F,$1F,$F8,$49,$BD,$FF,$E9,$7F,$03,$F9,$ED,$7F,$B5
	.dc.b	$27,$8B,$FE,$38,$6B,$FE,$E4,$6B,$BF,$0F,$FC,$4F,$1D,$21,$D1,$FC
	.dc.b	$13,$D8,$FE,$78,$7F,$88,$27,$53,$FE,$5B,$E4,$FF,$11,$FB,$86,$9F
	.dc.b	$D0,$FE,$2A,$CF,$FA,$82,$74,$D9,$FF,$1B,$C6,$23,$FC,$9F,$87,$F9
	.dc.b	$9F,$DD,$FF,$59,$68,$EC,$FF,$2D,$3F,$EF,$23,$1F,$15,$FE,$85,$99
	.dc.b	$7F,$E1,$D9,$CB,$6B,$8B,$EE,$C3,$FC,$53,$CD,$F8,$5F,$8F,$FC,$81
	.dc.b	$27,$FC,$D3,$ED,$D7,$FA,$74,$ED,$FC,$C4,$FF,$A5,$67,$FC,$8B,$F5
	.dc.b	$DF,$E1,$49,$FF,$0D,$B5,$FE,$90,$DE,$3F,$E6,$58,$7F,$87,$32,$54
	.dc.b	$AD,$46,$21,$18,$75,$C9,$FE,$A5,$F5,$7C,$2F,$F9,$8A,$7C,$13,$FE
	.dc.b	$7A,$6B,$CD,$AB,$C6,$7D,$73,$E4,$3C,$1F,$1F,$FE,$15,$9C,$C6,$3D
	.dc.b	$7F,$3C,$62,$FD,$35,$FE,$1B,$E4,$FF,$50,$7B,$3B,$9E,$D8,$F5,$77
	.dc.b	$87,$28,$C4,$2F,$C1,$18,$7A,$6D,$FE,$67,$7F,$C4,$DF,$5A,$FA,$3F
	.dc.b	$F0,$1D,$FC,$78,$C6,$7F,$EF,$06,$C7,$F5,$56,$BF,$E9,$EF,$91,$FC
	.dc.b	$D7,$D7,$7F,$BD,$9B,$1F,$D6,$DA,$EF,$A9,$3D,$7C,$7F,$D6,$DF,$F5
	.dc.b	$CE,$FC,$12,$FF,$DA,$7F,$5C,$F9,$CF,$F5,$9E,$BF,$36,$31,$CF,$F1
	.dc.b	$2B,$F2,$DF,$4F,$AE,$2F,$9F,$5A,$9F,$8D,$18,$C3,$51,$F6,$E7,$58
	.dc.b	$7C,$8B,$CE,$E9,$CE,$1F,$8D,$F5,$DF,$39,$FE,$C6,$EB,$FF,$50,$C6
	.dc.b	$34,$FF,$65,$FB,$B9,$78,$55,$D3,$F3,$35,$B7,$FC,$4C,$DB,$FC,$FD
	.dc.b	$EE,$7F,$C2,$F5,$6B,$57,$D0,$EE,$7F,$A1,$1C,$CB,$BD,$F5,$8F,$8D
	.dc.b	$F1,$46,$1F,$27,$5F,$FB,$C7,$E1,$F3,$5B,$C7,$FB,$CB,$70,$FD,$EB
	.dc.b	$74,$FD,$AD,$70,$C3,$9F,$53,$E8,$FE,$06,$2F,$47,$FC,$FA,$FF,$BC
	.dc.b	$23,$31,$8C,$BF,$CA,$D9,$AF,$95,$FF,$17,$AD,$65,$FC,$13,$0E,$3F
	.dc.b	$0B,$F1,$EF,$FD,$1B,$7F,$EA,$66,$E3,$D3,$83,$FB,$EC,$E7,$87,$81
	.dc.b	$1F,$F4,$F4,$6A,$7F,$CB,$55,$CB,$1F,$F6,$FF,$D7,$7F,$92,$AF,$7F
	.dc.b	$0F,$F5,$E3,$17,$DE,$31,$C5,$2B,$AF,$D7,$5A,$FF,$33,$5F,$0C,$3A
	.dc.b	$DF,$D4,$C9,$7B,$9F,$F0,$E7,$13,$DB,$D7,$D5,$FD,$86,$E7,$F6,$1C
	.dc.b	$3E,$CE,$6B,$A7,$DA,$9B,$87,$F9,$7F,$2F,$C4,$F3,$7E,$22,$7F,$07
	.dc.b	$E0,$78,$FC,$2B,$97,$C6,$6E,$3F,$84,$9F,$47,$81,$8C,$BC,$F1,$6F
	.dc.b	$F9,$CF,$3B,$3F,$F2,$EC,$7E,$3A,$7F,$23,$C9,$F0,$7E,$2E,$CF,$D4
	.dc.b	$6E,$5F,$92,$9F,$8D,$F3,$D8,$CB,$E2,$C5,$BE,$F8,$AB,$17,$FA,$5F
	.dc.b	$AD,$7D,$77,$CF,$BF,$D2,$C6,$1F,$C1,$98,$88,$FF,$B1,$3E,$36,$D7
	.dc.b	$F6,$0A,$E0,$DA,$FD,$6F,$C6,$7F,$B4,$AF,$FC,$C7,$5A,$CF,$FD,$1D
	.dc.b	$3F,$E3,$C9,$FF,$67,$FB,$9E,$2F,$97,$E2,$D8,$E3,$1D,$3B,$71,$97
	.dc.b	$EA,$2F,$F4,$B5,$FE,$39,$C5,$7B,$8E,$27,$FB,$73,$EB,$BC,$BF,$53
	.dc.b	$E6,$6C,$79,$BA,$7E,$19,$EA,$F7,$2C,$7F,$B9,$3C,$7C,$34,$FF,$73
	.dc.b	$C7,$F1,$9F,$AE,$FE,$A1,$1F,$EB,$BF,$AE,$7B,$B9,$EB,$FE,$49,$E6
	.dc.b	$A7,$35,$FD,$FB,$8C,$FF,$7D,$C3,$9F,$0F,$FB,$35,$B3,$A8,$5F,$3F
	.dc.b	$73,$B5,$F3,$39,$63,$8A,$FF,$63,$E2,$AC,$61,$F5,$3C,$AF,$B2,$7D
	.dc.b	$AE,$9B,$8F,$3C,$3F,$C8,$7C,$8A,$CC,$7D,$3A,$B7,$E5,$CD,$F5,$CE
	.dc.b	$CF,$82,$DC,$BF,$DC,$0F,$E9,$30,$FF,$6D,$B9,$6C,$F0,$38,$F7,$DA
	.dc.b	$F2,$4B,$EA,$FF,$96,$3F,$E2,$53,$1D,$3F,$1D,$C4,$69,$FF,$15,$35
	.dc.b	$FF,$17,$BF,$8C,$6C,$7F,$4A,$FA,$DF,$AF,$1F,$EF,$F9,$D3,$DA,$BC
	.dc.b	$71,$C7,$C3,$57,$0D,$3C,$53,$9F,$71,$D5,$FD,$59,$B8,$7C,$FF,$0D
	.dc.b	$C5,$FD,$13,$3F,$F2,$7D,$69,$B7,$C4,$FD,$8F,$99,$70,$E5,$7F,$E3
	.dc.b	$96,$FB,$1F,$B6,$FA,$2F,$DB,$B2,$32,$70,$8C,$61,$EF,$F7,$FF,$33
	.dc.b	$F5,$6F,$AD,$FE,$F1,$B2,$73,$18,$66,$1F,$89,$E6,$5F,$F0,$38,$FA
	.dc.b	$33,$AC,$3B,$46,$E7,$51,$F5,$27,$F8,$6E,$D2,$BD,$BF,$53,$6B,$F8
	.dc.b	$24,$DE,$1F,$E1,$F3,$6B,$FC,$D1,$F5,$AF,$88,$FF,$08,$36,$3C,$1E
	.dc.b	$D6,$26,$47,$E3,$A7,$58,$B0,$FC,$F6,$7F,$E3,$17,$F1,$D3,$A7,$F2
	.dc.b	$1E,$0F,$F8,$5B,$5F,$83,$F1,$7F,$E1,$F8,$C5,$FF,$71,$AD,$FA,$19
	.dc.b	$3E,$7F,$AB,$73,$FC,$FD,$FF,$88,$33,$FE,$1B,$5A,$9C,$BF,$E2,$BF
	.dc.b	$AC,$7C,$5D,$3F,$88,$FD,$6F,$E6,$23,$35,$FA,$A6,$DF,$B6,$3F,$D1
	.dc.b	$D1,$8C,$BF,$79,$5C,$6F,$FC,$71,$E8,$FF,$87,$EF,$F6,$6D,$C3,$91
	.dc.b	$FC,$5B,$EB,$5F,$0D,$C3,$DC,$97,$C1,$FE,$4B,$3A,$8C,$1F,$C4,$7C
	.dc.b	$17,$3B,$9F,$F5,$07,$F5,$0F,$A5,$B5,$B9,$D3,$67,$4D,$3F,$4B,$F1
	.dc.b	$78,$CF,$FA,$63,$77,$FE,$AE,$D9,$FE,$F7,$DE,$8C,$57,$F9,$47,$0F
	.dc.b	$D1,$57,$87,$AF,$FC,$8B,$DD,$FB,$D6,$46,$4F,$DF,$7B,$C9,$CD,$FE
	.dc.b	$FB,$E2,$BF,$82,$1F,$E4,$EF,$06,$DF,$B6,$2C,$FF,$92,$AF,$F7,$CF
	.dc.b	$19,$D1,$1D,$57,$0D,$3F,$A5,$6F,$F3,$30,$C5,$FF,$86,$30,$E7,$5A
	.dc.b	$CD,$5F,$E7,$DF,$82,$7D,$33,$E5,$70,$DC,$7D,$BC,$38,$99,$1F,$D2
	.dc.b	$EB,$8C,$7F,$9A,$A7,$3A,$FF,$BB,$3E,$B7,$F5,$57,$FE,$2A,$FA,$AF
	.dc.b	$1B,$FE,$E4,$D9,$3A,$A7,$92,$BD,$C7,$FA,$16,$75,$5C,$3C,$4F,$F4
	.dc.b	$1D,$7E,$1A,$BC,$11,$FE,$88,$AD,$9D,$6F,$A6,$39,$9F,$F4,$6F,$D6
	.dc.b	$7F,$86,$EF,$FC,$F7,$F5,$9F,$6F,$3F,$CC,$4E,$78,$7F,$BC,$62,$AD
	.dc.b	$5F,$E9,$B4,$5E,$F0,$F7,$36,$7D,$BA,$7E,$39,$E1,$19,$AE,$3E,$EF
	.dc.b	$90,$F6,$F3,$5C,$3C,$71,$6F,$FD,$23,$87,$3C,$FF,$07,$8F,$79,$FF
	.dc.b	$7F,$8F,$AD,$FD,$02,$27,$8F,$FE,$13,$8C,$C7,$06,$BF,$E1,$6C,$78
	.dc.b	$46,$23,$FD,$F8,$75,$E9,$FE,$F1,$5C,$DF,$EF,$51,$A9,$E7,$39,$8F
	.dc.b	$EF,$DF,$15,$D2,$B8,$DD,$9C,$9D,$4F,$41,$E3,$30,$79,$E3,$8B,$83
	.dc.b	$DD,$3C,$DF,$7C,$D7,$26,$EB,$C8,$F3,$3B,$B8,$77,$C9,$FE,$31,$7C
	.dc.b	$AD,$7E,$7B,$5D,$FA,$DE,$8F,$8D,$FF,$66,$69,$F4,$6B,$84,$62,$91
	.dc.b	$8F,$03,$FA,$B4,$62,$F9,$46,$27,$57,$ED,$B7,$CB,$7C,$6D,$7E,$DC
	.dc.b	$63,$1D,$61,$D2,$72,$DD,$7C,$D7,$C5,$B5,$C3,$08,$FD,$0C,$3A,$36
	.dc.b	$B8,$3F,$8D,$8F,$66,$1D,$AB,$CD,$5A,$A5,$72,$AD,$47,$5C,$3C,$8F
	.dc.b	$3A,$E1,$5B,$C3,$E1,$E1,$F8,$7E,$2F,$D8,$F7,$7E,$1A,$3F,$46,$DD
	.dc.b	$BF,$66,$DC,$7F,$67,$D7,$6F,$8C,$7F,$00,$9C,$70,$FF,$3D,$8F,$F1
	.dc.b	$E7,$17,$F2,$A7,$44,$3F,$77,$3D,$79,$BF,$8D,$61,$CA,$31,$CD,$C6
	.dc.b	$BF,$87,$CE,$5F,$EC,$DF,$5B,$FE,$A2,$6F,$C0,$F9,$63,$11,$FC,$00
	.dc.b	$C4,$47,$67,$98,$C1,$FD,$EB,$07,$D6,$BE,$37,$CC,$D9,$E2,$7F,$80
	.dc.b	$39,$7F,$A2,$FC,$4F,$C7,$6C,$7F,$41,$F0,$3F,$DF,$8F,$F6,$2F,$07
	.dc.b	$51,$88,$E4,$DE,$3F,$F5,$05,$72,$9F,$EC,$15,$F4,$7E,$B7,$C5,$1F
	.dc.b	$CA,$DB,$5F,$D7,$9B,$C9,$FD,$77,$C6,$E1,$A4,$BF,$F0,$75,$69,$FE
	.dc.b	$97,$E2,$2B,$51,$D6,$AE,$1F,$CF,$2B,$53,$CA,$73,$87,$F5,$7C,$3A
	.dc.b	$6C,$BD,$DF,$9F,$E3,$7D,$44,$7F,$C0,$BD,$15,$D1,$E6,$8F,$C3,$4F
	.dc.b	$F3,$0C,$BF,$DD,$BE,$0F,$E2,$7C,$7F,$C3,$47,$FB,$C3,$97,$13,$30
	.dc.b	$FF,$42,$C6,$CC,$3F,$27,$C6,$F9,$84,$76,$7C,$67,$0E,$CC,$9F,$E4
	.dc.b	$01,$F5,$9F,$EB,$3B,$FF,$59,$37,$6F,$E6,$6D,$9F,$E9,$9F,$5B,$FC
	.dc.b	$69,$FF,$5E,$C6,$5E,$67,$EC,$4A,$7F,$B0,$A3,$83,$E1,$C1,$F2,$BB
	.dc.b	$6C,$7F,$B2,$1A,$F9,$FE,$B5,$F3,$1B,$A9,$3B,$7C,$13,$A8,$FF,$68
	.dc.b	$FD,$6F,$FB,$75,$BF,$05,$E5,$8C,$47,$C3,$FA,$DF,$E0,$4F,$C5,$8C
	.dc.b	$EB,$FE,$D5,$6E,$BF,$D5,$9B,$3F,$DB,$EB,$8D,$6E,$FE,$14,$7F,$BC
	.dc.b	$E7,$FD,$C6,$D7,$B4,$7C,$7F,$AD,$F4,$47,$D5,$6D,$7F,$B9,$5B,$87
	.dc.b	$F6,$6F,$67,$FC,$EF,$5F,$EE,$E9,$CB,$FE,$01,$6A,$6C,$70,$8F,$E1
	.dc.b	$1C,$5F,$3A,$FF,$C4,$2D,$BF,$F3,$5E,$9F,$97,$D7,$F3,$0C,$BF,$F6
	.dc.b	$CB,$97,$FD,$FF,$F0,$8F,$92,$FC,$13,$63,$31,$F8,$5F,$AD,$7C,$36
	.dc.b	$A2,$4E,$8F,$F0,$D6,$6C,$BF,$C0,$B7,$C2,$74,$7E,$6D,$7C,$9F,$53
	.dc.b	$9C,$7B,$8F,$D8,$10,$FF,$78,$FA,$15,$D1,$FE,$F4,$3F,$6C,$FF,$19
	.dc.b	$F3,$A4,$7F,$12,$7F,$A1,$7C,$FF,$F1,$3F,$35,$E0,$27,$75,$FC,$7D
	.dc.b	$DF,$83,$FA,$5F,$A9,$7D,$65,$25,$7C,$0D,$7F,$C4,$7F,$55,$FA,$8D
	.dc.b	$FE,$14,$FF,$8F,$6F,$8B,$B3,$17,$C9,$FD,$F5,$71,$BF,$F1,$6D,$70
	.dc.b	$9F,$F1,$CB,$5F,$F1,$88,$70,$F0,$AF,$F1,$FA,$D5,$F5,$6F,$AC,$FE
	.dc.b	$E1,$18,$6C,$66,$7C,$D3,$C3,$3C,$D4,$9C,$D6,$DF,$F6,$AD,$6D,$F7
	.dc.b	$78,$CE,$07,$8B,$F4,$89,$3A,$8E,$5B,$7F,$9F,$1D,$C6,$67,$6F,$F6
	.dc.b	$98,$DE,$BF,$1B,$9B,$EB,$12,$7E,$70,$DD,$F4,$79,$8F,$E5,$9C,$0A
	.dc.b	$C6,$5F,$62,$78,$56,$30,$FB,$BE,$27,$F2,$7C,$7F,$D6,$D5,$DB,$1E
	.dc.b	$4F,$AF,$EA,$FF,$1F,$F5,$92,$30,$7F,$64,$F3,$3F,$B7,$11,$F8,$51
	.dc.b	$F9,$6C,$3F,$2F,$F5,$92,$7E,$34,$7E,$6B,$CF,$FF,$13,$7D,$53,$EB
	.dc.b	$29,$9A,$FF,$7F,$46,$5B,$1A,$7E,$84,$FE,$93,$69,$1F,$4D,$D3,$FE
	.dc.b	$AB,$A7,$C6,$78,$09,$1F,$67,$EB,$2B,$B8,$BC,$3F,$6E,$6F,$0F,$FA
	.dc.b	$7F,$97,$F0,$9F,$59,$FB,$43,$38,$3B,$99,$7C,$C8,$F1,$72,$FC,$03
	.dc.b	$F7,$7F,$45,$63,$DF,$15,$FC,$34,$5C,$FF,$19,$DD,$FC,$08,$C6,$7F
	.dc.b	$EE,$1F,$AC,$7C,$1F,$F9,$45,$B8,$7E,$DF,$34,$62,$3F,$25,$F0,$1B
	.dc.b	$A9,$1B,$9D,$63,$F9,$BA,$CC,$FE,$67,$E2,$FF,$89,$98,$9F,$E0,$F8
	.dc.b	$76,$F9,$EF,$11,$2B,$F4,$DF,$0D,$5F,$F6,$F6,$29,$C5,$FE,$A6,$AE
	.dc.b	$7F,$3F,$BB,$FA,$54,$FE,$B7,$E2,$F3,$8F,$FB,$47,$EB,$3F,$83,$3F
	.dc.b	$A8,$6B,$FE,$11,$6C,$FE,$D7,$CA,$FF,$04,$46,$30,$FF,$61,$B6,$BA
	.dc.b	$78,$7F,$CC,$2B,$6F,$F3,$2F,$91,$93,$DE,$3F,$C2,$5E,$D3,$92,$92
	.dc.b	$7F,$83,$7C,$45,$7F,$C4,$93,$B7,$F7,$ED,$7F,$83,$47,$F8,$F3,$DB
	.dc.b	$FF,$47,$67,$FC,$0B,$DC,$BD,$DF,$E3,$F8,$FC,$A6,$0A,$E1,$87,$F3
	.dc.b	$79,$E0,$DA,$E0,$FF,$97,$27,$F1,$1E,$F4,$8F,$C1,$3F,$E4,$CF,$6F
	.dc.b	$11,$F6,$22,$F0,$7F,$97,$E7,$FC,$65,$F5,$9F,$5B,$1F,$F7,$6B,$6C
	.dc.b	$1C,$77,$FF,$98,$7D,$5E,$3C,$92,$BF,$90,$D6,$B5,$FF,$06,$BC,$E3
	.dc.b	$94,$ED,$FF,$4A,$4F,$F8,$53,$C8,$7D,$63,$E3,$BF,$DF,$FF,$19,$FE
	.dc.b	$05,$9C,$C6,$39,$FF,$DD,$55,$BC,$3F,$DF,$F1,$F3,$3D,$8D,$47,$FB
	.dc.b	$B2,$74,$B0,$E9,$18,$6D,$F0,$D7,$FC,$FD,$7C,$AD,$F5,$6F,$A1,$FF
	.dc.b	$2B,$C7,$C6,$32,$E1,$F9,$31,$BC,$FF,$B2,$B6,$3F,$B5,$9F,$EE,$2E
	.dc.b	$87,$D6,$7F,$6A,$F8,$EF,$CA,$36,$33,$3C,$27,$2F,$2A,$FE,$99,$D7
	.dc.b	$DB,$C7,$FD,$5C,$F9,$67,$2E,$E3,$2F,$FA,$98,$FF,$04,$ED,$FE,$33
	.dc.b	$D7,$FD,$3D,$78,$E0,$FC,$BD,$53,$FD,$5B,$16,$7F,$D4,$91,$78,$7F
	.dc.b	$C0,$5C,$BF,$52,$3B,$36,$7A,$B7,$F2,$3F,$F0,$33,$FD,$E9,$B3,$FE
	.dc.b	$D9,$7F,$DB,$11,$9F,$5F,$FC,$3F,$D3,$FE,$20,$B5,$70,$8C,$36,$BF
	.dc.b	$A1,$3C,$AF,$A5,$F2,$9C,$B6,$17,$D2,$30,$F6,$F6,$2D,$F5,$AD,$AF
	.dc.b	$53,$26,$BD,$7B,$8F,$B6,$2D,$7B,$66,$E5,$E8,$F5,$3A,$9D,$BD,$5B
	.dc.b	$87,$C5,$8B,$7F,$16,$6D,$7C,$3F,$C9,$5F,$5B,$E8,$6E,$1F,$E2,$65
	.dc.b	$F5,$79,$99,$39,$78,$46,$30,$5F,$5F,$5F,$A1,$D7,$E5,$AE,$1D,$62
	.dc.b	$D7,$08,$FE,$E5,$8A,$B3,$7C,$E7,$81,$E5,$8C,$BC,$3E,$C7,$D4,$7A
	.dc.b	$FE,$7A,$C7,$AE,$2E,$1D,$CF,$B3,$82,$B3,$87,$C5,$AB,$7E,$49,$B7
	.dc.b	$C3,$EC,$3E,$41,$F2,$7E,$46,$D7,$C9,$8B,$3D,$F9,$19,$2F,$39,$7D
	.dc.b	$3B,$B8,$7C,$29,$B8,$79,$BD,$BF,$AC,$7D,$7F,$81,$B5,$F5,$E2,$CF
	.dc.b	$7C,$7F,$D0,$B3,$92,$46,$67,$94,$63,$1F,$C2,$7D,$7F,$E0,$CF,$C2
	.dc.b	$7C,$45,$9F,$C2,$45,$BE,$D3,$6B,$8D,$68,$FC,$54,$65,$FB,$BF,$3D
	.dc.b	$F8,$FF,$AC,$FC,$83,$63,$F2,$0E,$1F,$C9,$78,$17,$0F,$63,$F9,$73
	.dc.b	$F3,$1E,$35,$9F,$CC,$7C,$EF,$E6,$FE,$37,$F3,$8D,$8F,$F4,$0F,$27
	.dc.b	$FA,$F9,$BF,$13,$FA,$17,$F7,$31,$9C,$7E,$C3,$FE,$CE,$8E,$3F,$17
	.dc.b	$AF,$80,$DB,$FC,$0F,$EF,$3F,$9A,$EF,$7C,$B3,$78,$F0,$CC,$3F,$D6
	.dc.b	$15,$C1,$B5,$FE,$5C,$6C,$EF,$2F,$CC,$56,$DB,$5B,$6C,$FA,$BE,$BB
	.dc.b	$D7,$F4,$F7,$0F,$82,$DA,$F6,$B7,$93,$F9,$36,$A8,$FE,$3D,$2F,$A6
	.dc.b	$F8,$4F,$EA,$69,$F6,$5F,$C1,$B0,$FF,$4E,$D6,$A7,$F9,$8C,$E6,$BF
	.dc.b	$60,$73,$DB,$FD,$38,$FF,$65,$AE,$AF,$FB,$29,$AE,$63,$6F,$F3,$0C
	.dc.b	$5F,$5F,$F6,$3E,$0B,$EC,$B7,$2F,$BE,$DA,$FB,$6D,$CB,$DE,$D9,$FF
	.dc.b	$41,$5F,$47,$F0,$F7,$F0,$B1,$7D,$4F,$F4,$1F,$35,$F5,$CF,$53,$53
	.dc.b	$98,$C6,$1F,$12,$32,$FF,$B6,$BE,$A1,$F6,$78,$DB,$F1,$FF,$5F,$F9
	.dc.b	$2F,$AF,$FC,$9B,$6F,$F3,$2D,$9D,$C7,$E7,$27,$F2,$51,$FE,$71,$8F
	.dc.b	$C6,$B4,$FC,$48,$E4,$DE,$D2,$76,$F2,$8C,$4F,$08,$BC,$2A,$FF,$20
	.dc.b	$F3,$FF,$90,$AF,$31,$FE,$1F,$FA,$CF,$F6,$EB,$FE,$97,$FC,$05,$EF
	.dc.b	$7E,$32,$73,$97,$F3,$3F,$8C,$56,$AF,$FD,$CF,$5A,$AF,$F1,$50,$70
	.dc.b	$F7,$F8,$5F,$C2,$B9,$6F,$C4,$3E,$BB,$EC,$71,$3A,$3F,$3B,$39,$C7
	.dc.b	$EA,$46,$2A,$47,$D7,$F2,$EE,$3F,$D2,$61,$FA,$F3,$E5,$3F,$C0,$91
	.dc.b	$86,$C7,$BB,$8B,$E4,$57,$C9,$8E,$15,$FE,$7B,$6C,$75,$6C,$F0,$9E
	.dc.b	$FA,$74,$B9,$1C,$5E,$F1,$8B,$FD,$7F,$E4,$DE,$FA,$FF,$3B,$DF,$8B
	.dc.b	$AD,$F5,$FF,$85,$8F,$2D,$AF,$83,$FD,$9E,$B7,$CD,$FC,$12,$73,$5C
	.dc.b	$1B,$87,$FA,$12,$BA,$F7,$6F,$7F,$E1,$7F,$AF,$BC,$8F,$D5,$7E,$90
	.dc.b	$F8,$6C,$C6,$1B,$8F,$A0,$FF,$35,$57,$17,$1A,$B2,$7E,$C1,$FC,$3F
	.dc.b	$5F,$5D,$EE,$BF,$9D,$FD,$6F,$F3,$E3,$F9,$FF,$83,$71,$FE,$8B,$16
	.dc.b	$7F,$A0,$CD,$BF,$E4,$BF,$63,$FD,$18,$FE,$8F,$F1,$57,$0F,$E9,$B1
	.dc.b	$6F,$FA,$4C,$DA,$FF,$27,$FC,$E7,$F9,$10,$FF,$22,$FC,$ED,$C3,$FC
	.dc.b	$B9,$16,$FF,$CB,$93,$6B,$FC,$AF,$CF,$FE,$4C,$3F,$C9,$BC,$F7,$0F
	.dc.b	$F3,$24,$5B,$FF,$32,$4D,$AE,$35,$A3,$FC,$D5,$19,$7F,$CC,$35,$6B
	.dc.b	$FC,$AF,$37,$DA,$FF,$34,$FC,$67,$F9,$70,$FF,$2E,$FC,$ED,$C3,$FC
	.dc.b	$E9,$16,$FF,$CE,$93,$71,$FF,$37,$FC,$DF,$F9,$90,$FF,$32,$FC,$E5
	.dc.b	$C3,$FC,$F9,$16,$FF,$CF,$93,$6B,$8D,$68,$FF,$41,$46,$5F,$F3,$CF
	.dc.b	$CD,$7F,$9C,$0F,$F3,$8F,$CE,$58,$FF,$44,$45,$BF,$F4,$44,$DB,$FF
	.dc.b	$41,$CD,$AF,$F4,$1C,$DA,$FF,$3C,$CD,$CF,$FD,$17,$39,$9F,$F6,$98
	.dc.b	$FB,$17,$CC,$FF,$18,$E8,$FE,$03,$18,$9F,$E3,$7F,$23,$FE,$EA,$8C
	.dc.b	$47,$EA,$72,$C5,$FF,$B0,$AA,$CF,$FB,$2E,$F9,$CE,$63,$C0,$F7,$AD
	.dc.b	$C7,$92,$FA,$BF,$EB,$03,$67,$88,$A7,$84,$A7,$FB,$FA,$F9,$BC,$63
	.dc.b	$9B,$FE,$E7,$37,$7E,$47,$A4,$78,$23,$11,$E3,$BF,$33,$D8,$E5,$E8
	.dc.b	$BE,$6B,$D8,$D8,$F3,$35,$F9,$ED,$F9,$46,$FF,$E9,$29,$FE,$E9,$5F
	.dc.b	$EF,$19,$CE,$E9,$C9,$63,$2E,$03,$8B,$ED,$7D,$72,$FA,$F2,$E0,$FF
	.dc.b	$7F,$46,$1B,$97,$86,$B8,$37,$1F,$AB,$7C,$71,$B7,$C1,$F2,$5F,$06
	.dc.b	$D6,$98,$7D,$8A,$F4,$E1,$E4,$AE,$13,$EA,$BE,$8D,$3C,$31,$AA,$E3
	.dc.b	$87,$C7,$AD,$C7,$96,$BB,$57,$8F,$7F,$F7,$5F,$C9,$FC,$B9,$FF,$79
	.dc.b	$37,$2F,$F7,$6B,$77,$FF,$76,$F9,$D6,$BF,$DC,$73,$6F,$9C,$7F,$BF
	.dc.b	$63,$1F,$21,$FE,$08,$8C,$10,$EE,$7F,$42,$9F,$F7,$ED,$70,$A9,$3B
	.dc.b	$C3,$FD,$FF,$3D,$37,$93,$C7,$0F,$C6,$57,$F7,$6C,$10,$F2,$F7,$5F
	.dc.b	$EB,$B8,$DE,$5F,$8B,$AD,$E1,$E8,$BD,$E1,$CE,$F7,$82,$F6,$F8,$6B
	.dc.b	$E1,$7C,$DF,$F9,$12,$3D,$8D,$7F,$D9,$4D,$BE,$9C,$3C,$3D,$59,$5C
	.dc.b	$1F,$93,$1D,$3E,$32,$4F,$47,$BF,$9D,$88,$EB,$16,$FE,$16,$1F,$E2
	.dc.b	$99,$E0,$43,$67,$9A,$B8,$6B,$F0,$E5,$78,$AF,$81,$1C,$DF,$5F,$C5
	.dc.b	$FC,$02,$3F,$03,$E3,$6B,$A3,$E1,$C7,$C6,$8F,$C4,$79,$4F,$AB,$A4
	.dc.b	$AF,$F5,$FF,$CD,$D7,$FD,$77,$17,$4F,$F4,$FC,$F2,$F9,$C9,$5F,$EC
	.dc.b	$67,$AF,$4E,$23,$FD,$87,$16,$BF,$D4,$F3,$F9,$B9,$7E,$C2,$F8,$BC
	.dc.b	$36,$3F,$C2,$E6,$A3,$87,$E6,$8F,$93,$DC,$63,$E5,$65,$70,$BE,$4F
	.dc.b	$FA,$AB,$E7,$9E,$B7,$85,$B1,$D9,$AF,$FA,$C3,$EE,$BF,$D2,$CE,$67
	.dc.b	$D1,$1F,$9D,$8F,$2B,$5E,$D1,$ED,$F8,$6F,$34,$7A,$9A,$FF,$AF,$1B
	.dc.b	$AF,$F9,$29,$7C,$C7,$1D,$25,$71,$7A,$46,$23,$94,$5A,$E5,$3F,$37
	.dc.b	$EA,$5E,$85,$25,$FC,$CF,$BF,$FF,$3E,$3F,$3A,$76,$E7,$E6,$7E,$B4
	.dc.b	$9E,$31,$FE,$D4,$F9,$3F,$3B,$FE,$8E,$3F,$A3,$38,$70,$D7,$FD,$26
	.dc.b	$D7,$F9,$13,$BC,$7F,$CD,$7F,$1A,$E1,$FC,$33,$6B,$F0,$75,$BE,$DF
	.dc.b	$E5,$5F,$52,$F6,$3F,$18,$DA,$FC,$4B,$5F,$F7,$5B,$FE,$D4,$73,$1C
	.dc.b	$08,$F0,$F6,$2B,$C2,$2E,$5F,$E8,$0A,$E5,$28,$FE,$45,$D8,$FA,$2F
	.dc.b	$D7,$F3,$4A,$FE,$61,$39,$6C,$E5,$E5,$97,$FB,$B2,$BF,$6F,$3F,$A7
	.dc.b	$FA,$67,$7D,$25,$FE,$9B,$EF,$DF,$27,$88,$E6,$7F,$79,$87,$CE,$7E
	.dc.b	$C5,$B1,$FA,$EB,$EA,$F9,$BD,$3F,$E0,$FF,$01,$23,$04,$3A,$F9,$9F
	.dc.b	$EC,$58,$D4,$63,$6F,$F4,$73,$BD,$FE,$2C,$E9,$F8,$F7,$ED,$F9,$CF
	.dc.b	$C5,$57,$F8,$3F,$14,$62,$11,$87,$6D,$7F,$24,$DA,$FF,$5A,$FD,$47
	.dc.b	$ED,$5B,$E7,$77,$AF,$EC,$5F,$94,$BE,$87,$F9,$2C,$43,$2F,$F6,$F8
	.dc.b	$D4,$7D,$A3,$47,$C2,$72,$FD,$01,$7F,$E2,$E7,$F8,$DB,$5D,$36,$FC
	.dc.b	$FE,$EF,$F2,$58,$C7,$2F,$F8,$1D,$B5,$FE,$33,$6B,$DA,$78,$7A,$D2
	.dc.b	$3B,$BC,$E3,$D5,$D9,$27,$E0,$BE,$DF,$A2,$F7,$78,$72,$7E,$2B,$F0
	.dc.b	$7C,$3F,$85,$F0,$3F,$0F,$E8,$A5,$7C,$A8,$F8,$DF,$03,$23,$E6,$BF
	.dc.b	$27,$E0,$FE,$57,$C1,$49,$FA,$2E,$DF,$94,$FC,$FF,$83,$91,$A3,$EB
	.dc.b	$ED,$F4,$C8,$FD,$AD,$AB,$F4,$E2,$F0,$FF,$85,$78,$95,$8E,$2F,$F0
	.dc.b	$94,$7C,$CE,$A3,$4D,$E1,$F8,$6F,$82,$FC,$11,$1F,$C4,$FC,$05,$7F
	.dc.b	$05,$17,$1F,$04,$DC,$3B,$CD,$AF,$F0,$E5,$FF,$91,$EF,$CD,$EF,$7E
	.dc.b	$3C,$93,$F9,$6F,$83,$AF,$E3,$E2,$CF,$B6,$6E,$1E,$C9,$B5,$DA,$7F
	.dc.b	$35,$D3,$23,$F4,$2F,$E7,$7E,$03,$11,$F9,$B8,$B8,$7C,$99,$B8,$7C
	.dc.b	$89,$B5,$EA,$9F,$D3,$7B,$B2,$3F,$5A,$FE,$A7,$E0,$31,$1F,$A7,$8B
	.dc.b	$87,$D5,$9B,$87,$D4,$9B,$5F,$1A,$7F,$CC,$1F,$94,$FF,$88,$EF,$FC
	.dc.b	$A0,$DB,$FF,$26,$36,$BF,$C9,$0F,$F9,$06,$33,$EF,$F2,$24,$FF,$98
	.dc.b	$E7,$4F,$F8,$C3,$E8,$FF,$C9,$A4,$7F,$C9,$BE,$85,$7F,$C9,$B1,$70
	.dc.b	$FF,$20,$CF,$F9,$53,$E3,$E5,$7F,$95,$1F,$F2,$BF,$8D,$88,$FF,$2A
	.dc.b	$C5,$AF,$F2,$4C,$FF,$97,$3C,$49,$5F,$E5,$C7,$FC,$BF,$E4,$62,$3F
	.dc.b	$CB,$B1,$6B,$FC,$A3,$37,$4F,$F2,$64,$DA,$FF,$25,$CF,$F9,$A3,$D1
	.dc.b	$FF,$49,$7C,$33,$87,$FC,$A5,$E5,$E1,$B7,$FE,$DB,$7F,$CA,$D5,$98
	.dc.b	$FF,$2A,$B7,$A5,$F0,$3F,$D4,$A7,$81,$FE,$5A,$BD,$42,$B9,$4B,$E9
	.dc.b	$5F,$43,$FE,$30,$8F,$F5,$31,$97,$0F,$FA,$A6,$37,$CD,$FD,$21,$B1
	.dc.b	$FE,$13,$F8,$7F,$DF,$FC,$F7,$F4,$C6,$C6,$7B,$D3,$96,$C7,$F8,$6E
	.dc.b	$9E,$BF,$F5,$A8,$F1,$C6,$3C,$FF,$04,$E6,$7C,$8D,$7F,$D3,$1F,$4B
	.dc.b	$FE,$E9,$E4,$F8,$7C,$08,$5A,$73,$1B,$A7,$17,$D5,$79,$D6,$DC,$F1
	.dc.b	$7F,$AA,$EB,$85,$6F,$0E,$D7,$B8,$C4,$EA,$B8,$4E,$5F,$F5,$A7,$C4
	.dc.b	$BC,$DE,$DC,$FF,$EB,$36,$E7,$AF,$42,$E9,$FE,$B5,$BE,$13,$FE,$C5
	.dc.b	$AC,$E9,$FE,$C4,$6C,$FF,$6F,$9D,$12,$BF,$D8,$D3,$AB,$DB,$C3,$CD
	.dc.b	$95,$FE,$D1,$F3,$ED,$FF,$B3,$66,$E5,$FE,$C9,$9B,$5E,$CF,$3B,$C2
	.dc.b	$78,$BC,$FB,$1F,$ED,$B8,$B8,$7F,$B5,$E6,$DF,$FB,$77,$E9,$7F,$C6
	.dc.b	$37,$FE,$E9,$6D,$7F,$88,$9F,$F5,$CC,$67,$D5,$F8,$31,$AA,$E5,$18
	.dc.b	$F8,$9C,$BA,$F0,$A5,$F5,$F8,$9B,$8F,$59,$B8,$78,$E6,$D7,$FB,$AF
	.dc.b	$1F,$F3,$B3,$E5,$AF,$0C,$EA,$FF,$DD,$54,$F0,$BC,$A7,$9B,$C4,$B8
	.dc.b	$7B,$62,$D7,$B6,$6E,$3F,$0E,$6D,$78,$66,$D7,$6F,$0B,$E0,$9F,$0B
	.dc.b	$C4,$B5,$F2,$62,$CF,$C9,$9B,$87,$CF,$9B,$5E,$D9,$B5,$EA,$F0,$FE
	.dc.b	$69,$F3,$BC,$4B,$5F,$56,$2C,$FD,$59,$B8,$7D,$B9,$B5,$F2,$66,$D7
	.dc.b	$C6,$FA,$5F,$B2,$7D,$AF,$12,$D7,$E0,$62,$CF,$E0,$66,$E1,$F8,$79
	.dc.b	$B5,$F5,$66,$D7,$D2,$FA,$5F,$C2,$FC,$37,$E1,$9B,$5F,$8D,$70,$FE
	.dc.b	$3F,$C0,$B5,$F8,$F7,$F1,$A7,$E3,$BC,$6B,$3F,$95,$F9,$1F,$C8,$FD
	.dc.b	$2F,$E4,$9B,$1F,$E9,$2F,$A5,$FD,$AF,$19,$FC,$BC,$61,$CF,$D1,$C8
	.dc.b	$FF,$4C,$FD,$2D,$CF,$FD,$1B,$F4,$1F,$CA,$B7,$FE,$06,$DA,$4E,$BC
	.dc.b	$AF,$F4,$AC,$63,$8B,$F8,$D5,$6E,$B5,$7C,$AF,$FA,$35,$EC,$E7,$5F
	.dc.b	$CE,$39,$35,$52,$75,$6A,$E9,$87,$6C,$3C,$33,$DA,$BF,$D6,$7F,$41
	.dc.b	$F0,$BE,$97,$E1,$B6,$32,$7F,$AE,$23,$2D,$3B,$CB,$E9,$5A,$FF,$35
	.dc.b	$8F,$EB,$31,$89,$FF,$4E,$B6,$3F,$DD,$0D,$77,$3F,$CB,$1E,$8F,$F2
	.dc.b	$EF,$A2,$7D,$DB,$D6,$D5,$F3,$D3,$FD,$A3,$F3,$BF,$73,$E4,$3E,$EB
	.dc.b	$73,$FF,$68,$B7,$3F,$F5,$B3,$6B,$8C,$FF,$B5,$AB,$8B,$4F,$7D,$BE
	.dc.b	$33,$C2,$7F,$42,$F3,$DF,$42,$F9,$9C,$DF,$0A,$B9,$FF,$B5,$2F,$75
	.dc.b	$8C,$BF,$DD,$3F,$4B,$F9,$4F,$A5,$FC,$AB,$71,$FF,$74,$B7,$0F,$F6
	.dc.b	$73,$6B,$F1,$70,$D3,$F2,$6D,$AF,$C2,$F2,$FE,$53,$93,$35,$27,$35
	.dc.b	$B3,$F3,$A2,$74,$E2,$6D,$F8,$26,$D7,$E8,$29,$F4,$8F,$88,$FC,$24
	.dc.b	$66,$31,$3F,$EE,$06,$C7,$F8,$80,$E3,$3C,$1F,$D5,$C6,$AF,$F1,$CD
	.dc.b	$AF,$D7,$FE,$10,$FC,$21,$E2,$FE,$2E,$B5,$87,$ED,$39,$5F,$01,$F8
	.dc.b	$F6,$E3,$FA,$AE,$2F,$B7,$E4,$66,$A4,$FE,$82,$B7,$7F,$6F,$95,$7A
	.dc.b	$86,$BF,$E1,$7B,$7D,$33,$E2,$7F,$C1,$31,$86,$C7,$9A,$30,$7E,$F0
	.dc.b	$B5,$D1,$FF,$16,$47,$37,$4F,$13,$11,$B3,$FC,$5C,$E2,$BC,$8D,$9E
	.dc.b	$7C,$1E,$2B,$FD,$97,$D2,$FF,$67,$9F,$DF,$B7,$0F,$CD,$7D,$33,$AF
	.dc.b	$7E,$2C,$BD,$BF,$A5,$9D,$46,$0F,$E1,$3F,$48,$F1,$FF,$80,$36,$3E
	.dc.b	$97,$D2,$FF,$83,$31,$F5,$09,$D3,$86,$ED,$FC,$45,$B5,$FC,$CB,$8D
	.dc.b	$7C,$EB,$58,$7F,$42,$F9,$89,$7F,$CA,$BD,$EF,$F0,$37,$81,$8C,$FE
	.dc.b	$54,$7F,$82,$6B,$18,$7F,$94,$E3,$77,$C7,$55,$7F,$4E,$FA,$59,$39
	.dc.b	$F9,$CB,$8F,$4B,$56,$31,$E9,$0A,$C5,$FF,$95,$BE,$4B,$FC,$96,$7F
	.dc.b	$93,$3E,$4E,$C7,$A7,$E0,$93,$8C,$BD,$30,$AC,$63,$E9,$8F,$C1,$4F
	.dc.b	$F9,$6F,$97,$FC,$AF,$CD,$FE,$58,$66,$3D,$5C,$B4,$93,$DA,$32,$D8
	.dc.b	$CD,$F7,$9C,$B6,$3D,$1F,$0F,$FE,$66,$3F,$CC,$FF,$27,$63,$24,$BF
	.dc.b	$5C,$65,$F6,$FC,$97,$F9,$B8,$FF,$37,$FC,$9D,$AF,$89,$16,$7E,$1C
	.dc.b	$DC,$3E,$14,$DA,$FF,$41,$CD,$AF,$F3,$F7,$C8,$7F,$9E,$8F,$F3,$DF
	.dc.b	$C9,$DA,$FA,$11,$67,$E7,$CD,$C3,$E7,$4D,$AF,$F4,$7C,$DA,$EB,$1C
	.dc.b	$8F,$B5,$19,$8F,$F6,$B8,$FA,$57,$D1,$7F,$B1,$BA,$BF,$D4,$91,$89
	.dc.b	$FF,$64,$FC,$6F,$F8,$4A,$31,$1F,$A3,$C3,$17,$FC,$BE,$AC,$FF,$39
	.dc.b	$BE,$73,$98,$F0,$3D,$EB,$71,$E4,$BE,$AF,$FA,$F2,$32,$FF,$88,$AE
	.dc.b	$DF,$48,$F2,$9D,$0F,$04,$F0,$7F,$D8,$26,$4F,$41,$4E,$85,$8E,$8F
	.dc.b	$8E,$FC,$CE,$CE,$7F,$20,$F9,$3F,$63,$63,$C6,$D7,$E7,$B7,$E4,$5B
	.dc.b	$7F,$3C,$94,$E6,$30,$DD,$BF,$C8,$2D,$9D,$DF,$F6,$5B,$E7,$5F,$29
	.dc.b	$6F,$FD,$A7,$5F,$AF,$18,$BD,$63,$E1,$9C,$B7,$2E,$97,$C5,$B5,$C3
	.dc.b	$08,$FC,$BC,$3A,$36,$B8,$3F,$E5,$88,$F2,$BB,$BE,$CE,$67,$6A,$79
	.dc.b	$E1,$E4,$BE,$EF,$AA,$B7,$3A,$C3,$E1,$61,$D3,$AD,$CB,$FE,$FF,$CB
	.dc.b	$F8,$BF,$C7,$FF,$82,$F3,$FE,$2C,$DD,$FF,$DE,$FF,$29,$6B,$FD,$E5
	.dc.b	$37,$0E,$6D,$C3,$A6,$1F,$C7,$EF,$13,$2F,$F9,$43,$B8,$C4,$78,$8C
	.dc.b	$9F,$CB,$5E,$9B,$48,$FE,$60,$F7,$F2,$31,$1F,$CC,$22,$DF,$F2,$F9
	.dc.b	$B9,$7F,$A1,$B2,$F8,$35,$8B,$47,$C4,$AF,$67,$C9,$FF,$93,$67,$FA
	.dc.b	$B3,$5F,$EA,$AD,$C3,$F9,$E7,$0F,$0F,$52,$57,$07,$E2,$47,$4F,$8B
	.dc.b	$93,$D1,$EF,$E6,$62,$3A,$C5,$BF,$EC,$18,$7F,$8A,$A7,$81,$0D,$9E
	.dc.b	$6A,$E1,$A7,$BE,$57,$8A,$F7,$C7,$37,$D7,$F1,$3F,$00,$8F,$C0,$F8
	.dc.b	$BA,$E8,$F8,$71,$F1,$A3,$F0,$5E,$13,$E9,$69,$2B,$FB,$B7,$D2,$D7
	.dc.b	$FB,$94,$5D,$3F,$B8,$CF,$2F,$A5,$95,$FD,$F5,$EB,$CF,$88,$FE,$F5
	.dc.b	$16,$BF,$BC,$CF,$E5,$65,$F4,$8F,$8A,$C3,$63,$FD,$56,$D7,$FC,$3F
	.dc.b	$F4,$AF,$91,$DD,$49,$DB,$D6,$75,$1F,$EB,$4F,$A0,$FF,$41,$37,$DC
	.dc.b	$F1,$46,$23,$FC,$57,$F4,$BF,$C6,$9D,$10,$FF,$17,$C6,$F3,$FF,$5E
	.dc.b	$4E,$E7,$FC,$67,$F4,$BE,$06,$E5,$EF,$8C,$57,$FB,$0D,$7C,$BE,$66
	.dc.b	$57,$FB,$2A,$77,$CD,$FE,$89,$77,$19,$9D,$BD,$8E,$15,$F3,$7E,$97
	.dc.b	$FC,$4D,$1F,$ED,$46,$BF,$E4,$86,$E1,$FE,$D5,$FA,$57,$27,$1F,$42
	.dc.b	$5F,$17,$EC,$4E,$A3,$FD,$BF,$F2,$3E,$47,$FC,$A6,$DB,$FF,$70,$37
	.dc.b	$4F,$F5,$17,$C6,$7B,$28,$27,$93,$FE,$A7,$31,$9F,$FA,$85,$DC,$69
	.dc.b	$FF,$56,$35,$E7,$F4,$BF,$E4,$E7,$FD,$E6,$D7,$FC,$CE,$DC,$3F,$DE
	.dc.b	$8F,$FB,$66,$37,$EA,$FE,$56,$35,$8F,$FA,$E5,$B3,$DB,$D6,$D5,$48
	.dc.b	$E2,$78,$23,$F5,$FF,$4B,$FB,$0E,$39,$3F,$E6,$38,$CB,$FE,$65,$8C
	.dc.b	$BF,$E6,$18,$FD,$37,$D1,$BB,$69,$27,$E4,$E6,$CF,$FD,$9D,$F4,$AF
	.dc.b	$91,$D5,$7F,$A1,$B0,$E4,$F0,$AF,$F6,$A4,$62,$3F,$D1,$75,$C5,$FF
	.dc.b	$3C,$7A,$1F,$C2,$3E,$96,$46,$08,$7F,$9D,$DC,$BF,$E7,$98,$FE,$1E
	.dc.b	$63,$1F,$F3,$CC,$E2,$7C,$47,$E0,$AD,$F3,$1F,$00,$91,$A1,$F4,$BF
	.dc.b	$AF,$BF,$81,$1C,$67,$F0,$CA,$B9,$BF,$89,$6B,$FE,$EB,$6B,$FE,$F5
	.dc.b	$F2,$DF,$4A,$FA,$2F,$E4,$F3,$0F,$F5,$14,$65,$B1,$97,$FD,$30,$7F
	.dc.b	$6F,$8D,$3F,$EB,$6C,$7F,$C4,$EF,$AD,$AE,$7D,$2F,$E4,$90,$F6,$AC
	.dc.b	$49,$E0,$EA,$31,$1C,$4C,$9D,$5F,$17,$D0,$C8,$F4,$BE,$5F,$A2,$F3
	.dc.b	$78,$12,$7D,$AF,$A7,$AF,$D5,$EF,$FA,$FE,$8A,$57,$C2,$8F,$77,$C0
	.dc.b	$48,$F8,$AF,$C1,$F8,$1F,$85,$F0,$12,$7E,$4B,$B7,$FD,$93,$BA,$3E
	.dc.b	$47,$C0,$C9,$D9,$F4,$37,$F9,$84,$7E,$9E,$F5,$E7,$17,$97,$FD,$A5
	.dc.b	$87,$CB,$9E,$44,$36,$72,$3E,$EF,$C0,$7D,$92,$3F,$80,$F7,$AB,$E8
	.dc.b	$8B,$5D,$A6,$F5,$FF,$86,$6D,$58,$BF,$F0,$CC,$7F,$93,$AF,$CB,$ED
	.dc.b	$FE,$1C,$93,$F8,$EF,$81,$AF,$C6,$8B,$3E,$C9,$B8,$77,$9B,$5D,$A7
	.dc.b	$F2,$5E,$DC,$8F,$CC,$BF,$95,$F7,$71,$1F,$4A,$2E,$1F,$22,$6E,$1E
	.dc.b	$B9,$B5,$EA,$9F,$CF,$7B,$72,$3F,$4A,$FE,$87,$DD,$C4,$7D,$E8,$B8
	.dc.b	$7D,$49,$B8,$7C,$79,$B5,$F1,$A7,$FC,$B1,$F4,$BF,$E2,$1B,$FD,$FF
	.dc.b	$6A,$31,$87,$F9,$1D,$B7,$FE,$46,$7F,$C8,$11,$9F,$77,$99,$27,$F8
	.dc.b	$8D,$73,$F7,$BF,$C9,$7F,$43,$27,$FC,$96,$FF,$93,$BE,$8F,$11,$FE
	.dc.b	$4C,$8B,$97,$F9,$02,$7F,$CA,$7F,$1B,$2B,$FC,$A6,$FF,$95,$BD,$3C
	.dc.b	$47,$F9,$52,$2D,$7F,$92,$27,$FC,$B7,$E2,$CA,$FF,$2D,$BF,$E5,$EF
	.dc.b	$A2,$C4,$7F,$97,$22,$D7,$F9,$42,$6E,$BF,$E4,$B9,$B5,$FE,$4A,$9F
	.dc.b	$F3,$3F,$A5,$FD,$C2,$1F,$0C,$E3,$FF,$29,$79,$99,$8C,$4F,$F9,$4A
	.dc.b	$31,$3F,$E5,$7A,$B3,$FE,$56,$6F,$03,$E0,$BF,$A5,$4F,$6A,$EA,$FF
	.dc.b	$B8,$5E,$4D,$3F,$C0,$7F,$4A,$FA,$2F,$F1,$64,$79,$E3,$1C,$FF,$E0
	.dc.b	$36,$C7,$F8,$1C,$38,$7F,$C1,$5F,$0F,$FC,$AF,$E7,$FF,$C1,$4D,$8E
	.dc.b	$0D,$77,$E0,$A7,$B7,$B0,$F4,$69,$2B,$C1,$87,$F8,$66,$3C,$5A,$FC
	.dc.b	$1F,$80,$FE,$AB,$C3,$F2,$A3,$1C,$9E,$F7,$E3,$7C,$A6,$B8,$3F,$D0
	.dc.b	$57,$72,$FF,$40,$B6,$B8,$C7,$CB,$3F,$D0,$B5,$9E,$1F,$F3,$56,$1D
	.dc.b	$5F,$94,$FF,$A9,$3E,$8F,$FB,$3F,$3F,$FA,$5B,$8E,$CF,$FA,$46,$6E
	.dc.b	$1F,$E9,$09,$B8,$FD,$B9,$B8,$74,$AF,$F5,$6C,$62,$B2,$DB,$F0,$1F
	.dc.b	$EB,$4F,$A3,$E3,$C1,$C9,$B1,$93,$FD,$73,$19,$69,$FB,$6E,$F7,$67
	.dc.b	$8B,$E3,$BC,$B1,$89,$FF,$60,$C6,$27,$51,$C5,$B9,$35,$FE,$17,$7F
	.dc.b	$E6,$0A,$ED,$DD,$D3,$B7,$AB,$63,$27,$8E,$32,$F9,$7B,$3C,$7D,$DE
	.dc.b	$46,$C6,$4F,$54,$65,$A6,$71,$F2,$3F,$C6,$E3,$51,$83,$F8,$E3,$F0
	.dc.b	$3B,$7E,$0F,$75,$AF,$8C,$DA,$E1,$C7,$FC,$9E,$31,$A7,$C0,$6D,$7F
	.dc.b	$29,$6D,$7F,$29,$7E,$6F,$27,$CE,$EE,$B5,$F4,$DB,$5D,$B6,$46,$30
	.dc.b	$F9,$AD,$AF,$E6,$CD,$AF,$E6,$CF,$DA,$E3,$FB,$7C,$D6,$BF,$00,$DA
	.dc.b	$EC,$DB,$FB,$2D,$9F,$E8,$4D,$9F,$E8,$4F,$E1,$B8,$FF,$0F,$CB,$6B
	.dc.b	$F1,$8D,$AF,$53,$6F,$F0,$AD,$9F,$E9,$CD,$9F,$E9,$CF,$E4,$B9,$3F
	.dc.b	$27,$CB,$6B,$F3,$0F,$E5,$9C,$57,$C7,$6D,$FF,$5C,$FD,$89,$F4,$9B
	.dc.b	$9F,$EA,$B1,$87,$F4,$47,$ED,$17,$D1,$5F,$A4,$C3,$67,$8F,$CE,$EF
	.dc.b	$E9,$38,$36,$3A,$FC,$C7,$3F,$8A,$E8,$D8,$FF,$3A,$FC,$FF,$F2,$3E
	.dc.b	$2F,$E0,$AF,$F2,$5E,$6B,$E2,$7F,$10,$E0,$FC,$B4,$E6,$F8,$CE,$78
	.dc.b	$3F,$CE,$F3,$9A,$DC,$E6,$BF,$D8,$73,$9A,$DC,$E6,$93,$9A,$DC,$E6
	.dc.b	$31,$3F,$EC,$CF,$A2,$F8,$B5,$FE,$8E,$B7,$C5,$BE,$53,$F8,$FD,$AF
	.dc.b	$6F,$EB,$EB,$73,$A9,$CC,$FF,$B5,$5A,$FF,$6B,$6B,$E1,$E0,$FF,$4A
	.dc.b	$BF,$E8,$DE,$27,$D0,$FD,$9F,$A2,$FB,$4D,$CF,$EF,$B6,$BF,$73,$5C
	.dc.b	$27,$F0,$75,$F8,$1F,$8A,$91,$F8,$6F,$8A,$B7,$C2,$6E,$9F,$8D,$9B
	.dc.b	$5F,$EF,$0F,$01,$8A,$F3,$82,$73,$78,$FA,$47,$CB,$39,$BF,$25,$F1
	.dc.b	$FF,$D2,$27,$31,$8C,$FF,$D7,$B1,$86,$D7,$17,$F3,$BF,$35,$C6,$31
	.dc.b	$97,$C0,$7F,$3C,$7E,$7F,$E5,$ED,$7E,$92,$2C,$FC,$69,$BB,$7E,$3A
	.dc.b	$1B,$FE,$61,$74,$D8,$C3,$70,$FC,$FC,$E9,$B3,$A6,$9A,$9E,$13,$FA
	.dc.b	$03,$81,$F6,$3E,$5F,$F6,$DF,$2F,$FB,$76,$D7,$EF,$1B,$3B,$8F,$DF
	.dc.b	$CF,$E5,$3D,$97,$95,$FD,$D3,$D2,$7D,$03,$CD,$D4,$E6,$31,$87,$FB
	.dc.b	$9E,$F8,$B6,$75,$E6,$70,$8F,$F7,$88,$FA,$07,$B9,$FE,$02,$8C,$7B
	.dc.b	$F6,$73,$1C,$1F,$F4,$EB,$67,$53,$B6,$C6,$4E,$11,$96,$9F,$EF,$DF
	.dc.b	$8F,$FE,$3D,$B7,$F5,$FD,$5F,$3E,$EA,$FD,$BD,$6B,$93,$FE,$00,$D6
	.dc.b	$A7,$3E,$54,$C5,$8F,$3C,$7F,$3D,$1D,$E7,$F2,$1C,$7F,$CA,$FE,$52
	.dc.b	$CF,$F3,$06,$DF,$F7,$1E,$AF,$D0,$D7,$F3,$1F,$94,$CE,$9F,$CE,$74
	.dc.b	$FE,$69,$AF,$F3,$56,$E5,$FC,$E9,$B8,$FE,$F1,$B5,$FE,$9C,$F8,$87
	.dc.b	$C8,$FF,$3D,$AF,$F1,$8C,$62,$FF,$A1,$C6,$27,$51,$FE,$9A,$9D,$4F
	.dc.b	$F0,$83,$FC,$73,$F2,$9F,$1E,$7F,$80,$7B,$67,$3B,$FF,$8F,$1A,$7F
	.dc.b	$A9,$23,$4F,$FA,$9A,$79,$57,$F0,$59,$CB,$FE,$0E,$8F,$E5,$38,$FF
	.dc.b	$AD,$97,$A1,$E2,$F1,$FC,$6D,$C3,$A3,$6F,$93,$6F,$93,$67,$F8,$C5
	.dc.b	$F4,$7C,$97,$D1,$A7,$43,$D1,$E7,$FC,$2F,$43,$E1,$B6,$32,$7B,$63
	.dc.b	$2D,$3E,$07,$7F,$CA,$F1,$3E,$5B,$63,$27,$C5,$8C,$B4,$CE,$7F,$16
	.dc.b	$32,$D8,$F9,$9E,$4F,$D6,$F0,$BE,$BB,$63,$27,$D1,$8C,$B4,$CC,$FD
	.dc.b	$18,$CB,$63,$3A,$FD,$39,$CB,$63,$F4,$AF,$F5,$69,$FD,$3F,$BB,$FA
	.dc.b	$8F,$9D,$78,$9F,$EC,$79,$FF,$51,$5E,$63,$FD,$61,$56,$7F,$D9,$9C
	.dc.b	$BF,$8D,$57,$C9,$E5,$3B,$7F,$6B,$3B,$9D,$47,$FB,$44,$E3,$5E,$19
	.dc.b	$D3,$E0,$8C,$9E,$0B,$E8,$FA,$8C,$9D,$9C,$BF,$EE,$03,$27,$B0,$A6
	.dc.b	$6B,$C6,$E5,$AE,$67,$C9,$3D,$2F,$C3,$7E,$6A,$F7,$8F,$82,$7C,$2F
	.dc.b	$A5,$B1,$A6,$BF,$05,$BF,$06,$F8,$9F,$F4,$C4,$F3,$AF,$F7,$D4,$E7
	.dc.b	$4F,$E3,$F7,$C2,$F6,$D8,$CD,$70,$9E,$7A,$7D,$0B,$7C,$4F,$F2,$8C
	.dc.b	$7E,$4C,$61,$FB,$0F,$77,$35,$CB,$5F,$1D,$71,$BE,$8D,$8F,$06,$3C
	.dc.b	$AB,$85,$FF,$36,$C3,$B3,$E0,$BE,$0E,$67,$6A,$7C,$58,$76,$8F,$2D
	.dc.b	$70,$75,$5C,$2B,$78,$79,$A7,$67,$8F,$0F,$CD,$79,$CE,$6F,$F1,$24
	.dc.b	$FF,$21,$F0,$BE,$25,$7F,$1D,$6F,$07,$F1,$4F,$3E,$D7,$F8,$A6,$F8
	.dc.b	$BF,$C9,$6B,$38,$75,$6C,$EB,$0F,$E5,$17,$89,$97,$FC,$C6,$36,$61
	.dc.b	$E8,$64,$FE,$6D,$E7,$74,$F3,$E4,$60,$87,$F3,$77,$2F,$F3,$98,$F1
	.dc.b	$98,$BF,$E7,$53,$89,$FE,$85,$E7,$FA,$3C,$F9,$18,$21,$FD,$0D,$CB
	.dc.b	$FD,$16,$3D,$FE,$27,$C0,$DA,$4F,$F4,$D8,$CB,$FD,$3E,$32,$FF,$4D
	.dc.b	$8C,$BF,$D3,$E3,$38,$7F,$4F,$9C,$BF,$D4,$63,$3B,$7F,$48,$9F,$9F
	.dc.b	$E2,$7D,$0F,$1A,$61,$FD,$76,$32,$FF,$5F,$8C,$BF,$D7,$63,$EC,$F8
	.dc.b	$6F,$92,$BF,$25,$2B,$FB,$94,$68,$C3,$FE,$F2,$94,$E6,$3F,$BC,$7C
	.dc.b	$97,$4F,$92,$91,$82,$1C,$5C,$BC,$E3,$F1,$FE,$53,$F6,$37,$C4,$61
	.dc.b	$B5,$FE,$72,$6B,$E3,$F9,$3D,$4F,$B3,$E4,$E4,$EE,$3F,$36,$BF,$63
	.dc.b	$FF,$7E,$BF,$D2,$38,$7F,$83,$46,$21,$C7,$FE,$A0,$8C,$3F,$EA,$2E
	.dc.b	$5F,$F6,$BC,$F1,$3F,$C6,$B3,$AA,$FE,$21,$3F,$EE,$29,$D0,$70,$F6
	.dc.b	$8F,$0F,$C3,$79,$BD,$DF,$E5,$D3,$C1,$FE,$57,$87,$FA,$C9,$B5,$FC
	.dc.b	$75,$DC,$F3,$3F,$C8,$EE,$27,$87,$85,$D4,$ED,$E1,$CA,$ED,$19,$7C
	.dc.b	$11,$F2,$3E,$69,$E6,$52,$4F,$B3,$E6,$9F,$0B,$88,$F1,$08,$7C,$3F
	.dc.b	$D2,$6C,$7C,$4B,$EA,$FF,$90,$7E,$5B,$F2,$BD,$32,$30,$43,$F2,$4E
	.dc.b	$5F,$CA,$47,$DE,$F9,$A7,$65,$ED,$95,$F4,$A7,$2D,$9C,$BE,$88,$4F
	.dc.b	$39,$93,$C6,$37,$BF,$FB,$92,$B9,$CF,$E9,$7E,$5F,$F5,$A4,$7F,$CB
	.dc.b	$9F,$2F,$5D,$1F,$E5,$F8,$FD,$9C,$7E,$8B,$F6,$4F,$E7,$13,$FE,$D0
	.dc.b	$6B,$FE,$79,$6D,$FF,$9B,$FE,$53,$87,$93,$FB,$CF,$32,$5F,$18,$FD
	.dc.b	$FF,$BB,$23,$FC,$DC,$FF,$02,$F3,$B1,$1F,$BD,$8B,$87,$F9,$EE,$6F
	.dc.b	$4F,$FB,$BF,$85,$58,$BF,$F7,$74,$7F,$60,$BF,$F4,$17,$CB,$FF,$0E
	.dc.b	$F7,$E5,$7F,$9F,$63,$F8,$89,$87,$F8,$64,$61,$FD,$3F,$CD,$3E,$0B
	.dc.b	$FC,$71,$1F,$DE,$E3,$0E,$03,$87,$6D,$7F,$C8,$0E,$6F,$FC,$79,$0F
	.dc.b	$9A,$7C,$2F,$F0,$66,$C7,$8A,$F1,$8E,$71,$93,$FB,$CA,$4E,$A3,$13
	.dc.b	$A7,$F4,$73,$A7,$84,$E7,$0F,$1C,$FF,$B6,$27,$27,$FA,$8F,$0F,$E6
	.dc.b	$FE,$1C,$8C,$1F,$EA,$7E,$EF,$E7,$7D,$F2,$30,$7F,$15,$F4,$DF,$A7
	.dc.b	$3E,$6B,$FB,$CA,$F9,$AA,$62,$33,$F1,$BA,$E6,$98,$73,$B4,$E2,$3F
	.dc.b	$88,$61,$DB,$E6,$7A,$99,$DF,$FC,$D5,$19,$9E,$5F,$41,$E2,$36,$F5
	.dc.b	$F9,$9F,$2F,$CD,$48,$C1,$0E,$F9,$23,$51,$89,$F6,$98,$C3,$C3,$5B
	.dc.b	$9E,$51,$8F,$8F,$F8,$7F,$1D,$27,$DD,$1F,$18,$C3,$E0,$8C,$3A,$3E
	.dc.b	$23,$A9,$E9,$1F,$1F,$E6,$BE,$7F,$44,$8C,$10,$F5,$B9,$7D,$B1,$F5
	.dc.b	$8C,$4F,$AA,$71,$3F,$4F,$E5,$7E,$DF,$9D,$23,$04,$3E,$3B,$97,$E4
	.dc.b	$C7,$E0,$8C,$57,$C6,$9C,$4F,$DF,$F9,$9F,$C3,$FC,$BC,$8C,$10,$FA
	.dc.b	$6E,$5F,$AB,$1F,$8E,$31,$5F,$4A,$71,$3F,$8B,$F3,$7F,$27,$F3,$32
	.dc.b	$30,$43,$EF,$B9,$7F,$03,$1F,$9A,$F3,$FF,$CA,$3F,$23,$27,$F2,$D1
	.dc.b	$97,$FE,$00,$46,$5F,$C4,$C7,$F9,$61,$FF,$2C,$FC,$34,$7F,$51,$D1
	.dc.b	$FE,$C7,$CB,$F3,$2D,$F1,$7F,$DF,$35,$BA,$FD,$0C,$E5,$73,$BF,$18
	.dc.b	$FF,$13,$C7,$EC,$D7,$C9,$7F,$9A,$9B,$19,$F7,$B0,$66,$64,$64,$87
	.dc.b	$EC,$27,$8B,$63,$8B,$DF,$DE,$FF,$51,$7C,$4F,$F8,$AB,$0F,$24,$63
	.dc.b	$C3,$57,$A3,$04,$E6,$31,$F0,$16,$F8,$C7,$FB,$BA,$F8,$C6,$21,$87
	.dc.b	$7C,$39,$C6,$CE,$65,$1F,$38,$E3,$FF,$0D,$C7,$FA,$E6,$31,$E2,$FF
	.dc.b	$3F,$6C,$7F,$44,$6B,$FE,$FD,$F8,$97,$A0,$F9,$AF,$E8,$D1,$86,$C6
	.dc.b	$67,$FA,$6C,$E5,$E9,$CE,$E3,$F0,$36,$33,$18,$3F,$AA,$EA,$F9,$37
	.dc.b	$4B,$9B,$E0,$4E,$63,$17,$F1,$B8,$7E,$4D,$BB,$33,$C5,$F5,$E3,$9E
	.dc.b	$3C,$9B,$7C,$4F,$C7,$B8,$9E,$15,$CA,$FA,$E1,$6F,$7C,$3E,$2E,$8F
	.dc.b	$E6,$B8,$BA,$F1,$9F,$92,$AD,$F6,$9B,$A7,$BE,$6D,$F1,$BF,$EA,$0F
	.dc.b	$DB,$6D,$6A,$76,$D8,$C9,$C3,$E6,$F9,$EF,$D1,$B1,$93,$AC,$65,$A7
	.dc.b	$F5,$FF,$9C,$78,$FE,$4E,$DF,$3C,$62,$7C,$B1,$89,$D4,$71,$6E,$9F
	.dc.b	$D7,$29,$E2,$F2,$F1,$B9,$B6,$FC,$0D,$AE,$0D,$DF,$FB,$0D,$BC,$5F
	.dc.b	$1F,$8B,$E4,$6E,$1E,$96,$E1,$E1,$6D,$F2,$6C,$FE,$C2,$FA,$3F,$06
	.dc.b	$FD,$FE,$2F,$C0,$F1,$6C,$FC,$56,$E3,$ED,$6E,$3E,$66,$CF,$56,$CF
	.dc.b	$F1,$CC,$3F,$C6,$93,$F3,$F9,$BE,$6F,$8B,$F3,$9B,$19,$3E,$A4,$65
	.dc.b	$A6,$6B,$E7,$46,$5B,$1F,$6F,$8B,$EC,$F3,$7D,$A6,$C6,$4F,$C0,$46
	.dc.b	$5A,$66,$7E,$D4,$65,$B1,$F8,$7E,$2F,$C2,$F2,$7E,$19,$B1,$93,$F1
	.dc.b	$91,$96,$99,$9F,$C3,$46,$5B,$1F,$93,$E3,$FC,$8F,$27,$E4,$9B,$1F
	.dc.b	$97,$30,$53,$81,$FE,$39,$F9,$A7,$CB,$FE,$6A,$3F,$63,$18,$78,$5F
	.dc.b	$F8,$CA,$1F,$2D,$AF,$97,$DB,$6B,$A7,$C1,$72,$F9,$7E,$6D,$8F,$07
	.dc.b	$CB,$F6,$F9,$7E,$ED,$8F,$F7,$07,$CE,$7F,$2D,$E1,$FE,$0A,$FF,$2F
	.dc.b	$E4,$BE,$2F,$F1,$6E,$8F,$E3,$F3,$9B,$E3,$39,$7F,$A8,$C6,$77,$FF
	.dc.b	$6D,$3D,$63,$77,$D3,$0E,$AD,$8C,$9F,$EE,$EF,$9C,$F8,$2F,$FA,$15
	.dc.b	$B7,$FE,$8A,$6B,$FD,$80,$FE,$C2,$E1,$C3,$C3,$1F,$F4,$3B,$67,$FB
	.dc.b	$3F,$C7,$3E,$A7,$FD,$13,$1E,$1E,$0F,$AB,$F3,$5F,$5B,$E6,$AC,$FD
	.dc.b	$B6,$E5,$F5,$DB,$97,$F9,$9B,$91,$1F,$7B,$E6,$25,$FE,$0F,$E6,$6E
	.dc.b	$1F,$7E,$6E,$1C,$66,$D7,$F7,$4C,$15,$8B,$FF,$0A,$C2,$B1,$A3,$E5
	.dc.b	$DC,$9F,$8E,$F9,$8F,$E8,$33,$98,$C6,$5F,$E0,$8C,$FF,$AD,$CE,$63
	.dc.b	$07,$C6,$BE,$6F,$E6,$FE,$67,$9C,$63,$3F,$CC,$3F,$9C,$3F,$3B,$F2
	.dc.b	$56,$BF,$45,$16,$7F,$3D,$37,$0F,$C6,$C7,$C8,$AF,$CA,$AE,$7F,$CA
	.dc.b	$B9,$3F,$1A,$66,$37,$BE,$A7,$6D,$8C,$9D,$23,$2D,$3F,$6D,$F3,$1F
	.dc.b	$B4,$F9,$9F,$DA,$B6,$32,$7E,$EE,$32,$D3,$E3,$DF,$F8,$CB,$E0,$9E
	.dc.b	$47,$F7,$0F,$45,$F3,$BF,$CF,$A9,$A7,$FB,$6A,$F7,$87,$2B,$DF,$97
	.dc.b	$FD,$E2,$FF,$DB,$B8,$DB,$DC,$72,$BF,$F2,$0D,$F6,$F2,$B5,$7F,$E0
	.dc.b	$09,$7C,$EB,$E1,$F0,$BE,$2A,$C6,$1B,$3E,$C3,$8B,$8C,$7C,$07,$F8
	.dc.b	$96,$B9,$4E,$DE,$CD,$9E,$1A,$7F,$8A,$23,$99,$F2,$29,$F0,$1F,$C8
	.dc.b	$DE,$B9,$BE,$77,$F3,$D5,$BA,$D5,$FF,$8C,$AB,$5D,$5F,$1E,$73,$5F
	.dc.b	$C1,$5B,$87,$F9,$0F,$0F,$D5,$CF,$3F,$23,$F3,$7C,$7F,$CD,$3E,$4A
	.dc.b	$CF,$F3,$96,$DF,$C0,$EA,$FD,$75,$7F,$3A,$F9,$3C,$EB,$FC,$07,$5F
	.dc.b	$E7,$BA,$7F,$3E,$6E,$5F,$D0,$DB,$8F,$F3,$F6,$D7,$FB,$BB,$E3,$7F
	.dc.b	$D1,$B5,$FD,$26,$BF,$C6,$F1,$87,$FA,$31,$8F,$82,$FE,$96,$D9,$E0
	.dc.b	$7F,$A8,$B5,$FF,$1E,$BF,$E4,$0F,$84,$F8,$7E,$D7,$F9,$7E,$31,$3A
	.dc.b	$78,$1F,$E4,$23,$87,$8B,$C7,$CE,$B1,$FE,$AE,$6E,$1F,$EA,$A6,$E7
	.dc.b	$FE,$B0,$6C,$FF,$A3,$B1,$46,$2B,$FD,$5F,$18,$9F,$F5,$CD,$F7,$69
	.dc.b	$DC,$FF,$5E,$F9,$BE,$BF,$3B,$D8,$D8,$C9,$FE,$C2,$8C,$B4,$FF,$62
	.dc.b	$F7,$FC,$3F,$13,$E2,$36,$32,$7F,$B2,$A3,$2D,$33,$8F,$FB,$22,$32
	.dc.b	$D8,$FF,$68,$F9,$1F,$3F,$C2,$FA,$0D,$8C,$9F,$ED,$68,$CB,$4C,$CF
	.dc.b	$FB,$52,$32,$D8,$CE,$BF,$ED,$C9,$CB,$63,$F5,$93,$FE,$C4,$F7,$3F
	.dc.b	$D8,$BF,$3A,$F1,$3F,$B4,$4F,$FA,$96,$B3,$18,$8F,$F6,$08,$9E,$0F
	.dc.b	$F6,$FE,$5F,$F5,$0C,$62,$3F,$DA,$A6,$E7,$AC,$E9,$E9,$7D,$A7,$95
	.dc.b	$D7,$8D,$F3,$7C,$87,$33,$CF,$3B,$73,$1B,$7C,$EF,$1B,$F0,$47,$03
	.dc.b	$91,$FC,$9A,$B9,$3C,$2F,$CF,$18,$BE,$D5,$CA,$BD,$BF,$02,$F8,$BF
	.dc.b	$5C,$23,$0F,$76,$BC,$23,$FC,$2D,$F0,$0F,$90,$FE,$83,$3C,$67,$31
	.dc.b	$8A,$E3,$5B,$D3,$FD,$FB,$18,$9D,$57,$09,$F9,$54,$F9,$0F,$F1,$26
	.dc.b	$BF,$56,$31,$8E,$B0,$F0,$CE,$5B,$AF,$9A,$F8,$B6,$B8,$61,$1F,$97
	.dc.b	$87,$46,$D7,$04,$23,$D4,$F0,$BE,$EE,$A3,$96,$1E,$7A,$E5,$3C,$6B
	.dc.b	$6D,$BF,$7C,$EA,$F8,$E1,$F0,$EB,$F1,$1D,$0E,$6F,$F1,$4C,$7F,$76
	.dc.b	$8C,$78,$3F,$E4,$9C,$7F,$DF,$71,$89,$D5,$70,$7F,$AF,$C7,$3F,$3F
	.dc.b	$87,$0F,$F1,$BC,$2D,$7F,$1A,$6D,$6E,$39,$D6,$2F,$FC,$71,$1D,$AB
	.dc.b	$13,$23,$04,$3C,$33,$E3,$C3,$CE,$BC,$DE,$9E,$74,$AF,$4C,$77,$30
	.dc.b	$F2,$8C,$3A,$3D,$EE,$AB,$AC,$F9,$FC,$EF,$47,$9D,$2F,$E2,$C6,$5F
	.dc.b	$8F,$19,$7C,$F1,$EF,$F0,$FE,$06,$D2,$7E,$6C,$65,$F9,$F1,$97,$DF
	.dc.b	$19,$7E,$0C,$67,$0F,$83,$39,$7E,$1C,$67,$6F,$E9,$D3,$F3,$FC,$3F
	.dc.b	$A1,$E2,$CC,$3E,$EC,$65,$FB,$F1,$97,$E7,$C7,$DA,$F0,$DF,$0D,$7E
	.dc.b	$22,$57,$E2,$E7,$2D,$9C,$BF,$88,$84,$F3,$9E,$9F,$3B,$D7,$E2,$65
	.dc.b	$F3,$8C,$BD,$63,$2F,$38,$FC,$7F,$92,$F9,$D7,$C3,$E1,$B5,$FE,$54
	.dc.b	$9D,$4F,$AF,$E7,$74,$FB,$7E,$76,$56,$EF,$8B,$FE,$3E,$1F,$3A,$F8
	.dc.b	$4F,$CF,$C6,$1B,$19,$8F,$F1,$E7,$37,$FB,$69,$FF,$1F,$46,$5E,$91
	.dc.b	$FE,$E4,$77,$E6,$F2,$AE,$BF,$3B,$FC,$B6,$7F,$C8,$CF,$F2,$CE,$77
	.dc.b	$37,$FB,$06,$30,$DF,$13,$89,$CB,$C4,$97,$CA,$32,$F4,$8C,$BE,$FA
	.dc.b	$F0,$3F,$07,$E7,$1E,$6F,$C0,$6C,$7F,$96,$5A,$FC,$7F,$9D,$7C,$1E
	.dc.b	$EB,$E3,$C3,$E1,$A9,$9A,$FC,$57,$CE,$FE,$3F,$AE,$46,$08,$7E,$2D
	.dc.b	$CB,$F8,$D8,$FA,$7F,$3A,$ED,$BD,$D2,$BF,$CD,$D3,$96,$CE,$5F,$BD
	.dc.b	$09,$E7,$32,$78,$C6,$F8,$7E,$8C,$7F,$3E,$AE,$8F,$E7,$FE,$77,$F5
	.dc.b	$04,$7F,$4D,$F3,$B5,$D1,$FA,$98,$FD,$74,$7E,$73,$E7,$7F,$9C,$4F
	.dc.b	$E6,$5A,$FF,$9C,$5B,$7F,$E6,$7F,$93,$E1,$E4,$7E,$DF,$CE,$97,$C6
	.dc.b	$3F,$75,$EE,$48,$FD,$C3,$FB,$DF,$3F,$11,$FB,$78,$B8,$7F,$9C,$E6
	.dc.b	$F4,$FF,$5C,$F3,$9C,$5F,$EA,$9C,$39,$9F,$C0,$EB,$F8,$47,$CE,$7F
	.dc.b	$09,$F7,$E4,$FF,$07,$8C,$BF,$C2,$63,$2F,$F0,$78,$FD,$6F,$CE,$BE
	.dc.b	$0A,$92,$7F,$D5,$1F,$3B,$F9,$6D,$3E,$9C,$2B,$83,$FE,$3B,$31,$55
	.dc.b	$A6,$63,$6F,$F9,$06,$97,$FB,$E7,$FD,$91,$1E,$0C,$31,$32,$30,$43
	.dc.b	$F8,$44,$F2,$7C,$F1,$CE,$64,$60,$FE,$55,$F0,$9F,$CB,$FC,$79,$18
	.dc.b	$3F,$97,$78,$3F,$CC,$FC,$19,$18,$3F,$BE,$7C,$CB,$E7,$3F,$B8,$AF
	.dc.b	$9D,$A6,$23,$3F,$3B,$AE,$69,$7C,$A2,$BF,$E0,$0B,$7C,$CF,$43,$3C
	.dc.b	$5F,$E0,$88,$CC,$F8,$BE,$77,$C2,$6D,$F2,$FC,$CF,$93,$E6,$A4,$60
	.dc.b	$87,$77,$2F,$86,$3D,$66,$2F,$C1,$38,$9F,$83,$F1,$BF,$03,$E3,$64
	.dc.b	$60,$87,$AD,$CB,$ED,$8F,$8E,$62,$BD,$93,$89,$F9,$BF,$35,$F3,$3A
	.dc.b	$24,$60,$87,$C7,$72,$FC,$98,$FA,$66,$2B,$E4,$4E,$27,$EC,$FC,$9F
	.dc.b	$D8,$F3,$A4,$60,$87,$D3,$72,$FD,$58,$FB,$E6,$2B,$EA,$4E,$27,$F0
	.dc.b	$BF,$33,$F8,$4F,$95,$91,$82,$1F,$7D,$CB,$F8,$18,$FC,$59,$8A,$FC
	.dc.b	$04,$E2,$7F,$23,$E5,$FE,$43,$E6,$64,$60,$87,$E2,$DC,$BF,$8D,$8F
	.dc.b	$CB,$F9,$FF,$E4,$5F,$90,$93,$F9,$C8,$CB,$FE,$54,$8C,$BF,$95,$8F
	.dc.b	$F2,$63,$FE,$4D,$F8,$78,$FF,$96,$8C,$9F,$E5,$C7,$2F,$F9,$30,$FF
	.dc.b	$2B,$FC,$EF,$F8,$59,$FF,$33,$2F,$91,$C9,$AF,$80,$93,$A8,$CB,$C2
	.dc.b	$32,$FF,$9A,$27,$93,$B3,$47,$6F,$7F,$FD,$0E,$BE,$2B,$FC,$33,$E4
	.dc.b	$A9,$5A,$9C,$C6,$2B,$FD,$BD,$E1,$7F,$35,$8F,$F1,$9C,$73,$8C,$57
	.dc.b	$16,$C7,$01,$7D,$E7,$51,$8A,$E7,$7F,$E2,$4F,$9D,$6F,$FE,$06,$AF
	.dc.b	$4C,$63,$C3,$FE,$60,$D8,$FF,$76,$35,$FF,$78,$FC,$53,$CF,$7C,$D7
	.dc.b	$F3,$78,$C3,$63,$33,$FE,$F5,$9C,$BD,$38,$DC,$3E,$06,$C6,$63,$07
	.dc.b	$F4,$3F,$5B,$FC,$37,$E7,$DF,$67,$FA,$4D,$7D,$29,$CB,$F5,$0F,$F5
	.dc.b	$DB,$87,$0E,$BA,$3F,$A8,$B6,$BE,$97,$C0,$FF,$81,$A7,$37,$67,$B7
	.dc.b	$C6,$DC,$3E,$77,$C2,$57,$FC,$35,$19,$E5,$FC,$14,$E7,$24,$FA,$2B
	.dc.b	$CA,$D7,$8D,$FB,$E3,$13,$C2,$3F,$DA,$B8,$FA,$27,$B7,$23,$E0,$FB
	.dc.b	$47,$7F,$81,$97,$FA,$4F,$56,$E5,$FE,$65,$C9,$58,$CB,$FD,$97,$97
	.dc.b	$F8,$1B,$0E,$E7,$CD,$9E,$FE,$EF,$BF,$F2,$9D,$63,$FA,$78,$C6,$5A
	.dc.b	$BE,$33,$FE,$EC,$AD,$E3,$DA,$F7,$5A,$A7,$E5,$3F,$D8,$63,$8F,$33
	.dc.b	$E0,$9F,$21,$F3,$EB,$56,$C7,$FC,$7B,$7B,$AD,$7A,$7F,$36,$F7,$7F
	.dc.b	$85,$6B,$FB,$A6,$A3,$C7,$E9,$FD,$FB,$F1,$7C,$6F,$86,$31,$92,$31
	.dc.b	$9F,$A2,$30,$FB,$70,$FE,$FC,$E6,$77,$87,$FA,$5E,$72,$FA,$E7,$C5
	.dc.b	$87,$B7,$E3,$F7,$A7,$F8,$A2,$DE,$67,$97,$DE,$7A,$FF,$16,$30,$DC
	.dc.b	$7E,$5C,$E9,$F2,$4E,$9A,$4C,$BC,$47,$EF,$B2,$FF,$5B,$C6,$71,$F1
	.dc.b	$79,$9F,$17,$EE,$DE,$BF,$D5,$8C,$36,$BE,$DC,$E9,$F8,$93,$A6,$93
	.dc.b	$0F,$0E,$1F,$4F,$D5,$E5,$1F,$EB,$DF,$C9,$FE,$6B,$AF,$C2,$7D,$8E
	.dc.b	$78,$7E,$C4,$63,$4F,$F2,$7D,$70,$7F,$0F,$19,$C7,$F9,$DC,$E5,$FA
	.dc.b	$31,$9B,$FC,$24,$E5,$B7,$FC,$07,$F2,$8F,$C9,$E1,$DA,$3F,$82,$53
	.dc.b	$E4,$BF,$15,$D7,$F0,$70,$F5,$46,$37,$DB,$73,$FD,$25,$F1,$7F,$47
	.dc.b	$7C,$71,$F5,$D7,$09,$FD,$56,$4F,$73,$19,$7C,$AF,$BB,$E4,$DB,$F2
	.dc.b	$FA,$7F,$9D,$9C,$E3,$F1,$A3,$13,$AC,$7F,$67,$5A,$7F,$B0,$FC,$65
	.dc.b	$C3,$E2,$DF,$F2,$38,$C4,$34,$4F,$F2,$9F,$D2,$5F,$B2,$FE,$F3,$0F
	.dc.b	$E0,$F1,$8E,$EF,$27,$C6,$6E,$B5,$8F,$E0,$7C,$BC,$E1,$F1,$9B,$1F
	.dc.b	$F0,$07,$3F,$F6,$24,$FF,$0D,$6C,$7E,$72,$1E,$A7,$FB,$45,$D6,$1F
	.dc.b	$CE,$FE,$89,$F3,$9F,$CF,$F8,$7F,$89,$72,$E6,$31,$E0,$FD,$3C,$3F
	.dc.b	$C0,$77,$C1,$FE,$43,$7C,$1F,$E3,$D1,$9C,$BF,$5B,$39,$7F,$5E,$23
	.dc.b	$4F,$F5,$B6,$BD,$AF,$FA,$44,$62,$3F,$C0,$7F,$93,$75,$AF,$F6,$FF
	.dc.b	$9A,$87,$CF,$7F,$55,$75,$8F,$FB,$15,$CB,$FC,$E2,$78,$63,$BA,$E0
	.dc.b	$D9,$D5,$7F,$69,$C3,$C3,$F3,$DD,$7D,$0E,$CD,$AD,$B6,$7F,$86,$E3
	.dc.b	$E5,$8F,$ED,$9F,$49,$FE,$0C,$AB,$E0,$3D,$DB,$18,$6B,$FD,$3A,$7F
	.dc.b	$C2,$71,$8C,$3F,$8C,$C6,$27,$54,$79,$18,$2C,$FD,$2D,$9F,$33,$FD
	.dc.b	$FF,$95,$F4,$5F,$CD,$B8,$7F,$DB,$35,$AC,$3F,$AF,$D6,$BC,$BF,$C5
	.dc.b	$61,$F5,$89,$3F,$ED,$DA,$E1,$8F,$F8,$5F,$6F,$AD,$F4,$1F,$57,$E8
	.dc.b	$3E,$B3,$6F,$E7,$B6,$BE,$4F,$12,$31,$A7,$FB,$B3,$EC,$5E,$87,$F8
	.dc.b	$5F,$E7,$1F,$6F,$FE,$E6,$CB,$FB,$85,$69,$FE,$EF,$3A,$F9,$1F,$F0
	.dc.b	$E4,$E7,$3F,$E8,$73,$9A,$E3,$87,$5C,$3A,$57,$E6,$AF,$83,$FE,$33
	.dc.b	$6B,$FE,$68,$6B,$FD,$BD,$BA,$FE,$5F,$C0,$FC,$B7,$B1,$F9,$76,$E3
	.dc.b	$F9,$06,$DF,$C2,$6E,$9F,$76,$3F,$BD,$7C,$83,$E2,$3E,$D7,$37,$F6
	.dc.b	$8C,$3F,$C2,$13,$97,$EC,$C6,$6B,$F6,$73,$97,$FC,$98,$1C,$3F,$E4
	.dc.b	$F6,$BF,$E8,$46,$BF,$E0,$BF,$C2,$FF,$D2,$95,$FE,$24,$FC,$4D,$F8
	.dc.b	$BB,$E2,$7F,$C9,$55,$AB,$FE,$E3,$5A,$9C,$FC,$57,$F9,$46,$31,$87
	.dc.b	$F8,$A6,$F8,$69,$F0,$67,$38,$7F,$0C,$9E,$8F,$F9,$95,$AF,$FA,$69
	.dc.b	$AF,$9E,$FF,$CA,$9E,$53,$E8,$FF,$45,$F4,$DF,$E4,$8C,$BF,$C7,$13
	.dc.b	$97,$F4,$31,$9B,$FE,$37,$39,$7E,$4B,$5F,$F3,$B3,$5F,$F5,$5B,$57
	.dc.b	$33,$F6,$6F,$E6,$D1,$FC,$23,$E5,$7F,$75,$87,$E6,$27,$3A,$FF,$20
	.dc.b	$9C,$D7,$E6,$27,$8B,$FE,$B8,$BE,$9F,$07,$FE,$73,$6E,$1F,$79,$B3
	.dc.b	$7C,$BF,$F3,$F4,$7E,$E3,$1F,$C7,$4F,$9A,$73,$E0,$69,$FF,$3E,$57
	.dc.b	$F5,$09,$FF,$4A,$35,$F1,$47,$FA,$0F,$A3,$F9,$B7,$AD,$FC,$D4,$FE
	.dc.b	$6D,$F4,$D7,$2F,$E5,$F1,$6F,$F8,$0C,$DC,$BF,$41,$8B,$BB,$FD,$7F
	.dc.b	$87,$EF,$3F,$67,$FE,$0C,$FE,$FA,$33,$87,$EF,$27,$2F,$EF,$A3,$2F
	.dc.b	$C7,$6B,$FE,$65,$AE,$DF,$1D,$F2,$67,$33,$FE,$96,$F7,$FF,$D4,$5F
	.dc.b	$6A,$FA,$0F,$F5,$36,$69,$CC,$61,$FF,$51,$46,$73,$FD,$7D,$FF,$9F
	.dc.b	$67,$24,$8F,$F4,$44,$EE,$7F,$DF,$15,$CE,$73,$56,$73,$18,$3F,$8F
	.dc.b	$FE,$B4,$F3,$BF,$D7,$1F,$B4,$3E,$8F,$F9,$44,$E6,$BF,$B2,$4E,$5E
	.dc.b	$35,$F8,$99,$D5,$FD,$E6,$BF,$8F,$6B,$FE,$D2,$9F,$ED,$DF,$87,$70
	.dc.b	$CF,$FD,$B7,$CD,$FE,$C9,$8C,$4E,$EB,$FB,$3C,$61,$FF,$70,$47,$2C
	.dc.b	$3F,$DA,$51,$F9,$E3,$81,$FC,$62,$BF,$D9,$CE,$4F,$F7,$37,$CA,$FF
	.dc.b	$A7,$FE,$B1,$F5,$9F,$EA,$1E,$76,$5F,$BB,$CF,$FD,$B1,$7B,$AD,$4E
	.dc.b	$73,$FE,$4B,$39,$C5,$87,$2F,$83,$F2,$C2,$75,$39,$AF,$F6,$14,$FF
	.dc.b	$AC,$67,$93,$FE,$B3,$F4,$BF,$9E,$67,$E6,$8F,$48,$70,$EE,$B8,$BC
	.dc.b	$CD,$65,$FE,$BD,$74,$60,$E7,$3C,$95,$78,$27,$DB,$1B,$38,$CE,$DF
	.dc.b	$61,$93,$8D,$78,$9C,$C7,$5A,$E9,$1E,$0A,$F2,$3C,$0E,$47,$07,$31
	.dc.b	$C2,$3D,$57,$EE,$8C,$47,$BE,$BB,$5F,$86,$B4,$D9,$F3,$57,$07,$B1
	.dc.b	$F1,$5C,$4F,$FB,$8F,$E5,$FF,$D9,$B3,$F4,$1A,$F8,$83,$87,$E8,$3A
	.dc.b	$F9,$5F,$E0,$1A,$F3,$F8,$BF,$A5,$18,$AE,$31,$8E,$4F,$F7,$E4,$60
	.dc.b	$FF,$7A,$4E,$EF,$A4,$EB,$3F,$C0,$CB,$E9,$3E,$F4,$61,$BA,$F5,$CB
	.dc.b	$F1,$B5,$E1,$AD,$63,$8B,$DB,$E0,$C3,$B6,$1C,$63,$F0,$57,$DD,$E0
	.dc.b	$74,$9D,$BC,$70,$CB,$C2,$FA,$AA,$F1,$5F,$5C,$3C,$73,$B3,$A5,$C7
	.dc.b	$04,$8F,$83,$87,$82,$3A,$36,$FC,$D1,$BB,$EC,$47,$F6,$FC,$1F,$DF
	.dc.b	$DF,$E2,$BE,$27,$E9,$F1,$FE,$BF,$7E,$69,$F6,$1F,$A9,$E0,$FE,$0F
	.dc.b	$5B,$E0,$FE,$0A,$D9,$DE,$7F,$5B,$E8,$BF,$9E,$C7,$C1,$6E,$06,$92
	.dc.b	$3F,$D2,$9F,$98,$BE,$B7,$A6,$48,$C4,$8F,$F5,$5D,$F3,$71,$9F,$AB
	.dc.b	$0E,$35,$C2,$7F,$D2,$6D,$9D,$47,$F4,$6F,$0F,$FA,$9F,$7B,$A7,$FD
	.dc.b	$0B,$3F,$E0,$76,$BF,$D0,$67,$71,$8A,$FF,$59,$F5,$7D,$26,$F1,$FC
	.dc.b	$5C,$3A,$5F,$C7,$C3,$F8,$95,$7F,$83,$B8,$BF,$82,$B6,$3E,$0F,$CA
	.dc.b	$FF,$57,$7D,$BC,$5F,$C7,$6B,$6D,$8E,$91,$87,$F9,$AC,$F1,$7F,$DA
	.dc.b	$31,$F5,$1F,$E3,$87,$13,$FB,$9B,$93,$96,$EF,$97,$FF,$47,$CF,$FB
	.dc.b	$25,$AF,$43,$DD,$F8,$EB,$D4,$F9,$6D,$8F,$F1,$83,$5F,$F1,$3B,$5F
	.dc.b	$F1,$C3,$9C,$BF,$C3,$FE,$23,$BB,$E5,$36,$BF,$D3,$75,$BD,$3F,$DB
	.dc.b	$DD,$DF,$88,$F6,$4D,$31,$F1,$76,$FF,$DB,$B1,$5F,$F5,$7C,$7D,$5F
	.dc.b	$DA,$7F,$C5,$D1,$FE,$16,$6B,$FE,$86,$6B,$FC,$F1,$AF,$F5,$17,$79
	.dc.b	$FF,$92,$97,$C3,$F5,$3B,$78,$FF,$85,$FD,$AB,$E5,$DF,$F9,$65,$AF
	.dc.b	$F8,$CD,$AF,$F9,$A1,$BC,$A7,$E9,$E1,$FE,$A5,$E2,$68,$F8,$9F,$88
	.dc.b	$77,$F7,$7F,$25,$F9,$77,$F3,$BC,$BF,$B4,$35,$FD,$73,$53,$4F,$E8
	.dc.b	$B8,$7F,$99,$BE,$3B,$F5,$9C,$3F,$DC,$5B,$5C,$BC,$6F,$D0,$7E,$38
	.dc.b	$F9,$FF,$E0,$AD,$9F,$F2,$1B,$5F,$F5,$9B,$5F,$ED,$2E,$70,$FF,$3A
	.dc.b	$F8,$EF,$A3,$FD,$63,$6B,$F4,$E7,$F0,$7A,$E1,$97,$F8,$15,$B1,$FC
	.dc.b	$2E,$78,$2B,$EB,$7E,$8C,$3F,$71,$F8,$7B,$ED,$FF,$8E,$B6,$FF,$CB
	.dc.b	$6D,$7F,$D9,$6D,$7F,$C0,$2E,$6F,$FD,$1B,$F8,$7B,$E0,$FF,$82,$36
	.dc.b	$BD,$22,$B8,$69,$FE,$39,$E2,$FE,$20,$DF,$03,$F4,$07,$E8,$7C,$29
	.dc.b	$7F,$0B,$97,$F0,$73,$A8,$C5,$F2,$7F,$96,$7D,$1B,$E9,$7F,$9C,$B6
	.dc.b	$3F,$CE,$ED,$7F,$DC,$ED,$7F,$C3,$AE,$75,$FF,$54,$FC,$DF,$E0,$A4
	.dc.b	$D9,$D6,$7F,$E3,$76,$CF,$F3,$D9,$E0,$6E,$FF,$79,$E1,$7E,$F7,$DA
	.dc.b	$95,$EA,$8C,$BE,$C8,$CB,$F6,$F0,$F9,$13,$FC,$C7,$0F,$D9,$CF,$B3
	.dc.b	$BB,$DE,$47,$07,$D2,$FB,$1F,$D5,$37,$B3,$94,$FF,$3B,$FB,$C7,$C4
	.dc.b	$5F,$A8,$97,$F6,$A7,$2D,$9C,$BE,$53,$FA,$E4,$27,$A4,$C9,$E5,$1C
	.dc.b	$37,$FF,$6A,$D7,$49,$E5,$C5,$FC,$03,$D7,$FE,$45,$F0,$32,$7E,$FC
	.dc.b	$7F,$25,$30,$F0,$E3,$4E,$63,$F1,$1F,$8C,$3E,$13,$F9,$64,$43,$2F
	.dc.b	$F7,$68,$D4,$7E,$30,$D1,$E4,$72,$FB,$C7,$D8,$BE,$AF,$FC,$0E,$D8
	.dc.b	$FF,$62,$B5,$FF,$34,$9C,$23,$67,$F6,$BF,$B1,$7C,$65,$3F,$B7,$57
	.dc.b	$1D,$BF,$DF,$6D,$E6,$FF,$43,$57,$0C,$3F,$5E,$D7,$CD,$3B,$5F,$28
	.dc.b	$FD,$AF,$FB,$73,$63,$FC,$44,$D7,$F8,$8F,$CF,$7F,$B8,$FE,$BF,$FC
	.dc.b	$01,$F9,$77,$F8,$17,$2F,$F0,$D5,$BF,$2E,$7C,$B7,$26,$E3,$FC,$0A
	.dc.b	$F8,$2B,$FE,$67,$F3,$7F,$E3,$16,$C7,$F2,$06,$BF,$8D,$C3,$B6,$1E
	.dc.b	$6F,$33,$FB,$E4,$F5,$FB,$5F,$29,$9F,$6A,$C7,$F2,$AB,$F3,$7C,$77
	.dc.b	$94,$CC,$75,$D1,$19,$8E,$11,$87,$B1,$FC,$73,$B9,$F5,$9F,$E2,$36
	.dc.b	$C6,$63,$E5,$FE,$5D,$FC,$22,$30,$78,$FC,$EF,$EA,$3B,$7C,$A6,$C7
	.dc.b	$F9,$16,$F9,$BF,$8F,$6F,$37,$C9,$8F,$B7,$3C,$33,$FE,$A9,$B7,$8F
	.dc.b	$2F,$27,$CE,$CB,$F4,$3F,$E3,$9F,$39,$F1,$99,$9F,$9D,$F9,$73,$E9
	.dc.b	$3F,$06,$D8,$E2,$43,$71,$A3,$96,$9E,$58,$7E,$5D,$FC,$41,$FD,$43
	.dc.b	$63,$4D,$DF,$F1,$7E,$1B,$A7,$FC,$9D,$39,$BF,$CC,$FE,$8F,$FE,$BA
	.dc.b	$7E,$E3,$5E,$BE,$4F,$03,$F3,$71,$9C,$7E,$47,$0A,$31,$97,$FB,$BF
	.dc.b	$0E,$D8,$7A,$3E,$97,$C2,$47,$F1,$59,$BC,$CC,$47,$E5,$7E,$01,$F5
	.dc.b	$3F,$AA,$6C,$7F,$62,$C1,$87,$0D,$3C,$75,$AA,$E5,$79,$CB,$C9,$D5
	.dc.b	$F0,$89,$3F,$9B,$F9,$2F,$D0,$BF,$32,$3F,$73,$E2,$30,$A7,$09,$D1
	.dc.b	$FA,$38,$D3,$63,$58,$FD,$47,$FD,$43,$9E,$A7,$F8,$CB,$5F,$F4,$D1
	.dc.b	$FE,$9B,$70,$FC,$B8,$F1,$FA,$92,$3E,$14,$7F,$A1,$3A,$BF,$D1,$5F
	.dc.b	$A6,$FA,$B0,$FC,$43,$74,$FF,$53,$79,$AF,$88,$FF,$0F,$B6,$F8,$F8
	.dc.b	$33,$2F,$F0,$35,$7F,$89,$16,$1F,$79,$FF,$58,$35,$CF,$C9,$FF,$A5
	.dc.b	$BD,$DF,$1B,$23,$FC,$33,$3C,$71,$F1,$57,$FA,$87,$ED,$9F,$75,$49
	.dc.b	$7E,$2F,$8C,$AF,$F4,$09,$DB,$FD,$08,$38,$75,$3F,$ED,$08,$C6,$DF
	.dc.b	$BF,$E8,$7E,$69,$FE,$74,$E0,$FF,$65,$46,$3E,$23,$BD,$7E,$C6,$DF
	.dc.b	$73,$E7,$26,$5F,$1B,$CA,$AF,$96,$3C,$04,$7C,$D2,$F9,$BF,$D7,$47
	.dc.b	$E0,$23,$1C,$3A,$9C,$CF,$FB,$76,$72,$FF,$60,$F9,$4F,$F3,$47,$81
	.dc.b	$F0,$5B,$19,$E6,$FE,$AA,$D9,$F0,$F7,$B6,$A4,$9F,$99,$19,$7F,$B6
	.dc.b	$FB,$AF,$AA,$FE,$4D,$BF,$F1,$38,$C7,$0F,$E1,$BE,$E7,$2F,$E6,$A3
	.dc.b	$37,$FD,$F5,$B5,$C6,$F7,$3F,$ED,$4B,$AF,$FA,$F6,$2D,$7F,$B2,$6F
	.dc.b	$EF,$36,$7C,$BF,$25,$FC,$95,$F0,$35,$1F,$EC,$77,$99,$E0,$E8,$FD
	.dc.b	$17,$DC,$E2,$FF,$CD,$B7,$D7,$2F,$75,$EF,$AD,$7B,$C7,$D9,$F2,$3E
	.dc.b	$63,$5F,$73,$72,$E7,$16,$7F,$C6,$76,$AC,$63,$FE,$B5,$87,$C7,$FC
	.dc.b	$2F,$B9,$D6,$3F,$B2,$8C,$5F,$8D,$B1,$93,$FD,$0B,$3C,$30,$FF,$16
	.dc.b	$7D,$C7,$F6,$C9,$FD,$33,$63,$FC,$52,$D7,$F8,$88,$70,$FF,$14,$78
	.dc.b	$63,$FD,$4D,$B3,$FE,$18,$8F,$F0,$BB,$FE,$96,$8C,$CF,$D2,$5E,$8F
	.dc.b	$E0,$39,$1F,$63,$FE,$06,$BE,$13,$98,$C5,$FA,$A3,$EE,$3E,$56,$DF
	.dc.b	$4A,$F6,$C6,$0F,$C5,$C6,$67,$FC,$41,$83,$EE,$73,$5A,$FB,$9F,$27
	.dc.b	$86,$F5,$B1,$97,$F8,$7E,$30,$FB,$67,$C2,$47,$E6,$B7,$1F,$A2,$FB
	.dc.b	$F4,$FF,$7F,$CF,$13,$E3,$7C,$67,$C3,$FB,$9F,$F0,$34,$60,$FF,$17
	.dc.b	$38,$BF,$B7,$3F,$00,$93,$F7,$1B,$1A,$8F,$F1,$B2,$FB,$9F,$87,$3F
	.dc.b	$AE,$FB,$7F,$9B,$B7,$F7,$AF,$B9,$F6,$DF,$F9,$CE,$73,$87,$F9,$52
	.dc.b	$73,$5F,$E1,$39,$FA,$B5,$F8,$86,$BF,$D6,$9A,$FE,$C9,$A9,$EA,$E7
	.dc.b	$D0,$FE,$03,$F7,$3F,$88,$8D,$47,$9F,$CE,$FC,$47,$4F,$C3,$C3,$C5
	.dc.b	$18,$F6,$35,$9F,$F4,$3C,$BF,$BC,$36,$7F,$13,$18,$C7,$FD,$12,$DB
	.dc.b	$E7,$5F,$EE,$EC,$BF,$C2,$2F,$FB,$C2,$74,$FF,$9D,$BE,$62,$DF,$FA
	.dc.b	$09,$AF,$B7,$E2,$FF,$3D,$18,$AF,$F5,$5D,$79,$67,$C6,$47,$FA,$D7
	.dc.b	$B3,$B8,$FE,$46,$D8,$FF,$44,$FD,$CB,$DE,$FE,$53,$F0,$C8,$C4,$FF
	.dc.b	$9D,$71,$7C,$96,$A7,$39,$FD,$DF,$27,$E1,$92,$FE,$43,$FA,$EC,$B8
	.dc.b	$C7,$F2,$AD,$BF,$CA,$51,$8A,$FF,$6B,$4B,$E0,$9E,$37,$A3,$2F,$E7
	.dc.b	$7F,$6E,$FB,$9F,$E7,$F9,$7F,$00,$E4,$CC,$63,$C2,$FB,$99,$7F,$88
	.dc.b	$AF,$83,$FD,$F5,$B9,$FF,$B0,$1A,$FF,$18,$6B,$C6,$BA,$CF,$A4,$FE
	.dc.b	$87,$F3,$CF,$C0,$FF,$D8,$B1,$9A,$FE,$8F,$7D,$B2,$FF,$42,$79,$6F
	.dc.b	$92,$C5,$FF,$8A,$9F,$F2,$8B,$C6,$35,$97,$F4,$D9,$CE,$1F,$EE,$B8
	.dc.b	$C7,$C1,$65,$B7,$C6,$BF,$B2,$BD,$7E,$1B,$97,$9F,$CD,$B1,$B6,$E1
	.dc.b	$FE,$F7,$7F,$B5,$FD,$D7,$FB,$52,$6F,$88,$F8,$CB,$18,$6B,$F0,$67
	.dc.b	$FD,$B7,$18,$B6,$BF,$EA,$59,$CD,$6E,$FF,$DA,$0D,$BE,$15,$B9,$F0
	.dc.b	$4F,$CD,$B6,$CF,$99,$FF,$00,$F4,$3E,$DF,$F9,$B6,$3F,$E9,$FA,$D6
	.dc.b	$1F,$56,$B5,$E5,$FE,$33,$09,$39,$C3,$F1,$38,$FD,$A9,$FA,$5F,$1D
	.dc.b	$F4,$FE,$EA,$DF,$CF,$6D,$7B,$A3,$EE,$46,$33,$FC,$0F,$0A,$3F,$C2
	.dc.b	$BF,$73,$FD,$DA,$BF,$11,$CE,$FB,$A7,$C0,$EB,$D4,$FF,$0E,$CE,$63
	.dc.b	$15,$FE,$57,$BE,$59,$FF,$70,$0E,$1F,$EE,$6D,$7F,$D0,$AF,$0C,$FF
	.dc.b	$D8,$8D,$9F,$F0,$FF,$E1,$CF,$C3,$9F,$47,$E0,$C3,$DC,$3B,$3F,$39
	.dc.b	$E3,$FE,$5F,$DB,$FC,$C3,$71,$FC,$A3,$6B,$8D,$70,$9F,$CC,$57,$F8
	.dc.b	$E3,$E6,$1F,$3D,$F7,$39,$9E,$AE,$63,$0D,$BF,$C2,$57,$07,$ED,$7A
	.dc.b	$79,$CB,$CE,$3E,$E5,$F3,$BF,$E4,$8F,$BA,$7D,$E3,$EF,$3F,$C9,$35
	.dc.b	$AC,$7F,$79,$5A,$9C,$DF,$F8,$2A,$B5,$87,$D1,$6B,$FE,$22,$6B,$FE
	.dc.b	$B4,$77,$CB,$FE,$F3,$6E,$1F,$E8,$8D,$BF,$3F,$5E,$89,$CB,$FE,$52
	.dc.b	$F4,$9F,$69,$FA,$0F,$BC,$7B,$1F,$E6,$09,$CE,$3F,$9F,$9C,$BF,$A9
	.dc.b	$8C,$DF,$C4,$6E,$5F,$E9,$BA,$B5,$FE,$CA,$9B,$4E,$B7,$DD,$FF,$36
	.dc.b	$AF,$E3,$1F,$79,$FB,$AA,$FC,$2D,$6B,$D2,$FF,$33,$4E,$72,$5F,$F2
	.dc.b	$AC,$39,$61,$C1,$B3,$FA,$8F,$81,$FD,$0C,$62,$BF,$DB,$B8,$FF,$49
	.dc.b	$BF,$F2,$7B,$5C,$FA,$7F,$E3,$E8,$C4,$7F,$38,$8F,$F4,$2C,$7F,$51
	.dc.b	$C3,$F6,$BE,$E6,$27,$F9,$83,$2F,$F9,$67,$DE,$5C,$3F,$33,$E3,$7F
	.dc.b	$84,$3B,$7F,$9E,$5F,$F3,$7A,$D6,$3E,$EB,$F9,$1E,$1F,$FA,$DF,$1F
	.dc.b	$DB,$67,$FB,$EF,$8D,$FF,$47,$36,$33,$3B,$AF,$6B,$6F,$A5,$6E,$7F
	.dc.b	$57,$FB,$67,$F4,$C7,$BE,$1F,$EA,$2A,$7C,$1F,$F9,$81,$B8,$FF,$B7
	.dc.b	$DB,$5F,$C2,$5A,$FF,$9E,$9C,$9E,$B7,$F9,$DC,$FF,$A6,$FE,$F3,$B3
	.dc.b	$2B,$F9,$B5,$70,$7F,$D5,$3A,$BE,$25,$7B,$C3,$F6,$B7,$B7,$FC,$0D
	.dc.b	$87,$5C,$39,$BF,$C8,$BE,$D9,$EA,$FC,$86,$CF,$CC,$6D,$7E,$11,$AF
	.dc.b	$E2,$1C,$DF,$CA,$6C,$F9,$BD,$6F,$F2,$F3,$FC,$D2,$33,$5F,$ED,$38
	.dc.b	$FE,$ED,$F7,$78,$8F,$F6,$2F,$47,$FB,$0E,$31,$1F,$9C,$3F,$DC,$33
	.dc.b	$FC,$D5,$DB,$5D,$1F,$EA,$89,$FF,$6C,$BF,$9F,$3F,$97,$7E,$40,$F9
	.dc.b	$BC,$CE,$A5,$F3,$DF,$ED,$6F,$BC,$D4,$7F,$35,$C7,$F8,$0B,$FC,$88
	.dc.b	$FF,$6D,$CE,$EF,$EA,$CE,$B4,$FF,$70,$4F,$FB,$83,$E3,$39,$DF,$FB
	.dc.b	$8B,$D8,$FC,$E4,$62,$39,$9F,$E9,$FC,$3F,$D8,$33,$A7,$FA,$A5,$77
	.dc.b	$9F,$F7,$58,$9F,$D3,$63,$CE,$3D,$61,$ED,$1A,$AE,$73,$97,$BE,$9D
	.dc.b	$63,$AD,$D7,$B4,$57,$AD,$F4,$8C,$4E,$EF,$B5,$7F,$B6,$A3,$E1,$5F
	.dc.b	$68,$F7,$19,$29,$F1,$6B,$6E,$6F,$B3,$E7,$31,$3C,$E3,$E0,$57,$09
	.dc.b	$E4,$74,$3E,$6E,$1C,$8D,$B9,$C3,$E7,$3E,$38,$D5,$FC,$6F,$96,$FF
	.dc.b	$69,$4F,$9D,$AF,$C6,$0E,$1F,$B6,$EB,$E5,$7F,$84,$E6,$9D,$DF,$DA
	.dc.b	$9C,$C6,$0D,$38,$AE,$5C,$5F,$E4,$29,$D0,$9E,$35,$F8,$E5,$F3,$7F
	.dc.b	$86,$8C,$37,$1F,$05,$F3,$9E,$0D,$8F,$24,$EE,$FC,$6D,$7C,$13,$B7
	.dc.b	$CE,$F9,$74,$EF,$1F,$8B,$C3,$C6,$D9,$D5,$F8,$B2,$FB,$15,$AA,$FC
	.dc.b	$1B,$71,$F2,$CE,$AF,$8D,$79,$0E,$D8,$FE,$93,$1D,$C7,$E1,$CF,$3D
	.dc.b	$69,$86,$70,$FA,$33,$C0,$F4,$C6,$6B,$A6,$3E,$78,$E1,$87,$C1,$7E
	.dc.b	$94,$FC,$5B,$F4,$56,$F1,$E8,$FF,$06,$F1,$7F,$73,$97,$FA,$73,$7F
	.dc.b	$F0,$9E,$3F,$E7,$5C,$3C,$77,$F8,$5F,$BB,$DF,$05,$AD,$46,$1F,$E3
	.dc.b	$BC,$35,$FC,$24,$7F,$25,$B3,$49,$1F,$CA,$E3,$2F,$FA,$A7,$EE,$DF
	.dc.b	$79,$D7,$A9,$18,$9F,$09,$FC,$F7,$DD,$F1,$77,$3A,$3F,$D1,$93,$93
	.dc.b	$FD,$69,$D2,$FC,$03,$73,$A9,$FE,$9D,$7F,$CF,$6F,$93,$67,$53,$FC
	.dc.b	$80,$FE,$51,$C3,$FD,$AF,$83,$FB,$77,$07,$F2,$46,$CF,$F9,$B7,$B5
	.dc.b	$F7,$DF,$16,$73,$3C,$AB,$83,$EF,$AD,$BF,$EA,$79,$D9,$1D,$57,$37
	.dc.b	$FA,$94,$7F,$53,$7F,$99,$1D,$0F,$F0,$BF,$DE,$3D,$7F,$8E,$D8,$FF
	.dc.b	$4C,$B5,$FF,$09,$F8,$BA,$9F,$89,$39,$9F,$F1,$CB,$5F,$F7,$5B,$5F
	.dc.b	$F4,$67,$8D,$FE,$5C,$7F,$DB,$9E,$6F,$DB,$8C,$47,$F8,$CB,$E4,$B1
	.dc.b	$32,$36,$7F,$B8,$E3,$F1,$9F,$79,$FE,$AC,$8F,$F6,$53,$5F,$F4,$2B
	.dc.b	$76,$FF,$5B,$D7,$0F,$23,$FB,$B4,$EE,$BE,$D3,$5F,$EB,$6D,$7F,$CC
	.dc.b	$DF,$78,$F8,$0F,$C6,$B6,$3F,$D7,$ED,$7F,$C9,$7D,$DA,$9F,$77,$C7
	.dc.b	$C9,$DC,$7E,$97,$EF,$3F,$9C,$E7,$FE,$33,$6F,$27,$F6,$C6,$D7,$F3
	.dc.b	$7F,$92,$7D,$37,$1F,$0F,$F5,$D1,$FA,$39,$FF,$52,$B7,$0E,$93,$FB
	.dc.b	$3F,$BC,$7D,$0D,$25,$FF,$BC,$FC,$7F,$F5,$BB,$FE,$F5,$9D,$BF,$E1
	.dc.b	$F6,$BF,$ED,$56,$BF,$A3,$6B,$FE,$04,$77,$7F,$E4,$F9,$FF,$68,$CF
	.dc.b	$7F,$83,$F0,$BF,$EB,$CF,$BA,$7D,$E7,$F0,$96,$C7,$F5,$7B,$FF,$2C
	.dc.b	$46,$23,$FD,$87,$F7,$8F,$57,$85,$FF,$11,$87,$C3,$53,$35,$CF,$87
	.dc.b	$FC,$B7,$D5,$FC,$07,$2F,$F7,$EF,$77,$E5,$4F,$CB,$78,$52,$F8,$1F
	.dc.b	$9C,$8D,$38,$F1,$D7,$CF,$0F,$F1,$2F,$D2,$3E,$F3,$F9,$23,$6F,$98
	.dc.b	$70,$FF,$B7,$BE,$17,$EF,$BA,$8C,$BF,$C4,$5A,$F6,$9F,$F2,$94,$62
	.dc.b	$3F,$BF,$3D,$33,$FF,$40,$57,$E1,$6F,$FC,$72,$DA,$DD,$FE,$D7,$CE
	.dc.b	$FD,$B7,$B5,$27,$DD,$92,$75,$18,$9D,$3F,$EA,$C8,$F7,$FC,$CC,$8F
	.dc.b	$53,$FD,$17,$EF,$1F,$2D,$99,$FF,$7E,$C3,$BF,$46,$75,$FF,$4C,$7D
	.dc.b	$DF,$E5,$A9,$18,$FB,$CF,$EA,$12,$6F,$C6,$CA,$F9,$F3,$96,$CE,$5E
	.dc.b	$7C,$1F,$E5,$C9,$ED,$7D,$67,$67,$2B,$FE,$35,$ED,$7F,$1B,$F8,$49
	.dc.b	$3F,$1A,$32,$FC,$88,$CB,$FE,$D0,$8F,$EB,$9F,$78,$F8,$9F,$E4,$93
	.dc.b	$0F,$C6,$46,$5B,$19,$7F,$DB,$27,$27,$FB,$87,$DD,$BE,$F2,$92,$7F
	.dc.b	$1F,$F6,$7F,$E4,$67,$F2,$33,$B7,$FD,$38,$D7,$FA,$F0,$70,$FF,$61
	.dc.b	$6B,$FE,$82,$78,$67,$FE,$CD,$97,$C8,$7F,$80,$DB,$5E,$0B,$FF,$04
	.dc.b	$7C,$D3,$EF,$3F,$B3,$B6,$73,$1F,$DA,$BE,$8F,$FD,$C3,$F6,$FF,$E0
	.dc.b	$DF,$BC,$FF,$1B,$57,$F7,$36,$BF,$E9,$46,$BF,$E5,$F6,$F1,$BC,$E7
	.dc.b	$CF,$FF,$86,$63,$0D,$8C,$D7,$F0,$0C,$9F,$78,$F9,$EF,$4F,$9F,$FB
	.dc.b	$6A,$FC,$49,$DF,$EC,$FC,$C6,$BD,$2B,$1F,$EC,$9D,$13,$FE,$C9,$6C
	.dc.b	$66,$3F,$D9,$D8,$F8,$3D,$CF,$8F,$91,$A7,$4A,$F0,$8F,$01,$F7,$7F
	.dc.b	$E3,$36,$C7,$F1,$06,$BE,$8B,$7D,$E3,$C7,$E1,$7F,$ED,$DB,$DD,$FF
	.dc.b	$88,$5A,$FF,$B0,$9A,$FF,$A3,$5E,$BD,$0E,$2F,$F4,$86,$3F,$D4,$B7
	.dc.b	$FC,$0F,$A9,$FD,$7B,$0F,$46,$3E,$7F,$A2,$91,$E7,$7F,$C9,$9E,$73
	.dc.b	$E4,$33,$3F,$96,$D1,$5E,$72,$47,$FB,$DE,$32,$FD,$BF,$BC,$7D,$4F
	.dc.b	$E4,$9B,$1D,$CF,$E6,$BF,$75,$FC,$4D,$BE,$47,$F9,$76,$31,$1F,$27
	.dc.b	$C4,$76,$66,$BF,$53,$3F,$EA,$FA,$DB,$FE,$49,$6B,$FE,$EA,$6B,$B8
	.dc.b	$FD,$2F,$DE,$7E,$DE,$3E,$B3,$78,$FE,$64,$6C,$F3,$FD,$4F,$94,$8F
	.dc.b	$E6,$71,$FE,$9C,$E8,$F2,$E0,$9F,$3C,$23,$0E,$A3,$31,$88,$FC,$47
	.dc.b	$C1,$BE,$F3,$F7,$6D,$8E,$C7,$E5,$27,$F3,$0F,$CC,$E9,$F9,$47,$F0
	.dc.b	$2F,$95,$FD,$56,$5E,$C9,$F1,$4F,$80,$8F,$A1,$AF,$E6,$7C,$56,$1F
	.dc.b	$C3,$1B,$1D,$CF,$B7,$1F,$64,$8F,$EF,$6D,$F6,$B8,$8F,$E4,$38,$F2
	.dc.b	$6D,$7C,$23,$EE,$FA,$CF,$8A,$FF,$0F,$F6,$7F,$BB,$69,$F7,$9E,$C8
	.dc.b	$C3,$63,$FD,$09,$D7,$FE,$7F,$3F,$D5,$11,$98,$E3,$E1,$E5,$FF,$03
	.dc.b	$C7,$E0,$6B,$9F,$CB,$F0,$3F,$D1,$1F,$03,$6B,$71,$5F,$E2,$B9,$79
	.dc.b	$E7,$FD,$6D,$F7,$2F,$BC,$A4,$AF,$F0,$E6,$9F,$EE,$3E,$3F,$A9,$5B
	.dc.b	$9F,$F0,$CB,$5F,$E2,$2D,$8F,$CA,$E1,$F1,$7E,$63,$A5,$7F,$4C,$AE
	.dc.b	$0F,$FA,$AD,$AF,$C9,$6A,$0B,$EE,$FD,$24,$8F,$3F,$91,$5F,$1C,$5C
	.dc.b	$3C,$78,$FF,$2E,$9F,$3F,$CE,$FE,$B5,$F0,$35,$F0,$60,$8C,$47,$89
	.dc.b	$BF,$31,$FE,$6C,$F1,$FF,$DB,$5C,$66,$92,$BE,$07,$C0,$3B,$7F,$82
	.dc.b	$E3,$FE,$41,$6C,$66,$77,$1F,$C2,$BD,$D7,$CB,$7F,$00,$D7,$F4,$D1
	.dc.b	$8E,$5F,$AF,$F7,$79,$7F,$25,$19,$C3,$FA,$2B,$6B,$8D,$EE,$7F,$DA
	.dc.b	$D7,$5F,$F1,$F4,$5A,$FF,$64,$5F,$F9,$61,$B3,$F8,$CD,$DF,$77,$BF
	.dc.b	$94,$97,$B7,$B7,$2F,$E9,$E7,$37,$E4,$9B,$9F,$D6,$9B,$7E,$AF,$93
	.dc.b	$F1,$1C,$B1,$FD,$BC,$78,$E3,$2D,$3F,$33,$DD,$8C,$7C,$DF,$1F,$ED
	.dc.b	$FB,$BE,$31,$8A,$FF,$39,$C6,$32,$FC,$3D,$6E,$FC,$0D,$AF,$F3,$CF
	.dc.b	$DD,$FF,$6E,$6F,$A1,$F2,$F9,$7F,$CA,$BE,$B7,$CB,$AD,$67,$FD,$D9
	.dc.b	$B9,$7F,$1F,$6E,$1F,$D9,$A3,$FC,$1C,$F4,$3B,$F9,$FF,$C9,$6B,$F8
	.dc.b	$75,$7F,$60,$9F,$E9,$ED,$8F,$F1,$CB,$5F,$E2,$81,$D5,$78,$6B,$D0
	.dc.b	$BE,$93,$D1,$12,$FC,$27,$F8,$8E,$74,$E2,$FD,$9F,$59,$F0,$7C,$17
	.dc.b	$A9,$F0,$63,$0D,$AF,$93,$1F,$1F,$4F,$F0,$C4,$60,$97,$D1,$FF,$19
	.dc.b	$46,$A3,$07,$CB,$F5,$7E,$87,$DD,$FF,$7C,$AF,$AB,$18,$F4,$FE,$93
	.dc.b	$23,$FC,$79,$4F,$BB,$C3,$CE,$38,$9A,$8C,$E9,$FD,$57,$EE,$7F,$C6
	.dc.b	$53,$FE,$45,$FB,$BF,$7B,$FE,$BC,$8C,$DF,$F8,$B6,$72,$FF,$8D,$A3
	.dc.b	$35,$CF,$D0,$FD,$87,$DD,$FE,$0A,$BF,$90,$56,$B0,$FE,$98,$D8,$CD
	.dc.b	$70,$E5,$FD,$C4,$E6,$4F,$F6,$4A,$FF,$2D,$74,$FC,$59,$F1,$79,$DF
	.dc.b	$97,$8C,$65,$ED,$8E,$67,$93,$D2,$B7,$ED,$8B,$7F,$E5,$0C,$FF,$C0
	.dc.b	$53,$93,$FC,$D1,$F3,$1E,$19,$FF,$27,$37,$0F,$95,$F1,$1F,$A7,$8C
	.dc.b	$57,$BA,$3D,$67,$F1,$3F,$5F,$37,$F1,$23,$18,$EA,$BF,$DD,$75,$AA
	.dc.b	$F8,$35,$FC,$77,$E0,$9F,$76,$F7,$B0,$DC,$3C,$FF,$1F,$89,$F9,$7E
	.dc.b	$54,$9E,$B3,$71,$FF,$37,$E4,$8F,$31,$FC,$AE,$7F,$CE,$5E,$46,$1F
	.dc.b	$F7,$FF,$DC,$7F,$B8,$1F,$F2,$2B,$6F,$FD,$5E,$D7,$F6,$ED,$7F,$C1
	.dc.b	$8E,$F7,$FF,$13,$36,$7F,$B3,$7D,$13,$EE,$BF,$D6,$31,$E0,$F9,$2F
	.dc.b	$E7,$95,$BC,$3F,$B6,$36,$F8,$7C,$5E,$AB,$F9,$BF,$C9,$EF,$CE,$3F
	.dc.b	$E4,$7C,$7F,$BA,$67,$FD,$FA,$34,$67,$2F,$F0,$7B,$8F,$80,$E7,$87
	.dc.b	$47,$AF,$A1,$DB,$E1,$6E,$3D,$0F,$F0,$06,$5E,$08,$CE,$3D,$1B,$97
	.dc.b	$85,$FE,$B5,$F7,$5F,$DC,$6E,$F8,$BF,$EB,$E7,$A5,$A7,$51,$87,$FB
	.dc.b	$C7,$C5,$F0,$8E,$2D,$75,$5C,$EB,$FD,$7F,$5C,$9B,$19,$C3,$FC,$EB
	.dc.b	$39,$6C,$66,$BE,$DF,$CB,$FD,$43,$FB,$97,$33,$EC,$BF,$67,$3A,$3F
	.dc.b	$13,$3B,$7D,$73,$BF,$1F,$19,$7D,$0F,$3F,$EC,$FD,$D7,$DA,$6D,$FC
	.dc.b	$2D,$FE,$DC,$62,$B8,$4E,$78,$D7,$FE,$03,$FB,$97,$A5,$FD,$46,$FF
	.dc.b	$C6,$91,$8B,$FF,$4F,$B6,$3F,$A7,$87,$0F,$F5,$27,$5B,$FF,$9F,$DB
	.dc.b	$3F,$E3,$FF,$B9,$7D,$CB,$E8,$7C,$58,$CA,$FB,$78,$F5,$8F,$F0,$D5
	.dc.b	$F0,$AD,$4F,$E4,$BC,$0F,$CD,$FB,$5F,$9C,$6E,$1E,$F6,$CE,$E3,$E1
	.dc.b	$4F,$F8,$8F,$AD,$F3,$14,$FA,$FC,$BF,$EE,$7A,$F1,$CE,$9B,$3B,$7E
	.dc.b	$BC,$EE,$FF,$DE,$F5,$B7,$FD,$F3,$F6,$CF,$98,$F9,$D0,$8C,$5F,$FB
	.dc.b	$31,$B1,$F3,$9A,$FF,$7E,$73,$C1,$FE,$AB,$6C,$FF,$9A,$F9,$BF,$1D
	.dc.b	$D5,$FE,$3F,$9C,$DF,$F9,$BD,$AF,$87,$0F,$7C,$ED,$FF,$23,$78,$EF
	.dc.b	$A4,$A7,$E6,$BE,$DB,$19,$7D,$69,$B8,$33,$FF,$5F,$46,$1B,$7C,$4E
	.dc.b	$75,$FB,$8F,$09,$F6,$AF,$1B,$FC,$B1,$18,$7F,$8B,$FB,$78,$9D,$57
	.dc.b	$16,$E5,$E1,$8F,$E4,$F1,$8C,$7F,$8D,$57,$1B,$FF,$31,$61,$FC,$F6
	.dc.b	$FF,$1E,$DC,$3F,$8E,$FA,$7A,$F2,$65,$69,$FE,$53,$18,$9D,$CF,$EB
	.dc.b	$63,$F9,$6F,$D2,$6A,$30,$49,$FE,$53,$19,$7D,$02,$DF,$53,$FE,$95
	.dc.b	$AF,$8C,$D9,$FF,$5F,$FD,$AF,$F3,$D8,$C6,$8A,$FE,$8F,$18,$6F,$C4
	.dc.b	$FF,$9D,$AB,$83,$FE,$83,$A7,$BD,$4F,$F3,$97,$DA,$DB,$FD,$9C,$E9
	.dc.b	$B3,$A6,$93,$1F,$59,$F8,$38,$D3,$89,$E4,$D8,$EC,$FF,$A3,$EA,$CF
	.dc.b	$AE,$3E,$27,$D7,$3C,$BF,$C2,$B7,$0F,$E4,$1E,$67,$F8,$FD,$FC,$64
	.dc.b	$66,$BE,$6E,$A8,$C6,$1D,$63,$18,$7F,$A6,$B0,$E5,$7C,$2F,$78,$7F
	.dc.b	$9B,$B8,$7F,$73,$3F,$90,$3F,$91,$B8,$24,$7F,$1A,$8F,$C9,$3F,$D8
	.dc.b	$3E,$D5,$F4,$FF,$EB,$39,$C9,$FE,$BC,$9E,$15,$FE,$AC,$AD,$C6,$0E
	.dc.b	$92,$F9,$6F,$F7,$57,$C1,$F4,$8F,$F7,$61,$FE,$D3,$9C,$E1,$FC,$E2
	.dc.b	$AC,$EB,$BB,$E7,$5F,$18,$E3,$3F,$EF,$5F,$B5,$C4,$73,$0E,$1F,$11
	.dc.b	$8B,$FF,$49,$DF,$D9,$9D,$57,$39,$CE,$16,$FF,$D3,$33,$5F,$E3,$51
	.dc.b	$51,$DA,$BC,$77,$D6,$FD,$93,$E1,$3D,$95,$C5,$E1,$1E,$07,$47,$B8
	.dc.b	$A7,$B6,$BC,$EE,$2B,$B3,$E6,$31,$5D,$6F,$CD,$3A,$9E,$95,$E6,$AD
	.dc.b	$47,$03,$B9,$E4,$3C,$E6,$2B,$9C,$70,$9E,$53,$91,$F2,$9F,$EB,$27
	.dc.b	$C8,$D7,$D8,$D7,$E4,$D7,$CA,$8D,$7C,$87,$F0,$0E,$1C,$67,$F8,$88
	.dc.b	$C5,$F6,$8C,$71,$FF,$3D,$8C,$1F,$52,$77,$7D,$27,$58,$FE,$2E,$5F
	.dc.b	$27,$F9,$48,$C3,$75,$EB,$97,$E0,$6B,$C3,$5A,$C7,$17,$B7,$C1,$87
	.dc.b	$6C,$38,$C7,$DC,$BE,$EF,$03,$A4,$ED,$E3,$86,$5E,$17,$D5,$57,$8A
	.dc.b	$FA,$E1,$E3,$9D,$9D,$2E,$38,$24,$7C,$1C,$3C,$11,$D1,$B7,$E6,$8D
	.dc.b	$DF,$62,$3F,$37,$0F,$24,$EE,$FA,$61,$E0,$CB,$C8,$4C,$3E,$1B,$F4
	.dc.b	$7E,$D3,$F8,$1C,$FF,$2B,$C7,$DB,$18,$D7,$F0,$0D,$87,$A2,$E0,$FE
	.dc.b	$E9,$1F,$E3,$4C,$D5,$B8,$E6,$65,$E5,$5C,$78,$75,$5C,$2B,$FC,$92
	.dc.b	$70,$3F,$C9,$6E,$5B,$5A,$E5,$FE,$71,$E9,$F9,$5B,$3F,$CF,$A3,$BF
	.dc.b	$2B,$EC,$B5,$CC,$9A,$4A,$FE,$8F,$3E,$9E,$0F,$E4,$F1,$FD,$61,$BB
	.dc.b	$FC,$76,$EF,$FE,$98,$F8,$E7,$D9,$71,$B8,$7F,$96,$2B,$6F,$CA,$9D
	.dc.b	$B9,$36,$77,$3E,$3B,$A7,$06,$C8,$7F,$81,$B8,$11,$BC,$3F,$AD,$7D
	.dc.b	$93,$CD,$F8,$CD,$BF,$F1,$73,$5F,$F0,$4F,$83,$A9,$91,$A7,$EC,$F7
	.dc.b	$A7,$ED,$FA,$AF,$85,$A4,$BF,$F4,$67,$C1,$57,$FD,$8B,$17,$2F,$F6
	.dc.b	$54,$FE,$1F,$EC,$BF,$C3,$B5,$FE,$7E,$6B,$FE,$08,$6C,$FF,$B4,$6B
	.dc.b	$87,$89,$2B,$83,$D7,$87,$FD,$CB,$F6,$5F,$E9,$A9,$FF,$26,$35,$FF
	.dc.b	$00,$B6,$FF,$DB,$9E,$57,$E2,$E3,$15,$CB,$D6,$C1,$F0,$27,$F3,$DF
	.dc.b	$4D,$FC,$B6,$FF,$54,$DA,$FE,$DE,$DC,$3F,$AB,$FC,$33,$E7,$78,$F8
	.dc.b	$32,$7F,$D3,$B5,$B5,$5B,$7F,$D4,$3F,$64,$F9,$AF,$D7,$36,$33,$1F
	.dc.b	$B1,$F1,$BF,$D3,$31,$82,$1F,$E3,$29,$E0,$FF,$8D,$23,$6E,$67,$F6
	.dc.b	$6A,$76,$FE,$91,$AF,$F8,$31,$AF,$F7,$4F,$B2,$7D,$97,$F0,$36,$C7
	.dc.b	$FA,$E9,$AF,$F9,$13,$EC,$9E,$86,$EA,$4E,$DF,$E1,$93,$A8,$FF,$6C
	.dc.b	$78,$3F,$92,$EF,$91,$82,$1F,$A5,$E9,$FF,$0D,$71,$7F,$92,$7B,$D1
	.dc.b	$8E,$3F,$F0,$5C,$63,$0F,$E3,$5F,$38,$FB,$2A,$4C,$7F,$DA,$7F,$2B
	.dc.b	$F5,$5F,$F2,$EC,$ED,$FF,$6B,$B5,$CC,$FF,$25,$9E,$19,$7F,$31,$AD
	.dc.b	$57,$F1,$06,$BF,$E3,$86,$BF,$E2,$06,$E1,$EA,$F2,$BF,$67,$EA,$4A
	.dc.b	$FF,$7E,$47,$ED,$8C,$3F,$0E,$30,$F9,$BE,$A3,$DE,$47,$FD,$DF,$F6
	.dc.b	$4F,$8C,$C4,$7B,$A6,$E5,$FD,$26,$B5,$3F,$EF,$6F,$B1,$FC,$5C,$FF
	.dc.b	$4D,$5F,$65,$FD,$15,$B3,$BF,$83,$C1,$99,$91,$92,$1F,$E8,$A9,$E2
	.dc.b	$D8,$E2,$FF,$91,$2B,$EF,$71,$3D,$7F,$E2,$FE,$EC,$AD,$72,$7E,$02
	.dc.b	$B5,$18,$3F,$AF,$B8,$7F,$15,$3C,$5F,$C6,$57,$F5,$CF,$B2,$FF,$7D
	.dc.b	$47,$1E,$1F,$C6,$BF,$C9,$8D,$7C,$04,$75,$87,$F4,$49,$DC,$71,$AE
	.dc.b	$6F,$FA,$A3,$EC,$9F,$65,$FD,$B5,$B1,$E7,$0E,$1F,$F5,$7F,$D5,$FF
	.dc.b	$91,$1D,$46,$75,$FF,$00,$37,$5F,$F4,$2B,$67,$FC,$DB,$5C,$5F,$15
	.dc.b	$FF,$04,$9E,$D1,$FD,$99,$AF,$28,$FE,$F7,$F2,$0F,$B2,$FE,$C2,$D9
	.dc.b	$FF,$06,$35,$FA,$1F,$31,$FE,$D0,$FA,$EF,$F0,$57,$D9,$7F,$88,$F4
	.dc.b	$FF,$0B,$37,$4F,$F4,$DB,$6D,$87,$37,$FD,$D3,$2F,$21,$F1,$3F,$E1
	.dc.b	$B8,$C3,$63,$33,$F8,$1A,$7D,$93,$E5,$3C,$FE,$67,$EE,$31,$F0,$FD
	.dc.b	$67,$98,$CF,$9D,$71,$FA,$9A,$27,$11,$FE,$FC,$C9,$87,$38,$F4,$E5
	.dc.b	$FD,$B5,$F1,$35,$E0,$DD,$7E,$1D,$FF,$8B,$FB,$9F,$41,$49,$7E,$BF
	.dc.b	$B2,$78,$B8,$8F,$F7,$9D,$70,$7F,$14,$D7,$CD,$3C,$1C,$D7,$D8,$F3
	.dc.b	$BF,$AE,$69,$FD,$4F,$4E,$99,$FF,$92,$F1,$7D,$37,$94,$98,$FE,$5F
	.dc.b	$E4,$FF,$DF,$0F,$82,$32,$F9,$1F,$F2,$54,$6A,$30,$7D,$5F,$B2,$7C
	.dc.b	$D7,$E2,$9B,$1C,$CF,$D2,$7D,$8B,$E8,$3F,$CA,$8D,$8C,$C7,$E4,$BC
	.dc.b	$37,$56,$A2,$4E,$8F,$C2,$4E,$9C,$1F,$6F,$EC,$BF,$4D,$AB,$A3,$FD
	.dc.b	$AB,$C0,$9C,$C7,$FB,$CB,$2F,$15,$AF,$FC,$E5,$0F,$9F,$F3,$92,$7C
	.dc.b	$3E,$75,$78,$5F,$39,$C5,$F8,$67,$F4,$79,$2F,$37,$E4,$F7,$DF,$49
	.dc.b	$FB,$26,$DF,$C8,$C7,$E3,$FC,$0C,$8F,$96,$FE,$E3,$E4,$7F,$31,$0D
	.dc.b	$3F,$29,$8F,$B6,$B8,$4F,$E7,$EB,$75,$F9,$B8,$FF,$36,$B5,$E0,$7F
	.dc.b	$A0,$9C,$3F,$EF,$E5,$87,$DA,$24,$7E,$DE,$5F,$57,$88,$F8,$A7,$19
	.dc.b	$C6,$BF,$17,$75,$F0,$AF,$E7,$1F,$0F,$FE,$24,$AF,$F5,$27,$D9,$7B
	.dc.b	$8F,$DB,$79,$7F,$E6,$EE,$BF,$F3,$A9,$FE,$9F,$8D,$78,$58,$C3,$F9
	.dc.b	$A4,$7E,$01,$07,$0D,$F9,$3F,$F3,$FF,$BF,$CD,$91,$C9,$CB,$FC,$FF
	.dc.b	$EC,$1F,$65,$FE,$AF,$6C,$66,$3F,$D1,$3F,$2B,$96,$46,$4F,$7C,$61
	.dc.b	$F8,$E2,$79,$B8,$6E,$9F,$E2,$9D,$7F,$98,$4E,$FE,$CB,$C4,$47,$E5
	.dc.b	$F8,$D5,$E5,$16,$B9,$4F,$A7,$E6,$7F,$A7,$57,$C6,$3C,$4E,$1F,$93
	.dc.b	$82,$31,$1A,$CF,$FD,$93,$F2,$BF,$E6,$5E,$3B,$C3,$FD,$12,$BF,$D9
	.dc.b	$6D,$7F,$D7,$B3,$B7,$FD,$7E,$7C,$1E,$BF,$F0,$03,$E3,$C7,$F5,$5D
	.dc.b	$9F,$ED,$66,$C7,$06,$BE,$9C,$5F,$86,$3E,$07,$F6,$9B,$7E,$D6,$73
	.dc.b	$18,$EC,$FF,$65,$7C,$B5,$C1,$5A,$9F,$F0,$A3,$63,$F3,$AD,$BF,$A9
	.dc.b	$5C,$63,$15,$F9,$07,$C7,$F0,$1F,$EF,$FB,$C6,$28,$C4,$78,$23,$15
	.dc.b	$F1,$2B,$8B,$63,$33,$E2,$6E,$1E,$DF,$77,$F3,$F5,$E8,$9C,$BE,$0C
	.dc.b	$3C,$D1,$8B,$F7,$D7,$07,$A4,$FC,$3F,$96,$FE,$D5,$F4,$37,$90,$FC
	.dc.b	$7E,$CF,$F7,$0F,$C7,$66,$26,$7F,$E3,$BB,$E2,$FE,$FA,$73,$8F,$FB
	.dc.b	$A7,$DB,$7D,$AF,$35,$F7,$16,$30,$DB,$FF,$1B,$36,$B9,$3E,$5F,$89
	.dc.b	$F9,$78,$FF,$80,$5B,$A7,$FA,$92,$FF,$1F,$39,$6C,$66,$FF,$8C,$CE
	.dc.b	$5F,$F0,$47,$E1,$9F,$EF,$1A,$F6,$56,$F8,$7F,$A0,$5B,$F6,$0F,$EC
	.dc.b	$15,$FE,$4B,$F8,$9F,$95,$18,$C7,$E0,$46,$3C,$DF,$3C,$E6,$7F,$65
	.dc.b	$7C,$E7,$24,$BF,$99,$5C,$F0,$FE,$59,$93,$E4,$3A,$56,$BE,$27,$E6
	.dc.b	$78,$FE,$26,$C6,$4F,$87,$87,$FA,$3D,$B1,$E4,$23,$BC,$FF,$C5,$5F
	.dc.b	$27,$FE,$02,$8C,$65,$F0,$AF,$FC,$71,$1F,$E6,$BF,$B7,$F8,$3E,$F7
	.dc.b	$DD,$6C,$78,$23,$E0,$91,$C1,$EE,$7F,$69,$E1,$B0,$7E,$07,$CB,$2F
	.dc.b	$AF,$8F,$D5,$7F,$19,$F7,$1F,$43,$D1,$FC,$63,$72,$F7,$C7,$D0,$23
	.dc.b	$83,$E6,$BF,$9D,$F6,$3F,$BD,$46,$0F,$A3,$F1,$FF,$97,$F4,$BE,$EB
	.dc.b	$63,$27,$CC,$7E,$E9,$1C,$1A,$C3,$F1,$FF,$7C,$E2,$E7,$3F,$E9,$3F
	.dc.b	$96,$FD,$21,$F8,$AA,$B1,$F8,$27,$F1,$24,$7E,$DB,$FC,$33,$C8,$7E
	.dc.b	$2A,$F7,$B9,$57,$FA,$BE,$73,$18,$AF,$DA,$7C,$EF,$EC,$CF,$DA,$7D
	.dc.b	$25,$BF,$DE,$79,$DE,$77,$F5,$6D,$AF,$E9,$43,$B5,$19,$8F,$CE,$B5
	.dc.b	$FE,$A1,$F6,$1F,$91,$7F,$A1,$B5,$E6,$D7,$7D,$7F,$BE,$9F,$F0,$8D
	.dc.b	$FF,$AE,$7D,$0F,$E3,$11,$8A,$FE,$1D,$18,$E2,$E0,$4A,$F4,$B6,$7F
	.dc.b	$80,$D7,$1E,$CF,$F1,$E7,$90,$F9,$77,$DC,$73,$C7,$FD,$A1,$CF,$FC
	.dc.b	$0D,$FF,$07,$D3,$C8,$7B,$1F,$8B,$F1,$FF,$8F,$36,$BF,$DA,$0D,$D3
	.dc.b	$F9,$23,$FC,$8F,$D7,$93,$DB,$E2,$ED,$7F,$29,$9B,$AF,$9B,$3F,$F5
	.dc.b	$7D,$FF,$B6,$E7,$F9,$67,$93,$23,$FD,$DD,$F7,$DF,$ED,$C7,$18,$7F
	.dc.b	$33,$9B,$87,$F9,$6E,$B5,$5F,$EE,$59,$CB,$EE,$D3,$FD,$5C,$F4,$F8
	.dc.b	$FF,$F3,$93,$B3,$FD,$D9,$EA,$7F,$3D,$3F,$58,$D3,$E4,$97,$1F,$F2
	.dc.b	$4E,$ED,$39,$C7,$F4,$7A,$FE,$8D,$E7,$CB,$FF,$80,$1F,$AE,$7F,$23
	.dc.b	$7C,$91,$9A,$FE,$9C,$FF,$57,$9D,$43,$0F,$AD,$3F,$E0,$18,$C6,$F7
	.dc.b	$C0,$FE,$3D,$1F,$E3,$6A,$F9,$AF,$F8,$E2,$33,$BF,$FA,$06,$38,$E3
	.dc.b	$FE,$1C,$69,$FE,$1E,$AE,$2F,$E9,$9A,$FF,$8C,$5E,$B3,$CD,$B3,$FE
	.dc.b	$35,$EE,$7D,$2F,$9E,$73,$7E,$BF,$D6,$BF,$C6,$D7,$FE,$60,$C3,$F3
	.dc.b	$31,$89,$FE,$BB,$FB,$6F,$A2,$BF,$2F,$39,$AD,$FB,$9F,$DF,$63,$13
	.dc.b	$FE,$6A,$EF,$5F,$1C,$3C,$17,$FD,$E7,$C2,$97,$E1,$3F,$47,$EE,$BE
	.dc.b	$AF,$37,$F7,$A3,$13,$B8,$FA,$9C,$DF,$C5,$E3,$1A,$AB,$75,$FE,$09
	.dc.b	$C3,$EC,$7C,$BF,$F8,$E2,$F9,$4B,$F2,$3F,$F0,$BC,$62,$75,$3F,$E7
	.dc.b	$09,$CB,$FE,$72,$8C,$BF,$C6,$9A,$FF,$99,$5A,$FF,$A3,$1A,$9E,$BB
	.dc.b	$FE,$0D,$B5,$FE,$6D,$FD,$15,$EB,$FA,$3D,$EF,$F1,$07,$D8,$CA,$F4
	.dc.b	$FD,$B7,$DE,$7E,$0E,$E9,$FC,$B4,$E7,$3F,$F1,$54,$63,$97,$94,$7F
	.dc.b	$8B,$7E,$47,$FC,$58,$7F,$8D,$F8,$3F,$8D,$63,$D2,$31,$E2,$EA,$3E
	.dc.b	$7C,$EA,$B9,$47,$F8,$EF,$F6,$EF,$E1,$11,$FC,$FB,$9D,$18,$8F,$F0
	.dc.b	$CB,$5F,$C4,$35,$FE,$88,$F2,$E1,$FF,$14,$B6,$BF,$94,$FC,$37,$E0
	.dc.b	$6B,$FC,$67,$39,$7F,$A6,$7C,$13,$D5,$FF,$29,$56,$F0,$FF,$19,$CE
	.dc.b	$70,$FF,$66,$F3,$7F,$96,$E3,$18,$7A,$5A,$FF,$8C,$1A,$FE,$7D,$AF
	.dc.b	$A3,$E0,$7F,$C9,$B9,$7E,$22,$3D,$F1,$82,$5F,$C7,$7C,$BC,$1F,$E6
	.dc.b	$BA,$FC,$7E,$3F,$E9,$3F,$C3,$FF,$C7,$17,$FE,$7F,$6E,$1F,$DA,$B9
	.dc.b	$11,$88,$FF,$4F,$35,$FF,$47,$3C,$2B,$A7,$BD,$F1,$38,$BF,$D2,$93
	.dc.b	$9C,$37,$38,$34,$FF,$A7,$A3,$57,$FE,$A2,$6E,$1E,$2A,$DC,$FF,$A8
	.dc.b	$A7,$23,$E7,$FF,$D3,$75,$E8,$6C,$66,$FD,$D1,$83,$CA,$7D,$A9,$B1
	.dc.b	$FB,$AA,$FF,$7C,$FC,$5F,$FB,$EB,$EC,$3F,$D7,$91,$FC,$8E,$71,$F1
	.dc.b	$1F,$EC,$AA,$B3,$91,$5C,$1F,$F6,$F4,$16,$79,$5E,$1B,$0D,$4E,$65
	.dc.b	$58,$C3,$FC,$3F,$7B,$9E,$71,$E3,$7C,$2E,$AF,$CB,$1E,$23,$27,$9F
	.dc.b	$0E,$D7,$A8,$F1,$CD,$7C,$27,$03,$D5,$7E,$1A,$E3,$1E,$23,$E1,$61
	.dc.b	$83,$DD,$3E,$77,$73,$8A,$ED,$7E,$1A,$E9,$87,$AE,$B5,$87,$84,$FA
	.dc.b	$A5,$9D,$BF,$E4,$2A,$E5,$7D,$23,$4D,$C3,$99,$E5,$C3,$E3,$4F,$A4
	.dc.b	$F8,$D8,$78,$0F,$3C,$69,$E1,$1E,$E8,$FA,$13,$D1,$A7,$47,$71,$F2
	.dc.b	$2B,$ED,$3E,$73,$FE,$00,$C7,$D9,$C7,$F0,$B1,$96,$BF,$65,$BE,$8F
	.dc.b	$F8,$FB,$6F,$F8,$01,$5F,$8D,$AF,$F3,$0F,$9E,$9C,$C7,$E4,$6B,$83
	.dc.b	$E0,$31,$3B,$70,$72,$9F,$3B,$FE,$6C,$F3,$FB,$5F,$E7,$6D,$18,$CB
	.dc.b	$C5,$1D,$A7,$58,$75,$78,$56,$EF,$C8,$D7,$C0,$7D,$3A,$F3,$E7,$C2
	.dc.b	$3F,$37,$1E,$06,$DF,$A6,$7B,$E3,$E8,$AD,$57,$3B,$ED,$1E,$5C,$7C
	.dc.b	$F8,$71,$7D,$33,$F2,$F0,$F0,$5F,$E8,$6F,$C7,$5C,$30,$F1,$DE,$E3
	.dc.b	$8D,$79,$26,$3C,$67,$4F,$C0,$33,$7E,$87,$CC,$7A,$63,$CF,$87,$C9
	.dc.b	$BF,$04,$F3,$3C,$27,$27,$35,$E8,$23,$83,$8E,$3E,$C2,$38,$35,$3E
	.dc.b	$53,$B1,$F0,$6B,$E3,$2C,$9F,$86,$3E,$87,$F2,$59,$FF,$15,$9C,$C6
	.dc.b	$35,$FE,$45,$18,$BF,$F1,$7F,$E7,$EF,$99,$D4,$75,$C7,$A3,$FE,$BE
	.dc.b	$31,$87,$F9,$7E,$F9,$47,$F5,$99,$DB,$7E,$1B,$9C,$F9,$9B,$27,$BF
	.dc.b	$8E,$C5,$F3,$3A,$99,$87,$F9,$1F,$0E,$4F,$A2,$72,$FE,$BF,$3F,$6E
	.dc.b	$3D,$63,$83,$5F,$F4,$7E,$1D,$9C,$E1,$EB,$9D,$47,$EE,$F0,$EE,$FA
	.dc.b	$CF,$85,$DF,$FE,$22,$EC,$F8,$ED,$8F,$F1,$44,$7B,$30,$7C,$CE,$BA
	.dc.b	$D3,$FE,$87,$BE,$17,$FE,$2D,$9C,$BD,$9A,$FF,$86,$7B,$38,$57,$F9
	.dc.b	$16,$3C,$6D,$74,$62,$3C,$5F,$2D,$23,$FC,$E1,$7E,$29,$F2,$7B,$12
	.dc.b	$7F,$C8,$0E,$DF,$F1,$9B,$E7,$C0,$D2,$47,$FA,$7F,$C3,$7C,$3F,$DC
	.dc.b	$F2,$3F,$2D,$3F,$E5,$CD,$BE,$FD,$6A,$30,$43,$FD,$87,$3C,$1F,$F4
	.dc.b	$45,$7E,$51,$7E,$BD,$FD,$F7,$9F,$FD,$C5,$CF,$F8,$AF,$5F,$F3,$55
	.dc.b	$FE,$84,$AD,$72,$FF,$BD,$A7,$33,$FE,$D0,$6D,$F2,$8E,$1D,$5D,$FE
	.dc.b	$D7,$FD,$63,$1F,$E8,$E6,$BF,$8F,$3F,$C1,$8E,$27,$FD,$89,$FA,$BB
	.dc.b	$E1,$79,$79,$9F,$EC,$5B,$5F,$27,$FA,$41,$88,$D4,$66,$3F,$A6,$3D
	.dc.b	$72,$FE,$9D,$F8,$1F,$F2,$CC,$7F,$DA,$4D,$8D,$DF,$CE,$F2,$3E,$27
	.dc.b	$6C,$C7,$FC,$FF,$15,$F8,$51,$70,$FF,$0E,$77,$7F,$AC,$6B,$EB,$FC
	.dc.b	$14,$BF,$F4,$3B,$FA,$0F,$C6,$3F,$55,$1F,$EB,$D8,$D3,$4F,$C7,$C6
	.dc.b	$AF,$FD,$D1,$FA,$BB,$EA,$7E,$E3,$6B,$FD,$76,$D7,$E7,$07,$FD,$88
	.dc.b	$EA,$3E,$64,$6E,$3E,$5B,$5F,$EE,$8D,$7F,$C8,$4F,$59,$FF,$7D,$7E
	.dc.b	$DE,$F9,$AF,$C7,$36,$BB,$1F,$ED,$4E,$4F,$DD,$B7,$C1,$FC,$81,$FC
	.dc.b	$6B,$E0,$1A,$62,$74,$43,$F1,$1B,$27,$71,$FE,$E0,$7E,$F2,$85,$BC
	.dc.b	$1F,$CF,$1D,$F8,$6D,$79,$62,$BD,$AF,$9D,$F2,$E3,$FD,$31,$27,$FD
	.dc.b	$E5,$F0,$8F,$CB,$DF,$2F,$99,$FD,$35,$5E,$0F,$F3,$05,$EE,$B1,$CD
	.dc.b	$FA,$DA,$B6,$AD,$3F,$27,$87,$F6,$C4,$7E,$07,$B1,$5F,$DA,$45,$C7
	.dc.b	$F2,$7A,$2B,$E7,$7C,$A7,$EF,$FE,$4E,$61,$F2,$63,$2F,$D7,$8C,$BF
	.dc.b	$0E,$3F,$9B,$FD,$A7,$F5,$6C,$3F,$CF,$5E,$42,$31,$AF,$FA,$1E,$31
	.dc.b	$E6,$FF,$56,$F8,$E7,$C3,$F2,$8F,$E7,$57,$FD,$83,$2F,$F2,$15,$F0
	.dc.b	$7F,$A3,$C6,$5E,$8F,$67,$EF,$73,$3E,$E3,$F8,$D1,$D6,$3F,$D3,$19
	.dc.b	$FF,$A3,$D7,$D6,$3E,$2F,$F8,$63,$63,$FB,$B1,$A8,$C9,$89,$FA,$71
	.dc.b	$FC,$41,$AF,$F8,$23,$9B,$F0,$2F,$F7,$F8,$CB,$FE,$A9,$E4,$7E,$D1
	.dc.b	$FD,$0A,$3F,$12,$D7,$F9,$24,$6B,$CD,$FF,$05,$BC,$3E,$7F,$F9,$D7
	.dc.b	$D0,$BE,$5F,$E9,$FA,$D3,$0F,$7F,$2F,$F9,$AE,$B5,$1F,$D6,$38,$7F
	.dc.b	$BA,$BF,$2A,$BE,$89,$1F,$F0,$B7,$6F,$EA,$3D,$0F,$E8,$84,$8E,$53
	.dc.b	$F8,$2F,$95,$91,$D5,$FE,$39,$9E,$23,$F9,$AF,$2F,$E1,$67,$EF,$92
	.dc.b	$3E,$EB,$FD,$47,$E9,$64,$73,$3D,$1F,$23,$F9,$02,$3F,$BF,$FB,$8A
	.dc.b	$F6,$D1,$3F,$8C,$24,$7E,$29,$FE,$BD,$CB,$23,$A9,$F1,$7D,$5F,$CE
	.dc.b	$91,$F9,$63,$E7,$1F,$2F,$DA,$38,$1F,$96,$9F,$CD,$91,$C1,$FC,$87
	.dc.b	$E5,$BF,$B7,$1D,$5F,$EC,$5A,$BF,$34,$FE,$2D,$1F,$E0,$E6,$BA,$6F
	.dc.b	$4F,$FB,$A7,$1F,$1F,$C0,$7F,$8A,$A7,$75,$FA,$A6,$BF,$B3,$6A,$BF
	.dc.b	$07,$93,$FE,$2B,$9E,$BC,$FF,$DF,$FE,$AB,$FC,$08,$75,$9E,$DD,$F3
	.dc.b	$0F,$33,$FE,$60,$C9,$F9,$D3,$E4,$F5,$5F,$82,$CB,$F6,$CF,$F9,$09
	.dc.b	$AE,$6F,$C4,$7F,$84,$FE,$2E,$47,$63,$E2,$73,$7B,$C8,$FD,$6C,$7C
	.dc.b	$3D,$98,$8F,$04,$FB,$48,$FE,$7B,$C3,$FF,$22,$F0,$FF,$78,$7F,$AE
	.dc.b	$6F,$FE,$1F,$F4,$FF,$C4,$9F,$03,$2F,$FA,$D4,$7F,$89,$CC,$3F,$57
	.dc.b	$87,$E9,$11,$FF,$42,$F9,$D5,$D1,$F7,$63,$FC,$BD,$1F,$E7,$EF,$4D
	.dc.b	$F4,$54,$95,$FE,$79,$F6,$FF,$88,$65,$FE,$92,$AD,$CF,$F9,$F5,$AF
	.dc.b	$F8,$B7,$A5,$18,$D7,$FD,$39,$7C,$AB,$FD,$45,$3C,$1F,$F3,$C3,$5E
	.dc.b	$D8,$7F,$A5,$E3,$11,$C3,$E4,$7D,$04,$7F,$D4,$7E,$95,$7C,$52,$9C
	.dc.b	$61,$FE,$26,$F8,$4F,$F0,$35,$7F,$85,$BE,$95,$F2,$DF,$E5,$98,$12
	.dc.b	$33,$A7,$F3,$66,$C7,$FA,$B5,$AF,$F9,$13,$E2,$FF,$6F,$CD,$FC,$F1
	.dc.b	$B1,$FE,$B3,$6B,$BF,$03,$F7,$CF,$3F,$19,$F1,$BD,$DB,$19,$8C,$1F
	.dc.b	$EB,$DC,$7E,$7F,$90,$E3,$FA,$8D,$AF,$A9,$8B,$C7,$C6,$FF,$EC,$FE
	.dc.b	$1F,$F5,$E6,$1C,$09,$7C,$2B,$71,$F8,$1C,$2C,$FF,$B6,$FD,$CF,$C4
	.dc.b	$CF,$8D,$B1,$FE,$66,$CB,$FC,$85,$C9,$FD,$8E,$3F,$DB,$73,$A7,$FD
	.dc.b	$70,$F0,$CB,$FD,$C4,$DB,$FF,$70,$E6,$F9,$3F,$0E,$3D,$27,$FD,$9D
	.dc.b	$5B,$C3,$C9,$8E,$B0,$E1,$5A,$8F,$4F,$89,$FC,$87,$AA,$F8,$3F,$E5
	.dc.b	$DA,$FF,$57,$FE,$E0,$FA,$1C,$D4,$9F,$E4,$CF,$29,$D4,$60,$FF,$22
	.dc.b	$78,$9F,$1B,$CE,$FF,$21,$F4,$DF,$6F,$F8,$B5,$7F,$95,$63,$15,$BF
	.dc.b	$0B,$F2,$73,$C8,$D1,$FD,$13,$77,$E1,$4F,$AA,$F9,$91,$F9,$88,$C4
	.dc.b	$FF,$83,$1B,$1F,$E9,$83,$8C,$F0,$6C,$F5,$9D,$F2,$FF,$86,$7F,$1B
	.dc.b	$FF,$62,$CE,$6F,$F0,$37,$E2,$F8,$1F,$C1,$F7,$3E,$99,$F0,$BA,$CF
	.dc.b	$FC,$DB,$8F,$F2,$79,$CB,$DE,$FC,$53,$AC,$3C,$CD,$7F,$D5,$ED,$7F
	.dc.b	$6A,$78,$67,$91,$F1,$72,$FF,$0B,$38,$8F,$F4,$9C,$BC,$AF,$3F,$C6
	.dc.b	$7F,$BF,$AB,$E0,$46,$3E,$17,$D0,$F9,$09,$1F,$2D,$B7,$D1,$F8,$7F
	.dc.b	$85,$FF,$9F,$E7,$99,$F2,$EF,$BF,$C1,$FC,$36,$C6,$4F,$AC,$FC,$22
	.dc.b	$38,$35,$7F,$EA,$CA,$FF,$28,$46,$38,$3F,$C9,$DE,$E7,$F8,$1A,$3F
	.dc.b	$D4,$1E,$2F,$D6,$8C,$5F,$FB,$A6,$31,$E3,$7E,$06,$73,$3F,$ED,$C6
	.dc.b	$BF,$39,$AF,$F2,$47,$9D,$F8,$27,$F5,$7F,$85,$7F,$0A,$AF,$F6,$0D
	.dc.b	$6B,$0E,$5F,$27,$FB,$0F,$A7,$7D,$17,$96,$30,$D9,$FF,$31,$D2,$7C
	.dc.b	$C4,$8C,$BB,$BF,$F0,$DE,$0F,$77,$19,$FF,$21,$F2,$DF,$4E,$F9,$CF
	.dc.b	$E3,$31,$87,$F3,$7F,$99,$BD,$7C,$D6,$F4,$FD,$84,$E5,$FF,$3B,$B7
	.dc.b	$3F,$D6,$35,$FC,$A8,$70,$FF,$A5,$73,$FF,$23,$CF,$4B,$F9,$3E,$FF
	.dc.b	$89,$FE,$2B,$E5,$7C,$69,$FF,$21,$F9,$28,$C6,$0D,$7F,$9D,$FE,$2B
	.dc.b	$F9,$B8,$FF,$1E,$CE,$A7,$FD,$25,$F5,$3F,$A9,$9F,$EB,$AD,$4D,$35
	.dc.b	$6C,$DF,$33,$F7,$8F,$E8,$3F,$6C,$FC,$C7,$F2,$78,$7D,$29,$FE,$0B
	.dc.b	$E5,$3A,$BF,$91,$7A,$4F,$83,$CC,$61,$BC,$1F,$EA,$2A,$E0,$D9,$FE
	.dc.b	$2D,$18,$C3,$FD,$0F,$1C,$1F,$F2,$DF,$17,$C1,$DF,$FD,$61,$5B,$7F
	.dc.b	$DA,$7F,$41,$FC,$B7,$AB,$F9,$93,$71,$FF,$7D,$37,$0F,$F7,$93,$6B
	.dc.b	$FD,$B7,$1F,$CD,$7E,$A7,$F9,$D3,$6B,$FC,$EF,$83,$CE,$FF,$04,$C6
	.dc.b	$3C,$C5,$F0,$8C,$3F,$D1,$7F,$61,$FE,$B1,$F1,$7F,$EC,$59,$C7,$AB
	.dc.b	$FD,$BA,$31,$1D,$A7,$7F,$11,$FD,$FE,$31,$5B,$8C,$67,$F8,$59,$FF
	.dc.b	$0D,$E6,$BE,$31,$D8,$FF,$47,$E1,$DA,$BF,$D0,$BF,$06,$F9,$4F,$EA
	.dc.b	$95,$FD,$A2,$30,$FB,$4C,$44,$7C,$DF,$01,$6B,$C4,$31,$F8,$DE,$9F
	.dc.b	$F6,$3F,$AE,$FE,$C8,$DB,$F9,$7C,$1F,$D4,$9F,$4A,$9D,$C6,$9F,$EA
	.dc.b	$7E,$27,$F7,$EC,$3B,$9F,$E2,$AF,$96,$CF,$4E,$9B,$7F,$EF,$0A,$E1
	.dc.b	$5F,$B8,$6B,$FD,$39,$AF,$F9,$11,$ED,$C9,$FE,$FD,$6C,$FF,$54,$FA
	.dc.b	$77,$C9,$F6,$BF,$83,$AB,$F4,$0F,$F0,$B7,$82,$FA,$BF,$CC,$63,$F5
	.dc.b	$A7,$31,$8B,$F9,$11,$8F,$4F,$FC,$18,$DA,$FE,$0B,$F2,$CF,$A0,$FF
	.dc.b	$08,$7C,$9F,$F8,$3B,$0F,$F5,$FC,$E6,$BF,$2F,$39,$DF,$FC,$8D,$9A
	.dc.b	$B5,$7C,$30,$EF,$CB,$FE,$28,$9C,$DF,$F9,$36,$9B,$FF,$1C,$7E,$8B
	.dc.b	$FE,$71,$CB,$FC,$30,$E6,$78,$39,$9D,$F8,$1F,$E1,$38,$E5,$1F,$CA
	.dc.b	$9A,$FC,$D8,$D1,$FE,$12,$8E,$95,$FE,$04,$FA,$1F,$2B,$B3,$FC,$51
	.dc.b	$F1,$7F,$E4,$5B,$7C,$DB,$97,$FC,$37,$39,$8C,$57,$0E,$FC,$9A,$D6
	.dc.b	$C7,$D0,$F1,$7E,$7F,$47,$C5,$6C,$64,$99,$7E,$0B,$8D,$F3,$FB,$9F
	.dc.b	$E7,$BD,$CF,$A9,$FF,$1D,$F2,$A3,$13,$FE,$39,$8C,$4E,$A7,$38,$7F
	.dc.b	$50,$9C,$D7,$4A,$FF,$2F,$FE,$84,$F9,$EF,$DD,$46,$19,$3F,$C6,$BE
	.dc.b	$43,$F8,$96,$1F,$B8,$C1,$87,$1A,$FE,$A6,$D9,$D3,$79,$BF,$D5,$DE
	.dc.b	$2F,$49,$FE,$17,$ED,$79,$C9,$7E,$BD,$3F,$C1,$B8,$FB,$E7,$F8,$71
	.dc.b	$E8,$F8,$9B,$5F,$EB,$4C,$7F,$D3,$9F,$18,$FC,$13,$F6,$AD,$E3,$FF
	.dc.b	$31,$B6,$BF,$D4,$D3,$A3,$F6,$F9,$A7,$4F,$FB,$67,$E6,$BF,$D4,$D5
	.dc.b	$DC,$FF,$61,$CE,$B0,$F0,$DF,$28,$C4,$EA,$B8,$DF,$FB,$8E,$B4,$4C
	.dc.b	$3F,$D3,$33,$AC,$39,$C7,$FA,$72,$5F,$47,$EA,$6E,$1E,$A8,$C1,$1F
	.dc.b	$36,$1E,$B9,$DB,$EB,$F8,$3E,$95,$FE,$78,$8F,$F6,$EF,$D4,$FF,$B7
	.dc.b	$BE,$A3,$BC,$7F,$1F,$C3,$15,$FE,$87,$AB,$39,$15,$C1,$FF,$39,$C7
	.dc.b	$FB,$A3,$E7,$71,$1F,$ED,$AB,$C5,$F3,$84,$EB,$C9,$FF,$6B,$3C,$21
	.dc.b	$7E,$07,$F9,$85,$D7,$D2,$78,$AF,$F9,$C4,$66,$38,$99,$3D,$18,$75
	.dc.b	$AE,$58,$7F,$B9,$63,$13,$E4,$BE,$6D,$3F,$C7,$75,$EA,$7E,$16,$1C
	.dc.b	$2B,$D5,$87,$86,$BA,$B6,$BC,$47,$A8,$A7,$AA,$78,$B8,$3D,$2E,$0E
	.dc.b	$D7,$E4,$7C,$F7,$C0,$B1,$E8,$77,$3F,$02,$3E,$84,$F3,$7E,$94,$F1
	.dc.b	$69,$C4,$B3,$EE,$6B,$E9,$8F,$8F,$B7,$DA,$8F,$8A,$7E,$AE,$3E,$F4
	.dc.b	$66,$3F,$0A,$64,$1F,$E5,$BF,$81,$FF,$2D,$39,$31,$B7,$EC,$2B,$FD
	.dc.b	$7F,$ED,$FF,$A0,$E3,$13,$CA,$30,$F1,$6C,$FE,$55,$7D,$4F,$FB,$22
	.dc.b	$30,$DC,$7A,$E1,$E2,$9E,$0D,$8F,$0C,$EE,$FC,$0D,$7C,$E7,$E1,$CB
	.dc.b	$97,$67,$F3,$58,$78,$9B,$7F,$02,$7B,$E3,$F8,$D9,$FF,$3F,$5F,$5B
	.dc.b	$E1,$1D,$5F,$6E,$1E,$8A,$EB,$5F,$EE,$4B,$F0,$5F,$E4,$2F,$E2,$C7
	.dc.b	$68,$C4,$EE,$FD,$B1,$E4,$B8,$E8,$EA,$62,$B9,$63,$D8,$95,$E3,$9F
	.dc.b	$36,$3E,$D8,$EB,$1C,$CF,$24,$65,$FB,$4D,$7E,$09,$E0,$37,$7E,$4A
	.dc.b	$EF,$1F,$48,$F9,$18,$FC,$47,$9C,$F1,$C3,$E1,$9F,$04,$FE,$3F,$97
	.dc.b	$F4,$9F,$B2,$FC,$9D,$7F,$37,$6D,$7F,$3E,$6E,$9F,$CF,$73,$B5,$FC
	.dc.b	$BE,$7F,$D7,$EF,$F4,$6E,$04,$62,$34,$7F,$52,$71,$83,$EC,$1E,$0F
	.dc.b	$48,$C3,$63,$35,$FC,$FA,$5C,$8E,$2D,$5C,$AD,$5F,$07,$FA,$E4,$E5
	.dc.b	$FF,$39,$DF,$67,$75,$E1,$BE,$91,$83,$DB,$3B,$7F,$DB,$AD,$7E,$4B
	.dc.b	$5F,$DE,$F3,$FF,$86,$78,$BE,$3B,$63,$FC,$3F,$18,$EA,$7C,$46,$BA
	.dc.b	$BF,$3F,$5F,$52,$F8,$3F,$E1,$D8,$CB,$D9,$AA,$3F,$C2,$FA,$F5,$24
	.dc.b	$7F,$87,$7D,$9A,$FF,$86,$22,$F1,$38,$BF,$D4,$FD,$AB,$F2,$7B,$73
	.dc.b	$0F,$F1,$BB,$C2,$BC,$4F,$DF,$C7,$FC,$70,$DC,$BF,$DD,$35,$FE,$1F
	.dc.b	$8D,$7E,$0D,$F0,$23,$FD,$FA,$D7,$FC,$60,$DC,$BC,$91,$FE,$82,$EE
	.dc.b	$F9,$73,$9C,$7F,$D7,$BE,$75,$AF,$F5,$73,$72,$56,$65,$F5,$54,$9B
	.dc.b	$FE,$72,$3F,$27,$E0,$D9,$D7,$4D,$31,$5B,$E2,$FC,$FD,$FF,$84,$8F
	.dc.b	$A2,$57,$2F,$FA,$D3,$A7,$FD,$4A,$E7,$8F,$F4,$CF,$03,$F5,$73,$C1
	.dc.b	$FD,$A9,$92,$BD,$37,$FF,$49,$7C,$E7,$F9,$B2,$7C,$4D,$7F,$D8,$73
	.dc.b	$BD,$FF,$2B,$F7,$9F,$ED,$D9,$FF,$14,$EE,$8C,$47,$F9,$85,$B8,$7D
	.dc.b	$2F,$53,$FD,$75,$E8,$FE,$72,$73,$87,$D2,$F9,$2F,$E0,$B4,$E3,$46
	.dc.b	$FD,$AC,$F9,$7F,$D9,$72,$FF,$58,$FD,$4B,$EA,$BF,$4E,$DD,$7F,$C7
	.dc.b	$B9,$7F,$B2,$7C,$5E,$26,$AA,$4E,$B0,$E9,$A7,$F1,$CF,$9B,$7B,$1F
	.dc.b	$75,$B5,$F5,$DA,$FF,$BA,$38,$DD,$7B,$AF,$AB,$5A,$AF,$A6,$D7,$B4
	.dc.b	$68,$ED,$C1,$23,$B4,$7C,$4F,$71,$F8,$4D,$F9,$59,$5F,$F0,$02,$34
	.dc.b	$61,$FF,$74,$CA,$73,$1F,$63,$F7,$1F,$4F,$17,$FB,$69,$BA,$FF,$12
	.dc.b	$6C,$B8,$5F,$05,$FE,$DF,$8E,$59,$F9,$7D,$99,$18,$3E,$8C,$FE,$E9
	.dc.b	$63,$FB,$4F,$82,$7D,$57,$E0,$A5,$18,$E7,$FC,$1D,$F1,$BF,$C5,$37
	.dc.b	$4F,$F1,$1D,$6E,$FF,$CF,$6D,$7B,$C7,$F8,$A3,$3F,$4F,$89,$D0,$FE
	.dc.b	$A7,$5B,$C3,$C9,$97,$8E,$B7,$87,$AB,$9F,$E9,$7B,$52,$30,$43,$E7
	.dc.b	$B9,$7C,$27,$C9,$CF,$F1,$AF,$CB,$FB,$47,$C0,$7D,$DF,$62,$4F,$E6
	.dc.b	$27,$2D,$9C,$BF,$72,$7F,$4F,$3A,$8E,$E3,$9D,$F3,$9F,$99,$24,$7E
	.dc.b	$E2,$3F,$4C,$F0,$76,$FF,$71,$6B,$FD,$D9,$AF,$E6,$A3,$F3,$A7,$EE
	.dc.b	$3E,$B7,$F0,$91,$F0,$34,$46,$36,$FE,$67,$3E,$5A,$FE,$51,$F8,$07
	.dc.b	$FB,$2E,$3F,$91,$35,$F5,$B6,$7F,$9D,$8F,$D5,$3F,$47,$3F,$CA,$7E
	.dc.b	$BD,$E6,$7F,$24,$6C,$7F,$82,$DA,$FF,$84,$1A,$FF,$86,$38,$BF,$25
	.dc.b	$7F,$83,$E5,$FF,$74,$47,$E5,$65,$8F,$E4,$5F,$F3,$9F,$B7,$FC,$95
	.dc.b	$FF,$3A,$FE,$89,$E4,$7E,$9B,$70,$F8,$33,$B9,$4E,$DF,$E6,$ED,$7F
	.dc.b	$C5,$AD,$7A,$F7,$7F,$40,$97,$D7,$FF,$51,$6C,$FF,$4B,$6B,$E2,$9F
	.dc.b	$EA,$D5,$FD,$12,$FE,$9B,$72,$EF,$5F,$E1,$0E,$1F,$8F,$5A,$8F,$E9
	.dc.b	$F5,$FD,$41,$F4,$98,$21,$FD,$13,$C5,$7C,$7F,$F3,$43,$C3,$1F,$D9
	.dc.b	$FF,$01,$FE,$B6,$FE,$61,$B7,$F8,$66,$E3,$FE,$24,$C5,$E9,$F1,$F8
	.dc.b	$BF,$EC,$35,$CB,$0F,$EB,$ED,$7F,$CA,$AD,$7F,$D5,$2F,$47,$F8,$E7
	.dc.b	$88,$FA,$8F,$E5,$C7,$58,$FE,$E9,$3F,$91,$C7,$9F,$B3,$E9,$3F,$CB
	.dc.b	$7F,$46,$FC,$03,$FA,$C5,$F0,$9F,$29,$1F,$F2,$E5,$FE,$75,$F3,$C7
	.dc.b	$2F,$29,$F1,$B4,$ED,$3F,$E0,$5F,$9C,$FD,$F7,$07,$EA,$9B,$9F,$F9
	.dc.b	$C9,$B2,$CF,$C9,$F5,$BF,$E1,$59,$FE,$0D,$3F,$E0,$46,$BC,$7C,$2F
	.dc.b	$F0,$5C,$62,$3F,$D9,$EF,$4A,$FE,$BB,$E2,$F9,$CE,$95,$D7,$B6,$5F
	.dc.b	$9D,$F8,$BF,$B7,$3E,$5B,$33,$FE,$80,$C7,$F6,$2F,$A6,$BC,$24,$7F
	.dc.b	$DB,$FA,$7F,$8E,$B2,$7D,$8B,$EA,$3E,$8B,$6B,$C0,$7C,$7F,$8F,$F8
	.dc.b	$04,$79,$EF,$FE,$EA,$BF,$EB,$9C,$F9,$9F,$25,$7C,$22,$3F,$E9,$6F
	.dc.b	$19,$E3,$BE,$7B,$FB,$A7,$CC,$CA,$FE,$71,$1F,$CD,$0C,$3F,$DC,$A3
	.dc.b	$0F,$F8,$A3,$EA,$5F,$1F,$FE,$7E,$6C,$66,$3F,$D0,$3E,$2E,$23,$FA
	.dc.b	$5F,$BA,$AC,$70,$FF,$97,$A3,$17,$CC,$E2,$7F,$A9,$7C,$37,$D6,$52
	.dc.b	$61,$FE,$A5,$F0,$9F,$83,$62,$3F,$D3,$F5,$C1,$FF,$55,$35,$FF,$51
	.dc.b	$CF,$07,$35,$FE,$B6,$9D,$47,$FA,$79,$A8,$77,$A3,$31,$8F,$8F,$7A
	.dc.b	$5F,$EB,$D6,$C7,$FA,$ED,$AF,$8B,$AF,$53,$23,$4F,$FB,$17,$1F,$C9
	.dc.b	$47,$F9,$BF,$EB,$FF,$B2,$E3,$FE,$CF,$6C,$7F,$B5,$1A,$FF,$9A,$9A
	.dc.b	$FF,$B8,$30,$FF,$1F,$F3,$FF,$B5,$DB,$1F,$ED,$C6,$BF,$E7,$26,$BF
	.dc.b	$EE,$89,$FE,$4B,$F1,$DE,$1F,$27,$FD,$93,$3D,$DA,$ED,$AF,$C5,$F1
	.dc.b	$FF,$B1,$F5,$FC,$BC,$98,$FF,$1B,$8F,$F3,$0F,$C5,$BE,$A7,$78,$71
	.dc.b	$9F,$95,$86,$30,$E2,$EB,$AD,$87,$1A,$D9,$F4,$79,$3F,$BD,$E1,$E1
	.dc.b	$CB,$F3,$73,$FE,$E0,$6C,$7F,$B8,$1A,$FF,$B8,$1C,$F3,$7F,$AD,$B2
	.dc.b	$F4,$DF,$87,$07,$D9,$79,$6F,$78,$F7,$AC,$1B,$9B,$1F,$AD,$F9,$5F
	.dc.b	$7F,$47,$E9,$DE,$9F,$01,$8E,$A4,$52,$57,$C8,$9B,$AF,$F9,$AB,$1F
	.dc.b	$89,$E6,$FF,$11,$AF,$F4,$6B,$E4,$9F,$E9,$7E,$1E,$AB,$6D,$AD,$BF
	.dc.b	$4A,$5F,$59,$FE,$FA,$9F,$F5,$BC,$7E,$96,$30,$E1,$AF,$FA,$F1,$AF
	.dc.b	$EA,$CE,$7F,$17,$D6,$38,$F0,$FF,$AC,$71,$F7,$DF,$FB,$3E,$72,$F1
	.dc.b	$7F,$C1,$6F,$FB,$4F,$EB,$3F,$D1,$0E,$FC,$0F,$44,$63,$1F,$F2,$7D
	.dc.b	$70,$7C,$D3,$F8,$47,$D4,$D7,$DE,$D8,$E1,$87,$F9,$11,$B7,$FE,$77
	.dc.b	$C1,$EA,$7A,$EF,$E4,$7D,$63,$EC,$FC,$73,$86,$DF,$FB,$AE,$74,$FC
	.dc.b	$CC,$91,$87,$E7,$B5,$FF,$73,$B5,$F9,$8E,$F2,$FF,$2C,$B6,$7C,$DA
	.dc.b	$78,$5F,$07,$1F,$C6,$8F,$3B,$63,$27,$8E,$7F,$3A,$D8,$F8,$D1,$8B
	.dc.b	$F2,$C7,$29,$FC,$7F,$C4,$D9,$FE,$87,$F4,$DF,$CF,$27,$FA,$EC,$7F
	.dc.b	$2D,$F9,$3F,$F6,$9C,$7F,$BD,$A7,$53,$FE,$86,$9C,$BC,$BC,$5F,$B1
	.dc.b	$CE,$AD,$CF,$E3,$63,$13,$A9,$CC,$FF,$3B,$9C,$DF,$7A,$FD,$3F,$C8
	.dc.b	$FE,$CA,$FC,$3A,$FF,$57,$5F,$57,$FA,$3C,$9F,$35,$B8,$C3,$71,$FD
	.dc.b	$2E,$5E,$59,$F1,$12,$33,$7D,$1F,$81,$1F,$B8,$E2,$E5,$A7,$FA,$9B
	.dc.b	$EA,$BF,$BD,$57,$C8,$84,$61,$F9,$BE,$BF,$ED,$E3,$1A,$7A,$63,$E1
	.dc.b	$12,$BC,$2F,$E8,$7C,$07,$37,$F6,$C9,$FA,$53,$FC,$2B,$CB,$F8,$8F
	.dc.b	$F0,$3F,$A9,$FD,$66,$5F,$BB,$0E,$1F,$DF,$35,$FE,$D4,$DD,$BE,$C7
	.dc.b	$C4,$7D,$52,$61,$F1,$1F,$DA,$7D,$4B,$EA,$7F,$1D,$AA,$BF,$BA,$4E
	.dc.b	$79,$FF,$8F,$FC,$B7,$F4,$88,$C4,$FE,$3B,$1F,$F2,$BF,$C7,$7F,$1B
	.dc.b	$E0,$FF,$3F,$CE,$76,$4E,$7C,$AF,$E6,$71,$89,$FB,$6D,$DF,$E0,$61
	.dc.b	$D2,$73,$7F,$94,$AD,$69,$FE,$1F,$6C,$FF,$82,$C7,$E0,$EF,$BD,$EF
	.dc.b	$7F,$CE,$3C,$DF,$E7,$3E,$95,$BF,$E7,$2D,$CF,$FD,$B1,$CB,$FE,$57
	.dc.b	$9F,$E7,$B5,$C1,$FD,$6E,$CF,$8D,$7B,$3F,$D1,$30,$FE,$D1,$18,$AF
	.dc.b	$E8,$B1,$89,$D4,$72,$9D,$4F,$80,$F8,$BE,$BF,$2F,$9E,$B5,$B6,$CF
	.dc.b	$C6,$DF,$F2,$CF,$ED,$5A,$FE,$CD,$AB,$87,$FD,$1B,$7C,$B8,$7F,$01
	.dc.b	$2C,$FD,$9F,$21,$9C,$BF,$C6,$D1,$EF,$FA,$4F,$69,$EE,$FA,$EB,$3E
	.dc.b	$B6,$DF,$FC,$01,$D5,$18,$C7,$BE,$1F,$27,$E9,$3E,$3F,$8B,$F2,$1B
	.dc.b	$8F,$C5,$6E,$3C,$DB,$7E,$88,$7D,$47,$CA,$8C,$71,$7F,$73,$9C,$F1
	.dc.b	$7F,$6F,$8C,$57,$D0,$9F,$D7,$35,$FE,$D4,$D7,$FC,$AC,$D7,$9E,$1F
	.dc.b	$C1,$E3,$C0,$FF,$71,$FC,$69,$F0,$BF,$DF,$67,$35,$F3,$7E,$6A,$EA
	.dc.b	$5C,$BF,$E2,$F7,$26,$31,$FC,$7B,$88,$D6,$7B,$78,$70,$CA,$FC,$67
	.dc.b	$4D,$C7,$E9,$37,$6F,$A3,$C5,$FA,$5A,$EB,$7F,$E1,$3F,$A9,$F4,$C7
	.dc.b	$B7,$9F,$1B,$FA,$DB,$1F,$DA,$E7,$79,$7F,$86,$1A,$FF,$19,$6B,$FE
	.dc.b	$15,$77,$8F,$D9,$6C,$FE,$3B,$F5,$57,$DB,$F6,$B5,$FD,$FF,$AB,$67
	.dc.b	$FE,$29,$6E,$1F,$AB,$FA,$BC,$4F,$EC,$9B,$19,$3F,$9F,$4F,$0B,$EB
	.dc.b	$1F,$9E,$E1,$57,$D3,$3F,$77,$F3,$36,$7F,$70,$DA,$FC,$F3,$71,$FF
	.dc.b	$59,$7D,$1B,$E6,$FF,$B5,$B6,$FF,$AC,$B6,$FF,$8F,$B5,$FE,$DC,$E6
	.dc.b	$B9,$D7,$F9,$8F,$E3,$3F,$CB,$73,$9B,$FD,$AF,$C0,$27,$51,$FE,$80
	.dc.b	$EF,$B8,$FB,$8F,$DF,$56,$E7,$FD,$0B,$5F,$E6,$FA,$DE,$3E,$AF,$7B
	.dc.b	$D2,$4A,$F4,$B7,$0F,$F6,$9E,$3F,$E4,$C7,$B7,$4F,$FA,$23,$E2,$FC
	.dc.b	$F5,$FE,$8D,$FA,$8F,$F4,$A5,$7F,$5D,$F8,$34,$62,$3F,$CB,$51,$8F
	.dc.b	$84,$FF,$25,$37,$2F,$F3,$2B,$67,$D8,$DC,$FF,$86,$EC,$F9,$1F,$06
	.dc.b	$13,$0F,$F1,$2D,$ED,$F9,$33,$9C,$39,$E3,$E6,$C7,$FD,$28,$DA,$D9
	.dc.b	$C9,$C4,$FF,$B6,$9F,$F2,$BC,$79,$9B,$1E,$F6,$BE,$E6,$BF,$EB,$97
	.dc.b	$73,$E0,$F7,$BF,$D4,$5C,$FF,$25,$FF,$44,$57,$FB,$A7,$EA,$FF,$DD
	.dc.b	$5F,$56,$F7,$3F,$C1,$1E,$2F,$F3,$C8,$C5,$7F,$82,$EB,$75,$AC,$EC
	.dc.b	$E4,$7C,$9F,$07,$FD,$ED,$87,$F4,$68,$C4,$78,$4D,$D7,$36,$9C,$F0
	.dc.b	$DD,$78,$E7,$FD,$3B,$87,$82,$BA,$E1,$5E,$97,$C2,$31,$09,$D5,$F3
	.dc.b	$8F,$46,$1B,$7C,$27,$FA,$8E,$4E,$04,$FA,$F0,$F1,$C3,$0E,$47,$B9
	.dc.b	$F0,$19,$8F,$74,$66,$BC,$78,$79,$9B,$5B,$3E,$15,$70,$3B,$D7,$67
	.dc.b	$C3,$87,$A4,$E9,$3E,$7C,$3C,$CF,$23,$FD,$7C,$FC,$17,$DC,$77,$3E
	.dc.b	$59,$4F,$AD,$1D,$67,$B9,$F5,$DE,$E6,$63,$E5,$C6,$4F,$8B,$B7,$E1
	.dc.b	$23,$F0,$CD,$7C,$2D,$7F,$08,$E7,$E2,$3F,$CF,$7A,$FD,$28,$C7,$BB
	.dc.b	$FC,$F2,$91,$89,$D6,$1F,$ED,$BB,$E3,$5A,$8C,$5F,$17,$F2,$D2,$FA
	.dc.b	$AF,$F7,$14,$61,$B8,$F0,$C3,$F0,$73,$C2,$F9,$4F,$7A,$DE,$1F,$88
	.dc.b	$C3,$A3,$C2,$79,$3F,$8E,$8F,$2E,$7D,$E3,$F3,$91,$C1,$FA,$8E,$A7
	.dc.b	$8E,$1B,$9E,$57,$E0,$55,$F0,$27,$CD,$8F,$4A,$FF,$7F,$61,$C4,$CD
	.dc.b	$F6,$C7,$E6,$46,$71,$F6,$4F,$92,$73,$1E,$A9,$D4,$FB,$EB,$AC,$E4
	.dc.b	$F3,$9C,$A3,$CE,$7C,$33,$E3,$57,$CA,$6D,$F9,$D8,$79,$5F,$11,$8B
	.dc.b	$F0,$C7,$C0,$9F,$2B,$C8,$EE,$4C,$3D,$E7,$C0,$9E,$38,$FC,$F7,$18
	.dc.b	$7F,$2B,$3F,$80,$D7,$F3,$C6,$E1,$FC,$F1,$BA,$FF,$6A,$FB,$4B,$5F
	.dc.b	$CE,$E7,$FD,$64,$F3,$3F,$9E,$38,$74,$7F,$3F,$D7,$8E,$BF,$D7,$1B
	.dc.b	$C3,$FC,$42,$7F,$D1,$3C,$8F,$8E,$CD,$49,$FE,$98,$F8,$BA,$15,$B9
	.dc.b	$C9,$0F,$EA,$F3,$B8,$D4,$60,$F2,$10,$FE,$A5,$1E,$9F,$41,$D4,$F8
	.dc.b	$4F,$83,$18,$6C,$66,$BD,$3F,$41,$89,$E5,$39,$9F,$F0,$DC,$60,$F8
	.dc.b	$D8,$FA,$BE,$0F,$CC,$49,$F8,$7E,$C5,$7F,$C1,$39,$A7,$1C,$5F,$DF
	.dc.b	$2B,$C7,$ED,$4B,$FF,$69,$64,$8D,$F1,$FF,$40,$D7,$DB,$1E,$62,$57
	.dc.b	$F8,$C9,$F2,$F9,$8F,$82,$FA,$1E,$3B,$E5,$E9,$F4,$E7,$57,$FE,$74
	.dc.b	$8C,$B4,$FF,$5E,$56,$E3,$F1,$71,$FE,$7E,$F1,$6F,$A1,$FE,$37,$8F
	.dc.b	$F6,$14,$E2,$64,$7F,$58,$DB,$FD,$3D,$5B,$8F,$F5,$65,$74,$FA,$9F
	.dc.b	$F6,$8C,$6F,$5F,$CF,$C6,$23,$FC,$80,$1C,$5F,$FA,$CB,$CD,$7D,$3F
	.dc.b	$3F,$4F,$F3,$97,$B6,$DF,$38,$FD,$59,$93,$F4,$4F,$4C,$3F,$5F,$C3
	.dc.b	$F4,$E7,$C8,$7F,$72,$70,$F9,$27,$83,$FE,$D1,$D9,$E0,$FD,$2F,$A9
	.dc.b	$93,$FE,$E7,$8A,$FF,$93,$B8,$53,$8D,$7F,$D3,$B5,$F6,$BE,$0B,$ED
	.dc.b	$F8,$73,$0F,$F5,$5C,$65,$FF,$56,$C6,$5E,$91,$F6,$3E,$A3,$F0,$51
	.dc.b	$EA,$0E,$1F,$9A,$DB,$FF,$6A,$7D,$CB,$EA,$79,$71,$FC,$CA,$E1,$7F
	.dc.b	$29,$AF,$F1,$66,$BF,$E1,$87,$A7,$07,$F1,$AF,$A0,$7C,$17,$E3,$1B
	.dc.b	$5E,$5A,$DB,$F2,$FE,$B7,$F3,$1E,$54,$8F,$A7,$E0,$38,$D5,$F9,$D3
	.dc.b	$53,$FE,$F9,$AC,$3C,$25,$EC,$FE,$8F,$8E,$4E,$62,$BF,$F0,$06,$2F
	.dc.b	$07,$FB,$42,$7F,$86,$E5,$CA,$BF,$0B,$DD,$F3,$27,$27,$EB,$FE,$7A
	.dc.b	$CF,$F3,$29,$D3,$6B,$93,$F9,$48,$CB,$63,$37,$FE,$E0,$E3,$FD,$D1
	.dc.b	$D2,$7F,$D2,$F9,$7F,$84,$9F,$77,$27,$EF,$CF,$E0,$1E,$44,$8C,$90
	.dc.b	$E2,$FF,$37,$F9,$DF,$D5,$C7,$CE,$5E,$E6,$4D,$76,$C9,$D4,$65,$E1
	.dc.b	$1F,$CC,$B8,$F9,$C7,$FA,$E7,$54,$63,$47,$2F,$F1,$F8,$C3,$FE,$60
	.dc.b	$1F,$A7,$BE,$43,$A6,$3F,$D0,$6B,$AE,$3F,$D9,$B1,$EB,$39,$24,$7E
	.dc.b	$86,$5E,$4F,$E4,$4E,$71,$FB,$EC,$9F,$50,$FA,$5F,$ED,$71,$86,$C6
	.dc.b	$67,$F8,$2F,$D4,$BC,$3F,$34,$7E,$BF,$E9,$FE,$A3,$FD,$C9,$AF,$F1
	.dc.b	$26,$BF,$DF,$9A,$FF,$A0,$3E,$A7,$F9,$43,$FC,$61,$AF,$F8,$11,$AE
	.dc.b	$FC,$8F,$EA,$FC,$7F,$AC,$8F,$E5,$3E,$BF,$F1,$8D,$3E,$7C,$7F,$A3
	.dc.b	$23,$1C,$7F,$E8,$EE,$F5,$F1,$BD,$DA,$FF,$90,$BF,$CA,$9A,$FF,$85
	.dc.b	$9A,$F5,$F2,$FF,$99,$FD,$47,$C7,$8F,$E7,$6D,$AE,$AD,$C7,$95,$7E
	.dc.b	$27,$29,$7F,$AE,$E2,$FE,$95,$87,$AB,$3F,$3F,$CC,$6A,$BF,$55,$5F
	.dc.b	$13,$97,$F1,$F8,$7E,$1C,$93,$FC,$8F,$3F,$E8,$78,$F8,$A3,$93,$F1
	.dc.b	$AB,$F2,$FE,$9F,$E6,$C9,$3F,$A5,$FB,$9A,$F8,$23,$F2,$84,$7F,$80
	.dc.b	$63,$FE,$1C,$BF,$97,$F5,$1F,$D2,$24,$FF,$4A,$C3,$BC,$7F,$50,$FA
	.dc.b	$37,$D4,$BE,$87,$FC,$7B,$18,$6C,$66,$7F,$71,$F1,$18,$9F,$F1,$F5
	.dc.b	$6E,$7F,$C8,$6D,$7F,$B1,$B5,$FF,$27,$B9,$C7,$CF,$E0,$7F,$80,$B2
	.dc.b	$FC,$DC,$7E,$2A,$75,$C7,$FE,$D6,$7C,$FF,$03,$E4,$26,$1F,$C2,$FE
	.dc.b	$8E,$BE,$0D,$13,$E6,$6C,$79,$9A,$EE,$3F,$C2,$7E,$3F,$F1,$48,$FE
	.dc.b	$B8,$D5,$8F,$96,$B8,$BE,$7F,$A9,$7C,$AF,$CA,$6D,$7C,$F6,$BC,$B3
	.dc.b	$FD,$27,$53,$E6,$38,$5F,$CC,$BD,$DF,$C8,$6B,$F0,$DA,$B5,$47,$6C
	.dc.b	$7F,$5B,$9F,$E7,$32,$FF,$3F,$B6,$F8,$56,$E7,$E7,$74,$FF,$84,$30
	.dc.b	$34,$FF,$09,$9F,$BA,$E1,$FC,$3B,$FE,$21,$BE,$3D,$DF,$84,$8C,$47
	.dc.b	$95,$A7,$F0,$2E,$9F,$E8,$9F,$13,$27,$FA,$24,$7F,$47,$30,$FC,$73
	.dc.b	$C7,$BA,$75,$1F,$D9,$BE,$75,$F4,$7F,$E7,$66,$C6,$63,$FD,$05,$F6
	.dc.b	$79,$64,$64,$FF,$48,$C7,$FA,$5B,$AD,$F3,$D4,$93,$FE,$96,$DE,$BE
	.dc.b	$5E,$9F,$F4,$C4,$E3,$8F,$F8,$57,$C0,$FF,$1E,$7F,$92,$CE,$AB,$FD
	.dc.b	$52,$D7,$FC,$92,$D7,$FD,$6A,$D4,$1F,$F5,$37,$2F,$9C,$91,$FE,$1A
	.dc.b	$FC,$39,$F6,$78,$8F,$36,$7F,$7D,$E4,$DE,$25,$70,$9F,$F1,$E7,$D4
	.dc.b	$7F,$5C,$C7,$FD,$78,$D8,$FF,$60,$B5,$FF,$30,$B5,$FF,$66,$FC,$5F
	.dc.b	$FA,$1F,$9F,$FD,$88,$D8,$FF,$64,$B5,$FF,$34,$B5,$FF,$6A,$CB,$E1
	.dc.b	$3F,$D8,$D1,$D9,$AE,$9A,$FC,$68,$E1,$F0,$FF,$3B,$3F,$F0,$04,$FC
	.dc.b	$23,$FA,$BE,$DF,$53,$C3,$7C,$26,$38,$FE,$7E,$BF,$EC,$D7,$FC,$EE
	.dc.b	$DB,$E5,$18,$AE,$2E,$B0,$EF,$87,$E0,$3E,$BD,$F3,$BF,$D8,$30,$EB
	.dc.b	$18,$BF,$C1,$B6,$3F,$06,$D7,$F0,$6E,$6B,$A5,$79,$E7,$F0,$98,$3E
	.dc.b	$77,$19,$30,$E1,$8F,$83,$0C,$C6,$30,$E3,$78,$9E,$9E,$3B,$A9,$F2
	.dc.b	$58,$6E,$3F,$E5,$6F,$CE,$7F,$BD,$3F,$B3,$9D,$11,$D7,$9D,$72,$FF
	.dc.b	$1C,$5F,$F8,$C6,$7F,$C8,$9F,$29,$D7,$D7,$F6,$74,$FF,$82,$23,$13
	.dc.b	$FE,$5F,$AE,$17,$FE,$5D,$6B,$FE,$37,$6B,$FD,$27,$EE,$6B,$FC,$9D
	.dc.b	$BD,$26,$93,$1F,$99,$19,$7E,$74,$BE,$A7,$FD,$35,$87,$F6,$BB,$F1
	.dc.b	$F2,$7F,$59,$7F,$84,$C7,$F9,$7F,$CB,$E7,$18,$BF,$E3,$9A,$7F,$89
	.dc.b	$27,$3E,$07,$E1,$67,$CC,$DC,$FC,$CD,$BE,$53,$FD,$B1,$E9,$8F,$86
	.dc.b	$7C,$F9,$7F,$93,$F2,$F3,$47,$B3,$E3,$DE,$6F,$DE,$FC,$0D,$F3,$19
	.dc.b	$8C,$36,$FF,$BE,$D7,$06,$CF,$96,$31,$9F,$F8,$01,$B9,$7F,$7B,$6D
	.dc.b	$7F,$1A,$9F,$95,$1D,$5F,$E7,$DA,$21,$E4,$FE,$0B,$C9,$7C,$BF,$A6
	.dc.b	$30,$DC,$3C,$B1,$E3,$23,$83,$FC,$41,$5D,$1F,$F6,$C6,$DF,$61,$FF
	.dc.b	$07,$46,$71,$FF,$46,$7D,$D7,$EA,$E7,$FD,$67,$B7,$E8,$DF,$F5,$AC
	.dc.b	$66,$7F,$D2,$53,$97,$8F,$5F,$DB,$EA,$56,$E7,$FD,$18,$D7,$F9,$EC
	.dc.b	$6A,$78,$4F,$5D,$BF,$C9,$0D,$C3,$E5,$7C,$9B,$E3,$BF,$6F,$39,$C5
	.dc.b	$8F,$83,$E6,$BE,$57,$CB,$E9,$B8,$FA,$9E,$E4,$7F,$D9,$96,$8C,$5F
	.dc.b	$C4,$7F,$47,$8B,$E2,$FA,$5F,$FB,$63,$C9,$F1,$64,$7C,$9A,$3E,$E7
	.dc.b	$FD,$EF,$5B,$DB,$F8,$D4,$E7,$E3,$BF,$44,$D9,$FC,$46,$9F,$D8,$A3
	.dc.b	$0F,$F9,$BA,$F9,$BA,$1F,$51,$FB,$67,$0D,$79,$B5,$FC,$A3,$5F,$F0
	.dc.b	$7F,$91,$F9,$13,$F9,$77,$D3,$BF,$2F,$7C,$F6,$72,$FC,$1F,$C8,$FD
	.dc.b	$82,$3F,$29,$FD,$E7,$E2,$CF,$C5,$FF,$37,$3E,$B3,$F4,$9F,$8C,$3E
	.dc.b	$8B,$F9,$E5,$6F,$BB,$F8,$8C,$E6,$B8,$6B,$FE,$04,$F4,$9F,$71,$FC
	.dc.b	$5B,$C8,$7B,$5F,$C6,$27,$37,$C2,$FF,$C7,$B3,$9B,$FE,$21,$0A,$DB
	.dc.b	$F2,$DA,$FD,$26,$BF,$BB,$77,$C9,$FE,$BC,$6D,$E7,$DB,$EC,$FA,$31
	.dc.b	$FF,$48,$7D,$B7,$F3,$EF,$7C,$DC,$70,$CA,$E1,$3A,$6C,$E4,$FF,$23
	.dc.b	$75,$BE,$F3,$F0,$79,$FF,$AE,$FC,$7F,$F7,$AD,$FC,$8A,$D7,$5F,$F0
	.dc.b	$DF,$90,$FE,$BF,$87,$0A,$FD,$9B,$71,$FF,$31,$FC,$EF,$F7,$E7,$BC
	.dc.b	$67,$0F,$07,$C4,$7F,$17,$F7,$79,$B7,$0F,$05,$23,$F8,$23,$63,$F3
	.dc.b	$15,$C9,$B1,$9B,$FD,$0C,$E5,$F0,$D7,$F6,$6C,$3F,$BD,$4E,$74,$FF
	.dc.b	$6D,$6A,$FB,$4F,$9C,$7F,$7C,$EA,$7C,$AE,$34,$F9,$B3,$73,$F9,$75
	.dc.b	$A9,$C9,$EA,$F5,$BF,$9A,$78,$DF,$21,$B1,$92,$57,$F5,$ED,$3F,$81
	.dc.b	$EC,$AF,$ED,$7F,$54,$FA,$8F,$ED,$BF,$83,$7F,$97,$71,$FC,$B4,$E5
	.dc.b	$FC,$C4,$67,$1F,$F4,$20,$9D,$3F,$C0,$5A,$FF,$08,$6B,$FD,$45,$AF
	.dc.b	$F9,$B1,$E3,$BA,$B5,$7E,$6F,$9D,$F3,$BF,$86,$F3,$7F,$AA,$76,$DA
	.dc.b	$FC,$E7,$47,$EC,$1F,$BF,$39,$CB,$E1,$7D,$5D,$CB,$B4,$7F,$9F,$3E
	.dc.b	$DF,$F4,$ED,$EF,$FF,$07,$F4,$7E,$3E,$B7,$2A,$DB,$6B,$FC,$2B,$18
	.dc.b	$AE,$18,$FF,$6F,$8D,$FD,$63,$E6,$FF,$C5,$BE,$07,$FB,$46,$BF,$82
	.dc.b	$4E,$5F,$DB,$46,$71,$F9,$F3,$97,$F9,$73,$5F,$E6,$AD,$7F,$BF,$35
	.dc.b	$FF,$4C,$BC,$72,$F5,$B6,$BF,$C0,$9F,$58,$EB,$FF,$17,$CE,$6F,$FD
	.dc.b	$A7,$1F,$EB,$EA,$FE,$07,$87,$E6,$F9,$7F,$85,$32,$32,$7E,$FF,$A7
	.dc.b	$F2,$0E,$2F,$FC,$63,$7F,$E5,$A6,$FC,$0F,$E9,$78,$7F,$8A,$32,$BF
	.dc.b	$88,$B9,$69,$FC,$2F,$F2,$3F,$F1,$4B,$FB,$A6,$D7,$EE,$1A,$FF,$88
	.dc.b	$7B,$91,$88,$FF,$00,$BC,$36,$FC,$2C,$EB,$DC,$FD,$C7,$83,$FE,$74
	.dc.b	$AD,$61,$FE,$C4,$F1,$3F,$D7,$3F,$4B,$A8,$FF,$61,$37,$2F,$F2,$7F
	.dc.b	$C5,$FF,$63,$CF,$AE,$FF,$C9,$27,$35,$B9,$F0,$E3,$F1,$67,$D0,$D7
	.dc.b	$CF,$AF,$F8,$76,$31,$1F,$E1,$07,$E6,$4F,$29,$FF,$45,$FD,$F7,$FB
	.dc.b	$53,$2F,$91,$1F,$E9,$6F,$B6,$7D,$77,$F9,$87,$E2,$D1,$8A,$FF,$73
	.dc.b	$B7,$2F,$F7,$3B,$5F,$F7,$3B,$9C,$7F,$DD,$ED,$9F,$F5,$DF,$B6,$F9
	.dc.b	$0F,$06,$32,$7F,$D7,$B5,$BC,$3F,$DD,$D7,$B9,$CD,$6E,$7F,$E0,$05
	.dc.b	$6F,$1F,$F1,$EE,$19,$C7,$C3,$E6,$FA,$EB,$FB,$66,$3E,$A8,$C1,$1F
	.dc.b	$0B,$FE,$B8,$F1,$BF,$D6,$98,$7C,$28,$FF,$48,$E2,$1D,$77,$EB,$FE
	.dc.b	$4C,$7F,$73,$C3,$1A,$7F,$88,$EA,$CE,$45,$70,$7F,$62,$3E,$93,$0F
	.dc.b	$A0,$C5,$73,$69,$CE,$36,$F9,$0E,$B1,$FE,$07,$BF,$0C,$76,$AE,$B8
	.dc.b	$7F,$22,$78,$E1,$A9,$F0,$46,$1E,$67,$A5,$C6,$1C,$AF,$E2,$38,$C3
	.dc.b	$C1,$5E,$63,$C0,$E2,$BA,$1C,$63,$A9,$A3,$B4,$7B,$E3,$47,$07,$91
	.dc.b	$92,$C7,$F9,$5A,$FC,$71,$E2,$AF,$24,$FC,$78,$E0,$F6,$38,$1E,$23
	.dc.b	$E0,$BF,$3E,$3C,$C7,$28,$FF,$20,$5F,$67,$8C,$6A,$78,$1E,$13,$EC
	.dc.b	$63,$F7,$67,$BB,$5F,$C0,$B5,$FC,$75,$FE,$3E,$35,$EE,$7F,$A3,$F5
	.dc.b	$C6,$1F,$EB,$B8,$C5,$AB,$7E,$FA,$F7,$3F,$99,$AD,$D6,$A7,$D5,$3F
	.dc.b	$E8,$7F,$7B,$FD,$8D,$87,$E5,$6F,$FD,$1D,$3A,$7B,$C6,$70,$ED,$3F
	.dc.b	$7E,$76,$DA,$E7,$97,$9E,$B7,$7C,$70,$FC,$AE,$1C,$34,$F0,$61,$FE
	.dc.b	$D7,$75,$8F,$E1,$A7,$C3,$8F,$C1,$AE,$B3,$ED,$6B,$FE,$A1,$C7,$BC
	.dc.b	$F8,$70,$EF,$5D,$63,$95,$78,$AB,$D7,$87,$80,$F1,$C7,$03,$57,$1F
	.dc.b	$25,$F3,$7C,$46,$67,$D3,$8F,$28,$F8,$87,$CB,$BF,$2B,$F5,$5F,$0E
	.dc.b	$1E,$48,$EA,$73,$7D,$4E,$09,$5E,$78,$CD,$77,$23,$83,$8C,$78,$C8
	.dc.b	$E0,$FC,$87,$92,$F9,$0F,$E4,$17,$F8,$BD,$B3,$18,$CF,$FC,$57,$AA
	.dc.b	$F8,$5F,$FA,$9A,$3F,$D0,$9F,$65,$BC,$38,$CF,$FB,$1E,$78,$61,$FE
	.dc.b	$C9,$8F,$CD,$5F,$27,$FD,$A3,$3A,$78,$43,$EC,$BA,$47,$06,$CF,$8D
	.dc.b	$BC,$4F,$B0,$CD,$F1,$BF,$33,$5F,$71,$CC,$F5,$4F,$17,$75,$C8,$8E
	.dc.b	$AA,$BE,$8A,$E8,$FC,$AB,$F2,$3E,$AA,$F8,$1E,$07,$F8,$53,$E1,$7D
	.dc.b	$6D,$9F,$F0,$F4,$63,$A9,$D9,$AE,$C4,$F2,$6D,$7F,$88,$E3,$07,$CD
	.dc.b	$EF,$77,$BB,$BA,$7B,$B2,$FF,$D8,$D3,$96,$CE,$5F,$F5,$FE,$5F,$EF
	.dc.b	$27,$27,$F8,$CF,$E8,$7C,$84,$7F,$D8,$11,$B7,$FD,$8F,$92,$7D,$04
	.dc.b	$8F,$F1,$D3,$E7,$F1,$5F,$21,$F7,$B3,$FF,$2B,$36,$33,$5D,$BC,$7C
	.dc.b	$4C,$8F,$F6,$CB,$F2,$63,$86,$B2,$7F,$DB,$AF,$CD,$30,$FF,$95,$9B
	.dc.b	$D1,$FE,$44,$CF,$E6,$CF,$FA,$8B,$F2,$1F,$8C,$DD,$7F,$26,$A3,$35
	.dc.b	$FE,$8A,$FB,$07,$9F,$A9,$E3,$5F,$AE,$38,$5F,$E8,$DA,$7E,$96,$36
	.dc.b	$FE,$D1,$EB,$D5,$FE,$53,$F8,$87,$D5,$FC,$76,$D7,$F8,$67,$2F,$F4
	.dc.b	$DD,$F0,$8F,$F0,$E7,$DF,$FF,$84,$E7,$FD,$3E,$DC,$3F,$80,$B6,$77
	.dc.b	$AF,$BA,$31,$06,$92,$3F,$37,$E5,$FE,$A6,$7F,$5B,$D0,$8C,$61,$FE
	.dc.b	$70,$6E,$1F,$E6,$DF,$67,$FA,$5C,$E7,$B7,$F0,$5E,$AF,$FA,$D2,$3F
	.dc.b	$D7,$7C,$C8,$C5,$FF,$8D,$F1,$ED,$7F,$53,$EB,$5F,$09,$49,$A7,$C5
	.dc.b	$BF,$E3,$FC,$9F,$95,$8C,$1F,$E4,$1A,$E2,$FF,$17,$6B,$F6,$43,$87
	.dc.b	$EE,$35,$FE,$56,$F0,$CF,$FD,$9B,$F6,$1F,$84,$9E,$E1,$C3,$FE,$F0
	.dc.b	$9E,$17,$ED,$DB,$E2,$B7,$C9,$F9,$E7,$E9,$F7,$7D,$7E,$30,$F5,$F0
	.dc.b	$7F,$AF,$FE,$77,$F5,$46,$63,$FD,$6B,$F2,$FF,$AD,$23,$F0,$B8,$5F
	.dc.b	$7E,$F9,$BC,$CF,$03,$AF,$7A,$FF,$8D,$D6,$E7,$96,$DF,$E2,$1E,$1F
	.dc.b	$C0,$79,$BF,$86,$3D,$1F,$47,$6B,$E6,$67,$F8,$79,$FE,$4F,$7F,$E2
	.dc.b	$79,$F8,$51,$83,$F8,$27,$E0,$12,$3B,$C7,$F2,$CF,$BE,$FF,$79,$47
	.dc.b	$AD,$AF,$E5,$1B,$8F,$E3,$38,$7C,$DC,$7F,$8B,$9C,$BF,$EC,$1D,$BF
	.dc.b	$99,$DF,$C0,$F2,$3B,$7B,$BF,$C5,$7D,$09,$3F,$11,$FE,$35,$F9,$8B
	.dc.b	$E2,$33,$3F,$EE,$49,$BA,$7E,$72,$B5,$3F,$84,$FA,$1F,$F7,$B3,$72
	.dc.b	$FE,$1F,$18,$8F,$EB,$21,$7C,$5E,$A7,$F8,$64,$6E,$7F,$88,$35,$FE
	.dc.b	$FE,$D7,$FD,$40,$F5,$B6,$9A,$F7,$FC,$27,$5A,$FC,$CF,$CF,$3E,$5F
	.dc.b	$F9,$1F,$C6,$BE,$26,$9F,$C9,$A7,$55,$C3,$AF,$F9,$04,$7E,$1E,$3F
	.dc.b	$58,$D7,$F9,$33,$5F,$A5,$F6,$3F,$D0,$C8,$FE,$E3,$E5,$5F,$64,$F9
	.dc.b	$7F,$E6,$BF,$5E,$F9,$DF,$E5,$CD,$BF,$F0,$93,$5F,$DE,$35,$FE,$6D
	.dc.b	$F6,$1F,$CB,$1F,$E0,$0D,$73,$E8,$FF,$39,$8C,$47,$F3,$1F,$C0,$7D
	.dc.b	$9C,$7F,$2A,$3F,$CE,$D1,$8E,$3D,$6F,$FD,$2A,$FB,$5E,$FC,$17,$6B
	.dc.b	$87,$F3,$84,$CB,$F8,$4F,$D8,$3E,$1F,$11,$FD,$36,$F8,$D7,$F8,$42
	.dc.b	$F9,$4F,$EB,$9A,$ED,$B8,$7F,$2E,$E9,$FE,$26,$DF,$5B,$AB,$E5,$97
	.dc.b	$3F,$EB,$C9,$7E,$42,$9D,$5F,$11,$1F,$13,$FE,$22,$FB,$8F,$E8,$70
	.dc.b	$C9,$18,$7E,$1C,$FF,$90,$BE,$89,$F6,$D2,$FF,$9C,$C6,$23,$46,$63
	.dc.b	$F4,$FF,$63,$FB,$E5,$3B,$7F,$9E,$35,$FE,$EC,$D7,$FC,$CA,$D4,$72
	.dc.b	$9E,$D9,$7C,$9F,$93,$FF,$0C,$1E,$28,$F1,$FC,$2C,$AF,$3C,$7F,$9A
	.dc.b	$DF,$E8,$78,$7F,$AF,$5C,$CF,$F9,$9F,$0F,$7B,$6B,$84,$F9,$9A,$F9
	.dc.b	$9C,$E1,$F8,$BC,$FA,$4F,$F7,$46,$C7,$E5,$6F,$AC,$FC,$AF,$C0,$5F
	.dc.b	$25,$F3,$DB,$19,$8F,$F1,$0F,$9A,$FA,$ED,$47,$9E,$B7,$5F,$29,$AF
	.dc.b	$CA,$6B,$F2,$AD,$1C,$30,$FC,$95,$7F,$90,$4E,$B5,$E5,$F9,$CF,$21
	.dc.b	$1F,$ED,$78,$FE,$EE,$BD,$96,$AF,$21,$2B,$EF,$C6,$5F,$EA,$3F,$57
	.dc.b	$5F,$D5,$35,$FF,$52,$F0,$A3,$11,$FE,$11,$6B,$F9,$97,$95,$FE,$AF
	.dc.b	$E2,$3F,$CB,$47,$FA,$02,$33,$D7,$C7,$E1,$30,$4B,$FE,$E5,$F9,$4F
	.dc.b	$AE,$75,$1F,$DD,$69,$5B,$8F,$F1,$7D,$6D,$7C,$5F,$F8,$5B,$E5,$3F
	.dc.b	$D4,$FE,$E1,$A7,$FA,$AB,$A7,$F9,$C5,$7F,$AE,$E7,$4D,$9D,$BF,$EB
	.dc.b	$F9,$DB,$FE,$AA,$AE,$CF,$FA,$FE,$B8,$BF,$EB,$C6,$BF,$EA,$56,$BF
	.dc.b	$AA,$8F,$F1,$5B,$E6,$21,$DF,$E3,$EB,$DE,$7F,$CD,$5A,$FF,$B5,$1B
	.dc.b	$A7,$3C,$17,$FE,$76,$FA,$EF,$F2,$96,$3F,$ED,$16,$E3,$FE,$3B,$6B
	.dc.b	$FE,$75,$6B,$FE,$E4,$F9,$3F,$E1,$FC,$BF,$ED,$96,$C7,$F9,$0D,$AF
	.dc.b	$F9,$F5,$AF,$FB,$B2,$7F,$93,$F8,$1E,$1E,$DF,$E7,$13,$E1,$6B,$E1
	.dc.b	$6B,$F1,$7C,$FF,$F7,$B6,$9F,$EE,$F6,$C7,$D1,$C3,$F0,$73,$9F,$91
	.dc.b	$FE,$A1,$BF,$FB,$E2,$F6,$D8,$CC,$FD,$1A,$E3,$7F,$6E,$F7,$3E,$1F
	.dc.b	$3F,$FA,$D4,$FF,$B7,$67,$FC,$01,$3F,$EF,$D6,$C7,$FB,$E9,$AE,$E7
	.dc.b	$BB,$F9,$48,$DE,$1F,$E7,$CF,$6F,$FA,$D5,$FB,$6F,$7B,$7F,$87,$5B
	.dc.b	$7C,$27,$F9,$8F,$BD,$FE,$56,$D7,$35,$DB,$A3,$F6,$DF,$19,$7A,$3F
	.dc.b	$D9,$B8,$7F,$B8,$3D,$CC,$4C,$8F,$D6,$3E,$6C,$7F,$DD,$33,$9D,$9B
	.dc.b	$7F,$65,$FA,$BE,$DF,$5B,$FE,$61,$C7,$FC,$C7,$18,$F8,$0E,$5E,$4A
	.dc.b	$31,$3A,$3F,$CD,$D5,$C7,$0F,$F1,$DF,$D5,$FF,$C0,$09,$FF,$47,$42
	.dc.b	$30,$FF,$A4,$5A,$FF,$B2,$DA,$FF,$54,$38,$7B,$DF,$41,$B3,$FE,$90
	.dc.b	$FC,$05,$79,$BE,$B9,$FD,$E5,$EC,$7D,$8E,$B7,$D7,$7A,$F1,$FF,$4C
	.dc.b	$D6,$9F,$EE,$31,$9C,$BD,$77,$F8,$87,$C9,$5B,$7D,$51,$97,$FB,$5C
	.dc.b	$65,$F6,$47,$94,$F5,$65,$FC,$66,$FC,$31,$E9,$F8,$CF,$30,$FA,$EC
	.dc.b	$4F,$F8,$12,$30,$FA,$1F,$05,$FC,$D6,$E0,$69,$27,$81,$FE,$B8,$FC
	.dc.b	$05,$F6,$9D,$AF,$EC,$F1,$78,$BD,$AF,$98,$DB,$F8,$EF,$62,$3F,$7B
	.dc.b	$1F,$C1,$57,$DB,$9E,$33,$DF,$E8,$2C,$FF,$A0,$3E,$9D,$F3,$BF,$9F
	.dc.b	$FA,$0F,$D2,$4F,$F8,$52,$73,$7F,$EB,$B9,$CC,$A7,$2F,$FB,$6D,$AF
	.dc.b	$E2,$4D,$57,$C6,$8C,$C7,$E3,$5E,$3C,$7F,$C1,$1B,$7F,$EF,$3F,$A0
	.dc.b	$7C,$9F,$E4,$67,$37,$E5,$F9,$0F,$47,$CB,$E5,$B8,$7F,$B9,$29,$1E
	.dc.b	$12,$47,$E9,$23,$2D,$8E,$4F,$FB,$53,$E6,$AD,$FE,$E2,$3F,$C6,$F9
	.dc.b	$3E,$BF,$FD,$4F,$1F,$F0,$02,$31,$F1,$5F,$A7,$FA,$FF,$D1,$CF,$0A
	.dc.b	$DD,$FF,$AC,$27,$F5,$EF,$F9,$09,$A9,$A6,$BC,$46,$CF,$85,$ED,$DF
	.dc.b	$FC,$5E,$31,$8F,$98,$FD,$D5,$F8,$B0,$7C,$4B,$EB,$DE,$BF,$F2,$C8
	.dc.b	$C3,$F8,$3F,$A4,$C4,$EB,$0E,$97,$FE,$58,$FA,$C7,$CD,$7E,$89,$B3
	.dc.b	$FC,$41,$B3,$F8,$30,$E1,$FE,$1D,$F5,$CF,$AF,$7D,$7F,$F3,$E8,$C3
	.dc.b	$FC,$8B,$DB,$C4,$EB,$0F,$1D,$FF,$A6,$7D,$DF,$E2,$DE,$93,$E2,$2C
	.dc.b	$61,$B5,$FC,$92,$7F,$AB,$D7,$E5,$DA,$FF,$21,$6B,$FE,$76,$7F,$AE
	.dc.b	$F4,$7F,$5A,$6C,$FF,$27,$FA,$D7,$D9,$F6,$BF,$C4,$0F,$4B,$FD,$E7
	.dc.b	$5F,$29,$7A,$2F,$B0,$FE,$7F,$5F,$D0,$1B,$5B,$9D,$36,$74,$F3,$F9
	.dc.b	$AF,$ED,$F1,$FE,$A1,$FA,$9F,$C7,$E9,$FE,$C9,$C5,$5A,$ED,$56,$B0
	.dc.b	$FE,$C7,$39,$8C,$61,$E9,$8C,$77,$FF,$5D,$5F,$23,$FD,$9E,$9F,$5B
	.dc.b	$98,$C3,$70,$FC,$4E,$3E,$08,$C6,$3F,$1A,$31,$3E,$BA,$E2,$D3,$8D
	.dc.b	$F6,$79,$F5,$3E,$6F,$FB,$3F,$C4,$7F,$79,$8C,$4F,$C9,$8C,$4C,$8F
	.dc.b	$91,$F3,$17,$2F,$AF,$B7,$C6,$F7,$7D,$9E,$6F,$BD,$B7,$F3,$DB,$5C
	.dc.b	$37,$FE,$53,$B7,$CC,$BF,$ED,$1C,$AF,$8F,$B9,$E1,$B9,$7E,$6A,$74
	.dc.b	$FF,$78,$6E,$3F,$A9,$6B,$FC,$B5,$AF,$E4,$67,$A4,$FE,$93,$EB,$7E
	.dc.b	$0C,$FE,$48,$57,$4E,$4F,$C2,$66,$FA,$EF,$B5,$F3,$7F,$75,$B8,$FE
	.dc.b	$35,$B8,$7B,$DB,$7F,$E3,$CE,$5F,$C8,$CB,$E5,$3F,$65,$7F,$B6,$9C
	.dc.b	$C6,$27,$C9,$1C,$09,$AF,$F8,$3D,$E7,$F5,$DF,$A6,$BF,$67,$CF,$FF
	.dc.b	$84,$E3,$17,$ED,$D9,$39,$EA,$FE,$EB,$5A,$AF,$EA,$8D,$7F,$C3,$CD
	.dc.b	$7F,$DE,$6F,$4C,$FF,$CA,$ED,$C3,$FC,$8F,$FB,$93,$E8,$FB,$47,$E4
	.dc.b	$B8,$FF,$C6,$FE,$0F,$E4,$5B,$19,$26,$3F,$87,$F1,$F5,$1F,$9A,$E2
	.dc.b	$FF,$1F,$FD,$57,$E6,$5B,$19,$25,$7E,$3F,$D4,$D4,$7E,$8B,$E9,$3F
	.dc.b	$C4,$5D,$2E,$E4,$61,$B3,$FC,$CE,$75,$39,$3F,$50,$7F,$96,$7E,$B7
	.dc.b	$FA,$44,$6A,$7F,$8C,$47,$F4,$BE,$D3,$4F,$F1,$A7,$17,$F0,$F7,$1E
	.dc.b	$8F,$33,$A7,$C0,$DB,$FE,$07,$F5,$9F,$2D,$F2,$37,$5F,$F0,$56,$48
	.dc.b	$C4,$7F,$BA,$9A,$84,$6F,$8F,$FC,$F5,$3C,$F2,$F2,$79,$38,$9F,$F4
	.dc.b	$2F,$B7,$FE,$43,$C3,$FC,$31,$F1,$88,$C4,$7F,$98,$E3,$1E,$6F,$F9
	.dc.b	$3D,$B9,$FF,$5D,$6C,$FF,$B6,$EB,$A6,$3F,$0F,$BF,$BE,$13,$0F,$ED
	.dc.b	$78,$70,$AD,$46,$2F,$8B,$FE,$DD,$6E,$1F,$EF,$46,$F0,$7D,$2C,$7A
	.dc.b	$4F,$FB,$AF,$C7,$FE,$6B,$7E,$56,$C7,$B9,$AF,$B5,$AF,$FB,$AD,$DE
	.dc.b	$09,$C7,$91,$F8,$5F,$98,$FF,$56,$D6,$A7,$FB,$F7,$6F,$D3,$FA,$F7
	.dc.b	$D3,$7F,$6D,$9C,$3A,$C9,$3A,$7F,$B8,$4E,$9F,$DB,$8F,$A7,$D3,$FD
	.dc.b	$CE,$3F,$01,$18,$8F,$DC,$9B,$AF,$03,$4E,$75,$BF,$43,$D9,$1E,$29
	.dc.b	$F2,$4E,$6F,$C5,$3F,$C1,$8D,$1E,$62,$99,$8E,$0E,$0F,$1D,$E0,$F3
	.dc.b	$2C,$3D,$37,$E6,$9C,$C7,$82,$7F,$DB,$67,$33,$81,$D7,$0F,$4D,$F0
	.dc.b	$8C,$61,$C5,$DC,$79,$0D,$17,$0F,$79,$F1,$9E,$58,$79,$E3,$EB,$36
	.dc.b	$7E,$4C,$6C,$E2,$FC,$39,$E3,$1E,$B3,$89,$63,$ED,$CF,$37,$E6,$35
	.dc.b	$FB,$3C,$1F,$7D,$FC,$0B,$5F,$A4,$D7,$71,$FE,$D7,$F9,$2F,$F4,$DE
	.dc.b	$DF,$C0,$EB,$84,$62,$91,$8F,$81,$FF,$79,$C6,$2F,$94,$62,$75,$7F
	.dc.b	$5A,$BF,$87,$5F,$FA,$31,$C0,$70,$FE,$3C,$C5,$76,$BF,$CC,$CF,$8D
	.dc.b	$BE,$FF,$FB,$BD,$B5,$D2,$75,$8F,$93,$0F,$C9,$C7,$6C,$B7,$5F,$92
	.dc.b	$8F,$74,$76,$C7,$C1,$7F,$17,$1E,$CD,$BE,$D3,$A7,$DA,$FC,$4A,$EA
	.dc.b	$49,$F1,$4F,$79,$D4,$73,$8F,$15,$7F,$33,$75,$87,$A2,$B7,$1F,$2A
	.dc.b	$75,$7D,$63,$C3,$5B,$3E,$14,$FA,$23,$81,$E8,$88,$F8,$0F,$8B,$1D
	.dc.b	$2F,$CF,$5F,$0E,$B3,$50,$F2,$93,$1F,$96,$4A,$F7,$BC,$63,$B1,$2B
	.dc.b	$07,$F1,$9C,$BF,$CD,$7F,$2D,$F0,$70,$FE,$1B,$18,$AF,$E7,$FB,$AB
	.dc.b	$5A,$FF,$59,$E3,$FE,$D7,$ED,$62,$FF,$2F,$7B,$7F,$9D,$5F,$FA,$AD
	.dc.b	$CB,$70,$ED,$1F,$CF,$7E,$0B,$94,$7F,$9F,$9B,$5F,$EE,$C9,$FE,$F7
	.dc.b	$EA,$E2,$64,$79,$9F,$16,$3F,$DF,$A7,$33,$FD,$36,$6E,$7F,$DF,$E7
	.dc.b	$33,$FD,$2E,$7C,$97,$DE,$3C,$87,$F7,$EE,$EF,$77,$6F,$F7,$D6,$F6
	.dc.b	$9F,$0F,$7B,$D4,$CC,$7C,$1A,$E1,$7C,$C8,$7E,$44,$FF,$58,$53,$C5
	.dc.b	$E6,$48,$FF,$0D,$F9,$15,$FF,$40,$CE,$DF,$F0,$FC,$F8,$A7,$CB,$C3
	.dc.b	$27,$FC,$32,$F1,$DB,$FA,$5E,$1F,$E1,$78,$F3,$67,$FE,$EE,$AF,$F0
	.dc.b	$EC,$60,$CD,$FC,$02,$47,$FA,$EF,$5A,$FF,$A5,$3C,$A7,$C8,$7D,$46
	.dc.b	$C7,$F8,$AA,$BE,$4F,$C4,$C8,$FF,$16,$BF,$6B,$D2,$7C,$F6,$67,$A4
	.dc.b	$F3,$31,$AF,$F6,$48,$7D,$57,$FB,$AD,$FF,$77,$B5,$FF,$46,$36,$3F
	.dc.b	$D9,$DF,$49,$C5,$E5,$1B,$9F,$A8,$D7,$F8,$33,$5F,$F1,$D3,$5F,$0F
	.dc.b	$91,$FA,$35,$F5,$BF,$55,$FF,$4E,$36,$BB,$CE,$FE,$B5,$8D,$25,$FF
	.dc.b	$91,$EB,$DD,$F3,$2F,$8D,$CC,$FF,$A9,$EF,$FC,$1F,$5F,$0D,$BD,$DF
	.dc.b	$C7,$5E,$B3,$FE,$F1,$F0,$BF,$37,$E7,$7E,$7E,$FF,$06,$DB,$FF,$76
	.dc.b	$CF,$D3,$37,$C9,$2F,$91,$F3,$BE,$2B,$E0,$6E,$69,$F0,$FE,$F1,$F0
	.dc.b	$38,$AF,$E0,$95,$B6,$D7,$F0,$77,$ED,$B6,$35,$1F,$EC,$5F,$9E,$7A
	.dc.b	$DF,$3D,$B1,$E3,$3E,$46,$7F,$C5,$F9,$9F,$59,$F0,$35,$95,$A1,$E2
	.dc.b	$2B,$71,$98,$C5,$70,$87,$07,$C8,$BF,$CC,$4F,$F0,$6F,$9C,$7A,$14
	.dc.b	$97,$CF,$1F,$F5,$03,$E1,$F0,$15,$C2,$3C,$1D,$AF,$90,$FD,$51,$CA
	.dc.b	$3F,$DF,$5F,$45,$F6,$BC,$C9,$1D,$63,$F5,$46,$1F,$3E,$FF,$E9,$1C
	.dc.b	$7F,$D5,$F5,$A9,$E0,$78,$F9,$FF,$C3,$5B,$FE,$12,$75,$3C,$8F,$8F
	.dc.b	$A6,$B0,$E9,$F4,$1F,$ED,$C9,$F6,$35,$E1,$C3,$F3,$6F,$F7,$C7,$F2
	.dc.b	$C9,$D6,$1F,$C5,$1B,$7F,$66,$7F,$9E,$FD,$63,$E1,$FF,$86,$7A,$13
	.dc.b	$2F,$AD,$39,$6C,$E5,$FE,$29,$3C,$A3,$F6,$DF,$3E,$F8,$9A,$4A,$FC
	.dc.b	$20,$E7,$FE,$26,$E2,$78,$3F,$D7,$5A,$FF,$3B,$0E,$1F,$E7,$CD,$7F
	.dc.b	$C2,$0F,$0D,$3F,$8D,$2E,$3F,$29,$E3,$9F,$A0,$3E,$9D,$F1,$5F,$C3
	.dc.b	$DB,$1E,$86,$BF,$43,$4F,$96,$FF,$B4,$7E,$85,$F1,$1F,$AF,$6C,$7F
	.dc.b	$77,$6B,$EB,$6B,$FE,$27,$F9,$17,$D6,$3E,$B3,$FC,$05,$18,$6C,$66
	.dc.b	$7C,$9D,$9F,$C9,$7E,$AF,$FC,$BB,$1F,$3B,$EB,$3F,$8E,$37,$C8,$D7
	.dc.b	$B3,$FC,$AE,$73,$3F,$35,$AF,$F6,$96,$BF,$E3,$B6,$BF,$EE,$FA,$7C
	.dc.b	$C7,$84,$91,$DC,$FB,$FF,$58,$F8,$0F,$E7,$4D,$8C,$C7,$E4,$B9,$F2
	.dc.b	$FF,$17,$9D,$4F,$E0,$5A,$FF,$7D,$6B,$FE,$4F,$77,$CD,$F9,$FE,$3F
	.dc.b	$D1,$1D,$27,$C6,$78,$88,$7F,$07,$E6,$F8,$73,$FD,$86,$BF,$94,$49
	.dc.b	$A4,$7C,$C7,$EB,$F8,$A3,$D8,$FC,$FF,$D6,$7E,$ED,$07,$0D,$EB,$FF
	.dc.b	$20,$7D,$1B,$ED,$DF,$71,$FD,$86,$FF,$D4,$17,$FE,$6D,$6B,$D1,$AF
	.dc.b	$33,$FC,$8C,$F8,$6B,$FC,$03,$EB,$76,$3F,$63,$3F,$A2,$9F,$19,$F8
	.dc.b	$A7,$1C,$7E,$13,$CB,$87,$EA,$E3,$DB,$1F,$D3,$67,$84,$7F,$8C,$5A
	.dc.b	$F9,$E3,$47,$87,$E5,$A4,$72,$3F,$82,$F8,$EF,$AD,$FF,$00,$36,$33
	.dc.b	$1F,$1B,$EE,$1F,$51,$A8,$FE,$BD,$5B,$AF,$E1,$0D,$7F,$C8,$8D,$7F
	.dc.b	$D2,$CE,$F8,$5E,$2F,$D6,$D9,$F5,$CF,$8B,$FA,$AD,$C3,$C6,$7A,$3B
	.dc.b	$3C,$E4,$7F,$B1,$53,$47,$C7,$66,$7C,$B5,$E8,$23,$FC,$7F,$CA,$78
	.dc.b	$9F,$82,$8F,$EB,$9F,$59,$FE,$63,$6F,$C3,$7F,$3A,$E7,$91,$82,$1F
	.dc.b	$C5,$DF,$F3,$7F,$D7,$7E,$7B,$7F,$EB,$0D,$97,$A0,$8C,$64,$F8,$6F
	.dc.b	$F2,$7D,$7F,$57,$AD,$69,$D0,$91,$FD,$F6,$77,$C5,$FE,$A2,$F3,$FF
	.dc.b	$90,$B7,$CA,$FF,$4A,$C6,$33,$FF,$51,$66,$E5,$CD,$4C,$7F,$C1,$9D
	.dc.b	$2C,$38,$47,$F8,$3A,$F8,$3A,$8F,$E5,$AF,$EC,$27,$85,$7F,$A9,$E7
	.dc.b	$32,$9C,$BF,$D3,$DE,$DF,$15,$E6,$24,$78,$BC,$A7,$D6,$62,$3F,$BD
	.dc.b	$57,$07,$C5,$5F,$EB,$2A,$DC,$7F,$62,$FA,$E7,$A9,$FA,$68,$46,$2F
	.dc.b	$FD,$7C,$D8,$FE,$C4,$D7,$FD,$3D,$F1,$FF,$E7,$CE,$AF,$F6,$23,$63
	.dc.b	$FB,$43,$5D,$F8,$49,$EB,$E2,$7F,$2A,$9F,$0B,$6C,$D3,$F9,$7E,$0C
	.dc.b	$1F,$0D,$FA,$8B,$F8,$3E,$3B,$E0,$B1,$CB,$FE,$E2,$BD,$F2,$FD,$AA
	.dc.b	$D3,$F7,$23,$37,$CA,$F8,$CF,$D9,$9C,$DF,$FB,$B2,$F8,$DF,$83,$D2
	.dc.b	$DB,$F6,$E3,$FB,$DD,$FF,$93,$FD,$9F,$EA,$4F,$86,$B7,$D7,$FE,$46
	.dc.b	$F3,$38,$EB,$FE,$0E,$9C,$DF,$0C,$7B,$61,$CE,$F7,$8F,$69,$FF,$19
	.dc.b	$D6,$F2,$F3,$61,$DB,$A9,$F5,$DE,$6E,$D7,$81,$FE,$BC,$D3,$F5,$13
	.dc.b	$98,$C6,$5F,$ED,$48,$C7,$C4,$73,$6D,$7F,$97,$3A,$BD,$7E,$B7,$E5
	.dc.b	$23,$FC,$BF,$E4,$E2,$64,$6C,$C9,$70,$FF,$11,$FD,$4F,$FB,$C6,$3F
	.dc.b	$C6,$ED,$8F,$F1,$3F,$D6,$FF,$30,$9C,$FC,$77,$D0,$94,$E5,$E7,$19
	.dc.b	$8C,$70,$7F,$B3,$6F,$A3,$7E,$5F,$94,$1E,$1E,$4E,$19,$8F,$91,$CB
	.dc.b	$4F,$F2,$6D,$FF,$2E,$71,$8F,$A2,$B8,$BE,$EF,$89,$EE,$7A,$7E,$B1
	.dc.b	$F5,$98,$BF,$83,$36,$BF,$C8,$F5,$A9,$C9,$FE,$5C,$71,$1D,$63,$15
	.dc.b	$C7,$3F,$81,$F0,$5E,$E3,$C1,$F3,$B6,$BE,$7C,$5B,$EF,$36,$BF,$DE
	.dc.b	$98,$23,$17,$FE,$0C,$A4,$E6,$39,$9F,$E7,$0F,$AC,$7C,$BF,$D7,$6C
	.dc.b	$7F,$A9,$FE,$AD,$F1,$D9,$A9,$3C,$5F,$CC,$4E,$A3,$07,$19,$DE,$1F
	.dc.b	$D1,$67,$4F,$F9,$BA,$FB,$B7,$E2,$7C,$47,$E6,$3E,$32,$E1,$E7,$8B
	.dc.b	$3F,$E9,$EE,$7F,$F7,$14,$E4,$FF,$5D,$D6,$B5,$FE,$25,$E1,$7F,$9F
	.dc.b	$E3,$13,$F1,$7E,$AF,$F3,$58,$7C,$70,$E1,$D7,$3F,$F1,$B7,$51,$97
	.dc.b	$FC,$72,$FE,$62,$76,$7C,$9F,$81,$FD,$14,$66,$7F,$B5,$EB,$FE,$66
	.dc.b	$AD,$1F,$EE,$5F,$86,$7E,$03,$F8,$8D,$3F,$95,$78,$0B,$E1,$87,$F6
	.dc.b	$FF,$9D,$7C,$FF,$F0,$96,$E5,$FE,$AC,$FA,$BF,$F7,$CB,$FE,$E1,$8C
	.dc.b	$D7,$FA,$7A,$72,$DF,$90,$FE,$4F,$F5,$6F,$AA,$7D,$5F,$EE,$EF,$7A
	.dc.b	$FC,$CE,$5F,$D6,$63,$FE,$D7,$C3,$85,$FF,$10,$EF,$E7,$8F,$2D,$3F
	.dc.b	$D7,$FE,$67,$F3,$87,$F2,$CD,$C3,$FC,$C7,$F5,$6F,$96,$FE,$15,$39
	.dc.b	$9F,$F4,$47,$A5,$83,$F8,$75,$6E,$FA,$C6,$2F,$51,$FE,$CE,$9D,$57
	.dc.b	$FA,$FA,$FC,$1D,$5F,$E5,$9A,$FF,$39,$ED,$ED,$AF,$E5,$35,$FC,$A9
	.dc.b	$93,$A3,$FC,$F3,$5C,$63,$F2,$93,$A7,$FD,$E7,$F2,$9F,$D5,$2F,$F2
	.dc.b	$6D,$BE,$2F,$92,$33,$E9,$FF,$52,$8D,$53,$04,$61,$B9,$6E,$B8,$7C
	.dc.b	$84,$C7,$CD,$83,$EE,$DF,$A6,$FF,$9A,$E7,$38,$F4,$BE,$17,$E8,$7F
	.dc.b	$B4,$DF,$0F,$89,$CE,$9E,$A9,$B8,$7F,$A6,$AB,$4F,$4D,$13,$EB,$F8
	.dc.b	$0F,$EB,$DD,$2F,$95,$C5,$F9,$BB,$3F,$37,$18,$D3,$FC,$99,$5B,$AE
	.dc.b	$97,$CA,$BE,$37,$C0,$7C,$7F,$5A,$D7,$85,$B8,$7D,$26,$EB,$F4,$9F
	.dc.b	$C4,$FC,$5C,$AF,$8F,$4F,$AB,$F0,$2B,$FE,$05,$CC,$BE,$13,$9B,$FE
	.dc.b	$61,$39,$79,$DF,$2B,$E3,$7F,$D9,$63,$13,$C2,$31,$97,$4A,$FC,$3F
	.dc.b	$D6,$FE,$0C,$FC,$27,$DB,$5B,$F9,$D1,$72,$FC,$7C,$DA,$FE,$31,$F5
	.dc.b	$CD,$BF,$BA,$3F,$8C,$A3,$9C,$3E,$0B,$93,$11,$ED,$3F,$DA,$EE,$67
	.dc.b	$9F,$3F,$F7,$FF,$A8,$78,$14,$FF,$15,$FD,$6B,$EA,$F1,$97,$CE,$AD
	.dc.b	$B6,$B7,$3A,$CB,$C1,$D4,$C3,$95,$EC,$FC,$CB,$89,$F3,$0C,$3A,$BF
	.dc.b	$B2,$F1,$3F,$23,$E9,$7E,$B5,$B7,$F9,$76,$CE,$E3,$F2,$9C,$FF,$DA
	.dc.b	$63,$F7,$7F,$59,$F9,$AF,$AD,$FD,$B3,$6B,$F4,$0D,$9D,$C7,$29,$FD
	.dc.b	$0F,$C4,$C8,$FF,$52,$79,$F6,$BD,$FF,$03,$FD,$1F,$B7,$F9,$ED,$FF
	.dc.b	$91,$EB,$5C,$BF,$D7,$2B,$FC,$49,$87,$23,$FD,$5D,$ED,$ED,$A6,$71
	.dc.b	$EB,$19,$7F,$D7,$1F,$51,$F2,$6F,$FA,$7B,$6B,$F8,$ED,$BE,$97,$FC
	.dc.b	$AD,$39,$BF,$E2,$93,$9A,$FD,$D4,$E7,$8F,$FD,$8D,$7C,$39,$3F,$CD
	.dc.b	$7E,$77,$F4,$8C,$BF,$D0,$DF,$0C,$8C,$63,$E0,$FA,$DD,$F9,$BF,$E6
	.dc.b	$76,$C7,$FB,$2D,$BA,$FC,$86,$E1,$F1,$5B,$3C,$CF,$F6,$3D,$F6,$AF
	.dc.b	$4C,$E9,$FF,$60,$7D,$5F,$F1,$8A,$FF,$66,$B6,$3E,$AF,$CC,$7F,$A1
	.dc.b	$23,$18,$7D,$38,$C7,$8F,$FE,$7D,$6D,$FF,$0C,$B1,$F9,$51,$F5,$AF
	.dc.b	$A4,$FF,$09,$79,$FF,$EA,$98,$C4,$FF,$85,$7B,$78,$C6,$23,$9C,$EE
	.dc.b	$BF,$C3,$51,$D6,$74,$FF,$A5,$DA,$E6,$3B,$DF,$47,$8D,$D9,$FE,$EB
	.dc.b	$35,$FF,$26,$45,$7A,$4F,$07,$D6,$F4,$69,$E9,$29,$E8,$9E,$4F,$8C
	.dc.b	$EE,$68,$F8,$F3,$DE,$FD,$D1,$CE,$FD,$B1,$88,$F2,$9C,$5C,$3C,$8C
	.dc.b	$57,$82,$BA,$36,$32,$78,$4F,$21,$E7,$31,$1C,$1A,$70,$AF,$9B,$19
	.dc.b	$6C,$66,$FE,$34,$7A,$2B,$C0,$D9,$E6,$D3,$89,$67,$E7,$6E,$F9,$DA
	.dc.b	$4A,$F9,$D1,$97,$F1,$FF,$1B,$FE,$7E,$D7,$F2,$33,$99,$FF,$7C,$7C
	.dc.b	$07,$F5,$D5,$19,$AE,$11,$8B,$FF,$46,$46,$27,$55,$C7,$1F,$CA,$4B
	.dc.b	$DE,$FF,$49,$46,$1B,$97,$2A,$E0,$DC,$BF,$1D,$7C,$72,$B7,$C1,$F5
	.dc.b	$DF,$06,$D6,$98,$7E,$3A,$BD,$38,$79,$2B,$84,$A7,$B3,$BB,$EC,$FC
	.dc.b	$4B,$F1,$D7,$0A,$DD,$FA,$EB,$35,$1E,$47,$8A,$3C,$11,$97,$84,$68
	.dc.b	$F3,$57,$53,$B5,$F8,$AF,$BE,$1F,$16,$B7,$87,$58,$FA,$75,$BB,$E5
	.dc.b	$39,$AF,$C0,$CE,$63,$C5,$5C,$27,$47,$BC,$EA,$79,$70,$F3,$9E,$A7
	.dc.b	$F7,$7C,$CE,$87,$D5,$FF,$B3,$F0,$FE,$77,$C6,$F3,$F1,$C1,$F9,$08
	.dc.b	$C6,$3F,$E5,$0C,$38,$BF,$CB,$AB,$6F,$F3,$26,$EE,$AE,$33,$FD,$9B
	.dc.b	$1D,$E5,$E3,$FA,$DD,$D7,$FB,$62,$BF,$9C,$FC,$16,$B2,$55,$37,$12
	.dc.b	$7C,$91,$97,$89,$0F,$EE,$AE,$DF,$F1,$26,$3E,$59,$C6,$9F,$DD,$6B
	.dc.b	$C1,$F3,$5E,$CF,$05,$DC,$F0,$30,$DC,$BE,$25,$62,$64,$7F,$6C,$79
	.dc.b	$BF,$27,$D8,$F6,$7D,$6C,$8C,$10,$FF,$05,$B9,$7F,$C1,$B1,$72,$FE
	.dc.b	$FF,$C1,$E2,$31,$C5,$FD,$5B,$3F,$3C,$66,$FF,$BE,$BA,$7F,$D2,$9E
	.dc.b	$AF,$CF,$7F,$C1,$78,$7B,$0F,$AB,$19,$C3,$FD,$43,$B6,$6B,$FC,$0F
	.dc.b	$5A,$8D,$FC,$AF,$C4,$F8,$99,$5F,$E8,$B8,$CB,$FE,$0E,$8C,$BF,$EA
	.dc.b	$E8,$FC,$1F,$D6,$FF,$BD,$F0,$FF,$7C,$7D,$57,$F8,$66,$BF,$CF,$AD
	.dc.b	$7F,$C9,$B3,$BD,$7F,$BA,$C2,$75,$5C,$BC,$6F,$CA,$CE,$5E,$F5,$8A
	.dc.b	$DB,$C9,$B3,$FE,$00,$6F,$3B,$0F,$04,$FF,$81,$3D,$07,$CE,$76,$FA
	.dc.b	$C6,$B4,$F0,$CE,$AB,$FD,$2E,$D7,$FC,$C3,$3C,$1F,$F4,$DB,$5F,$F3
	.dc.b	$2C,$6F,$4F,$F6,$9F,$77,$D6,$24,$FF,$AE,$7B,$EB,$FE,$DA,$8B,$3F
	.dc.b	$ED,$A9,$F8,$7F,$59,$FE,$F7,$AF,$9B,$F1,$5F,$0C,$CF,$C0,$D8,$C9
	.dc.b	$0F,$F3,$BC,$F1,$7F,$CF,$3B,$AB,$84,$61,$E6,$FF,$B3,$7D,$BF,$C6
	.dc.b	$FD,$64,$8C,$10,$FF,$78,$B9,$7F,$DE,$71,$F9,$4F,$4F,$F6,$9F,$2F
	.dc.b	$FE,$D6,$9F,$F6,$93,$5F,$F7,$CC,$6D,$CF,$99,$C0,$C3,$FE,$FE,$F9
	.dc.b	$FF,$CF,$91,$FF,$78,$FC,$13,$EF,$DD,$59,$9E,$07,$FB,$9E,$BF,$90
	.dc.b	$D6,$E7,$5C,$2E,$9F,$F2,$4F,$4B,$B5,$86,$A7,$FD,$E5,$97,$6C,$BF
	.dc.b	$DE,$5E,$67,$EC,$88,$FF,$BD,$7E,$3A,$BF,$EA,$EC,$3F,$C0,$99,$77
	.dc.b	$8F,$DE,$D7,$EE,$AB,$F9,$7F,$D1,$7F,$BE,$67,$FC,$B7,$5B,$CD,$39
	.dc.b	$9F,$F5,$CB,$67,$58,$7F,$83,$A7,$8E,$5E,$CA,$D3,$7E,$0B,$B9,$FC
	.dc.b	$2B,$C9,$FD,$8F,$D6,$FB,$6F,$FD,$7F,$F3,$8F,$AA,$FD,$AB,$63,$FD
	.dc.b	$F7,$3B,$CF,$D9,$09,$D3,$CA,$B8,$9D,$B8,$6C,$79,$8F,$F3,$CC,$F0
	.dc.b	$AE,$6D,$CB,$F8,$4E,$1D,$DF,$67,$D7,$FF,$08,$FA,$B7,$C5,$53,$E5
	.dc.b	$C7,$F0,$CF,$A0,$FF,$7B,$65,$DC,$38,$7F,$DB,$AD,$CB,$D7,$7D,$27
	.dc.b	$83,$F9,$98,$CF,$3F,$84,$ED,$F0,$DF,$C6,$BE,$B3,$E9,$5F,$FB,$A1
	.dc.b	$B5,$F1,$9B,$9F,$DE,$FB,$E7,$C0,$FF,$22,$FA,$EF,$E1,$53,$FC,$A1
	.dc.b	$B8,$7F,$A7,$9A,$AF,$9B,$9F,$2B,$FD,$2E,$D2,$CF,$F1,$B7,$F9,$98
	.dc.b	$70,$FE,$1F,$8F,$F9,$9F,$D7,$3D,$EA,$7E,$5A,$2C,$FC,$B9,$D3,$67
	.dc.b	$4D,$20,$D3,$F0,$F3,$C2,$3F,$37,$5F,$DE,$F9,$3D,$B3,$E6,$AD,$D6
	.dc.b	$71,$F2,$B0,$F1,$19,$29,$0F,$06,$7F,$D4,$AB,$A3,$FD,$CE,$FE,$6F
	.dc.b	$D3,$3E,$B9,$F5,$7F,$D5,$3E,$BB,$F7,$37,$FD,$69,$B8,$AF,$85,$6F
	.dc.b	$C0,$93,$F8,$FA,$DD,$6A,$31,$3A,$7C,$A7,$FB,$6A,$77,$87,$8A,$3F
	.dc.b	$4D,$5E,$5F,$8E,$F2,$76,$49,$CD,$27,$11,$D0,$5F,$E1,$9B,$5F,$E4
	.dc.b	$4B,$E9,$1E,$78,$E3,$B7,$EF,$2B,$F1,$B5,$AA,$F3,$AF,$AD,$7D,$37
	.dc.b	$F7,$AF,$15,$F2,$D4,$ED,$3A,$BF,$D9,$4E,$9B,$3B,$7D,$F3,$B7,$D5
	.dc.b	$3B,$6C,$ED,$F1,$CE,$F1,$FE,$5D,$87,$97,$E2,$BC,$9C,$72,$FA,$C5
	.dc.b	$7A,$C5,$AE,$36,$AF,$E6,$BC,$12,$FF,$8C,$78,$BF,$CD,$23,$F9,$CF
	.dc.b	$DF,$66,$7F,$9A,$C5,$C3,$F9,$D7,$8A,$AF,$F3,$47,$D4,$FF,$96,$7E
	.dc.b	$BB,$F7,$BE,$F1,$FD,$FE,$5F,$E8,$39,$DE,$5F,$C9,$EB,$5A,$F1,$33
	.dc.b	$1F,$8A,$AE,$5C,$1F,$E7,$EF,$AE,$FE,$39,$D1,$FE,$D6,$6D,$2B,$6F
	.dc.b	$B3,$2F,$EC,$2F,$F3,$5E,$55,$70,$6F,$83,$FE,$86,$AD,$D7,$F9,$E2
	.dc.b	$B4,$49,$D6,$3F,$E9,$0F,$AE,$7D,$5F,$97,$E2,$66,$1D,$E7,$2D,$9C
	.dc.b	$BD,$67,$94,$7F,$92,$3E,$B9,$F2,$7F,$CC,$2B,$FC,$9F,$18,$D3,$F9
	.dc.b	$1B,$63,$F9,$08,$70,$FF,$24,$F8,$EF,$F0,$5F,$4F,$F2,$A6,$C7,$06
	.dc.b	$BB,$F3,$BF,$81,$3D,$2B,$FA,$6F,$89,$E0,$F2,$FF,$8F,$4F,$91,$A8
	.dc.b	$F1,$38,$C7,$F6,$FE,$47,$F8,$1F,$AB,$FC,$05,$F1,$7F,$E5,$8E,$6F
	.dc.b	$A9,$E0,$FF,$48,$7F,$D8,$1F,$03,$B7,$A6,$1C,$9C,$9C,$30,$D6,$5F
	.dc.b	$81,$C3,$A3,$FE,$B0,$FD,$17,$F7,$75,$F9,$EE,$67,$B3,$C3,$15,$6A
	.dc.b	$FF,$21,$5A,$BF,$C4,$4E,$74,$FF,$0D,$4E,$6B,$B7,$89,$E2,$EB,$4E
	.dc.b	$6B,$FC,$4D,$8F,$6A,$D4,$E7,$1F,$6E,$3D,$5F,$F1,$CC,$67,$17,$6F
	.dc.b	$CC,$9F,$F7,$27,$D5,$7C,$08,$C4,$FE,$9F,$54,$E7,$8F,$FC,$0B,$7C
	.dc.b	$0F,$D8,$38,$AE,$13,$F2,$BD,$87,$E7,$3F,$2B,$6F,$F1,$F4,$63,$8B
	.dc.b	$DD,$E1,$BE,$6B,$85,$EF,$3F,$C1,$5E,$DF,$F2,$EC,$67,$2F,$DB,$5F
	.dc.b	$19,$CE,$1F,$6A,$5F,$5C,$F8,$EF,$BB,$D8,$FC,$8D,$F0,$DB,$AD,$63
	.dc.b	$FD,$32,$72,$FF,$89,$63,$38,$7F,$3E,$9C,$BF,$D2,$E3,$3B,$7F,$2E
	.dc.b	$9C,$BF,$E4,$28,$7C,$67,$8A,$5F,$01,$E1,$8C,$37,$0F,$13,$70,$F0
	.dc.b	$37,$0E,$CD,$9F,$D4,$65,$ED,$8F,$67,$77,$C0,$FC,$73,$F8,$6D,$7A
	.dc.b	$E3,$1E,$1E,$A3,$FB,$5C,$EB,$1F,$65,$6A,$78,$1A,$AE,$13,$FD,$82
	.dc.b	$B8,$63,$EF,$8F,$9F,$C2,$F8,$7F,$9D,$18,$6E,$1F,$DE,$A7,$4F,$CF
	.dc.b	$9D,$34,$D4,$F0,$8F,$ED,$F3,$C1,$FE,$F9,$F8,$3F,$F6,$EA,$FF,$35
	.dc.b	$64,$FC,$27,$FD,$33,$1F,$E5,$F9,$D5,$7F,$83,$AB,$5E,$4F,$E2,$67
	.dc.b	$39,$7F,$A5,$67,$37,$C7,$D2,$FE,$0F,$18,$C3,$C3,$18,$C7,$BC,$62
	.dc.b	$BC,$71,$8B,$FF,$42,$FD,$73,$E4,$FF,$88,$C3,$EE,$7F,$D1,$D7,$FE
	.dc.b	$75,$AD,$3F,$E8,$59,$D7,$B7,$F9,$F9,$CE,$3F,$E2,$89,$CD,$71,$C3
	.dc.b	$AE,$1D,$2B,$FC,$45,$7C,$1F,$F3,$0B,$5F,$E6,$41,$C3,$FC,$D5,$B9
	.dc.b	$BF,$2B,$F8,$F8,$7F,$AB,$7E,$93,$E2,$C6,$2F,$F6,$F1,$FB,$5F,$9A
	.dc.b	$FE,$2D,$1F,$E6,$B9,$D2,$8C,$DF,$5D,$B8,$78,$D3,$1F,$BB,$83,$D4
	.dc.b	$C6,$3F,$DF,$FF,$5D,$7A,$DF,$C3,$BE,$F3,$F9,$64,$63,$0F,$F0,$BF
	.dc.b	$84,$9F,$E1,$1E,$F3,$E2,$7E,$EF,$EE,$6F,$A0,$CC,$61,$CF,$CD,$63
	.dc.b	$1F,$C7,$D2,$31,$A7,$F9,$3F,$E8,$1F,$5D,$FD,$3B,$F2,$57,$D6,$BE
	.dc.b	$B3,$F8,$C5,$6B,$2F,$F6,$15,$6A,$73,$7C,$6F,$85,$7F,$AE,$6F,$86
	.dc.b	$1F,$E6,$8C,$38,$4E,$6F,$86,$1C,$A3,$AE,$1C,$AF,$FD,$D3,$F5,$1F
	.dc.b	$CD,$2F,$F9,$AB,$6B,$E3,$B6,$77,$1F,$2A,$7F,$9D,$FB,$DF,$CF,$3E
	.dc.b	$02,$CF,$CE,$6D,$7F,$BD,$9B,$DF,$FE,$42,$7F,$A1,$FD,$AF,$F4,$4F
	.dc.b	$7A,$DF,$D6,$6D,$FF,$BF,$9B,$7F,$E4,$A6,$CF,$FB,$F5,$B3,$FD,$19
	.dc.b	$FE,$BD,$F9,$4B,$E4,$3F,$A8,$F6,$E1,$B7,$FE,$4B,$9D,$3F,$86,$9D
	.dc.b	$34,$FF,$2F,$F4,$BE,$AB,$FA,$DC,$7E,$39,$AE,$9B,$8F,$E5,$5B,$92
	.dc.b	$7F,$5F,$C5,$F8,$F7,$FC,$45,$19,$AD,$7E,$DB,$F9,$97,$FD,$35,$5B
	.dc.b	$7F,$D4,$F4,$9D,$57,$19,$56,$E3,$A5,$6F,$B3,$FB,$44,$E6,$3F,$C6
	.dc.b	$46,$4F,$19,$E6,$E8,$FF,$13,$B9,$BF,$1C,$63,$0F,$0C,$FF,$8E,$FD
	.dc.b	$F7,$D6,$C8,$FF,$43,$BF,$91,$CB,$DD,$3A,$8F,$F1,$11,$99,$E4,$F1
	.dc.b	$C3,$FC,$F8,$E5,$FF,$3C,$39,$AE,$6E,$7D,$D7,$E6,$DF,$E8,$FB,$FB
	.dc.b	$3E,$1F,$F8,$86,$31,$87,$F0,$98,$C7,$C5,$7F,$08,$6D,$70,$3E,$1C
	.dc.b	$FC,$6F,$86,$F8,$7E,$DF,$F5,$3C,$FE,$85,$27,$3E,$97,$F8,$56,$30
	.dc.b	$72,$15,$C5,$FE,$62,$6A,$FE,$EC,$65,$F0,$47,$18,$C4,$7E,$2E,$BB
	.dc.b	$CF,$9A,$1F,$6F,$FE,$37,$C3,$E7,$F8,$FF,$E1,$4A,$D4,$E5,$FB,$11
	.dc.b	$9A,$FF,$74,$4E,$5F,$F5,$34,$66,$BF,$1D,$39,$DF,$F9,$A7,$B1,$F3
	.dc.b	$7B,$BF,$98,$FA,$D7,$2F,$D4,$B7,$2F,$E7,$0D,$C7,$F0,$7D,$1F,$E0
	.dc.b	$CA,$FF,$42,$71,$3A,$9F,$63,$FE,$67,$F8,$27,$D6,$FE,$FA,$30,$DB
	.dc.b	$FF,$6A,$CE,$9F,$DF,$4E,$9F,$F3,$E7,$CD,$6F,$4F,$F3,$CB,$73,$FF
	.dc.b	$3B,$B6,$7F,$CD,$AD,$9F,$F3,$6E,$3F,$5E,$BF,$CB,$FF,$81,$FF,$6F
	.dc.b	$7F,$75,$9F,$A7,$8F,$FC,$CD,$39,$C7,$DB,$7C,$5F,$F3,$0C,$67,$0E
	.dc.b	$B5,$C3,$D6,$FF,$48,$7E,$37,$FD,$66,$FD,$BD,$18,$F3,$FE,$23,$63
	.dc.b	$27,$FB,$66,$78,$61,$FE,$7F,$9D,$36,$75,$5C,$9A,$72,$86,$9F,$E9
	.dc.b	$4B,$E6,$FF,$AD,$2F,$CB,$F0,$9F,$57,$8E,$E1,$F4,$9B,$87,$FA,$3D
	.dc.b	$B7,$FE,$BC,$7F,$CF,$51,$FE,$C3,$F3,$3F,$D8,$9E,$A7,$DE,$8F,$F0
	.dc.b	$AF,$A9,$FC,$96,$31,$1F,$E1,$89,$DF,$2F,$F8,$6A,$15,$B8,$FF,$6F
	.dc.b	$CE,$A7,$9E,$35,$E4,$7F,$84,$E7,$95,$71,$8E,$95,$B7,$AD,$55,$19
	.dc.b	$9E,$11,$D2,$FC,$EF,$03,$B1,$D0,$A7,$FB,$8E,$38,$BF,$0C,$C9,$C6
	.dc.b	$39,$BE,$03,$31,$C1,$C4,$78,$23,$13,$E5,$AF,$43,$6B,$A9,$C2,$FE
	.dc.b	$0C,$EA,$3A,$B6,$FA,$36,$77,$3C,$AB,$A5,$79,$9B,$3F,$4A,$79,$9B
	.dc.b	$2C,$ED,$B1,$B8,$EF,$1B,$69,$BF,$89,$7A,$7D,$5B,$3F,$35,$AF,$DE
	.dc.b	$6F,$C6,$BE,$9B,$F3,$B2,$9C,$C6,$1B,$AA,$75,$7F,$E6,$8B,$E5,$39
	.dc.b	$AD,$D7,$DC,$9F,$F6,$8B,$F8,$F6,$B9,$BF,$0E,$1F,$9C,$9E,$73,$C7
	.dc.b	$E1,$7F,$CE,$B3,$96,$E5,$E1,$BE,$2D,$C7,$B4,$EA,$FD,$18,$78,$30
	.dc.b	$ED,$87,$09,$CB,$54,$EF,$0F,$75,$F3,$9C,$D6,$EB,$CF,$7E,$59,$F0
	.dc.b	$9E,$F9,$D1,$BA,$E5,$5D,$67,$C3,$86,$CE,$2F,$C1,$30,$7C,$77,$37
	.dc.b	$DE,$30,$47,$04,$C3,$E2,$46,$5F,$5B,$C9,$CC,$7C,$49,$C4,$FC,$8B
	.dc.b	$F3,$E1,$EA,$9C,$CF,$C6,$9C,$C7,$F0,$DC,$5C,$2F,$47,$FD,$5B,$3F
	.dc.b	$E2,$0E,$B7,$D2,$63,$83,$FC,$51,$18,$CB,$F9,$C5,$E5,$FC,$A6,$1C
	.dc.b	$DF,$EB,$B7,$9F,$01,$7C,$67,$F9,$BE,$1F,$93,$BE,$11,$83,$9F,$E1
	.dc.b	$1C,$69,$F2,$B4,$F0,$69,$EA,$9B,$67,$57,$FD,$BA,$B3,$12,$30,$43
	.dc.b	$C1,$5C,$DF,$EA,$91,$C9,$F2,$63,$EA,$AE,$51,$E7,$ED,$FF,$01,$6D
	.dc.b	$F0,$FE,$29,$F8,$5F,$F3,$EA,$FF,$1A,$C6,$39,$FE,$45,$F0,$9F,$F3
	.dc.b	$F5,$F1,$BE,$11,$F0,$EF,$83,$A3,$FC,$7F,$8B,$E3,$BF,$83,$47,$FA
	.dc.b	$1F,$E8,$FA,$F8,$27,$E9,$4E,$A7,$D6,$DE,$4F,$F0,$1D,$6E,$79,$1D
	.dc.b	$A7,$5C,$5F,$E0,$4B,$FA,$9A,$CA,$D9,$F0,$EF,$93,$F2,$38,$BD,$78
	.dc.b	$FE,$36,$BC,$B7,$F0,$B9,$BF,$B3,$CF,$F8,$BA,$7F,$11,$E2,$70,$CB
	.dc.b	$FC,$5F,$5A,$6F,$46,$AA,$65,$FE,$62,$E1,$FE,$AF,$8F,$FA,$3F,$F6
	.dc.b	$87,$C7,$FE,$4F,$F7,$4F,$5D,$B8,$51,$88,$57,$07,$FC,$B0,$D7,$F0
	.dc.b	$71,$CA,$78,$F5,$FE,$8E,$78,$DD,$FB,$D7,$9D,$9F,$A9,$7B,$14,$D4
	.dc.b	$4B,$FF,$33,$4E,$9B,$39,$7F,$4C,$2B,$AB,$A2,$1A,$72,$FE,$5F,$F2
	.dc.b	$1F,$C4,$3F,$EC,$83,$E7,$9F,$94,$72,$EF,$9F,$2E,$CF,$9D,$1A,$DB
	.dc.b	$FD,$4F,$FA,$B3,$F0,$2F,$E0,$0B,$C9,$A6,$22,$5F,$E6,$A7,$4D,$9C
	.dc.b	$BE,$EF,$98,$CB,$F0,$3E,$B9,$F6,$7F,$3D,$B1,$FE,$8D,$9D,$E3,$FE
	.dc.b	$CF,$6F,$47,$F6,$59,$57,$07,$A5,$FD,$9A,$DC,$7F,$6B,$6B,$C2,$3C
	.dc.b	$06,$F0,$FC,$BF,$C7,$7E,$63,$F0,$49,$5F,$EF,$48,$CB,$FE,$F6,$8C
	.dc.b	$BF,$C3,$3B,$13,$FC,$13,$F2,$E7,$A1,$FC,$A7,$E3,$1E,$CD,$35,$13
	.dc.b	$0F,$E0,$73,$A6,$CE,$5E,$DF,$75,$FA,$CF,$B6,$91,$FC,$2B,$24,$FE
	.dc.b	$C0,$CC,$79,$67,$0E,$E3,$3C,$3F,$EA,$49,$CC,$FA,$FE,$33,$F7,$3F
	.dc.b	$05,$23,$B4,$7E,$F0,$C3,$F0,$63,$0F,$8F,$E7,$3F,$80,$FD,$CC,$8C
	.dc.b	$10,$F9,$0E,$5F,$95,$1F,$67,$97,$FA,$A7,$C2,$FF,$89,$FE,$4A,$FA
	.dc.b	$27,$57,$F8,$7A,$CC,$48,$C1,$0F,$EA,$95,$CD,$FF,$77,$B5,$FC,$B4
	.dc.b	$FF,$01,$F0,$7F,$AF,$6A,$8F,$C8,$7E,$1C,$F8,$8F,$E0,$CD,$BF,$F7
	.dc.b	$25,$F0,$AD,$D7,$3F,$72,$57,$FB,$A6,$F8,$3F,$89,$F2,$FF,$0D,$DF
	.dc.b	$23,$EE,$47,$F1,$3F,$8E,$7C,$C5,$24,$FF,$BC,$FF,$75,$7C,$46,$23
	.dc.b	$F8,$6C,$3E,$C7,$F8,$E3,$63,$FD,$F5,$7D,$1E,$53,$F9,$7F,$AA,$91
	.dc.b	$83,$FE,$00,$7D,$77,$E5,$5F,$F5,$10,$70,$FF,$A9,$5A,$FE,$DD,$AF
	.dc.b	$F9,$75,$E5,$C3,$F3,$63,$FC,$C1,$F5,$FF,$D4,$23,$10,$8C,$70,$F3
	.dc.b	$ED,$FE,$6D,$7C,$A3,$F8,$24,$EA,$B9,$FA,$BA,$BF,$E7,$9F,$5E,$FA
	.dc.b	$7F,$E7,$DC,$AE,$5A,$6A,$26,$1F,$E7,$3D,$3C,$75,$B8,$C6,$7F,$C4
	.dc.b	$E3,$11,$FC,$F3,$0E,$CE,$8F,$F3,$E3,$87,$FA,$0C,$70,$AE,$B8,$78
	.dc.b	$27,$FA,$2F,$D4,$7F,$5E,$FA,$47,$E4,$8F,$23,$FC,$71,$18,$7F,$AC
	.dc.b	$4F,$F6,$13,$84,$EE,$BC,$55,$B7,$C8,$27,$87,$A9,$D2,$B8,$4F,$F8
	.dc.b	$EE,$B8,$3E,$49,$EF,$EF,$EF,$C2,$93,$D3,$0F,$F4,$8E,$7F,$69,$BB
	.dc.b	$7F,$A7,$71,$F5,$DF,$0B,$F7,$E1,$D1,$F3,$E4,$9D,$63,$CE,$BF,$BB
	.dc.b	$74,$FF,$2E,$BF,$6F,$E4,$9F,$D0,$31,$3F,$D0,$27,$53,$FD,$C1,$B7
	.dc.b	$F9,$1A,$DF,$99,$F3,$2F,$87,$95,$D8,$C4,$4C,$3E,$4E,$1C,$B4,$FE
	.dc.b	$A8,$FE,$CA,$3C,$7E,$A7,$33,$CF,$DB,$23,$24,$3C,$8F,$F9,$47,$E9
	.dc.b	$3F,$AF,$47,$F9,$25,$BA,$7E,$9F,$E0,$11,$8C,$7F,$D9,$1E,$CB,$CD
	.dc.b	$7D,$E7,$3C,$61,$FE,$5A,$BE,$0F,$F8,$A6,$76,$FF,$8C,$7E,$A5,$E6
	.dc.b	$FF,$9E,$62,$19,$7F,$CC,$31,$A8,$FF,$2D,$FC,$0F,$4C,$7F,$D0,$F1
	.dc.b	$AA,$FF,$3F,$FB,$1F,$C8,$35,$FE,$01,$4F,$DD,$BF,$9E,$5F,$F9,$91
	.dc.b	$AF,$F9,$F2,$B8,$6B,$FE,$E1,$6C,$7F,$90,$67,$77,$D1,$B5,$BE,$DF
	.dc.b	$09,$FE,$9B,$8C,$D7,$1F,$37,$2F,$ED,$FC,$EF,$F5,$4F,$D1,$7F,$9C
	.dc.b	$3D,$E9,$39,$F8,$94,$FF,$A6,$9B,$1F,$E6,$F9,$DB,$FE,$94,$8C,$BF
	.dc.b	$E4,$0F,$AB,$79,$8C,$3F,$95,$CA,$73,$18,$6C,$7F,$30,$6B,$F8,$6F
	.dc.b	$8C,$7C,$8B,$BF,$9C,$61,$B1,$99,$FE,$73,$39,$7A,$7C,$F7,$FA,$61
	.dc.b	$FF,$17,$5A,$31,$03,$CC,$E1,$BF,$0F,$FD,$13,$1F,$E1,$9E,$57,$F3
	.dc.b	$DE,$3F,$E7,$D1,$8E,$3F,$F4,$A7,$DD,$E5,$E9,$19,$C3,$FD,$77,$E3
	.dc.b	$FF,$37,$9F,$A8,$D8,$FF,$50,$D7,$17,$E5,$46,$6F,$BF,$D6,$FF,$3D
	.dc.b	$BF,$23,$6B,$F5,$B8,$23,$11,$FE,$4D,$8C,$79,$7F,$D4,$A3,$CB,$3C
	.dc.b	$6F,$AC,$F8,$7A,$7F,$35,$E2,$78,$F0,$E3,$8F,$E2,$F0,$E0,$F2,$BE
	.dc.b	$1D,$1F,$E6,$D8,$D4,$7A,$FE,$27,$E6,$C7,$FA,$8F,$87,$E9,$3F,$EB
	.dc.b	$BF,$A0,$D4,$63,$5F,$F6,$D3,$75,$FE,$D9,$7C,$A7,$FD,$8F,$5C,$38
	.dc.b	$BD,$1E,$FB,$E5,$7E,$57,$8B,$FC,$33,$E7,$33,$18,$F1,$51,$8D,$7F
	.dc.b	$85,$37,$2F,$E2,$CF,$D3,$5E,$23,$DE,$F9,$51,$F8,$98,$C5,$7E,$81
	.dc.b	$B1,$FE,$0C,$6B,$FC,$89,$CE,$DF,$E0,$46,$CF,$FA,$0F,$F5,$37,$C3
	.dc.b	$78,$67,$37,$DF,$E7,$3F,$0D,$DE,$FA,$E7,$D7,$6E,$73,$18,$CB,$FB
	.dc.b	$2E,$09,$F4,$E1,$EC,$6C,$66,$7D,$51,$83,$A3,$FD,$BB,$2F,$F7,$D3
	.dc.b	$D2,$3C,$DD,$FE,$4F,$AE,$7C,$DE,$2B,$FB,$BC,$61,$F1,$CF,$80,$8F
	.dc.b	$C9,$6E,$3F,$EC,$CC,$51,$F4,$FE,$93,$94,$E6,$3E,$33,$F1,$38,$BF
	.dc.b	$DB,$B7,$F0,$23,$1E,$3E,$A3,$EB,$47,$C0,$23,$83,$81,$FE,$0C,$FD
	.dc.b	$DB,$9C,$78,$CF,$C2,$7D,$7F,$FB,$A2,$7F,$C4,$98,$3E,$BF,$FC,$BB
	.dc.b	$1F,$E1,$49,$D5,$39,$3F,$C8,$D7,$BF,$57,$9F,$AB,$F5,$2F,$77,$FD
	.dc.b	$A1,$AF,$FB,$1D,$AF,$F1,$97,$AD,$FF,$93,$9B,$87,$F1,$DE,$EF,$F7
	.dc.b	$A7,$3F,$68,$FF,$13,$FD,$77,$FA,$F9,$FF,$1E,$5B,$EB,$FF,$D7,$F1
	.dc.b	$AF,$90,$FE,$29,$39,$F6,$7F,$49,$7C,$73,$FF,$34,$B5,$FE,$FE,$D7
	.dc.b	$FD,$BC,$D7,$F9,$83,$C6,$7F,$CC,$AD,$BF,$47,$CA,$FE,$3A,$BE,$44
	.dc.b	$72,$C3,$F9,$36,$EF,$AF,$F7,$D7,$E1,$9A,$BE,$6B,$F8,$BB,$9E,$0F
	.dc.b	$DE,$47,$F2,$68,$FF,$79,$35,$FE,$88,$D7,$97,$8F,$FC,$02,$FC,$1F
	.dc.b	$09,$CA,$3C,$78,$7F,$B9,$32,$FF,$29,$E1,$E0,$F3,$7F,$C3,$5E,$BF
	.dc.b	$F2,$DF,$6F,$FD,$61,$5F,$88,$FA,$FF,$F2,$23,$FE,$E1,$9F,$E2,$1F
	.dc.b	$45,$FD,$D2,$3F,$B5,$B7,$5F,$F4,$BB,$5F,$F1,$6B,$5F,$EB,$9A,$AF
	.dc.b	$FC,$43,$5D,$FD,$CF,$E4,$FF,$71,$F2,$A7,$37,$F3,$8F,$EC,$9F,$48
	.dc.b	$FA,$8F,$EA,$9F,$28,$FA,$F7,$D8,$7F,$25,$AD,$67,$FE,$D2,$9C,$BF
	.dc.b	$E7,$48,$CE,$1F,$EA,$E6,$BF,$E4,$46,$BF,$DC,$C3,$87,$C7,$87,$73
	.dc.b	$C7,$5F,$C9,$AB,$B4,$7F,$42,$FA,$7F,$E8,$6D,$8F,$F4,$C7,$0B,$D2
	.dc.b	$F1,$F5,$FF,$8E,$15,$F0,$AD,$C6,$1B,$7F,$D3,$BD,$FF,$E9,$FF,$61
	.dc.b	$67,$6D,$CF,$E5,$65,$DE,$31,$97,$77,$FA,$BF,$D8,$7F,$58,$F5,$AE
	.dc.b	$1E,$06,$E1,$CB,$3F,$EC,$53,$FE,$64,$55,$C6,$3C,$ED,$9E,$0D,$C3
	.dc.b	$FB,$4F,$D2,$BD,$8F,$EC,$FD,$DF,$DD,$63,$18,$7C,$08,$C4,$C8,$F8
	.dc.b	$7F,$51,$71,$F6,$F5,$BC,$DF,$EE,$35,$F9,$D6,$BF,$42,$34,$FC,$C3
	.dc.b	$45,$3F,$D7,$3C,$DF,$A2,$AF,$C9,$43,$EC,$3F,$4F,$1F,$95,$13,$BD
	.dc.b	$3F,$DE,$15,$BE,$DF,$EF,$13,$CE,$BF,$D2,$2D,$7F,$1C,$D7,$FC,$84
	.dc.b	$D7,$AF,$C7,$72,$3F,$C6,$95,$E3,$F6,$70,$FE,$33,$EC,$07,$F8,$A4
	.dc.b	$FD,$0D,$7E,$09,$CC,$6F,$AF,$51,$B6,$C6,$4F,$F7,$77,$96,$FB,$07
	.dc.b	$C7,$7F,$09,$F1,$BF,$C5,$91,$8A,$F9,$D1,$89,$91,$E5,$F8,$8B,$97
	.dc.b	$C9,$8F,$F3,$69,$FE,$FB,$3F,$C4,$7E,$9F,$F8,$7D,$CC,$7D,$2D,$3F
	.dc.b	$B0,$F0,$27,$F6,$C6,$A7,$FD,$7A,$D7,$37,$FE,$0E,$8C,$47,$F9,$BD
	.dc.b	$EB,$3C,$E3,$13,$F3,$B8,$5F,$3D,$E3,$8F,$F1,$97,$D8,$7E,$8A,$9D
	.dc.b	$CF,$2F,$FC,$31,$1F,$85,$9D,$4E,$7D,$AE,$3D,$F3,$4F,$E0,$78,$7F
	.dc.b	$82,$B4,$F1,$62,$C7,$B7,$17,$D1,$F0,$BF,$9C,$7B,$5F,$41,$B8,$FD
	.dc.b	$6E,$4F,$E1,$6F,$F6,$F6,$E1,$E6,$E6,$FF,$19,$3F,$CB,$70,$E9,$5F
	.dc.b	$7B,$BD,$F6,$1F,$E4,$C9,$46,$27,$FD,$D3,$49,$CD,$6E,$B5,$3F,$EE
	.dc.b	$E9,$CD,$75,$AF,$F3,$E7,$D4,$3E,$C3,$F8,$04,$61,$93,$FC,$9A,$46
	.dc.b	$2F,$F8,$DB,$C6,$35,$3D,$23,$5B,$7F,$0E,$E3,$F2,$12,$BD,$38,$FB
	.dc.b	$23,$18,$7C,$3C,$BF,$07,$5F,$F0,$03,$07,$D8,$FF,$15,$BF,$F4,$97
	.dc.b	$D8,$7F,$A3,$70,$FF,$4F,$B5,$FF,$4F,$35,$FF,$58,$B5,$FF,$6A,$3C
	.dc.b	$72,$FF,$5C,$B7,$0F,$06,$9D,$BE,$2E,$47,$FB,$A6,$75,$7F,$90,$CB
	.dc.b	$E7,$E1,$CA,$F7,$3F,$DA,$7E,$2D,$7C,$39,$FE,$5E,$5A,$C3,$C3,$F0
	.dc.b	$FF,$E8,$AB,$DB,$6F,$FD,$35,$9F,$FB,$45,$B8,$FC,$09,$F8,$67,$DC
	.dc.b	$EF,$B5,$FE,$DB,$C3,$FD,$2B,$E7,$7F,$86,$BE,$8B,$FB,$66,$1F,$E6
	.dc.b	$AA,$C3,$FD,$C4,$FF,$13,$46,$BD,$3F,$DA,$7A,$A8,$FE,$F1,$2F,$57
	.dc.b	$0F,$F7,$7A,$FD,$D4,$62,$3F,$BD,$61,$FE,$ED,$8C,$47,$84,$E5,$5D
	.dc.b	$2F,$37,$E0,$8C,$1E,$3C,$3C,$F5,$D0,$F3,$9E,$17,$2F,$F8,$60,$C9
	.dc.b	$FE,$22,$72,$FF,$83,$0F,$09,$C4,$A7,$5B,$F1,$C7,$A6,$7C,$73,$BB
	.dc.b	$F5,$BC,$DD,$BE,$03,$C4,$E0,$DE,$1E,$6C,$3D,$D3,$F4,$67,$55,$DA
	.dc.b	$3D,$07,$33,$D1,$87,$63,$9B,$E8,$3E,$63,$89,$E7,$1D,$67,$E5,$3C
	.dc.b	$EF,$D8,$FC,$1A,$E6,$75,$9F,$BE,$D8,$FC,$33,$5F,$A5,$AF,$E0,$DF
	.dc.b	$C5,$35,$CC,$7E,$22,$31,$F2,$0F,$9A,$FF,$51,$52,$73,$18,$6E,$89
	.dc.b	$D5,$FF,$91,$6F,$94,$E6,$B7,$5F,$6A,$5F,$33,$FE,$4B,$BF,$CA,$46
	.dc.b	$27,$E0,$87,$9B,$C7,$0F,$2C,$F0,$AE,$58,$F8,$5B,$5E,$5B,$E3,$5D
	.dc.b	$AF,$DB,$9F,$1A,$FC,$CD,$78,$5F,$CA,$BE,$08,$E1,$97,$92,$35,$7C
	.dc.b	$6B,$C9,$39,$BE,$B8,$F8,$E7,$AE,$3F,$E9,$A8,$F8,$D3,$A9,$F3,$35
	.dc.b	$F8,$13,$B7,$D4,$D7,$E2,$46,$F0,$F3,$57,$79,$F4,$D6,$E7,$31,$DA
	.dc.b	$FD,$53,$F1,$A3,$B9,$E9,$88,$F2,$3D,$27,$C6,$BF,$01,$E1,$9E,$C7
	.dc.b	$77,$33,$DA,$61,$F1,$89,$8F,$C9,$26,$1F,$12,$7F,$90,$E5,$FC,$DB
	.dc.b	$D4,$FF,$48,$47,$F4,$46,$ED,$FD,$29,$B8,$7F,$41,$EB,$B7,$C6,$3F
	.dc.b	$9B,$4F,$F7,$67,$FA,$7B,$63,$51,$FD,$63,$CE,$E3,$C6,$69,$DB,$97
	.dc.b	$FB,$8F,$E0,$78,$AF,$EA,$53,$5F,$EE,$71,$73,$FE,$B7,$37,$3F,$ED
	.dc.b	$53,$BA,$FE,$CB,$3E,$3C,$3C,$11,$FD,$36,$3D,$1D,$9F,$E1,$0E,$0F
	.dc.b	$7B,$67,$FC,$2D,$1E,$5E,$97,$D5,$6B,$B7,$F3,$95,$FE,$B9,$BE,$0F
	.dc.b	$F8,$4E,$32,$FE,$EE,$32,$FF,$AF,$F5,$FF,$07,$F1,$F5,$24,$7F,$84
	.dc.b	$FB,$2B,$FE,$12,$8B,$BF,$F9,$26,$BF,$C2,$F5,$E3,$F9,$F9,$7F,$E2
	.dc.b	$D7,$7C,$1E,$1C,$91,$F0,$1B,$3F,$E1,$66,$BE,$98,$F8,$BC,$BF,$D7
	.dc.b	$71,$FF,$2C,$67,$FD,$39,$B1,$9C,$3F,$8C,$65,$FE,$84,$E7,$F9,$73
	.dc.b	$9C,$7F,$CB,$3E,$83,$AB,$F0,$0D,$8F,$F4,$0C,$ED,$7D,$7F,$F9,$91
	.dc.b	$FE,$FD,$8F,$FB,$9E,$31,$1F,$EA,$56,$E5,$F1,$7E,$BF,$EB,$CF,$03
	.dc.b	$F9,$64,$EA,$BC,$F3,$FC,$D6,$7F,$DC,$ED,$7A,$79,$5F,$A2,$5F,$00
	.dc.b	$F3,$B0,$D8,$EC,$6A,$38,$7D,$7F,$E4,$23,$FD,$75,$3B,$7F,$CB,$AD
	.dc.b	$7F,$C2,$ED,$7F,$A1,$1C,$CF,$D7,$3C,$72,$FF,$3C,$CB,$D7,$EE,$79
	.dc.b	$2F,$F3,$7F,$39,$FE,$5C,$C3,$FD,$5A,$D7,$FC,$E8,$DB,$FC,$67,$D7
	.dc.b	$FF,$AC,$27,$7D,$3F,$59,$86,$0F,$99,$C6,$97,$D4,$FE,$ED,$79,$14
	.dc.b	$C4,$4A,$E3,$3A,$6C,$E5,$FF,$70,$7C,$46,$5B,$1F,$4E,$31,$8F,$FA
	.dc.b	$6B,$EB,$FF,$B6,$57,$D2,$9E,$0F,$F9,$09,$AF,$CC,$6B,$FD,$BD,$A9
	.dc.b	$EB,$C5,$F4,$A7,$F1,$73,$FE,$F3,$6D,$7F,$AA,$AB,$73,$F9,$EF,$86
	.dc.b	$FC,$EF,$D2,$48,$F8,$FC,$E9,$FD,$09,$98,$5F,$EF,$65,$5A,$9F,$F7
	.dc.b	$2F,$D7,$BE,$2B,$F8,$BB,$63,$31,$FA,$CF,$9E,$CB,$FC,$B7,$E0,$FF
	.dc.b	$D9,$13,$F6,$1A,$FF,$AD,$6B,$93,$E2,$D7,$F7,$78,$BE,$BF,$F6,$FF
	.dc.b	$5F,$2B,$D7,$1F,$BA,$30,$F1,$3F,$7F,$19,$76,$7C,$27,$7A,$7B,$E7
	.dc.b	$F8,$1F,$C5,$FF,$04,$F8,$19,$7F,$22,$32,$FC,$A8,$CB,$DF,$95,$8A
	.dc.b	$BF,$86,$FC,$D7,$F0,$EF,$AF,$97,$F4,$A3,$2F,$D4,$8C,$BC,$E9,$3F
	.dc.b	$8F,$D2,$4F,$E3,$5F,$EC,$BF,$09,$FE,$3B,$FA,$3C,$CF,$FB,$63,$97
	.dc.b	$FD,$DA,$FE,$46,$BF,$B3,$3F,$8D,$6C,$7F,$BC,$B6,$47,$F9,$06,$FF
	.dc.b	$C8,$DF,$49,$DB,$D1,$3D,$AB,$FD,$F3,$E1,$3E,$93,$F8,$B3,$72,$FE
	.dc.b	$26,$D7,$D9,$F0,$7F,$C6,$FE,$BD,$F1,$3E,$46,$CF,$F7,$B6,$BF,$EF
	.dc.b	$26,$BF,$D6,$78,$3F,$9E,$71,$FD,$0F,$63,$F9,$D9,$E2,$F8,$9F,$E5
	.dc.b	$1F,$5F,$FC,$2E,$FF,$08,$3E,$3D,$F2,$FF,$CB,$5B,$1F,$CA,$A7,$86
	.dc.b	$FF,$12,$3F,$CC,$93,$A7,$FD,$7A,$73,$AF,$9B,$19,$8F,$F6,$0B,$5F
	.dc.b	$E5,$74,$FA,$FF,$29,$23,$F9,$CC,$F5,$BF,$04,$3E,$BB,$F4,$F1,$86
	.dc.b	$F2,$FF,$36,$6C,$FF,$9E,$7E,$BB,$F8,$9B,$F0,$23,$F5,$EF,$FB,$51
	.dc.b	$AF,$1F,$8B,$FC,$2C,$62,$3F,$D0,$4F,$45,$3F,$DD,$AF,$CA,$DB,$E5
	.dc.b	$3F,$D4,$BE,$B9,$E2,$7F,$59,$6C,$66,$3F,$AE,$71,$FE,$17,$0F,$7D
	.dc.b	$6A,$BE,$6B,$5F,$F7,$5B,$5F,$EA,$AD,$7F,$D3,$0F,$18,$F9,$32,$DF
	.dc.b	$B9,$E3,$BF,$ED,$9F,$53,$FD,$BE,$BE,$7B,$71,$FE,$C2,$DC,$5F,$4A
	.dc.b	$F3,$BA,$47,$F7,$09,$FE,$F0,$FF,$C0,$06,$BC,$BD,$CF,$D8,$C6,$23
	.dc.b	$FD,$5C,$F5,$A7,$D7,$7F,$30,$9F,$F0,$44,$63,$CA,$F3,$DF,$A3,$DF
	.dc.b	$95,$E1,$7F,$CC,$F5,$E5,$8C,$9F,$96,$C3,$FC,$86,$F8,$27,$C6,$47
	.dc.b	$BE,$0F,$1B,$9B,$63,$B1,$FE,$0E,$E7,$FE,$6F,$AF,$E5,$9B,$AF,$F8
	.dc.b	$21,$B3,$F2,$3E,$DF,$F6,$0F,$1F,$33,$F8,$9C,$FF,$40,$8F,$F0,$79
	.dc.b	$FB,$48,$C9,$1F,$87,$F5,$FF,$D7,$B5,$3F,$D8,$23,$81,$E5,$F9,$5F
	.dc.b	$F1,$45,$FE,$EB,$1F,$24,$7C,$6C,$7F,$C4,$99,$7A,$63,$C6,$4B,$FF
	.dc.b	$46,$5F,$F0,$97,$F8,$87,$D7,$FF,$A5,$3E,$A9,$F4,$BE,$8C,$3F,$83
	.dc.b	$46,$2B,$FC,$9B,$51,$FF,$21,$46,$5B,$5C,$72,$FE,$BF,$5F,$E8,$2F
	.dc.b	$A6,$FE,$95,$5F,$E4,$9F,$82,$46,$23,$FA,$F4,$63,$CF,$FF,$43,$71
	.dc.b	$F1,$64,$F1,$AE,$19,$FF,$9E,$7D,$9F,$F7,$23,$FC,$77,$CB,$FC,$E7
	.dc.b	$67,$FA,$BB,$AB,$F9,$8C,$E5,$BA,$FF,$9B,$BF,$35,$7C,$DE,$EE,$67
	.dc.b	$FE,$5F,$C7,$9B,$FE,$36,$AD,$B9,$C3,$F9,$DA,$13,$C1,$FC,$1B,$5F
	.dc.b	$F2,$9B,$5E,$0F,$F5,$1F,$43,$FD,$8B,$E6,$9A,$7F,$AA,$7D,$EB,$5E
	.dc.b	$59,$D3,$67,$4F,$F9,$AB,$EB,$3F,$A3,$5F,$FB,$35,$B1,$FD,$01,$AF
	.dc.b	$E6,$1A,$FF,$9A,$3E,$3D,$F1,$3F,$C2,$23,$FA,$3C,$63,$B3,$FD,$B0
	.dc.b	$D8,$CC,$ED,$E9,$F3,$5F,$E9,$56,$FC,$3F,$8E,$31,$1F,$E7,$6C,$11
	.dc.b	$88,$F9,$7E,$17,$F3,$2E,$2F,$A8,$D8,$FF,$0C,$72,$3E,$BF,$FA,$F7
	.dc.b	$2F,$E0,$EB,$4D,$F1,$B3,$87,$D0,$9C,$BF,$EB,$0B,$E4,$F6,$9D,$E1
	.dc.b	$F7,$3E,$BF,$F9,$94,$FE,$15,$B1,$FE,$29,$FA,$FF,$F3,$BC,$E7,$C9
	.dc.b	$E3,$C4,$BD,$E7,$F9,$76,$D7,$FB,$02,$F8,$E1,$FE,$41,$BE,$18,$72
	.dc.b	$BD,$CF,$8E,$3C,$BE,$92,$F8,$E3,$D6,$7D,$3D,$CE,$0F,$9B,$9F,$EC
	.dc.b	$DB,$8F,$CA,$FA,$0D,$C6,$30,$FF,$6B,$B6,$BF,$DE,$4D,$A6,$4D,$98
	.dc.b	$72,$E1,$F2,$FC,$03,$E5,$FE,$5F,$89,$FC,$43,$E8,$33,$18,$F1,$51
	.dc.b	$8D,$7F,$86,$B7,$4F,$F3,$25,$F1,$9F,$F6,$3F,$85,$FE,$4E,$6F,$BD
	.dc.b	$F6,$29,$F5,$BF,$E4,$77,$FC,$53,$5B,$6C,$EE,$31,$C3,$FE,$18,$AE
	.dc.b	$0F,$F8,$6A,$33,$7F,$EC,$C5,$ED,$F1,$3C,$14,$F9,$0C,$57,$9A,$6E
	.dc.b	$3F,$E3,$AA,$D4,$E4,$91,$E6,$9E,$75,$C3,$0F,$17,$67,$9C,$F7,$78
	.dc.b	$CF,$9F,$C6,$3F,$12,$6E,$3F,$E4,$8A,$D4,$E4,$91,$F1,$27,$9D,$70
	.dc.b	$BF,$81,$3F,$14,$F9,$9F,$59,$FE,$F9,$BF,$A3,$19,$69,$FE,$FA,$8D
	.dc.b	$47,$D0,$34,$58,$D7,$C6,$62,$BF,$C8,$FF,$57,$FE,$F3,$9F,$E5,$91
	.dc.b	$8F,$8E,$FC,$DF,$C7,$FD,$49,$FF,$7E,$D6,$F0,$FF,$32,$B6,$F8,$4F
	.dc.b	$F5,$1A,$E1,$5F,$CD,$EB,$53,$CA,$FF,$11,$18,$97,$62,$F8,$C2,$74
	.dc.b	$BE,$AB,$FD,$A9,$87,$F9,$B2,$31,$F1,$9F,$E9,$A7,$FA,$34,$E9,$FF
	.dc.b	$38,$B5,$FF,$04,$B9,$CB,$FC,$D2,$DA,$FE,$9D,$8F,$8F,$83,$F3,$D3
	.dc.b	$9B,$EB,$86,$0F,$8F,$36,$FE,$59,$FD,$5F,$77,$D5,$7B,$DB,$AF,$EA
	.dc.b	$7E,$53,$F8,$7C,$62,$B8,$E3,$FD,$7E,$7F,$89,$A7,$76,$F8,$0E,$5E
	.dc.b	$17,$F2,$AC,$3B,$C6,$30,$E7,$18,$7A,$E1,$CE,$BF,$8E,$FC,$2F,$FA
	.dc.b	$BA,$7F,$E0,$0F,$A5,$CF,$EA,$7F,$C7,$D1,$83,$8F,$CB,$7F,$69,$6F
	.dc.b	$A1,$BF,$57,$F8,$24,$E7,$2F,$F7,$76,$6C,$3F,$DC,$18,$79,$1B,$3F
	.dc.b	$EE,$5F,$0D,$F5,$5C,$EB,$CF,$F3,$EF,$A5,$FE,$20,$D8,$F8,$FF,$1C
	.dc.b	$FA,$AF,$C4,$CE,$71,$FF,$5D,$B7,$AB,$FD,$4E,$DB,$FF,$6D,$57,$0C
	.dc.b	$3A,$92,$3A,$C6,$5F,$AB,$E6,$33,$7D,$07,$F3,$2B,$FE,$66,$DA,$DC
	.dc.b	$E9,$B3,$A6,$9A,$F8,$AF,$E7,$7F,$55,$FC,$F1,$B1,$93,$9C,$65,$A7
	.dc.b	$6F,$AA,$FE,$87,$E7,$7F,$44,$6C,$64,$F1,$C6,$5A,$67,$6F,$A5,$19
	.dc.b	$6C,$7F,$93,$7E,$7B,$FA,$8B,$7D,$0F,$EA,$3F,$41,$FD,$8A,$73,$5F
	.dc.b	$E5,$06,$F0,$FF,$B9,$7C,$34,$63,$0F,$F1,$E7,$07,$CB,$C3,$96,$3F
	.dc.b	$94,$F2,$7F,$AE,$D5,$3E,$59,$72,$FC,$43,$72,$FF,$79,$61,$F5,$63
	.dc.b	$FA,$B7,$27,$E6,$AB,$FD,$FA,$DA,$FF,$7C,$4B,$EA,$BF,$B2,$BF,$A6
	.dc.b	$9D,$3F,$E5,$1A,$46,$1F,$F2,$A3,$63,$97,$A1,$E0,$F8,$A9,$1C,$DF
	.dc.b	$F0,$FB,$A3,$07,$E3,$3E,$90,$7F,$84,$07,$D2,$D7,$0D,$8E,$0D,$C1
	.dc.b	$97,$C2,$AE,$35,$FE,$85,$FA,$97,$C4,$FF,$86,$1B,$1E,$1F,$80,$C4
	.dc.b	$C8,$E0,$F8,$E3,$18,$76,$9F,$88,$7F,$96,$8F,$E1,$A5,$74,$1F,$9D
	.dc.b	$F2,$FF,$E0,$97,$58,$E6,$7F,$CF,$D1,$83,$AD,$7C,$8D,$97,$C6,$24
	.dc.b	$F8,$1D,$46,$2B,$C2,$3E,$AB,$FC,$58,$DE,$6F,$F1,$EF,$91,$FD,$E2
	.dc.b	$37,$3A,$24,$EB,$D3,$FF,$10,$47,$F8,$D7,$A1,$B7,$F0,$BB,$FE,$B6
	.dc.b	$DE,$4F,$B7,$F5,$FF,$3F,$B6,$6D,$F4,$DF,$BD,$18,$BF,$D4,$7D,$4F
	.dc.b	$EA,$CF,$F2,$8F,$BA,$FA,$BC,$61,$FB,$39,$B5,$FE,$76,$9B,$C7,$FE
	.dc.b	$B9,$78,$9C,$AE,$DF,$F1,$A7,$A3,$63,$4D,$3F,$80,$71,$7E,$F4,$F1
	.dc.b	$FD,$5D,$8C,$93,$0F,$F4,$3C,$65,$FF,$37,$7D,$57,$F9,$A2,$7F,$D6
	.dc.b	$6D,$8F,$F2,$EE,$0C,$BF,$CF,$73,$9F,$84,$34,$E1,$B7,$F9,$8E,$77
	.dc.b	$F4,$B3,$1F,$F6,$65,$6B,$0E,$91,$89,$D6,$3E,$47,$E6,$DE,$35,$FF
	.dc.b	$6F,$61,$FE,$6B,$BE,$7F,$05,$F4,$BA,$FC,$AC,$8C,$9F,$2E,$31,$8F
	.dc.b	$E4,$0F,$F6,$C7,$93,$FD,$EF,$E7,$7F,$85,$43,$CD,$FD,$07,$AA,$9C
	.dc.b	$CA,$B8,$3F,$C4,$6B,$73,$A9,$CC,$FF,$83,$F0,$AF,$80,$FF,$04,$4F
	.dc.b	$82,$B8,$C7,$8A,$B7,$18,$8F,$E3,$D1,$8A,$E7,$1E,$33,$99,$E3,$73
	.dc.b	$1B,$8C,$9C,$8A,$7B,$E3,$6E,$27,$C0,$F9,$EF,$E1,$3D,$4E,$27,$85
	.dc.b	$CB,$EF,$33,$5E,$18,$CC,$73,$BF,$81,$7E,$09,$CD,$F8,$2B,$A5,$69
	.dc.b	$E6,$79,$6B,$84,$66,$BA,$CF,$03,$CC,$E2,$FC,$53,$F5,$67,$91,$D2
	.dc.b	$7C,$35,$D2,$38,$CE,$E7,$31,$E2,$E0,$79,$FC,$DB,$1F,$64,$38,$7E
	.dc.b	$83,$7E,$29,$F2,$FF,$D1,$27,$75,$FD,$6E,$73,$BA,$72,$58,$CB,$80
	.dc.b	$E2,$FB,$5F,$5C,$3E,$AC,$BE,$47,$FB,$44,$61,$B9,$78,$6B,$83,$71
	.dc.b	$FC,$DD,$F1,$C6,$DF,$07,$C9,$7C,$1B,$5A,$61,$F9,$2A,$F4,$E1,$E4
	.dc.b	$AE,$13,$EA,$BE,$8D,$3C,$31,$AA,$E3,$87,$C7,$AD,$C7,$96,$BB,$56
	.dc.b	$CE,$13,$A3,$6E,$70,$F2,$E1,$F0,$EB,$BC,$47,$D1,$5C,$5C,$4F,$A6
	.dc.b	$78,$E1,$E3,$9C,$BF,$25,$E2,$68,$F8,$A7,$9A,$30,$FB,$CC,$12,$7C
	.dc.b	$8F,$23,$13,$FC,$E3,$CF,$FF,$11,$57,$F2,$66,$E9,$9A,$FE,$99,$E2
	.dc.b	$61,$FC,$AF,$17,$F9,$99,$FE,$6B,$59,$D7,$F9,$EB,$BE,$67,$37,$F8
	.dc.b	$8A,$BF,$9D,$F9,$5B,$C0,$D3,$84,$BE,$AF,$15,$C0,$87,$F8,$4A,$FA
	.dc.b	$C6,$A3,$FA,$AD,$5D,$3F,$C9,$37,$CE,$78,$C7,$F4,$49,$D6,$9F,$E5
	.dc.b	$8A,$D5,$7F,$97,$F7,$F8,$1D,$FF,$E5,$E6,$F9,$9F,$87,$C7,$AF,$D5
	.dc.b	$E2,$65,$FF,$9B,$5F,$5C,$7B,$38,$E4,$FF,$77,$7D,$FD,$B8,$8F,$66
	.dc.b	$5F,$BE,$74,$7F,$7D,$8F,$0F,$B9,$23,$FC,$F4,$F1,$C7,$FC,$E9,$92
	.dc.b	$35,$3F,$E7,$97,$5A,$7F,$B9,$39,$7F,$98,$DF,$F9,$ED,$B1,$9E,$5F
	.dc.b	$B5,$CB,$FC,$4E,$B5,$1F,$12,$B7,$87,$FC,$00,$6F,$17,$F9,$83,$9F
	.dc.b	$F1,$12,$FA,$BF,$F1,$FB,$FE,$D8,$6B,$FE,$3C,$6E,$1F,$E8,$0F,$AB
	.dc.b	$70,$F2,$9C,$E9,$FE,$79,$6D,$FF,$1F,$6E,$2B,$E3,$3F,$EE,$6F,$82
	.dc.b	$7C,$1F,$91,$B1,$98,$F3,$FC,$9F,$E5,$23,$04,$3F,$CA,$33,$C1,$FF
	.dc.b	$29,$C6,$DC,$CF,$52,$3F,$EF,$68,$D6,$7F,$EA,$9F,$94,$FD,$2C,$FF
	.dc.b	$37,$8C,$72,$7E,$42,$73,$3F,$9F,$FA,$AF,$F3,$FB,$BD,$7F,$9F,$74
	.dc.b	$2B,$78,$72,$AD,$09,$D6,$1F,$B5,$C3,$9C,$F5,$FA,$BC,$78,$B2,$3F
	.dc.b	$DC,$B2,$9E,$06,$63,$C9,$38,$7B,$8F,$A6,$47,$E0,$9B,$1A,$6B,$FE
	.dc.b	$7C,$FA,$BF,$EC,$2E,$67,$81,$1E,$B1,$A7,$FD,$07,$3C,$B2,$FF,$48
	.dc.b	$47,$86,$B8,$3D,$7C,$1F,$E2,$9F,$06,$F8,$7B,$F4,$12,$BE,$3F,$2A
	.dc.b	$B5,$18,$6B,$FC,$45,$4F,$18,$C5,$7F,$08,$FA,$B7,$D0,$52,$4F,$3F
	.dc.b	$97,$AF,$FA,$B3,$89,$5C,$23,$EC,$C6,$2B,$D3,$5E,$AF,$AA,$F5,$FD
	.dc.b	$5C,$BF,$34,$65,$F4,$46,$5F,$44,$65,$F5,$46,$77,$FF,$77,$E8,$AF
	.dc.b	$8F,$F1,$1F,$23,$DD,$98,$7C,$38,$CB,$F1,$63,$2F,$C5,$8F,$9F,$F2
	.dc.b	$9F,$43,$EA,$E4,$FC,$B8,$CB,$F3,$63,$2F,$CD,$8F,$C4,$75,$48,$FC
	.dc.b	$43,$FD,$4F,$E0,$7F,$C4,$FF,$17,$99,$FC,$1C,$DD,$3F,$DA,$55,$A9
	.dc.b	$E2,$43,$67,$E9,$9D,$3F,$EC,$EF,$7D,$F3,$DF,$C6,$9B,$1F,$9E,$6B
	.dc.b	$F8,$3F,$94,$FE,$3F,$F5,$4F,$8F,$A4,$9E,$17,$F9,$2F,$2A,$47,$E3
	.dc.b	$5E,$DF,$21,$FC,$A3,$EA,$BF,$6B,$A7,$E7,$3E,$35,$F4,$7F,$CB,$1B
	.dc.b	$87,$EC,$2B,$7B,$FF,$90,$61,$3A,$7F,$DD,$2D,$BE,$97,$DF,$E2,$7C
	.dc.b	$1F,$55,$2F,$A3,$DE,$7F,$9D,$FD,$53,$E2,$A9,$2B,$F8,$0F,$D5,$7E
	.dc.b	$D7,$1F,$D9,$D6,$E7,$F8,$2B,$5F,$F7,$DB,$63,$F7,$33,$C7,$3F,$54
	.dc.b	$EE,$BF,$D5,$ED,$73,$7F,$EA,$E7,$FC,$C1,$19,$7F,$CC,$5F,$54,$EE
	.dc.b	$FE,$A6,$D8,$FE,$2E,$D5,$8F,$F0,$AE,$2F,$BA,$FF,$0C,$9D,$36,$B9
	.dc.b	$9F,$D7,$A7,$78,$FD,$1A,$DB,$67,$77,$FC,$9A,$B6,$FF,$9D,$FE,$8F
	.dc.b	$FC,$00,$F8,$AB,$ED,$46,$30,$FE,$3A,$D8,$FF,$3F,$7D,$2F,$F4,$1E
	.dc.b	$2F,$ED,$F3,$9B,$E9,$ED,$E0,$FF,$1D,$56,$E3,$F9,$BC,$ED,$34,$E5
	.dc.b	$1B,$8F,$21,$E7,$F7,$3B,$91,$FE,$9F,$1F,$D4,$DC,$1C,$0E,$B1,$8C
	.dc.b	$58,$FA,$BD,$0F,$F7,$5C,$7B,$A7,$53,$F0,$9F,$4C,$EA,$3F,$3D,$C3
	.dc.b	$FC,$9E,$FB,$36,$FF,$A1,$B6,$7E,$3F,$D4,$7E,$92,$BF,$AB,$CE,$5F
	.dc.b	$F5,$43,$5F,$D7,$F8,$BF,$22,$78,$47,$68,$FF,$0E,$CE,$CE,$59,$2B
	.dc.b	$78,$78,$B9,$76,$4C,$3E,$3C,$FF,$81,$31,$F5,$CE,$67,$D6,$F8,$7D
	.dc.b	$EF,$EB,$3F,$51,$FE,$73,$FA,$27,$D1,$FC,$3B,$46,$1D,$9F,$D6,$67
	.dc.b	$84,$EE,$75,$39,$3F,$A7,$FD,$1B,$C6,$FF,$30,$B6,$33,$1F,$E7,$6E
	.dc.b	$1C,$B2,$32,$7F,$6A,$F5,$BF,$CC,$DA,$7F,$B4,$FD,$1F,$F0,$B7,$3F
	.dc.b	$FA,$1F,$93,$F9,$35,$FF,$85,$A7,$2F,$F7,$2F,$A8,$7C,$BE,$EE,$4E
	.dc.b	$EF,$8B,$FE,$83,$AD,$B9,$9D,$BF,$98,$BE,$AE,$EF,$FD,$31,$E1,$BC
	.dc.b	$8A,$49,$F5,$7B,$35,$E5,$3B,$7D,$0D,$7F,$CF,$31,$FE,$60,$FA,$5F
	.dc.b	$E6,$38,$7F,$15,$6C,$FF,$35,$3F,$DE,$CE,$1C,$35,$FF,$7E,$7C,$2B
	.dc.b	$A7,$F8,$F7,$C4,$A3,$13,$CC,$FF,$7D,$B8,$7B,$46,$E7,$C5,$F2,$CF
	.dc.b	$8C,$A4,$AE,$58,$27,$51,$E7,$C5,$F4,$EF,$0B,$F7,$33,$FE,$98,$F7
	.dc.b	$2F,$7E,$2F,$75,$8E,$5F,$F5,$75,$5B,$56,$A7,$F1,$13,$98,$C4,$F5
	.dc.b	$9F,$C3,$60,$FB,$3E,$5E,$E7,$FB,$3B,$E5,$38,$4E,$6B,$F2,$51,$8A
	.dc.b	$DF,$4A,$31,$3A,$3A,$CE,$A3,$38,$7F,$AD,$1F,$47,$B4,$FD,$37,$FB
	.dc.b	$F5,$FE,$7F,$C5,$CC,$62,$FF,$D7,$71,$8E,$AF,$0B,$6B,$81,$FE,$76
	.dc.b	$F0,$1F,$BB,$BA,$7E,$34,$61,$F7,$FC,$5F,$FA,$6A,$B8,$56,$DB,$3C
	.dc.b	$23,$18,$FF,$B9,$EB,$83,$FC,$EA,$FA,$5F,$F9,$E3,$C2,$7A,$BF,$57
	.dc.b	$A9,$F6,$5F,$1E,$B5,$39,$8C,$61,$FC,$F6,$F8,$D7,$F4,$29,$CF,$47
	.dc.b	$F9,$D2,$73,$5F,$D2,$B0,$EB,$C3,$F8,$38,$4B,$C6,$ED,$F1,$79,$8C
	.dc.b	$65,$D6,$31,$3A,$8F,$53,$6B,$C8,$DC,$7C,$4D,$9F,$F1,$35,$AB,$E0
	.dc.b	$78,$FF,$1A,$FD,$11,$8F,$87,$F1,$36,$FE,$43,$4F,$84,$59,$E3,$1F
	.dc.b	$D9,$2B,$A3,$FE,$35,$F9,$FB,$5F,$BD,$97,$CB,$FC,$39,$CE,$BF,$CF
	.dc.b	$A3,$18,$F9,$9B,$87,$56,$E1,$F3,$70,$E7,$3F,$EB,$4C,$39,$C6,$27
	.dc.b	$FB,$A7,$D3,$FF,$A8,$EB,$FD,$8F,$93,$EC,$3F,$D3,$31,$FE,$BC,$9D
	.dc.b	$57,$F8,$0A,$B5,$E8,$7D,$99,$CC,$7F,$02,$BE,$35,$DB,$E1,$BF,$80
	.dc.b	$4F,$0C,$3A,$47,$FA,$E2,$7F,$CB,$17,$D9,$FE,$4C,$D7,$31,$B7,$EC
	.dc.b	$7E,$18,$E6,$FD,$AF,$E0,$FF,$90,$C3,$FC,$C5,$39,$7F,$D5,$71,$9A
	.dc.b	$E3,$5C,$39,$3F,$25,$7B,$9F,$F5,$A5,$6D,$46,$72,$FF,$67,$B7,$0F
	.dc.b	$C5,$7E,$18,$FC,$31,$F3,$FD,$33,$FE,$41,$E5,$3E,$A9,$F1,$1F,$C6
	.dc.b	$BE,$2B,$F8,$46,$9F,$CB,$2B,$53,$9F,$0B,$F4,$DD,$9F,$D5,$F5,$FF
	.dc.b	$7E,$D6,$9F,$F7,$8C,$67,$7F,$F1,$BC,$E5,$FF,$19,$46,$70,$FF,$73
	.dc.b	$B6,$BF,$45,$5E,$5F,$57,$BF,$A5,$FC,$43,$E9,$BC,$DF,$4D,$6F,$FD
	.dc.b	$C8,$F9,$3E,$77,$F8,$AF,$D8,$D9,$FF,$75,$F0,$A3,$F8,$CF,$D1,$E7
	.dc.b	$6F,$E2,$8D,$DB,$FD,$F3,$F1,$9F,$13,$0F,$DA,$47,$F4,$E5,$F4,$FF
	.dc.b	$19,$FF,$4F,$4E,$EF,$FC,$AB,$1A,$E1,$FE,$19,$39,$BE,$57,$C6,$FF
	.dc.b	$80,$4E,$6F,$F8,$11,$C4,$F0,$9D,$B8,$BF,$AC,$39,$9F,$F2,$E4,$F1
	.dc.b	$7F,$CB,$75,$D7,$D1,$FC,$A7,$D2,$7F,$38,$F0,$FF,$9C,$B7,$0F,$8A
	.dc.b	$DB,$FE,$6C,$DC,$BF,$B6,$DF,$E7,$76,$FE,$83,$F4,$7F,$D0,$9B,$87
	.dc.b	$CE,$6E,$1F,$CF,$DB,$7C,$97,$CE,$FF,$4A,$FA,$1F,$E9,$6D,$BF,$AE
	.dc.b	$DB,$FE,$90,$DB,$E4,$D9,$FE,$51,$8F,$EC,$63,$14,$F4,$9F,$4A,$F8
	.dc.b	$4F,$CF,$7B,$5F,$DA,$A3,$18,$FC,$EB,$62,$ED,$FF,$07,$4E,$71,$E7
	.dc.b	$8F,$28,$C6,$1C,$B0,$FF,$4A,$E1,$C7,$1F,$EC,$FF,$07,$FD,$A3,$E1
	.dc.b	$EE,$1F,$94,$6D,$FF,$65,$6E,$BF,$1D,$FE,$A9,$3F,$E2,$F3,$CC,$3A
	.dc.b	$BF,$9C,$D7,$FA,$6B,$E0,$3F,$0B,$18,$9E,$13,$F9,$48,$C4,$EB,$CE
	.dc.b	$EA,$DA,$F8,$55,$CD,$EF,$5F,$D1,$B3,$FE,$3D,$EE,$FF,$80,$BD,$7F
	.dc.b	$F0,$1B,$6B,$F5,$ED,$BF,$F0,$03,$73,$FF,$1B,$C3,$E9,$BF,$C7,$D5
	.dc.b	$F4,$70,$FF,$7E,$7A,$DF,$DC,$E7,$35,$F3,$3E,$33,$FB,$CE,$1C,$A7
	.dc.b	$2F,$E5,$A3,$38,$7F,$6B,$AE,$DF,$01,$FE,$36,$9C,$C7,$F2,$6F,$BD
	.dc.b	$FF,$1C,$FE,$14,$F9,$BF,$F0,$17,$9C,$9C,$C6,$1B,$8F,$C6,$9D,$3E
	.dc.b	$9A,$E4,$E6,$77,$7F,$ED,$1A,$DB,$F8,$6A,$DB,$FE,$DC,$E8,$FC,$BF
	.dc.b	$9B,$FC,$CF,$EA,$7F,$2C,$DC,$7F,$1E,$DA,$E1,$D3,$FC,$FA,$7E,$57
	.dc.b	$5A,$FF,$3D,$F2,$7F,$CF,$7E,$22,$D6,$9B,$87,$E6,$2D,$18,$E1,$FE
	.dc.b	$03,$18,$AF,$CF,$FB,$5F,$E5,$AF,$A5,$FD,$53,$63,$27,$FA,$C6,$32
	.dc.b	$D3,$EF,$7D,$9B,$8F,$39,$7E,$A2,$31,$3C,$A3,$13,$AC,$7F,$89,$D6
	.dc.b	$A7,$E0,$E1,$FE,$64,$FA,$6F,$F0,$0E,$1F,$E2,$BF,$82,$7C,$C3,$3D
	.dc.b	$7B,$7F,$E5,$C8,$C5,$7E,$F6,$F8,$E3,$FE,$DA,$9C,$BF,$56,$33,$87
	.dc.b	$4C,$3F,$D0,$DF,$20,$F9,$CF,$F3,$EF,$B1,$86,$E5,$FE,$CD,$9D,$3F
	.dc.b	$ED,$59,$D3,$4D,$4F,$0C,$7F,$D9,$15,$C1,$FE,$1B,$F1,$4B,$E3,$18
	.dc.b	$6E,$FF,$CE,$27,$58,$78,$7C,$2F,$F4,$8F,$C6,$7A,$1B,$19,$3F,$DD
	.dc.b	$11,$96,$9F,$C1,$30,$FF,$2B,$7D,$4F,$F9,$5F,$6F,$9B,$1F,$AF,$F6
	.dc.b	$FF,$D6,$71,$88,$FD,$94,$EF,$83,$FD,$A1,$5A,$7F,$D9,$11,$99,$E5
	.dc.b	$75,$E2,$2F,$A3,$FE,$C9,$8D,$3E,$43,$55,$DA,$33,$0A,$DD,$F8,$63
	.dc.b	$FD,$C5,$18,$9D,$B6,$39,$9A,$29,$FE,$16,$8D,$B8,$3C,$85,$3C,$06
	.dc.b	$67,$C0,$FB,$2F,$D7,$18,$8F,$54,$62,$7A,$B6,$3A,$57,$96,$3E,$1C
	.dc.b	$76,$74,$62,$35,$7F,$3A,$7C,$55,$D2,$3E,$34,$74,$32,$5A,$E7,$1C
	.dc.b	$5B,$1F,$0D,$AF,$AE,$39,$C7,$86,$3F,$DD,$BA,$78,$71,$FC,$C4,$7C
	.dc.b	$B6,$BF,$79,$BF,$1E,$F2,$5B,$7F,$A9,$67,$31,$8A,$EA,$7F,$AA,$6B
	.dc.b	$85,$6E,$FF,$A1,$46,$27,$61,$C6,$1F,$97,$5E,$57,$F4,$98,$C3,$72
	.dc.b	$E7,$8F,$86,$78,$36,$3B,$CE,$EF,$C8,$D7,$C1,$3B,$7D,$0D,$7C,$91
	.dc.b	$BB,$F1,$C7,$E8,$DF,$3D,$78,$B1,$C5,$6C,$E1,$7C,$EB,$77,$E2,$36
	.dc.b	$73,$8D,$1B,$9E,$51,$9A,$8F,$E8,$2B,$C6,$D9,$EF,$3A,$C3,$E3,$CF
	.dc.b	$63,$85,$F6,$9C,$4E,$A7,$31,$C7,$0F,$93,$1A,$7D,$CE,$8F,$01,$93
	.dc.b	$E4,$C7,$17,$DC,$68,$E1,$3F,$C5,$B2,$70,$3C,$BF,$F0,$35,$FF,$4C
	.dc.b	$F6,$EE,$DF,$E4,$CE,$84,$63,$26,$3C,$B0,$C3,$63,$FA,$EC,$EF,$C8
	.dc.b	$5F,$29,$FE,$85,$7C,$32,$FF,$0B,$47,$F6,$1F,$99,$E1,$47,$51,$0F
	.dc.b	$ED,$2E,$8C,$3F,$D1,$8C,$1A,$3F,$A8,$1A,$CF,$FB,$4C,$6A,$38,$1F
	.dc.b	$D8,$7E,$7F,$DD,$D2,$FC,$7D,$EF,$61,$B5,$B2,$1F,$B6,$DF,$FB,$84
	.dc.b	$6A,$30,$2B,$A5,$7C,$BF,$17,$F9,$E6,$1F,$03,$BB,$1C,$4F,$9F,$A6
	.dc.b	$71,$FF,$04,$CE,$1F,$EF,$11,$97,$FB,$B3,$F2,$BE,$22,$46,$88,$7F
	.dc.b	$83,$DC,$BC,$0F,$F0,$17,$BB,$F6,$7C,$37,$C8,$D3,$CB,$12,$7F,$C5
	.dc.b	$F1,$97,$89,$FE,$0B,$E8,$E5,$DD,$23,$04,$3F,$C6,$0E,$5F,$F0,$74
	.dc.b	$7E,$3D,$FA,$5E,$0F,$E2,$FF,$09,$FC,$CC,$7F,$99,$DA,$FC,$E3,$67
	.dc.b	$F9,$EB,$8B,$FC,$28,$EF,$9F,$91,$D6,$33,$1C,$5B,$CD,$FE,$C0,$AE
	.dc.b	$AD,$AE,$17,$E5,$FA,$97,$D0,$FE,$7F,$EA,$7F,$9C,$65,$FE,$7B,$6B
	.dc.b	$E5,$9D,$BF,$E8,$E6,$BF,$EA,$88,$DE,$1F,$ED,$06,$F4,$7F,$9E,$EB
	.dc.b	$73,$FE,$D4,$FA,$8F,$D6,$5F,$FA,$C3,$DD,$FD,$AB,$FF,$00,$3C,$0F
	.dc.b	$D9,$4E,$AB,$87,$81,$FB,$5B,$DD,$F1,$BD,$CF,$D1,$FA,$97,$B7,$F4
	.dc.b	$FD,$A7,$D9,$53,$51,$30,$FF,$54,$CE,$9B,$39,$7E,$B5,$74,$72,$43
	.dc.b	$35,$E8,$E9,$F4,$F8,$32,$7F,$D4,$DA,$A7,$51,$89,$D3,$FE,$D2,$9D
	.dc.b	$3F,$ED,$9F,$9D,$79,$9F,$9D,$88,$65,$FD,$A4,$6A,$3F,$D1,$66,$8F
	.dc.b	$F4,$B3,$97,$B7,$DC,$7C,$EF,$62,$47,$9A,$3E,$89,$87,$C9,$18,$74
	.dc.b	$7B,$1D,$70,$FD,$D9,$FB,$3F,$57,$F6,$BE,$AA,$5F,$C2,$8C,$BF,$12
	.dc.b	$32,$FC,$18,$FC,$0E,$FF,$82,$FB,$C9,$3F,$2A,$32,$FC,$C8,$CB,$F2
	.dc.b	$63,$F6,$5F,$00,$BF,$DA,$7D,$63,$E1,$29,$A8,$97,$FE,$EA,$9D,$36
	.dc.b	$72,$FE,$DE,$BA,$39,$21,$9A,$FC,$CF,$D7,$FF,$03,$FC,$2F,$F6,$7D
	.dc.b	$7F,$ED,$D6,$D3,$D0,$71,$AF,$87,$81,$FA,$2F,$B4,$97,$FE,$FE,$C3
	.dc.b	$8E,$1C,$3D,$FF,$E4,$1E,$0C,$AF,$C7,$C7,$F1,$3F,$AD,$79,$34,$92
	.dc.b	$3F,$89,$57,$FB,$29,$C7,$37,$FC,$00,$AE,$31,$8D,$BF,$DE,$D5,$FD
	.dc.b	$CE,$30,$73,$70,$FF,$1D,$9E,$2F,$6F,$4F,$F9,$B7,$91,$23,$89,$FC
	.dc.b	$52,$3F,$95,$7C,$27,$E5,$F1,$63,$E7,$A4,$E6,$30,$D8,$F6,$6C,$F9
	.dc.b	$0E,$7D,$BF,$CD,$63,$FA,$F6,$1D,$2B,$7E,$9E,$9F,$1E,$1C,$D5,$70
	.dc.b	$C7,$C4,$D8,$F9,$53,$BF,$AB,$FF,$73,$47,$F8,$18,$F9,$CE,$1C,$35
	.dc.b	$FF,$05,$B5,$F7,$3B,$C3,$CD,$0F,$AC,$FF,$72,$3F,$21,$AF,$F4,$B9
	.dc.b	$E1,$87,$2F,$80,$FE,$A9,$5A,$BE,$6D,$9D,$5F,$D2,$FA,$A7,$C5,$FF
	.dc.b	$8A,$7E,$7D,$FA,$AF,$D0,$AF,$BF,$18,$D3,$FB,$05,$F0,$9E,$57,$C6
	.dc.b	$F8,$47,$F8,$32,$F8,$3A,$21,$98,$DE,$5F,$56,$77,$3F,$D9,$EF,$B3
	.dc.b	$7D,$DF,$37,$DD,$BE,$32,$9E,$48,$93,$E8,$9D,$36,$72,$F3,$30,$43
	.dc.b	$07,$F7,$7F,$A6,$FF,$1F,$3F,$0B,$EE,$FF,$9B,$BE,$D6,$E5,$F6,$27
	.dc.b	$73,$C7,$CB,$F9,$95,$AD,$FF,$65,$A2,$31,$5F,$B5,$BE,$38,$73,$C3
	.dc.b	$8F,$99,$2B,$F5,$35,$AC,$7F,$4C,$F2,$C7,$A7,$87,$E0,$3F,$C2,$D1
	.dc.b	$DA,$8E,$F2,$FE,$3B,$1E,$B9,$DE,$1F,$E4,$EF,$CA,$5F,$25,$49,$5F
	.dc.b	$D7,$BE,$6E,$BF,$E4,$AF,$83,$57,$08,$FE,$0F,$18,$AF,$EF,$F5,$FE
	.dc.b	$4D,$F8,$73,$FE,$6C,$89,$5F,$C3,$BE,$11,$5A,$8C,$35,$FF,$2C,$F8
	.dc.b	$BF,$E7,$AA,$E3,$1F,$E5,$CA,$E3,$7F,$E8,$2F,$49,$F2,$7F,$E7,$AF
	.dc.b	$4E,$5F,$F1,$CF,$61,$5A,$8C,$35,$EF,$F0,$CA,$E3,$1F,$E6,$AA,$E3
	.dc.b	$7D,$CC,$F3,$7F,$11,$D3,$FD,$1F,$79,$BE,$BE,$4F,$FA,$33,$D5,$91
	.dc.b	$82,$1E,$07,$2F,$7C,$FF,$97,$E1,$FC,$23,$EA,$DF,$1F,$FE,$28,$AF
	.dc.b	$E5,$51,$8D,$3F,$8B,$42,$30,$FF,$19,$6B,$FE,$A0,$F8,$AF,$E0,$BD
	.dc.b	$3C,$1B,$1F,$C7,$DA,$EF,$C1,$4F,$6A,$FE,$73,$F0,$77,$C3,$34,$97
	.dc.b	$FE,$AE,$C5,$F5,$7F,$EC,$1E,$EF,$A8,$D9,$FA,$3D,$9F,$CE,$F6,$FE
	.dc.b	$79,$7C,$3C,$CC,$61,$FD,$7E,$30,$F2,$8C,$DF,$F7,$3E,$6F,$F0,$44
	.dc.b	$E4,$EF,$F0,$BF,$CC,$A7,$96,$5F,$E8,$E9,$FE,$7D,$87,$9A,$B5,$5F
	.dc.b	$ED,$0A,$D3,$EB,$6E,$3E,$66,$DF,$FB,$67,$DB,$F1,$E1,$CE,$3C,$24
	.dc.b	$BF,$C7,$CE,$AB,$8C,$3A,$9F,$57,$FE,$00,$BF,$D4,$36,$7F,$2F,$EE
	.dc.b	$7C,$0A,$DF,$45,$AD,$67,$FD,$96,$73,$87,$F7,$EE,$AF,$F7,$2E,$1B
	.dc.b	$C3,$E4,$C3,$E0,$7F,$CE,$35,$FB,$EF,$AB,$7C,$0E,$A7,$3A,$7F,$05
	.dc.b	$9C,$D7,$59,$CE,$7F,$DD,$7C,$25,$6A,$FF,$D1,$1F,$01,$F6,$27,$38
	.dc.b	$FD,$A9,$7D,$4F,$EF,$32,$FC,$EB,$63,$FB,$A3,$5F,$F7,$DB,$5F,$F0
	.dc.b	$73,$BC,$7F,$B8,$B6,$7F,$C2,$5F,$78,$F9,$2F,$04,$E6,$FF,$C5,$8B
	.dc.b	$EA,$FF,$1B,$DA,$FA,$CF,$E3,$38,$FF,$88,$2B,$4F,$F6,$C8,$CD,$FF
	.dc.b	$B1,$67,$2F,$A2,$FF,$1A,$FB,$9B,$1A,$8F,$5C,$79,$0F,$89,$F8,$8B
	.dc.b	$E2,$31,$87,$86,$BC,$BF,$5B,$F4,$23,$E0,$C6,$38,$3D,$AF,$62,$47
	.dc.b	$CC,$69,$E7,$3E,$97,$DE,$F3,$7E,$19,$F5,$7E,$4B,$1A,$B0,$7D,$16
	.dc.b	$71,$F1,$B7,$1F,$E2,$B7,$F8,$8C,$3F,$C8,$41,$C3,$FE,$45,$6B,$FE
	.dc.b	$5D,$7C,$4E,$EB,$CB,$5D,$63,$FD,$E5,$F5,$AF,$9A,$FE,$07,$C6,$FA
	.dc.b	$EF,$F5,$C6,$5F,$E5,$6A,$D3,$FE,$60,$9D,$7A,$7F,$91,$9C,$E1,$FE
	.dc.b	$41,$6B,$E8,$6B,$FE,$5B,$6B,$FE,$73,$78,$CF,$F9,$01,$B5,$FE,$77
	.dc.b	$ED,$7C,$7F,$E4,$63,$C9,$97,$E8,$7E,$B3,$F6,$51,$FD,$9F,$EB,$5F
	.dc.b	$5D,$9E,$EF,$E0,$11,$89,$FF,$38,$D7,$0D,$3F,$CB,$0D,$7E,$7B,$5F
	.dc.b	$F3,$93,$5F,$F4,$53,$C7,$4F,$F2,$BB,$67,$FD,$1D,$F4,$6F,$AE,$7C
	.dc.b	$67,$8E,$7F,$D1,$3E,$6B,$EB,$9F,$59,$FC,$D2,$30,$FE,$F7,$DB,$FC
	.dc.b	$D5,$70,$AD,$BF,$E8,$A8,$CE,$9F,$E7,$16,$BF,$88,$6B,$FE,$8B,$6B
	.dc.b	$FD,$4A,$5F,$5D,$DF,$B7,$C5,$18,$BE,$D1,$FC,$6F,$E2,$A4,$7F,$19
	.dc.b	$FA,$EB,$83,$83,$FC,$F3,$97,$F8,$72,$BF,$B7,$B6,$7C,$DE,$EF,$EF
	.dc.b	$23,$11,$FE,$9F,$F2,$BF,$96,$7C,$6F,$F2,$93,$F9,$57,$C6,$DC,$3F
	.dc.b	$D5,$3F,$5D,$F0,$AF,$C1,$1F,$EB,$D1,$F5,$DF,$CF,$1F,$F5,$54,$EF
	.dc.b	$3F,$F5,$14,$6B,$CF,$FE,$55,$3C,$EF,$F5,$ED,$78,$F1,$7F,$12,$8C
	.dc.b	$47,$FB,$2D,$E9,$5A,$3F,$D8,$A7,$8B,$E9,$BA,$D7,$F4,$3F,$AE,$FE
	.dc.b	$89,$E1,$DA,$FE,$84,$DC,$3F,$C7,$7D,$9F,$C1,$A7,$FA,$35,$70,$7F
	.dc.b	$A4,$47,$F4,$DF,$AE,$FE,$9C,$D8,$C9,$FD,$36,$32,$D3,$FA,$87,$CD
	.dc.b	$FF,$55,$FA,$EF,$EA,$CD,$8C,$9F,$D5,$63,$2D,$33,$87,$9A,$32,$D8
	.dc.b	$FF,$2C,$7D,$67,$F9,$A1,$FE,$D5,$F4,$BF,$E1,$9E,$0F,$ED,$D3,$9A
	.dc.b	$FF,$2C,$CE,$6B,$F8,$04,$E6,$B7,$F1,$3F,$DA,$FE,$0F,$FB,$D4,$E6
	.dc.b	$BF,$DF,$9C,$DF,$DC,$FE,$A7,$FB,$87,$C8,$FF,$71,$6C,$FF,$6F,$6E
	.dc.b	$7F,$39,$B9,$7F,$B4,$27,$FD,$86,$7D,$AE,$3F,$E5,$6E,$2B,$F1,$E7
	.dc.b	$FB,$DA,$33,$97,$F5,$69,$D4,$FF,$38,$6B,$FE,$FB,$6B,$F9,$78,$FE
	.dc.b	$F5,$1D,$6B,$F2,$BC,$CF,$9A,$CD,$7F,$82,$FE,$7F,$FC,$19,$ED,$5C
	.dc.b	$3F,$C1,$4D,$C7,$F1,$2D,$C7,$FC,$E1,$C9,$FE,$0B,$7F,$DD,$BF,$5C
	.dc.b	$F8,$7F,$F0,$DF,$CC,$7F,$7B,$C1,$39,$C7,$FC,$45,$39,$8C,$4F,$C2
	.dc.b	$37,$5C,$BE,$B7,$FC,$29,$3A,$C3,$FD,$F7,$E5,$7F,$A1,$F1,$FF,$16
	.dc.b	$7D,$77,$FB,$65,$63,$CE,$74,$FF,$A7,$27,$4E,$7E,$43,$FC,$07,$3A
	.dc.b	$8F,$EC,$AD,$79,$87,$0F,$F3,$48,$FF,$14,$47,$6B,$FF,$4A,$57,$F5
	.dc.b	$0A,$D5,$FE,$6B,$CB,$FF,$1F,$F4,$B3,$FC,$8C,$61,$B3,$B3,$FC,$79
	.dc.b	$E7,$E2,$FF,$D6,$5D,$3F,$E9,$49,$FD,$1F,$D7,$7E,$6B,$E3,$FA,$B6
	.dc.b	$3F,$53,$1F,$E4,$9F,$95,$C5,$F2,$F6,$BF,$CD,$FE,$B3,$1F,$F0,$26
	.dc.b	$1F,$E0,$6A,$FF,$32,$46,$1F,$F0,$4C,$EE,$B8,$CE,$77,$FD,$9C,$7F
	.dc.b	$97,$3E,$31,$CB,$FA,$E8,$C3,$31,$FF,$32,$4B,$E9,$71,$97,$FB,$12
	.dc.b	$AE,$1F,$EB,$FA,$D5,$F0,$9F,$F3,$97,$D6,$7F,$80,$B3,$34,$9E,$C7
	.dc.b	$B9,$FA,$91,$A8,$F3,$7D,$27,$27,$FD,$13,$F4,$8F,$A9,$FF,$49,$FC
	.dc.b	$82,$31,$97,$FA,$61,$BA,$FF,$AB,$DA,$FE,$22,$78,$57,$43,$E2,$7D
	.dc.b	$15,$BF,$F7,$35,$F2,$DB,$FD,$62,$D8,$FE,$27,$B2,$B8,$63,$FE,$E5
	.dc.b	$AD,$2C,$7B,$DF,$CC,$AD,$E1,$D3,$C4,$F5,$E0,$69,$2B,$D8,$FC,$23
	.dc.b	$ED,$74,$58,$FD,$6D,$7F,$AD,$BE,$BB,$F5,$FC,$0F,$A4,$FF,$26,$FB
	.dc.b	$8B,$CC,$7E,$2E,$AE,$BF,$AC,$AD,$3F,$62,$74,$BD,$6E,$8E,$EF,$F7
	.dc.b	$B1,$88,$FD,$AD,$E2,$BC,$35,$67,$FC,$BF,$3E,$07,$F8,$2D,$F8,$30
	.dc.b	$E5,$1F,$C2,$27,$4F,$FB,$B2,$AA,$8C,$D7,$1B,$F3,$DF,$28,$C5,$78
	.dc.b	$0F,$54,$F9,$1A,$67,$0E,$18,$79,$E3,$D6,$7C,$33,$E0,$14,$CD,$F0
	.dc.b	$70,$7A,$4B,$7E,$27,$77,$E1,$9F,$7C,$7A,$CD,$D7,$BB,$0F,$6B,$63
	.dc.b	$CE,$7C,$87,$CC,$F7,$C3,$C4,$74,$9E,$31,$E4,$3B,$57,$9C,$D3,$D0
	.dc.b	$F9,$C5,$3C,$93,$D2,$FF,$06,$D8,$F9,$11,$F2,$E3,$9E,$DF,$87,$9F
	.dc.b	$A0,$D7,$E5,$B5,$F8,$CE,$7E,$61,$E5,$FF,$93,$E7,$F5,$55,$FD,$02
	.dc.b	$73,$A7,$F4,$CB,$E1,$7B,$6C,$66,$B8,$4F,$37,$F2,$D4,$F2,$FF,$CD
	.dc.b	$18,$F5,$6E,$3F,$D6,$27,$BB,$F4,$1B,$5C,$A7,$72,$9E,$8F,$E2,$5D
	.dc.b	$D7,$8B,$0F,$D0,$46,$A7,$F1,$4D,$CB,$CB,$8F,$C1,$6D,$73,$C3,$BD
	.dc.b	$71,$C7,$38,$79,$70,$E1,$8F,$AE,$3B,$DF,$F6,$8C,$76,$49,$D4,$F8
	.dc.b	$A7,$51,$E7,$8F,$2D,$74,$7E,$24,$E9,$A7,$B2,$34,$FC,$37,$E8,$4F
	.dc.b	$6B,$F9,$13,$F1,$67,$2F,$CF,$78,$E1,$F0,$EB,$CB,$87,$9E,$B5,$8F
	.dc.b	$B1,$87,$99,$F1,$1F,$1D,$87,$C5,$7C,$07,$08,$F1,$C7,$13,$F7,$D9
	.dc.b	$7F,$4D,$F9,$6F,$F0,$7D,$7F,$42,$6E,$BF,$91,$E1,$46,$2D,$E6,$B9
	.dc.b	$1E,$CE,$72,$E9,$5F,$D5,$EF,$FB,$E3,$FD,$1F,$0E,$6E,$4F,$EB,$87
	.dc.b	$17,$FC,$42,$FF,$4F,$3C,$18,$77,$FB,$9D,$44,$3F,$AD,$9A,$C5,$19
	.dc.b	$8E,$46,$8F,$F0,$0B,$97,$FB,$45,$79,$9F,$31,$E8,$C3,$FB,$93,$EA
	.dc.b	$EA,$FE,$FF,$F5,$FF,$0D,$B1,$FD,$5B,$4F,$F8,$01,$B2,$BC,$9E,$0E
	.dc.b	$E3,$15,$FE,$1F,$BE,$2F,$F8,$7A,$32,$DA,$F9,$3A,$7C,$FE,$FF,$71
	.dc.b	$27,$FC,$37,$E5,$57,$FC,$FB,$3B,$7F,$C3,$F3,$F1,$67,$C5,$F5,$F2
	.dc.b	$7D,$03,$1F,$F0,$DC,$73,$E2,$FE,$CF,$18,$9F,$46,$4A,$F8,$8D,$8C
	.dc.b	$C6,$3D,$AF,$80,$48,$EF,$93,$CD,$7C,$8E,$67,$FD,$57,$F3,$3F,$80
	.dc.b	$F2,$CE,$74,$F1,$E7,$FD,$CD,$BD,$1F,$34,$FC,$3E,$3F,$CC,$E3,$38
	.dc.b	$7F,$AE,$BB,$BA,$1B,$7E,$08,$FA,$FF,$F1,$93,$86,$BF,$E8,$76,$CF
	.dc.b	$F8,$2B,$47,$D7,$B8,$BF,$4D,$7C,$FC,$4F,$C6,$61,$CA,$3F,$3A,$D7
	.dc.b	$FC,$FE,$D7,$FD,$16,$F6,$E6,$FF,$62,$DF,$E4,$A7,$EE,$B7,$1F,$F0
	.dc.b	$DC,$77,$3D,$7F,$5D,$FE,$B1,$6F,$AD,$E4,$8C,$47,$E1,$BE,$BD,$ED
	.dc.b	$E6,$BF,$5F,$1F,$D5,$A7,$EB,$B5,$FF,$54,$57,$2B,$F8,$5F,$5C,$F8
	.dc.b	$6F,$A3,$DB,$2F,$EF,$CE,$5B,$39,$7F,$2F,$C5,$FE,$4B,$72,$7D,$4F
	.dc.b	$8C,$7C,$87,$CC,$6C,$66,$3E,$7F,$B5,$F0,$F8,$3F,$D5,$35,$AA,$FF
	.dc.b	$46,$35,$FE,$1A,$D7,$75,$FC,$06,$38,$67,$F3,$2F,$FD,$0D,$5F,$86
	.dc.b	$6B,$FE,$C9,$9D,$DF,$FB,$EB,$27,$C7,$FF,$18,$5F,$4D,$4C,$44,$BF
	.dc.b	$8B,$19,$7E,$3C,$7F,$0D,$FA,$27,$C4,$52,$4F,$CC,$F8,$2A,$FF,$00
	.dc.b	$EE,$FE,$53,$AF,$E5,$5B,$9F,$FB,$5B,$C2,$63,$D2,$FE,$AF,$D1,$F9
	.dc.b	$FD,$F9,$18,$21,$FC,$33,$64,$6A,$31,$3E,$F3,$19,$7E,$12,$71,$3F
	.dc.b	$82,$FA,$1F,$89,$F4,$52,$30,$43,$B3,$97,$C1,$1F,$2F,$DA,$F9,$9F
	.dc.b	$41,$27,$F1,$B1,$97,$F1,$F1,$97,$D1,$1F,$B4,$F8,$45,$D2,$4F,$F0
	.dc.b	$2F,$AF,$FC,$92,$CB,$F5,$B5,$C2,$3F,$AD,$B5,$FF,$6C,$35,$FE,$B2
	.dc.b	$D7,$FD,$F6,$EE,$3F,$6F,$3F,$B0,$9F,$07,$77,$89,$F9,$DE,$E7,$E2
	.dc.b	$48,$FE,$AB,$DD,$FE,$1B,$F5,$FF,$C0,$E3,$FB,$3B,$51,$1C,$37,$D1
	.dc.b	$FE,$97,$C2,$F6,$FF,$2C,$49,$FD,$BF,$BB,$FC,$6B,$C4,$7C,$0F,$F1
	.dc.b	$CF,$AF,$7A,$34,$D4,$5B,$FE,$FF,$3A,$6C,$FF,$6D,$8C,$47,$FB,$FE
	.dc.b	$BF,$C0,$6F,$CF,$EA,$F4,$A8,$CF,$0F,$F1,$E6,$CE,$AB,$9C,$FF,$22
	.dc.b	$F4,$FF,$39,$7C,$CF,$86,$E2,$35,$DD,$FC,$DB,$37,$C8,$7F,$38,$AE
	.dc.b	$37,$FE,$12,$6B,$E4,$6B,$D2,$BF,$C1,$11,$D7,$1F,$4D,$BE,$4B,$B1
	.dc.b	$D6,$BF,$A2,$FD,$7B,$8E,$93,$1F,$EA,$7F,$5F,$F9,$2D,$3F,$9E,$D6
	.dc.b	$E7,$FC,$56,$D7,$FC,$4E,$1C,$3F,$E2,$A6,$BF,$31,$E1,$1F,$D2,$27
	.dc.b	$FB,$8C,$F8,$3D,$6F,$13,$F2,$47,$D7,$3E,$37,$FA,$EB,$63,$FB,$4B
	.dc.b	$5F,$A1,$F3,$DF,$C3,$BE,$87,$75,$FE,$2D,$AD,$57,$F8,$F9,$AF,$E0
	.dc.b	$1A,$FE,$69,$E9,$8F,$FB,$E2,$DF,$5E,$E1,$FF,$2A,$46,$1E,$F7,$E4
	.dc.b	$F8,$4F,$39,$1F,$F0,$FC,$FF,$22,$F8,$2C,$47,$CA,$C7,$FA,$F3,$E5
	.dc.b	$AF,$41,$1E,$3F,$5E,$FA,$FF,$0B,$63,$C0,$7F,$83,$3E,$37,$FA,$0B
	.dc.b	$7C,$FF,$F0,$2C,$62,$3E,$77,$B4,$FB,$3C,$D7,$F9,$4E,$B5,$5F,$E6
	.dc.b	$46,$BF,$A6,$6B,$FC,$01,$E9,$BF,$C7,$FB,$0F,$EA,$73,$F0,$1A,$FD
	.dc.b	$D9,$DF,$57,$3F,$4B,$CA,$49,$F8,$FF,$66,$F8,$6C,$47,$E4,$2F,$9B
	.dc.b	$E5,$AF,$41,$FC,$3B,$E2,$7F,$D1,$1F,$56,$FB,$17,$AF,$86,$DF,$F8
	.dc.b	$1A,$BF,$B4,$19,$99,$19,$3F,$CA,$FE,$B3,$E3,$7F,$D2,$31,$86,$BF
	.dc.b	$C0,$63,$55,$FE,$85,$F2,$FF,$CF,$15,$AE,$0F,$F2,$EB,$77,$FE,$CC
	.dc.b	$DC,$3F,$D3,$FF,$19,$FE,$0D,$5E,$FF,$F6,$AF,$B1,$FF,$50,$7B,$46
	.dc.b	$9F,$E9,$3C,$2D,$7F,$9B,$AB,$83,$6B,$83,$FE,$64,$BF,$F1,$DE,$2B
	.dc.b	$FE,$77,$7C,$2B,$FD,$4A,$D7,$F9,$3B,$5D,$DF,$47,$6F,$F2,$69,$CF
	.dc.b	$99,$E8,$3F,$D5,$BF,$66,$FB,$3C,$5F,$96,$6B,$FE,$8C,$9D,$3F,$E6
	.dc.b	$8F,$B3,$FC,$26,$3F,$EB,$E6,$C7,$F1,$46,$BF,$EB,$C6,$BF,$CF,$BE
	.dc.b	$47,$F8,$67,$85,$FE,$C6,$6C,$7F,$1C,$6B,$FE,$C4,$6B,$FD,$1A,$7F
	.dc.b	$81,$7C,$5F,$87,$C4,$34,$93,$F2,$3E,$CF,$FD,$B7,$AF,$D3,$6C,$7F
	.dc.b	$AA,$38,$3F,$C0,$B3,$9F,$17,$E9,$4F,$F4,$BD,$17,$C3,$87,$31,$8A
	.dc.b	$DC,$7F,$A1,$70,$E7,$7C,$6F,$86,$1F,$71,$FF,$4E,$BE,$3F,$8F,$E2
	.dc.b	$FF,$7A,$F8,$DF,$F7,$77,$C0,$62,$FF,$23,$7C,$1E,$97,$9C,$FC,$AE
	.dc.b	$EF,$8E,$1D,$49,$5C,$1F,$C8,$61,$C7,$0E,$99,$7F,$B0,$3E,$CF,$FB
	.dc.b	$FD,$7E,$C1,$B7,$F1,$FE,$07,$C9,$D6,$9C,$DF,$1E,$5C,$C6,$30,$F6
	.dc.b	$6C,$C3,$35,$E7,$F8,$8F,$F3,$FD,$7F,$B5,$3B,$DF,$66,$F8,$8D,$7C
	.dc.b	$4F,$C3,$9C,$C6,$2B,$85,$6D,$FF,$67,$DF,$4D,$BF,$C0,$97,$C2,$7F
	.dc.b	$D8,$3F,$63,$FB,$96,$F7,$7C,$FA,$7D,$9F,$FB,$5A,$72,$48,$FE,$CD
	.dc.b	$5C,$5F,$ED,$13,$B7,$35,$CE,$5E,$9D,$3B,$7D,$97,$FA,$A7,$0F,$1C
	.dc.b	$65,$A7,$FB,$5A,$35,$7E,$09,$D3,$67,$5F,$1B,$8E,$3F,$F6,$2E,$1F
	.dc.b	$E4,$49,$F7,$D7,$B7,$EC,$BE,$25,$6A,$3C,$C4,$9F,$F7,$2C,$E9,$FF
	.dc.b	$1A,$D7,$39,$F8,$26,$CF,$F0,$44,$ED,$FF,$29,$C6,$EB,$FD,$AF,$F2
	.dc.b	$79,$9F,$F7,$47,$0F,$8F,$E9,$FE,$84,$E4,$91,$9C,$7F,$DE,$F1,$86
	.dc.b	$C6,$6B,$FC,$31,$3D,$7E,$C1,$F3,$DF,$EE,$69,$FF,$26,$C6,$2F,$FC
	.dc.b	$53,$08,$C3,$FE,$2C,$6B,$FE,$54,$35,$3F,$85,$8D,$E3,$FC,$1B,$A5
	.dc.b	$F2,$5F,$81,$8F,$F7,$AC,$EB,$0F,$0D,$F1,$3F,$53,$5B,$8C,$9F,$E7
	.dc.b	$AA,$FF,$7E,$FD,$7B,$97,$F5,$FF,$60,$FA,$2F,$CB,$4E,$6F,$FB,$2C
	.dc.b	$E5,$E4,$7F,$9B,$23,$5A,$7F,$8F,$5B,$87,$F1,$EF,$A5,$7D,$87,$FA
	.dc.b	$16,$F8,$DF,$FA,$32,$F7,$1C,$F2,$F0,$9F,$C5,$6F,$8F,$C6,$62,$78
	.dc.b	$F8,$EF,$AF,$7D,$6F,$F1,$B8,$C3,$FA,$FF,$47,$F3,$BA,$FF,$30,$9C
	.dc.b	$DF,$8B,$4F,$F0,$07,$4E,$CE,$58,$BE,$C3,$15,$E0,$8C,$3C,$71,$EA
	.dc.b	$FF,$A9,$B0,$F2,$57,$0C,$3F,$D1,$1E,$47,$CF,$F8,$AF,$E2,$87,$F1
	.dc.b	$5F,$B0,$B8,$73,$F5,$BD,$07,$F1,$9F,$8D,$B1,$FC,$41,$C3,$FC,$4B
	.dc.b	$E3,$2E,$3F,$E6,$DF,$B0,$F8,$73,$ED,$6C,$7F,$A6,$61,$F5,$AF,$A9
	.dc.b	$FE,$29,$5A,$E2,$FE,$AB,$68,$C5,$FF,$AA,$9B,$1F,$E0,$BB,$E5,$3D
	.dc.b	$F8,$F9,$91,$FB,$E7,$83,$0C,$57,$C4,$E6,$7D,$87,$F3,$7F,$0B,$F9
	.dc.b	$C3,$6F,$F9,$7B,$6B,$83,$73,$FE,$D7,$C3,$F9,$BC,$3F,$A0,$7D,$87
	.dc.b	$F4,$16,$E1,$FC,$E1,$B8,$78,$5B,$7C,$97,$CC,$FF,$49,$FB,$0F,$E9
	.dc.b	$4D,$BF,$E8,$2D,$BF,$53,$6F,$93,$67,$FB,$85,$F4,$7F,$A4,$DF,$F5
	.dc.b	$EF,$AD,$FF,$2C,$C7,$F9,$49,$B8,$7C,$5F,$90,$D7,$C1,$4A,$D3,$FD
	.dc.b	$76,$31,$3F,$2F,$EB,$3F,$B0,$7C,$7F,$F6,$16,$DF,$F5,$86,$CE,$E3
	.dc.b	$FA,$DC,$FF,$55,$7F,$D6,$87,$E9,$4B,$1F,$89,$E3,$FF,$75,$46,$6F
	.dc.b	$81,$1F,$BF,$1A,$9F,$F6,$04,$72,$D7,$F1,$AD,$3F,$DE,$A5,$8F,$C6
	.dc.b	$7C,$C7,$37,$F0,$DF,$3D,$FD,$EF,$D5,$FE,$F8,$DA,$FE,$E6,$D9,$DC
	.dc.b	$7F,$76,$9F,$EE,$1F,$61,$FE,$BD,$8F,$E0,$0D,$9F,$DA,$FC,$C3,$EB
	.dc.b	$F7,$F0,$52,$F6,$FF,$88,$67,$51,$83,$F8,$57,$8E,$FA,$3F,$EF,$CD
	.dc.b	$8F,$E3,$FD,$0F,$B0,$FE,$71,$90,$56,$DF,$F7,$33,$77,$FD,$9B,$6F
	.dc.b	$FC,$75,$5C,$AB,$8E,$1E,$02,$4F,$82,$75,$5C,$69,$E8,$B9,$BF,$9A
	.dc.b	$47,$FA,$75,$B8,$FE,$8F,$CE,$FC,$B7,$97,$32,$FC,$B3,$FE,$2D,$8C
	.dc.b	$5F,$1F,$8F,$FC,$E1,$FE,$31,$F8,$FB,$7F,$9D,$8B,$3F,$C0,$66,$E1
	.dc.b	$FD,$7F,$17,$A9,$FA,$53,$FC,$A1,$F0,$0E,$1C,$65,$FA,$79,$B8,$FF
	.dc.b	$9B,$E6,$E1,$FE,$67,$78,$9C,$BE,$92,$DF,$FA,$96,$2D,$7F,$9D,$E6
	.dc.b	$D7,$F7,$1F,$AC,$FA,$FB,$9F,$B3,$C6,$CD,$F4,$7F,$E5,$89,$D9,$3D
	.dc.b	$7F,$4C,$65,$FF,$36,$7C,$83,$E5,$E9,$EE,$F9,$FB,$BF,$FA,$F6,$74
	.dc.b	$D9,$D3,$49,$B7,$FA,$2E,$76,$FF,$A4,$6B,$9B,$9A,$BA,$FF,$A2,$E6
	.dc.b	$DF,$FB,$12,$FF,$9D,$DF,$3F,$2B,$E8,$F3,$7A,$19,$5E,$57,$2D,$3F
	.dc.b	$B6,$4F,$E7,$7E,$C3,$F3,$DA,$BE,$B7,$FC,$41,$EA,$FF,$A8,$A3,$15
	.dc.b	$FE,$23,$AD,$D6,$AE,$CE,$4F,$A5,$19,$69,$9A,$E6,$FF,$73,$9E,$2F
	.dc.b	$EB,$C5,$61,$FF,$5E,$C5,$72,$74,$39,$14,$F6,$16,$FD,$57,$FE,$3B
	.dc.b	$8E,$37,$E2,$7A,$9C,$4C,$C7,$A6,$35,$18,$3C,$D3,$D1,$F6,$4E,$CE
	.dc.b	$D7,$E9,$9D,$4F,$B6,$31,$7E,$B6,$C7,$85,$B7,$E0,$6D,$F4,$9F,$65
	.dc.b	$71,$69,$C4,$B5,$C5,$B3,$C6,$3E,$05,$F9,$27,$E3,$C6,$2B,$A3,$63
	.dc.b	$A3,$4E,$9F,$1A,$F5,$FC,$8D,$9F,$C1,$35,$F8,$01,$C7,$C8,$3C,$87
	.dc.b	$07,$F2,$F9,$CC,$62,$B8,$5F,$E1,$6F,$85,$EE,$31,$5B,$BE,$33,$F8
	.dc.b	$CA,$79,$3F,$8D,$C7,$F2,$31,$8C,$75,$87,$79,$CB,$74,$F0,$5F,$16
	.dc.b	$D7,$0C,$23,$F9,$3C,$3A,$36,$B8,$3F,$D2,$E3,$D2,$EE,$FB,$39,$9D
	.dc.b	$A9,$E7,$87,$92,$FB,$BE,$AA,$DC,$EB,$0F,$87,$5A,$7C,$F7,$E3,$30
	.dc.b	$78,$F0,$F0,$BD,$EF,$99,$D2,$33,$7E,$8C,$3C,$15,$F3,$6B,$58,$74
	.dc.b	$8E,$46,$4C,$57,$1C,$3C,$77,$CA,$7E,$15,$6A,$FC,$31,$C3,$0F,$0C
	.dc.b	$62,$36,$7F,$0C,$C5,$C0,$F8,$FF,$E1,$58,$35,$FE,$51,$18,$6D,$FF
	.dc.b	$80,$39,$9E,$B6,$63,$F2,$D8,$73,$7F,$9F,$5E,$71,$EE,$EF,$A3,$FA
	.dc.b	$75,$E6,$FF,$D9,$19,$78,$70,$E6,$64,$E1,$8B,$EC,$31,$1F,$D2,$CA
	.dc.b	$7F,$A5,$23,$4E,$48,$7F,$4B,$39,$C7,$09,$F0,$65,$F1,$27,$19,$FF
	.dc.b	$86,$F2,$FE,$C9,$7E,$FE,$AF,$ED,$1F,$61,$78,$0D,$3C,$FE,$0D,$C7
	.dc.b	$FD,$8B,$5C,$1B,$5C,$1F,$99,$3E,$EE,$F9,$1F,$3D,$F8,$3E,$36,$23
	.dc.b	$DD,$17,$5F,$F4,$3E,$CA,$F2,$7D,$84,$BE,$EE,$F6,$FF,$1B,$67,$FE
	.dc.b	$01,$8D,$4F,$F0,$C9,$CE,$69,$E1,$3F,$E2,$E8,$7C,$6F,$C1,$24,$7F
	.dc.b	$91,$3D,$27,$D1,$62,$3A,$1C,$67,$1A,$7F,$A5,$67,$E5,$74,$4A,$F8
	.dc.b	$AF,$E1,$EF,$F9,$25,$7F,$86,$63,$53,$D3,$3F,$EB,$93,$FE,$9F,$EA
	.dc.b	$FC,$A4,$BE,$C3,$FC,$28,$FF,$B9,$5A,$F9,$0D,$78,$3F,$95,$FB,$0F
	.dc.b	$F1,$BC,$6E,$39,$CF,$0B,$FF,$65,$B7,$7F,$C3,$5F,$0B,$E9,$39,$7F
	.dc.b	$CC,$8D,$7B,$9F,$E6,$A8,$E1,$5F,$ED,$6F,$AF,$7A,$5E,$A6,$D7,$FB
	.dc.b	$F5,$AF,$FB,$CB,$EC,$3F,$D7,$AE,$67,$81,$1F,$F3,$BC,$69,$FF,$7C
	.dc.b	$FD,$5B,$E1,$7E,$77,$64,$9F,$F3,$EC,$E5,$B3,$97,$FD,$BD,$3F,$BB
	.dc.b	$9D,$47,$CB,$F8,$97,$C5,$FD,$26,$C7,$FA,$17,$3F,$F7,$9C,$62,$3E
	.dc.b	$97,$AD,$FB,$17,$95,$71,$23,$C7,$7F,$F1,$04,$EE,$F9,$D6,$AB,$89
	.dc.b	$B3,$FD,$29,$3B,$AF,$DE,$60,$F9,$FF,$57,$D1,$4B,$F8,$31,$ED,$30
	.dc.b	$FC,$88,$C3,$FB,$CF,$9F,$7C,$2F,$F0,$56,$C6,$63,$E3,$7B,$D9,$64
	.dc.b	$64,$F9,$F1,$F2,$FE,$BF,$E6,$7B,$52,$7E,$9C,$65,$FA,$B1,$97,$ED
	.dc.b	$C6,$5F,$BB,$19,$DB,$F9,$67,$8C,$D1,$7F,$67,$E6,$FE,$D7,$D7,$CC
	.dc.b	$7F,$03,$19,$7F,$07,$19,$7F,$13,$1F,$81,$F5,$BF,$05,$F3,$52,$7F
	.dc.b	$13,$19,$7F,$17,$19,$7F,$23,$1F,$B6,$F8,$05,$D2,$4F,$D9,$1F,$5F
	.dc.b	$FE,$23,$71,$3C,$1B,$3C,$1F,$DE,$4E,$DF,$DF,$4E,$DB,$3B,$D3,$FA
	.dc.b	$84,$76,$E7,$FC,$E7,$B3,$23,$F4,$6F,$E7,$FD,$AF,$E1,$BF,$5F,$F8
	.dc.b	$B9,$FF,$70,$35,$39,$E0,$FA,$15,$FA,$7F,$66,$57,$EB,$9F,$D5,$7B
	.dc.b	$5F,$C6,$3B,$DF,$5F,$FC,$6B,$EB,$FF,$97,$DF,$C3,$6B,$FC,$66,$76
	.dc.b	$E7,$CB,$E0,$47,$E5,$46,$9E,$1E,$B7,$72,$3C,$4F,$A1,$E7,$FE,$69
	.dc.b	$FE,$4C,$DE,$0F,$8E,$DE,$FF,$F4,$96,$6F,$8E,$FE,$63,$39,$B1,$FD
	.dc.b	$57,$B7,$27,$F3,$6A,$E1,$1E,$08,$C7,$C8,$48,$C1,$F8,$3F,$AF,$FF
	.dc.b	$74,$37,$93,$F9,$FC,$62,$3F,$9E,$7D,$7F,$E0,$74,$FB,$B3,$96,$D7
	.dc.b	$03,$FA,$3C,$EF,$4F,$F5,$65,$73,$BF,$F5,$70,$70,$EA,$BF,$D4,$CF
	.dc.b	$BA,$32,$FC,$0F,$AE,$7C,$5F,$F5,$46,$C7,$E7,$5A,$FF,$66,$FA,$67
	.dc.b	$88,$F9,$DE,$17,$2B,$83,$FE,$31,$AD,$CE,$67,$FA,$E5,$F4,$D3,$F3
	.dc.b	$AD,$CF,$E9,$DF,$0A,$F1,$7C,$07,$94,$8F,$F2,$98,$F1,$F1,$E2,$3A
	.dc.b	$CD,$C3,$FC,$71,$5A,$9E,$64,$36,$73,$CF,$FC,$8C,$FF,$71,$F8,$77
	.dc.b	$91,$FD,$E1,$B1,$FC,$05,$AF,$F7,$7F,$49,$F5,$9B,$A9,$3B,$7F,$C0
	.dc.b	$33,$A8,$C1,$0F,$04,$6A,$7C,$1E,$5C,$9F,$03,$F3,$7E,$AD,$EF,$66
	.dc.b	$7D,$13,$74,$F8,$35,$A9,$FE,$D3,$F0,$5F,$E6,$5F,$A3,$7D,$5B,$D0
	.dc.b	$C3,$6B,$64,$3F,$C4,$F3,$C2,$76,$7F,$6F,$AF,$F2,$97,$B3,$FE,$72
	.dc.b	$AF,$F2,$6B,$5F,$F1,$23,$71,$FF,$22,$FC,$27,$0A,$95,$C1,$FF,$36
	.dc.b	$C7,$F9,$DF,$85,$F4,$74,$95,$FE,$81,$E6,$AF,$F8,$F3,$CA,$FC,$BC
	.dc.b	$E3,$1F,$F2,$9D,$53,$FD,$AD,$7D,$2F,$FB,$A5,$F0,$7A,$18,$8F,$E3
	.dc.b	$B5,$FE,$74,$6D,$72,$F1,$3F,$D1,$5F,$54,$FA,$BF,$44,$4C,$39,$3A
	.dc.b	$8F,$F2,$1F,$D5,$FD,$AC,$BF,$6E,$DA,$FD,$E8,$70,$FF,$00,$6B,$FE
	.dc.b	$EB,$EE,$75,$7F,$00,$F8,$AF,$DE,$B6,$33,$3D,$27,$2F,$6A,$FF,$3B
	.dc.b	$7C,$5F,$8B,$C7,$34,$93,$FC,$43,$17,$C0,$FF,$78,$DB,$FC,$39,$E7
	.dc.b	$B7,$7D,$77,$FA,$A7,$8B,$FD,$37,$7B,$C7,$FC,$E3,$7B,$AD,$61,$FC
	.dc.b	$D3,$1E,$BE,$C6,$2D,$7B,$AF,$AF,$87,$3C,$7F,$C1,$79,$78,$27,$FD
	.dc.b	$3B,$93,$ED,$B9,$60,$FC,$E5,$F3,$BD,$72,$F3,$D6,$A7,$38,$72,$8C
	.dc.b	$56,$F0,$E9,$3F,$02,$B7,$9F,$FA,$06,$72,$FB,$A3,$39,$FF,$B2,$F1
	.dc.b	$EF,$7C,$70,$F1,$7E,$25,$F4,$23,$E5,$E8,$F9,$3C,$F0,$7A,$A3,$13
	.dc.b	$FE,$BA,$AE,$19,$FC,$C9,$CB,$FE,$93,$8C,$E5,$F1,$B0,$F8,$7E,$F7
	.dc.b	$FB,$B7,$0F,$F3,$6F,$5E,$1C,$B5,$C9,$FE,$72,$32,$7F,$88,$FE,$39
	.dc.b	$F1,$5F,$5F,$A5,$EC,$3E,$8B,$ED,$6A,$BD,$E1,$FC,$EE,$B4,$F3,$C3
	.dc.b	$94,$62,$BF,$D4,$95,$C1,$FE,$A9,$19,$D3,$FC,$4F,$2F,$9C,$E3,$F2
	.dc.b	$BE,$08,$C6,$3D,$E3,$13,$A8,$F5,$37,$0E,$ED,$BE,$AD,$9F,$F0,$4C
	.dc.b	$BE,$FD,$F0,$5F,$97,$E3,$7E,$00,$F4,$3E,$0D,$6B,$0F,$F4,$15,$6A
	.dc.b	$73,$7C,$6F,$85,$7F,$93,$EF,$86,$1F,$ED,$8C,$38,$4E,$6F,$86,$1C
	.dc.b	$A3,$AE,$1C,$AF,$FD,$99,$F5,$3F,$84,$CF,$F6,$9F,$6D,$FE,$8F,$BF
	.dc.b	$F2,$E4,$E5,$FF,$52,$46,$6B,$8F,$2F,$EC,$ED,$39,$7F,$CA,$F7,$C6
	.dc.b	$91,$AF,$3B,$F0,$33,$F9,$49,$FF,$0C,$B5,$F9,$6D,$7F,$C6,$AD,$7F
	.dc.b	$80,$FC,$BF,$EE,$BF,$69,$FE,$5D,$97,$F9,$CF,$C0,$4E,$7C,$7F,$C7
	.dc.b	$C6,$27,$F8,$AD,$70,$BF,$F7,$74,$E5,$FE,$27,$19,$CB,$FC,$E3,$39
	.dc.b	$6E,$1E,$1F,$03,$AB,$DB,$D1,$FE,$1B,$F2,$DF,$E0,$8C,$BC,$91,$8F
	.dc.b	$53,$51,$FC,$9A,$75,$9F,$8A,$B5,$3C,$09,$1F,$C9,$BC,$DF,$C0,$E5
	.dc.b	$E4,$AF,$85,$C3,$2F,$FD,$29,$F3,$4F,$85,$C6,$3F,$4A,$7F,$2F,$39
	.dc.b	$7E,$1D,$7E,$D2,$31,$3F,$0A,$3E,$77,$AD,$9B,$B7,$D9,$FA,$5F,$33
	.dc.b	$23,$D9,$C9,$70,$F0,$1F,$E6,$EC,$5F,$05,$F1,$A7,$C0,$DA,$F8,$9E
	.dc.b	$AF,$CA,$3E,$F7,$97,$63,$FD,$61,$E3,$FF,$3D,$C7,$FA,$4F,$8F,$F5
	.dc.b	$67,$F6,$73,$9E,$0F,$F6,$6C,$E6,$BF,$CF,$F3,$9A,$FF,$77,$F2,$7F
	.dc.b	$14,$8C,$4F,$F4,$EA,$E0,$A3,$3C,$DF,$ED,$7A,$E1,$CD,$FC,$A9,$FC
	.dc.b	$A7,$B3,$F9,$5F,$1A,$E5,$F9,$06,$E3,$F8,$E6,$E7,$FE,$E2,$E9,$FE
	.dc.b	$35,$5F,$E9,$4C,$91,$FE,$20,$8F,$CC,$F4,$7E,$87,$A3,$F4,$4D,$8C
	.dc.b	$9F,$9E,$8C,$BF,$A0,$F4,$FF,$51,$EE,$FE,$A5,$B1,$93,$F4,$B1,$96
	.dc.b	$99,$CF,$F4,$B1,$96,$C7,$FA,$AB,$2F,$EB,$9E,$C7,$F0,$F7,$E8,$F5
	.dc.b	$FF,$6D,$8C,$4F,$B7,$8B,$FC,$B2,$F1,$3F,$D6,$7E,$72,$77,$87,$D3
	.dc.b	$BE,$5E,$17,$F5,$87,$F8,$17,$CB,$7F,$03,$F8,$5B,$8F,$EF,$9B,$7F
	.dc.b	$BC,$6E,$7F,$3D,$FC,$07,$CF,$FF,$83,$69,$F4,$28,$CC,$79,$6C,$EA
	.dc.b	$33,$1F,$9E,$8C,$1F,$96,$8E,$93,$CE,$30,$73,$37,$3F,$A6,$3F,$1D
	.dc.b	$3C,$0F,$E1,$7F,$1D,$FC,$33,$E3,$ED,$7F,$09,$6D,$7F,$8A,$B8,$D1
	.dc.b	$8C,$BC,$F3,$FE,$28,$9C,$D7,$AE,$5F,$B9,$BE,$27,$FC,$4B,$FB,$53
	.dc.b	$E7,$BF,$9F,$4E,$6F,$FC,$7B,$39,$79,$46,$71,$FF,$1D,$CE,$5F,$E0
	.dc.b	$38,$76,$C3,$AE,$1F,$E0,$2F,$6F,$F9,$0C,$F9,$B1,$FA,$3F,$6B,$9F
	.dc.b	$2B,$E4,$B6,$3E,$47,$02,$31,$97,$D1,$8C,$4F,$D7,$86,$1D,$A3,$FD
	.dc.b	$FE,$76,$34,$E6,$3E,$EC,$F2,$7E,$A7,$E9,$1F,$CC,$2B,$F6,$9F,$A6
	.dc.b	$BE,$23,$FC,$37,$39,$AF,$F2,$AC,$E6,$BC,$53,$9E,$F4,$E6,$BA,$FA
	.dc.b	$5F,$E5,$48,$C6,$BF,$E6,$7E,$1F,$F0,$BD,$7C,$CF,$97,$FA,$7E,$B7
	.dc.b	$D6,$6E,$7F,$51,$B3,$B8,$FC,$C7,$6F,$EA,$F7,$FF,$24,$79,$FF,$B3
	.dc.b	$6C,$EB,$BB,$E3,$BF,$E5,$09,$D3,$4F,$F2,$9F,$C5,$7F,$94,$4F,$F2
	.dc.b	$97,$6D,$8C,$93,$1F,$F2,$94,$65,$FE,$87,$8F,$D4,$E5,$FF,$2D,$79
	.dc.b	$FF,$E7,$1C,$FE,$97,$17,$FA,$26,$31,$3A,$C7,$D7,$5A,$F7,$95,$FE
	.dc.b	$85,$8F,$F5,$0F,$B9,$FE,$65,$F3,$71,$EE,$FF,$91,$F9,$3F,$D0,$B1
	.dc.b	$89,$D6,$3F,$EC,$0A,$D3,$FC,$C7,$1F,$F3,$CD,$FE,$AB,$0E,$B3,$FE
	.dc.b	$74,$C3,$93,$4E,$58,$7E,$C3,$5F,$F2,$2D,$70,$7F,$D5,$18,$FF,$96
	.dc.b	$5C,$B5,$FF,$50,$2F,$85,$FF,$3F,$7D,$57,$A9,$B8,$79,$DB,$5F,$ED
	.dc.b	$FA,$E1,$3F,$E8,$8A,$FF,$37,$CF,$F7,$11,$F4,$2E,$1F,$F2,$CC,$7F
	.dc.b	$A2,$23,$0E,$3D,$53,$C2,$77,$EE,$7F,$0C,$AD,$CF,$F8,$30,$D1,$FE
	.dc.b	$64,$8D,$4F,$FB,$6A,$B9,$D7,$0A,$F0,$4E,$5E,$27,$82,$B8,$BE,$33
	.dc.b	$27,$03,$13,$CE,$37,$1E,$38,$C6,$1E,$0A,$ED,$18,$9E,$62,$F4,$FA
	.dc.b	$6A,$BC,$EB,$E0,$C6,$23,$E6,$57,$33,$D8,$F3,$3C,$F7,$E4,$77,$5C
	.dc.b	$9F,$0D,$F3,$9D,$47,$1A,$E9,$5E,$83,$83,$DA,$B8,$5F,$8E,$B8,$56
	.dc.b	$8E,$0F,$AC,$EE,$79,$23,$84,$F2,$8C,$D7,$81,$F9,$B5,$CD,$FA,$59
	.dc.b	$BE,$2B,$EC,$36,$33,$18,$3F,$D4,$BF,$11,$FC,$9F,$83,$FD,$F9,$18
	.dc.b	$EF,$FF,$42,$D7,$F3,$68,$C4,$EA,$3F,$23,$5C,$5D,$18,$AE,$38,$FE
	.dc.b	$4A,$7F,$AC,$4F,$D8,$6B,$F8,$C7,$94,$6E,$78,$E3,$E6,$F0,$BF,$D7
	.dc.b	$73,$CD,$B1,$E4,$BE,$55,$C1,$F2,$CE,$9A,$78,$63,$4A,$FC,$35,$DE
	.dc.b	$3C,$D5,$C5,$D5,$70,$AD,$C7,$43,$D5,$3B,$8C,$C7,$0C,$39,$B0,$F3
	.dc.b	$BE,$33,$18,$7B,$63,$06,$8E,$A6,$B0,$F8,$11,$F1,$E7,$13,$E3,$8C
	.dc.b	$3C,$5E,$8F,$23,$13,$C8,$F8,$0E,$63,$89,$23,$2F,$23,$11,$FD,$EB
	.dc.b	$77,$5F,$FB,$C2,$FF,$C3,$BC,$EE,$57,$A3,$9E,$1F,$E7,$4D,$C7,$96
	.dc.b	$17,$81,$97,$49,$FF,$1F,$63,$CE,$F8,$36,$3F,$A3,$5F,$1A,$FF,$7E
	.dc.b	$47,$F2,$A3,$CD,$E6,$56,$99,$C3,$BC,$65,$B1,$E0,$C9,$E9,$78,$1B
	.dc.b	$1F,$D5,$6B,$C0,$66,$64,$64,$87,$FA,$C3,$1F,$89,$1A,$8F,$E8,$D3
	.dc.b	$F0,$FA,$FF,$C4,$BE,$C7,$7F,$26,$57,$F6,$E8,$D6,$47,$31,$D7,$25
	.dc.b	$6A,$7E,$3F,$C2,$BE,$46,$92,$FE,$6C,$3E,$B3,$11,$E7,$8E,$0E,$A7
	.dc.b	$9C,$F8,$BE,$6B,$E2,$7A,$D2,$BF,$C0,$71,$97,$E5,$46,$5F,$F0,$7E
	.dc.b	$69,$FC,$17,$23,$E7,$5F,$33,$F7,$BD,$0F,$B7,$87,$F9,$6B,$87,$75
	.dc.b	$A8,$FF,$34,$35,$D1,$FE,$64,$8D,$D7,$E3,$E1,$E2,$FF,$12,$C3,$F2
	.dc.b	$FF,$3C,$F9,$7F,$C6,$B6,$BF,$CB,$15,$BE,$AF,$EE,$31,$CB,$BB,$77
	.dc.b	$23,$71,$DF,$F3,$E7,$B5,$E1,$FC,$8F,$F1,$35,$FE,$8D,$6E,$7F,$E1
	.dc.b	$5A,$DC,$FF,$99,$DA,$EF,$AB,$91,$D6,$E4,$7F,$9B,$6F,$9E,$5E,$66
	.dc.b	$C7,$F8,$22,$F9,$E1,$FE,$92,$3E,$87,$C1,$7C,$FF,$A4,$95,$F3,$23
	.dc.b	$E7,$98,$7F,$D1,$71,$87,$F6,$BF,$B1,$7E,$F9,$BE,$4F,$ED,$A0,$DF
	.dc.b	$0A,$4E,$88,$7E,$96,$79,$3F,$EB,$1F,$81,$7D,$47,$C9,$F3,$E4,$FF
	.dc.b	$A2,$E7,$2D,$9C,$BF,$36,$7F,$01,$3A,$8E,$93,$AC,$FF,$C8,$9A,$FF
	.dc.b	$96,$27,$91,$FC,$43,$E0,$BF,$DA,$0D,$F4,$7F,$88,$C1,$BE,$F4,$9D
	.dc.b	$10,$FE,$01,$3C,$9F,$D0,$FB,$DF,$C6,$BE,$05,$F2,$74,$FD,$EC,$49
	.dc.b	$FE,$43,$3A,$6C,$E5,$EB,$E9,$79,$3D,$E9,$1E,$98,$F3,$98,$7C,$71
	.dc.b	$87,$47,$BD,$D6,$9F,$CE,$B1,$57,$EC,$FE,$8B,$E0,$DC,$3F,$DA,$2E
	.dc.b	$5F,$90,$64,$FF,$68,$BF,$BB,$F1,$BF,$79,$E3,$49,$FF,$6C,$C6,$5F
	.dc.b	$F6,$DC,$65,$FF,$6C,$C7,$EE,$7F,$0B,$7C,$4F,$C9,$F9,$19,$5F,$EC
	.dc.b	$D9,$CB,$67,$2F,$F3,$1D,$95,$CE,$64,$FC,$E8,$DE,$9F,$E8,$7B,$E7
	.dc.b	$3F,$95,$ED,$FC,$E1,$1F,$F7,$8F,$81,$FC,$37,$F1,$67,$C6,$FE,$87
	.dc.b	$E2,$65,$7F,$B8,$67,$2F,$FB,$E3,$E5,$7F,$8F,$11,$FD,$5C,$FF,$21
	.dc.b	$FB,$0E,$9F,$19,$27,$FD,$FF,$15,$FF,$7E,$FD,$87,$F2,$73,$1A,$FF
	.dc.b	$72,$F3,$51,$99,$FE,$D7,$18,$7C,$97,$FC,$97,$F1,$1F,$C7,$42,$FF
	.dc.b	$D5,$11,$86,$F4,$FF,$72,$F8,$5F,$E6,$B1,$8A,$FE,$07,$F2,$DF,$C0
	.dc.b	$4E,$95,$23,$24,$3B,$79,$3F,$CB,$30,$EC,$DE,$4F,$21,$27,$D3,$F0
	.dc.b	$75,$FE,$D9,$A3,$3F,$E9,$06,$73,$F4,$DE,$2B,$E0,$CE,$70,$FE,$77
	.dc.b	$F4,$0F,$A1,$FF,$03,$52,$30,$FF,$82,$5A,$FF,$83,$5A,$FF,$AE,$1C
	.dc.b	$E1,$FE,$CF,$FA,$27,$97,$FD,$45,$B5,$FE,$BA,$9F,$F6,$A5,$70,$C3
	.dc.b	$C7,$86,$F0,$97,$B9,$CB,$4F,$F6,$1F,$E0,$8F,$89,$FE,$BD,$CB,$FC
	.dc.b	$56,$74,$1C,$3F,$40,$E3,$1F,$D9,$67,$94,$70,$9D,$E5,$FE,$E9,$6D
	.dc.b	$FF,$62,$AD,$D7,$30,$E2,$B8,$78,$DE,$43,$B6,$1F,$DC,$FE,$8F,$FC
	.dc.b	$1B,$7E,$3F,$81,$FD,$EF,$6F,$F9,$15,$B1,$EA,$9D,$E1,$FE,$FB,$6E
	.dc.b	$5F,$EE,$4A,$DD,$73,$EF,$FE,$BF,$3F,$DE,$2B,$73,$FE,$EA,$6D,$70
	.dc.b	$B5,$73,$BE,$37,$BF,$07,$D1,$F2,$3E,$59,$F8,$73,$FE,$58,$6A,$C5
	.dc.b	$3C,$23,$EC,$63,$FC,$6B,$4E,$46,$63,$F3,$7F,$15,$FE,$21,$FB,$29
	.dc.b	$1F,$B8,$8F,$F1,$21,$87,$F4,$31,$87,$F0,$9F,$06,$1F,$19,$FE,$59
	.dc.b	$DE,$4E,$A7,$2D,$9C,$BF,$E4,$B9,$FF,$2A,$4E,$A3,$F1,$0B,$EA,$FE
	.dc.b	$2F,$C3,$3E,$0F,$F3,$B5,$FE,$49,$E6,$FF,$2E,$57,$CA,$BE,$18,$7F
	.dc.b	$95,$26,$57,$A2,$B4,$74,$BE,$7C,$7F,$E8,$C6,$F1,$AB,$DD,$5D,$7F
	.dc.b	$1B,$FE,$AF,$1F,$E7,$9E,$7F,$F3,$C1,$FE,$6D,$8C,$C7,$A9,$B8,$7F
	.dc.b	$A6,$3E,$1F,$F4,$D3,$C4,$F9,$73,$AC,$3F,$C8,$3F,$44,$F9,$87,$C3
	.dc.b	$7F,$A3,$63,$0D,$8C,$DF,$FA,$4E,$72,$FF,$A5,$7E,$35,$C9,$FE,$A0
	.dc.b	$8C,$35,$FF,$27,$9F,$ED,$77,$0F,$FB,$E9,$AF,$C3,$9F,$F0,$5F,$A5
	.dc.b	$E3,$F1,$BD,$F3,$E8,$6B,$B6,$E3,$F8,$7F,$A2,$FF,$22,$67,$F3,$63
	.dc.b	$0F,$FA,$8F,$C8,$CD,$FC,$0E,$55,$E2,$F7,$3F,$C3,$5B,$C1,$F3,$6F
	.dc.b	$57,$C3,$0F,$CB,$F9,$3F,$96,$7F,$D7,$10,$8C,$61,$F0,$A3,$1D,$1F
	.dc.b	$B6,$8C,$4E,$9F,$F3,$AC,$EA,$BA,$D7,$A3,$A3,$F2,$DF,$21,$EA,$AD
	.dc.b	$5F,$E7,$30,$EF,$97,$59,$CB,$F9,$1A,$E1,$3A,$C7,$F2,$54,$F9,$AF
	.dc.b	$F1,$3C,$7E,$C9,$B8,$7F,$AD,$7E,$6B,$C1,$18,$C7,$B3,$6B,$9B,$69
	.dc.b	$97,$FB,$AE,$F8,$F0,$7A,$FD,$4F,$E2,$F8,$7F,$91,$BE,$8B,$0E,$5A
	.dc.b	$E4,$FF,$25,$99,$3F,$A2,$F9,$6F,$37,$E6,$65,$F9,$08,$C4,$FF,$BA
	.dc.b	$5B,$1F,$EE,$D6,$BB,$9F,$C2,$BF,$E2,$E6,$FC,$B7,$FB,$52,$BB,$C6
	.dc.b	$AB,$FB,$C4,$7F,$24,$FA,$27,$1E,$7C,$EF,$C7,$F8,$A9,$CD,$7A,$27
	.dc.b	$CF,$97,$FA,$D3,$0E,$6F,$96,$32,$FF,$07,$8C,$BE,$D8,$ED,$3F,$D9
	.dc.b	$B2,$6D,$8A,$FE,$B9,$F4,$4F,$9A,$FD,$0F,$03,$E8,$9F,$1F,$F0,$2B
	.dc.b	$59,$7F,$75,$AD,$4E,$6F,$C7,$82,$F6,$FE,$F8,$C4,$6A,$33,$1F,$C0
	.dc.b	$9D,$F2,$FD,$66,$E1,$F9,$2F,$AC,$7D,$5F,$78,$FF,$00,$FD,$1B,$E8
	.dc.b	$FF,$73,$83,$E9,$1F,$31,$F6,$2B,$57,$FE,$0D,$AD,$4E,$6F,$D3,$5A
	.dc.b	$CF,$D0,$D7,$F8,$BB,$5F,$F7,$3B,$BA,$FC,$6B,$6F,$D1,$E6,$75,$7F
	.dc.b	$C2,$FD,$6F,$8A,$FD,$C7,$C2,$FF,$9B,$27,$31,$8F,$81,$FC,$7D,$6E
	.dc.b	$7E,$56,$5F,$A8,$AF,$9F,$39,$8F,$9C,$D7,$F9,$6B,$9C,$BF,$42,$DA
	.dc.b	$FE,$9C,$B0,$FE,$29,$1E,$23,$CF,$E5,$7F,$0D,$F9,$97,$CA,$79,$23
	.dc.b	$0D,$9F,$2D,$FF,$49,$6C,$F9,$88,$F4,$3F,$C7,$D0,$F9,$9F,$BF,$97
	.dc.b	$92,$BE,$2E,$F2,$FF,$C7,$FF,$3D,$FE,$A6,$7F,$DB,$51,$FB,$48,$C5
	.dc.b	$7E,$3D,$AF,$F4,$E6,$BF,$DB,$43,$87,$F8,$9E,$7F,$E4,$FA,$EF,$EC
	.dc.b	$FD,$6F,$6F,$B3,$F6,$3E,$3E,$4F,$AB,$6F,$C7,$BF,$EE,$D8,$F2,$7C
	.dc.b	$9F,$E9,$32,$34,$93,$F6,$29,$F4,$9F,$3C,$FC,$2F,$9B,$67,$FC,$BF
	.dc.b	$E5,$7D,$6D,$7F,$D8,$1C,$9F,$C5,$23,$1C,$9F,$D2,$78,$DE,$57,$DF
	.dc.b	$BF,$F0,$65,$71,$7F,$B8,$B5,$F8,$8D,$77,$E6,$7E,$F2,$3A,$E3,$A8
	.dc.b	$FC,$CF,$CA,$78,$5F,$CD,$FC,$B7,$E7,$3E,$56,$D7,$DE,$E5,$46,$31
	.dc.b	$FF,$3B,$67,$F9,$18,$C6,$9F,$EF,$57,$F4,$BD,$1F,$A6,$E1,$B8,$F3
	.dc.b	$6E,$1C,$B1,$FE,$64,$DA,$E1,$5E,$1F,$67,$F5,$FE,$E7,$EC,$1B,$19
	.dc.b	$3C,$D1,$96,$99,$AE,$D1,$96,$C7,$F9,$53,$E8,$3F,$93,$BF,$9C,$E1
	.dc.b	$FD,$14,$FF,$93,$27,$2F,$FB,$DF,$AF,$FB,$B4,$63,$5F,$EE,$8D,$BF
	.dc.b	$DA,$E7,$FD,$12,$F9,$78,$DF,$C2,$3E,$7F,$F8,$4F,$95,$6F,$E7,$36
	.dc.b	$BE,$23,$71,$FC,$26,$1F,$D0,$AB,$F1,$7F,$29,$FE,$20,$5F,$46,$8F
	.dc.b	$C6,$19,$71,$1C,$0F,$E2,$7C,$F8,$C7,$F9,$F3,$5F,$C6,$D7,$E2,$4E
	.dc.b	$73,$FE,$BD,$3B,$FD,$56,$27,$F0,$7C,$6F,$96,$FE,$21,$F4,$BF,$C4
	.dc.b	$5B,$7F,$AF,$6D,$70,$6E,$9F,$E2,$91,$E8,$7F,$A5,$AB,$F8,$7F,$C6
	.dc.b	$BE,$73,$FA,$3C,$E7,$0F,$F7,$7C,$63,$E2,$25,$FE,$F7,$C5,$B5,$FE
	.dc.b	$C3,$F7,$7F,$91,$CF,$AF,$85,$F3,$1F,$E0,$69,$CD,$7F,$B4,$EB,$E6
	.dc.b	$CE,$7C,$AF,$EC,$F5,$AA,$FF,$31,$B5,$FE,$2C,$D7,$FC,$B1,$1C,$6B
	.dc.b	$C1,$EA,$59,$FF,$6B,$7D,$2F,$F2,$AF,$05,$F4,$AF,$A5,$FF,$0C,$CE
	.dc.b	$63,$18,$7F,$BA,$A3,$F9,$8D,$9D,$7C,$0F,$0F,$63,$E3,$FB,$5F,$E4
	.dc.b	$FE,$47,$D2,$63,$1F,$9A,$DC,$7F,$DE,$5C,$DF,$4A,$FF,$A5,$78,$59
	.dc.b	$8F,$CF,$F0,$FF,$86,$BE,$09,$E1,$FF,$A4,$32,$FF,$33,$B7,$1F,$F3
	.dc.b	$33,$5F,$F3,$1B,$9B,$E7,$3F,$D3,$3E,$25,$F1,$BF,$AE,$C7,$FC,$CD
	.dc.b	$18,$9D,$4E,$3C,$4C,$B2,$32,$73,$8C,$63,$FE,$C6,$BF,$A9,$97,$FA
	.dc.b	$9B,$9B,$D5,$18,$23,$FE,$73,$7F,$07,$97,$FA,$CF,$C1,$C4,$FF,$A3
	.dc.b	$7E,$07,$FC,$9F,$87,$FA,$55,$AF,$FA,$C9,$B6,$69,$FE,$58,$BF,$F4
	.dc.b	$44,$67,$E0,$FD,$FE,$1F,$56,$69,$FE,$CD,$AD,$DF,$18,$FF,$01,$EC
	.dc.b	$BD,$E3,$FE,$9F,$AD,$3E,$3C,$3A,$DF,$33,$D0,$FA,$FE,$23,$FC,$AB
	.dc.b	$97,$9D,$B7,$FE,$C4,$AF,$F2,$24,$E5,$B1,$9B,$F0,$57,$D8,$EC,$E9
	.dc.b	$5D,$E7,$FB,$E0,$FA,$57,$07,$FB,$33,$B7,$FD,$55,$18,$9D,$B5,$FF
	.dc.b	$77,$46,$A7,$FC,$11,$0E,$9D,$3F,$E9,$E6,$BF,$E0,$83,$51,$DE,$FF
	.dc.b	$DF,$13,$AF,$5B,$FD,$F5,$87,$2C,$3B,$61,$E2,$C3,$83,$E9,$7C,$33
	.dc.b	$BC,$38,$C6,$0D,$C7,$59,$F0,$9F,$E1,$BA,$FF,$64,$61,$93,$94,$68
	.dc.b	$FF,$10,$61,$D1,$E1,$85,$72,$78,$CC,$94,$CD,$F4,$3C,$45,$C3,$CD
	.dc.b	$1C,$23,$31,$E8,$3E,$6D,$F9,$0F,$33,$F0,$9D,$CF,$18,$E6,$FD,$6A
	.dc.b	$F3,$23,$E9,$4F,$39,$EA,$73,$2C,$7A,$A3,$BB,$63,$EF,$E9,$F7,$E3
	.dc.b	$F1,$6D,$7E,$1B,$5F,$06,$1E,$18,$D7,$C5,$7F,$2D,$E1,$C6,$1F,$E4
	.dc.b	$D8,$C5,$AB,$7D,$CB,$DC,$FA,$EB,$75,$A9,$FC,$5E,$3C,$FC,$3F,$F2
	.dc.b	$C6,$3F,$99,$BF,$EB,$58,$68,$E5,$8F,$57,$F0,$11,$E2,$9E,$15,$CA
	.dc.b	$FA,$57,$11,$5B,$8F,$55,$F3,$D3,$8D,$7E,$1E,$BC,$2F,$E1,$5F,$7C
	.dc.b	$79,$67,$C5,$87,$0A,$F1,$E1,$E1,$7C,$F8,$F8,$70,$E3,$8F,$92,$7A
	.dc.b	$57,$F9,$DF,$1E,$97,$E1,$C7,$C9,$7B,$9E,$C6,$AF,$C9,$3A,$7E,$75
	.dc.b	$F3,$AE,$07,$BA,$FA,$1E,$38,$F3,$4F,$C4,$6E,$1F,$16,$3D,$27,$98
	.dc.b	$F8,$91,$D5,$CE,$1E,$33,$CB,$7E,$08,$CB,$E1,$7E,$29,$E1,$3E,$26
	.dc.b	$1F,$2A,$35,$3C,$A7,$D6,$7A,$0F,$DC,$65,$FD,$17,$AF,$FD,$3D,$97
	.dc.b	$F3,$E6,$EB,$FD,$35,$B4,$F0,$1C,$AE,$0C,$E3,$CE,$BF,$9A,$DF,$FB
	.dc.b	$4A,$31,$1F,$E1,$8E,$3F,$F0,$1B,$FC,$EB,$CB,$E3,$06,$9E,$0C,$7B
	.dc.b	$CE,$0F,$E7,$95,$FD,$76,$71,$1F,$E8,$6C,$7F,$D7,$0F,$F5,$38,$CE
	.dc.b	$7F,$D6,$9E,$93,$A8,$C4,$F8,$C8,$79,$E7,$FD,$D7,$D3,$FB,$5F,$7B
	.dc.b	$E1,$B6,$7F,$A7,$35,$FF,$4B,$FC,$6B,$E7,$77,$92,$F9,$B6,$BC,$E2
	.dc.b	$F9,$3F,$ED,$16,$BF,$EF,$2B,$7C,$8F,$C5,$6C,$7F,$B9,$9A,$F8,$BE
	.dc.b	$77,$53,$23,$4F,$18,$F2,$FD,$2C,$9F,$5E,$08,$E1,$1F,$E1,$B7,$D2
	.dc.b	$47,$D3,$87,$F1,$98,$F5,$C7,$F1,$9F,$A5,$7D,$2F,$E2,$B6,$46,$27
	.dc.b	$FB,$B4,$62,$7F,$01,$E8,$3E,$3B,$85,$EE,$FF,$D8,$7E,$47,$E1,$DF
	.dc.b	$F3,$BF,$0E,$B0,$FC,$BC,$3C,$3F,$E2,$39,$3A,$3F,$DF,$91,$86,$F6
	.dc.b	$7E,$3F,$E9,$5C,$3C,$AB,$58,$7E,$86,$5F,$41,$B3,$57,$E0,$FA,$57
	.dc.b	$B3,$49,$7F,$E1,$31,$F4,$6F,$73,$31,$F8,$2A,$E2,$FC,$A6,$B9,$AF
	.dc.b	$9A,$A3,$3A,$FB,$63,$FB,$05,$BD,$EF,$86,$D8,$F1,$57,$0D,$BF,$C7
	.dc.b	$DC,$4F,$84,$F9,$E4,$AF,$F1,$57,$73,$E9,$5F,$0F,$99,$FA,$7F,$15
	.dc.b	$FB,$7E,$73,$9D,$BC,$98,$71,$F8,$2F,$DE,$62,$F8,$4A,$62,$26,$3E
	.dc.b	$09,$D3,$67,$2F,$F6,$5D,$7F,$BF,$2F,$85,$7D,$2F,$E0,$9B,$1F,$45
	.dc.b	$AF,$13,$E3,$D6,$E7,$E6,$CE,$A7,$F8,$5B,$5E,$ED,$7F,$C1,$CF,$4C
	.dc.b	$BE,$94,$7D,$EF,$89,$FA,$27,$57,$E5,$FC,$1F,$FB,$A5,$BE,$B7,$F1
	.dc.b	$F8,$FF,$49,$C6,$BD,$6C,$19,$99,$19,$3F,$6F,$F0,$3F,$C3,$88,$F6
	.dc.b	$F5,$EB,$EF,$E0,$7C,$2B,$E5,$BF,$97,$36,$79,$9E,$6F,$43,$C9,$EB
	.dc.b	$C8,$C1,$0E,$9A,$7F,$B7,$1F,$6C,$7A,$E6,$46,$08,$70,$7F,$11,$E7
	.dc.b	$7C,$18,$87,$FB,$C9,$CB,$F2,$0C,$9F,$C0,$9F,$C8,$79,$9F,$91,$F1
	.dc.b	$64,$FF,$BE,$A3,$2F,$FB,$F2,$32,$FF,$0A,$8F,$E0,$FF,$4A,$F8,$2A
	.dc.b	$49,$FD,$0F,$5B,$E0,$F1,$1F,$CE,$2B,$83,$FC,$81,$AE,$6B,$F6,$F1
	.dc.b	$8E,$5F,$D1,$EC,$F9,$4F,$A2,$74,$8F,$D3,$F5,$FE,$D7,$0F,$EE,$0D
	.dc.b	$E2,$FD,$97,$89,$FA,$FF,$1F,$F8,$FF,$D2,$BE,$3F,$F8,$2B,$70,$FE
	.dc.b	$66,$D7,$FA,$7B,$5F,$D4,$DB,$77,$D2,$FE,$EC,$93,$FB,$3F,$8C,$FD
	.dc.b	$F7,$D2,$F8,$3E,$3A,$4F,$C6,$8A,$F6,$8F,$19,$1F,$EB,$F8,$77,$FA
	.dc.b	$5F,$E1,$04,$8F,$E4,$D9,$2B,$F9,$DF,$D2,$FE,$56,$7F,$70,$1C,$70
	.dc.b	$7E,$C1,$BE,$2F,$FA,$07,$17,$C4,$7F,$47,$BE,$35,$CD,$FF,$03,$B5
	.dc.b	$FE,$16,$D7,$FD,$14,$D7,$C1,$DD,$FC,$6B,$E6,$E4,$7F,$3D,$9F,$0B
	.dc.b	$C3,$97,$D2,$47,$F1,$7F,$0F,$5F,$F0,$0C,$FE,$0D,$FE,$D2,$D7,$33
	.dc.b	$FD,$02,$31,$9F,$F2,$CE,$2F,$6D,$7E,$5B,$2E,$93,$FC,$4B,$E9,$1F
	.dc.b	$49,$FD,$91,$B5,$FD,$4D,$AF,$CC,$FA,$5F,$F4,$93,$99,$FE,$9F,$3A
	.dc.b	$8F,$F1,$6B,$5F,$E6,$2D,$7F,$D6,$0F,$4A,$FE,$47,$5F,$D1,$27,$FB
	.dc.b	$13,$5F,$B7,$3B,$DF,$F9,$DF,$D2,$FF,$4E,$9F,$ED,$0D,$7F,$02,$DB
	.dc.b	$FE,$5F,$CA,$F1,$7F,$C0,$33,$BA,$FF,$21,$B5,$EB,$E7,$FF,$6A,$8C
	.dc.b	$47,$FB,$1D,$E9,$87,$F4,$CB,$FE,$67,$3F,$90,$6E,$1F,$CF,$AB,$73
	.dc.b	$FE,$CF,$FA,$5F,$F1,$8C,$FF,$42,$6D,$FF,$56,$6C,$BE,$11,$F6,$18
	.dc.b	$84,$FF,$8E,$9F,$F2,$DB,$5E,$5E,$27,$F8,$0A,$31,$1E,$18,$FF,$3B
	.dc.b	$D7,$F1,$BE,$FB,$E1,$FF,$0B,$AE,$B5,$FE,$26,$BF,$E3,$91,$FE,$15
	.dc.b	$6B,$FA,$E6,$E5,$FE,$0D,$C3,$CF,$D9,$C7,$C4,$F1,$12,$FF,$C8,$37
	.dc.b	$C9,$FE,$1B,$CA,$F8,$DF,$F0,$37,$D3,$49,$FE,$5B,$1F,$E0,$93,$0F
	.dc.b	$F8,$FA,$30,$FF,$9B,$BE,$0D,$F3,$3F,$E7,$26,$C6,$63,$FC,$C9,$8E
	.dc.b	$59,$19,$3F,$CC,$D1,$FE,$7A,$FA,$6F,$E5,$D4,$69,$FC,$BF,$D2,$7D
	.dc.b	$1E,$F8,$A6,$1F,$E8,$0A,$D4,$E2,$3F,$D2,$37,$CE,$BF,$D1,$35,$FE
	.dc.b	$3A,$9F,$F3,$E3,$51,$FE,$81,$FA,$6F,$F5,$63,$FE,$8C,$6B,$FE,$17
	.dc.b	$A4,$63,$87,$FC,$5B,$F1,$3F,$E5,$4A,$E1,$E8,$7F,$A5,$AB,$5C,$FD
	.dc.b	$B0,$E3,$3F,$E8,$9F,$2F,$FC,$CF,$F4,$CF,$95,$FE,$03,$8F,$FA,$AA
	.dc.b	$31,$F1,$7F,$E7,$C6,$C7,$FA,$61,$AF,$FA,$C3,$E4,$BE,$F7,$4F,$FA
	.dc.b	$09,$B1,$FE,$9E,$6B,$BE,$6F,$D1,$3C,$FC,$9F,$7B,$FE,$3C,$BF,$F0
	.dc.b	$B6,$7E,$19,$CC,$FC,$89,$CB,$FE,$C0,$C7,$FD,$55,$F4,$DF,$E6,$3E
	.dc.b	$1F,$F3,$24,$63,$7F,$F2,$DF,$C7,$BC,$BD,$47,$D2,$9D,$61,$FE,$45
	.dc.b	$6D,$F2,$E1,$63,$C2,$BF,$DC,$98,$F3,$C3,$FD,$AB,$5A,$7F,$D8,$BE
	.dc.b	$6F,$E6,$1B,$D5,$C7,$CD,$73,$3E,$53,$A6,$1F,$D8,$6A,$E1,$C7,$2F
	.dc.b	$C1,$37,$4F,$C0,$DE,$AA,$65,$D8,$FF,$62,$46,$2F,$F1,$5F,$3B,$FE
	.dc.b	$F9,$9F,$F6,$CB,$63,$FD,$D5,$F3,$DE,$28,$C6,$1F,$E8,$DC,$91,$8B
	.dc.b	$FC,$63,$6F,$E1,$DF,$1E,$3F,$77,$AD,$FE,$B1,$BF,$F4,$5B,$FE,$BD
	.dc.b	$69,$AF,$8F,$DB,$4C,$9C,$1F,$A1,$7F,$E2,$9F,$4A,$FA,$47,$3F,$3A
	.dc.b	$F9,$1C,$C4,$C3,$FD,$65,$5B,$7F,$D6,$95,$B7,$33,$D3,$D3,$91,$AD
	.dc.b	$2E,$3F,$E1,$2C,$3F,$C1,$B3,$D2,$BC,$35,$AD,$FF,$D0,$75,$AB,$E1
	.dc.b	$5A,$BE,$07,$FB,$17,$E9,$5F,$37,$EE,$6C,$7F,$06,$C3,$E9,$7D,$2B
	.dc.b	$E9,$75,$72,$B9,$3F,$8A,$9D,$46,$09,$8F,$F8,$AB,$0E,$AF,$F0,$D9
	.dc.b	$DB,$9A,$F9,$FF,$48,$FA,$4F,$F4,$EC,$FF,$AC,$23,$18,$7F,$C0,$06
	.dc.b	$C7,$FA,$D0,$FE,$3A,$E1,$C3,$AC,$BF,$D4,$ED,$9F,$E2,$D8,$7F,$B2
	.dc.b	$9F,$F7,$44,$66,$B9,$DF,$90,$F7,$7C,$65,$AF,$F7,$77,$5F,$FB,$0D
	.dc.b	$FF,$57,$FB,$FF,$88,$8C,$5F,$FA,$62,$31,$F1,$32,$BF,$D6,$BD,$96
	.dc.b	$BF,$C9,$F7,$C3,$4F,$F5,$EB,$6B,$FC,$9D,$F0,$DF,$E5,$77,$C3,$87
	.dc.b	$7F,$23,$F7,$DC,$EF,$95,$C5,$7F,$BB,$23,$0F,$A2,$73,$97,$FA,$DF
	.dc.b	$E4,$FE,$AE,$7E,$09,$CB,$6B,$FD,$EB,$3A,$7F,$CF,$73,$FC,$37,$2F
	.dc.b	$91,$99,$F9,$58,$3E,$7B,$FD,$41,$7F,$ED,$4E,$0F,$D6,$C6,$27,$FC
	.dc.b	$DE,$DC,$BF,$DF,$76,$F7,$37,$EC,$F7,$6E,$3F,$ED,$E6,$E1,$FE,$C6
	.dc.b	$F4,$FC,$87,$C4,$F9,$AB,$5F,$9D,$3F,$DF,$79,$3E,$47,$C6,$7C,$AF
	.dc.b	$A4,$B3,$F2,$5C,$3F,$EE,$CF,$1E,$E3,$FE,$86,$F1,$7E,$A3,$79,$BE
	.dc.b	$27,$C8,$FF,$0A,$9C,$F0,$7F,$BE,$1B,$87,$F7,$4B,$FE,$3F,$7C,$23
	.dc.b	$CD,$1F,$DE,$27,$6D,$9E,$86,$DF,$04,$FF,$C0,$0F,$96,$FA,$9F,$27
	.dc.b	$F5,$5B,$19,$3A,$F2,$7D,$79,$D4,$63,$8D,$1F,$77,$9B,$EF,$73,$DA
	.dc.b	$D3,$71,$FF,$7A,$B7,$4F,$B2,$FE,$17,$E0,$BF,$0D,$EC,$DB,$EC,$DB
	.dc.b	$FF,$7E,$B6,$FF,$02,$D9,$FD,$0B,$65,$9F,$E3,$E3,$FA,$DF,$D0,$3E
	.dc.b	$6E,$9F,$95,$F4,$EE,$1F,$33,$CE,$56,$E3,$0D,$C5,$97,$E4,$AF,$8C
	.dc.b	$62,$BF,$67,$F3,$FF,$A1,$F2,$BF,$44,$D8,$F9,$4D,$9E,$07,$A3,$85
	.dc.b	$19,$C3,$F0,$9F,$2B,$F8,$73,$FC,$01,$6F,$A3,$FF,$48,$3F,$84,$C3
	.dc.b	$FB,$2C,$EE,$31,$CC,$8E,$35,$C2,$BF,$2D,$1B,$6D,$7E,$38,$ED,$F1
	.dc.b	$19,$C1,$1F,$ED,$DE,$37,$CA,$FF,$0E,$FA,$4F,$E1,$ED,$9F,$E1,$CD
	.dc.b	$C3,$DF,$C9,$FD,$96,$31,$87,$F8,$9F,$D1,$7C,$1F,$F8,$76,$11,$8C
	.dc.b	$3F,$8C,$78,$99,$8C,$63,$F2,$AB,$77,$93,$F9,$0D,$5B,$F8,$DE,$D7
	.dc.b	$F2,$0B,$FF,$7D,$B6,$FD,$B1,$EE,$DF,$FC,$18,$A7,$55,$FB,$56,$C7
	.dc.b	$F8,$E7,$C8,$F9,$BE,$B7,$F8,$1A,$73,$3C,$67,$06,$AB,$C6,$FF,$83
	.dc.b	$3E,$8F,$F9,$54,$FF,$9D,$1B,$B3,$E9,$3F,$C6,$D7,$F7,$E7,$74,$69
	.dc.b	$C2,$77,$1F,$8E,$35,$E8,$FC,$9F,$52,$5F,$F4,$3F,$36,$D7,$CB,$E1
	.dc.b	$FF,$04,$6F,$F5,$6F,$FA,$5D,$6A,$3F,$C6,$9D,$9F,$17,$0F,$D3,$ED
	.dc.b	$FE,$4D,$F6,$DF,$33,$FB,$28,$93,$A3,$FC,$C7,$18,$CF,$F0,$B5,$C2
	.dc.b	$72,$7F,$9A,$3A,$FA,$54,$8C,$D7,$09,$C3,$FE,$97,$BF,$AB,$DF,$F4
	.dc.b	$6B,$FC,$E3,$D7,$FE,$5E,$9C,$93,$8F,$D1,$19,$7F,$CD,$D1,$FE,$E7
	.dc.b	$F6,$2F,$C1,$7F,$8E,$71,$F3,$F6,$E2,$66,$3F,$CD,$AF,$83,$FE,$80
	.dc.b	$9C,$D7,$F9,$F6,$F9,$6F,$FE,$86,$6D,$7F,$AD,$39,$9B,$FE,$DE,$B5
	.dc.b	$97,$26,$D7,$53,$C1,$7C,$B0,$FC,$C3,$FE,$E1,$8C,$FC,$0F,$D5,$F9
	.dc.b	$9F,$5B,$23,$27,$FA,$5A,$31,$5F,$89,$3F,$D8,$FE,$7F,$FB,$23,$E8
	.dc.b	$FF,$8B,$C7,$F5,$7E,$BF,$D4,$C6,$23,$FB,$58,$70,$FF,$9F,$DA,$FF
	.dc.b	$80,$CD,$71,$FF,$8E,$67,$9B,$FE,$3B,$8D,$4F,$29,$CC,$7F,$8F,$EB
	.dc.b	$C9,$5C,$70,$B3,$93,$A9,$E7,$72,$F9,$AF,$A4,$62,$3D,$E7,$03,$FD
	.dc.b	$CF,$7E,$07,$85,$77,$74,$70,$AE,$6E,$0F,$4C,$6E,$3E,$0D,$F7,$AE
	.dc.b	$95,$C6,$B4,$7C,$19,$D5,$F9,$67,$99,$EB,$3B,$46,$9A,$79,$2B,$AB
	.dc.b	$C0,$F5,$CF,$39,$E1,$7E,$49,$E1,$39,$69,$D2,$3D,$A6,$CB,$7E,$78
	.dc.b	$DC,$73,$E0,$FF,$67,$47,$D3,$6B,$F8,$66,$BF,$41,$CF,$C2,$3E,$93
	.dc.b	$FB,$8C,$FF,$33,$F8,$1F,$F1,$34,$E7,$7F,$F7,$54,$62,$B7,$7C,$9B
	.dc.b	$19,$9F,$C0,$D3,$E8,$3F,$C6,$18,$FE,$66,$31,$8E,$B0,$F0,$CE,$5B
	.dc.b	$97,$8A,$F8,$B6,$B8,$61,$1F,$C5,$61,$D1,$B5,$C1,$08,$F6,$3C,$2F
	.dc.b	$BB,$A8,$E5,$87,$9E,$B9,$4F,$1A,$DB,$6F,$DF,$3A,$BE,$38,$7C,$3A
	.dc.b	$ED,$5A,$61,$82,$61,$DD,$CB,$F0,$E3,$DC,$E6,$3D,$A6,$27,$B4,$64
	.dc.b	$8E,$09,$1E,$DB,$F3,$1E,$7C,$3D,$78,$7C,$4B,$E3,$5B,$3B,$1F,$16
	.dc.b	$77,$5C,$A3,$39,$38,$1F,$49,$FE,$6F,$CF,$F9,$93,$75,$FA,$75,$B8
	.dc.b	$C5,$FF,$93,$B8,$DC,$18,$AD,$4F,$F3,$2A,$C6,$3D,$A3,$FC,$DB,$7C
	.dc.b	$1F,$E6,$FE,$5E,$88,$E8,$A7,$F4,$76,$E5,$FE,$16,$E5,$FE,$8B,$38
	.dc.b	$CF,$CF,$87,$F9,$7E,$3A,$9C,$CA,$73,$2D,$73,$C3,$FD,$D0,$FF,$88
	.dc.b	$BA,$7F,$C4,$1E,$B7,$C0,$6D,$7F,$99,$F4,$4F,$C4,$C9,$F0,$8F,$9D
	.dc.b	$E1,$5A,$BE,$17,$83,$85,$6E,$3F,$D2,$2B,$E4,$72,$48,$CF,$CE,$D7
	.dc.b	$FC,$97,$3B,$78,$CF,$49,$F0,$FD,$24,$9F,$43,$C7,$4F,$76,$28,$D4
	.dc.b	$FF,$82,$5D,$6F,$FB,$4A,$7D,$03,$E9,$3F,$07,$18,$6C,$66,$FF,$DD
	.dc.b	$DF,$13,$89,$91,$FE,$BA,$AE,$0E,$8F,$F3,$D7,$83,$FC,$16,$3F,$CA
	.dc.b	$71,$FE,$F9,$E7,$FC,$8B,$4F,$F5,$BF,$D2,$35,$CD,$F1,$CF,$FD,$94
	.dc.b	$DA,$FE,$1A,$DE,$25,$F1,$9F,$F4,$57,$D1,$BD,$7E,$ED,$8C,$C7,$8B
	.dc.b	$E9,$1E,$D6,$A2,$4E,$8F,$F4,$A4,$E9,$C1,$FE,$53,$AE,$8D,$9E,$95
	.dc.b	$F2,$BE,$06,$4F,$C7,$E1,$FF,$7A,$3F,$ED,$6F,$A2,$7C,$26,$A3,$FD
	.dc.b	$05,$3A,$6D,$7E,$56,$B7,$1C,$7D,$EC,$7C,$0C,$8C,$10,$FF,$4B,$CA
	.dc.b	$35,$18,$9F,$A7,$F0,$CF,$A4,$A4,$9F,$83,$55,$FF,$39,$4E,$DF,$F5
	.dc.b	$C3,$5F,$F3,$B6,$DD,$DC,$D7,$1E,$5F,$C1,$D6,$DF,$C2,$D6,$EB,$FD
	.dc.b	$5B,$EF,$7F,$AF,$DF,$F7,$C7,$A1,$FB,$58,$FF,$39,$C7,$FB,$EF,$D0
	.dc.b	$C1,$99,$91,$93,$F7,$DE,$DF,$EE,$88,$FE,$8F,$9F,$FC,$AD,$6F,$47
	.dc.b	$33,$FE,$71,$AB,$97,$F1,$7A,$D4,$F1,$3F,$DC,$1E,$47,$97,$CF,$91
	.dc.b	$9E,$95,$7F,$27,$8F,$6D,$69,$F7,$D7,$B6,$24,$71,$94,$EE,$FC,$55
	.dc.b	$FC,$5F,$E6,$3F,$88,$5F,$C8,$3E,$31,$9F,$16,$3C,$CC,$9E,$58,$E3
	.dc.b	$5F,$CC,$DF,$E2,$51,$FE,$DE,$7F,$58,$DE,$DF,$07,$87,$23,$94,$78
	.dc.b	$CC,$3F,$EB,$7F,$A4,$7B,$9F,$C0,$1B,$1F,$EF,$F6,$BD,$FD,$DD,$4C
	.dc.b	$8D,$3E,$F8,$C4,$75,$8B,$8F,$A6,$7F,$4D,$DD,$2B,$F5,$CF,$C7,$F9
	.dc.b	$5F,$E2,$3F,$44,$F8,$CA,$4A,$D9,$F1,$7E,$93,$E7,$11,$FD,$C7,$C3
	.dc.b	$FD,$1F,$A4,$E7,$F1,$52,$7E,$64,$57,$FD,$B1,$82,$71,$A7,$CF,$AF
	.dc.b	$B5,$F3,$D2,$FE,$A3,$AE,$AF,$F1,$5C,$3E,$91,$F3,$5F,$CA,$9B,$1F
	.dc.b	$93,$35,$1F,$D6,$FE,$0B,$FB,$54,$7F,$BA,$67,$6D,$9E,$46,$FD,$7B
	.dc.b	$1F,$C5,$7E,$33,$F8,$81,$87,$97,$CB,$F8,$48,$FE,$2F,$E0,$AB,$A3
	.dc.b	$C9,$82,$7F,$31,$3A,$D7,$F2,$19,$7E,$6C,$C6,$7F,$03,$E8,$9F,$49
	.dc.b	$FD,$21,$B9,$7B,$DA,$AF,$FA,$7F,$D2,$3D,$8E,$17,$2B,$83,$FD,$4A
	.dc.b	$75,$18,$3F,$D6,$C2,$FA,$BF,$48,$C4,$E9,$A7,$0F,$A4,$FE,$6A,$F8
	.dc.b	$F1,$FB,$31,$8C,$7F,$B1,$B6,$3E,$EF,$2F,$F0,$CA,$E0,$7F,$9A,$A7
	.dc.b	$4D,$F4,$7F,$B4,$C6,$21,$3C,$6F,$AD,$68,$93,$F8,$59,$D5,$FE,$33
	.dc.b	$E8,$FF,$7F,$7F,$DD,$9B,$19,$F0,$3F,$C8,$FE,$FB,$E9,$33,$8F,$F9
	.dc.b	$FA,$EA,$69,$A9,$E1,$5F,$E0,$08,$CB,$4E,$B5,$C3,$BF,$44,$BF,$E5
	.dc.b	$D7,$E1,$E8,$C4,$7C,$63,$C1,$38,$C9,$97,$FA,$4A,$F8,$57,$F1,$FF
	.dc.b	$88,$FE,$51,$F4,$92,$30,$43,$F9,$03,$97,$F9,$14,$7F,$92,$7E,$05
	.dc.b	$F2,$D4,$93,$FE,$40,$CA,$BF,$6A,$47,$07,$FA,$43,$89,$FF,$24,$CF
	.dc.b	$F9,$57,$E8,$FF,$C1,$79,$FF,$96,$5B,$9F,$FA,$B2,$D1,$8C,$1E,$6B
	.dc.b	$E6,$5F,$49,$CF,$C7,$95,$E0,$BE,$17,$BA,$CC,$F9,$E7,$A6,$1D,$AF
	.dc.b	$FC,$C9,$87,$4A,$F4,$FD,$27,$F5,$BC,$FF,$CD,$CD,$CF,$CC,$D9,$F2
	.dc.b	$FC,$2B,$C9,$FF,$39,$4E,$72,$34,$E1,$AF,$F9,$CA,$7F,$C8,$FF,$48
	.dc.b	$F9,$1F,$DB,$E3,$FB,$38,$C7,$7F,$EE,$1B,$1F,$B7,$0E,$1F,$DD,$7C
	.dc.b	$6B,$93,$FD,$55,$1B,$8C,$73,$7F,$02,$6C,$66,$76,$F8,$7C,$7F,$EF
	.dc.b	$8F,$F3,$6B,$FE,$6B,$E9,$78,$E7,$33,$FE,$E4,$9C,$BF,$C2,$F6,$FE
	.dc.b	$89,$CF,$F9,$88,$FF,$41,$F9,$97,$8F,$FD,$35,$E5,$FE,$43,$93,$FD
	.dc.b	$1B,$39,$7F,$09,$19,$AE,$31,$F8,$E8,$C5,$7F,$AB,$F9,$17,$C6,$FF
	.dc.b	$2B,$4F,$9B,$C7,$27,$FA,$8B,$CC,$D6,$9F,$D6,$67,$38,$A7,$37,$C7
	.dc.b	$3F,$F5,$0C,$62,$B6,$D8,$C9,$E5,$8C,$BF,$E8,$BF,$BA,$FE,$7E,$FF
	.dc.b	$AA,$BB,$BC,$DA,$A7,$3A,$7F,$AA,$27,$31,$88,$FF,$71,$35,$FF,$50
	.dc.b	$35,$FE,$8C,$DD,$1F,$35,$FE,$A4,$C3,$FC,$49,$CD,$F0,$A3,$17,$FE
	.dc.b	$23,$8C,$7C,$1F,$C6,$F5,$7F,$A7,$5F,$F8,$FE,$F7,$9F,$F8,$AE,$FE
	.dc.b	$76,$1F,$D8,$7C,$BF,$A1,$D3,$FB,$6F,$3F,$3E,$72,$73,$E1,$27,$31
	.dc.b	$86,$E7,$FC,$92,$74,$FF,$80,$69,$E2,$79,$7E,$3F,$AB,$67,$4D,$CB
	.dc.b	$F9,$55,$70,$9F,$F2,$25,$70,$69,$C0,$B3,$FC,$C7,$E8,$1F,$3F,$F5
	.dc.b	$BE,$9B,$FD,$D7,$5F,$E1,$39,$CB,$FE,$C7,$8C,$D7,$1A,$E1,$F2,$1F
	.dc.b	$1A,$F7,$3F,$D8,$B8,$17,$C1,$FF,$65,$CE,$A7,$FB,$A3,$5F,$F1,$3B
	.dc.b	$50,$70,$BE,$DD,$EA,$7E,$7B,$EE,$7F,$D4,$59,$FF,$8A,$E7,$2F,$F3
	.dc.b	$88,$CD,$7F,$51,$9C,$BF,$CF,$E3,$38,$7F,$B8,$27,$2F,$EF,$7B,$93
	.dc.b	$AB,$FF,$1E,$D6,$9B,$87,$ED,$7E,$D1,$F7,$3F,$CF,$27,$C9,$3F,$8C
	.dc.b	$EA,$FE,$49,$1F,$91,$F9,$0C,$79,$3F,$99,$6C,$64,$FF,$23,$4F,$0D
	.dc.b	$7F,$C1,$EF,$E8,$A7,$71,$FD,$B6,$35,$AF,$F1,$57,$FD,$CF,$F3,$BA
	.dc.b	$8F,$F7,$EC,$7F,$95,$2F,$FD,$41,$F1,$FF,$E8,$07,$BF,$91,$FC,$32
	.dc.b	$31,$87,$58,$C7,$BF,$EB,$6D,$70,$24,$FF,$6B,$FA,$8F,$C1,$47,$F9
	.dc.b	$93,$0F,$04,$FA,$BB,$F2,$FF,$64,$6A,$F9,$C7,$C8,$FE,$F7,$A3,$F6
	.dc.b	$B1,$8C,$3F,$75,$18,$9D,$47,$EF,$7E,$97,$0D,$3F,$CE,$15,$C6,$FC
	.dc.b	$5F,$31,$FC,$24,$FE,$5D,$F7,$0F,$C2,$DE,$4E,$70,$E3,$E1,$6A,$73
	.dc.b	$18,$3F,$D9,$D5,$C6,$B5,$BC,$9C,$D3,$E1,$31,$97,$F8,$97,$E9,$9F
	.dc.b	$21,$FC,$FE,$DF,$4B,$87,$FD,$11,$36,$BF,$CA,$31,$AA,$FE,$33,$39
	.dc.b	$E7,$FE,$AD,$39,$38,$CF,$5B,$D4,$7F,$A6,$2A,$DF,$F8,$D2,$6A,$8C
	.dc.b	$CF,$38,$FB,$5F,$37,$FC,$C7,$EB,$3F,$99,$36,$7F,$97,$36,$77,$1F
	.dc.b	$CC,$67,$F9,$C7,$C8,$FF,$39,$FA,$7B,$3F,$CD,$5B,$5F,$87,$6E,$7F
	.dc.b	$9D,$E7,$47,$F4,$1F,$88,$FE,$85,$DB,$70,$FE,$7C,$DC,$3F,$20,$DC
	.dc.b	$39,$36,$7F,$88,$F5,$7F,$4E,$8C,$57,$E5,$5F,$EB,$BF,$4A,$F8,$3F
	.dc.b	$E9,$9F,$3D,$86,$E1,$FC,$6B,$1F,$F0,$25,$6E,$3F,$DF,$55,$B6,$9F
	.dc.b	$D4,$FA,$3F,$AD,$76,$7F,$5B,$6C,$64,$FE,$B1,$19,$69,$FE,$CB,$F4
	.dc.b	$9E,$17,$DD,$F9,$3F,$EB,$71,$89,$F2,$1F,$CB,$8F,$C6,$C6,$A7,$84
	.dc.b	$73,$6E,$BF,$E7,$9A,$E4,$FE,$43,$99,$7C,$6F,$AC,$FE,$26,$3F,$BA
	.dc.b	$7C,$C7,$F7,$5F,$AF,$B5,$FD,$C5,$B5,$FC,$0D,$BB,$7D,$D7,$FB,$9F
	.dc.b	$D4,$38,$7F,$C3,$FF,$50,$F4,$B3,$18,$6E,$3E,$2E,$45,$6E,$3F,$DE
	.dc.b	$17,$CD,$D4,$67,$2F,$76,$7F,$E2,$CF,$B8,$7C,$7F,$CA,$8C,$36,$FB
	.dc.b	$4E,$9F,$A5,$3A,$69,$A9,$E1,$5F,$CE,$AB,$83,$FD,$A7,$97,$FC,$19
	.dc.b	$5E,$9F,$69,$F4,$BF,$90,$F6,$BF,$C3,$4D,$9D,$36,$FF,$A2,$D7,$09
	.dc.b	$FA,$7C,$2C,$3F,$CD,$51,$D7,$C4,$FC,$DF,$C3,$7F,$89,$5B,$19,$38
	.dc.b	$CF,$0C,$3A,$CF,$E8,$BA,$BF,$C9,$BE,$85,$BF,$A1,$F0,$5F,$D9,$6F
	.dc.b	$FD,$FD,$82,$F7,$5D,$1B,$3F,$D7,$3E,$2B,$FA,$1F,$27,$FA,$97,$CE
	.dc.b	$FF,$2E,$C7,$E4,$35,$FE,$B8,$E8,$91,$AE,$CD,$BF,$B6,$F2,$3F,$CC
	.dc.b	$3C,$57,$3F,$2B,$72,$FE,$D7,$AF,$D0,$7E,$95,$70,$AF,$F6,$0F,$E7
	.dc.b	$1F,$5B,$3F,$F3,$A7,$BB,$FE,$6E,$8C,$4F,$F0,$8C,$3F,$C6,$98,$7F
	.dc.b	$B2,$E7,$2F,$FA,$B2,$B8,$73,$7F,$9D,$1B,$7D,$A3,$DF,$E7,$62,$B9
	.dc.b	$45,$C7,$E0,$CD,$AF,$F6,$76,$1F,$E7,$BB,$F5,$F2,$79,$7E,$0B,$FD
	.dc.b	$02,$DA,$FA,$AD,$9D,$C7,$EE,$1F,$F1,$89,$FD,$C4,$7D,$3B,$EA,$33
	.dc.b	$0F,$8A,$FF,$51,$4E,$63,$3E,$9F,$F8,$EA,$31,$1F,$E3,$C9,$DD,$7F
	.dc.b	$AF,$EB,$4F,$48,$CC,$F2,$3C,$31,$9B,$DC,$66,$FA,$BE,$38,$CB,$EA
	.dc.b	$BA,$E4,$E6,$71,$AE,$8E,$0E,$05,$38,$57,$67,$E2,$3E,$0B,$F3,$C7
	.dc.b	$08,$C5,$F8,$23,$75,$DA,$79,$5F,$92,$F8,$D7,$F9,$82,$31,$7C,$5B
	.dc.b	$1E,$33,$99,$F5,$A7,$8B,$83,$E7,$4F,$17,$D5,$1A,$9E,$75,$E2,$9C
	.dc.b	$D6,$E3,$D3,$1D,$E7,$77,$F5,$E3,$66,$4B,$3C,$CE,$27,$5F,$15,$F5
	.dc.b	$3C,$5B,$3F,$75,$AF,$D0,$6F,$C6,$7F,$3D,$DF,$FD,$07,$18,$F0,$D1
	.dc.b	$8B,$FC,$BD,$6E,$D7,$BA,$D4,$7F,$A3,$27,$58,$7E,$51,$FE,$29,$1F
	.dc.b	$88,$6B,$F8,$37,$78,$F6,$C7,$C3,$39,$EA,$FD,$3D,$F3,$6D,$6F,$0E
	.dc.b	$97,$C2,$B5,$7D,$EF,$6F,$96,$75,$7C,$63,$B1,$B3,$9D,$C7,$24,$BF
	.dc.b	$65,$72,$8E,$93,$C5,$D5,$75,$AD,$CE,$5F,$2C,$FC,$1A,$E4,$B0,$F8
	.dc.b	$6F,$CB,$C3,$D2,$E5,$B1,$A8,$E1,$3C,$A3,$E6,$D7,$63,$C3,$3D,$EF
	.dc.b	$C5,$87,$DC,$C3,$9E,$1C,$4E,$4F,$63,$33,$F0,$EF,$A5,$71,$BC,$3F
	.dc.b	$C9,$BE,$3F,$F8,$ED,$7F,$A9,$79,$7F,$C1,$55,$FD,$63,$1C,$ED,$B8
	.dc.b	$FE,$83,$78,$5D,$6B,$E3,$1F,$D2,$AF,$5F,$00,$E4,$61,$85,$F4,$14
	.dc.b	$97,$FE,$17,$ED,$FF,$2D,$BF,$D4,$23,$2F,$F4,$B7,$C5,$92,$77,$39
	.dc.b	$3F,$B8,$EF,$AC,$FF,$37,$AF,$FB,$6F,$1F,$0D,$75,$36,$43,$91,$FE
	.dc.b	$13,$E1,$E5,$E1,$48,$FF,$01,$C7,$C3,$AF,$F0,$63,$98,$FE,$D7,$39
	.dc.b	$EF,$FD,$9C,$7B,$7C,$EF,$4E,$66,$98,$F2,$2C,$EC,$87,$F8,$15,$F0
	.dc.b	$7C,$43,$DA,$F7,$78,$B2,$BF,$C3,$F3,$96,$CE,$5F,$EF,$F3,$CA,$39
	.dc.b	$F7,$7C,$CF,$19,$F4,$54,$CD,$F4,$9C,$3F,$E3,$58,$CB,$FC,$53,$A3
	.dc.b	$FD,$01,$8B,$D6,$E9,$E2,$4A,$F2,$C7,$A8,$C3,$FE,$38,$8C,$3F,$DC
	.dc.b	$FF,$2B,$7C,$47,$E0,$3E,$E1,$EF,$D3,$F2,$94,$AD,$BC,$6B,$4F,$F9
	.dc.b	$8E,$B8,$B9,$21,$E9,$3F,$3A,$E9,$C7,$52,$31,$CE,$D3,$FC,$0F,$87
	.dc.b	$82,$F9,$1F,$E9,$2E,$AF,$CB,$FD,$4B,$E4,$E9,$E8,$F3,$7F,$2B,$87
	.dc.b	$F8,$41,$B7,$FA,$EA,$E0,$FF,$9E,$DA,$F6,$8D,$E6,$FA,$87,$D0,$7E
	.dc.b	$9B,$DC,$F3,$5F,$FA,$0D,$AF,$B2,$76,$FF,$A1,$DA,$FC,$08,$DE,$BF
	.dc.b	$EC,$FC,$BF,$63,$5A,$8E,$7F,$53,$FB,$EB,$FF,$15,$7D,$3B,$CE,$FA
	.dc.b	$AD,$8F,$F1,$74,$ED,$F8,$ED,$AF,$2C,$EE,$7E,$5B,$63,$BF,$8B,$F8
	.dc.b	$0A,$E1,$8F,$F8,$E2,$B4,$DC,$35,$5E,$8F,$8F,$F4,$FC,$A4,$CB,$E0
	.dc.b	$46,$5F,$85,$19,$7E,$34,$7E,$C3,$DD,$FD,$37,$D7,$49,$FD,$7C,$65
	.dc.b	$FF,$4C,$46,$5F,$C6,$68,$EB,$57,$CA,$F9,$7F,$97,$E2,$4B,$FA,$51
	.dc.b	$97,$EA,$46,$5F,$B1,$19,$7E,$D4,$66,$FE,$FF,$1F,$E0,$EB,$EB,$FD
	.dc.b	$57,$D8,$FA,$A9,$87,$DF,$8C,$BF,$81,$8C,$BF,$85,$8F,$BF,$C9,$F8
	.dc.b	$0E,$09,$3F,$87,$8C,$BF,$89,$8C,$BF,$8D,$8F,$DA,$FB,$9F,$C3,$23
	.dc.b	$F0,$B9,$FE,$10,$CF,$83,$23,$24,$3F,$57,$3C,$5B,$1C,$5D,$69,$F9
	.dc.b	$C9,$D3,$FA,$0F,$8B,$FF,$11,$EF,$FD,$65,$B5,$FE,$F6,$6B,$FC,$05
	.dc.b	$7D,$6B,$C2,$FE,$1B,$E3,$FE,$8B,$9D,$6B,$FF,$17,$FC,$47,$5F,$34
	.dc.b	$F0,$AF,$D3,$FC,$74,$BF,$D7,$3B,$9F,$D5,$F5,$7F,$1A,$FA,$6F,$D6
	.dc.b	$5F,$FA,$49,$B7,$FD,$C3,$BB,$F3,$EF,$8A,$33,$D1,$FC,$07,$E8,$BF
	.dc.b	$6A,$73,$F0,$2D,$7E,$F6,$2B,$FC,$5A,$BA,$AF,$A6,$7C,$97,$F2,$C6
	.dc.b	$C7,$FA,$95,$AF,$F2,$7F,$5B,$87,$89,$BA,$FE,$EB,$F0,$3F,$4B,$5F
	.dc.b	$8F,$A2,$31,$87,$CA,$C7,$9F,$3F,$F1,$46,$DF,$0A,$DD,$F1,$E9,$ED
	.dc.b	$87,$8F,$CC,$F3,$11,$FA,$B8,$FE,$8A,$FE,$D4,$6C,$FE,$3D,$F3,$3F
	.dc.b	$C6,$8E,$0F,$F3,$8C,$BF,$9E,$E5,$E8,$E1,$56,$EB,$91,$FC,$FB,$17
	.dc.b	$C6,$7F,$52,$6C,$7F,$B2,$DA,$FF,$44,$F0,$FF,$9C,$E7,$FC,$CA,$72
	.dc.b	$FF,$34,$87,$D1,$7F,$5C,$6D,$73,$BE,$C6,$66,$46,$4F,$F6,$D7,$D2
	.dc.b	$FF,$83,$5F,$F6,$B9,$F9,$87,$0E,$1A,$FF,$B7,$1A,$FF,$B8,$9D,$DF
	.dc.b	$F4,$8F,$A6,$78,$FF,$DC,$1B,$5E,$0F,$5F,$FB,$9D,$F0,$DB,$FD,$D7
	.dc.b	$5A,$AF,$EC,$AD,$C3,$FA,$9D,$6E,$7F,$5B,$F4,$AF,$A8,$FF,$0A,$7C
	.dc.b	$1C,$9F,$F7,$B4,$E5,$B3,$97,$FB,$BF,$B1,$FE,$AC,$C3,$FA,$FD,$6E
	.dc.b	$7F,$BA,$B6,$3F,$93,$CE,$F2,$FC,$BB,$6B,$C7,$7F,$E0,$EB,$FA,$71
	.dc.b	$FE,$07,$F9,$F7,$CC,$FE,$D1,$BD,$FB,$E0,$95,$F9,$4C,$BF,$84,$D6
	.dc.b	$A3,$01,$C3,$E6,$9F,$ED,$FF,$07,$FD,$C3,$EB,$A4,$FF,$77,$8C,$BF
	.dc.b	$DE,$63,$2F,$F7,$D8,$FF,$2B,$7E,$E7,$FE,$08,$8F,$F4,$0B,$5F,$D3
	.dc.b	$B5,$FF,$66,$B7,$63,$F9,$8C,$7F,$77,$F0,$7F,$B4,$3F,$77,$3A,$F6
	.dc.b	$3F,$D0,$3F,$4E,$F4,$3F,$CE,$5E,$64,$CB,$DF,$E2,$2B,$51,$86,$BF
	.dc.b	$E3,$0C,$7F,$72,$DA,$EB,$72,$3E,$4F,$85,$FE,$93,$9E,$D1,$FC,$56
	.dc.b	$F9,$E5,$FE,$22,$C7,$C1,$C3,$FE,$85,$EA,$97,$E1,$8F,$49,$87,$C5
	.dc.b	$18,$7E,$B8,$FA,$6F,$E4,$FE,$5F,$EB,$A3,$1E,$2F,$FA,$01,$FE,$8E
	.dc.b	$7F,$48,$9D,$BF,$B4,$6B,$A3,$FD,$9D,$F1,$9F,$83,$7F,$73,$F1,$48
	.dc.b	$C4,$6E,$31,$CF,$E1,$6C,$75,$F0,$9F,$19,$E5,$C3,$FC,$D2,$D8,$CC
	.dc.b	$EE,$3F,$83,$EA,$F6,$9E,$CF,$F9,$2B,$83,$E5,$46,$39,$FE,$8F,$33
	.dc.b	$C0,$D4,$7F,$66,$C3,$96,$18,$C9,$18,$9F,$C1,$E1,$CE,$7F,$01,$7B
	.dc.b	$AD,$47,$F4,$F9,$FF,$26,$57,$99,$B1,$F2,$67,$D1,$EB,$FF,$97,$A3
	.dc.b	$C5,$3A,$7E,$9C,$EA,$BC,$35,$E5,$E7,$FF,$6E,$71,$F8,$70,$EB,$A7
	.dc.b	$C0,$AC,$1B,$D1,$58,$C3,$F7,$74,$FA,$4F,$F5,$F4,$61,$B9,$FF,$A3
	.dc.b	$3D,$8F,$F2,$24,$E7,$C6,$F8,$D7,$C5,$B7,$C2,$3F,$C7,$B5,$C2,$FF
	.dc.b	$C9,$B3,$9B,$FE,$DF,$D7,$9B,$FE,$3B,$8F,$B6,$31,$3F,$C7,$A3,$1E
	.dc.b	$BF,$C1,$F1,$7D,$73,$FE,$87,$8C,$4F,$D9,$C7,$FD,$67,$E6,$FC,$D7
	.dc.b	$FC,$47,$F0,$5F,$DF,$6F,$F4,$4D,$8F,$EA,$2D,$7F,$3C,$E7,$2F,$F4
	.dc.b	$9B,$67,$F4,$FF,$0E,$F1,$3C,$13,$9B,$EF,$F1,$FE,$7F,$8C,$CB,$70
	.dc.b	$FD,$AD,$23,$CD,$5F,$EC,$A6,$C6,$67,$CF,$18,$39,$BF,$BE,$C3,$F5
	.dc.b	$2F,$48,$F8,$5F,$46,$F8,$0F,$AF,$F4,$8F,$92,$CC,$61,$B7,$FC,$16
	.dc.b	$B8,$36,$7C,$B1,$8C,$BF,$BB,$B7,$5F,$F5,$BB,$6B,$F5,$9F,$61,$FD
	.dc.b	$C9,$FE,$27,$19,$AF,$F2,$AF,$D2,$3E,$57,$F5,$5F,$4B,$FE,$FC,$BF
	.dc.b	$E3,$33,$97,$F9,$9C,$66,$BF,$DB,$33,$97,$FC,$12,$D7,$F8,$B3,$5F
	.dc.b	$E3,$AD,$7C,$35,$F8,$78,$E5,$8F,$FB,$4B,$E9,$5F,$4A,$F9,$FD,$5F
	.dc.b	$E3,$7C,$0F,$F2,$8E,$7F,$ED,$7F,$7B,$1E,$5F,$E6,$5B,$1F,$CB,$E7
	.dc.b	$7A,$7F,$87,$9A,$FF,$2B,$6B,$FC,$C9,$DE,$7F,$EE,$26,$CF,$F9,$5B
	.dc.b	$E0,$DF,$27,$DA,$FF,$DF,$74,$F2,$7F,$83,$57,$F8,$02,$FF,$A1,$CE
	.dc.b	$5F,$EA,$71,$9A,$E3,$5F,$92,$AD,$47,$9E,$30,$4A,$E6,$FF,$21,$F8
	.dc.b	$6F,$C6,$46,$0F,$F7,$D7,$2F,$EE,$CF,$DA,$7C,$C7,$FB,$32,$32,$7F
	.dc.b	$56,$71,$7F,$E4,$26,$BF,$D3,$DA,$FF,$56,$6A,$C3,$FB,$D7,$CA,$76
	.dc.b	$F7,$FF,$85,$36,$FF,$B8,$F8,$6F,$94,$FE,$1C,$7F,$3F,$F8,$97,$D2
	.dc.b	$3E,$97,$37,$E5,$E6,$FD,$AE,$BF,$83,$8F,$ED,$4D,$D7,$D1,$3F,$C3
	.dc.b	$F0,$E7,$BF,$5C,$4D,$F6,$70,$FF,$17,$FA,$47,$D2,$FF,$51,$CD,$F4
	.dc.b	$78,$7F,$BD,$CD,$CD,$CE,$9F,$F5,$F9,$C2,$FA,$1F,$C3,$2B,$85,$7F
	.dc.b	$79,$6B,$FE,$00,$6B,$FE,$C8,$7B,$57,$58,$C4,$F9,$6D,$F5,$DD,$B0
	.dc.b	$F9,$BF,$35,$FC,$E7,$E4,$3F,$9D,$36,$32,$7D,$28,$CB,$4F,$AB,$F4
	.dc.b	$DF,$D0,$7E,$9B,$FA,$13,$63,$27,$DA,$8C,$B4,$FB,$BE,$9F,$F4,$9E
	.dc.b	$DF,$E9,$4D,$8C,$9F,$82,$8C,$B4,$CE,$9F,$31,$FE,$AB,$1A,$8C,$1F
	.dc.b	$8F,$E7,$FD,$8C,$FF,$6A,$F2,$BF,$D5,$AF,$F8,$FB,$D1,$7C,$D7,$F6
	.dc.b	$69,$CD,$FF,$A0,$23,$1F,$19,$2F,$F2,$BF,$11,$6B,$BF,$7F,$F9,$57
	.dc.b	$AB,$FB,$27,$7F,$F6,$56,$DF,$E7,$1B,$5C,$34,$FC,$A7,$37,$D9,$C3
	.dc.b	$FC,$BD,$87,$F6,$5F,$4D,$E2,$7D,$BF,$A6,$C3,$73,$FF,$24,$CE,$9F
	.dc.b	$F2,$43,$72,$FF,$5B,$B5,$FF,$22,$35,$FF,$25,$BB,$C7,$F1,$51,$FE
	.dc.b	$12,$AF,$C5,$7B,$DE,$0B,$FD,$AF,$CB,$7F,$80,$7E,$9B,$FC,$04,$D8
	.dc.b	$C9,$FB,$A8,$CB,$4F,$F2,$FF,$D3,$7D,$4C,$3E,$4F,$D3,$3D,$7F,$F0
	.dc.b	$24,$E7,$3F,$89,$39,$BE,$3C,$FA,$3F,$C5,$D1,$FE,$47,$8C,$9F,$13
	.dc.b	$2F,$F1,$97,$D3,$3E,$9B,$E6,$46,$1B,$3F,$22,$3F,$BE,$EE,$69,$23
	.dc.b	$A1,$FE,$AE,$F9,$B7,$EF,$1F,$E0,$BD,$BF,$C7,$35,$F8,$BF,$77,$FC
	.dc.b	$45,$F4,$97,$0F,$C8,$36,$BF,$2F,$B3,$B3,$FC,$29,$CF,$FE,$9F,$F4
	.dc.b	$B5,$3F,$E6,$79,$FF,$3C,$61,$F9,$9F,$23,$F3,$1F,$17,$FE,$2F,$6C
	.dc.b	$7E,$7E,$3F,$C6,$DE,$96,$32,$EF,$7F,$A0,$AF,$F7,$DE,$1E,$1F,$03
	.dc.b	$FC,$64,$D8,$CC,$EE,$BF,$D1,$93,$97,$9D,$7F,$97,$FE,$0B,$FA,$6C
	.dc.b	$62,$FF,$D5,$71,$8A,$FF,$27,$FA,$7B,$EC,$E0,$CA,$DB,$96,$9F,$B9
	.dc.b	$F3,$7C,$B9,$1A,$48,$F3,$BF,$BB,$3C,$9E,$9D,$8F,$F6,$3F,$D2,$FE
	.dc.b	$0D,$FF,$4A,$B7,$3F,$F4,$03,$5F,$F3,$EB,$5F,$F4,$93,$BC,$BF,$D8
	.dc.b	$4D,$94,$FF,$5F,$CF,$FA,$D4,$78,$3A,$3F,$D2,$55,$BA,$FF,$4C,$B7
	.dc.b	$89,$7C,$27,$F8,$DF,$C3,$AB,$83,$6F,$99,$F4,$7D,$5B,$8F,$BB,$E1
	.dc.b	$BF,$CD,$95,$C5,$B7,$FE,$99,$C3,$D7,$18,$3C,$A7,$DB,$F3,$6D,$FF
	.dc.b	$B5,$A7,$DB,$EA,$FB,$BE,$9D,$F2,$7F,$EC,$FF,$1B,$FA,$94,$62,$BF
	.dc.b	$C0,$81,$C3,$F1,$A3,$73,$FE,$0B,$8F,$F6,$03,$CE,$FE,$5F,$C5,$70
	.dc.b	$9F,$09,$A3,$FD,$C7,$5F,$D8,$63,$50,$C3,$CD,$7C,$E3,$C4,$F7,$C3
	.dc.b	$A5,$F8,$A7,$CF,$18,$C3,$8C,$62,$B9,$4F,$6B,$F2,$C7,$5C,$31,$5E
	.dc.b	$53,$DD,$7A,$79,$1E,$87,$AB,$FE,$27,$3C,$A6,$4A,$66,$FC,$8E,$0F
	.dc.b	$86,$E0,$F0,$96,$BD,$C7,$A0,$F9,$D5,$C8,$F3,$95,$DC,$F8,$67,$9C
	.dc.b	$71,$8F,$B7,$3B,$6C,$7C,$13,$A9,$5C,$9E,$B8,$F8,$BE,$0F,$FB,$F6
	.dc.b	$3F,$0A,$D7,$E1,$35,$FB,$EE,$7E,$37,$FA,$37,$1F,$EB,$A3,$1E,$3F
	.dc.b	$F8,$BA,$BF,$DD,$37,$FB,$25,$7C,$1C,$9F,$B6,$71,$5C,$AF,$EB,$4F
	.dc.b	$0E,$AE,$57,$F9,$8B,$FF,$03,$CE,$9E,$91,$9C,$3C,$4E,$EB,$8E,$3E
	.dc.b	$3A,$E3,$7F,$85,$78,$C6,$6F,$D5,$97,$1A,$FC,$2D,$70,$8E,$D8,$E2
	.dc.b	$BB,$DF,$5C,$3B,$E1,$ED,$6B,$F0,$A7,$6F,$C1,$7B,$CF,$18,$EF,$3F
	.dc.b	$26,$FB,$C7,$E4,$B0,$F9,$51,$E5,$8C,$4E,$EB,$D5,$87,$94,$E7,$30
	.dc.b	$E5,$87,$81,$F1,$47,$CB,$AF,$3E,$1F,$1B,$1F,$09,$E2,$C7,$C0,$4C
	.dc.b	$7B,$E1,$D3,$0E,$D3,$98,$EA,$64,$FA,$18,$76,$3B,$E3,$F1,$23,$E3
	.dc.b	$57,$2B,$F8,$91,$C6,$FF,$8E,$70,$FF,$33,$F9,$0F,$C7,$5F,$F4,$26
	.dc.b	$EB,$FD,$19,$B7,$FD,$0F,$D4,$B7,$C6,$3F,$A8,$CF,$E5,$DF,$E9,$ED
	.dc.b	$79,$1F,$D5,$BE,$9F,$8F,$21,$A7,$6E,$8F,$E9,$3E,$A6,$2B,$FA,$FC
	.dc.b	$D7,$FC,$AF,$D2,$9C,$69,$FD,$6A,$AE,$9F,$D6,$B0,$E7,$87,$23,$C4
	.dc.b	$43,$FB,$3C,$F9,$DF,$F4,$FF,$63,$93,$E0,$E1,$FE,$1D,$6C,$66,$7F
	.dc.b	$C0,$BE,$AE,$27,$95,$6E,$7F,$C2,$31,$8A,$E1,$1F,$EF,$3C,$DE,$AF
	.dc.b	$C1,$6C,$7F,$84,$23,$1E,$66,$59,$19,$3F,$C2,$B1,$E6,$F8,$A7,$7D
	.dc.b	$24,$EC,$FF,$1B,$C6,$9F,$F1,$4B,$5F,$F1,$2B,$5F,$C2,$B5,$FF,$02
	.dc.b	$E4,$8F,$F6,$F4,$63,$6C,$C6,$23,$BF,$77,$DD,$F4,$9F,$13,$E7,$8F
	.dc.b	$C0,$7A,$EF,$9D,$A6,$6F,$A1,$FE,$D2,$C1,$D1,$EF,$24,$7F,$BC,$3E
	.dc.b	$C2,$BF,$E9,$CB,$FF,$17,$3F,$EA,$7F,$17,$3A,$FF,$A1,$BE,$9F,$C7
	.dc.b	$3D,$83,$87,$FD,$66,$63,$4F,$D1,$FA,$2F,$D7,$1F,$4B,$FA,$6A,$D1
	.dc.b	$F8,$BB,$E4,$FE,$95,$AF,$F3,$D6,$BF,$E3,$07,$C1,$87,$FA,$AF,$EA
	.dc.b	$5E,$6F,$E9,$5B,$5E,$7A,$DF,$1F,$F4,$7F,$37,$F9,$AE,$1F,$EA,$26
	.dc.b	$C6,$7C,$5F,$D7,$7D,$4B,$E7,$BD,$15,$AC,$5E,$02,$9F,$37,$C0,$DE
	.dc.b	$1F,$B7,$FA,$27,$D3,$7D,$DC,$E6,$3F,$63,$15,$6A,$30,$D7,$FD,$95
	.dc.b	$3C,$1F,$F6,$9C,$67,$4F,$F5,$DF,$D4,$BC,$7F,$9C,$DA,$FB,$0D,$7E
	.dc.b	$BF,$D4,$BE,$33,$75,$FB,$D9,$56,$DF,$E3,$0D,$7F,$B8,$35,$FF,$31
	.dc.b	$BD,$6F,$FD,$9B,$6F,$4F,$F0,$0D,$AF,$1D,$FC,$5F,$94,$F8,$DF,$35
	.dc.b	$27,$E0,$C6,$5F,$87,$19,$7E,$0C,$7E,$EF,$E1,$24,$7E,$8B,$1F,$D8
	.dc.b	$47,$D1,$F5,$F3,$3F,$9F,$F9,$EF,$D5,$98,$AF,$F7,$A6,$CF,$9C,$FB
	.dc.b	$1E,$54,$BF,$A7,$1F,$6C,$C3,$F4,$A3,$0E,$8F,$B2,$EA,$BA,$CF,$E0
	.dc.b	$FE,$A7,$F0,$9D,$B2,$FE,$FC,$65,$FC,$0C,$65,$FB,$F1,$F8,$BF,$23
	.dc.b	$F1,$9C,$32,$7F,$0F,$19,$7F,$13,$19,$7F,$0F,$1F,$B1,$F8,$1F,$E1
	.dc.b	$51,$FB,$B6,$BF,$BA,$6E,$3F,$6B,$0E,$1E,$47,$E1,$67,$75,$FD,$6D
	.dc.b	$AF,$F8,$CD,$AF,$FA,$99,$E7,$CD,$FE,$37,$E4,$FF,$50,$CF,$F0,$B6
	.dc.b	$BE,$1A,$FE,$D7,$F2,$1F,$D3,$7C,$9B,$E7,$FF,$2B,$AF,$E3,$F3,$96
	.dc.b	$F8,$BF,$C6,$9F,$E2,$7E,$FB,$2F,$CE,$36,$3F,$AD,$35,$FE,$E4,$D7
	.dc.b	$FC,$99,$D9,$FB,$4E,$CF,$DD,$12,$3C,$38,$BD,$DF,$DF,$76,$7F,$1D
	.dc.b	$FA,$8F,$DA,$65,$FE,$55,$6D,$FF,$08,$E8,$FD,$34,$E7,$BF,$F6,$15
	.dc.b	$FB,$FA,$FF,$03,$35,$D1,$FE,$07,$CF,$F8,$3F,$D4,$FF,$0D,$24,$7F
	.dc.b	$35,$9E,$EB,$EA,$5F,$07,$FC,$C5,$B1,$FC,$ED,$AF,$FB,$49,$78,$DB
	.dc.b	$9F,$E0,$91,$A9,$FF,$0D,$35,$FF,$39,$35,$FF,$72,$BD,$3A,$3E,$86
	.dc.b	$2F,$85,$FC,$1D,$7F,$2A,$6B,$E7,$BF,$37,$97,$23,$CC,$FF,$88,$67
	.dc.b	$EF,$5F,$D9,$C3,$F9,$FD,$7F,$8A,$DA,$F9,$AB,$D2,$47,$F9,$6F,$D4
	.dc.b	$FF,$35,$3B,$3F,$D6,$A1,$F1,$FF,$D5,$DB,$1F,$D8,$5A,$FF,$60,$F1
	.dc.b	$3F,$A0,$B9,$9F,$E5,$53,$A8,$FF,$20,$35,$FF,$4C,$35,$1F,$D6,$A3
	.dc.b	$FB,$34,$BE,$9A,$9E,$38,$FE,$D5,$F5,$1F,$E2,$2A,$FE,$E2,$DC,$BF
	.dc.b	$38,$DE,$2F,$F5,$37,$D4,$FF,$8B,$27,$3E,$D7,$F3,$FA,$E5,$7F,$E5
	.dc.b	$56,$BF,$EA,$D6,$BE,$67,$A3,$E8,$A7,$8F,$FD,$F5,$B1,$FE,$07,$9D
	.dc.b	$AF,$A9,$FE,$D8,$FF,$82,$DA,$FF,$82,$9B,$87,$ED,$FE,$9D,$EC,$3E
	.dc.b	$4B,$FC,$15,$5C,$2F,$FC,$D8,$D7,$FD,$88,$D7,$E6,$B5,$F1,$7A,$7F
	.dc.b	$E2,$CF,$87,$91,$E3,$8E,$BC,$12,$3C,$CF,$F9,$FB,$17,$D6,$FD,$59
	.dc.b	$C9,$FE,$65,$C3,$0F,$F9,$F5,$AF,$9A,$BF,$CD,$5D,$1F,$E0,$0F,$87
	.dc.b	$FF,$03,$7D,$2C,$AF,$F0,$0C,$7F,$82,$4C,$3F,$E0,$08,$C3,$FD,$7F
	.dc.b	$EA,$BF,$22,$B7,$FE,$3C,$23,$2E,$1B,$C9,$FE,$81,$F4,$5F,$2F,$FE
	.dc.b	$7A,$9C,$DF,$36,$DF,$3F,$63,$FC,$E5,$5F,$E9,$EF,$AA,$7A,$34,$98
	.dc.b	$7F,$87,$71,$75,$E2,$3C,$D9,$FE,$83,$E2,$98,$72,$9F,$F2,$DC,$63
	.dc.b	$0F,$F4,$05,$F1,$9F,$F8,$01,$39,$7F,$D5,$4D,$7F,$D5,$CD,$77,$5F
	.dc.b	$CC,$14,$78,$3C,$7F,$51,$FE,$8D,$F9,$8B,$3E,$2A,$E1,$5B,$8F,$64
	.dc.b	$FF,$96,$FE,$AB,$F9,$6F,$97,$FE,$A5,$6C,$FF,$AF,$5A,$FF,$B0,$5A
	.dc.b	$FF,$00,$1F,$1E,$F9,$37,$4F,$FA,$BE,$30,$D8,$CC,$FF,$B2,$67,$2F
	.dc.b	$7F,$0D,$F2,$5E,$36,$C6,$63,$07,$C8,$F7,$7F,$C4,$1C,$3F,$41,$B3
	.dc.b	$F5,$8F,$F2,$FF,$D5,$3B,$FF,$1B,$5F,$66,$B7,$CF,$FD,$B7,$81,$18
	.dc.b	$9D,$7B,$9C,$31,$E7,$85,$BD,$CD,$BF,$F2,$45,$F8,$FD,$BE,$1D,$AC
	.dc.b	$FF,$DA,$13,$92,$65,$E7,$C7,$C3,$7C,$2B,$73,$9C,$3A,$B6,$7C,$06
	.dc.b	$F8,$AE,$1F,$E4,$BF,$AA,$FF,$68,$B7,$C8,$F1,$F5,$7F,$AF,$63,$04
	.dc.b	$C3,$E4,$6E,$BE,$31,$FE,$F4,$F6,$3E,$14,$E7,$0F,$6F,$5F,$3A,$FF
	.dc.b	$5A,$65,$8B,$F8,$3F,$11,$EA,$C3,$FB,$5D,$ED,$FE,$DB,$19,$CB,$FD
	.dc.b	$F9,$39,$7F,$DF,$D1,$9C,$3F,$A7,$FC,$3F,$F7,$39,$FF,$5E,$36,$3F
	.dc.b	$D7,$7F,$1A,$F0,$BE,$C4,$7F,$B5,$3D,$0B,$5B,$CB,$FC,$54,$DC,$3F
	.dc.b	$39,$87,$17,$AD,$70,$FA,$A9,$1F,$EC,$B9,$FC,$BB,$E2,$8C,$E1,$FE
	.dc.b	$CC,$B4,$78,$AB,$73,$98,$DE,$1E,$27,$5D,$18,$C3,$F5,$DF,$50,$F8
	.dc.b	$4F,$F1,$F6,$1F,$DC,$23,$13,$FB,$38,$DC,$E9,$4E,$9D,$18,$8F,$8B
	.dc.b	$19,$C3,$F4,$5F,$53,$FD,$92,$FF,$DD,$13,$97,$A9,$8A,$FD,$14,$EE
	.dc.b	$FF,$DD,$75,$B7,$FC,$0B,$F5,$0F,$A6,$FD,$47,$D4,$BE,$0F,$EA,$CE
	.dc.b	$72,$FE,$53,$39,$7F,$DF,$11,$9B,$FF,$7B,$B7,$3F,$F3,$27,$D4,$BE
	.dc.b	$A5,$E3,$F4,$CE,$63,$C9,$FD,$D5,$6E,$73,$8F,$F0,$73,$FD,$CB,$ED
	.dc.b	$FE,$73,$DC,$FC,$4F,$85,$FE,$52,$C7,$FD,$C1,$D3,$FE,$4E,$7F,$C4
	.dc.b	$3A,$7F,$9C,$E7,$58,$73,$6C,$7F,$92,$BE,$09,$F2,$BD,$32,$98,$72
	.dc.b	$3F,$CF,$2F,$F0,$89,$DC,$67,$0F,$F1,$7E,$1D,$23,$FC,$FF,$E3,$EF
	.dc.b	$CF,$F3,$B6,$38,$57,$FC,$00,$8C,$C7,$F5,$6F,$84,$7C,$EE,$2B,$D1
	.dc.b	$57,$4F,$E4,$95,$AB,$E7,$8A,$BF,$80,$E1,$27,$BF,$CE,$FF,$68,$7E
	.dc.b	$3C,$7E,$CA,$31,$8F,$CF,$6E,$3F,$C1,$3E,$5F,$5E,$9F,$F0,$C6,$E1
	.dc.b	$F4,$1B,$7F,$CF,$72,$FE,$F3,$F2,$FF,$C4,$8F,$E6,$DF,$14,$FA,$87
	.dc.b	$D4,$E7,$1F,$2F,$37,$8C,$DC,$B8,$31,$3F,$EB,$68,$C3,$63,$7F,$31
	.dc.b	$C1,$FE,$1B,$F5,$2F,$A9,$A7,$F4,$7A,$7D,$4B,$EA,$73,$1F,$52,$B8
	.dc.b	$36,$B8,$56,$E3,$04,$DF,$F2,$D5,$C1,$FE,$BD,$1D,$AF,$AC,$7C,$5E
	.dc.b	$A7,$CE,$FF,$36,$F9,$1F,$E6,$ED,$BF,$CD,$B6,$B8,$37,$97,$F9,$BF
	.dc.b	$CF,$7F,$3E,$FA,$0F,$E7,$ED,$BF,$D1,$B6,$FC,$0D,$AF,$E7,$FE,$97
	.dc.b	$F4,$6E,$FF,$E8,$ED,$BF,$D5,$B6,$FD,$0D,$AE,$33,$FB,$0A,$E2,$D3
	.dc.b	$F4,$90,$F3,$3F,$D4,$31,$86,$DF,$C3,$EE,$FE,$B3,$18,$AF,$F2,$7B
	.dc.b	$6B,$F7,$6D,$A7,$4F,$F9,$1A,$F8,$FC,$6F,$F6,$0E,$7F,$EC,$3D,$97
	.dc.b	$1F,$E0,$AD,$C3,$E9,$37,$2F,$EC,$2F,$F5,$4F,$4F,$F8,$54,$FF,$11
	.dc.b	$6C,$FF,$BF,$EB,$F1,$3E,$6E,$FC,$49,$1F,$BC,$AD,$BF,$EB,$23,$66
	.dc.b	$63,$77,$D5,$FC,$6D,$62,$FF,$B9,$FC,$B7,$F7,$AF,$A9,$FE,$F6,$DC
	.dc.b	$3F,$8F,$B7,$0F,$C6,$36,$93,$FE,$43,$FA,$9F,$E2,$F5,$FC,$AD,$B3
	.dc.b	$FE,$EA,$FA,$9F,$F1,$95,$7F,$14,$9C,$BC,$78,$BF,$BB,$C6,$2F,$E8
	.dc.b	$D7,$07,$A4,$FC,$9C,$25,$FF,$25,$F9,$87,$D4,$7F,$7B,$BF,$F6,$AE
	.dc.b	$CB,$DC,$7C,$F7,$FC,$79,$5C,$3D,$6C,$E9,$EC,$AE,$35,$8C,$7F,$DB
	.dc.b	$F7,$B7,$E9,$47,$E3,$2E,$47,$F4,$6F,$2E,$DF,$F8,$B3,$9B,$C9,$8F
	.dc.b	$E5,$EF,$FC,$53,$5A,$8F,$C9,$FC,$1F,$E5,$8F,$F1,$6F,$97,$71,$FC
	.dc.b	$E4,$5C,$7F,$7D,$36,$BF,$37,$F3,$FF,$9E,$3F,$C7,$FE,$65,$C3,$F4
	.dc.b	$91,$6F,$F9,$57,$BB,$FC,$F5,$F3,$47,$09,$FF,$51,$57,$F9,$23,$9B
	.dc.b	$15,$C2,$2E,$1F,$1F,$AD,$8F,$E5,$2B,$F5,$2F,$59,$FD,$7F,$8F,$FB
	.dc.b	$23,$FC,$A7,$E6,$5C,$BF,$71,$16,$FC,$B3,$6B,$FC,$ED,$F4,$AF,$9F
	.dc.b	$FF,$39,$37,$0F,$F5,$C7,$9D,$89,$91,$FE,$B4,$9D,$61,$F0,$2B,$2F
	.dc.b	$F9,$C6,$33,$7D,$AF,$FC,$CF,$5A,$8C,$5F,$39,$CB,$4E,$DF,$01,$E4
	.dc.b	$3F,$CD,$FE,$35,$8C,$93,$1F,$A3,$19,$7F,$C2,$40,$7A,$17,$E9,$9F
	.dc.b	$1D,$FE,$3F,$F0,$7F,$2D,$18,$AF,$F2,$15,$6E,$B5,$85,$9C,$8A,$E0
	.dc.b	$F7,$BE,$75,$FE,$46,$BE,$13,$9B,$AF,$96,$2D,$70,$8E,$B7,$D1,$FF
	.dc.b	$2F,$B5,$F5,$9E,$1B,$F2,$B4,$E2,$E2,$38,$57,$6B,$F0,$D7,$29,$ED
	.dc.b	$7E,$18,$FF,$6A,$56,$DE,$86,$0A,$79,$2F,$DC,$E2,$79,$3E,$D3,$17
	.dc.b	$E0,$AE,$97,$E4,$9C,$9D,$0F,$25,$FA,$4E,$B1,$E6,$AE,$53,$C2,$39
	.dc.b	$19,$79,$57,$6A,$E2,$73,$8F,$8A,$7F,$BC,$6F,$CE,$F8,$0C,$5F,$6E
	.dc.b	$CF,$F7,$04,$7E,$01,$AF,$C3,$6B,$EF,$73,$F1,$0F,$A7,$6D,$FC,$AB
	.dc.b	$E2,$7F,$E0,$04,$E6,$31,$7F,$97,$8C,$2B,$E7,$7C,$A3,$17,$CA,$31
	.dc.b	$A7,$E4,$AD,$D1,$FA,$1B,$FB,$31,$87,$F2,$6F,$69,$E1,$5C,$B4,$F2
	.dc.b	$DF,$19,$E8,$D8,$F0,$65,$CA,$B8,$61,$E9,$AD,$34,$D4,$C7,$C7,$1A
	.dc.b	$58,$79,$5F,$3D,$74,$72,$D8,$D6,$1E,$49,$E8,$FA,$AF,$DB,$7D,$6B
	.dc.b	$78,$7A,$2F,$AD,$69,$86,$09,$7D,$5F,$39,$DA,$79,$46,$48,$E0,$91
	.dc.b	$F0,$4E,$27,$47,$2F,$C0,$BF,$31,$E7,$C3,$DC,$7B,$DC,$C7,$84,$7C
	.dc.b	$43,$8D,$F1,$BF,$7F,$2F,$E7,$AD,$DB,$F9,$15,$6E,$31,$7F,$D3,$3C
	.dc.b	$EB,$39,$8F,$E7,$D3,$8C,$3A,$C7,$FA,$A7,$A7,$FA,$AD,$E7,$E9,$E4
	.dc.b	$FF,$45,$7A,$CF,$F5,$AF,$3F,$33,$FD,$16,$6E,$7F,$D7,$FA,$D5,$8D
	.dc.b	$3C,$F9,$2F,$18,$7F,$62,$AB,$A7,$C1,$9C,$E5,$DE,$BF,$CB,$1C,$7E
	.dc.b	$EC,$FF,$CB,$8D,$F0,$BF,$59,$3E,$BD,$1E,$7E,$6A,$4F,$23,$51,$EC
	.dc.b	$97,$9F,$49,$3C,$FC,$8A,$FF,$81,$22,$E3,$EF,$9F,$27,$C3,$BB,$E9
	.dc.b	$2F,$C9,$75,$FE,$13,$C3,$FE,$E6,$AE,$11,$FE,$78,$6B,$DC,$C7,$7F
	.dc.b	$62,$3F,$EF,$0F,$85,$AF,$43,$F8,$AC,$A7,$39,$7F,$84,$FD,$25,$FA
	.dc.b	$39,$66,$1E,$17,$EF,$CB,$E5,$B3,$3D,$6D,$8F,$FA,$AA,$38,$1F,$8B
	.dc.b	$AE,$0F,$FB,$BB,$E9,$7B,$B7,$1F,$CB,$A8,$CB,$E9,$F3,$1F,$4E,$FA
	.dc.b	$3D,$D6,$B2,$FF,$29,$5F,$1C,$32,$7E,$8B,$0E,$37,$89,$F0,$10,$ED
	.dc.b	$87,$57,$FC,$19,$7D,$3C,$9F,$E3,$0D,$ED,$F3,$C6,$38,$3F,$C2,$5D
	.dc.b	$0F,$A6,$7C,$CE,$AE,$63,$D2,$FF,$9C,$61,$C6,$3F,$D5,$37,$C1,$FD
	.dc.b	$97,$D1,$BE,$8F,$E8,$DC,$9C,$C7,$EC,$A7,$51,$F3,$E7,$4F,$FC,$01
	.dc.b	$8C,$BF,$83,$1F,$4D,$FC,$01,$C3,$74,$F9,$6D,$AF,$E9,$DF,$4A,$F8
	.dc.b	$5F,$AF,$39,$BF,$F1,$93,$5E,$BE,$27,$DC,$AE,$11,$F0,$4D,$C7,$22
	.dc.b	$39,$11,$C1,$F5,$7C,$77,$BF,$E4,$E4,$7C,$68,$F8,$46,$1F,$81,$18
	.dc.b	$7F,$55,$EE,$7E,$D0,$8F,$CE,$F4,$6B,$F9,$BE,$A4,$E3,$5F,$F3,$C6
	.dc.b	$DF,$E6,$BA,$FA,$BF,$2B,$F4,$7C,$29,$18,$21,$F4,$5C,$BF,$4E,$3E
	.dc.b	$C1,$8A,$EB,$38,$9F,$C0,$FD,$37,$DD,$E9,$91,$82,$1F,$75,$CB,$F7
	.dc.b	$E3,$F0,$9D,$BF,$85,$F4,$E4,$FE,$32,$32,$FE,$3A,$32,$FE,$1E,$3F
	.dc.b	$67,$ED,$3E,$3E,$92,$7E,$B7,$17,$E9,$DF,$C6,$4E,$DF,$B4,$D7,$33
	.dc.b	$F8,$3C,$79,$46,$0D,$C6,$5F,$C3,$7C,$57,$F8,$97,$83,$FA,$DB,$63
	.dc.b	$3E,$4F,$F1,$07,$7E,$03,$D7,$FE,$17,$F1,$1F,$C1,$38,$7F,$82,$B7
	.dc.b	$AF,$FC,$50,$DB,$63,$D5,$FD,$47,$5F,$EB,$48,$F8,$3D,$8F,$D8,$75
	.dc.b	$FF,$18,$FA,$5F,$D6,$E5,$FC,$49,$B9,$7F,$04,$E8,$FD,$24,$63,$0F
	.dc.b	$F4,$C3,$53,$9F,$0D,$3F,$AD,$6F,$D3,$7E,$FC,$95,$FA,$7A,$E5,$F4
	.dc.b	$CF,$7F,$F9,$5B,$67,$F5,$8D,$7F,$A2,$7A,$68,$78,$BC,$2A,$57,$07
	.dc.b	$FC,$13,$3A,$8C,$1F,$E0,$7F,$80,$FB,$DA,$FF,$B7,$5B,$AF,$F9,$56
	.dc.b	$BC,$13,$FC,$D9,$7B,$FE,$62,$5F,$3E,$4A,$FF,$17,$9D,$BE,$A9,$FF
	.dc.b	$08,$E1,$F7,$E3,$07,$F2,$8F,$A6,$FE,$48,$72,$C3,$FA,$2D,$3E,$1F
	.dc.b	$FA,$7B,$6B,$F8,$83,$5F,$E4,$DE,$1B,$E7,$B7,$52,$76,$FF,$58,$9D
	.dc.b	$46,$0F,$E3,$4B,$E8,$34,$71,$8F,$E6,$5F,$49,$FE,$0E,$7F,$90,$35
	.dc.b	$CF,$AB,$FD,$93,$E9,$7F,$C3,$71,$83,$71,$FD,$A2,$78,$67,$FD,$8A
	.dc.b	$B6,$FF,$64,$8E,$95,$FE,$87,$0E,$1F,$F4,$53,$5F,$F7,$63,$57,$D1
	.dc.b	$FF,$55,$7F,$A5,$36,$7F,$B8,$B7,$87,$FD,$E9,$F4,$1F,$E0,$7F,$A5
	.dc.b	$FE,$FB,$39,$BF,$F7,$C3,$5E,$BE,$9F,$F9,$62,$B8,$47,$F4,$39,$D1
	.dc.b	$27,$5E,$BA,$FA,$EB,$2F,$83,$CF,$E9,$7E,$94,$60,$87,$AA,$34,$7F
	.dc.b	$6B,$F8,$3F,$ED,$9F,$3F,$27,$FB,$9C,$65,$FE,$EB,$19,$7F,$B7,$47
	.dc.b	$F9,$4F,$E9,$7F,$07,$1F,$E2,$56,$EC,$DB,$FC,$1B,$18,$7F,$C1,$3E
	.dc.b	$83,$E4,$B9,$73,$4C,$3F,$B4,$60,$C3,$84,$3C,$16,$1C,$63,$FB,$77
	.dc.b	$7B,$E9,$5E,$87,$FA,$09,$B3,$A6,$BF,$E4,$7B,$E7,$D3,$BA,$93,$B7
	.dc.b	$FC,$CD,$18,$8F,$E1,$75,$CB,$2F,$74,$7F,$A6,$A3,$1A,$FB,$AB,$C5
	.dc.b	$E2,$CA,$F1,$3F,$E7,$1F,$93,$C4,$74,$8B,$87,$FA,$2E,$7F,$C8,$FF
	.dc.b	$4B,$FC,$7F,$CD,$FC,$E5,$84,$65,$CB,$5F,$D1,$35,$FF,$7B,$FC,$83
	.dc.b	$AF,$9C,$7F,$B6,$23,$0F,$6F,$8C,$46,$23,$F5,$91,$89,$E3,$5F,$C7
	.dc.b	$BE,$3A,$F8,$46,$92,$FF,$69,$C4,$FD,$83,$FD,$49,$C1,$FE,$99,$F9
	.dc.b	$5F,$97,$BF,$CC,$AD,$7C,$5E,$2B,$F1,$35,$70,$FF,$64,$E1,$C9,$EA
	.dc.b	$D7,$E3,$1A,$AF,$F3,$4B,$FE,$CA,$8D,$3F,$EC,$CF,$CA,$F9,$63,$FE
	.dc.b	$BB,$F6,$3B,$64,$9C,$E5,$F9,$A9,$CC,$62,$FF,$2D,$D0,$9C,$E5,$E7
	.dc.b	$E8,$FC,$CC,$63,$0F,$B7,$8F,$5A,$F2,$E1,$FE,$E4,$C7,$CA,$D8,$CE
	.dc.b	$4F,$DD,$3F,$55,$87,$EA,$1B,$5F,$9D,$FB,$8F,$F4,$54,$E7,$B7,$E4
	.dc.b	$5F,$16,$DF,$08,$FC,$DE,$8B,$E3,$87,$FA,$86,$B5,$87,$F7,$FF,$99
	.dc.b	$FA,$52,$FD,$35,$F0,$FE,$FF,$55,$5A,$9C,$DF,$FA,$02,$31,$5B,$6C
	.dc.b	$7F,$99,$7E,$E1,$F1,$FF,$7F,$9B,$C9,$E9,$F8,$27,$35,$FE,$6F,$9C
	.dc.b	$BC,$6B,$86,$BF,$E2,$78,$C1,$B7,$1A,$7F,$A1,$FE,$99,$F0,$FF,$1B
	.dc.b	$EC,$1F,$35,$EE,$9C,$E1,$FC,$5A,$72,$FF,$9E,$E3,$37,$FE,$C7,$9C
	.dc.b	$BD,$6F,$A5,$F3,$7E,$FC,$2B,$8B,$FD,$65,$AF,$E6,$9A,$FF,$01,$F9
	.dc.b	$9F,$C7,$61,$FE,$8F,$F9,$E7,$E8,$4E,$EF,$AB,$5A,$BF,$F7,$35,$6A
	.dc.b	$73,$D9,$FA,$B8,$C6,$1F,$E9,$5B,$E1,$D1,$FC,$7A,$73,$87,$FA,$B6
	.dc.b	$73,$5F,$AE,$6C,$7F,$37,$E6,$FE,$B1,$3C,$F0,$F1,$FB,$7F,$A3,$FD
	.dc.b	$89,$F2,$98,$AF,$F5,$47,$5A,$73,$1E,$3A,$E1,$5B,$C7,$F9,$D5,$69
	.dc.b	$F2,$4E,$74,$FC,$DC,$BE,$67,$18,$7F,$04,$F8,$4F,$CE,$D7,$F3,$A8
	.dc.b	$7D,$0B,$E0,$FF,$3B,$5A,$9F,$E6,$06,$A3,$3F,$15,$FC,$4E,$30,$7F
	.dc.b	$AF,$67,$86,$7F,$D8,$EF,$A4,$E6,$B8,$BF,$EC,$73,$C3,$F0,$18,$AF
	.dc.b	$F6,$D4,$7F,$32,$35,$0F,$B8,$FD,$8C,$66,$59,$7A,$EB,$53,$FE,$CF
	.dc.b	$35,$19,$AE,$3E,$6F,$A6,$B7,$7E,$08,$CD,$F9,$71,$7A,$F8,$AF,$F7
	.dc.b	$87,$77,$EF,$CF,$E0,$7F,$4C,$F8,$1C,$61,$FE,$DF,$8C,$3F,$F0,$02
	.dc.b	$77,$3A,$AF,$F7,$D1,$89,$FC,$1F,$37,$D0,$8F,$F7,$E5,$F2,$7F,$BD
	.dc.b	$7E,$D7,$E8,$DB,$F8,$CB,$6F,$FB,$37,$37,$E2,$F1,$FE,$65,$39,$BF
	.dc.b	$ED,$58,$7F,$1C,$7F,$C1,$DC,$9D,$71,$FE,$FB,$09,$D6,$1C,$F1,$FC
	.dc.b	$BD,$E0,$CE,$0F,$7F,$F3,$DE,$4F,$F2,$6F,$33,$F8,$2E,$7F,$E0,$8D
	.dc.b	$9F,$79,$ED,$8E,$F3,$AC,$BF,$BF,$56,$B5,$E5,$E0,$FF,$23,$C3,$85
	.dc.b	$7F,$5E,$BE,$15,$FD,$C6,$72,$FF,$81,$E3,$37,$CE,$BF,$C1,$58,$78
	.dc.b	$30,$E9,$7F,$1B,$DB,$F8,$FF,$09,$6F,$F6,$BA,$FF,$A3,$5B,$3A,$69
	.dc.b	$A9,$F9,$FF,$6D,$F4,$3E,$DE,$CF,$26,$E1,$F6,$9B,$97,$DA,$7E,$BF
	.dc.b	$D5,$FD,$8F,$AB,$B7,$E2,$6D,$FE,$09,$B7,$F8,$26,$CF,$E0,$B3,$46
	.dc.b	$2B,$FC,$99,$18,$9F,$F2,$8F,$5B,$E0,$7F,$AD,$7A,$7F,$DA,$23,$13
	.dc.b	$F9,$48,$C4,$EA,$38,$B6,$FF,$2B,$D9,$F9,$0F,$B0,$FC,$8B,$6B,$E5
	.dc.b	$B6,$77,$1F,$3A,$7F,$11,$DF,$FD,$87,$EE,$FF,$19,$3F,$60,$FC,$04
	.dc.b	$67,$A3,$71,$FD,$8E,$77,$7F,$C1,$23,$66,$67,$71,$93,$FC,$71,$3C
	.dc.b	$36,$FC,$9D,$F5,$BF,$CF,$FC,$3F,$F3,$9F,$9F,$FE,$74,$D8,$C9,$FA
	.dc.b	$48,$CB,$4F,$F3,$EF,$E6,$CF,$95,$FF,$12,$7A,$6F,$9C,$FF,$0A,$CE
	.dc.b	$6B,$FA,$1C,$E5,$E3,$5C,$2B,$6D,$8F,$89,$A7,$E7,$FE,$13,$FC,$3F
	.dc.b	$FB,$97,$F3,$28,$77,$27,$3A,$FE,$D5,$B7,$CA,$B7,$18,$7E,$31,$88
	.dc.b	$FF,$4B,$DF,$5B,$E9,$18,$BF,$F1,$2F,$3F,$AE,$FF,$D2,$1F,$71,$FD
	.dc.b	$99,$FC,$9E,$5F,$E4,$58,$C4,$EF,$83,$FC,$67,$18,$9D,$79,$FF,$1B
	.dc.b	$E9,$BE,$4F,$CE,$5C,$FE,$3B,$72,$FF,$4F,$70,$23,$19,$7F,$8F,$A3
	.dc.b	$15,$FA,$0F,$8F,$FF,$2B,$7C,$13,$D2,$FD,$04,$61,$B3,$FD,$DA,$74
	.dc.b	$F1,$31,$1F,$E5,$FF,$6F,$AE,$5F,$E5,$3C,$B1,$F3,$B9,$65,$FF,$80
	.dc.b	$2B,$79,$FF,$94,$95,$6A,$FF,$2D,$2F,$DD,$BF,$2D,$7F,$EC,$CF,$86
	.dc.b	$FF,$2E,$C6,$2B,$FC,$E7,$18,$EC,$F6,$B6,$BD,$4F,$F9,$FB,$F5,$37
	.dc.b	$CE,$7C,$2E,$CC,$36,$BF,$D9,$F3,$A7,$E1,$4E,$9A,$6A,$71,$87,$F9
	.dc.b	$CF,$E7,$2E,$7F,$E9,$DC,$39,$4F,$EC,$7A,$51,$89,$F8,$55,$C3,$0F
	.dc.b	$E7,$53,$FE,$58,$9F,$F3,$BF,$0B,$BB,$D5,$F0,$7F,$69,$B7,$F5,$DB
	.dc.b	$97,$FB,$86,$B8,$4F,$E7,$E3,$FD,$99,$F5,$BF,$EC,$DF,$AD,$FE,$45
	.dc.b	$1F,$EA,$AF,$3F,$FD,$BF,$18,$8F,$F5,$6C,$EF,$C6,$FF,$57,$C7,$FB
	.dc.b	$06,$F0,$E2,$6B,$FD,$72,$2B,$E0,$15,$D5,$E4,$62,$BF,$DC,$35,$C2
	.dc.b	$FC,$71,$89,$D1,$E7,$BE,$AA,$B8,$46,$1B,$1C,$0F,$59,$4F,$0D,$F8
	.dc.b	$DE,$46,$6F,$C2,$FB,$47,$BD,$E9,$85,$FF,$9D,$F0,$F3,$4E,$63,$15
	.dc.b	$C6,$F0,$F8,$4E,$EE,$0F,$16,$1E,$86,$C7,$8E,$7D,$8D,$39,$4E,$EF
	.dc.b	$E3,$57,$36,$9F,$63,$0F,$5D,$7C,$29,$D3,$E5,$9D,$5F,$69,$CC,$EE
	.dc.b	$72,$F8,$07,$C7,$FC,$47,$E3,$35,$F9,$ED,$7F,$04,$E7,$E3,$DF,$5B
	.dc.b	$FE,$5C,$9E,$79,$FF,$BF,$27,$3B,$A7,$25,$8C,$B8,$0E,$2F,$B5,$F5
	.dc.b	$C7,$F3,$93,$FE,$DA,$AF,$CB,$B5,$FC,$33,$D2,$37,$3C,$71,$F2,$7C
	.dc.b	$57,$8E,$F6,$DA,$DE,$3D,$2F,$85,$79,$EB,$57,$CA,$35,$51,$E6,$79
	.dc.b	$27,$6F,$BE,$B9,$BE,$E6,$E1,$E9,$C3,$DB,$5C,$2B,$57,$D6,$73,$1D
	.dc.b	$4C,$61,$F3,$F0,$FA,$59,$6A,$7F,$3F,$3E,$76,$9E,$3B,$FC,$EB,$D8
	.dc.b	$F4,$92,$FE,$1B,$F3,$2F,$CF,$87,$97,$1F,$46,$1E,$1C,$3C,$77,$F1
	.dc.b	$CF,$31,$C5,$CB,$F7,$AB,$91,$CE,$FE,$19,$DA,$38,$E1,$E2,$3F,$8F
	.dc.b	$F1,$FF,$32,$FB,$AF,$B5,$87,$F2,$ED,$69,$FE,$8C,$E0,$FF,$42,$C7
	.dc.b	$F3,$B9,$CE,$1F,$E0,$EE,$34,$63,$D9,$5F,$0C,$39,$1C,$31,$F0,$61
	.dc.b	$FD,$3B,$83,$18,$FE,$AA,$30,$7F,$67,$F7,$F3,$32,$32,$7F,$8E,$F1
	.dc.b	$76,$AB,$53,$98,$7D,$3F,$95,$B1,$CB,$5F,$EA,$D8,$66,$A4,$E7,$3F
	.dc.b	$F4,$4F,$6B,$D6,$E5,$E6,$4A,$FF,$19,$E3,$FD,$DE,$7E,$2B,$63,$FC
	.dc.b	$01,$3B,$7F,$BE,$D7,$37,$FC,$09,$E8,$F9,$48,$FF,$81,$F9,$6B,$EB
	.dc.b	$8B,$7F,$E0,$99,$EF,$5F,$E8,$0A,$FF,$07,$7B,$1E,$F3,$3E,$5C,$8C
	.dc.b	$90,$F1,$BE,$8F,$DD,$5D,$FF,$6F,$C8,$FE,$C5,$97,$F8,$AF,$2D,$4C
	.dc.b	$C3,$FC,$B3,$1B,$30,$FF,$92,$BE,$BF,$EA,$3F,$E5,$8E,$47,$E5,$8D
	.dc.b	$BF,$10,$D9,$FF,$2A,$57,$07,$97,$2E,$AA,$46,$A3,$F2,$6B,$EB,$5F
	.dc.b	$47,$8F,$A8,$E9,$3F,$E5,$C6,$EB,$FE,$E0,$AD,$CF,$F9,$8D,$AE,$F0
	.dc.b	$E4,$76,$8C,$CF,$1B,$CD,$7F,$9A,$EF,$AF,$2F,$F2,$5B,$DB,$FE,$D2
	.dc.b	$E0,$7C,$97,$E9,$FF,$71,$7B,$74,$FA,$51,$31,$F9,$13,$A6,$CE,$5F
	.dc.b	$D1,$43,$E0,$3A,$74,$C9,$E3,$1D,$CC,$3C,$23,$0F,$C1,$F8,$CF,$91
	.dc.b	$F1,$72,$30,$43,$E5,$39,$7F,$D0,$FA,$27,$E3,$19,$DF,$FC,$0B,$3F
	.dc.b	$D6,$76,$5E,$A3,$FD,$DF,$E2,$FF,$09,$F7,$FF,$B1,$C7,$FB,$CB,$C6
	.dc.b	$C4,$C8,$C1,$0F,$E0,$53,$FD,$67,$0E,$E7,$F9,$A7,$DF,$FE,$29,$F1
	.dc.b	$0F,$B5,$A7,$EA,$22,$5F,$E5,$67,$4D,$9C,$BD,$BE,$AB,$CD,$E8,$C8
	.dc.b	$FC,$E6,$29,$F5,$19,$8F,$2C,$E1,$DC,$67,$1F,$8B,$D8,$BF,$F3,$AD
	.dc.b	$68,$EB,$E5,$7C,$5F,$46,$46,$08,$7C,$27,$2F,$C4,$8F,$99,$E3,$FC
	.dc.b	$DF,$3A,$4F,$A2,$32,$FA,$A3,$2F,$CC,$8F,$D0,$7C,$57,$F4,$0F,$41
	.dc.b	$F0,$BF,$CF,$69,$18,$EE,$DC,$7E,$AE,$B8,$56,$EB,$53,$92,$1F,$EF
	.dc.b	$57,$2D,$8E,$CF,$EF,$67,$F5,$1C,$52,$3F,$4A,$FF,$51,$F8,$9F,$D7
	.dc.b	$C7,$F8,$9F,$F0,$AF,$F7,$83,$FA,$AE,$6F,$D8,$1F,$C2,$23,$31,$FD
	.dc.b	$2E,$5F,$29,$CC,$91,$C4,$FE,$35,$0F,$B2,$E5,$E6,$49,$CC,$57,$A7
	.dc.b	$D9,$79,$4C,$72,$FE,$02,$D1,$99,$EA,$EB,$7F,$F1,$77,$DD,$FF,$01
	.dc.b	$9F,$F7,$AB,$6F,$F5,$B0,$8C,$4F,$F8,$D7,$EF,$FF,$D5,$13,$FE,$FF
	.dc.b	$9C,$BF,$E4,$B8,$79,$DF,$CA,$9B,$1B,$F1,$7F,$96,$D7,$08,$4E,$96
	.dc.b	$1E,$28,$C3,$E4,$9E,$1D,$7F,$C1,$6B,$FC,$B5,$3E,$88,$C4,$72,$C7
	.dc.b	$FC,$99,$18,$9F,$6F,$E1,$5F,$E0,$5A,$D1,$FE,$63,$9D,$57,$F9,$4E
	.dc.b	$5F,$19,$FD,$11,$B5,$E0,$9E,$BE,$1F,$F4,$8B,$E3,$3F,$E6,$DA,$DB
	.dc.b	$D8,$38,$C3,$D1,$5F,$D4,$3F,$0A,$7B,$F4,$97,$FC,$E3,$0A,$FF,$36
	.dc.b	$F0,$3F,$A0,$18,$C3,$EC,$DF,$29,$FE,$7F,$3A,$7D,$07,$FA,$06,$34
	.dc.b	$F9,$57,$BF,$E3,$24,$70,$F9,$27,$D7,$BE,$67,$33,$F2,$A7,$3F,$13
	.dc.b	$FE,$3C,$6C,$78,$06,$1D,$9C,$35,$E5,$6A,$E4,$D8,$DE,$1F,$DD,$BE
	.dc.b	$B1,$DD,$FE,$EA,$9F,$F4,$54,$61,$C0,$70,$ED,$AF,$FB,$C9,$CE,$9F
	.dc.b	$A2,$FA,$DF,$F2,$9C,$FA,$5A,$FD,$B9,$DF,$17,$82,$FF,$C1,$35,$AB
	.dc.b	$E5,$E2,$79,$E7,$FC,$1B,$F8,$23,$ED,$3F,$C1,$FE,$6F,$F7,$EC,$FF
	.dc.b	$D6,$8D,$53,$B7,$E5,$35,$FC,$64,$6C,$E1,$19,$C3,$F8,$27,$D9,$7F
	.dc.b	$87,$3E,$8F,$FB,$1D,$FE,$4D,$B3,$A8,$FE,$27,$3B,$F1,$24,$60,$FF
	.dc.b	$2A,$FE,$EA,$FA,$AF,$F2,$AB,$63,$31,$FE,$65,$F3,$F2,$C8,$C9,$FE
	.dc.b	$39,$F8,$4F,$EC,$AF,$EC,$BE,$1B,$FC,$FF,$E3,$FD,$86,$C7,$DD,$6B
	.dc.b	$FE,$C7,$0E,$1F,$F6,$53,$AC,$FF,$B1,$D7,$C3,$E3,$FF,$70,$7A,$3F
	.dc.b	$E7,$EE,$1F,$F4,$57,$C3,$B2,$A7,$F9,$F6,$26,$5F,$A8,$CB,$A6,$1F
	.dc.b	$EE,$95,$F5,$FF,$CA,$27,$C2,$D7,$4D,$7C,$ED,$8E,$33,$FE,$8B,$C2
	.dc.b	$57,$7A,$FE,$AF,$F5,$4F,$91,$FE,$31,$3F,$EE,$78,$C7,$A5,$F9,$E6
	.dc.b	$C7,$E8,$DA,$FC,$7F,$8C,$7A,$2F,$AB,$FD,$2C,$61,$B1,$99,$FD,$64
	.dc.b	$E5,$E9,$E3,$7F,$76,$6F,$A9,$E1,$8C,$47,$D5,$C3,$F8,$DC,$FD,$9F
	.dc.b	$AB,$FF,$3B,$6D,$F2,$DB,$5F,$EA,$DF,$B5,$FF,$17,$CF,$CD,$9C,$F2
	.dc.b	$FF,$90,$38,$51,$89,$D7,$C2,$70,$C1,$CB,$FE,$BA,$C7,$18,$70,$AB
	.dc.b	$F5,$9E,$0B,$FC,$3B,$5F,$B8,$D7,$F1,$AD,$93,$E1,$8C,$CF,$86,$DF
	.dc.b	$50,$F8,$CE,$F3,$32,$FC,$56,$1C,$6F,$75,$9C,$3F,$A9,$E8,$FA,$C7
	.dc.b	$D2,$7B,$63,$0F,$7E,$B7,$CE,$E6,$B7,$9F,$FB,$29,$B8,$74,$6E,$1F
	.dc.b	$A4,$AE,$18,$7A,$7E,$07,$FD,$17,$87,$F3,$7F,$0D,$F0,$59,$F8,$8F
	.dc.b	$9D,$9A,$73,$5F,$E5,$F9,$CC,$62,$7F,$DD,$97,$D1,$FE,$21,$7C,$BE
	.dc.b	$B3,$FC,$0F,$87,$F3,$66,$CF,$F3,$B6,$BF,$BF,$6A,$23,$C3,$3F,$F0
	.dc.b	$CB,$67,$F6,$DE,$C3,$D5,$F0,$7B,$7E,$AF,$AB,$FE,$01,$5F,$EF,$0F
	.dc.b	$AC,$FF,$1B,$BF,$D5,$E3,$39,$7D,$49,$CB,$FD,$EE,$32,$FF,$13,$6B
	.dc.b	$E5,$0E,$1F,$43,$5C,$F8,$DE,$33,$F2,$F8,$BC,$FC,$61,$FD,$A7,$E1
	.dc.b	$3F,$B2,$3F,$DE,$BE,$B3,$F1,$FE,$CF,$D4,$9C,$E3,$FE,$1C,$9C,$D7
	.dc.b	$1F,$5F,$F8,$95,$6A,$BF,$95,$35,$F4,$35,$F9,$CF,$4E,$0F,$F2,$2B
	.dc.b	$6B,$F9,$5F,$13,$E7,$BC,$58,$78,$FC,$9F,$D2,$7C,$17,$FA,$EE,$FD
	.dc.b	$51,$8F,$77,$51,$FD,$DB,$1F,$E3,$CD,$9F,$39,$1E,$27,$F9,$A3,$EB
	.dc.b	$38,$E9,$FE,$2D,$FA,$CF,$D1,$CF,$F8,$78,$7D,$67,$FB,$19,$FF,$09
	.dc.b	$46,$A7,$FC,$83,$1A,$F6,$3F,$3B,$39,$9F,$E9,$CD,$7E,$E3,$5F,$C9
	.dc.b	$35,$3D,$77,$FF,$35,$B6,$BF,$30,$79,$3E,$4F,$BF,$7F,$EB,$23,$F9
	.dc.b	$A9,$E9,$8F,$F7,$DB,$FE,$29,$33,$7E,$88,$CB,$FE,$4D,$8C,$BF,$D9
	.dc.b	$5A,$FE,$4D,$AF,$8B,$CE,$FE,$A9,$1D,$30,$FF,$40,$B4,$F3,$E4,$FA
	.dc.b	$CF,$35,$7F,$07,$F3,$25,$7F,$8F,$7E,$07,$FC,$66,$FF,$B8,$63,$F8
	.dc.b	$3C,$63,$1F,$F1,$B3,$5F,$EF,$4D,$7F,$50,$D7,$F9,$DF,$2F,$C8,$7F
	.dc.b	$CD,$51,$CF,$2F,$F6,$FC,$E5,$FF,$4C,$C6,$6F,$FB,$FF,$C6,$FC,$69
	.dc.b	$FF,$72,$B6,$3F,$C0,$7C,$9B,$C3,$FA,$4C,$E5,$FE,$1D,$7F,$CA,$9F
	.dc.b	$3F,$89,$FD,$EE,$31,$1E,$88,$FF,$00,$79,$0B,$FE,$5F,$87,$69,$FE
	.dc.b	$8D,$83,$E8,$3F,$4B,$1F,$CB,$3C,$DF,$E1,$F9,$AF,$F8,$F6,$AF,$AC
	.dc.b	$F7,$47,$F8,$7B,$C0,$FE,$C1,$39,$D7,$94,$7F,$83,$49,$3F,$E2,$B6
	.dc.b	$C7,$FA,$4F,$0E,$F3,$F1,$BD,$6E,$2F,$FA,$1B,$66,$AF,$93,$F4,$FC
	.dc.b	$9F,$A9,$B5,$B9,$D3,$67,$4D,$35,$F5,$9F,$0B,$E4,$7E,$1B,$63,$27
	.dc.b	$38,$CB,$4E,$DF,$59,$F2,$BD,$1F,$96,$D8,$C9,$E3,$8C,$B4,$CE,$BF
	.dc.b	$EC,$0C,$FD,$B3,$A8,$C6,$5F,$DE,$6B,$FD,$4F,$C2,$F8,$E7,$CD,$FF
	.dc.b	$67,$F5,$7F,$BA,$46,$2B,$E4,$74,$FF,$81,$6F,$85,$E8,$FC,$0F,$D9
	.dc.b	$DA,$F9,$1D,$7F,$84,$F8,$FF,$C2,$B6,$32,$7C,$D8,$CB,$4F,$F5,$67
	.dc.b	$2F,$F0,$DD,$5C,$EF,$AD,$FC,$75,$6B,$C1,$FC,$3C,$62,$BF,$05,$1F
	.dc.b	$E5,$32,$4F,$F9,$C5,$B1,$A8,$FF,$64,$D7,$E0,$FD,$0F,$CC,$57,$F3
	.dc.b	$EF,$A5,$FE,$7F,$F1,$B6,$FF,$15,$E0,$7F,$B3,$1B,$3A,$69,$A9,$FF
	.dc.b	$0F,$7D,$6B,$EA,$BF,$C5,$98,$BE,$8F,$31,$86,$DF,$D6,$AE,$0F,$E6
	.dc.b	$6B,$F3,$BA,$38,$D5,$A8,$FD,$17,$D6,$FF,$8B,$7E,$B7,$E5,$47,$FA
	.dc.b	$32,$31,$B7,$ED,$5F,$F3,$C1,$31,$FF,$4B,$36,$79,$C7,$C6,$E0,$77
	.dc.b	$F1,$8F,$F7,$17,$5F,$87,$3F,$DF,$5F,$F9,$63,$EB,$7F,$B6,$3F,$E1
	.dc.b	$D9,$D5,$7F,$A6,$27,$33,$FE,$4D,$DF,$FC,$67,$5A,$8E,$5E,$76,$27
	.dc.b	$E5,$FD,$6F,$CA,$FA,$DB,$8F,$C5,$6D,$FF,$94,$38,$15,$F4,$3E,$C3
	.dc.b	$3A,$79,$8F,$D1,$7A,$BF,$E9,$D6,$E3,$FE,$F1,$6E,$1F,$EB,$56,$BF
	.dc.b	$E2,$37,$27,$3E,$FF,$F0,$44,$FF,$B0,$F2,$FD,$5D,$FF,$97,$59,$1F
	.dc.b	$EC,$7F,$AE,$B8,$7E,$E7,$E0,$D5,$8C,$7F,$D7,$F5,$6F,$FC,$8D,$0F
	.dc.b	$AA,$FD,$06,$1E,$36,$BA,$6B,$F8,$76,$C7,$1F,$3B,$FC,$E5,$1F,$A9
	.dc.b	$8F,$F6,$6E,$98,$9F,$F4,$7F,$D6,$7F,$93,$F0,$FF,$6C,$37,$0F,$F6
	.dc.b	$C7,$7A,$31,$1F,$ED,$86,$BF,$EC,$D7,$86,$1F,$EA,$16,$D7,$FB,$6B
	.dc.b	$47,$D5,$F8,$3E,$06,$4F,$FB,$5E,$B7,$87,$FA,$92,$F7,$39,$AD,$CF
	.dc.b	$FB,$C6,$B7,$E4,$AF,$73,$9A,$F6,$FD,$27,$FB,$77,$3F,$53,$70,$34
	.dc.b	$98,$F8,$DF,$F4,$C7,$5F,$28,$FA,$A7,$FB,$EF,$07,$D7,$7F,$9F,$7E
	.dc.b	$BB,$F9,$7C,$7F,$9A,$FC,$5F,$D3,$C6,$23,$FC,$DD,$3B,$F2,$7F,$CE
	.dc.b	$11,$FC,$8A,$F0,$E2,$7F,$CB,$73,$AB,$FD,$8F,$B9,$CE,$BF,$D1,$95
	.dc.b	$A7,$C1,$75,$EE,$30,$E1,$7E,$4C,$3D,$57,$E0,$AF,$EF,$37,$5E,$A7
	.dc.b	$19,$F1,$46,$63,$E0,$99,$3C,$25,$3C,$38,$78,$23,$C9,$1D,$23,$09
	.dc.b	$F6,$3D,$30,$C9,$8B,$F2,$61,$E9,$9F,$8C,$77,$3E,$51,$4E,$58,$7A
	.dc.b	$1C,$4F,$2C,$3D,$4D,$8C,$9E,$39,$E1,$87,$86,$30,$7B,$9C,$61,$E6
	.dc.b	$9F,$97,$3E,$3C,$3C,$CF,$C2,$8E,$66,$AB,$E0,$9D,$4A,$E4,$DC,$7B
	.dc.b	$BE,$4B,$E3,$36,$7B,$C6,$23,$EF,$B7,$E4,$5F,$31,$FD,$2E,$53,$98
	.dc.b	$C3,$79,$3F,$DA,$CD,$9D,$DF,$FB,$16,$F9,$D7,$E3,$5F,$F7,$7C,$DF
	.dc.b	$89,$FF,$59,$E3,$BC,$7C,$18,$7E,$86,$F8,$DF,$CA,$0F,$27,$0D,$8E
	.dc.b	$13,$CA,$BA,$61,$F9,$7A,$E1,$9F,$5C,$7F,$DB,$4F,$09,$D5,$F1,$C7
	.dc.b	$FD,$E5,$7D,$70,$F6,$61,$E8,$9C,$D6,$DF,$2B,$F2,$70,$ED,$1E,$19
	.dc.b	$EF,$5E,$48,$FD,$05,$FB,$AB,$71,$8A,$DB,$DB,$0F,$29,$CD,$87,$2C
	.dc.b	$3C,$18,$75,$C7,$99,$1C,$1E,$1C,$3B,$91,$C1,$2B,$E0,$E1,$E2,$7E
	.dc.b	$8B,$71,$F4,$4F,$49,$D9,$F7,$4F,$C0,$61,$F1,$67,$71,$A3,$DC,$7B
	.dc.b	$0F,$E0,$9C,$3F,$D2,$7E,$17,$FB,$66,$FF,$D1,$DB,$29,$D3,$FE,$7B
	.dc.b	$E4,$7C,$06,$A7,$33,$FD,$53,$91,$7F,$CC,$E3,$94,$61,$FE,$B7,$19
	.dc.b	$9F,$EB,$5C,$1C,$EB,$F4,$75,$FD,$46,$3C,$7C,$DF,$D2,$FD,$FD,$44
	.dc.b	$9D,$1E,$58,$C3,$FD,$4E,$2F,$17,$F8,$9A,$B9,$4F,$03,$D1,$3E,$3B
	.dc.b	$F0,$C6,$0F,$F0,$A7,$5F,$F8,$C9,$FF,$3D,$B5,$FF,$22,$6B,$FE,$01
	.dc.b	$F1,$BC,$FB,$7F,$B7,$E7,$71,$F0,$F2,$FF,$05,$36,$F9,$4F,$F8,$2E
	.dc.b	$3E,$46,$5F,$DA,$7D,$3E,$64,$9F,$A5,$C5,$5F,$F0,$D4,$5C,$3D,$D8
	.dc.b	$FF,$5D,$7F,$C6,$18,$BC,$6F,$BB,$1F,$E1,$B6,$C6,$6F,$EF,$5E,$27
	.dc.b	$CC,$48,$F8,$6F,$AF,$EA,$BF,$C1,$31,$FE,$9E,$6B,$F1,$0F,$F3,$43
	.dc.b	$8B,$FA,$5E,$67,$C5,$9D,$42,$75,$AA,$5A,$2F,$D9,$39,$8C,$69,$EC
	.dc.b	$BE,$37,$C7,$EB,$1C,$3F,$9C,$84,$61,$FF,$04,$9B,$8F,$C7,$FD,$6F
	.dc.b	$68,$C1,$FE,$DB,$5A,$E8,$E9,$5C,$E7,$E0,$0F,$AB,$7D,$6F,$85,$B1
	.dc.b	$E0,$35,$1E,$5F,$AC,$7C,$8E,$E3,$F5,$53,$A3,$F5,$95,$C9,$FF,$10
	.dc.b	$35,$FF,$0C,$B9,$E0,$FF,$42,$C2,$FD,$2F,$85,$AF,$A2,$78,$6F,$FE
	.dc.b	$60,$D7,$C7,$3F,$77,$1F,$9C,$7E,$D3,$83,$FC,$ED,$5B,$8F,$DB,$FD
	.dc.b	$6F,$FB,$7E,$BE,$6F,$0A,$31,$1F,$80,$C3,$AB,$FE,$EB,$87,$C0,$72
	.dc.b	$E6,$93,$CA,$72,$D9,$CB,$C6,$7C,$31,$F3,$FE,$2A,$4F,$CF,$D1,$1F
	.dc.b	$3F,$E1,$F3,$3F,$2A,$3E,$7C,$EA,$3E,$7C,$60,$FA,$DD,$B2,$3A,$47
	.dc.b	$F1,$6F,$7F,$FB,$47,$0F,$F0,$F7,$FB,$FC,$FC,$26,$BF,$A5,$D5,$1C
	.dc.b	$37,$FC,$E3,$67,$F4,$2F,$0A,$E9,$E0,$77,$F0,$7F,$73,$EE,$CB,$FE
	.dc.b	$2A,$F8,$FE,$A7,$11,$DF,$E0,$1F,$3F,$FC,$B1,$B1,$C0,$F3,$7D,$47
	.dc.b	$AF,$CD,$91,$82,$1E,$9D,$11,$A8,$C4,$FC,$33,$1A,$7C,$39,$C4,$FC
	.dc.b	$3F,$1B,$E4,$79,$B2,$30,$43,$B3,$97,$C1,$1F,$3F,$C6,$FA,$1E,$54
	.dc.b	$9F,$9D,$19,$7E,$84,$65,$F4,$47,$EC,$FE,$92,$47,$E1,$5F,$E9,$FF
	.dc.b	$02,$F6,$73,$3F,$AE,$9F,$C2,$91,$FD,$9B,$FD,$A3,$CE,$7C,$2F,$EC
	.dc.b	$35,$FF,$7A,$C6,$39,$B9,$C7,$F5,$3E,$E7,$2F,$EA,$09,$8F,$EA,$33
	.dc.b	$FD,$9B,$FC,$8F,$EB,$5F,$03,$FC,$11,$B1,$FD,$BD,$AF,$FA,$8D,$AF
	.dc.b	$F2,$B3,$F9,$5C,$BE,$B7,$DE,$48,$FD,$77,$57,$C2,$FA,$DF,$E5,$16
	.dc.b	$F8,$BA,$66,$BF,$8D,$46,$5F,$14,$79,$48,$FF,$AD,$7B,$58,$BE,$B7
	.dc.b	$E7,$92,$79,$D7,$F3,$7F,$96,$FE,$0F,$BF,$F1,$A6,$F0,$7A,$DB,$5F
	.dc.b	$D1,$FB,$9F,$11,$C7,$E0,$3F,$94,$D7,$F1,$0A,$FF,$5E,$B5,$E5,$1A
	.dc.b	$3F,$A3,$57,$F4,$F8,$FE,$6C,$D7,$F9,$EC,$F8,$1F,$D0,$F8,$5F,$D3
	.dc.b	$7B,$7F,$84,$47,$F4,$26,$E5,$FC,$CB,$EB,$7F,$C1,$AF,$F3,$48,$DC
	.dc.b	$FF,$B4,$1A,$FD,$76,$BF,$D8,$DA,$F8,$0F,$E9,$52,$F8,$AF,$EB,$CD
	.dc.b	$8F,$E9,$D3,$E2,$D3,$F0,$5F,$5A,$FA,$3F,$EC,$8D,$AF,$EA,$8D,$7F
	.dc.b	$9A,$56,$A7,$FA,$9C,$E6,$7F,$DC,$ED,$7A,$C6,$8F,$E8,$7E,$14,$8E
	.dc.b	$B1,$E0,$F8,$C9,$1E,$07,$FC,$A3,$F2,$6F,$AD,$7C,$D6,$AB,$FD,$FD
	.dc.b	$8A,$FC,$64,$8F,$F2,$84,$65,$FE,$FB,$7C,$FE,$67,$AE,$07,$A6,$FF
	.dc.b	$D9,$BE,$B5,$E1,$FF,$81,$DB,$87,$F7,$76,$BF,$DF,$BE,$B7,$ED,$F0
	.dc.b	$7E,$B6,$73,$3F,$AF,$0E,$1D,$57,$F7,$C8,$C4,$7F,$87,$DE,$B7,$FD
	.dc.b	$E2,$FF,$CD,$13,$DF,$C8,$F2,$5F,$F8,$6F,$EB,$5F,$55,$49,$5F,$E0
	.dc.b	$DF,$43,$FB,$6E,$3F,$E0,$2A,$DC,$FF,$9D,$5A,$F9,$C3,$87,$D4,$D7
	.dc.b	$FC,$61,$DC,$FA,$DF,$F1,$67,$D1,$FF,$55,$DB,$F8,$73,$6B,$F8,$7C
	.dc.b	$FF,$2C,$37,$E2,$48,$FE,$3D,$1F,$E5,$4F,$AD,$7D,$6D,$24,$FF,$A4
	.dc.b	$7D,$3A,$FF,$92,$27,$6F,$F8,$EB,$44,$62,$3F,$B7,$7C,$2F,$F5,$E7
	.dc.b	$FC,$A9,$F0,$EF,$AD,$FE,$31,$18,$6D,$B5,$FF,$29,$56,$FC,$6F,$F4
	.dc.b	$9F,$E9,$3F,$CD,$23,$18,$7F,$92,$9A,$F3,$9D,$FC,$0F,$FA,$9F,$1E
	.dc.b	$F3,$FC,$CE,$B6,$D8,$CB,$E0,$6B,$FE,$B3,$F9,$EF,$E9,$98,$F9,$1B
	.dc.b	$19,$8F,$F4,$8F,$C0,$7F,$AD,$67,$36,$69,$C3,$CC,$6B,$F8,$6B,$FE
	.dc.b	$8D,$F5,$AF,$98,$FF,$51,$E5,$86,$BF,$E7,$58,$D7,$B1,$FE,$BF,$6C
	.dc.b	$7E,$2F,$E3,$7F,$AE,$FD,$77,$FB,$0D,$B1,$9F,$59,$39,$6C,$7E,$4E
	.dc.b	$BF,$8B,$7B,$1D,$7C,$93,$49,$3F,$AF,$DD,$F5,$DF,$E7,$6D,$7E,$A3
	.dc.b	$67,$F0,$7F,$5D,$F7,$5F,$C1,$C6,$79,$F1,$7C,$BE,$2B,$51,$89,$D6
	.dc.b	$1F,$E9,$1E,$16,$3A,$8F,$C3,$CE,$B1,$FF,$75,$7D,$6D,$C8,$E7,$8D
	.dc.b	$F5,$DF,$EB,$5C,$3F,$15,$3B,$26,$9D,$70,$FF,$51,$D6,$DF,$CD,$CE
	.dc.b	$AB,$8D,$6E,$FA,$FC,$87,$1C,$7C,$33,$CF,$57,$D7,$3E,$9F,$0D,$C7
	.dc.b	$FE,$00,$7D,$16,$26,$47,$EB,$E7,$5A,$7E,$BA,$B2,$FF,$93,$FA,$BD
	.dc.b	$0B,$C4,$7C,$47,$97,$E2,$3F,$D3,$F8,$FF,$A8,$E7,$2F,$FA,$22,$33
	.dc.b	$5F,$EA,$C9,$CB,$D6,$78,$57,$6F,$AE,$FF,$63,$47,$FA,$B5,$B3,$FC
	.dc.b	$03,$DB,$FF,$2B,$CE,$7E,$3F,$81,$1F,$DD,$46,$A7,$FD,$77,$5C,$B6
	.dc.b	$FF,$55,$FD,$6F,$F8,$8E,$FF,$94,$52,$30,$FF,$2B,$6B,$FD,$0D,$AF
	.dc.b	$F8,$CD,$DC,$F4,$F1,$3E,$B6,$0F,$43,$E4,$D6,$B0,$E1,$58,$37,$36
	.dc.b	$3F,$D9,$9F,$13,$FD,$75,$FF,$6E,$FD,$6F,$E3,$BD,$AF,$A7,$39,$CF
	.dc.b	$FD,$ED,$39,$AE,$3E,$B6,$8F,$AD,$F4,$36,$32,$4A,$FF,$46,$46,$5F
	.dc.b	$F7,$0D,$FF,$09,$9F,$F2,$D7,$C0,$78,$3C,$8F,$CA,$7A,$5F,$EB,$29
	.dc.b	$F3,$C6,$5F,$F3,$07,$D6,$DC,$39,$CD,$E1,$FD,$C7,$77,$F9,$2A,$B1
	.dc.b	$7F,$E5,$DF,$AD,$75,$7E,$9E,$1F,$5A,$F5,$FF,$27,$5A,$9F,$F8,$02
	.dc.b	$6A,$33,$52,$33,$4F,$99,$C6,$9F,$EF,$EC,$39,$36,$FC,$15,$BB,$FF
	.dc.b	$7F,$56,$B3,$FF,$79,$C7,$FA,$00,$F0,$E0,$F8,$6F,$D3,$C6,$71,$57
	.dc.b	$F1,$78,$F1,$4E,$BC,$C9,$1E,$6F,$93,$B7,$C2,$3F,$D0,$07,$3B,$F0
	.dc.b	$D7,$9A,$B7,$1F,$33,$BF,$F9,$6D,$7E,$C7,$CB,$FD,$CF,$C0,$7F,$89
	.dc.b	$23,$03,$1E,$C5,$8D,$38,$0E,$2F,$B4,$74,$F7,$7F,$81,$36,$BF,$80
	.dc.b	$34,$FF,$67,$67,$FC,$6F,$E4,$BE,$34,$FF,$A9,$6B,$13,$E6,$F1,$A4
	.dc.b	$FF,$AA,$5B,$7F,$EE,$FE,$6F,$E2,$F8,$FE,$29,$B5,$FE,$F0,$FA,$1F
	.dc.b	$CA,$DF,$F1,$8F,$27,$F6,$F8,$7F,$80,$BC,$0F,$E9,$0F,$F4,$F6,$BF
	.dc.b	$DE,$DA,$EE,$FE,$36,$3C,$2B,$F7,$53,$FE,$FC,$F5,$7B,$5C,$BD,$9C
	.dc.b	$CB,$3B,$6A,$AE,$7D,$2D,$9F,$2B,$F1,$FE,$5B,$E4,$37,$0E,$75,$B8
	.dc.b	$C6,$1F,$EF,$B6,$E7,$FA,$EC,$1F,$59,$F4,$7E,$53,$E9,$37,$1E,$73
	.dc.b	$A6,$CE,$9A,$6B,$EB,$3E,$D7,$A1,$F6,$DB,$19,$39,$C6,$5A,$67,$2F
	.dc.b	$C7,$E7,$EF,$9D,$46,$30,$FE,$C9,$97,$F7,$8E,$37,$CE,$FF,$5F,$6E
	.dc.b	$3E,$8F,$AC,$C4,$C8,$FF,$25,$4E,$B0,$E9,$59,$7B,$75,$FE,$4F,$E3
	.dc.b	$FF,$28,$DB,$F8,$6D,$9D,$C7,$C6,$9F,$C6,$78,$1F,$86,$9F,$B4,$DE
	.dc.b	$3F,$E1,$71,$FE,$1B,$8C,$1C,$2B,$F0,$7D,$7F,$7B,$CE,$C5,$7E,$BE
	.dc.b	$FF,$87,$57,$F9,$36,$FF,$C1,$17,$B8,$FC,$4C,$EA,$9C,$1F,$EF,$4F
	.dc.b	$91,$FE,$6B,$F1,$BF,$CD,$9B,$7F,$88,$EC,$FF,$76,$B6,$74,$D3,$FC
	.dc.b	$E5,$F5,$2F,$99,$FF,$07,$36,$3F,$23,$AF,$E8,$B8,$D3,$9C,$FF,$85
	.dc.b	$C6,$2F,$89,$FE,$1B,$AE,$15,$B9,$FD,$67,$D2,$7C,$92,$61,$FD,$D3
	.dc.b	$54,$63,$1F,$8C,$FC,$4F,$AB,$76,$7E,$FE,$25,$FC,$67,$51,$8A,$FE
	.dc.b	$39,$D1,$E1,$AF,$E4,$50,$8C,$3F,$33,$EA,$FF,$B1,$3C,$35,$FF,$11
	.dc.b	$CE,$A3,$84,$7F,$8D,$E7,$79,$FF,$0E,$AE,$6F,$C6,$FA,$BF,$96,$7C
	.dc.b	$3F,$A9,$B8,$FD,$08,$B8,$FE,$9B,$75,$62,$FF,$49,$E8,$7E,$8C,$FF
	.dc.b	$23,$7C,$0B,$EA,$F1,$97,$E8,$A6,$E3,$FE,$AA,$9B,$7F,$DC,$AF,$85
	.dc.b	$FF,$0C,$6D,$7F,$AC,$3E,$AF,$AD,$CA,$D3,$E0,$8C,$4F,$E0,$3E,$9D
	.dc.b	$E4,$7F,$95,$5B,$7D,$73,$C4,$C8,$FF,$60,$4E,$AB,$FD,$7F,$59,$7F
	.dc.b	$D7,$FF,$51,$FE,$3B,$CF,$FC,$FA,$DB,$FF,$3D,$7D,$5F,$F9,$96,$31
	.dc.b	$5F,$CB,$5B,$CC,$73,$D7,$FE,$A8,$AE,$1F,$37,$30,$FF,$61,$56,$B7
	.dc.b	$56,$71,$F1,$DD,$CB,$E5,$4D,$C3,$FD,$9F,$83,$E9,$3C,$87,$D6,$E8
	.dc.b	$B3,$DA,$2E,$3F,$ED,$E9,$B8,$7F,$52,$FA,$BF,$EE,$DE,$CF,$F2,$2A
	.dc.b	$FC,$BF,$81,$FE,$A1,$8C,$47,$E6,$A7,$7E,$37,$E7,$23,$FC,$06,$27
	.dc.b	$6F,$F8,$06,$7F,$C0,$5F,$57,$8B,$FF,$56,$3A,$94,$EA,$FA,$56,$9E
	.dc.b	$D3,$A1,$5C,$F3,$E1,$1E,$3B,$AF,$F8,$46,$F9,$39,$3C,$D7,$DA,$BD
	.dc.b	$71,$9C,$3A,$3A,$38,$61,$83,$D1,$87,$68,$F8,$31,$87,$CB,$7D,$5A
	.dc.b	$72,$71,$1B,$3A,$3E,$E9,$E9,$87,$A2,$BE,$81,$C8,$F6,$61,$E1,$7E
	.dc.b	$25,$F2,$7B,$1F,$6A,$35,$5D,$1B,$1C,$CF,$4D,$F8,$5C,$5F,$85,$FB
	.dc.b	$46,$27,$8D,$F7,$8F,$2B,$5D,$B5,$FB,$1F,$1F,$F8,$57,$E9,$35,$F6
	.dc.b	$B5,$EE,$E7,$E3,$DF,$2F,$FE,$5E,$94,$E6,$30,$DD,$3F,$CA,$0D,$9D
	.dc.b	$E9,$FC,$CA,$F9,$D7,$E1,$9F,$F6,$7C,$DF,$85,$FE,$6F,$8F,$E8,$27
	.dc.b	$73,$F4,$5B,$19,$9D,$E9,$D5,$B5,$FE,$61,$AD,$CF,$87,$2F,$55,$F1
	.dc.b	$AF,$01,$E3,$8C,$CE,$F0,$CC,$F8,$70,$EA,$B0,$EA,$F7,$BE,$2F,$86
	.dc.b	$77,$3E,$E8,$F5,$4E,$9E,$4D,$8D,$61,$E5,$9D,$DF,$94,$F1,$C6,$4F
	.dc.b	$8D,$5C,$4E,$57,$E0,$3C,$2E,$67,$F5,$F5,$E9,$9F,$EC,$B8,$F7,$39
	.dc.b	$CE,$B0,$F6,$D6,$CF,$0E,$3F,$15,$F2,$9F,$14,$C9,$F0,$E7,$D0,$FE
	.dc.b	$EC,$E4,$59,$F8,$95,$E0,$8F,$E0,$1C,$BF,$CE,$FE,$0B,$FB,$CD,$FF
	.dc.b	$44,$6E,$BF,$CC,$9B,$5F,$E2,$6E,$5F,$F6,$53,$FC,$FA,$33,$7C,$E3
	.dc.b	$07,$F6,$3F,$73,$3C,$9F,$9C,$8F,$E8,$BC,$18,$3A,$F3,$B1,$FF,$67
	.dc.b	$4E,$88,$7F,$5A,$9D,$C6,$A3,$06,$B3,$FE,$BB,$3A,$F6,$BB,$4E,$74
	.dc.b	$FE,$9B,$39,$8F,$ED,$3E,$27,$BA,$DD,$7F,$E2,$49,$C9,$FD,$F7,$DA
	.dc.b	$D4,$66,$3F,$BC,$56,$0F,$2C,$F9,$BE,$26,$4F,$F8,$15,$F4,$F1,$E2
	.dc.b	$3F,$CF,$7A,$27,$18,$FF,$83,$6B,$BF,$89,$2F,$FB,$FB,$AD,$3F,$D5
	.dc.b	$B9,$FF,$8C,$67,$24,$3A,$E1,$ED,$FA,$8F,$EE,$71,$FE,$AC,$C5,$18
	.dc.b	$8F,$F2,$8B,$61,$7F,$ED,$FD,$DE,$23,$5E,$9D,$D3,$1E,$98,$F1,$C3
	.dc.b	$85,$EE,$7F,$D7,$3F,$55,$F6,$72,$FF,$65,$B6,$BF,$C9,$6D,$9F,$F3
	.dc.b	$AF,$D5,$FF,$76,$AD,$F1,$FE,$52,$55,$B6,$C7,$E3,$2B,$85,$FF,$95
	.dc.b	$BE,$A9,$F2,$DC,$5B,$7F,$EC,$C6,$BF,$EF,$9F,$AB,$7C,$76,$EA,$4E
	.dc.b	$DF,$14,$EA,$30,$7F,$7C,$15,$D1,$FF,$06,$35,$F2,$4E,$F3,$FA,$77
	.dc.b	$F5,$C9,$3F,$E8,$82,$9F,$59,$B1,$F5,$A3,$FD,$37,$A7,$F8,$66,$31
	.dc.b	$C0,$FA,$BF,$F7,$2D,$7F,$B8,$9F,$CD,$4E,$9B,$9F,$DF,$77,$3F,$ED
	.dc.b	$06,$CE,$F9,$FF,$73,$1C,$27,$FD,$56,$EA,$BE,$87,$C4,$7C,$F2,$4F
	.dc.b	$FB,$EB,$5F,$C2,$2C,$3D,$4F,$C8,$9D,$B9,$AE,$6F,$1E,$3C,$BF,$32
	.dc.b	$7E,$83,$C4,$FA,$5B,$23,$EA,$66,$F6,$BF,$A4,$63,$FE,$FF,$F0,$BF
	.dc.b	$89,$C7,$F7,$4C,$3F,$DF,$BE,$16,$0C,$CC,$8C,$9F,$E7,$2F,$63,$F8
	.dc.b	$29,$11,$FE,$F5,$FA,$6C,$3F,$E7,$F8,$B8,$7D,$6C,$15,$E2,$FA,$5F
	.dc.b	$1F,$97,$30,$E7,$7B,$9D,$46,$37,$FD,$B6,$0B,$E1,$3F,$24,$FF,$6F
	.dc.b	$C2,$B8,$C6,$74,$F8,$95,$C3,$0E,$D3,$D4,$E3,$E2,$7C,$6F,$2E,$46
	.dc.b	$6F,$E6,$4F,$CA,$33,$1D,$71,$FF,$3E,$3B,$51,$F4,$3C,$2F,$A3,$E3
	.dc.b	$4A,$D4,$65,$E1,$19,$7A,$47,$EE,$FE,$12,$47,$EE,$DF,$E9,$BE,$C7
	.dc.b	$FB,$CE,$30,$43,$E2,$1F,$B6,$8D,$FA,$BC,$A3,$FC,$EB,$CA,$9F,$C6
	.dc.b	$B6,$3F,$1B,$1F,$8C,$E8,$71,$FE,$90,$93,$F1,$70,$FE,$1B,$BF,$E7
	.dc.b	$BE,$A1,$F5,$7F,$C6,$DB,$5F,$3C,$38,$7F,$65,$7E,$CF,$AB,$91,$D5
	.dc.b	$FD,$9F,$3F,$C0,$FA,$BE,$9F,$0D,$2B,$C5,$15,$F8,$38,$27,$19,$7F
	.dc.b	$08,$CD,$7F,$2B,$EA,$E6,$1E,$B7,$7D,$3F,$9E,$87,$CA,$3E,$13,$F9
	.dc.b	$B3,$63,$D8,$D7,$E8,$F6,$3E,$07,$75,$27,$6F,$F3,$F9,$D4,$60,$F8
	.dc.b	$CF,$F5,$46,$F8,$7C,$5F,$BB,$DB,$FC,$05,$F9,$8D,$73,$E0,$7F,$4B
	.dc.b	$FA,$BF,$F0,$B4,$60,$DC,$7F,$4F,$9E,$1B,$FF,$16,$AD,$BF,$C6,$23
	.dc.b	$3A,$7C,$39,$7C,$27,$53,$95,$FF,$59,$FA,$B7,$CF,$52,$65,$FC,$D7
	.dc.b	$1A,$FE,$5A,$76,$FE,$15,$AF,$F3,$58,$DE,$5F,$CE,$FC,$3F,$E6,$A6
	.dc.b	$33,$FA,$D6,$F8,$5F,$11,$30,$FD,$F7,$C7,$3E,$AD,$F2,$F9,$9F,$15
	.dc.b	$5C,$BF,$A8,$5E,$EB,$18,$FF,$20,$8E,$5F,$2D,$FE,$4A,$6E,$1E,$9C
	.dc.b	$3F,$BD,$7D,$5F,$EA,$E3,$F6,$6D,$C7,$FA,$1B,$70,$FC,$77,$D5,$FD
	.dc.b	$07,$F2,$11,$9A,$FE,$5E,$D4,$D9,$E1,$1A,$F0,$BF,$A3,$CF,$86,$3F
	.dc.b	$E0,$03,$5E,$B1,$FD,$33,$EA,$9F,$57,$FE,$11,$6C,$FE,$DD,$AF,$F0
	.dc.b	$CF,$7B,$FC,$08,$FF,$0E,$8C,$E5,$FD,$79,$B9,$7E,$A5,$B3,$FD,$22
	.dc.b	$B8,$D7,$EC,$FE,$AF,$FC,$41,$F3,$FF,$D5,$6B,$F8,$CB,$67,$51,$FC
	.dc.b	$3E,$77,$E1,$C8,$C1,$FE,$41,$FA,$97,$D5,$FF,$98,$5B,$19,$8F,$F3
	.dc.b	$7F,$D5,$E5,$91,$93,$FA,$07,$C1,$27,$FD,$07,$F5,$5F,$C2,$E7,$FA
	.dc.b	$47,$C2,$A3,$11,$FD,$4E,$31,$E4,$7F,$A0,$BE,$AF,$F9,$2F,$0F,$F6
	.dc.b	$89,$CD,$72,$BE,$3B,$7F,$87,$E7,$33,$FD,$66,$F8,$A8,$CC,$EE,$B8
	.dc.b	$DF,$56,$D3,$D5,$61,$CB,$07,$CD,$FF,$85,$31,$F0,$37,$1F,$33,$70
	.dc.b	$FF,$53,$7B,$9F,$EA,$48,$C5,$F2,$AC,$1F,$E9,$A9,$FF,$2D,$7D,$53
	.dc.b	$E8,$BF,$88,$E1,$FC,$AE,$31,$EB,$7F,$92,$DB,$1F,$CB,$CF,$F0,$6B
	.dc.b	$87,$1F,$15,$FD,$4B,$EA,$FF,$CA,$4D,$8E,$07,$FB,$B7,$D4,$4E,$9B
	.dc.b	$1A,$AF,$E2,$BF,$1B,$E1,$F2,$0D,$24,$F8,$F5,$7D,$47,$FA,$6F,$6E
	.dc.b	$F1,$8F,$89,$FF,$1E,$D7,$DD,$AD,$3C,$5A,$FF,$4B,$9E,$15,$F8,$A9
	.dc.b	$DF,$0F,$F9,$73,$EA,$5F,$19,$FE,$61,$8F,$C2,$4E,$AB,$B7,$CB,$65
	.dc.b	$FF,$4D,$75,$23,$19,$7E,$1E,$B1,$5F,$9C,$9C,$E5,$EC,$7D,$35,$B6
	.dc.b	$C6,$67,$DF,$5C,$AB,$D9,$5C,$A3,$18,$7A,$7E,$77,$13,$E2,$87,$EE
	.dc.b	$CF,$0B,$0D,$C7,$FC,$5D,$AB,$E7,$B3,$52,$7F,$D7,$95,$BC,$FF,$C7
	.dc.b	$97,$B9,$CD,$7C,$0E,$4F,$F3,$87,$8B,$F0,$AD,$F8,$FB,$A3,$5A,$7F
	.dc.b	$20,$C7,$FC,$29,$39,$7F,$C8,$98,$72,$73,$5C,$30,$E9,$5B,$C3,$FC
	.dc.b	$5F,$F5,$3F,$E0,$FB,$FF,$69,$7D,$83,$E0,$3C,$53,$9A,$FF,$2C,$66
	.dc.b	$AD,$5F,$F8,$96,$B5,$5F,$A3,$6B,$FE,$13,$6B,$FE,$72,$6A,$FC,$09
	.dc.b	$C5,$F7,$BF,$03,$FF,$32,$5F,$F8,$C6,$72,$FF,$1A,$8C,$D7,$F9,$22
	.dc.b	$72,$FF,$1B,$8C,$E3,$FE,$9A,$9C,$BF,$C5,$E3,$35,$FE,$27,$9C,$B6
	.dc.b	$FF,$27,$E8,$3F,$03,$FE,$6B,$3E,$39,$FA,$DF,$11,$FC,$1E,$3F,$C9
	.dc.b	$9E,$CE,$3C,$2F,$CD,$F8,$CF,$83,$FF,$50,$57,$F9,$2E,$F8,$6D,$FE
	.dc.b	$7F,$9C,$BF,$E6,$28,$CE,$DF,$E4,$1C,$7A,$5F,$1C,$7A,$E0,$FA,$CC
	.dc.b	$E3,$C7,$C2,$CF,$D0,$F9,$9B,$87,$85,$B3,$B8,$FF,$32,$4E,$DD,$1F
	.dc.b	$E6,$7F,$7F,$13,$FE,$76,$8F,$F3,$69,$FE,$93,$1F,$29,$FE,$4F,$7F
	.dc.b	$CD,$06,$A3,$FD,$40,$EB,$E0,$BF,$3C,$E7,$5F,$F7,$43,$93,$8F,$91
	.dc.b	$FC,$11,$C5,$78,$63,$89,$9F,$AC,$7C,$8E,$2B,$FC,$F1,$CA,$F8,$7F
	.dc.b	$A7,$F8,$3F,$F9,$FA,$73,$18,$F8,$BF,$D8,$56,$E7,$FD,$55,$5B,$CB
	.dc.b	$FD,$2B,$39,$7F,$D6,$D1,$9C,$FF,$DC,$F3,$96,$E1,$E0,$F4,$FA,$BF
	.dc.b	$63,$DB,$98,$F9,$BE,$AE,$E5,$D9,$B8,$74,$6C,$FF,$B6,$BA,$7F,$19
	.dc.b	$1E,$83,$F0,$65,$8F,$F4,$F7,$AF,$F8,$63,$F0,$FF,$6D,$63,$E2,$B8
	.dc.b	$69,$FE,$A5,$7F,$1B,$F0,$32,$3F,$D5,$1F,$05,$72,$FF,$5B,$F9,$EF
	.dc.b	$A9,$FE,$73,$EA,$3E,$A1,$F4,$BF,$C6,$EB,$58,$FF,$B6,$6B,$53,$9B
	.dc.b	$E3,$7C,$2B,$FD,$61,$7C,$30,$FF,$74,$61,$C2,$73,$7F,$D0,$A7,$FC
	.dc.b	$1C,$DA,$E1,$1F,$D8,$E3,$FA,$3E,$DF,$B3,$ED,$FD,$4F,$81,$FA,$A6
	.dc.b	$E1,$FE,$C5,$6D,$FE,$D5,$BC,$3F,$EB,$3F,$8D,$FD,$8F,$C3,$7E,$C9
	.dc.b	$B7,$FE,$CD,$6D,$FE,$F5,$B4,$9F,$B5,$DE,$FB,$6F,$DD,$C7,$F0,$88
	.dc.b	$C3,$FE,$D7,$31,$1A,$7F,$DF,$D1,$AB,$FE,$15,$1A,$38,$1A,$9E,$11
	.dc.b	$FC,$EB,$EA,$9F,$2B,$DF,$B7,$FB,$04,$62,$FA,$C6,$27,$FB,$7E,$2A
	.dc.b	$D5,$FF,$88,$63,$13,$A9,$CC,$FF,$B4,$F6,$65,$FE,$51,$C7,$D3,$E8
	.dc.b	$7B,$FE,$07,$E0,$36,$32,$7B,$A3,$2F,$E1,$3E,$9B,$FB,$B7,$DA,$FE
	.dc.b	$17,$4F,$99,$97,$F9,$5E,$73,$C7,$F9,$16,$DF,$11,$3F,$D7,$2F,$94
	.dc.b	$7F,$87,$E7,$F1,$44,$7F,$C6,$F3,$FE,$5B,$8F,$C5,$E5,$C2,$3F,$CB
	.dc.b	$0B,$ED,$7E,$D6,$1F,$6D,$B7,$F6,$1B,$4C,$7D,$3A,$7C,$69,$FF,$01
	.dc.b	$7B,$DF,$3A,$73,$F5,$8E,$CC,$F0,$FF,$4E,$8C,$65,$EC,$AF,$87,$97
	.dc.b	$F8,$D2,$BF,$4D,$87,$D4,$8F,$82,$66,$BD,$B1,$98,$F8,$BF,$98,$3B
	.dc.b	$BE,$7F,$D5,$FF,$85,$63,$17,$DC,$F0,$70,$A7,$55,$C2,$75,$7D,$23
	.dc.b	$0F,$C1,$8F,$1D,$F4,$8F,$F3,$2B,$5F,$F3,$73,$5D,$CF,$E3,$FC,$CF
	.dc.b	$D4,$7D,$6F,$E6,$9B,$7E,$0D,$3F,$31,$1B,$8F,$F3,$8C,$61,$BC,$C9
	.dc.b	$FC,$D7,$87,$FB,$3F,$A7,$FD,$2B,$6B,$A5,$23,$71,$C9,$B9,$7E,$96
	.dc.b	$7F,$C1,$2F,$BC,$FF,$4F,$7C,$1F,$F9,$EA,$B9,$F8,$BF,$E0,$88,$C4
	.dc.b	$FE,$DE,$3F,$58,$49,$F3,$BF,$C7,$3D,$FF,$ED,$51,$FD,$15,$B3,$FE
	.dc.b	$AB,$5F,$2D,$FE,$14,$73,$CE,$69,$23,$5E,$7F,$F0,$4F,$07,$C9,$77
	.dc.b	$0F,$F6,$7B,$4F,$74,$7F,$01,$24,$7F,$A4,$3F,$10,$FF,$2C,$4F,$F9
	.dc.b	$8B,$D1,$FF,$3D,$C6,$38,$7F,$D1,$51,$8F,$03,$FC,$F7,$3C,$2B,$FD
	.dc.b	$24,$D7,$FD,$A8,$D7,$FD,$9C,$F3,$BC,$CF,$BF,$E8,$BA,$47,$FB,$8A
	.dc.b	$75,$5C,$2B,$53,$97,$CB,$C9,$F1,$FC,$AB,$3F,$E8,$06,$D7,$FA,$BE
	.dc.b	$D5,$E8,$25,$7F,$08,$9F,$E1,$FF,$5B,$FC,$43,$EB,$5F,$59,$FE,$C9
	.dc.b	$F3,$3F,$8C,$C6,$2B,$FD,$99,$5B,$AD,$5D,$9C,$8A,$E0,$FF,$1C,$BE
	.dc.b	$75,$FE,$CE,$BE,$13,$9B,$AF,$90,$7B,$5C,$9F,$C8,$B5,$4F,$F2,$88
	.dc.b	$D5,$FF,$30,$75,$1D,$23,$C3,$3E,$6C,$37,$7E,$08,$E6,$7A,$70,$F6
	.dc.b	$B4,$E0,$E2,$7A,$C6,$4F,$51,$4F,$53,$D9,$C1,$D9,$C1,$C8,$B1,$F1
	.dc.b	$63,$95,$7A,$1F,$E4,$B7,$EB,$BF,$43,$63,$AB,$6F,$A5,$F9,$CE,$8F
	.dc.b	$69,$F5,$9F,$CC,$2F,$B3,$E8,$9E,$0E,$8F,$6C,$F4,$78,$5F,$89,$FF
	.dc.b	$62,$61,$E9,$F8,$3F,$B3,$1F,$84,$6B,$EA,$36,$7F,$60,$F8,$C7,$C9
	.dc.b	$7F,$51,$8C,$88,$CF,$17,$F6,$6A,$E7,$18,$9D,$B6,$33,$5C,$F1,$FC
	.dc.b	$9C,$B8,$7F,$A7,$C6,$1B,$9F,$F5,$29,$E9,$3C,$2B,$94,$E6,$B7,$87
	.dc.b	$A5,$AF,$9E,$76,$FB,$5A,$FA,$E3,$77,$E3,$8F,$D1,$BE,$0A,$F1,$57
	.dc.b	$08,$D4,$FE,$3D,$B7,$E5,$C7,$8D,$6F,$0E,$A7,$67,$91,$9B,$EF,$19
	.dc.b	$9E,$26,$22,$1D,$E7,$8C,$62,$BE,$44,$F0,$35,$19,$C3,$E3,$46,$67
	.dc.b	$CC,$FE,$F2,$32,$D3,$E5,$1F,$BB,$7E,$0C,$D8,$F9,$4F,$57,$67,$80
	.dc.b	$93,$E4,$7E,$35,$FB,$5F,$DD,$7A,$0E,$07,$95,$FE,$2A,$8C,$36,$33
	.dc.b	$3F,$D1,$E3,$07,$F8,$5B,$0F,$F1,$2F,$43,$E3,$F1,$C1,$FD,$7D,$B7
	.dc.b	$FD,$0A,$EE,$3E,$1F,$8D,$B9,$1A,$70,$EB,$FE,$C5,$F3,$58,$C3,$FA
	.dc.b	$34,$D7,$FB,$C4,$E9,$C9,$FD,$76,$3D,$44,$70,$7F,$88,$23,$FB,$77
	.dc.b	$D3,$F7,$D0,$F5,$D7,$FC,$B2,$7F,$9B,$75,$FF,$27,$B6,$FF,$C2,$B1
	.dc.b	$8F,$03,$FD,$41,$39,$9F,$2D,$BE,$C7,$B1,$F1,$27,$47,$CA,$E6,$F1
	.dc.b	$12,$BC,$D4,$FA,$5C,$47,$49,$B5,$F4,$A7,$A7,$D6,$3D,$2F,$3B,$6F
	.dc.b	$FC,$5A,$D7,$FD,$0C,$D7,$EE,$39,$AF,$BB,$F6,$1F,$14,$EB,$5D,$BE
	.dc.b	$C3,$A9,$2F,$AB,$5F,$AB,$EA,$3E,$0F,$EE,$36,$3F,$C8,$58,$AB,$D3
	.dc.b	$E5,$7A,$09,$3E,$86,$BD,$3A,$F9,$9C,$1F,$F2,$D8,$FC,$D1,$F0,$1F
	.dc.b	$DF,$DB,$1F,$E8,$86,$BF,$E7,$03,$FD,$22,$E1,$C3,$AD,$7F,$07,$FB
	.dc.b	$A3,$0F,$B4,$DA,$F0,$D6,$F9,$3F,$CF,$6D,$8F,$F4,$64,$ED,$FF,$28
	.dc.b	$5A,$30,$FF,$A1,$A7,$84,$BE,$C3,$E4,$3E,$73,$1B,$7F,$86,$E3,$31
	.dc.b	$EB,$6B,$FE,$8C,$78,$D7,$FA,$73,$EC,$3F,$61,$3F,$A3,$6B,$FE,$63
	.dc.b	$3A,$9F,$ED,$CF,$9A,$FB,$3F,$39,$23,$04,$3E,$FB,$97,$F0,$31,$FC
	.dc.b	$09,$7D,$63,$EC,$3F,$C6,$4D,$8F,$F5,$DB,$5F,$73,$5F,$AC,$E7,$0F
	.dc.b	$F5,$5F,$D8,$7A,$1B,$DD,$DE,$75,$BF,$FA,$B1,$B7,$FE,$CE,$BE,$13
	.dc.b	$FE,$B0,$6C,$7F,$B4,$A7,$7D,$DF,$EE,$6F,$27,$F3,$3F,$39,$2B,$FD
	.dc.b	$FF,$1F,$9C,$30,$FE,$82,$30,$FE,$7B,$4F,$D9,$91,$FD,$9B,$5F,$E3
	.dc.b	$7F,$5C,$F4,$3F,$91,$36,$39,$1F,$C6,$5F,$E6,$0D,$F3,$F6,$70,$F2
	.dc.b	$24,$64,$87,$67,$2F,$82,$38,$BD,$63,$C3,$E1,$78,$BC,$49,$3E,$18
	.dc.b	$CB,$E3,$8C,$BE,$A8,$F4,$FD,$9F,$AB,$C1,$93,$E9,$8C,$BE,$B8,$CB
	.dc.b	$F0,$A3,$ED,$7C,$97,$E1,$89,$1F,$86,$6B,$F9,$7E,$17,$CE,$7E,$3D
	.dc.b	$B1,$FC,$13,$89,$5D,$23,$F8,$3C,$E9,$E5,$59,$7E,$9F,$CA,$7D,$22
	.dc.b	$3F,$47,$E0,$3F,$39,$F5,$EF,$88,$FA,$FF,$3F,$2B,$F8,$9C,$E5,$FB
	.dc.b	$1E,$DF,$DC,$23,$F6,$E7,$EF,$FB,$7D,$3E,$26,$4F,$DF,$8A,$FE,$1B
	.dc.b	$D6,$FC,$29,$8E,$E7,$6A,$5D,$AE,$9F,$F2,$6C,$7E,$26,$F8,$BE,$6C
	.dc.b	$7D,$7E,$D4,$8F,$4B,$FD,$CF,$E1,$9E,$EE,$67,$FB,$8C,$3E,$2B,$46
	.dc.b	$74,$F1,$CF,$8F,$83,$FD,$66,$FF,$71,$AE,$0D,$AE,$1A,$FA,$61,$F5
	.dc.b	$DF,$A1,$7F,$8A,$37,$0E,$B1,$AE,$FF,$E1,$DF,$5D,$D2,$37,$18,$F4
	.dc.b	$64,$7F,$46,$BE,$58,$77,$6C,$7C,$9B,$57,$08,$E6,$7F,$4E,$FA,$C7
	.dc.b	$C7,$FF,$1D,$6C,$66,$3F,$73,$EE,$E5,$FD,$DD,$C9,$D1,$0F,$E0,$2E
	.dc.b	$5F,$E6,$55,$D1,$BF,$1F,$DC,$8F,$F0,$3F,$AD,$AF,$F0,$AF,$73,$F8
	.dc.b	$49,$8D,$BE,$3D,$AF,$35,$FC,$2E,$B5,$AF,$D1,$9F,$F1,$1C,$FF,$6C
	.dc.b	$6B,$FD,$15,$B7,$F4,$BE,$27,$FC,$3D,$3B,$F1,$BF,$A0,$47,$F1,$68
	.dc.b	$FE,$98,$D9,$FB,$35,$CA,$FC,$17,$C3,$C7,$98,$73,$3F,$BE,$FD,$5B
	.dc.b	$E5,$3F,$DF,$4D,$8F,$EC,$81,$C3,$FD,$9D,$AF,$F7,$F7,$38,$FE,$2B
	.dc.b	$EB,$3F,$C5,$D3,$FE,$0A,$6B,$FD,$8E,$77,$8F,$EB,$7E,$93,$FD,$E3
	.dc.b	$E0,$FF,$5D,$FC,$8F,$FB,$6C,$FE,$65,$B6,$AD,$BF,$B5,$C7,$F4,$B1
	.dc.b	$88,$FE,$E1,$3C,$13,$CA,$35,$3F,$E5,$4F,$49,$F3,$1F,$E4,$B6,$DF
	.dc.b	$5C,$3B,$46,$23,$FC,$A7,$F1,$3B,$A9,$1B,$17,$FD,$D6,$3F,$CC,$93
	.dc.b	$FE,$78,$F8,$0F,$E3,$55,$AC,$3F,$87,$7C,$37,$F9,$3F,$C8,$D1,$FE
	.dc.b	$79,$FA,$97,$CA,$62,$FF,$C3,$15,$B6,$E1,$D1,$FE,$01,$19,$7F,$C6
	.dc.b	$8D,$CF,$FD,$1F,$E8,$FF,$9D,$CE,$93,$E3,$13,$FE,$61,$F8,$0F,$41
	.dc.b	$23,$FD,$01,$8F,$F8,$0A,$75,$1E,$29,$FF,$51,$61,$FE,$41,$8F,$6B
	.dc.b	$5F,$E5,$9F,$53,$FC,$03,$5F,$C7,$B6,$3F,$C6,$8D,$7F,$C9,$8D,$7F
	.dc.b	$DD,$9E,$67,$E0,$DF,$F1,$D7,$C6,$23,$11,$FE,$53,$8C,$71,$7E,$71
	.dc.b	$B1,$DB,$A9,$ED,$78,$9B,$19,$8C,$1F,$43,$EA,$7F,$D5,$DC,$7F,$7F
	.dc.b	$64,$62,$BF,$D2,$FE,$8F,$FA,$82,$BE,$EC,$E7,$8B,$F9,$DD,$F0,$F7
	.dc.b	$35,$87,$D0,$AD,$5F,$F5,$1F,$45,$F1,$2C,$B7,$8F,$1A,$F1,$5B,$DD
	.dc.b	$FE,$77,$8F,$81,$AF,$56,$D7,$A1,$B3,$C6,$B9,$C7,$93,$1F,$0F,$AB
	.dc.b	$E0,$AD,$61,$F8,$7C,$39,$61,$C0,$FF,$57,$D6,$F0,$F6,$E8,$FA,$9F
	.dc.b	$F0,$6D,$7E,$9E,$31,$EB,$F7,$D5,$EB,$F0,$BD,$E7,$FE,$7E,$6D,$FF
	.dc.b	$B3,$1B,$87,$C7,$BE,$2F,$58,$CD,$7F,$09,$A7,$C8,$7D,$08,$7D,$4B
	.dc.b	$A3,$E9,$F3,$7E,$EA,$B5,$39,$C7,$F7,$D3,$97,$FD,$4F,$19,$BF,$D5
	.dc.b	$4E,$5F,$F4,$14,$67,$0F,$ED,$FF,$4F,$FE,$AC,$BF,$E5,$ED,$AF,$F4
	.dc.b	$A3,$5F,$F5,$63,$5F,$EF,$8E,$E7,$9F,$5F,$E1,$3D,$1F,$F6,$F5,$FF
	.dc.b	$AE,$AB,$4F,$F8,$0F,$E9,$9C,$5F,$8A,$FA,$77,$9B,$F0,$E7,$35,$F9
	.dc.b	$D9,$CB,$FE,$C4,$8C,$E7,$FE,$A9,$6B,$FE,$C1,$6B,$DE,$FF,$22,$A3
	.dc.b	$8E,$7F,$E1,$06,$D7,$F8,$BF,$E8,$5F,$4E,$F9,$7F,$26,$3F,$5F,$E4
	.dc.b	$BF,$C0,$F7,$FC,$BF,$AD,$F4,$FF,$E5,$48,$FF,$6C,$CE,$B4,$FE,$39
	.dc.b	$5A,$F8,$3E,$5E,$BF,$92,$B5,$5F,$ED,$96,$BF,$E2,$96,$BF,$E4,$37
	.dc.b	$A6,$DF,$E2,$86,$E1,$FB,$7C,$5F,$4F,$DA,$3C,$7E,$1F,$EA,$E5,$F3
	.dc.b	$9E,$B8,$C3,$67,$FB,$8D,$FF,$83,$5B,$3E,$72,$3D,$0F,$85,$5F,$E2
	.dc.b	$E7,$18,$FF,$B4,$5F,$E4,$87,$F5,$51,$F1,$DF,$EA,$57,$FD,$FE,$6A
	.dc.b	$3F,$A1,$3A,$F7,$7F,$4C,$E6,$3F,$DF,$CD,$7F,$C9,$CD,$7F,$CD,$0D
	.dc.b	$7F,$CD,$7E,$47,$23,$F8,$FC,$3E,$9D,$F4,$F9,$8F,$ED,$54,$E5,$FE
	.dc.b	$EB,$1F,$7F,$E9,$DF,$0D,$FB,$99,$CD,$7D,$E9,$CD,$72,$9C,$CF,$F9
	.dc.b	$8C,$E9,$E9,$FE,$52,$33,$1F,$E6,$76,$BF,$93,$8E,$C7,$F2,$3A,$FD
	.dc.b	$5C,$66,$BE,$8F,$B3,$F7,$4F,$27,$CF,$D8,$F1,$3D,$C8,$FF,$8B,$B0
	.dc.b	$47,$F3,$6E,$7F,$CD,$BC,$A7,$EE,$B6,$FF,$37,$C1,$FA,$59,$FC,$0F
	.dc.b	$4F,$E3,$7D,$DF,$C7,$3F,$15,$F7,$12,$3C,$0F,$79,$7D,$1F,$E5,$4F
	.dc.b	$CB,$7D,$3D,$AF,$A7,$E6,$BE,$9B,$FA,$BF,$A4,$FA,$77,$C9,$7F,$2B
	.dc.b	$AD,$5F,$F8,$8E,$B5,$39,$BF,$E0,$75,$AD,$7F,$C3,$71,$82,$5F,$81
	.dc.b	$EF,$F2,$5F,$9D,$9F,$F0,$B1,$FE,$16,$8F,$D6,$75,$FE,$B7,$C0,$B7
	.dc.b	$FA,$46,$D7,$C8,$6F,$81,$FE,$AC,$E3,$FE,$13,$1F,$B5,$F8,$9F,$DB
	.dc.b	$7C,$65,$C3,$F5,$AD,$C3,$E8,$B7,$0E,$5C,$5F,$EB,$CA,$FE,$29,$F1
	.dc.b	$5F,$ED,$B7,$F6,$A7,$F6,$B7,$17,$F6,$A3,$1E,$FF,$F0,$56,$9C,$0E
	.dc.b	$6E,$5A,$66,$3F,$C7,$16,$F9,$EF,$F2,$3C,$F8,$BB,$7F,$B8,$C6,$2B
	.dc.b	$F9,$1E,$1F,$A6,$78,$C6,$6F,$FD,$9E,$73,$F8,$0F,$47,$9F,$C9,$F5
	.dc.b	$7B,$DE,$BF,$4E,$E7,$F0,$1B,$8F,$66,$E1,$F3,$F2,$FE,$E3,$5F,$8C
	.dc.b	$FA,$0F,$F0,$2F,$D3,$7E,$0E,$BF,$DA,$7C,$E9,$CF,$37,$E3,$5F,$C4
	.dc.b	$92,$FF,$D2,$F3,$B6,$C6,$DF,$C9,$75,$3E,$27,$1A,$7C,$2C,$3E,$A7
	.dc.b	$D3,$FD,$56,$DF,$DA,$6C,$EE,$3E,$EC,$FF,$82,$7D,$FF,$91,$1F,$E3
	.dc.b	$7F,$8F,$FD,$5B,$FC,$6A,$33,$4E,$A5,$6A,$3F,$37,$3A,$7F,$C7,$DC
	.dc.b	$5F,$E3,$19,$FE,$9B,$03,$37,$FC,$9B,$E9,$9E,$17,$CA,$FA,$7F,$F0
	.dc.b	$E4,$62,$7F,$C5,$DC,$49,$CD,$7F,$74,$AD,$5F,$F2,$50,$E1,$FE,$50
	.dc.b	$D7,$C4,$6A,$39,$DF,$E8,$3E,$86,$CF,$E3,$3D,$1F,$D0,$7C,$A7,$E4
	.dc.b	$9B,$5F,$E4,$1E,$AF,$EC,$6D,$8F,$D3,$65,$B8,$D7,$91,$FA,$9F,$A0
	.dc.b	$FC,$DB,$63,$24,$C7,$F3,$BC,$DA,$8E,$DE,$4F,$F9,$8B,$A7,$F4,$CD
	.dc.b	$CF,$F9,$55,$FF,$45,$9C,$D7,$C3,$6C,$7E,$9A,$BF,$CA,$DE,$87,$EB
	.dc.b	$1B,$5F,$E1,$3F,$81,$FE,$D5,$7F,$C3,$57,$C6,$EB,$0F,$F0,$9C,$6A
	.dc.b	$BF,$CB,$18,$7F,$98,$6B,$57,$FB,$2C,$7F,$86,$61,$E5,$6B,$F4,$A3
	.dc.b	$F8,$07,$89,$EB,$26,$1F,$35,$FF,$36,$7C,$B7,$F0,$27,$FD,$13,$F4
	.dc.b	$9F,$B2,$BF,$F6,$2B,$5F,$F6,$4B,$5F,$F6,$AB,$5F,$F6,$2B,$C6,$FF
	.dc.b	$D3,$6D,$BF,$F3,$97,$CB,$76,$F0,$A4,$7F,$A9,$E7,$57,$FA,$3D,$57
	.dc.b	$A8,$FF,$46,$4E,$BC,$13,$C3,$B7,$FD,$47,$5B,$CB,$C3,$CD,$FE,$9C
	.dc.b	$A3,$49,$97,$FB,$A2,$32,$F8,$E3,$FD,$21,$F2,$DF,$11,$FF,$42,$65
	.dc.b	$FA,$FF,$A6,$FD,$87,$D3,$3E,$93,$FC,$15,$E8,$B1,$CC,$7E,$A6,$AD
	.dc.b	$FE,$9A,$B4,$AF,$93,$FE,$13,$9F,$F2,$65,$74,$BB,$39,$1E,$E7,$07
	.dc.b	$F8,$16,$1D,$23,$11,$E7,$3B,$57,$36,$9C,$F0,$DD,$79,$1E,$30,$C3
	.dc.b	$D2,$F6,$C2,$BD,$67,$8C,$62,$38,$1E,$6B,$F0,$61,$D2,$FD,$73,$CC
	.dc.b	$D1,$E3,$9F,$35,$EE,$3C,$E6,$DC,$1E,$07,$27,$8E,$38,$9F,$45,$CE
	.dc.b	$1E,$D8,$C4,$F1,$3B,$5F,$32,$C7,$A0,$D9,$D2,$BC,$C7,$DB,$C3,$E0
	.dc.b	$61,$EC,$86,$1F,$4E,$BC,$57,$D5,$A7,$59,$FB,$91,$97,$EF,$7C,$47
	.dc.b	$DE,$6C,$7A,$E3,$11,$FC,$C2,$FF,$A1,$46,$3E,$3D,$CA,$DF,$38,$7F
	.dc.b	$36,$9C,$C6,$30,$FE,$95,$18,$7F,$DC,$13,$BA,$E5,$18,$36,$E2,$3F
	.dc.b	$25,$3F,$B8,$9F,$BA,$D7,$EE,$87,$B4,$70,$9E,$58,$F9,$78,$3F,$D5
	.dc.b	$CD,$BE,$73,$A7,$C3,$87,$E5,$30,$ED,$AF,$8B,$0F,$EA,$6F,$59,$D5
	.dc.b	$F1,$BF,$15,$76,$7E,$0B,$5F,$34,$6E,$FE,$25,$F5,$9F,$05,$76,$C3
	.dc.b	$E3,$61,$D8,$CC,$C3,$F3,$37,$CA,$77,$18,$AE,$4F,$88,$E4,$70,$C3
	.dc.b	$C5,$1E,$5A,$F9,$98,$FB,$6B,$E6,$C7,$57,$91,$D4,$E7,$8E,$CF,$53
	.dc.b	$93,$B6,$3E,$0A,$C9,$31,$F3,$BC,$5F,$A9,$3C,$6B,$0F,$F4,$2E,$AF
	.dc.b	$EA,$5B,$7F,$46,$6C,$64,$FE,$C9,$F3,$38,$E4,$FE,$77,$3F,$E3,$8E
	.dc.b	$3F,$F1,$F5,$7F,$35,$8C,$19,$F9,$0E,$12,$69,$DB,$95,$CC,$F8,$7C
	.dc.b	$E1,$FD,$3E,$AB,$FD,$C2,$2E,$FF,$D6,$26,$E9,$FE,$2D,$C3,$9D,$70
	.dc.b	$BF,$EC,$75,$E3,$7F,$A5,$3F,$E2,$77,$F6,$1D,$5F,$CF,$9F,$F3,$5D
	.dc.b	$FC,$2C,$3F,$CE,$8F,$9F,$17,$D9,$66,$F8,$5F,$F7,$B6,$BF,$E3,$FB
	.dc.b	$E2,$FF,$84,$E3,$2F,$F6,$9E,$1E,$84,$7F,$C2,$5F,$4F,$5F,$F0,$7C
	.dc.b	$5A,$FF,$0A,$4F,$97,$E9,$DE,$8D,$25,$FF,$86,$EE,$BF,$E2,$59,$DB
	.dc.b	$FE,$25,$6B,$99,$F8,$96,$97,$8D,$DC,$95,$FE,$A6,$F9,$1A,$FF,$A5
	.dc.b	$3E,$17,$EE,$7A,$67,$38,$7F,$A5,$30,$78,$FE,$92,$5F,$FA,$C3,$E0
	.dc.b	$EB,$FE,$9E,$EA,$E6,$62,$FE,$B7,$D3,$3D,$6F,$CB,$D2,$30,$FD,$1E
	.dc.b	$4F,$F2,$FC,$62,$3F,$D5,$3C,$BC,$23,$11,$FE,$64,$39,$57,$E1,$5A
	.dc.b	$7E,$1E,$36,$F4,$8F,$C3,$7D,$3F,$B2,$3C,$07,$F3,$B7,$0F,$FB,$82
	.dc.b	$7C,$58,$7E,$3F,$E9,$FE,$34,$F0,$0E,$1F,$F7,$3B,$63,$FD,$C1,$F4
	.dc.b	$FF,$E0,$B7,$96,$FF,$AB,$8F,$D7,$57,$8B,$A9,$18,$8E,$73,$F5,$BE
	.dc.b	$9B,$EB,$FD,$3C,$9F,$F7,$0C,$65,$FF,$71,$C6,$5F,$F6,$8C,$7D,$0F
	.dc.b	$A6,$7C,$F5,$24,$FA,$FE,$99,$58,$8F,$F4,$B5,$70,$7F,$64,$D7,$35
	.dc.b	$F4,$63,$19,$FE,$D6,$FD,$4D,$E0,$E8,$FF,$B5,$7C,$AF,$C8,$FC,$DC
	.dc.b	$C7,$FD,$F7,$1F,$94,$30,$FF,$BA,$A3,$0F,$D8,$9F,$D0,$65,$FA,$D2
	.dc.b	$47,$CB,$E8,$FF,$47,$72,$7E,$F7,$A3,$DB,$3F,$3F,$E9,$9F,$21,$FC
	.dc.b	$75,$B1,$D6,$76,$FF,$7F,$AF,$E3,$12,$FA,$4E,$1F,$15,$2B,$A4,$73
	.dc.b	$30,$F1,$3F,$DA,$92,$9D,$47,$03,$C8,$F0,$CF,$FC,$B7,$CA,$AF,$05
	.dc.b	$7A,$7E,$2B,$D1,$F0,$72,$30,$43,$C1,$5C,$63,$51,$8A,$57,$C1,$FA
	.dc.b	$1F,$85,$F4,$F2,$BE,$34,$65,$F9,$11,$97,$94,$7E,$37,$E5,$64,$7E
	.dc.b	$97,$99,$1F,$C0,$39,$9E,$06,$AB,$F1,$B3,$F8,$92,$3F,$8E,$78,$78
	.dc.b	$52,$3A,$47,$D7,$F9,$69,$1F,$6D,$FB,$5F,$0B,$FC,$3B,$E9,$DF,$17
	.dc.b	$FC,$15,$B5,$FD,$2D,$AF,$FB,$CD,$AF,$FB,$C3,$87,$E8,$7B,$BF,$85
	.dc.b	$24,$7E,$27,$E8,$3F,$11,$EF,$F7,$F8,$C9,$3F,$8D,$8A,$F5,$8F,$11
	.dc.b	$1F,$E3,$47,$E1,$3D,$AF,$C9,$11,$FE,$1D,$8A,$7D,$9E,$EC,$9C,$BF
	.dc.b	$DD,$7E,$21,$F0,$39,$9F,$64,$FC,$12,$3F,$05,$E1,$F1,$D2,$39,$9F
	.dc.b	$C7,$7E,$77,$F4,$6F,$E6,$43,$8E,$CF,$EE,$4D,$EF,$4C,$F8,$25,$FF
	.dc.b	$40,$AE,$73,$FC,$E5,$AF,$03,$AF,$85,$FC,$EE,$33,$1D,$1E,$CF,$5F
	.dc.b	$0F,$FA,$5D,$9E,$D5,$FD,$3B,$E9,$DF,$23,$49,$7F,$C0,$7E,$0A,$BF
	.dc.b	$4A,$76,$FF,$83,$DA,$E6,$7F,$A7,$46,$34,$FD,$AF,$84,$CB,$F8,$51
	.dc.b	$D2,$3F,$C2,$D2,$F9,$2F,$21,$30,$FE,$25,$F0,$35,$F1,$47,$62,$3E
	.dc.b	$37,$F8,$B7,$BD,$23,$FA,$95,$75,$7F,$AA,$4B,$C6,$FE,$DE,$D8,$CC
	.dc.b	$7F,$72,$F8,$BF,$E0,$31,$83,$FC,$5D,$3B,$7F,$2A,$D7,$FB,$83,$5F
	.dc.b	$B0,$EF,$1F,$F1,$A4,$FF,$16,$9F,$C4,$36,$3F,$C7,$13,$BE,$2F,$C0
	.dc.b	$7D,$33,$E5,$FF,$C0,$AD,$C3,$FB,$73,$5F,$C9,$7D,$3B,$E2,$37,$5F
	.dc.b	$E0,$1A,$D5,$7F,$81,$5A,$F6,$8F,$F2,$57,$D1,$7F,$BC,$7C,$0F,$F0
	.dc.b	$9F,$D3,$FF,$77,$CB,$F2,$4D,$B5,$6D,$FF,$03,$B5,$FD,$24,$6D,$3C
	.dc.b	$63,$5A,$2B,$FC,$9D,$EA,$BE,$73,$FC,$B0,$DC,$3F,$AE,$35,$FE,$73
	.dc.b	$F0,$DE,$F7,$33,$C0,$8F,$F8,$26,$34,$FF,$4F,$F8,$9F,$F2,$7F,$8F
	.dc.b	$FE,$83,$F8,$3F,$E3,$F9,$78,$E9,$5A,$7F,$C2,$5F,$4E,$F9,$5E,$18
	.dc.b	$49,$FF,$13,$D6,$DF,$E9,$B5,$B7,$38,$7F,$2A,$7F,$81,$CE,$DF,$DF
	.dc.b	$35,$CC,$6D,$F1,$BF,$E3,$0D,$33,$F0,$5E,$93,$FD,$25,$7F,$E0,$E8
	.dc.b	$CD,$78,$27,$4F,$AB,$5F,$F2,$7F,$D3,$FF,$02,$E4,$FF,$45,$36,$7F
	.dc.b	$98,$B5,$FF,$26,$35,$FF,$7E,$79,$5F,$82,$7F,$9B,$FC,$72,$31,$1F
	.dc.b	$E5,$38,$C7,$17,$FA,$5D,$B1,$DB,$A1,$EE,$F8,$9B,$19,$8C,$1F,$D1
	.dc.b	$37,$FF,$43,$EF,$F5,$F8,$BE,$C4,$63,$7F,$F2,$CF,$E0,$2F,$8A,$FE
	.dc.b	$DB,$1F,$4A,$F8,$3F,$39,$B8,$FC,$E6,$D7,$1F,$57,$38,$75,$AF,$C5
	.dc.b	$53,$E6,$3F,$D4,$F3,$FE,$9A,$6C,$67,$0F,$43,$6B,$FC,$ED,$5B,$9F
	.dc.b	$C2,$D7,$8B,$D2,$72,$FE,$8E,$72,$4B,$F0,$D6,$E7,$51,$89,$FF,$21
	.dc.b	$7C,$C7,$FA,$35,$BC,$DE,$AF,$A1,$79,$3F,$03,$81,$7B,$C3,$FC,$69
	.dc.b	$18,$BD,$4E,$63,$E7,$DF,$2C,$3A,$9F,$E8,$1F,$A5,$D3,$FE,$C9,$FA
	.dc.b	$A7,$D5,$FC,$EC,$95,$A5,$82,$B7,$7C,$F2,$FF,$67,$DF,$96,$FF,$D3
	.dc.b	$CD,$7F,$B3,$35,$E7,$93,$E7,$7F,$DB,$78,$FB,$9B,$19,$C3,$FC,$1F
	.dc.b	$39,$6C,$66,$BE,$1B,$67,$FD,$DD,$C0,$F9,$5F,$8B,$87,$C6,$F9,$BF
	.dc.b	$B5,$E7,$3E,$4F,$19,$FF,$9B,$5B,$9F,$FA,$DF,$2F,$A3,$7E,$42,$57
	.dc.b	$89,$DC,$7F,$BF,$7E,$75,$E5,$E2,$BF,$DE,$DA,$F8,$DF,$D0,$FE,$43
	.dc.b	$FE,$D9,$7F,$07,$AF,$E7,$23,$19,$7B,$23,$E1,$92,$7E,$0B,$63,$51
	.dc.b	$D3,$E0,$7F,$07,$3F,$84,$F8,$DF,$D3,$F3,$7E,$BA,$FE,$3C,$63,$AF
	.dc.b	$58,$7C,$C8,$FA,$24,$70,$7E,$61,$E7,$F1,$DC,$2B,$F2,$7F,$7D,$FE
	.dc.b	$85,$AF,$F4,$B7,$8D,$F9,$98,$C4,$FF,$9E,$63,$1F,$15,$F8,$09,$CC
	.dc.b	$FF,$8A,$F8,$51,$88,$FF,$1A,$C6,$27,$81,$CA,$FF,$85,$FA,$DF,$97
	.dc.b	$9D,$47,$FA,$BE,$DF,$07,$FB,$9F,$DE,$7F,$A3,$4E,$63,$1E,$6F,$9D
	.dc.b	$F2,$12,$3F,$23,$5C,$9B,$1B,$7F,$DC,$55,$FE,$D6,$7A,$18,$EE,$F0
	.dc.b	$1F,$0F,$82,$C7,$95,$F8,$44,$7B,$BF,$17,$D8,$91,$E2,$6E,$BF,$E5
	.dc.b	$11,$E1,$B0,$7D,$87,$BC,$F8,$1E,$8D,$C3,$E2,$76,$7D,$12,$3E,$E7
	.dc.b	$FB,$76,$1F,$E8,$BB,$FF,$51,$CF,$F3,$8F,$D7,$7F,$C4,$0F,$C0,$F8
	.dc.b	$8F,$E6,$71,$8A,$FF,$38,$78,$2F,$92,$FE,$18,$BE,$5E,$C6,$1B,$87
	.dc.b	$FA,$32,$BF,$BE,$CF,$89,$FE,$E2,$D7,$FB,$AB,$5D,$C7,$E1,$8D,$95
	.dc.b	$FE,$F7,$D3,$D9,$FF,$66,$E4,$FA,$2F,$C2,$1F,$85,$F2,$2D,$FD,$C8
	.dc.b	$B5,$C2,$6F,$3F,$F5,$DF,$DD,$FE,$2C,$FF,$B5,$7E,$77,$F9,$C4,$62
	.dc.b	$7F,$CD,$F1,$8D,$E6,$3F,$88,$F8,$3B,$87,$F3,$CC,$3F,$D0,$93,$F8
	.dc.b	$CF,$73,$9D,$7E,$65,$B1,$FC,$92,$77,$5F,$C5,$DA,$FE,$AD,$AF,$EA
	.dc.b	$9D,$E3,$E9,$BF,$D2,$7D,$1F,$4A,$F8,$BF,$47,$FA,$73,$F5,$1E,$3D
	.dc.b	$C7,$F3,$F1,$67,$E5,$CD,$C3,$E2,$7D,$37,$EB,$CF,$86,$5B,$FA,$90
	.dc.b	$C9,$3A,$9F,$D5,$FD,$3D,$C3,$F1,$7F,$35,$F8,$AF,$9A,$FC,$5B,$6B
	.dc.b	$EF,$4F,$F9,$12,$3F,$09,$CE,$69,$27,$F1,$D1,$97,$8B,$67,$FA,$FC
	.dc.b	$7F,$B1,$3D,$67,$DB,$31,$F0,$CF,$AB,$3F,$F5,$BB,$FE,$52,$F0,$7F
	.dc.b	$03,$F1,$3F,$E0,$16,$C6,$E3,$FC,$09,$DF,$8C,$BF,$CA,$7C,$FF,$E0
	.dc.b	$73,$FC,$11,$CD,$6F,$FC,$BB,$16,$7F,$CB,$33,$78,$7F,$A8,$71,$FF
	.dc.b	$83,$FA,$3F,$BD,$79,$E8,$C6,$3F,$1A,$30,$FE,$BD,$B8,$7F,$BF,$1A
	.dc.b	$FF,$0D,$6B,$FE,$DF,$77,$7E,$09,$FF,$38,$F3,$F8,$2F,$FC,$41,$CF
	.dc.b	$27,$FC,$EB,$D1,$71,$FF,$38,$CD,$C3,$FC,$F9,$E0,$FF,$8B,$2B,$17
	.dc.b	$FE,$F4,$8F,$F1,$7E,$1F,$E2,$B3,$FC,$59,$D3,$67,$FD,$03,$17,$1F
	.dc.b	$F3,$EC,$DC,$3F,$CA,$5E,$43,$D8,$FF,$07,$B7,$0F,$F0,$6B,$6F,$FC
	.dc.b	$00,$D7,$FA,$A3,$9A,$E7,$5F,$E3,$DE,$BF,$F1,$B4,$E6,$FD,$1E,$BF
	.dc.b	$E1,$DF,$F4,$AC,$FF,$90,$F9,$F8,$47,$FA,$61,$B8,$FF,$A5,$1B,$8F
	.dc.b	$A3,$E3,$7C,$E4,$AF,$3B,$6B,$FC,$87,$8A,$3F,$CA,$1C,$DE,$47,$FC
	.dc.b	$CB,$F6,$4F,$99,$FF,$31,$B7,$0F,$F3,$C3,$70,$FF,$30,$B5,$FF,$50
	.dc.b	$B9,$C7,$FC,$F6,$D9,$FA,$DF,$8C,$FF,$41,$9C,$DF,$80,$93,$FC,$CF
	.dc.b	$85,$5B,$8F,$F5,$7D,$F3,$C3,$FD,$7A,$DE,$E3,$9B,$E3,$C7,$D2,$BF
	.dc.b	$D9,$13,$FD,$D5,$F4,$F9,$7E,$E2,$61,$EE,$7E,$19,$FE,$75,$F7,$EC
	.dc.b	$7F,$0A,$90,$FA,$7B,$F4,$FF,$D6,$63,$27,$FA,$92,$73,$E4,$FF,$73
	.dc.b	$8E,$02,$78,$3D,$3E,$07,$FD,$B5,$18,$8E,$F7,$E0,$F3,$31,$5F,$ED
	.dc.b	$F7,$53,$D1,$CE,$1B,$C3,$C1,$7E,$2A,$F2,$DF,$69,$CC,$78,$DF,$F0
	.dc.b	$44,$66,$3F,$C1,$53,$C5,$F3,$55,$7C,$A7,$13,$AB,$97,$C9,$5D,$23
	.dc.b	$C9,$1D,$0F,$F3,$74,$65,$A6,$70,$F5,$BD,$4F,$9B,$5E,$1B,$F3,$46
	.dc.b	$0D,$1F,$44,$A7,$D0,$9E,$67,$C8,$9E,$57,$E7,$6C,$F0,$7C,$EF,$C0
	.dc.b	$3C,$A7,$7C,$3C,$A7,$37,$C3,$7E,$47,$E2,$DF,$83,$0F,$9F,$5C,$1A
	.dc.b	$7C,$17,$91,$99,$F8,$CD,$7D,$DE,$9F,$FA,$55,$F9,$4D,$73,$1F,$72
	.dc.b	$31,$F0,$4F,$97,$FF,$5B,$D2,$73,$18,$6F,$27,$E7,$9B,$3B,$BF,$CE
	.dc.b	$DF,$3A,$F8,$AF,$FB,$52,$6F,$A5,$FA,$2A,$FC,$EC,$EE,$7E,$4B,$63
	.dc.b	$33,$BC,$7A,$B6,$BF,$5F,$5B,$9F,$0E,$3E,$AC,$B8,$D7,$E5,$6B,$C4
	.dc.b	$FE,$1D,$F0,$47,$8A,$7E,$0E,$1C,$2B,$C7,$87,$75,$7D,$DF,$0D,$77
	.dc.b	$C3,$CB,$1D,$EB,$F7,$F8,$75,$C7,$C5,$7B,$AF,$4D,$6A,$F8,$C7,$96
	.dc.b	$74,$F8,$1B,$87,$AB,$0F,$31,$D6,$78,$61,$F3,$98,$7C,$CA,$F6,$DF
	.dc.b	$C6,$9F,$34,$C3,$89,$2B,$AC,$F9,$1E,$C7,$7C,$7E,$31,$23,$A4,$F7
	.dc.b	$AE,$95,$CA,$3E,$3C,$7F,$0F,$E3,$FE,$6B,$F3,$3F,$EB,$9D,$FF,$9D
	.dc.b	$B6,$BF,$9D,$B7,$91,$D8,$F9,$5D,$61,$CA,$FF,$A9,$DF,$F8,$DB,$94
	.dc.b	$D3,$1F,$1B,$C2,$0D,$3B,$7E,$F5,$FD,$8E,$BF,$D0,$57,$FD,$2A,$75
	.dc.b	$18,$AE,$2F,$F5,$5A,$E0,$FF,$4C,$9D,$B4,$D7,$5E,$1F,$EC,$B1,$E4
	.dc.b	$7F,$B5,$BF,$E9,$BD,$5D,$AF,$A5,$F8,$91,$86,$C6,$6B,$FB,$36,$AF
	.dc.b	$97,$CD,$73,$AD,$57,$F8,$5A,$30,$7F,$AA,$77,$F9,$BC,$1F,$3F,$D7
	.dc.b	$7C,$75,$3E,$96,$5F,$EB,$EC,$FF,$C6,$2D,$AF,$F1,$85,$70,$C7,$FB
	.dc.b	$C3,$5F,$F3,$29,$FE,$D0,$9E,$0A,$BD,$FE,$F7,$B0,$91,$CB,$0A,$FF
	.dc.b	$8B,$A3,$D8,$47,$FB,$56,$2C,$1F,$0F,$EB,$24,$F9,$3E,$66,$BE,$4A
	.dc.b	$E1,$38,$C7,$FC,$9B,$6B,$F8,$9E,$3C,$C3,$DA,$FC,$7F,$A9,$C4,$73
	.dc.b	$AF,$F2,$43,$A3,$E1,$C7,$E7,$57,$D2,$F5,$7F,$D2,$6D,$7E,$39,$AB
	.dc.b	$FA,$9F,$19,$C6,$75,$3F,$EF,$B6,$BF,$E0,$E6,$BF,$E1,$46,$BC,$CF
	.dc.b	$AB,$F1,$5F,$7C,$E9,$5F,$81,$E5,$93,$F6,$9F,$C2,$7D,$3E,$23,$DD
	.dc.b	$17,$4F,$8F,$3F,$96,$FC,$D7,$F3,$35,$FE,$B8,$6B,$F8,$29,$46,$1F
	.dc.b	$9F,$F1,$7C,$79,$BF,$1F,$F7,$52,$F9,$47,$C9,$FC,$B5,$F1,$14,$93
	.dc.b	$FE,$F9,$F7,$9F,$76,$F9,$DC,$CF,$F8,$BE,$F8,$BF,$E3,$10,$E1,$D5
	.dc.b	$E5,$CB,$E1,$F8,$89,$1F,$EC,$A9,$E1,$A3,$B3,$FC,$86,$FB,$BE,$9B
	.dc.b	$E2,$12,$BF,$CB,$9F,$0C,$F5,$71,$1F,$93,$87,$A3,$FC,$85,$B1,$F2
	.dc.b	$EB,$1E,$14,$8E,$6E,$FC,$8C,$47,$38,$B8,$FF,$AA,$2F,$F9,$54,$FD
	.dc.b	$5F,$A3,$FB,$04,$7C,$B9,$D7,$FC,$B2,$7E,$DF,$95,$3F,$C5,$DB,$1F
	.dc.b	$E5,$BB,$FB,$B3,$93,$F9,$D7,$ED,$7F,$C0,$E3,$FD,$72,$DB,$F8,$F7
	.dc.b	$CA,$FF,$0C,$BE,$63,$F1,$84,$8F,$F4,$25,$7E,$9E,$FE,$84,$79,$67
	.dc.b	$FA,$47,$CA,$3E,$5E,$9C,$67,$F8,$05,$BC,$0F,$29,$23,$AF,$87,$5D
	.dc.b	$C5,$C7,$D3,$3F,$93,$F8,$B9,$5F,$83,$7F,$2D,$CF,$88,$FC,$84,$7E
	.dc.b	$2F,$94,$D2,$47,$E2,$F2,$FC,$F1,$C7,$0F,$8F,$E2,$7F,$65,$EA,$78
	.dc.b	$CF,$A7,$C3,$70,$F8,$B5,$FB,$3F,$4B,$EB,$12,$3E,$B7,$17,$F9,$15
	.dc.b	$FF,$4A,$F9,$0E,$17,$8B,$F6,$09,$3F,$43,$C7,$FE,$A7,$FB,$C3,$E2
	.dc.b	$7F,$AC,$36,$7F,$04,$D7,$F1,$3B,$A3,$11,$FE,$A4,$8C,$79,$5C,$89
	.dc.b	$1C,$9F,$EA,$D8,$FF,$6E,$9F,$C4,$F9,$1F,$8C,$3C,$73,$E4,$F7,$25
	.dc.b	$FE,$CB,$AF,$F6,$78,$A7,$F2,$7E,$46,$AB,$CB,$39,$7D,$33,$EC,$23
	.dc.b	$F9,$0F,$3B,$F3,$67,$1C,$BF,$22,$7B,$7C,$4F,$8A,$49,$FE,$29,$FA
	.dc.b	$37,$E5,$F2,$F1,$CF,$C8,$24,$FC,$86,$BF,$96,$F9,$5F,$D4,$9C,$DF
	.dc.b	$E5,$5F,$5D,$CD,$FE,$D4,$DC,$0E,$7A,$5F,$95,$B8,$FF,$C1,$73,$FD
	.dc.b	$1A,$B7,$E7,$FF,$84,$63,$04,$7F,$21,$5F,$DB,$30,$FE,$5C,$D7,$C7
	.dc.b	$1D,$39,$64,$7D,$13,$CA,$FE,$67,$E5,$FF,$CE,$78,$7C,$8C,$3F,$1F
	.dc.b	$5F,$C2,$5A,$FE,$21,$AF,$E1,$DA,$FF,$2E,$77,$1F,$C0,$B2,$57,$FA
	.dc.b	$76,$BF,$A2,$35,$F9,$37,$F0,$B7,$91,$E7,$BF,$93,$1F,$DB,$FC,$1C
	.dc.b	$CF,$8E,$3E,$31,$1F,$43,$FC,$7B,$E6,$24,$73,$CB,$FC,$63,$7F,$C5
	.dc.b	$A1,$F5,$4F,$96,$FF,$02,$36,$3E,$41,$A8,$FF,$04,$7D,$5B,$E8,$77
	.dc.b	$1F,$D5,$6B,$85,$7F,$36,$38,$1F,$D7,$5C,$BF,$CE,$BE,$93,$F3,$CE
	.dc.b	$DA,$E7,$9F,$FB,$04,$62,$04,$78,$E3,$EB,$FA,$5F,$E5,$D3,$E8,$6B
	.dc.b	$FE,$1A,$9D,$F3,$FD,$7E,$1F,$C8,$92,$7F,$CD,$FF,$AD,$FF,$8E,$5F
	.dc.b	$91,$19,$7F,$3E,$FF,$A0,$23,$51,$FE,$4C,$F7,$DF,$4A,$F9,$1F,$F4
	.dc.b	$2B,$67,$FC,$F6,$D7,$FC,$6D,$C7,$A9,$91,$A7,$FC,$C5,$18,$8F,$EC
	.dc.b	$35,$DA,$7F,$C1,$0D,$8F,$A0,$D7,$74,$F9,$0F,$19,$23,$FC,$2D,$FB
	.dc.b	$4B,$E3,$F1,$1E,$BA,$E0,$FA,$2B,$FA,$05,$6E,$3F,$0B,$F4,$DF,$E0
	.dc.b	$9F,$8C,$FF,$0D,$36,$3F,$04,$D7,$FB,$CB,$5F,$F7,$BF,$A8,$FA,$6F
	.dc.b	$F1,$17,$C6,$7F,$6F,$6C,$66,$7F,$19,$39,$7A,$FB,$8F,$8E,$F0,$B6
	.dc.b	$33,$18,$3E,$2F,$A0,$3F,$93,$F2,$7E,$33,$83,$11,$8A,$78,$3F,$DE
	.dc.b	$7D,$1C,$56,$AF,$FD,$6B,$7C,$DA,$F1,$AE,$11,$83,$7E,$2F,$29,$FE
	.dc.b	$45,$1F,$EA,$8F,$2F,$13,$23,$66,$4B,$8F,$E3,$EB,$8B,$67,$D1,$E7
	.dc.b	$FF,$B5,$AF,$94,$FF,$97,$FE,$93,$FD,$27,$87,$F0,$D6,$C7,$E6,$7E
	.dc.b	$69,$F2,$1F,$07,$BD,$5A,$7F,$A6,$B6,$FF,$C9,$AD,$B6,$AC,$39,$D7
	.dc.b	$7F,$1F,$E6,$53,$E9,$7C,$2F,$F8,$F1,$AF,$EF,$8D,$47,$F1,$77,$11
	.dc.b	$A7,$F8,$1B,$E1,$77,$3F,$EE,$8F,$AB,$FF,$18,$3F,$C6,$A7,$CD,$97
	.dc.b	$B3,$CA,$F7,$67,$FE,$0E,$AF,$E3,$F9,$BC,$DC,$D4,$9F,$90,$FA,$A7
	.dc.b	$51,$83,$E6,$7A,$3F,$4B,$DD,$7A,$FD,$A3,$0D,$9D,$9F,$CB,$BE,$6B
	.dc.b	$F6,$CE,$36,$FB,$15,$B7,$FD,$67,$3A,$73,$3F,$6F,$2F,$C8,$7D,$37
	.dc.b	$F5,$08,$C1,$FE,$93,$71,$3F,$83,$9F,$82,$70,$F4,$2D,$70,$9D,$3C
	.dc.b	$63,$31,$F8,$BF,$A3,$FC,$F7,$27,$F9,$0A,$BF,$19,$18,$EA,$F9,$AC
	.dc.b	$8F,$F5,$24,$F8,$B8,$7F,$17,$39,$BE,$95,$AF,$27,$15,$FE,$1C,$FA
	.dc.b	$57,$C5,$7C,$DF,$13,$F2,$D1,$8B,$FE,$FD,$D6,$E3,$56,$AE,$57,$FA
	.dc.b	$D7,$E2,$7F,$07,$87,$1C,$FF,$D8,$8D,$CB,$C1,$7F,$BB,$AD,$BE,$0F
	.dc.b	$98,$FD,$7F,$C3,$3E,$9B,$19,$7F,$AB,$1B,$97,$4A,$DF,$AB,$8C,$BD
	.dc.b	$91,$6F,$FA,$EE,$CC,$7F,$B7,$47,$F8,$0F,$45,$78,$3D,$F9,$39,$E2
	.dc.b	$B9,$FA,$9F,$8A,$7C,$6F,$8B,$B3,$CD,$C3,$FE,$C2,$F8,$DB,$5F,$EC
	.dc.b	$1F,$A1,$F6,$9E,$EF,$12,$D7,$8F,$B7,$E8,$9F,$C7,$FE,$9B,$FC,$B9
	.dc.b	$19,$3F,$D9,$BF,$45,$9F,$67,$E3,$B6,$3F,$5F,$ED,$FE,$AF,$E6,$BF
	.dc.b	$58,$D8,$C9,$33,$72,$7F,$86,$7C,$37,$6F,$0A,$D4,$E6,$FF,$6D,$39
	.dc.b	$BF,$F6,$E7,$D1,$7D,$83,$EC,$F9,$57,$0F,$F7,$1C,$59,$FF,$71,$4D
	.dc.b	$C3,$FC,$DB,$F4,$6D,$3F,$93,$37,$0F,$F7,$5F,$D1,$E2,$64,$7F,$8E
	.dc.b	$E7,$58,$7F,$BB,$6B,$2F,$B2,$7F,$DE,$F2,$9F,$C6,$ED,$D9,$B3,$CD
	.dc.b	$C3,$FE,$FA,$F6,$AD,$F3,$3E,$3D,$F3,$9E,$06,$CB,$1F,$EF,$FF,$A3
	.dc.b	$FC,$D9,$F9,$CF,$66,$CF,$EB,$3B,$13,$8A,$ED,$DB,$F9,$7B,$C6,$BF
	.dc.b	$1B,$E8,$FF,$4A,$7C,$62,$E3,$F0,$DC,$3D,$7E,$8E,$E1,$E9,$3F,$A5
	.dc.b	$FD,$2F,$F2,$FF,$A6,$FE,$60,$DA,$FE,$6E,$E1,$FF,$3F,$7A,$CF,$A6
	.dc.b	$61,$9C,$3A,$74,$30,$C6,$7F,$E7,$7B,$DE,$3D,$0F,$B1,$F1,$5F,$78
	.dc.b	$FE,$DB,$F3,$CE,$FC,$4E,$89,$97,$C4,$9D,$46,$30,$3F,$BE,$8F,$EC
	.dc.b	$CF,$F1,$86,$CF,$1E,$7E,$FD,$12,$B4,$F8,$A3,$13,$FD,$13,$8F,$CB
	.dc.b	$D3,$FD,$BB,$D1,$FA,$F8,$75,$8C,$4C,$8F,$07,$4D,$BE,$51,$FD,$4E
	.dc.b	$7D,$F3,$9C,$FD,$34,$E6,$69,$FE,$5F,$C3,$D5,$D3,$F1,$BA,$BE,$3B
	.dc.b	$23,$27,$BE,$31,$7F,$1F,$2F,$9A,$7C,$EE,$8B,$5F,$16,$2C,$FC,$D9
	.dc.b	$B9,$7F,$B8,$3C,$9F,$F1,$94,$FE,$86,$3F,$A3,$FB,$3F,$63,$AB,$FC
	.dc.b	$39,$5A,$9E,$13,$83,$58,$FF,$AC,$67,$4F,$D7,$E8,$CE,$1F,$75,$B7
	.dc.b	$F1,$36,$F8,$71,$8A,$E0,$7D,$DC,$7C,$3F,$1D,$23,$F0,$1C,$D7,$1F
	.dc.b	$46,$4A,$C6,$5D,$78,$17,$8C,$3F,$CB,$1F,$43,$FE,$79,$7F,$DB,$DD
	.dc.b	$48,$C6,$5F,$B6,$FA,$6F,$E6,$AF,$FA,$5E,$33,$8F,$DC,$AE,$1F,$0B
	.dc.b	$FE,$5E,$9C,$E1,$FD,$53,$0E,$37,$B9,$CE,$1F,$E4,$06,$E1,$CA,$FF
	.dc.b	$B9,$DF,$07,$CF,$87,$52,$3A,$E3,$B8,$79,$7C,$BF,$11,$FB,$0F,$7E
	.dc.b	$D7,$BB,$2F,$EC,$6D,$9D,$34,$FF,$6C,$D7,$FB,$23,$E9,$BF,$D9,$3F
	.dc.b	$4A,$F1,$FF,$C9,$D3,$FE,$89,$AC,$C7,$FB,$66,$AC,$FF,$94,$FE,$07
	.dc.b	$F4,$37,$C6,$38,$CE,$9F,$F6,$AC,$EA,$B9,$42,$B8,$3A,$BE,$0F,$94
	.dc.b	$D4,$F9,$2B,$94,$76,$32,$78,$E7,$8A,$BE,$73,$C2,$3F,$CD,$51,$89
	.dc.b	$DB,$63,$99,$A2,$9F,$0C,$A7,$53,$33,$D5,$F4,$5F,$B1,$E0,$D7,$83
	.dc.b	$98,$E1,$18,$BF,$75,$F9,$23,$AB,$6B,$83,$67,$73,$E4,$BF,$55,$77
	.dc.b	$6C,$FD,$08,$E3,$3B,$3A,$95,$C9,$C5,$CB,$4C,$C7,$DA,$F5,$3F,$CF
	.dc.b	$73,$F4,$DA,$FC,$96,$BF,$6D,$CF,$C1,$3E,$9B,$FC,$33,$3F,$E5,$4F
	.dc.b	$37,$FD,$3D,$39,$E3,$4E,$4B,$19,$70,$1C,$5F,$6B,$EB,$8F,$E3,$A7
	.dc.b	$FD,$77,$5F,$95,$6B,$F8,$07,$A4,$6E,$78,$E3,$E4,$F4,$7F,$D6,$53
	.dc.b	$CD,$B1,$E8,$9D,$DF,$99,$AF,$96,$76,$FA,$9A,$FA,$63,$77,$E2,$AF
	.dc.b	$04,$F9,$EB,$73,$98,$ED,$8E,$58,$7A,$EF,$99,$D2,$B8,$9C,$9C,$CF
	.dc.b	$0C,$3D,$95,$F1,$67,$2F,$C5,$AE,$2E,$6A,$39,$39,$61,$F0,$E3,$37
	.dc.b	$D2,$74,$79,$63,$33,$EE,$8C,$CF,$D1,$9C,$C7,$C7,$3A,$9F,$00,$FE
	.dc.b	$03,$C9,$FC,$F7,$E9,$BF,$DC,$D3,$FC,$CD,$BB,$7F,$44,$6D,$7F,$77
	.dc.b	$E7,$7C,$FE,$30,$E5,$3F,$CA,$EB,$1C,$67,$8F,$6F,$F7,$EB,$CF,$D2
	.dc.b	$CB,$FE,$64,$EB,$E9,$BF,$AC,$46,$08,$7F,$4F,$8D,$76,$71,$8C,$1A
	.dc.b	$D7,$FB,$55,$72,$8C,$92,$3F,$A8,$57,$28,$FE,$ED,$F1,$BE,$E9,$FF
	.dc.b	$62,$B7,$B8,$FA,$B6,$7D,$0E,$62,$5F,$23,$FC,$65,$1B,$E0,$F8,$3E
	.dc.b	$AB,$E9,$A9,$9A,$FF,$1C,$4E,$1F,$F6,$8C,$ED,$C9,$FE,$0B,$F8,$0F
	.dc.b	$21,$1D,$D5,$7F,$C0,$31,$73,$FF,$06,$57,$39,$FF,$37,$AF,$89,$F2
	.dc.b	$92,$3C,$3F,$2D,$5D,$C5,$AE,$73,$EC,$F1,$E5,$7A,$1F,$7F,$D0,$E2
	.dc.b	$3B,$C5,$AF,$1C,$FE,$32,$5F,$4C,$BC,$36,$3F,$8D,$D7,$F9,$21,$DC
	.dc.b	$FC,$CF,$8D,$DB,$23,$7A,$FF,$22,$76,$7C,$C9,$D3,$F0,$BE,$2B,$EA
	.dc.b	$98,$BF,$8F,$CB,$F6,$09,$3F,$53,$E9,$AB,$F2,$A2,$CF,$CE,$9F,$C5
	.dc.b	$7C,$F3,$96,$92,$FE,$EF,$C6,$57,$FC,$9D,$3B,$7D,$27,$FB,$59,$C3
	.dc.b	$F6,$23,$E4,$7D,$13,$E2,$29,$27,$FD,$33,$F4,$CF,$A4,$C4,$7F,$A8
	.dc.b	$EB,$83,$D8,$38,$7F,$D8,$1E,$F2,$B9,$46,$E7,$5A,$7D,$35,$F0,$DE
	.dc.b	$43,$BC,$7F,$9F,$86,$6F,$A6,$F8,$A4,$BF,$F6,$27,$A7,$5F,$F6,$1F
	.dc.b	$C1,$3D,$1F,$E1,$2D,$8F,$8B,$87,$4B,$CF,$74,$9F,$F6,$73,$C3,$C3
	.dc.b	$C4,$74,$8B,$5F,$E6,$FB,$FE,$37,$3D,$32,$7D,$37,$DA,$24,$7F,$B6
	.dc.b	$7A,$2B,$F6,$22,$DF,$39,$A7,$F7,$2F,$15,$F4,$7F,$B1,$9F,$F3,$F3
	.dc.b	$67,$F3,$B8,$78,$30,$F1,$7C,$B7,$E2,$88,$FE,$13,$0E,$7D,$98,$8F
	.dc.b	$C4,$CD,$C7,$AC,$FF,$30,$C5,$DD,$49,$7C,$E3,$F8,$13,$63,$2F,$FB
	.dc.b	$D2,$38,$BB,$F0,$3F,$23,$F4,$D2,$3F,$08,$FE,$53,$A3,$11,$F8,$F8
	.dc.b	$B8,$FE,$1E,$6E,$BF,$B5,$E6,$7D,$37,$E7,$0C,$E1,$FE,$C5,$C9,$E1
	.dc.b	$7F,$52,$EA,$78,$8F,$A6,$C3,$6F,$D9,$5F,$B2,$F5,$24,$7D,$D7,$75
	.dc.b	$F1,$7C,$39,$3F,$0D,$F9,$1E,$27,$F4,$3F,$A6,$7C,$7D,$25,$7C,$9F
	.dc.b	$17,$5E,$44,$8F,$CA,$D2,$37,$1F,$2F,$C7,$FA,$84,$77,$1D,$FD,$A9
	.dc.b	$1C,$9F,$B5,$E3,$E2,$3F,$DA,$96,$9C,$73,$7E,$32,$AF,$2F,$E4,$78
	.dc.b	$FF,$AC,$E1,$F7,$FC,$BC,$AB,$63,$FE,$C4,$8F,$4F,$8B,$F0,$4F,$A1
	.dc.b	$F4,$DF,$CC,$E3,$C3,$19,$76,$7F,$18,$E6,$4E,$EF,$F8,$15,$F5,$AF
	.dc.b	$C6,$7C,$BF,$E5,$4E,$4F,$FB,$BF,$E9,$78,$69,$FD,$01,$B1,$E7,$6E
	.dc.b	$2F,$A6,$72,$7F,$79,$9D,$E4,$69,$C2,$77,$7F,$C2,$A3,$1D,$DF,$8D
	.dc.b	$9F,$D1,$CF,$8F,$C0,$E9,$CF,$23,$83,$F9,$3F,$98,$FF,$31,$E3,$FC
	.dc.b	$7F,$14,$EE,$BD,$8E,$FD,$9C,$53,$5F,$F3,$BE,$AC,$7F,$CF,$D8,$7C
	.dc.b	$B6,$BC,$8F,$47,$17,$BC,$8F,$F3,$BF,$43,$D3,$87,$8E,$73,$3F,$35
	.dc.b	$FE,$DB,$B6,$27,$F8,$9F,$CC,$FF,$07,$31,$97,$F6,$9F,$A5,$7C,$A7
	.dc.b	$F7,$76,$E3,$FD,$25,$AF,$E2,$E7,$FB,$87,$D3,$3E,$67,$85,$CA,$E0
	.dc.b	$FF,$80,$67,$51,$FD,$4F,$E9,$7F,$38,$DE,$CF,$F0,$34,$62,$3B,$1F
	.dc.b	$E9,$EF,$A6,$7C,$1E,$AB,$FD,$E7,$3F,$E0,$1A,$E0,$FF,$BD,$9A,$E6
	.dc.b	$36,$FE,$37,$87,$F1,$44,$7F,$B9,$FB,$EF,$A6,$FD,$E4,$7E,$26,$74
	.dc.b	$FE,$65,$FF,$32,$46,$A3,$CF,$2F,$A6,$7C,$7F,$F9,$A5,$B3,$D9,$AF
	.dc.b	$F8,$DB,$93,$53,$23,$4F,$F9,$32,$31,$1D,$A7,$8F,$1F,$E3,$61,$7E
	.dc.b	$1B,$E1,$1D,$FE,$3B,$C4,$47,$F7,$5F,$4C,$F8,$CC,$47,$59,$B8,$75
	.dc.b	$E9,$5E,$EB,$FC,$19,$F4,$CF,$13,$EA,$C7,$D4,$8C,$71,$7D,$06,$C6
	.dc.b	$67,$6F,$F8,$E4,$FF,$6B,$7A,$8F,$A6,$FA,$FF,$16,$8C,$4F,$E0,$23
	.dc.b	$13,$A9,$CC,$F3,$BF,$F0,$67,$C6,$F8,$7D,$F3,$49,$3E,$1D,$3F,$AA
	.dc.b	$FC,$EF,$F8,$6F,$B7,$FB,$E7,$9A,$F2,$71,$6B,$DD,$7F,$9E,$A7,$31
	.dc.b	$89,$FF,$38,$C6,$3D,$56,$1C,$BD,$AF,$E9,$13,$71,$FE,$09,$8F,$FA
	.dc.b	$13,$C3,$C7,$66,$5B,$19,$3F,$D1,$31,$96,$99,$CB,$FD,$03,$19,$7F
	.dc.b	$56,$3E,$97,$FD,$09,$3F,$E2,$EF,$A7,$FE,$C2,$FF,$91,$E3,$33,$FB
	.dc.b	$2E,$75,$6A,$39,$CE,$9F,$B4,$D7,$FC,$E8,$D4,$78,$7C,$6F,$F3,$F3
	.dc.b	$F3,$BD,$8F,$9B,$DA,$F4,$BE,$1C,$FF,$A6,$A7,$31,$8C,$7F,$D3,$F1
	.dc.b	$8C,$FE,$53,$73,$FF,$30,$7C,$8B,$E6,$BE,$77,$C8,$7D,$AC,$93,$9A
	.dc.b	$FF,$59,$CE,$63,$13,$FE,$AA,$8C,$57,$2A,$FF,$2A,$E9,$FE,$81,$7F
	.dc.b	$D7,$7E,$4F,$E1,$B3,$FF,$58,$7C,$4F,$DD,$8C,$56,$F1,$FF,$5A,$C6
	.dc.b	$27,$57,$D1,$B3,$F4,$6B,$85,$FF,$AF,$AF,$9F,$B5,$F8,$9E,$6B,$8F
	.dc.b	$FB,$31,$B8,$7F,$B0,$9B,$7F,$EC,$67,$31,$FE,$A9,$8F,$F6,$35,$7D
	.dc.b	$48,$FF,$04,$61,$FE,$31,$BF,$E5,$F5,$A6,$C3,$E3,$3F,$35,$F3,$BE
	.dc.b	$E6,$CF,$C6,$6D,$FA,$DB,$7F,$E8,$9B,$E5,$5B,$3F,$DC,$BD,$7C,$DF
	.dc.b	$9F,$F1,$7F,$A7,$C3,$E5,$B6,$32,$7F,$BB,$23,$2D,$32,$FC,$B8,$FD
	.dc.b	$97,$E6,$0F,$AC,$DC,$FC,$FF,$17,$3C,$1F,$61,$B8,$7E,$01,$B3,$B8
	.dc.b	$FF,$3C,$5F,$D6,$7F,$DB,$79,$7F,$58,$78,$1A,$FA,$FE,$8F,$0F,$AC
	.dc.b	$F1,$9F,$89,$FD,$DD,$F4,$38,$9F,$F5,$1C,$61,$E0,$62,$35,$8F,$F6
	.dc.b	$8F,$05,$97,$FA,$CE,$B4,$79,$BE,$3F,$B5,$FC,$2F,$36,$4F,$AB,$96
	.dc.b	$E3,$FC,$07,$2F,$EB,$73,$93,$C7,$E7,$7C,$AA,$F9,$7F,$2B,$27,$E1
	.dc.b	$F6,$5C,$3F,$88,$F3,$3E,$63,$E8,$FC,$77,$EC,$E3,$17,$F2,$A3,$13
	.dc.b	$A8,$FF,$6D,$4E,$AB,$DD,$BF,$FA,$EA,$73,$5B,$9E,$ED,$AF,$EC,$DC
	.dc.b	$1F,$13,$CC,$FC,$0F,$57,$E0,$9B,$87,$DA,$6E,$1D,$9B,$7C,$B1,$7B
	.dc.b	$9F,$8A,$CB,$F1,$6D,$C3,$F0,$6D,$C3,$CE,$DB,$E5,$F2,$5F,$92,$EB
	.dc.b	$FC,$9B,$6F,$F1,$6D,$BF,$7B,$6B,$8C,$FE,$42,$B8,$B4,$FF,$74,$FE
	.dc.b	$EE,$F9,$5F,$CF,$FC,$77,$F1,$58,$C4,$FE,$62,$31,$3A,$8E,$2D,$C7
	.dc.b	$E7,$F9,$DF,$A8,$F8,$FF,$D4,$B6,$BF,$42,$D9,$DC,$7E,$92,$7F,$5F
	.dc.b	$DD,$FB,$0F,$8D,$B3,$FA,$86,$D7,$DC,$6E,$9F,$3D,$B3,$FE,$EF,$E0
	.dc.b	$46,$29,$C7,$FA,$08,$FE,$DD,$F3,$7F,$E5,$5B,$FF,$76,$2F,$A9,$CF
	.dc.b	$C1,$69,$B3,$A6,$E1,$FA,$BA,$E1,$3C,$BE,$A3,$AF,$C8,$76,$6C,$64
	.dc.b	$EF,$19,$69,$E2,$F0,$3C,$BE,$A7,$99,$B1,$93,$CF,$19,$69,$9C,$3C
	.dc.b	$91,$96,$C7,$BB,$CC,$F8,$3F,$49,$F0,$9B,$19,$3E,$1C,$65,$A7,$C6
	.dc.b	$FA,$2F,$93,$F4,$7F,$29,$B1,$93,$E5,$C6,$5A,$7C,$EF,$84,$FE,$F7
	.dc.b	$F3,$DF,$E4,$D8,$C1,$FE,$6E,$71,$8F,$CD,$9D,$F8,$FF,$61,$B5,$F6
	.dc.b	$3E,$0B,$ED,$9F,$73,$C7,$B1,$92,$5F,$D7,$8C,$BF,$67,$E8,$7F,$07
	.dc.b	$F4,$7F,$84,$6D,$7E,$11,$B3,$B8,$FC,$3C,$FE,$2B,$DF,$FC,$5F,$D1
	.dc.b	$D9,$FC,$5B,$6B,$F0,$CD,$EB,$FE,$1F,$CC,$8C,$57,$F8,$8B,$A9,$2C
	.dc.b	$3F,$2F,$F0,$9A,$C7,$F3,$73,$9B,$FC,$CD,$69,$A7,$E7,$3E,$77,$F3
	.dc.b	$FF,$3D,$FA,$06,$C6,$4F,$D0,$C6,$5A,$7F,$8D,$33,$FF,$00,$7D,$1F
	.dc.b	$F8,$DB,$E8,$DE,$E7,$F9,$D7,$D9,$FF,$47,$46,$34,$FF,$3C,$56,$EB
	.dc.b	$58,$59,$C8,$AE,$0F,$F9,$0A,$F9,$D7,$FA,$EA,$F8,$4E,$6E,$BF,$E4
	.dc.b	$88,$B5,$C2,$3A,$DF,$47,$FC,$FB,$5C,$15,$70,$8C,$36,$3C,$97,$E1
	.dc.b	$69,$FE,$56,$39,$98,$31,$1C,$6B,$C1,$1D,$8D,$1C,$1C,$C6,$E7,$B4
	.dc.b	$79,$EF,$D5,$5D,$0E,$2E,$1F,$80,$62,$FD,$2D,$3B,$9E,$19,$E9,$1A
	.dc.b	$6C,$6A,$FC,$4F,$C8,$BF,$AE,$E4,$E9,$3E,$86,$C7,$A0,$DC,$64,$EA
	.dc.b	$6F,$C3,$FE,$67,$3F,$80,$6B,$ED,$6B,$EC,$73,$F1,$1F,$E0,$3C,$BF
	.dc.b	$D0,$B1,$8F,$35,$18,$F1,$3E,$A5,$6E,$D7,$BA,$D4,$7F,$A3,$A7,$58
	.dc.b	$7E,$41,$FE,$EF,$1F,$8E,$6B,$F7,$5D,$E3,$DB,$1F,$0C,$E7,$CD,$FC
	.dc.b	$25,$F3,$6D,$6F,$0E,$97,$C2,$B5,$7D,$EF,$6F,$96,$75,$7C,$63,$B1
	.dc.b	$B3,$9D,$C7,$24,$BF,$65,$72,$8E,$93,$C5,$D5,$75,$AD,$CE,$5F,$2C
	.dc.b	$FC,$1A,$E4,$B0,$F8,$6F,$CB,$C3,$D2,$E5,$B1,$A8,$E1,$3C,$A3,$E6
	.dc.b	$D7,$63,$C3,$3D,$EF,$C5,$87,$DC,$C3,$9E,$1C,$4E,$4F,$63,$33,$F0
	.dc.b	$EF,$A5,$71,$BC,$3F,$CB,$7E,$8F,$FC,$41,$1F,$E0,$0D,$7F,$D3,$C7
	.dc.b	$F3,$1E,$23,$88,$FF,$04,$CF,$07,$FC,$D4,$DC,$3F,$9D,$CE,$F9,$3F
	.dc.b	$9C,$B6,$74,$F8,$7E,$7F,$04,$BF,$E5,$D1,$E5,$F9,$6C,$47,$69,$B8
	.dc.b	$7F,$95,$AD,$58,$BF,$67,$B7,$DC,$C6,$1E,$8C,$30,$F6,$3D,$71,$AC
	.dc.b	$58,$3E,$63,$D1,$F3,$D2,$F8,$56,$A7,$D8,$66,$3E,$1C,$E1,$DD,$3E
	.dc.b	$7B,$E1,$12,$33,$F3,$15,$F9,$71,$75,$F7,$4F,$C9,$F9,$EF,$A5,$F1
	.dc.b	$2F,$17,$FD,$99,$87,$F8,$1F,$D9,$D3,$E0,$AD,$CE,$A3,$F8,$3F,$01
	.dc.b	$5F,$3B,$E3,$1F,$73,$E4,$FE,$C7,$C5,$48,$F5,$C7,$DB,$30,$FE,$0A
	.dc.b	$30,$FC,$0F,$8D,$FC,$07,$C7,$48,$C1,$0F,$C4,$B9,$7F,$17,$1F,$87
	.dc.b	$F8,$BF,$C4,$7C,$64,$9F,$97,$19,$7E,$6C,$65,$FC,$9C,$65,$FC,$AC
	.dc.b	$67,$7F,$C4,$4F,$E4,$FE,$33,$F2,$9D,$52,$FE,$BC,$65,$FB,$31,$97
	.dc.b	$F3,$F1,$F9,$BF,$A6,$FC,$E7,$C6,$C9,$FB,$F1,$97,$F0,$31,$97,$F4
	.dc.b	$F1,$FA,$3F,$13,$F4,$9F,$1B,$27,$F0,$F1,$97,$F1,$31,$97,$F5,$F1
	.dc.b	$97,$F6,$31,$9C,$BF,$49,$3F,$B1,$F8,$9F,$F7,$C5,$7E,$48,$FD,$91
	.dc.b	$9F,$8D,$8E,$48,$7F,$B9,$63,$8B,$F8,$83,$FA,$7F,$C1,$3A,$3B,$7C
	.dc.b	$6C,$9F,$F5,$5C,$7F,$07,$C5,$5B,$9F,$D4,$DA,$F8,$47,$5A,$DD,$FA
	.dc.b	$7E,$33,$D5,$F1,$B2,$B8,$C6,$5E,$71,$97,$D1,$19,$7D,$51,$9A,$FF
	.dc.b	$6E,$E0,$AF,$8D,$EB,$7C,$78,$97,$E6,$8C,$BE,$88,$CB,$F1,$63,$E7
	.dc.b	$7C,$57,$CF,$F8,$C9,$3E,$E8,$CB,$F0,$23,$2F,$CD,$8F,$E5,$8F,$F8
	.dc.b	$D7,$0F,$2F,$C6,$7D,$93,$3C,$9F,$E2,$F9,$FA,$46,$FB,$7F,$CC,$D9
	.dc.b	$2B,$84,$FF,$1D,$34,$7E,$FE,$B9,$5F,$68,$FA,$73,$D9,$FC,$47,$64
	.dc.b	$7E,$BB,$F8,$BF,$8C,$FB,$9C,$2E,$7F,$C7,$7C,$64,$BF,$F7,$5C,$7E
	.dc.b	$47,$CD,$91,$F8,$07,$4F,$E3,$7C,$DF,$C3,$7E,$ED,$F8,$ED,$3F,$A7
	.dc.b	$CF,$F8,$72,$31,$BF,$BA,$B7,$CF,$F9,$DF,$8C,$97,$FC,$EF,$06,$1C
	.dc.b	$23,$FD,$BF,$C2,$AD,$C7,$E7,$5E,$6F,$19,$71,$F9,$09,$1F,$98,$F8
	.dc.b	$DA,$FE,$97,$E3,$7F,$5C,$63,$1F,$49,$E2,$AC,$CF,$E8,$27,$5A,$7F
	.dc.b	$A2,$F0,$6B,$F0,$89,$5F,$A6,$F8,$BA,$FE,$CF,$E3,$3F,$78,$63,$2E
	.dc.b	$F7,$CA,$7F,$5B,$3A,$7D,$97,$F3,$AE,$47,$BD,$FE,$0D,$EB,$62,$3F
	.dc.b	$80,$7A,$DF,$C2,$4C,$63,$DA,$B1,$3F,$BA,$9C,$E5,$FE,$A5,$F8,$87
	.dc.b	$C5,$FE,$5E,$B0,$D7,$FC,$AF,$1A,$75,$82,$31,$1F,$96,$6C,$7E,$13
	.dc.b	$D2,$7E,$F1,$8C,$3F,$D8,$CD,$BE,$B8,$7F,$11,$C3,$17,$C2,$3F,$C2
	.dc.b	$95,$BC,$3C,$11,$BF,$89,$CD,$CC,$3F,$18,$A3,$51,$FE,$02,$BF,$43
	.dc.b	$67,$2F,$F1,$B7,$83,$92,$19,$9F,$E0,$3D,$DA,$8F,$E5,$D8,$FB,$BE
	.dc.b	$2F,$CE,$6A,$A4,$F8,$63,$2F,$EF,$EB,$96,$5F,$1F,$1B,$9F,$F7,$FF
	.dc.b	$3D,$F8,$57,$E9,$5B,$D3,$F1,$33,$99,$7F,$89,$67,$2D,$AF,$D6,$69
	.dc.b	$F4,$AB,$F8,$CF,$CC,$BE,$23,$EF,$5A,$30,$FE,$05,$AF,$E1,$5A,$FF
	.dc.b	$1D,$73,$A7,$F0,$28,$7C,$D3,$9B,$F9,$3B,$63,$F8,$F5,$70,$D7,$F0
	.dc.b	$7F,$19,$F7,$BE,$3A,$57,$D7,$8F,$C0,$98,$7E,$7F,$02,$73,$1B,$3E
	.dc.b	$F3,$BC,$BE,$65,$7F,$98,$FE,$55,$F3,$0F,$2F,$F1,$BF,$21,$31,$FF
	.dc.b	$27,$56,$A7,$85,$6E,$7F,$15,$5B,$7F,$B7,$56,$DF,$E7,$5F,$2E,$F8
	.dc.b	$AF,$F5,$DC,$7E,$B6,$31,$D7,$FA,$D6,$C6,$67,$6F,$F4,$9F,$86,$79
	.dc.b	$0F,$98,$FD,$A4,$61,$B1,$99,$FD,$D4,$E5,$F0,$7C,$5B,$D4,$F0,$36
	.dc.b	$33,$18,$3F,$81,$67,$F6,$7D,$7F,$E4,$BC,$9F,$E5,$1E,$AF,$D6,$7B
	.dc.b	$78,$C1,$7B,$AF,$F3,$7F,$5A,$B5,$39,$AF,$D3,$CE,$6B,$E7,$F5,$31
	.dc.b	$E9,$8F,$69,$E7,$ED,$62,$DE,$DD,$8C,$36,$7C,$2D,$C7,$BB,$7A,$3F
	.dc.b	$8A,$5F,$2A,$EA,$70,$3F,$B9,$7C,$BB,$E1,$7F,$39,$87,$F9,$C4,$46
	.dc.b	$67,$F9,$33,$63,$F9,$53,$5F,$F0,$13,$9B,$FD,$33,$67,$F7,$DF,$61
	.dc.b	$FA,$77,$C3,$7E,$FF,$7B,$E7,$78,$6F,$7D,$AE,$6F,$F8,$04,$61,$FA
	.dc.b	$35,$C2,$31,$9F,$D0,$8C,$57,$D9,$87,$8C,$EC,$F9,$DE,$CE,$63,$18
	.dc.b	$FD,$C8,$C4,$EA,$3F,$8D,$36,$BF,$C8,$99,$FF,$91,$E7,$F0,$7E,$AF
	.dc.b	$E1,$F7,$F9,$9E,$EF,$DB,$8C,$56,$EF,$F0,$B1,$89,$D5,$FF,$81,$EB
	.dc.b	$57,$F6,$AB,$53,$D8,$D7,$A5,$F8,$7E,$5E,$AD,$8C,$9F,$96,$8C,$B4
	.dc.b	$CE,$3E,$A9,$FC,$87,$BF,$A8,$F6,$FB,$5F,$93,$F7,$FD,$2D,$BF,$33
	.dc.b	$67,$71,$FA,$49,$DB,$A3,$F4,$FD,$98,$9F,$93,$EB,$FE,$7D,$F8,$AD
	.dc.b	$9F,$84,$D8,$DC,$7E,$C1,$DB,$A3,$FD,$C3,$F2,$CF,$3F,$33,$F5,$BC
	.dc.b	$6F,$D7,$CF,$CC,$AF,$F6,$76,$FF,$A1,$9C,$F0,$FD,$66,$CF,$D3,$6D
	.dc.b	$F3,$8C,$1E,$0F,$23,$83,$E2,$F7,$FB,$9F,$83,$F9,$8F,$E7,$51,$93
	.dc.b	$FC,$88,$E2,$BF,$B6,$E5,$FE,$81,$9C,$F9,$9C,$9F,$6F,$B7,$CE,$3D
	.dc.b	$6D,$8F,$83,$F0,$1E,$E3,$DF,$C9,$63,$BF,$C9,$E1,$A7,$C8,$8F,$8B
	.dc.b	$F0,$32,$3E,$5F,$65,$CF,$FC,$FB,$C3,$FE,$44,$9F,$9F,$F0,$3F,$AD
	.dc.b	$8C,$4F,$AA,$31,$EE,$7D,$06,$D7,$03,$FD,$05,$5C,$0F,$F1,$56,$C9
	.dc.b	$DD,$F3,$AF,$BB,$E6,$FD,$AE,$9F,$B6,$D8,$C9,$F8,$28,$CB,$4F,$C2
	.dc.b	$FC,$07,$E0,$B2,$FC,$1B,$63,$27,$E2,$A3,$2D,$3F,$1B,$F0,$1F,$8A
	.dc.b	$E9,$FC,$5B,$63,$27,$E4,$A3,$2D,$33,$8F,$DE,$8C,$B6,$3F,$D0,$DF
	.dc.b	$02,$F9,$8F,$CB,$FC,$17,$F0,$B8,$C4,$FE,$7A,$31,$3A,$8E,$2D,$AF
	.dc.b	$C5,$F9,$DF,$A0,$F8,$0F,$D0,$B6,$BF,$4E,$D9,$DC,$7E,$AA,$7F,$4F
	.dc.b	$D9,$FA,$8F,$81,$B3,$FA,$E6,$D7,$E6,$1B,$C7,$F6,$9B,$3F,$DA,$F8
	.dc.b	$D1,$8A,$7A,$09,$CD,$7E,$D0,$FE,$E5,$F1,$BF,$E7,$87,$15,$C2,$1F
	.dc.b	$31,$FD,$A9,$D4,$7F,$95,$9C,$CF,$0B,$46,$B1,$FF,$3C,$3A,$AE,$1F
	.dc.b	$07,$CF,$0E,$9F,$05,$D7,$E0,$EE,$5C,$F0,$FF,$42,$36,$74,$D3,$53
	.dc.b	$E5,$F0,$3C,$DE,$A5,$9E,$4D,$CB,$D0,$DC,$FC,$AD,$9F,$86,$D9,$F8
	.dc.b	$4F,$C1,$F3,$7E,$17,$BD,$6B,$CC,$DA,$F8,$8D,$BF,$82,$FC,$9F,$98
	.dc.b	$F9,$5F,$31,$6F,$DC,$DB,$F9,$8D,$BF,$92,$FD,$1F,$84,$FE,$F3,$F1
	.dc.b	$2F,$1B,$18,$7F,$9D,$E3,$0F,$C9,$9D,$CE,$B0,$61,$FE,$EE,$F5,$D8
	.dc.b	$70,$9F,$B9,$F3,$1F,$7C,$FC,$07,$91,$72,$FA,$91,$67,$F0,$53,$70
	.dc.b	$FB,$FF,$31,$F8,$7F,$98,$FC,$43,$70,$FB,$8D,$BF,$C5,$36,$BF,$0F
	.dc.b	$F0,$1F,$8F,$F9,$8F,$C8,$36,$FF,$08,$DB,$FC,$93,$6B,$8C,$E8,$FC
	.dc.b	$57,$C2,$72,$71,$5C,$1C,$98,$9F,$CB,$FC,$C7,$E7,$FE,$63,$F4,$0D
	.dc.b	$9F,$CA,$36,$BF,$44,$DA,$FF,$1E,$C3,$E6,$0B,$E6,$1E,$F7,$F9,$06
	.dc.b	$3F,$CD,$58,$6A,$73,$19,$AC,$56,$FD,$95,$E2,$B4,$75,$9D,$46,$63
	.dc.b	$FC,$9D,$55,$EE,$6C,$E3,$7E,$07,$CF,$7D,$9D,$1C,$8A,$72,$8E,$0F
	.dc.b	$BA,$7B,$4F,$19,$CC,$FC,$33,$8C,$66,$FE,$14,$66,$79,$CE,$63,$CB
	.dc.b	$3E,$C7,$C2,$6D,$C4,$F6,$AF,$0D,$FB,$EB,$47,$F9,$BE,$FA,$3E,$BA
	.dc.b	$EF,$3B,$AE,$33,$D9,$B3,$D6,$3D,$33,$B6,$9B,$2C,$6C,$EA,$E5,$A6
	.dc.b	$6F,$D2,$7C,$2A,$EE,$D9,$E6,$DF,$13,$F9,$EC,$7E,$01,$AF,$D5,$6B
	.dc.b	$F4,$C3,$AF,$87,$79,$BF,$C0,$A7,$3E,$37,$FA,$4E,$73,$A7,$E6,$2F
	.dc.b	$85,$ED,$B1,$9A,$E1,$3C,$F2,$FC,$55,$BC,$EF,$F4,$F4,$61,$B9,$70
	.dc.b	$C3,$F2,$D5,$E0,$CB,$F1,$EF,$17,$35,$C1,$FD,$5C,$6E,$FB,$E1,$E1
	.dc.b	$C3,$D5,$7B,$B5,$F3,$7A,$D6,$DE,$6D,$77,$3C,$67,$85,$F6,$C3,$CF
	.dc.b	$7B,$9C,$E1,$E0,$C3,$35,$1F,$5E,$1C,$CE,$97,$C4,$E4,$E6,$38,$61
	.dc.b	$DD,$E2,$7C,$23,$37,$F2,$5C,$92,$38,$12,$3C,$91,$97,$84,$66,$3E
	.dc.b	$5C,$F8,$5F,$8B,$7D,$9C,$9E,$13,$27,$F0,$9F,$19,$F2,$9F,$E2,$B8
	.dc.b	$FE,$F1,$B7,$FB,$76,$BF,$C0,$2D,$3F,$E0,$04,$EE,$7F,$A9,$C6,$3C
	.dc.b	$BD,$47,$F8,$46,$2D,$FF,$83,$1B,$5C,$6F,$F8,$B4,$F8,$BE,$5F,$24
	.dc.b	$BF,$F0,$A5,$3E,$63,$11,$E0,$3F,$C4,$B0,$9C,$DF,$48,$E1,$3C,$A7
	.dc.b	$C3,$F0,$52,$30,$43,$A6,$1F,$12,$35,$18,$AE,$CF,$93,$E6,$3D,$7F
	.dc.b	$31,$23,$04,$38,$39,$79,$47,$C2,$F9,$89,$1E,$E7,$E2,$FC,$B6,$23
	.dc.b	$C1,$17,$3F,$64,$FC,$AF,$98,$F9,$FE,$1F,$F8,$0B,$D8,$BE,$21,$D6
	.dc.b	$3F,$2A,$B3,$12,$30,$7C,$EF,$98,$FA,$BF,$17,$23,$04,$3E,$13,$97
	.dc.b	$E2,$47,$DC,$F8,$EF,$BB,$F2,$12,$7E,$CC,$65,$FB,$71,$97,$E6,$47
	.dc.b	$E1,$3E,$37,$F0,$BF,$1B,$27,$F0,$31,$97,$F0,$71,$97,$EA,$46,$5F
	.dc.b	$AD,$19,$DB,$F0,$53,$F9,$0F,$8D,$FC,$8F,$64,$BF,$C6,$C6,$5F,$C7
	.dc.b	$C6,$5F,$C0,$47,$E6,$3E,$33,$F3,$3F,$1B,$27,$F2,$B1,$97,$F2,$F1
	.dc.b	$97,$F1,$11,$FA,$0F,$13,$F4,$3F,$1B,$27,$F3,$B1,$97,$F3,$F1,$97
	.dc.b	$F2,$11,$97,$F2,$51,$9C,$BF,$3D,$3F,$AE,$F8,$8F,$F7,$E5,$7E,$A0
	.dc.b	$FD,$79,$9F,$8D,$8E,$48,$7E,$82,$38,$BF,$AD,$A7,$C1,$39,$FB,$7C
	.dc.b	$6C,$AF,$F5,$9C,$7F,$00,$AD,$B9,$8E,$93,$9A,$E5,$1F,$E8,$FB,$7C
	.dc.b	$67,$A7,$E3,$64,$E6,$3D,$86,$1F,$34,$61,$D1,$CD,$D6,$3F,$C3,$2B
	.dc.b	$AC,$F7,$F5,$7E,$34,$48,$C1,$0F,$84,$E5,$F8,$91,$F3,$7E,$2B,$E7
	.dc.b	$7C,$64,$9F,$54,$65,$F6,$46,$5F,$99,$1F,$C9,$1F,$F1,$7D,$F9,$7E
	.dc.b	$33,$EC,$19,$E4,$FF,$26,$47,$F8,$E6,$77,$E3,$FF,$94,$BC,$56,$6C
	.dc.b	$38,$D7,$FB,$00,$D9,$FE,$58,$C3,$9D,$F1,$AE,$13,$F4,$6B,$C0,$FE
	.dc.b	$27,$B6,$3F,$55,$FC,$67,$C6,$7D,$9E,$8F,$C7,$7C,$64,$AF,$F7,$9C
	.dc.b	$7E,$47,$CD,$91,$F7,$5D,$3F,$8D,$F3,$3F,$0D,$F3,$1F,$8E,$DF,$FC
	.dc.b	$AC,$1C,$6B,$E7,$9D,$F4,$7E,$77,$E3,$26,$1F,$ED,$EC,$38,$4E,$23
	.dc.b	$F3,$53,$67,$AE,$1D,$27,$FD,$52,$B8,$FC,$64,$8F,$CA,$7C,$6D,$7F
	.dc.b	$49,$F1,$BF,$AD,$31,$7D,$2B,$13,$F9,$B9,$CE,$7F,$E8,$8A,$F6,$69
	.dc.b	$F0,$09,$5F,$A1,$F8,$CA,$FE,$C7,$E3,$7F,$74,$62,$BD,$35,$89,$FD
	.dc.b	$44,$E6,$FB,$66,$C3,$E6,$92,$BF,$5D,$EA,$D7,$F7,$DE,$B7,$F0,$73
	.dc.b	$19,$7C,$6A,$C4,$FE,$D6,$73,$87,$AF,$E2,$1F,$19,$FB,$F6,$D7,$D3
	.dc.b	$6B,$FD,$AB,$E5,$5F,$1B,$BA,$FF,$01,$C6,$B2,$FF,$00,$B7,$0E,$D3
	.dc.b	$BB,$F9,$51,$FA,$19,$D4,$3B,$11,$8B,$63,$47,$C5,$78,$19,$7E,$78
	.dc.b	$F5,$18,$7D,$11,$87,$47,$88,$F0,$57,$67,$17,$E2,$9E,$11,$FE,$15
	.dc.b	$1D,$DE,$27,$74,$F8,$CF,$86,$7F,$05,$9D,$4F,$16,$E3,$F1,$A7,$73
	.dc.b	$FB,$F9,$CE,$5F,$17,$2F,$F0,$1F,$CC,$7E,$7F,$D3,$3F,$A1,$D3,$E3
	.dc.b	$CF,$D2,$37,$D3,$31,$F2,$C7,$F1,$7F,$97,$7C,$4D,$24,$FF,$0F,$F9
	.dc.b	$8F,$F4,$AB,$FC,$46,$76,$FF,$B1,$03,$87,$FD,$8E,$D7,$FD,$D4,$D7
	.dc.b	$EF,$7C,$6F,$E1,$BE,$3E,$47,$C8,$8F,$C4,$98,$7F,$1D,$18,$74,$7C
	.dc.b	$E7,$5C,$1F,$8E,$9F,$F2,$AF,$CA,$BE,$61,$E5,$7E,$53,$E4,$66,$3F
	.dc.b	$7E,$B5,$3C,$2B,$73,$F6,$AB,$6F,$F9,$4E,$B6,$FF,$33,$F9,$87,$C5
	.dc.b	$7C,$68,$10,$46,$BA,$FE,$43,$63,$FD,$DA,$D7,$81,$FE,$A4,$F1,$DF
	.dc.b	$31,$F3,$FE,$21,$18,$9F,$F7,$AC,$62,$75,$39,$9E,$74,$F8,$C7,$AB
	.dc.b	$E2,$6C,$F8,$9A,$F0,$6B,$EB,$FC,$35,$F2,$9F,$CF,$3B,$3D,$F6,$F8
	.dc.b	$2F,$E2,$D1,$FD,$5E,$B6,$FE,$B6,$75,$E2,$71,$C3,$E3,$D6,$9F,$82
	.dc.b	$1C,$3F,$11,$AF,$8F,$0F,$ED,$4F,$CF,$F9,$97,$C4,$75,$E7,$7C,$D7
	.dc.b	$F0,$8C,$FF,$B5,$56,$9F,$E9,$D1,$9B,$FD,$E4,$E5,$FF,$13,$46,$6B
	.dc.b	$FB,$C4,$E5,$FE,$BB,$19,$AF,$15,$F9,$2F,$F0,$77,$CD,$FB,$AD,$77
	.dc.b	$3F,$03,$BF,$F8,$FF,$CA,$7C,$56,$C7,$F0,$CE,$EC,$78,$AA,$A4,$9D
	.dc.b	$1F,$C4,$9D,$FC,$A7,$0F,$95,$B3,$FC,$59,$B8,$FF,$3B,$6E,$BF,$CE
	.dc.b	$1F,$AD,$1F,$C7,$9F,$0F,$CB,$63,$C7,$46,$2F,$F8,$C3,$6B,$F8,$F3
	.dc.b	$6F,$FC,$41,$4F,$94,$F7,$7C,$67,$F3,$AC,$3F,$90,$46,$3C,$6E,$CD
	.dc.b	$8C,$9F,$CD,$27,$86,$3F,$CA,$E7,$E8,$3F,$CD,$1B,$7D,$EB,$F8,$C3
	.dc.b	$67,$F5,$FE,$4B,$F4,$2F,$DD,$D7,$F3,$AF,$03,$F2,$D3,$97,$E8,$98
	.dc.b	$9D,$D7,$F6,$38,$C1,$AA,$FE,$99,$5C,$5B,$5C,$33,$FE,$CB,$D7,$C2
	.dc.b	$77,$F0,$4F,$92,$F8,$BE,$27,$8A,$31,$3F,$06,$31,$3A,$8E,$2D,$CF
	.dc.b	$FD,$D5,$97,$F9,$F6,$37,$3F,$27,$86,$D6,$B0,$6F,$F8,$ED,$FC,$0D
	.dc.b	$BF,$2B,$6B,$83,$74,$4F,$F2,$2C,$BF,$D1,$93,$CA,$3F,$B2,$FC,$AD
	.dc.b	$BF,$07,$C6,$7E,$0B,$8F,$DC,$DA,$F8,$6D,$9D,$C7,$E0,$67,$6E,$8F
	.dc.b	$C2,$72,$62,$7F,$C2,$D1,$FA,$83,$E5,$7C,$0D,$7E,$7B,$63,$E1,$37
	.dc.b	$3F,$5C,$7F,$84,$FF,$57,$FA,$2E,$AB,$FC,$0B,$18,$9D,$E9,$FE,$AB
	.dc.b	$BF,$DD,$4B,$E5,$29,$FA,$0E,$3B,$87,$E0,$3C,$2C,$34,$FC,$F4,$7E
	.dc.b	$9B,$82,$47,$83,$BA,$E9,$F7,$BF,$68,$FF,$57,$CF,$93,$E5,$7F,$33
	.dc.b	$18,$96,$DE,$CE,$F6,$AB,$DF,$C3,$FA,$1B,$8F,$33,$F3,$78,$F3,$C3
	.dc.b	$95,$F1,$9C,$DF,$17,$E2,$F3,$7C,$2E,$6F,$86,$D8,$C9,$F2,$A3,$2D
	.dc.b	$3E,$6F,$C9,$7C,$AF,$93,$F9,$6D,$8C,$9F,$4A,$32,$D3,$EA,$F3,$FD
	.dc.b	$2F,$93,$FA,$6D,$8C,$9F,$6A,$32,$D3,$3B,$FC,$87,$FD,$6D,$9A,$77
	.dc.b	$39,$3F,$0C,$FD,$0F,$82,$FC,$0F,$C9,$DB,$FC,$3B,$6B,$87,$0F,$D2
	.dc.b	$8C,$6F,$F3,$1F,$C4,$FC,$8F,$E2,$BE,$4A,$E1,$F9,$06,$E1,$CB,$34
	.dc.b	$63,$0F,$AA,$FF,$3A,$97,$C8,$FF,$4D,$FD,$73,$EC,$47,$F9,$6C,$CE
	.dc.b	$1F,$E6,$08,$CD,$FF,$6E,$AE,$1C,$BC,$DB,$5F,$ED,$5B,$E2,$7B,$2F
	.dc.b	$8C,$7F,$87,$27,$73,$9E,$0F,$17,$1F,$8D,$B5,$E1,$6C,$EE,$3C,$93
	.dc.b	$FC,$D3,$F6,$8F,$D6,$47,$9F,$8B,$1D,$7E,$D6,$C6,$4F,$E6,$73,$C2
	.dc.b	$FC,$D3,$F1,$57,$CC,$FF,$8E,$BE,$65,$DD,$FE,$F5,$D7,$E6,$E3,$EF
	.dc.b	$BD,$8A,$DB,$FE,$99,$9D,$BF,$34,$DC,$6B,$91,$3A,$E0,$FF,$04,$63
	.dc.b	$E1,$C7,$C1,$C3,$E5,$8C,$63,$D6,$30,$F8,$67,$79,$79,$6B,$73,$FE
	.dc.b	$36,$8A,$FF,$95,$A2,$F7,$7F,$1E,$CF,$D3,$3F,$E5,$99,$CD,$70,$8F
	.dc.b	$4D,$FF,$95,$AF,$83,$FE,$38,$AD,$CF,$8C,$FF,$2D,$B8,$BE,$B1,$E1
	.dc.b	$34,$79,$4A,$65,$DD,$76,$8E,$66,$9C,$3E,$73,$17,$E7,$AF,$05,$F9
	.dc.b	$67,$27,$63,$D6,$7F,$9D,$EB,$9B,$E1,$8F,$75,$73,$8E,$37,$EA,$78
	.dc.b	$9E,$97,$17,$E0,$8F,$3C,$7A,$A7,$93,$63,$C5,$3B,$7D,$51,$93,$9B
	.dc.b	$E0,$AF,$54,$65,$A7,$9A,$39,$9B,$3C,$2E,$63,$99,$C6,$32,$7C,$1A
	.dc.b	$F2,$CE,$CE,$B1,$93,$AC,$F1,$7C,$F1,$B7,$99,$F0,$DE,$2D,$32,$7F
	.dc.b	$A4,$A7,$BB,$D6,$7C,$B3,$CE,$3A,$99,$79,$BC,$4F,$1B,$B7,$EC,$1B
	.dc.b	$3B,$DF,$DC,$7F,$6F,$F0,$BF,$44,$F9,$A5,$9E,$B3,$BC,$3E,$A5,$74
	.dc.b	$3D,$47,$90,$A6,$4F,$25,$FA,$9F,$11,$B3,$35,$DB,$0F,$83,$1C,$8F
	.dc.b	$C5,$9A,$3D,$2E,$9C,$1F,$9F,$F6,$18,$FE,$75,$B1,$FA,$66,$BB,$8F
	.dc.b	$F6,$A7,$B0,$F8,$8F,$F0,$1C,$AB,$73,$9D,$7F,$DB,$71,$8A,$DB,$63
	.dc.b	$37,$D2,$79,$E5,$FC,$2A,$DF,$19,$FE,$E3,$8C,$37,$3E,$18,$7F,$16
	.dc.b	$AF,$06,$5F,$C1,$1E,$2E,$6B,$83,$FE,$EE,$8D,$DF,$7C,$3C,$38,$7B
	.dc.b	$AF,$76,$BE,$6F,$5A,$DB,$CD,$AE,$E7,$8C,$F0,$BE,$D8,$79,$EF,$73
	.dc.b	$9C,$3C,$18,$66,$A3,$EB,$C3,$99,$D2,$F8,$9F,$4B,$1F,$8B,$23,$EE
	.dc.b	$DF,$46,$9E,$7B,$D4,$F9,$23,$B1,$F1,$CF,$0E,$39,$3A,$BA,$3E,$14
	.dc.b	$EA,$BB,$4E,$8E,$0F,$D6,$9E,$11,$98,$C3,$A8,$FA,$98,$7B,$9E,$87
	.dc.b	$9C,$E3,$3D,$8E,$24,$78,$9E,$93,$CA,$49,$F0,$3A,$31,$1C,$5F,$B0
	.dc.b	$71,$8F,$23,$F8,$1C,$3E,$BD,$79,$63,$89,$E5,$9E,$AF,$23,$27,$1B
	.dc.b	$E9,$7E,$98,$E0,$F8,$71,$FC,$14,$79,$6B,$C3,$1F,$96,$9E,$13,$D6
	.dc.b	$73,$87,$92,$3A,$9D,$08,$6E,$78,$D7,$38,$CB,$E0,$74,$62,$7A,$3F
	.dc.b	$40,$C5,$F6,$76,$7B,$4F,$7C,$F3,$24,$FE,$5A,$78,$46,$1F,$BB,$C1
	.dc.b	$FE,$1A,$E9,$FC,$C5,$FF,$90,$5B,$3F,$EA,$A8,$ED,$8F,$E1,$EB,$84
	.dc.b	$7F,$93,$EB,$7A,$FC,$BC,$7F,$6B,$5A,$AE,$87,$E9,$70,$E9,$8F,$C3
	.dc.b	$F1,$1C,$8E,$4E,$77,$FE,$33,$F9,$6B,$8F,$F8,$C1,$CD,$F0,$AB,$7C
	.dc.b	$6F,$FC,$C5,$5C,$39,$4E,$AB,$13,$9E,$48,$FF,$9A,$4C,$9F,$E6,$B7
	.dc.b	$2F,$F9,$94,$F2,$9F,$C5,$7B,$7C,$BF,$A6,$BD,$FA,$6B,$54,$D3,$35
	.dc.b	$F0,$F8,$3F,$84,$6C,$C3,$94,$E4,$FE,$51,$5C,$5C,$4E,$F0,$E7,$1F
	.dc.b	$E9,$B3,$C8,$7F,$B7,$B1,$F9,$3D,$CF,$97,$A6,$22,$4F,$FB,$92,$32
	.dc.b	$F1,$3C,$5C,$0F,$CB,$5E,$6F,$D4,$FD,$51,$F0,$F4,$FA,$B3,$AB,$E2
	.dc.b	$D7,$FD,$53,$3C,$1F,$5B,$5F,$F5,$A5,$F7,$BE,$D9,$F7,$F9,$99,$5C
	.dc.b	$23,$2F,$28,$CB,$DA,$32,$F8,$23,$3C,$DF,$EE,$BC,$3E,$7C,$F9,$3E
	.dc.b	$5F,$D9,$E5,$48,$F2,$C7,$BC,$C3,$EA,$8C,$3E,$AF,$92,$F8,$9F,$31
	.dc.b	$23,$04,$3E,$0B,$97,$E1,$C7,$C8,$F4,$FE,$67,$CC,$49,$F9,$31,$97
	.dc.b	$E3,$C6,$5F,$8F,$BA,$75,$1F,$3A,$75,$9B,$9F,$FA,$B7,$07,$DB,$BF
	.dc.b	$CC,$D4,$BF,$C6,$F3,$A3,$EF,$7C,$BF,$EC,$3F,$71,$79,$9F,$A0,$89
	.dc.b	$7F,$86,$9C,$BF,$73,$E6,$7F,$0C,$47,$AC,$FE,$27,$E2,$3A,$7A,$32
	.dc.b	$7F,$07,$15,$FC,$3F,$87,$F8,$F3,$1B,$FD,$97,$11,$F8,$A7,$3B,$FB
	.dc.b	$B2,$FF,$76,$CF,$97,$CB,$F5,$1F,$B8,$F2,$2C,$7C,$0F,$98,$FD,$D9
	.dc.b	$87,$F2,$75,$C9,$FF,$80,$3C,$09,$DC,$75,$3E,$17,$A1,$F1,$C8,$FE
	.dc.b	$7A,$EB,$FA,$29,$FD,$51,$8C,$FE,$B4,$38,$D8,$B0,$DC,$FE,$A2,$F8
	.dc.b	$67,$F7,$3E,$35,$D7,$FC,$99,$B5,$FB,$A6,$BE,$AF,$97,$D4,$FF,$1A
	.dc.b	$F3,$64,$EE,$32,$FE,$12,$32,$FE,$DA,$B9,$E9,$E8,$D1,$0E,$9F,$11
	.dc.b	$2B,$F7,$FE,$2D,$7F,$81,$FC,$C7,$F0,$B3,$18,$73,$BE,$53,$FC,$1A
	.dc.b	$74,$FA,$EF,$F9,$9F,$C0,$BE,$7F,$FA,$DC,$C8,$C3,$83,$FC,$6C,$E1
	.dc.b	$C3,$5E,$11,$FD,$43,$F7,$47,$CC,$53,$BC,$49,$FE,$67,$3A,$6C,$E5
	.dc.b	$FE,$69,$4F,$88,$FE,$B7,$F1,$0F,$8B,$A7,$AA,$25,$7F,$3E,$9D,$36
	.dc.b	$72,$FF,$2A,$F5,$BF,$99,$70,$C8,$FE,$57,$1F,$CD,$4C,$3F,$CC,$23
	.dc.b	$0F,$F4,$AF,$77,$D0,$B8,$3F,$23,$E7,$A7,$31,$86,$F3,$7F,$58,$5F
	.dc.b	$33,$FD,$D9,$CD,$AF,$96,$D9,$35,$5F,$DA,$BE,$57,$E2,$57,$D5,$0E
	.dc.b	$1F,$B0,$D7,$EE,$B5,$FF,$0C,$BB,$C7,$FC,$31,$F2,$CE,$9F,$EE,$ED
	.dc.b	$AF,$ED,$87,$6E,$0F,$11,$A9,$91,$AB,$E9,$A7,$F8,$9B,$F1,$2F,$E3
	.dc.b	$B3,$F8,$B6,$BF,$90,$6B,$BF,$1B,$D8,$A3,$8E,$5F,$E2,$DF,$9E,$7C
	.dc.b	$EF,$F8,$31,$B5,$E6,$AD,$E1,$FE,$36,$F9,$57,$C7,$7F,$B0,$6B,$F3
	.dc.b	$91,$8E,$4F,$94,$D8,$CC,$ED,$FD,$2F,$B0,$F2,$1F,$2D,$F4,$63,$0D
	.dc.b	$8C,$CF,$EA,$67,$2F,$83,$E2,$9E,$CF,$81,$B1,$98,$C1,$F7,$7E,$53
	.dc.b	$F4,$1A,$5E,$BF,$F2,$75,$7F,$94,$A7,$2F,$ED,$DA,$FE,$F1,$CF,$0F
	.dc.b	$F9,$55,$B5,$FD,$73,$E9,$7F,$C5,$EF,$DE,$8C,$D7,$FA,$D7,$E5,$1F
	.dc.b	$19,$FC,$BB,$09,$F2,$AF,$9D,$C6,$3F,$E6,$4B,$DB,$FE,$33,$A5,$6E
	.dc.b	$3F,$17,$3A,$C3,$F8,$73,$5F,$E2,$6D,$7A,$61,$F0,$1F,$23,$E0,$6C
	.dc.b	$FF,$8C,$32,$FC,$D3,$DA,$75,$DC,$F0,$5F,$2D,$86,$D7,$F0,$9E,$DC
	.dc.b	$4C,$8F,$E1,$8E,$E3,$18,$FF,$01,$9F,$F2,$9D,$3E,$47,$A7,$C9,$75
	.dc.b	$6C,$FF,$15,$6D,$71,$3F,$80,$57,$19,$FE,$35,$5E,$D9,$78,$A8,$F4
	.dc.b	$7C,$06,$63,$15,$FC,$86,$31,$3F,$1E,$B8,$B6,$B7,$1F,$CE,$A3,$1F
	.dc.b	$25,$F1,$BC,$67,$CB,$75,$8C,$37,$1F,$E4,$4D,$BF,$56,$DF,$C8,$63
	.dc.b	$1C,$7F,$EA,$3C,$98,$7C,$FF,$A2,$CE,$3F,$E2,$CF,$97,$FD,$AB,$FC
	.dc.b	$E7,$BB,$EE,$7A,$29,$CE,$9D,$B8,$2E,$7F,$DA,$66,$E1,$F4,$B0,$E3
	.dc.b	$18,$BF,$9B,$18,$24,$FF,$47,$EC,$B3,$FE,$6A,$F9,$17,$CB,$7C,$3F
	.dc.b	$97,$F1,$C6,$2F,$FA,$14,$62,$75,$1F,$E8,$09,$D5,$7C,$4E,$25,$EE
	.dc.b	$B5,$7F,$3F,$93,$17,$E1,$BF,$CA,$70,$78,$1B,$7D,$5B,$3B,$8F,$EC
	.dc.b	$13,$B7,$47,$F6,$3F,$81,$C4,$FB,$7E,$2F,$EF,$73,$7B,$1B,$5E,$96
	.dc.b	$CE,$E3,$FB,$04,$ED,$D1,$FD,$8F,$9F,$13,$F5,$63,$F5,$87,$C8,$F9
	.dc.b	$1A,$FC,$56,$C7,$CF,$6F,$17,$F6,$AF,$8C,$F9,$2F,$F8,$97,$81,$3F
	.dc.b	$BA,$9C,$DF,$F9,$D2,$72,$FF,$AF,$A1,$F2,$7B,$3F,$39,$C5,$67,$EB
	.dc.b	$78,$98,$69,$FD,$F6,$3F,$45,$BC,$8E,$9D,$D7,$4F,$F5,$7F,$CB,$BE
	.dc.b	$27,$C9,$E7,$FE,$62,$31,$7E,$38,$C4,$EA,$3F,$D8,$B3,$AA,$FD,$4E
	.dc.b	$9F,$E8,$69,$CD,$7F,$1F,$EB,$5E,$EC,$7F,$A5,$1C,$36,$7F,$93,$77
	.dc.b	$B0,$6F,$F2,$79,$BE,$53,$72,$F9,$0D,$C3,$C6,$DE,$2F,$A9,$F0,$9F
	.dc.b	$47,$E5,$7E,$93,$6F,$E8,$36,$FD,$6D,$AF,$B9,$CF,$F6,$7E,$5F,$ED
	.dc.b	$36,$FE,$C3,$6F,$E1,$B6,$B8,$CF,$DD,$AE,$2D,$3E,$5F,$BD,$F8,$5F
	.dc.b	$97,$FA,$4D,$8C,$9F,$E5,$BE,$1F,$97,$1F,$37,$87,$FB,$0E,$5F,$E6
	.dc.b	$29,$CD,$7F,$98,$78,$3E,$57,$CA,$FE,$47,$E1,$FF,$24,$D9,$FC,$83
	.dc.b	$6F,$ED,$B7,$4F,$EE,$FF,$09,$FD,$4B,$E5,$3E,$F5,$7F,$41,$3E,$DC
	.dc.b	$67,$9B,$FA,$04,$7F,$A6,$70,$FB,$F7,$FE,$4E,$BE,$2F,$48,$C9,$F8
	.dc.b	$4C,$93,$A8,$FF,$3E,$E1,$FD,$0E,$FA,$1F,$E4,$97,$07,$E2,$25,$C5
	.dc.b	$E7,$E3,$F4,$36,$FB,$DF,$F9,$5E,$B7,$18,$C3,$C3,$7C,$23,$13,$F9
	.dc.b	$27,$8E,$AF,$95,$FD,$BC,$6B,$8F,$FA,$0C,$63,$1F,$E9,$B6,$9C,$F6
	.dc.b	$7E,$96,$B5,$5F,$A4,$3F,$D8,$4E,$1F,$F7,$CB,$5F,$F4,$FF,$CB,$FE
	.dc.b	$AB,$E5,$FF,$86,$C7,$AF,$8B,$E8,$A9,$D7,$7F,$CD,$C9,$5A,$9C,$D7
	.dc.b	$1C,$FF,$CC,$EF,$ED,$A3,$3B,$7D,$4A,$E1,$2F,$81,$F0,$BF,$BC,$F0
	.dc.b	$3F,$DB,$71,$88,$E4,$6F,$2F,$F5,$BC,$E9,$FF,$5B,$61,$E2,$9E,$58
	.dc.b	$D7,$8E,$7C,$EF,$FD,$CE,$F5,$8F,$71,$D4,$F9,$51,$D5,$F8,$E6,$4E
	.dc.b	$A5,$3A,$DF,$92,$BC,$11,$D2,$BC,$11,$8B,$FE,$33,$86,$2B,$91,$E2
	.dc.b	$70,$79,$63,$D5,$87,$A3,$0F,$7D,$30,$E4,$6D,$F2,$4F,$A5,$FC,$25
	.dc.b	$78,$2B,$CD,$1E,$08,$EB,$87,$96,$3A,$4E,$AF,$A1,$D9,$F0,$B6,$3E
	.dc.b	$2B,$5E,$E7,$23,$E5,$CF,$5C,$3D,$B5,$EA,$6C,$F2,$8F,$01,$D8,$AF
	.dc.b	$D2,$8D,$BD,$9A,$FD,$23,$B9,$E1,$3C,$75,$E5,$74,$7D,$A7,$DB,$3D
	.dc.b	$A3,$6D,$76,$7F,$A6,$2B,$D5,$1F,$2A,$BE,$81,$E2,$9F,$C0,$47,$16
	.dc.b	$9E,$4A,$EE,$D8,$F7,$34,$F7,$3D,$4F,$8C,$57,$E3,$1B,$3C,$E5,$3C
	.dc.b	$F8,$7E,$1D,$D9,$EF,$8E,$ED,$32,$7B,$4F,$90,$72,$BF,$C2,$57,$C9
	.dc.b	$8F,$59,$C4,$F0,$94,$E0,$72,$36,$75,$7E,$0B,$B3,$35,$E7,$79,$9B
	.dc.b	$AF,$3B,$4F,$3C,$6D,$EF,$3F,$A1,$F6,$7F,$86,$47,$EE,$1A,$FE,$51
	.dc.b	$AF,$F0,$C7,$3E,$B3,$E3,$1C,$DF,$E0,$E9,$CC,$62,$91,$89,$D6,$7F
	.dc.b	$CC,$6F,$94,$62,$BF,$C2,$F1,$87,$F2,$8B,$E2,$FF,$9C,$63,$FC,$8A
	.dc.b	$31,$3F,$C6,$DE,$11,$9C,$7C,$4E,$EB,$8E,$7E,$16,$DF,$82,$B7,$3E
	.dc.b	$56,$C7,$92,$76,$AF,$C2,$F9,$A7,$6E,$5B,$1A,$BF,$14,$F9,$E7,$26
	.dc.b	$AB,$AD,$6E,$FC,$58,$6C,$F7,$4E,$8D,$D7,$28,$CD,$C3,$C8,$F1,$7E
	.dc.b	$5C,$F1,$71,$87,$C7,$9E,$0F,$9A,$38,$93,$1C,$13,$0F,$2C,$7F,$3F
	.dc.b	$C7,$E8,$C7,$CD,$8E,$6E,$AB,$B4,$78,$8C,$9F,$46,$FC,$D7,$D6,$3E
	.dc.b	$81,$FC,$CC,$D1,$DF,$1F,$90,$F2,$3E,$84,$FB,$2F,$C5,$1D,$CF,$5D
	.dc.b	$F8,$8F,$71,$23,$CC,$7B,$2F,$C5,$3F,$7F,$1F,$F0,$44,$7A,$8E,$53
	.dc.b	$DA,$B9,$57,$99,$D4,$78,$A7,$9E,$3F,$9D,$78,$D7,$B8,$F3,$56,$1F
	.dc.b	$59,$8A,$EC,$79,$9E,$CE,$1E,$51,$C6,$BB,$4F,$84,$93,$F2,$AB,$57
	.dc.b	$F8,$38,$F8,$B3,$F0,$4D,$BF,$04,$FA,$47,$92,$78,$BF,$46,$BF,$02
	.dc.b	$F5,$9E,$31,$FA,$69,$F8,$73,$E8,$3C,$F3,$E9,$7D,$F7,$A7,$C0,$7E
	.dc.b	$36,$FE,$8C,$F9,$63,$96,$FF,$E5,$AE,$9F,$F1,$E6,$5F,$E6,$16,$ED
	.dc.b	$F1,$EB,$AD,$FF,$8D,$67,$31,$FE,$55,$6B,$A3,$FC,$8B,$C9,$FE,$47
	.dc.b	$7F,$C9,$1C,$BF,$E5,$38,$C4,$7F,$92,$1A,$F0,$3F,$D1,$5C,$9F,$E5
	.dc.b	$37,$FC,$BB,$F2,$FE,$0C,$FF,$CE,$2F,$8A,$FA,$4F,$F9,$88,$D7,$2B
	.dc.b	$8E,$98,$88,$7F,$9A,$1C,$BC,$4F,$F3,$89,$FC,$5F,$B3,$CA,$D3,$CF
	.dc.b	$C4,$8C,$B6,$33,$1E,$7F,$8D,$7C,$AD,$3E,$3D,$6E,$BF,$D2,$CD,$E1
	.dc.b	$70,$FF,$A3,$EF,$8D,$7F,$A0,$5A,$F0,$C3,$E2,$78,$7E,$F8,$FF,$7B
	.dc.b	$7C,$BE,$0C,$E7,$23,$24,$3F,$DC,$AF,$CB,$F9,$37,$9B,$F5,$37,$46
	.dc.b	$2B,$FD,$7C,$D7,$FD,$78,$D7,$EF,$B9,$D3,$D7,$0E,$17,$CB,$3E,$5F
	.dc.b	$EE,$36,$7F,$D3,$37,$C7,$7F,$F5,$84,$FC,$0C,$39,$E1,$CB,$E4,$38
	.dc.b	$9C,$B8,$A6,$3F,$EA,$7B,$4E,$A3,$13,$A7,$FB,$9E,$1D,$A7,$51,$83
	.dc.b	$D1,$5E,$19,$F1,$FC,$BF,$A3,$CB,$95,$DE,$3D,$66,$1E,$58,$FC,$09
	.dc.b	$CC,$79,$7E,$3F,$E0,$FC,$BC,$8C,$10,$E3,$8A,$35,$18,$9F,$B1,$E9
	.dc.b	$FD,$7D,$8D,$24,$FD,$87,$E7,$FC,$BC,$8F,$9F,$1C,$9F,$91,$19,$7E
	.dc.b	$17,$0F,$E0,$A7,$F3,$93,$27,$F2,$8F,$DA,$F9,$6F,$DA,$7C,$B3,$CC
	.dc.b	$FD,$14,$4A,$FF,$6A,$CE,$5F,$AD,$F2,$FF,$82,$23,$F9,$F9,$FC,$2F
	.dc.b	$C4,$74,$F4,$A4,$FD,$D8,$AF,$C6,$F2,$3F,$16,$63,$9F,$F0,$FC,$3F
	.dc.b	$E3,$19,$FC,$24,$E9,$F1,$DF,$A3,$CB,$91,$E6,$7F,$75,$E3,$E2,$3E
	.dc.b	$DF,$80,$F9,$7F,$DF,$19,$CB,$F3,$FC,$4D,$FF,$DB,$77,$FB,$9A,$DB
	.dc.b	$D4,$F8,$7E,$87,$C8,$23,$F9,$9C,$6B,$F8,$29,$FD,$39,$8C,$3B,$DA
	.dc.b	$F3,$5F,$A1,$AD,$75,$21,$8B,$E5,$FF,$4B,$19,$C7,$D3,$18,$8F,$C0
	.dc.b	$B6,$3F,$63,$9B,$E4,$7F,$99,$4E,$70,$FE,$31,$E6,$FF,$21,$33,$DD
	.dc.b	$31,$FE,$35,$19,$7C,$13,$E5,$CB,$C9,$5D,$3A,$A5,$68,$F4,$E0,$8F
	.dc.b	$E1,$3E,$2E,$A7,$F4,$FF,$2F,$FC,$30,$C6,$BE,$CA,$C4,$FF,$05,$9C
	.dc.b	$E5,$F1,$61,$F0,$6F,$97,$FE,$BA,$D8,$FE,$8A,$D7,$F6,$BF,$31,$A9
	.dc.b	$FE,$AD,$F3,$0F,$98,$A6,$6B,$8C,$E1,$F6,$78,$8A,$E1,$1A,$3B,$FC
	.dc.b	$4F,$F5,$2C,$4D,$24,$7F,$6A,$8C,$BD,$3E,$1D,$F2,$1F,$DA,$5B,$1D
	.dc.b	$0F,$E4,$DE,$BF,$F3,$2E,$59,$18,$21,$FC,$49,$CB,$FC,$52,$3F,$BA
	.dc.b	$7B,$FE,$88,$FE,$3E,$DE,$2F,$E6,$ED,$E0,$FC,$96,$A8,$7C,$C7,$2E
	.dc.b	$09,$87,$C4,$9E,$0F,$F4,$97,$6F,$F7,$6F,$97,$F7,$BF,$CD,$C3,$8C
	.dc.b	$7F,$A8,$C6,$37,$FF,$0B,$FC,$C7,$FA,$BA,$39,$47,$F7,$EA,$E7,$1F
	.dc.b	$E1,$96,$BF,$85,$6B,$FA,$B7,$B6,$7F,$89,$AF,$D2,$4F,$F8,$29,$AF
	.dc.b	$96,$76,$FE,$CF,$E5,$BF,$91,$3E,$96,$E1,$E8,$6D,$FE,$57,$E5,$FF
	.dc.b	$7D,$5F,$E1,$69,$E2,$FF,$8B,$DA,$FE,$6D,$AF,$7F,$27,$FC,$37,$F2
	.dc.b	$EF,$8D,$F8,$53,$FE,$37,$8C,$70,$FF,$8D,$1B,$1F,$E3,$86,$BF,$E3
	.dc.b	$DF,$61,$E5,$3E,$5F,$FC,$4A,$D9,$FF,$1F,$CE,$A7,$27,$5F,$8A,$FF
	.dc.b	$0C,$4F,$85,$AF,$EC,$ED,$18,$8F,$E3,$51,$8F,$63,$FC,$93,$F2,$FF
	.dc.b	$9E,$DF,$FC,$72,$DA,$FE,$B3,$F2,$9F,$E3,$79,$CF,$5F,$F8,$53,$8B
	.dc.b	$F0,$B7,$B6,$C6,$67,$F8,$A5,$71,$BF,$C7,$5E,$E7,$F8,$47,$CA,$BE
	.dc.b	$33,$F1,$F5,$86,$BF,$CC,$63,$53,$FE,$60,$6C,$7F,$99,$1B,$8F,$E6
	.dc.b	$1B,$3E,$2F,$23,$C9,$5C,$EE,$AA,$DB,$9A,$FE,$51,$E4,$32,$FE,$71
	.dc.b	$1F,$E5,$EA,$DF,$6B,$BD,$F2,$BD,$A3,$0D,$8F,$E2,$35,$FC,$36,$30
	.dc.b	$47,$F8,$87,$65,$CB,$FC,$43,$4F,$8F,$E5,$F2,$1C,$DB,$7F,$C5,$DB
	.dc.b	$5C,$1B,$BF,$EE,$BC,$35,$FC,$CA,$FE,$6E,$D8,$CD,$FF,$23,$8C,$1F
	.dc.b	$1A,$33,$3E,$4F,$17,$13,$9F,$90,$F8,$1E,$13,$E5,$71,$8F,$F1,$36
	.dc.b	$E1,$FE,$9A,$C5,$7E,$12,$57,$F3,$37,$79,$7F,$96,$34,$7C,$AE,$2B
	.dc.b	$FB,$07,$CA,$FE,$CD,$FE,$71,$E0,$7D,$EE,$54,$E7,$3F,$97,$39,$8C
	.dc.b	$47,$FA,$B9,$AF,$48,$FC,$8C,$F2,$6C,$F3,$C7,$FC,$2E,$DB,$FE,$6B
	.dc.b	$8F,$EC,$1F,$C4,$C6,$70,$FF,$04,$FC,$7B,$E5,$3E,$3F,$CA,$F9,$E3
	.dc.b	$17,$FC,$E6,$31,$3E,$29,$CD,$7F,$B1,$73,$46,$23,$FD,$8E,$D7,$FC
	.dc.b	$08,$F0,$CB,$FC,$5A,$DC,$3F,$A5,$F0,$7F,$A6,$1F,$CF,$C6,$6B,$CF
	.dc.b	$3F,$99,$DB,$CC,$DA,$DB,$E5,$23,$FE,$D6,$C3,$CB,$18,$AF,$F3,$1D
	.dc.b	$FF,$59,$AF,$ED,$DE,$FF,$47,$C3,$F1,$9F,$8A,$E3,$F8,$8D,$8C,$9D
	.dc.b	$69,$1F,$14,$91,$97,$8B,$FD,$8F,$96,$E1,$FA,$C8,$FD,$A1,$F4,$7E
	.dc.b	$3A,$BF,$09,$FA,$04,$7C,$0F,$F1,$4E,$97,$C5,$7C,$7D,$7E,$9C,$FF
	.dc.b	$01,$8C,$64,$8C,$7C,$8C,$C3,$FB,$B7,$89,$72,$F8,$BE,$1F,$DF,$24
	.dc.b	$FB,$DF,$07,$07,$EA,$23,$FB,$F7,$0F,$7F,$03,$F6,$2F,$E8,$7E,$4D
	.dc.b	$F0,$FE,$3F,$2F,$0D,$BF,$E3,$B3,$A7,$EC,$37,$2F,$C7,$35,$FF,$1F
	.dc.b	$70,$23,$11,$FE,$51,$78,$73,$7F,$9E,$9B,$5F,$E1,$8E,$17,$D0,$F8
	.dc.b	$A3,$FC,$73,$1F,$9A,$6E,$BF,$27,$8B,$E5,$36,$32,$70,$8C,$BC,$BE
	.dc.b	$07,$E8,$7C,$A7,$D1,$6C,$64,$EF,$19,$69,$E2,$E7,$FB,$1F,$2D,$F6
	.dc.b	$5B,$19,$3C,$F1,$96,$99,$D7,$FD,$AF,$92,$75,$18,$7E,$A7,$B7,$F8
	.dc.b	$5F,$96,$B7,$F0,$9B,$5F,$EC,$8D,$BF,$B4,$6F,$FE,$F8,$C5,$E9,$EA
	.dc.b	$31,$F2,$9F,$8F,$F8,$5F,$C8,$37,$2F,$9A,$DC,$7C,$AD,$AF,$ED,$FF
	.dc.b	$07,$FD,$5F,$E4,$BF,$DA,$17,$FE,$87,$BF,$F5,$04,$E7,$9F,$FB,$3C
	.dc.b	$7F,$A9,$E7,$FD,$00,$C9,$FE,$98,$E4,$C9,$CA,$2B,$4F,$F2,$81,$BB
	.dc.b	$F7,$76,$F6,$AF,$27,$17,$97,$92,$E1,$FE,$80,$CB,$F1,$18,$B1,$4F
	.dc.b	$F9,$9A,$57,$C6,$38,$36,$BF,$DE,$BF,$28,$EE,$F8,$5C,$BF,$D2,$63
	.dc.b	$17,$DE,$31,$32,$3F,$33,$F3,$36,$FD,$63,$E5,$FF,$D9,$1F,$2F,$FC
	.dc.b	$06,$1E,$13,$47,$1F,$CC,$9E,$02,$B8,$56,$E7,$57,$FE,$D3,$F8,$0E
	.dc.b	$97,$FE,$D5,$F0,$BE,$0A,$8C,$BF,$ED,$99,$E1,$C7,$F6,$2B,$75,$AA
	.dc.b	$F0,$E3,$C2,$79,$B9,$BF,$5D,$F1,$D2,$DF,$E0,$A6,$BE,$38,$AF,$84
	.dc.b	$F6,$4F,$16,$9C,$4A,$71,$29,$C6,$BC,$93,$D2,$3C,$F3,$CE,$7C,$D1
	.dc.b	$C2,$FD,$33,$F9,$2B,$F1,$C6,$27,$C1,$5D,$1E,$E7,$D2,$C3,$07,$42
	.dc.b	$DF,$38,$ED,$87,$8F,$0E,$AD,$8C,$9D,$23,$2D,$33,$7E,$83,$C4,$77
	.dc.b	$AF,$8D,$76,$7F,$DD,$D8,$7D,$89,$F1,$57,$06,$9F,$0C,$F3,$57,$A2
	.dc.b	$BE,$73,$5F,$F7,$C5,$F9,$2B,$C5,$5C,$67,$E2,$B5,$D9,$E8,$7B,$BE
	.dc.b	$4C,$3E,$04,$F8,$E3,$BC,$FA,$8F,$19,$5E,$A7,$13,$E2,$47,$17,$BB
	.dc.b	$5F,$C1,$47,$8D,$FC,$65,$FA,$DB,$1D,$5E,$2E,$DF,$C3,$9D,$E7,$DE
	.dc.b	$D8,$F5,$BF,$91,$8F,$A0,$FE,$4E,$6A,$C3,$E9,$3E,$A3,$85,$72,$75
	.dc.b	$7F,$12,$3C,$58,$7C,$F7,$55,$F9,$3A,$F3,$9E,$33,$E4,$9F,$16,$FC
	.dc.b	$31,$E0,$8D,$9E,$03,$C4,$78,$0F,$C7,$47,$C5,$7F,$39,$3F,$15,$F8
	.dc.b	$E7,$57,$EA,$B4,$FD,$A7,$B7,$FC,$2D,$B3,$FA,$18,$C4,$7E,$11,$BE
	.dc.b	$D3,$E2,$FF,$9A,$4A,$73,$18,$6F,$22,$75,$75,$5F,$39,$D5,$70,$9F
	.dc.b	$E3,$D8,$6E,$9F,$17,$C7,$0F,$E4,$B1,$8A,$FE,$42,$77,$7A,$3C,$8C
	.dc.b	$63,$E8,$9E,$35,$CE,$FB,$36,$B7,$8F,$47,$CF,$7F,$03,$5E,$31,$FC
	.dc.b	$93,$1F,$13,$FC,$79,$F0,$C7,$9A,$7C,$18,$76,$AE,$95,$C3,$1D,$61
	.dc.b	$8B,$ED,$7B,$BE,$98,$FC,$9C,$77,$1D,$EB,$6F,$C1,$9D,$57,$1A,$D9
	.dc.b	$1C,$12,$7D,$E7,$3B,$F2,$D7,$C3,$8F,$AE,$DC,$7E,$6E,$3C,$E7,$C0
	.dc.b	$7C,$39,$F0,$BD,$30,$D3,$C8,$D1,$F2,$71,$F4,$BF,$3B,$0F,$0B,$0F
	.dc.b	$B4,$64,$B8,$7C,$EC,$7D,$51,$D4,$F5,$12,$FE,$BB,$F2,$8F,$05,$47
	.dc.b	$9E,$1E,$D3,$DD,$87,$D8,$8F,$55,$F6,$C3,$EF,$E1,$EC,$7C,$53,$F0
	.dc.b	$8E,$C7,$86,$7F,$35,$5E,$E9,$E3,$1C,$E3,$07,$79,$F5,$63,$E8,$3E
	.dc.b	$F9,$DF,$0F,$9B,$1E,$FC,$3C,$55,$C7,$0F,$C6,$61,$F8,$9C,$38,$BE
	.dc.b	$2C,$7E,$8E,$1E,$6C,$38,$C6,$1F,$8E,$62,$35,$3E,$79,$F9,$15,$E5
	.dc.b	$8E,$C7,$59,$FA,$B8,$7B,$F1,$F3,$E3,$F0,$27,$E5,$90,$F5,$57,$57
	.dc.b	$E4,$1F,$2F,$1F,$1C,$75,$7E,$14,$79,$E3,$04,$78,$3C,$E3,$E0,$C7
	.dc.b	$D5,$C3,$D0,$FE,$81,$FC,$D6,$FF,$E3,$9E,$9F,$E2,$D9,$FF,$8C,$23
	.dc.b	$15,$FE,$69,$C7,$EF,$F8,$6E,$6F,$F6,$BC,$FF,$1F,$BF,$F1,$BC,$E7
	.dc.b	$9B,$9F,$17,$F2,$38,$C5,$7F,$A9,$BE,$5F,$2A,$32,$FF,$98,$67,$FC
	.dc.b	$C1,$3A,$CB,$FC,$9C,$F0,$AD,$E1,$E7,$E5,$FF,$3C,$B7,$8F,$84,$7F
	.dc.b	$5D,$C7,$FC,$DC,$60,$CC,$47,$27,$BF,$C0,$F2,$BF,$CE,$1B,$A9,$F2
	.dc.b	$77,$3B,$33,$1F,$E8,$4A,$E0,$7F,$A1,$AB,$83,$63,$51,$88,$F8,$07
	.dc.b	$F9,$B3,$E2,$9F,$29,$F0,$5B,$1E,$53,$FD,$9F,$8F,$8B,$E5,$E4,$60
	.dc.b	$87,$FA,$E9,$CB,$FE,$BC,$8F,$95,$F2,$6F,$97,$A4,$9F,$F6,$D9,$4F
	.dc.b	$2F,$0B,$E4,$5F,$2F,$F5,$5B,$3F,$83,$E4,$FC,$06,$1D,$63,$07,$FA
	.dc.b	$22,$7D,$6F,$45,$F2,$1C,$CE,$9A,$C9,$E3,$8F,$86,$75,$18,$AE,$7C
	.dc.b	$7F,$99,$6E,$3F,$DC,$EF,$85,$6E,$FA,$3E,$1F,$97,$F5,$79,$D2,$3A
	.dc.b	$47,$B4,$C3,$E7,$8C,$3E,$1F,$90,$F8,$5F,$2F,$23,$04,$3D,$CE,$5F
	.dc.b	$81,$1F,$4B,$D6,$91,$F1,$BA,$5B,$A7,$E7,$7B,$BA,$AF,$F6,$BD,$6A
	.dc.b	$7F,$DA,$6D,$8F,$F6,$BB,$5D,$D7,$37,$E4,$7C,$BF,$D9,$3A,$70,$7F
	.dc.b	$01,$E6,$47,$E4,$49,$3F,$3F,$E5,$BF,$65,$F2,$CF,$3B,$C2,$D9,$F0
	.dc.b	$B5,$F3,$B5,$FF,$12,$71,$7E,$2F,$E5,$FF,$14,$48,$FB,$F8,$B0,$FC
	.dc.b	$77,$C6,$F8,$3D,$59,$5F,$81,$8A,$F6,$8F,$19,$1F,$F7,$CE,$1F,$99
	.dc.b	$F2,$BF,$30,$48,$FA,$19,$2B,$DB,$E6,$C9,$CB,$FB,$3F,$2F,$11,$EC
	.dc.b	$8F,$82,$47,$E0,$BF,$B9,$F9,$79,$1C,$8E,$FE,$1F,$CA,$23,$F9,$BA
	.dc.b	$AF,$C8,$8F,$9A,$47,$E6,$BF,$B1,$99,$1D,$0F,$C5,$7C,$BF,$EB,$1B
	.dc.b	$D7,$FC,$A6,$31,$1F,$BC,$DD,$F1,$79,$AF,$C5,$D6,$AB,$FC,$B5,$CC
	.dc.b	$8C,$47,$F3,$D6,$BF,$CE,$DE,$BC,$BF,$8B,$F3,$3F,$8D,$1D,$67,$CB
	.dc.b	$D5,$E3,$23,$FB,$BF,$1A,$BF,$9E,$B4,$F9,$09,$1F,$9E,$8C,$BF,$C0
	.dc.b	$FE,$5F,$F8,$A1,$CB,$07,$81,$FE,$73,$ED,$7C,$47,$F3,$E6,$DF,$F1
	.dc.b	$60,$E1,$FE,$2D,$F2,$FB,$9F,$E7,$D3,$99,$FF,$3F,$1C,$23,$24,$7F
	.dc.b	$8C,$FC,$BB,$E5,$BF,$AE,$36,$3C,$15,$DF,$1F,$3C,$7F,$A2,$E1,$F1
	.dc.b	$9E,$72,$47,$F2,$1E,$2A,$F8,$8F,$4C,$79,$88,$FE,$8B,$C6,$FF,$4D
	.dc.b	$3F,$CA,$78,$7F,$9D,$1F,$CE,$FD,$59,$19,$21,$FC,$DD,$FE,$E9,$EF
	.dc.b	$FA,$63,$FA,$2B,$72,$FF,$05,$37,$7F,$E6,$4B,$E5,$F8,$ED,$2F,$F9
	.dc.b	$EC,$ED,$77,$A3,$84,$7D,$5F,$96,$F8,$6D,$D7,$FB,$BC,$62,$3F,$B5
	.dc.b	$8F,$97,$F7,$B8,$8D,$67,$F3,$A1,$18,$CF,$FD,$70,$D8,$FF,$5E,$57
	.dc.b	$2A,$E1,$3D,$E7,$F0,$2D,$78,$DF,$F8,$0E,$31,$1F,$EC,$5F,$97,$FE
	.dc.b	$49,$5E,$16,$E1,$FE,$C7,$6C,$FF,$B2,$FE,$03,$FC,$1C,$FE,$56,$32
	.dc.b	$9A,$62,$37,$1F,$0B,$E5,$9F,$1F,$F0,$A3,$FD,$A9,$18,$E0,$FC,$44
	.dc.b	$23,$0F,$E2,$DA,$FE,$7B,$D5,$78,$EF,$97,$FF,$0F,$36,$7F,$25,$3A
	.dc.b	$9C,$9D,$A9,$F1,$5F,$E3,$5A,$FD,$4C,$63,$D6,$D4,$E6,$7C,$ED,$7F
	.dc.b	$DD,$DF,$28,$F8,$BF,$D3,$F0,$4F,$98,$7C,$CE,$38,$BF,$DE,$57,$B7
	.dc.b	$FC,$27,$4A,$DC,$7F,$01,$9D,$61,$FA,$B6,$BF,$BE,$6B,$D3,$0F,$C5
	.dc.b	$3F,$EF,$FF,$C3,$3F,$B4,$D7,$F7,$BF,$C3,$5E,$1F,$8E,$82,$B6,$FF
	.dc.b	$90,$BC,$75,$70,$BF,$E2,$95,$AC,$FF,$CD,$F3,$9B,$FC,$FC,$E6,$BF
	.dc.b	$CE,$13,$9A,$FC,$F4,$F6,$AF,$E2,$ED,$7D,$1E,$A7,$A9,$F2,$C6,$57
	.dc.b	$23,$F1,$B7,$C8,$FA,$E3,$0F,$B7,$BB,$F7,$1C,$39,$AD,$E9,$FE,$42
	.dc.b	$BD,$B6,$33,$87,$F0,$AA,$E1,$87,$3E,$E7,$D1,$77,$FD,$97,$D9,$8F
	.dc.b	$F0,$C8,$C7,$0E,$DB,$1F,$C3,$1B,$8F,$F0,$4F,$19,$F2,$7D,$7E,$6B
	.dc.b	$EA,$4F,$F4,$08,$C5,$FF,$34,$6C,$71,$6F,$17,$FA,$83,$F6,$9F,$F3
	.dc.b	$BD,$7F,$1A,$FA,$3F,$BD,$CC,$9C,$E1,$FC,$9A,$73,$18,$BF,$E4,$F1
	.dc.b	$8F,$82,$F8,$0D,$C3,$FC,$77,$5C,$AF,$FB,$9B,$5F,$F5,$73,$5F,$F5
	.dc.b	$67,$C9,$3E,$63,$E5,$7C,$CF,$92,$31,$7F,$CC,$23,$13,$A8,$FF,$29
	.dc.b	$4E,$AB,$F9,$ED,$6A,$73,$5F,$57,$E1,$F1,$5F,$D0,$BE,$77,$EC,$7D
	.dc.b	$07,$D9,$6D,$FF,$3E,$6C,$EE,$3F,$A0,$CE,$DD,$1F,$D1,$3C,$BC,$4F
	.dc.b	$E9,$72,$7C,$77,$E1,$7E,$6F,$D8,$DB,$F4,$B6,$B8,$37,$83,$FD,$AF
	.dc.b	$7C,$AB,$67,$F5,$2E,$9E,$6F,$F5,$4F,$2B,$F2,$5E,$AF,$E4,$DB,$19
	.dc.b	$3F,$AC,$C6,$5A,$67,$1F,$EC,$B3,$F9,$BF,$9E,$D4,$7F,$68,$F1,$FF
	.dc.b	$3C,$7D,$5F,$22,$DF,$D2,$6C,$FC,$16,$E7,$FD,$BB,$A3,$85,$7F,$6A
	.dc.b	$6D,$7F,$74,$F8,$8F,$D5,$1F,$AB,$9A,$7F,$AB,$3E,$3F,$EF,$5F,$F7
	.dc.b	$88,$FD,$D4,$61,$FE,$F2,$63,$C8,$8F,$1F,$97,$B8,$7F,$70,$3F,$DC
	.dc.b	$91,$D7,$E2,$7F,$5B,$18,$3F,$D9,$77,$FE,$89,$9F,$F1,$84,$67,$0F
	.dc.b	$F7,$E5,$70,$F8,$2F,$1B,$6B,$A1,$C2,$32,$FA,$FC,$4F,$4F,$CE,$7A
	.dc.b	$9B,$19,$3E,$04,$65,$A7,$C3,$F9,$9F,$83,$CB,$F0,$9B,$19,$3E,$44
	.dc.b	$65,$A7,$CB,$F9,$CF,$93,$F1,$1F,$29,$B1,$93,$E8,$46,$5A,$67,$4F
	.dc.b	$8B,$19,$6C,$7F,$B1,$7E,$55,$BF,$D6,$E9,$FE,$1F,$18,$9F,$B9,$18
	.dc.b	$9D,$47,$16,$D7,$D2,$F5,$FE,$FF,$CE,$7E,$01,$B5,$F8,$56,$CE,$E3
	.dc.b	$F1,$13,$F7,$21,$F3,$5F,$D7,$3E,$65,$F3,$5F,$EF,$1E,$57,$C6,$EE
	.dc.b	$B5,$39,$8C,$5A,$F8,$BF,$E4,$2B,$E3,$8F,$E1,$A3,$B9,$BB,$E5,$19
	.dc.b	$C3,$AC,$F1,$C3,$AE,$1D,$3D,$36,$3D,$2B,$F3,$07,$49,$F0,$BF,$DD
	.dc.b	$BA,$FF,$CE,$71,$88,$FE,$F3,$87,$67,$58,$D7,$27,$94,$FE,$FE,$2B
	.dc.b	$9E,$5F,$E7,$7A,$E1,$18,$9F,$F4,$5C,$62,$BA,$47,$96,$BF,$D1,$F5
	.dc.b	$B8,$EB,$5B,$7C,$06,$0A,$78,$6B,$9C,$6E,$FD,$93,$D4,$F3,$D7,$57
	.dc.b	$D7,$1F,$06,$31,$1B,$3E,$09,$D8,$C1,$DC,$C4,$F7,$BF,$81,$1E,$49
	.dc.b	$E7,$39,$8F,$82,$66,$FC,$ED,$39,$96,$FD,$33,$C6,$3E,$04,$79,$E7
	.dc.b	$8B,$63,$8B,$4E,$2F,$3B,$F2,$C6,$5A,$7D,$09,$E6,$D9,$E6,$D9,$E6
	.dc.b	$71,$8F,$5B,$5F,$2C,$6C,$EA,$FD,$09,$EA,$D3,$A9,$67,$AC,$F1,$8E
	.dc.b	$AD,$7A,$9B,$7A,$BF,$5A,$38,$B9,$38,$94,$E2,$6E,$38,$BF,$6E,$38
	.dc.b	$B9,$38,$94,$E2,$6E,$38,$DF,$DE,$8F,$9C,$70,$34,$57,$53,$F1,$E3
	.dc.b	$F2,$33,$ED,$69,$E7,$29,$E1,$29,$D0,$F1,$CF,$48,$E5,$1F,$C2,$BE
	.dc.b	$0B,$F0,$87,$E0,$8B,$1E,$8B,$FA,$73,$D4,$FC,$ED,$73,$7C,$46,$23
	.dc.b	$6F,$84,$F1,$C6,$DE,$58,$7E,$5D,$CC,$F0,$9D,$47,$13,$F8,$FF,$A8
	.dc.b	$C3,$F5,$2D,$8F,$D3,$B5,$FE,$00,$DF,$71,$F1,$EE,$CF,$F4,$8C,$E6
	.dc.b	$31,$5F,$E7,$EB,$FE,$2F,$7C,$2F,$71,$8A,$DD,$F1,$9F,$E4,$54,$F8
	.dc.b	$EF,$E3,$FA,$FF,$2A,$8C,$63,$AC,$3B,$CE,$5B,$AF,$82,$F8,$B6,$B8
	.dc.b	$61,$1F,$E3,$B8,$74,$6D,$70,$7F,$D5,$B1,$E9,$77,$7D,$9C,$CE,$D4
	.dc.b	$F3,$C3,$C9,$7D,$DF,$55,$6E,$75,$87,$C3,$AD,$3E,$7B,$F1,$98,$3C
	.dc.b	$78,$7C,$EC,$B9,$C7,$FB,$09,$AF,$DD,$9E,$8D,$38,$D7,$CE,$C3,$C4
	.dc.b	$78,$EF,$D9,$3E,$A6,$E1,$E6,$8F,$79,$E3,$35,$5C,$D8,$7C,$67,$51
	.dc.b	$C6,$3E,$2D,$77,$9C,$BE,$28,$DD,$78,$23,$11,$CC,$98,$FB,$89,$87
	.dc.b	$B9,$F9,$F1,$F2,$08,$E0,$91,$F3,$5E,$31,$F1,$0E,$B1,$92,$39,$8E
	.dc.b	$2F,$C2,$3A,$C7,$84,$8E,$63,$AB,$F1,$8E,$A6,$63,$AB,$98,$EA,$FC
	.dc.b	$63,$E5,$B8,$C3,$F2,$4F,$47,$E6,$18,$9F,$0C,$F1,$3C,$A7,$A5,$CC
	.dc.b	$70,$AF,$A5,$3E,$43,$85,$75,$BF,$C0,$9F,$4E,$3C,$77,$CA,$31,$1D
	.dc.b	$8F,$54,$76,$AF,$B3,$5E,$9A,$DB,$D7,$0F,$C0,$DF,$43,$CC,$43,$C7
	.dc.b	$7F,$1B,$0F,$74,$65,$F5,$3F,$88,$9E,$4F,$03,$57,$E0,$AE,$4F,$B2
	.dc.b	$77,$5F,$2A,$31,$5F,$E2,$FD,$FF,$C7,$BF,$35,$FE,$30,$B6,$DF,$E5
	.dc.b	$38,$C3,$F0,$E3,$FC,$8F,$3C,$2B,$FD,$49,$5A,$7C,$73,$AE,$04,$E6
	.dc.b	$B8,$57,$F2,$DB,$FC,$5C,$7F,$93,$BE,$6F,$57,$FE,$4C,$C3,$C7,$7F
	.dc.b	$E5,$DC,$F3,$52,$3F,$D6,$91,$C0,$C3,$D4,$7A,$4E,$8F,$F7,$1C,$7F
	.dc.b	$AE,$FE,$6F,$B9,$E9,$9D,$74,$DA,$DD,$7F,$AC,$AB,$6F,$FA,$36,$7C
	.dc.b	$FF,$35,$E8,$F0,$64,$FF,$B1,$E3,$2F,$FB,$26,$32,$FF,$AF,$23,$2F
	.dc.b	$FA,$FA,$33,$9F,$C0,$D7,$FC,$E3,$5F,$12,$BE,$2F,$34,$7F,$DA,$86
	.dc.b	$4F,$F6,$B3,$97,$FD,$92,$7D,$1F,$91,$7C,$8D,$F9,$A9,$23,$E4,$60
	.dc.b	$AD,$C6,$71,$F2,$46,$23,$6A,$B9,$C6,$2F,$F0,$8F,$F4,$8F,$92,$F1
	.dc.b	$7B,$12,$3A,$72,$7F,$6A,$9D,$F3,$9D,$4F,$6A,$CC,$48,$E2,$33,$F4
	.dc.b	$4F,$0C,$FF,$13,$3E,$13,$73,$ED,$F9,$AF,$88,$48,$CE,$8F,$9A,$F9
	.dc.b	$1F,$35,$F2,$7E,$56,$56,$A3,$E6,$7B,$D2,$38,$BA,$7E,$2F,$C2,$FE
	.dc.b	$37,$E5,$DE,$75,$3E,$A4,$59,$F0,$FC,$C7,$D7,$30,$FC,$E7,$0E,$8F
	.dc.b	$83,$D6,$BE,$B1,$EA,$F0,$3F,$00,$47,$DB,$F3,$35,$FB,$3F,$1D,$F8
	.dc.b	$53,$1C,$5F,$7A,$B1,$3F,$12,$73,$CD,$F8,$7C,$7C,$1E,$47,$B4,$95
	.dc.b	$F2,$BC,$5A,$FE,$1B,$E6,$7F,$24,$63,$2F,$DE,$DF,$29,$FA,$33,$A7
	.dc.b	$DB,$A7,$FB,$91,$FA,$BF,$31,$F9,$B2,$3F,$63,$A6,$BF,$94,$F8,$7F
	.dc.b	$D0,$18,$C7,$F3,$35,$89,$FB,$F3,$9D,$3D,$9F,$33,$F9,$DB,$F2,$B6
	.dc.b	$33,$DB,$FA,$BF,$98,$FE,$11,$39,$C3,$F2,$5D,$5F,$E4,$DC,$3F,$4F
	.dc.b	$E5,$EB,$09,$5F,$8E,$8F,$DA,$18,$7F,$8C,$6E,$FD,$AD,$F0,$DF,$EE
	.dc.b	$49,$FF,$72,$CE,$63,$11,$F4,$E7,$6E,$27,$4D,$8C,$BF,$BE,$F9,$A7
	.dc.b	$CC,$FF,$35,$6C,$7E,$32,$77,$97,$D6,$6E,$DF,$C1,$FC,$15,$F0,$BE
	.dc.b	$98,$7F,$2B,$BD,$C7,$F0,$DE,$D7,$C0,$F7,$F9,$99,$3F,$BA,$13,$A7
	.dc.b	$0D,$7D,$11,$C1,$FE,$7C,$70,$9F,$57,$C7,$FF,$53,$F4,$DF,$B3,$FF
	.dc.b	$02,$DB,$FA,$5C,$63,$6D,$E1,$FC,$13,$56,$1C,$70,$E1,$5F,$E8,$26
	.dc.b	$D7,$F0,$FA,$DB,$FC,$4A,$DF,$8E,$FD,$7C,$7B,$B7,$0F,$F4,$5B,$6B
	.dc.b	$97,$C8,$79,$E3,$17,$CA,$70,$7D,$59,$FE,$57,$F0,$9F,$C5,$63,$F0
	.dc.b	$21,$C3,$F8,$46,$BF,$89,$6B,$FC,$FC,$E1,$E4,$7F,$36,$F8,$87,$97
	.dc.b	$4E,$F1,$27,$EB,$4E,$9B,$39,$7E,$9F,$CD,$3E,$2B,$CB,$08,$C3,$E8
	.dc.b	$6B,$F9,$A6,$BF,$9F,$73,$AF,$F3,$AF,$9B,$77,$7F,$4C,$6D,$7E,$26
	.dc.b	$B7,$BF,$B7,$E6,$5F,$16,$E0,$F8,$F1,$86,$C6,$6F,$F5,$73,$97,$F5
	.dc.b	$BE,$8B,$C9,$7C,$D7,$D0,$8C,$36,$33,$3C,$1A,$FE,$E6,$BE,$B7,$A9
	.dc.b	$E2,$E3,$34,$93,$FB,$CF,$99,$7C,$57,$F5,$A8,$FB,$11,$8D,$7F,$02
	.dc.b	$D8,$E1,$3B,$7F,$BE,$46,$78,$7F,$B1,$B6,$BF,$9D,$7D,$17,$CC,$7F
	.dc.b	$65,$19,$AF,$F0,$27,$CC,$BD,$3F,$F1,$9E,$1F,$CE,$3E,$69,$E5,$E3
	.dc.b	$1F,$C6,$5E,$DF,$EA,$B3,$A7,$F7,$D1,$9C,$3F,$8C,$B5,$FE,$3C,$D7
	.dc.b	$9E,$1E,$B5,$1C,$70,$FE,$FA,$DA,$FF,$0D,$7E,$20,$FC,$4D,$F2,$7E
	.dc.b	$5C,$3D,$78,$DE,$67,$CE,$6B,$1F,$DD,$36,$BF,$C4,$7F,$23,$FC,$12
	.dc.b	$73,$87,$1A,$D6,$9F,$C3,$1B,$3F,$C9,$6B,$85,$FF,$94,$F0,$EF,$87
	.dc.b	$8F,$CA,$53,$C1,$F2,$7C,$E7,$C7,$C3,$C3,$18,$E1,$E1,$5B,$CB,$F8
	.dc.b	$85,$EF,$4F,$F2,$44,$E7,$0F,$04,$E6,$BF,$80,$79,$2F,$94,$E7,$F3
	.dc.b	$9F,$16,$FF,$98,$C6,$2B,$F9,$8B,$F4,$6B,$84,$7F,$10,$6C,$FF,$52
	.dc.b	$6B,$FD,$5B,$E6,$FF,$85,$CF,$F8,$C3,$17,$CE,$7D,$3F,$41,$5A,$BF
	.dc.b	$9F,$39,$8C,$69,$EC,$8C,$78,$DF,$09,$B8,$FF,$88,$AB,$6F,$D1,$F5
	.dc.b	$73,$87,$F6,$CF,$94,$7C,$E7,$CA,$F5,$3C,$51,$8B,$F3,$46,$27,$51
	.dc.b	$FE,$2E,$9D,$57,$E1,$2B,$53,$9A,$FA,$BF,$39,$8A,$FC,$4F,$CE,$7D
	.dc.b	$8F,$9C,$FB,$2D,$BF,$88,$D9,$DC,$7C,$79,$DB,$A3,$E5,$79,$B8,$9F
	.dc.b	$67,$C5,$FE,$13,$CF,$F5,$B6,$BD,$0D,$9D,$C7,$D3,$9D,$BA,$3E,$B7
	.dc.b	$5E,$27,$F4,$3E,$67,$E4,$3D,$5F,$C8,$B6,$BE,$E3,$67,$71,$F7,$E7
	.dc.b	$6E,$8F,$C1,$7C,$E6,$27,$F5,$BE,$57,$E7,$0F,$A7,$E5,$5A,$FA,$0D
	.dc.b	$9F,$B4,$DE,$7F,$CF,$F5,$F0,$AF,$C7,$36,$BF,$73,$F3,$9F,$A8,$3F
	.dc.b	$53,$34,$FF,$41,$FC,$97,$DC,$BF,$D4,$C7,$ED,$63,$0F,$EE,$CC,$79
	.dc.b	$91,$E3,$F3,$17,$0F,$F0,$98,$F8,$FE,$5F,$15,$FA,$D8,$C5,$78,$23
	.dc.b	$13,$A8,$FF,$11,$4E,$8F,$F5,$9E,$1F,$E8,$88,$CC,$FF,$23,$9D,$1F
	.dc.b	$E4,$27,$07,$F4,$1F,$19,$5F,$C3,$AB,$DF,$F3,$9E,$DF,$0B,$DC,$D8
	.dc.b	$C9,$F1,$23,$2F,$C6,$F9,$CF,$89,$DD,$F1,$5B,$19,$3E,$5C,$65,$A7
	.dc.b	$CE,$F9,$CF,$99,$F3,$9F,$35,$B1,$93,$E9,$C6,$5A,$67,$4F,$87,$19
	.dc.b	$6C,$7F,$A8,$BE,$55,$BF,$D9,$EB,$FE,$1D,$18,$9F,$BF,$18,$9F,$EC
	.dc.b	$DE,$32,$B5,$7F,$E6,$F8,$C4,$EA,$73,$3F,$E2,$9D,$FE,$6F,$CE,$7E
	.dc.b	$1F,$E7,$3F,$10,$DC,$BF,$1A,$DA,$FB,$6D,$E5,$FE,$ED,$F3,$9F,$EF
	.dc.b	$1F,$9C,$7C,$D7,$F9,$B7,$95,$F2,$3F,$89,$C5,$7B,$AD,$4E,$63,$13
	.dc.b	$FE,$E3,$AE,$39,$7E,$32,$38,$8C,$3B,$BB,$BE,$4B,$E2,$B2,$8F,$F0
	.dc.b	$BD,$58,$CC,$67,$0C,$63,$FE,$97,$ED,$5E,$2B,$FD,$35,$5E,$67,$A4
	.dc.b	$E9,$F2,$57,$FA,$72,$75,$1E,$03,$27,$22,$9C,$AB,$6F,$BB,$1F,$2D
	.dc.b	$F5,$7C,$F1,$9B,$F3,$57,$0B,$EF,$7B,$9F,$15,$7C,$37,$6D,$3C,$58
	.dc.b	$76,$BF,$74,$7F,$9B,$6F,$9C,$75,$8F,$39,$EB,$C3,$C7,$3A,$8C,$9E
	.dc.b	$53,$FC,$ED,$87,$8D,$EF,$87,$8B,$0E,$31,$E3,$3E,$11,$4F,$A8,$78
	.dc.b	$27,$93,$E5,$31,$87,$96,$7D,$51,$C4,$FC,$3B,$83,$BE,$1A,$2D,$FC
	.dc.b	$47,$C7,$1B,$8F,$92,$D9,$FB,$13,$C6,$7E,$6B,$63,$DA,$7C,$0A,$F9
	.dc.b	$6F,$AA,$B9,$3E,$99,$E0,$DA,$F1,$1D,$0A,$F8,$1A,$EC,$EC,$7D,$0C
	.dc.b	$3E,$94,$FA,$A7,$E7,$1E,$A2,$CF,$33,$8C,$7C,$26,$C6,$DE,$B5,$CD
	.dc.b	$B1,$E5,$76,$73,$3E,$CC,$F3,$6C,$6D,$A6,$CF,$C5,$39,$69,$F2,$8F
	.dc.b	$C1,$CF,$85,$B1,$BC,$3F,$07,$1F,$86,$AF,$96,$D9,$F8,$ED,$3C,$E7
	.dc.b	$E2,$6B,$E9,$BF,$56,$3E,$2B,$F3,$4F,$39,$4F,$29,$6B,$C2,$FE,$7C
	.dc.b	$D9,$DC,$A6,$4E,$26,$4A,$65,$E7,$7F,$6A,$7D,$D8,$7E,$C7,$0F,$6D
	.dc.b	$EE,$31,$3E,$E8,$C1,$F8,$33,$F4,$A5,$9E,$35,$D6,$3E,$95,$F8,$1F
	.dc.b	$F6,$76,$1F,$C2,$3D,$3F,$E5,$51,$FC,$5D,$AF,$EC,$5B,$EF,$BE,$49
	.dc.b	$B7,$F2,$F9,$CC,$62,$BA,$9F,$CC,$EB,$85,$6E,$FF,$DF,$91,$89,$D8
	.dc.b	$71,$97,$F3,$C5,$F2,$5F,$CF,$67,$F9,$5C,$61,$FE,$6C,$F0,$D3,$AE
	.dc.b	$3E,$3B,$E5,$7C,$5B,$7C,$27,$55,$C3,$1E,$77,$B8,$F3,$B6,$BD,$95
	.dc.b	$C1,$B5,$C1,$87,$9A,$34,$FB,$5B,$87,$A6,$3D,$75,$D5,$D5,$79,$4E
	.dc.b	$78,$7A,$23,$A4,$65,$E2,$6B,$0F,$53,$A3,$24,$9F,$99,$87,$88,$F1
	.dc.b	$DF,$87,$0F,$24,$EB,$0E,$D8,$F8,$49,$8F,$67,$58,$F5,$8F,$8E,$F9
	.dc.b	$6B,$27,$A4,$91,$E6,$78,$CF,$49,$FC,$46,$1F,$7E,$33,$5B,$3A,$CF
	.dc.b	$73,$E8,$9E,$FA,$EA,$F2,$3C,$51,$C5,$C1,$C4,$F9,$E7,$C7,$9F,$13
	.dc.b	$F9,$3A,$F1,$4F,$BC,$E5,$19,$3E,$81,$E2,$AF,$3B,$DA,$B9,$E1,$F9
	.dc.b	$F7,$E1,$D7,$AA,$BC,$F5,$E2,$72,$F9,$9C,$D7,$E7,$63,$B4,$F1,$23
	.dc.b	$83,$E1,$D7,$82,$FD,$31,$89,$F4,$D7,$DE,$78,$C7,$85,$F7,$3E,$97
	.dc.b	$E7,$D7,$B2,$34,$79,$63,$8C,$F9,$63,$E5,$CF,$85,$FF,$07,$C7,$C1
	.dc.b	$9F,$0D,$7E,$3C,$FC,$86,$3F,$93,$C7,$71,$E8,$BF,$9C,$7F,$82,$6F
	.dc.b	$93,$EF,$33,$8F,$C6,$BE,$51,$F7,$67,$DC,$ED,$E2,$F4,$C7,$EA,$CF
	.dc.b	$F1,$97,$E8,$C7,$4C,$7D,$B3,$D2,$FE,$9E,$1E,$8C,$7F,$1B,$7F,$0C
	.dc.b	$87,$C6,$3E,$1D,$FC,$A2,$3F,$4C,$FB,$1C,$5F,$EA,$2F,$9C,$FE,$3D
	.dc.b	$5F,$E4,$F8,$C4,$FF,$97,$63,$FC,$A4,$F5,$CF,$F0,$93,$A8,$FF,$59
	.dc.b	$D7,$2D,$FF,$41,$3F,$EA,$EF,$9C,$C5,$7F,$9C,$71,$B3,$FE,$B2,$6F
	.dc.b	$12,$A9,$88,$95,$FE,$D4,$8E,$7C,$FF,$E9,$5E,$8F,$E6,$D3,$FC,$DE
	.dc.b	$30,$E0,$38,$7F,$D3,$5C,$E9,$D4,$6A,$30,$FF,$B8,$BE,$73,$CA,$78
	.dc.b	$23,$31,$FE,$D2,$F9,$CF,$4F,$83,$23,$FD,$D3,$1E,$C3,$0F,$FB,$62
	.dc.b	$30,$E8,$FF,$77,$BA,$E0,$FA,$5A,$2B,$E2,$D7,$C6,$E6,$8F,$FB,$D8
	.dc.b	$C9,$FE,$F8,$72,$FF,$B9,$8F,$9F,$F2,$4F,$92,$BF,$39,$27,$EF,$71
	.dc.b	$7F,$A6,$F8,$97,$B9,$C8,$8C,$DF,$18,$FE,$EB,$5C,$69,$7E,$0A,$FC
	.dc.b	$0D,$7F,$6F,$F9,$5F,$9E,$D3,$E8,$F2,$A3,$2D,$8C,$C7,$AB,$D9,$91
	.dc.b	$E4,$39,$1E,$23,$FC,$05,$1A,$71,$97,$B2,$7C,$75,$D6,$73,$8B,$E7
	.dc.b	$3E,$39,$23,$E0,$DF,$C4,$F8,$EF,$97,$F3,$7F,$33,$E7,$25,$6B,$04
	.dc.b	$FD,$0F,$7A,$4E,$5D,$BE,$0F,$84,$FC,$DF,$CC,$BC,$EA,$7D,$88,$B3
	.dc.b	$D7,$E7,$3E,$D9,$87,$D2,$E1,$D1,$F6,$3B,$17,$D6,$3C,$DE,$17,$E1
	.dc.b	$08,$FA,$7E,$72,$BF,$13,$E3,$FF,$12,$63,$AF,$F8,$36,$DF,$E3,$59
	.dc.b	$F8,$13,$A7,$A9,$E1,$C3,$FD,$EE,$DF,$23,$DE,$7E,$53,$C4,$B1,$F2
	.dc.b	$7E,$73,$F2,$E6,$1F,$B9,$A7,$AA,$B3,$1B,$3A,$1F,$47,$E7,$3F,$3E
	.dc.b	$47,$EA,$74,$D7,$EF,$FC,$47,$E9,$0C,$6B,$EC,$AC,$4F,$DB,$9C,$F6
	.dc.b	$FF,$BF,$F1,$7C,$27,$E9,$6B,$CA,$DC,$3F,$DF,$13,$AE,$DF,$D8,$FC
	.dc.b	$AF,$F1,$8A,$D6,$1F,$A7,$F2,$FF,$6A,$66,$A6,$3F,$8B,$8C,$BF,$6E
	.dc.b	$B8,$BF,$C7,$3E,$6D,$F0,$DF,$C9,$1B,$1F,$81,$6B,$F9,$EF,$3B,$53
	.dc.b	$FC,$CB,$E6,$1F,$39,$4C,$D7,$19,$C3,$FA,$29,$DB,$93,$F6,$5D,$AF
	.dc.b	$80,$FD,$EB,$63,$F9,$6B,$5F,$E6,$8D,$7F,$9C,$B9,$E0,$FE,$A5,$4F
	.dc.b	$9C,$F3,$1F,$C2,$A7,$5B,$FF,$17,$97,$CE,$7F,$4E,$D1,$F3,$9F,$D6
	.dc.b	$31,$5F,$FA,$26,$30,$F1,$AF,$84,$71,$8C,$E4,$BE,$75,$A8,$FE,$47
	.dc.b	$0F,$9C,$FE,$A3,$87,$69,$E1,$18,$AF,$F4,$C3,$63,$F9,$56,$2F,$91
	.dc.b	$F4,$4E,$70,$E9,$58,$3E,$B4,$FF,$2A,$F8,$47,$8F,$49,$3F,$C8,$FE
	.dc.b	$23,$F8,$D3,$FC,$96,$76,$FF,$5E,$6B,$FD,$95,$AF,$F6,$A6,$BF,$EA
	.dc.b	$17,$79,$FF,$80,$A5,$E6,$7F,$3C,$6D,$77,$BF,$E7,$FF,$36,$F8,$AA
	.dc.b	$49,$FE,$71,$F3,$9F,$35,$FE,$75,$3B,$7F,$B9,$B5,$FE,$F4,$D7,$FB
	.dc.b	$FB,$5F,$F5,$B7,$CD,$BE,$2F,$A4,$7F,$81,$23,$1B,$7F,$7E,$6C,$7F
	.dc.b	$81,$5A,$FF,$82,$BD,$17,$8E,$F9,$CF,$EA,$AD,$9F,$F0,$64,$EA,$72
	.dc.b	$48,$7C,$5F,$A6,$75,$18,$D3,$FD,$A7,$7C,$EB,$4D,$8F,$F0,$C7,$C0
	.dc.b	$7F,$66,$A3,$48,$FF,$6A,$6B,$FE,$1A,$6B,$C8,$D1,$0D,$14,$D1,$82
	.dc.b	$18,$21,$FE,$2E,$7A,$B4,$E4,$53,$89,$4E,$27,$FB,$AD,$E2,$D3,$89
	.dc.b	$4E,$25,$38,$83,$E1,$9C,$0C,$3D,$1A,$64,$D9,$D0,$C9,$4C,$9D,$8A
	.dc.b	$7A,$48,$78,$CA,$72,$38,$3E,$17,$89,$B2,$1B,$38,$9A,$30,$68,$C1
	.dc.b	$1E,$84,$7D,$8D,$7D,$4D,$7C,$66,$88,$68,$A6,$8C,$10,$C1,$0F,$3B
	.dc.b	$D5,$A7,$22,$9C,$4A,$71,$3D,$8F,$16,$9C,$4A,$71,$29,$C4,$F8,$A4
	.dc.b	$3E,$21,$C4,$C3,$D5,$A6,$4D,$9D,$4C,$94,$C9,$DC,$A7,$A8,$87,$90
	.dc.b	$A7,$23,$83,$E1,$78,$9B,$21,$B3,$89,$A3,$06,$8C,$11,$E8,$47,$D8
	.dc.b	$D7,$D4,$D7,$C6,$68,$86,$8A,$68,$C1,$0C,$10,$F3,$BD,$5A,$72,$29
	.dc.b	$C4,$A7,$13,$D8,$F1,$69,$C4,$A7,$12,$9C,$4F,$8A,$43,$E2,$1C,$4C
	.dc.b	$3D,$5A,$64,$D9,$D4,$C9,$4C,$9D,$CA,$7A,$88,$7E,$EE,$DE,$6B,$8B
	.dc.b	$FD,$32,$D9,$FE,$8A,$D7,$FA,$63,$5F,$E3,$42,$79,$CF,$22,$47,$26
	.dc.b	$BC,$9A,$F2,$8E,$2E,$C8,$6C,$A6,$E7,$C4,$FF,$2B,$8E,$2E,$C8,$6E
	.dc.b	$7B,$BB,$9E,$2F,$A5,$B3,$CA,$38,$1E,$03,$84,$F2,$78,$36,$38,$46
	.dc.b	$8E,$06,$8A,$E9,$C1,$E6,$38,$14,$C9,$E2,$30,$F1,$24,$FA,$09,$3E
	.dc.b	$36,$CF,$81,$AF,$26,$BD,$48,$78,$8A,$78,$8C,$10,$C1,$27,$C7,$3E
	.dc.b	$12,$47,$85,$AF,$85,$AF,$68,$E2,$EC,$86,$CA,$6E,$7D,$2F,$8E,$38
	.dc.b	$BB,$21,$B9,$EE,$EE,$78,$BE,$96,$CF,$28,$E0,$78,$0E,$13,$C9,$E0
	.dc.b	$D8,$E1,$1A,$38,$1A,$2B,$A7,$07,$98,$E0,$53,$27,$88,$C3,$C4,$93
	.dc.b	$E8,$24,$F8,$DB,$3E,$06,$BC,$9A,$F5,$21,$E2,$29,$E2,$30,$43,$04
	.dc.b	$9F,$1C,$F8,$49,$1E,$16,$BE,$16,$BD,$A3,$8B,$B2,$1B,$29,$B9,$F4
	.dc.b	$BE,$38,$E2,$EC,$86,$E7,$BB,$B9,$E2,$FA,$5B,$3C,$A3,$81,$E0,$38
	.dc.b	$4F,$27,$83,$63,$84,$68,$E0,$68,$AE,$9C,$1E,$63,$81,$4C,$9E,$23
	.dc.b	$0F,$12,$4F,$A0,$93,$E3,$C3,$FC,$E5,$5F,$C9,$5A,$FF,$37,$6D,$7F
	.dc.b	$9F,$31,$FF,$3C,$E5,$FE,$81,$AF,$F4,$55,$7F,$A1,$DB,$19,$7F,$D0
	.dc.b	$2D,$7B,$35,$EC,$DA,$FF,$44,$C7,$23,$FD,$23,$3D,$5E,$66,$27,$84
	.dc.b	$66,$3A,$99,$21,$CC,$87,$22,$9C,$0A,$79,$0E,$26,$C8,$6C,$A6,$CE
	.dc.b	$33,$EA,$72,$D7,$81,$0E,$06,$A3,$D4,$E0,$D4,$71,$70,$47,$04,$8E
	.dc.b	$2D,$9F,$8A,$49,$F5,$9D,$27,$64,$8D,$47,$AA,$30,$F0,$3C,$0E,$1D
	.dc.b	$47,$07,$D4,$D7,$B3,$5E,$2D,$9F,$34,$75,$6B,$D0,$C4,$F7,$7A,$98
	.dc.b	$9E,$51,$98,$EE,$64,$87,$32,$1C,$8A,$70,$29,$E6,$38,$9B,$21,$B2
	.dc.b	$9B,$38,$CF,$A9,$CB,$5E,$04,$38,$1A,$8F,$53,$83,$51,$C5,$C1,$1C
	.dc.b	$12,$38,$B6,$7E,$29,$27,$D6,$74,$9D,$92,$35,$1E,$A8,$C3,$C0,$F0
	.dc.b	$38,$75,$1C,$1F,$53,$5E,$CD,$78,$B6,$7C,$D1,$D5,$AF,$43,$13,$DD
	.dc.b	$EA,$62,$79,$46,$63,$B9,$92,$1C,$C8,$72,$29,$C0,$A7,$98,$E2,$6C
	.dc.b	$86,$CA,$6C,$E3,$3E,$A7,$2D,$78,$10,$E0,$6A,$3D,$4E,$0D,$47,$17
	.dc.b	$04,$70,$48,$E2,$D9,$F8,$A4,$9F,$59,$D2,$76,$48,$D4,$7A,$A3,$0F
	.dc.b	$03,$C0,$E1,$D4,$70,$7D,$4D,$7B,$35,$E2,$D9,$F3,$47,$56,$BD,$0C
	.dc.b	$4F,$77,$A9,$89,$E5,$19,$8E,$E6,$48,$73,$21,$C8,$A7,$02,$9E,$63
	.dc.b	$89,$B2,$1B,$29,$B3,$8C,$FA,$9C,$B5,$E0,$43,$81,$A8,$F5,$38,$35
	.dc.b	$1C,$5C,$11,$C1,$23,$8B,$67,$E2,$92,$7D,$67,$49,$D9,$23,$51,$EA
	.dc.b	$8C,$3C,$0F,$03,$87,$51,$C1,$F5,$35,$EC,$D7,$DF,$F9,$70,$80

;--------------------------------------------------------------------------------
;	FSGLDIV.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsgldiv88x_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#11,-(sp)
	peamsg	'FSGLDIV.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSGLDIV.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsgldiv88x_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsgldiv.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSGLDIV.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsgldiv.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSGLDIV.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsgldiv88x_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3D,$61,$F3,$7F,$DA,$BD,$E7,$99,$C6,$30,$DA,$FE
	.dc.b	$50,$D9,$DC,$7E,$72,$3F,$B8,$BF,$98,$29,$F9,$E6,$B4,$CF,$9D,$C9
	.dc.b	$E2,$D7,$FB,$DB,$5E,$86,$A3,$9C,$66,$3C,$47,$48,$C9,$1C,$10,$E4
	.dc.b	$53,$BB,$B7,$24,$32,$6C,$F1,$98,$34,$60,$87,$90,$C9,$4E,$04,$36
	.dc.b	$7C,$53,$65,$34,$71,$37,$1F,$05,$D1,$88,$F5,$3C,$0C,$9C,$48,$60
	.dc.b	$91,$CC,$91,$CA,$3A,$1F,$04,$E3,$1A,$70,$72,$8E,$0E,$9A,$E4,$DC
	.dc.b	$71,$7D,$ED,$7D,$0D,$7C,$26,$A3,$9C,$66,$3B,$1D,$23,$24,$70,$43
	.dc.b	$91,$4E,$EE,$DC,$90,$C9,$B3,$C6,$60,$D1,$82,$1E,$43,$25,$38,$10
	.dc.b	$D9,$F1,$4D,$94,$D1,$C4,$DC,$7C,$17,$46,$23,$D4,$F0,$32,$71,$21
	.dc.b	$82,$47,$32,$47,$28,$E8,$7C,$13,$8C,$69,$C1,$CA,$38,$3A,$6B,$93
	.dc.b	$71,$C5,$F7,$B5,$F4,$35,$F0,$9A,$8E,$71,$98,$EC,$74,$8C,$91,$C1
	.dc.b	$0E,$45,$3B,$BB,$72,$43,$26,$CF,$19,$83,$46,$08,$79,$0C,$94,$E0
	.dc.b	$43,$67,$C5,$36,$53,$47,$13,$71,$F0,$5D,$18,$8F,$53,$F8,$1C,$BF
	.dc.b	$D1,$17,$FC,$F2,$3F,$A1,$35,$FE,$97,$87,$F4,$47,$F9,$F9,$A2,$9A
	.dc.b	$8E,$13,$D2,$32,$DC,$3F,$9E,$36,$FC,$93,$F2,$4F,$51,$0F,$49,$70
	.dc.b	$F6,$12,$3D,$71,$C8,$F8,$B1,$97,$A1,$C8,$8E,$5D,$C7,$27,$B3,$5D
	.dc.b	$B6,$38,$47,$D0,$32,$76,$72,$F8,$CD,$38,$7A,$9B,$8E,$84,$70,$43
	.dc.b	$CA,$53,$9B,$B7,$24,$32,$6C,$F1,$10,$F0,$9B,$30,$7A,$8C,$1A,$39
	.dc.b	$98,$34,$60,$87,$28,$F7,$3A,$31,$1E,$77,$67,$53,$27,$9C,$C1,$0E
	.dc.b	$45,$8E,$A4,$8E,$91,$C8,$F3,$46,$5E,$87,$22,$39,$77,$1C,$9E,$CD
	.dc.b	$76,$D9,$FC,$C4,$7E,$86,$BF,$37,$87,$E9,$5B,$87,$E8,$A3,$F5,$12
	.dc.b	$3F,$C5,$05,$3F,$54,$53,$F5,$47,$F8,$A2,$BB,$B8,$AE,$AF,$EB,$27
	.dc.b	$C1,$7F,$E4,$2C,$9F,$37,$FE,$2E,$C3,$FC,$CA,$D8,$CD,$F5,$6B,$E1
	.dc.b	$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7C,$51,$B7,$2D,$72,$71,$29,$A3
	.dc.b	$07,$F9,$61,$E0,$D3,$27,$F9,$7A,$3C,$0E,$4E,$44,$3A,$90,$D1,$4E
	.dc.b	$51,$D9,$A7,$69,$F8,$6F,$33,$89,$0D,$99,$29,$D5,$E2,$D9,$F3,$BA
	.dc.b	$39,$46,$9B,$1A,$8C,$46,$C8,$E0,$91,$B2,$4F,$AE,$32,$D8,$CC,$FB
	.dc.b	$5A,$FA,$67,$6E,$67,$81,$1C,$10,$E1,$3C,$9F,$74,$6D,$CB,$5C,$9C
	.dc.b	$4A,$68,$C1,$F0,$9E,$0D,$32,$7C,$28,$F0,$39,$39,$10,$EA,$43,$45
	.dc.b	$39,$47,$66,$9D,$A7,$E1,$BC,$CE,$24,$36,$64,$A7,$57,$8B,$67,$CE
	.dc.b	$E8,$E5,$1A,$6C,$6A,$31,$1B,$23,$82,$46,$C9,$3E,$B8,$CB,$63,$33
	.dc.b	$ED,$6B,$E9,$9D,$B9,$9E,$04,$70,$43,$84,$F2,$7D,$D1,$B7,$2D,$72
	.dc.b	$71,$29,$A3,$07,$C2,$78,$34,$C9,$F0,$A3,$C0,$E4,$E4,$43,$A9,$0D
	.dc.b	$14,$E5,$1D,$9A,$76,$9F,$86,$F3,$38,$90,$D9,$92,$9D,$5E,$2D,$9F
	.dc.b	$3B,$A3,$94,$7E,$F9,$FE,$8D,$F2,$DF,$EE,$44,$D2,$3F,$EE,$86,$BF
	.dc.b	$EE,$C6,$BF,$EF,$06,$BF,$D3,$5A,$FF,$51,$6B,$E0,$3A,$10,$E8,$53
	.dc.b	$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8,$A7,$22,$9C,$8A,$72,$38,$10
	.dc.b	$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$A6,$7B,$4A,$79
	.dc.b	$8A,$78,$4A,$76,$29,$D4,$A7,$57,$E9,$B8,$23,$82,$3E,$16,$BE,$16
	.dc.b	$BE,$16,$BE,$16,$BE,$16,$BE,$13,$A1,$0E,$85,$3A,$14,$E8,$53,$A1
	.dc.b	$4E,$87,$22,$1C,$8A,$72,$29,$C8,$A7,$23,$81,$0E,$05,$38,$14,$E0
	.dc.b	$68,$86,$8A,$68,$C1,$0C,$10,$FA,$67,$B4,$A7,$98,$A7,$84,$A7,$62
	.dc.b	$9D,$4A,$75,$7E,$9B,$82,$38,$23,$E1,$6B,$E1,$6B,$E1,$6B,$E1,$6B
	.dc.b	$E1,$6B,$E1,$3A,$10,$E8,$53,$A1,$4E,$85,$3A,$14,$E8,$72,$21,$C8
	.dc.b	$A7,$22,$9C,$8A,$72,$38,$10,$E0,$53,$81,$4E,$06,$88,$68,$A6,$8C
	.dc.b	$10,$FE,$09,$1F,$E7,$22,$61,$FE,$70,$8F,$F4,$16,$1F,$E7,$C6,$1F
	.dc.b	$E7,$C2,$8F,$83,$7C,$0B,$E3,$FF,$D0,$13,$FE,$85,$9C,$C6,$23,$84
	.dc.b	$ED,$DC,$E9,$FF,$4C,$9F,$E8,$DC,$DF,$09,$FE,$8D,$DD,$8F,$FA,$3E
	.dc.b	$73,$5F,$E9,$39,$CB,$C6,$BF,$D2,$9F,$88,$F7,$8F,$F4,$5C,$7F,$A5
	.dc.b	$FD,$27,$9F,$CE,$3F,$D3,$5F,$06,$F4,$31,$7F,$E9,$F8,$C3,$D4,$FF
	.dc.b	$51,$5F,$27,$01,$C3,$F1,$1A,$FF,$A9,$5A,$F8,$EF,$E3,$3C,$5F,$0B
	.dc.b	$4E,$B1,$D3,$2F,$83,$0A,$78,$7E,$8D,$DF,$BD,$7F,$AC,$EF,$E3,$C6
	.dc.b	$2F,$53,$9C,$FF,$D7,$9F,$2C,$FD,$45,$F0,$BA,$B8,$7C,$1C,$39,$3F
	.dc.b	$0E,$B8,$4F,$C8,$79,$78,$1E,$3F,$D8,$FC,$F9,$F8,$9B,$97,$83,$25
	.dc.b	$71,$6E,$3D,$EB,$6F,$E3,$DB,$87,$68,$FA,$3E,$B3,$D5,$E5,$E1,$C9
	.dc.b	$CC,$7C,$C9,$D4,$7E,$55,$AF,$E6,$9A,$FF,$B6,$DA,$FF,$B8,$7E,$7D
	.dc.b	$FA,$13,$CF,$F4,$7B,$72,$BF,$0D,$5A,$9F,$C5,$B6,$3E,$46,$8A,$E1
	.dc.b	$1F,$96,$13,$B7,$0D,$DF,$F4,$2D,$DF,$C9,$8F,$8E,$BE,$A7,$EA,$7F
	.dc.b	$A1,$BD,$F9,$4F,$3F,$EB,$4B,$E9,$29,$E6,$9D,$5F,$A9,$AF,$97,$85
	.dc.b	$5C,$63,$BD,$71,$B9,$1C,$5A,$F1,$97,$BB,$E2,$F7,$7C,$7E,$9C,$AF
	.dc.b	$4C,$65,$F5,$C6,$5E,$51,$96,$9C,$0F,$83,$E5,$FD,$0F,$D2,$9F,$11
	.dc.b	$F0,$3E,$2A,$57,$D6,$9C,$B6,$72,$F9,$E7,$94,$48,$D1,$0F,$67,$A1
	.dc.b	$F3,$3C,$CF,$9B,$E5,$49,$FA,$71,$97,$EA,$C6,$5F,$89,$19,$69,$C0
	.dc.b	$FB,$7F,$31,$F6,$3E,$72,$47,$DE,$7D,$DF,$0D,$23,$04,$3F,$13,$9A
	.dc.b	$3F,$03,$F0,$3B,$9C,$C6,$1F,$C4,$F6,$FE,$18,$8F,$E3,$3F,$47,$FC
	.dc.b	$4F,$EA,$1F,$8A,$64,$7E,$42,$3F,$1B,$E3,$C8,$FC,$9B,$F9,$0F,$2B
	.dc.b	$F8,$EF,$A2,$E3,$FC,$9F,$99,$2F,$F9,$5C,$E5,$B3,$97,$F1,$53,$CA
	.dc.b	$24,$7F,$4F,$9E,$0F,$F5,$46,$A2,$07,$2F,$F2,$2E,$8F,$CF,$F6,$FF
	.dc.b	$4E,$EC,$FC,$FF,$C0,$7E,$A0,$FD,$4F,$64,$BF,$D7,$C6,$5F,$D8,$C7
	.dc.b	$F2,$3F,$D5,$3F,$97,$42,$90,$FA,$1D,$1B,$98,$7F,$2D,$8D,$36,$32
	.dc.b	$FE,$CE,$1F,$3B,$DF,$E7,$64,$F1,$8F,$11,$87,$84,$61,$D1,$DD,$D6
	.dc.b	$5F,$06,$BA,$CF,$F1,$CF,$D2,$5F,$39,$ED,$F9,$E9,$39,$8F,$E3,$F3
	.dc.b	$A8,$FF,$02,$F9,$08,$C4,$7D,$23,$F8,$FB,$87,$EA,$C6,$E7,$FC,$15
	.dc.b	$A7,$C8,$F9,$EF,$99,$EF,$CB,$F8,$D1,$F3,$CC,$3F,$16,$30,$FE,$33
	.dc.b	$BD,$A7,$E0,$FE,$AD,$5F,$D8,$B5,$7E,$15,$F6,$4F,$CB,$DF,$0C,$17
	.dc.b	$C1,$FC,$67,$2A,$75,$39,$CF,$EF,$DF,$3B,$F1,$65,$E1,$CB,$F2,$78
	.dc.b	$F1,$8F,$CB,$79,$3D,$6B,$F2,$DE,$27,$E6,$23,$14,$8C,$5F,$2C,$FC
	.dc.b	$D5,$A7,$C7,$1D,$A7,$F4,$39,$F2,$6C,$71,$C7,$C9,$8F,$B2,$F8,$BE
	.dc.b	$EA,$DB,$4D,$9F,$AB,$8D,$B6,$73,$8F,$9E,$2D,$72,$74,$7C,$CB,$F1
	.dc.b	$3F,$D5,$5F,$43,$F1,$1A,$E4,$F3,$56,$5A,$73,$3C,$C5,$3E,$59,$E4
	.dc.b	$BF,$3D,$FD,$18,$D3,$E7,$6D,$F9,$E3,$58,$7C,$DB,$F0,$57,$43,$C8
	.dc.b	$5C,$74,$E2,$B8,$E3,$EA,$BF,$A5,$3E,$3B,$8E,$F0,$FB,$AF,$AB,$0E
	.dc.b	$67,$4A,$E3,$1E,$63,$8B,$AB,$F8,$F3,$A3,$13,$F1,$A7,$89,$27,$E8
	.dc.b	$BA,$8F,$54,$72,$3E,$AC,$F8,$5F,$51,$23,$A3,$C8,$C3,$F1,$67,$83
	.dc.b	$A3,$EA,$BE,$17,$07,$E1,$27,$D1,$3E,$18,$FB,$B1,$C9,$E9,$1D,$B0
	.dc.b	$F9,$B1,$A8,$E0,$FA,$8C,$1D,$A3,$EF,$57,$86,$3F,$26,$F5,$8C,$CF
	.dc.b	$5C,$7F,$0E,$7E,$23,$0F,$C2,$1F,$85,$73,$8F,$E1,$2B,$F1,$58,$FA
	.dc.b	$B1,$F0,$9D,$B0,$F4,$61,$C6,$FE,$CC,$62,$69,$C2,$F9,$C7,$B1,$FC
	.dc.b	$C5,$F9,$EF,$C7,$5C,$0E,$C7,$CE,$AF,$55,$E6,$FD,$B3,$93,$E2,$61
	.dc.b	$ED,$7F,$6E,$6D,$C3,$F8,$F3,$F4,$EF,$C0,$71,$8F,$CB,$9F,$A7,$1E
	.dc.b	$23,$EB,$9D,$CC,$3A,$3C,$6E,$8C,$CF,$82,$7A,$3D,$FD,$5F,$F2,$94
	.dc.b	$FF,$0A,$6D,$7E,$ED,$B3,$FC,$E5,$DC,$FF,$80,$FD,$EF,$F0,$24,$62
	.dc.b	$11,$8C,$FF,$DB,$AD,$8F,$EB,$0D,$7F,$C1,$6F,$F2,$27,$FB,$E7,$9D
	.dc.b	$5E,$67,$F1,$13,$25,$33,$8F,$F3,$F7,$2D,$73,$9F,$F6,$7E,$7F,$F7
	.dc.b	$0C,$E6,$FC,$11,$89,$F3,$D7,$78,$C3,$E8,$31,$7E,$5A,$E2,$D8,$C9
	.dc.b	$E5,$8C,$B4,$CD,$F8,$5B,$7E,$08,$F5,$5F,$82,$7E,$29,$A3,$C5,$1A
	.dc.b	$BF,$03,$63,$B3,$4E,$C5,$AF,$6E,$1B,$3C,$45,$7C,$47,$C7,$AE,$AD
	.dc.b	$AE,$AD,$AE,$AD,$3A,$9B,$AE,$AD,$3A,$CF,$F9,$2A,$FD,$CF,$E2,$CF
	.dc.b	$71,$F0,$2B,$D0,$FC,$09,$EC,$D3,$95,$F9,$9A,$7C,$FB,$F6,$3F,$65
	.dc.b	$AF,$59,$E6,$D3,$8D,$7A,$1E,$73,$5F,$F3,$15,$AF,$FA,$CC,$7A,$6B
	.dc.b	$CE,$F3,$37,$1E,$56,$9F,$28,$EF,$7F,$11,$CC,$FC,$C7,$27,$A6,$3E
	.dc.b	$53,$4F,$2B,$C5,$FD,$0B,$6B,$E6,$C7,$EA,$AF,$E0,$BF,$3D,$AF,$E9
	.dc.b	$8E,$E7,$A4,$E6,$77,$3D,$2F,$C3,$7F,$66,$F9,$DF,$4C,$F8,$6F,$F4
	.dc.b	$93,$DB,$0F,$8F,$7F,$26,$FF,$BF,$9F,$26,$FE,$9B,$4F,$19,$4E,$A7
	.dc.b	$E0,$4A,$7E,$9A,$BB,$3D,$27,$B4,$79,$63,$D6,$7B,$63,$E2,$D7,$ED
	.dc.b	$B0,$D4,$62,$7F,$26,$78,$4E,$27,$C9,$BF,$A6,$F8,$5A,$FA,$77,$FF
	.dc.b	$4B,$F9,$1E,$FD,$BE,$04,$FE,$3E,$73,$9F,$FA,$B6,$73,$3C,$78,$7F
	.dc.b	$DE,$33,$93,$F5,$35,$C2,$7F,$D3,$5F,$2D,$D6,$7F,$CE,$5F,$0F,$67
	.dc.b	$FD,$51,$6A,$C5,$FF,$AA,$AA,$CE,$6F,$FD,$51,$18,$AC,$BC,$F9,$BF
	.dc.b	$D2,$5F,$3D,$E6,$C7,$FD,$22,$D8,$CD,$FF,$A6,$3A,$13,$96,$C7,$FA
	.dc.b	$D7,$A3,$A4,$62,$FF,$D6,$91,$E4,$FA,$0F,$2F,$CC,$5B,$FF,$5E,$36
	.dc.b	$BF,$D7,$4D,$DB,$FD,$60,$D9,$FF,$4A,$36,$7E,$1B,$F0,$3E,$63,$E0
	.dc.b	$FC,$8D,$AF,$F6,$43,$6B,$FD,$8C,$DB,$FF,$5E,$3F,$23,$E8,$3E,$4F
	.dc.b	$CC,$DB,$FF,$68,$36,$FF,$D9,$CD,$BF,$F6,$23,$F5,$63,$FD,$D9,$E0
	.dc.b	$7C,$B6,$D7,$FA,$F5,$FF,$77,$C6,$53,$4F,$B3,$E6,$73,$8C,$CF,$FB
	.dc.b	$86,$31,$3A,$8F,$C5,$4F,$47,$47,$FB,$9F,$E6,$7F,$06,$7E,$13,$82
	.dc.b	$C7,$FB,$AF,$E8,$3F,$13,$9F,$F2,$18,$C1,$D3,$99,$19,$C3,$FD,$8F
	.dc.b	$27,$FD,$F5,$0F,$A0,$E6,$74,$9A,$73,$87,$C5,$FE,$47,$7F,$F0,$0E
	.dc.b	$FF,$C2,$71,$EF,$39,$EB,$98,$71,$F8,$6B,$9F,$88,$F0,$BE,$A2,$47
	.dc.b	$C4,$E7,$65,$DA,$3D,$18,$FB,$2B,$E0,$FC,$5C,$9F,$37,$C9,$DC,$7B
	.dc.b	$5F,$C2,$AF,$8F,$F0,$59,$D7,$D1,$B2,$31,$8F,$68,$C3,$F0,$6B,$8B
	.dc.b	$E5,$CD,$D3,$F0,$CF,$A7,$F4,$16,$75,$1F,$5B,$E2,$71,$A7,$93,$14
	.dc.b	$63,$3F,$1B,$6F,$C4,$FD,$EF,$9C,$91,$F4,$FE,$36,$E1,$CE,$7F,$07
	.dc.b	$F3,$98,$C3,$DA,$DB,$F6,$36,$BF,$02,$FD,$BE,$69,$3F,$7F,$E5,$2D
	.dc.b	$7D,$7E,$4F,$A7,$5F,$90,$F8,$8C,$E3,$F2,$9B,$87,$C9,$6D,$FE,$39
	.dc.b	$FE,$31,$FA,$53,$D7,$A7,$E6,$BE,$1A,$DF,$E0,$67,$4F,$E4,$7E,$57
	.dc.b	$38,$7C,$96,$E3,$F6,$9B,$3F,$65,$FD,$E7,$BD,$F9,$88,$FD,$D5,$23
	.dc.b	$0F,$EF,$5A,$FE,$94,$FD,$33,$87,$0F,$0A,$FC,$4F,$D0,$F6,$ED,$FD
	.dc.b	$9B,$FB,$4F,$8F,$FE,$81,$E8,$BF,$64,$7C,$4E,$73,$FE,$03,$39,$8C
	.dc.b	$67,$FA,$58,$C4,$FE,$C6,$F9,$DF,$28,$E3,$FB,$DB,$AF,$BB,$63,$F7
	.dc.b	$1F,$23,$89,$91,$C1,$FE,$13,$18,$C7,$F6,$83,$E9,$DF,$0D,$FC,$95
	.dc.b	$B7,$FC,$63,$EA,$9F,$85,$F7,$BE,$35,$A6,$DF,$F2,$B8,$C4,$7C,$D3
	.dc.b	$B4,$64,$97,$FC,$07,$F0,$6F,$74,$F2,$8C,$6F,$FC,$DF,$EA,$9E,$65
	.dc.b	$3F,$B2,$7D,$13,$E4,$B1,$7F,$CB,$B0,$5F,$08,$FC,$1B,$63,$F9,$F3
	.dc.b	$5F,$33,$5F,$C2,$7C,$DB,$CC,$79,$BF,$8C,$8C,$36,$33,$3F,$D2,$67
	.dc.b	$2F,$F4,$BF,$80,$7C,$8B,$E7,$3F,$2B,$18,$6C,$66,$78,$35,$F9,$54
	.dc.b	$F3,$5D,$5E,$36,$CF,$85,$AF,$06,$BF,$9F,$F9,$C7,$95,$FA,$1E,$FF
	.dc.b	$F1,$2F,$CE,$7C,$BA,$F9,$4D,$7F,$0F,$5C,$1F,$9C,$D7,$FB,$43,$5F
	.dc.b	$C0,$35,$F0,$57,$ED,$1F,$99,$F1,$DE,$1E,$1F,$D9,$79,$BF,$C1,$E7
	.dc.b	$ED,$35,$FC,$BF,$02,$B8,$47,$F5,$D8,$C3,$C3,$C7,$CB,$FC,$C1,$B1
	.dc.b	$A8,$E7,$3E,$38,$C4,$7E,$5D,$AF,$F0,$56,$BF,$43,$E7,$3F,$39,$A7
	.dc.b	$F9,$AF,$64,$63,$2F,$F7,$AB,$69,$E9,$BE,$69,$E2,$7A,$B0,$E3,$56
	.dc.b	$F8,$46,$23,$FB,$F5,$70,$6F,$83,$C7,$E7,$3F,$CB,$B9,$F8,$DB,$9F
	.dc.b	$F6,$AB,$57,$C8,$CF,$C4,$F1,$99,$87,$43,$F5,$5E,$47,$F6,$77,$C3
	.dc.b	$F3,$9C,$70,$3F,$3A,$75,$3F,$C3,$1B,$97,$F8,$1E,$B7,$3C,$8E,$BE
	.dc.b	$B7,$DA,$F4,$1F,$36,$F4,$3B,$7C,$E4,$BF,$F0,$34,$E5,$B3,$97,$FC
	.dc.b	$03,$82,$BA,$56,$E7,$FA,$E3,$5F,$E8,$17,$E0,$BF,$E8,$ED,$8F,$F2
	.dc.b	$13,$5F,$E3,$4B,$D1,$7C,$8F,$CB,$F9,$37,$C6,$53,$D9,$3A,$BF,$F0
	.dc.b	$F3,$5F,$F0,$BC,$F0,$7F,$C7,$4D,$7F,$B9,$46,$F5,$FF,$5C,$61,$E1
	.dc.b	$C3,$A9,$2B,$8F,$8B,$E3,$F9,$CF,$27,$93,$27,$AC,$65,$EF,$19,$7C
	.dc.b	$71,$96,$9C,$0F,$2F,$CA,$FD,$7F,$99,$7C,$67,$C1,$F9,$C9,$5F,$E3
	.dc.b	$B9,$CB,$67,$2F,$C0,$9E,$51,$23,$44,$3E,$2F,$C9,$7C,$DF,$92,$F9
	.dc.b	$DE,$4C,$9F,$8F,$19,$7E,$4C,$65,$F9,$B1,$96,$9C,$0F,$9F,$F3,$9F
	.dc.b	$67,$E7,$24,$7D,$27,$DD,$F1,$32,$30,$43,$FC,$4D,$9A,$3F,$05,$EA
	.dc.b	$6E,$73,$18,$7E,$EF,$67,$E1,$C8,$FE,$03,$E7,$3F,$15,$F3,$9F,$8B
	.dc.b	$64,$7E,$12,$3F,$1D,$F3,$92,$3F,$0E,$FE,$47,$C9,$FE,$3F,$E9,$38
	.dc.b	$FF,$29,$E5,$CB,$FE,$C9,$39,$6C,$E5,$FC,$9C,$F2,$89,$1F,$94,$8D
	.dc.b	$75,$7F,$BA,$AB,$94,$FE,$6B,$A7,$F4,$5F,$39,$23,$F3,$71,$FA,$53
	.dc.b	$0F,$E8,$23,$0F,$FA,$67,$E5,$7F,$93,$79,$07,$FB,$5C,$7F,$9F,$1B
	.dc.b	$87,$FA,$45,$D9,$3D,$5F,$E3,$8E,$BC,$8F,$F4,$4B,$AF,$9C,$CF,$C7
	.dc.b	$CC,$3C,$11,$C0,$C3,$DE,$30,$E8,$F3,$3A,$C3,$C5,$3F,$C6,$7E,$53
	.dc.b	$FD,$0B,$5E,$BF,$9C,$F1,$99,$F8,$D9,$19,$21,$FE,$6F,$9E,$2F,$BF
	.dc.b	$E7,$3D,$7F,$33,$23,$E3,$47,$B8,$C3,$F1,$63,$0F,$DE,$EE,$65,$FE
	.dc.b	$C1,$F9,$A5,$7C,$4F,$9C,$F8,$E7,$E2,$2B,$76,$AD,$BF,$EC,$9E,$44
	.dc.b	$EA,$73,$C5,$F5,$EF,$9D,$F8,$71,$F0,$63,$FE,$6D,$C7,$8C,$62,$7F
	.dc.b	$B1,$F0,$2F,$8C,$7F,$95,$2B,$6F,$8E,$F9,$65,$E1,$7A,$78,$5F,$E7
	.dc.b	$9C,$FC,$F4,$B4,$E6,$B8,$E5,$E1,$AD,$61,$CD,$F7,$56,$DA,$6E,$FC
	.dc.b	$78,$76,$6E,$1E,$29,$D5,$71,$7A,$E1,$C7,$0E,$D1,$E5,$9D,$5F,$9A
	.dc.b	$35,$51,$F8,$07,$88,$A7,$13,$58,$78,$EF,$AD,$73,$9D,$B6,$75,$5C
	.dc.b	$30,$F4,$56,$B0,$F1,$E1,$C8,$E6,$F9,$8C,$D7,$19,$F2,$5F,$C3,$9C
	.dc.b	$C7,$82,$B8,$BF,$10,$E0,$7C,$43,$C0,$64,$E1,$5A,$C3,$D1,$87,$9B
	.dc.b	$0F,$4D,$68,$E2,$47,$24,$8E,$04,$8F,$CD,$5F,$B6,$FC,$15,$E9,$9F
	.dc.b	$04,$FA,$1E,$95,$F8,$87,$C1,$3B,$72,$74,$AF,$46,$1F,$1E,$39,$C7
	.dc.b	$6A,$F1,$47,$C1,$AF,$33,$F2,$A3,$38,$7E,$32,$3A,$1D,$8F,$C0,$1E
	.dc.b	$28,$F2,$4F,$8C,$8E,$09,$1D,$48,$78,$5C,$BD,$23,$8B,$E0,$7C,$AE
	.dc.b	$63,$CB,$87,$C9,$8F,$25,$7D,$67,$84,$F4,$8C,$47,$13,$24,$30,$47
	.dc.b	$E0,$3F,$85,$F2,$FE,$71,$27,$E9,$3E,$49,$EA,$4A,$F2,$5F,$AA,$FA
	.dc.b	$46,$2F,$E6,$E3,$F9,$6A,$E4,$FD,$7A,$F9,$87,$96,$FA,$C6,$EB,$D8
	.dc.b	$EE,$BC,$57,$FC,$A3,$D4,$FF,$1C,$5F,$F1,$86,$C6,$71,$FC,$54,$7F
	.dc.b	$8F,$71,$0F,$9C,$B1,$98,$C1,$FE,$BE,$9F,$F7,$53,$88,$D4,$F0,$9C
	.dc.b	$BF,$C7,$67,$FD,$AF,$CB,$5E,$78,$FF,$26,$72,$7F,$0C,$3C,$73,$C7
	.dc.b	$99,$5C,$63,$0D,$9E,$13,$A7,$91,$DB,$4F,$F8,$01,$19,$BF,$04,$E4
	.dc.b	$F8,$27,$89,$CD,$F6,$8E,$6D,$8C,$C7,$82,$FC,$51,$EB,$AE,$D7,$F0
	.dc.b	$AB,$47,$23,$FC,$EA,$F9,$2B,$E8,$E1,$D0,$E4,$E2,$39,$5E,$2B,$93
	.dc.b	$4F,$19,$63,$67,$58,$DB,$63,$6D,$37,$1F,$0A,$FD,$2F,$C4,$3E,$31
	.dc.b	$F8,$67,$2F,$C8,$C3,$E2,$4F,$28,$D5,$FC,$48,$ED,$1C,$E7,$6D,$3E
	.dc.b	$2D,$F9,$1F,$15,$71,$69,$EA,$2C,$6D,$EE,$64,$AE,$5F,$B9,$5E,$56
	.dc.b	$9E,$18,$F9,$0D,$8F,$3B,$ED,$3C,$AF,$1B,$F8,$31,$F9,$63,$27,$13
	.dc.b	$CB,$1B,$AF,$54,$79,$23,$0F,$9C,$F1,$9D,$EB,$C6,$D8,$F1,$BD,$E3
	.dc.b	$2D,$8C,$BF,$58,$E6,$64,$F1,$9E,$53,$C4,$53,$67,$1A,$F6,$B6,$BD
	.dc.b	$2E,$AF,$EC,$BE,$53,$A9,$4E,$A5,$38,$1F,$18,$C4,$F6,$8E,$71,$E7
	.dc.b	$38,$94,$E3,$7F,$01,$F2,$1F,$26,$79,$3D,$4E,$0E,$23,$84,$FF,$8F
	.dc.b	$F8,$3F,$CF,$9E,$37,$F0,$7D,$FF,$84,$61,$EB,$9C,$E3,$FC,$92,$73
	.dc.b	$3F,$2E,$72,$FF,$97,$27,$F5,$31,$89,$DC,$FF,$0D,$8C,$4E,$B8,$7F
	.dc.b	$DA,$EF,$FA,$3F,$E7,$3C,$4F,$FA,$1B,$E7,$2D,$FF,$95,$26,$DF,$FA
	.dc.b	$4E,$6D,$A6,$C6,$1B,$7E,$93,$F9,$EF,$9A,$E6,$7C,$C7,$2E,$8E,$F3
	.dc.b	$98,$C5,$71,$CD,$39,$8F,$F5,$35,$70,$BE,$71,$83,$FD,$55,$F3,$9E
	.dc.b	$6F,$9C,$F3,$B6,$32,$7F,$AB,$E3,$2D,$33,$87,$FA,$BE,$32,$D8,$FF
	.dc.b	$5B,$7C,$E7,$C2,$F9,$9F,$86,$D8,$C9,$FE,$BB,$8C,$B4,$FF,$5F,$7C
	.dc.b	$CF,$CA,$F9,$CF,$96,$D8,$C9,$FE,$C3,$8C,$B4,$F8,$97,$FD,$A3,$C1
	.dc.b	$FA,$7E,$5F,$D8,$8C,$4F,$FB,$2E,$31,$09,$DC,$F2,$8C,$3B,$31,$1A
	.dc.b	$8F,$C7,$1F,$D5,$9C,$72,$7F,$B4,$7E,$2F,$F0,$67,$E1,$38,$6D,$FF
	.dc.b	$B6,$FE,$5F,$F1,$39,$FF,$6A,$8C,$1F,$77,$99,$19,$CB,$BC,$63,$E7
	.dc.b	$38,$9C,$AA,$9C,$69,$F2,$DF,$8A,$DF,$FB,$DE,$9F,$DC,$F0,$EB,$39
	.dc.b	$EC,$98,$71,$F8,$5B,$9F,$83,$24,$7A,$09,$3F,$93,$E7,$67,$DA,$3F
	.dc.b	$C3,$39,$7A,$AB,$DF,$F1,$52,$7C,$DF,$39,$72,$ED,$3F,$13,$E0,$B1
	.dc.b	$B7,$9F,$44,$63,$0E,$B1,$87,$E0,$57,$17,$C9,$9B,$AB,$E1,$1F,$43
	.dc.b	$E5,$2C,$EA,$3E,$9F,$C4,$E3,$3F,$80,$DC,$FC,$4D,$AF,$0B,$F6,$BE
	.dc.b	$72,$47,$D1,$F8,$BB,$7F,$26,$7E,$FF,$CE,$63,$0F,$63,$6F,$D6,$DA
	.dc.b	$70,$FE,$0A,$3E,$D7,$34,$AF,$BD,$F3,$96,$FF,$27,$C9,$F4,$AB,$F1
	.dc.b	$BF,$0B,$9C,$BE,$53,$70,$F9,$2D,$BE,$FC,$48,$FE,$23,$F3,$8F,$57
	.dc.b	$F9,$4E,$48,$C6,$1F,$E5,$08,$C4,$EA,$73,$87,$F8,$5E,$73,$5F,$95
	.dc.b	$F9,$CE,$B7,$FA,$26,$C6,$4F,$C6,$5F,$2C,$BF,$05,$3F,$BB,$F7,$3F
	.dc.b	$3B,$5F,$9F,$F9,$CC,$76,$7E,$B1,$B1,$93,$F8,$2C,$F0,$CF,$F3,$93
	.dc.b	$FB,$2F,$90,$FE,$7B,$E8,$3E,$71,$F0,$F9,$C7,$F9,$84,$E6,$31,$87
	.dc.b	$F0,$28,$C4,$FE,$C6,$F9,$DF,$28,$E3,$F3,$8E,$BE,$ED,$8F,$DC,$7C
	.dc.b	$86,$26,$47,$07,$F8,$3C,$63,$1F,$E2,$23,$E6,$9F,$09,$FD,$7A,$30
	.dc.b	$D8,$CD,$FF,$A1,$A7,$2D,$8C,$C7,$E2,$BE,$6B,$D0,$F6,$8C,$F1,$7F
	.dc.b	$9E,$F9,$55,$A8,$FE,$EF,$3A,$BF,$EB,$B3,$97,$FC,$F7,$19,$BF,$EB
	.dc.b	$33,$97,$FD,$3F,$F3,$2F,$27,$D4,$DB,$F4,$36,$7F,$D3,$ED,$7F,$CC
	.dc.b	$8E,$70,$FF,$05,$7E,$34,$D3,$FA,$14,$E6,$FF,$C1,$53,$97,$FA,$4F
	.dc.b	$CC,$FF,$1E,$F2,$7E,$33,$71,$FF,$56,$B5,$FF,$39,$35,$FF,$4B,$FB
	.dc.b	$8F,$8C,$7C,$D7,$CF,$8C,$36,$33,$3F,$EB,$89,$CB,$DB,$C9,$73,$F8
	.dc.b	$9B,$19,$8C,$1F,$6F,$E9,$DE,$8F,$FA,$A3,$83,$FC,$53,$18,$AF,$EC
	.dc.b	$AD,$7F,$13,$EB,$FF,$61,$73,$5C,$76,$FF,$58,$D6,$EF,$9C,$E6,$BF
	.dc.b	$DA,$B7,$C5,$FC,$9F,$CD,$3C,$AF,$C7,$42,$30,$FE,$45,$AF,$FB,$61
	.dc.b	$AF,$FA,$75,$CD,$FF,$90,$9B,$8F,$FB,$11,$B3,$FE,$E6,$F9,$BF,$8A
	.dc.b	$F8,$AF,$CC,$F8,$5B,$3C,$21,$3C,$2F,$FA,$77,$CC,$FB,$7C,$BF,$EE
	.dc.b	$2D,$F2,$0D,$F9,$8C,$33,$0F,$EE,$0E,$5F,$EF,$5F,$AB,$F5,$D3,$FB
	.dc.b	$1A,$F8,$3E,$CD,$9D,$F0,$E0,$CC,$C8,$C9,$FC,$6B,$6F,$9B,$E3,$7C
	.dc.b	$88,$FF,$73,$B5,$07,$0D,$8F,$8B,$1F,$0C,$F9,$7A,$30,$7C,$B6,$BD
	.dc.b	$4F,$89,$F3,$5F,$0B,$1F,$F0,$B3,$5F,$E2,$78,$7F,$76,$7F,$97,$CF
	.dc.b	$98,$8F,$F1,$18,$D3,$FE,$DB,$3C,$D1,$DD,$E2,$7C,$AF,$97,$FC,$87
	.dc.b	$86,$F9,$87,$CA,$7C,$AF,$2E,$5F,$F8,$4E,$72,$D9,$FE,$41,$5C,$5B
	.dc.b	$3A,$9F,$EA,$CD,$7E,$49,$FE,$6E,$75,$CB,$FC,$85,$BC,$5F,$E7,$A6
	.dc.b	$CF,$F7,$8D,$9F,$23,$E8,$97,$C8,$FE,$37,$E6,$7F,$01,$7F,$E3,$C6
	.dc.b	$E3,$D6,$78,$1E,$62,$4F,$F6,$FA,$E6,$D8,$DB,$F0,$5E,$DC,$3F,$23
	.dc.b	$E5,$BF,$6D,$7F,$E4,$1C,$9F,$33,$F9,$23,$5F,$05,$31,$FE,$F9,$7F
	.dc.b	$BD,$7E,$1E,$E8,$C4,$7F,$A6,$9A,$FF,$A5,$1E,$FB,$66,$BA,$4F,$F9
	.dc.b	$6B,$07,$C7,$F9,$7E,$2F,$E2,$65,$2B,$FC,$99,$8A,$75,$1E,$47,$ED
	.dc.b	$C6,$63,$04,$3F,$C4,$6E,$5B,$1F,$46,$BF,$63,$F1,$F2,$7F,$6E,$FC
	.dc.b	$7F,$88,$91,$F0,$9E,$E4,$7F,$31,$D3,$FA,$7F,$1F,$71,$E0,$9F,$DE
	.dc.b	$C7,$EF,$BE,$66,$4F,$F0,$47,$F0,$5F,$33,$F5,$7C,$6F,$59,$27,$E8
	.dc.b	$BF,$DD,$F0,$73,$7F,$7E,$9F,$F4,$33,$5F,$C5,$7A,$7F,$C3,$2F,$F4
	.dc.b	$FF,$A8,$3E,$23,$F8,$7C,$CA,$FF,$3C,$4E,$5B,$3F,$A6,$8C,$47,$F8
	.dc.b	$E5,$AF,$E0,$9A,$FF,$B6,$9D,$F3,$7F,$81,$5B,$CB,$9A,$FE,$3F,$DF
	.dc.b	$FC,$83,$B3,$F9,$0F,$CD,$4C,$7F,$93,$C6,$5F,$E5,$31,$97,$F1,$11
	.dc.b	$FC,$A3,$E6,$BF,$94,$FC,$D4,$9F,$E5,$F1,$97,$F9,$8C,$65,$FC,$84
	.dc.b	$7F,$30,$F8,$1F,$E6,$3D,$D2,$7F,$9B,$C6,$5F,$E7,$31,$97,$F3,$11
	.dc.b	$97,$F3,$51,$9D,$7F,$59,$3F,$43,$E1,$3F,$9E,$FC,$D4,$BF,$9F,$19
	.dc.b	$69,$0F,$F5,$EF,$EB,$1F,$C4,$1B,$EB,$7F,$53,$87,$CD,$D3,$11,$27
	.dc.b	$FD,$91,$3A,$6C,$E5,$FF,$23,$FA,$BF,$C5,$1F,$C4,$E3,$FC,$4C,$FE
	.dc.b	$CF,$19,$8D,$F8,$DF,$DE,$A3,$11,$E5,$6B,$FE,$20,$8D,$1C,$88,$F9
	.dc.b	$3C,$3F,$ED,$9F,$2B,$ED,$D7,$FD,$A4,$B3,$C9,$AA,$87,$FB,$8E,$34
	.dc.b	$D8,$CB,$FE,$A0,$F9,$4F,$F5,$F3,$BF,$9B,$EE,$7F,$48,$8C,$C6,$FE
	.dc.b	$3B,$04,$7E,$4C,$FC,$2F,$09,$8F,$F6,$AF,$98,$7C,$CF,$FB,$73,$37
	.dc.b	$CD,$7B,$0E,$15,$B9,$FF,$21,$DF,$2C,$7F,$C0,$11,$FE,$46,$F1,$BF
	.dc.b	$A6,$FA,$7D,$27,$FB,$BF,$0F,$F2,$47,$FB,$CC,$65,$42,$FA,$4E,$63
	.dc.b	$18,$78,$6B,$9D,$FF,$5A,$CF,$4D,$C7,$CB,$5B,$AE,$B8,$F9,$EF,$71
	.dc.b	$9C,$3C,$8A,$3C,$13,$BB,$E7,$39,$7D,$CF,$1C,$3C,$0F,$BA,$78,$37
	.dc.b	$0D,$30,$E4,$64,$A7,$99,$D5,$F9,$A3,$38,$7C,$26,$D7,$C6,$C3,$C7
	.dc.b	$1F,$48,$F8,$A7,$CE,$7D,$38,$FB,$30,$EB,$1C,$30,$C5,$74,$9C,$BC
	.dc.b	$8C,$12,$3C,$84,$9F,$8A,$F2,$AF,$41,$23,$C2,$68,$F0,$12,$7E,$14
	.dc.b	$F1,$8F,$5C,$F1,$9F,$8C,$F9,$A7,$C1,$1E,$57,$13,$D6,$30,$F5,$7E
	.dc.b	$61,$E9,$71,$87,$E0,$27,$8C,$F6,$9F,$09,$1C,$1E,$67,$8E,$3F,$80
	.dc.b	$8E,$B3,$C9,$FC,$DC,$66,$BB,$57,$96,$7C,$77,$ED,$76,$7D,$73,$B9
	.dc.b	$D2,$3B,$BF,$97,$C7,$F1,$AE,$4F,$79,$0C,$1D,$A3,$D4,$F6,$3F,$0C
	.dc.b	$64,$F6,$1C,$6B,$CE,$E0,$E6,$E1,$E5,$87,$BE,$3F,$28,$FD,$48,$CC
	.dc.b	$FB,$E7,$F3,$17,$EF,$7F,$5E,$FD,$F9,$ED,$1D,$49,$7E,$E7,$8D,$7A
	.dc.b	$9E,$2F,$E3,$27,$E3,$C6,$1F,$CC,$57,$57,$F0,$3E,$A7,$F9,$52,$7F
	.dc.b	$54,$DB,$FE,$34,$DC,$BF,$D1,$AF,$F9,$6A,$BF,$D1,$FF,$35,$8C,$3F
	.dc.b	$CF,$8D,$8F,$EB,$8D,$7F,$CD,$AD,$78,$9F,$40,$FE,$89,$E0,$7D,$F7
	.dc.b	$66,$78,$91,$98,$E3,$5F,$C9,$DE,$27,$31,$5B,$AE,$31,$98,$FE,$44
	.dc.b	$F0,$D3,$93,$C0,$FE,$22,$67,$4F,$6B,$E3,$C3,$C4,$EF,$0F,$16,$1D
	.dc.b	$9E,$78,$76,$8F,$0E,$18,$AE,$B1,$8C,$3A,$CF,$2C,$3C,$B5,$E3,$8C
	.dc.b	$3E,$43,$13,$FE,$09,$AE,$51,$F4,$0E,$B5,$CA,$FE,$13,$E8,$6C,$7C
	.dc.b	$33,$67,$57,$D1,$1C,$63,$37,$E9,$3C,$67,$AF,$0F,$29,$70,$F2,$57
	.dc.b	$7A,$E3,$1E,$E3,$2F,$73,$27,$A6,$BA,$C7,$D0,$AE,$AD,$38,$BE,$37
	.dc.b	$9C,$59,$FF,$5D,$E1,$F1,$1F,$C5,$C7,$47,$B4,$F0,$3C,$8F,$0B,$F8
	.dc.b	$13,$B8,$FC,$9D,$F8,$9F,$9C,$D7,$AD,$73,$7A,$9C,$CF,$74,$6D,$FC
	.dc.b	$CC,$7A,$8F,$0B,$FE,$FA,$BF,$43,$FA,$6A,$F1,$36,$7B,$3E,$03,$B4
	.dc.b	$7C,$17,$95,$FA,$A7,$71,$E0,$3C,$97,$F5,$DB,$1E,$27,$F4,$07,$A8
	.dc.b	$E6,$7D,$7A,$F1,$BF,$92,$3C,$31,$C6,$FD,$31,$93,$8C,$F9,$5E,$A7
	.dc.b	$A4,$CC,$F1,$BF,$44,$74,$3E,$7C,$7C,$76,$9E,$83,$E7,$F8,$AE,$0F
	.dc.b	$F6,$55,$27,$FC,$7A,$FD,$E9,$D7,$95,$FC,$13,$5F,$E0,$BE,$73,$95
	.dc.b	$F2,$D9,$C3,$94,$E6,$31,$8B,$17,$5B,$E6,$F3,$7B,$C3,$C5,$97,$6C
	.dc.b	$BA,$C7,$F9,$7F,$A3,$71,$8C,$BC,$E7,$5D,$53,$E0,$CF,$DD,$F3,$36
	.dc.b	$5F,$2D,$63,$0D,$C7,$FC,$D8,$DC,$3A,$DF,$C7,$8F,$5C,$7C,$7E,$BA
	.dc.b	$9F,$06,$DF,$E6,$4B,$F5,$9F,$E9,$98,$D9,$FC,$BA,$7F,$D3,$87,$18
	.dc.b	$AF,$56,$B9,$33,$1E,$BF,$33,$9C,$FF,$A0,$DF,$0F,$5F,$8B,$8A,$CF
	.dc.b	$FA,$89,$B7,$FE,$AC,$6E,$FF,$D0,$33,$FA,$B1,$F5,$3F,$52,$77,$53
	.dc.b	$DB,$F0,$17,$1F,$6E,$2E,$35,$F0,$9F,$F5,$8F,$CE,$6E,$BF,$D7,$6F
	.dc.b	$C6,$E3,$F8,$FF,$39,$6B,$FD,$6E,$DA,$FF,$60,$B7,$5F,$37,$0A,$31
	.dc.b	$5F,$99,$DB,$F1,$B1,$8B,$FC,$85,$6E,$BF,$D8,$7E,$9F,$D6,$E4,$FA
	.dc.b	$ED,$8C,$9F,$EC,$88,$CB,$4F,$F6,$5F,$27,$DE,$F8,$0F,$BE,$D8,$C9
	.dc.b	$FE,$D0,$8C,$B4,$F8,$FB,$FC,$6F,$2B,$E7,$1F,$88,$F9,$CB,$3A,$E1
	.dc.b	$4F,$E1,$F4,$FD,$6C,$62,$7F,$13,$7C,$DB,$19,$9F,$F7,$44,$E5,$FF
	.dc.b	$75,$F3,$76,$6D,$7D,$AD,$D9,$3E,$4F,$F5,$4D,$9F,$F7,$5F,$C6,$62
	.dc.b	$7F,$73,$AF,$F8,$33,$0F,$E1,$47,$E0,$DC,$3F,$C1,$1A,$E7,$C0,$F6
	.dc.b	$3F,$5B,$BB,$F5,$27,$83,$E7,$2E,$3F,$EF,$AF,$98,$EF,$E0,$FE,$B9
	.dc.b	$B1,$F3,$31,$FD,$86,$1F,$01,$AF,$93,$7F,$E0,$37,$E0,$6D,$7E,$CB
	.dc.b	$67,$EB,$3F,$E2,$A8,$F8,$8D,$BF,$E1,$F5,$C1,$FE,$0B,$19,$E8,$FE
	.dc.b	$4B,$8F,$F1,$2B,$F2,$74,$FC,$AF,$92,$F8,$EC,$8F,$B3,$3B,$C3,$D1
	.dc.b	$3E,$58,$C4,$7A,$5D,$5F,$ED,$F8,$9F,$39,$CE,$BF,$8B,$C6,$30,$F8
	.dc.b	$8F,$D6,$3F,$92,$64,$F5,$B1,$3A,$3C,$73,$E1,$23,$83,$E1,$F1,$F0
	.dc.b	$BF,$82,$FE,$0B,$82,$4F,$E2,$E9,$E8,$62,$FE,$4E,$8B,$F7,$12,$FE
	.dc.b	$1B,$67,$D7,$F3,$5C,$75,$F8,$0F,$F4,$8F,$D5,$1F,$1D,$4F,$6F,$2D
	.dc.b	$C7,$F5,$F3,$A7,$F4,$ED,$C3,$F4,$AD,$7F,$0C,$D7,$FA,$DB,$BC,$3F
	.dc.b	$93,$6F,$6B,$38,$7C,$C3,$FA,$2F,$B3,$FA,$7C,$3F,$AA,$87,$17,$CB
	.dc.b	$E7,$B7,$C1,$30,$FE,$83,$5C,$6F,$F5,$67,$19,$FD,$3B,$67,$71,$FB
	.dc.b	$6F,$9A,$E5,$3F,$82,$7C,$7C,$72,$BF,$37,$F1,$36,$BC,$B8,$7E,$1D
	.dc.b	$FE,$A6,$D7,$C2,$D7,$F7,$AD,$7F,$BE,$68,$E6,$FC,$6E,$1F,$8E,$6E
	.dc.b	$1F,$AA,$6D,$7E,$A1,$BA,$7C,$CF,$9E,$FC,$A9,$F9,$6F,$8C,$B7,$FB
	.dc.b	$28,$B7,$FB,$09,$B5,$FD,$AF,$E5,$5D,$FF,$DD,$9B,$87,$F3,$0F,$A2
	.dc.b	$7C,$56,$6A,$4F,$F5,$AC,$57,$C1,$FD,$FD,$F0,$9D,$4E,$70,$F0,$5F
	.dc.b	$D4,$8E,$17,$F6,$1A,$FF,$24,$6B,$FE,$2F,$FB,$17,$D0,$7F,$83,$1B
	.dc.b	$87,$9F,$CE,$C4,$C8,$F3,$4E,$AB,$CB,$59,$7F,$1F,$F2,$EE,$97,$8B
	.dc.b	$FE,$0C,$8C,$36,$33,$3F,$E1,$29,$CB,$FE,$14,$F6,$DF,$14,$F9,$7F
	.dc.b	$F0,$AC,$61,$B1,$99,$E0,$D7,$F3,$35,$FB,$6F,$73,$C5,$DA,$69,$27
	.dc.b	$FC,$47,$F2,$EF,$2F,$FC,$42,$D8,$FF,$12,$B5,$DC,$68,$FE,$A3,$C7
	.dc.b	$FE,$2D,$9D,$36,$33,$C1,$FD,$5B,$D4,$FF,$1C,$5F,$E0,$7E,$5D,$D1
	.dc.b	$A8,$FF,$32,$7C,$BF,$E2,$1F,$C3,$35,$FD,$EC,$F0,$79,$4F,$F8,$E2
	.dc.b	$31,$1F,$E3,$E6,$BF,$BC,$6B,$FD,$A1,$DE,$3F,$E4,$56,$E1,$9A,$F2
	.dc.b	$52,$FE,$53,$FC,$BF,$E5,$BD,$1E,$6F,$F7,$A6,$F9,$46,$FC,$BE,$19
	.dc.b	$5C,$4F,$E7,$11,$CB,$E4,$FA,$C7,$F9,$89,$7C,$0E,$4D,$70,$C9,$FF
	.dc.b	$05,$4E,$9B,$39,$7F,$87,$E7,$E6,$F0,$FF,$43,$1F,$DB,$BB,$3D,$BE
	.dc.b	$27,$C0,$8F,$F4,$A3,$F1,$9B,$5F,$E6,$48,$C4,$7F,$2D,$8C,$78,$BF
	.dc.b	$0B,$E5,$FE,$04,$FF,$87,$5A,$FF,$20,$8E,$E6,$F5,$91,$FC,$0E,$D3
	.dc.b	$A8,$C3,$5F,$89,$F2,$FF,$87,$F4,$9F,$2C,$F8,$EF,$95,$F2,$F2,$FF
	.dc.b	$C8,$33,$96,$CE,$5F,$9D,$3C,$E7,$51,$FE,$10,$6F,$47,$FA,$C9,$BC
	.dc.b	$FF,$EB,$06,$CF,$FA,$BF,$8B,$FC,$F2,$F8,$3E,$3B,$F0,$BF,$29,$F5
	.dc.b	$63,$FD,$3A,$DC,$3A,$4F,$84,$DF,$24,$AE,$44,$3C,$91,$B7,$FC,$95
	.dc.b	$5F,$A5,$AF,$F6,$3F,$CB,$7E,$BA,$7F,$15,$F2,$FF,$8D,$3E,$44,$66
	.dc.b	$77,$F0,$D9,$AF,$C5,$D6,$A2,$47,$F9,$AA,$B8,$69,$FE,$D0,$6C,$7C
	.dc.b	$1F,$8D,$EB,$F0,$DF,$03,$09,$3C,$63,$2F,$A2,$32,$FF,$8F,$A9,$3A
	.dc.b	$8C,$37,$1F,$F5,$46,$4C,$3F,$51,$F1,$B2,$FF,$46,$FB,$BE,$12,$46
	.dc.b	$08,$77,$7F,$61,$E3,$EA,$30,$7E,$B1,$FD,$A7,$CB,$C8,$FD,$7B,$F7
	.dc.b	$3E,$5F,$E5,$F8,$D6,$7E,$9C,$61,$FA,$BF,$2F,$F8,$22,$3F,$B8,$8F
	.dc.b	$CD,$7C,$BF,$F9,$16,$3F,$77,$3F,$C0,$CC,$FC,$0C,$8C,$90,$FF,$32
	.dc.b	$CF,$16,$C7,$FB,$73,$A9,$39,$9F,$F3,$A3,$51,$FE,$CB,$70,$FE,$6F
	.dc.b	$AB,$F8,$A7,$CB,$C8,$FE,$1F,$1F,$C5,$CC,$3F,$C5,$E3,$0F,$F1,$4F
	.dc.b	$97,$FE,$39,$F1,$12,$30,$43,$F8,$F3,$97,$F9,$04,$7F,$23,$F9,$5F
	.dc.b	$E4,$9D,$52,$7F,$8F,$C6,$5F,$E4,$31,$97,$F9,$44,$65,$FE,$55,$19
	.dc.b	$E5,$FF,$78,$76,$2B,$E8,$FC,$2F,$F3,$1F,$97,$98,$7F,$B2,$35,$FF
	.dc.b	$35,$B6,$7E,$64,$62,$3F,$D7,$ED,$7E,$6C,$7F,$92,$9F,$F7,$37,$C6
	.dc.b	$FF,$80,$16,$3F,$EE,$CF,$97,$FE,$93,$26,$FB,$12,$BF,$D8,$D3,$96
	.dc.b	$CE,$5F,$F2,$FF,$AF,$FC,$41,$FE,$1B,$9F,$F1,$03,$FA,$CC,$67,$07
	.dc.b	$81,$FE,$C8,$8C,$3F,$EF,$BE,$DF,$89,$7F,$C3,$2F,$8F,$B3,$23,$FC
	.dc.b	$26,$FB,$5F,$C4,$DF,$85,$FF,$34,$B5,$FC,$6F,$A2,$F9,$2F,$F3,$6E
	.dc.b	$5E,$29,$4E,$75,$FE,$D7,$5F,$1E,$B5,$3C,$6B,$26,$AA,$3F,$EE,$BA
	.dc.b	$EB,$9F,$E7,$39,$DF,$26,$F8,$BE,$BF,$2F,$30,$FD,$0E,$7E,$3B,$DD
	.dc.b	$7F,$4D,$BE,$15,$FF,$00,$1A,$FF,$BF,$9A,$FF,$B7,$DA,$FC,$DF,$0B
	.dc.b	$EF,$AF,$96,$7C,$77,$C4,$D1,$F2,$BF,$00,$FA,$F7,$C2,$B7,$7F,$3A
	.dc.b	$31,$7C,$6B,$84,$FF,$9A,$F8,$15,$A8,$E1,$A7,$EF,$BC,$BE,$70,$AF
	.dc.b	$E5,$4E,$70,$FC,$27,$8F,$FC,$12,$3F,$C2,$31,$EB,$BD,$47,$43,$D9
	.dc.b	$1C,$1F,$E1,$79,$76,$BE,$B1,$A3,$84,$67,$1F,$8B,$8E,$F1,$F0,$36
	.dc.b	$BB,$9E,$79,$DE,$1E,$38,$E8,$7A,$EB,$D1,$7C,$63,$38,$7C,$4C,$7C
	.dc.b	$B7,$C7,$1D,$11,$C1,$D6,$F9,$11,$C1,$AB,$FD,$0E,$1E,$88,$F8,$51
	.dc.b	$F2,$0F,$8D,$5E,$FB,$F0,$C7,$02,$9E,$92,$E1,$E7,$7D,$38,$FD,$B7
	.dc.b	$47,$B3,$0F,$7E,$1E,$48,$F9,$AF,$8B,$1F,$44,$60,$F1,$C7,$9B,$1F
	.dc.b	$81,$18,$34,$76,$9E,$2F,$D1,$C7,$CB,$87,$D2,$8F,$51,$EF,$3C,$B5
	.dc.b	$EF,$71,$8F,$AE,$F9,$BF,$84,$74,$4C,$3D,$04,$9F,$1C,$F1,$8C,$47
	.dc.b	$95,$F8,$0F,$94,$EF,$8F,$D3,$7C,$B7,$E3,$3B,$3E,$8B,$F1,$E1,$F3
	.dc.b	$70,$ED,$39,$C3,$BC,$62,$3B,$5F,$C8,$7D,$F3,$F8,$37,$C5,$7E,$EC
	.dc.b	$3E,$AC,$7A,$B0,$ED,$18,$8F,$51,$9C,$7F,$03,$19,$C7,$E3,$61,$EE
	.dc.b	$7C,$46,$2B,$CC,$71,$3F,$08,$72,$3E,$BD,$47,$24,$C3,$E1,$BE,$8A
	.dc.b	$F7,$DF,$87,$0F,$88,$7A,$27,$A1,$F9,$4B,$E9,$7E,$F8,$B7,$FB,$5F
	.dc.b	$5B,$FC,$C1,$8F,$E6,$DB,$87,$F2,$46,$E1,$FE,$12,$7F,$CC,$D7,$FE
	.dc.b	$14,$F9,$7C,$67,$FD,$AD,$B1,$FC,$54,$38,$7F,$8C,$35,$E4,$7F,$06
	.dc.b	$3F,$DE,$BC,$5F,$B0,$73,$9F,$F2,$47,$26,$CE,$A7,$1E,$3F,$EB,$11
	.dc.b	$B8,$CB,$C6,$33,$A7,$9E,$78,$57,$77,$71,$E9,$9D,$C7,$AE,$FC,$91
	.dc.b	$88,$F5,$9C,$8F,$24,$6A,$BA,$36,$32,$6E,$FC,$D1,$96,$C6,$6F,$CD
	.dc.b	$1E,$9A,$E7,$39,$3C,$85,$3C,$71,$B7,$BB,$4E,$B8,$64,$B5,$EF,$AF
	.dc.b	$E7,$75,$F4,$23,$0F,$26,$D7,$98,$E8,$F9,$8C,$9F,$36,$BB,$34,$E6
	.dc.b	$F1,$77,$3E,$A9,$E6,$D8,$CB,$EC,$6B,$E5,$3F,$B1,$61,$F4,$DC,$1F
	.dc.b	$E6,$87,$0F,$E1,$AB,$C6,$FC,$86,$BD,$4E,$77,$F3,$AF,$C7,$1F,$7A
	.dc.b	$76,$FE,$61,$AF,$E2,$0E,$27,$C7,$2C,$7B,$DB,$7F,$1D,$A7,$BC,$A7
	.dc.b	$8C,$F5,$57,$E0,$2B,$BE,$1F,$5B,$0F,$F1,$B4,$75,$AD,$9E,$12,$E1
	.dc.b	$F7,$9F,$2C,$57,$6F,$17,$F5,$B1,$67,$E4,$D7,$7B,$FC,$2D,$7C,$BC
	.dc.b	$3E,$D9,$EB,$7B,$B5,$E6,$FE,$05,$F0,$B5,$F0,$BF,$0F,$93,$FC,$A5
	.dc.b	$E5,$7F,$94,$74,$FF,$29,$57,$D5,$9C,$DF,$F9,$10,$E1,$C3,$FE,$D4
	.dc.b	$71,$3F,$2E,$76,$FF,$9F,$23,$FC,$A1,$C4,$F9,$1E,$B3,$86,$AE,$5F
	.dc.b	$F2,$AC,$61,$B5,$FE,$4F,$8F,$33,$FE,$47,$E5,$B5,$C6,$FB,$4F,$F9
	.dc.b	$86,$73,$87,$F9,$BF,$E5,$2E,$4F,$92,$B1,$86,$DF,$F9,$A5,$B8,$FA
	.dc.b	$CF,$F3,$AD,$71,$9F,$83,$1F,$EC,$8F,$43,$D6,$D8,$FF,$42,$28,$CB
	.dc.b	$63,$31,$E4,$F2,$FD,$73,$C4,$F0,$63,$FE,$87,$8C,$C7,$13,$36,$8C
	.dc.b	$D7,$3B,$FF,$53,$F4,$FC,$2E,$FF,$B1,$87,$47,$FC,$DF,$3F,$EA,$CE
	.dc.b	$1D,$56,$DB,$19,$3E,$FF,$03,$E4,$DD,$BF,$5D,$B3,$FE,$BF,$9D,$4E
	.dc.b	$49,$1F,$EB,$AF,$96,$FB,$33,$A8,$C1,$FE,$BE,$E1,$F4,$7C,$B7,$A5
	.dc.b	$B1,$93,$FD,$87,$19,$69,$9D,$3D,$D1,$96,$C7,$FB,$2B,$D1,$F8,$9C
	.dc.b	$7F,$15,$B1,$93,$FD,$9F,$19,$69,$FE,$D2,$E3,$F9,$9E,$FF,$CD,$6C
	.dc.b	$64,$FF,$6B,$C6,$5A,$7C,$AD,$5E,$5F,$C9,$24,$7C,$98,$CB,$63,$ED
	.dc.b	$71,$FC,$89,$CE,$5F,$6E,$31,$3A,$AE,$CD,$9F,$C5,$DB,$E4,$7F,$4A
	.dc.b	$DA,$FF,$76,$FC,$5E,$27,$EA,$E5,$FE,$02,$EA,$FF,$68,$DB,$E2,$73
	.dc.b	$DD,$C9,$FC,$77,$64,$CB,$F8,$2F,$5A,$F8,$47,$FB,$EA,$7F,$29,$F2
	.dc.b	$BD,$BB,$AC,$FE,$3B,$3F,$F6,$2B,$63,$F7,$7F,$05,$FE,$F8,$8C,$1D
	.dc.b	$9C,$6F,$FA,$86,$EB,$FE,$E8,$6C,$FF,$B8,$EF,$AD,$6E,$3D,$FC,$D1
	.dc.b	$F4,$E6,$F9,$0C,$57,$F0,$3E,$8F,$E1,$57,$F2,$BE,$47,$39,$FC,$66
	.dc.b	$E3,$FF,$00,$31,$46,$30,$FF,$7F,$C6,$27,$E0,$3F,$40,$F7,$FA,$B6
	.dc.b	$32,$7B,$F8,$F1,$5E,$F6,$E1,$EF,$6C,$A5,$C2,$DB,$F0,$11,$EB,$D9
	.dc.b	$C1,$F8,$03,$F0,$BE,$75,$C3,$EA,$47,$E2,$3E,$57,$17,$F8,$C6,$E7
	.dc.b	$F8,$66,$D7,$C1,$7F,$7F,$F2,$AF,$8C,$A7,$DC,$E6,$B5,$FC,$3A,$74
	.dc.b	$FD,$D9,$CE,$1F,$75,$B8,$7D,$D6,$CF,$BF,$5F,$74,$7F,$2C,$F8,$3F
	.dc.b	$7D,$7F,$5E,$6D,$6F,$67,$CA,$EF,$82,$4F,$E0,$BE,$56,$E7,$FA,$A6
	.dc.b	$E3,$FA,$46,$CF,$DF,$7C,$1C,$9F,$D1,$32,$FB,$D7,$FE,$77,$0E,$1F
	.dc.b	$F3,$D3,$5F,$B0,$6A,$3F,$43,$C3,$D2,$FF,$40,$FE,$93,$9F,$F4,$B3
	.dc.b	$67,$F4,$8D,$AF,$D2,$37,$1F,$D2,$3F,$AC,$F9,$59,$5F,$AB,$F8,$9B
	.dc.b	$7F,$AB,$9B,$7F,$AB,$9B,$5E,$6F,$93,$77,$7F,$5A,$6C,$7F,$48,$F9
	.dc.b	$47,$CA,$E6,$A4,$FC,$3C,$97,$C1,$F8,$CD,$7F,$D3,$CD,$7C,$13,$FE
	.dc.b	$02,$C5,$F3,$0F,$99,$EB,$5F,$DE,$67,$31,$FD,$BB,$E5,$DF,$21,$4F
	.dc.b	$EA,$7A,$7F,$5A,$BF,$F0,$00,$9D,$E7,$FE,$00,$6C,$7D,$B6,$BF,$EB
	.dc.b	$36,$BF,$69,$AF,$EC,$FE,$5B,$F9,$57,$8B,$F0,$5B,$1F,$85,$6B,$FE
	.dc.b	$BD,$6B,$F8,$4F,$65,$F1,$2F,$97,$F9,$51,$86,$C6,$67,$F2,$13,$97
	.dc.b	$B7,$63,$DA,$F1,$36,$33,$18,$3E,$B7,$CB,$FF,$83,$E3,$FC,$43,$F2
	.dc.b	$EF,$99,$FC,$BF,$17,$F9,$2A,$31,$3F,$E2,$6A,$E0,$FC,$C6,$BF,$ED
	.dc.b	$96,$BF,$A0,$6B,$FC,$65,$AB,$E6,$1D,$1E,$0E,$1F,$D3,$FC,$C7,$E3
	.dc.b	$AB,$FC,$6A,$D7,$ED,$D7,$07,$FC,$72,$D7,$EF,$C6,$F1,$FF,$31,$B6
	.dc.b	$BF,$67,$5B,$9F,$F3,$3B,$C6,$BF,$DE,$4D,$77,$7F,$E4,$28,$C4,$7F
	.dc.b	$40,$F9,$6F,$43,$4F,$E9,$EE,$3C,$4F,$ED,$1F,$31,$7C,$C4,$D2,$57
	.dc.b	$F5,$69,$E1,$F1,$5F,$66,$3F,$C9,$2B,$E0,$B2,$6B,$86,$4F,$F7,$69
	.dc.b	$D3,$67,$2F,$E8,$BC,$4F,$81,$B7,$F2,$7E,$BF,$37,$91,$FA,$AA,$FF
	.dc.b	$31,$D6,$9F,$EA,$0D,$D7,$FA,$85,$70,$3E,$0C,$FF,$22,$DE,$D3,$E5
	.dc.b	$FE,$17,$CC,$7C,$0A,$FF,$05,$B5,$F2,$47,$03,$7B,$C8,$F1,$C6,$5F
	.dc.b	$E1,$D5,$D5,$B1,$FE,$76,$8C,$47,$F3,$03,$FD,$29,$F2,$CF,$91,$F9
	.dc.b	$BF,$2F,$27,$FC,$C7,$39,$6C,$E5,$FA,$33,$F7,$BD,$55,$6E,$7F,$C1
	.dc.b	$4D,$E6,$70,$7F,$77,$AD,$E1,$C6,$B5,$3F,$D7,$78,$BF,$CE,$38,$78
	.dc.b	$BE,$43,$F1,$1F,$1B,$F6,$A3,$FD,$12,$DC,$BA,$C7,$B2,$7A,$1C,$39
	.dc.b	$24,$73,$21,$E6,$8D,$BF,$E3,$CA,$FD,$23,$FE,$B9,$F9,$6F,$F4,$F3
	.dc.b	$F8,$BF,$97,$FC,$99,$F0,$63,$31,$FE,$29,$6E,$1E,$4E,$56,$4B,$E3
	.dc.b	$7F,$E9,$0B,$DE,$7F,$15,$B1,$E1,$F8,$7E,$C7,$E8,$BE,$37,$C1,$F0
	.dc.b	$DF,$13,$09,$5C,$E3,$2F,$AE,$32,$F9,$63,$2D,$38,$1F,$06,$7F,$5F
	.dc.b	$F1,$B2,$3F,$54,$FB,$FE,$12,$46,$08,$7C,$17,$F6,$FE,$36,$A3,$07
	.dc.b	$EC,$DF,$DD,$FC,$BC,$8F,$DB,$3F,$7B,$E5,$FE,$67,$C2,$59,$FA,$91
	.dc.b	$87,$EB,$7C,$BF,$E1,$08,$FE,$FE,$3F,$29,$F2,$8F,$82,$FE,$13,$52
	.dc.b	$BF,$D7,$F3,$96,$CE,$5F,$E1,$B3,$CA,$24,$7F,$0B,$8D,$72,$7E,$43
	.dc.b	$A9,$7C,$EB,$F8,$AF,$57,$F1,$8F,$97,$91,$FC,$4E,$3F,$8D,$98,$7F
	.dc.b	$8D,$C6,$1F,$E3,$1F,$2F,$FC,$83,$E2,$24,$60,$87,$F2,$27,$2F,$F2
	.dc.b	$48,$FE,$4F,$F2,$BF,$CA,$3B,$24,$FF,$23,$8C,$BF,$C9,$63,$2F,$F2
	.dc.b	$C8,$CB,$FC,$BA,$33,$9F,$E6,$FA,$95,$F3,$BD,$EF,$E6,$BF,$2F,$30
	.dc.b	$F9,$B1,$97,$F9,$94,$65,$FF,$47,$6A,$9F,$E8,$9F,$22,$F5,$AF,$CB
	.dc.b	$CB,$FF,$57,$CE,$5B,$39,$7F,$E0,$04,$F2,$8F,$E6,$3E,$AF,$F0,$A9
	.dc.b	$FE,$0F,$97,$F0,$A3,$3F,$23,$23,$F8,$04,$65,$FF,$46,$6B,$FC,$31
	.dc.b	$57,$38,$C3,$7C,$1F,$F3,$A7,$87,$F1,$2B,$99,$CA,$3F,$C3,$98,$7F
	.dc.b	$61,$F9,$3F,$E6,$58,$7F,$05,$7D,$94,$8C,$65,$B8,$87,$FB,$42,$77
	.dc.b	$3A,$8C,$1F,$CC,$FE,$51,$F1,$BD,$FE,$5E,$4F,$8C,$74,$AB,$71,$98
	.dc.b	$C5,$70,$8F,$26,$7F,$E9,$1C,$58,$77,$AF,$9F,$E0,$FA,$6D,$F2,$EF
	.dc.b	$90,$E1,$F2,$DF,$04,$DE,$BF,$EB,$99,$D5,$7F,$AE,$F8,$58,$73,$9C
	.dc.b	$F9,$4D,$18,$F3,$C3,$D8,$7D,$6B,$E1,$97,$F9,$2D,$B1,$FE,$C6,$AE
	.dc.b	$2F,$CB,$AF,$1B,$E8,$AE,$19,$79,$5F,$F2,$A7,$8D,$FE,$98,$CB,$CF
	.dc.b	$5F,$84,$8C,$4F,$08,$E4,$7A,$F1,$F0,$BB,$3A,$1E,$F9,$E2,$EE,$F2
	.dc.b	$FA,$DB,$9A,$77,$3C,$CE,$E6,$CB,$1F,$2A,$78,$CF,$08,$E5,$3D,$2B
	.dc.b	$CB,$1B,$8E,$33,$B3,$84,$E8,$DB,$98,$E5,$51,$E2,$78,$0C,$61,$F2
	.dc.b	$2F,$B3,$6B,$94,$FA,$2F,$B5,$F1,$8E,$66,$5E,$78,$7D,$48,$D5,$F5
	.dc.b	$AE,$31,$C2,$3B,$3C,$0F,$9C,$66,$79,$3D,$49,$1C,$1D,$1C,$1E,$D5
	.dc.b	$E0,$9E,$31,$89,$E3,$18,$23,$C5,$FC,$34,$75,$23,$82,$47,$E1,$4E
	.dc.b	$C7,$12,$1F,$81,$8E,$2E,$30,$F9,$B3,$E0,$8D,$3D,$1F,$C3,$4F,$07
	.dc.b	$27,$1B,$FB,$AF,$80,$DC,$FE,$77,$CB,$F8,$84,$70,$75,$BF,$C6,$CF
	.dc.b	$C8,$38,$9E,$5C,$7E,$FE,$1F,$1C,$95,$E7,$AE,$31,$8A,$E0,$78,$E3
	.dc.b	$DA,$E0,$EC,$E1,$EE,$7B,$08,$E4,$FC,$F7,$AB,$FE,$4B,$6E,$1F,$B9
	.dc.b	$8C,$63,$FC,$6D,$FF,$28,$50,$7C,$BD,$8C,$C6,$0F,$F6,$85,$7F,$9B
	.dc.b	$DC,$46,$A7,$84,$E5,$FD,$FC,$FF,$61,$E1,$AF,$3C,$7F,$87,$B9,$3F
	.dc.b	$7E,$78,$E7,$8F,$27,$F9,$8A,$77,$3C,$73,$FF,$40,$4E,$5F,$04,$72
	.dc.b	$8C,$3F,$E8,$5A,$E7,$7E,$26,$C7,$80,$E8,$7C,$33,$C9,$7E,$08,$DB
	.dc.b	$63,$B4,$F3,$69,$CE,$77,$7E,$4C,$3D,$F8,$7F,$9F,$1E,$46,$2F,$E9
	.dc.b	$3D,$70,$C6,$1C,$CE,$AE,$6F,$E0,$D7,$86,$36,$66,$7D,$57,$F0,$9F
	.dc.b	$66,$1E,$47,$E2,$1D,$2F,$E4,$D7,$26,$9D,$A3,$C5,$19,$78,$B5,$EA
	.dc.b	$73,$33,$1C,$63,$C9,$1F,$8F,$AF,$23,$F7,$CE,$07,$28,$E8,$E5,$AF
	.dc.b	$D6,$39,$5F,$C1,$AE,$C7,$A2,$3E,$0C,$65,$A7,$AA,$36,$F9,$8C,$CF
	.dc.b	$99,$FC,$0C,$73,$7B,$9C,$4D,$B9,$76,$73,$37,$1C,$EB,$DF,$1E,$93
	.dc.b	$67,$BE,$7C,$6D,$3A,$9C,$CD,$9F,$0D,$DB,$4D,$9C,$4F,$0B,$E5,$8F
	.dc.b	$0D,$FE,$06,$74,$74,$3C,$91,$A7,$E9,$9D,$4E,$67,$C7,$E0,$FF,$0C
	.dc.b	$79,$5F,$ED,$6D,$FF,$C9,$93,$E3,$9C,$E5,$FE,$1A,$8C,$70,$AF,$84
	.dc.b	$FD,$7A,$DD,$7F,$8D,$0E,$D5,$B9,$E5,$1F,$B6,$9D,$DF,$FA,$5F,$95
	.dc.b	$F2,$7D,$2B,$FC,$49,$C9,$79,$BF,$C8,$93,$9C,$3F,$C6,$13,$9B,$E3
	.dc.b	$8F,$F9,$4E,$31,$5B,$CB,$FC,$A3,$3F,$E5,$7F,$95,$C7,$CA,$65,$B1
	.dc.b	$93,$FC,$B9,$19,$69,$F0,$FC,$FE,$9E,$5F,$9F,$5F,$E6,$51,$89,$D1
	.dc.b	$FE,$65,$9D,$4A,$74,$FC,$73,$8C,$EC,$E2,$FC,$BD,$3F,$C9,$D5,$F0
	.dc.b	$7B,$FF,$C9,$F5,$FE,$92,$E0,$C7,$56,$5B,$19,$3F,$D2,$F1,$97,$FC
	.dc.b	$D9,$F2,$AE,$DE,$9F,$2F,$F3,$E3,$13,$FE,$9E,$8C,$4E,$A3,$8B,$73
	.dc.b	$F3,$63,$F3,$9B,$F2,$FE,$7E,$2F,$43,$6F,$FD,$52,$D9,$DC,$7F,$AB
	.dc.b	$27,$6D,$36,$7F,$AC,$3D,$1F,$85,$C3,$F0,$DB,$19,$3F,$D6,$B1,$96
	.dc.b	$9F,$EB,$8E,$1F,$95,$EF,$7C,$B6,$C6,$4F,$F5,$EC,$65,$A7,$C8,$F2
	.dc.b	$FE,$97,$0F,$C7,$6C,$64,$FF,$62,$C6,$5A,$67,$2F,$8F,$19,$6C,$7F
	.dc.b	$A8,$F8,$96,$F9,$2F,$D0,$36,$7F,$DA,$3F,$19,$89,$FD,$36,$5F,$DC
	.dc.b	$FA,$95,$FE,$05,$C3,$9F,$CA,$EB,$B3,$9B,$F8,$BE,$A9,$5E,$1C,$D3
	.dc.b	$8D,$3E,$EF,$CB,$F6,$EF,$FC,$93,$6F,$FD,$6D,$A7,$FA,$B5,$F5,$7C
	.dc.b	$0F,$89,$CD,$7E,$E6,$31,$7F,$9D,$6E,$BE,$88,$6A,$9C,$DF,$F8,$7F
	.dc.b	$BD,$8F,$7B,$E5,$3F,$0B,$96,$3E,$BF,$93,$B3,$FC,$56,$F8,$D7,$CC
	.dc.b	$F9,$3C,$E7,$E1,$CF,$E8,$C6,$31,$F9,$5B,$A3,$18,$F9,$A3,$13,$F0
	.dc.b	$9F,$A6,$7B,$FD,$6B,$19,$3E,$2F,$1E,$2B,$E1,$37,$1F,$16,$DF,$7E
	.dc.b	$7D,$1B,$7E,$12,$F8,$9F,$43,$83,$F0,$27,$E1,$BC,$EB,$19,$3E,$E7
	.dc.b	$CB,$E2,$FF,$86,$68,$8C,$65,$F0,$63,$15,$C6,$B7,$5F,$CC,$FE,$5D
	.dc.b	$F1,$94,$FB,$9C,$B6,$FF,$DF,$13,$A7,$F0,$73,$9B,$FC,$03,$71,$F7
	.dc.b	$B6,$7F,$31,$AF,$1A,$FE,$4D,$F0,$3F,$0D,$FF,$1F,$07,$0F,$F9,$05
	.dc.b	$AF,$F9,$95,$AE,$F7,$FD,$5F,$CB,$7F,$39,$9F,$9D,$8F,$C1,$8C,$4E
	.dc.b	$B0,$FF,$7D,$D6,$9F,$C1,$CF,$1E,$2F,$CE,$6F,$67,$F4,$2D,$C7,$F3
	.dc.b	$CD,$C7,$EE,$BF,$A3,$E6,$FD,$25,$DB,$FD,$4B,$6F,$F4,$CD,$BF,$C2
	.dc.b	$BF,$AB,$F9,$69,$5F,$B0,$F8,$7B,$7F,$AD,$9B,$7F,$8C,$9B,$5F,$E6
	.dc.b	$7F,$95,$76,$FF,$80,$B0,$FF,$46,$46,$27,$FD,$11,$08,$C3,$FE,$8B
	.dc.b	$6B,$F5,$1C,$DF,$F8,$37,$E5,$5F,$2B,$FD,$9E,$73,$7F,$E9,$9A,$56
	.dc.b	$A3,$07,$97,$0F,$F0,$24,$EA,$F9,$46,$2B,$86,$1F,$DA,$FE,$51,$F1
	.dc.b	$9E,$B8,$46,$31,$F5,$B6,$3F,$D4,$8D,$7F,$1A,$E7,$1F,$E2,$9E,$0B
	.dc.b	$F5,$F7,$CC,$FF,$80,$6B,$58,$7F,$99,$67,$2F,$AB,$0F,$F5,$B7,$CB
	.dc.b	$7F,$2C,$C3,$E5,$B6,$3F,$D6,$8D,$7F,$3A,$D7,$FD,$39,$EC,$3E,$21
	.dc.b	$F2,$DF,$52,$30,$D8,$CC,$FF,$AF,$E7,$2F,$6E,$B7,$B3,$E2,$6C,$66
	.dc.b	$30,$7F,$B1,$FE,$D9,$F8,$03,$E0,$3F,$D5,$DC,$3F,$C5,$BC,$2B,$BF
	.dc.b	$EE,$1B,$8B,$0E,$37,$86,$C7,$F4,$69,$DF,$8F,$FC,$FD,$D6,$3D,$E7
	.dc.b	$FA,$5F,$C7,$6B,$C2,$7D,$15,$31,$12,$7F,$A8,$C6,$5E,$24,$3F,$A8
	.dc.b	$39,$7F,$A9,$47,$A1,$7C,$AB,$B7,$F4,$31,$86,$C6,$63,$F4,$AD,$7F
	.dc.b	$DD,$CD,$DB,$FA,$E4,$71,$3F,$08,$53,$FC,$85,$F2,$DF,$49,$FE,$C9
	.dc.b	$E1,$FC,$AF,$35,$FA,$03,$E2,$30,$DA,$FE,$DF,$5F,$14,$CC,$C8,$C9
	.dc.b	$FC,$6A,$3E,$57,$D1,$C8,$C1,$0F,$EE,$8E,$5F,$E6,$31,$F7,$31,$FD
	.dc.b	$DB,$F4,$FC,$29,$3F,$9F,$9C,$B6,$13,$A8,$FF,$00,$47,$33,$2E,$E3
	.dc.b	$53,$FD,$8A,$75,$1B,$F9,$6F,$C1,$12,$BB,$39,$7F,$C2,$2E,$5F,$F4
	.dc.b	$A5,$7F,$12,$8D,$46,$29,$18,$9F,$EE,$3F,$31,$EF,$FD,$B7,$F7,$F7
	.dc.b	$FE,$35,$F8,$FF,$AD,$5F,$E0,$76,$F4,$FF,$1B,$AD,$D7,$81,$B5,$BE
	.dc.b	$AF,$CF,$DF,$0A,$4A,$B7,$7C,$FC,$3F,$11,$2B,$C5,$F0,$FE,$5F,$88
	.dc.b	$E7,$F3,$D2,$7E,$96,$09,$D4,$7A,$67,$4F,$96,$32,$D3,$8B,$E6,$BF
	.dc.b	$85,$F2,$5E,$8F,$98,$95,$DA,$32,$FC,$58,$CB,$F0,$63,$2D,$38,$1F
	.dc.b	$2F,$E0,$BE,$1F,$CB,$C8,$F9,$CF,$D1,$E5,$91,$82,$1F,$35,$F9,$7F
	.dc.b	$1B,$A8,$C1,$F5,$DF,$9F,$94,$8F,$B4,$FD,$2F,$1B,$F3,$9F,$B7,$3E
	.dc.b	$7B,$15,$F7,$E7,$31,$F6,$3E,$66,$4F,$E0,$DF,$B7,$F3,$3F,$C2,$7E
	.dc.b	$07,$FC,$2B,$CB,$F8,$5F,$99,$FB,$E6,$6E,$1F,$8B,$72,$FF,$04,$9E
	.dc.b	$3C,$C8,$91,$87,$FC,$CB,$CE,$AE,$31,$FE,$CB,$9D,$79,$0A,$FF,$13
	.dc.b	$4F,$E6,$78,$FF,$22,$7E,$4B,$DA,$91,$92,$1F,$ED,$19,$E2,$DF,$99
	.dc.b	$FC,$CC,$48,$FD,$24,$7E,$70,$C3,$C6,$30,$FE,$A7,$E1,$BF,$43,$E6
	.dc.b	$C8,$C1,$0F,$03,$97,$C5,$1F,$A8,$31,$AF,$EC,$67,$13,$FC,$47,$8B
	.dc.b	$FD,$0A,$FE,$DF,$E6,$7F,$5E,$7D,$EF,$83,$91,$92,$1F,$C1,$A7,$8B
	.dc.b	$F0,$79,$FF,$AA,$FE,$BB,$F3,$0F,$9B,$9F,$F3,$19,$D1,$FE,$B9,$8D
	.dc.b	$3F,$BE,$9E,$3E,$AF,$2F,$98,$93,$B8,$CB,$C6,$32,$FF,$C0,$1A,$4E
	.dc.b	$A3,$13,$AD,$3F,$D3,$15,$A9,$F1,$7C,$97,$A3,$E5,$E4,$60,$87,$57
	.dc.b	$2F,$78,$F7,$F4,$FC,$0E,$59,$3E,$D8,$CB,$EF,$8C,$BE,$78,$F8,$FF
	.dc.b	$25,$F2,$3E,$4E,$4F,$C5,$8C,$BF,$1E,$32,$FB,$E3,$2F,$C1,$8C,$E5
	.dc.b	$EE,$9F,$A7,$F3,$FF,$53,$E4,$E5,$FD,$18,$CB,$F4,$E3,$2F,$CB,$8F
	.dc.b	$B7,$F1,$BF,$73,$E4,$A4,$FD,$98,$CB,$F6,$E3,$2F,$D3,$8F,$C1,$FC
	.dc.b	$97,$F7,$5F,$80,$78,$34,$CD,$7E,$0E,$70,$FF,$90,$FD,$3F,$F2,$54
	.dc.b	$EA,$3E,$F7,$C9,$7E,$38,$8F,$E2,$FA,$6B,$A3,$F1,$D1,$F9,$48,$FC
	.dc.b	$AF,$C7,$7E,$5B,$E8,$A5,$7E,$4E,$32,$FE,$56,$32,$FE,$36,$3F,$3B
	.dc.b	$F2,$7F,$9E,$F8,$89,$3F,$9B,$8C,$BF,$9D,$8C,$BF,$95,$8C,$B4,$E0
	.dc.b	$7E,$8F,$E4,$F3,$C7,$F9,$69,$BE,$3F,$F1,$DF,$19,$D6,$BF,$D7,$FC
	.dc.b	$97,$F9,$13,$F5,$EF,$F1,$4E,$3F,$EF,$56,$E1,$C6,$D5,$C3,$2F,$F3
	.dc.b	$86,$2B,$DC,$23,$17,$B8,$FE,$E2,$FF,$9B,$BC,$CF,$EE,$70,$F9,$A7
	.dc.b	$6F,$F9,$CB,$CE,$FE,$69,$39,$8C,$3F,$CD,$E7,$6F,$F3,$9F,$87,$FE
	.dc.b	$E5,$AE,$6B,$B3,$67,$57,$D2,$F8,$C2,$D3,$9B,$E3,$3E,$13,$B6,$1E
	.dc.b	$8B,$E3,$7C,$27,$B1,$CE,$F9,$3C,$30,$CC,$F9,$A3,$A1,$B3,$33,$C1
	.dc.b	$F2,$5F,$7C,$3C,$78,$78,$2B,$57,$E1,$BE,$6F,$59,$F2,$DF,$19,$F8
	.dc.b	$27,$63,$C6,$E5,$EF,$86,$5E,$77,$DA,$75,$7D,$70,$FA,$75,$A3,$9D
	.dc.b	$C7,$04,$C3,$C7,$19,$7D,$51,$DD,$CB,$EA,$BE,$F8,$7B,$8F,$15,$71
	.dc.b	$BF,$0D,$F2,$BE,$13,$E1,$F5,$DE,$BF,$DA,$6C,$FD,$33,$FD,$6B,$19
	.dc.b	$3F,$1E,$FF,$80,$FE,$6B,$FC,$09,$3C,$23,$F2,$71,$89,$DC,$23,$0F
	.dc.b	$26,$BC,$0F,$BE,$7F,$4E,$D3,$F0,$EF,$73,$3E,$32,$33,$1D,$4C,$B8
	.dc.b	$9E,$33,$B9,$EA,$6F,$1F,$CE,$BD,$CC,$CF,$0A,$E5,$1D,$0C,$96,$78
	.dc.b	$BC,$F2,$F8,$B7,$DE,$73,$7C,$EB,$8C,$78,$27,$55,$F1,$AB,$53,$C2
	.dc.b	$BB,$4E,$5A,$7C,$92,$9F,$26,$3A,$5F,$A1,$B1,$E0,$9F,$2C,$E5,$B1
	.dc.b	$93,$F8,$5D,$F9,$1E,$D5,$C8,$F2,$16,$7C,$8D,$8D,$C7,$48,$FE,$23
	.dc.b	$B7,$F2,$1F,$96,$FA,$5A,$7F,$29,$F9,$4B,$97,$F3,$1E,$15,$63,$2F
	.dc.b	$E4,$D0,$AC,$61,$CE,$FF,$96,$CE,$5B,$7A,$8C,$5F,$F2,$C3,$9F,$CB
	.dc.b	$64,$D7,$C5,$DB,$E7,$16,$78,$61,$FC,$E6,$B1,$97,$F4,$1A,$ED,$F4
	.dc.b	$78,$C3,$D4,$DC,$3D,$2D,$C3,$FA,$13,$E5,$FA,$0F,$37,$CB,$5B,$F5
	.dc.b	$B6,$FC,$39,$A3,$19,$7F,$51,$7D,$DF,$2D,$2F,$E1,$FC,$B5,$C3,$8C
	.dc.b	$DB,$FE,$B1,$36,$BE,$87,$C5,$7C,$FF,$11,$E3,$3D,$7F,$95,$39,$8C
	.dc.b	$57,$D0,$8C,$4E,$A3,$FC,$19,$3A,$CB,$D3,$F2,$9F,$4F,$E5,$FE,$A3
	.dc.b	$6B,$EC,$B6,$77,$1F,$72,$7E,$D7,$C8,$FD,$BF,$8E,$B3,$F8,$06,$D7
	.dc.b	$C6,$6E,$BF,$DD,$5F,$C1,$7C,$7F,$E0,$FE,$2E,$DF,$E2,$1B,$7F,$39
	.dc.b	$B7,$FD,$F5,$B3,$F9,$56,$CF,$E5,$1F,$C7,$7C,$67,$E3,$FE,$32,$D7
	.dc.b	$E5,$1B,$5F,$69,$B7,$FE,$09,$7F,$2D,$F1,$9F,$97,$F8,$CB,$7F,$9C
	.dc.b	$6D,$FE,$09,$B7,$FE,$11,$7F,$47,$FA,$FB,$C4,$BE,$26,$1B,$87,$F8
	.dc.b	$4A,$74,$D9,$D3,$4D,$4F,$0C,$3F,$43,$5C,$1B,$5E,$0F,$8C,$F0,$FC
	.dc.b	$65,$C3,$B3,$6B,$8B,$72,$FE,$65,$DC,$C5,$3F,$BF,$97,$C0,$FF,$8A
	.dc.b	$B0,$F4,$7C,$66,$63,$18,$F9,$23,$1E,$67,$C2,$6D,$70,$3E,$07,$C6
	.dc.b	$7C,$6F,$8C,$F8,$ED,$8C,$9F,$16,$32,$D3,$37,$F0,$A3,$2D,$8F,$99
	.dc.b	$E6,$7D,$2F,$37,$E9,$B6,$32,$7D,$18,$CB,$4F,$A9,$F1,$9F,$6B,$D8
	.dc.b	$FB,$6D,$8C,$9F,$66,$32,$D3,$FD,$E9,$AF,$F9,$52,$BF,$07,$F1,$9F
	.dc.b	$CA,$71,$4A,$73,$5C,$0F,$F2,$4C,$65,$FC,$35,$71,$8C,$57,$E0,$A3
	.dc.b	$15,$CA,$3F,$96,$1D,$9C,$4F,$DE,$E0,$7C,$67,$E5,$8F,$CB,$FC,$5D
	.dc.b	$C7,$C5,$87,$9B,$F6,$F7,$8D,$F9,$CF,$8D,$FC,$24,$62,$BF,$35,$1F
	.dc.b	$A0,$F1,$64,$60,$FC,$F3,$FA,$4F,$8A,$E3,$E0,$D9,$EF,$AB,$04,$EA
	.dc.b	$7F,$4D,$F1,$B7,$3F,$D0,$BF,$AE,$24,$72,$AE,$2F,$03,$AF,$B7,$FA
	.dc.b	$7C,$3F,$AF,$B6,$3F,$D1,$DB,$BE,$37,$F4,$13,$9C,$7E,$BB,$7B,$3F
	.dc.b	$4D,$F1,$BA,$F0,$3F,$7E,$DC,$7F,$78,$DB,$FD,$BB,$D3,$A6,$4F,$F0
	.dc.b	$1F,$8D,$B5,$FD,$8F,$46,$4B,$FE,$DB,$E3,$2B,$75,$FB,$88,$C6,$1E
	.dc.b	$EA,$DD,$6A,$3F,$D5,$53,$A8,$ED,$7F,$ED,$49,$F8,$BD,$3D,$BD,$9F
	.dc.b	$82,$C8,$FF,$53,$DB,$E3,$71,$87,$F7,$28,$C4,$F2,$37,$3A,$CB,$CC
	.dc.b	$FC,$CE,$69,$3D,$3E,$32,$E1,$E4,$9F,$7F,$C6,$E3,$0F,$73,$6F,$D6
	.dc.b	$DA,$FF,$5C,$62,$8F,$B3,$BF,$C8,$F8,$DB,$7E,$A6,$DF,$A5,$B8,$7C
	.dc.b	$B6,$CF,$C8,$6C,$F6,$7F,$75,$D6,$F3,$7E,$AF,$BB,$86,$DF,$EA,$AD
	.dc.b	$5B,$8F,$97,$5B,$69,$F3,$BE,$33,$FC,$01,$F1,$4F,$8D,$FC,$0C,$61
	.dc.b	$B3,$FB,$6F,$59,$5B,$8E,$46,$5C,$63,$FC,$13,$E6,$7F,$1B,$3E,$56
	.dc.b	$C7,$F8,$CE,$9F,$4A,$F3,$BF,$8C,$56,$B7,$FF,$3C,$5A,$38,$BF,$E9
	.dc.b	$26,$BF,$EE,$26,$BF,$99,$6F,$63,$8F,$CB,$5F,$9B,$F8,$EF,$CF,$EB
	.dc.b	$FA,$06,$E9,$F9,$26,$E3,$F9,$16,$D7,$19,$FC,$BD,$71,$7F,$8D,$FC
	.dc.b	$B7,$AF,$97,$F5,$7F,$21,$FE,$5B,$8C,$4F,$E7,$A3,$1A,$FE,$2E,$30
	.dc.b	$72,$71,$97,$E8,$67,$F9,$E7,$CC,$3E,$09,$DB,$FE,$AE,$A4,$E6,$30
	.dc.b	$D8,$FF,$59,$B5,$FF,$5D,$FC,$43,$CE,$7C,$C7,$38,$C3,$63,$33,$FE
	.dc.b	$BC,$9C,$BD,$3E,$17,$FC,$24,$DF,$17,$C3,$18,$8F,$F6,$4E,$08,$C4
	.dc.b	$7F,$0F,$D9,$F3,$1F,$C9,$63,$FA,$2F,$2B,$E6,$3F,$45,$18,$6B,$FE
	.dc.b	$7D,$AE,$3E,$47,$F5,$CE,$2E,$98,$73,$AF,$F3,$DD,$F0,$E5,$FF,$09
	.dc.b	$B6,$B7,$7C,$69,$8F,$87,$DE,$C3,$E3,$F3,$6E,$1F,$D2,$AB,$C9,$58
	.dc.b	$33,$32,$32,$43,$C6,$E5,$F2,$C7,$85,$7C,$C3,$E0,$F1,$F3,$1F,$A7
	.dc.b	$9F,$E0,$0D,$7F,$B9,$CE,$DF,$E0,$8D,$7F,$DE,$6D,$7F,$DF,$6D,$7F
	.dc.b	$A0,$1E,$09,$FC,$66,$3F,$47,$B3,$E2,$77,$BE,$61,$E7,$61,$B8,$7D
	.dc.b	$5A,$F8,$A6,$66,$46,$4F,$E3,$1E,$2F,$ED,$5F,$B9,$F0,$5F,$30,$F8
	.dc.b	$51,$98,$FF,$07,$B7,$8B,$EC,$CE,$E7,$9D,$FD,$99,$7C,$16,$F8,$FF
	.dc.b	$AE,$69,$F8,$B8,$FB,$26,$CD,$5F,$F8,$A1,$B1,$B8,$FC,$84,$6E,$E4
	.dc.b	$60,$87,$F6,$D8,$E7,$7E,$2F,$82,$FC,$39,$27,$FC,$27,$28,$F1,$19
	.dc.b	$8F,$CB,$B8,$76,$FE,$FE,$7A,$47,$D8,$3F,$C8,$73,$CD,$C1,$EB,$9E
	.dc.b	$1F,$31,$FE,$70,$8F,$D6,$FA,$DF,$B5,$BF,$F3,$AB,$72,$FF,$26,$D6
	.dc.b	$EB,$FC,$10,$DA,$DF,$5F,$E9,$2F,$85,$7F,$94,$2B,$57,$CB,$E0,$7E
	.dc.b	$A4,$62,$75,$1C,$7E,$0B,$3E,$E7,$3F,$16,$57,$F8,$E2,$FF,$CB,$4D
	.dc.b	$9C,$BF,$EF,$6C,$7F,$45,$9B,$2F,$4D,$74,$AE,$04,$38,$65,$E4,$F4
	.dc.b	$7D,$5F,$01,$27,$C1,$19,$7E,$3C,$65,$F8,$71,$96,$9C,$0F,$9B,$F0
	.dc.b	$1F,$17,$D8,$91,$F4,$1F,$87,$CD,$23,$04,$3E,$7B,$F3,$7E,$03,$51
	.dc.b	$83,$EC,$BF,$47,$29,$1F,$71,$FA,$9E,$37,$E8,$FE,$09,$EA,$E2,$BF
	.dc.b	$03,$39,$8F,$B5,$EC,$C9,$FC,$2B,$F7,$7E,$07,$F8,$A7,$AD,$F8,$FE
	.dc.b	$2F,$C3,$FC,$0F,$E0,$4C,$DC,$3F,$1A,$E5,$FF,$0B,$4F,$1E,$3F,$E1
	.dc.b	$ED,$8F,$E5,$FC,$9F,$E8,$87,$89,$C0,$FC,$77,$77,$E3,$F8,$E4,$7E
	.dc.b	$6A,$32,$FE,$72,$32,$FE,$16,$3F,$2D,$F0,$1F,$97,$A9,$3F,$A2,$8C
	.dc.b	$BF,$A4,$8C,$BF,$8D,$8F,$CF,$7C,$07,$E7,$FC,$99,$3F,$AA,$8C,$BF
	.dc.b	$AC,$8C,$BF,$95,$8C,$BF,$97,$8C,$E9,$FB,$09,$FA,$DF,$2B,$FA,$BF
	.dc.b	$72,$5F,$FA,$63,$8D,$3A,$8F,$DC,$CE,$9F,$F3,$E4,$E9,$FD,$EF,$37
	.dc.b	$F5,$8F,$91,$F9,$DE,$19,$BA,$FF,$38,$AF,$D5,$56,$D8,$72,$3F,$B9
	.dc.b	$CF,$F8,$41,$BE,$FF,$53,$B7,$A9,$23,$24,$3F,$DA,$7C,$7F,$CC,$AB
	.dc.b	$94,$60,$FE,$DE,$E3,$6F,$DD,$46,$2F,$B3,$E2,$BF,$67,$BF,$ED,$F6
	.dc.b	$24,$F9,$A3,$2F,$A2,$32,$F7,$D3,$E1,$4F,$C5,$E9,$F8,$DC,$D2,$BD
	.dc.b	$F1,$97,$E0,$C6,$5E,$59,$FE,$B1,$F8,$BE,$F7,$CF,$F6,$A4,$60,$87
	.dc.b	$07,$2F,$28,$FA,$C6,$35,$F8,$39,$7D,$7A,$CD,$7D,$2F,$6F,$EE,$7B
	.dc.b	$72,$30,$43,$CA,$E5,$F3,$C7,$E0,$FE,$63,$F0,$9E,$FC,$9F,$BB,$19
	.dc.b	$7E,$FC,$65,$F7,$C7,$E2,$FE,$63,$FC,$07,$F1,$4F,$06,$99,$AF,$C3
	.dc.b	$CE,$1F,$F4,$54,$ED,$C9,$F8,$CF,$98,$FC,$B1,$1F,$C7,$F5,$D7,$E7
	.dc.b	$C5,$DB,$E2,$F0,$7E,$7E,$7E,$9F,$CB,$FE,$7F,$E6,$24,$7E,$62,$3F
	.dc.b	$44,$61,$FB,$31,$87,$F3,$BF,$31,$FA,$7F,$87,$91,$82,$1F,$80,$72
	.dc.b	$FE,$0A,$3F,$5C,$47,$78,$79,$BE,$63,$11,$B3,$3D,$9F,$A3,$8F,$DB
	.dc.b	$7C,$67,$F9,$3B,$E4,$BF,$D4,$37,$F8,$26,$D7,$EA,$67,$77,$FE,$66
	.dc.b	$9C,$BF,$E5,$86,$F3,$FE,$FE,$7F,$C9,$9F,$0A,$F3,$18,$FF,$BF,$FE
	.dc.b	$5D,$D9,$FE,$42,$A7,$9D,$FE,$44,$9C,$E6,$13,$A7,$95,$71,$C3,$AC
	.dc.b	$2B,$8B,$C2,$33,$18,$AE,$53,$BA,$F1,$5F,$85,$E9,$87,$69,$F1,$57
	.dc.b	$AE,$31,$1E,$4A,$E0,$E9,$AF,$86,$37,$3C,$63,$C0,$6C,$CC,$F0,$8D
	.dc.b	$5F,$9F,$0F,$4E,$1E,$6B,$D9,$D2,$32,$F9,$9A,$F9,$4D,$9E,$03,$C8
	.dc.b	$5A,$EF,$3D,$63,$17,$E3,$AE,$75,$A9,$F4,$1A,$3C,$58,$47,$04,$C3
	.dc.b	$D5,$19,$7E,$95,$FB,$CE,$D1,$9C,$3C,$53,$AA,$C1,$ED,$9F,$5C,$66
	.dc.b	$3D,$58,$79,$8F,$C1,$FA,$CF,$5B,$ED,$CA,$30,$F3,$31,$1A,$8C,$C7
	.dc.b	$FA,$5B,$E6,$3F,$25,$18,$8F,$CA,$46,$27,$FC,$C2,$D8,$FF,$33,$35
	.dc.b	$E4,$7D,$13,$FD,$D7,$BF,$E2,$9E,$66,$7C,$4F,$EA,$78,$FE,$6A,$32
	.dc.b	$D8,$CE,$7E,$09,$E1,$5C,$D5,$F5,$AD,$C7,$90,$D9,$E3,$36,$62,$FC
	.dc.b	$0E,$1F,$05,$75,$AE,$8D,$9E,$17,$E3,$AD,$C7,$5A,$DB,$4D,$9E,$38
	.dc.b	$D9,$E0,$8D,$B6,$36,$7C,$A7,$73,$DC,$FF,$4A,$5F,$BA,$75,$18,$BF
	.dc.b	$35,$70,$9F,$04,$E5,$F2,$46,$5E,$27,$FB,$5A,$BA,$3F,$4A,$78,$4F
	.dc.b	$F1,$6D,$DC,$0F,$97,$FF,$05,$ED,$FC,$AF,$E6,$2E,$7F,$EC,$4A,$D3
	.dc.b	$FC,$AE,$33,$97,$F2,$BB,$B8,$74,$9B,$7C,$2B,$FC,$18,$A3,$3F,$17
	.dc.b	$30,$FE,$6D,$F1,$36,$F7,$37,$2F,$E7,$D3,$6B,$C9,$3D,$BE,$63,$17
	.dc.b	$FC,$F5,$B9,$7B,$9B,$91,$F2,$FC,$C7,$9B,$E6,$2E,$3F,$D1,$DB,$87
	.dc.b	$A1,$B9,$FF,$4F,$7D,$BF,$31,$2B,$FA,$77,$CA,$DB,$F7,$CD,$BF,$EA
	.dc.b	$D3,$6B,$E6,$7C,$3F,$CA,$F8,$77,$88,$F5,$FE,$4C,$E6,$31,$5F,$D6
	.dc.b	$E3,$13,$A8,$FF,$04,$BF,$43,$E6,$3E,$8F,$CC,$59,$FE,$C4,$DD,$3E
	.dc.b	$9B,$73,$FE,$D4,$FD,$8F,$95,$FB,$3F,$1B,$6F,$FB,$53,$6F,$ED,$B6
	.dc.b	$FF,$B9,$3F,$80,$F8,$EF,$C0,$FC,$55,$BF,$EE,$4D,$BF,$C1,$B6,$FF
	.dc.b	$BD,$36,$7F,$20,$D9,$FE,$F0,$FE,$33,$E2,$FF,$1B,$F1,$76,$BF,$BF
	.dc.b	$36,$BF,$1E,$DB,$FF,$03,$3F,$94,$F8,$BF,$CA,$FC,$C5,$BF,$F0,$33
	.dc.b	$6F,$F2,$ED,$BF,$F0,$73,$F9,$EF,$97,$FE,$69,$5F,$DF,$E7,$FC,$A7
	.dc.b	$1F,$CE,$63,$07,$F8,$0F,$A9,$1C,$2F,$FC,$DC,$ED,$BE,$27,$5F,$13
	.dc.b	$B3,$6F,$9C,$AF,$9C,$E6,$38,$CE,$A7,$BD,$79,$FE,$37,$D1,$F1,$96
	.dc.b	$B5,$8A,$31,$87,$91,$B9,$7B,$1F,$DF,$AF,$81,$79,$7F,$0F,$E3,$33
	.dc.b	$18,$C7,$AC,$62,$75,$1F,$EB,$76,$E1,$F0,$BE,$33,$E6,$7C,$67,$CD
	.dc.b	$6D,$79,$9B,$3B,$8F,$4C,$ED,$A6,$CB,$1F,$3B,$CD,$FB,$1E,$6F,$D9
	.dc.b	$6C,$FC,$26,$CE,$E3,$E2,$CF,$DF,$F8,$CF,$C0,$7B,$36,$7E,$4B,$6B
	.dc.b	$ED,$B7,$5F,$C1,$BF,$C9,$A3,$FC,$97,$7F,$89,$F8,$CC,$57,$F9,$3E
	.dc.b	$FE,$9C,$62,$75,$1F,$80,$AE,$2E,$8C,$57,$18,$FC,$5D,$71,$9D,$47
	.dc.b	$E3,$BE,$33,$F3,$07,$E6,$7E,$2E,$CF,$E3,$2D,$F3,$0F,$1B,$F3,$DF
	.dc.b	$1B,$F8,$08,$C5,$7E,$22,$3F,$45,$E2,$48,$C1,$F8,$C7,$F4,$DF,$15
	.dc.b	$C7,$C1,$B3,$DE,$D1,$98,$FC,$9F,$C6,$5D,$7F,$4A,$FE,$BC,$91,$C6
	.dc.b	$6C,$FF,$B9,$73,$7B,$5F,$AD,$7F,$AD,$B6,$BF,$65,$D4,$F8,$DF,$CD
	.dc.b	$CE,$6F,$FC,$FE,$DE,$EF,$DB,$FC,$6E,$BC,$0F,$E0,$0D,$C7,$F5,$2D
	.dc.b	$BF,$DD,$BD,$3A,$64,$FE,$BB,$E3,$6D,$79,$F1,$57,$F6,$8F,$1B,$FD
	.dc.b	$11,$7F,$4E,$72,$FE,$EE,$33,$5C,$6B,$83,$D3,$5F,$F7,$1C,$E9,$F8
	.dc.b	$3D,$3E,$E3,$B7,$B3,$70,$ED,$1E,$1F,$8D,$C6,$1F,$F0,$07,$04,$63
	.dc.b	$0E,$AD,$BF,$E3,$93,$A7,$D1,$CC,$7F,$D4,$57,$2F,$EA,$CE,$63,$27
	.dc.b	$D0,$70,$EF,$34,$62,$3F,$97,$46,$2B,$E2,$7C,$6F,$DA,$AF,$94,$D8
	.dc.b	$C9,$E4,$2D,$7D,$1D,$BE,$87,$C6,$7E,$09,$B3,$F4,$5B,$3B,$8F,$A5
	.dc.b	$3B,$69,$B2,$C7,$FB,$21,$F8,$3D,$4F,$33,$EF,$7B,$9F,$C0,$E3,$13
	.dc.b	$F7,$63,$13,$A8,$FE,$01,$3A,$AF,$A9,$5F,$87,$F8,$CF,$F0,$47,$C4
	.dc.b	$BE,$33,$18,$7F,$46,$F5,$53,$98,$FC,$55,$70,$AD,$E1,$E7,$17,$C2
	.dc.b	$3F,$8C,$7C,$BF,$EB,$2F,$EB,$D7,$FB,$76,$30,$E1,$AF,$DC,$6B,$F8
	.dc.b	$27,$7C,$3F,$E4,$EF,$9A,$7A,$3D,$B8,$FF,$A6,$C6,$30,$FD,$06,$3F
	.dc.b	$A5,$F8,$DF,$D3,$36,$32,$7E,$96,$32,$D3,$39,$7E,$6E,$32,$D8,$FE
	.dc.b	$D7,$F2,$CF,$41,$CB,$FB,$3F,$8F,$FF,$2E,$CE,$63,$15,$C2,$B6,$FE
	.dc.b	$D6,$B7,$18,$BF,$D5,$47,$F4,$1F,$97,$7C,$0B,$B7,$E5,$D2,$73,$18
	.dc.b	$6C,$7D,$06,$BF,$A8,$F8,$77,$9C,$F9,$8E,$71,$86,$C6,$67,$EC,$CE
	.dc.b	$5E,$9F,$09,$FE,$CD,$6F,$89,$E1,$8C,$47,$ED,$B0,$46,$20,$7E,$23
	.dc.b	$F6,$27,$C0,$FE,$FF,$7C,$F4,$A7,$35,$F9,$56,$BF,$91,$E2,$7C,$CF
	.dc.b	$F9,$86,$35,$87,$37,$B3,$6F,$A3,$9A,$E3,$E7,$E7,$0F,$EB,$57,$DF
	.dc.b	$D0,$C1,$CF,$CA,$78,$CF,$93,$CD,$6C,$87,$F2,$37,$2F,$E9,$F3,$F6
	.dc.b	$5E,$3B,$BF,$98,$D7,$9D,$7E,$E1,$FB,$5B,$FE,$DC,$D7,$FA,$C3,$63
	.dc.b	$F9,$25,$BF,$57,$F2,$D7,$0E,$DE,$97,$BE,$1F,$F2,$83,$67,$F7,$35
	.dc.b	$C1,$FD,$E0,$F9,$97,$C2,$FE,$42,$30,$D8,$CC,$7E,$51,$AF,$F0,$06
	.dc.b	$E5,$FC,$D9,$B8,$7F,$3E,$6C,$FE,$13,$E6,$BF,$13,$5F,$CF,$3F,$1A
	.dc.b	$79,$7F,$6E,$91,$87,$FC,$3C,$D7,$D3,$F6,$CF,$84,$E1,$39,$BE,$97
	.dc.b	$83,$A5,$6E,$FE,$7F,$07,$DF,$FD,$15,$E2,$D3,$E5,$C4,$9D,$90,$F9
	.dc.b	$3F,$AA,$FF,$83,$1C,$63,$FE,$E6,$E1,$FC,$44,$63,$7F,$EB,$F5,$B8
	.dc.b	$FB,$BF,$35,$F9,$5E,$37,$BF,$4F,$B7,$12,$7F,$0A,$D7,$F4,$92,$F8
	.dc.b	$0C,$65,$E1,$AB,$8F,$EC,$BB,$3F,$A1,$57,$E1,$BE,$57,$F4,$9F,$91
	.dc.b	$78,$F0,$FE,$E1,$48,$C5,$FE,$2A,$B8,$65,$F9,$66,$BF,$B5,$97,$CC
	.dc.b	$E3,$0F,$85,$56,$FC,$55,$AB,$FF,$26,$7E,$E3,$FC,$21,$FF,$4A,$71
	.dc.b	$3C,$BF,$A8,$7E,$9E,$75,$3F,$E3,$B6,$E3,$FE,$07,$0E,$30,$FF,$58
	.dc.b	$36,$59,$BF,$1E,$67,$D7,$C5,$95,$E4,$7F,$84,$4E,$A3,$CF,$D9,$F8
	.dc.b	$7E,$59,$18,$21,$E5,$8F,$41,$F8,$DF,$52,$C6,$B4,$F4,$67,$FA,$4F
	.dc.b	$09,$BF,$F7,$4A,$F8,$7F,$3D,$F9,$52,$3F,$23,$C3,$F8,$84,$70,$7C
	.dc.b	$CF,$4F,$E4,$11,$C1,$F3,$BE,$7B,$F4,$1C,$F2,$3E,$8B,$FA,$3E,$6F
	.dc.b	$E2,$DF,$B9,$3E,$3A,$92,$BF,$B2,$FB,$1F,$E3,$27,$EC,$FC,$F7,$F2
	.dc.b	$F8,$C4,$7F,$39,$6F,$07,$F8,$F7,$7F,$7D,$7E,$FF,$EF,$3F,$D9,$D7
	.dc.b	$FE,$9B,$7F,$B8,$37,$D6,$D7,$C8,$48,$D3,$5F,$E3,$72,$F8,$FC,$63
	.dc.b	$FE,$48,$8E,$B3,$AA,$F0,$57,$F0,$1F,$9F,$93,$F9,$27,$F8,$27,$B9
	.dc.b	$88,$FF,$11,$67,$F5,$A7,$18,$FA,$F9,$7F,$4F,$48,$6F,$FD,$CA,$3F
	.dc.b	$D5,$4D,$7F,$D2,$8D,$8E,$5D,$9F,$43,$D7,$95,$C1,$FE,$2D,$1F,$53
	.dc.b	$E3,$BF,$8C,$7C,$84,$AE,$B1,$96,$90,$F4,$FE,$93,$FE,$6A,$C3,$F5
	.dc.b	$18,$AB,$85,$79,$1F,$EC,$55,$BF,$0B,$FB,$DE,$A8,$7C,$F6,$89,$7F
	.dc.b	$E0,$0C,$7E,$F4,$BE,$7B,$33,$E2,$AB,$5A,$9F,$C7,$FC,$8F,$83,$E0
	.dc.b	$25,$FC,$88,$CB,$48,$7B,$3C,$5F,$C5,$7C,$5F,$E2,$FE,$4E,$4F,$E2
	.dc.b	$E3,$2D,$21,$FE,$0D,$FC,$07,$F1,$0E,$32,$FC,$44,$F0,$8D,$73,$FE
	.dc.b	$93,$E2,$A4,$7D,$C7,$FC,$11,$E1,$FE,$E6,$3E,$EC,$69,$A7,$FA,$DF
	.dc.b	$C6,$DF,$8B,$23,$FD,$74,$F2,$F1,$31,$1F,$EA,$F8,$B9,$7F,$AE,$A7
	.dc.b	$FB,$37,$A1,$FC,$56,$BF,$D6,$9E,$2F,$88,$CF,$54,$8F,$A9,$CF,$FD
	.dc.b	$4A,$31,$1F,$EB,$BF,$16,$E3,$FE,$AE,$C3,$A3,$6B,$DB,$3E,$EF,$0E
	.dc.b	$5F,$FB,$45,$F8,$3D,$F8,$8F,$F6,$34,$5B,$FF,$68,$CF,$C6,$F0,$25
	.dc.b	$7F,$B6,$5F,$93,$86,$23,$FD,$9D,$16,$BF,$DB,$33,$FE,$29,$DD,$E2
	.dc.b	$52,$5F,$FB,$8F,$BE,$BF,$F0,$06,$76,$FC,$23,$F6,$CE,$1F,$F6,$E4
	.dc.b	$7F,$90,$7F,$65,$FD,$04,$3B,$BF,$C8,$73,$F7,$8D,$6F,$27,$C0,$43
	.dc.b	$F8,$3D,$72,$C9,$87,$F8,$12,$7C,$9E,$57,$AC,$F0,$77,$58,$F3,$57
	.dc.b	$09,$C4,$74,$9F,$F2,$F7,$CE,$3E,$3E,$92,$7D,$DA,$3E,$8D,$CF,$99
	.dc.b	$FF,$2B,$5F,$17,$FD,$88,$D7,$F8,$FC,$F0,$73,$7C,$72,$FF,$29,$D6
	.dc.b	$DF,$E7,$95,$BA,$EB,$5B,$D7,$FC,$FF,$7B,$7F,$85,$D6,$EF,$9F,$CB
	.dc.b	$FB,$09,$87,$AB,$6F,$E1,$4F,$9A,$32,$FA,$9F,$B1,$E1,$7F,$92,$7E
	.dc.b	$1F,$FC,$57,$7F,$E9,$2F,$A2,$7C,$77,$FA,$53,$05,$7F,$8A,$5D,$46
	.dc.b	$63,$10,$9E,$2F,$FA,$72,$3A,$B9,$9E,$1E,$C7,$F4,$18,$C1,$BF,$5F
	.dc.b	$C1,$7C,$A3,$C3,$18,$AE,$38,$78,$A7,$B5,$F3,$BE,$18,$79,$1B,$87
	.dc.b	$1A,$DD,$F1,$7D,$B7,$C7,$0F,$65,$6E,$7C,$26,$F0,$F5,$E1,$DA,$F8
	.dc.b	$4F,$A4,$D5,$71,$AD,$DF,$A2,$7D,$E7,$42,$C7,$9E,$35,$7D,$70,$E7
	.dc.b	$50,$DE,$1E,$07,$D3,$19,$9F,$7C,$79,$E7,$57,$E2,$8F,$31,$E0,$C3
	.dc.b	$CD,$3A,$36,$FA,$4C,$DF,$33,$B6,$1F,$1A,$B0,$79,$23,$2F,$6C,$3D
	.dc.b	$4E,$47,$C4,$FF,$89,$DF,$A6,$63,$D1,$F8,$D1,$98,$F6,$99,$3F,$28
	.dc.b	$3E,$8F,$FD,$D3,$1B,$8F,$F7,$5C,$62,$77,$1B,$39,$1F,$9D,$8D,$3D
	.dc.b	$1F,$C2,$2F,$8C,$70,$E2,$39,$61,$F9,$57,$51,$9A,$E3,$5C,$27,$EF
	.dc.b	$9E,$42,$B9,$AE,$51,$D3,$2F,$6C,$E5,$F4,$D7,$4B,$ED,$5C,$9B,$7E
	.dc.b	$3B,$E1,$3E,$B3,$47,$88,$EC,$7E,$F5,$DB,$F0,$CD,$C7,$92,$BC,$11
	.dc.b	$C5,$B1,$98,$DD,$F9,$9A,$78,$63,$C9,$7E,$69,$F6,$B6,$7C,$4D,$47
	.dc.b	$CC,$3C,$07,$43,$CD,$3D,$1C,$B4,$F2,$16,$7C,$8D,$87,$C2,$B8,$DF
	.dc.b	$21,$FE,$A5,$C3,$F9,$8B,$6B,$FA,$E7,$1B,$CD,$C6,$1C,$6A,$E1,$CB
	.dc.b	$0F,$E8,$B1,$C7,$A1,$87,$19,$C9,$FD,$2B,$E2,$78,$9B,$FA,$37,$B5
	.dc.b	$89,$D1,$34,$FF,$36,$CE,$A3,$1C,$7F,$D3,$AF,$85,$6A,$EC,$E6,$B6
	.dc.b	$7F,$95,$AF,$95,$7F,$5E,$BF,$81,$5F,$DC,$BA,$EF,$77,$F6,$4C,$3F
	.dc.b	$A7,$79,$F8,$99,$1F,$DC,$A7,$58,$7F,$7C,$AC,$BF,$DF,$BE,$8F,$E2
	.dc.b	$68,$FA,$3B,$18,$6E,$1F,$D6,$DB,$5F,$E0,$67,$C7,$E3,$C9,$D1,$E1
	.dc.b	$71,$3F,$E1,$B7,$A4,$64,$FE,$CD,$F4,$7E,$23,$C7,$F1,$D6,$32,$4A
	.dc.b	$FF,$0A,$C6,$5F,$F0,$BF,$C6,$FA,$8F,$5F,$0D,$AF,$EE,$51,$67,$FC
	.dc.b	$3B,$37,$7F,$E0,$90,$FD,$5D,$8F,$E3,$E3,$0F,$C6,$FA,$4C,$4E,$B1
	.dc.b	$FF,$1A,$56,$9B,$19,$8F,$F1,$5C,$F0,$A9,$5F,$E0,$5F,$A4,$B5,$FE
	.dc.b	$30,$7E,$9F,$CC,$49,$FF,$05,$7D,$25,$AF,$F1,$AC,$DC,$BF,$DA,$9C
	.dc.b	$2A,$C5,$FF,$A6,$FF,$73,$71,$FE,$29,$B1,$FE,$15,$FA,$0C,$4C,$8F
	.dc.b	$F1,$CC,$EB,$2F,$F2,$1D,$65,$FF,$63,$4F,$F8,$4F,$B1,$F3,$BF,$1D
	.dc.b	$B1,$FE,$57,$F8,$97,$E4,$8E,$8D,$5C,$AF,$93,$7C,$30,$EA,$30,$E3
	.dc.b	$7B,$8C,$9F,$C2,$2F,$C1,$F4,$F9,$DF,$ED,$E0,$8F,$81,$D3,$32,$FF
	.dc.b	$26,$FC,$0F,$8D,$F7,$C6,$71,$FF,$67,$B7,$0F,$F6,$1B,$FC,$26,$BE
	.dc.b	$E4,$7F,$B1,$B9,$31,$D3,$96,$61,$D9,$CB,$4F,$9B,$CD,$F4,$CE,$9D
	.dc.b	$36,$32,$4C,$BF,$DC,$31,$97,$E2,$E0,$E2,$5F,$97,$A7,$53,$FE,$DA
	.dc.b	$7F,$01,$CB,$31,$F8,$35,$70,$E7,$87,$E2,$67,$F0,$35,$EF,$AD,$61
	.dc.b	$FE,$E4,$E8,$F8,$47,$C3,$E8,$B7,$F3,$22,$E3,$FE,$F8,$9B,$5F,$53
	.dc.b	$E4,$BF,$0C,$7E,$1F,$E7,$2E,$1F,$82,$8B,$7F,$7B,$8D,$58,$BE,$57
	.dc.b	$F8,$8F,$1B,$0F,$E4,$DF,$AD,$A7,$F4,$7E,$77,$E5,$7F,$CB,$2F,$FD
	.dc.b	$FB,$39,$8C,$5F,$FC,$01,$FD,$D1,$E5,$FF,$30,$6C,$7E,$6E,$77,$A3
	.dc.b	$05,$6A,$3F,$33,$E2,$B9,$DD,$BF,$89,$3F,$15,$F0,$17,$2F,$F5,$3E
	.dc.b	$0F,$84,$F0,$1F,$8F,$FA,$9B,$3A,$F4,$3C,$86,$33,$FC,$1D,$A3,$F2
	.dc.b	$DF,$1D,$2B,$81,$FA,$07,$D9,$E5,$C8,$FD,$1F,$CC,$5C,$39,$CD,$DF
	.dc.b	$FC,$51,$BF,$AE,$B1,$7F,$1E,$BF,$49,$F5,$38,$C3,$FD,$A1,$C0,$8F
	.dc.b	$95,$E5,$4B,$FE,$09,$F0,$EF,$80,$C6,$1E,$AA,$BA,$7F,$8D,$F9,$D7
	.dc.b	$BC,$3E,$3D,$69,$FF,$0A,$EA,$FD,$3B,$FC,$B7,$1A,$3F,$D9,$6E,$31
	.dc.b	$FD,$4F,$EE,$BF,$8C,$9F,$F6,$BB,$63,$F8,$8E,$3F,$2F,$E5,$FF,$B8
	.dc.b	$CE,$73,$FF,$25,$07,$17,$FE,$26,$8C,$76,$7E,$47,$8D,$7C,$A2,$57
	.dc.b	$F2,$59,$7C,$16,$32,$F0,$C6,$27,$CB,$F2,$DE,$23,$F3,$9F,$9A,$3E
	.dc.b	$6B,$13,$A2,$5F,$AA,$75,$18,$CF,$D3,$A3,$0E,$7A,$72,$C3,$83,$F0
	.dc.b	$3E,$32,$61,$FD,$73,$CB,$B5,$F2,$A6,$E3,$FD,$AB,$D3,$56,$2F,$E2
	.dc.b	$64,$BF,$F1,$97,$0F,$E9,$F0,$F1,$C6,$3C,$1F,$9F,$F1,$73,$0F,$E8
	.dc.b	$BA,$DD,$3E,$CC,$DC,$3F,$9F,$CD,$AF,$F7,$37,$D2,$FE,$1E,$FF,$A3
	.dc.b	$B6,$BF,$D0,$5E,$33,$EC,$1F,$31,$FD,$FE,$B5,$87,$2E,$BF,$F0,$2E
	.dc.b	$36,$3F,$D3,$2A,$7B,$3F,$EA,$26,$BF,$02,$7F,$12,$7F,$95,$BC,$AF
	.dc.b	$DB,$57,$E4,$23,$2D,$3F,$14,$7F,$9C,$31,$4F,$0A,$FC,$CD,$7F,$98
	.dc.b	$3D,$87,$C6,$D3,$FC,$77,$E3,$DC,$79,$DF,$F9,$BA,$B7,$18,$C5,$18
	.dc.b	$9F,$E3,$3E,$CA,$F9,$47,$F7,$4A,$DE,$18,$DB,$FC,$23,$57,$1F,$46
	.dc.b	$3D,$1E,$CD,$C3,$F5,$BA,$FE,$75,$E9,$19,$BF,$AA,$DA,$E6,$FF,$3B
	.dc.b	$AE,$13,$A8,$FF,$1B,$7D,$27,$FA,$0F,$E0,$BE,$AB,$70,$FF,$67,$B5
	.dc.b	$FF,$6B,$B5,$FF,$4E,$8F,$85,$7C,$B3,$E9,$7E,$FC,$61,$B1,$99,$FF
	.dc.b	$6D,$CE,$5E,$FE,$F7,$FB,$B1,$BF,$0D,$E4,$8C,$47,$89,$AF,$DF,$FA
	.dc.b	$4F,$D4,$BF,$55,$AF,$D1,$6E,$9F,$C9,$7E,$95,$F0,$7E,$A8,$FC,$EC
	.dc.b	$EB,$83,$A7,$7E,$5B,$87,$F8,$D7,$0E,$2F,$7F,$73,$F3,$BC,$AC,$11
	.dc.b	$86,$CF,$FB,$EA,$75,$39,$3B,$47,$FA,$F2,$76,$FA,$63,$3E,$4F,$E3
	.dc.b	$7E,$3F,$77,$CA,$BC,$44,$3F,$B5,$F3,$27,$AE,$8F,$A0,$FE,$5B,$5F
	.dc.b	$9E,$8C,$78,$BA,$9C,$CF,$E6,$7E,$97,$F3,$8E,$E3,$35,$FC,$F1,$BB
	.dc.b	$7F,$3B,$6C,$FF,$B5,$AB,$8B,$FE,$52,$FA,$5F,$F6,$2B,$FE,$82,$FA
	.dc.b	$5F,$E1,$6D,$EB,$F8,$6A,$76,$F1,$F8,$1F,$E1,$6E,$5B,$1A,$8F,$CA
	.dc.b	$99,$3F,$88,$FD,$2B,$DF,$E2,$D8,$CC,$60,$8F,$FA,$5A,$34,$FF,$9A
	.dc.b	$9B,$A7,$FA,$8B,$E9,$5E,$C7,$C9,$E0,$FA,$FF,$41,$FD,$3F,$0F,$E1
	.dc.b	$7C,$88,$C4,$7F,$95,$38,$3E,$B3,$F3,$6B,$F9,$3E,$3F,$5D,$B3,$A9
	.dc.b	$F6,$CE,$5F,$79,$FE,$2D,$8D,$47,$D5,$9F,$24,$FF,$AC,$3E,$6B,$FC
	.dc.b	$14,$FD,$BF,$2B,$F2,$3E,$AF,$F7,$0F,$92,$BC,$C7,$F0,$93,$2B,$BE
	.dc.b	$3F,$8B,$8F,$29,$23,$39,$23,$84,$7B,$DE,$BE,$A7,$E3,$FE,$67,$C3
	.dc.b	$F4,$BE,$4A,$FF,$05,$B7,$0F,$C7,$4E,$E7,$AB,$F1,$08,$E0,$FC,$B3
	.dc.b	$B7,$9F,$93,$FA,$2F,$A4,$FF,$5E,$C7,$D4,$E0,$FE,$F8,$DF,$43,$5E
	.dc.b	$24,$8F,$7D,$70,$7E,$8B,$5F,$E1,$4D,$7F,$C3,$8E,$F1,$FF,$10,$B7
	.dc.b	$B7,$35,$F4,$3E,$95,$9F,$97,$AF,$F1,$3C,$B3,$2F,$DB,$63,$FA,$2C
	.dc.b	$B9,$D7,$E1,$A3,$F4,$13,$E0,$8C,$11,$E2,$F5,$7E,$07,$A7,$F9,$6F
	.dc.b	$91,$91,$F0,$DF,$8D,$E0,$C8,$F8,$2F,$72,$3F,$13,$E0,$7C,$44,$70
	.dc.b	$7E,$53,$AB,$F4,$1C,$BF,$A4,$24,$7D,$1E,$3F,$E2,$BF,$47,$FE,$8A
	.dc.b	$8F,$C6,$B5,$FC,$FB,$5F,$E7,$0D,$7F,$CB,$7C,$9F,$DD,$1F,$F7,$57
	.dc.b	$B8,$FA,$7F,$CE,$B6,$3F,$01,$E9,$7F,$0E,$FA,$67,$D3,$5F,$1A,$57
	.dc.b	$F8,$08,$4E,$9C,$35,$FD,$8B,$5F,$E2,$2D,$7F,$B3,$B5,$1D,$27,$C4
	.dc.b	$47,$AF,$D1,$FF,$0A,$FA,$29,$1F,$97,$7F,$86,$FB,$18,$8F,$F5,$4E
	.dc.b	$89,$C7,$3F,$FA,$4A,$3F,$55,$D9,$FE,$6F,$DF,$FB,$BD,$FF,$9F,$B4
	.dc.b	$FE,$9B,$EE,$FF,$32,$8F,$F7,$3C,$E9,$BE,$AF,$D6,$23,$FA,$9E,$CF
	.dc.b	$B3,$F4,$DF,$C8,$3E,$8E,$4F,$F4,$DB,$4E,$A3,$EE,$BF,$C7,$A3,$31
	.dc.b	$82,$1E,$28,$E0,$FF,$B1,$BE,$21,$9F,$F6,$45,$F4,$74,$93,$FC,$C2
	.dc.b	$1F,$41,$88,$E5,$37,$7F,$F5,$BD,$6A,$7D,$1F,$4D,$D7,$E7,$E4,$60
	.dc.b	$87,$C6,$8F,$CC,$9E,$3F,$17,$F1,$9F,$17,$F8,$DF,$90,$95,$F8,$88
	.dc.b	$CB,$FE,$64,$E9,$FF,$5B,$B6,$3F,$AE,$C6,$23,$F1,$B1,$C1,$FF,$75
	.dc.b	$6B,$F8,$C8,$FE,$73,$F4,$CE,$9F,$D5,$FC,$6F,$F8,$23,$E1,$1F,$23
	.dc.b	$4C,$E3,$F5,$66,$BF,$80,$7F,$C1,$FE,$2F,$FA,$46,$3F,$AC,$AF,$1B
	.dc.b	$24,$8C,$FD,$35,$7F,$C2,$31,$78,$FF,$DC,$74,$AF,$EF,$BE,$A7,$F2
	.dc.b	$0B,$EB,$E3,$78,$0C,$F5,$C8,$FF,$44,$CE,$DF,$EB,$ED,$7F,$BF,$B5
	.dc.b	$FF,$2A,$BB,$C3,$F0,$F1,$FB,$08,$C7,$06,$AB,$DF,$E3,$7C,$0F,$12
	.dc.b	$5F,$C0,$7E,$1F,$85,$88,$FF,$16,$C5,$C7,$DB,$3F,$23,$C3,$95,$F2
	.dc.b	$1F,$97,$86,$23,$FC,$6B,$16,$BE,$2C,$FF,$8D,$B7,$FF,$6E,$D7,$F6
	.dc.b	$E6,$BF,$E0,$06,$BF,$E5,$96,$BF,$EE,$B7,$73,$FE,$F0,$6E,$59,$F1
	.dc.b	$7E,$EF,$81,$F8,$58,$B7,$F7,$A3,$0F,$F8,$63,$8F,$E4,$BF,$E4,$C8
	.dc.b	$DB,$FE,$A0,$9F,$47,$99,$FE,$18,$C3,$FC,$2D,$3E,$58,$C1,$1F,$F2
	.dc.b	$84,$7F,$85,$A7,$D1,$3F,$97,$EE,$FF,$2E,$FD,$0B,$E9,$BF,$CD,$AD
	.dc.b	$C3,$FC,$D4,$D7,$FC,$7F,$EF,$23,$11,$FE,$96,$8C,$4F,$42,$47,$F8
	.dc.b	$D3,$E9,$BF,$38,$FF,$AF,$6B,$A6,$DF,$EE,$06,$E5,$FE,$CD,$6C,$BC
	.dc.b	$C5,$F2,$8F,$F6,$04,$EB,$1F,$0C,$75,$C7,$FC,$F7,$5B,$C7,$B7,$D3
	.dc.b	$7B,$89,$5E,$E6,$BD,$1F,$8A,$7E,$F7,$C5,$FF,$33,$FC,$3F,$F8,$DE
	.dc.b	$BF,$A6,$FD,$2B,$E6,$FF,$D6,$F2,$F6,$3F,$D7,$33,$9C,$7F,$A9,$CE
	.dc.b	$5E,$35,$C3,$DA,$61,$89,$DC,$61,$7C,$87,$37,$FD,$89,$8F,$FB,$1A
	.dc.b	$31,$87,$59,$DE,$1D,$70,$FE,$C7,$5A,$C3,$A6,$3C,$31,$F3,$DF,$A9
	.dc.b	$B8,$78,$B0,$F2,$3F,$ED,$3B,$E3,$5B,$8C,$E3,$E1,$C3,$B5,$E2,$FD
	.dc.b	$58,$F7,$BF,$1E,$1C,$5F,$39,$DC,$A7,$7A,$EB,$8F,$92,$3E,$09,$D6
	.dc.b	$22,$C3,$B3,$6B,$B4,$79,$23,$07,$77,$64,$8F,$91,$1A,$3C,$47,$93
	.dc.b	$1F,$9D,$87,$63,$A1,$D6,$B9,$4F,$E0,$30,$F4,$57,$E2,$63,$A3,$6B
	.dc.b	$95,$FC,$37,$9B,$E2,$73,$3C,$CF,$92,$7C,$77,$D9,$F1,$5F,$E3,$BA
	.dc.b	$EC,$DC,$BF,$14,$DC,$BF,$C4,$AF,$F8,$F7,$E9,$AD,$FF,$A7,$9B,$87
	.dc.b	$FA,$F9,$AF,$F8,$B0,$38,$7F,$C5,$EF,$91,$FF,$18,$FC,$77,$F4,$49
	.dc.b	$FC,$C4,$68,$FB,$A2,$35,$97,$86,$33,$18,$7C,$46,$39,$3B,$3C,$F5
	.dc.b	$EB,$3F,$D4,$9E,$F1,$E6,$3E,$D1,$DE,$7F,$36,$F8,$EF,$8C,$74,$8E
	.dc.b	$11,$87,$CE,$73,$9C,$1D,$B0,$E7,$7E,$1C,$3D,$53,$A6,$9F,$EE,$88
	.dc.b	$D5,$7B,$27,$4D,$9D,$61,$E6,$8F,$1C,$61,$F1,$DF,$C2,$BF,$8D,$1F
	.dc.b	$56,$3C,$75,$C5,$B1,$F0,$8F,$55,$73,$BF,$54,$7C,$6B,$F2,$BC,$CF
	.dc.b	$A0,$78,$0B,$87,$09,$F4,$C6,$DF,$5F,$17,$F4,$7F,$98,$FF,$02,$E9
	.dc.b	$FC,$FD,$B9,$FF,$9D,$B9,$5D,$4F,$3B,$39,$31,$E5,$7F,$D2,$70,$FE
	.dc.b	$CF,$3F,$D1,$DA,$EF,$9D,$18,$AE,$33,$FD,$6A,$5F,$1B,$DA,$7F,$AE
	.dc.b	$3E,$0C,$9F,$4C,$F7,$35,$7C,$32,$FE,$B5,$78,$38,$56,$E3,$24,$C7
	.dc.b	$FA,$BC,$EA,$F8,$4E,$67,$FB,$55,$F5,$7C,$D5,$E9,$8C,$47,$F8,$4F
	.dc.b	$C9,$F8,$BB,$3B,$BF,$C9,$18,$7F,$82,$FD,$0C,$C6,$2F,$FC,$B6,$7F
	.dc.b	$4E,$70,$F3,$9F,$F1,$15,$FC,$AB,$E8,$FC,$CD,$3E,$AF,$D2,$FC,$5E
	.dc.b	$17,$D1,$D8,$C3,$70,$FF,$0E,$37,$1F,$0D,$FD,$E8,$F2,$F8,$F2,$B6
	.dc.b	$7F,$86,$1C,$57,$F8,$CD,$F2,$66,$69,$E5,$C5,$F4,$7E,$53,$CD,$F2
	.dc.b	$37,$0F,$F1,$74,$5C,$3C,$74,$AC,$6B,$FE,$5A,$CF,$F2,$7F,$1F,$EF
	.dc.b	$3E,$07,$15,$CB,$FC,$71,$17,$1E,$73,$6F,$FC,$77,$F3,$1E,$4B,$FF
	.dc.b	$6F,$36,$BC,$3F,$45,$F9,$E5,$19,$AE,$3A,$7E,$72,$72,$D9,$E5,$1F
	.dc.b	$CC,$CC,$DD,$7F,$DC,$31,$CE,$3F,$5D,$A3,$E8,$BC,$13,$9A,$FF,$17
	.dc.b	$4E,$6B,$FC,$A9,$F2,$FF,$7C,$FC,$07,$D1,$58,$C9,$2B,$E5,$C6,$5F
	.dc.b	$C7,$E4,$FA,$37,$37,$E6,$1B,$7F,$E6,$1F,$A2,$C4,$C8,$FF,$32,$7C
	.dc.b	$3D,$C3,$EB,$BF,$1F,$B1,$F4,$B4,$FE,$0F,$F2,$0F,$A6,$75,$E7,$16
	.dc.b	$3F,$A2,$3F,$86,$5F,$CD,$C3,$9C,$FC,$99,$CB,$63,$31,$F3,$5B,$AF
	.dc.b	$F9,$E1,$7D,$37,$79,$ED,$18,$AF,$0D,$7F,$22,$DB,$F5,$31,$83,$2E
	.dc.b	$2B,$FD,$E5,$8F,$C7,$EC,$95,$FE,$EF,$6C,$FF,$BD,$DF,$E3,$1F,$07
	.dc.b	$F6,$FA,$8D,$3F,$3D,$5F,$EF,$7E,$7D,$91,$FF,$7D,$75,$DC,$3E,$CE
	.dc.b	$4E,$9F,$B0,$76,$EB,$B7,$E9,$AF,$F2,$CD,$6E,$3F,$25,$AF,$E0,$6F
	.dc.b	$F8,$64,$E6,$BF,$31,$39,$AD,$CE,$63,$F6,$59,$7F,$2B,$AF,$1D,$FB
	.dc.b	$BA,$FF,$15,$CF,$31,$F1,$6B,$F0,$2B,$17,$EA,$AF,$E7,$9F,$0E,$F8
	.dc.b	$AC,$E7,$FB,$F8,$C6,$3C,$B3,$FD,$16,$1F,$C6,$1B,$5E,$4E,$AF,$9F
	.dc.b	$DB,$2F,$83,$F4,$A3,$17,$F8,$2F,$96,$FC,$C1,$F9,$9F,$A3,$B7,$F2
	.dc.b	$67,$FA,$F4,$FE,$4C,$91,$F9,$16,$C7,$E3,$3C,$BE,$2D,$9F,$46,$BF
	.dc.b	$A2,$FA,$6F,$91,$9F,$CA,$8C,$70,$FF,$48,$FA,$6F,$F4,$5D,$FD,$66
	.dc.b	$E1,$F5,$1A,$FF,$8F,$ED,$18,$8F,$EF,$1B,$3C,$0F,$C8,$9F,$92,$F8
	.dc.b	$2B,$87,$E1,$70,$E3,$F1,$1F,$D3,$71,$34,$93,$E3,$F5,$34,$47,$07
	.dc.b	$F1,$DE,$6F,$CE,$F8,$5F,$A0,$3D,$93,$67,$C1,$7F,$E0,$26,$F9,$9F
	.dc.b	$A6,$F9,$7F,$84,$C8,$F8,$4F,$94,$8E,$0F,$C2,$7D,$36,$F7,$FC,$23
	.dc.b	$F2,$FC,$A9,$3F,$A3,$F1,$9F,$12,$F8,$1C,$E1,$FC,$93,$3F,$EF,$17
	.dc.b	$F4,$70,$F8,$84,$70,$7E,$73,$B1,$EB,$72,$E2,$FC,$C3,$FE,$02,$FA
	.dc.b	$47,$65,$3F,$91,$F0,$BE,$67,$18,$FD,$3A,$DB,$FB,$66,$E9,$FB,$46
	.dc.b	$BF,$81,$6B,$FC,$69,$DE,$29,$FE,$5B,$BF,$FB,$87,$E0,$BB,$DF,$F8
	.dc.b	$42,$31,$3E,$1F,$97,$F4,$1F,$CE,$3E,$95,$F4,$D8,$9D,$1E,$59,$FF
	.dc.b	$0A,$64,$69,$23,$E8,$57,$BD,$E3,$87,$86,$30,$FF,$54,$A7,$CB,$FC
	.dc.b	$93,$FB,$1F,$93,$6F,$E3,$47,$80,$8F,$C7,$6D,$7F,$87,$38,$7F,$53
	.dc.b	$AF,$F9,$7B,$0E,$B7,$F6,$7D,$BF,$AD,$F1,$72,$3F,$80,$79,$FE,$B7
	.dc.b	$EA,$46,$71,$FA,$4F,$C0,$24,$7F,$55,$FA,$5F,$C6,$37,$A3,$FC,$0D
	.dc.b	$18,$AF,$F0,$13,$5F,$F0,$33,$9C,$BF,$CE,$7F,$4E,$EE,$FF,$09,$4E
	.dc.b	$70,$FE,$AF,$39,$7F,$11,$87,$E3,$CF,$F3,$47,$92,$E7,$C4,$E8,$99
	.dc.b	$7F,$94,$34,$56,$E7,$39,$FE,$26,$BF,$CF,$DE,$C3,$E3,$3F,$CE,$8D
	.dc.b	$AF,$E7,$BE,$C3,$C5,$CD,$49,$FE,$A5,$97,$F9,$16,$3F,$CA,$D3,$9C
	.dc.b	$3C,$58,$73,$7F,$C5,$15,$BB,$FE,$F9,$39,$9F,$F6,$F4,$E5,$DC,$7F
	.dc.b	$94,$67,$BB,$63,$F3,$DF,$2E,$EB,$F4,$CC,$AD,$BF,$E1,$69,$D4,$60
	.dc.b	$FE,$85,$F4,$DF,$EA,$1F,$82,$FF,$09,$37,$0F,$F1,$E3,$5F,$F2,$03
	.dc.b	$5F,$F7,$A7,$C2,$BE,$61,$F4,$DF,$E1,$B8,$C3,$63,$33,$FE,$47,$9C
	.dc.b	$BD,$BD,$EF,$E1,$ED,$F1,$3C,$71,$88,$F0,$B5,$FF,$16,$ED,$FE,$75
	.dc.b	$E5,$FE,$91,$A6,$36,$FE,$45,$5F,$22,$72,$FF,$94,$3C,$AB,$7C,$63
	.dc.b	$44,$3F,$CE,$F1,$AA,$7C,$77,$F8,$D5,$B3,$F6,$A3,$04,$7E,$BB,$C4
	.dc.b	$F0,$F9,$9F,$ED,$0F,$2F,$13,$A2,$1F,$E8,$67,$CF,$CE,$C3,$F2,$53
	.dc.b	$FD,$75,$AF,$F9,$81,$B1,$FE,$8B,$FA,$5F,$90,$FF,$B8,$A3,$3A,$7F
	.dc.b	$91,$1B,$49,$ED,$CD,$E6,$BE,$37,$FE,$5A,$E2,$7C,$43,$E3,$7E,$5E
	.dc.b	$2F,$C3,$7F,$A0,$3F,$E0,$D6,$BF,$EC,$2A,$E3,$3F,$49,$B1,$BF,$2F
	.dc.b	$9D,$70,$CF,$F6,$D5,$AA,$FF,$35,$7D,$27,$FB,$95,$FF,$34,$35,$FF
	.dc.b	$49,$35,$FF,$4B,$35,$FF,$39,$87,$84,$FF,$03,$8C,$7A,$BE,$C9,$CE
	.dc.b	$7C,$3E,$97,$E1,$D7,$F9,$F3,$E2,$DF,$59,$F0,$BC,$C7,$D5,$D3,$E2
	.dc.b	$CE,$AB,$FC,$60,$D7,$FD,$50,$D7,$FD,$58,$D7,$CE,$D7,$C1,$5D,$08
	.dc.b	$FF,$8E,$E9,$F3,$BF,$53,$F2,$95,$FE,$43,$CE,$7C,$E5,$35,$13,$0F
	.dc.b	$F4,$95,$70,$6C,$7F,$A5,$54,$65,$FF,$60,$37,$7F,$94,$DD,$7E,$7B
	.dc.b	$65,$D2,$C7,$FC,$67,$1F,$E5,$39,$FF,$5B,$BF,$E1,$1F,$63,$D0,$7A
	.dc.b	$FB,$E4,$79,$9F,$24,$F0,$8F,$1C,$60,$DB,$FE,$B1,$8D,$63,$FE,$71
	.dc.b	$F2,$BF,$33,$F5,$6F,$AC,$FC,$DE,$3F,$E9,$1C,$3F,$CC,$CD,$7F,$D7
	.dc.b	$53,$B7,$FC,$D4,$D7,$FD,$C4,$D7,$FD,$CC,$D7,$F2,$2D,$7F,$D8,$AF
	.dc.b	$1B,$FC,$93,$74,$DD,$F9,$3E,$63,$F5,$FF,$86,$3E,$47,$F8,$5F,$E9
	.dc.b	$09,$A6,$A2,$65,$FE,$CD,$AE,$0D,$8F,$B9,$18,$8F,$F7,$D3,$72,$FD
	.dc.b	$23,$71,$FD,$4B,$67,$ED,$E0,$F9,$8F,$C7,$7C,$C4,$9F,$F2,$53,$E8
	.dc.b	$F9,$C9,$1E,$37,$C2,$47,$FC,$DD,$C5,$FD,$2D,$FF,$4B,$60,$9E,$31
	.dc.b	$FE,$21,$F3,$3F,$39,$E0,$C8,$FF,$2C,$3F,$9F,$FA,$3F,$8F,$E4,$7B
	.dc.b	$89,$3F,$0D,$FF,$78,$73,$FE,$4F,$E6,$FF,$4C,$49,$FF,$31,$7C,$27
	.dc.b	$F1,$EF,$AB,$7C,$4F,$F2,$0F,$AC,$FF,$24,$D7,$FA,$AD,$AF,$FB,$FE
	.dc.b	$76,$FF,$AB,$DA,$FF,$12,$01,$CB,$96,$BF,$C6,$1A,$F0,$79,$F2,$7F
	.dc.b	$19,$6F,$26,$EF,$CD,$EB,$7F,$01,$FA,$8F,$E0,$5F,$39,$31,$FF,$3F
	.dc.b	$46,$5F,$F4,$04,$65,$FF,$27,$47,$F0,$9F,$0B,$F8,$57,$89,$27,$FD
	.dc.b	$0D,$19,$7F,$D1,$11,$97,$FC,$AD,$1F,$C4,$7C,$5F,$E2,$5F,$11,$27
	.dc.b	$FD,$1D,$19,$7F,$D2,$11,$97,$FC,$BD,$19,$7F,$CC,$11,$9D,$7F,$DD
	.dc.b	$53,$FD,$9B,$F7,$77,$CC,$FF,$1F,$F0,$26,$1F,$EC,$39,$CB,$67,$81
	.dc.b	$1F,$F7,$DF,$CA,$7D,$5C,$3F,$C8,$DC,$A8,$7B,$7F,$DE,$97,$CD,$53
	.dc.b	$11,$2F,$9C,$65,$EB,$1F,$8B,$F5,$7F,$15,$F2,$72,$7F,$0D,$19,$7F
	.dc.b	$0C,$7D,$F7,$F9,$97,$37,$F8,$0E,$32,$7E,$26,$32,$71,$7F,$24,$FE
	.dc.b	$5F,$3F,$41,$E9,$F8,$D9,$3F,$02,$32,$FC,$28,$CB,$FE,$05,$E7,$FC
	.dc.b	$BE,$9F,$0E,$BF,$C0,$3F,$BB,$FF,$77,$9C,$FE,$9C,$CF,$FC,$3E,$DC
	.dc.b	$3F,$3F,$BF,$F6,$FB,$E1,$1F,$A2,$6A,$69,$AC,$54,$F9,$5F,$F0,$8F
	.dc.b	$D2,$FF,$8F,$AF,$CC,$7F,$86,$FE,$BD,$F4,$BF,$E2,$A6,$C7,$8B,$4F
	.dc.b	$F0,$1D,$71,$CF,$FC,$04,$F3,$9F,$DB,$35,$DF,$C0,$FF,$63,$8C,$47
	.dc.b	$EE,$9E,$55,$A8,$EC,$66,$7F,$11,$AF,$E5,$F1,$FC,$56,$52,$BF,$2D
	.dc.b	$19,$69,$0F,$6E,$9F,$93,$ED,$FD,$07,$D3,$FE,$4F,$DC,$95,$F9,$F8
	.dc.b	$CB,$48,$72,$E7,$FE,$2F,$B7,$FA,$8C,$74,$3E,$07,$F1,$B1,$86,$C6
	.dc.b	$6B,$F2,$B3,$96,$C6,$5F,$A7,$19,$6C,$66,$3F,$CA,$DF,$1D,$FE,$4F
	.dc.b	$24,$FC,$8C,$DF,$00,$9F,$F2,$A7,$C3,$EA,$BE,$1C,$E5,$E0,$D7,$31
	.dc.b	$FE,$6A,$FD,$E3,$FC,$7D,$1F,$E4,$2C,$11,$88,$FC,$AB,$5F,$F3,$23
	.dc.b	$63,$7C,$7C,$C9,$3E,$0F,$9C,$FF,$45,$FE,$DD,$F9,$F8,$FF,$5F,$B5
	.dc.b	$FF,$5E,$B5,$FF,$4C,$B5,$7C,$3A,$39,$5F,$FB,$5E,$31,$CB,$D8,$FF
	.dc.b	$3B,$7B,$D6,$FF,$DA,$77,$C3,$1E,$B5,$A9,$CB,$B7,$FD,$CD,$3B,$7F
	.dc.b	$C6,$BA,$7F,$91,$9E,$7F,$2B,$FC,$3B,$1F,$23,$68,$D2,$4F,$02,$1F
	.dc.b	$5B,$E6,$7F,$D0,$FF,$16,$F5,$FF,$D2,$12,$FA,$7F,$DE,$3F,$D9,$27
	.dc.b	$FD,$FC,$FF,$66,$8C,$A8,$CB,$C7,$AF,$FD,$A7,$5D,$2B,$75,$DF,$7F
	.dc.b	$ED,$FC,$78,$F7,$FF,$D0,$6D,$C7,$38,$F2,$AB,$1E,$3C,$7A,$4F,$FB
	.dc.b	$82,$FF,$D1,$78,$F1,$AF,$11,$BB,$F4,$57,$38,$C9,$E9,$AE,$31,$CE
	.dc.b	$36,$71,$37,$8F,$A5,$FD,$14,$6D,$A7,$C7,$9F,$5C,$77,$C7,$E0,$E3
	.dc.b	$E8,$AD,$1D,$8F,$11,$6F,$B6,$3C,$31,$C6,$3E,$6A,$E2,$79,$2F,$A3
	.dc.b	$F4,$DB,$7F,$52,$FE,$21,$CE,$78,$5F,$CC,$7C,$27,$A8,$B8,$6C,$E1
	.dc.b	$3C,$B1,$F8,$0E,$1F,$71,$F0,$E3,$E9,$1C,$5C,$CF,$A2,$33,$18,$8F
	.dc.b	$45,$7F,$8C,$BE,$33,$F8,$46,$1F,$3D,$B8,$FE,$59,$B3,$BC,$3F,$BA
	.dc.b	$FC,$55,$C3,$FC,$78,$D7,$FD,$9E,$D7,$FC,$7E,$D7,$FD,$D6,$7C,$13
	.dc.b	$FC,$1B,$C5,$F1,$5D,$99,$F2,$11,$98,$E3,$5F,$A4,$78,$9C,$CC,$E7
	.dc.b	$FE,$4F,$AE,$31,$9B,$FC,$13,$C3,$4E,$4F,$91,$E9,$3C,$1C,$CF,$96
	.dc.b	$39,$DE,$2B,$CD,$87,$96,$BA,$D7,$1C,$3E,$3D,$7E,$88,$F0,$38,$C3
	.dc.b	$BD,$F9,$27,$C5,$3E,$98,$C5,$F4,$6C,$F3,$BE,$CF,$D2,$3C,$67,$BC
	.dc.b	$B8,$79,$F0,$EF,$1E,$33,$B9,$FE,$74,$3C,$A7,$37,$67,$1C,$3C,$91
	.dc.b	$C7,$0F,$89,$87,$C1,$8E,$F3,$ED,$37,$5C,$63,$B6,$1E,$8E,$1F,$E8
	.dc.b	$1E,$F7,$F7,$3C,$3F,$A2,$B6,$BF,$9A,$FC,$0D,$AD,$D7,$F4,$0A,$FE
	.dc.b	$B5,$18,$AF,$EA,$31,$8E,$3F,$F6,$45,$7F,$93,$2B,$FA,$77,$D4,$F3
	.dc.b	$93,$DF,$AF,$FB,$8F,$1E,$62,$5F,$F4,$E7,$5C,$C9,$CE,$5F,$D3,$26
	.dc.b	$F0,$FF,$9A,$AB,$57,$C4,$95,$FD,$9A,$BC,$C4,$8F,$F0,$6F,$57,$FB
	.dc.b	$35,$FF,$15,$DF,$F7,$8F,$7F,$C3,$CB,$F1,$63,$13,$AA,$E1,$87,$F9
	.dc.b	$ED,$AF,$F8,$05,$AF,$F8,$02,$FF,$C0,$59,$FF,$84,$3D,$CF,$69,$EE
	.dc.b	$F7,$2E,$1F,$E0,$E8,$B7,$FE,$12,$9B,$8F,$F8,$AF,$17,$D1,$79,$5B
	.dc.b	$87,$13,$CD,$19,$3F,$C2,$53,$FE,$14,$6C,$F0,$BF,$54,$63,$E8,$BA
	.dc.b	$1F,$5F,$C3,$FE,$E7,$87,$C1,$8C,$BF,$E9,$CF,$1E,$DF,$F9,$05,$F2
	.dc.b	$7D,$0C,$9F,$F5,$0F,$CD,$5A,$FF,$1C,$E4,$AE,$DC,$D9,$D3,$FD,$4B
	.dc.b	$9F,$E3,$2F,$FC,$E2,$3E,$99,$F1,$7F,$99,$8C,$3E,$7F,$27,$13,$AD
	.dc.b	$3F,$CC,$F5,$A7,$FD,$3B,$19,$8F,$F5,$1B,$5F,$F7,$53,$5F,$DC,$3B
	.dc.b	$C3,$FC,$B9,$F3,$6D,$7B,$E3,$F8,$78,$C6,$1E,$0F,$80,$7C,$9F,$9D
	.dc.b	$B1,$FE,$53,$F8,$87,$C6,$66,$A4,$E7,$E9,$6E,$BF,$E7,$6B,$47,$D5
	.dc.b	$F9,$CF,$AD,$F2,$F6,$FF,$D9,$ED,$BF,$F3,$A3,$76,$F9,$AF,$C9,$F0
	.dc.b	$7E,$EC,$FC,$E6,$CF,$F3,$CF,$0D,$FA,$BB,$E0,$F8,$79,$B3,$2F,$F1
	.dc.b	$D3,$FE,$5D,$C3,$9E,$3F,$C7,$CF,$F2,$8C,$64,$F8,$D9,$F1,$8F,$F3
	.dc.b	$DF,$0A,$3F,$21,$E9,$FE,$47,$C3,$B8,$FF,$BA,$9B,$7F,$E8,$F6,$EB
	.dc.b	$F8,$87,$F4,$9C,$32,$BF,$DE,$3E,$B5,$BF,$C8,$CF,$E7,$7E,$2F,$18
	.dc.b	$7E,$41,$FD,$07,$C5,$7F,$17,$FD,$57,$FB,$2E,$3F,$EF,$A8,$C3,$A3
	.dc.b	$FD,$19,$1A,$C3,$FD,$45,$3A,$9E,$06,$8E,$11,$FE,$B0,$9E,$17,$FD
	.dc.b	$AB,$C5,$43,$E3,$FC,$5E,$3F,$8D,$B8,$78,$1B,$87,$36,$E1,$CE,$DF
	.dc.b	$25,$EB,$EA,$F6,$37,$0F,$4B,$70,$F2,$36,$FF,$BE,$C7,$FA,$2B,$87
	.dc.b	$F1,$07,$F4,$9E,$77,$C4,$3C,$1C,$E3,$FB,$0E,$7F,$B8,$FF,$9F,$EF
	.dc.b	$F0,$C4,$70,$7F,$B7,$2F,$F1,$FF,$11,$CB,$1F,$F7,$8C,$62,$7E,$7F
	.dc.b	$67,$D5,$3E,$B7,$D3,$D8,$FD,$FF,$E2,$3F,$B9,$B0,$70,$DF,$07,$FD
	.dc.b	$69,$1B,$E4,$FA,$3E,$57,$E0,$4F,$C1,$4D,$BF,$BF,$E5,$7F,$11,$B3
	.dc.b	$F8,$73,$04,$8F,$D1,$61,$E2,$9D,$C6,$23,$C9,$3E,$23,$FA,$3F,$23
	.dc.b	$E5,$FF,$24,$73,$9F,$C4,$3E,$1F,$9A,$97,$FE,$DD,$FD,$3D,$DB,$8B
	.dc.b	$FE,$53,$92,$FD,$E4,$BF,$14,$65,$B1,$FD,$DE,$DD,$9C,$F9,$FF,$2D
	.dc.b	$18,$AF,$E5,$A3,$CF,$FE,$68,$D3,$E8,$CF,$F9,$19,$7E,$FC,$F8,$3D
	.dc.b	$D4,$9F,$88,$71,$CF,$E4,$B5,$FE,$18,$D7,$FC,$58,$EE,$FB,$7C,$1B
	.dc.b	$E3,$7A,$52,$73,$1F,$D2,$7D,$CF,$D2,$DF,$F4,$66,$D7,$ED,$7B,$35
	.dc.b	$DD,$2B,$EC,$56,$AB,$F8,$F3,$5F,$F6,$0C,$FF,$BF,$E3,$9E,$1F,$B3
	.dc.b	$F4,$79,$57,$F0,$17,$C5,$E0,$4C,$3F,$85,$FC,$FD,$AF,$EC,$38,$7F
	.dc.b	$46,$9F,$0D,$75,$23,$83,$F8,$9F,$C9,$71,$C7,$F8,$13,$FD,$9F,$F0
	.dc.b	$C7,$CA,$D3,$FC,$85,$CC,$FC,$4F,$FC,$09,$7F,$C0,$2B,$78,$7F,$6A
	.dc.b	$6E,$7F,$9C,$6B,$FC,$FD,$AF,$F9,$7D,$AF,$EC,$9E,$33,$FE,$88,$F1
	.dc.b	$1F,$86,$BE,$1F,$C9,$87,$E2,$0F,$DC,$CE,$A3,$FC,$01,$F4,$9F,$9E
	.dc.b	$9F,$EF,$AD,$C7,$EF,$47,$0F,$9C,$7C,$5F,$0E,$E9,$3F,$B1,$BE,$17
	.dc.b	$FD,$5D,$AF,$F9,$DD,$AB,$E8,$3F,$15,$E8,$7F,$84,$7F,$1A,$7C,$4F
	.dc.b	$F9,$AE,$73,$87,$F8,$42,$72,$FA,$EF,$96,$FF,$E6,$48,$D5,$F1,$8C
	.dc.b	$4F,$F8,$82,$B8,$BE,$C3,$F3,$5F,$BD,$3E,$A7,$FC,$AC,$DC,$3B,$79
	.dc.b	$4F,$85,$7C,$F6,$AE,$57,$F9,$82,$F8,$61,$D7,$0E,$17,$A8,$F0,$71
	.dc.b	$BF,$72,$7C,$DF,$FA,$1D,$B8,$F8,$33,$FB,$AF,$D8,$8C,$DF,$26,$EB
	.dc.b	$F8,$16,$CF,$F3,$6C,$BD,$1D,$5D,$71,$93,$FE,$8D,$AE,$78,$7F,$AA
	.dc.b	$F2,$F2,$57,$FA,$5A,$73,$3E,$1F,$91,$F3,$1F,$E8,$CF,$81,$B8,$7D
	.dc.b	$3C,$3D,$93,$8A,$F8,$5A,$2F,$1A,$7F,$62,$FA,$6F,$ED,$7E,$A7,$F3
	.dc.b	$96,$DF,$F3,$E6,$BF,$E2,$16,$BF,$EC,$1F,$91,$FE,$4D,$CB,$FD,$05
	.dc.b	$B1,$FD,$19,$AF,$F8,$A5,$AF,$FB,$26,$7F,$C8,$FC,$BE,$1F,$07,$F2
	.dc.b	$33,$E1,$6B,$E1,$6B,$FA,$AF,$77,$FA,$2B,$FE,$B7,$6B,$FE,$37,$D9
	.dc.b	$18,$8F,$9C,$D7,$97,$57,$FB,$6F,$4F,$DD,$4F,$26,$C6,$78,$3E,$E7
	.dc.b	$D3,$BE,$73,$18,$31,$EF,$EF,$63,$9F,$F3,$1F,$3F,$79,$0F,$0A,$DD
	.dc.b	$FD,$9D,$58,$7F,$6B,$9C,$CF,$AE,$72,$FB,$E7,$51,$88,$FC,$EB,$75
	.dc.b	$F5,$CF,$8F,$3F,$F3,$CD,$7F,$B1,$B3,$FE,$D4,$DF,$4F,$D7,$7E,$D7
	.dc.b	$D1,$C1,$8A,$DD,$7E,$E6,$B4,$F1,$3F,$51,$3F,$13,$8D,$F2,$8F,$8E
	.dc.b	$F8,$F4,$FA,$77,$A3,$86,$C7,$F0,$DB,$E3,$5C,$EF,$94,$FF,$A8,$AB
	.dc.b	$6F,$FA,$B1,$B8,$B8,$FF,$C4,$F8,$72,$AF,$F5,$4F,$D3,$FF,$84,$DF
	.dc.b	$F5,$5F,$D3,$FF,$2C,$6F,$8D,$E6,$9D,$47,$FA,$E1,$AF,$FA,$0D,$AF
	.dc.b	$F8,$D0,$38,$7F,$C6,$ED,$7F,$CB,$3F,$11,$F0,$BD,$D7,$D1,$53,$E2
	.dc.b	$FA,$9F,$E9,$5A,$FF,$5F,$B5,$FF,$64,$CF,$07,$FD,$86,$D7,$FD,$9B
	.dc.b	$1B,$D3,$FC,$BE,$DD,$3C,$CD,$9F,$F2,$5B,$5F,$F6,$9F,$C2,$7D,$6F
	.dc.b	$A7,$FF,$37,$C7,$F3,$2F,$A7,$FE,$B8,$DF,$3F,$57,$23,$04,$3F,$DB
	.dc.b	$F5,$C9,$B1,$B6,$9D,$7E,$6F,$C1,$E0,$FE,$A7,$07,$CC,$D3,$35,$FE
	.dc.b	$E2,$C3,$FD,$9E,$F9,$63,$4E,$6B,$6F,$89,$BA,$7F,$A5,$6D,$F3,$1F
	.dc.b	$94,$F4,$BF,$C5,$D7,$FE,$EE,$C3,$FD,$04,$DF,$6F,$5E,$E4,$8C,$10
	.dc.b	$FF,$7E,$D7,$26,$C6,$DA,$75,$F8,$AF,$D1,$FD,$3B,$E5,$BF,$84,$7C
	.dc.b	$9F,$FA,$9E,$BF,$DD,$95,$A9,$91,$82,$1F,$DA,$C5,$73,$7F,$BB,$18
	.dc.b	$8E,$2E,$E3,$4F,$FB,$CF,$E3,$BE,$87,$CC,$C8,$FC,$03,$E6,$F8,$79
	.dc.b	$18,$21,$CD,$FB,$1F,$09,$A8,$C1,$F8,$97,$EE,$78,$52,$3F,$18,$FD
	.dc.b	$FF,$5F,$E1,$F9,$36,$7F,$21,$18,$7F,$25,$EB,$FE,$14,$8F,$E5,$7D
	.dc.b	$4F,$E1,$BF,$4A,$F8,$2F,$E1,$FF,$4F,$FE,$0D,$AF,$86,$D7,$E4,$4E
	.dc.b	$DF,$8E,$D7,$E6,$46,$FA,$7F,$D9,$2D,$F0,$3F,$80,$56,$E7,$FD,$9A
	.dc.b	$F1,$AF,$E0,$81,$C3,$C2,$FF,$D5,$0E,$5F,$F5,$57,$A6,$EB,$FE,$8F
	.dc.b	$3F,$B1,$FA,$7F,$D2,$1F,$A5,$F4,$E4,$64,$87,$F8,$86,$78,$B7,$C3
	.dc.b	$FD,$67,$8D,$23,$F7,$91,$FA,$F3,$0F,$18,$C3,$FC,$07,$C6,$FD,$A7
	.dc.b	$C8,$C8,$C1,$0F,$03,$97,$C5,$1F,$BB,$31,$AF,$EC,$67,$13,$FE,$4B
	.dc.b	$87,$D0,$BE,$9B,$F8,$17,$7C,$AD,$4E,$5B,$3F,$ED,$1A,$E2,$D9,$D4
	.dc.b	$FD,$29,$CB,$FE,$DA,$DB,$FD,$FD,$ED,$FF,$77,$F0,$BF,$C8,$73,$FE
	.dc.b	$50,$6B,$FE,$4B,$51,$B3,$87,$BD,$27,$84,$7E,$4B,$D4,$FC,$8F,$BB
	.dc.b	$27,$F1,$B1,$97,$F2,$71,$97,$F2,$71,$97,$FD,$DF,$CB,$FE,$0F,$7F
	.dc.b	$DD,$CD,$73,$1F,$88,$C3,$DF,$97,$F6,$19,$64,$97,$D3,$FB,$CF,$81
	.dc.b	$AC,$AF,$81,$19,$7E,$14,$7F,$78,$FA,$7F,$ED,$BD,$1F,$E5,$DF,$A7
	.dc.b	$6B,$FE,$1F,$6D,$FF,$10,$AD,$A0,$FF,$78,$7F,$4A,$D7,$FB,$53,$5E
	.dc.b	$0D,$5F,$4D,$FE,$35,$BF,$F7,$9F,$D3,$BE,$8F,$FC,$32,$DA,$FE,$43
	.dc.b	$5B,$7C,$4D,$D9,$3B,$9F,$F0,$E3,$63,$E5,$47,$83,$E0,$7F,$C3,$35
	.dc.b	$FE,$49,$8E,$71,$FB,$A8,$C4,$6C,$FE,$02,$7F,$BD,$F9,$3A,$63,$FE
	.dc.b	$85,$DD,$F4,$DF,$91,$3F,$C6,$33,$AA,$E1,$9F,$E6,$C8,$F4,$CB,$F4
	.dc.b	$F1,$F9,$FF,$A3,$FC,$D7,$BB,$2B,$F4,$F1,$D0,$D3,$C6,$D1,$88,$FD
	.dc.b	$34,$7F,$9A,$1F,$F5,$17,$7F,$F9,$64,$F2,$FC,$0C,$8E,$4F,$EB,$C8
	.dc.b	$FE,$B7,$E0,$DF,$21,$FE,$61,$24,$7F,$AB,$7D,$CA,$FF,$A5,$A2,$E9
	.dc.b	$F9,$9E,$8F,$F6,$05,$7F,$95,$3E,$93,$E9,$CE,$70,$78,$5F,$E6,$D3
	.dc.b	$5C,$53,$2F,$8D,$7F,$69,$FF,$3F,$35,$FC,$A5,$A3,$11,$FE,$7B,$FA
	.dc.b	$5F,$CD,$BF,$EA,$AF,$0F,$FC,$F4,$7F,$9E,$63,$35,$FF,$00,$5D,$CF
	.dc.b	$FA,$ED,$AF,$FC,$00,$6B,$D7,$AB,$FC,$FD,$1C,$B0,$D4,$75,$3F,$D1
	.dc.b	$1E,$E5,$AF,$F5,$F5,$6D,$E4,$7F,$B5,$E3,$C9,$F4,$3F,$E8,$FF,$16
	.dc.b	$4F,$08,$CB,$F9,$4F,$87,$5A,$7C,$64,$9F,$F5,$E7,$A4,$CF,$F7,$5F
	.dc.b	$4D,$F9,$89,$FD,$E7,$C5,$7F,$8A,$9C,$F5,$FF,$B1,$7D,$CF,$F3,$25
	.dc.b	$6E,$54,$C3,$A4,$27,$53,$9C,$39,$CF,$6B,$E7,$87,$58,$C9,$D3,$0E
	.dc.b	$78,$72,$7A,$F3,$74,$BF,$F3,$B3,$6B,$F8,$A3,$BC,$2E,$1F,$ED,$DB
	.dc.b	$E6,$F7,$AF,$33,$70,$E1,$18,$3E,$1C,$F0,$7D,$07,$9B,$1F,$8C,$EC
	.dc.b	$F4,$5E,$67,$84,$7C,$6C,$7D,$EF,$23,$85,$7C,$56,$CF,$08,$EB,$7E
	.dc.b	$D9,$CD,$F5,$3C,$57,$D2,$3B,$19,$36,$76,$75,$5D,$30,$F8,$95,$BB
	.dc.b	$EC,$48,$CD,$F2,$8C,$3E,$97,$E3,$92,$3E,$51,$23,$E4,$CF,$8F,$E1
	.dc.b	$FF,$84,$C6,$03,$87,$E4,$98,$CB,$C3,$19,$8F,$EA,$1F,$0F,$FD,$46
	.dc.b	$31,$1F,$ED,$43,$81,$FE,$36,$8D,$4F,$F3,$86,$C6,$4F,$B0,$7F,$80
	.dc.b	$B8,$BF,$22,$F6,$33,$E4,$FF,$71,$CB,$F4,$11,$96,$C6,$72,$ED,$3C
	.dc.b	$2B,$9B,$E5,$AD,$C6,$2F,$BD,$F1,$AF,$25,$69,$F2,$9C,$6F,$FC,$A3
	.dc.b	$3B,$8F,$F6,$B5,$72,$BE,$91,$E1,$38,$C6,$5A,$66,$FD,$6F,$53,$FD
	.dc.b	$C1,$5D,$5F,$75,$F8,$2F,$AD,$79,$4E,$37,$D5,$B5,$D6,$BE,$1C,$6D
	.dc.b	$F8,$B8,$7B,$4D,$95,$D9,$DE,$7C,$F3,$C5,$F8,$B1,$98,$EE,$67,$83
	.dc.b	$F9,$0F,$B5,$FD,$0D,$B8,$FF,$25,$E1,$FF,$5A,$36,$75,$18,$C3,$F9
	.dc.b	$24,$FF,$87,$7A,$9D,$2F,$37,$FC,$65,$C1,$FD,$17,$1B,$7B,$AD,$3F
	.dc.b	$CF,$3E,$3F,$38,$FF,$51,$6E,$3D,$38,$F3,$A7,$F3,$89,$BA,$1A,$70
	.dc.b	$C7,$FD,$BB,$8F,$52,$63,$FE,$94,$BE,$51,$EF,$AF,$69,$1F,$EC,$A7
	.dc.b	$9F,$CE,$EA,$E2,$BF,$C9,$B6,$9F,$4E,$9F,$DA,$FD,$AF,$ED,$CD,$BF
	.dc.b	$EE,$71,$8E,$C3,$4F,$7F,$D9,$7F,$91,$E7,$FD,$09,$3A,$7C,$1B,$27
	.dc.b	$59,$7D,$23,$C7,$ED,$7C,$13,$E1,$7B,$56,$F8,$E8,$9C,$57,$1C,$3E
	.dc.b	$CD,$E3,$3F,$B3,$93,$E7,$BD,$04,$CB,$FC,$1A,$DC,$39,$67,$FC,$15
	.dc.b	$F2,$5F,$F8,$36,$7D,$BD,$1F,$1B,$07,$CF,$7C,$32,$63,$FE,$18,$6E
	.dc.b	$1F,$E9,$1F,$27,$84,$7E,$13,$CA,$FF,$30,$57,$C6,$F9,$9F,$98,$47
	.dc.b	$FC,$42,$DB,$FF,$4D,$73,$F8,$63,$93,$F9,$08,$7D,$3B,$E2,$7F,$C4
	.dc.b	$6D,$BF,$F2,$DC,$EA,$72,$48,$FF,$5D,$7D,$23,$E9,$FA,$5F,$1B,$F8
	.dc.b	$73,$9A,$FC,$54,$73,$78,$D7,$0B,$FC,$5D,$6E,$BE,$F0,$F7,$9F,$4B
	.dc.b	$FE,$38,$6E,$1E,$66,$CF,$F9,$C9,$AF,$FB,$D1,$CF,$17,$F9,$3F,$E1
	.dc.b	$9F,$49,$F0,$E7,$37,$FB,$D9,$CB,$FE,$C5,$FA,$5F,$A5,$F4,$FF,$4D
	.dc.b	$B1,$93,$FD,$95,$19,$69,$F0,$3B,$FF,$C6,$0F,$CC,$6E,$5F,$E3,$5F
	.dc.b	$2B,$FC,$A7,$18,$8F,$DC,$47,$C4,$BF,$B1,$3C,$33,$FE,$1F,$AA,$73
	.dc.b	$5B,$C3,$C0,$DC,$3F,$DB,$9E,$8F,$E2,$FC,$1F,$C6,$36,$BF,$DC,$4D
	.dc.b	$9D,$C7,$FB,$96,$7F,$35,$BF,$EE,$C9,$1F,$E7,$06,$CF,$FB,$A7,$E1
	.dc.b	$F7,$3F,$EE,$D7,$F3,$5F,$0F,$FC,$7B,$E9,$53,$83,$F7,$EE,$1F,$F7
	.dc.b	$A4,$6D,$D5,$7E,$E6,$74,$FF,$BD,$E7,$53,$C0,$E6,$77,$F8,$AF,$07
	.dc.b	$89,$63,$BB,$6B,$A3,$77,$FF,$4A,$61,$FC,$4E,$7C,$1F,$21,$E9,$F8
	.dc.b	$EF,$53,$70,$F4,$B6,$77,$1E,$C9,$FB,$3C,$1F,$78,$91,$F8,$36,$CF
	.dc.b	$F1,$DF,$01,$F4,$FF,$D2,$23,$F4,$7C,$9F,$7E,$FF,$9A,$56,$BC,$BE
	.dc.b	$B1,$8B,$E1,$A7,$FB,$72,$F8,$61,$D6,$73,$5F,$D6,$F4,$FC,$FC,$FC
	.dc.b	$DE,$9F,$AD,$F4,$FF,$5D,$FE,$11,$F4,$5F,$BD,$CB,$E2,$F6,$60,$FB
	.dc.b	$87,$DD,$F1,$66,$5F,$72,$2E,$3F,$67,$C3,$E6,$7E,$0C,$B3,$FC,$1B
	.dc.b	$0F,$E4,$EF,$E1,$3E,$9F,$35,$F8,$17,$F1,$44,$8E,$15,$67,$F8,$4D
	.dc.b	$3E,$67,$D2,$7F,$6E,$FA,$77,$C1,$E3,$0F,$04,$E7,$AB,$18,$7F,$47
	.dc.b	$B4,$63,$5F,$F6,$C4,$62,$F8,$9F,$D6,$1C,$57,$F1,$A8,$FE,$A4,$7F
	.dc.b	$22,$2B,$FD,$6D,$7D,$3B,$E9,$FF,$9B,$46,$AF,$F4,$F5,$FD,$9F,$CF
	.dc.b	$FF,$04,$CB,$E1,$3F,$BA,$3D,$63,$31,$89,$F1,$46,$3D,$2E,$AD,$AF
	.dc.b	$E6,$BE,$BB,$B7,$FA,$33,$63,$F6,$BE,$6E,$26,$47,$EE,$3C,$DB,$8F
	.dc.b	$ED,$5B,$9F,$EA,$9B,$3F,$8F,$7A,$F7,$C8,$FE,$03,$F4,$F6,$FF,$A0
	.dc.b	$E4,$AF,$E1,$1F,$23,$9C,$7F,$6A,$DC,$3F,$34,$DB,$7A,$8E,$D4,$FF
	.dc.b	$61,$FA,$67,$CA,$53,$FC,$67,$D2,$FA,$7F,$F0,$16,$5F,$D6,$2B,$6D
	.dc.b	$AF,$F1,$AF,$D3,$BE,$0F,$94,$FF,$89,$FC,$44,$E6,$38,$19,$71,$AF
	.dc.b	$F5,$EC,$3A,$BF,$E1,$DA,$E3,$7F,$DD,$FE,$9F,$F2,$F8,$7F,$74,$6D
	.dc.b	$08,$57,$2E,$03,$87,$6F,$2D,$FF,$CB,$3F,$52,$FA,$AF,$07,$81,$FE
	.dc.b	$3E,$8C,$61,$FE,$43,$FA,$25,$FF,$94,$E3,$FB,$EF,$D4,$FF,$AE,$1D
	.dc.b	$46,$72,$FE,$F7,$39,$7F,$C7,$55,$CB,$D1,$FF,$0A,$4E,$6B,$F3,$F3
	.dc.b	$9B,$EB,$F0,$BF,$E5,$0F,$A8,$7D,$4D,$3F,$CC,$5F,$3D,$6B,$FC,$B1
	.dc.b	$82,$B7,$18,$6E,$7F,$E4,$3F,$83,$5F,$18,$FA,$5F,$52,$F9,$CF,$F3
	.dc.b	$9B,$70,$FF,$26,$F3,$E2,$64,$7C,$FA,$72,$AF,$84,$7F,$9C,$DA,$FD
	.dc.b	$88,$D5,$7F,$A5,$67,$79,$FF,$00,$EC,$FF,$31,$4F,$6B,$FF,$28,$BE
	.dc.b	$4F,$8E,$98,$FF,$98,$3D,$FB,$5E,$A9,$E8,$49,$F3,$65,$E5,$8F,$E1
	.dc.b	$B3,$F8,$7F,$A9,$FE,$05,$EB,$7F,$90,$9B,$1E,$D6,$BB,$F1,$53,$FE
	.dc.b	$48,$F8,$DF,$F3,$87,$2D,$F2,$0D,$25,$F3,$F1,$3F,$DF,$ED,$E7,$F0
	.dc.b	$46,$23,$C8,$D7,$F4,$1A,$3E,$3F,$D3,$C5,$FC,$9A,$31,$BF,$E9,$23
	.dc.b	$13,$A8,$FF,$5D,$7C,$43,$E6,$B5,$1D,$27,$58,$7D,$A6,$E1,$D2,$3F
	.dc.b	$D2,$B7,$C5,$FE,$01,$F0,$1F,$D3,$1B,$DF,$D1,$FC,$1C,$65,$FF,$49
	.dc.b	$7D,$4B,$E7,$75,$82,$B5,$A7,$AD,$B9,$FC,$06,$E1,$FC,$7B,$0E,$4F
	.dc.b	$80,$87,$F0,$B8,$DF,$42,$77,$3F,$E9,$8F,$DE,$FF,$D4,$0F,$EA,$1A
	.dc.b	$E7,$DA,$F5,$57,$FA,$CF,$E7,$BD,$99,$2F,$6F,$FB,$89,$B9,$7F,$15
	.dc.b	$8D,$5F,$2D,$97,$D3,$CD,$7D,$7F,$C7,$CD,$FB,$AB,$E5,$B0,$D8,$FF
	.dc.b	$23,$CF,$13,$33,$23,$27,$FA,$43,$F1,$2F,$F7,$43,$7C,$8E,$D1,$88
	.dc.b	$FF,$47,$F8,$FF,$6D,$FF,$48,$C6,$75,$FD,$2B,$76,$FE,$5E,$D9,$FE
	.dc.b	$65,$5C,$6B,$F5,$7F,$56,$F7,$FF,$D5,$74,$8C,$3F,$EA,$E6,$BF,$C2
	.dc.b	$5A,$FF,$39,$73,$3D,$27,$E3,$46,$30,$FE,$83,$F5,$9F,$A7,$AD,$5F
	.dc.b	$FA,$C2,$B4,$FF,$47,$6A,$9E,$15,$FD,$2B,$EA,$FF,$DE,$F3,$FF,$00
	.dc.b	$7D,$47,$DC,$FF,$AA,$23,$0D,$7F,$D6,$F3,$C1,$BF,$13,$F6,$E3,$19
	.dc.b	$7F,$0E,$6E,$1F,$D6,$70,$EB,$F0,$DB,$B9,$87,$FB,$22,$F6,$AB,$6E
	.dc.b	$7C,$6F,$CA,$FD,$3F,$F7,$38,$FC,$86,$3F,$D9,$7A,$93,$99,$F3,$D7
	.dc.b	$94,$8F,$33,$89,$FE,$0A,$87,$D6,$EA,$38,$9B,$3F,$B1,$DF,$79,$DC
	.dc.b	$7F,$B8,$27,$FC,$22,$FF,$B8,$BE,$D9,$F3,$3F,$9F,$6C,$6F,$DC,$FD
	.dc.b	$15,$70,$8F,$F7,$1B,$77,$FE,$6E,$DC,$7F,$C0,$37,$F6,$AB,$FC,$09
	.dc.b	$39,$6D,$F4,$3C,$1E,$3F,$EB,$FE,$B5,$F0,$5F,$B1,$FA,$E7,$C1,$D3
	.dc.b	$51,$30,$FF,$7D,$D7,$06,$C7,$1F,$90,$D4,$7F,$BE,$A7,$4D,$9D,$3F
	.dc.b	$EC,$0F,$8A,$FE,$04,$47,$E8,$8F,$AE,$FE,$3E,$DF,$2F,$F8,$34,$60
	.dc.b	$FE,$3B,$E1,$BE,$E7,$F8,$61,$A9,$95,$FE,$CE,$8F,$E2,$1E,$FC,$8F
	.dc.b	$F6,$8B,$F8,$AF,$0F,$AF,$C6,$49,$F1,$E3,$FE,$82,$7F,$0B,$D3,$F8
	.dc.b	$CF,$4F,$F2,$24,$8C,$F4,$A3,$F8,$F7,$9F,$FC,$77,$EB,$7F,$7B,$5F
	.dc.b	$E3,$86,$BF,$E3,$B6,$A6,$9F,$B8,$27,$89,$FC,$A7,$E5,$7A,$32,$7E
	.dc.b	$4C,$FF,$00,$C7,$FC,$C5,$5F,$CE,$BF,$5B,$FC,$34,$17,$C4,$E4,$D7
	.dc.b	$4C,$C3,$DF,$3A,$6C,$E5,$FD,$09,$FE,$70,$87,$D3,$F7,$FB,$29,$3C
	.dc.b	$A3,$C4,$61,$E3,$87,$8E,$73,$1E,$0F,$88,$F4,$77,$48,$C1,$0E,$38
	.dc.b	$7E,$C1,$B1,$EF,$99,$1E,$A8,$E4,$F0,$8F,$EB,$7F,$BE,$7D,$78,$FF
	.dc.b	$24,$B5,$F6,$F8,$48,$C3,$FE,$35,$B7,$7F,$2E,$29,$7C,$9F,$A1,$1F
	.dc.b	$47,$E9,$3E,$97,$3C,$AF,$9B,$19,$7E,$7C,$65,$F6,$C7,$E2,$7C,$1F
	.dc.b	$E9,$7C,$7F,$C1,$67,$FC,$FC,$D7,$F1,$91,$A3,$83,$F8,$47,$FB,$77
	.dc.b	$D9,$7F,$B3,$E3,$04,$3E,$D9,$F9,$18,$DF,$C9,$7F,$2E,$FA,$AF,$E5
	.dc.b	$FF,$55,$2B,$F3,$F1,$97,$F4,$31,$97,$EE,$F5,$27,$FC,$13,$F5,$9F
	.dc.b	$D9,$FC,$8F,$F0,$33,$6B,$FD,$1A,$DE,$6F,$E2,$71,$C3,$DF,$8F,$F9
	.dc.b	$FA,$75,$E4,$7F,$A7,$F9,$3F,$CB,$7B,$FF,$87,$AB,$E2,$F7,$7C,$82
	.dc.b	$3F,$E7,$73,$F5,$CF,$ED,$48,$FF,$12,$E0,$F2,$3A,$3F,$5B,$1A,$7E
	.dc.b	$3C,$6B,$EC,$3F,$DE,$91,$FC,$69,$AF,$CC,$6D,$FF,$9F,$7E,$C5,$F5
	.dc.b	$9F,$E2,$19,$CE,$DF,$EB,$C6,$DF,$FA,$FD,$BB,$2F,$8C,$FC,$3F,$A9
	.dc.b	$F9,$A4,$7F,$8B,$EB,$F7,$DF,$CD,$EC,$9D,$46,$1F,$F1,$FC,$EF,$36
	.dc.b	$5F,$E3,$BE,$A7,$C5,$7E,$A0,$98,$7F,$99,$3E,$2A,$BF,$A8,$9D,$BF
	.dc.b	$AC,$9F,$F2,$14,$C8,$E0,$43,$F8,$1F,$BF,$EC,$6C,$6D,$A7,$F9,$CB
	.dc.b	$7F,$DC,$F3,$48,$FD,$CB,$FE,$4F,$E7,$C4,$78,$E2,$E7,$FE,$AD,$C3
	.dc.b	$DD,$9F,$FA,$33,$F7,$AF,$ED,$5F,$0A,$69,$FC,$A2,$7E,$9C,$67,$C5
	.dc.b	$FF,$3C,$CE,$9F,$F2,$F7,$CD,$61,$99,$7F,$83,$2B,$83,$FE,$8A,$9C
	.dc.b	$BF,$F0,$06,$B9,$39,$38,$3F,$CB,$1B,$E2,$F8,$0F,$F3,$67,$BB,$63
	.dc.b	$24,$3E,$5B,$FE,$6D,$F8,$AF,$F3,$66,$89,$FE,$83,$BD,$52,$F0,$31
	.dc.b	$E4,$7D,$38,$C4,$F1,$8C,$78,$4C,$31,$87,$3B,$EF,$6F,$AD,$E5,$8F
	.dc.b	$76,$C7,$8D,$E7,$76,$F9,$61,$D6,$B9,$47,$5B,$E4,$D8,$E0,$7B,$70
	.dc.b	$F0,$E1,$D2,$3D,$E6,$6B,$DB,$5C,$63,$27,$C2,$BF,$06,$3C,$6F,$CF
	.dc.b	$3E,$23,$91,$E0,$38,$63,$71,$F0,$E1,$D6,$B8,$9C,$DF,$6C,$EF,$1F
	.dc.b	$5D,$F4,$3B,$1D,$0B,$7D,$1D,$5F,$4B,$E7,$1A,$C7,$E4,$4E,$63,$73
	.dc.b	$A8,$F1,$46,$08,$77,$72,$FE,$12,$FC,$AD,$3D,$D5,$DA,$BA,$1E,$F2
	.dc.b	$47,$D3,$8D,$37,$0F,$C9,$6F,$FE,$0D,$8F,$73,$71,$FC,$4B,$6B,$F0
	.dc.b	$4E,$E7,$F3,$5F,$5D,$CB,$3F,$CD,$46,$04,$ED,$4E,$DD,$B5,$FD,$13
	.dc.b	$F6,$D7,$C6,$FF,$22,$7E,$99,$D7,$4F,$E2,$CE,$B8,$F8,$3D,$8D,$CF
	.dc.b	$19,$DC,$65,$E3,$19,$D3,$D3,$39,$7D,$71,$98,$F3,$9D,$2B,$BD,$F9
	.dc.b	$E7,$53,$EA,$34,$7A,$0E,$87,$EF,$DD,$BE,$BA,$EF,$5D,$23,$8D,$FA
	.dc.b	$2F,$B4,$E6,$3D,$2D,$AE,$35,$DE,$B6,$D3,$9C,$7B,$A3,$F8,$3C,$78
	.dc.b	$E6,$BE,$9C,$7A,$BF,$B4,$3C,$C6,$CE,$97,$A3,$CC,$F6,$3E,$3B,$9B
	.dc.b	$F0,$47,$3C,$3E,$E5,$70,$3B,$9F,$DB,$38,$BF,$A1,$79,$7F,$81,$DB
	.dc.b	$F0,$51,$8C,$BF,$D0,$7C,$8F,$A6,$D4,$7F,$4B,$BB,$92,$F8,$3F,$CF
	.dc.b	$E3,$3D,$EC,$39,$5F,$F4,$09,$E3,$7F,$D4,$FC,$CB,$CD,$FE,$D4,$AF
	.dc.b	$0F,$76,$26,$47,$F4,$87,$AE,$1F,$DF,$E3,$04,$CB,$FA,$84,$EA,$B8
	.dc.b	$C6,$27,$5A,$7F,$67,$9C,$BF,$0F,$B5,$F4,$9E,$E8,$FF,$33,$3E,$0F
	.dc.b	$AA,$C4,$EA,$FF,$CD,$B3,$83,$4E,$0F,$97,$F5,$9E,$83,$D3,$F6,$16
	.dc.b	$32,$4C,$BF,$C0,$F1,$97,$FC,$EF,$ED,$3E,$53,$C8,$DA,$E1,$9F,$DB
	.dc.b	$8C,$57,$0A,$D9,$CE,$74,$FB,$7E,$B9,$F4,$54,$FB,$3E,$33,$6C,$65
	.dc.b	$FD,$B7,$0F,$C4,$5F,$08,$C1,$97,$18,$FF,$A8,$AF,$94,$EA,$3F,$B6
	.dc.b	$4E,$AB,$F8,$94,$7F,$8C,$BD,$5F,$C6,$B6,$BF,$DA,$BF,$5A,$F9,$5C
	.dc.b	$F9,$2B,$A4,$AD,$CD,$BF,$F4,$0F,$3F,$E4,$9B,$D5,$F9,$AE,$C7,$D4
	.dc.b	$BC,$8F,$0D,$EF,$5F,$2B,$72,$FF,$06,$37,$1F,$E5,$77,$C6,$7F,$DF
	.dc.b	$F3,$99,$FD,$36,$3E,$E7,$8B,$5F,$F0,$C4,$79,$5F,$E7,$ED,$7E,$66
	.dc.b	$7F,$5C,$FA,$BE,$F5,$BF,$99,$8B,$7F,$89,$18,$6D,$FD,$3A,$E0,$FF
	.dc.b	$B5,$BD,$F7,$C1,$7E,$E1,$7C,$45,$8C,$37,$0F,$B5,$E0,$AB,$71,$87
	.dc.b	$9B,$67,$E5,$FD,$54,$8F,$C1,$70,$FF,$AA,$79,$71,$97,$D6,$AB,$AF
	.dc.b	$F5,$5A,$D5,$FC,$C9,$DD,$7C,$67,$E5,$3F,$C4,$DB,$B9,$F9,$D9,$BD
	.dc.b	$C7,$B9,$FC,$59,$7C,$35,$8C,$37,$1F,$7C,$E9,$B3,$A7,$E8,$F6,$7F
	.dc.b	$A8,$1B,$E3,$F5,$F9,$CF,$DB,$B2,$32,$74,$8C,$69,$FF,$00,$3E,$73
	.dc.b	$C9,$D7,$E5,$6D,$78,$9B,$3B,$8F,$2C,$ED,$A6,$CB,$1F,$E5,$B7,$FC
	.dc.b	$CD,$A7,$C3,$FB,$3F,$85,$F6,$3F,$0D,$B3,$F0,$1B,$5C,$45,$71,$9F
	.dc.b	$8B,$5F,$2F,$DE,$F9,$9D,$B6,$7E,$4B,$6F,$95,$FD,$38,$C7,$27,$F8
	.dc.b	$4E,$31,$5F,$AA,$E9,$FD,$99,$F5,$FE,$6E,$C6,$49,$87,$4E,$24,$EA
	.dc.b	$3F,$80,$D7,$E1,$09,$5F,$EE,$7E,$54,$7E,$1B,$A5,$F5,$5F,$A3,$3F
	.dc.b	$0D,$E3,$DB,$FE,$35,$83,$E3,$B7,$18,$6C,$7E,$B6,$BF,$8D,$C6,$35
	.dc.b	$F7,$68,$E6,$FC,$91,$FB,$1F,$86,$B8,$FE,$BB,$F0,$9F,$F4,$E3,$D6
	.dc.b	$3F,$B6,$C6,$32,$FA,$87,$8E,$3B,$92,$79,$45,$3E,$8C,$FF,$8F,$5B
	.dc.b	$D9,$9C,$7E,$94,$E5,$FE,$6B,$5F,$A8,$3E,$09,$6B,$F0,$8B,$EA,$9E
	.dc.b	$E7,$F3,$A8,$D5,$72,$EC,$C1,$FE,$0A,$8E,$13,$87,$AC,$67,$DC,$F8
	.dc.b	$D1,$83,$FA,$50,$73,$87,$C2,$FC,$8F,$F8,$04,$FD,$66,$C7,$9B,$1E
	.dc.b	$13,$9F,$03,$FA,$7D,$6B,$2F,$EA,$6D,$9F,$A3,$5C,$2F,$B5,$F0,$3D
	.dc.b	$50,$F7,$31,$5E,$68,$C3,$F8,$8F,$5B,$FA,$D3,$7C,$AF,$E6,$1F,$57
	.dc.b	$FD,$82,$73,$8F,$DC,$6E,$7F,$D9,$9B,$7F,$BA,$E7,$5F,$1C,$24,$FF
	.dc.b	$33,$F8,$8B,$8F,$F3,$F7,$B7,$D7,$49,$FC,$77,$A3,$6B,$C1,$37,$3F
	.dc.b	$C1,$CD,$AF,$F7,$4C,$FF,$90,$BF,$24,$7D,$33,$E5,$F5,$8A,$73,$1F
	.dc.b	$DF,$3E,$AB,$F5,$F8,$7E,$A9,$B5,$FC,$37,$C0,$7C,$AB,$E1,$78,$78
	.dc.b	$53,$0E,$0F,$F9,$2A,$75,$18,$25,$FF,$A4,$30,$EA,$FF,$AA,$9A,$88
	.dc.b	$E1,$C7,$E1,$AF,$3F,$FC,$26,$DA,$FF,$0D,$65,$E6,$F7,$73,$52,$7F
	.dc.b	$82,$D6,$EF,$D9,$8F,$F9,$75,$FD,$BC,$67,$87,$F5,$FF,$1D,$FE,$20
	.dc.b	$FA,$07,$C4,$7E,$FA,$30,$DC,$7F,$88,$CE,$9F,$E0,$D3,$A7,$FC,$07
	.dc.b	$F8,$EF,$F0,$F7,$FC,$70,$DC,$7F,$9C,$FC,$17,$E7,$56,$29,$D3,$7C
	.dc.b	$9F,$F3,$14,$71,$F3,$3F,$CF,$9E,$06,$27,$FC,$F7,$F5,$2F,$8B,$A7
	.dc.b	$FA,$0B,$A5,$E9,$E3,$5F,$93,$5B,$6D,$6E,$75,$A7,$E0,$EB,$4D,$9D
	.dc.b	$57,$03,$E9,$FE,$F7,$AD,$7F,$D3,$6D,$8F,$EB,$FB,$3E,$D7,$FD,$23
	.dc.b	$39,$C3,$FD,$34,$DE,$73,$9D,$FF,$D3,$55,$C3,$2F,$F2,$AF,$D5,$BE
	.dc.b	$7B,$FD,$D7,$8F,$FB,$7A,$31,$E6,$FF,$B8,$21,$18,$75,$3C,$1F,$F7
	.dc.b	$EF,$C7,$FF,$85,$3E,$AF,$83,$63,$FD,$D2,$D7,$7E,$6A,$7B,$5F,$F9
	.dc.b	$97,$E3,$EF,$8E,$69,$2F,$FC,$9B,$EB,$7F,$9C,$38,$FF,$D8,$2D,$E3
	.dc.b	$FF,$29,$B6,$7F,$0F,$D5,$FA,$77,$F6,$D1,$9E,$3F,$ED,$F5,$9B,$E6
	.dc.b	$D3,$E8,$DD,$BF,$F6,$47,$D5,$FF,$70,$7A,$1F,$E6,$27,$0F,$F9,$89
	.dc.b	$AF,$F9,$C1,$AE,$FC,$3E,$98,$27,$8E,$1F,$ED,$C6,$E1,$FE,$20,$FA
	.dc.b	$B7,$CB,$F0,$9C,$E3,$E4,$C7,$07,$9F,$99,$7C,$27,$FC,$DF,$F5,$6F
	.dc.b	$AB,$F6,$66,$8C,$4E,$A7,$97,$53,$EA,$F7,$5A,$AF,$F1,$65,$F1,$BC
	.dc.b	$9C,$2B,$C5,$E8,$3E,$AF,$E5,$5B,$C4,$7C,$BE,$1B,$1F,$E7,$39,$E2
	.dc.b	$66,$64,$64,$FF,$0D,$FD,$53,$EA,$FF,$31,$08,$C3,$F9,$B6,$BF,$60
	.dc.b	$D4,$70,$8C,$E7,$FE,$BC,$6F,$1F,$E2,$5B,$5F,$CC,$7D,$77,$37,$B3
	.dc.b	$EA,$DE,$EF,$BB,$EA,$FF,$A1,$5F,$F8,$C1,$AF,$F8,$82,$76,$FF,$8D
	.dc.b	$1A,$F9,$5A,$FE,$39,$AF,$E6,$DA,$FF,$43,$78,$E5,$FE,$D4,$6E,$5B
	.dc.b	$BF,$27,$5F,$D2,$FA,$AF,$F5,$3E,$1F,$ED,$DF,$82,$7D,$5B,$D8,$FD
	.dc.b	$04,$61,$E1,$F1,$B2,$BF,$83,$5F,$27,$E6,$35,$FD,$2B,$5F,$D9,$3B
	.dc.b	$D3,$FD,$D2,$DC,$F3,$5E,$0C,$7F,$BA,$5F,$F9,$62,$BF,$8C,$3F,$EB
	.dc.b	$AF,$A3,$7D,$2F,$E7,$DB,$1B,$F2,$3D,$C7,$E1,$2B,$8C,$79,$A3,$04
	.dc.b	$78,$10,$FF,$3F,$CB,$D7,$FC,$C8,$F7,$7F,$37,$F5,$7F,$EB,$2A,$FF
	.dc.b	$36,$B5,$FF,$2E,$CE,$DF,$F3,$8B,$5F,$C9,$35,$FE,$16,$D7,$F8,$C3
	.dc.b	$5F,$F0,$03,$C7,$56,$BA,$BF,$29,$E4,$E0,$FD,$7F,$B1,$FE,$77,$C3
	.dc.b	$FD,$03,$F4,$DF,$E0,$56,$FB,$5A,$F8,$79,$1F,$D0,$6B,$83,$FA,$76
	.dc.b	$BF,$C8,$1A,$FF,$2D,$77,$97,$92,$3F,$C2,$D1,$8C,$B5,$5F,$C0,$B0
	.dc.b	$FF,$7C,$D7,$FB,$C7,$EA,$FF,$83,$91,$FF,$4B,$FD,$47,$F1,$E4,$FF
	.dc.b	$2A,$7F,$C4,$4D,$73,$E4,$7F,$02,$8D,$79,$1F,$EF,$FF,$AB,$FE,$23
	.dc.b	$1C,$CF,$E5,$31,$9A,$FF,$16,$69,$FE,$BB,$F1,$1F,$57,$FC,$68,$93
	.dc.b	$F9,$1F,$0F,$9F,$D5,$F7,$24,$7F,$69,$C5,$1F,$DA,$F1,$73,$7F,$BE
	.dc.b	$AB,$F3,$5E,$9F,$E6,$4F,$E4,$FE,$77,$F2,$6F,$AB,$FD,$CC,$BD,$67
	.dc.b	$C9,$7F,$5D,$6F,$2E,$FC,$39,$5F,$DE,$6F,$F8,$2A,$69,$23,$71,$FE
	.dc.b	$B2,$8E,$F9,$B8,$1F,$35,$FD,$2F,$E4,$3F,$BA,$5F,$C6,$C7,$FD,$E7
	.dc.b	$0F,$8B,$D1,$BE,$C9,$5B,$3F,$DE,$33,$B7,$18,$A3,$11,$FE,$E3,$9F
	.dc.b	$07,$C3,$F8,$7E,$96,$4F,$0A,$FE,$57,$CF,$FE,$D4,$AC,$CF,$FB,$AE
	.dc.b	$B8,$BB,$F1,$3D,$5E,$0C,$8C,$10,$FF,$79,$39,$7F,$DE,$91,$F0,$4C
	.dc.b	$5F,$FB,$8A,$71,$3E,$D8,$7D,$1F,$DD,$8F,$F6,$BF,$A5,$FD,$CF,$E4
	.dc.b	$BF,$B5,$46,$E7,$55,$FA,$73,$8F,$3F,$F7,$48,$CF,$07,$E5,$1B,$5F
	.dc.b	$2B,$C0,$FA,$9F,$11,$F5,$7B,$64,$FC,$58,$CB,$F1,$E3,$2F,$F9,$A2
	.dc.b	$FE,$CC,$FE,$3F,$C5,$FE,$EB,$F5,$7F,$EC,$3A,$FB,$01,$C3,$F8,$78
	.dc.b	$E0,$FE,$3D,$FE,$E3,$D1,$FC,$5E,$3F,$21,$18,$36,$F1,$3A,$CF,$FB
	.dc.b	$C1,$FE,$79,$F2,$DF,$CE,$FE,$5A,$46,$08,$7F,$34,$72,$FF,$36,$8F
	.dc.b	$F0,$DF,$D4,$FF,$74,$8F,$E8,$DE,$3F,$F4,$53,$3E,$9C,$8F,$B1,$3B
	.dc.b	$7F,$C3,$6D,$E1,$FE,$35,$D5,$FC,$9E,$7E,$1E,$9F,$E2,$B7,$E5,$1F
	.dc.b	$EB,$3C,$FF,$D1,$4F,$B3,$E1,$3F,$75,$E1,$48,$FD,$A3,$F1,$88,$E0
	.dc.b	$EA,$ED,$EA,$7D,$1C,$BF,$C5,$3F,$56,$F9,$6F,$F1,$F7,$D5,$F8,$2B
	.dc.b	$C7,$9A,$31,$A7,$F3,$39,$5F,$17,$EE,$B5,$EE,$D8,$E0,$DE,$3F,$F6
	.dc.b	$85,$F5,$F9,$DF,$9F,$F5,$1F,$E4,$FF,$41,$14,$CE,$3F,$3A,$6B,$F3
	.dc.b	$9E,$1F,$35,$F9,$FF,$8E,$FF,$48,$67,$FA,$A6,$F5,$FF,$AC,$AF,$FC
	.dc.b	$71,$39,$6C,$65,$FF,$29,$B6,$3F,$58,$DC,$FF,$D5,$7C,$3F,$E5,$C2
	.dc.b	$57,$F1,$2F,$83,$F8,$F3,$FE,$B0,$CF,$33,$F3,$27,$2F,$FB,$4A,$BF
	.dc.b	$CC,$B3,$FB,$7E,$8F,$39,$27,$CA,$D7,$CA,$ED,$F0,$67,$FE,$82,$FA
	.dc.b	$9F,$E5,$B1,$FE,$8C,$F8,$74,$62,$3F,$D3,$51,$8F,$1F,$FD,$6C,$D8
	.dc.b	$FF,$2E,$35,$FC,$AB,$75,$FE,$E8,$DD,$3F,$D9,$BC,$9E,$1B,$97,$FC
	.dc.b	$DE,$B4,$FF,$BC,$27,$2F,$4C,$3C,$F8,$74,$6E,$1D,$27,$94,$7F,$98
	.dc.b	$1F,$4B,$50,$BF,$7F,$8D,$EC,$9D,$4E,$5F,$E7,$F5,$C2,$B7,$3D,$AB
	.dc.b	$FD,$2F,$EF,$FF,$A3,$BE,$33,$FD,$1B,$8F,$6F,$AA,$FC,$AD,$F8,$25
	.dc.b	$E0,$FF,$B1,$67,$32,$9C,$BC,$78,$3F,$D9,$51,$D2,$13,$C2,$31,$E5
	.dc.b	$DB,$FF,$68,$36,$BD,$EE,$FC,$36,$38,$C7,$AC,$62,$BA,$C7,$82,$FC
	.dc.b	$CD,$8F,$21,$EB,$9E,$18,$F5,$BC,$CF,$AE,$F8,$BF,$05,$AF,$BE,$B8
	.dc.b	$DF,$8A,$BC,$93,$B2,$9C,$CB,$8F,$C0,$9E,$11,$E8,$38,$14,$E0,$58
	.dc.b	$F9,$58,$79,$E3,$18,$79,$27,$E8,$C6,$F0,$F2,$BE,$FC,$3D,$B3,$AB
	.dc.b	$F9,$D3,$B9,$CD,$7C,$BC,$72,$4C,$7C,$EF,$86,$F8,$92,$3D,$6F,$96
	.dc.b	$38,$5F,$B1,$C4,$F2,$3E,$A4,$7F,$88,$DB,$1A,$8F,$40,$E0,$FF,$0F
	.dc.b	$E3,$F8,$B6,$E1,$F8,$16,$E3,$FA,$67,$FC,$4B,$5F,$A8,$FA,$BC,$61
	.dc.b	$FE,$DB,$6C,$7E,$B0,$D1,$DE,$34,$D3,$42,$3E,$EB,$FE,$4F,$DE,$B9
	.dc.b	$3D,$87,$5F,$43,$FC,$FD,$5F,$A8,$8E,$ED,$8C,$E7,$E0,$8C,$57,$4B
	.dc.b	$FC,$AD,$F7,$3F,$0C,$7A,$34,$DB,$83,$C5,$58,$3C,$0E,$2F,$D7,$87
	.dc.b	$82,$3D,$B7,$DA,$3F,$D2,$97,$C2,$31,$87,$8A,$BD,$D3,$E0,$34,$F9
	.dc.b	$2B,$8D,$FC,$A9,$D3,$4F,$0D,$73,$BF,$97,$5E,$3B,$FA,$91,$89,$F8
	.dc.b	$51,$D5,$E0,$D3,$E0,$47,$99,$A6,$6F,$CB,$87,$C3,$C3,$E7,$1E,$29
	.dc.b	$FA,$E7,$80,$E0,$F9,$DF,$B5,$7F,$C2,$B8,$3F,$A7,$F9,$9F,$E5,$38
	.dc.b	$FE,$5D,$C4,$8C,$53,$B1,$D0,$FA,$5C,$E3,$CB,$1F,$E8,$38,$7F,$84
	.dc.b	$70,$E6,$D7,$F9,$E3,$67,$84,$7F,$61,$C7,$FD,$E1,$5F,$D2,$BC,$CF
	.dc.b	$0E,$5E,$0E,$DE,$F5,$AB,$FE,$A9,$58,$37,$9A,$B1,$87,$32,$57,$F7
	.dc.b	$D8,$CB,$CA,$3F,$B5,$DF,$58,$FE,$B3,$7E,$78,$C4,$7F,$9D,$7E,$7F
	.dc.b	$E2,$F1,$FF,$A0,$9F,$0F,$D4,$63,$B9,$14,$F8,$53,$BC,$7F,$C3,$71
	.dc.b	$89,$DC,$7E,$DA,$FE,$86,$6F,$AB,$F6,$9E,$EF,$AB,$B5,$FE,$1B,$8B
	.dc.b	$5F,$E1,$99,$B8,$FF,$A3,$FC,$A7,$CF,$FA,$1B,$87,$13,$FD,$38,$E2
	.dc.b	$FF,$CB,$F4,$8C,$47,$F0,$F8,$C5,$7B,$27,$33,$FE,$EC,$6B,$FD,$83
	.dc.b	$EA,$1F,$57,$F2,$1B,$87,$FB,$13,$C6,$6F,$9A,$93,$9D,$2E,$1F,$E6
	.dc.b	$0B,$F8,$93,$FE,$98,$F1,$BF,$DB,$2D,$F5,$BF,$13,$F5,$6F,$95,$FC
	.dc.b	$DE,$0B,$DB,$FE,$70,$6E,$FF,$E8,$2A,$7C,$4F,$09,$CE,$3F,$E6,$49
	.dc.b	$CB,$FE,$A8,$E9,$FD,$23,$7A,$FF,$53,$E0,$3E,$AD,$E5,$78,$AF,$7A
	.dc.b	$7F,$A0,$1A,$FF,$90,$EF,$F5,$0F,$F8,$9D,$B1,$A8,$FF,$22,$C7,$E4
	.dc.b	$BE,$D5,$F3,$FD,$63,$0F,$CF,$E0,$99,$FC,$FF,$85,$B8,$7D,$EE,$6F
	.dc.b	$6B,$FB,$FE,$37,$C1,$6B,$8F,$FD,$4B,$18,$C3,$FD,$F7,$EF,$BE,$B3
	.dc.b	$F6,$EB,$E3,$6C,$61,$B3,$C1,$B8,$FF,$A6,$AB,$84,$FC,$BD,$7E,$31
	.dc.b	$F2,$BD,$A6,$38,$9D,$1F,$E6,$49,$F8,$A4,$70,$7F,$94,$32,$3F,$E5
	.dc.b	$2E,$4F,$A9,$E0,$F3,$CB,$FD,$7B,$18,$9E,$07,$FA,$8B,$E0,$5E,$F7
	.dc.b	$F1,$45,$F5,$96,$30,$D9,$E3,$7F,$CA,$6B,$71,$8F,$21,$1D,$23,$0E
	.dc.b	$A7,$F6,$3E,$0F,$FA,$BA,$7F,$6B,$E5,$78,$7E,$B3,$F7,$4C,$8E,$9F
	.dc.b	$02,$F6,$31,$87,$FB,$1A,$31,$3D,$BE,$B3,$D5,$E0,$FA,$DB,$19,$3C
	.dc.b	$71,$96,$99,$AF,$F6,$8E,$7F,$AB,$8C,$70,$FF,$BA,$A3,$13,$AF,$AC
	.dc.b	$F9,$1F,$0F,$F2,$5B,$7F,$01,$B3,$B8,$F8,$73,$F3,$F9,$BE,$87,$83
	.dc.b	$67,$E3,$B6,$BC,$6D,$C7,$C4,$FE,$BF,$B6,$57,$CE,$F9,$DB,$7F,$B1
	.dc.b	$9F,$D6,$92,$7F,$D5,$ED,$8F,$C2,$FE,$C1,$FE,$9A,$9F,$E0,$13,$FB
	.dc.b	$E8,$C4,$FE,$6F,$E9,$7F,$46,$7E,$1F,$A6,$DF,$E1,$25,$F2,$AA,$FD
	.dc.b	$FB,$67,$FA,$14,$EA,$72,$7F,$6B,$E7,$FC,$4F,$3F,$E4,$CF,$D8,$FC
	.dc.b	$5D,$BF,$C1,$DF,$F2,$0C,$13,$FD,$B3,$97,$CF,$BF,$E9,$2B,$FD,$6B
	.dc.b	$5A,$BF,$33,$63,$FA,$A3,$5F,$F2,$5B,$5F,$F2,$93,$5F,$F0,$45,$7F
	.dc.b	$8F,$EB,$B6,$28,$C5,$F0,$C3,$C9,$D7,$FC,$DA,$3F,$49,$87,$FA,$A1
	.dc.b	$FF,$2F,$18,$C3,$D0,$BE,$B1,$F0,$3C,$7B,$64,$FC,$A3,$9D,$FF,$98
	.dc.b	$9A,$F4,$8F,$E5,$F1,$D4,$FB,$DF,$03,$FD,$85,$FB,$06,$67,$F8,$5C
	.dc.b	$65,$F3,$1F,$ED,$EF,$AD,$FE,$67,$3F,$7D,$B1,$A8,$F4,$4F,$F5,$49
	.dc.b	$CE,$3F,$4E,$91,$C1,$FF,$3C,$B5,$FF,$40,$35,$F0,$78,$1F,$D2,$23
	.dc.b	$9D,$FE,$5F,$27,$C0,$F7,$AF,$F1,$EC,$FC,$98,$FE,$CB,$ED,$FF,$67
	.dc.b	$8F,$CB,$37,$0F,$E6,$DF,$5B,$AF,$46,$57,$E5,$AB,$55,$FE,$8E,$6B
	.dc.b	$FE,$E4,$9F,$EB,$71,$CF,$3F,$C1,$F8,$CF,$8B,$E7,$7F,$D0,$DF,$1F
	.dc.b	$A5,$31,$FE,$3B,$EB,$5A,$F9,$E3,$1F,$EE,$6F,$F8,$A5,$AF,$8E,$BC
	.dc.b	$24,$AF,$E6,$BF,$5A,$FA,$97,$D6,$F5,$D3,$F4,$93,$98,$FF,$04,$FD
	.dc.b	$6B,$C0,$A7,$F9,$8B,$E6,$1F,$13,$8C,$BF,$7D,$5B,$7F,$76,$DD,$3F
	.dc.b	$C7,$CD,$7F,$D6,$AD,$7F,$D7,$8E,$F2,$FD,$B0,$FB,$07,$CB,$F7,$AF
	.dc.b	$F1,$BC,$62,$FF,$87,$7D,$77,$F6,$86,$FA,$5E,$7F,$84,$FF,$11,$CE
	.dc.b	$6F,$E5,$B5,$FF,$63,$B5,$FF,$67,$35,$07,$1E,$17,$EF,$7E,$3F,$FC
	.dc.b	$5F,$F4,$8F,$8E,$FE,$21,$18,$6E,$5F,$C8,$E7,$4F,$F5,$A9,$D3,$FD
	.dc.b	$AB,$57,$D7,$7F,$25,$CB,$FC,$AA,$D8,$FE,$1B,$F0,$DF,$A0,$94,$70
	.dc.b	$F2,$BF,$CE,$F5,$BB,$FF,$02,$36,$BF,$CD,$FE,$06,$AB,$FD,$3B,$F5
	.dc.b	$DF,$DA,$27,$F7,$BF,$02,$8C,$63,$FC,$73,$AB,$FA,$FB,$FC,$AE,$75
	.dc.b	$EB,$F2,$26,$7F,$E0,$1A,$DB,$F2,$A7,$6F,$F8,$21,$B8,$7F,$BE,$1A
	.dc.b	$FF,$BF,$DA,$F1,$C3,$CC,$E7,$EB,$BC,$B5,$D5,$B1,$FE,$8C,$FA,$EF
	.dc.b	$F5,$CC,$62,$CD,$37,$C5,$FE,$C0,$9D,$E5,$FE,$6A,$FA,$D7,$D1,$3C
	.dc.b	$DF,$EA,$F1,$86,$C6,$72,$FF,$4E,$4E,$5F,$F4,$F7,$CA,$7F,$8B,$3E
	.dc.b	$BB,$FA,$FB,$63,$3E,$72,$72,$D8,$E7,$E4,$BE,$53,$C2,$D8,$CC,$60
	.dc.b	$FF,$56,$6E,$F5,$FF,$97,$72,$7E,$23,$EB,$BF,$D3,$9C,$7F,$9C,$E2
	.dc.b	$FF,$6F,$C6,$31,$61,$C6,$B0,$D8,$EB,$87,$F9,$D8,$93,$FD,$EA,$55
	.dc.b	$FC,$26,$BF,$C8,$5D,$2F,$AA,$ED,$8F,$FB,$B9,$EB,$5B,$AF,$C8,$B7
	.dc.b	$2F,$CF,$CF,$0B,$E5,$5F,$89,$AD,$63,$F8,$D7,$94,$7E,$25,$AE,$7C
	.dc.b	$6F,$34,$62,$3F,$53,$F5,$AF,$AD,$F8,$59,$23,$13,$FB,$56,$BF,$D8
	.dc.b	$B3,$64,$FA,$1E,$35,$AB,$FE,$36,$DC,$FF,$CE,$CD,$C3,$FD,$C3,$5F
	.dc.b	$13,$D0,$7D,$6F,$D3,$97,$84,$F9,$BC,$36,$3F,$D2,$B3,$C4,$CC,$C8
	.dc.b	$C9,$FE,$33,$FA,$DF,$92,$FF,$4C,$F5,$7F,$CD,$4F,$F8,$DD,$AF,$F5
	.dc.b	$08,$DE,$9F,$DD,$DB,$A7,$F2,$49,$FE,$DF,$DB,$E7,$BE,$37,$BE,$3F
	.dc.b	$57,$D6,$7F,$8D,$2F,$FC,$89,$F5,$BF,$D2,$9B,$E7,$EB,$DC,$91,$82
	.dc.b	$1F,$D8,$AB,$93,$63,$FC,$92,$A3,$2F,$94,$C4,$6A,$33,$1E,$9F,$AC
	.dc.b	$FF,$5D,$3F,$E9,$AF,$7D,$F5,$BF,$EE,$48,$C3,$5F,$EE,$73,$C1,$BF
	.dc.b	$1D,$FA,$9A,$E3,$EC,$5B,$E1,$AF,$F8,$7F,$0E,$0D,$BF,$BD,$18,$E0
	.dc.b	$F2,$47,$1F,$5F,$3E,$37,$EA,$3E,$A5,$F5,$34,$CE,$BF,$E6,$FC,$7F
	.dc.b	$8B,$BF,$EB,$19,$DB,$67,$6F,$AA,$AE,$5E,$57,$B7,$A9,$F9,$13,$FC
	.dc.b	$31,$3F,$E7,$BF,$AD,$FF,$61,$B7,$E4,$75,$EF,$C8,$C1,$0F,$F0,$CD
	.dc.b	$72,$6C,$6D,$A7,$5F,$1F,$F4,$1E,$E3,$E1,$BF,$45,$F5,$5F,$EE,$4A
	.dc.b	$FF,$46,$35,$FF,$12,$CE,$DF,$F4,$83,$5F,$F1,$4C,$6F,$4F,$F2,$83
	.dc.b	$78,$BF,$35,$5B,$9F,$F2,$A3,$63,$37,$E4,$3C,$7F,$27,$FB,$7F,$AD
	.dc.b	$98,$FF,$B7,$FC,$34,$7E,$E7,$EB,$3F,$8A,$CF,$F8,$29,$BE,$42,$B7
	.dc.b	$3C,$6B,$87,$97,$FC,$0F,$EB,$6C,$FF,$AA,$F0,$E3,$69,$FE,$11,$F0
	.dc.b	$32,$BC,$4F,$E1,$FC,$6F,$17,$CB,$D9,$FC,$24,$61,$FC,$37,$A5,$F8
	.dc.b	$D2,$3E,$BF,$AA,$FE,$25,$F5,$9F,$B0,$E3,$FF,$55,$7B,$0F,$93,$FE
	.dc.b	$9B,$C9,$FE,$26,$8C,$3C,$3C,$69,$8F,$F8,$AA,$F8,$DF,$08,$FE,$D8
	.dc.b	$D5,$87,$7B,$FE,$1C,$D8,$DC,$7F,$57,$C3,$BC,$7F,$B3,$E7,$5C,$6F
	.dc.b	$95,$FE,$CD,$3F,$EF,$AF,$89,$FE,$A3,$1F,$E6,$B8,$F2,$61,$FD,$8E
	.dc.b	$37,$97,$FB,$F7,$AB,$77,$2F,$63,$0E,$D3,$D3,$C9,$EB,$F4,$F2,$B7
	.dc.b	$5E,$39,$D4,$63,$15,$6E,$3F,$A6,$47,$F2,$56,$F8,$5E,$8E,$F9,$1C
	.dc.b	$63,$D6,$61,$E3,$18,$74,$77,$75,$9F,$FB,$5A,$BF,$82,$CF,$FA,$F2
	.dc.b	$1F,$4A,$E3,$FE,$EF,$F2,$5F,$D7,$6B,$D6,$D7,$84,$ED,$FF,$2D,$35
	.dc.b	$FE,$CD,$1B,$AF,$F7,$2B,$77,$FC,$25,$6E,$7C,$9F,$0F,$F5,$B9,$A6
	.dc.b	$1F,$1E,$3E,$C9,$87,$E3,$C6,$1F,$C1,$78,$5F,$DA,$7E,$B3,$FD,$87
	.dc.b	$1F,$60,$38,$7F,$D5,$BC,$9F,$C5,$70,$FF,$6A,$37,$2F,$EB,$D5,$B9
	.dc.b	$FF,$6C,$36,$3F,$2B,$7F,$A6,$3C,$1E,$8C,$C3,$C6,$7F,$33,$F9,$7F
	.dc.b	$E7,$5F,$2F,$23,$04,$3F,$C1,$9E,$52,$35,$18,$9F,$F0,$AF,$D6,$BE
	.dc.b	$2F,$FA,$27,$8F,$2B,$FD,$15,$39,$6C,$E5,$EB,$3C,$A3,$F7,$BD,$FF
	.dc.b	$1F,$D6,$95,$FA,$B8,$CB,$F2,$76,$FE,$FB,$B7,$C4,$AC,$CF,$CA,$AE
	.dc.b	$2F,$FB,$62,$B6,$FF,$AA,$B8,$13,$FE,$1D,$FA,$C7,$CB,$7F,$91,$3E
	.dc.b	$B7,$7D,$C7,$8E,$0A,$DB,$F6,$DA,$FF,$8C,$E3,$89,$C2,$33,$C9,$FE
	.dc.b	$CC,$DD,$F5,$7F,$E5,$2F,$91,$7C,$FD,$3E,$6C,$4C,$7F,$73,$3A,$6C
	.dc.b	$E5,$FD,$0F,$CF,$7F,$95,$7C,$D9,$1F,$A2,$8F,$F2,$5F,$61,$CC,$7F
	.dc.b	$92,$7E,$0F,$FA,$2E,$7F,$B0,$9F,$F2,$F7,$1C,$AF,$F3,$4B,$FE,$61
	.dc.b	$C7,$11,$C6,$2D,$7F,$B1,$27,$BF,$47,$F9,$9F,$CA,$97,$FB,$48,$CB
	.dc.b	$FE,$74,$8C,$BF,$BA,$8F,$F3,$FF,$D5,$FF,$22,$E7,$FF,$43,$66,$FB
	.dc.b	$6F,$F3,$91,$AF,$22,$61,$FE,$7E,$8C,$BF,$EB,$DB,$E4,$FF,$A3,$1B
	.dc.b	$F1,$06,$9A,$E6,$FF,$35,$F3,$F4,$AF,$F6,$9C,$EB,$0F,$F4,$14,$F9
	.dc.b	$BC,$9F,$F4,$5F,$C4,$7F,$4B,$C3,$FD,$23,$A6,$71,$F1,$56,$A2,$46
	.dc.b	$0F,$E0,$9F,$19,$FE,$6D,$E1,$FF,$7F,$FD,$6F,$DF,$94,$FF,$BF,$5C
	.dc.b	$F1,$7F,$98,$63,$10,$9E,$11,$CE,$31,$3B,$F1,$7F,$DE,$F1,$80,$76
	.dc.b	$F1,$BF,$CD,$97,$C9,$B5,$C7,$0F,$0D,$70,$C3,$B5,$EE,$32,$73,$2C
	.dc.b	$6A,$B9,$47,$84,$D1,$63,$D7,$3E,$AC,$3C,$ED,$BE,$78,$79,$2F,$C4
	.dc.b	$D7,$E3,$D7,$0C,$3B,$F9,$BE,$59,$34,$E8,$E2,$3C,$B8,$6C,$8E,$09
	.dc.b	$1E,$03,$C6,$7B,$AB,$C0,$FC,$56,$BE,$E9,$E6,$E7,$0E,$07,$27,$E1
	.dc.b	$61,$EB,$C7,$E4,$46,$CC,$9B,$AF,$49,$30,$FC,$25,$71,$39,$61,$E2
	.dc.b	$BF,$C2,$6F,$FE,$11,$C3,$E3,$E3,$F8,$28,$C4,$EB,$2F,$AF,$1B,$AF
	.dc.b	$C3,$FD,$76,$E7,$FD,$3D,$08,$C3,$FE,$A1,$6B,$FE,$B7,$6B,$C8,$FC
	.dc.b	$10,$F3,$DA,$FE,$2A,$35,$AF,$E8,$DD,$46,$5F,$01,$CA,$78,$BD,$0C
	.dc.b	$CF,$18,$CB,$CE,$33,$AF,$FA,$DE,$72,$FB,$2F,$C5,$18,$9F,$61,$DE
	.dc.b	$FC,$53,$A6,$C6,$4F,$14,$65,$A6,$6F,$C5,$19,$6C,$7E,$4C,$FF,$6F
	.dc.b	$57,$B3,$0C,$9E,$BB,$FD,$CB,$C2,$B1,$5D,$9A,$67,$0F,$75,$79,$A3
	.dc.b	$31,$EF,$9F,$23,$63,$31,$F0,$1A,$FD,$B3,$66,$63,$AB,$E1,$6C,$7F
	.dc.b	$59,$E3,$FE,$61,$E7,$FD,$6D,$7E,$BF,$0B,$89,$EB,$E7,$3F,$E2,$B8
	.dc.b	$70,$7F,$CB,$73,$A7,$F9,$F6,$1B,$CF,$FC,$C9,$39,$7F,$9C,$C6,$70
	.dc.b	$ED,$86,$F0,$FE,$85,$E7,$DE,$63,$BE,$8F,$EE,$F9,$FF,$4B,$8D,$11
	.dc.b	$FF,$36,$C6,$B2,$FE,$8F,$39,$5A,$7B,$B4,$7B,$F9,$8C,$37,$0F,$E6
	.dc.b	$7A,$7C,$38,$C5,$FB,$1F,$75,$7F,$99,$BE,$83,$D9,$83,$DE,$CF,$80
	.dc.b	$9C,$C6,$1B,$97,$F9,$99,$B8,$FF,$80,$FE,$AF,$E0,$1F,$07,$EB,$AD
	.dc.b	$7B,$A2,$CF,$F8,$26,$6E,$5F,$6F,$CD,$7C,$F7,$93,$E4,$7C,$F1,$8C
	.dc.b	$7E,$CC,$62,$75,$1F,$E9,$CA,$FF,$78,$46,$30,$FF,$42,$7D,$3B,$EB
	.dc.b	$7E,$D7,$92,$D3,$E1,$CE,$71,$FF,$13,$C6,$2B,$6D,$8C,$9F,$E7,$3F
	.dc.b	$21,$F5,$3F,$85,$FA,$D7,$C5,$7E,$2F,$3F,$F1,$44,$E6,$31,$5C,$2F
	.dc.b	$FD,$5B,$7C,$1E,$95,$B7,$FD,$01,$D4,$FA,$BF,$B3,$F5,$8F,$9F,$ED
	.dc.b	$39,$AF,$85,$39,$78,$D7,$0A,$DB,$63,$27,$08,$CB,$4F,$99,$B7,$9F
	.dc.b	$4F,$C9,$C7,$E3,$B7,$79,$BF,$2B,$3F,$9A,$D9,$FF,$28,$B6,$B8,$9F
	.dc.b	$ED,$1A,$E3,$3F,$EC,$EF,$84,$FD,$87,$1F,$EF,$1F,$CF,$C6,$7C,$0F
	.dc.b	$D9,$46,$2F,$9F,$6A,$38,$D7,$E8,$9D,$5F,$3B,$E3,$DB,$F2,$35,$FE
	.dc.b	$01,$E2,$BE,$B7,$F4,$18,$FD,$09,$CC,$63,$3F,$48,$E2,$F9,$CF,$B2
	.dc.b	$F9,$76,$75,$C3,$FD,$DF,$18,$7A,$18,$BE,$99,$FF,$50,$BE,$98,$F8
	.dc.b	$30,$FE,$21,$F0,$1F,$ED,$BC,$3F,$DE,$1F,$15,$8F,$81,$CB,$63,$27
	.dc.b	$FB,$D6,$32,$FE,$97,$BB,$FD,$0E,$FF,$B8,$BD,$3F,$D6,$F9,$3F,$C7
	.dc.b	$7E,$05,$18,$9D,$1C,$67,$51,$E4,$FA,$CF,$2F,$6D,$AF,$F2,$1E,$FF
	.dc.b	$EF,$58,$C6,$FF,$5B,$5F,$F4,$55,$68,$FE,$2E,$AF,$9C,$66,$5C,$CA
	.dc.b	$D4,$E6,$31,$7D,$B0,$F0,$4E,$69,$F5,$9F,$2B,$E1,$7E,$5B,$67,$4D
	.dc.b	$C7,$CC,$DB,$FE,$9B,$6F,$81,$FA,$7E,$C7,$D4,$6E,$1D,$DB,$87,$BD
	.dc.b	$B7,$F7,$79,$DE,$CF,$E4,$3E,$B3,$FC,$D3,$87,$EF,$A3,$18,$7E,$F1
	.dc.b	$B1,$93,$F1,$73,$C2,$7F,$27,$5D,$27,$31,$C5,$A7,$EB,$7E,$8F,$F5
	.dc.b	$27,$E3,$39,$EC,$7C,$F9,$7C,$9A,$BF,$5F,$F1,$3F,$C6,$E3,$15,$FB
	.dc.b	$78,$FC,$AF,$3C,$8C,$1F,$45,$FC,$65,$F1,$9B,$3D,$E9,$19,$8F,$B9
	.dc.b	$CF,$FE,$36,$AF,$D7,$D6,$A7,$3D,$9F,$97,$23,$CE,$ED,$7E,$DA,$3F
	.dc.b	$50,$DE,$9F,$EF,$D0,$FA,$A7,$D6,$7F,$37,$8D,$F0,$FE,$CD,$A7,$D6
	.dc.b	$F7,$9A,$EF,$0E,$71,$89,$F9,$47,$0C,$7F,$CA,$71,$97,$C5,$1F,$9D
	.dc.b	$11,$9B,$FE,$05,$F5,$9F,$C3,$1B,$DF,$E8,$CF,$77,$F5,$EB,$4F,$FA
	.dc.b	$CB,$04,$61,$FF,$31,$CB,$E2,$BF,$AA,$4E,$6B,$E1,$46,$2F,$B3,$63
	.dc.b	$CD,$5C,$5F,$C3,$7B,$4F,$47,$FA,$F3,$63,$F9,$D7,$D6,$62,$64,$7F
	.dc.b	$3D,$F8,$CB,$9F,$F3,$8F,$21,$18,$D7,$FA,$03,$6B,$F9,$D3,$DB,$D0
	.dc.b	$91,$F8,$6F,$4E,$E1,$FC,$D3,$3F,$ED,$4F,$F4,$EF,$AC,$7D,$43,$E6
	.dc.b	$37,$9F,$F9,$FA,$73,$18,$C7,$FD,$B7,$18,$7D,$F9,$7F,$91,$63,$18
	.dc.b	$7F,$42,$FA,$CF,$D8,$3F,$E9,$A6,$BF,$4C,$38,$7E,$B3,$5F,$F7,$4B
	.dc.b	$CB,$0F,$F2,$E7,$D6,$3E,$4F,$FC,$15,$39,$C7,$F9,$6C,$E5,$FD,$C4
	.dc.b	$66,$FF,$C8,$F3,$97,$CD,$E1,$7F,$90,$23,$18,$7F,$4F,$FA,$C7,$A5
	.dc.b	$FD,$8B,$E0,$1F,$05,$FE,$19,$9C,$E7,$FD,$6A,$72,$FF,$07,$8C,$E5
	.dc.b	$FD,$86,$72,$FF,$66,$8C,$E1,$FE,$AC,$F0,$15,$A8,$FF,$3C,$F8,$4F
	.dc.b	$92,$FF,$14,$7D,$13,$E2,$BF,$AA,$46,$1B,$9F,$F1,$79,$D3,$FC,$22
	.dc.b	$74,$FF,$90,$FE,$B3,$F8,$BE,$9F,$E1,$0C,$51,$87,$FD,$13,$97,$E9
	.dc.b	$DC,$F8,$FF,$E6,$69,$D5,$FF,$5C,$6D,$FF,$54,$BE,$11,$8F,$25,$F5
	.dc.b	$9F,$DA,$AF,$FD,$BB,$18,$F5,$3F,$BF,$61,$FE,$94,$AD,$3F,$87,$9D
	.dc.b	$57,$09,$CE,$5F,$E8,$1E,$8F,$F4,$84,$62,$7F,$CF,$FF,$0B,$FA,$0A
	.dc.b	$FF,$6C,$DF,$2C,$3F,$C4,$B3,$9A,$F0,$63,$FE,$80,$C7,$FD,$35,$F5
	.dc.b	$9E,$2B,$FE,$15,$82,$31,$1F,$EA,$A6,$BF,$E2,$16,$C6,$EB,$FD,$17
	.dc.b	$D7,$D1,$B8,$7F,$4D,$FA,$BF,$F4,$77,$27,$FC,$00,$C5,$18,$7F,$50
	.dc.b	$63,$E1,$BF,$DF,$8E,$23,$4F,$EB,$7E,$35,$E5,$BE,$AF,$A4,$61,$B1
	.dc.b	$99,$EB,$39,$7A,$78,$CF,$8E,$F2,$61,$FC,$75,$B1,$99,$DC,$7D,$CF
	.dc.b	$BE,$FF,$3A,$47,$F8,$5F,$74,$62,$3F,$CB,$F1,$8F,$85,$FE,$6D,$C9
	.dc.b	$FE,$BB,$FA,$BB,$C1,$FE,$D6,$C3,$84,$FD,$A6,$C6,$6B,$78,$7F,$B5
	.dc.b	$AE,$E1,$E0,$FA,$FB,$81,$A7,$0F,$7B,$FD,$CD,$F6,$F8,$C3,$D1,$5B
	.dc.b	$6D,$6F,$A1,$5B,$9F,$5D,$6D,$B5,$B9,$F2,$E8,$FC,$6F,$FC,$51,$5F
	.dc.b	$EB,$56,$BF,$B3,$6C,$7F,$B7,$7B,$78,$C6,$36,$FF,$25,$B6,$BF,$DC
	.dc.b	$6D,$CF,$FE,$00,$57,$B7,$C1,$7C,$15,$F5,$E6,$1F,$E0,$C8,$D1,$87
	.dc.b	$FD,$13,$3F,$E0,$27,$47,$F8,$8F,$F3,$27,$D8,$7D,$08,$46,$1D,$47
	.dc.b	$D4,$E9,$7E,$57,$99,$FF,$43,$D5,$DB,$FC,$FC,$DD,$BF,$BC,$36,$7F
	.dc.b	$35,$55,$EF,$EE,$F8,$67,$FD,$E9,$1E,$58,$F0,$7C,$4F,$FB,$A2,$30
	.dc.b	$43,$C3,$1A,$F4,$F9,$32,$35,$8B,$14,$7E,$AE,$FA,$3F,$EA,$0F,$AB
	.dc.b	$FF,$1B,$BF,$B9,$6B,$FB,$86,$BF,$D8,$5A,$FF,$85,$5D,$D7,$FA,$7D
	.dc.b	$B5,$9F,$7B,$E7,$7D,$67,$EB,$EF,$C1,$1F,$46,$B7,$1E,$66,$A9,$E1
	.dc.b	$7F,$EA,$BF,$05,$F0,$7F,$67,$9F,$E6,$1F,$36,$33,$7F,$E6,$16,$BF
	.dc.b	$EB,$E6,$D7,$3A,$E5,$F0,$7F,$83,$9C,$E7,$FE,$31,$6D,$FC,$A6,$BF
	.dc.b	$AA,$9F,$C5,$7E,$4C,$9F,$CC,$7C,$0C,$AF,$F3,$94,$E5,$B3,$C0,$8F
	.dc.b	$FA,$87,$ED,$3F,$BB,$BA,$21,$FE,$D6,$9E,$4F,$FB,$9A,$36,$E7,$C3
	.dc.b	$FE,$09,$0F,$A6,$A7,$9A,$24,$F9,$1A,$FF,$1C,$A7,$D1,$E3,$5F,$F6
	.dc.b	$D4,$7F,$91,$67,$81,$FC,$95,$DB,$FE,$7E,$F8,$2F,$DC,$11,$FF,$49
	.dc.b	$7D,$73,$E8,$3F,$60,$D8,$CC,$60,$FE,$01,$E3,$C8,$C9,$0F,$DE,$BF
	.dc.b	$C1,$7E,$86,$47,$FA,$39,$FE,$57,$F0,$FF,$D0,$7E,$C7,$F7,$DF,$3B
	.dc.b	$FE,$F3,$D3,$FD,$E8,$DE,$4F,$F3,$73,$70,$FF,$37,$F9,$8F,$96,$E6
	.dc.b	$D8,$FF,$6D,$63,$FD,$36,$72,$78,$3F,$02,$FA,$78,$FF,$C0,$1F,$5B
	.dc.b	$E9,$5F,$FA,$CD,$AA,$35,$C9,$FC,$4F,$EA,$1F,$19,$FC,$56,$F7,$5A
	.dc.b	$BE,$57,$BA,$D3,$23,$27,$FB,$06,$F9,$BF,$EB,$2F,$37,$F9,$67,$CE
	.dc.b	$48,$C1,$0F,$E5,$2E,$5F,$E5,$71,$FC,$CF,$E7,$BF,$9A,$61,$27,$FD
	.dc.b	$7B,$19,$7F,$D7,$F1,$97,$F9,$9C,$7F,$3B,$FB,$5F,$E7,$9E,$94,$9F
	.dc.b	$F6,$2C,$65,$FF,$63,$C6,$5F,$E7,$71,$97,$F9,$EC,$67,$3F,$F2,$7C
	.dc.b	$FF,$56,$DB,$FA,$ED,$7F,$20,$9F,$E4,$4F,$CF,$8C,$BF,$E4,$96,$C7
	.dc.b	$F9,$63,$EC,$B9,$78,$52,$B9,$72,$73,$BD,$BF,$CB,$3D,$17,$9C,$C3
	.dc.b	$FA,$C7,$C3,$FF,$59,$F9,$C9,$8F,$FB,$8A,$32,$FF,$B9,$23,$2F,$F5
	.dc.b	$88,$FE,$A9,$F9,$27,$F8,$1A,$3F,$9C,$B5,$37,$8D,$D7,$C3,$B2,$57
	.dc.b	$FC,$01,$AD,$4E,$23,$F2,$1A,$7F,$B3,$67,$A1,$A8,$FF,$7B,$F6,$FF
	.dc.b	$83,$79,$7F,$49,$B7,$F8,$46,$FF,$BA,$4C,$AF,$8A,$27,$4E,$1A,$FF
	.dc.b	$74,$8F,$EE,$1E,$0F,$FA,$0E,$38,$9F,$C4,$3E,$17,$27,$F8,$77,$F5
	.dc.b	$CF,$EC,$B1,$E8,$3F,$AD,$CF,$17,$07,$13,$9B,$9E,$4F,$C0,$E0,$AE
	.dc.b	$71,$F8,$4A,$7D,$63,$E7,$FF,$C0,$8D,$8F,$2B,$5F,$17,$73,$F5,$4F
	.dc.b	$D6,$C7,$8A,$75,$7D,$5B,$AF,$FB,$ED,$B2,$BE,$5D,$9E,$0C,$64,$F8
	.dc.b	$1F,$F2,$2C,$EA,$3E,$7E,$1F,$C5,$7E,$4F,$F4,$04,$8F,$A1,$C0,$F4
	.dc.b	$71,$1F,$9F,$9B,$5F,$12,$B5,$3D,$3A,$9F,$63,$FA,$B2,$47,$4C,$AB
	.dc.b	$FA,$98,$B7,$CE,$7F,$65,$F6,$12,$BA,$3F,$E5,$1F,$4F,$11,$FB,$18
	.dc.b	$B5,$F5,$BC,$0F,$F0,$1F,$19,$A4,$BF,$F2,$D7,$C7,$F3,$C3,$F1,$D3
	.dc.b	$96,$C7,$F9,$63,$F7,$EF,$F0,$4F,$1F,$F0,$36,$CF,$F3,$16,$E1,$DF
	.dc.b	$E9,$F8,$79,$32,$BE,$7D,$6B,$15,$7F,$9D,$BE,$BD,$F1,$34,$97,$E4
	.dc.b	$F9,$AA,$FF,$A2,$27,$6F,$FA,$35,$AF,$FA,$F2,$37,$B7,$FB,$02,$72
	.dc.b	$DA,$FF,$3D,$4E,$B4,$FF,$1E,$FC,$2A,$F7,$18,$6E,$1F,$B5,$C9,$7E
	.dc.b	$9F,$06,$4F,$8D,$FF,$43,$73,$62,$3D,$11,$73,$FA,$D3,$FE,$69,$F8
	.dc.b	$87,$D6,$FF,$5B,$B7,$D6,$7F,$9E,$1F,$F2,$8C,$FE,$DE,$31,$08,$C7
	.dc.b	$C2,$7F,$95,$A3,$15,$CE,$B7,$5D,$B0,$FF,$2D,$C6,$3C,$2E,$0D,$BC
	.dc.b	$E1,$CE,$F8,$47,$92,$31,$87,$57,$57,$C9,$B3,$C8,$F4,$CE,$DB,$86
	.dc.b	$5F,$55,$71,$7D,$ED,$7D,$71,$BB,$F1,$E1,$E7,$9F,$21,$C0,$D5,$F7
	.dc.b	$78,$9A,$3B,$9C,$EF,$B5,$74,$C3,$DB,$86,$6A,$1C,$67,$4F,$A8,$C5
	.dc.b	$72,$39,$DF,$86,$77,$3E,$03,$67,$9A,$A3,$82,$61,$F6,$2B,$91,$CE
	.dc.b	$FB,$4F,$09,$D5,$79,$9F,$88,$62,$3D,$15,$F8,$DE,$37,$CD,$FB,$A7
	.dc.b	$FD,$FC,$D8,$CC,$7B,$5A,$FE,$37,$EB,$7F,$D4,$71,$88,$FF,$52,$C6
	.dc.b	$27,$F2,$51,$83,$A9,$88,$D3,$4F,$A8,$3E,$3B,$FC,$1F,$87,$E7,$A3
	.dc.b	$93,$63,$3A,$76,$95,$7E,$2D,$E5,$C5,$5C,$9D,$AB,$8B,$E1,$BE,$93
	.dc.b	$BA,$E5,$5A,$AF,$80,$71,$9D,$34,$F4,$9D,$AB,$9B,$8A,$E7,$3C,$27
	.dc.b	$27,$B6,$78,$5F,$C8,$BF,$04,$7C,$08,$C5,$71,$6C,$75,$3C,$55,$CE
	.dc.b	$FB,$C6,$63,$BC,$FC,$79,$CC,$6E,$78,$CE,$5A,$7C,$57,$BC,$6C,$E7
	.dc.b	$19,$3F,$88,$5F,$A1,$F5,$98,$BF,$E0,$5C,$7F,$CB,$7D,$1F,$F2,$5E
	.dc.b	$BF,$CD,$5B,$3F,$E4,$3E,$37,$AD,$8C,$18,$71,$AF,$E8,$17,$98,$C4
	.dc.b	$FF,$87,$7A,$7F,$B0,$5E,$79,$97,$7C,$E9,$8F,$F6,$0E,$FB,$8F,$F4
	.dc.b	$D9,$BA,$7F,$6B,$9B,$5F,$D0,$E6,$D7,$F4,$39,$B9,$FF,$49,$9C,$E7
	.dc.b	$FD,$C6,$77,$83,$AD,$F2,$FE,$D8,$C3,$EF,$AF,$EF,$1D,$F9,$AD,$E7
	.dc.b	$FD,$AA,$B0,$6F,$75,$62,$7F,$BD,$FC,$D6,$4D,$79,$56,$32,$4D,$BE
	.dc.b	$45,$FF,$B6,$23,$19,$FD,$6C,$9D,$3E,$33,$95,$FF,$92,$E7,$D9,$19
	.dc.b	$AF,$EF,$EE,$A3,$93,$EB,$F9,$09,$3F,$E2,$EF,$7E,$DF,$3C,$15,$8C
	.dc.b	$3F,$C5,$D5,$6F,$FC,$25,$EF,$FC,$53,$E4,$78,$EF,$7F,$13,$FE,$10
	.dc.b	$9B,$97,$E1,$6B,$53,$93,$FC,$D3,$DA,$F6,$BF,$21,$48,$C4,$FF,$9D
	.dc.b	$BC,$B7,$EA,$CC,$77,$5A,$BF,$F3,$5B,$74,$FF,$33,$37,$04,$FF,$96
	.dc.b	$6F,$94,$78,$B1,$E6,$47,$C1,$1A,$9E,$B3,$E6,$FD,$B1,$E4,$D3,$F1
	.dc.b	$5F,$AE,$FF,$3E,$C3,$FD,$0B,$3A,$6C,$ED,$FF,$47,$4E,$EF,$FD,$07
	.dc.b	$5B,$7E,$45,$FC,$EF,$94,$F9,$ED,$BF,$F3,$3B,$67,$71,$FE,$6A,$9F
	.dc.b	$8B,$CB,$23,$FD,$3E,$3E,$87,$F9,$BD,$7C,$89,$D3,$67,$73,$AD,$7E
	.dc.b	$5D,$69,$B3,$AF,$07,$0F,$FB,$37,$85,$F2,$9F,$86,$F7,$7F,$0E,$D9
	.dc.b	$FF,$3F,$B7,$2F,$C3,$37,$0F,$D1,$7C,$47,$EC,$CF,$C7,$FC,$95,$BF
	.dc.b	$F2,$EC,$5B,$FC,$74,$DA,$FF,$43,$7A,$3F,$96,$E2,$FC,$BB,$70,$FF
	.dc.b	$47,$B6,$FF,$2C,$DA,$E3,$3F,$E9,$4A,$E2,$D3,$FD,$2F,$E8,$FE,$8B
	.dc.b	$D0,$FD,$1B,$63,$27,$FA,$72,$32,$D3,$FD,$3F,$E7,$FE,$AB,$C9,$FD
	.dc.b	$5B,$63,$27,$FA,$92,$32,$D3,$FD,$A3,$F3,$EF,$BB,$FD,$4B,$63,$EC
	.dc.b	$FC,$E6,$26,$47,$07,$F9,$F3,$F8,$38,$CE,$BF,$ED,$4A,$E4,$DA,$DF
	.dc.b	$2A,$B7,$AB,$0F,$D6,$FC,$33,$F2,$E7,$C4,$7F,$4B,$8C,$36,$3F,$96
	.dc.b	$E5,$FE,$E5,$9C,$D7,$FB,$B3,$D2,$FC,$4B,$79,$78,$61,$FC,$1A,$73
	.dc.b	$7F,$C3,$7C,$CF,$E2,$67,$5F,$AA,$B8,$7F,$B3,$3C,$BE,$1D,$3F,$C6
	.dc.b	$1B,$1D,$E2,$DF,$F1,$79,$BA,$FF,$B3,$FC,$BF,$E3,$E6,$2F,$F3,$CF
	.dc.b	$F1,$8F,$35,$E4,$53,$FB,$36,$A9,$C5,$7F,$12,$BF,$F7,$D5,$F0,$8C
	.dc.b	$1F,$E8,$B7,$18,$A7,$E7,$3F,$DD,$2F,$A5,$70,$8F,$E3,$B5,$C2,$BA
	.dc.b	$E1,$FD,$77,$E0,$9D,$9E,$2B,$FE,$F1,$18,$C3,$94,$30,$FE,$6E,$F8
	.dc.b	$63,$C9,$E4,$F3,$EC,$F8,$0C,$8E,$35,$C3,$2F,$3A,$C3,$F8,$E4,$7C
	.dc.b	$0A,$E1,$87,$F4,$5F,$06,$CF,$FB,$F5,$FE,$87,$F8,$0B,$EB,$29,$FE
	.dc.b	$15,$FA,$87,$E0,$7F,$82,$BF,$8B,$82,$F8,$69,$EC,$6C,$7A,$CF,$E8
	.dc.b	$6E,$27,$E6,$DF,$4E,$7E,$F7,$C7,$1F,$F1,$07,$E0,$6F,$8E,$CE,$5C
	.dc.b	$FD,$64,$E6,$FE,$5F,$D3,$BE,$2F,$FA,$EB,$63,$EC,$53,$EE,$DF,$6F
	.dc.b	$AB,$95,$C9,$FF,$18,$4E,$A3,$04,$C7,$FC,$E9,$87,$56,$D7,$C4,$7F
	.dc.b	$DA,$D1,$97,$FD,$0F,$F0,$8F,$B1,$FE,$E2,$DB,$FC,$1F,$B1,$FD,$05
	.dc.b	$FF,$6F,$46,$6B,$E5,$07,$1C,$3F,$49,$B3,$FE,$E5,$AE,$55,$FD,$CA
	.dc.b	$72,$DB,$F0,$9F,$E0,$29,$DC,$75,$8F,$2F,$AD,$2F,$CB,$F4,$16,$BC
	.dc.b	$78,$7F,$77,$C7,$C5,$7C,$EB,$51,$F9,$4F,$41,$F0,$FF,$E4,$1F,$D8
	.dc.b	$BF,$C2,$19,$7E,$AF,$B1,$5A,$8C,$4E,$B3,$FF,$0E,$5F,$1F,$67,$FC
	.dc.b	$46,$DA,$FD,$A6,$2F,$DF,$DE,$9F,$F9,$45,$B3,$DF,$DE,$ED,$8F,$F8
	.dc.b	$F6,$72,$DA,$E0,$7F,$94,$70,$B7,$F8,$AF,$92,$FF,$37,$B7,$E0,$BF
	.dc.b	$CB,$9F,$BB,$F9,$F3,$FE,$13,$D2,$BD,$E5,$F9,$96,$CE,$9B,$87,$F0
	.dc.b	$CB,$E4,$FF,$8A,$AF,$96,$1F,$E8,$3A,$DE,$1F,$E8,$86,$F0,$30,$E3
	.dc.b	$39,$AE,$1F,$2B,$31,$FA,$15,$FE,$23,$CB,$C5,$39,$BF,$33,$EC,$F8
	.dc.b	$8C,$4F,$F0,$0F,$A6,$FE,$63,$EE,$7F,$B5,$6D,$18,$7F,$DB,$0D,$7F
	.dc.b	$DC,$ED,$7F,$C8,$A3,$E2,$FF,$CC,$7F,$47,$CD,$B1,$FE,$E1,$6B,$FE
	.dc.b	$F0,$6B,$CA,$7F,$CC,$9C,$9E,$36,$C7,$8B,$C8,$FF,$2F,$36,$7F,$DD
	.dc.b	$9C,$2F,$A6,$7C,$77,$93,$4F,$B3,$18,$BF,$E1,$2D,$7F,$D1,$7E,$36
	.dc.b	$F8,$7E,$B4,$E6,$B8,$DE,$6F,$FD,$7B,$39,$AF,$F6,$15,$E7,$0F,$17
	.dc.b	$57,$CB,$6D,$F9,$AF,$F4,$AD,$9D,$46,$23,$F9,$23,$71,$F4,$CF,$87
	.dc.b	$91,$D2,$F4,$BC,$17,$BC,$FD,$37,$83,$85,$6F,$1F,$D6,$FD,$33,$E7
	.dc.b	$FD,$99,$A3,$13,$A9,$E3,$E8,$6A,$73,$3F,$EA,$B6,$D7,$FB,$99,$B8
	.dc.b	$FF,$B9,$6B,$D1,$F3,$4F,$90,$BE,$0C,$C7,$F8,$44,$E5,$B3,$97,$FD
	.dc.b	$CD,$09,$E7,$3F,$4E,$5F,$48,$FA,$6F,$CD,$46,$1B,$19,$8F,$CF,$35
	.dc.b	$FC,$C9,$FE,$9E,$71,$9F,$F6,$46,$E5,$F9,$E6,$CB,$A5,$F4,$D8,$8F
	.dc.b	$1F,$D2,$FD,$7F,$15,$DD,$4F,$46,$09,$A7,$98,$95,$E6,$6B,$FD,$22
	.dc.b	$38,$3F,$E1,$06,$BF,$DD,$70,$7C,$BB,$E9,$BE,$37,$D3,$7F,$2A,$AF
	.dc.b	$F0,$BF,$D3,$7F,$68,$6F,$A3,$AF,$26,$47,$F0,$3A,$E0,$FC,$60,$E1
	.dc.b	$F9,$2D,$7F,$81,$BC,$38,$3F,$C1,$AD,$E0,$CD,$7D,$9C,$3F,$77,$7F
	.dc.b	$D9,$3D,$B7,$C6,$7D,$DE,$DF,$98,$7C,$D8,$CC,$FF,$8C,$1A,$F9,$5A
	.dc.b	$FD,$C6,$BF,$C5,$9A,$FF,$B7,$1E,$39,$7F,$88,$9B,$8E,$EF,$C5,$D5
	.dc.b	$F8,$FF,$A5,$FF,$2B,$E1,$F9,$8F,$A6,$FC,$E9,$9C,$24,$7F,$2B,$9D
	.dc.b	$BF,$25,$AF,$E4,$1A,$FF,$2A,$77,$8F,$F8,$D1,$B8,$E6,$BF,$81,$79
	.dc.b	$EF,$97,$F2,$78,$F8,$F6,$26,$5F,$E5,$4C,$53,$A8,$F2,$3F,$A6,$8C
	.dc.b	$C6,$08,$7F,$85,$9C,$B6,$3F,$8F,$43,$DD,$FD,$EF,$D3,$4A,$FD,$DB
	.dc.b	$FB,$5F,$A6,$7D,$1D,$3F,$6A,$E2,$7F,$A8,$35,$FC,$73,$5F,$F3,$E7
	.dc.b	$85,$FC,$BD,$08,$E1,$BF,$F9,$69,$BB,$FF,$9F,$2B,$87,$90,$7F,$CD
	.dc.b	$97,$FC,$43,$E0,$7F,$8A,$9F,$CF,$7E,$25,$E7,$3E,$3F,$F8,$77,$C2
	.dc.b	$FF,$90,$EF,$FA,$FB,$5F,$1B,$5F,$F5,$CE,$5F,$DA,$EF,$9F,$B1,$D7
	.dc.b	$0F,$F1,$4D,$7F,$A4,$67,$2D,$F9,$8F,$B8,$FF,$A3,$3E,$57,$FC,$D0
	.dc.b	$FF,$A3,$BE,$97,$FA,$0B,$7C,$FF,$E4,$53,$A8,$FF,$26,$B5,$FE,$3A
	.dc.b	$D7,$FB,$BB,$5F,$96,$D4,$F3,$E1,$FF,$42,$B7,$3E,$3C,$FF,$CC,$70
	.dc.b	$7D,$0F,$F3,$EF,$47,$F9,$FF,$B7,$32,$FE,$7B,$19,$7F,$9F,$C6,$5F
	.dc.b	$E7,$B1,$FD,$1B,$E9,$BF,$A3,$D4,$9F,$E8,$B1,$97,$FA,$3C,$65,$FE
	.dc.b	$8B,$1F,$D3,$BD,$BF,$E9,$FA,$C9,$FE,$9B,$19,$7F,$A7,$C6,$5F,$E9
	.dc.b	$B1,$97,$FA,$7C,$67,$5F,$F5,$9F,$2A,$BF,$B3,$F1,$7F,$6B,$BF,$F5
	.dc.b	$27,$33,$CE,$FE,$59,$18,$6E,$1F,$ED,$23,$FC,$D3,$E2,$72,$24,$FF
	.dc.b	$98,$2F,$A4,$F8,$3E,$7B,$FB,$4F,$C2,$7F,$6A,$F7,$E5,$FF,$68,$8C
	.dc.b	$BF,$DA,$A3,$2F,$F6,$88,$FE,$CD,$F4,$DF,$A9,$87,$1B,$8F,$ED,$1A
	.dc.b	$F7,$24,$FF,$90,$EB,$FC,$4C,$FF,$66,$6B,$FE,$40,$6B,$FB,$47,$B1
	.dc.b	$23,$F0,$DD,$BF,$E0,$2E,$FF,$F0,$D7,$37,$FB,$43,$5F,$F4,$67,$46
	.dc.b	$71,$FF,$01,$D6,$A2,$46,$0F,$F7,$7F,$82,$F9,$FF,$F1,$27,$D3,$7F
	.dc.b	$84,$27,$FD,$E4,$D7,$F8,$B4,$7F,$81,$E7,$84,$72,$8C,$1C,$BD,$4F
	.dc.b	$F0,$67,$D2,$FF,$AD,$67,$FD,$39,$F7,$1F,$CC,$9F,$F7,$F3,$54,$6D
	.dc.b	$FF,$5E,$B5,$FF,$06,$B5,$FF,$33,$B5,$FE,$36,$D7,$9F,$C3,$78,$88
	.dc.b	$FF,$20,$D1,$9F,$F2,$2F,$05,$F4,$FF,$B2,$AF,$C5,$7C,$E7,$F8,$47
	.dc.b	$CC,$91,$82,$1F,$EF,$BC,$3F,$D7,$73,$F9,$C3,$EA,$FA,$FF,$9E,$F9
	.dc.b	$6F,$F0,$D7,$8B,$2F,$F2,$D1,$96,$90,$FD,$1F,$27,$C3,$C3,$F4,$1D
	.dc.b	$AF,$95,$FF,$47,$47,$FB,$29,$B1,$9A,$FF,$07,$4E,$5B,$19,$7D,$0D
	.dc.b	$8E,$B3,$FE,$5E,$F5,$A5,$7F,$B1,$5A,$FE,$DB,$C5,$7C,$A6,$2F,$E5
	.dc.b	$CB,$C4,$FF,$36,$DE,$78,$BF,$A7,$C7,$F9,$C2,$FF,$CC,$FF,$1B,$FE
	.dc.b	$18,$BF,$F4,$73,$5F,$AC,$D5,$82,$30,$DF,$33,$99,$27,$FD,$B9,$F3
	.dc.b	$DF,$E8,$BF,$A4,$7D,$3F,$FA,$2D,$B3,$FE,$A0,$6B,$FE,$5A,$F8,$94
	.dc.b	$62,$3F,$D6,$B1,$89,$E8,$48,$FF,$75,$7D,$05,$7C,$FA,$FF,$B5,$EF
	.dc.b	$AE,$5F,$EE,$26,$C7,$C4,$DD,$7C,$6B,$58,$5A,$D4,$FA,$FC,$3F,$51
	.dc.b	$31,$F5,$35,$F0,$8C,$BD,$71,$C6,$3F,$E0,$07,$3F,$FA,$07,$E3,$5F
	.dc.b	$4F,$F3,$A5,$F4,$FF,$E8,$B7,$FD,$A9,$E1,$7F,$B5,$A3,$15,$FC,$0A
	.dc.b	$77,$F1,$3F,$ED,$A8,$E7,$B7,$F8,$3A,$B7,$87,$5F,$87,$E3,$8F,$FB
	.dc.b	$81,$B8,$FF,$85,$1D,$DD,$C7,$A5,$F9,$70,$DD,$F8,$0F,$F0,$DE,$3E
	.dc.b	$38,$E3,$19,$3F,$DD,$93,$C6,$BC,$D8,$FA,$DB,$3E,$77,$BE,$16,$3A
	.dc.b	$D7,$A6,$77,$1F,$E9,$46,$E3,$E0,$3C,$38,$79,$89,$1D,$27,$C9,$8F
	.dc.b	$C6,$6E,$3E,$37,$D6,$D7,$D3,$8F,$8D,$A7,$03,$13,$F0,$6B,$B5,$F8
	.dc.b	$DF,$29,$27,$D7,$5F,$35,$A7,$32,$4F,$AA,$7D,$18,$7D,$02,$38,$24
	.dc.b	$FA,$0E,$87,$E2,$0B,$7E,$F9,$FA,$BC,$7F,$EB,$C9,$F0,$B5,$F6,$B6
	.dc.b	$7E,$33,$98,$FF,$2C,$FC,$EF,$F9,$6A,$31,$08,$C5,$7F,$AB,$DB,$1F
	.dc.b	$ED,$66,$BF,$E6,$23,$81,$F6,$CC,$7C,$8F,$F1,$C9,$FD,$14,$78,$0D
	.dc.b	$1F,$E6,$A8,$D6,$1D,$E3,$22,$32,$F8,$4C,$B8,$7C,$55,$F7,$5E,$1C
	.dc.b	$35,$F2,$E5,$E2,$AF,$E5,$EF,$92,$3D,$38,$66,$BE,$8C,$79,$E3,$0F
	.dc.b	$90,$E7,$87,$FB,$E2,$F9,$BE,$9B,$D9,$D7,$0F,$24,$EA,$BD,$53,$AA
	.dc.b	$E0,$73,$C3,$D1,$7C,$1B,$5F,$01,$F2,$BE,$89,$E0,$F9,$67,$83,$FB
	.dc.b	$93,$E8,$57,$77,$C3,$3C,$23,$CD,$1B,$BF,$33,$F3,$0F,$01,$D4,$B8
	.dc.b	$70,$9E,$91,$B7,$B7,$07,$F4,$2F,$33,$FC,$A1,$97,$F3,$B6,$D7,$F5
	.dc.b	$FE,$4F,$F4,$F4,$FF,$83,$4D,$16,$3F,$99,$72,$FF,$91,$E3,$FA,$85
	.dc.b	$7F,$A8,$EF,$B4,$60,$FE,$A9,$E8,$76,$8F,$EA,$6F,$83,$A9,$8B,$BF
	.dc.b	$58,$AA,$93,$0F,$15,$F0,$AC,$65,$FD,$1B,$1E,$B5,$E7,$6C,$7F,$7F
	.dc.b	$9D,$F1,$7F,$59,$C3,$D1,$8F,$7C,$3C,$E4,$BF,$D6,$1F,$DE,$3A,$BF
	.dc.b	$DD,$D1,$F0,$EF,$FB,$D7,$85,$EB,$D3,$FC,$39,$39,$AD,$DF,$F8,$B5
	.dc.b	$AF,$F8,$1A,$BF,$D9,$57,$CA,$BF,$C3,$F5,$AA,$FE,$D9,$F3,$9D,$4E
	.dc.b	$DE,$55,$AF,$87,$8A,$71,$5F,$E1,$DA,$B8,$7F,$A3,$F0,$FF,$33,$61
	.dc.b	$E5,$E7,$CB,$FE,$16,$7F,$07,$3F,$E1,$57,$CB,$3C,$2F,$E1,$CE,$6B
	.dc.b	$CF,$5A,$FA,$7E,$36,$7E,$F7,$86,$F8,$2C,$C4,$BF,$B7,$EA,$FF,$B9
	.dc.b	$B2,$F8,$D3,$F8,$58,$C6,$5E,$9F,$82,$EA,$74,$F8,$2B,$87,$CF,$8E
	.dc.b	$1F,$7F,$8C,$7F,$D6,$0F,$E7,$9F,$F7,$DF,$BB,$7B,$BF,$DB,$B7,$F9
	.dc.b	$DF,$35,$F5,$0C,$75,$72,$BF,$07,$87,$F3,$78,$FF,$80,$18,$72,$8C
	.dc.b	$B5,$FC,$4B,$9C,$FF,$C9,$BD,$4F,$C1,$FF,$21,$5F,$E1,$E3,$CB,$87
	.dc.b	$E7,$A3,$35,$FE,$54,$97,$D4,$7F,$96,$2B,$F6,$91,$8F,$2B,$F5,$9F
	.dc.b	$52,$F0,$7E,$05,$EE,$BB,$37,$0F,$1D,$7F,$8A,$30,$E7,$39,$7F,$CB
	.dc.b	$F1,$9A,$FF,$3D,$E9,$F7,$3E,$A7,$EE,$B6,$BF,$6B,$CD,$EF,$7F,$CC
	.dc.b	$EA,$77,$1A,$7E,$97,$4C,$8F,$E1,$C3,$E9,$FF,$A4,$E1,$F1,$78,$15
	.dc.b	$B8,$F9,$EF,$D2,$25,$7D,$68,$4F,$08,$EB,$4F,$90,$E5,$AF,$FA,$02
	.dc.b	$7C,$11,$F9,$8F,$99,$FC,$CF,$C0,$D9,$F6,$B7,$1F,$F4,$EB,$6F,$EE
	.dc.b	$E8,$8F,$D1,$FC,$74,$BF,$8B,$F2,$F7,$0F,$E0,$93,$F9,$FE,$FC,$63
	.dc.b	$F8,$6E,$CF,$D8,$3D,$3D,$3F,$D5,$F3,$7E,$B1,$B9,$7D,$06,$E3,$FE
	.dc.b	$B0,$6D,$71,$9F,$AB,$5C,$5A,$7D,$8F,$4F,$F6,$FE,$97,$EE,$1B,$19
	.dc.b	$3E,$EC,$65,$A7,$E0,$3D,$1F,$DF,$FC,$57,$F0,$06,$C6,$4F,$C2,$C6
	.dc.b	$5A,$7E,$6B,$EA,$5F,$53,$FB,$76,$C7,$D8,$FA,$9C,$4C,$8F,$BD,$3A
	.dc.b	$D7,$F6,$AD,$7F,$9C,$B5,$FF,$13,$3B,$D3,$F0,$1C,$D6,$B3,$7F,$D9
	.dc.b	$1F,$C9,$3F,$D9,$BE,$31,$F5,$3F,$C5,$25,$8A,$73,$F1,$DB,$AF,$E8
	.dc.b	$38,$FF,$B3,$27,$37,$FB,$8F,$59,$D3,$FB,$DC,$7F,$C5,$2D,$7F,$C8
	.dc.b	$AD,$7F,$C8,$8E,$7C,$37,$57,$F6,$9B,$FD,$8F,$A3,$E1,$FA,$9F,$E6
	.dc.b	$2F,$F3,$2F,$73,$F9,$16,$5F,$95,$AF,$DB,$C6,$3B,$64,$7E,$A6,$75
	.dc.b	$9F,$F9,$46,$7C,$51,$88,$FF,$28,$65,$E2,$F3,$FF,$9F,$9C,$A3,$FE
	.dc.b	$00,$BF,$CC,$63,$FA,$15,$FF,$18,$6C,$FF,$79,$E0,$FE,$B5,$39,$F3
	.dc.b	$B8,$1F,$06,$B7,$5F,$DD,$1A,$F8,$A3,$47,$F8,$37,$E1,$DE,$07,$2B
	.dc.b	$FF,$04,$C6,$1F,$17,$9D,$E5,$38,$76,$D8,$C9,$FC,$2F,$14,$F8,$C9
	.dc.b	$19,$7F,$AF,$F8,$9C,$72,$FF,$6C,$D7,$D6,$8F,$EA,$5F,$50,$FA,$7A
	.dc.b	$7F,$8C,$BE,$A1,$F5,$3F,$84,$CB,$F9,$ED,$6D,$B5,$F8,$38,$C6,$3F
	.dc.b	$D5,$1A,$FF,$86,$1A,$FF,$A1,$5D,$F4,$AB,$F3,$3E,$CB,$EA,$5F,$09
	.dc.b	$E4,$9F,$F1,$54,$61,$FE,$9F,$F5,$0F,$90,$FE,$D8,$D8,$FF,$18,$7D
	.dc.b	$4B,$E7,$B3,$52,$7F,$9A,$56,$F4,$FE,$D6,$D7,$FC,$5E,$D7,$FD,$38
	.dc.b	$EF,$3F,$44,$17,$D8,$7E,$22,$3F,$B5,$4E,$EB,$F6,$B5,$AE,$9F,$F0
	.dc.b	$17,$C6,$3E,$AA,$9F,$E0,$5F,$7B,$FA,$7D,$7F,$6F,$9D,$5F,$F6,$E6
	.dc.b	$E3,$FE,$02,$6B,$FE,$46,$6B,$FE,$B3,$6B,$FC,$19,$E3,$3F,$CF,$7E
	.dc.b	$23,$C3,$5F,$E2,$27,$CB,$ED,$C8,$FE,$FB,$83,$E9,$71,$8F,$F8,$1F
	.dc.b	$1F,$F0,$B5,$7F,$5F,$6C,$79,$6F,$F8,$44,$62,$3F,$6F,$E2,$BE,$33
	.dc.b	$FC,$5D,$EB,$FF,$8D,$9B,$87,$F7,$A6,$D7,$E6,$AB,$84,$FF,$8E,$2B
	.dc.b	$FC,$8F,$F5,$4F,$52,$9F,$E6,$3C,$2D,$7F,$99,$E7,$4D,$9D,$34,$FF
	.dc.b	$3C,$7C,$BF,$FA,$37,$1F,$F3,$E3,$77,$FF,$14,$FD,$57,$F7,$CA,$DF
	.dc.b	$C4,$FF,$A3,$6B,$51,$F9,$C6,$DB,$C1,$5F,$18,$EB,$7C,$4F,$F4,$D5
	.dc.b	$78,$63,$FD,$D9,$D8,$8C,$47,$59,$DF,$CC,$F8,$4F,$F4,$E5,$FF,$8C
	.dc.b	$71,$F5,$C6,$5A,$7F,$A6,$A3,$55,$EE,$9F,$F2,$B7,$D5,$7F,$99,$FE
	.dc.b	$37,$FC,$34,$D9,$FF,$28,$B5,$FF,$4A,$B5,$FE,$A6,$3E,$3F,$FC,$E5
	.dc.b	$F4,$BF,$E2,$36,$C7,$F9,$69,$AF,$FA,$79,$AF,$29,$FE,$3F,$C5,$E2
	.dc.b	$F1,$FF,$C5,$73,$E2,$6B,$E2,$6B,$FC,$33,$EA,$9F,$25,$E4,$E2,$FB
	.dc.b	$3E,$0D,$DF,$FB,$4F,$1F,$D4,$8C,$60,$E3,$63,$CA,$3F,$D7,$8D,$7F
	.dc.b	$96,$F0,$FF,$0F,$7E,$65,$72,$AE,$39,$5B,$FE,$CF,$5F,$47,$E3,$BB
	.dc.b	$E3,$F8,$47,$5D,$12,$74,$43,$FD,$9D,$5C,$9B,$1C,$B2,$55,$39,$4F
	.dc.b	$F1,$7C,$FF,$89,$FD,$53,$E9,$78,$78,$F2,$BF,$D9,$93,$96,$CF,$F9
	.dc.b	$6F,$74,$E6,$7F,$D6,$6D,$7E,$73,$5F,$9B,$F3,$EF,$95,$F7,$BF,$C2
	.dc.b	$E3,$FD,$D6,$D7,$FA,$BB,$B7,$31,$C3,$C0,$91,$83,$FD,$23,$F5,$0F
	.dc.b	$AA,$FB,$3D,$3F,$DC,$38,$EF,$A7,$F5,$F4,$FF,$4D,$5E,$F0,$E3,$86
	.dc.b	$AE,$46,$0E,$BF,$3B,$F5,$BE,$1F,$FD,$39,$1F,$EC,$8F,$8F,$FC,$3A
	.dc.b	$C1,$58,$6C,$6C,$87,$FA,$BD,$CB,$63,$B4,$C8,$F0,$9F,$EA,$EF,$A9
	.dc.b	$FE,$3C,$FF,$AC,$3E,$A7,$F2,$0F,$FA,$E5,$B9,$FF,$AD,$E7,$73,$FE
	.dc.b	$BD,$6C,$7F,$AE,$E7,$7C,$9F,$B8,$6D,$7F,$93,$2B,$73,$FB,$C7,$9C
	.dc.b	$7E,$F8,$38,$75,$E2,$7F,$A8,$5C,$B7,$D9,$7C,$67,$C8,$F7,$7E,$95
	.dc.b	$7F,$B4,$5B,$97,$FB,$3E,$B7,$3F,$EC,$E6,$BB,$E0,$E4,$7D,$28,$CC
	.dc.b	$F1,$BC,$12,$BF,$DA,$F7,$F9,$99,$FF,$6E,$7D,$42,$FF,$2B,$F5,$32
	.dc.b	$FF,$DB,$B3,$96,$CE,$5F,$CB,$CF,$E0,$A7,$51,$FE,$EB,$6D,$76,$AD
	.dc.b	$1F,$C0,$3C,$DF,$DB,$D7,$6F,$0F,$1D,$92,$30,$43,$FC,$77,$3F,$E2
	.dc.b	$67,$FC,$1B,$BF,$F1,$A8,$C4,$78,$6B,$C8,$47,$96,$1F,$E1,$AF,$09
	.dc.b	$97,$FB,$72,$7F,$67,$F5,$3F,$B6,$23,$FE,$FE,$FA,$97,$D2,$FE,$81
	.dc.b	$B1,$C0,$87,$F0,$F3,$FD,$2A,$E9,$FD,$EF,$4F,$19,$91,$FC,$0A,$3F
	.dc.b	$82,$FC,$04,$8F,$E0,$AF,$F3,$4F,$87,$FF,$75,$FC,$6F,$EC,$3E,$0F
	.dc.b	$FC,$67,$5D,$E7,$2F,$C4,$3F,$D4,$9E,$5A,$B8,$70,$F3,$8F,$F0,$94
	.dc.b	$EE,$BF,$DE,$FE,$9E,$4F,$07,$CC,$FD,$69,$F6,$0F,$95,$F9,$FC,$5E
	.dc.b	$2F,$A7,$7C,$EF,$F0,$46,$D7,$8C,$E1,$CF,$FC,$22,$F8,$E9,$FD,$A2
	.dc.b	$B7,$5D,$9B,$87,$BA,$B7,$3F,$EC,$06,$A9,$DD,$73,$C3,$B6,$1F,$CC
	.dc.b	$3C,$DF,$E6,$3E,$AC,$C3,$F9,$8C,$65,$FE,$67,$19,$7F,$99,$C7,$F3
	.dc.b	$8F,$A8,$FE,$73,$F5,$32,$7F,$9C,$C6,$5F,$E7,$71,$97,$F9,$DC,$7F
	.dc.b	$40,$F5,$FF,$A0,$E3,$27,$FA,$0C,$65,$FE,$87,$19,$7F,$A1,$C6,$5F
	.dc.b	$E8,$B1,$9D,$7F,$CE,$7E,$9B,$CE,$5F,$F7,$A9,$FF,$27,$78,$3F,$CE
	.dc.b	$7D,$49,$8F,$E1,$67,$2D,$E3,$56,$A3,$FD,$53,$E5,$FF,$24,$AD,$C7
	.dc.b	$F5,$6F,$81,$FE,$B7,$EC,$C8,$FE,$B7,$1F,$D7,$8C,$3F,$D7,$23,$0F
	.dc.b	$E0,$FE,$A7,$FA,$A7,$27,$E0,$78,$FF,$10,$7F,$4C,$8C,$CF,$FB,$F9
	.dc.b	$AF,$DF,$8D,$BC,$FB,$78,$18,$7F,$15,$18,$7F,$B7,$F7,$7F,$84,$39
	.dc.b	$7F,$53,$9F,$CA,$6E,$3F,$DE,$27,$FB,$B9,$B6,$57,$23,$F3,$9F,$12
	.dc.b	$F9,$8F,$F0,$DF,$D4,$FF,$2D,$AF,$F1,$6B,$68,$2B,$83,$CA,$B8,$FA
	.dc.b	$F2,$30,$7E,$A7,$EA,$1F,$01,$FE,$28,$FA,$9F,$E5,$B9,$7F,$8E,$9B
	.dc.b	$5D,$AB,$6F,$1A,$E1,$E6,$48,$C1,$0F,$F7,$3D,$72,$EA,$3D,$21,$7D
	.dc.b	$1F,$E3,$DB,$FF,$A6,$FE,$67,$FC,$0D,$EE,$49,$FC,$64,$7E,$4F,$A4
	.dc.b	$E6,$3F,$1B,$39,$8D,$7C,$97,$F8,$3F,$C1,$95,$F9,$58,$CB,$FC,$77
	.dc.b	$37,$57,$F0,$46,$D6,$A7,$CB,$8B,$E4,$BF,$C8,$DE,$4C,$9F,$D0,$47
	.dc.b	$94,$C3,$FE,$5C,$8C,$3D,$FC,$1F,$F0,$1B,$FE,$9B,$6E,$1F,$C8,$3D
	.dc.b	$4F,$F2,$DC,$3E,$46,$C4,$AE,$4E,$A3,$F6,$DF,$0D,$FE,$7A,$E0,$FF
	.dc.b	$56,$FD,$17,$F9,$6C,$CF,$99,$30,$FF,$4C,$46,$5F,$DD,$57,$1E,$7F
	.dc.b	$C6,$FA,$4F,$A4,$7D,$3F,$F9,$97,$B2,$61,$FD,$16,$B5,$3F,$E7,$7A
	.dc.b	$DC,$FF,$A9,$6B,$6F,$F5,$BA,$DB,$C6,$76,$E5,$B9,$FF,$6D,$6E,$1F
	.dc.b	$E8,$ED,$9F,$3B,$DF,$29,$87,$37,$E4,$C7,$8F,$C3,$FF,$3E,$7C,$5C
	.dc.b	$AF,$04,$65,$FF,$5D,$46,$5F,$14,$7F,$95,$3E,$25,$2C,$FF,$63,$F4
	.dc.b	$FF,$E6,$DB,$FF,$5B,$6D,$FE,$B8,$8C,$5F,$ED,$25,$18,$BE,$51,$8F
	.dc.b	$86,$FF,$5F,$D7,$6C,$3A,$C6,$30,$E8,$DA,$C9,$FC,$0A,$B8,$3C,$CC
	.dc.b	$5F,$6C,$3C,$77,$CA,$78,$35,$FF,$67,$4E,$EF,$C6,$D8,$D9,$E0,$3D
	.dc.b	$E5,$BF,$0D,$FB,$70,$C6,$1D,$67,$E1,$4E,$6B,$CC,$78,$2B,$6C,$33
	.dc.b	$3D,$30,$F1,$CE,$6B,$D7,$5E,$1C,$3C,$CF,$33,$8E,$1E,$13,$C5,$1A
	.dc.b	$C3,$A4,$FB,$0D,$9C,$EE,$38,$25,$FA,$49,$87,$D3,$8C,$BE,$67,$D3
	.dc.b	$7D,$F0,$F1,$47,$4A,$F8,$5C,$5F,$EA,$28,$F4,$37,$1F,$B6,$DB,$FC
	.dc.b	$3B,$B9,$FE,$83,$F2,$FF,$D0,$A3,$10,$8C,$7C,$17,$FA,$36,$30,$27
	.dc.b	$73,$A8,$FF,$49,$3F,$75,$7C,$63,$87,$11,$CB,$2F,$D0,$3A,$8C,$D7
	.dc.b	$1A,$E1,$3F,$78,$F3,$95,$CD,$72,$8E,$9A,$7F,$A8,$E7,$2F,$A6,$BA
	.dc.b	$5F,$6A,$E4,$DB,$F1,$DF,$09,$F5,$9A,$3C,$47,$63,$FB,$49,$E9,$33
	.dc.b	$5D,$9A,$71,$9F,$34,$65,$B1,$9B,$F3,$34,$F0,$C7,$92,$FC,$D3,$ED
	.dc.b	$6C,$F8,$AF,$D0,$F9,$2F,$B5,$74,$BE,$53,$E4,$69,$E2,$2C,$6C,$ED
	.dc.b	$1B,$6C,$1F,$E7,$BC,$4F,$2B,$EB,$EF,$FE,$42,$F8,$A7,$D3,$63,$4D
	.dc.b	$D5,$C9,$D6,$BD,$D7,$F3,$BA,$D4,$E6,$F8,$73,$FF,$4B,$9C,$DF,$0C
	.dc.b	$3F,$A0,$7D,$3E,$B3,$3C,$BA,$1E,$16,$62,$69,$FD,$25,$D4,$62,$F9
	.dc.b	$CF,$F7,$DC,$1F,$21,$9D,$BD,$CD,$9F,$F1,$5E,$DE,$B8,$C5,$FB,$AF
	.dc.b	$8D,$7F,$66,$F1,$7D,$3F,$05,$F0,$F1,$76,$A7,$35,$B3,$C5,$5B,$9D
	.dc.b	$46,$0F,$F0,$27,$C8,$FB,$CF,$81,$E6,$D8,$C9,$32,$FF,$62,$69,$FE
	.dc.b	$A6,$9D,$47,$F8,$AA,$FF,$C9,$55,$F0,$7E,$5F,$C1,$DB,$F4,$1B,$19
	.dc.b	$3F,$C0,$31,$97,$FC,$07,$CB,$F6,$BC,$87,$8B,$E9,$8C,$36,$7F,$C5
	.dc.b	$53,$A7,$89,$88,$F8,$1A,$F3,$F0,$EC,$F0,$6E,$7F,$E8,$66,$F0,$7F
	.dc.b	$A5,$30,$47,$E4,$67,$F8,$D7,$B9,$F8,$EF,$A6,$FC,$3E,$1F,$EC,$5E
	.dc.b	$FF,$F6,$63,$FE,$3D,$8C,$D7,$FB,$42,$72,$FC,$A8,$CE,$3F,$EE,$29
	.dc.b	$CB,$FE,$A4,$8C,$D7,$F0,$89,$CB,$F8,$88,$CD,$7F,$33,$BF,$27,$3F
	.dc.b	$9F,$E9,$BF,$C7,$57,$FE,$4F,$FA,$6F,$F0,$5B,$FE,$DB,$8C,$DF,$DC
	.dc.b	$9C,$BF,$EE,$68,$CE,$5F,$EB,$69,$CB,$FE,$F8,$8C,$D7,$E6,$A7,$2F
	.dc.b	$E7,$3D,$5F,$3D,$FE,$E1,$7D,$37,$ED,$9C,$F3,$1C,$D7,$F9,$95,$C9
	.dc.b	$8A,$FC,$8B,$8A,$FA,$9A,$FF,$A4,$EB,$FD,$2B,$97,$F4,$25,$F4,$4F
	.dc.b	$A6,$DF,$47,$F0,$B9,$CC,$61,$E2,$7F,$A7,$8C,$9F,$76,$38,$4F,$38
	.dc.b	$C1,$C3,$2F,$EB,$11,$97,$F0,$13,$C2,$BE,$07,$D0,$FC,$1F,$76,$E1
	.dc.b	$EF,$6D,$FB,$9B,$8F,$FA,$F5,$FD,$D7,$C6,$FE,$97,$A6,$DF,$E8,$9B
	.dc.b	$7F,$EE,$06,$DF,$E6,$DF,$A1,$E8,$7D,$1F,$42,$DF,$CF,$6D,$FC,$E6
	.dc.b	$DF,$FB,$35,$B3,$F7,$5B,$3F,$75,$FB,$9F,$4D,$F7,$7D,$0B,$5F,$6D
	.dc.b	$B5,$F6,$9B,$7F,$ED,$97,$F0,$9E,$AF,$E1,$7E,$22,$DF,$E0,$DB,$7F
	.dc.b	$82,$6D,$FF,$B8,$5F,$DB,$7D,$2B,$E9,$7E,$B9,$82,$DF,$D7,$B4,$EE
	.dc.b	$3E,$D1,$F8,$89,$E0,$E3,$8F,$55,$FE,$73,$8F,$EF,$75,$8C,$3F,$21
	.dc.b	$56,$73,$7E,$0B,$E0,$3E,$29,$F4,$BF,$C6,$63,$31,$8F,$0B,$FB,$03
	.dc.b	$AF,$1E,$D6,$F2,$FF,$0F,$DF,$F8,$C7,$97,$3F,$4B,$FC,$5F,$D2,$46
	.dc.b	$34,$FD,$67,$99,$FB,$13,$17,$FA,$C7,$F6,$9E,$CF,$1E,$9B,$5F,$1F
	.dc.b	$04,$66,$3F,$6D,$E5,$DC,$3F,$6A,$FE,$F4,$91,$C6,$6C,$FC,$23,$FB
	.dc.b	$DF,$BD,$FC,$B7,$C1,$3F,$DE,$38,$58,$66,$26,$1F,$EE,$9F,$8D,$7C
	.dc.b	$C6,$31,$FE,$AB,$18,$9F,$BD,$8F,$47,$85,$6F,$83,$FC,$8F,$2A,$E1
	.dc.b	$7E,$58,$7C,$97,$94,$FF,$06,$CF,$F2,$09,$FE,$F3,$19,$C3,$F8,$C7
	.dc.b	$83,$67,$F8,$BB,$70,$EF,$E2,$7F,$17,$9E,$D7,$FC,$8E,$BF,$A7,$7D
	.dc.b	$1B,$CF,$A7,$F8,$87,$E9,$5F,$4B,$F7,$33,$FC,$D5,$6D,$B5,$FE,$20
	.dc.b	$FA,$57,$C1,$F2,$9F,$F7,$44,$61,$D9,$88,$D6,$7F,$EE,$79,$D3,$F5
	.dc.b	$FE,$8D,$F1,$BF,$90,$8F,$F2,$DC,$62,$FF,$D0,$10,$8C,$3F,$E8,$36
	.dc.b	$BF,$E6,$23,$53,$FE,$1B,$FA,$4F,$BF,$5F,$B5,$AE,$3C,$1F,$E4,$69
	.dc.b	$CD,$F8,$EF,$87,$2F,$F7,$5B,$DE,$1F,$D0,$AF,$78,$7E,$9B,$E2,$7F
	.dc.b	$C5,$78,$FF,$86,$BD,$5F,$EA,$E8,$46,$A7,$FC,$05,$F0,$7F,$E0,$18
	.dc.b	$C6,$1F,$E3,$8B,$E3,$7C,$3D,$BF,$F0,$C5,$ED,$B5,$FC,$3E,$F8,$D7
	.dc.b	$F9,$6A,$F8,$C3,$D8,$FF,$2F,$7C,$FF,$EE,$3E,$6B,$C1,$18,$F0,$91
	.dc.b	$8C,$7F,$A8,$56,$DE,$D5,$B9,$FE,$DD,$F1,$3F,$E2,$BF,$52,$D7,$F8
	.dc.b	$A1,$B9,$FC,$36,$ED,$FE,$21,$7F,$C9,$DF,$48,$F4,$BF,$01,$4A,$4E
	.dc.b	$6F,$FD,$81,$18,$AD,$B6,$33,$3B,$AF,$F2,$5F,$BE,$F2,$3B,$E1,$FE
	.dc.b	$76,$7F,$CF,$3F,$23,$FE,$87,$C3,$FC,$D9,$F6,$D8,$F8,$4F,$F3,$DB
	.dc.b	$63,$27,$F7,$3F,$85,$57,$1C,$BB,$D6,$DB,$5B,$9D,$69,$FE,$93,$AF
	.dc.b	$0F,$CA,$7F,$A2,$3E,$DE,$DF,$F9,$F5,$B8,$7A,$9B,$7F,$E9,$B7,$FC
	.dc.b	$91,$F4,$DF,$C7,$3E,$2F,$F4,$AD,$AF,$F7,$33,$5F,$D2,$35,$FF,$63
	.dc.b	$FC,$6F,$FA,$93,$9B,$F5,$AD,$8F,$F7,$73,$5F,$D6,$35,$FF,$67,$CF
	.dc.b	$F1,$AE,$3F,$0F,$8F,$FE,$F3,$9F,$0B,$5F,$0B,$5F,$D2,$6E,$FC,$BB
	.dc.b	$FD,$EB,$1F,$35,$AF,$F5,$C8,$D6,$5F,$ED,$ED,$3F,$D7,$9F,$27,$76
	.dc.b	$E7,$87,$09,$FF,$7B,$DF,$0C,$3A,$61,$B7,$FD,$45,$85,$97,$D3,$BE
	.dc.b	$47,$0D,$AF,$C0,$35,$FF,$4B,$FC,$6F,$F6,$27,$33,$C0,$8F,$F8,$E7
	.dc.b	$CF,$4E,$E3,$D9,$5C,$F2,$F3,$B7,$4F,$8A,$D9,$FF,$6A,$D7,$1A,$F6
	.dc.b	$E2,$E3,$F5,$36,$BF,$2C,$D7,$F4,$7F,$07,$A9,$91,$A7,$E2,$65,$FE
	.dc.b	$E1,$9B,$E0,$FF,$BA,$2B,$FD,$47,$E1,$60,$CF,$C6,$48,$C9,$0F,$E4
	.dc.b	$4F,$CF,$5F,$4A,$BC,$36,$3F,$C5,$0D,$7F,$C3,$7C,$EF,$AD,$FD,$D4
	.dc.b	$6A,$B8,$91,$FE,$2F,$1A,$7F,$C0,$54,$FA,$3E,$E4,$8D,$B5,$DB,$76
	.dc.b	$70,$FF,$39,$C7,$C5,$39,$8F,$E7,$95,$EC,$8C,$29,$DB,$FE,$43,$FC
	.dc.b	$13,$FE,$00,$B7,$E3,$7E,$24,$62,$3F,$C3,$1F,$46,$FA,$BC,$D4,$9E
	.dc.b	$07,$F9,$36,$74,$E0,$FF,$10,$57,$3C,$BF,$C7,$2D,$CB,$CF,$5C,$2B
	.dc.b	$FC,$4D,$F4,$DF,$CC,$5F,$ED,$AD,$7F,$B5,$07,$0F,$F6,$D6,$BF,$09
	.dc.b	$E1,$7F,$E4,$86,$D6,$7C,$4F,$C1,$7D,$3F,$F0,$3A,$D4,$7E,$16,$B7
	.dc.b	$7F,$D9,$DA,$AB,$85,$FD,$0F,$A7,$FE,$D7,$3F,$EF,$6F,$2F,$F2,$E7
	.dc.b	$E3,$E3,$31,$FE,$43,$6E,$3F,$E5,$C6,$D7,$3F,$37,$F3,$FE,$AF,$16
	.dc.b	$5F,$F9,$EE,$B6,$FF,$6D,$9E,$DE,$EF,$EA,$3E,$8B,$FB,$4C,$FE,$FD
	.dc.b	$AF,$86,$36,$F3,$6B,$FE,$78,$EA,$FE,$ED,$F0,$B8,$CF,$CD,$57,$3F
	.dc.b	$F0,$55,$6A,$FF,$86,$75,$FE,$7E,$FF,$C9,$ED,$AF,$F0,$57,$99,$FB
	.dc.b	$1F,$A7,$FE,$AF,$7F,$E9,$56,$BF,$E6,$76,$DF,$FA,$01,$B5,$FE,$13
	.dc.b	$F2,$FF,$81,$9C,$FC,$C9,$7E,$28,$FE,$0F,$E6,$48,$F2,$BF,$17,$9B
	.dc.b	$C5,$F4,$32,$7C,$4F,$C9,$F1,$A4,$7C,$96,$BF,$E2,$BF,$1F,$F8,$57
	.dc.b	$E5,$FF,$E7,$48,$FF,$57,$35,$FF,$40,$37,$5F,$87,$E1,$70,$F8,$29
	.dc.b	$5C,$1F,$E3,$B1,$88,$E3,$16,$78,$C7,$CE,$F0,$7F,$91,$FD,$4C,$8C
	.dc.b	$10,$FF,$20,$39,$7F,$C8,$51,$FC,$AB,$EE,$3F,$96,$FD,$DF,$F9,$56
	.dc.b	$FF,$D0,$AD,$E2,$FF,$20,$4E,$EF,$FA,$44,$E5,$73,$23,$0B,$99,$5D
	.dc.b	$BC,$DE,$FF,$2D,$30,$E7,$19,$7A,$C6,$5F,$C9,$61,$FD,$A7,$1F,$C7
	.dc.b	$B6,$3C,$2A,$78,$65,$EA,$BF,$ED,$5E,$8B,$E5,$29,$30,$F5,$F8,$F5
	.dc.b	$FE,$F9,$3B,$7F,$C5,$8D,$7C,$71,$F2,$3E,$3F,$E4,$FD,$14,$9F,$8B
	.dc.b	$19,$7E,$3C,$65,$F5,$C7,$F5,$1F,$9B,$FF,$4B,$C7,$FB,$D5,$AF,$FB
	.dc.b	$1D,$B9,$7F,$B8,$3E,$A9,$F5,$5C,$7E,$1A,$5F,$17,$FA,$E4,$EA,$30
	.dc.b	$43,$A4,$6A,$7F,$83,$7D,$57,$F0,$76,$4F,$F0,$78,$CB,$FC,$26,$32
	.dc.b	$FF,$05,$F1,$53,$FE,$19,$C1,$EF,$FF,$BA,$2B,$FD,$1D,$1F,$D6,$1F
	.dc.b	$F3,$54,$65,$FF,$36,$71,$EA,$31,$3F,$ED,$6B,$FE,$8D,$5C,$3E,$3F
	.dc.b	$B5,$6A,$F8,$47,$F1,$FF,$43,$FB,$E7,$D4,$FF,$85,$6C,$D3,$1A,$FF
	.dc.b	$B8,$63,$CB,$7F,$DA,$27,$5E,$07,$23,$E5,$FC,$7D,$AD,$9F,$EA,$99
	.dc.b	$DB,$81,$7D,$5F,$F8,$01,$F1,$CF,$9C,$FF,$08,$36,$38,$35,$FE,$61
	.dc.b	$F5,$2F,$03,$75,$27,$6F,$F8,$92,$75,$1E,$4E,$0F,$A3,$F3,$7F,$18
	.dc.b	$91,$FE,$3F,$F9,$27,$D8,$62,$3E,$89,$F5,$27,$15,$FE,$F8,$9F,$D4
	.dc.b	$FC,$BC,$AF,$D1,$BF,$E3,$4F,$88,$FF,$0D,$46,$08,$7E,$AA,$35,$E2
	.dc.b	$C8,$FF,$28,$4F,$07,$FD,$47,$1C,$F4,$D4,$60,$DC,$65,$FD,$9F,$C7
	.dc.b	$FE,$F0,$8F,$4F,$12,$BF,$BA,$8B,$97,$E4,$7C,$66,$3F,$C5,$2B,$ED
	.dc.b	$F8,$3F,$E2,$5F,$4F,$FC,$E4,$D8,$CF,$B8,$9F,$F2,$9F,$EF,$1F,$C8
	.dc.b	$6B,$55,$F2,$9B,$8F,$E7,$DB,$D3,$FD,$9A,$B8,$61,$CE,$FA,$E1,$E0
	.dc.b	$8C,$47,$F7,$19,$DC,$E4,$91,$FE,$79,$8E,$7F,$49,$C4,$91,$9F,$1E
	.dc.b	$BE,$5B,$4E,$32,$F8,$97,$FC,$AE,$9F,$14,$F8,$FF,$F7,$04,$3E,$A1
	.dc.b	$F5,$1F,$EB,$6E,$AF,$F2,$7C,$62,$47,$28,$CC,$FF,$AF,$2B,$9D,$6E
	.dc.b	$BB,$C6,$27,$A4,$72,$9D,$5F,$E9,$5B,$5B,$9E,$F3,$E7,$C7,$85,$F2
	.dc.b	$9F,$43,$6F,$8C,$EA,$78,$DF,$37,$D0,$64,$A7,$23,$71,$9B,$F0,$46
	.dc.b	$4D,$D7,$A3,$0F,$83,$1C,$0D,$4F,$09,$D4,$73,$9D,$BF,$12,$BB,$5F
	.dc.b	$96,$FE,$16,$1E,$B3,$E6,$4E,$8E,$55,$0E,$F5,$C4,$E5,$7D,$AF,$9C
	.dc.b	$F9,$70,$F8,$AF,$9A,$74,$6E,$B9,$5F,$5A,$F4,$D7,$68,$F1,$3C,$4D
	.dc.b	$61,$E6,$74,$37,$7D,$4E,$5B,$1F,$1D,$AF,$C3,$38,$47,$FA,$0B,$E6
	.dc.b	$FF,$1F,$3A,$9F,$C8,$C6,$27,$51,$C0,$FF,$52,$88,$D3,$87,$F0,$0B
	.dc.b	$7F,$EB,$18,$FE,$5E,$3A,$9F,$7C,$D7,$0F,$07,$A1,$99,$E3,$8F,$8A
	.dc.b	$B9,$46,$27,$C0,$6E,$B9,$5F,$8A,$75,$3F,$EB,$C8,$C4,$EA,$38,$1E
	.dc.b	$33,$C3,$5D,$A3,$4D,$8D,$4F,$26,$9C,$A3,$DC,$F2,$BF,$8F,$7E,$28
	.dc.b	$F8,$33,$CC,$DD,$73,$6C,$73,$AF,$80,$72,$BF,$1B,$63,$C3,$3C,$E3
	.dc.b	$71,$FE,$DA,$BF,$03,$E3,$31,$5C,$9A,$72,$9E,$95,$E2,$3A,$47,$E3
	.dc.b	$F8,$1F,$0F,$FE,$13,$D3,$F9,$B4,$63,$0F,$E5,$9E,$7E,$5F,$C6,$C6
	.dc.b	$6F,$F9,$55,$E5,$E8,$DE,$57,$17,$F9,$4E,$FF,$A4,$CE,$F7,$FF,$79
	.dc.b	$7B,$D7,$13,$C6,$FF,$DD,$D7,$FD,$42,$34,$47,$FA,$07,$9F,$72,$FE
	.dc.b	$79,$37,$4F,$E7,$73,$6B,$D3,$36,$BF,$CE,$F7,$D9,$E3,$3F,$E5,$EF
	.dc.b	$07,$D9,$F1,$5F,$B5,$7C,$9E,$7E,$34,$45,$25,$7F,$53,$9B,$7F,$D6
	.dc.b	$7D,$0F,$61,$ED,$E2,$B5,$FD,$66,$2C,$FF,$5F,$9B,$AF,$F9,$1F,$27
	.dc.b	$D6,$FF,$07,$C3,$FC,$63,$18,$CF,$71,$FD,$42,$30,$EA,$78,$61,$FE
	.dc.b	$1E,$9C,$BE,$59,$D5,$7D,$BB,$7D,$4B,$E1,$30,$DA,$FF,$15,$7C,$86
	.dc.b	$26,$47,$F8,$C6,$75,$87,$FB,$A7,$C7,$5E,$A3,$FC,$5E,$3E,$A1,$F4
	.dc.b	$DF,$8B,$8C,$3F,$8B,$FA,$8F,$E4,$B8,$FF,$93,$6B,$6D,$9E,$11,$8A
	.dc.b	$FF,$7C,$57,$07,$BC,$67,$3F,$F3,$94,$7D,$DB,$FF,$2C,$69,$F7,$1F
	.dc.b	$F2,$E8,$FA,$4F,$E3,$B3,$E8,$9D,$37,$8F,$D1,$18,$BF,$F3,$F5,$F1
	.dc.b	$EC,$CD,$48,$EE,$FE,$A5,$D1,$83,$E5,$7B,$BF,$40,$FD,$66,$CF,$77
	.dc.b	$13,$A2,$65,$F3,$E3,$2F,$D1,$F6,$F8,$FD,$3F,$26,$D7,$F9,$8D,$B3
	.dc.b	$B8,$FF,$33,$4F,$D3,$FA,$67,$4D,$3D,$1B,$FE,$A7,$6F,$97,$3A,$BF
	.dc.b	$F4,$13,$70,$FB,$ED,$9F,$8B,$5C,$6B,$FC,$4F,$4A,$D4,$61,$B1,$FE
	.dc.b	$2C,$6C,$7D,$5E,$CF,$23,$FE,$F4,$E9,$FE,$25,$EF,$3E,$BA,$C6,$1B
	.dc.b	$3F,$EF,$A6,$E1,$FE,$EE,$7F,$80,$F5,$FF,$97,$E7,$F5,$5E,$57,$4F
	.dc.b	$9B,$CB,$23,$27,$FA,$93,$C3,$64,$AD,$63,$FE,$FA,$F8,$5F,$2F,$B3
	.dc.b	$E6,$6E,$1C,$6D,$18,$AE,$13,$9A,$E0,$D3,$84,$EF,$83,$FC,$FD,$6F
	.dc.b	$67,$E1,$F3,$7C,$46,$DF,$76,$E3,$FE,$A2,$CF,$D3,$18,$C3,$9F,$CD
	.dc.b	$7C,$CF,$7F,$E6,$B7,$0F,$43,$70,$E5,$92,$31,$7F,$86,$F8,$9F,$C6
	.dc.b	$1F,$5B,$E0,$6E,$1F,$88,$8B,$87,$29,$B5,$C6,$B4,$7C,$78,$CB,$FE
	.dc.b	$70,$CD,$A3,$E9,$7F,$52,$7E,$13,$E5,$ED,$7F,$B5,$FC,$F7,$C0,$FF
	.dc.b	$11,$6C,$7F,$B1,$EB,$13,$F8,$CF,$98,$93,$FE,$CB,$F6,$2E,$BF,$03
	.dc.b	$E0,$B9,$78,$3F,$92,$6D,$77,$8B,$3B,$8C,$1F,$81,$FA,$0F,$CC,$9A
	.dc.b	$E0,$FB,$6F,$F2,$5D,$FF,$95,$CF,$FB,$BD,$B5,$F6,$79,$7F,$8F,$4F
	.dc.b	$EB,$27,$37,$FC,$81,$B1,$9F,$2B,$FB,$C5,$70,$BF,$95,$1D,$CF,$E2
	.dc.b	$3E,$2D,$AF,$F7,$53,$E3,$F8,$79,$1F,$99,$E5,$B7,$FC,$FB,$15,$63
	.dc.b	$2F,$CA,$D5,$BF,$34,$FE,$E3,$E2,$F1,$83,$0F,$D5,$C7,$F4,$5F,$A0
	.dc.b	$6B,$4F,$EE,$FF,$60,$FA,$FC,$65,$FE,$B2,$D3,$F9,$FD,$F0,$8C,$1F
	.dc.b	$C4,$1C,$65,$FD,$96,$F9,$4E,$A3,$FD,$77,$3A,$AE,$78,$7F,$4A,$9E
	.dc.b	$4D,$76,$D5,$3D,$9F,$F3,$8F,$DC,$7C,$F4,$FD,$1F,$4D,$8B,$EA,$3F
	.dc.b	$B2,$4E,$C9,$E9,$7C,$08,$CB,$33,$F4,$61,$D2,$35,$18,$AF,$07,$CB
	.dc.b	$FF,$76,$D3,$FB,$1B,$67,$F9,$9F,$A9,$FC,$A5,$E7,$19,$C3,$FD,$BA
	.dc.b	$D4,$D9,$E1,$1F,$DF,$A3,$44,$F5,$7F,$BF,$4E,$AF,$C3,$1F,$E0,$19
	.dc.b	$FF,$21,$DF,$83,$D8,$DF,$81,$96,$46,$4F,$7E,$1F,$C0,$1C,$63,$FC
	.dc.b	$6F,$EE,$D1,$FE,$55,$5F,$11,$63,$0D,$7F,$9D,$46,$9A,$68,$FF,$18
	.dc.b	$F9,$FF,$E0,$4D,$3F,$C6,$CD,$8F,$E2,$5E,$D7,$B5,$FF,$1A,$46,$70
	.dc.b	$71,$7F,$5F,$F5,$BF,$C9,$F3,$9E,$8F,$F3,$4F,$9F,$89,$FF,$29,$F5
	.dc.b	$BC,$6A,$7F,$9B,$FE,$13,$FC,$3D,$AF,$F9,$1E,$75,$A7,$F9,$19,$BA
	.dc.b	$7F,$90,$5B,$3F,$12,$3F,$5D,$5C,$AB,$FC,$95,$5A,$6D,$F8,$1F,$EB
	.dc.b	$D3,$BA,$F5,$D6,$DF,$F6,$E6,$8F,$AC,$FF,$40,$36,$FF,$B3,$6D,$FD
	.dc.b	$95,$FD,$77,$B3,$F1,$2B,$64,$8D,$FA,$17,$0F,$F2,$4F,$D1,$BE,$67
	.dc.b	$FD,$AD,$1F,$ED,$C1,$19,$DF,$11,$89,$DB,$FE,$E2,$9D,$3C,$7E,$1B
	.dc.b	$FD,$4B,$F0,$1C,$1B,$19,$F8,$14,$E5,$B1,$E1,$AF,$F2,$FE,$9E,$06
	.dc.b	$C6,$7C,$7F,$F8,$00,$D9,$F4,$6E,$FA,$3F,$2F,$27,$FA,$05,$BB,$FF
	.dc.b	$9F,$5B,$7F,$63,$E5,$BF,$C5,$35,$BD,$ED,$ED,$B5,$F9,$0B,$E3,$CE
	.dc.b	$E5,$63,$AC,$9F,$46,$F6,$3E,$4C,$61,$B1,$9C,$7D,$2D,$7F,$C7,$67
	.dc.b	$E7,$9C,$69,$FE,$9B,$6E,$9F,$E9,$76,$CF,$F3,$FF,$41,$F4,$BE,$39
	.dc.b	$CE,$3E,$1A,$FF,$50,$C7,$0D,$17,$C2,$7F,$D4,$B5,$FA,$76,$F5,$7B
	.dc.b	$63,$11,$FE,$1F,$D5,$F0,$59,$A9,$3C,$0F,$CF,$C7,$CA,$C1,$F2,$FF
	.dc.b	$EE,$A9,$FF,$20,$2F,$A4,$D3,$63,$3C,$D8,$33,$32,$32,$7F,$38,$FA
	.dc.b	$3F,$1B,$F9,$86,$BF,$9D,$6B,$F3,$9A,$FF,$9A,$9D,$F0,$7F,$A7,$9B
	.dc.b	$C1,$9C,$BF,$07,$D0,$FA,$47,$CF,$FE,$1B,$16,$0F,$92,$DF,$AD,$F5
	.dc.b	$B0,$FF,$49,$73,$6E,$B5,$7E,$9B,$FD,$7B,$C9,$B1,$A8,$F5,$46,$0E
	.dc.b	$44,$7F,$B8,$57,$96,$3C,$9F,$48,$F6,$3C,$DF,$02,$F8,$CA,$6A,$26
	.dc.b	$1F,$C0,$2B,$83,$63,$CB,$18,$8F,$E1,$0D,$E0,$FF,$62,$B7,$83,$FD
	.dc.b	$2A,$D9,$1F,$E7,$6F,$86,$F9,$1F,$46,$FA,$4F,$95,$F4,$9F,$E1,$3A
	.dc.b	$FE,$E6,$D7,$FD,$C3,$3B,$7F,$BB,$B5,$FE,$34,$D7,$F5,$6D,$7F,$D4
	.dc.b	$ED,$7C,$6F,$1C,$7F,$DB,$AD,$C7,$77,$E4,$F0,$BE,$F7,$D2,$7F,$78
	.dc.b	$C3,$F3,$3E,$A7,$E6,$0C,$FA,$92,$3F,$94,$CE,$DF,$E5,$CD,$7F,$7E
	.dc.b	$D7,$FD,$76,$EF,$2F,$F7,$7B,$71,$CD,$7E,$9B,$D3,$F1,$F9,$5F,$AD
	.dc.b	$F9,$FF,$F5,$46,$3F,$EA,$B6,$AC,$79,$CF,$08,$F2,$BF,$90,$C7,$FA
	.dc.b	$24,$60,$87,$1C,$7C,$B3,$C6,$3F,$D4,$19,$3E,$27,$F7,$3D,$52,$7F
	.dc.b	$7A,$FE,$97,$E9,$1E,$8D,$3F,$3E,$E2,$7F,$A7,$B5,$FE,$A8,$D7,$FC
	.dc.b	$E9,$E0,$BE,$93,$94,$7F,$8E,$27,$86,$BF,$DD,$7C,$3F,$E1,$F5,$FC
	.dc.b	$33,$E9,$25,$FF,$10,$7E,$47,$5F,$97,$E0,$7C,$64,$9F,$1B,$FD,$8A
	.dc.b	$DC,$5F,$CA,$49,$E0,$7F,$18,$F3,$3F,$81,$7D,$23,$C3,$FE,$41,$EC
	.dc.b	$7D,$9C,$BF,$CE,$8D,$53,$B7,$FC,$A2,$D7,$FB,$A3,$5F,$E8,$AD,$7F
	.dc.b	$E0,$0B,$5F,$D6,$BC,$76,$FF,$0B,$47,$6A,$F2,$9F,$CB,$BC,$6F,$E5
	.dc.b	$FE,$1F,$F3,$0F,$9D,$95,$FC,$D2,$32,$FF,$36,$8C,$BF,$ED,$FE,$3F
	.dc.b	$E2,$98,$7C,$F5,$F3,$BF,$CF,$FE,$3B,$FA,$DC,$FF,$BB,$2B,$F9,$EC
	.dc.b	$C7,$FD,$F0,$6C,$C1,$FE,$E2,$F4,$F9,$7C,$C4,$8F,$F7,$F4,$75,$30
	.dc.b	$FF,$BA,$23,$0E,$8F,$F4,$97,$13,$04,$F0,$C3,$FD,$FF,$7F,$DE,$7C
	.dc.b	$17,$CB,$7D,$8E,$E9,$97,$FA,$6F,$D1,$56,$A3,$0D,$7F,$4C,$7F,$95
	.dc.b	$9C,$61,$FE,$47,$9E,$37,$F1,$7C,$7F,$8D,$F2,$BF,$1F,$CB,$98,$FA
	.dc.b	$35,$4E,$A3,$13,$A7,$BE,$DE,$D9,$FE,$A5,$F1,$2F,$A4,$FE,$A9,$F4
	.dc.b	$5F,$DF,$AF,$FD,$3A,$D7,$CE,$7F,$B3,$A3,$89,$FD,$4A,$79,$47,$F9
	.dc.b	$09,$AF,$F9,$1D,$AF,$F8,$61,$CE,$BF,$90,$6F,$8D,$FD,$02,$7B,$1E
	.dc.b	$2F,$92,$FE,$31,$F2,$9F,$C6,$59,$1F,$C6,$63,$2F,$F1,$B8,$CB,$F0
	.dc.b	$3B,$D3,$FE,$36,$C1,$F4,$8F,$92,$FF,$05,$FC,$7C,$C3,$FD,$A9,$5A
	.dc.b	$9E,$15,$B9,$CB,$5F,$B7,$3B,$7F,$96,$7A,$7F,$DF,$FE,$93,$FC,$63
	.dc.b	$0D,$DF,$2D,$F4,$0D,$78,$B2,$7F,$CD,$75,$FE,$3F,$7F,$C7,$ED,$7E
	.dc.b	$33,$5F,$EC,$EF,$62,$47,$DC,$F9,$3F,$F0,$87,$C9,$BE,$83,$FC,$5D
	.dc.b	$F4,$9F,$E5,$4C,$3F,$42,$D7,$EF,$CE,$DF,$F6,$FB,$5C,$CF,$43,$9F
	.dc.b	$87,$F1,$BE,$93,$FC,$6B,$E9,$4A,$F8,$F1,$96,$90,$FC,$AF,$C8,$7F
	.dc.b	$B2,$1F,$D3,$FD,$27,$F9,$93,$C2,$34,$91,$FE,$7C,$8C,$BC,$08,$7F
	.dc.b	$94,$7C,$8F,$F2,$67,$C1,$FF,$12,$EB,$FF,$06,$F7,$FF,$9D,$67,$26
	.dc.b	$B4,$B7,$FE,$B2,$8A,$FF,$96,$E3,$F8,$17,$CC,$79,$49,$1E,$36,$BF
	.dc.b	$BE,$F1,$7F,$81,$C7,$08,$FF,$5E,$F9,$3F,$E8,$8C,$FF,$D3,$FB,$FF
	.dc.b	$AA,$9B,$3F,$EA,$F6,$BF,$E8,$06,$BF,$A7,$73,$D3,$FE,$CF,$6E,$9F
	.dc.b	$E2,$A6,$CB,$47,$C7,$7F,$30,$AD,$7C,$07,$F7,$5A,$FF,$63,$B5,$FE
	.dc.b	$2F,$7C,$AF,$8C,$7A,$67,$58,$78,$5B,$7D,$31,$FF,$3E,$63,$DE,$FF
	.dc.b	$92,$D3,$E9,$3E,$03,$63,$E0,$35,$F8,$0D,$7E,$03,$9C,$3C,$13,$E3
	.dc.b	$3B,$F9,$7F,$E6,$FF,$8D,$7D,$27,$F8,$77,$07,$D1,$BE,$7F,$F8,$85
	.dc.b	$7E,$76,$31,$3F,$E5,$28,$C4,$EA,$B9,$56,$A9,$5D,$B1,$FF,$2C,$FD
	.dc.b	$26,$63,$F5,$0D,$EA,$FE,$3C,$E7,$0F,$23,$6B,$FC,$C3,$86,$F0,$B8
	.dc.b	$F9,$B0,$FE,$4D,$3B,$9F,$5C,$F9,$6B,$71,$E9,$AD,$CF,$18,$C9,$E8
	.dc.b	$C3,$D7,$1C,$61,$87,$85,$B1,$C5,$A7,$C5,$9F,$4C,$71,$C3,$81,$1C
	.dc.b	$12,$7C,$87,$8B,$1F,$9A,$F3,$C7,$17,$E4,$8D,$9E,$E2,$DF,$C1,$3E
	.dc.b	$16,$1E,$72,$61,$EB,$74,$FA,$63,$C9,$3D,$CE,$87,$C3,$BF,$B3,$3E
	.dc.b	$FA,$E7,$3B,$7B,$C6,$49,$19,$7E,$57,$0F,$CA,$8E,$CD,$7B,$1F,$E0
	.dc.b	$69,$D9,$FE,$FC,$5F,$3F,$FB,$98,$C3,$63,$FE,$00,$B5,$FE,$D0,$D7
	.dc.b	$FD,$52,$FA,$9F,$ED,$9A,$38,$71,$1C,$0F,$B6,$7F,$AC,$23,$4E,$0E
	.dc.b	$65,$73,$5F,$A1,$8C,$B6,$33,$87,$A2,$7E,$68,$8E,$B9,$71,$8F,$79
	.dc.b	$A8,$CC,$F1,$AE,$97,$E8,$AE,$F3,$FE,$07,$7A,$46,$5E,$77,$C1,$F0
	.dc.b	$98,$9E,$17,$DA,$3E,$59,$E1,$AF,$D2,$BF,$02,$B9,$1E,$7C,$33,$5C
	.dc.b	$67,$E5,$4E,$AF,$D0,$76,$BF,$25,$FA,$6F,$E2,$DF,$82,$F8,$36,$F8
	.dc.b	$3E,$8A,$EF,$7E,$48,$F5,$1F,$71,$C4,$75,$7B,$C7,$D7,$6D,$76,$8F
	.dc.b	$3D,$76,$8F,$E3,$F8,$B7,$7C,$5F,$E0,$B4,$FE,$78,$DC,$7F,$B4,$F0
	.dc.b	$FF,$99,$9F,$E8,$F1,$9C,$BF,$9E,$BF,$E8,$C7,$FA,$6B,$5C,$C7,$F3
	.dc.b	$E8,$C6,$1E,$1F,$83,$E4,$D9,$FF,$3A,$DB,$D1,$CE,$4B,$A4,$AD,$CD
	.dc.b	$C7,$FA,$CD,$6E,$7F,$AE,$D7,$9F,$97,$FB,$1D,$EF,$3F,$EB,$53,$E4
	.dc.b	$BE,$18,$7F,$52,$8C,$1E,$BE,$8F,$F3,$FF,$7F,$C1,$6E,$5E,$8F,$3F
	.dc.b	$1A,$AA,$9C,$5C,$4F,$F8,$6E,$30,$7C,$7D,$3F,$BB,$FA,$1E,$7D,$8D
	.dc.b	$E1,$C3,$2F,$E3,$B9,$69,$FE,$95,$FA,$0F,$F4,$AB,$71,$E1,$AF,$F8
	.dc.b	$5E,$72,$D8,$F7,$46,$23,$FC,$ED,$A2,$31,$01,$19,$9F,$C0,$6E,$F9
	.dc.b	$97,$C2,$7E,$1E,$30,$FD,$3F,$0D,$F2,$79,$AD,$E7,$FE,$74,$6E,$5F
	.dc.b	$EC,$DE,$DD,$61,$F6,$FD,$9F,$F0,$65,$F8,$F2,$FF,$6E,$FD,$1B,$E7
	.dc.b	$3F,$2D,$8A,$31,$5F,$EC,$2F,$A4,$FE,$C2,$F1,$3F,$D2,$91,$A9,$FF
	.dc.b	$51,$B6,$FF,$93,$F2,$2F,$FC,$1F,$7C,$6F,$E2,$5E,$E3,$FC,$A3,$3A
	.dc.b	$9F,$F5,$6E,$FF,$82,$6F,$2F,$EA,$E9,$F4,$7F,$C7,$A7,$31,$D5,$FC
	.dc.b	$79,$FE,$47,$9F,$D7,$38,$3F,$C9,$0E,$2B,$B6,$5F,$CF,$AB,$AF,$BB
	.dc.b	$F1,$27,$31,$F6,$CC,$9F,$47,$DE,$FA,$87,$D7,$F7,$EC,$64,$FF,$75
	.dc.b	$EF,$FE,$8F,$71,$97,$FB,$4A,$30,$FF,$A0,$3D,$EF,$C0,$7D,$27,$E0
	.dc.b	$5B,$5F,$EF,$4E,$6F,$F7,$7B,$67,$4D,$3E,$17,$CE,$FF,$B8,$9B,$DB
	.dc.b	$F2,$78,$3F,$84,$46,$09,$8F,$F6,$FA,$E0,$DA,$F9,$31,$88,$FE,$9C
	.dc.b	$EB,$7F,$91,$76,$FF,$DB,$13,$6B,$FB,$74,$DE,$37,$C0,$7F,$93,$EF
	.dc.b	$FD,$F1,$F3,$BF,$C4,$63,$17,$FE,$F6,$51,$9E,$ED,$36,$B8,$9F,$BB
	.dc.b	$F2,$FF,$88,$77,$1E,$9F,$39,$86,$4F,$E4,$F2,$7C,$2F,$EB,$A3,$91
	.dc.b	$97,$19,$F8,$FE,$23,$CD,$EC,$79,$DB,$5C,$F5,$46,$2B,$84,$E6,$B8
	.dc.b	$34,$E1,$3B,$E0,$FD,$4D,$BD,$8F,$89,$C9,$F1,$5B,$7D,$DB,$8F,$C7
	.dc.b	$6D,$F2,$F6,$BE,$67,$C0,$7C,$D6,$DF,$9D,$B7,$F3,$DB,$5F,$E2,$08
	.dc.b	$72,$3E,$33,$F8,$3B,$6F,$E0,$7C,$1E,$26,$47,$F8,$E6,$B8,$C7,$E2
	.dc.b	$89,$19,$AE,$52,$BF,$E1,$3B,$BE,$13,$9C,$E6,$FF,$DA,$F3,$97,$F4
	.dc.b	$9F,$49,$FC,$9C,$FC,$57,$CC,$58,$FE,$09,$E9,$3E,$13,$F8,$1B,$63
	.dc.b	$CC,$DC,$3F,$C9,$B3,$FD,$3F,$89,$F3,$3F,$95,$3F,$9A,$7B,$77,$0F
	.dc.b	$C0,$7C,$27,$F2,$AC,$5E,$1F,$E6,$E3,$0C,$9E,$EF,$88,$8F,$F9,$7F
	.dc.b	$A7,$F4,$BF,$49,$FA,$33,$86,$9F,$A3,$7F,$9B,$C7,$E5,$38,$7E,$B3
	.dc.b	$6B,$FD,$F7,$CF,$FC,$B6,$FF,$DF,$BE,$7F,$03,$F9,$1D,$7E,$72,$3F
	.dc.b	$C9,$0D,$7F,$AC,$1F,$EE,$58,$D9,$FD,$4F,$C9,$E5,$7F,$C4,$DF,$1F
	.dc.b	$24,$8F,$D1,$73,$5B,$FB,$07,$F6,$3C,$3E,$C3,$EF,$AF,$19,$2B,$F1
	.dc.b	$13,$AC,$BF,$47,$F1,$BC,$DB,$3F,$CF,$FE,$91,$AF,$F5,$A6,$C7,$F4
	.dc.b	$FF,$49,$F4,$8F,$8A,$D5,$CA,$FC,$B5,$F0,$D3,$FD,$20,$D7,$FC,$D8
	.dc.b	$D7,$E2,$BB,$C7,$FC,$89,$F4,$CE,$5E,$95,$F0,$25,$39,$BF,$EA,$DF
	.dc.b	$49,$F5,$68,$FD,$7D,$3C,$FF,$4B,$FD,$D2,$34,$7F,$6F,$9F,$F3,$F3
	.dc.b	$5F,$AE,$D7,$FC,$20,$D7,$AF,$A9,$FE,$53,$F7,$DF,$71,$FC,$97,$1F
	.dc.b	$1C,$E6,$31,$5F,$60,$7C,$CF,$F8,$3B,$4F,$E0,$8D,$BF,$C3,$FB,$5F
	.dc.b	$CD,$9F,$EE,$51,$9A,$FF,$11,$3F,$E1,$67,$FD,$2C,$D7,$F1,$ED,$77
	.dc.b	$EC,$FF,$78,$8E,$5E,$F3,$E8,$9E,$77,$F8,$62,$73,$18,$C3,$C3,$EF
	.dc.b	$E4,$FF,$0E,$E4,$F3,$31,$3A,$3F,$8B,$5F,$F9,$73,$3F,$2D,$F8,$4F
	.dc.b	$F2,$87,$91,$FE,$C7,$AF,$E8,$9F,$4D,$FE,$67,$8C,$57,$E7,$BB,$3F
	.dc.b	$A5,$CE,$6B,$4D,$78,$9F,$95,$F2,$BF,$C3,$5C,$1F,$EC,$EE,$6F,$E5
	.dc.b	$FC,$9F,$E5,$DF,$4B,$FC,$D5,$5A,$BF,$F2,$65,$69,$E5,$5B,$69,$FA
	.dc.b	$1F,$01,$E5,$7F,$9C,$DB,$1F,$D8,$7E,$2F,$FC,$7B,$AF,$ED,$B0,$FE
	.dc.b	$7D,$7B,$25,$6F,$E1,$6D,$6F,$0E,$6F,$FB,$46,$F8,$D7,$FB,$E5,$B1
	.dc.b	$99,$DB,$FC,$3B,$DC,$FF,$32,$FD,$37,$FA,$7A,$8F,$F9,$F7,$9F,$0C
	.dc.b	$C3,$FC,$95,$5C,$1F,$F5,$1C,$E6,$FF,$31,$F4,$CF,$96,$FF,$77,$47
	.dc.b	$FB,$94,$46,$76,$FE,$AC,$D8,$D3,$5F,$F7,$67,$C5,$7F,$B0,$BA,$FF
	.dc.b	$AE,$36,$3A,$35,$DF,$C4,$7F,$B2,$5E,$7E,$5B,$4E,$ED,$8C,$C6,$0F
	.dc.b	$91,$F4,$DE,$88,$FC,$67,$C4,$7F,$83,$1F,$F6,$0F,$B3,$FC,$DB,$A3
	.dc.b	$F1,$15,$B6,$EB,$9E,$0F,$F6,$AD,$5B,$75,$31,$E9,$1F,$81,$C7,$95
	.dc.b	$FE,$3B,$E9,$9E,$DF,$1C,$1D,$3F,$CA,$F0,$FC,$57,$99,$D7,$9D,$87
	.dc.b	$09,$FE,$64,$D7,$F4,$D5,$C5,$B5,$C6,$75,$B7,$E8,$DB,$97,$F4,$9B
	.dc.b	$E1,$3F,$E8,$E6,$C7,$9F,$2F,$16,$3F,$A0,$C7,$FD,$CF,$1E,$9D,$39
	.dc.b	$C7,$FA,$D2,$74,$FE,$85,$B8,$7F,$A6,$A7,$79,$FC,$7B,$EB,$D5,$F0
	.dc.b	$FE,$0D,$C7,$7E,$9A,$67,$FE,$80,$8D,$18,$7F,$D0,$11,$87,$FD,$31
	.dc.b	$F4,$AF,$86,$FA,$DD,$1F,$E0,$AF,$8B,$BE,$BF,$FB,$7A,$F7,$5F,$6D
	.dc.b	$5F,$17,$FD,$5D,$7C,$5B,$5C,$5F,$E7,$93,$B6,$FC,$97,$6D,$64,$79
	.dc.b	$23,$B4,$9C,$C7,$F7,$CB,$FF,$0F,$D6,$A6,$4F,$82,$37,$C1,$FE,$4B
	.dc.b	$BE,$53,$FE,$B7,$FA,$57,$B5,$EB,$E9,$F3,$5F,$F9,$45,$BC,$BF,$E1
	.dc.b	$4A,$DD,$F1,$BE,$1B,$48,$C1,$0F,$EB,$57,$C9,$FF,$7A,$35,$57,$47
	.dc.b	$FB,$07,$D2,$3E,$93,$E7,$7D,$2F,$E3,$A7,$FD,$F8,$D5,$3B,$7F,$DA
	.dc.b	$4D,$7F,$B5,$46,$CE,$11,$9C,$FF,$CD,$CE,$E7,$FC,$62,$74,$AF,$F3
	.dc.b	$53,$4F,$F3,$64,$6D,$FF,$1B,$FD,$2F,$F7,$67,$FD,$9F,$E8,$FE,$54
	.dc.b	$FC,$34,$66,$3F,$BE,$36,$FF,$DE,$FC,$CA,$E1,$5F,$EE,$BA,$D6,$DD
	.dc.b	$BC,$EE,$64,$9F,$F6,$EF,$9D,$DB,$C6,$FE,$0D,$F3,$DF,$DD,$A7,$FD
	.dc.b	$5C,$D7,$C3,$1B,$73,$E4,$70,$30,$E8,$F1,$46,$34,$63,$FE,$4B,$CF
	.dc.b	$FD,$7D,$5F,$EE,$EF,$A5,$FD,$A1,$11,$FE,$1B,$FA,$5F,$F4,$D3,$8B
	.dc.b	$FF,$4F,$46,$23,$FD,$E9,$3C,$5E,$1E,$67,$EF,$FE,$97,$F6,$15,$BB
	.dc.b	$E3,$39,$AF,$F7,$FF,$D2,$FF,$05,$23,$F0,$B9,$B8,$FC,$07,$C6,$64
	.dc.b	$78,$F4,$FF,$38,$E2,$AF,$97,$D7,$2B,$E5,$BF,$C4,$7E,$97,$F8,$17
	.dc.b	$D2,$FF,$A4,$27,$FD,$B1,$C9,$FC,$DF,$D6,$FB,$31,$F0,$A7,$4F,$F8
	.dc.b	$F5,$BB,$7F,$AC,$2B,$81,$D3,$C1,$95,$D0,$87,$17,$F9,$17,$81,$FC
	.dc.b	$8F,$E7,$A4,$FF,$BB,$7A,$BF,$7E,$FF,$89,$5B,$87,$E4,$27,$73,$9A
	.dc.b	$FE,$59,$F3,$DF,$D3,$FE,$2B,$F4,$77,$FE,$E4,$AC,$9D,$67,$4C,$30
	.dc.b	$7F,$BC,$FD,$0E,$5F,$2F,$23,$04,$3F,$C7,$16,$8D,$46,$27,$C4,$62
	.dc.b	$FF,$DE,$B3,$89,$FF,$47,$7A,$4F,$97,$FA,$1D,$92,$73,$1F,$D9,$67
	.dc.b	$51,$F7,$CF,$EE,$7C,$3F,$E1,$77,$1E,$6A,$31,$1F,$E3,$59,$E2,$FA
	.dc.b	$3E,$5B,$E3,$78,$F2,$38,$EA,$9F,$94,$66,$3D,$F3,$87,$C4,$7F,$BB
	.dc.b	$FE,$97,$FA,$67,$D2,$7F,$B8,$23,$2D,$7F,$D0,$9F,$16,$9F,$EA,$4D
	.dc.b	$8D,$FC,$7F,$23,$33,$23,$24,$39,$BF,$C2,$3E,$53,$F8,$4F,$C9,$C9
	.dc.b	$FE,$13,$19,$7F,$85,$C6,$5F,$F3,$87,$8E,$9F,$F0,$F7,$1B,$E9,$7F
	.dc.b	$E0,$05,$7E,$4F,$DA,$FE,$F0,$6A,$33,$1F,$EB,$76,$CF,$F2,$49,$DC
	.dc.b	$FD,$78,$E9,$F2,$7F,$DC,$7E,$97,$FC,$17,$2D,$9F,$2D,$F2,$8D,$78
	.dc.b	$52,$7E,$5C,$65,$FA,$35,$C8,$E7,$5C,$A3,$66,$AB,$FC,$01,$F2,$6F
	.dc.b	$A2,$FF,$1A,$7D,$2B,$C5,$A6,$A2,$5F,$E3,$27,$4D,$9C,$BF,$07,$E9
	.dc.b	$7F,$C3,$9F,$35,$23,$E0,$70,$27,$E4,$19,$8F,$DE,$4E,$1F,$D1,$FD
	.dc.b	$2F,$F8,$BF,$E6,$64,$60,$87,$F9,$71,$CB,$F9,$D8,$FF,$27,$7C,$1F
	.dc.b	$F0,$6F,$37,$FC,$81,$E3,$60,$CF,$8F,$6F,$FD,$35,$15,$FF,$26,$C7
	.dc.b	$EB,$FE,$5B,$FC,$81,$F1,$F2,$7F,$6F,$19,$7F,$5F,$F0,$9F,$EC,$56
	.dc.b	$CE,$57,$47,$F5,$07,$FC,$F5,$F0,$1F,$D4,$1F,$F3,$AF,$47,$F9,$34
	.dc.b	$FF,$37,$78,$89,$D5,$71,$8C,$70,$6E,$FC,$37,$C2,$F7,$AF,$F5,$8B
	.dc.b	$DD,$68,$87,$92,$35,$3C,$23,$CB,$7F,$E9,$F6,$9F,$EA,$1A,$E6,$FF
	.dc.b	$83,$9A,$FF,$AC,$E5,$F4,$5F,$E6,$7E,$D9,$3F,$EB,$09,$CB,$67,$2F
	.dc.b	$FA,$5A,$7D,$51,$FE,$55,$F8,$A7,$D2,$FC,$AA,$7D,$2B,$E8,$3F,$BA
	.dc.b	$C2,$BF,$38,$FF,$78,$9D,$3F,$DE,$A3,$2F,$29,$E3,$1D,$B6,$E1,$18
	.dc.b	$BF,$05,$F2,$BE,$DD,$7E,$08,$C6,$1D,$1B,$5C,$2F,$96,$1D,$30,$F1
	.dc.b	$61,$C6,$7C,$2D,$8F,$35,$71,$6D,$71,$7C,$25,$C3,$C1,$7D,$A3,$D8
	.dc.b	$70,$29,$C0,$B1,$C1,$D6,$1E,$8A,$E9,$87,$8E,$78,$4E,$5E,$38,$79
	.dc.b	$5C,$E1,$EA,$7C,$91,$9C,$3E,$64,$78,$2F,$A4,$78,$63,$EC,$4E,$CE
	.dc.b	$13,$A3,$6E,$70,$F2,$CC,$3C,$6F,$80,$C6,$1F,$52,$B8,$C6,$67,$C4
	.dc.b	$6B,$0F,$91,$5E,$DD,$1F,$21,$F2,$DB,$5F,$2D,$AF,$A2,$FF,$5D,$3F
	.dc.b	$E7,$08,$7C,$F6,$1F,$F3,$9B,$5F,$CB,$B5,$FF,$26,$B5,$E2,$7D,$81
	.dc.b	$E8,$FF,$09,$C3,$F2,$51,$C0,$D4,$67,$1F,$15,$7E,$1E,$78,$F0,$D7
	.dc.b	$AD,$F2,$BE,$11,$E6,$39,$CE,$EB,$94,$EA,$7D,$26,$8F,$31,$D2,$FF
	.dc.b	$D2,$31,$B8,$F3,$5F,$7B,$E9,$1C,$63,$D6,$F1,$3F,$77,$3C,$5C,$1F
	.dc.b	$E7,$27,$17,$E4,$AE,$75,$D6,$BC,$73,$96,$9D,$EF,$E2,$B4,$F8,$13
	.dc.b	$C4,$F1,$CF,$16,$C6,$CF,$82,$EE,$79,$B6,$39,$C7,$C2,$AF,$2B,$63
	.dc.b	$F6,$B9,$7F,$3D,$F8,$AF,$E8,$99,$FF,$45,$F3,$DC,$CF,$23,$39,$7F
	.dc.b	$36,$BB,$8A,$F7,$58,$C7,$A5,$FF,$54,$8C,$72,$30,$E3,$7C,$2B,$F9
	.dc.b	$A7,$C0,$FF,$55,$C7,$F9,$D3,$5F,$E8,$AD,$7F,$AC,$B7,$1F,$EA,$CF
	.dc.b	$4C,$9F,$17,$E0,$8C,$36,$FF,$B3,$B6,$FF,$A9,$1E,$BA,$E3,$3E,$43
	.dc.b	$65,$8D,$E1,$F1,$AB,$E1,$FC,$63,$17,$8B,$E3,$F3,$73,$18,$C7,$FB
	.dc.b	$BC,$62,$75,$1F,$E7,$26,$E3,$DB,$27,$9F,$ED,$3D,$DC,$36,$FF,$C0
	.dc.b	$31,$6B,$84,$DE,$1F,$F0,$06,$6F,$A0,$F3,$61,$FE,$08,$F8,$AF,$4C
	.dc.b	$63,$0F,$04,$63,$5F,$98,$DA,$E0,$7F,$92,$FE,$5B,$FD,$E0,$FF,$A9
	.dc.b	$FB,$3F,$0B,$92,$73,$87,$FA,$A6,$73,$18,$9F,$F7,$CC,$63,$E1,$7F
	.dc.b	$0B,$ED,$FF,$7A,$C3,$FC,$FB,$F4,$8F,$9A,$FC,$4F,$D2,$7F,$74,$9C
	.dc.b	$C6,$3C,$0F,$CA,$E8,$BE,$15,$FE,$94,$BE,$13,$F8,$03,$57,$C6,$7F
	.dc.b	$D1,$7B,$7D,$79,$FC,$2C,$3E,$8F,$FD,$5C,$EB,$97,$F4,$6E,$7D,$EF
	.dc.b	$0D,$F1,$DF,$FC,$61,$39,$38,$1F,$EA,$93,$18,$78,$E7,$84,$7F,$C0
	.dc.b	$03,$33,$FA,$58,$CC,$A7,$21,$FF,$1F,$F0,$3E,$47,$E5,$F8,$7F,$41
	.dc.b	$B5,$F6,$5B,$5F,$A2,$C3,$EC,$46,$1F,$E0,$9E,$47,$FB,$16,$BF,$D7
	.dc.b	$AB,$E9,$3F,$76,$E6,$FF,$4A,$E6,$77,$EF,$EA,$B6,$D8,$FF,$5C,$FC
	.dc.b	$CF,$FA,$FD,$F8,$9B,$DC,$3E,$B4,$DC,$7E,$5C,$DA,$FF,$18,$CD,$AF
	.dc.b	$89,$35,$46,$67,$ED,$7B,$DF,$86,$8C,$56,$DB,$1F,$C5,$BD,$FF,$F1
	.dc.b	$F5,$7F,$B5,$3E,$6B,$1D,$B9,$6C,$64,$FF,$6C,$C6,$5F,$F7,$37,$6F
	.dc.b	$EC,$7C,$DF,$E4,$3F,$02,$F9,$8F,$D8,$C6,$1B,$5F,$EE,$A9,$D3,$CB
	.dc.b	$C0,$56,$E3,$FD,$CF,$F1,$5E,$7F,$73,$D0,$D8,$C9,$FE,$EC,$8C,$B4
	.dc.b	$CF,$07,$F1,$CD,$53,$A8,$C3,$F0,$FD,$8F,$89,$CB,$6F,$FD,$EE,$DA
	.dc.b	$E4,$DE,$3F,$E9,$98,$23,$E6,$7A,$7F,$37,$D3,$B8,$7F,$C0,$06,$E1
	.dc.b	$E3,$6E,$1C,$9F,$B1,$F1,$9F,$2B,$E1,$2D,$FC,$76,$DF,$F3,$6E,$04
	.dc.b	$63,$0F,$EB,$D1,$8A,$FC,$8D,$F3,$69,$CC,$FC,$FF,$D2,$7E,$78,$FC
	.dc.b	$27,$99,$63,$F6,$17,$FD,$67,$D3,$7C,$17,$E9,$3D,$8F,$E2,$71,$8A
	.dc.b	$FD,$4C,$7E,$37,$CF,$91,$83,$99,$FB,$BF,$41,$C4,$9F,$CA,$79,$DD
	.dc.b	$3E,$22,$DF,$0E,$1F,$F1,$DC,$66,$31,$87,$2F,$A3,$C5,$7D,$F7,$F3
	.dc.b	$C4,$9E,$51,$FE,$4A,$AC,$3F,$C9,$EB,$F2,$6D,$EB,$FE,$F7,$CF,$FC
	.dc.b	$BA,$31,$7C,$AB,$7B,$FF,$0A,$F2,$7F,$9A,$C6,$27,$DD,$D1,$FC,$72
	.dc.b	$FF,$C1,$98,$FF,$0C,$BD,$B6,$BA,$73,$72,$3F,$6D,$CB,$63,$27,$C0
	.dc.b	$E0,$7C,$66,$71,$FC,$F3,$74,$FC,$AB,$6B,$E8,$57,$07,$E1,$FD,$1B
	.dc.b	$83,$FC,$25,$82,$31,$3F,$ED,$76,$C7,$F9,$CD,$AF,$FB,$18,$E1,$3F
	.dc.b	$D9,$BE,$8D,$C9,$FD,$62,$B8,$E5,$FE,$34,$9C,$E1,$E2,$BE,$17,$B6
	.dc.b	$C7,$D8,$FA,$2F,$A5,$7F,$EF,$FF,$71,$F4,$5F,$C2,$2B,$5E,$87,$F8
	.dc.b	$EA,$73,$E1,$7F,$68,$C3,$8D,$7F,$97,$EF,$86,$BF,$D9,$E7,$2F,$FA
	.dc.b	$12,$33,$8F,$F8,$E6,$73,$87,$4B,$EE,$7E,$55,$C5,$FF,$71,$F8,$FF
	.dc.b	$F0,$14,$FF,$59,$F7,$1F,$44,$F9,$6F,$EF,$75,$AB,$56,$A7,$3E,$CF
	.dc.b	$F9,$52,$31,$87,$E6,$AF,$84,$62,$FC,$15,$BC,$3B,$5E,$E3,$18,$78
	.dc.b	$3D,$DD,$78,$B6,$BF,$5D,$8F,$9A,$31,$8F,$F9,$2A,$31,$3E,$1A,$E2
	.dc.b	$FF,$3B,$3F,$A0,$FD,$17,$F4,$26,$99,$3F,$A0,$39,$69,$FD,$2B,$CF
	.dc.b	$FF,$55,$DF,$F9,$17,$8F,$FC,$A3,$18,$E0,$FF,$3D,$7B,$DF,$8A,$BF
	.dc.b	$F0,$27,$9B,$FD,$8B,$87,$FC,$CD,$87,$06,$CF,$F5,$6E,$BF,$F3,$0E
	.dc.b	$5F,$E5,$EF,$86,$7C,$AF,$F9,$7A,$73,$9F,$FA,$1E,$73,$7C,$7C,$9F
	.dc.b	$F3,$EC,$63,$0F,$F3,$6E,$6C,$38,$DF,$F1,$DA,$D6,$1F,$E5,$69,$CD
	.dc.b	$FF,$A3,$67,$35,$EA,$9C,$F8,$69,$CD,$7F,$9C,$70,$7D,$17,$FA,$0E
	.dc.b	$FF,$D2,$91,$8C,$3F,$91,$ED,$E3,$7F,$94,$7B,$49,$D6,$1F,$E7,$1B
	.dc.b	$E3,$87,$3B,$FE,$4B,$F4,$4F,$94,$FF,$72,$DF,$FB,$C6,$31,$C1,$FE
	.dc.b	$F3,$84,$61,$D4,$F0,$7F,$82,$7C,$47,$FA,$97,$9B,$83,$63,$FD,$F8
	.dc.b	$D7,$7F,$0B,$FA,$27,$C1,$E3,$FF,$A4,$5B,$B7,$87,$1F,$F0,$73,$67
	.dc.b	$53,$FE,$47,$3F,$D3,$1F,$9A,$BD,$5C,$72,$7F,$93,$67,$31,$8D,$BF
	.dc.b	$C9,$D8,$2F,$85,$7C,$CF,$23,$4F,$03,$E6,$C6,$9F,$F5,$7C,$F1,$BD
	.dc.b	$F2,$FF,$B4,$2F,$8C,$F5,$F3,$2E,$06,$9C,$3F,$0D,$FF,$64,$67,$E7
	.dc.b	$AE,$19,$78,$2B,$6F,$F6,$2A,$DE,$7E,$AC,$3A,$4E,$C8,$7F,$AF,$A7
	.dc.b	$6F,$E1,$7E,$29,$F6,$1E,$EE,$74,$62,$7F,$C7,$ED,$79,$77,$6E,$31
	.dc.b	$A7,$FA,$D1,$B9,$7E,$D9,$B5,$FE,$ED,$C5,$7F,$1B,$D6,$CF,$B2,$FD
	.dc.b	$2D,$EC,$7D,$08,$C3,$BD,$66,$3F,$E0,$59,$CB,$6B,$8B,$FE,$4F,$9F
	.dc.b	$B2,$FF,$93,$BC,$DB,$E6,$7F,$56,$8F,$F0,$2B,$5F,$E4,$1F,$9B,$7F
	.dc.b	$96,$DC,$CF,$02,$3F,$CA,$23,$4F,$FA,$C6,$7F,$DC,$EF,$4F,$4F,$B9
	.dc.b	$E2,$F6,$9E,$BE,$27,$44,$3F,$D7,$51,$A7,$1C,$5F,$99,$7F,$CC,$F5
	.dc.b	$C5,$DD,$7B,$BE,$91,$F1,$34,$93,$FC,$FB,$9B,$FD,$E2,$FF,$89,$27
	.dc.b	$6D,$9D,$BF,$D0,$67,$6F,$FB,$36,$76,$D9,$DE,$1F,$AD,$F0,$3F,$55
	.dc.b	$F1,$9E,$3F,$02,$5F,$77,$F5,$73,$FA,$FF,$7B,$33,$D6,$6D,$F3,$CB
	.dc.b	$FC,$F3,$3C,$48,$6C,$F6,$57,$DE,$FA,$17,$D1,$7F,$17,$94,$61,$FE
	.dc.b	$34,$D7,$FC,$34,$6A,$38,$46,$6F,$E8,$37,$AF,$F8,$73,$6B,$FD,$AD
	.dc.b	$F4,$7F,$C4,$AB,$77,$E3,$24,$FF,$94,$A7,$57,$FC,$5F,$C6,$FF,$6E
	.dc.b	$4F,$F9,$5B,$E0,$5F,$27,$F9,$86,$C7,$F9,$6A,$76,$FF,$BF,$DB,$8F
	.dc.b	$E1,$5B,$7D,$7E,$27,$97,$CB,$FE,$8A,$73,$7F,$EE,$D6,$BF,$D0,$67
	.dc.b	$6F,$F9,$A3,$E1,$BF,$03,$87,$E4,$5B,$87,$F2,$F6,$CF,$FB,$DF,$F4
	.dc.b	$7F,$E4,$95,$F9,$76,$D7,$F3,$56,$CE,$FC,$9F,$DA,$DB,$EA,$E9,$FB
	.dc.b	$8A,$DD,$1B,$3B,$8F,$F6,$67,$B7,$FC,$2C,$8F,$FB,$3F,$5F,$E0,$3F
	.dc.b	$85,$FF,$10,$8F,$F4,$1B,$56,$7F,$AD,$8C,$3F,$EA,$CF,$86,$FE,$2C
	.dc.b	$49,$FF,$6D,$3F,$C6,$78,$BF,$86,$FD,$5B,$E6,$A9,$2B,$FD,$23,$F8
	.dc.b	$5F,$E0,$10,$9E,$0F,$FA,$59,$AF,$93,$83,$FA,$E4,$63,$1F,$E0,$55
	.dc.b	$CE,$B9,$5F,$77,$81,$DA,$B8,$5F,$F6,$5A,$F1,$7A,$0F,$25,$F0,$9F
	.dc.b	$CC,$FE,$3E,$63,$FE,$F7,$8F,$E6,$C6,$1F,$F6,$94,$61,$FF,$7E,$7C
	.dc.b	$87,$F3,$BC,$E4,$60,$87,$FB,$65,$CB,$FE,$DB,$8F,$E8,$5E,$C7,$F4
	.dc.b	$3F,$9B,$93,$FE,$BE,$C7,$FC,$42,$D9,$CB,$FE,$E3,$9E,$2E,$4E,$0E
	.dc.b	$B6,$FF,$72,$CF,$F5,$DE,$FF,$D0,$FA,$5F,$DF,$FA,$B0,$7F,$53,$F3
	.dc.b	$EE,$3F,$ED,$19,$D3,$6B,$F5,$3B,$FC,$9A,$FE,$B5,$E6,$7F,$5B,$F2
	.dc.b	$65,$FA,$63,$2F,$AE,$32,$FF,$BE,$E3,$FA,$AF,$D6,$BE,$92,$92,$7F
	.dc.b	$DB,$3F,$35,$F9,$4E,$6F,$F0,$F5,$6E,$7F,$DB,$AD,$7F,$02,$1C,$6F
	.dc.b	$FD,$26,$B9,$4F,$E4,$A7,$35,$D2,$FC,$13,$CA,$B7,$9F,$E1,$FE,$7B
	.dc.b	$FC,$3B,$F0,$BF,$8A,$3F,$17,$C6,$8F,$EF,$86,$BD,$99,$87,$DB,$8F
	.dc.b	$EE,$7F,$58,$F0,$69,$27,$FD,$FB,$F4,$0F,$AC,$C4,$7B,$2B,$83,$FE
	.dc.b	$FC,$6B,$FA,$0C,$D5,$C6,$3F,$E0,$0F,$E1,$6F,$53,$FC,$06,$D8,$CC
	.dc.b	$7F,$8E,$3E,$E9,$F1,$DA,$89,$3A,$3F,$73,$C9,$FE,$4B,$C5,$7C,$27
	.dc.b	$47,$D2,$C3,$B3,$6F,$B6,$1F,$E2,$FF,$A2,$7C,$CD,$25,$76,$E5,$FE
	.dc.b	$0F,$93,$8F,$FC,$6B,$3F,$95,$BE,$2D,$BE,$2F,$C3,$9D,$BF,$E9,$89
	.dc.b	$FE,$63,$D6,$69,$FC,$D2,$FB,$7D,$37,$97,$CD,$97,$DD,$FC,$0C,$7A
	.dc.b	$7E,$62,$4F,$8D,$FF,$2D,$69,$88,$F0,$45,$E0,$F9,$1F,$03,$F1,$E7
	.dc.b	$FC,$A9,$5C,$1F,$F2,$E3,$FE,$63,$F4,$31,$1A,$32,$7F,$98,$5F,$95
	.dc.b	$F3,$32,$3E,$23,$FE,$6B,$F8,$6C,$47,$C1,$8B,$87,$8A,$7F,$CE,$BF
	.dc.b	$07,$FE,$89,$AF,$F3,$DB,$67,$FC,$E4,$DE,$2F,$A7,$93,$E6,$78,$DC
	.dc.b	$3F,$D6,$35,$B7,$FC,$C1,$19,$7A,$96,$FE,$1C,$5B,$FF,$49,$D7,$1E
	.dc.b	$CF,$EA,$D7,$FE,$79,$AF,$3F,$B9,$23,$BB,$FE,$89,$F8,$2C,$47,$AA
	.dc.b	$2E,$3F,$42,$7F,$CC,$7F,$13,$FE,$46,$8F,$EA,$7C,$9F,$8A,$B4,$BE
	.dc.b	$77,$FD,$59,$39,$8C,$4F,$0E,$8F,$F5,$84,$73,$8F,$C6,$46,$3E,$03
	.dc.b	$FB,$6C,$60,$FE,$C3,$3C,$9C,$1E,$0B,$EF,$5C,$70,$ED,$3C,$2F,$9E
	.dc.b	$1E,$3B,$DC,$F4,$6B,$EA,$9D,$DF,$38,$EF,$87,$86,$F9,$47,$63,$86
	.dc.b	$16,$7D,$93,$A9,$E3,$87,$69,$CD,$F5,$AD,$11,$C1,$2B,$CC,$75,$AE
	.dc.b	$58,$79,$A3,$DD,$86,$58,$7A,$23,$4F,$98,$F1,$9A,$BF,$AD,$1A,$C3
	.dc.b	$CB,$3E,$03,$67,$3A,$F7,$4E,$EF,$D8,$79,$23,$27,$5C,$3E,$FC,$66
	.dc.b	$B9,$0D,$3F,$87,$E1,$F8,$36,$E1,$F8,$26,$E1,$FE,$F0,$F8,$7F,$F3
	.dc.b	$8F,$CD,$DC,$3F,$64,$D7,$FD,$F8,$D7,$FC,$52,$D4,$FD,$95,$F1,$9F
	.dc.b	$EF,$FD,$3F,$39,$1C,$0F,$C7,$9A,$E1,$E0,$F2,$33,$3C,$75,$F2,$C6
	.dc.b	$27,$77,$DE,$B5,$3E,$83,$47,$63,$A5,$F8,$A3,$71,$EC,$37,$1E,$67
	.dc.b	$77,$FC,$4E,$77,$1E,$99,$E5,$7E,$28,$E3,$5E,$69,$CC,$78,$6F,$CD
	.dc.b	$5B,$9C,$9D,$E3,$33,$B3,$99,$EB,$8D,$D7,$53,$C9,$5C,$DB,$3C,$67
	.dc.b	$6F,$18,$CC,$77,$8E,$A7,$C1,$76,$D8,$D9,$FC,$8B,$E5,$55,$FC,$B1
	.dc.b	$B3,$FD,$FF,$8D,$ED,$E3,$4F,$E6,$F5,$6F,$91,$C4,$FE,$D9,$6E,$C6
	.dc.b	$1C,$A7,$53,$FD,$0B,$BF,$93,$F9,$DE,$6F,$F0,$54,$F2,$BE,$3E,$DE
	.dc.b	$B8,$55,$AB,$E5,$7F,$D1,$30,$E7,$87,$08,$C1,$27,$67,$B7,$0E,$97
	.dc.b	$CF,$3F,$EA,$78,$60,$E7,$92,$B1,$87,$F8,$17,$E4,$FD,$3E,$07,$F7
	.dc.b	$F7,$7E,$DE,$38,$51,$49,$5E,$DC,$15,$8D,$3E,$87,$B7,$E0,$3C,$3E
	.dc.b	$CD,$BE,$11,$67,$EB,$4D,$D7,$EC,$E6,$FB,$0F,$F1,$7E,$1F,$E6,$F6
	.dc.b	$DF,$68,$C7,$83,$EA,$9C,$DF,$26,$CF,$F7,$4C,$75,$5B,$EF,$93,$F5
	.dc.b	$FC,$57,$C4,$E3,$2F,$F5,$3E,$7E,$3B,$C6,$7F,$E3,$DB,$DE,$1E,$59
	.dc.b	$F1,$73,$C8,$E9,$F5,$37,$1E,$7A,$FE,$2D,$B5,$F8,$0C,$15,$BC,$7F
	.dc.b	$37,$4F,$B5,$7B,$5F,$90,$8C,$3F,$88,$FB,$8F,$AF,$9F,$E6,$A7,$2D
	.dc.b	$AE,$2D,$AE,$1C,$9C,$E7,$35,$FE,$86,$6E,$1E,$26,$DF,$F9,$43,$0E
	.dc.b	$F5,$BC,$3C,$9F,$96,$BC,$8A,$7E,$73,$8D,$F4,$38,$C7,$CD,$5B,$6D
	.dc.b	$6E,$75,$87,$F9,$5F,$E0,$BE,$37,$AD,$F1,$DB,$7F,$65,$B3,$B8,$FB
	.dc.b	$93,$F1,$D7,$C5,$D3,$F7,$2B,$E8,$BF,$D2,$0F,$C9,$36,$58,$E0,$FF
	.dc.b	$06,$38,$7A,$D8,$BF,$F5,$6C,$70,$6C,$F6,$AE,$B1,$89,$FC,$87,$95
	.dc.b	$F8,$1F,$27,$F0,$4D,$8C,$9F,$95,$8C,$B4,$FD,$27,$BB,$FA,$13,$F1
	.dc.b	$58,$58,$C9,$2F,$FD,$01,$19,$7F,$D0,$5E,$BF,$E4,$7E,$77,$F2,$4D
	.dc.b	$AF,$D2,$36,$77,$1F,$A7,$9D,$B4,$D9,$63,$FD,$1D,$D1,$F9,$DF,$B0
	.dc.b	$FC,$F3,$67,$F6,$0D,$9D,$C7,$EC,$E7,$F4,$9F,$61,$FA,$5F,$9D,$B3
	.dc.b	$FB,$76,$D7,$FA,$6D,$BD,$3F,$B5,$7F,$4D,$F7,$2F,$AB,$A7,$D3,$F3
	.dc.b	$6E,$1F,$A7,$9D,$36,$74,$D2,$61,$FE,$EA,$9D,$AC,$3F,$A8,$CF,$E2
	.dc.b	$EF,$AC,$FF,$AB,$FE,$3B,$F7,$C7,$EF,$FE,$1A,$D7,$ED,$FD,$37,$AB
	.dc.b	$FB,$D6,$C7,$F1,$5A,$C4,$FF,$07,$F8,$F9,$3F,$C6,$7E,$B2,$E3,$E3
	.dc.b	$83,$FE,$20,$FC,$65,$F4,$7D,$7E,$13,$FA,$CF,$B2,$9C,$DF,$8E,$74
	.dc.b	$D3,$FB,$2C,$EE,$B3,$7D,$CE,$DC,$3D,$30,$F8,$BC,$1F,$C6,$B6,$65
	.dc.b	$9A,$FE,$35,$F7,$1F,$ED,$9C,$FF,$8C,$37,$1E,$9E,$3B,$EE,$7F,$94
	.dc.b	$65,$FC,$92,$B4,$DE,$8E,$67,$F7,$4A,$E1,$E5,$E3,$1E,$B3,$FD,$83
	.dc.b	$EE,$7F,$97,$C6,$0F,$F0,$0F,$9E,$8C,$E7,$ED,$1F,$55,$FC,$83,$2F
	.dc.b	$E6,$AD,$BF,$9F,$6F,$8B,$7C,$27,$F3,$7A,$D6,$1F,$51,$AF,$4F,$3F
	.dc.b	$FB,$9D,$70,$C3,$F9,$D5,$68,$FE,$5F,$5F,$ED,$36,$BF,$78,$E5,$1F
	.dc.b	$ED,$77,$B5,$7F,$4A,$75,$EA,$3E,$2F,$FA,$53,$6B,$FA,$47,$DC,$BC
	.dc.b	$0C,$D4,$9F,$E9,$95,$BB,$F8,$97,$B9,$CD,$6E,$7E,$45,$6F,$5F,$EB
	.dc.b	$77,$B9,$CD,$7F,$5B,$FC,$F7,$F2,$F8,$7F,$59,$6C,$F6,$9D,$32,$74
	.dc.b	$FF,$63,$AE,$D1,$F0,$EB,$8D,$FF,$6A,$6C,$66,$B8,$5F,$37,$EF,$4B
	.dc.b	$E6,$FC,$87,$F6,$FF,$E0,$0F,$CC,$DF,$96,$32,$D3,$FB,$84,$6A,$B9
	.dc.b	$CE,$9B,$3A,$E8,$C6,$3E,$5F,$A3,$F8,$47,$F7,$DF,$E0,$1F,$C1,$5F
	.dc.b	$92,$33,$7F,$15,$D7,$16,$27,$F6,$5F,$06,$E6,$F8,$70,$FA,$47,$E9
	.dc.b	$CF,$4F,$FC,$21,$C3,$FE,$5A,$9C,$C6,$33,$FF,$31,$C6,$1F,$F1,$34
	.dc.b	$EE,$B9,$47,$EA,$3F,$3C,$7C,$7F,$F9,$31,$B1,$FE,$47,$F9,$BC,$4C
	.dc.b	$8E,$0F,$F9,$5A,$31,$97,$F0,$BF,$89,$7E,$7C,$FB,$4F,$F3,$53,$70
	.dc.b	$FF,$29,$7C,$E6,$26,$47,$F9,$EA,$B8,$C7,$F9,$8A,$73,$7D,$6F,$93
	.dc.b	$67,$57,$C6,$73,$C9,$FE,$8C,$C3,$C7,$F2,$1E,$23,$FC,$DB,$F3,$B6
	.dc.b	$F8,$45,$C7,$FD,$25,$36,$FF,$C7,$7F,$55,$FE,$25,$F3,$3F,$6D,$48
	.dc.b	$C3,$FB,$96,$BF,$BD,$35,$F0,$1F,$ED,$A7,$F7,$DF,$25,$FE,$1C,$FA
	.dc.b	$5E,$6D,$8C,$FB,$69,$CB,$63,$B7,$C4,$FF,$BD,$5B,$B7,$93,$14,$E6
	.dc.b	$7C,$51,$89,$FE,$1C,$3E,$7B,$F0,$33,$F4,$5A,$FD,$B6,$E9,$FE,$B6
	.dc.b	$EB,$7D,$63,$EA,$FF,$8D,$4F,$CF,$AD,$F1,$7F,$5A,$C7,$B7,$3E,$A3
	.dc.b	$13,$AC,$BF,$D7,$18,$F3,$7C,$5D,$FF,$3B,$0F,$33,$68,$D2,$4F,$02
	.dc.b	$1F,$EC,$BB,$F0,$2A,$E6,$EB,$1F,$45,$6B,$3F,$F6,$4D,$F0,$F2,$DE
	.dc.b	$97,$F7,$EC,$3F,$A7,$B5,$FF,$47,$B7,$0F,$8D,$F2,$1F,$EC,$89,$DF
	.dc.b	$B1,$CE,$B5,$87,$F5,$06,$CF,$FB,$66,$B8,$3F,$C1,$3E,$7B,$3E,$0B
	.dc.b	$EA,$DF,$07,$EC,$8C,$3E,$99,$E7,$A9,$E3,$5C,$2F,$F8,$2B,$5E,$D9
	.dc.b	$FC,$67,$FD,$57,$F4,$1F,$9E,$69,$FA,$0F,$91,$46,$5B,$19,$8E,$51
	.dc.b	$88,$FE,$E4,$DD,$BF,$DF,$FD,$2F,$AB,$FC,$BD,$7F,$A8,$69,$F4,$5F
	.dc.b	$62,$9F,$AE,$71,$E2,$FF,$12,$C6,$32,$EF,$53,$0F,$33,$E5,$23,$83
	.dc.b	$B3,$B7,$FC,$0D,$F3,$2F,$AC,$F8,$5F,$58,$F8,$CF,$17,$EB,$AF,$AE
	.dc.b	$A6,$A2,$61,$FC,$12,$B8,$36,$3E,$54,$62,$3F,$BF,$37,$83,$FD,$8E
	.dc.b	$DD,$7F,$C0,$8D,$9F,$E0,$5D,$8F,$1F,$CB,$E1,$FC,$5F,$D7,$BF,$D7
	.dc.b	$15,$FE,$48,$C7,$E6,$C6,$23,$FC,$45,$87,$FA,$A9,$FF,$67,$35,$F1
	.dc.b	$D7,$98,$8F,$AD,$EE,$F0,$7E,$A6,$1F,$DB,$E3,$FD,$B1,$F5,$EF,$97
	.dc.b	$FC,$57,$D7,$FF,$8A,$A7,$FD,$B8,$D7,$FC,$69,$3B,$78,$CF,$E3,$E3
	.dc.b	$11,$FE,$23,$6B,$FC,$AD,$AF,$F8,$A5,$DE,$5F,$EF,$06,$E9,$9A,$F1
	.dc.b	$F8,$5F,$9D,$F2,$5F,$19,$F9,$FF,$A0,$7C,$65,$35,$13,$3F,$E7,$F5
	.dc.b	$C1,$B1,$FA,$58,$C4,$7F,$8E,$5B,$9B,$2F,$F1,$F4,$61,$B5,$E1,$F8
	.dc.b	$DF,$D8,$7C,$B7,$DD,$C3,$ED,$57,$F9,$16,$31,$1F,$E5,$B6,$E1,$D6
	.dc.b	$3F,$D0,$FF,$5D,$FC,$AD,$3F,$CE,$9F,$EA,$4D,$7F,$CA,$0D,$78,$F9
	.dc.b	$7F,$BF,$A7,$D7,$FF,$03,$AE,$87,$58,$CD,$79,$87,$8B,$FC,$3F,$E2
	.dc.b	$7F,$89,$12,$35,$1F,$C4,$F7,$FE,$1D,$F5,$FF,$C7,$A7,$FB,$33,$5F
	.dc.b	$F3,$23,$5F,$F5,$1B,$67,$F8,$E7,$D1,$7F,$1F,$24,$F6,$F2,$3F,$02
	.dc.b	$FF,$AA,$BE,$6D,$EB,$7F,$1D,$FA,$F7,$CD,$53,$F9,$04,$EA,$BD,$81
	.dc.b	$C3,$FE,$75,$6B,$FD,$F5,$AF,$F9,$ED,$AF,$86,$BA,$91,$EB,$5F,$DD
	.dc.b	$78,$7F,$B6,$60,$FA,$3F,$E7,$5E,$EC,$BF,$64,$7F,$3D,$30,$FF,$45
	.dc.b	$8C,$3F,$07,$DD,$FE,$85,$8C,$8C,$10,$FE,$98,$E5,$FE,$9D,$1F,$D2
	.dc.b	$7E,$BF,$FA,$57,$D0,$C9,$F9,$91,$97,$E7,$46,$5F,$EA,$D1,$97,$FA
	.dc.b	$C4,$67,$2F,$6C,$FF,$6B,$F0,$3F,$45,$5F,$88,$FD,$A3,$FC,$4D,$5F
	.dc.b	$E0,$D6,$F4,$FF,$AD,$2B,$7E,$BF,$F8,$5F,$B3,$91,$30,$FC,$07,$A9
	.dc.b	$FD,$87,$CE,$FE,$C5,$E7,$49,$FC,$24,$65,$FC,$34,$65,$FE,$DD,$1F
	.dc.b	$D8,$7E,$47,$FD,$5D,$1F,$82,$F8,$5F,$A8,$DE,$3D,$7C,$2C,$8F,$F2
	.dc.b	$1D,$70,$7F,$D7,$4D,$7F,$C8,$4D,$7F,$D8,$0E,$F7,$FD,$D4,$7F,$B4
	.dc.b	$67,$A3,$FE,$D9,$BF,$F0,$1C,$FE,$1B,$E2,$7F,$C2,$DF,$0F,$F8,$9E
	.dc.b	$05,$FE,$CA,$30,$FE,$D3,$9F,$FD,$F2,$FE,$96,$B8,$F8,$F2,$30,$7E
	.dc.b	$83,$EC,$1E,$17,$F8,$8F,$EC,$3F,$A9,$78,$07,$FA,$A5,$2B,$6F,$1A
	.dc.b	$FF,$05,$46,$23,$FD,$E4,$DD,$3F,$C9,$55,$FB,$29,$FD,$7F,$D7,$BD
	.dc.b	$6F,$F1,$97,$D8,$7C,$0D,$FD,$B9,$A3,$17,$F0,$2F,$83,$CA,$FF,$C3
	.dc.b	$11,$88,$FF,$76,$35,$FF,$3A,$35,$FF,$79,$3B,$C7,$E9,$B7,$0C,$D7
	.dc.b	$93,$E4,$BF,$C8,$9F,$54,$FA,$CF,$F2,$47,$6F,$F9,$F7,$2F,$F5,$DB
	.dc.b	$72,$FE,$6D,$D8,$BE,$15,$FC,$A6,$31,$7D,$09,$1F,$CA,$39,$3F,$CA
	.dc.b	$2E,$EB,$FC,$D1,$B3,$1F,$F5,$74,$FF,$2E,$6A,$C7,$C3,$5F,$E6,$FC
	.dc.b	$7C,$5E,$67,$A8,$95,$DF,$E3,$FD,$BF,$61,$FE,$9E,$AF,$43,$6C,$D2
	.dc.b	$57,$95,$E2,$FF,$40,$F8,$46,$DF,$E6,$D6,$C6,$63,$06,$7D,$1B,$19
	.dc.b	$21,$FE,$E6,$9F,$F0,$E6,$7E,$CF,$B0,$F2,$12,$7F,$D6,$03,$2F,$A1
	.dc.b	$19,$8F,$89,$C5,$FA,$89,$CC,$7C,$D9,$F3,$7C,$5F,$FA,$23,$A7,$FA
	.dc.b	$94,$FF,$A2,$7E,$BB,$FC,$C9,$18,$23,$FC,$62,$34,$71,$71,$C4,$CD
	.dc.b	$F6,$19,$A8,$7F,$B8,$6B,$73,$A8,$C1,$A3,$FD,$DD,$1A,$9E,$DE,$EF
	.dc.b	$FA,$9E,$3F,$C0,$F3,$FC,$10,$F4,$46,$48,$F7,$78,$BF,$5F,$E1,$BF
	.dc.b	$D0,$5F,$1D,$FE,$61,$DF,$F8,$6F,$1F,$E6,$29,$2F,$A9,$ED,$39,$8C
	.dc.b	$4F,$0E,$9F,$14,$73,$8F,$CF,$9F,$C6,$1C,$7C,$5F,$1B,$F4,$36,$7F
	.dc.b	$8E,$BD,$7E,$0B,$F3,$91,$FE,$B4,$BE,$B1,$8C,$38,$4F,$63,$CF,$7E
	.dc.b	$6A,$DE,$3E,$59,$F1,$5F,$19,$F3,$D6,$E7,$CC,$D7,$E0,$CE,$DE,$35
	.dc.b	$E6,$AE,$31,$4E,$25,$38,$C5,$C7,$E2,$61,$E8,$25,$79,$AB,$D5,$3A
	.dc.b	$8E,$B1,$8C,$3C,$93,$98,$F1,$99,$3D,$C7,$C2,$2B,$B7,$C6,$7C,$B3
	.dc.b	$CE,$53,$81,$5F,$AF,$3E,$5A,$F3,$CF,$07,$EB,$5F,$6C,$3E,$BE,$3F
	.dc.b	$50,$97,$ED,$77,$7D,$71,$FC,$0C,$FB,$A7,$26,$B8,$3F,$91,$D7,$A1
	.dc.b	$B9,$7E,$35,$B8,$7E,$19,$DC,$FF,$66,$FB,$0F,$EC,$F1,$88,$46,$2F
	.dc.b	$F8,$4B,$63,$F5,$46,$84,$F1,$8F,$B9,$0F,$91,$FF,$3A,$57,$E9,$E3
	.dc.b	$C0,$7C,$B3,$5C,$DC,$1E,$91,$B9,$E3,$97,$8A,$78,$46,$1F,$1D,$FD
	.dc.b	$C7,$99,$F9,$F3,$BE,$BC,$F0,$D4,$76,$BF,$5D,$F3,$8F,$E0,$D7,$EA
	.dc.b	$8C,$47,$29,$DD,$72,$AF,$05,$7F,$A4,$6F,$DD,$7F,$15,$F0,$3E,$33
	.dc.b	$17,$C9,$C5,$72,$BE,$91,$E9,$BE,$95,$F0,$70,$E6,$76,$BF,$35,$FC
	.dc.b	$97,$95,$F4,$39,$47,$4C,$3C,$67,$18,$C9,$BA,$EB,$7D,$CF,$33,$B9
	.dc.b	$E3,$1F,$4D,$B3,$FD,$0F,$E9,$BF,$D6,$5B,$7F,$AC,$F3,$70,$A7,$39
	.dc.b	$7F,$86,$B8,$DE,$DE,$31,$E5,$3F,$D0,$6B,$FC,$87,$1F,$96,$BE,$53
	.dc.b	$FD,$73,$0E,$B5,$AA,$E2,$7E,$5A,$3C,$3E,$07,$56,$D7,$F7,$3F,$6F
	.dc.b	$18,$7F,$8B,$E3,$FA,$85,$60,$F2,$73,$FF,$57,$AC,$4E,$89,$87,$F7
	.dc.b	$38,$CB,$FD,$52,$BC,$98,$78,$23,$C9,$87,$76,$CF,$A7,$AB,$FC,$49
	.dc.b	$F3,$BF,$0D,$B8,$FF,$82,$7D,$BC,$70,$2A,$9C,$5C,$4F,$F8,$02,$B8
	.dc.b	$E1,$FB,$86,$BF,$D8,$F4,$FE,$E5,$F0,$3E,$EE,$03,$7C,$FC,$33,$0F
	.dc.b	$F0,$D3,$96,$9F,$E3,$9F,$B0,$FF,$0E,$B7,$C1,$E1,$8F,$F8,$38,$FF
	.dc.b	$09,$4E,$9C,$64,$8C,$47,$F7,$78,$C4,$FC,$59,$CC,$FF,$85,$DA,$F4
	.dc.b	$EF,$F9,$10,$7E,$CF,$8C,$F8,$DC,$C4,$C7,$FC,$8D,$F0,$37,$0F,$F2
	.dc.b	$BB,$CB,$9E,$4F,$F9,$47,$EC,$2D,$7A,$32,$FC,$CC,$FF,$93,$FE,$BB
	.dc.b	$3D,$5F,$EB,$97,$F3,$94,$FA,$37,$B9,$86,$DF,$FA,$62,$CF,$F9,$FB
	.dc.b	$EC,$3E,$DD,$FC,$93,$95,$F1,$3F,$40,$E2,$7F,$C7,$ED,$3F,$C8,$37
	.dc.b	$D9,$B3,$F9,$CA,$7C,$6F,$3B,$F1,$1E,$17,$D3,$F6,$0F,$2E,$9F,$A4
	.dc.b	$E5,$69,$8B,$E3,$97,$FA,$CA,$F8,$46,$3A,$3F,$1B,$F2,$DB,$AF,$C2
	.dc.b	$46,$27,$FC,$FD,$F0,$BF,$63,$D9,$FB,$2D,$8C,$9F,$E8,$38,$CB,$4F
	.dc.b	$0D,$7C,$AF,$8D,$FA,$30,$7F,$87,$2F,$B0,$FF,$5E,$BC,$0D,$16,$3E
	.dc.b	$6B,$89,$FF,$35,$B5,$FE,$C6,$D7,$FC,$9C,$EF,$8B,$ED,$C7,$DC,$F1
	.dc.b	$5E,$E7,$92,$FC,$71,$8B,$FF,$4F,$79,$BF,$98,$F3,$3F,$32,$D8,$C9
	.dc.b	$FE,$A3,$8C,$B4,$FD,$97,$C0,$7E,$88,$FD,$0D,$D8,$C9,$FE,$A5,$F4
	.dc.b	$31,$97,$FA,$D9,$FD,$3F,$C4,$7E,$A3,$EC,$2D,$7F,$AC,$5B,$5F,$63
	.dc.b	$C1,$46,$36,$F7,$F1,$27,$37,$FE,$B9,$C3,$A3,$4E,$87,$FA,$F6,$3F
	.dc.b	$6F,$F6,$1F,$B8,$6C,$64,$FF,$61,$46,$5A,$7F,$B1,$7E,$C3,$F7,$FF
	.dc.b	$4F,$FC,$01,$B1,$93,$FD,$95,$19,$69,$F8,$8F,$B0,$7D,$6F,$EE,$1B
	.dc.b	$1F,$9C,$3F,$11,$C7,$99,$91,$FB,$09,$DE,$DF,$EB,$06,$BF,$E2,$86
	.dc.b	$BF,$E9,$37,$7A,$FE,$1A,$7F,$29,$EE,$74,$AF,$E5,$11,$83,$FD,$C5
	.dc.b	$F2,$3F,$C7,$0F,$E3,$BF,$13,$63,$F4,$3E,$B3,$DA,$FE,$0A,$D8,$FE
	.dc.b	$0A,$DB,$FF,$62,$B5,$FE,$89,$C0,$F9,$2F,$E5,$27,$F2,$AF,$B0,$B7
	.dc.b	$FB,$EF,$6D,$F5,$EF,$B0,$FE,$3F,$93,$E2,$7F,$BE,$56,$A7,$31,$2F
	.dc.b	$C6,$F9,$88,$FF,$B6,$27,$AF,$17,$2C,$3E,$3C,$FF,$A9,$2B,$13,$FC
	.dc.b	$BB,$EB,$FF,$21,$3F,$C0,$9B,$AF,$F5,$1F,$31,$F6,$1F,$CE,$23,$F9
	.dc.b	$64,$EB,$93,$F6,$F0,$8E,$2F,$F9,$81,$AF,$FB,$2D,$A8,$E1,$C1,$FC
	.dc.b	$D2,$3A,$61,$FE,$ED,$FB,$07,$A1,$E7,$AF,$EF,$D1,$89,$FE,$33,$F6
	.dc.b	$1F,$D0,$31,$F1,$B6,$3F,$DE,$DF,$1E,$F8,$8F,$E9,$B3,$9C,$FC,$4F
	.dc.b	$F4,$87,$FC,$EE,$D7,$FD,$C4,$D7,$7E,$EF,$F4,$A8,$E5,$83,$ED,$9F
	.dc.b	$23,$DA,$BF,$C6,$D3,$98,$FE,$BB,$ED,$FF,$19,$C3,$FA,$B3,$67,$E5
	.dc.b	$FD,$8B,$C1,$DF,$99,$36,$FE,$AF,$7C,$70,$FF,$46,$B5,$FF,$7C,$35
	.dc.b	$F9,$CF,$58,$FA,$1F,$44,$FD,$5B,$B4,$7F,$42,$9D,$52,$B5,$87,$F7
	.dc.b	$4F,$B2,$FB,$17,$FD,$21,$B5,$FD,$BF,$2E,$2A,$33,$9F,$E1,$9D,$C7
	.dc.b	$FA,$91,$AF,$E0,$CE,$07,$F7,$67,$2F,$E1,$DE,$75,$EA,$CB,$BC,$62
	.dc.b	$7D,$11,$8F,$B1,$F4,$1F,$E0,$CF,$B2,$FD,$75,$FA,$63,$2F,$F8,$0A
	.dc.b	$D1,$E1,$24,$65,$FF,$38,$66,$FC,$85,$E0,$72,$CB,$FC,$71,$39,$8F
	.dc.b	$91,$EB,$4C,$3F,$B7,$FD,$9F,$E4,$DF,$95,$19,$BF,$F6,$39,$F2,$5F
	.dc.b	$84,$48,$CF,$9D,$FD,$1F,$8B,$FD,$05,$F5,$8F,$B3,$7B,$1F,$E3,$A9
	.dc.b	$CC,$62,$FF,$C1,$9D,$BF,$7A,$F8,$5E,$E7,$FC,$27,$18,$7F,$C2,$BF
	.dc.b	$0C,$FB,$37,$9D,$FE,$6F,$6C,$FF,$85,$FE,$25,$F4,$59,$A9,$3C,$AB
	.dc.b	$7B,$FF,$86,$AF,$73,$9A,$FF,$48,$7D,$93,$EC,$E9,$FE,$6F,$FA,$3B
	.dc.b	$5F,$E2,$8F,$92,$56,$E3,$0D,$C7,$B5,$70,$7F,$CD,$95,$C2,$FF,$D1
	.dc.b	$F5,$B9,$FC,$94,$E6,$FF,$D5,$13,$99,$FF,$80,$2B,$DB,$F3,$9F,$E8
	.dc.b	$3F,$A4,$B5,$E2,$8B,$5E,$F9,$BD,$DF,$CA,$3E,$CF,$FC,$8D,$E8,$FF
	.dc.b	$94,$9B,$5F,$82,$6B,$FA,$50,$E1,$FD,$47,$CC,$7F,$8D,$FB,$3F,$CB
	.dc.b	$6D,$8F,$C5,$B5,$E4,$D7,$F6,$13,$FE,$81,$D3,$C5,$F1,$E6,$92,$7F
	.dc.b	$D8,$DF,$49,$F8,$27,$F0,$1F,$4B,$FB,$9C,$FF,$C3,$7C,$68,$C6,$FF
	.dc.b	$CF,$F0,$61,$C7,$0F,$A6,$DC,$38,$DF,$09,$76,$31,$E9,$1F,$81,$BE
	.dc.b	$18,$F5,$5D,$BC,$6F,$87,$C7,$5C,$3F,$D1,$78,$71,$AE,$D9,$F9,$1D
	.dc.b	$54,$9D,$10,$FC,$C3,$97,$D3,$1E,$7E,$9F,$E0,$7E,$D7,$7F,$84,$95
	.dc.b	$FA,$AF,$0B,$FC,$1C,$D9,$CB,$FA,$39,$E5,$3A,$8F,$F6,$17,$6B,$EC
	.dc.b	$55,$F1,$63,$0D,$8C,$CF,$FA,$27,$E9,$79,$71,$7F,$9B,$2F,$73,$FB
	.dc.b	$EA,$DB,$D9,$B3,$FE,$54,$AE,$3C,$3F,$CD,$7E,$53,$FA,$ED,$FF,$A6
	.dc.b	$39,$DF,$60,$EA,$FB,$11,$87,$ED,$4F,$FB,$2F,$CF,$B7,$CA,$3F,$93
	.dc.b	$DF,$17,$A3,$5F,$F5,$D4,$6F,$2F,$F0,$7A,$FA,$7E,$F5,$9C,$BF,$AE
	.dc.b	$B6,$FF,$C1,$D5,$B7,$FD,$3E,$D7,$C3,$87,$23,$CB,$72,$3C,$71,$97
	.dc.b	$8C,$7B,$3E,$B5,$F1,$7F,$03,$EB,$FF,$D8,$B5,$F1,$5A,$FF,$B4,$E7
	.dc.b	$6F,$FA,$C1,$AF,$F4,$88,$DE,$7F,$EA,$2D,$51,$8D,$3F,$B3,$5F,$0A
	.dc.b	$FF,$18,$B6,$3F,$BE,$CE,$EF,$FB,$D3,$54,$ED,$FF,$19,$3F,$EB,$F7
	.dc.b	$EE,$7D,$7F,$FB,$65,$BF,$29,$EF,$F8,$E9,$18,$21,$ED,$72,$FF,$62
	.dc.b	$9E,$4D,$9C,$E3,$FE,$48,$FA,$E7,$CA,$FE,$1B,$EB,$FF,$C6,$D7,$FE
	.dc.b	$D4,$6B,$FE,$FB,$9D,$BF,$ED,$86,$BF,$EF,$F8,$DD,$7F,$95,$1B,$A7
	.dc.b	$F8,$36,$B7,$3E,$1B,$FF,$80,$36,$9E,$6F,$F2,$F6,$BC,$63,$78,$3E
	.dc.b	$2B,$F3,$BF,$43,$FE,$81,$AF,$F7,$5B,$56,$1F,$E6,$6A,$E1,$1F,$EF
	.dc.b	$06,$BA,$C7,$FC,$45,$5C,$6B,$FC,$DA,$D8,$FF,$13,$57,$09,$E9,$7F
	.dc.b	$AE,$8C,$4E,$A3,$CB,$F1,$9F,$B2,$FA,$CF,$EB,$18,$FF,$9C,$70,$FF
	.dc.b	$68,$FD,$43,$E9,$3F,$94,$36,$75,$8F,$F8,$D6,$F8,$C7,$9C,$FF,$3D
	.dc.b	$7D,$7F,$F3,$27,$FC,$F9,$1A,$C3,$FD,$0A,$DA,$FF,$1F,$56,$E8,$D2
	.dc.b	$7A,$3E,$7F,$A0,$FE,$75,$DA,$FA,$0F,$2E,$33,$0F,$E3,$D1,$FC,$3B
	.dc.b	$EB,$DF,$4D,$C3,$8A,$57,$A8,$FF,$48,$4E,$9F,$E4,$DF,$4B,$C6,$A4
	.dc.b	$60,$E0,$3E,$7B,$F8,$A6,$FF,$EA,$0F,$AF,$7C,$E7,$F1,$E6,$D7,$0A
	.dc.b	$DF,$0F,$F9,$75,$B1,$FE,$A1,$9D,$BF,$2C,$38,$BF,$F5,$6C,$F0,$9F
	.dc.b	$17,$89,$DF,$09,$1D,$E3,$E3,$FD,$2F,$C8,$F7,$24,$FF,$43,$8C,$BF
	.dc.b	$D1,$63,$2F,$DB,$F4,$5C,$4A,$FA,$3E,$DF,$D2,$C2,$5F,$F4,$D8,$CB
	.dc.b	$FD,$3E,$32,$F3,$A4,$FD,$AF,$53,$ED,$FA,$32,$BF,$AB,$46,$5F,$EB
	.dc.b	$11,$97,$94,$65,$E9,$19,$D7,$FA,$B4,$FF,$71,$F5,$3F,$43,$5F,$E0
	.dc.b	$8E,$AF,$F1,$45,$7E,$7A,$35,$DB,$1F,$C2,$6A,$AD,$F1,$7F,$B5,$63
	.dc.b	$13,$B3,$D9,$E4,$FE,$53,$C9,$91,$FD,$BE,$3F,$2E,$61,$F8,$51,$87
	.dc.b	$F1,$7F,$1C,$F8,$1F,$EC,$BE,$FF,$CF,$9F,$C7,$B5,$FC,$24,$ED,$E3
	.dc.b	$3C,$38,$E4,$60,$87,$E2,$EB,$96,$BF,$EF,$66,$E7,$FE,$AF,$C3,$A5
	.dc.b	$7D,$7F,$95,$FF,$14,$7C,$0F,$F9,$06,$3F,$D0,$FE,$C7,$EC,$A6,$46
	.dc.b	$08,$71,$3F,$4F,$F5,$EF,$07,$FC,$65,$F5,$FF,$D3,$F8,$4F,$F5,$1A
	.dc.b	$56,$DE,$35,$C3,$D1,$91,$83,$F6,$7F,$5C,$FA,$AF,$F1,$F7,$D7,$FE
	.dc.b	$EE,$7F,$34,$DE,$17,$07,$FB,$36,$F8,$57,$2B,$E3,$E1,$48,$C1,$0E
	.dc.b	$6E,$5F,$DB,$4F,$27,$F9,$17,$B8,$FA,$BF,$E5,$11,$FA,$BE,$97,$C1
	.dc.b	$FF,$2C,$3F,$C8,$D3,$AC,$FC,$B4,$BE,$5E,$6E,$E3,$F2,$B8,$72,$7F
	.dc.b	$D6,$77,$C7,$44,$71,$8F,$E6,$5C,$5F,$9B,$F8,$6F,$E7,$67,$83,$CD
	.dc.b	$95,$D2,$32,$F4,$8F,$2F,$D7,$FF,$99,$FB,$A4,$F8,$23,$2F,$FA,$FE
	.dc.b	$32,$FC,$2C,$13,$FE,$6F,$F9,$2D,$CE,$63,$0F,$FB,$1E,$33,$B7,$C2
	.dc.b	$DB,$E5,$56,$AB,$A7,$D7,$FF,$97,$F8,$7F,$B1,$43,$DF,$46,$2F,$E9
	.dc.b	$57,$CE,$38,$7D,$0C,$BF,$93,$1F,$E8,$0E,$4F,$DF,$C7,$FA,$3B,$E1
	.dc.b	$7F,$C8,$99,$FF,$A6,$BC,$3D,$4C,$3F,$DB,$CF,$FA,$2C,$C3,$FE,$CC
	.dc.b	$AE,$7C,$5F,$E9,$79,$D4,$FF,$B5,$E7,$BF,$77,$E5,$E3,$FD,$C5,$3A
	.dc.b	$79,$1E,$5F,$89,$FF,$4C,$77,$C8,$E5,$19,$7F,$DE,$11,$97,$C5,$1F
	.dc.b	$E7,$0F,$89,$FF,$2A,$E5,$F8,$6E,$3F,$CA,$C3,$E9,$FF,$13,$7F,$90
	.dc.b	$3F,$17,$19,$78,$75,$7F,$B3,$BB,$BF,$C3,$71,$B1,$3D,$9C,$3C,$5B
	.dc.b	$5D,$AF,$C3,$5C,$9B,$87,$6A,$F0,$D6,$F0,$FF,$6E,$B6,$3C,$B7,$C6
	.dc.b	$7C,$95,$C6,$7C,$75,$B6,$99,$C3,$CD,$87,$82,$3C,$A6,$67,$84,$EA
	.dc.b	$FA,$CE,$89,$5C,$49,$1E,$68,$E4,$77,$C3,$CC,$FA,$30,$CB,$CA,$7A
	.dc.b	$D7,$28,$F8,$26,$9C,$3F,$3A,$B9,$BE,$C3,$91,$D4,$FA,$2E,$AF,$E9
	.dc.b	$4F,$E0,$30,$F6,$CE,$4D,$3F,$00,$CE,$1E,$93,$FD,$F1,$F0,$DF,$0B
	.dc.b	$53,$49,$8F,$E2,$87,$D7,$70,$70,$D7,$F1,$ED,$7F,$8C,$B5,$E2,$7E
	.dc.b	$34,$FE,$53,$D9,$F7,$5D,$99,$EE,$FE,$53,$9F,$F3,$08,$CB,$63,$3A
	.dc.b	$F9,$67,$AC,$E6,$BA,$46,$27,$D5,$5C,$23,$D2,$77,$9D,$DF,$8E,$75
	.dc.b	$3E,$C3,$47,$A4,$E8,$7E,$B9,$DB,$E6,$AE,$F5,$D2,$38,$DF,$AD,$E2
	.dc.b	$7F,$3E,$AE,$2E,$0F,$8A,$77,$BF,$25,$73,$AF,$74,$E4,$EB,$7F,$0E
	.dc.b	$B7,$19,$9D,$9E,$18,$CB,$63,$89,$F1,$1D,$CF,$38,$C9,$F0,$E3,$2E
	.dc.b	$E3,$DE,$FF,$0F,$C7,$FC,$EF,$F4,$BF,$D8,$B3,$FE,$C7,$C7,$FE,$AD
	.dc.b	$9C,$65,$FD,$79,$DB,$FE,$38,$F2,$91,$AC,$2F,$23,$1C,$61,$CF,$0F
	.dc.b	$E6,$51,$8C,$38,$C7,$F3,$5F,$07,$1B,$F3,$CB,$FD,$17,$18,$9D,$1E
	.dc.b	$19,$D6,$3F,$E1,$38,$C1,$AC,$BF,$CD,$98,$7E,$A2,$31,$9F,$BA,$31
	.dc.b	$3F,$D2,$70,$EC,$DA,$E6,$FF,$6D,$F9,$8F,$77,$67,$FA,$DF,$E6,$FE
	.dc.b	$07,$67,$F8,$46,$73,$C0,$9C,$F0,$A7,$31,$86,$E3,$FE,$62,$EA,$FF
	.dc.b	$79,$BF,$EA,$7C,$1F,$5C,$F0,$3E,$5D,$6B,$4F,$F1,$0C,$E6,$F9,$7C
	.dc.b	$07,$92,$31,$3F,$E8,$DA,$E1,$C1,$FD,$FE,$72,$DC,$3E,$A6,$7C,$5B
	.dc.b	$8F,$8E,$B6,$FF,$9D,$78,$3D,$CE,$67,$78,$F9,$3E,$3B,$F0,$3E,$33
	.dc.b	$DE,$F2,$46,$1B,$7F,$E2,$E6,$D7,$F8,$E5,$F5,$7B,$DC,$BE,$B2,$D7
	.dc.b	$F8,$D1,$B5,$FE,$3F,$6E,$BF,$EE,$C7,$F2,$90,$FA,$27,$AD,$F9,$4F
	.dc.b	$AE,$7D,$76,$A7,$38,$FF,$A5,$27,$2F,$FB,$4A,$33,$7F,$EB,$79,$CB
	.dc.b	$FE,$C9,$8C,$E3,$FA,$A9,$CB,$FE,$4E,$9F,$16,$1E,$1C,$3C,$0D,$C3
	.dc.b	$FD,$BF,$F5,$CF,$86,$FD,$4F,$2B,$E4,$FD,$73,$9A,$FF,$54,$CE,$5E
	.dc.b	$35,$C2,$B6,$D8,$FF,$57,$FA,$BF,$EE,$B6,$FA,$FF,$B6,$8C,$67,$FE
	.dc.b	$D3,$51,$9F,$81,$D3,$6B,$89,$FE,$F6,$1F,$17,$FC,$0D,$FE,$05,$ED
	.dc.b	$FD,$27,$27,$F9,$A3,$7F,$F3,$8B,$A3,$94,$F1,$F1,$7F,$D6,$D1,$83
	.dc.b	$81,$FE,$70,$38,$65,$FE,$F9,$7A,$AA,$EB,$3C,$A3,$55,$D6,$FE,$07
	.dc.b	$93,$F0,$7C,$AB,$8F,$C6,$6D,$FC,$36,$E3,$FC,$9B,$44,$7E,$BF,$E4
	.dc.b	$BF,$3F,$E6,$DC,$3F,$D1,$2D,$C3,$FD,$28,$DC,$3F,$97,$F5,$A3,$E9
	.dc.b	$7C,$1F,$D3,$FA,$DB,$87,$D8,$6E,$1F,$55,$B8,$78,$DB,$3F,$EE,$9C
	.dc.b	$D1,$8A,$79,$89,$CD,$7E,$11,$FC,$17,$D6,$FE,$0D,$B1,$93,$F1,$31
	.dc.b	$96,$9F,$8C,$FA,$DF,$C5,$7D,$17,$E2,$DB,$19,$3F,$23,$19,$69,$F7
	.dc.b	$30,$FF,$55,$FD,$6B,$EA,$BF,$15,$F5,$CF,$77,$F7,$73,$9B,$E0,$27
	.dc.b	$4F,$3B,$4E,$DF,$C2,$47,$E2,$67,$71,$83,$C3,$7E,$0B,$EE,$DA,$E9
	.dc.b	$7C,$F6,$FC,$CE,$1C,$9F,$C2,$7C,$87,$F1,$0F,$4D,$F3,$5F,$AD,$8C
	.dc.b	$33,$0F,$D9,$FD,$75,$D3,$F6,$0F,$EC,$FD,$DF,$E5,$3D,$96,$BF,$DA
	.dc.b	$6F,$5F,$6E,$47,$C9,$F6,$9F,$5A,$FA,$EC,$E5,$ED,$3F,$C9,$37,$FD
	.dc.b	$DE,$B7,$C3,$AD,$FF,$DC,$71,$8C,$7A,$63,$EF,$7F,$DB,$FC,$EE,$BF
	.dc.b	$81,$1F,$E2,$59,$D6,$5F,$8E,$9C,$BF,$EE,$6F,$AD,$FC,$85,$7D,$56
	.dc.b	$C7,$D1,$F2,$DF,$5C,$F3,$BF,$9C,$D6,$B4,$FE,$54,$DC,$3F,$DE,$3E
	.dc.b	$B3,$EB,$5F,$5D,$C3,$BF,$FB,$AC,$62,$BC,$11,$89,$D4,$7F,$6F,$9D
	.dc.b	$69,$F1,$FE,$BB,$FB,$C4,$FF,$96,$61,$18,$7F,$CB,$8D,$7F,$CE,$4D
	.dc.b	$7F,$D7,$AE,$F2,$FF,$17,$7D,$73,$E2,$3F,$AA,$CE,$6F,$F9,$24,$E5
	.dc.b	$FC,$D4,$66,$FF,$8E,$4E,$5F,$37,$B9,$FE,$31,$C1,$39,$7F,$A8,$7B
	.dc.b	$1F,$C8,$56,$0B,$7D,$03,$EB,$5E,$27,$F6,$8C,$38,$5E,$EB,$53,$9F
	.dc.b	$37,$FC,$81,$18,$9F,$D8,$57,$0C,$3A,$E1,$D2,$73,$7C,$23,$13,$F9
	.dc.b	$5A,$E1,$7C,$E3,$F3,$1F,$59,$FC,$FD,$FD,$B4,$79,$EF,$FB,$EC,$E7
	.dc.b	$2F,$6D,$78,$27,$2F,$76,$E3,$FC,$3E,$F9,$3F,$C1,$EF,$93,$EF,$C7
	.dc.b	$D5,$69,$CB,$D6,$78,$7D,$67,$C2,$FA,$CF,$F2,$5E,$5F,$DF,$ED,$39
	.dc.b	$F1,$7C,$ED,$BE,$27,$8E,$F8,$BF,$57,$0F,$04,$ED,$F8,$6D,$C3,$B7
	.dc.b	$AB,$E1,$E4,$F8,$6D,$BF,$63,$6F,$E5,$B6,$BE,$3F,$C2,$3E,$AF,$FB
	.dc.b	$FD,$FF,$33,$8C,$7A,$BF,$E0,$5E,$0F,$F3,$0D,$6B,$1F,$E5,$15,$AB
	.dc.b	$E1,$39,$C3,$FB,$35,$F0,$C3,$FC,$CF,$F5,$7F,$C1,$3C,$17,$C0,$BE
	.dc.b	$9F,$3E,$7F,$FA,$15,$B9,$FF,$A3,$DB,$AF,$F9,$F1,$FF,$43,$7D,$5B
	.dc.b	$E4,$BF,$D1,$9C,$38,$6D,$FF,$4D,$F8,$C5,$6E,$3F,$D3,$55,$B6,$9B
	.dc.b	$8D,$61,$FE,$8A,$9D,$36,$75,$EB,$FF,$A6,$FD,$CF,$1B,$63,$27,$FA
	.dc.b	$9E,$32,$D3,$FA,$B7,$D5,$FF,$8F,$FC,$BF,$E1,$CD,$BF,$F7,$43,$5F
	.dc.b	$F7,$F3,$5F,$C3,$7C,$7F,$F8,$C3,$B7,$F8,$B3,$63,$FD,$E0,$D7,$F1
	.dc.b	$81,$C3,$F8,$F9,$7C,$3F,$FA,$F2,$3C,$2D,$7C,$2D,$78,$35,$FF,$61
	.dc.b	$FC,$3B,$D0,$FC,$D7,$47,$ED,$38,$3F,$D0,$51,$8D,$3F,$D9,$95,$BB
	.dc.b	$FF,$66,$73,$E5,$E7,$56,$F8,$C7,$DB,$AE,$0F,$FA,$FA,$B5,$F3,$5F
	.dc.b	$E0,$78,$FC,$03,$5F,$EF,$D4,$8C,$63,$FE,$E0,$FB,$27,$D9,$F2,$E7
	.dc.b	$98,$72,$7C,$9D,$2A,$DC,$E4,$87,$6C,$3A,$EB,$FD,$A2,$77,$3F,$E0
	.dc.b	$FF,$33,$F8,$67,$A9,$EF,$6C,$7F,$B0,$7C,$4F,$EF,$4F,$FB,$BF,$E0
	.dc.b	$B5,$18,$8E,$4D,$D7,$F9,$0D,$72,$9F,$F7,$0D,$7A,$3D,$3F,$E2,$71
	.dc.b	$FB,$DF,$1F,$F0,$77,$FC,$09,$B9,$FF,$7E,$AD,$D7,$F8,$97,$5F,$F0
	.dc.b	$07,$8B,$F5,$7C,$4F,$E9,$13,$70,$E2,$FF,$80,$67,$ED,$3F,$BD,$F8
	.dc.b	$1B,$E0,$9A,$7D,$FE,$B7,$D9,$BE,$4B,$37,$FE,$3B,$BE,$0D,$BE,$15
	.dc.b	$B8,$D7,$A1,$D8,$8F,$F3,$F8,$FC,$EB,$A2,$1F,$E2,$5A,$FE,$E8,$EF
	.dc.b	$B3,$8B,$9A,$BC,$1F,$D3,$B2,$F3,$D7,$BB,$EA,$DC,$54,$98,$7F,$4F
	.dc.b	$E7,$F8,$AA,$76,$FF,$95,$DA,$FF,$8D,$DB,$87,$79,$E3,$18,$E3,$FE
	.dc.b	$8C,$DA,$FE,$F8,$DC,$7F,$3D,$1E,$5F,$85,$8F,$77,$E8,$5F,$FA,$46
	.dc.b	$30,$43,$C5,$1A,$F0,$F8,$C6,$0D,$63,$E5,$8F,$BF,$F8,$23,$ED,$29
	.dc.b	$2B,$FC,$E9,$D8,$F0,$31,$1F,$E5,$5A,$E0,$FF,$9E,$5A,$FF,$96,$67
	.dc.b	$83,$9A,$E3,$87,$FB,$B6,$B6,$FF,$BB,$EB,$75,$FE,$61,$FA,$B7,$C5
	.dc.b	$FF,$8E,$E1,$18,$7F,$C7,$CD,$7F,$93,$9A,$8E,$11,$9A,$FF,$7E,$37
	.dc.b	$7F,$EF,$0D,$AF,$F1,$77,$D6,$7F,$29,$AD,$DF,$E9,$2B,$55,$D4,$FF
	.dc.b	$4C,$56,$EF,$FC,$05,$F9,$5B,$E6,$BF,$56,$DA,$FF,$4F,$35,$FF,$3F
	.dc.b	$72,$7F,$6B,$7F,$D4,$4D,$73,$CD,$FB,$48,$C6,$7F,$E2,$89,$57,$EB
	.dc.b	$9F,$F4,$43,$63,$53,$C3,$0F,$F0,$B3,$6B,$FC,$99,$5B,$9F,$E1,$5F
	.dc.b	$31,$FC,$44,$93,$FE,$B5,$F9,$FF,$C8,$3F,$E9,$5F,$AD,$FE,$4E,$FF
	.dc.b	$95,$5A,$9B,$1F,$73,$1F,$43,$6F,$FC,$50,$DA,$FF,$2F,$37,$C0,$FF
	.dc.b	$4C,$7A,$BF,$C5,$3C,$EF,$81,$87,$C1,$6E,$1F,$E3,$16,$E1,$FE,$CA
	.dc.b	$F2,$7F,$A7,$D6,$EB,$91,$27,$FD,$B5,$F1,$6A,$DC,$F3,$35,$3F,$EB
	.dc.b	$3B,$EB,$5B,$F3,$3F,$9B,$7C,$FF,$F3,$7F,$9C,$97,$FE,$E5,$D3,$FD
	.dc.b	$61,$3A,$8C,$79,$0A,$DC,$7F,$37,$84,$FF,$41,$F2,$FF,$A1,$7C,$3C
	.dc.b	$AE,$31,$97,$9C,$65,$E5,$1F,$D2,$7E,$DF,$FA,$57,$CF,$C9,$F0,$C6
	.dc.b	$5F,$1C,$65,$F1,$46,$5F,$24,$67,$83,$FC,$EF,$9F,$FB,$2E,$BF,$B1
	.dc.b	$7E,$8D,$FC,$CA,$FF,$B3,$36,$BF,$AE,$B7,$83,$FA,$B7,$7F,$0F,$1E
	.dc.b	$5F,$FA,$76,$F7,$1F,$D8,$67,$FB,$17,$C1,$7F,$63,$F9,$99,$7F,$2A
	.dc.b	$32,$FC,$C8,$CB,$F2,$E3,$F1,$1F,$55,$23,$B6,$3F,$87,$8F,$EB,$FF
	.dc.b	$59,$FB,$99,$FF,$57,$73,$27,$84,$7E,$05,$DF,$3E,$34,$FC,$94,$5C
	.dc.b	$3F,$D0,$58,$73,$DD,$F5,$6F,$AB,$FF,$09,$BD,$72,$FC,$33,$5F,$EF
	.dc.b	$0E,$DC,$C7,$0C,$24,$60,$F5,$8F,$53,$FC,$4D,$EF,$FF,$85,$7C,$CF
	.dc.b	$F3,$BC,$EA,$72,$F8,$3D,$1D,$54,$8D,$57,$07,$C7,$3F,$E0,$4F,$A9
	.dc.b	$7C,$15,$24,$F8,$87,$B7,$FD,$F5,$70,$7F,$17,$AE,$31,$B6,$BA,$E0
	.dc.b	$FF,$55,$46,$27,$F8,$F5,$73,$79,$5F,$82,$F8,$1F,$EE,$5A,$E1,$7F
	.dc.b	$EB,$3B,$E0,$FF,$AD,$7E,$A1,$E1,$FF,$87,$9B,$7E,$96,$BF,$E2,$9F
	.dc.b	$A9,$7C,$E6,$EA,$4E,$DF,$F2,$6C,$EA,$3E,$77,$D3,$3C,$8F,$F2,$33
	.dc.b	$63,$31,$FE,$54,$F8,$F7,$C5,$6A,$24,$E8,$FE,$15,$3A,$7E,$D7,$E3
	.dc.b	$EF,$A6,$FF,$29,$36,$33,$1F,$E6,$0E,$47,$B1,$A8,$93,$A3,$F8,$B4
	.dc.b	$E9,$B3,$C4,$8E,$BE,$82,$BF,$E4,$D9,$E0,$D9,$F0,$4F,$93,$DC,$93
	.dc.b	$E8,$7C,$9B,$E2,$3F,$CD,$98,$FF,$88,$FA,$0F,$F8,$97,$B3,$FD,$9F
	.dc.b	$5F,$E8,$BF,$C8,$FE,$DD,$FF,$37,$6F,$79,$CD,$7F,$B5,$BE,$87,$87
	.dc.b	$C0,$4B,$FF,$66,$56,$B6,$56,$67,$FD,$1F,$3D,$0D,$46,$7A,$FF,$D4
	.dc.b	$91,$9B,$F9,$D3,$E5,$F9,$D9,$7D,$9F,$A5,$C5,$88,$EB,$17,$0F,$E5
	.dc.b	$78,$AB,$FC,$E3,$F1,$3F,$62,$56,$E6,$FC,$53,$FE,$C3,$E1,$CB,$9E
	.dc.b	$5F,$F6,$6C,$62,$77,$7C,$EB,$47,$47,$11,$A9,$E1,$5C,$AB,$C3,$3C
	.dc.b	$B9,$3F,$77,$1C,$8F,$F6,$A3,$88,$E5,$1C,$27,$93,$BB,$EF,$1F,$EE
	.dc.b	$2A,$E0,$EA,$B8,$D6,$E7,$2F,$9D,$AF,$C1,$3B,$1E,$BB,$F4,$BB,$C3
	.dc.b	$2F,$99,$B8,$7F,$9E,$2F,$B9,$ED,$8C,$DF,$69,$D6,$1E,$53,$D1,$3A
	.dc.b	$39,$E1,$E0,$C3,$C8,$7A,$63,$4F,$C1,$8D,$57,$48,$F3,$1B,$34,$74
	.dc.b	$38,$E1,$E5,$72,$47,$04,$BF,$34,$7C,$33,$8B,$E8,$7F,$C3,$DC,$D7
	.dc.b	$87,$F8,$64,$7C,$53,$27,$91,$CB,$F9,$2F,$A9,$FC,$9C,$62,$11,$89
	.dc.b	$FF,$38,$B6,$3F,$2A,$68,$EA,$7E,$61,$FA,$AB,$E2,$DD,$98,$8E,$18
	.dc.b	$FE,$81,$D4,$66,$B8,$D7,$09,$FC,$A1,$E7,$2B,$9A,$E5,$1D,$31,$F2
	.dc.b	$CE,$5F,$4C,$70,$BE,$CD,$3A,$9C,$4F,$D8,$CF,$17,$DF,$7E,$49,$E3
	.dc.b	$7E,$9A,$D5,$7C,$13,$A4,$79,$2B,$84,$76,$6C,$66,$37,$7E,$68,$E9
	.dc.b	$7D,$A3,$D6,$D9,$F2,$B5,$FE,$17,$7E,$27,$E3,$57,$27,$C0,$76,$3D
	.dc.b	$65,$3D,$0E,$CF,$21,$6B,$C8,$78,$A3,$F8,$7E,$3F,$CD,$7E,$BF,$EB
	.dc.b	$6B,$FC,$A9,$B8,$FF,$1E,$E5,$70,$BE,$8F,$38,$FF,$3F,$BB,$8F,$3C
	.dc.b	$3F,$9D,$C6,$B0,$E3,$7A,$E6,$4D,$F1,$25,$FF,$60,$FA,$3B,$97,$F4
	.dc.b	$09,$BA,$7F,$34,$9B,$5F,$A0,$C3,$FC,$88,$FF,$4A,$AF,$06,$8A,$DE
	.dc.b	$1F,$D5,$66,$E7,$EC,$9C,$F5,$FF,$5E,$EC,$F6,$B7,$1F,$EE,$7F,$2D
	.dc.b	$8E,$D5,$52,$4E,$8F,$F0,$1B,$E2,$FA,$29,$1F,$DF,$34,$BA,$7F,$77
	.dc.b	$9B,$B7,$F5,$09,$B5,$FD,$3B,$2F,$EE,$39,$F8,$4C,$E1,$FE,$4D,$CB
	.dc.b	$DD,$5F,$E1,$29,$D3,$FE,$0A,$C6,$D7,$F8,$D2,$DF,$52,$F9,$DF,$7B
	.dc.b	$6F,$FC,$3B,$F2,$78,$99,$1F,$E5,$79,$D5,$FF,$8F,$B7,$FE,$CD,$7A
	.dc.b	$8F,$F2,$08,$FA,$67,$D3,$7E,$22,$30,$FE,$17,$E1,$DF,$53,$9A,$DE
	.dc.b	$5F,$E5,$96,$DF,$46,$F0,$FF,$A8,$2F,$8D,$6F,$0E,$CF,$D6,$3F,$CC
	.dc.b	$45,$BF,$F2,$AA,$FA,$7F,$F2,$8B,$AE,$3E,$24,$7F,$CF,$66,$FC,$9C
	.dc.b	$57,$BA,$3A,$63,$FF,$00,$5A,$7F,$81,$8E,$07,$53,$E5,$79,$DF,$40
	.dc.b	$FD,$37,$89,$C2,$BF,$D1,$1C,$09,$D5,$70,$3F,$CA,$38,$E2,$BE,$95
	.dc.b	$FF,$85,$E7,$FC,$AB,$D7,$D2,$3A,$B6,$32,$7F,$96,$A3,$2F,$FA,$1F
	.dc.b	$E0,$5E,$EF,$D7,$6C,$7C,$2D,$5F,$4B,$9A,$93,$C5,$FE,$0A,$FC,$D8
	.dc.b	$CE,$3F,$84,$BE,$6D,$AD,$DF,$4B,$E4,$7F,$BD,$3E,$5F,$F8,$57,$ED
	.dc.b	$3F,$D4,$A3,$07,$FB,$ED,$C5,$FF,$BD,$E7,$73,$FC,$47,$E0,$91,$88
	.dc.b	$FF,$3B,$FD,$3F,$EE,$CE,$BE,$75,$8C,$92,$FF,$3D,$82,$75,$18,$FA
	.dc.b	$0F,$2F,$2F,$99,$B7,$E5,$6C,$EE,$3D,$13,$B6,$9B,$2C,$78,$3E,$2B
	.dc.b	$E0,$FD,$D7,$C2,$6C,$FC,$16,$CE,$E3,$E2,$4F,$C8,$F4,$3E,$4F,$75
	.dc.b	$9F,$90,$DA,$F6,$B7,$A3,$FB,$C7,$73,$5F,$DF,$C7,$F7,$BA,$FF,$06
	.dc.b	$D7,$EE,$FC,$CC,$E1,$FE,$64,$E4,$7E,$16,$F7,$B5,$72,$BF,$DA,$97
	.dc.b	$C3,$06,$1C,$2F,$51,$FC,$2A,$74,$F8,$2B,$FD,$DB,$5B,$79,$DF,$F5
	.dc.b	$4F,$39,$F5,$2F,$87,$C3,$6B,$F1,$1F,$69,$89,$FE,$29,$E9,$3E,$2E
	.dc.b	$99,$AF,$F7,$54,$65,$FF,$7E,$F0,$BA,$9F,$27,$F8,$F3,$BF,$6D,$AF
	.dc.b	$F4,$3F,$1F,$F8,$8F,$57,$E6,$0F,$1B,$F9,$27,$AA,$F9,$7F,$54,$E7
	.dc.b	$0F,$F7,$D4,$E5,$B3,$C0,$DD,$58,$EE,$7F,$83,$29,$F6,$5D,$0D,$72
	.dc.b	$FF,$BD,$B1,$F1,$C7,$F3,$5F,$A8,$7C,$45,$3F,$B1,$FC,$1F,$FB,$77
	.dc.b	$1E,$B8,$7F,$A8,$AB,$71,$88,$46,$1F,$E8,$6D,$D7,$FC,$89,$E8,$2F
	.dc.b	$A5,$6B,$0F,$06,$1F,$E5,$68,$ED,$1F,$91,$7F,$D5,$B8,$75,$7F,$CB
	.dc.b	$6D,$7F,$D6,$3F,$7F,$F0,$2B,$C8,$D8,$FE,$C5,$7F,$95,$9C,$E5,$FD
	.dc.b	$56,$B5,$97,$E7,$DB,$3F,$DC,$EB,$85,$FF,$5E,$9C,$BF,$EC,$46,$A7
	.dc.b	$39,$7B,$FB,$3B,$63,$FD,$8F,$E8,$7F,$A3,$5F,$F6,$86,$E5,$FA,$FF
	.dc.b	$BD,$7C,$3E,$EA,$69,$F3,$EF,$6F,$FB,$42,$B6,$E6,$76,$75,$3F,$B9
	.dc.b	$46,$9C,$47,$3D,$FC,$87,$F7,$7F,$27,$FB,$6C,$F9,$63,$33,$B3,$F0
	.dc.b	$13,$AC,$7F,$83,$61,$FE,$84,$BE,$15,$B7,$FC,$0B,$5B,$8F,$F7,$3C
	.dc.b	$EE,$B8,$FC,$A7,$C5,$3F,$C1,$1F,$79,$F8,$68,$CD,$7C,$19,$D3,$F9
	.dc.b	$19,$D3,$D2,$33,$87,$CD,$79,$FC,$36,$2F,$CF,$F2,$9F,$4C,$EF,$BD
	.dc.b	$C3,$E7,$45,$C3,$FC,$3B,$8F,$E0,$AF,$8D,$7F,$53,$F8,$0F,$F3,$2F
	.dc.b	$D1,$BE,$8F,$FA,$E6,$1F,$D2,$A7,$31,$8B,$FE,$5F,$E4,$FF,$99,$30
	.dc.b	$FD,$6E,$1C,$7D,$2F,$EC,$38,$70,$7F,$CA,$BF,$4B,$FE,$22,$AF,$F2
	.dc.b	$33,$72,$78,$5F,$E7,$8F,$4B,$7E,$84,$BF,$DC,$F4,$B0,$E1,$7B,$9C
	.dc.b	$C7,$3F,$84,$7D,$37,$F1,$1C,$9B,$FF,$AF,$67,$3E,$17,$F9,$FF,$EC
	.dc.b	$FF,$CE,$75,$AC,$7F,$DA,$B8,$71,$C3,$83,$63,$37,$C7,$E1,$7F,$D1
	.dc.b	$73,$9B,$63,$DA,$FF,$88,$B6,$3F,$CA,$58,$F8,$31,$EB,$8F,$FA,$53
	.dc.b	$E9,$1F,$2D,$4F,$F4,$2F,$D8,$DC,$B9,$CE,$9B,$3A,$69,$FC,$77,$5F
	.dc.b	$F2,$DF,$77,$F1,$5C,$FE,$57,$0F,$FB,$9D,$FE,$31,$3A,$6C,$65,$FF
	.dc.b	$78,$FC,$5F,$FA,$2B,$E9,$79,$36,$3F,$90,$35,$DF,$BE,$9F,$05,$BD
	.dc.b	$AF,$8B,$5E,$06,$BF,$EF,$E8,$D7,$0F,$09,$CC,$8F,$9B,$F4,$BF,$D8
	.dc.b	$5F,$9A,$D7,$ED,$1A,$F4,$B8,$3F,$E8,$1F,$8A,$7C,$A6,$39,$3F,$09
	.dc.b	$7C,$5B,$79,$E0,$F0,$64,$AD,$47,$FB,$3F,$0E,$78,$FD,$EF,$95,$FE
	.dc.b	$DA,$FF,$43,$B4,$62,$3C,$67,$E1,$9C,$3E,$68,$DF,$A7,$F5,$DB,$1D
	.dc.b	$63,$C8,$F1,$6F,$53,$EA,$3F,$C9,$33,$B2,$1F,$12,$B9,$BF,$EA,$38
	.dc.b	$DB,$FE,$AF,$F2,$3F,$C2,$4F,$EC,$1A,$E7,$D7,$EF,$93,$E4,$7C,$7D
	.dc.b	$EB,$DF,$07,$F6,$86,$ED,$FE,$C5,$6E,$3F,$C0,$30,$E4,$FF,$85,$BD
	.dc.b	$77,$D4,$7B,$BC,$5F,$E2,$93,$7C,$7F,$93,$97,$F9,$02,$F7,$3C,$6F
	.dc.b	$4C,$30,$7F,$7F,$FA,$3F,$F4,$B3,$FA,$36,$BF,$DE,$5A,$FD,$66,$BF
	.dc.b	$8A,$77,$5F,$E4,$D6,$E7,$9F,$88,$FA,$BD,$4F,$A8,$FF,$39,$61,$E1
	.dc.b	$F9,$EC,$FD,$DB,$C8,$A7,$DC,$96,$4A,$C3,$63,$D1,$1F,$BB,$76,$D8
	.dc.b	$D4,$7A,$E3,$89,$1F,$F1,$1E,$1E,$5E,$DF,$07,$89,$F4,$A7,$F5,$32
	.dc.b	$FB,$BF,$F5,$6C,$70,$6B,$FE,$E8,$9E,$1B,$7F,$9B,$DE,$51,$F9,$E6
	.dc.b	$BF,$B2,$6B,$FB,$B6,$A7,$9E,$DB,$E6,$F2,$9F,$33,$B3,$F0,$31,$CF
	.dc.b	$E6,$3E,$7E,$2F,$A0,$A7,$B2,$26,$3F,$C8,$71,$47,$C9,$6C,$66,$30
	.dc.b	$70,$7D,$CF,$87,$C9,$FA,$DF,$55,$FC,$DA,$7F,$C9,$5E,$07,$F9,$8D
	.dc.b	$B3,$AF,$26,$47,$F3,$1A,$E0,$FF,$91,$DA,$FF,$04,$6B,$FC,$61,$DE
	.dc.b	$DF,$E9,$B6,$E9,$9A,$FC,$46,$1F,$C5,$6F,$FC,$B9,$F5,$5F,$E7,$77
	.dc.b	$FC,$BD,$F5,$4F,$5B,$F1,$CD,$8D,$FC,$0F,$E4,$6B,$84,$7F,$96,$DA
	.dc.b	$FF,$18,$6B,$FC,$B1,$AF,$F3,$57,$8E,$7F,$EB,$16,$E5,$BA,$F0,$F8
	.dc.b	$9F,$9F,$FA,$97,$D5,$52,$63,$FE,$75,$E7,$FD,$1B,$7A,$7F,$41,$18
	.dc.b	$8F,$F3,$83,$5F,$E5,$AD,$7F,$A0,$D7,$ED,$9F,$E9,$51,$C7,$3F,$F6
	.dc.b	$33,$72,$DF,$6F,$F2,$CE,$F7,$CF,$52,$61,$FE,$89,$F9,$6F,$C0,$3F
	.dc.b	$D4,$07,$D5,$7F,$3C,$7F,$A8,$B5,$FE,$B4,$D7,$7C,$7F,$BB,$9F,$B5
	.dc.b	$1C,$DB,$9F,$FB,$6D,$B9,$7F,$63,$F2,$BF,$D7,$9E,$EF,$F0,$9F,$77
	.dc.b	$FD,$C1,$6A,$FF,$6F,$71,$3E,$E3,$FA,$4F,$C7,$7F,$85,$9E,$13,$33
	.dc.b	$FF,$03,$61,$FE,$99,$7F,$B4,$B5,$FE,$E4,$D7,$FD,$1A,$7C,$17,$C3
	.dc.b	$1C,$0E,$F7,$FD,$A2,$7F,$C7,$FB,$BD,$FF,$9D,$F5,$72,$7F,$A1,$47
	.dc.b	$D1,$30,$FF,$7C,$D7,$FD,$B5,$39,$8F,$E9,$1E,$7F,$D7,$F8,$B9,$18
	.dc.b	$21,$C7,$0F,$E3,$3B,$7F,$49,$9F,$BD,$5F,$7F,$EA,$BF,$01,$F5,$73
	.dc.b	$0F,$EA,$91,$97,$FA,$B4,$65,$E9,$19,$7B,$46,$6F,$FA,$B4,$FF,$66
	.dc.b	$FA,$BF,$E6,$B1,$FE,$6A,$EC,$FE,$F3,$F0,$7F,$A6,$C1,$EA,$2F,$7C
	.dc.b	$3F,$E2,$18,$C7,$99,$CC,$95,$FD,$A7,$C3,$FC,$C7,$C6,$7E,$67,$C9
	.dc.b	$93,$FD,$B6,$32,$FF,$6F,$8C,$BF,$16,$3F,$15,$E7,$7F,$53,$FA,$B7
	.dc.b	$D3,$D3,$35,$FE,$E9,$E9,$FE,$B4,$FE,$22,$BF,$AA,$46,$23,$FD,$C8
	.dc.b	$D7,$FC,$42,$E7,$83,$F2,$D1,$F8,$C9,$EA,$FE,$AE,$FF,$C8,$B3,$AE
	.dc.b	$77,$D5,$BE,$AF,$FC,$42,$F3,$E0,$FD,$23,$70,$FF,$46,$BB,$7F,$C9
	.dc.b	$57,$CE,$31,$1F,$8D,$78,$5F,$6C,$7A,$E1,$FE,$1E,$F4,$1E,$C7,$F0
	.dc.b	$5C,$21,$FE,$96,$8C,$B6,$38,$11,$F0,$E5,$E6,$F1,$7F,$C0,$1F,$1F
	.dc.b	$FE,$E8,$9C,$9F,$E5,$FF,$79,$F1,$DF,$E0,$56,$C6,$84,$F1,$7F,$89
	.dc.b	$35,$FF,$27,$B5,$FF,$31,$B5,$FF,$36,$35,$E7,$F1,$DE,$42,$3D,$E7
	.dc.b	$FC,$36,$FF,$1F,$FA,$B7,$8F,$FE,$52,$FA,$BF,$ED,$73,$FB,$76,$F1
	.dc.b	$7D,$E8,$57,$07,$D4,$D7,$A5,$7F,$88,$E3,$13,$F8,$1A,$FC,$9D,$7F
	.dc.b	$B1,$3E,$53,$F9,$AB,$FE,$49,$F8,$8F,$E7,$07,$F9,$52,$33,$1F,$E8
	.dc.b	$4E,$C4,$62,$FF,$23,$18,$F3,$78,$1F,$E6,$CF,$A0,$7C,$B7,$F9,$7B
	.dc.b	$EA,$3F,$B4,$63,$FE,$7A,$B4,$62,$3F,$D1,$F1,$8F,$5F,$FD,$60,$D8
	.dc.b	$FF,$40,$9F,$BE,$9E,$78,$F8,$9B,$C9,$FB,$F6,$CF,$F9,$9F,$DA,$FF
	.dc.b	$38,$91,$FC,$3F,$57,$E5,$23,$FC,$EB,$F1,$39,$9F,$F5,$04,$5C,$3F
	.dc.b	$55,$D0,$AF,$67,$BD,$E6,$C7,$FC,$6B,$9F,$B2,$30,$4C,$7F,$DE,$73
	.dc.b	$FE,$36,$7D,$51,$FE,$84,$F0,$3F,$D2,$BE,$7B,$EA,$BF,$0F,$5F,$E9
	.dc.b	$D6,$BF,$EA,$C6,$BF,$EE,$86,$C7,$1F,$A1,$E7,$1F,$90,$D3,$FD,$F4
	.dc.b	$DD,$FF,$D6,$5F,$57,$AA,$98,$7F,$BC,$2B,$4F,$FA,$3E,$72,$FF,$A8
	.dc.b	$E3,$38,$77,$8F,$97,$B2,$BC,$3F,$41,$FE,$B3,$B3,$49,$5F,$EB,$F8
	.dc.b	$CB,$E1,$8E,$2F,$4E,$4F,$F3,$F7,$C7,$7F,$94,$B4,$FF,$6E,$73,$FE
	.dc.b	$5A,$3F,$CE,$4D,$DF,$FD,$B2,$E7,$8D,$5C,$39,$FF,$DB,$B7,$CF,$1F
	.dc.b	$CB,$C6,$30,$FF,$70,$F3,$F3,$9F,$F7,$53,$CE,$ED,$72,$C3,$AD,$7E
	.dc.b	$8C,$FF,$78,$61,$E4,$9E,$91,$E1,$33,$3C,$31,$E9,$1C,$45,$71,$79
	.dc.b	$E3,$EC,$BF,$06,$1C,$E7,$78,$78,$B1,$F4,$47,$B0,$E2,$7A,$0F,$0C
	.dc.b	$D7,$8B,$CF,$0F,$04,$78,$B0,$F0,$3C,$0F,$34,$7C,$9C,$7B,$93,$1F
	.dc.b	$2C,$EB,$1F,$74,$7C,$28,$E3,$7D,$1F,$09,$BC,$3C,$98,$7F,$AB,$71
	.dc.b	$F5,$9F,$18,$F8,$F8,$F7,$AF,$69,$C4,$91,$C3,$0F,$03,$C5,$F2,$B9
	.dc.b	$8E,$2E,$5F,$76,$FF,$C6,$9B,$1E,$C8,$C4,$7A,$1A,$87,$CF,$D8,$CC
	.dc.b	$60,$38,$8D,$4F,$F1,$E6,$CF,$33,$A3,$F5,$D7,$4F,$F8,$0A,$7F,$9B
	.dc.b	$C6,$8F,$C0,$1E,$38,$D5,$F5,$8C,$9C,$B8,$FA,$3E,$5A,$E7,$5C,$23
	.dc.b	$C4,$7A,$71,$CD,$7E,$46,$3C,$11,$87,$F3,$47,$8E,$B0,$79,$9E,$D7
	.dc.b	$D2,$FB,$47,$BF,$0E,$91,$FE,$1E,$BE,$11,$8C,$3C,$15,$E3,$BE,$51
	.dc.b	$89,$D4,$F8,$8D,$1F,$2E,$FA,$57,$C3,$BF,$41,$DA,$7E,$2C,$F9,$1F
	.dc.b	$54,$F6,$7E,$49,$D0,$F9,$D7,$E1,$7E,$0C,$F6,$7E,$4D,$F9,$2F,$B1
	.dc.b	$E0,$BF,$A9,$87,$C6,$69,$DD,$DB,$E1,$8B,$5E,$BC,$3F,$86,$63,$FD
	.dc.b	$27,$E5,$BF,$C8,$BA,$7F,$40,$6C,$FF,$A8,$78,$DF,$49,$8C,$3F,$A5
	.dc.b	$55,$C3,$95,$FF,$35,$8F,$F4,$E4,$62,$7F,$06,$D7,$FA,$B7,$8D,$C7
	.dc.b	$90,$F5,$EB,$FF,$20,$FD,$56,$62,$63,$FD,$3D,$D6,$69,$CE,$1F,$D3
	.dc.b	$E6,$E9,$FD,$CA,$6D,$FF,$63,$9F,$2E,$3E,$08,$FE,$99,$18,$3F,$D6
	.dc.b	$9D,$7F,$E1,$0E,$CF,$80,$D9,$FE,$B3,$F5,$78,$ED,$FE,$FD,$1F,$DE
	.dc.b	$2B,$8C,$63,$3F,$F0,$D3,$5F,$F0,$CB,$5F,$9D,$F5,$7E,$C3,$DB,$AD
	.dc.b	$BF,$EC,$F1,$67,$FD,$6D,$3F,$E6,$69,$C9,$FE,$08,$D3,$C8,$6F,$5F
	.dc.b	$F1,$03,$88,$FF,$12,$E7,$FE,$2C,$8F,$39,$23,$3F,$50,$FA,$4F,$82
	.dc.b	$D8,$FF,$11,$F6,$3E,$AF,$35,$27,$FC,$4F,$ED,$BE,$AF,$1A,$FF,$B1
	.dc.b	$5F,$C5,$2F,$AB,$7C,$D6,$1B,$7F,$E3,$0F,$6B,$93,$FE,$73,$8C,$D7
	.dc.b	$F7,$68,$FE,$41,$18,$DF,$FD,$35,$4F,$83,$E9,$87,$E3,$63,$17,$FE
	.dc.b	$7D,$3E,$B5,$7F,$9A,$5A,$7F,$35,$1F,$57,$FE,$72,$78,$72,$FA,$9F
	.dc.b	$B5,$96,$E3,$FD,$F5,$18,$83,$4E,$AE,$3E,$81,$E7,$F3,$AF,$D7,$52
	.dc.b	$72,$FF,$A6,$39,$1C,$6F,$4B,$EA,$9F,$A9,$E4,$EB,$3A,$8C,$D7,$D4
	.dc.b	$F8,$FB,$9F,$FB,$DD,$E1,$7C,$7C,$0B,$5F,$79,$B5,$F3,$C7,$72,$BF
	.dc.b	$82,$D7,$E1,$1E,$4F,$FA,$A7,$E1,$9F,$07,$F8,$06,$C7,$FA,$C8,$F9
	.dc.b	$BA,$E6,$64,$7F,$9E,$B0,$57,$07,$FC,$FA,$D7,$FA,$7B,$5F,$EC,$4E
	.dc.b	$F8,$7F,$DC,$33,$F6,$BE,$1F,$AD,$7A,$E3,$07,$E6,$3E,$8B,$F8,$97
	.dc.b	$C1,$7D,$88,$C1,$FE,$82,$71,$3F,$0F,$47,$52,$F8,$DF,$F1,$B8,$C3
	.dc.b	$C4,$FE,$11,$F4,$31,$E9,$E8,$DC,$3F,$D5,$5E,$12,$BC,$7E,$CE,$75
	.dc.b	$FF,$7B,$69,$FE,$D8,$C7,$C9,$F5,$7E,$8E,$CF,$4B,$63,$27,$A6,$32
	.dc.b	$D3,$39,$7A,$EB,$F5,$51,$FE,$15,$D9,$0B,$E9,$5B,$F8,$CF,$8F,$F5
	.dc.b	$7F,$21,$B8,$7C,$76,$CE,$E3,$E5,$4F,$CE,$F5,$3E,$7F,$87,$67,$E7
	.dc.b	$36,$BC,$8D,$D3,$E0,$BF,$C2,$BD,$0F,$E2,$93,$F9,$53,$FC,$57,$1F
	.dc.b	$D5,$F9,$5F,$56,$F8,$2E,$13,$31,$FC,$76,$1C,$B0,$FE,$E3,$3F,$EE
	.dc.b	$B7,$FC,$1C,$D7,$FC,$5A,$F5,$BF,$D0,$C7,$86,$7C,$78,$FF,$00,$8F
	.dc.b	$CC,$CE,$A3,$1E,$9B,$EA,$DF,$17,$FD,$B6,$30,$FE,$4B,$EA,$F1,$3A
	.dc.b	$CF,$FC,$91,$7F,$EF,$FF,$59,$F4,$5E,$5C,$FF,$DD,$6D,$8C,$D7,$EA
	.dc.b	$DB,$3F,$E3,$6E,$5F,$CC,$F6,$FE,$5C,$EF,$DD,$6F,$B9,$FE,$45,$FA
	.dc.b	$B7,$95,$FC,$EB,$D7,$78,$BE,$99,$CD,$70,$CB,$FD,$63,$3B,$3C,$D5
	.dc.b	$C2,$7F,$7C,$D9,$FF,$7C,$B5,$FF,$19,$57,$E7,$DF,$E2,$5C,$28,$C6
	.dc.b	$DF,$3E,$55,$CE,$3E,$7C,$EA,$FF,$CC,$DF,$56,$F8,$9F,$EA,$6D,$8F
	.dc.b	$E5,$DF,$0A,$F8,$1C,$D4,$9F,$E2,$95,$BC,$BF,$CB,$E1,$C3,$FE,$62
	.dc.b	$6B,$E0,$9F,$F1,$56,$2F,$B0,$7C,$6F,$5A,$FE,$5D,$39,$8F,$EC,$FF
	.dc.b	$59,$E9,$9F,$F4,$43,$7C,$E6,$9F,$9D,$CF,$E8,$52,$B8,$37,$C5,$FC
	.dc.b	$34,$7C,$B8,$FF,$3B,$B5,$FF,$45,$35,$EF,$1F,$95,$FC,$01,$F3,$1E
	.dc.b	$48,$C3,$FD,$F3,$E9,$7F,$B4,$57,$EC,$DB,$A7,$D5,$FA,$DF,$EB,$73
	.dc.b	$93,$81,$F9,$99,$E2,$FF,$A2,$9A,$FF,$A7,$5A,$FF,$AA,$5E,$B9,$7F
	.dc.b	$AA,$FE,$85,$F4,$BD,$B0,$FC,$25,$6A,$D5,$AF,$5F,$D1,$CF,$31,$F9
	.dc.b	$3E,$8F,$F8,$42,$FD,$53,$97,$EF,$E1,$FE,$A3,$6C,$7A,$49,$5F,$A1
	.dc.b	$C1,$FA,$C7,$3C,$BC,$75,$FE,$6B,$9C,$FB,$9F,$27,$DB,$98,$7D,$AF
	.dc.b	$AE,$FC,$93,$F7,$33,$FF,$0C,$E7,$F3,$6B,$E4,$12,$7F,$B4,$4E,$A7
	.dc.b	$E5,$4F,$7F,$B2,$FF,$1C,$57,$38,$C7,$BF,$F6,$7E,$72,$4F,$87,$CD
	.dc.b	$B8,$7D,$B9,$FA,$C4,$9E,$EF,$F2,$5F,$17,$FA,$3F,$C4,$7F,$A1,$FE
	.dc.b	$A7,$E9,$FB,$7F,$DB,$23,$1C,$DF,$E1,$DC,$3F,$B6,$46,$27,$51,$FC
	.dc.b	$26,$B8,$BF,$E2,$1F,$AC,$FE,$AB,$93,$B9,$18,$C9,$F0,$BF,$DA,$6F
	.dc.b	$87,$B3,$FE,$7A,$AE,$37,$C2,$3F,$D1,$7E,$D6,$5F,$F4,$E7,$D7,$FF
	.dc.b	$09,$BF,$F7,$A3,$72,$FF,$10,$FD,$7F,$FA,$25,$FF,$12,$7C,$62,$74
	.dc.b	$DF,$33,$99,$FE,$9D,$AE,$1D,$9F,$D1,$67,$58,$7F,$27,$6E,$1F,$EA
	.dc.b	$8C,$3F,$91,$3E,$2C,$3F,$D5,$B1,$88,$4B,$EA,$1F,$3B,$FE,$B5,$6E
	.dc.b	$1F,$CC,$F8,$3E,$2F,$D7,$E6,$A4,$F8,$6B,$78,$7F,$A3,$AB,$2F,$F6
	.dc.b	$DE,$0F,$F3,$BF,$7F,$F8,$D5,$B1,$FC,$89,$A8,$87,$CF,$8F,$F1,$E7
	.dc.b	$C8,$BE,$27,$F6,$73,$FC,$AA,$31,$E0,$EA,$73,$3C,$84,$E9,$F0,$7C
	.dc.b	$07,$C2,$6E,$DE,$58,$C4,$71,$C5,$18,$8F,$34,$BE,$BF,$FB,$5C,$7C
	.dc.b	$7E,$7F,$F3,$AB,$7C,$6F,$5F,$27,$D9,$AD,$CF,$E0,$9B,$1F,$E0,$1E
	.dc.b	$2F,$C9,$D7,$0E,$1F,$F7,$63,$6F,$EC,$DF,$09,$F1,$FC,$DB,$D7,$D6
	.dc.b	$0E,$77,$D2,$61,$B8,$FE,$26,$F8,$36,$F5,$52,$30,$43,$FC,$B5,$5C
	.dc.b	$9B,$1E,$5D,$57,$D2,$7F,$D0,$5D,$6F,$17,$FC,$33,$1F,$8D,$FD,$EF
	.dc.b	$EB,$6F,$F0,$5B,$AF,$E0,$EF,$87,$BD,$F0,$5B,$7C,$4F,$7D,$70,$7F
	.dc.b	$7D,$87,$67,$FD,$25,$DF,$FC,$F9,$F9,$4D,$7F,$84,$78,$5C,$78,$3F
	.dc.b	$CE,$BF,$5F,$A8,$E9,$5B,$AE,$71,$BF,$8A,$57,$D5,$9F,$EB,$FF,$1B
	.dc.b	$F5,$FA,$DF,$5C,$F4,$B0,$DA,$FE,$E3,$E9,$7F,$AD,$5B,$7A,$A9,$18
	.dc.b	$37,$EB,$72,$F9,$59,$18,$21,$FE,$B1,$C7,$C3,$1A,$8F,$F3,$06,$1F
	.dc.b	$E0,$A7,$7C,$9F,$E5,$1A,$DC,$7F,$53,$C9,$7E,$DB,$E1,$3F,$E6,$9F
	.dc.b	$0D,$F3,$1E,$FF,$AE,$FF,$5C,$CF,$F5,$B6,$BB,$8F,$8D,$3C,$23,$5F
	.dc.b	$03,$C4,$CC,$C8,$C9,$0E,$47,$F8,$62,$5F,$35,$E1,$FA,$49,$39,$8E
	.dc.b	$E6,$1F,$89,$82,$73,$1B,$3D,$0E,$F5,$FF,$23,$5F,$69,$FF,$13,$FD
	.dc.b	$77,$F2,$87,$FC,$53,$E0,$7F,$97,$1B,$E6,$7E,$0E,$75,$1F,$E8,$F6
	.dc.b	$DF,$F9,$46,$B8,$4F,$FA,$4D,$B1,$FE,$55,$9D,$F6,$FF,$24,$6D,$7F
	.dc.b	$27,$AD,$CF,$F2,$86,$C7,$F2,$B9,$DA,$FA,$DF,$F3,$9B,$FE,$40,$FA
	.dc.b	$EF,$F2,$C3,$7E,$0F,$F3,$53,$A8,$FF,$53,$37,$1F,$F3,$45,$70,$9F
	.dc.b	$F5,$53,$63,$FC,$D9,$3B,$C7,$F9,$DB,$CE,$BF,$9A,$35,$CF,$C3,$7F
	.dc.b	$BE,$5F,$33,$7E,$B5,$F5,$DF,$C7,$3E,$51,$F7,$94,$FD,$84,$EA,$FF
	.dc.b	$CB,$7E,$6F,$F9,$41,$FF,$2E,$C6,$5F,$F6,$03,$73,$FF,$41,$4E,$E7
	.dc.b	$FD,$88,$D8,$CF,$3F,$EF,$63,$1A,$7F,$56,$BF,$DC,$5F,$FB,$F7,$BD
	.dc.b	$F4,$3E,$7F,$A8,$93,$FD,$AE,$72,$FF,$9C,$3E,$BF,$F9,$36,$7F,$E9
	.dc.b	$3F,$A8,$FB,$32,$6C,$EB,$2F,$EC,$B1,$FD,$46,$79,$72,$FF,$11,$AD
	.dc.b	$47,$F6,$B9,$D3,$FE,$DF,$F7,$BF,$D4,$18,$7F,$B3,$FE,$5B,$FC,$07
	.dc.b	$E7,$FF,$4D,$6D,$7F,$A1,$2B,$79,$7F,$74,$6C,$27,$6A,$FA,$79,$FF
	.dc.b	$D1,$5F,$F5,$75,$70,$9F,$F4,$81,$FE,$BF,$AE,$33,$C3,$0F,$E5,$55
	.dc.b	$DB,$CA,$6A,$F7,$FF,$AE,$AF,$AF,$A6,$22,$65,$F3,$E3,$2F,$D1,$8E
	.dc.b	$3E,$87,$2F,$12,$4F,$D5,$8C,$BF,$5E,$32,$FD,$D8,$F0,$FD,$77,$8B
	.dc.b	$EB,$E4,$FD,$D8,$CB,$F7,$E3,$2F,$E1,$63,$2F,$E1,$E3,$3A,$FF,$80
	.dc.b	$72,$57,$F5,$FF,$AF,$FE,$5F,$3F,$E0,$5E,$BF,$D6,$60,$D5,$18,$F4
	.dc.b	$D7,$C3,$8F,$FB,$53,$6F,$85,$72,$BE,$3D,$DD,$FC,$CF,$9B,$F2,$1F
	.dc.b	$3B,$C7,$98,$7E,$66,$32,$FE,$6E,$32,$FE,$86,$3E,$CF,$97,$FD,$8F
	.dc.b	$EB,$DF,$57,$4C,$D7,$E3,$67,$0F,$FB,$4A,$76,$E4,$87,$E3,$75,$FF
	.dc.b	$22,$C6,$A3,$07,$F9,$25,$C6,$FF,$90,$AF,$F1,$1F,$D7,$BB,$EC,$4B
	.dc.b	$FF,$2F,$4E,$5B,$39,$7A,$EC,$9F,$EC,$71,$A3,$FC,$B1,$6C,$7F,$C3
	.dc.b	$1E,$C7,$FA,$32,$7F,$80,$1F,$98,$3F,$C3,$4E,$5F,$F1,$2B,$67,$F0
	.dc.b	$51,$B3,$AC,$F4,$F7,$3F,$C1,$DF,$25,$FE,$4E,$9F,$82,$3E,$BF,$FC
	.dc.b	$18,$E1,$B7,$E5,$8F,$F3,$74,$F1,$8C,$42,$31,$7F,$E7,$2A,$E2,$DF
	.dc.b	$57,$C3,$EF,$4B,$F2,$9F,$E7,$9A,$DB,$67,$4F,$9B,$EB,$FF,$8E,$3F
	.dc.b	$E8,$DF,$AF,$FE,$D8,$FC,$E6,$E5,$E4,$9D,$CF,$1F,$17,$FC,$BB,$5A
	.dc.b	$CB,$FD,$11,$BA,$31,$5F,$E8,$BB,$E3,$87,$C1,$EA,$7C,$CF,$F3,$2A
	.dc.b	$FE,$71,$F0,$BF,$C3,$0F,$F2,$74,$66,$B8,$37,$2E,$FD,$CA,$E1,$7F
	.dc.b	$0B,$0E,$BE,$57,$F9,$5F,$E9,$3F,$B2,$61,$E6,$F8,$1E,$17,$FE,$61
	.dc.b	$C3,$5E,$A4,$8C,$10,$FF,$34,$DF,$27,$F8,$07,$2F,$BB,$24,$E5,$B5
	.dc.b	$FE,$82,$9F,$87,$C1,$FE,$DE,$F6,$7B,$61,$D2,$B5,$1E,$7F,$6B,$FC
	.dc.b	$CB,$F8,$5F,$9A,$BF,$CD,$2D,$E1,$F9,$59,$FF,$8E,$67,$CB,$5C,$34
	.dc.b	$91,$F0,$E7,$55,$FE,$D9,$FA,$F7,$D3,$FF,$9E,$3E,$CE,$5F,$FB,$6E
	.dc.b	$72,$D9,$CB,$FB,$2E,$05,$7F,$A3,$EB,$73,$FE,$EC,$6D,$FF,$A2,$E7
	.dc.b	$8B,$E1,$BE,$06,$38,$7E,$A4,$E2,$FC,$5F,$4D,$FE,$8C,$F8,$89,$18
	.dc.b	$21,$E3,$72,$F8,$63,$FD,$05,$F1,$5F,$E5,$4C,$7F,$DB,$9C,$FF,$8F
	.dc.b	$8F,$90,$DD,$FF,$DB,$EE,$78,$7F,$B5,$61,$F8,$A3,$FD,$C9,$5C,$DC
	.dc.b	$1F,$ED,$CA,$E3,$CD,$FE,$E8,$9E,$AD,$AE,$77,$D7,$0F,$1C,$EE,$F8
	.dc.b	$E1,$D9,$AF,$FB,$A2,$F8,$CF,$3A,$E3,$3B,$29,$E3,$2E,$1E,$7C,$3C
	.dc.b	$11,$E8,$8F,$2B,$C4,$C5,$F6,$7D,$ED,$7A,$1C,$70,$F2,$CF,$5C,$3C
	.dc.b	$93,$93,$C3,$3A,$AE,$55,$B9,$F0,$5F,$D1,$7C,$F3,$A3,$75,$D8,$E5
	.dc.b	$86,$8E,$B3,$0F,$4B,$97,$C4,$4A,$F8,$24,$C3,$E8,$47,$18,$C3,$F0
	.dc.b	$24,$BD,$47,$D0,$7B,$9B,$1B,$35,$19,$38,$B3,$E9,$79,$CF,$08,$C1
	.dc.b	$C0,$D8,$8D,$CE,$23,$06,$CE,$37,$FC,$A5,$FE,$45,$BF,$E0,$9F,$4D
	.dc.b	$5E,$4F,$EA,$15,$B9,$E6,$F1,$8C,$E5,$DE,$B8,$5F,$3B,$DC,$F9,$AB
	.dc.b	$71,$E8,$AD,$BF,$A4,$3C,$25,$8D,$BE,$89,$E5,$5D,$6B,$8D,$F9,$5B
	.dc.b	$3E,$3A,$E5,$5B,$69,$DC,$F2,$9F,$42,$FB,$BE,$2B,$F7,$4F,$16,$C7
	.dc.b	$1B,$EE,$FA,$CE,$B1,$E3,$6D,$F8,$E3,$89,$92,$B9,$3D,$B1,$D4,$E2
	.dc.b	$75,$8D,$B6,$37,$1F,$E9,$4C,$5C,$4F,$94,$FF,$04,$69,$FE,$09,$E8
	.dc.b	$72,$BE,$67,$3A,$FE,$3B,$0E,$0F,$F8,$12,$B7,$3A,$D3,$F3,$95,$A7
	.dc.b	$F9,$E4,$67,$2F,$0E,$5B,$BE,$38,$7F,$45,$F1,$F1,$93,$E0,$F8,$66
	.dc.b	$9C,$C6,$1B,$5C,$4F,$25,$71,$71,$9F,$F4,$DA,$E1,$1A,$6C,$66,$37
	.dc.b	$A7,$C3,$AE,$53,$FD,$DB,$E3,$FD,$DF,$18,$FC,$B1,$E8,$7C,$1F,$07
	.dc.b	$FB,$BD,$6A,$73,$97,$F6,$78,$C5,$6F,$B9,$18,$9F,$99,$E0,$BE,$C3
	.dc.b	$5F,$61,$F1,$E3,$13,$FD,$C6,$31,$3A,$8E,$2D,$D3,$FC,$1B,$3F,$33
	.dc.b	$9F,$1B,$FB,$5B,$7F,$E1,$96,$D7,$FA,$47,$7F,$F6,$FC,$FF,$88,$7E
	.dc.b	$B9,$F5,$5F,$67,$C6,$FC,$34,$E6,$31,$D3,$F0,$EB,$73,$FE,$E5,$8C
	.dc.b	$4E,$A3,$FC,$9B,$83,$EC,$1F,$33,$F8,$5F,$5D,$F3,$BF,$8D,$AD,$4E
	.dc.b	$63,$15,$C2,$FF,$29,$7C,$1E,$B1,$9B,$FF,$3F,$CE,$5F,$F3,$6C,$FD
	.dc.b	$23,$FC,$BE,$3E,$71,$E8,$F8,$63,$53,$FE,$94,$75,$19,$E4,$FD,$14
	.dc.b	$60,$F6,$9C,$38,$7F,$CE,$CE,$6B,$C3,$3C,$2F,$BC,$F0,$8C,$5F,$CB
	.dc.b	$F9,$3F,$A1,$D7,$6B,$F6,$5C,$3F,$96,$CB,$FC,$C5,$C2,$BD,$D7,$D9
	.dc.b	$6E,$7F,$EA,$0F,$AE,$1F,$B3,$1F,$5F,$5C,$36,$BF,$80,$46,$9A,$E9
	.dc.b	$FB,$7F,$0E,$F8,$3F,$87,$F6,$0F,$8A,$FB,$53,$9A,$F9,$F3,$97,$8D
	.dc.b	$70,$AD,$B6,$32,$70,$8C,$B4,$F8,$7F,$49,$FC,$37,$E0,$BF,$D7,$11
	.dc.b	$83,$F8,$C3,$8C,$FF,$D8,$3B,$AB,$85,$7E,$75,$AF,$F9,$7F,$E8,$3A
	.dc.b	$7B,$5F,$B9,$6C,$64,$EB,$19,$69,$E0,$FB,$0F,$27,$57,$95,$B1,$93
	.dc.b	$CB,$19,$69,$9C,$7F,$DA,$F9,$7E,$6F,$8F,$F9,$C2,$BE,$73,$AF,$8F
	.dc.b	$F8,$5F,$61,$F0,$DB,$87,$C2,$6C,$EE,$3E,$2C,$FC,$9F,$57,$E5,$77
	.dc.b	$59,$F9,$2D,$AF,$1B,$72,$F1,$BF,$B7,$F4,$3F,$80,$CF,$FB,$EB,$91
	.dc.b	$F6,$1F,$89,$8F,$F6,$EC,$EA,$FF,$DB,$55,$AE,$7E,$58,$7F,$4D,$8C
	.dc.b	$63,$FC,$22,$F8,$3F,$E2,$C8,$CE,$1D,$2B,$86,$3D,$F1,$FF,$71,$E1
	.dc.b	$C2,$97,$C2,$7F,$DD,$5E,$93,$EB,$DF,$19,$FC,$47,$D5,$7D,$7F,$F5
	.dc.b	$FA,$D4,$E6,$26,$BF,$EE,$89,$E5,$D5,$8C,$D9,$76,$CB,$F4,$51,$F9
	.dc.b	$2E,$AF,$CC,$7A,$3F,$CB,$5B,$8F,$F2,$6F,$5B,$F7,$CF,$FA,$33,$C1
	.dc.b	$F1,$7D,$7B,$C6,$FE,$4D,$5A,$C7,$FD,$2B,$5A,$EC,$CB,$70,$F3,$60
	.dc.b	$9F,$D4,$7C,$96,$6F,$F1,$EF,$F3,$1F,$AF,$7C,$63,$0F,$F2,$55,$27
	.dc.b	$31,$86,$C7,$F9,$3D,$AF,$F9,$51,$CE,$FF,$E1,$2F,$AF,$FE,$CF,$3F
	.dc.b	$1A,$BF,$CB,$F5,$A7,$FC,$D3,$4A,$DB,$E3,$AF,$F3,$6D,$6D,$F9,$F3
	.dc.b	$AA,$E7,$7F,$E1,$7B,$E5,$EA,$7F,$4D,$FA,$FF,$85,$5F,$67,$EB,$DF
	.dc.b	$33,$9D,$7F,$01,$18,$9F,$EB,$35,$C3,$3F,$40,$E6,$56,$E3,$22,$78
	.dc.b	$65,$F9,$09,$D3,$EB,$9D,$61,$E0,$C7,$C5,$8F,$95,$FE,$C9,$F4,$BF
	.dc.b	$E8,$C8,$F8,$A7,$FA,$3B,$E8,$DF,$4B,$FD,$33,$0F,$0D,$6D,$F3,$CE
	.dc.b	$BC,$FE,$71,$FE,$3B,$8C,$5F,$F7,$6A,$E1,$97,$CF,$9C,$D7,$43,$86
	.dc.b	$1D,$E7,$D5,$5C,$2F,$C9,$39,$7F,$D3,$71,$FE,$8E,$AF,$1F,$D7,$7F
	.dc.b	$83,$1F,$3E,$DE,$38,$C6,$1D,$23,$1E,$AF,$C2,$6E,$1E,$76,$DF,$28
	.dc.b	$F8,$AB,$26,$1D,$27,$C7,$EA,$FC,$1F,$AD,$7D,$6F,$CE,$8C,$36,$79
	.dc.b	$DA,$B7,$1F,$1E,$B6,$D3,$71,$AC,$3E,$9C,$E9,$FA,$BF,$37,$E1,$F8
	.dc.b	$6F,$BA,$DB,$FA,$4D,$9D,$C7,$49,$FF,$2A,$EF,$FE,$19,$8F,$7F,$C1
	.dc.b	$7E,$0B,$D9,$FE,$C9,$F0,$BA,$64,$EA,$B8,$61,$FD,$3A,$B2,$FE,$73
	.dc.b	$E0,$DF,$55,$FD,$46,$FF,$DB,$7F,$0D,$FE,$88,$F9,$1F,$F3,$14,$E6
	.dc.b	$FF,$C9,$51,$8F,$5B,$FC,$FA,$DB,$FF,$3B,$BF,$E8,$5F,$AD,$FE,$31
	.dc.b	$83,$E1,$5F,$3F,$9F,$1B,$FC,$C4,$DC,$3F,$CC,$ED,$CB,$F6,$37,$C6
	.dc.b	$31,$D5,$FE,$9D,$C3,$9D,$FE,$0B,$E2,$99,$78,$27,$38,$FE,$EA,$5F
	.dc.b	$4E,$FA,$DF,$F4,$AF,$1F,$FA,$6E,$31,$8F,$FA,$42,$31,$3A,$8E,$D3
	.dc.b	$AB,$F2,$FC,$37,$CE,$CF,$F7,$DC,$3F,$E6,$0E,$DF,$E5,$6D,$AF,$F7
	.dc.b	$B3,$5F,$F7,$DB,$51,$3E,$3B,$FD,$17,$D9,$FC,$D1,$B1,$FE,$FF,$6B
	.dc.b	$C0,$E1,$F0,$DF,$A3,$7B,$FB,$AD,$FC,$6D,$8C,$C6,$0E,$4F,$89,$7B
	.dc.b	$BF,$EC,$98,$F9,$DF,$A9,$3E,$CB,$C1,$AF,$F3,$7B,$4E,$6B,$EA,$5F
	.dc.b	$17,$E9,$36,$B8,$4E,$E7,$EB,$36,$3A,$4E,$F8,$3C,$93,$F8,$BA,$F1
	.dc.b	$F1,$F8,$A7,$2D,$87,$33,$F4,$EF,$E8,$31,$FE,$DF,$6B,$E0,$9E,$19
	.dc.b	$7A,$9E,$55,$C7,$C8,$B7,$C3,$3F,$4D,$70,$CB,$C9,$8F,$E6,$2B,$59
	.dc.b	$7E,$B3,$8F,$FC,$D1,$3F,$C6,$3B,$3C,$9E,$9F,$9D,$B1,$FE,$E0,$F5
	.dc.b	$3C,$35,$C2,$3F,$71,$87,$F8,$15,$E4,$D7,$F2,$73,$E8,$F8,$27,$DE
	.dc.b	$D2,$4F,$F9,$A3,$E2,$6B,$FE,$52,$8B,$B7,$F9,$4A,$7E,$54,$FF,$68
	.dc.b	$CE,$EF,$DE,$BF,$AE,$37,$5F,$EA,$DF,$A4,$3E,$57,$E9,$74,$A3,$18
	.dc.b	$FF,$B1,$AF,$83,$FA,$06,$CE,$EB,$BF,$A1,$2B,$83,$D2,$DE,$16,$67
	.dc.b	$FB,$9E,$1F,$E2,$17,$67,$F3,$C9,$DB,$67,$86,$5E,$AB,$E9,$1F,$03
	.dc.b	$BB,$F4,$D5,$FD,$F9,$B5,$FD,$ED,$B9,$FF,$75,$FB,$77,$D5,$71,$E9
	.dc.b	$98,$7F,$B9,$2F,$6F,$F5,$7A,$DB,$FD,$41,$AF,$79,$E2,$FF,$7C,$8C
	.dc.b	$E1,$E1,$F8,$1F,$31,$27,$97,$15,$7D,$11,$67,$FB,$F5,$7E,$12,$76
	.dc.b	$79,$67,$59,$7A,$EF,$F0,$3F,$5D,$FC,$13,$0F,$F0,$D3,$67,$F8,$83
	.dc.b	$6D,$D0,$FC,$47,$EE,$C2,$74,$D3,$A9,$FA,$18,$FC,$1C,$F1,$E4,$CC
	.dc.b	$FF,$7E,$38,$1F,$A9,$9E,$B5,$FE,$35,$8D,$35,$FF,$1A,$B7,$0F,$F5
	.dc.b	$6F,$B1,$EC,$F1,$25,$70,$7F,$51,$12,$30,$43,$FD,$6E,$FE,$93,$EB
	.dc.b	$BF,$61,$1F,$CA,$1A,$FF,$33,$6B,$FC,$F1,$A9,$F8,$A4,$C7,$F5,$FF
	.dc.b	$5F,$F8,$69,$34,$9C,$1F,$C7,$71,$7E,$F7,$FD,$6E,$BE,$37,$DC,$3E
	.dc.b	$3F,$F7,$2D,$8C,$C7,$EF,$3C,$FF,$E2,$F7,$FE,$55,$6F,$0F,$F4,$B6
	.dc.b	$D7,$F5,$46,$AB,$7C,$C7,$F0,$FF,$B1,$EE,$C9,$FF,$7A,$C7,$F0,$AE
	.dc.b	$27,$05,$24,$FD,$DF,$D2,$1E,$5E,$23,$FC,$DB,$5C,$1F,$C1,$35,$FF
	.dc.b	$38,$CF,$07,$94,$F1,$D3,$FC,$E1,$5B,$7F,$CE,$55,$BA,$FF,$78,$FC
	.dc.b	$57,$F2,$72,$3F,$E4,$CF,$25,$FB,$67,$EE,$17,$8B,$FE,$C8,$9D,$35
	.dc.b	$E6,$71,$9E,$03,$AF,$FA,$5B,$63,$51,$8F,$A3,$CB,$23,$27,$38,$E3
	.dc.b	$E1,$C8,$EC,$FF,$5F,$F6,$1F,$B4,$FB,$75,$FF,$55,$67,$EA,$AE,$17
	.dc.b	$FE,$2C,$CB,$FA,$AA,$C3,$8E,$FA,$C3,$8D,$6B,$0E,$57,$FE,$52,$C7
	.dc.b	$B7,$2F,$F9,$22,$BF,$CD,$B3,$FD,$77,$F5,$2F,$C8,$3F,$E4,$F6,$E1
	.dc.b	$D3,$DF,$F8,$67,$F6,$18,$CD,$FF,$88,$BC,$04,$62,$FF,$BD,$E1,$D9
	.dc.b	$F5,$7B,$1F,$33,$EA,$24,$7D,$18,$F9,$E6,$1F,$81,$18,$7F,$86,$7E
	.dc.b	$2C,$FA,$5F,$EB,$ED,$8C,$C7,$F6,$2F,$AE,$7C,$3E,$A2,$4E,$8F,$E2
	.dc.b	$D3,$A6,$CE,$9F,$F6,$3D,$75,$7F,$D9,$33,$B7,$EA,$7C,$0F,$E1,$88
	.dc.b	$FE,$2F,$E9,$AB,$A3,$F1,$D1,$F8,$C8,$FE,$EF,$F7,$7F,$CC,$3F,$35
	.dc.b	$7C,$07,$F7,$A6,$DF,$F2,$B9,$FE,$6A,$6E,$74,$72,$F2,$3E,$1F,$DA
	.dc.b	$CA,$F8,$EF,$F8,$93,$F0,$9F,$F6,$E7,$37,$FB,$77,$54,$EE,$BD,$EE
	.dc.b	$F9,$64,$72,$21,$FE,$7E,$BE,$71,$F8,$BE,$9F,$F1,$27,$EA,$7E,$8A
	.dc.b	$FF,$4C,$37,$1F,$CA,$CF,$EB,$0D,$E7,$2B,$F6,$51,$FB,$0F,$B4,$E7
	.dc.b	$F1,$52,$7F,$6D,$19,$7F,$C6,$98,$27,$51,$FA,$79,$F1,$FC,$D4,$8D
	.dc.b	$BF,$E3,$6F,$BD,$91,$82,$1F,$E8,$67,$FC,$77,$5A,$8C,$1F,$C3,$FD
	.dc.b	$FF,$03,$63,$59,$FE,$EA,$55,$CA,$BC,$FF,$27,$F1,$88,$F3,$F8,$DF
	.dc.b	$F2,$49,$87,$E4,$C6,$1F,$F5,$47,$E1,$6C,$FF,$CC,$2D,$8C,$C7,$F9
	.dc.b	$4B,$EE,$A4,$60,$87,$FA,$89,$CB,$FE,$AF,$9E,$4D,$9C,$E5,$EB,$AE
	.dc.b	$F1,$27,$E1,$47,$8B,$E1,$E4,$79,$9F,$F3,$1E,$B2,$30,$43,$FD,$66
	.dc.b	$FF,$96,$3E,$25,$D2,$C7,$FD,$5B,$F3,$4F,$AB,$FF,$54,$53,$B3,$FD
	.dc.b	$55,$39,$C7,$FD,$65,$18,$AD,$E1,$CE,$55,$B5,$7E,$0C,$3B,$DF,$6C
	.dc.b	$39,$57,$43,$27,$FA,$FF,$0E,$98,$73,$C3,$94,$F6,$BE,$4F,$03,$BD
	.dc.b	$F4,$7C,$37,$CA,$7C,$F8,$6E,$FD,$D5,$6F,$84,$76,$C3,$C1,$87,$1A
	.dc.b	$F3,$1E,$12,$C6,$AF,$A4,$EE,$7C,$67,$53,$31,$1E,$46,$8B,$5E,$46
	.dc.b	$CF,$38,$F2,$61,$F0,$27,$78,$7A,$2B,$46,$DF,$59,$9B,$F0,$61,$F2
	.dc.b	$E3,$35,$8B,$F8,$AF,$43,$D0,$78,$AB,$4F,$C0,$8F,$15,$7C,$18,$D1
	.dc.b	$FE,$ED,$F2,$F9,$FA,$FF,$DF,$63,$E9,$DF,$FB,$AE,$72,$7F,$BB,$A1
	.dc.b	$F4,$19,$7F,$DE,$2D,$7F,$DF,$8D,$47,$03,$C0,$70,$3E,$68,$F6,$DD
	.dc.b	$5F,$87,$8E,$33,$67,$31,$C0,$E9,$8F,$E6,$DE,$55,$C5,$CE,$5E,$27
	.dc.b	$C7,$5C,$AB,$84,$68,$D4,$66,$7E,$05,$F8,$6F,$A3,$6B,$C4,$DB,$E6
	.dc.b	$DA,$E3,$3E,$EB,$E8,$FC,$1A,$F4,$46,$23,$E4,$9B,$BE,$ED,$3B,$96
	.dc.b	$7B,$B6,$BB,$C7,$49,$E2,$D9,$E2,$D3,$89,$67,$89,$F1,$E3,$E5,$9C
	.dc.b	$1C,$57,$9E,$3E,$A9,$E1,$3D,$91,$AA,$F0,$B6,$3F,$69,$E5,$B9,$DF
	.dc.b	$3F,$F7,$34,$FB,$BA,$B9,$7F,$D0,$99,$7F,$4F,$AD,$3F,$8C,$9D,$7B
	.dc.b	$38,$C9,$18,$9F,$E8,$F3,$9C,$BC,$17,$BA,$D5,$FF,$2F,$BF,$CB,$D7
	.dc.b	$97,$EE,$B5,$F8,$33,$E1,$F9,$6A,$9C,$C6,$1B,$7F,$D9,$5B,$87,$51
	.dc.b	$87,$59,$FD,$2E,$BF,$03,$1E,$D1,$C4,$C9,$EF,$73,$5C,$F2,$46,$2B
	.dc.b	$96,$3E,$3B,$E3,$7F,$12,$B5,$1E,$EC,$BF,$C4,$2F,$F8,$2B,$EA,$3D
	.dc.b	$BE,$1F,$CB,$F3,$DE,$0A,$70,$DA,$E0,$7F,$80,$EB,$33,$F1,$BE,$4B
	.dc.b	$E3,$F7,$DA,$FF,$04,$37,$3F,$89,$BF,$E0,$E3,$1B,$7F,$82,$5E,$35
	.dc.b	$EA,$CD,$E8,$BE,$47,$38,$FC,$D9,$CC,$7A,$67,$55,$C5,$B5,$E0,$D9
	.dc.b	$86,$EF,$F0,$DE,$8B,$F0,$3E,$5F,$51,$86,$E5,$FE,$A7,$6E,$3F,$E9
	.dc.b	$56,$E9,$FE,$92,$FA,$CF,$17,$CC,$74,$6D,$FF,$AB,$DB,$7F,$E9,$D6
	.dc.b	$D7,$FA,$8B,$F0,$AF,$9D,$3F,$89,$3F,$98,$FC,$67,$E6,$E3,$33,$FC
	.dc.b	$92,$73,$7F,$80,$69,$C7,$C0,$5F,$1E,$4F,$D4,$4E,$AF,$95,$6A,$F9
	.dc.b	$36,$7E,$BE,$1E,$09,$FE,$7F,$87,$2F,$A3,$7C,$B5,$F4,$F0,$DA,$F2
	.dc.b	$4E,$9B,$3A,$69,$FE,$D3,$FA,$AF,$A6,$7D,$4F,$B0,$B1,$92,$65,$FE
	.dc.b	$C5,$8C,$BF,$1F,$F3,$07,$C1,$FB,$BE,$AD,$EF,$7C,$09,$CE,$1F,$1E
	.dc.b	$72,$D9,$E0,$6E,$AC,$64,$97,$DE,$32,$FF,$B5,$FC,$FF,$C5,$FC,$4F
	.dc.b	$E3,$1B,$5F,$EE,$F6,$CE,$E3,$FD,$E5,$3F,$92,$DB,$F3,$57,$67,$F3
	.dc.b	$4D,$AF,$D0,$76,$7A,$9F,$F7,$35,$70,$AF,$F7,$E7,$8F,$F9,$C3,$F9
	.dc.b	$37,$C1,$3F,$2C,$7C,$AE,$6B,$64,$C3,$F8,$C5,$6E,$75,$1F,$EF,$3B
	.dc.b	$BC,$FF,$DD,$2B,$57,$C6,$3F,$DE,$7C,$7F,$AC,$3F,$5B,$F0,$16,$7F
	.dc.b	$8A,$F8,$6E,$F5,$67,$2F,$F8,$01,$77,$0F,$F0,$0D,$FE,$6A,$7F,$9E
	.dc.b	$74,$7F,$83,$B1,$E6,$FF,$58,$AD,$B9,$9F,$E1,$ED,$8C,$C7,$2F,$3B
	.dc.b	$9B,$F8,$7C,$3F,$4B,$1F,$EA,$DB,$FC,$34,$63,$D9,$91,$FA,$F9,$D5
	.dc.b	$7E,$DF,$9B,$FD,$E3,$39,$AD,$CF,$EE,$EB,$78,$7E,$E6,$F7,$39,$AF
	.dc.b	$6F,$95,$FD,$77,$F0,$C7,$E1,$8F,$AF,$CE,$7F,$CF,$A7,$31,$FB,$CF
	.dc.b	$C3,$1E,$1D,$39,$5F,$1B,$FE,$0D,$3A,$7F,$97,$79,$DF,$CB,$9B,$D3
	.dc.b	$C5,$A7,$F3,$3F,$BD,$7C,$5F,$F4,$76,$C7,$F5,$0F,$C4,$9F,$6E,$FC
	.dc.b	$07,$57,$FC,$43,$EC,$3F,$98,$DF,$F1,$76,$F1,$7F,$3B,$C7,$9B,$FC
	.dc.b	$86,$33,$C9,$FC,$7D,$DC,$7F,$85,$DA,$FF,$8F,$DA,$FF,$94,$9B,$D1
	.dc.b	$FC,$8B,$E2,$FF,$90,$FD,$AF,$FA,$AF,$1F,$C4,$C6,$3D,$E9,$1F,$CA
	.dc.b	$E7,$58,$7E,$2A,$B2,$FF,$35,$8C,$E3,$FC,$CD,$B5,$FE,$0C,$EB,$5F
	.dc.b	$0A,$FF,$2D,$FA,$8D,$3F,$DB,$97,$DF,$DF,$F9,$9F,$07,$E6,$64,$64
	.dc.b	$FA,$51,$8C,$FC,$53,$70,$EB,$7D,$27,$37,$FD,$17,$F1,$17,$A5,$FD
	.dc.b	$0B,$EB,$DF,$57,$FD,$12,$73,$5F,$D2,$E7,$2D,$9E,$06,$EA,$C6,$49
	.dc.b	$7F,$D6,$6F,$9B,$FD,$6E,$F9,$B9,$E4,$F1,$7C,$57,$85,$B7,$E1,$8B
	.dc.b	$3B,$8C,$11,$E9,$1C,$63,$19,$FF,$65,$FC,$6F,$FC,$43,$5F,$E9,$4C
	.dc.b	$1E,$E7,$F6,$29,$CC,$A7,$37,$F0,$6B,$E3,$1B,$F0,$6C,$64,$95,$FD
	.dc.b	$DE,$F9,$DF,$28,$C7,$D1,$FC,$FF,$73,$C0,$DB,$F9,$D1,$67,$71,$83
	.dc.b	$F8,$EF,$D8,$7F,$20,$3F,$90,$D5,$8C,$92,$7F,$8F,$C6,$5F,$F1,$07
	.dc.b	$E0,$9F,$E3,$DF,$D5,$5F,$55,$FE,$5C,$8C,$36,$FF,$C6,$13,$A7,$F9
	.dc.b	$BF,$C0,$AB,$71,$FC,$EB,$E5,$FF,$98,$3F,$E7,$5F,$9C,$7A,$FF,$E7
	.dc.b	$19,$CE,$5F,$E7,$69,$CD,$71,$F3,$BF,$CE,$91,$89,$FF,$22,$D7,$0F
	.dc.b	$86,$4E,$79,$3F,$D0,$ED,$C3,$C1,$8F,$7C,$79,$BF,$E7,$FF,$8D,$FE
	.dc.b	$5F,$3E,$5F,$4B,$1E,$87,$81,$B1,$FE,$68,$CB,$BC,$63,$4F,$F3,$86
	.dc.b	$5F,$D5,$AB,$FC,$5D,$F5,$CF,$A1,$FF,$6D,$57,$FB,$9E,$31,$A7,$FB
	.dc.b	$A6,$11,$87,$53,$C1,$FF,$58,$66,$F9,$47,$D5,$71,$8C,$36,$33,$3F
	.dc.b	$EF,$59,$CB,$E1,$F8,$9F,$F5,$1B,$75,$F1,$63,$FE,$69,$6C,$EA,$7F
	.dc.b	$DC,$D0,$FA,$D7,$C0,$7F,$BA,$78,$BF,$D6,$B1,$8A,$F8,$CD,$7F,$D2
	.dc.b	$9F,$5C,$EA,$E1,$C1,$FE,$B7,$AD,$5F,$2C,$33,$87,$FA,$E2,$73,$5F
	.dc.b	$EB,$DC,$39,$4F,$6F,$80,$F9,$59,$79,$5B,$7F,$E5,$09,$FF,$6C,$1C
	.dc.b	$E3,$27,$69,$E1,$1C,$CF,$F0,$8F,$D7,$3E,$83,$57,$CE,$7F,$D5,$CD
	.dc.b	$CF,$E0,$E1,$E8,$F4,$18,$71,$BF,$F4,$B2,$F9,$CF,$6F,$9B,$A6,$C6
	.dc.b	$67,$F5,$7D,$F8,$99,$1C,$9E,$DD,$DF,$B6,$8F,$F1,$3F,$D0,$BE,$AB
	.dc.b	$AF,$82,$F9,$BA,$79,$67,$55,$C2,$B2,$FE,$F2,$75,$1F,$4A,$3E,$16
	.dc.b	$1F,$97,$69,$F9,$9E,$0F,$43,$F3,$9A,$88,$70,$3F,$D3,$0F,$0D,$FF
	.dc.b	$D2,$CD,$E3,$E1,$D4,$FA,$DF,$F7,$75,$FF,$A6,$6B,$FC,$AA,$FF,$B6
	.dc.b	$A5,$FB,$23,$E5,$FF,$13,$18,$7D,$55,$C2,$BC,$3F,$11,$EB,$8C,$47
	.dc.b	$98,$FC,$04,$64,$8F,$85,$E6,$76,$C1,$E3,$FC,$7F,$AE,$FD,$46,$8D
	.dc.b	$3D,$7E,$07,$FA,$3B,$1F,$43,$5F,$F5,$D5,$70,$9F,$8E,$FE,$C6,$3C
	.dc.b	$8D,$73,$D5,$F4,$E3,$11,$F3,$DE,$4D,$8F,$21,$F1,$3E,$BB,$FC,$CD
	.dc.b	$1E,$0F,$9D,$F8,$F8,$7F,$0D,$8F,$2B,$EB,$89,$1F,$C5,$ED,$1E,$12
	.dc.b	$46,$5D,$BD,$4F,$AD,$5F,$90,$7F,$DB,$1F,$5D,$F3,$F2,$FF,$0F,$F6
	.dc.b	$FF,$27,$9F,$F7,$E3,$5D,$9F,$93,$9F,$C9,$26,$91,$FC,$B3,$D7,$3F
	.dc.b	$C8,$B6,$3F,$28,$DE,$BF,$E9,$3C,$9D,$7C,$5F,$DA,$E0,$FA,$BA,$67
	.dc.b	$0F,$F7,$56,$1F,$CD,$5F,$C4,$35,$EF,$5E,$22,$3C,$1E,$4B,$2F,$F4
	.dc.b	$CF,$4B,$EA,$BF,$61,$F5,$BF,$E6,$AA,$FF,$7C,$35,$FF,$80,$0D,$AF
	.dc.b	$F7,$F7,$D7,$7E,$A6,$7F,$60,$FE,$C9,$FD,$8B,$5C,$E9,$FB,$B8,$C4
	.dc.b	$2B,$F1,$2A,$3A,$36,$7C,$BC,$7F,$CD,$BE,$11,$FA,$13,$C5,$C3,$6F
	.dc.b	$FD,$00,$D7,$E0,$71,$7F,$04,$E0,$7C,$8F,$F0,$FF,$33,$F9,$75,$7E
	.dc.b	$CD,$AF,$F0,$A6,$BF,$C3,$5A,$FD,$7F,$43,$F8,$C5,$7F,$83,$A3,$F8
	.dc.b	$96,$FF,$B3,$7F,$8A,$46,$5F,$18,$F1,$DC,$4F,$AE,$FE,$45,$E7,$FD
	.dc.b	$5A,$E7,$A7,$F1,$C8,$FD,$5E,$3D,$67,$FC,$27,$F0,$5C,$39,$64,$7F
	.dc.b	$7F,$AE,$4F,$F5,$46,$BF,$C4,$5A,$F5,$8E,$2F,$98,$FF,$0F,$BF,$CD
	.dc.b	$7E,$BB,$F7,$11,$EA,$CF,$D3,$F2,$BF,$D4,$F0,$7D,$25,$3F,$A1,$44
	.dc.b	$AF,$F7,$2C,$65,$FE,$97,$1F,$D6,$FE,$53,$F7,$F1,$FB,$4D,$7F,$4A
	.dc.b	$67,$BE,$46,$48,$78,$E7,$8B,$7C,$5F,$EC,$3E,$96,$37,$FF,$4F,$B7
	.dc.b	$83,$FA,$46,$BF,$F0,$06,$BC,$47,$0C,$66,$1F,$48,$FB,$18,$FF,$6F
	.dc.b	$7B,$19,$F0,$18,$F5,$84,$62,$BA,$C6,$FD,$7F,$C9,$BD,$71,$6F,$FE
	.dc.b	$A2,$8C,$35,$FC,$4D,$AB,$8C,$61,$BA,$7F,$AF,$1B,$A7,$F6,$5A,$E1
	.dc.b	$CF,$F3,$FE,$33,$E8,$7D,$6F,$C6,$F8,$A9,$97,$8F,$74,$EA,$3D,$DB
	.dc.b	$AB,$71,$F4,$67,$FB,$D7,$D5,$FF,$B3,$A3,$E4,$0E,$57,$C9,$7E,$2C
	.dc.b	$DF,$C5,$C9,$FF,$2B,$57,$17,$FC,$56,$D7,$FB,$43,$5F,$EF,$EE,$F8
	.dc.b	$3F,$48,$DD,$F3,$5F,$E0,$DF,$AD,$7C,$BF,$E3,$F4,$FC,$2F,$C6,$CC
	.dc.b	$BE,$04,$65,$F7,$C6,$5F,$C7,$47,$F8,$5F,$EB,$5F,$11,$7B,$25,$7C
	.dc.b	$09,$CB,$67,$81,$1F,$8B,$F0,$3F,$C8,$35,$7D,$2F,$F8,$8F,$F3,$47
	.dc.b	$D6,$D3,$E6,$C4,$BF,$8E,$FF,$89,$EB,$8C,$60,$87,$EE,$DC,$B6,$3F
	.dc.b	$C9,$5F,$5D,$FC,$2F,$8B,$F8,$C7,$53,$EA,$7D,$37,$FB,$FE,$DB,$5F
	.dc.b	$C7,$7B,$BF,$2F,$B7,$F9,$3A,$F3,$CD,$F8,$F6,$E1,$FA,$26,$E3,$D6
	.dc.b	$BF,$80,$7A,$BC,$7E,$1F,$F8,$57,$C0,$CC,$3F,$59,$82,$75,$18,$6B
	.dc.b	$FC,$07,$9B,$C5,$F5,$BF,$E9,$FC,$BD,$4D,$B3,$49,$5D,$9E,$2F,$FB
	.dc.b	$3B,$8B,$FC,$C2,$47,$07,$FB,$4B,$0C,$3F,$C8,$FE,$0B,$EB,$BA,$31
	.dc.b	$1F,$17,$E2,$BC,$44,$8F,$F4,$3E,$3F,$3E,$31,$1F,$BE,$C5,$1C,$63
	.dc.b	$FD,$C3,$F0,$5F,$E8,$BF,$AC,$69,$FE,$94,$6C,$FF,$A4,$BD,$44,$62
	.dc.b	$3F,$CA,$11,$89,$E0,$73,$24,$7F,$BB,$BE,$BA,$BF,$D6,$27,$83,$FE
	.dc.b	$D5,$AE,$9C,$9F,$83,$D5,$7C,$23,$FB,$15,$F0,$BE,$57,$C3,$8F,$F9
	.dc.b	$8E,$1C,$27,$A7,$C4,$7A,$C8,$FA,$EB,$FD,$7C,$FF,$B1,$9F,$F5,$BA
	.dc.b	$BF,$7C,$FC,$53,$FD,$33,$C3,$FE,$8D,$F8,$B7,$CB,$7F,$88,$ED,$F5
	.dc.b	$DF,$D6,$1F,$F4,$54,$FF,$1D,$7F,$D1,$91,$95,$19,$78,$F8,$1F,$E9
	.dc.b	$08,$EB,$AF,$08,$C5,$F7,$F3,$79,$57,$16,$F8,$5F,$E9,$87,$33,$BA
	.dc.b	$E3,$7C,$F0,$F2,$DF,$FA,$7B,$1E,$31,$93,$CB,$7E,$D8,$E7,$59,$C7
	.dc.b	$D1,$1C,$2B,$D2,$7C,$0A,$E1,$87,$82,$3F,$CE,$2F,$B7,$0F,$33,$63
	.dc.b	$FD,$61,$8F,$67,$67,$3C,$7C,$15,$E7,$AE,$13,$E0,$3C,$F3,$E5,$BF
	.dc.b	$0C,$7B,$8E,$26,$62,$3C,$71,$F5,$D7,$4C,$7C,$47,$8F,$1F,$71,$23
	.dc.b	$D8,$E9,$E0,$4A,$F4,$1D,$70,$FA,$4F,$8C,$F6,$9F,$82,$C7,$E1,$BC
	.dc.b	$2B,$C1,$5E,$E7,$C1,$19,$8F,$99,$A7,$F8,$5B,$1F,$C3,$CF,$E6,$6B
	.dc.b	$24,$7F,$D8,$DF,$55,$FE,$DD,$AF,$F0,$EC,$7F,$B7,$E3,$13,$FE,$CB
	.dc.b	$84,$61,$E2,$75,$2C,$7E,$6D,$7B,$DF,$E5,$78,$C4,$72,$3D,$E6,$B8
	.dc.b	$7E,$FE,$3D,$EE,$CF,$FB,$C2,$F9,$4E,$63,$8C,$61,$F1,$E9,$E1,$8F
	.dc.b	$F3,$33,$E0,$AF,$CD,$4F,$73,$F0,$C7,$BC,$E7,$7E,$07,$FD,$FF,$15
	.dc.b	$E1,$87,$37,$8D,$F6,$C3,$C9,$87,$87,$0E,$F5,$EF,$8C,$3E,$03,$13
	.dc.b	$FE,$F7,$AE,$51,$EA,$8C,$57,$2C,$3D,$35,$AA,$F2,$9E,$83,$E5,$9F
	.dc.b	$37,$0E,$E7,$9F,$0F,$09,$70,$EB,$3E,$3B,$EB,$87,$78,$F6,$B6,$F9
	.dc.b	$D7,$D1,$6C,$F9,$CF,$AD,$3C,$6F,$BD,$7E,$37,$83,$FD,$4B,$3F,$CF
	.dc.b	$7C,$EF,$F4,$D6,$FF,$CF,$7E,$1A,$CE,$E6,$E3,$FC,$C6,$F8,$BF,$D3
	.dc.b	$9A,$FF,$37,$8F,$F4,$C3,$C6,$7F,$AB,$7C,$EB,$E4,$F9,$EE,$8C,$4F
	.dc.b	$F9,$97,$95,$F5,$DA,$9C,$E1,$FD,$55,$BA,$7F,$50,$6D,$FF,$4E,$AE
	.dc.b	$18,$C8,$FF,$3E,$63,$DE,$3D,$7E,$13,$2E,$F5,$AC,$39,$E5,$FD,$3D
	.dc.b	$F0,$A3,$F6,$CE,$1B,$7F,$01,$E0,$DE,$DF,$F1,$77,$D6,$7C,$A6,$DF
	.dc.b	$BF,$C6,$C7,$7A,$A9,$C5,$C4,$FF,$82,$9B,$87,$8C,$FF,$19,$4F,$87
	.dc.b	$3F,$67,$7F,$BB,$4F,$F0,$6E,$1F,$37,$E5,$B1,$32,$3E,$83,$B8,$C6
	.dc.b	$1F,$E9,$2F,$53,$FD,$3F,$5F,$EA,$16,$C7,$F8,$8E,$BD,$91,$8E,$1F
	.dc.b	$F2,$6B,$A9,$F3,$6A,$8C,$47,$FA,$5D,$AF,$FA,$A9,$B1,$BF,$AE,$EA
	.dc.b	$7D,$8F,$55,$F5,$D8,$CB,$C9,$3F,$73,$CA,$FF,$35,$C6,$0F,$F5,$6F
	.dc.b	$8F,$E6,$F9,$B9,$1F,$EA,$1F,$52,$E9,$EF,$9E,$5E,$4E,$36,$FF,$31
	.dc.b	$63,$FE,$55,$C3,$FD,$3F,$E0,$7F,$AE,$9B,$1F,$8C,$FA,$EF,$AB,$5F
	.dc.b	$EF,$9A,$D6,$3F,$EE,$F6,$BF,$D0,$2B,$F1,$0F,$F3,$C6,$C6,$A3,$F9
	.dc.b	$0C,$7E,$C9,$7D,$5B,$DC,$C5,$76,$B6,$3F,$85,$AD,$3F,$EB,$6F,$AE
	.dc.b	$7A,$9C,$DB,$19,$F9,$7C,$4C,$8F,$F5,$EF,$81,$FE,$98,$79,$47,$8E
	.dc.b	$31,$9F,$FB,$D7,$EB,$BF,$3D,$E1,$B9,$3E,$C4,$61,$B8,$7F,$BA,$E7
	.dc.b	$4F,$FB,$4E,$74,$FF,$B3,$3E,$A9,$F5,$BF,$05,$B1,$FE,$D0,$D5,$F0
	.dc.b	$59,$A9,$3F,$D9,$F2,$5F,$07,$FB,$53,$5F,$D5,$1A,$9F,$A7,$1B,$DF
	.dc.b	$E5,$CF,$F0,$18,$6D,$DA,$13,$A7,$BC,$7E,$5F,$E1,$FE,$1F,$35,$C7
	.dc.b	$E0,$36,$BF,$DF,$FA,$A3,$18,$7F,$AD,$A3,$15,$FB,$1C,$FF,$5E,$7E
	.dc.b	$8F,$1B,$19,$3F,$8B,$57,$E6,$3D,$4C,$E3,$FE,$BD,$7F,$55,$E4,$7D
	.dc.b	$6E,$3B,$7F,$4D,$B5,$E4,$6D,$F8,$DB,$3F,$C8,$FB,$7F,$D2,$31,$8A
	.dc.b	$46,$27,$C6,$7F,$29,$F0,$9D,$DF,$90,$AF,$83,$F0,$9F,$BD,$E7,$FC
	.dc.b	$4B,$6F,$F0,$CD,$9D,$C7,$45,$6E,$AF,$E6,$D3,$98,$C7,$A9,$FC,$16
	.dc.b	$BF,$83,$36,$75,$5F,$E9,$C6,$CE,$9E,$CF,$F8,$AB,$B7,$F6,$5E,$7F
	.dc.b	$F0,$38,$FC,$0C,$FF,$33,$E6,$7B,$BA,$B9,$3F,$93,$BE,$17,$FE,$0F
	.dc.b	$9F,$E0,$31,$88,$FF,$17,$47,$6F,$27,$8C,$76,$3F,$A9,$4F,$EC,$38
	.dc.b	$7F,$2B,$F5,$6F,$AA,$7D,$5F,$F7,$88,$C3,$FC,$97,$CD,$C4,$EB,$3F
	.dc.b	$6F,$D5,$BC,$6F,$E2,$6D,$8F,$D5,$37,$3F,$F2,$93,$5F,$EA,$5F,$1E
	.dc.b	$FD,$CF,$FD,$E3,$5F,$56,$31,$ED,$7F,$61,$F4,$BF,$A1,$4E,$6B,$77
	.dc.b	$ED,$EA,$7D,$43,$EA,$FF,$90,$7D,$5F,$F5,$AA,$D4,$E7,$0F,$AB,$39
	.dc.b	$E2,$FD,$93,$BA,$FD,$EB,$5F,$ED,$0D,$7F,$BE,$36,$1F,$72,$EC,$FF
	.dc.b	$10,$C6,$1F,$1F,$A5,$89,$D6,$1F,$86,$FA,$C7,$BD,$FC,$DD,$B1,$F8
	.dc.b	$EF,$8A,$C4,$C8,$FE,$2D,$3A,$CF,$FC,$FC,$D7,$F8,$AB,$5F,$F0,$03
	.dc.b	$BC,$BF,$8A,$ED,$FE,$2B,$F5,$3A,$57,$CB,$E1,$FE,$CF,$87,$97,$E0
	.dc.b	$3C,$87,$F8,$4F,$81,$F0,$58,$9D,$1E,$49,$FE,$AB,$9F,$FB,$B2,$30
	.dc.b	$4C,$3B,$79,$BC,$5B,$1F,$A8,$FA,$87,$D6,$7F,$59,$6C,$7E,$B3,$EB
	.dc.b	$3F,$8E,$D7,$E7,$44,$E9,$C3,$5F,$E6,$ED,$7F,$C4,$4D,$7F,$C6,$CD
	.dc.b	$79,$D7,$F9,$6B,$E8,$DF,$12,$F9,$AF,$24,$E6,$31,$8F,$A5,$F4,$1F
	.dc.b	$DC,$71,$7D,$66,$27,$47,$F4,$FC,$BA,$B6,$BF,$BE,$C6,$32,$FF,$5D
	.dc.b	$37,$8F,$FA,$6E,$7E,$4F,$BB,$E9,$5F,$BF,$8C,$61,$F2,$FE,$03,$E5
	.dc.b	$1F,$E0,$3F,$AD,$FF,$2B,$D7,$C8,$8C,$CF,$8D,$F0,$91,$C1,$FC,$A6
	.dc.b	$9F,$AF,$BE,$B3,$8D,$F8,$63,$15,$F6,$FB,$3E,$D1,$E1,$FA,$6B,$19
	.dc.b	$3E,$14,$7C,$02,$3F,$ED,$9F,$47,$F9,$B6,$1F,$E6,$3A,$FB,$DA,$FF
	.dc.b	$8B,$A3,$1C,$5F,$E3,$B6,$DF,$F3,$66,$E5,$FC,$1F,$EA,$DF,$21,$FE
	.dc.b	$6D,$6D,$7F,$A2,$FE,$BB,$13,$23,$FA,$47,$72,$B7,$7F,$CB,$74,$FF
	.dc.b	$42,$61,$B8,$FE,$61,$F1,$6F,$9D,$75,$7F,$A6,$9B,$97,$F4,$FF,$9C
	.dc.b	$FF,$6E,$6F,$FE,$08,$BE,$17,$B2,$56,$FD,$1B,$8F,$6C,$79,$BF,$E1
	.dc.b	$4B,$E3,$5F,$F0,$05,$B1,$99,$DB,$FE,$83,$8F,$37,$B1,$E5,$3F,$D5
	.dc.b	$FF,$25,$FD,$5F,$2F,$25,$FF,$71,$6C,$EF,$0F,$F5,$3D,$6D,$FF,$50
	.dc.b	$5F,$47,$FB,$1F,$D7,$3E,$85,$9F,$F8,$BE,$30,$D8,$CE,$7F,$E2,$E1
	.dc.b	$3A,$71,$83,$E5,$9D,$3F,$E3,$98,$C3,$63,$33,$C1,$AF,$F8,$EA,$9F
	.dc.b	$3C,$CF,$C8,$D9,$F1,$4E,$A7,$27,$CD,$F9,$DF,$F2,$1E,$DF,$65,$BA
	.dc.b	$FC,$26,$DF,$5F,$A7,$7C,$CF,$FA,$CE,$F8,$4D,$BE,$0D,$3C,$97,$CB
	.dc.b	$06,$1C,$A7,$FD,$AF,$3F,$EE,$77,$C9,$EC,$7F,$96,$E9,$39,$9F,$C3
	.dc.b	$35,$FC,$13,$9C,$3D,$55,$E1,$AE,$2D,$BF,$F2,$DD,$F8,$23,$57,$C7
	.dc.b	$0F,$81,$1B,$F8,$5F,$57,$4F,$EA,$7D,$DB,$D4,$69,$AF,$AE,$FE,$5D
	.dc.b	$97,$EC,$27,$4D,$9D,$BF,$B4,$9D,$BF,$BD,$F8,$6F,$91,$D9,$FE,$3D
	.dc.b	$F6,$2F,$C1,$1F,$0C,$CA,$FC,$76,$5F,$BD,$8F,$F4,$13,$63,$39,$27
	.dc.b	$FD,$0D,$CC,$FA,$DC,$57,$E5,$E3,$13,$A9,$CC,$FF,$A2,$27,$2E,$E3
	.dc.b	$3C,$7F,$CF,$1B,$C0,$9F,$CE,$C6,$1C,$7D,$17,$C0,$AE,$1F,$33,$EE
	.dc.b	$24,$FF,$37,$A4,$7E,$2F,$0F,$43,$E5,$24,$F8,$DA,$FF,$4B,$9E,$2F
	.dc.b	$38,$FC,$A2,$B7,$D6,$F5,$EA,$FF,$58,$CF,$F2,$8F,$C0,$FF,$54,$FF
	.dc.b	$8D,$DB,$97,$F5,$A9,$DC,$FF,$59,$6B,$F0,$4F,$DE,$B8,$7E,$26,$5F
	.dc.b	$C0,$7E,$3F,$E9,$CE,$67,$71,$E3,$23,$FE,$74,$FA,$DF,$0E,$BC,$AB
	.dc.b	$F8,$54,$FF,$9F,$3E,$53,$F0,$E7,$AE,$24,$7D,$D7,$6F,$DB,$6B,$E7
	.dc.b	$C3,$F8,$A4,$79,$0E,$31,$ED,$7E,$4D,$BE,$7F,$F4,$BD,$7F,$C9,$6B
	.dc.b	$E4,$7B,$1F,$A8,$D7,$FC,$5F,$1F,$D4,$A7,$4F,$F9,$85,$AF,$C7,$E4
	.dc.b	$FE,$0B,$E1,$7E,$6A,$73,$3F,$EB,$16,$BF,$EB,$07,$9E,$FF,$E4,$46
	.dc.b	$ED,$BA,$FB,$FF,$59,$FC,$6B,$0F,$E2,$3F,$51,$FA,$CF,$8B,$99,$79
	.dc.b	$63,$D6,$78,$48,$FE,$BB,$14,$62,$3D,$4F,$07,$FD,$53,$7F,$D0,$9F
	.dc.b	$F4,$D6,$4F,$A2,$7C,$97,$F2,$58,$C3,$63,$33,$FE,$B2,$FA,$CF,$F4
	.dc.b	$23,$7E,$1F,$FD,$FC,$A3,$2F,$F4,$F6,$BF,$EF,$E6,$BF,$93,$FA,$CF
	.dc.b	$E1,$EF,$E5,$43,$87,$5E,$F7,$68,$C4,$7F,$B4,$3C,$07,$B1,$FC,$50
	.dc.b	$91,$FE,$A1,$FA,$B7,$D6,$7E,$BB,$7F,$5C,$FF,$63,$6B,$E1,$6B,$FA
	.dc.b	$76,$BF,$C2,$5F,$F0,$8B,$FE,$D4,$E3,$FE,$72,$FF,$9A,$5B,$BB,$93
	.dc.b	$97,$A5,$FB,$EF,$4B,$FD,$CD,$7F,$ED,$C5,$F3,$99,$35,$AC,$9F,$EF
	.dc.b	$07,$F8,$85,$F9,$F3,$CA,$3F,$BE,$35,$F9,$4E,$70,$F1,$47,$F2,$18
	.dc.b	$C6,$1F,$C0,$DB,$19,$F8,$AF,$E7,$BF,$4B,$23,$FD,$EA,$FF,$42,$FA
	.dc.b	$C7,$CD,$66,$7F,$C7,$19,$2B,$19,$78,$3C,$B6,$1C,$2F,$F6,$1F,$41
	.dc.b	$FD,$2B,$DF,$91,$82,$1F,$B3,$8F,$DD,$9F,$D4,$BE,$6B,$F7,$D1,$FE
	.dc.b	$4A,$F5,$EB,$E2,$F7,$3F,$AA,$9F,$D9,$A3,$39,$7F,$AD,$9D,$CF,$EF
	.dc.b	$5A,$FF,$45,$F0,$7F,$C3,$3E,$27,$F5,$89,$CC,$A7,$2F,$86,$7B,$1E
	.dc.b	$5F,$2F,$FB,$2F,$BF,$27,$FA,$86,$1F,$0F,$83,$FD,$67,$1F,$E6,$2F
	.dc.b	$77,$F2,$73,$9A,$E1,$3F,$31,$FE,$FF,$3A,$E5,$F3,$67,$FE,$DE,$AE
	.dc.b	$B1,$FD,$77,$85,$8B,$E9,$7F,$BB,$9F,$0F,$E3,$A5,$FC,$96,$BF,$E7
	.dc.b	$1E,$07,$D6,$7F,$54,$F6,$9F,$2D,$FD,$5B,$F5,$97,$97,$4D,$44,$C3
	.dc.b	$F2,$F5,$C1,$B1,$FC,$1A,$31,$1F,$E5,$36,$F0,$FE,$5A,$3F,$C5,$B3
	.dc.b	$CD,$FF,$6A,$7D,$53,$E4,$FE,$E7,$9F,$FE,$14,$FD,$BB,$86,$1F,$8F
	.dc.b	$6F,$17,$C6,$9D,$E1,$F8,$76,$BE,$FD,$FF,$B8,$FD,$6F,$E6,$63,$E5
	.dc.b	$9F,$E0,$77,$0F,$FB,$95,$B8,$79,$DF,$F6,$DC,$70,$0E,$AF,$FB,$83
	.dc.b	$63,$A7,$77,$83,$E3,$BF,$C0,$3F,$5A,$FA,$9F,$F1,$7F,$CB,$FF,$2A
	.dc.b	$CB,$F5,$4D,$CB,$FB,$D5,$6F,$1F,$03,$5F,$F3,$F3,$5F,$E5,$8D,$78
	.dc.b	$DF,$F3,$88,$E1,$7F,$45,$B5,$AA,$F6,$6F,$F8,$07,$E2,$F4,$7E,$3B
	.dc.b	$2F,$F2,$19,$23,$FD,$7B,$F5,$9F,$11,$FE,$0B,$19,$69,$A7,$FC,$97
	.dc.b	$F1,$92,$3A,$F5,$7F,$B3,$1E,$7F,$09,$97,$F8,$7C,$DD,$3F,$8E,$F8
	.dc.b	$2C,$7F,$CB,$B4,$9C,$FE,$1D,$F8,$98,$FF,$2A,$B5,$FE,$A1,$48,$C4
	.dc.b	$7F,$80,$5A,$FF,$05,$78,$4F,$4E,$6F,$F3,$9C,$63,$6F,$E1,$58,$7F
	.dc.b	$99,$6F,$E3,$FD,$6F,$CC,$7F,$85,$46,$5F,$34,$F0,$F8,$5F,$F3,$7D
	.dc.b	$6F,$73,$C7,$3F,$74,$FF,$A0,$B0,$E5,$9F,$F2,$FA,$D6,$5F,$47,$2F
	.dc.b	$56,$FF,$E7,$28,$C4,$7A,$E3,$07,$A8,$8F,$DA,$F1,$FF,$D3,$1F,$00
	.dc.b	$FB,$6F,$F5,$13,$67,$FD,$C8,$D7,$FD,$D0,$D7,$FD,$C8,$E7,$1F,$F7
	.dc.b	$A3,$76,$FF,$6E,$B6,$7F,$D9,$1F,$51,$FE,$36,$7F,$E0,$0C,$67,$E2
	.dc.b	$CD,$31,$3B,$BE,$B8,$7A,$09,$7F,$BC,$B5,$6F,$0F,$0F,$C8,$7B,$09
	.dc.b	$5E,$C6,$BE,$27,$E1,$9F,$EB,$B9,$FF,$64,$47,$DA,$3F,$D1,$DF,$1E
	.dc.b	$F9,$AF,$C9,$43,$EB,$FF,$85,$CF,$F3,$98,$C7,$6F,$F9,$D6,$31,$3A
	.dc.b	$AE,$2D,$D3,$FD,$35,$E6,$F2,$F0,$3F,$CF,$8D,$C7,$38,$F3,$C3,$C3
	.dc.b	$3C,$67,$B9,$E6,$AF,$F4,$26,$1C,$63,$23,$1F,$56,$3F,$EA,$99,$E4
	.dc.b	$6C,$98,$F7,$8D,$36,$BF,$CE,$8F,$B6,$FC,$ED,$8F,$39,$F2,$31,$F3
	.dc.b	$C7,$1B,$EB,$3E,$73,$A9,$BB,$AF,$9E,$BE,$3D,$75,$8F,$5E,$1F,$32
	.dc.b	$3A,$12,$3A,$1A,$C3,$E1,$C6,$63,$A5,$F9,$9F,$89,$3E,$A7,$8E,$3F
	.dc.b	$34,$F8,$B5,$E1,$C7,$D3,$3C,$8F,$1E,$3E,$E7,$8C,$F9,$EB,$93,$F2
	.dc.b	$5F,$A8,$47,$07,$CD,$D7,$FC,$43,$97,$D6,$A4,$62,$FF,$22,$DC,$BF
	.dc.b	$DC,$4F,$F8,$83,$EB,$6E,$3F,$E7,$1B,$46,$1F,$F3,$A3,$5F,$F6,$53
	.dc.b	$5F,$F7,$63,$F1,$57,$C9,$FF,$94,$27,$FB,$4C,$68,$FC,$19,$FE,$D3
	.dc.b	$8D,$65,$E1,$AE,$51,$89,$F0,$9C,$0F,$F7,$D3,$97,$C6,$7E,$60,$F0
	.dc.b	$EF,$F8,$47,$C1,$A7,$8A,$7F,$CC,$6F,$92,$3D,$E7,$53,$89,$4C,$E1
	.dc.b	$C6,$BD,$D1,$C2,$70,$F8,$4C,$61,$FE,$9B,$BE,$EF,$B6,$3C,$D1,$87
	.dc.b	$DB,$3C,$8F,$26,$1E,$99,$D4,$FB,$A7,$55,$CF,$0F,$09,$C2,$7D,$55
	.dc.b	$F1,$5F,$83,$3E,$3C,$3E,$24,$79,$8F,$A4,$E3,$0F,$03,$63,$CE,$DB
	.dc.b	$EF,$5E,$13,$A4,$FC,$7A,$E2,$FC,$36,$E1,$FE,$2D,$C5,$F3,$BF,$3B
	.dc.b	$3F,$F4,$AC,$63,$3F,$F4,$BC,$63,$7F,$F6,$AE,$9F,$CE,$6F,$83,$F9
	.dc.b	$A7,$EB,$E4,$31,$7F,$E4,$B6,$CF,$E7,$70,$EC,$DA,$FF,$4C,$3D,$73
	.dc.b	$F0,$BF,$A6,$F9,$E7,$C9,$E3,$0F,$D6,$DF,$07,$FA,$06,$1E,$CC,$34
	.dc.b	$DC,$3F,$A8,$61,$C9,$EC,$4A,$E3,$7C,$F3,$F8,$75,$B9,$CE,$1D,$2B
	.dc.b	$C1,$1E,$F3,$C7,$1E,$F6,$DF,$B6,$31,$D7,$FE,$1D,$F0,$FD,$4D,$9F
	.dc.b	$F1,$BF,$9D,$8E,$AF,$F2,$94,$7F,$98,$A7,$58,$7F,$89,$A7,$2D,$BE
	.dc.b	$B1,$88,$F2,$77,$7F,$86,$3A,$BB,$FD,$7B,$D0,$F9,$53,$9C,$BB,$4E
	.dc.b	$5F,$F1,$3D,$70,$AD,$F7,$23,$13,$E5,$C5,$F1,$FE,$93,$39,$7F,$73
	.dc.b	$D3,$FC,$79,$19,$AF,$F2,$0B,$E8,$23,$FC,$87,$3F,$F5,$4E,$9E,$6F
	.dc.b	$17,$DF,$C5,$FE,$4E,$DB,$FD,$EF,$7F,$E5,$6A,$D1,$1F,$5F,$B7,$73
	.dc.b	$FF,$72,$F8,$DC,$CF,$87,$F5,$96,$BF,$CB,$F1,$67,$D1,$82,$B1,$8F
	.dc.b	$FA,$97,$EB,$5E,$CF,$E4,$ED,$18,$C7,$FD,$25,$DB,$FE,$C2,$7F,$D2
	.dc.b	$91,$99,$FF,$0B,$1C,$2B,$F1,$0D,$9F,$E8,$4D,$3F,$A2,$4F,$17,$F2
	.dc.b	$06,$6F,$FD,$FD,$93,$E7,$FC,$95,$DE,$73,$1E,$7F,$91,$FF,$64,$E1
	.dc.b	$FE,$A7,$6E,$1E,$4F,$AD,$78,$BB,$F5,$65,$ED,$F8,$13,$A8,$C1,$F9
	.dc.b	$7F,$07,$EA,$DF,$E0,$5B,$92,$7F,$DC,$DE,$C3,$EB,$BF,$47,$5A,$9F
	.dc.b	$E1,$5D,$1F,$E4,$6E,$4F,$CB,$FA,$AF,$C2,$37,$18,$7E,$97,$D7,$3E
	.dc.b	$6E,$98,$F3,$AE,$3F,$EC,$59,$D3,$F3,$9B,$B7,$F7,$16,$BF,$39,$AF
	.dc.b	$EF,$DD,$ED,$FE,$E7,$FD,$C7,$FB,$FD,$75,$C3,$F0,$91,$8B,$F8,$5D
	.dc.b	$3E,$E3,$F6,$3C,$0F,$86,$C6,$1F,$26,$7D,$FF,$4D,$8B,$FF,$2E,$77
	.dc.b	$B8,$13,$FA,$5F,$9B,$98,$7E,$EB,$E1,$AE,$1C,$3C,$15,$63,$7F,$D5
	.dc.b	$55,$BF,$9F,$CA,$E7,$FA,$67,$9B,$E4,$EC,$FD,$D8,$FE,$37,$3B,$8F
	.dc.b	$F0,$C7,$3F,$F9,$F2,$72,$48,$FF,$6B,$EA,$F8,$AE,$7A,$FB,$23,$13
	.dc.b	$F8,$9F,$89,$FC,$19,$F1,$BB,$2C,$64,$E7,$E4,$E2,$F9,$3F,$C1,$FE
	.dc.b	$C2,$4F,$FB,$6A,$69,$F0,$E2,$EF,$F9,$38,$71,$B8,$BF,$AD,$5F,$FB
	.dc.b	$8A,$FF,$7D,$E8,$FE,$9B,$0E,$18,$7F,$8B,$BB,$F3,$E1,$E9,$98,$7E
	.dc.b	$EA,$FF,$5B,$5F,$E1,$39,$FD,$FC,$62,$3F,$C6,$51,$FC,$2F,$2F,$CF
	.dc.b	$57,$F6,$BC,$3B,$DF,$86,$3F,$9E,$F7,$7F,$63,$FA,$F7,$D7,$3E,$BF
	.dc.b	$F8,$A7,$D7,$BE,$1B,$53,$9C,$3A,$1F,$D6,$67,$4F,$F4,$FF,$51,$E6
	.dc.b	$7F,$58,$6C,$7D,$E6,$EF,$FE,$F8,$6B,$FD,$43,$99,$F2,$3F,$B8,$3F
	.dc.b	$BC,$7D,$7B,$EB,$DF,$05,$9C,$7C,$78,$7F,$82,$7E,$BF,$FA,$74,$7F
	.dc.b	$22,$6C,$7F,$99,$5A,$FF,$22,$6B,$F1,$FE,$BD,$D9,$FD,$CF,$95,$18
	.dc.b	$BF,$F3,$74,$62,$75,$39,$AF,$E5,$BD,$FF,$EB,$1F,$1B,$FB,$BF,$C0
	.dc.b	$7E,$DF,$1F,$3E,$3C,$5F,$EE,$D1,$9B,$FE,$C1,$C0,$F8,$9F,$EF,$ED
	.dc.b	$AF,$7C,$FE,$EB,$E0,$B3,$52,$7E,$73,$F4,$48,$E0,$FD,$5F,$9F,$BA
	.dc.b	$FD,$53,$DB,$29,$3D,$BC,$9F,$F7,$13,$F7,$63,$35,$F6,$5F,$BC,$48
	.dc.b	$FB,$CF,$F8,$5B,$F2,$27,$D5,$F0,$D7,$F5,$1B,$27,$37,$F6,$BE,$91
	.dc.b	$F0,$EF,$9C,$FF,$1C,$46,$1F,$ED,$1F,$5E,$F9,$5C,$D6,$EF,$93,$75
	.dc.b	$FE,$BF,$87,$19,$FF,$53,$36,$33,$3B,$7F,$C6,$91,$9C,$FF,$7D,$E2
	.dc.b	$7F,$29,$F0,$7B,$E7,$FD,$31,$F0,$FD,$7C,$CB,$EA,$E9,$EA,$7C,$71
	.dc.b	$9C,$BB,$BF,$E3,$FC,$BD,$ED,$CC,$D3,$E5,$7B,$FC,$74,$F9,$3B,$7F
	.dc.b	$8A,$31,$F9,$1F,$5C,$FA,$EF,$8E,$D8,$F1,$FC,$FE,$26,$47,$C2,$9D
	.dc.b	$69,$F1,$9A,$FC,$26,$BF,$E7,$5D,$11,$C3,$1F,$8B,$E1,$3E,$7B,$AD
	.dc.b	$FF,$35,$7C,$7C,$73,$1F,$8B,$F0,$B6,$98,$7F,$2B,$CB,$AD,$FC,$5C
	.dc.b	$BE,$F6,$1E,$78,$C1,$FE,$64,$E6,$CE,$1F,$B0,$6B,$FE,$6D,$E6,$FC
	.dc.b	$8C,$EC,$95,$FD,$33,$D8,$FF,$35,$DF,$2C,$BD,$71,$8F,$84,$95,$FE
	.dc.b	$6B,$FA,$B7,$C9,$E3,$3F,$F2,$B6,$3F,$E4,$EA,$FF,$08,$B6,$19,$AF
	.dc.b	$8D,$7F,$9D,$6B,$5F,$18,$FC,$2F,$FD,$4F,$87,$FA,$51,$BA,$7F,$AB
	.dc.b	$BE,$83,$F8,$33,$FE,$B0,$8C,$E1,$FE,$A9,$BE,$5E,$E7,$FA,$6A,$73
	.dc.b	$82,$58,$72,$BE,$15,$A9,$E3,$1F,$EA,$F8,$F1,$D7,$FA,$5E,$72,$FF
	.dc.b	$92,$9A,$E7,$CC,$FE,$C2,$FF,$AB,$1B,$8F,$A7,$D9,$FF,$56,$C6,$30
	.dc.b	$F4,$36,$BC,$CD,$BF,$3D,$70,$9F,$F3,$47,$D7,$FF,$80,$FD,$5F,$F2
	.dc.b	$FB,$63,$FC,$BA,$1C,$3F,$E6,$06,$BF,$EB,$1F,$25,$F5,$6E,$7F,$F3
	.dc.b	$4C,$61,$B1,$99,$EB,$39,$7B,$D7,$FA,$97,$1F,$1F,$B7,$F3,$E7,$2D
	.dc.b	$7C,$4D,$7E,$0F,$C8,$FF,$B0,$F6,$FF,$75,$EB,$FE,$D8,$9F,$EB,$B7
	.dc.b	$FD,$92,$B5,$B7,$FB,$9F,$C6,$7C,$73,$E3,$F3,$87,$CA,$BB,$87,$18
	.dc.b	$E7,$8F,$57,$FC,$8B,$7C,$5C,$F8,$1F,$E8,$AA,$FE,$E6,$DA,$FE,$1D
	.dc.b	$3F,$EF,$36,$BF,$83,$6E,$7E,$A9,$F0,$C6,$33,$F4,$7E,$57,$FE,$22
	.dc.b	$AD,$5F,$80,$87,$F9,$62,$35,$3F,$AC,$F7,$7D,$BD,$5F,$D3,$1B,$D2
	.dc.b	$7E,$0E,$EB,$FF,$40,$61,$FA,$F9,$D9,$1F,$F0,$4C,$EF,$DC,$E3,$E1
	.dc.b	$FF,$51,$9F,$F2,$8F,$D6,$E0,$CF,$99,$23,$24,$3F,$CA,$8F,$D2,$E8
	.dc.b	$FF,$22,$FD,$9F,$FA,$D1,$B3,$FD,$4B,$14,$E5,$B1,$98,$E5,$18,$8F
	.dc.b	$E4,$FB,$7F,$33,$8C,$6F,$FC,$CB,$E0,$7C,$1F,$6F,$F7,$63,$FC,$9F
	.dc.b	$F4,$9E,$BC,$3F,$D9,$2D,$8F,$E6,$13,$BC,$3F,$9F,$3F,$85,$AF,$EC
	.dc.b	$6D,$73,$CF,$EB,$8C,$47,$EA,$5E,$4D,$8F,$19,$F3,$3E,$43,$E2,$7D
	.dc.b	$BF,$F4,$79,$FF,$0D,$E2,$F8,$9F,$CC,$C2,$30,$FF,$81,$67,$83,$F0
	.dc.b	$8F,$DC,$38,$7F,$BC,$B5,$FE,$B0,$D7,$7E,$67,$D3,$8E,$79,$FF,$51
	.dc.b	$6E,$7C,$27,$F0,$7E,$87,$A3,$5F,$2F,$D8,$7E,$EF,$0F,$F3,$DB,$6F
	.dc.b	$BC,$EF,$1F,$F0,$7E,$3F,$E8,$56,$C7,$84,$8F,$B5,$E0,$FF,$A7,$7E
	.dc.b	$27,$F3,$FE,$3F,$F9,$92,$3F,$C8,$38,$7E,$03,$04,$48,$C3,$FD,$42
	.dc.b	$78,$3F,$1F,$AF,$FA,$0F,$67,$F2,$68,$D3,$63,$31,$FE,$B4,$9E,$2F
	.dc.b	$F6,$B7,$9E,$7F,$E3,$A6,$E1,$FE,$F0,$6C,$FF,$A9,$3A,$1F,$71,$EA
	.dc.b	$E8,$FB,$BF,$5F,$E8,$AF,$F2,$DB,$5F,$35,$FE,$57,$0E,$95,$E5,$8F
	.dc.b	$CD,$4F,$8E,$30,$47,$FD,$07,$97,$F5,$E8,$F0,$47,$FB,$2B,$EF,$BC
	.dc.b	$3E,$9C,$8F,$0B,$F9,$0F,$39,$F4,$34,$F8,$D3,$FE,$00,$6C,$6A,$3E
	.dc.b	$53,$5F,$57,$5F,$FB,$BF,$E0,$79,$47,$F9,$E2,$75,$C1,$FE,$28,$9F
	.dc.b	$CF,$FD,$6F,$B2,$BE,$23,$5F,$BA,$D4,$DF,$AE,$FE,$95,$1F,$2D,$AF
	.dc.b	$E1,$1A,$9C,$D3,$EF,$99,$FF,$48,$8C,$36,$33,$5F,$EA,$EF,$BC,$FE
	.dc.b	$2F,$EB,$FE,$53,$E7,$3F,$6F,$C4,$F3,$A9,$9B,$FA,$70,$9F,$B4,$48
	.dc.b	$FC,$73,$5F,$C0,$BC,$5F,$DF,$72,$3E,$C3,$FD,$01,$1F,$CE,$3A,$3A
	.dc.b	$37,$C1,$FE,$72,$7F,$3A,$B9,$19,$21,$FC,$27,$8D,$5C,$A3,$1F,$B8
	.dc.b	$7E,$ED,$FF,$7E,$AF,$A4,$FE,$9A,$D8,$FE,$92,$EE,$8D,$35,$1D,$27
	.dc.b	$8F,$9F,$FD,$6B,$EB,$1F,$3B,$D3,$EC,$A5,$FE,$6A,$72,$D9,$FD,$B4
	.dc.b	$62,$3F,$D9,$ED,$7F,$CC,$0D,$7F,$82,$F6,$A3,$87,$37,$FA,$C9,$BC
	.dc.b	$1A,$6D,$7C,$0F,$1F,$DF,$F2,$32,$3D,$EF,$C2,$C6,$46,$08,$7C,$A7
	.dc.b	$E3,$FD,$7F,$C8,$F8,$C9,$3F,$22,$32,$FC,$A8,$CB,$F4,$A3,$FA,$1E
	.dc.b	$1F,$E3,$68,$FF,$70,$F2,$FF,$BD,$9B,$DB,$AF,$6A,$47,$F1,$DA,$E0
	.dc.b	$FF,$91,$1A,$FF,$0F,$6B,$FC,$4D,$DE,$AD,$71,$5D,$2B,$FA,$80,$FA
	.dc.b	$F7,$C2,$7F,$84,$3E,$75,$C5,$79,$25,$FF,$83,$A7,$2D,$9C,$BF,$93
	.dc.b	$8F,$C3,$E5,$FE,$D0,$77,$19,$E1,$EC,$47,$F2,$3F,$53,$F9,$43,$0E
	.dc.b	$CC,$9F,$D1,$AB,$D8,$FF,$48,$FB,$CF,$E8,$EB,$EB,$7F,$8A,$ED,$FC
	.dc.b	$D5,$B1,$BF,$57,$FB,$AD,$7F,$84,$E3,$F9,$C3,$5F,$E7,$8D,$78,$1F
	.dc.b	$D3,$DE,$E4,$8F,$F1,$BF,$DF,$FE,$C7,$0F,$2D,$3E,$0F,$FA,$13,$FD
	.dc.b	$79,$AE,$7C,$BF,$D9,$F1,$48,$F7,$67,$FD,$36,$B6,$72,$ED,$FD,$AD
	.dc.b	$7D,$72,$57,$D5,$7F,$7B,$F6,$32,$30,$43,$F8,$2B,$FE,$52,$D7,$FA
	.dc.b	$04,$7F,$B3,$DA,$FE,$C1,$AF,$FB,$09,$BD,$9F,$3F,$CB,$E4,$49,$FF
	.dc.b	$29,$FB,$AF,$91,$FF,$2D,$B6,$33,$18,$33,$F4,$16,$32,$43,$FC,$D4
	.dc.b	$F9,$FE,$3F,$FB,$F6,$26,$92,$7F,$D5,$38,$FE,$0E,$38,$47,$F6,$C9
	.dc.b	$E0,$F9,$BC,$8F,$F4,$6F,$DB,$BE,$23,$FD,$10,$D9,$FF,$59,$F8,$A8
	.dc.b	$C4,$7F,$B2,$5A,$FF,$B3,$5C,$F2,$7E,$5A,$78,$57,$63,$E1,$FA,$B6
	.dc.b	$FF,$D4,$B7,$C3,$93,$FD,$C4,$D8,$71,$FF,$A0,$EB,$84,$7F,$92,$70
	.dc.b	$EA,$F2,$C3,$AB,$D3,$EB,$3D,$64,$7D,$78,$23,$11,$CE,$30,$F1,$3E
	.dc.b	$DF,$CA,$FF,$97,$3E,$29,$D8,$D7,$FC,$73,$F5,$DC,$2D,$4F,$23,$FC
	.dc.b	$4D,$39,$8C,$4F,$0C,$BF,$DB,$58,$73,$85,$6E,$31,$E1,$71,$C3,$FC
	.dc.b	$60,$DA,$FF,$24,$BB,$C6,$E1,$D1,$ED,$3F,$E4,$EC,$7C,$2F,$69,$FF
	.dc.b	$75,$5F,$99,$E1,$19,$38,$61,$E5,$C7,$C1,$19,$3D,$F5,$C5,$DE,$1E
	.dc.b	$07,$E1,$D7,$AA,$FB,$C7,$98,$F2,$57,$B2,$3C,$38,$FA,$E7,$B6,$3E
	.dc.b	$8C,$7C,$26,$49,$5D,$27,$77,$E3,$7C,$38,$F8,$EF,$EC,$B6,$75,$1D
	.dc.b	$6B,$E0,$DF,$87,$0F,$79,$23,$CA,$68,$F0,$47,$91,$C6,$1F,$08,$FF
	.dc.b	$08,$3E,$59,$E5,$19,$7B,$46,$49,$19,$7E,$9E,$BF,$E1,$2B,$F4,$CA
	.dc.b	$31,$97,$E2,$5B,$97,$FA,$31,$FF,$0C,$CF,$FA,$3B,$EB,$31,$8F,$F9
	.dc.b	$DE,$11,$87,$FC,$F4,$D7,$FC,$FC,$D7,$91,$EC,$3F,$D0,$5B,$7F,$54
	.dc.b	$9F,$D1,$46,$9B,$19,$CB,$C1,$5F,$81,$AF,$F5,$0C,$E5,$F0,$9D,$4C
	.dc.b	$F0,$ED,$F0,$E9,$D2,$3F,$1D,$18,$BF,$F4,$95,$7F,$66,$AF,$1C,$FB
	.dc.b	$8E,$87,$F1,$FA,$E8,$E7,$0E,$8E,$8E,$13,$83,$9E,$1E,$0B,$ED,$87
	.dc.b	$B2,$BF,$48,$75,$29,$C9,$F2,$E1,$E8,$8C,$1F,$29,$C5,$75,$6C,$7A
	.dc.b	$2B,$E3,$3F,$50,$C4,$7A,$A7,$84,$78,$1B,$7E,$88,$F2,$D7,$A1,$F9
	.dc.b	$F1,$ED,$BF,$56,$1F,$08,$E0,$57,$DD,$1B,$9E,$31,$EC,$7F,$84,$63
	.dc.b	$FD,$1F,$D7,$FF,$0B,$E7,$FD,$0F,$81,$18,$D3,$FD,$C3,$C4,$E9,$7C
	.dc.b	$B6,$72,$FC,$F6,$1C,$2F,$FA,$3E,$1F,$E0,$E9,$FE,$8E,$D7,$6D,$CB
	.dc.b	$C5,$3F,$D0,$3E,$73,$A4,$7F,$5B,$7B,$65,$E1,$F8,$2C,$E2,$9A,$4B
	.dc.b	$FE,$FB,$37,$1F,$E9,$13,$74,$FF,$1A,$E1,$CE,$FF,$AE,$CE,$EB,$FB
	.dc.b	$33,$E5,$C3,$B4,$F9,$4F,$EF,$3D,$0F,$92,$F8,$5D,$7F,$E3,$D7,$CF
	.dc.b	$BB,$E4,$F3,$5B,$BF,$EF,$D5,$F0,$EF,$84,$74,$8C,$4E,$F2,$FF,$6F
	.dc.b	$E0,$F0,$FA,$9D,$BE,$26,$E1,$FE,$1C,$8B,$5F,$EB,$1A,$56,$36,$FF
	.dc.b	$2A,$E5,$FE,$53,$F9,$6F,$29,$9C,$7F,$C5,$8E,$1F,$7B,$E4,$23,$FC
	.dc.b	$57,$7F,$F5,$9F,$CC,$FD,$6F,$2D,$F4,$3F,$63,$D6,$4E,$62,$5F,$F9
	.dc.b	$DD,$D4,$63,$5F,$F6,$2C,$7E,$35,$7C,$2B,$EA,$3A,$4E,$63,$17,$FE
	.dc.b	$54,$A1,$C2,$B4,$70,$8D,$46,$5B,$F5,$8F,$73,$C0,$D8,$E9,$A3,$CE
	.dc.b	$FD,$0C,$70,$AD,$9F,$9D,$B5,$F1,$7F,$9D,$35,$F2,$9E,$67,$0E,$1E
	.dc.b	$1C,$1F,$E7,$6F,$AA,$FF,$6C,$DF,$73,$D2,$E2,$BF,$DD,$7F,$57,$F4
	.dc.b	$73,$FA,$0D,$AF,$F5,$24,$59,$DC,$60,$FF,$54,$7A,$3F,$67,$D1,$FB
	.dc.b	$4D,$8C,$9F,$EA,$D8,$CB,$4F,$1C,$7D,$7C,$0F,$D9,$FD,$37,$D7,$3F
	.dc.b	$02,$72,$F0,$3F,$AF,$5A,$32,$FF,$61,$6B,$F2,$0D,$57,$CF,$8D,$E5
	.dc.b	$FE,$9C,$F1,$5F,$27,$E3,$AF,$C5,$46,$30,$FB,$7C,$8F,$B0,$FC,$B7
	.dc.b	$DD,$FE,$5D,$B3,$A6,$E7,$FE,$4B,$8F,$E4,$75,$C5,$E7,$5F,$9F,$FA
	.dc.b	$1F,$D0,$6F,$67,$BB,$86,$9E,$28,$C5,$F5,$1D,$4F,$80,$FD,$CD,$7E
	.dc.b	$76,$31,$D9,$FA,$DF,$9D,$FE,$23,$5A,$BE,$55,$AB,$FF,$47,$CE,$67
	.dc.b	$F2,$33,$97,$E6,$35,$CC,$73,$E0,$FE,$4F,$A7,$FB,$1E,$FC,$53,$97
	.dc.b	$E1,$F1,$FE,$FF,$83,$F9,$BF,$EE,$3F,$B2,$9F,$9F,$18,$C7,$51,$FA
	.dc.b	$C9,$D4,$7C,$1E,$96,$3F,$72,$BF,$A8,$D6,$B2,$FE,$72,$FF,$79,$6F
	.dc.b	$CA,$F8,$38,$FC,$2D,$AE,$CD,$C7,$B5,$A3,$E4,$FC,$0C,$AF,$E1,$43
	.dc.b	$F6,$57,$4E,$70,$F9,$BB,$AF,$F9,$D3,$6F,$E7,$C6,$08,$FF,$10,$D5
	.dc.b	$EC,$73,$D7,$FD,$E7,$18,$9F,$E3,$5F,$50,$FC,$05,$F3,$1F,$DB,$E3
	.dc.b	$0F,$CE,$F7,$71,$3A,$AF,$C3,$7D,$A3,$E0,$FF,$A7,$36,$3F,$DA,$2D
	.dc.b	$DB,$F0,$2D,$7F,$5D,$F5,$2F,$85,$FE,$38,$DA,$FD,$BB,$67,$FC,$AC
	.dc.b	$D7,$F8,$EE,$8E,$47,$CF,$FF,$66,$C7,$F8,$C6,$FF,$6A,$31,$C5,$32
	.dc.b	$F2,$BE,$82,$3F,$C1,$BF,$08,$FC,$86,$5F,$CF,$9B,$D6,$C7,$AD,$7F
	.dc.b	$2E,$9C,$C3,$E1,$3C,$47,$E5,$BE,$82,$D7,$C8,$8F,$8E,$47,$E3,$B7
	.dc.b	$0F,$DB,$0F,$85,$FF,$67,$CF,$13,$FA,$6B,$88,$FD,$9D,$BC,$3F,$84
	.dc.b	$7E,$9B,$DB,$B7,$F6,$23,$EB,$91,$FA,$ED,$AF,$0D,$BE,$1B,$8C,$61
	.dc.b	$FE,$D1,$F8,$0F,$E0,$E3,$CE,$DE,$0F,$D4,$FC,$96,$BA,$66,$5F,$CC
	.dc.b	$AB,$85,$FF,$2F,$6B,$FD,$61,$AF,$F6,$E7,$A4,$7F,$5C,$FA,$97,$C4
	.dc.b	$F4,$AF,$8F,$18,$9F,$EE,$34,$F9,$9F,$39,$FC,$22,$EC,$EA,$3F,$B6
	.dc.b	$4E,$E3,$CD,$18,$23,$FD,$F2,$BF,$CA,$57,$F0,$27,$FC,$15,$18,$C9
	.dc.b	$5F,$DC,$BF,$01,$71,$7F,$83,$DB,$1F,$C5,$7E,$07,$13,$23,$FA,$BC
	.dc.b	$EB,$1F,$F5,$B3,$5F,$EA,$CD,$7B,$4F,$8B,$AF,$94,$7C,$47,$C1,$A4
	.dc.b	$BF,$E4,$9E,$E5,$AF,$C7,$E1,$F2,$5E,$F5,$FE,$0F,$8C,$47,$82,$3F
	.dc.b	$CD,$DD,$7F,$B1,$F3,$FF,$99,$FD,$A7,$F3,$46,$E3,$FC,$C5,$B7,$FC
	.dc.b	$91,$BA,$FF,$91,$BF,$1F,$FF,$29,$37,$3F,$F2,$EF,$D0,$7F,$98,$27
	.dc.b	$38,$7F,$8C,$3B,$11,$8C,$7F,$AD,$B7,$3F,$F4,$55,$F1,$C3,$FC,$FF
	.dc.b	$F1,$5F,$DD,$6F,$FB,$17,$C2,$A3,$19,$FF,$9E,$3D,$FF,$C7,$D7,$FB
	.dc.b	$B6,$73,$A7,$FB,$AD,$B1,$9D,$7F,$CE,$7C,$FD,$99,$5F,$EE,$DC,$7B
	.dc.b	$47,$FA,$4E,$FC,$57,$FE,$36,$9C,$D6,$E7,$FD,$07,$3E,$66,$E3,$E5
	.dc.b	$FA,$2F,$5F,$C2,$CB,$FF,$2A,$61,$EB,$8F,$F4,$2C,$E7,$0F,$E1,$7F
	.dc.b	$51,$FD,$13,$D0,$FF,$11,$B6,$3F,$C4,$01,$C3,$FE,$23,$6B,$FE,$4C
	.dc.b	$CB,$F8,$27,$D2,$FF,$8B,$1B,$1C,$9A,$FF,$8A,$DA,$FF,$95,$25,$F2
	.dc.b	$2C,$BC,$6D,$9F,$1C,$EA,$72,$7C,$5F,$B7,$FF,$2E,$EB,$FA,$EA,$FA
	.dc.b	$D1,$8D,$7E,$B3,$63,$B7,$C9,$7F,$73,$7E,$3E,$49,$D6,$1F,$83,$6C
	.dc.b	$66,$7E,$6D,$71,$7F,$13,$F8,$07,$F2,$F7,$EF,$35,$FF,$25,$5A,$31
	.dc.b	$1F,$7D,$AF,$FA,$2A,$35,$E2,$78,$A7,$96,$5E,$D9,$CD,$78,$BC,$AD
	.dc.b	$5C,$8F,$C0,$DF,$0C,$FA,$63,$FD,$92,$7F,$C7,$9D,$1F,$E5,$86,$F1
	.dc.b	$FC,$5E,$9F,$D5,$B7,$BB,$53,$FE,$3E,$AB,$AF,$ED,$EB,$84,$63,$1F
	.dc.b	$DA,$5D,$B7,$53,$E6,$FF,$D8,$B1,$FE,$33,$FB,$6F,$A9,$3F,$DE,$5B
	.dc.b	$3A,$9F,$F1,$B3,$72,$FF,$19,$CF,$0F,$34,$F0,$F7,$3E,$F5,$BE,$73
	.dc.b	$FD,$C8,$FF,$A0,$1A,$FE,$79,$AF,$F9,$FD,$AF,$FA,$A1,$DD,$FE,$36
	.dc.b	$3A,$7C,$6F,$69,$CE,$9F,$EC,$2E,$9F,$7F,$B9,$9F,$8C,$FC,$84,$3F
	.dc.b	$01,$78,$98,$6E,$5F,$EC,$6C,$7F,$DE,$0D,$BF,$F4,$BB,$EB,$6C,$6A
	.dc.b	$3D,$51,$83,$C0,$F1,$3F,$DA,$D6,$FA,$9F,$CC,$69,$EB,$BF,$F6,$95
	.dc.b	$6B,$CE,$7C,$A7,$A1,$B1,$E5,$F8,$9F,$80,$78,$88,$FC,$D6,$BF,$E1
	.dc.b	$8A,$EA,$D8,$F8,$7F,$52,$F9,$3F,$AF,$F5,$3F,$BF,$AF,$F4,$A6,$9F
	.dc.b	$E0,$FF,$67,$C4,$6B,$C1,$93,$FE,$BA,$AE,$0F,$FA,$DD,$AF,$FB,$81
	.dc.b	$AF,$FB,$D9,$ED,$AF,$DB,$6F,$1E,$EB,$CB,$E6,$79,$39,$3F,$1D,$E4
	.dc.b	$CC,$7F,$DF,$78,$7F,$B4,$1B,$3F,$5A,$31,$1F,$CA,$9A,$F9,$63,$47
	.dc.b	$57,$89,$FE,$AF,$F9,$F7,$83,$F9,$AF,$A8,$FF,$69,$69,$FC,$77,$C2
	.dc.b	$FE,$15,$F0,$3F,$B7,$F1,$3F,$CF,$B8,$8F,$F2,$3D,$F0,$DF,$FE,$00
	.dc.b	$BF,$8A,$83,$48,$FE,$39,$E0,$F3,$6B,$C0,$FE,$0D,$97,$E4,$71,$FD
	.dc.b	$27,$A7,$FE,$1C,$9F,$14,$63,$C3,$FD,$9C,$E6,$7C,$BF,$4D,$ED,$3C
	.dc.b	$04,$7F,$DE,$B8,$23,$F4,$32,$F8,$1E,$15,$DB,$2F,$6D,$76,$9F,$E7
	.dc.b	$D8,$BF,$01,$7C,$56,$1B,$87,$11,$3B,$9F,$E1,$3F,$51,$FE,$9B,$9D
	.dc.b	$0F,$A9,$FD,$83,$F3,$CF,$A0,$E1,$C3,$5C,$F8,$3F,$BA,$9F,$C7,$C7
	.dc.b	$D7,$6A,$73,$6F,$A9,$F8,$53,$E2,$E0,$FE,$CE,$FA,$7C,$A7,$CB,$BE
	.dc.b	$8F,$F9,$5B,$67,$87,$A9,$FC,$92,$F8,$C7,$F7,$E6,$BE,$B0,$E1,$F7
	.dc.b	$35,$FC,$73,$CB,$1F,$E4,$8D,$E1,$DD,$7E,$1B,$E9,$FF,$9D,$F8,$7F
	.dc.b	$CD,$FE,$6E,$63,$E9,$F3,$FF,$82,$E5,$FD,$E7,$5F,$E8,$D7,$AA,$FF
	.dc.b	$74,$DF,$27,$E0,$7C,$87,$F1,$47,$FD,$DB,$EB,$7F,$43,$3F,$A5,$C6
	.dc.b	$70,$E8,$DD,$3F,$DF,$B3,$BA,$E7,$E5,$70,$F8,$AF,$F0,$37,$D4,$FE
	.dc.b	$0E,$7F,$0C,$D7,$F0,$F7,$FA,$FD,$18,$FE,$1E,$72,$7F,$60,$71,$97
	.dc.b	$F8,$56,$F9,$37,$A3,$F2,$75,$C2,$F9,$E3,$E3,$8F,$0C,$EA,$B8,$74
	.dc.b	$66,$7B,$F8,$0F,$AE,$7C,$7F,$F6,$5F,$B8,$95,$F9,$FA,$D4,$FF,$7F
	.dc.b	$AD,$CE,$5A,$F4,$6E,$1F,$2B,$A3,$FB,$87,$77,$F7,$1F,$84,$97,$F9
	.dc.b	$3F,$5F,$F2,$EF,$AD,$B7,$F7,$5F,$F7,$F4,$6E,$79,$BF,$9D,$C7,$F2
	.dc.b	$DE,$54,$8C,$10,$FC,$C4,$7F,$64,$EB,$F3,$C7,$E6,$5A,$FE,$62,$3F
	.dc.b	$40,$47,$F7,$7F,$53,$FE,$DB,$7F,$55,$F3,$FF,$85,$5B,$7F,$8D,$FB
	.dc.b	$3F,$97,$1F,$E1,$69,$D7,$21,$E1,$3B,$AF,$9D,$E3,$AF,$87,$4F,$F8
	.dc.b	$B7,$EF,$FE,$D7,$57,$E2,$76,$7C,$5F,$F8,$89,$B8,$FD,$7A,$DB,$E4
	.dc.b	$6E,$46,$9A,$D3,$C9,$C9,$FE,$22,$FA,$1F,$E5,$17,$F0,$3E,$87,$FC
	.dc.b	$4B,$C0,$75,$E4,$49,$FF,$40,$8A,$E2,$FF,$A1,$1A,$FF,$07,$6B,$FC
	.dc.b	$8D,$DF,$37,$F8,$9D,$BB,$66,$BF,$C6,$BF,$53,$FD,$9A,$BF,$C5,$9F
	.dc.b	$7F,$FE,$FA,$C3,$F9,$7A,$F9,$1F,$F1,$DB,$63,$3F,$19,$82,$38,$3F
	.dc.b	$9B,$FC,$DF,$F5,$5C,$BF,$D8,$DF,$7D,$D6,$7F,$5D,$C7,$72,$FE,$20
	.dc.b	$DA,$DE,$1F,$E6,$3B,$C5,$FE,$11,$BC,$67,$3C,$5F,$83,$F7,$7F,$CD
	.dc.b	$5E,$A7,$F0,$3B,$FE,$8D,$7F,$DF,$3C,$4F,$9A,$D8,$FF,$57,$B5,$FE
	.dc.b	$64,$DD,$3F,$C0,$15,$FC,$2B,$3F,$F7,$03,$6B,$F2,$FC,$5F,$39,$BC
	.dc.b	$5F,$E7,$B8,$C6,$1F,$6E,$3F,$CD,$FF,$1A,$F9,$3D,$44,$BF,$E0,$D7
	.dc.b	$CF,$55,$EE,$7F,$85,$62,$FA,$AF,$5B,$63,$D6,$D7,$D6,$D7,$E1,$B9
	.dc.b	$D3,$DB,$3F,$20,$F4,$70,$7F,$AC,$BE,$B5,$C7,$F9,$4F,$71,$18,$C7
	.dc.b	$FD,$D8,$D7,$FD,$5E,$D7,$FD,$D8,$E7,$0F,$F5,$53,$78,$3F,$D6,$0D
	.dc.b	$9F,$F6,$97,$D3,$BE,$3F,$C5,$8C,$9F,$F7,$6D,$6F,$15,$EE,$73,$5B
	.dc.b	$7F,$A1,$CF,$A2,$3F,$D9,$B3,$B9,$CF,$4F,$EA,$9F,$F7,$13,$68,$D2
	.dc.b	$70,$FF,$B5,$9F,$C0,$46,$A3,$AF,$4F,$FA,$8B,$E3,$9F,$57,$FE,$1E
	.dc.b	$C1,$F5,$5F,$AB,$53,$F5,$E3,$0D,$AF,$F7,$F5,$71,$9C,$BC,$FC,$2D
	.dc.b	$F1,$7F,$88,$DB,$87,$F9,$4D,$DF,$0D,$C3,$9B,$FE,$57,$C3,$75,$E4
	.dc.b	$9F,$F1,$66,$3E,$18,$FF,$2E,$3E,$AA,$E1,$3C,$A7,$81,$AC,$3F,$C6
	.dc.b	$97,$D0,$F0,$C6,$8F,$7E,$3E,$9A,$EF,$5E,$96,$CF,$9C,$F0,$CE,$DB
	.dc.b	$7E,$27,$C0,$53,$E5,$D7,$8A,$BC,$D7,$E2,$3C,$8F,$69,$F2,$61,$ED
	.dc.b	$6E,$1F,$3E,$75,$5C,$27,$C2,$6B,$0F,$33,$E0,$2D,$FE,$C7,$1F,$82
	.dc.b	$78,$6F,$99,$27,$9D,$EA,$3C,$93,$F5,$0C,$4F,$81,$C4,$68,$FA,$4F
	.dc.b	$CE,$CF,$FC,$69,$7E,$B6,$E3,$F9,$16,$E3,$FE,$97,$7F,$C7,$58,$7F
	.dc.b	$A6,$7E,$AF,$18,$7F,$94,$E1,$18,$7F,$CA,$CD,$7F,$CE,$4D,$79,$1F
	.dc.b	$ED,$AC,$7F,$CB,$9C,$2E,$2F,$8F,$3A,$BF,$D1,$3F,$BF,$8C,$BE,$03
	.dc.b	$91,$FE,$7F,$9E,$4E,$0F,$09,$4E,$45,$39,$69,$DA,$38,$BE,$08,$CD
	.dc.b	$F9,$AB,$F1,$51,$E5,$8C,$3C,$F0,$F4,$DF,$8B,$0E,$51,$C2,$3A,$D7
	.dc.b	$C9,$8C,$5F,$4B,$EE,$7A,$DC,$46,$A3,$F4,$C7,$CA,$C3,$D1,$3A,$8C
	.dc.b	$DF,$67,$E9,$9F,$05,$F2,$5F,$86,$31,$3F,$32,$FB,$5F,$A2,$79,$CF
	.dc.b	$58,$EF,$1E,$23,$CE,$7F,$B7,$6F,$C0,$FD,$0C,$36,$73,$2E,$1E,$1A
	.dc.b	$E6,$77,$2B,$F9,$1D,$FF,$C1,$D8,$7F,$B0,$F2,$FE,$5F,$3F,$7B,$3F
	.dc.b	$BF,$1F,$CF,$BD,$CC,$57,$1A,$B9,$7F,$38,$7F,$0A,$FF,$B1,$6B,$9C
	.dc.b	$FE,$6D,$B1,$FE,$AF,$6B,$BC,$1F,$1F,$D2,$7F,$A0,$B7,$89,$F0,$79
	.dc.b	$AD,$E1,$E9,$6E,$1F,$D3,$2F,$85,$6D,$95,$B7,$2D,$24,$F9,$6B,$9B
	.dc.b	$E5,$9F,$6B,$F2,$5A,$FF,$9A,$39,$DD,$DF,$13,$B5,$18,$C3,$FC,$1B
	.dc.b	$F2,$B9,$8C,$5F,$FA,$69,$AF,$F8,$1A,$FA,$DF,$26,$CE,$A4,$7F,$88
	.dc.b	$BD,$7F,$07,$71,$FA,$BD,$DF,$E1,$AF,$7B,$31,$33,$FF,$0F,$BA,$EF
	.dc.b	$4E,$70,$F2,$7D,$2F,$9C,$C7,$2F,$F8,$D1,$C6,$1F,$E2,$07,$CA,$47
	.dc.b	$FC,$7B,$E4,$3E,$3B,$E2,$36,$3F,$D2,$3E,$46,$26,$47,$FB,$F7,$C7
	.dc.b	$FB,$78,$7F,$10,$E1,$FF,$7F,$57,$B2,$73,$5F,$0B,$27,$D5,$7F,$8A
	.dc.b	$23,$E2,$E1,$F9,$48,$C4,$FF,$B5,$8D,$FD,$37,$E6,$9E,$3E,$25,$9D
	.dc.b	$D7,$E7,$3D,$C7,$B1,$9A,$E1,$18,$C7,$B9,$E3,$FA,$77,$C5,$75,$6C
	.dc.b	$6B,$2F,$B3,$F5,$2F,$80,$D5,$CA,$FF,$49,$DF,$0B,$FE,$88,$D7,$E0
	.dc.b	$07,$0F,$86,$7C,$1A,$F3,$8F,$81,$28,$FA,$77,$2F,$FD,$87,$16,$BF
	.dc.b	$DD,$BC,$5E,$59,$FA,$9F,$25,$9D,$BF,$D1,$2F,$E3,$BD,$7F,$8F,$3F
	.dc.b	$94,$E3,$7D,$E7,$FB,$D2,$3F,$D9,$FC,$1F,$A5,$8C,$65,$FF,$00,$6F
	.dc.b	$8B,$FE,$D2,$8C,$E3,$FD,$43,$AB,$FD,$A1,$7C,$7D,$9F,$B1,$E4,$FE
	.dc.b	$0B,$0F,$E4,$DE,$0F,$CF,$F8,$CC,$B3,$4F,$F5,$ED,$7E,$F2,$BF,$AE
	.dc.b	$4F,$EE,$9F,$F3,$03,$63,$A4,$7F,$98,$F9,$B9,$3F,$89,$8F,$C0,$3F
	.dc.b	$EF,$FF,$C0,$3F,$5F,$EF,$3E,$D2,$C6,$1B,$8F,$E0,$75,$56,$E3,$0A
	.dc.b	$B8,$47,$FA,$43,$E0,$BF,$84,$1F,$BC,$F8,$5F,$B5,$19,$26,$2E,$0F
	.dc.b	$91,$F2,$FF,$E0,$19,$CC,$61,$E5,$C3,$FC,$EB,$E1,$BF,$80,$C6,$0F
	.dc.b	$CB,$38,$7D,$18,$7D,$68,$C6,$1D,$E7,$F8,$9E,$8A,$D7,$87,$E1,$AD
	.dc.b	$09,$D6,$1E,$19,$D6,$3F,$7E,$F8,$E1,$F3,$30,$FE,$37,$93,$EC,$F1
	.dc.b	$3F,$2E,$30,$FC,$A9,$DC,$E9,$F0,$E4,$AD,$C6,$36,$7D,$5F,$6F,$6F
	.dc.b	$ED,$B7,$1F,$AA,$DA,$E0,$DC,$FF,$B0,$E2,$F9,$1F,$E1,$8D,$BF,$E0
	.dc.b	$1D,$5F,$C1,$3A,$33,$52,$7F,$C8,$18,$FC,$58,$FC,$A5,$F1,$8F,$F2
	.dc.b	$23,$5F,$E0,$2E,$70,$FD,$37,$3F,$F2,$D9,$FE,$3D,$39,$BF,$AB,$39
	.dc.b	$AF,$E5,$9F,$48,$F7,$BF,$AC,$36,$BF,$8B,$7D,$4B,$EA,$5F,$45,$AB
	.dc.b	$FE,$51,$F5,$2F,$8B,$FD,$53,$70,$F5,$B7,$5F,$F2,$E3,$5F,$E7,$1C
	.dc.b	$1F,$EB,$2F,$AC,$FE,$D1,$F0,$07,$FC,$1B,$B3,$EE,$7F,$81,$E3,$FC
	.dc.b	$36,$B4,$FC,$CD,$FF,$93,$57,$F8,$4B,$EA,$DF,$43,$EC,$F3,$3F,$9F
	.dc.b	$46,$2F,$F9,$96,$88,$C3,$FE,$71,$6B,$FC,$B9,$AF,$F4,$F6,$BF,$EE
	.dc.b	$7F,$C2,$5F,$41,$FE,$24,$F1,$51,$8B,$EF,$7D,$B2,$FF,$35,$D6,$AF
	.dc.b	$FA,$97,$5F,$FB,$62,$FC,$AD,$C7,$F5,$3C,$0D,$5F,$59,$C3,$E1,$A6
	.dc.b	$1F,$8A,$C3,$86,$1F,$D1,$9A,$FF,$67,$6B,$FE,$0A,$7B,$67,$FD,$BF
	.dc.b	$EE,$BE,$6C,$FF,$67,$9C,$D7,$F3,$D9,$CD,$7F,$90,$27,$FC,$36,$DF
	.dc.b	$77,$F1,$7F,$4D,$EC,$65,$F9,$5F,$E6,$15,$FE,$1B,$C3,$C5,$18,$3F
	.dc.b	$7F,$DD,$C2,$3F,$73,$E1,$7E,$6D,$FF,$10,$FD,$5B,$E3,$7F,$BD,$37
	.dc.b	$1F,$BB,$F5,$BF,$98,$7E,$FC,$66,$BF,$C0,$4D,$7F,$B5,$B5,$FF,$06
	.dc.b	$35,$EF,$5F,$D1,$E3,$96,$9F,$EA,$4F,$A3,$F1,$7D,$5F,$8A,$BF,$CC
	.dc.b	$B1,$8C,$3C,$3F,$15,$E6,$3F,$C2,$19,$3E,$B7,$13,$A3,$F8,$1E,$4A
	.dc.b	$F0,$12,$32,$F6,$C3,$F4,$B9,$FC,$48,$C7,$8F,$F1,$3C,$19,$7F,$A0
	.dc.b	$FA,$D7,$D2,$E3,$0F,$87,$5E,$D2,$57,$89,$B1,$FC,$EB,$57,$DC,$BE
	.dc.b	$67,$97,$0F,$C3,$8C,$61,$F4,$3D,$4F,$AA,$7F,$8B,$F3,$F9,$31,$F3
	.dc.b	$A3,$2F,$C0,$7E,$71,$1F,$E4,$3E,$73,$DA,$FE,$87,$F4,$3F,$DC,$1B
	.dc.b	$3F,$DA,$9B,$A7,$F3,$D6,$E5,$FE,$4B,$FC,$AD,$F6,$BF,$E7,$76,$DF
	.dc.b	$F3,$6F,$AE,$7C,$76,$6A,$4F,$F3,$AA,$DE,$3F,$B4,$BD,$CE,$6B,$FD
	.dc.b	$35,$F4,$1F,$E1,$CF,$93,$37,$CC,$FF,$0C,$46,$3D,$AD,$92,$FF,$D6
	.dc.b	$93,$AA,$E4,$7F,$96,$B1,$FF,$57,$35,$FF,$41,$DF,$3A,$FF,$58,$4E
	.dc.b	$65,$2F,$D7,$DE,$1F,$FA,$BD,$B7,$E0,$F9,$3C,$4C,$8F,$F6,$34,$EA
	.dc.b	$FF,$BD,$DB,$1F,$F5,$AD,$6A,$FF,$CD,$9F,$63,$FD,$8F,$CC,$FE,$BA
	.dc.b	$DC,$3F,$CB,$21,$C3,$FE,$5C,$6B,$FE,$7D,$BF,$F5,$6F,$4F,$F6,$66
	.dc.b	$C7,$26,$BF,$E6,$46,$BF,$E8,$59,$72,$7C,$28,$F0,$C6,$32,$F2,$CE
	.dc.b	$67,$C4,$D7,$E6,$E7,$FE,$69,$7E,$1F,$C3,$7F,$A4,$79,$3F,$DC,$F3
	.dc.b	$99,$FF,$07,$FC,$87,$FA,$51,$D9,$0F,$93,$1B,$9F,$F2,$F4,$EE,$73
	.dc.b	$C3,$FE,$04,$C7,$EC,$36,$CD,$24,$FD,$67,$C9,$BF,$A1,$B1,$F8,$2F
	.dc.b	$9F,$FF,$44,$E3,$E8,$9C,$E1,$DA,$B9,$1F,$E9,$75,$5C,$2F,$FC,$FD
	.dc.b	$CA,$FA,$DE,$BE,$7A,$31,$87,$E9,$E7,$C3,$EE,$B5,$7D,$CF,$0B,$DD
	.dc.b	$FF,$5E,$6E,$BF,$B7,$C3,$8D,$F0,$AF,$F2,$07,$CB,$74,$F2,$FF,$77
	.dc.b	$1F,$E4,$3E,$AC,$4C,$8C,$10,$FE,$96,$F8,$7D,$7F,$F1,$8E,$BE,$96
	.dc.b	$E7,$FE,$0A,$AF,$CB,$CE,$5B,$19,$7D,$13,$CD,$FF,$5F,$73,$BE,$C7
	.dc.b	$F9,$D4,$7F,$09,$F7,$7D,$1F,$6F,$23,$04,$3F,$D7,$B8,$FC,$B9,$F1
	.dc.b	$92,$3C,$2F,$C9,$23,$E0,$AF,$E4,$CF,$F1,$7F,$37,$E3,$6F,$23,$04
	.dc.b	$3B,$E7,$FD,$4D,$F6,$CF,$C2,$23,$BE,$0F,$E5,$51,$F8,$CC,$DE,$8B
	.dc.b	$E7,$7F,$07,$F3,$7F,$AB,$BF,$F4,$9F,$CB,$7D,$73,$3D,$33,$2F,$9B
	.dc.b	$5F,$AF,$7F,$D9,$4D,$5B,$A3,$0D,$9E,$53,$FE,$26,$9F,$07,$42,$19
	.dc.b	$F9,$BD,$DF,$29,$F8,$9F,$88,$93,$FE,$76,$BF,$F7,$CB,$67,$F0,$B1
	.dc.b	$88,$FA,$AD,$7F,$CB,$6D,$7D,$31,$D9,$E4,$7F,$AF,$BC,$4F,$8B,$ED
	.dc.b	$7E,$67,$9A,$4F,$86,$32,$FC,$47,$C2,$47,$06,$C8,$7F,$91,$BE,$9B
	.dc.b	$FD,$24,$FF,$A9,$7E,$5F,$F4,$C4,$7D,$DE,$3F,$D6,$3D,$A4,$7E,$9B
	.dc.b	$FE,$94,$7F,$C2,$5E,$96,$DE,$27,$F9,$AB,$D4,$FF,$05,$BF,$EC,$8F
	.dc.b	$CD,$9E,$DF,$F1,$0F,$8C,$46,$27,$53,$98,$F3,$C7,$FB,$1B,$99,$F5
	.dc.b	$4F,$8E,$FF,$52,$C6,$1B,$19,$9F,$E9,$73,$97,$FD,$57,$F5,$EE,$4F
	.dc.b	$53,$63,$31,$83,$EE,$FC,$97,$F0,$BF,$06,$4F,$E0,$E3,$F3,$66,$9F
	.dc.b	$CB,$B5,$FC,$1B,$A7,$81,$FE,$FF,$FA,$D7,$CB,$7F,$22,$F5,$DF,$3D
	.dc.b	$FC,$67,$E0,$A5,$FE,$5F,$47,$6F,$FB,$B2,$F7,$39,$8C,$5F,$08,$FF
	.dc.b	$32,$DF,$58,$FE,$49,$F3,$92,$7F,$DC,$CF,$87,$E7,$31,$1F,$E6,$C8
	.dc.b	$B8,$F8,$23,$FD,$ED,$8A,$BF,$97,$79,$D2,$7F,$DE,$0F,$DD,$E1,$FA
	.dc.b	$51,$FE,$76,$8D,$34,$F4,$0E,$5F,$AC,$E4,$98,$78,$7C,$FF,$F7,$B5
	.dc.b	$75,$E4,$FE,$7D,$E3,$CB,$FF,$80,$0F,$CB,$F9,$79,$18,$21,$FE,$FD
	.dc.b	$7F,$A3,$7C,$14,$8F,$1E,$0F,$41,$3F,$57,$E0,$B5,$5F,$E9,$08,$B9
	.dc.b	$F9,$77,$FF,$70,$4F,$FB,$2B,$DB,$FE,$A2,$A7,$E5,$3F,$E2,$FE,$1F
	.dc.b	$EE,$2D,$38,$7A,$58,$AE,$31,$FC,$5C,$FE,$30,$E1,$C3,$5C,$C7,$ED
	.dc.b	$63,$C3,$5E,$06,$F2,$F0,$9E,$DE,$0F,$F7,$23,$D3,$7E,$79,$FE,$4D
	.dc.b	$AA,$73,$87,$D3,$6C,$FF,$2B,$F2,$9E,$1B,$EA,$BA,$6F,$32,$F9,$B8
	.dc.b	$70,$7F,$9C,$D6,$DF,$A3,$C1,$5F,$F5,$ED,$FF,$7B,$3C,$DF,$07,$23
	.dc.b	$71,$97,$AC,$7C,$CF,$B9,$71,$FC,$EE,$67,$D3,$53,$51,$2F,$FC,$A2
	.dc.b	$2B,$8B,$86,$BF,$E5,$66,$BF,$54,$E1,$1C,$DF,$94,$7F,$4F,$8E,$3B
	.dc.b	$FE,$0B,$D8,$7D,$07,$F8,$5F,$E7,$A5,$FC,$27,$F0,$BF,$54,$FA,$5A
	.dc.b	$6A,$7F,$DC,$33,$87,$F3,$3F,$45,$FD,$79,$73,$23,$0D,$F5,$3F,$01
	.dc.b	$F1,$72,$B8,$C7,$F8,$D2,$B7,$CB,$FE,$EB,$9F,$DB,$F3,$BF,$46,$FE
	.dc.b	$71,$1F,$E1,$46,$BF,$97,$8D,$79,$BF,$E2,$A9,$CC,$FE,$1D,$E3,$C1
	.dc.b	$FC,$36,$B7,$3D,$A7,$3D,$5F,$92,$F7,$FF,$C2,$53,$F3,$29,$EF,$BE
	.dc.b	$67,$FC,$87,$E1,$CA,$F3,$CE,$5B,$3F,$E3,$C8,$C4,$7F,$8B,$1A,$FA
	.dc.b	$5A,$F9,$DF,$51,$1F,$F2,$0E,$3E,$26,$C7,$F9,$82,$31,$1E,$76,$E5
	.dc.b	$FE,$7C,$F0,$FF,$CD,$BF,$B0,$7F,$AC,$2B,$FD,$43,$C0,$8C,$47,$E3
	.dc.b	$63,$1E,$A7,$FA,$06,$73,$3F,$C1,$67,$2F,$F9,$D6,$38,$DF,$F3,$99
	.dc.b	$CC,$FF,$A4,$BB,$BC,$37,$23,$83,$FE,$6C,$8C,$47,$FA,$F2,$F9,$EF
	.dc.b	$FE,$CC,$6C,$7E,$3B,$7F,$DC,$63,$DE,$5E,$77,$9C,$95,$EE,$8C,$B6
	.dc.b	$3E,$19,$1F,$E6,$74,$E6,$FE,$4F,$F0,$EF,$93,$FE,$59,$4F,$A8,$FD
	.dc.b	$33,$FE,$3D,$9F,$DF,$BF,$E3,$F8,$CB,$FE,$D5,$F0,$3F,$96,$46,$27
	.dc.b	$A6,$7B,$EB,$FF,$67,$61,$DB,$DF,$D6,$1F,$CC,$DB,$8F,$F9,$2D,$E5
	.dc.b	$3B,$AE,$37,$CE,$BC,$97,$E0,$9F,$F6,$D3,$63,$D1,$5E,$CC,$79,$DF
	.dc.b	$19,$F4,$17,$1F,$6B,$F0,$70,$EE,$D3,$9C,$F9,$63,$BC,$53,$C6,$53
	.dc.b	$FD,$F1,$7E,$97,$89,$CC,$F0,$96,$BC,$37,$E6,$8C,$4F,$A6,$BB,$C6
	.dc.b	$F0,$F7,$3D,$B0,$F5,$56,$E3,$C6,$6C,$CC,$47,$FD,$15,$5D,$31,$FA
	.dc.b	$31,$83,$CE,$F8,$F0,$C9,$27,$E2,$BE,$78,$E2,$48,$F9,$2E,$5E,$C4
	.dc.b	$9E,$AF,$CE,$E1,$FF,$4E,$47,$C6,$6B,$ED,$0E,$32,$FC,$6C,$62,$BF
	.dc.b	$A4,$2F,$A9,$FF,$4A,$57,$28,$FF,$4B,$46,$27,$6D,$8F,$EA,$0D,$7F
	.dc.b	$C0,$CF,$DB,$5F,$17,$FE,$39,$9F,$D0,$C7,$43,$E4,$1F,$EA,$38,$D6
	.dc.b	$1D,$A3,$27,$2E,$1E,$8F,$9C,$CC,$F0,$7C,$67,$E1,$CE,$05,$38,$65
	.dc.b	$E2,$8F,$F2,$63,$E4,$8E,$97,$9A,$EF,$87,$AE,$FB,$57,$B8,$EC,$F2
	.dc.b	$C3,$D1,$87,$19,$F0,$57,$EA,$5F,$1C,$77,$8C,$57,$19,$F4,$CF,$13
	.dc.b	$4F,$6B,$E4,$F9,$AF,$FC,$01,$5E,$48,$F9,$98,$78,$4E,$37,$CE,$BC
	.dc.b	$A7,$0B,$E7,$87,$89,$B7,$DC,$F4,$9F,$39,$F8,$6D,$3E,$A9,$F3,$27
	.dc.b	$A3,$C2,$BC,$0F,$E4,$BF,$61,$7D,$0F,$F9,$EF,$5F,$E8,$91,$8C,$3F
	.dc.b	$CC,$3F,$BE,$7E,$61,$FF,$3A,$EA,$9D,$61,$FE,$1C,$8C,$61,$FE,$5C
	.dc.b	$8F,$CC,$CE,$9F,$EA,$8D,$73,$1B,$7C,$9E,$47,$8E,$F8,$B6,$3F,$C3
	.dc.b	$38,$35,$C6,$3E,$3A,$B8,$9A,$71,$F0,$97,$C7,$1F,$5D,$6D,$B5,$C1
	.dc.b	$F8,$38,$5C,$7D,$D5,$B7,$FB,$93,$5F,$6D,$7A,$F1,$F2,$5F,$F8,$13
	.dc.b	$E2,$7D,$5D,$FF,$E2,$77,$C9,$97,$F8,$5E,$7F,$C3,$D1,$AE,$C3,$F1
	.dc.b	$6B,$84,$F3,$26,$3F,$E1,$7F,$A9,$D7,$71,$A4,$9F,$E1,$B5,$FE,$26
	.dc.b	$8F,$0F,$CD,$F9,$2B,$85,$7E,$02,$3C,$5E,$E4,$8C,$1F,$E3,$37,$18
	.dc.b	$7F,$8F,$DF,$36,$5F,$EE,$B9,$FF,$11,$C7,$DD,$E0,$FF,$44,$7D,$1F
	.dc.b	$D3,$F1,$BE,$96,$1F,$E6,$49,$CD,$FF,$96,$38,$B4,$47,$FD,$0F,$76
	.dc.b	$BF,$CD,$67,$FB,$47,$CC,$E9,$E2,$7E,$1E,$31,$5F,$E6,$28,$CB,$4F
	.dc.b	$F6,$7C,$68,$F1,$63,$F4,$DE,$BF,$83,$7E,$85,$BC,$7D,$7E,$A9,$F2
	.dc.b	$3D,$EB,$5D,$3F,$EB,$C7,$85,$FF,$A0,$DA,$FF,$AE,$5A,$83,$86,$DF
	.dc.b	$AF,$F5,$4F,$C1,$E5,$E8,$AE,$EF,$FB,$0E,$33,$5F,$2F,$CB,$7C,$BF
	.dc.b	$CF,$E4,$FD,$14,$62,$FF,$D0,$B1,$96,$9F,$EB,$49,$C5,$7D,$5F,$B4
	.dc.b	$7C,$AF,$EA,$DB,$5F,$E8,$2F,$AB,$FF,$72,$BF,$EE,$88,$CD,$71,$AE
	.dc.b	$07,$E8,$AB,$71,$9A,$FA,$7C,$AF,$A9,$FB,$4D,$8F,$AF,$E2,$7F,$AE
	.dc.b	$9D,$CE,$8F,$E1,$55,$FE,$9E,$6D,$71,$8F,$ED,$AD,$7F,$D6,$CE,$78
	.dc.b	$3F,$DE,$7F,$4D,$E4,$7F,$DE,$B1,$9B,$FF,$78,$4E,$5F,$F7,$1F,$23
	.dc.b	$E1,$7F,$38,$7F,$0F,$F7,$DF,$57,$8A,$FF,$59,$CD,$CB,$F4,$15,$A9
	.dc.b	$C9,$FE,$E0,$3F,$4D,$F2,$1F,$C6,$7D,$C4,$60,$FF,$5E,$C5,$AF,$D5
	.dc.b	$4E,$A7,$27,$FB,$0B,$EC,$3F,$60,$7F,$2B,$F8,$4F,$DB,$46,$49,$5F
	.dc.b	$A2,$9D,$78,$1C,$E3,$13,$AD,$FC,$71,$FE,$CC,$D5,$7C,$2B,$B9,$FE
	.dc.b	$D2,$F3,$FF,$80,$1E,$BE,$1B,$19,$25,$FA,$A3,$2F,$B3,$D8,$FE,$11
	.dc.b	$E8,$7C,$36,$D7,$EA,$72,$F1,$F1,$B6,$56,$B0,$E5,$7F,$EE,$4C,$3F
	.dc.b	$99,$FA,$0F,$BC,$A7,$F4,$7F,$11,$F2,$B8,$BF,$F6,$FD,$6D,$FE,$56
	.dc.b	$DC,$3F,$E0,$07,$94,$8C,$47,$FB,$85,$AF,$EF,$E3,$58,$7F,$23,$F9
	.dc.b	$6E,$D3,$FC,$49,$FE,$35,$F5,$4F,$84,$FC,$EB,$71,$FF,$23,$D2,$30
	.dc.b	$FF,$92,$9A,$F8,$7A,$FF,$A5,$64,$69,$2B,$F4,$BC,$AF,$93,$FE,$6C
	.dc.b	$7F,$37,$F8,$EB,$3F,$3C,$FF,$23,$7D,$4B,$C4,$FE,$BB,$F8,$1F,$F1
	.dc.b	$B9,$D6,$7F,$E5,$0A,$D4,$63,$81,$5B,$AF,$F7,$D3,$5F,$E4,$82,$36
	.dc.b	$FF,$63,$8D,$3F,$2D,$AF,$E4,$FE,$13,$FA,$71,$E4,$6D,$78,$65,$F2
	.dc.b	$BE,$83,$FB,$AF,$D4,$7C,$59,$D4,$65,$FE,$D3,$3C,$1B,$1E,$32,$3F
	.dc.b	$03,$A5,$B7,$E1,$7A,$BF,$4F,$9F,$C0,$BF,$EC,$5C,$BF,$23,$E8,$A6
	.dc.b	$3C,$7A,$DF,$51,$8D,$BE,$2D,$7C,$12,$57,$91,$B1,$FD,$6B,$D6,$E1
	.dc.b	$9F,$E9,$34,$FE,$FB,$3F,$3B,$E4,$FE,$B9,$FE,$2D,$F9,$C7,$C4,$BE
	.dc.b	$57,$39,$FC,$7A,$FA,$04,$7E,$1B,$6F,$FB,$4F,$C4,$70,$6C,$FC,$FE
	.dc.b	$DF,$C2,$9D,$F0,$FE,$55,$3F,$77,$93,$F3,$3A,$7C,$67,$F3,$31,$F7
	.dc.b	$6B,$F0,$04,$BF,$A6,$FF,$82,$FA,$FA,$C7,$D1,$D7,$ED,$57,$E3,$3C
	.dc.b	$BF,$CA,$9E,$FF,$A6,$FF,$1D,$65,$DE,$33,$7F,$79,$FC,$71,$1C,$1D
	.dc.b	$7D,$6D,$DF,$F8,$21,$FF,$1E,$F9,$1F,$C4,$27,$F3,$CD,$95,$FF,$28
	.dc.b	$E1,$62,$E3,$E3,$E3,$4C,$7F,$C6,$F8,$F1,$C7,$F4,$7D,$3F,$CB,$AB
	.dc.b	$FC,$89,$39,$9F,$EA,$7F,$29,$FE,$2D,$FA,$1F,$F1,$73,$70,$F8,$8D
	.dc.b	$9F,$F0,$FB,$74,$FC,$AD,$FF,$82,$9B,$EF,$FF,$95,$BD,$0F,$6B,$30
	.dc.b	$F6,$B9,$69,$FE,$86,$FC,$A5,$F6,$7F,$E7,$66,$C7,$FA,$2B,$E0,$7F
	.dc.b	$80,$61,$FD,$2B,$DC,$56,$B4,$FF,$46,$36,$74,$DB,$FF,$72,$DF,$27
	.dc.b	$FD,$CF,$19,$AF,$F7,$7B,$CD,$FE,$17,$7D,$B0,$FF,$3B,$4E,$67,$8C
	.dc.b	$63,$E1,$FD,$07,$F9,$DB,$D4,$B8,$7B,$62,$D7,$F9,$67,$E0,$D5,$8C
	.dc.b	$FF,$B2,$CB,$E9,$7F,$77,$EC,$7F,$38,$6C,$7F,$C0,$06,$BF,$C8,$0E
	.dc.b	$1E,$1F,$FB,$E1,$C1,$FD,$37,$D0,$74,$FF,$42,$F3,$D1,$89,$DC,$62
	.dc.b	$75,$39,$9E,$D7,$FE,$3C,$E0,$F2,$7B,$46,$92,$7C,$DE,$EF,$F3,$4D
	.dc.b	$3D,$B6,$8C,$70,$7F,$38,$8C,$57,$FB,$2B,$9B,$2F,$F9,$7E,$32,$FF
	.dc.b	$9E,$9B,$7D,$2F,$7E,$5F,$FA,$E2,$F8,$DF,$CB,$F9,$8B,$81,$A7,$0E
	.dc.b	$27,$B1,$8A,$FF,$3F,$D6,$DB,$5B,$9D,$47,$8E,$2E,$5F,$E0,$CB,$E7
	.dc.b	$3F,$E8,$8F,$D6,$FE,$33,$75,$F2,$C6,$23,$F0,$9E,$C6,$26,$46,$CF
	.dc.b	$F7,$6C,$79,$F8,$38,$F8,$B2,$7F,$DE,$31,$97,$C1,$19,$7F,$21,$1F
	.dc.b	$13,$F1,$B7,$D0,$52,$4F,$F7,$2E,$47,$C6,$FF,$87,$E3,$FD,$43,$5C
	.dc.b	$1B,$5C,$1F,$E2,$FC,$9F,$E9,$5B,$E3,$E8,$F7,$F1,$7F,$C2,$34,$C3
	.dc.b	$F8,$D3,$FE,$14,$6B,$F9,$86,$BF,$9D,$73,$C1,$FA,$66,$E1,$FE,$00
	.dc.b	$6C,$FE,$7B,$2F,$F0,$C3,$E8,$BF,$13,$4F,$74,$69,$FF,$62,$CB,$EA
	.dc.b	$7F,$23,$97,$C6,$F9,$5B,$A1,$F9,$53,$2F,$E2,$B5,$F8,$67,$EC,$3B
	.dc.b	$3F,$BA,$BA,$21,$E0,$8E,$F1,$A8,$F1,$F9,$5E,$2F,$DA,$BF,$0D,$B7
	.dc.b	$F4,$36,$ED,$F0,$67,$EE,$12,$BE,$BB,$D3,$E8,$F1,$1D,$0C,$9F,$7F
	.dc.b	$04,$7E,$A3,$DA,$7C,$B7,$C2,$FC,$91,$F3,$14,$D4,$4C,$3F,$DC,$B5
	.dc.b	$C1,$B1,$FB,$2F,$65,$F7,$5B,$9F,$F7,$9D,$6D,$B5,$B9,$F1,$7C,$B7
	.dc.b	$D2,$E8,$7C,$8D,$33,$7E,$09,$AF,$47,$EC,$FE,$00,$FA,$AC,$F1,$FF
	.dc.b	$2E,$CD,$58,$E2,$FC,$C5,$EF,$0F,$D1,$7E,$49,$F8,$CE,$83,$4F,$C7
	.dc.b	$C3,$F0,$1F,$F4,$DB,$FE,$5A,$31,$E3,$7F,$12,$8C,$C7,$E1,$0C,$9F
	.dc.b	$AF,$D7,$F2,$5A,$7F,$7F,$6E,$5F,$E5,$CA,$E3,$3F,$D1,$9B,$1F,$AF
	.dc.b	$9D,$D7,$FB,$03,$EF,$7F,$DC,$8F,$F9,$B2,$51,$88,$FC,$AC,$63,$E2
	.dc.b	$FF,$3F,$93,$E0,$3D,$1E,$44,$BF,$66,$5F,$C7,$1F,$CE,$EA,$FC,$87
	.dc.b	$FA,$96,$BF,$E1,$76,$E7,$FC,$0D,$B5,$FD,$49,$BF,$72,$F1,$BF,$23
	.dc.b	$3F,$27,$EA,$DF,$1F,$FC,$21,$B1,$FE,$9A,$6B,$FD,$79,$AF,$F6,$37
	.dc.b	$38,$FF,$6C,$6E,$3F,$E8,$66,$C8,$7E,$03,$FA,$C7,$11,$A4,$D3,$11
	.dc.b	$BF,$3F,$FC,$F3,$F4,$3F,$D4,$BE,$63,$DF,$06,$FC,$1A,$7F,$AB,$B7
	.dc.b	$0F,$F4,$14,$F0,$37,$3C,$FC,$3E,$9F,$3F,$2B,$71,$97,$8C,$65,$FE
	.dc.b	$23,$87,$86,$7F,$B1,$FE,$D4,$F7,$6F,$D4,$CB,$FC,$C4,$E5,$B3,$97
	.dc.b	$F3,$10,$9E,$73,$27,$8C,$6F,$7F,$F0,$DD,$73,$9F,$D0,$7B,$5F,$EC
	.dc.b	$17,$F4,$CD,$73,$E0,$7F,$78,$FD,$AB,$FD,$B5,$1A,$8F,$EF,$73,$BB
	.dc.b	$FF,$13,$B7,$8B,$FD,$83,$7C,$2F,$E0,$EF,$FE,$4C,$7E,$6F,$E0,$DF
	.dc.b	$54,$FA,$DE,$FC,$8C,$90,$E2,$F8,$3E,$23,$FA,$CE,$26,$92,$7F,$DD
	.dc.b	$B1,$97,$DB,$1C,$5F,$E0,$9E,$2F,$ED,$9F,$F5,$BC,$65,$F8,$AF,$F8
	.dc.b	$62,$3F,$77,$C3,$F8,$9F,$8F,$FC,$67,$C0,$CB,$FC,$3C,$7E,$3C,$C3
	.dc.b	$F0,$23,$0F,$F1,$3F,$D8,$DF,$11,$FE,$24,$6C,$66,$3F,$C4,$FF,$84
	.dc.b	$3E,$67,$51,$27,$47,$F1,$D9,$D3,$FC,$7F,$F6,$4F,$F2,$5B,$FE,$50
	.dc.b	$0E,$1F,$F2,$9B,$5F,$F2,$C3,$5F,$F7,$CB,$BE,$3F,$F3,$1B,$78,$F3
	.dc.b	$E6,$FF,$8F,$3C,$0F,$C6,$56,$A3,$FC,$7F,$5B,$BF,$DE,$35,$FE,$1D
	.dc.b	$D3,$FE,$44,$F2,$7F,$A5,$C7,$F3,$76,$BF,$CD,$5B,$7F,$E7,$4F,$C7
	.dc.b	$3E,$C4,$63,$2F,$E1,$15,$FC,$B3,$CB,$E8,$7F,$94,$2B,$F9,$9F,$C6
	.dc.b	$BF,$05,$7C,$BF,$F9,$72,$30,$D8,$CD,$FF,$02,$FA,$7C,$4C,$8F,$E2
	.dc.b	$D9,$B2,$FE,$31,$87,$29,$E0,$7F,$51,$FB,$77,$C7,$FF,$98,$9B,$19
	.dc.b	$8F,$E2,$DF,$2B,$96,$46,$4F,$E2,$71,$A6,$90,$FD,$6F,$C0,$65,$FD
	.dc.b	$7C,$5C,$7C,$D0,$EA,$FE,$45,$5F,$07,$F0,$67,$DC,$52,$57,$BB,$4F
	.dc.b	$E6,$96,$FC,$3B,$FC,$7B,$AB,$0E,$37,$B1,$19,$C3,$9C,$7F,$22,$AF
	.dc.b	$EE,$FD,$9F,$D4,$A0,$D3,$F9,$BF,$C5,$BE,$7F,$FD,$13,$19,$9A,$7F
	.dc.b	$A1,$E2,$BF,$DF,$E3,$4F,$8F,$25,$73,$8D,$1C,$0F,$F0,$14,$F8,$2B
	.dc.b	$F0,$9F,$9C,$3C,$0A,$4A,$E9,$F5,$6F,$B9,$C4,$7F,$93,$B5,$5F,$18
	.dc.b	$F8,$91,$8B,$FF,$52,$DF,$19,$FF,$0B,$7C,$A7,$FB,$49,$FF,$62,$7C
	.dc.b	$AF,$F9,$DA,$31,$1F,$EC,$48,$C1,$FE,$AC,$8C,$F1,$FF,$B6,$1B,$C5
	.dc.b	$F6,$3F,$57,$FE,$61,$7F,$E1,$F9,$CB,$FD,$BD,$A9,$CF,$02,$78,$DF
	.dc.b	$5F,$04,$FF,$77,$ED,$FB,$38,$73,$F6,$FF,$67,$18,$8E,$91,$8E,$2F
	.dc.b	$DB,$E5,$FE,$0F,$BE,$51,$8C,$3A,$DB,$EC,$39,$5F,$F8,$7D,$B8,$7F
	.dc.b	$00,$73,$5C,$1B,$87,$F8,$2D,$DD,$DC,$3C,$37,$C5,$F0,$56,$67,$E2
	.dc.b	$5F,$39,$DC,$64,$F5,$97,$0F,$53,$E7,$8E,$6D,$3E,$21,$BB,$F0,$3D
	.dc.b	$AF,$95,$F7,$8D,$9C,$4E,$E5,$BE,$F8,$7B,$A3,$17,$D1,$DC,$64,$98
	.dc.b	$7A,$A3,$67,$83,$08,$E0,$93,$C1,$F3,$0F,$92,$FE,$6B,$87,$C0,$6E
	.dc.b	$3F,$61,$B7,$FE,$3B,$7F,$9D,$D7,$F8,$F7,$E3,$F1,$9F,$F3,$88,$C1
	.dc.b	$D5,$C3,$FE,$E0,$30,$6E,$38,$B4,$F8,$C3,$D4,$FE,$3F,$97,$E4,$63
	.dc.b	$AB,$63,$39,$F6,$AF,$B9,$5C,$77,$AF,$86,$F8,$46,$27,$77,$CA,$BC
	.dc.b	$55,$A7,$C6,$71,$BF,$74,$EE,$3F,$AB,$D7,$2B,$E9,$1E,$A9,$D3,$E8
	.dc.b	$AE,$33,$AB,$EC,$7B,$67,$B1,$FD,$8A,$B8,$B8,$3F,$CD,$8E,$2F,$C1
	.dc.b	$1E,$66,$CF,$36,$DF,$C5,$9F,$F5,$FB,$E2,$69,$F4,$70,$F8,$07,$CF
	.dc.b	$F3,$BF,$A7,$E7,$FD,$42,$7F,$CF,$91,$8C,$BF,$8A,$5D,$9F,$EF,$33
	.dc.b	$5D,$CF,$FA,$E5,$46,$F8,$3F,$8A,$56,$E7,$35,$BA,$F0,$47,$E3,$5A
	.dc.b	$F6,$8F,$F1,$23,$C0,$FC,$97,$27,$F3,$8F,$85,$F2,$36,$7F,$9F,$FD
	.dc.b	$B6,$39,$FF,$DD,$31,$FC,$F1,$B7,$FD,$05,$B9,$FF,$70,$CD,$CC,$C3
	.dc.b	$96,$5E,$FA,$FF,$58,$63,$FE,$D7,$6E,$3F,$E8,$8E,$A7,$D2,$EF,$92
	.dc.b	$5E,$DF,$89,$3A,$8C,$12,$BF,$DB,$D7,$D0,$F2,$4E,$E3,$35,$D1,$B5
	.dc.b	$FD,$A6,$75,$5F,$6C,$E5,$1F,$E2,$0F,$05,$F3,$3F,$21,$B1,$FD,$C7
	.dc.b	$B9,$F3,$F9,$A9,$3F,$D9,$2B,$78,$FF,$81,$6B,$2F,$FA,$23,$E5,$3F
	.dc.b	$01,$EB,$1B,$E8,$E1,$98,$7F,$AE,$1C,$B4,$FF,$0A,$57,$F8,$53,$D6
	.dc.b	$FC,$77,$29,$E3,$1F,$D7,$27,$FC,$97,$39,$78,$B6,$34,$7F,$90,$27
	.dc.b	$8E,$BF,$9F,$9D,$3F,$E3,$05,$F6,$DF,$8F,$8C,$36,$FF,$C5,$B9,$3E
	.dc.b	$B7,$F2,$33,$9C,$7F,$C5,$6D,$C3,$9F,$47,$C4,$AD,$BF,$A9,$F7,$5E
	.dc.b	$6D,$3F,$5D,$DE,$FD,$27,$3A,$7C,$4A,$B9,$FB,$03,$8D,$7F,$74,$D9
	.dc.b	$FE,$17,$56,$BF,$DF,$13,$57,$EC,$7F,$EE,$DA,$FA,$2D,$8F,$93,$83
	.dc.b	$F4,$27,$C6,$7F,$01,$AD,$65,$FC,$25,$B8,$7F,$9F,$9B,$87,$F3,$FB
	.dc.b	$E3,$3D,$BE,$0B,$91,$FC,$02,$7C,$6D,$AF,$14,$7F,$51,$6B,$FE,$8A
	.dc.b	$FB,$3F,$F7,$A5,$7E,$21,$B1,$E8,$B7,$BB,$FC,$5E,$73,$8F,$F1,$F6
	.dc.b	$DF,$3F,$63,$E8,$DF,$08,$E9,$1F,$9C,$F9,$2F,$CE,$FA,$36,$7F,$CB
	.dc.b	$1C,$DF,$E7,$46,$CE,$9A,$6A,$7F,$94,$7E,$2D,$FE,$9C,$8F,$CF,$B7
	.dc.b	$4F,$CB,$FE,$30,$FC,$61,$F4,$BC,$3D,$A9,$87,$07,$F9,$A5,$6E,$73
	.dc.b	$7F,$85,$C3,$A3,$FD,$E5,$AF,$FA,$C9,$A8,$38,$F2,$5A,$3E,$47,$F7
	.dc.b	$BF,$80,$7E,$F9,$B9,$FC,$26,$E1,$FB,$0B,$E3,$18,$E2,$FE,$9D,$F6
	.dc.b	$1F,$E7,$7B,$FD,$FB,$6F,$EC,$7C,$CF,$F4,$D8,$C5,$70,$F6,$30,$7D
	.dc.b	$69,$B7,$FE,$9B,$FC,$91,$F2,$9F,$EB,$F8,$C3,$F9,$0C,$7F,$81,$CE
	.dc.b	$A7,$FA,$E3,$C3,$AE,$D7,$1A,$DB,$C7,$1F,$F5,$AD,$71,$BE,$B7,$C1
	.dc.b	$E2,$D7,$27,$E1,$BD,$0F,$E4,$BE,$7F,$F2,$66,$C6,$4F,$C5,$C6,$5A
	.dc.b	$7E,$3B,$C9,$FE,$5B,$F3,$FF,$CB,$9B,$19,$3F,$27,$19,$69,$F3,$A3
	.dc.b	$E6,$1F,$DA,$3F,$0A,$76,$E2,$74,$4D,$7F,$D4,$D3,$A8,$C6,$9F,$6A
	.dc.b	$FF,$71,$C1,$FD,$22,$7F,$C8,$76,$8C,$3F,$E4,$66,$BF,$E4,$BF,$9F
	.dc.b	$FE,$63,$3F,$91,$84,$61,$FF,$7C,$7C,$7F,$F2,$D7,$3E,$5F,$F8,$47
	.dc.b	$F0,$9F,$F2,$14,$7F,$23,$6E,$5F,$E1,$BB,$E5,$29,$CE,$1F,$9F,$F9
	.dc.b	$77,$EC,$CF,$07,$FC,$3D,$18,$7F,$C4,$1C,$AF,$89,$F2,$4E,$6F,$FA
	.dc.b	$BE,$4F,$86,$F2,$79,$5F,$E2,$58,$C4,$FF,$5C,$CF,$F9,$BB,$67,$6F
	.dc.b	$2A,$BA,$7F,$58,$CB,$C1,$1E,$2C,$B9,$CF,$BE,$F1,$3F,$0C,$97,$FD
	.dc.b	$AB,$F0,$B7,$D2,$62,$FE,$35,$5D,$FF,$C8,$15,$AB,$F3,$4E,$5F,$E0
	.dc.b	$99,$3E,$25,$D9,$9C,$7F,$5F,$18,$AF,$77,$A9,$F5,$4F,$F2,$36,$2F
	.dc.b	$A0,$C4,$E8,$97,$FD,$6E,$75,$18,$CF,$FC,$A9,$BF,$F8,$2E,$3F,$8B
	.dc.b	$41,$F2,$FE,$46,$F8,$1E,$B8,$7C,$98,$C4,$FF,$83,$3F,$2D,$FE,$79
	.dc.b	$3F,$DA,$5B,$7C,$E3,$F8,$49,$FE,$41,$9E,$1E,$7F,$F9,$7E,$B7,$1F
	.dc.b	$97,$6C,$F0,$F8,$0D,$54,$AF,$39,$F2,$BF,$69,$F8,$D5,$E1,$9F,$C6
	.dc.b	$C6,$30,$FF,$39,$60,$FB,$9F,$01,$F9,$2E,$27,$66,$32,$F6,$4D,$CB
	.dc.b	$F1,$15,$A9,$C9,$23,$F1,$9F,$03,$F1,$1F,$85,$5F,$9E,$8C,$63,$E2
	.dc.b	$FA,$5F,$80,$7E,$89,$FF,$10,$5F,$CB,$8C,$B4,$F7,$DF,$F3,$29,$DB
	.dc.b	$F3,$E7,$6F,$FB,$23,$24,$7F,$80,$5B,$5F,$C2,$9B,$3B,$E2,$FA,$47
	.dc.b	$8F,$C9,$B5,$F7,$62,$CF,$EA,$70,$FE,$B1,$39,$3F,$CC,$DE,$87,$FA
	.dc.b	$67,$F2,$B7,$D5,$FF,$9D,$23,$0D,$9F,$F2,$9C,$E9,$E2,$62,$3F,$D4
	.dc.b	$5E,$E7,$E9,$63,$FC,$B2,$DC,$3F,$D0,$93,$AE,$99,$5A,$7F,$D0,$71
	.dc.b	$89,$FF,$1E,$57,$F8,$2D,$BE,$AF,$FA,$23,$EE,$3B,$B3,$0E,$EE,$5A
	.dc.b	$49,$E2,$FF,$A1,$67,$84,$76,$F3,$3D,$27,$FA,$53,$ED,$ED,$7A,$A2
	.dc.b	$CF,$FA,$76,$6E,$BF,$EC,$7F,$B1,$7C,$F7,$FB,$EE,$BF,$9D,$1F,$E0
	.dc.b	$B8,$CF,$BD,$FD,$11,$B1,$A0,$E1,$E3,$F1,$EF,$9A,$FD,$9C,$61,$AF
	.dc.b	$F4,$78,$D7,$4F,$F5,$06,$C7,$5A,$FF,$4F,$FC,$87,$87,$E2,$4D,$24
	.dc.b	$EF,$EC,$BF,$A9,$69,$ED,$6E,$FF,$D5,$70,$46,$29,$EC,$BE,$31,$F5
	.dc.b	$59,$9F,$6D,$5B,$E3,$1F,$E9,$E6,$BF,$EC,$99,$D7,$D9,$7F,$B8,$24
	.dc.b	$D2,$4F,$CB,$76,$F3,$7C,$98,$78,$FE,$53,$33,$FE,$87,$AE,$6E,$CF
	.dc.b	$F6,$44,$79,$FE,$CB,$E2,$C7,$FA,$C9,$AF,$FA,$3D,$BA,$7F,$B5,$3C
	.dc.b	$1E,$1B,$4A,$E0,$F8,$B5,$4D,$E2,$ED,$F6,$52,$FF,$DB,$11,$97,$C1
	.dc.b	$19,$7F,$DD,$91,$F2,$FE,$CB,$FB,$74,$7F,$83,$9A,$FE,$2D,$AF,$E4
	.dc.b	$9A,$FF,$83,$5D,$E3,$FC,$FE,$11,$8E,$1D,$79,$FF,$77,$95,$F2,$DF
	.dc.b	$C1,$AF,$FD,$AB,$7F,$ED,$D7,$F8,$FC,$BE,$A7,$94,$7F,$94,$9A,$F6
	.dc.b	$F4,$3D,$47,$F8,$99,$C1,$EB,$71,$5D,$63,$D9,$1C,$1F,$0B,$E5,$B6
	.dc.b	$9F,$17,$DE,$F4,$65,$FE,$00,$BF,$2B,$5F,$F6,$46,$09,$E1,$1F,$2D
	.dc.b	$F0,$91,$C1,$F0,$1D,$BF,$E1,$5E,$AF,$E0,$2F,$C4,$F5,$BE,$47,$D9
	.dc.b	$7E,$1B,$6F,$F0,$4B,$67,$FD,$1D,$87,$F8,$F9,$E0,$7D,$BA,$F8,$24
	.dc.b	$70,$68,$F4,$BF,$56,$DF,$65,$FA,$FF,$37,$FA,$5D,$7D,$4F,$B2,$FF
	.dc.b	$0B,$B7,$D6,$D7,$9D,$23,$FC,$AF,$5C,$1F,$E2,$8D,$7F,$8D,$B5,$FF
	.dc.b	$2B,$3B,$E4,$FC,$53,$75,$CD,$7F,$01,$F6,$DF,$05,$E5,$F3,$BF,$19
	.dc.b	$D1,$FE,$17,$CB,$F1,$2D,$7F,$DF,$98,$FF,$89,$DF,$ED,$67,$F8,$02
	.dc.b	$7F,$11,$19,$8F,$3C,$60,$8F,$08,$D6,$5E,$69,$FC,$E7,$D9,$7F,$B9
	.dc.b	$2B,$F3,$3F,$65,$F8,$FA,$FF,$11,$75,$23,$1A,$FF,$90,$9B,$5F,$EC
	.dc.b	$7F,$B2,$FE,$3E,$F9,$9A,$FF,$3D,$6B,$FD,$11,$AF,$F9,$FD,$AF,$FA
	.dc.b	$E3,$D9,$FC,$DB,$CF,$C5,$FE,$55,$3D,$31,$F2,$D6,$A9,$1F,$E3,$9C
	.dc.b	$1F,$4C,$F8,$7F,$E6,$D1,$86,$C6,$6B,$E0,$FA,$BD,$BC,$5F,$DA,$64
	.dc.b	$69,$27,$FA,$C6,$AF,$B2,$FE,$BF,$9F,$C8,$6E,$7F,$E6,$86,$D3,$E0
	.dc.b	$BF,$91,$F4,$FF,$45,$FB,$17,$D9,$7F,$47,$F1,$7F,$D6,$77,$F8,$06
	.dc.b	$BE,$EC,$7F,$D4,$D5,$C2,$3F,$0A,$D7,$FB,$63,$5F,$EE,$6D,$78,$D7
	.dc.b	$F1,$18,$E1,$97,$F2,$36,$D6,$AB,$C5,$E4,$7F,$4C,$FB,$27,$CB,$FF
	.dc.b	$4E,$F9,$0F,$6F,$9C,$6C,$EA,$FF,$C8,$72,$8C,$47,$AC,$FE,$D3,$3C
	.dc.b	$8E,$35,$D7,$C2,$ED,$F0,$D2,$B9,$D7,$F8,$3D,$E5,$87,$86,$B5,$18
	.dc.b	$95,$73,$8F,$D5,$FD,$8B,$E5,$BF,$BB,$2F,$B2,$A6,$22,$4F,$08,$CB
	.dc.b	$CA,$32,$FA,$C4,$F3,$78,$1B,$3F,$77,$EB,$FF,$9E,$9F,$DE,$FD,$97
	.dc.b	$F2,$37,$93,$78,$7F,$AB,$4E,$E7,$B3,$63,$3E,$17,$F8,$3E,$31,$C3
	.dc.b	$FE,$EB,$B7,$C7,$FD,$7F,$7B,$EC,$7C,$74,$CB,$E5,$C6,$5F,$9B,$19
	.dc.b	$7E,$54,$75,$FB,$2F,$F0,$E7,$C4,$3E,$D2,$99,$AF,$6C,$FF,$5D,$CC
	.dc.b	$D2,$47,$F9,$04,$FF,$24,$BC,$5F,$6B,$DB,$D4,$FC,$77,$CA,$FE,$3F
	.dc.b	$E9,$E5,$7E,$0A,$32,$FE,$12,$32,$FE,$06,$3F,$C6,$5F,$66,$FB,$3F
	.dc.b	$F1,$A7,$D9,$FF,$BF,$EB,$F9,$6B,$5F,$B5,$3B,$7F,$98,$B5,$CC,$F4
	.dc.b	$39,$FD,$37,$F8,$FF,$EC,$FF,$CB,$13,$FC,$E3,$DE,$7C,$F7,$F9,$0D
	.dc.b	$B1,$FC,$F2,$76,$FE,$25,$AF,$F9,$A9,$A8,$8E,$1B,$E9,$FF,$82,$E3
	.dc.b	$9F,$47,$F8,$75,$BC,$FC,$69,$18,$7A,$5F,$E8,$DF,$E9,$FE,$6F,$F5
	.dc.b	$3B,$FF,$13,$7C,$BF,$DF,$9F,$EA,$6D,$7F,$41,$3B,$C3,$B3,$63,$D0
	.dc.b	$D7,$F9,$AF,$A1,$FE,$65,$AF,$E7,$57,$FE,$F1,$FA,$67,$D9,$BE,$47
	.dc.b	$F8,$77,$BB,$31,$FF,$66,$D6,$A7,$FC,$E4,$D8,$CD,$71,$6C,$FF,$17
	.dc.b	$FA,$6F,$F3,$A7,$D9,$BE,$AB,$FC,$64,$DA,$FC,$A3,$5F,$F4,$B3,$5F
	.dc.b	$F4,$EB,$9C,$BF,$C9,$4D,$DB,$E7,$36,$7F,$8E,$FC,$9F,$F2,$72,$3F
	.dc.b	$C6,$9F,$E5,$5F,$0B,$88,$FE,$31,$16,$D7,$F8,$1F,$2F,$F9,$75,$FC
	.dc.b	$0F,$B3,$7C,$9F,$EA,$F7,$46,$32,$FF,$57,$46,$27,$53,$99,$FF,$5B
	.dc.b	$CE,$5F,$9B,$19,$C7,$EA,$37,$1F,$17,$D6,$7F,$80,$A9,$DD,$FE,$5B
	.dc.b	$69,$FE,$5D,$70,$47,$FD,$48,$FF,$64,$76,$7F,$46,$F8,$B7,$D9,$EA
	.dc.b	$21,$FE,$22,$9F,$E9,$F3,$A8,$C1,$A3,$F9,$94,$6A,$7F,$0B,$F5,$EF
	.dc.b	$07,$E3,$36,$BE,$2B,$5F,$86,$D7,$F5,$1A,$23,$5A,$7E,$5D,$B9,$FE
	.dc.b	$D6,$75,$7F,$EB,$2F,$95,$7D,$9F,$FB,$0F,$E5,$5F,$67,$FE,$B6,$FC
	.dc.b	$6B,$F5,$59,$FD,$ED,$3F,$D9,$71,$8C,$FF,$DB,$D5,$C4,$FF,$44,$5F
	.dc.b	$2A,$FF,$70,$46,$08,$FF,$92,$23,$4F,$F5,$56,$DF,$7E,$47,$DA,$3C
	.dc.b	$DF,$37,$83,$FE,$2C,$C1,$F6,$9F,$72,$DF,$AE,$3B,$7F,$D9,$ED,$AF
	.dc.b	$05,$F1,$BF,$BB,$9F,$FB,$0B,$3F,$E3,$5E,$16,$3E,$13,$F8,$A3,$70
	.dc.b	$CF,$3F,$FB,$F7,$1F,$14,$F2,$7A,$47,$FC,$01,$AF,$E3,$98,$F1,$BF
	.dc.b	$3C,$F1,$C3,$F9,$44,$F1,$6C,$71,$9A,$7F,$2D,$C7,$FC,$80,$E0,$3E
	.dc.b	$97,$A9,$BC,$3C,$6F,$1C,$3C,$75,$D6,$36,$72,$9D,$B6,$FD,$EF,$12
	.dc.b	$9E,$0A,$F8,$B5,$EB,$86,$3E,$B7,$AE,$1E,$86,$D7,$89,$F3,$5F,$AE
	.dc.b	$7D,$55,$F1,$C9,$7D,$5E,$86,$2B,$C8,$78,$23,$47,$8F,$DD,$FF,$4C
	.dc.b	$57,$C5,$6E,$3F,$89,$6E,$1F,$D6,$5F,$F4,$F5,$FF,$5B,$F9,$4C,$63
	.dc.b	$FD,$56,$30,$27,$6F,$FB,$C8,$C4,$6D,$C4,$6A,$7E,$4C,$3D,$CF,$D6
	.dc.b	$57,$F0,$E8,$EA,$D8,$CE,$5C,$EB,$8C,$62,$BC,$91,$8E,$3E,$B1,$E3
	.dc.b	$D7,$94,$7E,$1E,$3D,$31,$87,$F0,$27,$82,$FD,$D7,$D6,$3F,$BA,$5F
	.dc.b	$57,$E7,$1E,$03,$8C,$7F,$05,$AF,$54,$78,$6B,$9D,$F8,$63,$07,$F7
	.dc.b	$FA,$E1,$3E,$D8,$E7,$3B,$BF,$45,$6A,$7C,$17,$E9,$3B,$CF,$97,$0B
	.dc.b	$3E,$63,$FC,$29,$5E,$B7,$E0,$E1,$FD,$5F,$67,$81,$FC,$77,$4F,$E3
	.dc.b	$D5,$F0,$0F,$D5,$F0,$FF,$0C,$9D,$69,$F8,$3E,$2B,$5F,$E1,$96,$D7
	.dc.b	$F1,$8C,$73,$3F,$E2,$59,$AF,$F1,$BA,$FF,$27,$34,$FF,$28,$47,$89
	.dc.b	$ED,$1A,$8E,$B8,$FF,$BE,$23,$FD,$E3,$3C,$71,$FE,$51,$39,$8F,$F4
	.dc.b	$D7,$4F,$F4,$5F,$8D,$F3,$B6,$FF,$D4,$3D,$2F,$A2,$CF,$67,$1C,$BF
	.dc.b	$A5,$61,$E3,$C5,$8F,$17,$FA,$9C,$66,$31,$1F,$ED,$0E,$4F,$83,$47
	.dc.b	$E2,$F7,$3E,$D7,$31,$34,$FF,$25,$57,$D2,$38,$3F,$DC,$67,$57,$FE
	.dc.b	$9B,$9C,$B6,$BA,$CE,$A7,$94,$7A,$67,$FD,$E6,$D8,$E3,$1E,$7A,$FF
	.dc.b	$09,$F8,$AF,$9C,$F8,$ED,$8F,$F6,$3F,$85,$FE,$98,$7F,$C7,$F1,$9A
	.dc.b	$F9,$ED,$7F,$BA,$B5,$FE,$E4,$D7,$E5,$B7,$8F,$FD,$AD,$F3,$1F,$8A
	.dc.b	$F8,$13,$7D,$1C,$33,$0F,$F6,$7B,$96,$9F,$28,$FF,$02,$FB,$5F,$92
	.dc.b	$E0,$3C,$72,$FF,$62,$46,$27,$84,$7F,$81,$E7,$E4,$4E,$63,$FC,$6F
	.dc.b	$18,$3F,$01,$F6,$BF,$91,$AF,$F0,$73,$67,$FD,$D3,$F6,$BF,$90,$8C
	.dc.b	$59,$D6,$6F,$C8,$7F,$C9,$78,$7F,$9A,$A7,$31,$8C,$3F,$DB,$5F,$09
	.dc.b	$FD,$F6,$BC,$F1,$FE,$EA,$F1,$3F,$67,$DC,$FB,$6D,$4F,$09,$FD,$9D
	.dc.b	$E3,$93,$FC,$39,$7B,$EF,$CD,$72,$C3,$F8,$2E,$3D,$BE,$D9,$F0,$FF
	.dc.b	$BC,$6C,$7C,$8C,$1F,$6C,$F9,$3D,$5C,$AF,$F6,$65,$F0,$D7,$FA,$33
	.dc.b	$5F,$E9,$CD,$7F,$D4,$8E,$F4,$FF,$19,$7D,$F3,$F3,$E7,$DD,$75,$B5
	.dc.b	$6A,$73,$87,$C6,$FB,$47,$C6,$7F,$14,$6C,$7A,$75,$7C,$46,$6A,$4E
	.dc.b	$7E,$B2,$E5,$DA,$3F,$96,$DA,$7F,$31,$F3,$BF,$99,$F6,$6D,$7E,$33
	.dc.b	$A5,$DF,$FE,$84,$8C,$2A,$E1,$18,$6E,$3D,$CF,$F6,$77,$DB,$FF,$AC
	.dc.b	$67,$F0,$2D,$8D,$7D,$C3,$EE,$7F,$DB,$B5,$AF,$7F,$F9,$9D,$EE,$FF
	.dc.b	$BE,$35,$FF,$69,$35,$EF,$1F,$C2,$63,$B7,$17,$FB,$0F,$EF,$BF,$D6
	.dc.b	$D7,$D4,$FF,$60,$B8,$BF,$46,$9F,$03,$E7,$7F,$79,$F7,$5F,$BD,$6D
	.dc.b	$FC,$26,$D7,$0D,$91,$8C,$7F,$AA,$FD,$C7,$FA,$1E,$FF,$42,$DB,$F8
	.dc.b	$DD,$9F,$92,$FA,$5F,$EA,$33,$9B,$E3,$F0,$1F,$A6,$E5,$7D,$2E,$A3
	.dc.b	$0F,$F5,$CF,$BA,$7D,$0F,$EC,$6F,$FD,$05,$18,$F0,$75,$39,$D5,$AA
	.dc.b	$B5,$1F,$C3,$1E,$6F,$0A,$E2,$D3,$55,$CE,$B9,$3F,$AB,$8C,$9B,$8F
	.dc.b	$8D,$EB,$7F,$27,$F5,$3F,$94,$36,$7F,$1A,$D9,$DC,$7E,$42,$7F,$97
	.dc.b	$7C,$B7,$F2,$FF,$AA,$B3,$F9,$46,$D7,$D3,$6E,$7F,$E9,$5E,$34,$7C
	.dc.b	$EB,$97,$FC,$AB,$CC,$7D,$D3,$C2,$CE,$5F,$9B,$1C,$7F,$D2,$DF,$F2
	.dc.b	$43,$5F,$F2,$83,$5F,$9A,$70,$C3,$FA,$E4,$71,$CF,$FC,$35,$F8,$AB
	.dc.b	$97,$BC,$61,$FF,$09,$FC,$6F,$F3,$B8,$FF,$0F,$B7,$4F,$B2,$3E,$6F
	.dc.b	$F9,$D6,$49,$D3,$F7,$BE,$E3,$FC,$B5,$5F,$E5,$F8,$46,$1F,$F3,$13
	.dc.b	$5F,$F3,$53,$5F,$C3,$FD,$23,$EE,$FF,$81,$57,$F5,$28,$C7,$89,$FE
	.dc.b	$2E,$E6,$7C,$97,$8A,$73,$5F,$D0,$B2,$7C,$87,$F8,$C3,$37,$A7,$F0
	.dc.b	$23,$0C,$AF,$44,$E9,$B1,$FE,$07,$8C,$69,$FE,$81,$CB,$C3,$7E,$3C
	.dc.b	$BB,$E1,$FD,$4A,$F7,$B7,$FB,$9E,$F7,$39,$BF,$21,$FE,$44,$FB,$A7
	.dc.b	$DD,$E2,$74,$79,$27,$C4,$47,$07,$F7,$6A,$FF,$2C,$F0,$F1,$CF,$F9
	.dc.b	$34,$62,$7C,$9E,$D7,$C3,$3C,$1F,$5B,$63,$27,$C2,$8F,$80,$47,$07
	.dc.b	$7F,$BB,$FF,$5E,$BC,$2F,$BC,$61,$FC,$E7,$DD,$7F,$4A,$9F,$EE,$AD
	.dc.b	$8F,$29,$FE,$4F,$8C,$FA,$3F,$E6,$69,$D4,$FF,$AB,$DA,$FE,$81,$AF
	.dc.b	$F0,$E6,$A7,$B6,$DF,$47,$EE,$FF,$45,$7D,$8F,$90,$E2,$BC,$FE,$8F
	.dc.b	$84,$FC,$76,$9F,$A1,$AF,$4C,$65,$FC,$F6,$1F,$D7,$DB,$1F,$E4,$D8
	.dc.b	$C5,$7F,$B0,$5B,$87,$E0,$FE,$0F,$E3,$3F,$E8,$E8,$CD,$FF,$28,$C3
	.dc.b	$FC,$CD,$F1,$7F,$0F,$C7,$93,$F7,$ED,$EF,$E3,$2F,$9F,$5E,$12,$57
	.dc.b	$99,$B1,$F8,$4E,$DF,$F2,$5C,$FD,$49,$FD,$64,$62,$7E,$CF,$0F,$D3
	.dc.b	$3F,$61,$87,$F5,$2B,$FB,$71,$97,$E1,$BF,$04,$8F,$FB,$87,$CE,$7C
	.dc.b	$37,$F6,$88,$FE,$33,$F7,$6F,$BB,$FF,$41,$FA,$9F,$EA,$68,$C4,$FF
	.dc.b	$81,$EB,$86,$5F,$D3,$27,$2F,$F6,$C8,$CA,$8C,$E5,$FE,$BA,$F8,$CF
	.dc.b	$F1,$CC,$FF,$2F,$6C,$2F,$FA,$BE,$5E,$6C,$BF,$D6,$F5,$A9,$FF,$5E
	.dc.b	$7C,$0E,$0F,$F4,$CC,$D8,$FF,$28,$CF,$F8,$6F,$0E,$DE,$CF,$FA,$7B
	.dc.b	$EE,$F9,$B2,$B6,$E5,$A4,$9F,$E7,$A7,$FB,$1E,$3F,$D3,$15,$D6,$3F
	.dc.b	$DA,$11,$8F,$3F,$CE,$7F,$AB,$3E,$EE,$DF,$92,$2D,$7C,$79,$BA,$7F
	.dc.b	$B7,$BE,$E5,$F4,$EF,$5F,$F9,$D4,$61,$B1,$9B,$F2,$89,$D3,$8F,$95
	.dc.b	$7D,$0F,$F8,$22,$30,$D7,$DB,$1A,$EA,$FE,$90,$D8,$EB,$5F,$EB,$1F
	.dc.b	$96,$F1,$FC,$79,$A4,$9D,$FD,$DF,$FB,$AA,$8D,$24,$7B,$1F,$ED,$3A
	.dc.b	$7C,$16,$EF,$FD,$B2,$B8,$D7,$E4,$7E,$D3,$2F,$F8,$72,$1F,$77,$8D
	.dc.b	$DF,$9A,$BE,$4B,$0D,$BF,$F4,$AD,$F0,$6D,$F9,$C8,$FD,$06,$BF,$E0
	.dc.b	$58,$E0,$D8,$F1,$F4,$FE,$B3,$E6,$FC,$B2,$FB,$CF,$8C,$FF,$A9,$1A
	.dc.b	$FF,$83,$DB,$B7,$E8,$61,$5F,$B6,$F1,$79,$D6,$B6,$FF,$07,$B6,$BF
	.dc.b	$C7,$AD,$AE,$11,$FB,$7E,$4E,$3F,$57,$23,$04,$3C,$4E,$5F,$DD,$47
	.dc.b	$AB,$EE,$FF,$BC,$C7,$F5,$DF,$B5,$7C,$EF,$D2,$E8,$46,$27,$FD,$83
	.dc.b	$5C,$1F,$F1,$63,$5F,$D5,$B5,$FF,$14,$B5,$FF,$2B,$B5,$3F,$EF,$07
	.dc.b	$C5,$E8,$78,$67,$B6,$DF,$26,$77,$E7,$FA,$4D,$54,$8F,$2C,$65,$A4
	.dc.b	$3F,$BB,$FC,$F3,$4F,$85,$F0,$3E,$FB,$FF,$02,$D7,$F7,$F6,$BE,$E8
	.dc.b	$DB,$9F,$4F,$81,$87,$47,$C6,$8C,$63,$FE,$15,$E8,$FE,$03,$3F,$2B
	.dc.b	$D8,$F9,$DF,$79,$F7,$F7,$3F,$81,$9F,$8B,$3F,$5C,$DF,$6C,$8E,$44
	.dc.b	$3E,$9C,$6D,$FF,$0C,$5B,$ED,$FF,$33,$EF,$7F,$A4,$A7,$FC,$39,$F7
	.dc.b	$9F,$E5,$96,$FB,$1A,$F4,$64,$60,$87,$F9,$AE,$B9,$36,$36,$D3,$FD
	.dc.b	$2B,$AF,$D9,$F7,$9F,$0D,$E2,$F4,$BF,$1D,$D4,$F5,$69,$9C,$38,$F7
	.dc.b	$7F,$9C,$B3,$5E,$A7,$FC,$69,$7C,$5D,$9F,$EA,$49,$DB,$83,$FC,$DD
	.dc.b	$3D,$5F,$53,$F9,$FF,$BC,$FF,$79,$C7,$F8,$FF,$EF,$3F,$28,$FF,$8F
	.dc.b	$9B,$A7,$F9,$3D,$B8,$FF,$A1,$BE,$F1,$E4,$7F,$06,$6D,$F3,$ED,$FD
	.dc.b	$65,$F0,$8F,$F5,$CC,$E9,$FF,$25,$EB,$F9,$57,$D1,$19,$C7,$F4,$4D
	.dc.b	$C7,$FC,$35,$5B,$AF,$F7,$C4,$3E,$B9,$F1,$7F,$CB,$23,$0D,$8C,$CF
	.dc.b	$BF,$E3,$FD,$1E,$3F,$99,$91,$E2,$8F,$E0,$9F,$79,$FD,$A3,$3F,$F3
	.dc.b	$83,$71,$FF,$16,$36,$9F,$0C,$F2,$7F,$25,$1F,$4F,$EE,$DF,$79,$FC
	.dc.b	$B3,$C7,$FF,$5C,$4F,$D7,$6B,$FE,$B9,$9D,$BC,$67,$87,$C6,$48,$C1
	.dc.b	$0F,$ED,$33,$C1,$B3,$FE,$A9,$51,$97,$FD,$7F,$E5,$BE,$F1,$F3,$DF
	.dc.b	$D0,$BE,$47,$D9,$26,$CE,$AB,$F2,$33,$C6,$B8,$44,$33,$3B,$5B,$FF
	.dc.b	$B7,$7C,$3E,$DF,$19,$27,$9E,$3F,$E0,$37,$7C,$3F,$E2,$A8,$C5,$FF
	.dc.b	$B9,$6B,$85,$68,$FE,$9F,$3A,$AF,$F6,$DF,$DD,$BE,$6F,$FA,$DA,$FB
	.dc.b	$CA,$62,$25,$FE,$A6,$74,$D9,$CB,$FE,$ED,$8E,$8E,$48,$66,$7F,$B4
	.dc.b	$FB,$7F,$ED,$F8,$FD,$A7,$DE,$7F,$B7,$DB,$E5,$6B,$C3,$91,$FE,$DF
	.dc.b	$AE,$0F,$F8,$91,$AF,$F8,$E9,$AF,$C5,$F8,$0F,$B1,$F4,$D2,$3E,$5C
	.dc.b	$7D,$B3,$0F,$C7,$E3,$FB,$93,$98,$FE,$11,$4F,$BB,$FF,$00,$FC,$73
	.dc.b	$EF,$29,$EC,$89,$5F,$C3,$E7,$4D,$9C,$BF,$54,$FF,$21,$CF,$E4,$EB
	.dc.b	$B4,$7B,$E3,$07,$FB,$FF,$0F,$C8,$7C,$D7,$E4,$FD,$A9,$5F,$86,$8F
	.dc.b	$CB,$18,$7D,$9C,$7F,$DE,$5B,$1F,$E1,$AF,$BB,$7D,$DF,$F8,$73,$EF
	.dc.b	$1E,$85,$35,$12,$FF,$CA,$62,$B8,$B8,$6B,$FE,$58,$6B,$FE,$6E,$FB
	.dc.b	$BF,$F2,$BB,$FC,$CF,$C2,$7D,$E7,$F8,$B1,$B1,$FC,$DE,$76,$FC,$76
	.dc.b	$F6,$7F,$80,$E7,$73,$F2,$DB,$1F,$E0,$79,$DF,$5A,$A4,$74,$7C,$9E
	.dc.b	$8F,$F0,$38,$C4,$6C,$38,$79,$79,$3F,$D3,$DF,$B5,$F3,$3F,$79,$BC
	.dc.b	$BF,$E4,$69,$D6,$3F,$E1,$97,$84,$FF,$A4,$1A,$F7,$6B,$DB,$CC,$F9
	.dc.b	$FF,$3C,$FB,$C7,$C9,$FF,$13,$EE,$97,$FD,$82,$B5,$3F,$E6,$A6,$C7
	.dc.b	$FA,$29,$AF,$FA,$81,$AF,$BD,$AF,$C1,$FB,$CF,$F3,$63,$F9,$77,$F6
	.dc.b	$39,$7F,$8C,$DB,$BE,$FE,$27,$FC,$CB,$1C,$2B,$FD,$5A,$D7,$47,$E2
	.dc.b	$9C,$3F,$C5,$BE,$4B,$F9,$71,$1F,$E4,$3F,$07,$5F,$CD,$C5,$C3,$FC
	.dc.b	$D9,$D9,$FC,$CE,$BE,$07,$DE,$3E,$47,$E1,$7E,$46,$F1,$29,$A8,$99
	.dc.b	$7C,$5A,$E0,$D8,$FA,$31,$88,$FF,$59,$F8,$1F,$D0,$BE,$1F,$E9,$C7
	.dc.b	$F4,$06,$BF,$BD,$8F,$EC,$4F,$EC,$4E,$27,$F5,$17,$FA,$93,$93,$FD
	.dc.b	$A6,$F0,$23,$FE,$92,$F9,$67,$DE,$FF,$A4,$A7,$33,$FE,$1D,$7F,$AA
	.dc.b	$CE,$A3,$F0,$9F,$6C,$F0,$FF,$D3,$7F,$9C,$BD,$9A,$7E,$1E,$75,$5C
	.dc.b	$2B,$F2,$31,$88,$FC,$43,$5F,$C4,$35,$FD,$47,$CA,$3E,$F3,$FD,$53
	.dc.b	$F7,$BF,$A5,$E4,$FA,$ED,$E8,$DC,$FF,$AB,$2B,$85,$FF,$6A,$AD,$3F
	.dc.b	$EA,$86,$A8,$DB,$FC,$AD,$AF,$FB,$82,$7F,$BF,$46,$23,$B9,$FE,$E2
	.dc.b	$F0,$FF,$B8,$AF,$BC,$FF,$4F,$73,$BE,$F1,$F6,$DF,$EC,$26,$D7,$F9
	.dc.b	$2F,$0F,$F0,$E6,$9F,$C3,$70,$FE,$5D,$18,$9E,$9C,$5F,$C4,$71,$E9
	.dc.b	$8F,$33,$E0,$65,$E2,$C3,$8B,$6B,$E5,$F0,$7F,$13,$9E,$AD,$8F,$9F
	.dc.b	$3B,$C3,$F8,$F6,$3E,$1B,$DC,$79,$6B,$9B,$4D,$D7,$17,$E6,$36,$BF
	.dc.b	$0D,$1B,$3E,$85,$F8,$5E,$39,$66,$7C,$17,$DA,$7E,$09,$E4,$AE,$0F
	.dc.b	$6A,$E1,$7D,$E7,$37,$E0,$AE,$26,$CA,$F6,$7A,$1E,$18,$E3,$3E,$93
	.dc.b	$A4,$64,$EF,$51,$7C,$6F,$F3,$BB,$FB,$0D,$AF,$C2,$B6,$BF,$A5,$3F
	.dc.b	$CE,$3E,$4A,$F8,$7F,$D1,$F0,$FE,$9D,$18,$70,$D7,$FA,$63,$5D,$9E
	.dc.b	$E1,$EC,$B8,$BE,$84,$6B,$5F,$E1,$53,$98,$E0,$2B,$71,$99,$FE,$19
	.dc.b	$18,$AE,$75,$B3,$C1,$5B,$D7,$D5,$19,$BE,$F5,$C2,$F9,$4F,$81,$B3
	.dc.b	$DA,$39,$CE,$DA,$6C,$B1,$B3,$AB,$96,$99,$AF,$99,$7E,$59,$F9,$67
	.dc.b	$E3,$37,$FD,$B7,$77,$C3,$D3,$E2,$4F,$CE,$35,$C2,$9D,$65,$EE,$CF
	.dc.b	$F4,$58,$72,$8F,$F6,$35,$FC,$FC,$BF,$71,$87,$19,$CF,$67,$6A,$FC
	.dc.b	$03,$5F,$F0,$43,$5F,$F0,$83,$5F,$EA,$51,$FE,$86,$9F,$F7,$1B,$D0
	.dc.b	$C7,$C5,$79,$A7,$9C,$60,$D3,$FE,$57,$9E,$2F,$F9,$FF,$C2,$73,$E2
	.dc.b	$3F,$9A,$DF,$96,$72,$D9,$D4,$FF,$89,$67,$37,$FE,$9A,$E9,$FE,$9E
	.dc.b	$FF,$4A,$BF,$26,$69,$CD,$FF,$96,$F6,$FF,$5F,$5F,$FB,$BD,$B1,$FE
	.dc.b	$6D,$EA,$7D,$DF,$B6,$3F,$D4,$3C,$B6,$B6,$7D,$A2,$DF,$F2,$9B,$E2
	.dc.b	$F5,$9E,$07,$CF,$9E,$F5,$FE,$D5,$8C,$12,$7F,$D6,$BF,$95,$F2,$9F
	.dc.b	$44,$FB,$23,$15,$FA,$DD,$1E,$23,$E6,$FE,$53,$71,$FF,$16,$37,$1F
	.dc.b	$F1,$63,$5F,$F1,$63,$9B,$E9,$E1,$BE,$1B,$EF,$63,$DB,$1B,$39,$C7
	.dc.b	$FD,$83,$F1,$5F,$43,$D5,$37,$EE,$F0,$C9,$D9,$FE,$C8,$8C,$69,$FE
	.dc.b	$8F,$C5,$ED,$7E,$55,$B5,$FE,$4E,$8C,$11,$FF,$34,$69,$FA,$C9,$FF
	.dc.b	$05,$46,$0F,$F2,$44,$FD,$B8,$CD,$7E,$4F,$EE,$FF,$3E,$FF,$93,$DA
	.dc.b	$FF,$93,$DA,$FE,$45,$AA,$11,$CA,$FF,$C7,$FF,$77,$FC,$97,$0F,$F3
	.dc.b	$04,$E5,$FE,$DD,$19,$BF,$F5,$77,$C1,$7F,$7A,$9F,$F4,$CB,$63,$F8
	.dc.b	$FB,$5F,$E7,$4D,$7F,$D0,$4E,$EF,$F9,$87,$73,$EE,$FF,$82,$57,$6F
	.dc.b	$17,$F7,$13,$9F,$03,$F3,$DF,$76,$F8,$6F,$E0,$8D,$CF,$FD,$56,$DC
	.dc.b	$3F,$98,$B5,$FE,$98,$E7,$2F,$E8,$1F,$76,$FB,$A7,$DD,$F6,$C7,$F8
	.dc.b	$44,$E6,$31,$8F,$D7,$C5,$39,$8C,$43,$1E,$D1,$FE,$A7,$9D,$4F,$FB
	.dc.b	$5F,$E3,$1F,$0F,$FC,$75,$B5,$FE,$C4,$6C,$FF,$4A,$6B,$FD,$95,$CE
	.dc.b	$7F,$2B,$ED,$9F,$5B,$FC,$0E,$73,$7F,$EB,$7D,$D5,$A8,$F5,$78,$6E
	.dc.b	$AF,$F6,$BC,$61,$FF,$40,$37,$E0,$FF,$51,$EB,$7E,$A5,$B1,$93,$84
	.dc.b	$7F,$13,$FB,$9F,$F5,$6C,$7F,$B7,$9B,$AF,$F6,$46,$BF,$DF,$DA,$FF
	.dc.b	$B1,$DD,$EB,$FC,$27,$EE,$7F,$84,$D7,$CE,$9C,$BF,$CB,$E3,$34,$AF
	.dc.b	$EB,$35,$A8,$FD,$14,$EA,$73,$ED,$FF,$08,$8C,$5F,$F9,$37,$2F,$E1
	.dc.b	$17,$FF,$00,$3E,$E7,$F8,$47,$87,$FC,$25,$B5,$F1,$DB,$3B,$8F,$C1
	.dc.b	$FD,$B7,$EE,$A3,$FC,$38,$D7,$FD,$C8,$D7,$FD,$D8,$D4,$FF,$02,$CB
	.dc.b	$EF,$7C,$E7,$F3,$F8,$C5,$7D,$DE,$04,$FF,$5F,$FB,$97,$D1,$7E,$E7
	.dc.b	$0F,$F0,$F4,$63,$E6,$75,$39,$C3,$FD,$F9,$39,$7F,$3C,$F2,$77,$3C
	.dc.b	$1A,$6A,$79,$64,$AE,$4F,$D3,$9E,$B3,$A9,$CD,$FF,$B4,$35,$7D,$67
	.dc.b	$F3,$5F,$43,$F9,$B3,$70,$FC,$A3,$71,$E8,$39,$7F,$35,$19,$C3,$A7
	.dc.b	$DC,$7F,$3F,$FA,$5F,$E8,$0D,$C3,$F3,$CD,$C3,$9E,$1F,$A3,$8C,$5F
	.dc.b	$F8,$8B,$81,$1F,$E1,$36,$DF,$F8,$FA,$FE,$BC,$E4,$91,$FC,$1E,$9F
	.dc.b	$62,$FA,$3F,$B5,$5F,$E1,$59,$CC,$62,$B8,$56,$DF,$C5,$D6,$E3,$17
	.dc.b	$F0,$87,$C3,$FF,$52,$70,$DC,$3F,$CB,$D6,$F9,$5F,$E7,$D4,$38,$F9
	.dc.b	$9F,$E0,$BF,$B7,$FF,$34,$D7,$F1,$FE,$37,$C6,$F3,$8E,$16,$AD,$BF
	.dc.b	$52,$B7,$7F,$E3,$1F,$9E,$7D,$C7,$F0,$EA,$FB,$31,$8F,$03,$FC,$6D
	.dc.b	$7C,$31,$FE,$5D,$39,$7F,$96,$EA,$F8,$BF,$37,$9D,$FE,$21,$8C,$4F
	.dc.b	$F3,$2D,$13,$A8,$F5,$6B,$AC,$FE,$45,$6A,$73,$87,$8A,$B5,$BF,$FA
	.dc.b	$17,$3F,$83,$3C,$F0,$F7,$56,$AF,$E1,$E1,$67,$E6,$35,$7D,$C3,$EA
	.dc.b	$BE,$76,$12,$FE,$77,$9D,$F7,$33,$FF,$38,$E3,$CF,$3F,$F3,$87,$2E
	.dc.b	$F3,$F9,$D7,$BC,$3C,$38,$FC,$A8,$C4,$F8,$BD,$5F,$B8,$7F,$91,$F4
	.dc.b	$7D,$3E,$30,$F9,$13,$F4,$BE,$E1,$F0,$D9,$C3,$EE,$46,$2B,$67,$E1
	.dc.b	$5C,$47,$CC,$AE,$55,$FE,$55,$FB,$77,$CE,$D3,$EF,$D3,$EF,$71,$8F
	.dc.b	$F6,$AA,$DB,$F9,$1F,$B9,$F5,$D2,$72,$FE,$3E,$33,$18,$E2,$FB,$11
	.dc.b	$8C,$3F,$1A,$EB,$D0,$F2,$1F,$90,$E2,$FD,$26,$3E,$18,$CB,$FA,$1F
	.dc.b	$81,$F6,$E1,$FE,$A4,$C7,$FB,$56,$5F,$9F,$9C,$D7,$E9,$27,$35,$F8
	.dc.b	$99,$CD,$7E,$BE,$73,$5F,$B1,$C1,$5A,$AF,$77,$91,$F2,$4F,$D2,$E0
	.dc.b	$FB,$9C,$61,$F1,$E7,$F5,$3E,$07,$F9,$72,$30,$7E,$E5,$C5,$FF,$07
	.dc.b	$CB,$FA,$3C,$E7,$83,$9C,$FE,$E6,$B7,$8F,$BF,$8B,$FD,$65,$87,$E2
	.dc.b	$30,$FE,$40,$DB,$FF,$6F,$B5,$FF,$7D,$39,$AF,$D5,$5F,$F0,$17,$A5
	.dc.b	$79,$7B,$7F,$D4,$7F,$07,$FD,$DF,$07,$DC,$BE,$E7,$3E,$8F,$FA,$7E
	.dc.b	$31,$87,$F8,$6A,$F8,$46,$2F,$FD,$4D,$5B,$BF,$F5,$37,$C5,$7F,$93
	.dc.b	$B2,$FF,$6E,$B7,$1F,$F5,$24,$EB,$AA,$47,$F8,$F2,$75,$87,$F4,$FC
	.dc.b	$BB,$61,$CA,$FA,$FA,$93,$0F,$27,$DC,$DC,$3F,$9B,$63,$FE,$2C,$7F
	.dc.b	$94,$9F,$ED,$CA,$FF,$6F,$46,$63,$FD,$C7,$18,$3E,$2F,$CB,$73,$C5
	.dc.b	$87,$FA,$EE,$3D,$BE,$6C,$8F,$89,$F7,$37,$0F,$0C,$F6,$24,$FF,$59
	.dc.b	$C7,$FB,$8B,$FD,$6F,$ED,$DF,$4A,$F6,$B0,$DC,$3D,$22,$77,$3A,$3F
	.dc.b	$B2,$7C,$A3,$E8,$3F,$C2,$75,$C2,$31,$DB,$A9,$CC,$FC,$49,$CB,$DA
	.dc.b	$FF,$D9,$1F,$2B,$E5,$F8,$C3,$49,$3F,$33,$D4,$72,$BE,$97,$FA,$BF
	.dc.b	$CA,$FF,$61,$AF,$B5,$18,$F5,$3F,$DC,$8D,$8F,$BD,$F4,$17,$0E,$73
	.dc.b	$70,$E9,$5F,$DA,$70,$E9,$5A,$8C,$61,$FE,$EF,$C3,$87,$C3,$7F,$68
	.dc.b	$E7,$FB,$35,$F8,$BF,$87,$C7,$27,$FB,$D5,$B1,$F8,$FA,$B8,$FB,$1B
	.dc.b	$9F,$FA,$8E,$7F,$45,$58,$99,$18,$21,$FE,$1E,$9F,$F4,$FB,$FE,$82
	.dc.b	$F4,$7F,$27,$8F,$93,$0F,$EA,$DC,$9E,$09,$FC,$05,$7F,$89,$9D,$E5
	.dc.b	$FE,$09,$AF,$EB,$93,$BA,$8F,$53,$D3,$F1,$6F,$7F,$A7,$AB,$27,$FD
	.dc.b	$27,$39,$6C,$E5,$F0,$4F,$28,$E7,$D1,$D3,$92,$57,$28,$CB,$D2,$32
	.dc.b	$FF,$8F,$AF,$F3,$AF,$83,$E1,$BC,$BF,$51,$23,$04,$38,$39,$79,$47
	.dc.b	$B0,$C6,$DF,$EE,$C9,$C4,$FB,$3E,$8F,$E0,$FA,$12,$30,$43,$C8,$E5
	.dc.b	$F3,$47,$C8,$31,$5F,$EF,$89,$C4,$FC,$8F,$3F,$E6,$FD,$5C,$8C,$10
	.dc.b	$F8,$0E,$5F,$85,$1F,$50,$C5,$7E,$07,$25,$66,$BE,$AF,$D9,$FD,$AF
	.dc.b	$B9,$91,$82,$1F,$35,$CB,$F3,$E3,$F0,$26,$2F,$B4,$E2,$7F,$03,$F7
	.dc.b	$FF,$86,$FC,$0E,$46,$08,$7D,$97,$2F,$DB,$8F,$C6,$98,$AF,$54,$E2
	.dc.b	$7F,$1B,$F7,$5F,$92,$F5,$E4,$60,$FC,$8F,$47,$E5,$A6,$46,$0F,$CA
	.dc.b	$FB,$9F,$9A,$99,$18,$3F,$33,$EA,$FE,$7B,$D4,$91,$83,$FD,$DF,$E7
	.dc.b	$FF,$18,$6F,$B3,$FB,$C8,$FF,$17,$C6,$BC,$5C,$7C,$76,$7C,$49,$3C
	.dc.b	$A3,$F8,$2A,$FB,$17,$CC,$E1,$7D,$15,$31,$12,$7F,$D0,$1C,$5F,$C4
	.dc.b	$9E,$53,$B7,$3E,$1F,$F9,$FA,$FA,$C6,$0F,$F4,$43,$87,$FD,$24,$F8
	.dc.b	$1D,$1D,$0A,$EE,$BC,$1F,$63,$F8,$E9,$FF,$55,$35,$FE,$9A,$D7,$FB
	.dc.b	$3B,$5F,$F4,$8B,$B8,$E7,$D1,$F1,$23,$17,$E9,$D3,$E7,$65,$F3,$F2
	.dc.b	$98,$FD,$48,$CB,$F5,$A3,$2F,$FB,$13,$5F,$45,$FF,$1F,$5F,$6B,$F6
	.dc.b	$49,$3F,$7B,$E9,$2B,$C6,$2E,$9F,$CB,$69,$58,$BE,$95,$6F,$FD,$B9
	.dc.b	$5A,$7E,$87,$CD,$BF,$0D,$FE,$5F,$1E,$5F,$81,$FC,$51,$FD,$22,$75
	.dc.b	$CD,$6B,$79,$FF,$50,$9C,$DF,$49,$CE,$5E,$DF,$4B,$F2,$9F,$51,$2B
	.dc.b	$F3,$2F,$E5,$FE,$87,$11,$F1,$62,$E3,$ED,$9F,$CA,$FD,$8B,$D7,$BF
	.dc.b	$1D,$30,$FC,$2C,$E5,$B3,$97,$FD,$B9,$09,$E7,$32,$78,$C6,$EF,$F3
	.dc.b	$55,$CE,$7E,$C7,$BF,$FA,$DE,$09,$1F,$B2,$8F,$D8,$18,$7E,$F4,$61
	.dc.b	$FC,$CF,$EC,$CF,$6F,$F5,$FF,$B5,$7F,$48,$9F,$F7,$83,$71,$FD,$55
	.dc.b	$6E,$78,$D7,$ED,$A3,$11,$FE,$F7,$6C,$FE,$C6,$BF,$B0,$4F,$FB,$EF
	.dc.b	$E1,$3F,$83,$37,$BF,$F7,$F1,$FC,$1A,$35,$D5,$83,$33,$23,$F6,$33
	.dc.b	$CA,$FF,$7C,$DE,$2F,$F7,$CB,$CE,$33,$5C,$EB,$F7,$3F,$94,$FF,$AE
	.dc.b	$1F,$DD,$FC,$1B,$E3,$FF,$88,$B6,$37,$E7,$7F,$14,$AE,$10,$C3,$F8
	.dc.b	$74,$61,$B3,$CB,$E4,$BF,$31,$F8,$7F,$F8,$D6,$FF,$61,$F0,$DF,$C9
	.dc.b	$4C,$F8,$D2,$3F,$47,$3B,$C3,$F3,$8D,$DB,$F6,$DF,$09,$FC,$B7,$E1
	.dc.b	$7F,$97,$7C,$14,$C3,$F9,$9C,$65,$FE,$6B,$19,$7F,$80,$47,$F3,$5F
	.dc.b	$81,$FE,$6D,$F0,$12,$7F,$9D,$C6,$5F,$E7,$B1,$97,$F8,$44,$7F,$8A
	.dc.b	$7B,$3F,$C5,$5D,$92,$7F,$C6,$11,$97,$FC,$65,$19,$7F,$C5,$D1,$FE
	.dc.b	$31,$FA,$DF,$F1,$97,$D5,$C9,$FF,$1C,$46,$5F,$F1,$D4,$65,$FF,$1B
	.dc.b	$46,$5F,$F1,$C4,$67,$6F,$F1,$F7,$88,$AD,$46,$27,$5B,$7F,$8F,$6B
	.dc.b	$57,$FE,$47,$FA,$3F,$F2,$27,$34,$8C,$10,$FF,$23,$39,$7F,$C9,$11
	.dc.b	$FE,$4B,$F9,$2F,$F2,$67,$C7,$49,$FF,$29,$46,$5F,$F2,$A4,$65,$FF
	.dc.b	$28,$46,$5F,$F2,$94,$67,$2F,$5C,$FF,$95,$FE,$9B,$FC,$B1,$F2,$12
	.dc.b	$FF,$CB,$D1,$97,$FC,$C1,$19,$7F,$CB,$91,$97,$FC,$BD,$19,$AF,$8F
	.dc.b	$39,$7E,$4C,$67,$3F,$8D,$3F,$DD,$3E,$2F,$FD,$F7,$4B,$7D,$63,$6C
	.dc.b	$42,$BF,$9F,$3A,$8C,$C6,$27,$84,$F3,$8E,$AE,$CC,$79,$9F,$AF,$70
	.dc.b	$75,$F3,$BB,$D7,$28,$C1,$E5,$71,$5D,$1B,$5C,$DC,$D7,$2B,$F2,$B6
	.dc.b	$FC,$38,$73,$9F,$4D,$F1,$77,$39,$3D,$55,$C5,$F8,$53,$B7,$99,$EA
	.dc.b	$C3,$E1,$57,$28,$C4,$ED,$E3,$7F,$3F,$C2,$FE,$4D,$7F,$19,$B5,$F1
	.dc.b	$5B,$3B,$F0,$7F,$DB,$7F,$17,$70,$FF,$7A,$B5,$FF,$64,$9F,$ED,$08
	.dc.b	$DB,$4D,$9F,$40,$FF,$74,$6B,$EF,$78,$E9,$F4,$5C,$9B,$15,$B7,$31
	.dc.b	$B7,$27,$5F,$88,$FF,$64,$C6,$67,$84,$63,$4F,$2C,$F7,$8C,$57,$9A
	.dc.b	$31,$7D,$E3,$AC,$61,$F2,$5F,$49,$D3,$63,$F2,$1F,$12,$E5,$7C,$87
	.dc.b	$FB,$73,$4F,$E1,$BF,$1F,$71,$FE,$15,$37,$06,$1C,$6B,$0F,$39,$B8
	.dc.b	$26,$C6,$1B,$7F,$ED,$96,$DF,$FB,$55,$B8,$78,$A7,$FD,$BD,$5C,$5F
	.dc.b	$F7,$07,$D0,$76,$FA,$0E,$ED,$8C,$9F,$EE,$58,$CB,$4C,$E1,$FE,$E3
	.dc.b	$8C,$B6,$3F,$DD,$9E,$67,$AB,$E7,$7D,$6D,$8C,$9F,$EF,$18,$CB,$4F
	.dc.b	$F7,$A7,$8F,$F0,$BE,$77,$E1,$B6,$32,$7F,$BE,$63,$2D,$33,$7F,$EF
	.dc.b	$78,$CB,$63,$39,$FF,$42,$9C,$B6,$3F,$CF,$DB,$7D,$DF,$2F,$EA,$FE
	.dc.b	$CA,$9F,$AB,$18,$69,$C2,$34,$F2,$E0,$FF,$04,$39,$3E,$FF,$A0,$F9
	.dc.b	$2F,$C0,$FE,$E7,$FC,$9E,$7F,$D4,$F1,$8D,$36,$D8,$C9,$E6,$9E,$13
	.dc.b	$FE,$0F,$8F,$E6,$19,$27,$F5,$F9,$FA,$6B,$F2,$30,$F9,$2F,$0F,$3F
	.dc.b	$89,$B8,$75,$6D,$75,$38,$BF,$9B,$F6,$9E,$1F,$F6,$BA,$FF,$3B,$B6
	.dc.b	$B8,$65,$E3,$F1,$B8,$4E,$73,$FF,$3A,$36,$FD,$39,$FF,$98,$5E,$7E
	.dc.b	$37,$C6,$F1,$FF,$55,$97,$A7,$8A,$EB,$F3,$A3,$CD,$B3,$EA,$1E,$26
	.dc.b	$AE,$65,$FE,$BB,$AD,$78,$B8,$C3,$8E,$08,$FD,$C5,$3F,$1D,$FF,$66
	.dc.b	$BF,$EC,$33,$1E,$EE,$A3,$37,$F1,$E7,$26,$E3,$3C,$1F,$4E,$33,$3B
	.dc.b	$9C,$C7,$E0,$DF,$F5,$7B,$C7,$DE,$ED,$7D,$DB,$1F,$ED,$0F,$83,$7C
	.dc.b	$5F,$C6,$6C,$7B,$FE,$91,$F0,$99,$A9,$39,$F8,$DB,$9F,$CA,$B4,$7E
	.dc.b	$5B,$93,$F2,$FE,$45,$BF,$C9,$36,$FE,$FF,$37,$F7,$F9,$FC,$B5,$70
	.dc.b	$7F,$DD,$5F,$49,$FE,$05,$F1,$FF,$92,$3F,$E0,$7F,$1A,$E1,$FC,$86
	.dc.b	$B6,$A3,$35,$A6,$CE,$9B,$3F,$97,$AE,$93,$CA,$B8,$3F,$A9,$EA,$FD
	.dc.b	$89,$FB,$2F,$46,$C6,$49,$7E,$C8,$CB,$FA,$3E,$EF,$D0,$F8,$FF,$9C
	.dc.b	$3F,$77,$E8,$5B,$FD,$1C,$7E,$FB,$D1,$C5,$FC,$67,$F8,$0F,$6C,$AF
	.dc.b	$DE,$75,$5A,$F9,$73,$7A,$BF,$75,$E0,$FE,$F6,$B1,$7F,$A8,$AB,$7F
	.dc.b	$BE,$9B,$5F,$C0,$FC,$AF,$E1,$FB,$FF,$10,$6E,$1F,$C2,$DB,$7F,$69
	.dc.b	$B8,$FF,$09,$87,$D1,$BC,$9F,$D9,$B6,$FF,$DE,$FF,$82,$3B,$33,$52
	.dc.b	$7F,$C8,$37,$FE,$42,$BE,$0E,$0E,$B3,$DD,$D3,$57,$23,$E6,$DF,$2B
	.dc.b	$F8,$A5,$F2,$B6,$30,$DC,$7F,$23,$3A,$6C,$E9,$A6,$A7,$85,$FE,$2E
	.dc.b	$B8,$3F,$9E,$E8,$FE,$73,$C9,$FC,$E9,$B7,$FC,$D7,$CF,$FE,$7B,$E7
	.dc.b	$FF,$3E,$6C,$7F,$39,$D3,$FA,$0F,$9F,$FD,$09,$B1,$FC,$F7,$C8,$FF
	.dc.b	$02,$FE,$E9,$FD,$33,$3F,$D6,$C6,$26,$1F,$D6,$25,$F3,$58,$AF,$E6
	.dc.b	$31,$89,$D4,$F4,$FB,$9E,$AD,$8E,$5F,$37,$E0,$F9,$BF,$0B,$63,$27
	.dc.b	$78,$CB,$4C,$E1,$E0,$8C,$B6,$3C,$DF,$37,$EC,$E3,$F6,$B6,$32,$7A
	.dc.b	$E3,$2D,$33,$3E,$C8,$CB,$63,$E1,$71,$FC,$8F,$85,$F9,$2D,$8C,$9F
	.dc.b	$1E,$32,$D3,$33,$F2,$23,$2D,$8F,$9D,$F3,$7F,$53,$CB,$FA,$AD,$8C
	.dc.b	$9F,$4E,$32,$D3,$33,$F5,$23,$2D,$8F,$B5,$E4,$FE,$03,$CB,$FC,$0B
	.dc.b	$63,$27,$DF,$8C,$B4,$CC,$FE,$02,$32,$D8,$FC,$37,$CD,$7E,$33,$E6
	.dc.b	$BF,$1A,$D8,$C9,$F8,$B8,$CB,$4F,$C7,$7C,$CF,$F8,$93,$87,$F2,$AD
	.dc.b	$8C,$9F,$E2,$38,$CB,$4F,$F1,$37,$17,$F8,$B3,$8B,$FC,$5A,$D8,$C9
	.dc.b	$FE,$2B,$8C,$B4,$FF,$17,$71,$7F,$8D,$38,$BF,$C6,$AD,$8C,$9F,$E3
	.dc.b	$38,$CB,$4C,$E7,$FE,$14,$F6,$93,$A8,$C3,$FE,$40,$E3,$FF,$14,$7B
	.dc.b	$6F,$88,$FF,$16,$DF,$F1,$0C,$3F,$DB,$D3,$9C,$3F,$CF,$B3,$FC,$1A
	.dc.b	$31,$E3,$23,$84,$FF,$A8,$2F,$9E,$1D,$9B,$87,$47,$BF,$36,$6B,$AD
	.dc.b	$55,$F3,$AD,$5E,$BF,$F2,$B9,$CC,$61,$FE,$5D,$3B,$9D,$36,$3F,$99
	.dc.b	$7C,$13,$E4,$FF,$97,$46,$1A,$FF,$3A,$8D,$78,$7D,$DB,$1D,$B8,$5F
	.dc.b	$05,$D9,$B1,$98,$C1,$FD,$07,$81,$F3,$AF,$89,$57,$C3,$3F,$D5,$31
	.dc.b	$97,$13,$C2,$77,$1F,$34,$DF,$C3,$E3,$93,$E9,$DF,$19,$D5,$F1,$79
	.dc.b	$C6,$67,$75,$C7,$8B,$FC,$DB,$2F,$88,$EF,$E6,$FF,$74,$8F,$F3,$E6
	.dc.b	$1F,$E8,$A6,$BF,$E6,$27,$83,$99,$E3,$CB,$D3,$E0,$AC,$FB,$9B,$9F
	.dc.b	$C4,$AB,$7C,$AB,$A5,$60,$CC,$F0,$F8,$4F,$ED,$F8,$FF,$82,$1F,$C1
	.dc.b	$57,$F9,$D1,$B1,$98,$E0,$71,$E4,$8F,$FA,$6C,$F1,$FC,$EF,$F4,$E7
	.dc.b	$FD,$3D,$F0,$5D,$CC,$F6,$C8,$C9,$0F,$F4,$1C,$F1,$6F,$47,$4E,$49
	.dc.b	$1F,$EB,$29,$FC,$E3,$67,$2F,$29,$F1,$C7,$93,$B3,$CB,$D5,$2B,$94
	.dc.b	$65,$E9,$19,$7C,$71,$97,$CB,$19,$CF,$FC,$5F,$87,$F0,$69,$F2,$75
	.dc.b	$FC,$2F,$16,$46,$08,$7B,$9C,$BF,$02,$3E,$49,$8A,$EB,$38,$9F,$81
	.dc.b	$E2,$FC,$EF,$16,$46,$08,$7C,$A7,$2F,$CC,$8F,$AA,$62,$BD,$33,$89
	.dc.b	$F9,$9E,$2F,$DA,$F1,$64,$60,$87,$D6,$72,$FD,$88,$FC,$09,$8A,$F8
	.dc.b	$B3,$89,$FB,$1E,$2F,$E1,$BC,$59,$18,$21,$F8,$27,$2F,$E1,$23,$F1
	.dc.b	$A6,$2B,$E8,$CE,$27,$F0,$9E,$2F,$E4,$BE,$76,$46,$0F,$C4,$73,$7E
	.dc.b	$5A,$64,$60,$FC,$67,$CE,$FE,$6A,$64,$60,$FC,$87,$CE,$FE,$7B,$E7
	.dc.b	$64,$60,$FE,$27,$97,$F9,$CB,$E2,$7F,$8D,$C7,$FB,$79,$7C,$5F,$EE
	.dc.b	$1B,$1C,$3C,$0F,$F7,$AC,$70,$AE,$5E,$B6,$FC,$29,$7B,$1E,$12,$FF
	.dc.b	$CD,$91,$91,$F3,$AF,$9C,$FE,$13,$18,$F8,$7B,$26,$CC,$9E,$53,$96
	.dc.b	$CE,$5F,$F8,$00,$76,$9E,$71,$23,$8C,$78,$3E,$77,$FD,$9B,$1D,$32
	.dc.b	$F0,$99,$E8,$91,$FD,$56,$76,$FF,$A0,$5A,$FF,$A1,$9A,$FC,$8A,$F9
	.dc.b	$97,$23,$E3,$65,$F1,$67,$E8,$19,$8F,$A9,$38,$77,$93,$E2,$7E,$B1
	.dc.b	$23,$3F,$21,$5F,$B9,$17,$6F,$C0,$4D,$DF,$ED,$CD,$AF,$AF,$6A,$FE
	.dc.b	$77,$F0,$8F,$9D,$72,$FE,$17,$E4,$66,$1F,$7E,$B5,$3C,$2B,$73,$96
	.dc.b	$BF,$ED,$16,$E7,$F9,$3F,$4B,$F2,$1F,$3B,$2B,$DE,$FE,$4F,$E7,$71
	.dc.b	$1F,$97,$8B,$5F,$9B,$9F,$C4,$FC,$EB,$E4,$EF,$C9,$CC,$3F,$21,$39
	.dc.b	$6C,$E5,$FC,$3C,$27,$9C,$C9,$E3,$1B,$BF,$F6,$65,$73,$9F,$D5,$7C
	.dc.b	$EF,$EA,$78,$24,$7D,$88,$FD,$61,$87,$F6,$11,$87,$F4,$9F,$3A,$F8
	.dc.b	$1F,$D1,$FC,$EF,$F4,$A9,$FC,$DB,$71,$FF,$6E,$56,$E7,$8D,$7E,$AA
	.dc.b	$31,$1F,$A2,$6C,$FF,$B8,$EB,$FA,$D3,$FE,$E5,$F9,$DF,$DE,$73,$7F
	.dc.b	$84,$9F,$E0,$37,$FA,$A6,$CE,$FD,$AF,$DB,$C7,$09,$FF,$0F,$BA,$AE
	.dc.b	$5E,$06,$AA,$57,$F0,$1A,$D3,$C2,$72,$FE,$E3,$E7,$5E,$67,$F0,$4F
	.dc.b	$91,$7C,$1D,$3F,$83,$CE,$AB,$F8,$63,$5E,$55,$A3,$97,$C2,$7E,$73
	.dc.b	$E7,$3F,$CC,$13,$FC,$87,$E7,$7F,$90,$19,$F1,$E4,$7E,$76,$77,$BB
	.dc.b	$7F,$F1,$EF,$6A,$73,$1F,$C9,$BE,$77,$F9,$57,$CE,$FF,$2B,$F8,$29
	.dc.b	$8F,$EE,$A3,$2F,$EF,$23,$2F,$F3,$38,$FE,$65,$F0,$3F,$CC,$FD,$F9
	.dc.b	$3F,$C0,$A3,$2F,$F0,$48,$CB,$FC,$EE,$3F,$C5,$5D,$BF,$E2,$BE,$D9
	.dc.b	$3F,$C4,$3C,$2F,$CC,$6D,$FE,$2D,$AC,$CF,$F8,$B6,$B8,$BB,$F9,$7F
	.dc.b	$F1,$A7,$CC,$C8,$C1,$0F,$F1,$A3,$97,$FC,$6D,$1F,$E3,$B3,$17,$FE
	.dc.b	$3B,$9C,$4F,$18,$C3,$E5,$8C,$BF,$E3,$FF,$98,$FF,$22,$74,$48,$F5
	.dc.b	$47,$F9,$1C,$C3,$FE,$45,$8C,$3E,$FF,$98,$FF,$26,$7C,$C4,$8C,$10
	.dc.b	$FF,$26,$39,$7F,$C9,$D1,$FE,$53,$31,$87,$F9,$4E,$71,$3F,$2F,$E6
	.dc.b	$3F,$CB,$1F,$31,$23,$04,$3F,$CB,$0E,$5F,$F2,$D4,$7F,$97,$CC,$57
	.dc.b	$F9,$7E,$71,$3C,$63,$0F,$D9,$8C,$BF,$D7,$BE,$29,$E5,$AB,$FD,$23
	.dc.b	$F3,$AD,$7F,$D1,$B5,$FC,$F6,$F3,$18,$F3,$7F,$D2,$35,$B9,$FF,$4C
	.dc.b	$D7,$4B,$DC,$63,$0E,$CA,$B8,$2A,$EC,$E7,$0E,$CF,$23,$55,$CA,$B8
	.dc.b	$D7,$79,$E0,$78,$B0,$F2,$DF,$28,$C9,$CB,$0E,$95,$C6,$F9,$DF,$0A
	.dc.b	$E4,$6B,$0F,$06,$1F,$11,$B1,$A3,$AD,$FA,$AF,$CB,$3A,$9F,$01,$FE
	.dc.b	$C6,$F0,$79,$63,$FE,$C0,$F0,$FF,$94,$E1,$EB,$6D,$FC,$87,$F9,$6D
	.dc.b	$FF,$BC,$8E,$47,$FB,$08,$7C,$5F,$3C,$31,$18,$36,$7F,$B1,$4A,$73
	.dc.b	$C7,$E9,$BF,$EE,$CE,$BF,$A0,$E7,$3E,$DF,$13,$FC,$6E,$3E,$59,$93
	.dc.b	$FD,$A6,$E6,$B9,$C7,$85,$EB,$19,$9F,$2E,$1D,$A3,$15,$EB,$8C,$5F
	.dc.b	$8A,$39,$57,$8A,$31,$3A,$9F,$19,$A3,$C7,$1A,$7C,$F5,$DB,$D3,$FF
	.dc.b	$0A,$FC,$C7,$AF,$6F,$F6,$47,$CC,$DC,$3F,$DE,$75,$A7,$F9,$14,$67
	.dc.b	$0F,$F2,$BF,$53,$0C,$5F,$F8,$82,$AC,$E6,$F8,$9F,$CA,$EB,$33,$A6
	.dc.b	$C6,$4D,$DF,$FB,$66,$32,$FF,$B6,$FE,$57,$AF,$CB,$76,$6D,$7F,$37
	.dc.b	$6C,$EE,$3F,$9D,$4E,$DA,$6C,$B1,$FE,$E7,$F3,$7D,$3F,$2D,$EA,$6C
	.dc.b	$FF,$43,$6C,$EE,$3F,$A3,$4F,$C0,$F2,$3E,$0F,$CB,$59,$FE,$94,$DA
	.dc.b	$FF,$7A,$37,$6F,$F7,$C3,$67,$E8,$EA,$8C,$57,$D0,$6E,$DF,$E4,$E6
	.dc.b	$CE,$EB,$FA,$C7,$99,$F5,$3E,$53,$FB,$94,$60,$FF,$35,$B8,$C3,$F9
	.dc.b	$14,$78,$27,$83,$F6,$9A,$FF,$9B,$3E,$47,$EF,$FC,$D2,$BF,$03,$E9
	.dc.b	$A7,$31,$86,$D7,$13,$FC,$F7,$5C,$5B,$C9,$D3,$E4,$BA,$B6,$32,$70
	.dc.b	$8C,$B4,$FF,$36,$78,$6F,$13,$F2,$DE,$6A,$31,$3E,$0F,$1F,$31,$8C
	.dc.b	$FE,$08,$AE,$17,$EE,$F1,$F6,$DC,$3F,$D6,$98,$FF,$8B,$9B,$B7,$E9
	.dc.b	$73,$7C,$EB,$C5,$FD,$46,$6C,$38,$3F,$EB,$06,$F2,$FF,$0D,$BE,$3E
	.dc.b	$2F,$09,$CE,$5F,$3E,$72,$FF,$B2,$47,$CE,$FF,$B0,$5C,$3F,$EA,$16
	.dc.b	$FB,$DF,$B9,$77,$EF,$59,$E1,$97,$C2,$9C,$9C,$0E,$73,$BA,$E0,$D8
	.dc.b	$C9,$FE,$D7,$F8,$27,$C5,$FC,$06,$C7,$FB,$7F,$17,$CE,$BE,$13,$57
	.dc.b	$2B,$52,$F8,$DC,$67,$F5,$B0,$E3,$1F,$92,$E3,$FC,$9F,$8F,$70,$FC
	.dc.b	$6B,$6F,$EE,$F7,$BA,$BF,$BF,$57,$E4,$AF,$8B,$FE,$F3,$F9,$BF,$CE
	.dc.b	$78,$FF,$C7,$DF,$CF,$78,$D7,$14,$D8,$C3,$6B,$94,$E9,$B3,$A6,$9A
	.dc.b	$9E,$13,$E6,$C3,$F3,$13,$FA,$6E,$CF,$D7,$FA,$3F,$B0,$64,$64,$FD
	.dc.b	$5C,$62,$BF,$37,$DB,$FA,$1F,$1F,$F3,$E7,$EE,$3D,$1B,$7F,$A8,$8F
	.dc.b	$DD,$FA,$38,$BF,$13,$FB,$EE,$D9,$5F,$B8,$EC,$B5,$E7,$9B,$DB,$FD
	.dc.b	$97,$8B,$E7,$56,$2F,$FA,$5E,$DF,$A3,$7F,$80,$DF,$F0,$A9,$D6,$0B
	.dc.b	$FA,$DE,$57,$F0,$EE,$0F,$E1,$ED,$C3,$F8,$4B,$6B,$E4,$37,$2F,$EE
	.dc.b	$D0,$F9,$C7,$93,$88,$FE,$30,$65,$C3,$FC,$6E,$37,$7F,$E0,$A8,$D3
	.dc.b	$63,$31,$FC,$22,$9E,$EB,$E2,$FF,$7F,$5A,$9F,$ED,$D3,$9B,$FC,$77
	.dc.b	$CC,$BE,$53,$F1,$CB,$E4,$EC,$61,$B3,$C1,$B9,$7F,$18,$E9,$5F,$1A
	.dc.b	$EB,$7C,$5A,$71,$3F,$98,$F4,$FF,$3A,$E5,$FE,$76,$D8,$FE,$6B,$E8
	.dc.b	$7F,$3E,$F4,$3F,$9F,$B6,$3F,$9C,$F0,$7F,$42,$F4,$3F,$A1,$B6,$3F
	.dc.b	$9E,$F9,$5F,$E0,$4F,$9C,$FE,$9D,$C3,$F8,$5E,$9F,$E3,$7C,$1F,$7E
	.dc.b	$73,$7F,$E1,$06,$C6,$7E,$67,$94,$F3,$63,$FD,$46,$FF,$21,$4F,$9C
	.dc.b	$CD,$FF,$00,$8C,$56,$CE,$DF,$33,$E5,$F9,$AF,$33,$63,$55,$E5,$8C
	.dc.b	$3C,$DB,$87,$81,$B3,$D0,$E3,$D6,$C1,$87,$C3,$F9,$AF,$89,$C9,$6F
	.dc.b	$75,$F8,$36,$CE,$9A,$6A,$78,$65,$ED,$6E,$5F,$09,$FA,$1C,$FF,$47
	.dc.b	$E1,$AD,$F6,$6D,$FC,$56,$DF,$CA,$6C,$FD,$36,$CF,$D7,$7E,$E7,$CC
	.dc.b	$FD,$DF,$32,$D7,$A9,$B5,$F4,$5B,$7F,$59,$B3,$F7,$DB,$3F,$83,$7F
	.dc.b	$11,$E4,$FE,$27,$CC,$B5,$F1,$9B,$5F,$75,$B7,$F8,$26,$CF,$E2,$DB
	.dc.b	$3F,$8F,$7F,$29,$F3,$3F,$95,$F9,$9B,$5F,$49,$B5,$F8,$96,$DF,$E3
	.dc.b	$9F,$CE,$7C,$C7,$F8,$B7,$96,$DF,$F8,$B1,$B7,$FD,$43,$DA,$46,$30
	.dc.b	$FE,$4D,$18,$AF,$F1,$9F,$37,$F8,$DF,$9B,$FC,$70,$D8,$C9,$FE,$38
	.dc.b	$8C,$B4,$FF,$1D,$F2,$FF,$8F,$F9,$3F,$C8,$0D,$8C,$9F,$E4,$08,$CB
	.dc.b	$4C,$E5,$E7,$8C,$B6,$3F,$C8,$FC,$9F,$EA,$1F,$71,$F1,$3F,$E8,$6C
	.dc.b	$FF,$31,$7F,$EB,$28,$C7,$36,$DD,$5D,$9F,$E7,$D5,$C2,$F9,$C6,$27
	.dc.b	$5D,$0A,$FE,$39,$3D,$4D,$63,$FE,$51,$BE,$30,$F9,$C6,$AF,$4B,$F9
	.dc.b	$D4,$61,$B1,$99,$FF,$63,$CE,$5F,$F6,$4F,$C1,$BE,$4F,$F9,$6C,$61
	.dc.b	$AF,$FB,$46,$35,$E2,$FF,$47,$6C,$74,$E1,$7C,$1F,$89,$B1,$98,$C1
	.dc.b	$FE,$D8,$DB,$E0,$76,$FF,$56,$D7,$E2,$C7,$C3,$ED,$C7,$17,$D1,$6E
	.dc.b	$1F,$32,$ED,$72,$8F,$8D,$5B,$69,$F4,$AB,$7C,$2B,$BD,$BF,$D9,$B9
	.dc.b	$FF,$2E,$DE,$3F,$C0,$37,$0F,$B7,$DB,$73,$FB,$13,$71,$FF,$1F,$57
	.dc.b	$FA,$42,$31,$59,$64,$64,$FF,$4A,$C7,$1E,$D9,$1F,$E9,$87,$A7,$6E
	.dc.b	$23,$F5,$D1,$72,$FD,$6C,$F8,$7B,$65,$7F,$A8,$1F,$27,$6E,$23,$F6
	.dc.b	$D1,$6B,$F6,$B3,$EB,$EE,$7C,$CD,$25,$FF,$38,$F9,$9A,$FF,$4F,$8B
	.dc.b	$7F,$D5,$27,$E2,$7E,$B7,$F8,$0A,$FF,$29,$B6,$7F,$C5,$0D,$E0,$FF
	.dc.b	$78,$AF,$86,$E3,$CD,$2F,$FA,$6C,$ED,$FF,$14,$3B,$73,$1B,$7B,$BE
	.dc.b	$29,$F1,$FB,$72,$79,$BE,$6E,$6C,$47,$9A,$2E,$3E,$69,$B8,$F9,$A6
	.dc.b	$D7,$8A,$7D,$FF,$31,$23,$CE,$FC,$2F,$7B,$11,$F0,$A2,$E1,$F0,$A6
	.dc.b	$E1,$F0,$A6,$D7,$BA,$7E,$5F,$BD,$23,$E1,$BF,$3B,$E2,$71,$1F,$3A
	.dc.b	$2E,$1F,$3A,$6E,$1F,$3A,$6D,$7C,$A9,$FA,$FF,$13,$23,$E7,$BF,$6B
	.dc.b	$C0,$C4,$7D,$A8,$B8,$7D,$A9,$B8,$7D,$A9,$B5,$F5,$A7,$F0,$7F,$2F
	.dc.b	$F8,$4F,$97,$93,$F7,$23,$2D,$21,$F8,$8F,$96,$FC,$59,$1F,$C1,$BF
	.dc.b	$8E,$F8,$AF,$C7,$FC,$9F,$E4,$3E,$4A,$57,$E2,$23,$F2,$7F,$23,$23
	.dc.b	$F1,$6F,$E5,$BE,$4B,$F2,$FF,$23,$27,$F1,$EF,$E6,$BE,$4B,$F8,$04
	.dc.b	$3E,$69,$BE,$1B,$1F,$E2,$A3,$51,$9D,$F4,$C8,$D4,$F0,$7F,$DE,$67
	.dc.b	$59,$D3,$88,$D7,$0F,$F9,$F9,$BC,$3F,$E7,$C7,$F9,$C4,$FE,$9B,$E4
	.dc.b	$7F,$8A,$78,$3F,$E8,$08,$FF,$61,$B6,$FF,$63,$3F,$B6,$37,$30,$E4
	.dc.b	$7A,$FE,$6B,$F8,$DB,$CC,$38,$7B,$35,$FF,$47,$B5,$1C,$3C,$5F,$97
	.dc.b	$1D,$2B,$FD,$32,$DC,$7D,$9D,$BE,$AF,$43,$D7,$F1,$F3,$1F,$F2,$7E
	.dc.b	$7F,$D6,$DF,$6C,$E5,$F7,$CF,$17,$71,$9C,$7F,$D7,$14,$AD,$57,$D2
	.dc.b	$C5,$F2,$7E,$62,$47,$DD,$FB,$C7,$9B,$88,$E1,$37,$1F,$F5,$9D,$6A
	.dc.b	$7C,$FF,$29,$F5,$3E,$52,$46,$08,$7D,$67,$2F,$D8,$8F,$C8,$FC,$D3
	.dc.b	$E2,$E9,$27,$F2,$10,$F8,$BC,$47,$E3,$23,$83,$A9,$E7,$3A,$C7,$FD
	.dc.b	$93,$5A,$AE,$15,$A7,$F3,$9E,$C3,$AB,$FD,$8D,$1F,$E1,$06,$C6,$63
	.dc.b	$FD,$94,$D7,$F1,$0D,$BF,$F6,$BB,$78,$FF,$DA,$AD,$9F,$C3,$FA,$6E
	.dc.b	$CF,$D8,$FC,$D3,$E3,$69,$2B,$F6,$1F,$34,$E6,$C4,$7E,$B2,$B8,$3F
	.dc.b	$B5,$6B,$FB,$09,$E0,$FE,$E7,$E2,$BC,$0D,$C3,$F8,$0C,$62,$3F,$6E
	.dc.b	$F9,$DF,$F7,$1C,$68,$8F,$3D,$6B,$C2,$2E,$BF,$EF,$8A,$E5,$C4,$F8
	.dc.b	$9F,$E1,$ED,$BF,$E1,$8D,$7F,$83,$FE,$19,$FD,$8D,$46,$6D,$C1,$FF
	.dc.b	$00,$23,$0D,$BF,$E3,$77,$C7,$F1,$8F,$F7,$A6,$1F,$C6,$5A,$FF,$14
	.dc.b	$6E,$1D,$23,$F7,$7F,$86,$FC,$38,$C4,$27,$78,$7F,$00,$8C,$7C,$0F
	.dc.b	$F2,$A9,$CC,$72,$D3,$F8,$BF,$CB,$48,$FD,$EB,$FC,$6B,$E5,$F1,$1F
	.dc.b	$C6,$A2,$E3,$FB,$A9,$FE,$3F,$F2,$F2,$BF,$82,$BF,$C8,$BE,$5F,$11
	.dc.b	$FC,$8A,$2D,$7F,$02,$9F,$F1,$C7,$CF,$CA,$FF,$6D,$FC,$02,$3F,$C7
	.dc.b	$9D,$F9,$9F,$F4,$F4,$5B,$FF,$4E,$CF,$F9,$07,$E5,$25,$72,$7F,$C8
	.dc.b	$9E,$E6,$23,$FD,$4D,$16,$BF,$D4,$B3,$78,$3F,$D1,$D3,$6B,$FD,$13
	.dc.b	$37,$E0,$BF,$D0,$D3,$72,$FF,$1B,$D7,$A3,$DD,$FF,$29,$11,$FF,$7A
	.dc.b	$FE,$CB,$FE,$29,$BF,$F4,$74,$E6,$7F,$D7,$0F,$F9,$6E,$74,$7F,$8B
	.dc.b	$61,$97,$E0,$69,$39,$8C,$37,$2F,$E7,$F3,$AF,$73,$A7,$BB,$30,$F1
	.dc.b	$BD,$E3,$11,$DA,$B8,$DF,$38,$5F,$82,$F8,$47,$B3,$DD,$F4,$11,$F7
	.dc.b	$FB,$B5,$E8,$72,$BF,$F5,$4E,$5E,$8A,$F8,$15,$B8,$C5,$75,$39,$94
	.dc.b	$E7,$87,$C2,$AE,$18,$FA,$2F,$FC,$0F,$F0,$DF,$E0,$6C,$3F,$B0,$FC
	.dc.b	$EB,$97,$F4,$92,$F7,$7F,$4D,$39,$CF,$FB,$3C,$E5,$E3,$5C,$3D,$5F
	.dc.b	$A7,$0A,$E0,$AB,$C0,$E6,$BB,$DF,$66,$C7,$43,$BD,$F4,$BE,$77,$C9
	.dc.b	$B7,$C6,$BC,$AD,$7C,$15,$C6,$FC,$D8,$75,$9F,$25,$6F,$0F,$64,$79
	.dc.b	$4F,$0C,$EE,$3C,$15,$E6,$9D,$1C,$8D,$5F,$77,$89,$FE,$D0,$DB,$C2
	.dc.b	$D3,$3E,$8F,$E6,$EF,$E2,$BF,$9D,$8F,$F7,$61,$B3,$FD,$A4,$53,$2E
	.dc.b	$D7,$C3,$7F,$87,$23,$0F,$BE,$3F,$DE,$5A,$D7,$67,$A8,$C9,$4C,$E9
	.dc.b	$F6,$9C,$B5,$CE,$5F,$E3,$BF,$4B,$F9,$94,$E6,$FC,$11,$89,$FF,$1B
	.dc.b	$46,$2F,$C4,$DB,$F0,$36,$F9,$47,$96,$F9,$3F,$E4,$99,$D5,$FE,$D3
	.dc.b	$E0,$3F,$A1,$6B,$FC,$01,$B8,$FE,$E3,$E0,$2E,$1F,$AF,$9B,$C5,$FC
	.dc.b	$21,$BC,$66,$D6,$19,$7F,$C6,$1C,$B4,$97,$FB,$2A,$E5,$1F,$C7,$AB
	.dc.b	$8C,$E7,$87,$B9,$E0,$E1,$B5,$FC,$92,$2C,$F3,$9B,$97,$F1,$C9,$B5
	.dc.b	$FC,$52,$6D,$7F,$0B,$F6,$3D,$67,$B3,$DF,$B5,$FC,$C2,$2C,$F9,$E6
	.dc.b	$E1,$FC,$A3,$E0,$3E,$19,$F1,$3E,$02,$E1,$FC,$E2,$2D,$FB,$E6,$D7
	.dc.b	$1A,$D1,$FC,$FA,$32,$FF,$2A,$AB,$5F,$C8,$66,$ED,$FC,$EB,$E0,$3E
	.dc.b	$99,$F5,$3D,$FB,$87,$F4,$88,$B7,$F3,$E6,$E3,$FE,$6A,$F9,$9F,$BB
	.dc.b	$E8,$1B,$E8,$E1,$98,$FF,$9E,$1C,$B4,$FF,$3D,$FA,$9C,$8E,$7E,$B5
	.dc.b	$8C,$93,$2F,$F4,$DC,$65,$FF,$54,$FE,$6C,$F8,$FF,$C8,$C2,$31,$5F
	.dc.b	$E9,$2C,$FF,$C3,$5D,$4F,$B6,$78,$BC,$2F,$78,$74,$C7,$07,$E7,$2F
	.dc.b	$85,$63,$1F,$F6,$07,$ED,$3F,$E5,$16,$EF,$BF,$A0,$7C,$CF,$92,$B5
	.dc.b	$97,$FB,$01,$BA,$FF,$8A,$9B,$87,$FA,$A3,$26,$1C,$A3,$A9,$EB,$3F
	.dc.b	$D5,$A3,$E8,$1C,$FF,$B2,$8C,$36,$1C,$F6,$24,$F3,$CB,$FC,$93,$5C
	.dc.b	$A3,$07,$5A,$DC,$FF,$6C,$8E,$A6,$A7,$84,$65,$EF,$19,$AF,$AB,$E3
	.dc.b	$BC,$AA,$7C,$DF,$A0,$7C,$76,$2F,$FD,$D9,$5B,$6D,$6E,$75,$8F,$FB
	.dc.b	$AB,$EC,$BB,$B7,$3F,$E2,$7F,$87,$3F,$0E,$7C,$B7,$E1,$AF,$7C,$5F
	.dc.b	$E5,$A7,$8E,$1F,$EE,$96,$BF,$E5,$B6,$BF,$ED,$C6,$FC,$FF,$FB,$7B
	.dc.b	$DC,$FE,$42,$DD,$7E,$63,$FE,$E2,$F9,$EC,$DF,$F9,$FE,$72,$D3,$FC
	.dc.b	$2F,$3C,$5A,$27,$8D,$4A,$E2,$F4,$8C,$4F,$FC,$00,$FA,$1F,$DA,$FD
	.dc.b	$17,$ED,$9B,$7F,$EE,$A6,$CE,$E3,$FD,$DB,$3F,$AA,$FA,$2F,$F6,$64
	.dc.b	$7E,$16,$51,$87,$FC,$5B,$CE,$E4,$7A,$BC,$7C,$09,$87,$1C,$38,$6F
	.dc.b	$FB,$2B,$DC,$E6,$BF,$85,$F5,$FF,$0C,$DA,$CF,$FC,$00,$6E,$7F,$E3
	.dc.b	$2E,$CF,$91,$8F,$FB,$F2,$F8,$5F,$E6,$FC,$6F,$F6,$24,$62,$F8,$36
	.dc.b	$BF,$D3,$39,$BE,$83,$E0,$1F,$D6,$BE,$8D,$F5,$2E,$CC,$E7,$EB,$AF
	.dc.b	$F5,$0E,$5F,$E9,$FC,$39,$4F,$FA,$86,$31,$3A,$47,$CA,$71,$9F,$FB
	.dc.b	$67,$93,$F9,$5F,$D2,$7F,$2C,$6D,$FA,$F6,$FC,$13,$67,$4D,$3F,$1F
	.dc.b	$F3,$CF,$8B,$FC,$9B,$63,$3F,$21,$89,$91,$C1,$DC,$63,$3F,$C9,$D7
	.dc.b	$16,$CE,$AB,$81,$E5,$CF,$FA,$19,$FD,$13,$E9,$2C,$7A,$7E,$93,$FA
	.dc.b	$39,$FE,$04,$FC,$BF,$FD,$AF,$19,$3E,$C3,$FC,$CB,$8F,$FD,$20,$BE
	.dc.b	$93,$87,$95,$C5,$E5,$F2,$32,$F8,$FD,$15,$D7,$FC,$EF,$83,$E9,$3C
	.dc.b	$06,$72,$EA,$E2,$FC,$3F,$19,$6F,$8B,$E7,$F1,$24,$F9,$B8,$2D,$76
	.dc.b	$9B,$87,$A2,$6D,$7F,$A9,$31,$FE,$22,$FB,$AB,$E1,$4E,$AF,$FD,$3F
	.dc.b	$4E,$2F,$8C,$7C,$7F,$99,$B8,$7C,$68,$B5,$EE,$9B,$8F,$C8,$9B,$5E
	.dc.b	$19,$B5,$DB,$E8,$FE,$91,$F4,$FC,$AB,$5F,$4A,$2C,$FC,$A9,$B8,$7D
	.dc.b	$49,$B5,$ED,$9B,$5E,$AE,$7F,$BC,$7D,$FF,$A3,$B5,$F7,$A2,$CF,$D6
	.dc.b	$9B,$87,$E0,$26,$D7,$C9,$9B,$5F,$1B,$E8,$FF,$14,$7E,$2F,$E2,$ED
	.dc.b	$7E,$2A,$2C,$FE,$0A,$6E,$1F,$8C,$9B,$5F,$56,$6D,$7D,$2F,$8B,$FC
	.dc.b	$B1,$F9,$7F,$52,$D7,$E5,$A2,$CF,$E3,$A6,$E1,$FE,$E4,$FA,$1F,$F2
	.dc.b	$F9,$FE,$60,$F9,$0B,$87,$FB,$D2,$2D,$FF,$96,$E6,$D7,$FB,$B3,$E5
	.dc.b	$BF,$CC,$E7,$F9,$A3,$E8,$6E,$1F,$EF,$C8,$B7,$FE,$63,$9B,$5F,$EF
	.dc.b	$4F,$A1,$FF,$37,$9F,$E7,$0F,$A0,$B8,$7F,$25,$F8,$74,$E3,$0F,$F3
	.dc.b	$65,$5B,$E5,$7B,$AE,$4D,$3F,$97,$7E,$9B,$FE,$12,$BF,$F5,$2F,$C8
	.dc.b	$FF,$1E,$F1,$D3,$FE,$A0,$F8,$C4,$E6,$38,$46,$2B,$FD,$FF,$1B,$9D
	.dc.b	$42,$FF,$94,$D6,$EF,$AF,$A5,$9A,$F0,$9C,$8C,$4E,$F0,$E9,$1F,$CC
	.dc.b	$BE,$95,$E1,$2F,$F9,$C4,$E6,$30,$E2,$77,$3A,$84,$61,$E3,$F2,$0F
	.dc.b	$89,$79,$5B,$8C,$36,$33,$3C,$1A,$FF,$49,$A7,$8A,$F9,$0F,$04,$61
	.dc.b	$B1,$99,$F3,$35,$FE,$9E,$35,$7C,$9F,$F4,$58,$F9,$7C,$1F,$5E,$3F
	.dc.b	$88,$46,$0E,$0E,$1F,$9B,$3C,$BC,$4E,$BC,$1F,$4E,$BA,$56,$9E,$11
	.dc.b	$9F,$0F,$78,$77,$BB,$7D,$BC,$3F,$3F,$2F,$DB,$CB,$CF,$18,$E4,$F3
	.dc.b	$36,$BC,$BE,$1D,$AD,$C7,$F8,$07,$45,$65,$B1,$9A,$C4,$C8,$E0,$EE
	.dc.b	$38,$78,$72,$7A,$3C,$FC,$3C,$43,$2F,$46,$1F,$E5,$A8,$D9,$E3,$9F
	.dc.b	$17,$87,$23,$CC,$F9,$7C,$3C,$47,$3D,$3E,$A4,$E3,$0E,$B5,$EF,$FA
	.dc.b	$5F,$F3,$55,$FF,$87,$DA,$F1,$3F,$00,$E2,$7E,$63,$FE,$36,$8E,$1F
	.dc.b	$01,$CC,$8F,$3F,$67,$E6,$7D,$2F,$E2,$EF,$EA,$37,$2F,$DF,$36,$BF
	.dc.b	$1C,$BE,$53,$8F,$3C,$9F,$F1,$A4,$EF,$88,$EE,$33,$3B,$7D,$53,$A8
	.dc.b	$F1,$D7,$93,$D5,$95,$D1,$F3,$F2,$E2,$38,$E1,$FE,$39,$74,$78,$A3
	.dc.b	$59,$7B,$2A,$E5,$EC,$9F,$83,$F2,$B2,$FD,$2F,$C4,$F9,$3C,$47,$68
	.dc.b	$B8,$7C,$69,$B7,$F1,$A6,$D7,$C6,$9F,$9B,$F2,$72,$3E,$2B,$F4,$3E
	.dc.b	$4F,$11,$EA,$8B,$87,$D2,$9B,$87,$D2,$9B,$5F,$4A,$7E,$CF,$C9,$C8
	.dc.b	$FA,$2F,$DC,$EF,$C4,$7C,$68,$B8,$7D,$E9,$B8,$7D,$E9,$B5,$F7,$A7
	.dc.b	$F0,$BF,$27,$F8,$6F,$89,$93,$F7,$A3,$2D,$21,$F5,$3E,$27,$F1,$A4
	.dc.b	$7F,$0A,$FE,$43,$E4,$FF,$23,$F2,$7F,$92,$F4,$25,$7E,$2A,$3F,$2B
	.dc.b	$E8,$48,$FC,$6B,$F9,$8F,$43,$F3,$3E,$84,$9F,$C8,$BF,$9C,$F3,$BF
	.dc.b	$85,$43,$E9,$5B,$E1,$B1,$FE,$A1,$6B,$FA,$D3,$86,$FB,$99,$1B,$79
	.dc.b	$46,$23,$84,$F2,$E1,$FE,$B1,$1E,$1A,$DB,$FA,$6F,$43,$F8,$E7,$85
	.dc.b	$FE,$8B,$AF,$E4,$8D,$E3,$FD,$8C,$FE,$D8,$DD,$43,$91,$FE,$B9,$FA
	.dc.b	$4F,$E4,$2F,$8B,$AB,$DC,$7C,$28,$CE,$1E,$47,$73,$FE,$9B,$6B,$FD
	.dc.b	$31,$AF,$FA,$71,$BD,$CC,$7F,$B4,$CF,$FA,$DF,$E9,$7D,$67,$B3,$D0
	.dc.b	$91,$FE,$9D,$8C,$BF,$E9,$F8,$CB,$C4,$D1,$93,$D9,$EA,$FD,$AF,$A5
	.dc.b	$79,$54,$F6,$44,$9D,$90,$FB,$53,$FD,$77,$3F,$9B,$F4,$BF,$43,$E8
	.dc.b	$65,$7F,$AC,$63,$E9,$98,$7F,$DA,$51,$87,$FD,$CD,$F4,$AF,$98,$FC
	.dc.b	$6B,$63,$31,$8F,$8C,$CB,$23,$27,$28,$C3,$F8,$F9,$E4,$D9,$E4,$FF
	.dc.b	$BD,$7D,$87,$57,$E7,$3D,$27,$5D,$3F,$3F,$3A,$AE,$15,$FE,$1B,$7F
	.dc.b	$C4,$6D,$8D,$47,$F8,$91,$AF,$FB,$77,$E9,$5F,$37,$FA,$AF,$A5,$7A
	.dc.b	$34,$FD,$6C,$EA,$BF,$DC,$3C,$1F,$A0,$8C,$47,$EA,$DB,$19,$BE,$9C
	.dc.b	$DF,$B6,$FA,$57,$83,$49,$87,$4E,$DF,$F0,$FC,$FA,$30,$FD,$6B,$C2
	.dc.b	$74,$7F,$BC,$9C,$3F,$BB,$6C,$70,$D7,$AC,$E6,$EC,$E6,$39,$8E,$37
	.dc.b	$C8,$7F,$0C,$FA,$5F,$EA,$13,$FC,$1D,$B5,$FC,$02,$B7,$3C,$9A,$FF
	.dc.b	$0D,$6C,$6E,$BA,$7A,$3F,$C6,$3E,$91,$F0,$DF,$C6,$BE,$97,$EF,$E3
	.dc.b	$E4,$6B,$FC,$5E,$76,$FA,$1A,$E6,$7A,$1C,$F3,$FE,$23,$F4,$B2,$7E
	.dc.b	$4F,$83,$FB,$D8,$FE,$2B,$F4,$59,$9F,$EE,$FC,$49,$C6,$BF,$C6,$2B
	.dc.b	$F8,$DF,$D2,$CB,$E7,$68,$FE,$3F,$F4,$B9,$9E,$71,$6F,$F9,$0C,$FF
	.dc.b	$8E,$FE,$26,$57,$F8,$FD,$FF,$1F,$78,$58,$8F,$F0,$4F,$BC,$9C,$5F
	.dc.b	$F8,$DA,$BF,$C8,$9F,$13,$2F,$FC,$90,$FF,$91,$FD,$DC,$47,$28,$B5
	.dc.b	$FE,$3E,$9B,$D3,$FE,$69,$9B,$5F,$E6,$89,$BB,$FF,$9C,$26,$E5,$FE
	.dc.b	$33,$BF,$F3,$87,$BB,$FE,$55,$26,$1F,$EB,$5E,$5A,$FA,$A2,$E5,$FE
	.dc.b	$65,$F8,$9F,$F1,$51,$FE,$29,$9F,$F1,$3B,$63,$5E,$F7,$F9,$96,$3F
	.dc.b	$23,$87,$FA,$D5,$D3,$63,$46,$FC,$DB,$7F,$EC,$78,$AF,$68,$E9,$EE
	.dc.b	$C8,$FF,$66,$3D,$FD,$DC,$47,$92,$2E,$3E,$99,$B8,$78,$E6,$D7,$FA
	.dc.b	$5F,$15,$63,$64,$7F,$B1,$E7,$A1,$FE,$DB,$9F,$EF,$7F,$11,$FD,$EB
	.dc.b	$14,$BE,$95,$CF,$FE,$A4,$A7,$BB,$FE,$A6,$9C,$D7,$09,$CB,$CA,$32
	.dc.b	$A3,$B3,$D6,$BA,$7A,$9F,$E8,$38,$C1,$CE,$FB,$4F,$85,$B1,$DE,$78
	.dc.b	$5F,$6B,$EA,$DA,$DD,$78,$5B,$3A,$BF,$0E,$1E,$77,$B5,$53,$C4,$53
	.dc.b	$D5,$87,$82,$38,$C7,$A0,$EC,$6E,$BE,$1B,$97,$C0,$FA,$6B,$E7,$F0
	.dc.b	$FE,$4E,$7E,$2B,$6F,$E2,$36,$77,$C1,$FE,$F0,$F8,$6B,$87,$FB,$60
	.dc.b	$38,$7F,$D9,$E7,$FB,$8E,$38,$34,$E1,$1E,$D7,$FD,$F9,$C1,$5E,$67
	.dc.b	$B4,$C9,$4C,$E9,$F7,$9C,$B5,$CE,$7E,$5F,$49,$5D,$6F,$C1,$18,$AE
	.dc.b	$91,$89,$D4,$78,$67,$57,$D6,$B5,$3C,$0D,$57,$09,$F2,$57,$0B,$FD
	.dc.b	$D7,$C0,$77,$CB,$F8,$23,$6F,$F7,$DF,$01,$71,$FD,$64,$DE,$2F,$D5
	.dc.b	$B7,$88,$DA,$C3,$2F,$FA,$0C,$E9,$B3,$9B,$FE,$23,$57,$0F,$E2,$D3
	.dc.b	$6B,$F8,$87,$0F,$73,$C1,$C1,$63,$AC,$5C,$3F,$8D,$CD,$BF,$E4,$13
	.dc.b	$6B,$F8,$D4,$DA,$FE,$11,$EC,$FA,$CF,$67,$BF,$6B,$D3,$16,$7F,$93
	.dc.b	$CD,$C3,$F9,$67,$C0,$7C,$33,$E2,$7C,$05,$C3,$E0,$C5,$BF,$E5,$F3
	.dc.b	$6B,$8D,$68,$F8,$F1,$97,$F9,$85,$5A,$FE,$3D,$37,$6F,$E3,$BF,$01
	.dc.b	$F4,$CF,$A9,$EF,$DC,$3E,$8C,$5B,$FE,$7F,$37,$1F,$F4,$2F,$D2,$FD
	.dc.b	$DF,$10,$DF,$3F,$0C,$C7,$FB,$8B,$96,$9F,$DD,$3D,$6E,$47,$3F,$62
	.dc.b	$C6,$49,$97,$F8,$02,$32,$FF,$A9,$7E,$91,$F2,$3F,$91,$84,$62,$BF
	.dc.b	$C6,$7D,$2F,$A5,$78,$BB,$AD,$5F,$3C,$3F,$35,$6F,$CC,$5F,$01,$B9
	.dc.b	$CC,$63,$2F,$F1,$AF,$D3,$7F,$A4,$DB,$C1,$CB,$E9,$9F,$3D,$E5,$AD
	.dc.b	$67,$FE,$BD,$6F,$1F,$EE,$F4,$5F,$F8,$FB,$1E,$67,$A9,$C5,$AB,$F8
	.dc.b	$14,$BE,$99,$D3,$FB,$38,$C3,$61,$D3,$62,$4F,$F9,$60,$D9,$67,$6D
	.dc.b	$8D,$DF,$86,$3F,$B0,$9C,$1C,$1F,$C6,$27,$8B,$88,$D5,$FF,$B5,$3C
	.dc.b	$97,$99,$F6,$5B,$3F,$63,$E8,$1F,$1D,$9A,$93,$9E,$0B,$8F,$83,$66
	.dc.b	$1F,$32,$BF,$BA,$CB,$E9,$3C,$6F,$F4,$16,$BA,$6B,$FD,$F1,$AF,$1D
	.dc.b	$BF,$8F,$7D,$0B,$ED,$FF,$C1,$37,$FE,$76,$BF,$7E,$DF,$CD,$6B,$5E
	.dc.b	$77,$F2,$5F,$A5,$FF,$03,$FB,$97,$3F,$B7,$7F,$CE,$E7,$FC,$17,$F4
	.dc.b	$39,$A9,$97,$FA,$32,$78,$B6,$D5,$A7,$E8,$CE,$9A,$76,$9F,$F0,$BC
	.dc.b	$76,$79,$57,$EE,$FE,$97,$F7,$9F,$4B,$67,$FC,$D2,$DB,$FF,$38,$B7
	.dc.b	$1F,$EA,$5E,$37,$FB,$CE,$3F,$65,$F4,$8F,$0A,$9F,$AC,$E2,$7A,$D8
	.dc.b	$CB,$FD,$3B,$5B,$7F,$6F,$F8,$AB,$D5,$D6,$9F,$76,$31,$3F,$E7,$CE
	.dc.b	$FF,$E2,$5B,$FF,$13,$6C,$64,$FF,$40,$C6,$5A,$66,$BF,$D1,$71,$96
	.dc.b	$C7,$93,$E9,$3F,$B7,$47,$F8,$1F,$14,$62,$3F,$1B,$18,$9F,$F0,$64
	.dc.b	$7F,$91,$23,$0E,$1B,$BF,$F6,$BF,$86,$FE,$27,$5D,$CF,$F6,$4C,$23
	.dc.b	$35,$FE,$A0,$E3,$FE,$61,$F4,$BF,$CC,$5B,$7F,$E9,$E6,$CE,$E3,$FD
	.dc.b	$43,$3F,$91,$F9,$E7,$C7,$53,$1F,$39,$6B,$F1,$D3,$A6,$CE,$9A,$4B
	.dc.b	$FC,$7C,$ED,$B5,$CD,$FF,$5B,$E9,$FD,$1C,$FE,$91,$F4,$B6,$FF,$D6
	.dc.b	$BF,$4B,$FD,$2C,$FF,$04,$7D,$17,$F8,$82,$32,$7F,$B1,$1F,$E6,$FB
	.dc.b	$7E,$41,$7D,$2F,$0F,$2F,$8B,$CB,$E4,$A5,$F1,$FA,$1B,$A7,$F1,$6E
	.dc.b	$C6,$2F,$A5,$F0,$9A,$CF,$B3,$8C,$3C,$5F,$19,$70,$E5,$7F,$EE,$69
	.dc.b	$F3,$7D,$2F,$A8,$F5,$F1,$D8,$C9,$2B,$8C,$65,$E7,$37,$1F,$74,$DA
	.dc.b	$FF,$47,$E5,$E8,$E2,$F8,$E7,$C8,$FA,$3B,$3F,$1E,$2E,$3E,$69,$B8
	.dc.b	$7C,$99,$B5,$D3,$25,$62,$D8,$7C,$6F,$9B,$FA,$A7,$D6,$F3,$2E,$1F
	.dc.b	$56,$2D,$7C,$49,$B9,$7D,$79,$B5,$DE,$6D,$76,$C7,$F9,$C3,$7C,$BF
	.dc.b	$C1,$FD,$1F,$E1,$19,$7F,$83,$72,$D2,$4F,$D4,$8C,$B6,$3F,$15,$F2
	.dc.b	$FF,$8E,$3F,$1F,$F1,$B6,$32,$49,$FB,$B1,$97,$EF,$CD,$CF,$F2,$53
	.dc.b	$6B,$E5,$7C,$6F,$E6,$8F,$CD,$FB,$36,$3F,$35,$17,$0F,$C4,$CD,$BF
	.dc.b	$F7,$87,$D0,$FF,$99,$CF,$F3,$47,$C9,$5C,$3F,$CC,$B1,$6F,$FD,$E7
	.dc.b	$36,$BF,$DF,$1F,$2F,$FE,$6F,$3F,$CE,$1F,$35,$70,$FF,$36,$C5,$BF
	.dc.b	$F7,$DC,$DA,$FF,$80,$1F,$43,$FE,$77,$3F,$CF,$1F,$41,$70,$FF,$3A
	.dc.b	$C5,$BF,$D5,$7C,$62,$B1,$7C,$AF,$67,$F9,$F2,$32,$FE,$6B,$E9,$7F
	.dc.b	$C1,$37,$FE,$B2,$F9,$3C,$7B,$3F,$E9,$DF,$91,$46,$E7,$0E,$AF,$FD
	.dc.b	$E3,$3A,$9F,$CB,$DF,$48,$C7,$4A,$3F,$D6,$71,$87,$57,$CE,$BC,$47
	.dc.b	$FB,$F3,$E9,$5E,$13,$2F,$E7,$B2,$9C,$C6,$1B,$1F,$D0,$9A,$FF,$B3
	.dc.b	$FE,$51,$F1,$AF,$2B,$9C,$61,$B1,$99,$FE,$97,$39,$7A,$78,$BF,$D9
	.dc.b	$5B,$E9,$78,$63,$11,$FE,$DD,$C1,$18,$8F,$F7,$4F,$ED,$CF,$8B,$FF
	.dc.b	$12,$EF,$F2,$70,$4E,$6B,$E6,$ED,$C2,$F6,$FD,$9F,$87,$B3,$FD,$DE
	.dc.b	$B8,$3F,$2A,$F9,$DE,$9B,$1D,$0F,$AF,$5C,$97,$EA,$6F,$77,$FC,$73
	.dc.b	$CD,$FD,$CA,$31,$23,$FC,$9E,$70,$9D,$F2,$FE,$37,$E8,$AC,$EE,$B8
	.dc.b	$BC,$1B,$7D,$6F,$7C,$DF,$87,$BE,$37,$D4,$F0,$56,$DB,$87,$F9,$3B
	.dc.b	$0E,$6F,$86,$32,$3D,$F7,$ED,$AF,$9A,$F2,$C6,$1B,$19,$9E,$39,$7E
	.dc.b	$67,$E7,$F3,$52,$7A,$DF,$17,$7D,$4A,$F5,$AF,$9D,$F8,$79,$7F,$A0
	.dc.b	$BE,$66,$FC,$39,$DC,$CC,$3F,$D0,$93,$96,$CE,$5F,$A3,$7F,$9C,$7D
	.dc.b	$E7,$EB,$FA,$BE,$4C,$66,$3F,$B8,$7E,$92,$F9,$CC,$D4,$9E,$35,$BE
	.dc.b	$1F,$F2,$6D,$EE,$7A,$FC,$EF,$EC,$DF,$8B,$0F,$8C,$FB,$2D,$8D,$F8
	.dc.b	$1F,$6A,$B8,$47,$E2,$1B,$7F,$E3,$E6,$ED,$FD,$BA,$FF,$86,$56,$EF
	.dc.b	$91,$1F,$0F,$5F,$D8,$F8,$1F,$0F,$44,$9F,$AB,$CC,$9D,$46,$1A,$FC
	.dc.b	$4F,$9C,$BA,$7F,$A2,$6B,$84,$EC,$F7,$CF,$E1,$BD,$DF,$7F,$CE,$49
	.dc.b	$EF,$19,$69,$0F,$A3,$F3,$99,$7F,$D6,$FD,$1F,$94,$9C,$63,$F4,$2B
	.dc.b	$F2,$BF,$9C,$FF,$22,$BF,$9F,$F3,$BF,$38,$CF,$91,$23,$F2,$73,$B7
	.dc.b	$AB,$6F,$ED,$F3,$6E,$73,$56,$73,$1C,$BD,$AF,$B7,$F3,$B2,$30,$43
	.dc.b	$AC,$79,$8F,$C0,$FC,$A5,$8D,$57,$FA,$CF,$15,$6A,$7F,$13,$D9,$F8
	.dc.b	$82,$3E,$AF,$5B,$E0,$91,$C1,$F9,$0F,$9D,$FC,$79,$97,$DF,$18,$7F
	.dc.b	$2B,$F3,$BF,$94,$F5,$64,$60,$FD,$CF,$DD,$7F,$38,$CF,$FD,$90,$DE
	.dc.b	$CF,$E3,$DC,$88,$C5,$7F,$8D,$3B,$1E,$B7,$E8,$3D,$6F,$D0,$B2,$7F
	.dc.b	$4B,$1F,$C6,$BE,$A3,$E9,$63,$FE,$E4,$FA,$96,$BF,$C7,$9B,$7F,$BA
	.dc.b	$AD,$BF,$CB,$DA,$9A,$62,$1C,$1D,$F9,$65,$73,$7B,$47,$ED,$BD,$89
	.dc.b	$3F,$BA,$7F,$73,$EC,$62,$3F,$83,$C5,$E0,$FE,$07,$3F,$CF,$3C,$27
	.dc.b	$ED,$28,$C3,$6F,$F9,$8C,$EA,$73,$EC,$69,$93,$AA,$E0,$AD,$B7,$F4
	.dc.b	$FC,$3B,$E1,$F2,$FE,$0B,$FC,$07,$7F,$0D,$BB,$7F,$9A,$A3,$FC,$0E
	.dc.b	$D7,$FD,$8A,$DC,$3F,$D9,$AD,$C3,$FA,$E4,$F3,$3F,$95,$ED,$F8,$07
	.dc.b	$07,$F6,$09,$E0,$FD,$5B,$FB,$5E,$97,$F4,$CE,$47,$75,$25,$7F,$28
	.dc.b	$CE,$BF,$C0,$A7,$6F,$E5,$27,$F9,$5C,$FF,$64,$FD,$D7,$CB,$3F,$EA
	.dc.b	$96,$E9,$F1,$A3,$FD,$0F,$18,$D3,$FD,$2E,$FF,$2C,$8C,$9C,$C8,$7F
	.dc.b	$5E,$74,$FF,$8C,$3C,$7F,$EE,$3F,$54,$F8,$BA,$7F,$74,$E5,$FE,$6B
	.dc.b	$8F,$FB,$D9,$B8,$FC,$A6,$E1,$FE,$07,$6A,$BF,$F7,$DB,$FD,$62,$32
	.dc.b	$4C,$3C,$3F,$19,$F0,$7D,$69,$5E,$77,$E2,$7A,$D8,$8F,$89,$17,$2F
	.dc.b	$89,$EA,$B5,$5F,$F8,$33,$EA,$6E,$1F,$DA,$1A,$FF,$87,$1B,$7F,$36
	.dc.b	$3F,$DF,$3E,$77,$1A,$91,$C5,$FF,$0B,$FC,$47,$F8,$4A,$3F,$C4,$DF
	.dc.b	$87,$7F,$B9,$9F,$EE,$BF,$87,$BE,$6B,$FC,$3A,$D8,$FE,$F3,$3B,$7F
	.dc.b	$C5,$AD,$C7,$C4,$DE,$2F,$CB,$B6,$7F,$C6,$94,$FA,$5F,$F1,$4C,$E6
	.dc.b	$4D,$31,$3B,$8F,$F0,$4F,$B4,$E0,$FF,$18,$B6,$33,$1F,$E3,$4F,$BB
	.dc.b	$F8,$B8,$FE,$A2,$B5,$5F,$E4,$46,$E5,$F2,$5B,$3F,$AE,$AE,$35,$FE
	.dc.b	$4B,$FC,$E3,$F3,$0D,$3F,$C8,$2E,$17,$EB,$BF,$B3,$8C,$12,$3F,$C8
	.dc.b	$F1,$FC,$DB,$C9,$FE,$71,$E3,$7E,$CC,$93,$F9,$FC,$6B,$FB,$18,$B9
	.dc.b	$7E,$23,$93,$FC,$75,$3F,$A5,$F7,$7F,$65,$72,$30,$43,$F6,$8E,$5F
	.dc.b	$DB,$47,$F0,$3F,$1A,$47,$F5,$07,$FC,$8B,$EE,$E2,$3F,$81,$45,$C3
	.dc.b	$F3,$3E,$E7,$F9,$06,$7F,$83,$77,$FF,$0D,$23,$FE,$50,$FA,$BA,$E8
	.dc.b	$FE,$BB,$1F,$E5,$28,$CD,$7F,$5E,$8B,$C3,$FE,$C3,$9B,$5F,$E6,$7F
	.dc.b	$0B,$FD,$98,$F2,$3B,$1F,$E6,$CF,$0D,$FA,$63,$D1,$FF,$3F,$C6,$1B
	.dc.b	$19,$9F,$F3,$2F,$C0,$62,$64,$73,$9D,$3F,$D0,$71,$FF,$6B,$57,$F9
	.dc.b	$DB,$F3,$17,$C8,$FF,$A1,$BF,$24,$F2,$DE,$C7,$86,$8C,$EF,$FC,$97
	.dc.b	$9D,$87,$2B,$FF,$3C,$B5,$FF,$19,$B6,$78,$37,$2F,$F5,$06,$0F,$C8
	.dc.b	$3F,$2F,$87,$4C,$BC,$C4,$C7,$FD,$13,$3A,$FD,$A3,$F5,$D1,$E7,$B7
	.dc.b	$E2,$1F,$8E,$8C,$37,$3F,$45,$70,$6F,$83,$FE,$98,$8C,$67,$FE,$16
	.dc.b	$6D,$FC,$2C,$79,$4E,$6B,$81,$27,$CB,$D3,$5F,$F2,$9D,$F2,$D0,$F4
	.dc.b	$CF,$E0,$57,$FA,$D3,$E3,$7F,$AE,$E7,$FE,$C3,$FB,$1F,$CB,$43,$E0
	.dc.b	$3F,$B5,$C6,$1B,$D7,$FD,$BA,$78,$E6,$AE,$3E,$7F,$F8,$06,$31,$5B
	.dc.b	$BE,$31,$FE,$D1,$BE,$CF,$86,$7A,$57,$1B,$E7,$86,$6B,$86,$1C,$96
	.dc.b	$1C,$9E,$E7,$5B,$E5,$87,$4B,$E5,$3C,$2F,$95,$F1,$C3,$BC,$F2,$3A
	.dc.b	$E1,$C9,$F0,$9C,$A7,$86,$1D,$63,$84,$7C,$4E,$7F,$F5,$84,$72,$C7
	.dc.b	$FD,$E1,$18,$9D,$7C,$47,$18,$DD,$7F,$BD,$7E,$33,$73,$FE,$08,$84
	.dc.b	$61,$FF,$05,$B5,$FF,$10,$B5,$E4,$7A,$C7,$AB,$FE,$0F,$CF,$F8,$A4
	.dc.b	$69,$B1,$9C,$BC,$15,$F2,$2B,$8F,$15,$7C,$67,$5A,$E0,$F9,$0C,$46
	.dc.b	$E7,$94,$EA,$7C,$E6,$8F,$31,$D2,$FC,$75,$C6,$35,$1E,$87,$57,$FE
	.dc.b	$2C,$AF,$0F,$0F,$E8,$3E,$0F,$FC,$59,$AF,$EC,$9B,$8F,$F0,$2E,$4F
	.dc.b	$F7,$E2,$C7,$F5,$33,$A5,$7C,$E7,$57,$CD,$FE,$0B,$E0,$EE,$75,$5C
	.dc.b	$A7,$67,$EF,$6F,$A7,$1F,$F1,$1B,$E1,$C8,$9A,$6B,$9B,$FC,$05,$7C
	.dc.b	$B0,$F1,$D6,$E3,$18,$7F,$12,$E6,$F1,$7E,$D0,$F8,$6A,$78,$36,$7D
	.dc.b	$96,$30,$E9,$B2,$F8,$46,$1B,$9F,$F1,$FB,$E3,$39,$C3,$94,$E7,$E3
	.dc.b	$6F,$07,$F8,$E2,$D1,$F2,$F9,$BF,$A7,$F8,$F7,$C0,$FF,$54,$E5,$EA
	.dc.b	$F1,$7A,$E5,$FD,$0E,$F7,$3F,$D3,$6F,$81,$F2,$AB,$79,$7F,$9A,$6B
	.dc.b	$77,$FE,$F4,$F0,$5D,$DF,$55,$B8,$F7,$CF,$13,$23,$E1,$CE,$AB,$F9
	.dc.b	$F6,$3E,$A9,$CD,$6E,$7B,$3E,$3B,$C6,$3F,$D5,$F0,$E4,$D9,$D5,$F2
	.dc.b	$3B,$F3,$79,$4F,$67,$15,$8C,$93,$1F,$1C,$65,$FF,$45,$74,$7E,$33
	.dc.b	$C7,$78,$B6,$30,$DB,$FF,$43,$36,$BF,$D2,$4F,$C5,$E5,$93,$A3,$E1
	.dc.b	$B8,$9F,$98,$F4,$8C,$9F,$95,$87,$CF,$7F,$AA,$23,$0D,$8F,$EF,$DD
	.dc.b	$6F,$96,$7A,$7F,$AF,$9E,$3E,$15,$BE,$1C,$1F,$E7,$96,$E3,$FE,$CB
	.dc.b	$C3,$8B,$DA,$B9,$7E,$18,$F8,$BA,$7E,$EF,$37,$D7,$BD,$8C,$E1,$E9
	.dc.b	$BE,$0D,$BE,$15,$B8,$C1,$EF,$F9,$07,$CF,$FA,$5B,$1F,$EE,$8F,$A8
	.dc.b	$7B,$39,$A9,$3C,$5F,$24,$EA,$30,$4E,$0F,$8F,$0F,$8B,$E3,$18,$7E
	.dc.b	$47,$D6,$7E,$DB,$8B,$FD,$3C,$DB,$FF,$78,$5B,$E7,$77,$E1,$C9,$FB
	.dc.b	$15,$BC,$7F,$A2,$DE,$E7,$35,$F8,$DE,$D9,$1F,$EC,$FE,$7B,$7F,$ED
	.dc.b	$B9,$B9,$7F,$9C,$66,$D7,$FB,$D3,$E9,$BF,$97,$DF,$F8,$FC,$38,$AF
	.dc.b	$F7,$17,$23,$27,$CE,$7E,$16,$B5,$87,$3E,$EC,$1F,$2A,$6C,$7E,$4E
	.dc.b	$77,$8F,$B3,$EB,$B2,$FC,$C7,$F5,$1E,$0C,$AF,$F7,$57,$81,$6B,$FD
	.dc.b	$ED,$37,$4F,$F4,$64,$DA,$FE,$91,$E4,$37,$7D,$2F,$F0,$F3,$F6,$9F
	.dc.b	$57,$6B,$F8,$6C,$5C,$BF,$C9,$5E,$1A,$B1,$97,$FB,$5F,$5F,$C8,$57
	.dc.b	$C7,$6C,$7F,$98,$FE,$ED,$D5,$F1,$7B,$A6,$7F,$05,$FE,$A9,$3A,$8C
	.dc.b	$13,$0F,$F0,$66,$1D,$9F,$EC,$47,$F6,$37,$0F,$D9,$AF,$F7,$53,$FD
	.dc.b	$9F,$91,$F8,$C3,$EA,$3F,$5C,$DC,$3F,$AC,$4E,$BE,$23,$35,$27,$35
	.dc.b	$BB,$FE,$0A,$AB,$51,$F1,$BE,$17,$F8,$F1,$FC,$7F,$EB,$AD,$FF,$62
	.dc.b	$E0,$FF,$7A,$36,$74,$D3,$F8,$B7,$E6,$6F,$A1,$FE,$20,$D8,$F3,$FD
	.dc.b	$73,$DA,$CD,$49,$E2,$FF,$0B,$9D,$46,$09,$97,$F1,$E8,$FA,$0F,$E7
	.dc.b	$EF,$B3,$9E,$2F,$E7,$1F,$5D,$FC,$E7,$E8,$2E,$1E,$B7,$FC,$19,$F8
	.dc.b	$43,$EB,$69,$B9,$FE,$F3,$19,$7E,$17,$D7,$7F,$43,$EA,$FE,$88,$D8
	.dc.b	$F8,$DC,$DD,$3E,$0B,$FA,$43,$63,$9C,$5D,$3F,$2B,$37,$AB,$FD,$D9
	.dc.b	$EF,$7F,$4F,$31,$7F,$EA,$CE,$4F,$F0,$A5,$FF,$AE,$FB,$9A,$7E,$EA
	.dc.b	$BC,$FE,$17,$F5,$7F,$93,$F6,$B2,$32,$7D,$B8,$C5,$F3,$AE,$4F,$FA
	.dc.b	$C7,$A1,$5A,$CE,$CF,$FB,$2A,$73,$3F,$E1,$EF,$BA,$7C,$9F,$F8,$DD
	.dc.b	$B1,$FE,$2A,$9F,$EB,$2E,$E7,$44,$75,$EA,$5B,$F7,$5F,$27,$C9,$DD
	.dc.b	$9C,$7C,$7E,$FB,$2F,$CF,$47,$FB,$6E,$DF,$47,$E3,$3F,$C8,$B3,$FA
	.dc.b	$1C,$BD,$11,$96,$9E,$7E,$FC,$65,$E2,$9E,$3D,$B8,$C7,$C3,$87,$E4
	.dc.b	$1F,$07,$C4,$FC,$13,$BF,$A3,$72,$F8,$D1,$71,$F9,$D3,$6B,$8D,$7C
	.dc.b	$7E,$AC,$3F,$EE,$CC,$3F,$25,$7F,$CF,$7E,$AD,$F1,$BF,$DF,$5B,$5F
	.dc.b	$3F,$EB,$F1,$32,$3F,$C1,$53,$AA,$FB,$D5,$97,$A5,$7C,$6F,$57,$19
	.dc.b	$7B,$9F,$F0,$57,$BF,$F4,$E3,$F0,$74,$8C,$63,$E7,$F5,$B7,$CD,$23
	.dc.b	$FC,$35,$3A,$AF,$C5,$56,$5F,$F2,$1F,$99,$FE,$85,$A7,$B5,$FC,$D1
	.dc.b	$FF,$45,$73,$77,$9D,$4E,$69,$39,$AF,$F3,$E1,$CB,$BF,$FC,$B9,$7C
	.dc.b	$1F,$CB,$7C,$74,$9F,$F3,$37,$95,$70,$FC,$8C,$DD,$7F,$1F,$A7,$F2
	.dc.b	$29,$CD,$FF,$8D,$77,$FF,$0A,$C7,$F7,$16,$D7,$0E,$2F,$F1,$94,$62
	.dc.b	$BF,$A1,$B6,$BF,$D4,$CD,$D3,$E6,$57,$0E,$49,$7D,$DF,$F3,$B4,$62
	.dc.b	$FF,$CF,$9F,$97,$7F,$83,$5F,$F6,$DF,$0F,$FB,$92,$31,$F1,$5F,$B6
	.dc.b	$9F,$8F,$A3,$C8,$FD,$EC,$E7,$E0,$D7,$C3,$0E,$5E,$DE,$9B,$A7,$FB
	.dc.b	$AD,$FA,$B8,$F3,$8C,$64,$C3,$9D,$F1,$8F,$F6,$27,$D5,$FF,$53,$F4
	.dc.b	$7F,$DE,$50,$8C,$3F,$EF,$56,$BF,$DA,$8D,$77,$FF,$BC,$1F,$F7,$D7
	.dc.b	$C6,$7F,$00,$FA,$CE,$6D,$8C,$F8,$29,$CB,$63,$B7,$83,$E7,$6E,$7E
	.dc.b	$4C,$7F,$AE,$36,$75,$3F,$DF,$21,$F5,$9F,$EB,$48,$F8,$31,$8F,$67
	.dc.b	$FC,$03,$C5,$F3,$6B,$55,$F5,$8F,$A7,$F2,$DA,$E2,$FF,$74,$61,$C6
	.dc.b	$79,$D7,$0B,$E5,$86,$E7,$89,$F4,$07,$D6,$7F,$A0,$9F,$9A,$D7,$E5
	.dc.b	$1A,$F8,$6F,$A2,$D3,$6F,$F8,$63,$AF,$F7,$F3,$C5,$B1,$9E,$4F,$C0
	.dc.b	$FC,$D5,$BF,$F8,$01,$7B,$9F,$F1,$0B,$63,$35,$BE,$BF,$E1,$57,$C7
	.dc.b	$CE,$7D,$5F,$F9,$06,$FF,$DD,$6D,$7F,$B1,$37,$2F,$DA,$FC,$5F,$18
	.dc.b	$C6,$7F,$E4,$A6,$D2,$FA,$FB,$DE,$1E,$BF,$F6,$DC,$FF,$B9,$FE,$07
	.dc.b	$D5,$83,$E8,$DF,$19,$86,$D6,$C8,$34,$FE,$CF,$5C,$6B,$81,$DE,$FE
	.dc.b	$35,$7F,$29,$E9,$FA,$38,$FF,$58,$8F,$D1,$B5,$FA,$ED,$7F,$44,$E7
	.dc.b	$0F,$E6,$6D,$C3,$FC,$9E,$D9,$FE,$4B,$F4,$DF,$E3,$17,$FD,$99,$F5
	.dc.b	$9F,$A9,$7D,$15,$FC,$BA,$7F,$CC,$30,$32,$72,$F7,$7D,$35,$C7,$0F
	.dc.b	$E8,$6F,$39,$FD,$B3,$5F,$F3,$0B,$5E,$8D,$4F,$3B,$DF,$81,$D8,$FC
	.dc.b	$36,$1F,$C6,$6B,$F0,$3E,$EE,$39,$64,$7E,$1A,$3F,$04,$7F,$0C,$E1
	.dc.b	$FC,$5B,$78,$8D,$24,$F1,$21,$F0,$39,$58,$3E,$23,$F1,$DE,$DF,$93
	.dc.b	$E2,$E5,$FE,$5A,$32,$FE,$BB,$2F,$CD,$C7,$84,$91,$9C,$91,$C2,$3F
	.dc.b	$3A,$FC,$0F,$96,$FD,$27,$D5,$7F,$9E,$67,$EA,$70,$FF,$88,$B3,$FE
	.dc.b	$CF,$18,$E0,$FB,$5E,$8F,$0F,$1E,$4F,$8F,$0E,$4F,$F2,$BD,$11,$88
	.dc.b	$FF,$4B,$C6,$27,$E5,$4F,$0C,$7F,$C8,$73,$9A,$F0,$1F,$AD,$9D,$61
	.dc.b	$F8,$BE,$6F,$4F,$A5,$F6,$BE,$B2,$5F,$0C,$3C,$33,$A8,$F3,$3E,$02
	.dc.b	$38,$38,$61,$FE,$F9,$6C,$7E,$1B,$EB,$3F,$0F,$D9,$27,$CB,$C9,$F6
	.dc.b	$A3,$E5,$FD,$64,$9F,$88,$FB,$88,$FF,$26,$9F,$C0,$FB,$5B,$8F,$F3
	.dc.b	$96,$2A,$FC,$AC,$FE,$5B,$E5,$3F,$2F,$F5,$92,$FC,$3C,$69,$FE,$3D
	.dc.b	$F3,$DF,$D2,$A7,$E7,$B5,$FF,$70,$B5,$FE,$30,$DC,$BF,$27,$F5,$7F
	.dc.b	$A4,$FA,$C9,$5D,$A3,$F8,$6F,$D4,$FD,$28,$FD,$D7,$D5,$78,$5F,$AE
	.dc.b	$1C,$3F,$BE,$11,$C5,$BC,$FF,$D1,$A3,$F0,$D1,$FE,$DC,$6B,$FC,$1D
	.dc.b	$AF,$03,$BB,$CC,$E9,$3F,$CB,$63,$E2,$FB,$3F,$BE,$23,$F2,$3E,$B2
	.dc.b	$BF,$B3,$D7,$FC,$C2,$F0,$34,$7F,$B9,$3E,$B1,$3F,$D4,$FC,$34,$62
	.dc.b	$75,$39,$87,$B9,$62,$4E,$8E,$91,$F3,$3E,$B3,$F9,$CE,$5F,$C9,$A7
	.dc.b	$E4,$37,$B7,$FC,$CB,$A2,$B8,$57,$F8,$7A,$30,$D8,$F9,$84,$7E,$3B
	.dc.b	$C1,$EB,$E9,$7E,$23,$BF,$FB,$47,$B8,$FB,$0A,$66,$BF,$59,$CB,$FC
	.dc.b	$11,$FF,$64,$37,$2F,$DC,$B7,$C1,$FE,$81,$AF,$F7,$3F,$8C,$FF,$17
	.dc.b	$57,$FA,$B9,$B3,$FE,$D3,$BD,$70,$ED,$87,$EB,$A7,$6F,$F9,$D1,$FE
	.dc.b	$78,$FF,$4B,$E4,$FE,$75,$7F,$E8,$AF,$AD,$7C,$67,$F6,$46,$D7,$E5
	.dc.b	$AB,$78,$FF,$53,$84,$7F,$43,$7F,$C8,$8D,$74,$D7,$C1,$EC,$7F,$87
	.dc.b	$A3,$9C,$60,$ED,$1E,$1F,$83,$F8,$5E,$D4,$9F,$63,$F1,$7D,$6C,$47
	.dc.b	$F4,$5D,$13,$8D,$3E,$45,$7F,$7E,$FA,$D3,$FE,$00,$F8,$67,$9D,$4D
	.dc.b	$44,$C7,$F6,$B5,$C1,$B1,$C7,$D1,$FF,$04,$FD,$6B,$E3,$3F,$C1,$7F
	.dc.b	$5A,$F8,$8A,$6A,$26,$1F,$E6,$EA,$E0,$D8,$FF,$20,$C6,$23,$FC,$8A
	.dc.b	$DD,$3F,$D7,$ED,$E2,$FA,$4D,$9F,$F0,$36,$0F,$9A,$FF,$0E,$FB,$BF
	.dc.b	$12,$7F,$87,$FD,$6B,$DC,$FF,$10,$B6,$3F,$89,$CE,$DF,$E2,$4D,$CB
	.dc.b	$FD,$9E,$DC,$BF,$04,$D9,$FF,$0B,$57,$C7,$7F,$8D,$7D,$6F,$E4,$97
	.dc.b	$D6,$FE,$D6,$FF,$14,$D8,$FF,$0F,$4E,$C8,$FF,$8D,$9E,$51,$FE,$80
	.dc.b	$C3,$F5,$3E,$67,$F8,$37,$BB,$F9,$4D,$FF,$3A,$8C,$11,$FD,$9B,$FD
	.dc.b	$00,$FD,$76,$5F,$AF,$CE,$CF,$EA,$63,$0F,$EA,$E3,$2F,$E3,$E3,$2F
	.dc.b	$EF,$3C,$9F,$11,$1F,$F0,$85,$FF,$26,$7F,$BF,$37,$0F,$E0,$33,$FD
	.dc.b	$48,$FF,$31,$7B,$FF,$D5,$3C,$0F,$61,$2F,$9B,$5F,$E1,$6E,$DF,$ED
	.dc.b	$BE,$FB,$E9,$71,$1A,$3F,$89,$CF,$F9,$BE,$33,$87,$F1,$28,$BB,$FF
	.dc.b	$B2,$A7,$FD,$15,$E3,$3E,$99,$E9,$7F,$B1,$A3,$0D,$8C,$E3,$FE,$9C
	.dc.b	$9C,$B6,$32,$FF,$96,$A3,$2D,$8C,$E3,$FE,$FE,$F8,$2F,$F4,$97,$D6
	.dc.b	$3E,$63,$FD,$2D,$F5,$BF,$D1,$B0,$FF,$05,$B5,$FF,$32,$FB,$8A,$E1
	.dc.b	$1F,$E1,$D6,$BF,$E2,$F6,$BF,$E2,$E6,$BF,$E6,$F8,$E5,$5F,$E5,$99
	.dc.b	$E3,$87,$F8,$C1,$BA,$ED,$B6,$F5,$9F,$53,$FC,$CB,$17,$D5,$FE,$D6
	.dc.b	$BF,$C4,$7F,$59,$FE,$A0,$7F,$CB,$CD,$CB,$FC,$7B,$3B,$9F,$6B,$5F
	.dc.b	$F2,$2B,$5F,$43,$5F,$F5,$73,$6B,$C3,$DB,$FE,$BD,$8E,$95,$DA,$3F
	.dc.b	$D8,$58,$78,$73,$E7,$1B,$26,$1F,$E3,$A9,$D3,$FE,$95,$F9,$07,$A5
	.dc.b	$FE,$B1,$87,$D6,$FF,$AD,$94,$BE,$13,$FA,$54,$E6,$31,$3F,$E9,$F8
	.dc.b	$C5,$73,$6E,$5F,$EA,$3F,$A2,$E1,$87,$F5,$36,$DE,$7A,$BF,$AB,$63
	.dc.b	$E4,$BC,$4F,$9F,$1E,$95,$AC,$3C,$6D,$9F,$13,$CA,$77,$1C,$6B,$9B
	.dc.b	$EC,$AF,$15,$71,$86,$1E,$86,$C7,$3A,$E3,$87,$C0,$9C,$9F,$2E,$F9
	.dc.b	$BD,$CC,$9F,$2B,$0F,$9D,$CD,$FE,$CB,$8C,$5F,$FB,$1E,$31,$7F,$06
	.dc.b	$51,$8A,$E3,$8F,$F7,$B9,$FF,$67,$7C,$77,$18,$C4,$23,$0E,$DA,$FF
	.dc.b	$80,$DA,$F2,$3C,$E3,$D4,$6F,$F2,$23,$55,$F8,$57,$51,$97,$C0,$72
	.dc.b	$9E,$2F,$43,$33,$C6,$32,$F3,$8E,$3A,$E6,$BE,$DC,$7A,$4F,$CA,$EB
	.dc.b	$F0,$6F,$AD,$71,$C3,$CD,$1D,$27,$E1,$C6,$2F,$CC,$F4,$BF,$F7,$B5
	.dc.b	$72,$AD,$C7,$63,$FC,$51,$19,$7F,$6B,$F0,$FE,$CD,$7F,$7E,$DC,$3F
	.dc.b	$E0,$07,$27,$F0,$AF,$07,$17,$C2,$7F,$D7,$1C,$7F,$ED,$86,$DE,$AF
	.dc.b	$84,$60,$95,$FE,$CF,$AE,$0D,$AF,$9A,$FF,$0C,$6B,$FC,$31,$E9,$CB
	.dc.b	$FE,$0D,$AE,$78,$FD,$4C,$13,$9B,$FE,$67,$F5,$BF,$EF,$7B,$FF,$67
	.dc.b	$36,$3F,$DF,$9C,$8F,$A9,$F6,$F3,$AB,$59,$7F,$B7,$9E,$17,$FE,$E4
	.dc.b	$6B,$FC,$DD,$AA,$79,$7C,$5F,$AA,$31,$3D,$31,$F9,$B8,$BE,$43,$C3
	.dc.b	$39,$AF,$E7,$53,$9A,$FF,$76,$F3,$7F,$63,$F1,$BE,$9B,$63,$FB,$8E
	.dc.b	$98,$EF,$FF,$67,$47,$06,$E1,$FA,$FA,$E0,$D8,$CB,$FE,$6D,$8C,$BF
	.dc.b	$E6,$FD,$1D,$EF,$1B,$E3,$B7,$0F,$F0,$0E,$B8,$99,$1F,$EF,$D9,$D5
	.dc.b	$B7,$FB,$1E,$27,$B6,$3F,$B4,$CE,$A7,$F5,$F8,$F8,$30,$E7,$5C,$70
	.dc.b	$FC,$AE,$1C,$6F,$78,$71,$9F,$ED,$4F,$82,$BD,$4D,$73,$7F,$57,$0E
	.dc.b	$D5,$FC,$41,$F7,$F2,$FB,$8F,$CE,$63,$F8,$DC,$7B,$C6,$6B,$86,$3F
	.dc.b	$E5,$D6,$C7,$58,$C6,$3F,$C7,$F8,$DE,$A3,$A3,$F5,$0F,$F8,$A7,$0F
	.dc.b	$F5,$54,$63,$C4,$CB,$63,$27,$FA,$BE,$3E,$9F,$0C,$8C,$1F,$45,$C6
	.dc.b	$1F,$CB,$F5,$FB,$71,$E5,$24,$FF,$30,$97,$CF,$7F,$6C,$84,$7D,$12
	.dc.b	$C6,$5C,$75,$3E,$B5,$E8,$FF,$08,$9E,$7E,$05,$BE,$1C,$3F,$CB,$30
	.dc.b	$E1,$D5,$9A,$FE,$1F,$18,$CB,$C3,$5F,$C1,$BE,$B7,$FD,$53,$18,$3F
	.dc.b	$D8,$8E,$31,$F1,$FD,$6B,$D9,$F1,$4E,$6F,$97,$AF,$DA,$BF,$89,$C6
	.dc.b	$30,$E9,$87,$F4,$99,$CC,$7F,$B4,$7E,$4D,$F3,$9E,$D6,$C7,$0C,$DF
	.dc.b	$58,$F6,$F5,$72,$BE,$75,$F0,$C7,$FA,$FB,$5F,$EC,$8D,$7C,$13,$F1
	.dc.b	$2D,$F2,$1C,$E7,$E8,$C6,$2F,$D3,$0F,$AD,$FE,$18,$DE,$1E,$5F,$39
	.dc.b	$F5,$67,$27,$F1,$1D,$3D,$3F,$1F,$9A,$F0,$3F,$96,$ED,$98,$FF,$BF
	.dc.b	$FA,$2E,$5F,$EE,$39,$BC,$3F,$CD,$7B,$3F,$97,$5F,$CC,$ED,$FD,$0D
	.dc.b	$7E,$17,$EA,$FF,$4B,$7F,$D0,$9B,$5F,$D7,$37,$7D,$36,$FC,$09,$3F
	.dc.b	$97,$AF,$D6,$E7,$FE,$0B,$6B,$F9,$80,$E1,$FC,$DB,$D9,$FF,$03,$5B
	.dc.b	$EB,$7A,$C7,$CC,$7F,$6B,$E2,$4B,$F5,$6B,$FC,$0D,$FE,$CE,$FF,$13
	.dc.b	$6C,$6E,$BF,$51,$AF,$F0,$0F,$4B,$1A,$FF,$18,$FA,$9F,$E5,$67,$F0
	.dc.b	$2F,$A3,$B8,$FF,$29,$8F,$ED,$9F,$17,$FE,$93,$8C,$1F,$2D,$C6,$3F
	.dc.b	$B5,$3D,$FD,$FF,$34,$FE,$BB,$F5,$AE,$CC,$5F,$E1,$B8,$7F,$4C,$FC
	.dc.b	$BA,$FE,$B1,$7C,$A3,$04,$7F,$81,$66,$FB,$1F,$AB,$3C,$EB,$FC,$0F
	.dc.b	$39,$E4,$FD,$DF,$D7,$3E,$16,$98,$F8,$BB,$87,$EA,$67,$4D,$9D,$34
	.dc.b	$FA,$3F,$0F,$FC,$A0,$FE,$53,$BD,$8C,$93,$4F,$A3,$CA,$9D,$47,$C7
	.dc.b	$FA,$D7,$D1,$FF,$17,$6D,$FE,$EF,$EB,$BF,$CD,$6F,$EB,$23,$35,$C6
	.dc.b	$BE,$3C,$62,$3F,$CB,$8D,$7F,$24,$D7,$F9,$5B,$CB,$6F,$E4,$57,$F5
	.dc.b	$6F,$B5,$FF,$22,$8C,$5F,$AB,$EA,$3F,$A3,$FD,$17,$F4,$86,$DF,$F1
	.dc.b	$8F,$3D,$F5,$AF,$AC,$FE,$BC,$D9,$FE,$BC,$DC,$3F,$A2,$35,$FD,$47
	.dc.b	$67,$F5,$6F,$AE,$FE,$AB,$D7,$FD,$59,$B7,$FA,$2E,$7F,$F1,$4E,$1F
	.dc.b	$92,$DF,$F8,$04,$63,$E1,$65,$78,$1F,$19,$1F,$F3,$FE,$3F,$D9,$3D
	.dc.b	$EF,$EC,$67,$0C,$FE,$CB,$DB,$A6,$57,$EC,$BE,$56,$D7,$FB,$CB,$3F
	.dc.b	$C9,$BE,$CA,$F8,$24,$AF,$82,$FE,$B1,$E3,$9F,$F9,$2F,$97,$FD,$0F
	.dc.b	$1F,$2F,$EB,$9F,$2B,$FE,$46,$6C,$7E,$C7,$D5,$C4,$C8,$FF,$05,$4E
	.dc.b	$B1,$FE,$E0,$D7,$FC,$58,$D7,$FE,$00,$3B,$C3,$C5,$D7,$CE,$7F,$84
	.dc.b	$E3,$FA,$47,$FB,$97,$22,$DF,$4B,$E5,$3F,$83,$F7,$5C,$BD,$51,$FE
	.dc.b	$1F,$9D,$C7,$9A,$30,$47,$F8,$67,$57,$1C,$3C,$2F,$C1,$F8,$B9,$3F
	.dc.b	$C4,$7D,$2B,$5E,$4C,$15,$DC,$95,$F1,$1B,$3E,$CE,$9E,$3B,$7D,$17
	.dc.b	$FC,$11,$F5,$CF,$8E,$A7,$F2,$2F,$AE,$B8,$FF,$5E,$9D,$3F,$E1,$96
	.dc.b	$E1,$FE,$CA,$6B,$FE,$CF,$6B,$FD,$15,$DE,$1F,$81,$F5,$BA,$CF,$D0
	.dc.b	$7F,$C3,$5E,$FF,$D5,$BF,$E9,$8D,$97,$57,$F0,$DF,$5F,$7C,$D2,$FF
	.dc.b	$C5,$97,$F7,$EF,$FC,$56,$D4,$E6,$EF,$97,$FE,$B0,$A7,$B7,$FC,$DA
	.dc.b	$FF,$A5,$DA,$B5,$39,$CF,$FC,$DB,$18,$EF,$FF,$2C,$B7,$0F,$CE,$7C
	.dc.b	$7F,$E6,$CF,$F2,$EF,$91,$63,$24,$CB,$BC,$65,$FE,$9D,$C3,$FE,$19
	.dc.b	$C7,$FD,$72,$DA,$FF,$3A,$F5,$7F,$90,$63,$15,$FE,$4A,$6B,$F7,$79
	.dc.b	$BF,$CD,$B3,$B6,$57,$F9,$4E,$BF,$C4,$55,$C6,$7F,$AA,$C6,$3E,$09
	.dc.b	$F5,$CF,$6F,$FD,$EF,$F1,$5F,$B8,$C1,$5A,$9C,$F1,$FF,$A2,$BC,$8B
	.dc.b	$5A,$C3,$96,$7F,$E6,$0C,$79,$63,$C6,$31,$87,$F9,$92,$31,$6A,$F0
	.dc.b	$72,$7D,$CC,$7A,$D6,$A7,$C9,$4F,$AD,$77,$BD,$CF,$F3,$94,$61,$B1
	.dc.b	$99,$FF,$0E,$1F,$E7,$49,$D3,$81,$F1,$DF,$C1,$3E,$BB,$FC,$F8,$D8
	.dc.b	$D3,$5E,$6D,$7F,$C5,$13,$FD,$93,$1F,$1F,$83,$F2,$A7,$B4,$E5,$B1
	.dc.b	$97,$FD,$95,$F5,$DF,$ED,$D7,$E3,$9F,$A6,$EB,$79,$DF,$E3,$5D,$7E
	.dc.b	$3C,$62,$77,$8A,$F9,$47,$CD,$E0,$FA,$11,$8B,$FA,$AD,$8C,$FB,$1D
	.dc.b	$5F,$F6,$3F,$CE,$5B,$FF,$4D,$FD,$77,$FA,$59,$FF,$2A,$64,$F4,$3F
	.dc.b	$C9,$1D,$7F,$ED,$66,$F3,$7F,$92,$EF,$85,$FF,$8B,$DB,$5B,$8E,$75
	.dc.b	$C1,$EF,$F0,$D5,$57,$0B,$E9,$39,$8C,$4F,$C7,$F9,$CE,$5E,$5B,$EB
	.dc.b	$7F,$DE,$73,$FD,$FD,$AF,$F4,$46,$E5,$FE,$EE,$F8,$CE,$31,$8C,$BF
	.dc.b	$D7,$77,$CF,$E0,$B2,$79,$FB,$3F,$89,$DD,$D9,$F0,$5E,$BC,$1F,$57
	.dc.b	$FE,$D9,$D3,$FD,$4F,$18,$F8,$DE,$15,$27,$35,$B7,$FD,$95,$5B,$7E
	.dc.b	$8C,$FC,$5B,$FE,$B7,$1F,$E5,$DF,$B0,$F2,$78,$A7,$5E,$94,$9D,$10
	.dc.b	$FF,$31,$57,$27,$FC,$C9,$F5,$FF,$A9,$7F,$CE,$7F,$53,$FE,$63,$7F
	.dc.b	$D2,$8D,$77,$E1,$FD,$79,$E1,$C6,$D3,$FD,$C7,$C0,$C3,$84,$E7,$0F
	.dc.b	$E8,$71,$8A,$FF,$49,$5F,$26,$E1,$D4,$95,$E2,$E9,$FA,$FD,$DF,$4F
	.dc.b	$0F,$C1,$E5,$DF,$A3,$37,$F8,$1A,$D4,$48,$C1,$0F,$C2,$47,$07,$E0
	.dc.b	$FC,$57,$E1,$3D,$BE,$DF,$1B,$27,$F1,$91,$97,$E2,$C6,$5F,$C7,$46
	.dc.b	$5A,$70,$3E,$5F,$CC,$FE,$63,$EB,$7F,$8E,$C7,$CE,$F4,$BD,$86,$7C
	.dc.b	$99,$19,$21,$FE,$1E,$9E,$2D,$8D,$B4,$FB,$3C,$9D,$FD,$3F,$91,$F5
	.dc.b	$F2,$7F,$CC,$97,$FE,$E8,$6C,$E5,$FC,$1C,$F1,$72,$43,$C5,$1F,$4F
	.dc.b	$EB,$FE,$A7,$5C,$9F,$C4,$BE,$FF,$8F,$91,$82,$1E,$B7,$EE,$7B,$7A
	.dc.b	$8C,$1F,$92,$7F,$01,$F2,$DF,$81,$FA,$F9,$3F,$96,$8F,$E0,$7F,$41
	.dc.b	$FC,$BF,$B3,$F2,$1C,$FF,$93,$9E,$35,$FE,$83,$AD,$75,$FE,$63,$A5
	.dc.b	$F5,$DF,$8D,$FA,$F9,$7F,$A2,$8F,$E2,$DF,$59,$F4,$33,$FC,$E4,$BE
	.dc.b	$B5,$9F,$F1,$B6,$C0,$FF,$69,$57,$1C,$BF,$DE,$4D,$8D,$4F,$07,$FD
	.dc.b	$B0,$DB,$EF,$3B,$93,$67,$55,$FE,$FD,$6C,$F8,$67,$B6,$9F,$B5,$F6
	.dc.b	$7F,$54,$49,$FD,$FF,$D7,$57,$F8,$1C,$5C,$7E,$5E,$AA,$FE,$75,$E0
	.dc.b	$BE,$BF,$18,$7B,$EF,$F8,$04,$E6,$B7,$5A,$A5,$6A,$1E,$F6,$AA,$57
	.dc.b	$3D,$7E,$56,$1C,$E3,$07,$F8,$16,$F8,$B8,$3E,$2F,$D7,$FD,$3F,$94
	.dc.b	$91,$9D,$93,$F7,$0C,$C7,$D0,$9C,$3B,$8F,$F0,$53,$F2,$DA,$F6,$33
	.dc.b	$3D,$1F,$F0,$87,$9C,$E4,$FF,$2B,$C7,$A3,$83,$FD,$E1,$F5,$FF,$CE
	.dc.b	$8F,$CC,$F8,$12,$75,$E5,$A7,$51,$F9,$99,$FE,$A9,$F2,$5F,$E2,$0E
	.dc.b	$2F,$CF,$B7,$83,$F9,$9B,$6B,$FC,$E6,$FF,$2F,$69,$C8,$87,$95,$D3
	.dc.b	$FE,$32,$F2,$7F,$BB,$7D,$7B,$E3,$E9,$FD,$E7,$C5,$FF,$1C,$E3,$FA
	.dc.b	$46,$E3,$FA,$1C,$17,$C5,$FA,$CD,$7F,$07,$1B,$9E,$31,$FA,$96,$BC
	.dc.b	$AF,$C2,$FD,$F8,$DE,$AC,$9E,$BF,$C6,$25,$7B,$7D,$5A,$F8,$0F,$8B
	.dc.b	$18,$CF,$E5,$CF,$F8,$57,$EB,$CF,$F8,$5F,$E2,$1E,$65,$35,$13,$0F
	.dc.b	$C9,$4E,$9B,$39,$7C,$03,$EB,$DF,$1D,$FE,$23,$FA,$F7,$C4,$D3,$FC
	.dc.b	$4D,$3A,$AF,$CF,$35,$E7,$3C,$1F,$D1,$35,$EF,$1B,$CF,$F8,$CB,$78
	.dc.b	$7E,$B5,$6E,$7F,$8E,$B6,$3E,$D4,$EE,$BF,$8E,$7B,$3F,$39,$FD,$87
	.dc.b	$D7,$FF,$89,$1B,$EC,$FF,$8F,$67,$51,$F0,$1B,$8F,$F2,$AA,$E1,$3F
	.dc.b	$11,$B1,$FC,$BA,$77,$87,$E2,$FE,$BF,$F3,$96,$FA,$FF,$F0,$63,$FC
	.dc.b	$BD,$AA,$36,$70,$8C,$9A,$7F,$4D,$DB,$FB,$7F,$23,$FC,$5B,$F1,$72
	.dc.b	$FF,$A4,$47,$F4,$E3,$0F,$E2,$79,$DD,$DF,$EA,$F8,$D9,$FA,$BD,$BF
	.dc.b	$6F,$A4,$8C,$10,$E5,$D0,$8D,$46,$27,$FA,$EF,$91,$FE,$51,$F9,$79
	.dc.b	$3F,$D6,$E3,$2F,$F9,$E2,$32,$FE,$77,$4F,$E2,$53,$FD,$9B,$C8,$FF
	.dc.b	$2D,$7C,$7C,$AF,$EC,$91,$97,$FD,$03,$19,$79,$78,$7F,$CB,$9B,$3B
	.dc.b	$69,$FE,$88,$FA,$6C,$72,$7F,$90,$71,$FE,$45,$58,$E4,$FF,$23,$FB
	.dc.b	$CB,$FF,$42,$7C,$43,$E9,$DE,$9F,$F9,$BF,$DC,$9A,$7F,$40,$AD,$4F
	.dc.b	$0A,$DC,$FF,$A6,$2B,$6F,$EF,$6B,$6F,$F4,$8C,$DF,$5C,$F9,$AF,$F3
	.dc.b	$F7,$D7,$FB,$77,$FF,$04,$B7,$6F,$F4,$2D,$6F,$0E,$4D,$7F,$05,$3B
	.dc.b	$5D,$08,$C6,$3E,$0C,$7F,$D3,$37,$C7,$1E,$FC,$1E,$23,$33,$FE,$9E
	.dc.b	$FA,$D7,$D1,$FF,$A4,$FE,$B7,$F0,$DA,$7F,$4B,$6B,$F7,$E7,$6F,$BD
	.dc.b	$AF,$E5,$E3,$78,$7F,$89,$9B,$AF,$FA,$8A,$B7,$3F,$65,$B1,$9B,$F2
	.dc.b	$1F,$EA,$5F,$AC,$E3,$83,$45,$F0,$8C,$36,$33,$5C,$7A,$BF,$D8,$5F
	.dc.b	$26,$F4,$3F,$C0,$B8,$3E,$BB,$FD,$4D,$2F,$85,$FE,$99,$18,$6F,$C5
	.dc.b	$FF,$4F,$9E,$5B,$7F,$8C,$6B,$8E,$1D,$1E,$18,$66,$F8,$61,$CA,$3F
	.dc.b	$C7,$18,$72,$7C,$53,$D6,$F9,$3C,$23,$91,$C2,$B9,$2A,$E5,$7E,$2C
	.dc.b	$37,$39,$AD,$D7,$B6,$32,$F0,$BF,$05,$62,$B6,$70,$C3,$C7,$7E,$87
	.dc.b	$E4,$F2,$BE,$BF,$8C,$A3,$0E,$EF,$FC,$AD,$18,$8F,$45,$72,$AF,$EF
	.dc.b	$51,$FE,$5B,$F9,$3D,$C6,$21,$18,$76,$D7,$FC,$04,$D7,$91,$F3,$0F
	.dc.b	$F0,$37,$B7,$FE,$C7,$D3,$FD,$C9,$1A,$6C,$67,$3F,$05,$7C,$AA,$4E
	.dc.b	$5F,$09,$D4,$FF,$24,$F1,$6D,$F3,$99,$38,$B9,$9E,$93,$CE,$7D,$26
	.dc.b	$DC,$3E,$5A,$F8,$B1,$88,$EB,$7E,$8B,$E9,$1C,$5F,$C3,$F0,$7E,$9B
	.dc.b	$A7,$FD,$CD,$AF,$EB,$DB,$3F,$EE,$CE,$37,$85,$8C,$3F,$57,$57,$0E
	.dc.b	$57,$FB,$F8,$D7,$53,$0E,$33,$92,$5F,$F0,$3E,$4E,$5C,$AC,$38,$F3
	.dc.b	$7F,$10,$BD,$E9,$FC,$1E,$F7,$5A,$7C,$5F,$5E,$F8,$AF,$0E,$6F,$93
	.dc.b	$CC,$61,$B8,$FF,$98,$E7,$4F,$F3,$5A,$E4,$E6,$77,$A7,$F6,$FA,$DB
	.dc.b	$FD,$76,$B6,$E6,$7E,$37,$37,$F4,$BF,$23,$E1,$F8,$5F,$E5,$9C,$7F
	.dc.b	$A5,$78,$1F,$EB,$87,$FC,$E9,$19,$C3,$FA,$44,$E5,$FE,$FD,$19,$D9
	.dc.b	$5F,$A8,$AD,$47,$FB,$42,$75,$3F,$EC,$1E,$27,$C8,$7F,$7D,$8F,$EA
	.dc.b	$4D,$C3,$AE,$7F,$6E,$31,$87,$F6,$96,$D7,$36,$D7,$F4,$EC,$3E,$1D
	.dc.b	$6F,$B7,$91,$F8,$2F,$1F,$CF,$87,$82,$30,$E8,$FF,$4A,$46,$AF,$FC
	.dc.b	$11,$87,$6E,$54,$E7,$0F,$27,$37,$5F,$A0,$F4,$37,$1F,$6B,$71,$E6
	.dc.b	$DA,$FF,$4B,$F5,$7E,$5B,$C9,$FF,$01,$5F,$FA,$46,$31,$E3,$65,$B1
	.dc.b	$93,$FD,$2D,$1F,$1F,$EA,$FF,$3F,$E5,$BE,$23,$15,$FE,$AA,$6E,$7F
	.dc.b	$E3,$79,$D4,$FD,$2A,$E3,$5F,$AD,$A6,$9F,$D4,$EB,$FD,$9F,$DE,$FA
	.dc.b	$FF,$C7,$EB,$F8,$2B,$FF,$7F,$5E,$DF,$F4,$65,$6F,$AF,$A4,$E6,$BF
	.dc.b	$DC,$B7,$CB,$0F,$F5,$34,$E5,$FF,$6F,$C6,$70,$0F,$AF,$FF,$48,$CF
	.dc.b	$93,$EB,$FF,$70,$FF,$B8,$63,$33,$E2,$9C,$BF,$E9,$08,$CE,$1F,$EF
	.dc.b	$69,$CB,$EB,$9E,$DE,$C7,$F0,$FC,$D5,$AC,$3F,$D7,$9F,$08,$F9,$FD
	.dc.b	$7D,$73,$E4,$3E,$54,$E6,$BE,$4C,$E5,$FF,$7B,$46,$72,$F8,$F3,$97
	.dc.b	$FD,$4B,$19,$CF,$E7,$64,$AD,$47,$9E,$75,$7F,$1A,$B5,$87,$8B,$1F
	.dc.b	$2B,$F7,$7E,$BF,$E6,$4F,$3E,$1C,$70,$FA,$9B,$87,$A1,$BA,$79,$9F
	.dc.b	$CC,$75,$49,$FF,$74,$71,$DA,$FF,$3D,$CD,$D7,$FD,$EB,$36,$BF,$DC
	.dc.b	$9F,$5B,$FC,$D2,$FE,$0F,$D7,$FE,$46,$31,$7F,$03,$81,$39,$EA,$F9
	.dc.b	$ED,$C3,$E6,$B7,$0E,$71,$F8,$8B,$FB,$98,$74,$8F,$8D,$87,$47,$FE
	.dc.b	$00,$77,$FE,$CC,$FD,$A7,$8D,$63,$24,$C3,$FD,$2F,$19,$7F,$83,$F6
	.dc.b	$FE,$CB,$D0,$FE,$34,$DA,$FE,$3E,$D9,$DC,$79,$4F,$CE,$F7,$FF,$AA
	.dc.b	$AB,$E1,$7A,$DF,$93,$AF,$8E,$D7,$F4,$79,$A3,$10,$8C,$5F,$F8,$56
	.dc.b	$31,$3F,$DA,$74,$7D,$73,$EA,$BB,$DF,$8E,$D5,$A9,$CC,$62,$7E,$65
	.dc.b	$71,$78,$47,$67,$95,$71,$7F,$84,$FD,$77,$FB,$F9,$BF,$05,$FA,$D8
	.dc.b	$7C,$3D,$89,$3A,$38,$C6,$2F,$FD,$8F,$F0,$7F,$CA,$4F,$E5,$5E,$1D
	.dc.b	$AF,$56,$EF,$19,$59,$BF,$F6,$7D,$DC,$FF,$80,$7D,$73,$E8,$FF,$80
	.dc.b	$7D,$73,$C2,$FE,$49,$39,$D3,$F9,$1C,$E5,$F8,$31,$9C,$7F,$90,$CE
	.dc.b	$5F,$90,$7F,$1A,$8D,$61,$FB,$C9,$D3,$E2,$C3,$C3,$87,$81,$B8,$7B
	.dc.b	$33,$7D,$3F,$F4,$9F,$9E,$FE,$94,$D9,$FF,$5F,$7D,$73,$EB,$7F,$A7
	.dc.b	$7D,$37,$F8,$9E,$31,$3C,$E3,$2F,$5E,$AB,$BF,$FB,$D1,$FE,$AD,$D1
	.dc.b	$C7,$AA,$D7,$67,$0F,$97,$D7,$B7,$FE,$FC,$7F,$AF,$12,$38,$1B,$3C
	.dc.b	$FD,$5E,$53,$FB,$27,$C8,$58,$C9,$F7,$39,$7F,$B9,$3F,$E0,$DD,$3F
	.dc.b	$25,$87,$F5,$D8,$E0,$D7,$8B,$4F,$A7,$E9,$BE,$B9,$F2,$7F,$E4,$B6
	.dc.b	$D7,$7F,$47,$13,$23,$C5,$DB,$70,$FE,$D8,$DE,$8F,$EC,$9D,$8D,$3F
	.dc.b	$41,$5E,$0C,$BA,$BC,$30,$FF,$7A,$43,$E8,$BC,$47,$F7,$CE,$EB,$3A
	.dc.b	$8F,$EF,$FD,$58,$BF,$EF,$8D,$CB,$C3,$8F,$E3,$AF,$FC,$65,$8F,$57
	.dc.b	$E5,$57,$BF,$E1,$E4,$7C,$DF,$3E,$E3,$E7,$9F,$F0,$6F,$4E,$32,$EA
	.dc.b	$DB,$FE,$33,$97,$E3,$2F,$84,$7D,$79,$FF,$06,$7D,$6B,$E2,$A9,$FE
	.dc.b	$18,$FA,$EB,$5F,$C2,$A7,$4F,$DF,$F9,$2C,$E1,$FE,$19,$6E,$5E,$E6
	.dc.b	$CF,$89,$FF,$0D,$7B,$7F,$3A,$7E,$1F,$B1,$8E,$5F,$F1,$3B,$6F,$F1
	.dc.b	$8D,$C3,$FC,$4C,$FF,$9A,$3E,$BB,$FD,$21,$3F,$EF,$5E,$EC,$3F,$C9
	.dc.b	$CC,$1F,$E6,$A8,$DF,$C3,$FF,$32,$BF,$F3,$4C,$62,$BF,$1D,$F1,$52
	.dc.b	$FF,$CE,$3E,$4D,$BF,$C3,$4D,$CF,$FC,$D9,$36,$BF,$C4,$75,$F2,$F9
	.dc.b	$3F,$C7,$1C,$3F,$E0,$28,$FF,$32,$71,$E2,$65,$7F,$97,$BA,$AD,$7F
	.dc.b	$14,$EA,$FF,$2E,$C6,$32,$FC,$CF,$85,$F9,$87,$FC,$49,$97,$5A,$DE
	.dc.b	$1F,$D1,$FE,$05,$F5,$CF,$8F,$FF,$22,$7C,$4F,$EB,$34,$FF,$78,$4E
	.dc.b	$6F,$FD,$D5,$CB,$FD,$83,$9B,$F8,$DC,$62,$B7,$87,$2A,$FF,$7A,$DF
	.dc.b	$19,$FF,$77,$C7,$F6,$8D,$7F,$A4,$DF,$1C,$BC,$78,$71,$B6,$7F,$E7
	.dc.b	$AE,$FC,$43,$EB,$7F,$A8,$FB,$7F,$D9,$DB,$1F,$EF,$D6,$BF,$40,$E1
	.dc.b	$3F,$EF,$07,$03,$E3,$7F,$75,$F5,$DF,$DC,$5B,$1A,$6B,$CD,$AF,$DC
	.dc.b	$AF,$EA,$59,$78,$FC,$13,$49,$3E,$3C,$BF,$D2,$B3,$F8,$4E,$1F,$A1
	.dc.b	$18,$9F,$98,$D7,$E4,$9F,$13,$6F,$F2,$D3,$A7,$FB,$13,$6B,$A6,$FF
	.dc.b	$3D,$B1,$C7,$96,$DF,$FA,$9B,$C0,$FC,$AF,$2F,$E2,$BC,$0B,$C7,$F5
	.dc.b	$DB,$87,$86,$AD,$F1,$8F,$F5,$9C,$E3,$33,$6A,$4E,$88,$7E,$6E,$7F
	.dc.b	$B0,$3F,$EB,$CE,$FE,$5E,$FC,$8C,$10,$E0,$E5,$E5,$1E,$1F,$9D,$91
	.dc.b	$FE,$C6,$7C,$FD,$6F,$0B,$33,$E1,$9B,$AF,$F3,$29,$DF,$B5,$C3,$EC
	.dc.b	$A5,$FF,$5B,$8C,$BF,$D7,$63,$2F,$FA,$DE,$3E,$2F,$ED,$7F,$82,$8F
	.dc.b	$F0,$EA,$F9,$0C,$9A,$E6,$93,$A8,$CB,$C2,$32,$FF,$01,$9E,$4E,$CD
	.dc.b	$1D,$BD,$5F,$0F,$34,$8C,$10,$E4,$E5,$E9,$1E,$83,$19,$FF,$64,$AF
	.dc.b	$E0,$93,$A3,$D7,$F5,$7E,$EF,$83,$91,$82,$1E,$77,$2F,$A6,$3E,$29
	.dc.b	$8A,$ED,$38,$9F,$8F,$F0,$DF,$2B,$E4,$A4,$60,$87,$C3,$72,$FC,$58
	.dc.b	$FA,$26,$2B,$D5,$38,$9F,$A7,$E5,$7D,$6F,$3A,$46,$08,$7C,$F7,$2F
	.dc.b	$D1,$8F,$BA,$62,$BE,$34,$E2,$7E,$FF,$D5,$7E,$08,$8F,$E0,$BE,$AB
	.dc.b	$F0,$A6,$1F,$B9,$18,$74,$43,$F0,$11,$F8,$CF,$AA,$FC,$6F,$A1,$27
	.dc.b	$F1,$AF,$E4,$3E,$43,$F2,$3F,$55,$67,$F2,$31,$87,$F2,$7F,$55,$F9
	.dc.b	$62,$3F,$96,$FA,$AF,$DF,$AF,$AA,$4E,$14,$D3,$11,$27,$FD,$1B,$3A
	.dc.b	$6C,$E5,$FF,$36,$9D,$67,$93,$A3,$8B,$FA,$AF,$3F,$F5,$7F,$53,$27
	.dc.b	$F5,$71,$97,$F5,$B1,$97,$F4,$D1,$FB,$2F,$4F,$F8,$BF,$EE,$DF,$D0
	.dc.b	$F8,$7F,$C7,$ED,$F1,$FF,$D7,$7D,$0A,$E1,$87,$FA,$52,$B4,$FF,$54
	.dc.b	$6F,$8E,$AE,$13,$BF,$3F,$BF,$D3,$48,$C1,$0F,$EF,$56,$8D,$46,$27
	.dc.b	$CE,$63,$2F,$3C,$E2,$7F,$D7,$3F,$52,$F1,$3F,$D9,$10,$8C,$3F,$EC
	.dc.b	$B6,$BF,$D8,$1A,$FF,$65,$73,$4D,$3F,$C2,$F1,$86,$D7,$E0,$7C,$0F
	.dc.b	$97,$F4,$DF,$33,$E6,$26,$3F,$1E,$32,$FC,$98,$CB,$F0,$23,$F1,$BF
	.dc.b	$7F,$FD,$6A,$3F,$06,$FF,$42,$6F,$99,$AF,$3A,$47,$FB,$62,$75,$A7
	.dc.b	$9E,$02,$33,$9F,$F8,$C1,$ED,$3B,$9D,$57,$8A,$E4,$75,$8F,$C3,$FC
	.dc.b	$8F,$E2,$3E,$66,$4F,$E0,$E3,$2F,$E1,$63,$2F,$DC,$8F,$D4,$FA,$DF
	.dc.b	$A1,$8F,$F0,$7B,$5F,$F0,$A3,$5F,$F2,$2B,$5F,$5B,$C7,$0F,$43,$72
	.dc.b	$DF,$4F,$ED,$BC,$EF,$DA,$DF,$FB,$F3,$EF,$FD,$F5,$FE,$50,$7E,$1C
	.dc.b	$7C,$66,$A7,$6E,$73,$FE,$03,$1A,$23,$83,$3A,$49,$F0,$47,$F0,$8F
	.dc.b	$C7,$3F,$D4,$91,$FE,$2E,$0E,$1F,$F1,$8B,$5F,$F1,$B3,$5F,$F3,$3B
	.dc.b	$BC,$FE,$C3,$70,$CF,$89,$FC,$5B,$C5,$F4,$DB,$C9,$FF,$2E,$3E,$16
	.dc.b	$BF,$E4,$06,$BF,$E4,$56,$BF,$E7,$67,$78,$7E,$1D,$B7,$9F,$67,$F9
	.dc.b	$2F,$ED,$FF,$23,$B5,$D8,$F9,$CF,$F1,$3F,$E2,$6D,$FF,$C2,$1A,$7E
	.dc.b	$5B,$4F,$CC,$4E,$7D,$3F,$E2,$B7,$2B,$5A,$7E,$72,$B5,$18,$0E,$1F
	.dc.b	$E1,$73,$FC,$7B,$E6,$64,$7F,$19,$7F,$90,$FC,$CE,$23,$F8,$8C,$5C
	.dc.b	$7F,$91,$CF,$F8,$E7,$EB,$BF,$C7,$5F,$1D,$2F,$C1,$B3,$DA,$56,$E7
	.dc.b	$30,$AE,$11,$FE,$3D,$9F,$F2,$27,$D9,$FF,$91,$78,$A4,$F0,$8C,$BC
	.dc.b	$A3,$2F,$F9,$16,$32,$FF,$91,$E3,$39,$7F,$A8,$27,$2F,$FA,$8A,$33
	.dc.b	$B7,$FA,$3E,$7F,$CA,$9F,$39,$FE,$55,$F8,$99,$87,$A2,$32,$FA,$A3
	.dc.b	$2F,$F9,$56,$3F,$CB,$FF,$4F,$FE,$E3,$8F,$86,$BC,$CC,$9A,$F3,$24
	.dc.b	$EA,$32,$F0,$8C,$BF,$EB,$B9,$E4,$EC,$D1,$F4,$3E,$93,$FA,$33,$7C
	.dc.b	$5F,$F3,$8C,$1B,$E5,$49,$D1,$0F,$F6,$44,$F2,$6C,$72,$99,$1D,$E3
	.dc.b	$93,$C2,$32,$F0,$3F,$24,$7F,$AA,$7B,$1B,$FF,$32,$FA,$67,$8D,$FE
	.dc.b	$E0,$F8,$57,$B1,$FE,$E2,$9C,$E1,$FC,$E2,$72,$F1,$AE,$18,$74,$85
	.dc.b	$70,$7C,$18,$6D,$E5,$87,$5B,$F0,$36,$3B,$5F,$3B,$E5,$82,$31,$7C
	.dc.b	$2B,$A1,$FE,$C5,$BE,$4F,$80,$F1,$C7,$9B,$0D,$DF,$AA,$72,$79,$0C
	.dc.b	$DF,$67,$C7,$3C,$2F,$AD,$71,$C3,$CB,$7B,$BF,$07,$97,$FE,$80,$6F
	.dc.b	$BB,$E1,$7F,$D9,$ED,$8D,$47,$F6,$4F,$84,$FE,$CB,$18,$84,$62,$7F
	.dc.b	$DC,$6D,$8F,$F7,$43,$5F,$F6,$B1,$EC,$3F,$B7,$FA,$1F,$DC,$70,$FA
	.dc.b	$91,$DC,$D0,$8D,$69,$DA,$33,$18,$7B,$98,$E6,$FA,$2F,$3A,$EB,$3B
	.dc.b	$8F,$1C,$62,$77,$1E,$C3,$91,$E2,$31,$1B,$71,$7D,$AB,$A4,$77,$BF
	.dc.b	$F0,$5B,$EE,$BF,$C7,$75,$B8,$5F,$19,$FE,$11,$E0,$FF,$09,$4E,$7E
	.dc.b	$2F,$17,$FE,$14,$8E,$15,$8C,$11,$B9,$FD,$F4,$E6,$31,$3F,$B8,$8C
	.dc.b	$60,$BB,$18,$6D,$FF,$07,$6D,$7F,$00,$6E,$1E,$6B,$E5,$5F,$CF,$3E
	.dc.b	$95,$F3,$DE,$17,$DB,$19,$8C,$63,$96,$C6,$4E,$51,$96,$99,$C7,$F8
	.dc.b	$2B,$F0,$EB,$AC,$7F,$C0,$1F,$13,$E0,$78,$0F,$94,$B1,$86,$DF,$F4
	.dc.b	$E6,$D7,$43,$E5,$D7,$19,$FE,$A7,$F0,$9D,$3E,$17,$AB,$63,$27,$F5
	.dc.b	$88,$CB,$4C,$D7,$5D,$FF,$B0,$4E,$A3,$16,$8C,$6D,$F8,$28,$C5,$F2
	.dc.b	$8C,$4F,$DD,$FD,$F5,$CD,$79,$B0,$DB,$FE,$DF,$3A,$6C,$E9,$A7,$FA
	.dc.b	$57,$E8,$3F,$11,$E5,$9B,$C5,$86,$4E,$CF,$F4,$E4,$63,$4F,$F4,$9F
	.dc.b	$17,$E5,$7C,$A7,$E6,$9F,$E9,$5B,$FF,$51,$4E,$63,$15,$C2,$7F,$C2
	.dc.b	$95,$C3,$BF,$C7,$7B,$7F,$41,$4F,$D3,$9F,$37,$86,$E1,$FE,$1D,$EB
	.dc.b	$78,$CF,$84,$D5,$CA,$FF,$26,$5F,$0C,$3F,$6E,$7F,$A0,$9C,$38,$6B
	.dc.b	$FB,$CF,$A6,$FF,$5F,$4F,$F5,$A6,$C7,$E2,$F1,$7D,$23,$E6,$3D,$35
	.dc.b	$AE,$7F,$F4,$53,$C2,$FC,$2D,$7B,$7C,$17,$FA,$3D,$B8,$F9,$97,$DE
	.dc.b	$FF,$0E,$8F,$F1,$E3,$FC,$3E,$3F,$38,$E5,$A7,$F2,$B8,$DB,$FE,$A0
	.dc.b	$6E,$5F,$C3,$67,$86,$1F,$EA,$57,$A3,$6B,$81,$D2,$78,$38,$8D,$47
	.dc.b	$D8,$FA,$6F,$A5,$7F,$E0,$06,$D7,$EC,$7F,$09,$7E,$12,$F9,$4E,$1C
	.dc.b	$33,$0F,$F4,$16,$1C,$30,$FA,$ED,$7F,$97,$B5,$FF,$77,$78,$BF,$9C
	.dc.b	$F9,$3F,$CE,$B6,$BF,$DE,$8D,$9F,$F7,$5B,$74,$F8,$1F,$2F,$FE,$DB
	.dc.b	$AF,$9A,$FD,$B6,$9F,$1D,$C5,$7D,$07,$83,$78,$75,$3B,$6B,$D1,$B1
	.dc.b	$B8,$FF,$33,$F6,$7E,$56,$FF,$A2,$AF,$A2,$FD,$BF,$37,$EE,$1B,$5C
	.dc.b	$1B,$7F,$C4,$6B,$84,$FF,$8C,$3E,$A1,$F3,$BF,$AC,$6C,$7F,$A1,$7E
	.dc.b	$7D,$F7,$AF,$98,$D5,$CA,$FF,$5A,$5F,$0C,$7F,$28,$D7,$F5,$6D,$7F
	.dc.b	$AD,$62,$FA,$4F,$E1,$FE,$47,$F1,$06,$DF,$C0,$6D,$7A,$5B,$B7,$C2
	.dc.b	$6C,$F4,$6C,$FF,$64,$BF,$25,$7C,$8F,$A4,$FE,$3B,$97,$F1,$E6,$E1
	.dc.b	$F3,$9B,$7F,$1D,$FE,$47,$F4,$9F,$DA,$FE,$B7,$ED,$E1,$F4,$A3,$17
	.dc.b	$A3,$FD,$3B,$3A,$7D,$F1,$A7,$F1,$3F,$23,$F8,$7F,$80,$FC,$43,$63
	.dc.b	$27,$E3,$A3,$2D,$33,$8F,$E3,$63,$2D,$8F,$0F,$27,$9C,$F4,$7D,$1D
	.dc.b	$8F,$2F,$C7,$FA,$FE,$83,$D8,$D8,$F4,$FC,$47,$BC,$C6,$1F,$D6,$74
	.dc.b	$FF,$69,$4F,$BB,$E3,$7E,$21,$F1,$7B,$EC,$64,$F8,$7F,$1F,$8B,$E4
	.dc.b	$FC,$B2,$46,$E3,$2F,$CA,$FA,$0F,$A1,$BF,$D1,$6C,$64,$F9,$F1,$96
	.dc.b	$99,$9F,$44,$65,$B1,$F5,$B7,$FB,$9F,$11,$F7,$5B,$19,$3E,$DC,$65
	.dc.b	$A6,$67,$E2,$46,$5B,$1F,$82,$F8,$7F,$C4,$7C,$47,$E2,$5B,$19,$3F
	.dc.b	$0F,$19,$69,$99,$FA,$11,$96,$C7,$E3,$BE,$23,$F2,$9E,$77,$E5,$5B
	.dc.b	$19,$3F,$27,$19,$69,$99,$FB,$91,$96,$C7,$E6,$BD,$CF,$D0,$7B,$9F
	.dc.b	$A1,$6C,$64,$FC,$FC,$65,$A6,$67,$F1,$11,$96,$C7,$E9,$BD,$CF,$D6
	.dc.b	$7B,$7F,$AD,$6C,$64,$FD,$5C,$65,$A7,$F8,$73,$C1,$FF,$11,$1F,$E5
	.dc.b	$FF,$99,$FF,$0F,$5F,$F8,$BF,$DB,$FE,$31,$5B,$24,$7F,$8C,$E7,$7D
	.dc.b	$BF,$E7,$0A,$DC,$F1,$FA,$0F,$F1,$77,$C7,$FF,$8B,$DB,$87,$F8,$B1
	.dc.b	$B3,$B8,$FF,$17,$CF,$F8,$D7,$E8,$3F,$C6,$DF,$41,$67,$FC,$66,$DA
	.dc.b	$F3,$37,$1F,$13,$67,$FD,$01,$8A,$31,$5F,$E8,$08,$C4,$FF,$8F,$BE
	.dc.b	$24,$FF,$AD,$F2,$E3,$18,$AF,$F5,$BC,$74,$F8,$8C,$7C,$3A,$37,$7C
	.dc.b	$6B,$B5,$EA,$73,$1C,$8F,$15,$E5,$DD,$71,$5F,$42,$F0,$7F,$D7,$51
	.dc.b	$FE,$19,$8C,$7A,$FF,$E1,$B8,$DC,$E9,$FF,$19,$CE,$9F,$F1,$07,$C5
	.dc.b	$3E,$4D,$E9,$F1,$8C,$36,$33,$3C,$1A,$F4,$F1,$BF,$CD,$2F,$8D,$AE
	.dc.b	$7E,$1B,$C9,$82,$73,$3F,$EB,$3F,$3F,$0F,$C7,$F8,$74,$62,$3E,$34
	.dc.b	$62,$7F,$D8,$7C,$3F,$3E,$7F,$D1,$FE,$4E,$38,$7A,$55,$9F,$F4,$9B
	.dc.b	$6F,$94,$77,$C1,$79,$5E,$47,$FB,$43,$9B,$FD,$15,$E4,$5E,$3F,$43
	.dc.b	$70,$FF,$4F,$D5,$C7,$D2,$DC,$3F,$C6,$52,$3F,$D2,$F5,$99,$91,$9C
	.dc.b	$3B,$B7,$0E,$ED,$3B,$97,$0F,$8F,$E4,$77,$F2,$24,$E7,$24,$F8,$CC
	.dc.b	$C7,$FA,$A2,$70,$EF,$C8,$F4,$11,$ED,$EB,$BC,$3C,$47,$FA,$C2,$6E
	.dc.b	$7F,$EF,$0A,$D4,$F0,$F0,$BE,$24,$7F,$AD,$63,$9F,$85,$89,$91,$82
	.dc.b	$1F,$EB,$F9,$47,$CD,$F9,$BF,$C6,$DF,$FB,$F9,$7C,$46,$4D,$73,$CC
	.dc.b	$3E,$34,$E9,$B3,$97,$FD,$99,$1D,$38,$3E,$B7,$02,$38,$4E,$AB,$86
	.dc.b	$1E,$1F,$13,$C9,$CD,$2B,$BC,$79,$CC,$3D,$A3,$0E,$8F,$2B,$A9,$E9
	.dc.b	$1E,$8F,$0F,$E0,$78,$72,$30,$43,$D2,$E5,$F5,$C7,$C7,$31,$3F,$1E
	.dc.b	$71,$3F,$13,$C3,$F9,$9E,$14,$8C,$10,$F8,$AE,$5F,$8F,$1F,$4C,$C5
	.dc.b	$7D,$39,$C4,$FD,$0F,$0B,$EC,$77,$48,$C1,$0F,$A2,$E5,$FA,$71,$F7
	.dc.b	$CC,$57,$DF,$9C,$4F,$DC,$FA,$1F,$C2,$11,$FB,$FF,$43,$F8,$73,$0F
	.dc.b	$DE,$8C,$3A,$21,$F8,$28,$FC,$77,$BB,$F8,$FF,$A1,$93,$F8,$A7,$F2
	.dc.b	$5E,$47,$E4,$FE,$86,$CF,$E3,$A3,$0F,$E4,$3E,$83,$F3,$04,$7F,$27
	.dc.b	$F3,$9F,$C2,$57,$D0,$A7,$0A,$69,$88,$93,$FC,$0C,$E3,$1A,$71,$3C
	.dc.b	$DB,$39,$5A,$FF,$5B,$78,$9A,$3F,$4F,$F2,$DF,$AF,$F9,$79,$18,$21
	.dc.b	$FA,$87,$2F,$EA,$A3,$F6,$DF,$11,$FC,$8F,$DD,$FE,$41,$3F,$EB,$23
	.dc.b	$27,$ED,$23,$4C,$30,$7F,$AE,$3C,$FE,$5F,$19,$23,$04,$3F,$D6,$8E
	.dc.b	$5F,$F5,$B4,$78,$4C,$63,$FE,$C0,$9C,$4F,$C9,$FA,$17,$89,$F4,$3B
	.dc.b	$24,$E6,$3E,$B4,$EA,$3F,$B7,$35,$F6,$9F,$CF,$5C,$38,$6B,$FE,$B8
	.dc.b	$F9,$7F,$89,$F4,$32,$3F,$DA,$B1,$F1,$CC,$3F,$EC,$F8,$C3,$F7,$FE
	.dc.b	$87,$FA,$BB,$7C,$CF,$C5,$47,$F5,$88,$D7,$9D,$83,$33,$23,$24,$3F
	.dc.b	$07,$E2,$FD,$6A,$E5,$18,$3F,$AC,$38,$D7,$F1,$0F,$8A,$3F,$DC,$9F
	.dc.b	$43,$F7,$FE,$66,$47,$FB,$C2,$3F,$04,$61,$FF,$75,$46,$1F,$CD,$7A
	.dc.b	$DF,$9F,$7F,$39,$E7,$B9,$7F,$4C,$D8,$DF,$4F,$EA,$6B,$84,$7D,$23
	.dc.b	$FC,$4D,$1F,$E2,$36,$9F,$E2,$58,$DB,$F6,$3E,$87,$D5,$9F,$E8,$DF
	.dc.b	$76,$9F,$98,$6C,$FF,$79,$77,$19,$DB,$F7,$51,$A9,$34,$DB,$C6,$37
	.dc.b	$B4,$8F,$DD,$4E,$97,$D0,$FF,$8D,$5F,$DF,$78,$7F,$E0,$A7,$EA,$07
	.dc.b	$18,$FE,$EE,$78,$4E,$9A,$FE,$59,$AF,$ED,$1A,$FF,$85,$5A,$BC,$77
	.dc.b	$BB,$FC,$5F,$E8,$7B,$E9,$FB,$C6,$D7,$F8,$76,$B6,$FF,$15,$6B,$F5
	.dc.b	$5A,$FE,$A1,$AF,$F0,$66,$AF,$35,$DA,$FA,$1F,$F1,$87,$CC,$7F,$8E
	.dc.b	$3D,$4B,$C3,$FE,$10,$C3,$F3,$18,$70,$BF,$E2,$46,$AA,$4F,$F8,$A7
	.dc.b	$0E,$8D,$FA,$1F,$E3,$04,$75,$F4,$35,$EB,$29,$C7,$07,$FA,$17,$1F
	.dc.b	$3D,$FF,$90,$7E,$17,$FC,$85,$F0,$B3,$1F,$F1,$A4,$65,$FF,$1B,$46
	.dc.b	$5F,$4F,$B4,$9F,$F2,$57,$C2,$FF,$92,$F8,$E5,$7F,$8F,$63,$2F,$F8
	.dc.b	$FE,$32,$F2,$8C,$BD,$23,$39,$79,$ED,$5A,$8C,$4E,$B8,$FF,$CE,$D3
	.dc.b	$FE,$5B,$F8,$5F,$F2,$E7,$C2,$CC,$7F,$C9,$D1,$97,$FC,$A1,$19,$7D
	.dc.b	$71,$FE,$65,$FA,$1F,$DD,$47,$D6,$5E,$56,$4D,$77,$C9,$FF,$2F,$4E
	.dc.b	$9B,$39,$79,$4F,$17,$27,$07,$FC,$E9,$F4,$0F,$26,$F8,$D2,$BF,$CD
	.dc.b	$53,$96,$CE,$5F,$C5,$C2,$79,$CC,$9E,$31,$BD,$BF,$05,$5C,$E7,$8D
	.dc.b	$67,$5F,$F2,$6C,$F3,$9E,$E7,$F7,$BF,$85,$FD,$E5,$FF,$A1,$FE,$85
	.dc.b	$E2,$FF,$6A,$97,$AF,$FD,$B2,$73,$7F,$E8,$D9,$CB,$C6,$B8,$78,$3F
	.dc.b	$E0,$08,$57,$05,$5E,$07,$35,$DE,$FB,$36,$3A,$1D,$EF,$A5,$F3,$BE
	.dc.b	$4D,$BE,$35,$E5,$6B,$E0,$AE,$37,$E6,$C3,$AC,$F9,$2B,$78,$7B,$23
	.dc.b	$CA,$78,$67,$71,$E0,$AF,$34,$E8,$E4,$6A,$FB,$BC,$47,$BB,$FE,$7E
	.dc.b	$AF,$84,$D9,$F9,$0D,$BF,$F5,$D3,$FE,$84,$AF,$F5,$E7,$C2,$63,$DD
	.dc.b	$FE,$A4,$D8,$FF,$02,$B5,$FF,$6C,$9E,$08,$DB,$E7,$1E,$83,$6F,$87
	.dc.b	$1D,$8D,$46,$5E,$02,$B7,$19,$D7,$EF,$4F,$16,$C6,$74,$F0,$C6,$2F
	.dc.b	$BD,$73,$3B,$B8,$BE,$AD,$8C,$9E,$28,$CB,$4C,$CF,$E5,$BB,$1B,$BE
	.dc.b	$77,$D5,$AF,$EA,$3E,$5A,$E5,$FA,$69,$B8,$FF,$43,$B5,$63,$0E,$75
	.dc.b	$67,$37,$FB,$C8,$C5,$65,$B1,$93,$F8,$0C,$65,$A6,$6F,$F8,$0C,$77
	.dc.b	$FA,$17,$D0,$5B,$C3,$6F,$FD,$E9,$3A,$6C,$E9,$A6,$A7,$85,$FF,$58
	.dc.b	$AE,$0D,$AF,$77,$6F,$FB,$C6,$FF,$DF,$9F,$19,$8F,$13,$2D,$8C,$9F
	.dc.b	$F0,$06,$32,$AF,$FA,$6F,$C2,$74,$F8,$6E,$AD,$9D,$6F,$FD,$16,$31
	.dc.b	$5C,$27,$27,$FB,$9B,$15,$F2,$8C,$CF,$F7,$99,$CC,$63,$74,$7F,$BF
	.dc.b	$23,$0E,$A7,$B1,$FD,$8A,$BA,$36,$79,$7C,$67,$EE,$1B,$CB,$F8,$38
	.dc.b	$C4,$FF,$80,$94,$67,$9B,$4D,$AE,$27,$F8,$0F,$E8,$7F,$19,$E4,$1B
	.dc.b	$C5,$86,$4E,$CF,$F0,$54,$63,$1F,$F0,$2F,$0F,$E6,$3C,$B7,$D0,$FF
	.dc.b	$51,$BF,$F0,$84,$E6,$31,$5C,$27,$E4,$D7,$0E,$EF,$1D,$ED,$FD,$1D
	.dc.b	$3E,$85,$F3,$7F,$EB,$B8,$FF,$57,$C6,$30,$FF,$32,$46,$E7,$4F,$E0
	.dc.b	$E7,$4F,$F9,$A5,$BC,$9F,$C3,$7B,$5E,$33,$E0,$BC,$35,$AC,$7F,$19
	.dc.b	$5A,$9F,$2F,$D0,$3D,$FE,$F9,$23,$17,$E4,$6C,$7F,$9E,$5A,$FF,$AB
	.dc.b	$1C,$E5,$FC,$7F,$67,$E2,$FF,$E0,$8B,$FF,$71,$56,$9F,$CF,$4E,$BC
	.dc.b	$8F,$E1,$3F,$43,$FE,$D2,$9F,$CA,$BE,$86,$E3,$FA,$56,$E9,$C2,$30
	.dc.b	$7F,$A8,$E7,$83,$4C,$E5,$FC,$CE,$32,$D8,$CD,$7E,$C6,$72,$D8,$C8
	.dc.b	$8C,$BF,$B7,$F9,$F7,$C3,$FC,$A6,$D7,$DC,$6C,$FF,$A8,$5A,$FF,$B5
	.dc.b	$1C,$DF,$F4,$7F,$9E,$7E,$03,$FE,$7D,$BF,$DB,$DA,$F7,$1F,$C1,$EB
	.dc.b	$7E,$C7,$E9,$3C,$5F,$D2,$FC,$8D,$AF,$F3,$53,$70,$FF,$32,$B7,$2F
	.dc.b	$F3,$27,$62,$3E,$9F,$CB,$7F,$5D,$BF,$A6,$BE,$67,$31,$A6,$BA,$6E
	.dc.b	$1F,$4E,$78,$4F,$23,$81,$4E,$07,$F9,$FF,$A7,$F7,$BF,$41,$FB,$E6
	.dc.b	$C6,$4F,$F4,$24,$65,$A7,$C6,$F9,$F7,$CF,$7E,$71,$B1,$FA,$E6,$E1
	.dc.b	$FE,$D3,$6B,$FC,$C4,$FE,$64,$EB,$4F,$EC,$3F,$58,$F2,$FF,$7B,$39
	.dc.b	$BF,$E6,$73,$97,$FD,$31,$F4,$1F,$C5,$BC,$9F,$E2,$ED,$8C,$9F,$E9
	.dc.b	$D8,$CB,$4C,$E3,$FE,$9A,$8C,$B6,$3F,$D4,$99,$7F,$22,$FA,$0F,$E4
	.dc.b	$6D,$8C,$9F,$EA,$98,$CB,$4F,$F5,$67,$D0,$7F,$75,$F9,$FF,$BD,$7F
	.dc.b	$EA,$88,$C4,$FF,$2E,$6B,$C0,$FE,$8D,$3C,$35,$FD,$67,$03,$E3,$FF
	.dc.b	$13,$EF,$FE,$29,$B7,$F8,$A6,$E3,$F8,$56,$DF,$26,$CF,$EC,$EF,$A3
	.dc.b	$F9,$2B,$F4,$71,$C8,$F1,$FD,$05,$C7,$FD,$9B,$93,$E2,$BD,$DF,$3F
	.dc.b	$EF,$6D,$FA,$BE,$1F,$E1,$18,$BE,$2F,$C5,$F8,$B9,$3F,$03,$E7,$ED
	.dc.b	$76,$9F,$95,$F1,$D8,$E0,$FF,$67,$66,$8C,$E1,$F2,$1D,$3F,$DF,$FD
	.dc.b	$64,$EE,$3E,$6F,$CF,$FD,$5E,$2F,$AC,$D8,$C9,$F4,$A3,$2D,$33,$97
	.dc.b	$C2,$8C,$B6,$3E,$CF,$0F,$E0,$7E,$23,$F0,$4D,$8C,$9F,$7A,$32,$D3
	.dc.b	$33,$F3,$A3,$2D,$8F,$C2,$FC,$3F,$E3,$7E,$23,$F1,$CD,$8C,$9F,$8A
	.dc.b	$8C,$B4,$CC,$FD,$A8,$CB,$63,$F2,$3F,$11,$F9,$9F,$4B,$F3,$4D,$8C
	.dc.b	$9F,$96,$8C,$B4,$CC,$FE,$1A,$32,$D8,$FC,$EF,$BB,$FA,$5F,$03,$F4
	.dc.b	$CD,$8C,$9F,$A2,$8C,$B4,$CC,$FE,$4A,$32,$D8,$FD,$4F,$BB,$FB,$1F
	.dc.b	$77,$F6,$4D,$8C,$9F,$AE,$8C,$B4,$FF,$16,$FC,$A7,$F8,$AC,$FF,$37
	.dc.b	$7C,$BF,$F8,$B3,$97,$FD,$21,$C8,$F7,$D5,$BB,$E2,$7F,$8F,$EF,$83
	.dc.b	$F9,$39,$DB,$FE,$3A,$F9,$EF,$F1,$C7,$99,$FE,$39,$6C,$64,$FF,$1F
	.dc.b	$C6,$5A,$7F,$90,$BE,$7B,$FC,$81,$F3,$DF,$E4,$16,$C6,$4F,$F2,$3C
	.dc.b	$65,$A6,$76,$FF,$23,$46,$5B,$1F,$E4,$EF,$8A,$FF,$57,$7C,$3B,$1F
	.dc.b	$54,$61,$F6,$7C,$4D,$B6,$DF,$16,$B0,$DC,$3F,$D8,$F7,$C9,$B5,$C3
	.dc.b	$46,$1C,$1F,$F7,$FC,$EA,$5F,$41,$FC,$47,$D8,$FF,$1E,$B6,$3F,$A7
	.dc.b	$1F,$E7,$F7,$0E,$1A,$FF,$53,$F8,$B7,$CF,$BC,$9F,$F2,$3C,$61,$B1
	.dc.b	$99,$EA,$27,$4E,$29,$E3,$FF,$BB,$23,$0D,$7C,$4D,$7F,$B2,$47,$0F
	.dc.b	$88,$EB,$B7,$E3,$70,$FE,$D1,$C1,$F4,$6F,$E3,$7C,$FE,$38,$3F,$18
	.dc.b	$D8,$F9,$55,$6F,$94,$7C,$E7,$F3,$13,$9D,$CD,$AC,$7C,$FF,$F7,$7E
	.dc.b	$8F,$CF,$D2,$31,$C3,$F9,$B6,$E3,$F6,$7D,$CB,$97,$E9,$66,$DF,$1A
	.dc.b	$FF,$4C,$C6,$2B,$2C,$8C,$9F,$E9,$E8,$E3,$F0,$72,$3F,$D4,$2F,$4E
	.dc.b	$9C,$47,$FA,$7E,$2E,$5E,$89,$F0,$FC,$FF,$8B,$E7,$A5,$FF,$AA,$A3
	.dc.b	$2F,$FA,$B2,$32,$FF,$AA,$63,$D7,$2E,$BA,$49,$FD,$F7,$5D,$58,$7F
	.dc.b	$06,$AE,$11,$FC,$0E,$B9,$D7,$4E,$B9,$3F,$C1,$DE,$FD,$78,$8E,$7C
	.dc.b	$1F,$D6,$F0,$56,$71,$ED,$76,$B5,$36,$FF,$67,$6A,$F5,$F5,$C9,$FE
	.dc.b	$2C,$FB,$B8,$B1,$1E,$03,$8C,$63,$1F,$64,$DC,$7C,$93,$6B,$BC,$FC
	.dc.b	$8E,$39,$1F,$C8,$DF,$97,$C3,$88,$ED,$C6,$CF,$F2,$71,$B3,$F9,$54
	.dc.b	$EF,$2F,$99,$77,$2F,$89,$3F,$5B,$8A,$5F,$F3,$17,$EC,$F1,$62,$3B
	.dc.b	$E3,$FE,$8C,$74,$7F,$37,$8D,$63,$F6,$6A,$E1,$F4,$67,$F0,$77,$2F
	.dc.b	$F9,$F3,$F8,$6E,$1C,$47,$68,$B8,$7E,$0E,$6D,$FD,$B9,$B5,$F5,$A7
	.dc.b	$F1,$FC,$5F,$90,$F9,$A9,$3F,$D2,$E3,$2D,$21,$EC,$F9,$3F,$CB,$9A
	.dc.b	$7D,$F1,$87,$FA,$A7,$C1,$FE,$70,$8F,$F5,$7F,$87,$FC,$F9,$1C,$03
	.dc.b	$EB,$FD,$CF,$F2,$ED,$3F,$B1,$AF,$03,$F8,$0B,$9D,$7F,$C6,$4D,$AF
	.dc.b	$E8,$AE,$DF,$F5,$8F,$D0,$7F,$B9,$DB,$87,$F0,$C5,$19,$7F,$28,$3E
	.dc.b	$5F,$A3,$A9,$E1,$3B,$C5,$87,$FA,$46,$30,$FF,$A4,$EF,$8C,$62,$FD
	.dc.b	$CD,$7A,$1F,$EC,$68,$F5,$72,$31,$7C,$67,$C0,$25,$7C,$7C,$7E,$2F
	.dc.b	$CB,$62,$3E,$4C,$D9,$CC,$6F,$2F,$8F,$F3,$FF,$28,$91,$93,$E4,$FC
	.dc.b	$F6,$1F,$A7,$16,$BE,$9C,$FD,$2F,$98,$FA,$7F,$37,$2F,$96,$9E,$99
	.dc.b	$D4,$63,$45,$6E,$3E,$FC,$E9,$FC,$0C,$67,$0F,$C0,$CF,$F4,$2F,$8A
	.dc.b	$FA,$9F,$03,$77,$FF,$63,$E1,$E9,$9C,$BD,$FE,$6F,$55,$23,$55,$C1
	.dc.b	$FC,$7C,$FE,$0F,$E8,$3F,$D4,$51,$FD,$C8,$38,$7F,$BA,$B5,$FF,$01
	.dc.b	$B5,$FF,$69,$3B,$C7,$F0,$B1,$F9,$D8,$C6,$FA,$F6,$3F,$17,$E4,$FE
	.dc.b	$4F,$E8,$1E,$BD,$26,$1F,$54,$43,$D7,$CC,$7E,$62,$38,$BB,$6B,$98
	.dc.b	$DE,$5E,$48,$E7,$58,$C7,$FD,$D5,$7D,$67,$EC,$69,$FB,$02,$3F,$6F
	.dc.b	$E6,$2B,$FB,$88,$B8,$7E,$E2,$7F,$5D,$F0,$3F,$CC,$2B,$F5,$AD,$9F
	.dc.b	$13,$70,$FF,$7D,$7D,$CF,$E1,$AB,$E0,$4E,$5F,$E0,$CD,$7F,$D8,$DE
	.dc.b	$37,$EE,$E7,$84,$7E,$EB,$DF,$FF,$12,$61,$EF,$6E,$1F,$C3,$E7,$FD
	.dc.b	$7D,$B7,$F0,$39,$7C,$D5,$89,$3F,$46,$78,$BF,$C5,$DD,$B9,$8D,$BE
	.dc.b	$17,$A4,$FF,$0F,$FB,$77,$B9,$49,$5F,$77,$E2,$FF,$CA,$AF,$DF,$9D
	.dc.b	$B6,$76,$FF,$16,$9D,$BF,$C6,$3F,$1D,$73,$7F,$1D,$6C,$7E,$25,$AF
	.dc.b	$F1,$EF,$BE,$7C,$EE,$EA,$4E,$DF,$E4,$D3,$A8,$FD,$D4,$FF,$32,$23
	.dc.b	$FB,$DF,$9C,$AF,$F3,$A8,$BB,$FF,$3A,$9F,$E7,$3F,$2B,$FC,$E9,$95
	.dc.b	$FC,$16,$32,$F3,$21,$FD,$11,$FF,$1F,$F9,$72,$3F,$87,$FA,$E8,$FF
	.dc.b	$21,$F9,$99,$9F,$F0,$0C,$E5,$DC,$E9,$E3,$E5,$FF,$92,$88,$F4,$F2
	.dc.b	$AB,$CA,$2E,$5F,$E5,$89,$BC,$BF,$EB,$19,$B5,$FE,$8F,$9B,$E2,$FF
	.dc.b	$AE,$E6,$E3,$FD,$0E,$1F,$BF,$7F,$40,$25,$FF,$97,$9F,$EB,$4F,$F8
	.dc.b	$69,$CB,$FC,$B9,$E1,$E6,$6F,$CA,$93,$97,$97,$97,$88,$E5,$17,$4F
	.dc.b	$F6,$7D,$AB,$1A,$7F,$73,$C7,$CB,$78,$C3,$FD,$63,$7B,$7C,$5E,$5F
	.dc.b	$9C,$8F,$97,$CB,$AF,$A6,$2E,$5E,$09,$B8,$F7,$B5,$62,$FB,$D5,$C7
	.dc.b	$E3,$CD,$A4,$FF,$7B,$F7,$7F,$D7,$B8,$FF,$82,$FE,$85,$E5,$FF,$AC
	.dc.b	$FD,$7F,$E0,$11,$8C,$FF,$C2,$17,$FE,$59,$78,$57,$FA,$E7,$D7,$FF
	.dc.b	$37,$C2,$B8,$3F,$EB,$58,$CC,$62,$B9,$4E,$EB,$C0,$F4,$3B,$D7,$08
	.dc.b	$ED,$86,$67,$FD,$81,$87,$9A,$78,$C7,$33,$8C,$EE,$FD,$F1,$93,$89
	.dc.b	$9A,$E4,$F0,$35,$1C,$30,$F4,$4E,$A7,$89,$AC,$3C,$CF,$1F,$03,$F8
	.dc.b	$FB,$F0,$70,$FF,$1C,$46,$30,$F1,$B6,$BE,$4B,$C2,$BF,$C7,$DE,$E7
	.dc.b	$0F,$07,$FD,$BE,$D8,$FF,$76,$B5,$FF,$6A,$89,$E4,$A3,$E4,$3F,$E4
	.dc.b	$5E,$0A,$E4,$F8,$86,$4A,$67,$3F,$B6,$E5,$AE,$74,$FF,$7F,$F1,$FF
	.dc.b	$7A,$9C,$DF,$82,$31,$3E,$AA,$EF,$18,$7D,$66,$2F,$CB,$5C,$5B,$19
	.dc.b	$3C,$B1,$97,$F0,$1C,$3F,$C1,$AF,$FC,$17,$AF,$EE,$1B,$7F,$A9,$F3
	.dc.b	$ED,$FE,$DA,$6D,$F1,$AF,$E1,$11,$8E,$1F,$F1,$9C,$60,$95,$FC,$3E
	.dc.b	$75,$7B,$64,$EB,$0F,$E1,$33,$8A,$FE,$2D,$DD,$D4,$ED,$E4,$5B,$FE
	.dc.b	$31,$16,$7F,$8E,$4D,$BF,$E2,$73,$6B,$F8,$24,$DA,$FE,$23,$E4,$7A
	.dc.b	$4F,$57,$7D,$AF,$E4,$91,$67,$F9,$44,$DC,$3F,$8F,$F7,$FC,$13,$E1
	.dc.b	$77,$DC,$3F,$96,$45,$BF,$E6,$13,$6B,$8D,$68,$FE,$65,$19,$7F,$8D
	.dc.b	$55,$AF,$E4,$F3,$76,$FE,$7F,$F2,$DF,$44,$FA,$5F,$2D,$70,$FE,$79
	.dc.b	$16,$FF,$A0,$4D,$C7,$FD,$47,$3F,$73,$D2,$4D,$8C,$36,$FF,$CE,$0D
	.dc.b	$BF,$F5,$6B,$C7,$E5,$A4,$E8,$FF,$3B,$38,$9F,$F5,$9B,$D2,$32,$7F
	.dc.b	$94,$FF,$45,$FD,$A3,$79,$FF,$23,$CD,$FE,$1E,$F0,$BF,$27,$82,$F6
	.dc.b	$FF,$AB,$9B,$A7,$F9,$4D,$B8,$7F,$B0,$EF,$93,$D4,$FF,$31,$7E,$D2
	.dc.b	$DF,$AB,$63,$FD,$35,$2F,$AC,$7C,$26,$AE,$57,$27,$CF,$3A,$8C,$1F
	.dc.b	$EA,$3F,$9D,$FF,$6A,$74,$FE,$B1,$B1,$9D,$3F,$52,$34,$FF,$3F,$CE
	.dc.b	$A3,$13,$FE,$DD,$8C,$4E,$A7,$FA,$AC,$F3,$E9,$ED,$18,$38,$8A,$E3
	.dc.b	$3C,$E3,$55,$CE,$B8,$DF,$D0,$F9,$C7,$8F,$4F,$7F,$9B,$71,$FF,$46
	.dc.b	$4E,$9B,$3A,$69,$FE,$B9,$F9,$AF,$C8,$1F,$91,$F9,$AB,$19,$26,$7F
	.dc.b	$EF,$A8,$CB,$EA,$D1,$F1,$2F,$97,$F9,$8D,$BF,$A0,$BD,$6C,$CC,$9C
	.dc.b	$BC,$23,$17,$D6,$B9,$36,$75,$5C,$0F,$F5,$57,$CD,$BD,$DF,$CA,$B6
	.dc.b	$3F,$D9,$5F,$54,$F8,$DC,$D4,$9F,$F6,$25,$6F,$0F,$F4,$50,$0E,$5C
	.dc.b	$B5,$D7,$C0,$3F,$39,$ED,$5F,$ED,$96,$CF,$E0,$3C,$27,$85,$DF,$9F
	.dc.b	$FD,$97,$87,$1C,$FF,$9F,$B7,$1E,$51,$FC,$17,$3C,$E5,$D2,$FF,$DA
	.dc.b	$78,$F4,$F9,$DF,$E1,$5F,$3B,$FC,$2D,$B8,$FF,$BD,$DB,$A7,$8F,$44
	.dc.b	$63,$0E,$71,$8A,$FF,$7F,$5F,$36,$9F,$16,$7C,$3F,$BC,$A3,$FA,$AB
	.dc.b	$5D,$1F,$E5,$15,$F3,$DA,$88,$6B,$16,$4A,$E5,$1F,$25,$E3,$3B,$9D
	.dc.b	$3E,$0E,$18,$E6,$7F,$20,$D8,$CD,$73,$7C,$FC,$78,$C7,$F7,$FC,$6F
	.dc.b	$98,$FC,$D9,$F9,$CF,$83,$B8,$FE,$72,$2E,$3F,$97,$9B,$7C,$A6,$D7
	.dc.b	$F0,$7B,$56,$2F,$DF,$97,$F1,$77,$DD,$F3,$DF,$2B,$C1,$F9,$6F,$CC
	.dc.b	$F9,$D9,$8F,$B6,$2D,$7C,$5F,$9E,$FA,$26,$34,$FA,$AF,$D4,$F8,$3F
	.dc.b	$AB,$F3,$B7,$0F,$8C,$E1,$A7,$C9,$8C,$5F,$D1,$9F,$BB,$F1,$38,$C3
	.dc.b	$F0,$2F,$E0,$3E,$76,$61,$F4,$78,$2E,$1F,$66,$6D,$FE,$22,$6D,$7E
	.dc.b	$16,$6D,$7C,$49,$BC,$7F,$83,$E0,$FC,$69,$F8,$EF,$8A,$B8,$7D,$F8
	.dc.b	$B7,$F8,$79,$B8,$FE,$5A,$6D,$7E,$4E,$6D,$7D,$2F,$9D,$FC,$C9,$F9
	.dc.b	$AF,$16,$D7,$E2,$E2,$CF,$E4,$E6,$E1,$FA,$29,$B5,$F9,$F9,$B5,$F7
	.dc.b	$BE,$77,$F4,$A7,$E9,$BE,$76,$D7,$E5,$E2,$CF,$E7,$E6,$E1,$FA,$E9
	.dc.b	$B5,$FA,$B9,$B5,$F8,$AF,$9D,$FD,$89,$FB,$2F,$9D,$B5,$FA,$38,$B3
	.dc.b	$FA,$B9,$B8,$7E,$EA,$6D,$7E,$DE,$6D,$7E,$D3,$F7,$AF,$EF,$6D,$F4
	.dc.b	$FF,$CA,$70,$F7,$EC,$4B,$FD,$B4,$F0,$E0,$3B,$8C,$CF,$F9,$67,$D5
	.dc.b	$FF,$2D,$79,$76,$7F,$DC,$8D,$C7,$FD,$D0,$DD,$3F,$DB,$EF,$F9,$87
	.dc.b	$CB,$95,$FE,$EB,$F1,$AD,$FF,$BB,$E6,$DF,$FB,$9A,$6D,$7F,$B5,$FC
	.dc.b	$5F,$F3,$51,$FE,$6B,$F1,$2C,$7F,$BD,$62,$E1,$FE,$F9,$9B,$7F,$EF
	.dc.b	$09,$B5,$FE,$E1,$9B,$5F,$C7,$3F,$7F,$FE,$37,$EF,$BE,$7B,$FD,$E9
	.dc.b	$1F,$C7,$27,$FD,$45,$18,$BF,$E5,$91,$83,$97,$8F,$8B,$E5,$1D,$67
	.dc.b	$15,$D2,$AC,$E6,$7F,$D6,$F3,$96,$C7,$F5,$FF,$9F,$FE,$ED,$EB,$7F
	.dc.b	$62,$8C,$79,$08,$C4,$27,$6E,$DA,$FF,$6A,$F8,$47,$CE,$7F,$81,$A3
	.dc.b	$A4,$63,$CA,$E6,$D8,$CC,$ED,$FE,$E7,$5F,$E4,$0E,$5F,$0B,$63,$C1
	.dc.b	$E4,$FF,$B2,$DB,$3E,$9E,$87,$C5,$F8,$A7,$DF,$18,$F9,$3F,$EF,$D1
	.dc.b	$F3,$63,$0E,$A7,$83,$96,$F0,$FD,$09,$DC,$E6,$B9,$CE,$6B,$7C,$1F
	.dc.b	$E0,$BB,$F1,$FC,$E7,$83,$9B,$FB,$CD,$23,$1C,$5B,$8C,$65,$FD,$7D
	.dc.b	$7C,$E6,$A3,$FB,$3C,$69,$FF,$07,$36,$FA,$DE,$F0,$F8,$77,$6B,$58
	.dc.b	$77,$AC,$19,$9D,$7C,$E4,$C7,$C4,$F1,$F9,$CC,$47,$82,$2E,$5E,$C9
	.dc.b	$ED,$F3,$9D,$FE,$72,$5F,$AE,$32,$FB,$63,$2F,$AA,$3D,$12,$F9,$FA
	.dc.b	$49,$FF,$01,$43,$E7,$F1,$1F,$E2,$F8,$E0,$EA,$79,$CF,$4F,$9F,$93
	.dc.b	$C9,$EF,$F2,$38,$8E,$51,$72,$FF,$17,$E0,$AC,$6D,$F7,$61,$58,$BF
	.dc.b	$F0,$DC,$6D,$F3,$FC,$8F,$B0,$8F,$AB,$A2,$BE,$98,$B8,$F8,$26,$E1
	.dc.b	$DE,$6D,$76,$9F,$8F,$CB,$23,$E2,$BF,$2B,$87,$11,$F1,$62,$E1,$EC
	.dc.b	$9B,$87,$AE,$6D,$7A,$A7,$E9,$F0,$C8,$FA,$2F,$D6,$E1,$C4,$7D,$18
	.dc.b	$B8,$7C,$89,$B8,$7C,$79,$B5,$F1,$A7,$EF,$E7,$23,$EE,$BF,$82,$E0
	.dc.b	$C4,$7D,$D8,$B8,$7D,$49,$B8,$7D,$39,$B5,$F4,$A7,$F1,$7C,$3F,$8C
	.dc.b	$F9,$69,$3F,$8A,$8C,$B4,$87,$E2,$FE,$5F,$F2,$66,$9F,$C7,$46,$1F
	.dc.b	$C9,$FC,$17,$E6,$08,$FE,$5B,$E2,$BF,$36,$47,$07,$F9,$6F,$DD,$FF
	.dc.b	$48,$E9,$FE,$72,$F0,$3F,$71,$18,$D7,$F8,$53,$6B,$FC,$CE,$FF,$35
	.dc.b	$7F,$CD,$43,$E7,$D8,$7F,$07,$8C,$36,$33,$1C,$3E,$47,$A3,$A3,$9C
	.dc.b	$69,$FF,$6C,$99,$2C,$70,$8D,$63,$FD,$62,$74,$D8,$CE,$1F,$CC,$2B
	.dc.b	$D5,$D7,$FE,$A4,$F9,$7F,$79,$30,$F8,$9F,$31,$5F,$F4,$C4,$5C,$7F
	.dc.b	$D2,$F3,$F1,$BE,$62,$57,$CB,$7E,$4F,$CC,$62,$3F,$D4,$11,$6B,$FD
	.dc.b	$3F,$3F,$3B,$E2,$FE,$7F,$C6,$4B,$FA,$91,$97,$EB,$46,$5F,$F5,$54
	.dc.b	$65,$FF,$56,$46,$6B,$FD,$57,$3F,$CD,$FE,$21,$A7,$FB,$5E,$FF,$C4
	.dc.b	$BE,$0F,$F3,$5F,$4F,$F8,$BC,$30,$FF,$62,$D6,$E3,$1B,$FF,$B4,$23
	.dc.b	$13,$AF,$43,$B4,$62,$7F,$0F,$F2,$DF,$81,$F9,$F7,$C0,$FE,$07,$E2
	.dc.b	$A6,$5F,$88,$9C,$B6,$7F,$BF,$A8,$CB,$FE,$05,$6B,$FE,$E8,$6B,$FD
	.dc.b	$B3,$E7,$FF,$D3,$EF,$E5,$1A,$FE,$29,$B9,$FE,$1D,$7C,$07,$E5,$DC
	.dc.b	$E5,$CF,$E0,$70,$66,$6D,$F5,$8A,$F5,$8F,$D4,$E9,$23,$F5,$CF,$EB
	.dc.b	$3E,$63,$11,$FE,$E3,$8B,$8F,$FB,$8A,$7F,$51,$EA,$3C,$7F,$D5,$7C
	.dc.b	$FF,$DA,$C3,$F5,$ED,$7F,$4B,$3B,$7F,$66,$D7,$33,$D0,$E7,$CD,$FB
	.dc.b	$5F,$87,$FF,$0D,$4F,$EB,$9B,$8F,$EA,$DB,$AF,$F8,$61,$FF,$38,$4F
	.dc.b	$F0,$07,$28,$38,$6F,$A5,$DF,$DE,$F0,$32,$39,$39,$7F,$85,$FC,$FB
	.dc.b	$E1,$3F,$85,$7C,$FF,$AF,$2F,$DF,$37,$2F,$F1,$5D,$6E,$F8,$DF,$F1
	.dc.b	$18,$C4,$7F,$0A,$6D,$F2,$F4,$3F,$8B,$7C,$FB,$A7,$F8,$C7,$CF,$BE
	.dc.b	$5E,$9A,$89,$8F,$F2,$0A,$E0,$D8,$E3,$F3,$BF,$CA,$89,$1F,$CB,$DF
	.dc.b	$E5,$BF,$3B,$88,$FD,$69,$FC,$B2,$31,$97,$F9,$CB,$C7,$79,$4B,$FE
	.dc.b	$6B,$97,$2C,$3F,$9D,$46,$1F,$E7,$91,$97,$AD,$72,$9D,$47,$F8,$E3
	.dc.b	$CF,$FF,$1A,$11,$FF,$20,$79,$F5,$D1,$FE,$43,$8F,$F1,$D4,$7F,$8E
	.dc.b	$FC,$E9,$3F,$E4,$77,$FC,$7D,$E6,$E2,$3F,$C8,$B1,$73,$FF,$28,$CD
	.dc.b	$E3,$FF,$28,$4D,$AF,$F2,$6C,$DC,$BF,$CD,$73,$7D,$9F,$E7,$74,$97
	.dc.b	$CF,$7F,$3C,$24,$7F,$94,$5F,$EA,$CF,$C2,$72,$FE,$5F,$1F,$9D,$1C
	.dc.b	$7D,$0E,$1E,$7C,$AD,$67,$FE,$6B,$A4,$F5,$F3,$F5,$5D,$62,$E1,$D2
	.dc.b	$6D,$F1,$9B,$5E,$9B,$57,$9F,$CF,$93,$E0,$34,$FA,$FC,$FC,$C7,$AE
	.dc.b	$2E,$3E,$A9,$B8,$F9,$E6,$D7,$79,$B9,$FC,$89,$B8,$A7,$FB,$C7,$BE
	.dc.b	$E9,$FE,$89,$83,$E7,$DE,$8F,$FA,$6A,$BF,$C9,$D3,$98,$C7,$B7,$FE
	.dc.b	$9E,$AD,$CF,$F0,$5A,$E9,$AF,$F0,$79,$CC,$A7,$2F,$33,$85,$73,$BE
	.dc.b	$B1,$C5,$F1,$61,$E6,$BE,$77,$C9,$B7,$C6,$F9,$9C,$4F,$25,$71,$BE
	.dc.b	$78,$66,$7C,$95,$C5,$A6,$70,$ED,$5C,$23,$D0,$68,$B1,$F1,$B0,$F2
	.dc.b	$E1,$DE,$7C,$38,$7F,$29,$F0,$3F,$96,$C7,$C3,$9F,$A3,$18,$BF,$8A
	.dc.b	$D8,$CF,$83,$FE,$E9,$F7,$B9,$36,$D5,$D1,$FE,$6D,$3A,$7F,$DA,$61
	.dc.b	$C3,$FE,$CC,$3D,$67,$FB,$CF,$83,$E2,$3C,$8C,$F1,$A3,$31,$CC,$CB
	.dc.b	$89,$E3,$3B,$9E,$A6,$F3,$FC,$1B,$DC,$CC,$F0,$AE,$51,$D0,$C9,$67
	.dc.b	$8B,$CF,$4F,$6D,$F7,$9C,$DF,$3A,$FC,$77,$0F,$F0,$AB,$F2,$6B,$FC
	.dc.b	$11,$B8,$FF,$01,$F1,$2D,$FE,$E6,$6D,$F1,$AF,$DC,$36,$7F,$88,$56
	.dc.b	$B8,$7F,$53,$5A,$6C,$FF,$15,$BD,$56,$D9,$5B,$72,$D3,$97,$77,$53
	.dc.b	$B7,$93,$63,$24,$BF,$E4,$51,$97,$F9,$24,$DC,$BF,$91,$CD,$AF,$E4
	.dc.b	$1E,$4F,$A4,$F5,$78,$16,$3D,$51,$70,$FE,$5D,$36,$FF,$96,$F7,$7C
	.dc.b	$13,$E1,$77,$DC,$3E,$14,$5B,$FE,$6D,$36,$B8,$D6,$8F,$91,$19,$7F
	.dc.b	$9A,$55,$AF,$E5,$93,$76,$FE,$8B,$F3,$FF,$44,$FA,$5F,$3F,$70,$FA
	.dc.b	$51,$6F,$FA,$34,$DC,$7F,$B9,$78,$4A,$FB,$EA,$6C,$61,$B7,$FE,$AD
	.dc.b	$6D,$FF,$72,$AE,$13,$FE,$B1,$F8,$AE,$67,$17,$04,$9E,$11,$96,$C7
	.dc.b	$FA,$C3,$E7,$FF,$7A,$DE,$7F,$CA,$F3,$FF,$8A,$7C,$2F,$CB,$DA,$F6
	.dc.b	$FF,$AB,$1B,$A7,$EC,$E9,$F7,$3F,$E3,$BC,$7A,$1F,$D4,$E3,$37,$D3
	.dc.b	$E8,$3F,$C4,$D3,$FE,$C2,$6C,$7F,$87,$25,$F4,$0F,$86,$E1,$C5,$30
	.dc.b	$E9,$2F,$98,$C6,$1C,$63,$13,$FD,$8F,$E7,$FF,$C9,$DD,$5F,$B2,$6C
	.dc.b	$66,$9D,$79,$7F,$5B,$B7,$F0,$FA,$DC,$7F,$8C,$1B,$19,$8E,$35,$CE
	.dc.b	$F9,$11,$ED,$76,$7F,$DA,$BF,$40,$F2,$7E,$7B,$6B,$DB,$E7,$E2,$64
	.dc.b	$7C,$0F,$95,$B7,$EF,$7F,$1B,$F2,$52,$7F,$DE,$DF,$25,$6B,$F1,$53
	.dc.b	$74,$FC,$AC,$DA,$FF,$39,$6A,$F8,$E7,$CE,$FC,$76,$CF,$CB,$5E,$C6
	.dc.b	$66,$4E,$5E,$11,$8C,$BE,$45,$72,$6C,$EA,$B8,$1F,$E1,$98,$7C,$FB
	.dc.b	$E0,$7F,$DE,$D1,$FE,$87,$32,$E1,$FF,$45,$46,$E7,$FA,$AC,$62,$74
	.dc.b	$F0,$8F,$F1,$36,$6F,$61,$F2,$FF,$A0,$C3,$95,$FF,$9B,$AB,$53,$FC
	.dc.b	$EB,$E1,$5F,$3E,$F8,$4C,$37,$2F,$E7,$7E,$1F,$2C,$3F,$AC,$5F,$0B
	.dc.b	$D9,$FD,$0B,$E1,$5F,$21,$8C,$3F,$A1,$C6,$1E,$7F,$2D,$FC,$2B,$E5
	.dc.b	$BF,$85,$B6,$32,$78,$23,$2D,$33,$AF,$F0,$B8,$CB,$63,$FA,$8F,$CF
	.dc.b	$FF,$73,$69,$FD,$59,$7C,$B5,$88,$68,$E2,$E2,$7F,$D1,$33,$F7,$F4
	.dc.b	$FF,$4A,$36,$BF,$D8,$1E,$62,$3F,$17,$1E,$2E,$4F,$01,$1F,$2F,$35
	.dc.b	$C3,$F1,$7F,$37,$F9,$53,$F2,$DF,$11,$6B,$F2,$D1,$67,$F2,$33,$72
	.dc.b	$FD,$E7,$42,$B1,$7F,$A2,$85,$62,$FF,$D1,$19,$7F,$13,$AF,$8F,$E1
	.dc.b	$FC,$5F,$17,$E3,$3F,$1F,$E7,$E4,$E8,$F9,$AF,$CB,$F9,$F9,$18,$3E
	.dc.b	$8B,$F3,$FE,$1F,$E8,$7C,$FD,$9F,$AA,$E1,$A7,$D8,$8C,$65,$F4,$27
	.dc.b	$EC,$7C,$76,$37,$FA,$2F,$DB,$F9,$F9,$87,$E0,$38,$2E,$1F,$62,$6D
	.dc.b	$FD,$89,$B5,$F6,$36,$56,$2F,$E2,$D5,$E3,$FB,$9C,$1F,$89,$3F,$15
	.dc.b	$F1,$B7,$1F,$C7,$C5,$BF,$C3,$CD,$CB,$F0,$F3,$6B,$C1,$36,$BE,$9F
	.dc.b	$CF,$FE,$54,$FC,$B7,$91,$6B,$F3,$71,$67,$F2,$73,$70,$FC,$9C,$DA
	.dc.b	$F6,$4D,$AF,$BF,$E3,$FE,$84,$FD,$17,$83,$6B,$F4,$F1,$67,$F3,$F3
	.dc.b	$70,$FC,$FC,$DA,$F9,$13,$6B,$F1,$7F,$3F,$FA,$D3,$F5,$DF,$3F,$6B
	.dc.b	$F6,$71,$67,$F5,$73,$70,$FD,$5C,$DA,$FA,$93,$6B,$FD,$BD,$DD,$FD
	.dc.b	$DA,$7F,$5E,$D8,$CF,$C0,$FF,$11,$F5,$A5,$7F,$10,$7F,$8A,$46,$27
	.dc.b	$FD,$C5,$EB,$FF,$94,$FC,$CF,$F2,$A3,$6F,$FC,$A8,$D9,$DC,$7F,$95
	.dc.b	$E7,$FC,$B5,$E6,$48,$FF,$2D,$79,$76,$BF,$DD,$D3,$75,$FF,$76,$CD
	.dc.b	$AF,$F7,$57,$89,$FE,$63,$3F,$CC,$9E,$25,$8F,$F3,$24,$5C,$3F,$DF
	.dc.b	$13,$6F,$FD,$EF,$36,$BF,$DE,$93,$69,$CC,$F9,$FF,$E4,$51,$FE,$9D
	.dc.b	$F9,$EF,$E2,$F1,$FC,$8A,$FF,$D4,$71,$87,$FD,$29,$F0,$48,$DF,$8F
	.dc.b	$87,$53,$C2,$FB,$DF,$37,$B1,$88,$E0,$EA,$AA,$F9,$FF,$EE,$3E,$BF
	.dc.b	$F5,$E8,$C7,$90,$8C,$4E,$A7,$FB,$1B,$5F,$F5,$CF,$C3,$3E,$7B,$FC
	.dc.b	$01,$1F,$DA,$23,$1E,$57,$56,$C6,$67,$6F,$FB,$0A,$BF,$C6,$DC,$FE
	.dc.b	$0F,$2B,$FD,$93,$3E,$06,$BB,$6B,$FD,$EE,$5E,$3F,$F7,$DE,$0F,$95
	.dc.b	$23,$E0,$46,$78,$38,$36,$3E,$5F,$97,$5D,$9F,$3C,$A7,$CE,$2D,$76
	.dc.b	$E2,$FC,$04,$F1,$C3,$FA,$25,$7F,$84,$DA,$B9,$7F,$DA,$4F,$D3,$F1
	.dc.b	$3F,$C2,$91,$FE,$D5,$B7,$E3,$0F,$A0,$CF,$2F,$BE,$EE,$1E,$DC,$7A
	.dc.b	$E3,$D1,$B5,$FD,$9B,$0E,$55,$E3,$BE,$2F,$84,$C1,$FE,$CA,$8D,$BE
	.dc.b	$79,$F9,$5E,$87,$DF,$F0,$BC,$2D,$9F,$9D,$8F,$F8,$5B,$8B,$FC,$47
	.dc.b	$1F,$ED,$4E,$8F,$F1,$FB,$AC,$39,$CE,$A3,$B0,$BE,$8E,$1B,$7F,$16
	.dc.b	$F9,$CF,$0D,$7F,$61,$C5,$F1,$31,$7E,$81,$F9,$BA,$F0,$08,$CF,$3F
	.dc.b	$CA,$BE,$33,$E7,$BE,$57,$C6,$3C,$B7,$C1,$D1,$FE,$4F,$FC,$BD,$F1
	.dc.b	$1F,$31,$B1,$98,$FA,$FD,$2E,$57,$CB,$EE,$65,$6C,$F3,$D6,$E7,$A7
	.dc.b	$C4,$FE,$DE,$58,$23,$0D,$8F,$DE,$35,$FF,$2D,$B7,$7F,$C8,$B7,$2F
	.dc.b	$F3,$23,$67,$FC,$F5,$93,$E8,$3C,$31,$EF,$C2,$BF,$63,$2F,$64,$FC
	.dc.b	$1F,$07,$DF,$4F,$32,$99,$BF,$91,$35,$FC,$09,$EC,$7E,$3F,$0F,$E9
	.dc.b	$B4,$F3,$C6,$DB,$3F,$1A,$75,$A7,$AB,$C2,$F0,$F9,$52,$BE,$B4,$7D
	.dc.b	$52,$38,$39,$78,$38,$7C,$7B,$A7,$18,$FE,$26,$7F,$D3,$D8,$FE,$4A
	.dc.b	$3F,$77,$DF,$F9,$B7,$5B,$BF,$2E,$F8,$57,$F7,$1A,$FC,$0A,$E1,$A7
	.dc.b	$FA,$55,$B1,$FC,$65,$AF,$F4,$5E,$AF,$C5,$FE,$5D,$FE,$DD,$8F,$F5
	.dc.b	$04,$E8,$93,$AE,$7F,$DF,$C3,$CC,$FC,$9E,$3D,$B2,$F3,$5E,$A6,$61
	.dc.b	$D1,$AF,$D6,$F8,$37,$91,$8D,$7E,$0D,$5D,$7F,$D5,$D5,$AB,$FC,$DF
	.dc.b	$E5,$CF,$92,$FD,$1B,$63,$31,$FC,$53,$F2,$EF,$F4,$DC,$7F,$AE,$DA
	.dc.b	$FE,$19,$B8,$FF,$2A,$6E,$1F,$CF,$F4,$79,$CF,$A6,$FD,$8B,$70,$E9
	.dc.b	$EE,$77,$33,$3F,$C0,$FD,$17,$CA,$52,$57,$AB,$E4,$7C,$DE,$5C,$8F
	.dc.b	$33,$DB,$E8,$E4,$7E,$EB,$24,$62,$3F,$8C,$60,$F9,$5F,$E6,$7F,$A2
	.dc.b	$7C,$79,$FF,$68,$B5,$FF,$6C,$07,$0F,$FB,$6D,$AF,$F5,$F7,$97,$07
	.dc.b	$F4,$E6,$EF,$BE,$5F,$E8,$3E,$13,$E6,$3C,$31,$30,$FF,$75,$5F,$07
	.dc.b	$FB,$55,$6D,$FF,$66,$F9,$9F,$1F,$C7,$91,$82,$1E,$E7,$2F,$C0,$8F
	.dc.b	$9D,$C5,$F3,$FE,$9E,$4F,$FB,$62,$32,$FF,$B6,$A3,$2F,$C8,$8F,$EB
	.dc.b	$3F,$4B,$F3,$23,$FE,$00,$2F,$9E,$C9,$AE,$19,$3A,$8C,$BC,$23,$2F
	.dc.b	$F1,$29,$E4,$EC,$D1,$FD,$28,$7D,$27,$F0,$1D,$FF,$C1,$D1,$8F,$1B
	.dc.b	$FB,$6C,$E6,$3F,$C2,$54,$CB,$35,$27,$A5,$6D,$56,$DF,$F7,$DF,$95
	.dc.b	$F9,$5F,$A7,$91,$82,$1F,$8C,$72,$FE,$3A,$3C,$7F,$07,$FD,$26,$CD
	.dc.b	$24,$FF,$90,$E3,$2F,$92,$38,$BF,$3B,$D8,$AF,$F3,$89,$DB,$67,$EA
	.dc.b	$CF,$F6,$DF,$C9,$BF,$B5,$73,$7F,$87,$57,$CF,$E4,$D7,$95,$30,$FE
	.dc.b	$EF,$E9,$AB,$71,$86,$BF,$E0,$17,$AF,$07,$D2,$DF,$FC,$83,$3F,$D3
	.dc.b	$1B,$8F,$F8,$C1,$B3,$F6,$3E,$95,$E8,$7F,$89,$27,$37,$FE,$49,$6A
	.dc.b	$6C,$F0,$8F,$C2,$47,$E2,$1F,$DB,$9B,$E4,$FE,$F1,$19,$8F,$DD,$B5
	.dc.b	$F4,$47,$84,$8F,$E3,$FD,$BF,$CF,$3E,$B8,$CB,$4D,$72,$FF,$8F,$3E
	.dc.b	$95,$EE,$52,$4F,$E6,$7F,$0A,$7C,$FE,$23,$FB,$8D,$70,$7F,$3C,$D7
	.dc.b	$FB,$AC,$F0,$7F,$45,$F9,$77,$F2,$E6,$F8,$9F,$E4,$C8,$C4,$7F,$7C
	.dc.b	$FC,$23,$D3,$C3,$FE,$60,$9C,$B6,$B8,$1F,$E5,$49,$DF,$0F,$F9,$43
	.dc.b	$F5,$5F,$E4,$35,$FD,$C7,$DE,$46,$27,$58,$FF,$39,$FA,$CF,$F2,$AC
	.dc.b	$E6,$3F,$C9,$8D,$AE,$31,$B3,$F1,$FE,$42,$B1,$1F,$CD,$9F,$E8,$9C
	.dc.b	$18,$C7,$FC,$ED,$5B,$AE,$35,$A7,$FC,$9F,$6F,$91,$7C,$97,$F9,$A1
	.dc.b	$B3,$A6,$BF,$E5,$FF,$D0,$FF,$97,$3F,$E6,$18,$CE,$3F,$05,$BA,$7F
	.dc.b	$9D,$1B,$3F,$E6,$9A,$E3,$5F,$1F,$47,$D3,$FC,$42,$57,$F9,$D6,$EB
	.dc.b	$FE,$6F,$9D,$BE,$A6,$BF,$E6,$FF,$83,$F6,$39,$AE,$37,$DD,$FF,$36
	.dc.b	$4E,$A7,$A4,$6B,$1F,$B8,$EB,$0F,$F4,$DF,$7B,$D2,$FD,$8B,$7D,$AC
	.dc.b	$C7,$ED,$3B,$DF,$9A,$7D,$DA,$D6,$5F,$E7,$E6,$BF,$E8,$C6,$E9,$FE
	.dc.b	$AD,$9F,$C1,$7E,$16,$F8,$AF,$F7,$17,$12,$31,$87,$EE,$A3,$13,$A9
	.dc.b	$CC,$FF,$74,$9C,$BF,$80,$8C,$DF,$E5,$DB,$B7,$FB,$4F,$27,$DD,$FF
	.dc.b	$9B,$6F,$86,$3F,$EA,$CB,$DC,$7F,$A8,$DB,$5C,$3C,$8F,$F5,$CF,$C5
	.dc.b	$FF,$89,$2B,$FD,$E7,$EC,$7F,$A1,$63,$11,$FD,$51,$AF,$FB,$8D,$FF
	.dc.b	$13,$B6,$39,$9F,$EA,$3F,$BA,$FC,$4E,$7C,$63,$33,$FD,$86,$B8,$E0
	.dc.b	$BD,$CF,$33,$55,$C1,$FF,$5B,$46,$95,$7F,$88,$74,$FF,$7B,$7D,$5B
	.dc.b	$E8,$BE,$37,$DD,$BE,$8F,$FD,$6C,$DC,$3C,$5E,$D7,$FB,$AD,$5F,$17
	.dc.b	$DD,$19,$8C,$47,$83,$DD,$FE,$EF,$18,$CB,$FC,$7F,$3E,$2C,$7E,$45
	.dc.b	$F1,$6E,$59,$9F,$95,$87,$29,$F9,$CD,$8F,$0C,$ED,$F7,$D7,$37,$89
	.dc.b	$9B,$F0,$BE,$AB,$EE,$D3,$99,$4F,$51,$C6,$FC,$0F,$6B,$E9,$7F,$17
	.dc.b	$97,$F9,$9E,$1F,$3E,$91,$89,$D7,$85,$E0,$8F,$E7,$32,$3F,$DB,$07
	.dc.b	$08,$FE,$BB,$F1,$79,$9D,$CE,$9B,$19,$9D,$CF,$A4,$79,$FF,$EA,$BD
	.dc.b	$3E,$AC,$77,$AF,$90,$EB,$8A,$BE,$29,$DD,$71,$D7,$AC,$62,$B7,$7C
	.dc.b	$AB,$57,$FE,$6A,$8C,$4E,$A3,$C9,$3A,$7F,$C7,$73,$A7,$D0,$78,$6F
	.dc.b	$F2,$BF,$19,$FE,$C3,$D7,$FD,$89,$3C,$0F,$B1,$C6,$E0,$56,$F2,$FF
	.dc.b	$3D,$FB,$B6,$FF,$C9,$2D,$BF,$D6,$63,$99,$FA,$53,$5F,$F3,$C5,$7F
	.dc.b	$94,$DA,$7E,$BA,$75,$32,$74,$F5,$E3,$FD,$74,$60,$95,$FE,$18,$39
	.dc.b	$9F,$E1,$A9,$F6,$BE,$88,$FE,$49,$F8,$5F,$F9,$BD,$FE,$BE,$D8,$FD
	.dc.b	$FF,$C9,$FB,$7A,$D3,$9D,$7F,$BD,$36,$FF,$1D,$3A,$F3,$BC,$D7,$C6
	.dc.b	$B5,$82,$B5,$7E,$2B,$F2,$CE,$6B,$C4,$FF,$9D,$2B,$6F,$FA,$AB,$8F
	.dc.b	$FA,$77,$B3,$F4,$5B,$87,$F9,$43,$BB,$E5,$E7,$FE,$FF,$9C,$DF,$F4
	.dc.b	$BE,$FF,$F3,$D5,$FF,$4F,$6C,$2B,$FA,$76,$BD,$67,$31,$FD,$CF,$DD
	.dc.b	$F8,$9E,$29,$FC,$07,$8A,$F9,$DC,$C4,$CB,$FB,$93,$A8,$C6,$5F,$DD
	.dc.b	$E2,$E0,$AF,$E9,$D7,$CD,$E2,$4A,$E0,$FF,$9B,$A3,$51,$83,$FC,$21
	.dc.b	$ED,$7D,$9F,$18,$FD,$BF,$21,$F5,$59,$89,$87,$F8,$6D,$D4,$62,$FF
	.dc.b	$DB,$FC,$9F,$9D,$F3,$3E,$DF,$49,$7D,$7E,$66,$63,$FE,$34,$78,$46
	.dc.b	$2F,$FC,$2D,$F5,$BF,$28,$C6,$DF,$E1,$F7,$17,$FE,$19,$72,$FF,$92
	.dc.b	$DF,$D6,$7B,$7F,$D3,$63,$FD,$78,$DA,$FB,$51,$FC,$87,$C0,$7D,$56
	.dc.b	$70,$FD,$FD,$DA,$FF,$60,$B6,$FF,$94,$36,$F8,$F8,$FE,$03,$F7,$9E
	.dc.b	$7F,$F9,$16,$FF,$D9,$8D,$BF,$09,$FC,$9B,$EA,$5E,$CF,$0E,$49,$7C
	.dc.b	$1F,$E1,$58,$AB,$73,$93,$FD,$B1,$F5,$3F,$92,$7F,$DA,$ED,$C1,$F8
	.dc.b	$13,$F0,$27,$D2,$FF,$11,$AD,$51,$A7,$1A,$E1,$3B,$F0,$66,$1F,$C2
	.dc.b	$EB,$F4,$B7,$FE,$E6,$39,$4F,$EE,$1B,$3F,$C4,$7E,$A1,$F1,$DF,$C8
	.dc.b	$5B,$1F,$D4,$F8,$BF,$3D,$5F,$17,$E4,$B5,$72,$BF,$DE,$D7,$C2,$BE
	.dc.b	$5E,$DF,$16,$F5,$1F,$EC,$9F,$1F,$F9,$87,$C2,$3E,$32,$C6,$1B,$8F
	.dc.b	$FB,$45,$B5,$FE,$D1,$7C,$DF,$0B,$27,$FD,$03,$F6,$DF,$E8,$07,$C7
	.dc.b	$19,$AF,$BA,$33,$FF,$68,$CF,$0E,$6D,$6D,$FD,$32,$73,$5F,$EC,$C6
	.dc.b	$BF,$2D,$AF,$F7,$73,$A5,$FF,$24,$EE,$FD,$77,$D6,$7E,$C7,$EB,$FF
	.dc.b	$64,$DC,$BF,$D4,$2D,$BF,$C8,$D7,$09,$FF,$52,$D7,$F5,$3F,$C2,$5F
	.dc.b	$31,$4F,$C8,$FE,$12,$F5,$F1,$7F,$5C,$F2,$57,$07,$18,$FF,$83,$63
	.dc.b	$0F,$D6,$AE,$53,$A8,$4E,$B5,$FC,$0F,$D7,$FF,$A1,$FE,$37,$FB,$53
	.dc.b	$FE,$89,$FA,$AB,$E7,$A6,$C6,$1B,$8F,$8A,$B9,$36,$95,$B8,$C3,$71
	.dc.b	$FD,$AD,$77,$9F,$5E,$CF,$A0,$FF,$14,$46,$1F,$F0,$D5,$7C,$C7,$2D
	.dc.b	$E9,$FF,$00,$46,$33,$FC,$ED,$F3,$9C,$D6,$9D,$F8,$9C,$3D,$1B,$8F
	.dc.b	$76,$D7,$76,$E1,$D3,$1F,$EB,$B5,$F5,$33,$FB,$87,$F7,$7E,$77,$BF
	.dc.b	$89,$FA,$5C,$1F,$11,$B5,$B9,$D6,$5F,$9C,$AD,$3F,$BF,$3E,$CC,$FE
	.dc.b	$FE,$33,$1F,$C0,$DA,$FE,$67,$E9,$7F,$B2,$5F,$E0,$5B,$1F,$27,$E1
	.dc.b	$BF,$C4,$91,$8C,$FF,$C6,$3F,$94,$36,$FF,$14,$37,$0E,$5E,$9E,$27
	.dc.b	$E7,$F8,$32,$7F,$27,$C8,$F8,$AE,$91,$FE,$B7,$E1,$FE,$15,$97,$F8
	.dc.b	$FB,$37,$D9,$F6,$E3,$F1,$B6,$FF,$31,$AF,$CE,$DF,$FC,$D4,$D5,$7C
	.dc.b	$A7,$77,$FE,$43,$CF,$C5,$85,$CF,$58,$7F,$08,$C7,$A5,$6B,$0E,$FF
	.dc.b	$35,$F0,$0F,$F2,$47,$E5,$0F,$AA,$C4,$E8,$98,$7F,$12,$9D,$46,$35
	.dc.b	$F9,$77,$E3,$FC,$2B,$F6,$77,$E1,$8C,$76,$FC,$EF,$93,$98,$7E,$DF
	.dc.b	$F5,$3F,$E4,$17,$E7,$9C,$B6,$75,$E2,$E3,$2F,$B3,$5D,$FE,$F3,$15
	.dc.b	$FE,$91,$D9,$A2,$7F,$CC,$9F,$94,$BE,$CE,$9F,$26,$9F,$25,$8C,$3F
	.dc.b	$85,$56,$DB,$5B,$9D,$69,$FC,$D2,$B4,$FF,$72,$B7,$C6,$7C,$2D,$3F
	.dc.b	$CF,$F5,$A7,$FA,$9C,$EB,$7F,$2F,$89,$33,$FC,$6C,$BE,$87,$19,$77
	.dc.b	$AB,$87,$2A,$D5,$FA,$27,$2F,$C3,$CB,$E1,$39,$3F,$32,$E2,$FD,$5C
	.dc.b	$9F,$18,$FC,$66,$1F,$E3,$8A,$F8,$71,$96,$9F,$A2,$8D,$4F,$28,$CB
	.dc.b	$FB,$BC,$FF,$0E,$DA,$FE,$F5,$18,$8F,$EF,$DF,$94,$FF,$B3,$AB,$F9
	.dc.b	$D3,$63,$E6,$5F,$F6,$F9,$CF,$C2,$FF,$A7,$EB,$53,$FE,$AC,$6C,$FE
	.dc.b	$0E,$38,$57,$F6,$A9,$CA,$F1,$9F,$07,$FE,$1A,$BF,$E9,$E3,$F0,$DF
	.dc.b	$4E,$1F,$CB,$A7,$57,$FE,$30,$AD,$67,$FE,$74,$F5,$BF,$D6,$D1,$83
	.dc.b	$FC,$D0,$E3,$0E,$71,$FE,$9A,$FD,$73,$FA,$13,$FD,$7B,$E2,$51,$8C
	.dc.b	$3F,$B6,$E3,$FE,$36,$7F,$C7,$F3,$AF,$0B,$FD,$51,$5B,$CF,$FC,$3F
	.dc.b	$39,$C3,$FC,$F1,$EE,$EA,$FF,$D8,$DE,$13,$D5,$A7,$FB,$2F,$EB,$DF
	.dc.b	$31,$8C,$3F,$B6,$56,$DB,$5B,$9D,$69,$FE,$16,$AD,$36,$75,$5C,$0E
	.dc.b	$7E,$37,$F8,$D3,$E4,$7F,$63,$5F,$ED,$88,$C7,$67,$FA,$53,$B6,$CE
	.dc.b	$6F,$FE,$00,$4F,$FB,$8A,$B3,$2C,$7F,$93,$D6,$A3,$EF,$70,$FE,$DB
	.dc.b	$1E,$F1,$BB,$E3,$F6,$2F,$9F,$7A,$5F,$ED,$D8,$C3,$63,$35,$FD,$5E
	.dc.b	$72,$FF,$59,$F9,$57,$D2,$3C,$0F,$F7,$34,$61,$B1,$99,$E0,$D7,$FC
	.dc.b	$C1,$5F,$E2,$8E,$DF,$17,$AE,$69,$27,$C5,$CD,$FE,$3B,$D7,$DB,$7F
	.dc.b	$E8,$29,$F8,$41,$C2,$8F,$A3,$DB,$FA,$2F,$A0,$E5,$C7,$FE,$45,$BD
	.dc.b	$D7,$4C,$33,$80,$E8,$E3,$0E,$51,$E0,$A7,$D8,$FF,$7D,$8E,$11,$87
	.dc.b	$E7,$46,$5F,$33,$63,$51,$FE,$62,$6F,$5F,$F9,$F2,$36,$7F,$8B,$3A
	.dc.b	$7F,$94,$3F,$E4,$39,$DF,$2F,$AE,$FF,$CC,$17,$C6,$73,$5F,$C0,$EF
	.dc.b	$8B,$FA,$ED,$DF,$01,$C3,$1F,$81,$D9,$F7,$A3,$14,$8C,$3F,$EE,$7E
	.dc.b	$4C,$C6,$23,$A4,$E2,$FA,$C6,$CF,$F7,$77,$17,$4C,$1F,$63,$F9,$09
	.dc.b	$FE,$20,$DD,$3F,$8A,$57,$8B,$6F,$E3,$6F,$19,$D5,$FF,$93,$9B,$7F
	.dc.b	$A4,$9F,$4A,$FB,$07,$CC,$7D,$36,$C7,$D9,$6B,$FE,$37,$6B,$FE,$B7
	.dc.b	$73,$8F,$D5,$6E,$1F,$E8,$F6,$CF,$E9,$78,$9F,$51,$E2,$97,$AF,$F3
	.dc.b	$3A,$1F,$4E,$F9,$AC,$37,$1F,$F4,$05,$F0,$6D,$FA,$48,$FA,$5A,$FF
	.dc.b	$AB,$23,$83,$63,$C1,$8B,$E7,$BC,$9F,$09,$E3,$9F,$F3,$C5,$7F,$90
	.dc.b	$9A,$F3,$8D,$BF,$1D,$F9,$44,$70,$7A,$1D,$BF,$E0,$9E,$EF,$CE,$6D
	.dc.b	$F2,$7E,$C3,$C1,$F6,$32,$BD,$F1,$F1,$CF,$AC,$47,$EB,$35,$F7,$3B
	.dc.b	$7F,$CB,$35,$FC,$7D,$FF,$54,$76,$FE,$49,$FF,$0D,$7D,$87,$B3,$0F
	.dc.b	$F0,$BB,$78,$BF,$D9,$15,$B9,$FD,$93,$C6,$BF,$DF,$0D,$77,$D1,$FB
	.dc.b	$D8,$C4,$7F,$09,$79,$56,$A3,$C7,$3F,$81,$FB,$1F,$F6,$4C,$79,$BE
	.dc.b	$C7,$F0,$98,$F0,$8F,$E5,$CD,$7F,$C7,$71,$F9,$CC,$0F,$84,$91,$FC
	.dc.b	$4E,$32,$D8,$E0,$47,$FD,$99,$F6,$3F,$EF,$F7,$FD,$03,$F6,$2F,$81
	.dc.b	$FC,$B3,$63,$31,$FA,$CF,$B1,$FF,$52,$C7,$FB,$55,$AF,$FA,$7D,$AF
	.dc.b	$FB,$CC,$38,$7F,$DE,$CD,$7F,$A6,$7B,$0F,$B0,$ED,$F1,$12,$7F,$D2
	.dc.b	$B3,$97,$FD,$9F,$F3,$8F,$A7,$FE,$0C,$D8,$CC,$7A,$BE,$0B,$D1,$EE
	.dc.b	$FF,$33,$CB,$F9,$1D,$79,$5A,$F9,$23,$F9,$E7,$7F,$FB,$EF,$EA,$FF
	.dc.b	$71,$91,$A7,$83,$DD,$FE,$1F,$F6,$3F,$2F,$07,$47,$F3,$BE,$DF,$95
	.dc.b	$18,$E9,$FE,$9C,$FF,$04,$E2,$E3,$F2,$72,$3F,$A6,$61,$CD,$F9,$4D
	.dc.b	$78,$B5,$FE,$B0,$F8,$70,$FE,$94,$DB,$DD,$7F,$24,$C3,$FC,$19,$7F
	.dc.b	$2B,$C9,$F9,$3E,$2C,$8C,$10,$F4,$1F,$E4,$DE,$3F,$E3,$AF,$D2,$8F
	.dc.b	$A3,$3F,$4B,$97,$E9,$FC,$AC,$AF,$AF,$19,$7E,$CC,$65,$EB,$E5,$30
	.dc.b	$FB,$D1,$E0,$3F,$93,$F0,$7F,$3C,$F1,$9F,$3F,$4C,$44,$9F,$EF,$5C
	.dc.b	$1F,$DA,$EB,$71,$8D,$FE,$8C,$62,$3B,$3F,$75,$D1,$FE,$03,$70,$EA
	.dc.b	$FC,$0E,$8F,$29,$7E,$C7,$F8,$93,$D3,$E8,$BD,$07,$F5,$48,$CE,$1F
	.dc.b	$DF,$DB,$87,$E1,$1B,$5F,$E1,$36,$BF,$D2,$34,$FC,$FF,$8F,$FA,$0C
	.dc.b	$24,$FE,$92,$32,$FE,$9A,$32,$FC,$ED,$3F,$A1,$3E,$7F,$84,$F9,$9F
	.dc.b	$11,$23,$04,$3F,$74,$31,$4F,$F8,$75,$B1,$FE,$23,$6B,$C1,$B9,$FB
	.dc.b	$E7,$E1,$1F,$6F,$D7,$FE,$FD,$F3,$1F,$DE,$B8,$4B,$E3,$F2,$6B,$C9
	.dc.b	$99,$7D,$6A,$DC,$EA,$31,$5F,$E0,$47,$FC,$C5,$E3,$7F,$81,$3E,$C5
	.dc.b	$EE,$53,$FC,$0F,$EB,$7F,$9A,$B5,$FF,$19,$37,$93,$F6,$ED,$C3,$FC
	.dc.b	$8C,$D5,$6F,$3F,$9F,$37,$F8,$4B,$DA,$7D,$15,$33,$7F,$89,$C3,$F2
	.dc.b	$4F,$F2,$76,$BE,$0A,$FF,$25,$46,$23,$C3,$1C,$9F,$D3,$47,$E6,$F2
	.dc.b	$FF,$25,$5F,$E6,$1F,$F2,$67,$D8,$BD,$DF,$F1,$2F,$D8,$BC,$EA,$7F
	.dc.b	$8A,$27,$58,$FE,$75,$AF,$EC,$DB,$7D,$2F,$9F,$C7,$7F,$8C,$3E,$C5
	.dc.b	$E4,$FF,$8C,$BE,$4F,$E0,$E3,$FA,$76,$BF,$BD,$9D,$BF,$AB,$6B,$99
	.dc.b	$E8,$73,$F0,$BF,$CA,$3F,$53,$FE,$6D,$F7,$CD,$25,$7F,$9D,$63,$2F
	.dc.b	$02,$1F,$D2,$BE,$23,$FA,$49,$1C,$1F,$D3,$AB,$0F,$EA,$62,$E8,$E6
	.dc.b	$FF,$44,$75,$FF,$19,$BF,$F3,$2F,$D1,$BC,$1F,$13,$71,$FD,$5E,$48
	.dc.b	$C4,$7F,$67,$8C,$4F,$03,$99,$23,$FD,$D9,$F5,$1F,$CE,$9F,$F7,$4E
	.dc.b	$8F,$98,$F8,$0D,$8E,$ED,$7B,$35,$FE,$A7,$7F,$2A,$7E,$69,$23,$F5
	.dc.b	$DF,$11,$E7,$7F,$DF,$95,$CF,$5F,$F7,$AB,$75,$FF,$45,$B6,$7F,$DA
	.dc.b	$F9,$F3,$7C,$F8,$5B,$F9,$51,$A3,$F9,$8F,$85,$FE,$CF,$EF,$FF,$57
	.dc.b	$36,$3F,$89,$B5,$FF,$07,$35,$FE,$DA,$E6,$7A,$B7,$1F,$F5,$F3,$65
	.dc.b	$BB,$EA,$BF,$C3,$17,$E4,$D5,$F5,$EF,$B0,$FF,$4A,$7D,$8B,$CC,$A7
	.dc.b	$FA,$92,$75,$97,$F4,$AB,$E3,$7F,$93,$9C,$CF,$DF,$6B,$F4,$9A,$FF
	.dc.b	$AD,$1D,$E1,$E3,$6E,$99,$AF,$1F,$99,$FE,$A0,$F8,$CF,$F1,$9D,$FC
	.dc.b	$CD,$5F,$98,$FE,$C2,$30,$DC,$3F,$C2,$75,$C1,$BE,$EF,$FB,$66,$3F
	.dc.b	$C6,$D9,$9A,$48,$F0,$12,$3F,$9A,$7D,$87,$E5,$70,$FF,$6E,$57,$59
	.dc.b	$FF,$77,$35,$E9,$5F,$EB,$FA,$E1,$7E,$4F,$67,$FC,$57,$4F,$B2,$FF
	.dc.b	$53,$66,$FB,$27,$D6,$FF,$B1,$DB,$5F,$DD,$73,$61,$CB,$0F,$F5,$74
	.dc.b	$FA,$E3,$AF,$7E,$2B,$FC,$95,$3F,$C6,$BE,$2F,$E2,$FC,$17,$F8,$E2
	.dc.b	$B5,$7C,$27,$CD,$BA,$FC,$D1,$C1,$C6,$3D,$70,$FF,$20,$BC,$AB,$35
	.dc.b	$CA,$BC,$07,$C6,$AF,$04,$73,$8C,$9E,$99,$F0,$61,$CE,$33,$8F,$AA
	.dc.b	$3C,$11,$AA,$ED,$1B,$E8,$FE,$5F,$5F,$1A,$FF,$4B,$3E,$39,$CE,$5F
	.dc.b	$49,$FE,$6B,$87,$BF,$E2,$B6,$FF,$9F,$DA,$FF,$97,$9A,$EC,$47,$25
	.dc.b	$1E,$57,$FC,$CB,$E9,$FF,$04,$BF,$AF,$1C,$5B,$19,$CB,$FC,$2B,$5C
	.dc.b	$6B,$E9,$4E,$5F,$21,$F2,$03,$97,$CC,$7D,$83,$C3,$C1,$F5,$1E,$3A
	.dc.b	$F8,$23,$11,$C3,$0C,$3D,$4F,$59,$DE,$BB,$3F,$06,$BF,$33,$B3,$E3
	.dc.b	$3F,$90,$EB,$FC,$8A,$BC,$75,$A9,$CE,$3F,$EA,$BF,$92,$B5,$FE,$8A
	.dc.b	$6F,$13,$8D,$B6,$AB,$F2,$95,$53,$4E,$33,$C2,$FF,$CB,$93,$A6,$CF
	.dc.b	$96,$3C,$51,$FE,$44,$7A,$B5,$E0,$66,$3C,$B9,$BD,$FF,$05,$7F,$8E
	.dc.b	$63,$0F,$F2,$7F,$B2,$FF,$3E,$B7,$D1,$F2,$FC,$AF,$C0,$ED,$56,$B2
	.dc.b	$FF,$4A,$C6,$2B,$CD,$08,$C5,$FF,$70,$AE,$13,$FD,$81,$AF,$EE,$67
	.dc.b	$6F,$FA,$53,$3F,$2B,$DA,$7C,$1C,$1F,$CB,$67,$2F,$F4,$EF,$84,$FE
	.dc.b	$89,$87,$F5,$A6,$DF,$FA,$7B,$C1,$F8,$B5,$BF,$03,$ED,$F3,$AF,$77
	.dc.b	$FD,$69,$AF,$F5,$86,$BF,$AA,$6E,$7F,$E2,$2F,$82,$F8,$DE,$31,$FC
	.dc.b	$2F,$8E,$FB,$2C,$C4,$D7,$FB,$DB,$A8,$C6,$3F,$E0,$C8,$BA,$7F,$4B
	.dc.b	$BE,$57,$C0,$95,$FE,$0C,$AF,$A3,$8F,$8E,$7F,$C2,$31,$8F,$21,$EF
	.dc.b	$FD,$B6,$CF,$F8,$73,$EC,$B1,$32,$3F,$C4,$F3,$AC,$3F,$C4,$CD,$7F
	.dc.b	$C3,$2D,$7F,$C6,$7F,$01,$F5,$A6,$F2,$7E,$9A,$3F,$CE,$71,$FE,$89
	.dc.b	$75,$3F,$ED,$D9,$D1,$BF,$66,$D7,$F8,$83,$EC,$BE,$69,$8A,$FF,$18
	.dc.b	$B8,$8F,$F1,$7B,$F0,$B8,$0D,$3F,$2B,$3F,$E9,$DF,$17,$F6,$DE,$FF
	.dc.b	$F5,$AB,$FF,$68,$F8,$2F,$B2,$C5,$FF,$B8,$EA,$F1,$FF,$07,$6D,$FF
	.dc.b	$20,$AF,$E1,$8F,$FB,$56,$33,$5F,$C8,$27,$2F,$FB,$B7,$D1,$FF,$29
	.dc.b	$57,$FB,$6D,$B1,$FF,$00,$30,$7D,$93,$DC,$FE,$19,$5A,$CB,$F8,$C3
	.dc.b	$70,$FE,$63,$4F,$9A,$DF,$17,$F1,$8C,$79,$C7,$F1,$5F,$B0,$FD,$2E
	.dc.b	$3F,$C8,$5B,$87,$F6,$BF,$B1,$7D,$93,$E8,$78,$77,$CA,$FE,$25,$7F
	.dc.b	$A4,$DC,$D2,$4F,$F0,$9F,$C1,$1F,$4D,$C6,$BF,$84,$C6,$2F,$FD,$F9
	.dc.b	$F6,$6F,$91,$FE,$54,$D8,$FA,$BC,$2F,$99,$CD,$49,$CF,$CB,$DC,$7E
	.dc.b	$55,$A3,$F9,$BF,$C6,$3C,$AB,$E6,$61,$B8,$7F,$B6,$9B,$83,$8F,$FC
	.dc.b	$79,$7C,$6B,$FD,$BF,$7E,$6F,$8A,$91,$FE,$51,$FA,$FF,$F4,$93,$EC
	.dc.b	$8C,$E1,$FE,$FC,$DB,$FD,$B1,$3F,$CE,$DB,$19,$9F,$CB,$C6,$31,$FE
	.dc.b	$AF,$F7,$7F,$06,$FA,$1F,$E2,$18,$CE,$FF,$AF,$FB,$4F,$D8,$7D,$A5
	.dc.b	$AF,$D8,$37,$2F,$64,$7E,$E2,$31,$7F,$E5,$9F,$15,$1F,$D6,$FE,$C5
	.dc.b	$F4,$54,$FC,$67,$DA,$7F,$B4,$30,$F1,$1F,$E3,$3C,$55,$C2,$33,$1F
	.dc.b	$E3,$78,$C3,$F9,$F9,$F8,$EF,$90,$D5,$76,$F0,$3F,$B8,$FD,$A7,$F3
	.dc.b	$7F,$95,$BE,$8B,$ED,$2C,$61,$B9,$F8,$2B,$F9,$F4,$63,$1E,$D1,$89
	.dc.b	$D5,$AF,$94,$61,$B5,$E5,$F6,$9F,$55,$FE,$51,$AF,$6C,$63,$C5,$FF
	.dc.b	$02,$CE,$6B,$83,$FE,$22,$AD,$B9,$E1,$FF,$44,$39,$9D,$9C,$AB,$9F
	.dc.b	$93,$D3,$DA,$B8,$72,$6D,$76,$6E,$1D,$1F,$B5,$AC,$AF,$CB,$7C,$1D
	.dc.b	$BF,$A0,$7D,$89,$FE,$4F,$C5,$F3,$6B,$FC,$20,$D9,$FB,$C7,$DB,$9D
	.dc.b	$9F,$84,$F8,$9E,$98,$FF,$B3,$B3,$71,$A7,$FC,$4D,$F2,$FF,$D9,$F2
	.dc.b	$FC,$2E,$28,$C3,$F2,$3E,$2F,$7E,$B7,$F8,$B7,$EC,$DC,$1F,$CF,$1B
	.dc.b	$87,$F8,$95,$B7,$E6,$0E,$1F,$CB,$71,$7D,$AF,$17,$EB,$9F,$E3,$EE
	.dc.b	$57,$C8,$F8,$B8,$3E,$AC,$FA,$3E,$D3,$C5,$CB,$73,$F4,$E7,$FE,$0E
	.dc.b	$7F,$44,$FF,$8D,$33,$F8,$ED,$73,$3E,$93,$FC,$53,$97,$82,$FC,$59
	.dc.b	$1E,$F8,$7E,$AA,$74,$F8,$3E,$8B,$E2,$1F,$E5,$1F,$B4,$7D,$86,$27
	.dc.b	$47,$C5,$9F,$90,$47,$07,$5F,$B3,$FD,$E6,$7F,$84,$D1,$5F,$16,$73
	.dc.b	$B7,$F9,$A6,$5F,$2F,$F5,$4F,$F3,$17,$DA,$7E,$F2,$75,$19,$C7,$EB
	.dc.b	$3F,$64,$8E,$0F,$17,$9F,$BC,$BF,$82,$6A,$CB,$FA,$44,$7E,$63,$ED
	.dc.b	$3F,$B0,$5F,$E2,$9B,$1C,$2B,$F2,$13,$9F,$5B,$FC,$ED,$5A,$AF,$CD
	.dc.b	$B5,$FE,$34,$D7,$FD,$9C,$D4,$F6,$C9,$5F,$ED,$3B,$FF,$70,$FC,$67
	.dc.b	$6C,$3F,$19,$F1,$DF,$22,$BA,$F8,$5E,$23,$F2,$16,$FA,$6C,$4F,$AE
	.dc.b	$7F,$C3,$39,$7F,$15,$8F,$F4,$13,$73,$FC,$F3,$5F,$5C,$F8,$35,$E7
	.dc.b	$1F,$2B,$1F,$93,$A7,$FA,$86,$FF,$0F,$87,$C9,$E2,$FD,$14,$BE,$0F
	.dc.b	$E3,$C4,$9F,$0B,$E4,$23,$83,$F2,$DE,$63,$E7,$78,$65,$F6,$A3,$13
	.dc.b	$FC,$03,$ED,$3F,$DA,$AD,$F8,$9F,$0F,$A7,$FE,$AB,$9C,$DF,$F1,$66
	.dc.b	$BF,$C4,$9A,$FF,$51,$6A,$0E,$3B,$9F,$11,$FE,$1D,$C3,$FA,$A8,$FB
	.dc.b	$4F,$87,$7F,$CF,$E7,$4F,$F5,$59,$D6,$9F,$E7,$EF,$77,$FD,$8B,$18
	.dc.b	$3F,$A1,$D7,$0D,$FF,$0B,$7F,$86,$C3,$AC,$63,$E2,$5F,$69,$FD,$33
	.dc.b	$0F,$EC,$F1,$8F,$3B,$FD,$17,$F0,$7F,$EB,$3A,$D6,$3F,$EB,$26,$E1
	.dc.b	$CD,$B8,$7C,$1D,$7D,$F7,$C7,$0F,$F6,$B7,$89,$FE,$1B,$C7,$FA,$0B
	.dc.b	$71,$F3,$DB,$EC,$9F,$3D,$C3,$DD,$99,$FF,$38,$C3,$83,$DE,$B6,$E6
	.dc.b	$77,$8F,$F3,$3A,$F0,$3F,$EE,$D8,$FF,$3A,$C6,$23,$FD,$DD,$E6,$7C
	.dc.b	$8F,$95,$FE,$91,$1F,$EE,$C8,$C7,$9D,$ED,$E2,$FD,$FD,$6A,$BF,$DD
	.dc.b	$F5,$A7,$86,$3F,$83,$9D,$61,$FD,$3E,$31,$87,$3F,$07,$85,$F2,$9C
	.dc.b	$BF,$8B,$FB,$37,$D2,$3D,$2F,$F7,$A4,$61,$B1,$99,$FE,$A5,$39,$7F
	.dc.b	$AA,$7C,$C3,$EA,$7F,$DA,$51,$86,$BF,$DB,$A3,$5E,$37,$FC,$01,$6C
	.dc.b	$74,$F6,$9D,$FE,$26,$C6,$63,$07,$9B,$A3,$FC,$6D,$A7,$F7,$AC,$BF
	.dc.b	$D1,$D5,$FE,$17,$3D,$D3,$A7,$FC,$79,$1A,$7F,$36,$68,$F8,$A1,$D7
	.dc.b	$07,$D3,$6E,$FF,$5F,$E8,$FF,$C8,$D3,$9F,$B4,$FF,$24,$74,$FF,$B7
	.dc.b	$FC,$8C,$DF,$DD,$6E,$3F,$46,$B7,$9F,$91,$FE,$FD,$19,$3E,$9B,$87
	.dc.b	$C8,$DC,$8F,$28,$E0,$7F,$88,$A3,$2F,$F8,$F3,$A3,$F9,$A3,$FB,$4E
	.dc.b	$1F,$F4,$93,$7E,$03,$97,$77,$DB,$7A,$76,$7F,$0E,$E5,$CD,$F0,$BF
	.dc.b	$F2,$F5,$ED,$E4,$7F,$08,$9E,$3D,$5F,$E7,$08,$FE,$2B,$F6,$9F,$8B
	.dc.b	$7F,$89,$37,$83,$F6,$F3,$B9,$FE,$31,$CB,$ED,$DD,$7B,$7F,$CB,$0D
	.dc.b	$96,$7F,$C9,$2B,$E2,$43,$ED,$1F,$39,$F1,$FA,$5F,$56,$ED,$C3,$6F
	.dc.b	$FC,$77,$7C,$1B,$7F,$4E,$31,$1F,$E3,$56,$BF,$E4,$C6,$BF,$E8,$B5
	.dc.b	$ED,$76,$F9,$D9,$3D,$A3,$CF,$3A,$8F,$F1,$35,$AB,$71,$FE,$5E,$9D
	.dc.b	$36,$B8,$91,$FF,$17,$74,$BE,$87,$AF,$C4,$7C,$1A,$FF,$40,$D7,$F8
	.dc.b	$CD,$AF,$58,$DB,$9F,$17,$81,$87,$47,$C6,$8C,$71,$FF,$46,$DB,$FD
	.dc.b	$05,$3E,$7F,$B3,$F0,$7D,$A4,$8F,$9B,$1F,$00,$C3,$FE,$26,$C5,$39
	.dc.b	$8D,$90,$E8,$7F,$92,$FC,$3F,$E3,$6F,$F9,$CB,$BF,$F2,$D8,$F8,$FE
	.dc.b	$CF,$DA,$DD,$3F,$0B,$3A,$E7,$FF,$60,$36,$DA,$30,$E3,$5F,$E2,$56
	.dc.b	$CF,$F7,$5A,$E0,$FF,$A1,$7B,$3A,$91,$FF,$30,$7D,$A7,$E7,$72,$FE
	.dc.b	$23,$5F,$EF,$D9,$FF,$32,$5F,$87,$1F,$06,$1E,$6F,$8F,$95,$A8,$FC
	.dc.b	$D5,$78,$2F,$94,$FF,$A8,$9B,$87,$6B,$E3,$3F,$EA,$6F,$49,$E8,$FE
	.dc.b	$8F,$ED,$1F,$05,$49,$5F,$E9,$CF,$B4,$7C,$4F,$EB,$1B,$1F,$E7,$BA
	.dc.b	$E2,$DA,$D5,$7E,$BF,$DC,$7D,$96,$2B,$A4,$E6,$3F,$87,$FC,$73,$E9
	.dc.b	$E9,$2B,$B7,$C2,$78,$7E,$03,$D0,$C8,$F3,$47,$7F,$AD,$91,$FE,$8F
	.dc.b	$C9,$1D,$BE,$03,$F8,$2F,$DA,$7C,$2A,$FF,$5C,$F6,$FA,$9F,$F6,$5B
	.dc.b	$79,$3F,$A3,$CE,$E7,$FD,$9E,$D8,$FE,$97,$3B,$EA,$FF,$1C,$57,$F0
	.dc.b	$0E,$0E,$FC,$72,$3F,$9F,$E1,$CD,$FF,$69,$B5,$E2,$D7,$F8,$EE,$FF
	.dc.b	$DB,$DF,$F7,$BA,$FB,$4F,$92,$7C,$AF,$3A,$4E,$A3,$2F,$08,$F9,$FC
	.dc.b	$5F,$43,$E6,$64,$F4,$8C,$BD,$A3,$2B,$5F,$F1,$6E,$1F,$72,$3F,$DC
	.dc.b	$3E,$3F,$CE,$6F,$D1,$FF,$32,$83,$78,$64,$E8,$87,$FB,$EA,$79,$36
	.dc.b	$39,$4C,$8E,$F1,$FC,$FB,$ED,$3F,$80,$C7,$1F,$A5,$F0,$19,$F4,$64
	.dc.b	$7F,$6E,$9D,$BF,$58,$38,$7E,$CB,$5F,$A3,$E4,$7E,$5A,$A4,$7D,$28
	.dc.b	$FC,$C9,$87,$E3,$71,$FF,$96,$73,$56,$A7,$FC,$F9,$A7,$F8,$4B,$07
	.dc.b	$AD,$F1,$FE,$56,$56,$B1,$F1,$CF,$BC,$CC,$7C,$79,$C3,$B3,$FD,$11
	.dc.b	$5D,$1F,$81,$3F,$E0,$B8,$C4,$7C,$13,$E0,$C7,$F8,$4B,$D1,$7C,$D7
	.dc.b	$F6,$D6,$FC,$8F,$F8,$12,$0D,$F2,$A5,$F8,$6B,$6D,$AD,$B6,$7F,$BA
	.dc.b	$EF,$FE,$5F,$D3,$FC,$67,$F6,$8F,$80,$FE,$F2,$DB,$F6,$1F,$E5,$8F
	.dc.b	$1F,$89,$FE,$25,$BF,$F3,$44,$7D,$B6,$BF,$B2,$6B,$FE,$AB,$7F,$26
	.dc.b	$F4,$F7,$5F,$4D,$FE,$08,$6C,$78,$27,$6F,$DC,$39,$73,$6E,$A4,$6E
	.dc.b	$3D,$A7,$94,$8F,$F8,$FA,$79,$BF,$91,$FB,$37,$BF,$FE,$18,$FB,$4F
	.dc.b	$F6,$84,$FC,$06,$BE,$11,$3C,$1F,$8A,$D7,$EE,$35,$ED,$3F,$E4,$B8
	.dc.b	$C4,$7F,$16,$FB,$47,$8B,$FE,$28,$F9,$6F,$83,$3D,$DB,$C1,$FE,$68
	.dc.b	$AD,$CF,$D4,$6B,$F8,$B6,$BF,$AD,$6B,$DE,$BF,$CB,$FE,$FB,$EA,$7F
	.dc.b	$65,$7F,$EB,$8F,$DA,$FF,$0F,$7F,$3F,$99,$18,$9F,$F6,$85,$70,$EB
	.dc.b	$FF,$73,$61,$CF,$0E,$B3,$FE,$69,$8C,$47,$E7,$DA,$FF,$01,$C1,$DB
	.dc.b	$FE,$C6,$8F,$E5,$BF,$13,$FD,$50,$FE,$AB,$E0,$58,$C9,$0F,$34,$FF
	.dc.b	$18,$CB,$FD,$C9,$F4,$EE,$FF,$F3,$5F,$E7,$FF,$E1,$CB,$FD,$33,$5F
	.dc.b	$F3,$5C,$ED,$FF,$74,$B5,$FE,$2D,$9A,$31,$1F,$D3,$A3,$15,$F3,$27
	.dc.b	$33,$F1,$34,$7C,$DF,$F9,$D3,$E7,$64,$FF,$BD,$A7,$2D,$9F,$92,$47
	.dc.b	$E4,$46,$9F,$1B,$72,$FA,$73,$FE,$8A,$8C,$6F,$FE,$DD,$C3,$13,$2F
	.dc.b	$B1,$FC,$1F,$E1,$BD,$8F,$F1,$9F,$8A,$FF,$18,$F6,$FF,$A4,$BE,$B7
	.dc.b	$F4,$39,$FF,$BF,$1B,$92,$FF,$D9,$3E,$E7,$F0,$B8,$C6,$1C,$A7,$F8
	.dc.b	$F3,$5F,$E8,$0D,$7F,$BB,$B5,$FC,$5F,$DA,$3E,$D3,$FD,$47,$F6,$AF
	.dc.b	$36,$9F,$EA,$89,$D5,$7F,$31,$E3,$46,$23,$F3,$55,$C5,$FE,$4C,$D7
	.dc.b	$FD,$57,$1B,$D1,$97,$8B,$1F,$D3,$CF,$3A,$FC,$03,$5D,$1F,$EE,$DF
	.dc.b	$67,$FD,$73,$F6,$BF,$B8,$BF,$8D,$D4,$8C,$57,$C9,$F9,$0F,$F1,$DB
	.dc.b	$FC,$35,$B9,$78,$A3,$FD,$5B,$3C,$2B,$97,$8B,$FE,$AE,$BD,$E1,$FE
	.dc.b	$20,$6D,$F5,$C7,$47,$FB,$33,$EC,$BF,$D5,$D1,$DE,$32,$FF,$3E,$9E
	.dc.b	$38,$2B,$79,$7F,$BE,$25,$EC,$FF,$84,$ED,$F6,$BF,$EA,$77,$FD,$87
	.dc.b	$F6,$AF,$B2,$FF,$6D,$B6,$3B,$57,$1A,$E1,$0B,$EB,$87,$E3,$67,$A6
	.dc.b	$4C,$39,$5F,$F1,$56,$D7,$97,$0E,$9F,$1B,$F0,$DB,$5F,$3F,$E0,$3F
	.dc.b	$C8,$B8,$F8,$32,$FF,$30,$4F,$7B,$E1,$85,$8F,$AD,$87,$27,$CD,$87
	.dc.b	$2A,$FA,$0D,$8D,$D7,$17,$D9,$5C,$6F,$BC,$7C,$8B,$E2,$EE,$3E,$0B
	.dc.b	$93,$98,$E6,$FE,$7F,$1A,$6D,$7D,$36,$DF,$EC,$BC,$5F,$F0,$77,$C6
	.dc.b	$5C,$3F,$C6,$0D,$7F,$CC,$ED,$7F,$78,$D4,$FA,$D7,$A3,$F9,$5D,$3E
	.dc.b	$CC,$70,$3D,$07,$F0,$58,$D6,$7E,$28,$C9,$CB,$8B,$A3,$D8,$CC,$F0
	.dc.b	$7C,$E6,$23,$77,$E3,$9D,$4F,$A8,$F9,$1B,$FE,$1F,$E3,$BF,$55,$9F
	.dc.b	$EA,$E7,$59,$7E,$33,$81,$5A,$CB,$CD,$39,$8C,$57,$8A,$31,$3F,$E2
	.dc.b	$48,$7C,$7E,$9F,$E4,$8F,$0C,$7F,$55,$A6,$67,$75,$C6,$B0,$6F,$8D
	.dc.b	$58,$9D,$1F,$E3,$DC,$BF,$D9,$2E,$2F,$F4,$5E,$FE,$DB,$19,$D3,$F8
	.dc.b	$56,$5F,$A9,$AF,$5F,$5B,$ED,$7F,$D0,$B7,$FD,$D5,$B8,$7F,$60,$6B
	.dc.b	$FD,$7D,$AF,$F8,$01,$DE,$1D,$B3,$F2,$CF,$F8,$DE,$FB,$7A,$5E,$DC
	.dc.b	$7C,$38,$70,$C7,$F9,$2E,$1C,$31,$FE,$41,$87,$09,$FF,$2E,$4E,$5F
	.dc.b	$E9,$BD,$FF,$D0,$DF,$E9,$AD,$CB,$FA,$6B,$5F,$E9,$AD,$78,$F7,$FD
	.dc.b	$5F,$01,$F6,$FF,$7B,$9D,$7C,$F3,$FE,$C1,$3F,$23,$C6,$7C,$1D,$3E
	.dc.b	$F7,$8A,$FB,$6C,$6B,$FD,$F6,$6E,$BF,$E2,$BA,$D4,$E4,$91,$FD,$F2
	.dc.b	$7F,$2C,$FF,$82,$5A,$B7,$FF,$1C,$46,$1B,$3F,$DE,$FE,$5F,$BE,$7F
	.dc.b	$80,$C1,$1F,$80,$F2,$7F,$D4,$75,$FE,$33,$6E,$3F,$E1,$E6,$BF,$E1
	.dc.b	$D6,$BF,$E3,$E7,$77,$F8,$CF,$96,$79,$3D,$7D,$FF,$D0,$46,$30,$FF
	.dc.b	$73,$F3,$FE,$8F,$CF,$FC,$06,$5F,$EE,$CF,$1F,$13,$2B,$67,$FB,$BA
	.dc.b	$31,$3F,$8B,$E1,$7D,$B7,$C1,$6D,$7F,$89,$A3,$04,$7F,$D4,$9F,$A6
	.dc.b	$EA,$B9,$4F,$ED,$63,$1A,$FF,$05,$F7,$FF,$AC,$5F,$FA,$D6,$FF,$D5
	.dc.b	$FE,$4F,$D2,$6C,$66,$77,$1F,$EB,$B6,$D7,$F7,$9F,$11,$F6,$FF,$C3
	.dc.b	$AF,$FD,$57,$5A,$7F,$93,$46,$6F,$FD,$1D,$E9,$7F,$93,$E7,$C4,$D8
	.dc.b	$FF,$64,$35,$FF,$75,$35,$FF,$1A,$1F,$E6,$1E,$3F,$E3,$38,$3E,$CD
	.dc.b	$F3,$9F,$C5,$EF,$A6,$7E,$79,$CE,$3F,$EC,$0F,$B5,$FD,$46,$5F,$E3
	.dc.b	$F6,$A6,$93,$C0,$FE,$1B,$F6,$AF,$A8,$FE,$49,$3F,$AB,$BF,$F4,$3C
	.dc.b	$E5,$ED,$5C,$2B,$6D,$8F,$F6,$8F,$DA,$3E,$3F,$E9,$36,$3E,$B3,$71
	.dc.b	$FF,$76,$35,$F8,$61,$D6,$7E,$FF,$95,$FE,$71,$18,$AF,$EB,$1A,$27
	.dc.b	$F9,$D7,$C4,$3E,$86,$F9,$58,$6D,$FF,$BB,$DB,$97,$FA,$A6,$B8,$4F
	.dc.b	$FB,$CE,$BB,$FC,$EF,$A8,$91,$EA,$6D,$7F,$B7,$F3,$7D,$3B,$E6,$B8
	.dc.b	$DF,$F9,$C6,$7F,$AE,$CE,$63,$E3,$DF,$F5,$89,$E3,$5C,$2F,$A5,$ED
	.dc.b	$FF,$54,$FD,$A7,$ED,$7E,$D3,$F6,$CD,$C7,$B1,$FE,$5A,$F0,$51,$9A
	.dc.b	$E3,$3A,$5E,$87,$FA,$F2,$FF,$B2,$7D,$9B,$E8,$FF,$54,$DB,$FD,$73
	.dc.b	$6F,$EB,$87,$0F,$DA,$75,$A7,$EB,$7E,$CF,$FD,$7D,$5E,$1A,$E1,$7F
	.dc.b	$A0,$9C,$C2,$BF,$CD,$9F,$69,$FC,$FB,$E4,$FF,$BE,$BF,$D0,$3E,$1A
	.dc.b	$FA,$49,$B1,$86,$E5,$EF,$AE,$4D,$A5,$6E,$30,$DC,$3C,$51,$FE,$A9
	.dc.b	$AF,$03,$F2,$75,$7D,$17,$FA,$B6,$30,$FF,$89,$EB,$FD,$E8,$E5,$BE
	.dc.b	$0F,$F8,$3E,$31,$9F,$F8,$01,$E0,$D9,$D4,$EF,$C4,$E1,$ED,$5C,$3B
	.dc.b	$B6,$BB,$B6,$FA,$3F,$73,$3F,$B6,$4A,$FB,$6D,$AF,$79,$FE,$48,$DB
	.dc.b	$FA,$23,$83,$EF,$47,$D7,$F3,$BA,$57,$E2,$F2,$5F,$0C,$7F,$C7,$5C
	.dc.b	$3F,$3A,$B5,$39,$7F,$90,$FC,$A7,$F7,$67,$F0,$8D,$CB,$E4,$FC,$47
	.dc.b	$F8,$DE,$31,$87,$F9,$17,$EC,$9B,$7F,$90,$38,$9F,$1F,$AC,$BF,$96
	.dc.b	$D6,$9F,$D0,$4E,$BD,$6F,$A5,$E1,$C9,$CD,$F1,$D3,$F4,$BC,$7F,$C3
	.dc.b	$DF,$F7,$56,$8F,$B4,$FF,$25,$67,$FE,$46,$9F,$8D,$39,$C3,$F3,$D3
	.dc.b	$9D,$7D,$1C,$9E,$7C,$B5,$87,$CD,$AD,$5F,$FB,$DF,$5F,$5E,$5B,$24
	.dc.b	$74,$C2,$DF,$1F,$9F,$F1,$9F,$E5,$DF,$B4,$7D,$76,$2F,$C7,$3F,$1F
	.dc.b	$EC,$9F,$0F,$9C,$3F,$03,$9A,$73,$7C,$0F,$92,$E2,$7C,$11,$8A,$F8
	.dc.b	$7F,$67,$23,$F7,$DF,$69,$FB,$CE,$0F,$B1,$39,$BE,$F8,$7F,$4D,$7E
	.dc.b	$14,$79,$7C,$DD,$57,$FB,$B7,$2F,$24,$63,$1F,$8D,$87,$1A,$FF,$44
	.dc.b	$7D,$A7,$F6,$4C,$7F,$30,$D8,$E5,$87,$E4,$64,$27,$57,$FD,$D1,$B1
	.dc.b	$FE,$4E,$F8,$BF,$81,$7D,$BD,$3F,$C8,$46,$2F,$84,$63,$5F,$F4,$76
	.dc.b	$1C,$B1,$FF,$79,$C6,$3C,$29,$7F,$94,$97,$D2,$E3,$3F,$4D,$76,$D7
	.dc.b	$1A,$7C,$3D,$3E,$06,$3E,$7A,$D3,$FC,$B7,$3F,$E4,$8F,$E8,$27,$78
	.dc.b	$7C,$09,$FC,$2C,$EF,$87,$F5,$58,$3E,$D3,$D3,$12,$7F,$53,$E4,$BE
	.dc.b	$A7,$18,$FE,$22,$31,$3C,$CD,$CE,$B0,$FB,$53,$AA,$FE,$19,$F6,$9F
	.dc.b	$EE,$7B,$FF,$1D,$36,$3F,$A6,$B5,$FF,$02,$B5,$FF,$14,$BB,$AE,$BF
	.dc.b	$0D,$FE,$BD,$8C,$57,$EA,$EF,$FD,$12,$FF,$39,$F8,$47,$DA,$7F,$B2
	.dc.b	$21,$F6,$9F,$13,$2F,$F4,$14,$E6,$31,$97,$F9,$DE,$31,$A7,$FA,$31
	.dc.b	$B7,$FD,$B7,$E1,$DF,$69,$FD,$1A,$30,$DC,$7F,$D1,$DF,$47,$FE,$BE
	.dc.b	$8C,$63,$FE,$AC,$6D,$7E,$5A,$F9,$F8,$7F,$E9,$AB,$E0,$FF,$B6,$FB
	.dc.b	$FF,$C3,$38,$FF,$37,$6E,$3F,$E8,$1F,$9D,$D7,$B1,$32,$FF,$34,$57
	.dc.b	$0B,$FF,$18,$B5,$FF,$2A,$B5,$FF,$73,$CF,$FA,$AB,$EE,$D5,$FA,$AF
	.dc.b	$95,$FF,$46,$D6,$A3,$E7,$4E,$A1,$3A,$7F,$DE,$7E,$3E,$63,$18,$23
	.dc.b	$1E,$7F,$ED,$BD,$17,$27,$EF,$6F,$75,$AC,$3F,$DF,$6F,$7C,$BD,$6D
	.dc.b	$8F,$5B,$5D,$C7,$F7,$57,$F8,$1F,$DA,$BE,$95,$E9,$7F,$A3,$63,$0D
	.dc.b	$8C,$CF,$F8,$8A,$72,$FF,$89,$3E,$61,$F5,$7F,$ED,$68,$C3,$5F,$F2
	.dc.b	$FC,$6B,$C4,$FF,$4D,$36,$3A,$7A,$EF,$92,$F1,$36,$33,$18,$3F,$C6
	.dc.b	$7B,$FC,$6E,$6F,$F4,$6E,$FF,$52,$31,$5F,$5F,$87,$E6,$70,$FF,$86
	.dc.b	$A3,$07,$F9,$11,$7D,$DF,$29,$DE,$16,$B4,$DB,$EB,$1F,$7A,$7A,$D6
	.dc.b	$7F,$1C,$74,$DF,$2A,$5F,$83,$05,$6A,$30,$D7,$A1,$FE,$58,$7F,$D4
	.dc.b	$AE,$A7,$B5,$48,$C1,$0E,$35,$DA,$3D,$19,$3E,$F3,$FD,$67,$D7,$FE
	.dc.b	$F3,$6F,$37,$C6,$6E,$1F,$17,$ED,$6E,$3F,$94,$9B,$87,$3A,$FE,$1D
	.dc.b	$18,$AC,$FC,$04,$9F,$E2,$4F,$0F,$84,$C4,$7F,$8C,$B0,$FC,$FC,$E3
	.dc.b	$2F,$D5,$47,$E8,$AB,$C3,$C0,$F6,$29,$30,$FF,$29,$FA,$75,$FF,$29
	.dc.b	$C5,$C3,$FC,$A1,$3E,$AF,$CA,$7F,$D5,$55,$FE,$5D,$6B,$E3,$6C,$F8
	.dc.b	$9F,$F0,$6C,$6F,$0E,$44,$79,$78,$B5,$E5,$16,$B9,$4F,$AB,$D3,$95
	.dc.b	$FE,$6A,$7D,$BF,$21,$88,$FF,$35,$C5,$AF,$F3,$44,$DE,$1F,$E0,$DA
	.dc.b	$2B,$17,$FE,$3A,$AF,$AF,$F9,$4B,$E3,$69,$27,$FC,$F9,$1F,$C5,$71
	.dc.b	$F9,$EE,$E3,$33,$B7,$EB,$46,$F1,$FE,$2C,$EE,$33,$3B,$AE,$7D,$1F
	.dc.b	$63,$CE,$95,$FE,$88,$7E,$DF,$B3,$88,$FF,$44,$C5,$AF,$F4,$34,$DC
	.dc.b	$7D,$F3,$6B,$FC,$AF,$3E,$EF,$56,$47,$B1,$F8,$3F,$41,$88,$F4,$C5
	.dc.b	$C3,$DF,$37,$0F,$5C,$DA,$FF,$63,$E6,$AF,$C8,$FA,$5F,$92,$F4,$A5
	.dc.b	$7F,$A8,$A3,$2D,$21,$FE,$A4,$F2,$FF,$32,$47,$FD,$52,$FE,$73,$E2
	.dc.b	$3E,$BF,$D8,$FE,$7B,$E2,$65,$7D,$68,$FB,$7F,$65,$23,$EC,$BF,$7B
	.dc.b	$E3,$BE,$FF,$D8,$C9,$FB,$AF,$EA,$3E,$E3,$F7,$71,$FA,$9F,$95,$BD
	.dc.b	$C6,$98,$FA,$4B,$3F,$ED,$99,$D3,$67,$4D,$21,$FE,$A9,$8D,$B6,$7F
	.dc.b	$73,$E9,$7F,$0A,$FD,$C7,$FB,$34,$FF,$01,$6F,$37,$FB,$16,$77,$3F
	.dc.b	$ED,$19,$CF,$37,$F6,$56,$C7,$FA,$62,$7D,$BF,$6A,$E1,$FF,$7B,$CF
	.dc.b	$F9,$82,$30,$E1,$AF,$FB,$F1,$AF,$F9,$98,$E1,$1D,$08,$FC,$3F,$57
	.dc.b	$D5,$F5,$5E,$BF,$B1,$93,$E7,$8C,$BE,$98,$CB,$EF,$8C,$BF,$06,$33
	.dc.b	$9F,$F2,$0C,$95,$FC,$CB,$DA,$7E,$EB,$FF,$00,$70,$FF,$18,$35,$FF
	.dc.b	$6D,$C6,$BE,$0F,$E7,$7B,$3C,$6B,$58,$FF,$85,$6B,$57,$CA,$F6,$FC
	.dc.b	$FF,$57,$EB,$7C,$D4,$8C,$10,$FB,$0E,$5F,$B5,$1F,$8E,$FB,$67,$C2
	.dc.b	$D2,$4F,$FA,$4B,$1F,$C0,$CF,$F8,$22,$31,$C7,$F4,$63,$8C,$60,$FF
	.dc.b	$4C,$B8,$7F,$DE,$CD,$9E,$7E,$DF,$69,$CD,$D9,$CC,$73,$1F,$6D,$F2
	.dc.b	$5F,$C0,$18,$CF,$F1,$2D,$3F,$17,$1B,$7F,$0C,$FC,$CC,$6C,$FE,$BA
	.dc.b	$31,$3E,$8F,$6B,$1E,$9C,$8F,$14,$A3,$F6,$BF,$85,$7F,$7E,$96,$3F
	.dc.b	$A6,$AD,$47,$0A,$E4,$EB,$1F,$CC,$5F,$1B,$FD,$6F,$DB,$FF,$45,$AF
	.dc.b	$80,$D7,$81,$FE,$C4,$71,$1C,$3C,$DF,$E0,$F1,$8B,$FD,$23,$6B,$87
	.dc.b	$9D,$AA,$DB,$23,$BB,$97,$E8,$7E,$3C,$F5,$3F,$89,$B6,$33,$1F,$C5
	.dc.b	$BF,$15,$7C,$1E,$A2,$4E,$8F,$B5,$3A,$7E,$E7,$CA,$3C,$CF,$E3,$ED
	.dc.b	$8C,$C7,$F2,$2F,$88,$FF,$26,$47,$C2,$13,$B7,$0D,$DB,$F8,$BB,$76
	.dc.b	$FE,$0D,$5C,$A7,$F7,$1F,$53,$FC,$74,$97,$FC,$67,$E2,$6B,$FC,$7E
	.dc.b	$2C,$FE,$FE,$7F,$92,$7B,$F2,$BF,$90,$3F,$C9,$FE,$03,$11,$FC,$9E
	.dc.b	$2D,$7F,$07,$9F,$F1,$9F,$B5,$2B,$FD,$80,$FF,$8E,$3E,$97,$CD,$9F
	.dc.b	$F9,$8A,$35,$3F,$E3,$1F,$81,$FF,$0E,$BF,$E3,$3A,$DA,$F7,$D0,$F8
	.dc.b	$FE,$84,$9F,$DB,$4E,$BD,$DC,$47,$39,$B9,$FF,$8F,$2F,$FC,$10,$D8
	.dc.b	$CE,$5F,$EA,$19,$ED,$5D,$1D,$C6,$67,$73,$E5,$9D,$BE,$9B,$F8,$1E
	.dc.b	$EC,$8E,$4F,$C3,$F8,$FC,$47,$BE,$2E,$3E,$59,$F9,$1F,$61,$2B,$C4
	.dc.b	$FC,$BF,$77,$11,$F1,$E2,$D7,$B6,$6E,$7E,$B9,$B5,$E7,$9F,$F3,$5F
	.dc.b	$E4,$7E,$8B,$FF,$2F,$B6,$FF,$D8,$4D,$AF,$A4,$FC,$26,$E9,$C8,$8F
	.dc.b	$D2,$8D,$6A,$79,$36,$77,$3F,$E7,$19,$D9,$A2,$47,$9B,$4F,$3E,$92
	.dc.b	$BB,$3E,$AD,$31,$1C,$E2,$E5,$EC,$9B,$97,$B2,$6D,$7B,$27,$FC,$2B
	.dc.b	$F1,$3F,$D2,$2D,$2F,$B1,$7D,$8F,$F7,$5A,$FF,$1A,$4E,$A7,$2F,$08
	.dc.b	$CB,$C6,$78,$78,$DF,$CB,$23,$06,$DC,$57,$28,$57,$28,$C3,$6B,$9D
	.dc.b	$F8,$5B,$5F,$ED,$9D,$2F,$8B,$F9,$4B,$F6,$4E,$4F,$05,$FB,$A7,$F3
	.dc.b	$31,$F0,$3E,$2F,$17,$FD,$52,$11,$87,$FA,$BB,$5F,$F7,$09,$D4,$B8
	.dc.b	$FA,$DF,$F7,$EE,$BE,$77,$89,$F0,$8F,$9F,$F1,$C8,$D4,$72,$CF,$FB
	.dc.b	$2C,$E6,$38,$1E,$72,$B9,$AF,$A7,$C3,$F9,$1E,$3E,$1A,$7E,$79,$B8
	.dc.b	$FE,$7B,$8E,$E3,$F9,$E9,$B7,$C6,$BF,$39,$18,$E3,$5D,$89,$3A,$3F
	.dc.b	$45,$18,$BF,$CF,$7C,$17,$13,$97,$0D,$AF,$D4,$45,$9F,$D2,$4D,$CB
	.dc.b	$F6,$B3,$6B,$F6,$B3,$6B,$F6,$B3,$72,$FD,$84,$DE,$5F,$F4,$55,$FF
	.dc.b	$1B,$AF,$6C,$7F,$89,$A3,$04,$BF,$E6,$53,$AA,$DB,$27,$58,$7F,$2B
	.dc.b	$9C,$57,$F3,$9F,$2B,$A9,$DB,$DE,$B7,$FC,$AE,$2C,$FF,$3F,$9B,$7F
	.dc.b	$CD,$7D,$FF,$29,$E6,$F7,$EE,$1F,$CC,$E2,$DF,$F4,$79,$B5,$C6,$B4
	.dc.b	$7F,$39,$8C,$BF,$CF,$6A,$D7,$F3,$99,$BA,$7F,$4E,$F7,$FE,$29,$F1
	.dc.b	$BE,$B2,$E1,$FD,$0E,$2D,$FF,$57,$9B,$8F,$F9,$A7,$CF,$FC,$4F,$AE
	.dc.b	$6F,$CF,$61,$98,$FF,$A3,$5C,$B4,$FF,$48,$7D,$6F,$23,$9F,$97,$63
	.dc.b	$24,$CB,$FD,$35,$19,$7F,$C1,$9F,$93,$7F,$86,$A7,$FC,$A0,$D8,$FF
	.dc.b	$18,$F4,$BE,$A9,$F0,$FF,$A2,$95,$6F,$5F,$F5,$3B,$70,$FF,$4B,$36
	.dc.b	$58,$73,$BE,$53,$99,$78,$6F,$8C,$EC,$DC,$3F,$CD,$D8,$3E,$9D,$AE
	.dc.b	$AE,$57,$F9,$A6,$F8,$61,$EA,$BD,$CE,$6B,$F7,$7F,$47,$FC,$EB,$CF
	.dc.b	$B7,$FE,$35,$AF,$F3,$E4,$FF,$3E,$F8,$AC,$D4,$9C,$D6,$CF,$F2,$BD
	.dc.b	$6E,$33,$3B,$9C,$9B,$DB,$F8,$B5,$6E,$33,$3F,$37,$EA,$1E,$B5,$3F
	.dc.b	$88,$ED,$F0,$7C,$CC,$65,$FE,$87,$AD,$B6,$B7,$AA,$B7,$8F,$F9,$2E
	.dc.b	$DF,$17,$DC,$FE,$3B,$EF,$BF,$02,$FA,$79,$75,$9D,$3F,$ED,$8A,$E2
	.dc.b	$D8,$FF,$73,$DF,$FB,$56,$72,$FF,$A6,$35,$7C,$B3,$D2,$FA,$2D,$C3
	.dc.b	$E7,$AF,$4B,$33,$27,$2F,$08,$C5,$FD,$0A,$E4,$D9,$D5,$70,$3F,$DE
	.dc.b	$DF,$23,$FB,$0F,$96,$FD,$8B,$63,$27,$FB,$EE,$32,$D3,$FD,$7D,$AB
	.dc.b	$F2,$47,$CC,$FC,$86,$CF,$F9,$3B,$C0,$7C,$BE,$6A,$4F,$2A,$DE,$7F
	.dc.b	$E9,$EF,$11,$87,$0A,$D5,$FF,$06,$F8,$2F,$E0,$FF,$2F,$67,$B3,$70
	.dc.b	$E5,$68,$C7,$57,$FB,$DD,$B5,$FE,$18,$C7,$BC,$62,$FF,$CD,$57,$C6
	.dc.b	$BF,$DA,$5F,$4C,$F1,$7F,$80,$36,$3E,$1F,$3B,$D0,$CD,$49,$FF,$02
	.dc.b	$60,$BE,$0F,$F8,$1D,$AF,$FB,$71,$AF,$97,$0F,$DC,$E7,$FD,$EF,$3F
	.dc.b	$8B,$D3,$FC,$A3,$BF,$F9,$4B,$71,$FA,$0D,$C3,$E1,$BF,$8E,$E4,$93
	.dc.b	$F8,$EE,$4B,$5F,$89,$9B,$8F,$ED,$39,$95,$8B,$FF,$1F,$F4,$7E,$6A
	.dc.b	$F1,$82,$3F,$3B,$58,$CD,$F9,$2B,$E2,$3F,$C0,$ED,$C7,$FD,$27,$F6
	.dc.b	$EF,$67,$35,$AF,$8A,$95,$FE,$98,$B7,$C6,$63,$4F,$C4,$43,$E0,$7F
	.dc.b	$C2,$79,$7F,$A8,$6F,$84,$63,$DC,$CF,$91,$27,$C3,$EF,$DC,$7F,$25
	.dc.b	$83,$DF,$E6,$67,$93,$E9,$46,$6F,$D3,$3E,$0E,$AD,$C7,$B1,$F7,$F8
	.dc.b	$72,$3D,$5E,$85,$C7,$D3,$37,$1F,$BD,$B7,$F6,$3B,$F8,$13,$97,$F3
	.dc.b	$11,$FC,$15,$E6,$68,$F9,$7E,$D7,$C2,$3E,$1F,$B3,$63,$24,$C3,$DF
	.dc.b	$19,$7E,$0C,$DC,$3D,$F3,$6B,$C3,$E9,$7D,$A3,$ED,$F4,$D8,$FA,$D1
	.dc.b	$70,$FA,$B3,$6F,$DB,$36,$BD,$73,$6B,$FD,$B5,$B7,$F8,$BD,$FA,$7E
	.dc.b	$CF,$DC,$F6,$7E,$EB,$2F,$F1,$2E,$5A,$49,$E2,$FE,$3A,$78,$47,$E2
	.dc.b	$3C,$4F,$C9,$9F,$94,$F1,$2D,$7E,$3E,$2C,$FE,$3A,$6E,$7F,$5A,$6D
	.dc.b	$7D,$49,$B5,$EA,$F6,$7F,$26,$7E,$53,$D9,$B5,$FA,$08,$B3,$F9,$09
	.dc.b	$B8,$7F,$BC,$7C,$7F,$F2,$89,$FE,$52,$FB,$AB,$87,$FB,$BE,$2D,$FF
	.dc.b	$29,$F6,$D5,$8B,$FF,$7D,$7D,$97,$F9,$68,$FF,$2D,$CD,$C3,$FD,$F1
	.dc.b	$16,$F9,$4D,$AF,$E2,$1F,$8E,$7F,$AF,$AF,$FD,$41,$D5,$FC,$A6,$3F
	.dc.b	$D2,$F7,$FE,$A0,$8C,$3F,$E9,$1F,$81,$46,$FD,$5C,$3A,$BE,$74,$F1
	.dc.b	$18,$74,$C3,$8C,$E6,$3C,$4F,$73,$18,$74,$74,$5A,$E7,$16,$78,$D7
	.dc.b	$49,$CD,$F8,$96,$4F,$AD,$E4,$D9,$D2,$FA,$DC,$CC,$9C,$BC,$23,$19
	.dc.b	$79,$C7,$D0,$FF,$1A,$CF,$FA,$2B,$63,$FD,$94,$D7,$FD,$DA,$7F,$BB
	.dc.b	$9C,$3F,$EE,$DE,$5F,$F0,$5B,$FE,$D1,$F8,$54,$62,$38,$9F,$EE,$67
	.dc.b	$0F,$1F,$2F,$C3,$E4,$7F,$31,$8F,$0B,$5E,$37,$FC,$E7,$C9,$D9,$FE
	.dc.b	$DF,$D9,$F0,$CE,$DF,$3F,$0F,$CD,$8C,$61,$F3,$5A,$FD,$9A,$E3,$E5
	.dc.b	$F0,$E4,$FE,$E9,$5A,$BE,$58,$67,$0F,$C3,$4E,$6B,$77,$9C,$1F,$43
	.dc.b	$FE,$2D,$BF,$B8,$FF,$8A,$DB,$E2,$6E,$35,$1E,$36,$E5,$F8,$57,$84
	.dc.b	$66,$78,$7A,$F2,$39,$90,$E4,$F8,$F2,$FF,$10,$C0,$FF,$0F,$38,$6F
	.dc.b	$AD,$F7,$BA,$7F,$BF,$4B,$C0,$C7,$47,$C4,$AB,$7F,$09,$BA,$7D,$26
	.dc.b	$DF,$19,$F0,$61,$D3,$C0,$C9,$23,$38,$7D,$0F,$47,$11,$E6,$9B,$5E
	.dc.b	$49,$EF,$C2,$F9,$BA,$4B,$E6,$7E,$EE,$1F,$1B,$98,$FF,$1F,$47,$17
	.dc.b	$7A,$7E,$26,$BA,$D7,$B3,$E8,$7F,$D4,$55,$A6,$BF,$E4,$76,$E3,$FE
	.dc.b	$25,$7F,$CF,$D1,$BC,$79,$11,$E5,$EA,$57,$94,$5A,$E5,$3E,$EF,$90
	.dc.b	$95,$E6,$CD,$1F,$0B,$DF,$CC,$FB,$22,$DF,$F8,$DE,$BF,$D3,$33,$C6
	.dc.b	$73,$B7,$F9,$46,$79,$CF,$5B,$FB,$7F,$3F,$FE,$C2,$8F,$8A,$D7,$FC
	.dc.b	$CC,$DC,$3F,$C8,$8F,$F2,$68,$CF,$A5,$C8,$8F,$2E,$9A,$E8,$EB,$1E
	.dc.b	$08,$FB,$BF,$43,$27,$E0,$3F,$80,$FA,$1C,$47,$DA,$8B,$87,$BE,$6E
	.dc.b	$7E,$D9,$B5,$F8,$39,$F6,$FA,$12,$3D,$AF,$C0,$F4,$31,$1E,$E8,$B8
	.dc.b	$7B,$A6,$E1,$FE,$79,$CD,$58,$BF,$E9,$50,$AF,$CB,$FC,$3F,$E6,$3E
	.dc.b	$12,$57,$DB,$8C,$B4,$87,$E5,$BE,$17,$F3,$E4,$7F,$04,$FE,$8B,$E7
	.dc.b	$3E,$BF,$A1,$FA,$4F,$85,$95,$F6,$23,$ED,$FA,$52,$3E,$DB,$F7,$BE
	.dc.b	$33,$EF,$FC,$64,$9F,$BE,$FE,$BB,$E8,$7F,$7F,$1F,$AF,$FA,$1B,$DC
	.dc.b	$69,$8F,$A1,$B3,$FB,$23,$81,$67,$83,$63,$6B,$91,$C9,$FD,$42,$B7
	.dc.b	$39,$8C,$5F,$78,$FD,$27,$D0,$FF,$12,$F8,$5E,$31,$FE,$C1,$AF,$E0
	.dc.b	$F3,$23,$04,$3F,$81,$BE,$EF,$A1,$70,$FB,$BE,$AE,$57,$AE,$72,$D9
	.dc.b	$FF,$56,$A8,$CB,$FD,$80,$C4,$6A,$33,$1F,$DB,$3C,$4F,$F6,$7F,$D0
	.dc.b	$7A,$7E,$86,$4F,$FB,$66,$3D,$86,$1F,$F6,$6C,$61,$D1,$FE,$DF,$75
	.dc.b	$B7,$FA,$DF,$45,$7F,$38,$F8,$E7,$D0,$BE,$13,$F9,$CF,$C1,$3E,$56
	.dc.b	$9B,$9D,$67,$FE,$F0,$6B,$F4,$C5,$F4,$BE,$07,$D3,$F8,$AF,$A9,$EC
	.dc.b	$7D,$5F,$8B,$97,$FE,$F5,$8C,$BF,$EF,$78,$CB,$FE,$EA,$8F,$C8,$FD
	.dc.b	$0B,$E1,$A9,$27,$F1,$D1,$FE,$10,$6C,$65,$F3,$08,$E4,$F0,$F7,$75
	.dc.b	$C1,$F8,$88,$D5,$63,$9F,$FC,$43,$7D,$67,$9F,$D0,$FC,$2B,$E1,$3F
	.dc.b	$C8,$EB,$7E,$5F,$EA,$2B,$51,$FE,$2A,$7E,$4D,$FF,$89,$1A,$E7,$5F
	.dc.b	$C5,$C6,$23,$FB,$C3,$FD,$AD,$F5,$FD,$0B,$E4,$7F,$6A,$D8,$FD,$84
	.dc.b	$EF,$2F,$63,$70,$F1,$7B,$1E,$53,$8F,$B5,$FB,$EF,$A1,$FE,$87,$87
	.dc.b	$C8,$6E,$1F,$15,$B7,$FE,$44,$7F,$A6,$CE,$FD,$3E,$44,$7F,$81,$46
	.dc.b	$AB,$94,$E6,$78,$1A,$F4,$BF,$87,$FD,$0F,$F7,$88,$FE,$15,$F4,$2F
	.dc.b	$80,$FE,$26,$D8,$FE,$1D,$3B,$7E,$FB,$70,$FB,$8D,$CB,$FC,$B6,$D9
	.dc.b	$FF,$12,$53,$E3,$7F,$8F,$FD,$0B,$D1,$FE,$43,$F4,$3F,$76,$FF,$C5
	.dc.b	$4D,$53,$B7,$F9,$03,$5F,$C8,$B7,$2F,$C7,$37,$2F,$F3,$6D,$7F,$97
	.dc.b	$E7,$F7,$1F,$29,$FC,$6C,$97,$FE,$16,$F0,$9F,$43,$88,$FD,$97,$8A
	.dc.b	$95,$66,$FF,$81,$5F,$F2,$5F,$83,$98,$73,$7F,$94,$72,$62,$39,$E1
	.dc.b	$FE,$54,$BF,$E1,$55,$A9,$FF,$13,$FC,$17,$F8,$79,$FF,$25,$FB,$DF
	.dc.b	$E4,$03,$FC,$87,$F1,$BE,$BD,$BF,$CD,$71,$99,$E2,$EB,$E1,$F7,$F2
	.dc.b	$B2,$7F,$C6,$76,$8E,$7D,$F9,$9F,$F6,$74,$5C,$FF,$9F,$F0,$7F,$AA
	.dc.b	$2B,$19,$7F,$97,$E3,$FD,$57,$58,$BF,$F5,$F6,$1F,$6E,$73,$1E,$09
	.dc.b	$D4,$73,$8F,$1F,$C0,$7B,$08,$F9,$BE,$56,$BF,$EE,$18,$B9,$79,$30
	.dc.b	$57,$C4,$F9,$59,$7E,$E7,$E3,$FC,$06,$23,$FD,$D5,$16,$F9,$4D,$CF
	.dc.b	$D9,$82,$B1,$7E,$CA,$FF,$39,$FC,$FF,$C9,$8F,$F2,$EB,$5F,$C3,$B7
	.dc.b	$0F,$92,$FC,$33,$18,$7C,$96,$CF,$FB,$95,$B9,$9E,$1A,$78,$09,$3E
	.dc.b	$0D,$6B,$E0,$8B,$87,$82,$7C,$FA,$CA,$F3,$BE,$AD,$71,$1E,$88,$B5
	.dc.b	$E8,$9B,$97,$9E,$6D,$79,$E7,$FC,$33,$F1,$DF,$D2,$6F,$FD,$45,$F4
	.dc.b	$0F,$A1,$FF,$6E,$CF,$F8,$DE,$F3,$18,$F8,$3F,$F7,$0D,$6E,$7F,$D5
	.dc.b	$55,$CD,$46,$54,$66,$30,$F5,$8D,$D7,$57,$B9,$D3,$CA,$FF,$11,$3A
	.dc.b	$31,$87,$FA,$E7,$8F,$ED,$5F,$85,$B5,$C1,$B3,$BD,$BF,$C2,$5F,$1D
	.dc.b	$70,$FF,$79,$B5,$FF,$73,$8A,$FC,$33,$FE,$C7,$8D,$34,$F2,$1F,$EC
	.dc.b	$BF,$91,$FF,$17,$E7,$F0,$63,$4D,$8C,$E7,$E2,$9E,$22,$32,$F9,$0C
	.dc.b	$B8,$E2,$FC,$77,$27,$F8,$F3,$3F,$CB,$B7,$87,$F2,$DC,$97,$0F,$CB
	.dc.b	$4D,$BE,$35,$FA,$28,$C7,$27,$F9,$E2,$30,$4A,$FD,$2C,$EA,$F6,$F0
	.dc.b	$E1,$95,$FA,$CE,$0B,$7F,$AC,$9B,$97,$EB,$26,$D7,$EA,$E6,$D7,$EA
	.dc.b	$E6,$E5,$FA,$B9,$B9,$7F,$87,$EF,$F7,$31,$EB,$E5,$FF,$46,$47,$F2
	.dc.b	$68,$C5,$E5,$91,$93,$F9,$5C,$63,$0F,$E8,$F5,$C9,$B5,$CA,$B8,$46
	.dc.b	$BC,$DF,$01,$E1,$F8,$1B,$3F,$CD,$A2,$CF,$49,$B9,$FF,$21,$F8,$2F
	.dc.b	$41,$E9,$F8,$2B,$87,$F3,$E8,$B7,$E4,$9B,$5C,$6B,$47,$F4,$48,$CB
	.dc.b	$FC,$8A,$AD,$7F,$3C,$9B,$AB,$37,$C1,$7C,$93,$E5,$7A,$F7,$1F,$E9
	.dc.b	$F1,$70,$F8,$B3,$72,$FF,$35,$7A,$3F,$8B,$F0,$3F,$CD,$18,$7F,$73
	.dc.b	$F4,$71,$32,$3F,$BB,$3B,$8C,$61,$FD,$B3,$E7,$F9,$1C,$FC,$9B,$7F
	.dc.b	$DF,$62,$CF,$F7,$29,$B8,$FF,$95,$7E,$83,$FC,$2D,$7F,$EA,$D6,$D7
	.dc.b	$F9,$CF,$A1,$F4,$2F,$88,$FD,$24,$AB,$78,$73,$C7,$F5,$18,$70,$F3
	.dc.b	$37,$5A,$CB,$F7,$D9,$7F,$88,$DF,$F5,$E7,$88,$F8,$DE,$AD,$86,$7E
	.dc.b	$3F,$A1,$6D,$AB,$95,$AD,$7F,$D6,$FA,$2F,$C1,$87,$2C,$7A,$3C,$F5
	.dc.b	$FD,$F7,$CE,$F6,$DA,$CF,$94,$FF,$4A,$47,$8B,$E2,$F3,$33,$3F,$01
	.dc.b	$E6,$9E,$4E,$29,$1C,$63,$0E,$A3,$A9,$DF,$3F,$F6,$7C,$65,$B0,$BF
	.dc.b	$03,$FE,$92,$FA,$17,$AB,$F5,$1B,$87,$87,$5F,$85,$E5,$E6,$A4,$E7
	.dc.b	$EB,$7F,$9B,$3F,$EE,$DB,$E3,$87,$E0,$67,$2F,$7F,$8D,$FF,$4E,$3F
	.dc.b	$50,$4A,$33,$1F,$62,$31,$5F,$ED,$98,$DC,$E9,$B1,$F1,$F3,$7D,$2B
	.dc.b	$E5,$7D,$97,$FC,$CE,$73,$1F,$63,$E6,$5E,$95,$31,$E9,$DC,$7E,$7C
	.dc.b	$E9,$B3,$A6,$93,$2F,$BF,$3B,$6D,$73,$7F,$CB,$7F,$29,$FB,$9F,$6F
	.dc.b	$F7,$4D,$BF,$F3,$63,$67,$71,$FE,$6F,$9F,$E9,$9F,$0A,$FA,$37,$D0
	.dc.b	$61,$B5,$FE,$75,$F9,$9C,$4C,$8F,$C6,$78,$EF,$8A,$DE,$5F,$EE,$9C
	.dc.b	$7A,$5F,$F5,$79,$CF,$07,$FB,$E7,$E0,$FF,$88,$70,$FF,$11,$6E,$5F
	.dc.b	$E8,$36,$E3,$FE,$78,$6D,$71,$9F,$F4,$45,$71,$69,$F9,$FA,$7D,$17
	.dc.b	$FB,$2E,$3F,$00,$D8,$FC,$41,$FB,$C7,$0E,$1A,$FE,$35,$E1,$5F,$BC
	.dc.b	$E6,$7B,$7F,$C4,$A7,$37,$FE,$09,$9C,$BF,$E9,$DF,$07,$F9,$77,$57
	.dc.b	$F2,$F6,$C6,$4F,$F5,$14,$65,$A7,$E2,$38,$FF,$1C,$7E,$3F,$8E,$C6
	.dc.b	$49,$9F,$E2,$23,$2F,$E2,$A6,$E7,$FB,$CE,$5F,$F6,$85,$F0,$8F,$F4
	.dc.b	$7F,$D1,$3E,$27,$FC,$1C,$D9,$FB,$FC,$3F,$B2,$AD,$D6,$BE,$2A,$57
	.dc.b	$E1,$2D,$F1,$78,$CF,$FD,$77,$1F,$EB,$A9,$E7,$EE,$F6,$F8,$2F,$C3
	.dc.b	$3F,$87,$F1,$A4,$F8,$39,$2E,$1F,$EC,$4D,$7F,$D2,$33,$E4,$F8,$0C
	.dc.b	$ED,$FE,$A8,$9C,$D6,$E7,$F2,$5E,$D7,$E4,$DB,$7E,$7E,$EF,$79,$F0
	.dc.b	$3C,$FB,$19,$25,$FA,$A3,$2F,$B2,$6E,$3F,$2A,$6D,$7F,$AD,$F1,$FF
	.dc.b	$1A,$5F,$A7,$DB,$FD,$27,$B3,$FA,$56,$57,$CE,$72,$D2,$4F,$17,$E9
	.dc.b	$CF,$08,$F9,$BE,$5F,$D8,$3E,$CF,$35,$AF,$A9,$16,$7E,$9C,$DC,$BF
	.dc.b	$03,$36,$BF,$D9,$DC,$4A,$C5,$BC,$0F,$E0,$1E,$D7,$EF,$4F,$DF,$7B
	.dc.b	$57,$0F,$C3,$C5,$AF,$B1,$37,$2F,$AD,$36,$BE,$34,$DA,$ED,$8B,$C0
	.dc.b	$FC,$99,$F9,$4F,$02,$DF,$E3,$E2,$D7,$E3,$A6,$E3,$F9,$C9,$B5,$EC
	.dc.b	$9B,$5E,$BD,$DE,$D7,$F1,$73,$F8,$C7,$B5,$6F,$F4,$31,$6B,$F2,$53
	.dc.b	$71,$FF,$72,$78,$7F,$E5,$B3,$FC,$B9,$F4,$57,$0F,$F8,$03,$16,$FF
	.dc.b	$CA,$F3,$6B,$FD,$D9,$F4,$5F,$E6,$33,$FC,$C9,$ED,$DC,$3F,$97,$A8
	.dc.b	$C5,$FF,$98,$22,$D2,$BF,$C2,$FC,$2F,$A2,$FF,$0C,$C7,$FA,$87,$AF
	.dc.b	$31,$FE,$97,$8C,$61,$F7,$63,$6B,$E0,$7F,$8F,$B9,$F4,$7B,$39,$BE
	.dc.b	$6E,$27,$78,$78,$6B,$15,$AC,$3B,$CE,$5B,$87,$82,$F8,$BE,$5C,$3B
	.dc.b	$1E,$53,$99,$E3,$9E,$79,$3E,$8B,$38,$7F,$99,$DB,$0F,$A2,$B1,$2B
	.dc.b	$8B,$A8,$C5,$F8,$07,$D1,$7F,$1C,$D3,$FC,$14,$D8,$FE,$96,$6B,$C4
	.dc.b	$FF,$61,$38,$8D,$3F,$D4,$3E,$19,$D2,$F2,$3F,$C2,$91,$86,$C6,$67
	.dc.b	$AC,$E5,$E9,$E4,$7F,$B7,$5B,$F0,$DE,$48,$C4,$7F,$5E,$C3,$F9,$E4
	.dc.b	$F9,$B7,$7E,$4B,$FE,$E8,$E3,$FC,$3B,$E8,$BF,$9F,$3A,$D7,$FC,$5D
	.dc.b	$C7,$F7,$2B,$85,$7F,$03,$6D,$6E,$3E,$EC,$73,$9E,$11,$CA,$FF,$2B
	.dc.b	$F2,$FF,$D3,$E3,$F0,$6D,$7C,$2D,$9F,$C5,$AF,$2F,$94,$FF,$13,$8D
	.dc.b	$57,$4D,$33,$8A,$73,$5B,$BC,$C6,$27,$55,$83,$33,$CB,$17,$D7,$7F
	.dc.b	$93,$F9,$FF,$43,$7F,$2B,$E9,$B1,$CD,$F2,$9B,$9F,$C5,$BB,$5F,$41
	.dc.b	$B8,$2A,$98,$8C,$FC,$74,$BF,$F1,$7D,$23,$8F,$CD,$66,$59,$79,$2B
	.dc.b	$51,$C2,$B6,$FF,$7A,$E9,$F1,$F8,$1F,$A7,$8F,$EF,$BD,$58,$99,$18
	.dc.b	$21,$FD,$C5,$F5,$FD,$0F,$FA,$9E,$3F,$C8,$6B,$D2,$C9,$AF,$4E,$4E
	.dc.b	$A3,$2F,$08,$CB,$EA,$E0,$FE,$33,$5C,$E3,$81,$B3,$FC,$27,$DB,$F3
	.dc.b	$3E,$E3,$FC,$77,$B7,$D4,$6D,$FF,$71,$9F,$79,$BA,$87,$22,$1E,$13
	.dc.b	$FA,$F4,$F2,$69,$C8,$FF,$30,$FE,$38,$F4,$BE,$EB,$FC,$9A,$39,$72
	.dc.b	$23,$18,$FE,$21,$E1,$3C,$AB,$8F,$44,$8C,$10,$E7,$3C,$B2,$FE,$7A
	.dc.b	$D8,$FE,$E8,$F4,$7C,$38,$FE,$EF,$0F,$89,$C9,$F1,$CF,$91,$D5,$2B
	.dc.b	$FC,$F7,$5F,$9E,$6C,$E5,$E7,$3C,$5C,$9C,$1F,$83,$E7,$7C,$2F,$3A
	.dc.b	$4F,$C6,$8C,$BF,$22,$32,$FB,$23,$2F,$BA,$33,$87,$C7,$9F,$BD,$D1
	.dc.b	$2B,$CE,$FE,$07,$9A,$46,$08,$7C,$27,$2D,$3D,$EF,$E2,$7A,$3F,$15
	.dc.b	$F1,$D2,$7E,$1B,$F7,$3C,$9F,$BB,$E4,$D9,$FC,$0C,$61,$FC,$1F,$C6
	.dc.b	$FE,$08,$8F,$CC,$F9,$FF,$E0,$6B,$E8,$1F,$2F,$87,$FD,$D3,$1F,$E9
	.dc.b	$96,$BF,$93,$38,$46,$CE,$1F,$09,$23,$84,$65,$E5,$19,$79,$4F,$81
	.dc.b	$FB,$FF,$43,$FA,$98,$91,$82,$1F,$8A,$72,$FE,$32,$3F,$61,$F4,$DF
	.dc.b	$B1,$B9,$3F,$8A,$8C,$BF,$8C,$8C,$BF,$94,$8F,$DC,$5F,$EE,$7E,$96
	.dc.b	$4F,$E4,$A3,$2F,$E5,$23,$2F,$E7,$23,$2F,$E7,$A3,$3C,$5F,$C0,$67
	.dc.b	$E7,$FD,$17,$F4,$1A,$FF,$80,$0D,$7F,$BB,$B5,$FF,$61,$B5,$11,$E1
	.dc.b	$3F,$E2,$08,$46,$3A,$77,$18,$F1,$FE,$E7,$27,$F1,$5E,$FF,$E2,$DF
	.dc.b	$21,$2B,$F5,$F1,$97,$F6,$31,$97,$F7,$31,$F8,$8F,$9F,$7A,$77,$D3
	.dc.b	$95,$FE,$DF,$9C,$B6,$72,$FE,$26,$3F,$C9,$0E,$48,$66,$77,$A7,$B6
	.dc.b	$79,$CF,$F1,$0F,$0F,$FA,$CF,$E2,$BF,$C8,$FC,$43,$7C,$2F,$F1,$65
	.dc.b	$7F,$24,$AD,$B0,$C1,$FE,$F5,$F9,$F7,$BB,$FE,$28,$8F,$9F,$18,$70
	.dc.b	$D7,$FC,$5A,$D7,$FD,$F2,$E7,$2F,$A8,$DE,$1F,$A6,$D9,$FB,$79,$FF
	.dc.b	$89,$2B,$ED,$7D,$07,$F8,$2E,$7E,$FC,$63,$E0,$FF,$6F,$26,$F9,$F2
	.dc.b	$B7,$8F,$E0,$2B,$51,$83,$F1,$2E,$1F,$DD,$3D,$DB,$3D,$AA,$47,$11
	.dc.b	$3B,$BF,$C8,$7A,$FF,$6A,$7F,$C9,$ED,$7F,$22,$7F,$BC,$38,$3E,$63
	.dc.b	$4F,$9D,$1C,$67,$9C,$FF,$14,$8C,$63,$F9,$AF,$8A,$FC,$0D,$7E,$7F
	.dc.b	$F1,$CF,$F2,$BE,$1F,$5D,$AF,$F9,$81,$AF,$89,$AF,$E9,$1D,$C7,$3F
	.dc.b	$63,$F9,$1C,$62,$FF,$53,$EF,$FB,$DF,$F4,$CF,$98,$F9,$1F,$F1,$5F
	.dc.b	$E2,$FF,$E4,$1F,$46,$F0,$FF,$84,$74,$FE,$BF,$5F,$E0,$33,$33,$23
	.dc.b	$FA,$EC,$7F,$81,$FE,$46,$47,$F6,$07,$FC,$15,$DD,$88,$FE,$D3,$17
	.dc.b	$6F,$F0,$64,$FF,$8E,$3C,$AF,$F1,$CF,$8F,$2F,$FC,$6B,$19,$7F,$C6
	.dc.b	$F1,$97,$FC,$07,$EB,$A7,$FC,$83,$E3,$FF,$90,$BC,$89,$5F,$E3,$E8
	.dc.b	$CB,$FE,$40,$8C,$BC,$A3,$2F,$48,$CE,$3F,$E4,$19,$CB,$FE,$43,$8C
	.dc.b	$F0,$7F,$9F,$A7,$FC,$A3,$E3,$FF,$94,$BC,$89,$87,$F9,$3A,$32,$FF
	.dc.b	$94,$23,$2F,$AA,$3F,$CB,$3E,$3F,$F9,$6B,$C8,$93,$FE,$56,$8C,$BF
	.dc.b	$E5,$88,$CB,$F0,$A3,$2F,$C4,$8C,$E5,$FE,$59,$9F,$F3,$3F,$DF,$BD
	.dc.b	$6B,$EB,$CC,$3F,$5D,$E3,$2B,$51,$86,$BF,$E3,$27,$FC,$BC,$D9,$EB
	.dc.b	$52,$38,$47,$57,$84,$65,$E0,$68,$EC,$FF,$9F,$3E,$FB,$FC,$51,$1F
	.dc.b	$E4,$07,$FC,$F8,$F9,$9A,$E7,$D5,$E0,$71,$F3,$24,$6E,$32,$F1,$8C
	.dc.b	$BC,$4F,$E0,$DE,$F3,$B7,$FD,$03,$0F,$9E,$FC,$73,$FD,$6A,$5E,$B7
	.dc.b	$F5,$69,$CC,$A7,$2F,$F6,$08,$CC,$75,$DF,$1C,$DF,$D7,$A3,$15,$DE
	.dc.b	$31,$7D,$5B,$1C,$0F,$16,$1E,$6C,$3A,$BE,$8F,$79,$F1,$7D,$63,$0D
	.dc.b	$CB,$FB,$9C,$EA,$F9,$C6,$27,$FD,$E3,$EF,$D9,$FE,$46,$D7,$FD,$B2
	.dc.b	$2B,$93,$86,$BF,$E0,$08,$F4,$3F,$E0,$2F,$69,$D3,$88,$D1,$F0,$4F
	.dc.b	$F0,$0C,$69,$C1,$E0,$CB,$E7,$A8,$D4,$62,$15,$D6,$73,$9F,$F8,$4B
	.dc.b	$91,$C2,$E5,$ED,$AF,$FB,$4A,$1C,$D8,$D3,$84,$6D,$FC,$94,$67,$36
	.dc.b	$1D,$30,$C3,$DA,$6D,$F1,$AF,$CE,$46,$2B,$2D,$BE,$A7,$ED,$3E,$FB
	.dc.b	$FC,$BA,$D9,$EB,$18,$AF,$F5,$2B,$D7,$1F,$F7,$5B,$96,$CF,$0E,$9E
	.dc.b	$2D,$BE,$87,$78,$CB,$4C,$DF,$68,$CB,$FC,$CB,$D3,$F7,$F6,$BD,$4B
	.dc.b	$18,$6D,$FF,$BC,$1B,$5F,$CE,$9B,$8F,$FB,$BD,$B3,$F5,$EF,$A3,$FE
	.dc.b	$F8,$F0,$BB,$F8,$5E,$06,$C6,$4F,$F7,$EC,$65,$A7,$FC,$00,$F0,$BC
	.dc.b	$FF,$27,$E8,$6C,$64,$FE,$47,$69,$D4,$61,$B9,$7F,$52,$6E,$5F,$97
	.dc.b	$E1,$5F,$1B,$E9,$7B,$6D,$6E,$75,$3F,$D7,$3E,$AB,$E5,$FD,$57,$CC
	.dc.b	$6D,$F9,$1B,$3B,$8F,$3C,$FE,$43,$C9,$E7,$87,$FB,$E7,$2C,$61,$96
	.dc.b	$5F,$FB,$59,$CB,$4F,$F6,$CD,$FE,$77,$C9,$7E,$37,$FE,$7A,$9C,$9F
	.dc.b	$E3,$58,$CE,$5E,$1A,$E1,$5B,$8F,$F7,$4C,$61,$FD,$40,$FC,$65,$EB
	.dc.b	$61,$B8,$7F,$90,$7A,$9F,$36,$FA,$BD,$5C,$AF,$F3,$05,$F0,$C3,$FC
	.dc.b	$72,$75,$9F,$F4,$73,$FE,$8C,$8D,$3F,$E3,$EF,$BC,$7B,$DE,$46,$DF
	.dc.b	$F9,$47,$F0,$E7,$E1,$CF,$A9,$D5,$CA,$FF,$38,$5F,$0C,$3C,$2D,$7B
	.dc.b	$4F,$89,$7D,$03,$E3,$BF,$91,$DF,$F9,$7E,$31,$AF,$F0,$D7,$51,$67
	.dc.b	$E1,$3B,$78,$D3,$97,$F9,$45,$72,$AD,$1C,$6F,$93,$4C,$9C,$DF,$AB
	.dc.b	$F4,$1F,$8A,$8F,$F5,$9B,$6D,$F3,$AF,$B3,$7C,$6F,$0E,$89,$87,$FA
	.dc.b	$4F,$0F,$CE,$E1,$FE,$6C,$6B,$D2,$7E,$27,$67,$FA,$89,$FE,$53,$F0
	.dc.b	$AE,$BB,$18,$6E,$1F,$E1,$89,$D3,$67,$4F,$C5,$F8,$4F,$91,$F4,$BF
	.dc.b	$25,$B1,$93,$E5,$46,$5A,$67,$1F,$98,$FF,$B6,$B7,$FB,$13,$B9,$C9
	.dc.b	$F7,$38,$51,$FD,$0B,$C3,$FA,$BC,$5F,$C1,$5F,$B3,$5C,$35,$FE,$3B
	.dc.b	$8F,$F8,$CA,$73,$5B,$8E,$5C,$5C,$DA,$7F,$A7,$7E,$7F,$BB,$71,$FE
	.dc.b	$A9,$C7,$F5,$7E,$77,$F5,$F5,$AC,$FC,$E1,$C3,$FE,$A0,$6B,$FE,$F2
	.dc.b	$6B,$FE,$FA,$6E,$7F,$CF,$FE,$C3,$DD,$F0,$7E,$F6,$E5,$EB,$6D,$FB
	.dc.b	$F5,$46,$35,$F8,$8D,$AF,$EE,$D3,$A7,$E2,$5F,$CB,$F9,$4F,$99,$F0
	.dc.b	$36,$7E,$43,$71,$EC,$DC,$7E,$83,$F4,$FE,$1F,$EA,$76,$5B,$FA,$0D
	.dc.b	$BF,$33,$6F,$EC,$3F,$B1,$FA,$1F,$D9,$7C,$D5,$BF,$D9,$36,$FF,$07
	.dc.b	$CB,$FB,$D8,$C6,$1F,$B4,$E5,$46,$2F,$F1,$91,$87,$F7,$98,$76,$7F
	.dc.b	$CE,$BF,$8C,$FF,$11,$6E,$5F,$8D,$F8,$8F,$F0,$7C,$E7,$3F,$C6,$3F
	.dc.b	$EE,$FA,$E3,$3A,$D3,$6F,$E5,$7E,$7E,$4E,$8F,$C9,$3F,$99,$E9,$91
	.dc.b	$F9,$5E,$8B,$8F,$DB,$9F,$CF,$7D,$2E,$37,$FC,$F3,$8B,$FC,$E3,$97
	.dc.b	$EE,$BE,$8B,$F4,$B6,$3D,$ED,$BF,$5B,$6B,$FB,$76,$68,$C5,$7F,$8F
	.dc.b	$63,$0F,$C8,$BE,$AF,$EB,$7E,$83,$F6,$9F,$41,$FB,$56,$C6,$4F,$D9
	.dc.b	$46,$5A,$67,$1F,$CD,$46,$5B,$1F,$5B,$E8,$BE,$9F,$07,$D4,$6C,$64
	.dc.b	$FB,$71,$96,$99,$9F,$B3,$19,$6C,$7F,$05,$9F,$E1,$93,$FC,$35,$B1
	.dc.b	$93,$F8,$54,$65,$A6,$67,$F6,$51,$96,$C7,$E3,$BE,$2F,$F1,$7F,$37
	.dc.b	$F8,$C6,$C6,$4F,$C9,$C6,$5A,$66,$7F,$23,$19,$6C,$7F,$1D,$E4,$FE
	.dc.b	$49,$F4,$1F,$C9,$5B,$19,$3F,$91,$46,$5A,$7F,$B5,$BF,$60,$FE,$56
	.dc.b	$FF,$9E,$DB,$5F,$E7,$27,$F9,$4C,$64,$FF,$39,$BA,$ED,$34,$DC,$F1
	.dc.b	$9D,$F6,$F0,$EC,$B1,$E1,$6C,$FF,$39,$A4,$63,$8F,$FC,$85,$18,$AF
	.dc.b	$F7,$6F,$E3,$1F,$EC,$87,$FD,$49,$F3,$DF,$E9,$38,$C6,$1F,$E9,$88
	.dc.b	$C7,$C2,$FF,$A8,$7E,$09,$1B,$F8,$6C,$37,$0E,$77,$FE,$92,$C3,$9D
	.dc.b	$F0,$8C,$61,$C1,$D6,$17,$55,$F0,$C7,$A5,$78,$CE,$38,$F6,$AD,$DF
	.dc.b	$1C,$9E,$0D,$F0,$70,$DB,$E1,$3A,$6C,$E9,$A7,$AB,$07,$D0,$7C,$CF
	.dc.b	$57,$FC,$77,$18,$E0,$FF,$60,$3F,$E3,$E9,$D3,$63,$28,$FF,$77,$7C
	.dc.b	$9F,$F8,$BD,$FF,$22,$FC,$4A,$31,$1B,$3F,$DF,$CE,$1E,$3E,$27,$8A
	.dc.b	$FD,$FF,$11,$75,$FF,$71,$57,$8D,$AF,$33,$D6,$E1,$FF,$4E,$1D,$F8
	.dc.b	$FE,$B7,$07,$FA,$23,$4C,$70,$FF,$A8,$5B,$5F,$EC,$EF,$77,$FD,$E7
	.dc.b	$3C,$A3,$EA,$47,$C0,$BE,$6D,$9D,$4F,$DB,$B7,$BD,$F6,$23,$EE,$35
	.dc.b	$D3,$6F,$CA,$B8,$7C,$75,$CB,$2B,$3B,$C3,$FD,$DC,$DC,$3F,$DC,$AF
	.dc.b	$17,$BC,$F0,$23,$E2,$BE,$B8,$BE,$2F,$EA,$73,$FF,$A9,$FE,$2E,$F2
	.dc.b	$7F,$B1,$5B,$87,$C5,$AB,$A7,$FA,$C9,$B8,$2A,$98,$8C,$FC,$F4,$BF
	.dc.b	$F6,$0B,$C3,$E7,$B1,$1F,$E3,$7B,$4E,$32,$FF,$5D,$57,$83,$E8,$1F
	.dc.b	$1F,$7E,$42,$63,$F9,$3B,$FF,$63,$56,$A3,$1A,$FE,$5F,$54,$E6,$7F
	.dc.b	$C8,$CA,$F9,$C6,$30,$F7,$FC,$43,$C3,$BE,$1C,$AE,$A3,$15,$6E,$33
	.dc.b	$18,$AE,$11,$FE,$55,$95,$76,$99,$3C,$27,$2D,$9C,$BE,$8F,$81,$FA
	.dc.b	$7F,$0D,$FA,$3D,$7F,$DC,$CD,$D7,$E0,$6B,$FE,$C2,$BF,$1B,$5F,$F3
	.dc.b	$34,$ED,$F5,$37,$3F,$8F,$87,$D1,$3B,$CC,$39,$BF,$84,$F9,$BF,$F4
	.dc.b	$BC,$7C,$07,$F9,$33,$FE,$73,$6E,$3F,$EF,$B7,$71,$9F,$77,$97,$44
	.dc.b	$8E,$51,$D4,$C3,$CE,$30,$FC,$EC,$7F,$8A,$DF,$C2,$E4,$F8,$A7,$C6
	.dc.b	$EA,$95,$B8,$CB,$C6,$32,$FC,$48,$CB,$F1,$A3,$DD,$F4,$1E,$FF,$0E
	.dc.b	$4F,$BE,$32,$FC,$18,$CB,$F0,$63,$2F,$C3,$8C,$DF,$C3,$9F,$B7,$D1
	.dc.b	$2B,$D8,$FD,$EE,$79,$18,$21,$F6,$9C,$B4,$F8,$8F,$E1,$B9,$FF,$0F
	.dc.b	$F2,$12,$7E,$43,$F6,$7E,$7F,$ED,$7D,$05,$9F,$B5,$18,$7E,$E7,$C1
	.dc.b	$FD,$F2,$3F,$47,$E6,$7F,$83,$AF,$99,$7C,$86,$17,$C8,$D3,$11,$27
	.dc.b	$F5,$67,$18,$D3,$89,$E6,$D9,$CB,$FE,$B0,$8E,$AE,$4F,$C2,$7C,$CF
	.dc.b	$E9,$A2,$46,$08,$7E,$89,$CB,$FA,$48,$FD,$6F,$CC,$FE,$BA,$A4,$FE
	.dc.b	$3E,$32,$FE,$46,$32,$FE,$B2,$3F,$6B,$7F,$B6,$F9,$59,$3F,$97,$8C
	.dc.b	$BF,$99,$8C,$BF,$B4,$8C,$BF,$B6,$8C,$F1,$7E,$FA,$7E,$7F,$D0,$7F
	.dc.b	$50,$AF,$8D,$C1,$F3,$CC,$F8,$F2,$3F,$BF,$CE,$DF,$ED,$41,$C3,$FD
	.dc.b	$B5,$AF,$F2,$9E,$8F,$E3,$AF,$EB,$FE,$0B,$F8,$89,$FC,$4B,$C2,$91
	.dc.b	$92,$1F,$C3,$9F,$C2,$FC,$F3,$E8,$2F,$CF,$4A,$FC,$0C,$E5,$B3,$97
	.dc.b	$FD,$DB,$09,$E7,$32,$78,$C6,$F5,$78,$9F,$E1,$1B,$E9,$5F,$C4,$FC
	.dc.b	$2F,$EB,$7F,$17,$FC,$7C,$FE,$41,$E1,$FF,$25,$C3,$F8,$B4,$EA,$A1
	.dc.b	$FC,$75,$FC,$FF,$D0,$3E,$0B,$DB,$E2,$CA,$FC,$DC,$E5,$B3,$FE,$30
	.dc.b	$51,$97,$E9,$B5,$FF,$0F,$35,$FF,$15,$7D,$07,$F7,$F5,$87,$B9,$FF
	.dc.b	$1A,$43,$E1,$BF,$66,$D8,$FD,$6B,$C6,$78,$42,$31,$AF,$6A,$E2,$72
	.dc.b	$F3,$E5,$72,$8C,$BD,$23,$F8,$0F,$AF,$F5,$63,$F7,$7F,$0D,$FD,$BF
	.dc.b	$1E,$ED,$C3,$FC,$89,$5B,$AF,$5C,$63,$87,$E3,$4E,$67,$FC,$80,$D7
	.dc.b	$47,$F9,$6B,$E8,$3F,$CA,$0F,$F0,$AF,$71,$F1,$1F,$C5,$1B,$1F,$C3
	.dc.b	$A7,$6F,$80,$FF,$2E,$38,$7F,$C9,$CD,$7F,$CB,$4D,$77,$ED,$7F,$9A
	.dc.b	$BC,$F7,$53,$E6,$FF,$C5,$3F,$35,$FE,$3D,$F5,$2F,$0F,$F8,$17,$1F
	.dc.b	$EE,$35,$FD,$E8,$CC,$C8,$FE,$B9,$1F,$DF,$BE,$6E,$47,$F5,$F7,$FC
	.dc.b	$03,$E0,$62,$3F,$BD,$C5,$E3,$FF,$01,$CF,$F8,$DB,$CC,$FF,$1B,$F9
	.dc.b	$32,$FF,$C2,$BE,$C2,$75,$18,$9D,$3F,$E3,$58,$FF,$1F,$F9,$3F,$E4
	.dc.b	$0F,$2A,$4F,$28,$CB,$D2,$32,$FF,$8F,$63,$2F,$F8,$FE,$33,$8F,$E6
	.dc.b	$78,$18,$AF,$73,$98,$FE,$99,$7D,$4F,$01,$EA,$F2,$7F,$CA,$7E,$54
	.dc.b	$8C,$10,$FF,$27,$39,$7F,$CA,$11,$FE,$5A,$F2,$7F,$CB,$7E,$54,$9F
	.dc.b	$8B,$19,$7E,$3C,$65,$FF,$2C,$46,$5F,$F2,$D4,$67,$0F,$44,$FF,$9A
	.dc.b	$3E,$83,$F4,$15,$FE,$47,$7F,$CA,$0F,$F9,$19,$AF,$E4,$F3,$47,$06
	.dc.b	$FB,$1C,$FD,$99,$5C,$63,$B1,$87,$95,$74,$C3,$E7,$3C,$C9,$5B,$8C
	.dc.b	$D7,$F9,$FB,$E7,$BF,$D0,$D7,$FE,$5A,$7F,$CF,$CD,$F4,$F5,$EB,$C8
	.dc.b	$C1,$0F,$54,$F2,$6C,$75,$99,$1D,$63,$FC,$13,$F0,$0E,$EF,$E3,$B2
	.dc.b	$F9,$FF,$CB,$3F,$ED,$19,$7B,$1F,$EC,$D9,$CC,$A7,$2F,$FB,$5E,$33
	.dc.b	$1D,$73,$C7,$37,$FB,$56,$31,$5D,$E3,$17,$D5,$B1,$C0,$F1,$61,$E6
	.dc.b	$C3,$AB,$E8,$F7,$DF,$41,$D6,$30,$DC,$BF,$DD,$53,$AB,$E7,$18,$9F
	.dc.b	$EE,$BF,$01,$67,$FD,$C0,$D7,$FD,$BC,$2B,$93,$86,$BF,$EF,$78,$F4
	.dc.b	$3F,$EF,$9F,$73,$FA,$46,$3F,$22,$34,$7C,$13,$FD,$F1,$1A,$CB,$C7
	.dc.b	$19,$39,$79,$1D,$1E,$35,$CE,$B8,$47,$FA,$97,$83,$FC,$83,$5F,$8B
	.dc.b	$E6,$FE,$B1,$A7,$F0,$19,$CD,$7E,$22,$73,$5F,$EB,$99,$CD,$6E,$73
	.dc.b	$5D,$79,$AF,$32,$D5,$87,$8E,$1C,$E3,$15,$97,$8F,$D0,$7F,$96,$27
	.dc.b	$FD,$72,$F0,$C3,$FD,$7C,$E5,$FD,$B3,$9C,$0E,$71,$FE,$38,$E4,$C5
	.dc.b	$7F,$B4,$DC,$74,$F8,$31,$F0,$B6,$7C,$2D,$C3,$BB,$6F,$F9,$27,$AB
	.dc.b	$F0,$3A,$DE,$B5,$8C,$37,$0F,$E7,$ED,$C3,$FD,$E0,$DB,$FE,$5C,$D9
	.dc.b	$FF,$74,$5F,$47,$FA,$3F,$85,$DF,$C2,$F0,$36,$32,$7F,$4C,$8C,$B4
	.dc.b	$FE,$9F,$E1,$79,$FE,$5B,$D0,$D8,$C9,$FD,$52,$32,$D3,$39,$7F,$25
	.dc.b	$C1,$3A,$8C,$37,$2F,$E9,$0D,$AF,$5B,$F1,$FE,$5F,$E4,$7C,$FD,$9F
	.dc.b	$EC,$6D,$BE,$CD,$BF,$EA,$4F,$E3,$BC,$9F,$0D,$7F,$B4,$70,$C6,$19
	.dc.b	$64,$64,$FF,$6B,$46,$30,$FF,$78,$DF,$E6,$BC,$97,$D0,$27,$B4,$E6
	.dc.b	$31,$7F,$ED,$F8,$C4,$EA,$38,$65,$FE,$43,$BE,$73,$FA,$61,$F3,$EF
	.dc.b	$5F,$FC,$83,$1D,$CC,$B8,$7C,$51,$BB,$FF,$43,$C6,$9B,$19,$8F,$F1
	.dc.b	$55,$3F,$1B,$78,$DE,$1A,$D6,$3E,$59,$CE,$1F,$EC,$0F,$9E,$7C,$07
	.dc.b	$87,$24,$62,$7C,$2D,$8F,$F4,$CB,$5F,$51,$C2,$7F,$82,$FC,$DB,$F0
	.dc.b	$27,$D3,$7C,$3B,$E5,$97,$F9,$A6,$B5,$3F,$C3,$3E,$6D,$F1,$DF,$E8
	.dc.b	$AC,$FE,$69,$F5,$63,$3F,$3D,$B9,$D6,$9F,$E4,$B9,$CD,$70,$84,$74
	.dc.b	$8E,$15,$C5,$A6,$E3,$4D,$A4,$EE,$30,$FD,$DF,$9F,$FC,$9D,$FF,$AF
	.dc.b	$1A,$FC,$46,$BF,$7D,$AF,$E3,$DE,$39,$7F,$34,$FA,$97,$CA,$F5,$E8
	.dc.b	$FE,$5F,$29,$CE,$3E,$DF,$0D,$F0,$BC,$A3,$F3,$51,$8E,$AF,$E6,$B3
	.dc.b	$9C,$D3,$9A,$E5,$39,$8C,$4F,$F8,$8E,$30,$2B,$E8,$7C,$1F,$D1,$F9
	.dc.b	$BB,$87,$C7,$6D,$FD,$3D,$D1,$8C,$FF,$D5,$1C,$1F,$7A,$31,$7F,$76
	.dc.b	$B6,$FD,$1C,$3B,$3E,$FF,$0B,$EF,$E1,$FE,$59,$7E,$AC,$64,$FE,$A6
	.dc.b	$E3,$1F,$F1,$E3,$C2,$3E,$2B,$99,$E5,$1C,$38,$7F,$71,$F3,$FE,$2B
	.dc.b	$FF,$7B,$B5,$FC,$FB,$5F,$DD,$35,$FE,$14,$F1,$AF,$F6,$37,$94,$FA
	.dc.b	$87,$C8,$F6,$D7,$FB,$5C,$E6,$3D,$3F,$31,$EA,$F8,$3B,$A7,$C1,$6E
	.dc.b	$7E,$C6,$E5,$F0,$DB,$3E,$E6,$CF,$E0,$35,$47,$C6,$F9,$5F,$8F,$F0
	.dc.b	$36,$FE,$73,$6F,$E4,$B7,$0F,$A0,$FC,$EF,$87,$F9,$FF,$0B,$6F,$EB
	.dc.b	$36,$FE,$8B,$6F,$EC,$3F,$B9,$F9,$AF,$DD,$7C,$D5,$BF,$D8,$36,$FF
	.dc.b	$14,$DB,$FC,$4F,$2A,$31,$5F,$89,$6D,$7F,$00,$9D,$4F,$E9,$FE,$7B
	.dc.b	$FD,$AA,$DC,$7F,$0F,$F1,$5F,$E1,$A9,$CD,$FF,$A0,$DF,$F7,$E5,$71
	.dc.b	$9D,$67,$B7,$F1,$FF,$41,$27,$47,$E5,$DF,$C9,$F4,$48,$FC,$D7,$2D
	.dc.b	$C7,$F2,$B3,$F9,$8F,$A0,$C6,$DF,$9D,$71,$7F,$A1,$72,$FE,$69,$F4
	.dc.b	$5F,$A5,$B1,$EF,$6D,$FA,$DB,$5F,$EC,$1C,$D1,$8A,$FE,$13,$18,$7E
	.dc.b	$45,$F5,$7F,$65,$F4,$1F,$AE,$FA,$0F,$D7,$B6,$32,$7E,$DE,$32,$D3
	.dc.b	$38,$FE,$D2,$32,$D8,$FA,$DF,$41,$F4,$F8,$BE,$A3,$63,$27,$DB,$8C
	.dc.b	$B4,$CC,$FD,$98,$CB,$63,$F8,$54,$FF,$06,$9F,$E0,$ED,$8C,$9F,$C3
	.dc.b	$E3,$2D,$33,$3F,$C3,$23,$2D,$8F,$C7,$7C,$67,$E2,$FE,$77,$F1,$8D
	.dc.b	$8C,$9F,$93,$8C,$B4,$CC,$FE,$46,$32,$D8,$FE,$45,$C9,$FC,$7B,$C5
	.dc.b	$FE,$3E,$D8,$C9,$FC,$9E,$32,$D3,$FA,$47,$CF,$FF,$23,$AF,$E8,$6D
	.dc.b	$F1,$0F,$F2,$59,$FE,$96,$E4,$8F,$FA,$01,$D4,$F2,$8C,$76,$67,$AE
	.dc.b	$C7,$81,$B3,$FE,$EF,$AF,$E2,$73,$E4,$AE,$0F,$F5,$DF,$A0,$FF,$67
	.dc.b	$BF,$EA,$BE,$EF,$EB,$B1,$8B,$7C,$17,$F2,$C7,$FA,$EC,$7F,$B6,$EB
	.dc.b	$84,$63,$E1,$91,$8C,$D5,$CE,$F9,$46,$2B,$77,$DB,$0F,$0F,$C3,$62
	.dc.b	$F7,$87,$19,$F2,$DE,$5B,$8F,$79,$F2,$61,$CC,$F2,$47,$8F,$37,$81
	.dc.b	$C6,$7E,$AF,$81,$74,$FA,$B3,$69,$36,$30,$DB,$E8,$F3,$B7,$D0,$7F
	.dc.b	$42,$F0,$FF,$C9,$AD,$8F,$F1,$D9,$AE,$04,$E9,$B1,$AF,$88,$7C,$8F
	.dc.b	$F8,$E2,$30,$D7,$FC,$8B,$1A,$F1,$3F,$CB,$6D,$8E,$F8,$7C,$0F,$88
	.dc.b	$F1,$E9,$FE,$ED,$9C,$B5,$FF,$27,$CE,$DF,$F2,$8C,$BF,$4F,$7C,$6F
	.dc.b	$F9,$B7,$93,$F0,$93,$98,$C7,$1F,$F9,$FE,$31,$7F,$47,$C6,$CB,$FE
	.dc.b	$FD,$C3,$9B,$F8,$A9,$DB,$F6,$5B,$87,$E2,$38,$97,$CA,$BE,$7F,$8D
	.dc.b	$73,$34,$EF,$C8,$FB,$17,$4E,$6F,$F3,$27,$22,$DF,$27,$F3,$53,$C1
	.dc.b	$CD,$F1,$7F,$1C,$71,$3F,$21,$7D,$EF,$DF,$EA,$3F,$1F,$7C,$27,$FA
	.dc.b	$57,$C8,$FF,$3B,$46,$30,$FF,$36,$56,$EB,$5D,$5F,$C1,$BC,$BB,$4A
	.dc.b	$D3,$C2,$F9,$2C,$3A,$E1,$C3,$2F,$E2,$38,$5D,$FD,$D7,$BC,$BF,$D6
	.dc.b	$7F,$AB,$3E,$5F,$BC,$23,$13,$FC,$7B,$6E,$46,$A3,$31,$CC,$FE,$06
	.dc.b	$E3,$3F,$DF,$E5,$FC,$8A,$BD,$74,$FA,$A7,$83,$86,$C7,$FA,$3C,$D4
	.dc.b	$7A,$FF,$62,$79,$1B,$89,$3B,$AE,$2F,$8F,$05,$F0,$8E,$9F,$8A,$BE
	.dc.b	$5F,$FD,$87,$08,$C3,$FE,$C6,$6B,$FB,$A6,$BF,$EB,$67,$39,$FF,$1C
	.dc.b	$8E,$53,$EE,$9C,$EB,$FE,$E3,$EF,$FE,$5B,$F4,$79,$A9,$7E,$47,$EF
	.dc.b	$CE,$A3,$CF,$F1,$58,$F9,$9F,$F4,$0C,$7F,$AC,$E5,$18,$8F,$A4,$7A
	.dc.b	$E7,$C4,$47,$99,$C7,$F4,$57,$BD,$98,$E3,$9A,$B1,$A7,$F3,$8C,$58
	.dc.b	$70,$BF,$97,$87,$97,$D0,$91,$82,$1D,$23,$E2,$1E,$EF,$23,$E4,$61
	.dc.b	$F2,$72,$E9,$5B,$38,$C7,$86,$77,$B7,$FB,$FF,$C3,$FF,$2C,$57,$C3
	.dc.b	$FC,$57,$E1,$5F,$9D,$B5,$BF,$3B,$F2,$37,$C2,$FE,$3B,$6F,$ED,$36
	.dc.b	$BE,$C7,$E2,$DF,$ED,$8A,$FC,$A4,$E5,$1F,$E9,$2B,$EA,$7D,$A9,$E1
	.dc.b	$1D,$BC,$FF,$BB,$8B,$5A,$79,$22,$4F,$66,$BF,$AD,$DD,$7F,$D7,$23
	.dc.b	$44,$3B,$1F,$85,$74,$F1,$FC,$57,$CE,$FE,$25,$AE,$4F,$DA,$FE,$2C
	.dc.b	$F8,$4A,$48,$FD,$A4,$65,$FC,$97,$CD,$3D,$3F,$CC,$B6,$37,$5C,$5F
	.dc.b	$54,$65,$FD,$B7,$98,$8F,$E3,$0D,$8F,$F7,$A3,$5F,$57,$D4,$79,$BE
	.dc.b	$B6,$47,$91,$F0,$FD,$54,$8F,$2B,$5E,$FF,$4D,$FC,$4B,$EA,$9F,$33
	.dc.b	$FC,$09,$B3,$FD,$E0,$38,$7F,$BD,$B5,$F9,$2E,$79,$FF,$81,$37,$9F
	.dc.b	$E7,$B6,$7F,$0F,$DA,$F8,$3F,$16,$12,$7F,$90,$56,$DF,$DE,$CF,$C4
	.dc.b	$E9,$F8,$BC,$12,$7E,$2C,$65,$F8,$F1,$97,$E6,$C7,$CC,$D7,$E6,$FD
	.dc.b	$64,$9F,$9B,$19,$7E,$7C,$65,$FA,$B1,$FC,$E3,$2F,$20,$F9,$BC,$19
	.dc.b	$CE,$57,$F3,$99,$D3,$67,$2F,$F1,$68,$E7,$1C,$A3,$84,$FF,$60,$F3
	.dc.b	$DF,$92,$BE,$03,$0D,$AF,$ED,$2D,$7F,$17,$CD,$A9,$91,$A7,$F1,$B1
	.dc.b	$F8,$EF,$27,$F1,$FE,$DC,$AF,$C7,$C6,$5F,$C8,$C6,$5F,$CB,$C7,$E2
	.dc.b	$7C,$5F,$C4,$12,$3F,$11,$C5,$FC,$A5,$B1,$B7,$EE,$BF,$2F,$E8,$F1
	.dc.b	$1F,$CE,$27,$83,$67,$4F,$F0,$EE,$5F,$ED,$7F,$86,$3A,$3F,$BF,$2F
	.dc.b	$93,$A6,$22,$57,$1E,$8F,$E2,$FE,$62,$F8,$4E,$44,$66,$F8,$C7,$F4
	.dc.b	$CF,$AA,$7C,$07,$F7,$A6,$C7,$F7,$76,$BF,$D4,$BE,$B3,$AB,$C6,$78
	.dc.b	$11,$FE,$FD,$1A,$CF,$FB,$55,$3E,$2B,$FB,$AB,$6B,$F6,$8D,$7F,$CD
	.dc.b	$AD,$7F,$8D,$B9,$9E,$73,$E7,$3F,$C2,$35,$FE,$61,$8C,$D7,$A6,$74
	.dc.b	$F9,$3D,$FF,$F0,$BE,$1F,$EA,$38,$FE,$DA,$DA,$FF,$00,$37,$4F,$E5
	.dc.b	$BF,$80,$7F,$5A,$9C,$F6,$F2,$25,$FF,$87,$67,$55,$FC,$D3,$F1,$6F
	.dc.b	$F2,$84,$FF,$7A,$6D,$FF,$83,$5B,$3F,$CE,$BF,$17,$FE,$0D,$18,$AF
	.dc.b	$EA,$6D,$4D,$8F,$E2,$5D,$5F,$AC,$F8,$8F,$D5,$92,$BF,$97,$FA,$95
	.dc.b	$FF,$1E,$4E,$DF,$D7,$CF,$F3,$69,$91,$C0,$87,$F3,$CE,$4C,$BF,$CF
	.dc.b	$A2,$E5,$FE,$44,$F1,$BF,$35,$3F,$E6,$2E,$0F,$94,$FF,$97,$DB,$8F
	.dc.b	$E0,$F6,$FF,$2D,$7C,$77,$F0,$59,$C9,$23,$35,$CF,$C2,$D1,$1F,$F7
	.dc.b	$BC,$D7,$DA,$A3,$18,$FF,$96,$3D,$CF,$C5,$70,$7F,$7E,$CB,$F2,$95
	.dc.b	$A8,$C5,$F1,$C3,$A9,$E1,$21,$DA,$7F,$CE,$51,$FE,$7C,$E3,$FF,$2D
	.dc.b	$F4,$DF,$67,$F8,$0E,$7F,$EE,$69,$CE,$1F,$B5,$E9,$69,$FE,$1C,$84
	.dc.b	$E9,$FF,$39,$36,$B9,$5F,$4C,$FF,$80,$DF,$1A,$7D,$BB,$EF,$7F,$B0
	.dc.b	$46,$1B,$19,$C7,$FD,$B7,$39,$6C,$65,$F0,$B7,$1F,$53,$70,$F0,$FC
	.dc.b	$87,$C1,$8C,$57,$A9,$B7,$D6,$3F,$CF,$58,$AB,$FA,$27,$DA,$7F,$8B
	.dc.b	$27,$FC,$37,$C0,$8C,$47,$D7,$6B,$FE,$A9,$6C,$6F,$E5,$39,$92,$7E
	.dc.b	$77,$D6,$BE,$FB,$11,$FA,$2B,$FF,$1D,$EA,$D7,$B6,$1F,$EA,$FC,$38
	.dc.b	$5F,$FA,$F2,$F7,$5A,$8C,$56,$F0,$FF,$51,$FC,$77,$E2,$67,$FD,$31
	.dc.b	$F8,$A7,$FB,$99,$BF,$1D,$AE,$69,$1F,$EA,$5A,$E1,$A7,$FA,$45,$B9
	.dc.b	$7F,$3E,$FA,$A7,$8F,$C2,$FF,$D0,$B5,$A7,$FC,$B2,$DE,$65,$70,$9E
	.dc.b	$DE,$47,$FB,$35,$7D,$6B,$EB,$BF,$C9,$78,$7F,$8E,$E3,$13,$D3,$CC
	.dc.b	$FF,$27,$C6,$21,$3F,$EC,$0D,$3F,$CD,$33,$98,$C4,$F6,$BE,$B7,$C9
	.dc.b	$B1,$C1,$F0,$E1,$E5,$BE,$4F,$9F,$9F,$FC,$91,$87,$46,$C6,$7D,$7E
	.dc.b	$39,$7F,$9E,$A7,$FC,$9D,$8F,$F9,$F3,$E4,$71,$5F,$E6,$68,$46,$1F
	.dc.b	$F3,$4B,$5F,$F3,$6B,$5E,$47,$AC,$FF,$38,$FC,$0F,$E2,$5C,$1C,$0A
	.dc.b	$1D,$65,$F2,$63,$9B,$63,$39,$F9,$6B,$E2,$57,$6E,$0A,$FF,$A5,$B8
	.dc.b	$3C,$19,$FF,$B5,$E3,$F0,$B3,$FC,$37,$7F,$B8,$FF,$BC,$63,$27,$FB
	.dc.b	$30,$B3,$C6,$B9,$CF,$E3,$F0,$B5,$FE,$8E,$9A,$F8,$67,$F8,$D7,$D7
	.dc.b	$7F,$B3,$5E,$D3,$C3,$85,$39,$AF,$EA,$3F,$54,$F8,$CE,$BD,$68,$C6
	.dc.b	$3E,$9F,$AC,$CC,$62,$F8,$DE,$0E,$DA,$AB,$17,$FE,$21,$C3,$A5,$62
	.dc.b	$DA,$FD,$7C,$FF,$4B,$86,$63,$ED,$5F,$0C,$3F,$0C,$DB,$FF,$56,$B5
	.dc.b	$FE,$2F,$97,$95,$FF,$59,$F1,$3F,$7E,$7D,$87,$CA,$8C,$3F,$17,$C0
	.dc.b	$7C,$DE,$6B,$79,$7F,$8E,$6F,$6D,$8C,$E1,$FC,$AA,$B8,$61,$2B,$F9
	.dc.b	$DD,$6A,$BC,$6D,$7F,$48,$D7,$FA,$7F,$EE,$8E,$FF,$33,$6B,$FA,$B6
	.dc.b	$5F,$ED,$D7,$C9,$19,$AF,$EA,$AD,$D7,$FA,$F3,$67,$FA,$AF,$C3,$F4
	.dc.b	$C3,$EF,$F9,$0F,$32,$9F,$0F,$C8,$B7,$FE,$09,$8B,$97,$F4,$89,$B9
	.dc.b	$7D,$6B,$E5,$7B,$3C,$F3,$A7,$FB,$DD,$DB,$FF,$31,$E6,$AC,$70,$FF
	.dc.b	$95,$7C,$DF,$29,$F5,$7D,$9B,$8F,$F8,$6E,$2E,$1F,$D6,$A6,$E7,$FE
	.dc.b	$21,$F5,$BF,$41,$E7,$FF,$9E,$70,$FF,$6F,$F8,$F8,$99,$1F,$EE,$27
	.dc.b	$71,$8C,$3F,$C3,$3E,$5F,$C6,$31,$87,$BD,$C5,$FF,$8B,$5C,$BF,$E1
	.dc.b	$87,$F6,$BF,$02,$F9,$AA,$7E,$EB,$C2,$7D,$E3,$E4,$B3,$87,$FA,$1F
	.dc.b	$26,$1C,$63,$0D,$8F,$D4,$C1,$FE,$39,$5C,$E3,$78,$7F,$06,$FA,$67
	.dc.b	$C3,$53,$C5,$4F,$8B,$C5,$F8,$6B,$6D,$AD,$CE,$B0,$F0,$9E,$1A,$7C
	.dc.b	$7F,$73,$F8,$5F,$DD,$7A,$B0,$E9,$19,$69,$FC,$16,$35,$5D,$05,$74
	.dc.b	$9D,$C7,$F4,$78,$C3,$6B,$F9,$83,$E3,$6A,$FA,$37,$CF,$7F,$28,$6D
	.dc.b	$FF,$AD,$77,$7D,$1B,$CE,$D5,$CA,$FA,$97,$C3,$0F,$83,$87,$0B,$D4
	.dc.b	$7C,$4F,$1D,$E0,$7D,$5C,$3F,$C9,$CF,$F5,$16,$CF,$F3,$58,$C3,$C5
	.dc.b	$E0,$DB,$D3,$B7,$D5,$F4,$52,$BE,$57,$A7,$6B,$E5,$64,$AC,$73,$FF
	.dc.b	$2C,$E0,$FF,$3A,$61,$EC,$9E,$17,$F8,$B3,$F1,$97,$C6,$7F,$BB,$47
	.dc.b	$CB,$9F,$9B,$39,$8E,$07,$FB,$7B,$AB,$FC,$C1,$5F,$E6,$D7,$FC,$C6
	.dc.b	$DF,$67,$F6,$BD,$3F,$B6,$6C,$64,$E1,$1F,$C5,$BE,$B2,$47,$3F,$B9
	.dc.b	$B8,$7F,$A3,$A6,$E9,$FE,$A1,$9B,$5F,$A5,$FA,$67,$CA,$3C,$0F,$ED
	.dc.b	$71,$8F,$96,$B3,$9A,$FF,$7C,$4E,$5B,$19,$CB,$F4,$33,$96,$C6,$71
	.dc.b	$FF,$09,$4E,$5B,$19,$AF,$22,$BF,$0E,$1E,$27,$F8,$CF,$A1,$FC,$6B
	.dc.b	$DA,$B7,$F2,$DB,$97,$E3,$F7,$46,$32,$FF,$61,$3F,$C8,$B5,$FE,$47
	.dc.b	$F5,$F7,$0F,$A8,$DC,$39,$36,$FF,$D9,$4F,$E9,$FB,$65,$7E,$9F,$AD
	.dc.b	$F2,$58,$C3,$F1,$F5,$70,$FF,$19,$74,$7E,$B2,$B1,$87,$E3,$79,$7F
	.dc.b	$88,$3F,$8C,$6E,$1F,$E4,$36,$BF,$8C,$6B,$F9,$5F,$B0,$7D,$9F,$F8
	.dc.b	$89,$B5,$FE,$BB,$FD,$F7,$F5,$39,$7F,$90,$ED,$5A,$7F,$87,$7E,$23
	.dc.b	$FE,$30,$BF,$D0,$36,$3F,$96,$4E,$EB,$94,$E7,$B9,$87,$EE,$27,$FC
	.dc.b	$34,$DC,$3F,$9A,$73,$3F,$12,$E5,$1F,$CD,$EF,$8E,$DE,$9F,$0B,$FC
	.dc.b	$77,$93,$EF,$71,$7F,$D0,$E3,$0D,$8D,$9A,$D7,$D6,$FF,$91,$B2,$F0
	.dc.b	$5D,$CB,$F9,$3E,$9E,$AA,$C4,$F8,$3E,$56,$63,$FD,$47,$F1,$1F,$E6
	.dc.b	$77,$F0,$27,$39,$7C,$87,$FC,$A3,$3A,$AF,$93,$1A,$AF,$F3,$96,$4F
	.dc.b	$C4,$7F,$0F,$5E,$BA,$F3,$46,$30,$FA,$7F,$31,$F0,$8E,$BE,$25,$BF
	.dc.b	$85,$16,$7F,$A3,$65,$FC,$69,$E3,$3A,$3F,$B2,$65,$E2,$C1,$F0,$DC
	.dc.b	$63,$0F,$F9,$AB,$F1,$17,$AF,$4F,$95,$AB,$E6,$F1,$8F,$F7,$5A,$DB
	.dc.b	$6B,$73,$AC,$BF,$AD,$56,$9F,$EC,$F6,$F8,$7F,$81,$97,$E5,$6B,$4F
	.dc.b	$C8,$8C,$E3,$FC,$CE,$7F,$97,$4E,$FC,$8F,$2F,$CC,$FA,$59,$5E,$57
	.dc.b	$2D,$26,$3E,$2C,$3E,$1C,$62,$FC,$1A,$7F,$A3,$3F,$11,$7C,$0D,$3F
	.dc.b	$19,$3F,$E8,$93,$F5,$91,$AA,$FF,$0C,$37,$0F,$EF,$4D,$9F,$C1,$9C
	.dc.b	$67,$FA,$74,$65,$B5,$FD,$E2,$39,$57,$F4,$C6,$BA,$9F,$EA,$7F,$4D
	.dc.b	$EA,$8F,$F0,$43,$6F,$FD,$19,$FB,$17,$EE,$E5,$39,$AF,$EB,$CD,$8F
	.dc.b	$F1,$53,$7A,$D5,$C6,$31,$EE,$FF,$37,$BF,$C6,$9F,$DF,$7F,$34,$FF
	.dc.b	$1A,$63,$FE,$9B,$9D,$63,$FE,$26,$A5,$69,$FE,$05,$3A,$AF,$F6,$C4
	.dc.b	$6B,$D7,$FF,$4A,$79,$7F,$EC,$08,$C1,$FE,$AB,$FD,$57,$F8,$DB,$7C
	.dc.b	$2F,$F4,$D7,$8D,$FE,$BA,$9C,$E7,$FE,$36,$F7,$91,$8C,$BF,$D5,$AD
	.dc.b	$C3,$C7,$7C,$70,$FF,$66,$FC,$57,$2F,$8C,$7C,$A7,$FB,$46,$7F,$DD
	.dc.b	$D1,$8C,$DC,$4F,$AF,$7C,$B6,$B0,$E1,$9F,$FC,$01,$6D,$FF,$C0,$06
	.dc.b	$E1,$CF,$0E,$57,$C6,$BF,$DD,$D8,$D5,$97,$97,$0E,$27,$F2,$DC,$BC
	.dc.b	$39,$77,$AE,$2D,$E6,$7E,$A8,$F4,$3B,$46,$1F,$06,$8E,$47,$DA,$6A
	.dc.b	$F8,$61,$CD,$B8,$FA,$CF,$D4,$B8,$DF,$D9,$7C,$AB,$86,$3E,$BF,$B2
	.dc.b	$7C,$F3,$CD,$FF,$17,$C6,$1B,$19,$CB,$FC,$5E,$27,$4E,$3E,$55,$E0
	.dc.b	$FF,$88,$23,$0D,$7F,$C6,$F1,$AF,$0F,$FC,$7A,$D8,$EB,$5F,$E6,$8E
	.dc.b	$0F,$1F,$C0,$7F,$04,$9C,$B5,$FF,$20,$CE,$DF,$F2,$1C,$BD,$7F,$F0
	.dc.b	$AC,$7D,$16,$BF,$1E,$35,$EB,$FF,$94,$78,$3E,$3F,$D9,$DE,$6F,$F4
	.dc.b	$F5,$6A,$F9,$5E,$5F,$E2,$F8,$74,$C3,$8C,$7E,$04,$F8,$DF,$0D,$FE
	.dc.b	$5D,$E2,$F3,$B6,$3F,$CC,$19,$FF,$AC,$DF,$F3,$89,$FE,$AD,$70,$FF
	.dc.b	$82,$CE,$67,$57,$35,$FA,$38,$C9,$F8,$3E,$6F,$D2,$7D,$9F,$9E,$33
	.dc.b	$3B,$AE,$3A,$78,$B0,$F6,$61,$C2,$FF,$35,$DE,$FB,$37,$D9,$FF,$9D
	.dc.b	$FC,$16,$BF,$EC,$3A,$D4,$E6,$31,$5C,$2F,$F8,$C9,$FE,$F2,$FB,$3D
	.dc.b	$2B,$E2,$E9,$AF,$F0,$B9,$E4,$FF,$A7,$FA,$95,$CB,$2F,$D4,$67,$D3
	.dc.b	$17,$D9,$BD,$FC,$37,$1F,$1D,$F0,$6E,$3F,$43,$5F,$E4,$B5,$B9,$FE
	.dc.b	$58,$D8,$FE,$61,$8F,$AA,$7F,$9F,$F9,$9F,$E9,$86,$9F,$D2,$1C,$79
	.dc.b	$3F,$DC,$E3,$07,$FB,$C5,$C3,$FD,$32,$3D,$9F,$66,$F9,$0E,$DF,$62
	.dc.b	$FB,$3F,$93,$DC,$FA,$8F,$E2,$B8,$FC,$E8,$C7,$8F,$C2,$A4,$FF,$AB
	.dc.b	$2F,$93,$63,$D1,$18,$8F,$DF,$B7,$5F,$F2,$E3,$78,$BF,$D7,$0D,$9F
	.dc.b	$F4,$FF,$4B,$E5,$3C,$FF,$49,$E6,$FA,$DF,$F6,$25,$79,$5A,$FF,$BC
	.dc.b	$6D,$3C,$23,$CC,$F9,$08,$E0,$FC,$58,$8E,$2E,$1A,$FF,$88,$BE,$CD
	.dc.b	$EF,$FC,$A3,$C9,$F1,$F2,$78,$46,$5F,$8C,$FC,$42,$38,$36,$43,$FC
	.dc.b	$5F,$E4,$FF,$A0,$75,$FF,$41,$77,$FF,$AB,$E7,$EA,$7D,$9F,$C0,$9F
	.dc.b	$F2,$BB,$77,$FF,$71,$56,$E9,$5C,$BD,$3F,$C6,$4E,$67,$FD,$C2,$D7
	.dc.b	$8C,$68,$FC,$A4,$73,$9E,$53,$F8,$5F,$B3,$7C,$D6,$BD,$4F,$C5,$E4
	.dc.b	$D6,$99,$BF,$9F,$8A,$BC,$64,$8F,$F7,$BC,$65,$B1,$C5,$F5,$E7,$EE
	.dc.b	$8F,$CC,$E2,$F2,$FF,$37,$F6,$7F,$EF,$CD,$3F,$CF,$6D,$BF,$F4,$EB
	.dc.b	$70,$0F,$B3,$FF,$62,$BF,$84,$6B,$FD,$61,$AF,$03,$8B,$87,$0D,$7F
	.dc.b	$DF,$3F,$2B,$FE,$7A,$7A,$7C,$6F,$EF,$E7,$AE,$1F,$E9,$0A,$D5,$7F
	.dc.b	$B9,$5F,$E2,$9E,$61,$A7,$BB,$E5,$F9,$7C,$3F,$F0,$7C,$0D,$25,$FF
	.dc.b	$70,$EE,$E5,$F3,$9F,$C0,$30,$34,$EF,$F3,$5F,$C4,$FE,$C5,$F1,$7F
	.dc.b	$C5,$7B,$9F,$47,$4D,$44,$CB,$FB,$E8,$AE,$2E,$1A,$FF,$80,$9A,$FC
	.dc.b	$B3,$84,$73,$7E,$F9,$F5,$A3,$8E,$FF,$EE,$5B,$7D,$9F,$C6,$E7,$F8
	.dc.b	$FC,$53,$0F,$E6,$BC,$5F,$D7,$70,$FE,$9F,$1F,$E4,$08,$C7,$27,$FB
	.dc.b	$75,$FA,$6D,$AE,$93,$CB,$87,$E9,$FC,$4C,$8C,$10,$F3,$F9,$DF,$E0
	.dc.b	$98,$D4,$63,$0F,$AA,$FF,$10,$D3,$D7,$8F,$E2,$D7,$C6,$FC,$F3,$E8
	.dc.b	$6F,$2B,$84,$65,$E5,$19,$7E,$38,$9E,$6F,$03,$67,$F6,$5F,$B3,$7C
	.dc.b	$BF,$F6,$1F,$93,$FE,$83,$3E,$A6,$EF,$F2,$3D,$65,$F0,$AE,$57,$FD
	.dc.b	$CA,$72,$D9,$D4,$FC,$1F,$8B,$FC,$DF,$A1,$F9,$CF,$7E,$4F,$D6,$8C
	.dc.b	$BF,$62,$32,$FC,$98,$FC,$A7,$95,$F9,$3E,$E9,$3F,$CE,$BA,$3F,$9B
	.dc.b	$E9,$F7,$A3,$F2,$55,$FD,$AE,$7F,$9D,$C7,$E3,$DA,$FE,$ED,$AF,$E3
	.dc.b	$E3,$BB,$D5,$FD,$17,$9D,$FE,$07,$FB,$3F,$D4,$CF,$F8,$09,$7D,$0E
	.dc.b	$4D,$77,$49,$D4,$65,$E1,$1F,$E0,$1F,$B3,$FF,$35,$C7,$F5,$4F,$49
	.dc.b	$F3,$DF,$E0,$56,$C7,$F5,$89,$DB,$F8,$06,$E9,$F0,$5B,$D1,$FC,$5D
	.dc.b	$B3,$FC,$82,$BC,$EF,$3F,$92,$ED,$E1,$48,$C1,$0F,$F3,$CD,$27,$FC
	.dc.b	$2C,$D8,$FE,$02,$D7,$71,$ED,$23,$ED,$F8,$2F,$F1,$07,$C2,$FE,$CA
	.dc.b	$3F,$99,$FD,$9B,$E2,$FF,$C4,$8D,$8F,$EE,$F3,$B7,$FD,$23,$E2,$72
	.dc.b	$23,$FD,$F2,$7F,$BB,$BF,$DF,$BE,$CD,$F0,$BF,$E2,$FF,$B3,$7E,$01
	.dc.b	$4F,$F1,$94,$EA,$B8,$56,$08,$FF,$82,$2B,$F7,$2F,$F9,$47,$B5,$F6
	.dc.b	$9F,$C8,$BC,$0F,$D3,$B6,$33,$3F,$EA,$29,$CB,$63,$2F,$FB,$85,$BC
	.dc.b	$9F,$B1,$6F,$27,$F1,$7F,$6F,$FA,$29,$2B,$F8,$FB,$FD,$23,$9F,$11
	.dc.b	$FC,$46,$2D,$FF,$22,$F2,$3F,$D3,$D5,$FB,$9E,$4F,$F4,$15,$FF,$5D
	.dc.b	$CB,$EF,$70,$FA,$E7,$33,$FE,$77,$6B,$FE,$71,$7E,$D4,$F0,$8F,$F3
	.dc.b	$94,$7D,$DB,$FF,$6A,$7C,$87,$7F,$2B,$A1,$2B,$E2,$57,$F2,$67,$FA
	.dc.b	$25,$FE,$3E,$78,$47,$47,$FC,$F1,$F0,$9F,$8F,$CF,$F8,$B5,$F1,$6C
	.dc.b	$F0,$AF,$F4,$04,$D9,$CC,$73,$3F,$CF,$31,$FE,$8C,$E7,$FF,$34,$77
	.dc.b	$FF,$5C,$6C,$FF,$B6,$9A,$FF,$B2,$3E,$03,$F8,$4C,$62,$FF,$A7,$B7
	.dc.b	$83,$FD,$87,$3F,$C5,$9F,$F0,$8E,$BF,$EF,$7F,$96,$F0,$7D,$1F,$EF
	.dc.b	$B1,$FF,$4B,$72,$7B,$8C,$F7,$C8,$F1,$4E,$DF,$F0,$A6,$48,$C4,$7F
	.dc.b	$B4,$9A,$A3,$53,$CE,$3F,$B7,$C6,$1F,$F4,$DF,$DA,$7F,$3D,$FB,$3F
	.dc.b	$F1,$FC,$7F,$A8,$3E,$D3,$F8,$4B,$FD,$81,$BB,$7F,$47,$9D,$CF,$F7
	.dc.b	$AE,$34,$62,$3F,$1C,$D7,$FD,$84,$D7,$FD,$78,$D8,$E3,$EF,$78,$C9
	.dc.b	$3E,$3F,$AA,$AF,$7A,$F0,$DF,$8E,$BF,$D7,$D8,$2F,$FD,$61,$F2,$5F
	.dc.b	$87,$E1,$FA,$7F,$69,$FE,$FE,$C3,$F4,$CD,$8F,$F2,$04,$EF,$2F,$F2
	.dc.b	$4B,$F9,$77,$FC,$34,$D7,$7C,$DF,$EB,$B8,$C4,$70,$8F,$09,$1F,$F6
	.dc.b	$5F,$D8,$FF,$70,$AE,$FE,$6F,$F9,$EA,$7F,$C8,$5F,$67,$E2,$F4,$3F
	.dc.b	$CA,$5E,$DA,$31,$3F,$E5,$C8,$C3,$E6,$AE,$51,$88,$64,$9C,$D7,$29
	.dc.b	$EE,$7B,$AF,$BC,$79,$34,$EC,$D9,$FC,$66,$17,$8F,$FC,$B1,$1E,$1A
	.dc.b	$FF,$37,$C6,$23,$C3,$87,$B6,$31,$1F,$E7,$4F,$89,$DD,$FF,$91,$61
	.dc.b	$18,$7F,$C9,$0D,$7F,$CE,$8D,$79,$1E,$A3,$FC,$FF,$F0,$DF,$E4,$29
	.dc.b	$F8,$51,$A6,$C6,$71,$F0,$57,$BA,$BF,$D0,$13,$97,$C2,$75,$0E,$5F
	.dc.b	$F4,$6F,$A1,$FE,$50,$CB,$FC,$29,$08,$C6,$BF,$C0,$B6,$FB,$4F,$FB
	.dc.b	$52,$74,$D9,$E3,$1F,$E9,$16,$BD,$A3,$FC,$E2,$F0,$3F,$CE,$71,$D9
	.dc.b	$75,$BE,$CF,$FD,$CF,$5D,$4F,$F5,$1E,$1F,$D1,$63,$35,$FB,$9F,$B3
	.dc.b	$7C,$8F,$3E,$94,$63,$0F,$CC,$FC,$06,$63,$17,$C6,$F0,$78,$F2,$69
	.dc.b	$FE,$8C,$C3,$C2,$7F,$AD,$31,$EB,$5C,$23,$FB,$46,$1D,$30,$FE,$AC
	.dc.b	$D7,$83,$55,$72,$9F,$D6,$F8,$3C,$E7,$FC,$EF,$87,$BB,$4F,$E6,$FF
	.dc.b	$62,$FB,$2F,$91,$C8,$8C,$63,$E6,$C3,$E5,$F8,$AF,$B3,$DD,$6A,$7F
	.dc.b	$95,$B5,$FE,$6A,$D7,$F4,$6D,$7F,$9A,$36,$FC,$18,$7C,$7F,$E0,$0F
	.dc.b	$E7,$18,$F8,$67,$2F,$8F,$81,$3A,$C3,$F9,$DF,$DA,$7F,$BD,$AB,$83
	.dc.b	$63,$FD,$2F,$9F,$FB,$DE,$73,$E0,$7C,$0A,$D5,$7F,$4C,$6B,$FB,$A6
	.dc.b	$BF,$D2,$DB,$C1,$FE,$45,$F3,$3F,$0D,$E4,$9F,$C0,$62,$F2,$73,$13
	.dc.b	$2F,$F0,$3B,$A8,$C6,$3F,$D2,$6F,$95,$68,$F8,$17,$CD,$C4,$EE,$FF
	.dc.b	$4F,$87,$8E,$7F,$C1,$B1,$8B,$E7,$5F,$E0,$E7,$FD,$C7,$3F,$E1,$8F
	.dc.b	$33,$CF,$EE,$FA,$99,$1F,$E1,$67,$91,$1F,$E2,$DB,$3E,$0B,$FD,$07
	.dc.b	$1F,$ED,$2F,$6B,$F4,$11,$FD,$46,$30,$4C,$7F,$D8,$D3,$AF,$1B,$6F
	.dc.b	$CA,$F2,$E5,$6F,$9F,$FC,$3F,$18,$BF,$53,$ED,$24,$FF,$BD,$3D,$97
	.dc.b	$D1,$7F,$03,$A4,$62,$7F,$39,$E0,$BE,$D1,$F2,$FB,$AD,$5F,$FB,$39
	.dc.b	$B9,$7F,$BB,$2F,$87,$A5,$BC,$BF,$88,$46,$32,$FD,$46,$0F,$B4,$7C
	.dc.b	$0F,$8D,$B3,$E2,$D1,$F1,$F9,$A9,$39,$FA,$2B,$83,$4F,$16,$09,$E9
	.dc.b	$F2,$92,$FF,$71,$D2,$FB,$4F,$83,$1E,$4D,$15,$BC,$BF,$A8,$46,$27
	.dc.b	$FA,$94,$7A,$67,$F9,$8C,$E6,$37,$1F,$CB,$F2,$7D,$D7,$AA,$FB,$4E
	.dc.b	$63,$15,$FE,$F8,$FB,$17,$C9,$FD,$56,$DF,$D4,$FB,$27,$A3,$9A,$93
	.dc.b	$9E,$5B,$87,$C3,$8F,$EB,$F7,$EE,$8C,$1F,$D0,$BE,$58,$FF,$33,$7F
	.dc.b	$AC,$CF,$F0,$B7,$2D,$E4,$D4,$ED,$B1,$93,$EE,$E4,$F7,$7D,$47,$F4
	.dc.b	$7E,$07,$AF,$8A,$DC,$FF,$5A,$AE,$11,$F3,$9B,$87,$F8,$05,$AF,$A0
	.dc.b	$F4,$BA,$E0,$FC,$96,$6F,$7B,$B5,$7C,$0C,$D3,$9C,$3F,$94,$77,$7F
	.dc.b	$BB,$9B,$D9,$FB,$6F,$77,$F6,$ED,$8C,$9C,$23,$2F,$FA,$7B,$E9,$7F
	.dc.b	$8F,$9F,$BE,$FB,$5B,$19,$3F,$9B,$7C,$73,$EC,$33,$A7,$D4,$F1,$13
	.dc.b	$9A,$FE,$9D,$F6,$AF,$97,$79,$1F,$DE,$E3,$1F,$39,$67,$35,$C1,$46
	.dc.b	$6B,$4D,$9D,$37,$1F,$C2,$DF,$49,$E5,$5C,$1F,$8F,$EC,$7F,$1A,$FA
	.dc.b	$7F,$E3,$6D,$8C,$9F,$32,$32,$D3,$E7,$ED,$FC,$8B,$ED,$7F,$91,$B6
	.dc.b	$32,$7D,$48,$CB,$4F,$C6,$F8,$DF,$A5,$3F,$86,$F2,$3E,$67,$13,$A3
	.dc.b	$F1,$53,$F8,$D2,$38,$3F,$BD,$0F,$B1,$7C,$C7,$2D,$BF,$4F,$39,$8F
	.dc.b	$E5,$1F,$24,$FB,$5A,$7F,$86,$F8,$FF,$4F,$7F,$AE,$13,$BB,$FE,$43
	.dc.b	$F6,$BF,$E3,$59,$FC,$D0,$70,$FE,$75,$AF,$ED,$9B,$CB,$F8,$6E,$BF
	.dc.b	$01,$FE,$3D,$E6,$7D,$AF,$7B,$FC,$66,$5F,$BC,$7C,$99,$7F,$17,$EE
	.dc.b	$67,$DF,$A7,$DB,$E4,$7F,$45,$D7,$F2,$58,$7A,$DA,$FA,$1A,$F1,$AF
	.dc.b	$E4,$B1,$88,$FC,$C6,$5D,$E7,$F8,$7E,$7E,$6B,$ED,$87,$A6,$B5,$7E
	.dc.b	$FD,$FC,$47,$F5,$7F,$B5,$FE,$35,$9F,$A2,$75,$97,$CC,$AF,$F3,$1E
	.dc.b	$7F,$6A,$3C,$71,$82,$3F,$E3,$9C,$FF,$01,$C7,$DA,$31,$C5,$F8,$B7
	.dc.b	$EB,$FC,$EF,$C3,$3C,$5E,$25,$8C,$9F,$02,$3C,$24,$70,$74,$F8,$AE
	.dc.b	$19,$FF,$32,$C7,$F2,$11,$FD,$CF,$ED,$5E,$C5,$3E,$E4,$BE,$7F,$18
	.dc.b	$7F,$6A,$AD,$BF,$D5,$1B,$87,$FA,$9D,$AF,$E3,$9A,$FF,$29,$77,$BF
	.dc.b	$CB,$BF,$F6,$16,$BD,$AB,$CD,$F1,$9F,$0D,$FC,$E5,$FF,$2F,$C3,$97
	.dc.b	$93,$E6,$F9,$EF,$53,$23,$C8,$FF,$A5,$A7,$F9,$CD,$70,$9C,$9F,$4E
	.dc.b	$DA,$F1,$9F,$0B,$FE,$17,$FB,$5F,$F2,$E6,$9F,$81,$94,$62,$FF,$1B
	.dc.b	$3F,$E7,$B7,$FA,$34,$66,$7F,$AF,$BF,$E6,$57,$F6,$2D,$77,$F0,$7F
	.dc.b	$EA,$C8,$C4,$70,$8F,$D7,$7A,$4F,$AE,$F2,$46,$1F,$F1,$67,$D5,$FA
	.dc.b	$B0,$FF,$0F,$B6,$FF,$9F,$7D,$AF,$FA,$D7,$2F,$54,$E5,$B9,$FF,$AF
	.dc.b	$23,$85,$7F,$06,$6B,$FD,$85,$AF,$FA,$4B,$CA,$FF,$5F,$60,$FB,$5F
	.dc.b	$F1,$AB,$FD,$FB,$C2,$7C,$1F,$F4,$48,$FE,$43,$3A,$BE,$51,$8F,$77
	.dc.b	$FC,$EF,$87,$4A,$D7,$B4,$FB,$5F,$E3,$B7,$FE,$2C,$6E,$1F,$E8,$CF
	.dc.b	$23,$FD,$8D,$18,$DB,$F7,$17,$C7,$C5,$C9,$FE,$84,$AB,$87,$F9,$1B
	.dc.b	$E4,$79,$FC,$97,$F7,$BC,$BF,$DD,$77,$FC,$06,$31,$BB,$ED,$5F,$05
	.dc.b	$9F,$63,$FD,$CD,$77,$0E,$39,$7F,$88,$31,$E9,$1F,$EF,$39,$D6,$3C
	.dc.b	$F0,$E1,$39,$BE,$18,$74,$7F,$9F,$C7,$43,$FD,$FB,$97,$C2,$8F,$E7
	.dc.b	$AD,$8F,$F1,$7F,$3B,$ED,$5E,$8F,$96,$30,$F8,$B9,$7F,$CB,$58,$78
	.dc.b	$2D,$5A,$8C,$36,$1A,$FC,$78,$C3,$F1,$AB,$9C,$EB,$2F,$1E,$3E,$2C
	.dc.b	$3B,$4B,$ED,$1F,$46,$F4,$3F,$CC,$31,$86,$C6,$67,$FA,$D9,$C5,$C3
	.dc.b	$FD,$7B,$E6,$1E,$1F,$F8,$6E,$37,$18,$F1,$7F,$CD,$AD,$8C,$CE,$DF
	.dc.b	$EC,$F8,$7F,$99,$B6,$F1,$FC,$0F,$C4,$9F,$0C,$E5,$B1,$97,$FC,$91
	.dc.b	$3F,$EB,$5F,$5B,$FC,$95,$A6,$3D,$D4,$62,$BE,$BC,$61,$FB,$F5,$C5
	.dc.b	$B1,$97,$E6,$1F,$59,$C7,$0F,$D3,$6F,$07,$EC,$F8,$5F,$6B,$FE,$5B
	.dc.b	$AF,$C9,$E9,$FE,$B0,$C3,$F0,$1C,$DF,$EF,$CF,$83,$D6,$3F,$BB,$F9
	.dc.b	$0F,$F4,$1D,$7D,$8B,$E0,$F8,$8F,$B5,$F1,$3E,$B7,$FC,$FF,$1A,$DF
	.dc.b	$E5,$CE,$67,$67,$E7,$0A,$73,$9E,$B9,$79,$A7,$C7,$2C,$BF,$65,$F4
	.dc.b	$38,$EE,$7D,$AF,$F2,$7A,$FF,$58,$B5,$FF,$49,$C6,$BC,$8F,$F3,$E7
	.dc.b	$83,$FD,$F7,$0F,$E2,$FF,$6B,$8D,$7F,$D2,$F7,$BC,$39,$E1,$95,$19
	.dc.b	$7F,$CD,$9C,$4B,$D6,$1F,$C8,$B4,$E1,$8B,$ED,$5F,$01,$86,$E3,$FE
	.dc.b	$A6,$D3,$C9,$97,$CE,$6E,$1B,$B9,$5B,$C3,$EF,$4F,$96,$1E,$4B,$CE
	.dc.b	$F1,$FB,$BF,$EE,$29,$FF,$42,$FC,$6E,$EB,$51,$FD,$2A,$BF,$DA,$2F
	.dc.b	$26,$C6,$A3,$F6,$2D,$7F,$C9,$BF,$66,$FB,$4F,$8D,$E0,$BE,$AD,$E3
	.dc.b	$E1,$B7,$FE,$F6,$BE,$0D,$BD,$54,$8C,$10,$FF,$30,$4F,$06,$CF,$83
	.dc.b	$A1,$F2,$9F,$0E,$9F,$39,$E4,$FA,$EF,$F6,$46,$1F,$E3,$56,$BD,$23
	.dc.b	$81,$BF,$62,$47,$9E,$32,$D2,$1F,$E2,$FF,$B4,$FF,$7E,$BE,$DF,$85
	.dc.b	$F0,$7C,$84,$8C,$10,$FF,$1B,$60,$8D,$47,$BA,$7D,$44,$78,$73,$FF
	.dc.b	$8A,$FC,$CF,$F0,$B3,$FE,$17,$F0,$FF,$D6,$73,$FE,$51,$FB,$4F,$6C
	.dc.b	$FF,$8C,$5B,$D3,$FE,$DD,$AD,$D7,$FA,$2D,$B5,$BD,$7E,$DD,$F0,$AF
	.dc.b	$F7,$DD,$6A,$F9,$7A,$DF,$91,$8C,$4E,$A3,$F0,$7F,$68,$F9,$CD,$7A
	.dc.b	$DD,$36,$97,$F9,$B3,$AC,$69,$F5,$46,$9F,$34,$65,$A7,$13,$FD,$E8
	.dc.b	$F8,$3C,$EF,$C9,$FD,$A7,$FA,$8B,$6F,$F2,$93,$74,$FF,$43,$B6,$C3
	.dc.b	$ED,$3F,$D8,$0F,$56,$BF,$E6,$D8,$E1,$D6,$79,$47,$1F,$81,$FD,$1F
	.dc.b	$9B,$FE,$50,$8E,$51,$FB,$7B,$E7,$8F,$F9,$96,$B5,$5F,$EF,$7F,$3F
	.dc.b	$F9,$8B,$63,$F8,$6C,$62,$7D,$7E,$0F,$3F,$85,$F7,$B2,$FD,$11,$E2
	.dc.b	$F9,$C9,$1F,$E9,$5C,$11,$D3,$E5,$FF,$85,$7D,$9F,$FB,$5A,$BE,$67
	.dc.b	$5B,$E9,$7F,$87,$36,$3E,$84,$ED,$FF,$6A,$37,$3F,$F3,$54,$EE,$7F
	.dc.b	$DB,$0D,$8F,$F3,$74,$EF,$87,$E8,$D2,$3A,$3F,$ED,$A6,$BA,$6B,$F6
	.dc.b	$1B,$E6,$36,$FF,$18,$DF,$F3,$F7,$E7,$DF,$F9,$5D,$B7,$FD,$35,$DC
	.dc.b	$67,$A7,$E7,$F0,$CA,$FC,$B6,$1D,$67,$E9,$EF,$F5,$3E,$1E,$56,$B1
	.dc.b	$4E,$A3,$13,$A7,$EF,$47,$CC,$F8,$BF,$9B,$F1,$52,$7E,$8C,$65,$FA
	.dc.b	$71,$97,$E5,$C6,$5F,$9B,$19,$D3,$E8,$CF,$F5,$9F,$45,$F5,$AF,$94
	.dc.b	$FE,$8B,$E5,$4C,$7F,$65,$5A,$9E,$15,$B9,$FF,$18,$CE,$5B,$19,$7E
	.dc.b	$C9,$FD,$99,$C3,$F7,$36,$FA,$9E,$6F,$E6,$BD,$C9,$3F,$1E,$3F,$3A
	.dc.b	$61,$FD,$14,$61,$FC,$7F,$91,$F9,$2E,$A9,$18,$21,$F8,$E7,$2F,$E5
	.dc.b	$23,$F3,$A7,$F5,$EE,$5F,$E8,$B1,$FE,$4C,$3F,$21,$19,$31,$83,$EC
	.dc.b	$FF,$4F,$3F,$82,$5F,$3F,$FE,$00,$6C,$67,$AF,$06,$66,$46,$4F,$E3
	.dc.b	$1F,$67,$FE,$62,$7F,$8D,$7A,$0F,$9A,$FE,$F8,$D8,$DF,$C1,$7F,$7F
	.dc.b	$AE,$11,$FE,$64,$6B,$F4,$9A,$FE,$B1,$AF,$F1,$E5,$F2,$1D,$BC,$09
	.dc.b	$3E,$48,$EC,$61,$73,$FF,$B9,$70,$56,$A7,$81,$F2,$EF,$AB,$6F,$FC
	.dc.b	$1B,$18,$8F,$F3,$DE,$1F,$B0,$7F,$98,$FD,$9B,$E2,$3F,$C3,$8D,$8F
	.dc.b	$E7,$D3,$B7,$FD,$13,$DD,$FE,$18,$8F,$F5,$7C,$7E,$49,$AF,$F0,$C6
	.dc.b	$BF,$CF,$BE,$C9,$F0,$1F,$E2,$9F,$B2,$7D,$9D,$3F,$C5,$93,$AA,$FE
	.dc.b	$9C,$D7,$FD,$26,$D7,$F4,$2D,$7F,$8D,$B5,$FD,$27,$D9,$FF,$8D,$3E
	.dc.b	$BE,$47,$EC,$3B,$13,$FB,$03,$F2,$9D,$BF,$B8,$6C,$6A,$7F,$63,$1F
	.dc.b	$96,$F1,$3F,$C4,$31,$FC,$47,$D9,$FE,$8E,$7F,$48,$E8,$B1,$92,$1F
	.dc.b	$D3,$DF,$E0,$DC,$7F,$4F,$A9,$F0,$DF,$7A,$BF,$90,$B5,$FF,$36,$CF
	.dc.b	$D9,$CF,$F1,$7C,$5E,$88,$F9,$F1,$89,$FE,$13,$18,$3F,$92,$BD,$27
	.dc.b	$FC,$D3,$5C,$E7,$33,$DB,$C5,$E8,$47,$9B,$5D,$1F,$C6,$A3,$8F,$4F
	.dc.b	$F9,$B1,$F6,$CC,$9D,$3F,$E7,$8F,$81,$FC,$A5,$3C,$9F,$F7,$5F,$D4
	.dc.b	$7F,$06,$BF,$D0,$C7,$FA,$97,$CA,$F3,$9F,$E8,$5F,$89,$56,$EF,$95
	.dc.b	$FF,$BB,$23,$93,$FE,$C6,$6B,$DB,$C2,$FE,$25,$5C,$9B,$F4,$5F,$C0
	.dc.b	$67,$FD,$1F,$C9,$E7,$3C,$11,$9C,$3B,$DA,$31,$97,$FB,$0E,$B8,$5F
	.dc.b	$F8,$BE,$B5,$E0,$F2,$3B,$92,$3F,$DD,$F1,$A7,$FA,$0F,$D9,$7F,$8E
	.dc.b	$DF,$EC,$7F,$65,$FC,$81,$F8,$A7,$FA,$DD,$C6,$5F,$0B,$8D,$5C,$6B
	.dc.b	$FD,$6A,$D9,$FF,$5A,$57,$0C,$BF,$C6,$7C,$88,$C5,$FF,$AC,$AF,$85
	.dc.b	$7F,$AE,$1B,$19,$F6,$3F,$B0,$46,$2F,$97,$D5,$62,$65,$7F,$60,$AD
	.dc.b	$60,$AF,$F5,$67,$C7,$FE,$2B,$1F,$1F,$D9,$7F,$BC,$F3,$FF,$65,$36
	.dc.b	$3F,$C3,$33,$BC,$7F,$C9,$6D,$CD,$5B,$9F,$F6,$E3,$54,$F1,$6F,$2F
	.dc.b	$88,$8F,$F8,$4B,$EC,$3F,$D5,$6F,$D0,$38,$F9,$DF,$CD,$97,$DB,$3C
	.dc.b	$0F,$2F,$CD,$7F,$8B,$A3,$13,$E6,$8C,$4F,$AA,$B9,$56,$AB,$AE,$BF
	.dc.b	$E6,$68,$C6,$5F,$E5,$86,$E1,$CA,$BE,$0D,$F2,$C3,$A6,$1F,$E6,$EE
	.dc.b	$3F,$F2,$EE,$3F,$ED,$78,$C7,$B1,$AC,$15,$AC,$3D,$8F,$F9,$92,$3F
	.dc.b	$CE,$FF,$0F,$C1,$FF,$08,$B5,$FF,$25,$B5,$D9,$D2,$BF,$C2,$EA,$3D
	.dc.b	$98,$FF,$86,$BE,$13,$F7,$3A,$7C,$1A,$E2,$D8,$CC,$F2,$AE,$35,$FE
	.dc.b	$5A,$9C,$BE,$43,$E9,$CF,$5E,$1D,$C7,$FA,$43,$D5,$FE,$0F,$A7,$F0
	.dc.b	$88,$C4,$FD,$FA,$CB,$FE,$2E,$AB,$3F,$E6,$86,$F1,$3E,$CF,$FD,$C3
	.dc.b	$5F,$EA,$5A,$FF,$79,$BF,$E5,$A8,$CC,$7E,$24,$C9,$FB,$A9,$E5,$D9
	.dc.b	$DA,$3C,$66,$67,$8E,$3F,$BD,$FB,$37,$C7,$79,$3A,$D1,$8C,$BF,$31
	.dc.b	$EE,$E6,$31,$8F,$F0,$8A,$DD,$E4,$F5,$F3,$7F,$A4,$70,$FF,$2A,$DF
	.dc.b	$0C,$7C,$15,$AC,$F8,$57,$E9,$5B,$8F,$AF,$0F,$3C,$FF,$AF,$B1,$EE
	.dc.b	$EC,$FE,$49,$F6,$4F,$B2,$FA,$57,$F9,$78,$C4,$FF,$24,$6C,$7F,$98
	.dc.b	$39,$51,$88,$FF,$4D,$C6,$2B,$EA,$F8,$DF,$9C,$BE,$F1,$89,$FB,$66
	.dc.b	$E3,$FB,$7B,$8B,$EF,$A7,$F4,$CE,$0F,$EE,$B3,$C3,$EC,$BF,$DD,$D7
	.dc.b	$FC,$E1,$B9,$79,$5A,$F9,$1A,$FE,$CD,$DD,$F5,$CB,$FD,$33,$5F,$D1
	.dc.b	$E7,$2F,$F3,$48,$CD,$7F,$80,$BC,$AF,$C0,$79,$07,$F2,$7E,$47,$F6
	.dc.b	$4B,$FE,$E9,$1A,$23,$FE,$AC,$8D,$57,$38,$D9,$FE,$08,$9F,$A7,$0D
	.dc.b	$BF,$CB,$31,$86,$CE,$CF,$F0,$BF,$BF,$DA,$BF,$39,$8A,$3C,$FE,$5F
	.dc.b	$1C,$BF,$DC,$77,$FE,$16,$8C,$11,$FF,$10,$F7,$F0,$69,$FE,$BF,$F6
	.dc.b	$3F,$63,$E6,$FF,$A1,$23,$FD,$85,$E2,$62,$64,$6C,$C9,$73,$FF,$15
	.dc.b	$F9,$5E,$32,$47,$F8,$A1,$B1,$FB,$5F,$90,$7E,$07,$F4,$27,$F8,$0C
	.dc.b	$62,$95,$FD,$17,$D9,$7D,$07,$CA,$8F,$F3,$CB,$63,$33,$FE,$A3,$6C
	.dc.b	$FF,$26,$6F,$2F,$F8,$73,$C2,$7D,$9B,$E7,$BF,$83,$56,$B0,$F0,$D6
	.dc.b	$A7,$F8,$57,$D9,$3E,$07,$C4,$DA,$FF,$59,$37,$2F,$E6,$AD,$7F,$D9
	.dc.b	$AE,$72,$F1,$64,$FB,$3E,$B8,$7F,$24,$CB,$CB,$1E,$0F,$92,$EE,$4C
	.dc.b	$BB,$B7,$2F,$F0,$CE,$2F,$A5,$7C,$77,$1B,$F9,$73,$FC,$DA,$73,$18
	.dc.b	$3E,$7D,$34,$61,$FC,$DA,$FA,$61,$CE,$BB,$E1,$C9,$FE,$F5,$F6,$4F
	.dc.b	$8F,$FA,$CD,$CB,$FD,$AC,$D9,$FE,$B2,$D7,$FD,$E2,$E7,$0F,$AD,$F5
	.dc.b	$FF,$EB,$6A,$F3,$57,$E1,$2B,$51,$F5,$A7,$53,$FD,$07,$E4,$7F,$13
	.dc.b	$3C,$3C,$44,$37,$FE,$9E,$EA,$B8,$46,$1C,$B6,$B9,$4F,$17,$C7,$2C
	.dc.b	$1F,$2B,$F0,$09,$3F,$01,$B3,$FE,$4E,$B6,$2F,$7B,$FA,$B6,$1C,$B0
	.dc.b	$F7,$D6,$A7,$38,$7E,$16,$F8,$57,$8A,$F8,$61,$D6,$72,$FF,$63,$EC
	.dc.b	$FF,$7A,$57,$C3,$7F,$CD,$CD,$F6,$FF,$80,$F5,$7F,$02,$6C,$64,$E1
	.dc.b	$1F,$C9,$BE,$8F,$F8,$81,$23,$FD,$4A,$DB,$FD,$F7,$D7,$6E,$7E,$97
	.dc.b	$62,$3F,$C0,$3F,$62,$F9,$57,$C8,$61,$7C,$BD,$8C,$37,$0F,$C1,$D7
	.dc.b	$18,$C6,$3F,$E7,$A8,$C4,$EA,$FA,$B6,$B7,$CB,$FE,$9E,$AD,$DF,$C7
	.dc.b	$F4,$BF,$91,$FD,$27,$F2,$46,$E1,$F2,$5B,$3B,$8F,$99,$3F,$CA,$B5
	.dc.b	$FE,$57,$F6,$36,$7E,$83,$6B,$EA,$37,$1F,$EC,$9C,$28,$FC,$1F,$5F
	.dc.b	$EA,$89,$7F,$AA,$6D,$FF,$A6,$21,$F5,$AF,$95,$E3,$1F,$A8,$32,$E2
	.dc.b	$78,$4E,$E7,$F1,$55,$FE,$34,$76,$7E,$7B,$E4,$1F,$63,$FE,$2B,$6C
	.dc.b	$7D,$CB,$C4,$FF,$8B,$FE,$BF,$FC,$67,$A7,$FB,$2B,$89,$F6,$3C,$2F
	.dc.b	$FD,$01,$5A,$C3,$F9,$FD,$6B,$0F,$E8,$7D,$B2,$7F,$80,$CE,$5F,$C5
	.dc.b	$C7,$8F,$C2,$F0,$73,$D9,$FD,$AE,$1F,$8F,$6C,$7F,$94,$2F,$C7,$5C
	.dc.b	$27,$AE,$5F,$C8,$1B,$1B,$8C,$65,$E8,$BC,$4F,$BF,$79,$7F,$C1,$3E
	.dc.b	$C7,$F8,$9E,$BF,$E6,$3C,$FE,$76,$3F,$02,$FC,$5C,$0F,$8D,$E3,$3F
	.dc.b	$ED,$2D,$7E,$DC,$63,$0F,$95,$5B,$CB,$DB,$87,$2A,$FA,$1F,$35,$27
	.dc.b	$C5,$F1,$17,$2F,$9F,$3E,$0F,$47,$19,$74,$6D,$FD,$2C,$91,$8B,$FD
	.dc.b	$B4,$62,$7F,$90,$FD,$8B,$D4,$FF,$37,$36,$3E,$BE,$BF,$E1,$77,$73
	.dc.b	$A3,$E9,$7C,$47,$AF,$0F,$F6,$15,$F1,$7E,$14,$67,$2F,$F4,$64,$E5
	.dc.b	$FC,$64,$66,$FF,$AC,$CE,$5F,$C5,$C6,$6B,$FC,$17,$3F,$C9,$AB,$9F
	.dc.b	$8B,$F8,$3C,$5F,$37,$EA,$89,$8F,$E4,$31,$F8,$18,$F9,$A7,$D3,$A6
	.dc.b	$A3,$CE,$DC,$7C,$AF,$FA,$47,$EC,$7F,$CD,$13,$FE,$DF,$6C,$FE,$94
	.dc.b	$FF,$3E,$C7,$F8,$82,$33,$87,$F7,$46,$C7,$F2,$7F,$31,$F5,$3D,$7D
	.dc.b	$FF,$CE,$46,$2F,$FA,$AD,$F0,$D7,$FC,$09,$7B,$9F,$D5,$D6,$F0,$FF
	.dc.b	$79,$7D,$37,$B9,$FF,$01,$87,$0F,$F8,$19,$AF,$F8,$D9,$AE,$F3,$FF
	.dc.b	$1D,$FD,$37,$FA,$C6,$7E,$4E,$9F,$E8,$1F,$21,$3F,$E6,$EF,$1F,$F8
	.dc.b	$87,$D8,$7F,$8C,$71,$7A,$0F,$81,$FE,$87,$AF,$F1,$BA,$D4,$E6,$F8
	.dc.b	$FB,$3C,$EB,$4F,$FA,$EF,$EC,$1E,$2D,$3F,$CF,$BE,$35,$AF,$F0,$6F
	.dc.b	$C0,$AB,$71,$86,$E9,$FE,$CD,$D3,$FC,$41,$BA,$BF,$D6,$95,$A8,$FF
	.dc.b	$71,$7C,$5F,$0F,$8D,$FE,$FD,$87,$FB,$9F,$3E,$0F,$FB,$E7,$D5,$FF
	.dc.b	$2F,$EE,$FB,$0F,$F2,$A6,$7F,$E6,$2C,$38,$DB,$1E,$98,$7F,$BA,$71
	.dc.b	$E7,$7F,$EE,$D6,$D6,$63,$FC,$3C,$F8,$5A,$F3,$8F,$1D,$F9,$B1,$7C
	.dc.b	$23,$B7,$38,$F3,$C3,$D0,$7D,$69,$D6,$3E,$79,$FF,$3C,$F6,$3E,$C1
	.dc.b	$E8,$72,$C1,$39,$8C,$65,$F7,$FD,$CD,$4E,$71,$E9,$7C,$30,$E7,$1E
	.dc.b	$C9,$D4,$FC,$C3,$58,$7C,$26,$CF,$C0,$FA,$F7,$CF,$FF,$90,$E7,$F3
	.dc.b	$E2,$33,$E8,$62,$31,$3B,$7F,$4B,$3A,$78,$FC,$A7,$F0,$AF,$0B,$83
	.dc.b	$63,$3E,$2F,$F8,$7D,$B3,$D6,$BF,$CD,$19,$DF,$7B,$E2,$D7,$92,$72
	.dc.b	$D8,$CB,$FE,$78,$DF,$EB,$E1,$FE,$7A,$E3,$FB,$5C,$5F,$C7,$2B,$E4
	.dc.b	$D6,$9B,$1F,$ED,$F6,$D7,$1E,$2F,$F7,$33,$63,$EF,$36,$BF,$0B,$F1
	.dc.b	$37,$F0,$2F,$E0,$F8,$FE,$25,$F8,$CD,$E0,$D6,$10,$F1,$CE,$B1,$FD
	.dc.b	$0B,$63,$F8,$0B,$D2,$33,$1D,$E6,$3C,$8F,$26,$4F,$89,$FF,$4D,$F5
	.dc.b	$7E,$BB,$3F,$CD,$FC,$46,$39,$3F,$5F,$68,$C6,$1F,$9B,$C2,$DF,$E8
	.dc.b	$9B,$5B,$9F,$F4,$63,$5F,$F7,$AC,$6B,$E2,$64,$69,$E3,$F0,$18,$8F
	.dc.b	$F7,$DC,$5C,$BC,$D6,$AF,$0F,$8D,$FE,$3B,$BF,$F3,$4F,$A1,$83,$3E
	.dc.b	$9C,$8C,$90,$FF,$77,$3E,$AF,$98,$7B,$B7,$B2,$57,$FA,$96,$72,$D9
	.dc.b	$CA,$CF,$E5,$4F,$EE,$A3,$44,$35,$C1,$FC,$6A,$35,$18,$3F,$9D,$B8
	.dc.b	$9E,$75,$EE,$F9,$EF,$7F,$B9,$2F,$FD,$0F,$19,$7F,$D1,$31,$97,$A7
	.dc.b	$0B,$55,$6E,$73,$1F,$E5,$87,$89,$D0,$F8,$1E,$EF,$DE,$F9,$2F,$EB
	.dc.b	$F9,$7E,$1D,$BA,$7A,$E7,$E5,$1B,$A8,$72,$21,$E4,$9E,$2F,$F5,$A6
	.dc.b	$E2,$D3,$FC,$FD,$3F,$EA,$4F,$B7,$FB,$FE,$9C,$8C,$10,$F5,$B9,$7D
	.dc.b	$B1,$F8,$63,$19,$7F,$6D,$9C,$4F,$FB,$87,$0F,$E3,$13,$ED,$F9,$1F
	.dc.b	$80,$7C,$1E,$89,$5B,$8C,$BC,$63,$2F,$B6,$32,$FB,$E3,$F2,$FF,$29
	.dc.b	$23,$FD,$7E,$FE,$6B,$E5,$24,$60,$87,$D6,$72,$D3,$FD,$8E,$FE,$8B
	.dc.b	$E0,$3F,$47,$F7,$72,$7F,$D9,$6F,$D9,$F3,$3E,$D7,$C1,$59,$F8,$11
	.dc.b	$87,$E1,$7C,$6F,$DF,$23,$FE,$D4,$F3,$BF,$80,$AF,$A8,$63,$85,$8D
	.dc.b	$31,$12,$47,$E7,$CE,$51,$B7,$3B,$7F,$55,$AE,$B1,$8E,$5E,$2F,$86
	.dc.b	$F8,$3C,$E7,$C2,$EC,$EE,$F2,$7F,$47,$FB,$FF,$F6,$1B,$FE,$40,$BE
	.dc.b	$7B,$F8,$DB,$67,$97,$32,$38,$63,$FE,$12,$6C,$FF,$8C,$9E,$0F,$6A
	.dc.b	$E7,$3D,$3E,$93,$AF,$9F,$2B,$AD,$A7,$51,$89,$D3,$E8,$8F,$2F,$97
	.dc.b	$E6,$F6,$24,$F2,$8C,$BD,$23,$2F,$C0,$8C,$BF,$0A,$33,$9F,$C2,$9F
	.dc.b	$BB,$F5,$3F,$CA,$EB,$FB,$FB,$5F,$F2,$FB,$5F,$A6,$D7,$EE,$BC,$67
	.dc.b	$FC,$1C,$DC,$B7,$E1,$7E,$1F,$AF,$E7,$FD,$A7,$D0,$FB,$39,$8F,$C4
	.dc.b	$8C,$BF,$1A,$32,$FD,$A8,$FC,$85,$7E,$47,$C5,$93,$F8,$13,$F2,$D1
	.dc.b	$A7,$1C,$49,$DC,$7E,$26,$74,$FE,$2E,$33,$9F,$E4,$E7,$F9,$17,$E0
	.dc.b	$9F,$C8,$D9,$5F,$C2,$63,$2F,$F0,$B8,$CB,$FC,$7E,$3E,$6F,$57,$F8
	.dc.b	$9E,$1A,$A3,$0D,$8F,$F1,$63,$5F,$A7,$1C,$9F,$74,$66,$FF,$51,$D3
	.dc.b	$FB,$6F,$AB,$7D,$BD,$F9,$09,$87,$C5,$9C,$B6,$72,$FE,$D6,$13,$CE
	.dc.b	$64,$F1,$8D,$DF,$F9,$2E,$B9,$CF,$EF,$7E,$B3,$FD,$C0,$FF,$91,$5A
	.dc.b	$FF,$92,$9A,$FF,$AA,$9A,$FE,$F9,$DC,$FF,$95,$DB,$C5,$9F,$8A,$F7
	.dc.b	$7A,$FF,$C5,$3A,$3C,$F8,$7F,$94,$DA,$FF,$96,$1A,$FF,$AE,$1A,$FF
	.dc.b	$0C,$77,$87,$F9,$95,$B7,$9C,$BF,$90,$FD,$A7,$BA,$BF,$84,$BF,$C2
	.dc.b	$3E,$6E,$61,$FB,$C8,$FE,$16,$61,$FE,$23,$18,$7F,$80,$E5,$FC,$40
	.dc.b	$8F,$F0,$4F,$8B,$AF,$F1,$88,$B7,$FC,$62,$7F,$C5,$FF,$2D,$FE,$30
	.dc.b	$DE,$5F,$F8,$C2,$32,$FF,$8C,$A3,$2F,$F8,$13,$D2,$4F,$F8,$E3,$3F
	.dc.b	$F1,$CF,$A1,$2B,$FC,$73,$19,$7F,$C7,$71,$97,$94,$65,$E9,$19,$C7
	.dc.b	$FC,$6F,$39,$7F,$C7,$31,$9D,$7F,$C9,$B3,$FE,$48,$F4,$7F,$C9,$3E
	.dc.b	$74,$C3,$FC,$93,$19,$7F,$C9,$71,$97,$D5,$1F,$E5,$2F,$85,$FF,$04
	.dc.b	$9F,$DE,$AF,$FD,$8E,$D9,$FF,$73,$BA,$78,$79,$71,$FE,$7F,$19,$6C
	.dc.b	$65,$FF,$15,$C7,$F8,$5B,$D8,$FE,$15,$1F,$D1,$9F,$F1,$6B,$7D,$2F
	.dc.b	$F3,$2C,$6A,$3F,$C6,$6D,$D3,$FC,$E0,$F0,$8C,$CF,$0E,$D9,$1C,$C8
	.dc.b	$72,$72,$F0,$33,$5D,$23,$FC,$EB,$F4,$FF,$E5,$89,$FF,$21,$BF,$E4
	.dc.b	$DC,$BF,$DD,$6D,$87,$90,$8E,$17,$FE,$85,$6D,$70,$D3,$A9,$DB,$79
	.dc.b	$87,$F6,$8A,$E9,$87,$3A,$CC,$FF,$80,$7D,$FF,$F7,$46,$3F,$E0,$4F
	.dc.b	$A5,$7D,$2F,$F7,$E9,$7C,$6F,$F8,$02,$73,$7F,$E0,$A9,$CB,$C6,$B8
	.dc.b	$7C,$2F,$F9,$CE,$15,$C1,$57,$81,$CD,$77,$BE,$CD,$8E,$87,$7B,$E9
	.dc.b	$7C,$EF,$97,$6B,$E9,$BC,$39,$7F,$BF,$1B,$87,$FB,$E9,$B3,$B8,$FF
	.dc.b	$61,$99,$2B,$97,$FD,$AC,$64,$AE,$5F,$F1,$5F,$C0,$79,$9F,$F1,$6F
	.dc.b	$2B,$87,$11,$A3,$DE,$7F,$8A,$A3,$4E,$0E,$F9,$7C,$75,$1A,$8C,$42
	.dc.b	$BA,$CE,$74,$FF,$1E,$EE,$FA,$57,$8F,$DB,$5F,$F6,$74,$7F,$71,$CF
	.dc.b	$FC,$5A,$E9,$BC,$9B,$79,$C6,$73,$FF,$4D,$BC,$5B,$7B,$7F,$33,$19
	.dc.b	$BE,$F3,$9C,$3C,$98,$5A,$E9,$5E,$48,$D7,$33,$DC,$BE,$66,$1B,$9F
	.dc.b	$EF,$DB,$9F,$EF,$1B,$87,$EF,$1B,$3F,$EF,$86,$CF,$FB,$E1,$B9,$30
	.dc.b	$F0,$57,$F0,$BC,$3D,$1D,$2F,$62,$F8,$F8,$6D,$7F,$18,$6E,$5F,$C5
	.dc.b	$9B,$9F,$FB,$F6,$F9,$57,$F1,$EB,$E4,$D3,$F9,$0F,$AF,$E0,$F5,$FC
	.dc.b	$2D,$8C,$9F,$C9,$A3,$2D,$3F,$94,$F0,$FA,$3E,$8B,$D2,$D8,$C9,$FC
	.dc.b	$BA,$32,$D3,$39,$FF,$2E,$8C,$B6,$33,$97,$AE,$72,$D8,$FE,$73,$F1
	.dc.b	$FF,$23,$CF,$F9,$2D,$8C,$9F,$CF,$A3,$2D,$3F,$CE,$7F,$75,$F8,$7F
	.dc.b	$58,$DF,$BA,$C3,$27,$67,$F9,$EA,$31,$97,$FA,$3B,$E8,$79,$7D,$5F
	.dc.b	$36,$D7,$FA,$01,$B3,$B8,$FF,$41,$CF,$E6,$E1,$F7,$EF,$87,$C3,$6B
	.dc.b	$FD,$3D,$F8,$4B,$F0,$97,$B5,$AB,$95,$FE,$A8,$C3,$FD,$53,$87,$17
	.dc.b	$07,$69,$FE,$06,$EE,$34,$F1,$FA,$6F,$F5,$FD,$7F,$9B,$5B,$0E,$1F
	.dc.b	$E1,$7F,$2A,$FB,$D7,$D4,$FA,$EF,$75,$D9,$E3,$7D,$7E,$0F,$C2,$FF
	.dc.b	$AF,$A3,$2D,$8C,$AF,$A5,$7C,$9B,$D8,$FE,$05,$18,$F6,$2C,$E7,$5F
	.dc.b	$F6,$24,$E5,$B1,$9B,$F5,$47,$FB,$5A,$75,$18,$3C,$13,$BC,$BF,$94
	.dc.b	$CE,$9B,$19,$AE,$4A,$FC,$78,$79,$1F,$BD,$F5,$19,$AF,$F4,$C3,$71
	.dc.b	$F0,$6D,$F2,$FE,$87,$77,$2F,$EC,$D2,$F8,$3F,$7D,$AF,$FB,$A1,$AF
	.dc.b	$F7,$EF,$0D,$F0,$9F,$C9,$57,$C2,$D8,$C3,$6F,$FD,$C7,$3A,$6C,$E9
	.dc.b	$F9,$7F,$4F,$F4,$3E,$73,$E8,$B6,$32,$7D,$08,$CB,$4C,$E1,$F3,$F8
	.dc.b	$3E,$EC,$EA,$31,$BA,$3F,$6D,$7F,$B7,$6D,$FF,$BB,$1B,$5F,$F0,$01
	.dc.b	$BC,$1F,$F0,$01,$FA,$3E,$34,$A5,$7D,$FF,$39,$F5,$0F,$A4,$D4,$FE
	.dc.b	$0C,$FA,$17,$F9,$46,$DF,$D4,$E0,$4E,$5B,$1F,$B0,$F3,$1E,$1E,$2F
	.dc.b	$F1,$D1,$8F,$9B,$FE,$25,$87,$F1,$36,$DF,$9B,$5F,$EF,$35,$B8,$C1
	.dc.b	$FD,$CB,$0E,$93,$A5,$5C,$B0,$EF,$18,$AE,$31,$FE,$42,$8E,$31,$CE
	.dc.b	$7F,$92,$7D,$7F,$F6,$BF,$80,$7A,$D8,$AE,$39,$7F,$A3,$23,$18,$FB
	.dc.b	$2B,$75,$AC,$58,$73,$BE,$9E,$BF,$5E,$EB,$7D,$30,$FE,$16,$D9,$D3
	.dc.b	$4D,$4F,$F0,$0D,$7F,$80,$E9,$67,$F7,$ED,$CF,$F1,$6D,$D7,$F7,$2D
	.dc.b	$9F,$DC,$B6,$7F,$80,$BF,$A7,$E9,$93,$E5,$F9,$8B,$5F,$6B,$A7,$F8
	.dc.b	$13,$FA,$FF,$27,$F5,$DD,$DF,$AF,$7E,$57,$24,$C3,$F5,$B1,$6B,$E4
	.dc.b	$76,$FC,$F3,$E8,$7D,$5D,$AF,$89,$1F,$4F,$DC,$C5,$7C,$F7,$13,$F2
	.dc.b	$5C,$ED,$F8,$8E,$2F,$F3,$6C,$FE,$F3,$CB,$FE,$05,$E5,$FF,$03,$6C
	.dc.b	$64,$FE,$03,$19,$69,$9C,$3E,$04,$65,$B1,$F6,$7D,$CF,$C4,$7C,$DF
	.dc.b	$E2,$5B,$19,$3E,$F4,$65,$A6,$67,$F1,$51,$96,$C7,$F1,$4F,$37,$F8
	.dc.b	$D7,$81,$FC,$6D,$B1,$93,$F8,$CC,$65,$A6,$67,$EC,$46,$5B,$1F,$91
	.dc.b	$F9,$BF,$D0,$7C,$DF,$E8,$5B,$19,$3F,$2D,$19,$69,$99,$FD,$14,$65
	.dc.b	$B1,$FC,$B3,$C2,$FE,$65,$D3,$FC,$CD,$B1,$93,$F9,$8C,$65,$A6,$67
	.dc.b	$F2,$11,$96,$C7,$EA,$7C,$CF,$DC,$7C,$DF,$EE,$5B,$19,$3F,$5D,$19
	.dc.b	$69,$FE,$E4,$F9,$0F,$F2,$99,$FE,$54,$F7,$AC,$64,$97,$FE,$E6,$8C
	.dc.b	$BF,$EE,$8F,$77,$FC,$B7,$F0,$BF,$E5,$C6,$D7,$FB,$C1,$B3,$B8,$FF
	.dc.b	$79,$CF,$F9,$E7,$E8,$3F,$AB,$C7,$FA,$83,$BB,$FD,$1F,$1F,$D4,$EB
	.dc.b	$FD,$29,$18,$F3,$3F,$D3,$BF,$02,$8D,$F9,$D8,$75,$87,$3B,$FF,$70
	.dc.b	$E1,$CE,$F8,$46,$30,$E0,$EB,$0B,$E4,$AF,$CB,$3E,$23,$58,$75,$9D
	.dc.b	$5F,$0A,$ED,$8F,$FA,$A3,$83,$FD,$03,$87,$1B,$FE,$93,$F0,$1D,$A5
	.dc.b	$E1,$D8,$99,$75,$9E,$3A,$7A,$2F,$FA,$8D,$EB,$0F,$20,$F9,$D7,$C7
	.dc.b	$7F,$AF,$A3,$FC,$87,$18,$F3,$FF,$C1,$AD,$8C,$CE,$DF,$F2,$67,$C2
	.dc.b	$BB,$3F,$C6,$71,$FE,$4C,$8C,$7A,$3F,$E1,$66,$C6,$67,$6F,$F9,$56
	.dc.b	$FF,$C2,$7E,$6F,$83,$8B,$FA,$6C,$F9,$1A,$ED,$AF,$FB,$43,$E2,$FC
	.dc.b	$4F,$CF,$6B,$F2,$DB,$B7,$C5,$5D,$BF,$E0,$09,$D6,$36,$B4,$DE,$3E
	.dc.b	$CD,$97,$03,$0E,$71,$F7,$FE,$13,$FC,$E1,$87,$D3,$7E,$43,$78,$39
	.dc.b	$5F,$FB,$06,$AD,$70,$C7,$FD,$D6,$F0,$6D,$6A,$61,$DD,$E2,$61,$F8
	.dc.b	$1D,$DE,$DF,$17,$FC,$0F,$97,$FA,$2E,$7F,$BB,$B5,$FF,$08,$A9,$E1
	.dc.b	$CB,$F3,$5B,$3F,$02,$78,$57,$4E,$EC,$E7,$F0,$EE,$D3,$BA,$93,$0F
	.dc.b	$0F,$CE,$D7,$C3,$17,$1F,$3C,$F4,$F9,$D7,$7D,$EF,$98,$7F,$84,$27
	.dc.b	$2D,$9C,$BF,$E2,$C8,$4F,$39,$F3,$FC,$EB,$E3,$2F,$C5,$CC,$3F,$C3
	.dc.b	$73,$96,$CE,$5F,$F1,$BC,$27,$9C,$C9,$E3,$1B,$CB,$FC,$6F,$5C,$E7
	.dc.b	$C9,$E1,$7B,$7C,$09,$1E,$B8,$F8,$06,$1F,$54,$61,$D1,$F0,$5D,$57
	.dc.b	$F9,$13,$35,$7B,$BE,$37,$EA,$FC,$E7,$FB,$0E,$BF,$C9,$2D,$E4,$FF
	.dc.b	$30,$F1,$B3,$5F,$1C,$FF,$C7,$B7,$B7,$F4,$2D,$AE,$13,$B9,$FF,$2D
	.dc.b	$CE,$4D,$57,$48,$FB,$1E,$0F,$DD,$F9,$D9,$18,$21,$F6,$5C,$BF,$6E
	.dc.b	$3F,$08,$63,$2F,$81,$38,$9F,$3F,$3F,$C0,$F8,$C9,$18,$21,$E6,$72
	.dc.b	$FA,$23,$E3,$98,$AF,$8F,$38,$9F,$C8,$7C,$EF,$E5,$48,$FE,$4F,$E6
	.dc.b	$3F,$30,$61,$FC,$94,$61,$D1,$0F,$CB,$47,$E7,$FE,$07,$F4,$1F,$39
	.dc.b	$27,$F3,$CF,$DA,$F8,$FF,$B7,$F1,$F6,$7E,$AC,$61,$FA,$FF,$1F,$F8
	.dc.b	$02,$3F,$A7,$F9,$DF,$DD,$AF,$8D,$63,$85,$8D,$31,$12,$7F,$D0,$E7
	.dc.b	$18,$D3,$89,$E6,$D9,$CB,$FE,$A9,$8E,$AE,$4F,$F4,$27,$2F,$F0,$AF
	.dc.b	$83,$7C,$ED,$31,$12,$7F,$D1,$51,$97,$89,$FE,$8D,$F9,$DE,$5E,$7C
	.dc.b	$8C,$10,$FF,$50,$B9,$7F,$D4,$71,$E1,$F2,$FC,$5E,$94,$9F,$F4,$C4
	.dc.b	$65,$FF,$4D,$46,$5F,$F5,$5C,$65,$FF,$56,$C6,$79,$3F,$D6,$53,$F4
	.dc.b	$3E,$73,$F8,$BD,$7C,$3E,$4F,$A0,$67,$C2,$91,$FE,$FE,$9D,$BF,$D5
	.dc.b	$C3,$87,$FA,$D3,$5F,$F5,$D7,$CE,$FC,$BF,$8F,$91,$FE,$B1,$8F,$9C
	.dc.b	$61,$FF,$61,$C6,$1F,$B5,$5F,$83,$F1,$64,$60,$87,$DB,$72,$FD,$D8
	.dc.b	$FC,$51,$8C,$9B,$FF,$6F,$E1,$FE,$67,$5C,$A7,$F8,$9F,$AD,$FC,$5A
	.dc.b	$E4,$FF,$10,$8F,$E3,$26,$1F,$E1,$B1,$87,$F2,$79,$BA,$7F,$35,$CF
	.dc.b	$27,$31,$F4,$4F,$94,$FC,$86,$A3,$E6,$B8,$69,$AF,$9D,$7C,$D7,$EA
	.dc.b	$97,$CD,$53,$11,$27,$F4,$F3,$A6,$CE,$56,$7F,$19,$F1,$4F,$67,$64
	.dc.b	$37,$AF,$F8,$66,$35,$18,$3D,$5F,$3A,$F8,$7F,$37,$C5,$49,$FD,$CC
	.dc.b	$E5,$B3,$EF,$8C,$47,$C6,$6B,$F9,$16,$BF,$E2,$9E,$8F,$63,$E1,$1F
	.dc.b	$3B,$FD,$7F,$2F,$64,$63,$CD,$FE,$23,$5C,$27,$2F,$29,$FC,$F3,$5E
	.dc.b	$DA,$7D,$38,$C4,$7F,$8C,$BB,$3B,$BF,$66,$BE,$57,$92,$97,$CE,$FF
	.dc.b	$0D,$3F,$87,$65,$2F,$86,$FF,$D8,$AB,$71,$83,$B5,$6D,$F9,$D7,$D7
	.dc.b	$5F,$E3,$04,$8E,$11,$E1,$F8,$BC,$47,$0B,$F9,$B9,$79,$9B,$8F,$86
	.dc.b	$BF,$C5,$3F,$0D,$FE,$2A,$E2,$98,$7F,$8E,$23,$2F,$F8,$EA,$32,$FF
	.dc.b	$8E,$A3,$FC,$63,$B7,$F8,$CB,$79,$3F,$E4,$08,$CB,$FE,$42,$8C,$BF
	.dc.b	$E4,$28,$CB,$FE,$44,$8C,$E5,$FE,$4A,$9C,$BF,$E4,$C8,$CE,$3F,$E7
	.dc.b	$39,$FF,$20,$FA,$1F,$E4,$2F,$36,$61,$FE,$50,$8C,$BF,$E5,$28,$CB
	.dc.b	$FE,$52,$8F,$F2,$5F,$BF,$FE,$0D,$3F,$D2,$18,$FD,$C6,$E1,$FE,$A9
	.dc.b	$AD,$9C,$3C,$A8,$FD,$29,$E4,$E0,$FB,$2F,$27,$FC,$F6,$FF,$86,$FD
	.dc.b	$8F,$E2,$B1,$F7,$9F,$F1,$5B,$F7,$1A,$E7,$BB,$FC,$C3,$1C,$30,$FC
	.dc.b	$8B,$A6,$FA,$1A,$99,$5C,$E3,$89,$87,$A1,$92,$18,$3F,$23,$F3,$BF
	.dc.b	$AD,$CB,$FC,$72,$FF,$94,$E3,$F2,$2D,$77,$B7,$F9,$DA,$38,$47,$FA
	.dc.b	$69,$D6,$1F,$A8,$6D,$74,$AE,$07,$1D,$25,$F6,$8F,$F0,$5F,$BE,$F8
	.dc.b	$5F,$E3,$72,$F9,$D7,$CE,$7F,$AB,$E9,$F1,$5F,$EB,$19,$CD,$7F,$31
	.dc.b	$9C,$BF,$CC,$E3,$2A,$3B,$69,$9B,$E9,$87,$82,$F7,$5C,$AF,$C5,$5E
	.dc.b	$1C,$3C,$0F,$0C,$39,$4F,$87,$0F,$4F,$C1,$3E,$43,$52,$8C,$3B,$AF
	.dc.b	$E9,$11,$88,$F3,$57,$2A,$FE,$9F,$F0,$3A,$6C,$F3,$6D,$71,$8F,$EA
	.dc.b	$53,$B7,$FD,$96,$7A,$8F,$F6,$C7,$27,$F4,$7D,$3E,$2C,$70,$34,$7F
	.dc.b	$5B,$8D,$65,$E0,$9D,$C6,$27,$AC,$63,$7F,$87,$D6,$F9,$DF,$F5,$8C
	.dc.b	$7F,$1C,$7F,$D5,$6D,$F1,$FF,$AC,$6D,$FC,$C2,$73,$87,$F2,$67,$8B
	.dc.b	$79,$77,$7F,$E9,$09,$CD,$71,$9C,$D7,$FB,$BA,$73,$5B,$9C,$E1,$E5
	.dc.b	$C2,$D7,$83,$0F,$D8,$FB,$9A,$E7,$7B,$36,$30,$DC,$BF,$78,$DC,$FF
	.dc.b	$80,$36,$FF,$76,$D9,$FF,$7A,$B6,$59,$30,$EF,$5F,$C1,$F0,$EE,$FF
	.dc.b	$13,$F6,$3D,$3D,$4F,$5E,$C6,$1B,$3F,$C6,$5B,$8F,$F1,$86,$E7,$FC
	.dc.b	$8D,$B3,$FE,$BA,$BE,$8F,$F2,$1F,$63,$BF,$97,$E0,$6C,$64,$FE,$4D
	.dc.b	$19,$69,$FC,$A7,$8F,$CF,$F3,$BE,$86,$C6,$4F,$E5,$D1,$96,$99,$CB
	.dc.b	$F9,$7C,$65,$B1,$9C,$BF,$9E,$4E,$5B,$1F,$CE,$7E,$77,$E3,$FA,$1F
	.dc.b	$21,$B1,$93,$F9,$F4,$65,$A7,$F7,$2F,$19,$F2,$3F,$88,$6C,$67,$E3
	.dc.b	$F1,$32,$38,$3B,$8C,$65,$FD,$E7,$E3,$F9,$7C,$7F,$36,$DF,$76,$CE
	.dc.b	$E3,$C5,$3F,$9C,$87,$CD,$BE,$23,$FC,$4B,$1F,$E5,$13,$2E,$1F,$F2
	.dc.b	$A4,$6E,$BF,$CA,$ED,$8F,$F2,$DB,$75,$FF,$25,$7C,$CB,$F0,$27,$CA
	.dc.b	$F8,$6B,$58,$7F,$97,$EB,$53,$E5,$F9,$BF,$F6,$1C,$FF,$99,$1B,$8F
	.dc.b	$F9,$A1,$AF,$FA,$4D,$AF,$A5,$DE,$3F,$C1,$3E,$91,$CF,$D7,$E0,$FF
	.dc.b	$7B,$29,$CE,$1F,$E2,$FF,$9C,$7C,$9F,$F9,$66,$30,$BD,$4F,$E0,$B3
	.dc.b	$AF,$56,$CE,$6B,$71,$8A,$CB,$63,$27,$48,$CB,$4F,$F6,$EF,$CD,$F3
	.dc.b	$D7,$FD,$0E,$D7,$FD,$64,$D7,$EC,$35,$FF,$45,$3C,$72,$FF,$3E,$79
	.dc.b	$0F,$9C,$7C,$97,$E0,$EB,$58,$7F,$37,$9C,$BF,$E4,$AF,$19,$F0,$BF
	.dc.b	$C8,$17,$C3,$58,$C3,$67,$83,$73,$FF,$2B,$D7,$09,$F9,$7F,$31,$F2
	.dc.b	$3E,$63,$E4,$B6,$32,$7D,$08,$CB,$4C,$D7,$FA,$3F,$87,$EB,$CE,$A3
	.dc.b	$1C,$28,$FD,$85,$FE,$C5,$B7,$FB,$56,$D7,$F9,$95,$BB,$FF,$99,$5F
	.dc.b	$8D,$C9,$FA,$39,$6D,$F8,$87,$EA,$C6,$63,$11,$F5,$A3,$FA,$5D,$BE
	.dc.b	$9F,$FA,$5C,$FD,$69,$CC,$63,$CB,$FF,$44,$47,$F0,$7F,$9C,$FE,$10
	.dc.b	$DB,$FE,$18,$DB,$FF,$3E,$37,$2F,$F3,$E3,$67,$FD,$18,$D9,$FF,$4C
	.dc.b	$7C,$EF,$F5,$AF,$7B,$FC,$19,$5F,$C5,$E3,$1E,$A7,$F1,$76,$C6,$4F
	.dc.b	$F4,$6C,$F0,$AF,$F4,$9B,$FE,$17,$8C,$FC,$27,$5F,$5B,$B3,$63,$27
	.dc.b	$28,$CB,$FA,$EC,$7F,$57,$9F,$EB,$1B,$19,$3F,$67,$19,$69,$9C,$3F
	.dc.b	$03,$19,$6C,$7E,$6F,$A7,$F3,$47,$C1,$E8,$B1,$F9,$FE,$AF,$CF,$78
	.dc.b	$DF,$9F,$6C,$7C,$0E,$FF,$90,$63,$0F,$82,$FC,$BF,$9C,$93,$F1,$BE
	.dc.b	$76,$D7,$C6,$9F,$A1,$ED,$63,$8F,$FD,$8F,$C3,$FE,$54,$C3,$E6,$CE
	.dc.b	$9F,$9F,$56,$FF,$57,$E4,$FE,$CF,$C9,$FD,$A3,$6B,$F7,$0D,$9D,$C7
	.dc.b	$EE,$E7,$6D,$36,$58,$FB,$7E,$D7,$E1,$7E,$6B,$F0,$CD,$9F,$C0,$B6
	.dc.b	$77,$1F,$84,$9D,$B4,$D9,$63,$F8,$1F,$97,$FC,$2F,$B7,$F8,$63,$67
	.dc.b	$F8,$83,$67,$71,$FC,$4E,$76,$D3,$65,$8F,$C9,$FC,$DF,$E7,$7E,$6F
	.dc.b	$F3,$CD,$9F,$CC,$B6,$77,$1F,$9C,$9D,$B4,$D9,$63,$F8,$DF,$7F,$F2
	.dc.b	$3E,$8F,$E4,$8D,$9F,$E5,$0D,$9D,$C7,$F2,$B9,$DB,$4D,$96,$3F,$57
	.dc.b	$E6,$FE,$D7,$E6,$BF,$6C,$D9,$FD,$8B,$67,$71,$FB,$49,$FF,$1E,$FC
	.dc.b	$74,$8F,$F6,$97,$B9,$6B,$FD,$A3,$37,$97,$FD,$B9,$36,$BF,$C7,$5E
	.dc.b	$EF,$F9,$1B,$E1,$3F,$C8,$ED,$8F,$F6,$E3,$70,$FF,$6D,$B6,$D2,$9F
	.dc.b	$F2,$AF,$CE,$FF,$4A,$8F,$F4,$6F,$6E,$7C,$44,$FF,$45,$A7,$C0,$FF
	.dc.b	$56,$78,$F9,$9D,$1C,$DF,$1A,$E1,$5B,$BE,$77,$B8,$C4,$EA,$B8,$4E
	.dc.b	$5E,$5E,$65,$C3,$BE,$3E,$67,$D3,$19,$C7,$AC,$E5,$F4,$C6,$78,$5E
	.dc.b	$F5,$31,$E6,$DB,$E1,$3A,$6C,$E9,$A7,$87,$07,$CE,$BE,$31,$E8,$FF
	.dc.b	$8D,$E7,$FC,$71,$39,$8C,$47,$D1,$8C,$38,$6B,$C7,$E1,$1D,$BF,$E2
	.dc.b	$08,$E3,$18,$F4,$39,$C7,$FB,$86,$74,$F1,$6B,$E2,$E0,$FF,$6B,$3D
	.dc.b	$DA,$F9,$EF,$FC,$17,$E7,$74,$6C,$7A,$F8,$3F,$C6,$FF,$31,$FE,$53
	.dc.b	$E4,$FB,$9C,$0F,$94,$C7,$27,$FA,$3D,$B8,$7F,$A8,$A5,$78,$C3,$9C
	.dc.b	$FD,$0A,$C5,$73,$8F,$A7,$5B,$F9,$DF,$87,$1F,$75,$AF,$E4,$0F,$F5
	.dc.b	$7B,$8C,$7F,$20,$BC,$0E,$5B,$4A,$E4,$F5,$8C,$47,$3C,$95,$CF,$3F
	.dc.b	$D2,$36,$7B,$D7,$A3,$E4,$BF,$CE,$DD,$3F,$A3,$C3,$FC,$BB,$EC,$63
	.dc.b	$9B,$FD,$64,$DC,$3E,$25,$DA,$FF,$31,$B7,$65,$53,$11,$9F,$98,$97
	.dc.b	$FB,$BC,$7C,$16,$9E,$5F,$33,$AA,$56,$A7,$0F,$09,$F1,$79,$8F,$8B
	.dc.b	$BF,$19,$30,$FF,$06,$C6,$8C,$3F,$E0,$E8,$C3,$E3,$D7,$F8,$57,$CE
	.dc.b	$3C,$3F,$72,$F0,$A9,$88,$95,$C4,$FF,$74,$CE,$DC,$70,$23,$11,$EC
	.dc.b	$BE,$F5,$CA,$30,$72,$BE,$EF,$2A,$F2,$3F,$E2,$8F,$8F,$FA,$5F,$5D
	.dc.b	$FB,$9D,$BE,$D3,$71,$FE,$49,$AA,$BE,$39,$C2,$E1,$CC,$87,$8B,$35
	.dc.b	$F3,$84,$61,$E6,$79,$7C,$5F,$8F,$E1,$48,$C1,$0E,$99,$FF,$1C,$7D
	.dc.b	$91,$F3,$A6,$46,$08,$70,$7D,$FE,$0F,$C0,$F8,$C9,$3F,$3B,$1F,$E3
	.dc.b	$6D,$9C,$BE,$B8,$4F,$28,$D1,$C5,$DE,$BF,$2A,$BF,$03,$E2,$7E,$0B
	.dc.b	$B6,$5F,$D3,$8C,$BF,$56,$32,$FB,$A3,$2F,$C0,$8C,$D7,$EB,$F3,$57
	.dc.b	$DA,$F8,$C9,$7E,$B7,$EE,$FC,$64,$8C,$10,$F6,$39,$69,$F0,$DF,$C2
	.dc.b	$FC,$67,$E1,$BE,$32,$4F,$C7,$7F,$3B,$E1,$7E,$7B,$B2,$CF,$E3,$E3
	.dc.b	$0F,$E4,$7B,$7F,$46,$47,$E8,$7C,$67,$F0,$55,$FA,$3B,$E7,$30,$BE
	.dc.b	$72,$98,$89,$3F,$A9,$38,$C6,$9C,$4F,$36,$CE,$5F,$D3,$9C,$0C,$9F
	.dc.b	$83,$F8,$FF,$D2,$FC,$7C,$8C,$10,$FC,$23,$97,$F0,$D1,$FC,$7F,$91
	.dc.b	$F9,$2B,$DC,$C2,$F7,$A9,$88,$97,$FC,$0E,$B8,$CE,$A3,$15,$CD,$FC
	.dc.b	$C7,$A9,$FB,$AF,$0A,$46,$08,$7E,$65,$CB,$F9,$B8,$FE,$02,$63,$A3
	.dc.b	$F8,$0C,$E2,$7E,$67,$CD,$B8,$3F,$B8,$C2,$30,$FF,$74,$6B,$FE,$00
	.dc.b	$6B,$FE,$05,$73,$5F,$E2,$98,$E5,$3F,$81,$9C,$F8,$5F,$E3,$3E,$DF
	.dc.b	$E3,$F7,$FC,$53,$BB,$F8,$BF,$7C,$C7,$F7,$11,$FC,$68,$C3,$FB,$78
	.dc.b	$C3,$F8,$5E,$1F,$E9,$AD,$F1,$FF,$21,$1F,$D2,$63,$5C,$D8,$33,$32
	.dc.b	$32,$43,$F1,$33,$C5,$B1,$C5,$FE,$73,$F2,$3F,$99,$6F,$A9,$FD,$86
	.dc.b	$3F,$C5,$51,$AE,$8C,$19,$99,$19,$3E,$9F,$07,$D2,$DF,$F9,$DC,$7C
	.dc.b	$F8,$C1,$1D,$90,$F3,$F7,$FC,$08,$F8,$3E,$34,$9F,$8D,$19,$7E,$44
	.dc.b	$65,$F7,$C6,$5F,$83,$19,$DF,$EE,$69,$E4,$7E,$F6,$0E,$CF,$41,$2F
	.dc.b	$F1,$0D,$7F,$67,$1B,$7F,$23,$1F,$5B,$C0,$FE,$0D,$F3,$4F,$83,$FC
	.dc.b	$9B,$6B,$FC,$8A,$D7,$FC,$A2,$D7,$FC,$AC,$E7,$0F,$C9,$37,$97,$FD
	.dc.b	$0D,$28,$C3,$FE,$89,$F9,$BF,$F2,$4D,$7F,$66,$DB,$FB,$4F,$8B,$2B
	.dc.b	$FA,$54,$7F,$6C,$30,$FF,$49,$8C,$3F,$D3,$FC,$1F,$EE,$24,$7F,$A9
	.dc.b	$69,$5F,$EA,$11,$6F,$FB,$AC,$FF,$8A,$7F,$76,$FD,$54,$7F,$8C,$5F
	.dc.b	$E3,$D5,$FD,$C5,$B1,$9C,$F8,$1C,$7C,$19,$5F,$E0,$EA,$E7,$3F,$EA
	.dc.b	$37,$F3,$11,$CF,$8F,$C0,$78,$79,$24,$71,$8C,$BC,$E3,$2F,$13,$46
	.dc.b	$48,$64,$98,$7A,$E3,$DD,$EB,$7B,$FD,$99,$5E,$78,$CB,$E9,$8C,$BF
	.dc.b	$06,$3E,$37,$CF,$7C,$7F,$9E,$93,$EF,$8C,$BF,$06,$32,$FC,$98,$CB
	.dc.b	$F2,$E3,$38,$7C,$F9,$FF,$2C,$4F,$F9,$67,$C3,$8F,$F8,$E3,$D2,$FF
	.dc.b	$10,$79,$A6,$9F,$E0,$E8,$FF,$13,$36,$FF,$CB,$71,$D0,$E5,$5D,$1D
	.dc.b	$99,$38,$B9,$8E,$31,$FE,$48,$F9,$7F,$E5,$3B,$FF,$B7,$57,$83,$FE
	.dc.b	$71,$6C,$EF,$B3,$06,$66,$46,$48,$7F,$8F,$67,$8B,$63,$8B,$FD,$CF
	.dc.b	$DC,$7C,$27,$F9,$CA,$1F,$32,$F3,$7F,$5F,$4F,$88,$FD,$8C,$E6,$BF
	.dc.b	$DF,$13,$97,$FD,$F5,$19,$51,$DB,$4C,$DF,$4C,$3C,$17,$BA,$E5,$7E
	.dc.b	$2A,$F0,$E1,$E0,$78,$61,$CA,$7C,$38,$7A,$7D,$E7,$C6,$EA,$51,$87
	.dc.b	$74,$73,$3B,$9D,$47,$FB,$2C,$C9,$FE,$E9,$F7,$B8,$46,$A3,$94,$62
	.dc.b	$7F,$DD,$6D,$8E,$07,$AC,$FE,$33,$C0,$E0,$F5,$47,$7D,$7E,$43,$FE
	.dc.b	$F8,$8C,$BD,$8E,$47,$A6,$79,$38,$35,$3C,$A3,$2F,$D7,$E2,$70,$3C
	.dc.b	$9E,$3A,$7F,$B1,$63,$FC,$AF,$8F,$E5,$67,$2F,$E5,$7B,$EE,$5F,$E8
	.dc.b	$6A,$E5,$3F,$9B,$9C,$E1,$DA,$ED,$74,$BF,$CA,$46,$2B,$2D,$BF,$DA
	.dc.b	$7A,$BC,$B9,$5E,$25,$8C,$36,$7F,$74,$DC,$3F,$6E,$DC,$FF,$DA,$2D
	.dc.b	$9F,$C0,$67,$FB,$F8,$C5,$7F,$A8,$30,$EF,$5E,$23,$A9,$FC,$C7,$E7
	.dc.b	$BD,$BD,$CF,$9E,$B1,$86,$CF,$F3,$76,$DF,$F3,$F6,$E5,$FD,$0D,$B3
	.dc.b	$FC,$ED,$EB,$E3,$76,$F8,$EB,$3F,$D0,$5B,$7F,$D2,$5B,$7F,$D3,$5F
	.dc.b	$2F,$C7,$79,$BE,$36,$DF,$F4,$96,$DF,$F5,$16,$DF,$F5,$56,$CF,$F4
	.dc.b	$D6,$CF,$FB,$13,$34,$63,$64,$62,$BF,$AB,$61,$E1,$7F,$AC,$7C,$6F
	.dc.b	$C8,$F8,$DF,$92,$D8,$C9,$FD,$76,$32,$D3,$FD,$81,$E8,$FE,$33,$D2
	.dc.b	$37,$D2,$C3,$27,$67,$FB,$1E,$31,$A7,$FB,$17,$E0,$7F,$2B,$FC,$01
	.dc.b	$FB,$1A,$FF,$66,$46,$1E,$CD,$72,$7F,$BF,$A3,$84,$FF,$A5,$3E,$63
	.dc.b	$C7,$3F,$E9,$D6,$C7,$FB,$FB,$E6,$5F,$46,$F9,$4F,$D4,$CA,$B7,$C7
	.dc.b	$FE,$6F,$78,$C7,$F8,$78,$D6,$9F,$E7,$37,$FC,$43,$1A,$6C,$65,$FE
	.dc.b	$39,$F2,$EE,$5E,$ED,$CF,$F9,$06,$1F,$C2,$33,$F2,$FB,$DA,$B9,$5E
	.dc.b	$1C,$39,$C7,$55,$87,$37,$4D,$7F,$C6,$BF,$30,$F8,$E7,$CA,$7F,$08
	.dc.b	$8C,$7C,$9D,$9C,$E3,$E4,$3F,$9A,$4E,$9C,$5A,$3A,$5F,$1B,$E4,$DA
	.dc.b	$DC,$E9,$B6,$AD,$C6,$1F,$8B,$DD,$23,$FD,$79,$F3,$2F,$96,$C6,$5F
	.dc.b	$ED,$2C,$FF,$C0,$D5,$F1,$63,$E1,$DF,$C5,$9C,$C7,$C4,$8F,$A9,$D8
	.dc.b	$F8,$07,$CC,$7F,$26,$F8,$CF,$EA,$B8,$FF,$A4,$A3,$1E,$6F,$F2,$AA
	.dc.b	$DC,$FC,$2B,$5F,$0C,$5C,$9F,$D0,$A7,$B6,$3C,$AF,$E0,$E5,$F3,$FE
	.dc.b	$63,$E8,$36,$32,$7C,$68,$CB,$4C,$E1,$D9,$FA,$BC,$09,$DC,$E4,$FA
	.dc.b	$D1,$99,$FD,$47,$B7,$FB,$0F,$6B,$F6,$2D,$8C,$9F,$AD,$8C,$B4,$FC
	.dc.b	$0D,$23,$FB,$EC,$61,$B1,$E1,$C9,$D8,$F6,$FF,$5F,$5A,$D7,$FD,$B8
	.dc.b	$F0,$BF,$D7,$35,$FF,$46,$B5,$FF,$03,$36,$FA,$FB,$5F,$C2,$BD,$BF
	.dc.b	$E1,$6D,$CB,$F8,$33,$6F,$C9,$92,$31,$8F,$D7,$8C,$57,$F0,$DB,$E6
	.dc.b	$D3,$F4,$3E,$8B,$F3,$1F,$D3,$E3,$F3,$E3,$1D,$1F,$DD,$1D,$75,$5A
	.dc.b	$FD,$05,$F2,$78,$E5,$FE,$08,$AE,$33,$9A,$E1,$7F,$C9,$BD,$BF,$E4
	.dc.b	$FE,$E5,$AF,$E4,$4D,$CB,$DA,$DC,$7D,$8F,$EF,$79,$7F,$80,$47,$E5
	.dc.b	$CF,$F0,$BF,$67,$F0,$58,$CF,$4E,$F0,$FD,$CD,$6D,$FC,$7C,$E9,$FD
	.dc.b	$DB,$70,$FC,$85,$F3,$9E,$37,$FA,$DF,$88,$91,$FC,$E7,$CF,$B7,$F8
	.dc.b	$7E,$A7,$81,$FB,$3F,$6B,$F6,$8D,$BF,$E8,$1E,$87,$F4,$83,$17,$F7
	.dc.b	$1F,$E9,$7E,$84,$9F,$E8,$DE,$1D,$AF,$C0,$CF,$F5,$0F,$3B,$1C,$1F
	.dc.b	$81,$71,$7F,$D3,$5C,$BF,$E1,$6E,$84,$7F,$07,$F6,$7F,$84,$7C,$A5
	.dc.b	$AF,$E0,$9B,$23,$19,$7B,$5B,$8F,$F0,$59,$D5,$70,$BE,$6D,$39,$D7
	.dc.b	$F6,$6F,$4B,$FB,$3F,$A3,$67,$FB,$13,$71,$FC,$EB,$71,$FC,$E3,$67
	.dc.b	$D4,$D9,$77,$7E,$DA,$3F,$91,$7C,$97,$F2,$3F,$99,$B7,$EC,$6D,$FD
	.dc.b	$76,$E1,$EB,$6C,$FF,$8F,$34,$46,$2B,$BC,$62,$7F,$BD,$7A,$7F,$E0
	.dc.b	$2E,$DF,$F0,$1B,$63,$27,$F7,$F8,$CB,$4C,$D7,$EF,$23,$2D,$8F,$A7
	.dc.b	$F3,$3F,$CF,$7E,$67,$F9,$F3,$63,$27,$D8,$8C,$B4,$CC,$FE,$56,$32
	.dc.b	$D8,$FF,$09,$77,$FF,$86,$BC,$9F,$F0,$DB,$63,$27,$F8,$5E,$32,$D3
	.dc.b	$FA,$57,$AD,$FC,$09,$FF,$31,$B6,$BF,$C7,$D5,$FE,$5E,$C6,$57,$07
	.dc.b	$FC,$C3,$18,$9F,$E6,$15,$FD,$83,$5F,$EA,$4F,$F4,$18,$CE,$1F,$E6
	.dc.b	$8F,$2B,$FC,$D2,$DC,$7C,$6D,$9F,$F7,$7F,$B3,$FE,$9C,$F7,$5F,$2F
	.dc.b	$FE,$78,$AF,$F4,$7D,$27,$2F,$F9,$E7,$DC,$E1,$5B,$BB,$3F,$E7,$D6
	.dc.b	$F2,$FF,$BF,$DB,$87,$FA,$57,$0E,$73,$9C,$7F,$D7,$B8,$73,$6E,$1C
	.dc.b	$DB,$2B,$E5,$5B,$AF,$33,$F3,$BC,$4F,$0F,$EE,$FF,$DE,$71,$EF,$AF
	.dc.b	$F7,$EB,$D9,$FA,$69,$D5,$F1,$CB,$BC,$62,$7F,$CE,$98,$74,$F1,$3D
	.dc.b	$F0,$F9,$9F,$EC,$5E,$4F,$F8,$5D,$B1,$FE,$3D,$6B,$FE,$4F,$6B,$FF
	.dc.b	$00,$3B,$3F,$C7,$4F,$F9,$0F,$E0,$FF,$87,$7A,$3F,$E2,$16,$C6,$67
	.dc.b	$6F,$3F,$3B,$FD,$E2,$F7,$6B,$9C,$7C,$91,$88,$FB,$DF,$11,$FE,$F5
	.dc.b	$7E,$1F,$17,$CD,$F8,$8C,$71,$7D,$6E,$34,$63,$8B,$8D,$FD,$49,$61
	.dc.b	$98,$FF,$5F,$55,$C3,$A4,$6E,$71,$F3,$3F,$CA,$E3,$EE,$B5,$F1,$B7
	.dc.b	$1F,$C3,$AF,$03,$8E,$B2,$B8,$BD,$23,$11,$CA,$2D,$72,$9F,$26,$4F
	.dc.b	$89,$F8,$5D,$5F,$E3,$D8,$FF,$50,$E1,$FE,$43,$6B,$FE,$21,$78,$39
	.dc.b	$9E,$3C,$DF,$A9,$F6,$EC,$FF,$B7,$DB,$AF,$E6,$6A,$DB,$E0,$29,$2F
	.dc.b	$C1,$F0,$15,$F0,$45,$CB,$CD,$3D,$3C,$37,$CC,$DF,$99,$98,$7F,$58
	.dc.b	$8D,$18,$7F,$BA,$46,$1F,$F3,$07,$CC,$3E,$47,$CE,$BE,$36,$98,$89
	.dc.b	$3C,$23,$2F,$28,$CB,$FE,$47,$9E,$4E,$CD,$1F,$DB,$3E,$67,$E2,$FC
	.dc.b	$3F,$FA,$CF,$4F,$F3,$73,$78,$BF,$A8,$4F,$BC,$DD,$43,$91,$0F,$F2
	.dc.b	$B4,$F1,$7F,$B1,$B7,$87,$FC,$FB,$3E,$EF,$8E,$F7,$F8,$92,$FE,$1C
	.dc.b	$65,$F8,$B1,$97,$D9,$19,$7D,$D1,$9C,$7F,$D0,$D9,$2B,$DB,$E0,$FB
	.dc.b	$BE,$2A,$61,$FE,$8F,$CD,$3A,$8C,$4E,$9F,$5C,$65,$F6,$C6,$70,$F8
	.dc.b	$F3,$F6,$3C,$6F,$B3,$D9,$2F,$EE,$C6,$5F,$BF,$19,$7E,$B4,$65,$FB
	.dc.b	$11,$9A,$F5,$CF,$D6,$F8,$A9,$5E,$97,$EC,$FC,$CC,$8C,$10,$FA,$6E
	.dc.b	$5A,$7C,$17,$F0,$3F,$33,$F8,$2F,$90,$93,$F1,$5F,$C9,$F8,$3F,$94
	.dc.b	$EC,$B3,$F9,$88,$C3,$F9,$AE,$DF,$CC,$91,$F9,$DF,$2B,$FB,$45,$F3
	.dc.b	$2F,$87,$C3,$FE,$A1,$8F,$F5,$1B,$5F,$F5,$81,$C2,$36,$70,$F8,$69
	.dc.b	$1C,$23,$2F,$28,$CB,$CA,$7C,$0F,$E0,$7E,$67,$F4,$3F,$2F,$23,$04
	.dc.b	$3F,$36,$E5,$FC,$EC,$7F,$0E,$E3,$7C,$CF,$F9,$BA,$7F,$D3,$6B,$E1
	.dc.b	$34,$D8,$CF,$C1,$60,$CC,$C8,$C9,$F9,$5F,$99,$FD,$AF,$81,$23,$04
	.dc.b	$3F,$5E,$E5,$FD,$8C,$7E,$F0,$C7,$3F,$EF,$27,$13,$FE,$D0,$F9,$97
	.dc.b	$07,$F2,$6F,$A8,$FE,$B3,$7F,$EE,$F6,$F8,$5F,$D5,$EB,$77,$C6,$FF
	.dc.b	$AD,$32,$32,$E0,$38,$75,$19,$7F,$82,$F7,$7F,$12,$F0,$24,$7E,$D6
	.dc.b	$3F,$8A,$98,$7F,$86,$C6,$1F,$F7,$4F,$0F,$F4,$6C,$98,$FF,$6D,$8F
	.dc.b	$E7,$D3,$AF,$23,$F2,$51,$B9,$D7,$36,$26,$57,$48,$D9,$87,$F2,$75
	.dc.b	$D1,$B1,$FD,$3B,$E3,$BF,$31,$87,$8A,$7F,$D0,$71,$FE,$F7,$47,$FB
	.dc.b	$ED,$C3,$79,$B9,$1C,$FE,$3A,$47,$0A,$E9,$E2,$AF,$BE,$7F,$57,$87
	.dc.b	$DB,$D5,$30,$FB,$D1,$FB,$1E,$1F,$E9,$D1,$F4,$5A,$FD,$76,$BF,$6E
	.dc.b	$3E,$B2,$D7,$79,$7B,$63,$1E,$37,$CC,$3E,$6D,$48,$FA,$B1,$97,$EB
	.dc.b	$C6,$5F,$97,$19,$7E,$6C,$79,$FB,$3E,$BF,$74,$9F,$CC,$46,$5F,$53
	.dc.b	$E6,$C7,$FC,$0B,$1E,$B8,$F4,$FC,$D3,$E1,$BF,$89,$7C,$D3,$E5,$A9
	.dc.b	$FC,$56,$75,$7F,$3F,$7F,$F1,$B4,$62,$3F,$38,$D7,$F4,$2D,$77,$7F
	.dc.b	$9D,$E1,$79,$0F,$1B,$FB,$7F,$8F,$2F,$FA,$94,$7F,$72,$30,$FF,$6A
	.dc.b	$8C,$3F,$D5,$F8,$FF,$BB,$91,$FE,$B5,$AD,$7F,$B8,$C5,$E2,$FE,$F7
	.dc.b	$3F,$E2,$2F,$73,$EC,$47,$FA,$6B,$C5,$C1,$9D,$A6,$1F,$EA,$A8,$CB
	.dc.b	$FE,$AC,$8E,$5D,$1C,$FA,$24,$FF,$AC,$23,$2F,$FA,$CA,$32,$FF,$AA
	.dc.b	$A3,$2F,$FA,$B2,$33,$7F,$EA,$C9,$CB,$FE,$AE,$8C,$D7,$FA,$EA,$7D
	.dc.b	$5E,$B7,$AF,$E4,$66,$1F,$EC,$08,$CB,$FE,$C2,$8C,$BF,$EB,$A8,$F8
	.dc.b	$5F,$21,$F0,$FE,$42,$4F,$FB,$22,$32,$FF,$B2,$A3,$2F,$FB,$0A,$32
	.dc.b	$FF,$B1,$23,$39,$7F,$B1,$27,$FC,$9B,$3F,$E4,$EF,$1E,$3F,$E5,$B3
	.dc.b	$27,$F9,$75,$CB,$FE,$5D,$32,$7F,$97,$DC,$CF,$F9,$96,$32,$FF,$99
	.dc.b	$E3,$39,$7F,$B2,$67,$FC,$BF,$F3,$1F,$E4,$05,$5F,$C2,$23,$FC,$49
	.dc.b	$1F,$9D,$1C,$88,$CC,$70,$AF,$E2,$F3,$9E,$CE,$A7,$6F,$0A,$57,$39
	.dc.b	$D3,$6F,$C5,$87,$86,$B7,$3D,$4F,$0C,$EB,$0F,$F5,$27,$B8,$F8,$AF
	.dc.b	$4D,$3E,$65,$E7,$6B,$E4,$3F,$D5,$51,$89,$FE,$5B,$69,$CD,$74,$BE
	.dc.b	$77,$BB,$5F,$7C,$FF,$9A,$CE,$63,$94,$E2,$FC,$0D,$3B,$DF,$48,$ED
	.dc.b	$7F,$CF,$BD,$F7,$CD,$62,$D1,$87,$8C,$EE,$75,$39,$9F,$F6,$19,$C2
	.dc.b	$3F,$B6,$7B,$BB,$8F,$EE,$2D,$77,$3C,$5F,$E9,$F1,$E7,$7F,$A8,$F0
	.dc.b	$FF,$55,$CF,$E1,$46,$8F,$9D,$8F,$F7,$C9,$D3,$E0,$8C,$9F,$ED,$98
	.dc.b	$CB,$E2,$32,$76,$D7,$E3,$70,$7E,$37,$CA,$FE,$C9,$A7,$E4,$FB,$AD
	.dc.b	$7E,$3B,$85,$58,$C3,$F0,$B5,$70,$E5,$7F,$8D,$8C,$5F,$08,$C1,$F9
	.dc.b	$1B,$CC,$E9,$B1,$FE,$CD,$F5,$7A,$7B,$8F,$5A,$C6,$1B,$3F,$ED,$36
	.dc.b	$E5,$FE,$D3,$6E,$7F,$ED,$96,$CF,$EA,$2F,$A3,$DC,$E4,$5A,$E4,$FE
	.dc.b	$A7,$B9,$F2,$5E,$F5,$F2,$56,$30,$D9,$E0,$DA,$FD,$66,$CB,$E3,$5D
	.dc.b	$6F,$8B,$4E,$27,$87,$C5,$F0,$7C,$97,$85,$B1,$93,$CD,$19,$69,$E9
	.dc.b	$F9,$1F,$47,$C8,$FA,$5B,$19,$3D,$D1,$96,$99,$CB,$D3,$8A,$75,$18
	.dc.b	$6E,$DF,$E2,$29,$D5,$71,$8C,$4F,$C9,$F9,$1F,$95,$F2,$16,$BE,$73
	.dc.b	$70,$EE,$DB,$E8,$FE,$3F,$CA,$7A,$74,$C7,$A9,$70,$FF,$65,$45,$C3
	.dc.b	$FD,$C3,$37,$0F,$F7,$5F,$C1,$FE,$67,$E4,$BF,$CA,$B8,$7F,$B4,$63
	.dc.b	$0F,$76,$B9,$3F,$96,$47,$0B,$FF,$51,$FC,$CF,$92,$7F,$CD,$4D,$8F
	.dc.b	$F3,$83,$5F,$F3,$AB,$5F,$F2,$88,$78,$5F,$ED,$3E,$75,$CD,$FB,$1A
	.dc.b	$D6,$7F,$B3,$A5,$6B,$0F,$E9,$54,$F9,$77,$1F,$F9,$16,$3B,$C6,$27
	.dc.b	$FD,$00,$D8,$FF,$43,$35,$FF,$45,$B9,$BF,$F3,$57,$6B,$E6,$5F,$19
	.dc.b	$E5,$AD,$61,$FC,$72,$72,$FF,$5D,$F9,$77,$C7,$3E,$53,$F8,$64,$63
	.dc.b	$E5,$2C,$E6,$B8,$28,$CD,$69,$B3,$A6,$E9,$FD,$6E,$FA,$4F,$2A,$E0
	.dc.b	$FC,$6C,$7E,$45,$7C,$0E,$0F,$8C,$FF,$A1,$23,$2F,$C2,$8C,$BF,$35
	.dc.b	$FE,$49,$2F,$9D,$7B,$DF,$0E,$73,$18,$9F,$EE,$9F,$00,$F9,$3F,$E5
	.dc.b	$5F,$1F,$FD,$4A,$73,$18,$F3,$3F,$97,$56,$E7,$FB,$E5,$6F,$2F,$F4
	.dc.b	$C5,$EE,$7E,$3E,$3F,$2B,$E5,$7E,$5B,$63,$27,$CC,$8C,$B4,$CD,$7C
	.dc.b	$9E,$04,$EA,$30,$FE,$A7,$DA,$FD,$57,$B3,$6F,$F4,$AD,$AF,$E4,$3D
	.dc.b	$68,$C7,$3F,$FA,$A2,$31,$5F,$98,$9F,$EE,$AF,$FB,$4D,$B5,$FE,$D9
	.dc.b	$6B,$FE,$DE,$6B,$FA,$A7,$75,$FE,$E7,$F4,$3F,$51,$18,$AF,$F1,$4F
	.dc.b	$C2,$DB,$FF,$17,$E0,$F6,$7F,$82,$7B,$5F,$C1,$5B,$7F,$C0,$1B,$5E
	.dc.b	$C6,$E5,$EB,$6C,$FF,$8D,$AF,$A3,$FC,$22,$FF,$B5,$75,$BE,$67,$F3
	.dc.b	$D8,$FF,$8C,$63,$1D,$5A,$9C,$E5,$FB,$07,$F4,$58,$7F,$91,$EB,$97
	.dc.b	$4E,$EF,$83,$67,$F8,$BF,$B5,$FC,$7F,$DB,$FE,$40,$D8,$C9,$FC,$72
	.dc.b	$32,$D3,$F6,$3C,$BF,$B6,$CB,$F2,$07,$FB,$AF,$B3,$F7,$71,$9E,$9F
	.dc.b	$DC,$B6,$B8,$9F,$C0,$6B,$8B,$FE,$F0,$AE,$2E,$6B,$8B,$4F,$CF,$FC
	.dc.b	$47,$E9,$8F,$E7,$1E,$6D,$8F,$D1,$FC,$4F,$EA,$BD,$BF,$D5,$B6,$3F
	.dc.b	$9B,$FA,$3F,$CF,$CC,$71,$7E,$51,$FE,$85,$E8,$49,$FE,$7B,$E2,$DA
	.dc.b	$FC,$CC,$FF,$47,$F4,$31,$B7,$E6,$5C,$5F,$F4,$47,$2F,$E0,$5F,$DE
	.dc.b	$FB,$5F,$BE,$F9,$3B,$3F,$BA,$6E,$1F,$E0,$6F,$09,$18,$C1,$5F,$E8
	.dc.b	$09,$CB,$FC,$0A,$FA,$34,$E8,$7F,$56,$F4,$7F,$AF,$7A,$3F,$D7,$DB
	.dc.b	$19,$3F,$AD,$C6,$5A,$67,$1F,$DC,$C6,$5B,$1F,$C4,$BE,$4F,$F8,$CF
	.dc.b	$CC,$FF,$1A,$6C,$64,$FE,$2F,$19,$69,$99,$F8,$11,$96,$C7,$F6,$EF
	.dc.b	$4F,$FB,$B7,$57,$F7,$76,$C6,$4F,$EE,$71,$96,$99,$9F,$E2,$31,$96
	.dc.b	$C7,$F2,$AF,$99,$FE,$63,$F3,$3F,$CC,$9B,$19,$3F,$97,$C6,$5A,$66
	.dc.b	$7E,$C4,$65,$B1,$FE,$06,$ED,$FF,$07,$79,$3F,$E0,$F6,$C6,$4F,$F0
	.dc.b	$5C,$65,$A7,$F9,$47,$0F,$F2,$81,$FE,$51,$9B,$19,$25,$7F,$8D,$A3
	.dc.b	$2F,$F8,$E2,$BF,$CB,$1E,$47,$F9,$65,$B5,$FE,$5D,$6C,$EE,$3F,$CC
	.dc.b	$13,$FE,$72,$F6,$BF,$23,$2F,$7F,$FB,$FB,$FE,$D6,$BF,$F3,$F3,$9F
	.dc.b	$47,$FC,$DD,$EE,$70,$AD,$FB,$5F,$E1,$AC,$31,$5F,$E7,$36,$DF,$36
	.dc.b	$CF,$F8,$82,$EE,$1D,$B0,$E6,$DC,$39,$65,$FE,$AC,$7C,$2D,$C7,$A3
	.dc.b	$70,$FF,$39,$78,$5D,$BE,$67,$FB,$56,$3E,$6F,$6F,$FB,$76,$1E,$B3
	.dc.b	$FD,$E7,$3A,$C7,$FD,$E3,$3A,$AF,$6C,$E6,$BF,$DF,$33,$9A,$FF,$7B
	.dc.b	$CE,$65,$F3,$2F,$85,$79,$3F,$E4,$29,$FF,$21,$CE,$63,$10,$38,$CF
	.dc.b	$07,$33,$B7,$7F,$07,$FC,$2B,$D3,$E3,$1F,$EE,$CE,$FF,$F1,$8B,$A6
	.dc.b	$BE,$19,$E0,$F4,$F3,$FF,$DC,$AD,$F2,$79,$E1,$FD,$9D,$B3,$A9,$F5
	.dc.b	$4E,$5F,$94,$7C,$08,$E7,$B7,$E2,$A3,$15,$F5,$FF,$0B,$7E,$60,$F9
	.dc.b	$AF,$F3,$2D,$EF,$1E,$8D,$C7,$9F,$07,$C8,$6E,$4C,$79,$E1,$CA,$B5
	.dc.b	$97,$E1,$6B,$C2,$79,$B1,$EB,$3F,$3E,$F8,$CE,$E7,$2F,$D3,$6B,$CB
	.dc.b	$1F,$D2,$CF,$08,$E9,$8F,$91,$D5,$F8,$9F,$C7,$4E,$EB,$F1,$7E,$AF
	.dc.b	$FA,$13,$C5,$FD,$8F,$2B,$8D,$39,$E6,$FC,$6B,$71,$FC,$BF,$C6,$DB
	.dc.b	$E1,$36,$F8,$D7,$E2,$F0,$EB,$18,$8F,$F6,$BA,$BE,$19,$F1,$F9,$C9
	.dc.b	$5C,$72,$FC,$EC,$EA,$30,$7E,$A2,$76,$F4,$8F,$32,$FC,$21,$E1,$7E
	.dc.b	$5A,$30,$D8,$CC,$7E,$69,$AF,$E9,$1B,$97,$F1,$68,$E2,$7F,$A8,$A3
	.dc.b	$38,$7E,$CB,$CD,$F8,$DF,$3C,$FC,$21,$F2,$FF,$26,$9F,$1D,$FD,$CB
	.dc.b	$1F,$5C,$63,$AB,$8D,$F0,$9F,$7E,$1C,$F0,$FD,$7B,$63,$F7,$2D,$7F
	.dc.b	$7C,$D7,$FD,$82,$E7,$3F,$E5,$ED,$C3,$F8,$03,$67,$F8,$9F,$A9,$E2
	.dc.b	$F0,$FE,$E7,$A7,$27,$C7,$92,$75,$1E,$28,$C3,$63,$CF,$18,$8F,$2B
	.dc.b	$71,$F1,$B7,$1F,$1B,$67,$F9,$2F,$9D,$F1,$3E,$3F,$F1,$DA,$C9,$F1
	.dc.b	$C6,$5F,$13,$E0,$23,$83,$64,$3F,$27,$E6,$FD,$8F,$DA,$DE,$FF,$E6
	.dc.b	$FC,$F9,$5F,$4A,$72,$D9,$FA,$71,$88,$FA,$2D,$7E,$8B,$5F,$9E,$EF
	.dc.b	$4F,$A6,$DD,$33,$5F,$AB,$F3,$FC,$9E,$D7,$EB,$7D,$09,$8F,$8E,$32
	.dc.b	$FE,$0B,$04,$79,$09,$19,$76,$B1,$F0,$C7,$EE,$7E,$AF,$F7,$5E,$A4
	.dc.b	$AF,$D7,$F8,$1F,$C1,$63,$E3,$7A,$F2,$7C,$0F,$C2,$23,$FD,$32,$7F
	.dc.b	$86,$7D,$4E,$E3,$C1,$B7,$C2,$AF,$E1,$13,$FA,$8F,$D7,$9F,$55,$49
	.dc.b	$7E,$2F,$0F,$2F,$F0,$21,$FA,$6B,$83,$F7,$CD,$8F,$EA,$AD,$7F,$AF
	.dc.b	$B5,$FE,$C0,$7F,$72,$F2,$FF,$8D,$11,$F0,$7D,$27,$F0,$C9,$65,$FC
	.dc.b	$7F,$E9,$25,$78,$E7,$2D,$9F,$19,$1F,$1B,$5F,$86,$7F,$71,$8E,$2E
	.dc.b	$0F,$6C,$7F,$2F,$FC,$9B,$F8,$44,$70,$DD,$FB,$BF,$F7,$DE,$4F,$E7
	.dc.b	$7F,$0F,$FC,$B6,$BE,$CB,$5E,$13,$BE,$2F,$F0,$05,$F2,$9F,$E7,$95
	.dc.b	$BB,$E7,$E5,$7F,$3F,$C3,$85,$2D,$6F,$2B,$5E,$AC,$AF,$0C,$74,$30
	.dc.b	$F8,$A3,$0F,$E3,$AD,$F2,$BE,$2F,$16,$4E,$63,$C4,$61,$F8,$58,$27
	.dc.b	$31,$B3,$E0,$BB,$CF,$FA,$FD,$F6,$9F,$E5,$BF,$86,$FF,$81,$E7,$3F
	.dc.b	$2D,$EF,$3F,$9D,$C6,$76,$FE,$4E,$DA,$79,$08,$C3,$7C,$EF,$91,$F4
	.dc.b	$DF,$27,$E8,$7E,$6F,$8B,$30,$F8,$11,$97,$EB,$C6,$5F,$AF,$1F,$D7
	.dc.b	$3F,$0B,$67,$FD,$CF,$85,$F9,$43,$CA,$C3,$70,$FD,$35,$F0,$6D,$FF
	.dc.b	$63,$8C,$47,$F9,$09,$AF,$F9,$1D,$AF,$E9,$9D,$EF,$FD,$8D,$BB,$E6
	.dc.b	$BC,$DA,$7E,$2B,$C6,$FC,$5F,$74,$C7,$F2,$71,$97,$F2,$B1,$97,$ED
	.dc.b	$C7,$D0,$DB,$F2,$7E,$24,$9F,$AF,$19,$69,$0F,$C1,$F8,$9F,$1F,$E7
	.dc.b	$3E,$47,$0C,$9F,$87,$19,$7F,$CB,$3A,$A3,$E2,$12,$32,$ED,$FD,$A4
	.dc.b	$6D,$F9,$BF,$4B,$F8,$6C,$FF,$53,$DD,$27,$F1,$B1,$96,$90,$FC,$CF
	.dc.b	$89,$F8,$E9,$FD,$8F,$87,$27,$F2,$B1,$96,$90,$FD,$0F,$87,$FE,$20
	.dc.b	$FA,$27,$C5,$7F,$87,$DB,$3F,$E7,$66,$BF,$E8,$06,$BF,$E8,$57,$3C
	.dc.b	$BF,$E1,$F6,$F2,$FF,$26,$6C,$FF,$A1,$FE,$8F,$FC,$E9,$5F,$E2,$56
	.dc.b	$D7,$F2,$C6,$CE,$FD,$DF,$F2,$55,$7F,$08,$F3,$FF,$C6,$F5,$AA,$34
	.dc.b	$E3,$3B,$8F,$E8,$1F,$AB,$BE,$17,$FC,$78,$D8,$FF,$49,$B5,$FF,$4E
	.dc.b	$B5,$FF,$51,$39,$9F,$F1,$E3,$71,$FE,$86,$D9,$FF,$52,$7E,$AF,$FC
	.dc.b	$C2,$BF,$C5,$3E,$0B,$CB,$FF,$19,$B6,$BF,$C6,$6D,$7F,$C6,$6D,$7F
	.dc.b	$C6,$6E,$67,$9C,$FF,$0E,$3F,$CD,$9F,$BC,$3E,$2F,$17,$FC,$5A,$6B
	.dc.b	$FD,$06,$74,$E4,$87,$F4,$CF,$13,$F8,$A4,$F4,$AF,$F2,$AF,$B5,$FC
	.dc.b	$86,$BF,$C8,$ED,$7F,$C8,$ED,$7F,$C8,$AD,$7F,$8F,$BC,$70,$FF,$25
	.dc.b	$B7,$4D,$F9,$3F,$D1,$FA,$FF,$D1,$1C,$4F,$C9,$BF,$CD,$58,$7F,$32
	.dc.b	$EB,$FE,$98,$6A,$33,$87,$FB,$4E,$B8,$3F,$EC,$E6,$E5,$FE,$D8,$6B
	.dc.b	$FE,$DD,$6B,$FD,$F1,$CD,$A7,$A5,$76,$A5,$6A,$7F,$B1,$FE,$09,$F8
	.dc.b	$18,$F4,$7C,$1F,$F6,$1F,$AC,$97,$FE,$92,$9C,$B7,$05,$79,$DB,$1E
	.dc.b	$76,$BE,$76,$BE,$77,$38,$F9,$E3,$CB,$18,$C7,$D3,$5F,$DC,$FE,$AF
	.dc.b	$E4,$32,$BC,$4E,$5F,$27,$57,$F7,$9F,$7E,$46,$08,$78,$E3,$D0,$47
	.dc.b	$D0,$FD,$52,$3F,$DD,$FE,$77,$FD,$5F,$E0,$FF,$0A,$8F,$F1,$5B,$51
	.dc.b	$FE,$35,$9F,$F5,$94,$66,$3F,$C7,$1E,$CE,$08,$FF,$8D,$0E,$6F,$FA
	.dc.b	$E7,$E2,$7F,$AF,$6B,$F8,$5F,$85,$FF,$1C,$C6,$2B,$ED,$C6,$3C,$EF
	.dc.b	$F2,$F4,$60,$F1,$CE,$F8,$3F,$DA,$91,$E6,$F8,$6E,$8F,$85,$AF,$F9
	.dc.b	$19,$AE,$CE,$87,$F9,$47,$E6,$FF,$D8,$CF,$C2,$6B,$F0,$B3,$46,$23
	.dc.b	$87,$0F,$BD,$D7,$CB,$66,$8D,$24,$FE,$2F,$E2,$3F,$0D,$AF,$E1,$F2
	.dc.b	$FD,$74,$62,$7F,$C7,$BE,$AF,$FB,$41,$FC,$B4,$67,$1F,$C6,$BE,$6A
	.dc.b	$E3,$3D,$49,$3F,$97,$9E,$1E,$CF,$F9,$43,$EA,$39,$5F,$0A,$FF,$26
	.dc.b	$C6,$3F,$20,$FD,$6E,$1F,$A7,$6C,$FF,$94,$BE,$1F,$57,$2B,$4F,$6C
	.dc.b	$BF,$D9,$D1,$82,$57,$EF,$EF,$9B,$FB,$F6,$BF,$C2,$5A,$FE,$F9,$F0
	.dc.b	$E7,$FE,$42,$F9,$2F,$04,$FF,$99,$1F,$47,$BB,$2F,$FD,$43,$F5,$6F
	.dc.b	$95,$C5,$FF,$9E,$33,$65,$FE,$61,$BF,$E2,$8D,$9F,$36,$1F,$36,$31
	.dc.b	$1E,$78,$F7,$61,$D3,$2F,$F2,$AB,$F5,$7C,$67,$88,$F6,$70,$DC,$BF
	.dc.b	$CD,$1F,$57,$89,$91,$F8,$BA,$E3,$1F,$92,$6C,$66,$B9,$65,$FD,$0B
	.dc.b	$C2,$EA,$7D,$EF,$1D,$E2,$E2,$FF,$77,$A7,$F4,$0A,$C5,$FF,$A4,$EF
	.dc.b	$75,$A3,$FA,$4C,$EA,$B8,$47,$FB,$32,$77,$5F,$D5,$F0,$5F,$48,$FE
	.dc.b	$A6,$FA,$63,$35,$E9,$8F,$D5,$7B,$AF,$07,$F3,$96,$8C,$57,$FB,$1F
	.dc.b	$F1,$1F,$EF,$0F,$FB,$26,$33,$3F,$EA,$F6,$BF,$C4,$5A,$FF,$B1,$9A
	.dc.b	$FF,$1C,$6E,$8F,$B2,$7C,$1F,$D9,$6D,$FC,$3B,$7E,$92,$FB,$97,$CE
	.dc.b	$6F,$09,$7B,$C7,$A4,$7E,$DE,$F8,$56,$B0,$FF,$63,$7D,$53,$EA,$BE
	.dc.b	$0B,$71,$E5,$9F,$8A,$B7,$5A,$3F,$89,$56,$E3,$F8,$93,$5F,$E6,$4D
	.dc.b	$7F,$C0,$7E,$57,$E4,$8F,$E0,$7F,$11,$F3,$23,$ED,$46,$70,$F8,$1E
	.dc.b	$0B,$C2,$FE,$87,$1E,$DE,$45,$7F,$0C,$9C,$DF,$E0,$FC,$2F,$CF,$99
	.dc.b	$E2,$B8,$71,$8E,$1E,$8E,$2B,$FD,$C9,$7F,$EC,$AA,$FA,$10,$FC,$07
	.dc.b	$E1,$74,$FF,$6D,$8C,$71,$FE,$22,$5E,$5D,$89,$7F,$CD,$6B,$6D,$AF
	.dc.b	$99,$1F,$D5,$DA,$FF,$60,$36,$7F,$BE,$D7,$C4,$74,$8C,$3F,$BC,$EF
	.dc.b	$95,$F2,$32,$FF,$5F,$BF,$1E,$33,$5E,$EE,$EB,$9F,$F8,$1F,$87,$E7
	.dc.b	$47,$CB,$F8,$2F,$99,$EA,$5C,$7A,$37,$0F,$8E,$DC,$39,$65,$FE,$A3
	.dc.b	$AF,$93,$A3,$E5,$7E,$31,$F1,$FC,$3B,$3A,$8F,$EB,$93,$B8,$F8,$D1
	.dc.b	$82,$3F,$EF,$FF,$77,$8E,$4C,$7F,$DE,$B1,$FD,$07,$A7,$F0,$9E,$1D
	.dc.b	$BF,$88,$DB,$FB,$ED,$BF,$4B,$FC,$A7,$F6,$A7,$C9,$53,$FA,$67,$D7
	.dc.b	$5C,$3F,$5D,$3A,$7F,$2F,$E6,$3E,$BB,$59,$79,$67,$FB,$5C,$7F,$62
	.dc.b	$CF,$FC,$99,$1F,$5F,$C6,$C3,$F9,$B6,$B9,$3F,$47,$1C,$2B,$F2,$F3
	.dc.b	$A6,$CE,$8E,$47,$E0,$FA,$FC,$19,$FE,$9D,$B1,$93,$AC,$65,$A7,$83
	.dc.b	$D1,$FD,$77,$9D,$E9,$6C,$64,$FC,$84,$65,$A7,$E0,$BD,$1F,$D3,$1F
	.dc.b	$E0,$2F,$8F,$7C,$66,$27,$47,$E9,$67,$F0,$64,$70,$7C,$2F,$9E,$E1
	.dc.b	$9F,$CD,$1D,$9F,$C6,$E3,$E7,$7B,$52,$BE,$4F,$D4,$5B,$FD,$67,$C4
	.dc.b	$7E,$40,$93,$FB,$57,$FB,$07,$D2,$32,$FE,$D8,$D9,$FA,$7E,$E6,$27
	.dc.b	$EE,$7D,$3F,$E2,$7D,$4B,$5F,$86,$70,$FF,$6D,$E8,$FF,$5B,$36,$3F
	.dc.b	$B7,$46,$32,$FF,$2D,$6D,$FC,$D7,$DA,$FC,$41,$CA,$3E,$04,$66,$FF
	.dc.b	$A0,$7E,$D3,$FE,$16,$7F,$CD,$3E,$9A,$31,$5F,$1B,$A5,$F4,$FF,$E0
	.dc.b	$79,$CD,$6F,$DA,$FF,$00,$47,$F8,$8A,$31,$87,$F8,$16,$30,$74,$C7
	.dc.b	$FA,$57,$C0,$73,$BF,$CF,$3D,$7C,$39,$7F,$9A,$F9,$9B,$A7,$FB,$3F
	.dc.b	$1F,$E0,$0F,$F9,$E7,$2F,$45,$F8,$89,$7E,$17,$F5,$5C,$7C,$BC,$3F
	.dc.b	$E3,$5B,$23,$D9,$F2,$52,$FF,$53,$EB,$DC,$BC,$13,$EE,$24,$FB,$5B
	.dc.b	$1F,$3A,$DF,$4F,$C7,$1E,$EF,$C9,$F9,$49,$87,$ED,$BC,$BB,$87,$B6
	.dc.b	$7E,$59,$27,$E5,$36,$3B,$73,$70,$C3,$D6,$FD,$5E,$A9,$7F,$C0,$BE
	.dc.b	$02,$DF,$C9,$9F,$AE,$49,$FA,$CD,$8F,$D3,$43,$F7,$1F,$F3,$2C,$71
	.dc.b	$8F,$F0,$71,$9F,$82,$FF,$11,$FD,$0F,$E1,$A3,$FB,$1B,$67,$E2,$7C
	.dc.b	$7E,$BA,$A5,$7F,$5F,$AD,$57,$FA,$CD,$AF,$F5,$D6,$BF,$93,$FA,$8F
	.dc.b	$E4,$9F,$51,$FC,$95,$B5,$FC,$81,$B3,$F4,$DB,$B7,$F5,$9F,$DD,$3F
	.dc.b	$24,$4F,$6B,$FD,$1E,$E3,$B6,$98,$8B,$87,$F7,$EF,$A6,$78,$FF,$E4
	.dc.b	$1C,$9F,$13,$FD,$C1,$FF,$69,$CE,$BD,$EF,$F6,$07,$B9,$FE,$D6,$FA
	.dc.b	$67,$E3,$8F,$2F,$FD,$E5,$8F,$1C,$58,$F1,$7A,$E1,$FD,$FE,$31,$3B
	.dc.b	$6C,$75,$BE,$99,$F8,$2F,$C7,$9F,$9F,$0E,$0D,$BE,$8D,$96,$5E,$6C
	.dc.b	$BC,$B5,$AC,$FA,$57,$9A,$F8,$36,$3C,$17,$C7,$1D,$E7,$E6,$AF,$1D
	.dc.b	$62,$FC,$17,$B9,$F7,$9A,$29,$D4,$FF,$6D,$57,$C5,$7F,$DF,$4D,$7C
	.dc.b	$C7,$FB,$7F,$E3,$1F,$4A,$FA,$9F,$F1,$74,$61,$B1,$99,$E0,$D7,$AF
	.dc.b	$9E,$E1,$FF,$1B,$60,$8C,$4F,$91,$AF,$09,$DC,$6B,$7F,$B0,$DD,$BE
	.dc.b	$E3,$6B,$E8,$FD,$4B,$E7,$3F,$C7,$D7,$C3,$0F,$97,$7B,$BF,$C3,$BC
	.dc.b	$75,$61,$C6,$77,$51,$F9,$79,$7A,$63,$EB,$CE,$E7,$37,$D6,$B5,$18
	.dc.b	$AD,$E3,$E4,$CB,$D3,$1D,$72,$F3,$57,$DA,$CB,$C7,$7B,$8F,$EE,$3F
	.dc.b	$49,$FE,$62,$7F,$CF,$2D,$7F,$AB,$B7,$9B,$FC,$A0,$FF,$A8,$23,$7C
	.dc.b	$BF,$E6,$19,$CB,$78,$D3,$DA,$30,$F8,$3E,$16,$F9,$CC,$3F,$61,$87
	.dc.b	$6F,$81,$CC,$1B,$B4,$C3,$CF,$1A,$30,$FF,$83,$47,$67,$FA,$4D,$FF
	.dc.b	$4A,$F5,$7A,$5F,$EE,$6D,$C3,$FD,$0D,$3B,$9E,$4D,$7F,$92,$78,$BF
	.dc.b	$E8,$57,$AB,$63,$51,$FD,$EB,$EA,$3F,$D3,$EF,$F3,$6F,$89,$FD,$0B
	.dc.b	$79,$FE,$1F,$9C,$AD,$CF,$F8,$96,$31,$F1,$3F,$23,$89,$87,$2A,$FE
	.dc.b	$8A,$DC,$D3,$E1,$AE,$B8,$74,$9C,$DB,$0E,$AF,$FB,$46,$75,$7F,$63
	.dc.b	$83,$EA,$7D,$27,$2F,$17,$E9,$FD,$4C,$C3,$84,$65,$FC,$4C,$65,$F5
	.dc.b	$60,$9D,$46,$1B,$97,$E3,$3E,$57,$CF,$F0,$DF,$80,$DE,$5F,$96,$32
	.dc.b	$FE,$5A,$32,$F6,$8C,$B4,$E0,$7E,$73,$D0,$F9,$FF,$51,$FE,$86,$8F
	.dc.b	$CF,$FD,$4F,$E3,$8C,$FC,$14,$8C,$90,$FA,$93,$C5,$B1,$B6,$9F,$AA
	.dc.b	$F7,$FB,$FB,$FF,$9A,$F7,$E4,$FF,$BF,$ED,$3A,$8F,$D8,$CE,$9E,$F1
	.dc.b	$96,$9C,$5F,$D2,$FA,$9F,$A6,$FA,$59,$3F,$BA,$7D,$FE,$E4,$8C,$10
	.dc.b	$F5,$BF,$AE,$F5,$35,$18,$3F,$81,$BF,$96,$FA,$97,$D2,$D2,$4F,$F0
	.dc.b	$87,$F7,$AB,$EA,$7F,$CE,$4E,$48,$7E,$D5,$CB,$FC,$3B,$C6,$FD,$F9
	.dc.b	$1F,$E2,$7F,$4D,$D2,$58,$7F,$04,$FA,$09,$5F,$F0,$02,$72,$D9,$CB
	.dc.b	$FC,$6A,$79,$44,$8F,$F7,$F4,$F0,$7F,$DF,$47,$F1,$87,$0E,$1A,$FE
	.dc.b	$97,$E3,$37,$D1,$FE,$F3,$F9,$5F,$E1,$5D,$FF,$A9,$6D,$7F,$A7,$AB
	.dc.b	$73,$C6,$3F,$90,$CE,$BB,$CD,$AD,$CF,$FA,$9A,$3F,$91,$3F,$ED,$8F
	.dc.b	$53,$F9,$4C,$FF,$88,$63,$FD,$1B,$3A,$6F,$8F,$AA,$91,$FC,$D3,$1F
	.dc.b	$44,$F2,$F9,$5F,$1F,$85,$2B,$FD,$CB,$19,$78,$67,$ED,$9D,$47,$FA
	.dc.b	$73,$25,$6E,$3C,$73,$AC,$3F,$88,$E6,$BD,$D7,$43,$F8,$A7,$D1,$7F
	.dc.b	$42,$C7,$E7,$7D,$47,$B8,$FE,$4F,$19,$9D,$F9,$7F,$61,$FF,$54,$C7
	.dc.b	$F8,$E5,$FE,$A2,$68,$D7,$8C,$8D,$47,$DB,$F9,$5F,$A7,$E0,$C8,$F2
	.dc.b	$67,$E7,$9F,$9E,$66,$3E,$7C,$E1,$F5,$0F,$A8,$65,$FD,$83,$85,$F5
	.dc.b	$2F,$1B,$0D,$BE,$37,$C1,$B7,$AA,$91,$82,$1F,$06,$78,$36,$7F,$AB
	.dc.b	$46,$23,$E1,$E5,$F8,$EF,$0B,$F2,$3E,$0C,$9F,$7F,$37,$FB,$F9,$B3
	.dc.b	$97,$F1,$B3,$F9,$88,$FA,$7A,$7E,$6B,$C0,$95,$F6,$F8,$3E,$74,$EA
	.dc.b	$3E,$8F,$02,$3E,$C4,$7E,$01,$B3,$C2,$33,$9B,$E5,$FF,$86,$AE,$1A
	.dc.b	$62,$24,$FC,$A8,$CB,$F3,$23,$2F,$C7,$7E,$29,$1F,$2E,$7F,$B4,$EA
	.dc.b	$91,$EB,$CB,$F8,$BF,$37,$F6,$AC,$BF,$15,$7E,$89,$FE,$37,$5F,$90
	.dc.b	$9C,$BD,$0F,$83,$7F,$C0,$3B,$64,$75,$8F,$CE,$9C,$C8,$F0,$3E,$3F
	.dc.b	$D4,$BE,$33,$F9,$7F,$D4,$FF,$0E,$93,$67,$51,$F9,$46,$BF,$A7,$6B
	.dc.b	$FC,$81,$AF,$EA,$1A,$FF,$8E,$A3,$FC,$D7,$9F,$F8,$2A,$3B,$FB,$1E
	.dc.b	$23,$FC,$7B,$1F,$C6,$BC,$AC,$4F,$E9,$2B,$8F,$5F,$ED,$1B,$0A,$E0
	.dc.b	$FE,$AB,$EA,$5F,$0B,$FD,$31,$7C,$65,$31,$12,$7F,$17,$3C,$1B,$1F
	.dc.b	$E3,$78,$C4,$7F,$3E,$DF,$F8,$17,$89,$FE,$52,$F5,$BF,$86,$75,$9F
	.dc.b	$E1,$D3,$AA,$FC,$15,$7D,$F2,$3F,$9C,$6B,$FE,$39,$E3,$47,$F1,$83
	.dc.b	$FC,$BF,$F5,$0F,$8A,$C5,$74,$21,$F0,$F7,$FF,$19,$4E,$E3,$F5,$9E
	.dc.b	$B7,$F2,$87,$FC,$B3,$E3,$FA,$4F,$F2,$5C,$67,$83,$F2,$CE,$E7,$F2
	.dc.b	$8D,$7F,$CE,$B9,$FF,$93,$3C,$7F,$D8,$CE,$67,$FD,$5D,$F4,$FF,$7E
	.dc.b	$FF,$9C,$F8,$EF,$9B,$FE,$A2,$7F,$9F,$39,$D5,$BD,$7F,$D0,$31,$87
	.dc.b	$78,$7F,$2C,$BE,$4F,$F4,$19,$D3,$6B,$F1,$CF,$F3,$A8,$CB,$FD,$D9
	.dc.b	$AE,$63,$6F,$81,$FF,$62,$3D,$FE,$93,$FB,$39,$E7,$8C,$C7,$CF,$3D
	.dc.b	$FF,$05,$A7,$81,$E3,$78,$11,$C1,$EB,$FA,$7C,$3C,$EF,$E1,$CE,$33
	.dc.b	$F9,$35,$E1,$EC,$FE,$F3,$ED,$CC,$3E,$04,$65,$F8,$D1,$97,$A6,$3F
	.dc.b	$BA,$BF,$F3,$DE,$3F,$42,$F8,$56,$9F,$F0,$2F,$CD,$FF,$8D,$3C,$1E
	.dc.b	$1B,$7E,$33,$D3,$FC,$99,$9A,$D7,$AB,$FE,$B4,$FA,$87,$C6,$53,$39
	.dc.b	$7F,$5E,$F8,$9F,$EB,$5A,$FD,$EF,$84,$FE,$DF,$18,$9F,$F7,$B4,$63
	.dc.b	$D0,$FD,$F4,$60,$FF,$62,$CE,$F7,$FF,$68,$47,$FB,$23,$E1,$BA,$3F
	.dc.b	$E3,$46,$BF,$EF,$23,$91,$63,$FD,$B1,$BB,$E6,$FF,$D3,$F9,$7F,$84
	.dc.b	$3E,$87,$F6,$1C,$3F,$DD,$AB,$4F,$F7,$E8,$CE,$1F,$EC,$99,$7C,$C7
	.dc.b	$2A,$FD,$9F,$8F,$6B,$FD,$BB,$5A,$9F,$F0,$3F,$81,$FC,$59,$FC,$D7
	.dc.b	$C3,$6F,$4F,$D2,$BC,$7C,$6F,$F1,$17,$7E,$7A,$53,$4D,$77,$3E,$77
	.dc.b	$18,$FE,$A5,$BA,$7F,$8C,$5B,$C1,$FB,$67,$C1,$F5,$0D,$29,$D3,$E1
	.dc.b	$6E,$1F,$B1,$9D,$36,$74,$D2,$61,$FE,$3F,$F9,$EE,$2D,$8C,$E3,$FC
	.dc.b	$1A,$72,$D8,$ED,$E3,$7E,$DA,$D8,$7F,$BF,$67,$FC,$75,$18,$9D,$4E
	.dc.b	$70,$FE,$00,$DB,$FE,$27,$C9,$EE,$AF,$8E,$DC,$3F,$87,$36,$FF,$CA
	.dc.b	$CF,$D9,$F1,$BF,$DD,$7F,$45,$FC,$21,$B5,$FE,$46,$E1,$4E,$5B,$19
	.dc.b	$8F,$B5,$E5,$BE,$6F,$A5,$6B,$1F,$E1,$9C,$9F,$D2,$E7,$3E,$3C,$C7
	.dc.b	$FC,$9B,$E5,$7F,$2B,$7F,$9E,$C6,$6F,$F8,$A6,$7F,$F0,$02,$3F,$11
	.dc.b	$5B,$9C,$E2,$9E,$11,$FC,$F9,$F4,$12,$FD,$8F,$A3,$DF,$ED,$BF,$F9
	.dc.b	$D3,$44,$7E,$76,$DE,$F3,$C7,$FE,$5B,$18,$6C,$67,$1F,$F1,$DB,$67
	.dc.b	$FC,$E8,$DC,$3F,$C1,$0B,$EA,$3B,$39,$AF,$F3,$BB,$97,$FC,$95,$F4
	.dc.b	$AF,$8F,$FB,$0D,$8F,$F4,$47,$D3,$3E,$9D,$F3,$FA,$B9,$5C,$9F,$DC
	.dc.b	$56,$E7,$39,$FF,$41,$1F,$4E,$FA,$0F,$7B,$70,$FF,$2B,$B6,$7F,$D2
	.dc.b	$0D,$7F,$D9,$2E,$75,$FE,$23,$F6,$5F,$E1,$AA,$FF,$42,$CE,$5F,$F5
	.dc.b	$94,$66,$BE,$DF,$A3,$F8,$F2,$4F,$E4,$DB,$1F,$8C,$F0,$8F,$E4,$23
	.dc.b	$F8,$64,$62,$7F,$C0,$FE,$0F,$E6,$48,$FE,$75,$B5,$CF,$D5,$FC,$CC
	.dc.b	$72,$7F,$05,$F5,$1F,$3E,$7F,$D5,$CD,$7F,$DC,$ED,$7F,$D5,$ED,$7C
	.dc.b	$AF,$1C,$FF,$C9,$0B,$E2,$7A,$F0,$76,$6E,$3F,$EA,$A9,$D3,$67,$4F
	.dc.b	$F8,$8B,$BF,$F6,$84,$7F,$70,$DC,$3E,$37,$7F,$ED,$23,$E2,$BF,$1B
	.dc.b	$E0,$BE,$3F,$C1,$5B,$FF,$5F,$ED,$FE,$8F,$6C,$E9,$A6,$A7,$E2,$F8
	.dc.b	$1F,$A4,$CF,$F8,$14,$FF,$A7,$1B,$97,$C8,$9D,$1F,$CF,$3D,$FE,$35
	.dc.b	$FE,$23,$6C,$FF,$BA,$DB,$B7,$F8,$C1,$FE,$87,$C9,$F7,$BD,$AB,$7E
	.dc.b	$86,$DF,$E1,$5B,$7F,$EF,$0D,$3F,$84,$3F,$E9,$9F,$A4,$7C,$8F,$F2
	.dc.b	$16,$E1,$FE,$E6,$6E,$1F,$E6,$76,$BF,$E9,$D7,$35,$FE,$44,$FA,$7F
	.dc.b	$E5,$91,$8A,$FC,$AF,$8D,$FE,$FF,$9F,$EC,$19,$A7,$F2,$FE,$06,$1A
	.dc.b	$7F,$BC,$23,$4F,$D4,$8D,$34,$D1,$C3,$1F,$E5,$F3,$C1,$FD,$34,$FE
	.dc.b	$93,$9F,$C3,$9D,$BE,$CD,$AE,$2D,$CF,$CD,$0C,$BF,$82,$4F,$AB,$D2
	.dc.b	$FD,$7F,$97,$70,$FC,$4B,$6F,$F6,$CD,$C3,$E9,$3F,$AC,$F4,$3F,$10
	.dc.b	$4A,$FC,$43,$6B,$D7,$F3,$BB,$9F,$88,$D8,$CC,$7E,$9E,$7E,$6C,$FC
	.dc.b	$EF,$7A,$4F,$CB,$F9,$EB,$7F,$C7,$F9,$9F,$0D,$F5,$3D,$DF,$DF,$B6
	.dc.b	$FF,$DF,$5E,$B3,$3F,$ED,$0D,$8F,$AD,$58,$9F,$BB,$F4,$DF,$82,$F3
	.dc.b	$ED,$7F,$16,$DF,$F2,$11,$98,$FB,$DE,$C5,$D3,$EC,$BF,$86,$24,$71
	.dc.b	$3F,$D2,$FD,$1F,$E3,$6B,$FA,$BC,$9F,$CD,$EB,$FB,$E7,$D3,$7F,$86
	.dc.b	$B0,$FE,$D2,$DA,$DF,$43,$E8,$F7,$EB,$4C,$FF,$21,$F0,$37,$1F,$C5
	.dc.b	$B7,$0E,$2D,$9F,$CE,$E1,$FE,$F3,$7C,$D4,$F5,$59,$7F,$02,$C3,$F1
	.dc.b	$71,$8F,$07,$F4,$7F,$05,$E8,$66,$1F,$A1,$E4,$B8,$FE,$71,$B8,$7A
	.dc.b	$9B,$3E,$27,$E0,$FC,$64,$8F,$D5,$79,$D6,$FE,$1C,$FE,$CB,$E9,$B1
	.dc.b	$87,$C4,$6D,$FB,$1B,$5F,$A1,$D5,$1F,$3B,$E3,$A5,$7E,$DF,$C4,$B7
	.dc.b	$F4,$27,$F7,$FC,$D8,$C7,$E8,$B6,$FE,$4B,$6B,$B3,$F6,$BA,$A4,$FF
	.dc.b	$03,$FA,$6B,$5F,$72,$7F,$CE,$52,$FA,$67,$BD,$AC,$7F,$73,$39,$8C
	.dc.b	$3F,$D9,$67,$6F,$D8,$E3,$FA,$F7,$D2,$73,$87,$F8,$17,$E7,$FF,$11
	.dc.b	$7F,$DC,$22,$08,$CB,$83,$1D,$1F,$E2,$88,$CC,$70,$77,$97,$F8,$C7
	.dc.b	$E7,$FF,$C6,$11,$8B,$FE,$53,$EE,$27,$F9,$4F,$C4,$7F,$2A,$FA,$6B
	.dc.b	$5F,$C9,$DB,$8F,$F2,$26,$E3,$FA,$07,$FC,$AF,$D9,$2B,$F2,$FD,$CC
	.dc.b	$9F,$13,$9C,$7F,$CB,$31,$FA,$9A,$F1,$DF,$F9,$82,$B5,$8F,$F2,$6B
	.dc.b	$7D,$1B,$C8,$FF,$24,$C3,$E2,$3F,$C0,$0F,$F8,$62,$75,$EF,$FF,$AF
	.dc.b	$FD,$8F,$F0,$DF,$D1,$BE,$99,$E6,$7F,$32,$C7,$8E,$3D,$B2,$EB,$18
	.dc.b	$CB,$9C,$F7,$C7,$EB,$46,$27,$C3,$8F,$5C,$3A,$56,$B0,$E6,$DC,$BC
	.dc.b	$98,$F9,$E7,$2A,$F9,$6D,$A8,$F2,$55,$CB,$C1,$87,$39,$F2,$1C,$73
	.dc.b	$F6,$DE,$F1,$EB,$3C,$8D,$97,$3F,$4D,$70,$8C,$E1,$D9,$F3,$63,$F9
	.dc.b	$17,$91,$F9,$68,$F8,$AD,$BF,$AA,$FE,$6B,$E2,$DF,$44,$FA,$6E,$51
	.dc.b	$86,$C6,$67,$F3,$D3,$97,$F4,$16,$F4,$3F,$DA,$11,$86,$BE,$36,$BF
	.dc.b	$59,$AF,$E1,$5E,$39,$FD,$D6,$EB,$BE,$2F,$AB,$BB,$E9,$9F,$31,$FB
	.dc.b	$5C,$38,$57,$D3,$C3,$8E,$3D,$63,$8D,$F0,$8D,$44,$7E,$6C,$F0,$C3
	.dc.b	$B6,$1D,$71,$F1,$56,$AF,$86,$3F,$8E,$C7,$A4,$E4,$91,$E4,$7F,$07
	.dc.b	$87,$47,$F1,$FC,$DF,$B3,$F9,$0F,$E1,$DC,$BF,$EB,$56,$F2,$7F,$04
	.dc.b	$6D,$7F,$85,$79,$FE,$CB,$FB,$C5,$1A,$BE,$93,$98,$C4,$F0,$8F,$DF
	.dc.b	$F8,$F8,$FA,$49,$1C,$E3,$66,$1F,$E0,$B1,$87,$F9,$57,$CD,$FF,$01
	.dc.b	$6F,$37,$9B,$C2,$4E,$67,$F8,$6F,$C1,$66,$31,$AF,$F6,$66,$EB,$FE
	.dc.b	$9E,$BE,$38,$7A,$2F,$87,$5F,$AA,$E6,$3E,$28,$CB,$48,$31,$FE,$8B
	.dc.b	$F1,$BF,$05,$F0,$4F,$09,$FF,$78,$D7,$F1,$27,$FB,$EE,$4B,$E0,$F9
	.dc.b	$67,$FC,$19,$3F,$47,$D5,$F9,$A7,$B7,$C3,$91,$F1,$9A,$F8,$EB,$FC
	.dc.b	$CF,$F4,$DF,$CF,$E3,$44,$3C,$73,$F6,$E3,$51,$E6,$F3,$3E,$DF,$97
	.dc.b	$23,$E9,$C7,$CD,$23,$83,$FA,$0C,$FA,$9F,$F0,$C6,$3F,$DF,$9D,$1D
	.dc.b	$4D,$F7,$7D,$7F,$0F,$F1,$3F,$49,$27,$F0,$31,$96,$90,$F7,$1F,$E6
	.dc.b	$BB,$71,$EE,$3E,$7C,$DE,$0F,$F1,$ED,$6A,$7F,$BB,$7D,$23,$D2,$FC
	.dc.b	$CF,$97,$27,$31,$F8,$F9,$D4,$60,$87,$8B,$A9,$7F,$E5,$D8,$CE,$1F
	.dc.b	$E3,$6A,$E3,$1E,$8E,$74,$62,$7C,$BC,$7F,$AA,$FA,$49,$7F,$B1,$7C
	.dc.b	$5A,$C8,$F4,$B5,$F1,$62,$DB,$EE,$19,$C3,$C5,$58,$AF,$DE,$78,$3F
	.dc.b	$B8,$23,$FB,$FF,$A4,$F8,$18,$C8,$F8,$EF,$E1,$CF,$E0,$EB,$E9,$3E
	.dc.b	$03,$F5,$0C,$67,$A8,$CE,$3F,$4B,$0F,$99,$87,$F0,$8F,$9B,$98,$7D
	.dc.b	$48,$CB,$48,$7B,$A3,$2F,$CB,$D5,$38,$BF,$BF,$5F,$C5,$FE,$89,$D9
	.dc.b	$FC,$67,$F7,$8F,$F1,$7D,$7F,$A0,$1B,$D7,$FE,$CE,$AD,$E5,$FE,$1C
	.dc.b	$6B,$F7,$E7,$6F,$FA,$ED,$BC,$69,$DD,$FF,$A3,$1F,$BD,$7D,$79,$3C
	.dc.b	$87,$F2,$BF,$99,$FE,$D3,$5B,$8C,$E3,$E1,$75,$E6,$7A,$49,$1F,$E3
	.dc.b	$0C,$B8,$E1,$E9,$F9,$6C,$CF,$9E,$AE,$1E,$39,$B5,$FE,$EC,$BE,$37
	.dc.b	$EB,$BF,$91,$5E,$5F,$A4,$F9,$5E,$2C,$AF,$F1,$FE,$69,$D4,$7C,$79
	.dc.b	$D3,$FA,$59,$D3,$E7,$F9,$0F,$A4,$47,$E7,$78,$B5,$F9,$D1,$74,$F9
	.dc.b	$53,$F8,$BE,$2F,$B5,$8C,$BF,$F5,$39,$F9,$38,$D3,$8E,$5F,$C2,$B6
	.dc.b	$3F,$AF,$E8,$FA,$5F,$EB,$B3,$92,$1F,$9B,$AE,$2F,$F1,$09,$E1,$A7
	.dc.b	$F6,$3F,$69,$F5,$34,$95,$F8,$9E,$9A,$FF,$07,$9D,$BF,$E6,$86,$BF
	.dc.b	$C2,$F9,$3F,$4F,$C7,$F9,$3F,$13,$B7,$C6,$4A,$FC,$A4,$7E,$74,$C3
	.dc.b	$D6,$30,$FC,$4F,$D3,$9F,$1D,$FD,$E7,$E8,$3F,$CD,$DC,$9F,$EE,$06
	.dc.b	$E3,$FE,$93,$AD,$DF,$1B,$FD,$FC,$62,$3F,$7C,$DB,$FF,$3C,$7D,$3B
	.dc.b	$E4,$F7,$72,$F9,$1F,$5A,$74,$FE,$B7,$C6,$F2,$7C,$7C,$8C,$10,$FE
	.dc.b	$5D,$9A,$35,$1F,$1E,$7F,$7D,$E1,$7B,$7E,$3E,$4F,$EE,$63,$2F,$FA
	.dc.b	$5A,$D3,$A8,$FA,$13,$FE,$22,$CD,$F8,$DF,$F8,$C1,$6C,$AF,$F0,$04
	.dc.b	$E6,$31,$3C,$2B,$F9,$2F,$D3,$FF,$02,$9C,$D6,$EF,$8D,$77,$8C,$42
	.dc.b	$B8,$3D,$FE,$A1,$F1,$9F,$E7,$38,$46,$1F,$F3,$B3,$5F,$F6,$43,$5F
	.dc.b	$F6,$6B,$9C,$7F,$D5,$CD,$C7,$FD,$A4,$D9,$F7,$5F,$F9,$BA,$BF,$D6
	.dc.b	$9F,$4E,$F9,$1F,$E8,$2D,$AF,$D8,$B5,$FE,$84,$BE,$33,$73,$27,$6F
	.dc.b	$28,$FF,$1E,$7E,$8E,$E6,$A4,$AF,$F2,$97,$BD,$5F,$ED,$53,$B7,$FD
	.dc.b	$E0,$D7,$FD,$9F,$1B,$C7,$EC,$C3,$D1,$56,$E7,$31,$8B,$FF,$2D,$DF
	.dc.b	$1D,$7F,$C9,$17,$B6,$CE,$B0,$FF,$80,$1F,$9B,$FF,$93,$9E,$CD,$C0
	.dc.b	$E7,$CB,$FF,$73,$7D,$27,$F9,$6E,$34,$4F,$6F,$FC,$D7,$1F,$E7,$4F
	.dc.b	$80,$7D,$FF,$F9,$83,$1F,$E8,$1F,$39,$FD,$6C,$D4,$66,$3F,$C4,$B3
	.dc.b	$C3,$DA,$C9,$23,$F9,$CF,$75,$77,$5C,$78,$7E,$3D,$F1,$9D,$9F,$9E
	.dc.b	$DF,$F4,$51,$E4,$EC,$FE,$D9,$DF,$2B,$42,$74,$E1,$AF,$82,$7D,$27
	.dc.b	$F6,$3F,$A5,$B1,$BC,$53,$88,$FE,$99,$37,$1F,$3D,$6A,$78,$57,$37
	.dc.b	$C1,$F0,$0F,$5F,$FC,$05,$F3,$B2,$73,$1F,$3E,$75,$18,$21,$FD,$6F
	.dc.b	$E7,$F5,$32,$33,$8F,$E3,$B9,$1E,$9A,$FA,$7A,$12,$FC,$6F,$FA,$AF
	.dc.b	$EA,$1F,$2D,$E1,$F6,$A5,$7A,$1F,$E7,$1F,$1D,$FE,$30,$F8,$BF,$F5
	.dc.b	$9E,$9F,$4B,$E3,$BF,$8A,$5A,$72,$FD,$18,$CC,$63,$91,$18,$F4,$FF
	.dc.b	$D3,$55,$B3,$98,$F8,$9E,$CE,$1A,$FF,$B9,$4E,$05,$AE,$AF,$E0,$7D
	.dc.b	$CF,$E5,$1D,$3F,$81,$6C,$7F,$9A,$7E,$81,$F2,$BF,$B6,$D3,$F2,$35
	.dc.b	$A6,$FB,$19,$C3,$F2,$32,$8C,$60,$E5,$71,$7F,$11,$BF,$05,$6B,$36
	.dc.b	$1F,$E1,$99,$F2,$7B,$7C,$B0,$FB,$3F,$3D,$BA,$D6,$5F,$AC,$6E,$7E
	.dc.b	$6B,$E1,$CE,$BE,$11,$31,$FF,$1B,$77,$DC,$7F,$C9,$4F,$8F,$EA,$FF
	.dc.b	$90,$4F,$E9,$5B,$3F,$E2,$9F,$9F,$D7,$91,$2B,$4F,$82,$31,$3F,$C0
	.dc.b	$2B,$8E,$7F,$BA,$6D,$FF,$59,$D7,$F2,$B7,$C2,$B5,$87,$C1,$D5,$E0
	.dc.b	$E7,$0F,$F2,$33,$F2,$7E,$C5,$E3,$D3,$E0,$F2,$DA,$FD,$F4,$E9,$B3
	.dc.b	$A6,$93,$5F,$F3,$67,$D5,$F1,$6C,$7F,$2C,$F8,$FF,$BB,$E5,$FF,$2A
	.dc.b	$7E,$B7,$87,$FE,$B9,$9D,$12,$BF,$D5,$DE,$5A,$B7,$F0,$BC,$A3,$15
	.dc.b	$FC,$E3,$D4,$C1,$D3,$E1,$6C,$64,$98,$7F,$AF,$F1,$4E,$A3,$BD,$5E
	.dc.b	$2F,$E0,$55,$AB,$E3,$1F,$D3,$B0,$FE,$F7,$18,$3E,$A4,$EE,$F9,$C7
	.dc.b	$F9,$73,$93,$FC,$40,$DF,$6B,$F3,$F3,$FA,$1F,$A9,$B5,$A8,$FF,$59
	.dc.b	$37,$1F,$EC,$31,$FA,$5A,$E3,$DA,$BE,$31,$D3,$E9,$9F,$3B,$4F,$A3
	.dc.b	$F5,$3F,$EA,$AC,$FF,$CD,$F3,$A6,$CF,$6C,$7F,$89,$CE,$5F,$E1,$61
	.dc.b	$C3,$FC,$39,$AF,$F8,$B7,$8B,$F8,$D7,$9D,$F8,$0F,$1F,$FA,$0E,$1F
	.dc.b	$D3,$35,$7D,$3B,$BB,$57,$2B,$FA,$8E,$1C,$9E,$15,$B7,$38,$74,$C7
	.dc.b	$FC,$E8,$7F,$01,$F7,$FF,$30,$E2,$7F,$C9,$1D,$3F,$C1,$7E,$13,$F8
	.dc.b	$A4,$7E,$AE,$B7,$A7,$F4,$19,$CD,$FF,$13,$9C,$D7,$EC,$65,$F1,$5D
	.dc.b	$FE,$5F,$A3,$67,$FD,$08,$DB,$FF,$44,$37,$0F,$B7,$F9,$93,$DB,$FC
	.dc.b	$1B,$6F,$FD,$47,$F4,$D7,$0F,$F7,$A4,$DB,$E3,$7C,$0F,$C5,$D6,$6F
	.dc.b	$FA,$5D,$5B,$FF,$30,$37,$A1,$A7,$EA,$A7,$AE,$1F,$27,$BB,$E3,$7B
	.dc.b	$5F,$2F,$DE,$B9,$7F,$9C,$5B,$7F,$E7,$C6,$E9,$FE,$93,$7E,$97,$D3
	.dc.b	$4A,$FF,$53,$7B,$36,$FF,$D0,$73,$6F,$E8,$4D,$AF,$EC,$7B,$3E,$3B
	.dc.b	$DA,$7F,$36,$FA,$8B,$3F,$1A,$2E,$3F,$ED,$79,$B8,$7E,$A2,$F9,$5F
	.dc.b	$FB,$BF,$DA,$CB,$FE,$E9,$7F,$94,$7D,$44,$8F,$F5,$E7,$9D,$71,$61
	.dc.b	$FE,$05,$F3,$17,$98,$FC,$45,$5B,$FE,$6D,$E8,$FF,$38,$C3,$FA,$E7
	.dc.b	$47,$F8,$2E,$7F,$DE,$3E,$77,$E4,$F3,$FE,$A2,$D8,$C9,$FE,$CD,$8F
	.dc.b	$E7,$9F,$35,$23,$FD,$A1,$9D,$C3,$CB,$1F,$D1,$74,$56,$74,$FC,$E5
	.dc.b	$DA,$D4,$DB,$F2,$CD,$E7,$66,$FA,$3F,$D4,$1F,$DB,$7C,$3F,$E2,$98
	.dc.b	$FE,$8E,$32,$FF,$B8,$DD,$78,$B8,$C7,$FD,$D1,$E8,$FF,$55,$F4,$FF
	.dc.b	$64,$DB,$FF,$75,$36,$77,$1F,$E5,$2C,$BF,$BF,$D7,$F4,$AF,$84,$FB
	.dc.b	$75,$F0,$0F,$F4,$23,$8D,$BF,$C7,$59,$FE,$E2,$73,$CD,$C4,$98,$F5
	.dc.b	$9D,$78,$9F,$E9,$F9,$CD,$75,$CB,$FC,$1F,$AE,$31,$FC,$13,$FB,$0F
	.dc.b	$16,$65,$CF,$C2,$B8,$79,$72,$FF,$55,$7C,$7F,$A4,$F9,$5D,$B6,$FD
	.dc.b	$DC,$AE,$1F,$EF,$47,$F7,$BF,$4E,$CF,$9F,$E9,$FE,$17,$C0,$7D,$36
	.dc.b	$C7,$F2,$8E,$34,$E3,$4F,$57,$1A,$F1,$C3,$F9,$BF,$37,$FC,$10,$63
	.dc.b	$1F,$C5,$3F,$DE,$3F,$5B,$78,$74,$FC,$17,$A5,$6B,$F9,$BC,$E9,$B3
	.dc.b	$A6,$93,$2F,$99,$BB,$C2,$E5,$18,$7C,$1D,$D2,$BE,$B6,$F6,$BE,$66
	.dc.b	$3F,$AB,$6D,$7E,$56,$75,$A7,$FB,$12,$B4,$FE,$E7,$6C,$E1,$FA,$77
	.dc.b	$DD,$C1,$23,$F0,$9F,$39,$72,$EF,$36,$FF,$1B,$36,$BF,$B0,$60,$AF
	.dc.b	$F1,$97,$C9,$E7,$0F,$F2,$E6,$A8,$FF,$11,$FE,$C2,$C2,$9F,$AD,$F7
	.dc.b	$AB,$FD,$8A,$34,$D3,$45,$25,$FC,$CF,$0F,$8B,$63,$97,$C8,$78,$0F
	.dc.b	$F3,$7A,$FA,$6F,$DF,$3A,$24,$FE,$66,$D5,$C2,$B7,$39,$8E,$95,$FB
	.dc.b	$7F,$D7,$BC,$F3,$98,$C1,$F0,$3C,$C8,$FE,$F7,$D8,$7D,$36,$2B,$C1
	.dc.b	$57,$8B,$FC,$67,$36,$FF,$CC,$58,$7F,$12,$BF,$F4,$43,$E6,$8F,$17
	.dc.b	$8D,$FE,$83,$8C,$4F,$EE,$DE,$1E,$21,$CE,$1F,$EC,$17,$26,$2F,$BF
	.dc.b	$A7,$CC,$E9,$E1,$D8,$F8,$11,$70,$ED,$36,$FF,$C7,$3F,$49,$FE,$70
	.dc.b	$F6,$95,$FE,$B5,$F4,$BF,$D8,$33,$98,$C0,$FA,$37,$D3,$7F,$77,$F8
	.dc.b	$6F,$F0,$44,$F2,$AE,$11,$88,$46,$2F,$AE,$7F,$EA,$F9,$EB,$87,$4E
	.dc.b	$1F,$C9,$47,$80,$E3,$3B,$C3,$A6,$1C,$E7,$C1,$39,$C3,$D7,$87,$3D
	.dc.b	$7D,$B8,$70,$AD,$5F,$0C,$F3,$87,$91,$B1,$EF,$CF,$DD,$39,$8F,$26
	.dc.b	$5E,$17,$CB,$5C,$32,$F0,$46,$B0,$EE,$E8,$FF,$2C,$D7,$CD,$7F,$DC
	.dc.b	$AD,$7B,$1F,$EE,$8F,$65,$F4,$6F,$A6,$FF,$30,$CF,$FB,$D6,$30,$FF
	.dc.b	$99,$27,$73,$A6,$C7,$83,$81,$F0,$FC,$B0,$46,$27,$53,$98,$C4,$79
	.dc.b	$9D,$6D,$F4,$9B,$AF,$E2,$1B,$3F,$87,$FA,$67,$CC,$FE,$3A,$3E,$4C
	.dc.b	$E9,$FC,$2B,$71,$F1,$E3,$F2,$EF,$8D,$7D,$43,$71,$67,$71,$D2,$38
	.dc.b	$5F,$7C,$7C,$D1,$BC,$3B,$1E,$6D,$98,$F6,$C3,$91,$2F,$8B,$F9,$68
	.dc.b	$D4,$7E,$83,$95,$D0,$F8,$BF,$77,$37,$FA,$E1,$BC,$BF,$EB,$46,$E2
	.dc.b	$BE,$3C,$58,$79,$CD,$C3,$A5,$7F,$15,$8C,$56,$7D,$3D,$7D,$14,$AF
	.dc.b	$E3,$51,$97,$F8,$E4,$65,$FF,$68,$47,$8B,$D2,$7C,$A7,$FB,$CA,$7F
	.dc.b	$D5,$B1,$8E,$7F,$C7,$B6,$33,$3B,$7E,$E0,$8D,$3F,$E0,$F8,$D6,$BF
	.dc.b	$BC,$9C,$BF,$ED,$CF,$A1,$F1,$4F,$FB,$B6,$BF,$99,$3F,$04,$CF,$5C
	.dc.b	$8F,$0B,$E4,$8E,$11,$E2,$BF,$5C,$E4,$E1,$1E,$37,$E8,$EF,$FE,$54
	.dc.b	$D3,$C3,$F4,$DF,$3F,$D5,$97,$FE,$C0,$C5,$3A,$8F,$82,$F8,$48,$E0
	.dc.b	$E0,$ED,$FF,$1C,$7D,$2F,$F4,$C7,$A7,$97,$F7,$7E,$82,$46,$08,$7A
	.dc.b	$ED,$3E,$F2,$47,$D2,$7E,$49,$1F,$F2,$B5,$7F,$08,$7D,$5F,$4D,$F8
	.dc.b	$BF,$A6,$91,$82,$1D,$E3,$E6,$11,$FB,$EF,$D6,$23,$DB,$8F,$F1,$5F
	.dc.b	$4C,$F8,$FF,$CC,$7B,$92,$BF,$DD,$33,$96,$CF,$E3,$23,$11,$F6,$5A
	.dc.b	$FF,$07,$E2,$46,$23,$FD,$F8,$F0,$E4,$FC,$63,$78,$B3,$7F,$AB,$D1
	.dc.b	$C3,$FA,$BF,$99,$98,$FF,$7F,$7D,$3A,$48,$FD,$0E,$28,$F3,$12,$3F
	.dc.b	$3E,$7E,$E6,$DB,$71,$8F,$F3,$75,$71,$7E,$9C,$7E,$EF,$C0,$91,$FE
	.dc.b	$0F,$7F,$7D,$ED,$FC,$AC,$7E,$21,$27,$C6,$FF,$40,$5F,$4B,$EC,$7F
	.dc.b	$A1,$B5,$F2,$B5,$FD,$AB,$5F,$93,$3F,$C3,$7E,$66,$47,$D6,$8F,$B6
	.dc.b	$7D,$52,$3F,$11,$AF,$B9,$DB,$F9,$08,$FE,$35,$F4,$AE,$AF,$E3,$9E
	.dc.b	$A7,$F0,$6E,$2F,$F6,$C5,$EA,$63,$FE,$90,$36,$60,$FF,$50,$57,$DD
	.dc.b	$7F,$94,$37,$C0,$37,$E1,$BF,$C1,$93,$C3,$8F,$C6,$47,$FD,$75,$1A
	.dc.b	$C3,$FD,$D5,$E5,$F9,$89,$3F,$EB,$FB,$7C,$9E,$23,$FD,$0F,$36,$BF
	.dc.b	$C2,$F8,$F8,$AB,$1C,$DF,$EE,$1A,$B7,$E8,$9F,$8B,$E1,$48,$EC,$FF
	.dc.b	$41,$C9,$F2,$F4,$D4,$FC,$29,$C3,$FE,$B6,$F8,$FF,$9A,$47,$D1,$E2
	.dc.b	$57,$FD,$3B,$17,$4F,$3E,$AA,$FE,$A7,$5F,$8C,$E2,$FA,$F8,$C8,$FF
	.dc.b	$22,$DF,$F8,$D1,$D1,$8D,$7F,$BE,$71,$FF,$AF,$A3,$F2,$77,$F9,$78
	.dc.b	$C4,$7E,$52,$30,$78,$23,$35,$D5,$B5,$FE,$54,$D9,$F4,$DF,$D8,$B0
	.dc.b	$FF,$71,$7C,$47,$FA,$65,$FF,$7A,$35,$CF,$B5,$F9,$2E,$DF,$ED,$93
	.dc.b	$9C,$BE,$BB,$71,$FF,$30,$F1,$A3,$18,$7E,$3A,$F8,$FB,$32,$3B,$BF
	.dc.b	$9E,$F1,$24,$60,$87,$E6,$1F,$9B,$F4,$CF,$8C,$FE,$FD,$F2,$1F,$2F
	.dc.b	$4F,$C4,$B7,$4F,$F1,$35,$6F,$0F,$9C,$7F,$9D,$9C,$3F,$C4,$9A,$FF
	.dc.b	$72,$6B,$CA,$FF,$84,$47,$0B,$FB,$4D,$AD,$57,$61,$F4,$CF,$A0,$F3
	.dc.b	$FB,$52,$7C,$7B,$B3,$4F,$F0,$0F,$1E,$56,$A7,$2D,$9C,$AC,$FE,$0F
	.dc.b	$B9,$F1,$09,$1F,$B1,$F0,$FF,$85,$98,$78,$69,$F6,$30,$FF,$61,$C6
	.dc.b	$CF,$8F,$6F,$A6,$FF,$1E,$47,$F8,$D1,$AF,$E7,$1A,$F8,$DB,$1C,$6F
	.dc.b	$FC,$6E,$D9,$EB,$3F,$E4,$78,$FD,$27,$D2,$FF,$1F,$9F,$95,$F4,$CF
	.dc.b	$8A,$FD,$6C,$E6,$7D,$27,$FA,$C1,$C5,$7E,$B6,$B8,$4F,$2A,$FE,$81
	.dc.b	$18,$8F,$EA,$2D,$7F,$C6,$2D,$7F,$CD,$DF,$4C,$F8,$FF,$E9,$0D,$8F
	.dc.b	$F4,$4B,$5F,$EB,$6D,$7F,$C7,$6E,$6F,$F8,$03,$74,$EB,$5F,$C2,$21
	.dc.b	$F1,$3F,$E4,$BF,$A6,$74,$7F,$92,$1B,$8F,$F9,$31,$AF,$F8,$DD,$AF
	.dc.b	$F7,$F7,$38,$FF,$91,$DB,$8F,$FB,$5B,$D0,$46,$2B,$FC,$D5,$EC,$7F
	.dc.b	$B3,$EF,$86,$1E,$32,$4F,$FB,$F7,$DF,$AF,$D2,$C3,$9C,$E2,$BF,$9D
	.dc.b	$DF,$3C,$7C,$1F,$3B,$FE,$5F,$C3,$FC,$28,$D7,$3E,$D7,$F9,$9A,$31
	.dc.b	$1F,$E6,$37,$FC,$C3,$3A,$8F,$DD,$E2,$9C,$C7,$AB,$E9,$BF,$D0,$BF
	.dc.b	$00,$FA,$6F,$F3,$AC,$FE,$C3,$D1,$7C,$87,$F6,$C3,$73,$A9,$FC,$6D
	.dc.b	$71,$F6,$FF,$D8,$ED,$8F,$F0,$CB,$5F,$F3,$CB,$5F,$F5,$B3,$9D,$BF
	.dc.b	$DE,$6D,$D1,$3F,$EC,$AB,$F0,$57,$F6,$2F,$43,$F8,$64,$E7,$BB,$FC
	.dc.b	$03,$E0,$48,$FE,$37,$8F,$B6,$7D,$27,$9B,$03,$49,$3E,$88,$CB,$C9
	.dc.b	$AF,$C0,$6E,$1E,$3F,$A2,$EB,$3B,$AA,$FE,$CB,$C1,$64,$AF,$A3,$C6
	.dc.b	$F5,$BF,$C3,$3F,$35,$30,$F4,$CE,$5B,$3F,$3A,$31,$1F,$31,$AF,$CC
	.dc.b	$6B,$F2,$B4,$FF,$64,$CF,$80,$91,$FE,$16,$F9,$1F,$1F,$9F,$D4,$91
	.dc.b	$F6,$AF,$FB,$6B,$D7,$D8,$F1,$11,$D3,$FE,$49,$F8,$7F,$F6,$B6,$FE
	.dc.b	$1F,$A4,$FF,$7A,$46,$29,$9A,$73,$EB,$FF,$37,$9C,$C6,$2F,$94,$7F
	.dc.b	$B0,$BE,$12,$BF,$EF,$13,$66,$4A,$66,$FB,$3E,$0E,$8F,$C6,$F3,$7E
	.dc.b	$59,$B1,$F9,$66,$BF,$96,$6B,$F2,$D7,$A7,$FE,$17,$8C,$35,$EC,$DA
	.dc.b	$4F,$0A,$F0,$7A,$A7,$F2,$F1,$FE,$34,$7F,$C9,$3F,$31,$9C,$3F,$5F
	.dc.b	$39,$6D,$75,$8C,$47,$9E,$30,$4A,$FD,$E7,$A1,$CF,$3F,$D7,$DF,$37
	.dc.b	$FC,$9A,$FF,$82,$67,$F5,$3F,$3F,$FC,$B9,$BE,$17,$17,$FC,$37,$D0
	.dc.b	$F9,$8C,$67,$F7,$2F,$83,$FC,$19,$B8,$FE,$F9,$B7,$CA,$B8,$EA,$9C
	.dc.b	$DF,$3C,$7E,$54,$74,$8F,$E1,$2D,$7F,$3C,$D7,$85,$FF,$1B,$E9,$EB
	.dc.b	$3D,$3A,$BE,$9C,$78,$3E,$8D,$E2,$53,$D7,$3F,$B8,$F9,$4C,$61,$FC
	.dc.b	$32,$B6,$FF,$1E,$6E,$3F,$C7,$9A,$FF,$1E,$6B,$FB,$37,$7A,$FD,$0F
	.dc.b	$A3,$EB,$3F,$C1,$F8,$11,$F3,$3C,$6F,$E7,$5F,$15,$7C,$73,$E6,$ED
	.dc.b	$FF,$72,$67,$FD,$02,$72,$D9,$FC,$A5,$72,$8F,$F6,$6B,$71,$FE,$5E
	.dc.b	$7F,$52,$9D,$D7,$F8,$36,$BA,$79,$DD,$FE,$03,$C4,$C8,$C9,$FE,$38
	.dc.b	$C7,$FD,$36,$E3,$0F,$EC,$57,$CD,$B5,$CE,$75,$A7,$E7,$F2,$FE,$B1
	.dc.b	$3E,$22,$4F,$A2,$F9,$5F,$A0,$6F,$FE,$46,$9F,$F2,$73,$63,$FC,$46
	.dc.b	$7E,$E3,$E6,$7F,$2E,$F0,$F5,$AC,$EE,$3A,$35,$F8,$AD,$7F,$CF,$71
	.dc.b	$FB,$F5,$F4,$4F,$99,$C3,$70,$FF,$7D,$CE,$A7,$24,$8F,$F2,$5F,$D1
	.dc.b	$7E,$85,$FF,$33,$57,$1A,$FF,$1D,$EA,$AD,$47,$F8,$37,$C3,$FE,$B7
	.dc.b	$7F,$E0,$06,$C7,$09,$DD,$7C,$D9,$CB,$7C,$9F,$C2,$C7,$0A,$FC,$1E
	.dc.b	$E8,$C4,$7F,$89,$5A,$FD,$0F,$7B,$F9,$4D,$FF,$A3,$BA,$3F,$53,$18
	.dc.b	$CB,$FC,$D5,$1F,$AC,$F6,$7F,$43,$18,$3F,$D5,$B7,$99,$E3,$F0,$9B
	.dc.b	$F9,$1B,$5F,$E7,$F7,$0D,$3F,$D6,$B1,$8A,$FB,$FF,$44,$F5,$3E,$FB
	.dc.b	$6B,$FD,$25,$F4,$36,$B7,$5A,$3E,$E5,$66,$3F,$1C,$D7,$F4,$ED,$7F
	.dc.b	$C2,$EF,$3D,$3F,$D5,$FD,$5C,$F0,$FF,$4B,$3F,$13,$D4,$F8,$5E,$CD
	.dc.b	$9F,$F4,$D3,$86,$9F,$ED,$08,$C6,$1F,$E0,$BD,$9F,$45,$F2,$CF,$9F
	.dc.b	$E9,$DB,$FF,$52,$45,$AF,$F6,$DC,$DC,$BF,$DD,$DA,$FF,$BD,$6F,$E1
	.dc.b	$FC,$3F,$F2,$0F,$25,$F3,$DF,$A6,$AD,$63,$FA,$A6,$BF,$B1,$6B,$FC
	.dc.b	$29,$AF,$F8,$E9,$A9,$E7,$87,$F5,$0F,$53,$B6,$1F,$CC,$3C,$44,$5F
	.dc.b	$A1,$99,$FF,$BC,$BC,$DB,$97,$F2,$79,$B8,$FF,$BC,$26,$D7,$F5,$6F
	.dc.b	$2F,$F2,$78,$7B,$5B,$1F,$EF,$C8,$B8,$7F,$8D,$F5,$FE,$8D,$D3,$E2
	.dc.b	$AF,$F3,$27,$D1,$79,$4F,$CE,$E7,$63,$24,$BF,$F2,$05,$27,$51,$FC
	.dc.b	$F6,$AE,$7F,$DF,$7C,$25,$EF,$0E,$53,$98,$7D,$17,$EB,$0F,$EA,$BE
	.dc.b	$17,$F1,$D9,$D4,$66,$BF,$DD,$35,$C1,$FF,$75,$D6,$DF,$F3,$6D,$F8
	.dc.b	$2B,$F6,$FE,$7F,$D1,$F4,$6D,$7F,$A0,$DB,$9F,$F5,$56,$E5,$FE,$08
	.dc.b	$7F,$AB,$7B,$DF,$72,$BF,$D5,$4D,$AF,$E6,$B9,$7F,$00,$5E,$17,$FA
	.dc.b	$F3,$93,$89,$FD,$5A,$BF,$13,$19,$BF,$EA,$F1,$83,$A6,$9F,$E2,$CD
	.dc.b	$38,$CF,$FA,$65,$FC,$C7,$87,$FC,$D7,$9B,$F8,$BE,$5F,$91,$8C,$1D
	.dc.b	$FE,$22,$5F,$66,$CF,$F4,$2E,$7F,$F0,$F4,$FF,$A9,$78,$3F,$2D,$E7
	.dc.b	$FE,$5D,$FA,$7F,$45,$F0,$BD,$AB,$5F,$EB,$2D,$91,$98,$F5,$BE,$F2
	.dc.b	$3F,$D6,$F1,$FE,$B7,$8B,$CB,$FD,$21,$C6,$3F,$C0,$CF,$F8,$17,$E8
	.dc.b	$BF,$D6,$F5,$FE,$C0,$6C,$FF,$A4,$78,$FF,$B8,$72,$FF,$6D,$9C,$B7
	.dc.b	$C1,$FF,$02,$C7,$0A,$FE,$82,$D7,$FD,$C8,$D7,$EC,$BC,$AF,$F1,$78
	.dc.b	$FF,$08,$7C,$55,$FB,$BE,$EF,$37,$64,$9F,$C9,$EB,$71,$FF,$73,$63
	.dc.b	$FA,$F7,$FB,$E7,$32,$31,$1E,$4B,$F3,$92,$FF,$DB,$1A,$73,$D3,$FC
	.dc.b	$B6,$FC,$5E,$19,$3F,$A2,$F9,$6B,$87,$82,$7E,$09,$27,$FC,$4B,$8A
	.dc.b	$31,$3F,$E3,$5F,$8C,$E3,$8F,$F9,$A1,$FF,$1B,$7D,$0F,$F9,$EF,$0F
	.dc.b	$F7,$8B,$6B,$E9,$7D,$16,$A6,$1F,$E3,$29,$D4,$FF,$70,$6B,$FE,$E2
	.dc.b	$D5,$18,$8F,$D4,$BD,$70,$FF,$3C,$F8,$3D,$27,$FC,$94,$F9,$3E,$2E
	.dc.b	$61,$FE,$EC,$AF,$F4,$62,$FA,$2F,$E0,$D6,$9F,$F1,$96,$0B,$E3,$3F
	.dc.b	$52,$31,$7E,$99,$FF,$26,$C6,$27,$51,$FE,$6A,$FA,$1F,$35,$7A,$EB
	.dc.b	$57,$FC,$86,$73,$7E,$CF,$2B,$E5,$9F,$E5,$DF,$A2,$B8,$7C,$C8,$EE
	.dc.b	$47,$E5,$3F,$EB,$1F,$01,$7F,$C7,$EF,$FC,$8A,$F8,$2B,$F9,$0B,$99
	.dc.b	$DE,$5E,$7F,$53,$D0,$DA,$F9,$07,$07,$D9,$E8,$C8,$CE,$1D,$BD,$1C
	.dc.b	$3F,$F0,$07,$1F,$E4,$18,$F9,$2F,$E2,$56,$B4,$FF,$5F,$5B,$E8,$3F
	.dc.b	$D2,$7F,$00,$87,$C0,$FF,$86,$DC,$4E,$9C,$FA,$7B,$FA,$07,$D1,$7F
	.dc.b	$A0,$A7,$27,$0C,$FF,$DB,$2E,$0F,$03,$8C,$3B,$46,$31,$F1,$47,$86
	.dc.b	$FA,$F0,$FF,$B5,$63,$79,$78,$DB,$87,$3C,$BF,$D1,$B3,$97,$CD,$97
	.dc.b	$96,$B7,$8F,$69,$DD,$F1,$CB,$D1,$39,$59,$78,$B7,$C4,$79,$CF,$66
	.dc.b	$5D,$DE,$27,$C2,$9E,$0E,$AF,$B3,$C4,$FF,$46,$4F,$CD,$6C,$7C,$86
	.dc.b	$DF,$F9,$F3,$D7,$7D,$03,$E8,$B0,$DA,$FE,$79,$3A,$9C,$9F,$E8,$4C
	.dc.b	$1C,$0F,$84,$C3,$6F,$FA,$24,$EA,$72,$48,$F7,$CE,$F6,$FA,$6D,$DB
	.dc.b	$F0,$9F,$44,$F9,$4F,$F5,$FD,$6B,$0F,$93,$8F,$5C,$79,$C7,$CF,$DD
	.dc.b	$87,$19,$DE,$5D,$5F,$C9,$B5,$FC,$D3,$99,$E3,$8F,$83,$2F,$C1,$CE
	.dc.b	$F1,$F0,$65,$E3,$C3,$EB,$63,$DD,$F1,$92,$34,$7E,$4E,$34,$FE,$97
	.dc.b	$93,$F8,$0F,$1F,$FB,$73,$93,$F8,$73,$78,$FF,$80,$B7,$0F,$E0,$BE
	.dc.b	$D5,$BE,$31,$FB,$C6,$BF,$BB,$9D,$73,$2A,$FB,$38,$F3,$25,$F1,$AF
	.dc.b	$F3,$BB,$67,$2F,$F0,$59,$E2,$FF,$0A,$1F,$A0,$3E,$73,$FD,$93,$18
	.dc.b	$6C,$66,$3F,$D9,$AD,$7F,$C1,$4D,$D3,$F8,$6B,$74,$FF,$7F,$B6,$7F
	.dc.b	$92,$FD,$1F,$C3,$F0,$9F,$AE,$3C,$0F,$7F,$C2,$4B,$FE,$3F,$5A,$9E
	.dc.b	$15,$B9,$FF,$0D,$4E,$5B,$19,$7F,$BE,$B5,$FF,$67,$B9,$C9,$97,$F6
	.dc.b	$18,$C3,$6F,$E8,$FD,$47,$83,$E9,$3E,$9F,$9D,$31,$F2,$E0,$9D,$47
	.dc.b	$89,$FF,$79,$E5,$EE,$6C,$EE,$74,$72,$8E,$2F,$F3,$BF,$AB,$F7,$FD
	.dc.b	$1F,$E0,$BC,$E9,$3E,$08,$CB,$F0,$1E,$A4,$70,$6C,$87,$F4,$7F,$77
	.dc.b	$E5,$FC,$97,$E3,$BC,$D9,$3E,$C8,$CB,$F3,$1F,$49,$1C,$1B,$21,$FD
	.dc.b	$4F,$EA,$FE,$F7,$BE,$F5,$BF,$37,$F2,$12,$BE,$EC,$E5,$B3,$F2,$23
	.dc.b	$11,$FE,$36,$6B,$FD,$5C,$FF,$3E,$B8,$7F,$C3,$4F,$0E,$4F,$BB,$1A
	.dc.b	$9F,$19,$FA,$CF,$AA,$FD,$6F,$D4,$49,$FD,$53,$E8,$F9,$09,$1F,$8E
	.dc.b	$C1,$1E,$52,$47,$F9,$AF,$C1,$FF,$38,$F9,$3C,$23,$FC,$3D,$D3,$E4
	.dc.b	$AF,$E0,$1E,$EF,$F2,$F9,$C8,$BF,$F3,$CF,$CA,$68,$91,$FC,$16,$5F
	.dc.b	$4D,$F3,$3E,$13,$E3,$12,$7C,$CF,$F9,$F5,$E9,$3F,$DB,$B7,$FF,$61
	.dc.b	$37,$2F,$F5,$AB,$67,$7F,$37,$FB,$5F,$B3,$77,$7F,$17,$FA,$97,$4D
	.dc.b	$35,$12,$FF,$67,$5C,$1B,$1F,$A2,$8C,$47,$F9,$C1,$B9,$7D,$6E,$E4
	.dc.b	$63,$3F,$F2,$A3,$6B,$E3,$62,$F8,$6F,$97,$E2,$FC,$CF,$95,$95,$F1
	.dc.b	$E3,$2F,$C9,$8C,$BF,$43,$8B,$FD,$4F,$3F,$53,$C3,$FA,$BF,$33,$2B
	.dc.b	$E8,$46,$5F,$A5,$19,$79,$70,$FF,$08,$7E,$7F,$AD,$F7,$BE,$2A,$47
	.dc.b	$D9,$8F,$C0,$98,$78,$C6,$1D,$1F,$7D,$D7,$07,$D8,$9F,$E5,$1F,$73
	.dc.b	$FE,$67,$AF,$C2,$78,$BF,$8C,$33,$EE,$48,$FE,$53,$3B,$7F,$1F,$C6
	.dc.b	$8C,$61,$F8,$4F,$9E,$FC,$AF,$C5,$FE,$5B,$E9,$A6,$1F,$91,$8C,$BF
	.dc.b	$93,$8C,$BF,$0A,$3F,$AF,$2F,$A8,$47,$E2,$7E,$7A,$4F,$F5,$19,$CB
	.dc.b	$67,$FA,$8C,$62,$3F,$DA,$CD,$7F,$80,$1D,$63,$67,$82,$38,$61,$FE
	.dc.b	$5E,$6F,$47,$F9,$77,$91,$CF,$FD,$FE,$BF,$47,$D5,$FB,$4E,$89,$97
	.dc.b	$EB,$E3,$F6,$E6,$1F,$BF,$18,$7E,$6F,$9F,$FB,$CD,$24,$60,$87,$D0
	.dc.b	$DD,$3F,$68,$FE,$07,$ED,$FA,$77,$FE,$0B,$CD,$2F,$E3,$46,$5F,$DE
	.dc.b	$6E,$8F,$8E,$48,$CB,$B7,$F8,$69,$F4,$DF,$E2,$3E,$4F,$F8,$4F,$DD
	.dc.b	$7D,$CF,$E0,$B8,$BE,$37,$EE,$FF,$3B,$87,$F9,$4F,$1A,$31,$3F,$E8
	.dc.b	$7B,$E3,$CF,$CF,$0E,$59,$FF,$95,$5B,$3A,$8F,$F0,$F5,$FF,$97,$71
	.dc.b	$FE,$01,$C0,$F3,$FF,$C4,$92,$F9,$9A,$7E,$82,$25,$FE,$71,$AF,$EE
	.dc.b	$E5,$F4,$1F,$92,$FC,$81,$E1,$7F,$87,$1B,$5F,$C9,$5A,$FE,$30,$38
	.dc.b	$7F,$1E,$EB,$2F,$F5,$4B,$76,$FE,$A4,$D9,$FF,$3A,$79,$1F,$C5,$5F
	.dc.b	$F5,$BE,$A8,$C4,$7F,$40,$8C,$74,$7D,$39,$CE,$1F,$43,$F2,$07,$85
	.dc.b	$9A,$98,$74,$BD,$E5,$F9,$DB,$DC,$FF,$8F,$7E,$9D,$EC,$7F,$8F,$1B
	.dc.b	$1F,$D1,$5A,$FC,$16,$BF,$AE,$73,$97,$FB,$21,$B7,$FD,$C5,$B3,$FE
	.dc.b	$C9,$FA,$8F,$EB,$95,$F7,$3D,$4F,$F1,$0F,$21,$A4,$AF,$F1,$CC,$65
	.dc.b	$FE,$2F,$1F,$C7,$8F,$C1,$FA,$76,$32,$43,$F0,$0E,$5F,$D7,$45,$AF
	.dc.b	$CC,$F1,$FF,$80,$F2,$FF,$6F,$78,$9F,$C8,$7D,$2F,$29,$30,$F1,$B5
	.dc.b	$FD,$C3,$FC,$C0,$FC,$AF,$D3,$7F,$9E,$FD,$27,$ED,$1F,$A0,$ED,$FD
	.dc.b	$37,$CF,$7E,$68,$D4,$7E,$92,$77,$DD,$FD,$D5,$B8,$FE,$45,$AF,$F1
	.dc.b	$E6,$BF,$E1,$F7,$3A,$FF,$87,$1B,$A7,$FB,$E5,$B3,$FE,$0B,$D5,$F3
	.dc.b	$DE,$2F,$82,$FD,$6F,$D3,$4A,$FF,$80,$18,$F3,$CF,$FD,$1D,$18,$CB
	.dc.b	$A4,$62,$3C,$8D,$7C,$8D,$77,$7E,$D8,$E1,$5D,$63,$C5,$3F,$C0,$BD
	.dc.b	$7A,$FF,$88,$B3,$F9,$73,$8A,$FB,$35,$F2,$BD,$FF,$E0,$BF,$2F,$30
	.dc.b	$F3,$1F,$ED,$BC,$53,$B8,$F9,$91,$86,$C7,$94,$8F,$2C,$7E,$DC,$71
	.dc.b	$8F,$E1,$9F,$3D,$F6,$BD,$FF,$61,$23,$AB,$EF,$F7,$E4,$78,$1A,$FF
	.dc.b	$BD,$3C,$06,$9F,$5F,$BF,$FC,$B5,$F1,$BF,$BE,$8C,$50,$46,$5F,$F7
	.dc.b	$C4,$F0,$AE,$71,$83,$FB,$2F,$C4,$72,$7F,$DE,$26,$4B,$A7,$F1,$37
	.dc.b	$67,$17,$F3,$7D,$7F,$E6,$88,$FB,$AD,$7F,$06,$D7,$E8,$36,$3F,$47
	.dc.b	$F0,$37,$73,$49,$C7,$F9,$FF,$DC,$95,$FE,$76,$D7,$F3,$91,$89,$FE
	.dc.b	$8F,$F3,$F9,$7F,$3D,$19,$CB,$F1,$0D,$7F,$4C,$D7,$77,$E5,$8C,$47
	.dc.b	$81,$E5,$7F,$C1,$BE,$0B,$B4,$E6,$AF,$67,$F6,$19,$CC,$2F,$FA,$6E
	.dc.b	$1F,$89,$6D,$FF,$53,$F8,$1D,$C6,$31,$F6,$F9,$58,$3A,$F6,$2B,$18
	.dc.b	$73,$9F,$41,$1F,$90,$DC,$3F,$B0,$7C,$F7,$06,$CF,$8F,$DD,$F0,$9D
	.dc.b	$BE,$66,$C6,$4F,$A5,$1F,$14,$8E,$0F,$EE,$FF,$3D,$C2,$BF,$B7,$3F
	.dc.b	$1B,$C5,$7D,$17,$F5,$DA,$FF,$7B,$B6,$BF,$BA,$7C,$F6,$E3,$1B,$7F
	.dc.b	$9F,$DD,$4F,$FA,$EE,$F9,$F4,$FE,$35,$B8,$7F,$BF,$AB,$75,$FC,$E9
	.dc.b	$7C,$1F,$53,$E5,$FC,$F5,$AF,$1C,$5A,$FF,$03,$4D,$C7,$C5,$36,$BF
	.dc.b	$C4,$33,$69,$87,$BA,$BC,$98,$78,$2F,$C9,$18,$3F,$49,$4F,$A4,$61
	.dc.b	$F8,$C8,$C3,$F2,$7E,$C7,$FD,$77,$7F,$C6,$A7,$2D,$AE,$2F,$FA,$ED
	.dc.b	$B1,$FE,$17,$6B,$FE,$64,$6B,$FE,$E5,$F9,$EF,$F6,$BC,$F5,$6C,$7F
	.dc.b	$A8,$7D,$7F,$F7,$A4,$E7,$C6,$F3,$56,$B5,$FE,$04,$D9,$FF,$35,$57
	.dc.b	$0B,$FF,$26,$CE,$5F,$E3,$1F,$8C,$7F,$8E,$6B,$D8,$D8,$DE,$1F,$0E
	.dc.b	$31,$87,$F0,$F7,$53,$FE,$7E,$6B,$FE,$FF,$6B,$DB,$D9,$FA,$7C,$DF
	.dc.b	$8D,$3F,$57,$F4,$17,$1F,$C9,$C7,$EB,$B2,$C5,$FF,$91,$DF,$DF,$FB
	.dc.b	$1F,$C4,$BE,$32,$F6,$9F,$E0,$3E,$13,$1C,$C4,$7F,$A8,$F8,$1F,$E0
	.dc.b	$2F,$9E,$C6,$5F,$EC,$DB,$47,$E0,$7E,$8F,$FD,$D3,$7F,$7D,$B5,$BF
	.dc.b	$47,$5F,$0D,$2B,$FD,$F5,$5A,$AF,$F5,$0B,$5F,$E8,$C1,$C3,$FD,$25
	.dc.b	$EB,$C5,$FE,$7A,$C5,$F3,$DD,$6B,$CC,$FE,$D7,$E2,$E6,$1E,$CA,$7C
	.dc.b	$DE,$2F,$FD,$93,$D8,$BF,$8B,$F3,$39,$CB,$B3,$F2,$3D,$4F,$93,$F3
	.dc.b	$17,$3E,$8D,$C7,$E3,$36,$FC,$CF,$C7,$F8,$19,$5F,$3B,$BF,$C6,$FC
	.dc.b	$98,$CD,$FF,$9E,$77,$47,$EC,$9B,$3F,$2A,$30,$7F,$3C,$A6,$3C,$B1
	.dc.b	$FF,$4F,$C6,$2B,$E1,$7C,$C7,$E0,$BE,$2F,$F0,$6D,$8C,$9F,$1E,$32
	.dc.b	$D3,$F9,$1F,$C5,$FF,$21,$3F,$17,$F0,$76,$32,$7C,$FF,$93,$C6,$BF
	.dc.b	$89,$7F,$23,$ED,$7E,$4B,$E5,$AD,$7D,$56,$D7,$E3,$DB,$A7,$DF,$6C
	.dc.b	$FF,$B8,$79,$51,$8A,$75,$7F,$96,$A3,$F3,$FE,$27,$E8,$3D,$AB,$87
	.dc.b	$E0,$DB,$7F,$9D,$6E,$3F,$8D,$7F,$4F,$ED,$7E,$A3,$D7,$B7,$F8,$B6
	.dc.b	$DF,$E9,$5B,$7F,$95,$7F,$A6,$7D,$B2,$E9,$F5,$F9,$EE,$1F,$E4,$0F
	.dc.b	$31,$5B,$8F,$F5,$D4,$62,$BF,$A5,$4E,$71,$FD,$39,$FC,$F1,$C3,$FD
	.dc.b	$45,$D6,$3F,$82,$E4,$EF,$3F,$E3,$8C,$3F,$C6,$9B,$7F,$3C,$9F,$E0
	.dc.b	$BE,$C4,$9F,$D4,$7C,$BD,$BE,$07,$F4,$AF,$B0,$FE,$62,$FF,$5F,$6B
	.dc.b	$FE,$94,$6B,$BF,$33,$FC,$5D,$8F,$FA,$3B,$E1,$3F,$89,$9F,$E1,$4F
	.dc.b	$67,$FC,$33,$8F,$DD,$8C,$7A,$5F,$DD,$71,$FF,$06,$4F,$F1,$6A,$FE
	.dc.b	$E2,$F7,$F6,$34,$7F,$6F,$9D,$E5,$F1,$B8,$51,$88,$F4,$4F,$92,$FF
	.dc.b	$AC,$C7,$BB,$2F,$F1,$27,$CC,$74,$C7,$C0,$F8,$FB,$A4,$78,$FE,$7A
	.dc.b	$DF,$C6,$CF,$BD,$78,$C9,$5F,$73,$3F,$EB,$B5,$F2,$E7,$E1,$79,$7C
	.dc.b	$B8,$FF,$9B,$4F,$E7,$34,$FE,$75,$87,$E1,$FE,$55,$E5,$FF,$86,$5B
	.dc.b	$1F,$17,$D6,$C4,$C8,$FC,$5C,$EA,$FE,$23,$5F,$F0,$C3,$5F,$F6,$6B
	.dc.b	$BC,$3C,$7B,$FF,$91,$FD,$1E,$95,$F1,$F2,$FF,$22,$61,$C3,$AB,$C8
	.dc.b	$7D,$DF,$98,$B1,$E9,$8F,$F1,$26,$46,$92,$3C,$EF,$E1,$7E,$07,$8E
	.dc.b	$1D,$DF,$81,$E1,$CA,$FC,$27,$CC,$5A,$FA,$58,$2B,$B9,$2B,$E1,$B6
	.dc.b	$7D,$5E,$E7,$1D,$BD,$FA,$23,$E7,$78,$73,$0F,$C9,$79,$B7,$0E,$F3
	.dc.b	$EC,$24,$FD,$06,$C7,$C7,$F9,$8E,$18,$F6,$7F,$C5,$9F,$37,$F9,$C9
	.dc.b	$FF,$54,$37,$8B,$FC,$ED,$AE,$BB,$26,$5F,$C0,$AF,$71,$FE,$38,$BF
	.dc.b	$F2,$6C,$62,$75,$7F,$E7,$FF,$95,$FD,$87,$A9,$FB,$16,$CF,$EB,$5B
	.dc.b	$5F,$75,$B9,$FC,$4F,$3F,$F7,$07,$EE,$7E,$62,$DF,$ED,$62,$DF,$E1
	.dc.b	$66,$D7,$FA,$CF,$E7,$BF,$BC,$FD,$07,$FB,$17,$E2,$1E,$0F,$F8,$0A
	.dc.b	$73,$82,$72,$BE,$81,$F4,$2F,$5F,$F4,$F5,$AA,$5F,$28,$C3,$70,$ED
	.dc.b	$ED,$32,$F1,$47,$FB,$5F,$2E,$B8,$F4,$8D,$CE,$B1,$FF,$6E,$61,$CD
	.dc.b	$B8,$73,$7D,$15,$CF,$2F,$36,$DC,$67,$CB,$5B,$CF,$CA,$DB,$EF,$3C
	.dc.b	$8F,$81,$7C,$9C,$1C,$B1,$E9,$8F,$8E,$30,$71,$CF,$D7,$96,$B1,$ED
	.dc.b	$3E,$8B,$FC,$53,$C0,$FE,$9B,$7F,$49,$B3,$ED,$7F,$A8,$FC,$5B,$B1
	.dc.b	$EF,$72,$8C,$36,$33,$3F,$97,$9C,$BF,$99,$BF,$F1,$AF,$BD,$E2,$6C
	.dc.b	$67,$DD,$FF,$76,$B6,$75,$18,$DF,$EE,$B7,$7F,$C5,$36,$BF,$11,$F4
	.dc.b	$2F,$6B,$FC,$9F,$7C,$30,$F9,$37,$BB,$FC,$4B,$C7,$81,$87,$19,$DD
	.dc.b	$47,$E9,$E5,$E8,$C3,$F2,$F7,$C3,$1E,$F9,$79,$30,$E1,$5E,$1C,$3E
	.dc.b	$B6,$3D,$5E,$79,$78,$F2,$E9,$87,$19,$DE,$3D,$1F,$F1,$0F,$D0,$FE
	.dc.b	$8D,$FF,$3D,$B5,$FE,$B8,$DE,$4F,$E7,$6F,$F9,$22,$37,$C9,$FE,$72
	.dc.b	$9C,$B7,$8F,$FC,$0F,$5C,$5B,$EE,$E6,$FC,$3E,$2D,$EB,$5B,$D4,$C3
	.dc.b	$6B,$CB,$58,$33,$32,$32,$7F,$B2,$FB,$7F,$DC,$AF,$F8,$07,$4E,$DE
	.dc.b	$27,$C7,$D7,$FB,$3C,$AB,$6F,$F8,$60,$C7,$95,$F9,$E6,$9F,$A0,$8D
	.dc.b	$BF,$E2,$C6,$A3,$FD,$D3,$F4,$2F,$27,$D7,$EE,$C9,$FF,$1D,$4E,$5B
	.dc.b	$39,$7E,$8C,$FC,$7F,$41,$5B,$9F,$F0,$43,$71,$FF,$00,$CE,$AF,$84
	.dc.b	$E7,$87,$E9,$F9,$58,$F2,$7E,$6F,$BB,$32,$F1,$C6,$5F,$C1,$C6,$5E
	.dc.b	$F1,$96,$9C,$0F,$C4,$F9,$5E,$0F,$1F,$EC,$FB,$92,$7D,$F1,$97,$F1
	.dc.b	$F1,$97,$D7,$19,$69,$C0,$FC,$AF,$B9,$EC,$F9,$DF,$C2,$FB,$92,$7E
	.dc.b	$5C,$65,$FC,$DC,$65,$F8,$F1,$96,$9C,$0F,$D0,$FB,$BF,$3F,$E3,$9F
	.dc.b	$2D,$F9,$2F,$7A,$57,$F9,$C6,$72,$D9,$CB,$FA,$89,$E5,$12,$34,$43
	.dc.b	$F5,$BE,$F7,$E7,$BD,$D9,$1F,$B2,$7C,$5E,$F4,$8C,$10,$FF,$24,$60
	.dc.b	$8F,$D5,$7C,$FF,$F2,$09,$CC,$71,$7F,$CA,$D1,$9E,$8F,$F1,$AA,$F5
	.dc.b	$38,$92,$7F,$DF,$B0,$F7,$BE,$27,$BF,$D9,$93,$F0,$E3,$C3,$E3,$C8
	.dc.b	$E2,$F9,$3D,$DF,$CD,$7D,$0F,$F8,$B6,$7F,$4D,$97,$F9,$1B,$E8,$5C
	.dc.b	$DF,$C0,$DB,$3F,$E9,$FA,$E0,$FF,$99,$DB,$5C,$6B,$81,$DB,$AE,$57
	.dc.b	$FA,$92,$B8,$BF,$B5,$1D,$E8,$CC,$7E,$E6,$31,$3F,$D7,$BC,$86,$DF
	.dc.b	$C1,$B1,$F9,$3E,$F7,$D0,$3E,$8F,$89,$31,$F9,$31,$97,$E5,$C7,$D6
	.dc.b	$F0,$BE,$BF,$BD,$27,$E7,$C6,$5F,$A3,$19,$7E,$CC,$7D,$EF,$53,$EF
	.dc.b	$FB,$D2,$7E,$BC,$65,$FB,$31,$97,$F0,$31,$97,$F0,$71,$9C,$7F,$07
	.dc.b	$3F,$C3,$FE,$6B,$FD,$0F,$5F,$82,$F0,$FF,$1A,$67,$E0,$24,$7F,$B4
	.dc.b	$27,$6F,$FA,$A5,$AF,$CF,$E0,$46,$23,$FD,$97,$B3,$DF,$FC,$CF,$BF
	.dc.b	$27,$F2,$31,$F9,$C3,$0F,$E6,$A3,$0F,$E2,$BC,$8F,$B9,$F4,$3C,$E3
	.dc.b	$07,$FB,$FB,$9D,$EF,$FE,$34,$FE,$A5,$09,$E0,$FF,$41,$6E,$FF,$95
	.dc.b	$9E,$13,$D0,$E1,$8F,$2A,$E3,$1F,$EE,$E6,$BE,$48,$FE,$95,$18,$87
	.dc.b	$2F,$E3,$3A,$3F,$73,$F0,$12,$7F,$63,$1F,$BC,$30,$FE,$EA,$30,$FB
	.dc.b	$B7,$F9,$BF,$0D,$FC,$0B,$29,$39,$E0,$4F,$DD,$3F,$C7,$35,$C2,$7E
	.dc.b	$D4,$62,$77,$1F,$E3,$EC,$9B,$FF,$0E,$E0,$93,$F0,$E7,$2D,$9C,$BF
	.dc.b	$C4,$27,$E3,$44,$8E,$8F,$C4,$23,$FE,$95,$F7,$9F,$43,$F8,$CE,$0F
	.dc.b	$F6,$5F,$CA,$FE,$E6,$BF,$DE,$0D,$8F,$F7,$0D,$70,$E4,$FF,$07,$36
	.dc.b	$BA,$5F,$E4,$5F,$F2,$8B,$63,$51,$F9,$9F,$37,$FC,$29,$83,$E3,$A9
	.dc.b	$FA,$58,$93,$FA,$17,$E4,$11,$FF,$32,$FD,$03,$BB,$C7,$F9,$1B,$E6
	.dc.b	$A9,$F9,$29,$D5,$7E,$35,$AF,$FA,$DD,$AF,$C1,$6B,$FC,$6D,$AF,$F9
	.dc.b	$B3,$C7,$73,$7C,$26,$C7,$E9,$34,$FE,$81,$3F,$E3,$1F,$A1,$78,$3B
	.dc.b	$89,$5B,$BE,$2B,$87,$FD,$FB,$7C,$23,$F4,$9F,$42,$F5,$FF,$5B,$F4
	.dc.b	$3F,$D2,$67,$F5,$0D,$7F,$C5,$D3,$B7,$8C,$FF,$8B,$E3,$11,$FE,$D7
	.dc.b	$6B,$F8,$16,$BF,$CD,$FA,$1E,$97,$F8,$FB,$A2,$4F,$E8,$63,$F8,$E1
	.dc.b	$F9,$D2,$3F,$C6,$5A,$FB,$7C,$1E,$91,$C2,$3F,$85,$F7,$B9,$FF,$C0
	.dc.b	$31,$B3,$FA,$5E,$1E,$27,$C0,$43,$F2,$FE,$8F,$F2,$DF,$43,$FC,$A7
	.dc.b	$E9,$C9,$FD,$B4,$65,$FE,$58,$FE,$CC,$8F,$FB,$E7,$E1,$5F,$1F,$FA
	.dc.b	$5D,$3F,$CA,$19,$F9,$7E,$83,$F3,$27,$F9,$BE,$B7,$DF,$6F,$86,$1F
	.dc.b	$C4,$B5,$5F,$08,$FE,$53,$7C,$1B,$5F,$A9,$8C,$47,$EC,$DA,$FF,$6D
	.dc.b	$6B,$FE,$63,$7C,$FF,$05,$FA,$CF,$9F,$93,$D2,$3C,$86,$1F,$DE,$C6
	.dc.b	$1D,$10,$FA,$13,$D1,$FA,$B8,$7F,$88,$DD,$73,$34,$F1,$63,$FE,$99
	.dc.b	$9E,$2D,$9F,$E1,$BE,$AF,$C5,$F7,$3F,$80,$F7,$4B,$F8,$31,$97,$BC
	.dc.b	$65,$E7,$19,$69,$C0,$F2,$FC,$FF,$D1,$F7,$3F,$D4,$1E,$D7,$F1,$9C
	.dc.b	$FE,$CB,$69,$AF,$DE,$AF,$B1,$EE,$70,$F8,$D9,$5F,$52,$3F,$AD,$FC
	.dc.b	$67,$EE,$B6,$FA,$1E,$07,$F8,$7E,$31,$5D,$23,$1E,$17,$EE,$E3,$07
	.dc.b	$FA,$66,$77,$97,$F9,$16,$3F,$98,$7C,$5F,$46,$9C,$0F,$F7,$29,$4F
	.dc.b	$F5,$13,$B8,$FC,$BF,$6F,$D3,$9F,$D4,$74,$B6,$FE,$5D,$BF,$E3,$AB
	.dc.b	$5A,$6E,$B5,$E4,$E2,$BE,$DF,$BF,$71,$FE,$D7,$8B,$E8,$B1,$9F,$E9
	.dc.b	$AB,$58,$F7,$9C,$FC,$27,$CE,$C7,$F3,$ED,$C7,$FA,$6F,$D0,$6B,$19
	.dc.b	$5F,$8E,$F8,$2B,$5F,$D4,$9B,$AF,$F5,$66,$CF,$ED,$3A,$D1,$E0,$F3
	.dc.b	$64,$FE,$63,$E0,$6E,$1F,$9C,$CF,$FA,$44,$F9,$2B,$D3,$82,$F7,$8F
	.dc.b	$CB,$C9,$87,$0C,$7F,$B2,$FD,$06,$B2,$F1,$36,$3F,$80,$37,$6F,$E3
	.dc.b	$8F,$C4,$F7,$3F,$DC,$B3,$FB,$9B,$46,$23,$F2,$B1,$8B,$FE,$28,$D8
	.dc.b	$FE,$08,$DC,$7F,$BC,$FC,$FF,$5A,$FD,$BB,$6B,$FB,$DB,$67,$73,$FE
	.dc.b	$1D,$F9,$8F,$F4,$B3,$FE,$01,$AF,$F3,$CC,$61,$FF,$02,$7B,$08,$C4
	.dc.b	$7F,$94,$5A,$FF,$9F,$5E,$15,$F8,$DE,$4F,$F5,$FD,$F7,$EA,$FC,$3C
	.dc.b	$62,$FF,$9F,$61,$D7,$E0,$3F,$20,$DB,$FE,$23,$F2,$B8,$C9,$54,$93
	.dc.b	$A3,$D8,$E2,$FD,$33,$C2,$7D,$D7,$D6,$73,$5F,$46,$31,$97,$A7,$E8
	.dc.b	$B9,$36,$FF,$CA,$DF,$3C,$CB,$EC,$62,$8C,$4F,$CB,$8C,$4E,$A7,$31
	.dc.b	$F3,$DB,$5F,$B4,$FA,$97,$57,$EC,$6B,$8E,$1F,$E7,$0A,$56,$A3,$FC
	.dc.b	$D1,$F3,$BF,$ED,$76,$F8,$FF,$B4,$F6,$BE,$75,$FF,$05,$C1,$7B,$7F
	.dc.b	$B7,$18,$AF,$04,$66,$7F,$83,$57,$2B,$FF,$5F,$4E,$5F,$F6,$07,$CF
	.dc.b	$FF,$0E,$C3,$F1,$4D,$7F,$C9,$8D,$77,$DF,$F2,$23,$86,$1F,$E4,$EF
	.dc.b	$C6,$FA,$50,$4E,$AF,$1E,$D2,$B7,$C3,$F9,$5E,$EF,$B5,$C9,$FB,$7C
	.dc.b	$7F,$2D,$18,$26,$1F,$D7,$34,$FC,$64,$7F,$02,$7F,$90,$63,$FD,$1B
	.dc.b	$E2,$BF,$68,$DA,$FF,$35,$76,$A7,$2D,$8C,$C7,$F2,$4F,$35,$F3,$FD
	.dc.b	$2B,$59,$7F,$36,$9C,$DF,$FA,$87,$E7,$FF,$DE,$EF,$E2,$5B,$3F,$B5
	.dc.b	$6B,$FE,$77,$6B,$FE,$F1,$77,$C1,$FC,$F6,$9F,$37,$F8,$B9,$CD,$FF
	.dc.b	$87,$27,$2F,$87,$E1,$BD,$15,$B6,$C7,$F0,$7F,$89,$F9,$A4,$7F,$57
	.dc.b	$C0,$8C,$65,$ED,$F9,$7F,$DC,$47,$C3,$7E,$27,$AB,$F1,$7E,$5A,$E1
	.dc.b	$FE,$42,$CD,$18,$C3,$E3,$37,$5F,$63,$EB,$F8,$2F,$D8,$ED,$FE,$74
	.dc.b	$C3,$E4,$46,$08,$FF,$93,$BB,$99,$71,$8C,$37,$1F,$ED,$33,$A9,$FA
	.dc.b	$95,$FD,$7A,$31,$3F,$67,$E5,$BE,$F7,$C4,$FD,$F6,$DF,$AD,$B3,$B8
	.dc.b	$F7,$4F,$F1,$BF,$96,$FC,$B9,$23,$E5,$B6,$7F,$0D,$F2,$1B,$9F,$8C
	.dc.b	$FE,$37,$D9,$FC,$77,$CA,$59,$F9,$4D,$AF,$C7,$37,$7F,$C2,$36,$7F
	.dc.b	$04,$D9,$FF,$74,$F0,$23,$F3,$5E,$07,$E6,$FD,$9B,$7F,$5D,$B7,$F9
	.dc.b	$B6,$E1,$F9,$17,$F4,$5E,$C7,$E8,$FD,$5B,$7F,$7D,$B7,$FA,$36,$DF
	.dc.b	$E6,$5F,$E8,$7F,$3E,$BF,$E6,$CD,$BF,$E7,$6D,$BF,$EA,$ED,$7F,$85
	.dc.b	$0F,$05,$1B,$C3,$FC,$45,$C4,$E2,$EF,$8F,$DF,$6E,$3F,$89,$6E,$3F
	.dc.b	$86,$7F,$7D,$E3,$C9,$FC,$D7,$97,$6B,$FA,$EF,$CF,$FF,$1C,$F4,$BF
	.dc.b	$C1,$7E,$07,$F5,$E8,$C7,$A1,$FE,$25,$D7,$F7,$1F,$03,$FC,$24,$FF
	.dc.b	$16,$C6,$72,$FD,$C4,$63,$D0,$E5,$EA,$59,$FE,$51,$7F,$E4,$3E,$26
	.dc.b	$3F,$E1,$D9,$C9,$FC,$B6,$33,$5D,$BE,$5B,$9C,$FF,$AD,$32,$46,$63
	.dc.b	$0E,$B5,$FF,$27,$CE,$A7,$AC,$7C,$FC,$3F,$AE,$C7,$76,$C3,$5F,$F1
	.dc.b	$6F,$CB,$7C,$3E,$39,$7E,$1F,$22,$DF,$66,$EB,$FE,$59,$CB,$F9,$5B
	.dc.b	$EF,$CB,$AE,$5F,$E3,$8A,$FF,$05,$FC,$93,$E1,$29,$F3,$3D,$5B,$9F
	.dc.b	$A3,$81,$5B,$8F,$85,$E7,$6B,$5F,$7B,$74,$F1,$36,$7F,$D8,$FA,$FF
	.dc.b	$91,$E7,$E9,$5F,$F1,$C9,$EC,$DB,$F9,$3F,$2B,$E6,$F1,$E5,$7D,$1F
	.dc.b	$80,$B5,$F3,$9B,$87,$C1,$6C,$F8,$1F,$7F,$87,$23,$ED,$7C,$AD,$BF
	.dc.b	$F1,$7E,$4A,$FC,$6F,$B5,$9C,$7E,$63,$70,$F5,$B6,$FE,$C6,$9F,$E5
	.dc.b	$19,$FC,$56,$1F,$C1,$DF,$6F,$87,$F9,$8F,$33,$E8,$B2,$3F,$19,$F3
	.dc.b	$F7,$0F,$C4,$B7,$2F,$B8,$D9,$F9,$6F,$F8,$8F,$E6,$7F,$C9,$F5,$FE
	.dc.b	$05,$6A,$69,$3B,$3F,$C9,$FF,$11,$C5,$B7,$F9,$DF,$94,$FD,$6F,$A5
	.dc.b	$FA,$E6,$C6,$4F,$D1,$46,$5A,$7E,$97,$CE,$FD,$A9,$FB,$6F,$96,$B1
	.dc.b	$92,$65,$FA,$68,$CB,$FE,$30,$F9,$BF,$EE,$BF,$11,$FE,$BC,$F0,$7F
	.dc.b	$8B,$C6,$25,$F0,$E9,$FF,$12,$7C,$F3,$E7,$DE,$BF,$FB,$06,$B5,$7F
	.dc.b	$E2,$9C,$39,$46,$21,$18,$C3,$BF,$B2,$CB,$C5,$4C,$BB,$63,$D6,$F7
	.dc.b	$5A,$AE,$77,$C5,$F3,$CE,$9A,$73,$CB,$CF,$A7,$1C,$BC,$53,$E7,$BE
	.dc.b	$59,$F9,$E7,$39,$F8,$AF,$85,$6B,$1E,$D7,$B8,$C6,$1D,$F3,$F4,$3D
	.dc.b	$73,$E2,$7F,$9D,$72,$F9,$D1,$FD,$18,$F5,$C6,$6B,$D9,$19,$F8,$A7
	.dc.b	$63,$DC,$FF,$3F,$46,$1B,$19,$9F,$F5,$64,$E5,$FF,$57,$60,$F9,$E7
	.dc.b	$9B,$E3,$6C,$E9,$AF,$96,$78,$4E,$CF,$95,$3C,$34,$F9,$0D,$D7,$F0
	.dc.b	$FF,$3E,$F6,$7F,$B1,$D6,$AF,$89,$31,$FA,$FC,$2B,$E1,$19,$AE,$0F
	.dc.b	$DC,$3F,$17,$8F,$77,$F1,$13,$C1,$D6,$5E,$86,$CE,$DE,$38,$7E,$43
	.dc.b	$0F,$06,$3E,$6C,$BF,$47,$7B,$C3,$96,$5E,$23,$F3,$F7,$C3,$1F,$CC
	.dc.b	$F9,$4F,$41,$F4,$9F,$E0,$2E,$EF,$F6,$A3,$79,$7F,$D1,$D5,$B8,$C5
	.dc.b	$FF,$55,$B7,$C8,$62,$B5,$3F,$D7,$9A,$B0,$EA,$E6,$B8,$BB,$F9,$DF
	.dc.b	$ED,$6D,$ED,$E9,$06,$FA,$32,$74,$43,$B6,$7E,$8A,$E7,$1F,$EB,$AF
	.dc.b	$03,$E8,$3F,$9B,$6B,$FE,$22,$6B,$FC,$A7,$8B,$FD,$AC,$FF,$29,$7E
	.dc.b	$0B,$4F,$EF,$65,$3C,$27,$C5,$F8,$6F,$8F,$C8,$FD,$D5,$F2,$9F,$2B
	.dc.b	$27,$C8,$7F,$6C,$C3,$FC,$0B,$18,$F2,$38,$DF,$09,$FF,$04,$E1,$CF
	.dc.b	$0F,$F7,$53,$63,$F6,$C7,$EA,$27,$83,$FB,$98,$D3,$63,$3C,$0C,$3F
	.dc.b	$D9,$95,$A8,$C5,$73,$BF,$BD,$C0,$F9,$3F,$2E,$98,$F9,$39,$7E,$6C
	.dc.b	$D3,$A8,$F2,$E5,$F0,$E3,$D3,$3A,$9C,$B5,$F3,$47,$17,$E6,$CF,$17
	.dc.b	$F1,$5A,$FD,$4E,$87,$B5,$E6,$F9,$29,$5F,$5A,$72,$D9,$FB,$B1,$88
	.dc.b	$FE,$23,$18,$23,$B2,$1F,$7B,$9B,$F5,$5E,$D7,$E9,$3D,$57,$E7,$3F
	.dc.b	$C9,$F8,$BE,$4D,$7E,$BF,$E4,$BE,$49,$B9,$D7,$27,$F9,$3D,$FC,$7D
	.dc.b	$F2,$C7,$ED,$9F,$E7,$58,$C9,$1E,$44,$3F,$59,$F4,$DD,$7E,$57,$EB
	.dc.b	$FC,$A4,$9F,$F1,$D5,$27,$51,$F8,$8B,$F4,$C7,$81,$B1,$99,$E0,$D9
	.dc.b	$E2,$FE,$82,$76,$FC,$5E,$67,$D3,$7B,$FD,$19,$3E,$27,$E1,$78,$12
	.dc.b	$3B,$3E,$1C,$7F,$82,$C7,$F5,$1C,$3E,$6C,$FC,$4D,$BF,$8D,$7C,$07
	.dc.b	$1A,$F1,$5F,$0E,$0F,$A3,$9A,$BE,$87,$4C,$BF,$86,$FE,$73,$E9,$7E
	.dc.b	$A7,$37,$84,$93,$F0,$5F,$27,$27,$D9,$24,$7F,$1B,$F9,$0F,$E0,$BF
	.dc.b	$45,$FD,$53,$3F,$C0,$FE,$B7,$FA,$AA,$FC,$43,$63,$13,$BE,$6F,$F5
	.dc.b	$9B,$FD,$72,$BF,$B8,$35,$CF,$5F,$F1,$48,$C4,$03,$CD,$C3,$5C,$C7
	.dc.b	$F3,$7F,$C0,$9E,$F7,$F3,$9C,$5F,$27,$7E,$12,$61,$FC,$EE,$72,$D9
	.dc.b	$CB,$EC,$9E,$51,$CF,$D3,$E9,$E7,$4A,$E5,$19,$7A,$46,$5F,$8B,$1E
	.dc.b	$3F,$57,$C9,$F2,$32,$7C,$51,$97,$C9,$19,$7E,$6C,$65,$F9,$F1,$9D
	.dc.b	$FE,$9C,$FE,$DF,$F3,$AF,$A5,$C1,$FD,$7B,$47,$C1,$FC,$33,$51,$31
	.dc.b	$FF,$51,$57,$07,$FD,$D0,$DE,$A5,$FF,$B9,$E7,$E4,$7C,$97,$CD,$FA
	.dc.b	$09,$18,$21,$F8,$47,$2F,$E1,$A3,$F1,$7F,$0F,$FD,$C3,$DB,$7D,$C3
	.dc.b	$E2,$30,$DB,$FE,$A7,$D0,$BE,$31,$FC,$2A,$30,$D8,$FC,$51,$1E,$2F
	.dc.b	$07,$F2,$FF,$11,$F8,$5F,$94,$FC,$37,$C3,$49,$FC,$24,$65,$FC,$34
	.dc.b	$65,$FC,$FC,$7C,$FE,$1F,$C7,$7C,$9C,$9F,$97,$19,$69,$0F,$D3,$FC
	.dc.b	$9F,$EB,$3E,$07,$F2,$DF,$23,$27,$DF,$19,$7D,$CF,$E9,$48,$E0,$D9
	.dc.b	$0F,$D9,$FC,$9F,$E0,$F8,$BF,$45,$E8,$C9,$FB,$F1,$96,$90,$FD,$DF
	.dc.b	$91,$F8,$BD,$7F,$55,$F2,$32,$7F,$0F,$19,$69,$0F,$E9,$5F,$BA,$FE
	.dc.b	$DF,$CD,$75,$35,$FE,$64,$DA,$FE,$6C,$D7,$F9,$23,$5F,$E0,$47,$E5
	.dc.b	$23,$F3,$C4,$7F,$C6,$95,$FE,$2B,$6F,$BF,$FE,$39,$8C,$47,$92,$31
	.dc.b	$EF,$E5,$91,$C0,$E7,$1F,$D8,$71,$FE,$27,$89,$A4,$F1,$FF,$A6,$3E
	.dc.b	$F2,$3E,$1F,$83,$FF,$16,$FD,$13,$7F,$F1,$8B,$71,$FE,$C6,$D7,$FC
	.dc.b	$3E,$D7,$FD,$2C,$E7,$C2,$FE,$60,$DF,$07,$E2,$B6,$7F,$94,$7D,$1F
	.dc.b	$F2,$58,$C4,$1A,$62,$35,$7E,$62,$4F,$F9,$0A,$3F,$97,$78,$6F,$93
	.dc.b	$FF,$24,$B6,$7F,$BD,$B5,$FF,$1B,$B5,$FF,$57,$39,$C7,$FA,$43,$71
	.dc.b	$FB,$AD,$9F,$E8,$1F,$B0,$BD,$AF,$F2,$7F,$EE,$9F,$E0,$C9,$78,$EE
	.dc.b	$4E,$0D,$CF,$73,$0F,$F1,$AD,$7F,$98,$38,$0D,$24,$6C,$FF,$19,$BD
	.dc.b	$E9,$F6,$6E,$BF,$1F,$C1,$7E,$37,$E5,$A6,$1F,$E5,$AC,$3F,$CA,$3B
	.dc.b	$FF,$94,$1F,$F2,$BC,$7E,$6B,$2F,$0D,$FF,$97,$A7,$33,$B8,$C1,$D9
	.dc.b	$E8,$FE,$6F,$C4,$C4,$79,$A2,$F0,$FF,$8D,$36,$FE,$47,$3E,$AF,$83
	.dc.b	$FD,$17,$91,$23,$CD,$1F,$30,$F1,$11,$F8,$ED,$79,$1C,$0F,$D7,$FA
	.dc.b	$3F,$53,$D5,$F5,$12,$3E,$7B,$ED,$F5,$64,$7D,$26,$BF,$EE,$69,$FF
	.dc.b	$55,$3F,$EE,$4E,$EF,$DF,$FC,$67,$F7,$AC,$FF,$CF,$31,$8A,$DE,$FF
	.dc.b	$BC,$AD,$5F,$13,$FD,$F9,$F1,$7C,$1A,$64,$FF,$71,$94,$FF,$07,$3B
	.dc.b	$9F,$C2,$F3,$FF,$87,$67,$ED,$35,$FF,$7C,$F0,$FE,$57,$8B,$F1,$B3
	.dc.b	$9A,$FC,$6F,$C0,$DA,$34,$95,$F9,$4F,$6F,$FD,$27,$AF,$FA,$52,$FF
	.dc.b	$2B,$18,$9F,$F6,$BF,$13,$C8,$C6,$7E,$2A,$B7,$C2,$FA,$12,$7F,$D4
	.dc.b	$D9,$7F,$91,$E3,$18,$FE,$1F,$E5,$B8,$CE,$6B,$F9,$CC,$74,$FD,$CD
	.dc.b	$EB,$53,$C3,$D9,$FB,$1F,$96,$C6,$5F,$EB,$AA,$DB,$6B,$7B,$2B,$78
	.dc.b	$FF,$A8,$2F,$6F,$FA,$ED,$AF,$E9,$DA,$FF,$AE,$9A,$B4,$47,$2D,$3F
	.dc.b	$45,$F2,$DD,$9B,$97,$97,$CF,$F6,$1E,$AE,$1F,$F5,$6C,$64,$98,$F8
	.dc.b	$E7,$EE,$61,$FE,$2A,$9F,$F5,$FB,$67,$51,$FD,$4B,$E2,$BA,$B4,$FF
	.dc.b	$75,$FE,$05,$FE,$E6,$6F,$6F,$DA,$5F,$B6,$FF,$A6,$5C,$92,$FF,$DC
	.dc.b	$35,$C5,$B5,$F9,$35,$19,$7F,$B3,$8F,$A4,$FB,$9A,$7F,$9D,$DB,$7F
	.dc.b	$E6,$DF,$31,$F6,$2F,$3B,$F1,$B8,$71,$E8,$B8,$70,$8F,$F3,$A3,$6F
	.dc.b	$8D,$6A,$A6,$1F,$E7,$9C,$7C,$11,$F9,$8A,$DC,$E4,$FE,$FB,$7D,$70
	.dc.b	$F0,$47,$FB,$1A,$3F,$29,$1E,$C6,$BC,$2B,$F6,$95,$FD,$D7,$0E,$DE
	.dc.b	$77,$AF,$0F,$F0,$2F,$CD,$62,$64,$7F,$B1,$67,$58,$7F,$AC,$DA,$FF
	.dc.b	$84,$2F,$A4,$FC,$C8,$FE,$A3,$CC,$F9,$AF,$CE,$9F,$9E,$F9,$AB,$5F
	.dc.b	$E1,$48,$B5,$FE,$24,$9B,$AA,$BE,$96,$3F,$E6,$17,$E2,$53,$1F,$F0
	.dc.b	$87,$8D,$C2,$76,$FB,$8F,$83,$5C,$6B,$D5,$18,$33,$DD,$F1,$0F,$D9
	.dc.b	$FC,$D5,$BE,$F8,$7F,$87,$67,$F6,$FF,$27,$9C,$3F,$C7,$6F,$CC,$EA
	.dc.b	$97,$FE,$30,$FA,$AB,$5F,$EE,$7C,$D5,$FC,$03,$EA,$73,$AF,$F5,$57
	.dc.b	$F1,$BF,$44,$69,$FC,$33,$F6,$7F,$F6,$06,$5F,$0B,$15,$6E,$30,$D8
	.dc.b	$FA,$8D,$7F,$42,$D7,$FC,$92,$D7,$FD,$98,$F1,$BF,$A5,$1F,$E0,$CA
	.dc.b	$7C,$1F,$92,$FE,$9F,$86,$9C,$C7,$C8,$CF,$F0,$4D,$E3,$F5,$7D,$3F
	.dc.b	$E4,$99,$1C,$AB,$F7,$D8,$FC,$79,$F8,$1F,$27,$F0,$7B,$7E,$17,$D4
	.dc.b	$DC,$38,$36,$DA,$FA,$34,$F2,$4E,$63,$19,$23,$F7,$78,$B3,$A7,$CF
	.dc.b	$F9,$3B,$97,$CF,$9D,$3F,$9F,$C9,$18,$7F,$44,$D7,$F8,$7B,$5F,$F3
	.dc.b	$BB,$BC,$7C,$7F,$39,$DA,$7D,$35,$FE,$FD,$CF,$E6,$4F,$E0,$7E,$53
	.dc.b	$F0,$5F,$33,$6B,$E4,$36,$F8,$9F,$5D,$C6,$1F,$01,$FE,$1D,$F9,$C7
	.dc.b	$D6,$9D,$F4,$A3,$15,$F1,$BD,$5D,$F2,$4C,$7E,$B7,$CA,$DC,$7B,$EB
	.dc.b	$F8,$18,$FC,$B7,$CD,$7E,$5F,$D7,$B7,$F7,$5B,$7C,$39,$11,$8C,$FE
	.dc.b	$AB,$6B,$E1,$4E,$9F,$C3,$5F,$57,$F1,$1E,$BF,$E9,$7D,$3F,$D3,$36
	.dc.b	$32,$7E,$36,$32,$D3,$F2,$1E,$77,$EB,$7E,$67,$F5,$CD,$8C,$9F,$95
	.dc.b	$8C,$B4,$FC,$27,$1F,$E2,$8F,$F0,$0F,$CA,$BC,$EC,$4E,$8F,$C4,$73
	.dc.b	$AB,$FA,$B3,$63,$31,$83,$F8,$07,$CC,$F2,$E0,$F9,$0F,$DA,$E4,$95
	.dc.b	$FA,$9F,$2E,$D7,$DC,$E6,$70,$7F,$42,$6D,$7F,$37,$E7,$FC,$04,$E4
	.dc.b	$D7,$1F,$E1,$8F,$E2,$1F,$31,$6B,$F6,$DF,$31,$F8,$AE,$1F,$E2,$8D
	.dc.b	$8F,$E4,$F8,$A7,$C0,$48,$F0,$46,$5B,$AF,$E5,$0E,$59,$FA,$5F,$CB
	.dc.b	$E1,$27,$F8,$7B,$FD,$15,$EE,$E6,$7F,$BE,$E7,$F9,$B8,$F8,$04,$9F
	.dc.b	$0B,$FE,$36,$1F,$2F,$C7,$9F,$F8,$44,$62,$7F,$A0,$FE,$A9,$70,$FD
	.dc.b	$3B,$6B,$7F,$33,$FD,$E6,$31,$5F,$E0,$97,$85,$7F,$67,$6B,$9C,$3F
	.dc.b	$BA,$C6,$23,$FA,$E3,$C9,$FF,$5D,$E1,$FE,$39,$AE,$07,$C2,$F6,$7F
	.dc.b	$8C,$F5,$C8,$F8,$BF,$33,$6F,$F8,$19,$FC,$5A,$7F,$1E,$49,$FC,$9B
	.dc.b	$63,$F7,$76,$F8,$7E,$5A,$7C,$47,$FC,$D7,$F1,$C9,$A7,$CB,$F9,$9B
	.dc.b	$8F,$F7,$6F,$71,$5B,$8F,$F0,$34,$62,$BF,$D0,$93,$9C,$7F,$C6,$B3
	.dc.b	$9A,$DC,$E7,$1F,$AB,$EE,$76,$BF,$3B,$FE,$7E,$FD,$CF,$E3,$47,$F9
	.dc.b	$85,$BE,$1F,$9F,$E9,$B5,$C1,$32,$F1,$D7,$F9,$A3,$33,$49,$3F,$E5
	.dc.b	$3F,$B7,$73,$71,$CB,$F9,$7E,$09,$CB,$FD,$D7,$E8,$3F,$CF,$8D,$F8
	.dc.b	$1F,$C8,$7C,$E7,$F9,$2A,$73,$7F,$59,$AF,$F9,$59,$A8,$8E,$1B,$3F
	.dc.b	$EB,$1D,$5F,$78,$F8,$9F,$F4,$9B,$FE,$83,$C7,$FB,$BB,$96,$ED,$A9
	.dc.b	$DB,$63,$27,$EC,$3E,$47,$F5,$07,$EA,$7E,$6E,$C6,$49,$97,$E2,$A3
	.dc.b	$2F,$E4,$BE,$83,$FD,$4F,$E7,$7F,$B3,$3E,$57,$D9,$29,$CA,$F8,$B4
	.dc.b	$FC,$1F,$A3,$7D,$37,$E5,$72,$E5,$7F,$17,$0E,$38,$70,$8C,$72,$FF
	.dc.b	$86,$1E,$98,$F8,$2B,$6F,$5C,$3A,$4E,$8E,$39,$76,$7C,$D7,$CF,$0E
	.dc.b	$D1,$89,$F1,$E5,$E6,$ED,$FF,$14,$D6,$B2,$EF,$7D,$0E,$C7,$13,$C1
	.dc.b	$5C,$73,$F0,$4F,$09,$D5,$73,$C7,$C5,$9F,$A7,$1E,$EF,$B6,$33,$5D
	.dc.b	$CF,$F5,$0D,$FC,$E6,$C7,$C5,$6E,$5F,$E1,$DF,$8E,$FF,$76,$7D,$35
	.dc.b	$F8,$23,$49,$5F,$E8,$1C,$1E,$17,$FB,$46,$30,$D7,$C2,$D7,$E9,$35
	.dc.b	$FB,$4F,$1D,$3E,$43,$73,$DF,$17,$CE,$FA,$67,$C1,$FE,$2A,$F7,$3F
	.dc.b	$22,$F8,$61,$D2,$38,$56,$F0,$E7,$3F,$4B,$1F,$4D,$71,$37,$14,$D5
	.dc.b	$F4,$7F,$30,$F6,$E1,$65,$E4,$9E,$19,$79,$1E,$37,$D1,$FC,$77,$3B
	.dc.b	$E7,$1E,$D7,$E9,$3C,$2F,$F2,$D4,$62,$BF,$CD,$ED,$7F,$C6,$2F,$F8
	.dc.b	$2D,$CC,$F1,$E5,$FF,$25,$4E,$EB,$9C,$E6,$97,$E1,$F4,$B0,$BE,$9B
	.dc.b	$FD,$C1,$3F,$E7,$E7,$F8,$63,$7A,$F7,$1A,$8F,$F2,$1B,$70,$FF,$39
	.dc.b	$3C,$23,$33,$C3,$E2,$BC,$50,$EF,$F9,$0F,$F5,$2F,$A1,$F1,$F9,$07
	.dc.b	$73,$AC,$7F,$D2,$B5,$C7,$9B,$FC,$E8,$D3,$FC,$ED,$3D,$1F,$E5,$91
	.dc.b	$FE,$79,$E0,$7D,$33,$E0,$BE,$27,$17,$F8,$A3,$D9,$BE,$B7,$F8,$EF
	.dc.b	$1F,$EB,$97,$B6,$DE,$AA,$46,$08,$7F,$9D,$30,$EF,$3F,$45,$B1,$F7
	.dc.b	$3D,$8C,$5B,$DF,$FA,$7C,$CF,$42,$9A,$89,$7F,$84,$8C,$BE,$68,$CB
	.dc.b	$E1,$AE,$8E,$4F,$C5,$E7,$F4,$BB,$1F,$41,$E0,$F6,$65,$7F,$B3,$E7
	.dc.b	$2D,$9C,$BF,$92,$9E,$51,$23,$44,$3F,$2F,$EF,$FE,$47,$D7,$7D,$37
	.dc.b	$F4,$29,$FC,$DF,$C0,$7C,$03,$51,$9E,$6F,$F6,$9D,$70,$AD,$1F,$A4
	.dc.b	$AD,$B0,$FF,$7C,$D7,$1E,$4F,$ED,$55,$C6,$7F,$57,$F0,$1D,$BD,$FF
	.dc.b	$9F,$EE,$C9,$E3,$19,$7F,$63,$19,$7F,$D5,$38,$7F,$B0,$9D,$90,$CC
	.dc.b	$FD,$AF,$53,$D7,$F0,$52,$7C,$AF,$BB,$C2,$91,$82,$1D,$9F,$C1,$FC
	.dc.b	$06,$A3,$07,$F0,$47,$E3,$F3,$48,$F8,$2F,$E2,$BB,$FE,$5F,$1D,$9F
	.dc.b	$E1,$71,$87,$E3,$F1,$7D,$02,$3F,$C4,$3E,$03,$F6,$FF,$49,$FC,$EF
	.dc.b	$93,$FD,$DD,$F4,$DF,$B2,$6F,$26,$BB,$66,$1F,$E1,$99,$CB,$67,$FD
	.dc.b	$BD,$5C,$5F,$F0,$1D,$F5,$6D,$FE,$F2,$3F,$A3,$57,$F8,$8D,$A9,$CF
	.dc.b	$8C,$BF,$F3,$EC,$F7,$3F,$8F,$AF,$81,$FE,$3E,$D8,$D4,$F1,$F8,$1C
	.dc.b	$9A,$A9,$3A,$AE,$AF,$FB,$89,$7A,$7D,$3C,$E9,$3B,$8E,$E6,$1E,$11
	.dc.b	$87,$A7,$A9,$E4,$F8,$09,$18,$21,$DD,$CB,$E1,$8F,$59,$8D,$BE,$AC
	.dc.b	$E2,$7E,$17,$63,$E8,$FE,$84,$7C,$4F,$80,$F8,$67,$ED,$63,$33,$BE
	.dc.b	$1F,$DB,$57,$EA,$A7,$F9,$8B,$FE,$3A,$9D,$3F,$EC,$D8,$CB,$F7,$BD
	.dc.b	$DF,$9F,$ED,$C8,$F9,$51,$F4,$8C,$3F,$2E,$30,$FD,$93,$EF,$7C,$DF
	.dc.b	$F5,$9F,$69,$F4,$CF,$91,$C3,$6B,$87,$32,$FF,$32,$6A,$A4,$EA,$32
	.dc.b	$FB,$8E,$E5,$9E,$4D,$8D,$73,$7E,$3F,$D1,$FC,$5F,$C6,$4A,$FC,$34
	.dc.b	$7E,$38,$C3,$F8,$78,$C3,$F5,$35,$70,$7E,$53,$E9,$A4,$E6,$3E,$A1
	.dc.b	$E5,$E0,$FB,$4D,$AE,$33,$F4,$63,$F5,$1E,$EF,$E7,$BE,$96,$4F,$C4
	.dc.b	$8F,$DD,$98,$7F,$35,$18,$74,$7C,$18,$E6,$D3,$EF,$70,$FE,$AB,$CB
	.dc.b	$91,$EB,$8F,$C5,$1E,$52,$3C,$0F,$C3,$67,$FB,$2F,$87,$91,$F0,$E3
	.dc.b	$F2,$47,$B4,$8F,$03,$F1,$D8,$7F,$86,$F9,$FF,$D4,$3E,$2D,$F8,$3F
	.dc.b	$F1,$25,$27,$FA,$4E,$AB,$FC,$F1,$F2,$49,$3D,$0F,$CF,$F0,$BE,$57
	.dc.b	$FC,$35,$6E,$9F,$F5,$A4,$E7,$DA,$FF,$14,$D6,$A7,$2F,$F8,$8D,$B1
	.dc.b	$A2,$5F,$F1,$09,$D2,$E2,$6B,$FC,$8E,$3F,$C6,$FE,$2E,$1F,$E2,$ED
	.dc.b	$7D,$D8,$7F,$5E,$9E,$15,$CB,$C3,$FC,$44,$7F,$2A,$8C,$57,$EF,$1A
	.dc.b	$FE,$FB,$E9,$5C,$5F,$E3,$6F,$A5,$F2,$C9,$B3,$A8,$FD,$B3,$5E,$53
	.dc.b	$FE,$34,$8C,$47,$F7,$F6,$BF,$EC,$27,$86,$DF,$E3,$C6,$E5,$9A,$F1
	.dc.b	$65,$FE,$48,$9C,$DF,$FB,$95,$A9,$B8,$7C,$98,$78,$0F,$A4,$FF,$28
	.dc.b	$FD,$2B,$D1,$A7,$F9,$52,$75,$5F,$C4,$5A,$FF,$22,$6B,$FE,$19,$6B
	.dc.b	$FE,$DC,$6A,$C5,$F4,$3F,$E1,$38,$FF,$38,$F0,$F4,$DB,$F1,$91,$8C
	.dc.b	$BF,$C8,$D5,$BB,$E3,$3F,$8E,$98,$FF,$5D,$8C,$B6,$7F,$CC,$71,$88
	.dc.b	$FF,$17,$35,$FF,$78,$B5,$1E,$78,$FF,$2F,$3F,$8C,$3C,$FF,$47,$F9
	.dc.b	$E3,$CB,$EE,$48,$FC,$D4,$65,$F1,$46,$5F,$F3,$14,$F4,$C3,$DF,$7F
	.dc.b	$12,$BF,$53,$E0,$6E,$73,$5B,$E1,$F2,$DE,$F0,$FD,$45,$FC,$6F,$73
	.dc.b	$F6,$3E,$1C,$9F,$85,$19,$7F,$5D,$19,$7C,$71,$96,$9C,$0F,$DA,$7D
	.dc.b	$1F,$D2,$F4,$3F,$D2,$7E,$1C,$9F,$9D,$1F,$5B,$D0,$FF,$4B,$FD,$34
	.dc.b	$9F,$A5,$1F,$E9,$EF,$8A,$61,$FC,$4E,$7F,$09,$DB,$F4,$14,$EB,$83
	.dc.b	$DF,$39,$8C,$4F,$03,$EB,$B8,$DB,$FD,$5D,$1F,$6B,$E1,$B8,$BF,$E6
	.dc.b	$D6,$BF,$EE,$73,$91,$63,$FD,$7A,$7E,$5B,$DB,$FC,$9E,$1F,$94,$F0
	.dc.b	$1E,$07,$F2,$9E,$3F,$C7,$56,$9F,$E4,$71,$9C,$3E,$FF,$B7,$97,$F8
	.dc.b	$CF,$D2,$DC,$BF,$47,$39,$C7,$AC,$E7,$99,$EC,$53,$5D,$6F,$2B,$1A
	.dc.b	$7E,$7E,$73,$F4,$B8,$D3,$F3,$19,$7F,$39,$8C,$6F,$EB,$C5,$18,$C3
	.dc.b	$F9,$BB,$E5,$FA,$5F,$F1,$DC,$7E,$B5,$B8,$74,$A7,$C8,$6F,$D6,$97
	.dc.b	$B7,$CF,$18,$AF,$F6,$15,$F2,$C7,$C5,$F3,$B9,$6E,$1C,$4F,$E8,$91
	.dc.b	$97,$FA,$5B,$E5,$F4,$3F,$C5,$BC,$4B,$FF,$04,$D7,$F8,$DA,$31,$3A
	.dc.b	$9C,$E1,$FC,$09,$B8,$7C,$3F,$86,$E9,$87,$F5,$36,$D7,$BD,$B3,$B9
	.dc.b	$FF,$1A,$7D,$27,$FB,$A5,$FF,$14,$36,$BF,$C8,$ED,$7F,$D7,$CD,$7F
	.dc.b	$40,$EE,$7F,$90,$CB,$E9,$70,$EB,$C3,$FC,$53,$97,$F1,$96,$FA,$4F
	.dc.b	$C3,$CF,$F9,$1D,$B1,$FC,$67,$CD,$7D,$2B,$E3,$78,$D6,$B2,$FE,$63
	.dc.b	$7C,$F9,$B2,$7E,$67,$0E,$37,$8C,$7B,$4F,$F9,$7A,$B8,$47,$F9,$27
	.dc.b	$0F,$0C,$7F,$B7,$5A,$FE,$FD,$F0,$47,$FA,$97,$EB,$BF,$DE,$2F,$43
	.dc.b	$FD,$31,$8A,$33,$5E,$BF,$47,$D5,$5F,$DE,$5B,$1F,$EE,$BC,$BF,$C1
	.dc.b	$39,$7F,$9C,$9B,$5B,$8F,$D8,$76,$F4,$C7,$FD,$DB,$87,$27,$F8,$04
	.dc.b	$66,$BF,$C3,$7F,$41,$FB,$23,$F6,$7F,$4B,$63,$FC,$35,$17,$2F,$D3
	.dc.b	$4D,$BE,$BE,$57,$8F,$1F,$F3,$C6,$1E,$98,$C1,$1F,$83,$F5,$1F,$1B
	.dc.b	$0F,$F1,$25,$70,$7F,$AA,$5A,$75,$A7,$F8,$A3,$BB,$B1,$30,$F8,$CD
	.dc.b	$BF,$DD,$7C,$87,$DD,$C7,$F8,$65,$7F,$8C,$E3,$1D,$5E,$72,$47,$CF
	.dc.b	$6D,$7F,$03,$FA,$4D,$CF,$F8,$ED,$FC,$AF,$D2,$1F,$C3,$36,$3F,$D1
	.dc.b	$CD,$9F,$F5,$6B,$5F,$E8,$81,$D6,$9F,$CC,$FE,$89,$F0,$5D,$B0,$FE
	.dc.b	$55,$18,$C3,$FD,$51,$7C,$1F,$F2,$BC,$67,$0F,$EF,$87,$5A,$E3,$7F
	.dc.b	$0F,$87,$EB,$12,$FF,$1C,$DA,$FE,$5B,$6F,$8D,$E1,$5F,$EE,$A8,$C3
	.dc.b	$7A,$7C,$BF,$49,$E6,$6C,$64,$E1,$19,$69,$FE,$C1,$A7,$D2,$7E,$BD
	.dc.b	$B1,$FE,$B7,$6E,$1F,$ED,$66,$BF,$29,$CE,$BF,$27,$E6,$3F,$84,$C6
	.dc.b	$2B,$E2,$C6,$2F,$E5,$B6,$3D,$8F,$CE,$F8,$1F,$9F,$F2,$B6,$7E,$03
	.dc.b	$71,$F8,$AD,$C7,$EC,$3F,$C2,$FE,$8F,$E9,$4F,$FB,$A5,$AF,$F8,$CB
	.dc.b	$99,$18,$8F,$F4,$23,$63,$3C,$1F,$80,$F9,$0E,$8D,$C7,$E9,$7C,$AF
	.dc.b	$E1,$3D,$7F,$C2,$B6,$32,$7D,$78,$CB,$4C,$E3,$F7,$A3,$2D,$8F,$BD
	.dc.b	$EB,$7E,$43,$D2,$FC,$8B,$63,$27,$E0,$E3,$2D,$3F,$0D,$E9,$7E,$63
	.dc.b	$DE,$FC,$CB,$63,$27,$E2,$E3,$2D,$3F,$09,$C1,$F8,$32,$3F,$83,$6E
	.dc.b	$1F,$96,$F9,$3F,$BF,$1F,$A6,$6C,$FE,$5D,$B3,$B8,$FF,$31,$F0,$FF
	.dc.b	$9D,$37,$FA,$7E,$DF,$D9,$3F,$5F,$E3,$DA,$D9,$F6,$78,$9C,$1F,$B4
	.dc.b	$F1,$3F,$00,$D9,$DE,$5F,$73,$87,$F0,$86,$5F,$EB,$16,$8F,$DE,$FC
	.dc.b	$97,$E2,$77,$B7,$CB,$4F,$C3,$C6,$63,$16,$DB,$18,$F4,$7F,$24,$49
	.dc.b	$DD,$59,$F0,$E7,$F9,$73,$FA,$3C,$D3,$AB,$FD,$2B,$E4,$B1,$3F,$E3
	.dc.b	$08,$46,$37,$FE,$96,$DA,$FF,$23,$BF,$DB,$7E,$8F,$71,$EB,$6D,$F3
	.dc.b	$F9,$3D,$54,$AF,$F7,$0F,$A9,$6B,$C0,$DC,$3F,$AC,$B6,$7F,$CA,$AF
	.dc.b	$F0,$0E,$89,$1F,$EE,$AF,$92,$B7,$FC,$93,$B5,$5F,$D9,$BE,$13,$38
	.dc.b	$FF,$64,$6E,$1F,$E6,$06,$DF,$FB,$D9,$FF,$2F,$7C,$67,$F9,$A2,$7F
	.dc.b	$D6,$CD,$9F,$8D,$F2,$5A,$DE,$57,$FB,$F3,$D6,$B5,$F2,$DB,$8F,$F7
	.dc.b	$46,$CF,$F9,$C1,$FF,$36,$7D,$17,$F9,$36,$3F,$D4,$ED,$A7,$B9,$FE
	.dc.b	$7A,$FA,$3D,$F0,$4B,$FC,$98,$FA,$3F,$C4,$3F,$EC,$DC,$39,$39,$E6
	.dc.b	$E1,$8F,$FB,$16,$30,$6E,$FF,$C6,$F7,$CA,$7F,$A7,$FC,$E7,$F9,$A2
	.dc.b	$BF,$AE,$C7,$FB,$42,$30,$E1,$AF,$F6,$26,$BF,$EA,$77,$7A,$7F,$85
	.dc.b	$BE,$73,$FC,$87,$18,$AF,$F4,$9C,$7F,$A3,$7E,$49,$F0,$D7,$DF,$C3
	.dc.b	$6B,$F6,$7B,$AB,$71,$87,$FD,$F4,$D9,$E1,$97,$E7,$FE,$43,$FC,$56
	.dc.b	$7F,$8B,$3E,$8E,$D7,$4E,$34,$E2,$BF,$4D,$57,$5F,$F2,$8F,$CE,$38
	.dc.b	$3F,$BE,$D3,$E2,$1F,$43,$FE,$C7,$AD,$79,$9F,$EB,$F8,$FF,$7C,$FD
	.dc.b	$0B,$E8,$DD,$8D,$F5,$97,$FB,$FB,$1E,$2F,$5C,$3A,$67,$E0,$CF,$BC
	.dc.b	$EE,$31,$3D,$B1,$EB,$87,$4A,$61,$D3,$0E,$31,$8C,$38,$BC,$F3,$F1
	.dc.b	$E7,$E7,$CF,$CD,$3E,$33,$45,$39,$4F,$7C,$3B,$67,$E3,$AE,$38,$F8
	.dc.b	$27,$37,$CA,$31,$8F,$8B,$1F,$06,$5C,$E3,$CB,$5D,$8F,$CA,$DF,$C6
	.dc.b	$7F,$C6,$6D,$7E,$3E,$5F,$9A,$7F,$DF,$3E,$DB,$E8,$F1,$F1,$BF,$ED
	.dc.b	$76,$C6,$67,$F4,$33,$97,$BF,$84,$E7,$E9,$8A,$31,$3A,$9C,$C6,$23
	.dc.b	$CA,$EB,$5F,$84,$DD,$FF,$10,$D9,$6C,$FA,$37,$BA,$9F,$9D,$5B,$7E
	.dc.b	$EB,$72,$F2,$65,$F6,$AF,$8D,$70,$88,$FD,$99,$DB,$F7,$70,$F0,$61
	.dc.b	$DE,$35,$87,$52,$5F,$59,$D6,$5E,$C8,$EA,$DA,$E9,$8F,$89,$FC,$57
	.dc.b	$4B,$A9,$EA,$F7,$E6,$FF,$4E,$F8,$77,$97,$E0,$5B,$A9,$39,$C7,$F9
	.dc.b	$4E,$3C,$9F,$E0,$ED,$7F,$B0,$46,$EB,$93,$67,$58,$7E,$7A,$5F,$47
	.dc.b	$FE,$D3,$BF,$E2,$2F,$F1,$66,$F6,$73,$8D,$42,$B9,$46,$1B,$3B,$38
	.dc.b	$7C,$1F,$A6,$9D,$FF,$16,$BF,$D7,$BF,$39,$E7,$D0,$EE,$75,$AF,$F2
	.dc.b	$FE,$2D,$4C,$C3,$FC,$32,$74,$8F,$8D,$C4,$FA,$37,$B9,$F2,$70,$FF
	.dc.b	$11,$FC,$5D,$F5,$0F,$0A,$DE,$3F,$E2,$4C,$39,$61,$BA,$93,$9A,$DB
	.dc.b	$F0,$F0,$FC,$C5,$F0,$8F,$EB,$78,$76,$6D,$F7,$C3,$C1,$ED,$FD,$5F
	.dc.b	$32,$57,$58,$CB,$F7,$A3,$2F,$73,$F6,$D9,$27,$71,$9E,$3F,$C8,$57
	.dc.b	$0E,$0F,$E3,$B9,$FC,$5B,$FA,$9D,$6F,$9D,$FC,$37,$AD,$32,$F9,$B3
	.dc.b	$96,$CE,$5E,$DC,$FF,$E4,$57,$24,$3F,$D7,$B3,$C5,$B3,$F9,$3F,$73
	.dc.b	$F3,$9E,$B3,$E8,$FF,$D9,$FA,$7E,$57,$17,$BB,$F9,$93,$73,$AE,$4F
	.dc.b	$F4,$B3,$72,$FD,$2C,$ED,$FF,$2A,$46,$70,$ED,$1C,$88,$7E,$AB,$DD
	.dc.b	$EB,$EA,$7E,$9F,$D8,$93,$CA,$32,$FE,$C2,$32,$FD,$3A,$F5,$4E,$A3
	.dc.b	$17,$FE,$D5,$79,$1F,$B8,$F4,$7D,$7F,$47,$23,$CE,$FB,$BD,$E9,$18
	.dc.b	$21,$D1,$FD,$E7,$BB,$A8,$C1,$FC,$11,$F8,$FC,$D2,$3E,$1B,$FC,$0F
	.dc.b	$BF,$E5,$F2,$D9,$FE,$17,$18,$7E,$4F,$1F,$D0,$23,$FC,$43,$DD,$FD
	.dc.b	$E7,$D0,$7F,$41,$E5,$FE,$19,$F4,$7F,$DA,$AB,$F2,$4D,$8C,$4E,$F9
	.dc.b	$7F,$98,$36,$BF,$C8,$B5,$B9,$FE,$46,$D8,$FE,$89,$3B,$4F,$1E,$B4
	.dc.b	$ED,$B5,$E6,$EB,$FE,$49,$3F,$E8,$4B,$FF,$17,$AF,$77,$26,$BD,$D9
	.dc.b	$3A,$8C,$BC,$23,$9F,$A7,$D3,$CD,$93,$D2,$32,$F6,$8C,$BF,$12,$3C
	.dc.b	$7E,$A7,$93,$DD,$93,$E4,$8C,$BE,$68,$CB,$F3,$23,$2F,$CE,$8C,$ED
	.dc.b	$F4,$A7,$F5,$BF,$43,$F0,$F3,$7B,$DF,$08,$CE,$D3,$1F,$87,$19,$7F
	.dc.b	$D7,$95,$C7,$E8,$FE,$4F,$B7,$27,$E4,$C7,$CC,$30,$FE,$02,$30,$FF
	.dc.b	$36,$E2,$7C,$E7,$F5,$BF,$69,$F4,$6F,$8E,$C3,$6F,$85,$7E,$38,$CC
	.dc.b	$C8,$C9,$0F,$E4,$95,$D2,$35,$1F,$C9,$CF,$C3,$CF,$E0,$BC,$FF,$C1
	.dc.b	$FC,$6C,$9F,$C1,$C6,$5F,$C2,$C6,$5F,$CD,$C7,$CC,$D7,$F1,$7F,$3B
	.dc.b	$27,$E3,$C6,$5F,$C6,$C6,$5F,$E0,$DB,$7D,$27,$E2,$7B,$BF,$94,$FA
	.dc.b	$39,$18,$21,$FA,$A7,$2F,$AF,$64,$FE,$C8,$CF,$27,$EC,$AB,$15,$F6
	.dc.b	$B8,$3F,$43,$E6,$48,$C1,$0F,$DB,$B9,$7C,$F1,$F8,$AD,$7F,$53,$EC
	.dc.b	$49,$FC,$2C,$65,$FD,$5C,$65,$F7,$C7,$E4,$AF,$FC,$35,$87,$FA,$77
	.dc.b	$C4,$BF,$04,$75,$7F,$90,$AC,$C4,$8C,$1F,$EA,$6F,$99,$FF,$0B,$3F
	.dc.b	$D7,$DB,$7F,$E1,$D6,$F3,$FF,$8B,$1F,$F0,$B3,$7D,$9F,$F1,$34,$6A
	.dc.b	$BF,$DA,$AD,$F0,$4F,$08,$79,$8D,$3F,$88,$78,$B6,$7F,$D5,$D1,$FD
	.dc.b	$70,$FE,$39,$3B,$F0,$FF,$C3,$D1,$27,$FC,$BD,$9F,$FA,$42,$B5,$1F
	.dc.b	$E1,$CF,$A3,$FF,$5B,$BF,$E1,$EF,$A3,$F1,$26,$C6,$A3,$F3,$CD,$AF
	.dc.b	$E6,$53,$C0,$DF,$17,$F8,$CE,$78,$57,$F9,$84,$38,$7F,$CC,$8D,$7F
	.dc.b	$CE,$8D,$77,$1C,$C8,$F9,$F1,$FF,$20,$78,$0F,$A2,$FF,$21,$7D,$1B
	.dc.b	$D1,$A6,$A2,$5F,$F8,$F2,$74,$D9,$CB,$FE,$21,$B7,$CF,$FF,$80,$78
	.dc.b	$7E,$BE,$1F,$70,$FF,$21,$C7,$F8,$16,$77,$59,$3F,$15,$3A,$61,$82
	.dc.b	$1D,$E7,$91,$FE,$40,$D7,$F1,$DF,$47,$F8,$A3,$BF,$B5,$2B,$F2,$D1
	.dc.b	$97,$FC,$99,$5C,$79,$7F,$C8,$ED,$8F,$C8,$CE,$DF,$05,$73,$7F,$3D
	.dc.b	$19,$BF,$F1,$DD,$71,$C7,$E1,$FC,$7F,$E7,$3C,$39,$87,$82,$3E,$19
	.dc.b	$87,$F5,$11,$87,$47,$EB,$5D,$5F,$69,$F9,$7E,$7F,$F9,$DF,$C1,$97
	.dc.b	$EC,$8F,$9F,$E7,$FF,$9E,$FE,$8E,$4F,$C0,$8F,$F4,$17,$C4,$3E,$8F
	.dc.b	$FD,$15,$3F,$EC,$FE,$DF,$F5,$94,$E7,$3F,$F3,$5C,$62,$B7,$AF,$F0
	.dc.b	$0A,$D5,$F1,$3F,$35,$F0,$FC,$1A,$64,$FF,$5C,$94,$FF,$67,$3B,$9F
	.dc.b	$C5,$FB,$6E,$3F,$DA,$9F,$EB,$4E,$E7,$83,$9C,$3F,$09,$57,$4F,$F5
	.dc.b	$65,$6A,$FE,$FF,$B7,$97,$97,$D1,$DC,$3F,$DD,$DC,$AA,$C7,$4F,$FB
	.dc.b	$06,$B5,$A7,$E7,$BD,$8C,$F8,$6F,$A3,$CE,$29,$CC,$61,$B9,$FE,$41
	.dc.b	$B9,$FE,$6D,$BA,$FE,$29,$B3,$FE,$9A,$FA,$27,$AB,$E0,$F9,$BF,$1C
	.dc.b	$62,$7F,$99,$46,$27,$51,$FE,$A9,$9D,$61,$FC,$D6,$B5,$3C,$0D,$57
	.dc.b	$09,$FD,$E5,$70,$C3,$F9,$D5,$F0,$9E,$57,$D7,$CF,$F8,$7F,$09,$E4
	.dc.b	$6C,$64,$FE,$8D,$19,$69,$99,$FE,$91,$19,$6C,$7F,$B3,$3E,$8B,$FD
	.dc.b	$92,$FE,$0A,$3F,$09,$F4,$58,$7F,$C1,$93,$71,$FC,$25,$2B,$0A,$A5
	.dc.b	$61,$E7,$E1,$7E,$0D,$46,$67,$94,$E7,$E8,$7E,$F4,$7F,$B8,$1B,$5F
	.dc.b	$E9,$1F,$A2,$7C,$66,$F8,$25,$FF,$3D,$AF,$15,$F0,$7F,$A0,$35,$F3
	.dc.b	$07,$0F,$09,$FC,$C6,$6F,$A2,$FF,$64,$5F,$8E,$73,$18,$AF,$F1,$85
	.dc.b	$F9,$AF,$F4,$D3,$97,$FD,$0D,$19,$BF,$EC,$73,$97,$E4,$79,$DF,$DB
	.dc.b	$F1,$FD,$2F,$56,$71,$FE,$E7,$E7,$FC,$0B,$FA,$DD,$4A,$D6,$1C,$67
	.dc.b	$31,$8A,$E1,$7F,$EE,$78,$E9,$7F,$E0,$0F,$9E,$FD,$79,$FB,$0F,$A2
	.dc.b	$B8,$7F,$84,$62,$DF,$F8,$46,$6E,$BF,$DE,$F1,$FF,$2E,$5F,$E2,$61
	.dc.b	$F4,$5E,$FB,$F4,$F9,$5F,$32,$31,$A7,$F8,$36,$31,$C1,$FC,$1E,$31
	.dc.b	$87,$3A,$DE,$3F,$E4,$28,$FF,$12,$7C,$7E,$E3,$18,$FF,$89,$63,$1E
	.dc.b	$C7,$F0,$A6,$D7,$03,$F3,$D8,$7F,$87,$7A,$7F,$87,$FD,$0F,$36,$CE
	.dc.b	$B1,$46,$31,$FF,$71,$46,$27,$FC,$6D,$5F,$9D,$FA,$23,$FC,$7B,$E8
	.dc.b	$9F,$01,$98,$C3,$6B,$ED,$CE,$9F,$D9,$60,$BE,$71,$AA,$E1,$2E,$0F
	.dc.b	$E6,$DC,$6C,$39,$4F,$FA,$0F,$0E,$2E,$F0,$FD,$B5,$6F,$0F,$95,$C7
	.dc.b	$E7,$F8,$DB,$5E,$46,$DF,$11,$7D,$E3,$39,$78,$B2,$47,$C1,$EA,$F8
	.dc.b	$5F,$45,$71,$F7,$B7,$1E,$98,$FA,$63,$18,$F3,$7F,$84,$60,$FA,$2F
	.dc.b	$99,$F2,$F8,$6E,$3F,$22,$74,$FC,$C9,$D3,$4D,$4F,$0C,$7F,$D5,$15
	.dc.b	$C1,$F1,$E6,$BF,$B5,$EE,$FD,$BF,$93,$B8,$FD,$86,$DF,$06,$E7,$F0
	.dc.b	$DF,$CE,$7D,$17,$D4,$8F,$8D,$F1,$F8,$E1,$FC,$33,$70,$FC,$1B,$70
	.dc.b	$F1,$BF,$8B,$F9,$6F,$C6,$7A,$F6,$BF,$12,$DA,$F4,$B7,$0F,$A8,$D9
	.dc.b	$FC,$0E,$A8,$C5,$7D,$C8,$C4,$FE,$53,$D6,$FC,$D7,$A3,$F9,$B6,$C6
	.dc.b	$4F,$CC,$C6,$5A,$7E,$73,$D1,$FD,$17,$B9,$FA,$36,$C6,$4F,$D0,$C6
	.dc.b	$5A,$7D,$BE,$0F,$D5,$7C,$97,$E0,$9B,$19,$3F,$53,$19,$69,$9C,$BE
	.dc.b	$DC,$65,$B1,$FC,$0B,$81,$ED,$FD,$63,$F7,$1E,$15,$9D,$70,$7E,$EF
	.dc.b	$C5,$FB,$6D,$8E,$5C,$1F,$78,$C6,$3F,$A7,$F0,$19,$7C,$C8,$FC,$6F
	.dc.b	$C8,$FE,$13,$6F,$E0,$CD,$C3,$C9,$1F,$88,$D7,$15,$CE,$D1,$9A,$F5
	.dc.b	$3A,$E3,$FF,$57,$4E,$A7,$DD,$A7,$E4,$8F,$E9,$AD,$32,$7F,$48,$F9
	.dc.b	$0C,$61,$FD,$1D,$B7,$FE,$40,$6C,$FF,$91,$DF,$E7,$1F,$44,$69,$FD
	.dc.b	$6B,$E4,$AC,$FF,$30,$9D,$3F,$D5,$BD,$3C,$E1,$FD,$55,$B8,$7F,$93
	.dc.b	$9B,$3F,$E5,$57,$F8,$1F,$44,$8F,$EB,$FF,$25,$6F,$F7,$73,$FD,$AB
	.dc.b	$E1,$31,$87,$F9,$61,$B7,$FE,$5F,$6D,$7F,$1C,$D1,$1F,$E5,$4F,$89
	.dc.b	$FE,$DD,$5F,$DB,$3D,$5C,$57,$F7,$66,$C6,$4F,$F1,$57,$B4,$AE,$38
	.dc.b	$7F,$71,$AD,$B6,$B7,$3A,$C7,$FC,$DB,$5F,$E6,$4F,$A1,$FF,$23,$57
	.dc.b	$F8,$B4,$7D,$0F,$DD,$C3,$FD,$8D,$3A,$7F,$CF,$91,$9B,$FF,$3D,$62
	.dc.b	$AD,$5F,$2E,$3E,$DB,$FF,$9D,$E3,$07,$AA,$77,$87,$38,$FF,$1F,$5F
	.dc.b	$3A,$E1,$39,$C7,$FC,$85,$F3,$4F,$A3,$FF,$0D,$FC,$CF,$F9,$06,$31
	.dc.b	$8F,$F8,$5A,$31,$3A,$8F,$EC,$33,$AA,$FF,$0C,$57,$FA,$23,$E3,$9F
	.dc.b	$0D,$73,$C3,$70,$FF,$12,$CE,$9B,$3A,$69,$FE,$26,$F8,$CF,$F4,$9F
	.dc.b	$EE,$C9,$FF,$18,$C6,$1B,$3F,$E3,$09,$D3,$C4,$C4,$7F,$97,$7E,$6F
	.dc.b	$FD,$17,$E4,$7F,$AC,$3E,$83,$FC,$9F,$18,$D1,$F1,$09,$FF,$34,$7D
	.dc.b	$03,$E8,$BF,$23,$97,$1B,$FF,$36,$61,$C7,$0E,$11,$8E,$3F,$F0,$6B
	.dc.b	$D3,$1E,$F1,$89,$DE,$3D,$7A,$BF,$D7,$91,$CA,$D8,$75,$C3,$8E,$7E
	.dc.b	$19,$CE,$3C,$E7,$CF,$9F,$9A,$7C,$07,$8E,$B8,$5F,$38,$C1,$AC,$BC
	.dc.b	$59,$EB,$1E,$93,$AC,$BC,$97,$C6,$F7,$3D,$8F,$1E,$1C,$4F,$F4,$03
	.dc.b	$83,$FD,$05,$3F,$3D,$FF,$3F,$99,$2F,$C5,$BD,$B7,$D1,$72,$8C,$36
	.dc.b	$33,$3F,$E8,$D9,$CB,$DB,$BF,$F1,$0D,$E0,$F1,$E0,$9C,$CF,$8A,$31
	.dc.b	$3A,$9E,$39,$FC,$06,$EF,$F8,$86,$DB,$47,$D1,$3D,$DF,$F4,$DE,$1C
	.dc.b	$71,$FA,$38,$70,$C3,$EE,$BC,$B0,$E3,$7C,$31,$ED,$7C,$09,$7F,$5F
	.dc.b	$1E,$CE,$31,$EC,$FD,$3C,$BD,$6E,$F0,$E8,$FE,$63,$2F,$14,$EA,$B8
	.dc.b	$9F,$94,$BE,$B5,$F8,$AF,$5D,$D1,$FE,$31,$8F,$F4,$DF,$81,$FE,$B8
	.dc.b	$EF,$FF,$2F,$F8,$3F,$CB,$67,$FD,$47,$5B,$7F,$C9,$6D,$9D,$D7,$1E
	.dc.b	$5F,$F3,$0D,$F1,$BD,$FA,$99,$7F,$CB,$BF,$45,$FE,$CE,$7F,$D5,$EF
	.dc.b	$F1,$06,$F6,$EF,$34,$EE,$7F,$CB,$6D,$CB,$FC,$EA,$F1,$8C,$D7,$18
	.dc.b	$D6,$3F,$ED,$4E,$EF,$87,$7E,$EF,$9A,$ED,$99,$F3,$62,$AD,$DF,$1E
	.dc.b	$2D,$D4,$AF,$E5,$71,$ED,$E3,$7D,$13,$DB,$F8,$18,$3D,$07,$C5,$61
	.dc.b	$B8,$FF,$5C,$BE,$0D,$BD,$54,$8C,$10,$FF,$44,$5F,$58,$D4,$7F,$A2
	.dc.b	$F0,$F1,$D7,$96,$DE,$CF,$CD,$E4,$79,$54,$D4,$4C,$3F,$03,$19,$7C
	.dc.b	$D1,$97,$F0,$71,$97,$C1,$5D,$9F,$F5,$7F,$43,$E6,$7C,$1E,$9C,$9C
	.dc.b	$C7,$D0,$9D,$47,$E4,$27,$4F,$FA,$94,$FE,$57,$3B,$3F,$8A,$B8,$3F
	.dc.b	$82,$B8,$9E,$71,$FC,$5B,$D3,$7D,$17,$F1,$8E,$5F,$F5,$EE,$8F,$67
	.dc.b	$E2,$1F,$97,$AD,$DD,$BE,$1C,$3F,$EB,$EB,$E1,$3F,$A6,$BD,$30,$FF
	.dc.b	$71,$D7,$05,$7D,$1F,$D6,$FB,$3D,$7D,$CF,$A9,$EB,$49,$E3,$19,$7F
	.dc.b	$67,$19,$7F,$CC,$57,$FE,$C4,$76,$43,$33,$F7,$FD,$1F,$57,$D1,$49
	.dc.b	$F2,$BE,$DF,$6E,$46,$08,$76,$7F,$11,$ED,$6A,$30,$7F,$06,$7E,$37
	.dc.b	$24,$8F,$82,$FE,$3F,$B7,$E5,$72,$59,$FE,$1B,$18,$7E,$3F,$B9,$F3
	.dc.b	$C8,$FF,$12,$F6,$FF,$6B,$F3,$9F,$D0,$39,$7F,$DE,$BF,$42,$E6,$FE
	.dc.b	$10,$DB,$FF,$7F,$D6,$E5,$12,$30,$FF,$BC,$27,$83,$FE,$E8,$6D,$62
	.dc.b	$77,$27,$8F,$88,$AD,$DF,$68,$C3,$E8,$ED,$F2,$FB,$9E,$6F,$72,$63
	.dc.b	$E4,$CB,$D7,$3A,$8C,$5F,$3C,$57,$C2,$7C,$D5,$F0,$3D,$2F,$83,$E6
	.dc.b	$49,$E2,$7B,$63,$4E,$33,$F4,$4E,$E3,$E0,$4F,$CA,$F5,$3E,$5F,$D1
	.dc.b	$49,$E1,$97,$F9,$E9,$B3,$97,$E4,$4F,$17,$27,$07,$5A,$7D,$89,$FD
	.dc.b	$77,$CF,$FC,$8A,$FA,$5E,$EF,$DA,$33,$B4,$8F,$0C,$65,$F6,$F7,$2B
	.dc.b	$96,$BF,$4F,$E8,$BF,$07,$ED,$4B,$F4,$C7,$E1,$8C,$3F,$7E,$30,$FE
	.dc.b	$23,$E8,$7F,$AC,$7B,$2F,$A2,$7C,$6E,$1B,$5B,$20,$DF,$F2,$55,$C6
	.dc.b	$B8,$1F,$9A,$B9,$5F,$5A,$74,$DB,$EA,$FF,$21,$E7,$57,$E7,$BC,$FF
	.dc.b	$CF,$FC,$5C,$AF,$AB,$19,$7E,$BC,$65,$FC,$E4,$7D,$0E,$0F,$D3,$FC
	.dc.b	$EC,$9F,$97,$C0,$9D,$47,$E0,$27,$4F,$FB,$C6,$74,$FE,$BB,$DC,$FD
	.dc.b	$8F,$CD,$C8,$C1,$0F,$D6,$39,$7D,$91,$FB,$C3,$1B,$FE,$F2,$71,$3E
	.dc.b	$4E,$0F,$DE,$F9,$B2,$30,$43,$F7,$0E,$5F,$34,$7E,$3B,$4F,$E0,$BE
	.dc.b	$C4,$9F,$7C,$65,$FC,$EC,$65,$F7,$47,$E5,$AF,$FC,$35,$83,$E0,$5E
	.dc.b	$36,$1B,$7F,$2A,$B8,$3F,$B2,$AE,$15,$B3,$31,$F8,$7E,$8F,$E1,$1F
	.dc.b	$35,$FE,$1B,$AF,$E3,$5D,$08,$C4,$EB,$D9,$FF,$13,$47,$F8,$7A,$D5
	.dc.b	$B9,$C9,$23,$F8,$04,$EF,$A9,$A7,$F1,$AF,$16,$D7,$F0,$56,$BE,$53
	.dc.b	$FA,$CC,$7F,$28,$9E,$3E,$1F,$F8,$AA,$24,$FE,$D1,$FE,$4F,$F4,$2E
	.dc.b	$1F,$F1,$6F,$D1,$79,$E8,$D9,$D4,$7F,$18,$0E,$1F,$F0,$14,$71,$38
	.dc.b	$4F,$F8,$E2,$75,$B7,$F3,$A9,$D7,$CA,$66,$A5,$FF,$A7,$2F,$A3,$FE
	.dc.b	$9A,$6B,$FD,$18,$F0,$C7,$F8,$3B,$C0,$7D,$0F,$F9,$1B,$E8,$9E,$95
	.dc.b	$3F,$C9,$33,$AB,$F5,$35,$FF,$0D,$4F,$07,$DC,$D7,$FC,$3D,$1B,$CF
	.dc.b	$FA,$D6,$AF,$9E,$FF,$05,$71,$FE,$0F,$2F,$C0,$9F,$E5,$18,$FF,$07
	.dc.b	$4E,$F0,$C9,$F8,$09,$D3,$0C,$10,$FF,$24,$4F,$23,$FC,$A6,$FE,$06
	.dc.b	$BF,$25,$F4,$5F,$90,$3C,$1E,$DC,$9F,$C6,$C6,$5F,$EA,$38,$7A,$67
	.dc.b	$51,$CF,$91,$5B,$8F,$0C,$EB,$2F,$D0,$D6,$A7,$A9,$D3,$1F,$2C,$FC
	.dc.b	$65,$F1,$FF,$A4,$F1,$A5,$6E,$3E,$31,$87,$C5,$18,$74,$7E,$99,$D5
	.dc.b	$75,$9F,$9D,$E8,$7F,$A2,$3C,$19,$7E,$38,$FA,$5E,$87,$FA,$33,$E8
	.dc.b	$A4,$F9,$E3,$FD,$1B,$F1,$0F,$A2,$FE,$8D,$2E,$0F,$F4,$D4,$E7,$3F
	.dc.b	$F0,$BC,$62,$B6,$D8,$CF,$6F,$FB,$CA,$DF,$0F,$87,$FD,$EA,$D7,$FD
	.dc.b	$B6,$77,$8E,$11,$BA,$FC,$77,$B9,$FE,$29,$BF,$F6,$1F,$85,$7C,$1F
	.dc.b	$1F,$17,$F9,$9D,$FE,$0D,$19,$CB,$D5,$2F,$73,$1C,$EF,$09,$F4,$5B
	.dc.b	$D3,$F3,$95,$A9,$CE,$5F,$77,$D9,$E3,$D0,$FA,$2E,$19,$A7,$31,$86
	.dc.b	$D7,$13,$F8,$FC,$65,$A6,$74,$FC,$94,$78,$FE,$89,$EB,$78,$3E,$6F
	.dc.b	$0D,$BF,$CD,$CE,$9F,$E5,$73,$A6,$9A,$9E,$17,$FE,$B3,$AE,$0F,$E8
	.dc.b	$6B,$83,$6B,$81,$C9,$E1,$E7,$FC,$2F,$85,$F3,$36,$7C,$6D,$9D,$C7
	.dc.b	$F3,$E9,$DB,$4D,$96,$3F,$D5,$9F,$43,$FE,$D7,$8F,$F2,$38,$FA,$2C
	.dc.b	$4E,$BC,$1F,$BD,$39,$7F,$01,$0A,$DF,$37,$F9,$C6,$B4,$7F,$3B,$8C
	.dc.b	$BF,$E7,$23,$33,$FE,$BD,$FA,$1F,$B9,$1F,$9B,$84,$61,$FC,$F3,$5F
	.dc.b	$E6,$C6,$B7,$FF,$3C,$BF,$CE,$63,$85,$FE,$6F,$E8,$7F,$DA,$57,$FD
	.dc.b	$36,$72,$FF,$9C,$63,$35,$E3,$BF,$CE,$5A,$72,$FF,$A1,$AF,$8D,$7F
	.dc.b	$8F,$E7,$37,$E1,$BE,$31,$89,$FF,$48,$D7,$0B,$E9,$E9,$7E,$A7,$CF
	.dc.b	$FE,$E7,$E6,$FE,$8E,$31,$E7,$A3,$19,$FF,$6D,$6E,$3F,$DD,$6F,$8C
	.dc.b	$FF,$79,$BF,$D4,$FC,$EC,$8F,$F0,$0F,$D0,$DB,$FE,$65,$37,$2F,$E5
	.dc.b	$F3,$6B,$FD,$4F,$87,$F8,$1B,$C9,$F8,$7F,$43,$F0,$23,$13,$FE,$1D
	.dc.b	$8C,$69,$FE,$66,$8C,$6F,$EA,$BE,$2F,$B3,$D6,$FD,$EF,$C7,$5A,$FF
	.dc.b	$0A,$B7,$5F,$E8,$CD,$D3,$FA,$1B,$F1,$39,$FF,$82,$FD,$0D,$BF,$F0
	.dc.b	$EB,$6F,$FA,$73,$6F,$FA,$5B,$F9,$8F,$A1,$3F,$C5,$BE,$7D,$F0,$19
	.dc.b	$8C,$37,$0F,$F5,$44,$E9,$FF,$0B,$57,$27,$33,$BC,$3F,$B4,$56,$DF
	.dc.b	$F1,$50,$AE,$0E,$A7,$EF,$F1,$78,$FE,$2E,$E1,$E0,$6D,$74,$6E,$5F
	.dc.b	$E4,$6C,$51,$EB,$E9,$F6,$7D,$0D,$C3,$D2,$DC,$3C,$AD,$C3,$93,$FB
	.dc.b	$CB,$7D,$0F,$C5,$F9,$6C,$37,$0F,$F1,$B4,$E9,$F8,$B3,$A6,$9A,$9E
	.dc.b	$18,$7C,$2A,$E0,$FC,$6A,$FA,$5E,$EF,$D3,$F9,$3B,$87,$D0,$6D,$7C
	.dc.b	$C6,$E5,$F0,$5F,$D5,$7D,$0F,$CF,$8F,$89,$F1,$D8,$E0,$FB,$CD,$C3
	.dc.b	$ED,$B7,$0F,$AE,$FE,$0F,$E5,$7F,$09,$EB,$DA,$FC,$0B,$6B,$EE,$B7
	.dc.b	$0F,$A4,$D9,$4F,$F9,$D3,$85,$39,$8C,$3F,$91,$F5,$BF,$25,$E9,$5B
	.dc.b	$FC,$7B,$6B,$F1,$6D,$C3,$F0,$0F,$E6,$7C,$EF,$CD,$7B,$D6,$FF,$2E
	.dc.b	$DB,$FC,$9B,$6F,$F1,$0F,$DF,$E0,$FD,$17,$C8,$DB,$FC,$FB,$6F,$F3
	.dc.b	$6D,$BF,$C8,$36,$7D,$FE,$9B,$54,$E6,$FE,$13,$F5,$BD,$B9,$7F,$AC
	.dc.b	$EF,$B5,$FC,$6F,$8D,$C3,$FB,$5F,$4B,$EE,$B6,$FF,$E0,$0C,$FC,$C6
	.dc.b	$F1,$66,$F9,$3F,$BD,$F9,$0F,$C2,$F0,$5A,$E2,$E1,$E7,$A5,$BF,$13
	.dc.b	$F8,$C2,$47,$03,$67,$8B,$2F,$C8,$9F,$E1,$B6,$99,$3F,$D5,$9F,$1F
	.dc.b	$8A,$FF,$36,$4A,$31,$C1,$FE,$C0,$8C,$57,$1A,$DD,$7F,$47,$FA,$1C
	.dc.b	$C7,$F4,$16,$DF,$1F,$91,$D5,$4B,$FF,$5E,$7A,$77,$0F,$0B,$6F,$C0
	.dc.b	$D9,$FB,$78,$FF,$86,$25,$1F,$EA,$5E,$C7,$4F,$F7,$0C,$3F,$3F,$18
	.dc.b	$9F,$EF,$4D,$8F,$F2,$B3,$5F,$EF,$2E,$6B,$E0,$7C,$17,$F1,$1B,$FF
	.dc.b	$22,$36,$FF,$DA,$8D,$BF,$92,$FF,$92,$FE,$23,$FC,$9D,$3F,$ED,$9F
	.dc.b	$53,$19,$7F,$93,$5B,$19,$3F,$D2,$B3,$C2,$FE,$9C,$E9,$B3,$AA,$E4
	.dc.b	$FF,$A7,$3E,$87,$E2,$3F,$E1,$BF,$A1,$FC,$FB,$9D,$7F,$C4,$F1,$8B
	.dc.b	$FF,$75,$7B,$6B,$E3,$1F,$32,$75,$AF,$FB,$02,$73,$82,$72,$FF,$A1
	.dc.b	$23,$31,$8A,$FF,$45,$46,$38,$BF,$C6,$BF,$32,$FA,$1F,$F3,$77,$CC
	.dc.b	$61,$B8,$FF,$BB,$E7,$4F,$FC,$01,$9D,$34,$FE,$37,$97,$E0,$B1,$77
	.dc.b	$3E,$35,$F0,$7F,$E8,$07,$FD,$39,$7C,$0F,$E0,$AE,$9C,$57,$28,$DB
	.dc.b	$FE,$CF,$74,$DD,$BA,$57,$57,$FD,$11,$F1,$7F,$E8,$EF,$A1,$E3,$97
	.dc.b	$EA,$63,$13,$AC,$7B,$E4,$BD,$C7,$E7,$AB,$6F,$93,$17,$CD,$7F,$A1
	.dc.b	$3C,$8F,$E7,$7F,$0C,$F9,$EF,$E7,$B5,$AE,$5F,$F0,$4B,$FE,$B2,$8F
	.dc.b	$E8,$3F,$3E,$FA,$17,$5F,$C9,$C3,$95,$FF,$46,$CB,$B3,$C4,$C6,$3D
	.dc.b	$27,$71,$DF,$2F,$1E,$5D,$B1,$EB,$9E,$2B,$8D,$57,$81,$D4,$B7,$E2
	.dc.b	$CF,$CD,$9F,$97,$1F,$6E,$7E,$4A,$F0,$B6,$33,$9F,$95,$F1,$57,$1C
	.dc.b	$3C,$27,$7A,$E0,$E2,$FC,$19,$FA,$23,$58,$73,$78,$9F,$E7,$EA,$F8
	.dc.b	$AD,$8F,$9E,$DC,$BF,$D1,$1F,$16,$F6,$1F,$43,$86,$D7,$FB,$9E,$75
	.dc.b	$39,$3F,$D2,$18,$3E,$73,$FD,$E9,$5F,$EE,$D8,$C7,$1E,$A7,$33,$E5
	.dc.b	$6B,$F2,$1E,$3A,$7C,$86,$ED,$9A,$FA,$DB,$BE,$85,$ED,$FE,$33,$0E
	.dc.b	$15,$F6,$72,$F0,$E5,$F2,$F2,$E6,$B1,$E5,$87,$1C,$BB,$4E,$F0,$E5
	.dc.b	$7C,$2F,$73,$89,$8E,$CF,$C9,$46,$9C,$63,$E4,$6D,$FE,$7B,$2F,$4C
	.dc.b	$7E,$47,$97,$F5,$DC,$FF,$A6,$E7,$FE,$11,$18,$9F,$D9,$F3,$FF,$08
	.dc.b	$6F,$C1,$6F,$9F,$FC,$17,$39,$95,$F2,$8C,$37,$0E,$13,$FC,$52,$7F
	.dc.b	$D5,$0F,$6C,$B3,$8F,$F0,$1F,$92,$B4,$6C,$CA,$F2,$C6,$8C,$3F,$E7
	.dc.b	$AF,$C9,$1E,$97,$FB,$5E,$11,$87,$FD,$B4,$D7,$FD,$AE,$6A,$38,$46
	.dc.b	$72,$FE,$59,$39,$3F,$89,$46,$73,$F8,$7F,$27,$EE,$E0,$7D,$B3,$E0
	.dc.b	$7D,$7F,$35,$31,$FE,$5B,$5A,$9E,$15,$B9,$F1,$CE,$5B,$19,$7F,$DD
	.dc.b	$B1,$96,$C6,$72,$FE,$85,$39,$6C,$67,$1F,$B3,$7E,$0F,$82,$F9,$FF
	.dc.b	$27,$31,$F2,$DA,$75,$1E,$27,$E3,$BF,$9D,$8D,$39,$21,$F8,$4B,$FE
	.dc.b	$5F,$1A,$8C,$1C,$67,$EF,$F6,$7C,$2F,$A9,$FC,$0F,$C8,$4A,$F1,$C6
	.dc.b	$5F,$86,$F8,$B4,$F9,$ED,$C0,$D3,$84,$F8,$27,$6F,$E7,$BC,$7F,$B1
	.dc.b	$D3,$FC,$63,$6F,$CD,$67,$F9,$23,$3E,$04,$BF,$AB,$5F,$C7,$9F,$E9
	.dc.b	$F8,$23,$11,$FD,$39,$AF,$F0,$37,$B6,$B9,$AE,$93,$F1,$39,$FF,$B1
	.dc.b	$7C,$7F,$93,$E8,$7F,$47,$F5,$52,$BF,$13,$8A,$75,$1E,$47,$E2,$46
	.dc.b	$63,$04,$3E,$13,$96,$C7,$EC,$3E,$1B,$F6,$3F,$19,$27,$F8,$1B,$F1
	.dc.b	$BD,$39,$1F,$09,$EE,$47,$F5,$79,$7F,$7B,$E1,$DC,$78,$27,$F7,$B1
	.dc.b	$FB,$EF,$0E,$4F,$F1,$07,$F8,$0F,$D5,$7D,$4F,$81,$F5,$92,$7E,$8B
	.dc.b	$F9,$2C,$3F,$C1,$1F,$15,$FC,$20,$93,$FE,$0C,$FA,$5F,$F0,$1B,$FE
	.dc.b	$3E,$5F,$0D,$FC,$43,$D1,$FF,$33,$CF,$F8,$49,$AE,$E3,$F8,$CF,$90
	.dc.b	$AE,$33,$FC,$22,$31,$3F,$A4,$6C,$66,$76,$F8,$A3,$2F,$53,$FC,$5E
	.dc.b	$E3,$A3,$F9,$9D,$FF,$AB,$A7,$D5,$CB,$FC,$9A,$FF,$E0,$0C,$7B,$7E
	.dc.b	$23,$13,$23,$04,$3F,$19,$3F,$90,$7F,$C8,$AB,$DA,$E9,$F1,$52,$7E
	.dc.b	$44,$77,$30,$F2,$8C,$3F,$37,$E2,$BC,$9C,$D2,$30,$43,$C2,$E5,$F1
	.dc.b	$C7,$AC,$C6,$7F,$D4,$B0,$FC,$A3,$63,$F9,$67,$E1,$EF,$AA,$F8,$5F
	.dc.b	$19,$30,$FE,$57,$39,$6C,$F0,$23,$F8,$3F,$27,$E5,$70,$FC,$BF,$89
	.dc.b	$93,$F8,$78,$CB,$F8,$98,$CB,$F2,$A3,$FA,$DB,$FE,$52,$F9,$1F,$C5
	.dc.b	$7C,$3C,$8F,$F2,$B7,$A8,$9D,$B6,$3F,$C3,$71,$FD,$56,$31,$1F,$E1
	.dc.b	$B8,$C1,$C0,$8F,$E5,$7E,$37,$F0,$5D,$BF,$83,$F8,$A9,$3F,$9F,$8C
	.dc.b	$BF,$A1,$8C,$BF,$82,$8F,$97,$F3,$3F,$8B,$C6,$4F,$CA,$8C,$B4,$87
	.dc.b	$E2,$AF,$D7,$D1,$F9,$3F,$8E,$93,$F0,$23,$2F,$F8,$E1,$F7,$11,$C1
	.dc.b	$B2,$1F,$96,$ED,$FF,$03,$FB,$2F,$C9,$1E,$37,$E8,$7E,$2E,$5F,$FA
	.dc.b	$5A,$B5,$3C,$2B,$73,$96,$BF,$D4,$27,$6F,$F0,$5F,$83,$FF,$0A,$FC
	.dc.b	$9B,$F5,$3F,$F1,$3C,$FE,$A3,$E4,$3F,$60,$6A,$33,$D9,$FE,$AF,$AE
	.dc.b	$1E,$3E,$49,$7F,$C4,$7E,$27,$AF,$0F,$F8,$4E,$8D,$24,$FF,$6C,$8C
	.dc.b	$BC,$08,$7F,$1B,$F8,$FF,$E0,$44,$70,$7F,$AA,$BF,$76,$FF,$0B,$E3
	.dc.b	$FA,$B9,$D4,$FF,$02,$7F,$C5,$F3,$AF,$41,$2B,$F2,$F8,$3E,$3C,$4D
	.dc.b	$24,$FF,$8A,$27,$6F,$FB,$21,$AF,$F7,$E6,$E3,$FE,$4E,$C1,$F1,$DF
	.dc.b	$AA,$FC,$21,$A7,$F8,$DD,$B8,$7F,$83,$9A,$FF,$8D,$9A,$FF,$3C,$73
	.dc.b	$7F,$E3,$66,$E7,$FE,$C2,$6C,$FF,$A4,$3C,$87,$C3,$78,$70,$93,$FE
	.dc.b	$CD,$AD,$F1,$AF,$73,$9A,$DB,$FE,$1A,$15,$C2,$77,$19,$FA,$27,$C5
	.dc.b	$7F,$94,$DB,$1F,$E2,$A6,$BF,$E5,$26,$BF,$D6,$1C,$E7,$FE,$52,$6E
	.dc.b	$7F,$ED,$E6,$CF,$DA,$FA,$3F,$F5,$0D,$7F,$90,$B0,$7B,$7F,$80,$F9
	.dc.b	$29,$7F,$E5,$29,$CB,$67,$FC,$59,$18,$8F,$F2,$7B,$5F,$F0,$FB,$5F
	.dc.b	$F7,$5B,$BA,$F0,$4F,$F9,$8E,$31,$8F,$E3,$BD,$27,$0F,$C6,$BF,$37
	.dc.b	$5F,$E5,$3E,$26,$65,$E5,$8E,$47,$90,$8F,$F9,$9B,$1F,$DE,$BE,$48
	.dc.b	$DB,$C4,$FD,$27,$C6,$61,$FF,$21,$4F,$F0,$4B,$F0,$C6,$8F,$89,$EA
	.dc.b	$7E,$9B,$EA,$64,$60,$87,$C4,$8F,$8E,$47,$C8,$FC,$12,$3F,$B5,$F9
	.dc.b	$2F,$F4,$36,$9F,$DE,$A3,$FC,$F6,$D7,$FC,$96,$D7,$F6,$BE,$EA,$31
	.dc.b	$1F,$EC,$D7,$FB,$D3,$FE,$88,$6B,$D3,$D5,$FF,$33,$46,$23,$FC,$47
	.dc.b	$F1,$AF,$83,$FF,$5C,$C3,$D9,$FF,$5E,$4E,$74,$FE,$0B,$18,$AD,$B6
	.dc.b	$33,$F1,$7F,$EA,$AF,$7E,$E1,$FE,$30,$6B,$FE,$E9,$3B,$46,$DA,$7A
	.dc.b	$FD,$AF,$F5,$7B,$F0,$32,$46,$23,$5C,$BE,$F8,$C6,$BF,$17,$C1,$B7
	.dc.b	$FC,$62,$BE,$24,$60,$CF,$AA,$8F,$F5,$9C,$FF,$B2,$1F,$E5,$5E,$B6
	.dc.b	$71,$FF,$44,$4E,$5B,$5C,$AB,$E8,$36,$3E,$DB,$C7,$4F,$E4,$1F,$53
	.dc.b	$6B,$D9,$36,$BE,$E7,$4A,$B1,$9F,$DF,$F1,$6E,$5F,$D2,$31,$FE,$7D
	.dc.b	$DF,$89,$91,$EA,$7A,$4F,$EB,$A7,$38,$FA,$A7,$E2,$92,$BF,$10,$D8
	.dc.b	$FF,$49,$62,$FA,$EE,$3A,$7F,$39,$7C,$7E,$5C,$C3,$FA,$87,$D4,$5C
	.dc.b	$3E,$3C,$FD,$22,$4F,$E5,$5B,$1D,$BE,$6F,$86,$1F,$D1,$5F,$89,$0F
	.dc.b	$9D,$79,$D8,$6E,$1F,$D3,$FE,$B9,$F2,$59,$A9,$3F,$ED,$6A,$DD,$FF
	.dc.b	$BF,$9A,$FF,$A1,$1A,$FF,$62,$1F,$95,$7D,$39,$FE,$CB,$97,$E0,$E7
	.dc.b	$3C,$7F,$67,$CB,$0F,$92,$F4,$E1,$C6,$BF,$A3,$36,$F9,$37,$0F,$EE
	.dc.b	$37,$CA,$F8,$E1,$27,$C2,$79,$6B,$86,$1E,$0B,$F0,$4E,$A3,$FB,$DC
	.dc.b	$E9,$F3,$C7,$FB,$E2,$3C,$91,$FE,$A4,$6B,$D6,$BF,$5D,$E7,$FF,$85
	.dc.b	$BE,$07,$F2,$CD,$BF,$F0,$77,$D4,$63,$95,$54,$E2,$E2,$7F,$C0,$CD
	.dc.b	$C3,$FD,$67,$7C,$DF,$F5,$AF,$3F,$F4,$4F,$A8,$FC,$CF,$39,$BF,$51
	.dc.b	$86,$63,$FE,$1F,$72,$D3,$FC,$25,$8B,$DF,$F3,$1C,$31,$FF,$0A,$46
	.dc.b	$67,$73,$E8,$23,$E8,$7D,$71,$EF,$3C,$BE,$ED,$9E,$F1,$E8,$F8,$0C
	.dc.b	$5F,$EA,$9F,$C3,$76,$BD,$EA,$7B,$BD,$EB,$87,$F9,$6E,$74,$FF,$8E
	.dc.b	$FE,$77,$38,$7E,$D1,$FD,$2F,$94,$F8,$4A,$7F,$00,$F3,$7F,$9D,$5F
	.dc.b	$E0,$EF,$FD,$7D,$3A,$8F,$F7,$1C,$7F,$90,$2B,$94,$62,$36,$D8,$CC
	.dc.b	$70,$6F,$57,$F2,$5E,$36,$0F,$23,$B5,$AB,$53,$9C,$3F,$2D,$9B,$F6
	.dc.b	$EF,$CD,$B7,$A7,$D5,$DF,$FC,$8A,$72,$7E,$83,$3E,$FD,$B9,$AF,$80
	.dc.b	$FC,$1F,$97,$F8,$5F,$2D,$72,$FD,$25,$F9,$A3,$1A,$F9,$67,$FC,$29
	.dc.b	$A2,$73,$AF,$A1,$F9,$FF,$0B,$2B,$9F,$B8,$F5,$F1,$8F,$F9,$0F,$89
	.dc.b	$7F,$D5,$E3,$FD,$DC,$D8,$CD,$FC,$D8,$C1,$FC,$E2,$DE,$B7,$4C,$BE
	.dc.b	$74,$62,$BD,$79,$BE,$5B,$F0,$1F,$2D,$F8,$16,$CE,$9B,$9F,$BC,$FF
	.dc.b	$05,$57,$19,$E7,$5F,$3F,$AA,$47,$6F,$52,$DF,$1E,$6F,$B1,$5F,$8F
	.dc.b	$F9,$6C,$E9,$F7,$9F,$C9,$7B,$7F,$93,$F9,$5B,$97,$A1,$B8,$F8,$F9
	.dc.b	$3F,$13,$18,$C3,$F0,$ED,$AF,$AA,$DA,$F4,$BF,$9F,$F9,$5F,$D0,$7C
	.dc.b	$AD,$AF,$8A,$DA,$ED,$BA,$31,$87,$E5,$1F,$D4,$7B,$DF,$A9,$F9,$5B
	.dc.b	$87,$CE,$6E,$1C,$9B,$7F,$9C,$7F,$BD,$7B,$6F,$07,$FD,$37,$7F,$CC
	.dc.b	$E3,$1C,$BF,$77,$DE,$DD,$6B,$1F,$A9,$5A,$BF,$DA,$B6,$3F,$AA,$1F
	.dc.b	$E3,$59,$E0,$FF,$57,$8D,$36,$3A,$63,$F8,$8F,$7B,$BE,$3F,$8D,$7F
	.dc.b	$83,$7C,$AC,$C7,$F1,$7F,$2B,$71,$FB,$DF,$9A,$B8,$7F,$C5,$AD,$AF
	.dc.b	$E1,$3C,$E8,$C5,$79,$A3,$15,$FE,$37,$E4,$FE,$0B,$F1,$5F,$C5,$4F
	.dc.b	$E2,$DF,$2D,$71,$FC,$DF,$2F,$F8,$83,$26,$7F,$C6,$E3,$0C,$9F,$0C
	.dc.b	$E9,$B1,$E6,$3C,$1F,$53,$FE,$DC,$8F,$E0,$2D,$D3,$DB,$3F,$E2,$68
	.dc.b	$C7,$07,$F8,$FF,$5F,$E3,$5F,$0B,$FE,$32,$F2,$7F,$D8,$58,$F2,$8C
	.dc.b	$11,$F8,$71,$AC,$7E,$13,$5F,$F0,$65,$74,$9D,$D7,$87,$E5,$F9,$57
	.dc.b	$E8,$F1,$47,$87,$E5,$E5,$FA,$BC,$4B,$7F,$B3,$C9,$5F,$E6,$26,$D7
	.dc.b	$92,$30,$47,$E1,$FC,$2F,$2D,$3F,$9D,$65,$FE,$6E,$9F,$E6,$B7,$FC
	.dc.b	$5B,$0E,$DE,$2F,$C7,$F8,$5F,$88,$CA,$F8,$2F,$84,$8E,$0F,$A3,$96
	.dc.b	$EB,$DC,$FC,$DF,$92,$93,$F5,$BD,$9B,$5E,$D9,$F6,$12,$7E,$7B,$63
	.dc.b	$E4,$7C,$2F,$0D,$3D,$6F,$D9,$F8,$F9,$7F,$82,$F9,$5B,$7F,$26,$7E
	.dc.b	$41,$27,$ED,$B6,$3E,$A7,$A7,$C3,$0F,$8E,$FF,$95,$BE,$67,$E4,$4F
	.dc.b	$F2,$D0,$E2,$7E,$6E,$BB,$E1,$98,$FF,$6F,$AE,$37,$FB,$E6,$BF,$E5
	.dc.b	$36,$BF,$EC,$1E,$1F,$F0,$C3,$F8,$7F,$95,$FC,$D7,$B5,$F9,$B6,$DF
	.dc.b	$E8,$1B,$3B,$8F,$D1,$FB,$3F,$A1,$3F,$45,$EC,$D8,$C9,$27,$F2,$91
	.dc.b	$97,$DD,$F2,$CF,$92,$75,$25,$F1,$BF,$E6,$37,$15,$B9,$D4,$E7,$B1
	.dc.b	$F3,$9F,$D7,$6F,$8F,$B3,$FE,$73,$C3,$84,$60,$E7,$8F,$81,$C0,$8D
	.dc.b	$E1,$D6,$FA,$3E,$39,$D5,$71,$8C,$57,$0B,$F0,$4E,$5F,$F7,$8D,$F2
	.dc.b	$73,$8F,$A9,$B8,$78,$AF,$92,$C7,$DB,$EB,$7F,$B2,$F2,$EF,$97,$5A
	.dc.b	$F5,$E3,$CD,$A7,$BE,$F8,$E9,$98,$F4,$CD,$C7,$E2,$C7,$F4,$66,$BE
	.dc.b	$43,$17,$F1,$DF,$F8,$03,$E9,$7F,$A6,$1F,$F2,$1F,$C7,$A3,$11,$AF
	.dc.b	$86,$FF,$22,$4E,$5F,$F2,$35,$FF,$88,$7C,$EB,$EC,$9A,$4B,$FF,$27
	.dc.b	$CE,$DF,$8D,$3B,$6C,$FC,$1D,$1F,$38,$F2,$BF,$D6,$9B,$FD,$6C,$38
	.dc.b	$36,$FE,$14,$63,$87,$F2,$58,$72,$BE,$31,$F8,$B9,$D6,$3F,$94,$6B
	.dc.b	$F8,$97,$35,$C6,$3A,$DC,$7C,$E7,$D5,$C7,$CC,$E7,$1F,$23,$A7,$F3
	.dc.b	$D1,$A8,$FC,$D7,$96,$F9,$CF,$F3,$44,$7F,$95,$9A,$FF,$89,$5B,$DB
	.dc.b	$FE,$6F,$7F,$DB,$F1,$BF,$1B,$FD,$B1,$39,$6F,$27,$EC,$EF,$B4,$62
	.dc.b	$91,$E1,$F8,$0C,$3E,$3D,$AE,$2F,$7A,$98,$89,$7E,$78,$CB,$E9,$8E
	.dc.b	$CF,$F2,$2F,$13,$AF,$73,$C5,$A6,$BC,$04,$D3,$FD,$25,$5C,$4F,$E4
	.dc.b	$CE,$23,$71,$FE,$12,$F9,$C7,$BB,$E8,$F7,$A4,$FF,$A3,$A7,$2D,$9C
	.dc.b	$BF,$36,$7E,$27,$2A,$B7,$3F,$E5,$F6,$F1,$7F,$81,$5B,$C5,$F8,$E6
	.dc.b	$CF,$CC,$DE,$DF,$3F,$4B,$E5,$FB,$D2,$BC,$91,$E0,$BF,$F1,$74,$7E
	.dc.b	$06,$74,$F8,$67,$44,$78,$9F,$87,$ED,$E5,$E8,$7D,$8F,$76,$4F,$C1
	.dc.b	$8C,$BF,$8D,$8C,$BE,$D8,$CB,$4E,$07,$E4,$FD,$DF,$8B,$D5,$FC,$27
	.dc.b	$6F,$C9,$63,$F8,$63,$3E,$14,$BF,$CD,$46,$5F,$F5,$FD,$71,$DF,$F6
	.dc.b	$AD,$8F,$E4,$CD,$DB,$FA,$4D,$FE,$4B,$DD,$F0,$FA,$1F,$94,$F7,$66
	.dc.b	$1D,$23,$2F,$EA,$63,$2F,$FA,$5F,$0F,$F1,$2B,$B2,$19,$9F,$CF,$FB
	.dc.b	$BF,$A0,$F6,$64,$FE,$CD,$F8,$3E,$7C,$8C,$10,$EC,$FE,$A3,$83,$51
	.dc.b	$83,$F7,$8F,$EB,$3C,$49,$1F,$BF,$7F,$5F,$EC,$FC,$EE,$CB,$3F,$C0
	.dc.b	$E3,$0F,$F0,$5F,$67,$F6,$A4,$7F,$DB,$BF,$39,$FE,$86,$E4,$71,$FF
	.dc.b	$73,$F1,$92,$F5,$7F,$7F,$F0,$7F,$ED,$6C,$3F,$DC,$AD,$7F,$D2,$73
	.dc.b	$B7,$FD,$34,$D7,$FD,$71,$1B,$C3,$BC,$7F,$0C,$8C,$61,$E1,$AE,$15
	.dc.b	$AC,$3C,$78,$7F,$A7,$30,$D3,$5F,$E9,$55,$DD,$FF,$4F,$35,$FE,$38
	.dc.b	$FE,$27,$C5,$F6,$FB,$5F,$C6,$A3,$FC,$D3,$3F,$13,$9F,$06,$66,$46
	.dc.b	$4F,$93,$ED,$71,$F5,$A4,$60,$87,$FB,$5E,$D1,$A8,$C4,$F8,$7D,$AF
	.dc.b	$17,$3C,$9F,$AB,$19,$7E,$BC,$65,$E9,$19,$7B,$46,$75,$F9,$B9,$FD
	.dc.b	$AA,$FE,$2F,$F3,$8F,$94,$F8,$1E,$E4,$C7,$FC,$65,$87,$FA,$6E,$B5
	.dc.b	$18,$E1,$FE,$CB,$18,$8F,$F5,$C3,$5F,$F4,$BB,$5D,$DF,$D5,$C7,$EC
	.dc.b	$C7,$E4,$3D,$BF,$A0,$7D,$1E,$19,$19,$21,$F2,$5F,$E8,$BE,$93,$E0
	.dc.b	$BF,$27,$ED,$CA,$F3,$1F,$E4,$88,$D9,$97,$F2,$31,$87,$67,$F9,$06
	.dc.b	$3F,$A4,$CF,$58,$FF,$0B,$C6,$0F,$57,$17,$E5,$A3,$FB,$1A,$F7,$3F
	.dc.b	$14,$7E,$2F,$B6,$4F,$EB,$A3,$2F,$EC,$23,$E9,$FC,$37,$E4,$71,$93
	.dc.b	$EA,$F1,$1B,$AB,$73,$FE,$EA,$8C,$36,$3F,$45,$3C,$1F,$5F,$57,$E7
	.dc.b	$7D,$C9,$5F,$03,$54,$FC,$C3,$31,$F9,$79,$C3,$B8,$E6,$FF,$8E,$9A
	.dc.b	$FF,$9E,$4F,$F3,$D7,$B0,$F9,$B7,$8D,$FA,$DF,$9C,$95,$F5,$AB,$53
	.dc.b	$FE,$0D,$6C,$7E,$F5,$AF,$50,$E1,$F0,$35,$FF,$45,$7C,$63,$E7,$3F
	.dc.b	$C5,$1D,$3F,$A2,$E3,$7C,$8F,$EF,$CF,$F0,$E5,$6F,$C6,$B7,$FE,$2F
	.dc.b	$C3,$D9,$39,$7F,$88,$35,$F2,$35,$DF,$C1,$7F,$2B,$EC,$FF,$15,$7B
	.dc.b	$0D,$BF,$C3,$67,$82,$75,$4B,$47,$F8,$69,$B1,$FC,$75,$AF,$8A,$3F
	.dc.b	$C3,$FD,$9F,$CD,$BE,$0F,$F8,$D1,$27,$F4,$70,$F8,$EF,$E5,$F1,$FC
	.dc.b	$4E,$74,$D3,$71,$AD,$FF,$9A,$E1,$E4,$F0,$7F,$C6,$7F,$29,$2F,$F5
	.dc.b	$F3,$96,$E4,$AF,$43,$63,$D1,$18,$3F,$C6,$FF,$36,$F8,$9F,$C1,$78
	.dc.b	$3F,$9C,$A6,$5F,$D1,$B9,$7F,$A9,$75,$FF,$91,$B0,$E1,$5F,$BA,$6B
	.dc.b	$FD,$15,$AF,$E2,$5A,$FE,$5D,$AF,$FB,$45,$E3,$5F,$19,$BC,$5B,$C7
	.dc.b	$F2,$DE,$4F,$FB,$66,$FC,$11,$E3,$23,$F9,$D8,$D3,$FE,$FD,$F9,$B7
	.dc.b	$CC,$FF,$97,$BE,$6F,$FC,$E7,$3F,$F0,$05,$B9,$7D,$6A,$DC,$FF,$16
	.dc.b	$6B,$FD,$85,$AF,$E9,$5A,$FE,$CD,$AF,$FB,$2E,$DE,$C7,$F9,$E3,$8D
	.dc.b	$F3,$74,$FF,$2D,$CE,$AB,$F8,$93,$5F,$A7,$A7,$FA,$16,$B8,$C7,$E2
	.dc.b	$2B,$8D,$C8,$FE,$ED,$E1,$7F,$BF,$A3,$F8,$AE,$1C,$E7,$F8,$44,$E5
	.dc.b	$DC,$7F,$08,$7F,$27,$F0,$FE,$73,$CB,$DF,$23,$24,$3C,$A7,$BD,$D3
	.dc.b	$8A,$57,$58,$C3,$74,$FE,$CF,$93,$E2,$B8,$C7,$F8,$5A,$F8,$E1,$8D
	.dc.b	$3E,$15,$7C,$8F,$4B,$E6,$7C,$7C,$CF,$E4,$C6,$5F,$D6,$C6,$5F,$0C
	.dc.b	$65,$A7,$03,$F6,$7F,$1D,$FE,$8E,$ED,$FF,$0A,$C7,$B9,$77,$FF,$1C
	.dc.b	$8F,$DF,$B5,$F8,$11,$C3,$4F,$52,$F4,$7F,$D5,$BC,$D2,$BF,$A8,$4E
	.dc.b	$5F,$EC,$BF,$16,$F8,$0F,$D9,$C7,$FB,$DF,$E2,$BF,$A0,$4E,$71,$FF
	.dc.b	$24,$46,$2B,$7E,$BF,$FB,$52,$B5,$7C,$4F,$E7,$9E,$EF,$06,$99,$3F
	.dc.b	$DF,$25,$01,$E1,$C7,$FE,$D4,$F5,$FF,$C9,$79,$7F,$88,$36,$78,$B8
	.dc.b	$E4,$FF,$7E,$ED,$FE,$32,$7E,$3C,$67,$26,$3F,$ED,$EF,$0B,$16,$FA
	.dc.b	$9F,$F3,$2E,$1F,$1A,$73,$18,$BF,$9F,$F1,$1C,$F0,$FF,$73,$BC,$BA
	.dc.b	$7F,$C6,$18,$7C,$E8,$D6,$3F,$ED,$A8,$FE,$63,$F2,$58,$9E,$0D,$8C
	.dc.b	$9C,$4B,$87,$DE,$F1,$B9,$9D,$FC,$2B,$3F,$72,$2C,$FE,$27,$1F,$E2
	.dc.b	$73,$92,$46,$7E,$73,$7C,$1F,$A4,$6C,$FF,$01,$6C,$B3,$FE,$9D,$1E
	.dc.b	$CF,$27,$FA,$7C,$FD,$06,$BF,$3A,$51,$8D,$79,$36,$33,$7F,$D5,$3E
	.dc.b	$6F,$A3,$4F,$CB,$CF,$CD,$8C,$DF,$B5,$FB,$8B,$E6,$BF,$C0,$35,$FE
	.dc.b	$87,$6C,$7F,$A2,$DA,$FF,$83,$DA,$FF,$87,$5D,$E5,$F8,$EF,$A7,$FF
	.dc.b	$29,$57,$F9,$2A,$72,$FF,$9A,$A3,$35,$FC,$50,$7C,$DB,$83,$CA,$DA
	.dc.b	$FF,$12,$79,$6F,$8F,$CD,$49,$FF,$2B,$D6,$EF,$F8,$FD,$FF,$89,$2F
	.dc.b	$51,$DE,$74,$FF,$8A,$5A,$FF,$8D,$DA,$FE,$B1,$AF,$9A,$FF,$49,$5F
	.dc.b	$C0,$7B,$7A,$E7,$FD,$CF,$B5,$1F,$A9,$F3,$BF,$C3,$3E,$FF,$DE,$6E
	.dc.b	$9F,$E1,$78,$C7,$29,$A7,$EA,$BE,$7B,$94,$66,$7F,$C3,$3D,$29,$FD
	.dc.b	$57,$CD,$49,$FF,$0B,$FC,$DD,$DF,$F3,$93,$75,$FE,$77,$36,$BF,$C8
	.dc.b	$B8,$FF,$9D,$DF,$F1,$57,$BF,$E4,$23,$83,$C3,$E7,$79,$23,$F7,$EF
	.dc.b	$9A,$7D,$64,$9F,$F1,$A3,$71,$FD,$77,$C0,$6E,$7F,$C5,$AF,$CC,$EA
	.dc.b	$7B,$9F,$A4,$E3,$FE,$22,$D8,$CD,$7F,$98,$DB,$7F,$E6,$3A,$DC,$FF
	.dc.b	$8E,$3E,$67,$A5,$FF,$0F,$7F,$43,$E4,$BE,$0B,$F6,$38,$7F,$41,$6C
	.dc.b	$66,$FF,$DD,$4D,$9F,$F7,$73,$77,$FF,$65,$F9,$9F,$80,$AF,$F5,$4C
	.dc.b	$E5,$FE,$6B,$DE,$9D,$5F,$E8,$6B,$4F,$F0,$78,$CD,$FF,$95,$27,$35
	.dc.b	$C7,$9F,$F9,$57,$CD,$FE,$5B,$63,$49,$D5,$FC,$25,$FD,$26,$1E,$6E
	.dc.b	$CE,$33,$FC,$EB,$2F,$D6,$69,$FE,$64,$5F,$2B,$9F,$9B,$D3,$67,$4D
	.dc.b	$DB,$FB,$85,$70,$9E,$55,$F3,$BE,$17,$E5,$92,$3E,$5B,$6B,$FB,$C7
	.dc.b	$9F,$FE,$0D,$9F,$93,$3F,$D3,$A3,$13,$F1,$E3,$1E,$97,$5A,$FF,$3D
	.dc.b	$D6,$B4,$FA,$7F,$29,$ED,$F9,$5F,$73,$6B,$DC,$DC,$7D,$2D,$AF,$F0
	.dc.b	$E6,$FF,$C5,$A7,$FC,$66,$D7,$FB,$7F,$07,$D0,$E8,$F9,$C4,$C3,$E5
	.dc.b	$7C,$9F,$D6,$8F,$9C,$FC,$FF,$67,$E8,$7C,$9D,$AF,$A0,$DC,$3E,$5B
	.dc.b	$72,$FC,$6B,$67,$F1,$7C,$28,$C5,$7E,$25,$FB,$9F,$27,$F7,$7E,$4E
	.dc.b	$DF,$DD,$6D,$7D,$86,$E1,$F9,$A7,$F0,$9E,$D7,$E1,$7E,$4E,$DF,$E1
	.dc.b	$5B,$7F,$80,$6D,$FE,$89,$FF,$04,$7B,$2E,$EF,$F4,$D5,$FF,$B5,$ED
	.dc.b	$39,$E4,$FE,$C2,$D9,$D3,$67,$FD,$BC,$DC,$13,$FA,$CA,$7B,$5E,$0C
	.dc.b	$3B,$E1,$F8,$C8,$C4,$FE,$2D,$B5,$F8,$87,$F4,$5F,$27,$23,$F4,$5F
	.dc.b	$27,$6F,$FC,$CD,$F3,$6D,$FF,$51,$F2,$7F,$E1,$B8,$C5,$FE,$A6,$3F
	.dc.b	$59,$F1,$32,$30,$7E,$B5,$FD,$87,$1F,$1C,$6C,$FF,$8E,$B6,$FF,$36
	.dc.b	$60,$9D,$4F,$F9,$6B,$2A,$7F,$97,$FD,$CF,$F2,$0C,$64,$91,$E8,$9F
	.dc.b	$F1,$BC,$EA,$3F,$D0,$1E,$17,$FA,$02,$BC,$07,$E7,$1C,$47,$5F,$81
	.dc.b	$FF,$1C,$46,$31,$FF,$40,$36,$BF,$DF,$7F,$27,$AA,$EE,$D8,$FF,$38
	.dc.b	$B6,$FD,$0F,$4F,$93,$93,$FE,$76,$F1,$ED,$79,$A7,$D3,$F0,$18,$CB
	.dc.b	$FC,$B4,$DB,$FE,$5B,$9F,$F9,$66,$F8,$47,$FA,$02,$7D,$3E,$1C,$8F
	.dc.b	$F4,$1F,$BF,$6B,$E1,$CF,$C6,$C7,$18,$FF,$99,$DB,$7D,$9B,$5F,$E3
	.dc.b	$DC,$51,$F1,$BE,$52,$57,$FA,$43,$D5,$B7,$F4,$27,$E9,$FC,$06,$31
	.dc.b	$FF,$3A,$36,$FD,$6D,$AE,$CF,$D2,$F8,$19,$3F,$E9,$CF,$8A,$B5,$F7
	.dc.b	$27,$EF,$F9,$F8,$C3,$FD,$00,$DB,$F8,$ED,$AF,$53,$FE,$44,$F9,$7F
	.dc.b	$95,$1F,$E8,$C3,$FD,$1A,$E1,$C3,$5C,$F0,$7F,$6A,$8F,$A5,$7F,$B6
	.dc.b	$9F,$F3,$17,$B2,$F9,$AE,$D7,$FE,$86,$7F,$2B,$EB,$FE,$5B,$E5,$6E
	.dc.b	$7F,$EB,$86,$E1,$F9,$F6,$E3,$FE,$9B,$7F,$3B,$EB,$CA,$FF,$5F,$FA
	.dc.b	$D6,$FF,$4D,$36,$FF,$D4,$53,$6B,$DD,$F2,$9E,$5E,$AF,$F5,$5F,$C3
	.dc.b	$FF,$93,$F1,$FF,$56,$CE,$5F,$F5,$77,$AF,$FD,$0B,$E6,$FF,$AA,$D6
	.dc.b	$E5,$5B,$8C,$36,$BB,$ED,$FE,$F0,$AE,$67,$FA,$BA,$F9,$CE,$B0,$F1
	.dc.b	$B6,$76,$76,$AE,$58,$79,$FD,$16,$1C,$AB,$D3,$5A,$BE,$57,$E4,$8C
	.dc.b	$4F,$87,$D5,$FB,$13,$9C,$7C,$33,$C9,$AF,$85,$CE,$3E,$0A,$E3,$87
	.dc.b	$63,$D1,$5F,$8B,$38,$1F,$D0,$F2,$F9,$8E,$5A,$FF,$BA,$BD,$17,$B5
	.dc.b	$FD,$26,$1F,$12,$9C,$C7,$F4,$D8,$C4,$ED,$B1,$DF,$E2,$FF,$A8,$B7
	.dc.b	$CF,$E9,$82,$73,$3E,$36,$BF,$86,$78,$EB,$F7,$1B,$96,$6F,$E1,$FC
	.dc.b	$DB,$CA,$FD,$55,$EE,$7E,$4E,$3D,$DB,$87,$2C,$7F,$DD,$F8,$71,$73
	.dc.b	$31,$FC,$0D,$71,$7F,$13,$B2,$B8,$46,$AF,$A3,$F9,$46,$D7,$1C,$3F
	.dc.b	$11,$5C,$1F,$35,$71,$C3,$C1,$1A,$C7,$D7,$97,$BE,$B4,$49,$D3,$C2
	.dc.b	$35,$1D,$0C,$C6,$E3,$24,$7E,$0C,$F6,$7F,$63,$8F,$A0,$97,$F3,$A7
	.dc.b	$53,$ED,$8F,$03,$C1,$D3,$83,$47,$84,$EE,$67,$1F,$89,$7E,$3C,$BE
	.dc.b	$65,$F6,$9C,$D7,$53,$94,$E9,$E5,$87,$99,$FA,$B9,$7C,$DC,$3D,$D1
	.dc.b	$E2,$CB,$E2,$47,$F9,$32,$75,$3D,$67,$2D,$9D,$63,$E3,$BE,$D5,$AC
	.dc.b	$3C,$57,$ED,$3B,$5F,$AF,$0E,$B1,$83,$A1,$27,$8D,$78,$23,$46,$E3
	.dc.b	$35,$EA,$C7,$E1,$CE,$6F,$AC,$60,$98,$7E,$76,$F8,$D7,$80,$EB,$3F
	.dc.b	$32,$33,$18,$9F,$6D,$FC,$08,$F9,$B9,$7D,$EA,$DD,$F6,$8C,$61,$F2
	.dc.b	$A7,$38,$72,$BF,$BE,$7D,$4C,$BF,$1D,$3E,$42,$7E,$F2,$F1,$B3,$87
	.dc.b	$32,$57,$09,$D5,$FF,$22,$6B,$D4,$FC,$89,$FB,$67,$3A,$7F,$37,$6E
	.dc.b	$9F,$E8,$C6,$CB,$14,$63,$DE,$76,$7F,$17,$AF,$06,$7E,$32,$3F,$CE
	.dc.b	$27,$7A,$7E,$67,$E7,$FF,$95,$4F,$F5,$66,$EE,$69,$23,$FC,$8A,$7F
	.dc.b	$62,$7F,$C0,$9D,$3F,$DF,$63,$91,$AB,$7D,$9B,$DC,$DF,$57,$F6,$EC
	.dc.b	$2D,$F3,$3A,$9C,$86,$1F,$E0,$9C,$39,$DF,$07,$E0,$F3,$DD,$5F,$1F
	.dc.b	$4C,$44,$C3,$C9,$3A,$6C,$E5,$EF,$97,$FB,$DA,$7F,$D1,$71,$C3,$E6
	.dc.b	$FB,$1D,$FD,$59,$19,$21,$C8,$FF,$0F,$BE,$7F,$8D,$F0,$46,$72,$47
	.dc.b	$68,$C3,$74,$F8,$0F,$06,$E5,$CE,$E5,$7C,$FA,$D6,$48,$61,$DA,$F7
	.dc.b	$1D,$AF,$C9,$1F,$4A,$1F,$8F,$7F,$C0,$07,$E9,$B5,$F6,$1A,$F8,$3F
	.dc.b	$A5,$F9,$6F,$E7,$B3,$FC,$15,$6D,$B3,$D6,$31,$97,$97,$25,$F1,$8F
	.dc.b	$A1,$F0,$BF,$6F,$E4,$A6,$5F,$62,$3E,$F1,$87,$E5,$6E,$9C,$C7,$E2
	.dc.b	$B9,$1F,$13,$F9,$C2,$47,$F7,$E6,$BE,$C8,$DB,$C6,$3F,$1D,$CA,$F8
	.dc.b	$6F,$8D,$5A,$AE,$15,$83,$73,$F9,$3F,$91,$FC,$A7,$3C,$AF,$C8,$C6
	.dc.b	$5F,$C9,$C6,$5F,$5C,$7D,$3F,$C9,$5E,$85,$24,$FE,$6B,$E4,$AB,$F5
	.dc.b	$27,$6F,$EA,$9A,$FC,$68,$FB,$DF,$71,$FC,$02,$0D,$23,$FC,$11,$FF
	.dc.b	$1B,$72,$FE,$F5,$BC,$3F,$E8,$36,$CF,$F9,$F3,$E8,$FA,$B0,$FD,$DC
	.dc.b	$6B,$93,$F8,$BC,$FF,$00,$F0,$1F,$A9,$3E,$27,$0D,$C3,$F8,$7C,$EA
	.dc.b	$7F,$7D,$C3,$A6,$4E,$AB,$83,$F8,$49,$E7,$F1,$32,$34,$FF,$07,$F8
	.dc.b	$FC,$47,$ED,$F0,$F1,$CE,$35,$FE,$21,$CE,$BF,$2F,$C4,$CC,$3C,$0F
	.dc.b	$F1,$1F,$9A,$C4,$73,$CD,$38,$C3,$9D,$6E,$BF,$89,$B5,$FC,$DF,$C6
	.dc.b	$D9,$DC,$68,$FE,$33,$1F,$1B,$E1,$24,$7C,$17,$F9,$17,$D0,$62,$3C
	.dc.b	$51,$6B,$D7,$3F,$3B,$CC,$95,$F2,$5F,$E5,$5F,$19,$88,$F5,$45,$AF
	.dc.b	$87,$37,$97,$F8,$77,$07,$DA,$AC,$5F,$FA,$FA,$AC,$E6,$3B,$1F,$CC
	.dc.b	$A3,$EF,$FC,$44,$8F,$B4,$FF,$3D,$FE,$01,$FF,$38,$E3,$FE,$C9,$BF
	.dc.b	$C3,$CE,$EB,$E7,$3B,$F4,$7F,$8D,$7C,$D7,$F0,$39,$FF,$43,$35,$FF
	.dc.b	$48,$07,$0F,$FA,$4D,$AF,$FB,$7D,$E1,$7F,$D7,$5B,$9E,$7C,$2F,$E5
	.dc.b	$1F,$37,$E6,$8C,$5F,$F5,$AA,$DB,$FC,$CD,$AA,$78,$3F,$CD,$FD,$87
	.dc.b	$07,$F7,$46,$C7,$FA,$7D,$AF,$A9,$AF,$FA,$AD,$CE,$3F,$DC,$DB,$87
	.dc.b	$FB,$E5,$B3,$FD,$17,$CC,$7C,$77,$F4,$19,$CC,$FF,$BD,$70,$C1,$FD
	.dc.b	$16,$B7,$1F,$DD,$A7,$64,$9D,$57,$F4,$BF,$CF,$FF,$D6,$95,$FD,$F3
	.dc.b	$E6,$3F,$AE,$63,$FD,$56,$11,$87,$FA,$C4,$F0,$7F,$A7,$B7,$2F,$F0
	.dc.b	$73,$6B,$93,$67,$FB,$05,$3D,$0F,$17,$B5,$27,$FC,$1B,$5B,$7C,$33
	.dc.b	$FD,$87,$F3,$F7,$83,$49,$3F,$E1,$4F,$B5,$7C,$4E,$FA,$A4,$FF,$86
	.dc.b	$2D,$1F,$DA,$EB,$FA,$E4,$FF,$73,$F7,$3D,$FB,$7F,$75,$E6,$7C,$7F
	.dc.b	$C3,$9C,$D2,$72,$F1,$DF,$FC,$46,$E6,$B8,$46,$27,$73,$FE,$CE,$34
	.dc.b	$7F,$BB,$FD,$CF,$05,$7E,$89,$B1,$93,$63,$E6,$9F,$10,$E3,$FF,$1A
	.dc.b	$C6,$1F,$F2,$1F,$C0,$BF,$18,$FC,$CC,$6A,$B8,$5F,$F8,$22,$DF,$09
	.dc.b	$8D,$BF,$C9,$11,$8B,$E7,$CE,$F4,$75,$1F,$87,$9D,$61,$FE,$12,$F8
	.dc.b	$47,$81,$A8,$FF,$10,$46,$2B,$CB,$F9,$A3,$E7,$35,$4A,$D3,$CE,$F3
	.dc.b	$18,$CF,$D3,$BB,$0E,$58,$FF,$95,$AB,$4F,$E6,$DB,$87,$F9,$DF,$F3
	.dc.b	$F7,$CB,$74,$CB,$CD,$E2,$A7,$38,$7F,$98,$3F,$2C,$7C,$C7,$99,$B1
	.dc.b	$FE,$70,$FA,$87,$E6,$2F,$84,$D5,$CA,$E4,$FC,$3A,$DC,$E7,$4F,$F1
	.dc.b	$CE,$1D,$1B,$5B,$C3,$A6,$1C,$89,$3F,$CA,$2B,$83,$FB,$90,$E1,$FD
	.dc.b	$E3,$5F,$F2,$97,$D1,$BE,$A3,$E8,$B6,$BF,$D1,$7C,$8F,$AE,$7D,$76
	.dc.b	$AE,$57,$FA,$12,$F8,$61,$FE,$6D,$C3,$85,$EA,$3F,$CA,$B3,$A7,$CC
	.dc.b	$D7,$F8,$83,$5F,$2D,$FF,$A6,$5F,$EB,$72,$F9,$DF,$AE,$DC,$3F,$D3
	.dc.b	$9E,$87,$EE,$1F,$F0,$AC,$66,$BF,$C1,$4D,$E7,$FE,$B0,$D9,$FF,$30
	.dc.b	$57,$1A,$DD,$F6,$F9,$59,$5D,$AE,$D7,$49,$B8,$7F,$A8,$F0,$F8,$F7
	.dc.b	$E2,$9C,$BE,$A7,$F4,$9E,$EA,$77,$7C,$EB,$81,$1F,$F4,$A4,$6A,$FD
	.dc.b	$37,$F0,$BE,$76,$5F,$C2,$F9,$1B,$87,$C0,$9B,$7E,$59,$B5,$F0,$3E
	.dc.b	$2F,$C7,$89,$BF,$47,$86,$4F,$CE,$72,$D3,$E7,$F9,$9F,$50,$DE,$5F
	.dc.b	$41,$C4,$68,$E1,$19,$7E,$BF,$CD,$7D,$C3,$C7,$F2,$16,$32,$49,$F8
	.dc.b	$71,$97,$EE,$F8,$BF,$84,$3D,$7F,$13,$6B,$F0,$71,$67,$E5,$CD,$D7
	.dc.b	$FD,$CD,$F8,$23,$E8,$FF,$89,$B7,$0F,$C6,$F9,$B8,$99,$1F,$EE,$49
	.dc.b	$D6,$1F,$5A,$B2,$FF,$AD,$23,$2F,$ED,$5A,$FF,$6F,$6B,$FE,$C5,$6A
	.dc.b	$F9,$B7,$CD,$FF,$26,$6D,$FE,$83,$D7,$C4,$C8,$FF,$7B,$CE,$AB,$F1
	.dc.b	$35,$97,$F1,$78,$FE,$03,$E2,$3F,$02,$DB,$FD,$5B,$67,$71,$FA,$E9
	.dc.b	$FE,$6F,$F2,$CF,$9A,$A7,$E7,$3E,$45,$F4,$78,$BF,$F7,$06,$FF,$D4
	.dc.b	$31,$61,$C6,$72,$7F,$84,$5C,$47,$FB,$CE,$F9,$CE,$A3,$FD,$D5,$3A
	.dc.b	$AE,$8F,$08,$FE,$44,$D7,$A6,$1F,$C2,$DE,$2D,$EB,$FD,$97,$89,$FA
	.dc.b	$2F,$8B,$FD,$1B,$71,$FE,$1A,$DB,$FD,$9B,$71,$F6,$7C,$2F,$F8,$76
	.dc.b	$BC,$D4,$8C,$29,$FE,$E3,$87,$FB,$BB,$1F,$EC,$35,$A9,$E7,$EE,$E0
	.dc.b	$FE,$0F,$F2,$9E,$FC,$7F,$C6,$D5,$B7,$FD,$F1,$F8,$E3,$B7,$F8,$6A
	.dc.b	$EA,$B1,$86,$CF,$CB,$CD,$5B,$8C,$3C,$DB,$3C,$34,$F5,$47,$82,$31
	.dc.b	$9F,$B2,$B9,$57,$F1,$8E,$1F,$E1,$BF,$15,$FC,$39,$B8,$7F,$38,$6C
	.dc.b	$EE,$3F,$9D,$CF,$F1,$4F,$5B,$F8,$AF,$C1,$59,$FE,$7E,$DA,$FE,$4A
	.dc.b	$DC,$7F,$99,$3F,$96,$FA,$B9,$5E,$FE,$5F,$F6,$0B,$F5,$B9,$53,$AC
	.dc.b	$3D,$A7,$E3,$9F,$CE,$65,$9C,$7F,$1F,$1F,$8B,$9F,$CD,$B6,$B9,$DF
	.dc.b	$FA,$72,$72,$FF,$A7,$BD,$17,$E8,$8F,$A0,$FF,$19,$FB,$AF,$1F,$FD
	.dc.b	$73,$39,$B6,$0A,$D2,$E3,$FE,$95,$CB,$CB,$0E,$38,$DA,$65,$E0,$8C
	.dc.b	$37,$1E,$B8,$F4,$C7,$9C,$63,$0E,$6D,$C3,$9E,$5E,$57,$C9,$3F,$ED
	.dc.b	$0F,$5F,$F3,$F8,$79,$B1,$ED,$5A,$3C,$73,$AA,$E1,$D1,$E4,$8C,$4F
	.dc.b	$6C,$F1,$33,$2F,$16,$1F,$E4,$78,$F5,$E5,$E3,$CB,$C1,$8F,$2A,$D5
	.dc.b	$FA,$71,$7C,$BB,$E1,$DE,$DF,$4C,$3C,$35,$A9,$CB,$E9,$31,$3C,$EB
	.dc.b	$93,$C7,$3F,$A1,$3A,$38,$4E,$9B,$5D,$32,$F3,$D7,$0A,$D4,$7D,$EA
	.dc.b	$E4,$D8,$CE,$BF,$23,$2F,$16,$1C,$E3,$FD,$75,$3A,$BF,$C4,$B7,$3F
	.dc.b	$87,$1E,$2C,$B8,$91,$FA,$D1,$AC,$7C,$18,$75,$9E,$6D,$7C,$B8,$78
	.dc.b	$DE,$84,$9E,$37,$F1,$71,$ED,$18,$9D,$57,$18,$C1,$1F,$5D,$F8,$70
	.dc.b	$F2,$46,$30,$E6,$DB,$F4,$E7,$F6,$B2,$F8,$51,$E6,$3D,$35,$E3,$7C
	.dc.b	$26,$0B,$7D,$9C,$63,$E3,$6E,$1E,$73,$C1,$19,$8E,$D3,$C5,$F0,$12
	.dc.b	$FE,$96,$1E,$5C,$7E,$54,$79,$8C,$E7,$F8,$38,$C9,$32,$F4,$63,$DE
	.dc.b	$B8,$4F,$C5,$BE,$95,$F1,$30,$F3,$9F,$B0,$C3,$A4,$66,$3C,$F3,$C1
	.dc.b	$F8,$66,$27,$6E,$2F,$CB,$9F,$17,$C3,$7E,$4C,$FE,$19,$2B,$C3,$9F
	.dc.b	$C3,$AD,$C7,$CE,$C7,$C1,$9F,$AF,$3F,$56,$7E,$08,$FC,$35,$7F,$A7
	.dc.b	$6F,$F5,$4D,$AF,$DA,$B7,$2F,$E6,$1F,$04,$F0,$9F,$3F,$86,$D7,$FA
	.dc.b	$9C,$4E,$E7,$47,$F3,$9C,$1F,$3B,$FF,$00,$6B,$84,$63,$E1,$35,$39
	.dc.b	$9F,$33,$5F,$D9,$BC,$73,$FD,$D3,$77,$CD,$7F,$1C,$F9,$FF,$F0,$45
	.dc.b	$FF,$0D,$AD,$3F,$BA,$6B,$FC,$69,$CF,$02,$78,$CF,$F1,$09,$DE,$3E
	.dc.b	$78,$C5,$F7,$E5,$FE,$9B,$18,$26,$5E,$69,$D4,$78,$D8,$EB,$2F,$EA
	.dc.b	$B9,$78,$6F,$9D,$F1,$AD,$1E,$9B,$E6,$E0,$F5,$3D,$23,$FA,$46,$5E
	.dc.b	$A7,$91,$30,$F8,$F5,$FC,$5A,$75,$18,$9D,$4F,$73,$57,$E4,$9D,$12
	.dc.b	$35,$3E,$09,$D4,$62,$74,$F9,$8E,$67,$9D,$CE,$3F,$36,$B9,$4E,$4E
	.dc.b	$44,$C7,$E3,$39,$3A,$47,$A2,$33,$1F,$16,$7A,$57,$9E,$73,$5E,$EA
	.dc.b	$E3,$39,$C3,$E1,$C7,$80,$C9,$E1,$72,$E0,$97,$D9,$FF,$03,$C6,$A3
	.dc.b	$11,$AC,$BF,$1F,$39,$7F,$1A,$4C,$BE,$EE,$1E,$5B,$E2,$75,$C7,$D3
	.dc.b	$8F,$C8,$BE,$18,$FC,$C3,$E0,$38,$BF,$97,$18,$C3,$E6,$5F,$28,$F7
	.dc.b	$61,$DA,$FC,$B9,$7D,$A9,$F4,$5F,$0C,$BE,$2E,$1C,$67,$27,$87,$1E
	.dc.b	$FC,$5F,$A4,$9F,$A6,$DA,$FD,$6B,$70,$FD,$4B,$6F,$E0,$E7,$FA,$4A
	.dc.b	$EF,$F3,$FF,$EA,$09,$FF,$0C,$B5,$F0,$4E,$EF,$F2,$38,$2B,$58,$FF
	.dc.b	$8A,$DE,$18,$F9,$5B,$87,$EC,$9A,$FF,$90,$1B,$5C,$F9,$FA,$C7,$4C
	.dc.b	$78,$47,$E8,$39,$DF,$2B,$FC,$83,$0F,$F0,$F7,$CD,$FF,$06,$EB,$FF
	.dc.b	$4C,$F9,$4F,$91,$E0,$DA,$D9,$FE,$A2,$E3,$5F,$F9,$8E,$3F,$C6,$2D
	.dc.b	$8C,$C7,$F2,$F8,$C1,$DF,$6F,$F0,$DB,$63,$FC,$36,$DC,$9D,$FF,$E9
	.dc.b	$F9,$FF,$4E,$66,$EA,$D2,$F9,$EA,$62,$24,$F0,$8C,$BC,$A3,$8F,$C0
	.dc.b	$72,$F9,$09,$3D,$A3,$2F,$82,$32,$FF,$A2,$B1,$F9,$73,$E7,$F9,$FF
	.dc.b	$14,$FF,$AC,$54,$64,$D6,$12,$75,$19,$78,$46,$5E,$D3,$C9,$D9,$A3
	.dc.b	$FD,$A7,$F3,$EF,$63,$E4,$79,$9F,$EA,$09,$FF,$61,$35,$FF,$50,$89
	.dc.b	$E0,$F2,$9F,$A4,$E5,$B1,$A8,$F8,$13,$F9,$DE,$3F,$F6,$FE,$AF,$9F
	.dc.b	$FA,$7C,$92,$FE,$BC,$7D,$63,$0F,$C9,$9F,$C6,$36,$3C,$3C,$6F,$9B
	.dc.b	$FC,$27,$84,$FC,$03,$C5,$93,$9B,$FD,$53,$39,$DB,$8D,$CC,$BF,$BF
	.dc.b	$3F,$93,$23,$83,$F3,$4E,$DF,$B9,$CA,$F6,$3F,$1B,$F1,$BF,$8E,$E8
	.dc.b	$95,$F9,$28,$CB,$F9,$48,$CB,$ED,$8F,$97,$F3,$EF,$A0,$FC,$E7,$BE
	.dc.b	$F8,$DA,$6A,$25,$FF,$B1,$37,$F7,$3F,$EC,$6B,$E4,$E7,$7F,$EF,$B1
	.dc.b	$88,$F9,$9F,$41,$F6,$BE,$83,$FC,$03,$1F,$97,$C9,$F1,$FE,$23,$51
	.dc.b	$0F,$F6,$E4,$FE,$ED,$34,$8E,$CF,$C5,$BD,$B1,$FC,$1F,$CC,$7E,$EF
	.dc.b	$C5,$7C,$FB,$DE,$FD,$E7,$CE,$EF,$6F,$F7,$0F,$84,$8C,$67,$FE,$DC
	.dc.b	$BE,$0F,$BC,$C5,$F5,$8C,$C7,$FB,$F3,$0E,$F3,$FC,$13,$9B,$5E,$FF
	.dc.b	$83,$A6,$5F,$3B,$F2,$CE,$A3,$18,$23,$11,$F9,$7F,$A0,$F0,$7B,$1F
	.dc.b	$F0,$06,$FD,$AD,$B3,$49,$5D,$1E,$2F,$F1,$AF,$A0,$FE,$32,$47,$07
	.dc.b	$F1,$EF,$A0,$C3,$FA,$78,$BC,$3F,$B0,$9F,$85,$ED,$F8,$49,$5F,$E6
	.dc.b	$1C,$7E,$8C,$62,$3E,$36,$28,$E3,$1F,$CB,$3E,$73,$E6,$7A,$3E,$B2
	.dc.b	$47,$3A,$FF,$28,$BF,$DA,$5F,$F2,$9E,$3F,$96,$C3,$EF,$D7,$2A,$FE
	.dc.b	$6F,$C3,$5F,$0C,$5A,$FE,$13,$37,$8F,$FC,$9F,$87,$46,$D7,$F3,$E9
	.dc.b	$FC,$07,$37,$CF,$26,$1E,$56,$BE,$E7,$6F,$E2,$79,$1C,$FF,$E7,$BC
	.dc.b	$BF,$5B,$2F,$A0,$FE,$27,$18,$FA,$07,$C7,$FF,$24,$6D,$7E,$4C,$E1
	.dc.b	$E4,$7F,$28,$BF,$E3,$71,$FB,$06,$BF,$E7,$26,$A3,$89,$FE,$97,$29
	.dc.b	$FC,$D9,$F1,$EB,$C6,$7C,$5E,$B7,$F3,$6F,$6F,$ED,$5F,$8F,$D0,$F9
	.dc.b	$0F,$56,$A8,$DB,$FD,$09,$AF,$F0,$66,$BF,$E8,$86,$BE,$76,$BC,$F9
	.dc.b	$3C,$44,$7C,$5F,$01,$5E,$D1,$75,$F9,$F6,$AF,$E9,$FF,$41,$FE,$BB
	.dc.b	$BF,$99,$F1,$7F,$D7,$DF,$EA,$ED,$BF,$AF,$3B,$9F,$EB,$6D,$7F,$8F
	.dc.b	$B5,$FF,$50,$B5,$FA,$0D,$4F,$38,$EB,$EF,$79,$0F,$EC,$75,$FE,$0F
	.dc.b	$AF,$C0,$7D,$03,$C2,$FF,$10,$36,$3E,$D3,$5F,$F6,$93,$5F,$E6,$6B
	.dc.b	$AB,$89,$23,$8B,$FE,$2B,$F7,$D6,$C7,$F1,$1B,$BE,$7F,$FD,$7B,$5A
	.dc.b	$9C,$C6,$2B,$86,$0B,$E0,$FE,$46,$F8,$39,$AF,$F7,$47,$BB,$CA,$3F
	.dc.b	$A4,$B5,$FF,$60,$9B,$32,$51,$F2,$CF,$8F,$FD,$6C,$7F,$B4,$A3,$1C
	.dc.b	$BF,$9F,$6C,$7F,$BD,$DA,$FF,$58,$5F,$3D,$CA,$3F,$51,$18,$9E,$53
	.dc.b	$99,$F0,$B7,$E1,$3F,$AF,$CF,$ED,$DB,$3F,$2E,$3F,$C1,$3F,$3A,$F9
	.dc.b	$EF,$ED,$FB,$FF,$04,$BD,$B7,$BF,$A1,$2F,$F7,$13,$D3,$0F,$ED,$EF
	.dc.b	$08,$FF,$74,$FC,$F7,$F7,$F7,$C5,$5D,$9C,$FA,$36,$FF,$82,$7C,$E3
	.dc.b	$E4,$71,$08,$C5,$7F,$63,$F9,$E7,$CF,$7F,$93,$A3,$F8,$A4,$EA,$BF
	.dc.b	$C7,$ED,$D3,$F8,$FB,$67,$F8,$95,$F2,$F8,$3E,$F3,$9B,$F1,$E3,$F3
	.dc.b	$63,$A4,$7F,$83,$1A,$FF,$1B,$3C,$D1,$B3,$D9,$C2,$FC,$BF,$FC,$AD
	.dc.b	$8F,$9A,$73,$18,$AF,$F2,$AF,$CF,$3E,$7B,$E0,$B6,$FF,$C0,$58,$3E
	.dc.b	$7F,$F9,$AC,$7F,$33,$9D,$5F,$2B,$FC,$24,$62,$3F,$C4,$6D,$7D,$0D
	.dc.b	$7F,$9A,$3C,$B5,$FC,$1F,$D6,$3D,$0F,$1D,$FD,$79,$4E,$70,$FF,$06
	.dc.b	$79,$4F,$A0,$FB,$9E,$8F,$F4,$A7,$EF,$79,$DF,$D4,$BB,$B3,$5B,$BE
	.dc.b	$35,$F9,$FB,$E1,$1F,$DE,$E7,$5A,$FF,$8F,$27,$2D,$87,$D0,$74,$3B
	.dc.b	$F9,$FF,$1A,$7F,$C4,$D3,$AB,$F3,$D6,$DF,$EC,$D5,$BD,$3F,$BE,$56
	.dc.b	$EF,$8D,$ED,$FF,$26,$45,$9D,$CF,$A6,$38,$4F,$FB,$BD,$AF,$F6,$1C
	.dc.b	$3F,$B6,$CE,$FE,$83,$E1,$9E,$8F,$8F,$B9,$7C,$C8,$D3,$F0,$DB,$7F
	.dc.b	$E5,$ED,$7F,$BE,$FC,$27,$C0,$3E,$0F,$CD,$5B,$F6,$C5,$AF,$F3,$24
	.dc.b	$DD,$3D,$FF,$41,$F5,$0C,$5F,$9F,$54,$66,$3E,$A3,$EF,$23,$F5,$FE
	.dc.b	$83,$EE,$9F,$4B,$E7,$2C,$7E,$1A,$3F,$07,$EB,$FD,$CC,$3E,$7C,$63
	.dc.b	$8B,$F0,$BE,$3C,$BF,$AB,$F0,$F7,$5F,$B7,$3C,$BE,$4F,$18,$FF,$A2
	.dc.b	$38,$3F,$CD,$91,$FC,$2B,$E8,$1F,$3F,$4F,$C2,$F9,$F7,$3F,$F0,$E4
	.dc.b	$E9,$FF,$0A,$37,$2F,$F4,$58,$70,$FF,$A3,$9A,$FF,$03,$78,$63,$E3
	.dc.b	$DB,$F1,$4F,$F8,$C2,$33,$5F,$E8,$29,$CB,$FC,$2F,$E5,$5F,$41,$FC
	.dc.b	$89,$B1,$F9,$6F,$57,$13,$23,$FC,$7D,$4E,$7C,$65,$FE,$F7,$8C,$4F
	.dc.b	$E7,$B4,$FD,$2F,$C3,$7E,$99,$B1,$93,$F4,$71,$96,$9F,$E4,$1D,$9F
	.dc.b	$41,$FC,$D9,$B1,$FE,$4F,$F8,$E7,$CF,$E6,$A4,$FF,$8F,$EB,$79,$7C
	.dc.b	$C6,$BF,$EB,$86,$BF,$CB,$5D,$F0,$7D,$DC,$3F,$C5,$55,$E3,$BF,$E1
	.dc.b	$B1,$8B,$FF,$60,$79,$1F,$C0,$BE,$1F,$F8,$1B,$6F,$F7,$ED,$9D,$C7
	.dc.b	$F0,$29,$FE,$A3,$F0,$6F,$98,$A7,$F1,$AF,$95,$F1,$D7,$F4,$19,$D3
	.dc.b	$FE,$65,$6D,$FF,$6F,$1C,$EF,$97,$7D,$5F,$4B,$FF,$3B,$46,$1F,$F3
	.dc.b	$C7,$D0,$BC,$2F,$E5,$2B,$C0,$B1,$86,$CF,$06,$E3,$E1,$C5,$7C,$6B
	.dc.b	$AD,$F1,$69,$C4,$FE,$47,$C7,$FC,$AB,$E2,$FF,$95,$B6,$32,$7F,$28
	.dc.b	$8C,$B4,$FE,$57,$F0,$7F,$CC,$BE,$17,$F9,$9B,$63,$27,$F3,$08,$CB
	.dc.b	$4F,$C7,$7D,$0F,$E3,$4F,$BB,$56,$32,$7E,$2E,$3F,$C4,$5C,$E6,$92
	.dc.b	$3F,$D4,$5D,$CF,$91,$47,$FA,$4F,$E0,$1E,$57,$E3,$EF,$FA,$BD,$EE
	.dc.b	$B5,$3F,$CE,$7E,$81,$F4,$3C,$B8,$98,$F2,$5E,$72,$B7,$87,$29,$FC
	.dc.b	$1F,$9E,$E8,$E5,$87,$1C,$38,$63,$8C,$BC,$0F,$7C,$7F,$C2,$13,$A7
	.dc.b	$FD,$58,$63,$2E,$F9,$7F,$AF,$A7,$4D,$8C,$E1,$DF,$DF,$FC,$BB,$E1
	.dc.b	$8C,$E3,$FD,$5E,$72,$DE,$BF,$25,$FB,$A7,$2F,$C0,$AE,$2E,$67,$78
	.dc.b	$2B,$F4,$B8,$FA,$9B,$5D,$63,$D6,$D3,$D5,$1D,$97,$CC,$FF,$AF,$7E
	.dc.b	$4B,$E6,$5F,$5C,$3C,$B5,$AF,$8C,$57,$D1,$8C,$36,$F9,$47,$DA,$CF
	.dc.b	$C1,$A7,$DB,$8D,$99,$3F,$DB,$53,$C2,$34,$D9,$E2,$7F,$91,$5C,$4F
	.dc.b	$0B,$F1,$BC,$CE,$B5,$F4,$F0,$F4,$D7,$09,$CD,$F3,$BE,$37,$BC,$FD
	.dc.b	$51,$89,$D6,$38,$3E,$2E,$7F,$27,$1E,$84,$AE,$18,$FD,$7C,$38,$3E
	.dc.b	$59,$D5,$F2,$CB,$CB,$7C,$2B,$CA,$7B,$67,$57,$CB,$2F,$8B,$5B,$C3
	.dc.b	$E1,$E7,$F5,$EB,$59,$FA,$31,$F0,$5F,$92,$B5,$18,$C7,$6F,$73,$EA
	.dc.b	$46,$B1,$EF,$3A,$9F,$2C,$E6,$FB,$56,$A7,$A1,$CB,$3F,$8D,$87,$9C
	.dc.b	$98,$7B,$27,$75,$C4,$97,$F8,$08,$F2,$46,$1B,$19,$7E,$39,$32,$F8
	.dc.b	$78,$78,$2F,$D1,$18,$CF,$E7,$C6,$31,$F8,$95,$BB,$F1,$E1,$EC,$3B
	.dc.b	$61,$F6,$23,$27,$DD,$29,$E1,$AE,$78,$76,$8C,$CF,$87,$3F,$06,$3F
	.dc.b	$43,$3F,$95,$5E,$C2,$1F,$3B,$2F,$26,$7F,$06,$31,$8F,$58,$FA,$11
	.dc.b	$FE,$A9,$3F,$5A,$5C,$7F,$66,$DC,$BF,$4E,$EF,$E1,$9E,$53,$E8,$7F
	.dc.b	$C6,$B2,$9C,$C6,$1B,$1F,$E3,$A6,$BD,$FE,$85,$F1,$3C,$B1,$46,$27
	.dc.b	$FC,$83,$18,$9D,$4E,$67,$94,$7F,$0B,$8E,$19,$7F,$04,$6F,$07,$FA
	.dc.b	$07,$E8,$7F,$A0,$5E,$EE,$3F,$AF,$AD,$BF,$D2,$A7,$6E,$67,$77,$D5
	.dc.b	$FE,$43,$3A,$8C,$4E,$C8,$FF,$4D,$8D,$5F,$F1,$5B,$EF,$5F,$CB,$1B
	.dc.b	$87,$BE,$3F,$5B,$5C,$1F,$E4,$33,$A2,$61,$E2,$E4,$FF,$72,$63,$E2
	.dc.b	$26,$3D,$E7,$57,$EA,$8C,$3E,$67,$E3,$57,$27,$35,$CA,$7C,$35,$C9
	.dc.b	$E0,$60,$BC,$0C,$7D,$11,$ED,$9F,$03,$BA,$F8,$98,$77,$8C,$61,$E0
	.dc.b	$23,$F4,$E3,$4F,$33,$C2,$59,$E6,$F6,$31,$3C,$B0,$F2,$57,$2C,$BE
	.dc.b	$75,$FB,$6F,$67,$73,$94,$6A,$F9,$63,$EC,$8C,$4F,$A0,$98,$76,$78
	.dc.b	$E5,$F5,$9F,$B7,$8F,$86,$31,$8F,$CD,$AE,$15,$EC,$BE,$AF,$92,$33
	.dc.b	$3E,$79,$C9,$E2,$8C,$E3,$F2,$2B,$D6,$71,$C7,$D1,$87,$82,$7D,$78
	.dc.b	$F8,$F2,$F3,$9F,$77,$2E,$D5,$BC,$3E,$21,$E5,$71,$8F,$82,$33,$3C
	.dc.b	$2B,$96,$1F,$8C,$9E,$A7,$C2,$32,$70,$74,$FF,$4C,$FA,$17,$57,$E1
	.dc.b	$70,$46,$27,$8C,$78,$4C,$CE,$E7,$53,$DC,$98,$7F,$8F,$E3,$2D,$9E
	.dc.b	$04,$9F,$F2,$47,$82,$F9,$FF,$E2,$B1,$FE,$19,$F9,$D3,$FE,$5E,$F9
	.dc.b	$3F,$CD,$53,$B1,$1F,$8F,$AD,$D7,$FA,$54,$FF,$15,$B8,$F3,$7F,$DD
	.dc.b	$55,$C2,$7C,$2D,$7F,$C0,$93,$BE,$AF,$F0,$F3,$FC,$CA,$39,$9F,$CA
	.dc.b	$DC,$71,$9F,$E6,$F8,$FA,$FA,$39,$57,$FA,$9B,$37,$D0,$E2,$BE,$03
	.dc.b	$63,$FD,$69,$3F,$03,$3F,$F8,$01,$18,$8F,$F5,$03,$C9,$D6,$5F,$1A
	.dc.b	$F9,$C6,$0F,$06,$1D,$7E,$1F,$B7,$CA,$49,$D5,$7D,$19,$D4,$63,$05
	.dc.b	$6E,$3C,$33,$E9,$F2,$BC,$51,$FE,$B3,$51,$93,$59,$C9,$FF,$61,$C6
	.dc.b	$5F,$F6,$2C,$65,$E5,$19,$7A,$47,$CA,$5F,$42,$F7,$FE,$57,$9D,$FE
	.dc.b	$A5,$9F,$5B,$5F,$C1,$4E,$DF,$88,$72,$9F,$B9,$19,$8F,$EF,$06,$4F
	.dc.b	$EE,$2E,$5F,$C6,$E7,$F9,$7F,$A1,$FB,$1C,$72,$7E,$44,$7D,$B3,$0F
	.dc.b	$D8,$8C,$3E,$2E,$27,$CE,$7E,$1B,$C3,$7D,$0F,$8F,$93,$FD,$65,$18
	.dc.b	$D3,$E1,$5F,$0C,$39,$61,$C7,$0E,$1C,$9F,$A8,$C3,$85,$68,$87,$F8
	.dc.b	$02,$BF,$32,$47,$E0,$E9,$FB,$3F,$A1,$FC,$A7,$34,$9F,$C2,$47,$E5
	.dc.b	$CC,$3F,$94,$8C,$3F,$3B,$E8,$1F,$43,$F9,$FF,$29,$F2,$54,$FB,$93
	.dc.b	$AA,$E1,$5F,$AF,$8C,$47,$D4,$6B,$FE,$2D,$6B,$FC,$03,$E8,$7F,$C6
	.dc.b	$1C,$3F,$ED,$9C,$D8,$BE,$53,$C4,$7E,$2E,$B7,$BF,$1C,$32,$7E,$92
	.dc.b	$B8,$3F,$88,$9D,$BF,$C5,$1F,$E0,$89,$A4,$7F,$C7,$3E,$0B,$E7,$DE
	.dc.b	$EF,$F0,$1F,$7F,$9E,$67,$AD,$6E,$7F,$84,$B6,$FF,$2A,$D7,$FC,$9C
	.dc.b	$D7,$FD,$D3,$7D,$9D,$AF,$6F,$C1,$C9,$27,$F8,$35,$79,$67,$BD,$6E
	.dc.b	$78,$C7,$AA,$31,$1E,$06,$BE,$0F,$53,$F8,$FF,$C8,$BE,$63,$F8,$4C
	.dc.b	$FF,$13,$8C,$78,$5D,$AA,$4F,$99,$EC,$47,$F2,$7E,$1B,$E8,$3F,$93
	.dc.b	$12,$3F,$8B,$7C,$8D,$7F,$92,$C5,$E5,$FE,$53,$3E,$CF,$67,$E8,$7D
	.dc.b	$34,$BF,$83,$AA,$75,$1F,$21,$F8,$04,$7E,$5F,$A1,$FC,$E7,$F1,$E7
	.dc.b	$B7,$4F,$93,$12,$76,$43,$E4,$3F,$2C,$D2,$C9,$D3,$FD,$83,$8F,$F1
	.dc.b	$15,$FD,$17,$9B,$75,$AB,$56,$9F,$5D,$6D,$F6,$F5,$7E,$3F,$E3,$A4
	.dc.b	$60,$87,$DB,$8F,$BC,$47,$FD,$23,$F4,$3F,$D9,$70,$F0,$43,$E8,$9F
	.dc.b	$23,$FC,$D5,$B1,$FC,$BA,$B8,$69,$FD,$D1,$B1,$FD,$F2,$76,$FF,$77
	.dc.b	$6D,$F6,$9D,$C9,$FE,$EC,$FF,$AB,$5A,$F5,$F2,$3F,$BB,$46,$23,$FC
	.dc.b	$13,$ED,$FD,$D7,$FA,$0F,$47,$FA,$42,$FD,$6D,$C7,$FC,$13,$5B,$AF
	.dc.b	$F0,$53,$6B,$7A,$FF,$53,$BE,$15,$DB,$8B,$FA,$B5,$EF,$0E,$71,$83
	.dc.b	$87,$C0,$5A,$FE,$A7,$5B,$79,$1E,$68,$FE,$C1,$F4,$5F,$EC,$48,$FE
	.dc.b	$B3,$F1,$DF,$DB,$9B,$E8,$EB,$DE,$91,$82,$1F,$E2,$3A,$E4,$D8,$DB
	.dc.b	$4E,$BE,$A7,$F6,$FF,$A1,$7D,$17,$F7,$1F,$A5,$FE,$67,$C3,$F8,$26
	.dc.b	$EF,$FE,$30,$AD,$E1,$FD,$B9,$7D,$1F,$1F,$90,$93,$98,$FE,$F9,$F0
	.dc.b	$4D,$FF,$45,$3F,$ED,$8F,$98,$FF,$1D,$CE,$73,$FF,$21,$FC,$0A,$73
	.dc.b	$7C,$23,$17,$FE,$48,$BE,$0F,$33,$FD,$E9,$EF,$F2,$BF,$EF,$ED,$8C
	.dc.b	$9B,$3F,$D7,$E5,$F2,$1F,$29,$FB,$1C,$A7,$DE,$BE,$93,$1C,$1F,$E5
	.dc.b	$2B,$DB,$FE,$33,$A5,$6E,$3F,$C2,$73,$70,$FE,$05,$1F,$DE,$7E,$9B
	.dc.b	$FB,$3B,$AC,$3A,$CB,$CF,$56,$A3,$FC,$C1,$F1,$3F,$D7,$EF,$FC,$3E
	.dc.b	$D8,$EA,$71,$F3,$DF,$4B,$D2,$13,$BD,$3F,$7E,$DB,$FE,$29,$F4,$BD
	.dc.b	$E7,$3E,$37,$93,$C8,$4E,$6B,$FC,$E7,$7C,$70,$FE,$2F,$39,$7F,$CF
	.dc.b	$B1,$9A,$E3,$EC,$F8,$FE,$8D,$F2,$B4,$F8,$3F,$48,$FA,$5F,$F2,$7E
	.dc.b	$7F,$E3,$CA,$DB,$6B,$83,$FE,$63,$AE,$1E,$1E,$33,$FF,$43,$5F,$07
	.dc.b	$E2,$F2,$3E,$97,$FC,$9D,$87,$33,$FA,$D3,$89,$E3,$8F,$D4,$AE,$1F
	.dc.b	$09,$D7,$87,$F9,$B4,$63,$0F,$F2,$8F,$D2,$3E,$93,$F9,$5E,$3F,$E9
	.dc.b	$88,$C4,$FF,$94,$A1,$18,$7F,$CA,$AD,$7F,$D3,$E6,$A7,$F0,$B6,$FA
	.dc.b	$47,$B1,$F7,$6F,$94,$FE,$66,$73,$87,$92,$F8,$C6,$30,$FF,$55,$5F
	.dc.b	$0A,$FF,$4A,$4E,$6F,$FA,$97,$17,$F4,$2F,$A2,$F5,$E1,$FD,$53,$D3
	.dc.b	$46,$27,$58,$7F,$86,$A7,$2D,$9F,$CE,$FB,$FF,$E2,$3F,$A3,$FC,$65
	.dc.b	$6B,$0F,$CC,$CE,$6F,$FC,$33,$2F,$A4,$D9,$CB,$E0,$FF,$AA,$CE,$A3
	.dc.b	$2D,$3F,$C7,$D3,$8B,$FD,$2E,$5F,$D7,$2B,$D1,$3D,$6B,$D1,$18,$23
	.dc.b	$FE,$AA,$FA,$6F,$DF,$DF,$2A,$E0,$FA,$23,$3A,$7C,$89,$79,$5D,$89
	.dc.b	$7F,$15,$B7,$F0,$3D,$AF,$F3,$5C,$FC,$09,$F8,$F1,$89,$F8,$B9,$FE
	.dc.b	$57,$0F,$F3,$13,$6F,$FD,$8C,$BE,$73,$E3,$72,$C9,$FF,$30,$3F,$26
	.dc.b	$31,$7E,$BF,$6F,$E1,$92,$BE,$D3,$63,$F5,$3F,$36,$3E,$CC,$7E,$95
	.dc.b	$C4,$FD,$8E,$1F,$9A,$47,$F0,$8D,$AF,$81,$C5,$F8,$28,$F7,$BF,$84
	.dc.b	$F2,$3E,$A9,$2B,$F1,$8D,$AF,$91,$F2,$9B,$9F,$F6,$FE,$C8,$FE,$41
	.dc.b	$F4,$8F,$A3,$FC,$CD,$A3,$17,$FB,$36,$C7,$FB,$19,$AF,$EC,$9C,$F0
	.dc.b	$7F,$A4,$70,$FC,$0B,$FE,$28,$8C,$D7,$FB,$DE,$72,$AA,$4E,$97,$37
	.dc.b	$FA,$17,$97,$CF,$8F,$E0,$1F,$E7,$1F,$2C,$FA,$4F,$4B,$71,$FE,$14
	.dc.b	$DC,$3F,$DB,$AD,$7F,$84,$B9,$CF,$D1,$CB,$D2,$FF,$1A,$FE,$AF,$1F
	.dc.b	$D6,$7C,$55,$C7,$FD,$4C,$DC,$7F,$22,$DC,$3F,$22,$FF,$4D,$D9,$F4
	.dc.b	$9F,$39,$B7,$FC,$7D,$B7,$FE,$F2,$6B,$FC,$79,$CD,$FF,$AF,$7E,$91
	.dc.b	$EE,$F5,$C3,$F8,$4C,$63,$0F,$F7,$FD,$F0,$7E,$76,$E9,$D6,$3F,$D9
	.dc.b	$0F,$C2,$5F,$2C,$3F,$85,$78,$5F,$C2,$FE,$32,$E3,$FE,$C5,$6D,$FE
	.dc.b	$C1,$B9,$FE,$C1,$FE,$D9,$F0,$4F,$77,$F1,$AD,$BF,$E7,$CD,$BF,$E7
	.dc.b	$61,$C3,$FC,$F9,$D5,$FF,$B7,$7E,$65,$F4,$9D,$B0,$FE,$53,$18,$C3
	.dc.b	$F1,$95,$FE,$E3,$FA,$3F,$E0,$11,$83,$E1,$AF,$A4,$FE,$62,$EB,$2F
	.dc.b	$E5,$EE,$67,$84,$7E,$52,$35,$5D,$6B,$8F,$7F,$26,$DF,$0E,$15,$F3
	.dc.b	$C3,$AD,$EE,$BF,$DE,$1F,$49,$FC,$E7,$E4,$7F,$9D,$36,$32,$7F,$BD
	.dc.b	$63,$2D,$3F,$DF,$1C,$3F,$D0,$7E,$1F,$FA,$13,$63,$27,$FB,$F6,$32
	.dc.b	$D3,$F0,$FE,$0F,$E5,$48,$FE,$55,$B8,$FF,$52,$EB,$FF,$60,$CF,$EE
	.dc.b	$7D,$C7,$CF,$FE,$EE,$BF,$87,$5F,$F5,$CB,$DC,$2B,$6F,$F0,$1E,$9F
	.dc.b	$E4,$AF,$5F,$8D,$FF,$5F,$4E,$A3,$9E,$3F,$2E,$F9,$CF,$2F,$29,$F4
	.dc.b	$6F,$A4,$E3,$A7,$FB,$2E,$31,$87,$9E,$3F,$A1,$F4,$EB,$2F,$F6,$5D
	.dc.b	$69,$F3,$4F,$1B,$D6,$1F,$E4,$6E,$8B,$7B,$CB,$C5,$EC,$7E,$6D,$F2
	.dc.b	$46,$70,$FF,$6E,$4E,$5E,$99,$F8,$F2,$FF,$6D,$CE,$5B,$87,$82,$31
	.dc.b	$8F,$BB,$2C,$4C,$BE,$56,$F9,$C7,$C6,$7F,$99,$E3,$D6,$74,$9D,$CE
	.dc.b	$BE,$03,$C5,$8F,$B6,$73,$7C,$75,$F6,$E3,$F1,$F1,$ED,$39,$D3,$F0
	.dc.b	$19,$73,$9E,$E6,$B2,$EB,$A7,$B6,$30,$6A,$78,$63,$F7,$DB,$7F,$23
	.dc.b	$1F,$04,$62,$BF,$A4,$35,$E7,$5E,$6C,$BD,$38,$C9,$ED,$97,$9D,$F8
	.dc.b	$46,$23,$55,$E1,$9D,$67,$F0,$EB,$44,$9D,$57,$B3,$1F,$2D,$F5,$AF
	.dc.b	$1E,$5F,$12,$39,$E7,$F0,$73,$F4,$12,$33,$7C,$B0,$F1,$DF,$17,$B6
	.dc.b	$78,$C7,$E0,$DF,$7C,$3E,$26,$7E,$7C,$FC,$B5,$E7,$C7,$93,$F8,$29
	.dc.b	$D4,$62,$76,$47,$E8,$61,$DE,$BD,$67,$83,$1F,$1B,$D4,$C1,$6F,$AB
	.dc.b	$8B,$E8,$DB,$EA,$F1,$3A,$9D,$A7,$8B,$8C,$3E,$35,$F8,$2F,$9C,$E7
	.dc.b	$1F,$93,$8F,$9E,$F7,$3E,$FC,$FE,$85,$7C,$72,$61,$CF,$2F,$84,$4A
	.dc.b	$F3,$1F,$86,$C7,$BE,$1D,$1F,$C7,$47,$A6,$31,$3E,$C9,$DD,$F6,$AF
	.dc.b	$16,$7F,$75,$E1,$1E,$86,$E1,$F7,$73,$F8,$38,$FA,$A7,$FD,$3E,$FE
	.dc.b	$B5,$B5,$FB,$D6,$D7,$FA,$A7,$E1,$1F,$16,$FA,$4C,$36,$BF,$B7,$CE
	.dc.b	$A7,$27,$FA,$C7,$07,$D1,$BE,$23,$0D,$BF,$EE,$F3,$A9,$C9,$23,$F7
	.dc.b	$93,$BD,$7F,$78,$DD,$BF,$C8,$1A,$BE,$93,$FC,$01,$7F,$C5,$2B,$4F
	.dc.b	$EF,$9A,$FF,$1D,$73,$3C,$79,$3F,$A8,$61,$DB,$1E,$44,$BF,$D0,$61
	.dc.b	$D9,$F2,$C6,$67,$84,$6A,$B9,$E3,$FC,$73,$1F,$0E,$5E,$FA,$E2,$6C
	.dc.b	$E5,$3B,$C2,$C3,$2F,$33,$83,$D5,$1A,$8C,$C7,$4C,$7C,$64,$AF,$8F
	.dc.b	$97,$A6,$31,$3F,$16,$F9,$4E,$6B,$C5,$3C,$DB,$5F,$D9,$E7,$53,$C4
	.dc.b	$D1,$63,$59,$7C,$28,$C5,$72,$23,$ED,$AE,$6F,$B0,$E6,$5A,$E6,$FC
	.dc.b	$1C,$7E,$54,$7A,$B1,$F8,$2F,$B2,$78,$5F,$69,$CB,$E0,$8C,$9E,$43
	.dc.b	$C5,$87,$9D,$C9,$2F,$BC,$FC,$A9,$D4,$62,$79,$1D,$2B,$BB,$88,$D3
	.dc.b	$9C,$3D,$F1,$82,$65,$F9,$48,$CB,$F8,$82,$65,$FD,$BA,$32,$FC,$38
	.dc.b	$E1,$87,$91,$DD,$FD,$4C,$3C,$6F,$AE,$33,$97,$CC,$8F,$9F,$8F,$8B
	.dc.b	$0E,$D8,$73,$7B,$18,$9A,$75,$C7,$F0,$78,$F6,$BE,$53,$F9,$D2,$61
	.dc.b	$F5,$63,$E4,$4F,$C6,$38,$63,$E4,$CB,$E7,$5F,$73,$8B,$87,$FA,$97
	.dc.b	$D2,$7F,$B4,$DD,$61,$EB,$24,$FF,$06,$9D,$4F,$F8,$1F,$E4,$7F,$C9
	.dc.b	$DC,$9F,$C9,$1B,$1F,$E1,$26,$BF,$E0,$66,$BF,$E2,$57,$F0,$24,$7F
	.dc.b	$C6,$5E,$0B,$A3,$FD,$1B,$1F,$E2,$69,$D3,$E1,$37,$1A,$7F,$C5,$B2
	.dc.b	$F4,$1E,$8F,$EA,$63,$0D,$8C,$CF,$81,$7D,$56,$66,$4E,$5E,$11,$C7
	.dc.b	$D0,$E5,$F5,$12,$B9,$46,$5E,$91,$97,$CB,$7F,$EE,$77,$DF,$E4,$FF
	.dc.b	$3E,$6F,$07,$9E,$31,$1E,$C7,$FD,$74,$E8,$86,$63,$87,$5F,$16,$46
	.dc.b	$B8,$7F,$CF,$FA,$2F,$9C,$FF,$95,$39,$5F,$A1,$FC,$48,$C3,$6F,$FD
	.dc.b	$0F,$3A,$DB,$E6,$7D,$33,$D6,$FE,$2D,$39,$C3,$FC,$D8,$D7,$9D,$70
	.dc.b	$9F,$A5,$E8,$49,$FA,$4F,$D5,$F8,$DC,$47,$C8,$8B,$8F,$AB,$75,$7F
	.dc.b	$0A,$FA,$1F,$F6,$CD,$FF,$95,$1A,$FF,$9C,$DB,$D4,$69,$FE,$A7,$F3
	.dc.b	$DD,$1F,$E6,$A8,$C3,$6B,$FD,$95,$E9,$FF,$B5,$DF,$F6,$99,$D4,$B7
	.dc.b	$D5,$B5,$C9,$71,$7F,$C0,$1F,$43,$0B,$E7,$9F,$0F,$85,$D9,$4C,$44
	.dc.b	$BF,$F6,$E4,$E9,$B3,$95,$A7,$B1,$0F,$A6,$7B,$DE,$58,$C3,$63,$31
	.dc.b	$FE,$E9,$E6,$FE,$03,$5C,$E3,$67,$C2,$9D,$BE,$0D,$FF,$DA,$51,$89
	.dc.b	$F1,$66,$BE,$75,$FE,$98,$FA,$4F,$AA,$4C,$3F,$C4,$4D,$7A,$1F,$E9
	.dc.b	$07,$D1,$B9,$F9,$7C,$0F,$A8,$CC,$4A,$FA,$AE,$A3,$E0,$7D,$1B,$E9
	.dc.b	$29,$2B,$E6,$FA,$15,$F0,$89,$E0,$FE,$09,$AF,$9B,$4F,$B9,$3F,$81
	.dc.b	$F1,$9F,$42,$F7,$70,$DB,$FA,$BC,$1F,$5E,$B7,$5F,$83,$D3,$6C,$AD
	.dc.b	$EF,$FE,$3A,$C3,$94,$78,$79,$3E,$23,$A3,$FC,$45,$F5,$FF,$41,$BE
	.dc.b	$A7,$96,$31,$AF,$F2,$9F,$AB,$C4,$CA,$E5,$A2,$F7,$D8,$BF,$CD,$B6
	.dc.b	$BC,$18,$73,$6C,$65,$F8,$CB,$D6,$E3,$52,$76,$F4,$8F,$37,$C0,$C9
	.dc.b	$F1,$E0,$8F,$0F,$8B,$99,$FC,$3E,$5E,$79,$C7,$0F,$E4,$B2,$FE,$45
	.dc.b	$5C,$BE,$93,$E1,$11,$F3,$FC,$65,$79,$E4,$9C,$5F,$3A,$BC,$3F,$E6
	.dc.b	$CA,$D5,$F5,$39,$94,$E7,$7C,$1F,$E0,$B8,$2F,$EA,$7C,$34,$9F,$7B
	.dc.b	$F3,$BE,$BB,$11,$E4,$8B,$97,$B6,$7F,$79,$F9,$6F,$BA,$BF,$80,$91
	.dc.b	$C3,$80,$E1,$D6,$DF,$CA,$A3,$95,$7F,$01,$6E,$9A,$AF,$E4,$7D,$3F
	.dc.b	$8D,$79,$DF,$88,$8E,$B8,$FF,$88,$7D,$1B,$E9,$3F,$9B,$B6,$7F,$9D
	.dc.b	$B5,$F3,$07,$0F,$A5,$D6,$5F,$C5,$5B,$8F,$C0,$6C,$FF,$22,$FA,$5F
	.dc.b	$F5,$53,$FB,$C8,$CB,$D6,$FC,$A7,$F2,$59,$DD,$F5,$8C,$CF,$F4,$2A
	.dc.b	$E3,$5A,$BF,$9F,$EE,$3E,$03,$F9,$77,$8B,$FE,$A1,$CB,$F9,$53,$6B
	.dc.b	$73,$FC,$CE,$B8,$63,$FD,$39,$AF,$F1,$26,$DF,$2F,$81,$E1,$52,$B8
	.dc.b	$3E,$08,$FE,$7D,$F4,$5E,$E2,$69,$FD,$4B,$F2,$5F,$F1,$45,$FF,$5C
	.dc.b	$97,$D2,$F0,$24,$7F,$60,$FA,$57,$C8,$7F,$49,$F8,$17,$D2,$7F,$50
	.dc.b	$AF,$F6,$C7,$D2,$7E,$42,$73,$B7,$F5,$2F,$5D,$39,$BE,$11,$8B,$FE
	.dc.b	$AF,$7C,$1E,$67,$E5,$FD,$4E,$57,$F8,$46,$C6,$4D,$9F,$ED,$42,$FC
	.dc.b	$CF,$F8,$23,$8F,$FB,$CB,$63,$FA,$6F,$C3,$38,$9F,$55,$FD,$A7,$5F
	.dc.b	$EC,$17,$B6,$F7,$E7,$1F,$D6,$CF,$EE,$5F,$E4,$3F,$89,$BE,$1B,$59
	.dc.b	$FF,$72,$AD,$3F,$D3,$3E,$75,$F2,$7A,$84,$63,$0F,$E7,$5F,$3C,$F9
	.dc.b	$FF,$F0,$3C,$7F,$84,$A7,$55,$CE,$F9,$46,$34,$FF,$01,$61,$C7,$DA
	.dc.b	$CD,$4C,$3F,$A3,$64,$F9,$FF,$F0,$2C,$75,$9F,$4C,$E7,$DB,$F3,$7C
	.dc.b	$F7,$AB,$2F,$F0,$23,$6F,$DD,$F5,$EF,$9D,$DF,$14,$9F,$F1,$4D,$6F
	.dc.b	$76,$4C,$38,$56,$AF,$85,$7F,$72,$BE,$17,$BC,$33,$8F,$8A,$FF,$C2
	.dc.b	$D7,$FC,$82,$FE,$7B,$FC,$7D,$AE,$63,$6F,$F8,$9B,$F2,$AF,$C0,$E3
	.dc.b	$FE,$3E,$6C,$7F,$84,$B1,$F3,$7C,$BF,$D7,$9C,$E3,$FE,$25,$6D,$FF
	.dc.b	$30,$6D,$FB,$B3,$5F,$1E,$59,$7F,$E5,$AB,$EF,$7F,$CD,$5A,$F3,$9F
	.dc.b	$3F,$A5,$FD,$CF,$E3,$3F,$1A,$DB,$FE,$DF,$F2,$98,$F4,$FF,$D3,$B1
	.dc.b	$FD,$A2,$B0,$7E,$53,$D7,$FF,$54,$56,$27,$FB,$6F,$C6,$71,$35,$F7
	.dc.b	$BF,$E6,$68,$C9,$33,$FE,$CB,$3A,$8E,$06,$77,$FF,$0E,$64,$9D,$4E
	.dc.b	$4F,$ED,$0F,$0A,$FF,$15,$5F,$77,$AC,$FF,$80,$3E,$23,$D8,$79,$3E
	.dc.b	$9A,$C6,$49,$3C,$A3,$2F,$4F,$9B,$F5,$9E,$CF,$3A,$D7,$F8,$63,$04
	.dc.b	$E2,$BC,$75,$73,$F5,$FB,$9F,$34,$C6,$1F,$E0,$17,$0F,$F8,$59,$CD
	.dc.b	$7F,$8C,$9F,$AF,$D1,$FE,$D6,$9F,$F1,$6B,$67,$FC,$95,$F4,$1F,$E5
	.dc.b	$B9,$CF,$C6,$F0,$23,$FE,$2E,$8D,$4F,$F8,$82,$B9,$61,$FE,$77,$F9
	.dc.b	$27,$C9,$7F,$0D,$6E,$1C,$7E,$5F,$13,$23,$FD,$3B,$3A,$C3,$C7,$59
	.dc.b	$7F,$CC,$7E,$C3,$87,$F8,$C3,$6F,$FD,$4F,$F3,$78,$99,$1F,$EA,$89
	.dc.b	$D5,$7F,$99,$FE,$29,$7A,$8E,$B7,$FB,$45,$5C,$1F,$F0,$C3,$5F,$F0
	.dc.b	$EB,$5F,$F4,$FF,$90,$F9,$4F,$E5,$8D,$C7,$FD,$39,$78,$99,$1F,$EB
	.dc.b	$C9,$D5,$7F,$9E,$FE,$37,$CF,$B7,$FA,$86,$72,$FF,$A7,$38,$3D,$37
	.dc.b	$F4,$FE,$6F,$F5,$91,$8C,$3E,$DC,$63,$4C,$B6,$B8,$1F,$17,$E1,$BE
	.dc.b	$59,$FB,$5D,$93,$89,$D1,$2F,$E5,$46,$5F,$AB,$5F,$27,$6F,$F3,$75
	.dc.b	$7F,$B9,$9B,$3F,$E2,$C6,$BF,$E4,$6E,$C7,$43,$E5,$7F,$D9,$5E,$8B
	.dc.b	$E5,$FF,$D9,$75,$A9,$CC,$62,$BE,$4D,$6E,$FF,$DC,$31,$CF,$1F,$F6
	.dc.b	$9C,$E7,$0E,$51,$8B,$E5,$87,$4C,$38,$F1,$5B,$DC,$30,$ED,$EA,$BF
	.dc.b	$20,$FF,$0C,$E5,$CE,$93,$9F,$6B,$97,$2F,$F8,$07,$0E,$38,$70,$BE
	.dc.b	$9A,$7F,$85,$E3,$13,$AB,$E3,$F2,$BE,$5C,$7A,$65,$DF,$37,$CE,$FE
	.dc.b	$D3,$0F,$73,$72,$F9,$2F,$F8,$A2,$B8,$39,$CD,$7C,$BE,$4B,$E0,$4E
	.dc.b	$63,$AF,$B7,$89,$97,$C9,$DC,$62,$FD,$67,$D2,$FB,$2F,$95,$C5,$FE
	.dc.b	$0A,$8C,$E1,$F8,$09,$CF,$A9,$A6,$E3,$ED,$B4,$62,$F8,$C3,$E0,$FA
	.dc.b	$9D,$BE,$CA,$DF,$18,$B5,$C2,$6E,$DF,$87,$8F,$DF,$F9,$5F,$81,$6D
	.dc.b	$FC,$DF,$05,$79,$E8,$5D,$25,$6E,$6D,$FC,$23,$99,$FE,$06,$AE,$32
	.dc.b	$8C,$52,$30,$7F,$10,$71,$10,$79,$8F,$9E,$FD,$0F,$1F,$38,$C7,$4F
	.dc.b	$E6,$D7,$CF,$F8,$2F,$EA,$C6,$B1,$FC,$C3,$63,$38,$74,$7B,$D7,$1D
	.dc.b	$7F,$5D,$5B,$6D,$6E,$9F,$2E,$F2,$9E,$FF,$6A,$D4,$E6,$31,$9F,$D1
	.dc.b	$8C,$5F,$8A,$31,$7C,$A3,$19,$FE,$BD,$EF,$19,$C7,$C7,$5C,$30,$D6
	.dc.b	$1D,$27,$F0,$0D,$D3,$E4,$65,$E4,$C3,$94,$7F,$B6,$A7,$58,$FD,$76
	.dc.b	$E7,$F0,$67,$55,$C2,$B4,$D7,$CB,$8F,$BB,$3F,$1C,$E7,$0E,$B5,$A3
	.dc.b	$DA,$7E,$25,$C9,$23,$B6,$1E,$19,$EF,$7D,$08,$F2,$CB,$DF,$5D,$C9
	.dc.b	$3E,$39,$ED,$18,$6C,$66,$78,$BD,$1B,$1D,$71,$F8,$33,$96,$C6,$4F
	.dc.b	$29,$D0,$93,$F5,$B1,$E1,$1A,$8E,$A6,$5D,$CF,$48,$CB,$F6,$33,$FB
	.dc.b	$D3,$A3,$94,$66,$BA,$C7,$CD,$AE,$2F,$EA,$B3,$FC,$06,$1A,$CF,$EE
	.dc.b	$C6,$27,$A1,$C7,$0E,$8E,$27,$EA,$57,$C7,$9F,$31,$27,$C6,$77,$C7
	.dc.b	$CF,$87,$8A,$3A,$19,$C7,$C5,$97,$D3,$9F,$84,$6F,$1E,$F5,$F3,$4E
	.dc.b	$24,$7D,$37,$DB,$0F,$14,$F6,$CF,$F1,$0E,$F0,$ED,$3E,$CC,$3B,$46
	.dc.b	$27,$C8,$7B,$DC,$57,$2C,$3C,$73,$FC,$6C,$FB,$E5,$3F,$38,$57,$F7
	.dc.b	$AE,$7E,$05,$E4,$3E,$7F,$84,$FF,$BD,$23,$0E,$27,$73,$A8,$46,$1E
	.dc.b	$35,$FE,$0B,$EC,$D3,$63,$3E,$0F,$F7,$66,$CE,$A3,$19,$FF,$06,$6E
	.dc.b	$9F,$C0,$DB,$5F,$CC,$BE,$7F,$FD,$E0,$FF,$0E,$8C,$BF,$C8,$9B,$87
	.dc.b	$86,$FF,$94,$D7,$0A,$FD,$F5,$68,$98,$7F,$34,$9D,$57,$8E,$3B,$12
	.dc.b	$BF,$9D,$CE,$B0,$F3,$4E,$6B,$85,$6B,$C9,$FA,$D1,$82,$57,$09,$D1
	.dc.b	$C5,$C4,$6A,$3E,$04,$F2,$9C,$92,$3F,$93,$61,$F1,$B5,$63,$E5,$9C
	.dc.b	$E3,$D5,$A7,$80,$F8,$6E,$5F,$94,$FC,$3C,$3C,$11,$C3,$1F,$23,$A3
	.dc.b	$93,$8C,$BE,$86,$5E,$5C,$7C,$77,$CA,$F8,$5E,$F2,$F4,$65,$E7,$9C
	.dc.b	$E1,$ED,$69,$F0,$8F,$36,$3D,$9F,$24,$7F,$73,$9E,$53,$99,$E2,$72
	.dc.b	$C7,$C6,$F2,$BF,$C0,$BE,$AB,$F0,$57,$13,$B5,$70,$9F,$8E,$4C,$3E
	.dc.b	$B4,$65,$F6,$65,$F3,$E3,$17,$D3,$0E,$F5,$F9,$4C,$BE,$26,$3D,$E7
	.dc.b	$35,$EB,$9C,$C6,$31,$F7,$9F,$0C,$F1,$4F,$5C,$7E,$C4,$7B,$5F,$16
	.dc.b	$5F,$12,$B9,$4F,$C3,$8C,$4E,$E3,$E1,$1D,$8F,$93,$87,$C2,$9F,$C6
	.dc.b	$65,$EC,$9D,$4E,$71,$FC,$3C,$E6,$78,$E1,$F7,$A7,$37,$FB,$3E,$0F
	.dc.b	$E3,$57,$F3,$5B,$19,$BF,$84,$7F,$8D,$8A,$79,$A7,$C4,$3E,$7F,$FD
	.dc.b	$97,$1F,$E1,$2E,$9F,$CC,$F6,$BF,$1E,$7C,$2E,$1B,$5B,$21,$FE,$41
	.dc.b	$3F,$C5,$B5,$FC,$BF,$2F,$E5,$8D,$8F,$F0,$C3,$5F,$F0,$C3,$F7,$08
	.dc.b	$FF,$97,$7D,$2F,$1F,$C9,$BD,$2A,$49,$FF,$47,$43,$D7,$C4,$7F,$9E
	.dc.b	$A3,$83,$A9,$E7,$3D,$3D,$2E,$BE,$B4,$AE,$71,$97,$AC,$65,$E9,$1F
	.dc.b	$1B,$E3,$FF,$D8,$91,$FE,$9F,$6B,$FE,$8A,$6F,$3F,$FA,$F5,$FF,$79
	.dc.b	$46,$F9,$39,$11,$E5,$D9,$5E,$51,$6B,$94,$FF,$0E,$E7,$7D,$07,$BA
	.dc.b	$FF,$D3,$3E,$73,$D7,$FB,$6D,$8F,$F6,$15,$70,$DB,$95,$F1,$C7,$FD
	.dc.b	$92,$DE,$BF,$F4,$15,$6E,$FF,$D7,$98,$3E,$83,$EA,$FA,$F2,$7E,$9B
	.dc.b	$F6,$3C,$FC,$47,$D3,$8B,$97,$FB,$7F,$8D,$5F,$BC,$FA,$0F,$EE,$DE
	.dc.b	$C7,$FA,$33,$AD,$E5,$BE,$3F,$F2,$AD,$CF,$FC,$D3,$7C,$3A,$FF,$CC
	.dc.b	$8F,$F7,$18,$FF,$3F,$B7,$C2,$34,$91,$FE,$9F,$C7,$FD,$35,$3A,$8C
	.dc.b	$57,$46,$9F,$ED,$4F,$5D,$F4,$1E,$18,$FF,$6B,$FC,$D3,$D2,$C6,$A8
	.dc.b	$FF,$62,$4F,$28,$C7,$81,$A9,$E1,$39,$6A,$CF,$E2,$47,$28,$C6,$5F
	.dc.b	$EB,$7B,$E7,$1C,$70,$F2,$2F,$A0,$7C,$3F,$93,$99,$ED,$53,$73,$AA
	.dc.b	$F7,$9C,$6B,$24,$7D,$55,$FC,$29,$F1,$9F,$E9,$3F,$9E,$F8,$9E,$7B
	.dc.b	$EC,$FA,$4E,$A3,$17,$E6,$E2,$5D,$3E,$C1,$2B,$FC,$4E,$D7,$B9,$FE
	.dc.b	$BC,$78,$3F,$37,$89,$EC,$FC,$2F,$A1,$7C,$CF,$D4,$F8,$17,$D4,$D3
	.dc.b	$51,$30,$F8,$35,$C1,$B1,$F5,$FE,$8B,$F0,$1E,$83,$E8,$5F,$45,$F8
	.dc.b	$0F,$03,$FC,$9D,$D5,$F9,$B6,$F4,$FF,$B8,$EB,$79,$7E,$33,$0E,$58
	.dc.b	$74,$9C,$CF,$CC,$9C,$BF,$85,$E0,$F0,$FC,$47,$D1,$8F,$E8,$AD,$7F
	.dc.b	$DE,$0D,$7F,$DE,$AD,$93,$F6,$72,$FF,$7D,$C6,$3D,$DE,$A7,$AF,$CA
	.dc.b	$97,$F9,$9D,$3F,$86,$3F,$80,$EC,$46,$30,$F8,$1D,$7C,$2B,$53,$9D
	.dc.b	$BF,$30,$D8,$5F,$2A,$F6,$FD,$17,$F9,$AF,$3F,$3B,$74,$34,$97,$F2
	.dc.b	$E3,$2D,$8F,$A9,$E7,$7C,$8F,$A2,$E4,$49,$FF,$3E,$DF,$C4,$8C,$47
	.dc.b	$D1,$B4,$71,$8F,$85,$F2,$15,$D4,$5D,$7F,$01,$37,$5F,$A7,$36,$BD
	.dc.b	$FA,$AB,$ED,$FD,$17,$B4,$93,$E4,$C5,$18,$8F,$23,$C1,$FA,$9F,$45
	.dc.b	$FC,$03,$E8,$9A,$FF,$02,$E7,$F8,$75,$FA,$86,$EA,$D7,$FD,$4D,$7C
	.dc.b	$2B,$F8,$0B,$5F,$F5,$0B,$5F,$F5,$53,$5E,$37,$FC,$C6,$38,$57,$59
	.dc.b	$08,$CD,$70,$8F,$E1,$F7,$F9,$38,$FF,$5C,$7D,$17,$F3,$F9,$FD,$FF
	.dc.b	$D1,$3E,$3F,$F8,$EB,$63,$F8,$BC,$EF,$1F,$04,$23,$93,$FE,$C1,$6B
	.dc.b	$DB,$CC,$FE,$91,$18,$8F,$3B,$CF,$0D,$47,$69,$F0,$FA,$3F,$CB,$AB
	.dc.b	$53,$C2,$70,$69,$FE,$57,$F0,$6F,$A2,$FE,$6B,$E6,$7F,$AB,$F2,$F3
	.dc.b	$B7,$2F,$F6,$C5,$6E,$F8,$DF,$F5,$D8,$C4,$7F,$B5,$1A,$FF,$B7,$5A
	.dc.b	$F2,$8F,$E8,$71,$C2,$FE,$2B,$6B,$FD,$CD,$6F,$9E,$FE,$93,$F4,$3F
	.dc.b	$0F,$13,$49,$97,$F6,$F5,$F1,$F4,$93,$FE,$E3,$8C,$BF,$EE,$5F,$85
	.dc.b	$7C,$77,$F5,$D8,$6F,$FE,$4E,$7F,$DD,$31,$98,$C4,$F0,$D7,$FC,$2B
	.dc.b	$F0,$7F,$EF,$D8,$C5,$F2,$8F,$F7,$BF,$AF,$A8,$FF,$80,$0D,$7F,$D9
	.dc.b	$E6,$CC,$94,$7C,$C7,$F8,$5B,$7F,$F7,$E3,$63,$F3,$64,$70,$E0,$E3
	.dc.b	$EC,$7F,$99,$5F,$D0,$C7,$EF,$E7,$9F,$66,$67,$BB,$67,$51,$CC,$7C
	.dc.b	$EB,$E6,$FF,$59,$A7,$E7,$A3,$18,$7F,$85,$BE,$75,$F3,$DF,$E1,$38
	.dc.b	$FE,$F7,$3A,$C3,$FB,$4B,$5F,$43,$5F,$6B,$5F,$DC,$B5,$F1,$E1,$E2
	.dc.b	$9F,$6F,$CF,$7F,$83,$9E,$F1,$99,$F9,$73,$97,$FB,$CF,$CF,$7B,$B2
	.dc.b	$FD,$BB,$6F,$E3,$FC,$F7,$EF,$DF,$F0,$9C,$66,$BF,$85,$B5,$39,$D1
	.dc.b	$5C,$67,$BE,$BE,$37,$6F,$D1,$AF,$9F,$1F,$65,$AF,$DE,$6B,$E0,$8F
	.dc.b	$66,$0F,$A8,$7D,$4F,$5B,$F3,$4E,$63,$C1,$F3,$EF,$9F,$A7,$C5,$CF
	.dc.b	$CF,$F4,$18,$D7,$FC,$63,$5B,$7F,$05,$A2,$30,$FE,$15,$AF,$E4,$5A
	.dc.b	$FE,$59,$DE,$3F,$CC,$FD,$C7,$E5,$6F,$88,$F0,$5F,$F8,$D6,$31,$87
	.dc.b	$F6,$BF,$8F,$FB,$5E,$7F,$F8,$21,$FB,$9E,$97,$F8,$03,$E6,$73,$E7
	.dc.b	$FF,$AC,$EF,$81,$FD,$66,$BF,$45,$7C,$63,$07,$F5,$DF,$90,$E2,$74
	.dc.b	$F9,$BF,$EA,$91,$FD,$EE,$32,$D3,$FC,$D5,$1A,$CF,$FB,$B4,$E9,$B3
	.dc.b	$A3,$1B,$7F,$87,$B1,$57,$91,$B3,$E3,$25,$7E,$EF,$0F,$F1,$3F,$C5
	.dc.b	$FB,$8F,$37,$D0,$DA,$FF,$08,$47,$32,$3F,$DE,$5F,$E0,$5A,$B3,$7C
	.dc.b	$9F,$BC,$F8,$1E,$5D,$BF,$6C,$5C,$BF,$C5,$D4,$AC,$6B,$EB,$F6,$7E
	.dc.b	$89,$8C,$3F,$C8,$2E,$1F,$F0,$53,$EF,$23,$FE,$20,$F4,$3F,$DD,$CF
	.dc.b	$F9,$4D,$B8,$7F,$8F,$7E,$07,$FC,$D7,$39,$F9,$0F,$C1,$D6,$B0,$FF
	.dc.b	$2F,$7B,$98,$9F,$C2,$FB,$28,$FD,$87,$C7,$3E,$4E,$9F,$8A,$C9,$F3
	.dc.b	$38,$CF,$FD,$23,$5B,$7F,$CA,$1E,$F6,$75,$FF,$25,$BF,$BB,$C5,$C7
	.dc.b	$4F,$CA,$FC,$3B,$E8,$71,$87,$FA,$72,$B6,$FE,$B1,$B8,$7F,$2F,$6B
	.dc.b	$FC,$ED,$AF,$F4,$07,$7B,$FF,$B5,$EF,$FB,$45,$75,$C3,$C3,$EB,$7F
	.dc.b	$0F,$F9,$97,$CA,$D3,$F5,$3F,$1D,$CA,$FF,$A5,$CE,$9F,$F5,$8B,$73
	.dc.b	$FF,$37,$7C,$07,$E2,$27,$37,$FB,$28,$C4,$FD,$7D,$7E,$67,$D0,$FC
	.dc.b	$D6,$DF,$FB,$47,$8B,$E2,$36,$74,$D3,$E3,$7C,$2F,$CD,$3E,$5C,$58
	.dc.b	$C9,$F2,$63,$FA,$D7,$07,$FB,$32,$30,$4C,$3F,$D0,$DC,$4F,$89,$FF
	.dc.b	$22,$7A,$DF,$C4,$67,$FC,$8B,$7C,$5D,$CE,$BC,$6F,$F6,$F7,$6B,$E5
	.dc.b	$F1,$5F,$E8,$88,$C6,$5F,$E4,$EC,$7A,$78,$1F,$AF,$E1,$DB,$FB,$1C
	.dc.b	$2D,$F0,$C1,$7C,$23,$0D,$BD,$D7,$83,$DA,$7D,$0F,$F8,$83,$2F,$06
	.dc.b	$7F,$E5,$E9,$CF,$BD,$E0,$E6,$FF,$04,$65,$EF,$9C,$DF,$0D,$3F,$C4
	.dc.b	$38,$F3,$7C,$78,$FF,$BF,$79,$7F,$80,$56,$EF,$8D,$68,$FF,$54,$CE
	.dc.b	$9C,$6C,$FA,$1F,$DC,$C6,$1B,$97,$C6,$7F,$C6,$78,$78,$EB,$87,$C7
	.dc.b	$FC,$89,$DE,$5F,$E7,$A9,$CD,$6E,$78,$E1,$D6,$F8,$12,$75,$EF,$D9
	.dc.b	$F2,$F0,$BE,$87,$E9,$FC,$AF,$D4,$6D,$F8,$E7,$4D,$9D,$34,$D7,$C2
	.dc.b	$7D,$C3,$EE,$FD,$0D,$8C,$93,$0F,$0C,$65,$F9,$47,$F1,$7F,$43,$F1
	.dc.b	$AD,$9F,$89,$D2,$AC,$F4,$AE,$9F,$E9,$FA,$FE,$10,$DB,$FE,$10,$D7
	.dc.b	$F3,$ED,$7B,$1F,$63,$E7,$3E,$5F,$17,$E9,$63,$2F,$E7,$23,$3C,$5F
	.dc.b	$EA,$CF,$39,$F4,$1F,$C1,$EB,$F5,$91,$8E,$BF,$D3,$AF,$A1,$F4,$67
	.dc.b	$F9,$18,$D3,$6F,$F8,$83,$FE,$5E,$8C,$E5,$FE,$AC,$6C,$66,$77,$3E
	.dc.b	$37,$C5,$8F,$F9,$9B,$E5,$BF,$AC,$7A,$DD,$72,$F9,$B1,$8B,$FC,$D6
	.dc.b	$3D,$B3,$F8,$13,$97,$D7,$8F,$4A,$FF,$65,$CE,$71,$ED,$E8,$32,$FD
	.dc.b	$65,$2B,$4B,$1E,$D7,$C2,$B7,$5A,$AE,$17,$CD,$F2,$E1,$E1,$73,$5C
	.dc.b	$27,$58,$FA,$1B,$87,$C3,$CB,$DF,$18,$BF,$4B,$5E,$67,$68,$DD,$F5
	.dc.b	$24,$67,$3B,$87,$79,$B8,$77,$9B,$5D,$A6,$E7,$F0,$23,$07,$E8,$33
	.dc.b	$F8,$07,$C6,$8D,$47,$93,$0E,$CE,$F2,$F5,$4E,$AB,$8C,$7A,$DE,$59
	.dc.b	$FC,$77,$38,$78,$A3,$13,$B2,$47,$E1,$71,$F4,$E1,$E1,$CB,$EB,$3F
	.dc.b	$04,$F1,$4E,$B0,$EF,$9F,$C8,$9C,$93,$3F,$C4,$C6,$5F,$59,$2F,$E4
	.dc.b	$4E,$DE,$F1,$98,$FC,$46,$5E,$8C,$3C,$19,$7C,$28,$C4,$F1,$3E,$6D
	.dc.b	$78,$32,$DD,$FC,$1C,$7C,$58,$F6,$25,$F6,$8C,$BC,$A3,$39,$FC,$FC
	.dc.b	$BE,$3C,$F7,$8C,$5F,$68,$F8,$D8,$FA,$5F,$94,$73,$3B,$63,$B8,$F7
	.dc.b	$57,$29,$F5,$1C,$1C,$47,$3C,$FE,$7D,$77,$AF,$F7,$04,$7E,$ED,$B8
	.dc.b	$7E,$A5,$B8,$7F,$C0,$1F,$2D,$F4,$34,$93,$FE,$DF,$F8,$8F,$F2,$1C
	.dc.b	$7F,$A8,$30,$FF,$00,$77,$5F,$08,$D2,$5F,$FB,$9E,$76,$FE,$76,$76
	.dc.b	$D9,$FE,$41,$F4,$2F,$27,$FD,$F7,$B7,$F1,$0B,$DB,$6B,$F8,$2C,$63
	.dc.b	$83,$F9,$85,$F1,$C3,$35,$1E,$C7,$F1,$C8,$D3,$8B,$E8,$DA,$E7,$87
	.dc.b	$89,$FE,$51,$87,$79,$FE,$3E,$D7,$F8,$E3,$9E,$06,$1D,$B0,$EA,$4B
	.dc.b	$E7,$8F,$C0,$C7,$B6,$3D,$23,$D1,$5B,$8F,$E4,$D7,$CE,$F8,$C7,$9C
	.dc.b	$E0,$7B,$6F,$9B,$E4,$BE,$73,$C2,$FF,$A9,$DF,$3A,$E3,$1D,$DE,$07
	.dc.b	$33,$89,$A9,$E4,$7C,$67,$27,$08,$F4,$61,$F0,$B0,$F1,$E5,$F8,$29
	.dc.b	$EC,$DB,$E5,$7E,$B3,$93,$D9,$F8,$35,$CB,$2F,$A1,$5C,$67,$C8,$68
	.dc.b	$F1,$B9,$79,$99,$3B,$39,$39,$61,$E7,$8E,$33,$D8,$FB,$98,$FC,$78
	.dc.b	$E6,$4A,$E3,$19,$7A,$4F,$1C,$BE,$F5,$74,$7E,$5D,$F7,$8F,$1C,$ED
	.dc.b	$FA,$06,$27,$77,$D6,$75,$97,$CE,$8C,$4F,$5A,$E3,$3F,$3A,$B7,$97
	.dc.b	$CE,$8C,$57,$4C,$BE,$63,$C6,$3F,$0B,$1F,$94,$C7,$CD,$1C,$30,$EB
	.dc.b	$7C,$CF,$AC,$FC,$E9,$DC,$66,$3E,$A9,$9C,$7F,$21,$3A,$8F,$A9,$5E
	.dc.b	$07,$FC,$45,$F4,$3F,$ED,$77,$FC,$07,$D5,$FA,$6F,$05,$CE,$F8,$AC
	.dc.b	$36,$B6,$43,$FC,$75,$1A,$3F,$27,$52,$3F,$CD,$D3,$C5,$D9,$C7,$0F
	.dc.b	$F4,$CF,$A4,$F4,$FC,$4D,$9F,$F3,$B3,$5F,$F3,$FA,$F6,$B7,$32,$76
	.dc.b	$F2,$8E,$7E,$A7,$4F,$6A,$57,$48,$CB,$DA,$32,$F6,$8F,$89,$E4,$7F
	.dc.b	$AC,$E3,$FD,$10,$D7,$FD,$22,$DC,$7F,$D6,$6F,$FB,$8A,$37,$C9,$C8
	.dc.b	$8F,$2E,$CA,$F2,$8B,$5C,$A7,$F8,$CF,$2B,$E8,$7E,$14,$FF,$B0,$3C
	.dc.b	$B7,$B1,$F6,$5B,$8F,$FA,$A6,$B7,$CF,$FE,$B7,$6C,$39,$15,$C2,$3F
	.dc.b	$D5,$6D,$C3,$FD,$75,$5C,$29,$C2,$F5,$3E,$B9,$30,$FA,$FF,$17,$5F
	.dc.b	$AF,$17,$0F,$BF,$3F,$C6,$3E,$81,$53,$D8,$78,$7F,$EB,$1F,$25,$F1
	.dc.b	$DF,$95,$6D,$7F,$96,$AF,$8A,$BE,$2F,$F9,$A2,$33,$18,$D1,$5C,$AB
	.dc.b	$FD,$92,$DC,$BF,$CE,$77,$C3,$1F,$F3,$E3,$6B,$E4,$56,$F7,$F3,$65
	.dc.b	$EB,$9F,$55,$FF,$B2,$7E,$7D,$E6,$71,$CD,$0E,$CC,$36,$F8,$9F,$55
	.dc.b	$E2,$E2,$78,$D7,$0A,$93,$9A,$DB,$F0,$EF,$E2,$FC,$FB,$E7,$BC,$5C
	.dc.b	$8F,$5E,$9E,$DB,$56,$DF,$19,$8A,$DB,$E3,$AE,$4E,$A7,$83,$E1,$3F
	.dc.b	$D3,$3F,$35,$F2,$7C,$A7,$CF,$EB,$93,$EB,$C6,$31,$F1,$79,$0B,$A7
	.dc.b	$1B,$E1,$A7,$E2,$2F,$86,$3E,$6B,$E0,$DB,$E1,$84,$8C,$1E,$CF,$9E
	.dc.b	$7C,$C7,$D5,$F7,$5F,$3F,$4F,$89,$3A,$AE,$15,$97,$EA,$57,$07,$27
	.dc.b	$DA,$F2,$9F,$3C,$F9,$FF,$C1,$77,$FE,$62,$BE,$37,$47,$EB,$1F,$C7
	.dc.b	$5F,$0A,$B7,$C6,$7F,$05,$7C,$78,$BF,$38,$DE,$1F,$F7,$35,$6E,$BE
	.dc.b	$96,$0F,$88,$FE,$81,$1E,$0F,$67,$F4,$58,$FE,$95,$B8,$7F,$BB,$EB
	.dc.b	$73,$FA,$CB,$E5,$EC,$7E,$6E,$F7,$87,$F0,$B6,$CE,$A3,$CF,$E2,$7C
	.dc.b	$02,$3E,$AE,$9A,$FC,$2D,$3F,$89,$63,$EC,$AF,$8F,$5F,$13,$27,$C2
	.dc.b	$F8,$7E,$36,$63,$F3,$E7,$2D,$9C,$BE,$F9,$E5,$1C,$FE,$7F,$D5,$E1
	.dc.b	$CA,$E5,$19,$7F,$61,$AA,$75,$1F,$33,$05,$6E,$30,$DC,$FE,$AF,$C9
	.dc.b	$5E,$0F,$99,$36,$BF,$99,$6C,$AF,$77,$CF,$FC,$DF,$85,$95,$ED,$8C
	.dc.b	$BE,$48,$CB,$E3,$8F,$DD,$7C,$F3,$CF,$FD,$E7,$4F,$C4,$AF,$DA,$1F
	.dc.b	$A0,$E0,$DC,$C8,$FD,$FC,$EF,$5F,$E5,$71,$B3,$F8,$1D,$F5,$6D,$76
	.dc.b	$B9,$18,$3F,$84,$FC,$EF,$E1,$97,$CF,$3E,$37,$F8,$83,$63,$7E,$4F
	.dc.b	$F1,$3A,$E1,$39,$9F,$4B,$72,$F2,$CF,$53,$F9,$E4,$FE,$B5,$AF,$F4
	.dc.b	$66,$BF,$D4,$1A,$9E,$55,$E1,$3C,$BE,$87,$F2,$5F,$7D,$F3,$DF,$C9
	.dc.b	$FC,$BF,$F4,$FE,$3F,$19,$BC,$BF,$D5,$2B,$79,$7F,$2A,$6B,$F4,$A7
	.dc.b	$6F,$F6,$26,$A8,$DD,$7D,$26,$D7,$5B,$F2,$1E,$3F,$8D,$FE,$75,$F3
	.dc.b	$DF,$B4,$F2,$5F,$1B,$F6,$B1,$FE,$5E,$DC,$3F,$B4,$D7,$D5,$1F,$1B
	.dc.b	$FD,$0F,$E9,$3F,$90,$69,$F8,$86,$E1,$F8,$4A,$DC,$FF,$46,$F8,$47
	.dc.b	$C5,$FF,$73,$8F,$F6,$CF,$C0,$FF,$76,$9C,$DF,$F8,$03,$87,$FC,$89
	.dc.b	$18,$AD,$9F,$E0,$47,$11,$FE,$49,$8F,$F0,$3F,$AF,$C6,$FF,$95,$42
	.dc.b	$30,$FF,$2D,$6B,$FE,$E0,$3A,$8F,$96,$7B,$3F,$E0,$8C,$7F,$A9,$FD
	.dc.b	$77,$F4,$2D,$FF,$B0,$D6,$9F,$EE,$BC,$29,$D6,$1F,$AC,$AF,$DD,$56
	.dc.b	$BF,$3F,$FF,$0E,$BC,$BC,$1E,$CF,$F5,$98,$CD,$7E,$D3,$E7,$DF,$3F
	.dc.b	$A8,$FF,$12,$C6,$2B,$FC,$0B,$F4,$2F,$A1,$FF,$06,$C7,$F7,$69,$D5
	.dc.b	$73,$BE,$59,$3E,$87,$FC,$05,$7B,$9C,$C6,$2B,$84,$FF,$84,$AB,$87
	.dc.b	$AF,$E3,$96,$1D,$B7,$FF,$17,$D6,$AF,$FC,$0F,$F3,$FF,$0A,$FF,$87
	.dc.b	$B6,$B7,$F4,$3F,$C0,$6B,$7B,$F5,$AE,$17,$FD,$95,$AF,$F8,$F9,$AF
	.dc.b	$F9,$21,$A9,$ED,$3F,$E4,$BF,$A1,$7D,$0F,$5B,$F9,$98,$A7,$35,$E2
	.dc.b	$3F,$C6,$31,$9A,$E1,$18,$9D,$E1,$FC,$AB,$E8,$1E,$97,$D7,$E5,$7D
	.dc.b	$03,$E1,$7E,$45,$6B,$1F,$C4,$CE,$5F,$F1,$6C,$67,$0F,$F1,$45,$F2
	.dc.b	$BE,$38,$75,$C3,$85,$78,$63,$13,$AC,$3A,$BF,$89,$F5,$FF,$C0,$1F
	.dc.b	$1B,$F8,$BF,$71,$EB,$FE,$FA,$FF,$B4,$4E,$63,$18,$FF,$5D,$8C,$78
	.dc.b	$9C,$FD,$6F,$F5,$85,$FF,$64,$8C,$7C,$6C,$8F,$EC,$E5,$BF,$EC,$D1
	.dc.b	$78,$BF,$C1,$D3,$4F,$F0,$74,$78,$7E,$07,$15,$FE,$0B,$6E,$7E,$86
	.dc.b	$E7,$FE,$45,$BF,$F0,$07,$C4,$FA,$7E,$B3,$CC,$D8,$C9,$FD,$EA,$32
	.dc.b	$D3,$FB,$EF,$D0,$7C,$13,$E1,$79,$D6,$32,$4C,$BF,$C5,$11,$97,$FC
	.dc.b	$05,$EF,$7C,$9F,$B4,$F8,$CD,$AF,$F0,$A3,$67,$71,$FE,$0C,$9F,$AD
	.dc.b	$EA,$FF,$B8,$63,$FC,$7B,$F0,$18,$F9,$1F,$B1,$EF,$23,$13,$A3,$FC
	.dc.b	$9D,$5C,$70,$FF,$15,$4F,$ED,$FE,$35,$F2,$3C,$BD,$BC,$37,$0F,$F3
	.dc.b	$EC,$E9,$FF,$26,$CE,$9A,$7F,$A0,$AB,$FD,$80,$FF,$A4,$FE,$85,$EB
	.dc.b	$FF,$05,$9C,$E3,$FE,$5C,$9C,$D7,$1E,$2F,$C9,$46,$27,$FC,$27,$5C
	.dc.b	$39,$FF,$D6,$D3,$97,$FC,$23,$19,$BF,$F6,$3F,$32,$B5,$1F,$EC,$29
	.dc.b	$D5,$FF,$B7,$67,$2F,$FB,$3B,$E5,$9F,$25,$FA,$3F,$7D,$CD,$FC,$76
	.dc.b	$73,$8F,$FA,$BE,$72,$FF,$A8,$A3,$37,$FE,$78,$9C,$BF,$E2,$C8,$CE
	.dc.b	$7F,$EC,$F9,$CB,$FE,$D1,$F8,$97,$D0,$FE,$CD,$FE,$EB,$87,$FA,$2A
	.dc.b	$31,$EE,$65,$B1,$93,$FD,$1F,$1F,$3F,$E2,$3F,$79,$C6,$F8,$6F,$F7
	.dc.b	$0F,$B1,$FC,$47,$5F,$F7,$6F,$0F,$FA,$FF,$1F,$F6,$55,$ED,$FE,$D7
	.dc.b	$E2,$FC,$9C,$79,$E3,$CA,$31,$3C,$23,$18,$74,$AD,$CF,$28,$C3,$B3
	.dc.b	$1C,$7F,$C4,$3E,$53,$8B,$63,$3E,$07,$CC,$D3,$D1,$39,$BF,$EF,$99
	.dc.b	$7E,$02,$98,$72,$6E,$1C,$B9,$31,$3E,$6F,$6D,$F4,$3F,$C7,$33,$EB
	.dc.b	$C9,$FE,$2D,$C3,$D1,$5A,$F1,$7C,$B3,$9D,$3C,$22,$75,$5E,$C9,$D5
	.dc.b	$FC,$F9,$CD,$70,$D7,$FC,$E5,$E1,$7C,$0F,$A1,$7D,$0D,$31,$F0,$17
	.dc.b	$3F,$8F,$3A,$6C,$E9,$A4,$C3,$B4,$ED,$B5,$CD,$F1,$70,$3E,$87,$E4
	.dc.b	$7C,$BF,$C9,$6E,$1F,$49,$B5,$C1,$BA,$FD,$4B,$7C,$37,$D1,$3E,$97
	.dc.b	$D0,$DC,$3E,$DC,$5C,$3C,$33,$6F,$F2,$91,$FC,$6F,$E8,$7F,$07,$F3
	.dc.b	$7F,$E4,$8C,$3F,$15,$39,$7F,$13,$19,$8F,$A6,$64,$F9,$CE,$6B,$AF
	.dc.b	$5A,$FA,$EB,$F9,$8B,$E1,$5F,$8E,$BE,$1D,$1F,$C2,$6F,$78,$FE,$6F
	.dc.b	$D1,$7D,$07,$E5,$69,$8A,$73,$D5,$FA,$45,$F4,$3E,$1C,$39,$61,$C6
	.dc.b	$F8,$35,$7C,$C7,$F9,$FF,$D7,$DE,$3E,$78,$C4,$FE,$4B,$1E,$F8,$7C
	.dc.b	$69,$CE,$BF,$BF,$9C,$D7,$C6,$9C,$E7,$E8,$9C,$D7,$D2,$9C,$D7,$93
	.dc.b	$0F,$07,$3B,$2E,$CD,$9E,$8D,$BF,$06,$7E,$BC,$39,$D7,$B2,$B5,$97
	.dc.b	$C8,$8C,$61,$EE,$CF,$AB,$23,$AB,$96,$93,$1F,$45,$F4,$8F,$45,$71
	.dc.b	$9C,$D5,$BF,$81,$8F,$97,$0E,$18,$78,$8E,$93,$AC,$BE,$0D,$F1,$9F
	.dc.b	$BF,$97,$C6,$CF,$E4,$DF,$AE,$3D,$E4,$C3,$C9,$3A,$AE,$31,$E0,$9D
	.dc.b	$5F,$B7,$2F,$3D,$78,$8D,$C6,$70,$ED,$7C,$49,$19,$9D,$9C,$4F,$65
	.dc.b	$7D,$4C,$7D,$71,$9C,$3D,$59,$7C,$3B,$F7,$4F,$0C,$7E,$16,$78,$BF
	.dc.b	$3C,$FE,$46,$F8,$3C,$F2,$F7,$C6,$32,$F5,$93,$0E,$75,$BC,$7E,$66
	.dc.b	$7F,$16,$3C,$07,$AA,$B8,$3F,$70,$E0,$7C,$58,$E1,$8F,$92,$77,$18
	.dc.b	$C7,$E2,$D6,$F1,$F2,$63,$F2,$B1,$F0,$C6,$27,$C1,$39,$7D,$D9,$FC
	.dc.b	$8C,$7C,$71,$E6,$C7,$B6,$1F,$17,$1F,$83,$7B,$C3,$96,$1E,$4C,$7B
	.dc.b	$63,$EE,$39,$38,$CB,$F7,$B1,$FA,$53,$F2,$85,$72,$F9,$4D,$7C,$4B
	.dc.b	$E8,$5F,$43,$FE,$E2,$94,$E6,$30,$D8,$FF,$73,$B5,$ED,$E0,$7F,$5D
	.dc.b	$7F,$DD,$77,$FD,$BB,$B7,$A3,$63,$33,$B7,$8C,$FF,$13,$8E,$59,$7F
	.dc.b	$0D,$6E,$7F,$C4,$7E,$85,$E4,$FF,$03,$AD,$5F,$12,$61,$FC,$7E,$B6
	.dc.b	$E2,$76,$FF,$1D,$C3,$AB,$FC,$CE,$76,$E6,$B8,$61,$FB,$7A,$DF,$0A
	.dc.b	$F7,$87,$42,$56,$EF,$FA,$14,$EA,$31,$87,$52,$57,$F3,$19,$D4,$7F
	.dc.b	$08,$C3,$CA,$FF,$3E,$6E,$5F,$2E,$B8,$65,$F0,$F0,$EB,$1E,$2B,$E5
	.dc.b	$8F,$8A,$B5,$18,$3C,$35,$C0,$F9,$D1,$A8,$F0,$4F,$C2,$78,$57,$81
	.dc.b	$C1,$A3,$99,$F2,$E7,$95,$F9,$EF,$B4,$F2,$7B,$BC,$A3,$4E,$4E,$46
	.dc.b	$70,$F8,$27,$D2,$C7,$E3,$C7,$87,$0F,$80,$DB,$EA,$F9,$0C,$1D,$72
	.dc.b	$FB,$58,$F9,$B0,$EB,$97,$BE,$73,$5E,$18,$C6,$3E,$0C,$BC,$F3,$D0
	.dc.b	$EB,$5C,$1F,$04,$68,$F6,$C6,$72,$F3,$C6,$70,$F6,$63,$E3,$C7,$C1
	.dc.b	$87,$E6,$E3,$0F,$49,$DB,$AB,$FC,$06,$1D,$72,$F9,$F3,$E5,$C7,$C3
	.dc.b	$97,$A7,$0F,$7C,$FD,$8B,$E0,$FD,$13,$84,$76,$7C,$19,$7C,$2C,$BF
	.dc.b	$17,$5C,$27,$73,$F8,$F6,$1C,$0F,$51,$8C,$BE,$0D,$F5,$7F,$19,$1A
	.dc.b	$23,$FE,$15,$FA,$1F,$F6,$D4,$7F,$86,$7A,$BF,$37,$E0,$3A,$1F,$1B
	.dc.b	$86,$D6,$C8,$7F,$90,$23,$47,$E5,$AA,$47,$F9,$22,$32,$F0,$C3,$FD
	.dc.b	$1B,$E9,$BD,$4F,$0B,$67,$FD,$2E,$D7,$FC,$F4,$BD,$BD,$CC,$9D,$BC
	.dc.b	$A3,$9F,$A9,$D3,$DA,$95,$D2,$32,$F6,$8C,$BD,$A3,$E1,$F8,$FF,$E9
	.dc.b	$E8,$FF,$47,$6F,$83,$3D,$56,$32,$43,$FD,$21,$3F,$ED,$E7,$19,$FF
	.dc.b	$A8,$2B,$84,$6A,$BF,$85,$F0,$BE,$87,$E0,$5F,$FA,$CB,$E8,$1E,$D7
	.dc.b	$D6,$6E,$1F,$EA,$8A,$DF,$4F,$F9,$93,$0E,$5B,$EF,$09,$7B,$8F,$9F
	.dc.b	$EB,$48,$F9,$EF,$D2,$F8,$FC,$47,$D0,$8B,$9F,$D8,$9F,$E2,$7F,$35
	.dc.b	$FE,$4B,$F5,$9E,$27,$F9,$D3,$A5,$E9,$BE,$53,$F1,$AD,$CF,$FC,$D1
	.dc.b	$87,$1B,$FF,$1D,$36,$17,$CA,$F8,$9E,$08,$FF,$40,$E1,$E0,$7F,$C8
	.dc.b	$B5,$C6,$B8,$47,$FB,$83,$0F,$03,$FE,$73,$9D,$7D,$0B,$CF,$BD,$92
	.dc.b	$BA,$E7,$EB,$AD,$46,$0F,$2B,$85,$97,$8F,$24,$FF,$AD,$A7,$84,$62
	.dc.b	$FC,$38,$7B,$E5,$F4,$2F,$8D,$F5,$73,$3D,$6A,$6E,$75,$5E,$33,$8D
	.dc.b	$69,$EF,$5C,$5D,$4F,$07,$CF,$A7,$F9,$A7,$E0,$BE,$87,$98,$FA,$1D
	.dc.b	$6F,$F3,$63,$19,$79,$FC,$85,$D3,$8D,$F0,$C3,$FD,$6B,$7C,$31,$F3
	.dc.b	$5F,$06,$DF,$0C,$24,$60,$F6,$7D,$03,$E7,$3E,$E7,$BC,$FA,$1A,$7C
	.dc.b	$99,$D5,$70,$AC,$BF,$42,$B8,$39,$3E,$D7,$94,$FA,$07,$D0,$FE,$27
	.dc.b	$C9,$FF,$21,$70,$FD,$B6,$F0,$7F,$AE,$6B,$78,$72,$6B,$F3,$0F,$CC
	.dc.b	$CF,$06,$9F,$9D,$C3,$C1,$85,$F8,$9F,$E6,$73,$C7,$A1,$ED,$7E,$63
	.dc.b	$3F,$C6,$37,$3F,$F6,$55,$F0,$9F,$D1,$62,$8C,$4F,$EC,$EB,$86,$1F
	.dc.b	$AC,$6D,$6F,$D9,$F8,$1E,$1C,$AF,$CB,$E9,$FC,$39,$F2,$3F,$8B,$6C
	.dc.b	$6D,$A7,$0E,$9F,$67,$C3,$7E,$CF,$17,$C5,$53,$35,$F4,$B4,$56,$63
	.dc.b	$F4,$F5,$C1,$D4,$7C,$7F,$A1,$E9,$E0,$49,$E5,$19,$78,$C6,$5F,$A9
	.dc.b	$AF,$C5,$9D,$46,$30,$46,$37,$FC,$07,$C8,$5E,$0F,$9F,$36,$FE,$EC
	.dc.b	$FD,$9F,$A1,$F7,$7C,$2C,$9F,$64,$65,$F5,$C6,$5F,$04,$7F,$01,$FA
	.dc.b	$17,$A1,$FC,$0F,$A7,$E1,$D7,$EF,$5A,$FE,$34,$4F,$07,$F8,$13,$5D
	.dc.b	$47,$0C,$FF,$93,$47,$EA,$F7,$F0,$5C,$8E,$D1,$E1,$30,$F7,$FA,$0F
	.dc.b	$C4,$AF,$A1,$7C,$77,$F1,$96,$C7,$F1,$39,$D9,$C2,$33,$1E,$96,$E1
	.dc.b	$E4,$9D,$F9,$5C,$EE,$5F,$38,$FE,$4D,$3F,$CE,$A3,$FD,$6C,$D4,$47
	.dc.b	$0D,$8F,$F5,$A2,$F8,$07,$D0,$FF,$31,$F2,$BF,$D2,$F8,$FC,$66,$F1
	.dc.b	$F4,$BE,$15,$FC,$DE,$F8,$C7,$F3,$1A,$DB,$E2,$6B,$C0,$FE,$6D,$1C
	.dc.b	$1F,$F6,$71,$92,$B9,$7F,$DA,$4D,$7C,$5F,$41,$FB,$AF,$1D,$F1,$9F
	.dc.b	$73,$2F,$E6,$CD,$CB,$79,$FF,$4D,$B7,$C5,$FF,$4F,$FA,$1F,$E4,$F9
	.dc.b	$7E,$2D,$B8,$FE,$1A,$B7,$3F,$D3,$7E,$09,$F1,$1F,$D7,$A1,$EB,$FF
	.dc.b	$91,$1F,$F6,$FC,$66,$31,$3C,$2F,$FC,$29,$EF,$FC,$B8,$C5,$F2,$8F
	.dc.b	$87,$EB,$6A,$3F,$DE,$4D,$7F,$DA,$C6,$CC,$94,$7C,$D3,$D7,$FF,$02
	.dc.b	$61,$FD,$67,$E8,$7F,$A4,$70,$7F,$67,$AD,$3F,$DE,$F8,$93,$AC,$3F
	.dc.b	$5D,$4E,$FE,$77,$FD,$73,$E8,$7F,$1E,$F3,$8C,$D3,$05,$6A,$3F,$71
	.dc.b	$F3,$AF,$9F,$FD,$9C,$7E,$36,$31,$87,$E1,$21,$18,$7F,$0E,$D7,$F2
	.dc.b	$86,$A7,$C3,$8B,$E7,$FF,$C1,$57,$FE,$05,$BE,$5A,$7F,$89,$EB,$5E
	.dc.b	$A7,$A3,$1E,$57,$FE,$1D,$9C,$BC,$71,$EB,$8F,$4A,$FE,$25,$F3,$DF
	.dc.b	$16,$BD,$0D,$8F,$CE,$B5,$F3,$B5,$FD,$63,$BC,$7E,$BF,$CE,$BE,$7B
	.dc.b	$C9,$39,$BF,$F1,$FC,$E5,$FF,$10,$46,$6F,$FC,$63,$39,$7C,$BB,$FD
	.dc.b	$4C,$13,$97,$F9,$1D,$71,$73,$87,$82,$B8,$56,$F0,$FE,$05,$F3,$DF
	.dc.b	$D6,$9F,$C4,$7C,$F3,$E1,$3E,$54,$E7,$3F,$F0,$17,$12,$B5,$7F,$E3
	.dc.b	$7A,$D6,$7F,$E2,$D9,$CD,$FF,$3D,$9C,$D7,$F8,$AA,$73,$5F,$CE,$A7
	.dc.b	$35,$F8,$F9,$CD,$79,$BD,$0F,$C9,$E1,$E1,$C7,$F2,$BE,$77,$F8,$17
	.dc.b	$E2,$BF,$2F,$EB,$7F,$83,$BC,$0C,$FA,$1F,$EA,$DC,$BF,$B8,$C6,$2B
	.dc.b	$7E,$97,$F0,$A9,$FE,$F1,$18,$9E,$3F,$15,$27,$FB,$C9,$76,$FE,$DD
	.dc.b	$17,$83,$FC,$37,$34,$FF,$0D,$C7,$6F,$77,$15,$FE,$17,$6E,$7F,$E2
	.dc.b	$46,$E5,$FE,$52,$BF,$F0,$57,$C2,$F9,$BE,$7B,$D6,$D8,$C9,$FE,$0C
	.dc.b	$8C,$B4,$FF,$07,$FC,$DF,$B8,$F7,$F9,$56,$32,$4C,$BF,$C6,$D8,$7F
	.dc.b	$4B,$8C,$69,$FE,$32,$DF,$FD,$6F,$1F,$E1,$AF,$9E,$F9,$BE,$D7,$CA
	.dc.b	$6C,$64,$FF,$11,$46,$5F,$F2,$AF,$A1,$FE,$E5,$7F,$C6,$7E,$E7,$DA
	.dc.b	$8C,$65,$FE,$34,$8C,$7C,$4F,$DC,$F7,$51,$8B,$E2,$7F,$A1,$FE,$25
	.dc.b	$F1,$BB,$F5,$FF,$81,$46,$27,$FC,$83,$18,$9D,$47,$16,$E3,$FE,$13
	.dc.b	$B7,$1F,$E3,$3E,$75,$F3,$DF,$C3,$67,$38,$7F,$A8,$A7,$2F,$FA,$66
	.dc.b	$33,$7F,$E6,$19,$CB,$D6,$FA,$5F,$36,$DF,$16,$CF,$F8,$97,$E4,$9F
	.dc.b	$1F,$F9,$EF,$71,$CF,$FC,$7E,$73,$87,$FA,$D2,$72,$FF,$AA,$E3,$37
	.dc.b	$FE,$73,$9C,$BD,$6F,$A5,$FE,$AB,$E1,$9F,$3D,$7D,$AC,$37,$2F,$F4
	.dc.b	$93,$70,$FF,$3F,$57,$09,$FF,$4B,$57,$C7,$F8,$5F,$DA,$F1,$3E,$0F
	.dc.b	$FD,$BD,$EA,$FF,$0F,$C7,$FC,$3F,$C5,$FE,$C5,$C7,$FD,$9B,$7B,$7F
	.dc.b	$C4,$75,$BA,$FC,$1E,$1D,$30,$E7,$E2,$FF,$87,$71,$E5,$5F,$E9,$1C
	.dc.b	$79,$2C,$39,$4E,$BB,$3F,$DB,$D5,$A9,$CE,$3D,$72,$F0,$E3,$FE,$91
	.dc.b	$F9,$1E,$F9,$7F,$A4,$5B,$E1,$7F,$A7,$B8,$6F,$7A,$69,$30,$E7,$7F
	.dc.b	$E9,$9A,$C3,$E5,$C3,$FC,$83,$EC,$BE,$7B,$F8,$B4,$E4,$65,$FE,$6E
	.dc.b	$F0,$BC,$75,$BF,$1B,$C7,$CF,$FB,$BB,$E1,$39,$CF,$C3,$87,$17,$C7
	.dc.b	$19,$9E,$D3,$96,$FC,$EB,$E7,$BD,$ED,$BE,$6F,$F8,$B1,$B1,$A2,$3A
	.dc.b	$F7,$6D,$74,$9B,$C5,$F8,$3C,$18,$73,$AD,$61,$E8,$E1,$7C,$F7,$CE
	.dc.b	$F9,$3F,$9E,$D9,$D3,$6B,$D7,$5C,$27,$95,$7D,$6F,$83,$91,$D7,$E7
	.dc.b	$AD,$F0,$9B,$9F,$97,$05,$63,$8F,$F2,$1F,$3D,$F8,$2F,$98,$F4,$5F
	.dc.b	$E3,$A3,$1C,$1F,$88,$CB,$F8,$DC,$63,$3F,$C1,$C6,$1F,$AB,$7C,$B0
	.dc.b	$EB,$18,$BE,$5D,$6C,$39,$61,$F8,$5F,$45,$F3,$BF,$C0,$33,$FD,$74
	.dc.b	$63,$A7,$F3,$2B,$E7,$B9,$67,$F9,$48,$D3,$6F,$96,$9F,$E8,$9F,$95
	.dc.b	$FF,$48,$7A,$B9,$C7,$D9,$18,$CF,$C5,$7C,$1F,$F4,$DC,$67,$3F,$1C
	.dc.b	$E6,$53,$9C,$3F,$D4,$39,$78,$A7,$F4,$06,$BA,$FF,$CA,$8F,$6C,$7C
	.dc.b	$31,$8C,$7F,$83,$61,$CE,$BC,$33,$9C,$3A,$E3,$E1,$AD,$57,$5C,$BE
	.dc.b	$9C,$71,$CB,$C5,$8F,$4C,$79,$5F,$4B,$DC,$62,$B7,$8F,$C7,$8D,$D7
	.dc.b	$53,$D7,$87,$5C,$7C,$53,$A8,$F1,$99,$CB,$DB,$8F,$62,$5F,$C1,$CE
	.dc.b	$E7,$E5,$9B,$97,$BA,$F8,$DF,$83,$1E,$C7,$9F,$2F,$0D,$F0,$9F,$61
	.dc.b	$D4,$E0,$E5,$F5,$61,$F6,$AF,$AB,$E1,$8D,$9C,$5C,$13,$0F,$89,$5F
	.dc.b	$89,$C7,$A6,$5F,$03,$2F,$26,$3D,$2F,$78,$F4,$C7,$C1,$87,$0B,$F5
	.dc.b	$E3,$CE,$B5,$7C,$2F,$C1,$0C,$FE,$AB,$EE,$8C,$BE,$C9,$F9,$18,$75
	.dc.b	$BF,$89,$7C,$2B,$47,$1A,$EC,$70,$7C,$26,$49,$1C,$72,$F7,$57,$4C
	.dc.b	$FE,$8D,$70,$3A,$D6,$F2,$F6,$C7,$48,$C5,$79,$63,$E7,$61,$E2,$7C
	.dc.b	$07,$2B,$EB,$97,$BB,$1F,$C6,$65,$C6,$3E,$2E,$3D,$2F,$C5,$5A,$6E
	.dc.b	$5F,$B0,$7F,$D6,$F1,$FB,$F6,$C6,$5F,$AE,$7F,$AF,$3E,$19,$F3,$CF
	.dc.b	$9E,$FF,$62,$46,$1B,$19,$9F,$F6,$18,$9D,$38,$BF,$ED,$9D,$9F,$EC
	.dc.b	$D6,$C6,$7C,$0F,$F7,$DB,$67,$51,$8C,$FF,$82,$B7,$6F,$E3,$8D,$AF
	.dc.b	$C3,$7C,$F3,$CA,$47,$F0,$49,$D3,$FC,$61,$B8,$F8,$B0,$FE,$5F,$7C
	.dc.b	$6B,$86,$31,$FE,$03,$7D,$F6,$56,$E7,$35,$BB,$F3,$BF,$CD,$67,$51
	.dc.b	$89,$D9,$1F,$0E,$3E,$88,$C6,$5E,$EB,$E8,$FF,$2E,$BE,$6E,$67,$67
	.dc.b	$F3,$8C,$7C,$D3,$A8,$C5,$72,$69,$FD,$46,$34,$FC,$58,$D3,$4D,$5F
	.dc.b	$97,$0F,$51,$30,$F1,$3E,$A9,$E3,$97,$CD,$9C,$9E,$9A,$E4,$E2,$7C
	.dc.b	$75,$CD,$FB,$73,$A7,$33,$B7,$C9,$3D,$DF,$82,$62,$3A,$65,$F5,$AB
	.dc.b	$94,$E4,$E6,$48,$D3,$E8,$CB,$F0,$A7,$CA,$9E,$38,$7A,$E7,$89,$CE
	.dc.b	$B8,$BE,$28,$D3,$9C,$BE,$6C,$60,$97,$D6,$7C,$19,$7D,$5B,$E7,$97
	.dc.b	$C3,$9F,$41,$CF,$1F,$1D,$F9,$AF,$86,$3D,$A3,$15,$F8,$4C,$39,$E5
	.dc.b	$F2,$72,$F0,$E1,$F9,$5C,$79,$BE,$2B,$FB,$E5,$CB,$EA,$61,$CA,$3E
	.dc.b	$0D,$F2,$76,$7C,$1C,$7C,$35,$F5,$4F,$54,$ED,$C6,$3E,$39,$DB,$F8
	.dc.b	$5A,$F2,$1E,$17,$0F,$69,$E7,$3F,$26,$3F,$CC,$FF,$9C,$7E,$7F,$13
	.dc.b	$4F,$D1,$77,$BE,$63,$FC,$A7,$5F,$E6,$A9,$D6,$9F,$8F,$BE,$2F,$E4
	.dc.b	$6F,$87,$7E,$6B,$77,$C6,$B5,$87,$48,$FF,$2C,$FC,$0F,$1E,$E3,$4C
	.dc.b	$7D,$9D,$9F,$F3,$34,$57,$FC,$C3,$1C,$7D,$49,$1F,$E6,$B7,$C9,$F8
	.dc.b	$4F,$F2,$88,$FF,$33,$C6,$9A,$7F,$A2,$23,$5B,$FF,$9D,$FE,$77,$F8
	.dc.b	$74,$FF,$A2,$1F,$E8,$D1,$9E,$BF,$81,$1A,$8E,$13,$89,$8F,$FA,$52
	.dc.b	$75,$C5,$FC,$C3,$81,$18,$9F,$F5,$0B,$DA,$74,$D8,$C9,$FB,$FE,$44
	.dc.b	$BB,$F0,$D9,$FF,$30,$35,$FF,$50,$E4,$F8,$ED,$D4,$9D,$BF,$22,$3E
	.dc.b	$4F,$B3,$F2,$AE,$57,$C7,$8C,$BF,$26,$32,$F6,$8F,$E0,$7E,$2C,$8F
	.dc.b	$F2,$8B,$F8,$7F,$37,$FD,$FF,$18,$21,$FE,$4A,$8D,$7A,$7C,$63,$06
	.dc.b	$B7,$FF,$56,$57,$28,$FF,$5E,$F0,$3D,$7F,$CA,$2F,$7E,$98,$89,$3F
	.dc.b	$EC,$89,$D3,$67,$2F,$FB,$3B,$E8,$1E,$DF,$E8,$5B,$1F,$EC,$E6,$BF
	.dc.b	$EC,$EF,$9F,$7D,$4E,$EA,$4E,$DF,$D4,$4E,$A3,$FC,$EB,$F3,$9F,$D7
	.dc.b	$7D,$67,$91,$FE,$95,$8C,$36,$7F,$BC,$CE,$9F,$F7,$4F,$BA,$FA,$9D
	.dc.b	$C7,$FA,$5A,$EB,$FE,$94,$6F,$8A,$77,$D6,$BE,$55,$DB,$1F,$02,$C7
	.dc.b	$C0,$F3,$34,$77,$2D,$73,$C7,$C3,$D6,$F8,$FA,$4A,$DE,$9F,$97,$C3
	.dc.b	$AD,$6E,$7A,$9F,$A8,$AE,$0F,$8F,$1F,$D6,$5F,$17,$FC,$0D,$E0,$BE
	.dc.b	$6B,$E0,$B6,$B9,$B5,$E4,$33,$7D,$13,$DC,$E3,$72,$B8,$BF,$22,$75
	.dc.b	$1F,$0B,$C0,$7C,$D7,$E4,$E3,$0D,$7B,$C6,$BA,$3E,$8F,$83,$EF,$9C
	.dc.b	$D6,$F8,$3E,$F5,$EE,$79,$DF,$0D,$E4,$7D,$0C,$39,$5F,$E1,$74,$46
	.dc.b	$36,$FE,$24,$BE,$A3,$8D,$FC,$2A,$DE,$18,$C3,$95,$EE,$BD,$7F,$0B
	.dc.b	$F8,$EC,$FD,$ED,$D0,$FB,$38,$F0,$CA,$E2,$79,$E3,$87,$B9,$23,$E3
	.dc.b	$BF,$13,$DC,$C4,$7C,$48,$B9,$79,$26,$ED,$FE,$EE,$C5,$58,$BF,$B9
	.dc.b	$9A,$F1,$9F,$D6,$AD,$6B,$F1,$A7,$D9,$E9,$FE,$7B,$3F,$A2,$DC,$FF
	.dc.b	$DF,$6D,$CB,$AF,$B1,$C2,$64,$7E,$0E,$75,$87,$D2,$AF,$D9,$FC,$EB
	.dc.b	$E2,$E9,$2F,$F3,$FC,$AF,$98,$C4,$7E,$58,$57,$17,$F4,$AD,$75,$3C
	.dc.b	$5C,$D7,$2C,$FF,$73,$5B,$7F,$77,$5B,$AE,$D3,$F2,$9E,$BA,$FA,$0F
	.dc.b	$E1,$73,$FA,$F8,$FF,$45,$CE,$DE,$33,$B7,$D7,$E5,$57,$27,$43,$25
	.dc.b	$32,$4B,$F2,$C7,$F1,$AE,$17,$CD,$FF,$AB,$2B,$FD,$37,$18,$70,$D7
	.dc.b	$FD,$62,$D7,$FB,$EB,$9A,$FE,$36,$DD,$FF,$D1,$CD,$9F,$F5,$3F,$CE
	.dc.b	$FF,$82,$EB,$77,$E3,$24,$FF,$16,$9D,$5F,$FA,$5F,$E6,$DF,$01,$FC
	.dc.b	$C9,$B5,$FC,$71,$AF,$F1,$2F,$3D,$F9,$63,$EA,$B8,$65,$2B,$83,$FC
	.dc.b	$E3,$EB,$5E,$5F,$F3,$EF,$71,$CD,$FC,$DE,$DE,$D7,$F8,$05,$FF,$60
	.dc.b	$CE,$B8,$3F,$D8,$73,$99,$FF,$6B,$57,$18,$C5,$6E,$79,$FA,$9F,$ED
	.dc.b	$CF,$43,$8B,$63,$27,$F8,$94,$A7,$FA,$94,$BF,$36,$F9,$8F,$EB,$59
	.dc.b	$7F,$4C,$8C,$71,$FF,$4A,$FA,$CF,$F6,$C3,$FD,$0A,$33,$7F,$A3,$7E
	.dc.b	$34,$67,$83,$9B,$FE,$DE,$E2,$7C,$F7,$F7,$48,$D1,$FD,$33,$E6,$F9
	.dc.b	$E5,$FD,$71,$B1,$FE,$E6,$C1,$F5,$2F,$67,$AE,$7F,$EE,$AB,$DB,$75
	.dc.b	$CA,$8C,$DF,$2A,$CC,$70,$9D,$57,$F6,$C7,$FD,$BD,$1E,$E7,$FC,$A6
	.dc.b	$D7,$31,$B7,$FD,$F7,$F3,$6F,$9A,$F0,$B7,$2F,$EE,$B8,$FA,$BC,$BC
	.dc.b	$D4,$9F,$ED,$B5,$BB,$FE,$C7,$7B,$9C,$D6,$E7,$F8,$95,$6F,$4F,$F0
	.dc.b	$95,$EE,$73,$5B,$9F,$F0,$3D,$6E,$FE,$66,$9F,$D4,$5E,$A7,$8B,$DE
	.dc.b	$F4,$9F,$3B,$77,$CD,$FF,$82,$67,$44,$CB,$C7,$5B,$9D,$46,$27,$58
	.dc.b	$F8,$AB,$5E,$6F,$69,$FE,$6D,$E4,$7D,$CF,$28,$C8,$FE,$D9,$F0,$D9
	.dc.b	$8C,$63,$FD,$E9,$B8,$FF,$46,$7F,$0B,$E5,$7F,$70,$AE,$11,$C7,$E3
	.dc.b	$66,$1F,$DD,$FE,$BE,$E3,$CE,$6E,$3F,$DC,$26,$D7,$F5,$09,$B5,$FD
	.dc.b	$72,$6E,$5E,$89,$CE,$5F,$E0,$09,$CE,$3F,$C4,$B0,$FE,$A9,$E2,$79
	.dc.b	$BC,$E7,$D3,$7B,$62,$4E,$8F,$F0,$64,$63,$3F,$F0,$35,$3E,$83,$91
	.dc.b	$CF,$EB,$ED,$FF,$84,$E2,$D7,$09,$BB,$7F,$6C,$F5,$BE,$61,$8B,$FE
	.dc.b	$DA,$E1,$FE,$D4,$E6,$9B,$FD,$68,$FD,$7F,$C0,$FF,$A0,$6B,$FC,$8E
	.dc.b	$DA,$FF,$0E,$7C,$BB,$8B,$7E,$BC,$BD,$BF,$62,$75,$18,$3F,$CA,$3F
	.dc.b	$10,$F8,$FF,$DE,$B6,$3F,$CB,$3F,$33,$89,$91,$C1,$FC,$1F,$BA,$9C
	.dc.b	$E5,$FD,$EF,$D3,$7C,$BF,$F0,$86,$E1,$FE,$6A,$F8,$07,$CC,$66,$A4
	.dc.b	$FF,$99,$2B,$77,$FE,$39,$BD,$CE,$6B,$73,$DD,$FC,$AC,$EA,$31,$3F
	.dc.b	$E7,$AF,$39,$F3,$0F,$98,$FF,$47,$79,$FF,$ED,$0B,$FE,$97,$39,$B4
	.dc.b	$7F,$A8,$7D,$DD,$FC,$67,$F4,$CA,$D4,$AF,$96,$3D,$B1,$DC,$F4,$6F
	.dc.b	$80,$E0,$FF,$11,$67,$DB,$0E,$38,$F4,$7C,$94,$F8,$E7,$C8,$72,$5E
	.dc.b	$1D,$8C,$37,$0E,$9A,$FF,$BD,$6B,$71,$8D,$3F,$D0,$91,$D2,$31,$87
	.dc.b	$7B,$F4,$D7,$FB,$CF,$E5,$AF,$11,$F5,$F5,$7E,$C6,$7E,$16,$5F,$21
	.dc.b	$E1,$8F,$76,$9A,$3E,$6D,$F3,$8F,$0D,$73,$D3,$EA,$57,$06,$D6,$E3
	.dc.b	$91,$9B,$FA,$F9,$2F,$94,$78,$EB,$85,$FC,$CF,$01,$F3,$3F,$51,$FC
	.dc.b	$44,$66,$31,$CD,$96,$E7,$F1,$AB,$8B,$67,$EF,$F7,$7F,$96,$BE,$0E
	.dc.b	$FC,$0F,$F9,$D6,$B8,$7C,$D6,$26,$47,$D3,$AE,$0E,$8C,$57,$1D,$FE
	.dc.b	$46,$EF,$88,$7C,$A7,$E3,$1F,$DC,$4E,$9C,$DF,$DC,$73,$B7,$E3,$1C
	.dc.b	$CE,$FB,$B9,$5F,$36,$DF,$16,$CE,$E3,$E1,$7D,$2F,$C9,$3E,$57,$81
	.dc.b	$63,$24,$9F,$4F,$07,$F8,$81,$B5,$F4,$7E,$A6,$57,$CD,$FB,$3B,$5C
	.dc.b	$26,$E9,$F6,$66,$F2,$FF,$8F,$BF,$25,$57,$EC,$3B,$D1,$8C,$7E,$17
	.dc.b	$CB,$E6,$31,$97,$E4,$EB,$77,$93,$E6,$55,$BF,$D4,$4E,$9B,$7C,$B2
	.dc.b	$F9,$95,$BC,$3F,$3D,$C6,$F9,$BF,$A3,$9F,$E9,$7E,$4B,$FB,$9F,$AB
	.dc.b	$F8,$19,$CD,$FF,$10,$9C,$BC,$A3,$2A,$33,$8F,$F1,$9C,$3B,$71,$7E
	.dc.b	$E2,$73,$8F,$7F,$45,$97,$EC,$E7,$38,$FF,$50,$CB,$AC,$62,$3B,$57
	.dc.b	$07,$F5,$F1,$C3,$1E,$FA,$F3,$AD,$67,$FB,$8A,$E4,$6F,$0E,$B3,$AC
	.dc.b	$BC,$99,$71,$3E,$16,$7E,$99,$D6,$7F,$0D,$AF,$92,$77,$97,$18,$F2
	.dc.b	$CE,$B1,$F0,$5F,$3B,$D6,$5E,$5A,$EB,$9F,$AB,$0E,$35,$AC,$BC,$B1
	.dc.b	$F4,$70,$E4,$E8,$98,$7D,$1C,$FD,$B7,$C6,$7D,$0F,$99,$DD,$F4,$AF
	.dc.b	$14,$E5,$B3,$A3,$89,$E1,$CF,$A3,$F0,$4D,$63,$E9,$9E,$24,$BE,$C7
	.dc.b	$F5,$A8,$D3,$89,$EB,$3B,$7F,$05,$8F,$BA,$31,$1D,$DF,$74,$66,$79
	.dc.b	$4E,$5E,$D8,$FC,$28,$F5,$19,$3A,$E1,$EA,$CF,$F0,$B9,$FD,$2A,$E1
	.dc.b	$8F,$82,$7E,$CD,$6E,$BE,$E1,$BC,$7C,$58,$7A,$B1,$F6,$65,$EE,$8C
	.dc.b	$10,$F1,$39,$7C,$84,$C3,$BB,$E8,$7C,$B9,$FD,$5A,$F7,$C6,$27,$EA
	.dc.b	$1B,$C3,$CB,$5B,$8F,$04,$EF,$2F,$93,$97,$AE,$BF,$0D,$8F,$8E,$B5
	.dc.b	$8F,$AB,$3F,$86,$FE,$4B,$1F,$ED,$87,$E0,$4B,$3F,$C1,$9B,$97,$ED
	.dc.b	$DD,$FC,$17,$F9,$17,$E6,$F8,$C7,$FB,$9B,$CD,$46,$21,$3B,$76,$D7
	.dc.b	$BF,$3F,$FB,$C1,$E7,$87,$F5,$2E,$2E,$4D,$8C,$CE,$DE,$33,$FC,$46
	.dc.b	$39,$6F,$FC,$1D,$B9,$FE,$5F,$E6,$DD,$1F,$E0,$BA,$D5,$F1,$26,$1F
	.dc.b	$CB,$AB,$6E,$27,$6F,$F1,$8C,$3A,$BF,$E6,$E9,$DB,$9A,$E1,$87,$F3
	.dc.b	$5A,$DE,$EB,$DE,$1D,$09,$5B,$BF,$EF,$53,$A8,$C6,$1D,$49,$5F,$CD
	.dc.b	$E7,$51,$FC,$FF,$0F,$2B,$FC,$F5,$B9,$7C,$BA,$E1,$97,$C3,$C3,$AC
	.dc.b	$78,$AF,$96,$3E,$2A,$D4,$60,$F0,$D7,$03,$E7,$46,$A3,$C1,$3F,$09
	.dc.b	$E1,$5E,$07,$06,$8E,$67,$CB,$9E,$57,$E7,$BE,$D3,$C9,$EE,$F2,$8D
	.dc.b	$39,$39,$19,$C3,$E0,$9F,$4B,$1F,$8F,$1E,$1C,$3E,$03,$6F,$AB,$E4
	.dc.b	$30,$75,$CB,$ED,$63,$E6,$C3,$AE,$5E,$F9,$CD,$78,$63,$18,$F8,$32
	.dc.b	$F3,$CF,$43,$AD,$70,$7C,$11,$A3,$DB,$19,$CB,$CF,$19,$C3,$D9,$8F
	.dc.b	$8F,$1F,$06,$1F,$9B,$8C,$3D,$27,$6E,$AF,$F0,$18,$75,$CB,$E7,$CF
	.dc.b	$97,$1F,$0E,$5E,$9C,$3D,$F3,$F6,$2F,$83,$F4,$4E,$11,$D9,$F0,$65
	.dc.b	$F0,$B2,$F1,$E5,$EA,$AD,$36,$75,$5E,$5C,$BF,$03,$87,$63,$BC,$76
	.dc.b	$9F,$CB,$3F,$E2,$CF,$9B,$FD,$1F,$5B,$F7,$6F,$E3,$F8,$FF,$A8,$9B
	.dc.b	$8F,$FA,$D3,$05,$FE,$A9,$B9,$FF,$A4,$79,$3F,$D2,$51,$8C,$BF,$D7
	.dc.b	$71,$8A,$FE,$02,$DA,$E1,$3D,$67,$A1,$23,$F5,$79,$7F,$9C,$7C,$7B
	.dc.b	$E3,$BD,$9A,$4C,$3F,$D5,$70,$E9,$C4,$7F,$9E,$A3,$83,$A9,$E7,$3D
	.dc.b	$3E,$22,$4F,$27,$C3,$DF,$FC,$AF,$83,$FD,$8B,$82,$77,$5D,$5D,$ED
	.dc.b	$E8,$F9,$5F,$93,$2D,$11,$FD,$1F,$0F,$F6,$7B,$78,$3F,$D5,$6F,$09
	.dc.b	$DD,$73,$EB,$8F,$FB,$22,$74,$48,$CC,$7F,$A6,$23,$C3,$C6,$B7,$8F
	.dc.b	$F3,$39,$3C,$17,$E5,$6D,$6E,$7F,$8D,$B6,$7F,$D5,$15,$C9,$B3,$F1
	.dc.b	$BE,$23,$E3,$FC,$2F,$C8,$A9,$5F,$EB,$98,$CB,$FE,$BB,$8C,$BF,$E9
	.dc.b	$68,$FE,$37,$F3,$9F,$80,$F9,$E5,$D3,$35,$FE,$6B,$9F,$C4,$BF,$E7
	.dc.b	$96,$BF,$E7,$76,$F5,$9A,$72,$8F,$F3,$6B,$F8,$EF,$31,$F3,$BF,$90
	.dc.b	$E3,$7B,$17,$E7,$66,$1F,$EF,$89,$CB,$67,$2F,$FB,$22,$13,$CE,$7F
	.dc.b	$A2,$F2,$BE,$75,$F3,$9F,$9E,$F5,$5F,$35,$4D,$CE,$B2,$FF,$51,$B5
	.dc.b	$FF,$55,$37,$A7,$EF,$43,$0E,$D8,$75,$F9,$DF,$E8,$FF,$3A,$A9,$F0
	.dc.b	$0F,$13,$FD,$F7,$C2,$F9,$4F,$88,$70,$AD,$CF,$FA,$AE,$FF,$C3,$51
	.dc.b	$FE,$A5,$6F,$21,$A4,$8E,$27,$FA,$BD,$F1,$2C,$3C,$4E,$4D,$F7,$3D
	.dc.b	$D7,$CE,$70,$AD,$D7,$EB,$3E,$23,$19,$7B,$70,$71,$BE,$03,$37,$FB
	.dc.b	$38,$C3,$E4,$39,$17,$C1,$61,$C2,$32,$79,$F0,$F4,$CE,$A3,$CF,$5F
	.dc.b	$A5,$74,$7C,$FF,$59,$F3,$7E,$2C,$9F,$07,$FE,$EA,$AF,$7B,$5E,$22
	.dc.b	$B8,$BE,$76,$BA,$6E,$9F,$11,$B9,$7C,$3F,$9D,$7C,$2F,$C2,$F1,$5F
	.dc.b	$37,$F9,$AC,$7F,$CB,$F1,$8E,$DF,$F2,$B3,$63,$FC,$B6,$D7,$73,$C7
	.dc.b	$2F,$B3,$39,$6C,$67,$1F,$A9,$E5,$7C,$7B,$F8,$BA,$A3,$17,$F8,$28
	.dc.b	$C7,$0F,$C7,$9C,$DF,$2F,$91,$C4,$CC,$3F,$0F,$5A,$7F,$98,$29,$F8
	.dc.b	$18,$FE,$BA,$7D,$6D,$7D,$7D,$FF,$02,$31,$1F,$96,$CF,$E3,$CF,$22
	.dc.b	$47,$83,$97,$AF,$CE,$71,$24,$76,$6B,$F8,$4A,$FA,$31,$C2,$3D,$3E
	.dc.b	$FD,$75,$82,$71,$B7,$C5,$CD,$78,$D7,$FD,$A1,$7B,$C3,$F5,$F8,$F6
	.dc.b	$7A,$98,$F6,$3F,$52,$FB,$5A,$FF,$B6,$5A,$A1,$18,$6D,$7E,$2E,$3F
	.dc.b	$67,$8F,$E0,$9B,$1F,$80,$F9,$8F,$D5,$7C,$03,$E7,$7F,$59,$D1,$FB
	.dc.b	$2C,$BF,$3E,$DC,$7F,$7B,$5B,$9F,$DB,$35,$FF,$56,$B5,$FF,$5B,$B5
	.dc.b	$FE,$1A,$D7,$F9,$03,$C6,$3A,$D7,$90,$F1,$FC,$4F,$9B,$7F,$2F,$D0
	.dc.b	$4C,$3C,$31,$97,$FD,$75,$87,$CF,$7C,$71,$FC,$92,$31,$1E,$48,$E0
	.dc.b	$FE,$F2,$38,$61,$FC,$6B,$1F,$E1,$0F,$C0,$AD,$B6,$BF,$8B,$4E,$B3
	.dc.b	$FE,$53,$B7,$F3,$79,$FE,$35,$F3,$CF,$3F,$F8,$C3,$63,$F8,$F4,$ED
	.dc.b	$FE,$1A,$D7,$FD,$BE,$D7,$AF,$87,$FC,$FA,$31,$1F,$D2,$5E,$58,$FF
	.dc.b	$32,$6E,$1C,$27,$C7,$E3,$FF,$29,$F9,$E7,$C8,$FF,$48,$6C,$FF,$B8
	.dc.b	$DA,$FF,$BA,$1A,$FF,$BC,$9F,$0A,$69,$31,$FE,$A7,$EF,$38,$FF,$A6
	.dc.b	$57,$F6,$3F,$47,$FA,$7C,$E7,$4F,$F2,$E4,$62,$B7,$97,$F8,$42,$FF
	.dc.b	$B3,$5E,$DE,$47,$F9,$98,$A7,$FB,$9C,$A7,$F6,$DF,$33,$AB,$B9,$E8
	.dc.b	$BE,$5D,$F3,$1F,$AE,$6E,$DF,$D9,$27,$FB,$2C,$61,$C1,$C6,$78,$7C
	.dc.b	$DC,$AC,$CF,$58,$C3,$C3,$E6,$B9,$6B,$FD,$A1,$B5,$FE,$03,$C1,$F3
	.dc.b	$1F,$89,$8F,$EC,$93,$AB,$FF,$04,$3E,$37,$CC,$D7,$FB,$DB,$5D,$E3
	.dc.b	$FC,$2A,$39,$60,$FA,$5F,$F3,$D5,$F6,$AF,$F0,$0C,$C9,$CD,$FF,$62
	.dc.b	$F9,$A7,$CD,$EB,$0F,$F0,$4C,$61,$FF,$06,$E7,$EA,$F1,$35,$7F,$E1
	.dc.b	$FA,$D3,$CE,$F2,$A3,$38,$FF,$84,$AF,$96,$1F,$E0,$59,$CB,$E0,$BF
	.dc.b	$E4,$B1,$D2,$FE,$C3,$5D,$4F,$D4,$8E,$30,$F7,$9F,$37,$FE,$75,$AE
	.dc.b	$D7,$E8,$9C,$FB,$BE,$6C,$3E,$1E,$5F,$E2,$06,$DF,$F8,$EE,$B9,$F9
	.dc.b	$1B,$F8,$29,$3E,$17,$CE,$47,$07,$F8,$D0,$7E,$5D,$FC,$D5,$F8,$11
	.dc.b	$F2,$9C,$F8,$9F,$33,$C6,$FE,$87,$E3,$53,$FA,$1E,$FF,$DF,$3C,$87
	.dc.b	$C5,$66,$66,$1F,$E0,$A7,$75,$FD,$45,$FE,$A3,$7D,$27,$5F,$2B,$D4
	.dc.b	$FB,$7E,$6F,$F8,$1F,$E7,$31,$3F,$D6,$E6,$E1,$E1,$9B,$C3,$FD,$4E
	.dc.b	$6D,$7F,$84,$A6,$D7,$F5,$DC,$55,$E2,$6D,$78,$89,$3F,$E2,$68,$FE
	.dc.b	$37,$1F,$DA,$FC,$3F,$2E,$8F,$9C,$F8,$91,$27,$E0,$3D,$48,$FE,$F3
	.dc.b	$3F,$F1,$17,$CE,$72,$39,$FC,$E5,$9F,$F0,$EC,$5C,$BE,$74,$DD,$7F
	.dc.b	$B8,$7A,$7F,$4C,$C5,$FF,$7B,$70,$FD,$17,$DC,$47,$F8,$4F,$BB,$FE
	.dc.b	$A0,$7F,$CC,$8D,$BF,$F1,$BF,$AF,$F7,$27,$3E,$AF,$EA,$AB,$58,$7F
	.dc.b	$8E,$7E,$09,$E9,$66,$BE,$CC,$62,$FF,$CE,$1F,$16,$F9,$1F,$D9,$B6
	.dc.b	$3F,$CD,$7F,$39,$89,$91,$FE,$84,$F5,$3F,$D6,$EF,$28,$FC,$2C,$63
	.dc.b	$5F,$F3,$4F,$C4,$FF,$A0,$6B,$FC,$FC,$D8,$FF,$28,$FC,$03,$E6,$3F
	.dc.b	$81,$CE,$76,$FF,$46,$BB,$AF,$E6,$8D,$7F,$D3,$2D,$7F,$D5,$8D,$7F
	.dc.b	$DD,$FF,$03,$F9,$8E,$0F,$F3,$8E,$3F,$D1,$7C,$4E,$33,$9E,$3F,$F5
	.dc.b	$0F,$9F,$FE,$BA,$F3,$DF,$36,$F9,$CC,$E1,$FE,$B9,$9C,$C6,$33,$FF
	.dc.b	$4E,$7B,$96,$73,$9F,$79,$FF,$62,$61,$CE,$31,$D7,$FE,$4D,$EF,$E1
	.dc.b	$39,$A6,$7F,$EC,$CA,$D5,$F0,$F8,$E7,$C8,$78,$65,$1F,$A8,$BE,$3A
	.dc.b	$FF,$AC,$E3,$77,$C7,$C2,$E1,$18,$AD,$8B,$E7,$3A,$C3,$6F,$BE,$5C
	.dc.b	$4F,$9C,$C3,$71,$F6,$F1,$FE,$DA,$BA,$9C,$2B,$67,$FB,$42,$7E,$EB
	.dc.b	$63,$EE,$B5,$FB,$AD,$79,$47,$CA,$F8,$FF,$F0,$83,$F3,$E3,$27,$B3
	.dc.b	$24,$EA,$31,$E2,$3E,$73,$E5,$3F,$93,$BF,$AA,$E5,$BD,$1A,$9D,$B6
	.dc.b	$32,$7D,$FF,$1D,$F0,$3F,$E8,$68,$C3,$70,$F0,$FC,$E7,$E2,$6B,$7F
	.dc.b	$07,$6B,$5C,$9F,$81,$F9,$27,$C6,$E6,$BF,$01,$18,$CB,$F1,$7F,$12
	.dc.b	$F9,$5F,$C2,$3F,$BB,$D3,$E1,$39,$9D,$F8,$7A,$AD,$B6,$32,$74,$8C
	.dc.b	$B4,$CD,$7C,$AF,$9C,$F9,$27,$CA,$F1,$AC,$FC,$A8,$B3,$F4,$26,$E9
	.dc.b	$F9,$DE,$27,$CD,$FD,$23,$E9,$FC,$E5,$C7,$E9,$C5,$C3,$EC,$CD,$BF
	.dc.b	$DC,$7C,$E7,$2B,$FD,$87,$82,$8C,$5F,$C2,$F9,$8F,$DB,$46,$2F,$8E
	.dc.b	$3F,$8F,$BE,$0D,$9E,$38,$30,$FE,$05,$18,$6C,$FD,$4E,$CE,$AD,$AF
	.dc.b	$C2,$7C,$E7,$D3,$C5,$E0,$F3,$C1,$F2,$9F,$EB,$EF,$6F,$7C,$7F,$AF
	.dc.b	$8C,$4F,$EA,$9B,$7C,$A3,$C9,$18,$BF,$37,$4B,$1F,$04,$77,$9D,$57
	.dc.b	$0F,$17,$FC,$48,$F5,$6E,$1D,$70,$E5,$8F,$8E,$73,$87,$36,$DF,$3C
	.dc.b	$BD,$39,$FE,$CF,$2E,$F3,$EE,$AD,$B6,$B6,$F8,$5E,$11,$A9,$F3,$E3
	.dc.b	$E2,$BC,$D4,$BE,$07,$F1,$A9,$D3,$8C,$BE,$33,$DC,$C6,$3F,$1B,$3F
	.dc.b	$8B,$1E,$93,$25,$9F,$85,$1A,$8E,$51,$D3,$1F,$69,$E6,$9D,$38,$CB
	.dc.b	$D4,$7C,$6C,$3C,$B5,$C6,$33,$8F,$AE,$3B,$E7,$F3,$E3,$19,$7C,$F7
	.dc.b	$E3,$CF,$17,$39,$79,$E7,$D9,$5C,$63,$D6,$7A,$23,$39,$F0,$8C,$1C
	.dc.b	$46,$3E,$69,$F0,$47,$2C,$BD,$D7,$D7,$1F,$14,$FC,$8B,$E0,$F3,$3A
	.dc.b	$9D,$9C,$B8,$AF,$8D,$8F,$B3,$0E,$F7,$D0,$8F,$A2,$BC,$59,$7C,$87
	.dc.b	$F0,$19,$FD,$08,$C5,$70,$9F,$5C,$E6,$FE,$06,$1E,$AC,$EE,$1C,$31
	.dc.b	$ED,$8F,$1C,$7C,$D3,$F6,$71,$E5,$1E,$2C,$7D,$0F,$BF,$1F,$5B,$E4
	.dc.b	$CF,$E3,$63,$EC,$24,$79,$23,$2F,$1C,$3B,$46,$23,$C3,$7D,$9A,$7C
	.dc.b	$FC,$3C,$78,$FD,$28,$FF,$6D,$63,$FC,$21,$B8,$7E,$F9,$B7,$FE,$D8
	.dc.b	$F8,$6F,$F4,$3F,$CE,$5F,$64,$D2,$57,$FB,$8B,$0F,$E9,$DC,$5F,$EE
	.dc.b	$06,$C6,$7A,$13,$96,$C6,$63,$F8,$24,$63,$3F,$D5,$37,$6F,$E3,$3F
	.dc.b	$38,$E8,$FE,$4D,$5A,$BE,$24,$C3,$F8,$05,$6D,$C4,$ED,$FE,$51,$87
	.dc.b	$57,$F3,$B3,$B7,$35,$C3,$0F,$E4,$D5,$BE,$25,$EF,$0E,$84,$AD,$DF
	.dc.b	$F8,$06,$75,$18,$C3,$A9,$2B,$F9,$64,$EA,$3F,$9C,$E1,$E5,$7F,$A1
	.dc.b	$B7,$2F,$9B,$5C,$32,$F8,$78,$75,$8F,$15,$F2,$C7,$C5,$5A,$8C,$1E
	.dc.b	$1A,$E0,$7C,$E8,$D4,$78,$27,$E1,$3C,$2B,$C0,$E0,$D1,$CC,$F9,$73
	.dc.b	$CA,$FC,$F7,$DA,$79,$3D,$DE,$51,$A7,$27,$23,$38,$7C,$13,$E9,$63
	.dc.b	$F1,$E3,$C3,$87,$C0,$6D,$F5,$7C,$86,$0E,$B9,$7D,$AC,$7C,$D8,$75
	.dc.b	$CB,$DF,$39,$AF,$0C,$63,$1F,$06,$5E,$79,$E8,$75,$AE,$0F,$82,$34
	.dc.b	$7B,$63,$39,$79,$E3,$38,$7B,$31,$F1,$E3,$E0,$C3,$F3,$71,$87,$A4
	.dc.b	$ED,$D5,$FE,$03,$0E,$B9,$7C,$F9,$F2,$E3,$E1,$CB,$D3,$87,$BE,$7E
	.dc.b	$C5,$F0,$7E,$89,$C2,$3B,$3E,$0C,$BE,$16,$5E,$3C,$BD,$55,$A6,$CE
	.dc.b	$AB,$CB,$97,$E0,$70,$EC,$77,$8E,$D3,$F9,$67,$FC,$6F,$F3,$9F,$A3
	.dc.b	$F1,$1D,$BF,$E2,$59,$FF,$3D,$37,$1F,$F5,$B5,$7E,$A6,$CF,$19,$D6
	.dc.b	$3F,$E3,$7C,$39,$E1,$FA,$3B,$DE,$77,$0E,$13,$FA,$7A,$DB,$E4,$9C
	.dc.b	$79,$2F,$76,$92,$7F,$D6,$9D,$75,$FF,$48,$4E,$DF,$F5,$C4,$F2,$9E
	.dc.b	$7E,$A4,$9F,$F5,$E3,$E1,$EE,$FE,$4F,$1C,$A3,$4D,$3F,$D2,$B9,$F9
	.dc.b	$FE,$5B,$DF,$1F,$EB,$97,$F9,$F6,$FF,$EC,$96,$ED,$F0,$DD,$EA,$69
	.dc.b	$AA,$E5,$CF,$D5,$87,$FB,$62,$35,$3C,$67,$33,$FE,$31,$AF,$F6,$2B
	.dc.b	$FE,$DE,$E0,$D9,$D7,$87,$FC,$C3,$DB,$FC,$55,$B3,$BB,$FA,$55,$BB
	.dc.b	$E7,$3F,$EB,$8F,$91,$F8,$D1,$23,$FD,$6F,$1F,$24,$C3,$FE,$BE,$8C
	.dc.b	$3F,$E5,$3F,$9C,$FB,$DE,$53,$E3,$A9,$FC,$66,$24,$FF,$B3,$63,$2F
	.dc.b	$12,$1F,$EC,$D9,$E4,$EC,$CE,$DF,$EF,$9E,$07,$B3,$F8,$D5,$F3,$74
	.dc.b	$C4,$4A,$FF,$47,$57,$19,$D4,$63,$2F,$CE,$5F,$48,$FF,$4F,$FC,$DB
	.dc.b	$E6,$FF,$35,$F3,$4F,$99,$A7,$E8,$67,$55,$C2,$B2,$F5,$1C,$8B,$E5
	.dc.b	$1B,$3F,$CF,$9F,$37,$FE,$99,$F7,$1E,$47,$EB,$FD,$57,$CA,$FE,$AD
	.dc.b	$B7,$C0,$FF,$55,$57,$07,$FD,$33,$7C,$EF,$95,$F1,$3F,$A6,$46,$67
	.dc.b	$67,$29,$DB,$E9,$C5,$87,$82,$3F,$D5,$15,$C5,$FD,$07,$C5,$73,$AF
	.dc.b	$77,$BD,$73,$3B,$EC,$55,$35,$12,$FC,$47,$FB,$16,$76,$E0,$7A,$2F
	.dc.b	$99,$F0,$7C,$27,$FB,$82,$78,$E4,$8C,$43,$7F,$8C,$F2,$AF,$55,$F3
	.dc.b	$8E,$35,$C2,$BE,$65,$F3,$F4,$37,$87,$C9,$EF,$7C,$C7,$E2,$71,$ED
	.dc.b	$C3,$FC,$9F,$BD,$F6,$9F,$15,$B3,$EE,$BE,$3D,$9C,$F0,$E5,$A7,$82
	.dc.b	$F7,$7D,$0F,$C5,$B8,$8F,$C0,$5F,$29,$F8,$3E,$27,$C8,$78,$6A,$8C
	.dc.b	$57,$B1,$B9,$FF,$3D,$5F,$25,$F6,$DC,$DF,$E6,$9B,$7C,$F8,$F5,$5E
	.dc.b	$9E,$AF,$77,$C3,$CC,$3D,$DA,$FE,$81,$F5,$69,$F2,$2B,$51,$E6,$D5
	.dc.b	$7C,$E3,$8F,$C3,$7C,$9F,$AD,$91,$82,$1C,$5C,$BD,$B1,$4F,$CD,$24
	.dc.b	$70,$DF,$E1,$7B,$58,$8E,$13,$67,$31,$F5,$FD,$5F,$C4,$C7,$D5,$FB
	.dc.b	$4F,$B8,$7E,$0E,$33,$C4,$AF,$F6,$0C,$61,$FB,$33,$C1,$E9,$3F,$AB
	.dc.b	$F7,$5F,$1F,$FA,$DE,$7F,$D4,$D7,$E8,$5A,$FE,$0C,$4F,$07,$F4,$CD
	.dc.b	$75,$1C,$31,$FD,$0B,$6F,$CD,$3C,$AF,$C4,$78,$7B,$24,$74,$8F,$36
	.dc.b	$7E,$5D,$E4,$F8,$23,$2F,$9E,$32,$F9,$F2,$FA,$6E,$CD,$10,$D4,$F2
	.dc.b	$7F,$C3,$07,$B9,$AF,$CD,$8E,$6E,$9A,$FF,$87,$F6,$FE,$31,$9F,$F0
	.dc.b	$0F,$9C,$79,$7F,$C6,$5B,$5F,$C3,$6B,$7A,$7C,$46,$C7,$DB,$9D,$BF
	.dc.b	$45,$B7,$FC,$32,$77,$3F,$C1,$3C,$2F,$3B,$4F,$4C,$7F,$2B,$7C,$D1
	.dc.b	$AF,$9C,$7C,$7F,$F2,$BF,$C5,$1F,$55,$4F,$E6,$13,$AB,$FE,$48,$F9
	.dc.b	$BD,$07,$C2,$FE,$3D,$BA,$F0,$CB,$F9,$C7,$BC,$F9,$EF,$E4,$98,$FF
	.dc.b	$B3,$BD,$2F,$C1,$CE,$77,$FC,$D7,$AB,$FD,$D2,$93,$97,$81,$F9,$F7
	.dc.b	$11,$B8,$FE,$E2,$68,$FF,$68,$94,$CB,$B5,$E9,$78,$2F,$F2,$BF,$32
	.dc.b	$F9,$BF,$EB,$D8,$FF,$5F,$D5,$F4,$5F,$AF,$DB,$FA,$35,$ED,$FE,$97
	.dc.b	$3A,$CB,$8C,$74,$F9,$FB,$1F,$DD,$B8,$DF,$55,$88,$60,$AD,$F9,$0B
	.dc.b	$DE,$1F,$D5,$7E,$7F,$C1,$87,$E6,$1A,$FF,$9C,$5A,$FE,$C1,$AF,$F0
	.dc.b	$77,$8E,$9E,$EF,$9F,$FF,$33,$57,$5D,$BC,$F2,$9C,$DF,$F0,$3B,$E3
	.dc.b	$7F,$E0,$49,$CD,$7F,$60,$9C,$D7,$F6,$39,$CD,$7F,$02,$F9,$D7,$CF
	.dc.b	$7F,$66,$AC,$35,$FE,$39,$1A,$9F,$D9,$B6,$3F,$6E,$DC,$3E,$B7,$CF
	.dc.b	$BD,$BF,$99,$39,$BF,$E4,$DC,$2A,$D4,$7F,$7F,$9D,$60,$AD,$5F,$2C
	.dc.b	$7E,$CC,$62,$F8,$46,$30,$E7,$7F,$C9,$F0,$E7,$7B,$8C,$7C,$0F,$E0
	.dc.b	$B0,$7C,$FF,$F3,$27,$C5,$19,$8C,$4F,$09,$FF,$0B,$9C,$3C,$5F,$6D
	.dc.b	$6D,$B1,$AA,$E3,$A7,$F8,$4B,$1F,$A3,$87,$F9,$02,$73,$5F,$D2,$7C
	.dc.b	$E7,$95,$23,$FC,$3B,$E6,$BE,$BF,$F3,$35,$AA,$FC,$AC,$E6,$BF,$AC
	.dc.b	$46,$2B,$6F,$3C,$3F,$B8,$46,$3E,$56,$57,$F8,$2B,$E8,$2D,$7F,$78
	.dc.b	$B5,$63,$4F,$F0,$55,$5B,$F4,$56,$9F,$3F,$9B,$9C,$3F,$C3,$2D,$E8
	.dc.b	$FF,$6A,$B7,$3F,$F1,$E5,$7F,$6F,$F1,$BD,$7E,$F7,$A1,$B1,$93,$E1
	.dc.b	$C6,$5A,$7F,$88,$BE,$7F,$F6,$FE,$B1,$BF,$3F,$86,$4E,$CF,$F1,$4C
	.dc.b	$63,$3F,$81,$F3,$BF,$33,$97,$E4,$B6,$BF,$C0,$8D,$9D,$C7,$D6,$9F
	.dc.b	$E1,$5F,$03,$FE,$98,$8F,$F2,$AF,$C1,$FF,$87,$9F,$F2,$E4,$66,$BF
	.dc.b	$C6,$F3,$97,$FB,$F4,$67,$8B,$FC,$A7,$39,$7F,$CB,$71,$9A,$FF,$4D
	.dc.b	$FA,$7F,$C6,$7E,$29,$F2,$5F,$C6,$3D,$57,$D2,$66,$30,$DB,$FF,$34
	.dc.b	$CE,$9F,$F3,$85,$72,$73,$3F,$C9,$7E,$2B,$FC,$FB,$1F,$E7,$0C,$5B
	.dc.b	$7F,$A4,$FC,$CF,$E0,$1B,$7F,$88,$EF,$6F,$F9,$52,$B7,$7F,$EA,$78
	.dc.b	$4E,$B5,$FF,$38,$D6,$A7,$FA,$2F,$87,$FE,$18,$AD,$5F,$FA,$9F,$1F
	.dc.b	$04,$E7,$D8,$FE,$6F,$87,$1C,$BC,$78,$F6,$9C,$DF,$08,$79,$CF,$9C
	.dc.b	$7C,$F6,$7B,$3F,$C9,$7D,$BB,$AD,$4E,$78,$BF,$59,$9F,$C0,$9C,$ED
	.dc.b	$FE,$AB,$9C,$ED,$FD,$B3,$2E,$B3,$9C,$7B,$7C,$73,$E7,$6F,$83,$86
	.dc.b	$D7,$FA,$D6,$74,$D9,$D3,$4F,$99,$C4,$F9,$DF,$57,$C7,$7F,$80,$67
	.dc.b	$31,$8F,$5B,$E1,$71,$7D,$28,$C4,$E8,$FF,$62,$57,$19,$FA,$67,$DD
	.dc.b	$9D,$B8,$39,$4F,$0A,$F9,$5D,$EF,$9D,$BE,$CE,$1B,$5F,$ED,$19,$D3
	.dc.b	$67,$4D,$3F,$DD,$DE,$47,$F9,$8B,$E0,$7E,$B7,$8C,$F9,$CF,$B9,$39
	.dc.b	$AF,$F6,$E4,$E5,$E3,$F0,$2B,$E3,$7C,$23,$13,$C2,$3F,$05,$F1,$6F
	.dc.b	$94,$BE,$1E,$1B,$5F,$EE,$99,$D3,$67,$4D,$35,$3C,$37,$F9,$7D,$BF
	.dc.b	$ED,$E8,$C7,$67,$EA,$A3,$FD,$75,$F3,$9F,$28,$F9,$7E,$25,$C3,$E4
	.dc.b	$45,$C3,$9F,$0A,$B1,$87,$38,$7C,$67,$D4,$3E,$AF,$CE,$5C,$7E,$94
	.dc.b	$5C,$39,$CD,$BF,$CF,$7C,$E7,$2B,$F9,$BD,$3F,$A5,$8C,$5F,$B3,$8F
	.dc.b	$F7,$B5,$FE,$A5,$BE,$38,$FD,$8C,$38,$36,$75,$DF,$FE,$52,$C7,$9D
	.dc.b	$FC,$3C,$39,$E3,$F9,$2F,$05,$5A,$BF,$F7,$96,$5F,$82,$F9,$CF,$E3
	.dc.b	$57,$F9,$5F,$9A,$FE,$DB,$EB,$FE,$2A,$73,$9F,$F0,$69,$CD,$6F,$C2
	.dc.b	$FF,$09,$36,$F8,$72,$FE,$A2,$B7,$7C,$6F,$6D,$AD,$ED,$FC,$7E,$B7
	.dc.b	$9F,$FA,$C9,$B8,$FF,$B2,$5B,$D4,$CF,$CB,$5B,$CB,$AB,$71,$E9,$18
	.dc.b	$C1,$7C,$73,$F3,$4E,$65,$8F,$81,$B1,$C6,$B8,$65,$D3,$09,$5F,$04
	.dc.b	$FE,$03,$3A,$71,$9F,$BD,$B9,$7F,$86,$F3,$F4,$C7,$B7,$1F,$16,$5E
	.dc.b	$13,$DF,$3A,$9F,$1E,$3E,$0C,$3A,$1F,$D3,$31,$F0,$38,$C7,$E1,$57
	.dc.b	$80,$F8,$37,$CF,$0E,$D8,$F8,$0F,$74,$65,$C6,$3D,$EF,$D1,$9F,$C8
	.dc.b	$8C,$61,$E9,$7E,$DB,$DC,$F3,$9E,$1C,$BC,$B5,$E4,$CB,$CD,$3C,$4F
	.dc.b	$93,$8F,$86,$FA,$C7,$29,$F7,$4E,$63,$B6,$5F,$1B,$1F,$24,$63,$2F
	.dc.b	$5C,$62,$FB,$CF,$3C,$7D,$79,$F6,$C3,$E1,$E5,$EA,$BD,$C6,$2F,$CB
	.dc.b	$87,$3C,$3D,$44,$8F,$9B,$19,$7D,$95,$EE,$CF,$E0,$C6,$2B,$83,$E3
	.dc.b	$38,$E3,$E3,$CB,$E8,$C7,$82,$31,$97,$D0,$BE,$73,$C4,$99,$7C,$3A
	.dc.b	$DC,$7D,$1C,$FE,$46,$5E,$CC,$7E,$2D,$6A,$3C,$D3,$AC,$BC,$B8,$73
	.dc.b	$C3,$E6,$64,$61,$F8,$E2,$9F,$80,$2C,$7F,$08,$73,$F0,$6F,$9C,$7C
	.dc.b	$E7,$09,$FF,$74,$46,$1C,$4E,$E7,$51,$FE,$F0,$9D,$3D,$BA,$7F,$C2
	.dc.b	$EF,$3B,$FE,$A9,$C5,$C1,$B1,$99,$DB,$C6,$7F,$89,$47,$2D,$3F,$90
	.dc.b	$37,$3F,$F4,$97,$CE,$7F,$49,$BD,$DC,$7F,$95,$56,$DF,$E1,$F3,$B7
	.dc.b	$33,$BB,$EA,$FF,$78,$9D,$46,$27,$64,$7F,$7D,$1A,$BF,$E4,$D7,$DE
	.dc.b	$BF,$8C,$B7,$0F,$7D,$F0,$E8,$FA,$EF,$52,$5F,$2C,$7C,$53,$A8,$C6
	.dc.b	$1E,$16,$D7,$38,$F5,$46,$08,$FC,$09,$E4,$FF,$39,$C3,$C8,$4B,$EF
	.dc.b	$1C,$8D,$37,$85,$97,$BE,$72,$78,$CB,$8F,$8D,$C1,$F0,$AB,$94,$F1
	.dc.b	$8C,$1A,$CB,$E2,$E5,$F7,$32,$F3,$CE,$B0,$F8,$AF,$69,$D3,$E7,$34
	.dc.b	$FC,$0C,$7C,$F5,$CA,$FA,$D7,$12,$4E,$9E,$51,$A7,$99,$F2,$4C,$E5
	.dc.b	$F1,$DF,$E9,$B8,$FA,$E7,$81,$27,$F2,$15,$D6,$BD,$D1,$C7,$1F,$8C
	.dc.b	$7B,$23,$B9,$E0,$9E,$11,$9A,$E9,$97,$D6,$AE,$6F,$CC,$C7,$DD,$87
	.dc.b	$5C,$BE,$16,$5E,$58,$C5,$FC,$08,$C5,$FB,$0E,$98,$F7,$AD,$E5,$E2
	.dc.b	$C3,$F1,$78,$79,$B2,$F9,$73,$98,$F9,$86,$6B,$E0,$E3,$DA,$3C,$E7
	.dc.b	$C8,$CB,$CF,$97,$86,$30,$5C,$BE,$FE,$1D,$63,$A4,$FB,$DD,$65,$EE
	.dc.b	$BE,$71,$DA,$FE,$6C,$FA,$32,$F9,$33,$D8,$D4,$71,$79,$9F,$16,$B9
	.dc.b	$D7,$C3,$BF,$8B,$5C,$F2,$F8,$B9,$7B,$6B,$58,$75,$69,$EA,$8D,$57
	.dc.b	$1B,$E7,$5F,$23,$1F,$2D,$FA,$23,$18,$76,$C7,$C7,$87,$19,$EC,$DA
	.dc.b	$ED,$3A,$9E,$38,$76,$7E,$49,$FB,$7C,$BE,$66,$3E,$0C,$3E,$14,$FC
	.dc.b	$43,$E1,$1F,$10,$FC,$A6,$5E,$A7,$A9,$E2,$2E,$3E,$27,$85,$7E,$1A
	.dc.b	$78,$47,$8A,$7D,$CD,$AF,$0C,$7C,$86,$9E,$D2,$C7,$66,$FC,$1B,$CC
	.dc.b	$FF,$65,$4F,$F5,$98,$C7,$89,$FD,$59,$B1,$99,$DB,$FD,$9A,$32,$BC
	.dc.b	$E7,$27,$F6,$D9,$CC,$62,$51,$89,$FE,$E5,$DC,$FD,$15,$E9,$7F,$74
	.dc.b	$C3,$84,$E6,$31,$5C,$2B,$78,$7F,$7B,$BD,$C6,$23,$C9,$3D,$BB,$1F
	.dc.b	$27,$CA,$58,$F5,$79,$57,$5C,$75,$39,$9E,$67,$0B,$E6,$F0,$3F,$CE
	.dc.b	$BF,$77,$FE,$FF,$7F,$C0,$9C,$48,$C4,$7F,$83,$A3,$1E,$6F,$F8,$11
	.dc.b	$B1,$C2,$77,$A7,$C9,$9C,$CF,$89,$B1,$F1,$AD,$F3,$7E,$D9,$F0,$7C
	.dc.b	$E7,$5C,$0D,$26,$1D,$DE,$11,$A8,$F6,$4F,$27,$DD,$93,$07,$D1,$70
	.dc.b	$C2,$57,$0C,$39,$3F,$22,$F8,$56,$A3,$D5,$3A,$CB,$DC,$DD,$7E,$C5
	.dc.b	$79,$3D,$4F,$51,$2F,$E3,$35,$E6,$ED,$F3,$E4,$FA,$1F,$C0,$43,$F0
	.dc.b	$27,$BF,$F8,$3E,$F7,$C9,$BE,$5B,$0D,$BF,$BC,$71,$AE,$77,$CA,$3F
	.dc.b	$C7,$AD,$7F,$C8,$8D,$7F,$C9,$AD,$7F,$CA,$2E,$F6,$FD,$34,$75,$3B
	.dc.b	$C6,$75,$FF,$2C,$62,$EB,$7B,$DE,$2F,$80,$F3,$E5,$31,$E3,$19,$7C
	.dc.b	$AF,$F9,$49,$F8,$4D,$C4,$D3,$84,$7E,$80,$D9,$F9,$FC,$3C,$5F,$45
	.dc.b	$AC,$FD,$37,$CA,$BC,$26,$F2,$57,$C8,$F8,$0F,$9B,$D5,$2B,$DF,$19
	.dc.b	$7E,$5B,$E5,$23,$83,$64,$3F,$6B,$8F,$C5,$F3,$7E,$FF,$CA,$BE,$4F
	.dc.b	$E0,$36,$BF,$CE,$6D,$7F,$CF,$4D,$7F,$D0,$7E,$63,$D2,$F9,$0D,$8F
	.dc.b	$F3,$DB,$5F,$F4,$13,$5F,$F4,$3F,$B3,$FC,$95,$79,$0F,$93,$FF,$48
	.dc.b	$F9,$AA,$D4,$E6,$BF,$D2,$D3,$9E,$2F,$F4,$C4,$E6,$53,$97,$AD,$FF
	.dc.b	$A4,$F8,$76,$D8,$C9,$FE,$9E,$8C,$B4,$C8,$FB,$07,$C2,$FE,$52,$3F
	.dc.b	$D4,$F1,$8E,$8F,$F4,$DB,$63,$FD,$44,$D7,$FD,$57,$F1,$EF,$90,$E5
	.dc.b	$1F,$EB,$18,$C5,$7F,$A9,$1B,$1F,$EA,$B6,$BF,$EB,$4F,$AF,$7C,$77
	.dc.b	$F4,$BD,$BF,$4F,$18,$BF,$CA,$47,$EF,$1F,$DB,$C6,$6B,$F3,$4D,$7A
	.dc.b	$57,$91,$46,$5F,$F6,$2B,$5F,$F6,$43,$CE,$BF,$AE,$72,$BE,$45,$F3
	.dc.b	$3E,$5C,$78,$E3,$71,$FF,$63,$D6,$A7,$CB,$F4,$79,$9F,$0B,$BC,$3F
	.dc.b	$B6,$75,$67,$34,$D2,$5F,$C0,$9B,$7E,$E9,$F6,$57,$0A,$FF,$68,$B6
	.dc.b	$7F,$DB,$0D,$7E,$74,$F8,$70,$FE,$07,$5E,$63,$F8,$03,$89,$F9,$7E
	.dc.b	$67,$94,$F4,$69,$FC,$BE,$BE,$8C,$65,$F0,$BE,$42,$38,$3A,$F9,$7B
	.dc.b	$AF,$DD,$DF,$F5,$C9,$FE,$A7,$F3,$CF,$4B,$E6,$79,$A8,$C4,$FF,$40
	.dc.b	$F9,$CC,$C6,$30,$FE,$D0,$D7,$FE,$00,$D7,$1B,$FC,$AB,$6F,$FB,$CD
	.dc.b	$6E,$82,$7C,$1D,$5D,$8F,$BD,$E6,$FF,$65,$F4,$B1,$87,$A2,$6D,$FF
	.dc.b	$63,$F2,$FF,$C6,$53,$92,$47,$9A,$79,$CE,$A3,$13,$A7,$D3,$8F,$F0
	.dc.b	$98,$F3,$D7,$AA,$31,$3F,$E0,$F5,$F2,$1F,$10,$EB,$F1,$76,$FE,$6C
	.dc.b	$71,$23,$F0,$9F,$DC,$E4,$FB,$BF,$E0,$1C,$5F,$E3,$66,$D7,$F8,$BF
	.dc.b	$AB,$E1,$C6,$32,$FF,$17,$B6,$F9,$76,$7B,$AB,$6F,$E9,$FD,$57,$E9
	.dc.b	$EF,$85,$C3,$72,$FF,$09,$E3,$D6,$B7,$5A,$3F,$87,$B8,$9F,$F2,$F5
	.dc.b	$71,$CB,$FC,$B4,$D7,$EE,$9F,$7B,$DA,$78,$AF,$7B,$F0,$3E,$1F,$E4
	.dc.b	$6B,$53,$97,$FB,$A7,$3F,$E1,$B9,$DE,$FB,$E0,$3F,$CC,$58,$71,$B6
	.dc.b	$5D,$B2,$EB,$8F,$8A,$31,$3D,$0E,$F3,$A7,$F2,$58,$75,$D3,$11,$D3
	.dc.b	$3F,$26,$3D,$F0,$E6,$D8,$CE,$3E,$7C,$39,$5F,$73,$96,$3D,$30,$E3
	.dc.b	$5E,$5B,$E3,$5E,$2C,$BD,$B3,$9B,$E1,$5A,$AE,$67,$AF,$3F,$8B,$7C
	.dc.b	$23,$DA,$78,$70,$EE,$E3,$2F,$A1,$5C,$33,$EB,$8C,$AE,$18,$F9,$2B
	.dc.b	$51,$89,$D5,$FC,$17,$E5,$4E,$A3,$13,$B2,$3E,$0C,$FE,$2B,$A2,$67
	.dc.b	$EF,$9F,$64,$EA,$31,$3A,$9F,$54,$7C,$FC,$FE,$04,$7C,$49,$D3,$67
	.dc.b	$59,$7A,$1E,$64,$8F,$A1,$19,$7A,$93,$0F,$64,$EE,$3A,$19,$C3,$CC
	.dc.b	$F0,$32,$6A,$78,$B9,$CF,$E9,$B9,$9E,$2F,$08,$C9,$F6,$27,$8C,$67
	.dc.b	$3F,$C1,$4E,$6B,$8B,$C2,$32,$7D,$98,$CB,$8C,$FD,$D7,$D2,$B8,$46
	.dc.b	$2B,$6F,$C3,$BF,$36,$7F,$3A,$B7,$1F,$73,$1F,$9A,$D3,$DD,$8F,$86
	.dc.b	$B8,$3E,$18,$D6,$3E,$18,$C6,$7F,$76,$31,$8F,$87,$1F,$76,$1C,$A3
	.dc.b	$17,$D7,$3B,$1E,$99,$D6,$3E,$33,$D3,$3A,$7C,$39,$FB,$9D,$61,$DA
	.dc.b	$75,$97,$8E,$B4,$FE,$32,$75,$5C,$31,$F4,$E7,$F0,$AB,$51,$89,$FC
	.dc.b	$69,$BC,$BC,$D8,$F6,$BF,$C9,$67,$F0,$2B,$47,$E8,$B3,$F5,$E3,$D3
	.dc.b	$0F,$D7,$E7,$E4,$C3,$F2,$99,$F9,$2B,$58,$7B,$B0,$E3,$5A,$CF,$E4
	.dc.b	$4F,$6A,$F7,$56,$A3,$EC,$E3,$DA,$31,$1F,$91,$8C,$67,$E9,$8F,$D1
	.dc.b	$4E,$B1,$FC,$2D,$FE,$FE,$31,$87,$C3,$CF,$DD,$3E,$63,$58,$73,$C3
	.dc.b	$EB,$9E,$5C,$BE,$2E,$7E,$F8,$F2,$9F,$16,$78,$65,$F0,$6B,$DD,$9F
	.dc.b	$CE,$AE,$47,$9A,$B7,$97,$C2,$C7,$C5,$7D,$B3,$F3,$E1,$C6,$7E,$5E
	.dc.b	$1F,$D7,$0F,$E6,$25,$C7,$FA,$6B,$70,$FE,$B4,$EF,$C7,$7C,$03,$C3
	.dc.b	$E5,$3F,$E9,$48,$C3,$89,$DC,$EA,$3F,$D2,$B3,$A7,$C1,$F0,$1F,$EA
	.dc.b	$27,$9E,$08,$C4,$6B,$CA,$E5,$39,$6C,$F1,$CB,$FA,$E3,$75,$FE,$D0
	.dc.b	$DB,$FF,$1B,$7B,$EF,$80,$FE,$FB,$7C,$30,$FE,$B9,$7B,$BF,$EE,$6F
	.dc.b	$1E,$26,$1C,$67,$75,$1F,$EA,$D5,$B9,$D4,$62,$75,$7D,$E3,$FB,$04
	.dc.b	$EA,$30,$4C,$7D,$35,$FD,$52,$75,$18,$AE,$84,$AF,$EC,$73,$A8,$FF
	.dc.b	$0C,$5F,$AF,$2F,$91,$1B,$BF,$F0,$A5,$6E,$3D,$F8,$79,$A7,$B6,$1D
	.dc.b	$32,$F8,$F1,$82,$57,$6A,$F3,$4E,$A3,$07,$72,$D7,$37,$B1,$8A,$E7
	.dc.b	$97,$CE,$AE,$53,$E4,$39,$16,$B9,$38,$C7,$DB,$1F,$26,$3E,$FC,$F2
	.dc.b	$9C,$DF,$A2,$79,$3F,$0A,$75,$97,$C7,$26,$3E,$CA,$F3,$9C,$0F,$41
	.dc.b	$D8,$CD,$F6,$73,$8F,$AD,$CD,$74,$7F,$03,$7E,$0B,$C9,$DE,$3D,$93
	.dc.b	$BC,$7D,$E7,$B6,$3C,$47,$8E,$FB,$3C,$2B,$E6,$C6,$27,$86,$1E,$5A
	.dc.b	$E1,$97,$E0,$63,$04,$C3,$F3,$58,$F9,$6B,$A1,$E8,$3C,$E7,$DD,$C3
	.dc.b	$B6,$5F,$36,$F9,$65,$EB,$CB,$C3,$8F,$AE,$BE,$86,$1C,$A7,$EC,$61
	.dc.b	$CA,$3E,$69,$EE,$C3,$E6,$46,$63,$17,$CD,$FA,$47,$9A,$B9,$E5,$ED
	.dc.b	$9D,$4F,$C0,$CB,$D9,$5E,$4C,$BC,$D9,$7A,$6F,$A6,$3D,$F2,$F5,$E3
	.dc.b	$EF,$CB,$A6,$1E,$9C,$3D,$73,$E1,$7F,$37,$3A,$8C,$1E,$6C,$7C,$0F
	.dc.b	$3C,$BC,$75,$E1,$BF,$1C,$FC,$2C,$3C,$37,$F8,$2A,$D4,$F1,$3C,$47
	.dc.b	$E4,$30,$F0,$4F,$63,$84,$73,$CB,$DB,$3A,$AE,$75,$AA,$F9,$B8,$7E
	.dc.b	$2A,$B5,$1C,$63,$CA,$E8,$F1,$1E,$E3,$E7,$1E,$42,$99,$C3,$ED,$CF
	.dc.b	$48,$E5,$5E,$19,$F7,$DF,$CA,$C7,$E2,$D7,$84,$FC,$91,$CF,$1F,$CB
	.dc.b	$BD,$EB,$AC,$F0,$8E,$B5,$F5,$AF,$E2,$3D,$63,$FD,$55,$C8,$F0,$1D
	.dc.b	$3F,$D6,$63,$0D,$8C,$CF,$F6,$29,$CB,$63,$2B,$AD,$CA,$85,$70,$75
	.dc.b	$19,$8C,$4F,$09,$75,$BD,$F7,$8D,$FE,$D6,$AD,$4E,$63,$18,$77,$8C
	.dc.b	$5F,$86,$B7,$7C,$A3,$15,$BF,$09,$F3,$3C,$A7,$C3,$5B,$7C,$E6,$31
	.dc.b	$D6,$3D,$A3,$2A,$F9,$C7,$61,$C8,$FD,$8F,$FD,$F9,$1F,$E0,$C8,$C7
	.dc.b	$77,$F8,$25,$B1,$99,$DB,$FE,$1B,$8C,$BE,$46,$E9,$F2,$5B,$A7,$B7
	.dc.b	$E0,$BD,$91,$E1,$F1,$3B,$60,$69,$30,$F0,$3C,$23,$51,$EB,$9E,$4F
	.dc.b	$B7,$35,$FD,$98,$F2,$57,$06,$DF,$33,$E4,$CE,$F1,$F4,$56,$DB,$3B
	.dc.b	$AE,$D5,$BC,$7F,$19,$8F,$CB,$F6,$7D,$64,$AF,$90,$D7,$C6,$FA,$CF
	.dc.b	$0F,$DD,$FD,$F8,$7B,$2F,$96,$FC,$17,$73,$DD,$7C,$2E,$1B,$7F,$74
	.dc.b	$E3,$5C,$EF,$94,$7F,$90,$DA,$FF,$92,$9A,$FF,$95,$5A,$FE,$9C,$E6
	.dc.b	$75,$78,$ED,$F3,$5B,$A7,$EB,$71,$75,$3C,$9F,$16,$BE,$7F,$7E,$65
	.dc.b	$CE,$32,$F9,$5F,$F2,$C3,$F0,$9B,$89,$A7,$08,$FD,$19,$B3,$F3,$D8
	.dc.b	$78,$BD,$FD,$67,$E7,$AB,$9F,$9F,$35,$7C,$7F,$7F,$E6,$54,$C7,$E0
	.dc.b	$46,$5F,$94,$F9,$08,$E0,$D9,$0F,$DB,$E3,$F1,$3D,$FF,$BD,$F1,$4F
	.dc.b	$3F,$DE,$DA,$FF,$3C,$35,$FF,$3F,$B5,$FF,$45,$7B,$AF,$7B,$E3,$B6
	.dc.b	$3F,$D0,$0D,$7F,$D0,$ED,$7F,$D1,$9E,$97,$F2,$55,$E4,$BD,$DF,$C0
	.dc.b	$79,$8A,$D4,$E6,$BF,$95,$4E,$78,$BF,$96,$4E,$65,$39,$7A,$DF,$FA
	.dc.b	$63,$87,$6D,$8C,$9F,$EA,$58,$CB,$4C,$8F,$73,$FA,$2F,$1F,$FA,$75
	.dc.b	$B1,$FE,$A4,$6B,$FE,$AF,$6B,$F9,$BF,$98,$7C,$D7,$F0,$58,$C3,$5F
	.dc.b	$E8,$D1,$AA,$FF,$55,$36,$3F,$D6,$3E,$DF,$F4,$29,$FD,$83,$63,$EF
	.dc.b	$E8,$F7,$3F,$A5,$3F,$C4,$E7,$4D,$F1,$BC,$14,$9E,$4F,$FB,$1D,$AF
	.dc.b	$EC,$DA,$FF,$5B,$79,$DF,$EC,$7D,$A7,$3F,$82,$38,$5D,$BF,$F6,$55
	.dc.b	$6A,$7D,$3E,$E6,$67,$C4,$EF,$A3,$FB,$57,$C1,$E7,$24,$D2,$5F,$EC
	.dc.b	$33,$56,$30,$FD,$AD,$F0,$AF,$F6,$A3,$63,$E5,$DF,$C2,$8C,$47,$C1
	.dc.b	$8F,$16,$2F,$77,$A6,$3F,$DD,$72,$4E,$71,$FE,$45,$F8,$87,$E7,$1B
	.dc.b	$CB,$EB,$BF,$D8,$46,$09,$97,$FB,$B2,$B8,$36,$BE,$AC,$62,$3F,$DF
	.dc.b	$4E,$AB,$C5,$6F,$7B,$E5,$4E,$70,$FA,$91,$F4,$7C,$6F,$EE,$1F,$0D
	.dc.b	$7C,$A3,$F5,$7C,$BF,$EB,$3F,$2F,$98,$98,$7F,$71,$7F,$2D,$7C,$A3
	.dc.b	$07,$F7,$5F,$4F,$A1,$CF,$CC,$7B,$58,$9F,$ED,$19,$7E,$A3,$6F,$EB
	.dc.b	$F7,$BA,$D6,$FF,$DA,$22,$CF,$F8,$7B,$0E,$78,$72,$9D,$D7,$F4,$7C
	.dc.b	$7F,$C1,$91,$E6,$C7,$F6,$33,$FE,$10,$6C,$FF,$84,$6F,$E1,$1E,$7F
	.dc.b	$62,$E1,$E6,$8E,$44,$7F,$C2,$4F,$ED,$F5,$7B,$FF,$CE,$A7,$FC,$04
	.dc.b	$DC,$7F,$C4,$74,$F7,$7E,$74,$E7,$4F,$F1,$CB,$6F,$FC,$70,$DB,$FF
	.dc.b	$20,$DF,$19,$EB,$E4,$FE,$AF,$D6,$7B,$EF,$98,$C3,$6B,$67,$F8,$C6
	.dc.b	$77,$3A,$3F,$88,$38,$8F,$F2,$E4,$F1,$C3,$FB,$91,$FE,$58,$70,$E3
	.dc.b	$DC,$FB,$BE,$FF,$F9,$77,$C5,$7B,$9F,$DE,$67,$35,$FD,$D2,$3F,$C2
	.dc.b	$FF,$01,$EB,$8C,$1F,$E6,$59,$DD,$72,$8C,$73,$30,$ED,$87,$58,$C4
	.dc.b	$F2,$E4,$63,$D7,$0E,$52,$C7,$AE,$1C,$71,$E7,$3D,$9B,$87,$2E,$2F
	.dc.b	$F3,$96,$1C,$B0,$E3,$7B,$3D,$27,$5C,$BC,$73,$E9,$BE,$5A,$76,$AD
	.dc.b	$63,$70,$E7,$18,$3D,$59,$FA,$73,$F3,$47,$69,$D9,$E2,$C3,$AC,$67
	.dc.b	$2F,$0D,$70,$CB,$D1,$18,$9F,$F3,$8E,$1D,$5F,$54,$EB,$1F,$25,$6F
	.dc.b	$0E,$B3,$9A,$ED,$5A,$3E,$3D,$F3,$8F,$54,$EC,$98,$7E,$03,$3F,$96
	.dc.b	$F9,$49,$3E,$27,$59,$F5,$8E,$E4,$AF,$99,$7D,$F3,$F8,$79,$FB,$6B
	.dc.b	$C3,$87,$27,$83,$D9,$C9,$F0,$B1,$F1,$BE,$12,$1A,$CF,$E0,$CF,$1B
	.dc.b	$EF,$9F,$C0,$BE,$37,$D0,$93,$F5,$71,$F3,$63,$E5,$8E,$33,$F7,$AB
	.dc.b	$8C,$63,$2F,$8F,$3E,$5C,$FE,$14,$62,$3C,$98,$74,$C7,$39,$7C,$AC
	.dc.b	$FC,$F5,$EA,$C7,$A4,$FC,$3C,$39,$46,$27,$B1,$F1,$8E,$19,$FE,$07
	.dc.b	$1E,$D1,$93,$C3,$9F,$A2,$7D,$84,$CB,$E2,$67,$E4,$AE,$71,$87,$96
	.dc.b	$1E,$1C,$FC,$F1,$F2,$2F,$93,$F1,$6F,$93,$9B,$E5,$97,$AA,$B6,$FE
	.dc.b	$13,$1F,$24,$F0,$C7,$E9,$46,$2F,$C3,$7E,$D8,$C5,$F3,$BE,$B9,$FD
	.dc.b	$FA,$EA,$FC,$93,$11,$ED,$CB,$D3,$8F,$CB,$C7,$E2,$E5,$E1,$C3,$8C
	.dc.b	$7C,$EC,$7B,$3F,$3E,$F8,$CE,$A3,$89,$E6,$C3,$B6,$7F,$4A,$F8,$9E
	.dc.b	$FC,$3B,$4E,$B3,$F5,$E7,$E6,$C3,$E5,$1C,$CF,$06,$1C,$E3,$58,$73
	.dc.b	$9D,$3F,$92,$CB,$CB,$87,$69,$E7,$3E,$2A,$FC,$85,$F3,$9F,$BE,$6A
	.dc.b	$3A,$9A,$C3,$CB,$1A,$AF,$01,$CD,$FD,$96,$1E,$07,$C7,$5F,$45,$E0
	.dc.b	$F3,$C7,$C9,$5E,$7C,$3C,$55,$F8,$73,$C4,$70,$3A,$BA,$CF,$F2,$4D
	.dc.b	$CF,$EB,$E3,$E1,$BF,$E4,$17,$FD,$7E,$7F,$A7,$B7,$3F,$EC,$AF,$F4
	.dc.b	$33,$FB,$3F,$92,$F8,$07,$7E,$A7,$FD,$2B,$18,$78,$CE,$E7,$4D,$8E
	.dc.b	$B9,$B9,$FF,$D4,$F1,$86,$BF,$DC,$DA,$FF,$71,$6B,$FD,$B5,$AF,$F6
	.dc.b	$36,$C7,$5E,$FF,$33,$67,$B4,$62,$76,$47,$04,$8D,$90,$F2,$39,$69
	.dc.b	$93,$CC,$E5,$A6,$4F,$43,$96,$99,$3D,$4E,$5A,$66,$7D,$4F,$80,$87
	.dc.b	$62,$9D,$0A,$72,$2B,$C9,$E0,$43,$81,$4E,$05,$78,$3A,$21,$A2,$BA
	.dc.b	$70,$47,$04,$9F,$A1,$1E,$D6,$BE,$66,$BE,$16,$BD,$9A,$F5,$6C,$75
	.dc.b	$9F,$94,$D9,$ED,$18,$9D,$91,$C1,$23,$64,$3C,$8E,$5A,$64,$F3,$39
	.dc.b	$69,$93,$D0,$E5,$A6,$4F,$53,$96,$99,$9F,$53,$E0,$21,$D8,$A7,$42
	.dc.b	$9C,$8A,$F2,$78,$10,$E0,$53,$81,$5E,$0E,$88,$68,$AE,$9C,$11,$C1
	.dc.b	$27,$E8,$47,$B5,$AF,$99,$AF,$85,$AF,$66,$BD,$5B,$1D,$67,$E5,$36
	.dc.b	$7B,$46,$27,$64,$70,$48,$D9,$0F,$23,$96,$99,$3C,$CE,$5A,$64,$F4
	.dc.b	$39,$69,$93,$D4,$E5,$A6,$67,$D4,$F8,$08,$76,$29,$D0,$A7,$22,$BC
	.dc.b	$9E,$04,$38,$14,$E0,$57,$83,$FC,$02,$BF,$9E,$76,$3C,$D4,$C9,$CC
	.dc.b	$62,$3F,$A3,$E3,$AA,$E1,$1B,$CB,$FA,$54,$62,$36,$1C,$5F,$48,$E1
	.dc.b	$5C,$EF,$B4,$E6,$7F,$D2,$47,$F5,$D3,$FA,$8E,$9F,$D8,$1F,$F4,$99
	.dc.b	$92,$18,$21,$FD,$8C,$A7,$42,$E3,$FE,$C4,$78,$9B,$21,$B2,$BB,$78
	.dc.b	$C7,$8D,$CB,$5E,$04,$8E,$0E,$B4,$FE,$BB,$3B,$9C,$CE,$E3,$04,$9D
	.dc.b	$92,$3E,$1B,$5F,$49,$C5,$C3,$E6,$6B,$E1,$38,$90,$E3,$3E,$22,$4F
	.dc.b	$86,$34,$73,$29,$CC,$E2,$ED,$B3,$E0,$7C,$87,$02,$9A,$31,$3E,$08
	.dc.b	$E0,$D8,$CC,$F5,$7C,$47,$12,$18,$21,$E5,$29,$E0,$2C,$75,$78,$9B
	.dc.b	$21,$B2,$BB,$78,$C7,$8D,$CB,$5E,$04,$8E,$0E,$A7,$DB,$3B,$9C,$CE
	.dc.b	$E3,$04,$9D,$92,$3E,$1B,$5F,$49,$C5,$C3,$E6,$6B,$E1,$38,$90,$E3
	.dc.b	$3E,$22,$4F,$86,$34,$73,$29,$CC,$E2,$ED,$B3,$E0,$7C,$87,$02,$9A
	.dc.b	$31,$3E,$08,$E0,$D8,$CC,$F5,$7C,$47,$12,$18,$21,$E5,$29,$E0,$2C
	.dc.b	$75,$78,$9B,$21,$B2,$BB,$78,$C7,$8D,$CB,$5E,$04,$8E,$0E,$A7,$DB
	.dc.b	$3B,$9C,$CE,$E3,$04,$9D,$92,$3E,$1B,$5F,$49,$C5,$C3,$FA,$0A,$FF
	.dc.b	$3D,$3F,$E8,$4C,$BF,$D0,$0D,$C7,$FC,$FF,$18,$8F,$6D,$72,$7D,$07
	.dc.b	$80,$87,$83,$0F,$49,$30,$F4,$46,$8E,$65,$39,$9C,$5D,$B6,$7C,$0F
	.dc.b	$AC,$E0,$53,$46,$27,$C1,$1C,$1B,$19,$9E,$AF,$88,$E2,$43,$04,$3C
	.dc.b	$A5,$3C,$05,$8E,$AF,$13,$64,$36,$57,$6F,$18,$F1,$B9,$6B,$C0,$91
	.dc.b	$C1,$D4,$FB,$67,$73,$99,$DC,$60,$93,$B2,$47,$C3,$6B,$E9,$38,$B8
	.dc.b	$7C,$CD,$7C,$27,$12,$1C,$67,$C4,$49,$F0,$C6,$8E,$65,$39,$9C,$5D
	.dc.b	$B6,$7C,$0F,$90,$E0,$53,$47,$F0,$BC,$3F,$C1,$04,$BF,$F0,$9E,$3F
	.dc.b	$E0,$C6,$CF,$F8,$40,$FF,$0A,$46,$71,$FF,$09,$46,$72,$FF,$0B,$D7
	.dc.b	$08,$FF,$15,$9F,$E2,$E2,$18,$26,$5F,$E2,$E6,$E9,$FE,$55,$7F,$C8
	.dc.b	$4D,$7F,$C7,$6D,$7F,$C6,$CD,$7A,$C7,$32,$46,$63,$AE,$9F,$E4,$F9
	.dc.b	$CB,$67,$51,$89,$F0,$CE,$48,$E0,$87,$91,$CB,$4C,$9E,$67,$2D,$32
	.dc.b	$7A,$27,$C2,$48,$DC,$65,$AF,$A8,$E2,$53,$C0,$43,$B1,$4E,$47,$09
	.dc.b	$F7,$39,$39,$1C,$08,$70,$34,$F0,$72,$70,$34,$61,$D3,$82,$1B,$23
	.dc.b	$A6,$CF,$C1,$7C,$CD,$7C,$2D,$7B,$35,$EB,$1C,$C9,$19,$8E,$B3,$F2
	.dc.b	$A7,$2D,$9F,$C3,$65,$FD,$02,$BF,$3C,$7F,$9E,$CF,$D1,$1F,$D0,$4F
	.dc.b	$D6,$D3,$46,$9F,$E8,$4D,$5C,$3F,$E8,$3A,$D4,$E6,$3F,$D1,$3B,$B8
	.dc.b	$1C,$4E,$2D,$67,$FE,$8E,$C7,$8A,$CB,$AB,$9C,$FC,$2E,$72,$F0,$46
	.dc.b	$27,$79,$77,$C7,$B5,$EC,$E5,$9F,$81,$63,$E0,$BE,$F1,$8C,$3A,$63
	.dc.b	$EA,$CF,$D3,$9F,$1C,$38,$4F,$8B,$1C,$5F,$0C,$19,$F8,$F0,$CD,$49
	.dc.b	$F2,$1D,$DF,$87,$3B,$8C,$E1,$D9,$F4,$98,$BE,$D3,$BC,$BD,$35,$BA
	.dc.b	$E8,$6E,$FA,$E5,$EA,$8C,$12,$B8,$E1,$D2,$75,$18,$AF,$43,$6B,$E6
	.dc.b	$61,$E1,$CB,$E1,$47,$36,$CF,$D4,$BE,$71,$82,$4F,$B2,$75,$9F,$D8
	.dc.b	$7B,$12,$77,$7D,$EB,$84,$72,$7D,$F3,$CB,$3F,$3C,$E7,$3F,$95,$7C
	.dc.b	$6F,$73,$9C,$FE,$06,$3E,$18,$C6,$7C,$AB,$ED,$E7,$EA,$7C,$24,$39
	.dc.b	$9A,$AE,$4E,$71,$FA,$87,$8A,$79,$47,$02,$46,$67,$8C,$63,$0F,$C6
	.dc.b	$3D,$B1,$F5,$57,$49,$CB,$C7,$2F,$45,$71,$C7,$F0,$38,$74,$AE,$11
	.dc.b	$ED,$C3,$9E,$1C,$89,$1C,$30,$E7,$5C,$30,$FC,$0E,$3E,$A8,$C6,$7F
	.dc.b	$5E,$7E,$15,$70,$C3,$9C,$70,$CB,$C5,$18,$AE,$19,$78,$B2,$E0,$78
	.dc.b	$6F,$F1,$87,$CB,$AF,$AF,$86,$A3,$35,$C5,$D9,$F1,$F2,$F3,$57,$7C
	.dc.b	$FF,$17,$18,$C3,$96,$3F,$89,$BE,$39,$FB,$2F,$F8,$16,$7E,$8B,$DC
	.dc.b	$63,$3B,$7F,$84,$CF,$C5,$8F,$C3,$8F,$16,$1C,$5E,$B9,$FA,$9B,$8F
	.dc.b	$D3,$CF,$C9,$18,$3E,$46,$3D,$9E,$6D,$CF,$F0,$11,$BC,$7E,$5D,$72
	.dc.b	$7C,$53,$A6,$E1,$F0,$EB,$EB,$9C,$73,$F8,$B9,$79,$EB,$79,$7D,$36
	.dc.b	$D7,$D0,$C7,$B5,$FE,$A2,$3C,$E7,$33,$E2,$46,$9F,$B5,$5C,$E7,$85
	.dc.b	$7A,$2F,$C1,$18,$3E,$75,$73,$3E,$09,$71,$F6,$46,$E3,$BC,$F2,$7B
	.dc.b	$9E,$4A,$F4,$9D,$F2,$F9,$59,$7D,$38,$FC,$D7,$23,$91,$CB,$FD,$33
	.dc.b	$01,$27,$51,$98,$C4,$F0,$C3,$FB,$74,$2B,$71,$86,$C3,$A7,$FA,$CC
	.dc.b	$61,$FE,$B7,$D4,$EA,$74,$FF,$5E,$9D,$E3,$DF,$0E,$58,$F9,$B0,$E1
	.dc.b	$7B,$8C,$61,$C9,$F5,$E1,$DB,$AB,$FA,$33,$E0,$9D,$61,$D0,$61,$D2
	.dc.b	$F8,$9D,$30,$ED,$3A,$C3,$9C,$E4,$DC,$7F,$66,$EB,$4F,$C9,$8C,$37
	.dc.b	$91,$3A,$AF,$97,$18,$9D,$4E,$6B,$CD,$39,$AE,$F8,$76,$E6,$F2,$CE
	.dc.b	$6B,$76,$EC,$E6,$4C,$3B,$C6,$5B,$1C,$A3,$13,$F1,$70,$F5,$F0,$E6
	.dc.b	$A5,$FC,$0C,$3D,$71,$F1,$AB,$73,$9C,$39,$EB,$FE,$1A,$BD,$CE,$6F
	.dc.b	$86,$7F,$E3,$58,$C1,$C7,$B7,$E1,$91,$F4,$B7,$3F,$7B,$CB,$AF,$D4
	.dc.b	$FD,$88,$77,$BA,$FE,$DF,$6B,$C0,$78,$1B,$AD,$65,$F7,$2B,$53,$9B
	.dc.b	$FB,$98,$74,$6D,$F0,$8F,$D2,$B5,$FF,$2A,$39,$DB,$F6,$0D,$C3,$F6
	.dc.b	$38,$7E,$9F,$CF,$ED,$1F,$8C,$B7,$88,$F1,$3D,$3A,$3C,$4F,$74,$E6
	.dc.b	$FD,$13,$97,$94,$66,$FF,$72,$D7,$FC,$D0,$D7,$F8,$19,$E0,$BF,$7C
	.dc.b	$71,$CF,$F8,$43,$6B,$D3,$0F,$17,$C7,$52,$7C,$2F,$D0,$9D,$46,$0F
	.dc.b	$2F,$8D,$EC,$F1,$FD,$6D,$8C,$9F,$C4,$31,$F8,$71,$E4,$24,$67,$14
	.dc.b	$72,$8F,$B5,$6F,$23,$17,$FC,$04,$79,$2F,$2B,$EE,$B6,$FC,$0D,$9F
	.dc.b	$E4,$2D,$7F,$D2,$1E,$5F,$EA,$1B,$D1,$F8,$68,$C4,$7C,$C6,$BF,$E9
	.dc.b	$7F,$0F,$F9,$8F,$2B,$CC,$FE,$5B,$18,$6C,$FE,$65,$B3,$B8,$FE,$73
	.dc.b	$E5,$3C,$DF,$E6,$11,$B3,$2E,$27,$84,$27,$86,$DD,$E3,$6D,$5E,$73
	.dc.b	$CE,$FA,$51,$86,$C6,$63,$EB,$35,$FE,$8F,$E7,$BA,$78,$B6,$3E,$E3
	.dc.b	$5D,$C7,$6F,$49,$CD,$F9,$4C,$7F,$AC,$46,$2B,$F4,$CD,$7F,$9E,$75
	.dc.b	$FF,$AC,$7D,$1E,$1C,$9F,$9C,$AD,$57,$FB,$1D,$AE,$70,$F0,$11,$FE
	.dc.b	$05,$EB,$3D,$3E,$70,$C5,$39,$8F,$1D,$69,$5E,$5F,$C8,$37,$2F,$5E
	.dc.b	$1D,$30,$FF,$64,$79,$3C,$6B,$59,$7A,$6B,$06,$F9,$3F,$B4,$E3,$F0
	.dc.b	$A3,$E4,$11,$FD,$45,$78,$5F,$81,$1D,$F3,$7A,$94,$EB,$3F,$33,$D6
	.dc.b	$F1,$11,$FE,$49,$83,$D6,$C4,$7F,$81,$31,$57,$D9,$24,$78,$9F,$07
	.dc.b	$AF,$23,$B1,$FD,$1F,$2F,$9B,$E4,$FF,$66,$6F,$95,$F0,$A7,$FB,$97
	.dc.b	$B1,$89,$FE,$09,$5B,$9F,$EB,$B1,$8A,$F8,$11,$FE,$FA,$5E,$CF,$42
	.dc.b	$47,$F6,$CF,$33,$FC,$49,$C7,$FD,$5E,$B7,$3F,$D6,$EB,$6F,$F6,$5A
	.dc.b	$DF,$0F,$59,$CD,$F1,$7F,$0F,$E6,$FF,$7B,$BF,$EE,$B1,$E5,$23,$FD
	.dc.b	$D1,$F4,$7B,$5F,$00,$91,$FE,$00,$F6,$FF,$C1,$AF,$F6,$D8,$DB,$FD
	.dc.b	$F1,$AF,$F3,$A3,$FC,$3B,$EE,$7F,$24,$7F,$C1,$2D,$7F,$BF,$37,$0F
	.dc.b	$F1,$AD,$3D,$DF,$4C,$E7,$2F,$F1,$93,$6F,$AF,$91,$EE,$BE,$11,$FE
	.dc.b	$24,$F5,$9E,$F7,$F8,$4E,$30,$D7,$FC,$65,$1A,$F1,$FB,$DF,$EB,$A3
	.dc.b	$13,$FE,$4D,$AE,$0F,$F5,$68,$CB,$EF,$6B,$F0,$7D,$B7,$AD,$FE,$01
	.dc.b	$87,$53,$DB,$FE,$BF,$5A,$F1,$FF,$CC,$73,$9A,$FF,$32,$CE,$5E,$35
	.dc.b	$FE,$68,$E7,$7C,$03,$E0,$3F,$B4,$E1,$C6,$D8,$71,$8C,$37,$0F,$0E
	.dc.b	$5E,$59,$F0,$63,$D3,$0E,$75,$AC,$BC,$AD,$C7,$C6,$79,$B0,$E5,$1E
	.dc.b	$5C,$BC,$93,$C0,$D5,$57,$CB,$5C,$1C,$9D,$30,$F1,$3B,$CB,$D0,$EB
	.dc.b	$2F,$5D,$71,$AD,$CF,$90,$EB,$7C,$67,$B4,$60,$ED,$87,$99,$E1,$87
	.dc.b	$67,$58,$76,$BF,$56,$1D,$A3,$13,$C3,$0E,$B7,$E6,$AD,$63,$F1,$EF
	.dc.b	$64,$7E,$36,$5F,$3B,$0F,$64,$6C,$8A,$B9,$BA,$21,$E6,$8E,$6F,$BF
	.dc.b	$18,$F9,$5F,$4E,$5E,$FC,$3C,$53,$CB,$1F,$44,$E6,$FC,$33,$D4,$F0
	.dc.b	$61,$E2,$70,$4A,$E3,$8F,$AE,$31,$97,$C4,$3E,$B6,$1D,$63,$C1,$19
	.dc.b	$CB,$CB,$5C,$6F,$E9,$12,$78,$C6,$5E,$15,$ED,$3E,$DC,$66,$BE,$0E
	.dc.b	$5E,$EC,$7D,$18,$73,$9C,$E5,$F2,$30,$E5,$7F,$11,$F8,$17,$C3,$1E
	.dc.b	$F3,$CF,$2F,$55,$7B,$B0,$E6,$FD,$78,$CE,$3D,$E3,$AE,$3D,$EF,$B4
	.dc.b	$E5,$FD,$8D,$70,$C3,$BD,$7E,$0B,$1F,$16,$3D,$B1,$F2,$DE,$E3,$18
	.dc.b	$FD,$29,$B8,$72,$7C,$99,$7C,$57,$C6,$7D,$CB,$E4,$E0,$EB,$1B,$CB
	.dc.b	$CF,$5F,$4A,$7A,$E1,$F0,$31,$F2,$C7,$1C,$7C,$4D,$AF,$14,$EA,$F9
	.dc.b	$DF,$E0,$2B,$99,$E4,$9D,$65,$F9,$2C,$39,$DF,$0A,$D6,$5E,$88,$C1
	.dc.b	$F2,$B2,$F4,$5F,$1C,$BD,$17,$E3,$CB,$D1,$87,$1A,$D4,$78,$E7,$47
	.dc.b	$6B,$E6,$FE,$62,$35,$87,$82,$BE,$B5,$F3,$AD,$1F,$94,$3D,$B3,$F8
	.dc.b	$9B,$E7,$1F,$89,$34,$78,$23,$4F,$0A,$EB,$3C,$31,$FC,$44,$7E,$2C
	.dc.b	$F4,$CF,$9C,$E0,$7D,$D9,$E1,$8F,$D2,$C3,$CD,$3E,$1C,$3E,$43,$C0
	.dc.b	$FC,$09,$70,$FA,$38,$76,$9F,$2C,$FE,$C4,$FE,$C1,$7F,$D7,$DB,$8F
	.dc.b	$FA,$A6,$FF,$AD,$47,$F8,$17,$C9,$7C,$07,$F9,$6E,$3F,$C1,$07,$FA
	.dc.b	$15,$C3,$B6,$BF,$EC,$68,$E3,$E2,$75,$F6,$9E,$FD,$25,$74,$C1,$3A
	.dc.b	$8C,$37,$3F,$EC,$AD,$D3,$FB,$77,$C0,$3E,$01,$D1,$FD,$4E,$B5,$87
	.dc.b	$F7,$66,$BF,$DC,$1C,$F1,$2F,$95,$71,$CE,$5F,$F5,$0B,$DD,$6A,$31
	.dc.b	$3A,$CB,$CA,$FF,$52,$9D,$46,$27,$64,$7F,$BD,$C6,$B2,$FF,$07,$DF
	.dc.b	$9B,$3F,$87,$87,$59,$F7,$E7,$EE,$C7,$94,$70,$9D,$47,$F5,$3C,$3A
	.dc.b	$3F,$06,$32,$FC,$17,$CD,$5C,$F1,$F3,$12,$FA,$57,$F7,$FB,$E6,$F1
	.dc.b	$3C,$27,$53,$55,$CE,$78,$D7,$39,$D3,$F0,$27,$51,$89,$D9,$1F,$54
	.dc.b	$69,$F0,$4F,$36,$CF,$C4,$CB,$DB,$8F,$C8,$9F,$2E,$7F,$77,$0E,$53
	.dc.b	$9C,$39,$37,$0E,$57,$EC,$3A,$9E,$53,$C7,$87,$6C,$FE,$B1,$F8,$4C
	.dc.b	$3A,$B8,$CF,$E9,$E5,$E1,$AE,$11,$8C,$FE,$5E,$5E,$1C,$7B,$1E,$0B
	.dc.b	$E0,$F8,$E3,$4E,$67,$8E,$1D,$E3,$DF,$18,$BE,$F1,$F5,$30,$B9,$FD
	.dc.b	$3C,$7A,$BF,$0B,$0E,$4F,$D5,$9D,$E7,$F4,$EB,$6F,$D5,$BF,$8A,$78
	.dc.b	$F3,$F8,$38,$F0,$8F,$8E,$7A,$71,$EB,$5F,$3B,$1F,$05,$7C,$3B,$E3
	.dc.b	$7B,$8F,$CA,$E3,$CE,$3B,$37,$3F,$91,$97,$92,$7E,$4C,$7C,$63,$39
	.dc.b	$FC,$38,$E8,$7C,$32,$E3,$F0,$E3,$4F,$92,$34,$D3,$58,$F9,$9C,$65
	.dc.b	$EB,$3E,$25,$F8,$73,$FC,$26,$5E,$76,$E1,$E0,$AF,$99,$8F,$81,$E3
	.dc.b	$3A,$C3,$AD,$7E,$47,$2F,$73,$F4,$31,$ED,$9F,$C6,$8E,$86,$F1,$E8
	.dc.b	$FC,$4C,$3A,$63,$DA,$BE,$86,$5E,$57,$96,$5E,$0B,$F9,$E7,$42,$B9
	.dc.b	$C7,$E6,$E7,$F0,$F0,$E9,$5F,$21,$B5,$F2,$30,$ED,$39,$3C,$98,$78
	.dc.b	$9E,$D8,$74,$9D,$D7,$E4,$E7,$38,$FA,$9E,$27,$E8,$31,$F1,$2E,$37
	.dc.b	$BE,$EC,$FE,$51,$3F,$EC,$58,$C6,$3F,$D7,$2B,$75,$AA,$1C,$EB,$6F
	.dc.b	$FB,$2D,$BD,$9F,$CC,$EB,$8D,$F0,$8F,$E6,$DD,$CF,$7D,$F0,$3F,$CE
	.dc.b	$67,$86,$1C,$AF,$BD,$EE,$B5,$18,$BE,$2F,$FB,$FF,$B3,$39,$2F,$84
	.dc.b	$F8,$4E,$17,$A8,$CC,$B1,$FF,$7D,$DE,$70,$EB,$39,$C3,$FD,$B5,$E0
	.dc.b	$3C,$8F,$89,$3F,$E9,$F8,$C4,$FF,$81,$39,$51,$89,$F9,$27,$2A,$E3
	.dc.b	$3A,$7D,$2D,$AF,$44,$FA,$FB,$DE,$76,$B0,$E9,$9F,$F6,$1C,$3B,$FA
	.dc.b	$5D,$24,$D2,$63,$E5,$8C,$BD,$A7,$8B,$E0,$D5,$8B,$C0,$F0,$61,$2B
	.dc.b	$86,$1C,$B3,$F8,$C7,$9E,$B4,$E3,$1E,$AD,$9E,$19,$7E,$13,$1E,$D8
	.dc.b	$74,$C7,$DD,$D1,$EB,$24,$7A,$5B,$87,$37,$EE,$53,$DE,$E3,$5F,$2A
	.dc.b	$9E,$CB,$DA,$FC,$3F,$6B,$DC,$7B,$BF,$8A,$BD,$D6,$A7,$38,$73,$C3
	.dc.b	$93,$FE,$42,$AD,$B6,$33,$1F,$E4,$66,$BF,$82,$36,$7C,$82,$EB,$FE
	.dc.b	$1B,$C7,$F2,$9E,$EF,$8E,$B9,$CF,$EE,$17,$8F,$E5,$F5,$34,$DC,$7C
	.dc.b	$CF,$E5,$B1,$F8,$2D,$AE,$33,$E7,$8C,$1C,$9F,$C8,$E3,$FA,$5F,$77
	.dc.b	$17,$E7,$BE,$2D,$BF,$97,$8A,$B6,$F9,$FE,$0B,$E5,$7C,$1F,$AD,$B9
	.dc.b	$7C,$B7,$C8,$47,$E5,$37,$0F,$03,$DB,$D0,$E5,$3F,$77,$DD,$76,$FB
	.dc.b	$DB,$1F,$55,$B7,$FE,$78,$6B,$F4,$FE,$13,$F7,$ED,$EF,$F9,$11,$88
	.dc.b	$FD,$CB,$5F,$B5,$E2,$FF,$24,$F2,$DE,$F7,$FA,$12,$30,$D9,$4E,$9E
	.dc.b	$26,$23,$FD,$0D,$D7,$FC,$4E,$30,$7F,$30,$F2,$FF,$77,$C1,$D2,$31
	.dc.b	$3B,$BE,$90,$AD,$BC,$8F,$F4,$5F,$C2,$FE,$5D,$A7,$E6,$7C,$94,$65
	.dc.b	$B1,$F9,$DF,$79,$EF,$F1,$6C,$66,$30,$75,$F6,$9F,$0F,$FC,$83,$4F
	.dc.b	$E9,$F1,$8A,$FD,$A3,$5F,$E1,$3D,$4D,$DF,$0F,$C7,$93,$F9,$4D,$EE
	.dc.b	$FF,$54,$D7,$38,$F8,$48,$FE,$6B,$DE,$FC,$A3,$7A,$BA,$3F,$D7,$2B
	.dc.b	$C5,$9F,$9B,$B7,$32,$94,$61,$B1,$FA,$6C,$DF,$01,$9A,$93,$EE,$7B
	.dc.b	$EB,$ED,$9C,$CF,$FB,$2A,$F9,$B6,$BE,$51,$1F,$57,$57,$2C,$CF,$2C
	.dc.b	$BE,$6B,$E5,$24,$F9,$7E,$02,$BD,$E3,$EB,$11,$FE,$35,$87,$47,$E4
	.dc.b	$7C,$07,$73,$8D,$FF,$4F,$F8,$97,$C4,$FC,$AF,$2D,$18,$BF,$E9,$8D
	.dc.b	$7F,$A2,$F9,$BF,$D3,$3C,$5E,$13,$99,$FE,$CB,$5C,$AF,$F9,$63,$63
	.dc.b	$FB,$64,$ED,$FF,$09,$C7,$83,$E2,$B9,$92,$3F,$AD,$79,$6F,$88,$C4
	.dc.b	$7F,$53,$B5,$63,$3F,$EA,$55,$79,$FA,$D6,$AF,$D7,$7F,$88,$D3,$FB
	.dc.b	$2D,$F4,$9F,$4D,$6D,$FE,$7D,$97,$95,$F4,$DF,$B2,$5F,$17,$F0,$89
	.dc.b	$3E,$8F,$32,$BC,$A7,$6F,$C2,$9E,$47,$F5,$1C,$5E,$FF,$F5,$D9,$FE
	.dc.b	$F4,$D7,$E8,$B7,$4F,$EE,$1D,$3E,$B8,$C6,$BF,$E3,$06,$D7,$F8,$AD
	.dc.b	$B7,$FE,$04,$AF,$D4,$FA,$DF,$E0,$5F,$81,$BE,$99,$A7,$5F,$42,$E3
	.dc.b	$FE,$4D,$9D,$36,$7F,$B5,$57,$38,$FE,$D9,$ED,$3D,$EF,$E5,$31,$FC
	.dc.b	$17,$D4,$FF,$15,$CE,$7C,$8F,$F2,$AF,$82,$9C,$DF,$08,$C5,$B9,$5E
	.dc.b	$FB,$E0,$3F,$C6,$B5,$AA,$65,$D5,$F0,$1F,$E6,$27,$11,$B6,$E3,$F9
	.dc.b	$4B,$E9,$5E,$48,$F1,$E5,$63,$96,$5E,$59,$CC,$76,$CB,$CB,$97,$92
	.dc.b	$73,$97,$7B,$E1,$7C,$B0,$E9,$87,$08,$EC,$7A,$A7,$6D,$9C,$E5,$E8
	.dc.b	$8F,$89,$97,$A1,$F7,$9B,$2C,$6D,$C6,$1F,$33,$0F,$04,$6B,$0E,$F5
	.dc.b	$B3,$B4,$EA,$BD,$58,$78,$AB,$67,$B2,$75,$8F,$C0,$26,$1E,$F3,$E3
	.dc.b	$63,$E7,$C3,$9C,$78,$70,$E8,$7C,$49,$D5,$72,$9F,$4B,$D2,$FB,$C7
	.dc.b	$5B,$E6,$49,$FB,$99,$7D,$39,$D4,$73,$CB,$DB,$87,$69,$C9,$0E,$E7
	.dc.b	$CE,$AE,$4E,$09,$3C,$A3,$2F,$C9,$CB,$E2,$E5,$EB,$9C,$C6,$2B,$E0
	.dc.b	$1F,$12,$33,$3F,$0A,$33,$8F,$D7,$C3,$CD,$1E,$0B,$E5,$39,$BE,$33
	.dc.b	$C3,$1F,$8D,$7C,$31,$F2,$63,$D7,$0E,$17,$6F,$E2,$63,$F6,$F0,$E0
	.dc.b	$F6,$8E,$78,$F6,$BF,$9D,$7F,$23,$2F,$A6,$F9,$71,$E8,$FC,$92,$99
	.dc.b	$3E,$01,$8C,$BC,$AE,$0F,$51,$F4,$DF,$9C,$EF,$1F,$23,$D4,$D9,$D6
	.dc.b	$FD,$AE,$27,$AC,$60,$F4,$C6,$F1,$FC,$3B,$D6,$FD,$38,$7B,$2F,$8D
	.dc.b	$7C,$2A,$D1,$E6,$9D,$61,$E4,$8F,$C4,$C6,$2F,$78,$78,$6F,$E7,$C7
	.dc.b	$6C,$BE,$7B,$A8,$F1,$DF,$A7,$0F,$7E,$1C,$E3,$15,$F8,$D7,$81,$F4
	.dc.b	$EF,$9B,$A3,$E6,$61,$F0,$31,$F5,$C7,$C9,$8C,$61,$DA,$B8,$C6,$2F
	.dc.b	$B5,$73,$8F,$C0,$E1,$E3,$75,$1D,$B1,$F6,$BF,$16,$F9,$C6,$27,$E1
	.dc.b	$4F,$DD,$C3,$C1,$97,$E3,$6B,$47,$C2,$AE,$A7,$82,$F9,$DF,$1C,$7C
	.dc.b	$B3,$AA,$FA,$D3,$E7,$6C,$66,$3E,$35,$F7,$79,$1E,$92,$C7,$78,$D4
	.dc.b	$7E,$1C,$F9,$13,$FD,$09,$FF,$63,$65,$FD,$45,$B8,$FF,$62,$6E,$1F
	.dc.b	$ED,$1F,$29,$F0,$0F,$17,$0D,$AF,$F6,$81,$FE,$85,$70,$FF,$B4,$E7
	.dc.b	$8C,$FF,$70,$F2,$9E,$FD,$25,$72,$C9,$3A,$8C,$37,$5F,$EC,$8D,$D3
	.dc.b	$FC,$6F,$F0,$0F,$80,$74,$FF,$52,$AD,$61,$FD,$B5,$AF,$F7,$27,$3C
	.dc.b	$4B,$E5,$5C,$73,$97,$FD,$5A,$F7,$5A,$8C,$4E,$B2,$F2,$BF,$DA,$27
	.dc.b	$51,$89,$D9,$1F,$EB,$D1,$AC,$BF,$C2,$37,$E6,$CF,$E1,$E1,$D6,$7D
	.dc.b	$F9,$FB,$B1,$E5,$1C,$27,$51,$FE,$13,$C3,$A3,$F0,$63,$2F,$C1,$7C
	.dc.b	$D5,$CF,$1F,$31,$2F,$A5,$7F,$82,$6F,$9B,$C4,$F0,$9D,$4D,$57,$39
	.dc.b	$E3,$5C,$E7,$4F,$C0,$9D,$46,$27,$64,$7D,$51,$A7,$C1,$3C,$DB,$3F
	.dc.b	$13,$2F,$6E,$3F,$22,$7C,$B9,$FD,$DC,$39,$4E,$70,$E4,$DC,$39,$5F
	.dc.b	$B0,$EA,$79,$4F,$1E,$1D,$B3,$FA,$C7,$E1,$30,$EA,$E3,$3F,$A7,$97
	.dc.b	$86,$B8,$46,$33,$F9,$79,$78,$71,$EC,$78,$2F,$83,$E3,$8D,$39,$9E
	.dc.b	$38,$77,$8F,$7C,$62,$FB,$C7,$D4,$C2,$E7,$F4,$F1,$EA,$FC,$2C,$39
	.dc.b	$3F,$56,$77,$9F,$D3,$AD,$BF,$56,$FE,$29,$E3,$CF,$E0,$E3,$C2,$3E
	.dc.b	$39,$E9,$C7,$AD,$7C,$EC,$7C,$15,$F0,$EF,$8D,$EE,$3F,$2B,$8E,$49
	.dc.b	$1F,$90,$DD,$F1,$9F,$EF,$C9,$FF,$31,$74,$BC,$17,$C2,$FE,$3A,$F7
	.dc.b	$E6,$FE,$42,$B5,$39,$C3,$96,$3F,$E7,$49,$CE,$3F,$E6,$DC,$38,$BD
	.dc.b	$B1,$EA,$E6,$F8,$E9,$FE,$84,$BE,$38,$7F,$9F,$B0,$E2,$F5,$C3,$8B
	.dc.b	$DF,$E3,$B8,$EC,$85,$A7,$31,$85,$F1,$DE,$B7,$33,$BD,$70,$66,$6E
	.dc.b	$9F,$1A,$9E,$97,$13,$97,$4D,$9D,$45,$C7,$CD,$37,$5F,$77,$C7,$FF
	.dc.b	$AA,$EF,$E8,$BF,$C4,$33,$39,$BF,$98,$E4,$C5,$72,$E2,$FA,$91,$8D
	.dc.b	$7E,$15,$F2,$9C,$F5,$69,$B8,$F8,$1F,$E1,$7C,$AF,$8A,$7C,$87,$DF
	.dc.b	$E2,$7C,$3E,$A7,$38,$7D,$B8,$C5,$6D,$B1,$93,$E4,$7C,$97,$CF,$F9
	.dc.b	$2F,$A0,$D8,$C9,$F3,$63,$2D,$3F,$D5,$3D,$1F,$ED,$B7,$F2,$FF,$26
	.dc.b	$F6,$BF,$8F,$4E,$78,$FF,$25,$39,$AE,$3E,$27,$E6,$63,$13,$FE,$AC
	.dc.b	$AE,$1C,$9F,$91,$9C,$BF,$ED,$18,$CF,$1F,$CC,$9C,$BF,$8F,$AF,$D3
	.dc.b	$DF,$5C,$7F,$4D,$F2,$AF,$2E,$FC,$AE,$1B,$A7,$EC,$27,$4D,$9D,$34
	.dc.b	$FD,$57,$CB,$3E,$5F,$F6,$5F,$17,$FE,$00,$9C,$C6,$3C,$3F,$DC,$56
	.dc.b	$E7,$F7,$B5,$BD,$3F,$59,$39,$7F,$71,$19,$E0,$F8,$1E,$5F,$F6,$7F
	.dc.b	$3F,$FB,$47,$63,$E6,$1E,$AF,$D0,$AD,$67,$F9,$09,$CB,$F6,$23,$39
	.dc.b	$7E,$D2,$72,$FD,$18,$CD,$F5,$BE,$97,$FB,$BC,$3C,$38,$70,$C7,$CB
	.dc.b	$39,$3C,$15,$FD,$7F,$C0,$FF,$07,$5F,$A3,$37,$CB,$F8,$B4,$56,$B0
	.dc.b	$FD,$3D,$69,$E9,$3F,$C2,$E7,$4F,$F0,$C9,$D2,$BF,$F0,$63,$D3,$BF
	.dc.b	$D7,$3A,$9C,$E6,$F4,$7D,$FE,$8B,$E6,$3C,$93,$9C,$FC,$3B,$AB,$51
	.dc.b	$E4,$CF,$E9,$61,$CB,$0E,$2D,$9F,$4F,$85,$FE,$44,$DF,$E6,$F8,$4F
	.dc.b	$80,$FA,$73,$9A,$EF,$39,$78,$F2,$FD,$6D,$BF,$03,$39,$CF,$E0,$57
	.dc.b	$F2,$FB,$E5,$18,$3F,$98,$CE,$EB,$F1,$3E,$3B,$E1,$7E,$FB,$70,$F7
	.dc.b	$AF,$7B,$33,$27,$2F,$08,$C6,$1D,$F1,$FF,$30,$FC,$CF,$BB,$E0,$7D
	.dc.b	$BA,$23,$18,$FB,$1B,$7C,$67,$53,$F0,$AB,$8F,$67,$C8,$F8,$0B,$3F
	.dc.b	$19,$B7,$EF,$E4,$FE,$D1,$18,$DB,$E3,$B6,$BE,$36,$2C,$3E,$0F,$C8
	.dc.b	$67,$16,$9F,$E7,$A9,$C6,$5F,$D6,$6B,$53,$FE,$7E,$BF,$B7,$F2,$DE
	.dc.b	$E6,$E3,$F6,$1B,$8F,$B5,$B9,$FE,$5B,$E5,$DF,$19,$FB,$4B,$7C,$D7
	.dc.b	$D7,$7F,$8E,$4E,$9F,$D1,$9A,$E5,$FD,$54,$67,$1F,$C0,$4E,$4F,$CB
	.dc.b	$46,$70,$FF,$38,$46,$67,$F0,$C2,$75,$1F,$DE,$4D,$4F,$9F,$0F,$36
	.dc.b	$1E,$5B,$FE,$FD,$6F,$99,$FE,$1B,$EB,$BE,$6D,$E5,$79,$A7,$31,$8C
	.dc.b	$FC,$B1,$89,$D4,$7F,$00,$AE,$51,$8C,$3F,$39,$18,$9E,$26,$EF,$A4
	.dc.b	$FB,$30,$F2,$4F,$FA,$FA,$3D,$D8,$7F,$06,$7E,$6B,$5F,$5B,$57,$C7
	.dc.b	$3E,$3F,$F9,$4E,$AF,$93,$F8,$F5,$FE,$1F,$AD,$61,$FC,$62,$B5,$B7
	.dc.b	$F1,$F9,$CD,$7E,$D6,$73,$5F,$C2,$E7,$31,$FC,$FE,$FA,$D7,$8B,$0F
	.dc.b	$25,$72,$8F,$EC,$06,$6F,$DD,$19,$8F,$EC,$9F,$04,$F9,$8F,$E7,$BF
	.dc.b	$38,$F9,$DF,$E6,$53,$9C,$3F,$8F,$CE,$5F,$E3,$D1,$E4,$7F,$C7,$9C
	.dc.b	$0F,$9E,$FC,$36,$3F,$8A,$AD,$3F,$EE,$68,$CE,$BF,$37,$CB,$FF,$22
	.dc.b	$7B,$EF,$93,$FF,$78,$C6,$1B,$69,$D3,$EE,$9D,$3F,$E4,$DF,$85,$FF
	.dc.b	$08,$5E,$21,$18,$79,$5F,$F8,$16,$31,$7D,$27,$7E,$A7,$F4,$97,$67
	.dc.b	$F9,$67,$E7,$DF,$27,$FE,$0D,$C7,$FD,$B1,$AB,$DA,$FE,$9F,$18,$6E
	.dc.b	$FF,$E1,$3B,$E3,$87,$F4,$77,$F9,$13,$F3,$BE,$83,$A4,$FF,$91,$B8
	.dc.b	$51,$89,$F1,$78,$D5,$7C,$AF,$47,$F0,$26,$34,$FC,$F3,$4F,$D0,$46
	.dc.b	$DF,$C3,$BC,$73,$FE,$CF,$F2,$8F,$A1,$EC,$DA,$FE,$BD,$5B,$CF,$FC
	.dc.b	$5C,$D8,$FF,$3F,$4E,$DF,$E1,$6D,$BF,$EE,$73,$B9,$FF,$2E,$7A,$7E
	.dc.b	$6F,$82,$34,$91,$FB,$78,$CB,$FD,$BF,$CA,$FE,$EC,$DF,$4B,$E2,$4E
	.dc.b	$A3,$F8,$DB,$70,$FF,$01,$D7,$09,$FE,$3E,$D8,$FF,$03,$CE,$F2,$FF
	.dc.b	$25,$36,$BF,$D3,$55,$B9,$FF,$27,$36,$3F,$D3,$D3,$B7,$FC,$F1,$E2
	.dc.b	$FD,$79,$7B,$14,$F3,$C4,$9F,$64,$65,$FF,$06,$57,$F8,$15,$B3,$97
	.dc.b	$D3,$3C,$5F,$ED,$17,$F8,$7F,$2F,$FD,$91,$F3,$17,$D2,$3F,$8B,$F3
	.dc.b	$3F,$CA,$17,$C0,$FF,$07,$C6,$CD,$79,$7F,$9A,$8F,$EC,$7C,$1F,$0E
	.dc.b	$BF,$23,$C1,$CE,$3E,$36,$3F,$EB,$DF,$83,$C4,$CB,$FF,$10,$D6,$BD
	.dc.b	$3F,$F4,$DC,$62,$7C,$DF,$2B,$23,$04,$3F,$BE,$68,$8D,$47,$F8,$3E
	.dc.b	$FC,$83,$E6,$FE,$07,$24,$9F,$0C,$79,$CC,$3D,$23,$0F,$F8,$83,$E8
	.dc.b	$FE,$56,$46,$98,$FA,$3B,$3B,$21,$FE,$11,$94,$7C,$5F,$67,$E9,$7C
	.dc.b	$F4,$AF,$87,$19,$7F,$C7,$F1,$97,$D9,$1F,$BF,$F8,$A7,$47,$DD,$FA
	.dc.b	$47,$C7,$53,$51,$2F,$FC,$AF,$3A,$6C,$E5,$FE,$FF,$F4,$BE,$D7,$FC
	.dc.b	$BB,$F4,$DF,$96,$6F,$57,$F0,$F9,$D4,$70,$DB,$F1,$D5,$B8,$FF,$18
	.dc.b	$37,$B7,$FC,$E3,$5C,$1F,$F3,$9F,$C9,$BE,$9F,$F2,$DF,$07,$FC,$67
	.dc.b	$CE,$FF,$02,$E3,$FE,$93,$F1,$3F,$D7,$51,$88,$5F,$17,$FC,$E0,$D7
	.dc.b	$FC,$81,$5C,$FD,$AF,$F5,$05,$F2,$7F,$D8,$95,$BC,$7C,$78,$F8,$AB
	.dc.b	$58,$78,$40,$C7,$C5,$7C,$7C,$1F,$EB,$3F,$50,$F8,$0E,$3E,$03,$7F
	.dc.b	$C9,$63,$CB,$1F,$6C,$E7,$AF,$BF,$9E,$C7,$95,$EF,$7E,$B9,$74,$CB
	.dc.b	$9C,$F8,$E7,$CB,$8F,$38,$E0,$7F,$67,$BE,$6E,$C4,$7F,$A6,$7A,$3C
	.dc.b	$FF,$08,$F9,$BA,$62,$25,$74,$AF,$F5,$04,$EA,$31,$5F,$EA,$97,$C2
	.dc.b	$7D,$6F,$9C,$F8,$7F,$1F,$FD,$B7,$7F,$F7,$D7,$AB,$F3,$AB,$47,$0E
	.dc.b	$25,$F0,$99,$39,$8E,$13,$A8,$F2,$5F,$5D,$7E,$1B,$63,$EC,$CE,$DF
	.dc.b	$85,$9B,$C0,$FE,$AD,$97,$D4,$7F,$1C,$DC,$3E,$DC,$69,$F9,$ED,$AF
	.dc.b	$96,$F0,$8C,$9C,$3C,$DF,$C5,$F9,$1F,$E6,$0F,$90,$FC,$2F,$8E,$F5
	.dc.b	$29,$AB,$4D,$33,$8F,$D1,$AE,$0F,$DE,$BE,$8F,$D5,$ED,$FF,$12,$5F
	.dc.b	$E1,$1D,$C7,$C0,$6E,$7F,$92,$77,$1F,$84,$6E,$87,$75,$CF,$2F,$CE
	.dc.b	$C6,$88,$E0,$CD,$7E,$8B,$E3,$BF,$DC,$CF,$E2,$5A,$FD,$26,$E1,$F4
	.dc.b	$7A,$38,$71,$CA,$E0,$FD,$28,$E9,$EF,$C9,$E8,$F7,$FA,$2C,$47,$58
	.dc.b	$B7,$FB,$0C,$15,$FB,$0F,$A9,$7C,$6F,$D4,$C2,$1F,$AA,$C5,$3A,$8C
	.dc.b	$35,$FC,$5F,$07,$C8,$3E,$5C,$F0,$AE,$C7,$49,$D9,$83,$ED,$7C,$E7
	.dc.b	$3C,$BF,$86,$FC,$B7,$F9,$7E,$7F,$70,$7F,$20,$71,$1F,$C9,$B8,$95
	.dc.b	$C7,$87,$F8,$ED,$69,$B8,$FE,$B6,$31,$87,$FB,$72,$B8,$3F,$A6,$8C
	.dc.b	$BD,$E7,$81,$E6,$C7,$F6,$33,$AE,$63,$4E,$B8,$F9,$27,$75,$CA,$FD
	.dc.b	$77,$C2,$3A,$46,$72,$F8,$31,$72,$FE,$63,$87,$3A,$C1,$E7,$9B,$19
	.dc.b	$8E,$B8,$F8,$E7,$39,$7B,$A7,$D8,$6A,$33,$97,$9A,$30,$4C,$7E,$75
	.dc.b	$71,$CB,$D1,$39,$CB,$E6,$BF,$CB,$72,$F1,$5F,$AA,$34,$79,$CD,$5F
	.dc.b	$3C,$BC,$D1,$AA,$E3,$7D,$2F,$91,$1F,$87,$7C,$72,$F8,$77,$EC,$AF
	.dc.b	$3E,$5E,$B7,$84,$78,$5C,$BC,$32,$F5,$C6,$B1,$F0,$CE,$F1,$F1,$11
	.dc.b	$FC,$45,$F5,$8C,$47,$85,$F2,$C6,$63,$91,$27,$AB,$B3,$25,$73,$7E
	.dc.b	$4A,$EA,$D9,$F0,$9E,$9C,$BE,$4E,$1E,$C8,$F9,$E7,$53,$A6,$5F,$1D
	.dc.b	$CD,$73,$C3,$BD,$F2,$C3,$DD,$18,$CB,$B1,$F9,$6B,$E3,$5B,$7C,$66
	.dc.b	$23,$F1,$6F,$03,$EE,$61,$E1,$C7,$E9,$C7,$93,$2F,$8C,$F9,$F2,$F2
	.dc.b	$57,$9E,$3F,$DF,$71,$F9,$06,$E1,$FC,$15,$B8,$7F,$B0,$1F,$F6,$FF
	.dc.b	$CE,$3E,$73,$1E,$FA,$31,$3F,$EC,$68,$C4,$EA,$73,$3E,$0F,$57,$FA
	.dc.b	$4B,$FE,$28,$6B,$FD,$B1,$AF,$F8,$D1,$B8,$9F,$EA,$FA,$7F,$8A,$1B
	.dc.b	$B7,$EF,$3E,$71,$F2,$4F,$1B,$FC,$05,$5A,$C7,$B7,$46,$0F,$EC,$17
	.dc.b	$C2,$B1,$3F,$C3,$2B,$83,$E3,$AD,$DF,$58,$D4,$70,$C7,$C8,$EA,$F1
	.dc.b	$8F,$BE,$2D,$38,$99,$78,$63,$99,$30,$E6,$FF,$03,$8D,$46,$09,$7D
	.dc.b	$67,$58,$F9,$67,$31,$F0,$0E,$93,$F1,$E7,$4F,$19,$DC,$FB,$CE,$78
	.dc.b	$FA,$1C,$47,$4C,$3D,$37,$CA,$7B,$4F,$2C,$FE,$C6,$5F,$0E,$7C,$26
	.dc.b	$B1,$F0,$1E,$3C,$7B,$57,$09,$DE,$3D,$A7,$39,$FD,$4A,$D6,$3F,$D6
	.dc.b	$73,$F8,$B8,$F4,$C7,$B1,$31,$E4,$F8,$8F,$5C,$6E,$33,$5E,$9C,$3B
	.dc.b	$1F,$89,$8C,$CF,$5A,$E2,$75,$BF,$0B,$E0,$32,$6A,$78,$B9,$C3,$D3
	.dc.b	$5E,$1A,$E3,$3C,$27,$2F,$5C,$FE,$73,$9C,$3C,$B9,$FC,$DC,$BD,$0F
	.dc.b	$E3,$B2,$EF,$5B,$8C,$63,$DB,$1E,$B3,$93,$EF,$9F,$15,$C9,$E2,$AE
	.dc.b	$6E,$30,$F0,$CF,$0C,$3C,$2F,$D8,$BE,$D3,$A9,$FC,$44,$63,$0F,$B6
	.dc.b	$7A,$72,$F9,$98,$75,$8C,$61,$F0,$73,$F4,$E7,$F0,$AB,$E2,$EE,$F9
	.dc.b	$CF,$F6,$CD,$7F,$97,$7B,$3F,$29,$39,$F3,$BF,$2D,$DA,$E7,$FF,$32
	.dc.b	$DF,$F8,$D7,$0E,$39,$FE,$F4,$61,$C6,$7F,$D7,$27,$FA,$EA,$33,$3C
	.dc.b	$23,$13,$B9,$E4,$DD,$FF,$DA,$6D,$DB,$FD,$5F,$E8,$FF,$92,$AF,$C5
	.dc.b	$9F,$9B,$C0,$F1,$E1,$FE,$8B,$8C,$7C,$E6,$59,$5F,$EA,$BA,$DE,$1F
	.dc.b	$E8,$CA,$CE,$5F,$E9,$69,$7C,$E7,$53,$B7,$5D,$C3,$94,$5A,$E1,$37
	.dc.b	$2F,$F6,$7E,$7F,$E3,$6B,$FF,$5D,$3F,$A0,$DC,$E6,$FF,$D4,$AE,$4C
	.dc.b	$57,$2E,$2F,$A7,$18,$D3,$E1,$C6,$3A,$B2,$DC,$7B,$BF,$C3,$F9,$DE
	.dc.b	$EB,$DB,$FB,$BC,$8F,$9C,$D4,$E7,$0F,$F5,$04,$62,$A4,$7C,$9F,$A2
	.dc.b	$BB,$7C,$7F,$9D,$FA,$5F,$0D,$F4,$DB,$5F,$3D,$B3,$B8,$FA,$53,$FB
	.dc.b	$6E,$97,$CE,$FE,$43,$CA,$7C,$46,$63,$0D,$BF,$C6,$E1,$F8,$DB,$E3
	.dc.b	$18,$81,$FE,$BC,$AF,$EA,$51,$F9,$38,$C3,$B6,$9F,$E8,$09,$E6,$FF
	.dc.b	$B6,$3D,$77,$C2,$F8,$63,$FD,$55,$F3,$BF,$A6,$F2,$9F,$2F,$63,$0D
	.dc.b	$C7,$FD,$64,$DA,$FF,$59,$3F,$B0,$F9,$C7,$CE,$FE,$C3,$E1,$BF,$BD
	.dc.b	$5F,$FA,$8E,$72,$FB,$3B,$1D,$7E,$EE,$D5,$F0,$C9,$5F,$AA,$C3,$8C
	.dc.b	$62,$FE,$2F,$86,$9C,$D7,$4A,$60,$F6,$1F,$43,$E7,$D3,$EC,$7C,$07
	.dc.b	$F7,$DF,$7F,$FB,$F5,$B8,$DF,$3B,$AE,$0F,$A7,$5A,$9C,$C7,$AF,$7F
	.dc.b	$B5,$87,$F0,$BD,$2C,$FF,$8B,$79,$53,$99,$FB,$B1,$89,$D4,$F8,$DE
	.dc.b	$32,$C3,$CB,$F5,$4E,$AD,$4F,$2B,$AF,$FB,$53,$C2,$CE,$BF,$33,$E7
	.dc.b	$F3,$E3,$7F,$84,$72,$F7,$F3,$27,$37,$34,$F7,$9F,$57,$8D,$F4,$59
	.dc.b	$CF,$C9,$3F,$C9,$78,$5E,$F7,$83,$D9,$7C,$4F,$92,$73,$5D,$2F,$EA
	.dc.b	$D6,$A3,$11,$F4,$6F,$81,$EA,$F8,$6F,$F2,$94,$EA,$F9,$D7,$88,$FF
	.dc.b	$17,$7A,$9F,$51,$FD,$37,$95,$FE,$4B,$E8,$FA,$0D,$BF,$D5,$4F,$09
	.dc.b	$CD,$71,$E1,$FF,$20,$5E,$EB,$B7,$1B,$CE,$7D,$07,$D2,$AD,$5F,$E2
	.dc.b	$A7,$39,$7B,$EF,$85,$FD,$49,$7D,$05,$89,$5B,$9E,$38,$F8,$1D,$C6
	.dc.b	$67,$C1,$97,$EF,$23,$D7,$97,$F9,$A7,$C4,$F5,$92,$33,$3B,$C3,$E4
	.dc.b	$56,$E7,$C1,$F5,$7D,$2F,$AF,$87,$F1,$09,$3C,$9B,$7C,$B0,$FE,$13
	.dc.b	$3E,$9F,$A1,$E2,$FD,$FC,$7F,$CF,$97,$8B,$F6,$55,$DB,$E7,$E1,$CA
	.dc.b	$7A,$57,$C7,$AD,$61,$FE,$FC,$AF,$E0,$D5,$FD,$D7,$0F,$F4,$44,$7D
	.dc.b	$A2,$4F,$C0,$7E,$39,$F4,$7D,$9B,$1F,$90,$F9,$E7,$C0,$7E,$B2,$1F
	.dc.b	$40,$F9,$3F,$D6,$D6,$A7,$F3,$46,$A3,$35,$FB,$DC,$95,$B7,$F1,$ED
	.dc.b	$7F,$9D,$35,$FF,$07,$3B,$D7,$FC,$28,$DB,$FE,$87,$EE,$3E,$81,$F1
	.dc.b	$DE,$09,$F5,$7D,$07,$F0,$4A,$D7,$91,$E8,$8C,$5F,$F3,$0A,$D9,$FE
	.dc.b	$C1,$C1,$3E,$96,$C6,$63,$07,$37,$F7,$5F,$36,$FA,$C7,$CA,$67,$5F
	.dc.b	$DA,$1F,$0A,$F1,$AF,$FA,$AE,$F8,$56,$30,$FF,$17,$7C,$FB,$E8,$3F
	.dc.b	$8E,$DB,$E8,$7E,$16,$5F,$D9,$2B,$57,$FC,$2E,$B5,$C3,$FC,$8A,$73
	.dc.b	$5F,$BE,$6B,$FD,$A5,$AF,$83,$0F,$B1,$1C,$2B,$FC,$88,$DB,$FE,$E7
	.dc.b	$97,$BF,$E3,$7B,$5F,$F3,$08,$7D,$0F,$E0,$5B,$5F,$EE,$D6,$E3,$FC
	.dc.b	$41,$AF,$F7,$B7,$C6,$F6,$AE,$44,$9F,$F7,$B4,$7F,$69,$3F,$CA,$FF
	.dc.b	$21,$FC,$6B,$D4,$FE,$C6,$D8,$C9,$FE,$07,$8C,$BF,$E0,$CF,$8B,$FF
	.dc.b	$0A,$E9,$88,$46,$1E,$26,$3D,$BF,$EA,$4E,$23,$5E,$C3,$0F,$F3,$37
	.dc.b	$CF,$BE,$5B,$CE,$DB,$FE,$5E,$1C,$3F,$CC,$9A,$FF,$87,$5F,$F0,$BB
	.dc.b	$C1,$AE,$FC,$2E,$A7,$C4,$F9,$AE,$79,$FF,$9E,$DB,$AF,$F9,$E5,$B5
	.dc.b	$FE,$2C,$F9,$DF,$F3,$FB,$FD,$42,$3C,$2E,$F1,$E8,$47,$D3,$1C,$94
	.dc.b	$B8,$BF,$D0,$D8,$7F,$A2,$E3,$13,$D6,$F3,$87,$FA,$2F,$E1,$7F,$C3
	.dc.b	$F3,$FD,$81,$AF,$EA,$1B,$7F,$E9,$5F,$9A,$7B,$7C,$AB,$FC,$8B,$5A
	.dc.b	$BE,$9E,$BE,$67,$FC,$CE,$DB,$E3,$1D,$E3,$F1,$75,$E0,$8F,$F2,$63
	.dc.b	$5E,$67,$D8,$7B,$3F,$E5,$49,$FE,$FA,$FF,$78,$6B,$E9,$9F,$14,$FF
	.dc.b	$29,$8F,$77,$85,$83,$C9,$EE,$D8,$ED,$1E,$42,$38,$3B,$7B,$5E,$B3
	.dc.b	$83,$FD,$47,$1F,$89,$E4,$7F,$44,$D3,$FA,$5C,$7C,$7F,$3F,$FB,$C7
	.dc.b	$3E,$BC,$9F,$E6,$D7,$FD,$2B,$0F,$E9,$51,$83,$3C,$7F,$02,$4D,$3E
	.dc.b	$1D,$FF,$80,$FE,$17,$C3,$19,$DE,$9E,$1F,$3F,$FD,$31,$8F,$F6,$39
	.dc.b	$D3,$67,$55,$C2,$3F,$BD,$CF,$D3,$58,$78,$AB,$F5,$6D,$8D,$FC,$CF
	.dc.b	$88,$8F,$F8,$2F,$EB,$6B,$FE,$24,$B4,$F2,$6C,$72,$3F,$C4,$B9,$7F
	.dc.b	$87,$FC,$BF,$51,$23,$FC,$2D,$E2,$D7,$FC,$2D,$17,$83,$FC,$45,$3F
	.dc.b	$13,$D6,$95,$F0,$5D,$70,$FF,$A7,$B4,$F9,$B1,$E3,$24,$FC,$27,$F9
	.dc.b	$0F,$C5,$B9,$FF,$3C,$D9,$FF,$24,$B5,$FF,$1B,$FD,$77,$F9,$E9,$CC
	.dc.b	$FE,$0F,$D8,$7C,$7D,$39,$C7,$EA,$7E,$5F,$FB,$6D,$7F,$95,$1A,$FF
	.dc.b	$5C,$6E,$BF,$E8,$8F,$7B,$F5,$B3,$9D,$FF,$21,$1F,$E8,$08,$FF,$21
	.dc.b	$35,$D1,$FE,$60,$F9,$27,$CE,$7E,$E5,$B1,$98,$FC,$D7,$C2,$B9,$B5
	.dc.b	$1F,$D1,$2B,$75,$FE,$5C,$6B,$FE,$C7,$6B,$FE,$CC,$0F,$0C,$7F,$D3
	.dc.b	$9E,$0E,$7C,$9F,$F5,$57,$37,$FA,$8E,$31,$EA,$71,$BE,$1C,$DF,$EB
	.dc.b	$2F,$9B,$7C,$E6,$B3,$79,$69,$CC,$7F,$A9,$CC,$B8,$7F,$D7,$87,$FA
	.dc.b	$E3,$83,$FD,$F8,$EE,$79,$57,$5C,$FF,$D7,$9D,$96,$FF,$33,$7C,$AF
	.dc.b	$6F,$7A,$7C,$0F,$FA,$5A,$30,$BE,$73,$A3,$63,$3D,$78,$33,$32,$32
	.dc.b	$7A,$CF,$97,$F3,$9F,$05,$A6,$76,$47,$A7,$E0,$FF,$D6,$13,$98,$FA
	.dc.b	$9B,$FD,$76,$C7,$CF,$6A,$3A,$9F,$EC,$FA,$EA,$FB,$FE,$3F,$FB,$A3
	.dc.b	$7D,$0F,$24,$EA,$3E,$7E,$9F,$E9,$BF,$3B,$E6,$78,$DF,$EE,$99,$CB
	.dc.b	$79,$BF,$06,$FF,$5C,$8C,$CF,$0E,$59,$1F,$3E,$75,$5F,$EF,$3F,$0B
	.dc.b	$FD,$FF,$E6,$FE,$15,$B5,$F8,$36,$BE,$0C,$F5,$87,$EB,$2F,$F2,$7F
	.dc.b	$26,$F7,$7F,$04,$DB,$E5,$5B,$3E,$36,$8F,$03,$FC,$45,$18,$78,$4F
	.dc.b	$C2,$6E,$1F,$A0,$77,$0A,$E4,$FE,$46,$37,$5C,$F0,$C4,$C3,$F2,$13
	.dc.b	$A7,$93,$6B,$9C,$7E,$57,$E3,$FF,$2A,$DE,$6F,$CF,$46,$23,$E9,$F3
	.dc.b	$E2,$64,$6C,$FA,$11,$C3,$D9,$91,$D5,$E7,$F3,$98,$8F,$C0,$DF,$FA
	.dc.b	$81,$D1,$E3,$8F,$DD,$53,$E7,$1F,$1B,$86,$D7,$ED,$5A,$FE,$33,$CA
	.dc.b	$7C,$8E,$EB,$F3,$37,$CD,$B7,$C3,$AB,$F8,$A4,$E7,$0F,$CF,$46,$27
	.dc.b	$AF,$0F,$E0,$A7,$F0,$27,$53,$F1,$5F,$39,$AE,$2F,$E2,$DF,$2F,$FE
	.dc.b	$3E,$AF,$CE,$75,$23,$13,$FD,$46,$B8,$5F,$EB,$6B,$59,$7E,$E6,$F7
	.dc.b	$39,$6F,$3A,$F8,$CF,$F2,$C9,$DB,$6B,$91,$E6,$D3,$F6,$73,$AC,$3B
	.dc.b	$D7,$F2,$23,$81,$FC,$93,$0E,$D5,$C2,$F9,$CE,$EB,$8E,$5F,$AF,$C7
	.dc.b	$C7,$18,$CB,$E1,$DF,$0A,$D6,$1C,$F0,$E5,$97,$8C,$F6,$E3,$C9,$C5
	.dc.b	$F1,$C3,$E4,$5F,$1C,$3E,$6D,$6B,$0F,$46,$1C,$70,$E1,$1F,$25,$F8
	.dc.b	$4E,$0F,$7D,$72,$8C,$C7,$91,$D6,$3E,$B8,$FF,$11,$D7,$57,$44,$AF
	.dc.b	$6B,$C2,$FC,$91,$88,$D6,$3F,$11,$A7,$4C,$7D,$91,$99,$F2,$4F,$B2
	.dc.b	$72,$FA,$CE,$39,$7D,$67,$C1,$97,$C0,$8C,$47,$63,$DF,$7D,$4F,$96
	.dc.b	$F0,$39,$E3,$ED,$C3,$C2,$49,$DC,$F3,$CB,$E2,$3D,$0F,$51,$CE,$78
	.dc.b	$38,$38,$65,$F1,$2B,$9C,$EA,$79,$1C,$F0,$F4,$C6,$6F,$B9,$F4,$E3
	.dc.b	$35,$EB,$AF,$89,$5E,$09,$CC,$62,$74,$F5,$C3,$DD,$18,$BE,$91,$89
	.dc.b	$E0,$4A,$E6,$F9,$A7,$C7,$87,$E6,$32,$FA,$B3,$98,$C3,$FA,$83,$FC
	.dc.b	$48,$FF,$B0,$9B,$97,$FB,$5A,$BF,$67,$1F,$E2,$73,$FD,$CB,$F0,$CF
	.dc.b	$9C,$FF,$6D,$FB,$A8,$C4,$EC,$DC,$EA,$33,$1F,$E3,$3F,$57,$FA,$8B
	.dc.b	$FE,$2B,$6B,$FC,$45,$AF,$F8,$D9,$B6,$7F,$B0,$6D,$FC,$71,$BB,$7F
	.dc.b	$9A,$FE,$71,$F2,$2E,$2F,$F0,$95,$6B,$1E,$DD,$78,$3F,$B9,$5F,$0A
	.dc.b	$C4,$FF,$0D,$AE,$0F,$8E,$B7,$7D,$63,$51,$C3,$1F,$23,$AB,$C6,$3E
	.dc.b	$F8,$B5,$FE,$9C,$C7,$C0,$F2,$25,$F2,$7F,$91,$46,$A3,$04,$AE,$93
	.dc.b	$AC,$3C,$93,$98,$F7,$9D,$27,$E3,$4E,$9E,$33,$B9,$F7,$1C,$F0,$F3
	.dc.b	$B8,$8E,$9C,$2C,$3C,$97,$C8,$95,$E0,$AF,$F1,$14,$EA,$31,$3A,$7A
	.dc.b	$3E,$4B,$EF,$9F,$DF,$CB,$E3,$E3,$DF,$2F,$46,$1C,$88,$72,$3E,$63
	.dc.b	$A7,$04,$AE,$8F,$B4,$F8,$87,$27,$39,$7A,$27,$94,$63,$1F,$43,$6B
	.dc.b	$BE,$3E,$8C,$FF,$0F,$39,$C7,$D1,$39,$CB,$D7,$9F,$C3,$C3,$C9,$87
	.dc.b	$03,$F2,$D7,$C2,$3C,$A6,$AB,$F1,$8E,$30,$F1,$3F,$06,$78,$67,$F6
	.dc.b	$E7,$27,$8F,$0F,$03,$8C,$3C,$18,$74,$AE,$11,$8C,$3E,$34,$62,$7D
	.dc.b	$39,$FC,$4A,$D4,$FE,$13,$0E,$B1,$8C,$FF,$07,$97,$91,$F9,$31,$9C
	.dc.b	$7E,$49,$EB,$71,$3B,$8F,$BC,$D7,$DD,$5D,$B1,$F4,$5F,$68,$C4,$FF
	.dc.b	$A8,$FE,$73,$FD,$BF,$A7,$FA,$AD,$B1,$FE,$37,$E7,$FC,$5F,$0A,$73
	.dc.b	$A3,$1F,$F4,$84,$E7,$D2,$DD,$6A,$7F,$CD,$FD,$0C,$33,$53,$2F,$03
	.dc.b	$F9,$8E,$C5,$6E,$72,$7F,$A2,$67,$5E,$0F,$5D,$3F,$CE,$4B,$E7,$33
	.dc.b	$32,$B8,$1F,$E9,$18,$C5,$7F,$A6,$BE,$73,$99,$D3,$A6,$D7,$FA,$5E
	.dc.b	$2C,$FF,$A8,$26,$EF,$FE,$E2,$CB,$FC,$91,$7F,$EC,$B7,$F7,$59,$FA
	.dc.b	$BA,$7D,$91,$83,$78,$7F,$98,$F9,$73,$3A,$6D,$F3,$6C,$EE,$3F,$D4
	.dc.b	$9E,$FB,$DB,$FD,$FF,$32,$31,$3F,$EA,$FF,$9B,$CC,$62,$FF,$C9,$0D
	.dc.b	$D3,$FD,$64,$DE,$3F,$F5,$2F,$12,$F8,$E5,$F4,$3E,$6F,$E8,$FC,$C5
	.dc.b	$CF,$FD,$78,$DC,$FF,$D8,$AD,$CF,$FD,$8A,$FF,$0F,$F2,$1F,$37,$4F
	.dc.b	$C0,$79,$AF,$85,$C6,$3F,$EE,$9A,$DB,$6B,$73,$AC,$7F,$CC,$95,$A6
	.dc.b	$CE,$AB,$81,$FE,$AB,$F1,$7F,$1D,$E3,$3E,$4A,$C6,$1B,$3F,$EB,$16
	.dc.b	$DF,$FA,$91,$FD,$1F,$CD,$BE,$6E,$9F,$99,$F5,$1F,$0F,$8C,$3F,$DE
	.dc.b	$9C,$1F,$A7,$C5,$E1,$31,$E5,$5A,$E6,$FD,$5C,$62,$BF,$DE,$78,$F4
	.dc.b	$AD,$47,$29,$D6,$1F,$EF,$5D,$DF,$0D,$FD,$C7,$DB,$FE,$E5,$69,$CE
	.dc.b	$08,$4F,$F7,$7E,$E7,$BC,$E6,$E5,$7B,$BE,$58,$E1,$E5,$8F,$47,$9C
	.dc.b	$EE,$31,$8F,$5C,$3A,$4F,$86,$3C,$B3,$A8,$E1,$7D,$6F,$BC,$6C,$FF
	.dc.b	$7C,$DF,$28,$EE,$78,$A7,$87,$81,$FE,$08,$C3,$E0,$B6,$FF,$E0,$0E
	.dc.b	$AF,$9A,$DF,$7C,$9E,$D8,$FC,$5B,$E0,$FF,$BF,$EF,$84,$EA,$3F,$DC
	.dc.b	$35,$E7,$E9,$7B,$34,$F8,$DB,$BE,$0B,$FC,$89,$87,$1A,$DB,$6B,$83
	.dc.b	$E2,$AE,$1E,$6E,$31,$F7,$F8,$5F,$E3,$FF,$99,$FA,$4D,$BE,$1E,$13
	.dc.b	$E6,$33,$E7,$2E,$92,$B7,$37,$0F,$97,$93,$0E,$55,$A3,$AF,$CC,$FB
	.dc.b	$23,$35,$C1,$BC,$7F,$8B,$6D,$7F,$0C,$AE,$18,$79,$3C,$29,$3E,$8F
	.dc.b	$99,$B8,$FB,$B4,$FE,$09,$3F,$53,$5F,$F2,$F5,$F8,$6F,$79,$FF,$93
	.dc.b	$E7,$FC,$C5,$E0,$7C,$03,$CD,$F3,$37,$0E,$58,$FF,$01,$6C,$E9,$A4
	.dc.b	$CB,$C5,$19,$6C,$67,$0F,$05,$73,$9F,$85,$7D,$DF,$54,$7D,$23,$E4
	.dc.b	$70,$5B,$F1,$C5,$9F,$95,$36,$FF,$73,$F2,$EF,$63,$F4,$8D,$C3,$F2
	.dc.b	$FE,$9B,$E6,$1F,$11,$AB,$95,$F9,$7B,$E1,$97,$EA,$30,$E1,$7A,$8F
	.dc.b	$D5,$89,$DE,$3E,$2C,$3F,$80,$5F,$0C,$3A,$47,$93,$E5,$E4,$7E,$AB
	.dc.b	$E5,$1F,$30,$F8,$EC,$E7,$DB,$36,$18,$CF,$FD,$11,$87,$0B,$C6,$3F
	.dc.b	$C4,$6F,$6F,$AE,$FF,$89,$56,$E3,$04,$C3,$B4,$EB,$1F,$14,$E7,$1F
	.dc.b	$E4,$BF,$2E,$F9,$6F,$E4,$F8,$7F,$1F,$8C,$57,$FA,$E6,$11,$87,$FD
	.dc.b	$78,$D7,$FD,$88,$E6,$BA,$D7,$87,$E1,$7F,$8E,$4E,$6F,$B4,$E0,$D3
	.dc.b	$FC,$36,$35,$7F,$CD,$F2,$7C,$BF,$F2,$56,$D7,$E0,$BE,$89,$F3,$B9
	.dc.b	$AD,$63,$2B,$F8,$67,$D7,$BC,$6C,$65,$F8,$03,$FA,$B9,$FD,$5B,$E2
	.dc.b	$7F,$DB,$1E,$6F,$F8,$F9,$B1,$93,$FD,$C5,$19,$7F,$A6,$7C,$23,$E5
	.dc.b	$FF,$A6,$C6,$C4,$67,$87,$84,$62,$77,$E9,$30,$FE,$A1,$7B,$AE,$5F
	.dc.b	$34,$F9,$2F,$EA,$98,$FF,$5E,$8C,$4F,$F3,$56,$BF,$C7,$3E,$C1,$F1
	.dc.b	$7C,$37,$FE,$B9,$5F,$E1,$1A,$E1,$E2,$E5,$FE,$CD,$19,$7F,$51,$F6
	.dc.b	$3F,$49,$FE,$85,$F3,$3D,$B2,$FE,$DE,$DC,$3F,$B4,$36,$7F,$B8,$7C
	.dc.b	$E7,$F8,$09,$E3,$A7,$75,$3B,$6C,$66,$78,$47,$ED,$A7,$DA,$FF,$75
	.dc.b	$36,$7E,$5D,$CB,$E2,$9F,$CD,$3F,$E4,$7F,$9B,$F0,$7C,$19,$A7,$8F
	.dc.b	$E6,$BF,$C1,$13,$FD,$7E,$74,$D9,$DB,$FD,$8E,$76,$FF,$0A,$9D,$B6
	.dc.b	$76,$FF,$0E,$9D,$E5,$FE,$16,$AD,$B6,$B6,$FF,$86,$A7,$79,$7F,$84
	.dc.b	$E9,$F3,$9E,$A2,$61,$E3,$F7,$AB,$E3,$8B,$87,$39,$B7,$CE,$6D,$7F
	.dc.b	$89,$70,$FF,$1D,$BF,$D2,$27,$F0,$9E,$63,$CA,$FF,$16,$E3,$FD,$0A
	.dc.b	$7F,$11,$E8,$FF,$69,$E8,$DD,$6A,$7F,$C6,$17,$C5,$C9,$C3,$D1,$FD
	.dc.b	$46,$3F,$D7,$F8,$3E,$1C,$F0,$6C,$2B,$F0,$FB,$F1,$F8,$4F,$2D,$6A
	.dc.b	$73,$5C,$EB,$4D,$21,$FE,$0C,$9E,$B9,$FF,$80,$6F,$9D,$4A,$DE,$9F
	.dc.b	$DF,$67,$BF,$CB,$C9,$F4,$3E,$BF,$9B,$C4,$75,$8B,$5F,$E0,$59,$F8
	.dc.b	$1E,$6C,$AF,$F0,$CB,$F0,$FC,$BC,$47,$F8,$36,$2D,$7F,$83,$67,$DF
	.dc.b	$EB,$4A,$F9,$0E,$B6,$F9,$0E,$63,$DC,$E5,$FF,$31,$7C,$4B,$8F,$F6
	.dc.b	$4D,$8F,$F1,$F3,$5F,$F1,$2F,$CE,$7F,$8A,$DC,$CF,$02,$3F,$E3,$98
	.dc.b	$D3,$FD,$B3,$E5,$7F,$BA,$BF,$E4,$96,$B9,$CF,$F1,$3E,$EF,$E7,$A3
	.dc.b	$FC,$ED,$3A,$D3,$FC,$F6,$DC,$FF,$CE,$8D,$9F,$F2,$B7,$C0,$FF,$0B
	.dc.b	$F9,$27,$CE,$52,$4F,$F9,$37,$E0,$DC,$38,$8F,$F0,$0D,$70,$7F,$CB
	.dc.b	$CD,$7F,$C8,$B3,$C1,$CD,$71,$CB,$FD,$1F,$5B,$7F,$D0,$17,$FC,$AF
	.dc.b	$C7,$7C,$E3,$E7,$3F,$D1,$9E,$57,$FA,$96,$73,$18,$F8,$8F,$F4,$B4
	.dc.b	$63,$3F,$F5,$0F,$53,$87,$FA,$FC,$7F,$A9,$27,$4B,$0E,$98,$F7,$8D
	.dc.b	$35,$FF,$56,$FA,$1D,$D6,$1F,$D1,$B9,$F0,$FF,$48,$F9,$CB,$E5,$1B
	.dc.b	$32,$BC,$B3,$96,$CE,$5F,$26,$9F,$EA,$BE,$07,$CE,$78,$65,$18,$7F
	.dc.b	$D7,$E6,$23,$C7,$F0,$8F,$A0,$E1,$5B,$AE,$A7,$FA,$F5,$C6,$DF,$ED
	.dc.b	$18,$E4,$7C,$AA,$E4,$F6,$9D,$E1,$E5,$C3,$A9,$E0,$C3,$E1,$F8,$6F
	.dc.b	$62,$D4,$C3,$EA,$CE,$5B,$39,$7C,$BB,$7F,$A3,$6B,$9C,$FD,$1F,$11
	.dc.b	$1F,$D2,$B4,$FF,$74,$46,$3A,$75,$39,$3F,$DB,$57,$FE,$97,$F4,$73
	.dc.b	$33,$1F,$F7,$3D,$F1,$7C,$B3,$F7,$3A,$FF,$C3,$38,$72,$5E,$EE,$4D
	.dc.b	$54,$3F,$23,$1A,$6C,$65,$F6,$C7,$47,$F1,$6E,$DC,$C6,$F2,$F7,$C7
	.dc.b	$E4,$3E,$3B,$FC,$2A,$FE,$79,$AF,$C8,$6C,$70,$CF,$FD,$6D,$CF,$F9
	.dc.b	$69,$CE,$3F,$11,$B7,$D7,$93,$EC,$D6,$FD,$E9,$3D,$DE,$5F,$2D,$88
	.dc.b	$F0,$45,$CB,$C1,$82,$BF,$51,$6F,$9B,$FF,$26,$61,$FB,$36,$BF,$94
	.dc.b	$8D,$7A,$FF,$AC,$F1,$9F,$21,$CA,$F7,$97,$F0,$36,$EB,$FC,$09,$B8
	.dc.b	$FE,$DA,$F8,$BD,$F0,$E4,$47,$F8,$3C,$69,$FD,$2F,$13,$E6,$F9,$C7
	.dc.b	$EC,$3E,$57,$FC,$69,$97,$E9,$FA,$91,$8C,$FF,$BC,$DF,$8B,$0E,$59
	.dc.b	$7E,$CA,$B5,$8F,$EC,$5B,$3A,$8F,$D7,$46,$9F,$06,$B8,$3F,$2F,$7E
	.dc.b	$4C,$7C,$31,$99,$FD,$E9,$C0,$F5,$CE,$EB,$A5,$70,$9D,$B7,$2F,$D5
	.dc.b	$D6,$EF,$A4,$7B,$EF,$CF,$39,$BE,$31,$D0,$CE,$3E,$FB,$E3,$97,$F1
	.dc.b	$B9,$CD,$60,$F8,$53,$72,$F8,$51,$D2,$34,$FA,$F1,$F2,$E1,$D6,$34
	.dc.b	$48,$CB,$C3,$2F,$66,$3E,$1B,$F9,$37,$C6,$FD,$91,$E1,$CB,$D9,$18
	.dc.b	$C3,$A5,$FD,$CB,$E0,$F3,$8C,$9F,$E2,$FA,$E3,$19,$AE,$47,$92,$76
	.dc.b	$7A,$E3,$47,$57,$D1,$97,$CB,$C3,$CF,$18,$7C,$D8,$78,$A3,$E4,$9E
	.dc.b	$2C,$BE,$4E,$1E,$0C,$3A,$CE,$A3,$C4,$65,$C4,$6A,$78,$BC,$23,$CC
	.dc.b	$68,$F7,$57,$85,$C9,$A9,$E2,$EC,$CE,$5F,$49,$E2,$7A,$CF,$46,$5F
	.dc.b	$21,$E1,$1B,$9E,$71,$CB,$2F,$88,$FE,$72,$FC,$55,$D6,$3C,$57,$CD
	.dc.b	$E1,$97,$CA,$7E,$8C,$EC,$ED,$3B,$71,$7F,$96,$9F,$C6,$9F,$EA,$98
	.dc.b	$FD,$43,$63,$FD,$5E,$FE,$0C,$FF,$57,$7B,$4F,$85,$7C,$DF,$FA,$C6
	.dc.b	$7F,$DC,$31,$87,$13,$B9,$D4,$7F,$B8,$A7,$4F,$87,$D2,$FE,$8C,$FF
	.dc.b	$81,$5A,$FF,$7B,$6B,$FE,$06,$6E,$1F,$E9,$3E,$7F,$EF,$51,$FD,$57
	.dc.b	$7F,$ED,$3F,$36,$F8,$F7,$99,$AB,$98,$FF,$06,$C3,$87,$17,$FA,$5B
	.dc.b	$0E,$17,$AC,$7F,$92,$56,$AF,$C9,$87,$F6,$EB,$DC,$73,$8E,$18,$7A
	.dc.b	$DD,$63,$E2,$9D,$13,$0E,$B3,$AC,$78,$CF,$63,$57,$E5,$BE,$57,$C2
	.dc.b	$72,$6D,$F8,$31,$A9,$F1,$4E,$8F,$95,$7D,$DC,$65,$F2,$DE,$7C,$4C
	.dc.b	$BD,$B8,$F7,$26,$1D,$1E,$F9,$7A,$2F,$A9,$CA,$B8,$D6,$C9,$1E,$77
	.dc.b	$D9,$3C,$70,$F1,$5F,$6A,$FF,$60,$5F,$27,$81,$0E,$F1,$E2,$7D,$79
	.dc.b	$FD,$77,$38,$76,$73,$97,$CE,$CB,$DD,$9F,$C7,$9E,$99,$78,$AF,$CF
	.dc.b	$5A,$CF,$CD,$8F,$5A,$D5,$F9,$B0,$E4,$F8,$0E,$73,$F8,$9B,$E5,$1E
	.dc.b	$23,$E3,$4F,$0A,$E7,$1C,$4E,$53,$C3,$2F,$89,$19,$9F,$0C,$E7,$1F
	.dc.b	$9B,$39,$C3,$C1,$9F,$C9,$8C,$67,$F0,$63,$18,$78,$09,$5C,$32,$F2
	.dc.b	$3E,$F2,$57,$E2,$72,$F1,$D7,$08,$C6,$3E,$28,$C6,$3F,$E1,$2C,$7B
	.dc.b	$E1,$D7,$3F,$16,$5D,$B1,$E3,$7F,$7B,$3F,$63,$70,$E3,$3A,$8F,$8B
	.dc.b	$1F,$85,$9D,$8F,$B2,$7C,$B7,$DD,$8E,$74,$EF,$CE,$5F,$EA,$EF,$9E
	.dc.b	$7B,$1F,$6A,$73,$5F,$E7,$A9,$CD,$7F,$A5,$A7,$3C,$1F,$E5,$59,$CD
	.dc.b	$7F,$86,$27,$3C,$1F,$E7,$89,$CD,$7F,$A5,$67,$35,$FE,$B2,$9C,$D2
	.dc.b	$24,$EB,$BD,$5A,$9F,$F3,$57,$8B,$F9,$99,$77,$7F,$AB,$B2,$FF,$36
	.dc.b	$7C,$06,$63,$17,$FE,$6D,$BC,$19,$9B,$5F,$E7,$3E,$EE,$1F,$1B,$C5
	.dc.b	$B1,$93,$FC,$F1,$19,$69,$FE,$D8,$FB,$37,$B3,$E5,$6C,$67,$83,$13
	.dc.b	$23,$83,$B8,$C6,$5F,$EA,$AA,$E2,$D9,$D5,$70,$3F,$DC,$9F,$2E,$F9
	.dc.b	$8F,$89,$EE,$3A,$BF,$7D,$E0,$AB,$57,$FE,$EE,$AD,$3C,$AF,$8D,$F0
	.dc.b	$AD,$E3,$FE,$BC,$BD,$BF,$E9,$7F,$03,$E7,$1F,$3F,$B2,$E3,$FE,$9D
	.dc.b	$8B,$3F,$E9,$D9,$BC,$FF,$83,$F9,$7F,$E8,$DE,$B7,$F4,$7E,$67,$9E
	.dc.b	$F9,$5F,$CA,$56,$B3,$FF,$61,$CE,$5F,$CA,$46,$72,$FF,$7E,$CE,$5F
	.dc.b	$CA,$46,$6F,$AD,$F4,$BF,$C8,$61,$E1,$C3,$86,$3E,$59,$C9,$E0,$AF
	.dc.b	$C8,$7C,$B3,$E6,$7A,$2F,$99,$B1,$86,$D7,$2D,$19,$AF,$84,$E6,$31
	.dc.b	$7C,$63,$85,$3C,$9F,$EE,$3A,$7C,$7E,$07,$A7,$F0,$6B,$5C,$1E,$F9
	.dc.b	$CE,$1C,$B0,$E2,$FF,$BB,$23,$34,$C3,$FB,$C3,$CA,$7C,$93,$A9,$CF
	.dc.b	$7B,$E6,$7E,$67,$BA,$F9,$AF,$14,$E7,$0F,$64,$E5,$F5,$C6,$71,$F0
	.dc.b	$47,$FB,$FA,$75,$18,$3F,$81,$DF,$6B,$EA,$DA,$E6,$D9,$E0,$7E,$2B
	.dc.b	$D8,$FC,$07,$A6,$FA,$2F,$8D,$CC,$9C,$C6,$1B,$87,$D0,$9D,$3F,$56
	.dc.b	$B9,$3F,$6F,$C7,$7A,$EF,$77,$BE,$DF,$93,$9C,$C6,$2B,$85,$6D,$E9
	.dc.b	$5B,$8C,$67,$F0,$79,$58,$FF,$39,$BF,$07,$D7,$3E,$17,$F3,$9E,$4B
	.dc.b	$ED,$3C,$13,$9A,$FC,$64,$E5,$E3,$5C,$2B,$6D,$8D,$F4,$74,$3A,$F8
	.dc.b	$36,$32,$4C,$FC,$D8,$7F,$04,$E6,$FD,$D3,$F8,$79,$FE,$79,$C0,$CF
	.dc.b	$F0,$15,$AF,$AA,$F3,$F2,$FA,$9B,$7E,$86,$E3,$D9,$F6,$FB,$DF,$03
	.dc.b	$EA,$6D,$7B,$9B,$5E,$46,$E1,$E4,$6C,$F9,$31,$5F,$CA,$8C,$DB,$2F
	.dc.b	$44,$E3,$87,$FA,$7D,$6A,$7D,$9F,$15,$F4,$BC,$6F,$A2,$D9,$FA,$ED
	.dc.b	$C3,$F3,$95,$C2,$7E,$AD,$7E,$E7,$E6,$9F,$35,$FC,$6F,$D6,$7C,$DB
	.dc.b	$D4,$D4,$E6,$31,$87,$F0,$6B,$E2,$FF,$0F,$BE,$37,$C2,$31,$5C,$32
	.dc.b	$FD,$3D,$F0,$7F,$61,$7C,$2B,$A6,$1D,$F0,$FE,$1D,$F3,$4F,$97,$FE
	.dc.b	$61,$C4,$FA,$57,$C1,$EA,$73,$18,$C7,$F8,$DD,$F1,$7F,$92,$5F,$1B
	.dc.b	$E1,$18,$AE,$19,$FE,$F2,$F8,$3F,$C1,$6F,$85,$74,$C3,$BE,$1C,$AB
	.dc.b	$FD,$71,$7C,$1F,$F0,$71,$FE,$0F,$70,$E1,$AF,$F8,$4F,$EB,$DF,$41
	.dc.b	$FD,$37,$E7,$5F,$6A,$F9,$0F,$E6,$15,$AC,$FF,$9F,$56,$A7,$37,$C6
	.dc.b	$F8,$57,$F3,$CB,$E1,$87,$F1,$FC,$38,$4E,$6F,$F9,$9C,$FA,$2B,$6F
	.dc.b	$A8,$38,$75,$1C,$1F,$F1,$65,$7C,$8D,$3F,$D6,$9C,$0F,$9E,$FF,$18
	.dc.b	$63,$FD,$3A,$B4,$FF,$8C,$E7,$57,$FD,$36,$5F,$2D,$F9,$5F,$B6,$7D
	.dc.b	$36,$63,$0F,$F5,$3A,$FF,$1E,$7A,$BF,$E3,$DF,$65,$F1,$DF,$EE,$B8
	.dc.b	$C3,$65,$3A,$7D,$D3,$A7,$FC,$8D,$F1,$3F,$E5,$8D,$B1,$08,$C3,$CA
	.dc.b	$FE,$FC,$62,$FA,$4E,$FD,$EF,$EC,$0E,$CF,$F2,$97,$CE,$FF,$8A,$DF
	.dc.b	$EE,$3F,$6E,$F9,$8F,$C4,$E5,$FD,$9A,$31,$3F,$DD,$AB,$83,$FD,$49
	.dc.b	$7C,$E7,$0E,$2F,$99,$EE,$7F,$6E,$C3,$E9,$B6,$3F,$2B,$5C,$2F,$FC
	.dc.b	$03,$E5,$BE,$9B,$F0,$11,$F9,$A8,$C3,$83,$FC,$81,$15,$A6,$63,$6F
	.dc.b	$F6,$6F,$87,$FF,$06,$3F,$E0,$E6,$BF,$DA,$A3,$78,$FF,$81,$1B,$87
	.dc.b	$EA,$AB,$73,$FE,$08,$6C,$79,$6F,$E0,$5F,$F9,$F7,$E7,$3D,$8F,$A9
	.dc.b	$AF,$E6,$DA,$FE,$D9,$AF,$1C,$F9,$AF,$9E,$FE,$E2,$FF,$89,$9A,$FF
	.dc.b	$80,$A7,$BE,$3F,$E1,$C6,$E1,$FC,$0A,$B7,$3F,$E2,$06,$C7,$86,$FB
	.dc.b	$D7,$0B,$FF,$4B,$56,$9B,$EA,$6A,$BC,$BF,$3C,$F7,$7C,$FC,$3F,$E1
	.dc.b	$1B,$FF,$1D,$B5,$FF,$0B,$CE,$DF,$F1,$FB,$5F,$F0,$DC,$6F,$3F,$F1
	.dc.b	$AB,$73,$FE,$37,$5B,$9F,$F1,$CB,$63,$F8,$FC,$EE,$BF,$D5,$CD,$53
	.dc.b	$B7,$FB,$07,$9C,$FA,$FF,$B3,$E7,$7F,$AB,$BE,$0E,$FA,$9F,$E7,$5B
	.dc.b	$FF,$09,$D7,$DE,$F3,$FF,$BF,$3A,$A9,$3A,$3F,$C3,$3F,$5B,$BC,$4D
	.dc.b	$3B,$FB,$3F,$D3,$67,$64,$3F,$B4,$CF,$07,$FC,$69,$E5,$F1,$7F,$C3
	.dc.b	$FF,$3D,$E6,$30,$F0,$BC,$3A,$3A,$70,$38,$3F,$C0,$9E,$FF,$BB,$D0
	.dc.b	$95,$FE,$30,$8F,$41,$87,$FC,$43,$18,$7F,$C8,$DE,$D7,$A8,$8F,$F9
	.dc.b	$27,$E1,$AB,$FE,$3D,$D7,$FC,$74,$F0,$34,$7F,$8F,$BD,$DF,$9F,$CF
	.dc.b	$23,$04,$3F,$C6,$2E,$5F,$8B,$1F,$4F,$EE,$1F,$51,$F3,$26,$1F,$E5
	.dc.b	$98,$CB,$63,$2F,$94,$E4,$FE,$C3,$E6,$3F,$8F,$C7,$F9,$57,$E2,$7F
	.dc.b	$A0,$BF,$E9,$96,$BF,$E7,$28,$DE,$FF,$E8,$C6,$EF,$FE,$FF,$E8,$FF
	.dc.b	$51,$79,$5F,$8E,$C3,$8D,$EF,$AF,$F7,$FF,$2A,$F7,$BF,$85,$FA,$0F
	.dc.b	$B9,$FE,$E5,$AF,$F8,$7D,$AF,$FA,$0A,$B8,$3F,$EA,$A6,$BF,$E1,$E8
	.dc.b	$DE,$1F,$D9,$1B,$A7,$F8,$1A,$B7,$3F,$EA,$46,$C7,$F6,$59,$DD,$2F
	.dc.b	$C9,$EB,$7A,$31,$FE,$2F,$F3,$9A,$D9,$F4,$0E,$6F,$D6,$B7,$1F,$F5
	.dc.b	$FD,$74,$4F,$18,$D4,$7F,$A1,$5B,$AF,$FA,$E2,$7B,$7D,$13,$E4,$FF
	.dc.b	$72,$BC,$AA,$62,$26,$1F,$ED,$19,$D3,$67,$2F,$CB,$EE,$FE,$07,$F1
	.dc.b	$3F,$CE,$BD,$4F,$F2,$A4,$FF,$BC,$3A,$FF,$C5,$B1,$88,$72,$2F,$ED
	.dc.b	$9C,$AB,$7E,$FF,$FB,$F1,$B4,$38,$CF,$07,$FD,$CD,$87,$86,$B8,$47
	.dc.b	$F8,$39,$FB,$13,$FC,$2B,$E1,$B9,$F0,$FF,$AA,$9B,$DD,$FE,$D9,$BE
	.dc.b	$07,$FB,$D1,$FF,$0E,$B8,$9E,$0D,$8E,$0B,$3F,$83,$3A,$8C,$61,$F9
	.dc.b	$DA,$E7,$18,$3F,$C3,$BF,$43,$FE,$C7,$7B,$71,$7F,$02,$75,$7F,$86
	.dc.b	$F0,$3E,$1F,$AA,$BE,$15,$CE,$31,$F1,$5E,$EC,$39,$67,$E4,$6E,$7F
	.dc.b	$E3,$1C,$38,$CF,$A3,$CD,$7C,$8F,$C6,$6E,$5F,$05,$AF,$91,$7C,$86
	.dc.b	$E6,$4E,$DE,$51,$88,$E1,$16,$BC,$13,$F4,$BE,$37,$E9,$FC,$54,$BF
	.dc.b	$9D,$19,$7E,$84,$65,$F8,$BC,$1F,$EF,$97,$EA,$6B,$F7,$08,$FD,$7F
	.dc.b	$8A,$AF,$08,$B8,$FD,$8A,$57,$E4,$2B,$F2,$FC,$FF,$88,$F9,$7F,$F2
	.dc.b	$3D,$FD,$E6,$EF,$FC,$16,$78,$74,$7E,$A2,$FE,$CD,$69,$72,$23,$11
	.dc.b	$D8,$FC,$89,$4E,$3E,$F7,$84,$EB,$7F,$A4,$F4,$9F,$3B,$FE,$36,$CF
	.dc.b	$F6,$31,$8F,$92,$D4,$E7,$4F,$F7,$8F,$DD,$3C,$EC,$D7,$A3,$97,$FC
	.dc.b	$DF,$A7,$E5,$4F,$DB,$B8,$CB,$AE,$3C,$E7,$FD,$C8,$D7,$7D,$3E,$AE
	.dc.b	$0F,$C5,$BF,$AC,$D3,$C9,$36,$B8,$56,$A3,$C0,$DB,$EF,$8F,$97,$0E
	.dc.b	$98,$F8,$F4,$FE,$68,$DC,$3C,$75,$B7,$9C,$65,$F2,$9F,$B1,$CB,$C6
	.dc.b	$EA,$FA,$91,$D6,$5E,$6A,$E5,$97,$97,$1F,$0D,$FA,$F1,$F0,$65,$FC
	.dc.b	$7E,$B5,$8F,$F1,$96,$E5,$E0,$9D,$63,$D0,$98,$73,$BF,$14,$EA,$2E
	.dc.b	$3F,$C8,$EF,$93,$E2,$AF,$51,$2B,$E4,$BC,$0F,$15,$F6,$7D,$B8,$FA
	.dc.b	$2B,$95,$77,$9C,$9C,$4F,$92,$E4,$EA,$F7,$8E,$24,$7E,$99,$C7,$1F
	.dc.b	$86,$F2,$8F,$56,$5F,$37,$2D,$63,$F6,$F2,$F6,$DF,$A2,$72,$7C,$57
	.dc.b	$C0,$64,$B1,$C1,$EC,$67,$1F,$46,$1E,$0C,$BE,$04,$E6,$3D,$15,$C7
	.dc.b	$2F,$39,$31,$F2,$65,$E7,$BF,$09,$DA,$B8,$4F,$CD,$C7,$BC,$FC,$23
	.dc.b	$77,$CF,$2F,$54,$EA,$7C,$E7,$3C,$3B,$46,$6B,$F1,$31,$8C,$7E,$75
	.dc.b	$6F,$2F,$5B,$CA,$B7,$18,$7C,$11,$F7,$0F,$1E,$3E,$7A,$F8,$91,$FA
	.dc.b	$13,$FB,$81,$F3,$4A,$7E,$A8,$B3,$FB,$97,$3F,$02,$F9,$D7,$CF,$7F
	.dc.b	$B4,$E3,$0D,$8C,$CF,$FB,$6A,$72,$FF,$B7,$29,$EF,$FF,$8A,$E3,$0D
	.dc.b	$7C,$2D,$7F,$C6,$CD,$7F,$60,$F1,$D3,$F6,$8D,$C7,$F5,$5F,$3D,$FD
	.dc.b	$6B,$CD,$FE,$31,$39,$BE,$3E,$66,$0F,$E3,$55,$B8,$FE,$2F,$3B,$7F
	.dc.b	$C3,$AD,$C7,$D3,$5C,$3B,$7F,$D1,$95,$C2,$3A,$34,$FE,$29,$7D,$72
	.dc.b	$F3,$CE,$9B,$3A,$AE,$71,$82,$65,$F0,$AB,$C1,$97,$18,$C5,$F7,$AF
	.dc.b	$E8,$F8,$77,$9C,$C7,$29,$F8,$27,$7A,$E4,$E2,$35,$C4,$BE,$53,$9B
	.dc.b	$E6,$45,$3B,$C7,$D3,$39,$9F,$93,$8F,$6A,$D6,$3E,$09,$C9,$E9,$AF
	.dc.b	$B3,$7C,$A3,$C0,$43,$78,$FB,$A3,$51,$89,$F0,$65,$F5,$A3,$DC,$75
	.dc.b	$39,$39,$AE,$57,$E3,$8C,$57,$B7,$0F,$3D,$F1,$3B,$1E,$47,$37,$F3
	.dc.b	$63,$26,$A3,$33,$E4,$CF,$F1,$47,$E4,$E3,$33,$E4,$8C,$E5,$F7,$A3
	.dc.b	$C1,$87,$91,$F0,$67,$F5,$1E,$F1,$97,$C1,$19,$8C,$12,$BC,$11,$97
	.dc.b	$BC,$FC,$5B,$F1,$BD,$23,$38,$FC,$68,$F1,$C6,$31,$FA,$F9,$7C,$98
	.dc.b	$C4,$FE,$13,$0E,$D9,$64,$F6,$61,$F8,$E8,$CC,$FD,$68,$C6,$7F,$73
	.dc.b	$3F,$81,$9F,$91,$B7,$C2,$BC,$87,$0C,$BA,$E7,$FD,$AF,$E7,$9F,$3D
	.dc.b	$F8,$BC,$7D,$BC,$B8,$BF,$F4,$17,$4B,$E1,$BF,$17,$39,$DB,$FC,$7F
	.dc.b	$18,$C3,$F2,$9D,$7F,$E3,$F8,$C6,$1C,$9B,$5C,$23,$FC,$83,$7C,$F4
	.dc.b	$FF,$27,$57,$2F,$7B,$FD,$71,$5F,$E9,$36,$DF,$FA,$19,$7C,$F6,$FB
	.dc.b	$25,$6D,$E5,$18,$AE,$1F,$3D,$D7,$CD,$EC,$DB,$EC,$D9,$DC,$78,$67
	.dc.b	$DB,$F3,$CF,$9E,$A6,$33,$B5,$FE,$A9,$9D,$36,$74,$D2,$5F,$FA,$82
	.dc.b	$76,$DA,$E6,$FF,$A7,$FE,$79,$F3,$1F,$3B,$C9,$FE,$8F,$87,$FA,$EB
	.dc.b	$C2,$56,$A3,$FD,$35,$D9,$FC,$1E,$73,$7C,$63,$15,$BC,$3F,$D3,$D1
	.dc.b	$F4,$7C,$09,$87,$D1,$EC,$B5,$F4,$66,$F1,$FF,$B0,$B8,$FF,$91,$F2
	.dc.b	$9F,$F5,$6E,$1F,$7B,$D4,$FE,$87,$EA,$7F,$44,$A7,$A2,$F8,$3F,$C6
	.dc.b	$D6,$AF,$FD,$C5,$5A,$9C,$C7,$FB,$89,$B9,$B0,$E9,$87,$3A,$F0,$1B
	.dc.b	$2C,$ED,$EB,$1B,$8F,$F6,$8F,$B4,$F9,$EF,$14,$7F,$A6,$A1,$F1,$98
	.dc.b	$C9,$1B,$C3,$FD,$F1,$8F,$B6,$37,$18,$3D,$6E,$EF,$C4,$75,$BE,$91
	.dc.b	$9A,$E5,$1E,$29,$79,$1F,$DC,$E7,$E5,$72,$BE,$6F,$53,$9C,$7E,$4C
	.dc.b	$E5,$F1,$58,$E7,$C5,$EB,$9D,$47,$90,$CC,$F0,$CD,$3A,$C7,$FB,$ED
	.dc.b	$7F,$AB,$7D,$5F,$89,$F3,$CF,$9E,$F4,$78,$29,$CC,$61,$B9,$79,$2F
	.dc.b	$93,$DE,$F9,$39,$BE,$4E,$6F,$93,$4E,$5C,$2F,$53,$EB,$7A,$4F,$85
	.dc.b	$F9,$53,$9A,$F6,$4E,$5E,$35,$C2,$B6,$D8,$F9,$3E,$3B,$E2,$DE,$AF
	.dc.b	$6C,$FF,$0F,$39,$8C,$57,$0A,$DB,$E0,$AD,$C6,$35,$FC,$1F,$0B,$3F
	.dc.b	$F6,$C5,$F8,$3E,$25,$ED,$7E,$47,$B1,$F3,$DE,$09,$CD,$7D,$79,$CB
	.dc.b	$C6,$B8,$56,$DB,$1B,$E8,$E8,$75,$F0,$2C,$64,$99,$F9,$B0,$FE,$22
	.dc.b	$DA,$F1,$72,$F9,$7E,$7A,$DF,$8D,$B5,$C5,$BC,$FF,$C0,$F0,$47,$AF
	.dc.b	$DA,$F7,$7C,$0D,$C3,$D8,$DC,$3C,$4D,$C3,$93,$67,$FD,$91,$B2,$FE
	.dc.b	$44,$66,$FF,$C6,$B3,$98,$F9,$D3,$A6,$9F,$47,$DB,$FA,$1E,$37,$CF
	.dc.b	$6C,$64,$FA,$11,$96,$9F,$9D,$F9,$D7,$C5,$7F,$0C,$F5,$DF,$3C,$F5
	.dc.b	$BF,$7D,$5A,$BF,$DB,$D6,$A7,$37,$C6,$F8,$57,$58,$C4,$EA,$38,$CE
	.dc.b	$B4,$FD,$E6,$1D,$5F,$E2,$38,$75,$AF,$D9,$7C,$E3,$E3,$BF,$90,$F0
	.dc.b	$3E,$79,$F3,$DF,$C4,$EB,$57,$FC,$32,$B5,$39,$BE,$37,$C2,$BA,$C6
	.dc.b	$27,$51,$FE,$AE,$C7,$F8,$A6,$1D,$1C,$35,$FF,$59,$B5,$FF,$5B,$BB
	.dc.b	$C3,$F8,$9B,$67,$F8,$F7,$CD,$3E,$A7,$F7,$11,$E2,$3F,$88,$7C,$F3
	.dc.b	$E7,$BF,$9F,$D3,$E7,$DF,$1D,$FC,$C6,$B5,$7F,$CA,$AB,$53,$9B,$E3
	.dc.b	$7C,$2B,$AC,$62,$75,$18,$3F,$99,$FD,$E3,$EF,$1E,$E6,$B6,$FE,$65
	.dc.b	$97,$AD,$71,$3E,$87,$F9,$D6,$3F,$CA,$6B,$53,$9A,$FE,$91,$0F,$A0
	.dc.b	$FC,$87,$D0,$BE,$5B,$31,$87,$FA,$75,$7F,$45,$F6,$BF,$A3,$7B,$AF
	.dc.b	$8E,$FA,$D1,$86,$CA,$74,$FF,$BA,$27,$4F,$F4,$FF,$8A,$FF,$25,$73
	.dc.b	$E2,$11,$87,$95,$FF,$BC,$A3,$17,$D2,$77,$F0,$1F,$DF,$DD,$9F,$D6
	.dc.b	$FE,$83,$FC,$4A,$FF,$4B,$FA,$17,$CC,$7F,$BF,$32,$FE,$AF,$18,$9F
	.dc.b	$EA,$15,$C1,$FE,$C0,$BE,$57,$87,$17,$F6,$7E,$2F,$EF,$75,$FB,$4F
	.dc.b	$7B,$FB,$0E,$3F,$D8,$B0,$E1,$3F,$D6,$FC,$E7,$D0,$F2,$8F,$EE,$11
	.dc.b	$87,$51,$95,$F1,$5F,$DA,$5C,$9F,$D9,$5F,$F0,$0B,$7E,$03,$85,$73
	.dc.b	$8F,$EF,$0D,$C3,$FB,$AD,$70,$9F,$EF,$8D,$8F,$EF,$33,$BC,$FF,$C0
	.dc.b	$94,$FA,$1F,$5B,$FE,$03,$6D,$7F,$80,$1B,$3C,$3E,$87,$F0,$55,$FD
	.dc.b	$EE,$78,$BD,$0E,$9B,$7E,$FD,$FF,$09,$46,$9F,$0B,$5C,$C7,$C5,$9F
	.dc.b	$EE,$11,$FE,$07,$6B,$E5,$8F,$95,$3E,$48,$FF,$0F,$4E,$AF,$FC,$2E
	.dc.b	$DB,$EF,$7F,$47,$E8,$7F,$C3,$B7,$FE,$13,$E1,$FF,$04,$37,$DE,$D7
	.dc.b	$C1,$48,$C1,$0F,$F1,$45,$72,$6C,$7D,$DA,$E7,$1F,$E2,$D6,$E5,$FE
	.dc.b	$2A,$AE,$0B,$1F,$EB,$3F,$43,$F8,$9F,$6F,$FC,$D9,$F0,$F7,$D7,$FF
	.dc.b	$34,$65,$FE,$05,$AF,$C7,$7C,$03,$D6,$46,$EE,$56,$CF,$F0,$67,$D0
	.dc.b	$F0,$A3,$4F,$07,$BD,$FD,$0E,$76,$43,$FC,$B3,$87,$79,$DF,$A1,$C6
	.dc.b	$7F,$C2,$5F,$37,$E7,$30,$F0,$C3,$BB,$A2,$19,$AF,$27,$17,$BF,$D6
	.dc.b	$93,$FE,$23,$8C,$BF,$E1,$F8,$CB,$D6,$3D,$3E,$F4,$8F,$F1,$FB,$EC
	.dc.b	$F8,$AC,$47,$F8,$C3,$4F,$F1,$6B,$A3,$FC,$8B,$1F,$17,$E0,$BE,$87
	.dc.b	$3C,$9F,$F1,$A4,$65,$FF,$18,$46,$5F,$74,$7D,$4F,$A1,$7D,$0F,$CC
	.dc.b	$98,$7F,$93,$23,$2D,$8C,$BE,$53,$93,$FC,$1F,$E8,$7F,$8D,$47,$F9
	.dc.b	$23,$E3,$3F,$9E,$3F,$E7,$76,$BF,$E7,$B8,$DF,$47,$F9,$E1,$BD,$1F
	.dc.b	$E9,$4F,$2F,$F1,$D7,$C2,$B5,$D7,$FC,$57,$E6,$1F,$05,$FC,$6F,$C7
	.dc.b	$7D,$0F,$F6,$6D,$3F,$C2,$ED,$7F,$D1,$55,$C1,$FF,$3F,$35,$FF,$08
	.dc.b	$C6,$F0,$FD,$04,$7F,$5E,$8C,$67,$FC,$52,$B8,$57,$FA,$41,$B1,$FC
	.dc.b	$D6,$77,$5C,$7D,$6F,$46,$3F,$CB,$7E,$87,$3E,$03,$E8,$1C,$BF,$AD
	.dc.b	$6E,$1F,$E9,$3A,$E6,$9E,$31,$A8,$FF,$3B,$B7,$4F,$F5,$1C,$F6,$FA
	.dc.b	$1F,$EF,$75,$FE,$A9,$7F,$9E,$B7,$C8,$FD,$E4,$6A,$3F,$D0,$2D,$AF
	.dc.b	$F7,$33,$C2,$33,$3C,$3D,$7F,$A9,$DF,$FC,$17,$E3,$BF,$9B,$FA,$9F
	.dc.b	$E4,$BB,$6B,$FE,$EE,$EB,$FE,$7B,$39,$97,$1B,$0E,$71,$FE,$BF,$C3
	.dc.b	$9F,$C2,$7F,$B9,$B0,$E2,$FD,$3A,$DD,$7F,$A6,$2B,$6F,$81,$AF,$68
	.dc.b	$FE,$17,$F1,$3C,$A3,$CB,$DD,$FE,$C5,$7F,$C2,$17,$FF,$00,$23,$10
	.dc.b	$CB,$D9,$9F,$FA,$97,$D7,$E4,$FF,$A9,$F0,$B7,$C2,$B6,$DC,$3A,$5F
	.dc.b	$87,$D5,$7D,$0F,$FA,$EB,$2F,$09,$F2,$39,$3F,$81,$BE,$1B,$FF,$74
	.dc.b	$67,$F8,$9F,$0B,$95,$F1,$AD,$1F,$E2,$37,$19,$FA,$30,$E6,$DA,$E3
	.dc.b	$F1,$DE,$AF,$39,$F2,$94,$98,$F9,$32,$FF,$78,$FC,$9E,$23,$BC,$7F
	.dc.b	$BD,$27,$18,$FC,$6B,$5E,$34,$F4,$C7,$6C,$3A,$35,$E6,$F0,$7C,$1B
	.dc.b	$FF,$BA,$DB,$F1,$9F,$38,$F9,$FF,$23,$23,$24,$3C,$B3,$C5,$BA,$7D
	.dc.b	$62,$3D,$BE,$36,$BA,$3C,$31,$F7,$23,$F1,$11,$F9,$8E,$9F,$C0,$FC
	.dc.b	$DF,$F9,$17,$5F,$C2,$B7,$4F,$E1,$13,$C3,$A7,$F5,$55,$AF,$7F,$89
	.dc.b	$99,$98,$7C,$FC,$39,$3F,$AA,$6B,$E1,$3F,$0F,$E8,$BE,$83,$FC,$5D
	.dc.b	$87,$EC,$63,$1E,$D7,$E7,$DB,$5C,$3E,$85,$E7,$E7,$0E,$3A,$FF,$1F
	.dc.b	$E5,$FF,$15,$E5,$FB,$7D,$BF,$0B,$D5,$E0,$BD,$E4,$CF,$AB,$F9,$23
	.dc.b	$F7,$B8,$72,$C7,$A5,$F1,$D3,$C0,$7C,$59,$D5,$0E,$33,$B7,$F4,$A6
	.dc.b	$0E,$D8,$F9,$1B,$3C,$B0,$F1,$E5,$E9,$8E,$B3,$C6,$31,$7D,$63,$8B
	.dc.b	$DA,$FA,$4E,$A7,$F5,$B7,$D3,$1F,$5C,$70,$24,$FD,$1C,$BE,$2C,$77
	.dc.b	$9D,$CF,$33,$59,$7C,$1C,$7C,$19,$79,$B2,$F0,$DF,$C3,$CB,$C1,$8F
	.dc.b	$C5,$9C,$B6,$B8,$1F,$07,$2F,$0C,$EA,$7C,$46,$F0,$F8,$19,$78,$F1
	.dc.b	$F2,$92,$7E,$E5,$F2,$C3,$AD,$F9,$8E,$35,$BA,$EA,$76,$C7,$CE,$E3
	.dc.b	$1F,$11,$E8,$8C,$B8,$BE,$B3,$F3,$B0,$F0,$61,$DE,$3E,$6E,$1F,$0E
	.dc.b	$73,$18,$3D,$87,$1C,$BE,$63,$93,$C9,$1C,$32,$F8,$11,$C1,$FA,$6F
	.dc.b	$77,$2F,$09,$E7,$18,$C7,$D5,$5C,$EF,$E1,$46,$2B,$71,$8C,$7E,$A4
	.dc.b	$63,$0F,$04,$FC,$FC,$BE,$1D,$7C,$2C,$7C,$B1,$8C,$BC,$95,$BC,$BC
	.dc.b	$18,$72,$25,$6D,$F8,$19,$78,$DE,$26,$CF,$61,$83,$EB,$63,$E6,$7E
	.dc.b	$81,$FE,$A9,$8F,$D7,$36,$3F,$D5,$ED,$C3,$FD,$59,$F0,$2F,$A0,$7D
	.dc.b	$0E,$1B,$5F,$EB,$33,$FD,$E2,$E1,$FF,$5B,$4F,$19,$FF,$5B,$60,$F7
	.dc.b	$9E,$6E,$1B,$7C,$E7,$53,$92,$47,$F0,$69,$DE,$DF,$C9,$7E,$87,$FA
	.dc.b	$AF,$9D,$FB,$F9,$CF,$07,$F8,$FD,$B7,$FA,$F6,$EE,$BE,$33,$D7,$1E
	.dc.b	$44,$7F,$83,$C6,$B1,$FE,$55,$7D,$27,$FC,$42,$DC,$BD,$31,$C3,$B7
	.dc.b	$FC,$A7,$5C,$30,$E8,$D9,$D6,$78,$CF,$D1,$56,$7F,$9A,$65,$E1,$9C
	.dc.b	$C7,$9C,$97,$F0,$F0,$EF,$3A,$8E,$C4,$7F,$89,$57,$3A,$F8,$97,$DD
	.dc.b	$F5,$BF,$D5,$23,$51,$83,$BC,$F5,$39,$DF,$82,$39,$9A,$2D,$F9,$70
	.dc.b	$ED,$8F,$C6,$8F,$24,$FB,$AB,$93,$DE,$32,$DB,$EF,$7C,$B2,$FA,$07
	.dc.b	$DC,$C7,$CF,$8F,$5A,$F3,$5F,$89,$E2,$F9,$4E,$86,$4F,$15,$F5,$CB
	.dc.b	$E9,$67,$F8,$7C,$FC,$B8,$75,$BE,$07,$29,$D5,$F1,$8F,$85,$18,$7C
	.dc.b	$13,$C5,$E1,$8F,$A3,$2F,$5E,$3E,$08,$C3,$CF,$3F,$95,$5C,$AB,$6F
	.dc.b	$E2,$8C,$4F,$1C,$BC,$F3,$AA,$F9,$F8,$F8,$DF,$9B,$8F,$8C,$F6,$46
	.dc.b	$73,$F9,$38,$FC,$63,$CF,$87,$4C,$7C,$93,$F3,$30,$EB,$97,$A2,$FC
	.dc.b	$91,$8B,$E1,$18,$AF,$AB,$87,$C5,$CF,$FC,$11,$3F,$48,$FF,$5D,$95
	.dc.b	$CC,$6E,$3C,$13,$1F,$F0,$D7,$D0,$BE,$87,$F2,$95,$EB,$9F,$F0,$A7
	.dc.b	$2E,$70,$FF,$1B,$7C,$37,$E4,$23,$13,$FE,$64,$E8,$4E,$70,$FF,$66
	.dc.b	$DF,$9A,$BF,$C6,$31,$87,$D3,$7C,$A7,$4D,$8C,$CF,$39,$FF,$3A,$79
	.dc.b	$0F,$A1,$C5,$7F,$9D,$3E,$01,$D9,$FA,$DF,$11,$18,$C3,$FC,$F8,$F1
	.dc.b	$6C,$EE,$35,$5C,$E3,$17,$97,$5F,$43,$BF,$36,$E1,$D1,$B7,$E3,$6F
	.dc.b	$1F,$FA,$AB,$14,$79,$7E,$81,$F4,$34,$C7,$05,$C7,$FD,$25,$3A,$6C
	.dc.b	$E9,$A4,$C7,$FD,$2B,$3B,$6D,$73,$7F,$D4,$FD,$AF,$A1,$7C,$D7,$C7
	.dc.b	$F2,$BF,$A8,$63,$FE,$B4,$AD,$4E,$7B,$7F,$8C,$46,$27,$57,$C9,$FF
	.dc.b	$54,$79,$8F,$3B,$73,$9C,$BE,$B7,$87,$F3,$CF,$A1,$DB,$70,$FA,$B1
	.dc.b	$67,$EE,$4D,$E2,$FF,$75,$74,$BD,$56,$1F,$EE,$9F,$84,$FC,$A4,$63
	.dc.b	$0F,$F7,$34,$62,$7F,$B9,$47,$FA,$0A,$7F,$4E,$FE,$49,$AF,$E8,$1A
	.dc.b	$EE,$3F,$CE,$DE,$AF,$EB,$AF,$C5,$18,$3F,$DA,$FF,$40,$F8,$CF,$15
	.dc.b	$6F,$0B,$5F,$ED,$CF,$75,$F4,$38,$C9,$DC,$9E,$55,$D6,$73,$1C,$23
	.dc.b	$13,$BC,$79,$3C,$6E,$D7,$2B,$63,$F1,$6B,$58,$72,$AD,$4F,$F0,$AF
	.dc.b	$33,$FB,$FB,$BE,$77,$CF,$79,$F1,$56,$B0,$FF,$6F,$CE,$63,$13,$FE
	.dc.b	$9D,$6E,$2E,$57,$AF,$BE,$3E,$9C,$38,$CB,$DB,$C2,$FA,$1C,$DF,$BA
	.dc.b	$31,$3F,$17,$3F,$F5,$EC,$FD,$8B,$E5,$1F,$E3,$16,$BF,$C3,$A3,$47
	.dc.b	$49,$EF,$C1,$67,$E3,$C7,$F2,$EF,$0D,$EA,$FC,$EF,$55,$F0,$BF,$0E
	.dc.b	$73,$5F,$52,$72,$F9,$63,$37,$D0,$FF,$60,$38,$E3,$FA,$9C,$9F,$E4
	.dc.b	$BA,$E1,$8F,$E1,$7C,$97,$D0,$BD,$7F,$16,$F9,$8C,$5F,$86,$31,$32
	.dc.b	$3E,$F7,$89,$74,$F1,$60,$F8,$97,$B5,$FA,$28,$C3,$63,$37,$FE,$5B
	.dc.b	$9C,$B6,$33,$5D,$1B,$AF,$E3,$FC,$8F,$C4,$D7,$E6,$EF,$8D,$FC,$3E
	.dc.b	$8F,$E6,$FD,$5F,$9A,$BD,$CE,$5F,$E2,$AD,$8D,$12,$FF,$47,$3A,$C3
	.dc.b	$C1,$3F,$27,$1F,$C0,$DF,$FB,$35,$AF,$FB,$ED,$B8,$FC,$3E,$5F,$80
	.dc.b	$71,$3C,$9F,$77,$72,$F5,$53,$DB,$F8,$C4,$8F,$88,$DC,$3E,$23,$ED
	.dc.b	$8F,$23,$1E,$6E,$5A,$48,$E2,$FB,$8E,$0F,$CF,$F6,$FE,$B1,$23,$EA
	.dc.b	$36,$35,$1F,$95,$F3,$FE,$24,$FE,$EF,$E7,$DF,$15,$FC,$4F,$D6,$7D
	.dc.b	$03,$D6,$D4,$E6,$31,$8F,$EC,$6F,$8B,$FC,$B9,$B9,$7F,$A8,$5A,$FF
	.dc.b	$80,$DA,$FF,$86,$9D,$F0,$FF,$86,$9B,$29,$7C,$B3,$E7,$5F,$51,$E4
	.dc.b	$AF,$E0,$5F,$3E,$F9,$1F,$E5,$99,$3E,$85,$F4,$3F,$C6,$6B,$57,$FC
	.dc.b	$B6,$B5,$39,$BF,$E8,$35,$AC,$FF,$D7,$4D,$7F,$C3,$8D,$7F,$C6,$6E
	.dc.b	$EB,$FC,$66,$DB,$F5,$67,$E9,$EF,$ED,$1E,$8C,$2E,$1F,$C1,$6B,$53
	.dc.b	$93,$F8,$F7,$D0,$BE,$83,$FA,$44,$BE,$85,$F1,$FF,$CD,$6B,$57,$FD
	.dc.b	$26,$B5,$39,$BF,$EB,$B5,$AC,$3F,$DA,$CD,$7F,$C7,$0D,$7F,$C9,$EE
	.dc.b	$F4,$FF,$27,$B7,$0F,$AD,$F4,$2F,$A1,$FF,$0F,$C7,$A3,$0F,$E9,$19
	.dc.b	$7C,$7C,$7E,$1E,$9F,$DB,$A3,$11,$F7,$A3,$1E,$C4,$BF,$EF,$35,$FE
	.dc.b	$0D,$6C,$7F,$76,$AF,$C3,$7A,$DF,$87,$F6,$5F,$17,$FE,$F1,$8C,$36
	.dc.b	$53,$A7,$FC,$AD,$3A,$7F,$21,$F1,$1F,$E5,$CD,$F1,$08,$C3,$CA,$FF
	.dc.b	$CE,$71,$8B,$E9,$3B,$F7,$BF,$B8,$BB,$3F,$39,$F3,$DF,$E3,$27,$FC
	.dc.b	$2A,$D7,$FC,$D0,$1C,$3F,$E6,$B6,$BF,$E8,$25,$F1,$FF,$E1,$E8,$E0
	.dc.b	$D8,$CC,$76,$8D,$7B,$EF,$9E,$FF,$45,$E7,$FE,$8A,$8C,$57,$F6,$D6
	.dc.b	$BF,$D6,$4F,$F3,$7F,$AF,$C3,$7F,$F4,$9C,$EA,$7F,$D0,$0D,$73,$7E
	.dc.b	$02,$3F,$DE,$7E,$17,$FC,$16,$DF,$6F,$A3,$FC,$06,$BC,$38,$FF,$A7
	.dc.b	$BE,$7B,$D7,$3E,$36,$C6,$73,$E3,$2F,$9E,$FB,$EF,$F8,$1E,$B7,$9F
	.dc.b	$FA,$A9,$B1,$FC,$3D,$AF,$1D,$F9,$C7,$82,$3F,$D3,$8D,$79,$5F,$D1
	.dc.b	$8C,$47,$38,$F1,$DF,$F7,$A8,$FF,$0C,$35,$F5,$47,$A3,$E7,$9E,$FD
	.dc.b	$24,$FF,$87,$FE,$61,$EB,$E2,$3F,$0F,$5C,$1F,$F2,$FB,$5C,$D7,$E1
	.dc.b	$63,$1A,$7F,$B2,$2F,$FC,$37,$5D,$7D,$CE,$93,$FE,$6B,$F4,$BF,$C0
	.dc.b	$DF,$3D,$F5,$1B,$5F,$D8,$A3,$EB,$FA,$CF,$75,$F5,$9B,$F5,$15,$FE
	.dc.b	$72,$C3,$8E,$1F,$EC,$56,$BF,$ED,$89,$DB,$FE,$0D,$F9,$FE,$44,$7F
	.dc.b	$C5,$DE,$0D,$7F,$C2,$11,$73,$FF,$2F,$65,$FD,$0E,$7F,$BB,$CE,$79
	.dc.b	$7F,$C0,$4F,$A2,$75,$3C,$8F,$01,$0F,$36,$09,$F5,$3F,$22,$3F,$C4
	.dc.b	$1A,$FB,$CE,$9C,$16,$3F,$C4,$D1,$C0,$8F,$EF,$B6,$F9,$1E,$D7,$8C
	.dc.b	$91,$FE,$4B,$F8,$5A,$FF,$89,$F2,$FF,$1B,$BF,$E4,$F9,$F5,$4F,$CF
	.dc.b	$E3,$93,$FE,$16,$78,$EB,$F6,$F3,$FC,$4C,$FF,$8E,$67,$D6,$47,$FC
	.dc.b	$7D,$F3,$EF,$9F,$FC,$03,$63,$FC,$74,$D7,$FC,$75,$F3,$4F,$9C,$DD
	.dc.b	$7C,$D9,$87,$07,$F6,$BF,$3E,$E7,$A4,$9F,$F4,$ED,$BE,$2F,$11,$FE
	.dc.b	$47,$F8,$57,$CF,$FE,$41,$B1,$FE,$83,$C3,$A3,$FE,$85,$F9,$7F,$23
	.dc.b	$73,$FE,$0B,$18,$8F,$F2,$97,$C5,$7F,$BD,$1D,$1F,$EC,$88,$D3,$FE
	.dc.b	$13,$3F,$C2,$8E,$1C,$35,$CC,$FF,$1A,$F7,$9F,$3F,$FD,$DA,$7F,$D7
	.dc.b	$91,$8F,$53,$C7,$5F,$C6,$3E,$7F,$55,$FC,$BD,$BC,$9F,$ED,$D8,$46
	.dc.b	$3B,$3F,$C4,$DD,$3F,$EA,$DE,$5E,$58,$FE,$D3,$E3,$BF,$DD,$18,$7F
	.dc.b	$B2,$1A,$F5,$8F,$DC,$F1,$BE,$52,$F9,$52,$BE,$9C,$68,$C3,$F8,$08
	.dc.b	$C3,$F3,$FE,$7F,$E6,$91,$FF,$5D,$FD,$4D,$7F,$AE,$F1,$27,$FA,$C3
	.dc.b	$63,$E8,$46,$0F,$E1,$BF,$1F,$FD,$03,$D4,$FF,$2A,$4F,$FB,$A2,$31
	.dc.b	$F0,$BF,$4A,$BF,$79,$F1,$5F,$5E,$F7,$C1,$FC,$56,$31,$5F,$23,$C2
	.dc.b	$FF,$69,$E3,$FE,$0A,$8C,$7A,$FA,$8F,$F7,$AC,$E9,$61,$D2,$30,$DB
	.dc.b	$E1,$87,$F2,$3B,$E5,$E7,$BE,$83,$FD,$75,$8F,$73,$FC,$26,$E1,$EB
	.dc.b	$97,$F8,$6E,$31,$8F,$AB,$8D,$F5,$3D,$27,$35,$F8,$29,$FE,$0B,$7C
	.dc.b	$86,$5F,$EF,$1F,$1B,$87,$C8,$79,$BC,$97,$CA,$D2,$5F,$96,$FF,$A9
	.dc.b	$FC,$A6,$23,$A4,$7F,$56,$9C,$6B,$F3,$AD,$78,$CF,$F2,$B1,$DB,$0E
	.dc.b	$8D,$79,$BC,$1F,$07,$0F,$FB,$C1,$BF,$1D,$F3,$23,$FD,$61,$1A,$F9
	.dc.b	$3C,$19,$99,$19,$39,$E9,$C4,$8F,$6F,$8E,$AF,$B2,$FE,$5B,$F8,$09
	.dc.b	$F0,$4F,$E0,$E1,$CE,$E9,$C3,$63,$EE,$1A,$8F,$C1,$F4,$3E,$47,$71
	.dc.b	$FC,$2F,$BF,$F8,$65,$6C,$FF,$50,$38,$BE,$91,$FC,$04,$D1,$5C,$F0
	.dc.b	$7E,$7B,$E6,$FF,$CA,$2D,$F8,$1E,$F5,$F8,$FF,$51,$F4,$1F,$E3,$08
	.dc.b	$FC,$34,$63,$DD,$FD,$03,$63,$87,$D0,$FF,$48,$77,$5C,$73,$FE,$49
	.dc.b	$5B,$9F,$E2,$CD,$8C,$CE,$E7,$FD,$79,$3B,$C7,$F7,$71,$F3,$39,$FF
	.dc.b	$C6,$F1,$8D,$3F,$2B,$DB,$E9,$C3,$D1,$3F,$C0,$B0,$E5,$CF,$FC,$CB
	.dc.b	$0E,$2D,$BE,$1A,$7D,$69,$CE,$5F,$CE,$9A,$F8,$67,$6F,$DA,$CB,$C2
	.dc.b	$FE,$AC,$D1,$5C,$E3,$E3,$CB,$CB,$8F,$86,$B5,$87,$6B,$DC,$FC,$2A
	.dc.b	$DD,$78,$F0,$F0,$5E,$CF,$3E,$5E,$E9,$D1,$D2,$76,$7A,$1C,$13,$1F
	.dc.b	$95,$5C,$AF,$BC,$FA,$A7,$35,$CA,$3E,$EE,$5F,$1E,$31,$3A,$9C,$98
	.dc.b	$C7,$DD,$19,$3E,$15,$71,$9F,$14,$E8,$F8,$11,$99,$E1,$19,$96,$3E
	.dc.b	$F7,$CD,$19,$69,$AC,$3D,$79,$7C,$C8,$F8,$B7,$D1,$D6,$1D,$EF,$95
	.dc.b	$6F,$0E,$CF,$1A,$EA,$76,$8C,$E5,$F4,$F2,$F8,$13,$D0,$A7,$0A,$E9
	.dc.b	$97,$92,$B8,$4F,$C2,$38,$14,$E0,$FD,$5C,$7C,$D1,$EA,$C7,$E2,$63
	.dc.b	$E2,$C7,$BE,$3D,$AF,$77,$CA,$7F,$05,$97,$C1,$8C,$1F,$60,$F0,$57
	.dc.b	$BC,$E4,$F4,$C7,$CF,$3F,$6A,$31,$5F,$03,$0F,$14,$72,$AF,$43,$A3
	.dc.b	$DD,$1A,$7F,$8D,$47,$FA,$C9,$B3,$FE,$2D,$8C,$47,$FB,$AB,$E8,$5F
	.dc.b	$43,$FE,$E2,$8F,$F7,$84,$61,$FF,$72,$FC,$2A,$31,$1F,$CA,$A3,$13
	.dc.b	$FE,$39,$F8,$17,$9B,$49,$5F,$E1,$28,$CB,$63,$F8,$04,$63,$3F,$E1
	.dc.b	$DF,$43,$FD,$73,$87,$35,$2F,$F8,$15,$EF,$83,$F8,$ED,$EE,$73,$5B
	.dc.b	$C3,$F8,$E5,$EE,$7F,$3F,$5B,$9C,$D6,$EF,$FC,$9B,$7B,$D9,$7B,$9C
	.dc.b	$D6,$F0,$FF,$2A,$DE,$EB,$51,$89,$D3,$FD,$6A,$75,$DC,$CB,$3C,$09
	.dc.b	$F3,$56,$EF,$FA,$56,$1D,$23,$07,$B8,$99,$FC,$8C,$7A,$46,$32,$F3
	.dc.b	$B4,$F8,$31,$A9,$F8,$31,$93,$AB,$CF,$2F,$2E,$3E,$4C,$BC,$D9,$79
	.dc.b	$30,$E5,$8F,$6C,$39,$91,$FA,$B1,$AC,$7E,$21,$DE,$F9,$BF,$D9,$32
	.dc.b	$F7,$46,$2F,$94,$F0,$7C,$8F,$1C,$7C,$87,$CC,$CB,$CF,$97,$D5,$CF
	.dc.b	$EC,$E5,$E1,$BF,$6E,$7E,$72,$3F,$12,$F8,$BF,$1E,$B8,$C6,$31,$F1
	.dc.b	$C6,$27,$C5,$87,$7C,$FD,$39,$79,$23,$19,$78,$73,$F6,$67,$E0,$C3
	.dc.b	$C6,$78,$4E,$67,$87,$0E,$77,$E1,$9F,$97,$9D,$AF,$89,$18,$78,$E7
	.dc.b	$E6,$BE,$31,$89,$D4,$E7,$1F,$26,$5F,$26,$BC,$11,$8C,$7E,$F6,$3E
	.dc.b	$A2,$4E,$9C,$BF,$E1,$CF,$A1,$7C,$BF,$DF,$D7,$E3,$65,$F6,$09,$9F
	.dc.b	$F8,$52,$B7,$87,$29,$FF,$0A,$F9,$B8,$C7,$FC,$64,$D9,$FF,$1C,$B6
	.dc.b	$3F,$D8,$CD,$41,$E5,$3D,$31,$FC,$97,$84,$EC,$7D,$0F,$E1,$6F,$79
	.dc.b	$F4,$BC,$1C,$2B,$79,$7F,$94,$2B,$79,$7F,$A0,$3E,$6B,$FC,$AD,$3F
	.dc.b	$E7,$B6,$C7,$F9,$E9,$E2,$DE,$DF,$1C,$6B,$4F,$F5,$4B,$72,$FF,$67
	.dc.b	$BC,$23,$33,$C3,$E8,$78,$F9,$D7,$1F,$03,$71,$F0,$36,$FA,$BF,$0B
	.dc.b	$E8,$5F,$3B,$4C,$7C,$6D,$C3,$FD,$29,$3A,$6C,$E9,$A4,$C3,$FD,$2F
	.dc.b	$3B,$6D,$73,$7F,$D4,$DD,$AF,$A0,$7D,$0F,$F0,$58,$C3,$F5,$BC,$D7
	.dc.b	$A1,$9A,$DE,$5F,$EB,$BB,$DB,$63,$38,$7F,$AA,$2F,$8C,$E7,$0F,$9F
	.dc.b	$E0,$C8,$FA,$DD,$76,$FE,$B4,$DD,$FE,$9C,$DE,$3F,$C7,$FC,$1F,$FA
	.dc.b	$16,$FF,$1F,$CA,$8C,$47,$E4,$5B,$19,$C7,$FB,$73,$65,$EC,$7F,$38
	.dc.b	$78,$4F,$E6,$A3,$F9,$DF,$53,$DB,$7D,$0F,$F3,$E9,$E1,$87,$36,$E3
	.dc.b	$9C,$3C,$E7,$33,$F4,$B8,$73,$BD,$63,$EA,$AB,$2F,$79,$F4,$38,$7C
	.dc.b	$38,$73,$BF,$F6,$FC,$6A,$31,$5C,$70,$E1,$7C,$5B,$3A,$AE,$77,$F0
	.dc.b	$A3,$F7,$2B,$C8,$FE,$F9,$97,$C5,$6C,$FF,$BC,$7D,$7F,$4D,$6F,$7B
	.dc.b	$5A,$8F,$E0,$6D,$CD,$C6,$BD,$DF,$2D,$B1,$8F,$D6,$9F,$B2,$47,$07
	.dc.b	$C1,$C7,$F8,$74,$7C,$8D,$FE,$14,$7A,$CF,$C2,$C7,$52,$4F,$C6,$9D
	.dc.b	$E7,$F6,$A7,$D7,$DC,$F4,$E9,$F4,$7C,$AB,$87,$86,$74,$D9,$D3,$4F
	.dc.b	$AD,$E2,$3E,$61,$E9,$7E,$02,$30,$FD,$DE,$FC,$4E,$AB,$8B,$75,$F6
	.dc.b	$5F,$09,$FB,$55,$C2,$FA,$7C,$33,$D7,$A7,$E3,$3C,$5F,$F7,$A5,$F9
	.dc.b	$27,$4D,$9D,$BF,$8A,$9D,$E3,$E0,$AD,$BF,$37,$91,$3A,$E0,$FD,$3B
	.dc.b	$6F,$F9,$4F,$95,$E1,$F4,$E4,$75,$7C,$91,$8C,$3B,$71,$7A,$8F,$47
	.dc.b	$D0,$DB,$F4,$45,$9F,$16,$0A,$C6,$3F,$E3,$2D,$DA,$3D,$7F,$8A,$7C
	.dc.b	$3C,$A9,$F0,$DB,$8F,$48,$B7,$CE,$B9,$46,$CF,$8C,$68,$FF,$7F,$BA
	.dc.b	$9F,$91,$5D,$E7,$84,$7C,$BF,$5B,$EB,$9F,$C0,$3D,$7F,$EE,$18,$77
	.dc.b	$8C,$D7,$D7,$75,$E6,$62,$FF,$3C,$3E,$7D,$F4,$3F,$C8,$9B,$7F,$E8
	.dc.b	$D6,$CF,$EF,$8F,$EE,$EE,$1C,$3A,$C3,$F9,$23,$67,$FB,$EF,$CB,$3E
	.dc.b	$81,$F4,$1F,$C0,$AB,$58,$74,$C7,$07,$F1,$EB,$E1,$58,$C3,$F7,$5F
	.dc.b	$3C,$F9,$0F,$E7,$2D,$8F,$F5,$23,$70,$F4,$07,$0F,$AD,$D6,$7F,$CE
	.dc.b	$DB,$3E,$8C,$FF,$64,$F8,$2F,$F8,$CC,$EA,$FF,$C3,$2D,$C7,$DD,$D7
	.dc.b	$AA,$97,$C4,$FE,$65,$3A,$71,$7F,$53,$E7,$5F,$37,$F0,$7E,$79,$F1
	.dc.b	$1F,$CA,$27,$37,$CA,$72,$FF,$1D,$8C,$DF,$3B,$FE,$8F,$F3,$CF,$9E
	.dc.b	$7A,$FC,$E6,$61,$C5,$DC,$EA,$30,$7F,$29,$CF,$E0,$60,$7E,$17,$7B
	.dc.b	$EB,$3F,$A1,$56,$F6,$FE,$7F,$5F,$DF,$2F,$F2,$2D,$E2,$FF,$50,$FB
	.dc.b	$AF,$9E,$FC,$1D,$75,$F6,$3F,$A3,$BF,$E4,$6F,$53,$FC,$8F,$EC,$3E
	.dc.b	$17,$EE,$C6,$1B,$29,$D3,$F9,$99,$D3,$FE,$50,$F8,$37,$B7,$FD,$F2
	.dc.b	$38,$5F,$F5,$C9,$CE,$7C,$25,$18,$76,$73,$AE,$58,$78,$21,$F3,$FF
	.dc.b	$E2,$87,$FA,$93,$5F,$D9,$B6,$FF,$C1,$9F,$46,$FB,$4F,$ED,$B1,$FE
	.dc.b	$0F,$97,$CF,$63,$8B,$FB,$8C,$63,$0E,$B3,$95,$C8,$FA,$D7,$8B,$99
	.dc.b	$FF,$6D,$7D,$15,$F3,$4D,$35,$F0,$9F,$56,$7F,$B5,$4E,$9B,$3B,$7F
	.dc.b	$B7,$4E,$DF,$E1,$F3,$B7,$F8,$8C,$7F,$81,$DD,$FB,$5C,$5F,$E2,$BF
	.dc.b	$45,$EB,$6E,$7E,$78,$C6,$DF,$E3,$1F,$A2,$FB,$D5,$FD,$F6,$73,$BF
	.dc.b	$F8,$51,$B1,$FE,$8E,$6A,$9E,$2F,$63,$95,$F7,$3E,$34,$6A,$FF,$C8
	.dc.b	$15,$E5,$AF,$E4,$8D,$7C,$15,$FC,$41,$F0,$35,$7D,$13,$E4,$3E,$93
	.dc.b	$63,$FC,$2C,$D7,$F9,$BF,$CE,$7F,$A9,$5C,$CF,$02,$3F,$E1,$D8,$D3
	.dc.b	$FC,$FA,$79,$3F,$CE,$5A,$FF,$8E,$1A,$FF,$AC,$9B,$E8,$FF,$62,$FA
	.dc.b	$1F,$C5,$36,$7F,$C0,$51,$F8,$DE,$3F,$C5,$FD,$16,$BD,$34,$D2,$57
	.dc.b	$F9,$3B,$91,$5B,$FA,$29,$3F,$E0,$D7,$A4,$FF,$3D,$8C,$10,$FF,$06
	.dc.b	$46,$BC,$3E,$31,$83,$5C,$1F,$E7,$2A,$E5,$19,$24,$7F,$75,$AE,$51
	.dc.b	$FE,$1D,$FA,$2F,$51,$1F,$46,$F5,$FF,$0E,$45,$AE,$D3,$E8,$F6,$A5
	.dc.b	$7B,$DF,$5F,$C3,$E2,$3F,$C4,$58,$FF,$8B,$DD,$1F,$16,$3E,$57,$C1
	.dc.b	$48,$FF,$18,$3C,$70,$FF,$06,$39,$8F,$F1,$73,$9E,$6F,$F7,$5F,$0B
	.dc.b	$E8,$9F,$45,$F6,$9B,$7F,$E4,$E6,$BF,$E3,$5F,$9C,$7C,$F6,$EA,$4E
	.dc.b	$DF,$C0,$4E,$A3,$FC,$D9,$F4,$0E,$8F,$E1,$8D,$8C,$C7,$0F,$A2,$CB
	.dc.b	$23,$27,$FA,$4A,$93,$FC,$57,$E5,$FB,$4F,$F8,$61,$BB,$7F,$AC,$5B
	.dc.b	$AF,$FA,$9B,$E2,$7F,$DC,$B5,$FE,$7F,$9C,$A0,$E1,$BA,$7E,$6E,$78
	.dc.b	$C7,$78,$FD,$C4,$F8,$A3,$FC,$44,$D7,$FD,$42,$D7,$B7,$D0,$7F,$78
	.dc.b	$7F,$D2,$0D,$7F,$C6,$2D,$C3,$FD,$8F,$F1,$9F,$B8,$7F,$D9,$91,$9C
	.dc.b	$7F,$D5,$AD,$AE,$BE,$CF,$EC,$6F,$85,$7F,$B1,$B4,$FE,$FB,$5F,$2F
	.dc.b	$C3,$FA,$E7,$D8,$F5,$A4,$64,$87,$E0,$5F,$A1,$F3,$F2,$3E,$53,$F1
	.dc.b	$7E,$83,$11,$F4,$E2,$E9,$FE,$E0,$DD,$5A,$AF,$8F,$5F,$28,$8F,$F8
	.dc.b	$D3,$C1,$FE,$A7,$F0,$FF,$EC,$AF,$49,$F2,$5F,$EE,$0E,$CF,$F6,$05
	.dc.b	$FF,$80,$63,$15,$B7,$5E,$A5,$9F,$EB,$F5,$C1,$5F,$47,$37,$BC,$3C
	.dc.b	$31,$D4,$E3,$7C,$FD,$17,$CF,$FF,$AC,$A7,$51,$8F,$8B,$D4,$E6,$7B
	.dc.b	$E7,$FE,$F8,$E4,$7D,$06,$AE,$57,$2B,$E1,$A7,$97,$3F,$F6,$FD,$F9
	.dc.b	$7C,$37,$CA,$D2,$61,$E7,$AF,$E9,$5F,$25,$88,$F0,$9F,$E0,$0A,$E2
	.dc.b	$ED,$AE,$A7,$8B,$9A,$E5,$87,$8E,$B6,$DB,$F9,$1C,$1F,$ED,$FB,$F1
	.dc.b	$3F,$EA,$76,$FC,$5E,$BE,$42,$46,$08,$78,$67,$93,$75,$EA,$47,$AF
	.dc.b	$C6,$57,$47,$82,$3C,$71,$F8,$18,$FD,$EF,$4D,$EE,$34,$FC,$07,$C8
	.dc.b	$BE,$43,$15,$F8,$13,$89,$6B,$8B,$F3,$E7,$83,$9A,$FE,$18,$BB,$BF
	.dc.b	$82,$C6,$1B,$1F,$C4,$5A,$EC,$FE,$6F,$F3,$DF,$E2,$37,$EC,$F9,$08
	.dc.b	$C7,$07,$E6,$63,$1E,$B7,$E7,$D7,$CF,$FF,$80,$27,$55,$C3,$A7,$AD
	.dc.b	$EE,$31,$83,$0E,$3E,$5E,$6B,$E4,$73,$7F,$8B,$31,$FD,$45,$FF,$28
	.dc.b	$8C,$63,$D2,$3F,$51,$87,$91,$F9,$13,$A7,$F9,$86,$DF,$9A,$EB,$DE
	.dc.b	$7E,$2B,$E1,$5F,$98,$3B,$4F,$23,$79,$79,$1C,$46,$EF,$A3,$F2,$2F
	.dc.b	$DA,$F2,$35,$87,$8D,$C6,$5F,$1B,$0E,$91,$AC,$3A,$CE,$62,$E5,$FC
	.dc.b	$FA,$F9,$63,$E0,$AD,$57,$12,$3F,$22,$35,$5F,$1E,$72,$F3,$BE,$B1
	.dc.b	$F4,$F1,$F4,$D7,$08,$F9,$38,$78,$0F,$99,$97,$C8,$C7,$B4,$67,$0E
	.dc.b	$73,$9C,$7B,$5F,$80,$E7,$3F,$31,$F0,$9F,$12,$37,$7E,$C3,$AC,$EB
	.dc.b	$0F,$25,$F5,$BE,$67,$53,$91,$EB,$CB,$E6,$DF,$28,$F1,$4F,$17,$64
	.dc.b	$AE,$B9,$7C,$2C,$BD,$93,$C2,$31,$8F,$8E,$7E,$1C,$62,$7C,$55,$CB
	.dc.b	$1F,$0E,$5E,$C2,$4F,$94,$E5,$87,$C8,$BE,$59,$79,$C9,$5F,$65,$B5
	.dc.b	$F0,$1B,$97,$AF,$1F,$B1,$3E,$BC,$BC,$B9,$78,$A3,$86,$1E,$2B,$E1
	.dc.b	$5E,$EC,$39,$E3,$E2,$AF,$3D,$FC,$CB,$DE,$5E,$88,$FD,$06,$3D,$5F
	.dc.b	$04,$7D,$D7,$F2,$87,$F8,$63,$1F,$F4,$FB,$F8,$43,$11,$FE,$97,$F7
	.dc.b	$DF,$28,$F9,$FC,$36,$BF,$8F,$9F,$EE,$97,$0E,$27,$8C,$FF,$A8,$75
	.dc.b	$7B,$6F,$84,$C3,$6F,$99,$FE,$34,$70,$FF,$8B,$9A,$FF,$8B,$9E,$1B
	.dc.b	$FE,$B5,$BB,$7F,$98,$3E,$7D,$D9,$FC,$B6,$B5,$87,$F3,$BB,$DE,$1F
	.dc.b	$C5,$1B,$C0,$C3,$8F,$26,$EE,$5F,$F0,$FB,$DD,$6A,$31,$3A,$C3,$BC
	.dc.b	$7F,$3A,$9D,$46,$09,$97,$AE,$BF,$94,$CE,$A3,$15,$D0,$95,$FD,$06
	.dc.b	$75,$1E,$9A,$F0,$57,$0A,$D6,$5F,$06,$30,$4A,$E0,$79,$63,$4E,$27
	.dc.b	$CA,$D9,$F9,$93,$A8,$FA,$D5,$E0,$74,$7F,$4B,$BF,$2B,$A3,$D3,$5E
	.dc.b	$1C,$BE,$75,$F3,$CF,$F0,$B8,$73,$BE,$18,$FA,$B3,$C5,$6A,$79,$9D
	.dc.b	$EF,$96,$7F,$72,$38,$E5,$DF,$3F,$1D,$F1,$C7,$DB,$5A,$39,$63,$DB
	.dc.b	$0E,$55,$BC,$3E,$0C,$E7,$2F,$81,$9F,$AA,$31,$87,$92,$30,$F9,$8E
	.dc.b	$C7,$77,$38,$F8,$B0,$ED,$3F,$2A,$73,$87,$D7,$8E,$33,$A9,$CC,$7D
	.dc.b	$03,$38,$FE,$53,$3F,$87,$97,$93,$1F,$95,$5E,$3C,$3B,$BE,$3C,$FE
	.dc.b	$2E,$1F,$00,$97,$F9,$4C,$39,$D7,$19,$ED,$9F,$C2,$BE,$37,$B8,$C5
	.dc.b	$F0,$C7,$C2,$7E,$33,$2F,$49,$3E,$7D,$F2,$9F,$92,$8C,$37,$1F,$CF
	.dc.b	$C2,$77,$87,$7C,$FF,$C4,$F1,$87,$F3,$91,$FE,$14,$8C,$61,$E2,$D7
	.dc.b	$FD,$CB,$18,$6E,$3F,$A2,$F0,$1F,$30,$E5,$F2,$5E,$F2,$FF,$65,$73
	.dc.b	$BD,$5F,$F1,$D7,$97,$FE,$3B,$C5,$D9,$FE,$DA,$DB,$FC,$99,$7B,$7F
	.dc.b	$DA,$55,$B9,$D6,$1F,$E2,$1A,$D2,$9D,$7A,$3F,$B1,$9C,$C6,$2F,$AE
	.dc.b	$3F,$E4,$BB,$FE,$6B,$5C,$9B,$19,$7F,$CA,$F1,$97,$FD,$C9,$8F,$C0
	.dc.b	$F8,$8F,$5D,$E3,$57,$C1,$62,$D7,$BA,$F5,$9B,$9D,$46,$67,$D5,$39
	.dc.b	$36,$F2,$A5,$F0,$7B,$CE,$AB,$E9,$FC,$D6,$7E,$6B,$4D,$9D,$52,$31
	.dc.b	$87,$80,$EC,$E3,$4F,$AD,$F3,$AF,$96,$F3,$BF,$ED,$8C,$3C,$0E,$5B
	.dc.b	$E8,$EA,$76,$D8,$C9,$C2,$32,$D3,$38,$78,$CF,$C6,$74,$FC,$6F,$2F
	.dc.b	$FD,$25,$3F,$12,$72,$FC,$18,$CC,$71,$E5,$FE,$A0,$FF,$72,$6B,$FE
	.dc.b	$92,$6A,$39,$9F,$DE,$36,$FC,$EF,$C8,$7C,$FF,$AF,$FA,$0D,$C3,$E5
	.dc.b	$B6,$BE,$55,$70,$9F,$9F,$5F,$83,$97,$1B,$D2,$FC,$0F,$A9,$F7,$2F
	.dc.b	$D9,$39,$7E,$E4,$66,$BE,$24,$E5,$EF,$19,$E0,$FC,$E4,$E5,$FD,$74
	.dc.b	$66,$BF,$37,$7F,$9D,$BF,$DA,$B5,$F2,$B6,$FB,$47,$FA,$E3,$1F,$DF
	.dc.b	$FD,$7F,$E4,$7D,$DF,$C9,$37,$0F,$C6,$B6,$BE,$1F,$22,$31,$9F,$EC
	.dc.b	$7A,$9E,$87,$E8,$BD,$07,$C0,$7C,$79,$CE,$5F,$9D,$9C,$BF,$81,$8C
	.dc.b	$E5,$FE,$24,$6B,$FE,$30,$6B,$FE,$D0,$6B,$F7,$FE,$F7,$C9,$3D,$7E
	.dc.b	$D7,$EB,$46,$27,$ED,$DA,$73,$9F,$9D,$B7,$C4,$FC,$A5,$F1,$BF,$F2
	.dc.b	$0B,$5F,$F7,$0B,$5F,$15,$7D,$3F,$11,$F3,$0F,$53,$E2,$71,$7F,$0B
	.dc.b	$9C,$C6,$31,$FE,$0B,$18,$7F,$C8,$AD,$C3,$FC,$A2,$D7,$FD,$E2,$D7
	.dc.b	$F4,$0F,$83,$E1,$FE,$97,$AF,$DF,$14,$62,$FB,$C6,$1E,$ED,$BE,$ED
	.dc.b	$7B,$B5,$EE,$BE,$3F,$F9,$37,$C9,$FF,$45,$F1,$7C,$11,$87,$F9,$CD
	.dc.b	$F0,$8F,$8D,$5C,$63,$C3,$18,$3F,$39,$CF,$FD,$53,$E0,$3F,$93,$1F
	.dc.b	$C9,$FC,$0B,$5F,$C8,$A2,$E5,$F9,$69,$BC,$EA,$F3,$F3,$7E,$57,$E0
	.dc.b	$FE,$23,$63,$E1,$BE,$42,$3F,$11,$F3,$F2,$FF,$3E,$FA,$CF,$B9,$3F
	.dc.b	$2A,$31,$C7,$F3,$DD,$C7,$CC,$6E,$5F,$31,$D4,$7F,$3A,$DA,$CF,$F3
	.dc.b	$9E,$3F,$C4,$D6,$33,$FE,$DD,$D5,$F8,$CC,$3C,$07,$DA,$E4,$FD,$87
	.dc.b	$27,$DF,$6C,$64,$FB,$AF,$D0,$23,$FE,$92,$C5,$F3,$4F,$8A,$FE,$C1
	.dc.b	$8B,$EA,$5F,$53,$FC,$86,$B5,$87,$F5,$CA,$D4,$E6,$F8,$DF,$0A,$FE
	.dc.b	$7D,$7C,$30,$FE,$AD,$87,$09,$CD,$FF,$19,$AD,$57,$FA,$C1,$AF,$F3
	.dc.b	$D3,$CF,$1F,$C7,$FE,$6D,$E2,$7E,$4B,$E7,$1C,$FF,$CD,$E7,$38,$FF
	.dc.b	$72,$EB,$56,$A3,$FA,$94,$EB,$1F,$ED,$55,$AB,$E5,$87,$56,$DF,$08
	.dc.b	$FF,$03,$7C,$E3,$E7,$1F,$13,$AC,$7F,$B1,$65,$D7,$6F,$F0,$C3,$71
	.dc.b	$FE,$B1,$87,$1B,$E1,$5F,$E1,$DF,$5D,$F6,$5F,$C2,$7E,$F9,$F5,$7F
	.dc.b	$D4,$E7,$38,$FF,$0A,$9C,$BF,$DB,$63,$39,$FE,$36,$72,$FF,$75,$8C
	.dc.b	$E1,$FA,$78,$74,$AB,$73,$E3,$C7,$C0,$7F,$89,$7D,$B7,$DD,$BE,$93
	.dc.b	$39,$FF,$7D,$E3,$7C,$6E,$FE,$BB,$9C,$FD,$B9,$CB,$FE,$26,$C1,$3A
	.dc.b	$D7,$FB,$E5,$7E,$C1,$FC,$4F,$DF,$F0,$7F,$C1,$D1,$9D,$3F,$CA,$B3
	.dc.b	$97,$FB,$37,$C0,$7F,$81,$7E,$19,$F4,$5F,$EF,$28,$C3,$70,$4E,$9F
	.dc.b	$F8,$03,$3A,$7F,$C1,$DF,$0C,$FA,$2F,$F2,$C4,$70,$BF,$CB,$4E,$70
	.dc.b	$E1,$28,$C3,$B3,$9D,$72,$C3,$C1,$0F,$A1,$7D,$9F,$E8,$70,$FF,$30
	.dc.b	$7D,$1F,$F6,$EA,$FF,$15,$35,$FF,$18,$57,$07,$FC,$5C,$BE,$AB,$97
	.dc.b	$0F,$EA,$E9,$E6,$E2,$FF,$C5,$51,$87,$8D,$65,$7D,$DB,$E3,$70,$D8
	.dc.b	$FF,$1E,$35,$FF,$21,$6A,$F8,$0D,$D7,$38,$FD,$3C,$FF,$90,$DA,$FF
	.dc.b	$92,$6B,$96,$9F,$E7,$F6,$E7,$FC,$0E,$B7,$3F,$E8,$36,$C7,$F0,$79
	.dc.b	$DD,$7F,$91,$7E,$2F,$FC,$FA,$FF,$94,$5A,$FF,$95,$DB,$7F,$E8,$DF
	.dc.b	$91,$FF,$3B,$CE,$FE,$23,$E1,$56,$B0,$FE,$28,$D9,$FF,$29,$57,$07
	.dc.b	$A5,$7C,$98,$FF,$57,$47,$F9,$85,$AE,$8E,$C6,$7E,$0B,$81,$E0,$9F
	.dc.b	$A5,$F4,$6F,$A3,$FE,$75,$48,$C3,$FC,$F5,$AF,$F2,$93,$FC,$CC,$E1
	.dc.b	$C3,$AC,$BF,$D5,$9F,$49,$FE,$1A,$9F,$F3,$B3,$5F,$F3,$E4,$EF,$3F
	.dc.b	$E6,$8D,$8F,$F3,$94,$EE,$B9,$FA,$9E,$8C,$3F,$1D,$E1,$BE,$7F,$F8
	.dc.b	$EE,$1F,$E2,$06,$BF,$E2,$06,$BF,$E2,$06,$BF,$E2,$07,$77,$FE,$D3
	.dc.b	$F8,$0F,$F0,$A4,$FF,$93,$FD,$CE,$F5,$AD,$38,$6D,$82,$57,$FA,$C7
	.dc.b	$E7,$18,$62,$3F,$BA,$5F,$D6,$C7,$F9,$36,$5E,$9C,$3F,$51,$EE,$B1
	.dc.b	$E3,$87,$F8,$F3,$0F,$BA,$FA,$6B,$FB,$85,$F1,$8F,$81,$18,$9E,$D9
	.dc.b	$7F,$B5,$BF,$00,$7C,$7F,$66,$DF,$5B,$F6,$F1,$71,$24,$7C,$B6,$BF
	.dc.b	$E6,$2F,$A2,$F3,$EC,$78,$C7,$C1,$FB,$4F,$5E,$A6,$98,$F6,$6C,$EC
	.dc.b	$87,$FA,$A2,$7F,$C0,$EF,$C1,$F9,$0F,$61,$1F,$BC,$D7,$FB,$7F,$98
	.dc.b	$E8,$FE,$25,$E4,$7F,$4C,$7E,$CD,$7D,$AF,$8C,$F8,$C4,$BF,$C4,$35
	.dc.b	$FF,$3F,$F9,$7E,$E3,$83,$F4,$FE,$2B,$E7,$11,$FC,$83,$5F,$F4,$39
	.dc.b	$F1,$8F,$C4,$BF,$C8,$7F,$01,$FD,$7C,$1A,$47,$F6,$6F,$EB,$FA,$3F
	.dc.b	$C7,$3E,$87,$E1,$BE,$03,$F2,$55,$FE,$D7,$AD,$73,$FF,$87,$5B,$0E
	.dc.b	$A5,$70,$6D,$F8,$E7,$F9,$A7,$D1,$FF,$37,$7F,$7C,$1C,$3F,$C0,$5A
	.dc.b	$FF,$8A,$DA,$FF,$BA,$5E,$58,$FF,$94,$3D,$5F,$41,$DB,$C9,$FC,$2F
	.dc.b	$D5,$FE,$1B,$E8,$A6,$1F,$A9,$8C,$BF,$AB,$8C,$BF,$85,$8F,$E9,$5F
	.dc.b	$32,$FA,$4F,$E3,$2D,$9F,$F2,$A3,$5F,$5B,$5F,$E3,$4E,$71,$FF,$0C
	.dc.b	$70,$BE,$CF,$FC,$83,$5F,$E0,$B6,$A9,$FF,$0F,$2A,$E5,$C3,$FE,$4D
	.dc.b	$6C,$2F,$A7,$C0,$7F,$65,$6C,$0E,$95,$C7,$1F,$D4,$FD,$97,$EA,$BC
	.dc.b	$69,$5F,$C3,$63,$2F,$F0,$F8,$CB,$FA,$98,$FE,$DD,$F4,$DE,$88,$FE
	.dc.b	$70,$D7,$FC,$C4,$D7,$C4,$D4,$FC,$0C,$FF,$C7,$9C,$6F,$A6,$7C,$7F
	.dc.b	$F1,$86,$DF,$AE,$BE,$2D,$F1,$CF,$DD,$83,$E5,$DF,$65,$FC,$81,$FE
	.dc.b	$DF,$87,$E5,$DB,$3B,$9D,$1B,$F8,$29,$5D,$EF,$AD,$F4,$8C,$1C,$EB
	.dc.b	$6F,$CE,$3B,$57,$47,$F9,$C3,$FC,$63,$BA,$46,$08,$7F,$15,$72,$FF
	.dc.b	$17,$8F,$E3,$FE,$A7,$F2,$0A,$93,$FD,$0E,$32,$FF,$45,$8C,$BF,$C7
	.dc.b	$E3,$F3,$7C,$DF,$9A,$24,$7E,$69,$AF,$A7,$27,$CA,$FF,$74,$F7,$5F
	.dc.b	$3F,$F3,$A3,$0F,$E4,$EF,$AD,$FC,$4B,$E3,$C7,$FE,$A3,$C3,$8C,$FF
	.dc.b	$07,$C3,$8B,$F9,$5E,$76,$1F,$DC,$3E,$8D,$F2,$BC,$73,$FC,$D7,$A2
	.dc.b	$9C,$E1,$F9,$DC,$BA,$CE,$6F,$84,$FF,$15,$F2,$1F,$45,$8A,$D4,$F1
	.dc.b	$C7,$AF,$D1,$FF,$87,$B1,$E6,$D8,$F0,$E5,$E0,$A5,$78,$23,$0F,$0F
	.dc.b	$43,$B5,$F1,$AF,$F4,$BD,$6A,$FB,$D6,$89,$3E,$33,$79,$E1,$EF,$3F
	.dc.b	$C0,$72,$FF,$40,$C6,$0F,$94,$E3,$1F,$14,$73,$C9,$97,$8E,$3F,$86
	.dc.b	$56,$EF,$A7,$C0,$6E,$B5,$B7,$C8,$8F,$25,$F0,$6C,$66,$3C,$58,$76
	.dc.b	$CF,$E8,$C6,$CF,$31,$72,$FC,$1B,$71,$F8,$6F,$4C,$BC,$8F,$B3,$1C
	.dc.b	$1E,$5B,$EB,$8F,$87,$0E,$6D,$C3,$D7,$7E,$4C,$BD,$78,$F4,$8F,$01
	.dc.b	$D4,$D6,$1D,$23,$57,$CE,$75,$87,$CC,$7D,$71,$97,$99,$D3,$2F,$84
	.dc.b	$FA,$0D,$D7,$43,$DB,$1A,$8C,$E5,$BB,$F5,$D6,$B1,$F4,$CE,$72,$F6
	.dc.b	$D7,$8C,$DC,$66,$FC,$F3,$F3,$1E,$A4,$9D,$BF,$00,$E0,$7A,$B1,$F8
	.dc.b	$18,$F9,$67,$64,$AF,$09,$D5,$C5,$78,$09,$3C,$67,$71,$93,$E3,$61
	.dc.b	$EC,$7C,$91,$A9,$E1,$1B,$BF,$9C,$79,$2B,$85,$F8,$27,$9E,$5F,$67
	.dc.b	$1F,$5D,$EE,$31,$3D,$8F,$2D,$F2,$C7,$CB,$1C,$8E,$39,$7D,$7C,$3F
	.dc.b	$21,$18,$7E,$76,$5E,$68,$F9,$13,$B7,$E6,$1D,$CC,$CF,$27,$E9,$9F
	.dc.b	$E0,$68,$FD,$DB,$63,$FC,$1B,$18,$8F,$F7,$BF,$B8,$FA,$3F,$F1,$4C
	.dc.b	$7F,$BD,$E3,$0F,$F8,$B3,$E0,$BF,$DC,$13,$A6,$C7,$F8,$5B,$DC,$FF
	.dc.b	$04,$37,$C2,$FF,$0D,$C6,$23,$FC,$32,$D7,$FC,$20,$F0,$DF,$F5,$ED
	.dc.b	$D3,$F9,$AF,$C4,$FF,$BA,$7B,$7F,$82,$CE,$6F,$F9,$F3,$5F,$F0,$E3
	.dc.b	$9D,$95,$C6,$7F,$5B,$3B,$E1,$34,$E1,$3B,$C7,$C7,$97,$80,$97,$FC
	.dc.b	$52,$75,$97,$AA,$73,$5C,$2B,$5D,$2C,$F3,$7B,$9F,$1D,$FF,$2B,$C3
	.dc.b	$9C,$7B,$23,$6D,$39,$D6,$1F,$7C,$6B,$3F,$7E,$1D,$E7,$A9,$31,$F2
	.dc.b	$1F,$D7,$30,$EF,$1C,$8D,$63,$E5,$9D,$57,$37,$8C,$64,$92,$BF,$0E
	.dc.b	$3E,$5C,$FE,$6E,$5E,$5C,$BB,$C6,$2F,$B9,$ED,$C7,$BD,$F1,$CF,$EA
	.dc.b	$61,$F1,$73,$F1,$67,$E0,$C3,$D1,$87,$91,$F1,$65,$DB,$0E,$31,$DB
	.dc.b	$3F,$2B,$E0,$38,$DF,$AA,$33,$7C,$E7,$27,$96,$7F,$27,$87,$8A,$30
	.dc.b	$43,$75,$CA,$7C,$D9,$FB,$EF,$CD,$39,$AE,$13,$9A,$F1,$63,$E9,$CB
	.dc.b	$E1,$1F,$07,$3B,$7E,$B9,$D3,$EB,$BE,$4F,$C2,$CB,$CC,$FA,$89,$1A
	.dc.b	$C7,$C5,$18,$CF,$E3,$E7,$F4,$A3,$B5,$F2,$BE,$18,$78,$2B,$58,$FF
	.dc.b	$A0,$EF,$C0,$FF,$A1,$4C,$63,$F4,$1B,$5F,$B3,$6E,$1F,$2E,$7E,$20
	.dc.b	$FA,$27,$D1,$E6,$FF,$4B,$39,$8C,$57,$1A,$DD,$FC,$7C,$3F,$C1,$31
	.dc.b	$87,$FD,$43,$7C,$67,$CD,$1F,$E4,$29,$D4,$7A,$1B,$19,$BE,$C7,$E0
	.dc.b	$7A,$FF,$DA,$BC,$B8,$BF,$F0,$DF,$37,$FA,$39,$FF,$2C,$75,$3D,$AF
	.dc.b	$A7,$39,$C7,$FC,$C9,$39,$AE,$3E,$27,$FA,$DA,$7E,$65,$6D,$F9,$C7
	.dc.b	$F9,$6A,$34,$FF,$AC,$E3,$4F,$08,$FE,$05,$E1,$FF,$B0,$27,$C0,$FF
	.dc.b	$A9,$3E,$7D,$E9,$EB,$0E,$07,$F9,$BF,$D0,$7D,$1E,$72,$FF,$7A,$43
	.dc.b	$DB,$C6,$FF,$EF,$6A,$B3,$9B,$E5,$18,$AC,$B6,$3B,$7D,$1F,$0C,$78
	.dc.b	$B6,$32,$78,$E3,$2D,$3F,$D0,$9F,$46,$F9,$EF,$1A,$FA,$3B,$18,$6C
	.dc.b	$F0,$6E,$9F,$E9,$5A,$E1,$3D,$2B,$83,$4E,$07,$FA,$DA,$5C,$7F,$D3
	.dc.b	$77,$FF,$67,$F9,$6F,$A4,$F9,$15,$AE,$9F,$8F,$39,$C3,$96,$1C,$5F
	.dc.b	$D5,$36,$9C,$3F,$ED,$9F,$A4,$E7,$39,$87,$13,$E5,$FE,$8F,$B9,$F4
	.dc.b	$9B,$8F,$DB,$6E,$3F,$EB,$3A,$E1,$3F,$7E,$BC,$FC,$CC,$BE,$EF,$37
	.dc.b	$E0,$AB,$E0,$4E,$5F,$F5,$A4,$67,$0F,$F6,$B1,$FE,$AE,$AF,$CA,$B6
	.dc.b	$7F,$C1,$B5,$C6,$3F,$36,$64,$AE,$77,$FC,$C5,$78,$B9,$3C,$AF,$E3
	.dc.b	$BE,$8B,$F1,$FE,$ED,$C7,$F3,$0D,$C7,$DC,$DC,$3F,$B2,$6D,$F2,$23
	.dc.b	$EA,$F5,$3E,$93,$F3,$7F,$08,$FA,$3C,$C6,$1B,$97,$E0,$6B,$83,$FA
	.dc.b	$7F,$2B,$39,$FC,$2F,$A4,$6B,$D3,$E8,$BF,$07,$18,$BF,$F6,$E6,$69
	.dc.b	$CD,$7C,$8A,$D5,$FC,$66,$BF,$C0,$9A,$FF,$31,$77,$B7,$F8,$06,$11
	.dc.b	$8C,$11,$FB,$AF,$8F,$EF,$18,$AF,$F0,$47,$8A,$F8,$C7,$D1,$FC,$4E
	.dc.b	$5F,$E0,$53,$98,$C5,$F1,$3F,$DD,$75,$C1,$C6,$3F,$E0,$58,$C3,$F9
	.dc.b	$1F,$B1,$E7,$18,$C7,$C2,$BE,$93,$E9,$FD,$26,$9B,$87,$4C,$BC,$D1
	.dc.b	$98,$FE,$31,$E1,$DC,$98,$F4,$F9,$0F,$E4,$F8,$7B,$5B,$91,$A4,$C3
	.dc.b	$B4,$65,$FE,$49,$F0,$5E,$89,$FE,$3E,$C8,$FE,$51,$E2,$5C,$7F,$5D
	.dc.b	$37,$1F,$C8,$F7,$FF,$17,$BE,$67,$6E,$8F,$19,$1F,$D2,$E7,$F4,$E3
	.dc.b	$18,$7B,$27,$E7,$C7,$F3,$0F,$4B,$E5,$BF,$33,$9B,$DA,$49,$E7,$AF
	.dc.b	$E3,$DB,$1B,$7F,$A0,$45,$BE,$C7,$D9,$9C,$65,$EA,$9B,$5F,$80,$9B
	.dc.b	$5F,$D1,$A6,$E3,$F8,$BE,$0F,$9E,$4B,$F1,$B6,$BB,$CE,$E7,$FA,$97
	.dc.b	$CC,$7E,$1A,$BF,$95,$FD,$1B,$E8,$FF,$AA,$66,$FA,$47,$C8,$6A,$73
	.dc.b	$18,$C7,$F9,$C5,$F1,$7F,$D3,$ED,$CF,$F8,$9B,$5F,$ED,$ED,$7F,$C3
	.dc.b	$CE,$F0,$FD,$73,$67,$F5,$BF,$1E,$FA,$FF,$D5,$47,$83,$1F,$E8,$7F
	.dc.b	$47,$FC,$C3,$1F,$9B,$6F,$A5,$72,$EB,$CA,$FE,$E1,$39,$8C,$63,$FE
	.dc.b	$B6,$C3,$96,$1F,$E0,$46,$BF,$E3,$36,$BF,$C0,$1F,$06,$9F,$C1,$5B
	.dc.b	$3F,$C1,$3E,$91,$F4,$AF,$88,$EF,$87,$F3,$CB,$DB,$CF,$3F,$F0,$25
	.dc.b	$F0,$AD,$E3,$FC,$47,$DA,$7C,$9F,$F0,$8F,$A5,$7C,$9F,$F6,$49,$CE
	.dc.b	$BF,$BB,$9C,$BF,$DB,$63,$39,$FF,$46,$6B,$FE,$22,$6B,$FE,$51,$6B
	.dc.b	$FC,$79,$E3,$87,$8F,$0F,$F0,$D7,$C0,$BE,$59,$F4,$BE,$39,$FE,$EF
	.dc.b	$C4,$FA,$5F,$EB,$2D,$CF,$FD,$CA,$DC,$FF,$AC,$B5,$FF,$1A,$BF,$C7
	.dc.b	$9B,$E3,$F2,$24,$FF,$BB,$A3,$F9,$9F,$C1,$7F,$34,$F8,$87,$77,$F6
	.dc.b	$48,$C3,$65,$3A,$7F,$C8,$73,$A7,$F9,$EF,$C3,$3C,$CF,$E4,$31,$C2
	.dc.b	$FF,$CE,$D3,$9E,$0E,$12,$8C,$3B,$39,$D7,$2C,$3C,$10,$FA,$27,$D1
	.dc.b	$FA,$DB,$1F,$DF,$0D,$46,$4D,$7D,$14,$8D,$4A,$31,$1F,$E0,$4F,$A2
	.dc.b	$7A,$9F,$E8,$DC,$FF,$AE,$46,$27,$FC,$64,$D7,$FC,$23,$93,$DF,$E1
	.dc.b	$BF,$FA,$56,$B5,$5F,$E6,$A6,$B9,$C3,$EA,$46,$23,$FB,$3B,$D7,$0F
	.dc.b	$ED,$98,$7E,$96,$7F,$C8,$4D,$7F,$C4,$33,$BB,$FF,$52,$7C,$57,$F9
	.dc.b	$9E,$7F,$C9,$0D,$7F,$C4,$ED,$BF,$EF,$1F,$27,$FE,$70,$9D,$FC,$37
	.dc.b	$A9,$FE,$F3,$5B,$BE,$B3,$99,$57,$17,$AC,$7E,$3E,$BC,$11,$FE,$8A
	.dc.b	$6B,$C8,$FB,$0F,$57,$FC,$0F,$3F,$E5,$E7,$FC,$C0,$D7,$A9,$FE,$63
	.dc.b	$FA,$27,$C9,$F8,$DB,$19,$8F,$37,$D1,$7E,$6A,$30,$7E,$1E,$76,$FF
	.dc.b	$91,$1A,$FF,$A8,$1A,$FF,$B3,$DD,$E5,$FE,$1F,$9F,$F0,$C4,$FF,$9D
	.dc.b	$5A,$FF,$95,$27,$79,$7F,$88,$BC,$D7,$D1,$7F,$16,$AF,$F0,$63,$5F
	.dc.b	$EA,$6D,$C3,$FB,$CF,$27,$D3,$AE,$3D,$7F,$9F,$6D,$FF,$78,$BE,$4F
	.dc.b	$FB,$89,$AF,$85,$F9,$57,$FE,$30,$6E,$19,$F8,$5F,$C0,$7C,$0B,$E5
	.dc.b	$1E,$FF,$86,$F7,$8B,$2F,$5E,$19,$38,$DF,$09,$F6,$C6,$5A,$7F,$8D
	.dc.b	$B0,$ED,$86,$63,$FD,$9F,$7D,$1F,$65,$7B,$F0,$FC,$14,$7B,$78,$BC
	.dc.b	$84,$7A,$63,$FC,$11,$F9,$71,$C4,$8F,$C7,$7F,$84,$63,$EF,$FA,$0F
	.dc.b	$49,$27,$9F,$A5,$5F,$F0,$AC,$ED,$EF,$3F,$02,$7E,$37,$C5,$C9,$FB
	.dc.b	$AF,$1D,$BE,$83,$EB,$24,$7D,$97,$DD,$F1,$1F,$3C,$91,$EC,$FA,$6A
	.dc.b	$F8,$B7,$78,$FF,$61,$B1,$F8,$CA,$FB,$1F,$0D,$23,$F1,$EF,$D6,$F9
	.dc.b	$CC,$47,$33,$91,$8C,$BF,$CF,$9F,$43,$FE,$7A,$8F,$F6,$93,$5F,$F5
	.dc.b	$8B,$63,$FD,$3F,$C1,$FC,$B6,$76,$7F,$51,$9F,$EA,$49,$A4,$7F,$23
	.dc.b	$EF,$7E,$87,$97,$FD,$13,$C3,$FE,$D0,$E6,$FE,$21,$7D,$E7,$C9,$F4
	.dc.b	$2F,$A1,$FE,$28,$DC,$7F,$DD,$0D,$7E,$CE,$4F,$93,$DD,$78,$3D,$19
	.dc.b	$5C,$23,$F3,$7F,$43,$F9,$CF,$9C,$93,$F8,$E8,$CB,$F9,$08,$CB,$DA
	.dc.b	$3F,$93,$7C,$C3,$86,$92,$7F,$DB,$7F,$13,$5F,$F6,$AC,$ED,$FE,$B8
	.dc.b	$D7,$33,$FD,$EE,$31,$C4,$DB,$FD,$5F,$CC,$FA,$1F,$F1,$EE,$1F,$EA
	.dc.b	$48,$C7,$B9,$DE,$FE,$BF,$D0,$FE,$E3,$C5,$97,$FA,$A8,$FD,$D9,$87
	.dc.b	$E4,$C6,$1F,$F5,$78,$FA,$0E,$EF,$FB,$D1,$B5,$FE,$C5,$6E,$5F,$73
	.dc.b	$E8,$5F,$1D,$FD,$2E,$73,$6E,$95,$3E,$6B,$F8,$AB,$67,$C3,$5C,$29
	.dc.b	$F2,$EF,$A1,$FE,$36,$FF,$67,$CF,$FD,$E6,$DC,$FF,$1A,$EF,$0F,$E2
	.dc.b	$8D,$7F,$7E,$ED,$F2,$36,$FB,$DF,$63,$BF,$BF,$2B,$9C,$7F,$24,$8F
	.dc.b	$E4,$BE,$6C,$9F,$E2,$F1,$97,$F8,$CC,$65,$FC,$BC,$7F,$2C,$F4,$BF
	.dc.b	$96,$D4,$9F,$E3,$F1,$97,$F9,$0C,$65,$FC,$FC,$7E,$4B,$A2,$47,$D8
	.dc.b	$E5,$47,$F4,$0F,$98,$7C,$8E,$AB,$F2,$73,$F8,$F2,$3F,$94,$7F,$A3
	.dc.b	$FC,$AF,$F6,$AF,$6D,$F2,$1F,$B8,$8C,$3F,$C2,$71,$EF,$7F,$EE,$BF
	.dc.b	$A1,$FC,$0B,$FE,$54,$9F,$EC,$D1,$88,$5F,$47,$FA,$84,$66,$31,$BF
	.dc.b	$F5,$7F,$A0,$FE,$2D,$1B,$F4,$95,$CF,$8F,$F8,$D4,$63,$A3,$F8,$E4
	.dc.b	$62,$75,$8F,$8A,$72,$D3,$CC,$7F,$61,$FA,$1F,$F0,$A6,$1D,$DB,$9E
	.dc.b	$E7,$D2,$FF,$80,$63,$31,$C7,$D3,$F3,$CE,$5B,$1E,$EB,$B9,$2F,$FC
	.dc.b	$45,$97,$A9,$B5,$C4,$1F,$C0,$CC,$14,$F0,$1E,$A1,$AB,$4F,$F1,$36
	.dc.b	$7F,$3E,$73,$18,$CF,$9C,$E6,$B8,$7A,$9C,$DB,$7F,$00,$F1,$B8,$9F
	.dc.b	$8D,$A7,$BA,$7A,$D7,$1C,$3B,$47,$06,$E7,$F0,$6B,$86,$1C,$F1,$FC
	.dc.b	$65,$70,$BE,$67,$86,$F8,$3D,$70,$8F,$94,$DE,$3E,$47,$2F,$B6,$BA
	.dc.b	$61,$DE,$7E,$49,$C5,$C3,$CB,$0F,$1D,$FE,$1B,$0F,$0C,$EE,$F9,$92
	.dc.b	$75,$97,$E1,$27,$91,$F0,$63,$C7,$7F,$40,$EB,$1C,$E3,$2F,$93,$1F
	.dc.b	$44,$62,$39,$E1,$E2,$9C,$C7,$12,$3F,$05,$B5,$DA,$36,$64,$AE,$49
	.dc.b	$1E,$16,$BE,$A3,$C3,$7E,$DC,$7E,$39,$31,$FB,$B1,$97,$E2,$4F,$88
	.dc.b	$E8,$7C,$33,$07,$4C,$BF,$07,$1C,$30,$F2,$CF,$03,$C5,$7D,$F2,$F9
	.dc.b	$F7,$C1,$F8,$D1,$F7,$72,$F8,$D9,$7A,$A3,$85,$FE,$2F,$2F,$35,$F0
	.dc.b	$CB,$D3,$8F,$3A,$F2,$DF,$3B,$DC,$62,$B6,$F1,$BF,$B3,$8F,$82,$3D
	.dc.b	$D9,$7A,$9F,$71,$F9,$83,$FC,$61,$87,$EB,$9B,$7F,$E9,$D6,$DF,$FB
	.dc.b	$8F,$DE,$FF,$6D,$FD,$0D,$F7,$0D,$25,$7F,$84,$3D,$AF,$F2,$4B,$D9
	.dc.b	$AF,$F3,$96,$BF,$BC,$6B,$FC,$29,$FF,$28,$E5,$FC,$15,$BB,$7F,$A8
	.dc.b	$7E,$1B,$FD,$CD,$C1,$FC,$1E,$73,$7C,$FC,$4C,$1F,$A9,$AD,$C7,$EB
	.dc.b	$E7,$6F,$79,$DB,$FB,$F9,$DE,$1E,$19,$CB,$67,$5D,$3F,$C9,$63,$04
	.dc.b	$C7,$DD,$5E,$19,$D4,$62,$BC,$53,$D2,$3D,$93,$A9,$F5,$1A,$25,$7F
	.dc.b	$53,$BF,$2D,$73,$8F,$04,$F0,$75,$5C,$E3,$DC,$72,$9E,$07,$83,$1F
	.dc.b	$14,$71,$8C,$BF,$EB,$5B,$EE,$E4,$87,$4B,$EF,$5C,$A3,$15,$CB,$81
	.dc.b	$7C,$A7,$27,$2E,$06,$7F,$6A,$B7,$97,$A6,$73,$9F,$56,$DF,$D2,$CB
	.dc.b	$18,$7D,$C9,$CE,$7E,$4A,$F8,$07,$79,$ED,$97,$8E,$B7,$9F,$C0,$AF
	.dc.b	$2E,$7E,$2C,$7A,$E5,$F1,$A3,$13,$D0,$F5,$9C,$30,$E9,$8F,$8C,$95
	.dc.b	$F8,$3B,$E7,$5E,$4B,$E7,$97,$DA,$8E,$79,$79,$1F,$A5,$9F,$C3,$C7
	.dc.b	$9C,$71,$CB,$BE,$5D,$30,$F8,$07,$3C,$39,$57,$97,$0E,$55,$BC,$FF
	.dc.b	$05,$3C,$E7,$2F,$CD,$31,$9F,$A5,$FE,$C9,$F4,$2F,$A1,$FC,$C6,$BE
	.dc.b	$AC,$39,$E7,$E2,$7F,$C1,$70,$AE,$18,$73,$AF,$F1,$74,$63,$0F,$04
	.dc.b	$63,$0E,$BA,$7E,$4B,$0E,$51,$E0,$BE,$15,$FE,$3A,$F4,$9C,$B9,$A9
	.dc.b	$97,$F6,$BB,$DE,$7F,$EC,$4B,$DC,$E6,$BF,$51,$CD,$FE,$87,$8F,$F6
	.dc.b	$B6,$9F,$E2,$48,$C3,$86,$BF,$ED,$B6,$BF,$ED,$F7,$7E,$0F,$F8,$A5
	.dc.b	$B3,$FE,$D4,$ED,$7A,$FF,$E2,$FA,$D7,$8B,$6B,$93,$70,$F1,$37,$B1
	.dc.b	$87,$2A,$FF,$55,$E3,$DA,$FF,$D4,$59,$BE,$87,$82,$F6,$EC,$61,$B7
	.dc.b	$FE,$42,$CF,$FC,$85,$92,$F8,$4E,$4F,$95,$7C,$5C,$61,$E0,$F9,$FE
	.dc.b	$FA,$F8,$1B,$3A,$C3,$A4,$62,$98,$F8,$31,$4E,$63,$C7,$F4,$2F,$9A
	.dc.b	$F8,$4F,$FA,$43,$1E,$8E,$5B,$E2,$6A,$76,$D8,$C9,$C2,$32,$D3,$38
	.dc.b	$F9,$1F,$F3,$2B,$F8,$EE,$8F,$A8,$D8,$F6,$79,$4F,$A0,$CF,$3F,$F5
	.dc.b	$2C,$FF,$CD,$4D,$AE,$6D,$BF,$F6,$C5,$F1,$9F,$F7,$2C,$E7,$6F,$A5
	.dc.b	$F2,$1F,$6B,$DC,$FB,$6D,$AF,$99,$C0,$8C,$63,$E9,$3F,$06,$E3,$2F
	.dc.b	$CA,$67,$FE,$7D,$9F,$CC,$35,$FD,$03,$5F,$F7,$7B,$79,$3C,$D1,$D4
	.dc.b	$CE,$1F,$77,$97,$EC,$BF,$94,$AF,$8F,$5C,$2F,$C3,$3B,$E2,$CD,$7E
	.dc.b	$5B,$E7,$FF,$2F,$EE,$DA,$F6,$B7,$2F,$66,$DF,$95,$8C,$63,$F8,$97
	.dc.b	$EA,$F9,$2F,$A0,$A7,$E8,$BC,$5B,$8F,$E6,$31,$56,$E3,$0D,$8E,$FC
	.dc.b	$0A,$E3,$1F,$07,$E8,$1A,$7C,$76,$E1,$FC,$2A,$51,$87,$F8,$6B,$5F
	.dc.b	$C6,$9A,$9F,$C6,$C6,$F3,$FD,$27,$C4,$62,$65,$FA,$6B,$4F,$F1,$D9
	.dc.b	$CD,$FE,$07,$C2,$7C,$E3,$E3,$BF,$86,$46,$1F,$E0,$BE,$D6,$27,$58
	.dc.b	$7E,$F2,$B4,$D8,$CD,$FC,$4A,$E1,$7C,$7E,$82,$5F,$E5,$3B,$6D,$F8
	.dc.b	$F0,$56,$35,$FE,$01,$56,$FF,$A4,$D3,$E8,$3C,$87,$F1,$AF,$6A,$CF
	.dc.b	$8E,$2E,$3F,$5A,$6E,$1F,$C2,$DE,$15,$FA,$6E,$FB,$5B,$8F,$D4,$CF
	.dc.b	$F2,$69,$CD,$7F,$B8,$F5,$F9,$3F,$41,$F1,$4F,$E5,$7E,$3D,$C3,$E2
	.dc.b	$45,$C7,$F0,$F3,$6F,$96,$6F,$A0,$F9,$C7,$F3,$46,$E1,$F3,$62,$DF
	.dc.b	$E4,$26,$D7,$29,$B5,$FD,$6E,$6D,$31,$FB,$F5,$8E,$9F,$E7,$55,$A9
	.dc.b	$FC,$1F,$D0,$7D,$F3,$FA,$2F,$C9,$DC,$7E,$F4,$5C,$7F,$43,$37,$25
	.dc.b	$7F,$24,$F9,$F7,$CF,$7F,$07,$6C,$7E,$85,$B7,$FD,$B1,$AF,$E8,$1C
	.dc.b	$D7,$48,$08,$CC,$E3,$E5,$9F,$3F,$F9,$C8,$FD,$D4,$EA,$B8,$B5,$E5
	.dc.b	$87,$F6,$BB,$E3,$A7,$F3,$AA,$D7,$CE,$FF,$29,$C3,$F7,$0D,$C3,$FC
	.dc.b	$06,$D7,$F6,$ED,$7F,$A2,$BB,$CB,$D5,$08,$C4,$FB,$3E,$75,$F3,$CF
	.dc.b	$80,$FE,$97,$5A,$BF,$82,$D7,$97,$91,$FE,$03,$BE,$35,$C2,$26,$1F
	.dc.b	$C6,$EF,$95,$6A,$31,$3F,$D6,$3D,$57,$C7,$FF,$02,$F9,$E7,$C6,$66
	.dc.b	$30,$DB,$FE,$4D,$5C,$1F,$EF,$FE,$CB,$E7,$9F,$3D,$BD,$BF,$A2,$5E
	.dc.b	$DB,$8F,$F8,$6E,$F8,$56,$F8,$3E,$3F,$1B,$E7,$BF,$B6,$B7,$3E,$FE
	.dc.b	$BB,$E3,$B3,$5A,$F1,$65,$79,$3E,$9B,$FC,$2F,$AF,$F5,$5C,$7F,$8B
	.dc.b	$E5,$FD,$77,$DE,$FC,$6F,$C1,$BB,$7F,$C7,$F1,$86,$DA,$74,$FF,$36
	.dc.b	$9D,$3F,$96,$F8,$37,$97,$FE,$1A,$8E,$17,$FE,$08,$9C,$E1,$C2,$51
	.dc.b	$87,$67,$3A,$E5,$87,$82,$1F,$3C,$F9,$FF,$F0,$9E,$1F,$8E,$8C,$4F
	.dc.b	$F5,$B6,$BF,$DD,$7E,$95,$EC,$70,$DF,$FC,$37,$4F,$A0,$C5,$FE,$C2
	.dc.b	$30,$F3,$8C,$AB,$FE,$68,$FF,$69,$FA,$0F,$EC,$8D,$F5,$37,$18,$8F
	.dc.b	$F0,$36,$AF,$7F,$35,$27,$81,$FD,$CE,$74,$E0,$E3,$E1,$78,$A3,$F5
	.dc.b	$11,$9D,$7F,$5F,$F0,$FF,$E5,$D8,$FE,$F8,$D7,$FC,$2C,$D8,$FD,$AF
	.dc.b	$C7,$7F,$99,$67,$7F,$09,$EE,$AD,$6F,$FE,$40,$6C,$FF,$14,$AE,$0F
	.dc.b	$4A,$E0,$47,$FC,$19,$1A,$BF,$2C,$FF,$80,$E7,$FD,$C2,$DA,$F3,$47
	.dc.b	$F8,$09,$AF,$F9,$43,$E7,$DF,$25,$F4,$9B,$1F,$E1,$96,$BF,$E3,$3F
	.dc.b	$A0,$FC,$C3,$99,$E0,$47,$FC,$41,$1A,$7F,$C7,$53,$C9,$FF,$09,$B5
	.dc.b	$F1,$4F,$A6,$31,$1F,$D3,$FE,$83,$F8,$5B,$FE,$07,$6B,$FE,$11,$6E
	.dc.b	$3F,$E7,$DE,$6F,$C7,$FA,$09,$CE,$BF,$EC,$86,$DF,$5F,$0F,$F2,$BE
	.dc.b	$8F,$FA,$4F,$0E,$35,$FE,$C0,$AF,$07,$C2,$E0,$DF,$CD,$3E,$77,$13
	.dc.b	$A2,$1F,$E4,$A8,$D3,$8D,$3F,$C0,$17,$D6,$3D,$57,$D2,$F9,$C6,$0E
	.dc.b	$19,$FF,$58,$C9,$8F,$49,$F8,$5C,$3E,$82,$3F,$EA,$38,$FD,$BE,$5E
	.dc.b	$09,$CC,$F0,$35,$D9,$EC,$FA,$09,$1C,$5F,$7F,$D0,$62,$3D,$91,$6F
	.dc.b	$FD,$77,$82,$BE,$4F,$C6,$4B,$FB,$0E,$F3,$FA,$4E,$63,$ED,$39,$7F
	.dc.b	$01,$F1,$5F,$48,$8F,$AF,$E8,$2B,$F8,$18,$B8,$F8,$A7,$EC,$FC,$44
	.dc.b	$AF,$C4,$3F,$63,$E7,$31,$1F,$89,$8B,$5E,$A9,$FE,$27,$F4,$1F,$E6
	.dc.b	$88,$FE,$DE,$DC,$BF,$D6,$ED,$B7,$07,$D4,$F0,$B8,$C4,$3F,$D0,$53
	.dc.b	$B9,$D4,$60,$FD,$C4,$7E,$E5,$34,$C7,$D0,$3E,$7F,$F9,$3B,$67,$FD
	.dc.b	$1C,$D7,$F2,$1F,$41,$A9,$91,$A7,$F3,$31,$F9,$AF,$3F,$F3,$7F,$37
	.dc.b	$2B,$F1,$11,$97,$F1,$51,$97,$F3,$91,$FC,$E3,$E5,$DB,$52,$4F,$FA
	.dc.b	$DF,$E2,$AB,$FE,$BB,$9D,$BF,$EA,$56,$BF,$EB,$0F,$92,$4F,$08,$FF
	.dc.b	$4C,$CE,$9B,$5F,$C7,$E3,$11,$FA,$5E,$67,$CF,$FE,$DB,$C2,$93,$FA
	.dc.b	$38,$FD,$C9,$87,$F6,$D1,$87,$FD,$71,$F3,$EF,$8A,$FE,$AA,$D8,$CC
	.dc.b	$7F,$08,$F9,$5F,$EC,$31,$82,$1F,$ED,$99,$E0,$FF,$B3,$23,$6F,$FB
	.dc.b	$3B,$E7,$DF,$3F,$FC,$49,$FE,$A3,$19,$F3,$31,$32,$3F,$CB,$3D,$3F
	.dc.b	$E5,$BE,$75,$F0,$8C,$37,$A7,$F8,$43,$DA,$B8,$C6,$2F,$C1,$84,$9C
	.dc.b	$9F,$9C,$AE,$0F,$F1,$B8,$FE,$43,$E5,$C8,$FE,$1D,$1F,$C9,$0C,$3F
	.dc.b	$C8,$63,$0F,$F1,$3F,$3B,$F9,$4D,$C8,$C1,$0F,$E5,$0E,$5F,$E5,$51
	.dc.b	$F8,$FE,$69,$1F,$89,$7F,$9C,$F2,$BE,$3F,$33,$F8,$F9,$BA,$7C,$7A
	.dc.b	$D4,$F1,$3F,$DD,$5C,$8A,$FC,$61,$23,$FD,$D2,$7F,$A8,$7E,$7F,$EC
	.dc.b	$FA,$EF,$8D,$FF,$45,$F9,$A9,$CC,$61,$B5,$FC,$1B,$A1,$F3,$DF,$E9
	.dc.b	$28,$61,$CD,$ED,$EB,$E6,$31,$8F,$58,$C6,$3D,$5B,$5D,$6E,$D6,$E3
	.dc.b	$FD,$3F,$7C,$BE,$7F,$FA,$83,$7C,$DD,$E7,$E6,$B5,$6B,$2F,$6B,$70
	.dc.b	$FF,$61,$36,$7B,$47,$AA,$77,$7E,$2A,$D1,$BB,$EC,$F3,$A6,$BF,$13
	.dc.b	$CF,$F3,$E1,$F1,$63,$13,$D4,$F0,$E1,$F0,$1E,$39,$FF,$85,$33,$F9
	.dc.b	$51,$8C,$39,$BF,$EC,$0B,$E7,$3F,$06,$32,$6C,$EC,$57,$9C,$70,$C3
	.dc.b	$DB,$3C,$23,$0F,$5B,$F1,$4F,$A5,$B1,$E8,$38,$38,$7B,$46,$EF,$CF
	.dc.b	$9F,$D2,$C7,$DD,$87,$6B,$D9,$DE,$B7,$97,$CB,$9D,$13,$1F,$0E,$1C
	.dc.b	$EF,$C3,$71,$F7,$D7,$AA,$75,$18,$AF,$34,$F3,$AE,$F8,$79,$27,$24
	.dc.b	$8C,$D7,$5C,$BE,$4C,$E6,$3E,$D6,$1D,$A7,$35,$BC,$BE,$15,$6B,$2F
	.dc.b	$66,$1C,$6F,$73,$9C,$7C,$58,$71,$8F,$88,$64,$C1,$F2,$0F,$93,$8F
	.dc.b	$87,$1E,$F7,$D0,$8F,$DE,$9F,$85,$5C,$B2,$F8,$D9,$79,$23,$19,$7C
	.dc.b	$E8,$C5,$70,$26,$1F,$8F,$6E,$1E,$02,$3F,$85,$32,$7A,$A3,$C6,$E5
	.dc.b	$AF,$AF,$0F,$1E,$5F,$0A,$38,$9E,$EB,$F1,$D7,$0A,$DC,$FA,$31,$F9
	.dc.b	$19,$7C,$38,$F0,$63,$E4,$79,$63,$D7,$0E,$59,$7B,$E3,$C1,$3F,$DD
	.dc.b	$A3,$FD,$00,$FD,$A3,$18,$7F,$2A,$7F,$DA,$BE,$AB,$E7,$F1,$ED,$23
	.dc.b	$13,$FC,$C2,$31,$3A,$9C,$CF,$F8,$13,$D6,$7C,$17,$F3,$A6,$C7,$F3
	.dc.b	$8D,$91,$88,$E1,$3B,$3F,$75,$1C,$34,$FE,$5F,$F0,$7F,$EC,$BE,$DC
	.dc.b	$D4,$BF,$D3,$DE,$F7,$FE,$83,$7B,$9C,$D6,$F0,$FE,$13,$7B,$97,$02
	.dc.b	$F8,$56,$AF,$86,$1F,$C8,$70,$E1,$7B,$E9,$65,$AC,$BF,$8E,$57,$57
	.dc.b	$F9,$5B,$C3,$2E,$B8,$73,$C7,$F9,$1F,$73,$F0,$27,$CD,$F4,$C5,$D7
	.dc.b	$FE,$0D,$9C,$E7,$FE,$04,$BE,$1F,$07,$BA,$D6,$16,$BF,$D2,$D7,$C3
	.dc.b	$0E,$71,$89,$D4,$7F,$9A,$25,$78,$C3,$C9,$56,$5F,$37,$7E,$A7,$0D
	.dc.b	$C7,$C9,$3A,$6C,$E9,$A7,$F8,$73,$17,$CE,$3E,$57,$C1,$DD,$F8,$6A
	.dc.b	$D4,$E5,$FF,$55,$70,$FA,$2F,$8D,$FC,$77,$FB,$D1,$89,$FC,$F5,$71
	.dc.b	$8F,$F0,$4B,$5F,$F0,$07,$1B,$E1,$1F,$41,$EF,$F8,$8F,$8B,$39,$8C
	.dc.b	$61,$FE,$34,$8C,$3F,$02,$77,$5C,$A3,$06,$DC,$5F,$D2,$3F,$22,$E3
	.dc.b	$8B,$EF,$7A,$5C,$B1,$E7,$F5,$7F,$47,$E6,$EE,$3F,$39,$B8,$F3,$E4
	.dc.b	$FC,$54,$63,$1F,$C4,$36,$BF,$C4,$B9,$30,$FB,$9F,$37,$9C,$7C,$7F
	.dc.b	$55,$F7,$FE,$A3,$D6,$DC,$3D,$4D,$9F,$56,$FF,$26,$31,$9F,$E6,$63
	.dc.b	$15,$C0,$FC,$2F,$CE,$74,$7E,$26,$1F,$8D,$E2,$F9,$2D,$8C,$9F,$8C
	.dc.b	$8C,$B4,$CE,$3E,$2D,$15,$F9,$7F,$23,$73,$9E,$BF,$CD,$7C,$E7,$D6
	.dc.b	$6E,$1F,$55,$B3,$B8,$FC,$DC,$ED,$D1,$F9,$E9,$C4,$FD,$1E,$93,$FA
	.dc.b	$73,$F1,$3F,$81,$BE,$17,$17,$F8,$98,$C3,$FA,$89,$DC,$E9,$FB,$9F
	.dc.b	$10,$33,$3F,$C0,$5C,$3C,$E7,$F7,$10,$F2,$A9,$8F,$94,$B5,$E1,$F9
	.dc.b	$6C,$34,$E9,$1C,$BE,$4E,$47,$7C,$2E,$3F,$99,$F3,$1F,$14,$FA,$DF
	.dc.b	$E1,$03,$D7,$FE,$1B,$3A,$8C,$C6,$27,$86,$1F,$C6,$E7,$83,$FE,$A0
	.dc.b	$AE,$0E,$6B,$84,$FE,$FE,$FE,$7D,$69,$FE,$02,$D7,$FC,$F8,$D7,$EA
	.dc.b	$F6,$FC,$BF,$A3,$F8,$FF,$13,$F2,$1B,$8F,$CD,$6D,$7D,$26,$F5,$FF
	.dc.b	$AF,$F8,$5C,$1F,$43,$E2,$7E,$8B,$70,$FA,$CD,$C3,$ED,$B6,$F9,$7D
	.dc.b	$17,$D8,$F8,$9F,$B2,$DB,$FB,$CD,$BF,$C1,$B6,$B8,$CF,$E0,$EB,$8B
	.dc.b	$4F,$E2,$DF,$82,$3E,$33,$F0,$BE,$A7,$F1,$C8,$C4,$FE,$2E,$31,$3F
	.dc.b	$37,$1F,$EC,$B5,$A7,$F9,$6F,$82,$8C,$47,$F2,$A6,$BF,$3B,$5F,$F6
	.dc.b	$16,$1F,$97,$F8,$6F,$CA,$FB,$BF,$96,$6C,$64,$FC,$E4,$65,$FE,$8D
	.dc.b	$DE,$F1,$3F,$1B,$EA,$7F,$41,$8C,$4F,$E2,$63,$13,$A8,$E2,$DC,$7F
	.dc.b	$98,$0A,$E2,$FF,$B3,$1A,$FF,$4A,$6B,$FE,$C5,$6A,$E5,$FD,$A7,$C1
	.dc.b	$FE,$C3,$E8,$BF,$62,$DC,$3F,$6C,$DA,$FD,$DB,$78,$BF,$93,$6C,$F8
	.dc.b	$6F,$DC,$FA,$DF,$55,$B8,$7D,$A7,$EA,$70,$FD,$16,$D7,$19,$FF,$72
	.dc.b	$CE,$5F,$F1,$F6,$8F,$3F,$F8,$2F,$91,$F2,$5B,$3F,$39,$BA,$7C,$6A
	.dc.b	$E1,$3F,$C3,$EB,$83,$D2,$BF,$0D,$5F,$2D,$B5,$C2,$3F,$B9,$FB,$5F
	.dc.b	$E2,$EF,$5F,$FC,$19,$F8,$2F,$F0,$8C,$3F,$93,$ED,$FE,$1F,$AD,$47
	.dc.b	$F8,$76,$75,$7F,$D9,$E7,$2F,$4C,$FF,$C2,$F5,$FE,$00,$8F,$F2,$D4
	.dc.b	$61,$D6,$1F,$9B,$F9,$8F,$F1,$6F,$B3,$E2,$E9,$FF,$10,$C6,$30,$F0
	.dc.b	$46,$27,$FC,$51,$7F,$CF,$05,$6D,$FF,$07,$B5,$D3,$5F,$E8,$6F,$F8
	.dc.b	$F3,$D9,$FF,$0E,$FD,$67,$56,$E1,$F0,$8D,$DF,$F8,$A6,$FA,$B5,$EA
	.dc.b	$D7,$73,$FB,$7E,$66,$AF,$8A,$F9,$E7,$D0,$F9,$0B,$8F,$CD,$8B,$87
	.dc.b	$F8,$26,$95,$8E,$0F,$F2,$96,$9F,$E5,$4F,$53,$FC,$6B,$F3,$7F,$15
	.dc.b	$B9,$7A,$71,$F8,$BA,$A7,$E8,$36,$7E,$B5,$F3,$6C,$66,$BE,$D7,$D8
	.dc.b	$7E,$6D,$FC,$07,$CC,$7F,$8D,$6B,$FC,$91,$39,$CB,$F1,$13,$9A,$FD
	.dc.b	$EC,$F1,$7F,$C8,$6D,$7E,$D3,$5F,$EE,$ED,$4F,$5A,$DF,$4F,$F9,$9B
	.dc.b	$97,$FD,$4D,$D5,$FE,$51,$AD,$5F,$2A,$C1,$BA,$56,$2F,$FC,$C1,$F4
	.dc.b	$4F,$A4,$FF,$38,$E6,$FA,$DF,$E5,$CF,$F9,$6E,$75,$5F,$36,$35,$CF
	.dc.b	$FE,$61,$9C,$E7,$FA,$16,$BF,$E6,$26,$BF,$99,$6B,$FE,$A0,$E5,$7C
	.dc.b	$77,$ED,$3E,$4F,$FC,$BF,$18,$C7,$F6,$51,$89,$D4,$7F,$98,$67,$55
	.dc.b	$F8,$9A,$D4,$FF,$12,$9C,$CF,$F9,$B9,$AF,$F8,$9B,$AF,$F9,$55,$FF
	.dc.b	$1C,$9F,$F4,$B6,$5F,$EA,$17,$F3,$7F,$27,$E4,$E3,$FE,$12,$DC,$7F
	.dc.b	$83,$37,$1F,$CD,$36,$FF,$CA,$57,$D2,$BF,$C0,$3D,$0F,$93,$D4,$78
	.dc.b	$E7,$58,$7C,$5C,$7F,$8B,$D6,$A3,$E4,$7C,$07,$F1,$AE,$4F,$8E,$DC
	.dc.b	$3E,$23,$67,$FC,$CB,$82,$31,$97,$91,$E3,$5F,$C7,$FE,$46,$DF,$0F
	.dc.b	$80,$FF,$51,$CF,$D4,$6B,$F4,$7E,$03,$FD,$69,$F0,$5F,$61,$B1,$FE
	.dc.b	$AD,$AF,$F2,$9F,$C4,$FF,$89,$BC,$6F,$E4,$CD,$8F,$F5,$3B,$5F,$E3
	.dc.b	$CD,$F8,$B7,$6F,$FB,$4A,$5E,$BF,$FA,$9D,$FF,$2F,$CE,$B6,$E1,$5B
	.dc.b	$AD,$52,$B4,$F2,$C3,$E8,$78,$FF,$CB,$70,$FF,$6F,$7B,$17,$3F,$F6
	.dc.b	$CB,$6F,$F3,$15,$7A,$3F,$B9,$7C,$16,$78,$FF,$DD,$5D,$D7,$E0,$7F
	.dc.b	$5D,$87,$E5,$EB,$87,$C1,$7F,$9A,$E3,$FD,$74,$D7,$FD,$36,$DD,$3F
	.dc.b	$25,$D5,$FE,$5B,$9D,$FA,$9D,$E1,$F2,$34,$EB,$12,$7F,$AD,$46,$5E
	.dc.b	$27,$83,$DB,$F1,$91,$FE,$C1,$EA,$57,$C7,$17,$0F,$1C,$DE,$1F,$CD
	.dc.b	$FC,$6F,$8C,$C5,$B8,$1A,$31,$D4,$FF,$6D,$AD,$D6,$BD,$AF,$ED,$53
	.dc.b	$C6,$D1,$89,$D4,$FF,$83,$A7,$2D,$9F,$89,$E8,$48,$F9,$B1,$F4,$0C
	.dc.b	$3F,$D3,$E3,$0F,$F4,$91,$F3,$5F,$B5,$7C,$6D,$7F,$BF,$35,$FF,$7F
	.dc.b	$B5,$FE,$EC,$EF,$46,$7F,$DE,$7E,$6D,$F4,$BF,$8B,$B4,$61,$F1,$5F
	.dc.b	$1A,$FF,$00,$7E,$05,$FE,$9D,$9F,$33,$5F,$F2,$5B,$70,$FF,$0D,$66
	.dc.b	$F8,$DE,$D3,$9B,$E7,$D3,$83,$F3,$9E,$77,$D6,$AC,$63,$F4,$AF,$B3
	.dc.b	$FE,$11,$0E,$1E,$66,$EB,$9F,$C6,$3E,$6E,$9D,$25,$9F,$F7,$CF,$25
	.dc.b	$E2,$FE,$03,$D2,$46,$2B,$FB,$AC,$E6,$31,$1F,$E1,$48,$FD,$77,$99
	.dc.b	$23,$FC,$05,$5C,$A7,$F6,$9F,$83,$BC,$8A,$4A,$FF,$1F,$70,$57,$CA
	.dc.b	$FF,$AD,$A7,$84,$7F,$83,$7C,$14,$62,$3F,$C3,$13,$FC,$17,$0F,$4C
	.dc.b	$79,$C8,$7F,$83,$8A,$7F,$BD,$FE,$B3,$F8,$31,$C7,$3F,$9D,$A7,$F8
	.dc.b	$83,$F0,$7F,$F7,$55,$7F,$86,$1A,$FF,$90,$1A,$FF,$90,$1A,$FF,$90
	.dc.b	$1D,$C7,$F5,$BC,$7F,$89,$CF,$B8,$FF,$45,$38,$8F,$07,$C6,$FE,$13
	.dc.b	$E9,$FF,$71,$5F,$E5,$DE,$54,$62,$3F,$CD,$AD,$7F,$CA,$D1,$AE,$EF
	.dc.b	$DD,$B6,$3E,$56,$0C,$FE,$A9,$DE,$BF,$49,$F2,$CF,$A0,$FE,$07,$68
	.dc.b	$C3,$FC,$19,$AF,$F9,$C5,$AF,$E5,$DC,$E7,$F0,$7D,$FF,$C1,$9D,$AA
	.dc.b	$57,$92,$32,$FF,$03,$AE,$3C,$1F,$E0,$7F,$C2,$3F,$39,$3F,$E9,$E6
	.dc.b	$BF,$ED,$46,$BF,$E7,$7F,$05,$18,$8E,$5C,$FF,$94,$F9,$79,$1F,$E5
	.dc.b	$9A,$DC,$FE,$63,$E1,$BF,$33,$F0,$52,$BE,$4C,$65,$F9,$71,$97,$EE
	.dc.b	$73,$FE,$5A,$7E,$5F,$B7,$F2,$B4,$FF,$79,$57,$C9,$8C,$11,$FC,$F6
	.dc.b	$CC,$FF,$AE,$36,$3E,$24,$F4,$77,$3A,$30,$F8,$B8,$9D,$9F,$B2,$F8
	.dc.b	$39,$3F,$82,$8F,$DA,$98,$78,$46,$1F,$89,$BB,$D6,$FE,$84,$D8,$FA
	.dc.b	$51,$82,$3F,$8E,$F8,$0F,$E0,$27,$F7,$79,$E1,$1E,$5E,$EF,$E0,$BF
	.dc.b	$29,$23,$F2,$B1,$FC,$20,$C3,$EB,$8C,$3A,$3F,$36,$EB,$9B,$F5,$D3
	.dc.b	$FC,$43,$D2,$FE,$23,$EC,$4B,$FD,$0C,$65,$FD,$1C,$65,$F9,$11,$FC
	.dc.b	$63,$C6,$FE,$33,$E8,$C9,$FD,$4C,$65,$FD,$5C,$65,$FA,$11,$FE,$17
	.dc.b	$F5,$BF,$8B,$EF,$FD,$B1,$B3,$FE,$D7,$6B,$FC,$F5,$AF,$FB,$3D,$CE
	.dc.b	$5F,$E6,$3E,$37,$2F,$E1,$0E,$DC,$5F,$7A,$FF,$D6,$57,$C3,$A9,$0E
	.dc.b	$7F,$DE,$5F,$18,$DD,$F1,$7F,$33,$F8,$57,$F8,$86,$F8,$DA,$31,$3C
	.dc.b	$DB,$1F,$CD,$23,$FC,$69,$EC,$BE,$5B,$FC,$69,$4F,$95,$A6,$A3,$3E
	.dc.b	$24,$BF,$E7,$D5,$FC,$A3,$1F,$0B,$C6,$33,$F2,$9D,$B4,$91,$83,$F7
	.dc.b	$59,$7C,$67,$9F,$CA,$F8,$8D,$BE,$E7,$0E,$8E,$33,$EA,$E7,$91,$D1
	.dc.b	$F5,$FC,$9E,$23,$E3,$F8,$9E,$E3,$1C,$7F,$E1,$08,$FF,$19,$62,$BD
	.dc.b	$47,$BA,$B7,$A7,$FA,$8E,$BF,$C6,$F3,$EA,$E5,$F9,$67,$CA,$F9,$3B
	.dc.b	$1F,$1B,$E4,$FE,$69,$87,$FA,$67,$8A,$BE,$71,$F7,$27,$6F,$53,$F9
	.dc.b	$07,$E1,$4E,$0F,$F3,$63,$63,$31,$F6,$BE,$53,$2F,$DC,$F6,$24,$E8
	.dc.b	$87,$E2,$9C,$BE,$4A,$E8,$DB,$EE,$47,$F0,$3F,$29,$5F,$C9,$FC,$1F
	.dc.b	$E2,$8C,$70,$FC,$DD,$18,$30,$D5,$FE,$2A,$F7,$BF,$FB,$3E,$9F,$29
	.dc.b	$EE,$25,$7E,$41,$AF,$E8,$BE,$53,$F3,$26,$32,$F0,$57,$29,$FC,$BC
	.dc.b	$E9,$F9,$75,$FE,$4D,$F9,$8F,$F6,$04,$7F,$9B,$1A,$FF,$8E,$C3,$87
	.dc.b	$FC,$7C,$D7,$F7,$26,$BE,$1F,$FD,$39,$D7,$23,$FC,$D1,$3D,$31,$FF
	.dc.b	$74,$75,$3E,$3B,$F6,$3F,$29,$2F,$F5,$B1,$FB,$43,$0F,$F0,$08,$C3
	.dc.b	$FB,$2F,$88,$FD,$C9,$1F,$DA,$FC,$05,$7F,$83,$C5,$C3,$C3,$3F,$E0
	.dc.b	$AF,$82,$FF,$63,$D7,$E4,$F0,$F5,$FA,$29,$CF,$4F,$9E,$BF,$8B,$CE
	.dc.b	$7C,$5F,$F6,$CD,$FF,$1A,$E6,$7C,$DF,$F8,$CB,$0E,$18,$7F,$85,$71
	.dc.b	$E7,$7D,$9B,$1D,$1B,$7C,$BC,$3F,$F6,$FD,$6F,$1E,$F7,$B6,$BD,$5C
	.dc.b	$D7,$18,$CE,$5E,$77,$DC,$71,$29,$C4,$CC,$71,$73,$1C,$71,$F5,$B6
	.dc.b	$FC,$93,$C3,$2F,$81,$18,$8F,$F1,$76,$5F,$39,$B8,$FF,$8D,$63,$11
	.dc.b	$F8,$AF,$8A,$75,$3D,$8F,$F5,$FC,$61,$B1,$99,$FC,$B4,$E5,$FF,$1F
	.dc.b	$7C,$0F,$F0,$C7,$ED,$B5,$FB,$2D,$7F,$C9,$0D,$CF,$FD,$49,$E0,$FC
	.dc.b	$88,$FE,$3F,$AF,$E0,$76,$7C,$DB,$E1,$BF,$95,$4F,$C4,$AD,$EF,$E0
	.dc.b	$C3,$91,$F5,$2F,$87,$6E,$33,$F9,$B7,$C1,$F2,$D6,$F1,$EB,$87,$39
	.dc.b	$DC,$F1,$C7,$C9,$8F,$49,$C9,$23,$C9,$3C,$F1,$E1,$7F,$DA,$3E,$2B
	.dc.b	$FD,$57,$1F,$E9,$96,$DF,$F9,$19,$7B,$7F,$D1,$F9,$BF,$A4,$79,$89
	.dc.b	$DC,$61,$B7,$D7,$DF,$E0,$F7,$BB,$87,$F7,$FA,$DB,$7D,$BC,$E3,$D7
	.dc.b	$DA,$B4,$9B,$18,$6E,$1E,$96,$DF,$FA,$64,$FE,$33,$EA,$FF,$5F,$7F
	.dc.b	$8C,$B5,$FE,$33,$DE,$8C,$47,$F8,$52,$31,$5F,$C9,$67,$D3,$3D,$23
	.dc.b	$E4,$79,$D8,$AD,$CF,$F5,$EE,$EF,$F6,$65,$7C,$DF,$6B,$E0,$F3,$27
	.dc.b	$35,$F3,$E7,$31,$88,$FF,$43,$35,$FE,$1C,$D5,$3C,$FE,$07,$D2,$DB
	.dc.b	$FF,$44,$57,$6E,$1F,$AD,$EA,$F6,$BE,$FE,$27,$CE,$F1,$6E,$1F,$83
	.dc.b	$6E,$5F,$EF,$DC,$BD,$15,$DB,$2F,$F1,$55,$FF,$47,$9C,$B6,$33,$5F
	.dc.b	$0E,$BE,$DF,$B5,$D3,$3F,$EC,$D5,$FE,$48,$BF,$C9,$7A,$3F,$0B,$7F
	.dc.b	$C0,$B6,$3A,$47,$84,$8E,$0F,$7B,$F8,$8F,$67,$86,$5F,$4E,$BF,$17
	.dc.b	$C3,$F3,$9B,$7F,$19,$F6,$11,$F5,$36,$79,$47,$E9,$FD,$9E,$4F,$CE
	.dc.b	$EA,$FC,$BF,$B3,$F6,$9B,$19,$3E,$23,$F2,$48,$E0,$D5,$7E,$C6,$39
	.dc.b	$47,$ED,$63,$F4,$67,$E2,$BE,$6B,$FC,$49,$19,$3F,$D8,$CE,$23,$EC
	.dc.b	$35,$FF,$68,$35,$FE,$B0,$D7,$FC,$E5,$F1,$43,$B7,$A7,$FB,$47,$15
	.dc.b	$DA,$31,$E5,$48,$FF,$3D,$F0,$7F,$06,$AF,$F3,$64,$60,$FF,$3B,$A9
	.dc.b	$F5,$7A,$5F,$98,$C7,$FC,$4D,$5F,$8B,$6E,$3E,$A9,$FD,$AF,$6F,$97
	.dc.b	$D8,$F2,$1E,$5C,$EE,$1F,$E8,$EF,$9A,$7C,$37,$F0,$C1,$F3,$5F,$C4
	.dc.b	$BC,$E4,$EA,$BC,$D9,$FF,$A4,$EF,$71,$CD,$FE,$0B,$5C,$1B,$5C,$6B
	.dc.b	$FD,$4C,$D7,$FD,$4C,$E7,$7F,$F3,$85,$71,$BE,$F9,$F9,$BE,$33,$C9
	.dc.b	$7B,$9F,$8F,$ED,$FD,$1F,$73,$E9,$36,$38,$F1,$23,$19,$FF,$BF,$A3
	.dc.b	$13,$AA,$FB,$5C,$1F,$6F,$DC,$B3,$C9,$B7,$C5,$BA,$FD,$B7,$F0,$5E
	.dc.b	$DF,$E0,$FD,$CB,$7E,$26,$DF,$85,$B7,$F8,$36,$CF,$E0,$9B,$3F,$90
	.dc.b	$7F,$92,$FC,$CB,$E0,$FF,$21,$EA,$61,$B7,$F5,$BC,$35,$6E,$30,$DC
	.dc.b	$3E,$17,$BD,$C6,$7E,$3F,$BB,$F9,$BF,$77,$F3,$8D,$8C,$9F,$32,$32
	.dc.b	$D3,$F0,$B4,$EF,$FE,$30,$B5,$FE,$CB,$D1,$FD,$CD,$FE,$A1,$3A,$73
	.dc.b	$EA,$7F,$4A,$8E,$35,$FE,$6A,$3A,$78,$9F,$D4,$9B,$87,$F9,$CD,$A7
	.dc.b	$F9,$D6,$7A,$47,$F9,$C9,$B1,$F9,$9F,$67,$B5,$7E,$E7,$DC,$FD,$D7
	.dc.b	$CD,$5C,$3F,$12,$DB,$FC,$43,$72,$FD,$D3,$F5,$3E,$47,$F8,$17,$AD
	.dc.b	$6F,$F2,$2D,$BF,$F5,$EF,$0A,$BF,$84,$7B,$D9,$C7,$E1,$FA,$3F,$C2
	.dc.b	$FC,$CF,$84,$DC,$BE,$6E,$E8,$F8,$4D,$8F,$84,$DC,$BE,$14,$E8,$E7
	.dc.b	$3F,$A4,$F0,$AC,$FD,$58,$3F,$E4,$DE,$6F,$F0,$EC,$7E,$CB,$E4,$3F
	.dc.b	$81,$BF,$C0,$BD,$EF,$ED,$BE,$C7,$F8,$52,$73,$7C,$CE,$51,$FE,$1C
	.dc.b	$9D,$D7,$3F,$84,$C6,$1D,$AF,$87,$C5,$BD,$BF,$F1,$0F,$90,$CF,$F3
	.dc.b	$1C,$1F,$EF,$5A,$D5,$78,$7B,$BF,$C5,$31,$8C,$FF,$8B,$B7,$1F,$26
	.dc.b	$3C,$FE,$23,$78,$7F,$8B,$AB,$C1,$ED,$7F,$87,$9B,$87,$FA,$FD,$B8
	.dc.b	$7E,$4D,$AF,$FA,$F9,$CE,$5E,$5B,$FF,$1A,$57,$C9,$F9,$AE,$F8,$7C
	.dc.b	$76,$47,$CD,$F9,$AB,$8F,$9B,$55,$63,$1F,$F2,$6C,$2B,$17,$F6,$32
	.dc.b	$F9,$1E,$BF,$A7,$33,$49,$3E,$B8,$CB,$F1,$9B,$97,$87,$E8,$FE,$2D
	.dc.b	$7F,$8F,$3E,$6B,$F3,$17,$FA,$F6,$D7,$FB,$8D,$AF,$ED,$DA,$FF,$B8
	.dc.b	$9D,$E7,$FE,$4F,$6C,$FE,$43,$8F,$FD,$39,$D9,$DF,$AE,$4F,$F0,$CA
	.dc.b	$DE,$1F,$EA,$1B,$FF,$73,$4E,$6B,$73,$FE,$5D,$AD,$D7,$4A,$D1,$FC
	.dc.b	$5B,$E6,$5F,$2D,$FC,$2D,$B1,$FC,$2D,$B7,$FE,$FF,$6B,$FC,$4D,$CE
	.dc.b	$5F,$E6,$96,$CF,$D3,$FA,$D7,$77,$F9,$92,$73,$7F,$E5,$4E,$67,$C9
	.dc.b	$FE,$A7,$E6,$BF,$CC,$31,$8C,$7F,$61,$18,$9F,$F2,$CC,$E6,$BF,$CA
	.dc.b	$0D,$7F,$72,$75,$8D,$9C,$6B,$F7,$5F,$3F,$67,$F6,$DD,$6B,$FD,$7F
	.dc.b	$3B,$E6,$FC,$7C,$DF,$BD,$6E,$3F,$E5,$BA,$FF,$51,$E1,$E0,$C3,$95
	.dc.b	$F9,$89,$5F,$D4,$EB,$AF,$CD,$F2,$BF,$0C,$7F,$A7,$AF,$8D,$7F,$0F
	.dc.b	$AD,$70,$23,$F8,$AC,$F9,$FE,$0F,$F8,$87,$2F,$C8,$6C,$F7,$7E,$39
	.dc.b	$1F,$43,$71,$F0,$BF,$C6,$7E,$53,$94,$F9,$FE,$0F,$E7,$91,$F8,$0F
	.dc.b	$D2,$F8,$39,$1F,$0D,$B7,$FD,$4C,$7C,$57,$F2,$CF,$17,$FD,$39,$1F
	.dc.b	$D6,$E3,$1E,$47,$F5,$D8,$D8,$9D,$BF,$72,$76,$FF,$36,$F8,$E7,$A7
	.dc.b	$FD,$B6,$38,$5F,$E0,$A7,$3B,$F0,$94,$61,$D9,$CE,$B9,$61,$E6,$DF
	.dc.b	$F0,$9E,$57,$F6,$D6,$D7,$E0,$63,$1C,$5F,$DF,$E3,$F0,$9F,$09,$82
	.dc.b	$38,$3F,$D4,$5F,$09,$FE,$69,$7F,$C2,$CD,$7F,$26,$DC,$FF,$8F,$FC
	.dc.b	$07,$E7,$1F,$C1,$4F,$FA,$F2,$78,$3F,$91,$6B,$F8,$66,$BB,$F7,$7D
	.dc.b	$51,$D7,$87,$FA,$FE,$3F,$D3,$7E,$03,$B4,$63,$A9,$18,$C3,$91,$E8
	.dc.b	$F7,$F1,$B3,$1F,$F2,$74,$E4,$FE,$D1,$EB,$6A,$33,$1E,$EA,$C1,$D7
	.dc.b	$2F,$EB,$D5,$89,$F8,$F3,$E8,$23,$83,$FB,$A7,$AD,$C7,$23,$C6,$3E
	.dc.b	$5F,$B9,$E1,$23,$FD,$EF,$E3,$6B,$FE,$D9,$C9,$3F,$18,$91,$F6,$1E
	.dc.b	$DF,$37,$23,$EC,$C7,$C5,$5F,$34,$EF,$C3,$63,$FC,$8A,$D7,$FC,$0C
	.dc.b	$3E,$6F,$F4,$0E,$A7,$F5,$53,$B8,$FF,$21,$35,$FE,$3E,$D7,$FC,$33
	.dc.b	$F3,$7F,$E9,$C7,$FC,$A6,$D7,$BB,$79,$3F,$17,$E5,$FD,$9D,$3D,$53
	.dc.b	$9E,$4F,$E5,$6D,$BE,$BE,$EF,$E0,$EF,$84,$76,$8E,$24,$7F,$C4,$DB
	.dc.b	$30,$C4,$7E,$FE,$FC,$8F,$F9,$75,$AF,$82,$BF,$CB,$9E,$8F,$F8,$8F
	.dc.b	$C1,$FC,$93,$67,$FB,$C4,$7E,$07,$E7,$73,$E3,$23,$8B,$67,$FC,$01
	.dc.b	$18,$3F,$CD,$1C,$EF,$98,$7C,$77,$E4,$DB,$5F,$EB,$26,$BF,$E1,$5D
	.dc.b	$F5,$32,$34,$FE,$4E,$3C,$71,$0F,$F0,$71,$B3,$FC,$7F,$87,$F2,$7C
	.dc.b	$BF,$C8,$11,$E8,$21,$FE,$31,$F9,$CF,$F7,$23,$FE,$3C,$6B,$FE,$48
	.dc.b	$6F,$0F,$F0,$6C,$BD,$9A,$FC,$89,$CE,$3F,$C0,$33,$FF,$00,$E1,$ED
	.dc.b	$F4,$75,$F1,$BF,$77,$E7,$3F,$6F,$9F,$E1,$9B,$87,$F8,$5D,$B7,$FD
	.dc.b	$A7,$0F,$EF,$6F,$F9,$96,$33,$9F,$CD,$F9,$CF,$98,$75,$ED,$FD,$47
	.dc.b	$CB,$BD,$FA,$4C,$3E,$87,$C9,$57,$FA,$34,$ED,$FF,$50,$35,$CC,$FF
	.dc.b	$51,$8C,$69,$F0,$F3,$7C,$0F,$DD,$3A,$47,$F2,$CF,$0D,$F3,$8F,$81
	.dc.b	$FE,$3A,$DC,$BF,$D9,$8D,$7F,$CD,$78,$7D,$87,$FC,$17,$19,$BF,$07
	.dc.b	$CB,$7E,$2C,$E7,$7F,$8D,$F8,$8F,$C7,$7C,$1C,$C3,$F1,$D1,$97,$F2
	.dc.b	$11,$97,$BC,$7C,$8F,$84,$91,$FE,$D0,$D7,$FD,$C8,$FF,$A6,$B1,$FC
	.dc.b	$16,$DE,$57,$FA,$4C,$27,$6F,$CC,$AF,$90,$47,$E6,$BF,$CD,$69,$95
	.dc.b	$3C,$33,$F0,$BE,$57,$F5,$1F,$0D,$23,$F5,$11,$FA,$B3,$0F,$C8,$8C
	.dc.b	$3F,$33,$D7,$F9,$E4,$7F,$61,$EC,$57,$E7,$FC,$1F,$ED,$4C,$69,$F7
	.dc.b	$7C,$1F,$DC,$7C,$AC,$AF,$DC,$47,$EE,$CC,$3F,$62,$30,$E8,$FD,$F3
	.dc.b	$A9,$E9,$1F,$C0,$7D,$4F,$E0,$9E,$D4,$8C,$10,$FC,$1B,$97,$F0,$B1
	.dc.b	$FC,$2F,$C8,$FE,$19,$EA,$C9,$FE,$19,$19,$7F,$87,$46,$5F,$C6,$C7
	.dc.b	$F7,$1F,$5B,$F8,$B6,$F7,$A0,$D3,$EB,$73,$5B,$FF,$16,$57,$06,$D7
	.dc.b	$06,$D7,$F8,$C5,$FF,$18,$46,$5F,$F6,$3F,$1B,$9B,$F0,$47,$58,$FF
	.dc.b	$65,$FC,$D7,$F8,$2D,$FF,$35,$37,$BC,$DB,$FF,$3A,$CF,$F9,$C3,$DB
	.dc.b	$77,$7F,$8B,$5B,$1F,$E4,$3E,$DF,$F2,$4C,$E4,$FE,$6F,$7F,$C6,$A7
	.dc.b	$BF,$87,$FE,$09,$8C,$63,$E0,$F9,$CC,$7C,$AF,$F8,$5B,$2F,$56,$1F
	.dc.b	$A5,$F1,$7F,$CD,$71,$88,$F5,$CE,$F5,$F2,$12,$3C,$93,$B7,$FD,$1F
	.dc.b	$87,$BB,$E5,$BC,$44,$8E,$C3,$5F,$F3,$1D,$F8,$FA,$24,$EA,$7F,$CC
	.dc.b	$F1,$EB,$F9,$5C,$CF,$C8,$9C,$BE,$29,$F9,$04,$7B,$E1,$F4,$3C,$5F
	.dc.b	$88,$72,$EA,$FE,$3D,$9F,$F8,$DE,$BC,$0D,$7E,$A7,$3F,$CB,$F9,$4F
	.dc.b	$90,$FD,$52,$47,$C2,$6B,$F8,$2F,$95,$FA,$47,$07,$D9,$F3,$8E,$0F
	.dc.b	$F2,$3B,$63,$31,$F6,$7E,$5B,$2F,$F9,$BE,$75,$3F,$05,$AF,$FA,$BD
	.dc.b	$AF,$FA,$91,$DE,$5F,$D8,$79,$3F,$C7,$73,$F1,$1A,$FF,$B0,$27,$7C
	.dc.b	$7C,$71,$F2,$92,$78,$7C,$B5,$7C,$51,$E5,$23,$F9,$7A,$47,$E4,$FE
	.dc.b	$0A,$4F,$C2,$8F,$83,$F2,$B2,$3E,$2B,$F9,$93,$0F,$BD,$F8,$44,$7B
	.dc.b	$3F,$9F,$F9,$69,$1C,$8F,$C7,$8F,$99,$78,$DF,$E5,$56,$C6,$9A,$FE
	.dc.b	$EB,$E5,$B5,$3F,$EB,$99,$CC,$FE,$E1,$AF,$FB,$B9,$AF,$F0,$0F,$92
	.dc.b	$FD,$97,$CB,$C8,$FA,$11,$FB,$53,$0F,$F0,$88,$C3,$F5,$7E,$2B,$F7
	.dc.b	$44,$7E,$C7,$C8,$D7,$F8,$7C,$5E,$2F,$BD,$3F,$E0,$AF,$85,$FF,$34
	.dc.b	$F4,$31,$61,$3D,$1F,$F4,$2F,$8F,$8A,$D4,$E6,$F8,$DF,$FB,$BB,$E7
	.dc.b	$3F,$C5,$B5,$AA,$FE,$D1,$5A,$78,$5F,$3B,$EF,$1A,$8C,$3F,$C0,$6F
	.dc.b	$9C,$EB,$0F,$1C,$FF,$B9,$A3,$17,$D6,$39,$19,$39,$78,$9F,$EE,$39
	.dc.b	$DE,$1D,$A7,$78,$F9,$F0,$EA,$FB,$BC,$06,$5E,$2A,$F4,$9D,$4E,$38
	.dc.b	$74,$7C,$13,$C3,$1F,$04,$70,$3F,$9D,$E1,$EA,$6E,$5F,$E9,$A6,$E5
	.dc.b	$FD,$37,$E3,$1D,$0F,$82,$C3,$6B,$FA,$29,$FE,$F6,$70,$FF,$48,$9E
	.dc.b	$33,$FE,$40,$F8,$2F,$E3,$13,$F2,$DA,$FD,$F6,$BF,$E1,$F6,$E3,$FE
	.dc.b	$93,$EF,$F9,$0D,$9F,$7F,$CE,$3E,$2B,$FD,$DF,$C3,$F2,$F0,$FE,$50
	.dc.b	$DC,$38,$C6,$38,$3F,$1F,$C0,$C3,$97,$76,$AE,$65,$F3,$F3,$F0,$65
	.dc.b	$F5,$6F,$73,$9C,$BA,$3E,$BB,$E3,$5B,$CF,$CB,$9F,$C1,$9C,$B6,$BD
	.dc.b	$73,$A7,$91,$F8,$BF,$11,$EF,$FF,$B5,$F4,$FE,$4D,$C2,$7F,$A5,$C6
	.dc.b	$6B,$FB,$FC,$E5,$73,$7F,$A3,$E7,$4F,$F3,$19,$D3,$9F,$86,$B7,$FB
	.dc.b	$4C,$3F,$C0,$B5,$8C,$10,$FC,$2D,$D3,$98,$C5,$59,$CD,$FF,$60,$51
	.dc.b	$9A,$D3,$67,$4F,$7B,$7E,$18,$F1,$BB,$69,$FE,$93,$BF,$E5,$73,$97
	.dc.b	$E1,$FA,$1C,$63,$17,$F0,$A7,$D9,$5C,$63,$FD,$FB,$3A,$D7,$E3,$5F
	.dc.b	$BF,$17,$DE,$FF,$B9,$71,$FF,$03,$8F,$C3,$5F,$2B,$B9,$D6,$5D,$CF
	.dc.b	$AD,$3A,$AF,$F0,$7D,$F4,$E5,$F9,$91,$CF,$0E,$51,$8A,$E1,$7C,$EF
	.dc.b	$85,$F8,$A7,$39,$7D,$F9,$CD,$78,$6B,$A1,$E0,$F9,$EF,$C5,$BF,$88
	.dc.b	$F9,$3F,$74,$62,$7D,$1C,$5F,$90,$9C,$EF,$F4,$1B,$7C,$4E,$91,$96
	.dc.b	$99,$AF,$84,$3D,$9F,$F2,$B4,$7E,$63,$97,$EF,$46,$27,$C3,$5F,$AE
	.dc.b	$CF,$FD,$47,$39,$7F,$35,$5C,$3D,$2F,$C1,$B7,$0F,$F3,$35,$73,$C3
	.dc.b	$FC,$B6,$D7,$FA,$0B,$57,$C4,$3F,$0E,$7C,$67,$E3,$B8,$7F,$41,$39
	.dc.b	$8C,$61,$FE,$48,$8C,$3F,$B2,$9D,$D7,$28,$C1,$B7,$17,$F8,$5A,$E2
	.dc.b	$FC,$0F,$91,$E3,$87,$83,$F0,$FF,$A9,$AF,$EE,$63,$1B,$7E,$89,$B9
	.dc.b	$7F,$02,$6E,$1F,$29,$B5,$FC,$0F,$16,$1D,$2F,$F4,$66,$8F,$EE,$71
	.dc.b	$AE,$4F,$11,$E3,$F8,$9B,$3F,$06,$91,$8C,$7D,$91,$89,$D5,$7F,$05
	.dc.b	$FC,$41,$F1,$FC,$E7,$FC,$C7,$18,$7A,$61,$F0,$7E,$1E,$46,$B5,$B9
	.dc.b	$76,$F8,$1E,$91,$F1,$DB,$3E,$5F,$87,$F9,$47,$87,$E3,$6C,$7F,$8C
	.dc.b	$B9,$3F,$C6,$2F,$8B,$5F,$C9,$FA,$5E,$76,$C6,$4E,$11,$96,$99,$CB
	.dc.b	$F3,$F5,$D1,$FD,$5E,$BF,$B6,$9F,$E5,$8F,$FA,$B1,$F8,$3A,$FC,$5F
	.dc.b	$81,$F8,$CD,$8C,$9F,$1A,$32,$FC,$8F,$8F,$F9,$9F,$1B,$F3,$5B,$19
	.dc.b	$3E,$6C,$65,$A7,$D0,$DB,$EA,$70,$7D,$56,$C6,$4F,$AB,$19,$69,$9C
	.dc.b	$BD,$B1,$96,$C7,$E1,$7C,$8F,$C0,$75,$FE,$6D,$B1,$93,$F0,$31,$96
	.dc.b	$9F,$84,$DB,$F1,$1F,$0B,$F8,$96,$C6,$4F,$C4,$C6,$5A,$7D,$DF,$86
	.dc.b	$FC,$87,$C4,$FF,$0E,$6C,$64,$FC,$8C,$65,$A6,$71,$FB,$31,$96,$C7
	.dc.b	$E6,$3E,$23,$F3,$9F,$13,$F9,$D6,$C6,$4F,$CE,$C6,$5A,$7F,$09,$F9
	.dc.b	$A7,$C3,$FE,$97,$97,$F8,$7C,$62,$7F,$4B,$18,$9D,$47,$16,$E1,$FC
	.dc.b	$BB,$F1,$07,$CC,$7E,$BF,$83,$FB,$54,$62,$FF,$5F,$18,$9D,$47,$FA
	.dc.b	$52,$75,$5F,$98,$AD,$4F,$03,$55,$C2,$7F,$8D,$63,$FE,$AC,$9F,$F6
	.dc.b	$EF,$E2,$5C,$BC,$0F,$E2,$13,$9A,$FE,$9B,$39,$AF,$E1,$F3,$9A,$FF
	.dc.b	$0C,$78,$EA,$D5,$FF,$36,$AD,$5F,$F8,$56,$73,$E0,$A7,$FC,$B3,$3E
	.dc.b	$2F,$6F,$E5,$FB,$5F,$E1,$5B,$46,$27,$47,$D3,$9D,$47,$D2,$F8,$2F
	.dc.b	$F1,$06,$6F,$BE,$78,$39,$C3,$FD,$33,$3F,$E2,$A9,$CC,$7D,$DA,$E3
	.dc.b	$18,$D3,$E3,$C6,$2F,$C9,$C3,$2F,$E4,$FC,$65,$BF,$8F,$37,$1F,$8B
	.dc.b	$A2,$B1,$7F,$2B,$E0,$3F,$C4,$BF,$21,$F8,$A6,$C7,$E3,$DB,$8F,$D8
	.dc.b	$6E,$1F,$E4,$78,$7C,$F7,$9E,$3F,$C8,$BF,$4E,$C3,$FC,$69,$39,$C3
	.dc.b	$E0,$CE,$5F,$F6,$EC,$66,$FF,$CA,$33,$97,$FC,$07,$19,$B4,$E5,$F1
	.dc.b	$46,$63,$15,$FE,$B8,$8C,$5F,$8B,$4F,$46,$19,$6B,$FE,$CC,$9F,$A0
	.dc.b	$FF,$B0,$0F,$6C,$69,$7E,$24,$EA,$FF,$19,$FE,$15,$71,$FF,$7B,$CD
	.dc.b	$AF,$F1,$64,$DC,$FE,$44,$DA,$FF,$6D,$89,$C5,$7F,$B7,$A6,$CE,$6F
	.dc.b	$FC,$9D,$F5,$4F,$4F,$F2,$DF,$21,$86,$D7,$F9,$22,$74,$FE,$BA,$74
	.dc.b	$D3,$53,$C3,$0F,$F2,$45,$70,$7F,$63,$5E,$0F,$4F,$F4,$9F,$19,$70
	.dc.b	$FD,$CB,$6B,$F6,$ED,$D3,$F2,$2D,$9F,$D9,$36,$7F,$D0,$BC,$7F,$B6
	.dc.b	$7F,$CF,$F8,$FF,$99,$2F,$8D,$FA,$F0,$F0,$BF,$E6,$4F,$AA,$FC,$47
	.dc.b	$67,$F0,$1F,$9E,$FF,$39,$46,$32,$46,$38,$7F,$78,$DB,$FE,$22,$DC
	.dc.b	$FF,$87,$B7,$2F,$D4,$BD,$3C,$CF,$F4,$5F,$CC,$5B,$FF,$55,$BE,$0F
	.dc.b	$33,$FD,$1F,$F4,$F6,$7F,$D5,$EF,$FA,$57,$10,$F8,$5B,$EE,$7F,$A9
	.dc.b	$23,$F2,$11,$8F,$82,$FF,$74,$B6,$32,$7F,$A9,$23,$9A,$F8,$A7,$C8
	.dc.b	$FF,$AE,$A3,$47,$82,$33,$B7,$38,$E1,$3B,$C3,$C5,$18,$9F,$04,$38
	.dc.b	$5E,$9F,$82,$7F,$D8,$DF,$51,$75,$FF,$53,$B7,$0F,$F5,$CD,$F0,$7E
	.dc.b	$C7,$A7,$9E,$5F,$F6,$BF,$D4,$5F,$68,$F1,$D3,$C1,$57,$07,$B5,$C7
	.dc.b	$A1,$E3,$D3,$51,$33,$FE,$9B,$19,$7F,$11,$19,$7F,$A6,$C6,$5F,$F5
	.dc.b	$45,$76,$7F,$1D,$DF,$DF,$E4,$A4,$60,$87,$F5,$67,$2F,$28,$F4,$18
	.dc.b	$E0,$FE,$67,$38,$9E,$31,$87,$FB,$1C,$65,$F1,$9E,$FF,$91,$8F,$E8
	.dc.b	$4F,$88,$60,$FE,$CE,$E1,$D1,$FD,$B9,$D4,$FF,$B0,$6F,$F8,$F3,$FC
	.dc.b	$EF,$E0,$93,$F6,$69,$F6,$7F,$E7,$87,$FA,$CF,$1F,$E4,$23,$11,$FE
	.dc.b	$B8,$AE,$2E,$8F,$CF,$B8,$7F,$D7,$B1,$C2,$FF,$A5,$36,$BF,$37,$5B
	.dc.b	$9F,$E9,$DF,$42,$F8,$CF,$E6,$98,$7E,$82,$30,$E0,$38,$76,$D7,$F9
	.dc.b	$E3,$99,$FE,$B5,$F4,$4F,$33,$E0,$36,$BD,$47,$4B,$F8,$97,$C7,$1F
	.dc.b	$D8,$D6,$EB,$FD,$B2,$DC,$3D,$37,$C0,$EB,$84,$AE,$B8,$73,$7F,$C3
	.dc.b	$1C,$7F,$A3,$F4,$A4,$7F,$90,$63,$F5,$9E,$9F,$F8,$65,$CC,$7F,$B8
	.dc.b	$A7,$22,$3F,$4D,$DB,$27,$FC,$5E,$FE,$A7,$0C,$47,$F8,$BA,$2E,$5F
	.dc.b	$E0,$89,$EF,$F8,$9B,$C0,$FD,$E4,$BE,$39,$EF,$61,$B8,$FF,$B7,$EF
	.dc.b	$83,$6F,$55,$23,$06,$F3,$FE,$09,$F8,$AB,$E2,$A9,$E2,$89,$3F,$8D
	.dc.b	$EF,$56,$E3,$0D,$7F,$DD,$18,$FF,$8F,$DF,$27,$0F,$6F,$4A,$46,$08
	.dc.b	$35,$FF,$26,$DF,$F8,$6E,$3A,$CF,$C4,$AF,$D3,$7E,$2A,$F6,$7D,$3B
	.dc.b	$4B,$FC,$4C,$E5,$B3,$97,$B7,$37,$F4,$4D,$3D,$77,$D2,$A5,$7B,$23
	.dc.b	$86,$7F,$E3,$1C,$3A,$4F,$BB,$D1,$F9,$3E,$14,$8F,$A5,$1F,$68,$C3
	.dc.b	$E0,$18,$FC,$A9,$D4,$7C,$AF,$AA,$FA,$7C,$12,$30,$43,$95,$7D,$28
	.dc.b	$D4,$63,$75,$7C,$6F,$9A,$FE,$89,$EE,$BF,$01,$FE,$73,$7F,$C1,$E3
	.dc.b	$1E,$1E,$E3,$E4,$D7,$0A,$DC,$7A,$6B,$6E,$88,$66,$37,$B3,$87,$FB
	.dc.b	$8D,$70,$AF,$84,$37,$61,$F9,$2F,$AA,$FC,$9F,$1C,$AF,$C6,$C6,$5F
	.dc.b	$C7,$C6,$5F,$81,$1F,$A8,$C3,$F3,$7C,$32,$7F,$49,$19,$7F,$2F,$19
	.dc.b	$7F,$2F,$1F,$A2,$EF,$FD,$1F,$D5,$49,$FC,$EC,$65,$FC,$FC,$65,$FA
	.dc.b	$11,$97,$E9,$46,$74,$FC,$DC,$FE,$BB,$D4,$FD,$7F,$CD,$CB,$FD,$4C
	.dc.b	$65,$FD,$5C,$65,$FB,$91,$FB,$6F,$8F,$FD,$BF,$C8,$49,$FD,$8C,$65
	.dc.b	$FD,$9C,$65,$FC,$24,$7E,$A3,$C1,$FD,$FE,$72,$7F,$4F,$19,$7F,$77
	.dc.b	$19,$7E,$BF,$07,$F1,$D7,$66,$88,$6A,$7F,$93,$FE,$2D,$FE,$11,$8F
	.dc.b	$D0,$78,$68,$C6,$BF,$85,$94,$E7,$D0,$FF,$0B,$7C,$AB,$E4,$79,$FA
	.dc.b	$18,$F4,$66,$1F,$D7,$F0,$E9,$3E,$0F,$8D,$93,$9C,$D1,$C2,$3F,$53
	.dc.b	$8F,$FB,$19,$D7,$8D,$E8,$3C,$3E,$DD,$8E,$5F,$47,$EC,$30,$FB,$30
	.dc.b	$46,$63,$67,$43,$C5,$F1,$5F,$10,$8F,$F2,$BC,$1F,$25,$88,$F2,$FC
	.dc.b	$8F,$C2,$31,$C3,$F1,$F6,$F9,$58,$66,$B9,$56,$B9,$3F,$CB,$F5,$F3
	.dc.b	$E7,$E9,$90,$F0,$7B,$D5,$F8,$7E,$EF,$D0,$31,$8F,$7D,$7F,$C2,$F8
	.dc.b	$7F,$27,$D1,$5B,$BF,$9C,$7F,$86,$7E,$6F,$F9,$C3,$FE,$08,$6B,$FE
	.dc.b	$BD,$6B,$FE,$6A,$6B,$FE,$76,$78,$E1,$F8,$5F,$21,$6E,$3F,$C2,$9D
	.dc.b	$FC,$C9,$87,$CB,$8C,$BF,$E0,$AC,$39,$E5,$FE,$80,$FA,$57,$C9,$7E
	.dc.b	$3B,$9A,$5F,$F8,$A7,$A9,$5A,$8C,$35,$FC,$4C,$F0,$7F,$9E,$4F,$0A
	.dc.b	$FF,$35,$56,$F5,$F5,$69,$F9,$2B,$FC,$06,$1A,$BD,$FB,$5F,$E7,$9C
	.dc.b	$9F,$26,$F8,$6C,$37,$0F,$1D,$70,$7F,$C8,$15,$C2,$B6,$66,$3C,$B9
	.dc.b	$FF,$BA,$5D,$DF,$37,$FA,$A5,$71,$7A,$9F,$E3,$2F,$99,$7C,$7F,$FA
	.dc.b	$1F,$CE,$93,$98,$FF,$2C,$4E,$A3,$FC,$0D,$DD,$FD,$49,$B5,$AA,$FF
	.dc.b	$49,$7E,$2F,$FE,$37,$C3,$FB,$93,$FE,$A2,$E0,$FC,$03,$63,$3E,$C7
	.dc.b	$03,$8F,$B3,$31,$FF,$0A,$57,$37,$FB,$6F,$13,$E5,$3F,$D3,$DE,$CC
	.dc.b	$9F,$D8,$47,$F0,$E3,$0F,$F6,$CD,$7F,$C3,$59,$FD,$89,$FC,$E7,$95
	.dc.b	$FC,$87,$E1,$9F,$39,$F9,$98,$C3,$7A,$3F,$D9,$F3,$AC,$5F,$3C,$FA
	.dc.b	$0F,$8D,$1F,$ED,$3B,$E0,$D3,$F4,$58,$74,$6E,$1D,$27,$FD,$B7,$7C
	.dc.b	$F8,$BF,$DA,$78,$F7,$B7,$5B,$1E,$51,$86,$C6,$71,$E9,$3E,$6C,$BC
	.dc.b	$B7,$CE,$B5,$3E,$53,$95,$70,$69,$E9,$C7,$C8,$E6,$78,$D7,$3C,$3C
	.dc.b	$27,$B6,$B7,$5D,$6B,$BB,$FC,$1E,$7F,$A8,$36,$FF,$D2,$13,$A8,$FF
	.dc.b	$79,$FC,$4F,$F9,$0F,$E8,$3F,$D2,$8D,$8F,$F4,$CB,$5F,$F7,$C9,$FE
	.dc.b	$F9,$F7,$93,$D3,$D1,$7B,$14,$95,$FC,$6E,$32,$D8,$F7,$46,$34,$FC
	.dc.b	$06,$8F,$A0,$7A,$9F,$D7,$E7,$E6,$56,$F8,$7F,$91,$37,$1F,$E4,$4D
	.dc.b	$C3,$95,$F1,$AD,$E3,$E0,$25,$FD,$1C,$7B,$F5,$63,$2E,$F5,$6F,$F1
	.dc.b	$35,$AB,$FA,$19,$78,$5F,$C4,$B7,$63,$E3,$9E,$18,$74,$AF,$CC,$78
	.dc.b	$AE,$6F,$F6,$46,$7F,$E8,$7F,$A3,$FF,$52,$3F,$C1,$3D,$9F,$F0,$07
	.dc.b	$47,$F3,$D9,$CE,$1F,$B1,$9C,$D7,$F9,$D6,$73,$5F,$D0,$EF,$AD,$FF
	.dc.b	$98,$A7,$35,$FE,$7B,$F8,$27,$6F,$2A,$FF,$37,$C6,$9B,$19,$CB,$FD
	.dc.b	$59,$CD,$6B,$8D,$FF,$A7,$DE,$13,$FE,$B1,$AD,$57,$5B,$78,$0F,$A1
	.dc.b	$E9,$9F,$F5,$8C,$BF,$95,$46,$3B,$B8,$56,$E7,$F9,$3C,$62,$75,$1F
	.dc.b	$D4,$FE,$87,$E8,$BF,$CF,$BE,$87,$FC,$0E,$E7,$8F,$DC,$A3,$39,$7F
	.dc.b	$54,$AE,$2F,$F3,$E8,$CF,$07,$F8,$C7,$45,$6A,$3F,$C6,$97,$CF,$D2
	.dc.b	$F9,$55,$AC,$7B,$E1,$CA,$98,$FF,$B8,$71,$E5,$18,$39,$1F,$DB,$7C
	.dc.b	$1F,$C3,$3F,$D1,$BD,$CE,$B1,$8A,$FE,$65,$18,$DF,$E5,$B6,$B8,$1C
	.dc.b	$E3,$2D,$33,$3F,$E8,$EF,$4F,$FC,$83,$1F,$D5,$3A,$FE,$C4,$62,$7F
	.dc.b	$A0,$46,$3C,$CF,$B2,$DA,$E0,$7B,$B7,$FF,$2C,$52,$F9,$46,$1B,$3F
	.dc.b	$EC,$DC,$3F,$99,$3C,$4F,$8F,$5C,$70,$FC,$B7,$A4,$FA,$17,$B7,$F8
	.dc.b	$AE,$0C,$C6,$30,$FE,$DB,$18,$9D,$47,$F9,$8E,$75,$87,$F5,$BA,$D4
	.dc.b	$E6,$BF,$2B,$EE,$62,$BF,$CE,$BF,$43,$F4,$75,$FD,$75,$79,$23,$17
	.dc.b	$EB,$8C,$69,$F9,$E6,$C7,$F8,$01,$B7,$D2,$7F,$47,$38,$BF,$F2,$D5
	.dc.b	$FE,$26,$BF,$D0,$2D,$CC,$BD,$7F,$17,$2F,$8D,$B8,$7F,$A0,$67,$4D
	.dc.b	$9F,$DE,$7D,$0F,$FA,$26,$35,$58,$F7,$39,$47,$1F,$57,$DC,$70,$D2
	.dc.b	$C7,$5F,$47,$0D,$3E,$D6,$9F,$EF,$AC,$7E,$2B,$EA,$F8,$FF,$90,$78
	.dc.b	$BD,$CB,$5C,$27,$F1,$1E,$EF,$9B,$5F,$C9,$B6,$75,$CC,$8C,$57,$E6
	.dc.b	$DB,$C3,$F8,$F6,$D7,$E3,$2B,$83,$E6,$F0,$59,$FC,$F6,$FB,$9F,$1B
	.dc.b	$87,$E3,$B6,$B8,$52,$31,$93,$2F,$F3,$A4,$E6,$B7,$7F,$43,$DE,$FA
	.dc.b	$3E,$E5,$9E,$6D,$C3,$94,$7F,$08,$CB,$CB,$39,$C7,$AB,$F6,$B8,$3E
	.dc.b	$DF,$15,$BF,$2B,$70,$E6,$7D,$37,$17,$E6,$6C,$F7,$CF,$FD,$14,$FB
	.dc.b	$AF,$AB,$F8,$BF,$33,$F1,$3D,$7F,$A1,$6C,$64,$F8,$B1,$96,$9F,$23
	.dc.b	$83,$F2,$3E,$DF,$E4,$9B,$19,$3E,$6C,$65,$A7,$E0,$3E,$87,$F3,$3E
	.dc.b	$B7,$F1,$66,$C6,$4F,$AB,$19,$69,$9D,$7E,$1E,$FF,$80,$9D,$46,$34
	.dc.b	$FF,$32,$57,$DE,$F7,$3F,$4F,$F4,$3F,$A8,$6C,$64,$FC,$1C,$65,$A7
	.dc.b	$FB,$C7,$E2,$1F,$3D,$FB,$0E,$7F,$E3,$71,$89,$FC,$5C,$62,$75,$1C
	.dc.b	$5B,$8F,$DA,$1F,$42,$FA,$1F,$DD,$F1,$7F,$6F,$8C,$5F,$E5,$A3,$13
	.dc.b	$A8,$FD,$74,$EA,$BE,$05,$6A,$78,$1A,$AE,$13,$F8,$AF,$01,$93,$E6
	.dc.b	$FF,$C0,$DF,$2D,$CF,$A3,$FD,$E5,$39,$C7,$F5,$D3,$97,$FD,$E3,$19
	.dc.b	$C3,$E9,$CE,$56,$1D,$B0,$EB,$8F,$C8,$F6,$BB,$FB,$D7,$2E,$AD,$C3
	.dc.b	$9B,$73,$FF,$73,$65,$FE,$B6,$AF,$A9,$F1,$9F,$E2,$19,$7D,$0B,$C2
	.dc.b	$FA,$B3,$98,$C5,$70,$3F,$8E,$D7,$0A,$DC,$62,$77,$8F,$B3,$87,$E2
	.dc.b	$9F,$1B,$E8,$6D,$7C,$48,$B3,$F0,$A6,$E1,$F2,$BE,$0F,$FC,$4F,$F0
	.dc.b	$FF,$88,$6E,$1F,$8D,$6D,$FD,$5E,$24,$62,$FF,$B9,$FD,$07,$91,$FF
	.dc.b	$26,$FD,$0B,$1F,$F1,$B4,$E7,$1F,$EF,$53,$97,$F9,$A4,$67,$1F,$E3
	.dc.b	$D3,$97,$F2,$91,$9B,$4E,$5F,$0E,$BF,$E0,$4A,$FE,$FB,$5F,$19,$FF
	.dc.b	$47,$B5,$CC,$6D,$FF,$14,$C7,$85,$3E,$3E,$47,$D7,$7C,$0C,$7C,$F7
	.dc.b	$FE,$51,$FA,$17,$5F,$F9,$0B,$E8,$AD,$7F,$4F,$9B,$97,$FB,$FA,$6E
	.dc.b	$7F,$E1,$A9,$B5,$FE,$37,$9B,$5F,$E3,$C9,$B2,$9C,$FB,$0F,$A2,$7A
	.dc.b	$DF,$A6,$F8,$DF,$F2,$3C,$62,$FF,$4B,$18,$9D,$47,$F6,$AA,$E5,$18
	.dc.b	$C7,$F9,$DC,$62,$78,$9B,$BE,$93,$FA,$CB,$EF,$EA,$FE,$DB,$B6,$DF
	.dc.b	$ED,$1B,$5F,$B2,$6E,$9F,$A8,$6C,$FF,$53,$EA,$FD,$C4,$FF,$99,$71
	.dc.b	$FF,$2D,$DB,$C4,$56,$AF,$FC,$C3,$9F,$C3,$CB,$C5,$1E,$76,$DF,$B3
	.dc.b	$E7,$7F,$1D,$87,$2F,$A2,$FF,$36,$C6,$31,$46,$38,$BF,$88,$B7,$1F
	.dc.b	$E1,$ED,$C3,$F8,$6B,$74,$FE,$06,$F4,$F4,$3F,$D2,$7F,$45,$6F,$FC
	.dc.b	$09,$EA,$23,$C3,$E8,$7F,$A6,$3E,$6A,$D6,$9F,$F4,$F6,$5F,$E7,$EF
	.dc.b	$86,$F7,$FC,$1F,$FB,$AA,$11,$87,$FD,$DA,$D7,$FC,$BE,$DF,$8C,$7D
	.dc.b	$17,$F8,$2E,$5C,$9F,$EB,$87,$FD,$17,$3A,$DF,$85,$6E,$B5,$4A,$D3
	.dc.b	$CB,$0F,$A7,$C4,$F5,$BF,$CD,$F8,$FF,$B2,$7E,$8A,$EB,$FE,$75,$6E
	.dc.b	$1E,$2B,$E0,$FE,$55,$7A,$FA,$E6,$FF,$6F,$7C,$E5,$E3,$3C,$B6,$5F
	.dc.b	$0C,$39,$7B,$7C,$7A,$1E,$5D,$38,$52,$69,$9A,$FE,$9B,$38,$79,$98
	.dc.b	$21,$F8,$D8,$E7,$E0,$F7,$FA,$29,$3E,$B8,$CB,$FD,$62,$32,$FE,$56
	.dc.b	$32,$FF,$58,$8C,$F0,$7F,$B1,$F0,$FE,$1F,$3C,$CE,$44,$39,$1F,$24
	.dc.b	$F8,$1F,$45,$1C,$10,$FE,$CA,$E5,$FD,$31,$F1,$0C,$4F,$F5,$E8,$C4
	.dc.b	$7F,$9A,$FE,$0D,$1F,$6A,$1F,$45,$FE,$89,$7F,$CE,$0D,$7F,$C1,$D3
	.dc.b	$C1,$E5,$3C,$67,$55,$FE,$BE,$6F,$1F,$FA,$E2,$F8,$5F,$F8,$5F,$E8
	.dc.b	$5F,$1D,$E1,$6C,$7F,$9F,$DA,$FF,$87,$7E,$8B,$F3,$EE,$67,$F5,$4F
	.dc.b	$FB,$AA,$76,$FF,$B1,$83,$87,$FD,$92,$D7,$84,$7C,$19,$FE,$CB,$1F
	.dc.b	$E8,$F6,$BE,$78,$F3,$1B,$F3,$A4,$78,$CF,$F1,$E7,$17,$E7,$7C,$F9
	.dc.b	$18,$21,$FE,$4E,$E5,$FF,$13,$BF,$C3,$23,$F5,$7E,$67,$F8,$AA,$BF
	.dc.b	$41,$D7,$2B,$FC,$4E,$FE,$8F,$0C,$47,$F8,$9A,$2E,$FF,$E1,$C9,$E7
	.dc.b	$F4,$4F,$03,$F6,$F2,$F8,$FF,$F5,$36,$1F,$E1,$86,$BF,$E3,$BA,$E0
	.dc.b	$FF,$AE,$32,$46,$23,$FD,$13,$5C,$FE,$03,$AD,$F2,$87,$12,$BF,$C7
	.dc.b	$37,$C6,$73,$18,$D7,$F8,$37,$D1,$3E,$2E,$9E,$A8,$93,$FE,$B5,$9D
	.dc.b	$36,$72,$FA,$78,$79,$FA,$92,$3D,$58,$FE,$33,$1F,$0E,$BE,$CB,$D5
	.dc.b	$7F,$A0,$B1,$FF,$11,$3C,$23,$F4,$DF,$3C,$F6,$BC,$FC,$12,$BF,$DA
	.dc.b	$D3,$96,$CE,$5E,$93,$CA,$24,$7A,$E3,$58,$FC,$48,$57,$37,$EA,$7A
	.dc.b	$3F,$1B,$C1,$91,$EC,$8F,$BE,$61,$F5,$46,$1F,$91,$F3,$7F,$3B,$82
	.dc.b	$46,$08,$7C,$17,$2F,$C3,$8F,$83,$F4,$3F,$D0,$3D,$E7,$D1,$7F,$34
	.dc.b	$AF,$F7,$EC,$63,$C4,$DC,$7C,$4A,$E1,$5B,$8F,$F6,$8D,$6D,$D1,$0C
	.dc.b	$C6,$F4,$FC,$3C,$EE,$7F,$D4,$1A,$2F,$F1,$3F,$45,$F8,$AE,$09,$5F
	.dc.b	$87,$8C,$BF,$89,$8C,$BF,$7E,$3F,$4F,$CD,$F9,$2D,$E4,$FE,$22,$32
	.dc.b	$FE,$46,$32,$FE,$9E,$3F,$33,$DF,$F9,$AF,$A2,$93,$F9,$78,$CB,$F9
	.dc.b	$98,$CB,$F8,$F8,$CB,$F9,$18,$CE,$9F,$A6,$F2,$1C,$AB,$FD,$3F,$A9
	.dc.b	$FA,$8F,$9D,$98,$FE,$96,$32,$FE,$9E,$32,$FE,$76,$3F,$5F,$F2,$1F
	.dc.b	$B0,$F9,$19,$3F,$AD,$8C,$BF,$AF,$8C,$BF,$A5,$8F,$D1,$F8,$5F,$B8
	.dc.b	$D6,$4F,$E8,$A3,$2F,$ED,$E3,$2F,$D8,$8C,$BF,$AF,$8C,$E9,$FA,$1E
	.dc.b	$2F,$E7,$F3,$FE,$CD,$FA,$1F,$F0,$6B,$7D,$0F,$E4,$F1,$8A,$F8,$DF
	.dc.b	$2E,$F9,$2C,$FA,$3F,$E1,$18,$79,$14,$CE,$3F,$C7,$70,$7C,$7F,$42
	.dc.b	$47,$F1,$9C,$7A,$D7,$3F,$6F,$C3,$E4,$4A,$D6,$DF,$C7,$F9,$99,$FF
	.dc.b	$B2,$67,$B5,$7A,$3E,$8A,$46,$0F,$13,$87,$E0,$3D,$5F,$85,$F1,$7F
	.dc.b	$00,$8F,$80,$FF,$37,$53,$E4,$F3,$0D,$3E,$2F,$C9,$FC,$33,$2F,$F2
	.dc.b	$2D,$D8,$66,$BA,$67,$EA,$BD,$E9,$FC,$EF,$C8,$61,$FD,$BF,$1E,$58
	.dc.b	$7D,$43,$E8,$7B,$D5,$E7,$A3,$DD,$FA,$86,$63,$EF,$18,$36,$E6,$BF
	.dc.b	$92,$7C,$E7,$F3,$49,$F6,$65,$FD,$35,$46,$67,$7E,$6F,$E0,$79,$FF
	.dc.b	$C3,$D3,$9B,$FA,$D8,$FF,$93,$F0,$E6,$7C,$AE,$6D,$E1,$23,$D3,$1F
	.dc.b	$8B,$30,$FF,$AC,$3E,$89,$F2,$9F,$8F,$E8,$93,$FE,$B7,$9C,$B6,$72
	.dc.b	$F8,$38,$BF,$0F,$5F,$E6,$3A,$DC,$F5,$AD,$EB,$F9,$EA,$D4,$6F,$DA
	.dc.b	$FF,$39,$E8,$F9,$57,$C3,$61,$B8,$F8,$EB,$83,$F8,$9A,$E1,$5B,$33
	.dc.b	$1E,$D9,$C3,$BA,$E6,$E4,$FC,$90,$F9,$B7,$C8,$7F,$9F,$FC,$F9,$3A
	.dc.b	$9C,$B6,$72,$FF,$89,$67,$FC,$A9,$3A,$8F,$8D,$F4,$5F,$E3,$4D,$7F
	.dc.b	$92,$3F,$E9,$C6,$FB,$3F,$E8,$F8,$D6,$9F,$B3,$6D,$7F,$B4,$DE,$11
	.dc.b	$99,$E1,$EC,$FF,$07,$F9,$6F,$F4,$B7,$B3,$30,$FC,$CC,$65,$FF,$61
	.dc.b	$C6,$5F,$B5,$CC,$9F,$C7,$F9,$5F,$E5,$8F,$86,$7C,$FF,$E5,$FA,$93
	.dc.b	$98,$C3,$70,$7D,$03,$E8,$BE,$16,$1C,$A7,$FC,$59,$87,$2F,$2F,$FD
	.dc.b	$9D,$8F,$26,$CE,$B1,$E9,$8F,$3E,$4F,$F6,$6D,$EE,$BF,$DA,$6D,$C7
	.dc.b	$B4,$7E,$B2,$F8,$E5,$E2,$AD,$12,$7B,$63,$E0,$C7,$B4,$78,$8E,$51
	.dc.b	$72,$F4,$47,$80,$E2,$59,$E2,$E4,$E2,$7F,$BA,$63,$8B,$FE,$7F,$6B
	.dc.b	$EB,$3F,$D0,$5F,$11,$FE,$37,$FA,$2F,$F4,$3B,$63,$FD,$E4,$D7,$7E
	.dc.b	$F2,$7F,$D1,$B4,$F4,$5E,$CE,$1B,$7E,$36,$BE,$36,$BF,$7D,$DE,$DF
	.dc.b	$32,$BA,$57,$C7,$FA,$2F,$EF,$9C,$DF,$16,$F7,$B7,$1A,$C1,$BF,$53
	.dc.b	$FC,$D3,$8F,$3A,$DE,$5F,$84,$6B,$FC,$91,$CE,$CC,$BB,$CF,$CD,$CF
	.dc.b	$C6,$DA,$E3,$18,$BF,$C8,$5F,$1C,$3C,$39,$79,$70,$E1,$97,$99,$FB
	.dc.b	$59,$76,$C3,$CD,$3F,$94,$F0,$BF,$B9,$FB,$3F,$E8,$2E,$BF,$F0,$F7
	.dc.b	$A3,$FE,$C4,$DF,$FA,$8A,$FA,$ED,$47,$F6,$D8,$D6,$7F,$EA,$BB,$CA
	.dc.b	$C3,$AD,$72,$BC,$C6,$27,$F9,$87,$C1,$3B,$72,$B9,$AC,$61,$B8,$7F
	.dc.b	$64,$9D,$36,$74,$FF,$1D,$FA,$2E,$F9,$FF,$80,$A3,$07,$B3,$B3,$CF
	.dc.b	$8F,$F2,$7A,$FF,$40,$3F,$D6,$FC,$04,$62,$3F,$D5,$51,$8A,$FF,$70
	.dc.b	$B6,$3F,$DC,$19,$3E,$87,$F3,$15,$FC,$F7,$E8,$BF,$C0,$8F,$F8,$1E
	.dc.b	$75,$D1,$F2,$21,$3A,$BF,$EC,$75,$A9,$E5,$E9,$FC,$E8,$C4,$FF,$68
	.dc.b	$9E,$33,$9C,$3C,$33,$C9,$BC,$5F,$7B,$EA,$75,$97,$3F,$0B,$FC,$6B
	.dc.b	$8F,$E0,$3D,$EC,$70,$7C,$A6,$C6,$4F,$54,$F0,$CB,$FD,$D9,$3A,$7D
	.dc.b	$31,$9C,$7F,$1B,$EF,$65,$FF,$57,$7A,$DF,$E4,$0C,$3F,$1F,$D9,$F5
	.dc.b	$23,$18,$7F,$BD,$E3,$1E,$6F,$DC,$6D,$70,$3F,$B0,$D7,$0B,$EB,$5F
	.dc.b	$82,$F1,$37,$87,$F8,$3B,$D5,$7D,$1B,$DD,$FC,$3F,$07,$E6,$A7,$31
	.dc.b	$8C,$FF,$DE,$55,$B7,$F4,$B5,$B8,$C5,$FF,$82,$F8,$3E,$7D,$7F,$8D
	.dc.b	$E7,$39,$FF,$B8,$1B,$1F,$AB,$F7,$B9,$63,$E1,$FA,$3F,$91,$3F,$AF
	.dc.b	$8C,$61,$F9,$C6,$C7,$ED,$9B,$8F,$85,$B8,$FF,$BB,$AF,$9C,$FE,$D6
	.dc.b	$D7,$A3,$FC,$99,$3F,$91,$AE,$1C,$BE,$3F,$6B,$C8,$DA,$FF,$30,$6A
	.dc.b	$BE,$11,$CA,$B6,$FF,$96,$BE,$8B,$FD,$08,$F0,$94,$FF,$04,$9C,$E5
	.dc.b	$F5,$67,$35,$BF,$5B,$E2,$1B,$CE,$C7,$3F,$4B,$0D,$3C,$11,$F2,$FE
	.dc.b	$42,$47,$8F,$DF,$B8,$FF,$94,$0F,$F3,$8F,$BF,$E3,$DB,$F2,$4D,$AF
	.dc.b	$C6,$B6,$78,$73,$23,$1C,$1F,$9C,$7F,$84,$4F,$4A,$FF,$33,$3F,$E6
	.dc.b	$7A,$E1,$18,$81,$B3,$7E,$23,$77,$BB,$F1,$F8,$7E,$43,$6F,$E7,$36
	.dc.b	$FF,$CB,$95,$C2,$7E,$95,$7C,$FF,$73,$E8,$7B,$B6,$7E,$B3,$6F,$8B
	.dc.b	$74,$FF,$71,$61,$FE,$89,$AF,$B9,$C3,$F6,$78,$BE,$D3,$63,$27,$E0
	.dc.b	$63,$2D,$33,$87,$9A,$32,$D8,$FC,$37,$95,$F8,$5E,$BF,$D2,$36,$32
	.dc.b	$7C,$9E,$6F,$C9,$47,$F3,$76,$C6,$79,$BF,$C1,$2F,$CF,$E0,$F9,$9C
	.dc.b	$5F,$91,$F6,$FF,$24,$D9,$FC,$C3,$72,$F9,$6D,$DB,$F8,$CF,$D1,$FE
	.dc.b	$67,$D7,$FE,$30,$DB,$FE,$26,$DB,$FC,$3B,$6B,$8C,$FE,$06,$7F,$D8
	.dc.b	$33,$F8,$8E,$53,$4F,$EA,$59,$7E,$0F,$DB,$FD,$47,$D1,$FE,$A5,$B8
	.dc.b	$7E,$BD,$B8,$7E,$09,$B8,$FF,$00,$86,$CF,$89,$7D,$07,$EC,$F4,$FE
	.dc.b	$45,$18,$C7,$F7,$51,$89,$F9,$D3,$9B,$FC,$ED,$F3,$9C,$9F,$A2,$FA
	.dc.b	$37,$D1,$FE,$FF,$8F,$FB,$8C,$63,$2F,$E0,$D1,$89,$D4,$7F,$A7,$27
	.dc.b	$58,$7E,$72,$B5,$3F,$26,$73,$3F,$AE,$6B,$FE,$D5,$6B,$F0,$31,$7C
	.dc.b	$E7,$F8,$53,$E5,$FA,$EB,$F1,$74,$F8,$93,$9C,$FE,$1B,$F0,$09,$97
	.dc.b	$E4,$70,$E6,$DF,$73,$A7,$BF,$F2,$9B,$1F,$CA,$25,$1F,$28,$91,$C4
	.dc.b	$E9,$F1,$DF,$E2,$CC,$9F,$46,$F0,$FE,$AC,$E6,$31,$5D,$63,$EA,$91
	.dc.b	$E6,$7E,$87,$5F,$B1,$7F,$83,$E5,$46,$27,$F8,$9C,$63,$E8,$FE,$37
	.dc.b	$1C,$9F,$E2,$F7,$C6,$F8,$70,$FF,$90,$55,$EE,$3F,$B8,$6A,$F8,$3F
	.dc.b	$F1,$AF,$C4,$7E,$39,$B8,$FC,$57,$F1,$A4,$7E,$73,$FC,$B3,$E8,$BD
	.dc.b	$F6,$CB,$FC,$B5,$F4,$6C,$3F,$C7,$B5,$A9,$CC,$63,$83,$91,$31,$FD
	.dc.b	$95,$74,$C3,$F9,$BB,$C6,$3F,$9A,$C2,$30,$FE,$AB,$E8,$FD,$93,$FE
	.dc.b	$F2,$9C,$CE,$EB,$C1,$3F,$DA,$2B,$73,$FD,$46,$B7,$F4,$6E,$BF,$F2
	.dc.b	$27,$D1,$DC,$3F,$DF,$B3,$6F,$FC,$81,$37,$3F,$DE,$42,$74,$FF,$50
	.dc.b	$6B,$EA,$6B,$FD,$41,$BE,$CB,$E8,$DE,$C7,$E9,$7E,$37,$FC,$A1,$18
	.dc.b	$CB,$F0,$27,$E1,$B9,$13,$AA,$E1,$3F,$E5,$FA,$E1,$8F,$D7,$8E,$15
	.dc.b	$B7,$FB,$16,$1E,$1B,$FF,$38,$36,$33,$7C,$AF,$E9,$CF,$9F,$D5,$FD
	.dc.b	$D7,$67,$99,$B7,$E1,$6D,$7F,$9B,$2F,$FC,$6A,$FE,$5E,$B8,$56,$CF
	.dc.b	$6C,$7E,$6F,$DE,$CB,$C6,$EE,$BF,$A1,$9B,$5F,$DE,$B1,$7C,$F7,$E1
	.dc.b	$BA,$BF,$5B,$F4,$7F,$E7,$98,$C6,$E8,$C7,$2F,$F1,$36,$E1,$E2,$BE
	.dc.b	$2F,$FA,$26,$FF,$23,$87,$6B,$FE,$35,$DD,$C3,$5F,$F4,$4D,$7F,$80
	.dc.b	$5F,$F2,$A7,$D1,$F8,$7D,$3F,$F5,$2B,$E3,$F4,$7F,$D4,$FF,$37,$73
	.dc.b	$FF,$60,$BF,$EA,$6B,$FF,$0A,$FC,$47,$B7,$E1,$7F,$C3,$0D,$8F,$F0
	.dc.b	$EB,$5F,$D3,$B7,$E3,$9F,$47,$FB,$79,$FF,$5F,$57,$EB,$27,$3C,$9B
	.dc.b	$AD,$70,$7F,$B1,$6F,$FC,$4B,$5A,$78,$9F,$EC,$7D,$9E,$BF,$ED,$B0
	.dc.b	$FF,$6D,$FD,$1D,$D7,$FD,$B0,$DB,$E1,$57,$1F,$AF,$EB,$67,$91,$EE
	.dc.b	$3E,$77,$0D,$C3,$E9,$57,$1D,$1E,$ED,$24,$FD,$9E,$9F,$82,$FD,$98
	.dc.b	$ED,$0F,$2E,$9C,$27,$EC,$CE,$1E,$51,$9D,$FF,$BD,$69,$F8,$EF,$0F
	.dc.b	$C4,$4A,$F0,$7C,$E5,$7F,$07,$16,$BF,$9C,$4D,$E1,$FD,$D7,$D1,$F9
	.dc.b	$0C,$5F,$CD,$BC,$3E,$B8,$CE,$1F,$9E,$8F,$F0,$D3,$5F,$DD,$E9,$FE
	.dc.b	$1D,$7E,$B1,$A7,$FC,$C7,$8F,$F4,$9F,$A3,$F8,$67,$CB,$98,$7C,$97
	.dc.b	$2F,$E8,$4F,$B4,$3E,$15,$1D,$DB,$1F,$E2,$D6,$BF,$E6,$F6,$BF,$E7
	.dc.b	$47,$33,$FC,$AB,$E8,$FF,$D2,$B3,$FC,$08,$38,$7F,$D1,$13,$C3,$3F
	.dc.b	$E8,$BF,$46,$F9,$1F,$2B,$6B,$93,$5F,$F4,$87,$D1,$B4,$DD,$49,$DB
	.dc.b	$EE,$9D,$46,$0F,$F2,$45,$73,$7F,$D1,$AD,$7C,$73,$F4,$63,$C1,$97
	.dc.b	$F4,$6A,$FE,$E5,$5E,$D6,$BF,$EA,$39,$DE,$1F,$E9,$4F,$5F,$FC,$8D
	.dc.b	$8F,$F9,$5D,$B5,$FE,$21,$8C,$7A,$5E,$68,$FF,$21,$B7,$1F,$F0,$77
	.dc.b	$37,$F9,$3E,$7F,$8E,$35,$FF,$02,$D7,$E9,$3B,$64,$7E,$A5,$FD,$3F
	.dc.b	$A5,$88,$FE,$94,$7F,$83,$E3,$17,$FE,$13,$9F,$DF,$62,$FA,$37,$83
	.dc.b	$E9,$6E,$1F,$E8,$96,$BF,$E5,$A6,$BF,$E0,$09,$E5,$3F,$C2,$6D,$F2
	.dc.b	$5F,$EC,$1A,$FE,$8F,$18,$F8,$0F,$0E,$1D,$23,$FC,$98,$FB,$E7,$A3
	.dc.b	$5F,$9C,$D7,$FC,$EA,$DC,$3D,$5A,$FF,$15,$FA,$17,$C7,$53,$BC,$78
	.dc.b	$38,$FD,$C4,$C3,$DC,$D7,$E4,$F8,$0B,$FF,$14,$47,$F4,$F6,$BF,$3B
	.dc.b	$0F,$95,$5F,$B0,$FA,$1F,$F2,$CC,$FF,$A6,$5A,$FF,$B6,$DA,$FF,$B8
	.dc.b	$1A,$FF,$8A,$9D,$E7,$FE,$51,$BF,$8D,$C7,$FE,$50,$6D,$7F,$B9,$2B
	.dc.b	$73,$C7,$CA,$F8,$5C,$12,$BF,$6D,$5A,$7E,$4F,$A3,$F7,$C8,$FD,$F6
	.dc.b	$BF,$0B,$0F,$E7,$8D,$F0,$FE,$CC,$FC,$FF,$9D,$FA,$1B,$CA,$F8,$D1
	.dc.b	$97,$E4,$46,$5E,$91,$F1,$FC,$0F,$E4,$58,$9A,$49,$F9,$4F,$F5,$5F
	.dc.b	$80,$7D,$1F,$F4,$59,$FF,$09,$C6,$3C,$3E,$91,$F0,$2B,$85,$6E,$3F
	.dc.b	$4B,$5B,$74,$7D,$DF,$A3,$FC,$57,$14,$8C,$10,$F8,$8E,$5F,$8D,$1F
	.dc.b	$9C,$E1,$FD,$51,$23,$F5,$4D,$7E,$66,$BF,$97,$38,$3F,$8D,$EF,$FC
	.dc.b	$D7,$C9,$48,$C1,$0F,$A8,$E5,$FA,$D1,$FA,$13,$1C,$9F,$89,$9C,$4F
	.dc.b	$E6,$7D,$4F,$D3,$7C,$94,$8C,$10,$FC,$03,$97,$F0,$51,$FA,$DF,$90
	.dc.b	$FD,$77,$C8,$C9,$FD,$24,$65,$FD,$34,$65,$FC,$54,$7E,$97,$C1,$FB
	.dc.b	$E4,$8F,$D3,$35,$FC,$86,$9F,$BB,$38,$3F,$9A,$8C,$3F,$B3,$8C,$BF
	.dc.b	$B3,$FA,$1F,$F0,$FB,$FE,$56,$EB,$46,$23,$5E,$77,$F8,$6F,$E6,$1F
	.dc.b	$47,$FC,$92,$9E,$86,$7C,$79,$87,$39,$CB,$7E,$43,$81,$1F,$F2,$A6
	.dc.b	$1F,$E2,$66,$FB,$3D,$7C,$69,$18,$3A,$3E,$1F,$A3,$91,$83,$C0,$E1
	.dc.b	$F7,$3C,$9F,$81,$F1,$9E,$92,$3E,$2F,$95,$AF,$F1,$CC,$3F,$BB,$73
	.dc.b	$FF,$B7,$5B,$C3,$FD,$6F,$46,$1C,$2B,$5F,$2B,$F1,$CC,$56,$F0,$F9
	.dc.b	$04,$3E,$0F,$BB,$5F,$07,$B9,$F3,$CC,$5A,$B9,$4F,$C8,$9D,$3F,$3C
	.dc.b	$FF,$0D,$FC,$FF,$F3,$0A,$FF,$5D,$B6,$BF,$DA,$2D,$97,$96,$D3,$F0
	.dc.b	$17,$C3,$E2,$3F,$01,$C9,$2F,$FC,$35,$87,$14,$F3,$C3,$94,$EE,$7F
	.dc.b	$B6,$9A,$AF,$F3,$37,$D1,$BE,$52,$92,$7F,$07,$F2,$D5,$FA,$5C,$DF
	.dc.b	$91,$31,$9F,$F9,$62,$F9,$4F,$E2,$A7,$4F,$FA,$C3,$AD,$3A,$8F,$EF
	.dc.b	$F1,$8F,$6F,$C6,$48,$FC,$9F,$B9,$5F,$F0,$2F,$03,$2F,$F6,$AC,$7F
	.dc.b	$AF,$23,$1F,$0F,$BC,$FC,$98,$7F,$A1,$67,$35,$BB,$E2,$F8,$B2,$FF
	.dc.b	$7B,$3F,$E4,$5F,$9C,$7C,$8F,$F9,$7D,$B0,$21,$1C,$B8,$34,$E1,$C3
	.dc.b	$AC,$7F,$C5,$3E,$7F,$FA,$48,$FF,$30,$C6,$71,$ED,$C4,$FA,$3F,$F1
	.dc.b	$DD,$FE,$47,$0F,$C3,$C7,$FA,$AD,$FF,$00,$B6,$35,$3C,$7D,$9F,$F5
	.dc.b	$14,$F2,$BE,$AE,$DB,$ED,$6A,$7F,$87,$7C,$BC,$C7,$F8,$2B,$FC,$47
	.dc.b	$C0,$C4,$7E,$D6,$2E,$5F,$B5,$9F,$D0,$F9,$3F,$E5,$EF,$86,$7D,$07
	.dc.b	$E8,$23,$0D,$E6,$FF,$69,$BF,$37,$E8,$5F,$47,$F3,$2F,$8E,$7F,$47
	.dc.b	$B5,$8F,$28,$C3,$5F,$D1,$B9,$BE,$7C,$9F,$ED,$AC,$7C,$14,$C3,$AB
	.dc.b	$C3,$2F,$1B,$D1,$B7,$CA,$32,$6C,$E5,$5D,$2B,$9E,$3E,$6B,$E3,$97
	.dc.b	$AE,$B4,$4B,$EC,$72,$18,$77,$70,$77,$3F,$1D,$7D,$5F,$C8,$19,$3D
	.dc.b	$A7,$E4,$FE,$21,$EE,$BE,$8F,$72,$9C,$C6,$1B,$1D,$1A,$F8,$3D,$07
	.dc.b	$AD,$CB,$04,$62,$7C,$71,$89,$D4,$E6,$79,$47,$CA,$9D,$47,$C1,$C5
	.dc.b	$F4,$7F,$EC,$3E,$6D,$61,$D7,$7F,$9A,$DD,$7F,$95,$B7,$0F,$93,$7C
	.dc.b	$5F,$93,$8F,$6B,$E3,$97,$43,$EE,$D6,$E3,$35,$C2,$FA,$3F,$CB,$9E
	.dc.b	$D9,$79,$BD,$4F,$DA,$57,$02,$65,$E4,$C7,$AC,$EA,$31,$87,$DF,$E9
	.dc.b	$79,$DF,$EC,$FD,$7F,$D5,$7E,$BF,$EF,$F9,$FF,$D5,$FF,$47,$67,$F8
	.dc.b	$2D,$6F,$0F,$F4,$0B,$6B,$FD,$6B,$7B,$9F,$E9,$B2,$BC,$61,$E0,$AB
	.dc.b	$F0,$6E,$CB,$CF,$86,$E7,$FC,$87,$3F,$E1,$B5,$B8,$C1,$FC,$61,$FF
	.dc.b	$71,$38,$CF,$FA,$EC,$6D,$FF,$4B,$BA,$7F,$AA,$F7,$3E,$3F,$F8,$F4
	.dc.b	$7F,$28,$6B,$FC,$BD,$AF,$F2,$E6,$A8,$E5,$5F,$CE,$E7,$DF,$AB,$DA
	.dc.b	$F8,$97,$B9,$F8,$D9,$BE,$8B,$F3,$78,$7F,$4D,$FA,$37,$A7,$A9,$CE
	.dc.b	$1F,$D5,$27,$35,$FD,$7E,$73,$3F,$D1,$AF,$A7,$3F,$A1,$B5,$FB,$86
	.dc.b	$CF,$F4,$96,$ED,$FC,$FD,$B3,$F8,$0F,$A3,$F1,$5F,$A3,$E1,$BF,$13
	.dc.b	$3F,$EF,$2F,$7B,$1B,$FC,$06,$C7,$79,$DE,$5D,$DA,$F9,$DA,$FF,$5F
	.dc.b	$C7,$FD,$AB,$7C,$A3,$F1,$FE,$F5,$CF,$EB,$FA,$CF,$3F,$E7,$F5,$7C
	.dc.b	$E8,$C6,$7F,$83,$E8,$FC,$FC,$E6,$BE,$CD,$6B,$1F,$E3,$6D,$7F,$B6
	.dc.b	$35,$FE,$E2,$EF,$0F,$0C,$FD,$EF,$17,$A6,$3F,$E6,$BF,$A3,$7B,$DF
	.dc.b	$82,$E1,$FD,$27,$BC,$9C,$D7,$97,$7F,$BB,$97,$F6,$FA,$D6,$1F,$90
	.dc.b	$6D,$7F,$30,$6E,$9F,$E0,$16,$BF,$E0,$67,$35,$CF,$0E,$F3,$FB,$8F
	.dc.b	$7F,$C7,$18,$9F,$07,$D1,$7D,$67,$FB,$DE,$59,$C7,$FC,$25,$8A,$B5
	.dc.b	$1F,$C1,$32,$F6,$C7,$DA,$C9,$5B,$BF,$54,$E6,$31,$7E,$B8,$E7,$5C
	.dc.b	$1B,$C1,$F8,$E9,$FF,$4B,$46,$AB,$C5,$D1,$E0,$3C,$3E,$DD,$AD,$BE
	.dc.b	$02,$3D,$DF,$DD,$DB,$E8,$FF,$D3,$B3,$C0,$FC,$43,$8C,$5A,$7C,$6B
	.dc.b	$F7,$FB,$32,$F7,$A5,$BD,$C7,$C8,$D3,$FC,$94,$FB,$FD,$4F,$86,$47
	.dc.b	$E1,$7A,$FF,$28,$FE,$4F,$9B,$DF,$C4,$FF,$1D,$C3,$FD,$23,$E8,$EB
	.dc.b	$7F,$D0,$36,$75,$CE,$8F,$CF,$92,$3F,$B9,$C6,$5B,$1C,$9F,$D4,$7D
	.dc.b	$9B,$D9,$C6,$1E,$C3,$FD,$13,$93,$DE,$F3,$F0,$FA,$1B,$87,$CA,$6D
	.dc.b	$71,$E3,$FA,$11,$8E,$3F,$F2,$4E,$0F,$82,$F8,$3E,$FF,$C2,$6E,$3F
	.dc.b	$51,$B8,$F3,$AE,$13,$F6,$2B,$E4,$F1,$7C,$AE,$5B,$3F,$71,$B7,$C5
	.dc.b	$B8,$F8,$9B,$3F,$E6,$4C,$3F,$DB,$EF,$E1,$2F,$A3,$F9,$4F,$37,$EB
	.dc.b	$F6,$FE,$89,$B1,$92,$63,$E7,$E5,$D4,$7E,$3B,$8B,$EF,$FB,$BF,$80
	.dc.b	$6C,$64,$FC,$9C,$65,$A7,$DE,$FA,$4F,$C3,$FB,$3F,$C8,$5B,$19,$3E
	.dc.b	$FB,$FC,$80,$8E,$0D,$61,$F2,$FE,$0B,$94,$7E,$8B,$DF,$FC,$9F,$D2
	.dc.b	$7E,$51,$B1,$93,$F4,$F1,$96,$9F,$DC,$FB,$9F,$12,$FA,$1F,$CE,$F6
	.dc.b	$7F,$24,$8C,$57,$EC,$23,$13,$23,$F6,$79,$DC,$FF,$D9,$BF,$48,$FA
	.dc.b	$4F,$D3,$F1,$7F,$6E,$8C,$5F,$EE,$A3,$13,$FD,$E2,$73,$5F,$E5,$A6
	.dc.b	$BF,$EA,$B6,$BF,$EB,$07,$7C,$FF,$E0,$66,$DF,$E1,$67,$AB,$EF,$CD
	.dc.b	$F3,$9F,$E2,$0F,$98,$DD,$7F,$24,$9C,$D7,$F5,$59,$CE,$DF,$1A,$FE
	.dc.b	$21,$C7,$EA,$AD,$75,$F8,$1E,$9F,$07,$F2,$5B,$19,$26,$5F,$AF,$C5
	.dc.b	$F5,$3B,$9F,$37,$C7,$FF,$4D,$D0,$D2,$47,$F4,$F8,$CB,$C0,$F4,$71
	.dc.b	$7D,$99,$F7,$32,$3D,$9F,$4B,$71,$F1,$4D,$E0,$FF,$63,$E4,$AC,$5F
	.dc.b	$F4,$EC,$5F,$0F,$E5,$24,$78,$DB,$8F,$F1,$3E,$0F,$C6,$C7,$CF,$FA
	.dc.b	$AF,$C6,$3F,$E3,$0F,$A4,$FE,$55,$7F,$E9,$26,$DF,$FA,$69,$AF,$FB
	.dc.b	$AD,$AF,$FB,$C1,$DE,$7F,$E4,$66,$CF,$F6,$4F,$A5,$F3,$BF,$CD,$63
	.dc.b	$35,$FE,$F9,$6A,$73,$8A,$B8,$CF,$F2,$D3,$5E,$67,$A4,$97,$FA,$B9
	.dc.b	$DE,$1E,$39,$FF,$26,$FD,$23,$B3,$FC,$AA,$DB,$FF,$58,$B7,$0F,$F5
	.dc.b	$CB,$51,$03,$BB,$FF,$32,$B6,$76,$3E,$95,$D9,$E2,$C3,$FC,$BB,$F4
	.dc.b	$AF,$6F,$F4,$5E,$76,$1B,$87,$F9,$DF,$DA,$56,$E3,$FC,$AC,$DD,$3F
	.dc.b	$CA,$CD,$7F,$CB,$CD,$7F,$C9,$0E,$F4,$FF,$2D,$36,$7D,$BE,$77,$F5
	.dc.b	$7F,$91,$F1,$47,$86,$7F,$C6,$2E,$B9,$FF,$6B,$87,$F5,$FC,$7C,$BE
	.dc.b	$C7,$EE,$FB,$FD,$AF,$F9,$B7,$0F,$F3,$8D,$78,$A3,$04,$C3,$B3,$FC
	.dc.b	$0F,$E0,$38,$C6,$09,$3F,$B7,$8C,$BE,$1F,$A0,$FC,$5E,$8C,$DF,$4B
	.dc.b	$FE,$8A,$9C,$C6,$30,$F9,$91,$F9,$AB,$68,$8C,$37,$97,$FD,$0D,$18
	.dc.b	$3F,$33,$19,$AF,$E2,$DE,$07,$7A,$F0,$7A,$7C,$09,$3F,$E8,$DB,$FF
	.dc.b	$02,$37,$D2,$EA,$47,$07,$E3,$F1,$83,$E2,$BF,$D0,$5F,$0F,$86,$C6
	.dc.b	$DA,$FE,$5D,$BF,$1C,$FA,$4F,$F1,$BC,$FC,$0A,$FC,$E4,$E7,$9B,$75
	.dc.b	$AD,$BF,$DA,$57,$FA,$2A,$D3,$C4,$F9,$39,$FF,$85,$5F,$77,$AF,$E8
	.dc.b	$72,$DA,$F9,$CD,$97,$AF,$49,$5F,$1D,$D7,$A9,$FB,$1C,$BF,$D6,$54
	.dc.b	$8C,$6B,$FE,$A5,$8C,$5F,$F4,$9C,$F2,$F9,$AB,$9B,$F4,$DA,$FF,$AF
	.dc.b	$E7,$8B,$FD,$17,$C8,$E3,$CB,$F8,$D7,$9E,$9F,$87,$BB,$7F,$C3,$6F
	.dc.b	$F1,$BD,$B8,$38,$F2,$BE,$67,$13,$A2,$1F,$D4,$DC,$BE,$39,$F8,$C4
	.dc.b	$7F,$14,$F4,$FA,$39,$1C,$B6,$F8,$AF,$90,$87,$FA,$FF,$F1,$47,$D3
	.dc.b	$62,$3E,$01,$F4,$C8,$7E,$5C,$F4,$FD,$2C,$7F,$52,$7C,$C1,$F0,$C8
	.dc.b	$C3,$63,$FC,$E8,$D7,$F9,$D7,$D2,$BA,$77,$5F,$87,$9E,$6D,$AE,$11
	.dc.b	$FA,$36,$BF,$CC,$FE,$91,$F2,$3E,$36,$C7,$FA,$0D,$AF,$F4,$9F,$A5
	.dc.b	$FD,$FB,$99,$E0,$47,$FD,$15,$1A,$7F,$A8,$4F,$2D,$1C,$FE,$1B,$E1
	.dc.b	$5D,$79,$2B,$FB,$9B,$F0,$B6,$7D,$D1,$88,$FE,$F9,$3F,$96,$F3,$3F
	.dc.b	$C8,$6D,$F4,$3C,$58,$FF,$81,$3E,$53,$13,$FA,$DA,$DD,$FF,$84,$E3
	.dc.b	$17,$E9,$8F,$EB,$1D,$3F,$C9,$7B,$BF,$26,$48,$FF,$17,$7A,$15,$FF
	.dc.b	$12,$45,$CB,$FC,$27,$3E,$5F,$A5,$7D,$2D,$25,$FF,$8D,$EE,$BF,$EF
	.dc.b	$19,$DB,$FE,$F5,$6B,$99,$F5,$47,$F9,$8A,$3F,$CC,$8D,$7F,$C3,$4D
	.dc.b	$BF,$F1,$2E,$5E,$3C,$BF,$6B,$F4,$AF,$90,$A6,$70,$E7,$C1,$E9,$23
	.dc.b	$FD,$F3,$0F,$B6,$B5,$5B,$E8,$37,$3F,$EB,$58,$79,$DE,$03,$37,$CA
	.dc.b	$5F,$48,$F2,$BF,$4C,$D8,$FF,$6B,$35,$E9,$D7,$A9,$FD,$14,$E6,$7F
	.dc.b	$A1,$35,$FF,$09,$35,$FF,$0A,$BC,$EB,$D7,$97,$AB,$83,$E2,$9D,$6B
	.dc.b	$E6,$FA,$1F,$6C,$8F,$C3,$E7,$FF,$01,$E8,$FA,$6C,$CF,$C1,$FA,$6F
	.dc.b	$9E,$63,$2E,$F4,$FA,$2F,$A7,$C7,$2F,$A1,$F3,$67,$4E,$34,$F9,$F5
	.dc.b	$B8,$F9,$D5,$F7,$23,$E4,$F8,$72,$7E,$4B,$F8,$0F,$9C,$C4,$7F,$66
	.dc.b	$DF,$F9,$73,$FD,$45,$AE,$67,$E6,$9F,$D1,$3E,$0D,$F4,$DF,$E7,$E9
	.dc.b	$F3,$C6,$3C,$AE,$BA,$FB,$BE,$6F,$F1,$FD,$12,$FD,$5B,$A7,$F2,$66
	.dc.b	$63,$F1,$13,$87,$F3,$BC,$5F,$AB,$23,$CB,$E6,$6B,$F8,$FE,$4F,$CE
	.dc.b	$98,$C7,$E7,$F8,$7F,$A0,$F9,$99,$5E,$28,$FD,$19,$87,$F3,$11,$87
	.dc.b	$47,$A1,$D4,$F4,$8F,$5F,$AF,$FA,$CF,$99,$91,$82,$1F,$A3,$72,$FE
	.dc.b	$96,$3F,$67,$F2,$FF,$B4,$F9,$99,$3F,$1A,$32,$FC,$88,$CB,$FA,$D8
	.dc.b	$FD,$3F,$83,$23,$F4,$EF,$F7,$2F,$95,$FF,$0D,$46,$0F,$BD,$1F,$A8
	.dc.b	$23,$83,$EA,$ED,$FC,$18,$E8,$FF,$A9,$7E,$97,$FC,$30,$DF,$5B,$F9
	.dc.b	$94,$63,$4F,$D5,$F8,$6F,$A6,$CF,$AF,$FC,$C6,$BF,$C3,$38,$7F,$8D
	.dc.b	$BC,$A4,$62,$3F,$C7,$51,$89,$E1,$F2,$9D,$89,$3F,$A3,$C1,$F3,$5F
	.dc.b	$E1,$05,$EE,$F9,$3E,$56,$56,$E2,$BD,$B2,$4F,$94,$91,$E5,$87,$C9
	.dc.b	$FB,$89,$1F,$B6,$CF,$D3,$5F,$0B,$E1,$64,$E6,$FF,$D3,$6F,$9F,$E6
	.dc.b	$72,$F8,$67,$E2,$11,$F0,$3E,$BF,$99,$91,$CE,$3E,$7B,$0E,$C7,$C1
	.dc.b	$F8,$2C,$3E,$D3,$E8,$10,$F6,$1F,$1F,$D8,$8F,$23,$F8,$88,$FA,$2F
	.dc.b	$E7,$2E,$1A,$FF,$B3,$5B,$C7,$F8,$0E,$4F,$F0,$F4,$6A,$7F,$CE,$8D
	.dc.b	$7F,$D8,$8D,$7F,$D8,$ED,$7C,$1E,$8F,$DB,$E4,$91,$E1,$8F,$F2,$5F
	.dc.b	$D3,$3E,$66,$92,$7F,$09,$F3,$55,$FF,$42,$75,$2B,$84,$7F,$A1,$A3
	.dc.b	$15,$FE,$6F,$9C,$CF,$06,$E7,$F9,$CE,$8F,$C7,$1D,$5F,$EC,$77,$D7
	.dc.b	$E0,$3C,$A4,$BF,$CB,$7C,$1D,$7A,$47,$72,$3E,$3C,$3F,$CC,$8A,$FF
	.dc.b	$B7,$F6,$33,$7C,$5F,$F9,$DA,$30,$F5,$9F,$ED,$9F,$40,$F9,$5F,$F2
	.dc.b	$83,$70,$FB,$AD,$7B,$5F,$FB,$73,$B7,$DE,$FF,$BE,$67,$53,$FE,$AD
	.dc.b	$6B,$9A,$FF,$29,$46,$23,$FC,$09,$F4,$CF,$7B,$FD,$06,$FF,$A9,$23
	.dc.b	$B9,$AF,$7B,$33,$23,$FA,$F4,$EF,$2F,$EE,$AD,$C7,$FD,$E2,$F5,$9F
	.dc.b	$E0,$5F,$39,$FC,$04,$8F,$F0,$5E,$EA,$F8,$3D,$8F,$F0,$B4,$E3,$0F
	.dc.b	$E2,$F5,$F9,$8F,$27,$FC,$BF,$F1,$6F,$81,$FC,$BC,$61,$BC,$CC,$BF
	.dc.b	$D9,$D1,$7E,$8D,$F4,$DF,$37,$0E,$55,$F9,$EC,$39,$37,$0E,$53,$A8
	.dc.b	$FD,$84,$EA,$F9,$F1,$7F,$B5,$71,$F0,$D7,$80,$E4,$53,$8E,$1D,$23
	.dc.b	$C9,$8F,$8D,$E3,$8F,$7E,$D6,$5E,$4C,$38,$9D,$72,$F0,$BC,$CE,$F5
	.dc.b	$C6,$78,$57,$2C,$75,$7C,$2F,$C9,$39,$7F,$DE,$AD,$7F,$82,$D7,$F9
	.dc.b	$C5,$F5,$9F,$EF,$BF,$8D,$FF,$22,$7D,$37,$FB,$95,$B1,$FE,$FE,$6B
	.dc.b	$BF,$82,$4F,$FB,$AE,$DE,$83,$D9,$C3,$6F,$C6,$D7,$F8,$BB,$5F,$E2
	.dc.b	$EE,$F4,$F9,$0D,$D7,$F0,$1F,$4D,$FE,$BC,$E6,$F8,$D7,$FC,$A6,$B4
	.dc.b	$FF,$1E,$6B,$9C,$3E,$3D,$71,$8F,$8F,$3A,$C7,$D3,$5D,$1B,$3F,$56
	.dc.b	$B8,$63,$E4,$AD,$5F,$13,$C3,$B2,$B7,$97,$B2,$B6,$4C,$3A,$7A,$9F
	.dc.b	$E7,$EC,$7A,$47,$E3,$F9,$9F,$50,$F9,$AF,$F4,$1F,$3F,$F9,$AA,$31
	.dc.b	$BF,$F9,$5D,$7C,$E6,$A3,$F5,$17,$CA,$F1,$87,$FA,$2E,$F8,$BF,$B7
	.dc.b	$6E,$1D,$70,$DD,$F1,$C3,$B7,$CD,$E2,$DF,$39,$63,$0D,$9F,$23,$71
	.dc.b	$EC,$F8,$97,$7B,$BF,$BF,$81,$FE,$02,$BF,$E5,$31,$8F,$4B,$85,$6E
	.dc.b	$7F,$82,$5F,$F2,$7A,$FF,$4F,$3F,$CA,$DA,$FF,$09,$9E,$0F,$F8,$83
	.dc.b	$C2,$7C,$2F,$AF,$E6,$3E,$2F,$A9,$FD,$02,$31,$8F,$F8,$46,$B6,$FC
	.dc.b	$FA,$DC,$62,$FE,$54,$62,$A5,$7D,$5F,$97,$B5,$FC,$D3,$0F,$E9,$BF
	.dc.b	$0D,$F2,$FE,$5F,$9B,$6F,$C0,$DA,$E1,$92,$31,$C9,$FE,$C6,$C3,$9D
	.dc.b	$F0,$3F,$09,$DF,$D1,$F5,$7B,$FF,$67,$E5,$FD,$2D,$8C,$9F,$8B,$8C
	.dc.b	$B4,$CE,$5E,$29,$FC,$17,$C2,$EA,$3C,$BF,$01,$F8,$5F,$87,$F8,$AD
	.dc.b	$BF,$94,$D9,$DC,$7E,$66,$76,$E8,$FC,$E7,$CA,$E2,$7E,$47,$DC,$7F
	.dc.b	$94,$27,$F4,$1F,$29,$F6,$23,$13,$F9,$B8,$C7,$C0,$7E,$5D,$B5,$C0
	.dc.b	$FF,$26,$57,$0D,$FE,$53,$63,$F5,$3D,$9F,$D5,$5D,$DF,$E8,$7D,$19
	.dc.b	$1C,$BD,$AB,$97,$99,$B8,$F4,$6C,$F4,$EA,$F4,$47,$EA,$DB,$1F,$ED
	.dc.b	$1F,$C6,$5F,$25,$FB,$26,$C7,$EF,$FC,$FC,$4F,$C1,$F6,$BF,$69,$16
	.dc.b	$74,$73,$F6,$F9,$1C,$FC,$1B,$1F,$E6,$DF,$C6,$5F,$25,$FC,$03,$81
	.dc.b	$F1,$5F,$A4,$9C,$D7,$FB,$96,$72,$F1,$AE,$15,$B6,$C6,$4F,$F4,$95
	.dc.b	$72,$E0,$FF,$48,$B5,$FD,$DF,$62,$31,$1F,$C6,$B6,$E1,$C2,$BC,$7C
	.dc.b	$9F,$C5,$C7,$E3,$36,$3E,$53,$6F,$FC,$29,$68,$F9,$9F,$25,$F3,$7E
	.dc.b	$4E,$DF,$D1,$6D,$F0,$6E,$3E,$18,$79,$4C,$55,$F5,$FE,$4F,$EC,$78
	.dc.b	$D7,$1F,$B8,$DC,$3C,$6D,$C7,$A6,$3C,$E3,$15,$F0,$5B,$5E,$0C,$7F
	.dc.b	$88,$5F,$F0,$1F,$C0,$FF,$CA,$57,$F8,$9F,$13,$F8,$DC,$62,$FE,$0C
	.dc.b	$63,$DB,$FE,$4B,$C2,$AD,$57,$D2,$6B,$FE,$31,$8D,$4F,$F2,$CC,$FE
	.dc.b	$07,$17,$E5,$7E,$47,$F2,$CD,$CB,$F3,$4D,$C3,$E8,$B7,$3F,$95,$0F
	.dc.b	$A2,$FE,$A3,$5F,$D6,$39,$5F,$23,$FD,$1E,$73,$5F,$42,$73,$5F,$EA
	.dc.b	$A9,$CE,$5F,$ED,$4B,$E1,$E9,$7E,$6D,$B5,$FD,$23,$0F,$E9,$37,$FD
	.dc.b	$89,$C3,$5C,$D7,$BF,$83,$BC,$7E,$2B,$C7,$FD,$A6,$1F,$B5,$6E,$1F
	.dc.b	$B9,$6C,$EE,$3F,$DB,$5C,$0F,$33,$F7,$9E,$A7,$D6,$6C,$E9,$FA,$B3
	.dc.b	$B8,$FF,$7D,$36,$BF,$DF,$2D,$7D,$3E,$8F,$F0,$7F,$67,$E4,$36,$BE
	.dc.b	$33,$67,$FD,$0A,$7D,$17,$1C,$1F,$53,$6F,$EF,$B5,$F1,$EB,$FC,$E5
	.dc.b	$18,$70,$D7,$C3,$1B,$F6,$DE,$03,$C8,$FE,$35,$F2,$3F,$E1,$29,$CC
	.dc.b	$62,$BF,$CF,$D8,$FE,$B6,$73,$1C,$EF,$93,$6B,$FB,$EC,$EB,$0F,$87
	.dc.b	$E5,$FF,$86,$3F,$1C,$7C,$8F,$E8,$A3,$0D,$C3,$E5,$C7,$CE,$23,$EF
	.dc.b	$A7,$E3,$8F,$A6,$F4,$7E,$39,$FE,$E3,$AF,$F7,$CC,$E5,$F4,$9F,$E9
	.dc.b	$7C,$93,$BB,$F9,$F5,$B7,$E6,$35,$FC,$A3,$5F,$F0,$EB,$5F,$F5,$16
	.dc.b	$3C,$F5,$FD,$4B,$70,$FF,$1F,$F8,$6F,$8A,$FF,$17,$4E,$6F,$8F,$95
	.dc.b	$83,$FC,$7B,$5B,$BF,$4C,$E9,$AB,$E9,$7F,$B2,$3F,$EE,$7F,$AF,$79
	.dc.b	$9F,$E3,$B9,$CE,$69,$CD,$F2,$9F,$EE,$B7,$FB,$16,$BF,$E3,$A6,$BF
	.dc.b	$EB,$F6,$F8,$9F,$C5,$7D,$3F,$F2,$2F,$D4,$FF,$1C,$6D,$FE,$B1,$FD
	.dc.b	$39,$1F,$D3,$B6,$FF,$C6,$1D,$68,$FC,$37,$87,$CE,$B9,$12,$BF,$DB
	.dc.b	$5C,$DF,$03,$1F,$14,$FE,$EB,$D0,$FF,$2B,$F1,$7F,$2E,$6C,$64,$FD
	.dc.b	$BB,$FB,$A2,$3F,$AB,$E9,$FF,$3E,$FB,$5F,$97,$F9,$4F,$F3,$AC,$62
	.dc.b	$7E,$9F,$8C,$C9,$5A,$F5,$58,$71,$BF,$B3,$87,$1D,$3D,$37,$BA,$FF
	.dc.b	$31,$CE,$67,$FC,$D2,$D7,$FD,$DC,$D7,$F7,$36,$EC,$FD,$67,$C9,$78
	.dc.b	$5B,$8F,$82,$74,$FA,$32,$56,$E3,$0D,$B7,$2F,$F6,$BC,$7F,$D3,$F8
	.dc.b	$FE,$E7,$D5,$D4,$E7,$2E,$B7,$C3,$1F,$3C,$62,$7F,$B4,$E4,$BE,$35
	.dc.b	$AC,$7F,$82,$7C,$97,$C9,$6C,$7A,$5B,$8F,$53,$E0,$B8,$C3,$C4,$F1
	.dc.b	$9F,$77,$C7,$DA,$E1,$3F,$E0,$3E,$9F,$F4,$DF,$CA,$7D,$CF,$2F,$F4
	.dc.b	$75,$D7,$89,$EB,$AB,$5E,$AF,$FA,$8B,$C4,$BA,$7E,$0A,$BF,$84,$6B
	.dc.b	$FE,$9C,$F8,$5F,$AF,$CF,$FC,$D5,$B9,$FF,$5D,$6B,$FE,$A6,$6F,$C5
	.dc.b	$BD,$FF,$A3,$3F,$CD,$AB,$FD,$57,$39,$BD,$D6,$B8,$3F,$D6,$D6,$15
	.dc.b	$B7,$91,$FD,$03,$7F,$F2,$5B,$E9,$F8,$7F,$F5,$BE,$1F,$E1,$EC,$D1
	.dc.b	$8D,$3F,$23,$7C,$2F,$E5,$BF,$E2,$C7,$E1,$1B,$3F,$C6,$35,$FD,$EE
	.dc.b	$FF,$15,$3B,$9E,$3E,$C6,$F2,$FF,$1B,$36,$4F,$3E,$1F,$D7,$7C,$9E
	.dc.b	$63,$1C,$EF,$93,$A6,$71,$FD,$7E,$1C,$9F,$D0,$46,$9C,$D7,$7F,$52
	.dc.b	$47,$07,$AF,$91,$88,$F7,$C5,$D3,$F3,$73,$76,$FF,$7D,$E4,$F9,$0F
	.dc.b	$29,$9C,$3E,$86,$18,$7F,$65,$19,$DB,$FC,$AD,$F0,$5F,$D7,$5B,$3F
	.dc.b	$C4,$F5,$4E,$5B,$19,$8F,$87,$F1,$12,$3E,$0C,$7C,$B3,$0F,$F2,$88
	.dc.b	$C3,$FD,$33,$C4,$7E,$3B,$FE,$9C,$8F,$F0,$64,$63,$E0,$BF,$B8,$36
	.dc.b	$33,$3B,$7F,$C7,$11,$97,$F5,$5E,$0B,$EA,$BE,$D3,$63,$C1,$5C,$78
	.dc.b	$7F,$DC,$CF,$FA,$06,$33,$DB,$F8,$BC,$3C,$AF,$E0,$3C,$7C,$56,$EB
	.dc.b	$FD,$F0,$DB,$E7,$5B,$9E,$67,$A1,$F5,$CF,$18,$FF,$4E,$B5,$FF,$36
	.dc.b	$3C,$F8,$BF,$D9,$FF,$55,$EA,$3C,$35,$F9,$FA,$7D,$0B,$E7,$7F,$C2
	.dc.b	$D1,$86,$C6,$63,$FC,$34,$D7,$FC,$74,$DC,$7F,$CE,$3F,$44,$F8,$8E
	.dc.b	$CD,$AF,$F0,$0D,$6F,$2F,$F0,$E7,$82,$F8,$0F,$D9,$7B,$08,$C5,$FF
	.dc.b	$82,$A3,$13,$FA,$DF,$C7,$BF,$88,$4E,$6B,$6F,$F2,$FA,$DB,$9A,$E1
	.dc.b	$E2,$7C,$1F,$C7,$FF,$A9,$4F,$F8,$D5,$AF,$F9,$55,$AF,$F9,$F5,$AF
	.dc.b	$F1,$70,$F0,$D7,$FC,$9D,$F9,$07,$EF,$67,$D0,$D7,$FC,$CB,$3B,$CF
	.dc.b	$EB,$FB,$1F,$70,$93,$F9,$66,$BF,$E0,$8F,$27,$F8,$49,$F8,$A8,$F8
	.dc.b	$FF,$6D,$FA,$C9,$FF,$49,$68,$8C,$47,$9E,$31,$D9,$FE,$53,$6C,$7F
	.dc.b	$9B,$67,$96,$7F,$D8,$23,$F8,$B9,$DA,$BF,$11,$F4,$AF,$94,$FC,$3B
	.dc.b	$67,$FC,$BC,$D7,$FC,$FC,$D7,$FD,$4C,$E6,$BF,$DA,$BC,$E3,$F9,$41
	.dc.b	$67,$FC,$8F,$3B,$D3,$E8,$36,$3F,$DC,$F3,$B7,$FB,$13,$6B,$FD,$87
	.dc.b	$3B,$EA,$FE,$0F,$F1,$1F,$AA,$25,$FF,$04,$6B,$CB,$E5,$7C,$67,$EE
	.dc.b	$23,$CB,$F0,$9E,$6E,$49,$5F,$E2,$CC,$93,$A8,$C4,$E9,$F2,$C7,$CA
	.dc.b	$F9,$3F,$92,$48,$F9,$2D,$7F,$BE,$7C,$EB,$1F,$E9,$8D,$8F,$8B,$3C
	.dc.b	$DF,$F1,$EC,$65,$E3,$39,$7E,$2C,$FC,$ED,$A4,$7A,$63,$E8,$98,$7E
	.dc.b	$5C,$61,$FE,$43,$BF,$F1,$22,$3F,$CB,$1A,$FD,$2F,$88,$FB,$67,$F2
	.dc.b	$08,$FB,$BF,$29,$F7,$BD,$B9,$5F,$26,$32,$FC,$B8,$CB,$F7,$23,$2F
	.dc.b	$DE,$8C,$F0,$FE,$2A,$7F,$13,$F2,$9F,$8A,$EA,$97,$F5,$63,$2F,$D7
	.dc.b	$8C,$BF,$88,$8F,$C8,$FB,$FF,$92,$F7,$24,$FD,$D8,$CB,$F7,$E3,$2F
	.dc.b	$E4,$23,$FA,$07,$E2,$6F,$90,$FE,$A2,$D9,$FE,$AA,$D7,$FA,$8B,$5F
	.dc.b	$E1,$9C,$C8,$D6,$6E,$8F,$F1,$37,$AA,$E4,$FC,$7E,$1F,$80,$E5,$F0
	.dc.b	$E5,$23,$04,$3F,$CA,$59,$75,$8F,$D7,$FC,$C4,$9F,$CD,$3F,$B2,$F8
	.dc.b	$AF,$EB,$5E,$FF,$F9,$86,$7F,$3D,$E3,$FE,$D8,$CF,$C0,$7E,$E3,$C5
	.dc.b	$93,$FA,$37,$4F,$E9,$7D,$2E,$7E,$B4,$8C,$1F,$DE,$F0,$47,$F8,$A7
	.dc.b	$F2,$2F,$ED,$DC,$3F,$01,$B4,$D9,$7C,$23,$FD,$3B,$3A,$E1,$34,$E3
	.dc.b	$5C,$30,$F6,$F9,$F2,$FC,$4F,$87,$DB,$C4,$7F,$9B,$6F,$FD,$AA,$F6
	.dc.b	$F8,$0C,$F2,$FD,$58,$F2,$67,$FE,$E7,$74,$43,$E8,$B9,$7E,$9F,$E4
	.dc.b	$5F,$E0,$D6,$FA,$9F,$E0,$98,$C5,$75,$D7,$D7,$E9,$66,$BE,$37,$CB
	.dc.b	$4C,$3F,$C4,$98,$70,$E0,$FF,$36,$70,$23,$12,$C7,$E2,$DF,$18,$F0
	.dc.b	$9C,$BA,$3C,$44,$7E,$6F,$CB,$57,$E2,$79,$BF,$5C,$C6,$BF,$0F,$2F
	.dc.b	$05,$E6,$BE,$AD,$6B,$3F,$2E,$6F,$13,$E0,$12,$BE,$DF,$CB,$57,$E9
	.dc.b	$7B,$DF,$85,$31,$9F,$7B,$E5,$3F,$82,$9D,$3F,$12,$FE,$67,$85,$23
	.dc.b	$E3,$BF,$8F,$EE,$C4,$7D,$FF,$5B,$F2,$46,$31,$ED,$58,$9F,$C7,$CE
	.dc.b	$73,$FF,$02,$7C,$D3,$C3,$FF,$15,$B6,$BF,$C1,$06,$A3,$24,$7F,$DE
	.dc.b	$BE,$CB,$E5,$FF,$CA,$27,$F8,$DE,$B9,$4F,$E4,$FE,$5B,$F4,$DF,$2F
	.dc.b	$23,$F4,$31,$FA,$93,$0F,$E6,$63,$0F,$E9,$BE,$5F,$F5,$C4,$7F,$53
	.dc.b	$F0,$55,$FD,$04,$5D,$BF,$25,$DC,$AF,$F2,$57,$CE,$BE,$79,$E2,$7F
	.dc.b	$AE,$FE,$29,$E5,$7F,$89,$EB,$53,$98,$C5,$70,$CD,$7C,$17,$CF,$3E
	.dc.b	$7F,$E2,$47,$13,$FD,$91,$8F,$47,$8E,$1D,$31,$D4,$27,$58,$75,$C3
	.dc.b	$A7,$0F,$FB,$6F,$0E,$17,$C8,$EB,$97,$96,$FC,$91,$CE,$31,$3F,$ED
	.dc.b	$9B,$E7,$3E,$0C,$3B,$CF,$4B,$E3,$18,$9F,$6E,$5E,$C7,$C0,$75,$CB
	.dc.b	$96,$3D,$A3,$C7,$1F,$07,$1E,$F1,$C2,$FE,$13,$F8,$E6,$BE,$C3,$15
	.dc.b	$F3,$DF,$C9,$FC,$63,$D2,$7B,$BF,$96,$9F,$F7,$A4,$E6,$31,$1F,$EF
	.dc.b	$69,$DB,$B9,$D3,$C6,$DF,$37,$FE,$F4,$8C,$35,$F3,$35,$FC,$33,$5F
	.dc.b	$BA,$F1,$D3,$E4,$B7,$5F,$A1,$F3,$EE,$3F,$E6,$55,$AF,$8E,$FA,$57
	.dc.b	$BC,$39,$C7,$C9,$BE,$1D,$5A,$89,$3F,$2E,$B8,$6E,$BD,$CE,$6B,$78
	.dc.b	$77,$8F,$C4,$56,$F1,$F1,$E5,$EC,$BD,$B6,$B7,$87,$4C,$39,$1F,$A1
	.dc.b	$F1,$DD,$5F,$EA,$5D,$3F,$82,$FB,$9F,$D3,$FA,$3F,$A2,$4F,$F9,$16
	.dc.b	$72,$4C,$7F,$DB,$75,$B8,$FF,$04,$CE,$E7,$31,$FC,$43,$DA,$D4,$7F
	.dc.b	$45,$9F,$F2,$DC,$E4,$FE,$2B,$E3,$6B,$1F,$2D,$DA,$DD,$FF,$BA,$A3
	.dc.b	$15,$D2,$1D,$BF,$EB,$DC,$3F,$16,$D8,$FC,$83,$5F,$E4,$ED,$7F,$96
	.dc.b	$FA,$3C,$E3,$86,$7F,$0B,$DB,$C4,$FA,$BC,$94,$3C,$8F,$37,$CF,$E6
	.dc.b	$31,$97,$F3,$08,$C4,$FF,$7B,$CB,$FC,$67,$5A,$7F,$9C,$DF,$49,$D4
	.dc.b	$7F,$6E,$9D,$3C,$23,$9D,$FF,$3C,$F1,$EC,$FF,$8C,$AB,$FC,$8D,$3F
	.dc.b	$D0,$3D,$BF,$27,$91,$F2,$DB,$1F,$B3,$C7,$D7,$3E,$0B,$FF,$11,$5F
	.dc.b	$27,$CA,$D7,$D3,$1C,$96,$9F,$88,$EC,$C3,$FD,$DE,$9E,$2F,$DB,$F6
	.dc.b	$FE,$33,$72,$61,$E9,$8F,$16,$5F,$39,$B7,$CA,$BD,$0D,$AF,$2C,$F4
	.dc.b	$9F,$EB,$FE,$4D,$9F,$89,$EE,$7E,$2F,$DC,$FA,$CD,$AF,$0E,$7E,$78
	.dc.b	$EE,$48,$EE,$DA,$E7,$1F,$DB,$3C,$AE,$6F,$23,$FC,$4B,$F3,$EF,$95
	.dc.b	$FC,$D7,$B9,$F8,$98,$C4,$FF,$74,$8C,$4F,$C0,$9C,$DF,$E1,$1A,$FF
	.dc.b	$9E,$9A,$F6,$8E,$35,$FD,$FB,$BE,$CF,$D4,$8F,$DA,$FB,$9F,$AB,$3C
	.dc.b	$5E,$E5,$8F,$44,$FF,$98,$70,$4F,$8C,$93,$F8,$78,$CB,$FC,$EF,$BB
	.dc.b	$F8,$5F,$57,$48,$FF,$07,$B6,$FC,$47,$F9,$17,$E3,$5F,$19,$FB,$F6
	.dc.b	$D7,$F8,$5B,$CC,$C4,$FC,$7F,$7B,$E2,$12,$7C,$4F,$E5,$F1,$5B,$DC
	.dc.b	$E8,$75,$F7,$6D,$FB,$BA,$5F,$3E,$F6,$FF,$88,$71,$BD,$6F,$DF,$4E
	.dc.b	$6F,$E2,$E7,$F6,$E7,$37,$F9,$6A,$D6,$FF,$BA,$6B,$FE,$B0,$6B,$FE
	.dc.b	$68,$77,$D1,$FE,$5B,$6E,$1F,$EB,$2C,$7F,$39,$39,$BF,$8C,$7C,$7F
	.dc.b	$76,$D7,$93,$64,$63,$1F,$9F,$9F,$C4,$8C,$61,$FE,$7F,$AD,$DF,$1F
	.dc.b	$6F,$E9,$7B,$9F,$4D,$B1,$93,$B4,$65,$A7,$87,$DB,$FB,$5E,$37,$DB
	.dc.b	$6C,$64,$F3,$46,$5A,$67,$3F,$3E,$1F,$C8,$F1,$F3,$5F,$29,$FF,$66
	.dc.b	$FC,$BF,$F9,$66,$FE,$07,$8B,$FC,$CE,$31,$3C,$A3,$1E,$CC,$AF,$8D
	.dc.b	$EE,$5A,$EF,$C7,$F9,$1F,$85,$FC,$93,$6B,$E6,$36,$77,$1F,$3E,$7F
	.dc.b	$A8,$7C,$FF,$F6,$99,$F9,$FC,$9F,$E5,$9A,$FD,$5C,$27,$5A,$FF,$AB
	.dc.b	$EB,$53,$FA,$43,$5E,$8F,$F3,$16,$DF,$F8,$92,$FB,$47,$F9,$F9,$AF
	.dc.b	$F2,$83,$67,$37,$BC,$7E,$2F,$DB,$F3,$D7,$EB,$FD,$DF,$D8,$79,$36
	.dc.b	$BF,$14,$DC,$3E,$9B,$72,$FC,$43,$F3,$FC,$EF,$DC,$7A,$F6,$FF,$24
	.dc.b	$DB,$FA,$F3,$FB,$DC,$B1,$87,$F6,$1C,$FF,$BD,$BF,$93,$F6,$7E,$5F
	.dc.b	$A7,$FC,$0D,$B9,$7C,$18,$F8,$95,$FE,$02,$C7,$E1,$46,$0D,$BF,$8A
	.dc.b	$F2,$EE,$1F,$EE,$9F,$73,$FD,$13,$E5,$FF,$0F,$F9,$FF,$F0,$9C,$63
	.dc.b	$2F,$D5,$46,$3C,$75,$53,$F8,$B4,$D3,$F8,$57,$4B,$CE,$FF,$0C,$7C
	.dc.b	$9B,$E7,$7F,$29,$18,$6D,$70,$8D,$FD,$2E,$36,$FC,$17,$CF,$BE,$3B
	.dc.b	$F0,$4D,$BF,$F3,$2B,$67,$FC,$D4,$D7,$FD,$82,$E7,$7F,$F0,$73,$67
	.dc.b	$F0,$D8,$BC,$CF,$F1,$44,$E6,$FF,$B8,$CE,$6B,$F3,$2D,$8F,$E1,$7E
	.dc.b	$2B,$E2,$3F,$C6,$13,$9C,$7F,$C3,$8D,$BE,$F8,$FF,$8E,$EF,$84,$74
	.dc.b	$8F,$F1,$C7,$CF,$7F,$5A,$C7,$FC,$02,$D4,$D2,$77,$70,$3F,$C6,$5F
	.dc.b	$62,$F5,$31,$87,$D4,$F5,$BF,$C1,$3C,$7F,$EE,$78,$FF,$0F,$37,$1F
	.dc.b	$F1,$7B,$5F,$D5,$79,$E8,$D6,$CA,$FF,$0A,$CE,$6F,$EF,$E5,$FC,$45
	.dc.b	$93,$E1,$3F,$A3,$F2,$7F,$7A,$BE,$58,$B0,$E5,$5B,$8F,$E3,$B7,$D2
	.dc.b	$31,$5C,$2B,$C9,$E9,$FB,$09,$3E,$C6,$DF,$F2,$67,$F2,$7F,$4B,$FC
	.dc.b	$95,$FF,$38,$F4,$7F,$A0,$3D,$CF,$CB,$FC,$FE,$3E,$05,$18,$C7,$FC
	.dc.b	$07,$5B,$7F,$D7,$0D,$C7,$F8,$81,$FE,$F4,$AE,$33,$A6,$C6,$A7,$86
	.dc.b	$7F,$09,$B5,$FE,$40,$F4,$9B,$3E,$97,$C9,$97,$97,$B3,$F5,$BF,$41
	.dc.b	$70,$F9,$AD,$CF,$B6,$9F,$B3,$9F,$15,$7F,$96,$F3,$FF,$4F,$B6,$75
	.dc.b	$5F,$E1,$49,$CD,$7D,$8F,$6F,$A6,$BE,$7B,$DB,$70,$FD,$FC,$EA,$BF
	.dc.b	$CB,$37,$FE,$AC,$3F,$81,$FC,$77,$CD,$6C,$7C,$CC,$11,$E3,$24,$7C
	.dc.b	$96,$E3,$E5,$9D,$CF,$E3,$3E,$83,$A3,$F5,$BD,$4F,$F1,$03,$F5,$9B
	.dc.b	$5F,$45,$AF,$F4,$7B,$FF,$13,$FB,$7F,$16,$7F,$C0,$73,$BF,$63,$04
	.dc.b	$BF,$F7,$AC,$7F,$9D,$BE,$1B,$DF,$CB,$FE,$FB,$6C,$7E,$11,$AF,$F5
	.dc.b	$56,$FC,$5B,$E0,$FF,$D0,$F3,$F0,$AB,$FA,$D4,$E7,$87,$75,$AD,$15
	.dc.b	$A8,$C3,$F4,$33,$FF,$1D,$61,$F0,$7D,$CF,$CA,$B9,$6B,$F4,$23,$6F
	.dc.b	$C5,$F7,$33,$87,$F9,$29,$B2,$69,$8F,$6F,$FB,$86,$5F,$9B,$94,$63
	.dc.b	$5F,$23,$70,$74,$FE,$7B,$E7,$F1,$97,$F9,$62,$BA,$57,$D9,$6B,$F6
	.dc.b	$5E,$55,$FA,$6F,$9F,$BD,$67,$9E,$5E,$2B,$FC,$27,$AF,$82,$3F,$C5
	.dc.b	$3C,$6A,$FE,$86,$53,$8D,$7F,$8B,$E3,$FB,$4B,$DD,$FE,$26,$F6,$FE
	.dc.b	$3E,$FE,$39,$1F,$E4,$3F,$3F,$E6,$3A,$6D,$F5,$FD,$1F,$39,$27,$F9
	.dc.b	$3F,$D2,$57,$F3,$11,$F4,$88,$FE,$75,$F8,$3F,$41,$23,$F4,$51,$F3
	.dc.b	$FC,$47,$D0,$7F,$A3,$A7,$FD,$00,$D7,$FD,$27,$1A,$F8,$5F,$97,$F4
	.dc.b	$5C,$6B,$55,$FD,$11,$AF,$F4,$26,$BF,$CE,$DB,$5E,$0E,$CE,$44,$9F
	.dc.b	$E8,$7E,$85,$7C,$35,$FE,$78,$7C,$B3,$F7,$27,$5C,$5F,$D0,$6B,$55
	.dc.b	$C2,$B4,$F0,$3F,$DF,$51,$A7,$D1,$1F,$A0,$BF,$24,$7E,$8B,$E8,$1F
	.dc.b	$05,$F9,$58,$46,$1F,$F2,$CB,$5F,$F2,$B7,$D0,$FF,$18,$73,$3F,$C1
	.dc.b	$27,$51,$FD,$11,$AF,$F9,$55,$AB,$D9,$FD,$DF,$C0,$FE,$95,$B1,$FE
	.dc.b	$D2,$6B,$FE,$9A,$FA,$1F,$E1,$0E,$7C,$05,$71,$6C,$66,$3F,$4F,$3C
	.dc.b	$23,$FC,$DD,$F4,$1F,$E3,$77,$FC,$EA,$D7,$3E,$57,$3A,$F8,$93,$99
	.dc.b	$FE,$BE,$D7,$FC,$E0,$D4,$1C,$69,$F9,$2F,$5B,$F1,$84,$9F,$F0,$2F
	.dc.b	$D1,$57,$FD,$6F,$DF,$FE,$41,$7F,$C1,$3E,$D6,$78,$FF,$CB,$5F,$42
	.dc.b	$E8,$F9,$0D,$AF,$90,$D7,$9F,$CB,$6A,$7F,$83,$C4,$36,$FE,$07,$E8
	.dc.b	$5F,$2D,$49,$3F,$EE,$0F,$A1,$7C,$F6,$23,$F9,$6D,$70,$7F,$96,$B5
	.dc.b	$CD,$7F,$48,$8C,$65,$FC,$32,$C7,$F1,$C2,$D7,$FA,$A2,$77,$E0,$7E
	.dc.b	$BF,$E2,$FF,$68,$49,$F2,$7C,$6D,$7F,$DB,$F6,$FA,$1E,$86,$71,$F0
	.dc.b	$7C,$3F,$7E,$69,$7E,$D8,$F1,$18,$79,$C6,$1F,$27,$CB,$7C,$72,$3F
	.dc.b	$80,$AF,$E4,$FF,$40,$CB,$33,$F1,$AB,$E1,$11,$F8,$EF,$F3,$C9,$F4
	.dc.b	$47,$3D,$3F,$C4,$91,$88,$FF,$63,$6A,$AF,$90,$7C,$9D,$A4,$FD,$78
	.dc.b	$CB,$F6,$63,$2F,$C6,$8F,$E3,$DC,$12,$3F,$92,$BF,$4F,$E6,$71,$1F
	.dc.b	$33,$E3,$BE,$B1,$8D,$3E,$D7,$CC,$FD,$9F,$7A,$57,$E1,$E3,$EE,$18
	.dc.b	$7E,$AC,$61,$D1,$F8,$D7,$53,$D2,$3F,$21,$F3,$3F,$85,$F5,$E4,$60
	.dc.b	$87,$E0,$1C,$BF,$82,$8F,$C6,$7C,$1F,$E3,$7C,$09,$3F,$9B,$8C,$BF
	.dc.b	$9D,$8C,$BF,$8A,$8F,$E9,$5F,$42,$F9,$5A,$49,$FA,$C7,$F7,$CE,$2F
	.dc.b	$E2,$2F,$F4,$C9,$E0,$FF,$84,$BC,$14,$62,$3F,$CD,$31,$8A,$FE,$6D
	.dc.b	$39,$9F,$F4,$9B,$BC,$FF,$1B,$CF,$F8,$73,$C5,$3F,$A0,$F9,$AF,$D5
	.dc.b	$11,$FD,$C7,$C7,$FF,$63,$F8,$27,$C1,$FF,$43,$6C,$F9,$5A,$F9,$1A
	.dc.b	$FF,$1C,$3F,$C3,$3D,$BF,$E3,$27,$F8,$27,$91,$FB,$93,$F7,$5F,$15
	.dc.b	$FB,$BF,$73,$C1,$EC,$CA,$FE,$15,$15,$ED,$1E,$32,$3F,$E3,$8F,$69
	.dc.b	$F4,$1F,$DD,$F7,$FF,$09,$B7,$C0,$FF,$44,$36,$BF,$D5,$9A,$7F,$B7
	.dc.b	$5F,$AB,$C6,$8F,$F6,$C4,$61,$AF,$F8,$09,$B7,$DB,$E6,$BE,$47,$AB
	.dc.b	$F1,$C9,$5F,$1D,$AA,$7C,$BF,$09,$FE,$36,$C4,$F2,$9F,$1F,$D0,$FF
	.dc.b	$2A,$6F,$93,$FE,$17,$8C,$47,$3D,$3D,$BE,$B6,$6B,$FB,$D5,$6A,$BF
	.dc.b	$DA,$ED,$7F,$DA,$AD,$7F,$D9,$CF,$4C,$7E,$4F,$C1,$F9,$4E,$93,$E4
	.dc.b	$E8,$F1,$11,$FC,$EF,$07,$D2,$F8,$3C,$47,$E6,$30,$57,$90,$91,$F9
	.dc.b	$F8,$CB,$CB,$57,$A5,$F1,$4E,$79,$7F,$B7,$69,$E2,$7C,$32,$5F,$1F
	.dc.b	$9A,$AF,$86,$3E,$19,$1F,$03,$F4,$7E,$0E,$47,$D0,$8F,$9D,$E1,$C8
	.dc.b	$FA,$6F,$D7,$F0,$31,$1E,$D8,$F9,$E4,$7D,$8F,$DD,$F6,$64,$72,$3F
	.dc.b	$81,$8F,$9D,$78,$9F,$E2,$D6,$C6,$9A,$FD,$F3,$FD,$0B,$B3,$D1,$E1
	.dc.b	$5F,$EB,$0A,$DD,$7F,$BC,$1A,$FE,$D5,$AF,$2F,$9A,$FC,$97,$CD,$C8
	.dc.b	$FA,$51,$F9,$53,$0F,$87,$93,$F5,$F3,$98,$FB,$1F,$37,$F9,$B2,$3F
	.dc.b	$6F,$E1,$6B,$C7,$34,$E3,$8F,$F6,$F5,$FE,$4D,$FA,$07,$CE,$7F,$92
	.dc.b	$7E,$31,$E5,$7F,$0D,$9C,$F8,$BF,$EC,$D8,$C5,$6D,$FF,$16,$FD,$03
	.dc.b	$E8,$7F,$B7,$6B,$B3,$FC,$63,$87,$25,$8F,$67,$38,$73,$6E,$3C,$70
	.dc.b	$E9,$EC,$7F,$B5,$F2,$F0,$CF,$5A,$63,$D7,$1E,$91,$8C,$7A,$63,$CA
	.dc.b	$38,$56,$F0,$E7,$1E,$46,$DF,$96,$B8,$4E,$4F,$29,$4E,$D8,$74,$C3
	.dc.b	$BE,$5F,$0A,$75,$8F,$C1,$7F,$9C,$1C,$0F,$F7,$1B,$97,$C6,$6B,$E3
	.dc.b	$9E,$A7,$F9,$D6,$3F,$DD,$31,$8F,$83,$FF,$74,$B6,$33,$3B,$7B,$D3
	.dc.b	$E7,$9E,$F7,$FB,$C2,$30,$D8,$CC,$F9,$1A,$F2,$73,$AF,$CE,$AE,$33
	.dc.b	$F1,$75,$7D,$0F,$E4,$3C,$8F,$9F,$87,$0E,$0F,$C6,$36,$FF,$0A,$DB
	.dc.b	$E3,$C8,$BE,$39,$CB,$F9,$77,$BC,$7B,$46,$27,$59,$F9,$6B,$58,$73
	.dc.b	$AD,$12,$75,$5F,$8F,$BE,$71,$89,$FC,$07,$3F,$F6,$0C,$39,$72,$FF
	.dc.b	$60,$6E,$5F,$D8,$38,$AD,$FF,$08,$9B,$7C,$6B,$F9,$84,$63,$9B,$F6
	.dc.b	$31,$83,$F9,$A7,$CF,$67,$DB,$D3,$63,$27,$F3,$98,$CB,$4F,$ED,$BF
	.dc.b	$91,$FC,$4C,$7F,$C0,$DE,$82,$31,$1F,$EA,$48,$C7,$07,$C5,$8D,$CE
	.dc.b	$94,$EA,$BA,$D7,$B3,$EE,$1E,$17,$C0,$AD,$5F,$FB,$EB,$0F,$F5,$26
	.dc.b	$1F,$D3,$CD,$38,$8F,$E2,$D1,$8C,$7C,$51,$E5,$9D,$3E,$48,$CB,$E4
	.dc.b	$8C,$BB,$8E,$53,$FE,$42,$1F,$71,$FE,$2E,$7E,$2C,$78,$9F,$AB,$F9
	.dc.b	$2B,$E8,$31,$1F,$DB,$A3,$0F,$CC,$3C,$38,$FC,$B6,$E3,$F9,$77,$FC
	.dc.b	$98,$DC,$7F,$C9,$6D,$E2,$FC,$95,$71,$AF,$EF,$7F,$3A,$F3,$FC,$38
	.dc.b	$7F,$51,$A7,$E4,$AF,$9C,$F3,$FD,$1F,$CF,$8C,$63,$E7,$8C,$4F,$9A
	.dc.b	$73,$7F,$E5,$C6,$BF,$84,$6B,$F8,$37,$79,$79,$9B,$7F,$EE,$AE,$66
	.dc.b	$7F,$E0,$88,$FF,$1D,$CE,$AF,$F0,$EB,$E6,$7E,$6D,$79,$DB,$5C,$1F
	.dc.b	$31,$1F,$F3,$A4,$F1,$6C,$71,$7E,$3F,$9D,$6F,$E3,$53,$D7,$FB,$7F
	.dc.b	$3F,$F1,$5B,$7E,$67,$E2,$11,$EE,$DA,$E6,$FE,$E3,$17,$9D,$CE,$BF
	.dc.b	$09,$F3,$FF,$88,$3F,$41,$97,$FB,$A2,$32,$7F,$97,$9C,$61,$FE,$8C
	.dc.b	$6B,$FB,$46,$BF,$B3,$6A,$D3,$E7,$F9,$7D,$BD,$FF,$23,$6F,$F2,$DF
	.dc.b	$3F,$F9,$A3,$F3,$7C,$36,$3C,$5F,$13,$F9,$8C,$BF,$DA,$D1,$FE,$A1
	.dc.b	$6E,$5E,$59,$FD,$6A,$FA,$47,$C6,$E1,$B7,$FD,$6A,$46,$E7,$47,$7F
	.dc.b	$8A,$FD,$6F,$C2,$FC,$66,$D7,$C5,$74,$F8,$DB,$5F,$1D,$E1,$1F,$A6
	.dc.b	$F1,$2C,$FF,$89,$7D,$D6,$BF,$BA,$F8,$2F,$E5,$F1,$8B,$FD,$D4,$62
	.dc.b	$75,$1F,$E6,$A9,$D5,$7F,$8B,$FB,$BF,$D2,$93,$9C,$B9,$9F,$83,$F1
	.dc.b	$6C,$64,$95,$FE,$37,$E6,$4E,$A3,$A7,$CF,$FF,$0E,$3C,$BF,$3F,$6F
	.dc.b	$F8,$74,$59,$F6,$4D,$ED,$FE,$B1,$CE,$CD,$79,$C3,$F6,$9A,$30,$C6
	.dc.b	$3F,$89,$8F,$D5,$3F,$1F,$B7,$FA,$9F,$E4,$DF,$C7,$23,$17,$FE,$8C
	.dc.b	$8C,$67,$E0,$BE,$2F,$8A,$33,$3E,$0B,$FE,$53,$84,$8F,$99,$F3,$D7
	.dc.b	$1F,$99,$37,$0E,$93,$6B,$FD,$B5,$CA,$E4,$FE,$67,$8F,$F6,$AC,$7F
	.dc.b	$25,$18,$AF,$F3,$DD,$6D,$FC,$94,$E9,$B1,$98,$FC,$B3,$5F,$F1,$8B
	.dc.b	$9C,$3F,$B0,$36,$7F,$8F,$64,$EE,$F1,$4E,$6F,$F6,$9F,$64,$F6,$BF
	.dc.b	$BE,$B7,$0F,$C6,$7C,$FE,$26,$47,$FA,$E2,$75,$5F,$99,$AC,$BF,$E7
	.dc.b	$F8,$7B,$7D,$0F,$CD,$F8,$56,$FF,$30,$DA,$FD,$13,$74,$FE,$F5,$F4
	.dc.b	$7F,$E7,$AA,$F0,$79,$DF,$2E,$31,$7D,$E3,$1E,$AF,$F0,$09,$CC,$FF
	.dc.b	$95,$9A,$FE,$F8,$38,$7F,$80,$BD,$2F,$C3,$3F,$20,$EB,$7F,$07,$D1
	.dc.b	$F8,$E7,$ED,$BE,$5E,$C7,$EC,$DB,$87,$EF,$5B,$7F,$A2,$D3,$FC,$B5
	.dc.b	$5F,$2F,$BF,$F7,$F7,$F4,$9B,$19,$3F,$7D,$19,$7F,$8B,$D7,$F7,$2F
	.dc.b	$67,$F8,$DC,$FF,$9D,$3C,$B4,$62,$3F,$C5,$2D,$7F,$C5,$0D,$7F,$C4
	.dc.b	$AF,$29,$E9,$E1,$7F,$90,$97,$D5,$3E,$2B,$FC,$4D,$5B,$C7,$E6,$7D
	.dc.b	$37,$F6,$C7,$CA,$79,$AB,$83,$FE,$DA,$8D,$63,$FE,$D5,$C3,$AE,$1F
	.dc.b	$A7,$9C,$CF,$FA,$39,$AF,$6F,$0B,$F9,$EC,$62,$3F,$BE,$78,$5F,$D0
	.dc.b	$3C,$2F,$E3,$38,$FE,$B1,$B3,$FC,$FD,$AF,$FA,$69,$CE,$3F,$E4,$96
	.dc.b	$CF,$E1,$FD,$97,$99,$FB,$B9,$CD,$FF,$93,$3F,$27,$FE,$DB,$7F,$D2
	.dc.b	$F8,$11,$89,$FA,$8D,$7F,$CB,$6D,$7F,$CB,$4E,$F1,$FF,$16,$B6,$BF
	.dc.b	$C3,$1E,$BB,$E5,$3F,$CA,$D3,$9B,$F1,$7A,$18,$3F,$CB,$93,$6F,$E9
	.dc.b	$CE,$9B,$8F,$9A,$B7,$18,$BE,$93,$EA,$EA,$FF,$23,$C7,$F9,$AE,$31
	.dc.b	$3E,$F8,$C1,$FE,$9A,$CB,$D3,$5F,$E6,$DF,$84,$F3,$BF,$E8,$CF,$82
	.dc.b	$7E,$51,$FA,$2C,$FF,$AC,$7C,$86,$63,$18,$7F,$48,$F3,$13,$9E,$AF
	.dc.b	$F3,$4E,$1C,$9B,$7C,$F0,$FD,$F4,$E7,$4F,$EF,$CD,$86,$3F,$EA,$B8
	.dc.b	$C3,$FD,$17,$D2,$ED,$5F,$22,$7F,$41,$EC,$FF,$9B,$BE,$97,$D5,$D5
	.dc.b	$FD,$2B,$1F,$F0,$F6,$A9,$CD,$79,$AB,$FA,$B5,$79,$1A,$FF,$84,$2F
	.dc.b	$AE,$1E,$69,$E9,$18,$9F,$3C,$FF,$3C,$F4,$7A,$BE,$5A,$D9,$E3,$E6
	.dc.b	$79,$5F,$E0,$4D,$7D,$D3,$D3,$47,$CF,$7F,$A2,$3C,$EF,$A9,$18,$C7
	.dc.b	$FD,$11,$18,$9F,$9F,$39,$CF,$FE,$00,$E9,$FD,$EE,$72,$4B,$F9,$15
	.dc.b	$BC,$7F,$B8,$7A,$7C,$E7,$3D,$9D,$08,$FD,$6A,$FF,$7E,$B7,$1F,$09
	.dc.b	$1C,$1F,$DA,$1F,$F1,$CF,$C0,$7C,$BF,$94,$FF,$1E,$36,$3F,$A0,$07
	.dc.b	$0F,$F4,$36,$F8,$6F,$98,$FC,$2C,$B5,$5F,$F4,$9A,$D6,$9C,$2F,$75
	.dc.b	$AB,$FF,$25,$D6,$9E,$57,$F7,$FB,$DF,$25,$FE,$00,$B4,$62,$BF,$2E
	.dc.b	$D7,$E7,$B5,$FC,$EA,$F6,$F9,$36,$38,$C6,$0C,$FC,$E7,$F6,$1C,$FF
	.dc.b	$41,$48,$C6,$BE,$56,$DF,$F9,$E7,$93,$FD,$44,$FE,$8E,$B8,$3E,$69
	.dc.b	$EA,$D2,$1F,$E5,$3F,$43,$9F,$37,$F9,$05,$EB,$9F,$C0,$AB,$F8,$F7
	.dc.b	$F9,$DA,$FF,$60,$D7,$F4,$27,$D3,$71,$9F,$F8,$EF,$AB,$8F,$95,$2B
	.dc.b	$8B,$E0,$D3,$F6,$91,$83,$F8,$1D,$F5,$7F,$83,$B5,$FD,$A3,$5F,$04
	.dc.b	$FA,$B9,$1F,$4B,$4F,$04,$F9,$B3,$5B,$E7,$3D,$A4,$9F,$F2,$B7,$55
	.dc.b	$7F,$57,$3B,$7C,$B3,$F8,$58,$C4,$6A,$3C,$5E,$94,$8F,$8F,$1F,$63
	.dc.b	$C6,$FE,$A3,$E1,$5F,$18,$D3,$E9,$F2,$3E,$6B,$18,$7F,$3B,$9F,$F2
	.dc.b	$9D,$F2,$8C,$1F,$D7,$7F,$29,$78,$7D,$DB,$1F,$D3,$1A,$FD,$8E,$7F
	.dc.b	$CE,$FA,$7B,$A9,$3B,$7C,$71,$88,$FB,$F6,$BE,$7C,$CE,$9F,$F5,$BC
	.dc.b	$E7,$0E,$37,$CC,$F3,$3E,$AB,$F1,$C7,$F9,$06,$78,$3F,$E7,$0F,$CA
	.dc.b	$BF,$CA,$EF,$F6,$60,$E1,$FC,$89,$8F,$5B,$F4,$AE,$BE,$06,$CE,$E3
	.dc.b	$F8,$7B,$5F,$F2,$83,$5F,$F6,$7B,$57,$D2,$FF,$BD,$9F,$F4,$1B,$5F
	.dc.b	$F6,$B3,$5F,$E3,$6D,$7F,$CB,$8E,$F3,$F8,$D1,$FE,$DC,$FA,$67,$CF
	.dc.b	$7A,$1B,$1E,$FA,$E1,$AF,$F9,$33,$F2,$07,$CA,$FF,$02,$C5,$18,$BF
	.dc.b	$F0,$5C,$62,$7A,$63,$BA,$D5,$7F,$30,$AD,$3F,$EE,$F6,$B9,$8D,$BF
	.dc.b	$E7,$5E,$17,$E4,$BC,$AB,$FC,$D1,$F1,$BF,$C2,$BD,$B3,$FC,$32,$75
	.dc.b	$9F,$FB,$EA,$D5,$AC,$3F,$D3,$17,$C4,$E1,$9F,$1F,$03,$A7,$B1,$2B
	.dc.b	$FC,$4F,$87,$8A,$71,$1F,$E9,$D9,$B9,$FF,$B7,$33,$FE,$27,$9F,$FA
	.dc.b	$37,$EC,$1D,$FF,$86,$6E,$1F,$8D,$6B,$F8,$CF,$C2,$3F,$85,$B9,$9F
	.dc.b	$EB,$93,$A8,$FE,$B2,$D7,$B4,$68,$F8,$95,$23,$91,$EA,$B7,$D2,$FF
	.dc.b	$32,$F8,$8F,$D5,$D7,$FB,$C3,$AB,$E5,$4E,$59,$7F,$0A,$B6,$B6,$FA
	.dc.b	$B5,$CF,$E3,$A4,$ED,$FA,$9F,$19,$88,$FE,$05,$69,$C6,$DF,$26,$BF
	.dc.b	$57,$F4,$DF,$D6,$6F,$F5,$ED,$9F,$CD,$C2,$31,$9A,$1E,$9B,$EE,$3F
	.dc.b	$BD,$47,$EE,$EB,$A3,$FE,$02,$6B,$FE,$C3,$6B,$CD,$AF,$7F,$3F,$FA
	.dc.b	$07,$AD,$23,$A4,$7F,$43,$F8,$49,$1F,$88,$F3,$51,$F9,$5F,$8A,$CC
	.dc.b	$FF,$23,$9C,$BB,$9D,$3F,$90,$F3,$FF,$C6,$AF,$D9,$F5,$BF,$3C,$7F
	.dc.b	$5B,$C6,$C6,$48,$7F,$2F,$9E,$2F,$7D,$3F,$AC,$11,$FC,$27,$C0,$D7
	.dc.b	$47,$A6,$3F,$59,$19,$DF,$FC,$E5,$17,$7F,$D7,$73,$7F,$A4,$EF,$F1
	.dc.b	$DF,$01,$FB,$32,$3E,$BA,$AF,$F4,$39,$4E,$2F,$F8,$BF,$37,$F9,$62
	.dc.b	$FF,$79,$E0,$FF,$6F,$C2,$1F,$9B,$72,$FC,$F3,$27,$57,$FB,$17,$E5
	.dc.b	$7F,$86,$87,$90,$8C,$37,$8F,$FC,$2B,$F5,$1F,$E2,$FC,$BF,$94,$CE
	.dc.b	$6F,$FD,$79,$7F,$86,$8C,$47,$FB,$BE,$30,$77,$CB,$FC,$77,$EC,$7F
	.dc.b	$6E,$3C,$78,$FE,$7B,$A3,$FC,$B1,$CE,$69,$30,$EF,$EB,$FF,$72,$C7
	.dc.b	$FC,$5F,$1D,$E3,$07,$FB,$77,$D0,$FE,$43,$1F,$B5,$6E,$5F,$E4,$CF
	.dc.b	$13,$FB,$FE,$5E,$EF,$A5,$F8,$BF,$45,$30,$FE,$3F,$C0,$9C,$47,$B6
	.dc.b	$7C,$44,$7F,$DD,$58,$F9,$38,$3E,$CB,$79,$7F,$DA,$FA,$72,$9F,$97
	.dc.b	$AF,$E0,$2F,$FD,$A2,$D7,$FB,$FB,$63,$F0,$79,$7C,$59,$DF,$27,$C3
	.dc.b	$23,$F5,$1A,$FF,$83,$F1,$FE,$66,$FC,$3F,$1B,$F0,$73,$FE,$25,$F5
	.dc.b	$1F,$5D,$49,$5F,$0B,$5A,$FF,$A2,$39,$7F,$C7,$53,$FE,$24,$AE,$33
	.dc.b	$FE,$FF,$6B,$B6,$E7,$F1,$37,$FF,$17,$9D,$1F,$C6,$C3,$E9,$BC,$84
	.dc.b	$AF,$A7,$F3,$F5,$E9,$1E,$12,$3E,$17,$FC,$77,$E2,$C8,$FF,$7E,$D7
	.dc.b	$56,$FD,$1F,$C1,$23,$F8,$0F,$9C,$AF,$A6,$3D,$C4,$7D,$CF,$F9,$23
	.dc.b	$A2,$47,$43,$BF,$C4,$FC,$D2,$3F,$CD,$73,$FF,$09,$F3,$3E,$59,$AE
	.dc.b	$AB,$E4,$5F,$CD,$23,$F3,$5F,$F2,$BE,$72,$3B,$9E,$EF,$67,$EE,$11
	.dc.b	$FE,$F3,$CB,$FE,$1E,$7F,$C9,$73,$FE,$62,$7E,$C4,$FD,$B2,$3F,$9C
	.dc.b	$F1,$FF,$CC,$BC,$92,$39,$C7,$F9,$A0,$C3,$F8,$F8,$C3,$E0,$F1,$7F
	.dc.b	$CD,$A4,$7C,$79,$57,$F2,$D1,$75,$FD,$4C,$FF,$93,$3E,$81,$F2,$3F
	.dc.b	$ED,$1B,$FF,$6C,$7B,$7F,$D7,$D4,$EB,$E0,$3F,$DC,$33,$98,$C4,$F0
	.dc.b	$1F,$40,$FA,$2F,$F1,$AF,$C4,$2F,$86,$3D,$E3,$17,$FD,$B7,$0E,$71
	.dc.b	$88,$F0,$CE,$EB,$FB,$96,$1D,$7C,$3F,$F7,$64,$74,$CB,$CB,$7C,$EF
	.dc.b	$83,$63,$37,$C6,$36,$76,$8B,$8F,$8A,$FA,$4F,$B6,$B8,$3E,$F3,$81
	.dc.b	$67,$83,$DC,$E1,$97,$C2,$AE,$71,$F3,$4E,$25,$9F,$96,$DC,$3E,$A3
	.dc.b	$BF,$88,$FC,$DF,$D1,$71,$8F,$F8,$03,$AA,$31,$09,$DB,$B6,$BD,$F8
	.dc.b	$BF,$C3,$6F,$3C,$3F,$D7,$3E,$57,$26,$C6,$67,$6F,$19,$F8,$F1,$CB
	.dc.b	$7F,$B7,$AB,$E8,$9C,$FF,$E8,$69,$FB,$98,$F4,$6D,$F1,$DB,$F1,$8D
	.dc.b	$CB,$F1,$78,$72,$E9,$EF,$9C,$9F,$A1,$5B,$CF,$C1,$9F,$7C,$79,$DE
	.dc.b	$F2,$EE,$FE,$0B,$2E,$98,$F2,$8D,$57,$08,$FC,$04,$EA,$B8,$C7,$E7
	.dc.b	$FC,$87,$D1,$7F,$B8,$34,$FF,$3D,$B6,$7F,$DA,$8F,$EB,$A3,$3A,$7F
	.dc.b	$AA,$39,$7F,$C2,$73,$B6,$C6,$67,$FD,$1D,$3C,$6B,$9C,$E5,$FF,$33
	.dc.b	$43,$DA,$C6,$1E,$4F,$6E,$CE,$70,$F2,$C6,$2B,$2D,$8C,$9F,$CA,$CF
	.dc.b	$F7,$0F,$D0,$FC,$6C,$BF,$C2,$ED,$8F,$F5,$17,$6B,$C1,$F2,$78,$CA
	.dc.b	$D6,$9F,$E2,$46,$C6,$78,$BC,$F7,$C7,$0F,$05,$EE,$79,$DF,$FA,$CA
	.dc.b	$7E,$69,$8C,$BD,$AD,$38,$E1,$E1,$8F,$8D,$0F,$A2,$7C,$FE,$F1,$FF
	.dc.b	$80,$38,$75,$6D,$F0,$AD,$DF,$4A,$D6,$9F,$0A,$3E,$7F,$AE,$F6,$FC
	.dc.b	$F1,$86,$BC,$4F,$AB,$F3,$0F,$A2,$CE,$7E,$29,$B8,$7F,$95,$AB,$51
	.dc.b	$C4,$F6,$FD,$13,$E5,$7C,$3E,$F7,$D0,$8C,$4F,$FA,$16,$31,$32,$3F
	.dc.b	$D9,$F5,$FE,$03,$79,$C6,$70,$61,$FE,$54,$C3,$9C,$62,$BF,$DC,$17
	.dc.b	$C6,$30,$7F,$98,$27,$75,$C7,$E8,$3E,$2D,$79,$5B,$7E,$A7,$0F,$93
	.dc.b	$E2,$6E,$1D,$38,$98,$AF,$38,$FC,$5B,$B8,$7F,$AD,$6F,$7F,$1D,$F5
	.dc.b	$3E,$83,$E2,$36,$7E,$4B,$87,$E1,$FC,$D5,$CF,$C1,$9A,$B1,$97,$D3
	.dc.b	$AB,$7F,$F0,$07,$85,$F3,$7F,$82,$3F,$43,$A3,$DB,$C5,$7E,$52,$30
	.dc.b	$D8,$FC,$1F,$D1,$70,$6C,$64,$FA,$F1,$97,$FD,$81,$F3,$5F,$90,$3F
	.dc.b	$23,$C5,$63,$B7,$A7,$E2,$3F,$1F,$F2,$16,$3F,$5B,$C0,$FA,$17,$B9
	.dc.b	$FC,$2E,$30,$FE,$5B,$CE,$C4,$EA,$BD,$5E,$A7,$E8,$3B,$FD,$ED,$8C
	.dc.b	$93,$0F,$C2,$F8,$AE,$AF,$E3,$D3,$F5,$F8,$57,$CA,$38,$72,$FE,$87
	.dc.b	$C6,$C6,$3F,$EF,$CE,$1F,$D7,$FC,$D7,$F1,$EF,$8A,$46,$30,$FE,$13
	.dc.b	$A2,$73,$1F,$EE,$7A,$E1,$5B,$BF,$2D,$7E,$F9,$BA,$FE,$DB,$C7,$EC
	.dc.b	$CC,$3A,$D6,$F3,$FD,$B5,$66,$FF,$DE,$1E,$C7,$F0,$43,$CD,$F4,$57
	.dc.b	$0F,$F7,$C4,$59,$F0,$CD,$BF,$F6,$AF,$57,$F9,$66,$F6,$78,$27,$4A
	.dc.b	$9A,$BE,$23,$F8,$94,$F9,$BC,$1F,$E8,$71,$8A,$F7,$C6,$3C,$4F,$94
	.dc.b	$DA,$E0,$7C,$AC,$58,$FF,$1E,$3E,$87,$C6,$59,$D4,$5C,$3E,$5C,$DC
	.dc.b	$D5,$FE,$4B,$E6,$FB,$EF,$5C,$7F,$1B,$C9,$FC,$A9,$B1,$F9,$69,$DE
	.dc.b	$1F,$8A,$D1,$E1,$6A,$3F,$88,$9F,$9B,$C3,$18,$FF,$BF,$AB,$78,$73
	.dc.b	$9F,$EC,$BF,$1E,$FA,$2A,$7E,$2F,$E6,$ED,$7E,$2A,$74,$D9,$D3,$4F
	.dc.b	$C1,$F8,$9D,$3B,$BF,$28,$D8,$C9,$CE,$32,$D3,$F6,$7F,$45,$FE,$65
	.dc.b	$7C,$1E,$6F,$EC,$E3,$1A,$FE,$AB,$1F,$ED,$6A,$33,$AF,$CE,$8F,$C2
	.dc.b	$46,$2F,$E8,$7A,$7C,$C8,$F8,$1F,$DE,$98,$9F,$D1,$FC,$7F,$C6,$F0
	.dc.b	$7F,$5C,$DA,$FD,$0B,$67,$71,$F1,$E7,$F6,$91,$F3,$BB,$EC,$FC,$B6
	.dc.b	$C7,$AD,$B8,$FA,$7A,$59,$FF,$06,$A7,$95,$FE,$5B,$F0,$FF,$C1,$BE
	.dc.b	$C7,$F1,$39,$FE,$A3,$83,$E8,$9F,$3F,$AB,$9A,$7F,$81,$30,$E1,$7F
	.dc.b	$E3,$1B,$DC,$E6,$BF,$91,$7D,$17,$F1,$78,$FF,$35,$B7,$1F,$F3,$83
	.dc.b	$5F,$F3,$D3,$5F,$F5,$43,$BB,$FA,$90,$8C,$4F,$F5,$7F,$B1,$7C,$87
	.dc.b	$83,$C9,$FE,$65,$BF,$F8,$DF,$0F,$CE,$B7,$1E,$7E,$DF,$F5,$0C,$7F
	.dc.b	$9D,$CE,$5F,$F5,$C3,$6B,$F8,$33,$54,$F1,$7A,$F8,$BF,$E3,$8F,$9D
	.dc.b	$7A,$34,$FF,$1A,$FC,$9D,$C7,$FC,$5B,$3A,$6C,$E9,$A4,$BF,$EF,$53
	.dc.b	$B7,$FA,$1F,$B2,$BE,$91,$AB,$B9,$FE,$0E,$B5,$87,$38,$FC,$F7,$23
	.dc.b	$1F,$F2,$59,$FE,$4E,$D1,$6F,$8D,$F6,$76,$4C,$7D,$2F,$F9,$42,$31
	.dc.b	$9F,$F8,$6B,$E1,$1F,$45,$F9,$1C,$5E,$53,$E8,$FF,$CE,$15,$AE,$CF
	.dc.b	$F2,$0C,$62,$FF,$31,$39,$CF,$FA,$9F,$8D,$FE,$24,$7F,$9C,$46,$72
	.dc.b	$F0,$57,$03,$C1,$5B,$84,$EF,$83,$F7,$CF,$E7,$BC,$CF,$F2,$8F,$D1
	.dc.b	$79,$3B,$BF,$9F,$57,$F4,$A8,$C4,$CC,$BC,$7E,$8D,$C7,$E0,$61,$C2
	.dc.b	$7C,$AD,$8F,$E9,$B9,$7F,$6A,$77,$7C,$71,$F1,$D6,$F7,$FE,$A5,$8E
	.dc.b	$71,$F7,$EE,$F9,$7F,$F3,$3F,$9D,$F2,$A3,$13,$FD,$3E,$31,$32,$3E
	.dc.b	$4F,$A7,$74,$FA,$D7,$C1,$B5,$FD,$7A,$75,$9F,$C6,$AF,$CF,$DF,$FA
	.dc.b	$33,$7F,$B7,$D7,$FE,$A6,$FC,$79,$EA,$F8,$E3,$C9,$9C,$BF,$F0,$5F
	.dc.b	$AF,$FE,$72,$8F,$09,$FE,$27,$3F,$B9,$7C,$1F,$DC,$FA,$4F,$F7,$A3
	.dc.b	$63,$FD,$FC,$D7,$FB,$C0,$71,$E3,$BE,$33,$FC,$E3,$3F,$80,$AD,$CE
	.dc.b	$76,$DD,$6B,$55,$6A,$30,$FE,$3B,$B5,$F3,$1E,$0B,$FC,$4C,$63,$1F
	.dc.b	$88,$D7,$FA,$77,$D5,$6F,$5E,$5E,$F5,$9F,$C9,$D7,$F9,$75,$E3,$19
	.dc.b	$E7,$71,$3E,$9F,$86,$DF,$B4,$F9,$9B,$89,$A7,$1E,$CF,$F3,$F6,$5E
	.dc.b	$48,$CB,$4F,$D5,$FB,$58,$7F,$C9,$BC,$EF,$5F,$C4,$D8,$E0,$7E,$F7
	.dc.b	$E9,$7F,$CB,$EF,$E0,$DA,$E7,$CF,$F1,$78,$3E,$89,$CF,$1F,$EC,$1B
	.dc.b	$C5,$FB,$A6,$DF,$E9,$AB,$85,$C8,$EC,$F4,$69,$FE,$2D,$FA,$6F,$39
	.dc.b	$B7,$CD,$F2,$9E,$12,$3F,$C2,$BC,$0A,$E8,$F5,$47,$C9,$8C,$E5,$FE
	.dc.b	$4A,$F6,$7E,$19,$8C,$FF,$D4,$DE,$53,$C8,$FB,$9E,$5F,$F2,$B6,$C6
	.dc.b	$72,$FF,$35,$F2,$3E,$9B,$35,$27,$FA,$1B,$F8,$19,$D4,$7F,$BC,$BE
	.dc.b	$9B,$FD,$C8,$DF,$27,$BC,$62,$3F,$A1,$FB,$38,$99,$1B,$3C,$31,$87
	.dc.b	$FD,$89,$3C,$75,$FE,$86,$DD,$3F,$AF,$E0,$BE,$15,$AC,$2D,$6A,$7F
	.dc.b	$39,$83,$E9,$9E,$E6,$1B,$1F,$EA,$86,$BF,$E5,$07,$F8,$4B,$98,$E0
	.dc.b	$47,$87,$91,$DB,$E9,$7F,$B6,$C7,$FA,$C9,$AF,$F9,$69,$B9,$7F,$82
	.dc.b	$3E,$9B,$FA,$0C,$EF,$D3,$FE,$12,$CA,$E2,$D8,$CB,$FC,$3A,$32,$FF
	.dc.b	$A1,$E9,$F4,$CF,$99,$FD,$E4,$61,$B1,$99,$FF,$63,$E1,$FE,$16,$79
	.dc.b	$6B,$CE,$75,$1F,$E3,$26,$DF,$F9,$4C,$E9,$1C,$BD,$9F,$E0,$DF,$47
	.dc.b	$AB,$FF,$0A,$FB,$CF,$92,$FE,$14,$DB,$FF,$5D,$D6,$DF,$F4,$FB,$7C
	.dc.b	$8F,$F2,$3D,$70,$3A,$E9,$DB,$BB,$9F,$B7,$30,$E7,$19,$7F,$CF,$D1
	.dc.b	$97,$FD,$AF,$87,$EF,$17,$D2,$BB,$BF,$0E,$D8,$FF,$75,$B5,$FF,$49
	.dc.b	$FD,$33,$C1,$DD,$49,$DB,$F9,$09,$D4,$60,$FF,$48,$6D,$F0,$8E,$0F
	.dc.b	$93,$E7,$FF,$97,$7C,$67,$E7,$E3,$FD,$82,$DD,$D9,$AB,$E7,$9C,$3A
	.dc.b	$E5,$FC,$C8,$E7,$F2,$52,$3F,$DB,$AF,$D6,$F8,$FC,$47,$FB,$9A,$2F
	.dc.b	$2F,$F9,$DF,$05,$7E,$AB,$E9,$7F,$A2,$5F,$FA,$D1,$B5,$FE,$62,$6E
	.dc.b	$9F,$ED,$9F,$65,$F4,$DF,$D8,$6F,$F1,$A7,$FB,$76,$B6,$DF,$4F,$B1
	.dc.b	$FD,$4E,$B8,$FA,$F8,$9F,$F6,$C7,$B9,$FD,$08,$C3,$FD,$1B,$E2,$7F
	.dc.b	$A2,$91,$FC,$6F,$C6,$57,$47,$FB,$86,$3F,$2F,$1F,$D5,$67,$F3,$35
	.dc.b	$2B,$FC,$7B,$19,$7F,$2F,$19,$7F,$DB,$11,$FD,$52,$E4,$7F,$BB,$5F
	.dc.b	$D1,$FC,$2E,$23,$FA,$DC,$5D,$7E,$74,$DD,$7F,$CF,$7E,$8F,$F9,$EA
	.dc.b	$F9,$1F,$A6,$9F,$D6,$FC,$24,$8F,$D4,$BF,$DB,$70,$C4,$7E,$7A,$2E
	.dc.b	$1F,$EE,$79,$FD,$AF,$81,$FD,$AA,$A1,$FB,$27,$2F,$D6,$E6,$FF,$32
	.dc.b	$BA,$3F,$53,$1F,$D3,$3E,$9B,$EC,$F9,$E6,$9F,$72,$9F,$4D,$FE,$11
	.dc.b	$C7,$FC,$FB,$3A,$E2,$E4,$7F,$83,$67,$5B,$32,$FF,$62,$D7,$3F,$33
	.dc.b	$30,$C7,$FD,$07,$EE,$F2,$8F,$C6,$8E,$5F,$CC,$B8,$6B,$F9,$9E,$7F
	.dc.b	$07,$87,$EB,$F6,$24,$FE,$5B,$7F,$64,$79,$4F,$EC,$53,$27,$FB,$4B
	.dc.b	$FD,$97,$6F,$57,$D2,$FC,$5F,$A3,$95,$F2,$F8,$13,$88,$70,$FC,$4A
	.dc.b	$C3,$FE,$E7,$AD,$CF,$F7,$6C,$7F,$B9,$11,$F3,$70,$2F,$EA,$72,$49
	.dc.b	$CB,$F3,$75,$C4,$78,$38,$BE,$76,$DF,$66,$B3,$9F,$86,$ED,$7F,$70
	.dc.b	$F1,$BF,$C0,$46,$30,$FF,$7C,$FA,$CF,$A6,$FF,$07,$B6,$3E,$53,$5F
	.dc.b	$01,$F6,$76,$F4,$3A,$9E,$24,$7E,$46,$95,$FF,$66,$EF,$FE,$13,$31
	.dc.b	$9F,$FB,$23,$07,$D3,$78,$89,$5F,$47,$E9,$2B,$E2,$DD,$5C,$23,$CF
	.dc.b	$5F,$93,$AD,$DF,$FB,$5F,$C3,$FF,$11,$18,$DB,$E4,$E0,$FA,$6F,$82
	.dc.b	$4C,$7F,$01,$F4,$15,$F0,$4E,$DF,$8B,$3F,$A1,$9D,$61,$FE,$E5,$E1
	.dc.b	$FF,$18,$18,$C3,$BF,$C7,$7C,$E2,$61,$FC,$B2,$1F,$32,$D3,$33,$E0
	.dc.b	$AB,$82,$FF,$5C,$70,$AC,$BF,$E2,$EC,$FF,$C8,$06,$2F,$EF,$FB,$5F
	.dc.b	$78,$8F,$F9,$2B,$BA,$BE,$E8,$B8,$7F,$BF,$E7,$FC,$95,$E2,$7F,$92
	.dc.b	$F8,$E5,$FF,$94,$A3,$2F,$F9,$52,$32,$FE,$23,$BB,$FE,$00,$CF,$F9
	.dc.b	$4F,$C4,$93,$FE,$5A,$7F,$CA,$DA,$62,$38,$78,$1F,$D4,$5B,$3A,$7F
	.dc.b	$83,$FD,$23,$E5,$7F,$C6,$7F,$17,$FE,$4D,$9F,$E2,$53,$9F,$86,$FF
	.dc.b	$61,$FB,$BF,$EB,$E8,$C4,$FF,$AF,$E3,$F8,$C7,$D2,$3E,$9B,$FC,$53
	.dc.b	$87,$3A,$5F,$08,$C3,$6F,$A6,$3E,$27,$9D,$F2,$73,$8F,$8B,$0F,$2D
	.dc.b	$6D,$B1,$9C,$7C,$5E,$2F,$FB,$53,$2E,$F7,$BB,$F1,$B7,$2F,$14,$2F
	.dc.b	$8B,$DC,$DE,$16,$77,$18,$3E,$11,$C7,$2F,$43,$E4,$3F,$CB,$31,$F2
	.dc.b	$1F,$F2,$FB,$5F,$69,$FE,$5E,$F8,$DF,$CD,$7D,$37,$F9,$82,$11,$87
	.dc.b	$5B,$A7,$53,$99,$EB,$CB,$FD,$AD,$BE,$57,$3C,$3F,$D4,$CD,$9D,$4F
	.dc.b	$DD,$7F,$B9,$6B,$F5,$9B,$BF,$DE,$FA,$67,$57,$EA,$EF,$E0,$9F,$7A
	.dc.b	$B6,$FE,$29,$AE,$70,$FA,$D5,$CB,$AB,$E6,$56,$B2,$F2,$DF,$46,$E5
	.dc.b	$F9,$5E,$06,$1C,$6B,$57,$C4,$98,$F6,$CB,$CD,$87,$AE,$3B,$92,$BE
	.dc.b	$CE,$1D,$2B,$79,$FC,$3A,$FC,$87,$33,$95,$C5,$FE,$83,$DB,$FD,$6A
	.dc.b	$DC,$BF,$BB,$4B,$E3,$F1,$7C,$6A,$CF,$2B,$FF,$69,$C6,$2B,$2F,$0F
	.dc.b	$96,$7C,$85,$F1,$B0,$DB,$FF,$0C,$4E,$9B,$3A,$69,$FE,$06,$FC,$73
	.dc.b	$F9,$C3,$FE,$E2,$F0,$BD,$1A,$27,$2F,$F0,$A8,$CC,$62,$3F,$D6,$ED
	.dc.b	$7F,$C3,$07,$1F,$0F,$F0,$8E,$9A,$FF,$70,$FC,$5D,$E5,$7C,$7C,$1F
	.dc.b	$4E,$FA,$2F,$95,$E5,$2F,$75,$FC,$FE,$B4,$FF,$85,$E3,$38,$7F,$6E
	.dc.b	$8E,$77,$CA,$73,$E2,$FE,$5E,$72,$71,$C7,$47,$CF,$FA,$CB,$1D,$7C
	.dc.b	$AF,$A5,$F4,$FC,$DB,$19,$3E,$8C,$65,$A6,$72,$FE,$E1,$8F,$FA,$66
	.dc.b	$3E,$BF,$BB,$B9,$CF,$8F,$F7,$FC,$DF,$43,$6B,$DC,$D9,$DC,$7E,$02
	.dc.b	$76,$E8,$FC,$1F,$D5,$E2,$7C,$3E,$8F,$E2,$FE,$2B,$E2,$36,$BE,$53
	.dc.b	$67,$71,$F8,$C9,$DB,$A3,$F1,$FF,$2D,$89,$F6,$F5,$7E,$5F,$E2,$FE
	.dc.b	$83,$6B,$EB,$36,$77,$1F,$98,$9D,$BA,$3F,$37,$CD,$89,$FE,$C5,$D4
	.dc.b	$FA,$6F,$D2,$1F,$86,$FB,$67,$16,$30,$FF,$2D,$46,$1F,$D2,$CE,$E7
	.dc.b	$57,$D3,$E9,$31,$5F,$AA,$6C,$7F,$89,$F5,$FF,$61,$CF,$F9,$95,$B1
	.dc.b	$E4,$F8,$FF,$CB,$C6,$3A,$3F,$43,$F4,$DF,$90,$BF,$F3,$E7,$86,$D5
	.dc.b	$87,$EC,$2B,$58,$FF,$9F,$67,$37,$FB,$CF,$43,$F8,$11,$FC,$0F,$C7
	.dc.b	$B1,$FE,$80,$BF,$03,$DA,$9E,$57,$5F,$5F,$F8,$53,$67,$F8,$3B,$71
	.dc.b	$F9,$AD,$7F,$76,$2F,$DE,$E1,$AF,$F0,$06,$BE,$9F,$8C,$FE,$3F,$D6
	.dc.b	$F9,$EF,$E3,$51,$86,$E1,$FE,$AF,$9D,$3F,$C6,$A7,$4E,$7E,$7B,$F9
	.dc.b	$17,$C7,$74,$6C,$64,$FE,$43,$19,$69,$FC,$90,$F1,$FD,$3F,$F2,$B6
	.dc.b	$99,$3C,$4E,$5A,$67,$1F,$F1,$96,$09,$D4,$61,$FC,$95,$7F,$3E,$F9
	.dc.b	$D7,$C6,$63,$0F,$F6,$64,$61,$FC,$AC,$FF,$24,$B7,$90,$EA,$4E,$63
	.dc.b	$0D,$8C,$D7,$D6,$AF,$F6,$25,$FE,$36,$3E,$67,$C7,$78,$F8,$BE,$9B
	.dc.b	$63,$27,$92,$32,$FE,$2F,$C0,$F5,$7D,$3F,$E5,$5B,$19,$3D,$71,$96
	.dc.b	$99,$C3,$C9,$C4,$AF,$74,$7E,$2D,$B1,$99,$F5,$74,$38,$1F,$21,$F8
	.dc.b	$AF,$9D,$FC,$5B,$6B,$F1,$0D,$C7,$EA,$B7,$AF,$FC,$29,$C3,$FE,$84
	.dc.b	$FA,$0F,$C9,$FC,$E7,$F4,$CF,$84,$46,$31,$FF,$70,$46,$1F,$99,$5C
	.dc.b	$5B,$19,$C7,$F4,$B3,$97,$DF,$F0,$7F,$BB,$F6,$3F,$78,$DB,$FD,$B3
	.dc.b	$67,$FD,$86,$7E,$B5,$C6,$5F,$C2,$1F,$8A,$49,$FF,$21,$6E,$8F,$C2
	.dc.b	$62,$E9,$FD,$7F,$C7,$FB,$5B,$7E,$C7,$64,$7D,$6F,$2F,$9F,$FD,$BE
	.dc.b	$5F,$15,$B3,$F1,$1D,$3F,$09,$B1,$F9,$1F,$A5,$FC,$83,$FE,$DE,$CD
	.dc.b	$F4,$CF,$A3,$FF,$06,$D6,$A7,$FC,$90,$6A,$33,$D1,$F8,$69,$E7,$7F
	.dc.b	$C4,$9A,$FF,$83,$9A,$FE,$A3,$E9,$BE,$26,$48,$FF,$6F,$46,$1F,$F7
	.dc.b	$8F,$E0,$6F,$CB,$1E,$8F,$F8,$3A,$B5,$9F,$E2,$1B,$87,$3E,$44,$63
	.dc.b	$16,$3C,$F0,$E5,$C5,$FE,$2A,$F8,$0F,$E0,$D8,$F8,$F2,$FF,$11,$C3
	.dc.b	$EA,$B5,$E2,$6D,$B5,$FE,$7F,$9E,$39,$7E,$F5,$AF,$F3,$46,$BE,$5E
	.dc.b	$57,$D2,$FF,$93,$3E,$5B,$F3,$CD,$BF,$CE,$BF,$88,$23,$F9,$C6,$D3
	.dc.b	$2F,$D4,$DF,$4A,$F5,$1F,$D6,$BD,$07,$87,$DA,$3A,$92,$FF,$DB,$13
	.dc.b	$A8,$C5,$F4,$ED,$FF,$30,$F8,$5F,$B3,$6C,$FE,$C9,$FC,$E9,$1F,$D8
	.dc.b	$BF,$C9,$FD,$47,$E5,$AF,$A5,$FE,$61,$E0,$FF,$9C,$67,$31,$8B,$F4
	.dc.b	$E7,$FE,$78,$BE,$11,$FC,$D2,$F8,$46,$27,$FD,$E1,$E3,$FE,$FF,$3F
	.dc.b	$F4,$F6,$3D,$30,$FE,$60,$DA,$FE,$CE,$D7,$FC,$C6,$D7,$F2,$8F,$A0
	.dc.b	$FE,$8D,$E6,$78,$1B,$3E,$67,$4F,$F4,$88,$D3,$4D,$61,$FE,$89,$AF
	.dc.b	$B1,$1D,$A3,$17,$FE,$B5,$F1,$D8,$FF,$9A,$2F,$67,$C3,$C7,$AB,$87
	.dc.b	$C2,$FC,$2F,$7B,$37,$F0,$BF,$2E,$FE,$6D,$97,$F5,$CF,$0F,$E8,$46
	.dc.b	$2F,$E4,$C6,$3C,$9F,$EC,$8D,$AE,$07,$BE,$B8,$5F,$5A,$FE,$D5,$C9
	.dc.b	$BC,$3F,$93,$FA,$1F,$E4,$1C,$3E,$DE,$AF,$A9,$7A,$BF,$E9,$AA,$D5
	.dc.b	$FB,$AB,$53,$9F,$47,$FD,$67,$C7,$FE,$45,$AE,$B1,$8E,$7F,$F5,$1C
	.dc.b	$7F,$0B,$F8,$9F,$B3,$F0,$FF,$DB,$DB,$1F,$E2,$66,$BF,$EA,$26,$FC
	.dc.b	$63,$E4,$7F,$59,$3F,$EC,$8A,$FF,$53,$CE,$78,$77,$5A,$E5,$FF,$64
	.dc.b	$5F,$F1,$BA,$D3,$C4,$FE,$3D,$B7,$F9,$29,$F8,$7F,$45,$FE,$C6,$6F
	.dc.b	$C5,$7F,$AD,$31,$FF,$59,$D6,$E7,$E5,$FB,$F9,$E3,$FF,$2A,$F1,$BE
	.dc.b	$03,$14,$E4,$F9,$D1,$88,$60,$C3,$7A,$B9,$F1,$5F,$76,$73,$18,$8F
	.dc.b	$D6,$F8,$B8,$9E,$F8,$FE,$5B,$D0,$BD,$87,$57,$CC,$FF,$15,$BA,$32
	.dc.b	$FE,$67,$07,$B9,$E0,$24,$7F,$8C,$BE,$AE,$BB,$8B,$C1,$FA,$7C,$15
	.dc.b	$8E,$0F,$F3,$CC,$7F,$2E,$BF,$F1,$DF,$8B,$95,$B7,$C1,$C7,$D3,$32
	.dc.b	$7F,$3D,$1F,$0A,$FF,$C7,$11,$C3,$2F,$F1,$2D,$FC,$99,$CB,$67,$F8
	.dc.b	$7C,$FF,$BB,$32,$F3,$78,$BF,$18,$E9,$3F,$AB,$BC,$3F,$E5,$08,$CB
	.dc.b	$FC,$03,$CD,$77,$3D,$3F,$1C,$61,$B1,$99,$FF,$10,$C6,$0F,$AF,$F1
	.dc.b	$9F,$CC,$63,$F6,$6D,$7B,$4F,$17,$C1,$F9,$7B,$E7,$7F,$1B,$82,$31
	.dc.b	$3F,$CD,$A3,$13,$F8,$4E,$8F,$F5,$EC,$E6,$B6,$F4,$9E,$71,$8D,$3F
	.dc.b	$90,$57,$0A,$D1,$FE,$3C,$9E,$2F,$F8,$1C,$FF,$38,$B8,$7F,$C0,$0D
	.dc.b	$73,$5F,$03,$D8,$91,$E3,$8F,$D2,$7D,$47,$F8,$7E,$3E,$EB,$5F,$F3
	.dc.b	$B3,$5F,$F5,$6B,$5F,$13,$BC,$1B,$63,$C4,$EA,$76,$F5,$FF,$65,$F4
	.dc.b	$FF,$90,$C3,$FA,$E8,$70,$FF,$93,$9B,$7F,$E3,$0F,$A3,$7B,$DC,$B9
	.dc.b	$65,$F2,$BD,$E5,$FE,$19,$BD,$CF,$6F,$A8,$FD,$D3,$FE,$C6,$6B,$F2
	.dc.b	$DA,$FF,$8C,$DA,$9F,$E0,$58,$FB,$A3,$FC,$71,$7F,$E5,$D7,$C5,$EA
	.dc.b	$FF,$0F,$AD,$E1,$F5,$3E,$3B,$F2,$FC,$26,$92,$FE,$F3,$FC,$33,$27
	.dc.b	$CC,$D3,$F0,$F1,$F8,$8F,$0B,$99,$2B,$F0,$CD,$7F,$A5,$57,$77,$FA
	.dc.b	$8B,$5F,$CE,$C7,$F2,$5F,$A6,$7C,$17,$EF,$A5,$18,$7F,$80,$B5,$FF
	.dc.b	$5D,$35,$FF,$78,$B9,$DB,$E3,$47,$2F,$82,$EC,$77,$A9,$1F,$82,$AF
	.dc.b	$F5,$4B,$67,$2F,$EC,$BE,$37,$E4,$91,$FD,$23,$5F,$7F,$C7,$FC,$93
	.dc.b	$F5,$B1,$F2,$FE,$73,$E6,$65,$2B,$E1,$C6,$5F,$8B,$19,$7E,$4C,$7C
	.dc.b	$5F,$A4,$F8,$98,$9A,$49,$F8,$AF,$F5,$2F,$75,$F4,$BF,$DA,$E7,$FC
	.dc.b	$77,$18,$F4,$3A,$47,$C1,$AE,$15,$B8,$FF,$0F,$D6,$DD,$1F,$EF,$4D
	.dc.b	$3F,$0D,$9C,$8C,$10,$FC,$0B,$97,$FC,$8B,$1F,$C4,$3B,$7F,$26,$48
	.dc.b	$FE,$1C,$D7,$FC,$97,$87,$F5,$C3,$83,$FE,$1A,$F1,$FF,$93,$70,$7F
	.dc.b	$8D,$F8,$5E,$F7,$E6,$4F,$EC,$7E,$24,$BF,$C8,$C6,$5E,$51,$97,$F2
	.dc.b	$F1,$97,$FC,$B7,$1F,$A3,$F7,$3F,$B6,$EB,$27,$F3,$71,$97,$C9,$19
	.dc.b	$7F,$43,$1F,$DD,$1F,$D6,$75,$47,$D4,$64,$FD,$2B,$97,$FC,$D6,$7F
	.dc.b	$82,$7E,$93,$F5,$0F,$F3,$10,$E1,$FE,$68,$D7,$FD,$78,$D7,$7C,$FF
	.dc.b	$E0,$8F,$A1,$FE,$81,$D7,$FE,$C1,$F3,$91,$8A,$E3,$5E,$13,$87,$89
	.dc.b	$23,$F9,$4E,$1F,$BB,$DF,$FC,$D1,$7F,$C2,$27,$FC,$9E,$D7,$FC,$98
	.dc.b	$D8,$DF,$D2,$7B,$7A,$BC,$9F,$49,$2F,$AC,$7F,$60,$FA,$44,$D2,$4F
	.dc.b	$67,$D9,$9F,$F6,$6F,$B3,$FF,$1A,$4F,$F6,$56,$BE,$28,$D7,$AB,$F0
	.dc.b	$FE,$96,$4E,$E2,$BE,$88,$B9,$FF,$31,$F3,$18,$74,$8C,$3F,$E4,$76
	.dc.b	$FC,$A7,$8C,$FA,$1E,$15,$C3,$E2,$EB,$F4,$CC,$3F,$E0,$FB,$E5,$C7
	.dc.b	$F5,$78,$3F,$C9,$33,$C0,$F6,$61,$EF,$F9,$19,$1E,$F7,$EF,$78,$78
	.dc.b	$8F,$A3,$F4,$BF,$81,$31,$A7,$A2,$B1,$3F,$62,$73,$8F,$FB,$EF,$F2
	.dc.b	$5F,$EB,$73,$FE,$23,$6D,$FC,$36,$CF,$9F,$C2,$FF,$0B,$C6,$2B,$FC
	.dc.b	$C9,$9B,$E9,$7F,$24,$6B,$CD,$93,$F8,$88,$CB,$FE,$1D,$AE,$5C,$25
	.dc.b	$F0,$DD,$89,$3F,$90,$F1,$EB,$F9,$0F,$A0,$FD,$01,$8D,$7D,$15,$89
	.dc.b	$FC,$BC,$E7,$5F,$81,$ED,$FA,$89,$3F,$9C,$F0,$EB,$FE,$C4,$9F,$F2
	.dc.b	$2E,$9F,$E8,$26,$C7,$A6,$B7,$8F,$F7,$99,$CC,$F1,$F5,$7E,$23,$B3
	.dc.b	$FC,$2B,$F4,$0F,$87,$FE,$C8,$D8,$FF,$31,$B5,$FE,$08,$1C,$3F,$C1
	.dc.b	$DD,$5F,$FB,$77,$D8,$FF,$04,$4F,$FA,$FD,$B1,$FD,$A2,$77,$87,$E4
	.dc.b	$7C,$BF,$E1,$1E,$34,$C3,$F7,$B1,$FC,$2C,$C3,$FB,$E8,$C3,$FC,$0B
	.dc.b	$E0,$FF,$88,$11,$FE,$0B,$E3,$D7,$F8,$2C,$5C,$7F,$85,$CF,$FA,$A7
	.dc.b	$E9,$3F,$D2,$BE,$B7,$FB,$37,$1F,$F2,$94,$63,$07,$C2,$A7,$3E,$F7
	.dc.b	$F8,$FE,$31,$2F,$A3,$7D,$37,$F1,$59,$CC,$B0,$E9,$18,$6E,$1D,$31
	.dc.b	$F1,$57,$3B,$E5,$39,$8F,$1B,$6F,$96,$3E,$38,$4F,$1F,$5F,$FD,$C5
	.dc.b	$1B,$24,$6B,$1F,$0B,$B3,$C2,$79,$EB,$A3,$B6,$B9,$C7,$CF,$97,$C0
	.dc.b	$C3,$BC,$EE,$7F,$D5,$B7,$F2,$E3,$FA,$73,$5F,$61,$FD,$47,$E1,$FF
	.dc.b	$75,$F4,$DF,$EB,$46,$C7,$E4,$9A,$EF,$9D,$3D,$BE,$71,$E9,$52,$57
	.dc.b	$E6,$B0,$4E,$A3,$0D,$DB,$EA,$37,$6F,$97,$AB,$E9,$9F,$3D,$FE,$67
	.dc.b	$BD,$E5,$F2,$32,$EA,$DC,$79,$65,$F8,$5C,$38,$F3,$66,$A4,$FD,$C3
	.dc.b	$EA,$57,$0C,$BC,$19,$77,$CF,$F2,$B3,$D2,$BF,$18,$DC,$BC,$31,$CE
	.dc.b	$F8,$E3,$DB,$3F,$3C,$F7,$38,$E5,$F6,$B3,$F2,$C6,$70,$E7,$5F,$9E
	.dc.b	$F5,$1D,$3F,$EE,$EC,$7F,$D9,$FC,$3F,$E8,$9E,$6F,$E0,$5C,$F6,$7F
	.dc.b	$C0,$55,$BC,$BF,$82,$5E,$56,$1D,$23,$0F,$6B,$B5,$C7,$0F,$F4,$B4
	.dc.b	$BE,$69,$F1,$7C,$17,$D3,$58,$C3,$73,$FD,$2D,$F2,$AD,$C6,$0F,$F0
	.dc.b	$C3,$D1,$C6,$7E,$68,$DA,$9F,$E3,$9F,$45,$FE,$84,$8F,$F6,$D7,$6A
	.dc.b	$31,$1F,$B4,$8C,$5F,$F2,$E9,$CC,$FF,$2D,$7E,$23,$63,$C3,$97,$BF
	.dc.b	$E1,$33,$7C,$AB,$C5,$5F,$C0,$3C,$77,$D2,$7F,$AC,$E3,$D2,$3E,$9B
	.dc.b	$FA,$EB,$FE,$FD,$8D,$57,$F4,$F9,$CB,$D0,$57,$4F,$5B,$D6,$DB,$FE
	.dc.b	$8A,$D9,$FF,$25,$35,$FF,$7E,$39,$D3,$FC,$3D,$87,$81,$FA,$5F,$4D
	.dc.b	$F8,$39,$D5,$FA,$E9,$EA,$FD,$DF,$92,$D4,$62,$7E,$2F,$02,$73,$5E
	.dc.b	$6A,$D6,$5E,$46,$BE,$46,$BD,$E3,$8D,$F4,$F1,$AC,$FC,$3C,$7E,$6F
	.dc.b	$A1,$F8,$FF,$A6,$F1,$B6,$FE,$63,$E1,$23,$F1,$B1,$46,$2F,$A4,$62
	.dc.b	$7D,$7F,$39,$D1,$FA,$DF,$37,$F9,$DF,$8C,$F8,$2D,$8C,$9D,$5F,$80
	.dc.b	$47,$06,$AF,$EE,$72,$FE,$A7,$DC,$E7,$3F,$83,$EC,$FD,$57,$C9,$FC
	.dc.b	$E6,$C6,$4F,$53,$F3,$48,$E0,$D5,$F6,$E2,$FC,$B7,$27,$E0,$2B,$F6
	.dc.b	$DF,$4D,$2F,$F5,$BC,$2F,$A4,$FB,$51,$F7,$27,$F0,$91,$8C,$7F,$1A
	.dc.b	$D7,$F1,$AD,$7F,$44,$D7,$F7,$1F,$4D,$CB,$9B,$F8,$33,$6F,$C3,$C0
	.dc.b	$F7,$FF,$24,$D8,$FC,$6F,$BD,$89,$FD,$AF,$D2,$FE,$A3,$1F,$F5,$43
	.dc.b	$5F,$EE,$21,$C3,$FD,$D1,$BC,$5F,$C6,$7D,$0F,$E3,$07,$F1,$9F,$32
	.dc.b	$E3,$FC,$42,$F9,$61,$FC,$77,$CB,$F0,$B6,$75,$8F,$9E,$BF,$C6,$ED
	.dc.b	$7C,$B3,$FC,$3A,$31,$1E,$67,$87,$12,$B1,$5F,$CC,$BF,$01,$7C,$07
	.dc.b	$82,$7F,$5F,$F3,$7F,$CD,$BB,$9F,$3F,$FC,$C6,$30,$D9,$FE,$33,$3A
	.dc.b	$78,$98,$8E,$5F,$3D,$FC,$E7,$E0,$EC,$FE,$F1,$BA,$FE,$2D,$BA,$7F
	.dc.b	$39,$7F,$A0,$4F,$8F,$E7,$2B,$E0,$6D,$7E,$83,$14,$62,$FF,$A5,$C6
	.dc.b	$2B,$F8,$59,$CC,$A7,$33,$F2,$FC,$EC,$3F,$AB,$FC,$73,$E9,$FF,$2F
	.dc.b	$18,$6D,$70,$8D,$FD,$4E,$33,$FD,$9D,$FF,$5B,$AF,$8F,$F3,$7D,$3B
	.dc.b	$FE,$73,$63,$27,$F4,$49,$E1,$5E,$D9,$FC,$AF,$85,$E4,$F8,$6B,$7B
	.dc.b	$6D,$7E,$1A,$7F,$08,$49,$FF,$3F,$C7,$E6,$63,$7B,$7E,$16,$3F,$35
	.dc.b	$EF,$E2,$BF,$DF,$DB,$BE,$4F,$F0,$9F,$2D,$F8,$56,$CF,$E0,$9B,$7F
	.dc.b	$4D,$B8,$7F,$B8,$7E,$09,$F4,$1F,$8D,$FA,$2F,$E8,$91,$8C,$3F,$17
	.dc.b	$18,$99,$1F,$8E,$F9,$6B,$8F,$F5,$93,$F6,$5F,$0B,$FB,$8F,$80,$FD
	.dc.b	$CB,$6B,$F5,$2F,$E9,$48,$FC,$2E,$F4,$62,$BA,$46,$27,$F6,$3E,$DF
	.dc.b	$47,$F8,$DE,$6E,$BF,$D3,$7C,$AF,$A9,$B3,$A7,$0F,$E9,$BE,$02,$E3
	.dc.b	$FD,$3E,$DF,$0F,$FA,$F8,$F8,$6D,$AE,$EE,$1F,$F0,$3F,$CB,$3E,$5B
	.dc.b	$15,$F0,$FE,$9F,$F0,$EF,$F1,$60,$E1,$FE,$32,$D7,$FD,$16,$D7,$7D
	.dc.b	$9F,$E1,$68,$FC,$8F,$27,$E1,$7D,$07,$D3,$7C,$5B,$F0,$5F,$F7,$D8
	.dc.b	$C5,$6D,$B9,$7F,$A6,$3E,$99,$F4,$EF,$5F,$87,$2C,$AF,$E8,$F7,$C3
	.dc.b	$26,$1C,$2F,$51,$F8,$1E,$1F,$DE,$3F,$ED,$A6,$BF,$05,$AF,$F3,$86
	.dc.b	$BF,$EA,$A7,$8E,$5F,$A1,$8F,$F4,$6F,$8D,$FE,$36,$FC,$3F,$E7,$65
	.dc.b	$FC,$DE,$F6,$FF,$8C,$F8,$DF,$49,$FE,$28,$6C,$7F,$8A,$1B,$87,$F8
	.dc.b	$A1,$AF,$F8,$A1,$CE,$3F,$A5,$AF,$F3,$2F,$AA,$F1,$3F,$4B,$13,$0F
	.dc.b	$C8,$BA,$8C,$61,$F8,$B8,$B7,$FE,$6A,$BE,$57,$C0,$FC,$CF,$77,$9C
	.dc.b	$8F,$9D,$B7,$F9,$57,$F9,$0F,$D4,$7E,$D9,$FF,$3E,$7B,$0F,$A8,$78
	.dc.b	$BF,$C9,$BC,$2C,$C6,$31,$FE,$35,$18,$9F,$F3,$3D,$71,$6D,$6F,$1F
	.dc.b	$CC,$4E,$6F,$F3,$50,$8C,$3F,$E5,$96,$E7,$FE,$88,$C1,$F0,$1D,$EF
	.dc.b	$FC,$FB,$5E,$73,$F9,$DF,$9D,$E6,$6C,$6B,$1F,$F4,$D6,$3F,$DF,$A7
	.dc.b	$84,$F8,$9B,$1E,$69,$DB,$FE,$F4,$8C,$E3,$FC,$AD,$B5,$FE,$F8,$F5
	.dc.b	$9E,$5F,$F9,$CA,$B5,$87,$FA,$86,$B9,$37,$0F,$80,$DB,$E3,$F2,$9E
	.dc.b	$9A,$F6,$FD,$33,$CB,$FE,$AF,$E2,$7D,$68,$C4,$FF,$4B,$8C,$4F,$D2
	.dc.b	$9C,$E3,$F4,$1A,$FC,$16,$BE,$D7,$79,$7B,$5B,$7E,$DC,$DC,$DD,$A3
	.dc.b	$FC,$3D,$EA,$BD,$CF,$F5,$1B,$63,$FA,$9B,$71,$FB,$ED,$7F,$8D,$2F
	.dc.b	$53,$89,$27,$8B,$FE,$AE,$90,$F8,$BF,$A9,$E5,$7E,$29,$B1,$FC,$95
	.dc.b	$AF,$F8,$F1,$BF,$22,$F9,$6F,$F6,$14,$FE,$46,$BF,$C8,$33,$9D,$B7
	.dc.b	$5A,$E4,$56,$A3,$0F,$F9,$23,$37,$BB,$FD,$CB,$1F,$F2,$A3,$63,$31
	.dc.b	$F1,$23,$1F,$01,$23,$E3,$B5,$FB,$0F,$F9,$53,$1F,$A8,$DD,$7E,$C3
	.dc.b	$67,$FD,$E7,$C1,$FE,$26,$7F,$13,$1F,$E0,$4C,$DC,$28,$C3,$67,$F4
	.dc.b	$4D,$7C,$27,$F8,$5F,$CD,$ED,$CD,$F9,$E7,$C1,$7D,$2E,$DB,$DE,$A4
	.dc.b	$C3,$FC,$41,$F4,$F5,$FD,$F4,$ED,$FC,$FC,$F6,$C9,$5B,$9F,$CC,$4F
	.dc.b	$A0,$8F,$A1,$E1,$E0,$C8,$E6,$7E,$AB,$4F,$29,$1F,$F7,$AF,$D4,$3E
	.dc.b	$2F,$11,$FA,$79,$F8,$A4,$7E,$2B,$E5,$F1,$24,$75,$3F,$C1,$3E,$97
	.dc.b	$D3,$F3,$FF,$4E,$DE,$B3,$ED,$F5,$F0,$FF,$8E,$27,$75,$FD,$36,$31
	.dc.b	$5C,$23,$FD,$EF,$F4,$9F,$E0,$57,$FD,$62,$D7,$FD,$BC,$DE,$5F,$F2
	.dc.b	$07,$2B,$EA,$3F,$DD,$D5,$C3,$D4,$F3,$5E,$F5,$FD,$6E,$68,$C5,$30
	.dc.b	$E5,$E5,$70,$C2,$4F,$33,$F6,$B9,$78,$32,$EF,$1E,$49,$F7,$61,$E7
	.dc.b	$8F,$F5,$7E,$3D,$E7,$27,$EA,$7E,$A3,$EA,$4F,$FA,$AD,$AF,$F9,$15
	.dc.b	$B8,$7F,$7D,$FC,$C1,$F3,$1F,$A3,$A5,$70,$7F,$C5,$0D,$7F,$D7,$ED
	.dc.b	$7F,$86,$B5,$7D,$4F,$F5,$57,$FD,$78,$D7,$FC,$AC,$DB,$FF,$37,$0F
	.dc.b	$65,$EF,$F8,$E7,$5C,$1F,$EC,$E6,$DF,$F1,$C6,$DF,$1A,$F5,$7D,$4F
	.dc.b	$EC,$E3,$FC,$34,$DB,$FF,$6B,$B6,$F9,$77,$70,$EE,$FE,$25,$3F,$BB
	.dc.b	$9F,$F2,$43,$5E,$31,$F5,$3E,$43,$F0,$64,$8F,$F2,$9F,$D3,$55,$8F
	.dc.b	$E3,$34,$7C,$E7,$F0,$46,$CF,$E5,$2F,$F1,$BE,$24,$8F,$CB,$BB,$FA
	.dc.b	$5C,$47,$4D,$7F,$B5,$5A,$FA,$1A,$C7,$FC,$75,$F5,$2E,$CF,$BC,$DC
	.dc.b	$3F,$DE,$0D,$7F,$D1,$4F,$F8,$BD,$CC,$7F,$35,$8D,$47,$F9,$AD,$AF
	.dc.b	$F4,$E0,$E1,$FE,$A2,$F5,$9F,$1F,$C2,$78,$8F,$1D,$7E,$B3,$E4,$A5
	.dc.b	$FE,$D9,$F8,$DF,$4D,$88,$F8,$07,$EB,$72,$7C,$AF,$CB,$35,$8F,$CF
	.dc.b	$F8,$DF,$9D,$BC,$BF,$F4,$3C,$7D,$13,$0F,$4E,$FF,$C9,$4E,$63,$E2
	.dc.b	$FC,$47,$C8,$23,$F2,$3D,$BA,$FF,$8D,$37,$4F,$C7,$24,$7F,$B3,$A3
	.dc.b	$2F,$F0,$0E,$1F,$F3,$87,$D4,$FF,$6D,$8C,$0F,$79,$19,$8F,$AB,$7F
	.dc.b	$CF,$F8,$BF,$15,$C3,$27,$FD,$CD,$19,$7F,$D4,$B1,$97,$FD,$E3,$1F
	.dc.b	$C6,$BE,$26,$47,$F2,$47,$FA,$56,$78,$8F,$F7,$BD,$FF,$4D,$31,$BF
	.dc.b	$F2,$6F,$37,$F9,$74,$FC,$9E,$27,$D1,$7E,$74,$FE,$A9,$E4,$49,$FF
	.dc.b	$5B,$C6,$5F,$F7,$E4,$65,$E5,$19,$7E,$A5,$76,$BF,$D5,$7B,$FF,$D7
	.dc.b	$F7,$95,$FE,$C2,$8F,$D6,$18,$78,$C6,$1F,$9B,$C7,$FE,$50,$7F,$65
	.dc.b	$F0,$D1,$CE,$1F,$EC,$7C,$3D,$D1,$93,$DD,$3F,$DB,$DF,$E5,$1F,$53
	.dc.b	$FB,$17,$EC,$87,$0F,$FB,$D5,$B3,$FD,$3B,$E4,$7F,$98,$CF,$09,$DB
	.dc.b	$FE,$D7,$6B,$FA,$D6,$BF,$E3,$C6,$BD,$FA,$FF,$C1,$FF,$4D,$FD,$4B
	.dc.b	$23,$FD,$52,$74,$7C,$27,$F8,$8B,$FA,$16,$E1,$FD,$89,$BC,$DF,$E7
	.dc.b	$75,$DF,$F0,$3A,$7F,$C7,$19,$1A,$4C,$3F,$D2,$FE,$4B,$EA,$5E,$07
	.dc.b	$F7,$F6,$CF,$E8,$27,$53,$EA,$3D,$7A,$FF,$81,$7E,$A5,$F3,$34,$CD
	.dc.b	$F1,$8C,$BE,$06,$BF,$F0,$06,$D1,$88,$FE,$D6,$D7,$DD,$E0,$FC,$A2
	.dc.b	$3D,$6B,$C5,$EC,$C8,$F0,$3F,$43,$C2,$C4,$77,$BF,$F2,$CB,$67,$FC
	.dc.b	$31,$18,$8F,$0C,$7D,$8D,$24,$74,$3D,$9F,$27,$F0,$48,$FD,$9F,$0E
	.dc.b	$BD,$E3,$A9,$1F,$84,$FE,$13,$EA,$64,$74,$3E,$87,$D4,$FF,$70,$6F
	.dc.b	$97,$FE,$33,$8C,$47,$E1,$7C,$4C,$4F,$F8,$C6,$73,$3F,$E1,$46,$BF
	.dc.b	$E2,$46,$AD,$D1,$D3,$9F,$F1,$DE,$77,$E6,$0E,$B5,$E4,$F8,$5F,$11
	.dc.b	$1F,$CC,$7B,$F5,$FB,$96,$9F,$21,$23,$F2,$31,$97,$F4,$1F,$53,$FA
	.dc.b	$63,$96,$0E,$BF,$47,$BF,$F0,$49,$5F,$A7,$F0,$EB,$E0,$8F,$84,$47
	.dc.b	$BB,$FE,$8D,$9F,$F6,$0E,$77,$D8,$FD,$FD,$7F,$5A,$D7,$E1,$DF,$F8
	.dc.b	$C7,$E8,$3F,$DB,$13,$FC,$29,$AF,$F6,$00,$E3,$2F,$DB,$7C,$B7,$1F
	.dc.b	$87,$FF,$55,$4F,$0A,$FE,$E8,$70,$8C,$98,$7F,$79,$E6,$7F,$08,$F2
	.dc.b	$24,$7F,$05,$8F,$E1,$66,$1F,$E0,$D1,$87,$F8,$57,$C1,$FF,$10,$23
	.dc.b	$FC,$37,$C7,$AF,$F0,$D8,$B9,$7F,$0D,$9F,$F2,$87,$D2,$FF,$A8,$7D
	.dc.b	$6F,$F6,$BF,$C3,$32,$FF,$06,$4E,$71,$FF,$6D,$46,$3E,$0F,$FC,$9D
	.dc.b	$5B,$97,$D3,$BE,$A7,$F8,$EC,$E6,$58,$73,$8C,$37,$0E,$98,$F8,$9F
	.dc.b	$F7,$4E,$1C,$FD,$8F,$F7,$24,$6B,$0F,$19,$FE,$EE,$BE,$4D,$CB,$CB
	.dc.b	$3C,$B2,$F1,$BC,$72,$F0,$CE,$AB,$86,$5E,$17,$D7,$7C,$1C,$9E,$03
	.dc.b	$D7,$3C,$5B,$3C,$72,$F4,$47,$33,$27,$E0,$71,$F9,$8F,$E1,$9A,$FB
	.dc.b	$4F,$C4,$FC,$47,$FA,$1B,$EA,$7F,$11,$08,$C3,$F8,$B6,$BC,$39,$D3
	.dc.b	$C2,$9F,$3D,$FE,$1D,$AF,$13,$5F,$13,$5F,$C0,$B5,$FF,$56,$7D,$8B
	.dc.b	$E9,$FF,$71,$1F,$4B,$D2,$FE,$BF,$A6,$37,$FC,$24,$2F,$87,$67,$CD
	.dc.b	$8D,$B6,$33,$7C,$FD,$26,$1C,$F8,$93,$E2,$CB,$0F,$FA,$FE,$B7,$87
	.dc.b	$5F,$88,$FE,$28,$DF,$0B,$79,$FF,$BA,$7E,$46,$D6,$B0,$F5,$37,$5F
	.dc.b	$4B,$73,$F0,$DF,$1C,$7A,$12,$F8,$E3,$D3,$1E,$AF,$83,$5F,$F7,$C5
	.dc.b	$7F,$88,$97,$D4,$7F,$AC,$A3,$AC,$6A,$FF,$CC,$8E,$A3,$86,$5B,$9F
	.dc.b	$A7,$57,$1F,$9D,$36,$F8,$57,$2F,$89,$C1,$9F,$93,$B8,$F8,$E2,$CF
	.dc.b	$82,$6F,$5F,$FB,$D3,$F3,$14,$77,$7C,$D1,$98,$C6,$39,$6E,$1F,$E4
	.dc.b	$0A,$E2,$D9,$F4,$7A,$6F,$3A,$F9,$F8,$6D,$7F,$92,$27,$4D,$9D,$34
	.dc.b	$D4,$F0,$CB,$FD,$19,$5C,$1B,$5E,$9F,$0B,$D5,$A5,$C3,$CC,$DA,$FF
	.dc.b	$4A,$64,$8C,$69,$F0,$DB,$5F,$84,$9D,$3E,$FB,$EA,$FA,$7C,$DF,$7F
	.dc.b	$9D,$F0,$1B,$19,$3D,$D1,$96,$99,$C7,$D7,$19,$6C,$7C,$DF,$99,$FA
	.dc.b	$9F,$4F,$F5,$5B,$19,$3E,$94,$65,$A6,$67,$DF,$19,$6C,$7D,$1F,$3F
	.dc.b	$EB,$F7,$FD,$86,$C6,$4F,$AD,$19,$69,$99,$FA,$71,$96,$C7,$E1,$7E
	.dc.b	$97,$F1,$87,$E3,$7E,$1E,$C7,$E2,$5B,$3F,$51,$BB,$FE,$55,$E1,$3F
	.dc.b	$8F,$F5,$ED,$7D,$AF,$A9,$FC,$C1,$F9,$0F,$4A,$D7,$E3,$7C,$EC,$34
	.dc.b	$FC,$84,$7E,$5B,$D2,$91,$F9,$BF,$96,$B8,$7F,$B9,$3E,$E7,$F9,$7F
	.dc.b	$89,$FB,$55,$AF,$F1,$49,$D7,$91,$FC,$7A,$3F,$94,$CE,$BC,$6E,$13
	.dc.b	$9C,$7F,$8F,$CE,$5F,$F6,$AB,$D9,$B5,$D4,$E7,$3C,$6D,$5C,$5F,$0E
	.dc.b	$1E,$5A,$E1,$7F,$99,$F6,$3F,$79,$F5,$1F,$BD,$6C,$7E,$E5,$B7,$FA
	.dc.b	$46,$D7,$F4,$2F,$83,$7D,$EF,$E8,$AB,$E0,$DF,$F3,$89,$CF,$85,$FC
	.dc.b	$65,$DF,$75,$BF,$E0,$B7,$C0,$FF,$79,$3C,$F0,$EC,$E9,$B3,$C6,$FF
	.dc.b	$8A,$79,$FF,$C5,$7E,$A6,$E5,$FC,$45,$B8,$FE,$F1,$B8,$7F,$1D,$6C
	.dc.b	$FE,$ED,$B2,$F1,$9E,$0A,$7F,$99,$4B,$E6,$9E,$AF,$F3,$3F,$88,$B8
	.dc.b	$FF,$B5,$26,$D7,$FB,$DE,$B8,$C6,$23,$FC,$8E,$D7,$FA,$F9,$D6,$3F
	.dc.b	$B7,$BA,$6B,$FC,$47,$DE,$FE,$69,$3F,$CD,$5B,$5F,$CC,$5B,$3F,$C7
	.dc.b	$1B,$9F,$E2,$3C,$9F,$C8,$F9,$7F,$92,$6D,$FE,$43,$9F,$F9,$FB,$F9
	.dc.b	$47,$FA,$6D,$71,$8D,$57,$16,$9C,$6B,$9E,$1F,$77,$67,$E5,$9F,$97
	.dc.b	$5E,$57,$E5,$27,$E8,$4E,$63,$13,$C3,$A3,$FA,$4F,$57,$F6,$1A,$D3
	.dc.b	$F5,$3F,$29,$F3,$E1,$FE,$8D,$FC,$B9,$F4,$79,$CF,$FB,$56,$89,$CB
	.dc.b	$FE,$0E,$DB,$E7,$E1,$FD,$CF,$0F,$F3,$DB,$5F,$AA,$D7,$FC,$2A,$DE
	.dc.b	$DF,$F1,$8F,$4F,$EF,$3A,$7F,$7A,$DC,$3E,$63,$6F,$F8,$2B,$70,$FD
	.dc.b	$DB,$67,$E8,$36,$7B,$F8,$7F,$AB,$FC,$AF,$FC,$F1,$5E,$93,$CB,$19
	.dc.b	$F4,$FE,$23,$F8,$8A,$E3,$3F,$E4,$CC,$FF,$DF,$AD,$9D,$4B,$EA,$5F
	.dc.b	$53,$FE,$11,$CD,$F6,$5F,$3A,$B5,$39,$8C,$57,$0C,$3F,$CB,$2D,$AF
	.dc.b	$F5,$13,$5F,$F5,$3B,$5F,$F1,$7F,$E6,$4F,$AE,$FF,$08,$57,$F8,$7E
	.dc.b	$30,$FF,$83,$4C,$46,$9F,$85,$3F,$B9,$9D,$BF,$DF,$23,$96,$11,$DF
	.dc.b	$3A,$9F,$5D,$9B,$FC,$C4,$7A,$73,$FF,$30,$D6,$F0,$F2,$5E,$E7,$3E
	.dc.b	$BB,$D1,$FF,$1A,$7A,$BF,$E3,$56,$E7,$FE,$43,$6E,$7F,$E4,$06,$DF
	.dc.b	$2F,$43,$FC,$79,$E9,$FF,$8F,$5B,$7F,$E4,$B6,$DF,$F9,$21,$B5,$C6
	.dc.b	$7F,$CA,$15,$C5,$A7,$12,$D7,$16,$CF,$AB,$D1,$FF,$26,$71,$7F,$93
	.dc.b	$5B,$3F,$E5,$B6,$D7,$F9,$61,$B5,$F0,$BC,$DF,$F2,$A7,$A3,$FE,$55
	.dc.b	$6D,$FF,$98,$DB,$7F,$E6,$06,$D7,$19,$FF,$34,$57,$16,$9F,$63,$B1
	.dc.b	$F2,$7F,$E6,$35,$F2,$96,$30,$D9,$E0,$DC,$7F,$8A,$57,$09,$E9,$5C
	.dc.b	$1A,$70,$2C,$F0,$AE,$58,$7E,$12,$F9,$4F,$F1,$FF,$C7,$FF,$D1,$D3
	.dc.b	$F8,$85,$EE,$7F,$9E,$5C,$E1,$FE,$8E,$72,$FE,$51,$CD,$7F,$53,$72
	.dc.b	$F5,$F7,$39,$57,$36,$D7,$16,$CF,$F6,$D3,$FA,$6F,$BB,$FE,$AC,$FC
	.dc.b	$CD,$F3,$7F,$D4,$25,$EA,$7F,$AF,$A7,$31,$9A,$E1,$5B,$AD,$3C,$65
	.dc.b	$3A,$7F,$AD,$CE,$9E,$42,$BA,$BF,$E0,$7C,$38,$3E,$63,$FD,$87,$E7
	.dc.b	$F5,$BF,$F6,$1F,$91,$FD,$C2,$79,$B5,$E0,$DF,$83,$7B,$9F,$E3,$49
	.dc.b	$E6,$7F,$B3,$23,$3B,$F4,$C3,$FB,$8C,$62,$76,$7F,$BC,$6B,$95,$F5
	.dc.b	$7A,$65,$EA,$A7,$B5,$FD,$EA,$30,$DD,$7A,$D7,$06,$E5,$F0,$2F,$8E
	.dc.b	$36,$F8,$3E,$7B,$E0,$FE,$0F,$E3,$3F,$C0,$F7,$F8,$DF,$A8,$BB,$7E
	.dc.b	$2D,$B5,$B3,$F1,$95,$8F,$05,$C6,$AC,$36,$B6,$43,$F3,$91,$A3,$75
	.dc.b	$CF,$93,$F1,$58,$FE,$8F,$8B,$06,$6E,$1F,$B9,$72,$FE,$EC,$C9,$FB
	.dc.b	$47,$2F,$ED,$8C,$C7,$57,$91,$C4,$FF,$77,$7B,$5E,$DF,$01,$ED,$D3
	.dc.b	$11,$27,$FD,$E9,$19,$78,$90,$FF,$79,$B9,$7F,$DE,$B1,$D7,$DA,$ED
	.dc.b	$EC,$C9,$FF,$7F,$46,$5F,$F8,$01,$19,$7F,$DF,$B1,$E5,$F6,$3C,$DF
	.dc.b	$09,$27,$FA,$9E,$3F,$E4,$D6,$CE,$5F,$F1,$54,$27,$94,$68,$E2,$EF
	.dc.b	$2F,$EA,$75,$A9,$E3,$39,$E1,$47,$F9,$7E,$FC,$0F,$68,$F9,$5E,$CC
	.dc.b	$8F,$2C,$7C,$D3,$0F,$92,$30,$FF,$6C,$F2,$5F,$57,$F8,$F6,$C6,$63
	.dc.b	$1F,$21,$96,$46,$4E,$51,$C7,$BF,$97,$D5,$C9,$ED,$19,$7C,$11,$97
	.dc.b	$FB,$CC,$7E,$89,$7C,$FB,$C6,$FE,$31,$1F,$0A,$30,$E1,$AF,$F1,$D6
	.dc.b	$BF,$C9,$DC,$EF,$FE,$76,$FA,$07,$C7,$77,$6D,$7F,$85,$0F,$0D,$F8
	.dc.b	$FF,$2D,$67,$E2,$94,$61,$F2,$B5,$FE,$5C,$D7,$F9,$C3,$9C,$3F,$D0
	.dc.b	$FF,$54,$FA,$0F,$80,$DA,$F1,$56,$F8,$7F,$D1,$F3,$FE,$66,$9F,$32
	.dc.b	$EB,$E2,$72,$F7,$64,$EA,$32,$F0,$8C,$BF,$D3,$E7,$93,$B3,$47,$8A
	.dc.b	$3E,$05,$F8,$63,$DB,$18,$23,$E4,$E1,$F5,$9F,$26,$3F,$8D,$7B,$DF
	.dc.b	$98,$CF,$FC,$6A,$FE,$96,$BF,$C7,$AD,$7F,$D2,$4E,$F2,$FD,$0F,$81
	.dc.b	$FE,$94,$E8,$E7,$D1,$2F,$8C,$7C,$AF,$53,$E5,$FA,$92,$7E,$3C,$65
	.dc.b	$F9,$31,$97,$E1,$46,$5F,$89,$19,$CB,$EB,$CF,$E9,$7C,$1F,$D3,$79
	.dc.b	$72,$FF,$55,$19,$7F,$59,$19,$7F,$CA,$B1,$F8,$BF,$A8,$F9,$64,$8F
	.dc.b	$C4,$B5,$FF,$59,$7C,$23,$CB,$FE,$84,$D8,$FC,$2C,$FE,$DF,$E5,$7F
	.dc.b	$71,$C3,$27,$F7,$91,$97,$F7,$D1,$97,$FC,$CF,$19,$7F,$CD,$31,$9C
	.dc.b	$FF,$CD,$3D,$CA,$FE,$AB,$E2,$BF,$34,$7C,$56,$17,$C5,$53,$11,$32
	.dc.b	$FF,$3F,$D7,$19,$D4,$62,$B9,$BF,$C4,$FE,$AF,$F8,$97,$54,$8C,$10
	.dc.b	$FF,$3F,$B9,$7F,$D0,$31,$FC,$27,$F3,$57,$8B,$79,$A5,$7E,$F6,$72
	.dc.b	$D9,$CB,$FE,$31,$8F,$C8,$3F,$CC,$1D,$B9,$8D,$ED,$FF,$00,$27,$72
	.dc.b	$77,$1F,$E9,$2F,$CD,$BF,$D6,$B3,$FE,$9A,$6B,$FE,$EF,$6C,$7D,$0E
	.dc.b	$1F,$F0,$04,$7F,$01,$97,$4A,$3F,$9A,$C7,$1E,$CD,$F8,$1F,$CE,$7D
	.dc.b	$19,$3F,$CF,$5E,$73,$FC,$F7,$E9,$64,$FF,$41,$7F,$A0,$78,$58,$8F
	.dc.b	$F5,$5F,$AD,$FD,$0C,$C6,$9F,$84,$71,$1F,$D2,$5C,$F8,$1F,$B6,$CF
	.dc.b	$FC,$D3,$3E,$FF,$27,$D6,$7B,$3C,$99,$19,$21,$F0,$DC,$BF,$16,$38
	.dc.b	$BF,$26,$3F,$AB,$73,$FF,$57,$F5,$E4,$FF,$5B,$8C,$BF,$D7,$63,$2F
	.dc.b	$FB,$1A,$32,$FF,$B2,$23,$39,$7E,$82,$7E,$8F,$7F,$D2,$F0,$25,$FD
	.dc.b	$98,$CB,$F6,$E3,$2F,$DB,$8C,$BF,$76,$33,$5F,$83,$9F,$F1,$E7,$C5
	.dc.b	$3E,$AB,$FB,$85,$C3,$FD,$BF,$19,$6C,$65,$FF,$6F,$9C,$98,$FF,$7B
	.dc.b	$35,$5F,$EE,$29,$E5,$1F,$8F,$F0,$3F,$1B,$E4,$48,$FC,$A4,$7E,$40
	.dc.b	$C3,$F9,$58,$C3,$A3,$F3,$4E,$AB,$F3,$D3,$EC,$F9,$9F,$6F,$D2,$4B
	.dc.b	$F5,$C6,$5F,$6C,$65,$57,$9E,$7F,$CA,$DE,$3C,$9F,$F6,$53,$FE,$59
	.dc.b	$F8,$3C,$47,$FB,$12,$2D,$7F,$AF,$67,$FC,$BD,$D3,$FE,$5F,$F2,$25
	.dc.b	$FF,$B4,$E3,$2F,$FB,$56,$32,$FF,$B3,$A3,$FC,$E1,$F7,$EF,$93,$FF
	.dc.b	$59,$FC,$4F,$F5,$28,$C5,$7F,$8D,$3D,$F4,$E6,$B7,$3F,$B8,$A7,$CA
	.dc.b	$E3,$C0,$47,$29,$FF,$59,$DF,$17,$A3,$5C,$9C,$EB,$94,$F1,$C3,$C3
	.dc.b	$3D,$F0,$C9,$AA,$F0,$7B,$4E,$AB,$E0,$4B,$EE,$7F,$0E,$9D,$38,$CD
	.dc.b	$18,$8F,$1C,$62,$79,$46,$27,$58,$F9,$21,$F3,$2F,$63,$FA,$D4,$7F
	.dc.b	$B6,$E3,$1E,$87,$F8,$F9,$B1,$FE,$D7,$6B,$FE,$E1,$F7,$FE,$A7,$C9
	.dc.b	$FF,$91,$1B,$1F,$ED,$D6,$BB,$D9,$3D,$7D,$3F,$F7,$4C,$F5,$6B,$E3
	.dc.b	$6B,$FE,$70,$3D,$2E,$1F,$F3,$97,$C9,$3C,$4F,$8D,$F3,$3F,$DD,$DF
	.dc.b	$91,$EE,$27,$55,$F2,$2B,$4F,$CA,$9D,$57,$2D,$37,$B7,$DC,$E1,$56
	.dc.b	$AF,$95,$6B,$97,$FC,$4B,$C2,$F0,$B1,$5F,$6D,$B8,$7D,$AF,$16,$D6
	.dc.b	$F3,$FB,$CD,$BF,$B5,$7A,$AD,$B2,$B6,$E5,$A7,$C0,$AF,$7F,$5F,$C0
	.dc.b	$C3,$D3,$F2,$98,$AF,$F1,$DE,$9F,$8C,$7F,$D5,$71,$9C,$BF,$D3,$D8
	.dc.b	$5B,$E3,$5F,$EB,$08,$C5,$AE,$FC,$8C,$AF,$F5,$AF,$C9,$5C,$7F,$D6
	.dc.b	$73,$72,$F2,$60,$AC,$5F,$F9,$3B,$E5,$70,$FF,$AC,$2B,$D5,$1E,$7C
	.dc.b	$3F,$8B,$CE,$EB,$CD,$39,$AD,$C6,$32,$CB,$6B,$F9,$EF,$CC,$FB,$7C
	.dc.b	$67,$CC,$D8,$C3,$67,$FA,$1B,$72,$F0,$9F,$E0,$4A,$E3,$3D,$0D,$96
	.dc.b	$37,$B2,$77,$18,$CD,$F2,$3E,$AD,$BD,$6D,$C3,$E1,$1F,$15,$C5,$71
	.dc.b	$6E,$BF,$19,$B3,$B3,$8C,$FF,$86,$AD,$7F,$07,$BF,$E1,$7C,$CD,$BF
	.dc.b	$EC,$2D,$BF,$4E,$48,$C6,$5F,$D7,$5B,$5F,$E8,$69,$D3,$FD,$AA,$FA
	.dc.b	$BE,$2C,$D9,$3E,$43,$EC,$FC,$97,$DA,$6D,$6E,$74,$D9,$D3,$4D,$4F
	.dc.b	$0C,$FE,$63,$73,$FE,$E5,$F1,$7F,$7F,$BB,$F0,$0D,$BF,$EF,$ED,$BF
	.dc.b	$7B,$6B,$8C,$FF,$81,$2B,$8B,$4F,$37,$C9,$7E,$48,$FC,$9F,$C2,$D8
	.dc.b	$F5,$37,$0F,$C0,$36,$FF,$3A,$F0,$9F,$81,$F2,$56,$BF,$0D,$F2,$3F
	.dc.b	$9E,$3F,$31,$F1,$F6,$BF,$C2,$FF,$07,$86,$9F,$E1,$A8,$FC,$F7,$C1
	.dc.b	$C8,$F9,$9F,$25,$70,$FC,$DF,$CB,$FF,$36,$F1,$7F,$81,$3F,$E0,$4F
	.dc.b	$0E,$F8,$FF,$C8,$77,$FE,$57,$39,$C7,$F1,$AF,$16,$C7,$F2,$E3,$73
	.dc.b	$C6,$18,$2B,$93,$E2,$9D,$39,$AE,$55,$C2,$BB,$4F,$A2,$72,$FE,$77
	.dc.b	$E3,$FF,$81,$FC,$87,$F0,$46,$DF,$E2,$DB,$3F,$B1,$6D,$7F,$46,$F6
	.dc.b	$9F,$33,$FA,$6B,$FF,$54,$BF,$C5,$2F,$F8,$0C,$EB,$BB,$F9,$43,$C3
	.dc.b	$C2,$B5,$C6,$B6,$F1,$C3,$B3,$96,$CF,$F1,$AF,$3B,$F8,$DF,$C8,$DC
	.dc.b	$BF,$44,$DC,$3F,$84,$36,$FF,$92,$B6,$7F,$95,$36,$7F,$80,$3F,$CF
	.dc.b	$57,$C3,$3E,$53,$FB,$2C,$7F,$6F,$32,$E1,$FE,$E5,$1B,$AD,$47,$DA
	.dc.b	$8C,$38,$6D,$F4,$6C,$FF,$3D,$F3,$FC,$51,$8A,$FE,$29,$F0,$FF,$CE
	.dc.b	$A3,$F9,$DB,$6B,$F8,$13,$67,$F9,$43,$6F,$F1,$BE,$4F,$E5,$BC,$BF
	.dc.b	$CB,$B6,$FF,$2B,$C9,$FD,$15,$FC,$53,$FD,$65,$CC,$6F,$0F,$C9,$36
	.dc.b	$FA,$CF,$2B,$FE,$C9,$F1,$DF,$96,$F1,$7F,$9E,$73,$E3,$A3,$FA,$63
	.dc.b	$6F,$FD,$16,$DC,$BF,$95,$BF,$4F,$E5,$FF,$AE,$47,$D1,$6B,$FE,$21
	.dc.b	$6B,$FA,$86,$A7,$D8,$E7,$3F,$C9,$FC,$C3,$E7,$3B,$47,$EE,$BA,$3F
	.dc.b	$77,$CD,$72,$F8,$AD,$C3,$F6,$4D,$DF,$F7,$2D,$9F,$92,$D9,$FD,$D7
	.dc.b	$07,$E7,$67,$D1,$E0,$FE,$26,$7D,$0D,$7D,$0D,$D0,$D3,$E2,$1F,$E1
	.dc.b	$3F,$9E,$7D,$9B,$C2,$CE,$1F,$D7,$30,$7C,$D3,$DE,$EC,$D9,$FC,$BC
	.dc.b	$EA,$72,$48,$E5,$3B,$9F,$EF,$B8,$78,$3E,$5F,$F5,$13,$9D,$BF,$C5
	.dc.b	$9F,$28,$F8,$1F,$F0,$8E,$3F,$E2,$78,$C3,$FD,$64,$C4,$69,$FE,$C5
	.dc.b	$1A,$C7,$FC,$05,$1C,$72,$8E,$BE,$0A,$CF,$F8,$2A,$BC,$53,$FC,$AB
	.dc.b	$A1,$18,$94,$62,$B8,$E3,$FE,$30,$F3,$FF,$C7,$3E,$A7,$F8,$E9,$B8
	.dc.b	$FF,$75,$F5,$D1,$8C,$7F,$C3,$31,$89,$D5,$7F,$90,$BC,$EF,$F2,$1F
	.dc.b	$A3,$67,$93,$6F,$8B,$71,$FF,$20,$B6,$7F,$CA,$6D,$9F,$4E,$1F,$E0
	.dc.b	$07,$C3,$7D,$AB,$CC,$D8,$F2,$F9,$FF,$E5,$4E,$5F,$F2,$AB,$6B,$D8
	.dc.b	$D9,$DC,$7B,$E7,$FC,$B7,$E6,$7F,$97,$3C,$FB,$3F,$0D,$B5,$F0,$9B
	.dc.b	$9F,$F9,$69,$B3,$FE,$6C,$6C,$FA,$9F,$F3,$57,$63,$E2,$6F,$C5,$E1
	.dc.b	$B7,$FE,$25,$9D,$36,$74,$D3,$53,$C2,$FF,$C9,$B5,$C1,$B5,$C1,$B5
	.dc.b	$C0,$E4,$FF,$95,$3E,$67,$FD,$21,$8F,$FA,$1D,$7B,$9F,$E8,$17,$35
	.dc.b	$F8,$F7,$2F,$F8,$ED,$CD,$7E,$41,$CB,$D7,$DC,$E5,$5C,$DB,$5C,$5B
	.dc.b	$3F,$E6,$C3,$FD,$81,$EF,$7F,$81,$BE,$69,$F2,$1F,$D5,$7D,$2F,$F5
	.dc.b	$DC,$62,$BF,$AC,$56,$EB,$58,$59,$C8,$AE,$0F,$F8,$4A,$F9,$D7,$F5
	.dc.b	$CB,$E1,$39,$BA,$FF,$86,$23,$FC,$33,$E4,$7F,$69,$8E,$AD,$7D,$21
	.dc.b	$C3,$C4,$C7,$9F,$FE,$DD,$F8,$27,$B4,$DF,$AC,$E6,$31,$5C,$0F,$F1
	.dc.b	$4D,$70,$AD,$C6,$27,$77,$E3,$AD,$4F,$C1,$CB,$C1,$EC,$FF,$BB,$75
	.dc.b	$F8,$74,$8C,$4F,$87,$1F,$87,$3C,$AB,$75,$AB,$F1,$B5,$4E,$EF,$F1
	.dc.b	$1F,$17,$FE,$FB,$C3,$F1,$DF,$1B,$76,$FC,$63,$70,$FC,$55,$5F,$01
	.dc.b	$C6,$8C,$2A,$A6,$22,$57,$E5,$63,$2F,$E5,$E3,$A7,$2F,$F8,$7F,$1F
	.dc.b	$F7,$32,$E2,$C9,$AA,$87,$ED,$5C,$BF,$B7,$32,$7E,$C9,$CB,$FB,$43
	.dc.b	$33,$DA,$32,$F8,$23,$DD,$D8,$F6,$2F,$B3,$27,$F9,$A4,$68,$C3,$FC
	.dc.b	$D2,$30,$E8,$FE,$74,$EA,$BF,$DD,$F3,$DB,$D8,$EF,$EC,$4B,$FE,$81
	.dc.b	$19,$7F,$A1,$46,$5F,$E8,$31,$E6,$F5,$FC,$FE,$C4,$9F,$E9,$11,$97
	.dc.b	$FA,$54,$65,$FE,$93,$19,$7F,$A5,$C6,$70,$70,$FF,$5B,$CF,$FC,$AB
	.dc.b	$5D,$0E,$64,$39,$9F,$D5,$E3,$E4,$FB,$12,$30,$43,$FA,$CB,$97,$FA
	.dc.b	$DC,$7E,$37,$C8,$7C,$35,$24,$FF,$76,$F8,$9A,$FF,$71,$8B,$8F,$F8
	.dc.b	$0A,$79,$77,$F3,$F8,$89,$7F,$DF,$A3,$2F,$F8,$02,$32,$FF,$7B,$8F
	.dc.b	$D0,$2F,$9A,$78,$F8,$6C,$7F,$87,$DA,$FF,$27,$F9,$AF,$E1,$CE,$67
	.dc.b	$FC,$3E,$F3,$6C,$F0,$8F,$F3,$F3,$5F,$F5,$98,$F9,$A6,$3E,$46,$C7
	.dc.b	$F8,$BD,$AF,$F3,$7F,$97,$78,$DB,$AF,$0D,$6A,$BF,$D1,$8D,$7B,$B5
	.dc.b	$FF,$61,$D3,$E4,$3E,$1B,$FB,$F8,$FF,$1F,$35,$F0,$46,$8D,$F5,$C8
	.dc.b	$DC,$65,$E3,$19,$78,$9E,$B8,$F7,$11,$EF,$A5,$7F,$D3,$F7,$E5,$7F
	.dc.b	$C9,$6D,$73,$3F,$E4,$EE,$E7,$BF,$F9,$A9,$FE,$79,$C7,$FE,$37,$87
	.dc.b	$37,$F1,$D6,$CF,$FB,$93,$8B,$F4,$6F,$F8,$EA,$39,$46,$2F,$AF,$3F
	.dc.b	$1B,$97,$C7,$0E,$D3,$F2,$BD,$2F,$97,$E9,$4A,$F9,$71,$97,$E6,$C6
	.dc.b	$5F,$B9,$19,$7E,$F4,$67,$1F,$F6,$3E,$FF,$E6,$F9,$FF,$2D,$F8,$1F
	.dc.b	$AB,$F2,$A4,$60,$87,$F9,$65,$CB,$FE,$5B,$8F,$C7,$7C,$AC,$8F,$C7
	.dc.b	$3F,$B5,$E2,$C4,$7D,$28,$FC,$69,$1F,$F6,$9E,$DF,$E6,$2F,$99,$FD
	.dc.b	$E7,$0C,$9F,$F3,$7C,$7E,$FC,$C3,$FE,$6B,$8C,$3A,$3F,$CE,$AE,$B3
	.dc.b	$FF,$3F,$CF,$F5,$5F,$09,$F3,$3F,$E9,$3B,$FA,$07,$F4,$95,$F0,$DB
	.dc.b	$6C,$6A,$78,$FB,$79,$35,$52,$75,$5D,$5F,$F4,$07,$CC,$FF,$17,$E0
	.dc.b	$91,$FE,$8A,$8F,$E3,$46,$1F,$F4,$34,$61,$FF,$53,$7C,$CB,$C2,$FE
	.dc.b	$1E,$FE,$33,$9B,$F8,$0B,$64,$1E,$0F,$29,$E3,$CD,$23,$04,$3F,$C7
	.dc.b	$53,$CD,$E5,$1E,$09,$FE,$F9,$F2,$FF,$89,$9F,$D2,$9F,$E9,$87,$1B
	.dc.b	$FE,$7D,$B8,$7F,$A7,$FC,$37,$1F,$2E,$DE,$7E,$1C,$C7,$E9,$F2,$B1
	.dc.b	$FE,$2F,$97,$FA,$E3,$CB,$FE,$80,$4A,$DC,$FF,$42,$7F,$D6,$BF,$33
	.dc.b	$FD,$14,$8F,$FA,$E3,$E6,$6B,$FE,$B3,$F0,$BF,$A5,$18,$C3,$FD,$8A
	.dc.b	$E2,$3F,$D7,$EE,$76,$FF,$5C,$61,$FE,$6E,$9F,$0F,$8D,$F1,$0F,$8B
	.dc.b	$E1,$C8,$C9,$0F,$8F,$A7,$B6,$35,$18,$A5,$73,$8F,$6C,$FF,$5F,$E9
	.dc.b	$FE,$C1,$EA,$C9,$FF,$69,$C6,$5F,$F6,$AC,$65,$FF,$68,$46,$5F,$F6
	.dc.b	$94,$67,$3F,$F7,$1C,$FD,$CE,$EF,$BB,$DF,$2F,$E6,$C6,$5F,$9F,$19
	.dc.b	$7D,$11,$97,$D5,$19,$AF,$C1,$4F,$F9,$0F,$E3,$DF,$33,$FD,$E6,$E1
	.dc.b	$F8,$08,$CB,$63,$2F,$FB,$A4,$E4,$C7,$FD,$E6,$6A,$BF,$97,$4F,$28
	.dc.b	$F9,$9E,$07,$E6,$3C,$19,$1F,$88,$8F,$CD,$98,$7E,$94,$61,$D1,$F8
	.dc.b	$E7,$55,$F9,$D9,$F6,$7C,$C7,$B7,$E6,$65,$FA,$E3,$2F,$B6,$32,$FE
	.dc.b	$0B,$04,$FF,$97,$3C,$49,$3F,$EC,$B7,$FC,$BF,$F0,$38,$8F,$F6,$3C
	.dc.b	$5C,$7F,$CC,$33,$FE,$64,$EA,$FF,$32,$F8,$D2,$FF,$DA,$91,$97,$FD
	.dc.b	$AD,$19,$7F,$DA,$31,$FE,$77,$F9,$97,$C8,$FF,$AE,$BE,$5F,$F8,$5F
	.dc.b	$42,$73,$7F,$EA,$B9,$CC,$62,$1E,$EA,$73,$F2,$9C,$E5,$7C,$F0,$C3
	.dc.b	$63,$A1,$AC,$3A,$61,$CE,$31,$7C,$27,$38,$78,$B0,$C1,$99,$EB,$EB
	.dc.b	$FE,$B2,$FA,$E2,$8F,$E1,$11,$E8,$BF,$7C,$63,$1F,$05,$6F,$AB,$99
	.dc.b	$D3,$C1,$98,$74,$BE,$CF,$33,$F9,$9F,$CC,$BD,$5F,$F1,$FC,$7F,$8E
	.dc.b	$04,$67,$DD,$C4,$70,$9D,$BF,$E3,$E9,$D3,$AF,$6F,$EB,$FC,$67,$08
	.dc.b	$FF,$5C,$EE,$8C,$4E,$A7,$84,$E5,$EF,$5F,$E1,$CE,$EE,$4D,$8C,$FA
	.dc.b	$5F,$EA,$D6,$CF,$F7,$8C,$DD,$FF,$E2,$0E,$0F,$8F,$18,$F0,$FE,$27
	.dc.b	$85,$9E,$3F,$9B,$77,$1F,$C1,$DE,$DF,$91,$C7,$F5,$67,$4F,$D7,$BB
	.dc.b	$5B,$C7,$EA,$FE,$4B,$F3,$72,$FB,$AB,$DA,$E2,$E7,$2E,$AE,$67,$7C
	.dc.b	$5A,$AD,$B6,$32,$74,$8C,$B4,$FC,$07,$83,$FE,$40,$EC,$FF,$69,$F8
	.dc.b	$76,$7F,$17,$37,$1F,$7C,$DD,$7F,$25,$37,$4F,$F7,$25,$FF,$62,$7F
	.dc.b	$DD,$0D,$AD,$7A,$92,$35,$E1,$DC,$7F,$A6,$DA,$B1,$97,$FB,$9A,$AD
	.dc.b	$FE,$F3,$C2,$7C,$CF,$8D,$7C,$CD,$8C,$36,$7F,$9B,$37,$2F,$E3,$4F
	.dc.b	$AF,$89,$F2,$17,$4C,$36,$FF,$BE,$CE,$9B,$3A,$69,$A9,$E1,$7F,$DA
	.dc.b	$2B,$83,$6B,$D1,$DF,$E9,$F9,$9B,$87,$F4,$B6,$D7,$F2,$F6,$EB,$FD
	.dc.b	$19,$B3,$F0,$DB,$3F,$C5,$B4,$47,$B7,$E2,$FD,$DF,$25,$6F,$D6,$DB
	.dc.b	$F3,$37,$0F,$F4,$06,$68,$C5,$7F,$2C,$8C,$3F,$12,$FA,$BF,$D8,$FE
	.dc.b	$1B,$EA,$77,$7D,$56,$C6,$4F,$ED,$11,$96,$99,$C7,$FA,$74,$65,$B1
	.dc.b	$F4,$BE,$33,$EC,$7C,$6F,$D9,$6C,$64,$FA,$F1,$96,$99,$9F,$A9,$19
	.dc.b	$6C,$7F,$7B,$F8,$7F,$C6,$1F,$8D,$F7,$AC,$7F,$7F,$6C,$FF,$63,$6E
	.dc.b	$BF,$DE,$9E,$13,$FE,$05,$F7,$AD,$7F,$69,$F9,$3F,$CC,$1F,$91,$F8
	.dc.b	$EB,$5F,$8E,$F8,$4C,$34,$FC,$8C,$7E,$5F,$E3,$A4,$7F,$84,$7C,$4B
	.dc.b	$87,$F9,$73,$E6,$3F,$95,$74,$7E,$E5,$FE,$59,$F2,$F7,$CB,$FE,$95
	.dc.b	$87,$FA,$2A,$72,$D8,$CD,$F3,$55,$C6,$F9,$36,$B8,$36,$B8,$3F,$B5
	.dc.b	$F8,$EF,$DB,$7C,$95,$AF,$F1,$43,$6B,$FC,$18,$DE,$1F,$F1,$1B,$FC
	.dc.b	$3F,$E1,$DF,$2F,$FB,$FF,$43,$0D,$C3,$FC,$E3,$3A,$7F,$C6,$93,$A6
	.dc.b	$9A,$CF,$A6,$7D,$5B,$19,$39,$C6,$5A,$67,$2F,$F3,$1D,$FC,$48,$C6
	.dc.b	$5F,$EE,$78,$C4,$FF,$A3,$87,$CB,$7C,$C9,$FF,$77,$D3,$E7,$5E,$2F
	.dc.b	$F2,$FB,$DE,$5F,$CC,$6B,$53,$9C,$3E,$AD,$7F,$73,$6B,$FE,$AA,$38
	.dc.b	$CF,$F6,$D8,$CC,$7F,$BF,$7B,$FE,$67,$87,$F3,$5B,$5F,$29,$B3,$EE
	.dc.b	$6E,$49,$FF,$5B,$F1,$7F,$77,$9F,$C6,$5F,$E3,$5B,$3F,$8B,$6D,$FE
	.dc.b	$2D,$B7,$CE,$17,$CE,$BF,$25,$7C,$DA,$7F,$32,$F2,$FF,$9C,$79,$7F
	.dc.b	$CE,$5B,$19,$3F,$9B,$C6,$5A,$7F,$BA,$2F,$E9,$F3,$7F,$5D,$F3,$DF
	.dc.b	$31,$FE,$41,$AD,$6A,$AD,$4E,$6F,$EE,$4E,$6B,$E7,$57,$AE,$7E,$A3
	.dc.b	$5F,$A9,$3B,$7F,$C5,$59,$FF,$87,$FA,$BF,$69,$5F,$B5,$6D,$FE,$CD
	.dc.b	$B5,$FB,$36,$EB,$F2,$6F,$95,$7E,$E6,$F9,$34,$F8,$1F,$2C,$F6,$7F
	.dc.b	$BE,$5B,$E7,$7F,$29,$5F,$89,$AD,$3F,$EF,$49,$FF,$03,$4E,$B2,$FF
	.dc.b	$7B,$D6,$9F,$F2,$0B,$5F,$F2,$33,$5E,$97,$F0,$DF,$F7,$FF,$E5,$7F
	.dc.b	$DE,$B0,$EC,$21,$F3,$2F,$CB,$DE,$0F,$E9,$6F,$85,$7F,$BE,$AB,$53
	.dc.b	$9F,$3F,$B4,$AB,$6F,$F9,$65,$AF,$E9,$DA,$FF,$9F,$BA,$9E,$37,$F1
	.dc.b	$1F,$2B,$F8,$93,$73,$FE,$1E,$DA,$FE,$06,$DD,$FF,$81,$78,$9F,$C6
	.dc.b	$4F,$E3,$5E,$45,$BF,$E2,$F1,$6F,$F8,$5C,$DA,$FD,$27,$A7,$FE,$7F
	.dc.b	$F8,$FF,$F1,$77,$C7,$FF,$8B,$DB,$8F,$F8,$BD,$B8,$7F,$8A,$9B,$7C
	.dc.b	$BC,$07,$C5,$7F,$8D,$FE,$2F,$FC,$70,$DC,$3F,$C7,$0D,$C3,$FC,$66
	.dc.b	$DB,$E4,$D9,$FF,$47,$B6,$7F,$CC,$6D,$CB,$FD,$2C,$DC,$1F,$17,$FE
	.dc.b	$47,$F8,$BF,$F2,$43,$70,$FF,$24,$37,$0F,$F2,$1B,$71,$F5,$7C,$4F
	.dc.b	$F9,$3F,$E2,$FF,$CA,$0D,$BF,$F2,$83,$6F,$FC,$96,$DA,$E3,$3F,$E5
	.dc.b	$5A,$E2,$D3,$F2,$35,$FE,$BE,$F9,$7F,$F3,$E5,$7D,$25,$EA,$FF,$98
	.dc.b	$5C,$D7,$FB,$39,$CB,$CB,$54,$6A,$FA,$57,$1F,$7F,$93,$6F,$85,$72
	.dc.b	$A6,$1E,$1B,$DD,$69,$B3,$A8,$F1,$46,$27,$F1,$3F,$9B,$FF,$4C,$9F
	.dc.b	$23,$FE,$74,$AE,$07,$FB,$71,$D3,$8C,$51,$BA,$E7,$E5,$7F,$A0,$63
	.dc.b	$17,$FE,$EE,$78,$CF,$83,$DE,$D5,$6D,$B5,$E2,$7F,$AF,$61,$FB,$FF
	.dc.b	$80,$FE,$01,$F3,$3F,$96,$8F,$EB,$5E,$4F,$F4,$58,$C4,$7F,$5D,$9D
	.dc.b	$F2,$7F,$B4,$EB,$4F,$FB,$4A,$33,$3C,$B0,$AF,$11,$7D,$15,$F2,$7F
	.dc.b	$DB,$D5,$FC,$4F,$D0,$7A,$7D,$65,$18,$76,$62,$35,$19,$87,$C2,$3D
	.dc.b	$BF,$BB,$1B,$3F,$90,$46,$76,$F0,$5F,$F7,$B8,$C4,$EC,$E7,$5C,$AF
	.dc.b	$B4,$EB,$2F,$8E,$BD,$AF,$F0,$24,$61,$BA,$F5,$BF,$1C,$F0,$6C,$78
	.dc.b	$67,$78,$79,$9A,$F8,$B2,$F9,$FC,$2F,$CE,$1F,$19,$FE,$13,$9F,$E3
	.dc.b	$70,$F8,$DC,$70,$FF,$3F,$BE,$11,$8C,$11,$CF,$0C,$36,$3F,$5D,$7C
	.dc.b	$DB,$5A,$8F,$0C,$6A,$B1,$32,$30,$7E,$C3,$8B,$AF,$33,$E1,$29,$88
	.dc.b	$93,$FB,$68,$CB,$C4,$87,$EA,$67,$93,$B3,$39,$FF,$0B,$AE,$31,$BA
	.dc.b	$EA,$FF,$B8,$8F,$7F,$6F,$E5,$1F,$F7,$67,$2E,$18,$60,$87,$13,$FD
	.dc.b	$E4,$72,$63,$83,$F8,$B7,$C5,$7C,$97,$F8,$02,$F8,$EF,$9E,$D8,$DE
	.dc.b	$78,$33,$32,$32,$74,$FA,$9E,$3A,$C8,$C1,$0E,$EE,$5F,$0C,$78,$0C
	.dc.b	$69,$FD,$8B,$0F,$B7,$3A,$39,$46,$6B,$7C,$1F,$8B,$A5,$FA,$BE,$13
	.dc.b	$D7,$F5,$33,$1F,$83,$19,$7E,$1C,$65,$F8,$71,$F9,$0F,$29,$EA,$D2
	.dc.b	$4F,$FA,$43,$D4,$AF,$FA,$12,$2E,$9F,$E8,$39,$E5,$F3,$DC,$FE,$7A
	.dc.b	$5F,$FA,$62,$32,$FF,$A6,$A3,$2F,$FA,$3A,$3F,$48,$BE,$33,$DA,$FF
	.dc.b	$9F,$CF,$F4,$03,$87,$06,$F4,$AF,$F8,$B6,$34,$F1,$78,$61,$FE,$59
	.dc.b	$AF,$9E,$DD,$FC,$33,$AD,$FF,$D1,$DF,$35,$F1,$30,$EA,$D7,$C2,$D7
	.dc.b	$CC,$D7,$8E,$9E,$CF,$9B,$7C,$25,$3D,$F5,$CF,$2F,$F3,$64,$B8,$78
	.dc.b	$6B,$31,$FF,$3E,$4E,$5B,$39,$7F,$D6,$51,$E9,$7F,$C9,$CE,$DC,$C6
	.dc.b	$EB,$BD,$79,$63,$E8,$FC,$DB,$E4,$FE,$5D,$A3,$0F,$CE,$6B,$F4,$9A
	.dc.b	$FD,$77,$37,$FE,$87,$F4,$1E,$77,$DE,$6D,$7F,$A5,$EB,$7B,$7F,$27
	.dc.b	$FC,$E5,$F2,$1F,$C9,$1F,$F6,$4D,$7F,$84,$5A,$FF,$7B,$76,$E6,$38
	.dc.b	$79,$52,$30,$7F,$A9,$70,$E3,$97,$97,$BF,$C8,$79,$6A,$56,$E3,$2F
	.dc.b	$18,$CB,$FE,$E1,$A4,$EA,$31,$3F,$AD,$F9,$EF,$D7,$7C,$FC,$9F,$F7
	.dc.b	$DC,$65,$FF,$7E,$C6,$5F,$F7,$8C,$7D,$49,$FA,$7A,$7C,$09,$F9,$D1
	.dc.b	$82,$3E,$DF,$95,$F9,$C7,$17,$FC,$71,$D6,$F8,$6F,$E0,$1F,$0D,$27
	.dc.b	$31,$FC,$0C,$C3,$FE,$27,$94,$E6,$36,$74,$77,$AF,$F9,$F2,$7B,$4F
	.dc.b	$83,$E7,$3F,$87,$75,$48,$C1,$0E,$CE,$5F,$04,$7F,$15,$E2,$FE,$2D
	.dc.b	$F3,$92,$7D,$71,$97,$DB,$19,$7D,$11,$F8,$FF,$83,$FC,$87,$CD,$49
	.dc.b	$FC,$14,$65,$FC,$24,$65,$FB,$91,$97,$EF,$46,$72,$FC,$E4,$FE,$57
	.dc.b	$9A,$57,$E5,$5F,$EF,$7F,$36,$F9,$BF,$CD,$7B,$7F,$9B,$F6,$A5,$FE
	.dc.b	$6E,$37,$94,$8F,$F3,$75,$A3,$71,$FE,$59,$8F,$D3,$BF,$E0,$A6,$F9
	.dc.b	$E6,$9F,$EE,$38,$E7,$D9,$E4,$E4,$91,$83,$BB,$87,$B3,$EA,$8F,$D9
	.dc.b	$79,$7F,$B3,$F6,$E5,$7E,$CE,$32,$FE,$D6,$32,$FE,$D6,$32,$FE,$DE
	.dc.b	$33,$BF,$FB,$0B,$BB,$FC,$05,$3E,$DF,$6F,$E4,$FC,$DC,$8F,$87,$1F
	.dc.b	$30,$C3,$F0,$E3,$0E,$8F,$92,$EA,$BF,$D9,$F9,$FF,$A7,$A7,$FD,$35
	.dc.b	$F7,$5F,$E0,$17,$F8,$51,$FC,$3C,$FF,$19,$B9,$9F,$F6,$BB,$79,$3F
	.dc.b	$8B,$4E,$CE,$7E,$77,$03,$0F,$43,$27,$DC,$F4,$7F,$0B,$EE,$C8,$C1
	.dc.b	$0F,$BC,$E5,$FC,$04,$7E,$30,$C6,$1E,$C9,$C4,$FB,$3C,$BF,$4F,$A3
	.dc.b	$23,$04,$3F,$D4,$59,$7F,$89,$6F,$DB,$3F,$02,$A5,$7C,$68,$E4,$F0
	.dc.b	$8F,$CE,$F8,$1F,$9E,$F1,$24,$FE,$62,$32,$FE,$6A,$32,$FE,$66,$3F
	.dc.b	$C6,$3F,$35,$23,$FC,$6E,$FF,$8D,$3E,$AF,$11,$FE,$38,$8B,$8F,$F8
	.dc.b	$9E,$7F,$C7,$3F,$35,$FE,$3A,$F8,$89,$7F,$E4,$08,$CB,$FE,$42,$8C
	.dc.b	$BF,$E4,$28,$FF,$31,$BF,$EA,$FF,$97,$FF,$6F,$73,$D8,$FF,$73,$37
	.dc.b	$AB,$FD,$BF,$17,$D8,$FF,$54,$D7,$08,$FF,$5F,$35,$7C,$0A,$3F,$DD
	.dc.b	$51,$C4,$E5,$71,$E2,$7F,$BD,$4F,$07,$B0,$F9,$AA,$48,$DE,$29,$7C
	.dc.b	$DE,$67,$C3,$87,$F9,$1E,$F1,$8F,$8E,$7F,$D2,$37,$C1,$F0,$8F,$9D
	.dc.b	$FE,$8D,$F0,$7F,$E2,$C6,$C7,$F8,$C1,$AF,$F8,$D5,$AF,$FA,$EE,$DF
	.dc.b	$35,$FE,$34,$F8,$4F,$53,$63,$33,$FE,$3A,$9C,$BD,$7C,$CF,$E0,$68
	.dc.b	$97,$FE,$45,$DB,$C9,$3A,$9C,$B7,$67,$BE,$EB,$77,$7C,$88,$C3,$7C
	.dc.b	$1C,$E1,$F2,$0F,$F5,$AC,$E9,$C5,$F4,$BD,$F0,$7C,$B8,$C6,$1C,$CE
	.dc.b	$D3,$AE,$1F,$F6,$17,$CE,$FF,$71,$9F,$B2,$FF,$6C,$6F,$A3,$BC,$7F
	.dc.b	$D8,$B3,$97,$C2,$F1,$6F,$A3,$BC,$53,$9C,$BC,$8F,$16,$DE,$E3,$18
	.dc.b	$F9,$23,$17,$CF,$3F,$F0,$DE,$5F,$47,$AB,$E2,$E5,$F3,$3C,$1F,$8F
	.dc.b	$18,$AD,$E8,$8C,$4E,$AF,$FD,$C5,$5A,$9C,$C7,$53,$3D,$DB,$BF,$A3
	.dc.b	$77,$0F,$B7,$EB,$5C,$7F,$C8,$B8,$71,$BF,$AF,$DF,$9A,$93,$9B,$E2
	.dc.b	$DB,$56,$AF,$D1,$3F,$EF,$7F,$80,$77,$78,$57,$81,$63,$0D,$9E,$D3
	.dc.b	$A6,$CE,$9F,$17,$23,$E5,$FD,$8F,$E5,$A3,$31,$8C,$B2,$D8,$C9,$CA
	.dc.b	$32,$D3,$38,$FA,$86,$6B,$ED,$19,$86,$3F,$C2,$2B,$D3,$E0,$7B,$FB
	.dc.b	$FE,$03,$63,$27,$BA,$32,$D3,$37,$D6,$BA,$36,$BD,$FF,$1D,$F0,$3E
	.dc.b	$62,$DF,$C4,$6D,$79,$72,$46,$38,$FF,$23,$18,$AF,$97,$7C,$DA,$73
	.dc.b	$3E,$9F,$C3,$FD,$9E,$FF,$B4,$D8,$C9,$F6,$23,$2D,$33,$8F,$AE,$32
	.dc.b	$D8,$FB,$7F,$1B,$F6,$BE,$3F,$ED,$B6,$32,$7E,$02,$32,$D3,$33,$EF
	.dc.b	$8C,$B6,$3F,$17,$F1,$1F,$91,$3F,$25,$F3,$B6,$3F,$1E,$D9,$FA,$0D
	.dc.b	$D3,$F6,$FB,$A3,$8D,$7E,$5B,$E7,$6D,$FD,$7F,$93,$FC,$F1,$F9,$4F
	.dc.b	$92,$B5,$F9,$8F,$88,$C3,$4F,$CD,$C7,$E6,$BE,$2E,$47,$E8,$BC,$1B
	.dc.b	$87,$E4,$FE,$63,$FA,$1F,$4F,$EF,$5F,$E8,$BF,$39,$7C,$BF,$CA,$F0
	.dc.b	$7F,$32,$9C,$E3,$FB,$D7,$8B,$65,$5C,$6F,$93,$67,$4D,$C3,$C5,$5C
	.dc.b	$27,$F6,$FF,$23,$FB,$DF,$8C,$FD,$F3,$63,$27,$EF,$23,$2D,$3F,$D5
	.dc.b	$7F,$3B,$FC,$17,$CF,$FE,$2D,$F0,$A8,$C4,$E8,$FE,$0B,$3A,$8C,$E3
	.dc.b	$FE,$AD,$CF,$8E,$7C,$9B,$7F,$EB,$86,$CE,$E3,$FD,$77,$3B,$69,$B2
	.dc.b	$C7,$F0,$31,$F3,$7F,$C0,$9F,$ED,$87,$F6,$D7,$0E,$1A,$E6,$BF,$B2
	.dc.b	$BC,$0E,$6F,$F3,$76,$CF,$F3,$3F,$3D,$F3,$8E,$5D,$61,$F7,$AB,$FB
	.dc.b	$E7,$86,$D7,$F7,$98,$FF,$B3,$FC,$4F,$97,$E1,$7C,$C6,$DF,$FB,$69
	.dc.b	$B3,$B8,$FC,$24,$FE,$2E,$7F,$18,$D8,$C9,$F8,$98,$CB,$4C,$CF,$E1
	.dc.b	$A3,$2D,$8F,$E7,$3E,$67,$F3,$8F,$33,$F9,$CB,$63,$27,$F3,$E8,$CB
	.dc.b	$4F,$EA,$D3,$F4,$BA,$3E,$5B,$F3,$DA,$FD,$36,$BB,$8F,$A0,$E5,$E7
	.dc.b	$9A,$7F,$B4,$7C,$CB,$E8,$BF,$80,$56,$A7,$F5,$DD,$9F,$AF,$6C,$7E
	.dc.b	$AD,$B8,$7E,$99,$BD,$5F,$B3,$E9,$ED,$5F,$B1,$BE,$4D,$39,$1E,$EF
	.dc.b	$9C,$7C,$47,$A9,$B1,$EA,$C5,$18,$7D,$AD,$7F,$1C,$E7,$0F,$C8,$B6
	.dc.b	$7F,$C8,$99,$FB,$FA,$BF,$25,$39,$BF,$F2,$47,$CD,$FF,$7A,$BF,$03
	.dc.b	$63,$E5,$B5,$FC,$E3,$5F,$D2,$BB,$C3,$F4,$2D,$9F,$F2,$A7,$E4,$6F
	.dc.b	$CC,$1F,$21,$FA,$3A,$D6,$1F,$DC,$B9,$9E,$3F,$F0,$6F,$2F,$F8,$3B
	.dc.b	$72,$FF,$0B,$4E,$9B,$3A,$69,$AF,$13,$F8,$79,$FC,$43,$C9,$B1,$92
	.dc.b	$61,$FD,$FE,$D3,$A8,$FF,$01,$77,$BE,$77,$FC,$59,$F2,$5F,$E2,$D6
	.dc.b	$E1,$FE,$27,$6D,$70,$6F,$07,$F9,$07,$E6,$BF,$C6,$9F,$3B,$FE,$35
	.dc.b	$6D,$FF,$8B,$DB,$7E,$06,$D7,$19,$FF,$1A,$57,$16,$9C,$4B,$5C,$5B
	.dc.b	$3F,$E5,$1F,$8D,$FF,$22,$7C,$87,$F9,$15,$B3,$FE,$3F,$6D,$7C,$06
	.dc.b	$D7,$F9,$67,$E2,$FF,$C9,$9F,$1F,$FE,$4D,$6D,$FF,$91,$DB,$7F,$21
	.dc.b	$B5,$C6,$7F,$C9,$95,$C5,$A7,$F9,$47,$E1,$3F,$D1,$53,$FE,$6A,$7F
	.dc.b	$C6,$91,$9F,$80,$FF,$2F,$46,$30,$FF,$1B,$3C,$27,$F9,$3B,$9A,$E5
	.dc.b	$5C,$3D,$6E,$2D,$AE,$87,$86,$32,$D3,$37,$DA,$3F,$CD,$DF,$3B,$FE
	.dc.b	$64,$9F,$F3,$43,$FE,$77,$7F,$9B,$B9,$9F,$F2,$53,$97,$FC,$9E,$E6
	.dc.b	$FF,$D0,$EE,$7C,$83,$9A,$E7,$EF,$F8,$3D,$DF,$0B,$63,$C2,$FF,$4D
	.dc.b	$C3,$FD,$85,$F0,$3F,$EC,$3F,$9D,$71,$FF,$53,$F3,$FF,$D7,$31,$8A
	.dc.b	$FE,$AD,$5B,$AD,$61,$67,$22,$B8,$3F,$EC,$FB,$E7,$5F,$CC,$2F,$84
	.dc.b	$E6,$EB,$E4,$1E,$93,$D3,$E1,$18,$6C,$66,$3C,$2D,$5F,$0A,$F7,$3F
	.dc.b	$D5,$51,$C4,$FF,$71,$46,$78,$3F,$DC,$95,$C6,$31,$3B,$39,$D7,$29
	.dc.b	$D3,$D3,$4F,$7D,$3D,$BF,$F0,$1C,$61,$BA,$7F,$81,$A7,$AC,$F0,$AE
	.dc.b	$53,$9A,$DE,$1E,$66,$BE,$2C,$7E,$87,$0B,$E7,$5F,$13,$FE,$12,$9F
	.dc.b	$F6,$6C,$3E,$2B,$1B,$FF,$81,$A3,$18,$7F,$AE,$DE,$2D,$BD,$BF,$EA
	.dc.b	$08,$DB,$FE,$9F,$6D,$F6,$9D,$E1,$E3,$6C,$EA,$FC,$18,$62,$77,$CE
	.dc.b	$E3,$BF,$0B,$33,$FD,$F4,$68,$C3,$FB,$A8,$C3,$A3,$F8,$23,$AC,$FF
	.dc.b	$82,$4E,$A7,$81,$A3,$81,$FC,$E0,$F4,$F8,$0E,$5A,$62,$21,$FE,$DF
	.dc.b	$72,$FF,$B8,$4C,$9F,$D0,$9C,$BD,$4F,$F7,$47,$C4,$FC,$37,$FD,$96
	.dc.b	$BE,$33,$E4,$B6,$37,$A6,$0C,$CC,$8C,$9D,$3E,$67,$8E,$B2,$30,$43
	.dc.b	$FD,$AD,$7F,$79,$B1,$E0,$99,$1E,$68,$E4,$F0,$8C,$BC,$0D,$1D,$73
	.dc.b	$6D,$FE,$52,$BF,$F7,$1C,$FC,$0F,$99,$F6,$9E,$EF,$83,$91,$92,$1E
	.dc.b	$67,$F1,$7E,$3B,$D3,$A4,$9F,$EF,$BE,$AD,$7F,$BA,$45,$CB,$C7,$1F
	.dc.b	$E0,$79,$E7,$F3,$9D,$3E,$76,$4F,$F8,$1A,$32,$FF,$82,$23,$2F,$F7
	.dc.b	$F8,$FD,$1A,$F8,$D6,$D8,$6C,$7F,$95,$1A,$FF,$A1,$E5,$A6,$EB,$FC
	.dc.b	$71,$3C,$DB,$5C,$23,$FD,$08,$D7,$FD,$1B,$F3,$BF,$0D,$FF,$30,$B5
	.dc.b	$FF,$4A,$37,$2F,$F4,$6F,$CE,$FF,$2A,$9D,$E9,$E0,$AE,$11,$FE,$91
	.dc.b	$6B,$DA,$3C,$B3,$E8,$8F,$F3,$7B,$EC,$AE,$B8,$7C,$48,$D4,$FF,$A9
	.dc.b	$5B,$3A,$8E,$33,$BC,$64,$60,$87,$17,$E7,$FC,$E3,$E5,$E9,$27,$FD
	.dc.b	$03,$E7,$BC,$AC,$47,$F9,$F2,$B8,$3F,$E6,$56,$B9,$AF,$F3,$47,$73
	.dc.b	$E5,$DF,$23,$FC,$69,$79,$34,$C4,$4B,$0D,$7F,$1F,$7C,$EB,$84,$EA
	.dc.b	$73,$7D,$A3,$D3,$3F,$8C,$CF,$F9,$14,$7B,$3B,$B1,$32,$30,$43,$FD
	.dc.b	$A5,$3E,$67,$99,$9B,$FF,$66,$C7,$48,$DC,$FE,$B3,$E7,$7F,$5B,$EC
	.dc.b	$49,$FF,$33,$C6,$5F,$F3,$4C,$65,$FF,$30,$47,$D4,$99,$1F,$EB,$2D
	.dc.b	$7C,$4F,$93,$CD,$CB,$F4,$E7,$F3,$F3,$A8,$FF,$70,$35,$FF,$6C,$F4
	.dc.b	$BE,$0F,$F8,$0F,$C1,$C9,$FF,$3D,$C7,$F0,$43,$0F,$F9,$CE,$30,$E8
	.dc.b	$FF,$40,$BA,$D3,$EA,$D7,$59,$FF,$42,$FC,$EF,$F0,$EE,$99,$18,$21
	.dc.b	$FE,$7F,$72,$FF,$A0,$63,$F8,$AF,$7F,$F1,$6F,$9D,$93,$FE,$90,$8C
	.dc.b	$BF,$E9,$28,$CB,$FE,$89,$8F,$C0,$7C,$17,$E0,$7E,$72,$4F,$E7,$23
	.dc.b	$2F,$E7,$A3,$2F,$E0,$23,$2F,$E0,$A3,$39,$7E,$02,$7F,$35,$EB,$CA
	.dc.b	$FC,$95,$FE,$63,$9B,$F4,$13,$FD,$E3,$E6,$9F,$3B,$FA,$1F,$67,$F4
	.dc.b	$5D,$93,$1E,$23,$0F,$D4,$4F,$1D,$24,$EA,$58,$27,$94,$FE,$9F,$2F
	.dc.b	$14,$FE,$D3,$4F,$D4,$B6,$33,$D9,$E5,$D7,$CC,$F9,$FE,$76,$61,$DE
	.dc.b	$F8,$3E,$55,$1A,$8F,$01,$97,$9B,$E7,$EC,$FF,$00,$37,$DA,$FE,$04
	.dc.b	$7F,$03,$F3,$24,$64,$87,$4D,$7F,$85,$46,$A3,$14,$AE,$71,$DE,$7E
	.dc.b	$87,$57,$D1,$F9,$D9,$3E,$B8,$CB,$ED,$8C,$BD,$F2,$4E,$A3,$13,$AD
	.dc.b	$BE,$5C,$FF,$8D,$3E,$77,$FC,$19,$5E,$43,$F8,$F1,$9F,$36,$3F,$0C
	.dc.b	$C9,$F6,$E3,$8D,$7F,$1E,$6F,$07,$E2,$F4,$FE,$47,$5C,$4E,$17,$F8
	.dc.b	$DF,$43,$F1,$DE,$E4,$9F,$B1,$19,$7E,$D4,$65,$F7,$46,$5F,$81,$19
	.dc.b	$C7,$F0,$F3,$ED,$F3,$FD,$DE,$84,$BF,$44,$65,$F5,$46,$5F,$B1,$9F
	.dc.b	$F8,$8E,$FD,$95,$97,$DD,$5C,$6B,$E8,$78,$3F,$A7,$F1,$64,$7E,$46
	.dc.b	$3F,$54,$61,$FA,$D1,$87,$FC,$83,$F3,$5F,$E3,$E2,$3F,$E4,$4F,$9D
	.dc.b	$AF,$F8,$DA,$2E,$1F,$D3,$3D,$25,$7F,$91,$BE,$67,$FC,$8F,$F3,$73
	.dc.b	$0F,$F2,$6C,$65,$FF,$27,$C6,$5F,$F2,$0C,$7F,$9C,$1F,$AF,$EF,$BE
	.dc.b	$5F,$FD,$C7,$D1,$86,$CF,$FB,$19,$FF,$73,$B9,$3F,$81,$7B,$1F,$EC
	.dc.b	$48,$DC,$2B,$A3,$AA,$E1,$D6,$BE,$13,$C8,$CB,$FB,$27,$95,$F4,$BE
	.dc.b	$75,$C6,$79,$E6,$F9,$AA,$48,$F4,$1F,$E6,$EF,$9B,$C3,$D0,$FF,$38
	.dc.b	$C6,$2F,$AD,$2B,$DD,$4F,$9D,$FE,$A9,$F0,$5F,$9E,$8F,$CF,$C6,$1C
	.dc.b	$1B,$F3,$7F,$C6,$0E,$E3,$4F,$1F,$82,$F4,$7C,$8F,$08,$FF,$6B,$E6
	.dc.b	$8C,$4E,$A7,$AB,$5E,$9E,$6F,$C7,$7C,$6D,$7F,$C2,$B7,$FE,$57,$E0
	.dc.b	$EC,$D8,$F5,$64,$FA,$8F,$F6,$37,$81,$FE,$1C,$3F,$77,$19,$EA,$FF
	.dc.b	$35,$C6,$27,$75,$FE,$6E,$3F,$DE,$EE,$23,$53,$FE,$72,$E0,$FC,$1D
	.dc.b	$F3,$9F,$AB,$C2,$F9,$E7,$CD,$6F,$93,$FD,$E7,$97,$6C,$7F,$17,$9F
	.dc.b	$F8,$FF,$CD,$D3,$FE,$AA,$F8,$5A,$E7,$3E,$BC,$DF,$F0,$05,$FC,$1C
	.dc.b	$64,$F4,$15,$FB,$9E,$77,$F6,$27,$E1,$47,$F1,$C8,$C3,$C1,$B8,$79
	.dc.b	$1B,$97,$F6,$DF,$C9,$1F,$01,$E4,$9C,$DB,$E5,$5F,$9D,$74,$F0,$3F
	.dc.b	$D2,$F1,$8D,$BF,$D3,$31,$FE,$4C,$E1,$7C,$2E,$36,$FF,$28,$DE,$DF
	.dc.b	$EE,$AD,$C7,$FA,$FB,$6F,$95,$71,$9F,$16,$3E,$1B,$FF,$02,$60,$F9
	.dc.b	$9D,$43,$F3,$B7,$B5,$C7,$9B,$EC,$56,$A7,$35,$FE,$01,$BE,$4F,$F8
	.dc.b	$20,$E5,$7C,$63,$38,$FF,$74,$9C,$9F,$E0,$E5,$F9,$6B,$E2,$BC,$1E
	.dc.b	$93,$E4,$3D,$55,$A9,$CC,$62,$B8,$5F,$EB,$5B,$5F,$B6,$6B,$FB,$B6
	.dc.b	$BF,$D6,$BF,$01,$76,$7C,$7F,$09,$F4,$AF,$57,$C3,$5A,$C7,$FC,$4B
	.dc.b	$5A,$9C,$DF,$1B,$E1,$5E,$DB,$E1,$87,$F8,$53,$0E,$13,$9B,$E2,$F4
	.dc.b	$E4,$FF,$51,$39,$3E,$37,$E5,$4F,$8C,$F3,$77,$FB,$AB,$F6,$93,$9A
	.dc.b	$F2,$CE,$5F,$EE,$F1,$9C,$7C,$53,$97,$FD,$3B,$19,$AF,$F7,$5E,$2A
	.dc.b	$D5,$2B,$57,$D3,$0F,$91,$7F,$CD,$63,$CF,$1F,$62,$7F,$DF,$91,$E4
	.dc.b	$FA,$0C,$7C,$5F,$A9,$B1,$93,$DF,$19,$69,$30,$F9,$58,$3E,$93,$8C
	.dc.b	$7F,$90,$07,$E3,$0F,$4F,$EE,$7D,$47,$AA,$31,$87,$DA,$8C,$4E,$B9
	.dc.b	$7F,$1F,$5A,$BF,$B1,$5A,$BF,$CC,$4E,$77,$FE,$49,$F9,$E3,$EA,$38
	.dc.b	$E5,$CB,$D7,$F0,$EF,$71,$F0,$37,$3E,$AD,$C7,$E8,$E6,$8C,$57,$F8
	.dc.b	$C3,$3F,$F6,$BB,$FA,$9C,$DF,$45,$AC,$7F,$A6,$7E,$36,$F8,$4F,$CE
	.dc.b	$36,$FD,$3F,$0D,$89,$F7,$B6,$32,$7E,$7A,$78,$57,$CD,$F1,$AE,$BF
	.dc.b	$D0,$F3,$61,$FB,$69,$CE,$1E,$78,$FD,$DF,$A3,$23,$A7,$BF,$71,$FB
	.dc.b	$18,$7F,$BF,$67,$27,$CE,$F2,$7D,$35,$D4,$91,$FE,$83,$AA,$7E,$07
	.dc.b	$E9,$7F,$94,$C7,$F1,$3E,$1F,$8F,$39,$F2,$7F,$8F,$C6,$35,$FE,$B9
	.dc.b	$7C,$2B,$F8,$B4,$E6,$FF,$CB,$F3,$9A,$FE,$CB,$39,$A0,$9E,$4F,$F8
	.dc.b	$25,$AF,$2B,$FE,$A6,$F0,$8C,$BF,$EA,$FF,$CE,$3F,$B7,$DB,$0F,$F5
	.dc.b	$D4,$3E,$0B,$F9,$A3,$A8,$CC,$63,$D4,$CB,$63,$FD,$6D,$3C,$5B,$F2
	.dc.b	$7C,$BE,$07,$9B,$63,$27,$08,$CB,$4F,$F6,$2F,$D5,$7F,$86,$7E,$A3
	.dc.b	$FA,$73,$FE,$1B,$F3,$AF,$80,$9B,$18,$6D,$72,$9D,$36,$74,$D3,$53
	.dc.b	$C3,$5F,$EB,$FE,$0F,$C1,$7F,$D8,$5C,$DF,$D7,$FD,$14,$E7,$D0,$FF
	.dc.b	$6A,$CE,$63,$FC,$37,$5F,$E9,$AA,$DF,$07,$D4,$6E,$3F,$E3,$E6,$CF
	.dc.b	$F9,$0F,$06,$FF,$E7,$6A,$F9,$BF,$33,$F5,$3E,$0F,$EA,$B6,$32,$7D
	.dc.b	$28,$CB,$FD,$87,$93,$F9,$5E,$0E,$A7,$D4,$FF,$9D,$EB,$5E,$47,$F7
	.dc.b	$B9,$CF,$AF,$FD,$93,$0E,$31,$8A,$D9,$FE,$FA,$71,$87,$E9,$6F,$F2
	.dc.b	$2F,$E5,$DB,$5F,$D1,$63,$F3,$11,$F9,$6D,$1F,$49,$FE,$26,$E9,$FC
	.dc.b	$A5,$27,$38,$7F,$81,$EB,$58,$FE,$92,$72,$FE,$9A,$F8,$C7,$4B,$E1
	.dc.b	$7C,$5B,$5F,$17,$E0,$7E,$06,$9F,$C7,$A3,$07,$B5,$C4,$FF,$01,$9F
	.dc.b	$59,$27,$FC,$E1,$9A,$3E,$0F,$D4,$7F,$16,$A6,$5F,$E0,$0E,$CF,$F0
	.dc.b	$7C,$E6,$31,$87,$F8,$5E,$31,$E8,$7F,$03,$8F,$F7,$1C,$62,$B8,$F8
	.dc.b	$9D,$89,$3F,$E8,$19,$FE,$57,$87,$FA,$35,$7C,$87,$DD,$FC,$CF,$FC
	.dc.b	$CB,$5D,$E3,$1F,$01,$E4,$7D,$24,$8E,$8F,$DC,$F2,$3F,$C6,$B8,$7B
	.dc.b	$27,$F8,$AF,$03,$0F,$F2,$BC,$7F,$8C,$6B,$7A,$FD,$CA,$D7,$B9,$FE
	.dc.b	$30,$9E,$78,$7F,$A8,$DA,$FF,$20,$3F,$A8,$FB,$5E,$13,$F5,$5D,$8E
	.dc.b	$9F,$F1,$DF,$D1,$FE,$C5,$B9,$FF,$96,$A7,$4F,$EB,$DB,$7F,$EB,$1E
	.dc.b	$CF,$DD,$D7,$43,$F2,$9D,$8F,$89,$FE,$46,$7F,$24,$F8,$AB,$8F,$F2
	.dc.b	$08,$B8,$7C,$33,$F9,$74,$E3,$7F,$F3,$9E,$4F,$8A,$FF,$30,$7C,$1F
	.dc.b	$F8,$97,$4F,$E1,$71,$8F,$57,$F8,$53,$EE,$26,$7E,$97,$E0,$FC,$4F
	.dc.b	$2F,$5F,$F8,$83,$67,$E0,$3A,$7E,$03,$74,$F8,$19,$7F,$B5,$67,$67
	.dc.b	$F9,$BF,$E9,$B9,$B8,$9F,$E4,$75,$FE,$6F,$9C,$C7,$BF,$DB,$7E,$06
	.dc.b	$F9,$5F,$6F,$B7,$FE,$85,$9C,$C6,$31,$FD,$B1,$FD,$72,$32,$FF,$B8
	.dc.b	$1B,$7F,$EC,$CB,$E9,$39,$6C,$7F,$B4,$3E,$B9,$F2,$7F,$32,$FC,$11
	.dc.b	$8B,$EF,$5F,$C2,$F8,$7F,$CC,$FC,$5F,$C3,$AD,$7B,$AF,$B5,$E3,$2F
	.dc.b	$78,$79,$EB,$58,$7F,$0E,$9C,$DF,$F9,$FA,$79,$3F,$EE,$06,$BC,$71
	.dc.b	$FE,$EB,$8F,$81,$FE,$89,$F2,$4F,$4F,$FD,$49,$F4,$DF,$E7,$6B,$FF
	.dc.b	$58,$CE,$73,$F5,$E8,$AD,$69,$FE,$C0,$AD,$5F,$AA,$73,$7F,$EB,$E9
	.dc.b	$CD,$7F,$26,$9C,$D7,$FB,$2E,$73,$5B,$C3,$C7,$F2,$BF,$D8,$67,$35
	.dc.b	$B9,$C1,$AA,$FF,$64,$CE,$A7,$FC,$33,$EC,$FF,$AF,$FF,$3D,$EF,$18
	.dc.b	$AF,$5C,$62,$3F,$C6,$C7,$D2,$71,$3C,$7E,$5F,$FD,$6F,$3A,$CF,$FD
	.dc.b	$D5,$7F,$13,$37,$D2,$EB,$0D,$DB,$E3,$FF,$C5,$7F,$4D,$A8,$FE,$1D
	.dc.b	$E3,$7F,$50,$8C,$47,$F1,$19,$DD,$FF,$13,$8F,$ED,$58,$E1,$C4,$D7
	.dc.b	$F9,$1C,$57,$FC,$73,$16,$FB,$57,$39,$F2,$1F,$EC,$6F,$A7,$C5,$7E
	.dc.b	$95,$B1,$93,$B9,$47,$CA,$FF,$AA,$F4,$E1,$1F,$E4,$88,$C3,$B6,$F2
	.dc.b	$2B,$F9,$74,$61,$BF,$13,$FD,$62,$9A,$7C,$4A,$FF,$75,$47,$BD,$AE
	.dc.b	$B2,$E3,$8E,$6F,$95,$F1,$F0,$3F,$9E,$36,$F5,$5C,$FA,$FF,$A0,$3F
	.dc.b	$E0,$28,$F6,$C6,$36,$F3,$B6,$3F,$C0,$DC,$0E,$0C,$47,$03,$F4,$FF
	.dc.b	$9F,$3D,$FE,$1D,$08,$C4,$FF,$06,$38,$1F,$EB,$FD,$FF,$D6,$0F,$2A
	.dc.b	$E3,$3C,$23,$58,$F8,$FC,$37,$CC,$FF,$23,$9F,$E2,$31,$8E,$FD,$4E
	.dc.b	$67,$F8,$7F,$D4,$3C,$6C,$D4,$9E,$35,$BE,$0F,$F0,$ED,$EE,$73,$5E
	.dc.b	$C8,$FE,$B1,$1F,$C1,$9A,$FF,$55,$C3,$C4,$F0,$35,$E3,$F1,$3C,$0F
	.dc.b	$F3,$DF,$25,$ED,$FE,$5E,$1D,$3F,$DC,$63,$0D,$8F,$E7,$53,$C1,$B3
	.dc.b	$F4,$3E,$49,$F5,$F4,$FB,$57,$C6,$BC,$15,$97,$F8,$FD,$70,$7F,$9D
	.dc.b	$CE,$F4,$FE,$75,$E3,$BB,$FC,$6D,$AF,$E9,$8D,$7F,$75,$48,$C4,$7F
	.dc.b	$83,$E3,$13,$F8,$6F,$1B,$FD,$CD,$3F,$CC,$1A,$FF,$61,$AE,$97,$FD
	.dc.b	$FB,$8B,$F0,$5E,$4F,$E2,$2B,$FD,$17,$1F,$06,$32,$C8,$FF,$0B,$39
	.dc.b	$69,$FE,$17,$7E,$86,$FF,$E9,$06,$CF,$59,$FF,$47,$3D,$EB,$FA,$A7
	.dc.b	$BB,$E9,$3B,$4F,$FA,$71,$E2,$47,$FC,$48,$D7,$FB,$0E,$3F,$F0,$01
	.dc.b	$F8,$1E,$87,$B2,$7F,$53,$E7,$3C,$CA,$4A,$FF,$61,$F9,$D5,$FE,$D7
	.dc.b	$3B,$7F,$D8,$F3,$CA,$7F,$51,$F9,$FF,$DD,$9F,$F1,$A3,$5F,$F5,$53
	.dc.b	$5F,$F7,$1E,$A8,$C4,$23,$1C,$7F,$BE,$F9,$4F,$E7,$7A,$FF,$A1,$1B
	.dc.b	$E0,$F2,$C3,$C1,$3F,$C3,$FF,$40,$7D,$0F,$E7,$F2,$46,$1F,$D1,$35
	.dc.b	$FE,$74,$D7,$FA,$33,$99,$FE,$35,$FA,$0B,$E8,$3F,$8A,$B6,$BF,$88
	.dc.b	$56,$F1,$FF,$30,$7C,$F3,$87,$C4,$DA,$FD,$93,$5F,$E9,$4D,$7F,$AB
	.dc.b	$39,$C7,$F9,$57,$E8,$4F,$90,$FE,$4A,$DA,$FE,$41,$5B,$C3,$FC,$E1
	.dc.b	$F5,$9F,$ED,$E9,$FE,$00,$D7,$FA,$C3,$5F,$EC,$8D,$78,$F3,$FF,$32
	.dc.b	$FA,$CF,$C5,$47,$F9,$94,$38,$CB,$F9,$F5,$75,$6D,$71,$E7,$97,$FC
	.dc.b	$D6,$B5,$87,$9B,$1F,$AD,$06,$92,$7D,$2F,$8F,$BA,$47,$38,$FE,$87
	.dc.b	$F9,$07,$F9,$D6,$3F,$C5,$AF,$F2,$8C,$BF,$D9,$6D,$C3,$FD,$A0,$EE
	.dc.b	$7F,$C8,$ED,$9D,$F8,$3D,$0E,$BD,$1F,$D5,$3E,$A3,$F8,$7D,$FF,$A7
	.dc.b	$5A,$83,$86,$F7,$FE,$4E,$3F,$87,$E3,$FE,$A5,$FA,$BF,$F5,$ED,$7F
	.dc.b	$59,$6B,$F3,$E7,$6F,$56,$A8,$DE,$3F,$DB,$DB,$5E,$5B,$FD,$FF,$A1
	.dc.b	$FC,$03,$E9,$66,$1F,$E5,$2E,$F4,$EA,$31,$3A,$7F,$80,$C7,$EF,$BE
	.dc.b	$3F,$F8,$DF,$61,$A4,$9F,$F1,$27,$7F,$EF,$63,$F7,$5E,$7C,$9F,$ED
	.dc.b	$75,$C5,$F8,$1F,$35,$FE,$00,$D5,$F3,$74,$C4,$49,$F8,$D1,$97,$89
	.dc.b	$0F,$CB,$4F,$27,$66,$77,$6F,$F5,$FD,$16,$FF,$CC,$63,$0B,$E0,$3F
	.dc.b	$93,$9F,$CA,$3C,$79,$87,$C9,$8C,$BF,$2E,$3F,$96,$F7,$FF,$2E,$F9
	.dc.b	$B9,$3F,$3E,$32,$FD,$18,$CB,$FC,$BA,$3F,$1D,$AF,$E1,$C9,$1F,$DE
	.dc.b	$79,$51,$88,$F7,$FB,$DF,$24,$E2,$FF,$88,$E7,$6F,$E3,$63,$4E,$67
	.dc.b	$FC,$65,$F2,$9F,$E4,$BE,$8B,$ED,$7E,$8F,$6F,$CE,$4E,$6F,$F1,$1D
	.dc.b	$FF,$E3,$6F,$C8,$3E,$44,$25,$F9,$1F,$E4,$A3,$0D,$53,$C3,$8B,$E8
	.dc.b	$E3,$FD,$50,$93,$FE,$C0,$BE,$7F,$13,$E4,$F7,$3C,$1C,$32,$B5,$7F
	.dc.b	$E5,$AE,$5F,$F2,$86,$5F,$DF,$9F,$BF,$3E,$23,$DF,$9F,$FA,$0E,$37
	.dc.b	$78,$98,$78,$67,$5A,$FE,$F6,$3D,$33,$FE,$56,$E2,$7C,$6D,$24,$FA
	.dc.b	$7F,$35,$7C,$86,$21,$97,$93,$77,$E8,$6F,$94,$FA,$6D,$AF,$F2,$E6
	.dc.b	$5D,$9B,$96,$AA,$46,$0D,$FC,$C7,$94,$8F,$8F,$DB,$AF,$8B,$6F,$BB
	.dc.b	$8B,$E5,$BC,$26,$B4,$FF,$3D,$66,$C7,$38,$7D,$5B,$E0,$F8,$0F,$F3
	.dc.b	$97,$E1,$4F,$94,$FF,$3D,$B6,$33,$1F,$90,$F8,$8F,$F4,$36,$3F,$E8
	.dc.b	$7A,$D5,$79,$4F,$C6,$78,$3F,$0D,$C7,$07,$F2,$1D,$FA,$E3,$27,$F1
	.dc.b	$15,$BC,$3E,$73,$76,$FB,$58,$F6,$9F,$07,$93,$E6,$25,$FE,$47,$F1
	.dc.b	$1D,$5F,$09,$FD,$3E,$1F,$02,$BC,$6D,$BE,$71,$F0,$A7,$7C,$12,$30
	.dc.b	$7C,$6F,$A6,$F8,$A4,$7F,$41,$F1,$5F,$E2,$F7,$FB,$86,$69,$FD,$BF
	.dc.b	$A7,$FE,$A7,$9F,$D3,$35,$DC,$73,$7E,$3C,$67,$1F,$C8,$37,$3F,$87
	.dc.b	$5F,$EB,$0F,$A6,$FE,$87,$7F,$E8,$76,$BF,$ED,$F6,$BF,$EE,$B6,$BF
	.dc.b	$A5,$77,$7F,$E3,$7E,$EF,$F1,$D1,$D7,$CD,$FF,$45,$7D,$A7,$FB,$07
	.dc.b	$17,$E7,$EF,$90,$C3,$72,$FF,$44,$D7,$07,$FB,$76,$3F,$BC,$BE,$37
	.dc.b	$C0,$D4,$FF,$A9,$FF,$1B,$7D,$4F,$FA,$F5,$B1,$FE,$BC,$6B,$F8,$FF
	.dc.b	$15,$F0,$CF,$B4,$E1,$72,$B8,$3F,$EA,$38,$FF,$6B,$FD,$7B,$4F,$E0
	.dc.b	$DC,$9F,$E3,$D8,$C5,$FF,$81,$63,$13,$AE,$FF,$EA,$5F,$5D,$FE,$53
	.dc.b	$9C,$D7,$F8,$2F,$6F,$E2,$0B,$0E,$6F,$FB,$B4,$C6,$1F,$C5,$67,$73
	.dc.b	$A6,$C7,$79,$68,$F8,$77,$93,$98,$C3,$B9,$EA,$7A,$73,$FF,$0F,$7C
	.dc.b	$4F,$F6,$8F,$8F,$E8,$D9,$D1,$FE,$CE,$F5,$3F,$DD,$DA,$FF,$98,$71
	.dc.b	$F9,$BF,$69,$C2,$7F,$DE,$5A,$FF,$AE,$5B,$7D,$9A,$F3,$C7,$DC,$FF
	.dc.b	$33,$DB,$F9,$6E,$5F,$ED,$D6,$C7,$C8,$3E,$47,$93,$FE,$E0,$71,$1A
	.dc.b	$7E,$6F,$C8,$BD,$CF,$EB,$51,$C6,$31,$E3,$7F,$BA,$5B,$19,$9D,$BF
	.dc.b	$43,$DF,$79,$1E,$36,$C7,$81,$AE,$E3,$CD,$93,$EC,$1F,$49,$FD,$63
	.dc.b	$8B,$FD,$11,$39,$8C,$71,$FF,$86,$7A,$9F,$41,$8C,$BF,$DF,$B8,$F2
	.dc.b	$E3,$FA,$78,$F1,$6C,$66,$59,$7D,$5C,$39,$D6,$B2,$E9,$4F,$7B,$FD
	.dc.b	$25,$3F,$95,$8F,$EC,$91,$87,$F2,$8D,$8F,$CC,$B6,$37,$F0,$FD,$23
	.dc.b	$86,$1F,$85,$EB,$7C,$FE,$6A,$61,$F8,$CC,$3C,$51,$EF,$E9,$5F,$0A
	.dc.b	$D6,$3D,$37,$FF,$67,$FD,$87,$97,$C2,$FF,$01,$FC,$1A,$31,$7F,$2F
	.dc.b	$D4,$FF,$03,$D6,$A7,$3A,$7F,$B5,$5B,$7C,$74,$DE,$1C,$4F,$E2,$A7
	.dc.b	$81,$CD,$7E,$C3,$83,$FB,$54,$EA,$3F,$CA,$7F,$60,$FA,$1D,$EF,$FC
	.dc.b	$4B,$D5,$E3,$6A,$D3,$C6,$74,$F3,$35,$5C,$B0,$EF,$87,$F6,$B9,$C9
	.dc.b	$E1,$5E,$5B,$DB,$FE,$19,$1F,$E9,$B8,$C4,$FF,$43,$9C,$CF,$06,$CE
	.dc.b	$A3,$FD,$8C,$DD,$BF,$C6,$38,$BF,$10,$FC,$56,$1E,$1A,$D3,$FE,$2D
	.dc.b	$FB,$17,$4F,$C8,$EF,$7D,$37,$F3,$AA,$F4,$56,$9F,$F2,$34,$67,$3F
	.dc.b	$F6,$B4,$E5,$FE,$04,$D7,$F9,$5B,$5F,$05,$FB,$23,$86,$7F,$E4,$89
	.dc.b	$46,$2F,$FC,$C7,$2F,$53,$EE,$D6,$AE,$FD,$8F,$F8,$AE,$FF,$CB,$5E
	.dc.b	$87,$BA,$BF,$D8,$D3,$97,$FC,$37,$19,$51,$97,$FA,$0B,$5F,$F2,$93
	.dc.b	$5E,$7E,$AF,$D4,$8E,$98,$FA,$5B,$7F,$E7,$AD,$3F,$8F,$3F,$81,$8C
	.dc.b	$DF,$93,$E3,$BE,$86,$2F,$7B,$D7,$18,$6D,$FF,$2D,$AF,$3C,$F8,$49
	.dc.b	$19,$B4,$70,$9F,$F1,$CE,$1F,$EA,$DA,$F1,$4F,$CA,$FA,$FF,$F6,$0C
	.dc.b	$7F,$A0,$7E,$BB,$E2,$46,$2F,$FC,$83,$18,$F8,$AF,$C9,$CE,$67,$FB
	.dc.b	$3B,$5F,$F3,$C3,$5F,$F7,$B3,$CF,$4F,$BC,$DB,$FF,$3F,$7D,$8B,$E2
	.dc.b	$BC,$98,$79,$7E,$23,$F2,$99,$7C,$5B,$FD,$56,$1F,$EB,$47,$D1,$B6
	.dc.b	$DF,$8D,$19,$CF,$FA,$A3,$5F,$3D,$FA,$8E,$0F,$F9,$84,$64,$F8,$7C
	.dc.b	$E9,$FE,$89,$FB,$1F,$F3,$76,$3F,$EA,$76,$DF,$1C,$5F,$5D,$FA,$B9
	.dc.b	$CD,$FC,$08,$F9,$24,$BF,$3B,$FB,$EF,$03,$71,$F2,$A3,$F3,$F1,$8C
	.dc.b	$3E,$5C,$7F,$18,$F9,$9F,$EC,$35,$FC,$55,$B3,$FE,$78,$C1,$F6,$3B
	.dc.b	$F3,$BF,$01,$E4,$7D,$A2,$61,$F2,$1F,$14,$F8,$0E,$D7,$67,$F8,$D7
	.dc.b	$D7,$7F,$AC,$9F,$E4,$5C,$5F,$3F,$5F,$DD,$4E,$6B,$F9,$14,$E6,$BF
	.dc.b	$D7,$93,$99,$FF,$17,$B5,$CF,$91,$FC,$82,$31,$00,$F5,$BE,$F3,$FD
	.dc.b	$03,$27,$D7,$FE,$1E,$FE,$E5,$EF,$0E,$5E,$13,$D7,$FE,$E9,$97,$F2
	.dc.b	$85,$F6,$3F,$D1,$1C,$D7,$07,$2F,$F4,$37,$38,$7F,$8F,$9C,$BF,$E5
	.dc.b	$D7,$37,$DE,$BA,$7C,$07,$5F,$66,$E1,$CD,$B8,$71,$6E,$3E,$57,$FA
	.dc.b	$EF,$D8,$FF,$C0,$1F,$95,$BD,$FF,$D4,$2B,$FD,$E3,$39,$C3,$FB,$02
	.dc.b	$9D,$47,$CB,$8D,$61,$FE,$C1,$75,$E6,$F4,$C3,$AB,$6B,$9D,$F2,$B5
	.dc.b	$F2,$C3,$C3,$39,$F3,$BF,$89,$FA,$9F,$97,$E9,$FE,$EB,$18,$C7,$F3
	.dc.b	$11,$8F,$0B,$E5,$CE,$72,$F9,$5F,$6F,$9C,$3E,$A7,$CC,$FD,$5F,$A3
	.dc.b	$B9,$7D,$36,$DF,$D1,$6E,$9F,$75,$FE,$93,$CB,$FC,$CB,$9F,$F0,$FE
	.dc.b	$A6,$3C,$9F,$C5,$74,$A3,$19,$7F,$31,$BF,$E8,$59,$7E,$4A,$FF,$2C
	.dc.b	$49,$FD,$2B,$EE,$FB,$DF,$E3,$91,$83,$FC,$15,$F1,$FF,$E3,$BD,$3F
	.dc.b	$BB,$BF,$B8,$6D,$FF,$85,$23,$13,$F9,$87,$83,$73,$D4,$ED,$F6,$FC
	.dc.b	$07,$AB,$C5,$B5,$FE,$21,$9F,$D2,$C6,$2B,$F4,$B5,$F0,$3E,$0B,$18
	.dc.b	$FF,$73,$FB,$3F,$E2,$0F,$F8,$13,$C0,$FE,$E3,$18,$F4,$BF,$59,$87
	.dc.b	$ED,$63,$F9,$1B,$FE,$AA,$6B,$FB,$90,$E1,$E1,$F2,$1F,$E1,$E8,$FE
	.dc.b	$3D,$1A,$8F,$95,$97,$81,$F0,$FC,$5F,$EF,$FE,$CF,$FC,$D3,$5F,$CE
	.dc.b	$6D,$39,$F8,$1D,$EF,$FE,$31,$9C,$BF,$0E,$FD,$37,$E4,$BF,$F0,$6E
	.dc.b	$3F,$E4,$08,$68,$BF,$E2,$5B,$B0,$FF,$31,$57,$F4,$9B,$E1,$7B,$8F
	.dc.b	$F4,$AC,$EA,$3F,$C3,$D3,$AB,$F8,$8D,$AD,$FC,$47,$8B,$DC,$FD,$27
	.dc.b	$47,$F9,$1B,$B3,$FB,$EB,$7B,$4D,$26,$3F,$E5,$F9,$DB,$FC,$5F,$E7
	.dc.b	$39,$B4,$98,$FF,$17,$FA,$9F,$E4,$13,$27,$3A,$7F,$25,$9C,$57,$C4
	.dc.b	$9F,$E6,$97,$FE,$10,$F9,$9F,$21,$23,$FD,$89,$19,$6E,$49,$FE,$00
	.dc.b	$7F,$9A,$7C,$5B,$3E,$99,$7D,$37,$BC,$91,$C9,$B8,$78,$C6,$1F,$E2
	.dc.b	$28,$FF,$3A,$7D,$3F,$F0,$D6,$57,$57,$2D,$25,$73,$AE,$51,$FE,$81
	.dc.b	$F8,$15,$F2,$AF,$07,$B8,$FB,$27,$CC,$FF,$9F,$23,$0D,$8C,$E1,$FE
	.dc.b	$D2,$6D,$7E,$94,$FF,$39,$B8,$DF,$F1,$57,$CA,$BC,$5D,$FF,$E9,$E8
	.dc.b	$C5,$F8,$BE,$C9,$F6,$5F,$E9,$CF,$91,$FF,$35,$4F,$F9,$FA,$73,$87
	.dc.b	$FA,$6E,$73,$69,$CD,$78,$E3,$17,$EB,$9C,$CF,$EE,$E7,$37,$FE,$EC
	.dc.b	$9E,$38,$FA,$1B,$1F,$EE,$9F,$B9,$F4,$46,$2F,$EB,$FD,$9B,$D2,$FF
	.dc.b	$59,$FD,$87,$F9,$FF,$0F,$CD,$4E,$5F,$F5,$94,$66,$BE,$B4,$E5,$F9
	.dc.b	$5C,$48,$C4,$7F,$11,$8C,$5F,$EA,$E7,$33,$F9,$A7,$96,$5F,$ED,$86
	.dc.b	$E1,$FE,$08,$C1,$F4,$5F,$D6,$A7,$37,$EA,$C3,$07,$AA,$6D,$FF,$B0
	.dc.b	$67,$FD,$66,$7E,$5F,$D6,$7D,$9F,$AF,$2F,$F7,$DD,$7F,$B3,$E3,$0F
	.dc.b	$FB,$F0,$FF,$41,$57,$23,$F1,$05,$34,$7F,$A1,$FD,$BF,$F5,$3B,$E0
	.dc.b	$AF,$F6,$2E,$6F,$B3,$CD,$6A,$9F,$23,$FE,$2F,$F2,$F5,$1F,$CE,$7C
	.dc.b	$5F,$E3,$91,$88,$FE,$79,$3B,$BF,$CD,$56,$9F,$2C,$66,$79,$65,$5E
	.dc.b	$22,$FA,$3F,$E5,$A6,$BE,$43,$FD,$AE,$F4,$69,$FA,$99,$FF,$23,$71
	.dc.b	$F3,$8C,$1A,$1F,$2A,$F8,$9F,$E7,$B1,$98,$C6,$BC,$B0,$FE,$B1,$18
	.dc.b	$9D,$57,$0B,$ED,$1D,$23,$9D,$7C,$5A,$FF,$7F,$4F,$BD,$AF,$BF,$2E
	.dc.b	$B5,$C1,$B8,$6F,$E3,$BF,$B4,$61,$CD,$B3,$AB,$E4,$7F,$99,$70,$FF
	.dc.b	$55,$3F,$E6,$7B,$FF,$76,$46,$23,$FD,$3C,$D8,$FC,$1E,$BF,$DD,$5D
	.dc.b	$FC,$4F,$E3,$A7,$85,$7F,$15,$6B,$F1,$23,$47,$1D,$7F,$CD,$BF,$64
	.dc.b	$F7,$B1,$4E,$34,$E6,$3C,$D5,$CD,$FE,$B1,$F6,$5F,$CA,$9B,$DF,$CB
	.dc.b	$A5,$39,$9F,$D5,$7D,$93,$C7,$D4,$E7,$6F,$EE,$57,$C3,$0C,$9E,$4C
	.dc.b	$D8,$71,$C7,$B9,$FD,$63,$0E,$CF,$F0,$96,$BF,$D8,$1A,$FF,$50,$7C
	.dc.b	$19,$7F,$A3,$A5,$E4,$77,$6D,$76,$3F,$A4,$F9,$2F,$6B,$E3,$F3,$FF
	.dc.b	$7C,$6C,$66,$7F,$A9,$7D,$D6,$27,$F8,$0C,$2B,$83,$FC,$11,$AF,$F3
	.dc.b	$BA,$E4,$D9,$F8,$5E,$3B,$EC,$E9,$D6,$7E,$A9,$D7,$C8,$7C,$87,$99
	.dc.b	$B1,$AC,$7F,$B2,$CF,$4F,$6B,$4F,$9A,$33,$3E,$5A,$E2,$FF,$4F,$6B
	.dc.b	$FC,$DF,$CC,$FC,$B6,$FF,$E0,$E6,$C7,$F7,$E6,$F0,$B3,$47,$E1,$CD
	.dc.b	$F1,$E3,$A3,$07,$C0,$8B,$7F,$E0,$F8,$AF,$F3,$5A,$E6,$E4,$FF,$0C
	.dc.b	$C7,$D0,$AE,$91,$83,$FC,$3B,$EE,$78,$48,$FF,$82,$7E,$5E,$BF,$D5
	.dc.b	$B0,$FF,$47,$3C,$08,$7F,$80,$F1,$FF,$6B,$79,$DF,$97,$F8,$6F,$F1
	.dc.b	$1C,$FF,$B4,$63,$1E,$8E,$59,$19,$3F,$DA,$D1,$FA,$9F,$AC,$FB,$D1
	.dc.b	$FE,$45,$6B,$FE,$1F,$6F,$5F,$FB,$8E,$B8,$72,$7B,$E7,$75,$FD,$AD
	.dc.b	$AF,$F9,$09,$AF,$FB,$DF,$EC,$1F,$67,$FC,$0A,$11,$87,$F1,$A6,$23
	.dc.b	$5F,$64,$F9,$2E,$15,$BA,$FF,$1F,$09,$7D,$9F,$03,$77,$DF,$E7,$BE
	.dc.b	$6D,$70,$6E,$9F,$E7,$58,$46,$39,$BF,$D2,$DF,$66,$F8,$1E,$5C,$BE
	.dc.b	$98,$7C,$9D,$33,$87,$AB,$EC,$FF,$0E,$F9,$9B,$97,$76,$E3,$FE,$A3
	.dc.b	$FA,$9F,$DC,$57,$1E,$7F,$EF,$B5,$CA,$3F,$C8,$8D,$7F,$D5,$4D,$41
	.dc.b	$EA,$E2,$5F,$01,$F4,$5B,$1D,$CF,$46,$BF,$56,$BE,$1C,$60,$FA,$B3
	.dc.b	$FC,$61,$F6,$4F,$3F,$06,$46,$CF,$F0,$E7,$D9,$FF,$9F,$5F,$E1,$3D
	.dc.b	$48,$E5,$A7,$FA,$96,$31,$E3,$7F,$33,$9E,$15,$CB,$AB,$13,$23,$9C
	.dc.b	$EA,$FE,$D7,$D9,$7F,$0A,$6F,$87,$F8,$F8,$C7,$07,$FB,$47,$EB,$5E
	.dc.b	$7E,$6B,$FC,$57,$5F,$C4,$31,$FF,$6B,$35,$F9,$AD,$4F,$47,$E8,$D7
	.dc.b	$FB,$41,$FC,$DB,$5F,$CB,$C6,$F4,$77,$FF,$84,$67,$E2,$AF,$A5,$FD
	.dc.b	$F1,$FB,$FF,$52,$4E,$A3,$2F,$08,$FD,$E7,$C1,$48,$FD,$E3,$FC,$1F
	.dc.b	$D4,$C4,$7F,$93,$FE,$67,$E0,$BF,$51,$AF,$F0,$19,$FD,$B6,$7F,$B1
	.dc.b	$8F,$5F,$CE,$E0,$CF,$85,$23,$24,$3F,$6B,$3C,$5B,$1C,$5D,$70,$FE
	.dc.b	$C6,$74,$FE,$0F,$4F,$E7,$63,$EC,$FF,$90,$1F,$C8,$7C,$99,$7F,$22
	.dc.b	$32,$FC,$A8,$FE,$4D,$DF,$FC,$9F,$E7,$A4,$FC,$E8,$CB,$F4,$23,$2F
	.dc.b	$F9,$CE,$FF,$81,$3F,$80,$CB,$F1,$44,$7E,$EF,$DB,$D7,$FC,$E3,$C7
	.dc.b	$EF,$7F,$D7,$4D,$7F,$17,$3F,$8E,$3E,$6F,$BF,$23,$F9,$0C,$FC,$2E
	.dc.b	$7F,$DD,$FC,$BF,$F9,$8F,$AF,$FC,$52,$DA,$FE,$E6,$D7,$D1,$E0,$7F
	.dc.b	$2D,$FB,$4F,$F1,$AC,$7F,$04,$6A,$69,$AE,$9F,$C9,$FB,$BF,$D3,$49
	.dc.b	$3D,$BE,$D3,$F3,$5C,$8C,$FC,$B5,$FE,$5A,$8F,$E5,$58,$38,$7C,$47
	.dc.b	$A2,$75,$5F,$C5,$B0,$F1,$4F,$FA,$6A,$B8,$47,$F8,$69,$AF,$FA,$71
	.dc.b	$AF,$1C,$D8,$FF,$B9,$F1,$F2,$4A,$72,$FC,$89,$D6,$1E,$1B,$EB,$7F
	.dc.b	$D0,$78,$9F,$25,$FD,$55,$B8,$7F,$54,$6A,$9E,$F8,$BE,$D1,$F2,$9C
	.dc.b	$70,$FF,$5A,$E1,$C3,$0F,$F7,$E3,$5F,$84,$D7,$F9,$CB,$DB,$0E,$BF
	.dc.b	$68,$F8,$DF,$C0,$B6,$BC,$97,$E7,$F9,$D9,$1E,$77,$F0,$F9,$7F,$94
	.dc.b	$A3,$AC,$79,$08,$FF,$92,$31,$46,$23,$DD,$AF,$E4,$B0,$7B,$5E,$23
	.dc.b	$BC,$FF,$72,$FB,$3F,$EC,$CF,$F7,$A6,$B9,$F3,$FF,$33,$F0,$FF,$DB
	.dc.b	$F6,$FE,$E1,$5A,$AF,$C2,$07,$0F,$E1,$DA,$FF,$5E,$6B,$BA,$FD,$2F
	.dc.b	$7F,$C5,$8F,$49,$F8,$37,$15,$E0,$9F,$F0,$95,$7F,$39,$AF,$07,$8D
	.dc.b	$83,$F5,$FF,$69,$EA,$BF,$3C,$7F,$BC,$2B,$FA,$11,$1C,$1E,$2C,$A5
	.dc.b	$FC,$38,$EF,$F6,$92,$3E,$33,$FB,$CF,$89,$FF,$1D,$C7,$CB,$8E,$C4
	.dc.b	$7F,$CF,$B6,$8C,$47,$BB,$95,$E7,$FF,$3A,$CB,$F2,$AD,$E0,$F9,$57
	.dc.b	$FE,$31,$FA,$9F,$EA,$18,$7F,$8A,$9B,$7F,$E2,$36,$D7,$F8,$33,$B1
	.dc.b	$F8,$43,$E0,$39,$79,$9F,$E9,$0C,$58,$74,$7E,$DB,$5F,$E0,$0D,$7F
	.dc.b	$D8,$5F,$00,$F9,$6F,$F6,$2E,$4F,$B5,$7C,$86,$1B,$7F,$E0,$AB,$E0
	.dc.b	$DB,$D5,$48,$C1,$FE,$98,$F8,$37,$DA,$BE,$D7,$FD,$94,$D9,$D3,$5F
	.dc.b	$F4,$07,$DA,$EA,$64,$69,$FF,$53,$C7,$F9,$6F,$ED,$3C,$72,$E7,$FF
	.dc.b	$27,$C6,$1B,$97,$F2,$17,$FC,$FF,$C4,$FB,$3F,$F2,$E4,$6A,$FF,$C7
	.dc.b	$6D,$7A,$78,$3F,$CA,$70,$E8,$FD,$5A,$E2,$E7,$0F,$E5,$DC,$1F,$AC
	.dc.b	$AF,$2B,$63,$37,$B3,$FC,$93,$1F,$CD,$7E,$27,$FC,$2B,$F2,$3E,$7C
	.dc.b	$3F,$C9,$B1,$8A,$F2,$C6,$3D,$2F,$F7,$D4,$79,$8F,$BB,$F6,$BC,$5F
	.dc.b	$F6,$A4,$6B,$4F,$F4,$FD,$F3,$9E,$0D,$8F,$8B,$C6,$BE,$54,$D3,$F9
	.dc.b	$7D,$FF,$B8,$1B,$1F,$C2,$8D,$79,$3F,$E6,$C7,$E1,$35,$FE,$A5,$F2
	.dc.b	$7F,$E2,$FE,$EF,$F7,$4B,$63,$93,$5F,$E2,$AD,$7E,$97,$C0,$3C,$6F
	.dc.b	$13,$63,$C5,$87,$FA,$32,$72,$7F,$BC,$FE,$D5,$F4,$FF,$D6,$36,$F8
	.dc.b	$71,$8E,$2F,$9B,$87,$E9,$BE,$8F,$5C,$5F,$4A,$B5,$3F,$21,$B7,$CB
	.dc.b	$0D,$CF,$1C,$39,$EE,$BD,$E1,$FD,$CF,$E0,$3F,$D3,$75,$E0,$BF,$F3
	.dc.b	$2F,$C3,$72,$E8,$FC,$B7,$C7,$F4,$9C,$D6,$EF,$8C,$7E,$04,$C9,$F8
	.dc.b	$3A,$E3,$3B,$BE,$55,$C3,$4F,$F5,$AB,$71,$FC,$E6,$3A,$AF,$5F,$6B
	.dc.b	$ED,$7B,$F8,$BF,$E0,$08,$C4,$FF,$80,$A3,$FD,$19,$D1,$FD,$F3,$3F
	.dc.b	$E0,$57,$C1,$FF,$53,$DF,$07,$37,$C5,$FE,$41,$5C,$75,$FE,$97,$5B
	.dc.b	$7F,$A6,$EA,$9F,$ED,$35,$FE,$B2,$C7,$FC,$26,$BE,$D5,$F4,$5D,$70
	.dc.b	$FE,$43,$E8,$DA,$D4,$E7,$1E,$93,$9B,$E3,$5F,$C8,$63,$15,$CE,$FF
	.dc.b	$98,$FC,$25,$F1,$8F,$3F,$B5,$FF,$4F,$E1,$FE,$EE,$9C,$B6,$78,$1B
	.dc.b	$E2,$B1,$FE,$F4,$FB,$57,$37,$FB,$9F,$1F,$F1,$D4,$62,$7F,$61,$08
	.dc.b	$C3,$FB,$36,$BF,$DF,$9C,$ED,$FE,$F1,$6C,$FF,$98,$21,$E7,$FB,$67
	.dc.b	$35,$FE,$DE,$6B,$DE,$FE,$67,$77,$F8,$D6,$B8,$47,$C5,$FB,$5F,$F0
	.dc.b	$CE,$3F,$E6,$8F,$2F,$E0,$E1,$FE,$06,$13,$A7,$FB,$64,$6B,$2F,$F2
	.dc.b	$6C,$6B,$D1,$E8,$78,$32,$FF,$5C,$61,$F8,$13,$FC,$BD,$87,$3B,$E2
	.dc.b	$D8,$CD,$71,$F8,$8E,$71,$C7,$0F,$F7,$DD,$3E,$37,$DB,$97,$F9,$B6
	.dc.b	$B5,$5E,$89,$CF,$BB,$EF,$64,$FF,$A1,$B3,$FF,$55,$3C,$30,$E8,$D9
	.dc.b	$DD,$FF,$82,$EB,$6F,$C5,$AD,$BF,$1F,$ED,$1F,$11,$F4,$3E,$CB,$E3
	.dc.b	$63,$FE,$7D,$9C,$BF,$E9,$68,$CD,$04,$E9,$E9,$5F,$9B,$FB,$57,$C4
	.dc.b	$FE,$76,$B5,$C1,$F4,$9B,$8F,$7B,$E1,$5B,$78,$7C,$47,$97,$6B,$5E
	.dc.b	$36,$E5,$EB,$9F,$07,$C2,$E3,$2E,$D1,$6F,$FD,$31,$9A,$B1,$7F,$E8
	.dc.b	$FF,$B5,$7B,$DF,$92,$6D,$FD,$6D,$5F,$65,$9A,$F6,$32,$7E,$DF,$87
	.dc.b	$C3,$0F,$9D,$3F,$BD,$8C,$69,$F0,$63,$17,$C6,$BF,$DF,$9E,$5F,$F0
	.dc.b	$9F,$8E,$FE,$A9,$8F,$FB,$1B,$37,$DA,$E6,$BE,$67,$6C,$AF,$0C,$DC
	.dc.b	$3F,$49,$A7,$09,$C9,$F9,$5F,$11,$97,$F8,$22,$B7,$7F,$C4,$FE,$C7
	.dc.b	$F9,$44,$7F,$9F,$1A,$FF,$B0,$5A,$FE,$0D,$A9,$FE,$61,$B7,$EC,$1B
	.dc.b	$87,$E2,$73,$7D,$97,$E8,$EF,$F7,$55,$A7,$A7,$C3,$F8,$CF,$E6,$15
	.dc.b	$C3,$97,$18,$7A,$FD,$6F,$ED,$79,$7F,$B2,$B5,$43,$ED,$7F,$9A,$3A
	.dc.b	$AF,$F3,$23,$97,$97,$A3,$A9,$DB,$63,$B7,$BD,$CB,$D9,$E6,$D8,$C9
	.dc.b	$E3,$8C,$B4,$FE,$09,$F6,$BF,$ED,$DF,$93,$FE,$94,$FF,$B7,$FC,$DB
	.dc.b	$E1,$26,$C6,$1B,$5C,$A7,$4D,$9D,$34,$D4,$F0,$E1,$FE,$C3,$E1,$FC
	.dc.b	$15,$E8,$7F,$15,$D3,$E2,$FD,$AF,$F5,$D8,$C6,$1F,$E7,$1B,$E1,$18
	.dc.b	$BF,$ED,$3B,$AF,$85,$75,$E0,$FA,$3F,$2F,$F4,$3E,$87,$E8,$B7,$2F
	.dc.b	$B0,$DC,$79,$B7,$8F,$F1,$1E,$57,$F5,$79,$FC,$73,$5F,$C7,$37,$A3
	.dc.b	$FA,$04,$7E,$5E,$3F,$34,$7A,$FE,$D7,$F8,$9E,$1F,$CC,$2B,$F9,$D5
	.dc.b	$74,$8C,$56,$FC,$7E,$31,$FA,$7F,$8F,$FD,$36,$3F,$8D,$7F,$6B,$7F
	.dc.b	$E0,$A7,$33,$FE,$AF,$73,$3B,$9F,$F0,$24,$62,$75,$18,$EA,$FD,$87
	.dc.b	$BF,$27,$FA,$54,$7E,$9B,$E4,$B1,$5F,$A6,$AB,$82,$FF,$59,$DC,$BC
	.dc.b	$C7,$F3,$7B,$7D,$A7,$F1,$57,$FD,$EE,$D7,$4D,$7F,$9F,$35,$E3,$E2
	.dc.b	$FF,$85,$E3,$F8,$DC,$FF,$8C,$B7,$FE,$58,$FF,$5C,$8C,$D7,$2A,$E0
	.dc.b	$7D,$3F,$22,$D6,$FE,$D3,$A1,$FB,$EF,$B4,$FE,$05,$7E,$88,$CC,$EC
	.dc.b	$F8,$B3,$AC,$13,$FE,$2B,$BF,$F1,$7D,$FF,$81,$5B,$A7,$C7,$8F,$F4
	.dc.b	$05,$FF,$B6,$70,$FE,$E7,$5A,$7E,$43,$67,$FB,$03,$76,$57,$19,$ED
	.dc.b	$EC,$7E,$8B,$96,$61,$FC,$57,$E7,$2D,$7E,$CA,$6E,$5F,$C5,$FA,$95
	.dc.b	$8B,$FE,$2A,$FF,$18,$3F,$8C,$FD,$55,$8E,$84,$C3,$F6,$71,$97,$F7
	.dc.b	$FF,$31,$FB,$43,$FC,$A9,$EE,$FF,$08,$8C,$D7,$EF,$67,$4F,$FB,$D7
	.dc.b	$C0,$B3,$FC,$23,$E9,$BF,$7A,$72,$F8,$FB,$5F,$C1,$62,$CF,$F9,$B3
	.dc.b	$05,$63,$4F,$F3,$F7,$7A,$F1,$87,$F9,$DA,$AE,$1F,$C5,$2B,$53,$D4
	.dc.b	$FE,$B1,$87,$47,$12,$C3,$EE,$7B,$2F,$B4,$7D,$5F,$F9,$DE,$30,$FC
	.dc.b	$2E,$DC,$4E,$B0,$FF,$1C,$56,$9B,$19,$C3,$D9,$5C,$2F,$A7,$D8,$BE
	.dc.b	$CF,$F8,$B6,$28,$C6,$3F,$ED,$B8,$C4,$EA,$73,$87,$ED,$4F,$F3,$AB
	.dc.b	$8C,$BD,$6D,$BF,$F4,$DF,$D9,$F8,$2E,$47,$A2,$75,$6F,$17,$FD,$39
	.dc.b	$7A,$8F,$F5,$0F,$D9,$BC,$EF,$E5,$9E,$4A,$31,$8F,$FB,$C6,$31,$3A
	.dc.b	$9C,$D7,$9E,$73,$5F,$E5,$F6,$C7,$FA,$B3,$EC,$9F,$45,$FD,$2E,$F8
	.dc.b	$EB,$FE,$BF,$6E,$1F,$EB,$A6,$DF,$A6,$F8,$CF,$59,$CE,$1D,$AF,$96
	.dc.b	$3F,$EA,$FB,$DE,$1E,$E8,$FF,$56,$3F,$EA,$83,$FD,$7F,$F6,$7F,$E9
	.dc.b	$4C,$7F,$CE,$31,$9B,$FF,$04,$4E,$63,$FD,$18,$66,$7E,$B7,$B1,$DF
	.dc.b	$EC,$FA,$B2,$39,$B9,$69,$DA,$5F,$1B,$FE,$69,$FB,$65,$FE,$07,$C3
	.dc.b	$FE,$29,$18,$AF,$C2,$56,$EB,$57,$67,$22,$B8,$3F,$E2,$D8,$D3,$FD
	.dc.b	$06,$AB,$FD,$62,$76,$FF,$9E,$E3,$FC,$F9,$C9,$FD,$26,$3A,$B5,$E4
	.dc.b	$7F,$B2,$BC,$9F,$CC,$3E,$C3,$07,$FA,$17,$E4,$DF,$0B,$FB,$58,$C8
	.dc.b	$8C,$ED,$E0,$AF,$EA,$D1,$89,$DB,$6F,$C4,$E7,$1F,$7B,$FE,$E6,$9F
	.dc.b	$7B,$5F,$89,$9F,$4C,$7C,$55,$C1,$B1,$E0,$F8,$9F,$EC,$D3,$D1,$B1
	.dc.b	$C7,$2E,$BE,$7F,$F6,$B7,$FD,$97,$8F,$FA,$7B,$1F,$F5,$DB,$5F,$D1
	.dc.b	$6F,$BD,$BD,$B7,$67,$E3,$35,$F8,$8E,$79,$DC,$6B,$D6,$1D,$71,$FF
	.dc.b	$2F,$7D,$A3,$D9,$E0,$DC,$BC,$79,$77,$E6,$7D,$A7,$F2,$49,$FD,$C3
	.dc.b	$5F,$E5,$5A,$FF,$AF,$67,$7D,$BE,$3F,$B4,$78,$FC,$F0,$E1,$A7,$F6
	.dc.b	$BB,$DE,$3D,$30,$E1,$7B,$64,$64,$FF,$49,$61,$D1,$C1,$FC,$37,$0F
	.dc.b	$23,$FE,$19,$BE,$AE,$6B,$F2,$2B,$95,$E5,$3D,$6F,$9D,$18,$6C,$66
	.dc.b	$7F,$A5,$FD,$A6,$26,$47,$F4,$29,$D3,$FD,$6E,$72,$FF,$37,$AE,$8F
	.dc.b	$F3,$D9,$DB,$FE,$2F,$8C,$BB,$F1,$DF,$19,$E4,$6C,$7F,$50,$6B,$FD
	.dc.b	$A7,$D7,$D4,$C8,$D3,$F8,$5F,$21,$3F,$77,$C9,$FC,$27,$17,$D9,$6D
	.dc.b	$FF,$6B,$E3,$E1,$1B,$64,$7F,$9F,$5C,$B4,$87,$FA,$21,$CB,$4C,$F1
	.dc.b	$FF,$7C,$8C,$BF,$DF,$BE,$4F,$C8,$49,$FE,$F3,$F2,$75,$FF,$49,$C5
	.dc.b	$9F,$F5,$C6,$4A,$FD,$17,$A1,$FE,$07,$BF,$F5,$F7,$9B,$83,$3E,$7D
	.dc.b	$8C,$90,$FF,$4A,$BF,$A1,$FB,$37,$DA,$52,$4E,$F2,$FC,$D4,$EA,$3F
	.dc.b	$AE,$35,$FD,$29,$FE,$A8,$71,$1F,$E3,$9B,$FD,$D7,$DA,$3E,$CA,$92
	.dc.b	$C7,$21,$F6,$8F,$8F,$D4,$66,$B9,$3C,$30,$FF,$2C,$28,$CB,$CE,$B9
	.dc.b	$3B,$E2,$FF,$40,$7C,$E3,$87,$F8,$73,$67,$47,$91,$C3,$E5,$FB,$37
	.dc.b	$BD,$C2,$A4,$F0,$15,$C3,$2F,$2D,$F0,$9D,$7D,$9B,$9B,$F8,$F3,$63
	.dc.b	$31,$FC,$87,$EA,$3F,$75,$18,$21,$F1,$E7,$83,$F2,$63,$6E,$67,$F8
	.dc.b	$1A,$9D,$BF,$E2,$D6,$BF,$EA,$36,$BF,$51,$F5,$BE,$8C,$3C,$E7,$97
	.dc.b	$DB,$B1,$92,$1E,$37,$F9,$C7,$D9,$B9,$6C,$49,$F9,$FE,$5A,$B5,$18
	.dc.b	$6B,$F8,$E8,$D4,$F1,$8C,$4E,$B4,$FC,$1F,$D8,$FF,$08,$6F,$6F,$F4
	.dc.b	$B8,$C7,$0F,$FB,$07,$EB,$7F,$A5,$61,$F6,$67,$2F,$E5,$1A,$9B,$1C
	.dc.b	$23,$F3,$11,$FC,$59,$FF,$2C,$35,$FF,$67,$35,$E9,$5F,$CD,$B2,$75
	.dc.b	$7F,$81,$27,$E1,$AF,$9A,$FD,$F9,$FC,$03,$D2,$93,$A8,$CB,$C2,$3F
	.dc.b	$79,$EF,$C8,$FD,$BB,$FC,$23,$D2,$C4,$7E,$E3,$C9,$77,$2B,$3B,$FF
	.dc.b	$AC,$6F,$FB,$96,$DF,$AA,$C3,$F6,$CB,$E6,$32,$6B,$B6,$4F,$E8,$E7
	.dc.b	$4D,$9C,$BC,$A7,$8B,$93,$83,$FC,$7B,$C3,$FE,$3F,$F6,$72,$7E,$2C
	.dc.b	$65,$F8,$F1,$97,$F0,$DB,$FF,$3E,$E0,$F9,$BD,$9F,$CA,$3E,$46,$57
	.dc.b	$CE,$8F,$E5,$66,$1E,$31,$87,$F0,$98,$FE,$24,$8F,$B7,$EC,$EB,$F8
	.dc.b	$28,$BB,$7F,$14,$D5,$5A,$AF,$C2,$D7,$ED,$39,$7E,$64,$60,$FF,$24
	.dc.b	$FC,$83,$D9,$FC,$E7,$47,$F2,$68,$C7,$5F,$F8,$9D,$B5,$F8,$8F,$87
	.dc.b	$FF,$14,$7D,$97,$F5,$E8,$34,$FE,$C3,$3C,$23,$E3,$E1,$FD,$30,$8F
	.dc.b	$F2,$28,$7D,$9B,$E2,$BC,$15,$FE,$4B,$DF,$F9,$8E,$0E,$0F,$09,$AE
	.dc.b	$7B,$97,$F2,$59,$D3,$67,$4F,$7C,$7F,$D8,$16,$D7,$E0,$46,$1B,$3D
	.dc.b	$0F,$E9,$1C,$4F,$95,$FF,$24,$36,$3F,$9A,$35,$FE,$AF,$AB,$EC,$DF
	.dc.b	$21,$C2,$E5,$70,$7E,$74,$EA,$30,$7C,$5F,$B3,$7C,$5F,$D3,$6C,$68
	.dc.b	$F2,$FC,$AF,$90,$8F,$CD,$AF,$F2,$3E,$7F,$D8,$A3,$FA,$A4,$E9,$E2
	.dc.b	$6F,$D9,$C6,$DF,$16,$1A,$31,$FF,$10,$C7,$97,$D4,$DE,$3D,$7E,$CB
	.dc.b	$FA,$94,$FF,$6F,$6E,$1F,$DA,$9B,$3F,$8F,$F8,$6F,$ED,$B3,$9F,$3F
	.dc.b	$FC,$E9,$5A,$BF,$EF,$2D,$C9,$7C,$30,$F0,$5F,$E5,$FB,$7E,$2E,$1F
	.dc.b	$49,$B8,$F5,$8F,$ED,$58,$FF,$30,$27,$91,$F9,$FF,$B3,$F0,$4A,$72
	.dc.b	$F0,$9D,$61,$E6,$7C,$1A,$49,$D1,$DB,$EC,$FD,$64,$7F,$41,$F1,$1F
	.dc.b	$E2,$D7,$E3,$46,$5A,$6B,$D6,$C3,$F6,$4F,$24,$7F,$70,$C4,$D3,$FB
	.dc.b	$9F,$D2,$FF,$3F,$9F,$F1,$03,$74,$FF,$0D,$B7,$6F,$F1,$2F,$88,$FB
	.dc.b	$37,$B9,$FE,$A5,$AD,$61,$FE,$2A,$6A,$73,$5C,$67,$FC,$CD,$F5,$DF
	.dc.b	$EA,$FC,$1F,$66,$F8,$EC,$37,$1F,$EE,$BE,$FA,$F8,$C7,$F8,$9A,$F8
	.dc.b	$DF,$03,$31,$FD,$37,$EC,$DF,$67,$FE,$AF,$6C,$7F,$8B,$BD,$34,$62
	.dc.b	$3F,$BF,$62,$FB,$3E,$17,$27,$85,$6F,$83,$FC,$CB,$5F,$EC,$7F,$B1
	.dc.b	$F3,$57,$FB,$E7,$C4,$FD,$64,$62,$1C,$DF,$E3,$C8,$C3,$FE,$41,$E1
	.dc.b	$7D,$9F,$F9,$2A,$35,$C1,$FB,$56,$DF,$53,$FC,$B3,$7C,$B0,$FA,$F7
	.dc.b	$C5,$F0,$DF,$16,$D7,$11,$5C,$56,$1E,$2F,$82,$E1,$1F,$EB,$BF,$4B
	.dc.b	$FD,$B4,$FA,$AF,$FC,$D3,$F0,$BF,$E0,$AF,$8D,$DB,$67,$CD,$83,$85
	.dc.b	$F6,$7A,$C7,$9E,$DF,$E6,$1C,$38,$61,$D4,$F0,$63,$E2,$C7,$96,$1E
	.dc.b	$49,$F4,$F9,$5F,$E5,$07,$E7,$B5,$F8,$AD,$77,$7F,$ED,$6F,$8F,$FF
	.dc.b	$0F,$78,$B7,$DD,$FE,$37,$5F,$AB,$13,$A7,$0D,$7F,$61,$C1,$FE,$70
	.dc.b	$F1,$BF,$DC,$2D,$8E,$4D,$77,$F0,$1F,$C5,$9F,$DC,$69,$F1,$73,$FE
	.dc.b	$59,$B7,$E0,$F8,$1F,$37,$8E,$2F,$D9,$C7,$F4,$8E,$CF,$9A,$E9,$B7
	.dc.b	$BC,$7E,$DC,$E5,$F9,$98,$F7,$AE,$38,$6E,$1F,$84,$3D,$3C,$36,$BE
	.dc.b	$DA,$F8,$0C,$CC,$9C,$BC,$23,$18,$78,$2B,$93,$67,$55,$C0,$F5,$F3
	.dc.b	$FF,$4D,$EB,$FF,$68,$EF,$67,$E0,$CD,$BF,$6C,$DC,$BF,$1D,$37,$4F
	.dc.b	$F5,$1D,$FF,$0A,$8C,$56,$62,$1F,$C3,$BC,$DB,$87,$F0,$28,$B8,$FF
	.dc.b	$B8,$A6,$C7,$F2,$4F,$6D,$E7,$F8,$5F,$F0,$14,$7F,$A3,$1C,$B7,$D2
	.dc.b	$D4,$ED,$B1,$93,$FD,$27,$FA,$23,$ED,$7D,$8F,$FB,$9A,$33,$18,$F8
	.dc.b	$BC,$B6,$32,$72,$8C,$B4,$CE,$7F,$E9,$BA,$E8,$FF,$A7,$71,$FE,$7B
	.dc.b	$3E,$9F,$82,$F6,$79,$3E,$D6,$C6,$4F,$74,$65,$A6,$6B,$A4,$65,$F0
	.dc.b	$3E,$C9,$F6,$B6,$7E,$0B,$67,$71,$F1,$27,$6D,$36,$58,$F5,$7C,$DF
	.dc.b	$D3,$F5,$FE,$A3,$67,$EA,$36,$77,$1F,$5E,$76,$D3,$65,$8F,$91,$E8
	.dc.b	$FD,$4F,$8F,$FA,$AD,$9F,$B2,$D9,$DC,$7D,$C9,$DB,$4D,$96,$3E,$94
	.dc.b	$7E,$2C,$FC,$67,$C4,$D9,$FC,$63,$67,$EC,$37,$8B,$FC,$69,$CD,$FE
	.dc.b	$80,$8E,$35,$F9,$3C,$2D,$FD,$FF,$8E,$FC,$C1,$F8,$EF,$22,$D7,$E4
	.dc.b	$BC,$6C,$34,$FC,$AC,$7E,$53,$CA,$91,$F9,$F8,$B6,$CF,$FC,$8D,$F4
	.dc.b	$DF,$E4,$6E,$0F,$DC,$3F,$E4,$8E,$7B,$E3,$FF,$B0,$71,$61,$C5,$B5
	.dc.b	$B7,$38,$74,$AE,$15,$B1,$5B,$70,$F8,$6B,$89,$D2,$7F,$6B,$3A,$79
	.dc.b	$D7,$5B,$F2,$E1,$E6,$7F,$7D,$9F,$EF,$F3,$B8,$FE,$FD,$B8,$FE,$AD
	.dc.b	$B8,$7D,$4E,$4F,$F0,$65,$7F,$05,$F0,$7F,$84,$78,$3F,$C2,$5B,$19
	.dc.b	$3F,$85,$46,$5A,$7F,$0D,$ED,$FE,$21,$DB,$FC,$45,$B1,$93,$F8,$94
	.dc.b	$65,$A6,$72,$FE,$01,$19,$6C,$7C,$67,$E9,$FD,$23,$8F,$F9,$87,$8C
	.dc.b	$FA,$47,$8B,$FC,$CE,$D7,$BC,$39,$5E,$EB,$53,$9A,$FE,$DC,$D7,$FC
	.dc.b	$B4,$D7,$FD,$28,$D7,$FC,$2B,$AF,$FA,$D6,$FF,$97,$78,$3F,$CC,$5F
	.dc.b	$E6,$4D,$72,$7F,$33,$32,$7B,$79,$7F,$D3,$3D,$1F,$92,$D3,$F2,$6D
	.dc.b	$9D,$73,$23,$18,$FE,$1D,$BC,$5F,$93,$6D,$7E,$35,$B3,$FE,$2C,$FD
	.dc.b	$15,$ED,$7F,$57,$5F,$79,$63,$0D,$BF,$A5,$D8,$AD,$C6,$1E,$6D,$9F
	.dc.b	$A5,$F4,$B8,$C3,$F4,$99,$3E,$EB,$FA,$AC,$7E,$9C,$4E,$E7,$FB,$84
	.dc.b	$E9,$FF,$5F,$4E,$B2,$AF,$FB,$1E,$BB,$3F,$E9,$16,$B9,$8D,$BF,$B4
	.dc.b	$D3,$EC,$3F,$AF,$DF,$F8,$0E,$5F,$C0,$9A,$FD,$06,$C6,$E3,$E9,$9B
	.dc.b	$29,$B3,$F3,$0B,$EF,$DF,$23,$FE,$07,$F5,$DF,$81,$3E,$1B,$F3,$15
	.dc.b	$AC,$3F,$D0,$93,$97,$9C,$67,$0F,$F2,$23,$5F,$F5,$83,$5F,$F7,$53
	.dc.b	$5F,$E3,$7F,$8D,$BE,$B7,$FC,$31,$F8,$53,$F3,$E7,$D7,$FE,$A6,$B5
	.dc.b	$8F,$EC,$6B,$53,$9B,$F9,$78,$2F,$6F,$F9,$5D,$AF,$FB,$0D,$AF,$FA
	.dc.b	$EB,$EA,$1F,$65,$E2,$FC,$FB,$F9,$7E,$5F,$8E,$9C,$BE,$18,$CD,$7F
	.dc.b	$2D,$E9,$56,$A3,$F9,$86,$6A,$DF,$37,$F5,$2A,$D2,$C3,$F9,$ED,$6E
	.dc.b	$FF,$C7,$15,$F9,$19,$F3,$72,$E3,$A3,$2C,$8F,$33,$96,$9F,$A6,$FC
	.dc.b	$0B,$F4,$B7,$E7,$7C,$31,$FE,$DA,$73,$8F,$F6,$67,$33,$FE,$A0,$73
	.dc.b	$3B,$D3,$FC,$99,$18,$9D,$46,$3E,$2B,$38,$5B,$F1,$36,$BC,$4D,$C3
	.dc.b	$C0,$D9,$F2,$36,$7C,$4D,$D5,$8F,$C5,$9F,$36,$1E,$0B,$F8,$1F,$07
	.dc.b	$E6,$FA,$CF,$3B,$63,$27,$C5,$8C,$B4,$F9,$1F,$67,$EE,$F0,$BD,$ED
	.dc.b	$8C,$9F,$36,$32,$D3,$38,$7C,$F8,$CB,$63,$F8,$CF,$D5,$3E,$67,$FC
	.dc.b	$EC,$BE,$72,$C6,$1B,$3C,$1B,$5F,$91,$AE,$13,$D2,$B8,$34,$E0,$59
	.dc.b	$E1,$5C,$B7,$FF,$12,$DF,$29,$FC,$E7,$D6,$D9,$FF,$14,$AF,$C0,$3F
	.dc.b	$D1,$8E,$70,$FF,$0A,$F0,$23,$51,$FC,$D9,$D5,$FF,$BA,$9C,$BD,$B1
	.dc.b	$E7,$7D,$1B,$5C,$5B,$3F,$CD,$4F,$F1,$1F,$C1,$7F,$89,$3E,$B9,$F5
	.dc.b	$DF,$E2,$69,$FF,$13,$F9,$BF,$AD,$53,$A7,$42,$B9,$56,$2B,$85,$6E
	.dc.b	$BA,$56,$9E,$D3,$AA,$FF,$72,$4E,$67,$C3,$85,$7C,$E3,$E2,$1F,$35
	.dc.b	$C2,$30,$D8,$CC,$79,$AB,$FC,$83,$18,$8F,$F7,$8F,$C3,$7F,$92,$DE
	.dc.b	$27,$F9,$1B,$E2,$3F,$DA,$5C,$1C,$30,$F8,$91,$89,$D9,$CA,$74,$F1
	.dc.b	$AE,$9A,$7C,$1A,$FF,$70,$4F,$B5,$AF,$CC,$CF,$AD,$70,$6D,$EF,$E1
	.dc.b	$BE,$A6,$1C,$DB,$3A,$BE,$47,$E1,$7E,$C3,$ED,$5F,$E4,$3E,$BE,$ED
	.dc.b	$F8,$EE,$24,$62,$F8,$1F,$8F,$BC,$71,$FE,$4E,$6D,$70,$AD,$10,$FC
	.dc.b	$FC,$68,$DD,$72,$FA,$D7,$C6,$DF,$1A,$5F,$F8,$4A,$72,$D9,$CB,$FD
	.dc.b	$16,$3F,$AD,$B9,$21,$99,$DE,$7F,$E9,$69,$E7,$3C,$6B,$39,$FE,$BA
	.dc.b	$79,$CE,$4F,$E7,$DC,$BF,$03,$BB,$F6,$D1,$FD,$07,$9F,$13,$23,$04
	.dc.b	$3F,$A2,$39,$7F,$A3,$47,$33,$18,$FF,$45,$FC,$13,$B7,$54,$AF,$E9
	.dc.b	$D1,$E1,$30,$FF,$4F,$8C,$3F,$D4,$FE,$5B,$CD,$F3,$52,$30,$43,$FA
	.dc.b	$B3,$97,$FA,$C4,$7B,$4C,$5F,$F5,$79,$C4,$F1,$8C,$3F,$D8,$63,$2F
	.dc.b	$F6,$1E,$3F,$8D,$F8,$1C,$8F,$EC,$D1,$F2,$4C,$3F,$D9,$E3,$0F,$F8
	.dc.b	$0F,$C0,$FC,$6F,$8C,$69,$8F,$0E,$CE,$C8,$7F,$81,$E5,$1C,$BC,$3F
	.dc.b	$CD,$78,$EF,$79,$F1,$98,$6E,$1F,$E1,$3A,$E0,$FF,$88,$E3,$2F,$F8
	.dc.b	$41,$BA,$FF,$83,$E7,$F4,$D2,$FC,$F9,$F2,$5F,$C8,$A3,$F9,$44,$61
	.dc.b	$C1,$B8,$AD,$33,$1B,$7F,$CD,$1E,$03,$E0,$FB,$B6,$3F,$CA,$33,$BC
	.dc.b	$3F,$CD,$3F,$02,$F9,$6E,$AD,$AF,$0B,$5F,$33,$5F,$E7,$2E,$67,$FC
	.dc.b	$E7,$F1,$1F,$A8,$6F,$6F,$C0,$9D,$76,$FF,$45,$FD,$1B,$E7,$61,$FE
	.dc.b	$64,$5F,$0F,$F1,$9B,$19,$E0,$C1,$99,$91,$92,$1F,$E8,$99,$E2,$D8
	.dc.b	$E2,$FF,$9C,$BE,$AB,$FD,$50,$FC,$E6,$BF,$4D,$AF,$F5,$66,$BF,$D7
	.dc.b	$5D,$E5,$FE,$92,$FA,$B7,$D2,$FD,$E6,$D7,$7F,$83,$FE,$47,$FA,$3F
	.dc.b	$F9,$BE,$3F,$D4,$C7,$F2,$6F,$84,$FE,$C7,$87,$F6,$7F,$29,$5B,$AF
	.dc.b	$F2,$F3,$67,$55,$C3,$6F,$F0,$0B,$69,$7C,$BD,$AF,$27,$DC,$F9,$79
	.dc.b	$25,$7F,$A5,$34,$4E,$A3,$13,$A7,$FD,$69,$3A,$78,$4E,$72,$F8,$55
	.dc.b	$C6,$7F,$5D,$EB,$7E,$DB,$D6,$91,$FE,$7B,$8F,$DC,$98,$7F,$69,$18
	.dc.b	$7F,$D4,$FF,$53,$FB,$67,$F4,$6D,$7F,$50,$D7,$FC,$16,$D7,$FC,$24
	.dc.b	$EE,$FF,$D8,$FE,$23,$EA,$3F,$66,$DA,$EF,$E0,$FF,$0C,$F2,$BF,$86
	.dc.b	$F9,$52,$7F,$D2,$51,$97,$FD,$29,$19,$7F,$86,$46,$5F,$E1,$D1,$9C
	.dc.b	$3F,$D2,$13,$FC,$63,$CA,$FE,$33,$E5,$4B,$FF,$4F,$46,$5F,$F5,$04
	.dc.b	$65,$FE,$31,$1F,$C8,$3C,$AF,$E4,$39,$49,$FF,$53,$46,$5F,$F5,$44
	.dc.b	$65,$FE,$41,$1F,$02,$BE,$0D,$C9,$FF,$6A,$D7,$F9,$8D,$B3,$97,$FD
	.dc.b	$E9,$1F,$E6,$57,$24,$33,$3B,$C3,$E7,$D7,$E8,$3E,$0E,$50,$FC,$DB
	.dc.b	$FC,$EB,$CA,$FD,$2F,$D4,$FE,$9B,$E9,$25,$6F,$0F,$D2,$75,$2B,$FA
	.dc.b	$0F,$9B,$2B,$FD,$8C,$F0,$7F,$9F,$7D,$E7,$F4,$62,$3F,$EC,$BE,$2A
	.dc.b	$FF,$43,$F4,$FF,$A5,$98,$C7,$FD,$8C,$E2,$3F,$DA,$4E,$7B,$7F,$CD
	.dc.b	$F9,$3E,$AF,$F7,$BE,$B4,$BF,$45,$6A,$7F,$80,$99,$8F,$DD,$CE,$1D
	.dc.b	$9F,$B8,$AE,$8E,$3B,$BF,$81,$CF,$1C,$7E,$1D,$FF,$62,$F7,$3F,$B1
	.dc.b	$FA,$D3,$0F,$F7,$44,$65,$FF,$75,$46,$5F,$EC,$51,$97,$FB,$24,$66
	.dc.b	$BF,$DC,$D3,$FC,$63,$C3,$FE,$33,$F1,$F2,$FC,$99,$FF,$1E,$7F,$8A
	.dc.b	$CE,$5F,$E2,$F3,$C5,$DC,$66,$FE,$DC,$E6,$7F,$DF,$5F,$69,$FC,$77
	.dc.b	$5F,$F4,$DF,$A9,$FC,$6F,$6F,$F7,$ED,$7F,$93,$2B,$4F,$F1,$E6,$C6
	.dc.b	$A3,$E4,$4F,$19,$C9,$A7,$A1,$9F,$13,$FC,$AF,$E0,$FF,$78,$C3,$FD
	.dc.b	$56,$BE,$73,$26,$BD,$49,$8F,$D5,$9D,$36,$72,$B4,$FE,$79,$3D,$63
	.dc.b	$F9,$AB,$C1,$D5,$F8,$2B,$8D,$7F,$43,$CB,$FB,$64,$FF,$AF,$9F,$F2
	.dc.b	$8C,$67,$D9,$F8,$9E,$7C,$C3,$F1,$15,$AB,$E1,$59,$9E,$E7,$F4,$6F
	.dc.b	$61,$D1,$FE,$59,$24,$7E,$46,$3F,$93,$3C,$37,$FF,$80,$33,$A8,$C4
	.dc.b	$F2,$78,$74,$7F,$98,$F9,$7F,$CC,$9D,$12,$7A,$46,$5E,$D1,$97,$9C
	.dc.b	$7F,$A2,$BC,$47,$CA,$7F,$B0,$B9,$7F,$DB,$F1,$8A,$FF,$69,$C6,$27
	.dc.b	$51,$FD,$8A,$7F,$C1,$EF,$FB,$3E,$74,$F0,$35,$3C,$33,$FF,$22,$B6
	.dc.b	$B8,$FC,$02,$F8,$DF,$33,$89,$CE,$F8,$B6,$B8,$CF,$FB,$DB,$3F,$F6
	.dc.b	$63,$7A,$7C,$10,$6D,$43,$A4,$65,$B1,$97,$C4,$3E,$83,$E3,$FC,$0F
	.dc.b	$F9,$29,$B1,$FE,$4F,$6B,$FE,$F2,$6B,$FE,$D6,$F6,$5F,$1D,$FE,$52
	.dc.b	$F8,$1F,$F2,$BB,$63,$33,$FE,$59,$9C,$BD,$7D,$7F,$33,$D5,$AF,$F4
	.dc.b	$13,$CC,$E1,$FE,$89,$68,$C6,$AF,$9F,$FC,$0F,$37,$CD,$7F,$CE,$CD
	.dc.b	$E1,$39,$F6,$2E,$3F,$56,$F9,$56,$38,$BE,$D4,$FF,$89,$B0,$E6,$ED
	.dc.b	$EE,$DB,$E1,$C3,$FD,$67,$07,$D0,$3C,$4C,$DF,$F8,$C2,$31,$9F,$E0
	.dc.b	$D7,$B9,$AB,$E3,$76,$77,$7E,$76,$E7,$E6,$AE,$F7,$BB,$98,$75,$79
	.dc.b	$46,$30,$F5,$7C,$BF,$C2,$EE,$FF,$26,$6B,$F3,$70,$FE,$98,$A7,$51
	.dc.b	$ED,$8D,$57,$EF,$DD,$47,$0D,$F7,$3F,$26,$AE,$5F,$5E,$6D,$FD,$4B
	.dc.b	$EE,$63,$E4,$72,$C3,$27,$91,$C6,$1F,$C3,$FD,$47,$CC,$76,$5F,$41
	.dc.b	$63,$0D,$BF,$ED,$13,$A6,$CE,$9F,$ED,$9F,$41,$FE,$BE,$6F,$89,$EA
	.dc.b	$8C,$61,$FC,$7D,$FE,$85,$3C,$3C,$6D,$56,$DB,$19,$39,$96,$B9,$4F
	.dc.b	$A7,$E8,$3D,$5F,$39,$6B,$CE,$DA,$F8,$0D,$E1,$FF,$42,$66,$8C,$57
	.dc.b	$E4,$E3,$0F,$C0,$BE,$AF,$B9,$F6,$4F,$B5,$B1,$93,$E1,$C6,$5A,$66
	.dc.b	$FE,$14,$65,$B1,$F3,$BE,$83,$EA,$7D,$07,$D5,$6C,$64,$FA,$71,$96
	.dc.b	$99,$9F,$BD,$19,$6C,$7D,$6F,$9B,$FA,$9F,$41,$F5,$5B,$19,$3E,$DC
	.dc.b	$65,$A6,$67,$ED,$46,$5B,$1F,$86,$8F,$C6,$1F,$8D,$F8,$2B,$1F,$8A
	.dc.b	$6C,$FE,$4D,$BA,$7D,$47,$84,$FE,$42,$ED,$7E,$67,$E8,$3F,$30,$7E
	.dc.b	$37,$E1,$ED,$7E,$4B,$E1,$B0,$D3,$F2,$B1,$F9,$3F,$89,$91,$F9,$C8
	.dc.b	$B7,$FD,$EB,$E7,$BF,$9B,$EF,$FB,$47,$F9,$CF,$2D,$F1,$BF,$0B,$5B
	.dc.b	$7F,$57,$87,$F9,$42,$B8,$B6,$33,$3B,$BE,$57,$CD,$B5,$B6,$CF,$ED
	.dc.b	$F2,$FD,$B6,$5F,$B7,$6D,$7E,$D1,$B3,$B8,$FD,$BC,$FE,$F7,$C3,$FD
	.dc.b	$F7,$87,$67,$F7,$6D,$AF,$E0,$8D,$E0,$FD,$2B,$FC,$17,$BB,$F8,$37
	.dc.b	$6D,$BF,$E0,$6D,$BF,$E1,$8D,$BF,$D6,$B6,$7F,$58,$D9,$FF,$75,$F3
	.dc.b	$FF,$24,$AF,$E9,$B0,$F9,$F7,$1E,$23,$FA,$F4,$63,$0F,$EC,$11,$89
	.dc.b	$D3,$FE,$A3,$9D,$3A,$3A,$47,$F2,$E6,$CF,$FA,$1B,$E7,$5F,$74,$FA
	.dc.b	$ED,$61,$CB,$1F,$E0,$F7,$FC,$9F,$C0,$FE,$50,$D3,$F9,$29,$71,$FE
	.dc.b	$5C,$D9,$FE,$D3,$CB,$FE,$02,$E8,$FC,$56,$9F,$8B,$6D,$7E,$1D,$B7
	.dc.b	$F8,$D6,$DF,$2E,$64,$62,$BF,$D8,$58,$75,$7F,$21,$87,$F4,$AF,$8B
	.dc.b	$7C,$BD,$F9,$9C,$36,$BF,$D4,$93,$A6,$CE,$9A,$7F,$AB,$33,$7D,$06
	.dc.b	$DF,$9C,$D7,$EA,$B5,$FF,$0B,$35,$FF,$6C,$3C,$74,$FC,$0B,$67,$FD
	.dc.b	$B9,$C3,$F5,$3D,$DF,$B1,$39,$B6,$9F,$E2,$CB,$FD,$A6,$FF,$B5,$69
	.dc.b	$FB,$12,$CF,$ED,$DB,$8F,$CB,$69,$F2,$B7,$FE,$DD,$3F,$BD,$AF,$C4
	.dc.b	$7C,$F3,$E1,$7F,$1F,$28,$C5,$FE,$3E,$31,$3A,$9C,$C7,$E4,$DB,$87
	.dc.b	$E8,$1B,$5F,$EF,$BF,$C2,$5F,$84,$BB,$3C,$38,$FE,$67,$E7,$DF,$03
	.dc.b	$E2,$6E,$1E,$16,$E7,$FA,$16,$BF,$A8,$73,$9F,$EC,$1B,$29,$FF,$34
	.dc.b	$7D,$E3,$5F,$06,$1F,$E1,$4F,$83,$7C,$FF,$F8,$63,$E6,$5F,$41,$C7
	.dc.b	$2F,$F6,$3C,$E6,$31,$97,$FA,$7E,$F8,$D7,$FB,$DA,$F8,$CF,$5B,$E0
	.dc.b	$F7,$E9,$95,$FE,$D2,$E8,$B3,$CE,$6E,$DF,$E5,$6F,$49,$92,$F3,$87
	.dc.b	$FA,$1D,$FF,$16,$FB,$BE,$2C,$7F,$D5,$F1,$8C,$B2,$DA,$FF,$5C,$36
	.dc.b	$7F,$D6,$BE,$F7,$2C,$F9,$B6,$BF,$D7,$CD,$9D,$C7,$FB,$06,$76,$D3
	.dc.b	$65,$8D,$9F,$EC,$28,$DB,$63,$FD,$95,$F3,$5E,$AF,$9B,$F5,$B6,$32
	.dc.b	$7F,$B3,$E3,$2D,$3F,$DA,$5F,$11,$F0,$BC,$5F,$86,$D8,$C9,$FE,$D7
	.dc.b	$8C,$B4,$CD,$FF,$B5,$A3,$2D,$8F,$C5,$6F,$F4,$FE,$81,$F2,$1F,$E5
	.dc.b	$C5,$F2,$36,$30,$DA,$E2,$7F,$A1,$67,$6E,$2D,$C4,$9E,$B3,$98,$C5
	.dc.b	$73,$AF,$F7,$75,$72,$6D,$F8,$2F,$84,$F2,$9D,$D7,$1C,$3D,$4F,$F9
	.dc.b	$D3,$E7,$B9,$47,$E3,$D7,$CB,$66,$74,$DC,$3C,$93,$A6,$CE,$9A,$6A
	.dc.b	$78,$5A,$FB,$47,$F5,$98,$FD,$B7,$BF,$FB,$7F,$9F,$7C,$8F,$F6,$4F
	.dc.b	$88,$FF,$5B,$C6,$2B,$FB,$3D,$6E,$B5,$8D,$9C,$8A,$E0,$FF,$03,$BE
	.dc.b	$75,$FD,$AE,$F8,$4E,$6E,$BE,$41,$F1,$0F,$A0,$E1,$18,$6C,$66,$3C
	.dc.b	$2D,$5F,$11,$FE,$BE,$D3,$F8,$A4,$E6,$31,$3C,$CF,$E3,$0E,$23,$5F
	.dc.b	$09,$F2,$67,$47,$47,$1A,$7B,$AB,$FD,$A1,$3E,$C6,$BF,$0F,$3F,$F7
	.dc.b	$74,$78,$27,$85,$72,$9C,$FC,$2F,$FB,$CE,$7A,$36,$38,$CE,$EF,$F3
	.dc.b	$3C,$0F,$8A,$FF,$7D,$67,$F8,$BF,$8B,$BB,$7F,$B9,$69,$18,$BE,$37
	.dc.b	$79,$55,$CE,$7F,$1F,$59,$8E,$31,$F9,$2A,$C1,$99,$91,$93,$F5,$5F
	.dc.b	$40,$F9,$1E,$CB,$E4,$29,$88,$93,$C2,$32,$F2,$8C,$BF,$9B,$CB,$FE
	.dc.b	$00,$57,$38,$E0,$6C,$86,$CF,$DF,$BB,$71,$92,$31,$9F,$F4,$17,$D1
	.dc.b	$3F,$C0,$FC,$1F,$E1,$ED,$E6,$F9,$11,$FB,$88,$D7,$3E,$0C,$CC,$8C
	.dc.b	$90,$F0,$CF,$16,$C7,$17,$C1,$F3,$3D,$BA,$A4,$60,$87,$27,$2F,$48
	.dc.b	$F2,$FC,$CF,$9B,$E6,$64,$FA,$E3,$2F,$B6,$32,$F9,$23,$2F,$9A,$33
	.dc.b	$9F,$F6,$5C,$55,$A8,$C4,$EB,$5F,$F0,$16,$AB,$DE,$1F,$2F,$93,$E4
	.dc.b	$FC,$CC,$8C,$10,$F8,$AE,$5F,$8F,$1F,$97,$F1,$9E,$0D,$24,$FF,$39
	.dc.b	$87,$89,$88,$FF,$04,$4D,$D3,$FB,$F5,$6A,$78,$78,$7F,$A5,$F1,$DE
	.dc.b	$FB,$E2,$F0,$DA,$D9,$0F,$C7,$4F,$09,$D9,$E7,$AF,$D8,$4B,$E8,$1F
	.dc.b	$27,$86,$C7,$F9,$95,$AF,$E6,$47,$77,$91,$D4,$FF,$91,$DE,$8D,$9E
	.dc.b	$31,$FE,$9E,$6B,$FE,$35,$1F,$02,$F9,$CF,$2B,$63,$FC,$EC,$D7,$C5
	.dc.b	$F1,$0F,$9D,$DD,$78,$6B,$55,$FE,$AC,$6B,$DD,$AF,$F9,$17,$E8,$3D
	.dc.b	$4F,$B5,$7C,$47,$C6,$6C,$7F,$A1,$DD,$F0,$FF,$AF,$9B,$D1,$EB,$77
	.dc.b	$3C,$EB,$97,$0C,$8E,$51,$D4,$C3,$F3,$7E,$83,$FD,$46,$FF,$A5,$5A
	.dc.b	$E7,$E1,$3E,$CF,$D0,$7F,$54,$8D,$47,$D1,$9D,$C7,$FB,$31,$AF,$CF
	.dc.b	$6B,$FE,$10,$F9,$CF,$F3,$8B,$FE,$A8,$3F,$99,$7C,$2F,$FB,$53,$4F
	.dc.b	$F2,$94,$EA,$7F,$8B,$B6,$3F,$CA,$D3,$BD,$3F,$5E,$DA,$ED,$EC,$7A
	.dc.b	$38,$3D,$3C,$12,$7F,$D6,$5A,$7D,$B6,$CE,$5F,$C8,$47,$E0,$71,$FF
	.dc.b	$76,$BB,$8C,$CE,$E7,$35,$DE,$7F,$65,$EA,$7E,$DF,$D4,$91,$82,$1F
	.dc.b	$E7,$B7,$2F,$F9,$F6,$3F,$4F,$F4,$1F,$B5,$8F,$D1,$36,$FF,$DF,$ED
	.dc.b	$D3,$F3,$9E,$17,$F5,$6A,$DF,$77,$E9,$6B,$94,$7F,$C0,$06,$BF,$A3
	.dc.b	$6B,$FE,$44,$ED,$FE,$7E,$FF,$A3,$FC,$6F,$E2,$07,$F1,$1F,$1E,$47
	.dc.b	$F0,$F8,$CB,$FC,$46,$32,$F1,$34,$64,$FE,$2D,$E3,$7F,$1B,$F1,$E4
	.dc.b	$60,$87,$FA,$7D,$CB,$FE,$A1,$8F,$E4,$5E,$3F,$F2,$3C,$E4,$FF,$21
	.dc.b	$8C,$BF,$C8,$E3,$2F,$FA,$A6,$3F,$89,$C7,$F9,$B2,$3F,$DE,$8F,$C0
	.dc.b	$D5,$AF,$CC,$C7,$78,$CB,$DC,$72,$30,$FF,$17,$BE,$8D,$8D,$3D,$A7
	.dc.b	$FC,$21,$CB,$FE,$E3,$8F,$F3,$83,$5F,$CF,$FB,$BF,$CF,$FC,$6D,$FC
	.dc.b	$FF,$0F,$A0,$95,$C2,$3F,$A2,$79,$12,$3F,$A1,$3A,$7F,$D9,$1F,$3D
	.dc.b	$FD,$24,$8F,$F4,$7F,$9E,$AF,$FB,$37,$D0,$FE,$9C,$63,$83,$FA,$3B
	.dc.b	$88,$FE,$9C,$E7,$83,$FA,$FE,$AF,$A0,$F6,$FA,$92,$FD,$B1,$F0,$0C
	.dc.b	$3F,$BC,$8C,$3A,$3E,$1B,$AC,$3F,$81,$CF,$F6,$1F,$6B,$FB,$17,$A9
	.dc.b	$2F,$FB,$04,$65,$FE,$C5,$19,$7F,$DD,$11,$97,$FD,$D5,$19,$AF,$EC
	.dc.b	$73,$F5,$3E,$2B,$EA,$FC,$F4,$BF,$AB,$19,$7E,$BC,$65,$FE,$2B,$19
	.dc.b	$7F,$8B,$C6,$6B,$F8,$DC,$FF,$93,$FE,$83,$F8,$D6,$9F,$E0,$6E,$BF
	.dc.b	$E3,$1E,$A7,$FB,$C3,$1D,$78,$91,$FE,$F8,$7F,$80,$4C,$1F,$E3,$AB
	.dc.b	$E6,$FE,$E4,$D1,$FE,$BB,$73,$2F,$03,$FB,$D4,$FF,$90,$97,$CD,$FF
	.dc.b	$98,$DB,$1F,$EC,$37,$79,$FE,$92,$3F,$D9,$31,$8C,$BF,$C3,$EF,$08
	.dc.b	$E9,$5C,$FD,$29,$1C,$E3,$B1,$87,$D5,$EB,$FC,$5F,$62,$47,$A6,$3E
	.dc.b	$41,$87,$E0,$46,$1F,$F2,$B7,$3F,$F9,$50,$8F,$F9,$67,$9E,$BF,$E5
	.dc.b	$08,$B9,$FF,$8E,$E7,$FC,$B7,$C7,$FE,$5C,$E7,$97,$FE,$61,$8C,$BF
	.dc.b	$E6,$38,$CB,$FE,$5A,$8F,$F4,$57,$88,$F9,$4F,$F0,$77,$BE,$E6,$FF
	.dc.b	$21,$CE,$6F,$FD,$A3,$18,$AD,$BA,$F3,$EC,$FF,$BD,$AB,$82,$BE,$8E
	.dc.b	$6F,$7C,$7F,$E4,$06,$D7,$F8,$3A,$B7,$39,$AF,$09,$C4,$AE,$5E,$D8
	.dc.b	$78,$DC,$FB,$3F,$EE,$28,$FF,$2E,$AE,$9F,$0B,$67,$66,$0C,$C4,$72
	.dc.b	$78,$87,$D0,$7C,$9F,$7F,$FC,$6D,$1F,$E3,$78,$C3,$83,$79,$FF,$84
	.dc.b	$DD,$C6,$9E,$3E,$FF,$F9,$3B,$E3,$78,$47,$FB,$93,$D9,$46,$27,$53
	.dc.b	$D5,$AF,$4F,$5F,$D8,$DC,$B9,$E1,$FE,$4A,$6C,$EA,$7F,$D0,$3C,$FF
	.dc.b	$EB,$FE,$BF,$9A,$D8,$F9,$A3,$7F,$F6,$37,$6F,$EB,$78,$7E,$BD,$6D
	.dc.b	$BD,$39,$E2,$EA,$DC,$39,$52,$F1,$87,$FB,$46,$AF,$D5,$2E,$E9,$86
	.dc.b	$E7,$F8,$29,$D3,$67,$4D,$35,$3C,$0F,$5D,$AB,$8C,$61,$FC,$34,$FC
	.dc.b	$2F,$47,$F8,$57,$57,$F0,$BB,$FF,$73,$46,$27,$E4,$E7,$70,$F6,$CD
	.dc.b	$AD,$D7,$AE,$F9,$5F,$F9,$EE,$30,$72,$2D,$F2,$6D,$7F,$BD,$7E,$DF
	.dc.b	$A5,$1B,$E9,$E1,$95,$C2,$74,$D9,$CD,$FF,$33,$F2,$7C,$DE,$11,$BE
	.dc.b	$4E,$19,$7F,$CD,$5C,$B4,$F2,$FE,$70,$F9,$AF,$80,$D8,$CF,$D0,$E2
	.dc.b	$64,$70,$77,$18,$C7,$FD,$FF,$5C,$5B,$3A,$AE,$07,$F4,$7F,$7F,$D2
	.dc.b	$7A,$BA,$EC,$64,$95,$FD,$46,$32,$FF,$53,$9B,$C1,$F7,$74,$FF,$5F
	.dc.b	$DF,$08,$FC,$06,$2F,$9F,$F6,$9E,$EF,$9F,$B5,$F0,$A2,$E3,$EA,$9B
	.dc.b	$7C,$A6,$D7,$FB,$C2,$D5,$8B,$FD,$76,$1F,$E1,$5A,$FE,$C9,$E6,$7D
	.dc.b	$6F,$37,$EB,$B2,$32,$7F,$69,$8C,$57,$F6,$F9,$B9,$7B,$66,$D7,$87
	.dc.b	$E7,$BE,$B9,$F6,$3D,$EB,$1F,$72,$2E,$1F,$52,$6D,$FC,$49,$B5,$EB
	.dc.b	$9B,$5E,$AF,$A6,$FC,$6F,$CE,$7E,$39,$B5,$FD,$FD,$C3,$FE,$01,$F4
	.dc.b	$6D,$7F,$81,$DF,$CA,$9F,$96,$EF,$B3,$F9,$0F,$81,$FC,$CF,$C5,$FE
	.dc.b	$41,$B1,$F9,$4F,$2F,$F2,$67,$E5,$3C,$4B,$1F,$98,$F2,$3F,$2E,$7E
	.dc.b	$8B,$C9,$B1,$FB,$2F,$98,$FE,$63,$CB,$FB,$D6,$C6,$7C,$A7,$D0,$E5
	.dc.b	$FB,$71,$9A,$C1,$99,$B1,$92,$6E,$E4,$FF,$0C,$61,$FD,$02,$B7,$39
	.dc.b	$AF,$E4,$9E,$1B,$F4,$87,$A9,$FC,$3D,$FF,$12,$CF,$F0,$87,$2D,$F5
	.dc.b	$35,$3B,$6C,$64,$FE,$23,$6F,$9E,$7C,$F7,$F1,$DF,$2D,$F5,$3F,$E0
	.dc.b	$7B,$DD,$69,$FE,$33,$3A,$E7,$FE,$49,$87,$3C,$3F,$C6,$8F,$F6,$8B
	.dc.b	$E0,$FF,$67,$6B,$98,$DB,$F0,$33,$FE,$09,$BF,$F3,$35,$E2,$FC,$1F
	.dc.b	$17,$E1,$37,$3F,$84,$DB,$F9,$47,$F5,$BA,$E3,$3F,$1E,$B8,$B4,$E2
	.dc.b	$7F,$34,$F9,$F7,$E8,$9F,$CB,$2F,$FC,$B9,$E6,$7F,$41,$9C,$BF,$CB
	.dc.b	$0C,$4E,$F2,$FF,$2A,$4E,$AF,$EB,$7D,$1E,$AF,$ED,$73,$7D,$BE,$5B
	.dc.b	$5F,$6D,$B9,$7A,$B9,$7F,$05,$18,$CB,$E2,$3F,$93,$F8,$D9,$7F,$97
	.dc.b	$F0,$EE,$1F,$0B,$A3,$FC,$1B,$58,$C3,$F2,$3D,$2B,$C6,$1F,$DC,$EF
	.dc.b	$6F,$E8,$6F,$F4,$13,$AC,$3F,$BC,$7D,$0B,$E5,$FF,$B2,$64,$FA,$AF
	.dc.b	$9F,$5F,$48,$56,$F1,$FE,$C3,$5B,$E5,$FC,$4C,$E6,$BF,$C5,$8D,$7F
	.dc.b	$C6,$2D,$7F,$C1,$07,$4C,$3F,$12,$F9,$3E,$1B,$F3,$5D,$6E,$EE,$B1
	.dc.b	$86,$E5,$FA,$2C,$13,$FA,$1F,$9D,$FD,$6E,$1E,$99,$CD,$7F,$07,$87
	.dc.b	$C9,$FF,$07,$6C,$67,$E5,$71,$32,$38,$3B,$8C,$4F,$D5,$38,$96,$35
	.dc.b	$3C,$0F,$DB,$F8,$FF,$BE,$9F,$CE,$F8,$0C,$7F,$AC,$56,$BC,$2F,$57
	.dc.b	$9F,$FB,$4C,$38,$70,$FE,$D1,$B5,$FE,$31,$C3,$97,$D2,$EF,$0F,$DD
	.dc.b	$FE,$89,$D3,$3E,$FF,$A3,$79,$B2,$FD,$EC,$E5,$FF,$59,$60,$9D,$63
	.dc.b	$F5,$AB,$4D,$CF,$F8,$74,$BE,$4B,$BC,$7F,$1D,$CE,$57,$F1,$D8,$B9
	.dc.b	$7E,$8F,$C3,$56,$33,$FE,$0D,$56,$FF,$85,$79,$FF,$C9,$CF,$E5,$1E
	.dc.b	$2D,$8F,$E5,$11,$71,$E9,$36,$FF,$C7,$3F,$7D,$FE,$32,$3F,$C6,$7F
	.dc.b	$7D,$70,$FF,$18,$45,$BF,$DB,$7B,$6A,$C5,$FF,$90,$BD,$AF,$F1,$D9
	.dc.b	$FE,$3C,$F8,$EB,$8F,$F8,$E6,$2E,$1C,$A6,$D7,$1A,$D1,$FE,$3F,$8C
	.dc.b	$BF,$E7,$CA,$B5,$FE,$70,$9B,$E2,$31,$FF,$0F,$CF,$9B,$DA,$FF,$26
	.dc.b	$FB,$7F,$E4,$E6,$63,$FE,$4B,$72,$D3,$FC,$9D,$EB,$FF,$95,$0F,$F2
	.dc.b	$AF,$B7,$63,$24,$BF,$87,$19,$7E,$2C,$DD,$3F,$CC,$B3,$6B,$FA,$FF
	.dc.b	$E4,$0A,$FF,$36,$36,$3F,$CC,$AA,$73,$32,$72,$F0,$8C,$5F,$F2,$9A
	.dc.b	$E4,$D9,$D5,$70,$24,$65,$EF,$5C,$AF,$FD,$09,$6F,$67,$CA,$79,$BD
	.dc.b	$AB,$7E,$68,$B8,$78,$66,$DF,$29,$B5,$F0,$26,$D7,$FB,$3B,$E0,$BF
	.dc.b	$D4,$9F,$3C,$F8,$FF,$F6,$8F,$6B,$0C,$C7,$FB,$46,$AD,$FF,$24,$AD
	.dc.b	$2B,$E4,$F5,$31,$5D,$2F,$8C,$7F,$2C,$BA,$FF,$A9,$8D,$9C,$7C,$57
	.dc.b	$B9,$E1,$8F,$E5,$CD,$8C,$C6,$E3,$F9,$BF,$C3,$7D,$9D,$7F,$D8,$15
	.dc.b	$AA,$4E,$5E,$67,$FB,$15,$C4,$6D,$B5,$D3,$7F,$F1,$65,$79,$8E,$7C
	.dc.b	$1F,$D1,$B4,$F8,$31,$83,$DF,$7D,$E3,$2D,$D3,$CF,$5C,$5B,$5C,$2F
	.dc.b	$BD,$FC,$CE,$47,$0B,$D8,$FE,$AB,$AF,$E2,$1B,$BF,$E7,$9B,$8F,$FB
	.dc.b	$BE,$9E,$F6,$30,$E7,$56,$7A,$57,$E8,$A3,$15,$96,$46,$4F,$D3,$C7
	.dc.b	$1D,$A4,$7E,$A9,$E9,$E2,$62,$3F,$4D,$17,$2F,$CF,$CD,$CB,$F6,$B3
	.dc.b	$6B,$D1,$37,$2F,$83,$37,$24,$FA,$F9,$FF,$51,$7F,$ED,$7E,$5C,$19
	.dc.b	$E5,$B1,$92,$1F,$CB,$1C,$BF,$CB,$A2,$E7,$FE,$EE,$9E,$DA,$CA,$FF
	.dc.b	$72,$3E,$1D,$71,$1F,$CD,$A2,$D7,$FB,$DA,$7C,$D9,$CA,$FF,$76,$3E
	.dc.b	$9C,$B1,$1F,$CF,$A2,$D7,$FB,$FA,$6E,$BF,$E7,$8C,$FF,$39,$3F,$B2
	.dc.b	$D7,$F9,$95,$F3,$C3,$A9,$0E,$87,$FB,$EF,$0F,$C1,$7D,$8B,$DB,$A4
	.dc.b	$9F,$F8,$01,$CB,$5F,$F2,$34,$ED,$18,$9F,$A5,$3F,$8D,$F0,$9E,$8D
	.dc.b	$25,$7F,$A2,$FD,$1A,$FF,$6F,$8B,$A7,$F9,$CE,$79,$7B,$F2,$BF,$D2
	.dc.b	$4F,$5F,$A6,$C4,$7F,$77,$8B,$5F,$E7,$B9,$FD,$0C,$BE,$3B,$FC,$56
	.dc.b	$FF,$A5,$1A,$FF,$92,$5B,$3F,$E8,$CF,$D2,$7F,$C7,$93,$BF,$3B,$9C
	.dc.b	$7E,$42,$3F,$D3,$AD,$79,$47,$83,$F4,$B7,$CC,$52,$4F,$FA,$93,$F4
	.dc.b	$B7,$97,$88,$FF,$2D,$47,$FB,$09,$D7,$B7,$9C,$BF,$D3,$FF,$80,$3E
	.dc.b	$63,$DA,$DA,$FF,$57,$B5,$FF,$32,$3F,$E5,$07,$31,$C0,$8F,$0F,$5E
	.dc.b	$BC,$22,$D7,$82,$7E,$5F,$CF,$BE,$12,$92,$FF,$D7,$FF,$41,$FE,$7F
	.dc.b	$7F,$D8,$33,$B7,$3E,$8F,$D4,$9F,$F6,$1C,$6A,$BA,$FA,$5F,$E7,$08
	.dc.b	$96,$8C,$47,$F9,$D2,$30,$E0,$38,$76,$DC,$7F,$D8,$FF,$3F,$F5,$DB
	.dc.b	$DB,$E0,$AF,$F6,$BC,$FF,$80,$5B,$E4,$7E,$2A,$31,$1F,$E8,$F5,$E7
	.dc.b	$66,$64,$F0,$39,$C6,$1E,$11,$6F,$C7,$3F,$A7,$F8,$8F,$D4,$7C,$54
	.dc.b	$BF,$F7,$84,$65,$FF,$79,$46,$5F,$F2,$B4,$7F,$3E,$F8,$77,$E9,$9E
	.dc.b	$FF,$1F,$F6,$B1,$8D,$B5,$39,$BF,$F6,$C7,$C5,$E2,$7F,$A2,$7C,$67
	.dc.b	$F9,$36,$7F,$7B,$A7,$F0,$D8,$C4,$7F,$B8,$30,$E8,$EB,$1F,$F7,$67
	.dc.b	$D5,$7F,$01,$FB,$09,$39,$3F,$DC,$D1,$FC,$14,$CB,$FE,$6F,$8C,$3B
	.dc.b	$F2,$9E,$32,$B7,$97,$FB,$DE,$F7,$7F,$EC,$1F,$A2,$7C,$C7,$E0,$E1
	.dc.b	$18,$7F,$0C,$D7,$F1,$6D,$7F,$20,$E6,$BF,$C3,$B0,$FA,$07,$CE,$7F
	.dc.b	$0C,$6C,$7B,$4E,$F7,$FD,$B7,$E3,$9F,$55,$FC,$8F,$69,$5E,$D9,$CB
	.dc.b	$67,$2F,$FA,$46,$79,$47,$7F,$A1,$93,$DF,$04,$78,$FE,$8B,$33,$FE
	.dc.b	$E3,$9C,$BB,$9D,$34,$86,$DC,$AC,$FF,$A6,$7E,$74,$F8,$4F,$F0,$43
	.dc.b	$67,$EB,$35,$E1,$97,$F4,$4E,$26,$8F,$AB,$E3,$F0,$52,$F8,$BD,$3E
	.dc.b	$0F,$AF,$C2,$7C,$9F,$8F,$95,$DA,$3E,$67,$C6,$C8,$F9,$4F,$CF,$E3
	.dc.b	$FA,$1F,$13,$27,$E7,$3B,$7C,$A3,$4F,$2C,$7D,$7F,$46,$4F,$03,$DB
	.dc.b	$EF,$FC,$02,$3F,$03,$E1,$2B,$D2,$DC,$2A,$CF,$1F,$CA,$BB,$C5,$F1
	.dc.b	$6B,$58,$76,$3A,$1F,$85,$FA,$5F,$C5,$11,$FC,$47,$D1,$D7,$C1,$C8
	.dc.b	$9C,$63,$FE,$03,$85,$63,$0F,$A9,$1B,$BE,$C7,$43,$ED,$F9,$3F,$78
	.dc.b	$8F,$DE,$F8,$2A,$FB,$F2,$FE,$55,$7E,$BA,$FC,$2D,$6B,$1F,$C4,$C5
	.dc.b	$9F,$81,$3F,$A5,$F4,$26,$1F,$A3,$7F,$51,$E8,$62,$3D,$B1,$70,$F6
	.dc.b	$4D,$BF,$5C,$DA,$FB,$99,$7F,$5F,$7E,$A7,$E7,$2F,$8F,$FF,$30,$36
	.dc.b	$3F,$87,$35,$DF,$C0,$6A,$64,$69,$E3,$18,$8E,$B1,$72,$FF,$19,$57
	.dc.b	$F8,$96,$7E,$AF,$77,$F0,$32,$3F,$C0,$7D,$8A,$E8,$FE,$09,$1F,$C2
	.dc.b	$63,$FC,$6B,$F2,$72,$7F,$92,$7C,$12,$3F,$C7,$38,$E6,$7F,$C6,$B1
	.dc.b	$71,$FF,$22,$57,$B5,$BF,$31,$FE,$40,$25,$74,$F5,$6B,$FE,$3E,$8B
	.dc.b	$3C,$A7,$FD,$11,$F9,$37,$F8,$F2,$BF,$C8,$3F,$1B,$FE,$88,$EA,$4F
	.dc.b	$F8,$06,$53,$98,$C3,$7E,$13,$FD,$85,$CB,$FC,$1E,$7A,$57,$08,$C7
	.dc.b	$C2,$6A,$3F,$71,$3C,$A7,$85,$67,$0F,$0B,$67,$55,$C2,$F1,$3B,$D3
	.dc.b	$F8,$AC,$FA,$B0,$FE,$31,$18,$8F,$11,$FC,$1D,$C6,$3E,$F7,$FC,$6D
	.dc.b	$1A,$C5,$1A,$AE,$7C,$7D,$CF,$07,$BB,$EB,$B7,$D1,$3D,$6F,$F2,$7C
	.dc.b	$7F,$88,$63,$1F,$0D,$FE,$1D,$6C,$7F,$88,$9A,$FF,$89,$FE,$3B,$FC
	.dc.b	$27,$F3,$3F,$E2,$36,$C7,$F8,$A1,$AE,$FD,$B4,$F5,$F6,$5F,$21,$49
	.dc.b	$5D,$E3,$2F,$F8,$D7,$77,$C5,$BE,$85,$E0,$7C,$BA,$FE,$5D,$39,$8C
	.dc.b	$57,$0F,$A2,$DD,$7C,$9A,$D3,$6F,$3C,$9F,$8D,$9C,$B6,$B3,$C7,$FE
	.dc.b	$A3,$AB,$7F,$CF,$7E,$8B,$FD,$2B,$97,$81,$F6,$B6,$37,$92,$73,$1F
	.dc.b	$6D,$E2,$DD,$37,$18,$CB,$F4,$0F,$06,$CE,$A7,$6D,$BF,$83,$9B,$E8
	.dc.b	$BE,$0F,$5F,$FA,$E6,$31,$3F,$E2,$17,$E6,$B6,$76,$FF,$B0,$5B,$5F
	.dc.b	$39,$B9,$7D,$4A,$A7,$49,$FE,$AC,$BE,$7B,$26,$BE,$8A,$E1,$E1,$8B
	.dc.b	$5C,$26,$E3,$FE,$43,$F5,$3B,$FC,$01,$BE,$9E,$19,$8F,$F8,$EE,$74
	.dc.b	$D9,$CE,$1F,$DA,$7E,$8B,$FD,$19,$3F,$E4,$16,$C6,$7C,$3F,$6C,$63
	.dc.b	$1F,$F7,$03,$6B,$FC,$20,$DA,$FE,$FC,$FF,$27,$8E,$9E,$17,$22,$3D
	.dc.b	$EE,$D7,$F7,$BC,$5F,$0B,$EE,$3D,$FD,$36,$FD,$91,$6B,$84,$DC,$BF
	.dc.b	$CD,$13,$6B,$FC,$49,$7F,$E8,$87,$89,$EF,$F8,$CF,$69,$EE,$F0,$2C
	.dc.b	$64,$98,$FC,$78,$CB,$F2,$66,$E1,$F1,$26,$D7,$83,$D2,$FA,$E7,$D8
	.dc.b	$F4,$AC,$7D,$58,$B8,$7C,$19,$B7,$FE,$81,$9B,$5E,$A9,$B5,$F3,$BC
	.dc.b	$5F,$AA,$7D,$6F,$84,$B5,$F6,$E2,$CF,$DF,$9B,$87,$DA,$9B,$5F,$6A
	.dc.b	$6D,$7F,$8D,$F3,$7C,$9F,$E4,$3E,$53,$F2,$2D,$BF,$C7,$38,$7F,$21
	.dc.b	$E9,$5B,$FB,$8F,$E6,$0F,$CC,$F7,$D9,$FC,$87,$C9,$FE,$73,$E5,$3F
	.dc.b	$1E,$D8,$FC,$A7,$C5,$7E,$48,$FC,$9F,$89,$63,$F3,$1E,$4F,$E5,$8F
	.dc.b	$D2,$F9,$36,$3F,$D4,$5F,$41,$FC,$C3,$97,$F7,$2D,$8C,$F9,$0F,$2F
	.dc.b	$2F,$F9,$AA,$33,$58,$33,$36,$32,$4D,$FF,$CB,$F7,$FC,$FA,$75,$1F
	.dc.b	$E5,$7F,$89,$FF,$12,$57,$FA,$BD,$7A,$BF,$C2,$9C,$CE,$E3,$1E,$96
	.dc.b	$5B,$5C,$5F,$E2,$72,$FA,$27,$C8,$7F,$AA,$A0,$48,$CD,$7F,$BE,$5B
	.dc.b	$1F,$EF,$E6,$BC,$0E,$13,$DA,$37,$D5,$FD,$2B,$E3,$1F,$66,$E9,$D5
	.dc.b	$FA,$FC,$3F,$67,$87,$73,$F8,$0D,$CF,$E2,$37,$2F,$EA,$BE,$57,$FB
	.dc.b	$4A,$31,$5F,$E9,$DD,$FF,$86,$5F,$CB,$C3,$AC,$FF,$B7,$F4,$7D,$1F
	.dc.b	$F2,$F8,$C3,$6B,$E9,$7A,$FF,$CE,$AB,$7E,$5D,$AD,$69,$ED,$6E,$1E
	.dc.b	$1B,$D5,$7D,$7E,$5F,$B1,$C9,$6B,$EE,$37,$2F,$C0,$37,$1F,$6E,$9F
	.dc.b	$8F,$AF,$C8,$FA,$DF,$90,$3F,$23,$E1,$D8,$C9,$30,$FC,$6F,$42,$75
	.dc.b	$18,$AB,$87,$F7,$EE,$8F,$F0,$7D,$63,$0F,$B9,$F4,$2F,$85,$FC,$36
	.dc.b	$3F,$86,$8C,$4F,$E0,$63,$73,$A5,$3A,$74,$62,$3F,$27,$19,$C3,$F0
	.dc.b	$DC,$EF,$0B,$35,$AF,$7A,$63,$F9,$6B,$FF,$0A,$FA,$4F,$A1,$FD,$2C
	.dc.b	$FF,$BF,$F2,$5F,$0C,$FF,$D9,$A7,$91,$FE,$EB,$7C,$7B,$75,$8F,$C3
	.dc.b	$A7,$C1,$FF,$02,$6E,$3F,$C1,$97,$C3,$66,$64,$E5,$E1,$18,$9F,$DF
	.dc.b	$9C,$8B,$1A,$9E,$07,$B4,$FD,$66,$BF,$BE,$BF,$F1,$9D,$7C,$5F,$03
	.dc.b	$D3,$AB,$E7,$FF,$7B,$7B,$C3,$8D,$60,$DC,$D8,$ED,$7F,$C0,$3E,$85
	.dc.b	$E7,$FF,$06,$6C,$EB,$67,$D0,$BE,$23,$57,$2B,$9E,$0C,$38,$BF,$EF
	.dc.b	$CC,$38,$DE,$E3,$F8,$F6,$3F,$C6,$CF,$E3,$98,$5C,$3F,$90,$45,$9F
	.dc.b	$E4,$93,$75,$FD,$2F,$9F,$FC,$8C,$FE,$49,$E2,$5C,$3F,$94,$45,$BF
	.dc.b	$E5,$93,$6B,$FC,$79,$F1,$9F,$E3,$23,$FC,$67,$F4,$37,$0F,$F1,$F4
	.dc.b	$5B,$FF,$1A,$CD,$AF,$F2,$27,$AF,$FE,$3A,$3F,$C7,$7F,$19,$70,$FF
	.dc.b	$23,$45,$BF,$F1,$EC,$DA,$E3,$5A,$3F,$C9,$91,$97,$F9,$9F,$91,$FE
	.dc.b	$7A,$BC,$5F,$F9,$E2,$AC,$FF,$9F,$BD,$7F,$F2,$59,$FE,$4C,$F6,$6E
	.dc.b	$1F,$E5,$78,$B7,$FE,$4E,$9B,$8F,$F9,$77,$D6,$FF,$29,$9F,$E5,$4F
	.dc.b	$62,$E1,$FE,$5F,$8B,$7F,$E5,$69,$B5,$C6,$B4,$7F,$99,$63,$2F,$AB
	.dc.b	$E7,$D7,$FE,$62,$6C,$7E,$4B,$24,$7F,$8F,$23,$0E,$B0,$61,$FD,$02
	.dc.b	$30,$DB,$E9,$87,$56,$5F,$27,$2D,$25,$F4,$7C,$D1,$A8,$C0,$C3,$CB
	.dc.b	$1E,$9F,$3A,$57,$9B,$D9,$B7,$C2,$FD,$75,$8C,$7D,$15,$6F,$E2,$4D
	.dc.b	$BE,$D9,$2B,$11,$FD,$B7,$DE,$FE,$EB,$F4,$0F,$A1,$FE,$61,$D3,$FE
	.dc.b	$B6,$8C,$57,$F3,$3A,$DD,$6B,$2B,$39,$15,$C1,$FF,$00,$DF,$3A,$FE
	.dc.b	$6F,$7C,$27,$37,$5F,$F0,$3C,$7F,$82,$3D,$47,$B3,$C2,$51,$87,$C0
	.dc.b	$D7,$D8,$6A,$3F,$C1,$5F,$09,$FE,$7C,$DB,$A4,$63,$87,$F0,$35,$B3
	.dc.b	$FC,$2F,$5B,$70,$1C,$5F,$77,$5A,$7B,$DF,$C9,$CF,$B5,$AF,$C4,$CF
	.dc.b	$C1,$8E,$70,$F0,$DF,$1E,$1F,$F1,$03,$6F,$55,$F3,$B8,$7F,$37,$ED
	.dc.b	$FF,$89,$63,$F3,$0D,$DB,$F4,$8D,$AF,$C0,$FB,$96,$F8,$C7,$E8,$A7
	.dc.b	$1C,$46,$D6,$19,$3B,$3F,$53,$1C,$37,$91,$FA,$C7,$9F,$85,$88,$FC
	.dc.b	$C5,$FD,$5C,$BF,$5D,$5A,$9E,$04,$36,$E5,$A6,$63,$8B,$FB,$AA,$F6
	.dc.b	$7C,$07,$F9,$06,$3F,$DD,$5C,$B8,$33,$CB,$63,$24,$3F,$DC,$AE,$5F
	.dc.b	$F7,$3C,$5B,$FF,$7B,$4F,$6C,$E5,$7F,$BD,$DF,$0E,$78,$8F,$F7,$7C
	.dc.b	$5A,$FF,$7F,$4F,$9B,$29,$5F,$EF,$F7,$D3,$96,$23,$FD,$EF,$16,$BF
	.dc.b	$CA,$79,$7F,$86,$67,$81,$0D,$9F,$C7,$DD,$38,$24,$72,$9E,$F1,$FC
	.dc.b	$97,$E1,$BF,$D0,$2D,$F2,$3F,$0D,$18,$8F,$E2,$17,$FB,$07,$44,$3F
	.dc.b	$8A,$46,$EF,$CF,$CB,$F8,$FF,$45,$E8,$52,$5F,$FA,$47,$CF,$AF,$FA
	.dc.b	$56,$2D,$FF,$95,$A7,$97,$BD,$2B,$FD,$30,$F5,$FA,$1C,$47,$FA,$76
	.dc.b	$2D,$7F,$97,$A7,$F4,$52,$FA,$1F,$F1,$8B,$FE,$0C,$6B,$FE,$3A,$6C
	.dc.b	$FF,$89,$3E,$87,$E0,$4E,$7C,$EE,$75,$AE,$8F,$F4,$B3,$67,$FD,$3F
	.dc.b	$5C,$2F,$FC,$5B,$F4,$2F,$8C,$F1,$B6,$BF,$C3,$AD,$7F,$C9,$7F,$43
	.dc.b	$FE,$28,$73,$3C,$08,$FF,$89,$23,$4F,$F9,$4B,$E8,$5F,$23,$F0,$1B
	.dc.b	$1F,$E2,$A6,$BB,$F5,$75,$32,$34,$F1,$8C,$47,$58,$B8,$FF,$92,$2B
	.dc.b	$FC,$C7,$3F,$E3,$6F,$A0,$7C,$17,$D0,$6C,$66,$3E,$9F,$D0,$FF,$A2
	.dc.b	$63,$04,$3F,$CD,$73,$C1,$FF,$36,$C6,$DC,$FA,$1D,$48,$FF,$91,$BE
	.dc.b	$87,$E8,$70,$FF,$B2,$6D,$5F,$E8,$6A,$FF,$69,$C6,$1C,$35,$FF,$45
	.dc.b	$B5,$FF,$49,$53,$C6,$FC,$5B,$63,$C1,$18,$F3,$32,$C8,$C9,$E5,$8C
	.dc.b	$3F,$E8,$89,$E4,$D9,$E4,$FF,$BD,$7E,$1F,$F5,$5F,$13,$23,$04,$3F
	.dc.b	$DF,$8E,$5F,$F7,$F4,$7F,$41,$F8,$67,$D0,$FB,$78,$3F,$DD,$0D,$D7
	.dc.b	$FD,$EB,$3A,$D7,$F7,$1F,$15,$C6,$B5,$8A,$B4,$FF,$B7,$E3,$3E,$53
	.dc.b	$BD,$7F,$C0,$7E,$87,$F8,$17,$D0,$CC,$3F,$C6,$94,$9D,$46,$27,$4F
	.dc.b	$F8,$7E,$74,$F0,$9C,$E9,$FE,$DA,$AE,$33,$FE,$B2,$FA,$17,$CE,$7F
	.dc.b	$00,$6C,$7F,$A2,$9A,$FF,$AE,$3E,$61,$F4,$DB,$AF,$C3,$4A,$B6,$FF
	.dc.b	$B4,$DA,$88,$E1,$FF,$1E,$7C,$5B,$E8,$7F,$91,$71,$49,$FF,$2A,$4E
	.dc.b	$5B,$39,$7E,$04,$FF,$77,$9D,$47,$9F,$E8,$3F,$8D,$E4,$69,$8F,$A1
	.dc.b	$B3,$FE,$A8,$9D,$36,$74,$D2,$1D,$4E,$13,$FE,$07,$F9,$E7,$C0,$52
	.dc.b	$4F,$AB,$E8,$75,$EF,$48,$FF,$38,$52,$39,$7B,$FC,$FE,$03,$D9,$F1
	.dc.b	$92,$F7,$1E,$FF,$8B,$91,$F1,$9F,$85,$C5,$F0,$FE,$1E,$4F,$CA,$76
	.dc.b	$FF,$99,$78,$3F,$D9,$4F,$CE,$F3,$BE,$51,$C1,$FF,$35,$69,$FE,$63
	.dc.b	$7C,$FE,$DF,$BC,$F8,$1E,$FD,$8C,$90,$E2,$E5,$E7,$17,$A5,$CB,$FE
	.dc.b	$AE,$BF,$F2,$0C,$62,$BF,$0B,$F4,$3F,$74,$8F,$E1,$3E,$86,$BE,$B8
	.dc.b	$B3,$C7,$BD,$A2,$F3,$8F,$FA,$F2,$17,$8C,$3F,$27,$5F,$6F,$DC,$91
	.dc.b	$F5,$9F,$BD,$EF,$62,$3E,$2C,$5C,$BC,$19,$FF,$91,$67,$81,$0D,$9F
	.dc.b	$80,$AF,$CC,$79,$D2,$3F,$3E,$FE,$6F,$CF,$C4,$7D,$78,$B7,$ED,$CF
	.dc.b	$FA,$5C,$E4,$E2,$43,$EF,$57,$F9,$6F,$E8,$5F,$1B,$49,$3F,$EC,$18
	.dc.b	$FF,$0A,$6D,$FE,$E0,$77,$19,$9D,$BF,$EA,$D8,$DB,$FE,$B5,$76,$E6
	.dc.b	$37,$87,$D3,$9D,$EF,$F5,$3B,$BF,$70,$4B,$FD,$FF,$A7,$5F,$CB,$45
	.dc.b	$BF,$81,$3F,$E3,$0F,$97,$95,$FE,$35,$7F,$C6,$78,$E2,$3F,$C5,$F1
	.dc.b	$6B,$F0,$9E,$EF,$99,$BF,$2D,$FE,$3A,$25,$7F,$8F,$FD,$4A,$FF,$8E
	.dc.b	$22,$CF,$29,$FF,$43,$7D,$0F,$F8,$E2,$BF,$D6,$7C,$5F,$C7,$F5,$4F
	.dc.b	$F7,$3F,$89,$FC,$24,$63,$DE,$FF,$66,$56,$EB,$95,$AF,$71,$86,$D7
	.dc.b	$09,$FF,$72,$4F,$5A,$E5,$EF,$E7,$0F,$0B,$67,$55,$C2,$F1,$3B,$D3
	.dc.b	$F4,$F3,$EA,$3F,$DE,$0E,$1F,$26,$1E,$D8,$C6,$3E,$C5,$F2,$DC,$F9
	.dc.b	$65,$F3,$CD,$5C,$30,$E5,$7E,$BA,$7D,$0B,$D4,$FF,$25,$C7,$F4,$93
	.dc.b	$2E,$1F,$E9,$91,$BF,$82,$FF,$0C,$36,$38,$1F,$ED,$6F,$65,$F2,$FF
	.dc.b	$E1,$DF,$8C,$46,$27,$67,$FB,$65,$C3,$CA,$37,$3D,$BD,$57,$C6,$D2
	.dc.b	$57,$2C,$3F,$BE,$C7,$F8,$AF,$D3,$FE,$BD,$DB,$FD,$8C,$F2,$B8,$AF
	.dc.b	$E4,$07,$F6,$77,$11,$AF,$4B,$F9,$23,$4F,$AD,$3F,$DD,$23,$F9,$47
	.dc.b	$1F,$D3,$E5,$BC,$3F,$27,$0E,$D3,$FC,$56,$FF,$BD,$F9,$5F,$CB,$FA
	.dc.b	$7F,$4B,$F2,$1C,$27,$C5,$39,$CF,$F0,$13,$99,$E3,$1E,$D9,$F4,$C7
	.dc.b	$76,$BC,$A3,$47,$F8,$0E,$BD,$F8,$2B,$B5,$39,$9F,$90,$BE,$67,$33
	.dc.b	$E0,$18,$7F,$47,$F6,$DF,$A6,$FF,$40,$CF,$B1,$FE,$38,$71,$5F,$EB
	.dc.b	$2E,$CE,$F5,$C3,$B2,$D6,$B0,$E8,$DC,$3F,$41,$7A,$AF,$3E,$7F,$9C
	.dc.b	$F8,$8F,$4D,$3E,$1D,$F1,$7A,$EF,$56,$A7,$35,$FE,$CD,$8C,$56,$DB
	.dc.b	$1E,$AD,$7F,$C5,$5F,$89,$3B,$BE,$3F,$63,$E9,$DE,$97,$C6,$AD,$61
	.dc.b	$C6,$B5,$39,$BF,$AB,$82,$F6,$FE,$89,$AF,$EA,$9A,$FE,$CF,$F1,$2F
	.dc.b	$4B,$F4,$B0,$FF,$1F,$61,$E3,$F4,$7E,$BD,$6B,$8F,$FB,$E4,$E6,$BD
	.dc.b	$96,$AD,$47,$2C,$FF,$BA,$BA,$8C,$CF,$EE,$A7,$2F,$84,$FF,$27,$46
	.dc.b	$9E,$9A,$FC,$6C,$1E,$A7,$92,$3C,$BF,$07,$F2,$BE,$15,$F2,$18,$CF
	.dc.b	$E4,$46,$1F,$14,$EE,$75,$8F,$87,$4F,$F5,$04,$63,$2F,$DD,$E7,$FE
	.dc.b	$AC,$C3,$E9,$53,$D6,$D4,$7C,$BF,$07,$C5,$F2,$1F,$21,$B1,$F0,$5B
	.dc.b	$97,$BC,$F9,$0E,$31,$EE,$F1,$9E,$FE,$15,$AE,$1F,$88,$7F,$B8,$27
	.dc.b	$F0,$9F,$89,$7B,$67,$3E,$B7,$E3,$23,$18,$FF,$14,$BE,$15,$EF,$9C
	.dc.b	$DF,$F3,$59,$CD,$7D,$E9,$CD,$7F,$A0,$67,$35,$BB,$FD,$DF,$E2,$4F
	.dc.b	$67,$F7,$B5,$AB,$E1,$EA,$F8,$F5,$CB,$71,$FC,$4D,$23,$15,$EE,$6E
	.dc.b	$7E,$97,$8D,$7E,$43,$F4,$EF,$C3,$5F,$18,$FD,$37,$E2,$7F,$E4,$76
	.dc.b	$FB,$9E,$4F,$27,$F5,$F3,$9C,$BE,$5C,$62,$75,$5C,$0F,$CE,$F9,$9B
	.dc.b	$CF,$E5,$57,$F1,$A8,$C4,$FC,$98,$FE,$17,$E2,$3E,$B2,$9E,$0F,$6A
	.dc.b	$D7,$C0,$FA,$FC,$34,$F4,$C7,$52,$3F,$EA,$AF,$95,$BA,$7E,$27,$E9
	.dc.b	$9F,$3D,$FC,$C3,$D3,$7D,$73,$AB,$F9,$7D,$6B,$1F,$E2,$55,$A9,$CD
	.dc.b	$F1,$BE,$15,$FB,$CB,$E1,$87,$ED,$70,$E1,$39,$BF,$E2,$35,$AA,$FE
	.dc.b	$EE,$D7,$FC,$00,$D7,$FD,$8F,$9F,$C6,$DB,$E3,$B6,$FF,$D8,$ED,$9F
	.dc.b	$F4,$F3,$78,$3F,$CD,$11,$F3,$B1,$F9,$ED,$AF,$F6,$7B,$4F,$F5,$31
	.dc.b	$67,$FD,$29,$F4,$DF,$E0,$AF,$07,$FA,$BB,$FE,$0C,$CE,$F7,$FE,$FE
	.dc.b	$FF,$8A,$4E,$70,$FE,$74,$F1,$6C,$FC,$6C,$3A,$BC,$5B,$7D,$30,$EA
	.dc.b	$DA,$E2,$FD,$6F,$6B,$E9,$75,$D9,$FA,$FB,$FE,$B5,$FA,$93,$F4,$08
	.dc.b	$FF,$88,$33,$6F,$FE,$31,$8F,$F1,$1F,$9C,$FC,$35,$F5,$DF,$D8,$17
	.dc.b	$D7,$D8,$C3,$6B,$FA,$C5,$71,$8C,$61,$FC,$DE,$31,$3F,$D0,$BE,$A1
	.dc.b	$F1,$7F,$E0,$2E,$0F,$EB,$9F,$09,$FE,$00,$9C,$D7,$F6,$39,$CB,$C6
	.dc.b	$B8,$79,$AB,$E1,$18,$6C,$FB,$EF,$9E,$3F,$E4,$06,$BF,$E4,$86,$BF
	.dc.b	$E7,$BF,$CC,$BF,$C2,$D3,$EC,$5F,$17,$FE,$02,$73,$3E,$D7,$33,$BF
	.dc.b	$73,$55,$B6,$C7,$C1,$EE,$FD,$97,$D4,$FF,$AD,$A3,$07,$EC,$1C,$61
	.dc.b	$FD,$07,$1F,$E0,$D3,$EB,$AF,$EF,$71,$89,$F6,$47,$F7,$EF,$BE,$7C
	.dc.b	$7F,$F8,$83,$F0,$3F,$E6,$B4,$9C,$E1,$FE,$0F,$F7,$95,$A9,$CD,$7F
	.dc.b	$6F,$9C,$FA,$5F,$E3,$49,$CF,$17,$F8,$CE,$F8,$63,$E1,$26,$3F,$8F
	.dc.b	$AE,$78,$FF,$9D,$BA,$9F,$37,$F0,$BD,$1F,$F1,$FC,$62,$FE,$14,$62
	.dc.b	$7F,$C3,$33,$9A,$FF,$3C,$B5,$E9,$1F,$E1,$EF,$09,$F2,$DF,$C7,$FE
	.dc.b	$AD,$F2,$FF,$E4,$C9,$CE,$5F,$C7,$A7,$2F,$F8,$5A,$33,$87,$FA,$19
	.dc.b	$AF,$FA,$39,$AF,$FA,$5D,$AB,$81,$FA,$83,$E4,$BF,$C9,$3F,$A8,$5C
	.dc.b	$3E,$84,$DB,$FD,$34,$DE,$2F,$F2,$F4,$E6,$7F,$D4,$98,$74,$C3,$FC
	.dc.b	$DE,$DC,$3F,$D3,$6D,$9F,$F5,$1F,$6F,$EE,$FE,$CF,$F7,$47,$EE,$FE
	.dc.b	$9E,$DF,$EE,$3C,$24,$E2,$FF,$71,$0A,$C6,$7F,$E7,$1E,$17,$D6,$7F
	.dc.b	$99,$3E,$2B,$FA,$1E,$1F,$87,$9C,$E5,$FC,$12,$73,$7C,$6F,$D3,$5A
	.dc.b	$BF,$F5,$CE,$9F,$84,$9C,$9E,$ED,$9E,$87,$F9,$B7,$EB,$7F,$CF,$71
	.dc.b	$8C,$FF,$88,$FC,$42,$73,$5C,$49,$3F,$1A,$B8,$63,$E0,$6E,$7F,$42
	.dc.b	$BF,$CF,$5F,$53,$CA,$FE,$BC,$5B,$FE,$39,$8A,$B1,$6B,$57,$F9,$E7
	.dc.b	$CE,$FD,$3C,$63,$89,$F9,$47,$F9,$EA,$3F,$D0,$1F,$0F,$FE,$D8,$EE
	.dc.b	$FF,$47,$CE,$70,$FF,$30,$4E,$6B,$FD,$B2,$D8,$FF,$70,$B5,$FF,$75
	.dc.b	$35,$F1,$F0,$3E,$DF,$EF,$D7,$F3,$8F,$BE,$FE,$33,$5F,$BF,$9C,$BF
	.dc.b	$CE,$63,$3A,$7F,$AC,$67,$2F,$BE,$7F,$D7,$53,$AA,$EF,$5A,$7E,$04
	.dc.b	$66,$7F,$4D,$3D,$9F,$CE,$B5,$E3,$CF,$F9,$E8,$C4,$7F,$B7,$6B,$FA
	.dc.b	$7F,$DF,$73,$8F,$F7,$54,$63,$C8,$FF,$64,$C3,$D6,$FF,$5C,$D6,$E3
	.dc.b	$15,$B3,$FD,$2A,$E0,$FF,$38,$17,$2F,$76,$EC,$DF,$63,$DB,$19,$87
	.dc.b	$27,$FD,$8F,$3A,$8C,$1D,$B0,$7C,$9F,$E1,$BE,$C7,$9C,$7F,$B8,$F3
	.dc.b	$FE,$26,$F8,$8D,$88,$DE,$49,$DE,$7F,$EF,$2B,$E9,$9F,$E3,$E7,$38
	.dc.b	$FE,$53,$19,$5A,$7C,$1E,$17,$C8,$7B,$E5,$E4,$7F,$37,$EC,$FD,$F8
	.dc.b	$C4,$7E,$7A,$B9,$3A,$BA,$E4,$F4,$63,$E1,$57,$CE,$31,$1E,$07,$51
	.dc.b	$98,$E1,$5D,$23,$11,$EA,$3A,$9E,$D7,$2F,$87,$89,$F0,$7D,$AB,$F6
	.dc.b	$B1,$89,$EA,$7F,$BB,$E7,$4D,$3F,$77,$F2,$EF,$8A,$6D,$CA,$BE,$64
	.dc.b	$E6,$31,$5C,$2B,$6F,$31,$3B,$70,$DD,$3E,$82,$FA,$0F,$B1,$18,$6E
	.dc.b	$9E,$0B,$EB,$3C,$1B,$1E,$49,$DD,$FA,$9A,$F8,$A7,$6F,$9F,$0F,$B7
	.dc.b	$87,$86,$2D,$70,$8F,$35,$F8,$6B,$F1,$5C,$1F,$93,$F2,$3F,$1D,$1F
	.dc.b	$A8,$E1,$FE,$E7,$8F,$F2,$48,$FE,$10,$F8,$5E,$F9,$78,$AF,$94,$FE
	.dc.b	$FA,$F9,$F0,$77,$8F,$BE,$7C,$AF,$B0,$CB,$F5,$9A,$E4,$FE,$3D,$C9
	.dc.b	$8D,$13,$49,$3A,$3C,$D5,$F7,$23,$AE,$0F,$C5,$3F,$B2,$57,$F9,$81
	.dc.b	$AF,$FB,$DE,$FA,$BC,$8F,$21,$BF,$17,$9C,$F0,$D3,$E1,$D7,$0A,$FF
	.dc.b	$18,$7D,$B3,$EA,$BF,$CF,$1D,$88,$C4,$FF,$80,$1A,$FF,$6A,$6B,$FE
	.dc.b	$08,$73,$7F,$C6,$BE,$FF,$FC,$8B,$3F,$E7,$86,$BE,$0B,$F9,$77,$DD
	.dc.b	$F8,$33,$AA,$E5,$DF,$AC,$3A,$ED,$FA,$CF,$03,$FD,$99,$97,$F5,$66
	.dc.b	$BF,$D5,$9A,$FF,$56,$6B,$C7,$E0,$FE,$7F,$82,$F9,$3A,$7D,$EE,$B5
	.dc.b	$F4,$CF,$FC,$4F,$F0,$BE,$8F,$1C,$D3,$EC,$F8,$DF,$E5,$0C,$FF,$C6
	.dc.b	$31,$96,$9F,$E0,$AF,$2B,$0F,$F8,$16,$3E,$8D,$74,$8F,$F0,$54,$60
	.dc.b	$EB,$F0,$72,$39,$9F,$E2,$3F,$4B,$EC,$D9,$A4,$8F,$F1,$CC,$65,$FF
	.dc.b	$55,$FE,$9E,$FA,$2F,$C8,$B6,$3F,$CF,$13,$B7,$FC,$6F,$F3,$9F,$96
	.dc.b	$F4,$4D,$31,$F7,$56,$7F,$C8,$11,$5F,$F1,$FC,$7C,$97,$DE,$43,$FC
	.dc.b	$50,$53,$FD,$79,$FA,$8B,$D6,$FD,$2B,$63,$FC,$BA,$7E,$CB,$F5,$2F
	.dc.b	$E5,$CF,$F3,$E6,$BF,$D0,$DA,$FF,$55,$6B,$FE,$2A,$78,$F0,$FF,$08
	.dc.b	$F8,$FF,$D4,$74,$7F,$5B,$6D,$6F,$CA,$FE,$03,$F8,$BB,$EC,$3F,$DC
	.dc.b	$12,$8C,$3F,$EE,$36,$BF,$D4,$9A,$FF,$57,$73,$87,$F1,$8F,$C5,$FF
	.dc.b	$29,$3F,$BD,$0E,$1F,$E8,$B8,$FF,$6F,$AE,$2F,$F7,$1F,$A5,$FF,$37
	.dc.b	$CF,$FB,$CD,$AF,$F6,$26,$BF,$D9,$DA,$F1,$F5,$7F,$8E,$FD,$17,$C0
	.dc.b	$83,$67,$9E,$3F,$CB,$9B,$7F,$E0,$19,$E0,$6E,$FF,$93,$CF,$EC,$AF
	.dc.b	$FC,$B2,$D4,$D2,$77,$FF,$2E,$FB,$77,$CA,$53,$9C,$7F,$28,$F8,$9F
	.dc.b	$EF,$F7,$FE,$EB,$E3,$C1,$9F,$56,$46,$48,$7F,$BB,$5F,$E7,$7F,$6C
	.dc.b	$F9,$3F,$B6,$D9,$FF,$04,$87,$0F,$F8,$31,$AF,$F8,$45,$CE,$FF,$D4
	.dc.b	$BE,$DB,$F5,$4D,$EE,$EF,$72,$FE,$F5,$6B,$87,$FC,$54,$D8,$FF,$3F
	.dc.b	$E1,$D2,$7F,$CF,$DD,$6F,$B3,$FD,$A7,$C0,$CB,$FF,$58,$47,$ED,$CC
	.dc.b	$3F,$EB,$08,$C3,$F8,$DB,$FA,$2F,$B9,$AF,$B7,$8B,$E1,$91,$F3,$CF
	.dc.b	$F6,$2C,$7F,$D6,$F1,$CE,$71,$D9,$9E,$49,$3E,$E8,$CB,$F0,$23,$2F
	.dc.b	$E7,$0F,$F4,$ED,$27,$71,$91,$1F,$90,$77,$3A,$DB,$FC,$D7,$9B,$EC
	.dc.b	$7F,$8C,$79,$72,$FF,$DA,$91,$FC,$6C,$C3,$FE,$D4,$8C,$3F,$ED,$9F
	.dc.b	$AB,$FE,$41,$F6,$52,$30,$43,$FD,$B6,$E5,$FF,$6E,$C7,$F8,$1B,$ED
	.dc.b	$1F,$03,$FC,$19,$B3,$F8,$C6,$BF,$E5,$66,$BF,$E5,$B7,$39,$7F,$87
	.dc.b	$3E,$11,$F1,$DF,$E1,$26,$D7,$F0,$7A,$DF,$27,$F9,$CD,$B1,$FE,$C8
	.dc.b	$9D,$BF,$2D,$B7,$FE,$28,$9D,$F9,$BF,$E5,$CE,$1F,$F1,$0B,$6F,$FC
	.dc.b	$20,$D7,$FD,$F7,$EE,$7F,$89,$3F,$16,$7D,$27,$0F,$C5,$9F,$45,$4D
	.dc.b	$47,$F4,$BF,$83,$98,$7F,$86,$FD,$E7,$37,$F9,$16,$3F,$09,$9B,$1E
	.dc.b	$71,$E2,$F2,$7E,$F3,$FC,$9F,$3F,$F0,$F6,$FF,$E4,$26,$F0,$FF,$BA
	.dc.b	$30,$FF,$3B,$F3,$F9,$FD,$FF,$37,$24,$C7,$94,$57,$C3,$9F,$C6,$C3
	.dc.b	$93,$63,$FD,$E7,$8F,$67,$FD,$EB,$E2,$3E,$CF,$FC,$8E,$D8,$FE,$78
	.dc.b	$D7,$F3,$9E,$C7,$D6,$51,$9D,$7F,$DC,$5A,$BF,$15,$7C,$C7,$D4,$6D
	.dc.b	$F4,$C7,$B9,$99,$91,$93,$C2,$F9,$08,$FD,$CF,$C5,$9F,$3F,$88,$FA
	.dc.b	$36,$FA,$2F,$01,$9C,$BF,$CD,$F9,$F8,$AB,$73,$A7,$B9,$FD,$6F,$F1
	.dc.b	$57,$DC,$FF,$9A,$9B,$19,$8F,$C7,$F9,$7F,$C4,$B0,$FF,$3D,$D6,$AB
	.dc.b	$F0,$7D,$FF,$E1,$76,$FB,$5C,$F0,$91,$F9,$69,$D5,$FC,$A6,$F2,$7F
	.dc.b	$B5,$F0,$E9,$3D,$BE,$7B,$C8,$4B,$FC,$FE,$5F,$E8,$C9,$D4,$75,$CF
	.dc.b	$F4,$B8,$FC,$58,$F7,$5F,$17,$CF,$3A,$31,$5E,$CA,$F8,$44,$9F,$D6
	.dc.b	$FC,$77,$F3,$D7,$E4,$47,$EC,$BC,$8F,$F7,$45,$7E,$11,$B9,$FA,$E7
	.dc.b	$95,$7E,$A2,$72,$FF,$97,$1A,$B3,$4F,$D3,$F8,$FF,$F0,$DC,$1A,$49
	.dc.b	$FE,$C7,$19,$7F,$81,$D3,$E2,$FC,$C7,$38,$FD,$E7,$C1,$7E,$31,$FF
	.dc.b	$4C,$62,$FC,$DF,$FB,$CC,$7F,$41,$8C,$6F,$FE,$6C,$E4,$93,$C2,$B7
	.dc.b	$AA,$D3,$67,$31,$8F,$05,$8F,$86,$BF,$A7,$7C,$4B,$CA,$FC,$C6,$3C
	.dc.b	$E5,$39,$E9,$FF,$4B,$D6,$B1,$E9,$87,$E5,$5F,$0D,$6E,$FF,$D6,$8D
	.dc.b	$7B,$D7,$15,$8F,$93,$CE,$FF,$65,$FD,$AB,$E5,$69,$2B,$F9,$27,$CB
	.dc.b	$7F,$2C,$7F,$C8,$93,$B7,$FC,$2A,$D7,$E1,$91,$C3,$80,$E1,$D4,$FF
	.dc.b	$A6,$BD,$5E,$3C,$FF,$EA,$EC,$7E,$4C,$63,$B3,$77,$E2,$F6,$F0,$FD
	.dc.b	$2C,$7F,$C8,$EF,$D4,$C3,$FD,$6B,$18,$84,$75,$DF,$FC,$CA,$EA,$BD
	.dc.b	$3B,$F9,$EF,$83,$63,$37,$ED,$6B,$E1,$E3,$7C,$47,$C4,$6C,$7C,$13
	.dc.b	$E0,$B8,$7B,$35,$F9,$3F,$1B,$F8,$5F,$0F,$FC,$82,$D8,$E4,$D7,$E6
	.dc.b	$35,$FA,$1C,$3F,$9C,$6F,$3F,$D6,$8C,$47,$C7,$6B,$FA,$9F,$73,$F2
	.dc.b	$5A,$FF,$B7,$3A,$7F,$CC,$3A,$FC,$BE,$0F,$C8,$FC,$D6,$9F,$BF,$3A
	.dc.b	$C3,$E6,$C6,$30,$D4,$7E,$9A,$75,$87,$D1,$BC,$CF,$87,$0E,$CE,$FE
	.dc.b	$83,$F4,$8F,$E2,$6D,$18,$8F,$0B,$5E,$13,$FE,$68,$9D,$BF,$E9,$B6
	.dc.b	$C7,$F8,$FF,$91,$CD,$FE,$CC,$AD,$5F,$EC,$FA,$BC,$38,$70,$6C,$66
	.dc.b	$7F,$7D,$5C,$70,$FF,$23,$5E,$E7,$F6,$BF,$69,$FD,$65,$FD,$AB,$67
	.dc.b	$C5,$CD,$FE,$9E,$9C,$F8,$3D,$67,$F8,$1D,$6F,$2F,$E0,$DD,$68,$C5
	.dc.b	$FF,$04,$E7,$E1,$39,$C3,$96,$BF,$C4,$BE,$CF,$A5,$BE,$D1,$F6,$3A
	.dc.b	$D5,$5A,$9C,$D7,$F2,$A8,$C5,$6D,$B1,$FE,$C7,$FB,$3F,$D0,$BF,$E9
	.dc.b	$86,$ED,$FA,$16,$BF,$C0,$DA,$FF,$08,$77,$B7,$E0,$1B,$3F,$E2,$98
	.dc.b	$FF,$2D,$BD,$FA,$FE,$35,$3E,$CB,$CE,$FF,$8E,$3E,$9D,$E7,$6A,$73
	.dc.b	$87,$F8,$72,$73,$87,$FB,$7E,$78,$DF,$F1,$76,$BF,$C7,$1A,$FF,$40
	.dc.b	$6B,$E1,$CF,$B1,$F6,$7B,$58,$3E,$CF,$38,$77,$F0,$75,$ED,$3E,$9B
	.dc.b	$CD,$18,$6C,$FF,$B0,$B3,$FF,$63,$36,$7F,$41,$9F,$C1,$6C,$EA,$3F
	.dc.b	$D3,$0D,$C7,$FC,$77,$97,$F9,$9A,$FC,$47,$C1,$EB,$F1,$FA,$FF,$1D
	.dc.b	$B1,$93,$C4,$F6,$23,$83,$58,$7F,$9A,$2B,$FC,$97,$18,$BF,$DD,$FD
	.dc.b	$7B,$EC,$BE,$37,$D9,$FC,$1A,$FE,$D7,$5A,$9C,$EB,$F7,$AF,$85,$7F
	.dc.b	$B9,$1B,$8F,$F2,$D6,$BF,$D3,$DA,$FF,$54,$77,$7D,$67,$B9,$FD,$63
	.dc.b	$EC,$DF,$13,$DF,$1F,$07,$C0,$79,$B6,$B7,$FE,$92,$D1,$18,$CF,$C5
	.dc.b	$5F,$53,$3F,$F4,$8D,$F9,$E3,$06,$F3,$FE,$3B,$F6,$7F,$94,$9F,$34
	.dc.b	$FE,$27,$E4,$5F,$67,$4F,$85,$E1,$DA,$F4,$E5,$FE,$12,$CB,$FC,$45
	.dc.b	$1F,$16,$7C,$44,$7F,$A9,$4B,$C7,$EB,$1C,$A1,$5C,$9F,$F7,$04,$E9
	.dc.b	$FF,$71,$46,$73,$FF,$0C,$78,$1F,$E7,$5F,$7F,$F8,$7F,$85,$FE,$C6
	.dc.b	$C7,$FD,$6F,$A7,$F8,$F9,$FF,$57,$46,$71,$E7,$F1,$BF,$1C,$93,$F7
	.dc.b	$5F,$CE,$E7,$C6,$7C,$7E,$BF,$6F,$99,$FE,$03,$39,$D3,$FA,$3F,$D8
	.dc.b	$BE,$6B,$F9,$57,$A0,$FB,$2F,$C9,$E5,$FE,$FB,$AD,$69,$F2,$AB,$5E
	.dc.b	$87,$F0,$C9,$CD,$7F,$79,$6B,$FE,$1B,$6B,$FE,$20,$6B,$FE,$44,$78
	.dc.b	$E3,$E4,$C7,$F9,$AF,$E0,$4F,$85,$ED,$1F,$23,$7F,$93,$CB,$74,$F8
	.dc.b	$ED,$CB,$FA,$C7,$0F,$D1,$9F,$F2,$72,$AE,$31,$F3,$A7,$E9,$EB,$F5
	.dc.b	$1B,$19,$3E,$9C,$65,$A7,$E3,$3E,$D1,$E0,$7F,$09,$8F,$F2,$FC,$63
	.dc.b	$2F,$EA,$CE,$BC,$3B,$5F,$02,$B6,$F1,$C1,$5C,$67,$B4,$E6,$31,$3F
	.dc.b	$CA,$27,$8D,$74,$BE,$AD,$8F,$CC,$6F,$FE,$C8,$7E,$D7,$0F,$EE,$9C
	.dc.b	$FA,$DF,$82,$8C,$61,$F8,$18,$FB,$C6,$7B,$6D,$7E,$B3,$A5,$F1,$0F
	.dc.b	$B4,$FE,$D6,$BE,$D2,$C6,$1B,$5C,$67,$4D,$9D,$3F,$AF,$FB,$37,$C1
	.dc.b	$FF,$7F,$E8,$79,$BF,$EE,$EA,$D4,$E6,$31,$AF,$C5,$3F,$D2,$37,$CA
	.dc.b	$BF,$7A,$D8,$FF,$28,$35,$FF,$39,$35,$FF,$3B,$BB,$BF,$F1,$7B,$67
	.dc.b	$FD,$1B,$D3,$F2,$27,$38,$7F,$83,$3E,$5F,$FC,$3D,$87,$FA,$31,$FE
	.dc.b	$EA,$FF,$8B,$5C,$EA,$73,$5F,$C2,$1C,$98,$C3,$FD,$50,$E3,$E4,$7F
	.dc.b	$BE,$E5,$FC,$1E,$BF,$D4,$DC,$EF,$B4,$7C,$5E,$BC,$3F,$DF,$C6,$32
	.dc.b	$F9,$D1,$97,$F7,$D1,$F2,$AF,$DF,$3F,$E0,$AF,$B3,$7C,$AF,$F8,$AB
	.dc.b	$ED,$3F,$AA,$6F,$F3,$C4,$EB,$1F,$F5,$7C,$EB,$0F,$F2,$BC,$E6,$BF
	.dc.b	$D0,$0D,$7F,$D4,$4D,$7F,$D4,$ED,$7F,$D9,$6F,$18,$E7,$7F,$E4,$3F
	.dc.b	$C3,$9F,$87,$7F,$8D,$B1,$FE,$71,$7B,$7F,$D6,$3E,$7B,$E9,$3E,$1F
	.dc.b	$A7,$FE,$26,$8C,$4F,$C2,$8C,$4C,$8F,$8B,$EF,$DD,$3E,$07,$88,$FA
	.dc.b	$DF,$E5,$7F,$6D,$FE,$F5,$BF,$E9,$13,$97,$F9,$54,$66,$B8,$D7,$F8
	.dc.b	$DA,$B5,$1F,$99,$E4,$56,$38,$3F,$CF,$F5,$70,$FF,$6E,$FD,$43,$ED
	.dc.b	$7F,$CF,$DF,$06,$8C,$4F,$FA,$B6,$31,$3A,$9C,$E5,$FE,$D1,$9C,$D7
	.dc.b	$FB,$9A,$73,$87,$F5,$AF,$01,$F3,$DF,$E6,$A9,$CD,$7F,$BE,$E7,$2F
	.dc.b	$EF,$67,$F8,$1F,$D7,$FF,$04,$64,$64,$FE,$05,$18,$C7,$FE,$00,$F0
	.dc.b	$3E,$83,$D6,$DA,$F5,$36,$7D,$0D,$7F,$DD,$B9,$FD,$DA,$FF,$3F,$CE
	.dc.b	$F3,$FF,$42,$FE,$A7,$FE,$4F,$C3,$FC,$0B,$7B,$7E,$07,$A9,$F1,$48
	.dc.b	$FC,$9D,$7F,$B7,$65,$D9,$FE,$D9,$1F,$4A,$7F,$D0,$DF,$6D,$F1,$59
	.dc.b	$1D,$0F,$E5,$1F,$10,$9C,$DF,$F1,$7C,$3E,$C5,$E3,$2F,$F0,$4D,$EF
	.dc.b	$0E,$57,$FC,$77,$E1,$7F,$A8,$F6,$7F,$AF,$5B,$3F,$EE,$B6,$C8,$87
	.dc.b	$FB,$CB,$D1,$FD,$64,$6B,$05,$6E,$30,$F1,$F1,$5F,$5B,$FD,$B7,$0E
	.dc.b	$58,$FF,$AB,$3E,$DB,$F0,$31,$FE,$C9,$FB,$6F,$E3,$0B,$64,$EB,$3F
	.dc.b	$F1,$2D,$6A,$73,$CB,$F9,$AA,$E7,$7F,$71,$AF,$AE,$77,$9F,$E7,$67
	.dc.b	$A3,$96,$CB,$36,$1F,$CD,$2F,$8C,$FF,$19,$3D,$92,$FB,$4F,$4C,$7F
	.dc.b	$BD,$70,$FE,$C1,$18,$8F,$F2,$11,$F3,$0D,$1F,$EE,$D8,$FF,$25,$BA
	.dc.b	$3F,$DE,$CF,$F9,$2D,$EF,$6F,$B5,$FF,$5B,$C7,$CD,$39,$E7,$FE,$FB
	.dc.b	$F6,$3F,$DA,$F5,$BB,$E5,$1C,$29,$F2,$3F,$7F,$ED,$BA,$47,$C1,$CB
	.dc.b	$F8,$86,$3F,$7A,$3D,$15,$BC,$BE,$F5,$6B,$1F,$65,$74,$3F,$CF,$73
	.dc.b	$EF,$EF,$E5,$E9,$F6,$9F,$C6,$E1,$F8,$F8,$C4,$27,$75,$D2,$BC,$13
	.dc.b	$F9,$49,$F2,$65,$C2,$3C,$0F,$28,$CD,$71,$9F,$19,$C7,$0F,$04,$67
	.dc.b	$0E,$11,$89,$F7,$9C,$0F,$D2,$57,$91,$CC,$F0,$74,$7E,$A7,$0F,$D2
	.dc.b	$47,$87,$8F,$FD,$3B,$3D,$1A,$FC,$20,$E3,$E5,$1F,$15,$FE,$E0,$9C
	.dc.b	$D7,$ED,$27,$3A,$F6,$9F,$DC,$46,$1D,$98,$AE,$57,$D5,$E9,$7F,$4E
	.dc.b	$9F,$49,$FC,$02,$30,$DD,$3C,$15,$C1,$B8,$FC,$7C,$58,$72,$70,$D7
	.dc.b	$E5,$3E,$5C,$3C,$17,$ED,$D7,$BD,$7D,$E3,$9E,$3F,$C4,$A3,$78,$67
	.dc.b	$0E,$58,$FC,$3A,$F3,$5F,$C2,$6B,$E4,$C3,$D0,$FC,$3B,$FC,$5F,$0F
	.dc.b	$EE,$7C,$4F,$E3,$F1,$FE,$B2,$E0,$FE,$ED,$5F,$EB,$BC,$BE,$66,$7F
	.dc.b	$1F,$DE,$DF,$0F,$D3,$6D,$7D,$26,$E5,$FC,$2B,$8B,$33,$F1,$A6,$E1
	.dc.b	$FE,$99,$8E,$D1,$C3,$8D,$5A,$8F,$E0,$B5,$BA,$FD,$7B,$5F,$E1,$8D
	.dc.b	$7C,$87,$F9,$C7,$ED,$9E,$17,$91,$B1,$E0,$3F,$AE,$FD,$A3,$ED,$BD
	.dc.b	$7D,$08,$C4,$EA,$7D,$BF,$6C,$FA,$ED,$D6,$AB,$FC,$0C,$D7,$FB,$7B
	.dc.b	$5F,$E6,$8D,$7F,$A0,$3C,$78,$BF,$DC,$75,$FE,$4A,$9F,$F4,$53,$5F
	.dc.b	$EC,$53,$BE,$0F,$EA,$5D,$EF,$84,$FA,$7D,$5F,$EB,$06,$C6,$6F,$FD
	.dc.b	$29,$E0,$7F,$33,$CB,$F8,$F5,$EE,$7F,$B2,$56,$DF,$E7,$11,$96,$C6
	.dc.b	$5F,$EF,$DF,$0D,$F2,$3C,$53,$4F,$C1,$F8,$7F,$E4,$B9,$FF,$14,$46
	.dc.b	$5A,$7F,$80,$3C,$7C,$3F,$DF,$EB,$9B,$A3,$FD,$31,$D0,$BE,$91,$BB
	.dc.b	$FE,$66,$FF,$6B,$AE,$2F,$F8,$E5,$AF,$A0,$FF,$13,$F9,$FF,$80,$C0
	.dc.b	$D3,$F0,$7E,$47,$E5,$27,$FC,$47,$3A,$78,$90,$FF,$17,$CF,$8B,$1F
	.dc.b	$F6,$0F,$DA,$FE,$93,$CB,$34,$C7,$D6,$5B,$FF,$60,$45,$7D,$F2,$9F
	.dc.b	$9D,$10,$FF,$29,$1B,$DF,$FA,$C7,$3F,$F9,$76,$3E,$1B,$5F,$F1,$B1
	.dc.b	$8D,$3F,$B8,$AF,$B4,$FE,$6B,$5F,$D8,$63,$1E,$2F,$EE,$A7,$3C,$BF
	.dc.b	$E2,$6F,$8A,$F2,$57,$F8,$A6,$F8,$56,$A3,$FC,$59,$3F,$B0,$7F,$D4
	.dc.b	$4D,$9D,$DF,$F0,$6F,$B2,$7D,$A5,$25,$F5,$C7,$FC,$6B,$F6,$8F,$93
	.dc.b	$CC,$FF,$AB,$61,$F3,$3F,$C4,$1B,$1C,$4F,$F1,$F6,$1D,$9C,$0F,$A0
	.dc.b	$FF,$3C,$B8,$6B,$FE,$B9,$6E,$7F,$17,$D8,$F8,$33,$9F,$47,$FC,$37
	.dc.b	$5C,$23,$FC,$40,$D7,$FC,$9A,$D7,$FC,$A0,$F4,$CF,$FC,$A3,$5F,$15
	.dc.b	$BE,$27,$F2,$A9,$D5,$70,$EF,$C1,$1F,$5F,$DA,$3E,$17,$11,$FE,$D1
	.dc.b	$87,$C8,$FF,$34,$6C,$7B,$EB,$F9,$CF,$C3,$7F,$84,$E3,$E4,$75,$FF
	.dc.b	$BC,$7D,$BC,$9A,$F5,$64,$EA,$93,$A8,$C4,$FF,$42,$FB,$47,$C7,$D2
	.dc.b	$4F,$D2,$FA,$8F,$F0,$1B,$F5,$27,$6F,$EF,$1A,$FF,$8C,$1A,$FF,$93
	.dc.b	$5A,$FF,$9E,$9D,$F1,$FE,$6E,$7F,$71,$3F,$EF,$36,$C7,$E7,$E7,$7C
	.dc.b	$3F,$EB,$7F,$2F,$F7,$1E,$EC,$C3,$E1,$F8,$3F,$D0,$5B,$39,$7F,$6F
	.dc.b	$3F,$BF,$8F,$81,$C7,$27,$D8,$FF,$64,$CB,$F9,$E6,$5E,$18,$C5,$CA
	.dc.b	$D9,$F1,$7C,$1E,$1C,$F2,$30,$43,$D8,$E5,$F7,$47,$73,$17,$F0,$E7
	.dc.b	$13,$EC,$FB,$1F,$E2,$DE,$E4,$8C,$10,$FE,$26,$E5,$FE,$2B,$1F,$C7
	.dc.b	$7E,$9F,$F8,$F7,$DA,$49,$F8,$F1,$97,$E4,$C6,$5F,$E3,$B1,$FE,$06
	.dc.b	$FB,$47,$C6,$D2,$4F,$ED,$BE,$07,$F9,$92,$CF,$F8,$B5,$70,$8F,$E8
	.dc.b	$2D,$7F,$CE,$AD,$7F,$D2,$4D,$7F,$D7,$EE,$E3,$F9,$04,$BE,$2B,$FC
	.dc.b	$32,$DA,$FE,$0F,$5B,$E3,$FF,$64,$7A,$0F,$1B,$F8,$1E,$3F,$E8,$98
	.dc.b	$C7,$07,$F8,$99,$B8,$7E,$2B,$DE,$FF,$12,$FD,$93,$ED,$3F,$D4,$6D
	.dc.b	$9F,$EB,$6D,$7F,$D1,$ED,$7F,$DA,$FD,$CE,$9F,$F2,$86,$5F,$E9,$98
	.dc.b	$C7,$C0,$FF,$51,$3C,$5F,$68,$F9,$DF,$26,$CE,$6F,$1F,$BD,$33,$FE
	.dc.b	$51,$BF,$B6,$71,$1E,$29,$F3,$91,$FF,$51,$CB,$0F,$91,$1F,$5F,$87
	.dc.b	$E3,$C6,$34,$F8,$CD,$8F,$F7,$8E,$DF,$E3,$F9,$FE,$A5,$E1,$BE,$43
	.dc.b	$FC,$96,$DC,$3F,$9C,$B5,$EC,$7F,$BE,$FC,$CF,$AC,$FF,$BF,$63,$4F
	.dc.b	$FB,$4D,$AF,$FC,$00,$6B,$B9,$FF,$6A,$C7,$5A,$FB,$9F,$66,$FA,$EF
	.dc.b	$B8,$DB,$F1,$DF,$9A,$64,$79,$9F,$C1,$7D,$A3,$E6,$73,$3F,$EC,$AC
	.dc.b	$79,$36,$BC,$24,$7E,$9E,$0F,$9E,$F0,$9C,$B2,$FE,$F1,$F6,$2F,$B3
	.dc.b	$FF,$38,$B6,$FF,$AF,$B5,$F8,$FC,$3F,$ED,$1F,$17,$75,$FE,$F4,$AD
	.dc.b	$57,$FB,$E5,$A8,$8E,$1B,$1F,$9F,$EE,$7D,$A5,$3A,$CF,$8B,$D0,$F2
	.dc.b	$91,$FC,$7F,$99,$FE,$11,$CF,$D7,$3E,$32,$4F,$F8,$FB,$1F,$EB,$CD
	.dc.b	$8F,$06,$92,$3B,$1F,$03,$1F,$86,$47,$F3,$BF,$1D,$FC,$ED,$FC,$48
	.dc.b	$B4,$FC,$32,$47,$85,$AF,$CA,$F2,$BF,$DF,$D8,$FF,$01,$6F,$3F,$C7
	.dc.b	$9E,$95,$F5,$3E,$D2,$47,$D4,$7F,$83,$62,$F2,$33,$3C,$4F,$9F,$3F
	.dc.b	$54,$8E,$47,$81,$FE,$65,$E6,$FF,$5A,$E2,$FB,$4F,$EF,$95,$FA,$68
	.dc.b	$C7,$07,$F9,$DA,$A4,$FF,$89,$E7,$2D,$AE,$3A,$FF,$0D,$F8,$3F,$CD
	.dc.b	$4F,$F8,$B7,$DE,$79,$7F,$9A,$AF,$F2,$9B,$70,$FC,$44,$EB,$A7,$FD
	.dc.b	$0B,$CF,$FE,$B0,$AD,$60,$AD,$3D,$63,$3A,$FF,$A0,$6B,$FD,$9B,$F6
	.dc.b	$6F,$B4,$A4,$BF,$F0,$CF,$D9,$FF,$29,$D3,$FC,$B9,$5B,$9F,$F2,$0B
	.dc.b	$5F,$C0,$B6,$3F,$95,$4F,$1D,$3E,$D4,$EE,$BE,$A3,$5C,$DF,$FA,$49
	.dc.b	$46,$5F,$D4,$2F,$6B,$F0,$9C,$9F,$EA,$F8,$C4,$FE,$B6,$31,$3A,$EE
	.dc.b	$FD,$86,$9F,$EA,$1F,$7B,$17,$D2,$AB,$E0,$6E,$1D,$21,$87,$FB,$0E
	.dc.b	$F9,$47,$7C,$3D,$BC,$7F,$E3,$28,$F6,$B5,$CF,$C3,$FF,$AB,$B0,$FE
	.dc.b	$34,$FF,$AC,$3E,$37,$F0,$3E,$05,$E2,$34,$97,$F9,$7B,$FF,$38,$73
	.dc.b	$F3,$E3,$FC,$AC,$FC,$E6,$BF,$21,$AB,$E3,$BF,$CB,$7C,$5E,$98,$FE
	.dc.b	$29,$18,$7D,$FA,$FC,$8E,$77,$CC,$2F,$85,$73,$F7,$B0,$F0,$39,$B8
	.dc.b	$E2,$FF,$72,$63,$D5,$F9,$58,$74,$BE,$73,$9C,$7F,$80,$2F,$B4,$EB
	.dc.b	$D1,$FE,$AD,$8C,$57,$E5,$BE,$1F,$2F,$0A,$B9,$FB,$1A,$FC,$17,$3C
	.dc.b	$AE,$9F,$F6,$6D,$7F,$A7,$30,$EB,$7F,$C7,$1B,$8F,$87,$0E,$37,$C2
	.dc.b	$B5,$1E,$0E,$A7,$DA,$7F,$4F,$C7,$F7,$CD,$AF,$F7,$9E,$BF,$9B,$9C
	.dc.b	$F8,$7C,$6B,$59,$7F,$B5,$1B,$3F,$4F,$AB,$55,$EA,$B7,$D9,$FC,$0A
	.dc.b	$7D,$A3,$EC,$F5,$39,$8C,$4F,$FA,$5A,$B8,$BF,$C6,$63,$39,$F9,$AB
	.dc.b	$85,$FC,$AF,$B4,$76,$D3,$E7,$F5,$FF,$93,$7D,$CC,$65,$FE,$C6,$AD
	.dc.b	$B6,$B7,$BF,$F9,$62,$30,$7F,$B3,$3E,$CF,$C3,$9F,$F3,$96,$BF,$D0
	.dc.b	$DA,$FF,$6B,$6A,$7D,$33,$FE,$F3,$6C,$FF,$94,$21,$F6,$9F,$84,$9C
	.dc.b	$D7,$3C,$F0,$79,$A6,$DF,$97,$E1,$3D,$FE,$E3,$ED,$39,$46,$1B,$3B
	.dc.b	$3F,$C8,$7B,$7F,$95,$DC,$63,$FD,$A3,$2F,$F5,$EC,$63,$3F,$F3,$15
	.dc.b	$6E,$B5,$87,$9F,$BB,$C3,$EC,$DC,$3F,$C9,$F9,$FF,$9A,$DF,$95,$39
	.dc.b	$BF,$F0,$BE,$0E,$DC,$65,$F1,$6A,$DF,$F9,$D2,$B4,$F2,$3F,$DF,$FF
	.dc.b	$60,$F8,$DF,$BF,$F6,$7F,$06,$73,$18,$DF,$F5,$F5,$B9,$C9,$FE,$21
	.dc.b	$C7,$F8,$3E,$6F,$B4,$7C,$9F,$0C,$65,$F0,$AF,$F0,$6D,$6A,$31,$7C
	.dc.b	$B0,$FF,$3A,$47,$F9,$C3,$2F,$85,$7F,$E1,$49,$CB,$FE,$38,$8C,$D7
	.dc.b	$1F,$7F,$CE,$7F,$06,$FB,$4F,$C1,$D2,$F8,$D6,$F0,$F8,$77,$BE,$5E
	.dc.b	$31,$FE,$80,$F9,$27,$DA,$7E,$91,$B1,$F2,$BC,$4C,$4F,$C6,$64,$7C
	.dc.b	$53,$FC,$41,$E4,$70,$CF,$D5,$3F,$C6,$A3,$19,$7C,$18,$C5,$F2,$AF
	.dc.b	$F3,$DF,$8D,$FE,$5F,$F8,$8F,$E3,$FE,$17,$FA,$D7,$2F,$37,$AD,$89
	.dc.b	$FA,$FF,$1D,$27,$C5,$87,$F9,$5F,$4F,$F1,$DC,$7E,$0A,$7C,$57,$23
	.dc.b	$78,$FF,$9B,$7E,$63,$18,$7F,$91,$3E,$C9,$F6,$7F,$80,$8F,$F1,$E4
	.dc.b	$62,$FF,$C6,$50,$8C,$3F,$E3,$56,$BF,$E3,$B7,$3B,$FE,$29,$B3,$FE
	.dc.b	$26,$FA,$F7,$C3,$FF,$40,$9C,$DF,$F9,$C5,$AF,$7C,$3F,$9C,$57,$0B
	.dc.b	$F8,$33,$97,$DB,$BB,$5F,$97,$C3,$F3,$1B,$3A,$6D,$7D,$2A,$E1,$3C
	.dc.b	$AB,$E9,$CF,$D4,$F4,$AC,$F6,$6C,$71,$6E,$9F,$ED,$5B,$47,$F5,$EF
	.dc.b	$B4,$FF,$03,$F8,$77,$BF,$DF,$18,$C7,$ED,$BF,$E0,$CD,$75,$5B,$6C
	.dc.b	$EA,$B8,$9F,$C5,$32,$FF,$62,$DF,$19,$D5,$F1,$7E,$3E,$DF,$EB,$56
	.dc.b	$FA,$FF,$DC,$B2,$7D,$A7,$EC,$27,$F0,$33,$F5,$48,$FE,$42,$77,$E2
	.dc.b	$E3,$2F,$CD,$FC,$63,$EC,$FF,$BC,$AF,$B4,$B1,$86,$E1,$F9,$49,$D3
	.dc.b	$67,$4F,$DE,$FB,$37,$C2,$FE,$57,$0F,$F5,$B4,$62,$7F,$2A,$D8,$FF
	.dc.b	$44,$35,$FF,$47,$39,$E3,$FE,$3A,$D9,$FC,$A7,$8D,$FE,$12,$8C,$57
	.dc.b	$6B,$C1,$F9,$29,$B5,$FA,$CF,$98,$FF,$08,$BF,$E5,$A7,$FB,$45,$7E
	.dc.b	$F9,$CB,$FE,$9A,$73,$87,$F8,$D3,$AD,$1A,$6F,$A1,$FE,$22,$8C,$1C
	.dc.b	$2F,$FD,$61,$97,$F0,$7C,$3F,$D7,$AD,$C3,$FC,$7F,$BB,$EC,$FF,$DF
	.dc.b	$97,$F2,$FC,$0F,$71,$1F,$B9,$1A,$C3,$9C,$FF,$8D,$3E,$C1,$F2,$DF
	.dc.b	$46,$51,$8B,$FE,$0A,$D8,$FF,$5C,$35,$FF,$5F,$39,$C7,$FA,$6B,$67
	.dc.b	$FD,$E1,$F6,$4F,$B2,$FF,$0F,$DF,$FB,$62,$B4,$FF,$BD,$5A,$F8,$B0
	.dc.b	$FF,$25,$D7,$1F,$5B,$D1,$1F,$E4,$CE,$17,$D1,$D3,$F4,$1A,$FF,$8F
	.dc.b	$70,$FF,$72,$CE,$B3,$FE,$1A,$D8,$CD,$70,$F6,$B9,$57,$F1,$2F,$09
	.dc.b	$F5,$BF,$E5,$A6,$C7,$E9,$FE,$CA,$E1,$FE,$F2,$9B,$C1,$FD,$B9,$B5
	.dc.b	$FE,$EB,$6C,$FF,$0A,$AA,$FF,$1D,$BF,$F2,$D7,$D2,$BE,$CA,$9F,$B0
	.dc.b	$F9,$EB,$8F,$FB,$66,$74,$D9,$D3,$4F,$D9,$4E,$2F,$F7,$7F,$47,$73
	.dc.b	$FC,$B7,$6A,$B1,$7F,$EB,$88,$56,$2F,$FD,$BD,$B3,$E8,$3F,$CE,$0D
	.dc.b	$BF,$F7,$67,$D3,$FF,$8F,$9F,$F7,$74,$66,$99,$7F,$A6,$3E,$13,$EC
	.dc.b	$E1,$C8,$FF,$3D,$61,$C6,$B7,$87,$FB,$CB,$0E,$58,$71,$D7,$FB,$2D
	.dc.b	$68,$FF,$79,$38,$8F,$F8,$01,$F6,$5F,$33,$D0,$98,$7C,$97,$FC,$F5
	.dc.b	$18,$BE,$59,$B0,$E5,$69,$EB,$7C,$23,$91,$23,$8C,$F0,$5E,$03,$E1
	.dc.b	$18,$F7,$CB,$FD,$73,$F0,$5F,$ED,$0E,$9F,$F5,$4B,$72,$FE,$FB,$F5
	.dc.b	$3F,$EC,$97,$FC,$0B,$19,$CB,$E0,$36,$F9,$F9,$BF,$AC,$CD,$87,$1B
	.dc.b	$DC,$FE,$C3,$67,$D9,$7E,$07,$2F,$75,$49,$CF,$2F,$F3,$56,$CE,$A7
	.dc.b	$E2,$46,$1C,$37,$2F,$BD,$7D,$67,$FC,$4B,$F5,$FF,$CA,$2B,$57,$CA
	.dc.b	$B5,$7C,$30,$FF,$63,$DE,$E7,$35,$E1,$F4,$BF,$8C,$63,$ED,$3C,$85
	.dc.b	$35,$1F,$E1,$E7,$E1,$61,$FE,$23,$9E,$BF,$65,$FE,$D6,$9C,$D7,$07
	.dc.b	$F1,$87,$FB,$D9,$FF,$17,$E0,$9E,$15,$27,$FC,$33,$5C,$8F,$C0,$D6
	.dc.b	$E3,$33,$DE,$5F,$1B,$F8,$DF,$B1,$E9,$1F,$88,$FB,$27,$CC,$6E,$B5
	.dc.b	$97,$E4,$F8,$55,$AC,$BC,$38,$74,$18,$F4,$7F,$35,$E4,$A7,$9C,$7D
	.dc.b	$BC,$6B,$A8,$AF,$E8,$62,$CF,$18,$EB,$8F,$D0,$9C,$D7,$09,$F1,$5F
	.dc.b	$85,$FF,$14,$C6,$67,$71,$C8,$F1,$14,$F1,$14,$F1,$16,$FC,$55,$D6
	.dc.b	$3E,$07,$1B,$D0,$F2,$47,$FB,$A1,$B1,$98,$EE,$D5,$F2,$8F,$89,$FE
	.dc.b	$2F,$1C,$4F,$E0,$51,$9E,$1F,$E0,$95,$C6,$31,$3B,$39,$D7,$29,$D3
	.dc.b	$D3,$0F,$A9,$4F,$A5,$FD,$8C,$61,$BA,$FE,$D2,$7A,$CF,$0A,$E5,$39
	.dc.b	$AD,$E1,$E6,$6B,$E2,$C7,$EE,$E1,$CA,$B8,$4E,$2B,$73,$98,$F7,$DF
	.dc.b	$67,$A1,$B3,$9B,$97,$B1,$F6,$F7,$FC,$AF,$87,$FC,$33,$1F,$F1,$27
	.dc.b	$07,$F6,$CA,$FF,$3F,$E1,$DD,$FA,$3E,$E6,$F8,$7E,$03,$63,$E8,$B7
	.dc.b	$17,$23,$93,$55,$C8,$EB,$58,$C3,$FD,$2B,$1F,$BA,$9C,$4C,$8D,$99
	.dc.b	$29,$FC,$32,$78,$B6,$3F,$86,$46,$23,$F9,$47,$D7,$BE,$C7,$D1,$D2
	.dc.b	$8C,$4F,$FA,$FA,$31,$3E,$BF,$B1,$7D,$67,$0B,$DD,$FF,$54,$6E,$FF
	.dc.b	$DA,$5B,$B7,$F0,$3B,$E3,$5D,$67,$8B,$E4,$C3,$FD,$7F,$8F,$82,$3F
	.dc.b	$93,$B5,$FE,$79,$CA,$EC,$7C,$77,$D1,$8C,$36,$33,$3F,$D5,$8C,$44
	.dc.b	$7F,$AB,$C6,$9F,$F7,$2F,$77,$F4,$9D,$3F,$BD,$7C,$17,$C6,$F0,$CD
	.dc.b	$3E,$F7,$87,$FE,$3E,$BF,$F1,$04,$65,$A7,$FB,$8A,$34,$FF,$72,$8D
	.dc.b	$34,$D7,$8B,$8B,$F0,$C7,$23,$FD,$F3,$5C,$A3,$FC,$2D,$2F,$3B,$ED
	.dc.b	$B6,$BF,$C3,$B1,$F7,$BC,$7F,$C9,$46,$08,$7F,$83,$27,$83,$FE,$0E
	.dc.b	$8D,$BF,$EA,$EF,$96,$FC,$E7,$96,$69,$8F,$AB,$B3,$FE,$AD,$8A,$FF
	.dc.b	$AD,$A3,$E5,$30,$FF,$23,$1A,$C7,$FC,$50,$E4,$FF,$0B,$99,$E5,$FF
	.dc.b	$00,$FD,$8F,$F2,$39,$FE,$7D,$28,$C4,$EB,$C3,$FD,$7F,$C4,$30,$FD
	.dc.b	$8D,$6A,$FF,$C5,$0D,$79,$57,$09,$FD,$BF,$D8,$3E,$76,$92,$BF,$C6
	.dc.b	$19,$78,$BE,$C5,$F2,$59,$9F,$EE,$B7,$C5,$DE,$9F,$E1,$A8,$C4,$7F
	.dc.b	$7A,$AE,$2E,$8F,$F1,$E7,$CE,$7F,$98,$5C,$07,$0F,$F8,$09,$B8,$7C
	.dc.b	$4F,$5F,$E0,$4F,$09,$CE,$1F,$E2,$A6,$C7,$6F,$3F,$F8,$AD,$70,$8E
	.dc.b	$91,$FE,$15,$9E,$F1,$FE,$2D,$6B,$CE,$35,$D5,$FC,$7A,$7F,$59,$87
	.dc.b	$D6,$6E,$1F,$E3,$AF,$B1,$FF,$76,$E1,$E5,$BE,$04,$7D,$F1,$A7,$FD
	.dc.b	$B9,$E6,$FF,$2D,$F8,$07,$B3,$4C,$44,$9F,$F7,$34,$65,$FF,$74,$47
	.dc.b	$F3,$2F,$B1,$7C,$4D,$24,$FC,$EF,$A5,$78,$58,$8F,$F1,$7D,$70,$7E
	.dc.b	$9B,$5F,$F1,$9C,$F0,$73,$5F,$E3,$D9,$D4,$7F,$95,$5A,$FF,$97,$1A
	.dc.b	$FF,$9D,$3A,$BF,$93,$BF,$E8,$9F,$31,$1F,$E9,$D8,$FC,$34,$63,$D9
	.dc.b	$FD,$AF,$93,$27,$35,$B7,$93,$78,$BF,$D5,$B3,$F0,$F8,$A5,$7B,$1F
	.dc.b	$EA,$F9,$FF,$BC,$A3,$04,$3E,$1C,$6B,$D7,$DF,$57,$0E,$59,$5F,$2E
	.dc.b	$32,$FC,$D8,$CB,$F2,$23,$2F,$CA,$8C,$E3,$FB,$EC,$3F,$28,$FF,$B1
	.dc.b	$3E,$BF,$F8,$9E,$12,$FD,$F1,$FC,$58,$C3,$F0,$23,$0F,$C4,$FA,$2F
	.dc.b	$E3,$7F,$4D,$23,$04,$3E,$3B,$97,$E4,$C7,$F7,$AF,$B1,$7C,$45,$24
	.dc.b	$FE,$DB,$DE,$FE,$53,$97,$F1,$6A,$DC,$FE,$E9,$AF,$F9,$E9,$B1,$FC
	.dc.b	$72,$78,$F5,$FF,$A8,$27,$75,$FE,$9D,$6B,$9B,$FF,$45,$46,$23,$FB
	.dc.b	$37,$83,$F2,$2B,$FD,$8D,$E6,$7F,$94,$78,$3F,$C3,$0D,$AF,$C3,$FA
	.dc.b	$FF,$E1,$6F,$AF,$7D,$8D,$24,$FF,$15,$AE,$0F,$E3,$3D,$AF,$E8,$A4
	.dc.b	$75,$F6,$2F,$99,$EF,$A3,$93,$B7,$B1,$2F,$A4,$E5,$FE,$3F,$9F,$F8
	.dc.b	$73,$8B,$F9,$FB,$77,$FF,$61,$B6,$7D,$F8,$7B,$32,$7C,$5D,$24,$F4
	.dc.b	$CB,$FC,$67,$E2,$BE,$C2,$92,$7F,$97,$FD,$8F,$D4,$CF,$FD,$A3,$5B
	.dc.b	$9F,$E6,$6D,$7F,$D6,$6D,$D3,$FA,$AD,$71,$9F,$F7,$B4,$E6,$BE,$9F
	.dc.b	$D8,$3E,$B3,$E9,$B6,$BA,$E3,$E6,$99,$1E,$07,$EC,$FD,$8B,$E5,$73
	.dc.b	$3E,$29,$B9,$FF,$A8,$F9,$17,$BA,$E4,$7D,$BC,$5F,$35,$E2,$38,$E7
	.dc.b	$FE,$F0,$FA,$F7,$D8,$FF,$96,$9B,$7F,$D5,$1A,$B2,$F0,$79,$7F,$D5
	.dc.b	$E3,$55,$C4,$8F,$E2,$E3,$4F,$FB,$72,$B9,$D7,$A7,$B9,$F6,$3F,$98
	.dc.b	$6D,$74,$3C,$5F,$2F,$E5,$23,$D7,$2A,$F5,$8B,$7E,$69,$A7,$E9,$AB
	.dc.b	$E1,$D7,$E7,$DA,$F1,$3E,$03,$F0,$C8,$FE,$97,$E2,$FF,$9C,$D7,$F8
	.dc.b	$5B,$15,$6A,$BD,$CF,$EC,$AB,$77,$F3,$E1,$30,$82,$B9,$B8,$36,$E1
	.dc.b	$C3,$5E,$2F,$FB,$E1,$FB,$1F,$19,$F6,$CF,$17,$89,$63,$24,$3E,$E5
	.dc.b	$FF,$0A,$C1,$3F,$E6,$2E,$3F,$F5,$14,$BE,$C5,$B6,$1B,$87,$F9,$8E
	.dc.b	$B8,$3F,$E1,$FA,$E1,$5B,$33,$1E,$01,$F0,$6F,$1B,$F4,$D1,$98,$C7
	.dc.b	$37,$FA,$72,$73,$BB,$0F,$97,$E7,$3E,$03,$F1,$F3,$9C,$7A,$DF,$0B
	.dc.b	$FA,$38,$F4,$9F,$F5,$24,$E5,$61,$CF,$CC,$FF,$5E,$FD,$83,$EC,$69
	.dc.b	$2B,$FB,$BF,$D8,$FF,$2C,$7F,$C7,$73,$B6,$CE,$DF,$F0,$04,$ED,$F8
	.dc.b	$42,$78,$38,$6E,$7F,$E6,$46,$E7,$F0,$2B,$94,$E7,$D9,$FC,$67,$1F
	.dc.b	$FA,$86,$D1,$8E,$FF,$C8,$46,$31,$FC,$9D,$7F,$AB,$34,$F9,$5E,$EE
	.dc.b	$E7,$F2,$F5,$B5,$5D,$70,$D4,$7A,$5E,$17,$CA,$3A,$E1,$E0,$E3,$FF
	.dc.b	$15,$47,$E8,$30,$FE,$3D,$EB,$FE,$7E,$3E,$14,$E9,$E1,$19,$F8,$AF
	.dc.b	$C3,$78,$37,$8B,$FD,$91,$5F,$EE,$76,$BF,$AC,$9D,$BF,$AB,$C3,$FC
	.dc.b	$DD,$CD,$DD,$B1,$9E,$3F,$CD,$BE,$56,$BF,$8E,$E5,$FC,$87,$1F,$B2
	.dc.b	$3F,$0F,$1F,$EF,$18,$C6,$DF,$93,$7F,$25,$C1,$F3,$67,$55,$F2,$B7
	.dc.b	$E9,$5A,$7B,$7C,$45,$9E,$B3,$70,$FC,$F4,$DF,$CF,$3F,$9C,$CF,$F9
	.dc.b	$21,$B5,$FE,$4D,$5F,$49,$FA,$3E,$5F,$D2,$60,$9D,$C6,$1B,$7F,$AC
	.dc.b	$6D,$FE,$91,$E7,$1D,$FC,$1E,$8E,$DB,$3B,$6D,$F7,$C2,$DF,$48,$F1
	.dc.b	$7A,$78,$33,$E0,$DC,$7D,$91,$71,$F6,$4D,$CB,$D1,$F7,$3D,$3E,$CF
	.dc.b	$AB,$70,$F8,$8D,$BF,$88,$DA,$F8,$7E,$D7,$FA,$12,$BF,$CD,$BD,$1E
	.dc.b	$88,$7A,$16,$24,$E8,$FF,$39,$C6,$30,$FF,$3E,$7E,$A9,$FE,$11,$9F
	.dc.b	$F3,$DB,$63,$3F,$53,$F1,$23,$18,$FD,$46,$D7,$F1,$C6,$D7,$F1,$45
	.dc.b	$5C,$7C,$C9,$5D,$9E,$91,$8A,$F1,$F9,$3D,$CF,$07,$89,$6F,$CB,$16
	.dc.b	$7D,$33,$71,$F2,$4D,$AF,$24,$DA,$F2,$7D,$E7,$AC,$F6,$7D,$F5,$AF
	.dc.b	$83,$16,$7E,$2C,$DC,$3E,$04,$DA,$F8,$13,$6B,$E0,$7D,$EF,$C7,$3E
	.dc.b	$47,$9B,$6B,$E6,$C5,$9F,$A3,$37,$0F,$99,$36,$BE,$64,$DA,$F9,$9F
	.dc.b	$55,$F4,$CF,$A9,$E4,$5A,$FB,$31,$67,$EE,$CD,$C3,$EC,$4D,$AF,$B1
	.dc.b	$36,$BE,$C7,$D9,$7D,$FF,$84,$FC,$03,$6B,$F0,$AE,$1F,$C3,$F9,$16
	.dc.b	$BF,$16,$DC,$7F,$0C,$D9,$FC,$37,$D1,$7E,$2C,$FC,$67,$CF,$58,$FC
	.dc.b	$8B,$87,$F2,$7F,$23,$70,$FC,$BF,$97,$F9,$38,$FC,$A3,$67,$FC,$2F
	.dc.b	$8F,$F6,$BF,$AB,$79,$9F,$D8,$E7,$27,$F8,$72,$33,$B7,$F0,$27,$7C
	.dc.b	$36,$B8,$65,$FE,$01,$9D,$DF,$28,$46,$1D,$4A,$B9,$5C,$BE,$17,$BC
	.dc.b	$3C,$D7,$B9,$CB,$E0,$8F,$D7,$79,$BF,$B1,$3F,$65,$E4,$DA,$FD,$BC
	.dc.b	$5C,$BF,$77,$36,$BF,$6D,$5F,$B9,$3F,$75,$E3,$57,$F7,$E5,$AF,$E0
	.dc.b	$71,$67,$F9,$8F,$E4,$DF,$C8,$4F,$E5,$73,$FE,$23,$F1,$11,$9F,$1B
	.dc.b	$F9,$4B,$70,$F8,$D3,$AF,$1F,$35,$A6,$CE,$AB,$85,$FF,$8A,$FF,$55
	.dc.b	$7C,$AF,$F1,$DF,$6F,$F9,$BF,$A6,$9C,$E1,$DB,$C6,$56,$A7,$38,$73
	.dc.b	$C3,$CB,$E4,$E5,$FE,$83,$0F,$A9,$4F,$F4,$4D,$DF,$79,$FE,$A5,$CF
	.dc.b	$F0,$17,$BA,$D4,$E6,$3F,$C4,$B7,$CA,$BF,$C0,$CD,$8F,$F3,$33,$5F
	.dc.b	$F3,$5B,$5F,$CA,$72,$39,$3F,$24,$7E,$4F,$92,$DF,$E1,$A2,$D7,$E4
	.dc.b	$26,$E5,$F8,$C9,$B5,$D2,$6D,$7F,$92,$38,$1E,$5F,$D8,$3E,$CF,$99
	.dc.b	$6F,$EA,$45,$AF,$C0,$52,$B1,$8F,$F9,$66,$5F,$52,$FA,$BF,$ED,$BE
	.dc.b	$A3,$EA,$7F,$D5,$39,$FC,$AA,$D3,$FA,$B9,$D7,$95,$F5,$27,$38,$7F
	.dc.b	$94,$1B,$8F,$F9,$9E,$BA,$79,$1B,$CB,$F7,$56,$FA,$4A,$63,$E9,$2E
	.dc.b	$7F,$B2,$9D,$36,$74,$D2,$47,$F0,$13,$65,$9E,$6F,$F9,$93,$F5,$27
	.dc.b	$C1,$FC,$0D,$9E,$2F,$E3,$A7,$38,$7B,$E7,$2F,$E3,$23,$37,$CE,$F9
	.dc.b	$78,$BC,$5F,$CB,$FE,$AD,$FE,$01,$9F,$F4,$A6,$CF,$AB,$FF,$5B,$E3
	.dc.b	$F9,$6A,$D7,$07,$C7,$9C,$DF,$F9,$EA,$73,$5F,$EB,$D6,$C7,$FA,$A3
	.dc.b	$81,$F5,$DE,$0F,$5B,$F7,$BE,$BC,$C3,$F7,$1F,$73,$70,$FC,$2C,$DC
	.dc.b	$FF,$84,$CD,$AF,$F5,$0F,$43,$EC,$FF,$83,$CF,$F0,$86,$CF,$F0,$36
	.dc.b	$E1,$F9,$06,$DF,$F9,$05,$FF,$1A,$9F,$E3,$6A,$B5,$FE,$4D,$8B,$5F
	.dc.b	$E4,$D9,$B5,$FE,$49,$FA,$EF,$F1,$E9,$FE,$3E,$E0,$B8,$7F,$95,$62
	.dc.b	$DF,$F9,$56,$6D,$71,$AD,$1F,$E5,$B8,$CB,$FE,$53,$AB,$5F,$E5,$39
	.dc.b	$BE,$4F,$F9,$6F,$E3,$BF,$C9,$A7,$F9,$3B,$E7,$EE,$1F,$E6,$58,$B7
	.dc.b	$FE,$65,$9B,$8F,$F9,$87,$E1,$FF,$CA,$A7,$F9,$5B,$E2,$6E,$1F,$E6
	.dc.b	$D8,$B7,$FE,$6D,$9B,$5C,$6B,$47,$F9,$CE,$32,$FF,$9A,$FA,$FF,$CC
	.dc.b	$27,$F9,$8B,$AE,$C7,$F9,$E6,$2D,$FF,$9E,$66,$DF,$F9,$D6,$6D,$7F
	.dc.b	$9D,$66,$D7,$F9,$D6,$6E,$7F,$E7,$D9,$CC,$FF,$B1,$07,$D6,$7E,$ED
	.dc.b	$13,$A3,$FD,$D1,$19,$8E,$0D,$7F,$E0,$0E,$1C,$23,$11,$FE,$C4,$3A
	.dc.b	$E1,$D1,$E9,$3B,$8C,$47,$FB,$40,$D9,$BA,$F0,$C6,$A3,$A5,$78,$67
	.dc.b	$31,$CB,$E0,$FA,$3C,$6E,$BE,$8B,$EF,$5C,$23,$06,$CF,$09,$83,$CE
	.dc.b	$60,$F3,$5F,$86,$BA,$C7,$A7,$97,$FD,$0A,$F8,$C3,$87,$E8,$35,$E1
	.dc.b	$F0,$1F,$17,$E1,$5F,$26,$DB,$E9,$DF,$19,$CC,$62,$B8,$56,$EF,$BC
	.dc.b	$62,$75,$1F,$72,$B8,$BE,$E5,$F2,$5F,$57,$2F,$BD,$18,$9F,$92,$F0
	.dc.b	$8C,$E1,$DD,$DD,$71,$D3,$D4,$DB,$F0,$56,$E7,$C9,$1E,$19,$D2,$BF
	.dc.b	$03,$E5,$AE,$0F,$2C,$3C,$75,$C2,$B7,$87,$92,$7C,$F7,$C2,$FA,$5E
	.dc.b	$CE,$11,$D3,$7F,$F7,$07,$D2,$FE,$56,$FF,$DC,$0D,$D7,$FD,$C0,$DC
	.dc.b	$3F,$DC,$1F,$21,$6F,$8C,$7F,$BB,$E7,$1F,$0C,$6D,$61,$93,$B3,$FD
	.dc.b	$ED,$18,$7F,$DE,$D1,$70,$FF,$79,$4F,$4F,$9D,$95,$FE,$FD,$7B,$FD
	.dc.b	$56,$23,$FD,$FD,$16,$BF,$DF,$53,$72,$FF,$7D,$4D,$AF,$F7,$D4,$FA
	.dc.b	$18,$7F,$93,$31,$FD,$AD,$AF,$F8,$2E,$49,$71,$70,$BF,$F2,$C4,$70
	.dc.b	$AD,$B6,$15,$C6,$BE,$57,$E3,$1F,$E1,$B9,$FF,$2D,$B5,$FF,$31,$37
	.dc.b	$0F,$14,$72,$8F,$ED,$2F,$0F,$23,$99,$1E,$7E,$0D,$79,$C5,$AE,$78
	.dc.b	$2F,$9D,$68,$E8,$43,$C7,$7E,$8F,$9C,$91,$E8,$7D,$7F,$5F,$88,$F4
	.dc.b	$C5,$C3,$BD,$AB,$F1,$5E,$17,$F8,$72,$FF,$CE,$DE,$06,$0C,$F8,$12
	.dc.b	$32,$43,$FC,$8E,$F2,$EF,$91,$FE,$7F,$79,$34,$87,$F9,$39,$CB,$FC
	.dc.b	$3F,$83,$F9,$B7,$EA,$EF,$03,$F4,$12,$8C,$3F,$E7,$66,$BC,$4F,$E9
	.dc.b	$3F,$AC,$2F,$8C,$EB,$5F,$F2,$D4,$FB,$EB,$80,$71,$1F,$E5,$F9,$E1
	.dc.b	$87,$F4,$9F,$A6,$78,$FE,$46,$E1,$FE,$88,$6B,$D3,$D6,$FE,$56,$E6
	.dc.b	$7D,$70,$F9,$5A,$73,$8F,$7F,$C3,$7F,$38,$AF,$F4,$9B,$5F,$3B,$72
	.dc.b	$FF,$3A,$3F,$CF,$5B,$E1,$F2,$24,$F2,$F0,$EB,$CA,$2C,$F2,$9F,$9F
	.dc.b	$EA,$F9,$6B,$FD,$46,$D7,$E3,$B6,$7F,$D0,$9F,$4A,$F1,$F8,$E7,$F6
	.dc.b	$21,$E4,$53,$38,$7F,$B0,$DF,$C8,$91,$FF,$63,$79,$35,$FF,$49,$C5
	.dc.b	$C3,$FD,$AF,$3F,$84,$FC,$2B,$FD,$E5,$5F,$EB,$86,$BF,$71,$B3,$FE
	.dc.b	$99,$5F,$19,$C7,$E1,$E5,$71,$7A,$46,$23,$94,$5A,$E5,$3F,$94,$FA
	.dc.b	$57,$C6,$7F,$81,$6F,$FC,$33,$18,$70,$D7,$FC,$14,$D7,$FC,$22,$E6
	.dc.b	$BF,$D5,$DF,$47,$FE,$E4,$9F,$F6,$9B,$5F,$CB,$CE,$F9,$BF,$C4,$FF
	.dc.b	$0B,$F3,$BE,$07,$FA,$33,$6B,$FC,$57,$87,$F1,$88,$C4,$7F,$AF,$BB
	.dc.b	$37,$BF,$FB,$92,$3F,$DD,$DF,$79,$FD,$37,$E1,$8E,$B1,$FE,$0F,$E9
	.dc.b	$7F,$08,$E5,$97,$FE,$FD,$8C,$BF,$EF,$F8,$CB,$FE,$CC,$8C,$BF,$EC
	.dc.b	$E8,$CE,$7F,$C8,$BC,$2F,$F1,$F3,$7C,$EF,$E2,$A4,$AF,$5E,$DF,$E8
	.dc.b	$EF,$3B,$11,$FE,$D8,$9B,$5C,$E7,$F8,$EF,$9D,$2B,$89,$FE,$E5,$7F
	.dc.b	$90,$F9,$D9,$8F,$F7,$0C,$5A,$F2,$4F,$F1,$2E,$3F,$E6,$B1,$F5,$BC
	.dc.b	$AC,$1F,$C5,$B8,$7F,$A7,$60,$AD,$E7,$FE,$51,$6C,$7F,$97,$9A,$FE
	.dc.b	$98,$E6,$7E,$02,$AB,$FC,$5A,$BF,$AD,$79,$6D,$7F,$B0,$CF,$B2,$30
	.dc.b	$FF,$3C,$3F,$9E,$F9,$BF,$E0,$7F,$A6,$F0,$C9,$A6,$2B,$AB,$FD,$7F
	.dc.b	$2F,$ED,$ED,$BF,$ED,$2D,$AE,$58,$F5,$F2,$FB,$7D,$1C,$BF,$25,$FF
	.dc.b	$4F,$6C,$65,$FE,$71,$7E,$38,$E5,$1A,$38,$F9,$DE,$93,$86,$DF,$DA
	.dc.b	$2F,$D5,$3D,$9F,$71,$EF,$FA,$BB,$19,$21,$E0,$84,$6A,$31,$16,$7B
	.dc.b	$4F,$C9,$FA,$69,$5E,$C7,$E6,$7D,$3E,$23,$A4,$5A,$F3,$4D,$D7,$E5
	.dc.b	$CD,$AF,$97,$3F,$57,$C4,$91,$F2,$1F,$B1,$F4,$98,$8F,$44,$5C,$3E
	.dc.b	$14,$DC,$3E,$BC,$DA,$FA,$F3,$F8,$1F,$56,$47,$D4,$7F,$09,$F3,$D8
	.dc.b	$8F,$89,$17,$0F,$9D,$37,$0F,$C1,$CD,$AF,$C1,$CF,$E3,$7E,$9A,$47
	.dc.b	$E0,$1F,$C8,$7B,$18,$8F,$A1,$17,$0F,$B5,$37,$0F,$C7,$CD,$AF,$C7
	.dc.b	$CF,$E6,$7B,$E4,$7E,$31,$FC,$E7,$C7,$62,$3E,$E4,$5C,$3F,$0D,$3F
	.dc.b	$E3,$ED,$FF,$C7,$FC,$52,$FF,$C7,$F1,$97,$FC,$83,$19,$7F,$C8,$31
	.dc.b	$FE,$46,$DE,$47,$F9,$19,$FF,$24,$F0,$E2,$3F,$C8,$F1,$70,$FF,$17
	.dc.b	$CF,$F9,$3B,$79,$5F,$E4,$E7,$FC,$A3,$C3,$88,$FF,$27,$C5,$AF,$F1
	.dc.b	$BC,$DD,$BF,$CC,$3E,$BA,$B1,$7F,$E5,$3A,$FF,$2D,$F0,$7F,$B7,$FD
	.dc.b	$C7,$CE,$F0,$E4,$C3,$6F,$FD,$DF,$3A,$7F,$CB,$95,$A6,$9A,$9E,$0A
	.dc.b	$78,$73,$7F,$96,$6F,$83,$FE,$5B,$BE,$54,$F9,$F7,$C2,$7F,$86,$E3
	.dc.b	$FD,$E9,$18,$F4,$BF,$B5,$36,$3F,$DE,$0D,$7F,$DE,$FE,$FA,$DF,$1B
	.dc.b	$FD,$C6,$30,$D8,$CC,$FF,$BE,$67,$2F,$4F,$4F,$F9,$4A,$3C,$6E,$1B
	.dc.b	$E0,$79,$23,$11,$FE,$8A,$F9,$FF,$F4,$A7,$1F,$C5,$8C,$7C,$A7,$FA
	.dc.b	$1B,$83,$E2,$D6,$A7,$E4,$70,$FF,$A6,$A1,$5B,$8C,$5D,$AE,$6F,$5E
	.dc.b	$05,$F3,$C3,$AE,$1B,$9F,$F5,$27,$C9,$FF,$A5,$B9,$BF,$D1,$BF,$04
	.dc.b	$8C,$63,$FE,$A4,$F9,$4B,$9F,$FA,$92,$B5,$3F,$EB,$1A,$CE,$1D,$23
	.dc.b	$FD,$69,$78,$AD,$DF,$FA,$C2,$F0,$66,$6D,$7F,$AE,$FE,$3F,$81,$C7
	.dc.b	$CF,$B1,$92,$5F,$FA,$EA,$32,$FF,$AF,$3E,$23,$C1,$F2,$1E,$16,$D7
	.dc.b	$FB,$1D,$B3,$B8,$FF,$65,$4F,$AB,$D2,$74,$53,$1E,$E5,$AF,$ED,$53
	.dc.b	$A6,$CE,$9A,$7F,$6F,$F9,$F7,$CF,$FC,$66,$C6,$7E,$7F,$13,$23,$83
	.dc.b	$B8,$C6,$5F,$D8,$AB,$8B,$67,$55,$C0,$F0,$7C,$9F,$73,$C1,$F3,$F6
	.dc.b	$32,$4A,$E9,$19,$7B,$4D,$E4,$FF,$68,$E0,$AC,$5F,$FB,$BA,$1F,$3F
	.dc.b	$ED,$3D,$DF,$09,$67,$E1,$45,$C7,$D5,$37,$0E,$B3,$6B,$A4,$DA,$FF
	.dc.b	$79,$DB,$E0,$3E,$51,$F2,$FE,$02,$DF,$CF,$8B,$5F,$1E,$6E,$3E,$A9
	.dc.b	$B5,$E9,$9B,$5D,$BE,$7F,$EB,$1F,$5F,$E7,$2D,$7D,$B8,$B3,$F4,$E6
	.dc.b	$E1,$F1,$A6,$D7,$C5,$9B,$5E,$AF,$9F,$FC,$17,$C1,$FE,$0D,$B5,$F8
	.dc.b	$77,$0F,$E2,$7E,$1E,$D7,$E0,$5B,$8F,$D4,$6C,$FD,$3F,$9F,$FC,$71
	.dc.b	$F8,$FF,$9F,$B1,$F9,$37,$0F,$E5,$7E,$66,$E1,$F8,$DF,$2F,$F2,$D1
	.dc.b	$F9,$76,$CF,$F9,$93,$E7,$FF,$95,$EB,$FC,$11,$B1,$9F,$2F,$EB,$36
	.dc.b	$75,$06,$98,$E1,$B3,$B2,$63,$FE,$F7,$BF,$E7,$33,$A8,$FD,$17,$CE
	.dc.b	$7E,$AC,$FD,$67,$93,$6F,$F6,$11,$67,$F5,$13,$77,$FC,$DD,$7E,$CC
	.dc.b	$FD,$A7,$8D,$5F,$DC,$16,$BF,$61,$16,$7F,$D2,$5E,$27,$F2,$0F,$79
	.dc.b	$E2,$D8,$C3,$70,$FF,$38,$4E,$9B,$3A,$69,$A9,$E1,$87,$F9,$82,$B8
	.dc.b	$2B,$FF,$5B,$7C,$FB,$E7,$FF,$9A,$FC,$52,$31,$8F,$83,$C7,$CC,$63
	.dc.b	$0E,$53,$FE,$C4,$8C,$61,$D3,$0F,$65,$F0,$7F,$A0,$DB,$E7,$D2,$C3
	.dc.b	$FD,$B9,$18,$6C,$66,$7F,$B4,$12,$74,$E0,$FE,$D8,$E2,$F8,$47,$79
	.dc.b	$F6,$FD,$C7,$9A,$F9,$47,$E3,$38,$24,$7E,$47,$82,$DF,$E2,$79,$15
	.dc.b	$8C,$7C,$9C,$7F,$E0,$4B,$C6,$1F,$DF,$63,$FC,$0F,$58,$C1,$1B,$9F
	.dc.b	$CF,$4F,$E0,$7C,$89,$1F,$7B,$CA,$B7,$F6,$66,$E3,$FE,$D9,$9B,$5F
	.dc.b	$13,$E8,$1F,$33,$F3,$9B,$1F,$ED,$2F,$49,$F4,$0F,$17,$57,$2B,$FD
	.dc.b	$AF,$9B,$0E,$2F,$DE,$18,$72,$BE,$11,$FA,$D1,$F1,$DF,$BB,$6D,$FE
	.dc.b	$FD,$7C,$76,$66,$4E,$5E,$11,$89,$E9,$17,$4F,$9F,$1C,$AB,$81,$E8
	.dc.b	$FA,$07,$CD,$7E,$39,$B1,$F0,$75,$FF,$21,$BB,$9D,$11,$D7,$81,$6F
	.dc.b	$E0,$0F,$9F,$FE,$EC,$E1,$B8,$FE,$57,$5F,$99,$F1,$7F,$9E,$9C,$E1
	.dc.b	$F9,$86,$DF,$F9,$05,$B5,$C2,$7E,$8C,$6B,$D6,$FF,$78,$7A,$BF,$C1
	.dc.b	$4F,$E0,$DF,$3D,$6B,$F8,$1C,$5C,$3F,$7D,$37,$2F,$D4,$5F,$FB,$E7
	.dc.b	$E7,$F8,$E1,$FC,$29,$B7,$FC,$15,$B5,$C3,$C1,$46,$2B,$FC,$B1,$6F
	.dc.b	$63,$31,$FE,$5B,$BF,$F2,$0D,$7F,$8D,$F1,$FF,$1C,$32,$74,$72,$AE
	.dc.b	$38,$FF,$BD,$15,$6E,$31,$F3,$DF,$E4,$13,$FC,$85,$C1,$6F,$9C,$FF
	.dc.b	$92,$E7,$0F,$39,$B8,$FF,$95,$26,$D7,$86,$6D,$71,$8E,$79,$7B,$6B
	.dc.b	$BC,$63,$14,$7F,$94,$FE,$62,$4E,$7E,$7A,$E1,$EB,$9B,$97,$F9,$92
	.dc.b	$6D,$7C,$1F,$97,$FF,$2E,$1F,$E5,$DF,$92,$B1,$DE,$2E,$1F,$16,$6D
	.dc.b	$FF,$9B,$A6,$D7,$CB,$9B,$5F,$E6,$DE,$7F,$F3,$41,$FE,$69,$E8,$B5
	.dc.b	$EB,$8B,$3F,$46,$6E,$1F,$E7,$A9,$B5,$F5,$E6,$D7,$58,$E4,$7C,$58
	.dc.b	$CC,$7F,$B0,$C7,$CF,$FF,$04,$7F,$9C,$F2,$FF,$3B,$8C,$46,$DA,$FF
	.dc.b	$0E,$AF,$F5,$34,$62,$3F,$D8,$47,$4B,$FE,$81,$5C,$DF,$E8,$51,$A9
	.dc.b	$E7,$39,$8F,$E8,$DE,$FF,$4A,$E3,$76,$72,$75,$3C,$EE,$5F,$34,$70
	.dc.b	$74,$60,$D9,$83,$FD,$BE,$5A,$F1,$3B,$3F,$D8,$4E,$1D,$FB,$AF,$7F
	.dc.b	$D7,$87,$FB,$1E,$31,$3E,$66,$BF,$3D,$BF,$04,$F9,$1F,$EF,$D3,$FE
	.dc.b	$92,$BF,$ED,$33,$9D,$7A,$46,$2B,$77,$C9,$B1,$99,$F8,$54,$F9,$0F
	.dc.b	$EE,$7A,$FD,$A8,$C6,$3A,$C3,$C3,$39,$6E,$9E,$2B,$E2,$DA,$E1,$84
	.dc.b	$7E,$6E,$1D,$1B,$5C,$1F,$F7,$54,$7A,$DD,$DF,$67,$33,$B5,$3C,$F0
	.dc.b	$F2,$5F,$77,$D5,$5B,$9D,$61,$F0,$6F,$F0,$9C,$BF,$C8,$7E,$63,$FC
	.dc.b	$27,$8F,$FB,$81,$BB,$FF,$1D,$6D,$7F,$B7,$FE,$62,$DF,$18,$FE,$19
	.dc.b	$38,$E7,$36,$B0,$C9,$D9,$FC,$4E,$30,$FF,$29,$8B,$87,$FB,$CA,$7A
	.dc.b	$7C,$C4,$AF,$E3,$4F,$7F,$98,$C4,$7F,$31,$8B,$5F,$EF,$A9,$B9,$7F
	.dc.b	$37,$9B,$5F,$EF,$99,$F4,$30,$FE,$4C,$7A,$F8,$F0,$FF,$3D,$2D,$FF
	.dc.b	$54,$CB,$FC,$17,$1F,$CC,$3E,$3B,$FC,$32,$DF,$1F,$E3,$C6,$23,$FC
	.dc.b	$25,$AA,$1E,$16,$AA,$57,$13,$A4,$61,$E3,$3C,$F0,$FE,$CA,$DD,$3F
	.dc.b	$AE,$BD,$27,$91,$C2,$F9,$57,$0B,$F2,$FC,$B4,$AF,$53,$E8,$F9,$7C
	.dc.b	$47,$82,$2D,$78,$27,$F0,$7E,$17,$F8,$62,$BF,$CF,$BD,$F8,$33,$E0
	.dc.b	$48,$C9,$0F,$F3,$E3,$CB,$BE,$47,$FA,$19,$E4,$D2,$1F,$E8,$37,$2F
	.dc.b	$F8,$7B,$E7,$9E,$07,$E6,$21,$18,$7F,$C8,$AD,$7F,$C4,$BF,$3E,$AE
	.dc.b	$13,$9C,$FF,$CE,$2D,$E0,$FF,$3C,$B6,$C2,$F9,$3F,$E5,$2F,$9F,$FF
	.dc.b	$2B,$37,$CB,$F2,$46,$23,$FC,$6B,$F3,$FF,$C7,$DD,$10,$FF,$1C,$C6
	.dc.b	$F1,$FF,$41,$CE,$E7,$FC,$7B,$F0,$4F,$0B,$E0,$36,$3F,$CC,$6D,$77
	.dc.b	$E1,$6A,$64,$69,$E3,$18,$8E,$B1,$6F,$FD,$27,$5F,$CE,$1F,$F4,$AF
	.dc.b	$A6,$F9,$2F,$A0,$DB,$FF,$39,$B5,$FF,$29,$FC,$FF,$F9,$A9,$CC,$F0
	.dc.b	$23,$FE,$7A,$8D,$3F,$EC,$1F,$23,$F1,$E4,$7F,$D9,$7E,$4D,$74,$7F
	.dc.b	$B3,$A3,$F2,$B1,$F8,$3F,$9F,$7C,$0D,$25,$7F,$A2,$21,$F1,$38,$8F
	.dc.b	$F3,$4C,$70,$75,$3C,$E7,$59,$FF,$AD,$AB,$55,$C2,$B4,$FF,$A5,$3E
	.dc.b	$7D,$F1,$3F,$95,$6C,$66,$3F,$31,$F3,$FF,$5F,$0F,$F6,$0D,$6A,$BF
	.dc.b	$C0,$AD,$7F,$D9,$0D,$7F,$D9,$AD,$5F,$04,$D3,$FD,$AF,$1F,$E5,$08
	.dc.b	$C7,$BD,$FC,$F9,$B1,$FE,$F6,$EB,$C5,$70,$8F,$F5,$54,$E9,$B1,$9C
	.dc.b	$3F,$CE,$15,$FC,$13,$CB,$FE,$0B,$E8,$CB,$FF,$80,$31,$97,$FC,$AF
	.dc.b	$E0,$7D,$EB,$FF,$3C,$56,$67,$FE,$00,$57,$17,$F9,$4D,$6F,$B7,$FC
	.dc.b	$F6,$FF,$8C,$AF,$FD,$01,$E6,$7F,$14,$25,$76,$F3,$2B,$CA,$2C,$FF
	.dc.b	$A1,$E7,$F8,$DF,$99,$2B,$CA,$FF,$1E,$F3,$31,$1E,$28,$B5,$FE,$8F
	.dc.b	$9F,$B3,$E3,$CA,$FF,$69,$F1,$7F,$33,$7F,$D6,$7B,$FF,$BF,$EF,$FD
	.dc.b	$FB,$39,$6D,$71,$AD,$96,$32,$F4,$8F,$E3,$D5,$CF,$4F,$87,$E6,$7F
	.dc.b	$34,$A8,$7C,$E7,$FB,$CF,$CC,$F7,$9F,$F5,$1B,$65,$E6,$23,$0B,$7E
	.dc.b	$D3,$A9,$FE,$A0,$DF,$2F,$A9,$DB,$2E,$FE,$57,$83,$E7,$25,$7F,$A5
	.dc.b	$AD,$38,$8F,$EC,$B0,$9C,$6B,$FE,$C2,$8D,$CF,$19,$F5,$F9,$92,$30
	.dc.b	$75,$9F,$79,$1F,$07,$CE,$57,$BC,$5B,$ED,$37,$5F,$8B,$36,$BE,$04
	.dc.b	$FC,$AF,$9A,$91,$EB,$7E,$6F,$CE,$62,$3D,$71,$70,$F5,$4D,$C3,$E8
	.dc.b	$CD,$AF,$99,$3F,$5B,$C3,$91,$F1,$DF,$B3,$F3,$98,$8F,$8F,$17,$0F
	.dc.b	$8D,$37,$0F,$BB,$36,$BE,$C4,$FE,$0B,$D6,$91,$F4,$DF,$C2,$FC,$DE
	.dc.b	$23,$E9,$C5,$C3,$E9,$4D,$C3,$F1,$33,$6B,$F0,$93,$F8,$EF,$9F,$91
	.dc.b	$F7,$DF,$C8,$FA,$F8,$8F,$BF,$17,$0F,$BD,$37,$0F,$CA,$CD,$AF,$C8
	.dc.b	$4F,$E6,$BE,$7E,$47,$E2,$DF,$CE,$FC,$FE,$23,$F1,$71,$70,$FC,$54
	.dc.b	$FF,$8D,$37,$FF,$1A,$F1,$4B,$FF,$1D,$46,$5F,$F1,$E4,$65,$FF,$18
	.dc.b	$C7,$F8,$F3,$79,$1F,$E4,$17,$FC,$7F,$C3,$88,$FF,$1B,$C5,$C3,$FC
	.dc.b	$6F,$3F,$E4,$4D,$E5,$7F,$92,$5F,$F2,$3F,$0E,$23,$FC,$7F,$16,$BF
	.dc.b	$C7,$F3,$76,$FF,$1F,$CD,$AF,$DC,$FA,$EA,$FF,$28,$F0,$7F,$85,$A5
	.dc.b	$ED,$BE,$7B,$FC,$8F,$C9,$98,$C4,$F2,$8C,$4F,$F9,$32,$AC,$F6,$6F
	.dc.b	$57,$F9,$6E,$AE,$3F,$DF,$F0,$E7,$3F,$E5,$FA,$E3,$6F,$9F,$7C,$17
	.dc.b	$F8,$56,$3C,$D1,$8F,$4B,$FD,$C7,$1B,$9D,$3F,$E3,$09,$D3,$FE,$E8
	.dc.b	$F7,$56,$F8,$AE,$31,$86,$C6,$67,$83,$5F,$8C,$79,$7D,$3F,$F3,$EB
	.dc.b	$8B,$46,$23,$CE,$D7,$C8,$DC,$BF,$CF,$DD,$2E,$2F,$F2,$3E,$BF,$13
	.dc.b	$6F,$F6,$94,$63,$83,$FC,$F3,$D7,$BC,$BE,$BC,$7F,$9F,$30,$C6,$1C
	.dc.b	$5B,$19,$BE,$55,$F8,$2B,$CB,$F8,$4A,$B8,$FF,$A4,$7A,$7B,$72,$FF
	.dc.b	$A2,$BA,$EE,$3F,$E8,$A9,$B8,$7E,$32,$6D,$F1,$AF,$F5,$04,$63,$45
	.dc.b	$D8,$C3,$67,$FD,$4C,$DB,$FF,$4D,$E7,$C7,$AF,$93,$6B,$FD,$58,$D9
	.dc.b	$DC,$7F,$AB,$E7,$C1,$F1,$B2,$3F,$A6,$72,$DA,$FF,$53,$CD,$D3,$FA
	.dc.b	$54,$DA,$FF,$3C,$E8,$FC,$65,$F3,$BE,$E5,$F3,$D6,$30,$DA,$FF,$40
	.dc.b	$E1,$F5,$EF,$94,$E7,$2F,$F5,$FE,$1C,$AF,$67,$81,$7C,$F7,$7F,$9E
	.dc.b	$F0,$36,$77,$5F,$DD,$2B,$71,$8C,$3D,$17,$C2,$31,$28,$C5,$78,$9B
	.dc.b	$8F,$9E,$75,$5C,$3E,$7B,$E0,$7B,$FF,$05,$B1,$93,$AC,$65,$A6,$70
	.dc.b	$F7,$C6,$5B,$1E,$4F,$5F,$E6,$7B,$3F,$35,$B1,$93,$D3,$19,$69,$99
	.dc.b	$F9,$71,$96,$C7,$C0,$F9,$DF,$B1,$F3,$3F,$65,$B1,$93,$E2,$C6,$5A
	.dc.b	$66,$7E,$BC,$65,$B1,$F3,$3E,$77,$F0,$9E,$DF,$E1,$5B,$19,$3E,$8C
	.dc.b	$65,$A6,$67,$F0,$71,$96,$C7,$D8,$F7,$FF,$20,$7E,$47,$E7,$2C,$7D
	.dc.b	$C6,$CF,$E3,$1B,$BF,$DB,$78,$4F,$E0,$BE,$76,$D7,$E4,$BE,$77,$F3
	.dc.b	$87,$E7,$7C,$6B,$5F,$88,$F0,$B0,$D3,$F1,$71,$FA,$3F,$9D,$91,$F8
	.dc.b	$EF,$97,$B8,$33,$FE,$31,$F3,$1F,$CC,$76,$FE,$12,$FF,$33,$E1,$BE
	.dc.b	$47,$FA,$8B,$1F,$CE,$CE,$5B,$19,$C3,$A4,$7F,$BB,$A7,$51,$80,$F7
	.dc.b	$8E,$35,$DA,$FB,$B6,$32,$7E,$A7,$F1,$57,$CE,$FE,$FE,$3F,$8E,$C6
	.dc.b	$1F,$D5,$98,$8D,$3C,$5B,$87,$8F,$AD,$4F,$83,$E7,$F3,$DD,$1B,$87
	.dc.b	$85,$B7,$C6,$B8,$4F,$96,$BC,$7F,$37,$E4,$F9,$BB,$3E,$96,$DF,$16
	.dc.b	$EF,$FC,$4F,$04,$62,$BF,$8F,$C6,$1F,$85,$7D,$5F,$E1,$F6,$F9,$DF
	.dc.b	$1B,$F3,$F0,$7C,$6F,$F0,$88,$FA,$46,$F3,$FE,$8F,$3A,$F0,$7F,$A4
	.dc.b	$CE,$70,$FE,$F7,$87,$48,$FE,$DE,$D7,$FC,$2C,$D7,$FD,$27,$93,$E5
	.dc.b	$BE,$AF,$CC,$7D,$66,$DF,$DB,$6D,$7C,$B6,$E3,$F9,$4E,$4F,$EE,$55
	.dc.b	$F8,$EE,$6F,$C9,$BF,$94,$6B,$93,$F2,$66,$4A,$66,$FF,$25,$CB,$FE
	.dc.b	$0F,$8D,$47,$DB,$F9,$4F,$C5,$FC,$C7,$E3,$1B,$3F,$91,$6C,$EE,$3F
	.dc.b	$29,$3F,$23,$E7,$3F,$C7,$51,$FE,$F7,$FC,$11,$E6,$E7,$D9,$FE,$EB
	.dc.b	$F1,$C9,$CD,$7D,$AB,$E3,$8F,$F8,$70,$38,$7F,$C3,$ED,$7F,$C8,$CD
	.dc.b	$7F,$5A,$0F,$DD,$8F,$85,$AE,$1B,$3F,$BC,$8D,$35,$D3,$4D,$47,$08
	.dc.b	$F9,$67,$03,$F1,$BF,$8C,$BE,$13,$FC,$27,$F8,$C3,$F1,$B7,$C3,$FE
	.dc.b	$3A,$B5,$8F,$F6,$8A,$D4,$E6,$FF,$C5,$98,$7F,$8F,$AF,$6F,$C5,$6B
	.dc.b	$98,$DB,$FE,$69,$FC,$C3,$F2,$58,$7F,$74,$C9,$E7,$BE,$13,$F3,$F5
	.dc.b	$AB,$FE,$2D,$5A,$9C,$F4,$74,$9F,$99,$5B,$7F,$CA,$2D,$7C,$11,$BF
	.dc.b	$0B,$F8,$6F,$CE,$7C,$2F,$C6,$DF,$1F,$FC,$62,$30,$DD,$3E,$14,$E9
	.dc.b	$FE,$41,$3A,$7F,$91,$7C,$BF,$F2,$03,$F9,$0F,$CE,$D8,$C9,$32,$FE
	.dc.b	$2D,$19,$7F,$90,$7C,$03,$DC,$FF,$21,$7C,$D7,$F9,$0D,$B7,$FE,$17
	.dc.b	$9D,$36,$74,$D3,$FC,$39,$E4,$FF,$8C,$9B,$EE,$7F,$94,$23,$18,$7F
	.dc.b	$2F,$7F,$C5,$8D,$F7,$75,$3B,$6D,$74,$6C,$EE,$3A,$99,$2B,$93,$B4
	.dc.b	$F8,$7D,$DF,$17,$93,$6B,$CE,$DA,$F2,$37,$C6,$FF,$22,$63,$FC,$9D
	.dc.b	$FF,$23,$7D,$D3,$6F,$F3,$5A,$E0,$B1,$86,$E3,$FE,$3F,$9D,$36,$74
	.dc.b	$D3,$53,$C3,$1F,$EA,$95,$C1,$FE,$AD,$F7,$77,$0F,$F2,$C2,$F1,$7F
	.dc.b	$CF,$AE,$70,$FF,$32,$B9,$7F,$CA,$0E,$69,$CA,$8D,$47,$6C,$79,$DF
	.dc.b	$46,$DF,$26,$CE,$E3,$AC,$ED,$A7,$F5,$48,$FF,$54,$8F,$A1,$7C,$C7
	.dc.b	$FA,$E3,$AF,$FD,$51,$18,$9F,$F5,$DF,$C4,$7F,$19,$8C,$47,$F1,$B9
	.dc.b	$DD,$7F,$25,$AD,$3C,$EE,$D7,$23,$C1,$19,$69,$9B,$EA,$FF,$19,$6B
	.dc.b	$FE,$F0,$36,$78,$67,$AB,$FE,$F1,$32,$71,$9F,$23,$8B,$F3,$BD,$E3
	.dc.b	$C5,$5D,$AF,$9C,$F1,$3D,$1C,$CE,$5E,$6D,$8C,$C6,$0F,$E7,$CB,$E2
	.dc.b	$DF,$13,$FE,$3D,$8F,$DF,$4E,$A7,$3C,$1F,$D1,$6F,$A5,$71,$BE,$11
	.dc.b	$8A,$DC,$F2,$3F,$A6,$38,$C7,$EA,$AF,$89,$F1,$C6,$1B,$9F,$87,$0F
	.dc.b	$24,$F0,$6C,$79,$67,$77,$D9,$AF,$97,$1F,$A3,$87,$86,$B8,$4E,$2B
	.dc.b	$73,$98,$F3,$DF,$67,$A1,$B3,$9B,$97,$B1,$F8,$08,$F6,$3B,$31,$1C
	.dc.b	$F8,$7F,$55,$D9,$F4,$B1,$FD,$AF,$CB,$5E,$0F,$D9,$B7,$0F,$D8,$F1
	.dc.b	$FF,$42,$7F,$6B,$79,$C3,$F6,$95,$68,$D9,$95,$FB,$C8,$D1,$87,$F8
	.dc.b	$3C,$61,$D1,$D8,$FF,$69,$7D,$DB,$E5,$3C,$0F,$F7,$48,$CF,$8D,$89
	.dc.b	$91,$82,$1F,$E6,$99,$E4,$F4,$8E,$73,$23,$07,$F4,$6F,$87,$F7,$75
	.dc.b	$7F,$72,$8F,$E9,$3E,$2E,$26,$46,$08,$7F,$49,$78,$FC,$3F,$2F,$87
	.dc.b	$93,$FD,$4A,$32,$FF,$54,$8C,$BF,$D4,$63,$2F,$F5,$38,$CF,$07,$F4
	.dc.b	$09,$CB,$FD,$0A,$33,$C1,$FE,$2B,$E0,$FF,$3A,$D7,$AB,$DD,$F5,$FC
	.dc.b	$3C,$C7,$FB,$1C,$65,$FE,$CB,$19,$7F,$B1,$47,$DF,$F6,$1F,$37,$49
	.dc.b	$3F,$E9,$FF,$97,$AF,$FA,$3E,$2E,$9F,$E8,$79,$E5,$F3,$7F,$8D,$F4
	.dc.b	$1F,$0C,$F1,$F0,$DC,$7F,$D5,$55,$C1,$FF,$18,$46,$5F,$F5,$B5,$72
	.dc.b	$AD,$47,$FA,$23,$D7,$70,$FF,$95,$63,$FC,$9D,$18,$70,$D7,$FC,$B4
	.dc.b	$D7,$F8,$B9,$FE,$64,$8F,$D2,$2F,$D6,$5F,$11,$86,$C7,$FA,$7A,$BA
	.dc.b	$69,$FE,$44,$FC,$71,$F2,$7E,$3A,$46,$1F,$33,$5F,$F3,$63,$5F,$1B
	.dc.b	$99,$FF,$27,$F7,$F8,$27,$FD,$4E,$D8,$F0,$57,$A3,$F1,$4F,$9F,$5F
	.dc.b	$EB,$25,$F1,$99,$35,$BC,$9D,$46,$5E,$11,$97,$FD,$6B,$3C,$9D,$9A
	.dc.b	$3F,$D7,$FF,$18,$F9,$AF,$80,$D8,$FA,$8D,$7F,$D1,$4D,$7E,$8B,$9C
	.dc.b	$7F,$CD,$BF,$1D,$FB,$A9,$FF,$62,$B6,$3F,$CE,$33,$BE,$BF,$F6,$9F
	.dc.b	$CE,$FE,$67,$C8,$98,$7F,$BD,$63,$F3,$86,$1F,$F7,$34,$61,$FF,$6B
	.dc.b	$FD,$07,$C8,$7F,$DA,$6F,$BD,$BA,$FE,$12,$35,$8F,$FA,$09,$B8,$7F
	.dc.b	$60,$78,$46,$67,$86,$B2,$FB,$C7,$33,$0F,$FB,$E7,$C6,$FD,$97,$CF
	.dc.b	$C8,$FF,$54,$5F,$FB,$CE,$7F,$6C,$66,$3F,$E0,$04,$E1,$FF,$78,$FD
	.dc.b	$03,$E4,$3F,$24,$D8,$FD,$33,$5F,$F5,$E3,$5F,$D1,$B9,$BF,$F4,$FF
	.dc.b	$D0,$3E,$5B,$F4,$AD,$AE,$F7,$FC,$27,$D3,$FE,$15,$E6,$49,$F4,$EF
	.dc.b	$FC,$5F,$C2,$FF,$60,$46,$0F,$F7,$AD,$6D,$64,$AE,$17,$D6,$78,$78
	.dc.b	$CC,$39,$C7,$0A,$DB,$A8,$FE,$B8,$E4,$EE,$72,$FC,$73,$FD,$88,$EC
	.dc.b	$FE,$44,$7F,$61,$72,$F1,$A4,$63,$0F,$34,$FF,$96,$AB,$F8,$5C,$62
	.dc.b	$BC,$D3,$CF,$E0,$BA,$79,$BD,$7C,$D9,$3E,$38,$CB,$E5,$8C,$BE,$1C
	.dc.b	$7F,$45,$97,$F8,$BB,$67,$D0,$7D,$4E,$B9,$7F,$34,$4E,$9C,$35,$F8
	.dc.b	$86,$A3,$8B,$88,$D6,$9F,$82,$AC,$BF,$84,$AF,$C1,$C7,$C7,$25,$7C
	.dc.b	$D7,$E5,$47,$CB,$FA,$0F,$99,$F4,$12,$BE,$94,$7C,$FF,$85,$91,$F5
	.dc.b	$5D,$3F,$07,$E8,$3E,$A1,$1F,$B5,$F1,$95,$F8,$DF,$41,$F6,$4C,$67
	.dc.b	$F4,$F6,$46,$67,$F0,$4E,$B6,$FE,$D9,$3F,$E5,$89,$FC,$37,$87,$F8
	.dc.b	$33,$F0,$9E,$24,$8C,$90,$FA,$CE,$5F,$B1,$1C,$5F,$C8,$47,$E3,$7E
	.dc.b	$7F,$F1,$DE,$CC,$9F,$CA,$46,$5F,$CB,$46,$5F,$C1,$46,$5F,$C2,$46
	.dc.b	$72,$F7,$4F,$E6,$7C,$8F,$CD,$7C,$FC,$BF,$D0,$46,$5F,$D1,$46,$5F
	.dc.b	$C7,$46,$5F,$C8,$46,$6B,$E5,$4F,$E9,$7B,$7F,$4D,$F3,$F2,$FF,$59
	.dc.b	$19,$7F,$5D,$19,$7F,$35,$19,$7F,$39,$19,$AF,$AD,$3F,$B1,$F7,$7F
	.dc.b	$65,$EF,$4B,$FD,$C4,$65,$FD,$D4,$65,$FD,$34,$65,$FD,$44,$66,$BF
	.dc.b	$05,$3F,$E5,$DF,$91,$7C,$3D,$F8,$79,$87,$F0,$E9,$CB,$67,$2F,$F0
	.dc.b	$38,$4F,$39,$FF,$25,$F8,$72,$BF,$8B,$FB,$BF,$CA,$F9,$3F,$C7,$18
	.dc.b	$7F,$87,$72,$FF,$7E,$DE,$EF,$F8,$75,$ED,$FF,$72,$D6,$EF,$FC,$81
	.dc.b	$2F,$3B,$FC,$B3,$F3,$D2,$FB,$C7,$F9,$70,$C3,$D2,$30,$F9,$3C,$AF
	.dc.b	$F3,$0F,$A7,$23,$04,$3C,$6E,$5F,$2C,$7F,$9A,$0C,$67,$FC,$7F,$0F
	.dc.b	$F0,$A6,$3E,$69,$FF,$36,$FC,$F7,$FB,$13,$E1,$1C,$7E,$A8,$C3,$71
	.dc.b	$09,$DA,$F1,$D8,$FF,$77,$9C,$C6,$2F,$9D,$6E,$91,$89,$D4,$F6,$AE
	.dc.b	$35,$CE,$FF,$DB,$3F,$41,$C3,$8F,$FC,$98,$DB,$FF,$23,$87,$0F,$F9
	.dc.b	$29,$AF,$FB,$77,$AD,$F2,$FF,$E5,$5F,$85,$FF,$22,$B6,$33,$3D,$27
	.dc.b	$2F,$6A,$7A,$BF,$D7,$A3,$CA,$D7,$E3,$C6,$BC,$EF,$3C,$E7,$57,$C6
	.dc.b	$7F,$B8,$F8,$7E,$06,$BF,$E8,$98,$C6,$3F,$3E,$31,$E0,$7F,$5C,$E0
	.dc.b	$F9,$B5,$A6,$F8,$19,$AE,$4D,$BF,$AD,$5F,$F0,$03,$0E,$F1,$89,$D6
	.dc.b	$3E,$4E,$FF,$ED,$3C,$BF,$D9,$F0,$78,$18,$C7,$CD,$57,$1E,$B3,$70
	.dc.b	$E5,$5F,$E0,$4B,$E2,$DB,$D4,$62,$FF,$BD,$28,$CD,$E9,$B5,$CF,$2E
	.dc.b	$3E,$07,$26,$C6,$4F,$04,$65,$A7,$8F,$C0,$F0,$9E,$7F,$1D,$CD,$89
	.dc.b	$D1,$33,$FF,$07,$4E,$A3,$1A,$FF,$82,$FE,$83,$F8,$BC,$7F,$88,$9F
	.dc.b	$E2,$2D,$F0,$3E,$24,$62,$FF,$C3,$0F,$06,$F8,$3A,$9D,$BC,$3D,$CE
	.dc.b	$3F,$1B,$70,$EC,$DB,$E6,$DC,$7C,$8D,$9F,$F6,$A6,$28,$C5,$7E,$E6
	.dc.b	$31,$3E,$9F,$85,$F3,$FC,$37,$A1,$B1,$93,$DD,$19,$69,$9A,$F6,$C6
	.dc.b	$5B,$1F,$17,$E1,$BE,$1F,$C0,$7C,$46,$C6,$4F,$95,$19,$69,$99,$F9
	.dc.b	$31,$96,$C7,$D1,$F8,$5F,$9F,$F0,$DF,$41,$B1,$93,$EB,$46,$5A,$66
	.dc.b	$7E,$AC,$65,$B1,$F7,$7C,$6F,$B7,$E4,$FD,$C6,$C6,$4F,$C1,$46,$5A
	.dc.b	$66,$7F,$03,$19,$6C,$7E,$27,$C8,$FC,$39,$F8,$8F,$1E,$C7,$E3,$5B
	.dc.b	$3F,$8A,$6F,$07,$E4,$DE,$13,$F9,$3F,$22,$D7,$E4,$3D,$CF,$C9,$9F
	.dc.b	$94,$F2,$6D,$7E,$67,$C3,$C3,$4F,$CE,$47,$E6,$BE,$62,$47,$E7,$FE
	.dc.b	$82,$E1,$FE,$7B,$F9,$9F,$E5,$BB,$7F,$05,$7F,$97,$F0,$DF,$2F,$FD
	.dc.b	$45,$87,$FA,$3E,$72,$D8,$CD,$F3,$55,$C6,$F9,$36,$B8,$36,$B8,$3F
	.dc.b	$C4,$FD,$B7,$D0,$7E,$CA,$B0,$D7,$FD,$28,$DA,$FF,$46,$4F,$09,$FD
	.dc.b	$E1,$CF,$E7,$7A,$7D,$05,$8F,$F4,$E3,$6B,$8B,$78,$FF,$D2,$16,$8F
	.dc.b	$27,$D0,$79,$7D,$1B,$87,$FA,$95,$B8,$78,$9B,$87,$26,$CF,$FA,$E5
	.dc.b	$B3,$FE,$9F,$7F,$9A,$2F,$9F,$65,$FE,$A4,$8F,$EF,$67,$F7,$C8,$CD
	.dc.b	$7F,$6A,$8E,$13,$B7,$73,$A7,$FB,$7B,$6F,$A3,$6B,$FC,$0F,$F3,$CF
	.dc.b	$AA,$FF,$6B,$47,$FA,$86,$75,$7F,$EB,$5F,$9A,$FA,$DF,$41,$F5,$DB
	.dc.b	$5F,$EC,$76,$CF,$D4,$6E,$3F,$EB,$EE,$3F,$EF,$75,$F8,$DE,$5F,$C7
	.dc.b	$34,$FC,$61,$6F,$F1,$8D,$AE,$5C,$9F,$ED,$D8,$FC,$95,$72,$69,$FE
	.dc.b	$D9,$F9,$6F,$C6,$7B,$1F,$8D,$6C,$64,$FF,$6F,$46,$5A,$7F,$B0,$7E
	.dc.b	$7D,$EE,$FC,$DC,$BF,$C5,$B1,$89,$FA,$4D,$8F,$F0,$EB,$5F,$B2,$1D
	.dc.b	$65,$F7,$5B,$3F,$E3,$8F,$C0,$5F,$81,$3B,$FF,$0B,$5A,$C1,$B7,$E1
	.dc.b	$30,$FD,$A4,$3D,$9F,$DD,$3F,$E4,$28,$CC,$62,$32,$D8,$C9,$CA,$32
	.dc.b	$D3,$2F,$FB,$7B,$E8,$1F,$13,$F0,$E1,$18,$9F,$CA,$36,$3F,$C7,$ED
	.dc.b	$7E,$1B,$9D,$3F,$3C,$D9,$FF,$2C,$DB,$EA,$DC,$3F,$A6,$AD,$61,$FE
	.dc.b	$5E,$FA,$17,$27,$8D,$B8,$FE,$95,$B3,$FE,$50,$6B,$F5,$1C,$E5,$FA
	.dc.b	$F6,$CF,$8B,$CB,$FA,$14,$9C,$BF,$E6,$BE,$E7,$CE,$FF,$8B,$7E,$5D
	.dc.b	$F4,$2F,$8F,$FE,$39,$39,$8C,$65,$FE,$76,$CF,$F8,$7C,$E6,$B7,$39
	.dc.b	$3F,$CF,$77,$D2,$3F,$93,$7D,$0C,$9D,$65,$E3,$FA,$1C,$3F,$C9,$26
	.dc.b	$E3,$FC,$72,$6D,$7F,$96,$7D,$9F,$F1,$EF,$C3,$FF,$8F,$9B,$1F,$E4
	.dc.b	$36,$E1,$FE,$40,$6D,$FF,$A5,$FD,$9F,$F2,$57,$8B,$FE,$27,$8C,$5F
	.dc.b	$FA,$E3,$D9,$CD,$69,$B1,$93,$FD,$7B,$19,$69,$9C,$3F,$D7,$71,$96
	.dc.b	$C6,$6B,$FD,$51,$EC,$F8,$BD,$9F,$1B,$67,$FD,$92,$DA,$FF,$62,$B6
	.dc.b	$93,$FE,$A8,$F8,$17,$D0,$B6,$FF,$33,$3F,$7A,$7E,$7C,$EA,$73,$C5
	.dc.b	$A6,$CE,$AB,$93,$67,$4D,$35,$3C,$31,$61,$E4,$7F,$D7,$1F,$41,$E1
	.dc.b	$73,$5E,$78,$78,$BF,$E7,$B7,$55,$C7,$24,$6A,$BD,$2E,$9E,$55,$C7
	.dc.b	$1E,$4D,$8D,$B7,$0F,$09,$E4,$71,$87,$7B,$ED,$3A,$C3,$C3,$87,$4C
	.dc.b	$3F,$2F,$3F,$ED,$B1,$F4,$2F,$99,$FE,$4F,$DF,$F0,$63,$13,$FC,$AB
	.dc.b	$E1,$FF,$DB,$B1,$88,$FE,$0F,$78,$BF,$F6,$6D,$59,$FE,$D3,$7C,$E7
	.dc.b	$31,$E0,$7B,$D6,$E3,$C9,$7D,$5F,$E5,$91,$97,$FD,$C0,$71,$3A,$B9
	.dc.b	$7F,$DE,$06,$4F,$29,$4E,$31,$D6,$BC,$91,$ED,$34,$E1,$F1,$D7,$99
	.dc.b	$E4,$76,$E4,$72,$FB,$1B,$1C,$DA,$FC,$F6,$FC,$53,$E2,$7F,$1F,$3F
	.dc.b	$EA,$CA,$FF,$07,$CE,$77,$4E,$4B,$19,$70,$1C,$5F,$6B,$EB,$97,$D7
	.dc.b	$97,$C4,$FF,$87,$23,$0D,$CB,$C3,$5C,$1B,$8F,$D5,$BE,$38,$DB,$E0
	.dc.b	$F9,$2F,$83,$6B,$4C,$3E,$A5,$7A,$70,$F2,$57,$09,$F5,$5F,$46,$9E
	.dc.b	$18,$D5,$71,$C3,$E3,$D6,$E3,$CB,$5D,$AB,$C7,$BF,$EC,$BA,$BF,$C7
	.dc.b	$79,$7E,$9F,$E6,$2E,$FF,$BB,$6E,$1F,$C0,$2A,$DF,$18,$FD,$CC,$E3
	.dc.b	$88,$DA,$93,$A2,$1F,$B3,$9F,$EA,$8F,$F0,$28,$E5,$32,$30,$7F,$1E
	.dc.b	$FA,$17,$CB,$F8,$97,$CB,$D3,$11,$27,$84,$65,$E5,$19,$7F,$82,$4F
	.dc.b	$27,$66,$8F,$E5,$BD,$AF,$85,$F8,$2B,$C4,$A6,$22,$4F,$08,$CB,$CA
	.dc.b	$38,$FC,$2F,$2F,$85,$93,$DA,$32,$F8,$23,$2F,$F1,$9B,$4E,$A3,$13
	.dc.b	$AE,$0F,$E5,$D5,$A9,$E8,$73,$21,$CE,$BD,$7E,$E7,$B3,$E1,$64,$FC
	.dc.b	$08,$CB,$F0,$A3,$2F,$8E,$3F,$07,$E3,$BE,$86,$92,$7F,$AA,$C3,$E8
	.dc.b	$71,$1F,$EA,$69,$B9,$7F,$A7,$A7,$9F,$D0,$FE,$43,$C9,$7C,$2B,$C6
	.dc.b	$C3,$71,$EB,$5C,$1F,$F0,$8D,$70,$AD,$99,$8F,$E8,$E3,$E3,$1C,$1F
	.dc.b	$9E,$8C,$36,$33,$1F,$D1,$BE,$87,$C6,$FF,$48,$8C,$BF,$E6,$C3,$89
	.dc.b	$FE,$A3,$72,$FF,$9C,$1A,$FF,$AA,$BE,$85,$F2,$9E,$16,$C7,$F5,$46
	.dc.b	$BE,$8E,$EF,$0B,$EA,$8C,$BF,$E7,$C6,$BD,$67,$FD,$68,$A3,$2F,$F2
	.dc.b	$AF,$A1,$E7,$BF,$FA,$F5,$7C,$5F,$C3,$6C,$FF,$66,$77,$E0,$7F,$B0
	.dc.b	$9B,$1C,$E7,$81,$C7,$82,$47,$18,$CB,$CE,$3E,$87,$C5,$BE,$6E,$92
	.dc.b	$7F,$BA,$FC,$6B,$E7,$31,$1F,$3E,$B8,$3F,$ED,$16,$B9,$AF,$F6,$A7
	.dc.b	$88,$E8,$7D,$0F,$E6,$3C,$89,$7F,$96,$8F,$CD,$98,$7F,$DF,$71,$87
	.dc.b	$FC,$0F,$F4,$3F,$01,$3F,$19,$FC,$6B,$78,$3F,$DD,$0E,$C9,$B7,$E3
	.dc.b	$DD,$70,$7F,$BB,$1D,$6F,$99,$97,$E0,$8F,$D7,$F8,$BF,$B0,$FA,$19
	.dc.b	$3F,$AE,$8C,$BF,$B0,$8C,$BF,$E2,$1D,$3F,$9A,$BF,$E2,$0F,$A1,$7C
	.dc.b	$87,$E8,$5B,$19,$8F,$D2,$7D,$0B,$CF,$D4,$7E,$32,$B7,$5F,$A1,$6B
	.dc.b	$FE,$CE,$6B,$FE,$5B,$BF,$F6,$23,$FC,$13,$CB,$FE,$12,$7F,$0A,$F4
	.dc.b	$64,$64,$87,$A5,$CB,$EB,$8E,$2F,$F0,$E8,$FE,$AB,$F3,$BF,$EC,$48
	.dc.b	$FE,$22,$7F,$16,$33,$F0,$51,$C9,$0F,$F2,$34,$71,$6F,$9B,$D3,$CE
	.dc.b	$91,$FE,$70,$9F,$EA,$6D,$9C,$BC,$A7,$C7,$1F,$2B,$AB,$E5,$F4,$CA
	.dc.b	$FF,$28,$EF,$F9,$D6,$CE,$5F,$E2,$F1,$F9,$CE,$6F,$F3,$94,$23,$83
	.dc.b	$A9,$E1,$3A,$9E,$1C,$DF,$E2,$8A,$E1,$5F,$18,$8F,$AD,$F9,$37,$F2
	.dc.b	$BE,$87,$E5,$FC,$C4,$AF,$81,$1F,$3B,$E8,$64,$7C,$37,$4F,$CA,$FA
	.dc.b	$1F,$A6,$47,$E4,$7C,$65,$7E,$7F,$D0,$FD,$83,$1A,$FF,$A1,$F8,$11
	.dc.b	$99,$FA,$0E,$B5,$FF,$53,$4F,$F9,$26,$7E,$A7,$89,$F8,$23,$F0,$7E
	.dc.b	$2C,$8C,$90,$FB,$EE,$5F,$C0,$C7,$17,$EF,$47,$E3,$3E,$7F,$F1,$BE
	.dc.b	$C4,$9F,$C1,$46,$5F,$C2,$46,$5F,$C5,$C6,$5F,$C6,$C6,$72,$F7,$4F
	.dc.b	$E6,$3C,$8F,$CC,$FC,$FC,$BF,$C7,$46,$5F,$C8,$46,$5F,$CB,$C6,$5F
	.dc.b	$CC,$C6,$6B,$E5,$4F,$E9,$3E,$7B,$F4,$BF,$3F,$2F,$F3,$51,$97,$F3
	.dc.b	$91,$97,$F4,$71,$97,$F4,$B1,$9A,$FA,$D3,$FB,$0E,$FF,$D8,$FB,$B2
	.dc.b	$FF,$4D,$19,$7F,$51,$19,$7F,$5F,$19,$7F,$63,$19,$AF,$C1,$4F,$F9
	.dc.b	$43,$E4,$9F,$0B,$7E,$1A,$61,$FE,$E2,$9C,$B6,$72,$FF,$B0,$61,$3C
	.dc.b	$E7,$FC,$77,$E1,$CA,$FF,$1F,$BD,$FD,$2C,$47,$F8,$C6,$2E,$3F,$ED
	.dc.b	$BC,$3F,$C2,$53,$FE,$35,$F3,$BF,$C8,$BF,$3D,$23,$FC,$93,$1F,$E4
	.dc.b	$83,$0F,$F8,$EE,$30,$FF,$93,$BC,$AF,$F2,$6F,$C8,$C8,$C1,$0F,$F2
	.dc.b	$13,$97,$FC,$89,$1F,$E5,$43,$18,$7F,$90,$A7,$13,$FE,$5A,$F9,$EF
	.dc.b	$E0,$5F,$08,$E2,$FF,$55,$FA,$2F,$79,$5A,$9C,$C6,$2F,$FD,$63,$7C
	.dc.b	$55,$F1,$F3,$F8,$46,$2B,$82,$AE,$15,$DF,$0D,$B6,$7A,$E1,$E0,$C3
	.dc.b	$9C,$3E,$87,$CB,$C9,$FE,$2F,$8C,$61,$FD,$55,$FF,$19,$4E,$9B,$19
	.dc.b	$7F,$91,$1F,$EC,$5E,$C7,$CB,$73,$F8,$5F,$F1,$9B,$63,$33,$FE,$3D
	.dc.b	$9C,$BD,$BD,$6F,$EB,$6F,$4C,$11,$88,$FE,$61,$18,$F3,$BD,$0D,$8F
	.dc.b	$1E,$4F,$C7,$FF,$CA,$7E,$1F,$C2,$9F,$95,$82,$31,$5B,$F6,$F1,$5F
	.dc.b	$37,$7F,$AF,$78,$E1,$FA,$D7,$BC,$38,$E3,$DB,$0E,$07,$DA,$FC,$71
	.dc.b	$F2,$3C,$72,$FF,$77,$E1,$C6,$B6,$7F,$AC,$B8,$BC,$3C,$DF,$E9,$0F
	.dc.b	$A1,$B9,$FF,$4F,$F4,$7F,$CF,$2F,$FA,$AA,$33,$7F,$EF,$A9,$CB,$FE
	.dc.b	$9D,$8C,$BD,$67,$81,$FF,$00,$2B,$FB,$0D,$70,$6D,$F4,$E6,$68,$EC
	.dc.b	$E7,$7B,$C1,$8B,$E7,$5F,$86,$BD,$5F,$2C,$61,$E9,$8B,$E3,$B3,$5B
	.dc.b	$C3,$C8,$74,$BE,$0E,$30,$ED,$3B,$5A,$FD,$8D,$3C,$F8,$74,$CB,$BF
	.dc.b	$C7,$7C,$6C,$5E,$63,$E7,$F3,$1D,$4C,$B8,$CB,$AC,$7B,$8C,$9E,$3C
	.dc.b	$7E,$2D,$F5,$BE,$93,$E7,$AF,$93,$C7,$F2,$F8,$3F,$DB,$D3,$9D,$FF
	.dc.b	$84,$3F,$CB,$FE,$33,$EB,$C6,$2F,$F9,$74,$62,$F9,$3F,$73,$C6,$FF
	.dc.b	$7C,$DF,$F3,$7F,$17,$1C,$7D,$DB,$19,$3C,$93,$C3,$0F,$E7,$53,$A7
	.dc.b	$FD,$77,$19,$E1,$FE,$69,$E1,$E5,$FD,$2F,$27,$DE,$F4,$FD,$9E,$2F
	.dc.b	$DA,$6E,$5F,$5E,$74,$FF,$4B,$9D,$34,$D7,$32,$B3,$97,$F4,$DE,$7B
	.dc.b	$8F,$82,$BF,$D7,$B7,$FD,$57,$C4,$FC,$4F,$87,$F3,$5B,$19,$3A,$C6
	.dc.b	$5A,$4C,$3F,$AD,$77,$F0,$7E,$F9,$F8,$EE,$2F,$B1,$D7,$F9,$66,$D7
	.dc.b	$A1,$B3,$C3,$6F,$D0,$C6,$33,$F4,$3C,$6B,$FB,$67,$CB,$DB,$E1,$F8
	.dc.b	$53,$E5,$3F,$0D,$5F,$A8,$8C,$3F,$DC,$CC,$46,$9F,$9D,$1A,$9F,$EE
	.dc.b	$91,$A3,$33,$FB,$BF,$D4,$DF,$49,$9A,$F7,$F6,$A5,$F4,$9B,$3C,$5F
	.dc.b	$21,$71,$EE,$DB,$F4,$B7,$5F,$D9,$5F,$E8,$A7,$35,$FA,$78,$C4,$FE
	.dc.b	$2B,$35,$F9,$7E,$93,$CD,$E6,$DC,$38,$3F,$0F,$A3,$19,$7C,$86,$E3
	.dc.b	$E0,$F1,$BD,$DF,$45,$EF,$6E,$1F,$B3,$FC,$27,$FB,$9B,$FC,$2F,$CD
	.dc.b	$7E,$12,$F5,$BF,$93,$D6,$B3,$FC,$1D,$6A,$73,$D7,$FC,$C2,$31,$87
	.dc.b	$58,$C6,$1C,$E3,$07,$77,$17,$F8,$E1,$87,$7C,$FF,$A8,$E1,$D2,$7E
	.dc.b	$AF,$0F,$F2,$7C,$3E,$57,$95,$F6,$FC,$6F,$B8,$C8,$C9,$F3,$E3,$15
	.dc.b	$EA,$FC,$C1,$D9,$E9,$6D,$7C,$37,$FB,$7B,$63,$44,$75,$DB,$6B,$D5
	.dc.b	$F4,$DF,$72,$7F,$A6,$7D,$67,$E4,$F5,$FF,$56,$CE,$6B,$F9,$BC,$E6
	.dc.b	$BF,$D5,$53,$9A,$4E,$6B,$FA,$6C,$E6,$31,$08,$FC,$73,$FE,$12,$6B
	.dc.b	$D3,$CC,$FF,$5E,$B6,$FF,$C2,$EF,$F5,$7F,$96,$FF,$1B,$47,$F9,$7F
	.dc.b	$A7,$FA,$BC,$62,$7F,$CB,$D1,$8E,$8F,$E7,$4D,$AE,$07,$E7,$7E,$7F
	.dc.b	$F5,$C7,$EB,$FE,$7A,$C6,$49,$7F,$72,$32,$FC,$AF,$92,$FC,$67,$D1
	.dc.b	$FF,$9D,$2B,$FA,$56,$89,$CF,$B9,$FD,$F2,$31,$3F,$9B,$AE,$17,$CE
	.dc.b	$31,$5B,$BE,$31,$D1,$B3,$F8,$ED,$3F,$83,$1F,$C1,$FE,$66,$D7,$EE
	.dc.b	$62,$CF,$E5,$26,$F0,$7F,$BA,$BE,$95,$F0,$5E,$AF,$9D,$FF,$08,$C6
	.dc.b	$31,$F3,$46,$27,$51,$FD,$9E,$75,$87,$EC,$34,$5F,$F8,$6B,$BB,$FC
	.dc.b	$21,$2E,$B5,$3D,$FD,$79,$DF,$29,$B9,$7C,$56,$DF,$56,$E9,$FB,$0B
	.dc.b	$E5,$5F,$DE,$71,$EF,$8F,$6B,$FF,$15,$FD,$23,$E1,$7F,$81,$7D,$CF
	.dc.b	$F2,$8C,$3F,$C0,$13,$97,$F8,$74,$67,$3F,$E0,$13,$97,$F8,$6C,$67
	.dc.b	$3F,$F2,$8C,$27,$51,$FC,$12,$75,$5F,$E3,$0A,$79,$BF,$E1,$C8,$FE
	.dc.b	$39,$EB,$FC,$AB,$E7,$39,$AF,$E3,$73,$9A,$E5,$39,$AF,$E2,$33,$9A
	.dc.b	$FF,$1E,$CE,$6B,$FC,$7F,$9A,$B5,$E8,$39,$1C,$9F,$E1,$FE,$2F,$D5
	.dc.b	$B7,$0F,$F0,$24,$E9,$FF,$67,$4E,$9A,$7F,$47,$F9,$8F,$EB,$1F,$3B
	.dc.b	$FD,$65,$B1,$93,$FA,$64,$65,$A7,$6E,$7F,$F2,$D7,$6F,$C6,$8C,$1F
	.dc.b	$E3,$97,$1B,$7F,$79,$C7,$F5,$6F,$F6,$48,$CE,$FF,$E2,$09,$CB,$FE
	.dc.b	$4F,$8C,$D7,$F9,$5E,$5F,$2F,$E0,$8F,$E1,$6D,$7E,$0B,$6F,$C0,$DB
	.dc.b	$EC,$F0,$9F,$F2,$77,$B6,$EE,$C5,$F0,$9E,$66,$A3,$37,$FE,$9A,$F7
	.dc.b	$DF,$92,$BE,$47,$FC,$E3,$EA,$FF,$A2,$27,$31,$8C,$3F,$CD,$D1,$87
	.dc.b	$C3,$3B,$AE,$51,$83,$FC,$7D,$F8,$5F,$E1,$FD,$67,$D2,$F9,$E7,$3E
	.dc.b	$17,$FA,$2E,$31,$AF,$F8,$9E,$F8,$46,$34,$FF,$26,$C6,$35,$FF,$44
	.dc.b	$B7,$0E,$67,$F9,$C6,$B8,$3F,$E8,$7F,$AC,$FD,$2E,$1F,$EE,$AF,$A5
	.dc.b	$78,$FF,$E9,$29,$CE,$89,$CD,$FF,$95,$E7,$31,$8B,$FF,$1E,$46,$2F
	.dc.b	$97,$B3,$2F,$C5,$CD,$6F,$73,$75,$FF,$4A,$67,$F8,$8B,$F3,$4F,$FA
	.dc.b	$D2,$BF,$E0,$08,$FA,$67,$4F,$FB,$B3,$BB,$FD,$A7,$18,$9F,$F7,$7F
	.dc.b	$C4,$7F,$AC,$23,$11,$FB,$4C,$31,$7F,$D7,$AA,$CF,$F0,$7B,$E7,$39
	.dc.b	$8F,$03,$DE,$B7,$1F,$C3,$6A,$BF,$CB,$E2,$A8,$CD,$74,$8F,$29,$D0
	.dc.b	$F0,$4F,$09,$DB,$63,$47,$A4,$A7,$C4,$BF,$6B,$E1,$32,$78,$47,$2B
	.dc.b	$E3,$FD,$8D,$8F,$1B,$5F,$61,$C2,$3F,$95,$7C,$67,$FB,$13,$5F,$F7
	.dc.b	$AC,$63,$E1,$FF,$D9,$B5,$B3,$FD,$BF,$5B,$70,$1C,$5F,$77,$58,$7D
	.dc.b	$65,$F2,$3F,$CE,$63,$0D,$D7,$C1,$97,$D2,$AE,$31,$E6,$9D,$61,$E2
	.dc.b	$6B,$E1,$C7,$CF,$87,$76,$D7,$19,$E9,$5C,$2F,$AC,$7B,$2B,$6F,$53
	.dc.b	$67,$8D,$CB,$DF,$0C,$BF,$6A,$7D,$AE,$E3,$F2,$3E,$13,$8F,$F6,$73
	.dc.b	$FA,$FE,$3F,$EB,$55,$FE,$CC,$9C,$CF,$FA,$53,$8B,$F9,$5B,$B2,$1F
	.dc.b	$B4,$AE,$3B,$A1,$CB,$FC,$4B,$B7,$30,$CB,$C5,$78,$AF,$E0,$55,$79
	.dc.b	$BB,$DF,$53,$BE,$6C,$7B,$5F,$09,$F0,$4E,$5F,$E2,$17,$E0,$C7,$BC
	.dc.b	$7F,$B8,$FF,$0B,$FF,$4F,$BC,$30,$E7,$F4,$EF,$8C,$F2,$36,$78,$7A
	.dc.b	$5C,$EF,$8C,$7F,$AE,$27,$4D,$AD,$54,$8E,$C7,$C4,$9D,$E5,$E7,$CD
	.dc.b	$18,$C7,$E5,$64,$CB,$BD,$FF,$4B,$EE,$FF,$7B,$63,$FD,$53,$AB,$E5
	.dc.b	$78,$3F,$EC,$88,$FE,$93,$3A,$7F,$DD,$7D,$1F,$D4,$BE,$1F,$77,$27
	.dc.b	$71,$F1,$FF,$08,$F1,$47,$D7,$6E,$67,$3B,$32,$78,$DC,$78,$66,$1F
	.dc.b	$03,$0E,$11,$E6,$9C,$CF,$FB,$82,$9E,$AF,$A8,$CE,$DF,$6F,$0E,$35
	.dc.b	$FD,$CA,$74,$F5,$CB,$CA,$3E,$33,$E1,$F9,$12,$75,$1E,$D3,$0F,$86
	.dc.b	$30,$FF,$B9,$7C,$2F,$CA,$7B,$BF,$92,$CB,$FD,$D7,$CF,$89,$95,$FE
	.dc.b	$EB,$8D,$98,$53,$E9,$C7,$FC,$01,$1F,$82,$DD,$EE,$3E,$3F,$F4,$4D
	.dc.b	$9F,$67,$95,$F4,$CC,$CC,$8C,$9F,$8C,$1F,$AD,$7F,$94,$DF,$F6,$E7
	.dc.b	$EB,$7F,$EE,$B7,$4D,$7F,$DC,$11,$BC,$BF,$CB,$ED,$EA,$FF,$2E,$56
	.dc.b	$E7,$FC,$C6,$D8,$FF,$30,$4E,$EB,$F8,$06,$8F,$A0,$FD,$FC,$3F,$5C
	.dc.b	$FF,$54,$3E,$76,$BF,$EF,$19,$E0,$FA,$DA,$FF,$BD,$63,$78,$FF,$9D
	.dc.b	$1B,$87,$F9,$C6,$B7,$3F,$E7,$8F,$B7,$FF,$76,$DF,$C5,$FC,$29,$F0
	.dc.b	$9E,$46,$C7,$C9,$9D,$BF,$63,$0F,$F3,$E4,$62,$FF,$D0,$75,$C2,$B4
	.dc.b	$7F,$1F,$71,$1E,$1F,$83,$F8,$35,$C6,$FF,$D0,$D5,$A6,$E1,$AA,$F8
	.dc.b	$FF,$8F,$3E,$9F,$E4,$FC,$AF,$FA,$77,$3F,$BC,$D7,$C2,$27,$83,$F8
	.dc.b	$36,$BA,$8E,$19,$FF,$A7,$1B,$8F,$FA,$66,$B7,$3F,$C4,$F9,$FE,$5C
	.dc.b	$FF,$2E,$6D,$FF,$3C,$6F,$29,$E3,$3B,$E0,$FD,$49,$2B,$F4,$13,$A7
	.dc.b	$FA,$37,$C3,$3E,$47,$F0,$FF,$72,$F6,$29,$F8,$A9,$D5,$7E,$7D,$AF
	.dc.b	$C2,$9E,$0F,$E8,$DA,$FC,$68,$DE,$7F,$EB,$C6,$E5,$FE,$B9,$AD,$CF
	.dc.b	$FB,$01,$B1,$FE,$BD,$9D,$D7,$C1,$3F,$82,$90,$F6,$FA,$B5,$FC,$2E
	.dc.b	$9F,$15,$C6,$1F,$36,$7F,$97,$7A,$1F,$C0,$3E,$BA,$57,$F1,$B8,$CB
	.dc.b	$FC,$1F,$B1,$3A,$8F,$C8,$F8,$5F,$CC,$EB,$F5,$FF,$52,$F9,$DF,$E0
	.dc.b	$55,$FE,$D5,$8C,$38,$0E,$1D,$B5,$FE,$14,$E7,$2F,$F7,$1F,$D3,$FF
	.dc.b	$01,$9F,$EF,$CD,$79,$D7,$88,$FE,$03,$19,$35,$5C,$F6,$FF,$6C,$FD
	.dc.b	$CB,$EA,$7F,$C0,$BB,$23,$18,$FF,$68,$8C,$4E,$A7,$FA,$AC,$62,$75
	.dc.b	$1C,$1F,$DA,$C6,$5B,$1F,$D6,$3F,$54,$FE,$8D,$1F,$CC,$DA,$FD,$86
	.dc.b	$E7,$FD,$D5,$E8,$DF,$1F,$FC,$31,$1A,$C7,$FC,$26,$DA,$FF,$10,$BC
	.dc.b	$23,$E9,$71,$7C,$4F,$12,$57,$F3,$9D,$13,$F6,$8C,$C7,$EF,$B5,$FE
	.dc.b	$92,$FF,$00,$9D,$BC,$67,$79,$7D,$6B,$E5,$3E,$DF,$A9,$F3,$FD,$54
	.dc.b	$8C,$1F,$DA,$7D,$0F,$5D,$48,$C1,$F1,$5F,$7F,$D5,$48,$C1,$F2,$7D
	.dc.b	$0F,$F1,$77,$EB,$AF,$85,$A7,$C6,$8B,$3F,$3F,$EA,$BE,$49,$87,$E8
	.dc.b	$38,$74,$7F,$82,$6B,$A3,$7E,$37,$B9,$1F,$B5,$F5,$75,$FB,$5F,$57
	.dc.b	$F5,$8C,$6F,$FE,$0B,$B5,$E6,$BF,$05,$5A,$E3,$FF,$18,$D7,$CE,$9F
	.dc.b	$D7,$B6,$F9,$70,$FE,$0B,$E8,$7E,$0B,$F8,$4E,$B9,$7E,$98,$CB,$F8
	.dc.b	$F8,$CB,$D2,$3F,$CA,$DF,$79,$FE,$39,$8F,$DE,$B6,$FF,$C4,$4D,$C7
	.dc.b	$F1,$FF,$07,$C3,$DA,$FC,$9F,$C8,$CC,$3F,$C6,$17,$B7,$FC,$E0,$DF
	.dc.b	$0D,$D7,$FD,$0A,$7F,$2F,$5C,$DE,$3F,$3D,$E4,$23,$FA,$7F,$8C,$AF
	.dc.b	$E9,$FE,$AF,$F4,$66,$35,$F3,$D6,$27,$F5,$D3,$9D,$7F,$D0,$D0,$FC
	.dc.b	$71,$E1,$7F,$82,$1B,$1F,$E1,$D6,$BF,$B5,$F2,$B5,$3F,$B2,$F9,$59
	.dc.b	$3B,$8C,$BF,$BD,$8C,$BF,$DA,$2B,$9E,$5F,$E8,$78,$FF,$2D,$72,$7F
	.dc.b	$8C,$BE,$D2,$47,$F8,$B7,$DF,$4F,$F8,$71,$B1,$FC,$D6,$76,$FF,$74
	.dc.b	$8C,$BF,$E5,$DF,$77,$FD,$43,$A3,$B1,$F1,$78,$6D,$7F,$91,$6F,$FC
	.dc.b	$3A,$FF,$AA,$9B,$5C,$6B,$81,$FE,$80,$BF,$C4,$FC,$8C,$BF,$C0,$BF
	.dc.b	$E8,$4F,$6B,$11,$F8,$C8,$BA,$7F,$A4,$71,$FF,$2F,$63,$F5,$3F,$25
	.dc.b	$71,$7F,$8E,$9B,$87,$F9,$3D,$AF,$F9,$E7,$D7,$D4,$FF,$A4,$BE,$92
	.dc.b	$4E,$E3,$2F,$B6,$32,$FF,$86,$AB,$9E,$3F,$EB,$9F,$1D,$F9,$23,$E5
	.dc.b	$3F,$0F,$F2,$7F,$E0,$29,$CC,$63,$A3,$FD,$A1,$18,$C7,$FD,$A5,$5F
	.dc.b	$E0,$8E,$DF,$F4,$B7,$B3,$B9,$56,$D5,$6F,$81,$D7,$BC,$3F,$D3,$CF
	.dc.b	$1B,$E1,$5F,$E1,$2F,$AD,$7D,$27,$4C,$7F,$B8,$C6,$2B,$FB,$83,$63
	.dc.b	$33,$B7,$FC,$F1,$F2,$7F,$8E,$FA,$EF,$EE,$CD,$8F,$EF,$6D,$77,$D8
	.dc.b	$9E,$BE,$37,$E9,$A7,$F5,$B8,$7F,$4C,$F0,$32,$48,$FF,$43,$64,$FA
	.dc.b	$EF,$F4,$97,$4F,$C8,$E2,$F9,$71,$8A,$F8,$B1,$FE,$F0,$E1,$7D,$3F
	.dc.b	$FB,$BF,$83,$EA,$5E,$DB,$F1,$39,$BE,$6D,$C3,$F3,$97,$72,$F1,$4E
	.dc.b	$67,$FD,$75,$F4,$1F,$B5,$7F,$83,$FB,$08,$C4,$0D,$E9,$FE,$1D,$9D
	.dc.b	$36,$4F,$FA,$9E,$33,$3F,$ED,$1F,$7B,$2F,$F9,$B2,$33,$D5,$FC,$15
	.dc.b	$B5,$E2,$BD,$CF,$EA,$6E,$CE,$4F,$F7,$07,$D5,$BE,$43,$38,$BD,$F4
	.dc.b	$E6,$BF,$80,$7C,$3E,$A7,$37,$6B,$57,$FE,$40,$6E,$3F,$B9,$BD,$57
	.dc.b	$6E,$BF,$F7,$A7,$D7,$78,$32,$7A,$5F,$EE,$E7,$F8,$74,$66,$31,$97
	.dc.b	$FB,$CE,$31,$F2,$1C,$F9,$91,$8B,$FE,$49,$C9,$F4,$AD,$5F,$E7,$B9
	.dc.b	$FF,$27,$36,$3F,$CA,$6D,$7F,$DA,$CE,$74,$FE,$64,$D9,$FB,$99,$7F
	.dc.b	$B9,$BD,$6F,$93,$7D,$B0,$BF,$5D,$FE,$FF,$BF,$F3,$37,$C4,$E3,$93
	.dc.b	$E1,$71,$A3,$13,$FE,$61,$AE,$18,$7F,$99,$9A,$FF,$BA,$5A,$FF,$56
	.dc.b	$77,$A7,$F4,$76,$CF,$9A,$DF,$5D,$DA,$FE,$B7,$B7,$E2,$FA,$CF,$B4
	.dc.b	$DB,$FC,$2B,$F9,$AE,$6F,$C1,$B7,$98,$D2,$4F,$03,$FD,$09,$CF,$D1
	.dc.b	$FC,$7F,$83,$F0,$3C,$3F,$C4,$36,$32,$7F,$A1,$30,$FC,$8C,$77,$24
	.dc.b	$67,$0F,$EB,$EF,$0C,$BF,$D3,$7A,$7F,$95,$A3,$1B,$7E,$A3,$0F,$CD
	.dc.b	$F0,$FD,$0F,$8C,$FC,$C3,$6B,$F4,$6F,$B4,$8F,$76,$E1,$F9,$D7,$EC
	.dc.b	$FD,$17,$29,$FD,$F6,$1D,$FE,$B5,$F5,$DF,$7A,$7F,$51,$18,$7F,$D5
	.dc.b	$06,$23,$E7,$46,$70,$FF,$08,$9F,$B3,$70,$FF,$85,$32,$46,$23,$F8
	.dc.b	$74,$62,$7A,$57,$FB,$73,$E3,$6D,$7F,$B0,$3E,$39,$EF,$7E,$CD,$B5
	.dc.b	$EA,$39,$E2,$F9,$BD,$57,$A9,$F3,$92,$7E,$C4,$ED,$FF,$5E,$F5,$FF
	.dc.b	$BC,$7B,$B8,$47,$FB,$B9,$B8,$F8,$FD,$1F,$17,$CB,$F8,$DB,$5E,$39
	.dc.b	$F8,$5D,$3F,$18,$91,$E0,$7F,$97,$E1,$FE,$FF,$F1,$BD,$9E,$77,$B5
	.dc.b	$B5,$CE,$DF,$5B,$FD,$92,$3F,$D7,$5F,$5C,$F7,$BF,$77,$39,$AF,$F5
	.dc.b	$BC,$E6,$B9,$4E,$67,$EB,$B5,$FF,$24,$B5,$FE,$84,$D7,$C3,$D9,$D8
	.dc.b	$FE,$63,$E6,$AD,$F4,$79,$C3,$E1,$68,$E4,$79,$5F,$64,$FB,$5D,$B6
	.dc.b	$B7,$3A,$6C,$E9,$A7,$FB,$67,$EB,$9D,$DE,$26,$C6,$7D,$1C,$4C,$8E
	.dc.b	$0E,$E3,$1B,$3B,$7F,$CB,$1F,$5B,$FC,$E1,$FE,$A7,$F5,$DF,$9D,$73
	.dc.b	$D3,$FD,$41,$46,$71,$FF,$38,$D7,$F4,$0B,$FD,$1B,$5C,$DF,$F3,$D8
	.dc.b	$C4,$7F,$6F,$7C,$13,$A3,$F1,$5D,$EF,$BC,$FB,$55,$E7,$9D,$34,$FE
	.dc.b	$C5,$EE,$BB,$3F,$9B,$F7,$7F,$61,$8C,$57,$D4,$E2,$FD,$05,$F1,$C5
	.dc.b	$5A,$8C,$37,$5F,$EF,$7C,$7F,$6F,$CE,$E9,$9F,$EC,$FD,$99,$9F,$DA
	.dc.b	$FA,$4B,$97,$EE,$66,$E3,$CB,$99,$58,$BF,$93,$F0,$1F,$8C,$FA,$D7
	.dc.b	$D7,$FF,$56,$E3,$4E,$63,$0F,$CB,$25,$7F,$12,$CF,$E6,$46,$1F,$8C
	.dc.b	$2B,$9C,$FE,$DF,$B1,$87,$DB,$AE,$33,$F9,$4E,$0F,$E2,$07,$F1,$1F
	.dc.b	$AB,$B1,$92,$4F,$F1,$78,$CB,$E0,$FA,$F7,$C4,$FB,$7E,$47,$FB,$8C
	.dc.b	$62,$FC,$58,$FF,$BF,$DF,$E0,$11,$9B,$E5,$7F,$17,$DA,$E2,$FF,$7A
	.dc.b	$F1,$3F,$C5,$73,$FE,$C0,$C3,$FD,$E7,$87,$F0,$99,$CE,$9E,$EF,$5F
	.dc.b	$87,$A5,$FE,$0F,$9F,$F1,$CE,$1F,$CE,$9A,$FF,$B3,$1A,$FF,$91,$5E
	.dc.b	$EF,$F1,$18,$FB,$1E,$B7,$4C,$3F,$8C,$7D,$7B,$E3,$3F,$83,$FD,$7F
	.dc.b	$F3,$8C,$3E,$9C,$E6,$BE,$94,$E7,$A3,$FD,$95,$39,$AD,$DF,$2F,$07
	.dc.b	$FC,$33,$39,$AF,$F6,$87,$A7,$FE,$34,$C7,$EE,$7D,$7F,$C9,$9C,$FB
	.dc.b	$BF,$C6,$23,$13,$FE,$E8,$AE,$19,$7E,$19,$B7,$FE,$5A,$E0,$5F,$FA
	.dc.b	$EA,$FF,$9A,$56,$9F,$EB,$7C,$EE,$7F,$F0,$93,$63,$FC,$0C,$DC,$7F
	.dc.b	$B3,$F8,$3F,$B0,$AD,$9F,$B7,$3F,$AF,$7D,$3D,$BF,$E4,$BF,$19,$86
	.dc.b	$9F,$CA,$23,$FB,$24,$66,$FF,$B4,$37,$4F,$F0,$66,$3F,$A0,$9F,$0B
	.dc.b	$5F,$EE,$CD,$7D,$7D,$9E,$A3,$F8,$51,$FE,$59,$EE,$B8,$FF,$16,$A7
	.dc.b	$C6,$7C,$02,$47,$C0,$6D,$FF,$0C,$7F,$CC,$5E,$0F,$F1,$56,$4F,$37
	.dc.b	$2D,$24,$F1,$7F,$DF,$93,$C2,$3F,$91,$7C,$13,$EA,$1F,$31,$FE,$9C
	.dc.b	$8C,$36,$33,$7F,$C1,$E7,$2D,$8C,$D7,$46,$E9,$FE,$FD,$DF,$FD,$23
	.dc.b	$7C,$2B,$F9,$57,$D7,$FE,$31,$79,$0F,$AD,$FF,$78,$65,$FC,$F2,$B5
	.dc.b	$39,$D3,$FD,$2D,$1F,$BF,$BF,$F1,$43,$5F,$F1,$3B,$5F,$F5,$03,$78
	.dc.b	$CF,$85,$BE,$0F,$FA,$67,$1F,$C8,$69,$FE,$4D,$AD,$7D,$77,$EB,$F3
	.dc.b	$FF,$42,$FD,$77,$F3,$47,$FA,$8C,$67,$2F,$5C,$E5,$FE,$91,$1F,$BB
	.dc.b	$7F,$C7,$CD,$7F,$C7,$AD,$7F,$D6,$CD,$C8,$F5,$6F,$BB,$E7,$E5,$FF
	.dc.b	$5C,$4E,$49,$19,$C3,$CA,$FF,$A6,$AF,$FD,$B1,$F1,$3F,$ED,$9F,$AF
	.dc.b	$74,$7F,$95,$3B,$3F,$DA,$71,$8A,$FF,$2B,$D6,$EB,$58,$59,$C9,$FE
	.dc.b	$1C,$8C,$B4,$CF,$AF,$CD,$FF,$74,$61,$FE,$E0,$8C,$47,$90,$61,$89
	.dc.b	$F1,$D7,$F5,$7C,$3C,$15,$CB,$0A,$F1,$C3,$84,$7F,$BD,$5E,$B0,$9C
	.dc.b	$57,$9A,$73,$18,$9F,$F7,$E0,$BF,$35,$78,$63,$D8,$78,$4A,$7A,$79
	.dc.b	$3F,$C7,$AF,$C4,$6B,$D2,$8A,$32,$F1,$BF,$0F,$C7,$7F,$BA,$38,$77
	.dc.b	$5C,$A3,$15,$C6,$B7,$DF,$FE,$35,$8C,$1D,$A7,$75,$C6,$75,$5F,$4E
	.dc.b	$5F,$33,$F0,$23,$0D,$CB,$AE,$3F,$17,$15,$F8,$AF,$71,$98,$E6,$DC
	.dc.b	$3B,$DF,$28,$FB,$AF,$BB,$0F,$9F,$AF,$29,$C9,$E9,$AF,$64,$79,$DE
	.dc.b	$71,$ED,$9F,$33,$4F,$06,$31,$E6,$7A,$23,$BD,$F8,$63,$F1,$78,$7C
	.dc.b	$CC,$3F,$41,$C3,$FB,$0E,$DF,$C1,$E3,$FE,$CF,$E1,$FF,$19,$E1,$FC
	.dc.b	$02,$73,$C1,$F7,$3A,$3F,$8D,$76,$62,$BF,$77,$57,$1E,$91,$C4,$FF
	.dc.b	$40,$F3,$FF,$44,$70,$7F,$60,$C7,$BB,$6B,$FC,$70,$F4,$3F,$DD,$91
	.dc.b	$AC,$BF,$CA,$3C,$DC,$E7,$FC,$C1,$92,$31,$3E,$BB,$F2,$7D,$77,$FB
	.dc.b	$0E,$FF,$CD,$6D,$7F,$CC,$0D,$9F,$F2,$EE,$3F,$ED,$A7,$FD,$E5,$1B
	.dc.b	$9F,$F6,$D3,$5F,$F6,$CB,$56,$48,$C3,$7D,$4F,$91,$93,$EB,$DF,$1D
	.dc.b	$86,$E1,$E6,$AF,$9D,$CD,$FD,$6F,$E2,$AF,$80,$69,$D3,$E3,$BF,$CF
	.dc.b	$37,$FE,$D1,$AE,$0F,$F5,$B9,$F9,$B1,$88,$FF,$7B,$B5,$FF,$7B,$7D
	.dc.b	$7F,$89,$FF,$42,$37,$67,$27,$FB,$E2,$30,$FE,$17,$07,$8B,$F8,$39
	.dc.b	$CD,$FC,$46,$EE,$73,$BA,$F8,$D7,$FA,$82,$76,$4D,$FD,$D8,$75,$7F
	.dc.b	$0F,$87,$36,$CF,$A6,$BD,$71,$8F,$8F,$F2,$12,$3F,$DA,$B8,$7F,$39
	.dc.b	$CF,$DB,$1E,$79,$E0,$FF,$A2,$F0,$F4,$46,$23,$F9,$59,$FD,$FB,$D7
	.dc.b	$FF,$4A,$BF,$ED,$AE,$7F,$BF,$1F,$94,$D7,$F5,$4E,$A7,$F3,$2D,$C3
	.dc.b	$FD,$49,$5C,$0E,$3E,$0F,$CB,$F1,$A5,$FC,$57,$4D,$3C,$DA,$FF,$88
	.dc.b	$1B,$1F,$10,$FD,$6A,$FA,$EF,$CF,$65,$FD,$11,$B1,$FC,$E9,$AB,$EB
	.dc.b	$FF,$9E,$3F,$CF,$B2,$F8,$EF,$33,$F8,$21,$BF,$4A,$C9,$CD,$74,$CD
	.dc.b	$F2,$BC,$1B,$7D,$71,$FE,$AF,$18,$8F,$F5,$CF,$D7,$BE,$6B,$85,$F8
	.dc.b	$23,$E5,$5F,$F5,$E6,$D7,$F5,$4B,$E5,$9F,$A3,$EB,$DF,$09,$E8,$6D
	.dc.b	$78,$4F,$EB,$6E,$1F,$F6,$4F,$D7,$FF,$13,$C7,$FB,$5C,$E6,$7F,$06
	.dc.b	$D7,$35,$F4,$23,$11,$D2,$3E,$3D,$3E,$1E,$9E,$38,$F9,$7F,$5F,$EF
	.dc.b	$AF,$20,$70,$FF,$B7,$1B,$8F,$D4,$F6,$5F,$3B,$CB,$7F,$AB,$1F,$EB
	.dc.b	$3B,$E6,$D7,$94,$7E,$63,$79,$1F,$0F,$B1,$CF,$F2,$B7,$FC,$DD,$EA
	.dc.b	$BF,$37,$77,$2F,$F7,$F4,$FE,$13,$E2,$1F,$29,$49,$7E,$CF,$AF,$7B
	.dc.b	$78,$8F,$F7,$D5,$70,$7F,$A2,$B5,$CD,$7E,$72,$31,$87,$E5,$2D,$F2
	.dc.b	$9F,$94,$6D,$FC,$AA,$DE,$7F,$B4,$8F,$DB,$90,$F8,$1A,$7F,$89,$DA
	.dc.b	$64,$FD,$C9,$FB,$EF,$4F,$F9,$07,$A3,$23,$F8,$D3,$FD,$03,$E8,$BF
	.dc.b	$8D,$BF,$D0,$BE,$29,$F5,$F4,$E1,$4D,$7F,$5B,$F5,$6F,$AF,$FD,$6B
	.dc.b	$67,$F0,$6D,$78,$69,$FE,$3E,$FA,$DF,$DB,$C6,$0F,$F1,$AC,$ED,$FE
	.dc.b	$32,$D7,$FC,$8A,$D7,$75,$FC,$22,$3B,$5F,$F0,$0A,$FD,$E5,$7E,$51
	.dc.b	$AF,$71,$3C,$2F,$FC,$CB,$F5,$CF,$AF,$FF,$05,$6C,$8C,$5F,$EC,$63
	.dc.b	$13,$A9,$FE,$97,$18,$9D,$47,$07,$DD,$19,$6C,$7F,$4E,$D9,$E3,$FD
	.dc.b	$32,$4F,$E7,$63,$96,$7F,$D0,$BA,$11,$C2,$75,$5C,$1F,$F0,$8C,$70
	.dc.b	$CB,$E4,$F1,$7D,$62,$4F,$C5,$FA,$6A,$FF,$48,$3E,$2C,$7C,$D2,$3F
	.dc.b	$E2,$2D,$51,$FC,$B3,$6F,$E7,$93,$F3,$5B,$9F,$E1,$A7,$F1,$DF,$5F
	.dc.b	$F8,$FF,$AF,$97,$ED,$8F,$EE,$9E,$7F,$F6,$62,$47,$F4,$87,$FC,$63
	.dc.b	$B7,$C0,$F2,$7F,$31,$F5,$F2,$BE,$34,$7E,$6F,$D1,$FF,$20,$7D,$6B
	.dc.b	$E4,$69,$9A,$FE,$55,$E8,$27,$51,$FD,$AD,$AF,$F6,$A8,$D4,$FF,$7C
	.dc.b	$9C,$CF,$FA,$63,$1F,$A7,$C3,$FA,$B2,$47,$69,$E7,$BF,$F7,$AC,$3C
	.dc.b	$1F,$1D,$E6,$25,$7D,$FF,$AF,$AF,$86,$D3,$E1,$24,$79,$DF,$DD,$7D
	.dc.b	$7C,$8F,$04,$7C,$2E,$29,$1E,$37,$F8,$07,$D2,$E2,$3B,$47,$B4,$8F
	.dc.b	$C3,$7F,$83,$77,$C8,$E4,$7C,$AF,$AF,$7C,$3F,$F9,$BD,$B1,$98,$FE
	.dc.b	$1D,$F0,$B9,$7F,$C4,$B3,$A9,$F9,$CD,$7F,$D8,$7A,$A3,$11,$FE,$BA
	.dc.b	$8C,$4F,$F1,$9F,$90,$93,$DC,$F2,$7D,$0F,$8C,$8F,$EA,$FE,$2E,$BE
	.dc.b	$08,$F1,$91,$FF,$16,$60,$8F,$E4,$DE,$74,$9E,$A7,$F6,$0F,$AF,$FE
	.dc.b	$64,$FF,$0D,$6F,$1F,$F8,$49,$BD,$FF,$E9,$7F,$03,$F9,$ED,$7F,$B5
	.dc.b	$27,$8B,$FE,$38,$6B,$FE,$E4,$6B,$BF,$0F,$FC,$4F,$1D,$21,$D1,$FC
	.dc.b	$13,$D8,$FE,$78,$7F,$88,$27,$53,$FE,$5B,$E4,$FF,$11,$FB,$86,$9F
	.dc.b	$D0,$FE,$2A,$CF,$FA,$82,$74,$D9,$FF,$1B,$C6,$23,$FC,$9F,$87,$F9
	.dc.b	$9F,$DD,$FF,$59,$68,$EC,$FF,$2D,$3F,$EF,$23,$1F,$15,$FE,$85,$99
	.dc.b	$7F,$E1,$D9,$CB,$6B,$8B,$EE,$C3,$FC,$53,$CD,$F8,$5F,$8F,$FC,$81
	.dc.b	$27,$FC,$D3,$ED,$D7,$FA,$74,$ED,$FC,$C4,$FF,$A5,$67,$FC,$8B,$F5
	.dc.b	$DF,$E1,$49,$FF,$0D,$B5,$FE,$90,$DE,$3F,$E6,$58,$7F,$87,$32,$54
	.dc.b	$AD,$46,$21,$18,$75,$C9,$FE,$A5,$F5,$7C,$2F,$F9,$8A,$7C,$13,$FE
	.dc.b	$7A,$6B,$CD,$AB,$C6,$7D,$73,$E4,$3C,$1F,$1F,$FE,$15,$9C,$C6,$3D
	.dc.b	$7F,$3C,$62,$FD,$35,$FE,$1B,$E4,$FF,$50,$7B,$3B,$9E,$D8,$F5,$77
	.dc.b	$87,$28,$C4,$2F,$C1,$18,$7A,$6D,$FE,$67,$7F,$C4,$DF,$5A,$FA,$3F
	.dc.b	$F0,$1D,$FC,$78,$C6,$7F,$EF,$06,$C7,$F5,$56,$BF,$E9,$EF,$91,$FC
	.dc.b	$D7,$D7,$7F,$BD,$9B,$1F,$D6,$DA,$EF,$A9,$3D,$7C,$7F,$D6,$DF,$F5
	.dc.b	$CE,$FC,$12,$FF,$DA,$7F,$5C,$F9,$CF,$F5,$9E,$BF,$36,$31,$CF,$F1
	.dc.b	$2B,$F2,$DF,$4F,$AE,$2F,$9F,$5A,$9F,$8D,$18,$C3,$51,$F6,$E7,$58
	.dc.b	$7C,$8B,$CE,$E9,$CE,$1F,$8D,$F5,$DF,$39,$FE,$C6,$EB,$FF,$50,$C6
	.dc.b	$34,$FF,$65,$FB,$B9,$78,$55,$D3,$F3,$35,$B7,$FC,$4C,$DB,$FC,$FD
	.dc.b	$EE,$7F,$C2,$F5,$6B,$57,$D0,$EE,$7F,$A1,$1C,$CB,$BD,$F5,$8F,$8D
	.dc.b	$F1,$46,$1F,$27,$5F,$FB,$C7,$E1,$F3,$5B,$C7,$FB,$CB,$70,$FD,$EB
	.dc.b	$74,$FD,$AD,$70,$C3,$9F,$53,$E8,$FE,$06,$2F,$47,$FC,$FA,$FF,$BC
	.dc.b	$23,$31,$8C,$BF,$CA,$D9,$AF,$95,$FF,$17,$AD,$65,$FC,$13,$0E,$3F
	.dc.b	$0B,$F1,$EF,$FD,$1B,$7F,$EA,$66,$E3,$D3,$83,$FB,$EC,$E7,$87,$81
	.dc.b	$1F,$F4,$F4,$6A,$7F,$CB,$55,$CB,$1F,$F6,$FF,$D7,$7F,$92,$AF,$7F
	.dc.b	$0F,$F5,$E3,$17,$DE,$31,$C5,$2B,$AF,$D7,$5A,$FF,$33,$5F,$0C,$3A
	.dc.b	$DF,$D4,$C9,$7B,$9F,$F0,$E7,$13,$DB,$D7,$D5,$FD,$86,$E7,$F6,$1C
	.dc.b	$3E,$CE,$6B,$A7,$DA,$9B,$87,$F9,$7F,$2F,$C4,$F3,$7E,$22,$7F,$07
	.dc.b	$E0,$78,$FC,$2B,$97,$C6,$6E,$3F,$84,$9F,$47,$81,$8C,$BC,$F1,$6F
	.dc.b	$F9,$CF,$3B,$3F,$F2,$EC,$7E,$3A,$7F,$23,$C9,$F0,$7E,$2E,$CF,$D4
	.dc.b	$6E,$5F,$92,$9F,$8D,$F3,$D8,$CB,$E2,$C5,$BE,$F8,$AB,$17,$FA,$5F
	.dc.b	$AD,$7D,$77,$CF,$BF,$D2,$C6,$1F,$C1,$98,$88,$FF,$B1,$3E,$36,$D7
	.dc.b	$F6,$0A,$E0,$DA,$FD,$6F,$C6,$7F,$B4,$AF,$FC,$C7,$5A,$CF,$FD,$1D
	.dc.b	$3F,$E3,$C9,$FF,$67,$FB,$9E,$2F,$97,$E2,$D8,$E3,$1D,$3B,$71,$97
	.dc.b	$EA,$2F,$F4,$B5,$FE,$39,$C5,$7B,$8E,$27,$FB,$73,$EB,$BC,$BF,$53
	.dc.b	$E6,$6C,$79,$BA,$7E,$19,$EA,$F7,$2C,$7F,$B9,$3C,$7C,$34,$FF,$73
	.dc.b	$C7,$F1,$9F,$AE,$FE,$A1,$1F,$EB,$BF,$AE,$7B,$B9,$EB,$FE,$49,$E6
	.dc.b	$A7,$35,$FD,$FB,$8C,$FF,$7D,$C3,$9F,$0F,$FB,$35,$B3,$A8,$5F,$3F
	.dc.b	$73,$B5,$F3,$39,$63,$8A,$FF,$63,$E2,$AC,$61,$F5,$3C,$AF,$B2,$7D
	.dc.b	$AE,$9B,$8F,$3C,$3F,$C8,$7C,$8A,$CC,$7D,$3A,$B7,$E5,$CD,$F5,$CE
	.dc.b	$CF,$82,$DC,$BF,$DC,$0F,$E9,$30,$FF,$6D,$B9,$6C,$F0,$38,$F7,$DA
	.dc.b	$F2,$4B,$EA,$FF,$96,$3F,$E2,$53,$1D,$3F,$1D,$C4,$69,$FF,$15,$35
	.dc.b	$FF,$17,$BF,$8C,$6C,$7F,$4A,$FA,$DF,$AF,$1F,$EF,$F9,$D3,$DA,$BC
	.dc.b	$71,$C7,$C3,$57,$0D,$3C,$53,$9F,$71,$D5,$FD,$59,$B8,$7C,$FF,$0D
	.dc.b	$C5,$FD,$13,$3F,$F2,$7D,$69,$B7,$C4,$FD,$8F,$99,$70,$E5,$7F,$E3
	.dc.b	$96,$FB,$1F,$B6,$FA,$2F,$DB,$B2,$32,$70,$8C,$61,$EF,$F7,$FF,$33
	.dc.b	$F5,$6F,$AD,$FE,$F1,$B2,$73,$18,$66,$1F,$89,$E6,$5F,$F0,$38,$FA
	.dc.b	$33,$AC,$3B,$46,$E7,$51,$F5,$27,$F8,$6E,$D2,$BD,$BF,$53,$6B,$F8
	.dc.b	$24,$DE,$1F,$E1,$F3,$6B,$FC,$D1,$F5,$AF,$88,$FF,$08,$36,$3C,$1E
	.dc.b	$D6,$26,$47,$E3,$A7,$58,$B0,$FC,$F6,$7F,$E3,$17,$F1,$D3,$A7,$F2
	.dc.b	$1E,$0F,$F8,$5B,$5F,$83,$F1,$7F,$E1,$F8,$C5,$FF,$71,$AD,$FA,$19
	.dc.b	$3E,$7F,$AB,$73,$FC,$FD,$FF,$88,$33,$FE,$1B,$5A,$9C,$BF,$E2,$BF
	.dc.b	$AC,$7C,$5D,$3F,$88,$FD,$6F,$E6,$23,$35,$FA,$A6,$DF,$B6,$3F,$D1
	.dc.b	$D1,$8C,$BF,$79,$5C,$6F,$FC,$71,$E8,$FF,$87,$EF,$F6,$6D,$C3,$91
	.dc.b	$FC,$5B,$EB,$5F,$0D,$C3,$DC,$97,$C1,$FE,$4B,$3A,$8C,$1F,$C4,$7C
	.dc.b	$17,$3B,$9F,$F5,$07,$F5,$0F,$A5,$B5,$B9,$D3,$67,$4D,$3F,$4B,$F1
	.dc.b	$78,$CF,$FA,$63,$77,$FE,$AE,$D9,$FE,$F7,$DE,$8C,$57,$F9,$47,$0F
	.dc.b	$D1,$57,$87,$AF,$FC,$8B,$DD,$FB,$D6,$46,$4F,$DF,$7B,$C9,$CD,$FE
	.dc.b	$FB,$E2,$BF,$82,$1F,$E4,$EF,$06,$DF,$B6,$2C,$FF,$92,$AF,$F7,$CF
	.dc.b	$19,$D1,$1D,$57,$0D,$3F,$A5,$6F,$F3,$30,$C5,$FF,$86,$30,$E7,$5A
	.dc.b	$CD,$5F,$E7,$DF,$82,$7D,$33,$E5,$70,$DC,$7D,$BC,$38,$99,$1F,$D2
	.dc.b	$EB,$8C,$7F,$9A,$A7,$3A,$FF,$BB,$3E,$B7,$F5,$57,$FE,$2A,$FA,$AF
	.dc.b	$1B,$FE,$E4,$D9,$3A,$A7,$92,$BD,$C7,$FA,$16,$75,$5C,$3C,$4F,$F4
	.dc.b	$1D,$7E,$1A,$BC,$11,$FE,$88,$AD,$9D,$6F,$A6,$39,$9F,$F4,$6F,$D6
	.dc.b	$7F,$86,$EF,$FC,$F7,$F5,$9F,$6F,$3F,$CC,$4E,$78,$7F,$BC,$62,$AD
	.dc.b	$5F,$E9,$B4,$5E,$F0,$F7,$36,$7D,$BA,$7E,$39,$E1,$19,$AE,$3E,$EF
	.dc.b	$90,$F6,$F3,$5C,$3C,$71,$6F,$FD,$23,$87,$3C,$FF,$07,$8F,$79,$FF
	.dc.b	$7F,$8F,$AD,$FD,$02,$27,$8F,$FE,$13,$8C,$C7,$06,$BF,$E1,$6C,$78
	.dc.b	$46,$23,$FD,$F8,$75,$E9,$FE,$F1,$5C,$DF,$EF,$51,$A9,$E7,$39,$8F
	.dc.b	$EF,$DF,$15,$D2,$B8,$DD,$9C,$9D,$4F,$41,$E3,$30,$79,$E3,$8B,$83
	.dc.b	$DD,$3C,$DF,$7C,$D7,$26,$EB,$C8,$F3,$3B,$B8,$77,$C9,$FE,$31,$7C
	.dc.b	$AD,$7E,$7B,$5D,$FA,$DE,$8F,$8D,$FF,$66,$69,$F4,$6B,$84,$62,$91
	.dc.b	$8F,$03,$FA,$B4,$62,$F9,$46,$27,$57,$ED,$B7,$CB,$7C,$6D,$7E,$DC
	.dc.b	$63,$1D,$61,$D2,$72,$DD,$7C,$D7,$C5,$B5,$C3,$08,$FD,$0C,$3A,$36
	.dc.b	$B8,$3F,$8D,$8F,$66,$1D,$AB,$CD,$5A,$A5,$72,$AD,$47,$5C,$3C,$8F
	.dc.b	$3A,$E1,$5B,$C3,$E1,$E1,$F8,$7E,$2F,$D8,$F7,$7E,$1A,$3F,$46,$DD
	.dc.b	$BF,$66,$DC,$7F,$67,$D7,$6F,$8C,$7F,$00,$9C,$70,$FF,$3D,$8F,$F1
	.dc.b	$E7,$17,$F2,$A7,$44,$3F,$77,$3D,$79,$BF,$8D,$61,$CA,$31,$CD,$C6
	.dc.b	$BF,$87,$CE,$5F,$EC,$DF,$5B,$FE,$A2,$6F,$C0,$F9,$63,$11,$FC,$00
	.dc.b	$C4,$47,$67,$98,$C1,$FD,$EB,$07,$D6,$BE,$37,$CC,$D9,$E2,$7F,$80
	.dc.b	$39,$7F,$A2,$FC,$4F,$C7,$6C,$7F,$41,$F0,$3F,$DF,$8F,$F6,$2F,$07
	.dc.b	$51,$88,$E4,$DE,$3F,$F5,$05,$72,$9F,$EC,$15,$F4,$7E,$B7,$C5,$1F
	.dc.b	$CA,$DB,$5F,$D7,$9B,$C9,$FD,$77,$C6,$E1,$A4,$BF,$F0,$75,$69,$FE
	.dc.b	$97,$E2,$2B,$51,$D6,$AE,$1F,$CF,$2B,$53,$CA,$73,$87,$F5,$7C,$3A
	.dc.b	$6C,$BD,$DF,$9F,$E3,$7D,$44,$7F,$C0,$BD,$15,$D1,$E6,$8F,$C3,$4F
	.dc.b	$F3,$0C,$BF,$DD,$BE,$0F,$E2,$7C,$7F,$C3,$47,$FB,$C3,$97,$13,$30
	.dc.b	$FF,$42,$C6,$CC,$3F,$27,$C6,$F9,$84,$76,$7C,$67,$0E,$CC,$9F,$E4
	.dc.b	$01,$F5,$9F,$EB,$3B,$FF,$59,$37,$6F,$E6,$6D,$9F,$E9,$9F,$5B,$FC
	.dc.b	$69,$FF,$5E,$C6,$5E,$67,$EC,$4A,$7F,$B0,$A3,$83,$E1,$C1,$F2,$BB
	.dc.b	$6C,$7F,$B2,$1A,$F9,$FE,$B5,$F3,$1B,$A9,$3B,$7C,$13,$A8,$FF,$68
	.dc.b	$FD,$6F,$FB,$75,$BF,$05,$E5,$8C,$47,$C3,$FA,$DF,$E0,$4F,$C5,$8C
	.dc.b	$EB,$FE,$D5,$6E,$BF,$D5,$9B,$3F,$DB,$EB,$8D,$6E,$FE,$14,$7F,$BC
	.dc.b	$E7,$FD,$C6,$D7,$B4,$7C,$7F,$AD,$F4,$47,$D5,$6D,$7F,$B9,$5B,$87
	.dc.b	$F6,$6F,$67,$FC,$EF,$5F,$EE,$E9,$CB,$FE,$01,$6A,$6C,$70,$8F,$E1
	.dc.b	$1C,$5F,$3A,$FF,$C4,$2D,$BF,$F3,$5E,$9F,$97,$D7,$F3,$0C,$BF,$F6
	.dc.b	$CB,$97,$FD,$FF,$F0,$8F,$92,$FC,$13,$63,$31,$F8,$5F,$AD,$7C,$36
	.dc.b	$A2,$4E,$8F,$F0,$D6,$6C,$BF,$C0,$B7,$C2,$74,$7E,$6D,$7C,$9F,$53
	.dc.b	$9C,$7B,$8F,$D8,$10,$FF,$78,$FA,$15,$D1,$FE,$F4,$3F,$6C,$FF,$19
	.dc.b	$F3,$A4,$7F,$12,$7F,$A1,$7C,$FF,$F1,$3F,$35,$E0,$27,$75,$FC,$7D
	.dc.b	$DF,$83,$FA,$5F,$A9,$7D,$65,$25,$7C,$0D,$7F,$C4,$7F,$55,$FA,$8D
	.dc.b	$FE,$14,$FF,$8F,$6F,$8B,$B3,$17,$C9,$FD,$F5,$71,$BF,$F1,$6D,$70
	.dc.b	$9F,$F1,$CB,$5F,$F1,$88,$70,$F0,$AF,$F1,$FA,$D5,$F5,$6F,$AC,$FE
	.dc.b	$E1,$18,$6C,$66,$7C,$D3,$C3,$3C,$D4,$9C,$D6,$DF,$F6,$AD,$6D,$F7
	.dc.b	$78,$CE,$07,$8B,$F4,$89,$3A,$8E,$5B,$7F,$9F,$1D,$C6,$67,$6F,$F6
	.dc.b	$98,$DE,$BF,$1B,$9B,$EB,$12,$7E,$70,$DD,$F4,$79,$8F,$E5,$9C,$0A
	.dc.b	$C6,$5F,$62,$78,$56,$30,$FB,$BE,$27,$F2,$7C,$7F,$D6,$D5,$DB,$1E
	.dc.b	$4F,$AF,$EA,$FF,$1F,$F5,$92,$30,$7F,$64,$F3,$3F,$B7,$11,$F8,$51
	.dc.b	$F9,$6C,$3F,$2F,$F5,$92,$7E,$34,$7E,$6B,$CF,$FF,$13,$7D,$53,$EB
	.dc.b	$29,$9A,$FF,$7F,$46,$5B,$1A,$7E,$84,$FE,$93,$69,$1F,$4D,$D3,$FE
	.dc.b	$AB,$A7,$C6,$78,$09,$1F,$67,$EB,$2B,$B8,$BC,$3F,$6E,$6F,$0F,$FA
	.dc.b	$7F,$97,$F0,$9F,$59,$FB,$43,$38,$3B,$99,$7C,$C8,$F1,$72,$FC,$03
	.dc.b	$F7,$7F,$45,$63,$DF,$15,$FC,$34,$5C,$FF,$19,$DD,$FC,$08,$C6,$7F
	.dc.b	$EE,$1F,$AC,$7C,$1F,$F9,$45,$B8,$7E,$DF,$34,$62,$3F,$25,$F0,$1B
	.dc.b	$A9,$1B,$9D,$63,$F9,$BA,$CC,$FE,$67,$E2,$FF,$89,$98,$9F,$E0,$F8
	.dc.b	$76,$F9,$EF,$11,$2B,$F4,$DF,$0D,$5F,$F6,$F6,$29,$C5,$FE,$A6,$AE
	.dc.b	$7F,$3F,$BB,$FA,$54,$FE,$B7,$E2,$F3,$8F,$FB,$47,$EB,$3F,$83,$3F
	.dc.b	$A8,$6B,$FE,$11,$6C,$FE,$D7,$CA,$FF,$04,$46,$30,$FF,$61,$B6,$BA
	.dc.b	$78,$7F,$CC,$2B,$6F,$F3,$2F,$91,$93,$DE,$3F,$C2,$5E,$D3,$92,$92
	.dc.b	$7F,$83,$7C,$45,$7F,$C4,$93,$B7,$F7,$ED,$7F,$83,$47,$F8,$F3,$DB
	.dc.b	$FF,$47,$67,$FC,$0B,$DC,$BD,$DF,$E3,$F8,$FC,$A6,$0A,$E1,$87,$F3
	.dc.b	$79,$E0,$DA,$E0,$FF,$97,$27,$F1,$1E,$F4,$8F,$C1,$3F,$E4,$CF,$6F
	.dc.b	$11,$F6,$22,$F0,$7F,$97,$E7,$FC,$65,$F5,$9F,$5B,$1F,$F7,$6B,$6C
	.dc.b	$1C,$77,$FF,$98,$7D,$5E,$3C,$92,$BF,$90,$D6,$B5,$FF,$06,$BC,$E3
	.dc.b	$94,$ED,$FF,$4A,$4F,$F8,$53,$C8,$7D,$63,$E3,$BF,$DF,$FF,$19,$FE
	.dc.b	$05,$9C,$C6,$39,$FF,$DD,$55,$BC,$3F,$DF,$F1,$F3,$3D,$8D,$47,$FB
	.dc.b	$B2,$74,$B0,$E9,$18,$6D,$F0,$D7,$FC,$FD,$7C,$AD,$F5,$6F,$A1,$FF
	.dc.b	$2B,$C7,$C6,$32,$E1,$F9,$31,$BC,$FF,$B2,$B6,$3F,$B5,$9F,$EE,$2E
	.dc.b	$87,$D6,$7F,$6A,$F8,$EF,$CA,$36,$33,$3C,$27,$2F,$2A,$FE,$99,$D7
	.dc.b	$DB,$C7,$FD,$5C,$F9,$67,$2E,$E3,$2F,$FA,$98,$FF,$04,$ED,$FE,$33
	.dc.b	$D7,$FD,$3D,$78,$E0,$FC,$BD,$53,$FD,$5B,$16,$7F,$D4,$91,$78,$7F
	.dc.b	$C0,$5C,$BF,$52,$3B,$36,$7A,$B7,$F2,$3F,$F0,$33,$FD,$E9,$B3,$FE
	.dc.b	$D9,$7F,$DB,$11,$9F,$5F,$FC,$3F,$D3,$FE,$20,$B5,$70,$8C,$36,$BF
	.dc.b	$A1,$3C,$AF,$A5,$F2,$9C,$B6,$17,$D2,$30,$F6,$F6,$2D,$F5,$AD,$AF
	.dc.b	$53,$26,$BD,$7B,$8F,$B6,$2D,$7B,$66,$E5,$E8,$F5,$3A,$9D,$BD,$5B
	.dc.b	$87,$C5,$8B,$7F,$16,$6D,$7C,$3F,$C9,$5F,$5B,$E8,$6E,$1F,$E2,$65
	.dc.b	$F5,$79,$99,$39,$78,$46,$30,$5F,$5F,$5F,$A1,$D7,$E5,$AE,$1D,$62
	.dc.b	$D7,$08,$FE,$E5,$8A,$B3,$7C,$E7,$81,$E5,$8C,$BC,$3E,$C7,$D4,$7A
	.dc.b	$FE,$7A,$C7,$AE,$2E,$1D,$CF,$B3,$82,$B3,$87,$C5,$AB,$7E,$49,$B7
	.dc.b	$C3,$EC,$3E,$41,$F2,$7E,$46,$D7,$C9,$8B,$3D,$F9,$19,$2F,$39,$7D
	.dc.b	$3B,$B8,$7C,$29,$B8,$79,$BD,$BF,$AC,$7D,$7F,$81,$B5,$F5,$E2,$CF
	.dc.b	$7C,$7F,$D0,$B3,$92,$46,$67,$94,$63,$1F,$C2,$7D,$7F,$E0,$CF,$C2
	.dc.b	$7C,$45,$9F,$C2,$45,$BE,$D3,$6B,$8D,$68,$FC,$54,$65,$FB,$BF,$3D
	.dc.b	$F8,$FF,$AC,$FC,$83,$63,$F2,$0E,$1F,$C9,$78,$17,$0F,$63,$F9,$73
	.dc.b	$F3,$1E,$35,$9F,$CC,$7C,$EF,$E6,$FE,$37,$F3,$8D,$8F,$F4,$0F,$27
	.dc.b	$FA,$F9,$BF,$13,$FA,$17,$F7,$31,$9C,$7E,$C3,$FE,$CE,$8E,$3F,$17
	.dc.b	$AF,$80,$DB,$FC,$0F,$EF,$3F,$9A,$EF,$7C,$B3,$78,$F0,$CC,$3F,$D6
	.dc.b	$15,$C1,$B5,$FE,$5C,$6C,$EF,$2F,$CC,$56,$DB,$5B,$6C,$FA,$BE,$BB
	.dc.b	$D7,$F4,$F7,$0F,$82,$DA,$F6,$B7,$93,$F9,$36,$A8,$FE,$3D,$2F,$A6
	.dc.b	$F8,$4F,$EA,$69,$F6,$5F,$C1,$B0,$FF,$4E,$D6,$A7,$F9,$8C,$E6,$BF
	.dc.b	$60,$73,$DB,$FD,$38,$FF,$65,$AE,$AF,$FB,$29,$AE,$63,$6F,$F3,$0C
	.dc.b	$5F,$5F,$F6,$3E,$0B,$EC,$B7,$2F,$BE,$DA,$FB,$6D,$CB,$DE,$D9,$FF
	.dc.b	$41,$5F,$47,$F0,$F7,$F0,$B1,$7D,$4F,$F4,$1F,$35,$F5,$CF,$53,$53
	.dc.b	$98,$C6,$1F,$12,$32,$FF,$B6,$BE,$A1,$F6,$78,$DB,$F1,$FF,$5F,$F9
	.dc.b	$2F,$AF,$FC,$9B,$6F,$F3,$2D,$9D,$C7,$E7,$27,$F2,$51,$FE,$71,$8F
	.dc.b	$C6,$B4,$FC,$48,$E4,$DE,$D2,$76,$F2,$8C,$4F,$08,$BC,$2A,$FF,$20
	.dc.b	$F3,$FF,$90,$AF,$31,$FE,$1F,$FA,$CF,$F6,$EB,$FE,$97,$FC,$05,$EF
	.dc.b	$7E,$32,$73,$97,$F3,$3F,$8C,$56,$AF,$FD,$CF,$5A,$AF,$F1,$50,$70
	.dc.b	$F7,$F8,$5F,$C2,$B9,$6F,$C4,$3E,$BB,$EC,$71,$3A,$3F,$3B,$39,$C7
	.dc.b	$EA,$46,$2A,$47,$D7,$F2,$EE,$3F,$D2,$61,$FA,$F3,$E5,$3F,$C0,$91
	.dc.b	$86,$C7,$BB,$8B,$E4,$57,$C9,$8E,$15,$FE,$7B,$6C,$75,$6C,$F0,$9E
	.dc.b	$FA,$74,$B9,$1C,$5E,$F1,$8B,$FD,$7F,$E4,$DE,$FA,$FF,$3B,$DF,$8B
	.dc.b	$AD,$F5,$FF,$85,$8F,$2D,$AF,$83,$FD,$9E,$B7,$CD,$FC,$12,$73,$5C
	.dc.b	$1B,$87,$FA,$12,$BA,$F7,$6F,$7F,$E1,$7F,$AF,$BC,$8F,$D5,$7E,$90
	.dc.b	$F8,$6C,$C6,$1B,$8F,$A0,$FF,$35,$57,$17,$1A,$B2,$7E,$C1,$FC,$3F
	.dc.b	$5F,$5D,$EE,$BF,$9D,$FD,$6F,$F3,$E3,$F9,$FF,$83,$71,$FE,$8B,$16
	.dc.b	$7F,$A0,$CD,$BF,$E4,$BF,$63,$FD,$18,$FE,$8F,$F1,$57,$0F,$E9,$B1
	.dc.b	$6F,$FA,$4C,$DA,$FF,$27,$FC,$E7,$F9,$10,$FF,$22,$FC,$ED,$C3,$FC
	.dc.b	$B9,$16,$FF,$CB,$93,$6B,$FC,$AF,$CF,$FE,$4C,$3F,$C9,$BC,$F7,$0F
	.dc.b	$F3,$24,$5B,$FF,$32,$4D,$AE,$35,$A3,$FC,$D5,$19,$7F,$CC,$35,$6B
	.dc.b	$FC,$AF,$37,$DA,$FF,$34,$FC,$67,$F9,$70,$FF,$2E,$FC,$ED,$C3,$FC
	.dc.b	$E9,$16,$FF,$CE,$93,$71,$FF,$37,$FC,$DF,$F9,$90,$FF,$32,$FC,$E5
	.dc.b	$C3,$FC,$F9,$16,$FF,$CF,$93,$6B,$8D,$68,$FF,$41,$46,$5F,$F3,$CF
	.dc.b	$CD,$7F,$9C,$0F,$F3,$8F,$CE,$58,$FF,$44,$45,$BF,$F4,$44,$DB,$FF
	.dc.b	$41,$CD,$AF,$F4,$1C,$DA,$FF,$3C,$CD,$CF,$FD,$17,$39,$9F,$F6,$98
	.dc.b	$FB,$17,$CC,$FF,$18,$E8,$FE,$03,$18,$9F,$E3,$7F,$23,$FE,$EA,$8C
	.dc.b	$47,$EA,$72,$C5,$FF,$B0,$AA,$CF,$FB,$2E,$F9,$CE,$63,$C0,$F7,$AD
	.dc.b	$C7,$92,$FA,$BF,$EB,$03,$67,$88,$A7,$84,$A7,$FB,$FA,$F9,$BC,$63
	.dc.b	$9B,$FE,$E7,$37,$7E,$47,$A4,$78,$23,$11,$E3,$BF,$33,$D8,$E5,$E8
	.dc.b	$BE,$6B,$D8,$D8,$F3,$35,$F9,$ED,$F9,$46,$FF,$E9,$29,$FE,$E9,$5F
	.dc.b	$EF,$19,$CE,$E9,$C9,$63,$2E,$03,$8B,$ED,$7D,$72,$FA,$F2,$E0,$FF
	.dc.b	$7F,$46,$1B,$97,$86,$B8,$37,$1F,$AB,$7C,$71,$B7,$C1,$F2,$5F,$06
	.dc.b	$D6,$98,$7D,$8A,$F4,$E1,$E4,$AE,$13,$EA,$BE,$8D,$3C,$31,$AA,$E3
	.dc.b	$87,$C7,$AD,$C7,$96,$BB,$57,$8F,$7F,$F7,$5F,$C9,$FC,$B9,$FF,$79
	.dc.b	$37,$2F,$F7,$6B,$77,$FF,$76,$F9,$D6,$BF,$DC,$73,$6F,$9C,$7F,$BF
	.dc.b	$63,$1F,$21,$FE,$08,$8C,$10,$EE,$7F,$42,$9F,$F7,$ED,$70,$A9,$3B
	.dc.b	$C3,$FD,$FF,$3D,$37,$93,$C7,$0F,$C6,$57,$F7,$6C,$10,$F2,$F7,$5F
	.dc.b	$EB,$B8,$DE,$5F,$8B,$AD,$E1,$E8,$BD,$E1,$CE,$F7,$82,$F6,$F8,$6B
	.dc.b	$E1,$7C,$DF,$F9,$12,$3D,$8D,$7F,$D9,$4D,$BE,$9C,$3C,$3D,$59,$5C
	.dc.b	$1F,$93,$1D,$3E,$32,$4F,$47,$BF,$9D,$88,$EB,$16,$FE,$16,$1F,$E2
	.dc.b	$99,$E0,$43,$67,$9A,$B8,$6B,$F0,$E5,$78,$AF,$81,$1C,$DF,$5F,$C5
	.dc.b	$FC,$02,$3F,$03,$E3,$6B,$A3,$E1,$C7,$C6,$8F,$C4,$79,$4F,$AB,$A4
	.dc.b	$AF,$F5,$FF,$CD,$D7,$FD,$77,$17,$4F,$F4,$FC,$F2,$F9,$C9,$5F,$EC
	.dc.b	$67,$AF,$4E,$23,$FD,$87,$16,$BF,$D4,$F3,$F9,$B9,$7E,$C2,$F8,$BC
	.dc.b	$36,$3F,$C2,$E6,$A3,$87,$E6,$8F,$93,$DC,$63,$E5,$65,$70,$BE,$4F
	.dc.b	$FA,$AB,$E7,$9E,$B7,$85,$B1,$D9,$AF,$FA,$C3,$EE,$BF,$D2,$CE,$67
	.dc.b	$D1,$1F,$9D,$8F,$2B,$5E,$D1,$ED,$F8,$6F,$34,$7A,$9A,$FF,$AF,$1B
	.dc.b	$AF,$F9,$29,$7C,$C7,$1D,$25,$71,$7A,$46,$23,$94,$5A,$E5,$3F,$37
	.dc.b	$EA,$5E,$85,$25,$FC,$CF,$BF,$FF,$3E,$3F,$3A,$76,$E7,$E6,$7E,$B4
	.dc.b	$9E,$31,$FE,$D4,$F9,$3F,$3B,$FE,$8E,$3F,$A3,$38,$70,$D7,$FD,$26
	.dc.b	$D7,$F9,$13,$BC,$7F,$CD,$7F,$1A,$E1,$FC,$33,$6B,$F0,$75,$BE,$DF
	.dc.b	$E5,$5F,$52,$F6,$3F,$18,$DA,$FC,$4B,$5F,$F7,$5B,$FE,$D4,$73,$1C
	.dc.b	$08,$F0,$F6,$2B,$C2,$2E,$5F,$E8,$0A,$E5,$28,$FE,$45,$D8,$FA,$2F
	.dc.b	$D7,$F3,$4A,$FE,$61,$39,$6C,$E5,$E5,$97,$FB,$B2,$BF,$6F,$3F,$A7
	.dc.b	$FA,$67,$7D,$25,$FE,$9B,$EF,$DF,$27,$88,$E6,$7F,$79,$87,$CE,$7E
	.dc.b	$C5,$B1,$FA,$EB,$EA,$F9,$BD,$3F,$E0,$FF,$01,$23,$04,$3A,$F9,$9F
	.dc.b	$EC,$58,$D4,$63,$6F,$F4,$73,$BD,$FE,$2C,$E9,$F8,$F7,$ED,$F9,$CF
	.dc.b	$C5,$57,$F8,$3F,$14,$62,$11,$87,$6D,$7F,$24,$DA,$FF,$5A,$FD,$47
	.dc.b	$ED,$5B,$E7,$77,$AF,$EC,$5F,$94,$BE,$87,$F9,$2C,$43,$2F,$F6,$F8
	.dc.b	$D4,$7D,$A3,$47,$C2,$72,$FD,$01,$7F,$E2,$E7,$F8,$DB,$5D,$36,$FC
	.dc.b	$FE,$EF,$F2,$58,$C7,$2F,$F8,$1D,$B5,$FE,$33,$6B,$DA,$78,$7A,$D2
	.dc.b	$3B,$BC,$E3,$D5,$D9,$27,$E0,$BE,$DF,$A2,$F7,$78,$72,$7E,$2B,$F0
	.dc.b	$7C,$3F,$85,$F0,$3F,$0F,$E8,$A5,$7C,$A8,$F8,$DF,$03,$23,$E6,$BF
	.dc.b	$27,$E0,$FE,$57,$C1,$49,$FA,$2E,$DF,$94,$FC,$FF,$83,$91,$A3,$EB
	.dc.b	$ED,$F4,$C8,$FD,$AD,$AB,$F4,$E2,$F0,$FF,$85,$78,$95,$8E,$2F,$F0
	.dc.b	$94,$7C,$CE,$A3,$4D,$E1,$F8,$6F,$82,$FC,$11,$1F,$C4,$FC,$05,$7F
	.dc.b	$05,$17,$1F,$04,$DC,$3B,$CD,$AF,$F0,$E5,$FF,$91,$EF,$CD,$EF,$7E
	.dc.b	$3C,$93,$F9,$6F,$83,$AF,$E3,$E2,$CF,$B6,$6E,$1E,$C9,$B5,$DA,$7F
	.dc.b	$35,$D3,$23,$F4,$2F,$E7,$7E,$03,$11,$F9,$B8,$B8,$7C,$99,$B8,$7C
	.dc.b	$89,$B5,$EA,$9F,$D3,$7B,$B2,$3F,$5A,$FE,$A7,$E0,$31,$1F,$A7,$8B
	.dc.b	$87,$D5,$9B,$87,$D4,$9B,$5F,$1A,$7F,$CC,$1F,$94,$FF,$88,$EF,$FC
	.dc.b	$A0,$DB,$FF,$26,$36,$BF,$C9,$0F,$F9,$06,$33,$EF,$F2,$24,$FF,$98
	.dc.b	$E7,$4F,$F8,$C3,$E8,$FF,$C9,$A4,$7F,$C9,$BE,$85,$7F,$C9,$B1,$70
	.dc.b	$FF,$20,$CF,$F9,$53,$E3,$E5,$7F,$95,$1F,$F2,$BF,$8D,$88,$FF,$2A
	.dc.b	$C5,$AF,$F2,$4C,$FF,$97,$3C,$49,$5F,$E5,$C7,$FC,$BF,$E4,$62,$3F
	.dc.b	$CB,$B1,$6B,$FC,$A3,$37,$4F,$F2,$64,$DA,$FF,$25,$CF,$F9,$A3,$D1
	.dc.b	$FF,$49,$7C,$33,$87,$FC,$A5,$E5,$E1,$B7,$FE,$DB,$7F,$CA,$D5,$98
	.dc.b	$FF,$2A,$B7,$A5,$F0,$3F,$D4,$A7,$81,$FE,$5A,$BD,$42,$B9,$4B,$E9
	.dc.b	$5F,$43,$FE,$30,$8F,$F5,$31,$97,$0F,$FA,$A6,$37,$CD,$FD,$21,$B1
	.dc.b	$FE,$13,$F8,$7F,$DF,$FC,$F7,$F4,$C6,$C6,$7B,$D3,$96,$C7,$F8,$6E
	.dc.b	$9E,$BF,$F5,$A8,$F1,$C6,$3C,$FF,$04,$E6,$7C,$8D,$7F,$D3,$1F,$4B
	.dc.b	$FE,$E9,$E4,$F8,$7C,$08,$5A,$73,$1B,$A7,$17,$D5,$79,$D6,$DC,$F1
	.dc.b	$7F,$AA,$EB,$85,$6F,$0E,$D7,$B8,$C4,$EA,$B8,$4E,$5F,$F5,$A7,$C4
	.dc.b	$BC,$DE,$DC,$FF,$EB,$36,$E7,$AF,$42,$E9,$FE,$B5,$BE,$13,$FE,$C5
	.dc.b	$AC,$E9,$FE,$C4,$6C,$FF,$6F,$9D,$12,$BF,$D8,$D3,$AB,$DB,$C3,$CD
	.dc.b	$95,$FE,$D1,$F3,$ED,$FF,$B3,$66,$E5,$FE,$C9,$9B,$5E,$CF,$3B,$C2
	.dc.b	$78,$BC,$FB,$1F,$ED,$B8,$B8,$7F,$B5,$E6,$DF,$FB,$77,$E9,$7F,$C6
	.dc.b	$37,$FE,$E9,$6D,$7F,$88,$9F,$F5,$CC,$67,$D5,$F8,$31,$AA,$E5,$18
	.dc.b	$F8,$9C,$BA,$F0,$A5,$F5,$F8,$9B,$8F,$59,$B8,$78,$E6,$D7,$FB,$AF
	.dc.b	$1F,$F3,$B3,$E5,$AF,$0C,$EA,$FF,$DD,$54,$F0,$BC,$A7,$9B,$C4,$B8
	.dc.b	$7B,$62,$D7,$B6,$6E,$3F,$0E,$6D,$78,$66,$D7,$6F,$0B,$E0,$9F,$0B
	.dc.b	$C4,$B5,$F2,$62,$CF,$C9,$9B,$87,$CF,$9B,$5E,$D9,$B5,$EA,$F0,$FE
	.dc.b	$69,$F3,$BC,$4B,$5F,$56,$2C,$FD,$59,$B8,$7D,$B9,$B5,$F2,$66,$D7
	.dc.b	$C6,$FA,$5F,$B2,$7D,$AF,$12,$D7,$E0,$62,$CF,$E0,$66,$E1,$F8,$79
	.dc.b	$B5,$F5,$66,$D7,$D2,$FA,$5F,$C2,$FC,$37,$E1,$9B,$5F,$8D,$70,$FE
	.dc.b	$3F,$C0,$B5,$F8,$F7,$F1,$A7,$E3,$BC,$6B,$3F,$95,$F9,$1F,$C8,$FD
	.dc.b	$2F,$E4,$9B,$1F,$E9,$2F,$A5,$FD,$AF,$19,$FC,$BC,$61,$CF,$D1,$C8
	.dc.b	$FF,$4C,$FD,$2D,$CF,$FD,$1B,$F4,$1F,$CA,$B7,$FE,$06,$DA,$4E,$BC
	.dc.b	$AF,$F4,$AC,$63,$8B,$F8,$D5,$6E,$B5,$7C,$AF,$FA,$35,$EC,$E7,$5F
	.dc.b	$CE,$39,$35,$52,$75,$6A,$E9,$87,$6C,$3C,$33,$DA,$BF,$D6,$7F,$41
	.dc.b	$F0,$BE,$97,$E1,$B6,$32,$7F,$AE,$23,$2D,$3B,$CB,$E9,$5A,$FF,$35
	.dc.b	$8F,$EB,$31,$89,$FF,$4E,$B6,$3F,$DD,$0D,$77,$3F,$CB,$1E,$8F,$F2
	.dc.b	$EF,$A2,$7D,$DB,$D6,$D5,$F3,$D3,$FD,$A3,$F3,$BF,$73,$E4,$3E,$EB
	.dc.b	$73,$FF,$68,$B7,$3F,$F5,$B3,$6B,$8C,$FF,$B5,$AB,$8B,$4F,$7D,$BE
	.dc.b	$33,$C2,$7F,$42,$F3,$DF,$42,$F9,$9C,$DF,$0A,$B9,$FF,$B5,$2F,$75
	.dc.b	$8C,$BF,$DD,$3F,$4B,$F9,$4F,$A5,$FC,$AB,$71,$FF,$74,$B7,$0F,$F6
	.dc.b	$73,$6B,$F1,$70,$D3,$F2,$6D,$AF,$C2,$F2,$FE,$53,$93,$35,$27,$35
	.dc.b	$B3,$F3,$A2,$74,$E2,$6D,$F8,$26,$D7,$E8,$29,$F4,$8F,$88,$FC,$24
	.dc.b	$66,$31,$3F,$EE,$06,$C7,$F8,$80,$E3,$3C,$1F,$D5,$C6,$AF,$F1,$CD
	.dc.b	$AF,$D7,$FE,$10,$FC,$21,$E2,$FE,$2E,$B5,$87,$ED,$39,$5F,$01,$F8
	.dc.b	$F6,$E3,$FA,$AE,$2F,$B7,$E4,$66,$A4,$FE,$82,$B7,$7F,$6F,$95,$7A
	.dc.b	$86,$BF,$E1,$7B,$7D,$33,$E2,$7F,$C1,$31,$86,$C7,$9A,$30,$7E,$F0
	.dc.b	$B5,$D1,$FF,$16,$47,$37,$4F,$13,$11,$B3,$FC,$5C,$E2,$BC,$8D,$9E
	.dc.b	$7C,$1E,$2B,$FD,$97,$D2,$FF,$67,$9F,$DF,$B7,$0F,$CD,$7D,$33,$AF
	.dc.b	$7E,$2C,$BD,$BF,$A5,$9D,$46,$0F,$E1,$3F,$48,$F1,$FF,$80,$36,$3E
	.dc.b	$97,$D2,$FF,$83,$31,$F5,$09,$D3,$86,$ED,$FC,$45,$B5,$FC,$CB,$8D
	.dc.b	$7C,$EB,$58,$7F,$42,$F9,$89,$7F,$CA,$BD,$EF,$F0,$37,$81,$8C,$FE
	.dc.b	$54,$7F,$82,$6B,$18,$7F,$94,$E3,$77,$C7,$55,$7F,$4E,$FA,$59,$39
	.dc.b	$F9,$CB,$8F,$4B,$56,$31,$E9,$0A,$C5,$FF,$95,$BE,$4B,$FC,$96,$7F
	.dc.b	$93,$3E,$4E,$C7,$A7,$E0,$93,$8C,$BD,$30,$AC,$63,$E9,$8F,$C1,$4F
	.dc.b	$F9,$6F,$97,$FC,$AF,$CD,$FE,$58,$66,$3D,$5C,$B4,$93,$DA,$32,$D8
	.dc.b	$CD,$F7,$9C,$B6,$3D,$1F,$0F,$FE,$66,$3F,$CC,$FF,$27,$63,$24,$BF
	.dc.b	$5C,$65,$F6,$FC,$97,$F9,$B8,$FF,$37,$FC,$9D,$AF,$89,$16,$7E,$1C
	.dc.b	$DC,$3E,$14,$DA,$FF,$41,$CD,$AF,$F3,$F7,$C8,$7F,$9E,$8F,$F3,$DF
	.dc.b	$C9,$DA,$FA,$11,$67,$E7,$CD,$C3,$E7,$4D,$AF,$F4,$7C,$DA,$EB,$1C
	.dc.b	$8F,$B5,$19,$8F,$F6,$B8,$FA,$57,$D1,$7F,$B1,$BA,$BF,$D4,$91,$89
	.dc.b	$FF,$64,$FC,$6F,$F8,$4A,$31,$1F,$A3,$C3,$17,$FC,$BE,$AC,$FF,$39
	.dc.b	$BE,$73,$98,$F0,$3D,$EB,$71,$E4,$BE,$AF,$FA,$F2,$32,$FF,$88,$AE
	.dc.b	$DF,$48,$F2,$9D,$0F,$04,$F0,$7F,$D8,$26,$4F,$41,$4E,$85,$8E,$8F
	.dc.b	$8E,$FC,$CE,$CE,$7F,$20,$F9,$3F,$63,$63,$C6,$D7,$E7,$B7,$E4,$5B
	.dc.b	$7F,$3C,$94,$E6,$30,$DD,$BF,$C8,$2D,$9D,$DF,$F6,$5B,$E7,$5F,$29
	.dc.b	$6F,$FD,$A7,$5F,$AF,$18,$BD,$63,$E1,$9C,$B7,$2E,$97,$C5,$B5,$C3
	.dc.b	$08,$FC,$BC,$3A,$36,$B8,$3F,$E5,$88,$F2,$BB,$BE,$CE,$67,$6A,$79
	.dc.b	$E1,$E4,$BE,$EF,$AA,$B7,$3A,$C3,$E1,$61,$D3,$AD,$CB,$FE,$FF,$CB
	.dc.b	$F8,$BF,$C7,$FF,$82,$F3,$FE,$2C,$DD,$FF,$DE,$FF,$29,$6B,$FD,$E5
	.dc.b	$37,$0E,$6D,$C3,$A6,$1F,$C7,$EF,$13,$2F,$F9,$43,$B8,$C4,$78,$8C
	.dc.b	$9F,$CB,$5E,$9B,$48,$FE,$60,$F7,$F2,$31,$1F,$CC,$22,$DF,$F2,$F9
	.dc.b	$B9,$7F,$A1,$B2,$F8,$35,$8B,$47,$C4,$AF,$67,$C9,$FF,$93,$67,$FA
	.dc.b	$B3,$5F,$EA,$AD,$C3,$F9,$E7,$0F,$0F,$52,$57,$07,$E2,$47,$4F,$8B
	.dc.b	$93,$D1,$EF,$E6,$62,$3A,$C5,$BF,$EC,$18,$7F,$8A,$A7,$81,$0D,$9E
	.dc.b	$6A,$E1,$A7,$BE,$57,$8A,$F7,$C7,$37,$D7,$F1,$3F,$00,$8F,$C0,$F8
	.dc.b	$BA,$E8,$F8,$71,$F1,$A3,$F0,$5E,$13,$E9,$69,$2B,$FB,$B7,$D2,$D7
	.dc.b	$FB,$94,$5D,$3F,$B8,$CF,$2F,$A5,$95,$FD,$F5,$EB,$CF,$88,$FE,$F5
	.dc.b	$16,$BF,$BC,$CF,$E5,$65,$F4,$8F,$8A,$C3,$63,$FD,$56,$D7,$FC,$3F
	.dc.b	$F4,$AF,$91,$DD,$49,$DB,$D6,$75,$1F,$EB,$4F,$A0,$FF,$41,$37,$DC
	.dc.b	$F1,$46,$23,$FC,$57,$F4,$BF,$C6,$9D,$10,$FF,$17,$C6,$F3,$FF,$5E
	.dc.b	$4E,$E7,$FC,$67,$F4,$BE,$06,$E5,$EF,$8C,$57,$FB,$0D,$7C,$BE,$66
	.dc.b	$57,$FB,$2A,$77,$CD,$FE,$89,$77,$19,$9D,$BD,$8E,$15,$F3,$7E,$97
	.dc.b	$FC,$4D,$1F,$ED,$46,$BF,$E4,$86,$E1,$FE,$D5,$FA,$57,$27,$1F,$42
	.dc.b	$5F,$17,$EC,$4E,$A3,$FD,$BF,$F2,$3E,$47,$FC,$A6,$DB,$FF,$70,$37
	.dc.b	$4F,$F5,$17,$C6,$7B,$28,$27,$93,$FE,$A7,$31,$9F,$FA,$85,$DC,$69
	.dc.b	$FF,$56,$35,$E7,$F4,$BF,$E4,$E7,$FD,$E6,$D7,$FC,$CE,$DC,$3F,$DE
	.dc.b	$8F,$FB,$66,$37,$EA,$FE,$56,$35,$8F,$FA,$E5,$B3,$DB,$D6,$D5,$48
	.dc.b	$E2,$78,$23,$F5,$FF,$4B,$FB,$0E,$39,$3F,$E6,$38,$CB,$FE,$65,$8C
	.dc.b	$BF,$E6,$18,$FD,$37,$D1,$BB,$69,$27,$E4,$E6,$CF,$FD,$9D,$F4,$AF
	.dc.b	$91,$D5,$7F,$A1,$B0,$E4,$F0,$AF,$F6,$A4,$62,$3F,$D1,$75,$C5,$FF
	.dc.b	$3C,$7A,$1F,$C2,$3E,$96,$46,$08,$7F,$9D,$DC,$BF,$E7,$98,$FE,$1E
	.dc.b	$63,$1F,$F3,$CC,$E2,$7C,$47,$E0,$AD,$F3,$1F,$00,$91,$A1,$F4,$BF
	.dc.b	$AF,$BF,$81,$1C,$67,$F0,$CA,$B9,$BF,$89,$6B,$FE,$EB,$6B,$FE,$F5
	.dc.b	$F2,$DF,$4A,$FA,$2F,$E4,$F3,$0F,$F5,$14,$65,$B1,$97,$FD,$30,$7F
	.dc.b	$6F,$8D,$3F,$EB,$6C,$7F,$C4,$EF,$AD,$AE,$7D,$2F,$E4,$90,$F6,$AC
	.dc.b	$49,$E0,$EA,$31,$1C,$4C,$9D,$5F,$17,$D0,$C8,$F4,$BE,$5F,$A2,$F3
	.dc.b	$78,$12,$7D,$AF,$A7,$AF,$D5,$EF,$FA,$FE,$8A,$57,$C2,$8F,$77,$C0
	.dc.b	$48,$F8,$AF,$C1,$F8,$1F,$85,$F0,$12,$7E,$4B,$B7,$FD,$93,$BA,$3E
	.dc.b	$47,$C0,$C9,$D9,$F4,$37,$F9,$84,$7E,$9E,$F5,$E7,$17,$97,$FD,$A5
	.dc.b	$87,$CB,$9E,$44,$36,$72,$3E,$EF,$C0,$7D,$92,$3F,$80,$F7,$AB,$E8
	.dc.b	$8B,$5D,$A6,$F5,$FF,$86,$6D,$58,$BF,$F0,$CC,$7F,$93,$AF,$CB,$ED
	.dc.b	$FE,$1C,$93,$F8,$EF,$81,$AF,$C6,$8B,$3E,$C9,$B8,$77,$9B,$5D,$A7
	.dc.b	$F2,$5E,$DC,$8F,$CC,$BF,$95,$F7,$71,$1F,$4A,$2E,$1F,$22,$6E,$1E
	.dc.b	$B9,$B5,$EA,$9F,$CF,$7B,$72,$3F,$4A,$FE,$87,$DD,$C4,$7D,$E8,$B8
	.dc.b	$7D,$49,$B8,$7C,$79,$B5,$F1,$A7,$FC,$B1,$F4,$BF,$E2,$1B,$FD,$FF
	.dc.b	$6A,$31,$87,$F9,$1D,$B7,$FE,$46,$7F,$C8,$11,$9F,$77,$99,$27,$F8
	.dc.b	$8D,$73,$F7,$BF,$C9,$7F,$43,$27,$FC,$96,$FF,$93,$BE,$8F,$11,$FE
	.dc.b	$4C,$8B,$97,$F9,$02,$7F,$CA,$7F,$1B,$2B,$FC,$A6,$FF,$95,$BD,$3C
	.dc.b	$47,$F9,$52,$2D,$7F,$92,$27,$FC,$B7,$E2,$CA,$FF,$2D,$BF,$E5,$EF
	.dc.b	$A2,$C4,$7F,$97,$22,$D7,$F9,$42,$6E,$BF,$E4,$B9,$B5,$FE,$4A,$9F
	.dc.b	$F3,$3F,$A5,$FD,$C2,$1F,$0C,$E3,$FF,$29,$79,$99,$8C,$4F,$F9,$4A
	.dc.b	$31,$3F,$E5,$7A,$B3,$FE,$56,$6F,$03,$E0,$BF,$A5,$4F,$6A,$EA,$FF
	.dc.b	$B8,$5E,$4D,$3F,$C0,$7F,$4A,$FA,$2F,$F1,$64,$79,$E3,$1C,$FF,$E0
	.dc.b	$36,$C7,$F8,$1C,$38,$7F,$C1,$5F,$0F,$FC,$AF,$E7,$FF,$C1,$4D,$8E
	.dc.b	$0D,$77,$E0,$A7,$B7,$B0,$F4,$69,$2B,$C1,$87,$F8,$66,$3C,$5A,$FC
	.dc.b	$1F,$80,$FE,$AB,$C3,$F2,$A3,$1C,$9E,$F7,$E3,$7C,$A6,$B8,$3F,$D0
	.dc.b	$57,$72,$FF,$40,$B6,$B8,$C7,$CB,$3F,$D0,$B5,$9E,$1F,$F3,$56,$1D
	.dc.b	$5F,$94,$FF,$A9,$3E,$8F,$FB,$3F,$3F,$FA,$5B,$8E,$CF,$FA,$46,$6E
	.dc.b	$1F,$E9,$09,$B8,$FD,$B9,$B8,$74,$AF,$F5,$6C,$62,$B2,$DB,$F0,$1F
	.dc.b	$EB,$4F,$A3,$E3,$C1,$C9,$B1,$93,$FD,$73,$19,$69,$FB,$6E,$F7,$67
	.dc.b	$8B,$E3,$BC,$B1,$89,$FF,$60,$C6,$27,$51,$C5,$B9,$35,$FE,$17,$7F
	.dc.b	$E6,$0A,$ED,$DD,$D3,$B7,$AB,$63,$27,$8E,$32,$F9,$7B,$3C,$7D,$DE
	.dc.b	$46,$C6,$4F,$54,$65,$A6,$71,$F2,$3F,$C6,$E3,$51,$83,$F8,$E3,$F0
	.dc.b	$3B,$7E,$0F,$75,$AF,$8C,$DA,$E1,$C7,$FC,$9E,$31,$A7,$C0,$6D,$7F
	.dc.b	$29,$6D,$7F,$29,$7E,$6F,$27,$CE,$EE,$B5,$F4,$DB,$5D,$B6,$46,$30
	.dc.b	$F9,$AD,$AF,$E6,$CD,$AF,$E6,$CF,$DA,$E3,$FB,$7C,$D6,$BF,$00,$DA
	.dc.b	$EC,$DB,$FB,$2D,$9F,$E8,$4D,$9F,$E8,$4F,$E1,$B8,$FF,$0F,$CB,$6B
	.dc.b	$F1,$8D,$AF,$53,$6F,$F0,$AD,$9F,$E9,$CD,$9F,$E9,$CF,$E4,$B9,$3F
	.dc.b	$27,$CB,$6B,$F3,$0F,$E5,$9C,$57,$C7,$6D,$FF,$5C,$FD,$89,$F4,$9B
	.dc.b	$9F,$EA,$B1,$87,$F4,$47,$ED,$17,$D1,$5F,$A4,$C3,$67,$8F,$CE,$EF
	.dc.b	$E9,$38,$36,$3A,$FC,$C7,$3F,$8A,$E8,$D8,$FF,$3A,$FC,$FF,$F2,$3E
	.dc.b	$2F,$E0,$AF,$F2,$5E,$6B,$E2,$7F,$10,$E0,$FC,$B4,$E6,$F8,$CE,$78
	.dc.b	$3F,$CE,$F3,$9A,$DC,$E6,$BF,$D8,$73,$9A,$DC,$E6,$93,$9A,$DC,$E6
	.dc.b	$31,$3F,$EC,$CF,$A2,$F8,$B5,$FE,$8E,$B7,$C5,$BE,$53,$F8,$FD,$AF
	.dc.b	$6F,$EB,$EB,$73,$A9,$CC,$FF,$B5,$5A,$FF,$6B,$6B,$E1,$E0,$FF,$4A
	.dc.b	$BF,$E8,$DE,$27,$D0,$FD,$9F,$A2,$FB,$4D,$CF,$EF,$B6,$BF,$73,$5C
	.dc.b	$27,$F0,$75,$F8,$1F,$8A,$91,$F8,$6F,$8A,$B7,$C2,$6E,$9F,$8D,$9B
	.dc.b	$5F,$EF,$0F,$01,$8A,$F3,$82,$73,$78,$FA,$47,$CB,$39,$BF,$25,$F1
	.dc.b	$FF,$D2,$27,$31,$8C,$FF,$D7,$B1,$86,$D7,$17,$F3,$BF,$35,$C6,$31
	.dc.b	$97,$C0,$7F,$3C,$7E,$7F,$E5,$ED,$7E,$92,$2C,$FC,$69,$BB,$7E,$3A
	.dc.b	$1B,$FE,$61,$74,$D8,$C3,$70,$FC,$FC,$E9,$B3,$A6,$9A,$9E,$13,$FA
	.dc.b	$03,$81,$F6,$3E,$5F,$F6,$DF,$2F,$FB,$76,$D7,$EF,$1B,$3B,$8F,$DF
	.dc.b	$CF,$E5,$3D,$97,$95,$FD,$D3,$D2,$7D,$03,$CD,$D4,$E6,$31,$87,$FB
	.dc.b	$9E,$F8,$B6,$75,$E6,$70,$8F,$F7,$88,$FA,$07,$B9,$FE,$02,$8C,$7B
	.dc.b	$F6,$73,$1C,$1F,$F4,$EB,$67,$53,$B6,$C6,$4E,$11,$96,$9F,$EF,$DF
	.dc.b	$8F,$FE,$3D,$B7,$F5,$FD,$5F,$3E,$EA,$FD,$BD,$6B,$93,$FE,$00,$D6
	.dc.b	$A7,$3E,$54,$C5,$8F,$3C,$7F,$3D,$1D,$E7,$F2,$1C,$7F,$CA,$FE,$52
	.dc.b	$CF,$F3,$06,$DF,$F7,$1E,$AF,$D0,$D7,$F3,$1F,$94,$CE,$9F,$CE,$74
	.dc.b	$FE,$69,$AF,$F3,$56,$E5,$FC,$E9,$B8,$FE,$F1,$B5,$FE,$9C,$F8,$87
	.dc.b	$C8,$FF,$3D,$AF,$F1,$8C,$62,$FF,$A1,$C6,$27,$51,$FE,$9A,$9D,$4F
	.dc.b	$F0,$83,$FC,$73,$F2,$9F,$1E,$7F,$80,$7B,$67,$3B,$FF,$8F,$1A,$7F
	.dc.b	$A9,$23,$4F,$FA,$9A,$79,$57,$F0,$59,$CB,$FE,$0E,$8F,$E5,$38,$FF
	.dc.b	$AD,$97,$A1,$E2,$F1,$FC,$6D,$C3,$A3,$6F,$93,$6F,$93,$67,$F8,$C5
	.dc.b	$F4,$7C,$97,$D1,$A7,$43,$D1,$E7,$FC,$2F,$43,$E1,$B6,$32,$7B,$63
	.dc.b	$2D,$3E,$07,$7F,$CA,$F1,$3E,$5B,$63,$27,$C5,$8C,$B4,$CE,$7F,$16
	.dc.b	$32,$D8,$F9,$9E,$4F,$D6,$F0,$BE,$BB,$63,$27,$D1,$8C,$B4,$CC,$FD
	.dc.b	$18,$CB,$63,$3A,$FD,$39,$CB,$63,$F4,$AF,$F5,$69,$FD,$3F,$BB,$FA
	.dc.b	$8F,$9D,$78,$9F,$EC,$79,$FF,$51,$5E,$63,$FD,$61,$56,$7F,$D9,$9C
	.dc.b	$BF,$8D,$57,$C9,$E5,$3B,$7F,$6B,$3B,$9D,$47,$FB,$44,$E3,$5E,$19
	.dc.b	$D3,$E0,$8C,$9E,$0B,$E8,$FA,$8C,$9D,$9C,$BF,$EE,$03,$27,$B0,$A6
	.dc.b	$6B,$C6,$E5,$AE,$67,$C9,$3D,$2F,$C3,$7E,$6A,$F7,$8F,$82,$7C,$2F
	.dc.b	$A5,$B1,$A6,$BF,$05,$BF,$06,$F8,$9F,$F4,$C4,$F3,$AF,$F7,$D4,$E7
	.dc.b	$4F,$E3,$F7,$C2,$F6,$D8,$CD,$70,$9E,$7A,$7D,$0B,$7C,$4F,$F2,$8C
	.dc.b	$7E,$4C,$61,$FB,$0F,$77,$35,$CB,$5F,$1D,$71,$BE,$8D,$8F,$06,$3C
	.dc.b	$AB,$85,$FF,$36,$C3,$B3,$E0,$BE,$0E,$67,$6A,$7C,$58,$76,$8F,$2D
	.dc.b	$70,$75,$5C,$2B,$78,$79,$A7,$67,$8F,$0F,$CD,$79,$CE,$6F,$F1,$24
	.dc.b	$FF,$21,$F0,$BE,$25,$7F,$1D,$6F,$07,$F1,$4F,$3E,$D7,$F8,$A6,$F8
	.dc.b	$BF,$C9,$6B,$38,$75,$6C,$EB,$0F,$E5,$17,$89,$97,$FC,$C6,$36,$61
	.dc.b	$E8,$64,$FE,$6D,$E7,$74,$F3,$E4,$60,$87,$F3,$77,$2F,$F3,$98,$F1
	.dc.b	$98,$BF,$E7,$53,$89,$FE,$85,$E7,$FA,$3C,$F9,$18,$21,$FD,$0D,$CB
	.dc.b	$FD,$16,$3D,$FE,$27,$C0,$DA,$4F,$F4,$D8,$CB,$FD,$3E,$32,$FF,$4D
	.dc.b	$8C,$BF,$D3,$E3,$38,$7F,$4F,$9C,$BF,$D4,$63,$3B,$7F,$48,$9F,$9F
	.dc.b	$E2,$7D,$0F,$1A,$61,$FD,$76,$32,$FF,$5F,$8C,$BF,$D7,$63,$EC,$F8
	.dc.b	$6F,$92,$BF,$25,$2B,$FB,$94,$68,$C3,$FE,$F2,$94,$E6,$3F,$BC,$7C
	.dc.b	$97,$4F,$92,$91,$82,$1C,$5C,$BC,$E3,$F1,$FE,$53,$F6,$37,$C4,$61
	.dc.b	$B5,$FE,$72,$6B,$E3,$F9,$3D,$4F,$B3,$E4,$E4,$EE,$3F,$36,$BF,$63
	.dc.b	$FF,$7E,$BF,$D2,$38,$7F,$83,$46,$21,$C7,$FE,$A0,$8C,$3F,$EA,$2E
	.dc.b	$5F,$F6,$BC,$F1,$3F,$C6,$B3,$AA,$FE,$21,$3F,$EE,$29,$D0,$70,$F6
	.dc.b	$8F,$0F,$C3,$79,$BD,$DF,$E5,$D3,$C1,$FE,$57,$87,$FA,$C9,$B5,$FC
	.dc.b	$75,$DC,$F3,$3F,$C8,$EE,$27,$87,$85,$D4,$ED,$E1,$CA,$ED,$19,$7C
	.dc.b	$11,$F2,$3E,$69,$E6,$52,$4F,$B3,$E6,$9F,$0B,$88,$F1,$08,$7C,$3F
	.dc.b	$D2,$6C,$7C,$4B,$EA,$FF,$90,$7E,$5B,$F2,$BD,$32,$30,$43,$F2,$4E
	.dc.b	$5F,$CA,$47,$DE,$F9,$A7,$65,$ED,$95,$F4,$A7,$2D,$9C,$BE,$88,$4F
	.dc.b	$39,$93,$C6,$37,$BF,$FB,$92,$B9,$CF,$E9,$7E,$5F,$F5,$A4,$7F,$CB
	.dc.b	$9F,$2F,$5D,$1F,$E5,$F8,$FD,$9C,$7E,$8B,$F6,$4F,$E7,$13,$FE,$D0
	.dc.b	$6B,$FE,$79,$6D,$FF,$9B,$FE,$53,$87,$93,$FB,$CF,$32,$5F,$18,$FD
	.dc.b	$FF,$BB,$23,$FC,$DC,$FF,$02,$F3,$B1,$1F,$BD,$8B,$87,$F9,$EE,$6F
	.dc.b	$4F,$FB,$BF,$85,$58,$BF,$F7,$74,$7F,$60,$BF,$F4,$17,$CB,$FF,$0E
	.dc.b	$F7,$E5,$7F,$9F,$63,$F8,$89,$87,$F8,$64,$61,$FD,$3F,$CD,$3E,$0B
	.dc.b	$FC,$71,$1F,$DE,$E3,$0E,$03,$87,$6D,$7F,$C8,$0E,$6F,$FC,$79,$0F
	.dc.b	$9A,$7C,$2F,$F0,$66,$C7,$8A,$F1,$8E,$71,$93,$FB,$CA,$4E,$A3,$13
	.dc.b	$A7,$F4,$73,$A7,$84,$E7,$0F,$1C,$FF,$B6,$27,$27,$FA,$8F,$0F,$E6
	.dc.b	$FE,$1C,$8C,$1F,$EA,$7E,$EF,$E7,$7D,$F2,$30,$7F,$15,$F4,$DF,$A7
	.dc.b	$3E,$6B,$FB,$CA,$F9,$AA,$62,$33,$F1,$BA,$E6,$98,$73,$B4,$E2,$3F
	.dc.b	$88,$61,$DB,$E6,$7A,$99,$DF,$FC,$D5,$19,$9E,$5F,$41,$E2,$36,$F5
	.dc.b	$F9,$9F,$2F,$CD,$48,$C1,$0E,$F9,$23,$51,$89,$F6,$98,$C3,$C3,$5B
	.dc.b	$9E,$51,$8F,$8F,$F8,$7F,$1D,$27,$DD,$1F,$18,$C3,$E0,$8C,$3A,$3E
	.dc.b	$23,$A9,$E9,$1F,$1F,$E6,$BE,$7F,$44,$8C,$10,$F5,$B9,$7D,$B1,$F5
	.dc.b	$8C,$4F,$AA,$71,$3F,$4F,$E5,$7E,$DF,$9D,$23,$04,$3E,$3B,$97,$E4
	.dc.b	$C7,$E0,$8C,$57,$C6,$9C,$4F,$DF,$F9,$9F,$C3,$FC,$BC,$8C,$10,$FA
	.dc.b	$6E,$5F,$AB,$1F,$8E,$31,$5F,$4A,$71,$3F,$8B,$F3,$7F,$27,$F3,$32
	.dc.b	$30,$43,$EF,$B9,$7F,$03,$1F,$9A,$F3,$FF,$CA,$3F,$23,$27,$F2,$D1
	.dc.b	$97,$FE,$00,$46,$5F,$C4,$C7,$F9,$61,$FF,$2C,$FC,$34,$7F,$51,$D1
	.dc.b	$FE,$C7,$CB,$F3,$2D,$F1,$7F,$DF,$35,$BA,$FD,$0C,$E5,$73,$BF,$18
	.dc.b	$FF,$13,$C7,$EC,$D7,$C9,$7F,$9A,$9B,$19,$F7,$B0,$66,$64,$64,$87
	.dc.b	$EC,$27,$8B,$63,$8B,$DF,$DE,$FF,$51,$7C,$4F,$F8,$AB,$0F,$24,$63
	.dc.b	$C3,$57,$A3,$04,$E6,$31,$F0,$16,$F8,$C7,$FB,$BA,$F8,$C6,$21,$87
	.dc.b	$7C,$39,$C6,$CE,$65,$1F,$38,$E3,$FF,$0D,$C7,$FA,$E6,$31,$E2,$FF
	.dc.b	$3F,$6C,$7F,$44,$6B,$FE,$FD,$F8,$97,$A0,$F9,$AF,$E8,$D1,$86,$C6
	.dc.b	$67,$FA,$6C,$E5,$E9,$CE,$E3,$F0,$36,$33,$18,$3F,$AA,$EA,$F9,$37
	.dc.b	$4B,$9B,$E0,$4E,$63,$17,$F1,$B8,$7E,$4D,$BB,$33,$C5,$F5,$E3,$9E
	.dc.b	$3C,$9B,$7C,$4F,$C7,$B8,$9E,$15,$CA,$FA,$E1,$6F,$7C,$3E,$2E,$8F
	.dc.b	$E6,$B8,$BA,$F1,$9F,$92,$AD,$F6,$9B,$A7,$BE,$6D,$F1,$BF,$EA,$0F
	.dc.b	$DB,$6D,$6A,$76,$D8,$C9,$C3,$E6,$F9,$EF,$D1,$B1,$93,$AC,$65,$A7
	.dc.b	$F5,$FF,$9C,$78,$FE,$4E,$DF,$3C,$62,$7C,$B1,$89,$D4,$71,$6E,$9F
	.dc.b	$D7,$29,$E2,$F2,$F1,$B9,$B6,$FC,$0D,$AE,$0D,$DF,$FB,$0D,$BC,$5F
	.dc.b	$1F,$8B,$E4,$6E,$1E,$96,$E1,$E1,$6D,$F2,$6C,$FE,$C2,$FA,$3F,$06
	.dc.b	$FD,$FE,$2F,$C0,$F1,$6C,$FC,$56,$E3,$ED,$6E,$3E,$66,$CF,$56,$CF
	.dc.b	$F1,$CC,$3F,$C6,$93,$F3,$F9,$BE,$6F,$8B,$F3,$9B,$19,$3E,$A4,$65
	.dc.b	$A6,$6B,$E7,$46,$5B,$1F,$6F,$8B,$EC,$F3,$7D,$A6,$C6,$4F,$C0,$46
	.dc.b	$5A,$66,$7E,$D4,$65,$B1,$F8,$7E,$2F,$C2,$F2,$7E,$19,$B1,$93,$F1
	.dc.b	$91,$96,$99,$9F,$C3,$46,$5B,$1F,$93,$E3,$FC,$8F,$27,$E4,$9B,$1F
	.dc.b	$97,$30,$53,$81,$FE,$39,$F9,$A7,$CB,$FE,$6A,$3F,$63,$18,$78,$5F
	.dc.b	$F8,$CA,$1F,$2D,$AF,$97,$DB,$6B,$A7,$C1,$72,$F9,$7E,$6D,$8F,$07
	.dc.b	$CB,$F6,$F9,$7E,$ED,$8F,$F7,$07,$CE,$7F,$2D,$E1,$FE,$0A,$FF,$2F
	.dc.b	$E4,$BE,$2F,$F1,$6E,$8F,$E3,$F3,$9B,$E3,$39,$7F,$A8,$C6,$77,$FF
	.dc.b	$6D,$3D,$63,$77,$D3,$0E,$AD,$8C,$9F,$EE,$EF,$9C,$F8,$2F,$FA,$15
	.dc.b	$B7,$FE,$8A,$6B,$FD,$80,$FE,$C2,$E1,$C3,$C3,$1F,$F4,$3B,$67,$FB
	.dc.b	$3F,$C7,$3E,$A7,$FD,$13,$1E,$1E,$0F,$AB,$F3,$5F,$5B,$E6,$AC,$FD
	.dc.b	$B6,$E5,$F5,$DB,$97,$F9,$9B,$91,$1F,$7B,$E6,$25,$FE,$0F,$E6,$6E
	.dc.b	$1F,$7E,$6E,$1C,$66,$D7,$F7,$4C,$15,$8B,$FF,$0A,$C2,$B1,$A3,$E5
	.dc.b	$DC,$9F,$8E,$F9,$8F,$E8,$33,$98,$C6,$5F,$E0,$8C,$FF,$AD,$CE,$63
	.dc.b	$07,$C6,$BE,$6F,$E6,$FE,$67,$9C,$63,$3F,$CC,$3F,$9C,$3F,$3B,$F2
	.dc.b	$56,$BF,$45,$16,$7F,$3D,$37,$0F,$C6,$C7,$C8,$AF,$CA,$AE,$7F,$CA
	.dc.b	$B9,$3F,$1A,$66,$37,$BE,$A7,$6D,$8C,$9D,$23,$2D,$3F,$6D,$F3,$1F
	.dc.b	$B4,$F9,$9F,$DA,$B6,$32,$7E,$EE,$32,$D3,$E3,$DF,$F8,$CB,$E0,$9E
	.dc.b	$47,$F7,$0F,$45,$F3,$BF,$CF,$A9,$A7,$FB,$6A,$F7,$87,$2B,$DF,$97
	.dc.b	$FD,$E2,$FF,$DB,$B8,$DB,$DC,$72,$BF,$F2,$0D,$F6,$F2,$B5,$7F,$E0
	.dc.b	$09,$7C,$EB,$E1,$F0,$BE,$2A,$C6,$1B,$3E,$C3,$8B,$8C,$7C,$07,$F8
	.dc.b	$96,$B9,$4E,$DE,$CD,$9E,$1A,$7F,$8A,$23,$99,$F2,$29,$F0,$1F,$C8
	.dc.b	$DE,$B9,$BE,$77,$F3,$D5,$BA,$D5,$FF,$8C,$AB,$5D,$5F,$1E,$73,$5F
	.dc.b	$C1,$5B,$87,$F9,$0F,$0F,$D5,$CF,$3F,$23,$F3,$7C,$7F,$CD,$3E,$4A
	.dc.b	$CF,$F3,$96,$DF,$C0,$EA,$FD,$75,$7F,$3A,$F9,$3C,$EB,$FC,$07,$5F
	.dc.b	$E7,$BA,$7F,$3E,$6E,$5F,$D0,$DB,$8F,$F3,$F6,$D7,$FB,$BB,$E3,$7F
	.dc.b	$D1,$B5,$FD,$26,$BF,$C6,$F1,$87,$FA,$31,$8F,$82,$FE,$96,$D9,$E0
	.dc.b	$7F,$A8,$B5,$FF,$1E,$BF,$E4,$0F,$84,$F8,$7E,$D7,$F9,$7E,$31,$3A
	.dc.b	$78,$1F,$E4,$23,$87,$8B,$C7,$CE,$B1,$FE,$AE,$6E,$1F,$EA,$A6,$E7
	.dc.b	$FE,$B0,$6C,$FF,$A3,$B1,$46,$2B,$FD,$5F,$18,$9F,$F5,$CD,$F7,$69
	.dc.b	$DC,$FF,$5E,$F9,$BE,$BF,$3B,$D8,$D8,$C9,$FE,$C2,$8C,$B4,$FF,$62
	.dc.b	$F7,$FC,$3F,$13,$E2,$36,$32,$7F,$B2,$A3,$2D,$33,$8F,$FB,$22,$32
	.dc.b	$D8,$FF,$68,$F9,$1F,$3F,$C2,$FA,$0D,$8C,$9F,$ED,$68,$CB,$4C,$CF
	.dc.b	$FB,$52,$32,$D8,$CE,$BF,$ED,$C9,$CB,$63,$F5,$93,$FE,$C4,$F7,$3F
	.dc.b	$D8,$BF,$3A,$F1,$3F,$B4,$4F,$FA,$96,$B3,$18,$8F,$F6,$08,$9E,$0F
	.dc.b	$F6,$FE,$5F,$F5,$0C,$62,$3F,$DA,$A6,$E7,$AC,$E9,$E9,$7D,$A7,$95
	.dc.b	$D7,$8D,$F3,$7C,$87,$33,$CF,$3B,$73,$1B,$7C,$EF,$1B,$F0,$47,$03
	.dc.b	$91,$FC,$9A,$B9,$3C,$2F,$CF,$18,$BE,$D5,$CA,$BD,$BF,$02,$F8,$BF
	.dc.b	$5C,$23,$0F,$76,$BC,$23,$FC,$2D,$F0,$0F,$90,$FE,$83,$3C,$67,$31
	.dc.b	$8A,$E3,$5B,$D3,$FD,$FB,$18,$9D,$57,$09,$F9,$54,$F9,$0F,$F1,$26
	.dc.b	$BF,$56,$31,$8E,$B0,$F0,$CE,$5B,$AF,$9A,$F8,$B6,$B8,$61,$1F,$97
	.dc.b	$87,$46,$D7,$04,$23,$D4,$F0,$BE,$EE,$A3,$96,$1E,$7A,$E5,$3C,$6B
	.dc.b	$6D,$BF,$7C,$EA,$F8,$E1,$F0,$EB,$F1,$1D,$0E,$6F,$F1,$4C,$7F,$76
	.dc.b	$8C,$78,$3F,$E4,$9C,$7F,$DF,$71,$89,$D5,$70,$7F,$AF,$C7,$3F,$3F
	.dc.b	$87,$0F,$F1,$BC,$2D,$7F,$1A,$6D,$6E,$39,$D6,$2F,$FC,$71,$1D,$AB
	.dc.b	$13,$23,$04,$3C,$33,$E3,$C3,$CE,$BC,$DE,$9E,$74,$AF,$4C,$77,$30
	.dc.b	$F2,$8C,$3A,$3D,$EE,$AB,$AC,$F9,$FC,$EF,$47,$9D,$2F,$E2,$C6,$5F
	.dc.b	$8F,$19,$7C,$F1,$EF,$F0,$FE,$06,$D2,$7E,$6C,$65,$F9,$F1,$97,$DF
	.dc.b	$19,$7E,$0C,$67,$0F,$83,$39,$7E,$1C,$67,$6F,$E9,$D3,$F3,$FC,$3F
	.dc.b	$A1,$E2,$CC,$3E,$EC,$65,$FB,$F1,$97,$E7,$C7,$DA,$F0,$DF,$0D,$7E
	.dc.b	$22,$57,$E2,$E7,$2D,$9C,$BF,$88,$84,$F3,$9E,$9F,$3B,$D7,$E2,$65
	.dc.b	$F3,$8C,$BD,$63,$2F,$38,$FC,$7F,$92,$F9,$D7,$C3,$E1,$B5,$FE,$54
	.dc.b	$9D,$4F,$AF,$E7,$74,$FB,$7E,$76,$56,$EF,$8B,$FE,$3E,$1F,$3A,$F8
	.dc.b	$4F,$CF,$C6,$1B,$19,$8F,$F1,$E7,$37,$FB,$69,$FF,$1F,$46,$5E,$91
	.dc.b	$FE,$E4,$77,$E6,$F2,$AE,$BF,$3B,$FC,$B6,$7F,$C8,$CF,$F2,$CE,$77
	.dc.b	$37,$FB,$06,$30,$DF,$13,$89,$CB,$C4,$97,$CA,$32,$F4,$8C,$BE,$FA
	.dc.b	$F0,$3F,$07,$E7,$1E,$6F,$C0,$6C,$7F,$96,$5A,$FC,$7F,$9D,$7C,$1E
	.dc.b	$EB,$E3,$C3,$E1,$A9,$9A,$FC,$57,$CE,$FE,$3F,$AE,$46,$08,$7E,$2D
	.dc.b	$CB,$F8,$D8,$FA,$7F,$3A,$ED,$BD,$D2,$BF,$CD,$D3,$96,$CE,$5F,$BD
	.dc.b	$09,$E7,$32,$78,$C6,$F8,$7E,$8C,$7F,$3E,$AE,$8F,$E7,$FE,$77,$F5
	.dc.b	$04,$7F,$4D,$F3,$B5,$D1,$FA,$98,$FD,$74,$7E,$73,$E7,$7F,$9C,$4F
	.dc.b	$E6,$5A,$FF,$9C,$5B,$7F,$E6,$7F,$93,$E1,$E4,$7E,$DF,$CE,$97,$C6
	.dc.b	$3F,$75,$EE,$48,$FD,$C3,$FB,$DF,$3F,$11,$FB,$78,$B8,$7F,$9C,$E6
	.dc.b	$F4,$FF,$5C,$F3,$9C,$5F,$EA,$9C,$39,$9F,$C0,$EB,$F8,$47,$CE,$7F
	.dc.b	$09,$F7,$E4,$FF,$07,$8C,$BF,$C2,$63,$2F,$F0,$78,$FD,$6F,$CE,$BE
	.dc.b	$0A,$92,$7F,$D5,$1F,$3B,$F9,$6D,$3E,$9C,$2B,$83,$FE,$3B,$31,$55
	.dc.b	$A6,$63,$6F,$F9,$06,$97,$FB,$E7,$FD,$91,$1E,$0C,$31,$32,$30,$43
	.dc.b	$F8,$44,$F2,$7C,$F1,$CE,$64,$60,$FE,$55,$F0,$9F,$CB,$FC,$79,$18
	.dc.b	$3F,$97,$78,$3F,$CC,$FC,$19,$18,$3F,$BE,$7C,$CB,$E7,$3F,$B8,$AF
	.dc.b	$9D,$A6,$23,$3F,$3B,$AE,$69,$7C,$A2,$BF,$E0,$0B,$7C,$CF,$43,$3C
	.dc.b	$5F,$E0,$88,$CC,$F8,$BE,$77,$C2,$6D,$F2,$FC,$CF,$93,$E6,$A4,$60
	.dc.b	$87,$77,$2F,$86,$3D,$66,$2F,$C1,$38,$9F,$83,$F1,$BF,$03,$E3,$64
	.dc.b	$60,$87,$AD,$CB,$ED,$8F,$8E,$62,$BD,$93,$89,$F9,$BF,$35,$F3,$3A
	.dc.b	$24,$60,$87,$C7,$72,$FC,$98,$FA,$66,$2B,$E4,$4E,$27,$EC,$FC,$9F
	.dc.b	$D8,$F3,$A4,$60,$87,$D3,$72,$FD,$58,$FB,$E6,$2B,$EA,$4E,$27,$F0
	.dc.b	$BF,$33,$F8,$4F,$95,$91,$82,$1F,$7D,$CB,$F8,$18,$FC,$59,$8A,$FC
	.dc.b	$04,$E2,$7F,$23,$E5,$FE,$43,$E6,$64,$60,$87,$E2,$DC,$BF,$8D,$8F
	.dc.b	$CB,$F9,$FF,$E4,$5F,$90,$93,$F9,$C8,$CB,$FE,$54,$8C,$BF,$95,$8F
	.dc.b	$F2,$63,$FE,$4D,$F8,$78,$FF,$96,$8C,$9F,$E5,$C7,$2F,$F9,$30,$FF
	.dc.b	$2B,$FC,$EF,$F8,$59,$FF,$33,$2F,$91,$C9,$AF,$80,$93,$A8,$CB,$C2
	.dc.b	$32,$FF,$9A,$27,$93,$B3,$47,$6F,$7F,$FD,$0E,$BE,$2B,$FC,$33,$E4
	.dc.b	$A9,$5A,$9C,$C6,$2B,$FD,$BD,$E1,$7F,$35,$8F,$F1,$9C,$73,$8C,$57
	.dc.b	$16,$C7,$01,$7D,$E7,$51,$8A,$E7,$7F,$E2,$4F,$9D,$6F,$FE,$06,$AF
	.dc.b	$4C,$63,$C3,$FE,$60,$D8,$FF,$76,$35,$FF,$78,$FC,$53,$CF,$7C,$D7
	.dc.b	$F3,$78,$C3,$63,$33,$FE,$F5,$9C,$BD,$38,$DC,$3E,$06,$C6,$63,$07
	.dc.b	$F4,$3F,$5B,$FC,$37,$E7,$DF,$67,$FA,$4D,$7D,$29,$CB,$F5,$0F,$F5
	.dc.b	$DB,$87,$0E,$BA,$3F,$A8,$B6,$BE,$97,$C0,$FF,$81,$A7,$37,$67,$B7
	.dc.b	$C6,$DC,$3E,$77,$C2,$57,$FC,$35,$19,$E5,$FC,$14,$E7,$24,$FA,$2B
	.dc.b	$CA,$D7,$8D,$FB,$E3,$13,$C2,$3F,$DA,$B8,$FA,$27,$B7,$23,$E0,$FB
	.dc.b	$47,$7F,$81,$97,$FA,$4F,$56,$E5,$FE,$65,$C9,$58,$CB,$FD,$97,$97
	.dc.b	$F8,$1B,$0E,$E7,$CD,$9E,$FE,$EF,$BF,$F2,$9D,$63,$FA,$78,$C6,$5A
	.dc.b	$BE,$33,$FE,$EC,$AD,$E3,$DA,$F7,$5A,$A7,$E5,$3F,$D8,$63,$8F,$33
	.dc.b	$E0,$9F,$21,$F3,$EB,$56,$C7,$FC,$7B,$7B,$AD,$7A,$7F,$36,$F7,$7F
	.dc.b	$85,$6B,$FB,$A6,$A3,$C7,$E9,$FD,$FB,$F1,$7C,$6F,$86,$31,$92,$31
	.dc.b	$9F,$A2,$30,$FB,$70,$FE,$FC,$E6,$77,$87,$FA,$5E,$72,$FA,$E7,$C5
	.dc.b	$87,$B7,$E3,$F7,$A7,$F8,$A2,$DE,$67,$97,$DE,$7A,$FF,$16,$30,$DC
	.dc.b	$7E,$5C,$E9,$F2,$4E,$9A,$4C,$BC,$47,$EF,$B2,$FF,$5B,$C6,$71,$F1
	.dc.b	$79,$9F,$17,$EE,$DE,$BF,$D5,$8C,$36,$BE,$DC,$E9,$F8,$93,$A6,$93
	.dc.b	$0F,$0E,$1F,$4F,$D5,$E5,$1F,$EB,$DF,$C9,$FE,$6B,$AF,$C2,$7D,$8E
	.dc.b	$78,$7E,$C4,$63,$4F,$F2,$7D,$70,$7F,$0F,$19,$C7,$F9,$DC,$E5,$FA
	.dc.b	$31,$9B,$FC,$24,$E5,$B7,$FC,$07,$F2,$8F,$C9,$E1,$DA,$3F,$82,$53
	.dc.b	$E4,$BF,$15,$D7,$F0,$70,$F5,$46,$37,$DB,$73,$FD,$25,$F1,$7F,$47
	.dc.b	$7C,$71,$F5,$D7,$09,$FD,$56,$4F,$73,$19,$7C,$AF,$BB,$E4,$DB,$F2
	.dc.b	$FA,$7F,$9D,$9C,$E3,$F1,$A3,$13,$AC,$7F,$67,$5A,$7F,$B0,$FC,$65
	.dc.b	$C3,$E2,$DF,$F2,$38,$C4,$34,$4F,$F2,$9F,$D2,$5F,$B2,$FE,$F3,$0F
	.dc.b	$E0,$F1,$8E,$EF,$27,$C6,$6E,$B5,$8F,$E0,$7C,$BC,$E1,$F1,$9B,$1F
	.dc.b	$F0,$07,$3F,$F6,$24,$FF,$0D,$6C,$7E,$72,$1E,$A7,$FB,$45,$D6,$1F
	.dc.b	$CE,$FE,$89,$F3,$9F,$CF,$F8,$7F,$89,$72,$E6,$31,$E0,$FD,$3C,$3F
	.dc.b	$C0,$77,$C1,$FE,$43,$7C,$1F,$E3,$D1,$9C,$BF,$5B,$39,$7F,$5E,$23
	.dc.b	$4F,$F5,$B6,$BD,$AF,$FA,$44,$62,$3F,$C0,$7F,$93,$75,$AF,$F6,$FF
	.dc.b	$9A,$87,$CF,$7F,$55,$75,$8F,$FB,$15,$CB,$FC,$E2,$78,$63,$BA,$E0
	.dc.b	$D9,$D5,$7F,$69,$C3,$C3,$F3,$DD,$7D,$0E,$CD,$AD,$B6,$7F,$86,$E3
	.dc.b	$E5,$8F,$ED,$9F,$49,$FE,$0C,$AB,$E0,$3D,$DB,$18,$6B,$FD,$3A,$7F
	.dc.b	$C2,$71,$8C,$3F,$8C,$C6,$27,$54,$79,$18,$2C,$FD,$2D,$9F,$33,$FD
	.dc.b	$FF,$95,$F4,$5F,$CD,$B8,$7F,$DB,$35,$AC,$3F,$AF,$D6,$BC,$BF,$C5
	.dc.b	$61,$F5,$89,$3F,$ED,$DA,$E1,$8F,$F8,$5F,$6F,$AD,$F4,$1F,$57,$E8
	.dc.b	$3E,$B3,$6F,$E7,$B6,$BE,$4F,$12,$31,$A7,$FB,$B3,$EC,$5E,$87,$F8
	.dc.b	$5F,$E7,$1F,$6F,$FE,$E6,$CB,$FB,$85,$69,$FE,$EF,$3A,$F9,$1F,$F0
	.dc.b	$E4,$E7,$3F,$E8,$73,$9A,$E3,$87,$5C,$3A,$57,$E6,$AF,$83,$FE,$33
	.dc.b	$6B,$FE,$68,$6B,$FD,$BD,$BA,$FE,$5F,$C0,$FC,$B7,$B1,$F9,$76,$E3
	.dc.b	$F9,$06,$DF,$C2,$6E,$9F,$76,$3F,$BD,$7C,$83,$E2,$3E,$D7,$37,$F6
	.dc.b	$8C,$3F,$C2,$13,$97,$EC,$C6,$6B,$F6,$73,$97,$FC,$98,$1C,$3F,$E4
	.dc.b	$F6,$BF,$E8,$46,$BF,$E0,$BF,$C2,$FF,$D2,$95,$FE,$24,$FC,$4D,$F8
	.dc.b	$BB,$E2,$7F,$C9,$55,$AB,$FE,$E3,$5A,$9C,$FC,$57,$F9,$46,$31,$87
	.dc.b	$F8,$A6,$F8,$69,$F0,$67,$38,$7F,$0C,$9E,$8F,$F9,$95,$AF,$FA,$69
	.dc.b	$AF,$9E,$FF,$CA,$9E,$53,$E8,$FF,$45,$F4,$DF,$E4,$8C,$BF,$C7,$13
	.dc.b	$97,$F4,$31,$9B,$FE,$37,$39,$7E,$4B,$5F,$F3,$B3,$5F,$F5,$5B,$57
	.dc.b	$33,$F6,$6F,$E6,$D1,$FC,$23,$E5,$7F,$75,$87,$E6,$27,$3A,$FF,$20
	.dc.b	$9C,$D7,$E6,$27,$8B,$FE,$B8,$BE,$9F,$07,$FE,$73,$6E,$1F,$79,$B3
	.dc.b	$7C,$BF,$F3,$F4,$7E,$E3,$1F,$C7,$4F,$9A,$73,$E0,$69,$FF,$3E,$57
	.dc.b	$F5,$09,$FF,$4A,$35,$F1,$47,$FA,$0F,$A3,$F9,$B7,$AD,$FC,$D4,$FE
	.dc.b	$6D,$F4,$D7,$2F,$E5,$F1,$6F,$F8,$0C,$DC,$BF,$41,$8B,$BB,$FD,$7F
	.dc.b	$87,$EF,$3F,$67,$FE,$0C,$FE,$FA,$33,$87,$EF,$27,$2F,$EF,$A3,$2F
	.dc.b	$C7,$6B,$FE,$65,$AE,$DF,$1D,$F2,$67,$33,$FE,$96,$F7,$FF,$D4,$5F
	.dc.b	$6A,$FA,$0F,$F5,$36,$69,$CC,$61,$FF,$51,$46,$73,$FD,$7D,$FF,$9F
	.dc.b	$67,$24,$8F,$F4,$44,$EE,$7F,$DF,$15,$CE,$73,$56,$73,$18,$3F,$8F
	.dc.b	$FE,$B4,$F3,$BF,$D7,$1F,$B4,$3E,$8F,$F9,$44,$E6,$BF,$B2,$4E,$5E
	.dc.b	$35,$F8,$99,$D5,$FD,$E6,$BF,$8F,$6B,$FE,$D2,$9F,$ED,$DF,$87,$70
	.dc.b	$CF,$FD,$B7,$CD,$FE,$C9,$8C,$4E,$EB,$FB,$3C,$61,$FF,$70,$47,$2C
	.dc.b	$3F,$DA,$51,$F9,$E3,$81,$FC,$62,$BF,$D9,$CE,$4F,$F7,$37,$CA,$FF
	.dc.b	$A7,$FE,$B1,$F5,$9F,$EA,$1E,$76,$5F,$BB,$CF,$FD,$B1,$7B,$AD,$4E
	.dc.b	$73,$FE,$4B,$39,$C5,$87,$2F,$83,$F2,$C2,$75,$39,$AF,$F6,$14,$FF
	.dc.b	$AC,$67,$93,$FE,$B3,$F4,$BF,$9E,$67,$E6,$8F,$48,$70,$EE,$B8,$BC
	.dc.b	$CD,$65,$FE,$BD,$74,$60,$E7,$3C,$95,$78,$27,$DB,$1B,$38,$CE,$DF
	.dc.b	$61,$93,$8D,$78,$9C,$C7,$5A,$E9,$1E,$0A,$F2,$3C,$0E,$47,$07,$31
	.dc.b	$C2,$3D,$57,$EE,$8C,$47,$BE,$BB,$5F,$86,$B4,$D9,$F3,$57,$07,$B1
	.dc.b	$F1,$5C,$4F,$FB,$8F,$E5,$FF,$D9,$B3,$F4,$1A,$F8,$83,$87,$E8,$3A
	.dc.b	$F9,$5F,$E0,$1A,$F3,$F8,$BF,$A5,$18,$AE,$31,$8E,$4F,$F7,$E4,$60
	.dc.b	$FF,$7A,$4E,$EF,$A4,$EB,$3F,$C0,$CB,$E9,$3E,$F4,$61,$BA,$F5,$CB
	.dc.b	$F1,$B5,$E1,$AD,$63,$8B,$DB,$E0,$C3,$B6,$1C,$63,$F0,$57,$DD,$E0
	.dc.b	$74,$9D,$BC,$70,$CB,$C2,$FA,$AA,$F1,$5F,$5C,$3C,$73,$B3,$A5,$C7
	.dc.b	$04,$8F,$83,$87,$82,$3A,$36,$FC,$D1,$BB,$EC,$47,$F6,$FC,$1F,$DF
	.dc.b	$DF,$E2,$BE,$27,$E9,$F1,$FE,$BF,$7E,$69,$F6,$1F,$A9,$E0,$FE,$0F
	.dc.b	$5B,$E0,$FE,$0A,$D9,$DE,$7F,$5B,$E8,$BF,$9E,$C7,$C1,$6E,$06,$92
	.dc.b	$3F,$D2,$9F,$98,$BE,$B7,$A6,$48,$C4,$8F,$F5,$5D,$F3,$71,$9F,$AB
	.dc.b	$0E,$35,$C2,$7F,$D2,$6D,$9D,$47,$F4,$6F,$0F,$FA,$9F,$7B,$A7,$FD
	.dc.b	$0B,$3F,$E0,$76,$BF,$D0,$67,$71,$8A,$FF,$59,$F5,$7D,$26,$F1,$FC
	.dc.b	$5C,$3A,$5F,$C7,$C3,$F8,$95,$7F,$83,$B8,$BF,$82,$B6,$3E,$0F,$CA
	.dc.b	$FF,$57,$7D,$BC,$5F,$C7,$6B,$6D,$8E,$91,$87,$F9,$AC,$F1,$7F,$DA
	.dc.b	$31,$F5,$1F,$E3,$87,$13,$FB,$9B,$93,$96,$EF,$97,$FF,$47,$CF,$FB
	.dc.b	$25,$AF,$43,$DD,$F8,$EB,$D4,$F9,$6D,$8F,$F1,$83,$5F,$F1,$3B,$5F
	.dc.b	$F1,$C3,$9C,$BF,$C3,$FE,$23,$BB,$E5,$36,$BF,$D3,$75,$BD,$3F,$DB
	.dc.b	$DD,$DF,$88,$F6,$4D,$31,$F1,$76,$FF,$DB,$B1,$5F,$F5,$7C,$7D,$5F
	.dc.b	$DA,$7F,$C5,$D1,$FE,$16,$6B,$FE,$86,$6B,$FC,$F1,$AF,$F5,$17,$79
	.dc.b	$FF,$92,$97,$C3,$F5,$3B,$78,$FF,$85,$FD,$AB,$E5,$DF,$F9,$65,$AF
	.dc.b	$F8,$CD,$AF,$F9,$A1,$BC,$A7,$E9,$E1,$FE,$A5,$E2,$68,$F8,$9F,$88
	.dc.b	$77,$F7,$7F,$25,$F9,$77,$F3,$BC,$BF,$B4,$35,$FD,$73,$53,$4F,$E8
	.dc.b	$B8,$7F,$99,$BE,$3B,$F5,$9C,$3F,$DC,$5B,$5C,$BC,$6F,$D0,$7E,$38
	.dc.b	$F9,$FF,$E0,$AD,$9F,$F2,$1B,$5F,$F5,$9B,$5F,$ED,$2E,$70,$FF,$3A
	.dc.b	$F8,$EF,$A3,$FD,$63,$6B,$F4,$E7,$F0,$7A,$E1,$97,$F8,$15,$B1,$FC
	.dc.b	$2E,$78,$2B,$EB,$7E,$8C,$3F,$71,$F8,$7B,$ED,$FF,$8E,$B6,$FF,$CB
	.dc.b	$6D,$7F,$D9,$6D,$7F,$C0,$2E,$6F,$FD,$1B,$F8,$7B,$E0,$FF,$82,$36
	.dc.b	$BD,$22,$B8,$69,$FE,$39,$E2,$FE,$20,$DF,$03,$F4,$07,$E8,$7C,$29
	.dc.b	$7F,$0B,$97,$F0,$73,$A8,$C5,$F2,$7F,$96,$7D,$1B,$E9,$7F,$9C,$B6
	.dc.b	$3F,$CE,$ED,$7F,$DC,$ED,$7F,$C3,$AE,$75,$FF,$54,$FC,$DF,$E0,$A4
	.dc.b	$D9,$D6,$7F,$E3,$76,$CF,$F3,$D9,$E0,$6E,$FF,$79,$E1,$7E,$F7,$DA
	.dc.b	$95,$EA,$8C,$BE,$C8,$CB,$F6,$F0,$F9,$13,$FC,$C7,$0F,$D9,$CF,$B3
	.dc.b	$BB,$DE,$47,$07,$D2,$FB,$1F,$D5,$37,$B3,$94,$FF,$3B,$FB,$C7,$C4
	.dc.b	$5F,$A8,$97,$F6,$A7,$2D,$9C,$BE,$53,$FA,$E4,$27,$A4,$C9,$E5,$1C
	.dc.b	$37,$FF,$6A,$D7,$49,$E5,$C5,$FC,$03,$D7,$FE,$45,$F0,$32,$7E,$FC
	.dc.b	$7F,$25,$30,$F0,$E3,$4E,$63,$F1,$1F,$8C,$3E,$13,$F9,$64,$43,$2F
	.dc.b	$F7,$68,$D4,$7E,$30,$D1,$E4,$72,$FB,$C7,$D8,$BE,$AF,$FC,$0E,$D8
	.dc.b	$FF,$62,$B5,$FF,$34,$9C,$23,$67,$F6,$BF,$B1,$7C,$65,$3F,$B7,$57
	.dc.b	$1D,$BF,$DF,$6D,$E6,$FF,$43,$57,$0C,$3F,$5E,$D7,$CD,$3B,$5F,$28
	.dc.b	$FD,$AF,$FB,$73,$63,$FC,$44,$D7,$F8,$8F,$CF,$7F,$B8,$FE,$BF,$FC
	.dc.b	$01,$F9,$77,$F8,$17,$2F,$F0,$D5,$BF,$2E,$7C,$B7,$26,$E3,$FC,$0A
	.dc.b	$F8,$2B,$FE,$67,$F3,$7F,$E3,$16,$C7,$F2,$06,$BF,$8D,$C3,$B6,$1E
	.dc.b	$6F,$33,$FB,$E4,$F5,$FB,$5F,$29,$9F,$6A,$C7,$F2,$AB,$F3,$7C,$77
	.dc.b	$94,$CC,$75,$D1,$19,$8E,$11,$87,$B1,$FC,$73,$B9,$F5,$9F,$E2,$36
	.dc.b	$C6,$63,$E5,$FE,$5D,$FC,$22,$30,$78,$FC,$EF,$EA,$3B,$7C,$A6,$C7
	.dc.b	$F9,$16,$F9,$BF,$8F,$6F,$37,$C9,$8F,$B7,$3C,$33,$FE,$A9,$B7,$8F
	.dc.b	$2F,$27,$CE,$CB,$F4,$3F,$E3,$9F,$39,$F1,$99,$9F,$9D,$F9,$73,$E9
	.dc.b	$3F,$06,$D8,$E2,$43,$71,$A3,$96,$9E,$58,$7E,$5D,$FC,$41,$FD,$43
	.dc.b	$63,$4D,$DF,$F1,$7E,$1B,$A7,$FC,$9D,$39,$BF,$CC,$FE,$8F,$FE,$BA
	.dc.b	$7E,$E3,$5E,$BE,$4F,$03,$F3,$71,$9C,$7E,$47,$0A,$31,$97,$FB,$BF
	.dc.b	$0E,$D8,$7A,$3E,$97,$C2,$47,$F1,$59,$BC,$CC,$47,$E5,$7E,$01,$F5
	.dc.b	$3F,$AA,$6C,$7F,$62,$C1,$87,$0D,$3C,$75,$AA,$E5,$79,$CB,$C9,$D5
	.dc.b	$F0,$89,$3F,$9B,$F9,$2F,$D0,$BF,$32,$3F,$73,$E2,$30,$A7,$09,$D1
	.dc.b	$FA,$38,$D3,$63,$58,$FD,$47,$FD,$43,$9E,$A7,$F8,$CB,$5F,$F4,$D1
	.dc.b	$FE,$9B,$70,$FC,$B8,$F1,$FA,$92,$3E,$14,$7F,$A1,$3A,$BF,$D1,$5F
	.dc.b	$A6,$FA,$B0,$FC,$43,$74,$FF,$53,$79,$AF,$88,$FF,$0F,$B6,$F8,$F8
	.dc.b	$33,$2F,$F0,$35,$7F,$89,$16,$1F,$79,$FF,$58,$35,$CF,$C9,$FF,$A5
	.dc.b	$BD,$DF,$1B,$23,$FC,$33,$3C,$71,$F1,$57,$FA,$87,$ED,$9F,$75,$49
	.dc.b	$7E,$2F,$8C,$AF,$F4,$09,$DB,$FD,$08,$38,$75,$3F,$ED,$08,$C6,$DF
	.dc.b	$BF,$E8,$7E,$69,$FE,$74,$E0,$FF,$65,$46,$3E,$23,$BD,$7E,$C6,$DF
	.dc.b	$73,$E7,$26,$5F,$1B,$CA,$AF,$96,$3C,$04,$7C,$D2,$F9,$BF,$D7,$47
	.dc.b	$E0,$23,$1C,$3A,$9C,$CF,$FB,$76,$72,$FF,$60,$F9,$4F,$F3,$47,$81
	.dc.b	$F0,$5B,$19,$E6,$FE,$AA,$D9,$F0,$F7,$B6,$A4,$9F,$99,$19,$7F,$B6
	.dc.b	$FB,$AF,$AA,$FE,$4D,$BF,$F1,$38,$C7,$0F,$E1,$BE,$E7,$2F,$E6,$A3
	.dc.b	$37,$FD,$F5,$B5,$C6,$F7,$3F,$ED,$4B,$AF,$FA,$F6,$2D,$7F,$B2,$6F
	.dc.b	$EF,$36,$7C,$BF,$25,$FC,$95,$F0,$35,$1F,$EC,$77,$99,$E0,$E8,$FD
	.dc.b	$17,$DC,$E2,$FF,$CD,$B7,$D7,$2F,$75,$EF,$AD,$7B,$C7,$D9,$F2,$3E
	.dc.b	$63,$5F,$73,$72,$E7,$16,$7F,$C6,$76,$AC,$63,$FE,$B5,$87,$C7,$FC
	.dc.b	$2F,$B9,$D6,$3F,$B2,$8C,$5F,$8D,$B1,$93,$FD,$0B,$3C,$30,$FF,$16
	.dc.b	$7D,$C7,$F6,$C9,$FD,$33,$63,$FC,$52,$D7,$F8,$88,$70,$FF,$14,$78
	.dc.b	$63,$FD,$4D,$B3,$FE,$18,$8F,$F0,$BB,$FE,$96,$8C,$CF,$D2,$5E,$8F
	.dc.b	$E0,$39,$1F,$63,$FE,$06,$BE,$13,$98,$C5,$FA,$A3,$EE,$3E,$56,$DF
	.dc.b	$4A,$F6,$C6,$0F,$C5,$C6,$67,$FC,$41,$83,$EE,$73,$5A,$FB,$9F,$27
	.dc.b	$86,$F5,$B1,$97,$F8,$7E,$30,$FB,$67,$C2,$47,$E6,$B7,$1F,$A2,$FB
	.dc.b	$F4,$FF,$7F,$CF,$13,$E3,$7C,$67,$C3,$FB,$9F,$F0,$34,$60,$FF,$17
	.dc.b	$38,$BF,$B7,$3F,$00,$93,$F7,$1B,$1A,$8F,$F1,$B2,$FB,$9F,$87,$3F
	.dc.b	$AE,$FB,$7F,$9B,$B7,$F7,$AF,$B9,$F6,$DF,$F9,$CE,$73,$87,$F9,$52
	.dc.b	$73,$5F,$E1,$39,$FA,$B5,$F8,$86,$BF,$D6,$9A,$FE,$C9,$A9,$EA,$E7
	.dc.b	$D0,$FE,$03,$F7,$3F,$88,$8D,$47,$9F,$CE,$FC,$47,$4F,$C3,$C3,$C5
	.dc.b	$18,$F6,$35,$9F,$F4,$3C,$BF,$BC,$36,$7F,$13,$18,$C7,$FD,$12,$DB
	.dc.b	$E7,$5F,$EE,$EC,$BF,$C2,$2F,$FB,$C2,$74,$FF,$9D,$BE,$62,$DF,$FA
	.dc.b	$09,$AF,$B7,$E2,$FF,$3D,$18,$AF,$F5,$5D,$79,$67,$C6,$47,$FA,$D7
	.dc.b	$B3,$B8,$FE,$46,$D8,$FF,$44,$FD,$CB,$DE,$FE,$53,$F0,$C8,$C4,$FF
	.dc.b	$9D,$71,$7C,$96,$A7,$39,$FD,$DF,$27,$E1,$92,$FE,$43,$FA,$EC,$B8
	.dc.b	$C7,$F2,$AD,$BF,$CA,$51,$8A,$FF,$6B,$4B,$E0,$9E,$37,$A3,$2F,$E7
	.dc.b	$7F,$6E,$FB,$9F,$E7,$F9,$7F,$00,$E4,$CC,$63,$C2,$FB,$99,$7F,$88
	.dc.b	$AF,$83,$FD,$F5,$B9,$FF,$B0,$1A,$FF,$18,$6B,$C6,$BA,$CF,$A4,$FE
	.dc.b	$87,$F3,$CF,$C0,$FF,$D8,$B1,$9A,$FE,$8F,$7D,$B2,$FF,$42,$79,$6F
	.dc.b	$92,$C5,$FF,$8A,$9F,$F2,$8B,$C6,$35,$97,$F4,$D9,$CE,$1F,$EE,$B8
	.dc.b	$C7,$C1,$65,$B7,$C6,$BF,$B2,$BD,$7E,$1B,$97,$9F,$CD,$B1,$B6,$E1
	.dc.b	$FE,$F7,$7F,$B5,$FD,$D7,$FB,$52,$6F,$88,$F8,$CB,$18,$6B,$F0,$67
	.dc.b	$FD,$B7,$18,$B6,$BF,$EA,$59,$CD,$6E,$FF,$DA,$0D,$BE,$15,$B9,$F0
	.dc.b	$4F,$CD,$B6,$CF,$99,$FF,$00,$F4,$3E,$DF,$F9,$B6,$3F,$E9,$FA,$D6
	.dc.b	$1F,$56,$B5,$E5,$FE,$33,$09,$39,$C3,$F1,$38,$FD,$A9,$FA,$5F,$1D
	.dc.b	$F4,$FE,$EA,$DF,$CF,$6D,$7B,$A3,$EE,$46,$33,$FC,$0F,$0A,$3F,$C2
	.dc.b	$BF,$73,$FD,$DA,$BF,$11,$CE,$FB,$A7,$C0,$EB,$D4,$FF,$0E,$CE,$63
	.dc.b	$15,$FE,$57,$BE,$59,$FF,$70,$0E,$1F,$EE,$6D,$7F,$D0,$AF,$0C,$FF
	.dc.b	$D8,$8D,$9F,$F0,$FF,$E1,$CF,$C3,$9F,$47,$E0,$C3,$DC,$3B,$3F,$39
	.dc.b	$E3,$FE,$5F,$DB,$FC,$C3,$71,$FC,$A3,$6B,$8D,$70,$9F,$CC,$57,$F8
	.dc.b	$E3,$E6,$1F,$3D,$F7,$39,$9E,$AE,$63,$0D,$BF,$C2,$57,$07,$ED,$7A
	.dc.b	$79,$CB,$CE,$3E,$E5,$F3,$BF,$E4,$8F,$BA,$7D,$E3,$EF,$3F,$C9,$35
	.dc.b	$AC,$7F,$79,$5A,$9C,$DF,$F8,$2A,$B5,$87,$D1,$6B,$FE,$22,$6B,$FE
	.dc.b	$B4,$77,$CB,$FE,$F3,$6E,$1F,$E8,$8D,$BF,$3F,$5E,$89,$CB,$FE,$52
	.dc.b	$F4,$9F,$69,$FA,$0F,$BC,$7B,$1F,$E6,$09,$CE,$3F,$9F,$9C,$BF,$A9
	.dc.b	$8C,$DF,$C4,$6E,$5F,$E9,$BA,$B5,$FE,$CA,$9B,$4E,$B7,$DD,$FF,$36
	.dc.b	$AF,$E3,$1F,$79,$FB,$AA,$FC,$2D,$6B,$D2,$FF,$33,$4E,$72,$5F,$F2
	.dc.b	$AC,$39,$61,$C1,$B3,$FA,$8F,$81,$FD,$0C,$62,$BF,$DB,$B8,$FF,$49
	.dc.b	$BF,$F2,$7B,$5C,$FA,$7F,$E3,$E8,$C4,$7F,$38,$8F,$F4,$2C,$7F,$51
	.dc.b	$C3,$F6,$BE,$E6,$27,$F9,$83,$2F,$F9,$67,$DE,$5C,$3F,$33,$E3,$7F
	.dc.b	$84,$3B,$7F,$9E,$5F,$F3,$7A,$D6,$3E,$EB,$F9,$1E,$1F,$FA,$DF,$1F
	.dc.b	$DB,$67,$FB,$EF,$8D,$FF,$47,$36,$33,$3B,$AF,$6B,$6F,$A5,$6E,$7F
	.dc.b	$57,$FB,$67,$F4,$C7,$BE,$1F,$EA,$2A,$7C,$1F,$F9,$81,$B8,$FF,$B7
	.dc.b	$DB,$5F,$C2,$5A,$FF,$9E,$9C,$9E,$B7,$F9,$DC,$FF,$A6,$FE,$F3,$B3
	.dc.b	$2B,$F9,$B5,$70,$7F,$D5,$3A,$BE,$25,$7B,$C3,$F6,$B7,$B7,$FC,$0D
	.dc.b	$87,$5C,$39,$BF,$C8,$BE,$D9,$EA,$FC,$86,$CF,$CC,$6D,$7E,$11,$AF
	.dc.b	$E2,$1C,$DF,$CA,$6C,$F9,$BD,$6F,$F2,$F3,$FC,$D2,$33,$5F,$ED,$38
	.dc.b	$FE,$ED,$F7,$78,$8F,$F6,$2F,$47,$FB,$0E,$31,$1F,$9C,$3F,$DC,$33
	.dc.b	$FC,$D5,$DB,$5D,$1F,$EA,$89,$FF,$6C,$BF,$9F,$3F,$97,$7E,$40,$F9
	.dc.b	$BC,$CE,$A5,$F3,$DF,$ED,$6F,$BC,$D4,$7F,$35,$C7,$F8,$0B,$FC,$88
	.dc.b	$FF,$6D,$CE,$EF,$EA,$CE,$B4,$FF,$70,$4F,$FB,$83,$E3,$39,$DF,$FB
	.dc.b	$8B,$D8,$FC,$E4,$62,$39,$9F,$E9,$FC,$3F,$D8,$33,$A7,$FA,$A5,$77
	.dc.b	$9F,$F7,$58,$9F,$D3,$63,$CE,$3D,$61,$ED,$1A,$AE,$73,$97,$BE,$9D
	.dc.b	$63,$AD,$D7,$B4,$57,$AD,$F4,$8C,$4E,$EF,$B5,$7F,$B6,$A3,$E1,$5F
	.dc.b	$68,$F7,$19,$29,$F1,$6B,$6E,$6F,$B3,$E7,$31,$3C,$E3,$E0,$57,$09
	.dc.b	$E4,$74,$3E,$6E,$1C,$8D,$B9,$C3,$E7,$3E,$38,$D5,$FC,$6F,$96,$FF
	.dc.b	$69,$4F,$9D,$AF,$C6,$0E,$1F,$B6,$EB,$E5,$7F,$84,$E6,$9D,$DF,$DA
	.dc.b	$9C,$C6,$0D,$38,$AE,$5C,$5F,$E4,$29,$D0,$9E,$35,$F8,$E5,$F3,$7F
	.dc.b	$86,$8C,$37,$1F,$05,$F3,$9E,$0D,$8F,$24,$EE,$FC,$6D,$7C,$13,$B7
	.dc.b	$CE,$F9,$74,$EF,$1F,$8B,$C3,$C6,$D9,$D5,$F8,$B2,$FB,$15,$AA,$FC
	.dc.b	$1B,$71,$F2,$CE,$AF,$8D,$79,$0E,$D8,$FE,$93,$1D,$C7,$E1,$CF,$3D
	.dc.b	$69,$86,$70,$FA,$33,$C0,$F4,$C6,$6B,$A6,$3E,$78,$E1,$87,$C1,$7E
	.dc.b	$94,$FC,$5B,$F4,$56,$F1,$E8,$FF,$06,$F1,$7F,$73,$97,$FA,$73,$7F
	.dc.b	$F0,$9E,$3F,$E7,$5C,$3C,$77,$F8,$5F,$BB,$DF,$05,$AD,$46,$1F,$E3
	.dc.b	$BC,$35,$FC,$24,$7F,$25,$B3,$49,$1F,$CA,$E3,$2F,$FA,$A7,$EE,$DF
	.dc.b	$79,$D7,$A9,$18,$9F,$09,$FC,$F7,$DD,$F1,$77,$3A,$3F,$D1,$93,$93
	.dc.b	$FD,$69,$D2,$FC,$03,$73,$A9,$FE,$9D,$7F,$CF,$6F,$93,$67,$53,$FC
	.dc.b	$80,$FE,$51,$C3,$FD,$AF,$83,$FB,$77,$07,$F2,$46,$CF,$F9,$B7,$B5
	.dc.b	$F7,$DF,$16,$73,$3C,$AB,$83,$EF,$AD,$BF,$EA,$79,$D9,$1D,$57,$37
	.dc.b	$FA,$94,$7F,$53,$7F,$99,$1D,$0F,$F0,$BF,$DE,$3D,$7F,$8E,$D8,$FF
	.dc.b	$4C,$B5,$FF,$09,$F8,$BA,$9F,$89,$39,$9F,$F1,$CB,$5F,$F7,$5B,$5F
	.dc.b	$F4,$67,$8D,$FE,$5C,$7F,$DB,$9E,$6F,$DB,$8C,$47,$F8,$CB,$E4,$B1
	.dc.b	$32,$36,$7F,$B8,$E3,$F1,$9F,$79,$FE,$AC,$8F,$F6,$53,$5F,$F4,$2B
	.dc.b	$76,$FF,$5B,$D7,$0F,$23,$FB,$B4,$EE,$BE,$D3,$5F,$EB,$6D,$7F,$CC
	.dc.b	$DF,$78,$F8,$0F,$C6,$B6,$3F,$D7,$ED,$7F,$C9,$7D,$DA,$9F,$77,$C7
	.dc.b	$C9,$DC,$7E,$97,$EF,$3F,$9C,$E7,$FE,$33,$6F,$27,$F6,$C6,$D7,$F3
	.dc.b	$7F,$92,$7D,$37,$1F,$0F,$F5,$D1,$FA,$39,$FF,$52,$B7,$0E,$93,$FB
	.dc.b	$3F,$BC,$7D,$0D,$25,$FF,$BC,$FC,$7F,$F5,$BB,$FE,$F5,$9D,$BF,$E1
	.dc.b	$F6,$BF,$ED,$56,$BF,$A3,$6B,$FE,$04,$77,$7F,$E4,$F9,$FF,$68,$CF
	.dc.b	$7F,$83,$F0,$BF,$EB,$CF,$BA,$7D,$E7,$F0,$96,$C7,$F5,$7B,$FF,$2C
	.dc.b	$46,$23,$FD,$87,$F7,$8F,$57,$85,$FF,$11,$87,$C3,$53,$35,$CF,$87
	.dc.b	$FC,$B7,$D5,$FC,$07,$2F,$F7,$EF,$77,$E5,$4F,$CB,$78,$52,$F8,$1F
	.dc.b	$9C,$8D,$38,$F1,$D7,$CF,$0F,$F1,$2F,$D2,$3E,$F3,$F9,$23,$6F,$98
	.dc.b	$70,$FF,$B7,$BE,$17,$EF,$BA,$8C,$BF,$C4,$5A,$F6,$9F,$F2,$94,$62
	.dc.b	$3F,$BF,$3D,$33,$FF,$40,$57,$E1,$6F,$FC,$72,$DA,$DD,$FE,$D7,$CE
	.dc.b	$FD,$B7,$B5,$27,$DD,$92,$75,$18,$9D,$3F,$EA,$C8,$F7,$FC,$CC,$8F
	.dc.b	$53,$FD,$17,$EF,$1F,$2D,$99,$FF,$7E,$C3,$BF,$46,$75,$FF,$4C,$7D
	.dc.b	$DF,$E5,$A9,$18,$FB,$CF,$EA,$12,$6F,$C6,$CA,$F9,$F3,$96,$CE,$5E
	.dc.b	$7C,$1F,$E5,$C9,$ED,$7D,$67,$67,$2B,$FE,$35,$ED,$7F,$1B,$F8,$49
	.dc.b	$3F,$1A,$32,$FC,$88,$CB,$FE,$D0,$8F,$EB,$9F,$78,$F8,$9F,$E4,$93
	.dc.b	$0F,$C6,$46,$5B,$19,$7F,$DB,$27,$27,$FB,$87,$DD,$BE,$F2,$92,$7F
	.dc.b	$1F,$F6,$7F,$E4,$67,$F2,$33,$B7,$FD,$38,$D7,$FA,$F0,$70,$FF,$61
	.dc.b	$6B,$FE,$82,$78,$67,$FE,$CD,$97,$C8,$7F,$80,$DB,$5E,$0B,$FF,$04
	.dc.b	$7C,$D3,$EF,$3F,$B3,$B6,$73,$1F,$DA,$BE,$8F,$FD,$C3,$F6,$FF,$E0
	.dc.b	$DF,$BC,$FF,$1B,$57,$F7,$36,$BF,$E9,$46,$BF,$E5,$F6,$F1,$BC,$E7
	.dc.b	$CF,$FF,$86,$63,$0D,$8C,$D7,$F0,$0C,$9F,$78,$F9,$EF,$4F,$9F,$FB
	.dc.b	$6A,$FC,$49,$DF,$EC,$FC,$C6,$BD,$2B,$1F,$EC,$9D,$13,$FE,$C9,$6C
	.dc.b	$66,$3F,$D9,$D8,$F8,$3D,$CF,$8F,$91,$A7,$4A,$F0,$8F,$01,$F7,$7F
	.dc.b	$E3,$36,$C7,$F1,$06,$BE,$8B,$7D,$E3,$C7,$E1,$7F,$ED,$DB,$DD,$FF
	.dc.b	$88,$5A,$FF,$B0,$9A,$FF,$A3,$5E,$BD,$0E,$2F,$F4,$86,$3F,$D4,$B7
	.dc.b	$FC,$0F,$A9,$FD,$7B,$0F,$46,$3E,$7F,$A2,$91,$E7,$7F,$C9,$9E,$73
	.dc.b	$E4,$33,$3F,$96,$D1,$5E,$72,$47,$FB,$DE,$32,$FD,$BF,$BC,$7D,$4F
	.dc.b	$E4,$9B,$1D,$CF,$E6,$BF,$75,$FC,$4D,$BE,$47,$F9,$76,$31,$1F,$27
	.dc.b	$C4,$76,$66,$BF,$53,$3F,$EA,$FA,$DB,$FE,$49,$6B,$FE,$EA,$6B,$B8
	.dc.b	$FD,$2F,$DE,$7E,$DE,$3E,$B3,$78,$FE,$64,$6C,$F3,$FD,$4F,$94,$8F
	.dc.b	$E6,$71,$FE,$9C,$E8,$F2,$E0,$9F,$3C,$23,$0E,$A3,$31,$88,$FC,$47
	.dc.b	$C1,$BE,$F3,$F7,$6D,$8E,$C7,$E5,$27,$F3,$0F,$CC,$E9,$F9,$47,$F0
	.dc.b	$2F,$95,$FD,$56,$5E,$C9,$F1,$4F,$80,$8F,$A1,$AF,$E6,$7C,$56,$1F
	.dc.b	$C3,$1B,$1D,$CF,$B7,$1F,$64,$8F,$EF,$6D,$F6,$B8,$8F,$E4,$38,$F2
	.dc.b	$6D,$7C,$23,$EE,$FA,$CF,$8A,$FF,$0F,$F6,$7F,$BB,$69,$F7,$9E,$C8
	.dc.b	$C3,$63,$FD,$09,$D7,$FE,$7F,$3F,$D5,$11,$98,$E3,$E1,$E5,$FF,$03
	.dc.b	$C7,$E0,$6B,$9F,$CB,$F0,$3F,$D1,$1F,$03,$6B,$71,$5F,$E2,$B9,$79
	.dc.b	$E7,$FD,$6D,$F7,$2F,$BC,$A4,$AF,$F0,$E6,$9F,$EE,$3E,$3F,$A9,$5B
	.dc.b	$9F,$F0,$CB,$5F,$E2,$2D,$8F,$CA,$E1,$F1,$7E,$63,$A5,$7F,$4C,$AE
	.dc.b	$0F,$FA,$AD,$AF,$C9,$6A,$0B,$EE,$FD,$24,$8F,$3F,$91,$5F,$1C,$5C
	.dc.b	$3C,$78,$FF,$2E,$9F,$3F,$CE,$FE,$B5,$F0,$35,$F0,$60,$8C,$47,$89
	.dc.b	$BF,$31,$FE,$6C,$F1,$FF,$DB,$5C,$66,$92,$BE,$07,$C0,$3B,$7F,$82
	.dc.b	$E3,$FE,$41,$6C,$66,$77,$1F,$C2,$BD,$D7,$CB,$7F,$00,$D7,$F4,$D1
	.dc.b	$8E,$5F,$AF,$F7,$79,$7F,$25,$19,$C3,$FA,$2B,$6B,$8D,$EE,$7F,$DA
	.dc.b	$D7,$5F,$F1,$F4,$5A,$FF,$64,$5F,$F9,$61,$B3,$F8,$CD,$DF,$77,$BF
	.dc.b	$94,$97,$B7,$B7,$2F,$E9,$E7,$37,$E4,$9B,$9F,$D6,$9B,$7E,$AF,$93
	.dc.b	$F1,$1C,$B1,$FD,$BC,$78,$E3,$2D,$3F,$33,$DD,$8C,$7C,$DF,$1F,$ED
	.dc.b	$FB,$BE,$31,$8A,$FF,$39,$C6,$32,$FC,$3D,$6E,$FC,$0D,$AF,$F3,$CF
	.dc.b	$DD,$FF,$6E,$6F,$A1,$F2,$F9,$7F,$CA,$BE,$B7,$CB,$AD,$67,$FD,$D9
	.dc.b	$B9,$7F,$1F,$6E,$1F,$D9,$A3,$FC,$1C,$F4,$3B,$F9,$FF,$C9,$6B,$F8
	.dc.b	$75,$7F,$60,$9F,$E9,$ED,$8F,$F1,$CB,$5F,$E2,$81,$D5,$78,$6B,$D0
	.dc.b	$BE,$93,$D1,$12,$FC,$27,$F8,$8E,$74,$E2,$FD,$9F,$59,$F0,$7C,$17
	.dc.b	$A9,$F0,$63,$0D,$AF,$93,$1F,$1F,$4F,$F0,$C4,$60,$97,$D1,$FF,$19
	.dc.b	$46,$A3,$07,$CB,$F5,$7E,$87,$DD,$FF,$7C,$AF,$AB,$18,$F4,$FE,$93
	.dc.b	$23,$FC,$79,$4F,$BB,$C3,$CE,$38,$9A,$8C,$E9,$FD,$57,$EE,$7F,$C6
	.dc.b	$53,$FE,$45,$FB,$BF,$7B,$FE,$BC,$8C,$DF,$F8,$B6,$72,$FF,$8D,$A3
	.dc.b	$35,$CF,$D0,$FD,$87,$DD,$FE,$0A,$BF,$90,$56,$B0,$FE,$98,$D8,$CD
	.dc.b	$70,$E5,$FD,$C4,$E6,$4F,$F6,$4A,$FF,$2D,$74,$FC,$59,$F1,$79,$DF
	.dc.b	$97,$8C,$65,$ED,$8E,$67,$93,$D2,$B7,$ED,$8B,$7F,$E5,$0C,$FF,$C0
	.dc.b	$53,$93,$FC,$D1,$F3,$1E,$19,$FF,$27,$37,$0F,$95,$F1,$1F,$A7,$8C
	.dc.b	$57,$BA,$3D,$67,$F1,$3F,$5F,$37,$F1,$23,$18,$EA,$BF,$DD,$75,$AA
	.dc.b	$F8,$35,$FC,$77,$E0,$9F,$76,$F7,$B0,$DC,$3C,$FF,$1F,$89,$F9,$7E
	.dc.b	$54,$9E,$B3,$71,$FF,$37,$E4,$8F,$31,$FC,$AE,$7F,$CE,$5E,$46,$1F
	.dc.b	$F7,$FF,$DC,$7F,$B8,$1F,$F2,$2B,$6F,$FD,$5E,$D7,$F6,$ED,$7F,$C1
	.dc.b	$8E,$F7,$FF,$13,$36,$7F,$B3,$7D,$13,$EE,$BF,$D6,$31,$E0,$F9,$2F
	.dc.b	$E7,$95,$BC,$3F,$B6,$36,$F8,$7C,$5E,$AB,$F9,$BF,$C9,$EF,$CE,$3F
	.dc.b	$E4,$7C,$7F,$BA,$67,$FD,$FA,$34,$67,$2F,$F0,$7B,$8F,$80,$E7,$87
	.dc.b	$47,$AF,$A1,$DB,$E1,$6E,$3D,$0F,$F0,$06,$5E,$08,$CE,$3D,$1B,$97
	.dc.b	$85,$FE,$B5,$F7,$5F,$DC,$6E,$F8,$BF,$EB,$E7,$A5,$A7,$51,$87,$FB
	.dc.b	$C7,$C5,$F0,$8E,$2D,$75,$5C,$EB,$FD,$7F,$5C,$9B,$19,$C3,$FC,$EB
	.dc.b	$39,$6C,$66,$BE,$DF,$CB,$FD,$43,$FB,$97,$33,$EC,$BF,$67,$3A,$3F
	.dc.b	$13,$3B,$7D,$73,$BF,$1F,$19,$7D,$0F,$3F,$EC,$FD,$D7,$DA,$6D,$FC
	.dc.b	$2D,$FE,$DC,$62,$B8,$4E,$78,$D7,$FE,$03,$FB,$97,$A5,$FD,$46,$FF
	.dc.b	$C6,$91,$8B,$FF,$4F,$B6,$3F,$A7,$87,$0F,$F5,$27,$5B,$FF,$9F,$DB
	.dc.b	$3F,$E3,$FF,$B9,$7D,$CB,$E8,$7C,$58,$CA,$FB,$78,$F5,$8F,$F0,$D5
	.dc.b	$F0,$AD,$4F,$E4,$BC,$0F,$CD,$FB,$5F,$9C,$6E,$1E,$F6,$CE,$E3,$E1
	.dc.b	$4F,$F8,$8F,$AD,$F3,$14,$FA,$FC,$BF,$EE,$7A,$F1,$CE,$9B,$3B,$7E
	.dc.b	$BC,$EE,$FF,$DE,$F5,$B7,$FD,$F3,$F6,$CF,$98,$F9,$D0,$8C,$5F,$FB
	.dc.b	$31,$B1,$F3,$9A,$FF,$7E,$73,$C1,$FE,$AB,$6C,$FF,$9A,$F9,$BF,$1D
	.dc.b	$D5,$FE,$3F,$9C,$DF,$F9,$BD,$AF,$87,$0F,$7C,$ED,$FF,$23,$78,$EF
	.dc.b	$A4,$A7,$E6,$BE,$DB,$19,$7D,$69,$B8,$33,$FF,$5F,$46,$1B,$7C,$4E
	.dc.b	$75,$FB,$8F,$09,$F6,$AF,$1B,$FC,$B1,$18,$7F,$8B,$FB,$78,$9D,$57
	.dc.b	$16,$E5,$E1,$8F,$E4,$F1,$8C,$7F,$8D,$57,$1B,$FF,$31,$61,$FC,$F6
	.dc.b	$FF,$1E,$DC,$3F,$8E,$FA,$7A,$F2,$65,$69,$FE,$53,$18,$9D,$CF,$EB
	.dc.b	$63,$F9,$6F,$D2,$6A,$30,$49,$FE,$53,$19,$7D,$02,$DF,$53,$FE,$95
	.dc.b	$AF,$8C,$D9,$FF,$5F,$FD,$AF,$F3,$D8,$C6,$8A,$FE,$8F,$18,$6F,$C4
	.dc.b	$FF,$9D,$AB,$83,$FE,$83,$A7,$BD,$4F,$F3,$97,$DA,$DB,$FD,$9C,$E9
	.dc.b	$B3,$A6,$93,$1F,$59,$F8,$38,$D3,$89,$E4,$D8,$EC,$FF,$A3,$EA,$CF
	.dc.b	$AE,$3E,$27,$D7,$3C,$BF,$C2,$B7,$0F,$E4,$1E,$67,$F8,$FD,$FC,$64
	.dc.b	$66,$BE,$6E,$A8,$C6,$1D,$63,$18,$7F,$A6,$B0,$E5,$7C,$2F,$78,$7F
	.dc.b	$9B,$B8,$7F,$73,$3F,$90,$3F,$91,$B8,$24,$7F,$1A,$8F,$C9,$3F,$D8
	.dc.b	$3E,$D5,$F4,$FF,$EB,$39,$C9,$FE,$BC,$9E,$15,$FE,$AC,$AD,$C6,$0E
	.dc.b	$92,$F9,$6F,$F7,$57,$C1,$F4,$8F,$F7,$61,$FE,$D3,$9C,$E1,$FC,$E2
	.dc.b	$AC,$EB,$BB,$E7,$5F,$18,$E3,$3F,$EF,$5F,$B5,$C4,$73,$0E,$1F,$11
	.dc.b	$8B,$FF,$49,$DF,$D9,$9D,$57,$39,$CE,$16,$FF,$D3,$33,$5F,$E3,$51
	.dc.b	$51,$DA,$BC,$77,$D6,$FD,$93,$E1,$3D,$95,$C5,$E1,$1E,$07,$47,$B8
	.dc.b	$A7,$B6,$BC,$EE,$2B,$B3,$E6,$31,$5D,$6F,$CD,$3A,$9E,$95,$E6,$AD
	.dc.b	$47,$03,$B9,$E4,$3C,$E6,$2B,$9C,$70,$9E,$53,$91,$F2,$9F,$EB,$27
	.dc.b	$C8,$D7,$D8,$D7,$E4,$D7,$CA,$8D,$7C,$87,$F0,$0E,$1C,$67,$F8,$88
	.dc.b	$C5,$F6,$8C,$71,$FF,$3D,$8C,$1F,$52,$77,$7D,$27,$58,$FE,$2E,$5F
	.dc.b	$27,$F9,$48,$C3,$75,$EB,$97,$E0,$6B,$C3,$5A,$C7,$17,$B7,$C1,$87
	.dc.b	$6C,$38,$C7,$DC,$BE,$EF,$03,$A4,$ED,$E3,$86,$5E,$17,$D5,$57,$8A
	.dc.b	$FA,$E1,$E3,$9D,$9D,$2E,$38,$24,$7C,$1C,$3C,$11,$D1,$B7,$E6,$8D
	.dc.b	$DF,$62,$3F,$37,$0F,$24,$EE,$FA,$61,$E0,$CB,$C8,$4C,$3E,$1B,$F4
	.dc.b	$7E,$D3,$F8,$1C,$FF,$2B,$C7,$DB,$18,$D7,$F0,$0D,$87,$A2,$E0,$FE
	.dc.b	$E9,$1F,$E3,$4C,$D5,$B8,$E6,$65,$E5,$5C,$78,$75,$5C,$2B,$FC,$92
	.dc.b	$70,$3F,$C9,$6E,$5B,$5A,$E5,$FE,$71,$E9,$F9,$5B,$3F,$CF,$A3,$BF
	.dc.b	$2B,$EC,$B5,$CC,$9A,$4A,$FE,$8F,$3E,$9E,$0F,$E4,$F1,$FD,$61,$BB
	.dc.b	$FC,$76,$EF,$FE,$98,$F8,$E7,$D9,$71,$B8,$7F,$96,$2B,$6F,$CA,$9D
	.dc.b	$B9,$36,$77,$3E,$3B,$A7,$06,$C8,$7F,$81,$B8,$11,$BC,$3F,$AD,$7D
	.dc.b	$93,$CD,$F8,$CD,$BF,$F1,$73,$5F,$F0,$4F,$83,$A9,$91,$A7,$EC,$F7
	.dc.b	$A7,$ED,$FA,$AF,$85,$A4,$BF,$F4,$67,$C1,$57,$FD,$8B,$17,$2F,$F6
	.dc.b	$54,$FE,$1F,$EC,$BF,$C3,$B5,$FE,$7E,$6B,$FE,$08,$6C,$FF,$B4,$6B
	.dc.b	$87,$89,$2B,$83,$D7,$87,$FD,$CB,$F6,$5F,$E9,$A9,$FF,$26,$35,$FF
	.dc.b	$00,$B6,$FF,$DB,$9E,$57,$E2,$E3,$15,$CB,$D6,$C1,$F0,$27,$F3,$DF
	.dc.b	$4D,$FC,$B6,$FF,$54,$DA,$FE,$DE,$DC,$3F,$AB,$FC,$33,$E7,$78,$F8
	.dc.b	$32,$7F,$D3,$B5,$B5,$5B,$7F,$D4,$3F,$64,$F9,$AF,$D7,$36,$33,$1F
	.dc.b	$B1,$F1,$BF,$D3,$31,$82,$1F,$E3,$29,$E0,$FF,$8D,$23,$6E,$67,$F6
	.dc.b	$6A,$76,$FE,$91,$AF,$F8,$31,$AF,$F7,$4F,$B2,$7D,$97,$F0,$36,$C7
	.dc.b	$FA,$E9,$AF,$F9,$13,$EC,$9E,$86,$EA,$4E,$DF,$E1,$93,$A8,$FF,$6C
	.dc.b	$78,$3F,$92,$EF,$91,$82,$1F,$A5,$E9,$FF,$0D,$71,$7F,$92,$7B,$D1
	.dc.b	$8E,$3F,$F0,$5C,$63,$0F,$E3,$5F,$38,$FB,$2A,$4C,$7F,$DA,$7F,$2B
	.dc.b	$F5,$5F,$F2,$EC,$ED,$FF,$6B,$B5,$CC,$FF,$25,$9E,$19,$7F,$31,$AD
	.dc.b	$57,$F1,$06,$BF,$E3,$86,$BF,$E2,$06,$E1,$EA,$F2,$BF,$67,$EA,$4A
	.dc.b	$FF,$7E,$47,$ED,$8C,$3F,$0E,$30,$F9,$BE,$A3,$DE,$47,$FD,$DF,$F6
	.dc.b	$4F,$8C,$C4,$7B,$A6,$E5,$FD,$26,$B5,$3F,$EF,$6F,$B1,$FC,$5C,$FF
	.dc.b	$4D,$5F,$65,$FD,$15,$B3,$BF,$83,$C1,$99,$91,$92,$1F,$E8,$A9,$E2
	.dc.b	$D8,$E2,$FF,$91,$2B,$EF,$71,$3D,$7F,$E2,$FE,$EC,$AD,$72,$7E,$02
	.dc.b	$B5,$18,$3F,$AF,$B8,$7F,$15,$3C,$5F,$C6,$57,$F5,$CF,$B2,$FF,$7D
	.dc.b	$47,$1E,$1F,$C6,$BF,$C9,$8D,$7C,$04,$75,$87,$F4,$49,$DC,$71,$AE
	.dc.b	$6F,$FA,$A3,$EC,$9F,$65,$FD,$B5,$B1,$E7,$0E,$1F,$F5,$7F,$D5,$FF
	.dc.b	$91,$1D,$46,$75,$FF,$00,$37,$5F,$F4,$2B,$67,$FC,$DB,$5C,$5F,$15
	.dc.b	$FF,$04,$9E,$D1,$FD,$99,$AF,$28,$FE,$F7,$F2,$0F,$B2,$FE,$C2,$D9
	.dc.b	$FF,$06,$35,$FA,$1F,$31,$FE,$D0,$FA,$EF,$F0,$57,$D9,$7F,$88,$F4
	.dc.b	$FF,$0B,$37,$4F,$F4,$DB,$6D,$87,$37,$FD,$D3,$2F,$21,$F1,$3F,$E1
	.dc.b	$B8,$C3,$63,$33,$F8,$1A,$7D,$93,$E5,$3C,$FE,$67,$EE,$31,$F0,$FD
	.dc.b	$67,$98,$CF,$9D,$71,$FA,$9A,$27,$11,$FE,$FC,$C9,$87,$38,$F4,$E5
	.dc.b	$FD,$B5,$F1,$35,$E0,$DD,$7E,$1D,$FF,$8B,$FB,$9F,$41,$49,$7E,$BF
	.dc.b	$B2,$78,$B8,$8F,$F7,$9D,$70,$7F,$14,$D7,$CD,$3C,$1C,$D7,$D8,$F3
	.dc.b	$BF,$AE,$69,$FD,$4F,$4E,$99,$FF,$92,$F1,$7D,$37,$94,$98,$FE,$5F
	.dc.b	$E4,$FF,$DF,$0F,$82,$32,$F9,$1F,$F2,$54,$6A,$30,$7D,$5F,$B2,$7C
	.dc.b	$D7,$E2,$9B,$1C,$CF,$D2,$7D,$8B,$E8,$3F,$CA,$8D,$8C,$C7,$E4,$BC
	.dc.b	$37,$56,$A2,$4E,$8F,$C2,$4E,$9C,$1F,$6F,$EC,$BF,$4D,$AB,$A3,$FD
	.dc.b	$AB,$C0,$9C,$C7,$FB,$CB,$2F,$15,$AF,$FC,$E5,$0F,$9F,$F3,$92,$7C
	.dc.b	$3E,$75,$78,$5F,$39,$C5,$F8,$67,$F4,$79,$2F,$37,$E4,$F7,$DF,$49
	.dc.b	$FB,$26,$DF,$C8,$C7,$E3,$FC,$0C,$8F,$96,$FE,$E3,$E4,$7F,$31,$0D
	.dc.b	$3F,$29,$8F,$B6,$B8,$4F,$E7,$EB,$75,$F9,$B8,$FF,$36,$B5,$E0,$7F
	.dc.b	$A0,$9C,$3F,$EF,$E5,$87,$DA,$24,$7E,$DE,$5F,$57,$88,$F8,$A7,$19
	.dc.b	$C6,$BF,$17,$75,$F0,$AF,$E7,$1F,$0F,$FE,$24,$AF,$F5,$27,$D9,$7B
	.dc.b	$8F,$DB,$79,$7F,$E6,$EE,$BF,$F3,$A9,$FE,$9F,$8D,$78,$58,$C3,$F9
	.dc.b	$A4,$7E,$01,$07,$0D,$F9,$3F,$F3,$FF,$BF,$CD,$91,$C9,$CB,$FC,$FF
	.dc.b	$EC,$1F,$65,$FE,$AF,$6C,$66,$3F,$D1,$3F,$2B,$96,$46,$4F,$7C,$61
	.dc.b	$F8,$E2,$79,$B8,$6E,$9F,$E2,$9D,$7F,$98,$4E,$FE,$CB,$C4,$47,$E5
	.dc.b	$F8,$D5,$E5,$16,$B9,$4F,$A7,$E6,$7F,$A7,$57,$C6,$3C,$4E,$1F,$93
	.dc.b	$82,$31,$1A,$CF,$FD,$93,$F2,$BF,$E6,$5E,$3B,$C3,$FD,$12,$BF,$D9
	.dc.b	$6D,$7F,$D7,$B3,$B7,$FD,$7E,$7C,$1E,$BF,$F0,$03,$E3,$C7,$F5,$5D
	.dc.b	$9F,$ED,$66,$C7,$06,$BE,$9C,$5F,$86,$3E,$07,$F6,$9B,$7E,$D6,$73
	.dc.b	$18,$EC,$FF,$65,$7C,$B5,$C1,$5A,$9F,$F0,$A3,$63,$F3,$AD,$BF,$A9
	.dc.b	$5C,$63,$15,$F9,$07,$C7,$F0,$1F,$EF,$FB,$C6,$28,$C4,$78,$23,$15
	.dc.b	$F1,$2B,$8B,$63,$33,$E2,$6E,$1E,$DF,$77,$F3,$F5,$E8,$9C,$BE,$0C
	.dc.b	$3C,$D1,$8B,$F7,$D7,$07,$A4,$FC,$3F,$96,$FE,$D5,$F4,$37,$90,$FC
	.dc.b	$7E,$CF,$F7,$0F,$C7,$66,$26,$7F,$E3,$BB,$E2,$FE,$FA,$73,$8F,$FB
	.dc.b	$A7,$DB,$7D,$AF,$35,$F7,$16,$30,$DB,$FF,$1B,$36,$B9,$3E,$5F,$89
	.dc.b	$F9,$78,$FF,$80,$5B,$A7,$FA,$92,$FF,$1F,$39,$6C,$66,$FF,$8C,$CE
	.dc.b	$5F,$F0,$47,$E1,$9F,$EF,$1A,$F6,$56,$F8,$7F,$A0,$5B,$F6,$0F,$EC
	.dc.b	$15,$FE,$4B,$F8,$9F,$95,$18,$C7,$E0,$46,$3C,$DF,$3C,$E6,$7F,$65
	.dc.b	$7C,$E7,$24,$BF,$99,$5C,$F0,$FE,$59,$93,$E4,$3A,$56,$BE,$27,$E6
	.dc.b	$78,$FE,$26,$C6,$4F,$87,$87,$FA,$3D,$B1,$E4,$23,$BC,$FF,$C5,$5F
	.dc.b	$27,$FE,$02,$8C,$65,$F0,$AF,$FC,$71,$1F,$E6,$BF,$B7,$F8,$3E,$F7
	.dc.b	$DD,$6C,$78,$23,$E0,$91,$C1,$EE,$7F,$69,$E1,$B0,$7E,$07,$CB,$2F
	.dc.b	$AF,$8F,$D5,$7F,$19,$F7,$1F,$43,$D1,$FC,$63,$72,$F7,$C7,$D0,$23
	.dc.b	$83,$E6,$BF,$9D,$F6,$3F,$BD,$46,$0F,$A3,$F1,$FF,$97,$F4,$BE,$EB
	.dc.b	$63,$27,$CC,$7E,$E9,$1C,$1A,$C3,$F1,$FF,$7C,$E2,$E7,$3F,$E9,$3F
	.dc.b	$96,$FD,$21,$F8,$AA,$B1,$F8,$27,$F1,$24,$7E,$DB,$FC,$33,$C8,$7E
	.dc.b	$2A,$F7,$B9,$57,$FA,$BE,$73,$18,$AF,$DA,$7C,$EF,$EC,$CF,$DA,$7D
	.dc.b	$25,$BF,$DE,$79,$DE,$77,$F5,$6D,$AF,$E9,$43,$B5,$19,$8F,$CE,$B5
	.dc.b	$FE,$A1,$F6,$1F,$91,$7F,$A1,$B5,$E6,$D7,$7D,$7F,$BE,$9F,$F0,$8D
	.dc.b	$FF,$AE,$7D,$0F,$E3,$11,$8A,$FE,$1D,$18,$E2,$E0,$4A,$F4,$B6,$7F
	.dc.b	$80,$D7,$1E,$CF,$F1,$E7,$90,$F9,$77,$DC,$73,$C7,$FD,$A1,$CF,$FC
	.dc.b	$0D,$FF,$07,$D3,$C8,$7B,$1F,$8B,$F1,$FF,$8F,$36,$BF,$DA,$0D,$D3
	.dc.b	$F9,$23,$FC,$8F,$D7,$93,$DB,$E2,$ED,$7F,$29,$9B,$AF,$9B,$3F,$F5
	.dc.b	$7D,$FF,$B6,$E7,$F9,$67,$93,$23,$FD,$DD,$F7,$DF,$ED,$C7,$18,$7F
	.dc.b	$33,$9B,$87,$F9,$6E,$B5,$5F,$EE,$59,$CB,$EE,$D3,$FD,$5C,$F4,$F8
	.dc.b	$FF,$F3,$93,$B3,$FD,$D9,$EA,$7F,$3D,$3F,$58,$D3,$E4,$97,$1F,$F2
	.dc.b	$4E,$ED,$39,$C7,$F4,$7A,$FE,$8D,$E7,$CB,$FF,$80,$1F,$AE,$7F,$23
	.dc.b	$7C,$91,$9A,$FE,$9C,$FF,$57,$9D,$43,$0F,$AD,$3F,$E0,$18,$C6,$F7
	.dc.b	$C0,$FE,$3D,$1F,$E3,$6A,$F9,$AF,$F8,$E2,$33,$BF,$FA,$06,$38,$E3
	.dc.b	$FE,$1C,$69,$FE,$1E,$AE,$2F,$E9,$9A,$FF,$8C,$5E,$B3,$CD,$B3,$FE
	.dc.b	$35,$EE,$7D,$2F,$9E,$73,$7E,$BF,$D6,$BF,$C6,$D7,$FE,$60,$C3,$F3
	.dc.b	$31,$89,$FE,$BB,$FB,$6F,$A2,$BF,$2F,$39,$AD,$FB,$9F,$DF,$63,$13
	.dc.b	$FE,$6A,$EF,$5F,$1C,$3C,$17,$FD,$E7,$C2,$97,$E1,$3F,$47,$EE,$BE
	.dc.b	$AF,$37,$F7,$A3,$13,$B8,$FA,$9C,$DF,$C5,$E3,$1A,$AB,$75,$FE,$09
	.dc.b	$C3,$EC,$7C,$BF,$F8,$E2,$F9,$4B,$F2,$3F,$F0,$BC,$62,$75,$3F,$E7
	.dc.b	$09,$CB,$FE,$72,$8C,$BF,$C6,$9A,$FF,$99,$5A,$FF,$A3,$1A,$9E,$BB
	.dc.b	$FE,$0D,$B5,$FE,$6D,$FD,$15,$EB,$FA,$3D,$EF,$F1,$07,$D8,$CA,$F4
	.dc.b	$FD,$B7,$DE,$7E,$0E,$E9,$FC,$B4,$E7,$3F,$F1,$54,$63,$97,$94,$7F
	.dc.b	$8B,$7E,$47,$FC,$58,$7F,$8D,$F8,$3F,$8D,$63,$D2,$31,$E2,$EA,$3E
	.dc.b	$7C,$EA,$B9,$47,$F8,$EF,$F6,$EF,$E1,$11,$FC,$FB,$9D,$18,$8F,$F0
	.dc.b	$CB,$5F,$C4,$35,$FE,$88,$F2,$E1,$FF,$14,$B6,$BF,$94,$FC,$37,$E0
	.dc.b	$6B,$FC,$67,$39,$7F,$A6,$7C,$13,$D5,$FF,$29,$56,$F0,$FF,$19,$CE
	.dc.b	$70,$FF,$66,$F3,$7F,$96,$E3,$18,$7A,$5A,$FF,$8C,$1A,$FE,$7D,$AF
	.dc.b	$A3,$E0,$7F,$C9,$B9,$7E,$22,$3D,$F1,$82,$5F,$C7,$7C,$BC,$1F,$E6
	.dc.b	$BA,$FC,$7E,$3F,$E9,$3F,$C3,$FF,$C7,$17,$FE,$7F,$6E,$1F,$DA,$B9
	.dc.b	$11,$88,$FF,$4F,$35,$FF,$47,$3C,$2B,$A7,$BD,$F1,$38,$BF,$D2,$93
	.dc.b	$9C,$37,$38,$34,$FF,$A7,$A3,$57,$FE,$A2,$6E,$1E,$2A,$DC,$FF,$A8
	.dc.b	$A7,$23,$E7,$FF,$D3,$75,$E8,$6C,$66,$FD,$D1,$83,$CA,$7D,$A9,$B1
	.dc.b	$FB,$AA,$FF,$7C,$FC,$5F,$FB,$EB,$EC,$3F,$D7,$91,$FC,$8E,$71,$F1
	.dc.b	$1F,$EC,$AA,$B3,$91,$5C,$1F,$F6,$F4,$16,$79,$5E,$1B,$0D,$4E,$65
	.dc.b	$58,$C3,$FC,$3F,$7B,$9E,$71,$E3,$7C,$2E,$AF,$CB,$1E,$23,$27,$9F
	.dc.b	$0E,$D7,$A8,$F1,$CD,$7C,$27,$03,$D5,$7E,$1A,$E3,$1E,$23,$E1,$61
	.dc.b	$83,$DD,$3E,$77,$73,$8A,$ED,$7E,$1A,$E9,$87,$AE,$B5,$87,$84,$FA
	.dc.b	$A5,$9D,$BF,$E4,$2A,$E5,$7D,$23,$4D,$C3,$99,$E5,$C3,$E3,$4F,$A4
	.dc.b	$F8,$D8,$78,$0F,$3C,$69,$E1,$1E,$E8,$FA,$13,$D1,$A7,$47,$71,$F2
	.dc.b	$2B,$ED,$3E,$73,$FE,$00,$C7,$D9,$C7,$F0,$B1,$96,$BF,$65,$BE,$8F
	.dc.b	$F8,$FB,$6F,$F8,$01,$5F,$8D,$AF,$F3,$0F,$9E,$9C,$C7,$E4,$6B,$83
	.dc.b	$E0,$31,$3B,$70,$72,$9F,$3B,$FE,$6C,$F3,$FB,$5F,$E7,$6D,$18,$CB
	.dc.b	$C5,$1D,$A7,$58,$75,$78,$56,$EF,$C8,$D7,$C0,$7D,$3A,$F3,$E7,$C2
	.dc.b	$3F,$37,$1E,$06,$DF,$A6,$7B,$E3,$E8,$AD,$57,$3B,$ED,$1E,$5C,$7C
	.dc.b	$F8,$71,$7D,$33,$F2,$F0,$F0,$5F,$E8,$6F,$C7,$5C,$30,$F1,$DE,$E3
	.dc.b	$8D,$79,$26,$3C,$67,$4F,$C0,$33,$7E,$87,$CC,$7A,$63,$CF,$87,$C9
	.dc.b	$BF,$04,$F3,$3C,$27,$27,$35,$E8,$23,$83,$8E,$3E,$C2,$38,$35,$3E
	.dc.b	$53,$B1,$F0,$6B,$E3,$2C,$9F,$86,$3E,$87,$F2,$59,$FF,$15,$9C,$C6
	.dc.b	$35,$FE,$45,$18,$BF,$F1,$7F,$E7,$EF,$99,$D4,$75,$C7,$A3,$FE,$BE
	.dc.b	$31,$87,$F9,$7E,$F9,$47,$F5,$99,$DB,$7E,$1B,$9C,$F9,$9B,$27,$BF
	.dc.b	$8E,$C5,$F3,$3A,$99,$87,$F9,$1F,$0E,$4F,$A2,$72,$FE,$BF,$3F,$6E
	.dc.b	$3D,$63,$83,$5F,$F4,$7E,$1D,$9C,$E1,$EB,$9D,$47,$EE,$F0,$EE,$FA
	.dc.b	$CF,$85,$DF,$FE,$22,$EC,$F8,$ED,$8F,$F1,$44,$7B,$30,$7C,$CE,$BA
	.dc.b	$D3,$FE,$87,$BE,$17,$FE,$2D,$9C,$BD,$9A,$FF,$86,$7B,$38,$57,$F9
	.dc.b	$16,$3C,$6D,$74,$62,$3C,$5F,$2D,$23,$FC,$E1,$7E,$29,$F2,$7B,$12
	.dc.b	$7F,$C8,$0E,$DF,$F1,$9B,$E7,$C0,$D2,$47,$FA,$7F,$C3,$7C,$3F,$DC
	.dc.b	$F2,$3F,$2D,$3F,$E5,$CD,$BE,$FD,$6A,$30,$43,$FD,$87,$3C,$1F,$F4
	.dc.b	$45,$7E,$51,$7E,$BD,$FD,$F7,$9F,$FD,$C5,$CF,$F8,$AF,$5F,$F3,$55
	.dc.b	$FE,$84,$AD,$72,$FF,$BD,$A7,$33,$FE,$D0,$6D,$F2,$8E,$1D,$5D,$FE
	.dc.b	$D7,$FD,$63,$1F,$E8,$E6,$BF,$8F,$3F,$C1,$8E,$27,$FD,$89,$FA,$BB
	.dc.b	$E1,$79,$79,$9F,$EC,$5B,$5F,$27,$FA,$41,$88,$D4,$66,$3F,$A6,$3D
	.dc.b	$72,$FE,$9D,$F8,$1F,$F2,$CC,$7F,$DA,$4D,$8D,$DF,$CE,$F2,$3E,$27
	.dc.b	$6C,$C7,$FC,$FF,$15,$F8,$51,$70,$FF,$0E,$77,$7F,$AC,$6B,$EB,$FC
	.dc.b	$14,$BF,$F4,$3B,$FA,$0F,$C6,$3F,$55,$1F,$EB,$D8,$D3,$4F,$C7,$C6
	.dc.b	$AF,$FD,$D1,$FA,$BB,$EA,$7E,$E3,$6B,$FD,$76,$D7,$E7,$07,$FD,$88
	.dc.b	$EA,$3E,$64,$6E,$3E,$5B,$5F,$EE,$8D,$7F,$C8,$4F,$59,$FF,$7D,$7E
	.dc.b	$DE,$F9,$AF,$C7,$36,$BB,$1F,$ED,$4E,$4F,$DD,$B7,$C1,$FC,$81,$FC
	.dc.b	$6B,$E0,$1A,$62,$74,$43,$F1,$1B,$27,$71,$FE,$E0,$7E,$F2,$85,$BC
	.dc.b	$1F,$CF,$1D,$F8,$6D,$79,$62,$BD,$AF,$9D,$F2,$E3,$FD,$31,$27,$FD
	.dc.b	$E5,$F0,$8F,$CB,$DF,$2F,$99,$FD,$35,$5E,$0F,$F3,$05,$EE,$B1,$CD
	.dc.b	$FA,$DA,$B6,$AD,$3F,$27,$87,$F6,$C4,$7E,$07,$B1,$5F,$DA,$45,$C7
	.dc.b	$F2,$7A,$2B,$E7,$7C,$A7,$EF,$FE,$4E,$61,$F2,$63,$2F,$D7,$8C,$BF
	.dc.b	$0E,$3F,$9B,$FD,$A7,$F5,$6C,$3F,$CF,$5E,$42,$31,$AF,$FA,$1E,$31
	.dc.b	$E6,$FF,$56,$F8,$E7,$C3,$F2,$8F,$E7,$57,$FD,$83,$2F,$F2,$15,$F0
	.dc.b	$7F,$A3,$C6,$5E,$8F,$67,$EF,$73,$3E,$E3,$F8,$D1,$D6,$3F,$D3,$19
	.dc.b	$FF,$A3,$D7,$D6,$3E,$2F,$F8,$63,$63,$FB,$B1,$A8,$C9,$89,$FA,$71
	.dc.b	$FC,$41,$AF,$F8,$23,$9B,$F0,$2F,$F7,$F8,$CB,$FE,$A9,$E4,$7E,$D1
	.dc.b	$FD,$0A,$3F,$12,$D7,$F9,$24,$6B,$CD,$FF,$05,$BC,$3E,$7F,$F9,$D7
	.dc.b	$D0,$BE,$5F,$E9,$FA,$D3,$0F,$7F,$2F,$F9,$AE,$B5,$1F,$D6,$38,$7F
	.dc.b	$BA,$BF,$2A,$BE,$89,$1F,$F0,$B7,$6F,$EA,$3D,$0F,$E8,$84,$8E,$53
	.dc.b	$F8,$2F,$95,$91,$D5,$FE,$39,$9E,$23,$F9,$AF,$2F,$E1,$67,$EF,$92
	.dc.b	$3E,$EB,$FD,$47,$E9,$64,$73,$3D,$1F,$23,$F9,$02,$3F,$BF,$FB,$8A
	.dc.b	$F6,$D1,$3F,$8C,$24,$7E,$29,$FE,$BD,$CB,$23,$A9,$F1,$7D,$5F,$CE
	.dc.b	$91,$F9,$63,$E7,$1F,$2F,$DA,$38,$1F,$96,$9F,$CD,$91,$C1,$FC,$87
	.dc.b	$E5,$BF,$B7,$1D,$5F,$EC,$5A,$BF,$34,$FE,$2D,$1F,$E0,$E6,$BA,$6F
	.dc.b	$4F,$FB,$A7,$1F,$1F,$C0,$7F,$8A,$A7,$75,$FA,$A6,$BF,$B3,$6A,$BF
	.dc.b	$07,$93,$FE,$2B,$9E,$BC,$FF,$DF,$FE,$AB,$FC,$08,$75,$9E,$DD,$F3
	.dc.b	$0F,$33,$FE,$60,$C9,$F9,$D3,$E4,$F5,$5F,$82,$CB,$F6,$CF,$F9,$09
	.dc.b	$AE,$6F,$C4,$7F,$84,$FE,$2E,$47,$63,$E2,$73,$7B,$C8,$FD,$6C,$7C
	.dc.b	$3D,$98,$8F,$04,$FB,$48,$FE,$7B,$C3,$FF,$22,$F0,$FF,$78,$7F,$AE
	.dc.b	$6F,$FE,$1F,$F4,$FF,$C4,$9F,$03,$2F,$FA,$D4,$7F,$89,$CC,$3F,$57
	.dc.b	$87,$E9,$11,$FF,$42,$F9,$D5,$D1,$F7,$63,$FC,$BD,$1F,$E7,$EF,$4D
	.dc.b	$F4,$54,$95,$FE,$79,$F6,$FF,$88,$65,$FE,$92,$AD,$CF,$F9,$F5,$AF
	.dc.b	$F8,$B7,$A5,$18,$D7,$FD,$39,$7C,$AB,$FD,$45,$3C,$1F,$F3,$C3,$5E
	.dc.b	$D8,$7F,$A5,$E3,$11,$C3,$E4,$7D,$04,$7F,$D4,$7E,$95,$7C,$52,$9C
	.dc.b	$61,$FE,$26,$F8,$4F,$F0,$35,$7F,$85,$BE,$95,$F2,$DF,$E5,$98,$12
	.dc.b	$33,$A7,$F3,$66,$C7,$FA,$B5,$AF,$F9,$13,$E2,$FF,$6F,$CD,$FC,$F1
	.dc.b	$B1,$FE,$B3,$6B,$BF,$03,$F7,$CF,$3F,$19,$F1,$BD,$DB,$19,$8C,$1F
	.dc.b	$EB,$DC,$7E,$7F,$90,$E3,$FA,$8D,$AF,$A9,$8B,$C7,$C6,$FF,$EC,$FE
	.dc.b	$1F,$F5,$E6,$1C,$09,$7C,$2B,$71,$F8,$1C,$2C,$FF,$B6,$FD,$CF,$C4
	.dc.b	$CF,$8D,$B1,$FE,$66,$CB,$FC,$85,$C9,$FD,$8E,$3F,$DB,$73,$A7,$FD
	.dc.b	$70,$F0,$CB,$FD,$C4,$DB,$FF,$70,$E6,$F9,$3F,$0E,$3D,$27,$FD,$9D
	.dc.b	$5B,$C3,$C9,$8E,$B0,$E1,$5A,$8F,$4F,$89,$FC,$87,$AA,$F8,$3F,$E5
	.dc.b	$DA,$FF,$57,$FE,$E0,$FA,$1C,$D4,$9F,$E4,$CF,$29,$D4,$60,$FF,$22
	.dc.b	$78,$9F,$1B,$CE,$FF,$21,$F4,$DF,$6F,$F8,$B5,$7F,$95,$63,$15,$BF
	.dc.b	$0B,$F2,$73,$C8,$D1,$FD,$13,$77,$E1,$4F,$AA,$F9,$91,$F9,$88,$C4
	.dc.b	$FF,$83,$1B,$1F,$E9,$83,$8C,$F0,$6C,$F5,$9D,$F2,$FF,$86,$7F,$1B
	.dc.b	$FF,$62,$CE,$6F,$F0,$37,$E2,$F8,$1F,$C1,$F7,$3E,$99,$F0,$BA,$CF
	.dc.b	$FC,$DB,$8F,$F2,$79,$CB,$DE,$FC,$53,$AC,$3C,$CD,$7F,$D5,$ED,$7F
	.dc.b	$6A,$78,$67,$91,$F1,$72,$FF,$0B,$38,$8F,$F4,$9C,$BC,$AF,$3F,$C6
	.dc.b	$7F,$BF,$AB,$E0,$46,$3E,$17,$D0,$F9,$09,$1F,$2D,$B7,$D1,$F8,$7F
	.dc.b	$85,$FF,$9F,$E7,$99,$F2,$EF,$BF,$C1,$FC,$36,$C6,$4F,$AC,$FC,$22
	.dc.b	$38,$35,$7F,$EA,$CA,$FF,$28,$46,$38,$3F,$C9,$DE,$E7,$F8,$1A,$3F
	.dc.b	$D4,$1E,$2F,$D6,$8C,$5F,$FB,$A6,$31,$E3,$7E,$06,$73,$3F,$ED,$C6
	.dc.b	$BF,$39,$AF,$F2,$47,$9D,$F8,$27,$F5,$7F,$85,$7F,$0A,$AF,$F6,$0D
	.dc.b	$6B,$0E,$5F,$27,$FB,$0F,$A7,$7D,$17,$96,$30,$D9,$FF,$31,$D2,$7C
	.dc.b	$C4,$8C,$BB,$BF,$F0,$DE,$0F,$77,$19,$FF,$21,$F2,$DF,$4E,$F9,$CF
	.dc.b	$E3,$31,$87,$F3,$7F,$99,$BD,$7C,$D6,$F4,$FD,$84,$E5,$FF,$3B,$B7
	.dc.b	$3F,$D6,$35,$FC,$A8,$70,$FF,$A5,$73,$FF,$23,$CF,$4B,$F9,$3E,$FF
	.dc.b	$89,$FE,$2B,$E5,$7C,$69,$FF,$21,$F9,$28,$C6,$0D,$7F,$9D,$FE,$2B
	.dc.b	$F9,$B8,$FF,$1E,$CE,$A7,$FD,$25,$F5,$3F,$A9,$9F,$EB,$AD,$4D,$35
	.dc.b	$6C,$DF,$33,$F7,$8F,$E8,$3F,$6C,$FC,$C7,$F2,$78,$7D,$29,$FE,$0B
	.dc.b	$E5,$3A,$BF,$91,$7A,$4F,$83,$CC,$61,$BC,$1F,$EA,$2A,$E0,$D9,$FE
	.dc.b	$2D,$18,$C3,$FD,$0F,$1C,$1F,$F2,$DF,$17,$C1,$DF,$FD,$61,$5B,$7F
	.dc.b	$DA,$7F,$41,$FC,$B7,$AB,$F9,$93,$71,$FF,$7D,$37,$0F,$F7,$93,$6B
	.dc.b	$FD,$B7,$1F,$CD,$7E,$A7,$F9,$D3,$6B,$FC,$EF,$83,$CE,$FF,$04,$C6
	.dc.b	$3C,$C5,$F0,$8C,$3F,$D1,$7F,$61,$FE,$B1,$F1,$7F,$EC,$59,$C7,$AB
	.dc.b	$FD,$BA,$31,$1D,$A7,$7F,$11,$FD,$FE,$31,$5B,$8C,$67,$F8,$59,$FF
	.dc.b	$0D,$E6,$BE,$31,$D8,$FF,$47,$E1,$DA,$BF,$D0,$BF,$06,$F9,$4F,$EA
	.dc.b	$95,$FD,$A2,$30,$FB,$4C,$44,$7C,$DF,$01,$6B,$C4,$31,$F8,$DE,$9F
	.dc.b	$F6,$3F,$AE,$FE,$C8,$DB,$F9,$7C,$1F,$D4,$9F,$4A,$9D,$C6,$9F,$EA
	.dc.b	$7E,$27,$F7,$EC,$3B,$9F,$E2,$AF,$96,$CF,$4E,$9B,$7F,$EF,$0A,$E1
	.dc.b	$5F,$B8,$6B,$FD,$39,$AF,$F9,$11,$ED,$C9,$FE,$FD,$6C,$FF,$54,$FA
	.dc.b	$77,$C9,$F6,$BF,$83,$AB,$F4,$0F,$F0,$B7,$82,$FA,$BF,$CC,$63,$F5
	.dc.b	$A7,$31,$8B,$F9,$11,$8F,$4F,$FC,$18,$DA,$FE,$0B,$F2,$CF,$A0,$FF
	.dc.b	$08,$7C,$9F,$F8,$3B,$0F,$F5,$FC,$E6,$BF,$2F,$39,$DF,$FC,$8D,$9A
	.dc.b	$B5,$7C,$30,$EF,$CB,$FE,$28,$9C,$DF,$F9,$36,$9B,$FF,$1C,$7E,$8B
	.dc.b	$FE,$71,$CB,$FC,$30,$E6,$78,$39,$9D,$F8,$1F,$E1,$38,$E5,$1F,$CA
	.dc.b	$9A,$FC,$D8,$D1,$FE,$12,$8E,$95,$FE,$04,$FA,$1F,$2B,$B3,$FC,$51
	.dc.b	$F1,$7F,$E4,$5B,$7C,$DB,$97,$FC,$37,$39,$8C,$57,$0E,$FC,$9A,$D6
	.dc.b	$C7,$D0,$F1,$7E,$7F,$47,$C5,$6C,$64,$99,$7E,$0B,$8D,$F3,$FB,$9F
	.dc.b	$E7,$BD,$CF,$A9,$FF,$1D,$F2,$A3,$13,$FE,$39,$8C,$4E,$A7,$38,$7F
	.dc.b	$50,$9C,$D7,$4A,$FF,$2F,$FE,$84,$F9,$EF,$DD,$46,$19,$3F,$C6,$BE
	.dc.b	$43,$F8,$96,$1F,$B8,$C1,$87,$1A,$FE,$A6,$D9,$D3,$79,$BF,$D5,$DE
	.dc.b	$2F,$49,$FE,$17,$ED,$79,$C9,$7E,$BD,$3F,$C1,$B8,$FB,$E7,$F8,$71
	.dc.b	$E8,$F8,$9B,$5F,$EB,$4C,$7F,$D3,$9F,$18,$FC,$13,$F6,$AD,$E3,$FF
	.dc.b	$31,$B6,$BF,$D4,$D3,$A3,$F6,$F9,$A7,$4F,$FB,$67,$E6,$BF,$D4,$D5
	.dc.b	$DC,$FF,$61,$CE,$B0,$F0,$DF,$28,$C4,$EA,$B8,$DF,$FB,$8E,$B4,$4C
	.dc.b	$3F,$D3,$33,$AC,$39,$C7,$FA,$72,$5F,$47,$EA,$6E,$1E,$A8,$C1,$1F
	.dc.b	$36,$1E,$B9,$DB,$EB,$F8,$3E,$95,$FE,$78,$8F,$F6,$EF,$D4,$FF,$B7
	.dc.b	$BE,$A3,$BC,$7F,$1F,$C3,$15,$FE,$87,$AB,$39,$15,$C1,$FF,$39,$C7
	.dc.b	$FB,$A3,$E7,$71,$1F,$ED,$AB,$C5,$F3,$84,$EB,$C9,$FF,$6B,$3C,$21
	.dc.b	$7E,$07,$F9,$85,$D7,$D2,$78,$AF,$F9,$C4,$66,$38,$99,$3D,$18,$75
	.dc.b	$AE,$58,$7F,$B9,$63,$13,$E4,$BE,$6D,$3F,$C7,$75,$EA,$7E,$16,$1C
	.dc.b	$2B,$D5,$87,$86,$BA,$B6,$BC,$47,$A8,$A7,$AA,$78,$B8,$3D,$2E,$0E
	.dc.b	$D7,$E4,$7C,$F7,$C0,$B1,$E8,$77,$3F,$02,$3E,$84,$F3,$7E,$94,$F1
	.dc.b	$69,$C4,$B3,$EE,$6B,$E9,$8F,$8F,$B7,$DA,$8F,$8A,$7E,$AE,$3E,$F4
	.dc.b	$66,$3F,$0A,$64,$1F,$E5,$BF,$81,$FF,$2D,$39,$31,$B7,$EC,$2B,$FD
	.dc.b	$7F,$ED,$FF,$A0,$E3,$13,$CA,$30,$F1,$6C,$FE,$55,$7D,$4F,$FB,$22
	.dc.b	$30,$DC,$7A,$E1,$E2,$9E,$0D,$8F,$0C,$EE,$FC,$0D,$7C,$E7,$E1,$CB
	.dc.b	$97,$67,$F3,$58,$78,$9B,$7F,$02,$7B,$E3,$F8,$D9,$FF,$3F,$5F,$5B
	.dc.b	$E1,$1D,$5F,$6E,$1E,$8A,$EB,$5F,$EE,$4B,$F0,$5F,$E4,$2F,$E2,$C7
	.dc.b	$68,$C4,$EE,$FD,$B1,$E4,$B8,$E8,$EA,$62,$B9,$63,$D8,$95,$E3,$9F
	.dc.b	$36,$3E,$D8,$EB,$1C,$CF,$24,$65,$FB,$4D,$7E,$09,$E0,$37,$7E,$4A
	.dc.b	$EF,$1F,$48,$F9,$18,$FC,$47,$9C,$F1,$C3,$E1,$9F,$04,$FE,$3F,$97
	.dc.b	$F4,$9F,$B2,$FC,$9D,$7F,$37,$6D,$7F,$3E,$6E,$9F,$CF,$73,$B5,$FC
	.dc.b	$BE,$7F,$D7,$EF,$F4,$6E,$04,$62,$34,$7F,$52,$71,$83,$EC,$1E,$0F
	.dc.b	$48,$C3,$63,$35,$FC,$FA,$5C,$8E,$2D,$5C,$AD,$5F,$07,$FA,$E4,$E5
	.dc.b	$FF,$39,$DF,$67,$75,$E1,$BE,$91,$83,$DB,$3B,$7F,$DB,$AD,$7E,$4B
	.dc.b	$5F,$DE,$F3,$FF,$86,$78,$BE,$3B,$63,$FC,$3F,$18,$EA,$7C,$46,$BA
	.dc.b	$BF,$3F,$5F,$52,$F8,$3F,$E1,$D8,$CB,$D9,$AA,$3F,$C2,$FA,$F5,$24
	.dc.b	$7F,$87,$7D,$9A,$FF,$86,$22,$F1,$38,$BF,$D4,$FD,$AB,$F2,$7B,$73
	.dc.b	$0F,$F1,$BB,$C2,$BC,$4F,$DF,$C7,$FC,$70,$DC,$BF,$DD,$35,$FE,$1F
	.dc.b	$8D,$7E,$0D,$F0,$23,$FD,$FA,$D7,$FC,$60,$DC,$BC,$91,$FE,$82,$EE
	.dc.b	$F9,$73,$9C,$7F,$D7,$BE,$75,$AF,$F5,$73,$72,$56,$65,$F5,$54,$9B
	.dc.b	$FE,$72,$3F,$27,$E0,$D9,$D7,$4D,$31,$5B,$E2,$FC,$FD,$FF,$84,$8F
	.dc.b	$A2,$57,$2F,$FA,$D3,$A7,$FD,$4A,$E7,$8F,$F4,$CF,$03,$F5,$73,$C1
	.dc.b	$FD,$A9,$92,$BD,$37,$FF,$49,$7C,$E7,$F9,$B2,$7C,$4D,$7F,$D8,$73
	.dc.b	$BD,$FF,$2B,$F7,$9F,$ED,$D9,$FF,$14,$EE,$8C,$47,$F9,$85,$B8,$7D
	.dc.b	$2F,$53,$FD,$75,$E8,$FE,$72,$73,$87,$D2,$F9,$2F,$E0,$B4,$E3,$46
	.dc.b	$FD,$AC,$F9,$7F,$D9,$72,$FF,$58,$FD,$4B,$EA,$BF,$4E,$DD,$7F,$C7
	.dc.b	$B9,$7F,$B2,$7C,$5E,$26,$AA,$4E,$B0,$E9,$A7,$F1,$CF,$9B,$7B,$1F
	.dc.b	$75,$B5,$F5,$DA,$FF,$BA,$38,$DD,$7B,$AF,$AB,$5A,$AF,$A6,$D7,$B4
	.dc.b	$68,$ED,$C1,$23,$B4,$7C,$4F,$71,$F8,$4D,$F9,$59,$5F,$F0,$02,$34
	.dc.b	$61,$FF,$74,$CA,$73,$1F,$63,$F7,$1F,$4F,$17,$FB,$69,$BA,$FF,$12
	.dc.b	$6C,$B8,$5F,$05,$FE,$DF,$8E,$59,$F9,$7D,$99,$18,$3E,$8C,$FE,$E9
	.dc.b	$63,$FB,$4F,$82,$7D,$57,$E0,$A5,$18,$E7,$FC,$1D,$F1,$BF,$C5,$37
	.dc.b	$4F,$F1,$1D,$6E,$FF,$CF,$6D,$7B,$C7,$F8,$A3,$3F,$4F,$89,$D0,$FE
	.dc.b	$A7,$5B,$C3,$C9,$97,$8E,$B7,$87,$AB,$9F,$E9,$7B,$52,$30,$43,$E7
	.dc.b	$B9,$7C,$27,$C9,$CF,$F1,$AF,$CB,$FB,$47,$C0,$7D,$DF,$62,$4F,$E6
	.dc.b	$27,$2D,$9C,$BF,$72,$7F,$4F,$3A,$8E,$E3,$9D,$F3,$9F,$99,$24,$7E
	.dc.b	$E2,$3F,$4C,$F0,$76,$FF,$71,$6B,$FD,$D9,$AF,$E6,$A3,$F3,$A7,$EE
	.dc.b	$3E,$B7,$F0,$91,$F0,$34,$46,$36,$FE,$67,$3E,$5A,$FE,$51,$F8,$07
	.dc.b	$FB,$2E,$3F,$91,$35,$F5,$B6,$7F,$9D,$8F,$D5,$3F,$47,$3F,$CA,$7E
	.dc.b	$BD,$E6,$7F,$24,$6C,$7F,$82,$DA,$FF,$84,$1A,$FF,$86,$38,$BF,$25
	.dc.b	$7F,$83,$E5,$FF,$74,$47,$E5,$65,$8F,$E4,$5F,$F3,$9F,$B7,$FC,$95
	.dc.b	$FF,$3A,$FE,$89,$E4,$7E,$9B,$70,$F8,$33,$B9,$4E,$DF,$E6,$ED,$7F
	.dc.b	$C5,$AD,$7A,$F7,$7F,$40,$97,$D7,$FF,$51,$6C,$FF,$4B,$6B,$E2,$9F
	.dc.b	$EA,$D5,$FD,$12,$FE,$9B,$72,$EF,$5F,$E1,$0E,$1F,$8F,$5A,$8F,$E9
	.dc.b	$F5,$FD,$41,$F4,$98,$21,$FD,$13,$C5,$7C,$7F,$F3,$43,$C3,$1F,$D9
	.dc.b	$FF,$01,$FE,$B6,$FE,$61,$B7,$F8,$66,$E3,$FE,$24,$C5,$E9,$F1,$F8
	.dc.b	$BF,$EC,$35,$CB,$0F,$EB,$ED,$7F,$CA,$AD,$7F,$D5,$2F,$47,$F8,$E7
	.dc.b	$88,$FA,$8F,$E5,$C7,$58,$FE,$E9,$3F,$91,$C7,$9F,$B3,$E9,$3F,$CB
	.dc.b	$7F,$46,$FC,$03,$FA,$C5,$F0,$9F,$29,$1F,$F2,$E5,$FE,$75,$F3,$C7
	.dc.b	$2F,$29,$F1,$B4,$ED,$3F,$E0,$5F,$9C,$FD,$F7,$07,$EA,$9B,$9F,$F9
	.dc.b	$C9,$B2,$CF,$C9,$F5,$BF,$E1,$59,$FE,$0D,$3F,$E0,$46,$BC,$7C,$2F
	.dc.b	$F0,$5C,$62,$3F,$D9,$EF,$4A,$FE,$BB,$E2,$F9,$CE,$95,$D7,$B6,$5F
	.dc.b	$9D,$F8,$BF,$B7,$3E,$5B,$33,$FE,$80,$C7,$F6,$2F,$A6,$BC,$24,$7F
	.dc.b	$DB,$FA,$7F,$8E,$B2,$7D,$8B,$EA,$3E,$8B,$6B,$C0,$7C,$7F,$8F,$F8
	.dc.b	$04,$79,$EF,$FE,$EA,$BF,$EB,$9C,$F9,$9F,$25,$7C,$22,$3F,$E9,$6F
	.dc.b	$19,$E3,$BE,$7B,$FB,$A7,$CC,$CA,$FE,$71,$1F,$CD,$0C,$3F,$DC,$A3
	.dc.b	$0F,$F8,$A3,$EA,$5F,$1F,$FE,$7E,$6C,$66,$3F,$D0,$3E,$2E,$23,$FA
	.dc.b	$5F,$BA,$AC,$70,$FF,$97,$A3,$17,$CC,$E2,$7F,$A9,$7C,$37,$D6,$52
	.dc.b	$61,$FE,$A5,$F0,$9F,$83,$62,$3F,$D3,$F5,$C1,$FF,$55,$35,$FF,$51
	.dc.b	$CF,$07,$35,$FE,$B6,$9D,$47,$FA,$79,$A8,$77,$A3,$31,$8F,$8F,$7A
	.dc.b	$5F,$EB,$D6,$C7,$FA,$ED,$AF,$8B,$AF,$53,$23,$4F,$FB,$17,$1F,$C9
	.dc.b	$47,$F9,$BF,$EB,$FF,$B2,$E3,$FE,$CF,$6C,$7F,$B5,$1A,$FF,$9A,$9A
	.dc.b	$FF,$B8,$30,$FF,$1F,$F3,$FF,$B5,$DB,$1F,$ED,$C6,$BF,$E7,$26,$BF
	.dc.b	$EE,$89,$FE,$4B,$F1,$DE,$1F,$27,$FD,$93,$3D,$DA,$ED,$AF,$C5,$F1
	.dc.b	$FF,$B1,$F5,$FC,$BC,$98,$FF,$1B,$8F,$F3,$0F,$C5,$BE,$A7,$78,$71
	.dc.b	$9F,$95,$86,$30,$E2,$EB,$AD,$87,$1A,$D9,$F4,$79,$3F,$BD,$E1,$E1
	.dc.b	$CB,$F3,$73,$FE,$E0,$6C,$7F,$B8,$1A,$FF,$B8,$1C,$F3,$7F,$AD,$B2
	.dc.b	$F4,$DF,$87,$07,$D9,$79,$6F,$78,$F7,$AC,$1B,$9B,$1F,$AD,$F9,$5F
	.dc.b	$7F,$47,$E9,$DE,$9F,$01,$8E,$A4,$52,$57,$C8,$9B,$AF,$F9,$AB,$1F
	.dc.b	$89,$E6,$FF,$11,$AF,$F4,$6B,$E4,$9F,$E9,$7E,$1E,$AB,$6D,$AD,$BF
	.dc.b	$4A,$5F,$59,$FE,$FA,$9F,$F5,$BC,$7E,$96,$30,$E1,$AF,$FA,$F1,$AF
	.dc.b	$EA,$CE,$7F,$17,$D6,$38,$F0,$FF,$AC,$71,$F7,$DF,$FB,$3E,$72,$F1
	.dc.b	$7F,$C1,$6F,$FB,$4F,$EB,$3F,$D1,$0E,$FC,$0F,$44,$63,$1F,$F2,$7D
	.dc.b	$70,$7C,$D3,$F8,$47,$D4,$D7,$DE,$D8,$E1,$87,$F9,$11,$B7,$FE,$77
	.dc.b	$C1,$EA,$7A,$EF,$E4,$7D,$63,$EC,$FC,$73,$86,$DF,$FB,$AE,$74,$FC
	.dc.b	$CC,$91,$87,$E7,$B5,$FF,$73,$B5,$F9,$8E,$F2,$FF,$2C,$B6,$7C,$DA
	.dc.b	$78,$5F,$07,$1F,$C6,$8F,$3B,$63,$27,$8E,$7F,$3A,$D8,$F8,$D1,$8B
	.dc.b	$F2,$C7,$29,$FC,$7F,$C4,$D9,$FE,$87,$F4,$DF,$CF,$27,$FA,$EC,$7F
	.dc.b	$2D,$F9,$3F,$F6,$9C,$7F,$BD,$A7,$53,$FE,$86,$9C,$BC,$BC,$5F,$B1
	.dc.b	$CE,$AD,$CF,$E3,$63,$13,$A9,$CC,$FF,$3B,$9C,$DF,$7A,$FD,$3F,$C8
	.dc.b	$FE,$CA,$FC,$3A,$FF,$57,$5F,$57,$FA,$3C,$9F,$35,$B8,$C3,$71,$FD
	.dc.b	$2E,$5E,$59,$F1,$12,$33,$7D,$1F,$81,$1F,$B8,$E2,$E5,$A7,$FA,$9B
	.dc.b	$EA,$BF,$BD,$57,$C8,$84,$61,$F9,$BE,$BF,$ED,$E3,$1A,$7A,$63,$E1
	.dc.b	$12,$BC,$2F,$E8,$7C,$07,$37,$F6,$C9,$FA,$53,$FC,$2B,$CB,$F8,$8F
	.dc.b	$F0,$3F,$A9,$FD,$66,$5F,$BB,$0E,$1F,$DF,$35,$FE,$D4,$DD,$BE,$C7
	.dc.b	$C4,$7D,$52,$61,$F1,$1F,$DA,$7D,$4B,$EA,$7F,$1D,$AA,$BF,$BA,$4E
	.dc.b	$79,$FF,$8F,$FC,$B7,$F4,$88,$C4,$FE,$3B,$1F,$F2,$BF,$C7,$7F,$1B
	.dc.b	$E0,$FF,$3F,$CE,$76,$4E,$7C,$AF,$E6,$71,$89,$FB,$6D,$DF,$E0,$61
	.dc.b	$D2,$73,$7F,$94,$AD,$69,$FE,$1F,$6C,$FF,$82,$C7,$E0,$EF,$BD,$EF
	.dc.b	$7F,$CE,$3C,$DF,$E7,$3E,$95,$BF,$E7,$2D,$CF,$FD,$B1,$CB,$FE,$57
	.dc.b	$9F,$E7,$B5,$C1,$FD,$6E,$CF,$8D,$7B,$3F,$D1,$30,$FE,$D1,$18,$AF
	.dc.b	$E8,$B1,$89,$D4,$72,$9D,$4F,$80,$F8,$BE,$BF,$2F,$9E,$B5,$B6,$CF
	.dc.b	$C6,$DF,$F2,$CF,$ED,$5A,$FE,$CD,$AB,$87,$FD,$1B,$7C,$B8,$7F,$01
	.dc.b	$2C,$FD,$9F,$21,$9C,$BF,$C6,$D1,$EF,$FA,$4F,$69,$EE,$FA,$EB,$3E
	.dc.b	$B6,$DF,$FC,$01,$D5,$18,$C7,$BE,$1F,$27,$E9,$3E,$3F,$8B,$F2,$1B
	.dc.b	$8F,$C5,$6E,$3C,$DB,$7E,$88,$7D,$47,$CA,$8C,$71,$7F,$73,$9C,$F1
	.dc.b	$7F,$6F,$8C,$57,$D0,$9F,$D7,$35,$FE,$D4,$D7,$FC,$AC,$D7,$9E,$1F
	.dc.b	$C1,$E3,$C0,$FF,$71,$FC,$69,$F0,$BF,$DF,$67,$35,$F3,$7E,$6A,$EA
	.dc.b	$5C,$BF,$E2,$F7,$26,$31,$FC,$7B,$88,$D6,$7B,$78,$70,$CA,$FC,$67
	.dc.b	$4D,$C7,$E9,$37,$6F,$A3,$C5,$FA,$5A,$EB,$7F,$E1,$3F,$A9,$F4,$C7
	.dc.b	$B7,$9F,$1B,$FA,$DB,$1F,$DA,$E7,$79,$7F,$86,$1A,$FF,$19,$6B,$FE
	.dc.b	$15,$77,$8F,$D9,$6C,$FE,$3B,$F5,$57,$DB,$F6,$B5,$FD,$FF,$AB,$67
	.dc.b	$FE,$29,$6E,$1F,$AB,$FA,$BC,$4F,$EC,$9B,$19,$3F,$9F,$4F,$0B,$EB
	.dc.b	$1F,$9E,$E1,$57,$D3,$3F,$77,$F3,$36,$7F,$70,$DA,$FC,$F3,$71,$FF
	.dc.b	$59,$7D,$1B,$E6,$FF,$B5,$B6,$FF,$AC,$B6,$FF,$8F,$B5,$FE,$DC,$E6
	.dc.b	$B9,$D7,$F9,$8F,$E3,$3F,$CB,$73,$9B,$FD,$AF,$C0,$27,$51,$FE,$80
	.dc.b	$EF,$B8,$FB,$8F,$DF,$56,$E7,$FD,$0B,$5F,$E6,$FA,$DE,$3E,$AF,$7B
	.dc.b	$D2,$4A,$F4,$B7,$0F,$F6,$9E,$3F,$E4,$C7,$B7,$4F,$FA,$23,$E2,$FC
	.dc.b	$F5,$FE,$8D,$FA,$8F,$F4,$A5,$7F,$5D,$F8,$34,$62,$3F,$CB,$51,$8F
	.dc.b	$84,$FF,$25,$37,$2F,$F3,$2B,$67,$D8,$DC,$FF,$86,$EC,$F9,$1F,$06
	.dc.b	$13,$0F,$F1,$2D,$ED,$F9,$33,$9C,$39,$E3,$E6,$C7,$FD,$28,$DA,$D9
	.dc.b	$C9,$C4,$FF,$B6,$9F,$F2,$BC,$79,$9B,$1E,$F6,$BE,$E6,$BF,$EB,$97
	.dc.b	$73,$E0,$F7,$BF,$D4,$5C,$FF,$25,$FF,$44,$57,$FB,$A7,$EA,$FF,$DD
	.dc.b	$5F,$56,$F7,$3F,$C1,$1E,$2F,$F3,$C8,$C5,$7F,$82,$EB,$75,$AC,$EC
	.dc.b	$E4,$7C,$9F,$07,$FD,$ED,$87,$F4,$68,$C4,$78,$4D,$D7,$36,$9C,$F0
	.dc.b	$DD,$78,$E7,$FD,$3B,$87,$82,$BA,$E1,$5E,$97,$C2,$31,$09,$D5,$F3
	.dc.b	$8F,$46,$1B,$7C,$27,$FA,$8E,$4E,$04,$FA,$F0,$F1,$C3,$0E,$47,$B9
	.dc.b	$F0,$19,$8F,$74,$66,$BC,$78,$79,$9B,$5B,$3E,$15,$70,$3B,$D7,$67
	.dc.b	$C3,$87,$A4,$E9,$3E,$7C,$3C,$CF,$23,$FD,$7C,$FC,$17,$DC,$77,$3E
	.dc.b	$59,$4F,$AD,$1D,$67,$B9,$F5,$DE,$E6,$63,$E5,$C6,$4F,$8B,$B7,$E1
	.dc.b	$23,$F0,$CD,$7C,$2D,$7F,$08,$E7,$E2,$3F,$CF,$7A,$FD,$28,$C7,$BB
	.dc.b	$FC,$F2,$91,$89,$D6,$1F,$ED,$BB,$E3,$5A,$8C,$5F,$17,$F2,$D2,$FA
	.dc.b	$AF,$F7,$14,$61,$B8,$F0,$C3,$F0,$73,$C2,$F9,$4F,$7A,$DE,$1F,$88
	.dc.b	$C3,$A3,$C2,$79,$3F,$8E,$8F,$2E,$7D,$E3,$F3,$91,$C1,$FA,$8E,$A7
	.dc.b	$8E,$1B,$9E,$57,$E0,$55,$F0,$27,$CD,$8F,$4A,$FF,$7F,$61,$C4,$CD
	.dc.b	$F6,$C7,$E6,$46,$71,$F6,$4F,$92,$73,$1E,$A9,$D4,$FB,$EB,$AC,$E4
	.dc.b	$F3,$9C,$A3,$CE,$7C,$33,$E3,$57,$CA,$6D,$F9,$D8,$79,$5F,$11,$8B
	.dc.b	$F0,$C7,$C0,$9F,$2B,$C8,$EE,$4C,$3D,$E7,$C0,$9E,$38,$FC,$F7,$18
	.dc.b	$7F,$2B,$3F,$80,$D7,$F3,$C6,$E1,$FC,$F1,$BA,$FF,$6A,$FB,$4B,$5F
	.dc.b	$CE,$E7,$FD,$64,$F3,$3F,$9E,$38,$74,$7F,$3F,$D7,$8E,$BF,$D7,$1B
	.dc.b	$C3,$FC,$42,$7F,$D1,$3C,$8F,$8E,$CD,$49,$FE,$98,$F8,$BA,$15,$B9
	.dc.b	$C9,$0F,$EA,$F3,$B8,$D4,$60,$F2,$10,$FE,$A5,$1E,$9F,$41,$D4,$F8
	.dc.b	$4F,$83,$18,$6C,$66,$BD,$3F,$41,$89,$E5,$39,$9F,$F0,$DC,$60,$F8
	.dc.b	$D8,$FA,$BE,$0F,$CC,$49,$F8,$7E,$C5,$7F,$C1,$39,$A7,$1C,$5F,$DF
	.dc.b	$2B,$C7,$ED,$4B,$FF,$69,$64,$8D,$F1,$FF,$40,$D7,$DB,$1E,$62,$57
	.dc.b	$F8,$C9,$F2,$F9,$8F,$82,$FA,$1E,$3B,$E5,$E9,$F4,$E7,$57,$FE,$74
	.dc.b	$8C,$B4,$FF,$5E,$56,$E3,$F1,$71,$FE,$7E,$F1,$6F,$A1,$FE,$37,$8F
	.dc.b	$F6,$14,$E2,$64,$7F,$58,$DB,$FD,$3D,$5B,$8F,$F5,$65,$74,$FA,$9F
	.dc.b	$F6,$8C,$6F,$5F,$CF,$C6,$23,$FC,$80,$1C,$5F,$FA,$CB,$CD,$7D,$3F
	.dc.b	$3F,$4F,$F3,$97,$B6,$DF,$38,$FD,$59,$93,$F4,$4F,$4C,$3F,$5F,$C3
	.dc.b	$F4,$E7,$C8,$7F,$72,$70,$F9,$27,$83,$FE,$D1,$D9,$E0,$FD,$2F,$A9
	.dc.b	$93,$FE,$E7,$8A,$FF,$93,$B8,$53,$8D,$7F,$D3,$B5,$F6,$BE,$0B,$ED
	.dc.b	$F8,$73,$0F,$F5,$5C,$65,$FF,$56,$C6,$5E,$91,$F6,$3E,$A3,$F0,$51
	.dc.b	$EA,$0E,$1F,$9A,$DB,$FF,$6A,$7D,$CB,$EA,$79,$71,$FC,$CA,$E1,$7F
	.dc.b	$29,$AF,$F1,$66,$BF,$E1,$87,$A7,$07,$F1,$AF,$A0,$7C,$17,$E3,$1B
	.dc.b	$5E,$5A,$DB,$F2,$FE,$B7,$F3,$1E,$54,$8F,$A7,$E0,$38,$D5,$F9,$D3
	.dc.b	$53,$FE,$F9,$AC,$3C,$25,$EC,$FE,$8F,$8E,$4E,$62,$BF,$F0,$06,$2F
	.dc.b	$07,$FB,$42,$7F,$86,$E5,$CA,$BF,$0B,$DD,$F3,$27,$27,$EB,$FE,$7A
	.dc.b	$CF,$F3,$29,$D3,$6B,$93,$F9,$48,$CB,$63,$37,$FE,$E0,$E3,$FD,$D1
	.dc.b	$D2,$7F,$D2,$F9,$7F,$84,$9F,$77,$27,$EF,$CF,$E0,$1E,$44,$8C,$90
	.dc.b	$E2,$FF,$37,$F9,$DF,$D5,$C7,$CE,$5E,$E6,$4D,$76,$C9,$D4,$65,$E1
	.dc.b	$1F,$CC,$B8,$F9,$C7,$FA,$E7,$54,$63,$47,$2F,$F1,$F8,$C3,$FE,$60
	.dc.b	$1F,$A7,$BE,$43,$A6,$3F,$D0,$6B,$AE,$3F,$D9,$B1,$EB,$39,$24,$7E
	.dc.b	$86,$5E,$4F,$E4,$4E,$71,$FB,$EC,$9F,$50,$FA,$5F,$ED,$71,$86,$C6
	.dc.b	$67,$F8,$2F,$D4,$BC,$3F,$34,$7E,$BF,$E9,$FE,$A3,$FD,$C9,$AF,$F1
	.dc.b	$26,$BF,$DF,$9A,$FF,$A0,$3E,$A7,$F9,$43,$FC,$61,$AF,$F8,$11,$AE
	.dc.b	$FC,$8F,$EA,$FC,$7F,$AC,$8F,$E5,$3E,$BF,$F1,$8D,$3E,$7C,$7F,$A3
	.dc.b	$23,$1C,$7F,$E8,$EE,$F5,$F1,$BD,$DA,$FF,$90,$BF,$CA,$9A,$FF,$85
	.dc.b	$9A,$F5,$F2,$FF,$99,$FD,$47,$C7,$8F,$E7,$6D,$AE,$AD,$C7,$95,$7E
	.dc.b	$27,$29,$7F,$AE,$E2,$FE,$95,$87,$AB,$3F,$3F,$CC,$6A,$BF,$55,$5F
	.dc.b	$13,$97,$F1,$F8,$7E,$1C,$93,$FC,$8F,$3F,$E8,$78,$F8,$A3,$93,$F1
	.dc.b	$AB,$F2,$FE,$9F,$E6,$C9,$3F,$A5,$FB,$9A,$F8,$23,$F2,$84,$7F,$80
	.dc.b	$63,$FE,$1C,$BF,$97,$F5,$1F,$D2,$24,$FF,$4A,$C3,$BC,$7F,$50,$FA
	.dc.b	$37,$D4,$BE,$87,$FC,$7B,$18,$6C,$66,$7F,$71,$F1,$18,$9F,$F1,$F5
	.dc.b	$6E,$7F,$C8,$6D,$7F,$B1,$B5,$FF,$27,$B9,$C7,$CF,$E0,$7F,$80,$B2
	.dc.b	$FC,$DC,$7E,$2A,$75,$C7,$FE,$D6,$7C,$FF,$03,$E4,$26,$1F,$C2,$FE
	.dc.b	$8E,$BE,$0D,$13,$E6,$6C,$79,$9A,$EE,$3F,$C2,$7E,$3F,$F1,$48,$FE
	.dc.b	$B8,$D5,$8F,$96,$B8,$BE,$7F,$A9,$7C,$AF,$CA,$6D,$7C,$F6,$BC,$B3
	.dc.b	$FD,$27,$53,$E6,$38,$5F,$CC,$BD,$DF,$C8,$6B,$F0,$DA,$B5,$47,$6C
	.dc.b	$7F,$5B,$9F,$E7,$32,$FF,$3F,$B6,$F8,$56,$E7,$E7,$74,$FF,$84,$30
	.dc.b	$34,$FF,$09,$9F,$BA,$E1,$FC,$3B,$FE,$21,$BE,$3D,$DF,$84,$8C,$47
	.dc.b	$95,$A7,$F0,$2E,$9F,$E8,$9F,$13,$27,$FA,$24,$7F,$47,$30,$FC,$73
	.dc.b	$C7,$BA,$75,$1F,$D9,$BE,$75,$F4,$7F,$E7,$66,$C6,$63,$FD,$05,$F6
	.dc.b	$79,$64,$64,$FF,$48,$C7,$FA,$5B,$AD,$F3,$D4,$93,$FE,$96,$DE,$BE
	.dc.b	$5E,$9F,$F4,$C4,$E3,$8F,$F8,$57,$C0,$FF,$1E,$7F,$92,$CE,$AB,$FD
	.dc.b	$52,$D7,$FC,$92,$D7,$FD,$6A,$D4,$1F,$F5,$37,$2F,$9C,$91,$FE,$1A
	.dc.b	$FC,$39,$F6,$78,$8F,$36,$7F,$7D,$E4,$DE,$25,$70,$9F,$F1,$E7,$D4
	.dc.b	$7F,$5C,$C7,$FD,$78,$D8,$FF,$60,$B5,$FF,$30,$B5,$FF,$66,$FC,$5F
	.dc.b	$FA,$1F,$9F,$FD,$88,$D8,$FF,$64,$B5,$FF,$34,$B5,$FF,$6A,$CB,$E1
	.dc.b	$3F,$D8,$D1,$D9,$AE,$9A,$FC,$68,$E1,$F0,$FF,$3B,$3F,$F0,$04,$FC
	.dc.b	$23,$FA,$BE,$DF,$53,$C3,$7C,$26,$38,$FE,$7E,$BF,$EC,$D7,$FC,$EE
	.dc.b	$DB,$E5,$18,$AE,$2E,$B0,$EF,$87,$E0,$3E,$BD,$F3,$BF,$D8,$30,$EB
	.dc.b	$18,$BF,$C1,$B6,$3F,$06,$D7,$F0,$6E,$6B,$A5,$79,$E7,$F0,$98,$3E
	.dc.b	$77,$19,$30,$E1,$8F,$83,$0C,$C6,$30,$E3,$78,$9E,$9E,$3B,$A9,$F2
	.dc.b	$58,$6E,$3F,$E5,$6F,$CE,$7F,$BD,$3F,$B3,$9D,$11,$D7,$9D,$72,$FF
	.dc.b	$1C,$5F,$F8,$C6,$7F,$C8,$9F,$29,$D7,$D7,$F6,$74,$FF,$82,$23,$13
	.dc.b	$FE,$5F,$AE,$17,$FE,$5D,$6B,$FE,$37,$6B,$FD,$27,$EE,$6B,$FC,$9D
	.dc.b	$BD,$26,$93,$1F,$99,$19,$7E,$74,$BE,$A7,$FD,$35,$87,$F6,$BB,$F1
	.dc.b	$F2,$7F,$59,$7F,$84,$C7,$F9,$7F,$CB,$E7,$18,$BF,$E3,$9A,$7F,$89
	.dc.b	$27,$3E,$07,$E1,$67,$CC,$DC,$FC,$CD,$BE,$53,$FD,$B1,$E9,$8F,$86
	.dc.b	$7C,$F9,$7F,$93,$F2,$F3,$47,$B3,$E3,$DE,$6F,$DE,$FC,$0D,$F3,$19
	.dc.b	$8C,$36,$FF,$BE,$D7,$06,$CF,$96,$31,$9F,$F8,$01,$B9,$7F,$7B,$6D
	.dc.b	$7F,$1A,$9F,$95,$1D,$5F,$E7,$DA,$21,$E4,$FE,$0B,$C9,$7C,$BF,$A6
	.dc.b	$30,$DC,$3C,$B1,$E3,$23,$83,$FC,$41,$5D,$1F,$F6,$C6,$DF,$61,$FF
	.dc.b	$07,$46,$71,$FF,$46,$7D,$D7,$EA,$E7,$FD,$67,$B7,$E8,$DF,$F5,$AC
	.dc.b	$66,$7F,$D2,$53,$97,$8F,$5F,$DB,$EA,$56,$E7,$FD,$18,$D7,$F9,$EC
	.dc.b	$6A,$78,$4F,$5D,$BF,$C9,$0D,$C3,$E5,$7C,$9B,$E3,$BF,$6F,$39,$C5
	.dc.b	$8F,$83,$E6,$BE,$57,$CB,$E9,$B8,$FA,$9E,$E4,$7F,$D9,$96,$8C,$5F
	.dc.b	$C4,$7F,$47,$8B,$E2,$FA,$5F,$FB,$63,$C9,$F1,$64,$7C,$9A,$3E,$E7
	.dc.b	$FD,$EF,$5B,$DB,$F8,$D4,$E7,$E3,$BF,$44,$D9,$FC,$46,$9F,$D8,$A3
	.dc.b	$0F,$F9,$BA,$F9,$BA,$1F,$51,$FB,$67,$0D,$79,$B5,$FC,$A3,$5F,$F0
	.dc.b	$7F,$91,$F9,$13,$F9,$77,$D3,$BF,$2F,$7C,$F6,$72,$FC,$1F,$C8,$FD
	.dc.b	$82,$3F,$29,$FD,$E7,$E2,$CF,$C5,$FF,$37,$3E,$B3,$F4,$9F,$8C,$3E
	.dc.b	$8B,$F9,$E5,$6F,$BB,$F8,$8C,$E6,$B8,$6B,$FE,$04,$F4,$9F,$71,$FC
	.dc.b	$5B,$C8,$7B,$5F,$C6,$27,$37,$C2,$FF,$C7,$B3,$9B,$FE,$21,$0A,$DB
	.dc.b	$F2,$DA,$FD,$26,$BF,$BB,$77,$C9,$FE,$BC,$6D,$E7,$DB,$EC,$FA,$31
	.dc.b	$FF,$48,$7D,$B7,$F3,$EF,$7C,$DC,$70,$CA,$E1,$3A,$6C,$E4,$FF,$23
	.dc.b	$75,$BE,$F3,$F0,$79,$FF,$AE,$FC,$7F,$F7,$AD,$FC,$8A,$D7,$5F,$F0
	.dc.b	$DF,$90,$FE,$BF,$87,$0A,$FD,$9B,$71,$FF,$31,$FC,$EF,$F7,$E7,$BC
	.dc.b	$67,$0F,$07,$C4,$7F,$17,$F7,$79,$B7,$0F,$05,$23,$F8,$23,$63,$F3
	.dc.b	$15,$C9,$B1,$9B,$FD,$0C,$E5,$F0,$D7,$F6,$6C,$3F,$BD,$4E,$74,$FF
	.dc.b	$6D,$6A,$FB,$4F,$9C,$7F,$7C,$EA,$7C,$AE,$34,$F9,$B3,$73,$F9,$75
	.dc.b	$A9,$C9,$EA,$F5,$BF,$9A,$78,$DF,$21,$B1,$92,$57,$F5,$ED,$3F,$81
	.dc.b	$EC,$AF,$ED,$7F,$54,$FA,$8F,$ED,$BF,$83,$7F,$97,$71,$FC,$B4,$E5
	.dc.b	$FC,$C4,$67,$1F,$F4,$20,$9D,$3F,$C0,$5A,$FF,$08,$6B,$FD,$45,$AF
	.dc.b	$F9,$B1,$E3,$BA,$B5,$7E,$6F,$9D,$F3,$BF,$86,$F3,$7F,$AA,$76,$DA
	.dc.b	$FC,$E7,$47,$EC,$1F,$BF,$39,$CB,$E1,$7D,$5D,$CB,$B4,$7F,$9F,$3E
	.dc.b	$DF,$F4,$ED,$EF,$FF,$07,$F4,$7E,$3E,$B7,$2A,$DB,$6B,$FC,$2B,$18
	.dc.b	$AE,$18,$FF,$6F,$8D,$FD,$63,$E6,$FF,$C5,$BE,$07,$FB,$46,$BF,$82
	.dc.b	$4E,$5F,$DB,$46,$71,$F9,$F3,$97,$F9,$73,$5F,$E6,$AD,$7F,$BF,$35
	.dc.b	$FF,$4C,$BC,$72,$F5,$B6,$BF,$C0,$9F,$58,$EB,$FF,$17,$CE,$6F,$FD
	.dc.b	$A7,$1F,$EB,$EA,$FE,$07,$87,$E6,$F9,$7F,$85,$32,$32,$7E,$FF,$A7
	.dc.b	$F2,$0E,$2F,$FC,$63,$7F,$E5,$A6,$FC,$0F,$E9,$78,$7F,$8A,$32,$BF
	.dc.b	$88,$B9,$69,$FC,$2F,$F2,$3F,$F1,$4B,$FB,$A6,$D7,$EE,$1A,$FF,$88
	.dc.b	$7B,$91,$88,$FF,$00,$BC,$36,$FC,$2C,$EB,$DC,$FD,$C7,$83,$FE,$74
	.dc.b	$AD,$61,$FE,$C4,$F1,$3F,$D7,$3F,$4B,$A8,$FF,$61,$37,$2F,$F2,$7F
	.dc.b	$C5,$FF,$63,$CF,$AE,$FF,$C9,$27,$35,$B9,$F0,$E3,$F1,$67,$D0,$D7
	.dc.b	$CF,$AF,$F8,$76,$31,$1F,$E1,$07,$E6,$4F,$29,$FF,$45,$FD,$F7,$FB
	.dc.b	$53,$2F,$91,$1F,$E9,$6F,$B6,$7D,$77,$F9,$87,$E2,$D1,$8A,$FF,$73
	.dc.b	$B7,$2F,$F7,$3B,$5F,$F7,$3B,$9C,$7F,$DD,$ED,$9F,$F5,$DF,$B6,$F9
	.dc.b	$0F,$06,$32,$7F,$D7,$B5,$BC,$3F,$DD,$D7,$B9,$CD,$6E,$7F,$E0,$05
	.dc.b	$6F,$1F,$F1,$EE,$19,$C7,$C3,$E6,$FA,$EB,$FB,$66,$3E,$A8,$C1,$1F
	.dc.b	$0B,$FE,$B8,$F1,$BF,$D6,$98,$7C,$28,$FF,$48,$E2,$1D,$77,$EB,$FE
	.dc.b	$4C,$7F,$73,$C3,$1A,$7F,$88,$EA,$CE,$45,$70,$7F,$62,$3E,$93,$0F
	.dc.b	$A0,$C5,$73,$69,$CE,$36,$F9,$0E,$B1,$FE,$07,$BF,$0C,$76,$AE,$B8
	.dc.b	$7F,$22,$78,$E1,$A9,$F0,$46,$1E,$67,$A5,$C6,$1C,$AF,$E2,$38,$C3
	.dc.b	$C1,$5E,$63,$C0,$E2,$BA,$1C,$63,$A9,$A3,$B4,$7B,$E3,$47,$07,$91
	.dc.b	$92,$C7,$F9,$5A,$FC,$71,$E2,$AF,$24,$FC,$78,$E0,$F6,$38,$1E,$23
	.dc.b	$E0,$BF,$3E,$3C,$C7,$28,$FF,$20,$5F,$67,$8C,$6A,$78,$1E,$13,$EC
	.dc.b	$63,$F7,$67,$BB,$5F,$C0,$B5,$FC,$75,$FE,$3E,$35,$EE,$7F,$A3,$F5
	.dc.b	$C6,$1F,$EB,$B8,$C5,$AB,$7E,$FA,$F7,$3F,$99,$AD,$D6,$A7,$D5,$3F
	.dc.b	$E8,$7F,$7B,$FD,$8D,$87,$E5,$6F,$FD,$1D,$3A,$7B,$C6,$70,$ED,$3F
	.dc.b	$7E,$76,$DA,$E7,$97,$9E,$B7,$7C,$70,$FC,$AE,$1C,$34,$F0,$61,$FE
	.dc.b	$D7,$75,$8F,$E1,$A7,$C3,$8F,$C1,$AE,$B3,$ED,$6B,$FE,$A1,$C7,$BC
	.dc.b	$F8,$70,$EF,$5D,$63,$95,$78,$AB,$D7,$87,$80,$F1,$C7,$03,$57,$1F
	.dc.b	$25,$F3,$7C,$46,$67,$D3,$8F,$28,$F8,$87,$CB,$BF,$2B,$F5,$5F,$0E
	.dc.b	$1E,$48,$EA,$73,$7D,$4E,$09,$5E,$78,$CD,$77,$23,$83,$8C,$78,$C8
	.dc.b	$E0,$FC,$87,$92,$F9,$0F,$E4,$17,$F8,$BD,$B3,$18,$CF,$FC,$57,$AA
	.dc.b	$F8,$5F,$FA,$9A,$3F,$D0,$9F,$65,$BC,$38,$CF,$FB,$1E,$78,$61,$FE
	.dc.b	$C9,$8F,$CD,$5F,$27,$FD,$A3,$3A,$78,$43,$EC,$BA,$47,$06,$CF,$8D
	.dc.b	$BC,$4F,$B0,$CD,$F1,$BF,$33,$5F,$71,$CC,$F5,$4F,$17,$75,$C8,$8E
	.dc.b	$AA,$BE,$8A,$E8,$FC,$AB,$F2,$3E,$AA,$F8,$1E,$07,$F8,$53,$E1,$7D
	.dc.b	$6D,$9F,$F0,$F4,$63,$A9,$D9,$AE,$C4,$F2,$6D,$7F,$88,$E3,$07,$CD
	.dc.b	$EF,$77,$BB,$BA,$7B,$B2,$FF,$D8,$D3,$96,$CE,$5F,$F5,$FE,$5F,$EF
	.dc.b	$27,$27,$F8,$CF,$E8,$7C,$84,$7F,$D8,$11,$B7,$FD,$8F,$92,$7D,$04
	.dc.b	$8F,$F1,$D3,$E7,$F1,$5F,$21,$F7,$B3,$FF,$2B,$36,$33,$5D,$BC,$7C
	.dc.b	$4C,$8F,$F6,$CB,$F2,$63,$86,$B2,$7F,$DB,$AF,$CD,$30,$FF,$95,$9B
	.dc.b	$D1,$FE,$44,$CF,$E6,$CF,$FA,$8B,$F2,$1F,$8C,$DD,$7F,$26,$A3,$35
	.dc.b	$FE,$8A,$FB,$07,$9F,$A9,$E3,$5F,$AE,$38,$5F,$E8,$DA,$7E,$96,$36
	.dc.b	$FE,$D1,$EB,$D5,$FE,$53,$F8,$87,$D5,$FC,$76,$D7,$F8,$67,$2F,$F4
	.dc.b	$DD,$F0,$8F,$F0,$E7,$DF,$FF,$84,$E7,$FD,$3E,$DC,$3F,$80,$B6,$77
	.dc.b	$AF,$BA,$31,$06,$92,$3F,$37,$E5,$FE,$A6,$7F,$5B,$D0,$8C,$61,$FE
	.dc.b	$70,$6E,$1F,$E6,$DF,$67,$FA,$5C,$E7,$B7,$F0,$5E,$AF,$FA,$D2,$3F
	.dc.b	$D7,$7C,$C8,$C5,$FF,$8D,$F1,$ED,$7F,$53,$EB,$5F,$09,$49,$A7,$C5
	.dc.b	$BF,$E3,$FC,$9F,$95,$8C,$1F,$E4,$1A,$E2,$FF,$17,$6B,$F6,$43,$87
	.dc.b	$EE,$35,$FE,$56,$F0,$CF,$FD,$9B,$F6,$1F,$84,$9E,$E1,$C3,$FE,$F0
	.dc.b	$9E,$17,$ED,$DB,$E2,$B7,$C9,$F9,$E7,$E9,$F7,$7D,$7E,$30,$F5,$F0
	.dc.b	$7F,$AF,$FE,$77,$F5,$46,$63,$FD,$6B,$F2,$FF,$AD,$23,$F0,$B8,$5F
	.dc.b	$7E,$F9,$BC,$CF,$03,$AF,$7A,$FF,$8D,$D6,$E7,$96,$DF,$E2,$1E,$1F
	.dc.b	$C0,$79,$BF,$86,$3D,$1F,$47,$6B,$E6,$67,$F8,$79,$FE,$4F,$7F,$E2
	.dc.b	$79,$F8,$51,$83,$F8,$27,$E0,$12,$3B,$C7,$F2,$CF,$BE,$FF,$79,$47
	.dc.b	$AD,$AF,$E5,$1B,$8F,$E3,$38,$7C,$DC,$7F,$8B,$9C,$BF,$EC,$1D,$BF
	.dc.b	$99,$DF,$C0,$F2,$3B,$7B,$BF,$C5,$7D,$09,$3F,$11,$FE,$35,$F9,$8B
	.dc.b	$E2,$33,$3F,$EE,$49,$BA,$7E,$72,$B5,$3F,$84,$FA,$1F,$F7,$B3,$72
	.dc.b	$FE,$1F,$18,$8F,$EB,$21,$7C,$5E,$A7,$F8,$64,$6E,$7F,$88,$35,$FE
	.dc.b	$FE,$D7,$FD,$40,$F5,$B6,$9A,$F7,$FC,$27,$5A,$FC,$CF,$CF,$3E,$5F
	.dc.b	$F9,$1F,$C6,$BE,$26,$9F,$C9,$A7,$55,$C3,$AF,$F9,$04,$7E,$1E,$3F
	.dc.b	$58,$D7,$F9,$33,$5F,$A5,$F6,$3F,$D0,$C8,$FE,$E3,$E5,$5F,$64,$F9
	.dc.b	$7F,$E6,$BF,$5E,$F9,$DF,$E5,$CD,$BF,$F0,$93,$5F,$DE,$35,$FE,$6D
	.dc.b	$F6,$1F,$CB,$1F,$E0,$0D,$73,$E8,$FF,$39,$8C,$47,$F3,$1F,$C0,$7D
	.dc.b	$9C,$7F,$2A,$3F,$CE,$D1,$8E,$3D,$6F,$FD,$2A,$FB,$5E,$FC,$17,$6B
	.dc.b	$87,$F3,$84,$CB,$F8,$4F,$D8,$3E,$1F,$11,$FD,$36,$F8,$D7,$F8,$42
	.dc.b	$F9,$4F,$EB,$9A,$ED,$B8,$7F,$2E,$E9,$FE,$26,$DF,$5B,$AB,$E5,$97
	.dc.b	$3F,$EB,$C9,$7E,$42,$9D,$5F,$11,$1F,$13,$FE,$22,$FB,$8F,$E8,$70
	.dc.b	$C9,$18,$7E,$1C,$FF,$90,$BE,$89,$F6,$D2,$FF,$9C,$C6,$23,$46,$63
	.dc.b	$F4,$FF,$63,$FB,$E5,$3B,$7F,$9E,$35,$FE,$EC,$D7,$FC,$CA,$D4,$72
	.dc.b	$9E,$D9,$7C,$9F,$93,$FF,$0C,$1E,$28,$F1,$FC,$2C,$AF,$3C,$7F,$9A
	.dc.b	$DF,$E8,$78,$7F,$AF,$5C,$CF,$F9,$9F,$0F,$7B,$6B,$84,$F9,$9A,$F9
	.dc.b	$9C,$E1,$F8,$BC,$FA,$4F,$F7,$46,$C7,$E5,$6F,$AC,$FC,$AF,$C0,$5F
	.dc.b	$25,$F3,$DB,$19,$8F,$F1,$0F,$9A,$FA,$ED,$47,$9E,$B7,$5F,$29,$AF
	.dc.b	$CA,$6B,$F2,$AD,$1C,$30,$FC,$95,$7F,$90,$4E,$B5,$E5,$F9,$CF,$21
	.dc.b	$1F,$ED,$78,$FE,$EE,$BD,$96,$AF,$21,$2B,$EF,$C6,$5F,$EA,$3F,$57
	.dc.b	$5F,$D5,$35,$FF,$52,$F0,$A3,$11,$FE,$11,$6B,$F9,$97,$95,$FE,$AF
	.dc.b	$E2,$3F,$CB,$47,$FA,$02,$33,$D7,$C7,$E1,$30,$4B,$FE,$E5,$F9,$4F
	.dc.b	$AE,$75,$1F,$DD,$69,$5B,$8F,$F1,$7D,$6D,$7C,$5F,$F8,$5B,$E5,$3F
	.dc.b	$D4,$FE,$E1,$A7,$FA,$AB,$A7,$F9,$C5,$7F,$AE,$E7,$4D,$9D,$BF,$EB
	.dc.b	$F9,$DB,$FE,$AA,$AE,$CF,$FA,$FE,$B8,$BF,$EB,$C6,$BF,$EA,$56,$BF
	.dc.b	$AA,$8F,$F1,$5B,$E6,$21,$DF,$E3,$EB,$DE,$7F,$CD,$5A,$FF,$B5,$1B
	.dc.b	$A7,$3C,$17,$FE,$76,$FA,$EF,$F2,$96,$3F,$ED,$16,$E3,$FE,$3B,$6B
	.dc.b	$FE,$75,$6B,$FE,$E4,$F9,$3F,$E1,$FC,$BF,$ED,$96,$C7,$F9,$0D,$AF
	.dc.b	$F9,$F5,$AF,$FB,$B2,$7F,$93,$F8,$1E,$1E,$DF,$E7,$13,$E1,$6B,$E1
	.dc.b	$6B,$F1,$7C,$FF,$F7,$B6,$9F,$EE,$F6,$C7,$D1,$C3,$F0,$73,$9F,$91
	.dc.b	$FE,$A1,$BF,$FB,$E2,$F6,$D8,$CC,$FD,$1A,$E3,$7F,$6E,$F7,$3E,$1F
	.dc.b	$3F,$FA,$D4,$FF,$B7,$67,$FC,$01,$3F,$EF,$D6,$C7,$FB,$E9,$AE,$E7
	.dc.b	$BB,$F9,$48,$DE,$1F,$E7,$CF,$6F,$FA,$D5,$FB,$6F,$7B,$7F,$87,$5B
	.dc.b	$7C,$27,$F9,$8F,$BD,$FE,$56,$D7,$35,$DB,$A3,$F6,$DF,$19,$7A,$3F
	.dc.b	$D9,$B8,$7F,$B8,$3D,$CC,$4C,$8F,$D6,$3E,$6C,$7F,$DD,$33,$9D,$9B
	.dc.b	$7F,$65,$FA,$BE,$DF,$5B,$FE,$61,$C7,$FC,$C7,$18,$F8,$0E,$5E,$4A
	.dc.b	$31,$3A,$3F,$CD,$D5,$C7,$0F,$F1,$DF,$D5,$FF,$C0,$09,$FF,$47,$42
	.dc.b	$30,$FF,$A4,$5A,$FF,$B2,$DA,$FF,$54,$38,$7B,$DF,$41,$B3,$FE,$90
	.dc.b	$FC,$05,$79,$BE,$B9,$FD,$E5,$EC,$7D,$8E,$B7,$D7,$7A,$F1,$FF,$4C
	.dc.b	$D6,$9F,$EE,$31,$9C,$BD,$77,$F8,$87,$C9,$5B,$7D,$51,$97,$FB,$5C
	.dc.b	$65,$F6,$47,$94,$F5,$65,$FC,$66,$FC,$31,$E9,$F8,$CF,$30,$FA,$EC
	.dc.b	$4F,$F8,$12,$30,$FA,$1F,$05,$FC,$D6,$E0,$69,$27,$81,$FE,$B8,$FC
	.dc.b	$05,$F6,$9D,$AF,$EC,$F1,$78,$BD,$AF,$98,$DB,$F8,$EF,$62,$3F,$7B
	.dc.b	$1F,$C1,$57,$DB,$9E,$33,$DF,$E8,$2C,$FF,$A0,$3E,$9D,$F3,$BF,$9F
	.dc.b	$FA,$0F,$D2,$4F,$F8,$52,$73,$7F,$EB,$B9,$CC,$A7,$2F,$FB,$6D,$AF
	.dc.b	$E2,$4D,$57,$C6,$8C,$C7,$E3,$5E,$3C,$7F,$C1,$1B,$7F,$EF,$3F,$A0
	.dc.b	$7C,$9F,$E4,$67,$37,$E5,$F9,$0F,$47,$CB,$E5,$B8,$7F,$B9,$29,$1E
	.dc.b	$12,$47,$E9,$23,$2D,$8E,$4F,$FB,$53,$E6,$AD,$FE,$E2,$3F,$C6,$F9
	.dc.b	$3E,$BF,$FD,$4F,$1F,$F0,$02,$31,$F1,$5F,$A7,$FA,$FF,$D1,$CF,$0A
	.dc.b	$DD,$FF,$AC,$27,$F5,$EF,$F9,$09,$A9,$A6,$BC,$46,$CF,$85,$ED,$DF
	.dc.b	$FC,$5E,$31,$8F,$98,$FD,$D5,$F8,$B0,$7C,$4B,$EB,$DE,$BF,$F2,$C8
	.dc.b	$C3,$F8,$3F,$A4,$C4,$EB,$0E,$97,$FE,$58,$FA,$C7,$CD,$7E,$89,$B3
	.dc.b	$FC,$41,$B3,$F8,$30,$E1,$FE,$1D,$F5,$CF,$AF,$7D,$7F,$F3,$E8,$C3
	.dc.b	$FC,$8B,$DB,$C4,$EB,$0F,$1D,$FF,$A6,$7D,$DF,$E2,$DE,$93,$E2,$2C
	.dc.b	$61,$B5,$FC,$92,$7F,$AB,$D7,$E5,$DA,$FF,$21,$6B,$FE,$76,$7F,$AE
	.dc.b	$F4,$7F,$5A,$6C,$FF,$27,$FA,$D7,$D9,$F6,$BF,$C4,$0F,$4B,$FD,$E7
	.dc.b	$5F,$29,$7A,$2F,$B0,$FE,$7F,$5F,$D0,$1B,$5B,$9D,$36,$74,$F3,$F9
	.dc.b	$AF,$ED,$F1,$FE,$A1,$FA,$9F,$C7,$E9,$FE,$C9,$C5,$5A,$ED,$56,$B0
	.dc.b	$FE,$C7,$39,$8C,$61,$E9,$8C,$77,$FF,$5D,$5F,$23,$FD,$9E,$9F,$5B
	.dc.b	$98,$C3,$70,$FC,$4E,$3E,$08,$C6,$3F,$1A,$31,$3E,$BA,$E2,$D3,$8D
	.dc.b	$F6,$79,$F5,$3E,$6F,$FB,$3F,$C4,$7F,$79,$8C,$4F,$C9,$8C,$4C,$8F
	.dc.b	$91,$F3,$17,$2F,$AF,$B7,$C6,$F7,$7D,$9E,$6F,$BD,$B7,$F3,$DB,$5C
	.dc.b	$37,$FE,$53,$B7,$CC,$BF,$ED,$1C,$AF,$8F,$B9,$E1,$B9,$7E,$6A,$74
	.dc.b	$FF,$78,$6E,$3F,$A9,$6B,$FC,$B5,$AF,$E4,$67,$A4,$FE,$93,$EB,$7E
	.dc.b	$0C,$FE,$48,$57,$4E,$4F,$C2,$66,$FA,$EF,$B5,$F3,$7F,$75,$B8,$FE
	.dc.b	$35,$B8,$7B,$DB,$7F,$E3,$CE,$5F,$C8,$CB,$E5,$3F,$65,$7F,$B6,$9C
	.dc.b	$C6,$27,$C9,$1C,$09,$AF,$F8,$3D,$E7,$F5,$DF,$A6,$BF,$67,$CF,$FF
	.dc.b	$84,$E3,$17,$ED,$D9,$39,$EA,$FE,$EB,$5A,$AF,$EA,$8D,$7F,$C3,$CD
	.dc.b	$7F,$DE,$6F,$4C,$FF,$CA,$ED,$C3,$FC,$8F,$FB,$93,$E8,$FB,$47,$E4
	.dc.b	$B8,$FF,$C6,$FE,$0F,$E4,$5B,$19,$26,$3F,$87,$F1,$F5,$1F,$9A,$E2
	.dc.b	$FF,$1F,$FD,$57,$E6,$5B,$19,$25,$7E,$3F,$D4,$D4,$7E,$8B,$E9,$3F
	.dc.b	$C4,$5D,$2E,$E4,$61,$B3,$FC,$CE,$75,$39,$3F,$50,$7F,$96,$7E,$B7
	.dc.b	$FA,$44,$6A,$7F,$8C,$47,$F4,$BE,$D3,$4F,$F1,$A7,$17,$F0,$F7,$1E
	.dc.b	$8F,$33,$A7,$C0,$DB,$FE,$07,$F5,$9F,$2D,$F2,$37,$5F,$F0,$56,$48
	.dc.b	$C4,$7F,$BA,$9A,$84,$6F,$8F,$FC,$F5,$3C,$F2,$F2,$79,$38,$9F,$F4
	.dc.b	$2F,$B7,$FE,$43,$C3,$FC,$31,$F1,$88,$C4,$7F,$98,$E3,$1E,$6F,$F9
	.dc.b	$3D,$B9,$FF,$5D,$6C,$FF,$B6,$EB,$A6,$3F,$0F,$BF,$BE,$13,$0F,$ED
	.dc.b	$78,$70,$AD,$46,$2F,$8B,$FE,$DD,$6E,$1F,$EF,$46,$F0,$7D,$2C,$7A
	.dc.b	$4F,$FB,$AF,$C7,$FE,$6B,$7E,$56,$C7,$B9,$AF,$B5,$AF,$FB,$AD,$DE
	.dc.b	$09,$C7,$91,$F8,$5F,$98,$FF,$56,$D6,$A7,$FB,$F7,$6F,$D3,$FA,$F7
	.dc.b	$D3,$7F,$6D,$9C,$3A,$C9,$3A,$7F,$B8,$4E,$9F,$DB,$8F,$A7,$D3,$FD
	.dc.b	$CE,$3F,$01,$18,$8F,$DC,$9B,$AF,$03,$4E,$75,$BF,$43,$D9,$1E,$29
	.dc.b	$F2,$4E,$6F,$C5,$3F,$C1,$8D,$1E,$62,$99,$8E,$0E,$0F,$1D,$E0,$F3
	.dc.b	$2C,$3D,$37,$E6,$9C,$C7,$82,$7F,$DB,$67,$33,$81,$D7,$0F,$4D,$F0
	.dc.b	$8C,$61,$C5,$DC,$79,$0D,$17,$0F,$79,$F1,$9E,$58,$79,$E3,$EB,$36
	.dc.b	$7E,$4C,$6C,$E2,$FC,$39,$E3,$1E,$B3,$89,$63,$ED,$CF,$37,$E6,$35
	.dc.b	$FB,$3C,$1F,$7D,$FC,$0B,$5F,$A4,$D7,$71,$FE,$D7,$F9,$2F,$F4,$DE
	.dc.b	$DF,$C0,$EB,$84,$62,$91,$8F,$81,$FF,$79,$C6,$2F,$94,$62,$75,$7F
	.dc.b	$5A,$BF,$87,$5F,$FA,$31,$C0,$70,$FE,$3C,$C5,$76,$BF,$CC,$CF,$8D
	.dc.b	$BE,$FF,$FB,$BD,$B5,$D2,$75,$8F,$93,$0F,$C9,$C7,$6C,$B7,$5F,$92
	.dc.b	$8F,$74,$76,$C7,$C1,$7F,$17,$1E,$CD,$BE,$D3,$A7,$DA,$FC,$4A,$EA
	.dc.b	$49,$F1,$4F,$79,$D4,$73,$8F,$15,$7F,$33,$75,$87,$A2,$B7,$1F,$2A
	.dc.b	$75,$7D,$63,$C3,$5B,$3E,$14,$FA,$23,$81,$E8,$88,$F8,$0F,$8B,$1D
	.dc.b	$2F,$CF,$5F,$0E,$B3,$50,$F2,$93,$1F,$96,$4A,$F7,$BC,$63,$B1,$2B
	.dc.b	$07,$F1,$9C,$BF,$CD,$7F,$2D,$F0,$70,$FE,$1B,$18,$AF,$E7,$FB,$AB
	.dc.b	$5A,$FF,$59,$E3,$FE,$D7,$ED,$62,$FF,$2F,$7B,$7F,$9D,$5F,$FA,$AD
	.dc.b	$CB,$70,$ED,$1F,$CF,$7E,$0B,$94,$7F,$9F,$9B,$5F,$EE,$C9,$FE,$F7
	.dc.b	$EA,$E2,$64,$79,$9F,$16,$3F,$DF,$A7,$33,$FD,$36,$6E,$7F,$DF,$E7
	.dc.b	$33,$FD,$2E,$7C,$97,$DE,$3C,$87,$F7,$EE,$EF,$77,$6F,$F7,$D6,$F6
	.dc.b	$9F,$0F,$7B,$D4,$CC,$7C,$1A,$E1,$7C,$C8,$7E,$44,$FF,$58,$53,$C5
	.dc.b	$E6,$48,$FF,$0D,$F9,$15,$FF,$40,$CE,$DF,$F0,$FC,$F8,$A7,$CB,$C3
	.dc.b	$27,$FC,$32,$F1,$DB,$FA,$5E,$1F,$E1,$78,$F3,$67,$FE,$EE,$AF,$F0
	.dc.b	$EC,$60,$CD,$FC,$02,$47,$FA,$EF,$5A,$FF,$A5,$3C,$A7,$C8,$7D,$46
	.dc.b	$C7,$F8,$AA,$BE,$4F,$C4,$C8,$FF,$16,$BF,$6B,$D2,$7C,$F6,$67,$A4
	.dc.b	$F3,$31,$AF,$F6,$48,$7D,$57,$FB,$AD,$FF,$77,$B5,$FF,$46,$36,$3F
	.dc.b	$D9,$DF,$49,$C5,$E5,$1B,$9F,$A8,$D7,$F8,$33,$5F,$F1,$D3,$5F,$0F
	.dc.b	$91,$FA,$35,$F5,$BF,$55,$FF,$4E,$36,$BB,$CE,$FE,$B5,$8D,$25,$FF
	.dc.b	$91,$EB,$DD,$F3,$2F,$8D,$CC,$FF,$A9,$EF,$FC,$1F,$5F,$0D,$BD,$DF
	.dc.b	$C7,$5E,$B3,$FE,$F1,$F0,$BF,$37,$E7,$7E,$7E,$FF,$06,$DB,$FF,$76
	.dc.b	$CF,$D3,$37,$C9,$2F,$91,$F3,$BE,$2B,$E0,$6E,$69,$F0,$FE,$F1,$F0
	.dc.b	$38,$AF,$E0,$95,$B6,$D7,$F0,$77,$ED,$B6,$35,$1F,$EC,$5F,$9E,$7A
	.dc.b	$DF,$3D,$B1,$E3,$3E,$46,$7F,$C5,$F9,$9F,$59,$F0,$35,$95,$A1,$E2
	.dc.b	$2B,$71,$98,$C5,$70,$87,$07,$C8,$BF,$CC,$4F,$F0,$6F,$9C,$7A,$14
	.dc.b	$97,$CF,$1F,$F5,$03,$E1,$F0,$15,$C2,$3C,$1D,$AF,$90,$FD,$51,$CA
	.dc.b	$3F,$DF,$5F,$45,$F6,$BC,$C9,$1D,$63,$F5,$46,$1F,$3E,$FF,$E9,$1C
	.dc.b	$7F,$D5,$F5,$A9,$E0,$78,$F9,$FF,$C3,$5B,$FE,$12,$75,$3C,$8F,$8F
	.dc.b	$A6,$B0,$E9,$F4,$1F,$ED,$C9,$F6,$35,$E1,$C3,$F3,$6F,$F7,$C7,$F2
	.dc.b	$C9,$D6,$1F,$C5,$1B,$7F,$66,$7F,$9E,$FD,$63,$E1,$FF,$86,$7A,$13
	.dc.b	$2F,$AD,$39,$6C,$E5,$FE,$29,$3C,$A3,$F6,$DF,$3E,$F8,$9A,$4A,$FC
	.dc.b	$20,$E7,$FE,$26,$E2,$78,$3F,$D7,$5A,$FF,$3B,$0E,$1F,$E7,$CD,$7F
	.dc.b	$C2,$0F,$0D,$3F,$8D,$2E,$3F,$29,$E3,$9F,$A0,$3E,$9D,$F1,$5F,$C3
	.dc.b	$DB,$1E,$86,$BF,$43,$4F,$96,$FF,$B4,$7E,$85,$F1,$1F,$AF,$6C,$7F
	.dc.b	$77,$6B,$EB,$6B,$FE,$27,$F9,$17,$D6,$3E,$B3,$FC,$05,$18,$6C,$66
	.dc.b	$7C,$9D,$9F,$C9,$7E,$AF,$FC,$BB,$1F,$3B,$EB,$3F,$8E,$37,$C8,$D7
	.dc.b	$B3,$FC,$AE,$73,$3F,$35,$AF,$F6,$96,$BF,$E3,$B6,$BF,$EE,$FA,$7C
	.dc.b	$C7,$84,$91,$DC,$FB,$FF,$58,$F8,$0F,$E7,$4D,$8C,$C7,$E4,$B9,$F2
	.dc.b	$FF,$17,$9D,$4F,$E0,$5A,$FF,$7D,$6B,$FE,$4F,$77,$CD,$F9,$FE,$3F
	.dc.b	$D1,$1D,$27,$C6,$78,$88,$7F,$07,$E6,$F8,$73,$FD,$86,$BF,$94,$49
	.dc.b	$A4,$7C,$C7,$EB,$F8,$A3,$D8,$FC,$FF,$D6,$7E,$ED,$07,$0D,$EB,$FF
	.dc.b	$20,$7D,$1B,$ED,$DF,$71,$FD,$86,$FF,$D4,$17,$FE,$6D,$6B,$D1,$AF
	.dc.b	$33,$FC,$8C,$F8,$6B,$FC,$03,$EB,$76,$3F,$63,$3F,$A2,$9F,$19,$F8
	.dc.b	$A7,$1C,$7E,$13,$CB,$87,$EA,$E3,$DB,$1F,$D3,$67,$84,$7F,$8C,$5A
	.dc.b	$F9,$E3,$47,$87,$E5,$A4,$72,$3F,$82,$F8,$EF,$AD,$FF,$00,$36,$33
	.dc.b	$1F,$1B,$EE,$1F,$51,$A8,$FE,$BD,$5B,$AF,$E1,$0D,$7F,$C8,$8D,$7F
	.dc.b	$D2,$CE,$F8,$5E,$2F,$D6,$D9,$F5,$CF,$8B,$FA,$AD,$C3,$C6,$7A,$3B
	.dc.b	$3C,$E4,$7F,$B1,$53,$47,$C7,$66,$7C,$B5,$E8,$23,$FC,$7F,$CA,$78
	.dc.b	$9F,$82,$8F,$EB,$9F,$59,$FE,$63,$6F,$C3,$7F,$3A,$E7,$91,$82,$1F
	.dc.b	$C5,$DF,$F3,$7F,$D7,$7E,$7B,$7F,$EB,$0D,$97,$A0,$8C,$64,$F8,$6F
	.dc.b	$F2,$7D,$7F,$57,$AD,$69,$D0,$91,$FD,$F6,$77,$C5,$FE,$A2,$F3,$FF
	.dc.b	$90,$B7,$CA,$FF,$4A,$C6,$33,$FF,$51,$66,$E5,$CD,$4C,$7F,$C1,$9D
	.dc.b	$2C,$38,$47,$F8,$3A,$F8,$3A,$8F,$E5,$AF,$EC,$27,$85,$7F,$A9,$E7
	.dc.b	$32,$9C,$BF,$D3,$DE,$DF,$15,$E6,$24,$78,$BC,$A7,$D6,$62,$3F,$BD
	.dc.b	$57,$07,$C5,$5F,$EB,$2A,$DC,$7F,$62,$FA,$E7,$A9,$FA,$68,$46,$2F
	.dc.b	$FD,$7C,$D8,$FE,$C4,$D7,$FD,$3D,$F1,$FF,$E7,$CE,$AF,$F6,$23,$63
	.dc.b	$FB,$43,$5D,$F8,$49,$EB,$E2,$7F,$2A,$9F,$0B,$6C,$D3,$F9,$7E,$0C
	.dc.b	$1F,$0D,$FA,$8B,$F8,$3E,$3B,$E0,$B1,$CB,$FE,$E2,$BD,$F2,$FD,$AA
	.dc.b	$D3,$F7,$23,$37,$CA,$F8,$CF,$D9,$9C,$DF,$FB,$B2,$F8,$DF,$83,$D2
	.dc.b	$DB,$F6,$E3,$FB,$DD,$FF,$93,$FD,$9F,$EA,$4F,$86,$B7,$D7,$FE,$46
	.dc.b	$F3,$38,$EB,$FE,$0E,$9C,$DF,$0C,$7B,$61,$CE,$F7,$8F,$69,$FF,$19
	.dc.b	$D6,$F2,$F3,$61,$DB,$A9,$F5,$DE,$6E,$D7,$81,$FE,$BC,$D3,$F5,$13
	.dc.b	$98,$C6,$5F,$ED,$48,$C7,$C4,$73,$6D,$7F,$97,$3A,$BD,$7E,$B7,$E5
	.dc.b	$23,$FC,$BF,$E4,$E2,$64,$6C,$C9,$70,$FF,$11,$FD,$4F,$FB,$C6,$3F
	.dc.b	$C6,$ED,$8F,$F1,$3F,$D6,$FF,$30,$9C,$FC,$77,$D0,$94,$E5,$E7,$19
	.dc.b	$8C,$70,$7F,$B3,$6F,$A3,$7E,$5F,$94,$1E,$1E,$4E,$19,$8F,$91,$CB
	.dc.b	$4F,$F2,$6D,$FF,$2E,$71,$8F,$A2,$B8,$BE,$EF,$89,$EE,$7A,$7E,$B1
	.dc.b	$F5,$98,$BF,$83,$36,$BF,$C8,$F5,$A9,$C9,$FE,$5C,$71,$1D,$63,$15
	.dc.b	$C7,$3F,$81,$F0,$5E,$E3,$C1,$F3,$B6,$BE,$7C,$5B,$EF,$36,$BF,$DE
	.dc.b	$98,$23,$17,$FE,$0C,$A4,$E6,$39,$9F,$E7,$0F,$AC,$7C,$BF,$D7,$6C
	.dc.b	$7F,$A9,$FE,$AD,$F1,$D9,$A9,$3C,$5F,$CC,$4E,$A3,$07,$19,$DE,$1F
	.dc.b	$D1,$67,$4F,$F9,$BA,$FB,$B7,$E2,$7C,$47,$E6,$3E,$32,$E1,$E7,$8B
	.dc.b	$3F,$E9,$EE,$7F,$F7,$14,$E4,$FF,$5D,$D6,$B5,$FE,$25,$E1,$7F,$9F
	.dc.b	$E3,$13,$F1,$7E,$AF,$F3,$58,$7C,$70,$E1,$D7,$3F,$F1,$B7,$51,$97
	.dc.b	$FC,$72,$FE,$62,$76,$7C,$9F,$81,$FD,$14,$66,$7F,$B5,$EB,$FE,$66
	.dc.b	$AD,$1F,$EE,$5F,$86,$7E,$03,$F8,$8D,$3F,$95,$78,$0B,$E1,$87,$F6
	.dc.b	$FF,$9D,$7C,$FF,$F0,$96,$E5,$FE,$AC,$FA,$BF,$F7,$CB,$FE,$E1,$8C
	.dc.b	$D7,$FA,$7A,$72,$DF,$90,$FE,$4F,$F5,$6F,$AA,$7D,$5F,$EE,$EF,$7A
	.dc.b	$FC,$CE,$5F,$D6,$63,$FE,$D7,$C3,$85,$FF,$10,$EF,$E7,$8F,$2D,$3F
	.dc.b	$D7,$FE,$67,$F3,$87,$F2,$CD,$C3,$FC,$C7,$F5,$6F,$96,$FE,$15,$39
	.dc.b	$9F,$F4,$47,$A5,$83,$F8,$75,$6E,$FA,$C6,$2F,$51,$FE,$CE,$9D,$57
	.dc.b	$FA,$FA,$FC,$1D,$5F,$E5,$9A,$FF,$39,$ED,$ED,$AF,$E5,$35,$FC,$A9
	.dc.b	$93,$A3,$FC,$F3,$5C,$63,$F2,$93,$A7,$FD,$E7,$F2,$9F,$D5,$2F,$F2
	.dc.b	$6D,$BE,$2F,$92,$33,$E9,$FF,$52,$8D,$53,$04,$61,$B9,$6E,$B8,$7C
	.dc.b	$84,$C7,$CD,$83,$EE,$DF,$A6,$FF,$9A,$E7,$38,$F4,$BE,$17,$E8,$7F
	.dc.b	$B4,$DF,$0F,$89,$CE,$9E,$A9,$B8,$7F,$A6,$AB,$4F,$4D,$13,$EB,$F8
	.dc.b	$0F,$EB,$DD,$2F,$95,$C5,$F9,$BB,$3F,$37,$18,$D3,$FC,$99,$5B,$AE
	.dc.b	$97,$CA,$BE,$37,$C0,$7C,$7F,$5A,$D7,$85,$B8,$7D,$26,$EB,$F4,$9F
	.dc.b	$C4,$FC,$5C,$AF,$8F,$4F,$AB,$F0,$2B,$FE,$05,$CC,$BE,$13,$9B,$FE
	.dc.b	$61,$39,$79,$DF,$2B,$E3,$7F,$D9,$63,$13,$C2,$31,$97,$4A,$FC,$3F
	.dc.b	$D6,$FE,$0C,$FC,$27,$DB,$5B,$F9,$D1,$72,$FC,$7C,$DA,$FE,$31,$F5
	.dc.b	$CD,$BF,$BA,$3F,$8C,$A3,$9C,$3E,$0B,$93,$11,$ED,$3F,$DA,$EE,$67
	.dc.b	$9F,$3F,$F7,$FF,$A8,$78,$14,$FF,$15,$FD,$6B,$EA,$F1,$97,$CE,$AD
	.dc.b	$B6,$B7,$3A,$CB,$C1,$D4,$C3,$95,$EC,$FC,$CB,$89,$F3,$0C,$3A,$BF
	.dc.b	$B2,$F1,$3F,$23,$E9,$7E,$B5,$B7,$F9,$76,$CE,$E3,$F2,$9C,$FF,$DA
	.dc.b	$63,$F7,$7F,$59,$F9,$AF,$AD,$FD,$B3,$6B,$F4,$0D,$9D,$C7,$29,$FD
	.dc.b	$0F,$C4,$C8,$FF,$52,$79,$F6,$BD,$FF,$03,$FD,$1F,$B7,$F9,$ED,$FF
	.dc.b	$91,$EB,$5C,$BF,$D7,$2B,$FC,$49,$87,$23,$FD,$5D,$ED,$ED,$A6,$71
	.dc.b	$EB,$19,$7F,$D7,$1F,$51,$F2,$6F,$FA,$7B,$6B,$F8,$ED,$BE,$97,$FC
	.dc.b	$AD,$39,$BF,$E2,$93,$9A,$FD,$D4,$E7,$8F,$FD,$8D,$7C,$39,$3F,$CD
	.dc.b	$7E,$77,$F4,$8C,$BF,$D0,$DF,$0C,$8C,$63,$E0,$FA,$DD,$F9,$BF,$E6
	.dc.b	$76,$C7,$FB,$2D,$BA,$FC,$86,$E1,$F1,$5B,$3C,$CF,$F6,$3D,$F6,$AF
	.dc.b	$4C,$E9,$FF,$60,$7D,$5F,$F1,$8A,$FF,$66,$B6,$3E,$AF,$CC,$7F,$A1
	.dc.b	$23,$18,$7D,$38,$C7,$8F,$FE,$7D,$6D,$FF,$0C,$B1,$F9,$51,$F5,$AF
	.dc.b	$A4,$FF,$09,$79,$FF,$EA,$98,$C4,$FF,$85,$7B,$78,$C6,$23,$9C,$EE
	.dc.b	$BF,$C3,$51,$D6,$74,$FF,$A5,$DA,$E6,$3B,$DF,$47,$8D,$D9,$FE,$EB
	.dc.b	$35,$FF,$26,$45,$7A,$4F,$07,$D6,$F4,$69,$E9,$29,$E8,$9E,$4F,$8C
	.dc.b	$EE,$68,$F8,$F3,$DE,$FD,$D1,$CE,$FD,$B1,$88,$F2,$9C,$5C,$3C,$8C
	.dc.b	$57,$82,$BA,$36,$32,$78,$4F,$21,$E7,$31,$1C,$1A,$70,$AF,$9B,$19
	.dc.b	$6C,$66,$FE,$34,$7A,$2B,$C0,$D9,$E6,$D3,$89,$67,$E7,$6E,$F9,$DA
	.dc.b	$4A,$F9,$D1,$97,$F1,$FF,$1B,$FE,$7E,$D7,$F2,$33,$99,$FF,$7C,$7C
	.dc.b	$07,$F5,$D5,$19,$AE,$11,$8B,$FF,$46,$46,$27,$55,$C7,$1F,$CA,$4B
	.dc.b	$DE,$FF,$49,$46,$1B,$97,$2A,$E0,$DC,$BF,$1D,$7C,$72,$B7,$C1,$F5
	.dc.b	$DF,$06,$D6,$98,$7E,$3A,$BD,$38,$79,$2B,$84,$A7,$B3,$BB,$EC,$FC
	.dc.b	$4B,$F1,$D7,$0A,$DD,$FA,$EB,$35,$1E,$47,$8A,$3C,$11,$97,$84,$68
	.dc.b	$F3,$57,$53,$B5,$F8,$AF,$BE,$1F,$16,$B7,$87,$58,$FA,$75,$BB,$E5
	.dc.b	$39,$AF,$C0,$CE,$63,$C5,$5C,$27,$47,$BC,$EA,$79,$70,$F3,$9E,$A7
	.dc.b	$F7,$7C,$CE,$87,$D5,$FF,$B3,$F0,$FE,$77,$C6,$F3,$F1,$C1,$F9,$08
	.dc.b	$C6,$3F,$E5,$0C,$38,$BF,$CB,$AB,$6F,$F3,$26,$EE,$AE,$33,$FD,$9B
	.dc.b	$1D,$E5,$E3,$FA,$DD,$D7,$FB,$62,$BF,$9C,$FC,$16,$B2,$55,$37,$12
	.dc.b	$7C,$91,$97,$89,$0F,$EE,$AE,$DF,$F1,$26,$3E,$59,$C6,$9F,$DD,$6B
	.dc.b	$C1,$F3,$5E,$CF,$05,$DC,$F0,$30,$DC,$BE,$25,$62,$64,$7F,$6C,$79
	.dc.b	$BF,$27,$D8,$F6,$7D,$6C,$8C,$10,$FF,$05,$B9,$7F,$C1,$B1,$72,$FE
	.dc.b	$FF,$C1,$E2,$31,$C5,$FD,$5B,$3F,$3C,$66,$FF,$BE,$BA,$7F,$D2,$9E
	.dc.b	$AF,$CF,$7F,$C1,$78,$7B,$0F,$AB,$19,$C3,$FD,$43,$B6,$6B,$FC,$0F
	.dc.b	$5A,$8D,$FC,$AF,$C4,$F8,$99,$5F,$E8,$B8,$CB,$FE,$0E,$8C,$BF,$EA
	.dc.b	$E8,$FC,$1F,$D6,$FF,$BD,$F0,$FF,$7C,$7D,$57,$F8,$66,$BF,$CF,$AD
	.dc.b	$7F,$C9,$B3,$BD,$7F,$BA,$C2,$75,$5C,$BC,$6F,$CA,$CE,$5E,$F5,$8A
	.dc.b	$DB,$C9,$B3,$FE,$00,$6F,$3B,$0F,$04,$FF,$81,$3D,$07,$CE,$76,$FA
	.dc.b	$C6,$B4,$F0,$CE,$AB,$FD,$2E,$D7,$FC,$C3,$3C,$1F,$F4,$DB,$5F,$F3
	.dc.b	$2C,$6F,$4F,$F6,$9F,$77,$D6,$24,$FF,$AE,$7B,$EB,$FE,$DA,$8B,$3F
	.dc.b	$ED,$A9,$F8,$7F,$59,$FE,$F7,$AF,$9B,$F1,$5F,$0C,$CF,$C0,$D8,$C9
	.dc.b	$0F,$F3,$BC,$F1,$7F,$CF,$3B,$AB,$84,$61,$E6,$FF,$B3,$7D,$BF,$C6
	.dc.b	$FD,$64,$8C,$10,$FF,$78,$B9,$7F,$DE,$71,$F9,$4F,$4F,$F6,$9F,$2F
	.dc.b	$FE,$D6,$9F,$F6,$93,$5F,$F7,$CC,$6D,$CF,$99,$C0,$C3,$FE,$FE,$F9
	.dc.b	$FF,$CF,$91,$FF,$78,$FC,$13,$EF,$DD,$59,$9E,$07,$FB,$9E,$BF,$90
	.dc.b	$D6,$E7,$5C,$2E,$9F,$F2,$4F,$4B,$B5,$86,$A7,$FD,$E5,$97,$6C,$BF
	.dc.b	$DE,$5E,$67,$EC,$88,$FF,$BD,$7E,$3A,$BF,$EA,$EC,$3F,$C0,$99,$77
	.dc.b	$8F,$DE,$D7,$EE,$AB,$F9,$7F,$D1,$7F,$BE,$67,$FC,$B7,$5B,$CD,$39
	.dc.b	$9F,$F5,$CB,$67,$58,$7F,$83,$A7,$8E,$5E,$CA,$D3,$7E,$0B,$B9,$FC
	.dc.b	$2B,$C9,$FD,$8F,$D6,$FB,$6F,$FD,$7F,$F3,$8F,$AA,$FD,$AB,$63,$FD
	.dc.b	$F7,$3B,$CF,$D9,$09,$D3,$CA,$B8,$9D,$B8,$6C,$79,$8F,$F3,$CC,$F0
	.dc.b	$AE,$6D,$CB,$F8,$4E,$1D,$DF,$67,$D7,$FF,$08,$FA,$B7,$C5,$53,$E5
	.dc.b	$C7,$F0,$CF,$A0,$FF,$7B,$65,$DC,$38,$7F,$DB,$AD,$CB,$D7,$7D,$27
	.dc.b	$83,$F9,$98,$CF,$3F,$84,$ED,$F0,$DF,$C6,$BE,$B3,$E9,$5F,$FB,$A1
	.dc.b	$B5,$F1,$9B,$9F,$DE,$FB,$E7,$C0,$FF,$22,$FA,$EF,$E1,$53,$FC,$A1
	.dc.b	$B8,$7F,$A7,$9A,$AF,$9B,$9F,$2B,$FD,$2E,$D2,$CF,$F1,$B7,$F9,$98
	.dc.b	$70,$FE,$1F,$8F,$F9,$9F,$D7,$3D,$EA,$7E,$5A,$2C,$FC,$B9,$D3,$67
	.dc.b	$4D,$20,$D3,$F0,$F3,$C2,$3F,$37,$5F,$DE,$F9,$3D,$B3,$E6,$AD,$D6
	.dc.b	$71,$F2,$B0,$F1,$19,$29,$0F,$06,$7F,$D4,$AB,$A3,$FD,$CE,$FE,$6F
	.dc.b	$D3,$3E,$B9,$F5,$7F,$D5,$3E,$BB,$F7,$37,$FD,$69,$B8,$AF,$85,$6F
	.dc.b	$C0,$93,$F8,$FA,$DD,$6A,$31,$3A,$7C,$A7,$FB,$6A,$77,$87,$8A,$3F
	.dc.b	$4D,$5E,$5F,$8E,$F2,$76,$49,$CD,$27,$11,$D0,$5F,$E1,$9B,$5F,$E4
	.dc.b	$4B,$E9,$1E,$78,$E3,$B7,$EF,$2B,$F1,$B5,$AA,$F3,$AF,$AD,$7D,$37
	.dc.b	$F7,$AF,$15,$F2,$D4,$ED,$3A,$BF,$D9,$4E,$9B,$3B,$7D,$F3,$B7,$D5
	.dc.b	$3B,$6C,$ED,$F1,$CE,$F1,$FE,$5D,$87,$97,$E2,$BC,$9C,$72,$FA,$C5
	.dc.b	$7A,$C5,$AE,$36,$AF,$E6,$BC,$12,$FF,$8C,$78,$BF,$CD,$23,$F9,$CF
	.dc.b	$DF,$66,$7F,$9A,$C5,$C3,$F9,$D7,$8A,$AF,$F3,$47,$D4,$FF,$96,$7E
	.dc.b	$BB,$F7,$BE,$F1,$FD,$FE,$5F,$E8,$39,$DE,$5F,$C9,$EB,$5A,$F1,$33
	.dc.b	$1F,$8A,$AE,$5C,$1F,$E7,$EF,$AE,$FE,$39,$D1,$FE,$D6,$6D,$2B,$6F
	.dc.b	$B3,$2F,$EC,$2F,$F3,$5E,$55,$70,$6F,$83,$FE,$86,$AD,$D7,$F9,$E2
	.dc.b	$B4,$49,$D6,$3F,$E9,$0F,$AE,$7D,$5F,$97,$E2,$66,$1D,$E7,$2D,$9C
	.dc.b	$BD,$67,$94,$7F,$92,$3E,$B9,$F2,$7F,$CC,$2B,$FC,$9F,$18,$D3,$F9
	.dc.b	$1B,$63,$F9,$08,$70,$FF,$24,$F8,$EF,$F0,$5F,$4F,$F2,$A6,$C7,$06
	.dc.b	$BB,$F3,$BF,$81,$3D,$2B,$FA,$6F,$89,$E0,$F2,$FF,$8F,$4F,$91,$A8
	.dc.b	$F1,$38,$C7,$F6,$FE,$47,$F8,$1F,$AB,$FC,$05,$F1,$7F,$E5,$8E,$6F
	.dc.b	$A9,$E0,$FF,$48,$7F,$D8,$1F,$03,$B7,$A6,$1C,$9C,$9C,$30,$D6,$5F
	.dc.b	$81,$C3,$A3,$FE,$B0,$FD,$17,$F7,$75,$F9,$EE,$67,$B3,$C3,$15,$6A
	.dc.b	$FF,$21,$5A,$BF,$C4,$4E,$74,$FF,$0D,$4E,$6B,$B7,$89,$E2,$EB,$4E
	.dc.b	$6B,$FC,$4D,$8F,$6A,$D4,$E7,$1F,$6E,$3D,$5F,$F1,$CC,$67,$17,$6F
	.dc.b	$CC,$9F,$F7,$27,$D5,$7C,$08,$C4,$FE,$9F,$54,$E7,$8F,$FC,$0B,$7C
	.dc.b	$0F,$D8,$38,$AE,$13,$F2,$BD,$87,$E7,$3F,$2B,$6F,$F1,$F4,$63,$8B
	.dc.b	$DD,$E1,$BE,$6B,$85,$EF,$3F,$C1,$5E,$DF,$F2,$EC,$67,$2F,$DB,$5F
	.dc.b	$19,$CE,$1F,$6A,$5F,$5C,$F8,$EF,$BB,$D8,$FC,$8D,$F0,$DB,$AD,$63
	.dc.b	$FD,$32,$72,$FF,$89,$63,$38,$7F,$3E,$9C,$BF,$D2,$E3,$3B,$7F,$2E
	.dc.b	$9C,$BF,$E4,$28,$7C,$67,$8A,$5F,$01,$E1,$8C,$37,$0F,$13,$70,$F0
	.dc.b	$37,$0E,$CD,$9F,$D4,$65,$ED,$8F,$67,$77,$C0,$FC,$73,$F8,$6D,$7A
	.dc.b	$E3,$1E,$1E,$A3,$FB,$5C,$EB,$1F,$65,$6A,$78,$1A,$AE,$13,$FD,$82
	.dc.b	$B8,$63,$EF,$8F,$9F,$C2,$F8,$7F,$9D,$18,$6E,$1F,$DE,$A7,$4F,$CF
	.dc.b	$9D,$34,$D4,$F0,$8F,$ED,$F3,$C1,$FE,$F9,$F8,$3F,$F6,$EA,$FF,$35
	.dc.b	$64,$FC,$27,$FD,$33,$1F,$E5,$F9,$D5,$7F,$83,$AB,$5E,$4F,$E2,$67
	.dc.b	$39,$7F,$A5,$67,$37,$C7,$D2,$FE,$0F,$18,$C3,$C3,$18,$C7,$BC,$62
	.dc.b	$BC,$71,$8B,$FF,$42,$FD,$73,$E4,$FF,$88,$C3,$EE,$7F,$D1,$D7,$FE
	.dc.b	$75,$AD,$3F,$E8,$59,$D7,$B7,$F9,$F9,$CE,$3F,$E2,$89,$CD,$71,$C3
	.dc.b	$AE,$1D,$2B,$FC,$45,$7C,$1F,$F3,$0B,$5F,$E6,$41,$C3,$FC,$D5,$B9
	.dc.b	$BF,$2B,$F8,$F8,$7F,$AB,$7E,$93,$E2,$C6,$2F,$F6,$F1,$FB,$5F,$9A
	.dc.b	$FE,$2D,$1F,$E6,$B9,$D2,$8C,$DF,$5D,$B8,$78,$D3,$1F,$BB,$83,$D4
	.dc.b	$C6,$3F,$DF,$FF,$5D,$7A,$DF,$C3,$BE,$F3,$F9,$64,$63,$0F,$F0,$BF
	.dc.b	$84,$9F,$E1,$1E,$F3,$E2,$7E,$EF,$EE,$6F,$A0,$CC,$61,$CF,$CD,$63
	.dc.b	$1F,$C7,$D2,$31,$A7,$F9,$3F,$E8,$1F,$5D,$FD,$3B,$F2,$57,$D6,$BE
	.dc.b	$B3,$F8,$C5,$6B,$2F,$F6,$15,$6A,$73,$7C,$6F,$85,$7F,$AE,$6F,$86
	.dc.b	$1F,$E6,$8C,$38,$4E,$6F,$86,$1C,$A3,$AE,$1C,$AF,$FD,$D3,$F5,$1F
	.dc.b	$CD,$2F,$F9,$AB,$6B,$E3,$B6,$77,$1F,$2A,$7F,$9D,$FB,$DF,$CF,$3E
	.dc.b	$02,$CF,$CE,$6D,$7F,$BD,$9B,$DF,$FE,$42,$7F,$A1,$FD,$AF,$F4,$4F
	.dc.b	$7A,$DF,$D6,$6D,$FF,$BF,$9B,$7F,$E4,$A6,$CF,$FB,$F5,$B3,$FD,$19
	.dc.b	$FE,$BD,$F9,$4B,$E4,$3F,$A8,$F6,$E1,$B7,$FE,$4B,$9D,$3F,$86,$9D
	.dc.b	$34,$FF,$2F,$F4,$BE,$AB,$FA,$DC,$7E,$39,$AE,$9B,$8F,$E5,$5B,$92
	.dc.b	$7F,$5F,$C5,$F8,$F7,$FC,$45,$19,$AD,$7E,$DB,$F9,$97,$FD,$35,$5B
	.dc.b	$7F,$D4,$F4,$9D,$57,$19,$56,$E3,$A5,$6F,$B3,$FB,$44,$E6,$3F,$C6
	.dc.b	$46,$4F,$19,$E6,$E8,$FF,$13,$B9,$BF,$1C,$63,$0F,$0C,$FF,$8E,$FD
	.dc.b	$F7,$D6,$C8,$FF,$43,$BF,$91,$CB,$DD,$3A,$8F,$F1,$11,$99,$E4,$F1
	.dc.b	$C3,$FC,$F8,$E5,$FF,$3C,$39,$AE,$6E,$7D,$D7,$E6,$DF,$E8,$FB,$FB
	.dc.b	$3E,$1F,$F8,$86,$31,$87,$F0,$98,$C7,$C5,$7F,$08,$6D,$70,$3E,$1C
	.dc.b	$FC,$6F,$86,$F8,$7E,$DF,$F5,$3C,$FE,$85,$27,$3E,$97,$F8,$56,$30
	.dc.b	$72,$15,$C5,$FE,$62,$6A,$FE,$EC,$65,$F0,$47,$18,$C4,$7E,$2E,$BB
	.dc.b	$CF,$9A,$1F,$6F,$FE,$37,$C3,$E7,$F8,$FF,$E1,$4A,$D4,$E5,$FB,$11
	.dc.b	$9A,$FF,$74,$4E,$5F,$F5,$34,$66,$BF,$1D,$39,$DF,$F9,$A7,$B1,$F3
	.dc.b	$7B,$BF,$98,$FA,$D7,$2F,$D4,$B7,$2F,$E7,$0D,$C7,$F0,$7D,$1F,$E0
	.dc.b	$CA,$FF,$42,$71,$3A,$9F,$63,$FE,$67,$F8,$27,$D6,$FE,$FA,$30,$DB
	.dc.b	$FF,$6A,$CE,$9F,$DF,$4E,$9F,$F3,$E7,$CD,$6F,$4F,$F3,$CB,$73,$FF
	.dc.b	$3B,$B6,$7F,$CD,$AD,$9F,$F3,$6E,$3F,$5E,$BF,$CB,$FF,$81,$FF,$6F
	.dc.b	$7F,$75,$9F,$A7,$8F,$FC,$CD,$39,$C7,$DB,$7C,$5F,$F3,$0C,$67,$0E
	.dc.b	$B5,$C3,$D6,$FF,$48,$7E,$37,$FD,$66,$FD,$BD,$18,$F3,$FE,$23,$63
	.dc.b	$27,$FB,$66,$78,$61,$FE,$7F,$9D,$36,$75,$5C,$9A,$72,$86,$9F,$E9
	.dc.b	$4B,$E6,$FF,$AD,$2F,$CB,$F0,$9F,$57,$8E,$E1,$F4,$9B,$87,$FA,$3D
	.dc.b	$B7,$FE,$BC,$7F,$CF,$51,$FE,$C3,$F3,$3F,$D8,$9E,$A7,$DE,$8F,$F0
	.dc.b	$AF,$A9,$FC,$96,$31,$1F,$E1,$89,$DF,$2F,$F8,$6A,$15,$B8,$FF,$6F
	.dc.b	$CE,$A7,$9E,$35,$E4,$7F,$84,$E7,$95,$71,$8E,$95,$B7,$AD,$55,$19
	.dc.b	$9E,$11,$D2,$FC,$EF,$03,$B1,$D0,$A7,$FB,$8E,$38,$BF,$0C,$C9,$C6
	.dc.b	$39,$BE,$03,$31,$C1,$C4,$78,$23,$13,$E5,$AF,$43,$6B,$A9,$C2,$FE
	.dc.b	$0C,$EA,$3A,$B6,$FA,$36,$77,$3C,$AB,$A5,$79,$9B,$3F,$4A,$79,$9B
	.dc.b	$2C,$ED,$B1,$B8,$EF,$1B,$69,$BF,$89,$7A,$7D,$5B,$3F,$35,$AF,$DE
	.dc.b	$6F,$C6,$BE,$9B,$F3,$B2,$9C,$C6,$1B,$AA,$75,$7F,$E6,$8B,$E5,$39
	.dc.b	$AD,$D7,$DC,$9F,$F6,$8B,$F8,$F6,$B9,$BF,$0E,$1F,$9C,$9E,$73,$C7
	.dc.b	$E1,$7F,$CE,$B3,$96,$E5,$E1,$BE,$2D,$C7,$B4,$EA,$FD,$18,$78,$30
	.dc.b	$ED,$87,$09,$CB,$54,$EF,$0F,$75,$F3,$9C,$D6,$EB,$CF,$7E,$59,$F0
	.dc.b	$9E,$F9,$D1,$BA,$E5,$5D,$67,$C3,$86,$CE,$2F,$C1,$30,$7C,$77,$37
	.dc.b	$DE,$30,$47,$04,$C3,$E2,$46,$5F,$5B,$C9,$CC,$7C,$49,$C4,$FC,$8B
	.dc.b	$F3,$E1,$EA,$9C,$CF,$C6,$9C,$C7,$F0,$DC,$5C,$2F,$47,$FD,$5B,$3F
	.dc.b	$E2,$0E,$B7,$D2,$63,$83,$FC,$51,$18,$CB,$F9,$C5,$E5,$FC,$A6,$1C
	.dc.b	$DF,$EB,$B7,$9F,$01,$7C,$67,$F9,$BE,$1F,$93,$BE,$11,$83,$9F,$E1
	.dc.b	$1C,$69,$F2,$B4,$F0,$69,$EA,$9B,$67,$57,$FD,$BA,$B3,$12,$30,$43
	.dc.b	$C1,$5C,$DF,$EA,$91,$C9,$F2,$63,$EA,$AE,$51,$E7,$ED,$FF,$01,$6D
	.dc.b	$F0,$FE,$29,$F8,$5F,$F3,$EA,$FF,$1A,$C6,$39,$FE,$45,$F0,$9F,$F3
	.dc.b	$F5,$F1,$BE,$11,$F0,$EF,$83,$A3,$FC,$7F,$8B,$E3,$BF,$83,$47,$FA
	.dc.b	$1F,$E8,$FA,$F8,$27,$E9,$4E,$A7,$D6,$DE,$4F,$F0,$1D,$6E,$79,$1D
	.dc.b	$A7,$5C,$5F,$E0,$4B,$FA,$9A,$CA,$D9,$F0,$EF,$93,$F2,$38,$BD,$78
	.dc.b	$FE,$36,$BC,$B7,$F0,$B9,$BF,$B3,$CF,$F8,$BA,$7F,$11,$E2,$70,$CB
	.dc.b	$FC,$5F,$5A,$6F,$46,$AA,$65,$FE,$62,$E1,$FE,$AF,$8F,$FA,$3F,$F6
	.dc.b	$87,$C7,$FE,$4F,$F7,$4F,$5D,$B8,$51,$88,$57,$07,$FC,$B0,$D7,$F0
	.dc.b	$71,$CA,$78,$F5,$FE,$8E,$78,$DD,$FB,$D7,$9D,$9F,$A9,$7B,$14,$D4
	.dc.b	$4B,$FF,$33,$4E,$9B,$39,$7F,$4C,$2B,$AB,$A2,$1A,$72,$FE,$5F,$F2
	.dc.b	$1F,$C4,$3F,$EC,$83,$E7,$9F,$94,$72,$EF,$9F,$2E,$CF,$9D,$1A,$DB
	.dc.b	$FD,$4F,$FA,$B3,$F0,$2F,$E0,$0B,$C9,$A6,$22,$5F,$E6,$A7,$4D,$9C
	.dc.b	$BE,$EF,$98,$CB,$F0,$3E,$B9,$F6,$7F,$3D,$B1,$FE,$8D,$9D,$E3,$FE
	.dc.b	$CF,$6F,$47,$F6,$59,$57,$07,$A5,$FD,$9A,$DC,$7F,$6B,$6B,$C2,$3C
	.dc.b	$06,$F0,$FC,$BF,$C7,$7E,$63,$F0,$49,$5F,$EF,$48,$CB,$FE,$F6,$8C
	.dc.b	$BF,$C3,$3B,$13,$FC,$13,$F2,$E7,$A1,$FC,$A7,$E3,$1E,$CD,$35,$13
	.dc.b	$0F,$E0,$73,$A6,$CE,$5E,$DF,$75,$FA,$CF,$B6,$91,$FC,$2B,$24,$FE
	.dc.b	$C0,$CC,$79,$67,$0E,$E3,$3C,$3F,$EA,$49,$CC,$FA,$FE,$33,$F7,$3F
	.dc.b	$05,$23,$B4,$7E,$F0,$C3,$F0,$63,$0F,$8F,$E7,$3F,$80,$FD,$CC,$8C
	.dc.b	$10,$F9,$0E,$5F,$95,$1F,$67,$97,$FA,$A7,$C2,$FF,$89,$FE,$4A,$FA
	.dc.b	$27,$57,$F8,$7A,$CC,$48,$C1,$0F,$EA,$95,$CD,$FF,$77,$B5,$FC,$B4
	.dc.b	$FF,$01,$F0,$7F,$AF,$6A,$8F,$C8,$7E,$1C,$F8,$8F,$E0,$CD,$BF,$F7
	.dc.b	$25,$F0,$AD,$D7,$3F,$72,$57,$FB,$A6,$F8,$3F,$89,$F2,$FF,$0D,$DF
	.dc.b	$23,$EE,$47,$F1,$3F,$8E,$7C,$C5,$24,$FF,$BC,$FF,$75,$7C,$46,$23
	.dc.b	$F8,$6C,$3E,$C7,$F8,$E3,$63,$FD,$F5,$7D,$1E,$53,$F9,$7F,$AA,$91
	.dc.b	$83,$FE,$00,$7D,$77,$E5,$5F,$F5,$10,$70,$FF,$A9,$5A,$FE,$DD,$AF
	.dc.b	$F9,$75,$E5,$C3,$F3,$63,$FC,$C1,$F5,$FF,$D4,$23,$10,$8C,$70,$F3
	.dc.b	$ED,$FE,$6D,$7C,$A3,$F8,$24,$EA,$B9,$FA,$BA,$BF,$E7,$9F,$5E,$FA
	.dc.b	$7F,$E7,$DC,$AE,$5A,$6A,$26,$1F,$E7,$3D,$3C,$75,$B8,$C6,$7F,$C4
	.dc.b	$E3,$11,$FC,$F3,$0E,$CE,$8F,$F3,$E3,$87,$FA,$0C,$70,$AE,$B8,$78
	.dc.b	$27,$FA,$2F,$D4,$7F,$5E,$FA,$47,$E4,$8F,$23,$FC,$71,$18,$7F,$AC
	.dc.b	$4F,$F6,$13,$84,$EE,$BC,$55,$B7,$C8,$27,$87,$A9,$D2,$B8,$4F,$F8
	.dc.b	$EE,$B8,$3E,$49,$EF,$EF,$EF,$C2,$93,$D3,$0F,$F4,$8E,$7F,$69,$BB
	.dc.b	$7F,$A7,$71,$F5,$DF,$0B,$F7,$E1,$D1,$F3,$E4,$9D,$63,$CE,$BF,$BB
	.dc.b	$74,$FF,$2E,$BF,$6F,$E4,$9F,$D0,$31,$3F,$D0,$27,$53,$FD,$C1,$B7
	.dc.b	$F9,$1A,$DF,$99,$F3,$2F,$87,$95,$D8,$C4,$4C,$3E,$4E,$1C,$B4,$FE
	.dc.b	$A8,$FE,$CA,$3C,$7E,$A7,$33,$CF,$DB,$23,$24,$3C,$8F,$F9,$47,$E9
	.dc.b	$3F,$AF,$47,$F9,$25,$BA,$7E,$9F,$E0,$11,$8C,$7F,$D9,$1E,$CB,$CD
	.dc.b	$7D,$E7,$3C,$61,$FE,$5A,$BE,$0F,$F8,$A6,$76,$FF,$8C,$7E,$A5,$E6
	.dc.b	$FF,$9E,$62,$19,$7F,$CC,$31,$A8,$FF,$2D,$FC,$0F,$4C,$7F,$D0,$F1
	.dc.b	$AA,$FF,$3F,$FB,$1F,$C8,$35,$FE,$01,$4F,$DD,$BF,$9E,$5F,$F9,$91
	.dc.b	$AF,$F9,$F2,$B8,$6B,$FE,$E1,$6C,$7F,$90,$67,$77,$D1,$B5,$BE,$DF
	.dc.b	$09,$FE,$9B,$8C,$D7,$1F,$37,$2F,$ED,$FC,$EF,$F5,$4F,$D1,$7F,$9C
	.dc.b	$3D,$E9,$39,$F8,$94,$FF,$A6,$9B,$1F,$E6,$F9,$DB,$FE,$94,$8C,$BF
	.dc.b	$E4,$0F,$AB,$79,$8C,$3F,$95,$CA,$73,$18,$6C,$7F,$30,$6B,$F8,$6F
	.dc.b	$8C,$7C,$8B,$BF,$9C,$61,$B1,$99,$FE,$73,$39,$7A,$7C,$F7,$FA,$61
	.dc.b	$FF,$17,$5A,$31,$03,$CC,$E1,$BF,$0F,$FD,$13,$1F,$E1,$9E,$57,$F3
	.dc.b	$DE,$3F,$E7,$D1,$8E,$3F,$F4,$A7,$DD,$E5,$E9,$19,$C3,$FD,$77,$E3
	.dc.b	$FF,$37,$9F,$A8,$D8,$FF,$50,$D7,$17,$E5,$46,$6F,$BF,$D6,$FF,$3D
	.dc.b	$BF,$23,$6B,$F5,$B8,$23,$11,$FE,$4D,$8C,$79,$7F,$D4,$A3,$CB,$3C
	.dc.b	$6F,$AC,$F8,$7A,$7F,$35,$E2,$78,$F0,$E3,$8F,$E2,$F0,$E0,$F2,$BE
	.dc.b	$1D,$1F,$E6,$D8,$D4,$7A,$FE,$27,$E6,$C7,$FA,$8F,$87,$E9,$3F,$EB
	.dc.b	$BF,$A0,$D4,$63,$5F,$F6,$D3,$75,$FE,$D9,$7C,$A7,$FD,$8F,$5C,$38
	.dc.b	$BD,$1E,$FB,$E5,$7E,$57,$8B,$FC,$33,$E7,$33,$18,$F1,$51,$8D,$7F
	.dc.b	$85,$37,$2F,$E2,$CF,$D3,$5E,$23,$DE,$F9,$51,$F8,$98,$C5,$7E,$81
	.dc.b	$B1,$FE,$0C,$6B,$FC,$89,$CE,$DF,$E0,$46,$CF,$FA,$0F,$F5,$37,$C3
	.dc.b	$78,$67,$37,$DF,$E7,$3F,$0D,$DE,$FA,$E7,$D7,$6E,$73,$18,$CB,$FB
	.dc.b	$2E,$09,$F4,$E1,$EC,$6C,$66,$7D,$51,$83,$A3,$FD,$BB,$2F,$F7,$D3
	.dc.b	$D2,$3C,$DD,$FE,$4F,$AE,$7C,$DE,$2B,$FB,$BC,$61,$F1,$CF,$80,$8F
	.dc.b	$C9,$6E,$3F,$EC,$CC,$51,$F4,$FE,$93,$94,$E6,$3E,$33,$F1,$38,$BF
	.dc.b	$DB,$B7,$F0,$23,$1E,$3E,$A3,$EB,$47,$C0,$23,$83,$81,$FE,$0C,$FD
	.dc.b	$DB,$9C,$78,$CF,$C2,$7D,$7F,$FB,$A2,$7F,$C4,$98,$3E,$BF,$FC,$BB
	.dc.b	$1F,$E1,$49,$D5,$39,$3F,$C8,$D7,$BF,$57,$9F,$AB,$F5,$2F,$77,$FD
	.dc.b	$A1,$AF,$FB,$1D,$AF,$F1,$97,$AD,$FF,$93,$9B,$87,$F1,$DE,$EF,$F7
	.dc.b	$A7,$3F,$68,$FF,$13,$FD,$77,$FA,$F9,$FF,$1E,$5B,$EB,$FF,$D7,$F1
	.dc.b	$AF,$90,$FE,$29,$39,$F6,$7F,$49,$7C,$73,$FF,$34,$B5,$FE,$FE,$D7
	.dc.b	$FD,$BC,$D7,$F9,$83,$C6,$7F,$CC,$AD,$BF,$47,$CA,$FE,$3A,$BE,$44
	.dc.b	$72,$C3,$F9,$36,$EF,$AF,$F7,$D7,$E1,$9A,$BE,$6B,$F8,$BB,$9E,$0F
	.dc.b	$DE,$47,$F2,$68,$FF,$79,$35,$FE,$88,$D7,$97,$8F,$FC,$02,$FC,$1F
	.dc.b	$09,$CA,$3C,$78,$7F,$B9,$32,$FF,$29,$E1,$E0,$F3,$7F,$C3,$5E,$BF
	.dc.b	$F2,$DF,$6F,$FD,$61,$5F,$88,$FA,$FF,$F2,$23,$FE,$E1,$9F,$E2,$1F
	.dc.b	$45,$FD,$D2,$3F,$B5,$B7,$5F,$F4,$BB,$5F,$F1,$6B,$5F,$EB,$9A,$AF
	.dc.b	$FC,$43,$5D,$FD,$CF,$E4,$FF,$71,$F2,$A7,$37,$F3,$8F,$EC,$9F,$48
	.dc.b	$FA,$8F,$EA,$9F,$28,$FA,$F7,$D8,$7F,$25,$AD,$67,$FE,$D2,$9C,$BF
	.dc.b	$E7,$48,$CE,$1F,$EA,$E6,$BF,$E4,$46,$BF,$DC,$C3,$87,$C7,$87,$73
	.dc.b	$C7,$5F,$C9,$AB,$B4,$7F,$42,$FA,$7F,$E8,$6D,$8F,$F4,$C7,$0B,$D2
	.dc.b	$F1,$F5,$FF,$8E,$15,$F0,$AD,$C6,$1B,$7F,$D3,$BD,$FF,$E9,$FF,$61
	.dc.b	$67,$6D,$CF,$E5,$65,$DE,$31,$97,$77,$FA,$BF,$D8,$7F,$58,$F5,$AE
	.dc.b	$1E,$06,$E1,$CB,$3F,$EC,$53,$FE,$64,$55,$C6,$3C,$ED,$9E,$0D,$C3
	.dc.b	$FB,$4F,$D2,$BD,$8F,$EC,$FD,$DF,$DD,$63,$18,$7C,$08,$C4,$C8,$F8
	.dc.b	$7F,$51,$71,$F6,$F5,$BC,$DF,$EE,$35,$F9,$D6,$BF,$42,$34,$FC,$C3
	.dc.b	$45,$3F,$D7,$3C,$DF,$A2,$AF,$C9,$43,$EC,$3F,$4F,$1F,$95,$13,$BD
	.dc.b	$3F,$DE,$15,$BE,$DF,$EF,$13,$CE,$BF,$D2,$2D,$7F,$1C,$D7,$FC,$84
	.dc.b	$D7,$AF,$C7,$72,$3F,$C6,$95,$E3,$F6,$70,$FE,$33,$EC,$07,$F8,$A4
	.dc.b	$FD,$0D,$7E,$09,$CC,$6F,$AF,$51,$B6,$C6,$4F,$F7,$77,$96,$FB,$07
	.dc.b	$C7,$7F,$09,$F1,$BF,$C5,$91,$8A,$F9,$D1,$89,$91,$E5,$F8,$8B,$97
	.dc.b	$C9,$8F,$F3,$69,$FE,$FB,$3F,$C4,$7E,$9F,$F8,$7D,$CC,$7D,$2D,$3F
	.dc.b	$B0,$F0,$27,$F6,$C6,$A7,$FD,$7A,$D7,$37,$FE,$0E,$8C,$47,$F9,$BD
	.dc.b	$EB,$3C,$E3,$13,$F3,$B8,$5F,$3D,$E3,$8F,$F1,$97,$D8,$7E,$8A,$9D
	.dc.b	$CF,$2F,$FC,$31,$1F,$85,$9D,$4E,$7D,$AE,$3D,$F3,$4F,$E0,$78,$7F
	.dc.b	$82,$B4,$F1,$62,$C7,$B7,$17,$D1,$F0,$BF,$9C,$7B,$5F,$41,$B8,$FD
	.dc.b	$6E,$4F,$E1,$6F,$F6,$F6,$E1,$E6,$E6,$FF,$19,$3F,$CB,$70,$E9,$5F
	.dc.b	$7B,$BD,$F6,$1F,$E4,$C9,$46,$27,$FD,$D3,$49,$CD,$6E,$B5,$3F,$EE
	.dc.b	$E9,$CD,$75,$AF,$F3,$E7,$D4,$3E,$C3,$F8,$04,$61,$93,$FC,$9A,$46
	.dc.b	$2F,$F8,$DB,$C6,$35,$3D,$23,$5B,$7F,$0E,$E3,$F2,$12,$BD,$38,$FB
	.dc.b	$23,$18,$7C,$3C,$BF,$07,$5F,$F0,$03,$07,$D8,$FF,$15,$BF,$F4,$97
	.dc.b	$D8,$7F,$A3,$70,$FF,$4F,$B5,$FF,$4F,$35,$FF,$58,$B5,$FF,$6A,$3C
	.dc.b	$72,$FF,$5C,$B7,$0F,$06,$9D,$BE,$2E,$47,$FB,$A6,$75,$7F,$90,$CB
	.dc.b	$E7,$E1,$CA,$F7,$3F,$DA,$7E,$2D,$7C,$39,$FE,$5E,$5A,$C3,$C3,$F0
	.dc.b	$FF,$E8,$AB,$DB,$6F,$FD,$35,$9F,$FB,$45,$B8,$FC,$09,$F8,$67,$DC
	.dc.b	$EF,$B5,$FE,$DB,$C3,$FD,$2B,$E7,$7F,$86,$BE,$8B,$FB,$66,$1F,$E6
	.dc.b	$AA,$C3,$FD,$C4,$FF,$13,$46,$BD,$3F,$DA,$7A,$A8,$FE,$F1,$2F,$57
	.dc.b	$0F,$F7,$7A,$FD,$D4,$62,$3F,$BD,$61,$FE,$ED,$8C,$47,$84,$E5,$5D
	.dc.b	$2F,$37,$E0,$8C,$1E,$3C,$3C,$F5,$D0,$F3,$9E,$17,$2F,$F8,$60,$C9
	.dc.b	$FE,$22,$72,$FF,$83,$0F,$09,$C4,$A7,$5B,$F1,$C7,$A6,$7C,$73,$BB
	.dc.b	$F5,$BC,$DD,$BE,$03,$C4,$E0,$DE,$1E,$6C,$3D,$D3,$F4,$67,$55,$DA
	.dc.b	$3D,$07,$33,$D1,$87,$63,$9B,$E8,$3E,$63,$89,$E7,$1D,$67,$E5,$3C
	.dc.b	$EF,$D8,$FC,$1A,$E6,$75,$9F,$BE,$D8,$FC,$33,$5F,$A5,$AF,$E0,$DF
	.dc.b	$C5,$35,$CC,$7E,$22,$31,$F2,$0F,$9A,$FF,$51,$52,$73,$18,$6E,$89
	.dc.b	$D5,$FF,$91,$6F,$94,$E6,$B7,$5F,$6A,$5F,$33,$FE,$4B,$BF,$CA,$46
	.dc.b	$27,$E0,$87,$9B,$C7,$0F,$2C,$F0,$AE,$58,$F8,$5B,$5E,$5B,$E3,$5D
	.dc.b	$AF,$DB,$9F,$1A,$FC,$CD,$78,$5F,$CA,$BE,$08,$E1,$97,$92,$35,$7C
	.dc.b	$6B,$C9,$39,$BE,$B8,$F8,$E7,$AE,$3F,$E9,$A8,$F8,$D3,$A9,$F3,$35
	.dc.b	$F8,$13,$B7,$D4,$D7,$E2,$46,$F0,$F3,$57,$79,$F4,$D6,$E7,$31,$DA
	.dc.b	$FD,$53,$F1,$A3,$B9,$E9,$88,$F2,$3D,$27,$C6,$BF,$01,$E1,$9E,$C7
	.dc.b	$77,$33,$DA,$61,$F1,$89,$8F,$C9,$26,$1F,$12,$7F,$90,$E5,$FC,$DB
	.dc.b	$D4,$FF,$48,$47,$F4,$46,$ED,$FD,$29,$B8,$7F,$41,$EB,$B7,$C6,$3F
	.dc.b	$9B,$4F,$F7,$67,$FA,$7B,$63,$51,$FD,$63,$CE,$E3,$C6,$69,$DB,$97
	.dc.b	$FB,$8F,$E0,$78,$AF,$EA,$53,$5F,$EE,$71,$73,$FE,$B7,$37,$3F,$ED
	.dc.b	$53,$BA,$FE,$CB,$3E,$3C,$3C,$11,$FD,$36,$3D,$1D,$9F,$E1,$0E,$0F
	.dc.b	$7B,$67,$FC,$2D,$1E,$5E,$97,$D5,$6B,$B7,$F3,$95,$FE,$B9,$BE,$0F
	.dc.b	$F8,$4E,$32,$FE,$EE,$32,$FF,$AF,$F5,$FF,$07,$F1,$F5,$24,$7F,$84
	.dc.b	$FB,$2B,$FE,$12,$8B,$BF,$F9,$26,$BF,$C2,$F5,$E3,$F9,$F9,$7F,$E2
	.dc.b	$D7,$7C,$1E,$1C,$91,$F0,$1B,$3F,$E1,$66,$BE,$98,$F8,$BC,$BF,$D7
	.dc.b	$71,$FF,$2C,$67,$FD,$39,$B1,$9C,$3F,$8C,$65,$FE,$84,$E7,$F9,$73
	.dc.b	$9C,$7F,$CB,$3E,$83,$AB,$F0,$0D,$8F,$F4,$0C,$ED,$7D,$7F,$F9,$91
	.dc.b	$FE,$FD,$8F,$FB,$9E,$31,$1F,$EA,$56,$E5,$F1,$7E,$BF,$EB,$CF,$03
	.dc.b	$F9,$64,$EA,$BC,$F3,$FC,$D6,$7F,$DC,$ED,$7A,$79,$5F,$A2,$5F,$00
	.dc.b	$F3,$B0,$D8,$EC,$6A,$38,$7D,$7F,$E4,$23,$FD,$75,$3B,$7F,$CB,$AD
	.dc.b	$7F,$C2,$ED,$7F,$A1,$1C,$CF,$D7,$3C,$72,$FF,$3C,$CB,$D7,$EE,$79
	.dc.b	$2F,$F3,$7F,$39,$FE,$5C,$C3,$FD,$5A,$D7,$FC,$E8,$DB,$FC,$67,$D7
	.dc.b	$FF,$AC,$27,$7D,$3F,$59,$86,$0F,$99,$C6,$97,$D4,$FE,$ED,$79,$14
	.dc.b	$C4,$4A,$E3,$3A,$6C,$E5,$FF,$70,$7C,$46,$5B,$1F,$4E,$31,$8F,$FA
	.dc.b	$6B,$EB,$FF,$B6,$57,$D2,$9E,$0F,$F9,$09,$AF,$CC,$6B,$FD,$BD,$A9
	.dc.b	$EB,$C5,$F4,$A7,$F1,$73,$FE,$F3,$6D,$7F,$AA,$AB,$73,$F9,$EF,$86
	.dc.b	$FC,$EF,$D2,$48,$F8,$FC,$E9,$FD,$09,$98,$5F,$EF,$65,$5A,$9F,$F7
	.dc.b	$2F,$D7,$BE,$2B,$F8,$BB,$63,$31,$FA,$CF,$9E,$CB,$FC,$B7,$E0,$FF
	.dc.b	$D9,$13,$F6,$1A,$FF,$AD,$6B,$93,$E2,$D7,$F7,$78,$BE,$BF,$F6,$FF
	.dc.b	$5F,$2B,$D7,$1F,$BA,$30,$F1,$3F,$7F,$19,$76,$7C,$27,$7A,$7B,$E7
	.dc.b	$F8,$1F,$C5,$FF,$04,$F8,$19,$7F,$22,$32,$FC,$A8,$CB,$DF,$95,$8A
	.dc.b	$BF,$86,$FC,$D7,$F0,$EF,$AF,$97,$F4,$A3,$2F,$D4,$8C,$BC,$E9,$3F
	.dc.b	$8F,$D2,$4F,$E3,$5F,$EC,$BF,$09,$FE,$3B,$FA,$3C,$CF,$FB,$63,$97
	.dc.b	$FD,$DA,$FE,$46,$BF,$B3,$3F,$8D,$6C,$7F,$BC,$B6,$47,$F9,$06,$FF
	.dc.b	$C8,$DF,$49,$DB,$D1,$3D,$AB,$FD,$F3,$E1,$3E,$93,$F8,$B3,$72,$FE
	.dc.b	$26,$D7,$D9,$F0,$7F,$C6,$FE,$BD,$F1,$3E,$46,$CF,$F7,$B6,$BF,$EF
	.dc.b	$26,$BF,$D6,$78,$3F,$9E,$71,$FD,$0F,$63,$F9,$D9,$E2,$F8,$9F,$E5
	.dc.b	$1F,$5F,$FC,$2E,$FF,$08,$3E,$3D,$F2,$FF,$CB,$5B,$1F,$CA,$A7,$86
	.dc.b	$FF,$12,$3F,$CC,$93,$A7,$FD,$7A,$73,$AF,$9B,$19,$8F,$F6,$0B,$5F
	.dc.b	$E5,$74,$FA,$FF,$29,$23,$F9,$CC,$F5,$BF,$04,$3E,$BB,$F4,$F1,$86
	.dc.b	$F2,$FF,$36,$6C,$FF,$9E,$7E,$BB,$F8,$9B,$F0,$23,$F5,$EF,$FB,$51
	.dc.b	$AF,$1F,$8B,$FC,$2C,$62,$3F,$D0,$4F,$45,$3F,$DD,$AF,$CA,$DB,$E5
	.dc.b	$3F,$D4,$BE,$B9,$E2,$7F,$59,$6C,$66,$3F,$AE,$71,$FE,$17,$0F,$7D
	.dc.b	$6A,$BE,$6B,$5F,$F7,$5B,$5F,$EA,$AD,$7F,$D3,$0F,$18,$F9,$32,$DF
	.dc.b	$B9,$E3,$BF,$ED,$9F,$53,$FD,$BE,$BE,$7B,$71,$FE,$C2,$DC,$5F,$4A
	.dc.b	$F3,$BA,$47,$F7,$09,$FE,$F0,$FF,$C0,$06,$BC,$BD,$CF,$D8,$C6,$23
	.dc.b	$FD,$5C,$F5,$A7,$D7,$7F,$30,$9F,$F0,$44,$63,$CA,$F3,$DF,$A3,$DF
	.dc.b	$95,$E1,$7F,$CC,$F5,$E5,$8C,$9F,$96,$C3,$FC,$86,$F8,$27,$C6,$47
	.dc.b	$BE,$0F,$1B,$9B,$63,$B1,$FE,$0E,$E7,$FE,$6F,$AF,$E5,$9B,$AF,$F8
	.dc.b	$21,$B3,$F2,$3E,$DF,$F6,$0F,$1F,$33,$F8,$9C,$FF,$40,$8F,$F0,$79
	.dc.b	$FB,$48,$C9,$1F,$87,$F5,$FF,$D7,$B5,$3F,$D8,$23,$81,$E5,$F9,$5F
	.dc.b	$F1,$45,$FE,$EB,$1F,$24,$7C,$6C,$7F,$C4,$99,$7A,$63,$C6,$4B,$FF
	.dc.b	$46,$5F,$F0,$97,$F8,$87,$D7,$FF,$A5,$3E,$A9,$F4,$BE,$8C,$3F,$83
	.dc.b	$46,$2B,$FC,$9B,$51,$FF,$21,$46,$5B,$5C,$72,$FE,$BF,$5F,$E8,$2F
	.dc.b	$A6,$FE,$95,$5F,$E4,$9F,$82,$46,$23,$FA,$F4,$63,$CF,$FF,$43,$71
	.dc.b	$F1,$64,$F1,$AE,$19,$FF,$9E,$7D,$9F,$F7,$23,$FC,$77,$CB,$FC,$E7
	.dc.b	$67,$FA,$BB,$AB,$F9,$8C,$E5,$BA,$FF,$9B,$BF,$35,$7C,$DE,$EE,$67
	.dc.b	$FE,$5F,$C7,$9B,$FE,$36,$AD,$B9,$C3,$F9,$DA,$13,$C1,$FC,$1B,$5F
	.dc.b	$F2,$9B,$5E,$0F,$F5,$1F,$43,$FD,$8B,$E6,$9A,$7F,$AA,$7D,$EB,$5E
	.dc.b	$59,$D3,$67,$4F,$F9,$AB,$EB,$3F,$A3,$5F,$FB,$35,$B1,$FD,$01,$AF
	.dc.b	$E6,$1A,$FF,$9A,$3E,$3D,$F1,$3F,$C2,$23,$FA,$3C,$63,$B3,$FD,$B0
	.dc.b	$D8,$CC,$ED,$E9,$F3,$5F,$E9,$56,$FC,$3F,$8E,$31,$1F,$E7,$6C,$11
	.dc.b	$88,$F9,$7E,$17,$F3,$2E,$2F,$A8,$D8,$FF,$0C,$72,$3E,$BF,$FA,$F7
	.dc.b	$2F,$E0,$EB,$4D,$F1,$B3,$87,$D0,$9C,$BF,$EB,$0B,$E4,$F6,$9D,$E1
	.dc.b	$F7,$3E,$BF,$F9,$94,$FE,$15,$B1,$FE,$29,$FA,$FF,$F3,$BC,$E7,$C9
	.dc.b	$E3,$C4,$BD,$E7,$F9,$76,$D7,$FB,$02,$F8,$E1,$FE,$41,$BE,$18,$72
	.dc.b	$BD,$CF,$8E,$3C,$BE,$92,$F8,$E3,$D6,$7D,$3D,$CE,$0F,$9B,$9F,$EC
	.dc.b	$DB,$8F,$CA,$FA,$0D,$C6,$30,$FF,$6B,$B6,$BF,$DE,$4D,$A6,$4D,$98
	.dc.b	$72,$E1,$F2,$FC,$03,$E5,$FE,$5F,$89,$FC,$43,$E8,$33,$18,$F1,$51
	.dc.b	$8D,$7F,$86,$B7,$4F,$F3,$25,$F1,$9F,$F6,$3F,$85,$FE,$4E,$6F,$BD
	.dc.b	$F6,$29,$F5,$BF,$E4,$77,$FC,$53,$5B,$6C,$EE,$31,$C3,$FE,$18,$AE
	.dc.b	$0F,$F8,$6A,$33,$7F,$EC,$C5,$ED,$F1,$3C,$14,$F9,$0C,$57,$9A,$6E
	.dc.b	$3F,$E3,$AA,$D4,$E4,$91,$E6,$9E,$75,$C3,$0F,$17,$67,$9C,$F7,$78
	.dc.b	$CF,$9F,$C6,$3F,$12,$6E,$3F,$E4,$8A,$D4,$E4,$91,$F1,$27,$9D,$70
	.dc.b	$BF,$81,$3F,$14,$F9,$9F,$59,$FE,$F9,$BF,$A3,$19,$69,$FE,$FA,$8D
	.dc.b	$47,$D0,$34,$58,$D7,$C6,$62,$BF,$C8,$FF,$57,$FE,$F3,$9F,$E5,$91
	.dc.b	$8F,$8E,$FC,$DF,$C7,$FD,$49,$FF,$7E,$D6,$F0,$FF,$32,$B6,$F8,$4F
	.dc.b	$F5,$1A,$E1,$5F,$CD,$EB,$53,$CA,$FF,$11,$18,$97,$62,$F8,$C2,$74
	.dc.b	$BE,$AB,$FD,$A9,$87,$F9,$B2,$31,$F1,$9F,$E9,$A7,$FA,$34,$E9,$FF
	.dc.b	$38,$B5,$FF,$04,$B9,$CB,$FC,$D2,$DA,$FE,$9D,$8F,$8F,$83,$F3,$D3
	.dc.b	$9B,$EB,$86,$0F,$8F,$36,$FE,$59,$FD,$5F,$77,$D5,$7B,$DB,$AF,$EA
	.dc.b	$7E,$53,$F8,$7C,$62,$B8,$E3,$FD,$7E,$7F,$89,$A7,$76,$F8,$0E,$5E
	.dc.b	$17,$F2,$AC,$3B,$C6,$30,$E7,$18,$7A,$E1,$CE,$BF,$8E,$FC,$2F,$FA
	.dc.b	$BA,$7F,$E0,$0F,$A5,$CF,$EA,$7F,$C7,$D1,$83,$8F,$CB,$7F,$69,$6F
	.dc.b	$A1,$BF,$57,$F8,$24,$E7,$2F,$F7,$76,$6C,$3F,$DC,$18,$79,$1B,$3F
	.dc.b	$EE,$5F,$0D,$F5,$5C,$EB,$CF,$F3,$EF,$A5,$FE,$20,$D8,$F8,$FF,$1C
	.dc.b	$FA,$AF,$C4,$CE,$71,$FF,$5D,$B7,$AB,$FD,$4E,$DB,$FF,$6D,$57,$0C
	.dc.b	$3A,$92,$3A,$C6,$5F,$AB,$E6,$33,$7D,$07,$F3,$2B,$FE,$66,$DA,$DC
	.dc.b	$E9,$B3,$A6,$9A,$F8,$AF,$E7,$7F,$55,$FC,$F1,$B1,$93,$9C,$65,$A7
	.dc.b	$6F,$AA,$FE,$87,$E7,$7F,$44,$6C,$64,$F1,$C6,$5A,$67,$6F,$A5,$19
	.dc.b	$6C,$7F,$93,$7E,$7B,$FA,$8B,$7D,$0F,$EA,$3F,$41,$FD,$8A,$73,$5F
	.dc.b	$E5,$06,$F0,$FF,$B9,$7C,$34,$63,$0F,$F1,$E7,$07,$CB,$C3,$96,$3F
	.dc.b	$94,$F2,$7F,$AE,$D5,$3E,$59,$72,$FC,$43,$72,$FF,$79,$61,$F5,$63
	.dc.b	$FA,$B7,$27,$E6,$AB,$FD,$FA,$DA,$FF,$7C,$4B,$EA,$BF,$B2,$BF,$A6
	.dc.b	$9D,$3F,$E5,$1A,$46,$1F,$F2,$A3,$63,$97,$A1,$E0,$F8,$A9,$1C,$DF
	.dc.b	$F0,$FB,$A3,$07,$E3,$3E,$90,$7F,$84,$07,$D2,$D7,$0D,$8E,$0D,$C1
	.dc.b	$97,$C2,$AE,$35,$FE,$85,$FA,$97,$C4,$FF,$86,$1B,$1E,$1F,$80,$C4
	.dc.b	$C8,$E0,$F8,$E3,$18,$76,$9F,$88,$7F,$96,$8F,$E1,$A5,$74,$1F,$9D
	.dc.b	$F2,$FF,$E0,$97,$58,$E6,$7F,$CF,$D1,$83,$AD,$7C,$8D,$97,$C6,$24
	.dc.b	$F8,$1D,$46,$2B,$C2,$3E,$AB,$FC,$58,$DE,$6F,$F1,$EF,$91,$FD,$E2
	.dc.b	$37,$3A,$24,$EB,$D3,$FF,$10,$47,$F8,$D7,$A1,$B7,$F0,$BB,$FE,$B6
	.dc.b	$DE,$4F,$B7,$F5,$FF,$3F,$B6,$6D,$F4,$DF,$BD,$18,$BF,$D4,$7D,$4F
	.dc.b	$EA,$CF,$F2,$8F,$BA,$FA,$BC,$61,$FB,$39,$B5,$FE,$76,$9B,$C7,$FE
	.dc.b	$B9,$78,$9C,$AE,$DF,$F1,$A7,$A3,$63,$4D,$3F,$80,$71,$7E,$F4,$F1
	.dc.b	$FD,$5D,$8C,$93,$0F,$F4,$3C,$65,$FF,$37,$7D,$57,$F9,$A2,$7F,$D6
	.dc.b	$6D,$8F,$F2,$EE,$0C,$BF,$CF,$73,$9F,$84,$34,$E1,$B7,$F9,$8E,$77
	.dc.b	$F4,$B3,$1F,$F6,$65,$6B,$0E,$91,$89,$D6,$3E,$47,$E6,$DE,$35,$FF
	.dc.b	$6F,$61,$FE,$6B,$BE,$7F,$05,$F4,$BA,$FC,$AC,$8C,$9F,$2E,$31,$8F
	.dc.b	$E4,$0F,$F6,$C7,$93,$FD,$EF,$E7,$7F,$85,$43,$CD,$FD,$07,$AA,$9C
	.dc.b	$CA,$B8,$3F,$C4,$6B,$73,$A9,$CC,$FF,$83,$F0,$AF,$80,$FF,$04,$4F
	.dc.b	$82,$B8,$C7,$8A,$B7,$18,$8F,$E3,$D1,$8A,$E7,$1E,$33,$99,$E3,$73
	.dc.b	$1B,$8C,$9C,$8A,$7B,$E3,$6E,$27,$C0,$F9,$EF,$E1,$3D,$4E,$27,$85
	.dc.b	$CB,$EF,$33,$5E,$18,$CC,$73,$BF,$81,$7E,$09,$CD,$F8,$2B,$A5,$69
	.dc.b	$E6,$79,$6B,$84,$66,$BA,$CF,$03,$CC,$E2,$FC,$53,$F5,$67,$91,$D2
	.dc.b	$7C,$35,$D2,$38,$CE,$E7,$31,$E2,$E0,$79,$FC,$DB,$1F,$64,$38,$7E
	.dc.b	$83,$7E,$29,$F2,$FF,$D1,$27,$75,$FD,$6E,$73,$BA,$72,$58,$CB,$80
	.dc.b	$E2,$FB,$5F,$5C,$3E,$AC,$BE,$47,$FB,$44,$61,$B9,$78,$6B,$83,$71
	.dc.b	$FC,$DD,$F1,$C6,$DF,$07,$C9,$7C,$1B,$5A,$61,$F9,$2A,$F4,$E1,$E4
	.dc.b	$AE,$13,$EA,$BE,$8D,$3C,$31,$AA,$E3,$87,$C7,$AD,$C7,$96,$BB,$56
	.dc.b	$CE,$13,$A3,$6E,$70,$F2,$E1,$F0,$EB,$BC,$47,$D1,$5C,$5C,$4F,$A6
	.dc.b	$78,$E1,$E3,$9C,$BF,$25,$E2,$68,$F8,$A7,$9A,$30,$FB,$CC,$12,$7C
	.dc.b	$8F,$23,$13,$FC,$E3,$CF,$FF,$11,$57,$F2,$66,$E9,$9A,$FE,$99,$E2
	.dc.b	$61,$FC,$AF,$17,$F9,$99,$FE,$6B,$59,$D7,$F9,$EB,$BE,$67,$37,$F8
	.dc.b	$8A,$BF,$9D,$F9,$5B,$C0,$D3,$84,$BE,$AF,$15,$C0,$87,$F8,$4A,$FA
	.dc.b	$C6,$A3,$FA,$AD,$5D,$3F,$C9,$37,$CE,$78,$C7,$F4,$49,$D6,$9F,$E5
	.dc.b	$8A,$D5,$7F,$97,$F7,$F8,$1D,$FF,$E5,$E6,$F9,$9F,$87,$C7,$AF,$D5
	.dc.b	$E2,$65,$FF,$9B,$5F,$5C,$7B,$38,$E4,$FF,$77,$7D,$FD,$B8,$8F,$66
	.dc.b	$5F,$BE,$74,$7F,$7D,$8F,$0F,$B9,$23,$FC,$F4,$F1,$C7,$FC,$E9,$92
	.dc.b	$35,$3F,$E7,$97,$5A,$7F,$B9,$39,$7F,$98,$DF,$F9,$ED,$B1,$9E,$5F
	.dc.b	$B5,$CB,$FC,$4E,$B5,$1F,$12,$B7,$87,$FC,$00,$6F,$17,$F9,$83,$9F
	.dc.b	$F1,$12,$FA,$BF,$F1,$FB,$FE,$D8,$6B,$FE,$3C,$6E,$1F,$E8,$0F,$AB
	.dc.b	$70,$F2,$9C,$E9,$FE,$79,$6D,$FF,$1F,$6E,$2B,$E3,$3F,$EE,$6F,$82
	.dc.b	$7C,$1F,$91,$B1,$98,$F3,$FC,$9F,$E5,$23,$04,$3F,$CA,$33,$C1,$FF
	.dc.b	$29,$C6,$DC,$CF,$52,$3F,$EF,$68,$D6,$7F,$EA,$9F,$94,$FD,$2C,$FF
	.dc.b	$37,$8C,$72,$7E,$42,$73,$3F,$9F,$FA,$AF,$F3,$FB,$BD,$7F,$9F,$74
	.dc.b	$2B,$78,$72,$AD,$09,$D6,$1F,$B5,$C3,$9C,$F5,$FA,$BC,$78,$B2,$3F
	.dc.b	$DC,$B2,$9E,$06,$63,$C9,$38,$7B,$8F,$A6,$47,$E0,$9B,$1A,$6B,$FE
	.dc.b	$7C,$FA,$BF,$EC,$2E,$67,$81,$1E,$B1,$A7,$FD,$07,$3C,$B2,$FF,$48
	.dc.b	$47,$86,$B8,$3D,$7C,$1F,$E2,$9F,$06,$F8,$7B,$F4,$12,$BE,$3F,$2A
	.dc.b	$B5,$18,$6B,$FC,$45,$4F,$18,$C5,$7F,$08,$FA,$B7,$D0,$52,$4F,$3F
	.dc.b	$97,$AF,$FA,$B3,$89,$5C,$23,$EC,$C6,$2B,$D3,$5E,$AF,$AA,$F5,$FD
	.dc.b	$5C,$BF,$34,$65,$F4,$46,$5F,$44,$65,$F5,$46,$77,$FF,$77,$E8,$AF
	.dc.b	$8F,$F1,$1F,$23,$DD,$98,$7C,$38,$CB,$F1,$63,$2F,$C5,$8F,$9F,$F2
	.dc.b	$9F,$43,$EA,$E4,$FC,$B8,$CB,$F3,$63,$2F,$CD,$8F,$C4,$75,$48,$FC
	.dc.b	$43,$FD,$4F,$E0,$7F,$C4,$FF,$17,$99,$FC,$1C,$DD,$3F,$DA,$55,$A9
	.dc.b	$E2,$43,$67,$E9,$9D,$3F,$EC,$EF,$7D,$F3,$DF,$C6,$9B,$1F,$9E,$6B
	.dc.b	$F8,$3F,$94,$FE,$3F,$F5,$4F,$8F,$A4,$9E,$17,$F9,$2F,$2A,$47,$E3
	.dc.b	$5E,$DF,$21,$FC,$A3,$EA,$BF,$6B,$A7,$E7,$3E,$35,$F4,$7F,$CB,$1B
	.dc.b	$87,$EC,$2B,$7B,$FF,$90,$61,$3A,$7F,$DD,$2D,$BE,$97,$DF,$E2,$7C
	.dc.b	$1F,$55,$2F,$A3,$DE,$7F,$9D,$FD,$53,$E2,$A9,$2B,$F8,$0F,$D5,$7E
	.dc.b	$D7,$1F,$D9,$D6,$E7,$F8,$2B,$5F,$F7,$DB,$63,$F7,$33,$C7,$3F,$54
	.dc.b	$EE,$BF,$D5,$ED,$73,$7F,$EA,$E7,$FC,$C1,$19,$7F,$CC,$5F,$54,$EE
	.dc.b	$FE,$A6,$D8,$FE,$2E,$D5,$8F,$F0,$AE,$2F,$BA,$FF,$0C,$9D,$36,$B9
	.dc.b	$9F,$D7,$A7,$78,$FD,$1A,$DB,$67,$77,$FC,$9A,$B6,$FF,$9D,$FE,$8F
	.dc.b	$FC,$00,$F8,$AB,$ED,$46,$30,$FE,$3A,$D8,$FF,$3F,$7D,$2F,$F4,$1E
	.dc.b	$2F,$ED,$F3,$9B,$E9,$ED,$E0,$FF,$1D,$56,$E3,$F9,$BC,$ED,$34,$E5
	.dc.b	$1B,$8F,$21,$E7,$F7,$3B,$91,$FE,$9F,$1F,$D4,$DC,$1C,$0E,$B1,$8C
	.dc.b	$58,$FA,$BD,$0F,$F7,$5C,$7B,$A7,$53,$F0,$9F,$4C,$EA,$3F,$3D,$C3
	.dc.b	$FC,$9E,$FB,$36,$FF,$A1,$B6,$7E,$3F,$D4,$7E,$92,$BF,$AB,$CE,$5F
	.dc.b	$F5,$43,$5F,$D7,$F8,$BF,$22,$78,$47,$68,$FF,$0E,$CE,$CE,$59,$2B
	.dc.b	$78,$78,$B9,$76,$4C,$3E,$3C,$FF,$81,$31,$F5,$CE,$67,$D6,$F8,$7D
	.dc.b	$EF,$EB,$3F,$51,$FE,$73,$FA,$27,$D1,$FC,$3B,$46,$1D,$9F,$D6,$67
	.dc.b	$84,$EE,$75,$39,$3F,$A7,$FD,$1B,$C6,$FF,$30,$B6,$33,$1F,$E7,$6E
	.dc.b	$1C,$B2,$32,$7F,$6A,$F5,$BF,$CC,$DA,$7F,$B4,$FD,$1F,$F0,$B7,$3F
	.dc.b	$FA,$1F,$93,$F9,$35,$FF,$85,$A7,$2F,$F7,$2F,$A8,$7C,$BE,$EE,$4E
	.dc.b	$EF,$8B,$FE,$83,$AD,$B9,$9D,$BF,$98,$BE,$AE,$EF,$FD,$31,$E1,$BC
	.dc.b	$8A,$49,$F5,$7B,$35,$E5,$3B,$7D,$0D,$7F,$CF,$31,$FE,$60,$FA,$5F
	.dc.b	$E6,$38,$7F,$15,$6C,$FF,$35,$3F,$DE,$CE,$1C,$35,$FF,$7E,$7C,$2B
	.dc.b	$A7,$F8,$F7,$C4,$A3,$13,$CC,$FF,$7D,$B8,$7B,$46,$E7,$C5,$F2,$CF
	.dc.b	$8C,$A4,$AE,$58,$27,$51,$E7,$C5,$F4,$EF,$0B,$F7,$33,$FE,$98,$F7
	.dc.b	$2F,$7E,$2F,$75,$8E,$5F,$F5,$75,$5B,$56,$A7,$F1,$13,$98,$C4,$F5
	.dc.b	$9F,$C3,$60,$FB,$3E,$5E,$E7,$FB,$3B,$E5,$38,$4E,$6B,$F2,$51,$8A
	.dc.b	$DF,$4A,$31,$3A,$3A,$CE,$A3,$38,$7F,$AD,$1F,$47,$B4,$FD,$37,$FB
	.dc.b	$F5,$FE,$7F,$C5,$CC,$62,$FF,$D7,$71,$8E,$AF,$0B,$6B,$81,$FE,$76
	.dc.b	$F0,$1F,$BB,$BA,$7E,$34,$61,$F7,$FC,$5F,$FA,$6A,$B8,$56,$DB,$3C
	.dc.b	$23,$18,$FF,$B9,$EB,$83,$FC,$EA,$FA,$5F,$F9,$E3,$C2,$7A,$BF,$57
	.dc.b	$A9,$F6,$5F,$1E,$B5,$39,$8C,$61,$FC,$F6,$F8,$D7,$F4,$29,$CF,$47
	.dc.b	$F9,$D2,$73,$5F,$D2,$B0,$EB,$C3,$F8,$38,$4B,$C6,$ED,$F1,$79,$8C
	.dc.b	$65,$D6,$31,$3A,$8F,$53,$6B,$C8,$DC,$7C,$4D,$9F,$F1,$35,$AB,$E0
	.dc.b	$78,$FF,$1A,$FD,$11,$8F,$87,$F1,$36,$FE,$43,$4F,$84,$59,$E3,$1F
	.dc.b	$D9,$2B,$A3,$FE,$35,$F9,$FB,$5F,$BD,$97,$CB,$FC,$39,$CE,$BF,$CF
	.dc.b	$A3,$18,$F9,$9B,$87,$56,$E1,$F3,$70,$E7,$3F,$EB,$4C,$39,$C6,$27
	.dc.b	$FB,$A7,$D3,$FF,$A8,$EB,$FD,$8F,$93,$EC,$3F,$D3,$31,$FE,$BC,$9D
	.dc.b	$57,$F8,$0A,$B5,$E8,$7D,$99,$CC,$7F,$02,$BE,$35,$DB,$E1,$BF,$80
	.dc.b	$4F,$0C,$3A,$47,$FA,$E2,$7F,$CB,$17,$D9,$FE,$4C,$D7,$31,$B7,$EC
	.dc.b	$7E,$18,$E6,$FD,$AF,$E0,$FF,$90,$C3,$FC,$C5,$39,$7F,$D5,$71,$9A
	.dc.b	$E3,$5C,$39,$3F,$25,$7B,$9F,$F5,$A5,$6D,$46,$72,$FF,$67,$B7,$0F
	.dc.b	$C5,$7E,$18,$FC,$31,$F3,$FD,$33,$FE,$41,$E5,$3E,$A9,$F1,$1F,$C6
	.dc.b	$BE,$2B,$F8,$46,$9F,$CB,$2B,$53,$9F,$0B,$F4,$DD,$9F,$D5,$F5,$FF
	.dc.b	$7E,$D6,$9F,$F7,$8C,$67,$7F,$F1,$BC,$E5,$FF,$19,$46,$70,$FF,$73
	.dc.b	$B6,$BF,$45,$5E,$5F,$57,$BF,$A5,$FC,$43,$E9,$BC,$DF,$4D,$6F,$FD
	.dc.b	$C8,$F9,$3E,$77,$F8,$AF,$D8,$D9,$FF,$75,$F0,$A3,$F8,$CF,$D1,$E7
	.dc.b	$6F,$E2,$8D,$DB,$FD,$F3,$F1,$9F,$13,$0F,$DA,$47,$F4,$E5,$F4,$FF
	.dc.b	$19,$FF,$4F,$4E,$EF,$FC,$AB,$1A,$E1,$FE,$19,$39,$BE,$57,$C6,$FF
	.dc.b	$80,$4E,$6F,$F8,$11,$C4,$F0,$9D,$B8,$BF,$AC,$39,$9F,$F2,$E4,$F1
	.dc.b	$7F,$CB,$75,$D7,$D1,$FC,$A7,$D2,$7F,$38,$F0,$FF,$9C,$B7,$0F,$8A
	.dc.b	$DB,$FE,$6C,$DC,$BF,$B6,$DF,$E7,$76,$FE,$83,$F4,$7F,$D0,$9B,$87
	.dc.b	$CE,$6E,$1F,$CF,$DB,$7C,$97,$CE,$FF,$4A,$FA,$1F,$E9,$6D,$BF,$AE
	.dc.b	$DB,$FE,$90,$DB,$E4,$D9,$FE,$51,$8F,$EC,$63,$14,$F4,$9F,$4A,$F8
	.dc.b	$4F,$CF,$7B,$5F,$DA,$A3,$18,$FC,$EB,$62,$ED,$FF,$07,$4E,$71,$E7
	.dc.b	$8F,$28,$C6,$1C,$B0,$FF,$4A,$E1,$C7,$1F,$EC,$FF,$07,$FD,$A3,$E1
	.dc.b	$EE,$1F,$94,$6D,$FF,$65,$6E,$BF,$1D,$FE,$A9,$3F,$E2,$F3,$CC,$3A
	.dc.b	$BF,$9C,$D7,$FA,$6B,$E0,$3F,$0B,$18,$9E,$13,$F9,$48,$C4,$EB,$CE
	.dc.b	$EA,$DA,$F8,$55,$CD,$EF,$5F,$D1,$B3,$FE,$3D,$EE,$FF,$80,$BD,$7F
	.dc.b	$F0,$1B,$6B,$F5,$ED,$BF,$F0,$03,$73,$FF,$1B,$C3,$E9,$BF,$C7,$D5
	.dc.b	$F4,$70,$FF,$7E,$7A,$DF,$DC,$E7,$35,$F3,$3E,$33,$FB,$CE,$1C,$A7
	.dc.b	$2F,$E5,$A3,$38,$7F,$6B,$AE,$DF,$01,$FE,$36,$9C,$C7,$F2,$6F,$BD
	.dc.b	$FF,$1C,$FE,$14,$F9,$BF,$F0,$17,$9C,$9C,$C6,$1B,$8F,$C6,$9D,$3E
	.dc.b	$9A,$E4,$E6,$77,$7F,$ED,$1A,$DB,$F8,$6A,$DB,$FE,$DC,$E8,$FC,$BF
	.dc.b	$9B,$FC,$CF,$EA,$7F,$2C,$DC,$7F,$1E,$DA,$E1,$D3,$FC,$FA,$7E,$57
	.dc.b	$5A,$FF,$3D,$F2,$7F,$CF,$7E,$22,$D6,$9B,$87,$E6,$2D,$18,$E1,$FE
	.dc.b	$03,$18,$AF,$CF,$FB,$5F,$E5,$AF,$A5,$FD,$53,$63,$27,$FA,$C6,$32
	.dc.b	$D3,$EF,$7D,$9B,$8F,$39,$7E,$A2,$31,$3C,$A3,$13,$AC,$7F,$89,$D6
	.dc.b	$A7,$E0,$E1,$FE,$64,$FA,$6F,$F0,$0E,$1F,$E2,$BF,$82,$7C,$C3,$3D
	.dc.b	$7B,$7F,$E5,$C8,$C5,$7E,$F6,$F8,$E3,$FE,$DA,$9C,$BF,$56,$33,$87
	.dc.b	$4C,$3F,$D0,$DF,$20,$F9,$CF,$F3,$EF,$B1,$86,$E5,$FE,$CD,$9D,$3F
	.dc.b	$ED,$59,$D3,$4D,$4F,$0C,$7F,$D9,$15,$C1,$FE,$1B,$F1,$4B,$E3,$18
	.dc.b	$6E,$FF,$CE,$27,$58,$78,$7C,$2F,$F4,$8F,$C6,$7A,$1B,$19,$3F,$DD
	.dc.b	$11,$96,$9F,$C1,$30,$FF,$2B,$7D,$4F,$F9,$5F,$6F,$9B,$1F,$AF,$F6
	.dc.b	$FF,$D6,$71,$88,$FD,$94,$EF,$83,$FD,$A1,$5A,$7F,$D9,$11,$99,$E5
	.dc.b	$75,$E2,$2F,$A3,$FE,$C9,$8D,$3E,$43,$55,$DA,$33,$0A,$DD,$F8,$63
	.dc.b	$FD,$C5,$18,$9D,$B6,$39,$9A,$29,$FE,$16,$8D,$B8,$3C,$85,$3C,$06
	.dc.b	$67,$C0,$FB,$2F,$D7,$18,$8F,$54,$62,$7A,$B6,$3A,$57,$96,$3E,$1C
	.dc.b	$76,$74,$62,$35,$7F,$3A,$7C,$55,$D2,$3E,$34,$74,$32,$5A,$E7,$1C
	.dc.b	$5B,$1F,$0D,$AF,$AE,$39,$C7,$86,$3F,$DD,$BA,$78,$71,$FC,$C4,$7C
	.dc.b	$B6,$BF,$79,$BF,$1E,$F2,$5B,$7F,$A9,$67,$31,$8A,$EA,$7F,$AA,$6B
	.dc.b	$85,$6E,$FF,$A1,$46,$27,$61,$C6,$1F,$97,$5E,$57,$F4,$98,$C3,$72
	.dc.b	$E7,$8F,$86,$78,$36,$3B,$CE,$EF,$C8,$D7,$C1,$3B,$7D,$0D,$7C,$91
	.dc.b	$BB,$F1,$C7,$E8,$DF,$3D,$78,$B1,$C5,$6C,$E1,$7C,$EB,$77,$E2,$36
	.dc.b	$73,$8D,$1B,$9E,$51,$9A,$8F,$E8,$2B,$C6,$D9,$EF,$3A,$C3,$E3,$CF
	.dc.b	$63,$85,$F6,$9C,$4E,$A7,$31,$C7,$0F,$93,$1A,$7D,$CE,$8F,$01,$93
	.dc.b	$E4,$C7,$17,$DC,$68,$E1,$3F,$C5,$B2,$70,$3C,$BF,$F0,$35,$FF,$4C
	.dc.b	$F6,$EE,$DF,$E4,$CE,$84,$63,$26,$3C,$B0,$C3,$63,$FA,$EC,$EF,$C8
	.dc.b	$5F,$29,$FE,$85,$7C,$32,$FF,$0B,$47,$F6,$1F,$99,$E1,$47,$51,$0F
	.dc.b	$ED,$2E,$8C,$3F,$D1,$8C,$1A,$3F,$A8,$1A,$CF,$FB,$4C,$6A,$38,$1F
	.dc.b	$D8,$7E,$7F,$DD,$D2,$FC,$7D,$EF,$61,$B5,$B2,$1F,$B6,$DF,$FB,$84
	.dc.b	$6A,$30,$2B,$A5,$7C,$BF,$17,$F9,$E6,$1F,$03,$BB,$1C,$4F,$9F,$A6
	.dc.b	$71,$FF,$04,$CE,$1F,$EF,$11,$97,$FB,$B3,$F2,$BE,$22,$46,$88,$7F
	.dc.b	$83,$DC,$BC,$0F,$F0,$17,$BB,$F6,$7C,$37,$C8,$D3,$CB,$12,$7F,$C5
	.dc.b	$F1,$97,$89,$FE,$0B,$E8,$E5,$DD,$23,$04,$3F,$C6,$0E,$5F,$F0,$74
	.dc.b	$7E,$3D,$FA,$5E,$0F,$E2,$FF,$09,$FC,$CC,$7F,$99,$DA,$FC,$E3,$67
	.dc.b	$F9,$EB,$8B,$FC,$28,$EF,$9F,$91,$D6,$33,$1C,$5B,$CD,$FE,$C0,$AE
	.dc.b	$AD,$AE,$17,$E5,$FA,$97,$D0,$FE,$7F,$EA,$7F,$9C,$65,$FE,$7B,$6B
	.dc.b	$E5,$9D,$BF,$E8,$E6,$BF,$EA,$88,$DE,$1F,$ED,$06,$F4,$7F,$9E,$EB
	.dc.b	$73,$FE,$D4,$FA,$8F,$D6,$5F,$FA,$C3,$DD,$FD,$AB,$FF,$00,$3C,$0F
	.dc.b	$D9,$4E,$AB,$87,$81,$FB,$5B,$DD,$F1,$BD,$CF,$D1,$FA,$97,$B7,$F4
	.dc.b	$FD,$A7,$D9,$53,$51,$30,$FF,$54,$CE,$9B,$39,$7E,$B5,$74,$72,$43
	.dc.b	$35,$E8,$E9,$F4,$F8,$32,$7F,$D4,$DA,$A7,$51,$89,$D3,$FE,$D2,$9D
	.dc.b	$3F,$ED,$9F,$9D,$79,$9F,$9D,$88,$65,$FD,$A4,$6A,$3F,$D1,$66,$8F
	.dc.b	$F4,$B3,$97,$B7,$DC,$7C,$EF,$62,$47,$9A,$3E,$89,$87,$C9,$18,$74
	.dc.b	$7B,$1D,$70,$FD,$D9,$FB,$3F,$57,$F6,$BE,$AA,$5F,$C2,$8C,$BF,$12
	.dc.b	$32,$FC,$18,$FC,$0E,$FF,$82,$FB,$C9,$3F,$2A,$32,$FC,$C8,$CB,$F2
	.dc.b	$63,$F6,$5F,$00,$BF,$DA,$7D,$63,$E1,$29,$A8,$97,$FE,$EA,$9D,$36
	.dc.b	$72,$FE,$DE,$BA,$39,$21,$9A,$FC,$CF,$D7,$FF,$03,$FC,$2F,$F6,$7D
	.dc.b	$7F,$ED,$D6,$D3,$D0,$71,$AF,$87,$81,$FA,$2F,$B4,$97,$FE,$FE,$C3
	.dc.b	$8E,$1C,$3D,$FF,$E4,$1E,$0C,$AF,$C7,$C7,$F1,$3F,$AD,$79,$34,$92
	.dc.b	$3F,$89,$57,$FB,$29,$C7,$37,$FC,$00,$AE,$31,$8D,$BF,$DE,$D5,$FD
	.dc.b	$CE,$30,$73,$70,$FF,$1D,$9E,$2F,$6F,$4F,$F9,$B7,$91,$23,$89,$FC
	.dc.b	$52,$3F,$95,$7C,$27,$E5,$F1,$63,$E7,$A4,$E6,$30,$D8,$F6,$6C,$F9
	.dc.b	$0E,$7D,$BF,$CD,$63,$FA,$F6,$1D,$2B,$7E,$9E,$9F,$1E,$1C,$D5,$70
	.dc.b	$C7,$C4,$D8,$F9,$53,$BF,$AB,$FF,$73,$47,$F8,$18,$F9,$CE,$1C,$35
	.dc.b	$FF,$05,$B5,$F7,$3B,$C3,$CD,$0F,$AC,$FF,$72,$3F,$21,$AF,$F4,$B9
	.dc.b	$E1,$87,$2F,$80,$FE,$A9,$5A,$BE,$6D,$9D,$5F,$D2,$FA,$A7,$C5,$FF
	.dc.b	$8A,$7E,$7D,$FA,$AF,$D0,$AF,$BF,$18,$D3,$FB,$05,$F0,$9E,$57,$C6
	.dc.b	$F8,$47,$F8,$32,$F8,$3A,$21,$98,$DE,$5F,$56,$77,$3F,$D9,$EF,$B3
	.dc.b	$7D,$DF,$37,$DD,$BE,$32,$9E,$48,$93,$E8,$9D,$36,$72,$F3,$30,$43
	.dc.b	$07,$F7,$7F,$A6,$FF,$1F,$3F,$0B,$EE,$FF,$9B,$BE,$D6,$E5,$F6,$27
	.dc.b	$73,$C7,$CB,$F9,$95,$AD,$FF,$65,$A2,$31,$5F,$B5,$BE,$38,$73,$C3
	.dc.b	$8F,$99,$2B,$F5,$35,$AC,$7F,$4C,$F2,$C7,$A7,$87,$E0,$3F,$C2,$D1
	.dc.b	$DA,$8E,$F2,$FE,$3B,$1E,$B9,$DE,$1F,$E4,$EF,$CA,$5F,$25,$49,$5F
	.dc.b	$D7,$BE,$6E,$BF,$E4,$AF,$83,$57,$08,$FE,$0F,$18,$AF,$EF,$F5,$FE
	.dc.b	$4D,$F8,$73,$FE,$6C,$89,$5F,$C3,$BE,$11,$5A,$8C,$35,$FF,$2C,$F8
	.dc.b	$BF,$E7,$AA,$E3,$1F,$E5,$CA,$E3,$7F,$E8,$2F,$49,$F2,$7F,$E7,$AF
	.dc.b	$4E,$5F,$F1,$CF,$61,$5A,$8C,$35,$EF,$F0,$CA,$E3,$1F,$E6,$AA,$E3
	.dc.b	$7D,$CC,$F3,$7F,$11,$D3,$FD,$1F,$79,$BE,$BE,$4F,$FA,$33,$D5,$91
	.dc.b	$82,$1E,$07,$2F,$7C,$FF,$97,$E1,$FC,$23,$EA,$DF,$1F,$FE,$28,$AF
	.dc.b	$E5,$51,$8D,$3F,$8B,$42,$30,$FF,$19,$6B,$FE,$A0,$F8,$AF,$E0,$BD
	.dc.b	$3C,$1B,$1F,$C7,$DA,$EF,$C1,$4F,$6A,$FE,$73,$F0,$77,$C3,$34,$97
	.dc.b	$FE,$AE,$C5,$F5,$7F,$EC,$1E,$EF,$A8,$D9,$FA,$3D,$9F,$CE,$F6,$FE
	.dc.b	$79,$7C,$3C,$CC,$61,$FD,$7E,$30,$F2,$8C,$DF,$F7,$3E,$6F,$F0,$44
	.dc.b	$E4,$EF,$F0,$BF,$CC,$A7,$96,$5F,$E8,$E9,$FE,$7D,$87,$9A,$B5,$5F
	.dc.b	$ED,$0A,$D3,$EB,$6E,$3E,$66,$DF,$FB,$67,$DB,$F1,$E1,$CE,$3C,$24
	.dc.b	$BF,$C7,$CE,$AB,$8C,$3A,$9F,$57,$FE,$00,$BF,$D4,$36,$7F,$2F,$EE
	.dc.b	$7C,$0A,$DF,$45,$AD,$67,$FD,$96,$73,$87,$F7,$EE,$AF,$F7,$2E,$1B
	.dc.b	$C3,$E4,$C3,$E0,$7F,$CE,$35,$FB,$EF,$AB,$7C,$0E,$A7,$3A,$7F,$05
	.dc.b	$9C,$D7,$59,$CE,$7F,$DD,$7C,$25,$6A,$FF,$D1,$1F,$01,$F6,$27,$38
	.dc.b	$FD,$A9,$7D,$4F,$EF,$32,$FC,$EB,$63,$FB,$A3,$5F,$F7,$DB,$5F,$F0
	.dc.b	$73,$BC,$7F,$B8,$B6,$7F,$C2,$5F,$78,$F9,$2F,$04,$E6,$FF,$C5,$8B
	.dc.b	$EA,$FF,$1B,$DA,$FA,$CF,$E3,$38,$FF,$88,$2B,$4F,$F6,$C8,$CD,$FF
	.dc.b	$B1,$67,$2F,$A2,$FF,$1A,$FB,$9B,$1A,$8F,$5C,$79,$0F,$89,$F8,$8B
	.dc.b	$E2,$31,$87,$86,$BC,$BF,$5B,$F4,$23,$E0,$C6,$38,$3D,$AF,$62,$47
	.dc.b	$CC,$69,$E7,$3E,$97,$DE,$F3,$7E,$19,$F5,$7E,$4B,$1A,$B0,$7D,$16
	.dc.b	$71,$F1,$B7,$1F,$E2,$B7,$F8,$8C,$3F,$C8,$41,$C3,$FE,$45,$6B,$FE
	.dc.b	$5D,$7C,$4E,$EB,$CB,$5D,$63,$FD,$E5,$F5,$AF,$9A,$FE,$07,$C6,$FA
	.dc.b	$EF,$F5,$C6,$5F,$E5,$6A,$D3,$FE,$60,$9D,$7A,$7F,$91,$9C,$E1,$FE
	.dc.b	$41,$6B,$E8,$6B,$FE,$5B,$6B,$FE,$73,$78,$CF,$F9,$01,$B5,$FE,$77
	.dc.b	$ED,$7C,$7F,$E4,$63,$C9,$97,$E8,$7E,$B3,$F6,$51,$FD,$9F,$EB,$5F
	.dc.b	$5D,$9E,$EF,$E0,$11,$89,$FF,$38,$D7,$0D,$3F,$CB,$0D,$7E,$7B,$5F
	.dc.b	$F3,$93,$5F,$F4,$53,$C7,$4F,$F2,$BB,$67,$FD,$1D,$F4,$6F,$AE,$7C
	.dc.b	$67,$8E,$7F,$D1,$3E,$6B,$EB,$9F,$59,$FC,$D2,$30,$FE,$F7,$DB,$FC
	.dc.b	$D5,$70,$AD,$BF,$E8,$A8,$CE,$9F,$E7,$16,$BF,$88,$6B,$FE,$8B,$6B
	.dc.b	$FD,$4A,$5F,$5D,$DF,$B7,$C5,$18,$BE,$D1,$FC,$6F,$E2,$A4,$7F,$19
	.dc.b	$FA,$EB,$83,$83,$FC,$F3,$97,$F8,$72,$BF,$B7,$B6,$7C,$DE,$EF,$EF
	.dc.b	$23,$11,$FE,$9F,$F2,$BF,$96,$7C,$6F,$F2,$93,$F9,$57,$C6,$DC,$3F
	.dc.b	$D5,$3F,$5D,$F0,$AF,$C1,$1F,$EB,$D1,$F5,$DF,$CF,$1F,$F5,$54,$EF
	.dc.b	$3F,$F5,$14,$6B,$CF,$FE,$55,$3C,$EF,$F5,$ED,$78,$F1,$7F,$12,$8C
	.dc.b	$47,$FB,$2D,$E9,$5A,$3F,$D8,$A7,$8B,$E9,$BA,$D7,$F4,$3F,$AE,$FE
	.dc.b	$89,$E1,$DA,$FE,$84,$DC,$3F,$C7,$7D,$9F,$C1,$A7,$FA,$35,$70,$7F
	.dc.b	$A4,$47,$F4,$DF,$AE,$FE,$9C,$D8,$C9,$FD,$36,$32,$D3,$FA,$87,$CD
	.dc.b	$FF,$55,$FA,$EF,$EA,$CD,$8C,$9F,$D5,$63,$2D,$33,$87,$9A,$32,$D8
	.dc.b	$FF,$2C,$7D,$67,$F9,$A1,$FE,$D5,$F4,$BF,$E1,$9E,$0F,$ED,$D3,$9A
	.dc.b	$FF,$2C,$CE,$6B,$F8,$04,$E6,$B7,$F1,$3F,$DA,$FE,$0F,$FB,$D4,$E6
	.dc.b	$BF,$DF,$9C,$DF,$DC,$FE,$A7,$FB,$87,$C8,$FF,$71,$6C,$FF,$6F,$6E
	.dc.b	$7F,$39,$B9,$7F,$B4,$27,$FD,$86,$7D,$AE,$3F,$E5,$6E,$2B,$F1,$E7
	.dc.b	$FB,$DA,$33,$97,$F5,$69,$D4,$FF,$38,$6B,$FE,$FB,$6B,$F9,$78,$FE
	.dc.b	$F5,$1D,$6B,$F2,$BC,$CF,$9A,$CD,$7F,$82,$FE,$7F,$FC,$19,$ED,$5C
	.dc.b	$3F,$C1,$4D,$C7,$F1,$2D,$C7,$FC,$E1,$C9,$FE,$0B,$7F,$DD,$BF,$5C
	.dc.b	$F8,$7F,$F0,$DF,$CC,$7F,$7B,$C1,$39,$C7,$FC,$45,$39,$8C,$4F,$C2
	.dc.b	$37,$5C,$BE,$B7,$FC,$29,$3A,$C3,$FD,$F7,$E5,$7F,$A1,$F1,$FF,$16
	.dc.b	$7D,$77,$FB,$65,$63,$CE,$74,$FF,$A7,$27,$4E,$7E,$43,$FC,$07,$3A
	.dc.b	$8F,$EC,$AD,$79,$87,$0F,$F3,$48,$FF,$14,$47,$6B,$FF,$4A,$57,$F5
	.dc.b	$0A,$D5,$FE,$6B,$CB,$FF,$1F,$F4,$B3,$FC,$8C,$61,$B3,$B3,$FC,$79
	.dc.b	$E7,$E2,$FF,$D6,$5D,$3F,$E9,$49,$FD,$1F,$D7,$7E,$6B,$E3,$FA,$B6
	.dc.b	$3F,$53,$1F,$E4,$9F,$95,$C5,$F2,$F6,$BF,$CD,$FE,$B3,$1F,$F0,$26
	.dc.b	$1F,$E0,$6A,$FF,$32,$46,$1F,$F0,$4C,$EE,$B8,$CE,$77,$FD,$9C,$7F
	.dc.b	$97,$3E,$31,$CB,$FA,$E8,$C3,$31,$FF,$32,$4B,$E9,$71,$97,$FB,$12
	.dc.b	$AE,$1F,$EB,$FA,$D5,$F0,$9F,$F3,$97,$D6,$7F,$80,$B3,$34,$9E,$C7
	.dc.b	$B9,$FA,$91,$A8,$F3,$7D,$27,$27,$FD,$13,$F4,$8F,$A9,$FF,$49,$FC
	.dc.b	$82,$31,$97,$FA,$61,$BA,$FF,$AB,$DA,$FE,$22,$78,$57,$43,$E2,$7D
	.dc.b	$15,$BF,$F7,$35,$F2,$DB,$FD,$62,$D8,$FE,$27,$B2,$B8,$63,$FE,$E5
	.dc.b	$AD,$2C,$7B,$DF,$CC,$AD,$E1,$D3,$C4,$F5,$E0,$69,$2B,$D8,$FC,$23
	.dc.b	$ED,$74,$58,$FD,$6D,$7F,$AD,$BE,$BB,$F5,$FC,$0F,$A4,$FF,$26,$FB
	.dc.b	$8B,$CC,$7E,$2E,$AE,$BF,$AC,$AD,$3F,$62,$74,$BD,$6E,$8E,$EF,$F7
	.dc.b	$B1,$88,$FD,$AD,$E2,$BC,$35,$67,$FC,$BF,$3E,$07,$F8,$2D,$F8,$30
	.dc.b	$E5,$1F,$C2,$27,$4F,$FB,$B2,$AA,$8C,$D7,$1B,$F3,$DF,$28,$C5,$78
	.dc.b	$0F,$54,$F9,$1A,$67,$0E,$18,$79,$E3,$D6,$7C,$33,$E0,$14,$CD,$F0
	.dc.b	$70,$7A,$4B,$7E,$27,$77,$E1,$9F,$7C,$7A,$CD,$D7,$BB,$0F,$6B,$63
	.dc.b	$CE,$7C,$87,$CC,$F7,$C3,$C4,$74,$9E,$31,$E4,$3B,$57,$9C,$D3,$D0
	.dc.b	$F9,$C5,$3C,$93,$D2,$FF,$06,$D8,$F9,$11,$F2,$E3,$9E,$DF,$87,$9F
	.dc.b	$A0,$D7,$E5,$B5,$F8,$CE,$7E,$61,$E5,$FF,$93,$E7,$F5,$55,$FD,$02
	.dc.b	$73,$A7,$F4,$CB,$E1,$7B,$6C,$66,$B8,$4F,$37,$F2,$D4,$F2,$FF,$CD
	.dc.b	$18,$F5,$6E,$3F,$D6,$27,$BB,$F4,$1B,$5C,$A7,$72,$9E,$8F,$E2,$5D
	.dc.b	$D7,$8B,$0F,$D0,$46,$A7,$F1,$4D,$CB,$CB,$8F,$C1,$6D,$73,$C3,$BD
	.dc.b	$71,$C7,$38,$79,$70,$E1,$8F,$AE,$3B,$DF,$F6,$8C,$76,$49,$D4,$F8
	.dc.b	$A7,$51,$E7,$8F,$2D,$74,$7E,$24,$E9,$A7,$B2,$34,$FC,$37,$E8,$4F
	.dc.b	$6B,$F9,$13,$F1,$67,$2F,$CF,$78,$E1,$F0,$EB,$CB,$87,$9E,$B5,$8F
	.dc.b	$B1,$87,$99,$F1,$1F,$1D,$87,$C5,$7C,$07,$08,$F1,$C7,$13,$F7,$D9
	.dc.b	$7F,$4D,$F9,$6F,$F0,$7D,$7F,$42,$6E,$BF,$91,$E1,$46,$2D,$E6,$B9
	.dc.b	$1E,$CE,$72,$E9,$5F,$D5,$EF,$FB,$E3,$FD,$1F,$0E,$6E,$4F,$EB,$87
	.dc.b	$17,$FC,$42,$FF,$4F,$3C,$18,$77,$FB,$9D,$44,$3F,$AD,$9A,$C5,$19
	.dc.b	$8E,$46,$8F,$F0,$0B,$97,$FB,$45,$79,$9F,$31,$E8,$C3,$FB,$93,$EA
	.dc.b	$EA,$FE,$FF,$F5,$FF,$0D,$B1,$FD,$5B,$4F,$F8,$01,$B2,$BC,$9E,$0E
	.dc.b	$E3,$15,$FE,$1F,$BE,$2F,$F8,$7A,$32,$DA,$F9,$3A,$7C,$FE,$FF,$71
	.dc.b	$27,$FC,$37,$E5,$57,$FC,$FB,$3B,$7F,$C3,$F3,$F1,$67,$C5,$F5,$F2
	.dc.b	$7D,$03,$1F,$F0,$DC,$73,$E2,$FE,$CF,$18,$9F,$46,$4A,$F8,$8D,$8C
	.dc.b	$C6,$3D,$AF,$80,$48,$EF,$93,$CD,$7C,$8E,$67,$FD,$57,$F3,$3F,$80
	.dc.b	$F2,$CE,$74,$F1,$E7,$FD,$CD,$BD,$1F,$34,$FC,$3E,$3F,$CC,$E3,$38
	.dc.b	$7F,$AE,$BB,$BA,$1B,$7E,$08,$FA,$FF,$F1,$93,$86,$BF,$E8,$76,$CF
	.dc.b	$F8,$2B,$47,$D7,$B8,$BF,$4D,$7C,$FC,$4F,$C6,$61,$CA,$3F,$3A,$D7
	.dc.b	$FC,$FE,$D7,$FD,$16,$F6,$E6,$FF,$62,$DF,$E4,$A7,$EE,$B7,$1F,$F0
	.dc.b	$DC,$77,$3D,$7F,$5D,$FE,$B1,$6F,$AD,$E4,$8C,$47,$E1,$BE,$BD,$ED
	.dc.b	$E6,$BF,$5F,$1F,$D5,$A7,$EB,$B5,$FF,$54,$57,$2B,$F8,$5F,$5C,$F8
	.dc.b	$6F,$A3,$DB,$2F,$EF,$CE,$5B,$39,$7F,$2F,$C5,$FE,$4B,$72,$7D,$4F
	.dc.b	$8C,$7C,$87,$CC,$6C,$66,$3E,$7F,$B5,$F0,$F8,$3F,$D5,$35,$AA,$FF
	.dc.b	$46,$35,$FE,$1A,$D7,$75,$FC,$06,$38,$67,$F3,$2F,$FD,$0D,$5F,$86
	.dc.b	$6B,$FE,$C9,$9D,$DF,$FB,$EB,$27,$C7,$FF,$18,$5F,$4D,$4C,$44,$BF
	.dc.b	$8B,$19,$7E,$3C,$7F,$0D,$FA,$27,$C4,$52,$4F,$CC,$F8,$2A,$FF,$00
	.dc.b	$EE,$FE,$53,$AF,$E5,$5B,$9F,$FB,$5B,$C2,$63,$D2,$FE,$AF,$D1,$F9
	.dc.b	$FD,$F9,$18,$21,$FC,$33,$64,$6A,$31,$3E,$F3,$19,$7E,$12,$71,$3F
	.dc.b	$82,$FA,$1F,$89,$F4,$52,$30,$43,$B3,$97,$C1,$1F,$2F,$DA,$F9,$9F
	.dc.b	$41,$27,$F1,$B1,$97,$F1,$F1,$97,$D1,$1F,$B4,$F8,$45,$D2,$4F,$F0
	.dc.b	$2F,$AF,$FC,$92,$CB,$F5,$B5,$C2,$3F,$AD,$B5,$FF,$6C,$35,$FE,$B2
	.dc.b	$D7,$FD,$F6,$EE,$3F,$6F,$3F,$B0,$9F,$07,$77,$89,$F9,$DE,$E7,$E2
	.dc.b	$48,$FE,$AB,$DD,$FE,$1B,$F5,$FF,$C0,$E3,$FB,$3B,$51,$1C,$37,$D1
	.dc.b	$FE,$97,$C2,$F6,$FF,$2C,$49,$FD,$BF,$BB,$FC,$6B,$C4,$7C,$0F,$F1
	.dc.b	$CF,$AF,$7A,$34,$D4,$5B,$FE,$FF,$3A,$6C,$FF,$6D,$8C,$47,$FB,$FE
	.dc.b	$BF,$C0,$6F,$CF,$EA,$F4,$A8,$CF,$0F,$F1,$E6,$CE,$AB,$9C,$FF,$22
	.dc.b	$F4,$FF,$39,$7C,$CF,$86,$E2,$35,$DD,$FC,$DB,$37,$C8,$7F,$38,$AE
	.dc.b	$37,$FE,$12,$6B,$E4,$6B,$D2,$BF,$C1,$11,$D7,$1F,$4D,$BE,$4B,$B1
	.dc.b	$D6,$BF,$A2,$FD,$7B,$8E,$93,$1F,$EA,$7F,$5F,$F9,$2D,$3F,$9E,$D6
	.dc.b	$E7,$FC,$56,$D7,$FC,$4E,$1C,$3F,$E2,$A6,$BF,$31,$E1,$1F,$D2,$27
	.dc.b	$FB,$8C,$F8,$3D,$6F,$13,$F2,$47,$D7,$3E,$37,$FA,$EB,$63,$FB,$4B
	.dc.b	$5F,$A1,$F3,$DF,$C3,$BE,$87,$75,$FE,$2D,$AD,$57,$F8,$F9,$AF,$E0
	.dc.b	$1A,$FE,$69,$E9,$8F,$FB,$E2,$DF,$5E,$E1,$FF,$2A,$46,$1E,$F7,$E4
	.dc.b	$F8,$4F,$39,$1F,$F0,$FC,$FF,$22,$F8,$2C,$47,$CA,$C7,$FA,$F3,$E5
	.dc.b	$AF,$41,$1E,$3F,$5E,$FA,$FF,$0B,$63,$C0,$7F,$83,$3E,$37,$FA,$0B
	.dc.b	$7C,$FF,$F0,$2C,$62,$3E,$77,$B4,$FB,$3C,$D7,$F9,$4E,$B5,$5F,$E6
	.dc.b	$46,$BF,$A6,$6B,$FC,$01,$E9,$BF,$C7,$FB,$0F,$EA,$73,$F0,$1A,$FD
	.dc.b	$D9,$DF,$57,$3F,$4B,$CA,$49,$F8,$FF,$66,$F8,$6C,$47,$E4,$2F,$9B
	.dc.b	$E5,$AF,$41,$FC,$3B,$E2,$7F,$D1,$1F,$56,$FB,$17,$AF,$86,$DF,$F8
	.dc.b	$1A,$BF,$B4,$19,$99,$19,$3F,$CA,$FE,$B3,$E3,$7F,$D2,$31,$86,$BF
	.dc.b	$C0,$63,$55,$FE,$85,$F2,$FF,$CF,$15,$AE,$0F,$F2,$EB,$77,$FE,$CC
	.dc.b	$DC,$3F,$D3,$FF,$19,$FE,$0D,$5E,$FF,$F6,$AF,$B1,$FF,$50,$7B,$46
	.dc.b	$9F,$E9,$3C,$2D,$7F,$9B,$AB,$83,$6B,$83,$FE,$64,$BF,$F1,$DE,$2B
	.dc.b	$FE,$77,$7C,$2B,$FD,$4A,$D7,$F9,$3B,$5D,$DF,$47,$6F,$F2,$69,$CF
	.dc.b	$99,$E8,$3F,$D5,$BF,$66,$FB,$3C,$5F,$96,$6B,$FE,$8C,$9D,$3F,$E6
	.dc.b	$8F,$B3,$FC,$26,$3F,$EB,$E6,$C7,$F1,$46,$BF,$EB,$C6,$BF,$CF,$BE
	.dc.b	$47,$F8,$67,$85,$FE,$C6,$6C,$7F,$1C,$6B,$FE,$C4,$6B,$FD,$1A,$7F
	.dc.b	$81,$7C,$5F,$87,$C4,$34,$93,$F2,$3E,$CF,$FD,$B7,$AF,$D3,$6C,$7F
	.dc.b	$AA,$38,$3F,$C0,$B3,$9F,$17,$E9,$4F,$F4,$BD,$17,$C3,$87,$31,$8A
	.dc.b	$DC,$7F,$A1,$70,$E7,$7C,$6F,$86,$1F,$71,$FF,$4E,$BE,$3F,$8F,$E2
	.dc.b	$FF,$7A,$F8,$DF,$F7,$77,$C0,$62,$FF,$23,$7C,$1E,$97,$9C,$FC,$AE
	.dc.b	$EF,$8E,$1D,$49,$5C,$1F,$C8,$61,$C7,$0E,$99,$7F,$B0,$3E,$CF,$FB
	.dc.b	$FD,$7E,$C1,$B7,$F1,$FE,$07,$C9,$D6,$9C,$DF,$1E,$5C,$C6,$30,$F6
	.dc.b	$6C,$C3,$35,$E7,$F8,$8F,$F3,$FD,$7F,$B5,$3B,$DF,$66,$F8,$8D,$7C
	.dc.b	$4F,$C3,$9C,$C6,$2B,$85,$6D,$FF,$67,$DF,$4D,$BF,$C0,$97,$C2,$7F
	.dc.b	$D8,$3F,$63,$FB,$96,$F7,$7C,$FA,$7D,$9F,$FB,$5A,$72,$48,$FE,$CD
	.dc.b	$5C,$5F,$ED,$13,$B7,$35,$CE,$5E,$9D,$3B,$7D,$97,$FA,$A7,$0F,$1C
	.dc.b	$65,$A7,$FB,$5A,$35,$7E,$09,$D3,$67,$5F,$1B,$8E,$3F,$F6,$2E,$1F
	.dc.b	$E4,$49,$F7,$D7,$B7,$EC,$BE,$25,$6A,$3C,$C4,$9F,$F7,$2C,$E9,$FF
	.dc.b	$1A,$D7,$39,$F8,$26,$CF,$F0,$44,$ED,$FF,$29,$C6,$EB,$FD,$AF,$F2
	.dc.b	$79,$9F,$F7,$47,$0F,$8F,$E9,$FE,$84,$E4,$91,$9C,$7F,$DE,$F1,$86
	.dc.b	$C6,$6B,$FC,$31,$3D,$7E,$C1,$F3,$DF,$EE,$69,$FF,$26,$C6,$2F,$FC
	.dc.b	$53,$08,$C3,$FE,$2C,$6B,$FE,$54,$35,$3F,$85,$8D,$E3,$FC,$1B,$A5
	.dc.b	$F2,$5F,$81,$8F,$F7,$AC,$EB,$0F,$0D,$F1,$3F,$53,$5B,$8C,$9F,$E7
	.dc.b	$AA,$FF,$7E,$FD,$7B,$97,$F5,$FF,$60,$FA,$2F,$CB,$4E,$6F,$FB,$2C
	.dc.b	$E5,$E4,$7F,$9B,$23,$5A,$7F,$8F,$5B,$87,$F1,$EF,$A5,$7D,$87,$FA
	.dc.b	$16,$F8,$DF,$FA,$32,$F7,$1C,$F2,$F0,$9F,$C5,$6F,$8F,$C6,$62,$78
	.dc.b	$F8,$EF,$AF,$7D,$6F,$F1,$B8,$C3,$FA,$FF,$47,$F3,$BA,$FF,$30,$9C
	.dc.b	$DF,$8B,$4F,$F0,$07,$4E,$CE,$58,$BE,$C3,$15,$E0,$8C,$3C,$71,$EA
	.dc.b	$FF,$A9,$B0,$F2,$57,$0C,$3F,$D1,$1E,$47,$CF,$F8,$AF,$E2,$87,$F1
	.dc.b	$5F,$B0,$B8,$73,$F5,$BD,$07,$F1,$9F,$8D,$B1,$FC,$41,$C3,$FC,$4B
	.dc.b	$E3,$2E,$3F,$E6,$DF,$B0,$F8,$73,$ED,$6C,$7F,$A6,$61,$F5,$AF,$A9
	.dc.b	$FE,$29,$5A,$E2,$FE,$AB,$68,$C5,$FF,$AA,$9B,$1F,$E0,$BB,$E5,$3D
	.dc.b	$F8,$F9,$91,$FB,$E7,$83,$0C,$57,$C4,$E6,$7D,$87,$F3,$7F,$0B,$F9
	.dc.b	$C3,$6F,$F9,$7B,$6B,$83,$73,$FE,$D7,$C3,$F9,$BC,$3F,$A0,$7D,$87
	.dc.b	$F4,$16,$E1,$FC,$E1,$B8,$78,$5B,$7C,$97,$CC,$FF,$49,$FB,$0F,$E9
	.dc.b	$4D,$BF,$E8,$2D,$BF,$53,$6F,$93,$67,$FB,$85,$F4,$7F,$A4,$DF,$F5
	.dc.b	$EF,$AD,$FF,$2C,$C7,$F9,$49,$B8,$7C,$5F,$90,$D7,$C1,$4A,$D3,$FD
	.dc.b	$76,$31,$3F,$2F,$EB,$3F,$B0,$7C,$7F,$F6,$16,$DF,$F5,$86,$CE,$E3
	.dc.b	$FA,$DC,$FF,$55,$7F,$D6,$87,$E9,$4B,$1F,$89,$E3,$FF,$75,$46,$6F
	.dc.b	$81,$1F,$BF,$1A,$9F,$F6,$04,$72,$D7,$F1,$AD,$3F,$DE,$A5,$8F,$C6
	.dc.b	$7C,$C7,$37,$F0,$DF,$3D,$FD,$EF,$D5,$FE,$F8,$DA,$FE,$E6,$D9,$DC
	.dc.b	$7F,$76,$9F,$EE,$1F,$61,$FE,$BD,$8F,$E0,$0D,$9F,$DA,$FC,$C3,$EB
	.dc.b	$F7,$F0,$52,$F6,$FF,$88,$67,$51,$83,$F8,$57,$8E,$FA,$3F,$EF,$CD
	.dc.b	$8F,$E3,$FD,$0F,$B0,$FE,$71,$90,$56,$DF,$F7,$33,$77,$FD,$9B,$6F
	.dc.b	$FC,$75,$5C,$AB,$8E,$1E,$02,$4F,$82,$75,$5C,$69,$E8,$B9,$BF,$9A
	.dc.b	$47,$FA,$75,$B8,$FE,$8F,$CE,$FC,$B7,$97,$32,$FC,$B3,$FE,$2D,$8C
	.dc.b	$5F,$1F,$8F,$FC,$E1,$FE,$31,$F8,$FB,$7F,$9D,$8B,$3F,$C0,$66,$E1
	.dc.b	$FD,$7F,$17,$A9,$FA,$53,$FC,$A1,$F0,$0E,$1C,$65,$FA,$79,$B8,$FF
	.dc.b	$9B,$E6,$E1,$FE,$67,$78,$9C,$BE,$92,$DF,$FA,$96,$2D,$7F,$9D,$E6
	.dc.b	$D7,$F7,$1F,$AC,$FA,$FB,$9F,$B3,$C6,$CD,$F4,$7F,$E5,$89,$D9,$3D
	.dc.b	$7F,$4C,$65,$FF,$36,$7C,$83,$E5,$E9,$EE,$F9,$FB,$BF,$FA,$F6,$74
	.dc.b	$D9,$D3,$49,$B7,$FA,$2E,$76,$FF,$A4,$6B,$9B,$9A,$BA,$FF,$A2,$E6
	.dc.b	$DF,$FB,$12,$FF,$9D,$DF,$3F,$2B,$E8,$F3,$7A,$19,$5E,$57,$2D,$3F
	.dc.b	$B6,$4F,$E7,$7E,$C3,$F3,$DA,$BE,$B7,$FC,$41,$EA,$FF,$A8,$A3,$15
	.dc.b	$FE,$23,$AD,$D6,$AE,$CE,$4F,$A5,$19,$69,$9A,$E6,$FF,$73,$9E,$2F
	.dc.b	$EB,$C5,$61,$FF,$5E,$C5,$72,$74,$39,$14,$F6,$16,$FD,$57,$FE,$3B
	.dc.b	$8E,$37,$E2,$7A,$9C,$4C,$C7,$A6,$35,$18,$3C,$D3,$D1,$F6,$4E,$CE
	.dc.b	$D7,$E9,$9D,$4F,$B6,$31,$7E,$B6,$C7,$85,$B7,$E0,$6D,$F4,$9F,$65
	.dc.b	$71,$69,$C4,$B5,$C5,$B3,$C6,$3E,$05,$F9,$27,$E3,$C6,$2B,$A3,$63
	.dc.b	$A3,$4E,$9F,$1A,$F5,$FC,$8D,$9F,$C1,$35,$F8,$01,$C7,$C8,$3C,$87
	.dc.b	$07,$F2,$F9,$CC,$62,$B8,$5F,$E1,$6F,$85,$EE,$31,$5B,$BE,$33,$F8
	.dc.b	$CA,$79,$3F,$8D,$C7,$F2,$31,$8C,$75,$87,$79,$CB,$74,$F0,$5F,$16
	.dc.b	$D7,$0C,$23,$F9,$3C,$3A,$36,$B8,$3F,$D2,$E3,$D2,$EE,$FB,$39,$9D
	.dc.b	$A9,$E7,$87,$92,$FB,$BE,$AA,$DC,$EB,$0F,$87,$5A,$7C,$F7,$E3,$30
	.dc.b	$78,$F0,$F0,$BD,$EF,$99,$D2,$33,$7E,$8C,$3C,$15,$F3,$6B,$58,$74
	.dc.b	$8E,$46,$4C,$57,$1C,$3C,$77,$CA,$7E,$15,$6A,$FC,$31,$C3,$0F,$0C
	.dc.b	$62,$36,$7F,$0C,$C5,$C0,$F8,$FF,$E1,$58,$35,$FE,$51,$18,$6D,$FF
	.dc.b	$80,$39,$9E,$B6,$63,$F2,$D8,$73,$7F,$9F,$5E,$71,$EE,$EF,$A3,$FA
	.dc.b	$75,$E6,$FF,$D9,$19,$78,$70,$E6,$64,$E1,$8B,$EC,$31,$1F,$D2,$CA
	.dc.b	$7F,$A5,$23,$4E,$48,$7F,$4B,$39,$C7,$09,$F0,$65,$F1,$27,$19,$FF
	.dc.b	$86,$F2,$FE,$C9,$7E,$FE,$AF,$ED,$1F,$61,$78,$0D,$3C,$FE,$0D,$C7
	.dc.b	$FD,$8B,$5C,$1B,$5C,$1F,$99,$3E,$EE,$F9,$1F,$3D,$F8,$3E,$36,$23
	.dc.b	$DD,$17,$5F,$F4,$3E,$CA,$F2,$7D,$84,$BE,$EE,$F6,$FF,$1B,$67,$FE
	.dc.b	$01,$8D,$4F,$F0,$C9,$CE,$69,$E1,$3F,$E2,$E8,$7C,$6F,$C1,$24,$7F
	.dc.b	$91,$3D,$27,$D1,$62,$3A,$1C,$67,$1A,$7F,$A5,$67,$E5,$74,$4A,$F8
	.dc.b	$AF,$E1,$EF,$F9,$25,$7F,$86,$63,$53,$D3,$3F,$EB,$93,$FE,$9F,$EA
	.dc.b	$FC,$A4,$BE,$C3,$FC,$28,$FF,$B9,$5A,$F9,$0D,$78,$3F,$95,$FB,$0F
	.dc.b	$F1,$BC,$6E,$39,$CF,$0B,$FF,$65,$B7,$7F,$C3,$5F,$0B,$E9,$39,$7F
	.dc.b	$CC,$8D,$7B,$9F,$E6,$A8,$E1,$5F,$ED,$6F,$AF,$7A,$5E,$A6,$D7,$FB
	.dc.b	$F5,$AF,$FB,$CB,$EC,$3F,$D7,$AE,$67,$81,$1F,$F3,$BC,$69,$FF,$7C
	.dc.b	$FD,$5B,$E1,$7E,$77,$64,$9F,$F3,$EC,$E5,$B3,$97,$FD,$BD,$3F,$BB
	.dc.b	$9D,$47,$CB,$F8,$97,$C5,$FD,$26,$C7,$FA,$17,$3F,$F7,$9C,$62,$3E
	.dc.b	$97,$AD,$FB,$17,$95,$71,$23,$C7,$7F,$F1,$04,$EE,$F9,$D6,$AB,$89
	.dc.b	$B3,$FD,$29,$3B,$AF,$DE,$60,$F9,$FF,$57,$D1,$4B,$F8,$31,$ED,$30
	.dc.b	$FC,$88,$C3,$FB,$CF,$9F,$7C,$2F,$F0,$56,$C6,$63,$E3,$7B,$D9,$64
	.dc.b	$64,$F9,$F1,$F2,$FE,$BF,$E6,$7B,$52,$7E,$9C,$65,$FA,$B1,$97,$ED
	.dc.b	$C6,$5F,$BB,$19,$DB,$F9,$67,$8C,$D1,$7F,$67,$E6,$FE,$D7,$D7,$CC
	.dc.b	$7F,$03,$19,$7F,$07,$19,$7F,$13,$1F,$81,$F5,$BF,$05,$F3,$52,$7F
	.dc.b	$13,$19,$7F,$17,$19,$7F,$23,$1F,$B6,$F8,$05,$D2,$4F,$D9,$1F,$5F
	.dc.b	$FE,$23,$71,$3C,$1B,$3C,$1F,$DE,$4E,$DF,$DF,$4E,$DB,$3B,$D3,$FA
	.dc.b	$84,$76,$E7,$FC,$E7,$B3,$23,$F4,$6F,$E7,$FD,$AF,$E1,$BF,$5F,$F8
	.dc.b	$B9,$FF,$70,$35,$39,$E0,$FA,$15,$FA,$7F,$66,$57,$EB,$9F,$D5,$7B
	.dc.b	$5F,$C6,$3B,$DF,$5F,$FC,$6B,$EB,$FF,$97,$DF,$C3,$6B,$FC,$66,$76
	.dc.b	$E7,$CB,$E0,$47,$E5,$46,$9E,$1E,$B7,$72,$3C,$4F,$A1,$E7,$FE,$69
	.dc.b	$FE,$4C,$DE,$0F,$8E,$DE,$FF,$F4,$96,$6F,$8E,$FE,$63,$39,$B1,$FD
	.dc.b	$57,$B7,$27,$F3,$6A,$E1,$1E,$08,$C7,$C8,$48,$C1,$F8,$3F,$AF,$FF
	.dc.b	$74,$37,$93,$F9,$FC,$62,$3F,$9E,$7D,$7F,$E0,$74,$FB,$B3,$96,$D7
	.dc.b	$03,$FA,$3C,$EF,$4F,$F5,$65,$73,$BF,$F5,$70,$70,$EA,$BF,$D4,$CF
	.dc.b	$BA,$32,$FC,$0F,$AE,$7C,$5F,$F5,$46,$C7,$E7,$5A,$FF,$66,$FA,$67
	.dc.b	$88,$F9,$DE,$17,$2B,$83,$FE,$31,$AD,$CE,$67,$FA,$E5,$F4,$D3,$F3
	.dc.b	$AD,$CF,$E9,$DF,$0A,$F1,$7C,$07,$94,$8F,$F2,$98,$F1,$F1,$E2,$3A
	.dc.b	$CD,$C3,$FC,$71,$5A,$9E,$64,$36,$73,$CF,$FC,$8C,$FF,$71,$F8,$77
	.dc.b	$91,$FD,$E1,$B1,$FC,$05,$AF,$F7,$7F,$49,$F5,$9B,$A9,$3B,$7F,$C0
	.dc.b	$33,$A8,$C1,$0F,$04,$6A,$7C,$1E,$5C,$9F,$03,$F3,$7E,$AD,$EF,$66
	.dc.b	$7D,$13,$74,$F8,$35,$A9,$FE,$D3,$F0,$5F,$E6,$5F,$A3,$7D,$5B,$D0
	.dc.b	$C3,$6B,$64,$3F,$C4,$F3,$C2,$76,$7F,$6F,$AF,$F2,$97,$B3,$FE,$72
	.dc.b	$AF,$F2,$6B,$5F,$F1,$23,$71,$FF,$22,$FC,$27,$0A,$95,$C1,$FF,$36
	.dc.b	$C7,$F9,$DF,$85,$F4,$74,$95,$FE,$81,$E6,$AF,$F8,$F3,$CA,$FC,$BC
	.dc.b	$E3,$1F,$F2,$9D,$53,$FD,$AD,$7D,$2F,$FB,$A5,$F0,$7A,$18,$8F,$E3
	.dc.b	$B5,$FE,$74,$6D,$72,$F1,$3F,$D1,$5F,$54,$FA,$BF,$44,$4C,$39,$3A
	.dc.b	$8F,$F2,$1F,$D5,$FD,$AC,$BF,$6E,$DA,$FD,$E8,$70,$FF,$00,$6B,$FE
	.dc.b	$EB,$EE,$75,$7F,$00,$F8,$AF,$DE,$B6,$33,$3D,$27,$2F,$6A,$FF,$3B
	.dc.b	$7C,$5F,$8B,$C7,$34,$93,$FC,$43,$17,$C0,$FF,$78,$DB,$FC,$39,$E7
	.dc.b	$B7,$7D,$77,$FA,$A7,$8B,$FD,$37,$7B,$C7,$FC,$E3,$7B,$AD,$61,$FC
	.dc.b	$D3,$1E,$BE,$C6,$2D,$7B,$AF,$AF,$87,$3C,$7F,$C1,$79,$78,$27,$FD
	.dc.b	$3B,$93,$ED,$B9,$60,$FC,$E5,$F3,$BD,$72,$F3,$D6,$A7,$38,$72,$8C
	.dc.b	$56,$F0,$E9,$3F,$02,$B7,$9F,$FA,$06,$72,$FB,$A3,$39,$FF,$B2,$F1
	.dc.b	$EF,$7C,$70,$F1,$7E,$25,$F4,$23,$E5,$E8,$F9,$3C,$F0,$7A,$A3,$13
	.dc.b	$FE,$BA,$AE,$19,$FC,$C9,$CB,$FE,$93,$8C,$E5,$F1,$B0,$F8,$7E,$F7
	.dc.b	$FB,$B7,$0F,$F3,$6F,$5E,$1C,$B5,$C9,$FE,$72,$32,$7F,$88,$FE,$39
	.dc.b	$F1,$5F,$5F,$A5,$EC,$3E,$8B,$ED,$6A,$BD,$E1,$FC,$EE,$B4,$F3,$C3
	.dc.b	$94,$62,$BF,$D4,$95,$C1,$FE,$A9,$19,$D3,$FC,$4F,$2F,$9C,$E3,$F2
	.dc.b	$BE,$08,$C6,$3D,$E3,$13,$A8,$F5,$37,$0E,$ED,$BE,$AD,$9F,$F0,$4C
	.dc.b	$BE,$FD,$F0,$5F,$97,$E3,$7E,$00,$F4,$3E,$0D,$6B,$0F,$F4,$15,$6A
	.dc.b	$73,$7C,$6F,$85,$7F,$93,$EF,$86,$1F,$ED,$8C,$38,$4E,$6F,$86,$1C
	.dc.b	$A3,$AE,$1C,$AF,$FD,$99,$F5,$3F,$84,$CF,$F6,$9F,$6D,$FE,$8F,$BF
	.dc.b	$F2,$E4,$E5,$FF,$52,$46,$6B,$8F,$2F,$EC,$ED,$39,$7F,$CA,$F7,$C6
	.dc.b	$91,$AF,$3B,$F0,$33,$F9,$49,$FF,$0C,$B5,$F9,$6D,$7F,$C6,$AD,$7F
	.dc.b	$80,$FC,$BF,$EE,$BF,$69,$FE,$5D,$97,$F9,$CF,$C0,$4E,$7C,$7F,$C7
	.dc.b	$C6,$27,$F8,$AD,$70,$BF,$F7,$74,$E5,$FE,$27,$19,$CB,$FC,$E3,$39
	.dc.b	$6E,$1E,$1F,$03,$AB,$DB,$D1,$FE,$1B,$F2,$DF,$E0,$8C,$BC,$91,$8F
	.dc.b	$53,$51,$FC,$9A,$75,$9F,$8A,$B5,$3C,$09,$1F,$C9,$BC,$DF,$C0,$E5
	.dc.b	$E4,$AF,$85,$C3,$2F,$FD,$29,$F3,$4F,$85,$C6,$3F,$4A,$7F,$2F,$39
	.dc.b	$7E,$1D,$7E,$D2,$31,$3F,$0A,$3E,$77,$AD,$9B,$B7,$D9,$FA,$5F,$33
	.dc.b	$23,$D9,$C9,$70,$F0,$1F,$E6,$EC,$5F,$05,$F1,$A7,$C0,$DA,$F8,$9E
	.dc.b	$AF,$CA,$3E,$F7,$97,$63,$FD,$61,$E3,$FF,$3D,$C7,$FA,$4F,$8F,$F5
	.dc.b	$67,$F6,$73,$9E,$0F,$F6,$6C,$E6,$BF,$CF,$F3,$9A,$FF,$77,$F2,$7F
	.dc.b	$14,$8C,$4F,$F4,$EA,$E0,$A3,$3C,$DF,$ED,$7A,$E1,$CD,$FC,$A9,$FC
	.dc.b	$A7,$B3,$F9,$5F,$1A,$E5,$F9,$06,$E3,$F8,$E6,$E7,$FE,$E2,$E9,$FE
	.dc.b	$35,$5F,$E9,$4C,$91,$FE,$20,$8F,$CC,$F4,$7E,$87,$A3,$F4,$4D,$8C
	.dc.b	$9F,$9E,$8C,$BF,$A0,$F4,$FF,$51,$EE,$FE,$A5,$B1,$93,$F4,$B1,$96
	.dc.b	$99,$CF,$F4,$B1,$96,$C7,$FA,$AB,$2F,$EB,$9E,$C7,$F0,$F7,$E8,$F5
	.dc.b	$FF,$6D,$8C,$4F,$B7,$8B,$FC,$B2,$F1,$3F,$D6,$7E,$72,$77,$87,$D3
	.dc.b	$BE,$5E,$17,$F5,$87,$F8,$17,$CB,$7F,$03,$F8,$5B,$8F,$EF,$9B,$7F
	.dc.b	$BC,$6E,$7F,$3D,$FC,$07,$CF,$FF,$83,$69,$F4,$28,$CC,$79,$6C,$EA
	.dc.b	$33,$1F,$9E,$8C,$1F,$96,$8E,$93,$CE,$30,$73,$37,$3F,$A6,$3F,$1D
	.dc.b	$3C,$0F,$E1,$7F,$1D,$FC,$33,$E3,$ED,$7F,$09,$6D,$7F,$8A,$B8,$D1
	.dc.b	$8C,$BC,$F3,$FE,$28,$9C,$D7,$AE,$5F,$B9,$BE,$27,$FC,$4B,$FB,$53
	.dc.b	$E7,$BF,$9F,$4E,$6F,$FC,$7B,$39,$79,$46,$71,$FF,$1D,$CE,$5F,$E0
	.dc.b	$38,$76,$C3,$AE,$1F,$E0,$2F,$6F,$F9,$0C,$F9,$B1,$FA,$3F,$6B,$9F
	.dc.b	$2B,$E4,$B6,$3E,$47,$02,$31,$97,$D1,$8C,$4F,$D7,$86,$1D,$A3,$FD
	.dc.b	$FE,$76,$34,$E6,$3E,$EC,$F2,$7E,$A7,$E9,$1F,$CC,$2B,$F6,$9F,$A6
	.dc.b	$BE,$23,$FC,$37,$39,$AF,$F2,$AC,$E6,$BC,$53,$9E,$F4,$E6,$BA,$FA
	.dc.b	$5F,$E5,$48,$C6,$BF,$E6,$7E,$1F,$F0,$BD,$7C,$CF,$97,$FA,$7E,$B7
	.dc.b	$D6,$6E,$7F,$51,$B3,$B8,$FC,$C7,$6F,$EA,$F7,$FF,$24,$79,$FF,$B3
	.dc.b	$6C,$EB,$BB,$E3,$BF,$E5,$09,$D3,$4F,$F2,$9F,$C5,$7F,$94,$4F,$F2
	.dc.b	$97,$6D,$8C,$93,$1F,$F2,$94,$65,$FE,$87,$8F,$D4,$E5,$FF,$2D,$79
	.dc.b	$FF,$E7,$1C,$FE,$97,$17,$FA,$26,$31,$3A,$C7,$D7,$5A,$F7,$95,$FE
	.dc.b	$85,$8F,$F5,$0F,$B9,$FE,$65,$F3,$71,$EE,$FF,$91,$F9,$3F,$D0,$B1
	.dc.b	$89,$D6,$3F,$EC,$0A,$D3,$FC,$C7,$1F,$F3,$CD,$FE,$AB,$0E,$B3,$FE
	.dc.b	$74,$C3,$93,$4E,$58,$7E,$C3,$5F,$F2,$2D,$70,$7F,$D5,$18,$FF,$96
	.dc.b	$5C,$B5,$FF,$50,$2F,$85,$FF,$3F,$7D,$57,$A9,$B8,$79,$DB,$5F,$ED
	.dc.b	$FA,$E1,$3F,$E8,$8A,$FF,$37,$CF,$F7,$11,$F4,$2E,$1F,$F2,$CC,$7F
	.dc.b	$A2,$23,$0E,$3D,$53,$C2,$77,$EE,$7F,$0C,$AD,$CF,$F8,$30,$D1,$FE
	.dc.b	$64,$8D,$4F,$FB,$6A,$B9,$D7,$0A,$F0,$4E,$5E,$27,$82,$B8,$BE,$33
	.dc.b	$27,$03,$13,$CE,$37,$1E,$38,$C6,$1E,$0A,$ED,$18,$9E,$62,$F4,$FA
	.dc.b	$6A,$BC,$EB,$E0,$C6,$23,$E6,$57,$33,$D8,$F3,$3C,$F7,$E4,$77,$5C
	.dc.b	$9F,$0D,$F3,$9D,$47,$1A,$E9,$5E,$83,$83,$DA,$B8,$5F,$8E,$B8,$56
	.dc.b	$8E,$0F,$AC,$EE,$79,$23,$84,$F2,$8C,$D7,$81,$F9,$B5,$CD,$FA,$59
	.dc.b	$BE,$2B,$EC,$36,$33,$18,$3F,$D4,$BF,$11,$FC,$9F,$83,$FD,$F9,$18
	.dc.b	$EF,$FF,$42,$D7,$F3,$68,$C4,$EA,$3F,$23,$5C,$5D,$18,$AE,$38,$FE
	.dc.b	$4A,$7F,$AC,$4F,$D8,$6B,$F8,$C7,$94,$6E,$78,$E3,$E6,$F0,$BF,$D7
	.dc.b	$73,$CD,$B1,$E4,$BE,$55,$C1,$F2,$CE,$9A,$78,$63,$4A,$FC,$35,$DE
	.dc.b	$3C,$D5,$C5,$D5,$70,$AD,$C7,$43,$D5,$3B,$8C,$C7,$0C,$39,$B0,$F3
	.dc.b	$BE,$33,$18,$7B,$63,$06,$8E,$A6,$B0,$F8,$11,$F1,$E7,$13,$E3,$8C
	.dc.b	$3C,$5E,$8F,$23,$13,$C8,$F8,$0E,$63,$89,$23,$2F,$23,$11,$FD,$EB
	.dc.b	$77,$5F,$FB,$C2,$FF,$C3,$BC,$EE,$57,$A3,$9E,$1F,$E7,$4D,$C7,$96
	.dc.b	$17,$81,$97,$49,$FF,$1F,$63,$CE,$F8,$36,$3F,$A3,$5F,$1A,$FF,$7E
	.dc.b	$47,$F2,$A3,$CD,$E6,$56,$99,$C3,$BC,$65,$B1,$E0,$C9,$E9,$78,$1B
	.dc.b	$1F,$D5,$6B,$C0,$66,$64,$64,$87,$FA,$C3,$1F,$89,$1A,$8F,$E8,$D3
	.dc.b	$F0,$FA,$FF,$C4,$BE,$C7,$7F,$26,$57,$F6,$E8,$D6,$47,$31,$D7,$25
	.dc.b	$6A,$7E,$3F,$C2,$BE,$46,$92,$FE,$6C,$3E,$B3,$11,$E7,$8E,$0E,$A7
	.dc.b	$9C,$F8,$BE,$6B,$E2,$7A,$D2,$BF,$C0,$71,$97,$E5,$46,$5F,$F0,$7E
	.dc.b	$69,$FC,$17,$23,$E7,$5F,$33,$F7,$BD,$0F,$B7,$87,$F9,$6B,$87,$75
	.dc.b	$A8,$FF,$34,$35,$D1,$FE,$64,$8D,$D7,$E3,$E1,$E2,$FF,$12,$C3,$F2
	.dc.b	$FF,$3C,$F9,$7F,$C6,$B6,$BF,$CB,$15,$BE,$AF,$EE,$31,$CB,$BB,$77
	.dc.b	$23,$71,$DF,$F3,$E7,$B5,$E1,$FC,$8F,$F1,$35,$FE,$8D,$6E,$7F,$E1
	.dc.b	$5A,$DC,$FF,$99,$DA,$EF,$AB,$91,$D6,$E4,$7F,$9B,$6F,$9E,$5E,$66
	.dc.b	$C7,$F8,$22,$F9,$E1,$FE,$92,$3E,$87,$C1,$7C,$FF,$A4,$95,$F3,$23
	.dc.b	$E7,$98,$7F,$D1,$71,$87,$F6,$BF,$B1,$7E,$F9,$BE,$4F,$ED,$A0,$DF
	.dc.b	$0A,$4E,$88,$7E,$96,$79,$3F,$EB,$1F,$81,$7D,$47,$C9,$F3,$E4,$FF
	.dc.b	$A2,$E7,$2D,$9C,$BF,$36,$7F,$01,$3A,$8E,$93,$AC,$FF,$C8,$9A,$FF
	.dc.b	$96,$27,$91,$FC,$43,$E0,$BF,$DA,$0D,$F4,$7F,$88,$C1,$BE,$F4,$9D
	.dc.b	$10,$FE,$01,$3C,$9F,$D0,$FB,$DF,$C6,$BE,$05,$F2,$74,$FD,$EC,$49
	.dc.b	$FE,$43,$3A,$6C,$E5,$EB,$E9,$79,$3D,$E9,$1E,$98,$F3,$98,$7C,$71
	.dc.b	$87,$47,$BD,$D6,$9F,$CE,$B1,$57,$EC,$FE,$8B,$E0,$DC,$3F,$DA,$2E
	.dc.b	$5F,$90,$64,$FF,$68,$BF,$BB,$F1,$BF,$79,$E3,$49,$FF,$6C,$C6,$5F
	.dc.b	$F6,$DC,$65,$FF,$6C,$C7,$EE,$7F,$0B,$7C,$4F,$C9,$F9,$19,$5F,$EC
	.dc.b	$D9,$CB,$67,$2F,$F3,$1D,$95,$CE,$64,$FC,$E8,$DE,$9F,$E8,$7B,$E7
	.dc.b	$3F,$95,$ED,$FC,$E1,$1F,$F7,$8F,$81,$FC,$37,$F1,$67,$C6,$FE,$87
	.dc.b	$E2,$65,$7F,$B8,$67,$2F,$FB,$E3,$E5,$7F,$8F,$11,$FD,$5C,$FF,$21
	.dc.b	$FB,$0E,$9F,$19,$27,$FD,$FF,$15,$FF,$7E,$FD,$87,$F2,$73,$1A,$FF
	.dc.b	$72,$F3,$51,$99,$FE,$D7,$18,$7C,$97,$FC,$97,$F1,$1F,$C7,$42,$FF
	.dc.b	$D5,$11,$86,$F4,$FF,$72,$F8,$5F,$E6,$B1,$8A,$FE,$07,$F2,$DF,$C0
	.dc.b	$4E,$95,$23,$24,$3B,$79,$3F,$CB,$30,$EC,$DE,$4F,$21,$27,$D3,$F0
	.dc.b	$75,$FE,$D9,$A3,$3F,$E9,$06,$73,$F4,$DE,$2B,$E0,$CE,$70,$FE,$77
	.dc.b	$F4,$0F,$A1,$FF,$03,$52,$30,$FF,$82,$5A,$FF,$83,$5A,$FF,$AE,$1C
	.dc.b	$E1,$FE,$CF,$FA,$27,$97,$FD,$45,$B5,$FE,$BA,$9F,$F6,$A5,$70,$C3
	.dc.b	$C7,$86,$F0,$97,$B9,$CB,$4F,$F6,$1F,$E0,$8F,$89,$FE,$BD,$CB,$FC
	.dc.b	$56,$74,$1C,$3F,$40,$E3,$1F,$D9,$67,$94,$70,$9D,$E5,$FE,$E9,$6D
	.dc.b	$FF,$62,$AD,$D7,$30,$E2,$B8,$78,$DE,$43,$B6,$1F,$DC,$FE,$8F,$FC
	.dc.b	$1B,$7E,$3F,$81,$FD,$EF,$6F,$F9,$15,$B1,$EA,$9D,$E1,$FE,$FB,$6E
	.dc.b	$5F,$EE,$4A,$DD,$73,$EF,$FE,$BF,$3F,$DE,$2B,$73,$FE,$EA,$6D,$70
	.dc.b	$B5,$73,$BE,$37,$BF,$07,$D1,$F2,$3E,$59,$F8,$73,$FE,$58,$6A,$C5
	.dc.b	$3C,$23,$EC,$63,$FC,$6B,$4E,$46,$63,$F3,$7F,$15,$FE,$21,$FB,$29
	.dc.b	$1F,$B8,$8F,$F1,$21,$87,$F4,$31,$87,$F0,$9F,$06,$1F,$19,$FE,$59
	.dc.b	$DE,$4E,$A7,$2D,$9C,$BF,$E4,$B9,$FF,$2A,$4E,$A3,$F1,$0B,$EA,$FE
	.dc.b	$2F,$C3,$3E,$0F,$F3,$B5,$FE,$49,$E6,$FF,$2E,$57,$CA,$BE,$18,$7F
	.dc.b	$95,$26,$57,$A2,$B4,$74,$BE,$7C,$7F,$E8,$C6,$F1,$AB,$DD,$5D,$7F
	.dc.b	$1B,$FE,$AF,$1F,$E7,$9E,$7F,$F3,$C1,$FE,$6D,$8C,$C7,$A9,$B8,$7F
	.dc.b	$A6,$3E,$1F,$F4,$D3,$C4,$F9,$73,$AC,$3F,$C8,$3F,$44,$F9,$87,$C3
	.dc.b	$7F,$A3,$63,$0D,$8C,$DF,$FA,$4E,$72,$FF,$A5,$7E,$35,$C9,$FE,$A0
	.dc.b	$8C,$35,$FF,$27,$9F,$ED,$77,$0F,$FB,$E9,$AF,$C3,$9F,$F0,$5F,$A5
	.dc.b	$E3,$F1,$BD,$F3,$E8,$6B,$B6,$E3,$F8,$7F,$A2,$FF,$22,$67,$F3,$63
	.dc.b	$0F,$FA,$8F,$C8,$CD,$FC,$0E,$55,$E2,$F7,$3F,$C3,$5B,$C1,$F3,$6F
	.dc.b	$57,$C3,$0F,$CB,$F9,$3F,$96,$7F,$D7,$10,$8C,$61,$F0,$A3,$1D,$1F
	.dc.b	$B6,$8C,$4E,$9F,$F3,$AC,$EA,$BA,$D7,$A3,$A3,$F2,$DF,$21,$EA,$AD
	.dc.b	$5F,$E7,$30,$EF,$97,$59,$CB,$F9,$1A,$E1,$3A,$C7,$F2,$54,$F9,$AF
	.dc.b	$F1,$3C,$7E,$C9,$B8,$7F,$AD,$7E,$6B,$C1,$18,$C7,$B3,$6B,$9B,$69
	.dc.b	$97,$FB,$AE,$F8,$F0,$7A,$FD,$4F,$E2,$F8,$7F,$91,$BE,$8B,$0E,$5A
	.dc.b	$E4,$FF,$25,$99,$3F,$A2,$F9,$6F,$37,$E6,$65,$F9,$08,$C4,$FF,$BA
	.dc.b	$5B,$1F,$EE,$D6,$BB,$9F,$C2,$BF,$E2,$E6,$FC,$B7,$FB,$52,$BB,$C6
	.dc.b	$AB,$FB,$C4,$7F,$24,$FA,$27,$1E,$7C,$EF,$C7,$F8,$A9,$CD,$7A,$27
	.dc.b	$CF,$97,$FA,$D3,$0E,$6F,$96,$32,$FF,$07,$8C,$BE,$D8,$ED,$3F,$D9
	.dc.b	$B2,$6D,$8A,$FE,$B9,$F4,$4F,$9A,$FD,$0F,$03,$E8,$9F,$1F,$F0,$2B
	.dc.b	$59,$7F,$75,$AD,$4E,$6F,$C7,$82,$F6,$FE,$F8,$C4,$6A,$33,$1F,$C0
	.dc.b	$9D,$F2,$FD,$66,$E1,$F9,$2F,$AC,$7D,$5F,$78,$FF,$00,$FD,$1B,$E8
	.dc.b	$FF,$73,$83,$E9,$1F,$31,$F6,$2B,$57,$FE,$0D,$AD,$4E,$6F,$D3,$5A
	.dc.b	$CF,$D0,$D7,$F8,$BB,$5F,$F7,$3B,$BA,$FC,$6B,$6F,$D1,$E6,$75,$7F
	.dc.b	$C2,$FD,$6F,$8A,$FD,$C7,$C2,$FF,$9B,$27,$31,$8F,$81,$FC,$7D,$6E
	.dc.b	$7E,$56,$5F,$A8,$AF,$9F,$39,$8F,$9C,$D7,$F9,$6B,$9C,$BF,$42,$DA
	.dc.b	$FE,$9C,$B0,$FE,$29,$1E,$23,$CF,$E5,$7F,$0D,$F9,$97,$CA,$79,$23
	.dc.b	$0D,$9F,$2D,$FF,$49,$6C,$F9,$88,$F4,$3F,$C7,$D0,$F9,$9F,$BF,$97
	.dc.b	$92,$BE,$2E,$F2,$FF,$C7,$FF,$3D,$FE,$A6,$7F,$DB,$51,$FB,$48,$C5
	.dc.b	$7E,$3D,$AF,$F4,$E6,$BF,$DB,$43,$87,$F8,$9E,$7F,$E4,$FA,$EF,$EC
	.dc.b	$FD,$6F,$6F,$B3,$F6,$3E,$3E,$4F,$AB,$6F,$C7,$BF,$EE,$D8,$F2,$7C
	.dc.b	$9F,$E9,$32,$34,$93,$F6,$29,$F4,$9F,$3C,$FC,$2F,$9B,$67,$FC,$BF
	.dc.b	$E5,$7D,$6D,$7F,$D8,$1C,$9F,$C5,$23,$1C,$9F,$D2,$78,$DE,$57,$DF
	.dc.b	$BF,$F0,$65,$71,$7F,$B8,$B5,$F8,$8D,$77,$E6,$7E,$F2,$3A,$E3,$A8
	.dc.b	$FC,$CF,$CA,$78,$5F,$CD,$FC,$B7,$E7,$3E,$56,$D7,$DE,$E5,$46,$31
	.dc.b	$FF,$3B,$67,$F9,$18,$C6,$9F,$EF,$57,$F4,$BD,$1F,$A6,$E1,$B8,$F3
	.dc.b	$6E,$1C,$B1,$FE,$64,$DA,$E1,$5E,$1F,$67,$F5,$FE,$E7,$EC,$1B,$19
	.dc.b	$3C,$D1,$96,$99,$AE,$D1,$96,$C7,$F9,$53,$E8,$3F,$93,$BF,$9C,$E1
	.dc.b	$FD,$14,$FF,$93,$27,$2F,$FB,$DF,$AF,$FB,$B4,$63,$5F,$EE,$8D,$BF
	.dc.b	$DA,$E7,$FD,$12,$F9,$78,$DF,$C2,$3E,$7F,$F8,$4F,$95,$6F,$E7,$36
	.dc.b	$BE,$23,$71,$FC,$26,$1F,$D0,$AB,$F1,$7F,$29,$FE,$20,$5F,$46,$8F
	.dc.b	$C6,$19,$71,$1C,$0F,$E2,$7C,$F8,$C7,$F9,$F3,$5F,$C6,$D7,$E2,$4E
	.dc.b	$73,$FE,$BD,$3B,$FD,$56,$27,$F0,$7C,$6F,$96,$FE,$21,$F4,$BF,$C4
	.dc.b	$5B,$7F,$AF,$6D,$70,$6E,$9F,$E2,$91,$E8,$7F,$A5,$AB,$F8,$7F,$C6
	.dc.b	$BE,$73,$FA,$3C,$E7,$0F,$F7,$7C,$63,$E2,$25,$FE,$F7,$C5,$B5,$FE
	.dc.b	$C3,$F7,$7F,$91,$CF,$AF,$85,$F3,$1F,$E0,$69,$CD,$7F,$B4,$EB,$E6
	.dc.b	$CE,$7C,$AF,$EC,$F5,$AA,$FF,$31,$B5,$FE,$2C,$D7,$FC,$B1,$1C,$6B
	.dc.b	$C1,$EA,$59,$FF,$6B,$7D,$2F,$F2,$AF,$05,$F4,$AF,$A5,$FF,$0C,$CE
	.dc.b	$63,$18,$7F,$BA,$A3,$F9,$8D,$9D,$7C,$0F,$0F,$63,$E3,$FB,$5F,$E4
	.dc.b	$FE,$47,$D2,$63,$1F,$9A,$DC,$7F,$DE,$5C,$DF,$4A,$FF,$A5,$78,$59
	.dc.b	$8F,$CF,$F0,$FF,$86,$BE,$09,$E1,$FF,$A4,$32,$FF,$33,$B7,$1F,$F3
	.dc.b	$33,$5F,$F3,$1B,$9B,$E7,$3F,$D3,$3E,$25,$F1,$BF,$AE,$C7,$FC,$CD
	.dc.b	$18,$9D,$4E,$3C,$4C,$B2,$32,$73,$8C,$63,$FE,$C6,$BF,$A9,$97,$FA
	.dc.b	$9B,$9B,$D5,$18,$23,$FE,$73,$7F,$07,$97,$FA,$CF,$C1,$C4,$FF,$A3
	.dc.b	$7E,$07,$FC,$9F,$87,$FA,$55,$AF,$FA,$C9,$B6,$69,$FE,$58,$BF,$F4
	.dc.b	$44,$67,$E0,$FD,$FE,$1F,$56,$69,$FE,$CD,$AD,$DF,$18,$FF,$01,$EC
	.dc.b	$BD,$E3,$FE,$9F,$AD,$3E,$3C,$3A,$DF,$33,$D0,$FA,$FE,$23,$FC,$AB
	.dc.b	$97,$9D,$B7,$FE,$C4,$AF,$F2,$24,$E5,$B1,$9B,$F0,$57,$D8,$EC,$E9
	.dc.b	$5D,$E7,$FB,$E0,$FA,$57,$07,$FB,$33,$B7,$FD,$55,$18,$9D,$B5,$FF
	.dc.b	$77,$46,$A7,$FC,$11,$0E,$9D,$3F,$E9,$E6,$BF,$E0,$83,$51,$DE,$FF
	.dc.b	$DF,$13,$AF,$5B,$FD,$F5,$87,$2C,$3B,$61,$E2,$C3,$83,$E9,$7C,$33
	.dc.b	$BC,$38,$C6,$0D,$C7,$59,$F0,$9F,$E1,$BA,$FF,$64,$61,$93,$94,$68
	.dc.b	$FF,$10,$61,$D1,$E1,$85,$72,$78,$CC,$94,$CD,$F4,$3C,$45,$C3,$CD
	.dc.b	$1C,$23,$31,$E8,$3E,$6D,$F9,$0F,$33,$F0,$9D,$CF,$18,$E6,$FD,$6A
	.dc.b	$F3,$23,$E9,$4F,$39,$EA,$73,$2C,$7A,$A3,$BB,$63,$EF,$E9,$F7,$E3
	.dc.b	$F1,$6D,$7E,$1B,$5F,$06,$1E,$18,$D7,$C5,$7F,$2D,$E1,$C6,$1F,$E4
	.dc.b	$D8,$C5,$AB,$7D,$CB,$DC,$FA,$EB,$75,$A9,$FC,$5E,$3C,$FC,$3F,$F2
	.dc.b	$C6,$3F,$99,$BF,$EB,$58,$68,$E5,$8F,$57,$F0,$11,$E2,$9E,$15,$CA
	.dc.b	$FA,$57,$11,$5B,$8F,$55,$F3,$D3,$8D,$7E,$1E,$BC,$2F,$E1,$5F,$7C
	.dc.b	$79,$67,$C5,$87,$0A,$F1,$E1,$E1,$7C,$F8,$F8,$70,$E3,$8F,$92,$7A
	.dc.b	$57,$F9,$DF,$1E,$97,$E1,$C7,$C9,$7B,$9E,$C6,$AF,$C9,$3A,$7E,$75
	.dc.b	$F3,$AE,$07,$BA,$FA,$1E,$38,$F3,$4F,$C4,$6E,$1F,$16,$3D,$27,$98
	.dc.b	$F8,$91,$D5,$CE,$1E,$33,$CB,$7E,$08,$CB,$E1,$7E,$29,$E1,$3E,$26
	.dc.b	$1F,$2A,$35,$3C,$A7,$D6,$7A,$0F,$DC,$65,$FD,$17,$AF,$FD,$3D,$97
	.dc.b	$F3,$E6,$EB,$FD,$35,$B4,$F0,$1C,$AE,$0C,$E3,$CE,$BF,$9A,$DF,$FB
	.dc.b	$4A,$31,$1F,$E1,$8E,$3F,$F0,$1B,$FC,$EB,$CB,$E3,$06,$9E,$0C,$7B
	.dc.b	$CE,$0F,$E7,$95,$FD,$76,$71,$1F,$E8,$6C,$7F,$D7,$0F,$F5,$38,$CE
	.dc.b	$7F,$D6,$9E,$93,$A8,$C4,$F8,$C8,$79,$E7,$FD,$D7,$D3,$FB,$5F,$7B
	.dc.b	$E1,$B6,$7F,$A7,$35,$FF,$4B,$FC,$6B,$E7,$77,$92,$F9,$B6,$BC,$E2
	.dc.b	$F9,$3F,$ED,$16,$BF,$EF,$2B,$7C,$8F,$C5,$6C,$7F,$B9,$9A,$F8,$BE
	.dc.b	$77,$53,$23,$4F,$18,$F2,$FD,$2C,$9F,$5E,$08,$E1,$1F,$E1,$B7,$D2
	.dc.b	$47,$D3,$87,$F1,$98,$F5,$C7,$F1,$9F,$A5,$7D,$2F,$E2,$B6,$46,$27
	.dc.b	$FB,$B4,$62,$7F,$01,$E8,$3E,$3B,$85,$EE,$FF,$D8,$7E,$47,$E1,$DF
	.dc.b	$F3,$BF,$0E,$B0,$FC,$BC,$3C,$3F,$E2,$39,$3A,$3F,$DF,$91,$86,$F6
	.dc.b	$7E,$3F,$E9,$5C,$3C,$AB,$58,$7E,$86,$5F,$41,$B3,$57,$E0,$FA,$57
	.dc.b	$B3,$49,$7F,$E1,$31,$F4,$6F,$73,$31,$F8,$2A,$E2,$FC,$A6,$B9,$AF
	.dc.b	$9A,$A3,$3A,$FB,$63,$FB,$05,$BD,$EF,$86,$D8,$F1,$57,$0D,$BF,$C7
	.dc.b	$DC,$4F,$84,$F9,$E4,$AF,$F1,$57,$73,$E9,$5F,$0F,$99,$FA,$7F,$15
	.dc.b	$FB,$7E,$73,$9D,$BC,$98,$71,$F8,$2F,$DE,$62,$F8,$4A,$62,$26,$3E
	.dc.b	$09,$D3,$67,$2F,$F6,$5D,$7F,$BF,$2F,$85,$7D,$2F,$E0,$9B,$1F,$45
	.dc.b	$AF,$13,$E3,$D6,$E7,$E6,$CE,$A7,$F8,$5B,$5E,$ED,$7F,$C1,$CF,$4C
	.dc.b	$BE,$94,$7D,$EF,$89,$FA,$27,$57,$E5,$FC,$1F,$FB,$A5,$BE,$B7,$F1
	.dc.b	$F8,$FF,$49,$C6,$BD,$6C,$19,$99,$19,$3F,$6F,$F0,$3F,$C3,$88,$F6
	.dc.b	$F5,$EB,$EF,$E0,$7C,$2B,$E5,$BF,$97,$36,$79,$9E,$6F,$43,$C9,$EB
	.dc.b	$C8,$C1,$0E,$9A,$7F,$B7,$1F,$6C,$7A,$E6,$46,$08,$70,$7F,$11,$E7
	.dc.b	$7C,$18,$87,$FB,$C9,$CB,$F2,$0C,$9F,$C0,$9F,$C8,$79,$9F,$91,$F1
	.dc.b	$64,$FF,$BE,$A3,$2F,$FB,$F2,$32,$FF,$0A,$8F,$E0,$FF,$4A,$F8,$2A
	.dc.b	$49,$FD,$0F,$5B,$E0,$F1,$1F,$CE,$2B,$83,$FC,$81,$AE,$6B,$F6,$F1
	.dc.b	$8E,$5F,$D1,$EC,$F9,$4F,$A2,$74,$8F,$D3,$F5,$FE,$D7,$0F,$EE,$0D
	.dc.b	$E2,$FD,$97,$89,$FA,$FF,$1F,$F8,$FF,$D2,$BE,$3F,$F8,$2B,$70,$FE
	.dc.b	$66,$D7,$FA,$7B,$5F,$D4,$DB,$77,$D2,$FE,$EC,$93,$FB,$3F,$8C,$FD
	.dc.b	$F7,$D2,$F8,$3E,$3A,$4F,$C6,$8A,$F6,$8F,$19,$1F,$EB,$F8,$77,$FA
	.dc.b	$5F,$E1,$04,$8F,$E4,$D9,$2B,$F9,$DF,$D2,$FE,$56,$7F,$70,$1C,$70
	.dc.b	$7E,$C1,$BE,$2F,$FA,$07,$17,$C4,$7F,$47,$BE,$35,$CD,$FF,$03,$B5
	.dc.b	$FE,$16,$D7,$FD,$14,$D7,$C1,$DD,$FC,$6B,$E6,$E4,$7F,$3D,$9F,$0B
	.dc.b	$C3,$97,$D2,$47,$F1,$7F,$0F,$5F,$F0,$0C,$FE,$0D,$FE,$D2,$D7,$33
	.dc.b	$FD,$02,$31,$9F,$F2,$CE,$2F,$6D,$7E,$5B,$2E,$93,$FC,$4B,$E9,$1F
	.dc.b	$49,$FD,$91,$B5,$FD,$4D,$AF,$CC,$FA,$5F,$F4,$93,$99,$FE,$9F,$3A
	.dc.b	$8F,$F1,$6B,$5F,$E6,$2D,$7F,$D6,$0F,$4A,$FE,$47,$5F,$D1,$27,$FB
	.dc.b	$13,$5F,$B7,$3B,$DF,$F9,$DF,$D2,$FF,$4E,$9F,$ED,$0D,$7F,$02,$DB
	.dc.b	$FE,$5F,$CA,$F1,$7F,$C0,$33,$BA,$FF,$21,$B5,$EB,$E7,$FF,$6A,$8C
	.dc.b	$47,$FB,$1D,$E9,$87,$F4,$CB,$FE,$67,$3F,$90,$6E,$1F,$CF,$AB,$73
	.dc.b	$FE,$CF,$FA,$5F,$F1,$8C,$FF,$42,$6D,$FF,$56,$6C,$BE,$11,$F6,$18
	.dc.b	$84,$FF,$8E,$9F,$F2,$DB,$5E,$5E,$27,$F8,$0A,$31,$1E,$18,$FF,$3B
	.dc.b	$D7,$F1,$BE,$FB,$E1,$FF,$0B,$AE,$B5,$FE,$26,$BF,$E3,$91,$FE,$15
	.dc.b	$6B,$FA,$E6,$E5,$FE,$0D,$C3,$CF,$D9,$C7,$C4,$F1,$12,$FF,$C8,$37
	.dc.b	$C9,$FE,$1B,$CA,$F8,$DF,$F0,$37,$D3,$49,$FE,$5B,$1F,$E0,$93,$0F
	.dc.b	$F8,$FA,$30,$FF,$9B,$BE,$0D,$F3,$3F,$E7,$26,$C6,$63,$FC,$C9,$8E
	.dc.b	$59,$19,$3F,$CC,$D1,$FE,$7A,$FA,$6F,$E5,$D4,$69,$FC,$BF,$D2,$7D
	.dc.b	$1E,$F8,$A6,$1F,$E8,$0A,$D4,$E2,$3F,$D2,$37,$CE,$BF,$D1,$35,$FE
	.dc.b	$3A,$9F,$F3,$E3,$51,$FE,$81,$FA,$6F,$F5,$63,$FE,$8C,$6B,$FE,$17
	.dc.b	$A4,$63,$87,$FC,$5B,$F1,$3F,$E5,$4A,$E1,$E8,$7F,$A5,$AB,$5C,$FD
	.dc.b	$B0,$E3,$3F,$E8,$9F,$2F,$FC,$CF,$F4,$CF,$95,$FE,$03,$8F,$FA,$AA
	.dc.b	$31,$F1,$7F,$E7,$C6,$C7,$FA,$61,$AF,$FA,$C3,$E4,$BE,$F7,$4F,$FA
	.dc.b	$09,$B1,$FE,$9E,$6B,$BE,$6F,$D1,$3C,$FC,$9F,$7B,$FE,$3C,$BF,$F0
	.dc.b	$B6,$7E,$19,$CC,$FC,$89,$CB,$FE,$C0,$C7,$FD,$55,$F4,$DF,$E6,$3E
	.dc.b	$1F,$F3,$24,$63,$7F,$F2,$DF,$C7,$BC,$BD,$47,$D2,$9D,$61,$FE,$45
	.dc.b	$6D,$F2,$E1,$63,$C2,$BF,$DC,$98,$F3,$C3,$FD,$AB,$5A,$7F,$D8,$BE
	.dc.b	$6F,$E6,$1B,$D5,$C7,$CD,$73,$3E,$53,$A6,$1F,$D8,$6A,$E1,$C7,$2F
	.dc.b	$C1,$37,$4F,$C0,$DE,$AA,$65,$D8,$FF,$62,$46,$2F,$F1,$5F,$3B,$FE
	.dc.b	$F9,$9F,$F6,$CB,$63,$FD,$D5,$F3,$DE,$28,$C6,$1F,$E8,$DC,$91,$8B
	.dc.b	$FC,$63,$6F,$E1,$DF,$1E,$3F,$77,$AD,$FE,$B1,$BF,$F4,$5B,$FE,$BD
	.dc.b	$69,$AF,$8F,$DB,$4C,$9C,$1F,$A1,$7F,$E2,$9F,$4A,$FA,$47,$3F,$3A
	.dc.b	$F9,$1C,$C4,$C3,$FD,$65,$5B,$7F,$D6,$95,$B7,$33,$D3,$D3,$91,$AD
	.dc.b	$2E,$3F,$E1,$2C,$3F,$C1,$B3,$D2,$BC,$35,$AD,$FF,$D0,$75,$AB,$E1
	.dc.b	$5A,$BE,$07,$FB,$17,$E9,$5F,$37,$EE,$6C,$7F,$06,$C3,$E9,$7D,$2B
	.dc.b	$E9,$75,$72,$B9,$3F,$8A,$9D,$46,$09,$8F,$F8,$AB,$0E,$AF,$F0,$D9
	.dc.b	$DB,$9A,$F9,$FF,$48,$FA,$4F,$F4,$EC,$FF,$AC,$23,$18,$7F,$C0,$06
	.dc.b	$C7,$FA,$D0,$FE,$3A,$E1,$C3,$AC,$BF,$D4,$ED,$9F,$E2,$D8,$7F,$B2
	.dc.b	$9F,$F7,$44,$66,$B9,$DF,$90,$F7,$7C,$65,$AF,$F7,$77,$5F,$FB,$0D
	.dc.b	$FF,$57,$FB,$FF,$88,$8C,$5F,$FA,$62,$31,$F1,$32,$BF,$D6,$BD,$96
	.dc.b	$BF,$C9,$F7,$C3,$4F,$F5,$EB,$6B,$FC,$9D,$F0,$DF,$E5,$77,$C3,$87
	.dc.b	$7F,$23,$F7,$DC,$EF,$95,$C5,$7F,$BB,$23,$0F,$A2,$73,$97,$FA,$DF
	.dc.b	$E4,$FE,$AE,$7E,$09,$CB,$6B,$FD,$EB,$3A,$7F,$CF,$73,$FC,$37,$2F
	.dc.b	$91,$99,$F9,$58,$3E,$7B,$FD,$41,$7F,$ED,$4E,$0F,$D6,$C6,$27,$FC
	.dc.b	$DE,$DC,$BF,$DF,$76,$F7,$37,$EC,$F7,$6E,$3F,$ED,$E6,$E1,$FE,$C6
	.dc.b	$F4,$FC,$87,$C4,$F9,$AB,$5F,$9D,$3F,$DF,$79,$3E,$47,$C6,$7C,$AF
	.dc.b	$A4,$B3,$F2,$5C,$3F,$EE,$CF,$1E,$E3,$FE,$86,$F1,$7E,$A3,$79,$BE
	.dc.b	$27,$C8,$FF,$0A,$9C,$F0,$7F,$BE,$1B,$87,$F7,$4B,$FE,$3F,$7C,$23
	.dc.b	$CD,$1F,$DE,$27,$6D,$9E,$86,$DF,$04,$FF,$C0,$0F,$96,$FA,$9F,$27
	.dc.b	$F5,$5B,$19,$3A,$F2,$7D,$79,$D4,$63,$8D,$1F,$77,$9B,$EF,$73,$DA
	.dc.b	$D3,$71,$FF,$7A,$B7,$4F,$B2,$FE,$17,$E0,$BF,$0D,$EC,$DB,$EC,$DB
	.dc.b	$FF,$7E,$B6,$FF,$02,$D9,$FD,$0B,$65,$9F,$E3,$E3,$FA,$DF,$D0,$3E
	.dc.b	$6E,$9F,$95,$F4,$EE,$1F,$33,$CE,$56,$E3,$0D,$C5,$97,$E4,$AF,$8C
	.dc.b	$62,$BF,$67,$F3,$FF,$A1,$F2,$BF,$44,$D8,$F9,$4D,$9E,$07,$A3,$85
	.dc.b	$19,$C3,$F0,$9F,$2B,$F8,$73,$FC,$01,$6F,$A3,$FF,$48,$3F,$84,$C3
	.dc.b	$FB,$2C,$EE,$31,$CC,$8E,$35,$C2,$BF,$2D,$1B,$6D,$7E,$38,$ED,$F1
	.dc.b	$19,$C1,$1F,$ED,$DE,$37,$CA,$FF,$0E,$FA,$4F,$E1,$ED,$9F,$E1,$CD
	.dc.b	$C3,$DF,$C9,$FD,$96,$31,$87,$F8,$9F,$D1,$7C,$1F,$F8,$76,$11,$8C
	.dc.b	$3F,$8C,$78,$99,$8C,$63,$F2,$AB,$77,$93,$F9,$0D,$5B,$F8,$DE,$D7
	.dc.b	$F2,$0B,$FF,$7D,$B6,$FD,$B1,$EE,$DF,$FC,$18,$A7,$55,$FB,$56,$C7
	.dc.b	$F8,$E7,$C8,$F9,$BE,$B7,$F8,$1A,$73,$3C,$67,$06,$AB,$C6,$FF,$83
	.dc.b	$3E,$8F,$F9,$54,$FF,$9D,$1B,$B3,$E9,$3F,$C6,$D7,$F7,$E7,$74,$69
	.dc.b	$C2,$77,$1F,$8E,$35,$E8,$FC,$9F,$52,$5F,$F4,$3F,$36,$D7,$CB,$E1
	.dc.b	$FF,$04,$6F,$F5,$6F,$FA,$5D,$6A,$3F,$C6,$9D,$9F,$17,$0F,$D3,$ED
	.dc.b	$FE,$4D,$F6,$DF,$33,$FB,$28,$93,$A3,$FC,$C7,$18,$CF,$F0,$B5,$C2
	.dc.b	$72,$7F,$9A,$3A,$FA,$54,$8C,$D7,$09,$C3,$FE,$97,$BF,$AB,$DF,$F4
	.dc.b	$6B,$FC,$E3,$D7,$FE,$5E,$9C,$93,$8F,$D1,$19,$7F,$CD,$D1,$FE,$E7
	.dc.b	$F6,$2F,$C1,$7F,$8E,$71,$F3,$F6,$E2,$66,$3F,$CD,$AF,$83,$FE,$80
	.dc.b	$9C,$D7,$F9,$F6,$F9,$6F,$FE,$86,$6D,$7F,$AD,$39,$9B,$FE,$DE,$B5
	.dc.b	$97,$26,$D7,$53,$C1,$7C,$B0,$FC,$C3,$FE,$E1,$8C,$FC,$0F,$D5,$F9
	.dc.b	$9F,$5B,$23,$27,$FA,$5A,$31,$5F,$89,$3F,$D8,$FE,$7F,$FB,$23,$E8
	.dc.b	$FF,$8B,$C7,$F5,$7E,$BF,$D4,$C6,$23,$FB,$58,$70,$FF,$9F,$DA,$FF
	.dc.b	$80,$CD,$71,$FF,$8E,$67,$9B,$FE,$3B,$8D,$4F,$29,$CC,$7F,$8F,$EB
	.dc.b	$C9,$5C,$70,$B3,$93,$A9,$E7,$72,$F9,$AF,$A4,$62,$3D,$E7,$03,$FD
	.dc.b	$CF,$7E,$07,$85,$77,$74,$70,$AE,$6E,$0F,$4C,$6E,$3E,$0D,$F7,$AE
	.dc.b	$95,$C6,$B4,$7C,$19,$D5,$F9,$67,$99,$EB,$3B,$46,$9A,$79,$2B,$AB
	.dc.b	$C0,$F5,$CF,$39,$E1,$7E,$49,$E1,$39,$69,$D2,$3D,$A6,$CB,$7E,$78
	.dc.b	$DC,$73,$E0,$FF,$67,$47,$D3,$6B,$F8,$66,$BF,$41,$CF,$C2,$3E,$93
	.dc.b	$FB,$8C,$FF,$33,$F8,$1F,$F1,$34,$E7,$7F,$F7,$54,$62,$B7,$7C,$9B
	.dc.b	$19,$9F,$C0,$D3,$E8,$3F,$C6,$18,$FE,$66,$31,$8E,$B0,$F0,$CE,$5B
	.dc.b	$97,$8A,$F8,$B6,$B8,$61,$1F,$C5,$61,$D1,$B5,$C1,$08,$F6,$3C,$2F
	.dc.b	$BB,$A8,$E5,$87,$9E,$B9,$4F,$1A,$DB,$6F,$DF,$3A,$BE,$38,$7C,$3A
	.dc.b	$ED,$5A,$61,$82,$61,$DD,$CB,$F0,$E3,$DC,$E6,$3D,$A6,$27,$B4,$64
	.dc.b	$8E,$09,$1E,$DB,$F3,$1E,$7C,$3D,$78,$7C,$4B,$E3,$5B,$3B,$1F,$16
	.dc.b	$77,$5C,$A3,$39,$38,$1F,$49,$FE,$6F,$CF,$F9,$93,$75,$FA,$75,$B8
	.dc.b	$C5,$FF,$93,$B8,$DC,$18,$AD,$4F,$F3,$2A,$C6,$3D,$A3,$FC,$DB,$7C
	.dc.b	$1F,$E6,$FE,$5E,$88,$E8,$A7,$F4,$76,$E5,$FE,$16,$E5,$FE,$8B,$38
	.dc.b	$CF,$CF,$87,$F9,$7E,$3A,$9C,$CA,$73,$2D,$73,$C3,$FD,$D0,$FF,$88
	.dc.b	$BA,$7F,$C4,$1E,$B7,$C0,$6D,$7F,$99,$F4,$4F,$C4,$C9,$F0,$8F,$9D
	.dc.b	$E1,$5A,$BE,$17,$83,$85,$6E,$3F,$D2,$2B,$E4,$72,$48,$CF,$CE,$D7
	.dc.b	$FC,$97,$3B,$78,$CF,$49,$F0,$FD,$24,$9F,$43,$C7,$4F,$76,$28,$D4
	.dc.b	$FF,$82,$5D,$6F,$FB,$4A,$7D,$03,$E9,$3F,$07,$18,$6C,$66,$FF,$DD
	.dc.b	$DF,$13,$89,$91,$FE,$BA,$AE,$0E,$8F,$F3,$D7,$83,$FC,$16,$3F,$CA
	.dc.b	$71,$FE,$F9,$E7,$FC,$8B,$4F,$F5,$BF,$D2,$35,$CD,$F1,$CF,$FD,$94
	.dc.b	$DA,$FE,$1A,$DE,$25,$F1,$9F,$F4,$57,$D1,$BD,$7E,$ED,$8C,$C7,$8B
	.dc.b	$E9,$1E,$D6,$A2,$4E,$8F,$F4,$A4,$E9,$C1,$FE,$53,$AE,$8D,$9E,$95
	.dc.b	$F2,$BE,$06,$4F,$C7,$E1,$FF,$7A,$3F,$ED,$6F,$A2,$7C,$26,$A3,$FD
	.dc.b	$05,$3A,$6D,$7E,$56,$B7,$1C,$7D,$EC,$7C,$0C,$8C,$10,$FF,$4B,$CA
	.dc.b	$35,$18,$9F,$A7,$F0,$CF,$A4,$A4,$9F,$83,$55,$FF,$39,$4E,$DF,$F5
	.dc.b	$C3,$5F,$F3,$B6,$DD,$DC,$D7,$1E,$5F,$C1,$D6,$DF,$C2,$D6,$EB,$FD
	.dc.b	$5B,$EF,$7F,$AF,$DF,$F7,$C7,$A1,$FB,$58,$FF,$39,$C7,$FB,$EF,$D0
	.dc.b	$C1,$99,$91,$93,$F7,$DE,$DF,$EE,$88,$FE,$8F,$9F,$FC,$AD,$6F,$47
	.dc.b	$33,$FE,$71,$AB,$97,$F1,$7A,$D4,$F1,$3F,$DC,$1E,$47,$97,$CF,$91
	.dc.b	$9E,$95,$7F,$27,$8F,$6D,$69,$F7,$D7,$B6,$24,$71,$94,$EE,$FC,$55
	.dc.b	$FC,$5F,$E6,$3F,$88,$5F,$C8,$3E,$31,$9F,$16,$3C,$CC,$9E,$58,$E3
	.dc.b	$5F,$CC,$DF,$E2,$51,$FE,$DE,$7F,$58,$DE,$DF,$07,$87,$23,$94,$78
	.dc.b	$CC,$3F,$EB,$7F,$A4,$7B,$9F,$C0,$1B,$1F,$EF,$F6,$BD,$FD,$DD,$4C
	.dc.b	$8D,$3E,$F8,$C4,$75,$8B,$8F,$A6,$7F,$4D,$DD,$2B,$F5,$CF,$C7,$F9
	.dc.b	$5F,$E2,$3F,$44,$F8,$CA,$4A,$D9,$F1,$7E,$93,$E7,$11,$FD,$C7,$C3
	.dc.b	$FD,$1F,$A4,$E7,$F1,$52,$7E,$64,$57,$FD,$B1,$82,$71,$A7,$CF,$AF
	.dc.b	$B5,$F3,$D2,$FE,$A3,$AE,$AF,$F1,$5C,$3E,$91,$F3,$5F,$CA,$9B,$1F
	.dc.b	$93,$35,$1F,$D6,$FE,$0B,$FB,$54,$7F,$BA,$67,$6D,$9E,$46,$FD,$7B
	.dc.b	$1F,$C5,$7E,$33,$F8,$81,$87,$97,$CB,$F8,$48,$FE,$2F,$E0,$AB,$A3
	.dc.b	$C9,$82,$7F,$31,$3A,$D7,$F2,$19,$7E,$6C,$C6,$7F,$03,$E8,$9F,$49
	.dc.b	$FD,$21,$B9,$7B,$DA,$AF,$FA,$7F,$D2,$3D,$8E,$17,$2B,$83,$FD,$4A
	.dc.b	$75,$18,$3F,$D6,$C2,$FA,$BF,$48,$C4,$E9,$A7,$0F,$A4,$FE,$6A,$F8
	.dc.b	$F1,$FB,$31,$8C,$7F,$B1,$B6,$3E,$EF,$2F,$F0,$CA,$E0,$7F,$9A,$A7
	.dc.b	$4D,$F4,$7F,$B4,$C6,$21,$3C,$6F,$AD,$68,$93,$F8,$59,$D5,$FE,$33
	.dc.b	$E8,$FF,$7F,$7F,$DD,$9B,$19,$F0,$3F,$C8,$FE,$FB,$E9,$33,$8F,$F9
	.dc.b	$FA,$EA,$69,$A9,$E1,$5F,$E0,$08,$CB,$4E,$B5,$C3,$BF,$44,$BF,$E5
	.dc.b	$D7,$E1,$E8,$C4,$7C,$63,$C1,$38,$C9,$97,$FA,$4A,$F8,$57,$F1,$FF
	.dc.b	$88,$FE,$51,$F4,$92,$30,$43,$F9,$03,$97,$F9,$14,$7F,$92,$7E,$05
	.dc.b	$F2,$D4,$93,$FE,$40,$CA,$BF,$6A,$47,$07,$FA,$43,$89,$FF,$24,$CF
	.dc.b	$F9,$57,$E8,$FF,$C1,$79,$FF,$96,$5B,$9F,$FA,$B2,$D1,$8C,$1E,$6B
	.dc.b	$E6,$5F,$49,$CF,$C7,$95,$E0,$BE,$17,$BA,$CC,$F9,$E7,$A6,$1D,$AF
	.dc.b	$FC,$C9,$87,$4A,$F4,$FD,$27,$F5,$BC,$FF,$CD,$CD,$CF,$CC,$D9,$F2
	.dc.b	$FC,$2B,$C9,$FF,$39,$4E,$72,$34,$E1,$AF,$F9,$CA,$7F,$C8,$FF,$48
	.dc.b	$F9,$1F,$DB,$E3,$FB,$38,$C7,$7F,$EE,$1B,$1F,$B7,$0E,$1F,$DD,$7C
	.dc.b	$6B,$93,$FD,$55,$1B,$8C,$73,$7F,$02,$6C,$66,$76,$F8,$7C,$7F,$EF
	.dc.b	$8F,$F3,$6B,$FE,$6B,$E9,$78,$E7,$33,$FE,$E4,$9C,$BF,$C2,$F6,$FE
	.dc.b	$89,$CF,$F9,$88,$FF,$41,$F9,$97,$8F,$FD,$35,$E5,$FE,$43,$93,$FD
	.dc.b	$1B,$39,$7F,$09,$19,$AE,$31,$F8,$E8,$C5,$7F,$AB,$F9,$17,$C6,$FF
	.dc.b	$2B,$4F,$9B,$C7,$27,$FA,$8B,$CC,$D6,$9F,$D6,$67,$38,$A7,$37,$C7
	.dc.b	$3F,$F5,$0C,$62,$B6,$D8,$C9,$E5,$8C,$BF,$E8,$BF,$BA,$FE,$7E,$FF
	.dc.b	$AA,$BB,$BC,$DA,$A7,$3A,$7F,$AA,$27,$31,$88,$FF,$71,$35,$FF,$50
	.dc.b	$35,$FE,$8C,$DD,$1F,$35,$FE,$A4,$C3,$FC,$49,$CD,$F0,$A3,$17,$FE
	.dc.b	$23,$8C,$7C,$1F,$C6,$F5,$7F,$A7,$5F,$F8,$FE,$F7,$9F,$F8,$AE,$FE
	.dc.b	$76,$1F,$D8,$7C,$BF,$A1,$D3,$FB,$6F,$3F,$3E,$72,$73,$E1,$27,$31
	.dc.b	$86,$E7,$FC,$92,$74,$FF,$80,$69,$E2,$79,$7E,$3F,$AB,$67,$4D,$CB
	.dc.b	$F9,$55,$70,$9F,$F2,$25,$70,$69,$C0,$B3,$FC,$C7,$E8,$1F,$3F,$F5
	.dc.b	$BE,$9B,$FD,$D7,$5F,$E1,$39,$CB,$FE,$C7,$8C,$D7,$1A,$E1,$F2,$1F
	.dc.b	$1A,$F7,$3F,$D8,$B8,$17,$C1,$FF,$65,$CE,$A7,$FB,$A3,$5F,$F1,$3B
	.dc.b	$50,$70,$BE,$DD,$EA,$7E,$7B,$EE,$7F,$D4,$59,$FF,$8A,$E7,$2F,$F3
	.dc.b	$88,$CD,$7F,$51,$9C,$BF,$CF,$E3,$38,$7F,$B8,$27,$2F,$EF,$7B,$93
	.dc.b	$AB,$FF,$1E,$D6,$9B,$87,$ED,$7E,$D1,$F7,$3F,$CF,$27,$C9,$3F,$8C
	.dc.b	$EA,$FE,$49,$1F,$91,$F9,$0C,$79,$3F,$99,$6C,$64,$FF,$23,$4F,$0D
	.dc.b	$7F,$C1,$EF,$E8,$A7,$71,$FD,$B6,$35,$AF,$F1,$57,$FD,$CF,$F3,$BA
	.dc.b	$8F,$F7,$EC,$7F,$95,$2F,$FD,$41,$F1,$FF,$E8,$07,$BF,$91,$FC,$32
	.dc.b	$31,$87,$58,$C7,$BF,$EB,$6D,$70,$24,$FF,$6B,$FA,$8F,$C1,$47,$F9
	.dc.b	$93,$0F,$04,$FA,$BB,$F2,$FF,$64,$6A,$F9,$C7,$C8,$FE,$F7,$A3,$F6
	.dc.b	$B1,$8C,$3F,$75,$18,$9D,$47,$EF,$7E,$97,$0D,$3F,$CE,$15,$C6,$FC
	.dc.b	$5F,$31,$FC,$24,$FE,$5D,$F7,$0F,$C2,$DE,$4E,$70,$E3,$E1,$6A,$73
	.dc.b	$18,$3F,$D9,$D5,$C6,$B5,$BC,$9C,$D3,$E1,$31,$97,$F8,$97,$E9,$9F
	.dc.b	$21,$FC,$FE,$DF,$4B,$87,$FD,$11,$36,$BF,$CA,$31,$AA,$FE,$33,$39
	.dc.b	$E7,$FE,$AD,$39,$38,$CF,$5B,$D4,$7F,$A6,$2A,$DF,$F8,$D2,$6A,$8C
	.dc.b	$CF,$38,$FB,$5F,$37,$FC,$C7,$EB,$3F,$99,$36,$7F,$97,$36,$77,$1F
	.dc.b	$CC,$67,$F9,$C7,$C8,$FF,$39,$FA,$7B,$3F,$CD,$5B,$5F,$87,$6E,$7F
	.dc.b	$9D,$E7,$47,$F4,$1F,$88,$FE,$85,$DB,$70,$FE,$7C,$DC,$3F,$20,$DC
	.dc.b	$39,$36,$7F,$88,$F5,$7F,$4E,$8C,$57,$E5,$5F,$EB,$BF,$4A,$F8,$3F
	.dc.b	$E9,$9F,$3D,$86,$E1,$FC,$6B,$1F,$F0,$25,$6E,$3F,$DF,$55,$B6,$9F
	.dc.b	$D4,$FA,$3F,$AD,$76,$7F,$5B,$6C,$64,$FE,$B1,$19,$69,$FE,$CB,$F4
	.dc.b	$9E,$17,$DD,$F9,$3F,$EB,$71,$89,$F2,$1F,$CB,$8F,$C6,$C6,$A7,$84
	.dc.b	$73,$6E,$BF,$E7,$9A,$E4,$FE,$43,$99,$7C,$6F,$AC,$FE,$26,$3F,$BA
	.dc.b	$7C,$C7,$F7,$5F,$AF,$B5,$FD,$C5,$B5,$FC,$0D,$BB,$7D,$D7,$FB,$9F
	.dc.b	$D4,$38,$7F,$C3,$FF,$50,$F4,$B3,$18,$6E,$3E,$2E,$45,$6E,$3F,$DE
	.dc.b	$17,$CD,$D4,$67,$2F,$76,$7F,$E2,$CF,$B8,$7C,$7F,$CA,$8C,$36,$FB
	.dc.b	$4E,$9F,$A5,$3A,$69,$A9,$E1,$5F,$CE,$AB,$83,$FD,$A7,$97,$FC,$19
	.dc.b	$5E,$9F,$69,$F4,$BF,$90,$F6,$BF,$C3,$4D,$9D,$36,$FF,$A2,$D7,$09
	.dc.b	$FA,$7C,$2C,$3F,$CD,$51,$D7,$C4,$FC,$DF,$C3,$7F,$89,$5B,$19,$38
	.dc.b	$CF,$0C,$3A,$CF,$E8,$BA,$BF,$C9,$BE,$85,$BF,$A1,$F0,$5F,$D9,$6F
	.dc.b	$FD,$FD,$82,$F7,$5D,$1B,$3F,$D7,$3E,$2B,$FA,$1F,$27,$FA,$97,$CE
	.dc.b	$FF,$2E,$C7,$E4,$35,$FE,$B8,$E8,$91,$AE,$CD,$BF,$B6,$F2,$3F,$CC
	.dc.b	$3C,$57,$3F,$2B,$72,$FE,$D7,$AF,$D0,$7E,$95,$70,$AF,$F6,$0F,$E7
	.dc.b	$1F,$5B,$3F,$F3,$A7,$BB,$FE,$6E,$8C,$4F,$F0,$8C,$3F,$C6,$98,$7F
	.dc.b	$B2,$E7,$2F,$FA,$B2,$B8,$73,$7F,$9D,$1B,$7D,$A3,$DF,$E7,$62,$B9
	.dc.b	$45,$C7,$E0,$CD,$AF,$F6,$76,$1F,$E7,$BB,$F5,$F2,$79,$7E,$0B,$FD
	.dc.b	$02,$DA,$FA,$AD,$9D,$C7,$EE,$1F,$F1,$89,$FD,$C4,$7D,$3B,$EA,$33
	.dc.b	$0F,$8A,$FF,$51,$4E,$63,$3E,$9F,$F8,$EA,$31,$1F,$E3,$C9,$DD,$7F
	.dc.b	$AF,$EB,$4F,$48,$CC,$F2,$3C,$31,$9B,$DC,$66,$FA,$BE,$38,$CB,$EA
	.dc.b	$BA,$E4,$E6,$71,$AE,$8E,$0E,$05,$38,$57,$67,$E2,$3E,$0B,$F3,$C7
	.dc.b	$08,$C5,$F8,$23,$75,$DA,$79,$5F,$92,$F8,$D7,$F9,$82,$31,$7C,$5B
	.dc.b	$1E,$33,$99,$F5,$A7,$8B,$83,$E7,$4F,$17,$D5,$1A,$9E,$75,$E2,$9C
	.dc.b	$D6,$E3,$D3,$1D,$E7,$77,$F5,$E3,$66,$4B,$3C,$CE,$27,$5F,$15,$F5
	.dc.b	$3C,$5B,$3F,$75,$AF,$D0,$6F,$C6,$7F,$3D,$DF,$FD,$07,$18,$F0,$D1
	.dc.b	$8B,$FC,$BD,$6E,$D7,$BA,$D4,$7F,$A3,$27,$58,$7E,$51,$FE,$29,$1F
	.dc.b	$88,$6B,$F8,$37,$78,$F6,$C7,$C3,$39,$EA,$FD,$3D,$F3,$6D,$6F,$0E
	.dc.b	$97,$C2,$B5,$7D,$EF,$6F,$96,$75,$7C,$63,$B1,$B3,$9D,$C7,$24,$BF
	.dc.b	$65,$72,$8E,$93,$C5,$D5,$75,$AD,$CE,$5F,$2C,$FC,$1A,$E4,$B0,$F8
	.dc.b	$6F,$CB,$C3,$D2,$E5,$B1,$A8,$E1,$3C,$A3,$E6,$D7,$63,$C3,$3D,$EF
	.dc.b	$C5,$87,$DC,$C3,$9E,$1C,$4E,$4F,$63,$33,$F0,$EF,$A5,$71,$BC,$3F
	.dc.b	$C9,$BE,$3F,$F8,$ED,$7F,$A9,$79,$7F,$C1,$55,$FD,$63,$1C,$ED,$B8
	.dc.b	$FE,$83,$78,$5D,$6B,$E3,$1F,$D2,$AF,$5F,$00,$E4,$61,$85,$F4,$14
	.dc.b	$97,$FE,$17,$ED,$FF,$2D,$BF,$D4,$23,$2F,$F4,$B7,$C5,$92,$77,$39
	.dc.b	$3F,$B8,$EF,$AC,$FF,$37,$AF,$FB,$6F,$1F,$0D,$75,$36,$43,$91,$FE
	.dc.b	$13,$E1,$E5,$E1,$48,$FF,$01,$C7,$C3,$AF,$F0,$63,$98,$FE,$D7,$39
	.dc.b	$EF,$FD,$9C,$7B,$7C,$EF,$4E,$66,$98,$F2,$2C,$EC,$87,$F8,$15,$F0
	.dc.b	$7C,$43,$DA,$F7,$78,$B2,$BF,$C3,$F3,$96,$CE,$5F,$EF,$F3,$CA,$39
	.dc.b	$F7,$7C,$CF,$19,$F4,$54,$CD,$F4,$9C,$3F,$E3,$58,$CB,$FC,$53,$A3
	.dc.b	$FD,$01,$8B,$D6,$E9,$E2,$4A,$F2,$C7,$A8,$C3,$FE,$38,$8C,$3F,$DC
	.dc.b	$FF,$2B,$7C,$47,$E0,$3E,$E1,$EF,$D3,$F2,$94,$AD,$BC,$6B,$4F,$F9
	.dc.b	$8E,$B8,$B9,$21,$E9,$3F,$3A,$E9,$C7,$52,$31,$CE,$D3,$FC,$0F,$87
	.dc.b	$82,$F9,$1F,$E9,$2E,$AF,$CB,$FD,$4B,$E4,$E9,$E8,$F3,$7F,$2B,$87
	.dc.b	$F8,$41,$B7,$FA,$EA,$E0,$FF,$9E,$DA,$F6,$8D,$E6,$FA,$87,$D0,$7E
	.dc.b	$9B,$DC,$F3,$5F,$FA,$0D,$AF,$B2,$76,$FF,$A1,$DA,$FC,$08,$DE,$BF
	.dc.b	$EC,$FC,$BF,$63,$5A,$8E,$7F,$53,$FB,$EB,$FF,$15,$7D,$3B,$CE,$FA
	.dc.b	$AD,$8F,$F1,$74,$ED,$F8,$ED,$AF,$2C,$EE,$7E,$5B,$63,$BF,$8B,$F8
	.dc.b	$0A,$E1,$8F,$F8,$E2,$B4,$DC,$35,$5E,$8F,$8F,$F4,$FC,$A4,$CB,$E0
	.dc.b	$46,$5F,$85,$19,$7E,$34,$7E,$C3,$DD,$FD,$37,$D7,$49,$FD,$7C,$65
	.dc.b	$FF,$4C,$46,$5F,$C6,$68,$EB,$57,$CA,$F9,$7F,$97,$E2,$4B,$FA,$51
	.dc.b	$97,$EA,$46,$5F,$B1,$19,$7E,$D4,$66,$FE,$FF,$1F,$E0,$EB,$EB,$FD
	.dc.b	$57,$D8,$FA,$A9,$87,$DF,$8C,$BF,$81,$8C,$BF,$85,$8F,$BF,$C9,$F8
	.dc.b	$0E,$09,$3F,$87,$8C,$BF,$89,$8C,$BF,$8D,$8F,$DA,$FB,$9F,$C3,$23
	.dc.b	$F0,$B9,$FE,$10,$CF,$83,$23,$24,$3F,$57,$3C,$5B,$1C,$5D,$69,$F9
	.dc.b	$C9,$D3,$FA,$0F,$8B,$FF,$11,$EF,$FD,$65,$B5,$FE,$F6,$6B,$FC,$05
	.dc.b	$7D,$6B,$C2,$FE,$1B,$E3,$FE,$8B,$9D,$6B,$FF,$17,$FC,$47,$5F,$34
	.dc.b	$F0,$AF,$D3,$FC,$74,$BF,$D7,$3B,$9F,$D5,$F5,$7F,$1A,$FA,$6F,$D6
	.dc.b	$5F,$FA,$49,$B7,$FD,$C3,$BB,$F3,$EF,$8A,$33,$D1,$FC,$07,$E8,$BF
	.dc.b	$6A,$73,$F0,$2D,$7E,$F6,$2B,$FC,$5A,$BA,$AF,$A6,$7C,$97,$F2,$C6
	.dc.b	$C7,$FA,$95,$AF,$F2,$7F,$5B,$87,$89,$BA,$FE,$EB,$F0,$3F,$4B,$5F
	.dc.b	$8F,$A2,$31,$87,$CA,$C7,$9F,$3F,$F1,$46,$DF,$0A,$DD,$F1,$E9,$ED
	.dc.b	$87,$8F,$CC,$F3,$11,$FA,$B8,$FE,$8A,$FE,$D4,$6C,$FE,$3D,$F3,$3F
	.dc.b	$C6,$8E,$0F,$F3,$8C,$BF,$9E,$E5,$E8,$E1,$56,$EB,$91,$FC,$FB,$17
	.dc.b	$C6,$7F,$52,$6C,$7F,$B2,$DA,$FF,$44,$F0,$FF,$9C,$E7,$FC,$CA,$72
	.dc.b	$FF,$34,$87,$D1,$7F,$5C,$6D,$73,$BE,$C6,$66,$46,$4F,$F6,$D7,$D2
	.dc.b	$FF,$83,$5F,$F6,$B9,$F9,$87,$0E,$1A,$FF,$B7,$1A,$FF,$B8,$9D,$DF
	.dc.b	$F4,$8F,$A6,$78,$FF,$DC,$1B,$5E,$0F,$5F,$FB,$9D,$F0,$DB,$FD,$D7
	.dc.b	$5A,$AF,$EC,$AD,$C3,$FA,$9D,$6E,$7F,$5B,$F4,$AF,$A8,$FF,$0A,$7C
	.dc.b	$1C,$9F,$F7,$B4,$E5,$B3,$97,$FB,$BF,$B1,$FE,$AC,$C3,$FA,$FD,$6E
	.dc.b	$7F,$BA,$B6,$3F,$93,$CE,$F2,$FC,$BB,$6B,$C7,$7F,$E0,$EB,$FA,$71
	.dc.b	$FE,$07,$F9,$F7,$CC,$FE,$D1,$BD,$FB,$E0,$95,$F9,$4C,$BF,$84,$D6
	.dc.b	$A3,$01,$C3,$E6,$9F,$ED,$FF,$07,$FD,$C3,$EB,$A4,$FF,$77,$8C,$BF
	.dc.b	$DE,$63,$2F,$F7,$D8,$FF,$2B,$7E,$E7,$FE,$08,$8F,$F4,$0B,$5F,$D3
	.dc.b	$B5,$FF,$66,$B7,$63,$F9,$8C,$7F,$77,$F0,$7F,$B4,$3F,$77,$3A,$F6
	.dc.b	$3F,$D0,$3F,$4E,$F4,$3F,$CE,$5E,$64,$CB,$DF,$E2,$2B,$51,$86,$BF
	.dc.b	$E3,$0C,$7F,$72,$DA,$EB,$72,$3E,$4F,$85,$FE,$93,$9E,$D1,$FC,$56
	.dc.b	$F9,$E5,$FE,$22,$C7,$C1,$C3,$FE,$85,$EA,$97,$E1,$8F,$49,$87,$C5
	.dc.b	$18,$7E,$B8,$FA,$6F,$E4,$FE,$5F,$EB,$A3,$1E,$2F,$FA,$01,$FE,$8E
	.dc.b	$7F,$48,$9D,$BF,$B4,$6B,$A3,$FD,$9D,$F1,$9F,$83,$7F,$73,$F1,$48
	.dc.b	$C4,$6E,$31,$CF,$E1,$6C,$75,$F0,$9F,$19,$E5,$C3,$FC,$D2,$D8,$CC
	.dc.b	$EE,$3F,$83,$EA,$F6,$9E,$CF,$F9,$2B,$83,$E5,$46,$39,$FE,$8F,$33
	.dc.b	$C0,$D4,$7F,$66,$C3,$96,$18,$C9,$18,$9F,$C1,$E1,$CE,$7F,$01,$7B
	.dc.b	$AD,$47,$F4,$F9,$FF,$26,$57,$99,$B1,$F2,$67,$D1,$EB,$FF,$97,$A3
	.dc.b	$C5,$3A,$7E,$9C,$EA,$BC,$35,$E5,$E7,$FF,$6E,$71,$F8,$70,$EB,$A7
	.dc.b	$C0,$AC,$1B,$D1,$58,$C3,$F7,$74,$FA,$4F,$F5,$F4,$61,$B9,$FF,$A3
	.dc.b	$3D,$8F,$F2,$24,$E7,$C6,$F8,$D7,$C5,$B7,$C2,$3F,$C7,$B5,$C2,$FF
	.dc.b	$C9,$B3,$9B,$FE,$DF,$D7,$9B,$FE,$3B,$8F,$B6,$31,$3F,$C7,$A3,$1E
	.dc.b	$BF,$C1,$F1,$7D,$73,$FE,$87,$8C,$4F,$D9,$C7,$FD,$67,$E6,$FC,$D7
	.dc.b	$FC,$47,$F0,$5F,$DF,$6F,$F4,$4D,$8F,$EA,$2D,$7F,$3C,$E7,$2F,$F4
	.dc.b	$9B,$67,$F4,$FF,$0E,$F1,$3C,$13,$9B,$EF,$F1,$FE,$7F,$8C,$CB,$70
	.dc.b	$FD,$AD,$23,$CD,$5F,$EC,$A6,$C6,$67,$CF,$18,$39,$BF,$BE,$C3,$F5
	.dc.b	$2F,$48,$F8,$5F,$46,$F8,$0F,$AF,$F4,$8F,$92,$CC,$61,$B7,$FC,$16
	.dc.b	$B8,$36,$7C,$B1,$8C,$BF,$BB,$B7,$5F,$F5,$BB,$6B,$F5,$9F,$61,$FD
	.dc.b	$C9,$FE,$27,$19,$AF,$F2,$AF,$D2,$3E,$57,$F5,$5F,$4B,$FE,$FC,$BF
	.dc.b	$E3,$33,$97,$F9,$9C,$66,$BF,$DB,$33,$97,$FC,$12,$D7,$F8,$B3,$5F
	.dc.b	$E3,$AD,$7C,$35,$F8,$78,$E5,$8F,$FB,$4B,$E9,$5F,$4A,$F9,$FD,$5F
	.dc.b	$E3,$7C,$0F,$F2,$8E,$7F,$ED,$7F,$7B,$1E,$5F,$E6,$5B,$1F,$CB,$E7
	.dc.b	$7A,$7F,$87,$9A,$FF,$2B,$6B,$FC,$C9,$DE,$7F,$EE,$26,$CF,$F9,$5B
	.dc.b	$E0,$DF,$27,$DA,$FF,$DF,$74,$F2,$7F,$83,$57,$F8,$02,$FF,$A1,$CE
	.dc.b	$5F,$EA,$71,$9A,$E3,$5F,$92,$AD,$47,$9E,$30,$4A,$E6,$FF,$21,$F8
	.dc.b	$6F,$C6,$46,$0F,$F7,$D7,$2F,$EE,$CF,$DA,$7C,$C7,$FB,$32,$32,$7F
	.dc.b	$56,$71,$7F,$E4,$26,$BF,$D3,$DA,$FF,$56,$6A,$C3,$FB,$D7,$CA,$76
	.dc.b	$F7,$FF,$85,$36,$FF,$B8,$F8,$6F,$94,$FE,$1C,$7F,$3F,$F8,$97,$D2
	.dc.b	$3E,$97,$37,$E5,$E6,$FD,$AE,$BF,$83,$8F,$ED,$4D,$D7,$D1,$3F,$C3
	.dc.b	$F0,$E7,$BF,$5C,$4D,$F6,$70,$FF,$17,$FA,$47,$D2,$FF,$51,$CD,$F4
	.dc.b	$78,$7F,$BD,$CD,$CD,$CE,$9F,$F5,$F9,$C2,$FA,$1F,$C3,$2B,$85,$7F
	.dc.b	$79,$6B,$FE,$00,$6B,$FE,$C8,$7B,$57,$58,$C4,$F9,$6D,$F5,$DD,$B0
	.dc.b	$F9,$BF,$35,$FC,$E7,$E4,$3F,$9D,$36,$32,$7D,$28,$CB,$4F,$AB,$F4
	.dc.b	$DF,$D0,$7E,$9B,$FA,$13,$63,$27,$DA,$8C,$B4,$FB,$BE,$9F,$F4,$9E
	.dc.b	$DF,$E9,$4D,$8C,$9F,$82,$8C,$B4,$CE,$9F,$31,$FE,$AB,$1A,$8C,$1F
	.dc.b	$8F,$E7,$FD,$8C,$FF,$6A,$F2,$BF,$D5,$AF,$F8,$FB,$D1,$7C,$D7,$F6
	.dc.b	$69,$CD,$FF,$A0,$23,$1F,$19,$2F,$F2,$BF,$11,$6B,$BF,$7F,$F9,$57
	.dc.b	$AB,$FB,$27,$7F,$F6,$56,$DF,$E7,$1B,$5C,$34,$FC,$A7,$37,$D9,$C3
	.dc.b	$FC,$BD,$87,$F6,$5F,$4D,$E2,$7D,$BF,$A6,$C3,$73,$FF,$24,$CE,$9F
	.dc.b	$F2,$43,$72,$FF,$5B,$B5,$FF,$22,$35,$FF,$25,$BB,$C7,$F1,$51,$FE
	.dc.b	$12,$AF,$C5,$7B,$DE,$0B,$FD,$AF,$CB,$7F,$80,$7E,$9B,$FC,$04,$D8
	.dc.b	$C9,$FB,$A8,$CB,$4F,$F2,$FF,$D3,$7D,$4C,$3E,$4F,$D3,$3D,$7F,$F0
	.dc.b	$24,$E7,$3F,$89,$39,$BE,$3C,$FA,$3F,$C5,$D1,$FE,$47,$8C,$9F,$13
	.dc.b	$2F,$F1,$97,$D3,$3E,$9B,$E6,$46,$1B,$3F,$22,$3F,$BE,$EE,$69,$23
	.dc.b	$A1,$FE,$AE,$F9,$B7,$EF,$1F,$E0,$BD,$BF,$C7,$35,$F8,$BF,$77,$FC
	.dc.b	$45,$F4,$97,$0F,$C8,$36,$BF,$2F,$B3,$B3,$FC,$29,$CF,$FE,$9F,$F4
	.dc.b	$B5,$3F,$E6,$79,$FF,$3C,$61,$F9,$9F,$23,$F3,$1F,$17,$FE,$2F,$6C
	.dc.b	$7E,$7E,$3F,$C6,$DE,$96,$32,$EF,$7F,$A0,$AF,$F7,$DE,$1E,$1F,$03
	.dc.b	$FC,$64,$D8,$CC,$EE,$BF,$D1,$93,$97,$9D,$7F,$97,$FE,$0B,$FA,$6C
	.dc.b	$62,$FF,$D5,$71,$8A,$FF,$27,$FA,$7B,$EC,$E0,$CA,$DB,$96,$9F,$B9
	.dc.b	$F3,$7C,$B9,$1A,$48,$F3,$BF,$BB,$3C,$9E,$9D,$8F,$F6,$3F,$D2,$FE
	.dc.b	$0D,$FF,$4A,$B7,$3F,$F4,$03,$5F,$F3,$EB,$5F,$F4,$93,$BC,$BF,$D8
	.dc.b	$4D,$94,$FF,$5F,$CF,$FA,$D4,$78,$3A,$3F,$D2,$55,$BA,$FF,$4C,$B7
	.dc.b	$89,$7C,$27,$F8,$DF,$C3,$AB,$83,$6F,$99,$F4,$7D,$5B,$8F,$BB,$E1
	.dc.b	$BF,$CD,$95,$C5,$B7,$FE,$99,$C3,$D7,$18,$3C,$A7,$DB,$F3,$6D,$FF
	.dc.b	$B5,$A7,$DB,$EA,$FB,$BE,$9D,$F2,$7F,$EC,$FF,$1B,$FA,$94,$62,$BF
	.dc.b	$C0,$81,$C3,$F1,$A3,$73,$FE,$0B,$8F,$F6,$03,$CE,$FE,$5F,$C5,$70
	.dc.b	$9F,$09,$A3,$FD,$C7,$5F,$D8,$63,$50,$C3,$CD,$7C,$E3,$C4,$F7,$C3
	.dc.b	$A5,$F8,$A7,$CF,$18,$C3,$8C,$62,$B9,$4F,$6B,$F2,$C7,$5C,$31,$5E
	.dc.b	$53,$DD,$7A,$79,$1E,$87,$AB,$FE,$27,$3C,$A6,$4A,$66,$FC,$8E,$0F
	.dc.b	$86,$E0,$F0,$96,$BD,$C7,$A0,$F9,$D5,$C8,$F3,$95,$DC,$F8,$67,$9C
	.dc.b	$71,$8F,$B7,$3B,$6C,$7C,$13,$A9,$5C,$9E,$B8,$F8,$BE,$0F,$FB,$F6
	.dc.b	$3F,$0A,$D7,$E1,$35,$FB,$EE,$7E,$37,$FA,$37,$1F,$EB,$A3,$1E,$3F
	.dc.b	$F8,$BA,$BF,$DD,$37,$FB,$25,$7C,$1C,$9F,$B6,$71,$5C,$AF,$EB,$4F
	.dc.b	$0E,$AE,$57,$F9,$8B,$FF,$03,$CE,$9E,$91,$9C,$3C,$4E,$EB,$8E,$3E
	.dc.b	$3A,$E3,$7F,$85,$78,$C6,$6F,$D5,$97,$1A,$FC,$2D,$70,$8E,$D8,$E2
	.dc.b	$BB,$DF,$5C,$3B,$E1,$ED,$6B,$F0,$A7,$6F,$C1,$7B,$CF,$18,$EF,$3F
	.dc.b	$26,$FB,$C7,$E4,$B0,$F9,$51,$E5,$8C,$4E,$EB,$D5,$87,$94,$E7,$30
	.dc.b	$E5,$87,$81,$F1,$47,$CB,$AF,$3E,$1F,$1B,$1F,$09,$E2,$C7,$C0,$4C
	.dc.b	$7B,$E1,$D3,$0E,$D3,$98,$EA,$64,$FA,$18,$76,$3B,$E3,$F1,$23,$E3
	.dc.b	$57,$2B,$F8,$91,$C6,$FF,$8E,$70,$FF,$33,$F9,$0F,$C7,$5F,$F4,$26
	.dc.b	$EB,$FD,$19,$B7,$FD,$0F,$D4,$B7,$C6,$3F,$A8,$CF,$E5,$DF,$E9,$ED
	.dc.b	$79,$1F,$D5,$BE,$9F,$8F,$21,$A7,$6E,$8F,$E9,$3E,$A6,$2B,$FA,$FC
	.dc.b	$D7,$FC,$AF,$D2,$9C,$69,$FD,$6A,$AE,$9F,$D6,$B0,$E7,$87,$23,$C4
	.dc.b	$43,$FB,$3C,$F9,$DF,$F4,$FF,$63,$93,$E0,$E1,$FE,$1D,$6C,$66,$7F
	.dc.b	$C0,$BE,$AE,$27,$95,$6E,$7F,$C2,$31,$8A,$E1,$1F,$EF,$3C,$DE,$AF
	.dc.b	$C1,$6C,$7F,$84,$23,$1E,$66,$59,$19,$3F,$C2,$B1,$E6,$F8,$A7,$7D
	.dc.b	$24,$EC,$FF,$1B,$C6,$9F,$F1,$4B,$5F,$F1,$2B,$5F,$C2,$B5,$FF,$02
	.dc.b	$E4,$8F,$F6,$F4,$63,$6C,$C6,$23,$BF,$77,$DD,$F4,$9F,$13,$E7,$8F
	.dc.b	$C0,$7A,$EF,$9D,$A6,$6F,$A1,$FE,$D2,$C1,$D1,$EF,$24,$7F,$BC,$3E
	.dc.b	$C2,$BF,$E9,$CB,$FF,$17,$3F,$EA,$7F,$17,$3A,$FF,$A1,$BE,$9F,$C7
	.dc.b	$3D,$83,$87,$FD,$66,$63,$4F,$D1,$FA,$2F,$D7,$1F,$4B,$FA,$6A,$D1
	.dc.b	$F8,$BB,$E4,$FE,$95,$AF,$F3,$D6,$BF,$E3,$07,$C1,$87,$FA,$AF,$EA
	.dc.b	$5E,$6F,$E9,$5B,$5E,$7A,$DF,$1F,$F4,$7F,$37,$F9,$AE,$1F,$EA,$26
	.dc.b	$C6,$7C,$5F,$D7,$7D,$4B,$E7,$BD,$15,$AC,$5E,$02,$9F,$37,$C0,$DE
	.dc.b	$1F,$B7,$FA,$27,$D3,$7D,$DC,$E6,$3F,$63,$15,$6A,$30,$D7,$FD,$95
	.dc.b	$3C,$1F,$F6,$9C,$67,$4F,$F5,$DF,$D4,$BC,$7F,$9C,$DA,$FB,$0D,$7E
	.dc.b	$BF,$D4,$BE,$33,$75,$FB,$D9,$56,$DF,$E3,$0D,$7F,$B8,$35,$FF,$31
	.dc.b	$BD,$6F,$FD,$9B,$6F,$4F,$F0,$0D,$AF,$1D,$FC,$5F,$94,$F8,$DF,$35
	.dc.b	$27,$E0,$C6,$5F,$87,$19,$7E,$0C,$7E,$EF,$E1,$24,$7E,$8B,$1F,$D8
	.dc.b	$47,$D1,$F5,$F3,$3F,$9F,$F9,$EF,$D5,$98,$AF,$F7,$A6,$CF,$9C,$FB
	.dc.b	$1E,$54,$BF,$A7,$1F,$6C,$C3,$F4,$A3,$0E,$8F,$B2,$EA,$BA,$CF,$E0
	.dc.b	$FE,$A7,$F0,$9D,$B2,$FE,$FC,$65,$FC,$0C,$65,$FB,$F1,$F8,$BF,$23
	.dc.b	$F1,$9C,$32,$7F,$0F,$19,$7F,$13,$19,$7F,$0F,$1F,$B1,$F8,$1F,$E1
	.dc.b	$51,$FB,$B6,$BF,$BA,$6E,$3F,$6B,$0E,$1E,$47,$E1,$67,$75,$FD,$6D
	.dc.b	$AF,$F8,$CD,$AF,$FA,$99,$E7,$CD,$FE,$37,$E4,$FF,$50,$CF,$F0,$B6
	.dc.b	$BE,$1A,$FE,$D7,$F2,$1F,$D3,$7C,$9B,$E7,$FF,$2B,$AF,$E3,$F3,$96
	.dc.b	$F8,$BF,$C6,$9F,$E2,$7E,$FB,$2F,$CE,$36,$3F,$AD,$35,$FE,$E4,$D7
	.dc.b	$FC,$99,$D9,$FB,$4E,$CF,$DD,$12,$3C,$38,$BD,$DF,$DF,$76,$7F,$1D
	.dc.b	$FA,$8F,$DA,$65,$FE,$55,$6D,$FF,$08,$E8,$FD,$34,$E7,$BF,$F6,$15
	.dc.b	$FB,$FA,$FF,$03,$35,$D1,$FE,$07,$CF,$F8,$3F,$D4,$FF,$0D,$24,$7F
	.dc.b	$35,$9E,$EB,$EA,$5F,$07,$FC,$C5,$B1,$FC,$ED,$AF,$FB,$49,$78,$DB
	.dc.b	$9F,$E0,$91,$A9,$FF,$0D,$35,$FF,$39,$35,$FF,$72,$BD,$3A,$3E,$86
	.dc.b	$2F,$85,$FC,$1D,$7F,$2A,$6B,$E7,$BF,$37,$97,$23,$CC,$FF,$88,$67
	.dc.b	$EF,$5F,$D9,$C3,$F9,$FD,$7F,$8A,$DA,$F9,$AB,$D2,$47,$F9,$6F,$D4
	.dc.b	$FF,$35,$3B,$3F,$D6,$A1,$F1,$FF,$D5,$DB,$1F,$D8,$5A,$FF,$60,$F1
	.dc.b	$3F,$A0,$B9,$9F,$E5,$53,$A8,$FF,$20,$35,$FF,$4C,$35,$1F,$D6,$A3
	.dc.b	$FB,$34,$BE,$9A,$9E,$38,$FE,$D5,$F5,$1F,$E2,$2A,$FE,$E2,$DC,$BF
	.dc.b	$38,$DE,$2F,$F5,$37,$D4,$FF,$8B,$27,$3E,$D7,$F3,$FA,$E5,$7F,$E5
	.dc.b	$56,$BF,$EA,$D6,$BE,$67,$A3,$E8,$A7,$8F,$FD,$F5,$B1,$FE,$07,$9D
	.dc.b	$AF,$A9,$FE,$D8,$FF,$82,$DA,$FF,$82,$9B,$87,$ED,$FE,$9D,$EC,$3E
	.dc.b	$4B,$FC,$15,$5C,$2F,$FC,$D8,$D7,$FD,$88,$D7,$E6,$B5,$F1,$7A,$7F
	.dc.b	$E2,$CF,$87,$91,$E3,$8E,$BC,$12,$3C,$CF,$F9,$FB,$17,$D6,$FD,$59
	.dc.b	$C9,$FE,$65,$C3,$0F,$F9,$F5,$AF,$9A,$BF,$CD,$5D,$1F,$E0,$0F,$87
	.dc.b	$FF,$03,$7D,$2C,$AF,$F0,$0C,$7F,$82,$4C,$3F,$E0,$08,$C3,$FD,$7F
	.dc.b	$EA,$BF,$22,$B7,$FE,$3C,$23,$2E,$1B,$C9,$FE,$81,$F4,$5F,$2F,$FE
	.dc.b	$7A,$9C,$DF,$36,$DF,$3F,$63,$FC,$E5,$5F,$E9,$EF,$AA,$7A,$34,$98
	.dc.b	$7F,$87,$71,$75,$E2,$3C,$D9,$FE,$83,$E2,$98,$72,$9F,$F2,$DC,$63
	.dc.b	$0F,$F4,$05,$F1,$9F,$F8,$01,$39,$7F,$D5,$4D,$7F,$D5,$CD,$77,$5F
	.dc.b	$CC,$14,$78,$3C,$7F,$51,$FE,$8D,$F9,$8B,$3E,$2A,$E1,$5B,$8F,$64
	.dc.b	$FF,$96,$FE,$AB,$F9,$6F,$97,$FE,$A5,$6C,$FF,$AF,$5A,$FF,$B0,$5A
	.dc.b	$FF,$00,$1F,$1E,$F9,$37,$4F,$FA,$BE,$30,$D8,$CC,$FF,$B2,$67,$2F
	.dc.b	$7F,$0D,$F2,$5E,$36,$C6,$63,$07,$C8,$F7,$7F,$C4,$1C,$3F,$41,$B3
	.dc.b	$F5,$8F,$F2,$FF,$D5,$3B,$FF,$1B,$5F,$66,$B7,$CF,$FD,$B7,$81,$18
	.dc.b	$9D,$7B,$9C,$31,$E7,$85,$BD,$CD,$BF,$F2,$45,$F8,$FD,$BE,$1D,$AC
	.dc.b	$FF,$DA,$13,$92,$65,$E7,$C7,$C3,$7C,$2B,$73,$9C,$3A,$B6,$7C,$06
	.dc.b	$F8,$AE,$1F,$E4,$BF,$AA,$FF,$68,$B7,$C8,$F1,$F5,$7F,$AF,$63,$04
	.dc.b	$C3,$E4,$6E,$BE,$31,$FE,$F4,$F6,$3E,$14,$E7,$0F,$6F,$5F,$3A,$FF
	.dc.b	$5A,$65,$8B,$F8,$3F,$11,$EA,$C3,$FB,$5D,$ED,$FE,$DB,$19,$CB,$FD
	.dc.b	$F9,$39,$7F,$DF,$D1,$9C,$3F,$A7,$FC,$3F,$F7,$39,$FF,$5E,$36,$3F
	.dc.b	$D7,$7F,$1A,$F0,$BE,$C4,$7F,$B5,$3D,$0B,$5B,$CB,$FC,$54,$DC,$3F
	.dc.b	$39,$87,$17,$AD,$70,$FA,$A9,$1F,$EC,$B9,$FC,$BB,$E2,$8C,$E1,$FE
	.dc.b	$CC,$B4,$78,$AB,$73,$98,$DE,$1E,$27,$5D,$18,$C3,$F5,$DF,$50,$F8
	.dc.b	$4F,$F1,$F6,$1F,$DC,$23,$13,$FB,$38,$DC,$E9,$4E,$9D,$18,$8F,$8B
	.dc.b	$19,$C3,$F4,$5F,$53,$FD,$92,$FF,$DD,$13,$97,$A9,$8A,$FD,$14,$EE
	.dc.b	$FF,$DD,$75,$B7,$FC,$0B,$F5,$0F,$A6,$FD,$47,$D4,$BE,$0F,$EA,$CE
	.dc.b	$72,$FE,$53,$39,$7F,$DF,$11,$9B,$FF,$7B,$B7,$3F,$F3,$27,$D4,$BE
	.dc.b	$A5,$E3,$F4,$CE,$63,$C9,$FD,$D5,$6E,$73,$8F,$F0,$73,$FD,$CB,$ED
	.dc.b	$FE,$73,$DC,$FC,$4F,$85,$FE,$52,$C7,$FD,$C1,$D3,$FE,$4E,$7F,$C4
	.dc.b	$3A,$7F,$9C,$E7,$58,$73,$6C,$7F,$92,$BE,$09,$F2,$BD,$32,$98,$72
	.dc.b	$3F,$CF,$2F,$F0,$89,$DC,$67,$0F,$F1,$7E,$1D,$23,$FC,$FF,$E3,$EF
	.dc.b	$CF,$F3,$B6,$38,$57,$FC,$00,$8C,$C7,$F5,$6F,$84,$7C,$EE,$2B,$D1
	.dc.b	$57,$4F,$E4,$95,$AB,$E7,$8A,$BF,$80,$E1,$27,$BF,$CE,$FF,$68,$7E
	.dc.b	$3C,$7E,$CA,$31,$8F,$CF,$6E,$3F,$C1,$3E,$5F,$5E,$9F,$F0,$C6,$E1
	.dc.b	$F4,$1B,$7F,$CF,$72,$FE,$F3,$F2,$FF,$C4,$8F,$E6,$DF,$14,$FA,$87
	.dc.b	$D4,$E7,$1F,$2F,$37,$8C,$DC,$B8,$31,$3F,$EB,$68,$C3,$63,$7F,$31
	.dc.b	$C1,$FE,$1B,$F5,$2F,$A9,$A7,$F4,$7A,$7D,$4B,$EA,$73,$1F,$52,$B8
	.dc.b	$36,$B8,$56,$E3,$04,$DF,$F2,$D5,$C1,$FE,$BD,$1D,$AF,$AC,$7C,$5E
	.dc.b	$A7,$CE,$FF,$36,$F9,$1F,$E6,$ED,$BF,$CD,$B6,$B8,$37,$97,$F9,$BF
	.dc.b	$CF,$7F,$3E,$FA,$0F,$E7,$ED,$BF,$D1,$B6,$FC,$0D,$AF,$E7,$FE,$97
	.dc.b	$F4,$6E,$FF,$E8,$ED,$BF,$D5,$B6,$FD,$0D,$AE,$33,$FB,$0A,$E2,$D3
	.dc.b	$F4,$90,$F3,$3F,$D4,$31,$86,$DF,$C3,$EE,$FE,$B3,$18,$AF,$F2,$7B
	.dc.b	$6B,$F7,$6D,$A7,$4F,$F9,$1A,$F8,$FC,$6F,$F6,$0E,$7F,$EC,$3D,$97
	.dc.b	$1F,$E0,$AD,$C3,$E9,$37,$2F,$EC,$2F,$F5,$4F,$4F,$F8,$54,$FF,$11
	.dc.b	$6C,$FF,$BF,$EB,$F1,$3E,$6E,$FC,$49,$1F,$BC,$AD,$BF,$EB,$23,$66
	.dc.b	$63,$77,$D5,$FC,$6D,$62,$FF,$B9,$FC,$B7,$F7,$AF,$A9,$FE,$F6,$DC
	.dc.b	$3F,$8F,$B7,$0F,$C6,$36,$93,$FE,$43,$FA,$9F,$E2,$F5,$FC,$AD,$B3
	.dc.b	$FE,$EA,$FA,$9F,$F1,$95,$7F,$14,$9C,$BC,$78,$BF,$BB,$C6,$2F,$E8
	.dc.b	$D7,$07,$A4,$FC,$9C,$25,$FF,$25,$F9,$87,$D4,$7F,$7B,$BF,$F6,$AE
	.dc.b	$CB,$DC,$7C,$F7,$FC,$79,$5C,$3D,$6C,$E9,$EC,$AE,$35,$8C,$7F,$DB
	.dc.b	$F7,$B7,$E9,$47,$E3,$2E,$47,$F4,$6F,$2E,$DF,$F8,$B3,$9B,$C9,$8F
	.dc.b	$E5,$EF,$FC,$53,$5A,$8F,$C9,$FC,$1F,$E5,$8F,$F1,$6F,$97,$71,$FC
	.dc.b	$E4,$5C,$7F,$7D,$36,$BF,$37,$F3,$FF,$9E,$3F,$C7,$FE,$65,$C3,$F4
	.dc.b	$91,$6F,$F9,$57,$BB,$FC,$F5,$F3,$47,$09,$FF,$51,$57,$F9,$23,$9B
	.dc.b	$15,$C2,$2E,$1F,$1F,$AD,$8F,$E5,$2B,$F5,$2F,$59,$FD,$7F,$8F,$FB
	.dc.b	$23,$FC,$A7,$E6,$5C,$BF,$71,$16,$FC,$B3,$6B,$FC,$ED,$F4,$AF,$9F
	.dc.b	$FF,$39,$37,$0F,$F5,$C7,$9D,$89,$91,$FE,$B4,$9D,$61,$F0,$2B,$2F
	.dc.b	$F9,$C6,$33,$7D,$AF,$FC,$CF,$5A,$8C,$5F,$39,$CB,$4E,$DF,$01,$E4
	.dc.b	$3F,$CD,$FE,$35,$8C,$93,$1F,$A3,$19,$7F,$C2,$40,$7A,$17,$E9,$9F
	.dc.b	$1D,$FE,$3F,$F0,$7F,$2D,$18,$AF,$F2,$15,$6E,$B5,$85,$9C,$8A,$E0
	.dc.b	$F7,$BE,$75,$FE,$46,$BE,$13,$9B,$AF,$96,$2D,$70,$8E,$B7,$D1,$FF
	.dc.b	$2F,$B5,$F5,$9E,$1B,$F2,$B4,$E2,$E2,$38,$57,$6B,$F0,$D7,$29,$ED
	.dc.b	$7E,$18,$FF,$6A,$56,$DE,$86,$0A,$79,$2F,$DC,$E2,$79,$3E,$D3,$17
	.dc.b	$E0,$AE,$97,$E4,$9C,$9D,$0F,$25,$FA,$4E,$B1,$E6,$AE,$53,$C2,$39
	.dc.b	$19,$79,$57,$6A,$E2,$73,$8F,$8A,$7F,$BC,$6F,$CE,$F8,$0C,$5F,$6E
	.dc.b	$CF,$F7,$04,$7E,$01,$AF,$C3,$6B,$EF,$73,$F1,$0F,$A7,$6D,$FC,$AB
	.dc.b	$E2,$7F,$E0,$04,$E6,$31,$7F,$97,$8C,$2B,$E7,$7C,$A3,$17,$CA,$31
	.dc.b	$A7,$E4,$AD,$D1,$FA,$1B,$FB,$31,$87,$F2,$6F,$69,$E1,$5C,$B4,$F2
	.dc.b	$DF,$19,$E8,$D8,$F0,$65,$CA,$B8,$61,$E9,$AD,$34,$D4,$C7,$C7,$1A
	.dc.b	$58,$79,$5F,$3D,$74,$72,$D8,$D6,$1E,$49,$E8,$FA,$AF,$DB,$7D,$6B
	.dc.b	$78,$7A,$2F,$AD,$69,$86,$09,$7D,$5F,$39,$DA,$79,$46,$48,$E0,$91
	.dc.b	$F0,$4E,$27,$47,$2F,$C0,$BF,$31,$E7,$C3,$DC,$7B,$DC,$C7,$84,$7C
	.dc.b	$43,$8D,$F1,$BF,$7F,$2F,$E7,$AD,$DB,$F9,$15,$6E,$31,$7F,$D3,$3C
	.dc.b	$EB,$39,$8F,$E7,$D3,$8C,$3A,$C7,$FA,$A7,$A7,$FA,$AD,$E7,$E9,$E4
	.dc.b	$FF,$45,$7A,$CF,$F5,$AF,$3F,$33,$FD,$16,$6E,$7F,$D7,$FA,$D5,$8D
	.dc.b	$3C,$F9,$2F,$18,$7F,$62,$AB,$A7,$C1,$9C,$E5,$DE,$BF,$CB,$1C,$7E
	.dc.b	$EC,$FF,$CB,$8D,$F0,$BF,$59,$3E,$BD,$1E,$7E,$6A,$4F,$23,$51,$EC
	.dc.b	$97,$9F,$49,$3C,$FC,$8A,$FF,$81,$22,$E3,$EF,$9F,$27,$C3,$BB,$E9
	.dc.b	$2F,$C9,$75,$FE,$13,$C3,$FE,$E6,$AE,$11,$FE,$78,$6B,$DC,$C7,$7F
	.dc.b	$62,$3F,$EF,$0F,$85,$AF,$43,$F8,$AC,$A7,$39,$7F,$84,$FD,$25,$FA
	.dc.b	$39,$66,$1E,$17,$EF,$CB,$E5,$B3,$3D,$6D,$8F,$FA,$AA,$38,$1F,$8B
	.dc.b	$AE,$0F,$FB,$BB,$E9,$7B,$B7,$1F,$CB,$A8,$CB,$E9,$F3,$1F,$4E,$FA
	.dc.b	$3D,$D6,$B2,$FF,$29,$5F,$1C,$32,$7E,$8B,$0E,$37,$89,$F0,$10,$ED
	.dc.b	$87,$57,$FC,$19,$7D,$3C,$9F,$E3,$0D,$ED,$F3,$C6,$38,$3F,$C2,$5D
	.dc.b	$0F,$A6,$7C,$CE,$AE,$63,$D2,$FF,$9C,$61,$C6,$3F,$D5,$37,$C1,$FD
	.dc.b	$97,$D1,$BE,$8F,$E8,$DC,$9C,$C7,$EC,$A7,$51,$F3,$E7,$4F,$FC,$01
	.dc.b	$8C,$BF,$83,$1F,$4D,$FC,$01,$C3,$74,$F9,$6D,$AF,$E9,$DF,$4A,$F8
	.dc.b	$5F,$AF,$39,$BF,$F1,$93,$5E,$BE,$27,$DC,$AE,$11,$F0,$4D,$C7,$22
	.dc.b	$39,$11,$C1,$F5,$7C,$77,$BF,$E4,$E4,$7C,$68,$F8,$46,$1F,$81,$18
	.dc.b	$7F,$55,$EE,$7E,$D0,$8F,$CE,$F4,$6B,$F9,$BE,$A4,$E3,$5F,$F3,$C6
	.dc.b	$DF,$E6,$BA,$FA,$BF,$2B,$F4,$7C,$29,$18,$21,$F4,$5C,$BF,$4E,$3E
	.dc.b	$C1,$8A,$EB,$38,$9F,$C0,$FD,$37,$DD,$E9,$91,$82,$1F,$75,$CB,$F7
	.dc.b	$E3,$F0,$9D,$BF,$85,$F4,$E4,$FE,$32,$32,$FE,$3A,$32,$FE,$1E,$3F
	.dc.b	$67,$ED,$3E,$3E,$92,$7E,$B7,$17,$E9,$DF,$C6,$4E,$DF,$B4,$D7,$33
	.dc.b	$F8,$3C,$79,$46,$0D,$C6,$5F,$C3,$7C,$57,$F8,$97,$83,$FA,$DB,$63
	.dc.b	$3E,$4F,$F1,$07,$7E,$03,$D7,$FE,$17,$F1,$1F,$C1,$38,$7F,$82,$B7
	.dc.b	$AF,$FC,$50,$DB,$63,$D5,$FD,$47,$5F,$EB,$48,$F8,$3D,$8F,$D8,$75
	.dc.b	$FF,$18,$FA,$5F,$D6,$E5,$FC,$49,$B9,$7F,$04,$E8,$FD,$24,$63,$0F
	.dc.b	$F4,$C3,$53,$9F,$0D,$3F,$AD,$6F,$D3,$7E,$FC,$95,$FA,$7A,$E5,$F4
	.dc.b	$CF,$7F,$F9,$5B,$67,$F5,$8D,$7F,$A2,$7A,$68,$78,$BC,$2A,$57,$07
	.dc.b	$FC,$13,$3A,$8C,$1F,$E0,$7F,$80,$FB,$DA,$FF,$B7,$5B,$AF,$F9,$56
	.dc.b	$BC,$13,$FC,$D9,$7B,$FE,$62,$5F,$3E,$4A,$FF,$17,$9D,$BE,$A9,$FF
	.dc.b	$08,$E1,$F7,$E3,$07,$F2,$8F,$A6,$FE,$48,$72,$C3,$FA,$2D,$3E,$1F
	.dc.b	$FA,$7B,$6B,$F8,$83,$5F,$E4,$DE,$1B,$E7,$B7,$52,$76,$FF,$58,$9D
	.dc.b	$46,$0F,$E3,$4B,$E8,$34,$71,$8F,$E6,$5F,$49,$FE,$0E,$7F,$90,$35
	.dc.b	$CF,$AB,$FD,$93,$E9,$7F,$C3,$71,$83,$71,$FD,$A2,$78,$67,$FD,$8A
	.dc.b	$B6,$FF,$64,$8E,$95,$FE,$87,$0E,$1F,$F4,$53,$5F,$F7,$63,$57,$D1
	.dc.b	$FF,$55,$7F,$A5,$36,$7F,$B8,$B7,$87,$FD,$E9,$F4,$1F,$E0,$7F,$A5
	.dc.b	$FE,$FB,$39,$BF,$F7,$C3,$5E,$BE,$9F,$F9,$62,$B8,$47,$F4,$39,$D1
	.dc.b	$27,$5E,$BA,$FA,$EB,$2F,$83,$CF,$E9,$7E,$94,$60,$87,$AA,$34,$7F
	.dc.b	$6B,$F8,$3F,$ED,$9F,$3F,$27,$FB,$9C,$65,$FE,$EB,$19,$7F,$B7,$47
	.dc.b	$F9,$4F,$E9,$7F,$07,$1F,$E2,$56,$EC,$DB,$FC,$1B,$18,$7F,$C1,$3E
	.dc.b	$83,$E4,$B9,$73,$4C,$3F,$B4,$60,$C3,$84,$3C,$16,$1C,$63,$FB,$77
	.dc.b	$7B,$E9,$5E,$87,$FA,$09,$B3,$A6,$BF,$E4,$7B,$E7,$D3,$BA,$93,$B7
	.dc.b	$FC,$CD,$18,$8F,$E1,$75,$CB,$2F,$74,$7F,$A6,$A3,$1A,$FB,$AB,$C5
	.dc.b	$E2,$CA,$F1,$3F,$E7,$1F,$93,$C4,$74,$8B,$87,$FA,$2E,$7F,$C8,$FF
	.dc.b	$4B,$FC,$7F,$CD,$FC,$E5,$84,$65,$CB,$5F,$D1,$35,$FF,$7B,$FC,$83
	.dc.b	$AF,$9C,$7F,$B6,$23,$0F,$6F,$8C,$46,$23,$F5,$91,$89,$E3,$5F,$C7
	.dc.b	$BE,$3A,$F8,$46,$92,$FF,$69,$C4,$FD,$83,$FD,$49,$C1,$FE,$99,$F9
	.dc.b	$5F,$97,$BF,$CC,$AD,$7C,$5E,$2B,$F1,$35,$70,$FF,$64,$E1,$C9,$EA
	.dc.b	$D7,$E3,$1A,$AF,$F3,$4B,$FE,$CA,$8D,$3F,$EC,$CF,$CA,$F9,$63,$FE
	.dc.b	$BB,$F6,$3B,$64,$9C,$E5,$F9,$A9,$CC,$62,$FF,$2D,$D0,$9C,$E5,$E7
	.dc.b	$E8,$FC,$CC,$63,$0F,$B7,$8F,$5A,$F2,$E1,$FE,$E4,$C7,$CA,$D8,$CE
	.dc.b	$4F,$DD,$3F,$55,$87,$EA,$1B,$5F,$9D,$FB,$8F,$F4,$54,$E7,$B7,$E4
	.dc.b	$5F,$16,$DF,$08,$FC,$DE,$8B,$E3,$87,$FA,$86,$B5,$87,$F7,$FF,$99
	.dc.b	$FA,$52,$FD,$35,$F0,$FE,$FF,$55,$5A,$9C,$DF,$FA,$02,$31,$5B,$6C
	.dc.b	$7F,$99,$7E,$E1,$F1,$FF,$7F,$9B,$C9,$E9,$F8,$27,$35,$FE,$6F,$9C
	.dc.b	$BC,$6B,$86,$BF,$E2,$78,$C1,$B7,$1A,$7F,$A1,$FE,$99,$F0,$FF,$1B
	.dc.b	$EC,$1F,$35,$EE,$9C,$E1,$FC,$5A,$72,$FF,$9E,$E3,$37,$FE,$C7,$9C
	.dc.b	$BD,$6F,$A5,$F3,$7E,$FC,$2B,$8B,$FD,$65,$AF,$E6,$9A,$FF,$01,$F9
	.dc.b	$9F,$C7,$61,$FE,$8F,$F9,$E7,$E8,$4E,$EF,$AB,$5A,$BF,$F7,$35,$6A
	.dc.b	$73,$D9,$FA,$B8,$C6,$1F,$E9,$5B,$E1,$D1,$FC,$7A,$73,$87,$FA,$B6
	.dc.b	$73,$5F,$AE,$6C,$7F,$37,$E6,$FE,$B1,$3C,$F0,$F1,$FB,$7F,$A3,$FD
	.dc.b	$89,$F2,$98,$AF,$F5,$47,$5A,$73,$1E,$3A,$E1,$5B,$C7,$F9,$D5,$69
	.dc.b	$F2,$4E,$74,$FC,$DC,$BE,$67,$18,$7F,$04,$F8,$4F,$CE,$D7,$F3,$A8
	.dc.b	$7D,$0B,$E0,$FF,$3B,$5A,$9F,$E6,$06,$A3,$3F,$15,$FC,$4E,$30,$7F
	.dc.b	$AF,$67,$86,$7F,$D8,$EF,$A4,$E6,$B8,$BF,$EC,$73,$C3,$F0,$18,$AF
	.dc.b	$F6,$D4,$7F,$32,$35,$0F,$B8,$FD,$8C,$66,$59,$7A,$EB,$53,$FE,$CF
	.dc.b	$35,$19,$AE,$3E,$6F,$A6,$B7,$7E,$08,$CD,$F9,$71,$7A,$F8,$AF,$F7
	.dc.b	$87,$77,$EF,$CF,$E0,$7F,$4C,$F8,$1C,$61,$FE,$DF,$8C,$3F,$F0,$02
	.dc.b	$77,$3A,$AF,$F7,$D1,$89,$FC,$1F,$37,$D0,$8F,$F7,$E5,$F2,$7F,$BD
	.dc.b	$7E,$D7,$E8,$DB,$F8,$CB,$6F,$FB,$37,$37,$E2,$F1,$FE,$65,$39,$BF
	.dc.b	$ED,$58,$7F,$1C,$7F,$C1,$DC,$9D,$71,$FE,$FB,$09,$D6,$1C,$F1,$FC
	.dc.b	$BD,$E0,$CE,$0F,$7F,$F3,$DE,$4F,$F2,$6F,$33,$F8,$2E,$7F,$E0,$8D
	.dc.b	$9F,$79,$ED,$8E,$F3,$AC,$BF,$BF,$56,$B5,$E5,$E0,$FF,$23,$C3,$85
	.dc.b	$7F,$5E,$BE,$15,$FD,$C6,$72,$FF,$81,$E3,$37,$CE,$BF,$C1,$58,$78
	.dc.b	$30,$E9,$7F,$1B,$DB,$F8,$FF,$09,$6F,$F6,$BA,$FF,$A3,$5B,$3A,$69
	.dc.b	$A9,$F9,$FF,$6D,$F4,$3E,$DE,$CF,$26,$E1,$F6,$9B,$97,$DA,$7E,$BF
	.dc.b	$D5,$FD,$8F,$AB,$B7,$E2,$6D,$FE,$09,$B7,$F8,$26,$CF,$E0,$B3,$46
	.dc.b	$2B,$FC,$99,$18,$9F,$F2,$8F,$5B,$E0,$7F,$AD,$7A,$7F,$DA,$23,$13
	.dc.b	$F9,$48,$C4,$EA,$38,$B6,$FF,$2B,$D9,$F9,$0F,$B0,$FC,$8B,$6B,$E5
	.dc.b	$B6,$77,$1F,$3A,$7F,$11,$DF,$FD,$87,$EE,$FF,$19,$3F,$60,$FC,$04
	.dc.b	$67,$A3,$71,$FD,$8E,$77,$7F,$C1,$23,$66,$67,$71,$93,$FC,$71,$3C
	.dc.b	$36,$FC,$9D,$F5,$BF,$CF,$FC,$3F,$F3,$9F,$9F,$FE,$74,$D8,$C9,$FA
	.dc.b	$48,$CB,$4F,$F3,$EF,$E6,$CF,$95,$FF,$12,$7A,$6F,$9C,$FF,$0A,$CE
	.dc.b	$6B,$FA,$1C,$E5,$E3,$5C,$2B,$6D,$8F,$89,$A7,$E7,$FE,$13,$FC,$3F
	.dc.b	$FB,$97,$F3,$28,$77,$27,$3A,$FE,$D5,$B7,$CA,$B7,$18,$7E,$31,$88
	.dc.b	$FF,$4B,$DF,$5B,$E9,$18,$BF,$F1,$2F,$3F,$AE,$FF,$D2,$1F,$71,$FD
	.dc.b	$99,$FC,$9E,$5F,$E4,$58,$C4,$EF,$83,$FC,$67,$18,$9D,$79,$FF,$1B
	.dc.b	$E9,$BE,$4F,$CE,$5C,$FE,$3B,$72,$FF,$4F,$70,$23,$19,$7F,$8F,$A3
	.dc.b	$15,$FA,$0F,$8F,$FF,$2B,$7C,$13,$D2,$FD,$04,$61,$B3,$FD,$DA,$74
	.dc.b	$F1,$31,$1F,$E5,$FF,$6F,$AE,$5F,$E5,$3C,$B1,$F3,$B9,$65,$FF,$80
	.dc.b	$2B,$79,$FF,$94,$95,$6A,$FF,$2D,$2F,$DD,$BF,$2D,$7F,$EC,$CF,$86
	.dc.b	$FF,$2E,$C6,$2B,$FC,$E7,$18,$EC,$F6,$B6,$BD,$4F,$F9,$FB,$F5,$37
	.dc.b	$CE,$7C,$2E,$CC,$36,$BF,$D9,$F3,$A7,$E1,$4E,$9A,$6A,$71,$87,$F9
	.dc.b	$CF,$E7,$2E,$7F,$E9,$DC,$39,$4F,$EC,$7A,$51,$89,$F8,$55,$C3,$0F
	.dc.b	$E7,$53,$FE,$58,$9F,$F3,$BF,$0B,$BB,$D5,$F0,$7F,$69,$B7,$F5,$DB
	.dc.b	$97,$FB,$86,$B8,$4F,$E7,$E3,$FD,$99,$F5,$BF,$EC,$DF,$AD,$FE,$45
	.dc.b	$1F,$EA,$AF,$3F,$FD,$BF,$18,$8F,$F5,$6C,$EF,$C6,$FF,$57,$C7,$FB
	.dc.b	$06,$F0,$E2,$6B,$FD,$72,$2B,$E0,$15,$D5,$E4,$62,$BF,$DC,$35,$C2
	.dc.b	$FC,$71,$89,$D1,$E7,$BE,$AA,$B8,$46,$1B,$1C,$0F,$59,$4F,$0D,$F8
	.dc.b	$DE,$46,$6F,$C2,$FB,$47,$BD,$E9,$85,$FF,$9D,$F0,$F3,$4E,$63,$15
	.dc.b	$C6,$F0,$F8,$4E,$EE,$0F,$16,$1E,$86,$C7,$8E,$7D,$8D,$39,$4E,$EF
	.dc.b	$E3,$57,$36,$9F,$63,$0F,$5D,$7C,$29,$D3,$E5,$9D,$5F,$69,$CC,$EE
	.dc.b	$72,$F8,$07,$C7,$FC,$47,$E3,$35,$F9,$ED,$7F,$04,$E7,$E3,$DF,$5B
	.dc.b	$FE,$5C,$9E,$79,$FF,$BF,$27,$3B,$A7,$25,$8C,$B8,$0E,$2F,$B5,$F5
	.dc.b	$C7,$F3,$93,$FE,$DA,$AF,$CB,$B5,$FC,$33,$D2,$37,$3C,$71,$F2,$7C
	.dc.b	$57,$8E,$F6,$DA,$DE,$3D,$2F,$85,$79,$EB,$57,$CA,$35,$51,$E6,$79
	.dc.b	$27,$6F,$BE,$B9,$BE,$E6,$E1,$E9,$C3,$DB,$5C,$2B,$57,$D6,$73,$1D
	.dc.b	$4C,$61,$F3,$F0,$FA,$59,$6A,$7F,$3F,$3E,$76,$9E,$3B,$FC,$EB,$D8
	.dc.b	$F4,$92,$FE,$1B,$F3,$2F,$CF,$87,$97,$1F,$46,$1E,$1C,$3C,$77,$F1
	.dc.b	$CF,$31,$C5,$CB,$F7,$AB,$91,$CE,$FE,$19,$DA,$38,$E1,$E2,$3F,$8F
	.dc.b	$F1,$FF,$32,$FB,$AF,$B5,$87,$F2,$ED,$69,$FE,$8C,$E0,$FF,$42,$C7
	.dc.b	$F3,$B9,$CE,$1F,$E0,$EE,$34,$63,$D9,$5F,$0C,$39,$1C,$31,$F0,$61
	.dc.b	$FD,$3B,$83,$18,$FE,$AA,$30,$7F,$67,$F7,$F3,$32,$32,$7F,$8E,$F1
	.dc.b	$76,$AB,$53,$98,$7D,$3F,$95,$B1,$CB,$5F,$EA,$D8,$66,$A4,$E7,$3F
	.dc.b	$F4,$4F,$6B,$D6,$E5,$E6,$4A,$FF,$19,$E3,$FD,$DE,$7E,$2B,$63,$FC
	.dc.b	$01,$3B,$7F,$BE,$D7,$37,$FC,$09,$E8,$F9,$48,$FF,$81,$F9,$6B,$EB
	.dc.b	$8B,$7F,$E0,$99,$EF,$5F,$E8,$0A,$FF,$07,$7B,$1E,$F3,$3E,$5C,$8C
	.dc.b	$90,$F1,$BE,$8F,$DD,$5D,$FF,$6F,$C8,$FE,$C5,$97,$F8,$AF,$2D,$4C
	.dc.b	$C3,$FC,$B3,$1B,$30,$FF,$92,$BE,$BF,$EA,$3F,$E5,$8E,$47,$E5,$8D
	.dc.b	$BF,$10,$D9,$FF,$2A,$57,$07,$97,$2E,$AA,$46,$A3,$F2,$6B,$EB,$5F
	.dc.b	$47,$8F,$A8,$E9,$3F,$E5,$C6,$EB,$FE,$E0,$AD,$CF,$F9,$8D,$AE,$F0
	.dc.b	$E4,$76,$8C,$CF,$1B,$CD,$7F,$9A,$EF,$AF,$2F,$F2,$5B,$DB,$FE,$D2
	.dc.b	$E0,$7C,$97,$E9,$FF,$71,$7B,$74,$FA,$51,$31,$F9,$13,$A6,$CE,$5F
	.dc.b	$D1,$43,$E0,$3A,$74,$C9,$E3,$1D,$CC,$3C,$23,$0F,$C1,$F8,$CF,$91
	.dc.b	$F1,$72,$30,$43,$E5,$39,$7F,$D0,$FA,$27,$E3,$19,$DF,$FC,$0B,$3F
	.dc.b	$D6,$76,$5E,$A3,$FD,$DF,$E2,$FF,$09,$F7,$FF,$B1,$C7,$FB,$CB,$C6
	.dc.b	$C4,$C8,$C1,$0F,$E0,$53,$FD,$67,$0E,$E7,$F9,$A7,$DF,$FE,$29,$F1
	.dc.b	$0F,$B5,$A7,$EA,$22,$5F,$E5,$67,$4D,$9C,$BD,$BE,$AB,$CD,$E8,$C8
	.dc.b	$FC,$E6,$29,$F5,$19,$8F,$2C,$E1,$DC,$67,$1F,$8B,$D8,$BF,$F3,$AD
	.dc.b	$68,$EB,$E5,$7C,$5F,$46,$46,$08,$7C,$27,$2F,$C4,$8F,$99,$E3,$FC
	.dc.b	$DF,$3A,$4F,$A2,$32,$FA,$A3,$2F,$CC,$8F,$D0,$7C,$57,$F4,$0F,$41
	.dc.b	$F0,$BF,$CF,$69,$18,$EE,$DC,$7E,$AE,$B8,$56,$EB,$53,$92,$1F,$EF
	.dc.b	$57,$2D,$8E,$CF,$EF,$67,$F5,$1C,$52,$3F,$4A,$FF,$51,$F8,$9F,$D7
	.dc.b	$C7,$F8,$9F,$F0,$AF,$F7,$83,$FA,$AE,$6F,$D8,$1F,$C2,$23,$31,$FD
	.dc.b	$2E,$5F,$29,$CC,$91,$C4,$FE,$35,$0F,$B2,$E5,$E6,$49,$CC,$57,$A7
	.dc.b	$D9,$79,$4C,$72,$FE,$02,$D1,$99,$EA,$EB,$7F,$F1,$77,$DD,$FF,$01
	.dc.b	$9F,$F7,$AB,$6F,$F5,$B0,$8C,$4F,$F8,$D7,$EF,$FF,$D5,$13,$FE,$FF
	.dc.b	$9C,$BF,$E4,$B8,$79,$DF,$CA,$9B,$1B,$F1,$7F,$96,$D7,$08,$4E,$96
	.dc.b	$1E,$28,$C3,$E4,$9E,$1D,$7F,$C1,$6B,$FC,$B5,$3E,$88,$C4,$72,$C7
	.dc.b	$FC,$99,$18,$9F,$6F,$E1,$5F,$E0,$5A,$D1,$FE,$63,$9D,$57,$F9,$4E
	.dc.b	$5F,$19,$FD,$11,$B5,$E0,$9E,$BE,$1F,$F4,$8B,$E3,$3F,$E6,$DA,$DB
	.dc.b	$D8,$38,$C3,$D1,$5F,$D4,$3F,$0A,$7B,$F4,$97,$FC,$E3,$0A,$FF,$36
	.dc.b	$F0,$3F,$A0,$18,$C3,$EC,$DF,$29,$FE,$7F,$3A,$7D,$07,$FA,$06,$34
	.dc.b	$F9,$57,$BF,$E3,$24,$70,$F9,$27,$D7,$BE,$67,$33,$F2,$A7,$3F,$13
	.dc.b	$FE,$3C,$6C,$78,$06,$1D,$9C,$35,$E5,$6A,$E4,$D8,$DE,$1F,$DD,$BE
	.dc.b	$B1,$DD,$FE,$EA,$9F,$F4,$54,$61,$C0,$70,$ED,$AF,$FB,$C9,$CE,$9F
	.dc.b	$A2,$FA,$DF,$F2,$9C,$FA,$5A,$FD,$B9,$DF,$17,$82,$FF,$C1,$35,$AB
	.dc.b	$E5,$E2,$79,$E7,$FC,$1B,$F8,$23,$ED,$3F,$C1,$FE,$6F,$F7,$EC,$FF
	.dc.b	$D6,$8D,$53,$B7,$E5,$35,$FC,$64,$6C,$E1,$19,$C3,$F8,$27,$D9,$7F
	.dc.b	$87,$3E,$8F,$FB,$1D,$FE,$4D,$B3,$A8,$FE,$27,$3B,$F1,$24,$60,$FF
	.dc.b	$2A,$FE,$EA,$FA,$AF,$F2,$AB,$63,$31,$FE,$65,$F3,$F2,$C8,$C9,$FE
	.dc.b	$39,$F8,$4F,$EC,$AF,$EC,$BE,$1B,$FC,$FF,$E3,$FD,$86,$C7,$DD,$6B
	.dc.b	$FE,$C7,$0E,$1F,$F6,$53,$AC,$FF,$B1,$D7,$C3,$E3,$FF,$70,$7A,$3F
	.dc.b	$E7,$EE,$1F,$F4,$57,$C3,$B2,$A7,$F9,$F6,$26,$5F,$A8,$CB,$A6,$1F
	.dc.b	$EE,$95,$F5,$FF,$CA,$27,$C2,$D7,$4D,$7C,$ED,$8E,$33,$FE,$8B,$C2
	.dc.b	$57,$7A,$FE,$AF,$F5,$4F,$91,$FE,$31,$3F,$EE,$78,$C7,$A5,$F9,$E6
	.dc.b	$C7,$E8,$DA,$FC,$7F,$8C,$7A,$2F,$AB,$FD,$2C,$61,$B1,$99,$FD,$64
	.dc.b	$E5,$E9,$E3,$7F,$76,$6F,$A9,$E1,$8C,$47,$D5,$C3,$F8,$DC,$FD,$9F
	.dc.b	$AB,$FF,$3B,$6D,$F2,$DB,$5F,$EA,$DF,$B5,$FF,$17,$CF,$CD,$9C,$F2
	.dc.b	$FF,$90,$38,$51,$89,$D7,$C2,$70,$C1,$CB,$FE,$BA,$C7,$18,$70,$AB
	.dc.b	$F5,$9E,$0B,$FC,$3B,$5F,$B8,$D7,$F1,$AD,$93,$E1,$8C,$CF,$86,$DF
	.dc.b	$50,$F8,$CE,$F3,$32,$FC,$56,$1C,$6F,$75,$9C,$3F,$A9,$E8,$FA,$C7
	.dc.b	$D2,$7B,$63,$0F,$7E,$B7,$CE,$E6,$B7,$9F,$FB,$29,$B8,$74,$6E,$1F
	.dc.b	$A4,$AE,$18,$7A,$7E,$07,$FD,$17,$87,$F3,$7F,$0D,$F0,$59,$F8,$8F
	.dc.b	$9D,$9A,$73,$5F,$E5,$F9,$CC,$62,$7F,$DD,$97,$D1,$FE,$21,$7C,$BE
	.dc.b	$B3,$FC,$0F,$87,$F3,$66,$CF,$F3,$B6,$BF,$BF,$6A,$23,$C3,$3F,$F0
	.dc.b	$CB,$67,$F6,$DE,$C3,$D5,$F0,$7B,$7E,$AF,$AB,$FE,$01,$5F,$EF,$0F
	.dc.b	$AC,$FF,$1B,$BF,$D5,$E3,$39,$7D,$49,$CB,$FD,$EE,$32,$FF,$13,$6B
	.dc.b	$E5,$0E,$1F,$43,$5C,$F8,$DE,$33,$F2,$F8,$BC,$FC,$61,$FD,$A7,$E1
	.dc.b	$3F,$B2,$3F,$DE,$BE,$B3,$F1,$FE,$CF,$D4,$9C,$E3,$FE,$1C,$9C,$D7
	.dc.b	$1F,$5F,$F8,$95,$6A,$BF,$95,$35,$F4,$35,$F9,$CF,$4E,$0F,$F2,$2B
	.dc.b	$6B,$F9,$5F,$13,$E7,$BC,$58,$78,$FC,$9F,$D2,$7C,$17,$FA,$EE,$FD
	.dc.b	$51,$8F,$77,$51,$FD,$DB,$1F,$E3,$CD,$9F,$39,$1E,$27,$F9,$A3,$EB
	.dc.b	$38,$E9,$FE,$2D,$FA,$CF,$D1,$CF,$F8,$78,$7D,$67,$FB,$19,$FF,$09
	.dc.b	$46,$A7,$FC,$83,$1A,$F6,$3F,$3B,$39,$9F,$E9,$CD,$7E,$E3,$5F,$C9
	.dc.b	$35,$3D,$77,$FF,$35,$B6,$BF,$30,$79,$3E,$4F,$BF,$7F,$EB,$23,$F9
	.dc.b	$A9,$E9,$8F,$F7,$DB,$FE,$29,$33,$7E,$88,$CB,$FE,$4D,$8C,$BF,$D9
	.dc.b	$5A,$FE,$4D,$AF,$8B,$CE,$FE,$A9,$1D,$30,$FF,$40,$B4,$F3,$E4,$FA
	.dc.b	$CF,$35,$7F,$07,$F3,$25,$7F,$8F,$7E,$07,$FC,$66,$FF,$B8,$63,$F8
	.dc.b	$3C,$63,$1F,$F1,$B3,$5F,$EF,$4D,$7F,$50,$D7,$F9,$DF,$2F,$C8,$7F
	.dc.b	$CD,$51,$CF,$2F,$F6,$FC,$E5,$FF,$4C,$C6,$6F,$FB,$FF,$C6,$FC,$69
	.dc.b	$FF,$72,$B6,$3F,$C0,$7C,$9B,$C3,$FA,$4C,$E5,$FE,$1D,$7F,$CA,$9F
	.dc.b	$3F,$89,$FD,$EE,$31,$1E,$88,$FF,$00,$79,$0B,$FE,$5F,$87,$69,$FE
	.dc.b	$8D,$83,$E8,$3F,$4B,$1F,$CB,$3C,$DF,$E1,$F9,$AF,$F8,$F6,$AF,$AC
	.dc.b	$F7,$47,$F8,$7B,$C0,$FE,$C1,$39,$D7,$94,$7F,$83,$49,$3F,$E2,$B6
	.dc.b	$C7,$FA,$4F,$0E,$F3,$F1,$BD,$6E,$2F,$FA,$1B,$66,$AF,$93,$F4,$FC
	.dc.b	$9F,$A9,$B5,$B9,$D3,$67,$4D,$35,$F5,$9F,$0B,$E4,$7E,$1B,$63,$27
	.dc.b	$38,$CB,$4E,$DF,$59,$F2,$BD,$1F,$96,$D8,$C9,$E3,$8C,$B4,$CE,$BF
	.dc.b	$EC,$0C,$FD,$B3,$A8,$C6,$5F,$DE,$6B,$FD,$4F,$C2,$F8,$E7,$CD,$FF
	.dc.b	$67,$F5,$7F,$BA,$46,$2B,$E4,$74,$FF,$81,$6F,$85,$E8,$FC,$0F,$D9
	.dc.b	$DA,$F9,$1D,$7F,$84,$F8,$FF,$C2,$B6,$32,$7C,$D8,$CB,$4F,$F5,$67
	.dc.b	$2F,$F0,$DD,$5C,$EF,$AD,$FC,$75,$6B,$C1,$FC,$3C,$62,$BF,$05,$1F
	.dc.b	$E5,$32,$4F,$F9,$C5,$B1,$A8,$FF,$64,$D7,$E0,$FD,$0F,$CC,$57,$F3
	.dc.b	$EF,$A5,$FE,$7F,$F1,$B6,$FF,$15,$E0,$7F,$B3,$1B,$3A,$69,$A9,$FF
	.dc.b	$0F,$7D,$6B,$EA,$BF,$C5,$98,$BE,$8F,$31,$86,$DF,$D6,$AE,$0F,$E6
	.dc.b	$6B,$F3,$BA,$38,$D5,$A8,$FD,$17,$D6,$FF,$8B,$7E,$B7,$E5,$47,$FA
	.dc.b	$32,$31,$B7,$ED,$5F,$F3,$C1,$31,$FF,$4B,$36,$79,$C7,$C6,$E0,$77
	.dc.b	$F1,$8F,$F7,$17,$5F,$87,$3F,$DF,$5F,$F9,$63,$EB,$7F,$B6,$3F,$E1
	.dc.b	$D9,$D5,$7F,$A6,$27,$33,$FE,$4D,$DF,$FC,$67,$5A,$8E,$5E,$76,$27
	.dc.b	$E5,$FD,$6F,$CA,$FA,$DB,$8F,$C5,$6D,$FF,$94,$38,$15,$F4,$3E,$C3
	.dc.b	$3A,$79,$8F,$D1,$7A,$BF,$E9,$D6,$E3,$FE,$F1,$6E,$1F,$EB,$56,$BF
	.dc.b	$E2,$37,$27,$3E,$FF,$F0,$44,$FF,$B0,$F2,$FD,$5D,$FF,$97,$59,$1F
	.dc.b	$EC,$7F,$AE,$B8,$7E,$E7,$E0,$D5,$8C,$7F,$D7,$F5,$6F,$FC,$8D,$0F
	.dc.b	$AA,$FD,$06,$1E,$36,$BA,$6B,$F8,$76,$C7,$1F,$3B,$FC,$E5,$1F,$A9
	.dc.b	$8F,$F6,$6E,$98,$9F,$F4,$7F,$D6,$7F,$93,$F0,$FF,$6C,$37,$0F,$F6
	.dc.b	$C7,$7A,$31,$1F,$ED,$86,$BF,$EC,$D7,$86,$1F,$EA,$16,$D7,$FB,$6B
	.dc.b	$47,$D5,$F8,$3E,$06,$4F,$FB,$5E,$B7,$87,$FA,$92,$F7,$39,$AD,$CF
	.dc.b	$FB,$C6,$B7,$E4,$AF,$73,$9A,$F6,$FD,$27,$FB,$77,$3F,$53,$70,$34
	.dc.b	$98,$F8,$DF,$F4,$C7,$5F,$28,$FA,$A7,$FB,$EF,$07,$D7,$7F,$9F,$7E
	.dc.b	$BB,$F9,$7C,$7F,$9A,$FC,$5F,$D3,$C6,$23,$FC,$DD,$3B,$F2,$7F,$CE
	.dc.b	$11,$FC,$8A,$F0,$E2,$7F,$CB,$73,$AB,$FD,$8F,$B9,$CE,$BF,$D1,$95
	.dc.b	$A7,$C1,$75,$EE,$30,$E1,$7E,$4C,$3D,$57,$E0,$AF,$EF,$37,$5E,$A7
	.dc.b	$19,$F1,$46,$63,$E0,$99,$3C,$25,$3C,$38,$78,$23,$C9,$1D,$23,$09
	.dc.b	$F6,$3D,$30,$C9,$8B,$F2,$61,$E9,$9F,$8C,$77,$3E,$51,$4E,$58,$7A
	.dc.b	$1C,$4F,$2C,$3D,$4D,$8C,$9E,$39,$E1,$87,$86,$30,$7B,$9C,$61,$E6
	.dc.b	$9F,$97,$3E,$3C,$3C,$CF,$C2,$8E,$66,$AB,$E0,$9D,$4A,$E4,$DC,$7B
	.dc.b	$BE,$4B,$E3,$36,$7B,$C6,$23,$EF,$B7,$E4,$5F,$31,$FD,$2E,$53,$98
	.dc.b	$C3,$79,$3F,$DA,$CD,$9D,$DF,$FB,$16,$F9,$D7,$E3,$5F,$F7,$7C,$DF
	.dc.b	$89,$FF,$59,$E3,$BC,$7C,$18,$7E,$86,$F8,$DF,$CA,$0F,$27,$0D,$8E
	.dc.b	$13,$CA,$BA,$61,$F9,$7A,$E1,$9F,$5C,$7F,$DB,$4F,$09,$D5,$F1,$C7
	.dc.b	$FD,$E5,$7D,$70,$F6,$61,$E8,$9C,$D6,$DF,$2B,$F2,$70,$ED,$1E,$19
	.dc.b	$EF,$5E,$48,$FD,$05,$FB,$AB,$71,$8A,$DB,$DB,$0F,$29,$CD,$87,$2C
	.dc.b	$3C,$18,$75,$C7,$99,$1C,$1E,$1C,$3B,$91,$C1,$2B,$E0,$E1,$E2,$7E
	.dc.b	$8B,$71,$F4,$4F,$49,$D9,$F7,$4F,$C0,$61,$F1,$67,$71,$A3,$DC,$7B
	.dc.b	$0F,$E0,$9C,$3F,$D2,$7E,$17,$FB,$66,$FF,$D1,$DB,$29,$D3,$FE,$7B
	.dc.b	$E4,$7C,$06,$A7,$33,$FD,$53,$91,$7F,$CC,$E3,$94,$61,$FE,$B7,$19
	.dc.b	$9F,$EB,$5C,$1C,$EB,$F4,$75,$FD,$46,$3C,$7C,$DF,$D2,$FD,$FD,$44
	.dc.b	$9D,$1E,$58,$C3,$FD,$4E,$2F,$17,$F8,$9A,$B9,$4F,$03,$D1,$3E,$3B
	.dc.b	$F0,$C6,$0F,$F0,$A7,$5F,$F8,$C9,$FF,$3D,$B5,$FF,$22,$6B,$FE,$01
	.dc.b	$F1,$BC,$FB,$7F,$B7,$E7,$71,$F0,$F2,$FF,$05,$36,$F9,$4F,$F8,$2E
	.dc.b	$3E,$46,$5F,$DA,$7D,$3E,$64,$9F,$A5,$C5,$5F,$F0,$D4,$5C,$3D,$D8
	.dc.b	$FF,$5D,$7F,$C6,$18,$BC,$6F,$BB,$1F,$E1,$B6,$C6,$6F,$EF,$5E,$27
	.dc.b	$CC,$48,$F8,$6F,$AF,$EA,$BF,$C1,$31,$FE,$9E,$6B,$F1,$0F,$F3,$43
	.dc.b	$8B,$FA,$5E,$67,$C5,$9D,$42,$75,$AA,$5A,$2F,$D9,$39,$8C,$69,$EC
	.dc.b	$BE,$37,$C7,$EB,$1C,$3F,$9C,$84,$61,$FF,$04,$9B,$8F,$C7,$FD,$6F
	.dc.b	$68,$C1,$FE,$DB,$5A,$E8,$E9,$5C,$E7,$E0,$0F,$AB,$7D,$6F,$85,$B1
	.dc.b	$E0,$35,$1E,$5F,$AC,$7C,$8E,$E3,$F5,$53,$A3,$F5,$95,$C9,$FF,$10
	.dc.b	$35,$FF,$0C,$B9,$E0,$FF,$42,$C2,$FD,$2F,$85,$AF,$A2,$78,$6F,$FE
	.dc.b	$60,$D7,$C7,$3F,$77,$1F,$9C,$7E,$D3,$83,$FC,$ED,$5B,$8F,$DB,$FD
	.dc.b	$6F,$FB,$7E,$BE,$6F,$0A,$31,$1F,$80,$C3,$AB,$FE,$EB,$87,$C0,$72
	.dc.b	$E6,$93,$CA,$72,$D9,$CB,$C6,$7C,$31,$F3,$FE,$2A,$4F,$CF,$D1,$1F
	.dc.b	$3F,$E1,$F3,$3F,$2A,$3E,$7C,$EA,$3E,$7C,$60,$FA,$DD,$B2,$3A,$47
	.dc.b	$F1,$6F,$7F,$FB,$47,$0F,$F0,$F7,$FB,$FC,$FC,$26,$BF,$A5,$D5,$1C
	.dc.b	$37,$FC,$E3,$67,$F4,$2F,$0A,$E9,$E0,$77,$F0,$7F,$73,$EE,$CB,$FE
	.dc.b	$2A,$F8,$FE,$A7,$11,$DF,$E0,$1F,$3F,$FC,$B1,$B1,$C0,$F3,$7D,$47
	.dc.b	$AF,$CD,$91,$82,$1E,$9D,$11,$A8,$C4,$FC,$33,$1A,$7C,$39,$C4,$FC
	.dc.b	$3F,$1B,$E4,$79,$B2,$30,$43,$B3,$97,$C1,$1F,$3F,$C6,$FA,$1E,$54
	.dc.b	$9F,$9D,$19,$7E,$84,$65,$F4,$47,$EC,$FE,$92,$47,$E1,$5F,$E9,$FF
	.dc.b	$02,$F6,$73,$3F,$AE,$9F,$C2,$91,$FD,$9B,$FD,$A3,$CE,$7C,$2F,$EC
	.dc.b	$35,$FF,$7A,$C6,$39,$B9,$C7,$F5,$3E,$E7,$2F,$EA,$09,$8F,$EA,$33
	.dc.b	$FD,$9B,$FC,$8F,$EB,$5F,$03,$FC,$11,$B1,$FD,$BD,$AF,$FA,$8D,$AF
	.dc.b	$F2,$B3,$F9,$5C,$BE,$B7,$DE,$48,$FD,$77,$57,$C2,$FA,$DF,$E5,$16
	.dc.b	$F8,$BA,$66,$BF,$8D,$46,$5F,$14,$79,$48,$FF,$AD,$7B,$58,$BE,$B7
	.dc.b	$E7,$92,$79,$D7,$F3,$7F,$96,$FE,$0F,$BF,$F1,$A6,$F0,$7A,$DB,$5F
	.dc.b	$D1,$FB,$9F,$11,$C7,$E0,$3F,$94,$D7,$F1,$0A,$FF,$5E,$B5,$E5,$1A
	.dc.b	$3F,$A3,$57,$F4,$F8,$FE,$6C,$D7,$F9,$EC,$F8,$1F,$D0,$F8,$5F,$D3
	.dc.b	$7B,$7F,$84,$47,$F4,$26,$E5,$FC,$CB,$EB,$7F,$C1,$AF,$F3,$48,$DC
	.dc.b	$FF,$B4,$1A,$FD,$76,$BF,$D8,$DA,$F8,$0F,$E9,$52,$F8,$AF,$EB,$CD
	.dc.b	$8F,$E9,$D3,$E2,$D3,$F0,$5F,$5A,$FA,$3F,$EC,$8D,$AF,$EA,$8D,$7F
	.dc.b	$9A,$56,$A7,$FA,$9C,$E6,$7F,$DC,$ED,$7A,$C6,$8F,$E8,$7E,$14,$8E
	.dc.b	$B1,$E0,$F8,$C9,$1E,$07,$FC,$A3,$F2,$6F,$AD,$7C,$D6,$AB,$FD,$FD
	.dc.b	$8A,$FC,$64,$8F,$F2,$84,$65,$FE,$FB,$7C,$FE,$67,$AE,$07,$A6,$FF
	.dc.b	$D9,$BE,$B5,$E1,$FF,$81,$DB,$87,$F7,$76,$BF,$DF,$BE,$B7,$ED,$F0
	.dc.b	$7E,$B6,$73,$3F,$AF,$0E,$1D,$57,$F7,$C8,$C4,$7F,$87,$DE,$B7,$FD
	.dc.b	$E2,$FF,$CD,$13,$DF,$C8,$F2,$5F,$F8,$6F,$EB,$5F,$55,$49,$5F,$E0
	.dc.b	$DF,$43,$FB,$6E,$3F,$E0,$2A,$DC,$FF,$9D,$5A,$F9,$C3,$87,$D4,$D7
	.dc.b	$FC,$61,$DC,$FA,$DF,$F1,$67,$D1,$FF,$55,$DB,$F8,$73,$6B,$F8,$7C
	.dc.b	$FF,$2C,$37,$E2,$48,$FE,$3D,$1F,$E5,$4F,$AD,$7D,$6D,$24,$FF,$A4
	.dc.b	$7D,$3A,$FF,$92,$27,$6F,$F8,$EB,$44,$62,$3F,$B7,$7C,$2F,$F5,$E7
	.dc.b	$FC,$A9,$F0,$EF,$AD,$FE,$31,$18,$6D,$B5,$FF,$29,$56,$FC,$6F,$F4
	.dc.b	$9F,$E9,$3F,$CD,$23,$18,$7F,$92,$9A,$F3,$9D,$FC,$0F,$FA,$9F,$1E
	.dc.b	$F3,$FC,$CE,$B6,$D8,$CB,$E0,$6B,$FE,$B3,$F9,$EF,$E9,$98,$F9,$1B
	.dc.b	$19,$8F,$F4,$8F,$C0,$7F,$AD,$67,$36,$69,$C3,$CC,$6B,$F8,$6B,$FE
	.dc.b	$8D,$F5,$AF,$98,$FF,$51,$E5,$86,$BF,$E7,$58,$D7,$B1,$FE,$BF,$6C
	.dc.b	$7E,$2F,$E3,$7F,$AE,$FD,$77,$FB,$0D,$B1,$9F,$59,$39,$6C,$7E,$4E
	.dc.b	$BF,$8B,$7B,$1D,$7C,$93,$49,$3F,$AF,$DD,$F5,$DF,$E7,$6D,$7E,$A3
	.dc.b	$67,$F0,$7F,$5D,$F7,$5F,$C1,$C6,$79,$F1,$7C,$BE,$2B,$51,$89,$D6
	.dc.b	$1F,$E9,$1E,$16,$3A,$8F,$C3,$CE,$B1,$FF,$75,$7D,$6D,$C8,$E7,$8D
	.dc.b	$F5,$DF,$EB,$5C,$3F,$15,$3B,$26,$9D,$70,$FF,$51,$D6,$DF,$CD,$CE
	.dc.b	$AB,$8D,$6E,$FA,$FC,$87,$1C,$7C,$33,$CF,$57,$D7,$3E,$9F,$0D,$C7
	.dc.b	$FE,$00,$7D,$16,$26,$47,$EB,$E7,$5A,$7E,$BA,$B2,$FF,$93,$FA,$BD
	.dc.b	$0B,$C4,$7C,$47,$97,$E2,$3F,$D3,$F8,$FF,$A8,$E7,$2F,$FA,$22,$33
	.dc.b	$5F,$EA,$C9,$CB,$D6,$78,$57,$6F,$AE,$FF,$63,$47,$FA,$B5,$B3,$FC
	.dc.b	$03,$DB,$FF,$2B,$CE,$7E,$3F,$81,$1F,$DD,$46,$A7,$FD,$77,$5C,$B6
	.dc.b	$FF,$55,$FD,$6F,$F8,$8E,$FF,$94,$52,$30,$FF,$2B,$6B,$FD,$0D,$AF
	.dc.b	$F8,$CD,$DC,$F4,$F1,$3E,$B6,$0F,$43,$E4,$D6,$B0,$E1,$58,$37,$36
	.dc.b	$3F,$D9,$9F,$13,$FD,$75,$FF,$6E,$FD,$6F,$E3,$BD,$AF,$A7,$39,$CF
	.dc.b	$FD,$ED,$39,$AE,$3E,$B6,$8F,$AD,$F4,$36,$32,$4A,$FF,$46,$46,$5F
	.dc.b	$F7,$0D,$FF,$09,$9F,$F2,$D7,$C0,$78,$3C,$8F,$CA,$7A,$5F,$EB,$29
	.dc.b	$F3,$C6,$5F,$F3,$07,$D6,$DC,$39,$CD,$E1,$FD,$C7,$77,$F9,$2A,$B1
	.dc.b	$7F,$E5,$DF,$AD,$75,$7E,$9E,$1F,$5A,$F5,$FF,$27,$5A,$9F,$F8,$02
	.dc.b	$6A,$33,$52,$33,$4F,$99,$C6,$9F,$EF,$EC,$39,$36,$FC,$15,$BB,$FF
	.dc.b	$7F,$56,$B3,$FF,$79,$C7,$FA,$00,$F0,$E0,$F8,$6F,$D3,$C6,$71,$57
	.dc.b	$F1,$78,$F1,$4E,$BC,$C9,$1E,$6F,$93,$B7,$C2,$3F,$D0,$07,$3B,$F0
	.dc.b	$D7,$9A,$B7,$1F,$33,$BF,$F9,$6D,$7E,$C7,$CB,$FD,$CF,$C0,$7F,$89
	.dc.b	$23,$03,$1E,$C5,$8D,$38,$0E,$2F,$B4,$74,$F7,$7F,$81,$36,$BF,$80
	.dc.b	$34,$FF,$67,$67,$FC,$6F,$E4,$BE,$34,$FF,$A9,$6B,$13,$E6,$F1,$A4
	.dc.b	$FF,$AA,$5B,$7F,$EE,$FE,$6F,$E2,$F8,$FE,$29,$B5,$FE,$F0,$FA,$1F
	.dc.b	$CA,$DF,$F1,$8F,$27,$F6,$F8,$7F,$80,$BC,$0F,$E9,$0F,$F4,$F6,$BF
	.dc.b	$DE,$DA,$EE,$FE,$36,$3C,$2B,$F7,$53,$FE,$FC,$F5,$7B,$5C,$BD,$9C
	.dc.b	$CB,$3B,$6A,$AE,$7D,$2D,$9F,$2B,$F1,$FE,$5B,$E4,$37,$0E,$75,$B8
	.dc.b	$C6,$1F,$EF,$B6,$E7,$FA,$EC,$1F,$59,$F4,$7E,$53,$E9,$37,$1E,$73
	.dc.b	$A6,$CE,$9A,$6B,$EB,$3E,$D7,$A1,$F6,$DB,$19,$39,$C6,$5A,$67,$2F
	.dc.b	$C7,$E7,$EF,$9D,$46,$30,$FE,$C9,$97,$F7,$8E,$37,$CE,$FF,$5F,$6E
	.dc.b	$3E,$8F,$AC,$C4,$C8,$FF,$25,$4E,$B0,$E9,$59,$7B,$75,$FE,$4F,$E3
	.dc.b	$FF,$28,$DB,$F8,$6D,$9D,$C7,$C6,$9F,$C6,$78,$1F,$86,$9F,$B4,$DE
	.dc.b	$3F,$E1,$71,$FE,$1B,$8C,$1C,$2B,$F0,$7D,$7F,$7B,$CE,$C5,$7E,$BE
	.dc.b	$FF,$87,$57,$F9,$36,$FF,$C1,$17,$B8,$FC,$4C,$EA,$9C,$1F,$EF,$4F
	.dc.b	$91,$FE,$6B,$F1,$BF,$CD,$9B,$7F,$88,$EC,$FF,$76,$B6,$74,$D3,$FC
	.dc.b	$E5,$F5,$2F,$99,$FF,$07,$36,$3F,$23,$AF,$E8,$B8,$D3,$9C,$FF,$85
	.dc.b	$C6,$2F,$89,$FE,$1B,$AE,$15,$B9,$FD,$67,$D2,$7C,$92,$61,$FD,$D3
	.dc.b	$54,$63,$1F,$8C,$FC,$4F,$AB,$76,$7E,$FE,$25,$FC,$67,$51,$8A,$FE
	.dc.b	$39,$D1,$E1,$AF,$E4,$50,$8C,$3F,$33,$EA,$FF,$B1,$3C,$35,$FF,$11
	.dc.b	$CE,$A3,$84,$7F,$8D,$E7,$79,$FF,$0E,$AE,$6F,$C6,$FA,$BF,$96,$7C
	.dc.b	$3F,$A9,$B8,$FD,$08,$B8,$FE,$9B,$75,$62,$FF,$49,$E8,$7E,$8C,$FF
	.dc.b	$23,$7C,$0B,$EA,$F1,$97,$E8,$A6,$E3,$FE,$AA,$9B,$7F,$DC,$AF,$85
	.dc.b	$FF,$0C,$6D,$7F,$AC,$3E,$AF,$AD,$CA,$D3,$E0,$8C,$4F,$E0,$3E,$9D
	.dc.b	$E4,$7F,$95,$5B,$7D,$73,$C4,$C8,$FF,$60,$4E,$AB,$FD,$7F,$59,$7F
	.dc.b	$D7,$FF,$51,$FE,$3B,$CF,$FC,$FA,$DB,$FF,$3D,$7D,$5F,$F9,$96,$31
	.dc.b	$5F,$CB,$5B,$CC,$73,$D7,$FE,$A8,$AE,$1F,$37,$30,$FF,$61,$56,$B7
	.dc.b	$56,$71,$F1,$DD,$CB,$E5,$4D,$C3,$FD,$9F,$83,$E9,$3C,$87,$D6,$E8
	.dc.b	$B3,$DA,$2E,$3F,$ED,$E9,$B8,$7F,$52,$FA,$BF,$EE,$DE,$CF,$F2,$2A
	.dc.b	$FC,$BF,$81,$FE,$A1,$8C,$47,$E6,$A7,$7E,$37,$E7,$23,$FC,$06,$27
	.dc.b	$6F,$F8,$06,$7F,$C0,$5F,$57,$8B,$FF,$56,$3A,$94,$EA,$FA,$56,$9E
	.dc.b	$D3,$A1,$5C,$F3,$E1,$1E,$3B,$AF,$F8,$46,$F9,$39,$3C,$D7,$DA,$BD
	.dc.b	$71,$9C,$3A,$3A,$38,$61,$83,$D1,$87,$68,$F8,$31,$87,$CB,$7D,$5A
	.dc.b	$72,$71,$1B,$3A,$3E,$E9,$E9,$87,$A2,$BE,$81,$C8,$F6,$61,$E1,$7E
	.dc.b	$25,$F2,$7B,$1F,$6A,$35,$5D,$1B,$1C,$CF,$4D,$F8,$5C,$5F,$85,$FB
	.dc.b	$46,$27,$8D,$F7,$8F,$2B,$5D,$B5,$FB,$1F,$1F,$F8,$57,$E9,$35,$F6
	.dc.b	$B5,$EE,$E7,$E3,$DF,$2F,$FE,$5E,$94,$E6,$30,$DD,$3F,$CA,$0D,$9D
	.dc.b	$E9,$FC,$CA,$F9,$D7,$E1,$9F,$F6,$7C,$DF,$85,$FE,$6F,$8F,$E8,$27
	.dc.b	$73,$F4,$5B,$19,$9D,$E9,$D5,$B5,$FE,$61,$AD,$CF,$87,$2F,$55,$F1
	.dc.b	$AF,$01,$E3,$8C,$CE,$F0,$CC,$F8,$70,$EA,$B0,$EA,$F7,$BE,$2F,$86
	.dc.b	$77,$3E,$E8,$F5,$4E,$9E,$4D,$8D,$61,$E5,$9D,$DF,$94,$F1,$C6,$4F
	.dc.b	$8D,$5C,$4E,$57,$E0,$3C,$2E,$67,$F5,$F5,$E9,$9F,$EC,$B8,$F7,$39
	.dc.b	$CE,$B0,$F6,$D6,$CF,$0E,$3F,$15,$F2,$9F,$14,$C9,$F0,$E7,$D0,$FE
	.dc.b	$EC,$E4,$59,$F8,$95,$E0,$8F,$E0,$1C,$BF,$CE,$FE,$0B,$FB,$CD,$FF
	.dc.b	$44,$6E,$BF,$CC,$9B,$5F,$E2,$6E,$5F,$F6,$53,$FC,$FA,$33,$7C,$E3
	.dc.b	$07,$F6,$3F,$73,$3C,$9F,$9C,$8F,$E8,$BC,$18,$3A,$F3,$B1,$FF,$67
	.dc.b	$4E,$88,$7F,$5A,$9D,$C6,$A3,$06,$B3,$FE,$BB,$3A,$F6,$BB,$4E,$74
	.dc.b	$FE,$9B,$39,$8F,$ED,$3E,$27,$BA,$DD,$7F,$E2,$49,$C9,$FD,$F7,$DA
	.dc.b	$D4,$66,$3F,$BC,$56,$0F,$2C,$F9,$BE,$26,$4F,$F8,$15,$F4,$F1,$E2
	.dc.b	$3F,$CF,$7A,$27,$18,$FF,$83,$6B,$BF,$89,$2F,$FB,$FB,$AD,$3F,$D5
	.dc.b	$B9,$FF,$8C,$67,$24,$3A,$E1,$ED,$FA,$8F,$EE,$71,$FE,$AC,$C5,$18
	.dc.b	$8F,$F2,$8B,$61,$7F,$ED,$FD,$DE,$23,$5E,$9D,$D3,$1E,$98,$F1,$C3
	.dc.b	$85,$EE,$7F,$D7,$3F,$55,$F6,$72,$FF,$65,$B6,$BF,$C9,$6D,$9F,$F3
	.dc.b	$AF,$D5,$FF,$76,$AD,$F1,$FE,$52,$55,$B6,$C7,$E3,$2B,$85,$FF,$95
	.dc.b	$BE,$A9,$F2,$DC,$5B,$7F,$EC,$C6,$BF,$EF,$9F,$AB,$7C,$76,$EA,$4E
	.dc.b	$DF,$14,$EA,$30,$7F,$7C,$15,$D1,$FF,$06,$35,$F2,$4E,$F3,$FA,$77
	.dc.b	$F5,$C9,$3F,$E8,$82,$9F,$59,$B1,$F5,$A3,$FD,$37,$A7,$F8,$66,$31
	.dc.b	$C0,$FA,$BF,$F7,$2D,$7F,$B8,$9F,$CD,$4E,$9B,$9F,$DF,$77,$3F,$ED
	.dc.b	$06,$CE,$F9,$FF,$73,$1C,$27,$FD,$56,$EA,$BE,$87,$C4,$7C,$F2,$4F
	.dc.b	$FB,$EB,$5F,$C2,$2C,$3D,$4F,$C8,$9D,$B9,$AE,$6F,$1E,$3C,$BF,$32
	.dc.b	$7E,$83,$C4,$FA,$5B,$23,$EA,$66,$F6,$BF,$A4,$63,$FE,$FF,$F0,$BF
	.dc.b	$89,$C7,$F7,$4C,$3F,$DF,$BE,$16,$0C,$CC,$8C,$9F,$E7,$2F,$63,$F8
	.dc.b	$29,$11,$FE,$F5,$FA,$6C,$3F,$E7,$F8,$B8,$7D,$6C,$15,$E2,$FA,$5F
	.dc.b	$1F,$97,$30,$E7,$7B,$9D,$46,$37,$FD,$B6,$0B,$E1,$3F,$24,$FF,$6F
	.dc.b	$C2,$B8,$C6,$74,$F8,$95,$C3,$0E,$D3,$D4,$E3,$E2,$7C,$6F,$2E,$46
	.dc.b	$6F,$E6,$4F,$CA,$33,$1D,$71,$FF,$3E,$3B,$51,$F4,$3C,$2F,$A3,$E3
	.dc.b	$4A,$D4,$65,$E1,$19,$7A,$47,$EE,$FE,$12,$47,$EE,$DF,$E9,$BE,$C7
	.dc.b	$FB,$CE,$30,$43,$E2,$1F,$B6,$8D,$FA,$BC,$A3,$FC,$EB,$CA,$9F,$C6
	.dc.b	$B6,$3F,$1B,$1F,$8C,$E8,$71,$FE,$90,$93,$F1,$70,$FE,$1B,$BF,$E7
	.dc.b	$BE,$A1,$F5,$7F,$C6,$DB,$5F,$3C,$38,$7F,$65,$7E,$CF,$AB,$91,$D5
	.dc.b	$FD,$9F,$3F,$C0,$FA,$BE,$9F,$0D,$2B,$C5,$15,$F8,$38,$27,$19,$7F
	.dc.b	$08,$CD,$7F,$2B,$EA,$E6,$1E,$B7,$7D,$3F,$9E,$87,$CA,$3E,$13,$F9
	.dc.b	$B3,$63,$D8,$D7,$E8,$F6,$3E,$07,$75,$27,$6F,$F3,$F9,$D4,$60,$F8
	.dc.b	$CF,$F5,$46,$F8,$7C,$5F,$BB,$DB,$FC,$05,$F9,$8D,$73,$E0,$7F,$4B
	.dc.b	$FA,$BF,$F0,$B4,$60,$DC,$7F,$4F,$9E,$1B,$FF,$16,$AD,$BF,$C6,$23
	.dc.b	$3A,$7C,$39,$7C,$27,$53,$95,$FF,$59,$FA,$B7,$CF,$52,$65,$FC,$D7
	.dc.b	$1A,$FE,$5A,$76,$FE,$15,$AF,$F3,$58,$DE,$5F,$CE,$FC,$3F,$E6,$A6
	.dc.b	$33,$FA,$D6,$F8,$5F,$11,$30,$FD,$F7,$C7,$3E,$AD,$F2,$F9,$9F,$15
	.dc.b	$5C,$BF,$A8,$5E,$EB,$18,$FF,$20,$8E,$5F,$2D,$FE,$4A,$6E,$1E,$9C
	.dc.b	$3F,$BD,$7D,$5F,$EA,$E3,$F6,$6D,$C7,$FA,$1B,$70,$FC,$77,$D5,$FD
	.dc.b	$07,$F2,$11,$9A,$FE,$5E,$D4,$D9,$E1,$1A,$F0,$BF,$A3,$CF,$86,$3F
	.dc.b	$E0,$03,$5E,$B1,$FD,$33,$EA,$9F,$57,$FE,$11,$6C,$FE,$DD,$AF,$F0
	.dc.b	$CF,$7B,$FC,$08,$FF,$0E,$8C,$E5,$FD,$79,$B9,$7E,$A5,$B3,$FD,$22
	.dc.b	$B8,$D7,$EC,$FE,$AF,$FC,$41,$F3,$FF,$D5,$6B,$F8,$CB,$67,$51,$FC
	.dc.b	$3E,$77,$E1,$C8,$C1,$FE,$41,$FA,$97,$D5,$FF,$98,$5B,$19,$8F,$F3
	.dc.b	$7F,$D5,$E5,$91,$93,$FA,$07,$C1,$27,$FD,$07,$F5,$5F,$C2,$E7,$FA
	.dc.b	$47,$C2,$A3,$11,$FD,$4E,$31,$E4,$7F,$A0,$BE,$AF,$F9,$2F,$0F,$F6
	.dc.b	$89,$CD,$72,$BE,$3B,$7F,$87,$E7,$33,$FD,$66,$F8,$A8,$CC,$EE,$B8
	.dc.b	$DF,$56,$D3,$D5,$61,$CB,$07,$CD,$FF,$85,$31,$F0,$37,$1F,$33,$70
	.dc.b	$FF,$53,$7B,$9F,$EA,$48,$C5,$F2,$AC,$1F,$E9,$A9,$FF,$2D,$7D,$53
	.dc.b	$E8,$BF,$88,$E1,$FC,$AE,$31,$EB,$7F,$92,$DB,$1F,$CB,$CF,$F0,$6B
	.dc.b	$87,$1F,$15,$FD,$4B,$EA,$FF,$CA,$4D,$8E,$07,$FB,$B7,$D4,$4E,$9B
	.dc.b	$1A,$AF,$E2,$BF,$1B,$E1,$F2,$0D,$24,$F8,$F5,$7D,$47,$FA,$6F,$6E
	.dc.b	$F1,$8F,$89,$FF,$1E,$D7,$DD,$AD,$3C,$5A,$FF,$4B,$9E,$15,$F8,$A9
	.dc.b	$DF,$0F,$F9,$73,$EA,$5F,$19,$FE,$61,$8F,$C2,$4E,$AB,$B7,$CB,$65
	.dc.b	$FF,$4D,$75,$23,$19,$7E,$1E,$B1,$5F,$9C,$9C,$E5,$EC,$7D,$35,$B6
	.dc.b	$C6,$67,$DF,$5C,$AB,$D9,$5C,$A3,$18,$7A,$7E,$77,$13,$E2,$87,$EE
	.dc.b	$CF,$0B,$0D,$C7,$FC,$5D,$AB,$E7,$B3,$52,$7F,$D7,$95,$BC,$FF,$C7
	.dc.b	$97,$B9,$CD,$7C,$0E,$4F,$F3,$87,$8B,$F0,$AD,$F8,$FB,$A3,$5A,$7F
	.dc.b	$20,$C7,$FC,$29,$39,$7F,$C8,$98,$72,$73,$5C,$30,$E9,$5B,$C3,$FC
	.dc.b	$5F,$F5,$3F,$E0,$FB,$FF,$69,$7D,$83,$E0,$3C,$53,$9A,$FF,$2C,$66
	.dc.b	$AD,$5F,$F8,$96,$B5,$5F,$A3,$6B,$FE,$13,$6B,$FE,$72,$6A,$FC,$09
	.dc.b	$C5,$F7,$BF,$03,$FF,$32,$5F,$F8,$C6,$72,$FF,$1A,$8C,$D7,$F9,$22
	.dc.b	$72,$FF,$1B,$8C,$E3,$FE,$9A,$9C,$BF,$C5,$E3,$35,$FE,$27,$9C,$B6
	.dc.b	$FF,$27,$E8,$3F,$03,$FE,$6B,$3E,$39,$FA,$DF,$11,$FC,$1E,$3F,$C9
	.dc.b	$9E,$CE,$3C,$2F,$CD,$F8,$CF,$83,$FF,$50,$57,$F9,$2E,$F8,$6D,$FE
	.dc.b	$7F,$9C,$BF,$E6,$28,$CE,$DF,$E4,$1C,$7A,$5F,$1C,$7A,$E0,$FA,$CC
	.dc.b	$E3,$C7,$C2,$CF,$D0,$F9,$9B,$87,$85,$B3,$B8,$FF,$32,$4E,$DD,$1F
	.dc.b	$E6,$7F,$7F,$13,$FE,$76,$8F,$F3,$69,$FE,$93,$1F,$29,$FE,$4F,$7F
	.dc.b	$CD,$06,$A3,$FD,$40,$EB,$E0,$BF,$3C,$E7,$5F,$F7,$43,$93,$8F,$91
	.dc.b	$FC,$11,$C5,$78,$63,$89,$9F,$AC,$7C,$8E,$2B,$FC,$F1,$CA,$F8,$7F
	.dc.b	$A7,$F8,$3F,$F9,$FA,$73,$18,$F8,$BF,$D8,$56,$E7,$FD,$55,$5B,$CB
	.dc.b	$FD,$2B,$39,$7F,$D6,$D1,$9C,$FF,$DC,$F3,$96,$E1,$E0,$F4,$FA,$BF
	.dc.b	$63,$DB,$98,$F9,$BE,$AE,$E5,$D9,$B8,$74,$6C,$FF,$B6,$BA,$7F,$19
	.dc.b	$1E,$83,$F0,$65,$8F,$F4,$F7,$AF,$F8,$63,$F0,$FF,$6D,$63,$E2,$B8
	.dc.b	$69,$FE,$A5,$7F,$1B,$F0,$32,$3F,$D5,$1F,$05,$72,$FF,$5B,$F9,$EF
	.dc.b	$A9,$FE,$73,$EA,$3E,$A1,$F4,$BF,$C6,$EB,$58,$FF,$B6,$6B,$53,$9B
	.dc.b	$E3,$7C,$2B,$FD,$61,$7C,$30,$FF,$74,$61,$C2,$73,$7F,$D0,$A7,$FC
	.dc.b	$1C,$DA,$E1,$1F,$D8,$E3,$FA,$3E,$DF,$B3,$ED,$FD,$4F,$81,$FA,$A6
	.dc.b	$E1,$FE,$C5,$6D,$FE,$D5,$BC,$3F,$EB,$3F,$8D,$FD,$8F,$C3,$7E,$C9
	.dc.b	$B7,$FE,$CD,$6D,$FE,$F5,$B4,$9F,$B5,$DE,$FB,$6F,$DD,$C7,$F0,$88
	.dc.b	$C3,$FE,$D7,$31,$1A,$7F,$DF,$D1,$AB,$FE,$15,$1A,$38,$1A,$9E,$11
	.dc.b	$FC,$EB,$EA,$9F,$2B,$DF,$B7,$FB,$04,$62,$FA,$C6,$27,$FB,$7E,$2A
	.dc.b	$D5,$FF,$88,$63,$13,$A9,$CC,$FF,$B4,$F6,$65,$FE,$51,$C7,$D3,$E8
	.dc.b	$7B,$FE,$07,$E0,$36,$32,$7B,$A3,$2F,$E1,$3E,$9B,$FB,$B7,$DA,$FE
	.dc.b	$17,$4F,$99,$97,$F9,$5E,$73,$C7,$F9,$16,$DF,$11,$3F,$D7,$2F,$94
	.dc.b	$7F,$87,$E7,$F1,$44,$7F,$C6,$F3,$FE,$5B,$8F,$C5,$E5,$C2,$3F,$CB
	.dc.b	$0B,$ED,$7E,$D6,$1F,$6D,$B7,$F6,$1B,$4C,$7D,$3A,$7C,$69,$FF,$01
	.dc.b	$7B,$DF,$3A,$73,$F5,$8E,$CC,$F0,$FF,$4E,$8C,$65,$EC,$AF,$87,$97
	.dc.b	$F8,$D2,$BF,$4D,$87,$D4,$8F,$82,$66,$BD,$B1,$98,$F8,$BF,$98,$3B
	.dc.b	$BE,$7F,$D5,$FF,$85,$63,$17,$DC,$F0,$70,$A7,$55,$C2,$75,$7D,$23
	.dc.b	$0F,$C1,$8F,$1D,$F4,$8F,$F3,$2B,$5F,$F3,$73,$5D,$CF,$E3,$FC,$CF
	.dc.b	$D4,$7D,$6F,$E6,$9B,$7E,$0D,$3F,$31,$1B,$8F,$F3,$8C,$61,$BC,$C9
	.dc.b	$FC,$D7,$87,$FB,$3F,$A7,$FD,$2B,$6B,$A5,$23,$71,$C9,$B9,$7E,$96
	.dc.b	$7F,$C1,$2F,$BC,$FF,$4F,$7C,$1F,$F9,$EA,$B9,$F8,$BF,$E0,$88,$C4
	.dc.b	$FE,$DE,$3F,$58,$49,$F3,$BF,$C7,$3D,$FF,$ED,$51,$FD,$15,$B3,$FE
	.dc.b	$AB,$5F,$2D,$FE,$14,$73,$CE,$69,$23,$5E,$7F,$F0,$4F,$07,$C9,$77
	.dc.b	$0F,$F6,$7B,$4F,$74,$7F,$01,$24,$7F,$A4,$3F,$10,$FF,$2C,$4F,$F9
	.dc.b	$8B,$D1,$FF,$3D,$C6,$38,$7F,$D1,$51,$8F,$03,$FC,$F7,$3C,$2B,$FD
	.dc.b	$24,$D7,$FD,$A8,$D7,$FD,$9C,$F3,$BC,$CF,$BF,$E8,$BA,$47,$FB,$8A
	.dc.b	$75,$5C,$2B,$53,$97,$CB,$C9,$F1,$FC,$AB,$3F,$E8,$06,$D7,$FA,$BE
	.dc.b	$D5,$E8,$25,$7F,$08,$9F,$E1,$FF,$5B,$FC,$43,$EB,$5F,$59,$FE,$C9
	.dc.b	$F3,$3F,$8C,$C6,$2B,$FD,$99,$5B,$AD,$5D,$9C,$8A,$E0,$FF,$1C,$BE
	.dc.b	$75,$FE,$CE,$BE,$13,$9B,$AF,$90,$7B,$5C,$9F,$C8,$B5,$4F,$F2,$88
	.dc.b	$D5,$FF,$30,$75,$1D,$23,$C3,$3E,$6C,$37,$7E,$08,$E6,$7A,$70,$F6
	.dc.b	$B4,$E0,$E2,$7A,$C6,$4F,$51,$4F,$53,$D9,$C1,$D9,$C1,$C8,$B1,$F1
	.dc.b	$63,$95,$7A,$1F,$E4,$B7,$EB,$BF,$43,$63,$AB,$6F,$A5,$F9,$CE,$8F
	.dc.b	$69,$F5,$9F,$CC,$2F,$B3,$E8,$9E,$0E,$8F,$6C,$F4,$78,$5F,$89,$FF
	.dc.b	$62,$61,$E9,$F8,$3F,$B3,$1F,$84,$6B,$EA,$36,$7F,$60,$F8,$C7,$C9
	.dc.b	$7F,$51,$8C,$88,$CF,$17,$F6,$6A,$E7,$18,$9D,$B6,$33,$5C,$F1,$FC
	.dc.b	$9C,$B8,$7F,$A7,$C6,$1B,$9F,$F5,$29,$E9,$3C,$2B,$94,$E6,$B7,$87
	.dc.b	$A5,$AF,$9E,$76,$FB,$5A,$FA,$E3,$77,$E3,$8F,$D1,$BE,$0A,$F1,$57
	.dc.b	$08,$D4,$FE,$3D,$B7,$E5,$C7,$8D,$6F,$0E,$A7,$67,$91,$9B,$EF,$19
	.dc.b	$9E,$26,$22,$1D,$E7,$8C,$62,$BE,$44,$F0,$35,$19,$C3,$E3,$46,$67
	.dc.b	$CC,$FE,$F2,$32,$D3,$E5,$1F,$BB,$7E,$0C,$D8,$F9,$4F,$57,$67,$80
	.dc.b	$93,$E4,$7E,$35,$FB,$5F,$DD,$7A,$0E,$07,$95,$FE,$2A,$8C,$36,$33
	.dc.b	$3F,$D1,$E3,$07,$F8,$5B,$0F,$F1,$2F,$43,$E3,$F1,$C1,$FD,$7D,$B7
	.dc.b	$FD,$0A,$EE,$3E,$1F,$8D,$B9,$1A,$70,$EB,$FE,$C5,$F3,$58,$C3,$FA
	.dc.b	$34,$D7,$FB,$C4,$E9,$C9,$FD,$76,$3D,$44,$70,$7F,$88,$23,$FB,$77
	.dc.b	$D3,$F7,$D0,$F5,$D7,$FC,$B2,$7F,$9B,$75,$FF,$27,$B6,$FF,$C2,$B1
	.dc.b	$8F,$03,$FD,$41,$39,$9F,$2D,$BE,$C7,$B1,$F1,$27,$47,$CA,$E6,$F1
	.dc.b	$12,$BC,$D4,$FA,$5C,$47,$49,$B5,$F4,$A7,$A7,$D6,$3D,$2F,$3B,$6F
	.dc.b	$FC,$5A,$D7,$FD,$0C,$D7,$EE,$39,$AF,$BB,$F6,$1F,$14,$EB,$5D,$BE
	.dc.b	$C3,$A9,$2F,$AB,$5F,$AB,$EA,$3E,$0F,$EE,$36,$3F,$C8,$58,$AB,$D3
	.dc.b	$E5,$7A,$09,$3E,$86,$BD,$3A,$F9,$9C,$1F,$F2,$D8,$FC,$D1,$F0,$1F
	.dc.b	$DF,$DB,$1F,$E8,$86,$BF,$E7,$03,$FD,$22,$E1,$C3,$AD,$7F,$07,$FB
	.dc.b	$A3,$0F,$B4,$DA,$F0,$D6,$F9,$3F,$CF,$6D,$8F,$F4,$64,$ED,$FF,$28
	.dc.b	$5A,$30,$FF,$A1,$A7,$84,$BE,$C3,$E4,$3E,$73,$1B,$7F,$86,$E3,$31
	.dc.b	$EB,$6B,$FE,$8C,$78,$D7,$FA,$73,$EC,$3F,$61,$3F,$A3,$6B,$FE,$63
	.dc.b	$3A,$9F,$ED,$CF,$9A,$FB,$3F,$39,$23,$04,$3E,$FB,$97,$F0,$31,$FC
	.dc.b	$09,$7D,$63,$EC,$3F,$C6,$4D,$8F,$F5,$DB,$5F,$73,$5F,$AC,$E7,$0F
	.dc.b	$F5,$5F,$D8,$7A,$1B,$DD,$DE,$75,$BF,$FA,$B1,$B7,$FE,$CE,$BE,$13
	.dc.b	$FE,$B0,$6C,$7F,$B4,$A7,$7D,$DF,$EE,$6F,$27,$F3,$3F,$39,$2B,$FD
	.dc.b	$FF,$1F,$9C,$30,$FE,$82,$30,$FE,$7B,$4F,$D9,$91,$FD,$9B,$5F,$E3
	.dc.b	$7F,$5C,$F4,$3F,$91,$36,$39,$1F,$C6,$5F,$E6,$0D,$F3,$F6,$70,$F2
	.dc.b	$24,$64,$87,$67,$2F,$82,$38,$BD,$63,$C3,$E1,$78,$BC,$49,$3E,$18
	.dc.b	$CB,$E3,$8C,$BE,$A8,$F4,$FD,$9F,$AB,$C1,$93,$E9,$8C,$BE,$B8,$CB
	.dc.b	$F0,$A3,$ED,$7C,$97,$E1,$89,$1F,$86,$6B,$F9,$7E,$17,$CE,$7E,$3D
	.dc.b	$B1,$FC,$13,$89,$5D,$23,$F8,$3C,$E9,$E5,$59,$7E,$9F,$CA,$7D,$22
	.dc.b	$3F,$47,$E0,$3F,$39,$F5,$EF,$88,$FA,$FF,$3F,$2B,$F8,$9C,$E5,$FB
	.dc.b	$1E,$DF,$DC,$23,$F6,$E7,$EF,$FB,$7D,$3E,$26,$4F,$DF,$8A,$FE,$1B
	.dc.b	$D6,$FC,$29,$8E,$E7,$6A,$5D,$AE,$9F,$F2,$6C,$7E,$26,$F8,$BE,$6C
	.dc.b	$7D,$7E,$D4,$8F,$4B,$FD,$CF,$E1,$9E,$EE,$67,$FB,$8C,$3E,$2B,$46
	.dc.b	$74,$F1,$CF,$8F,$83,$FD,$66,$FF,$71,$AE,$0D,$AE,$1A,$FA,$61,$F5
	.dc.b	$DF,$A1,$7F,$8A,$37,$0E,$B1,$AE,$FF,$E1,$DF,$5D,$D2,$37,$18,$F4
	.dc.b	$64,$7F,$46,$BE,$58,$77,$6C,$7C,$9B,$57,$08,$E6,$7F,$4E,$FA,$C7
	.dc.b	$C7,$FF,$1D,$6C,$66,$3F,$73,$EE,$E5,$FD,$DD,$C9,$D1,$0F,$E0,$2E
	.dc.b	$5F,$E6,$55,$D1,$BF,$1F,$DC,$8F,$F0,$3F,$AD,$AF,$F0,$AF,$73,$F8
	.dc.b	$49,$8D,$BE,$3D,$AF,$35,$FC,$2E,$B5,$AF,$D1,$9F,$F1,$1C,$FF,$6C
	.dc.b	$6B,$FD,$15,$B7,$F4,$BE,$27,$FC,$3D,$3B,$F1,$BF,$A0,$47,$F1,$68
	.dc.b	$FE,$98,$D9,$FB,$35,$CA,$FC,$17,$C3,$C7,$98,$73,$3F,$BE,$FD,$5B
	.dc.b	$E5,$3F,$DF,$4D,$8F,$EC,$81,$C3,$FD,$9D,$AF,$F7,$F7,$38,$FE,$2B
	.dc.b	$EB,$3F,$C5,$D3,$FE,$0A,$6B,$FD,$8E,$77,$8F,$EB,$7E,$93,$FD,$E3
	.dc.b	$E0,$FF,$5D,$FC,$8F,$FB,$6C,$FE,$65,$B6,$AD,$BF,$B5,$C7,$F4,$B1
	.dc.b	$88,$FE,$E1,$3C,$13,$CA,$35,$3F,$E5,$4F,$49,$F3,$1F,$E4,$B6,$DF
	.dc.b	$5C,$3B,$46,$23,$FC,$A7,$F1,$3B,$A9,$1B,$17,$FD,$D6,$3F,$CC,$93
	.dc.b	$FE,$78,$F8,$0F,$E3,$55,$AC,$3F,$87,$7C,$37,$F9,$3F,$C8,$D1,$FE
	.dc.b	$79,$FA,$97,$CA,$62,$FF,$C3,$15,$B6,$E1,$D1,$FE,$01,$19,$7F,$C6
	.dc.b	$8D,$CF,$FD,$1F,$E8,$FF,$9D,$CE,$93,$E3,$13,$FE,$61,$F8,$0F,$41
	.dc.b	$23,$FD,$01,$8F,$F8,$0A,$75,$1E,$29,$FF,$51,$61,$FE,$41,$8F,$6B
	.dc.b	$5F,$E5,$9F,$53,$FC,$03,$5F,$C7,$B6,$3F,$C6,$8D,$7F,$C9,$8D,$7F
	.dc.b	$DD,$9E,$67,$E0,$DF,$F1,$D7,$C6,$23,$11,$FE,$53,$8C,$71,$7E,$71
	.dc.b	$B1,$DB,$A9,$ED,$78,$9B,$19,$8C,$1F,$43,$EA,$7F,$D5,$DC,$7F,$7F
	.dc.b	$64,$62,$BF,$D2,$FE,$8F,$FA,$82,$BE,$EC,$E7,$8B,$F9,$DD,$F0,$F7
	.dc.b	$35,$87,$D0,$AD,$5F,$F5,$1F,$45,$F1,$2C,$B7,$8F,$1A,$F1,$5B,$DD
	.dc.b	$FE,$77,$8F,$81,$AF,$56,$D7,$A1,$B3,$C6,$B9,$C7,$93,$1F,$0F,$AB
	.dc.b	$E0,$AD,$61,$F8,$7C,$39,$61,$C0,$FF,$57,$D6,$F0,$F6,$E8,$FA,$9F
	.dc.b	$F0,$6D,$7E,$9E,$31,$EB,$F7,$D5,$EB,$F0,$BD,$E7,$FE,$7E,$6D,$FF
	.dc.b	$B3,$1B,$87,$C7,$BE,$2F,$58,$CD,$7F,$09,$A7,$C8,$7D,$08,$7D,$4B
	.dc.b	$A3,$E9,$F3,$7E,$EA,$B5,$39,$C7,$F7,$D3,$97,$FD,$4F,$19,$BF,$D5
	.dc.b	$4E,$5F,$F4,$14,$67,$0F,$ED,$FF,$4F,$FE,$AC,$BF,$E5,$ED,$AF,$F4
	.dc.b	$A3,$5F,$F5,$63,$5F,$EF,$8E,$E7,$9F,$5F,$E1,$3D,$1F,$F6,$F5,$FF
	.dc.b	$AE,$AB,$4F,$F8,$0F,$E9,$9C,$5F,$8A,$FA,$77,$9B,$F0,$E7,$35,$F9
	.dc.b	$D9,$CB,$FE,$C4,$8C,$E7,$FE,$A9,$6B,$FE,$C1,$6B,$DE,$FF,$22,$A3
	.dc.b	$8E,$7F,$E1,$06,$D7,$F8,$BF,$E8,$5F,$4E,$F9,$7F,$26,$3F,$5F,$E4
	.dc.b	$BF,$C0,$F7,$FC,$BF,$AD,$F4,$FF,$E5,$48,$FF,$6C,$CE,$B4,$FE,$39
	.dc.b	$5A,$F8,$3E,$5E,$BF,$92,$B5,$5F,$ED,$96,$BF,$E2,$96,$BF,$E4,$37
	.dc.b	$A6,$DF,$E2,$86,$E1,$FB,$7C,$5F,$4F,$DA,$3C,$7E,$1F,$EA,$E5,$F3
	.dc.b	$9E,$B8,$C3,$67,$FB,$8D,$FF,$83,$5B,$3E,$72,$3D,$0F,$85,$5F,$E2
	.dc.b	$E7,$18,$FF,$B4,$5F,$E4,$87,$F5,$51,$F1,$DF,$EA,$57,$FD,$FE,$6A
	.dc.b	$3F,$A1,$3A,$F7,$7F,$4C,$E6,$3F,$DF,$CD,$7F,$C9,$CD,$7F,$CD,$0D
	.dc.b	$7F,$CD,$7E,$47,$23,$F8,$FC,$3E,$9D,$F4,$F9,$8F,$ED,$54,$E5,$FE
	.dc.b	$EB,$1F,$7F,$E9,$DF,$0D,$FB,$99,$CD,$7D,$E9,$CD,$72,$9C,$CF,$F9
	.dc.b	$8C,$E9,$E9,$FE,$52,$33,$1F,$E6,$76,$BF,$93,$8E,$C7,$F2,$3A,$FD
	.dc.b	$5C,$66,$BE,$8F,$B3,$F7,$4F,$27,$CF,$D8,$F1,$3D,$C8,$FF,$8B,$B0
	.dc.b	$47,$F3,$6E,$7F,$CD,$BC,$A7,$EE,$B6,$FF,$37,$C1,$FA,$59,$FC,$0F
	.dc.b	$4F,$E3,$7D,$DF,$C7,$3F,$15,$F7,$12,$3C,$0F,$79,$7D,$1F,$E5,$4F
	.dc.b	$CB,$7D,$3D,$AF,$A7,$E6,$BE,$9B,$FA,$BF,$A4,$FA,$77,$C9,$7F,$2B
	.dc.b	$AD,$5F,$F8,$8E,$B5,$39,$BF,$E0,$75,$AD,$7F,$C3,$71,$82,$5F,$81
	.dc.b	$EF,$F2,$5F,$9D,$9F,$F0,$B1,$FE,$16,$8F,$D6,$75,$FE,$B7,$C0,$B7
	.dc.b	$FA,$46,$D7,$C8,$6F,$81,$FE,$AC,$E3,$FE,$13,$1F,$B5,$F8,$9F,$DB
	.dc.b	$7C,$65,$C3,$F5,$AD,$C3,$E8,$B7,$0E,$5C,$5F,$EB,$CA,$FE,$29,$F1
	.dc.b	$5F,$ED,$B7,$F6,$A7,$F6,$B7,$17,$F6,$A3,$1E,$FF,$F0,$56,$9C,$0E
	.dc.b	$6E,$5A,$66,$3F,$C7,$16,$F9,$EF,$F2,$3C,$F8,$BB,$7F,$B8,$C6,$2B
	.dc.b	$F9,$1E,$1F,$A6,$78,$C6,$6F,$FD,$9E,$73,$F8,$0F,$47,$9F,$C9,$F5
	.dc.b	$7B,$DE,$BF,$4E,$E7,$F0,$1B,$8F,$66,$E1,$F3,$F2,$FE,$E3,$5F,$8C
	.dc.b	$FA,$0F,$F0,$2F,$D3,$7E,$0E,$BF,$DA,$7C,$E9,$CF,$37,$E3,$5F,$C4
	.dc.b	$92,$FF,$D2,$F3,$B6,$C6,$DF,$C9,$75,$3E,$27,$1A,$7C,$2C,$3E,$A7
	.dc.b	$D3,$FD,$56,$DF,$DA,$6C,$EE,$3E,$EC,$FF,$82,$7D,$FF,$91,$1F,$E3
	.dc.b	$7F,$8F,$FD,$5B,$FC,$6A,$33,$4E,$A5,$6A,$3F,$37,$3A,$7F,$C7,$DC
	.dc.b	$5F,$E3,$19,$FE,$9B,$03,$37,$FC,$9B,$E9,$9E,$17,$CA,$FA,$7F,$F0
	.dc.b	$E4,$62,$7F,$C5,$DC,$49,$CD,$7F,$74,$AD,$5F,$F2,$50,$E1,$FE,$50
	.dc.b	$D7,$C4,$6A,$39,$DF,$E8,$3E,$86,$CF,$E3,$3D,$1F,$D0,$7C,$A7,$E4
	.dc.b	$9B,$5F,$E4,$1E,$AF,$EC,$6D,$8F,$D3,$65,$B8,$D7,$91,$FA,$9F,$A0
	.dc.b	$FC,$DB,$63,$24,$C7,$F3,$BC,$DA,$8E,$DE,$4F,$F9,$8B,$A7,$F4,$CD
	.dc.b	$CF,$F9,$55,$FF,$45,$9C,$D7,$C3,$6C,$7E,$9A,$BF,$CA,$DE,$87,$EB
	.dc.b	$1B,$5F,$E1,$3F,$81,$FE,$D5,$7F,$C3,$57,$C6,$EB,$0F,$F0,$9C,$6A
	.dc.b	$BF,$CB,$18,$7F,$98,$6B,$57,$FB,$2C,$7F,$86,$61,$E5,$6B,$F4,$A3
	.dc.b	$F8,$07,$89,$EB,$26,$1F,$35,$FF,$36,$7C,$B7,$F0,$27,$FD,$13,$F4
	.dc.b	$9F,$B2,$BF,$F6,$2B,$5F,$F6,$4B,$5F,$F6,$AB,$5F,$F6,$2B,$C6,$FF
	.dc.b	$D3,$6D,$BF,$F3,$97,$CB,$76,$F0,$A4,$7F,$A9,$E7,$57,$FA,$3D,$57
	.dc.b	$A8,$FF,$46,$4E,$BC,$13,$C3,$B7,$FD,$47,$5B,$CB,$C3,$CD,$FE,$9C
	.dc.b	$A3,$49,$97,$FB,$A2,$32,$F8,$E3,$FD,$21,$F2,$DF,$11,$FF,$42,$65
	.dc.b	$FA,$FF,$A6,$FD,$87,$D3,$3E,$93,$FC,$15,$E8,$B1,$CC,$7E,$A6,$AD
	.dc.b	$FE,$9A,$B4,$AF,$93,$FE,$13,$9F,$F2,$65,$74,$BB,$39,$1E,$E7,$07
	.dc.b	$F8,$16,$1D,$23,$11,$E7,$3B,$57,$36,$9C,$F0,$DD,$79,$1E,$30,$C3
	.dc.b	$D2,$F6,$C2,$BD,$67,$8C,$62,$38,$1E,$6B,$F0,$61,$D2,$FD,$73,$CC
	.dc.b	$D1,$E3,$9F,$35,$EE,$3C,$E6,$DC,$1E,$07,$27,$8E,$38,$9F,$45,$CE
	.dc.b	$1E,$D8,$C4,$F1,$3B,$5F,$32,$C7,$A0,$D9,$D2,$BC,$C7,$DB,$C3,$E0
	.dc.b	$61,$EC,$86,$1F,$4E,$BC,$57,$D5,$A7,$59,$FB,$91,$97,$EF,$7C,$47
	.dc.b	$DE,$6C,$7A,$E3,$11,$FC,$C2,$FF,$A1,$46,$3E,$3D,$CA,$DF,$38,$7F
	.dc.b	$36,$9C,$C6,$30,$FE,$95,$18,$7F,$DC,$13,$BA,$E5,$18,$36,$E2,$3F
	.dc.b	$25,$3F,$B8,$9F,$BA,$D7,$EE,$87,$B4,$70,$9E,$58,$F9,$78,$3F,$D5
	.dc.b	$CD,$BE,$73,$A7,$C3,$87,$E5,$30,$ED,$AF,$8B,$0F,$EA,$6F,$59,$D5
	.dc.b	$F1,$BF,$15,$76,$7E,$0B,$5F,$34,$6E,$FE,$25,$F5,$9F,$05,$76,$C3
	.dc.b	$E3,$61,$D8,$CC,$C3,$F3,$37,$CA,$77,$18,$AE,$4F,$88,$E4,$70,$C3
	.dc.b	$C5,$1E,$5A,$F9,$98,$FB,$6B,$E6,$C7,$57,$91,$D4,$E7,$8E,$CF,$53
	.dc.b	$93,$B6,$3E,$0A,$C9,$31,$F3,$BC,$5F,$A9,$3C,$6B,$0F,$F4,$2E,$AF
	.dc.b	$EA,$5B,$7F,$46,$6C,$64,$FE,$C9,$F3,$38,$E4,$FE,$77,$3F,$E3,$8E
	.dc.b	$3F,$F1,$F5,$7F,$35,$8C,$19,$F9,$0E,$12,$69,$DB,$95,$CC,$F8,$7C
	.dc.b	$E1,$FD,$3E,$AB,$FD,$C2,$2E,$FF,$D6,$26,$E9,$FE,$2D,$C3,$9D,$70
	.dc.b	$BF,$EC,$75,$E3,$7F,$A5,$3F,$E2,$77,$F6,$1D,$5F,$CF,$9F,$F3,$5D
	.dc.b	$FC,$2C,$3F,$CE,$8F,$9F,$17,$D9,$66,$F8,$5F,$F7,$B6,$BF,$E3,$FB
	.dc.b	$E2,$FF,$84,$E3,$2F,$F6,$9E,$1E,$84,$7F,$C2,$5F,$4F,$5F,$F0,$7C
	.dc.b	$5A,$FF,$0A,$4F,$97,$E9,$DE,$8D,$25,$FF,$86,$EE,$BF,$E2,$59,$DB
	.dc.b	$FE,$25,$6B,$99,$F8,$96,$97,$8D,$DC,$95,$FE,$A6,$F9,$1A,$FF,$A5
	.dc.b	$3E,$17,$EE,$7A,$67,$38,$7F,$A5,$30,$78,$FE,$92,$5F,$FA,$C3,$E0
	.dc.b	$EB,$FE,$9E,$EA,$E6,$62,$FE,$B7,$D3,$3D,$6F,$CB,$D2,$30,$FD,$1E
	.dc.b	$4F,$F2,$FC,$62,$3F,$D5,$3C,$BC,$23,$11,$FE,$64,$39,$57,$E1,$5A
	.dc.b	$7E,$1E,$36,$F4,$8F,$C3,$7D,$3F,$B2,$3C,$07,$F3,$B7,$0F,$FB,$82
	.dc.b	$7C,$58,$7E,$3F,$E9,$FE,$34,$F0,$0E,$1F,$F7,$3B,$63,$FD,$C1,$F4
	.dc.b	$FF,$E0,$B7,$96,$FF,$AB,$8F,$D7,$57,$8B,$A9,$18,$8E,$73,$F5,$BE
	.dc.b	$9B,$EB,$FD,$3C,$9F,$F7,$0C,$65,$FF,$71,$C6,$5F,$F6,$8C,$7D,$0F
	.dc.b	$A6,$7C,$F5,$24,$FA,$FE,$99,$58,$8F,$F4,$B5,$70,$7F,$64,$D7,$35
	.dc.b	$F4,$63,$19,$FE,$D6,$FD,$4D,$E0,$E8,$FF,$B5,$7C,$AF,$C8,$FC,$DC
	.dc.b	$C7,$FD,$F7,$1F,$94,$30,$FF,$BA,$A3,$0F,$D8,$9F,$D0,$65,$FA,$D2
	.dc.b	$47,$CB,$E8,$FF,$47,$72,$7E,$F7,$A3,$DB,$3F,$3F,$E9,$9F,$21,$FC
	.dc.b	$75,$B1,$D6,$76,$FF,$7F,$AF,$E3,$12,$FA,$4E,$1F,$15,$2B,$A4,$73
	.dc.b	$30,$F1,$3F,$DA,$92,$9D,$47,$03,$C8,$F0,$CF,$FC,$B7,$CA,$AF,$05
	.dc.b	$7A,$7E,$2B,$D1,$F0,$72,$30,$43,$C1,$5C,$63,$51,$8A,$57,$C1,$FA
	.dc.b	$1F,$85,$F4,$F2,$BE,$34,$65,$F9,$11,$97,$94,$7E,$37,$E5,$64,$7E
	.dc.b	$97,$99,$1F,$C0,$39,$9E,$06,$AB,$F1,$B3,$F8,$92,$3F,$8E,$78,$78
	.dc.b	$52,$3A,$47,$D7,$F9,$69,$1F,$6D,$FB,$5F,$0B,$FC,$3B,$E9,$DF,$17
	.dc.b	$FC,$15,$B5,$FD,$2D,$AF,$FB,$CD,$AF,$FB,$C3,$87,$E8,$7B,$BF,$85
	.dc.b	$24,$7E,$27,$E8,$3F,$11,$EF,$F7,$F8,$C9,$3F,$8D,$8A,$F5,$8F,$11
	.dc.b	$1F,$E3,$47,$E1,$3D,$AF,$C9,$11,$FE,$1D,$8A,$7D,$9E,$EC,$9C,$BF
	.dc.b	$DD,$7E,$21,$F0,$39,$9F,$64,$FC,$12,$3F,$05,$E1,$F1,$D2,$39,$9F
	.dc.b	$C7,$7E,$77,$F4,$6F,$E6,$43,$8E,$CF,$EE,$4D,$EF,$4C,$F8,$25,$FF
	.dc.b	$40,$AE,$73,$FC,$E5,$AF,$03,$AF,$85,$FC,$EE,$33,$1D,$1E,$CF,$5F
	.dc.b	$0F,$FA,$5D,$9E,$D5,$FD,$3B,$E9,$DF,$23,$49,$7F,$C0,$7E,$0A,$BF
	.dc.b	$4A,$76,$FF,$83,$DA,$E6,$7F,$A7,$46,$34,$FD,$AF,$84,$CB,$F8,$51
	.dc.b	$D2,$3F,$C2,$D2,$F9,$2F,$21,$30,$FE,$25,$F0,$35,$F1,$47,$62,$3E
	.dc.b	$37,$F8,$B7,$BD,$23,$FA,$95,$75,$7F,$AA,$4B,$C6,$FE,$DE,$D8,$CC
	.dc.b	$7F,$72,$F8,$BF,$E0,$31,$83,$FC,$5D,$3B,$7F,$2A,$D7,$FB,$83,$5F
	.dc.b	$B0,$EF,$1F,$F1,$A4,$FF,$16,$9F,$C4,$36,$3F,$C7,$13,$BE,$2F,$C0
	.dc.b	$7D,$33,$E5,$FF,$C0,$AD,$C3,$FB,$73,$5F,$C9,$7D,$3B,$E2,$37,$5F
	.dc.b	$E0,$1A,$D5,$7F,$81,$5A,$F6,$8F,$F2,$57,$D1,$7F,$BC,$7C,$0F,$F0
	.dc.b	$9F,$D3,$FF,$77,$CB,$F2,$4D,$B5,$6D,$FF,$03,$B5,$FD,$24,$6D,$3C
	.dc.b	$63,$5A,$2B,$FC,$9D,$EA,$BE,$73,$FC,$B0,$DC,$3F,$AE,$35,$FE,$73
	.dc.b	$F0,$DE,$F7,$33,$C0,$8F,$F8,$26,$34,$FF,$4F,$F8,$9F,$F2,$7F,$8F
	.dc.b	$FE,$83,$F8,$3F,$E3,$F9,$78,$E9,$5A,$7F,$C2,$5F,$4E,$F9,$5E,$18
	.dc.b	$49,$FF,$13,$D6,$DF,$E9,$B5,$B7,$38,$7F,$2A,$7F,$81,$CE,$DF,$DF
	.dc.b	$35,$CC,$6D,$F1,$BF,$E3,$0D,$33,$F0,$5E,$93,$FD,$25,$7F,$E0,$E8
	.dc.b	$CD,$78,$27,$4F,$AB,$5F,$F2,$7F,$D3,$FF,$02,$E4,$FF,$45,$36,$7F
	.dc.b	$98,$B5,$FF,$26,$35,$FF,$7E,$79,$5F,$82,$7F,$9B,$FC,$72,$31,$1F
	.dc.b	$E5,$38,$C7,$17,$FA,$5D,$B1,$DB,$A1,$EE,$F8,$9B,$19,$8C,$1F,$D1
	.dc.b	$37,$FF,$43,$EF,$F5,$F8,$BE,$C4,$63,$7F,$F2,$CF,$E0,$2F,$8A,$FE
	.dc.b	$DB,$1F,$4A,$F8,$3F,$39,$B8,$FC,$E6,$D7,$1F,$57,$38,$75,$AF,$C5
	.dc.b	$53,$E6,$3F,$D4,$F3,$FE,$9A,$6C,$67,$0F,$43,$6B,$FC,$ED,$5B,$9F
	.dc.b	$C2,$D7,$8B,$D2,$72,$FE,$8E,$72,$4B,$F0,$D6,$E7,$51,$89,$FF,$21
	.dc.b	$7C,$C7,$FA,$35,$BC,$DE,$AF,$A1,$79,$3F,$03,$81,$7B,$C3,$FC,$69
	.dc.b	$18,$BD,$4E,$63,$E7,$DF,$2C,$3A,$9F,$E8,$1F,$A5,$D3,$FE,$C9,$FA
	.dc.b	$A7,$D5,$FC,$EC,$95,$A5,$82,$B7,$7C,$F2,$FF,$67,$DF,$96,$FF,$D3
	.dc.b	$CD,$7F,$B3,$35,$E7,$93,$E7,$7F,$DB,$78,$FB,$9B,$19,$C3,$FC,$1F
	.dc.b	$39,$6C,$66,$BE,$1B,$67,$FD,$DD,$C0,$F9,$5F,$8B,$87,$C6,$F9,$BF
	.dc.b	$B5,$E7,$3E,$4F,$19,$FF,$9B,$5B,$9F,$FA,$DF,$2F,$A3,$7E,$42,$57
	.dc.b	$89,$DC,$7F,$BF,$7E,$75,$E5,$E2,$BF,$DE,$DA,$F8,$DF,$D0,$FE,$43
	.dc.b	$FE,$D9,$7F,$07,$AF,$E7,$23,$19,$7B,$23,$E1,$92,$7E,$0B,$63,$51
	.dc.b	$D3,$E0,$7F,$07,$3F,$84,$F8,$DF,$D3,$F3,$7E,$BA,$FE,$3C,$63,$AF
	.dc.b	$58,$7C,$C8,$FA,$24,$70,$7E,$61,$E7,$F1,$DC,$2B,$F2,$7F,$7D,$FE
	.dc.b	$85,$AF,$F4,$B7,$8D,$F9,$98,$C4,$FF,$9E,$63,$1F,$15,$F8,$09,$CC
	.dc.b	$FF,$8A,$F8,$51,$88,$FF,$1A,$C6,$27,$81,$CA,$FF,$85,$FA,$DF,$97
	.dc.b	$9D,$47,$FA,$BE,$DF,$07,$FB,$9F,$DE,$7F,$A3,$4E,$63,$1E,$6F,$9D
	.dc.b	$F2,$12,$3F,$23,$5C,$9B,$1B,$7F,$DC,$55,$FE,$D6,$7A,$18,$EE,$F0
	.dc.b	$1F,$0F,$82,$C7,$95,$F8,$44,$7B,$BF,$17,$D8,$91,$E2,$6E,$BF,$E5
	.dc.b	$11,$E1,$B0,$7D,$87,$BC,$F8,$1E,$8D,$C3,$E2,$76,$7D,$12,$3E,$E7
	.dc.b	$FB,$76,$1F,$E8,$BB,$FF,$51,$CF,$F3,$8F,$D7,$7F,$C4,$0F,$C0,$F8
	.dc.b	$8F,$E6,$71,$8A,$FF,$38,$78,$2F,$92,$FE,$18,$BE,$5E,$C6,$1B,$87
	.dc.b	$FA,$32,$BF,$BE,$CF,$89,$FE,$E2,$D7,$FB,$AB,$5D,$C7,$E1,$8D,$95
	.dc.b	$FE,$F7,$D3,$D9,$FF,$66,$E4,$FA,$2F,$C2,$1F,$85,$F2,$2D,$FD,$C8
	.dc.b	$B5,$C2,$6F,$3F,$F5,$DF,$DD,$FE,$2C,$FF,$B5,$7E,$77,$F9,$C4,$62
	.dc.b	$7F,$CD,$F1,$8D,$E6,$3F,$88,$F8,$3B,$87,$F3,$CC,$3F,$D0,$93,$F8
	.dc.b	$CF,$73,$9D,$7E,$65,$B1,$FC,$92,$77,$5F,$C5,$DA,$FE,$AD,$AF,$EA
	.dc.b	$9D,$E3,$E9,$BF,$D2,$7D,$1F,$4A,$F8,$BF,$47,$FA,$73,$F5,$1E,$3D
	.dc.b	$C7,$F3,$F1,$67,$E5,$CD,$C3,$E2,$7D,$37,$EB,$CF,$86,$5B,$FA,$90
	.dc.b	$C9,$3A,$9F,$D5,$FD,$3D,$C3,$F1,$7F,$35,$F8,$AF,$9A,$FC,$5B,$6B
	.dc.b	$EF,$4F,$F9,$12,$3F,$09,$CE,$69,$27,$F1,$D1,$97,$8B,$67,$FA,$FC
	.dc.b	$7F,$B1,$3D,$67,$DB,$31,$F0,$CF,$AB,$3F,$F5,$BB,$FE,$52,$F0,$7F
	.dc.b	$03,$F1,$3F,$E0,$16,$C6,$E3,$FC,$09,$DF,$8C,$BF,$CA,$7C,$FF,$E0
	.dc.b	$73,$FC,$11,$CD,$6F,$FC,$BB,$16,$7F,$CB,$33,$78,$7F,$A8,$71,$FF
	.dc.b	$83,$FA,$3F,$BD,$79,$E8,$C6,$3F,$1A,$30,$FE,$BD,$B8,$7F,$BF,$1A
	.dc.b	$FF,$0D,$6B,$FE,$DF,$77,$7E,$09,$FF,$38,$F3,$F8,$2F,$FC,$41,$CF
	.dc.b	$27,$FC,$EB,$D1,$71,$FF,$38,$CD,$C3,$FC,$F9,$E0,$FF,$8B,$2B,$17
	.dc.b	$FE,$F4,$8F,$F1,$7E,$1F,$E2,$B3,$FC,$59,$D3,$67,$FD,$03,$17,$1F
	.dc.b	$F3,$EC,$DC,$3F,$CA,$5E,$43,$D8,$FF,$07,$B7,$0F,$F0,$6B,$6F,$FC
	.dc.b	$00,$D7,$FA,$A3,$9A,$E7,$5F,$E3,$DE,$BF,$F1,$B4,$E6,$FD,$1E,$BF
	.dc.b	$E1,$DF,$F4,$AC,$FF,$90,$F9,$F8,$47,$FA,$61,$B8,$FF,$A5,$1B,$8F
	.dc.b	$A3,$E3,$7C,$E4,$AF,$3B,$6B,$FC,$87,$8A,$3F,$CA,$1C,$DE,$47,$FC
	.dc.b	$CB,$F6,$4F,$99,$FF,$31,$B7,$0F,$F3,$C3,$70,$FF,$30,$B5,$FF,$50
	.dc.b	$B9,$C7,$FC,$F6,$D9,$FA,$DF,$8C,$FF,$41,$9C,$DF,$80,$93,$FC,$CF
	.dc.b	$85,$5B,$8F,$F5,$7D,$F3,$C3,$FD,$7A,$DE,$E3,$9B,$E3,$C7,$D2,$BF
	.dc.b	$D9,$13,$FD,$D5,$F4,$F9,$7E,$E2,$61,$EE,$7E,$19,$FE,$75,$F7,$EC
	.dc.b	$7F,$0A,$90,$FA,$7B,$F4,$FF,$D6,$63,$27,$FA,$92,$73,$E4,$FF,$73
	.dc.b	$8E,$02,$78,$3D,$3E,$07,$FD,$B5,$18,$8E,$F7,$E0,$F3,$31,$5F,$ED
	.dc.b	$F7,$53,$D1,$CE,$1B,$C3,$C1,$7E,$2A,$F2,$DF,$69,$CC,$78,$DF,$F0
	.dc.b	$44,$66,$3F,$C1,$53,$C5,$F3,$55,$7C,$A7,$13,$AB,$97,$C9,$5D,$23
	.dc.b	$C9,$1D,$0F,$F3,$74,$65,$A6,$70,$F5,$BD,$4F,$9B,$5E,$1B,$F3,$46
	.dc.b	$0D,$1F,$44,$A7,$D0,$9E,$67,$C8,$9E,$57,$E7,$6C,$F0,$7C,$EF,$C0
	.dc.b	$3C,$A7,$7C,$3C,$A7,$37,$C3,$7E,$47,$E2,$DF,$83,$0F,$9F,$5C,$1A
	.dc.b	$7C,$17,$91,$99,$F8,$CD,$7D,$DE,$9F,$FA,$55,$F9,$4D,$73,$1F,$72
	.dc.b	$31,$F0,$4F,$97,$FF,$5B,$D2,$73,$18,$6F,$27,$E7,$9B,$3B,$BF,$CE
	.dc.b	$DF,$3A,$F8,$AF,$FB,$52,$6F,$A5,$FA,$2A,$FC,$EC,$EE,$7E,$4B,$63
	.dc.b	$33,$BC,$7A,$B6,$BF,$5F,$5B,$9F,$0E,$3E,$AC,$B8,$D7,$E5,$6B,$C4
	.dc.b	$FE,$1D,$F0,$47,$8A,$7E,$0E,$1C,$2B,$C7,$87,$75,$7D,$DF,$0D,$77
	.dc.b	$C3,$CB,$1D,$EB,$F7,$F8,$75,$C7,$C5,$7B,$AF,$4D,$6A,$F8,$C7,$96
	.dc.b	$74,$F8,$1B,$87,$AB,$0F,$31,$D6,$78,$61,$F3,$98,$7C,$CA,$F6,$DF
	.dc.b	$C6,$9F,$34,$C3,$89,$2B,$AC,$F9,$1E,$C7,$7C,$7E,$31,$23,$A4,$F7
	.dc.b	$AE,$95,$CA,$3E,$3C,$7F,$0F,$E3,$FE,$6B,$F3,$3F,$EB,$9D,$FF,$9D
	.dc.b	$B6,$BF,$9D,$B7,$91,$D8,$F9,$5D,$61,$CA,$FF,$A9,$DF,$F8,$DB,$94
	.dc.b	$D3,$1F,$1B,$C2,$0D,$3B,$7E,$F5,$FD,$8E,$BF,$D0,$57,$FD,$2A,$75
	.dc.b	$18,$AE,$2F,$F5,$5A,$E0,$FF,$4C,$9D,$B4,$D7,$5E,$1F,$EC,$B1,$E4
	.dc.b	$7F,$B5,$BF,$E9,$BD,$5D,$AF,$A5,$F8,$91,$86,$C6,$6B,$FB,$36,$AF
	.dc.b	$97,$CD,$73,$AD,$57,$F8,$5A,$30,$7F,$AA,$77,$F9,$BC,$1F,$3F,$D7
	.dc.b	$7C,$75,$3E,$96,$5F,$EB,$EC,$FF,$C6,$2D,$AF,$F1,$85,$70,$C7,$FB
	.dc.b	$C3,$5F,$F3,$29,$FE,$D0,$9E,$0A,$BD,$FE,$F7,$B0,$91,$CB,$0A,$FF
	.dc.b	$8B,$A3,$D8,$47,$FB,$56,$2C,$1F,$0F,$EB,$24,$F9,$3E,$66,$BE,$4A
	.dc.b	$E1,$38,$C7,$FC,$9B,$6B,$F8,$9E,$3C,$C3,$DA,$FC,$7F,$A9,$C4,$73
	.dc.b	$AF,$F2,$43,$A3,$E1,$C7,$E7,$57,$D2,$F5,$7F,$D2,$6D,$7E,$39,$AB
	.dc.b	$FA,$9F,$19,$C6,$75,$3F,$EF,$B6,$BF,$E0,$E6,$BF,$E1,$46,$BC,$CF
	.dc.b	$AB,$F1,$5F,$7C,$E9,$5F,$81,$E5,$93,$F6,$9F,$C2,$7D,$3E,$23,$DD
	.dc.b	$17,$4F,$8F,$3F,$96,$FC,$D7,$F3,$35,$FE,$B8,$6B,$F8,$29,$46,$1F
	.dc.b	$9F,$F1,$7C,$79,$BF,$1F,$F7,$52,$F9,$47,$C9,$FC,$B5,$F1,$14,$93
	.dc.b	$FE,$F9,$F7,$9F,$76,$F9,$DC,$CF,$F8,$BE,$F8,$BF,$E3,$10,$E1,$D5
	.dc.b	$E5,$CB,$E1,$F8,$89,$1F,$EC,$A9,$E1,$A3,$B3,$FC,$86,$FB,$BE,$9B
	.dc.b	$E2,$12,$BF,$CB,$9F,$0C,$F5,$71,$1F,$93,$87,$A3,$FC,$85,$B1,$F2
	.dc.b	$EB,$1E,$14,$8E,$6E,$FC,$8C,$47,$38,$B8,$FF,$AA,$2F,$F9,$54,$FD
	.dc.b	$5F,$A3,$FB,$04,$7C,$B9,$D7,$FC,$B2,$7E,$DF,$95,$3F,$C5,$DB,$1F
	.dc.b	$E5,$BB,$FB,$B3,$93,$F9,$D7,$ED,$7F,$C0,$E3,$FD,$72,$DB,$F8,$F7
	.dc.b	$CA,$FF,$0C,$BE,$63,$F1,$84,$8F,$F4,$25,$7E,$9E,$FE,$84,$79,$67
	.dc.b	$FA,$47,$CA,$3E,$5E,$9C,$67,$F8,$05,$BC,$0F,$29,$23,$AF,$87,$5D
	.dc.b	$C5,$C7,$D3,$3F,$93,$F8,$B9,$5F,$83,$7F,$2D,$CF,$88,$FC,$84,$7E
	.dc.b	$2F,$94,$D2,$47,$E2,$F2,$FC,$F1,$C7,$0F,$8F,$E2,$7F,$65,$EA,$78
	.dc.b	$CF,$A7,$C3,$70,$F8,$B5,$FB,$3F,$4B,$EB,$12,$3E,$B7,$17,$F9,$15
	.dc.b	$FF,$4A,$F9,$0E,$17,$8B,$F6,$09,$3F,$43,$C7,$FE,$A7,$FB,$C3,$E2
	.dc.b	$7F,$AC,$36,$7F,$04,$D7,$F1,$3B,$A3,$11,$FE,$A4,$8C,$79,$5C,$89
	.dc.b	$1C,$9F,$EA,$D8,$FF,$6E,$9F,$C4,$F9,$1F,$8C,$3C,$73,$E4,$F7,$25
	.dc.b	$FE,$CB,$AF,$F6,$78,$A7,$F2,$7E,$46,$AB,$CB,$39,$7D,$33,$EC,$23
	.dc.b	$F9,$0F,$3B,$F3,$67,$1C,$BF,$22,$7B,$7C,$4F,$8A,$49,$FE,$29,$FA
	.dc.b	$37,$E5,$F2,$F1,$CF,$C8,$24,$FC,$86,$BF,$96,$F9,$5F,$D4,$9C,$DF
	.dc.b	$E5,$5F,$5D,$CD,$FE,$D4,$DC,$0E,$7A,$5F,$95,$B8,$FF,$C1,$73,$FD
	.dc.b	$1A,$B7,$E7,$FF,$84,$63,$04,$7F,$21,$5F,$DB,$30,$FE,$5C,$D7,$C7
	.dc.b	$1D,$39,$64,$7D,$13,$CA,$FE,$67,$E5,$FF,$CE,$78,$7C,$8C,$3F,$1F
	.dc.b	$5F,$C2,$5A,$FE,$21,$AF,$E1,$DA,$FF,$2E,$77,$1F,$C0,$B2,$57,$FA
	.dc.b	$76,$BF,$A2,$35,$F9,$37,$F0,$B7,$91,$E7,$BF,$93,$1F,$DB,$FC,$1C
	.dc.b	$CF,$8E,$3E,$31,$1F,$43,$FC,$7B,$E6,$24,$73,$CB,$FC,$63,$7F,$C5
	.dc.b	$A1,$F5,$4F,$96,$FF,$02,$36,$3E,$41,$A8,$FF,$04,$7D,$5B,$E8,$77
	.dc.b	$1F,$D5,$6B,$85,$7F,$36,$38,$1F,$D7,$5C,$BF,$CE,$BE,$93,$F3,$CE
	.dc.b	$DA,$E7,$9F,$FB,$04,$62,$04,$78,$E3,$EB,$FA,$5F,$E5,$D3,$E8,$6B
	.dc.b	$FE,$1A,$9D,$F3,$FD,$7E,$1F,$C8,$92,$7F,$CD,$FF,$AD,$FF,$8E,$5F
	.dc.b	$91,$19,$7F,$3E,$FF,$A0,$23,$51,$FE,$4C,$F7,$DF,$4A,$F9,$1F,$F4
	.dc.b	$2B,$67,$FC,$F6,$D7,$FC,$6D,$C7,$A9,$91,$A7,$FC,$C5,$18,$8F,$EC
	.dc.b	$35,$DA,$7F,$C1,$0D,$8F,$A0,$D7,$74,$F9,$0F,$19,$23,$FC,$2D,$FB
	.dc.b	$4B,$E3,$F1,$1E,$BA,$E0,$FA,$2B,$FA,$05,$6E,$3F,$0B,$F4,$DF,$E0
	.dc.b	$9F,$8C,$FF,$0D,$36,$3F,$04,$D7,$FB,$CB,$5F,$F7,$BF,$A8,$FA,$6F
	.dc.b	$F1,$17,$C6,$7F,$6F,$6C,$66,$7F,$19,$39,$7A,$FB,$8F,$8E,$F0,$B6
	.dc.b	$33,$18,$3E,$2F,$A0,$3F,$93,$F2,$7E,$33,$83,$11,$8A,$78,$3F,$DE
	.dc.b	$7D,$1C,$56,$AF,$FD,$6B,$7C,$DA,$F1,$AE,$11,$83,$7E,$2F,$29,$FE
	.dc.b	$45,$1F,$EA,$8F,$2F,$13,$23,$66,$4B,$8F,$E3,$EB,$8B,$67,$D1,$E7
	.dc.b	$FF,$B5,$AF,$94,$FF,$97,$FE,$93,$FD,$27,$87,$F0,$D6,$C7,$E6,$7E
	.dc.b	$69,$F2,$1F,$07,$BD,$5A,$7F,$A6,$B6,$FF,$C9,$AD,$B6,$AC,$39,$D7
	.dc.b	$7F,$1F,$E6,$53,$E9,$7C,$2F,$F8,$F1,$AF,$EF,$8D,$47,$F1,$77,$11
	.dc.b	$A7,$F8,$1B,$E1,$77,$3F,$EE,$8F,$AB,$FF,$18,$3F,$C6,$A7,$CD,$97
	.dc.b	$B3,$CA,$F7,$67,$FE,$0E,$AF,$E3,$F9,$BC,$DC,$D4,$9F,$90,$FA,$A7
	.dc.b	$51,$83,$E6,$7A,$3F,$4B,$DD,$7A,$FD,$A3,$0D,$9D,$9F,$CB,$BE,$6B
	.dc.b	$F6,$CE,$36,$FB,$15,$B7,$FD,$67,$3A,$73,$3F,$6F,$2F,$C8,$7D,$37
	.dc.b	$F5,$08,$C1,$FE,$93,$71,$3F,$83,$9F,$82,$70,$F4,$2D,$70,$9D,$3C
	.dc.b	$63,$31,$F8,$BF,$A3,$FC,$F7,$27,$F9,$0A,$BF,$19,$18,$EA,$F9,$AC
	.dc.b	$8F,$F5,$24,$F8,$B8,$7F,$17,$39,$BE,$95,$AF,$27,$15,$FE,$1C,$FA
	.dc.b	$57,$C5,$7C,$DF,$13,$F2,$D1,$8B,$FE,$FD,$D6,$E3,$56,$AE,$57,$FA
	.dc.b	$D7,$E2,$7F,$07,$87,$1C,$FF,$D8,$8D,$CB,$C1,$7F,$BB,$AD,$BE,$0F
	.dc.b	$98,$FD,$7F,$C3,$3E,$9B,$19,$7F,$AB,$1B,$97,$4A,$DF,$AB,$8C,$BD
	.dc.b	$91,$6F,$FA,$EE,$CC,$7F,$B7,$47,$F8,$0F,$45,$78,$3D,$F9,$39,$E2
	.dc.b	$B9,$FA,$9F,$8A,$7C,$6F,$8B,$B3,$CD,$C3,$FE,$C2,$F8,$DB,$5F,$EC
	.dc.b	$1F,$A1,$F6,$9E,$EF,$12,$D7,$8F,$B7,$E8,$9F,$C7,$FE,$9B,$FC,$B9
	.dc.b	$19,$3F,$D9,$BF,$45,$9F,$67,$E3,$B6,$3F,$5F,$ED,$FE,$AF,$E6,$BF
	.dc.b	$58,$D8,$C9,$33,$72,$7F,$86,$7C,$37,$6F,$0A,$D4,$E6,$FF,$6D,$39
	.dc.b	$BF,$F6,$E7,$D1,$7D,$83,$EC,$F9,$57,$0F,$F7,$1C,$59,$FF,$71,$4D
	.dc.b	$C3,$FC,$DB,$F4,$6D,$3F,$93,$37,$0F,$F7,$5F,$D1,$E2,$64,$7F,$8E
	.dc.b	$E7,$58,$7F,$BB,$6B,$2F,$B2,$7F,$DE,$F2,$9F,$C6,$ED,$D9,$B3,$CD
	.dc.b	$C3,$FE,$FA,$F6,$AD,$F3,$3E,$3D,$F3,$9E,$06,$CB,$1F,$EF,$FF,$A3
	.dc.b	$FC,$D9,$F9,$CF,$66,$CF,$EB,$3B,$13,$8A,$ED,$DB,$F9,$7B,$C6,$BF
	.dc.b	$1B,$E8,$FF,$4A,$7C,$62,$E3,$F0,$DC,$3D,$7E,$8E,$E1,$E9,$3F,$A5
	.dc.b	$FD,$2F,$F2,$FF,$A6,$FE,$60,$DA,$FE,$6E,$E1,$FF,$3F,$7A,$CF,$A6
	.dc.b	$61,$9C,$3A,$74,$30,$C6,$7F,$E7,$7B,$DE,$3D,$0F,$B1,$F1,$5F,$78
	.dc.b	$FE,$DB,$F3,$CE,$FC,$4E,$89,$97,$C4,$9D,$46,$30,$3F,$BE,$8F,$EC
	.dc.b	$CF,$F1,$86,$CF,$1E,$7E,$FD,$12,$B4,$F8,$A3,$13,$FD,$13,$8F,$CB
	.dc.b	$D3,$FD,$BB,$D1,$FA,$F8,$75,$8C,$4C,$8F,$07,$4D,$BE,$51,$FD,$4E
	.dc.b	$7D,$F3,$9C,$FD,$34,$E6,$69,$FE,$5F,$C3,$D5,$D3,$F1,$BA,$BE,$3B
	.dc.b	$23,$27,$BE,$31,$7F,$1F,$2F,$9A,$7C,$EE,$8B,$5F,$16,$2C,$FC,$D9
	.dc.b	$B9,$7F,$B8,$3C,$9F,$F1,$94,$FE,$86,$3F,$A3,$FB,$3F,$63,$AB,$FC
	.dc.b	$39,$5A,$9E,$13,$83,$58,$FF,$AC,$67,$4F,$D7,$E8,$CE,$1F,$75,$B7
	.dc.b	$F1,$36,$F8,$71,$8A,$E0,$7D,$DC,$7C,$3F,$1D,$23,$F0,$1C,$D7,$1F
	.dc.b	$46,$4A,$C6,$5D,$78,$17,$8C,$3F,$CB,$1F,$43,$FE,$79,$7F,$DB,$DD
	.dc.b	$48,$C6,$5F,$B6,$FA,$6F,$E6,$AF,$FA,$5E,$33,$8F,$DC,$AE,$1F,$0B
	.dc.b	$FE,$5E,$9C,$E1,$FD,$53,$0E,$37,$B9,$CE,$1F,$E4,$06,$E1,$CA,$FF
	.dc.b	$B9,$DF,$07,$CF,$87,$52,$3A,$E3,$B8,$79,$7C,$BF,$11,$FB,$0F,$7E
	.dc.b	$D7,$BB,$2F,$EC,$6D,$9D,$34,$FF,$6C,$D7,$FB,$23,$E9,$BF,$D9,$3F
	.dc.b	$4A,$F1,$FF,$C9,$D3,$FE,$89,$AC,$C7,$FB,$66,$AC,$FF,$94,$FE,$07
	.dc.b	$F4,$37,$C6,$38,$CE,$9F,$F6,$AC,$EA,$B9,$42,$B8,$3A,$BE,$0F,$94
	.dc.b	$D4,$F9,$2B,$94,$76,$32,$78,$E7,$8A,$BE,$73,$C2,$3F,$CD,$51,$89
	.dc.b	$DB,$63,$99,$A2,$9F,$0C,$A7,$53,$33,$D5,$F4,$5F,$B1,$E0,$D7,$83
	.dc.b	$98,$E1,$18,$BF,$75,$F9,$23,$AB,$6B,$83,$67,$73,$E4,$BF,$55,$77
	.dc.b	$6C,$FD,$08,$E3,$3B,$3A,$95,$C9,$C5,$CB,$4C,$C7,$DA,$F5,$3F,$CF
	.dc.b	$73,$F4,$DA,$FC,$96,$BF,$6D,$CF,$C1,$3E,$9B,$FC,$33,$3F,$E5,$4F
	.dc.b	$37,$FD,$3D,$39,$E3,$4E,$4B,$19,$70,$1C,$5F,$6B,$EB,$8F,$E3,$A7
	.dc.b	$FD,$77,$5F,$95,$6B,$F8,$07,$A4,$6E,$78,$E3,$E4,$F4,$7F,$D6,$53
	.dc.b	$CD,$B1,$E8,$9D,$DF,$99,$AF,$96,$76,$FA,$9A,$FA,$63,$77,$E2,$AF
	.dc.b	$04,$F9,$EB,$73,$98,$ED,$8E,$58,$7A,$EF,$99,$D2,$B8,$9C,$9C,$CF
	.dc.b	$0C,$3D,$95,$F1,$67,$2F,$C5,$AE,$2E,$6A,$39,$39,$61,$F0,$E3,$37
	.dc.b	$D2,$74,$79,$63,$33,$EE,$8C,$CF,$D1,$9C,$C7,$C7,$3A,$9F,$00,$FE
	.dc.b	$03,$C9,$FC,$F7,$E9,$BF,$DC,$D3,$FC,$CD,$BB,$7F,$44,$6D,$7F,$77
	.dc.b	$E7,$7C,$FE,$30,$E5,$3F,$CA,$EB,$1C,$67,$8F,$6F,$F7,$EB,$CF,$D2
	.dc.b	$CB,$FE,$64,$EB,$E9,$BF,$AC,$46,$08,$7F,$4F,$8D,$76,$71,$8C,$1A
	.dc.b	$D7,$FB,$55,$72,$8C,$92,$3F,$A8,$57,$28,$FE,$ED,$F1,$BE,$E9,$FF
	.dc.b	$62,$B7,$B8,$FA,$B6,$7D,$0E,$62,$5F,$23,$FC,$65,$1B,$E0,$F8,$3E
	.dc.b	$AB,$E9,$A9,$9A,$FF,$1C,$4E,$1F,$F6,$8C,$ED,$C9,$FE,$0B,$F8,$0F
	.dc.b	$21,$1D,$D5,$7F,$C0,$31,$73,$FF,$06,$57,$39,$FF,$37,$AF,$89,$F2
	.dc.b	$92,$3C,$3F,$2D,$5D,$C5,$AE,$73,$EC,$F1,$E5,$7A,$1F,$7F,$D0,$E2
	.dc.b	$3B,$C5,$AF,$1C,$FE,$32,$5F,$4C,$BC,$36,$3F,$8D,$D7,$F9,$21,$DC
	.dc.b	$FC,$CF,$8D,$DB,$23,$7A,$FF,$22,$76,$7C,$C9,$D3,$F0,$BE,$2B,$EA
	.dc.b	$98,$BF,$8F,$CB,$F6,$09,$3F,$53,$E9,$AB,$F2,$A2,$CF,$CE,$9F,$C5
	.dc.b	$7C,$F3,$96,$92,$FE,$EF,$C6,$57,$FC,$9D,$3B,$7D,$27,$FB,$59,$C3
	.dc.b	$F6,$23,$E4,$7D,$13,$E2,$29,$27,$FD,$33,$F4,$CF,$A4,$C4,$7F,$A8
	.dc.b	$EB,$83,$D8,$38,$7F,$D8,$1E,$F2,$B9,$46,$E7,$5A,$7D,$35,$F0,$DE
	.dc.b	$43,$BC,$7F,$9F,$86,$6F,$A6,$F8,$A4,$BF,$F6,$27,$A7,$5F,$F6,$1F
	.dc.b	$C1,$3D,$1F,$E1,$2D,$8F,$8B,$87,$4B,$CF,$74,$9F,$F6,$73,$C3,$C3
	.dc.b	$C4,$74,$8B,$5F,$E6,$FB,$FE,$37,$3D,$32,$7D,$37,$DA,$24,$7F,$B6
	.dc.b	$7A,$2B,$F6,$22,$DF,$39,$A7,$F7,$2F,$15,$F4,$7F,$B1,$9F,$F3,$F3
	.dc.b	$67,$F3,$B8,$78,$30,$F1,$7C,$B7,$E2,$88,$FE,$13,$0E,$7D,$98,$8F
	.dc.b	$C4,$CD,$C7,$AC,$FF,$30,$C5,$DD,$49,$7C,$E3,$F8,$13,$63,$2F,$FB
	.dc.b	$D2,$38,$BB,$F0,$3F,$23,$F4,$D2,$3F,$08,$FE,$53,$A3,$11,$F8,$F8
	.dc.b	$B8,$FE,$1E,$6E,$BF,$B5,$E6,$7D,$37,$E7,$0C,$E1,$FE,$C5,$C9,$E1
	.dc.b	$7F,$52,$EA,$78,$8F,$A6,$C3,$6F,$D9,$5F,$B2,$F5,$24,$7D,$D7,$75
	.dc.b	$F1,$7C,$39,$3F,$0D,$F9,$1E,$27,$F4,$3F,$A6,$7C,$7D,$25,$7C,$9F
	.dc.b	$17,$5E,$44,$8F,$CA,$D2,$37,$1F,$2F,$C7,$FA,$84,$77,$1D,$FD,$A9
	.dc.b	$1C,$9F,$B5,$E3,$E2,$3F,$DA,$96,$9C,$73,$7E,$32,$AF,$2F,$E4,$78
	.dc.b	$FF,$AC,$E1,$F7,$FC,$BC,$AB,$63,$FE,$C4,$8F,$4F,$8B,$F0,$4F,$A1
	.dc.b	$F4,$DF,$CC,$E3,$C3,$19,$76,$7F,$18,$E6,$4E,$EF,$F8,$15,$F5,$AF
	.dc.b	$C6,$7C,$BF,$E5,$4E,$4F,$FB,$BF,$E9,$78,$69,$FD,$01,$B1,$E7,$6E
	.dc.b	$2F,$A6,$72,$7F,$79,$9D,$E4,$69,$C2,$77,$7F,$C2,$A3,$1D,$DF,$8D
	.dc.b	$9F,$D1,$CF,$8F,$C0,$E9,$CF,$23,$83,$F9,$3F,$98,$FF,$31,$E3,$FC
	.dc.b	$7F,$14,$EE,$BD,$8E,$FD,$9C,$53,$5F,$F3,$BE,$AC,$7F,$CF,$D8,$7C
	.dc.b	$B6,$BC,$8F,$47,$17,$BC,$8F,$F3,$BF,$43,$D3,$87,$8E,$73,$3F,$35
	.dc.b	$FE,$DB,$B6,$27,$F8,$9F,$CC,$FF,$07,$31,$97,$F6,$9F,$A5,$7C,$A7
	.dc.b	$F7,$76,$E3,$FD,$25,$AF,$E2,$E7,$FB,$87,$D3,$3E,$67,$85,$CA,$E0
	.dc.b	$FF,$80,$67,$51,$FD,$4F,$E9,$7F,$38,$DE,$CF,$F0,$34,$62,$3B,$1F
	.dc.b	$E9,$EF,$A6,$7C,$1E,$AB,$FD,$E7,$3F,$E0,$1A,$E0,$FF,$BD,$9A,$E6
	.dc.b	$36,$FE,$37,$87,$F1,$44,$7F,$B9,$FB,$EF,$A6,$FD,$E4,$7E,$26,$74
	.dc.b	$FE,$65,$FF,$32,$46,$A3,$CF,$2F,$A6,$7C,$7F,$F9,$A5,$B3,$D9,$AF
	.dc.b	$F8,$DB,$93,$53,$23,$4F,$F9,$32,$31,$1D,$A7,$8F,$1F,$E3,$61,$7E
	.dc.b	$1B,$E1,$1D,$FE,$3B,$C4,$47,$F7,$5F,$4C,$F8,$CC,$47,$59,$B8,$75
	.dc.b	$E9,$5E,$EB,$FC,$19,$F4,$CF,$13,$EA,$C7,$D4,$8C,$71,$7D,$06,$C6
	.dc.b	$67,$6F,$F8,$E4,$FF,$6B,$7A,$8F,$A6,$FA,$FF,$16,$8C,$4F,$E0,$23
	.dc.b	$13,$A9,$CC,$F3,$BF,$F0,$67,$C6,$F8,$7D,$F3,$49,$3E,$1D,$3F,$AA
	.dc.b	$FC,$EF,$F8,$6F,$B7,$FB,$E7,$9A,$F2,$71,$6B,$DD,$7F,$9E,$A7,$31
	.dc.b	$89,$FF,$38,$C6,$3D,$56,$1C,$BD,$AF,$E9,$13,$71,$FE,$09,$8F,$FA
	.dc.b	$13,$C3,$C7,$66,$5B,$19,$3F,$D1,$31,$96,$99,$CB,$FD,$03,$19,$7F
	.dc.b	$56,$3E,$97,$FD,$09,$3F,$E2,$EF,$A7,$FE,$C2,$FF,$91,$E3,$33,$FB
	.dc.b	$2E,$75,$6A,$39,$CE,$9F,$B4,$D7,$FC,$E8,$D4,$78,$7C,$6F,$F3,$F3
	.dc.b	$F3,$BD,$8F,$9B,$DA,$F4,$BE,$1C,$FF,$A6,$A7,$31,$8C,$7F,$D3,$F1
	.dc.b	$8C,$FE,$53,$73,$FF,$30,$7C,$8B,$E6,$BE,$77,$C8,$7D,$AC,$93,$9A
	.dc.b	$FF,$59,$CE,$63,$13,$FE,$AA,$8C,$57,$2A,$FF,$2A,$E9,$FE,$81,$7F
	.dc.b	$D7,$7E,$4F,$E1,$B3,$FF,$58,$7C,$4F,$DD,$8C,$56,$F1,$FF,$5A,$C6
	.dc.b	$27,$57,$D1,$B3,$F4,$6B,$85,$FF,$AF,$AF,$9F,$B5,$F8,$9E,$6B,$8F
	.dc.b	$FB,$31,$B8,$7F,$B0,$9B,$7F,$EC,$67,$31,$FE,$A9,$8F,$F6,$35,$7D
	.dc.b	$48,$FF,$04,$61,$FE,$31,$BF,$E5,$F5,$A6,$C3,$E3,$3F,$35,$F3,$BE
	.dc.b	$E6,$CF,$C6,$6D,$FA,$DB,$7F,$E8,$9B,$E5,$5B,$3F,$DC,$BD,$7C,$DF
	.dc.b	$9F,$F1,$7F,$A7,$C3,$E5,$B6,$32,$7F,$BB,$23,$2D,$32,$FC,$B8,$FD
	.dc.b	$97,$E6,$0F,$AC,$DC,$FC,$FF,$17,$3C,$1F,$61,$B8,$7E,$01,$B3,$B8
	.dc.b	$FF,$3C,$5F,$D6,$7F,$DB,$79,$7F,$58,$78,$1A,$FA,$FE,$8F,$0F,$AC
	.dc.b	$F1,$9F,$89,$FD,$DD,$F4,$38,$9F,$F5,$1C,$61,$E0,$62,$35,$8F,$F6
	.dc.b	$8F,$05,$97,$FA,$CE,$B4,$79,$BE,$3F,$B5,$FC,$2F,$36,$4F,$AB,$96
	.dc.b	$E3,$FC,$07,$2F,$EB,$73,$93,$C7,$E7,$7C,$AA,$F9,$7F,$2B,$27,$E1
	.dc.b	$F6,$5C,$3F,$88,$F3,$3E,$63,$E8,$FC,$77,$EC,$E3,$17,$F2,$A3,$13
	.dc.b	$A8,$FF,$6D,$4E,$AB,$DD,$BF,$FA,$EA,$73,$5B,$9E,$ED,$AF,$EC,$DC
	.dc.b	$1F,$13,$CC,$FC,$0F,$57,$E0,$9B,$87,$DA,$6E,$1D,$9B,$7C,$B1,$7B
	.dc.b	$9F,$8A,$CB,$F1,$6D,$C3,$F0,$6D,$C3,$CE,$DB,$E5,$F2,$5F,$92,$EB
	.dc.b	$FC,$9B,$6F,$F1,$6D,$BF,$7B,$6B,$8C,$FE,$42,$B8,$B4,$FF,$74,$FE
	.dc.b	$EE,$F9,$5F,$CF,$FC,$77,$F1,$58,$C4,$FE,$62,$31,$3A,$8E,$2D,$C7
	.dc.b	$E7,$F9,$DF,$A8,$F8,$FF,$D4,$B6,$BF,$42,$D9,$DC,$7E,$92,$7F,$5F
	.dc.b	$DD,$FB,$0F,$8D,$B3,$FA,$86,$D7,$DC,$6E,$9F,$3D,$B3,$FE,$EF,$E0
	.dc.b	$46,$29,$C7,$FA,$08,$FE,$DD,$F3,$7F,$E5,$5B,$FF,$76,$2F,$A9,$CF
	.dc.b	$C1,$69,$B3,$A6,$E1,$FA,$BA,$E1,$3C,$BE,$A3,$AF,$C8,$76,$6C,$64
	.dc.b	$EF,$19,$69,$E2,$F0,$3C,$BE,$A7,$99,$B1,$93,$CF,$19,$69,$9C,$3C
	.dc.b	$91,$96,$C7,$BB,$CC,$F8,$3F,$49,$F0,$9B,$19,$3E,$1C,$65,$A7,$C6
	.dc.b	$FA,$2F,$93,$F4,$7F,$29,$B1,$93,$E5,$C6,$5A,$7C,$EF,$84,$FE,$F7
	.dc.b	$F3,$DF,$E4,$D8,$C1,$FE,$6E,$71,$8F,$CD,$9D,$F8,$FF,$61,$B5,$F6
	.dc.b	$3E,$0B,$ED,$9F,$73,$C7,$B1,$92,$5F,$D7,$8C,$BF,$67,$E8,$7F,$07
	.dc.b	$F4,$7F,$84,$6D,$7E,$11,$B3,$B8,$FC,$3C,$FE,$2B,$DF,$FC,$5F,$D1
	.dc.b	$D9,$FC,$5B,$6B,$F0,$CD,$EB,$FE,$1F,$CC,$8C,$57,$F8,$8B,$A9,$2C
	.dc.b	$3F,$2F,$F0,$9A,$C7,$F3,$73,$9B,$FC,$CD,$69,$A7,$E7,$3E,$77,$F3
	.dc.b	$FF,$3D,$FA,$06,$C6,$4F,$D0,$C6,$5A,$7F,$8D,$33,$FF,$00,$7D,$1F
	.dc.b	$F8,$DB,$E8,$DE,$E7,$F9,$D7,$D9,$FF,$47,$46,$34,$FF,$3C,$56,$EB
	.dc.b	$58,$59,$C8,$AE,$0F,$F9,$0A,$F9,$D7,$FA,$EA,$F8,$4E,$6E,$BF,$E4
	.dc.b	$88,$B5,$C2,$3A,$DF,$47,$FC,$FB,$5C,$15,$70,$8C,$36,$3C,$97,$E1
	.dc.b	$69,$FE,$56,$39,$98,$31,$1C,$6B,$C1,$1D,$8D,$1C,$1C,$C6,$E7,$B4
	.dc.b	$79,$EF,$D5,$5D,$0E,$2E,$1F,$80,$62,$FD,$2D,$3B,$9E,$19,$E9,$1A
	.dc.b	$6C,$6A,$FC,$4F,$C8,$BF,$AE,$E4,$E9,$3E,$86,$C7,$A0,$DC,$64,$EA
	.dc.b	$6F,$C3,$FE,$67,$3F,$80,$6B,$ED,$6B,$EC,$73,$F1,$1F,$E0,$3C,$BF
	.dc.b	$D0,$B1,$8F,$35,$18,$F1,$3E,$A5,$6E,$D7,$BA,$D4,$7F,$A3,$A7,$58
	.dc.b	$7E,$41,$FE,$EF,$1F,$8E,$6B,$F7,$5D,$E3,$DB,$1F,$0C,$E7,$CD,$FC
	.dc.b	$25,$F3,$6D,$6F,$0E,$97,$C2,$B5,$7D,$EF,$6F,$96,$75,$7C,$63,$B1
	.dc.b	$B3,$9D,$C7,$24,$BF,$65,$72,$8E,$93,$C5,$D5,$75,$AD,$CE,$5F,$2C
	.dc.b	$FC,$1A,$E4,$B0,$F8,$6F,$CB,$C3,$D2,$E5,$B1,$A8,$E1,$3C,$A3,$E6
	.dc.b	$D7,$63,$C3,$3D,$EF,$C5,$87,$DC,$C3,$9E,$1C,$4E,$4F,$63,$33,$F0
	.dc.b	$EF,$A5,$71,$BC,$3F,$CB,$7E,$8F,$FC,$41,$1F,$E0,$0D,$7F,$D3,$C7
	.dc.b	$F3,$1E,$23,$88,$FF,$04,$CF,$07,$FC,$D4,$DC,$3F,$9D,$CE,$F9,$3F
	.dc.b	$9C,$B6,$74,$F8,$7E,$7F,$04,$BF,$E5,$D1,$E5,$F9,$6C,$47,$69,$B8
	.dc.b	$7F,$95,$AD,$58,$BF,$67,$B7,$DC,$C6,$1E,$8C,$30,$F6,$3D,$71,$AC
	.dc.b	$58,$3E,$63,$D1,$F3,$D2,$F8,$56,$A7,$D8,$66,$3E,$1C,$E1,$DD,$3E
	.dc.b	$7B,$E1,$12,$33,$F3,$15,$F9,$71,$75,$F7,$4F,$C9,$F9,$EF,$A5,$F1
	.dc.b	$2F,$17,$FD,$99,$87,$F8,$1F,$D9,$D3,$E0,$AD,$CE,$A3,$F8,$3F,$01
	.dc.b	$5F,$3B,$E3,$1F,$73,$E4,$FE,$C7,$C5,$48,$F5,$C7,$DB,$30,$FE,$0A
	.dc.b	$30,$FC,$0F,$8D,$FC,$07,$C7,$48,$C1,$0F,$C4,$B9,$7F,$17,$1F,$87
	.dc.b	$F8,$BF,$C4,$7C,$64,$9F,$97,$19,$7E,$6C,$65,$FC,$9C,$65,$FC,$AC
	.dc.b	$67,$7F,$C4,$4F,$E4,$FE,$33,$F2,$9D,$52,$FE,$BC,$65,$FB,$31,$97
	.dc.b	$F3,$F1,$F9,$BF,$A6,$FC,$E7,$C6,$C9,$FB,$F1,$97,$F0,$31,$97,$F4
	.dc.b	$F1,$FA,$3F,$13,$F4,$9F,$1B,$27,$F0,$F1,$97,$F1,$31,$97,$F5,$F1
	.dc.b	$97,$F6,$31,$9C,$BF,$49,$3F,$B1,$F8,$9F,$F7,$C5,$7E,$48,$FD,$91
	.dc.b	$9F,$8D,$8E,$48,$7F,$B9,$63,$8B,$F8,$83,$FA,$7F,$C1,$3A,$3B,$7C
	.dc.b	$6C,$9F,$F5,$5C,$7F,$07,$C5,$5B,$9F,$D4,$DA,$F8,$47,$5A,$DD,$FA
	.dc.b	$7E,$33,$D5,$F1,$B2,$B8,$C6,$5E,$71,$97,$D1,$19,$7D,$51,$9A,$FF
	.dc.b	$6E,$E0,$AF,$8D,$EB,$7C,$78,$97,$E6,$8C,$BE,$88,$CB,$F1,$63,$E7
	.dc.b	$7C,$57,$CF,$F8,$C9,$3E,$E8,$CB,$F0,$23,$2F,$CD,$8F,$E5,$8F,$F8
	.dc.b	$D7,$0F,$2F,$C6,$7D,$93,$3C,$9F,$E2,$F9,$FA,$46,$FB,$7F,$CC,$D9
	.dc.b	$2B,$84,$FF,$1D,$34,$7E,$FE,$B9,$5F,$68,$FA,$73,$D9,$FC,$47,$64
	.dc.b	$7E,$BB,$F8,$BF,$8C,$FB,$9C,$2E,$7F,$C7,$7C,$64,$BF,$F7,$5C,$7E
	.dc.b	$47,$CD,$91,$F8,$07,$4F,$E3,$7C,$DF,$C3,$7E,$ED,$F8,$ED,$3F,$A7
	.dc.b	$CF,$F8,$72,$31,$BF,$BA,$B7,$CF,$F9,$DF,$8C,$97,$FC,$EF,$06,$1C
	.dc.b	$23,$FD,$BF,$C2,$AD,$C7,$E7,$5E,$6F,$19,$71,$F9,$09,$1F,$98,$F8
	.dc.b	$DA,$FE,$97,$E3,$7F,$5C,$63,$1F,$49,$E2,$AC,$CF,$E8,$27,$5A,$7F
	.dc.b	$A2,$F0,$6B,$F0,$89,$5F,$A6,$F8,$BA,$FE,$CF,$E3,$3F,$78,$63,$2E
	.dc.b	$F7,$CA,$7F,$5B,$3A,$7D,$97,$F3,$AE,$47,$BD,$FE,$0D,$EB,$62,$3F
	.dc.b	$80,$7A,$DF,$C2,$4C,$63,$DA,$B1,$3F,$BA,$9C,$E5,$FE,$A5,$F8,$87
	.dc.b	$C5,$FE,$5E,$B0,$D7,$FC,$AF,$1A,$75,$82,$31,$1F,$96,$6C,$7E,$13
	.dc.b	$D2,$7E,$F1,$8C,$3F,$D8,$CD,$BE,$B8,$7F,$11,$C3,$17,$C2,$3F,$C2
	.dc.b	$95,$BC,$3C,$11,$BF,$89,$CD,$CC,$3F,$18,$A3,$51,$FE,$02,$BF,$43
	.dc.b	$67,$2F,$F1,$B7,$83,$92,$19,$9F,$E0,$3D,$DA,$8F,$E5,$D8,$FB,$BE
	.dc.b	$2F,$CE,$6A,$A4,$F8,$63,$2F,$EF,$EB,$96,$5F,$1F,$1B,$9F,$F7,$FF
	.dc.b	$3D,$F8,$57,$E9,$5B,$D3,$F1,$33,$99,$7F,$89,$67,$2D,$AF,$D6,$69
	.dc.b	$F4,$AB,$F8,$CF,$CC,$BE,$23,$EF,$5A,$30,$FE,$05,$AF,$E1,$5A,$FF
	.dc.b	$1D,$73,$A7,$F0,$28,$7C,$D3,$9B,$F9,$3B,$63,$F8,$F5,$70,$D7,$F0
	.dc.b	$7F,$19,$F7,$BE,$3A,$57,$D7,$8F,$C0,$98,$7E,$7F,$02,$73,$1B,$3E
	.dc.b	$F3,$BC,$BE,$65,$7F,$98,$FE,$55,$F3,$0F,$2F,$F1,$BF,$21,$31,$FF
	.dc.b	$27,$56,$A7,$85,$6E,$7F,$15,$5B,$7F,$B7,$56,$DF,$E7,$5F,$2E,$F8
	.dc.b	$AF,$F5,$DC,$7E,$B6,$31,$D7,$FA,$D6,$C6,$67,$6F,$F4,$9F,$86,$79
	.dc.b	$0F,$98,$FD,$A4,$61,$B1,$99,$FD,$D4,$E5,$F0,$7C,$5B,$D4,$F0,$36
	.dc.b	$33,$18,$3F,$81,$67,$F6,$7D,$7F,$E4,$BC,$9F,$E5,$1E,$AF,$D6,$7B
	.dc.b	$78,$C1,$7B,$AF,$F3,$7F,$5A,$B5,$39,$AF,$D3,$CE,$6B,$E7,$F5,$31
	.dc.b	$E9,$8F,$69,$E7,$ED,$62,$DE,$DD,$8C,$36,$7C,$2D,$C7,$BB,$7A,$3F
	.dc.b	$8A,$5F,$2A,$EA,$70,$3F,$B9,$7C,$BB,$E1,$7F,$39,$87,$F9,$C4,$46
	.dc.b	$67,$F9,$33,$63,$F9,$53,$5F,$F0,$13,$9B,$FD,$33,$67,$F7,$DF,$61
	.dc.b	$FA,$77,$C3,$7E,$FF,$7B,$E7,$78,$6F,$7D,$AE,$6F,$F8,$04,$61,$FA
	.dc.b	$35,$C2,$31,$9F,$D0,$8C,$57,$D9,$87,$8C,$EC,$F9,$DE,$CE,$63,$18
	.dc.b	$FD,$C8,$C4,$EA,$3F,$8D,$36,$BF,$C8,$99,$FF,$91,$E7,$F0,$7E,$AF
	.dc.b	$E1,$F7,$F9,$9E,$EF,$DB,$8C,$56,$EF,$F0,$B1,$89,$D5,$FF,$81,$EB
	.dc.b	$57,$F6,$AB,$53,$D8,$D7,$A5,$F8,$7E,$5E,$AD,$8C,$9F,$96,$8C,$B4
	.dc.b	$CE,$3E,$A9,$FC,$87,$BF,$A8,$F6,$FB,$5F,$93,$F7,$FD,$2D,$BF,$33
	.dc.b	$67,$71,$FA,$49,$DB,$A3,$F4,$FD,$98,$9F,$93,$EB,$FE,$7D,$F8,$AD
	.dc.b	$9F,$84,$D8,$DC,$7E,$C1,$DB,$A3,$FD,$C3,$F2,$CF,$3F,$33,$F5,$BC
	.dc.b	$6F,$D7,$CF,$CC,$AF,$F6,$76,$FF,$A1,$9C,$F0,$FD,$66,$CF,$D3,$6D
	.dc.b	$F3,$8C,$1E,$0F,$23,$83,$E2,$F7,$FB,$9F,$83,$F9,$8F,$E7,$51,$93
	.dc.b	$FC,$88,$E2,$BF,$B6,$E5,$FE,$81,$9C,$F9,$9C,$9F,$6F,$B7,$CE,$3D
	.dc.b	$6D,$8F,$83,$F0,$1E,$E3,$DF,$C9,$63,$BF,$C9,$E1,$A7,$C8,$8F,$8B
	.dc.b	$F0,$32,$3E,$5F,$65,$CF,$FC,$FB,$C3,$FE,$44,$9F,$9F,$F0,$3F,$AD
	.dc.b	$8C,$4F,$AA,$31,$EE,$7D,$06,$D7,$03,$FD,$05,$5C,$0F,$F1,$56,$C9
	.dc.b	$DD,$F3,$AF,$BB,$E6,$FD,$AE,$9F,$B6,$D8,$C9,$F8,$28,$CB,$4F,$C2
	.dc.b	$FC,$07,$E0,$B2,$FC,$1B,$63,$27,$E2,$A3,$2D,$3F,$1B,$F0,$1F,$8A
	.dc.b	$E9,$FC,$5B,$63,$27,$E4,$A3,$2D,$33,$8F,$DE,$8C,$B6,$3F,$D0,$DF
	.dc.b	$02,$F9,$8F,$CB,$FC,$17,$F0,$B8,$C4,$FE,$7A,$31,$3A,$8E,$2D,$AF
	.dc.b	$C5,$F9,$DF,$A0,$F8,$0F,$D0,$B6,$BF,$4E,$D9,$DC,$7E,$AA,$7F,$4F
	.dc.b	$D9,$FA,$8F,$81,$B3,$FA,$E6,$D7,$E6,$1B,$C7,$F6,$9B,$3F,$DA,$F8
	.dc.b	$D1,$8A,$7A,$09,$CD,$7E,$D0,$FE,$E5,$F1,$BF,$E7,$87,$15,$C2,$1F
	.dc.b	$31,$FD,$A9,$D4,$7F,$95,$9C,$CF,$0B,$46,$B1,$FF,$3C,$3A,$AE,$1F
	.dc.b	$07,$CF,$0E,$9F,$05,$D7,$E0,$EE,$5C,$F0,$FF,$42,$36,$74,$D3,$53
	.dc.b	$E5,$F0,$3C,$DE,$A5,$9E,$4D,$CB,$D0,$DC,$FC,$AD,$9F,$86,$D9,$F8
	.dc.b	$4F,$C1,$F3,$7E,$17,$BD,$6B,$CC,$DA,$F8,$8D,$BF,$82,$FC,$9F,$98
	.dc.b	$F9,$5F,$31,$6F,$DC,$DB,$F9,$8D,$BF,$92,$FD,$1F,$84,$FE,$F3,$F1
	.dc.b	$2F,$1B,$18,$7F,$9D,$E3,$0F,$C9,$9D,$CE,$B0,$61,$FE,$EE,$F5,$D8
	.dc.b	$70,$9F,$B9,$F3,$1F,$7C,$FC,$07,$91,$72,$FA,$91,$67,$F0,$53,$70
	.dc.b	$FB,$FF,$31,$F8,$7F,$98,$FC,$43,$70,$FB,$8D,$BF,$C5,$36,$BF,$0F
	.dc.b	$F0,$1F,$8F,$F9,$8F,$C8,$36,$FF,$08,$DB,$FC,$93,$6B,$8C,$E8,$FC
	.dc.b	$57,$C2,$72,$71,$5C,$1C,$98,$9F,$CB,$FC,$C7,$E7,$FE,$63,$F4,$0D
	.dc.b	$9F,$CA,$36,$BF,$44,$DA,$FF,$1E,$C3,$E6,$0B,$E6,$1E,$F7,$F9,$06
	.dc.b	$3F,$CD,$58,$6A,$73,$19,$AC,$56,$FD,$95,$E2,$B4,$75,$9D,$46,$63
	.dc.b	$FC,$9D,$55,$EE,$6C,$E3,$7E,$07,$CF,$7D,$9D,$1C,$8A,$72,$8E,$0F
	.dc.b	$BA,$7B,$4F,$19,$CC,$FC,$33,$8C,$66,$FE,$14,$66,$79,$CE,$63,$CB
	.dc.b	$3E,$C7,$C2,$6D,$C4,$F6,$AF,$0D,$FB,$EB,$47,$F9,$BE,$FA,$3E,$BA
	.dc.b	$EF,$3B,$AE,$33,$D9,$B3,$D6,$3D,$33,$B6,$9B,$2C,$6C,$EA,$E5,$A6
	.dc.b	$6F,$D2,$7C,$2A,$EE,$D9,$E6,$DF,$13,$F9,$EC,$7E,$01,$AF,$D5,$6B
	.dc.b	$F4,$C3,$AF,$87,$79,$BF,$C0,$A7,$3E,$37,$FA,$4E,$73,$A7,$E6,$2F
	.dc.b	$85,$ED,$B1,$9A,$E1,$3C,$F2,$FC,$55,$BC,$EF,$F4,$F4,$61,$B9,$70
	.dc.b	$C3,$F2,$D5,$E0,$CB,$F1,$EF,$17,$35,$C1,$FD,$5C,$6E,$FB,$E1,$E1
	.dc.b	$C3,$D5,$7B,$B5,$F3,$7A,$D6,$DE,$6D,$77,$3C,$67,$85,$F6,$C3,$CF
	.dc.b	$7B,$9C,$E1,$E0,$C3,$35,$1F,$5E,$1C,$CE,$97,$C4,$E4,$E6,$38,$61
	.dc.b	$DD,$E2,$7C,$23,$37,$F2,$5C,$92,$38,$12,$3C,$91,$97,$84,$66,$3E
	.dc.b	$5C,$F8,$5F,$8B,$7D,$9C,$9E,$13,$27,$F0,$9F,$19,$F2,$9F,$E2,$B8
	.dc.b	$FE,$F1,$B7,$FB,$76,$BF,$C0,$2D,$3F,$E0,$04,$EE,$7F,$A9,$C6,$3C
	.dc.b	$BD,$47,$F8,$46,$2D,$FF,$83,$1B,$5C,$6F,$F8,$B4,$F8,$BE,$5F,$24
	.dc.b	$BF,$F0,$A5,$3E,$63,$11,$E0,$3F,$C4,$B0,$9C,$DF,$48,$E1,$3C,$A7
	.dc.b	$C3,$F0,$52,$30,$43,$A6,$1F,$12,$35,$18,$AE,$CF,$93,$E6,$3D,$7F
	.dc.b	$31,$23,$04,$38,$39,$79,$47,$C2,$F9,$89,$1E,$E7,$E2,$FC,$B6,$23
	.dc.b	$C1,$17,$3F,$64,$FC,$AF,$98,$F9,$FE,$1F,$F8,$0B,$D8,$BE,$21,$D6
	.dc.b	$3F,$2A,$B3,$12,$30,$7C,$EF,$98,$FA,$BF,$17,$23,$04,$3E,$13,$97
	.dc.b	$E2,$47,$DC,$F8,$EF,$BB,$F2,$12,$7E,$CC,$65,$FB,$71,$97,$E6,$47
	.dc.b	$E1,$3E,$37,$F0,$BF,$1B,$27,$F0,$31,$97,$F0,$71,$97,$EA,$46,$5F
	.dc.b	$AD,$19,$DB,$F0,$53,$F9,$0F,$8D,$FC,$8F,$64,$BF,$C6,$C6,$5F,$C7
	.dc.b	$C6,$5F,$C0,$47,$E6,$3E,$33,$F3,$3F,$1B,$27,$F2,$B1,$97,$F2,$F1
	.dc.b	$97,$F1,$11,$FA,$0F,$13,$F4,$3F,$1B,$27,$F3,$B1,$97,$F3,$F1,$97
	.dc.b	$F2,$11,$97,$F2,$51,$9C,$BF,$3D,$3F,$AE,$F8,$8F,$F7,$E5,$7E,$A0
	.dc.b	$FD,$79,$9F,$8D,$8E,$48,$7E,$82,$38,$BF,$AD,$A7,$C1,$39,$FB,$7C
	.dc.b	$6C,$AF,$F5,$9C,$7F,$00,$AD,$B9,$8E,$93,$9A,$E5,$1F,$E8,$FB,$7C
	.dc.b	$67,$A7,$E3,$64,$E6,$3D,$86,$1F,$34,$61,$D1,$CD,$D6,$3F,$C3,$2B
	.dc.b	$AC,$F7,$F5,$7E,$34,$48,$C1,$0F,$84,$E5,$F8,$91,$F3,$7E,$2B,$E7
	.dc.b	$7C,$64,$9F,$54,$65,$F6,$46,$5F,$99,$1F,$C9,$1F,$F1,$7D,$F9,$7E
	.dc.b	$33,$EC,$19,$E4,$FF,$26,$47,$F8,$E6,$77,$E3,$FF,$94,$BC,$56,$6C
	.dc.b	$38,$D7,$FB,$00,$D9,$FE,$58,$C3,$9D,$F1,$AE,$13,$F4,$6B,$C0,$FE
	.dc.b	$27,$B6,$3F,$55,$FC,$67,$C6,$7D,$9E,$8F,$C7,$7C,$64,$AF,$F7,$9C
	.dc.b	$7E,$47,$CD,$91,$F7,$5D,$3F,$8D,$F3,$3F,$0D,$F3,$1F,$8E,$DF,$FC
	.dc.b	$AC,$1C,$6B,$E7,$9D,$F4,$7E,$77,$E3,$26,$1F,$ED,$EC,$38,$4E,$23
	.dc.b	$F3,$53,$67,$AE,$1D,$27,$FD,$52,$B8,$FC,$64,$8F,$CA,$7C,$6D,$7F
	.dc.b	$49,$F1,$BF,$AD,$31,$7D,$2B,$13,$F9,$B9,$CE,$7F,$E8,$8A,$F6,$69
	.dc.b	$F0,$09,$5F,$A1,$F8,$CA,$FE,$C7,$E3,$7F,$74,$62,$BD,$35,$89,$FD
	.dc.b	$44,$E6,$FB,$66,$C3,$E6,$92,$BF,$5D,$EA,$D7,$F7,$DE,$B7,$F0,$73
	.dc.b	$19,$7C,$6A,$C4,$FE,$D6,$73,$87,$AF,$E2,$1F,$19,$FB,$F6,$D7,$D3
	.dc.b	$6B,$FD,$AB,$E5,$5F,$1B,$BA,$FF,$01,$C6,$B2,$FF,$00,$B7,$0E,$D3
	.dc.b	$BB,$F9,$51,$FA,$19,$D4,$3B,$11,$8B,$63,$47,$C5,$78,$19,$7E,$78
	.dc.b	$F5,$18,$7D,$11,$87,$47,$88,$F0,$57,$67,$17,$E2,$9E,$11,$FE,$15
	.dc.b	$1D,$DE,$27,$74,$F8,$CF,$86,$7F,$05,$9D,$4F,$16,$E3,$F1,$A7,$73
	.dc.b	$FB,$F9,$CE,$5F,$17,$2F,$F0,$1F,$CC,$7E,$7F,$D3,$3F,$A1,$D3,$E3
	.dc.b	$CF,$D2,$37,$D3,$31,$F2,$C7,$F1,$7F,$97,$7C,$4D,$24,$FF,$0F,$F9
	.dc.b	$8F,$F4,$AB,$FC,$46,$76,$FF,$B1,$03,$87,$FD,$8E,$D7,$FD,$D4,$D7
	.dc.b	$EF,$7C,$6F,$E1,$BE,$3E,$47,$C8,$8F,$C4,$98,$7F,$1D,$18,$74,$7C
	.dc.b	$E7,$5C,$1F,$8E,$9F,$F2,$AF,$CA,$BE,$61,$E5,$7E,$53,$E4,$66,$3F
	.dc.b	$7E,$B5,$3C,$2B,$73,$F6,$AB,$6F,$F9,$4E,$B6,$FF,$33,$F9,$87,$C5
	.dc.b	$7C,$68,$10,$46,$BA,$FE,$43,$63,$FD,$DA,$D7,$81,$FE,$A4,$F1,$DF
	.dc.b	$31,$F3,$FE,$21,$18,$9F,$F7,$AC,$62,$75,$39,$9E,$74,$F8,$C7,$AB
	.dc.b	$E2,$6C,$F8,$9A,$F0,$6B,$EB,$FC,$35,$F2,$9F,$CF,$3B,$3D,$F6,$F8
	.dc.b	$2F,$E2,$D1,$FD,$5E,$B6,$FE,$B6,$75,$E2,$71,$C3,$E3,$D6,$9F,$82
	.dc.b	$1C,$3F,$11,$AF,$8F,$0F,$ED,$4F,$CF,$F9,$97,$C4,$75,$E7,$7C,$D7
	.dc.b	$F0,$8C,$FF,$B5,$56,$9F,$E9,$D1,$9B,$FD,$E4,$E5,$FF,$13,$46,$6B
	.dc.b	$FB,$C4,$E5,$FE,$BB,$19,$AF,$15,$F9,$2F,$F0,$77,$CD,$FB,$AD,$77
	.dc.b	$3F,$03,$BF,$F8,$FF,$CA,$7C,$56,$C7,$F0,$CE,$EC,$78,$AA,$A4,$9D
	.dc.b	$1F,$C4,$9D,$FC,$A7,$0F,$95,$B3,$FC,$59,$B8,$FF,$3B,$6E,$BF,$CE
	.dc.b	$1F,$AD,$1F,$C7,$9F,$0F,$CB,$63,$C7,$46,$2F,$F8,$C3,$6B,$F8,$F3
	.dc.b	$6F,$FC,$41,$4F,$94,$F7,$7C,$67,$F3,$AC,$3F,$90,$46,$3C,$6E,$CD
	.dc.b	$8C,$9F,$CD,$27,$86,$3F,$CA,$E7,$E8,$3F,$CD,$1B,$7D,$EB,$F8,$C3
	.dc.b	$67,$F5,$FE,$4B,$F4,$2F,$DD,$D7,$F3,$AF,$03,$F2,$D3,$97,$E8,$98
	.dc.b	$9D,$D7,$F6,$38,$C1,$AA,$FE,$99,$5C,$5B,$5C,$33,$FE,$CB,$D7,$C2
	.dc.b	$77,$F0,$4F,$92,$F8,$BE,$27,$8A,$31,$3F,$06,$31,$3A,$8E,$2D,$CF
	.dc.b	$FD,$D5,$97,$F9,$F6,$37,$3F,$27,$86,$D6,$B0,$6F,$F8,$ED,$FC,$0D
	.dc.b	$BF,$2B,$6B,$83,$74,$4F,$F2,$2C,$BF,$D1,$93,$CA,$3F,$B2,$FC,$AD
	.dc.b	$BF,$07,$C6,$7E,$0B,$8F,$DC,$DA,$F8,$6D,$9D,$C7,$E0,$67,$6E,$8F
	.dc.b	$C2,$72,$62,$7F,$C2,$D1,$FA,$83,$E5,$7C,$0D,$7E,$7B,$63,$E1,$37
	.dc.b	$3F,$5C,$7F,$84,$FF,$57,$FA,$2E,$AB,$FC,$0B,$18,$9D,$E9,$FE,$AB
	.dc.b	$BF,$DD,$4B,$E5,$29,$FA,$0E,$3B,$87,$E0,$3C,$2C,$34,$FC,$F4,$7E
	.dc.b	$9B,$82,$47,$83,$BA,$E9,$F7,$BF,$68,$FF,$57,$CF,$93,$E5,$7F,$33
	.dc.b	$18,$96,$DE,$CE,$F6,$AB,$DF,$C3,$FA,$1B,$8F,$33,$F3,$78,$F3,$C3
	.dc.b	$95,$F1,$9C,$DF,$17,$E2,$F3,$7C,$2E,$6F,$86,$D8,$C9,$F2,$A3,$2D
	.dc.b	$3E,$6F,$C9,$7C,$AF,$93,$F9,$6D,$8C,$9F,$4A,$32,$D3,$EA,$F3,$FD
	.dc.b	$2F,$93,$FA,$6D,$8C,$9F,$6A,$32,$D3,$3B,$FC,$87,$FD,$6D,$9A,$77
	.dc.b	$39,$3F,$0C,$FD,$0F,$82,$FC,$0F,$C9,$DB,$FC,$3B,$6B,$87,$0F,$D2
	.dc.b	$8C,$6F,$F3,$1F,$C4,$FC,$8F,$E2,$BE,$4A,$E1,$F9,$06,$E1,$CB,$34
	.dc.b	$63,$0F,$AA,$FF,$3A,$97,$C8,$FF,$4D,$FD,$73,$EC,$47,$F9,$6C,$CE
	.dc.b	$1F,$E6,$08,$CD,$FF,$6E,$AE,$1C,$BC,$DB,$5F,$ED,$5B,$E2,$7B,$2F
	.dc.b	$8C,$7F,$87,$27,$73,$9E,$0F,$17,$1F,$8D,$B5,$E1,$6C,$EE,$3C,$93
	.dc.b	$FC,$D3,$F6,$8F,$D6,$47,$9F,$8B,$1D,$7E,$D6,$C6,$4F,$E6,$73,$C2
	.dc.b	$FC,$D3,$F1,$57,$CC,$FF,$8E,$BE,$65,$DD,$FE,$F5,$D7,$E6,$E3,$EF
	.dc.b	$BD,$8A,$DB,$FE,$99,$9D,$BF,$34,$DC,$6B,$91,$3A,$E0,$FF,$04,$63
	.dc.b	$E1,$C7,$C1,$C3,$E5,$8C,$63,$D6,$30,$F8,$67,$79,$79,$6B,$73,$FE
	.dc.b	$36,$8A,$FF,$95,$A2,$F7,$7F,$1E,$CF,$D3,$3F,$E5,$99,$CD,$70,$8F
	.dc.b	$4D,$FF,$95,$AF,$83,$FE,$38,$AD,$CF,$8C,$FF,$2D,$B8,$BE,$B1,$E1
	.dc.b	$34,$79,$4A,$65,$DD,$76,$8E,$66,$9C,$3E,$73,$17,$E7,$AF,$05,$F9
	.dc.b	$67,$27,$63,$D6,$7F,$9D,$EB,$9B,$E1,$8F,$75,$73,$8E,$37,$EA,$78
	.dc.b	$9E,$97,$17,$E0,$8F,$3C,$7A,$A7,$93,$63,$C5,$3B,$7D,$51,$93,$9B
	.dc.b	$E0,$AF,$54,$65,$A7,$9A,$39,$9B,$3C,$2E,$63,$99,$C6,$32,$7C,$1A
	.dc.b	$F2,$CE,$CE,$B1,$93,$AC,$F1,$7C,$F1,$B7,$99,$F0,$DE,$2D,$32,$7F
	.dc.b	$A4,$A7,$BB,$D6,$7C,$B3,$CE,$3A,$99,$79,$BC,$4F,$1B,$B7,$EC,$1B
	.dc.b	$3B,$DF,$DC,$7F,$6F,$F0,$BF,$44,$F9,$A5,$9E,$B3,$BC,$3E,$A5,$74
	.dc.b	$3D,$47,$90,$A6,$4F,$25,$FA,$9F,$11,$B3,$35,$DB,$0F,$83,$1C,$8F
	.dc.b	$C5,$9A,$3D,$2E,$9C,$1F,$9F,$F6,$18,$FE,$75,$B1,$FA,$66,$BB,$8F
	.dc.b	$F6,$A7,$B0,$F8,$8F,$F0,$1C,$AB,$73,$9D,$7F,$DB,$71,$8A,$DB,$63
	.dc.b	$37,$D2,$79,$E5,$FC,$2A,$DF,$19,$FE,$E3,$8C,$37,$3E,$18,$7F,$16
	.dc.b	$AF,$06,$5F,$C1,$1E,$2E,$6B,$83,$FE,$EE,$8D,$DF,$7C,$3C,$38,$7B
	.dc.b	$AF,$76,$BE,$6F,$5A,$DB,$CD,$AE,$E7,$8C,$F0,$BE,$D8,$79,$EF,$73
	.dc.b	$9C,$3C,$18,$66,$A3,$EB,$C3,$99,$D2,$F8,$9F,$4B,$1F,$8B,$23,$EE
	.dc.b	$DF,$46,$9E,$7B,$D4,$F9,$23,$B1,$F1,$CF,$0E,$39,$3A,$BA,$3E,$14
	.dc.b	$EA,$BB,$4E,$8E,$0F,$D6,$9E,$11,$98,$C3,$A8,$FA,$98,$7B,$9E,$87
	.dc.b	$9C,$E3,$3D,$8E,$24,$78,$9E,$93,$CA,$49,$F0,$3A,$31,$1C,$5F,$B0
	.dc.b	$71,$8F,$23,$F8,$1C,$3E,$BD,$79,$63,$89,$E5,$9E,$AF,$23,$27,$1B
	.dc.b	$E9,$7E,$98,$E0,$F8,$71,$FC,$14,$79,$6B,$C3,$1F,$96,$9E,$13,$D6
	.dc.b	$73,$87,$92,$3A,$9D,$08,$6E,$78,$D7,$38,$CB,$E0,$74,$62,$7A,$3F
	.dc.b	$40,$C5,$F6,$76,$7B,$4F,$7C,$F3,$24,$FE,$5A,$78,$46,$1F,$BB,$C1
	.dc.b	$FE,$1A,$E9,$FC,$C5,$FF,$90,$5B,$3F,$EA,$A8,$ED,$8F,$E1,$EB,$84
	.dc.b	$7F,$93,$EB,$7A,$FC,$BC,$7F,$6B,$5A,$AE,$87,$E9,$70,$E9,$8F,$C3
	.dc.b	$F1,$1C,$8E,$4E,$77,$FE,$33,$F9,$6B,$8F,$F8,$C1,$CD,$F0,$AB,$7C
	.dc.b	$6F,$FC,$C5,$5C,$39,$4E,$AB,$13,$9E,$48,$FF,$9A,$4C,$9F,$E6,$B7
	.dc.b	$2F,$F9,$94,$F2,$9F,$C5,$7B,$7C,$BF,$A6,$BD,$FA,$6B,$54,$D3,$35
	.dc.b	$F0,$F8,$3F,$84,$6C,$C3,$94,$E4,$FE,$51,$5C,$5C,$4E,$F0,$E7,$1F
	.dc.b	$E9,$B3,$C8,$7F,$B7,$B1,$F9,$3D,$CF,$97,$A6,$22,$4F,$FB,$92,$32
	.dc.b	$F1,$3C,$5C,$0F,$CB,$5E,$6F,$D4,$FD,$51,$F0,$F4,$FA,$B3,$AB,$E2
	.dc.b	$D7,$FD,$53,$3C,$1F,$5B,$5F,$F5,$A5,$F7,$BE,$D9,$F7,$F9,$99,$5C
	.dc.b	$23,$2F,$28,$CB,$DA,$32,$F8,$23,$3C,$DF,$EE,$BC,$3E,$7C,$F9,$3E
	.dc.b	$5F,$D9,$E5,$48,$F2,$C7,$BC,$C3,$EA,$8C,$3E,$AF,$92,$F8,$9F,$31
	.dc.b	$23,$04,$3E,$0B,$97,$E1,$C7,$C8,$F4,$FE,$67,$CC,$49,$F9,$31,$97
	.dc.b	$E3,$C6,$5F,$8F,$BA,$75,$1F,$3A,$75,$9B,$9F,$FA,$B7,$07,$DB,$BF
	.dc.b	$CC,$D4,$BF,$C6,$F3,$A3,$EF,$7C,$BF,$EC,$3F,$71,$79,$9F,$A0,$89
	.dc.b	$7F,$86,$9C,$BF,$73,$E6,$7F,$0C,$47,$AC,$FE,$27,$E2,$3A,$7A,$32
	.dc.b	$7F,$07,$15,$FC,$3F,$87,$F8,$F3,$1B,$FD,$97,$11,$F8,$A7,$3B,$FB
	.dc.b	$B2,$FF,$76,$CF,$97,$CB,$F5,$1F,$B8,$F2,$2C,$7C,$0F,$98,$FD,$D9
	.dc.b	$87,$F2,$75,$C9,$FF,$80,$3C,$09,$DC,$75,$3E,$17,$A1,$F1,$C8,$FE
	.dc.b	$7A,$EB,$FA,$29,$FD,$51,$8C,$FE,$B4,$38,$D8,$B0,$DC,$FE,$A2,$F8
	.dc.b	$67,$F7,$3E,$35,$D7,$FC,$99,$B5,$FB,$A6,$BE,$AF,$97,$D4,$FF,$1A
	.dc.b	$F3,$64,$EE,$32,$FE,$12,$32,$FE,$DA,$B9,$E9,$E8,$D1,$0E,$9F,$11
	.dc.b	$2B,$F7,$FE,$2D,$7F,$81,$FC,$C7,$F0,$B3,$18,$73,$BE,$53,$FC,$1A
	.dc.b	$74,$FA,$EF,$F9,$9F,$C0,$BE,$7F,$FA,$DC,$C8,$C3,$83,$FC,$6C,$E1
	.dc.b	$C3,$5E,$11,$FD,$43,$F7,$47,$CC,$53,$BC,$49,$FE,$67,$3A,$6C,$E5
	.dc.b	$FE,$69,$4F,$88,$FE,$B7,$F1,$0F,$8B,$A7,$AA,$25,$7F,$3E,$9D,$36
	.dc.b	$72,$FF,$2A,$F5,$BF,$99,$70,$C8,$FE,$57,$1F,$CD,$4C,$3F,$CC,$23
	.dc.b	$0F,$F4,$AF,$77,$D0,$B8,$3F,$23,$E7,$A7,$31,$86,$F3,$7F,$58,$5F
	.dc.b	$33,$FD,$D9,$CD,$AF,$96,$D9,$35,$5F,$DA,$BE,$57,$E2,$57,$D5,$0E
	.dc.b	$1F,$B0,$D7,$EE,$B5,$FF,$0C,$BB,$C7,$FC,$31,$F2,$CE,$9F,$EE,$ED
	.dc.b	$AF,$ED,$87,$6E,$0F,$11,$A9,$91,$AB,$E9,$A7,$F8,$9B,$F1,$2F,$E3
	.dc.b	$B3,$F8,$B6,$BF,$90,$6B,$BF,$1B,$D8,$A3,$8E,$5F,$E2,$DF,$9E,$7C
	.dc.b	$EF,$F8,$31,$B5,$E6,$AD,$E1,$FE,$36,$F9,$57,$C7,$7F,$B0,$6B,$F3
	.dc.b	$91,$8E,$4F,$94,$D8,$CC,$ED,$FD,$2F,$B0,$F2,$1F,$2D,$F4,$63,$0D
	.dc.b	$8C,$CF,$EA,$67,$2F,$83,$E2,$9E,$CF,$81,$B1,$98,$C1,$F7,$7E,$53
	.dc.b	$F4,$1A,$5E,$BF,$F2,$75,$7F,$94,$A7,$2F,$ED,$DA,$FE,$F1,$CF,$0F
	.dc.b	$F9,$55,$B5,$FD,$73,$E9,$7F,$C5,$EF,$DE,$8C,$D7,$FA,$D7,$E5,$1F
	.dc.b	$19,$FC,$BB,$09,$F2,$AF,$9D,$C6,$3F,$E6,$4B,$DB,$FE,$33,$A5,$6E
	.dc.b	$3F,$17,$3A,$C3,$F8,$73,$5F,$E2,$6D,$7A,$61,$F0,$1F,$23,$E0,$6C
	.dc.b	$FF,$8C,$32,$FC,$D3,$DA,$75,$DC,$F0,$5F,$2D,$86,$D7,$F0,$9E,$DC
	.dc.b	$4C,$8F,$E1,$8E,$E3,$18,$FF,$01,$9F,$F2,$9D,$3E,$47,$A7,$C9,$75
	.dc.b	$6C,$FF,$15,$6D,$71,$3F,$80,$57,$19,$FE,$35,$5E,$D9,$78,$A8,$F4
	.dc.b	$7C,$06,$63,$15,$FC,$86,$31,$3F,$1E,$B8,$B6,$B7,$1F,$CE,$A3,$1F
	.dc.b	$25,$F1,$BC,$67,$CB,$75,$8C,$37,$1F,$E4,$4D,$BF,$56,$DF,$C8,$63
	.dc.b	$1C,$7F,$EA,$3C,$98,$7C,$FF,$A2,$CE,$3F,$E2,$CF,$97,$FD,$AB,$FC
	.dc.b	$E7,$BB,$EE,$7A,$29,$CE,$9D,$B8,$2E,$7F,$DA,$66,$E1,$F4,$B0,$E3
	.dc.b	$18,$BF,$9B,$18,$24,$FF,$47,$EC,$B3,$FE,$6A,$F9,$17,$CB,$7C,$3F
	.dc.b	$97,$F1,$C6,$2F,$FA,$14,$62,$75,$1F,$E8,$09,$D5,$7C,$4E,$25,$EE
	.dc.b	$B5,$7F,$3F,$93,$17,$E1,$BF,$CA,$70,$78,$1B,$7D,$5B,$3B,$8F,$EC
	.dc.b	$13,$B7,$47,$F6,$3F,$81,$C4,$FB,$7E,$2F,$EF,$73,$7B,$1B,$5E,$96
	.dc.b	$CE,$E3,$FB,$04,$ED,$D1,$FD,$8F,$9F,$13,$F5,$63,$F5,$87,$C8,$F9
	.dc.b	$1A,$FC,$56,$C7,$CF,$6F,$17,$F6,$AF,$8C,$F9,$2F,$F8,$97,$81,$3F
	.dc.b	$BA,$9C,$DF,$F9,$D2,$72,$FF,$AF,$A1,$F2,$7B,$3F,$39,$C5,$67,$EB
	.dc.b	$78,$98,$69,$FD,$F6,$3F,$45,$BC,$8E,$9D,$D7,$4F,$F5,$7F,$CB,$BE
	.dc.b	$27,$C9,$E7,$FE,$62,$31,$7E,$38,$C4,$EA,$3F,$D8,$B3,$AA,$FD,$4E
	.dc.b	$9F,$E8,$69,$CD,$7F,$1F,$EB,$5E,$EC,$7F,$A5,$1C,$36,$7F,$93,$77
	.dc.b	$B0,$6F,$F2,$79,$BE,$53,$72,$F9,$0D,$C3,$C6,$DE,$2F,$A9,$F0,$9F
	.dc.b	$47,$E5,$7E,$93,$6F,$E8,$36,$FD,$6D,$AF,$B9,$CF,$F6,$7E,$5F,$ED
	.dc.b	$36,$FE,$C3,$6F,$E1,$B6,$B8,$CF,$DD,$AE,$2D,$3E,$5F,$BD,$F8,$5F
	.dc.b	$97,$FA,$4D,$8C,$9F,$E5,$BE,$1F,$97,$1F,$37,$87,$FB,$0E,$5F,$E6
	.dc.b	$29,$CD,$7F,$98,$78,$3E,$57,$CA,$FE,$47,$E1,$FF,$24,$D9,$FC,$83
	.dc.b	$6F,$ED,$B7,$4F,$EE,$FF,$09,$FD,$4B,$E5,$3E,$F5,$7F,$41,$3E,$DC
	.dc.b	$67,$9B,$FA,$04,$7F,$A6,$70,$FB,$F7,$FE,$4E,$BE,$2F,$48,$C9,$F8
	.dc.b	$4C,$93,$A8,$FF,$3E,$E1,$FD,$0E,$FA,$1F,$E4,$97,$07,$E2,$25,$C5
	.dc.b	$E7,$E3,$F4,$36,$FB,$DF,$F9,$5E,$B7,$18,$C3,$C3,$7C,$23,$13,$F9
	.dc.b	$27,$8E,$AF,$95,$FD,$BC,$6B,$8F,$FA,$0C,$63,$1F,$E9,$B6,$9C,$F6
	.dc.b	$7E,$96,$B5,$5F,$A4,$3F,$D8,$4E,$1F,$F7,$CB,$5F,$F4,$FF,$CB,$FE
	.dc.b	$AB,$E5,$FF,$86,$C7,$AF,$8B,$E8,$A9,$D7,$7F,$CD,$C9,$5A,$9C,$D7
	.dc.b	$1C,$FF,$CC,$EF,$ED,$A3,$3B,$7D,$4A,$E1,$2F,$81,$F0,$BF,$BC,$F0
	.dc.b	$3F,$DB,$71,$88,$E4,$6F,$2F,$F5,$BC,$E9,$FF,$5B,$61,$E2,$9E,$58
	.dc.b	$D7,$8E,$7C,$EF,$FD,$CE,$F5,$8F,$71,$D4,$F9,$51,$D5,$F8,$E6,$4E
	.dc.b	$A5,$3A,$DF,$92,$BC,$11,$D2,$BC,$11,$8B,$FE,$33,$86,$2B,$91,$E2
	.dc.b	$70,$79,$63,$D5,$87,$A3,$0F,$7D,$30,$E4,$6D,$F2,$4F,$A5,$FC,$25
	.dc.b	$78,$2B,$CD,$1E,$08,$EB,$87,$96,$3A,$4E,$AF,$A1,$D9,$F0,$B6,$3E
	.dc.b	$2B,$5E,$E7,$23,$E5,$CF,$5C,$3D,$B5,$EA,$6C,$F2,$8F,$01,$D8,$AF
	.dc.b	$D2,$8D,$BD,$9A,$FD,$23,$B9,$E1,$3C,$75,$E5,$74,$7D,$A7,$DB,$3D
	.dc.b	$A3,$6D,$76,$7F,$A6,$2B,$D5,$1F,$2A,$BE,$81,$E2,$9F,$C0,$47,$16
	.dc.b	$9E,$4A,$EE,$D8,$F7,$34,$F7,$3D,$4F,$8C,$57,$E3,$1B,$3C,$E5,$3C
	.dc.b	$F8,$7E,$1D,$D9,$EF,$8E,$ED,$32,$7B,$4F,$90,$72,$BF,$C2,$57,$C9
	.dc.b	$8F,$59,$C4,$F0,$94,$E0,$72,$36,$75,$7E,$0B,$B3,$35,$E7,$79,$9B
	.dc.b	$AF,$3B,$4F,$3C,$6D,$EF,$3F,$A1,$F6,$7F,$86,$47,$EE,$1A,$FE,$51
	.dc.b	$AF,$F0,$C7,$3E,$B3,$E3,$1C,$DF,$E0,$E9,$CC,$62,$91,$89,$D6,$7F
	.dc.b	$CC,$6F,$94,$62,$BF,$C2,$F1,$87,$F2,$8B,$E2,$FF,$9C,$63,$FC,$8A
	.dc.b	$31,$3F,$C6,$DE,$11,$9C,$7C,$4E,$EB,$8E,$7E,$16,$DF,$82,$B7,$3E
	.dc.b	$56,$C7,$92,$76,$AF,$C2,$F9,$A7,$6E,$5B,$1A,$BF,$14,$F9,$E7,$26
	.dc.b	$AB,$AD,$6E,$FC,$58,$6C,$F7,$4E,$8D,$D7,$28,$CD,$C3,$C8,$F1,$7E
	.dc.b	$5C,$F1,$71,$87,$C7,$9E,$0F,$9A,$38,$93,$1C,$13,$0F,$2C,$7F,$3F
	.dc.b	$C7,$E8,$C7,$CD,$8E,$6E,$AB,$B4,$78,$8C,$9F,$46,$FC,$D7,$D6,$3E
	.dc.b	$81,$FC,$CC,$D1,$DF,$1F,$90,$F2,$3E,$84,$FB,$2F,$C5,$1D,$CF,$5D
	.dc.b	$F8,$8F,$71,$23,$CC,$7B,$2F,$C5,$3F,$7F,$1F,$F0,$44,$7A,$8E,$53
	.dc.b	$DA,$B9,$57,$99,$D4,$78,$A7,$9E,$3F,$9D,$78,$D7,$B8,$F3,$56,$1F
	.dc.b	$59,$8A,$EC,$79,$9E,$CE,$1E,$51,$C6,$BB,$4F,$84,$93,$F2,$AB,$57
	.dc.b	$F8,$38,$F8,$B3,$F0,$4D,$BF,$04,$FA,$47,$92,$78,$BF,$46,$BF,$02
	.dc.b	$F5,$9E,$31,$FA,$69,$F8,$73,$E8,$3C,$F3,$E9,$7D,$F7,$A7,$C0,$7E
	.dc.b	$36,$FE,$8C,$F9,$63,$96,$FF,$E5,$AE,$9F,$F1,$E6,$5F,$E6,$16,$ED
	.dc.b	$F1,$EB,$AD,$FF,$8D,$67,$31,$FE,$55,$6B,$A3,$FC,$8B,$C9,$FE,$47
	.dc.b	$7F,$C9,$1C,$BF,$E5,$38,$C4,$7F,$92,$1A,$F0,$3F,$D1,$5C,$9F,$E5
	.dc.b	$37,$FC,$BB,$F2,$FE,$0C,$FF,$CE,$2F,$8A,$FA,$4F,$F9,$88,$D7,$2B
	.dc.b	$8E,$98,$88,$7F,$9A,$1C,$BC,$4F,$F3,$89,$FC,$5F,$B3,$CA,$D3,$CF
	.dc.b	$C4,$8C,$B6,$33,$1E,$7F,$8D,$7C,$AD,$3E,$3D,$6E,$BF,$D2,$CD,$E1
	.dc.b	$70,$FF,$A3,$EF,$8D,$7F,$A0,$5A,$F0,$C3,$E2,$78,$7E,$F8,$FF,$7B
	.dc.b	$7C,$BE,$0C,$E7,$23,$24,$3F,$DC,$AF,$CB,$F9,$37,$9B,$F5,$37,$46
	.dc.b	$2B,$FD,$7C,$D7,$FD,$78,$D7,$EF,$B9,$D3,$D7,$0E,$17,$CB,$3E,$5F
	.dc.b	$EE,$36,$7F,$D3,$37,$C7,$7F,$F5,$84,$FC,$0C,$39,$E1,$CB,$E4,$38
	.dc.b	$9C,$B8,$A6,$3F,$EA,$7B,$4E,$A3,$13,$A7,$FB,$9E,$1D,$A7,$51,$83
	.dc.b	$D1,$5E,$19,$F1,$FC,$BF,$A3,$CB,$95,$DE,$3D,$66,$1E,$58,$FC,$09
	.dc.b	$CC,$79,$7E,$3F,$E0,$FC,$BC,$8C,$10,$E3,$8A,$35,$18,$9F,$B1,$E9
	.dc.b	$FD,$7D,$8D,$24,$FD,$87,$E7,$FC,$BC,$8F,$9F,$1C,$9F,$91,$19,$7E
	.dc.b	$17,$0F,$E0,$A7,$F3,$93,$27,$F2,$8F,$DA,$F9,$6F,$DA,$7C,$B3,$CC
	.dc.b	$FD,$14,$4A,$FF,$6A,$CE,$5F,$AD,$F2,$FF,$82,$23,$F9,$F9,$FC,$2F
	.dc.b	$C4,$74,$F4,$A4,$FD,$D8,$AF,$C6,$F2,$3F,$16,$63,$9F,$F0,$FC,$3F
	.dc.b	$E3,$19,$FC,$24,$E9,$F1,$DF,$A3,$CB,$91,$E6,$7F,$75,$E3,$E2,$3E
	.dc.b	$DF,$80,$F9,$7F,$DF,$19,$CB,$F3,$FC,$4D,$FF,$DB,$77,$FB,$9A,$DB
	.dc.b	$D4,$F8,$7E,$87,$C8,$23,$F9,$9C,$6B,$F8,$29,$FD,$39,$8C,$3B,$DA
	.dc.b	$F3,$5F,$A1,$AD,$75,$21,$8B,$E5,$FF,$4B,$19,$C7,$D3,$18,$8F,$C0
	.dc.b	$B6,$3F,$63,$9B,$E4,$7F,$99,$4E,$70,$FE,$31,$E6,$FF,$21,$33,$DD
	.dc.b	$31,$FE,$35,$19,$7C,$13,$E5,$CB,$C9,$5D,$3A,$A5,$68,$F4,$E0,$8F
	.dc.b	$E1,$3E,$2E,$A7,$F4,$FF,$2F,$FC,$30,$C6,$BE,$CA,$C4,$FF,$05,$9C
	.dc.b	$E5,$F1,$61,$F0,$6F,$97,$FE,$BA,$D8,$FE,$8A,$D7,$F6,$BF,$31,$A9
	.dc.b	$FE,$AD,$F3,$0F,$98,$A6,$6B,$8C,$E1,$F6,$78,$8A,$E1,$1A,$3B,$FC
	.dc.b	$4F,$F5,$2C,$4D,$24,$7F,$6A,$8C,$BD,$3E,$1D,$F2,$1F,$DA,$5B,$1D
	.dc.b	$0F,$E4,$DE,$BF,$F3,$2E,$59,$18,$21,$FC,$49,$CB,$FC,$52,$3F,$BA
	.dc.b	$7B,$FE,$88,$FE,$3E,$DE,$2F,$E6,$ED,$E0,$FC,$96,$A8,$7C,$C7,$2E
	.dc.b	$09,$87,$C4,$9E,$0F,$F4,$97,$6F,$F7,$6F,$97,$F7,$BF,$CD,$C3,$8C
	.dc.b	$7F,$A8,$C6,$37,$FF,$0B,$FC,$C7,$FA,$BA,$39,$47,$F7,$EA,$E7,$1F
	.dc.b	$E1,$96,$BF,$85,$6B,$FA,$B7,$B6,$7F,$89,$AF,$D2,$4F,$F8,$29,$AF
	.dc.b	$96,$76,$FE,$CF,$E5,$BF,$91,$3E,$96,$E1,$E8,$6D,$FE,$57,$E5,$FF
	.dc.b	$7D,$5F,$E1,$69,$E2,$FF,$8B,$DA,$FE,$6D,$AF,$7F,$27,$FC,$37,$F2
	.dc.b	$EF,$8D,$F8,$53,$FE,$37,$8C,$70,$FF,$8D,$1B,$1F,$E3,$86,$BF,$E3
	.dc.b	$DF,$61,$E5,$3E,$5F,$FC,$4A,$D9,$FF,$1F,$CE,$A7,$27,$5F,$8A,$FF
	.dc.b	$0C,$4F,$85,$AF,$EC,$ED,$18,$8F,$E3,$51,$8F,$63,$FC,$93,$F2,$FF
	.dc.b	$9E,$DF,$FC,$72,$DA,$FE,$B3,$F2,$9F,$E3,$79,$CF,$5F,$F8,$53,$8B
	.dc.b	$F0,$B7,$B6,$C6,$67,$F8,$A5,$71,$BF,$C7,$5E,$E7,$F8,$47,$CA,$BE
	.dc.b	$33,$F1,$F5,$86,$BF,$CC,$63,$53,$FE,$60,$6C,$7F,$99,$1B,$8F,$E6
	.dc.b	$1B,$3E,$2F,$23,$C9,$5C,$EE,$AA,$DB,$9A,$FE,$51,$E4,$32,$FE,$71
	.dc.b	$1F,$E5,$EA,$DF,$6B,$BD,$F2,$BD,$A3,$0D,$8F,$E2,$35,$FC,$36,$30
	.dc.b	$47,$F8,$87,$65,$CB,$FC,$43,$4F,$8F,$E5,$F2,$1C,$DB,$7F,$C5,$DB
	.dc.b	$5C,$1B,$BF,$EE,$BC,$35,$FC,$CA,$FE,$6E,$D8,$CD,$FF,$23,$8C,$1F
	.dc.b	$1A,$33,$3E,$4F,$17,$13,$9F,$90,$F8,$1E,$13,$E5,$71,$8F,$F1,$36
	.dc.b	$E1,$FE,$9A,$C5,$7E,$12,$57,$F3,$37,$79,$7F,$96,$34,$7C,$AE,$2B
	.dc.b	$FB,$07,$CA,$FE,$CD,$FE,$71,$E0,$7D,$EE,$54,$E7,$3F,$97,$39,$8C
	.dc.b	$47,$FA,$B9,$AF,$48,$FC,$8C,$F2,$6C,$F3,$C7,$FC,$2E,$DB,$FE,$6B
	.dc.b	$8F,$EC,$1F,$C4,$C6,$70,$FF,$04,$FC,$7B,$E5,$3E,$3F,$CA,$F9,$E3
	.dc.b	$17,$FC,$E6,$31,$3E,$29,$CD,$7F,$B1,$73,$46,$23,$FD,$8E,$D7,$FC
	.dc.b	$08,$F0,$CB,$FC,$5A,$DC,$3F,$A5,$F0,$7F,$A6,$1F,$CF,$C6,$6B,$CF
	.dc.b	$3F,$99,$DB,$CC,$DA,$DB,$E5,$23,$FE,$D6,$C3,$CB,$18,$AF,$F3,$1D
	.dc.b	$FF,$59,$AF,$ED,$DE,$FF,$47,$C3,$F1,$9F,$8A,$E3,$F8,$8D,$8C,$9D
	.dc.b	$69,$1F,$14,$91,$97,$8B,$FD,$8F,$96,$E1,$FA,$C8,$FD,$A1,$F4,$7E
	.dc.b	$3A,$BF,$09,$FA,$04,$7C,$0F,$F1,$4E,$97,$C5,$7C,$7D,$7E,$9C,$FF
	.dc.b	$01,$8C,$64,$8C,$7C,$8C,$C3,$FB,$B7,$89,$72,$F8,$BE,$1F,$DF,$24
	.dc.b	$FB,$DF,$07,$07,$EA,$23,$FB,$F7,$0F,$7F,$03,$F6,$2F,$E8,$7E,$4D
	.dc.b	$F0,$FE,$3F,$2F,$0D,$BF,$E3,$B3,$A7,$EC,$37,$2F,$C7,$35,$FF,$1F
	.dc.b	$70,$23,$11,$FE,$51,$78,$73,$7F,$9E,$9B,$5F,$E1,$8E,$17,$D0,$F8
	.dc.b	$A3,$FC,$73,$1F,$9A,$6E,$BF,$27,$8B,$E5,$36,$32,$70,$8C,$BC,$BE
	.dc.b	$07,$E8,$7C,$A7,$D1,$6C,$64,$EF,$19,$69,$E2,$E7,$FB,$1F,$2D,$F6
	.dc.b	$5B,$19,$3C,$F1,$96,$99,$D7,$FD,$AF,$92,$75,$18,$7E,$A7,$B7,$F8
	.dc.b	$5F,$96,$B7,$F0,$9B,$5F,$EC,$8D,$BF,$B4,$6F,$FE,$F8,$C5,$E9,$EA
	.dc.b	$31,$F2,$9F,$8F,$F8,$5F,$C8,$37,$2F,$9A,$DC,$7C,$AD,$AF,$ED,$FF
	.dc.b	$07,$FD,$5F,$E4,$BF,$DA,$17,$FE,$87,$BF,$F5,$04,$E7,$9F,$FB,$3C
	.dc.b	$7F,$A9,$E7,$FD,$00,$C9,$FE,$98,$E4,$C9,$CA,$2B,$4F,$F2,$81,$BB
	.dc.b	$F7,$76,$F6,$AF,$27,$17,$97,$92,$E1,$FE,$80,$CB,$F1,$18,$B1,$4F
	.dc.b	$F9,$9A,$57,$C6,$38,$36,$BF,$DE,$BF,$28,$EE,$F8,$5C,$BF,$D2,$63
	.dc.b	$17,$DE,$31,$32,$3F,$33,$F3,$36,$FD,$63,$E5,$FF,$D9,$1F,$2F,$FC
	.dc.b	$06,$1E,$13,$47,$1F,$CC,$9E,$02,$B8,$56,$E7,$57,$FE,$D3,$F8,$0E
	.dc.b	$97,$FE,$D5,$F0,$BE,$0A,$8C,$BF,$ED,$99,$E1,$C7,$F6,$2B,$75,$AA
	.dc.b	$F0,$E3,$C2,$79,$B9,$BF,$5D,$F1,$D2,$DF,$E0,$A6,$BE,$38,$AF,$84
	.dc.b	$F6,$4F,$16,$9C,$4A,$71,$29,$C6,$BC,$93,$D2,$3C,$F3,$CE,$7C,$D1
	.dc.b	$C2,$FD,$33,$F9,$2B,$F1,$C6,$27,$C1,$5D,$1E,$E7,$D2,$C3,$07,$42
	.dc.b	$DF,$38,$ED,$87,$8F,$0E,$AD,$8C,$9D,$23,$2D,$33,$7E,$83,$C4,$77
	.dc.b	$AF,$8D,$76,$7F,$DD,$D8,$7D,$89,$F1,$57,$06,$9F,$0C,$F3,$57,$A2
	.dc.b	$BE,$73,$5F,$F7,$C5,$F9,$2B,$C5,$5C,$67,$E2,$B5,$D9,$E8,$7B,$BE
	.dc.b	$4C,$3E,$04,$F8,$E3,$BC,$FA,$8F,$19,$5E,$A7,$13,$E2,$47,$17,$BB
	.dc.b	$5F,$C1,$47,$8D,$FC,$65,$FA,$DB,$1D,$5E,$2E,$DF,$C3,$9D,$E7,$DE
	.dc.b	$D8,$F5,$BF,$91,$8F,$A0,$FE,$4E,$6A,$C3,$E9,$3E,$A3,$85,$72,$75
	.dc.b	$7F,$12,$3C,$58,$7C,$F7,$55,$F9,$3A,$F3,$9E,$33,$E4,$9F,$16,$FC
	.dc.b	$31,$E0,$8D,$9E,$03,$C4,$78,$0F,$C7,$47,$C5,$7F,$39,$3F,$15,$F8
	.dc.b	$E7,$57,$EA,$B4,$FD,$A7,$B7,$FC,$2D,$B3,$FA,$18,$C4,$7E,$11,$BE
	.dc.b	$D3,$E2,$FF,$9A,$4A,$73,$18,$6F,$22,$75,$75,$5F,$39,$D5,$70,$9F
	.dc.b	$E3,$D8,$6E,$9F,$17,$C7,$0F,$E4,$B1,$8A,$FE,$42,$77,$7A,$3C,$8C
	.dc.b	$63,$E8,$9E,$35,$CE,$FB,$36,$B7,$8F,$47,$CF,$7F,$03,$5E,$31,$FC
	.dc.b	$93,$1F,$13,$FC,$79,$F0,$C7,$9A,$7C,$18,$76,$AE,$95,$C3,$1D,$61
	.dc.b	$8B,$ED,$7B,$BE,$98,$FC,$9C,$77,$1D,$EB,$6F,$C1,$9D,$57,$1A,$D9
	.dc.b	$1C,$12,$7D,$E7,$3B,$F2,$D7,$C3,$8F,$AE,$DC,$7E,$6E,$3C,$E7,$C0
	.dc.b	$7C,$39,$F0,$BD,$30,$D3,$C8,$D1,$F2,$71,$F4,$BF,$3B,$0F,$0B,$0F
	.dc.b	$B4,$64,$B8,$7C,$EC,$7D,$51,$D4,$F5,$12,$FE,$BB,$F2,$8F,$05,$47
	.dc.b	$9E,$1E,$D3,$DD,$87,$D8,$8F,$55,$F6,$C3,$EF,$E1,$EC,$7C,$53,$F0
	.dc.b	$8E,$C7,$86,$7F,$35,$5E,$E9,$E3,$1C,$E3,$07,$79,$F5,$63,$E8,$3E
	.dc.b	$F9,$DF,$0F,$9B,$1E,$FC,$3C,$55,$C7,$0F,$C6,$61,$F8,$9C,$38,$BE
	.dc.b	$2C,$7E,$8E,$1E,$6C,$38,$C6,$1F,$8E,$62,$35,$3E,$79,$F9,$15,$E5
	.dc.b	$8E,$C7,$59,$FA,$B8,$7B,$F1,$F3,$E3,$F0,$27,$E5,$90,$F5,$57,$57
	.dc.b	$E4,$1F,$2F,$1F,$1C,$75,$7E,$14,$79,$E3,$04,$78,$3C,$E3,$E0,$C7
	.dc.b	$D5,$C3,$D0,$FE,$81,$FC,$D6,$FF,$E3,$9E,$9F,$E2,$D9,$FF,$8C,$23
	.dc.b	$15,$FE,$69,$C7,$EF,$F8,$6E,$6F,$F6,$BC,$FF,$1F,$BF,$F1,$BC,$E7
	.dc.b	$9B,$9F,$17,$F2,$38,$C5,$7F,$A9,$BE,$5F,$2A,$32,$FF,$98,$67,$FC
	.dc.b	$C1,$3A,$CB,$FC,$9C,$F0,$AD,$E1,$E7,$E5,$FF,$3C,$B7,$8F,$84,$7F
	.dc.b	$5D,$C7,$FC,$DC,$60,$CC,$47,$27,$BF,$C0,$F2,$BF,$CE,$1B,$A9,$F2
	.dc.b	$77,$3B,$33,$1F,$E8,$4A,$E0,$7F,$A1,$AB,$83,$63,$51,$88,$F8,$07
	.dc.b	$F9,$B3,$E2,$9F,$29,$F0,$5B,$1E,$53,$FD,$9F,$8F,$8B,$E5,$E4,$60
	.dc.b	$87,$FA,$E9,$CB,$FE,$BC,$8F,$95,$F2,$6F,$97,$A4,$9F,$F6,$D9,$4F
	.dc.b	$2F,$0B,$E4,$5F,$2F,$F5,$5B,$3F,$83,$E4,$FC,$06,$1D,$63,$07,$FA
	.dc.b	$22,$7D,$6F,$45,$F2,$1C,$CE,$9A,$C9,$E3,$8F,$86,$75,$18,$AE,$7C
	.dc.b	$7F,$99,$6E,$3F,$DC,$EF,$85,$6E,$FA,$3E,$1F,$97,$F5,$79,$D2,$3A
	.dc.b	$47,$B4,$C3,$E7,$8C,$3E,$1F,$90,$F8,$5F,$2F,$23,$04,$3D,$CE,$5F
	.dc.b	$81,$1F,$4B,$D6,$91,$F1,$BA,$5B,$A7,$E7,$7B,$BA,$AF,$F6,$BD,$6A
	.dc.b	$7F,$DA,$6D,$8F,$F6,$BB,$5D,$D7,$37,$E4,$7C,$BF,$D9,$3A,$70,$7F
	.dc.b	$01,$E6,$47,$E4,$49,$3F,$3F,$E5,$BF,$65,$F2,$CF,$3B,$C2,$D9,$F0
	.dc.b	$B5,$F3,$B5,$FF,$12,$71,$7E,$2F,$E5,$FF,$14,$48,$FB,$F8,$B0,$FC
	.dc.b	$77,$C6,$F8,$3D,$59,$5F,$81,$8A,$F6,$8F,$19,$1F,$F7,$CE,$1F,$99
	.dc.b	$F2,$BF,$30,$48,$FA,$19,$2B,$DB,$E6,$C9,$CB,$FB,$3F,$2F,$11,$EC
	.dc.b	$8F,$82,$47,$E0,$BF,$B9,$F9,$79,$1C,$8E,$FE,$1F,$CA,$23,$F9,$BA
	.dc.b	$AF,$C8,$8F,$9A,$47,$E6,$BF,$B1,$99,$1D,$0F,$C5,$7C,$BF,$EB,$1B
	.dc.b	$D7,$FC,$A6,$31,$1F,$BC,$DD,$F1,$79,$AF,$C5,$D6,$AB,$FC,$B5,$CC
	.dc.b	$8C,$47,$F3,$D6,$BF,$CE,$DE,$BC,$BF,$8B,$F3,$3F,$8D,$1D,$67,$CB
	.dc.b	$D5,$E3,$23,$FB,$BF,$1A,$BF,$9E,$B4,$F9,$09,$1F,$9E,$8C,$BF,$C0
	.dc.b	$FE,$5F,$F8,$A1,$CB,$07,$81,$FE,$73,$ED,$7C,$47,$F3,$E6,$DF,$F1
	.dc.b	$60,$E1,$FE,$2D,$F2,$FB,$9F,$E7,$D3,$99,$FF,$3F,$1C,$23,$24,$7F
	.dc.b	$8C,$FC,$BB,$E5,$BF,$AE,$36,$3C,$15,$DF,$1F,$3C,$7F,$A2,$E1,$F1
	.dc.b	$9E,$72,$47,$F2,$1E,$2A,$F8,$8F,$4C,$79,$88,$FE,$8B,$C6,$FF,$4D
	.dc.b	$3F,$CA,$78,$7F,$9D,$1F,$CE,$FD,$59,$19,$21,$FC,$DD,$FE,$E9,$EF
	.dc.b	$FA,$63,$FA,$2B,$72,$FF,$05,$37,$7F,$E6,$4B,$E5,$F8,$ED,$2F,$F9
	.dc.b	$EC,$ED,$77,$A3,$84,$7D,$5F,$96,$F8,$6D,$D7,$FB,$BC,$62,$3F,$B5
	.dc.b	$8F,$97,$F7,$B8,$8D,$67,$F3,$A1,$18,$CF,$FD,$70,$D8,$FF,$5E,$57
	.dc.b	$2A,$E1,$3D,$E7,$F0,$2D,$78,$DF,$F8,$0E,$31,$1F,$EC,$5F,$97,$FE
	.dc.b	$49,$5E,$16,$E1,$FE,$C7,$6C,$FF,$B2,$FE,$03,$FC,$1C,$FE,$56,$32
	.dc.b	$9A,$62,$37,$1F,$0B,$E5,$9F,$1F,$F0,$A3,$FD,$A9,$18,$E0,$FC,$44
	.dc.b	$23,$0F,$E2,$DA,$FE,$7B,$D5,$78,$EF,$97,$FF,$0F,$36,$7F,$25,$3A
	.dc.b	$9C,$9D,$A9,$F1,$5F,$E3,$5A,$FD,$4C,$63,$D6,$D4,$E6,$7C,$ED,$7F
	.dc.b	$DD,$DF,$28,$F8,$BF,$D3,$F0,$4F,$98,$7C,$CE,$38,$BF,$DE,$57,$B7
	.dc.b	$FC,$27,$4A,$DC,$7F,$01,$9D,$61,$FA,$B6,$BF,$BE,$6B,$D3,$0F,$C5
	.dc.b	$3F,$EF,$FF,$C3,$3F,$B4,$D7,$F7,$BF,$C3,$5E,$1F,$8E,$82,$B6,$FF
	.dc.b	$90,$BC,$75,$70,$BF,$E2,$95,$AC,$FF,$CD,$F3,$9B,$FC,$FC,$E6,$BF
	.dc.b	$CE,$13,$9A,$FC,$F4,$F6,$AF,$E2,$ED,$7D,$1E,$A7,$A9,$F2,$C6,$57
	.dc.b	$23,$F1,$B7,$C8,$FA,$E3,$0F,$B7,$BB,$F7,$1C,$39,$AD,$E9,$FE,$42
	.dc.b	$BD,$B6,$33,$87,$F0,$AA,$E1,$87,$3E,$E7,$D1,$77,$FD,$97,$D9,$8F
	.dc.b	$F0,$C8,$C7,$0E,$DB,$1F,$C3,$1B,$8F,$F0,$4F,$19,$F2,$7D,$7E,$6B
	.dc.b	$EA,$4F,$F4,$08,$C5,$FF,$34,$6C,$71,$6F,$17,$FA,$83,$F6,$9F,$F3
	.dc.b	$BD,$7F,$1A,$FA,$3F,$BD,$CC,$9C,$E1,$FC,$9A,$73,$18,$BF,$E4,$F1
	.dc.b	$8F,$82,$F8,$0D,$C3,$FC,$77,$5C,$AF,$FB,$9B,$5F,$F5,$73,$5F,$F5
	.dc.b	$67,$C9,$3E,$63,$E5,$7C,$CF,$92,$31,$7F,$CC,$23,$13,$A8,$FF,$29
	.dc.b	$4E,$AB,$F9,$ED,$6A,$73,$5F,$57,$E1,$F1,$5F,$D0,$BE,$77,$EC,$7D
	.dc.b	$07,$D9,$6D,$FF,$3E,$6C,$EE,$3F,$A0,$CE,$DD,$1F,$D1,$3C,$BC,$4F
	.dc.b	$E9,$72,$7C,$77,$E1,$7E,$6F,$D8,$DB,$F4,$B6,$B8,$37,$83,$FD,$AF
	.dc.b	$7C,$AB,$67,$F5,$2E,$9E,$6F,$F5,$4F,$2B,$F2,$5E,$AF,$E4,$DB,$19
	.dc.b	$3F,$AC,$C6,$5A,$67,$1F,$EC,$B3,$F9,$BF,$9E,$D4,$7F,$68,$F1,$FF
	.dc.b	$3C,$7D,$5F,$22,$DF,$D2,$6C,$FC,$16,$E7,$FD,$BB,$A3,$85,$7F,$6A
	.dc.b	$6D,$7F,$74,$F8,$8F,$D5,$1F,$AB,$9A,$7F,$AB,$3E,$3F,$EF,$5F,$F7
	.dc.b	$88,$FD,$D4,$61,$FE,$F2,$63,$C8,$8F,$1F,$97,$B8,$7F,$70,$3F,$DC
	.dc.b	$91,$D7,$E2,$7F,$5B,$18,$3F,$D9,$77,$FE,$89,$9F,$F1,$84,$67,$0F
	.dc.b	$F7,$E5,$70,$F8,$2F,$1B,$6B,$A1,$C2,$32,$FA,$FC,$4F,$4F,$CE,$7A
	.dc.b	$9B,$19,$3E,$04,$65,$A7,$C3,$F9,$9F,$83,$CB,$F0,$9B,$19,$3E,$44
	.dc.b	$65,$A7,$CB,$F9,$CF,$93,$F1,$1F,$29,$B1,$93,$E8,$46,$5A,$67,$4F
	.dc.b	$8B,$19,$6C,$7F,$B1,$7E,$55,$BF,$D6,$E9,$FE,$1F,$18,$9F,$B9,$18
	.dc.b	$9D,$47,$16,$D7,$D2,$F5,$FE,$FF,$CE,$7E,$01,$B5,$F8,$56,$CE,$E3
	.dc.b	$F1,$13,$F7,$21,$F3,$5F,$D7,$3E,$65,$F3,$5F,$EF,$1E,$57,$C6,$EE
	.dc.b	$B5,$39,$8C,$5A,$F8,$BF,$E4,$2B,$E3,$8F,$E1,$A3,$B9,$BB,$E5,$19
	.dc.b	$C3,$AC,$F1,$C3,$AE,$1D,$3D,$36,$3D,$2B,$F3,$07,$49,$F0,$BF,$DD
	.dc.b	$BA,$FF,$CE,$71,$88,$FE,$F3,$87,$67,$58,$D7,$27,$94,$FE,$FE,$2B
	.dc.b	$9E,$5F,$E7,$7A,$E1,$18,$9F,$F4,$5C,$62,$BA,$47,$96,$BF,$D1,$F5
	.dc.b	$B8,$EB,$5B,$7C,$06,$0A,$78,$6B,$9C,$6E,$FD,$93,$D4,$F3,$D7,$57
	.dc.b	$D7,$1F,$06,$31,$1B,$3E,$09,$D8,$C1,$DC,$C4,$F7,$BF,$81,$1E,$49
	.dc.b	$E7,$39,$8F,$82,$66,$FC,$ED,$39,$96,$FD,$33,$C6,$3E,$04,$79,$E7
	.dc.b	$8B,$63,$8B,$4E,$2F,$3B,$F2,$C6,$5A,$7D,$09,$E6,$D9,$E6,$D9,$E6
	.dc.b	$71,$8F,$5B,$5F,$2C,$6C,$EA,$FD,$09,$EA,$D3,$A9,$67,$AC,$F1,$8E
	.dc.b	$AD,$7A,$9B,$7A,$BF,$5A,$38,$B9,$38,$94,$E2,$6E,$38,$BF,$6E,$38
	.dc.b	$B9,$38,$94,$E2,$6E,$38,$DF,$DE,$8F,$9C,$70,$34,$57,$53,$F1,$E3
	.dc.b	$F2,$33,$ED,$69,$E7,$29,$E1,$29,$D0,$F1,$CF,$48,$E5,$1F,$C2,$BE
	.dc.b	$0B,$F0,$87,$E0,$8B,$1E,$8B,$FA,$73,$D4,$FC,$ED,$73,$7C,$46,$23
	.dc.b	$6F,$84,$F1,$C6,$DE,$58,$7E,$5D,$CC,$F0,$9D,$47,$13,$F8,$FF,$A8
	.dc.b	$C3,$F5,$2D,$8F,$D3,$B5,$FE,$00,$DF,$71,$F1,$EE,$CF,$F4,$8C,$E6
	.dc.b	$31,$5F,$E7,$EB,$FE,$2F,$7C,$2F,$71,$8A,$DD,$F1,$9F,$E4,$54,$F8
	.dc.b	$EF,$E3,$FA,$FF,$2A,$8C,$63,$AC,$3B,$CE,$5B,$AF,$82,$F8,$B6,$B8
	.dc.b	$61,$1F,$E3,$B8,$74,$6D,$70,$7F,$D5,$B1,$E9,$77,$7D,$9C,$CE,$D4
	.dc.b	$F3,$C3,$C9,$7D,$DF,$55,$6E,$75,$87,$C3,$AD,$3E,$7B,$F1,$98,$3C
	.dc.b	$78,$7C,$EC,$B9,$C7,$FB,$09,$AF,$DD,$9E,$8D,$38,$D7,$CE,$C3,$C4
	.dc.b	$78,$EF,$D9,$3E,$A6,$E1,$E6,$8F,$79,$E3,$35,$5C,$D8,$7C,$67,$51
	.dc.b	$C6,$3E,$2D,$77,$9C,$BE,$28,$DD,$78,$23,$11,$CC,$98,$FB,$89,$87
	.dc.b	$B9,$F9,$F1,$F2,$08,$E0,$91,$F3,$5E,$31,$F1,$0E,$B1,$92,$39,$8E
	.dc.b	$2F,$C2,$3A,$C7,$84,$8E,$63,$AB,$F1,$8E,$A6,$63,$AB,$98,$EA,$FC
	.dc.b	$63,$E5,$B8,$C3,$F2,$4F,$47,$E6,$18,$9F,$0C,$F1,$3C,$A7,$A5,$CC
	.dc.b	$70,$AF,$A5,$3E,$43,$85,$75,$BF,$C0,$9F,$4E,$3C,$77,$CA,$31,$1D
	.dc.b	$8F,$54,$76,$AF,$B3,$5E,$9A,$DB,$D7,$0F,$C0,$DF,$43,$CC,$43,$C7
	.dc.b	$7F,$1B,$0F,$74,$65,$F5,$3F,$88,$9E,$4F,$03,$57,$E0,$AE,$4F,$B2
	.dc.b	$77,$5F,$2A,$31,$5F,$E2,$FD,$FF,$C7,$BF,$35,$FE,$30,$B6,$DF,$E5
	.dc.b	$38,$C3,$F0,$E3,$FC,$8F,$3C,$2B,$FD,$49,$5A,$7C,$73,$AE,$04,$E6
	.dc.b	$B8,$57,$F2,$DB,$FC,$5C,$7F,$93,$BE,$6F,$57,$FE,$4C,$C3,$C7,$7F
	.dc.b	$E5,$DC,$F3,$52,$3F,$D6,$91,$C0,$C3,$D4,$7A,$4E,$8F,$F7,$1C,$7F
	.dc.b	$AE,$FE,$6F,$B9,$E9,$9D,$74,$DA,$DD,$7F,$AC,$AB,$6F,$FA,$36,$7C
	.dc.b	$FF,$35,$E8,$F0,$64,$FF,$B1,$E3,$2F,$FB,$26,$32,$FF,$AF,$23,$2F
	.dc.b	$FA,$FA,$33,$9F,$C0,$D7,$FC,$E3,$5F,$12,$BE,$2F,$34,$7F,$DA,$86
	.dc.b	$4F,$F6,$B3,$97,$FD,$92,$7D,$1F,$91,$7C,$8D,$F9,$A9,$23,$E4,$60
	.dc.b	$AD,$C6,$71,$F2,$46,$23,$6A,$B9,$C6,$2F,$F0,$8F,$F4,$8F,$92,$F1
	.dc.b	$7B,$12,$3A,$72,$7F,$6A,$9D,$F3,$9D,$4F,$6A,$CC,$48,$E2,$33,$F4
	.dc.b	$4F,$0C,$FF,$13,$3E,$13,$73,$ED,$F9,$AF,$88,$48,$CE,$8F,$9A,$F9
	.dc.b	$1F,$35,$F2,$7E,$56,$56,$A3,$E6,$7B,$D2,$38,$BA,$7E,$2F,$C2,$FE
	.dc.b	$37,$E5,$DE,$75,$3E,$A4,$59,$F0,$FC,$C7,$D7,$30,$FC,$E7,$0E,$8F
	.dc.b	$83,$D6,$BE,$B1,$EA,$F0,$3F,$00,$47,$DB,$F3,$35,$FB,$3F,$1D,$F8
	.dc.b	$53,$1C,$5F,$7A,$B1,$3F,$12,$73,$CD,$F8,$7C,$7C,$1E,$47,$B4,$95
	.dc.b	$F2,$BC,$5A,$FE,$1B,$E6,$7F,$24,$63,$2F,$DE,$DF,$29,$FA,$33,$A7
	.dc.b	$DB,$A7,$FB,$91,$FA,$BF,$31,$F9,$B2,$3F,$63,$A6,$BF,$94,$F8,$7F
	.dc.b	$D0,$18,$C7,$F3,$35,$89,$FB,$F3,$9D,$3D,$9F,$33,$F9,$DB,$F2,$B6
	.dc.b	$33,$DB,$FA,$BF,$98,$FE,$11,$39,$C3,$F2,$5D,$5F,$E4,$DC,$3F,$4F
	.dc.b	$E5,$EB,$09,$5F,$8E,$8F,$DA,$18,$7F,$8C,$6E,$FD,$AD,$F0,$DF,$EE
	.dc.b	$49,$FF,$72,$CE,$63,$11,$F4,$E7,$6E,$27,$4D,$8C,$BF,$BE,$F9,$A7
	.dc.b	$CC,$FF,$35,$6C,$7E,$32,$77,$97,$D6,$6E,$DF,$C1,$FC,$15,$F0,$BE
	.dc.b	$98,$7F,$2B,$BD,$C7,$F0,$DE,$D7,$C0,$F7,$F9,$99,$3F,$BA,$13,$A7
	.dc.b	$0D,$7D,$11,$C1,$FE,$7C,$70,$9F,$57,$C7,$FF,$53,$F4,$DF,$B3,$FF
	.dc.b	$02,$DB,$FA,$5C,$63,$6D,$E1,$FC,$13,$56,$1C,$70,$E1,$5F,$E8,$26
	.dc.b	$D7,$F0,$FA,$DB,$FC,$4A,$DF,$8E,$FD,$7C,$7B,$B7,$0F,$F4,$5B,$6B
	.dc.b	$97,$C8,$79,$E3,$17,$CA,$70,$7D,$59,$FE,$57,$F0,$9F,$C5,$63,$F0
	.dc.b	$21,$C3,$F8,$46,$BF,$89,$6B,$FC,$FC,$E1,$E4,$7F,$36,$F8,$87,$97
	.dc.b	$4E,$F1,$27,$EB,$4E,$9B,$39,$7E,$9F,$CD,$3E,$2B,$CB,$08,$C3,$E8
	.dc.b	$6B,$F9,$A6,$BF,$9F,$73,$AF,$F3,$AF,$9B,$77,$7F,$4C,$6D,$7E,$26
	.dc.b	$B7,$BF,$B7,$E6,$5F,$16,$E0,$F8,$F1,$86,$C6,$6F,$F5,$73,$97,$F5
	.dc.b	$BE,$8B,$C9,$7C,$D7,$D0,$8C,$36,$33,$3C,$1A,$FE,$E6,$BE,$B7,$A9
	.dc.b	$E2,$E3,$34,$93,$FB,$CF,$99,$7C,$57,$F5,$A8,$FB,$11,$8D,$7F,$02
	.dc.b	$D8,$E1,$3B,$7F,$BE,$46,$78,$7F,$B1,$B6,$BF,$9D,$7D,$17,$CC,$7F
	.dc.b	$65,$19,$AF,$F0,$27,$CC,$BD,$3F,$F1,$9E,$1F,$CE,$3E,$69,$E5,$E3
	.dc.b	$1F,$C6,$5E,$DF,$EA,$B3,$A7,$F7,$D1,$9C,$3F,$8C,$B5,$FE,$3C,$D7
	.dc.b	$9E,$1E,$B5,$1C,$70,$FE,$FA,$DA,$FF,$0D,$7E,$20,$FC,$4D,$F2,$7E
	.dc.b	$5C,$3D,$78,$DE,$67,$CE,$6B,$1F,$DD,$36,$BF,$C4,$7F,$23,$FC,$12
	.dc.b	$73,$87,$1A,$D6,$9F,$C3,$1B,$3F,$C9,$6B,$85,$FF,$94,$F0,$EF,$87
	.dc.b	$8F,$CA,$53,$C1,$F2,$7C,$E7,$C7,$C3,$C3,$18,$E1,$E1,$5B,$CB,$F8
	.dc.b	$85,$EF,$4F,$F2,$44,$E7,$0F,$04,$E6,$BF,$80,$79,$2F,$94,$E7,$F3
	.dc.b	$9F,$16,$FF,$98,$C6,$2B,$F9,$8B,$F4,$6B,$84,$7F,$10,$6C,$FF,$52
	.dc.b	$6B,$FD,$5B,$E6,$FF,$85,$CF,$F8,$C3,$17,$CE,$7D,$3F,$41,$5A,$BF
	.dc.b	$9F,$39,$8C,$69,$EC,$8C,$78,$DF,$09,$B8,$FF,$88,$AB,$6F,$D1,$F5
	.dc.b	$73,$87,$F6,$CF,$94,$7C,$E7,$CA,$F5,$3C,$51,$8B,$F3,$46,$27,$51
	.dc.b	$FE,$2E,$9D,$57,$E1,$2B,$53,$9A,$FA,$BF,$39,$8A,$FC,$4F,$CE,$7D
	.dc.b	$8F,$9C,$FB,$2D,$BF,$88,$D9,$DC,$7C,$79,$DB,$A3,$E5,$79,$B8,$9F
	.dc.b	$67,$C5,$FE,$13,$CF,$F5,$B6,$BD,$0D,$9D,$C7,$D3,$9D,$BA,$3E,$B7
	.dc.b	$5E,$27,$F4,$3E,$67,$E4,$3D,$5F,$C8,$B6,$BE,$E3,$67,$71,$F7,$E7
	.dc.b	$6E,$8F,$C1,$7C,$E6,$27,$F5,$BE,$57,$E7,$0F,$A7,$E5,$5A,$FA,$0D
	.dc.b	$9F,$B4,$DE,$7F,$CF,$F5,$F0,$AF,$C7,$36,$BF,$73,$F3,$9F,$A8,$3F
	.dc.b	$53,$34,$FF,$41,$FC,$97,$DC,$BF,$D4,$C7,$ED,$63,$0F,$EE,$CC,$79
	.dc.b	$91,$E3,$F3,$17,$0F,$F0,$98,$F8,$FE,$5F,$15,$FA,$D8,$C5,$78,$23
	.dc.b	$13,$A8,$FF,$11,$4E,$8F,$F5,$9E,$1F,$E8,$88,$CC,$FF,$23,$9D,$1F
	.dc.b	$E4,$27,$07,$F4,$1F,$19,$5F,$C3,$AB,$DF,$F3,$9E,$DF,$0B,$DC,$D8
	.dc.b	$C9,$F1,$23,$2F,$C6,$F9,$CF,$89,$DD,$F1,$5B,$19,$3E,$5C,$65,$A7
	.dc.b	$CE,$F9,$CF,$99,$F3,$9F,$35,$B1,$93,$E9,$C6,$5A,$67,$4F,$87,$19
	.dc.b	$6C,$7F,$A8,$BE,$55,$BF,$D9,$EB,$FE,$1D,$18,$9F,$BF,$18,$9F,$EC
	.dc.b	$DE,$32,$B5,$7F,$E6,$F8,$C4,$EA,$73,$3F,$E2,$9D,$FE,$6F,$CE,$7E
	.dc.b	$1F,$E7,$3F,$10,$DC,$BF,$1A,$DA,$FB,$6D,$E5,$FE,$ED,$F3,$9F,$EF
	.dc.b	$1F,$9C,$7C,$D7,$F9,$B7,$95,$F2,$3F,$89,$C5,$7B,$AD,$4E,$63,$13
	.dc.b	$FE,$E3,$AE,$39,$7E,$32,$38,$8C,$3B,$BB,$BE,$4B,$E2,$B2,$8F,$F0
	.dc.b	$BD,$58,$CC,$67,$0C,$63,$FE,$97,$ED,$5E,$2B,$FD,$35,$5E,$67,$A4
	.dc.b	$E9,$F2,$57,$FA,$72,$75,$1E,$03,$27,$22,$9C,$AB,$6F,$BB,$1F,$2D
	.dc.b	$F5,$7C,$F1,$9B,$F3,$57,$0B,$EF,$7B,$9F,$15,$7C,$37,$6D,$3C,$58
	.dc.b	$76,$BF,$74,$7F,$9B,$6F,$9C,$75,$8F,$39,$EB,$C3,$C7,$3A,$8C,$9E
	.dc.b	$53,$FC,$ED,$87,$8D,$EF,$87,$8B,$0E,$31,$E3,$3E,$11,$4F,$A8,$78
	.dc.b	$27,$93,$E5,$31,$87,$96,$7D,$51,$C4,$FC,$3B,$83,$BE,$1A,$2D,$FC
	.dc.b	$47,$C7,$1B,$8F,$92,$D9,$FB,$13,$C6,$7E,$6B,$63,$DA,$7C,$0A,$F9
	.dc.b	$6F,$AA,$B9,$3E,$99,$E0,$DA,$F1,$1D,$0A,$F8,$1A,$EC,$EC,$7D,$0C
	.dc.b	$3E,$94,$FA,$A7,$E7,$1E,$A2,$CF,$33,$8C,$7C,$26,$C6,$DE,$B5,$CD
	.dc.b	$B1,$E5,$76,$73,$3E,$CC,$F3,$6C,$6D,$A6,$CF,$C5,$39,$69,$F2,$8F
	.dc.b	$C1,$CF,$85,$B1,$BC,$3F,$07,$1F,$86,$AF,$96,$D9,$F8,$ED,$3C,$E7
	.dc.b	$E2,$6B,$E9,$BF,$56,$3E,$2B,$F3,$4F,$39,$4F,$29,$6B,$C2,$FE,$7C
	.dc.b	$D9,$DC,$A6,$4E,$26,$4A,$65,$E7,$7F,$6A,$7D,$D8,$7E,$C7,$0F,$6D
	.dc.b	$EE,$31,$3E,$E8,$C1,$F8,$33,$F4,$A5,$9E,$35,$D6,$3E,$95,$F8,$1F
	.dc.b	$F6,$76,$1F,$C2,$3D,$3F,$E5,$51,$FC,$5D,$AF,$EC,$5B,$EF,$BE,$49
	.dc.b	$B7,$F2,$F9,$CC,$62,$BA,$9F,$CC,$EB,$85,$6E,$FF,$DF,$91,$89,$D8
	.dc.b	$71,$97,$F3,$C5,$F2,$5F,$CF,$67,$F9,$5C,$61,$FE,$6C,$F0,$D3,$AE
	.dc.b	$3E,$3B,$E5,$7C,$5B,$7C,$27,$55,$C3,$1E,$77,$B8,$F3,$B6,$BD,$95
	.dc.b	$C1,$B5,$C1,$87,$9A,$34,$FB,$5B,$87,$A6,$3D,$75,$D5,$D5,$79,$4E
	.dc.b	$78,$7A,$23,$A4,$65,$E2,$6B,$0F,$53,$A3,$24,$9F,$99,$87,$88,$F1
	.dc.b	$DF,$87,$0F,$24,$EB,$0E,$D8,$F8,$49,$8F,$67,$58,$F5,$8F,$8E,$F9
	.dc.b	$6B,$27,$A4,$91,$E6,$78,$CF,$49,$FC,$46,$1F,$7E,$33,$5B,$3A,$CF
	.dc.b	$73,$E8,$9E,$FA,$EA,$F2,$3C,$51,$C5,$C1,$C4,$F9,$E7,$C7,$9F,$13
	.dc.b	$F9,$3A,$F1,$4F,$BC,$E5,$19,$3E,$81,$E2,$AF,$3B,$DA,$B9,$E1,$F9
	.dc.b	$F7,$E1,$D7,$AA,$BC,$F5,$E2,$72,$F9,$9C,$D7,$E7,$63,$B4,$F1,$23
	.dc.b	$83,$E1,$D7,$82,$FD,$31,$89,$F4,$D7,$DE,$78,$C7,$85,$F7,$3E,$97
	.dc.b	$E7,$D7,$B2,$34,$79,$63,$8C,$F9,$63,$E5,$CF,$85,$FF,$07,$C7,$C1
	.dc.b	$9F,$0D,$7E,$3C,$FC,$86,$3F,$93,$C7,$71,$E8,$BF,$9C,$7F,$82,$6F
	.dc.b	$93,$EF,$33,$8F,$C6,$BE,$51,$F7,$67,$DC,$ED,$E2,$F4,$C7,$EA,$CF
	.dc.b	$F1,$97,$E8,$C7,$4C,$7D,$B3,$D2,$FE,$9E,$1E,$8C,$7F,$1B,$7F,$0C
	.dc.b	$87,$C6,$3E,$1D,$FC,$A2,$3F,$4C,$FB,$1C,$5F,$EA,$2F,$9C,$FE,$3D
	.dc.b	$5F,$E4,$F8,$C4,$FF,$97,$63,$FC,$A4,$F5,$CF,$F0,$93,$A8,$FF,$59
	.dc.b	$D7,$2D,$FF,$41,$3F,$EA,$EF,$9C,$C5,$7F,$9C,$71,$B3,$FE,$B2,$6F
	.dc.b	$12,$A9,$88,$95,$FE,$D4,$8E,$7C,$FF,$E9,$5E,$8F,$E6,$D3,$FC,$DE
	.dc.b	$30,$E0,$38,$7F,$D3,$5C,$E9,$D4,$6A,$30,$FF,$B8,$BE,$73,$CA,$78
	.dc.b	$23,$31,$FE,$D2,$F9,$CF,$4F,$83,$23,$FD,$D3,$1E,$C3,$0F,$FB,$62
	.dc.b	$30,$E8,$FF,$77,$BA,$E0,$FA,$5A,$2B,$E2,$D7,$C6,$E6,$8F,$FB,$D8
	.dc.b	$C9,$FE,$F8,$72,$FF,$B9,$8F,$9F,$F2,$4F,$92,$BF,$39,$27,$EF,$71
	.dc.b	$7F,$A6,$F8,$97,$B9,$C8,$8C,$DF,$18,$FE,$EB,$5C,$69,$7E,$0A,$FC
	.dc.b	$0D,$7F,$6F,$F9,$5F,$9E,$D3,$E8,$F2,$A3,$2D,$8C,$C7,$AB,$D9,$91
	.dc.b	$E4,$39,$1E,$23,$FC,$05,$1A,$71,$97,$B2,$7C,$75,$D6,$73,$8B,$E7
	.dc.b	$3E,$39,$23,$E0,$DF,$C4,$F8,$EF,$97,$F3,$7F,$33,$E7,$25,$6B,$04
	.dc.b	$FD,$0F,$7A,$4E,$5D,$BE,$0F,$84,$FC,$DF,$CC,$BC,$EA,$7D,$88,$B3
	.dc.b	$D7,$E7,$3E,$D9,$87,$D2,$E1,$D1,$F6,$3B,$17,$D6,$3C,$DE,$17,$E1
	.dc.b	$08,$FA,$7E,$72,$BF,$13,$E3,$FF,$12,$63,$AF,$F8,$36,$DF,$E3,$59
	.dc.b	$F8,$13,$A7,$A9,$E1,$C3,$FD,$EE,$DF,$23,$DE,$7E,$53,$C4,$B1,$F2
	.dc.b	$7E,$73,$F2,$E6,$1F,$B9,$A7,$AA,$B3,$1B,$3A,$1F,$47,$E7,$3F,$3E
	.dc.b	$47,$EA,$74,$D7,$EF,$FC,$47,$E9,$0C,$6B,$EC,$AC,$4F,$DB,$9C,$F6
	.dc.b	$FF,$BF,$F1,$7C,$27,$E9,$6B,$CA,$DC,$3F,$DF,$13,$AE,$DF,$D8,$FC
	.dc.b	$AF,$F1,$8A,$D6,$1F,$A7,$F2,$FF,$6A,$66,$A6,$3F,$8B,$8C,$BF,$6E
	.dc.b	$B8,$BF,$C7,$3E,$6D,$F0,$DF,$C9,$1B,$1F,$81,$6B,$F9,$EF,$3B,$53
	.dc.b	$FC,$CB,$E6,$1F,$39,$4C,$D7,$19,$C3,$FA,$29,$DB,$93,$F6,$5D,$AF
	.dc.b	$80,$FD,$EB,$63,$F9,$6B,$5F,$E6,$8D,$7F,$9C,$B9,$E0,$FE,$A5,$4F
	.dc.b	$9C,$F3,$1F,$C2,$A7,$5B,$FF,$17,$97,$CE,$7F,$4E,$D1,$F3,$9F,$D6
	.dc.b	$31,$5F,$FA,$26,$30,$F1,$AF,$84,$71,$8C,$E4,$BE,$75,$A8,$FE,$47
	.dc.b	$0F,$9C,$FE,$A3,$87,$69,$E1,$18,$AF,$F4,$C3,$63,$F9,$56,$2F,$91
	.dc.b	$F4,$4E,$70,$E9,$58,$3E,$B4,$FF,$2A,$F8,$47,$8F,$49,$3F,$C8,$FE
	.dc.b	$23,$F8,$D3,$FC,$96,$76,$FF,$5E,$6B,$FD,$95,$AF,$F6,$A6,$BF,$EA
	.dc.b	$17,$79,$FF,$80,$A5,$E6,$7F,$3C,$6D,$77,$BF,$E7,$FF,$36,$F8,$AA
	.dc.b	$49,$FE,$71,$F3,$9F,$35,$FE,$75,$3B,$7F,$B9,$B5,$FE,$F4,$D7,$FB
	.dc.b	$FB,$5F,$F5,$B7,$CD,$BE,$2F,$A4,$7F,$81,$23,$1B,$7F,$7E,$6C,$7F
	.dc.b	$81,$5A,$FF,$82,$BD,$17,$8E,$F9,$CF,$EA,$AD,$9F,$F0,$64,$EA,$72
	.dc.b	$48,$7C,$5F,$A6,$75,$18,$D3,$FD,$A7,$7C,$EB,$4D,$8F,$F0,$C7,$C0
	.dc.b	$7F,$66,$A3,$48,$FF,$6A,$6B,$FE,$1A,$6B,$C8,$D1,$0D,$14,$D1,$82
	.dc.b	$18,$21,$FE,$2E,$7A,$B4,$E4,$53,$89,$4E,$27,$FB,$AD,$E2,$D3,$89
	.dc.b	$4E,$25,$38,$83,$E1,$9C,$0C,$3D,$1A,$64,$D9,$D0,$C9,$4C,$9D,$8A
	.dc.b	$7A,$48,$78,$CA,$72,$38,$3E,$17,$89,$B2,$1B,$38,$9A,$30,$68,$C1
	.dc.b	$1E,$84,$7D,$8D,$7D,$4D,$7C,$66,$88,$68,$A6,$8C,$10,$C1,$0F,$3B
	.dc.b	$D5,$A7,$22,$9C,$4A,$71,$3D,$8F,$16,$9C,$4A,$71,$29,$C4,$F8,$A4
	.dc.b	$3E,$21,$C4,$C3,$D5,$A6,$4D,$9D,$4C,$94,$C9,$DC,$A7,$A8,$87,$90
	.dc.b	$A7,$23,$83,$E1,$78,$9B,$21,$B3,$89,$A3,$06,$8C,$11,$E8,$47,$D8
	.dc.b	$D7,$D4,$D7,$C6,$68,$86,$8A,$68,$C1,$0C,$10,$F3,$BD,$5A,$72,$29
	.dc.b	$C4,$A7,$13,$D8,$F1,$69,$C4,$A7,$12,$9C,$4F,$8A,$43,$E2,$1C,$4C
	.dc.b	$3D,$5A,$64,$D9,$D4,$C9,$4C,$9D,$CA,$7A,$88,$7E,$EE,$DE,$6B,$8B
	.dc.b	$FD,$32,$D9,$FE,$8A,$D7,$FA,$63,$5F,$E3,$42,$79,$CF,$22,$47,$26
	.dc.b	$BC,$9A,$F2,$8E,$2E,$C8,$6C,$A6,$E7,$C4,$FF,$2B,$8E,$2E,$C8,$6E
	.dc.b	$7B,$BB,$9E,$2F,$A5,$B3,$CA,$38,$1E,$03,$84,$F2,$78,$36,$38,$46
	.dc.b	$8E,$06,$8A,$E9,$C1,$E6,$38,$14,$C9,$E2,$30,$F1,$24,$FA,$09,$3E
	.dc.b	$36,$CF,$81,$AF,$26,$BD,$48,$78,$8A,$78,$8C,$10,$C1,$27,$C7,$3E
	.dc.b	$12,$47,$85,$AF,$85,$AF,$68,$E2,$EC,$86,$CA,$6E,$7D,$2F,$8E,$38
	.dc.b	$BB,$21,$B9,$EE,$EE,$78,$BE,$96,$CF,$28,$E0,$78,$0E,$13,$C9,$E0
	.dc.b	$D8,$E1,$1A,$38,$1A,$2B,$A7,$07,$98,$E0,$53,$27,$88,$C3,$C4,$93
	.dc.b	$E8,$24,$F8,$DB,$3E,$06,$BC,$9A,$F5,$21,$E2,$29,$E2,$30,$43,$04
	.dc.b	$9F,$1C,$F8,$49,$1E,$16,$BE,$16,$BD,$A3,$8B,$B2,$1B,$29,$B9,$F4
	.dc.b	$BE,$38,$E2,$EC,$86,$E7,$BB,$B9,$E2,$FA,$5B,$3C,$A3,$81,$E0,$38
	.dc.b	$4F,$27,$83,$63,$84,$68,$E0,$68,$AE,$9C,$1E,$63,$81,$4C,$9E,$23
	.dc.b	$0F,$12,$4F,$A0,$93,$E3,$C3,$FC,$E5,$5F,$C9,$5A,$FF,$37,$6D,$7F
	.dc.b	$9F,$31,$FF,$3C,$E5,$FE,$81,$AF,$F4,$55,$7F,$A1,$DB,$19,$7F,$D0
	.dc.b	$2D,$7B,$35,$EC,$DA,$FF,$44,$C7,$23,$FD,$23,$3D,$5E,$66,$27,$84
	.dc.b	$66,$3A,$99,$21,$CC,$87,$22,$9C,$0A,$79,$0E,$26,$C8,$6C,$A6,$CE
	.dc.b	$33,$EA,$72,$D7,$81,$0E,$06,$A3,$D4,$E0,$D4,$71,$70,$47,$04,$8E
	.dc.b	$2D,$9F,$8A,$49,$F5,$9D,$27,$64,$8D,$47,$AA,$30,$F0,$3C,$0E,$1D
	.dc.b	$47,$07,$D4,$D7,$B3,$5E,$2D,$9F,$34,$75,$6B,$D0,$C4,$F7,$7A,$98
	.dc.b	$9E,$51,$98,$EE,$64,$87,$32,$1C,$8A,$70,$29,$E6,$38,$9B,$21,$B2
	.dc.b	$9B,$38,$CF,$A9,$CB,$5E,$04,$38,$1A,$8F,$53,$83,$51,$C5,$C1,$1C
	.dc.b	$12,$38,$B6,$7E,$29,$27,$D6,$74,$9D,$92,$35,$1E,$A8,$C3,$C0,$F0
	.dc.b	$38,$75,$1C,$1F,$53,$5E,$CD,$78,$B6,$7C,$D1,$D5,$AF,$43,$13,$DD
	.dc.b	$EA,$62,$79,$46,$63,$B9,$92,$1C,$C8,$72,$29,$C0,$A7,$98,$E2,$6C
	.dc.b	$86,$CA,$6C,$E3,$3E,$A7,$2D,$78,$10,$E0,$6A,$3D,$4E,$0D,$47,$17
	.dc.b	$04,$70,$48,$E2,$D9,$F8,$A4,$9F,$59,$D2,$76,$48,$D4,$7A,$A3,$0F
	.dc.b	$03,$C0,$E1,$D4,$70,$7D,$4D,$7B,$35,$E2,$D9,$F3,$47,$56,$BD,$0C
	.dc.b	$4F,$77,$A9,$89,$E5,$19,$8E,$E6,$48,$73,$21,$C8,$A7,$02,$9E,$63
	.dc.b	$89,$B2,$1B,$29,$B3,$8C,$FA,$9C,$B5,$E0,$43,$81,$A8,$F5,$38,$35
	.dc.b	$1C,$5C,$11,$C1,$23,$8B,$67,$E2,$92,$7D,$67,$49,$D9,$23,$51,$EA
	.dc.b	$8C,$3C,$0F,$03,$87,$51,$C1,$F5,$35,$EC,$D7,$DF,$F9,$70,$80

;--------------------------------------------------------------------------------
;	FSGLMUL.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68060
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsglmul060_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#48,-(sp)
	peamsg	'FSGLMUL.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSGLMUL.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsglmul060_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsglmul.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSGLMUL.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsglmul.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSGLMUL.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsglmul060_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3A,$F9,$AF,$EC,$ED,$2D,$FC,$E8,$CC,$76,$B4,$62
	.dc.b	$75,$2F,$A0,$F5,$BE,$9C,$FF,$BA,$46,$23,$CF,$18,$CF,$D7,$87,$49
	.dc.b	$E5,$18,$C3,$FB,$ED,$F1,$C3,$DD,$5A,$3B,$1A,$AE,$0E,$4F,$45,$7F
	.dc.b	$81,$BE,$7F,$96,$3F,$E0,$76,$CE,$6B,$8C,$77,$C7,$C9,$8E,$AB,$A6
	.dc.b	$3E,$28,$D6,$3E,$19,$F9,$51,$B2,$61,$E4,$7E,$2F,$CA,$3E,$17,$F5
	.dc.b	$75,$FE,$1A,$BF,$E0,$D3,$9F,$17,$FC,$2B,$F0,$9F,$E1,$E8,$C5,$6E
	.dc.b	$31,$7F,$E0,$E8,$C0,$9D,$BF,$62,$77,$5D,$23,$55,$E2,$74,$63,$E0
	.dc.b	$75,$7E,$7F,$81,$B4,$69,$8F,$4E,$DF,$8A,$B8,$36,$B8,$36,$7C,$5F
	.dc.b	$37,$FC,$13,$C0,$FC,$7B,$67,$F2,$AD,$7F,$C7,$AD,$7F,$C7,$FC,$DF
	.dc.b	$DB,$F5,$FD,$C3,$75,$FD,$9F,$37,$BE,$93,$FA,$9A,$FE,$69,$F3,$8F
	.dc.b	$C5,$BF,$3F,$48,$C3,$FA,$26,$BF,$A8,$6B,$FE,$50,$6B,$FE,$49,$8E
	.dc.b	$71,$88,$FF,$2B,$B5,$FA,$3E,$B7,$CC,$23,$F2,$7F,$27,$FE,$0D,$7F
	.dc.b	$D2,$73,$FE,$53,$CD,$F1,$BC,$CF,$E4,$77,$F1,$9A,$FC,$66,$BF,$E5
	.dc.b	$06,$BF,$E6,$89,$E6,$F1,$9E,$87,$F0,$0E,$1F,$CB,$7A,$FF,$80,$24
	.dc.b	$FD,$BF,$07,$F1,$58,$BF,$6F,$FF,$2B,$4F,$F0,$C8,$C7,$81,$FC,$29
	.dc.b	$B1,$99,$DB,$FE,$78,$8C,$B6,$32,$FF,$9D,$9A,$FE,$9D,$AF,$E7,$F4
	.dc.b	$7B,$DF,$98,$6C,$7C,$B8,$F3,$CB,$E1,$BF,$09,$E2,$BC,$3F,$C0,$B6
	.dc.b	$FF,$90,$B5,$FE,$4E,$D7,$FD,$1E,$D7,$FA,$11,$C2,$39,$9F,$2F,$A7
	.dc.b	$FA,$A4,$7E,$8F,$CB,$FD,$59,$F9,$5F,$BB,$7C,$87,$93,$97,$C6,$4A
	.dc.b	$F1,$B5,$F1,$B5,$F1,$BD,$FC,$8F,$DB,$92,$3F,$82,$FC,$DB,$F4,$DF
	.dc.b	$EB,$DC,$3F,$D7,$FD,$0F,$E0,$55,$FE,$AA,$6B,$FE,$AD,$6B,$FD,$BB
	.dc.b	$C3,$FD,$3C,$62,$0D,$24,$7C,$1F,$2F,$F8,$49,$27,$F8,$BF,$53,$F2
	.dc.b	$07,$B3,$FC,$3B,$92,$57,$F0,$97,$F8,$8F,$C5,$FE,$EF,$27,$C6,$78
	.dc.b	$89,$5F,$D6,$23,$2F,$F5,$EE,$DF,$E6,$1D,$7F,$C6,$89,$1F,$C5,$79
	.dc.b	$FF,$83,$7D,$9F,$F8,$6E,$3F,$B2,$B5,$FF,$66,$B5,$FF,$68,$35,$FF
	.dc.b	$69,$B5,$34,$9E,$87,$E0,$FE,$0B,$F9,$59,$27,$F9,$C9,$FE,$1E,$F2
	.dc.b	$FF,$98,$FF,$00,$7F,$A3,$63,$F9,$62,$F4,$32,$6B,$E6,$E4,$EA,$32
	.dc.b	$F0,$8E,$7F,$07,$D3,$E0,$E4,$F4,$8C,$BD,$A3,$2F,$F4,$4D,$98,$7F
	.dc.b	$93,$63,$CF,$F6,$7F,$E3,$D7,$FD,$E2,$D7,$FD,$E8,$D7,$FD,$EE,$D4
	.dc.b	$D2,$7C,$CF,$F3,$27,$5E,$97,$F3,$7F,$43,$FB,$15,$FF,$89,$2F,$C8
	.dc.b	$DE,$CF,$F1,$C5,$6E,$7F,$B1,$4F,$F4,$36,$FA,$1F,$30,$F9,$BF,$21
	.dc.b	$63,$E6,$C5,$7A,$C7,$D2,$F5,$FE,$9F,$9D,$27,$E9,$C6,$5F,$AB,$19
	.dc.b	$7C,$B1,$FD,$47,$E2,$1E,$2F,$F4,$E6,$CF,$F8,$B9,$AF,$F8,$D1,$A8
	.dc.b	$FF,$41,$9B,$8E,$44,$7F,$98,$7E,$47,$FA,$A7,$FC,$07,$A2,$31,$1F
	.dc.b	$DF,$A3,$1D,$FF,$D9,$7F,$18,$FB,$93,$F3,$3A,$7F,$23,$F3,$B2,$3F
	.dc.b	$23,$1F,$94,$30,$FD,$28,$C3,$F9,$6F,$A1,$FC,$C9,$1F,$F0,$27,$E9
	.dc.b	$7F,$CD,$1F,$B3,$19,$69,$AF,$47,$FB,$C7,$C9,$FF,$9F,$63,$FC,$AE
	.dc.b	$D7,$E2,$1F,$EB,$7F,$9D,$FE,$F3,$19,$8F,$F5,$C4,$60,$8F,$F9,$72
	.dc.b	$34,$FE,$1B,$E2,$BF,$5E,$47,$F6,$7F,$28,$F9,$3F,$D9,$FC,$84,$9F
	.dc.b	$D9,$BF,$E2,$6F,$D3,$FC,$DC,$3F,$E3,$AE,$94,$EE,$BF,$2E,$EF,$3F
	.dc.b	$DE,$FE,$8F,$27,$F7,$AF,$F0,$0F,$47,$11,$F9,$E8,$B8,$75,$9F,$F1
	.dc.b	$4F,$5B,$E5,$7F,$C5,$2D,$BF,$F1,$4B,$5F,$F1,$4B,$5F,$93,$E8,$3F
	.dc.b	$33,$FE,$E5,$E0,$7F,$81,$5B,$29,$D3,$FC,$53,$E3,$3F,$8A,$FC,$A4
	.dc.b	$8F,$E2,$AF,$F1,$8E,$BC,$47,$ED,$A2,$E9,$F1,$67,$FC,$83,$F2,$3F
	.dc.b	$EA,$FA,$F1,$B5,$FF,$0B,$35,$F1,$37,$C6,$FE,$27,$F1,$DF,$C9,$89
	.dc.b	$3F,$C9,$BE,$47,$FC,$BD,$EC,$BF,$61,$FE,$55,$3F,$88,$11,$9F,$37
	.dc.b	$FD,$3E,$D8,$FF,$52,$B5,$E1,$83,$F8,$07,$E2,$B6,$3F,$CC,$F1,$83
	.dc.b	$AB,$FE,$63,$70,$E8,$DC,$D7,$45,$3F,$98,$47,$27,$FC,$EB,$1E,$BF
	.dc.b	$1F,$CF,$87,$FA,$92,$7F,$47,$39,$7D,$31,$97,$FD,$15,$5F,$E4,$CE
	.dc.b	$2B,$F4,$9E,$83,$D3,$E6,$FF,$99,$FE,$83,$FB,$6E,$9F,$C4,$7E,$AB
	.dc.b	$FD,$AB,$8A,$72,$A3,$31,$87,$FD,$B7,$CA,$FB,$3F,$E2,$31,$AF,$6B
	.dc.b	$FA,$E3,$70,$ED,$E9,$7E,$F3,$0E,$4F,$F9,$5E,$F8,$B9,$BE,$2D,$AE
	.dc.b	$38,$FF,$27,$BE,$2F,$E6,$6F,$8B,$9B,$E2,$DA,$E2,$FB,$63,$38,$FA
	.dc.b	$79,$7F,$87,$E5,$E7,$AF,$81,$97,$85,$A7,$32,$D7,$17,$AE,$5E,$47
	.dc.b	$B6,$38,$C7,$BE,$5E,$69,$C9,$D0,$B8,$F8,$4F,$9B,$1B,$BE,$84,$AE
	.dc.b	$79,$7A,$32,$F3,$5F,$16,$9F,$36,$35,$87,$A6,$74,$D9,$D3,$E7,$C6
	.dc.b	$3E,$5C,$3E,$7E,$5E,$F8,$C6,$3F,$39,$F4,$DF,$81,$FA,$98,$F8,$1F
	.dc.b	$79,$FB,$5B,$FC,$6B,$6B,$F7,$2F,$CF,$31,$87,$EE,$DF,$A0,$60,$C4
	.dc.b	$F2,$7C,$07,$88,$F6,$1E,$23,$D7,$8F,$C0,$75,$8F,$AE,$31,$3F,$36
	.dc.b	$BF,$1F,$39,$C7,$AC,$E6,$FF,$03,$97,$D5,$C7,$E1,$4E,$71,$FB,$39
	.dc.b	$7C,$18,$C6,$1F,$17,$1E,$98,$F9,$8F,$41,$E4,$C7,$A6,$5E,$8A,$E1
	.dc.b	$18,$9F,$A3,$39,$7D,$C6,$09,$8F,$DE,$BE,$F1,$DB,$0F,$1B,$C2,$7E
	.dc.b	$0E,$5E,$A3,$F1,$B1,$FB,$0E,$2F,$E4,$9A,$FF,$0B,$8F,$E4,$0F,$F0
	.dc.b	$C7,$FD,$23,$C1,$FC,$4D,$FE,$62,$DE,$1F,$E3,$11,$FC,$D5,$AF,$F7
	.dc.b	$EA,$FE,$5C,$FF,$26,$21,$FC,$9A,$3F,$94,$73,$38,$7F,$8F,$CE,$1F
	.dc.b	$E4,$33,$61,$5D,$1F,$E5,$E7,$F3,$43,$CF,$AF,$F2,$E2,$3F,$CB,$1F
	.dc.b	$F7,$37,$80,$E6,$FE,$A3,$3F,$E5,$47,$FC,$F3,$CD,$FD,$7B,$1F,$CB
	.dc.b	$4F,$F5,$A6,$F0,$39,$FF,$AF,$36,$3A,$F4,$7F,$68,$7F,$CC,$2F,$F6
	.dc.b	$DA,$FE,$DE,$43,$FB,$71,$FE,$86,$E5,$FF,$12,$ED,$F0,$E3,$FB,$D7
	.dc.b	$57,$F7,$CC,$3F,$BD,$C7,$FA,$CA,$3F,$AD,$E8,$EA,$FF,$06,$F6,$7F
	.dc.b	$83,$09,$5F,$E0,$C7,$E0,$F7,$7F,$85,$34,$FF,$3A,$BF,$E1,$99,$FF
	.dc.b	$0D,$90,$FF,$0D,$1F,$E9,$AE,$EF,$F1,$26,$7F,$E2,$7F,$02,$DF,$F9
	.dc.b	$7F,$8D,$E0,$FF,$8B,$23,$FD,$03,$3A,$C3,$FC,$FA,$FF,$8A,$DF,$F5
	.dc.b	$2F,$89,$FE,$39,$7F,$D0,$AF,$C2,$9F,$F2,$3F,$89,$FE,$3B,$9F,$F5
	.dc.b	$7F,$8B,$FE,$42,$4D,$27,$89,$FE,$48,$8F,$F2,$8B,$5F,$F0,$54,$7F
	.dc.b	$92,$B2,$FF,$52,$F8,$FF,$E5,$BD,$3F,$C8,$4F,$86,$BF,$1F,$E4,$7F
	.dc.b	$98,$67,$31,$FE,$67,$21,$FE,$AF,$8D,$65,$FE,$A0,$F2,$BF,$CD,$F5
	.dc.b	$FE,$4D,$F2,$BF,$CD,$F7,$FE,$77,$3F,$D8,$AE,$AB,$84,$7F,$A0,$BC
	.dc.b	$CF,$F4,$0D,$7E,$0C,$FF,$40,$C6,$7A,$3F,$D0,$5E,$9F,$FA,$5F,$CE
	.dc.b	$FF,$46,$4F,$FA,$4D,$AF,$F8,$A7,$CF,$FF,$4A,$BF,$E9,$43,$47,$FA
	.dc.b	$54,$FF,$78,$74,$BC,$BF,$F5,$26,$3F,$E6,$5F,$43,$FD,$51,$5F,$EA
	.dc.b	$63,$FD,$52,$43,$07,$FA,$FB,$CD,$FF,$5A,$4F,$F9,$B5,$FF,$5C,$C7
	.dc.b	$FA,$E0,$86,$CF,$F7,$3F,$A3,$F4,$57,$A5,$FE,$C5,$E4,$FF,$6D,$FC
	.dc.b	$77,$FB,$B8,$D4,$7F,$B2,$67,$FD,$AB,$E7,$FF,$B4,$C8,$FE,$EB,$D4
	.dc.b	$FF,$6B,$B6,$32,$FF,$53,$8F,$F7,$05,$FF,$BA,$38,$3F,$DD,$79,$FF
	.dc.b	$A1,$7D,$0F,$F7,$7D,$FF,$B6,$A7,$3C,$9F,$ED,$E9,$CD,$7F,$A2,$FC
	.dc.b	$9F,$F7,$DD,$FF,$B8,$3D,$5F,$F7,$FB,$FE,$E2,$8C,$CF,$FA,$46,$72
	.dc.b	$FF,$01,$C5,$F0,$AF,$63,$F2,$91,$86,$C6,$70,$FF,$68,$57,$F0,$95
	.dc.b	$C4,$F6,$3E,$67,$B2,$EC,$FE,$3B,$48,$C3,$FC,$81,$AF,$F7,$76,$BF
	.dc.b	$E7,$36,$BF,$EB,$C6,$BF,$EE,$16,$BD,$63,$F9,$5E,$5E,$CF,$6F,$D2
	.dc.b	$49,$F4,$EF,$FC,$2F,$37,$B7,$F8,$3F,$71,$EA,$FA,$1B,$7F,$CD,$9A
	.dc.b	$FF,$83,$5A,$FF,$53,$A7,$BD,$F2,$09,$1E,$03,$FA,$6C,$3D,$EF,$9B
	.dc.b	$F0,$12,$75,$A7,$D4,$8F,$E0,$DE,$8B,$E0,$3F,$6F,$AF,$FB,$7B,$37
	.dc.b	$C0,$F9,$C9,$8F,$9D,$AF,$9D,$AF,$FA,$7E,$38,$BF,$E2,$7C,$7F,$07
	.dc.b	$F0,$5F,$80,$24,$7A,$7D,$97,$A3,$F8,$1E,$7F,$C1,$7C,$37,$F0,$0A
	.dc.b	$07,$0D,$F1,$FF,$07,$18,$23,$82,$47,$8F,$85,$F0,$DF,$92,$24,$7C
	.dc.b	$B1,$F0,$CF,$86,$FD,$E6,$1F,$72,$7F,$B6,$B5,$FF,$22,$35,$FF,$60
	.dc.b	$35,$F4,$9E,$9E,$7F,$39,$1E,$6F,$E3,$BE,$23,$F9,$9C,$7F,$82,$1B
	.dc.b	$B9,$CE,$8F,$88,$FC,$2D,$7E,$0E,$31,$D5,$FE,$02,$6C,$7F,$96,$5A
	.dc.b	$FF,$B5,$5A,$FF,$7A,$94,$F1,$6C,$71,$6B,$C5,$AF,$E5,$FE,$2B,$F7
	.dc.b	$A4,$7F,$23,$B3,$E2,$BF,$23,$F1,$7F,$BE,$9F,$F3,$67,$4A,$31,$1F
	.dc.b	$EE,$98,$C7,$83,$FE,$1A,$9C,$CF,$CC,$9C,$BC,$A3,$2F,$E2,$23,$2F
	.dc.b	$38,$CB,$FA,$8E,$3F,$E2,$64,$7F,$4F,$AF,$FB,$65,$7C,$63,$E3,$7F
	.dc.b	$88,$B6,$3F,$C5,$ED,$7F,$D0,$CD,$7F,$02,$72,$8E,$24,$78,$BF,$C7
	.dc.b	$17,$C7,$3C,$FF,$E9,$8D,$8F,$CB,$35,$70,$FF,$29,$87,$9F,$F3,$BC
	.dc.b	$FF,$90,$4A,$F8,$CF,$F1,$CE,$0F,$EA,$F1,$F5,$BC,$FF,$E6,$A7,$F3
	.dc.b	$6F,$8F,$B1,$92,$1E,$0F,$07,$F9,$04,$7F,$3C,$F0,$64,$FE,$01,$FE
	.dc.b	$7F,$E4,$E2,$38,$1F,$C6,$23,$1A,$7F,$49,$C7,$FA,$04,$FE,$CF,$E4
	.dc.b	$BF,$A1,$CF,$F0,$2F,$91,$7C,$97,$EB,$27,$FC,$01,$59,$8C,$7A,$9F
	.dc.b	$B0,$AD,$CF,$F0,$AA,$E6,$A3,$2A,$33,$18,$7A,$C6,$EB,$AB,$DC,$E9
	.dc.b	$D1,$F4,$9D,$18,$C3,$F8,$CF,$9F,$CF,$1F,$E0,$6D,$9C,$DF,$AE,$F9
	.dc.b	$4F,$85,$E9,$8E,$EB,$A9,$E7,$BF,$75,$F1,$C3,$C9,$3E,$BB,$E3,$3E
	.dc.b	$88,$E3,$8F,$B3,$0E,$D7,$C2,$3E,$2C,$FE,$61,$AF,$53,$FD,$C4,$E1
	.dc.b	$F9,$9E,$6B,$EB,$FF,$D2,$33,$FD,$DD,$B1,$99,$E2,$D7,$B6,$3F,$25
	.dc.b	$B7,$CA,$31,$1E,$13,$27,$89,$CB,$F3,$CC,$47,$12,$38,$21,$C8,$87
	.dc.b	$52,$1C,$CC,$94,$C9,$F3,$5E,$AE,$8C,$1D,$8D,$DF,$C7,$72,$6D,$C1
	.dc.b	$27,$D7,$1A,$38,$B8,$8D,$4F,$82,$72,$D7,$CC,$43,$91,$63,$8B,$E5
	.dc.b	$31,$3D,$49,$3C,$E3,$89,$27,$8B,$E9,$9D,$BC,$CD,$38,$75,$3C,$E7
	.dc.b	$8C,$7A,$0E,$0E,$1F,$1C,$6E,$79,$46,$CE,$67,$07,$2F,$03,$11,$C4
	.dc.b	$8E,$08,$72,$21,$D4,$87,$33,$25,$32,$7B,$9E,$AE,$8C,$1D,$8D,$CF
	.dc.b	$B5,$C9,$B7,$04,$9F,$5C,$68,$E2,$E2,$35,$3E,$09,$CB,$5F,$31,$0E
	.dc.b	$45,$8E,$2F,$94,$C4,$F5,$24,$F3,$8E,$24,$9E,$2F,$A6,$76,$F3,$34
	.dc.b	$E1,$D4,$F3,$9E,$31,$E8,$38,$38,$7C,$71,$B9,$E5,$1B,$39,$9C,$1C
	.dc.b	$BC,$0C,$47,$12,$38,$21,$C8,$87,$52,$1C,$CC,$94,$C9,$EE,$7A,$BA
	.dc.b	$30,$76,$37,$3E,$D7,$26,$DC,$12,$7D,$71,$A3,$8B,$88,$D4,$F8,$27
	.dc.b	$2D,$7C,$C4,$39,$1F,$CD,$AF,$FC,$D7,$F4,$4E,$AF,$ED,$58,$FF,$35
	.dc.b	$8C,$56,$C3,$87,$FB,$34,$6E,$78,$D7,$46,$CF,$06,$D7,$1C,$3F,$A2
	.dc.b	$4E,$4F,$F1,$2F,$C7,$7F,$88,$48,$FF,$AF,$9A,$FF,$8A,$8F,$F4,$1B
	.dc.b	$87,$0D,$79,$9C,$48,$71,$0E,$DF,$F3,$AB,$5E,$A7,$03,$A9,$4E,$A7
	.dc.b	$37,$83,$D0,$D1,$D0,$A7,$43,$81,$83,$81,$82,$1E,$02,$9E,$03,$04
	.dc.b	$30,$43,$C0,$70,$29,$B2,$9B,$3C,$86,$4A,$64,$F2,$99,$29,$97,$E0
	.dc.b	$91,$F6,$35,$F1,$1D,$0F,$09,$0C,$10,$EA,$64,$A6,$5E,$8F,$8D,$AF
	.dc.b	$8C,$F0,$90,$F0,$9C,$88,$F0,$7B,$35,$EC,$71,$3B,$14,$EC,$74,$78
	.dc.b	$3D,$0D,$1D,$0A,$74,$38,$18,$38,$18,$21,$E0,$29,$E0,$30,$43,$04
	.dc.b	$3C,$07,$02,$9B,$29,$B3,$C8,$64,$A6,$4F,$29,$92,$99,$7E,$09,$1F
	.dc.b	$63,$5F,$11,$D0,$F0,$90,$C1,$0E,$A6,$4A,$65,$E8,$F8,$DA,$F8,$CF
	.dc.b	$09,$0F,$09,$C8,$8F,$07,$B3,$5E,$C7,$13,$B1,$4E,$C7,$47,$83,$D0
	.dc.b	$D1,$D0,$A7,$43,$81,$83,$81,$82,$1E,$02,$9E,$03,$04,$30,$43,$C0
	.dc.b	$70,$29,$B2,$9B,$3C,$86,$4A,$7F,$0B,$8F,$E6,$EF,$FA,$0F,$0F,$F3
	.dc.b	$C3,$FC,$F8,$C6,$1F,$D4,$5F,$F1,$0E,$7F,$E9,$72,$69,$FE,$89,$7F
	.dc.b	$D1,$F1,$FE,$92,$D7,$FD,$32,$FF,$AD,$5A,$FF,$9A,$A3,$7B,$7F,$A8
	.dc.b	$49,$5F,$E9,$37,$9C,$76,$9E,$47,$FA,$6F,$6F,$F5,$0B,$FE,$85,$8D
	.dc.b	$3F,$EA,$93,$5B,$47,$C8,$7F,$A8,$F0,$F1,$E7,$FE,$C4,$C7,$D4,$7F
	.dc.b	$AA,$EF,$9C,$66,$BB,$4E,$48,$E0,$87,$84,$EE,$7F,$AD,$30,$F3,$BF
	.dc.b	$E8,$AC,$7D,$EF,$B0,$3F,$12,$7C,$71,$9B,$F1,$D7,$2B,$FF,$6F,$DF
	.dc.b	$2A,$F0,$3C,$5F,$F5,$43,$98,$FF,$6F,$B9,$8F,$A5,$7E,$A7,$D1,$7F
	.dc.b	$42,$3E,$91,$FE,$CD,$9E,$57,$FE,$DA,$8C,$D7,$A6,$73,$8F,$FB,$82
	.dc.b	$3E,$1E,$1F,$ED,$3A,$EC,$FC,$B8,$F8,$51,$D4,$8F,$A1,$EA,$FB,$1C
	.dc.b	$BC,$0E,$B8,$FD,$A3,$ED,$E1,$F1,$4F,$34,$79,$67,$8D,$7C,$87,$AB
	.dc.b	$F1,$1C,$CF,$09,$F8,$A7,$D1,$8F,$4C,$75,$BF,$74,$F9,$4E,$C7,$A4
	.dc.b	$86,$49,$5E,$17,$D2,$7E,$4C,$FC,$B6,$3F,$56,$3C,$91,$D0,$F8,$04
	.dc.b	$72,$F4,$AF,$44,$7C,$D3,$E8,$47,$A4,$FA,$71,$F8,$9C,$3E,$44,$78
	.dc.b	$30,$F8,$70,$8F,$6D,$76,$3D,$0F,$C4,$3F,$0F,$3F,$13,$0F,$D7,$63
	.dc.b	$F8,$FB,$E5,$5F,$9C,$AF,$9F,$87,$9A,$73,$5F,$88,$AD,$1F,$10,$FD
	.dc.b	$4E,$3F,$93,$70,$FE,$8B,$0F,$44,$7B,$4F,$19,$EE,$7E,$63,$CB,$0F
	.dc.b	$D5,$BE,$9C,$7E,$2B,$C0,$C6,$3F,$15,$E3,$7E,$17,$86,$3F,$21,$FD
	.dc.b	$E6,$1E,$27,$95,$F4,$3E,$95,$F3,$8F,$11,$E5,$29,$C1,$FC,$63,$88
	.dc.b	$E5,$18,$BF,$E2,$77,$F9,$EB,$FA,$D0,$C3,$EC,$78,$4F,$D5,$9F,$87
	.dc.b	$7F,$99,$63,$0D,$8C,$E7,$FE,$71,$9C,$B6,$32,$FD,$A1,$1A,$70,$D7
	.dc.b	$EF,$B5,$F1,$FC,$9B,$EA,$7F,$07,$93,$A3,$31,$86,$D7,$E4,$67,$4F
	.dc.b	$83,$0F,$1C,$63,$06,$0F,$27,$95,$F9,$A3,$0F,$3C,$3C,$73,$AA,$7E
	.dc.b	$CC,$F6,$71,$87,$0C,$18,$F7,$8C,$37,$0E,$57,$C6,$BB,$1E,$1D,$79
	.dc.b	$91,$CE,$3F,$57,$DC,$C4,$76,$CF,$E3,$D6,$31,$F7,$68,$BF,$47,$EA
	.dc.b	$EF,$80,$FB,$8D,$C7,$ED,$35,$FA,$AD,$7E,$AB,$5F,$C3,$B5,$FB,$AD
	.dc.b	$C8,$DD,$7E,$8E,$7F,$73,$83,$E3,$9E,$4C,$3E,$4F,$D2,$CC,$7D,$F6
	.dc.b	$8F,$A1,$FA,$C3,$E5,$35,$5F,$04,$FB,$13,$8D,$BE,$F5,$6E,$79,$FC
	.dc.b	$57,$D6,$23,$F3,$3F,$6B,$7C,$DE,$23,$A6,$FF,$86,$AC,$61,$F8,$9B
	.dc.b	$DD,$7D,$4F,$DC,$1F,$3F,$FA,$76,$C7,$E9,$5A,$FE,$81,$AF,$F1,$EF
	.dc.b	$84,$7E,$E2,$F9,$7F,$C9,$C6,$1B,$19,$9F,$05,$FE,$4F,$D6,$FD,$D4
	.dc.b	$7E,$19,$AF,$A9,$B9,$7E,$93,$0E,$93,$BE,$8E,$44,$7F,$27,$1A,$7E
	.dc.b	$1F,$EE,$7E,$87,$F8,$3B,$5F,$DE,$B7,$63,$78,$FE,$0E,$1F,$A3,$CB
	.dc.b	$F4,$E7,$EA,$E7,$3C,$DF,$C4,$BF,$77,$78,$1F,$C4,$5B,$97,$F0,$F6
	.dc.b	$BF,$C2,$9A,$FF,$0A,$C5,$F2,$5F,$B4,$24,$7E,$AA,$78,$E5,$FB,$59
	.dc.b	$CC,$FE,$02,$B8,$BB,$BC,$D6,$E3,$F2,$B5,$FC,$0E,$F9,$E1,$FC,$AB
	.dc.b	$F7,$B7,$99,$FC,$A5,$B7,$FC,$9D,$AF,$F2,$26,$BE,$9F,$B2,$FE,$16
	.dc.b	$47,$F8,$0C,$70,$C7,$F8,$EC,$3C,$57,$9D,$FC,$BE,$3F,$B4,$C6,$1C
	.dc.b	$1B,$8A,$D3,$31,$B7,$93,$5F,$E7,$4D,$7A,$C7,$F4,$5F,$39,$F1,$7F
	.dc.b	$C9,$5B,$3F,$C6,$1A,$FE,$DC,$F0,$17,$D1,$FE,$14,$9A,$4F,$03,$CE
	.dc.b	$D8,$F5,$B5,$FE,$9D,$F7,$4F,$D6,$5F,$51,$FC,$F2,$30,$D8,$CC,$FF
	.dc.b	$30,$3C,$17,$99,$91,$FC,$D2,$B8,$BB,$BF,$E4,$DA,$7D,$2F,$A6,$7C
	.dc.b	$87,$CC,$6D,$7D,$06,$BF,$55,$AF,$F6,$56,$BF,$E2,$06,$BF,$E2,$B6
	.dc.b	$A7,$32,$F9,$4F,$1F,$1F,$F6,$16,$D7,$F5,$26,$BC,$6F,$FA,$FF,$A5
	.dc.b	$CF,$8F,$F0,$AD,$9F,$C5,$35,$FE,$F4,$D7,$FC,$6A,$F8,$FC,$3E,$39
	.dc.b	$1A,$49,$FC,$8F,$D2,$3E,$45,$DF,$86,$CF,$E6,$67,$53,$93,$8F,$D6
	.dc.b	$BE,$32,$92,$B8,$46,$5F,$3F,$E4,$BF,$CD,$37,$F5,$B6,$33,$E0,$FE
	.dc.b	$A9,$FF,$2C,$6A,$9D,$47,$B9,$FF,$11,$D7,$2F,$37,$84,$AB,$84,$61
	.dc.b	$B3,$FB,$3F,$82,$7A,$EF,$A1,$FF,$0A,$F2,$FF,$85,$A7,$31,$8A,$FF
	.dc.b	$0D,$64,$F8,$9C,$66,$BD,$D7,$F8,$23,$0F,$45,$79,$1B,$1F,$C0,$DB
	.dc.b	$5F,$E3,$36,$BF,$E2,$CB,$ED,$3B,$6C,$6D,$AE,$EB,$E0,$F4,$3F,$4F
	.dc.b	$7D,$8F,$9A,$30,$FA,$32,$7E,$26,$F8,$3D,$5F,$0C,$FF,$5C,$DC,$BF
	.dc.b	$BE,$63,$CA,$7F,$C4,$B8,$72,$BE,$38,$F5,$C7,$A5,$7C,$F6,$E1,$FE
	.dc.b	$51,$BE,$13,$9B,$FB,$31,$89,$D5,$70,$7F,$D2,$31,$96,$C6,$5F,$F2
	.dc.b	$AF,$A5,$F5,$3C,$CF,$F2,$5B,$F5,$BE,$29,$F5,$BF,$E1,$7C,$BF,$C3
	.dc.b	$73,$98,$C5,$FF,$99,$6B,$7E,$BF,$E3,$1B,$7F,$CE,$DB,$3B,$AE,$2F
	.dc.b	$FA,$9A,$32,$D8,$CB,$FE,$B3,$8C,$BF,$E6,$CD,$3D,$1E,$77,$CF,$E2
	.dc.b	$FF,$32,$79,$4F,$85,$CD,$7B,$63,$19,$7F,$59,$9F,$D0,$EA,$9C,$E3
	.dc.b	$FD,$6E,$73,$8B,$EB,$FB,$9F,$1F,$CE,$7C,$F6,$32,$FF,$1F,$CD,$C3
	.dc.b	$FC,$63,$5A,$9C,$92,$3F,$BB,$4F,$3A,$E1,$A7,$F9,$F6,$F8,$57,$F8
	.dc.b	$16,$7C,$F1,$88,$FF,$47,$B5,$FF,$47,$FD,$BF,$C4,$3E,$E7,$A7,$FE
	.dc.b	$44,$C7,$FC,$A7,$18,$E2,$F8,$33,$9A,$FF,$4D,$B5,$FF,$4D,$B5,$FF
	.dc.b	$56,$FA,$BF,$C2,$9F,$F1,$A7,$77,$EF,$3C,$24,$FF,$9D,$F1,$FF,$1D
	.dc.b	$4F,$E3,$63,$86,$9F,$E9,$96,$CF,$FB,$BE,$B8,$FB,$BB,$BE,$95,$CF
	.dc.b	$0F,$F3,$35,$6A,$FF,$CB,$19,$6B,$0F,$CD,$7B,$3E,$7B,$FF,$5B,$B7
	.dc.b	$8C,$D3,$1C,$2F,$AE,$FC,$59,$FA,$3F,$6F,$F0,$19,$7F,$9E,$A3,$2D
	.dc.b	$3F,$CE,$1C,$78,$C7,$FC,$61,$F5,$1F,$94,$3F,$59,$FA,$FB,$CD,$C5
	.dc.b	$FF,$A1,$26,$D7,$F9,$DA,$B5,$39,$3F,$D1,$3F,$69,$F9,$E3,$F6,$DF
	.dc.b	$04,$BC,$4F,$FA,$B2,$30,$FF,$B6,$63,$F3,$B1,$8B,$FF,$6B,$B5,$FF
	.dc.b	$6B,$B5,$FF,$74,$B5,$FF,$68,$47,$23,$F4,$7F,$71,$FB,$03,$C7,$C3
	.dc.b	$6F,$FD,$3F,$16,$BF,$D2,$37,$F8,$DE,$5F,$F6,$54,$FE,$E3,$D7,$97
	.dc.b	$FE,$C4,$F8,$47,$C8,$63,$1F,$F2,$D5,$5C,$3F,$DC,$15,$AB,$FF,$56
	.dc.b	$CF,$F1,$9F,$B2,$FF,$76,$47,$FB,$39,$B7,$E2,$F8,$5D,$4F,$CA,$7E
	.dc.b	$29,$1F,$F8,$00,$D8,$F8,$ED,$7E,$3B,$FD,$A1,$13,$E3,$6F,$C5,$FF
	.dc.b	$A6,$23,$FB,$F5,$63,$E1,$3F,$CB,$95,$67,$89,$A1,$5C,$54,$6D,$EF
	.dc.b	$EA,$2F,$11,$E1,$38,$DF,$47,$A5,$F3,$BD,$DF,$19,$D9,$B8,$CD,$F4
	.dc.b	$8F,$09,$A3,$C2,$E6,$37,$19,$3C,$55,$F0,$6B,$B3,$E1,$8D,$46,$2F
	.dc.b	$E1,$C6,$2F,$A3,$4E,$87,$8A,$BA,$46,$9B,$1A,$BF,$1C,$77,$32,$5A
	.dc.b	$F8,$97,$E1,$AF,$80,$77,$9D,$3F,$52,$34,$FD,$6A,$EB,$5E,$8A,$EA
	.dc.b	$D3,$AD,$F9,$5B,$5E,$53,$B3,$F0,$3E,$33,$D2,$7C,$52,$B9,$38,$B9
	.dc.b	$69,$DE,$3C,$D3,$DD,$A7,$79,$E6,$FB,$9A,$FA,$CF,$19,$F0,$AB,$D6
	.dc.b	$DA,$F4,$C7,$C4,$8F,$C0,$9D,$63,$C6,$F5,$6B,$D6,$7B,$BF,$95,$6B
	.dc.b	$F6,$E3,$A9,$EB,$7A,$CF,$BD,$F5,$CD,$7C,$2D,$7F,$50,$DC,$3F,$38
	.dc.b	$F9,$E3,$F4,$35,$D5,$A7,$9C,$EA,$FC,$F6,$99,$3B,$B9,$69,$F1,$4F
	.dc.b	$9F,$5D,$EB,$F0,$F1,$EF,$9F,$86,$D3,$A9,$F1,$5C,$BF,$AE,$AF,$8B
	.dc.b	$87,$E3,$A3,$B1,$F1,$27,$C4,$FC,$06,$D7,$C0,$6E,$1F,$C5,$2F,$E2
	.dc.b	$36,$3E,$14,$F9,$AB,$E9,$1E,$A8,$E2,$AB,$8B,$63,$31,$83,$E0,$57
	.dc.b	$E4,$67,$CC,$7C,$93,$CC,$5A,$F3,$3E,$C3,$E9,$9F,$93,$C3,$F2,$2D
	.dc.b	$BF,$C2,$BB,$9F,$46,$1F,$16,$34,$7A,$4D,$5F,$8D,$EA,$7F,$14,$AE
	.dc.b	$6F,$03,$CE,$53,$67,$C4,$32,$7D,$93,$F0,$27,$D9,$3E,$D9,$D4,$F1
	.dc.b	$1C,$EB,$CE,$FE,$B8,$C4,$FF,$1E,$7F,$9B,$1F,$CE,$4A,$7F,$30,$2D
	.dc.b	$63,$D2,$7C,$6F,$FA,$DE,$78,$C6,$3D,$4D,$4E,$67,$FA,$14,$E5,$DF
	.dc.b	$C1,$B8,$7A,$36,$33,$18,$23,$FD,$06,$5F,$0D,$FD,$63,$56,$BF,$D6
	.dc.b	$A7,$35,$FD,$5D,$61,$D6,$B7,$3F,$DA,$27,$38,$6F,$0E,$98,$73,$6D
	.dc.b	$6F,$0E,$53,$DD,$F4,$6D,$D2,$7F,$DA,$B5,$FD,$B2,$31,$87,$92,$7F
	.dc.b	$34,$73,$C3,$F4,$8D,$3D,$D5,$E0,$79,$9E,$BB,$F0,$BB,$C7,$AC,$EE
	.dc.b	$FB,$47,$82,$B7,$5F,$EE,$86,$CE,$B1,$F7,$4F,$5C,$23,$E2,$3E,$21
	.dc.b	$8B,$E9,$18,$8D,$1F,$56,$FA,$9D,$B0,$F2,$1E,$57,$33,$F0,$27,$DA
	.dc.b	$7C,$42,$E3,$D9,$C6,$3F,$59,$EF,$87,$C0,$7C,$98,$FD,$3B,$E9,$5F
	.dc.b	$3E,$F5,$5C,$B0,$FA,$F7,$B8,$EB,$8F,$D7,$8F,$87,$1B,$C3,$9D,$7A
	.dc.b	$E3,$E0,$63,$E5,$3C,$D8,$76,$25,$F5,$72,$F4,$73,$3F,$7D,$E5,$1E
	.dc.b	$09,$F3,$98,$9D,$47,$12,$38,$21,$F2,$1C,$BF,$7E,$FD,$27,$A2,$3B
	.dc.b	$1F,$06,$7C,$B3,$FA,$3C,$3E,$13,$4E,$27,$C3,$3E,$39,$F3,$9D,$BD
	.dc.b	$CD,$47,$AA,$7B,$C7,$E5,$DF,$88,$79,$B1,$FA,$CE,$23,$D7,$18,$7C
	.dc.b	$47,$87,$0F,$88,$FE,$22,$BB,$3D,$4E,$37,$E4,$AE,$51,$E5,$AF,$7B
	.dc.b	$F1,$CE,$AE,$23,$F1,$D5,$D2,$7B,$5F,$C7,$23,$E4,$8D,$4F,$09,$F5
	.dc.b	$4F,$52,$47,$E6,$6B,$D2,$FE,$C6,$39,$4F,$47,$CC,$FC,$5B,$F5,$63
	.dc.b	$F9,$89,$F6,$9F,$44,$FA,$B7,$D2,$7F,$5F,$1F,$A4,$3E,$21,$EA,$3B
	.dc.b	$98,$BF,$43,$83,$47,$DB,$C3,$A3,$F8,$93,$85,$FE,$8D,$FD,$9E,$1E
	.dc.b	$9A,$E7,$1C,$9F,$73,$CE,$35,$7E,$68,$FE,$11,$8F,$9D,$E1,$1F,$C1
	.dc.b	$E3,$15,$C8,$97,$FC,$2A,$BC,$27,$A6,$3C,$24,$AE,$D8,$7D,$0A,$F4
	.dc.b	$3F,$61,$FE,$57,$E0,$39,$9E,$27,$F6,$28,$C3,$63,$38,$FF,$68,$9C
	.dc.b	$B6,$32,$FF,$B7,$A7,$81,$1E,$1E,$03,$E3,$7F,$D6,$3A,$7F,$A8,$F0
	.dc.b	$4E,$6B,$FD,$7B,$39,$8C,$47,$FB,$72,$F8,$F1,$7F,$B0,$AF,$8F,$C6
	.dc.b	$62,$91,$87,$9C,$EE,$D5,$BA,$74,$3E,$37,$15,$AC,$38,$9D,$1C,$3F
	.dc.b	$DE,$F0,$E9,$87,$3B,$E5,$3C,$30,$F6,$F0,$F4,$24,$67,$57,$8B,$88
	.dc.b	$E8,$77,$9C,$6B,$F1,$EB,$73,$D7,$E3,$5F,$0B,$F6,$5B,$1F,$5D,$AF
	.dc.b	$FB,$57,$14,$62,$3F,$DC,$71,$89,$E4,$48,$E4,$FE,$17,$5F,$BF,$87
	.dc.b	$D5,$DF,$E3,$9F,$2F,$0E,$F1,$F6,$7A,$BE,$59,$30,$F8,$9F,$1A,$F5
	.dc.b	$31,$1F,$03,$3F,$C1,$D6,$2F,$F0,$D7,$BA,$F9,$99,$3E,$07,$EB,$92
	.dc.b	$3E,$87,$92,$F0,$F1,$1D,$6B,$83,$CA,$BF,$05,$5B,$8F,$B1,$E5,$3D
	.dc.b	$EF,$D2,$36,$3F,$44,$D7,$E6,$35,$13,$CC,$7C,$63,$E3,$7F,$2D,$18
	.dc.b	$6C,$66,$7C,$37,$F9,$5F,$3F,$FC,$5D,$B7,$E9,$DB,$D5,$FA,$16,$D3
	.dc.b	$1E,$8F,$E4,$63,$3C,$26,$9C,$27,$73,$F4,$7D,$0F,$53,$FA,$D6,$BF
	.dc.b	$AC,$6D,$1B,$BF,$C3,$C7,$F7,$19,$7E,$A8,$FD,$74,$E7,$2F,$DC,$7A
	.dc.b	$5F,$E4,$5C,$7F,$22,$D7,$3E,$0F,$F1,$38,$C4,$7E,$C7,$E3,$7F,$6A
	.dc.b	$47,$F5,$D1,$FE,$4F,$C7,$F8,$44,$E6,$7F,$00,$FE,$FE,$75,$7F,$C0
	.dc.b	$E3,$55,$C2,$B9,$DF,$F0,$5F,$59,$F1,$BF,$C9,$5B,$3F,$C8,$DA,$FE
	.dc.b	$B5,$AF,$EB,$34,$7A,$5F,$C3,$09,$1F,$C1,$67,$8E,$0C,$BF,$13,$3F
	.dc.b	$E7,$AF,$8D,$FF,$4A,$B4,$FF,$4B,$F8,$A8,$CB,$63,$31,$82,$38,$3F
	.dc.b	$9D,$79,$CF,$8D,$FE,$44,$D8,$FE,$30,$D7,$F8,$99,$E1,$2F,$C5,$FF
	.dc.b	$09,$7C,$ED,$7F,$42,$D4,$D2,$78,$3E,$C6,$C7,$FA,$D3,$E0,$5F,$1A
	.dc.b	$F8,$DF,$E7,$91,$86,$C6,$67,$F9,$A9,$DA,$F3,$32,$3F,$9C,$D7,$17
	.dc.b	$77,$FC,$5B,$4F,$9F,$F1,$6F,$8C,$F9,$ED,$AF,$F6,$2B,$5F,$F6,$4B
	.dc.b	$5F,$AE,$1C,$3D,$23,$99,$1F,$F6,$8E,$EF,$8D,$F1,$47,$FB,$15,$B1
	.dc.b	$9C,$FF,$AE,$4F,$5A,$FE,$BD,$E9,$F7,$E3,$FF,$6C,$B6,$7F,$DB,$AD
	.dc.b	$7C,$AD,$7F,$1A,$F9,$FC,$2E,$39,$7F,$24,$9D,$B5,$DB,$5F,$CA,$FC
	.dc.b	$63,$E3,$7F,$88,$47,$FB,$B2,$31,$DF,$FE,$EC,$6C,$66,$76,$FC,$88
	.dc.b	$7C,$6B,$D8,$E0,$D9,$E8,$D7,$83,$5F,$E5,$F9,$3E,$33,$F9,$C7,$0F
	.dc.b	$C2,$6C,$FC,$26,$BF,$09,$AF,$27,$FC,$47,$1C,$7C,$DF,$DA,$4F,$16
	.dc.b	$BC,$5A,$FF,$40,$F8,$97,$AE,$F8,$DE,$3C,$9C,$A7,$31,$8C,$3F,$C3
	.dc.b	$F8,$BE,$37,$18,$AF,$75,$FE,$12,$C3,$DD,$5E,$0A,$46,$27,$F1,$B7
	.dc.b	$CD,$B5,$B9,$D5,$7E,$61,$AF,$F8,$D2,$B8,$39,$6B,$FC,$53,$A1,$F1
	.dc.b	$AF,$89,$F3,$C6,$1F,$4E,$4F,$8C,$7C,$6E,$AF,$86,$9F,$B4,$6E,$3F
	.dc.b	$E2,$0C,$79,$4F,$F8,$C3,$0E,$57,$C7,$1E,$B8,$F4,$AF,$D8,$E1,$F1
	.dc.b	$E3,$0F,$ED,$2F,$8C,$E9,$B1,$99,$DB,$FC,$C2,$32,$D8,$CB,$F7,$5A
	.dc.b	$FF,$90,$FD,$1F,$AF,$E6,$7F,$96,$9F,$B3,$F1,$8F,$8D,$FF,$0F,$63
	.dc.b	$FE,$18,$9C,$C6,$2F,$F8,$4D,$6F,$D6,$FB,$8D,$BF,$6B,$67,$75,$C5
	.dc.b	$FE,$8F,$19,$6C,$65,$FF,$09,$C6,$5F,$F3,$16,$7E,$8F,$27,$E8,$F0
	.dc.b	$FF,$9C,$FC,$97,$C6,$E6,$BD,$B1,$8C,$BF,$B0,$4F,$E4,$75,$4E,$71
	.dc.b	$FE,$C5,$39,$C5,$F1,$5D,$CF,$8F,$E7,$BE,$37,$19,$7F,$8F,$26,$E1
	.dc.b	$FE,$43,$AD,$4E,$49,$1F,$E3,$B9,$E7,$5C,$34,$FF,$46,$5F,$0A,$FF
	.dc.b	$30,$35,$FF,$07,$CF,$A6,$31,$1F,$E8,$5F,$8D,$F8,$87,$DC,$F4,$BF
	.dc.b	$C9,$58,$FF,$94,$23,$1C,$5F,$06,$73,$5F,$E6,$D6,$BF,$E6,$C6,$BF
	.dc.b	$E3,$7E,$CF,$C3,$FA,$9F,$C3,$A3,$FC,$FF,$E1,$FF,$8D,$7D,$EC,$D4
	.dc.b	$CB,$FC,$74,$FF,$04,$9D,$46,$0F,$F4,$AB,$8A,$E7,$7F,$E8,$C9,$FF
	.dc.b	$2D,$E3,$C6,$31,$3F,$E6,$FF,$63,$CA,$FF,$AA,$1B,$C6,$69,$8E,$17
	.dc.b	$C6,$FE,$24,$FD,$0F,$B5,$F7,$B2,$FF,$3A,$C6,$5A,$7F,$9D,$38,$F1
	.dc.b	$8F,$F9,$CF,$E2,$FF,$24,$7E,$AB,$E3,$7F,$C4,$F5,$FE,$7E,$8C,$B4
	.dc.b	$FF,$56,$46,$A7,$94,$7F,$A1,$3E,$37,$F3,$87,$ED,$3E,$05,$78,$9F
	.dc.b	$F5,$83,$F9,$98,$CC,$7F,$B6,$E3,$04,$70,$4A,$FF,$55,$4F,$07,$FD
	.dc.b	$59,$9B,$E3,$7F,$58,$78,$78,$2D,$FF,$A6,$22,$D7,$FA,$7E,$FF,$13
	.dc.b	$D1,$FE,$DF,$9F,$DA,$7A,$D2,$FF,$D9,$5F,$08,$F8,$DC,$63,$FE,$9B
	.dc.b	$AB,$87,$FB,$AE,$B5,$7F,$EA,$69,$FE,$2B,$F1,$BF,$E4,$D8,$FF,$6A
	.dc.b	$36,$FC,$5F,$09,$A9,$F9,$0F,$C5,$23,$FE,$FF,$6C,$7C,$26,$BF,$EC
	.dc.b	$AC,$D1,$FD,$56,$3F,$CB,$7F,$1B,$FE,$A4,$F8,$B7,$C4,$7F,$8A,$3D
	.dc.b	$2F,$ED,$B1,$8A,$FF,$15,$D6,$EB,$58,$D9,$C8,$AE,$0F,$FA,$C2,$F9
	.dc.b	$D7,$F4,$4B,$E1,$39,$BA,$FF,$AD,$A2,$D7,$08,$EB,$7D,$1F,$F0,$A5
	.dc.b	$70,$55,$C2,$30,$D8,$F2,$5F,$85,$A7,$F9,$F2,$FB,$3F,$10,$C9,$E5
	.dc.b	$29,$E5,$2B,$C1,$C5,$F9,$63,$11,$E4,$37,$7E,$46,$9D,$0D,$4F,$18
	.dc.b	$D3,$63,$57,$EB,$8F,$85,$1C,$2F,$E9,$BA,$3A,$C6,$A7,$84,$FC,$18
	.dc.b	$E8,$66,$77,$19,$39,$C6,$4D,$BF,$27,$E3,$3C,$87,$12,$B9,$3C,$6E
	.dc.b	$5A,$75,$9F,$A1,$7E,$E8,$EE,$71,$8C,$89,$E4,$E1,$F2,$47,$08,$FC
	.dc.b	$2B,$5F,$B5,$5E,$87,$27,$47,$9B,$F7,$66,$BF,$8C,$3C,$13,$E3,$74
	.dc.b	$D7,$51,$E2,$7A,$47,$B5,$FC,$44,$F4,$C3,$FC,$AF,$56,$D5,$D8,$FD
	.dc.b	$4C,$7C,$A6,$DF,$C8,$7D,$27,$47,$B3,$E7,$BE,$8F,$33,$F3,$53,$CD
	.dc.b	$B3,$CD,$A7,$CB,$79,$DF,$E6,$30,$FC,$3C,$F8,$8D,$14,$F8,$65,$3E
	.dc.b	$1C,$FC,$59,$F7,$BD,$30,$FB,$D3,$F1,$8E,$13,$A6,$C7,$C4,$BF,$59
	.dc.b	$2B,$94,$FE,$77,$0F,$B8,$FA,$8F,$19,$4F,$19,$D0,$F8,$87,$EE,$67
	.dc.b	$E2,$3F,$B8,$9F,$C4,$34,$F3,$1C,$EF,$CC,$DB,$F3,$4F,$92,$FF,$05
	.dc.b	$3F,$54,$FA,$E7,$D5,$8D,$11,$CD,$7C,$F7,$F6,$26,$2B,$E3,$BC,$CF
	.dc.b	$CD,$CF,$99,$FE,$00,$79,$8F,$41,$B3,$DC,$FD,$F6,$9F,$68,$E6,$71
	.dc.b	$39,$9C,$4A,$7A,$4F,$21,$FC,$34,$FF,$3E,$9F,$C7,$4B,$86,$1F,$E4
	.dc.b	$C7,$FB,$CB,$D1,$7C,$6B,$D3,$E3,$18,$6C,$66,$7F,$D1,$33,$96,$FC
	.dc.b	$17,$FA,$B1,$BC,$5C,$A3,$11,$DD,$AF,$EB,$8F,$F6,$0B,$C7,$5F,$EA
	.dc.b	$B4,$F8,$4D,$46,$1B,$9F,$19,$57,$18,$C3,$6F,$F7,$73,$CD,$FC,$29
	.dc.b	$8A,$E6,$EA,$78,$3E,$4C,$36,$FF,$B5,$A7,$6D,$33,$5E,$29,$B5,$CA
	.dc.b	$3C,$13,$FC,$05,$B8,$7B,$A7,$46,$EF,$AC,$EB,$0F,$5E,$1D,$0E,$F3
	.dc.b	$A3,$75,$CA,$33,$84,$3C,$11,$97,$C5,$3C,$5E,$86,$A7,$86,$1F,$3E
	.dc.b	$32,$F9,$E3,$37,$F4,$6B,$C3,$1A,$8C,$56,$8F,$99,$19,$7E,$81,$93
	.dc.b	$E7,$1E,$13,$CA,$67,$0F,$98,$E6,$FC,$8E,$8C,$57,$C6,$71,$1A,$3D
	.dc.b	$0E,$8F,$09,$A8,$E1,$1A,$38,$3E,$A8,$E0,$79,$A3,$2F,$27,$84,$72
	.dc.b	$38,$3A,$8E,$11,$97,$83,$A3,$0E,$8C,$90,$C1,$27,$F0,$8E,$5F,$8B
	.dc.b	$3E,$28,$C5,$FD,$47,$D3,$3C,$1F,$CC,$18,$C3,$F2,$0F,$CF,$7F,$0D
	.dc.b	$5E,$99,$E0,$76,$BF,$3B,$8B,$ED,$3E,$1B,$EA,$F1,$BE,$77,$EF,$9F
	.dc.b	$44,$61,$F3,$9F,$85,$3D,$D5,$F9,$57,$CD,$87,$E0,$E3,$BD,$78,$9D
	.dc.b	$5F,$87,$0F,$35,$F4,$3F,$01,$DE,$E1,$78,$5F,$C9,$A3,$0D,$8C,$CF
	.dc.b	$F1,$53,$FC,$07,$3A,$70,$17,$D0,$3D,$CE,$8D,$9D,$CE,$A7,$24,$8C
	.dc.b	$CE,$DF,$04,$ED,$CC,$F9,$F9,$DF,$1B,$49,$3D,$A3,$2D,$8C,$BD,$23
	.dc.b	$2D,$8F,$E8,$FE,$93,$F5,$57,$C4,$61,$B5,$F0,$27,$53,$92,$46,$67
	.dc.b	$6F,$C9,$9D,$B9,$9F,$9D,$F1,$F2,$3E,$2E,$88,$C4,$7F,$B2,$3C,$AF
	.dc.b	$A8,$47,$6D,$74,$60,$87,$CA,$FD,$BF,$A3,$F4,$9A,$E6,$7F,$05,$18
	.dc.b	$8F,$EA,$9A,$A3,$11,$F8,$26,$BF,$81,$6C,$6E,$7F,$B7,$70,$7F,$B4
	.dc.b	$6B,$FB,$5F,$7E,$08,$E0,$87,$F6,$C7,$2F,$E0,$BE,$3D,$F0,$9F,$97
	.dc.b	$6C,$7E,$31,$AE,$E3,$F0,$DC,$92,$3F,$08,$D7,$F0,$4D,$7D,$FC,$3F
	.dc.b	$E7,$B7,$67,$E6,$3E,$3D,$EC,$7E,$9D,$B1,$98,$C1,$1C,$10,$FC,$DC
	.dc.b	$70,$7F,$09,$FA,$DB,$EE,$FF,$22,$D8,$FC,$73,$5D,$C6,$88,$E9,$AF
	.dc.b	$6F,$AD,$F3,$11,$F5,$3F,$BF,$F3,$DE,$4D,$24,$FE,$DA,$32,$D8,$CB
	.dc.b	$FB,$48,$CB,$63,$F8,$57,$C0,$BC,$5A,$49,$FE,$03,$19,$6C,$65,$FD
	.dc.b	$FC,$65,$B1,$FE,$3A,$F4,$DF,$B6,$BE,$1B,$0D,$AF,$E2,$73,$A9,$C9
	.dc.b	$FC,$77,$D4,$79,$74,$95,$FC,$52,$32,$F2,$FD,$C9,$F0,$BE,$16,$C6
	.dc.b	$63,$04,$70,$43,$F9,$46,$09,$E3,$18,$FD,$D9,$F2,$5E,$D6,$C7,$A9
	.dc.b	$AE,$E3,$F9,$B7,$B5,$E6,$8F,$E4,$AD,$7F,$6E,$D7,$F9,$BB,$55,$BF
	.dc.b	$7C,$FD,$CB,$7B,$7E,$74,$60,$FD,$DF,$BB,$FB,$9A,$FE,$66,$D7,$F8
	.dc.b	$2B,$5F,$E8,$AD,$7F,$A2,$37,$BC,$D2,$77,$7F,$51,$6C,$FF,$46,$6B
	.dc.b	$FD,$47,$E0,$BF,$19,$F6,$9F,$E7,$C6,$C7,$F5,$26,$BB,$E4,$4F,$F5
	.dc.b	$DF,$82,$FC,$34,$FF,$4C,$6B,$FC,$79,$AF,$F5,$E6,$BF,$EE,$2F,$8A
	.dc.b	$7C,$BF,$FA,$25,$B1,$FE,$8E,$6B,$FD,$91,$AF,$F6,$96,$BF,$E7,$F6
	.dc.b	$BD,$63,$A1,$FE,$95,$F9,$1F,$E3,$71,$FD,$C5,$AF,$FA,$1D,$AF,$FB
	.dc.b	$D9,$AB,$EB,$DF,$37,$FE,$9F,$6C,$7F,$A9,$1A,$FF,$7A,$6B,$FD,$FD
	.dc.b	$AF,$FA,$55,$A8,$ED,$1D,$4F,$F5,$67,$C8,$3E,$4A,$92,$7F,$D6,$71
	.dc.b	$96,$C6,$5F,$F0,$5C,$65,$B1,$FE,$B7,$FB,$27,$C1,$52,$4F,$FA,$FA
	.dc.b	$32,$D8,$CB,$FE,$16,$8C,$B6,$3F,$D8,$5F,$26,$F9,$BA,$49,$FF,$64
	.dc.b	$C6,$5F,$F6,$5F,$DF,$3E,$6F,$FD,$94,$D8,$CC,$60,$CC,$C8,$E2,$D7
	.dc.b	$FC,$53,$F2,$DF,$C6,$FA,$BF,$DA,$ED,$8C,$F8,$5F,$B8,$6C,$EB,$8B
	.dc.b	$F9,$34,$7F,$B2,$69,$18,$8F,$AD,$18,$E5,$E0,$7F,$B9,$3F,$11,$7C
	.dc.b	$DB,$E1,$30,$DB,$FF,$1E,$CE,$A7,$27,$1F,$A6,$D4,$FF,$91,$65,$18
	.dc.b	$8F,$F6,$E4,$62,$78,$1F,$E4,$8B,$7E,$D7,$FE,$3A,$8C,$36,$BF,$CA
	.dc.b	$13,$AF,$73,$A4,$7F,$4D,$9D,$3F,$E5,$5F,$C0,$DF,$51,$A8,$46,$30
	.dc.b	$FF,$2B,$C6,$27,$53,$98,$FF,$2D,$B5,$E9,$1A,$24,$FF,$98,$3E,$11
	.dc.b	$F3,$8F,$3B,$CB,$18,$6C,$66,$78,$36,$FF,$CD,$95,$B9,$CD,$70,$23
	.dc.b	$E8,$FC,$11,$F2,$1E,$D6,$C6,$63,$04,$70,$4B,$FF,$3B,$CF,$07,$FC
	.dc.b	$A9,$F3,$4F,$9B,$F9,$BF,$06,$8C,$5F,$F9,$7D,$B5,$C2,$BE,$77,$E0
	.dc.b	$AF,$5E,$92,$FE,$7C,$65,$B1,$9A,$FF,$44,$CE,$5B,$1F,$7B,$CB,$78
	.dc.b	$FF,$4E,$7F,$9A,$70,$66,$31,$1F,$E9,$16,$F9,$3F,$E9,$3F,$9B,$FC
	.dc.b	$3B,$72,$FE,$8D,$82,$72,$D8,$FE,$A1,$8F,$87,$CB,$F2,$63,$FD,$7A
	.dc.b	$30,$49,$D9,$23,$FB,$2C,$65,$B1,$97,$93,$FE,$16,$FD,$F7,$F6,$ED
	.dc.b	$ED,$FC,$EF,$72,$73,$5F,$E0,$A6,$BF,$DF,$BE,$5F,$E3,$7A,$26,$93
	.dc.b	$0F,$F1,$1F,$EF,$2F,$A7,$EA,$D8,$CC,$60,$F0,$F4,$3E,$5A,$92,$7F
	.dc.b	$C5,$B1,$96,$C6,$71,$FF,$0F,$CE,$5B,$1F,$6F,$D9,$F3,$57,$F8,$BF
	.dc.b	$1C,$11,$FF,$1C,$FC,$17,$F8,$59,$BA,$FC,$48,$C6,$1F,$E2,$C6,$BF
	.dc.b	$E9,$6C,$5F,$11,$F8,$96,$DF,$F9,$42,$30,$47,$04,$9F,$F2,$8C,$7E
	.dc.b	$47,$DF,$F3,$C7,$F9,$7F,$1C,$12,$BF,$CC,$CD,$8D,$46,$0F,$F6,$FF
	.dc.b	$CC,$FF,$58,$6F,$9D,$F9,$D8,$C4,$FF,$B7,$9A,$FF,$6D,$1D,$AF,$B9
	.dc.b	$FF,$08,$DF,$FB,$45,$B3,$FE,$CD,$6B,$C6,$78,$7A,$DF,$BC,$9C,$CF
	.dc.b	$F7,$99,$CD,$73,$AD,$1E,$5F,$8E,$91,$EA,$6D,$79,$DA,$FF,$B1,$B0
	.dc.b	$FF,$80,$13,$B3,$FD,$E5,$F2,$3F,$BB,$DC,$D2,$46,$89,$5F,$EF,$58
	.dc.b	$CB,$F0,$C7,$53,$E3,$7E,$3C,$7F,$84,$E3,$1A,$FF,$09,$9C,$CF,$06
	.dc.b	$BF,$E1,$9E,$0F,$E8,$CF,$E1,$8F,$E8,$AE,$09,$87,$F4,$88,$D1,$E3
	.dc.b	$29,$E5,$8F,$E3,$1F,$1A,$F9,$2A,$4A,$FF,$78,$36,$BF,$DE,$8D,$7F
	.dc.b	$DD,$F3,$C6,$7B,$63,$EF,$F8,$6F,$61,$23,$38,$FF,$AE,$9B,$19,$BF
	.dc.b	$91,$5C,$67,$27,$0F,$8F,$F8,$C4,$70,$7B,$78,$5F,$20,$F9,$9F,$E8
	.dc.b	$AD,$9D,$37,$0F,$AD,$5C,$2B,$64,$AD,$B6,$7C,$BC,$0F,$81,$FC,$31
	.dc.b	$2B,$E0,$C6,$5B,$19,$9E,$2D,$7F,$51,$93,$E4,$BF,$9E,$47,$C2,$6B
	.dc.b	$A6,$BF,$89,$6B,$F8,$E6,$C7,$3F,$13,$FA,$A4,$60,$CE,$DF,$EC,$CC
	.dc.b	$3E,$9F,$C4,$FE,$44,$8F,$E2,$70,$7C,$BF,$F9,$A7,$E4,$3F,$0D,$1F
	.dc.b	$BF,$F5,$7F,$C6,$11,$88,$FE,$05,$3B,$E4,$FF,$11,$DD,$9F,$D2,$4D
	.dc.b	$7F,$CE,$97,$CA,$BF,$2D,$39,$9F,$F7,$A5,$57,$C6,$2F,$C0,$E8,$C5
	.dc.b	$72,$69,$E7,$9E,$55,$D6,$FC,$13,$E4,$36,$79,$27,$6E,$0E,$D7,$E0
	.dc.b	$7C,$6F,$8A,$79,$C7,$C2,$38,$38,$7F,$D2,$55,$E1,$AE,$B5,$C5,$B7
	.dc.b	$F2,$2B,$C0,$76,$71,$3E,$1A,$E5,$3A,$6C,$66,$3C,$35,$D1,$F8,$46
	.dc.b	$CC,$96,$3C,$4D,$AE,$E7,$CB,$8C,$9D,$4A,$E4,$DC,$7C,$B8,$CB,$63
	.dc.b	$2F,$C5,$8C,$B6,$32,$FC,$63,$CB,$1D,$E3,$31,$F0,$CF,$5C,$65,$B1
	.dc.b	$E3,$69,$E1,$29,$DC,$A7,$78,$F1,$B6,$3B,$D7,$CC,$7A,$47,$E1,$4F
	.dc.b	$8C,$72,$3C,$86,$67,$C8,$D8,$E2,$FE,$29,$B7,$F8,$58,$E2,$64,$F4
	.dc.b	$1C,$4A,$64,$F5,$19,$29,$93,$D4,$79,$6B,$D4,$F2,$8F,$04,$62,$38
	.dc.b	$C6,$1F,$A3,$3E,$6B,$FC,$54,$77,$3A,$94,$F4,$CF,$1B,$F9,$71,$89
	.dc.b	$E2,$F9,$4C,$1F,$28,$E5,$7E,$37,$A9,$B7,$0F,$F0,$23,$FD,$4A,$FF
	.dc.b	$17,$23,$FD,$7F,$E4,$DF,$2C,$F9,$7F,$F5,$5C,$61,$B1,$99,$E0,$D7
	.dc.b	$FB,$D4,$BE,$5D,$EA,$F2,$6C,$F7,$6B,$C1,$AF,$F8,$96,$79,$4A,$1F
	.dc.b	$31,$FC,$A6,$7F,$C1,$F1,$98,$C7,$6F,$F8,$B5,$D5,$F8,$1B,$7B,$DB
	.dc.b	$B5,$6D,$B7,$93,$C1,$87,$89,$F0,$C7,$97,$0F,$5E,$BE,$09,$FE,$6D
	.dc.b	$5E,$47,$F9,$2B,$C6,$3C,$31,$C8,$EB,$7E,$16,$9E,$11,$3E,$1C,$78
	.dc.b	$9D,$E3,$91,$A9,$E4,$F0,$BF,$81,$87,$9F,$1E,$98,$79,$70,$E4,$74
	.dc.b	$9D,$5F,$49,$D4,$7C,$48,$C5,$F4,$7E,$2B,$5F,$11,$DB,$0F,$1D,$7A
	.dc.b	$6B,$AD,$F8,$1E,$8F,$86,$B9,$E1,$F2,$AF,$FC,$CD,$8F,$BC,$F9,$26
	.dc.b	$E7,$B9,$E7,$8E,$F8,$F8,$F1,$FA,$18,$FA,$6F,$93,$EE,$79,$B9,$25
	.dc.b	$F8,$8F,$B3,$5C,$EF,$D6,$76,$C3,$E5,$C7,$E0,$AB,$8D,$79,$2F,$AB
	.dc.b	$CD,$F0,$BF,$78,$EB,$5F,$88,$AE,$D1,$F2,$4E,$B1,$1F,$C5,$DF,$C4
	.dc.b	$BE,$F8,$7B,$A7,$B4,$62,$75,$18,$3D,$98,$FD,$17,$C4,$FC,$BA,$E5
	.dc.b	$1E,$9A,$EB,$87,$B6,$39,$19,$3D,$58,$FB,$9F,$C8,$63,$E9,$7E,$A1
	.dc.b	$E9,$C3,$58,$7E,$96,$BE,$1C,$F0,$3E,$19,$EE,$3F,$04,$7C,$5A,$8F
	.dc.b	$42,$57,$A6,$FC,$D1,$E5,$25,$FB,$09,$8F,$C4,$8C,$CF,$09,$FC,$B4
	.dc.b	$7C,$4C,$7F,$2A,$7E,$60,$8E,$48,$7A,$6F,$DB,$7D,$6B,$DD,$5F,$8F
	.dc.b	$38,$D7,$E6,$2B,$E9,$3C,$63,$FD,$3D,$C0,$F9,$8F,$F4,$94,$FF,$A7
	.dc.b	$5B,$C9,$FE,$A5,$6C,$EE,$7F,$D0,$F1,$82,$38,$3F,$D1,$FF,$34,$F8
	.dc.b	$EF,$F6,$96,$AE,$2F,$F5,$74,$E6,$BF,$98,$4E,$5F,$F5,$AC,$65,$FE
	.dc.b	$7B,$3E,$8B,$E6,$FF,$AD,$FE,$69,$F2,$B8,$AF,$F5,$EC,$61,$EB,$3B
	.dc.b	$9D,$67,$FE,$7D,$F1,$2D,$EE,$FB,$9C,$AB,$27,$0E,$85,$72,$C9,$F1
	.dc.b	$14,$97,$E0,$3F,$D9,$FD,$8F,$80,$CC,$76,$CD,$7C,$A3,$84,$62,$FC
	.dc.b	$F7,$C6,$73,$C1,$F1,$77,$F9,$17,$FE,$DB,$E0,$F6,$CE,$77,$FF,$7B
	.dc.b	$78,$0F,$07,$35,$31,$E6,$FC,$9F,$07,$FD,$6F,$3F,$13,$EA,$25,$79
	.dc.b	$38,$D1,$C6,$3D,$5B,$FF,$A3,$F5,$F1,$B7,$3F,$BF,$CB,$E4,$F1,$7E
	.dc.b	$43,$2F,$DF,$5B,$7F,$DD,$D3,$F5,$3D,$49,$1E,$17,$EC,$F8,$DE,$D8
	.dc.b	$FB,$F1,$A7,$E6,$66,$8D,$ED,$F8,$0F,$9B,$7C,$DF,$E7,$9B,$7F,$EE
	.dc.b	$66,$BF,$EE,$06,$BF,$EC,$46,$BF,$E0,$7E,$34,$62,$11,$89,$EC,$7E
	.dc.b	$B3,$3F,$D5,$4F,$C8,$EA,$FC,$51,$E5,$9F,$C9,$FC,$44,$9F,$A0,$FE
	.dc.b	$0B,$C0,$C4,$7B,$A2,$F0,$FE,$0B,$9F,$FD,$65,$3F,$33,$C5,$FC,$B1
	.dc.b	$1F,$C6,$FC,$D3,$E6,$F1,$1F,$6A,$6D,$FE,$32,$B5,$3C,$3D,$AF,$E0
	.dc.b	$B0,$69,$23,$FC,$54,$7F,$BA,$A3,$4E,$0F,$90,$E1,$C6,$C8,$C4,$71
	.dc.b	$9F,$E2,$7C,$4F,$A5,$FD,$B9,$27,$C8,$7E,$2B,$E3,$3F,$68,$47,$F5
	.dc.b	$7E,$A3,$CF,$C4,$7E,$62,$6E,$7F,$AA,$AD,$4F,$E7,$FE,$5F,$F8,$29
	.dc.b	$1F,$DA,$FA,$CF,$3B,$11,$F9,$09,$B7,$FB,$4A,$D4,$FE,$F3,$D8,$70
	.dc.b	$7F,$2E,$6C,$7C,$53,$F1,$2E,$1F,$8A,$D7,$E6,$FB,$0F,$99,$7C,$D7
	.dc.b	$F2,$18,$C3,$63,$33,$E1,$BF,$E4,$3F,$3D,$FC,$3E,$3F,$89,$37,$2F
	.dc.b	$E1,$6D,$F0,$BF,$CC,$98,$F4,$AF,$E2,$B3,$97,$F3,$ED,$E5,$39,$9E
	.dc.b	$13,$FC,$7F,$DF,$F5,$3F,$A9,$6B,$FA,$86,$D1,$BC,$8F,$A7,$FE,$6E
	.dc.b	$4A,$FE,$4F,$3F,$1F,$2F,$E4,$7F,$04,$F6,$3F,$AD,$B6,$BE,$76,$28
	.dc.b	$C4,$7E,$49,$AF,$E6,$A1,$8B,$E6,$7F,$A3,$92,$7F,$9B,$D7,$2C,$3F
	.dc.b	$99,$D6,$E7,$F9,$45,$6D,$FE,$81,$5B,$D7,$FA,$D4,$E6,$F8,$BF,$D5
	.dc.b	$A7,$55,$FC,$33,$E1,$5D,$DF,$DD,$9B,$1E,$C6,$BF,$AC,$6B,$EB,$C5
	.dc.b	$F4,$5F,$D8,$09,$1F,$D4,$27,$8E,$3F,$DC,$E3,$E0,$78,$3F,$E0,$66
	.dc.b	$9F,$E0,$8C,$BF,$90,$38,$3F,$34,$E0,$91,$C5,$FE,$4D,$F4,$8F,$A7
	.dc.b	$FD,$E5,$7F,$67,$8C,$77,$7F,$7B,$6C,$7F,$6B,$3B,$61,$E5,$F9,$A7
	.dc.b	$CD,$79,$5B,$3F,$E1,$C6,$BF,$AD,$6B,$FC,$4D,$AF,$F1,$86,$BF,$EC
	.dc.b	$36,$A7,$32,$FA,$7F,$E8,$EF,$F8,$64,$C6,$7F,$E1,$36,$D7,$A6,$BC
	.dc.b	$71,$BA,$E3,$C3,$FE,$22,$6D,$7F,$B7,$1A,$F8,$70,$FF,$0F,$2F,$03
	.dc.b	$F7,$4F,$F2,$D6,$BF,$CC,$9A,$FF,$B7,$5A,$F1,$F5,$3E,$E4,$60,$8F
	.dc.b	$F9,$13,$D2,$7D,$4B,$EC,$7F,$C0,$CD,$BF,$F0,$F4,$EA,$7C,$B7,$FC
	.dc.b	$17,$E0,$3E,$43,$FE,$4E,$6D,$7F,$1F,$6B,$FD,$15,$AF,$F4,$A7,$E8
	.dc.b	$7A,$7C,$76,$34,$93,$FE,$5A,$F9,$AF,$D2,$F7,$5E,$93,$49,$3F,$CB
	.dc.b	$E7,$FC,$0D,$E3,$71,$F4,$BF,$9B,$CF,$F5,$79,$CB,$B8,$CA,$19,$3E
	.dc.b	$57,$FA,$77,$1F,$B5,$B3,$ED,$6B,$ED,$6B,$CE,$79,$7C,$DF,$EC,$63
	.dc.b	$12,$8C,$47,$82,$31,$5D,$E7,$3C,$0F,$B0,$F3,$4F,$D6,$6D,$FF,$97
	.dc.b	$B3,$7C,$CF,$F9,$56,$B5,$3F,$9B,$7D,$D3,$AF,$77,$C2,$D8,$FE,$B0
	.dc.b	$D9,$FE,$FC,$D7,$FC,$06,$D7,$9D,$77,$76,$D7,$6D,$7E,$76,$0F,$82
	.dc.b	$FE,$BF,$7F,$90,$6C,$FF,$9E,$3A,$7E,$97,$99,$E9,$85,$F1,$F2,$F0
	.dc.b	$DC,$39,$46,$EF,$8C,$60,$FF,$3D,$38,$AF,$F0,$D3,$5F,$F4,$EE,$CC
	.dc.b	$3E,$AC,$E6,$7F,$D3,$B3,$9A,$FE,$FD,$39,$93,$4C,$76,$BE,$A9,$F1
	.dc.b	$1E,$B8,$C3,$EA,$F9,$77,$CB,$E6,$B7,$8F,$ED,$27,$2F,$F8,$37,$54
	.dc.b	$63,$2F,$F1,$A4,$62,$FF,$13,$39,$9F,$F1,$DC,$E6,$B9,$4E,$67,$EA
	.dc.b	$B5,$F9,$ED,$F2,$7F,$B4,$FD,$07,$D3,$6D,$7F,$4F,$F9,$7C,$79,$4A
	.dc.b	$D7,$92,$9F,$CF,$DA,$31,$7F,$D7,$E7,$31,$8C,$7F,$C9,$78,$5D,$FF
	.dc.b	$B9,$D7,$87,$CB,$FF,$2C,$61,$FE,$9B,$8F,$09,$1F,$EF,$7E,$B5,$C3
	.dc.b	$FD,$17,$36,$FF,$C1,$B3,$6B,$FD,$05,$7D,$2F,$FC,$15,$3E,$88,$C4
	.dc.b	$7F,$82,$DA,$FF,$B6,$7E,$BF,$E0,$9F,$17,$CF,$7C,$9E,$30,$FF,$0E
	.dc.b	$BF,$0A,$33,$1F,$E1,$A6,$D7,$F9,$A9,$AF,$F8,$67,$D2,$FF,$27,$CF
	.dc.b	$F8,$E3,$D3,$FD,$27,$87,$F9,$7F,$94,$FA,$95,$AB,$FF,$21,$D6,$0D
	.dc.b	$F5,$2B,$18,$7F,$B7,$70,$E4,$D9,$D3,$FE,$5F,$CB,$8D,$7E,$07,$D6
	.dc.b	$F2,$E3,$FE,$A9,$6B,$FE,$60,$9F,$BF,$18,$E3,$FF,$18,$FD,$8F,$E1
	.dc.b	$8F,$C5,$62,$F8,$4C,$65,$FE,$D5,$9B,$8F,$FB,$B6,$B5,$39,$3F,$D9
	.dc.b	$FF,$53,$F9,$3F,$82,$3F,$95,$F7,$3E,$5F,$2F,$FB,$6A,$34,$4A,$FF
	.dc.b	$45,$EB,$71,$FF,$67,$FD,$97,$E6,$8F,$CF,$7C,$AB,$E5,$71,$7F,$EF
	.dc.b	$C8,$C3,$FE,$52,$6D,$7F,$A8,$9A,$FF,$94,$5A,$FF,$B3,$DA,$FF,$B3
	.dc.b	$9B,$B9,$A6,$34,$EB,$9F,$FB,$D3,$0F,$2F,$C6,$FC,$16,$63,$FE,$F2
	.dc.b	$72,$D3,$FE,$00,$7D,$AF,$ED,$8F,$DD,$7C,$2B,$E4,$B1,$3A,$26,$9F
	.dc.b	$C7,$1F,$F7,$34,$ED,$CE,$27,$FD,$D5,$93,$E7,$FE,$A1,$FC,$1F,$E4
	.dc.b	$3D,$33,$A8,$C6,$7F,$C1,$27,$E0,$91,$C1,$27,$FC,$FF,$19,$7F,$D0
	.dc.b	$3E,$03,$ED,$BF,$88,$9F,$C5,$7E,$3B,$FC,$1B,$3F,$DF,$72,$FF,$3D
	.dc.b	$CE,$A3,$07,$FC,$01,$AE,$4E,$34,$FF,$80,$37,$C9,$F8,$43,$87,$FD
	.dc.b	$17,$7F,$91,$F9,$89,$7C,$65,$97,$F8,$23,$2F,$8D,$87,$1D,$58,$70
	.dc.b	$8E,$B6,$9C,$E1,$CF,$0E,$4F,$CE,$BF,$E5,$FF,$27,$23,$5F,$25,$71
	.dc.b	$E7,$68,$FA,$47,$F6,$8F,$93,$EB,$3A,$8C,$D7,$23,$FC,$87,$3A,$CF
	.dc.b	$D3,$9F,$B2,$BF,$61,$DD,$2F,$E0,$71,$7F,$AD,$DF,$85,$AF,$F0,$AC
	.dc.b	$BD,$AD,$C7,$9B,$67,$F1,$BF,$11,$CE,$BE,$1F,$0F,$F6,$F6,$F6,$7F
	.dc.b	$51,$DF,$FB,$2C,$62,$FE,$85,$6F,$1F,$F5,$CD,$EF,$CE,$FE,$AA,$FF
	.dc.b	$56,$A8,$7C,$DF,$0A,$E1,$F2,$1F,$CA,$12,$3F,$22,$D8,$FF,$24,$1F
	.dc.b	$E4,$9D,$FE,$FC,$66,$3F,$DA,$1C,$3F,$CC,$DD,$9C,$3E,$63,$FC,$A1
	.dc.b	$F2,$DF,$E6,$28,$FF,$16,$E1,$8F,$7B,$70,$AC,$3A,$15,$C5,$E8,$3E
	.dc.b	$47,$0B,$D8,$FC,$8C,$62,$3A,$1B,$BF,$F1,$FE,$1F,$E6,$39,$D6,$1E
	.dc.b	$0A,$FF,$24,$CD,$7F,$CD,$11,$6F,$F3,$B8,$70,$9F,$2E,$1D,$1A,$7C
	.dc.b	$2A,$E1,$7C,$E3,$C7,$87,$A3,$0F,$3D,$FC,$2B,$F2,$3C,$E7,$B1,$BB
	.dc.b	$CD,$F5,$3D,$87,$C7,$8F,$26,$1B,$8F,$19,$F2,$1C,$1D,$2B,$B1,$E1
	.dc.b	$8E,$67,$C2,$29,$F0,$8B,$87,$A1,$F1,$57,$73,$C3,$3F,$0E,$AC,$FD
	.dc.b	$B7,$C0,$F9,$E3,$C1,$3C,$5A,$79,$23,$AB,$5F,$53,$D4,$F4,$4F,$B1
	.dc.b	$E3,$1D,$4E,$25,$3F,$13,$3D,$DC,$9E,$F7,$CE,$FC,$19,$AF,$E3,$6F
	.dc.b	$ED,$61,$F4,$A3,$85,$F8,$1A,$74,$38,$1F,$28,$FC,$0B,$EB,$7E,$65
	.dc.b	$7A,$5F,$B8,$D7,$CF,$1B,$78,$9E,$B3,$EA,$5F,$AD,$D9,$93,$F4,$33
	.dc.b	$CD,$EF,$3F,$25,$F3,$61,$F6,$DE,$27,$A8,$F9,$E7,$D9,$8F,$9E,$71
	.dc.b	$72,$7C,$77,$2E,$CF,$5B,$B7,$C2,$64,$E2,$7C,$69,$F9,$F3,$FA,$E9
	.dc.b	$E7,$3D,$4E,$67,$13,$99,$93,$9C,$71,$7D,$B8,$7D,$77,$E3,$CF,$CB
	.dc.b	$C3,$F6,$B1,$FA,$63,$99,$E3,$8F,$0B,$4F,$09,$DC,$FD,$21,$FC,$68
	.dc.b	$FE,$1E,$58,$C3,$A3,$F9,$5F,$AE,$F9,$8F,$F7,$9C,$7F,$22,$8C,$7B
	.dc.b	$BC,$DB,$19,$9D,$BF,$CB,$A1,$F3,$1F,$EB,$48,$E4,$D7,$4D,$7C,$2D
	.dc.b	$47,$FB,$5D,$E5,$9F,$F2,$F9,$FF,$14,$4F,$F2,$56,$BF,$C3,$1F,$0F
	.dc.b	$16,$6B,$8E,$5E,$EF,$03,$FA,$4E,$1C,$DB,$1E,$2A,$E0,$7E,$AE,$FA
	.dc.b	$ED,$BC,$71,$8F,$B6,$3C,$8F,$68,$F4,$CF,$93,$1E,$11,$E5,$9D,$3E
	.dc.b	$DB,$F5,$63,$E2,$8E,$2D,$74,$71,$8D,$B0,$C1,$2B,$B9,$31,$F6,$46
	.dc.b	$5F,$84,$F1,$8F,$4C,$FC,$08,$F9,$55,$E1,$C3,$D5,$7C,$AF,$75,$FC
	.dc.b	$D2,$7E,$2C,$F4,$8C,$61,$F1,$63,$81,$9A,$F2,$4F,$C6,$9F,$87,$5F
	.dc.b	$62,$3A,$5F,$F7,$EC,$7E,$49,$E0,$26,$1D,$DC,$BE,$E7,$CB,$87,$6A
	.dc.b	$5F,$A9,$F0,$63,$EF,$AE,$31,$F4,$0F,$BB,$11,$F1,$9F,$83,$C3,$C9
	.dc.b	$87,$73,$DB,$1C,$4F,$92,$7C,$BC,$3C,$EE,$6F,$BB,$F7,$1C,$BE,$6A
	.dc.b	$EA,$EE,$BB,$3E,$83,$A9,$2B,$DF,$1A,$8E,$47,$9C,$F5,$57,$6C,$3E
	.dc.b	$BC,$75,$3A,$47,$27,$D5,$5F,$26,$7E,$A3,$C6,$FC,$B3,$D8,$FB,$F8
	.dc.b	$76,$C3,$EA,$12,$3A,$BF,$19,$E1,$5D,$71,$FC,$09,$F8,$78,$F2,$1D
	.dc.b	$71,$F4,$46,$63,$15,$D5,$F9,$4F,$5C,$7E,$F4,$76,$8C,$4F,$E3,$B1
	.dc.b	$FC,$E1,$F9,$D7,$C3,$8F,$C7,$AF,$BE,$7B,$09,$8F,$C3,$AE,$13,$FE
	.dc.b	$44,$E1,$FF,$47,$FC,$A7,$F2,$F6,$DF,$FA,$A2,$31,$C2,$69,$23,$44
	.dc.b	$3F,$CA,$F3,$FE,$93,$F9,$6F,$F2,$FD,$FF,$9E,$1B,$C3,$FE,$C0,$6D
	.dc.b	$7F,$A9,$70,$EC,$FF,$9D,$3E,$63,$38,$39,$D1,$86,$CF,$FA,$A7,$A1
	.dc.b	$E2,$E6,$A4,$F3,$35,$1D,$3A,$DF,$13,$49,$3F,$E9,$4F,$80,$AF,$FA
	.dc.b	$4A,$2E,$DF,$E9,$CE,$E5,$79,$B5,$6D,$49,$87,$FA,$93,$C0,$AF,$FA
	.dc.b	$BA,$2E,$1F,$EB,$EF,$69,$E1,$7F,$AC,$6B,$FD,$2B,$F3,$1E,$F2,$3F
	.dc.b	$EB,$DB,$F5,$74,$48,$FF,$54,$BE,$3F,$43,$11,$FE,$AA,$8B,$8F,$FA
	.dc.b	$D3,$35,$7C,$BF,$84,$97,$FE,$A4,$7E,$8F,$91,$EC,$8C,$10,$FF,$57
	.dc.b	$C6,$B1,$FA,$FF,$30,$F8,$EF,$C6,$36,$BF,$D2,$BD,$08,$C4,$7E,$21
	.dc.b	$AF,$FA,$9A,$35,$3C,$89,$1C,$A7,$6F,$FA,$AF,$93,$F1,$2F,$FB,$23
	.dc.b	$AB,$EF,$1E,$4A,$FC,$47,$C4,$49,$FF,$60,$BF,$8E,$E5,$FF,$51,$46
	.dc.b	$08,$7F,$B3,$23,$5E,$07,$E2,$7C,$59,$3F,$ED,$87,$F2,$FF,$2E,$F9
	.dc.b	$8C,$CF,$FB,$6A,$6F,$1F,$FB,$92,$B5,$3F,$F0,$05,$7B,$BF,$B3,$6C
	.dc.b	$7C,$86,$BF,$AE,$6B,$F1,$DA,$FF,$B3,$5A,$B7,$46,$1B,$3F,$C0,$7B
	.dc.b	$DF,$23,$FA,$D2,$5F,$96,$7F,$51,$F1,$B2,$3F,$DE,$EF,$ED,$BD,$87
	.dc.b	$CC,$66,$7F,$DF,$33,$73,$78,$9F,$EE,$1B,$DD,$7F,$BC,$BE,$63,$F7
	.dc.b	$84,$7F,$C5,$75,$C1,$F5,$79,$DF,$ED,$57,$81,$0F,$F8,$03,$3F,$C2
	.dc.b	$23,$50,$9E,$5A,$BD,$C7,$07,$F2,$16,$D7,$E5,$9A,$FF,$1C,$6B,$FE
	.dc.b	$FA,$F7,$5F,$30,$F8,$BF,$E3,$71,$86,$C6,$67,$D7,$AF,$FC,$01,$BE
	.dc.b	$B9,$79,$3E,$03,$F8,$9D,$68,$FB,$0E,$1F,$55,$7F,$93,$B3,$4E,$6B
	.dc.b	$E3,$65,$E1,$BE,$55,$AF,$23,$F9,$5D,$EF,$0E,$4D,$A5,$F1,$7B,$9F
	.dc.b	$89,$F8,$1F,$82,$F8,$5A,$E7,$5F,$E9,$51,$8C,$7F,$9C,$7C,$C7,$F3
	.dc.b	$72,$57,$92,$7E,$0E,$3E,$BF,$84,$7B,$1F,$D4,$DB,$5F,$C2,$1A,$FF
	.dc.b	$4E,$6B,$FC,$1F,$A1,$F2,$FF,$D1,$49,$1F,$0A,$78,$EF,$F8,$2A,$DC
	.dc.b	$FC,$66,$BF,$03,$3F,$EA,$17,$FD,$4A,$F8,$E3,$F4,$6F,$78,$73,$AE
	.dc.b	$F8,$71,$BD,$BF,$B1,$F8,$77,$77,$F6,$F6,$C7,$F2,$46,$BF,$DA,$9A
	.dc.b	$FB,$7E,$63,$FB,$01,$1F,$BB,$1C,$30,$FD,$34,$7C,$2F,$07,$FD,$4F
	.dc.b	$8F,$F2,$68,$FF,$4A,$B5,$D1,$FE,$AF,$70,$48,$E2,$7F,$AC,$FE,$35
	.dc.b	$F3,$1F,$C1,$AB,$F2,$D1,$8E,$FF,$EF,$6D,$8F,$CD,$1D,$8F,$C5,$8F
	.dc.b	$94,$7C,$BF,$FA,$F1,$B1,$FE,$C1,$6B,$FE,$AC,$23,$87,$07,$38,$E4
	.dc.b	$47,$93,$FE,$CB,$C9,$F3,$1F,$D4,$1F,$E1,$86,$32,$FF,$09,$B6,$BD
	.dc.b	$37,$E2,$8D,$D7,$1E,$1F,$E2,$AD,$AF,$E1,$CD,$7C,$87,$C2,$F0,$3F
	.dc.b	$74,$FA,$1A,$FB,$5A,$FF,$B2,$9A,$9A,$4E,$AF,$F7,$1B,$C3,$D2,$7C
	.dc.b	$C3,$E6,$3F,$C5,$B1,$86,$C6,$6F,$C9,$7F,$E3,$2F,$7F,$EA,$61,$FE
	.dc.b	$F1,$6C,$FC,$E0,$E1,$FA,$2D,$7E,$B3,$F7,$7D,$3E,$5A,$9A,$49,$FF
	.dc.b	$7F,$7C,$C7,$E9,$3B,$EF,$47,$F1,$29,$F0,$4E,$5B,$19,$7F,$03,$0F
	.dc.b	$47,$F9,$FC,$7E,$1E,$31,$F0,$5D,$A7,$33,$FE,$F6,$9C,$BE,$AF,$95
	.dc.b	$FE,$B3,$C5,$EF,$D1,$18,$9D,$79,$5F,$AE,$7F,$DF,$F3,$A8,$C4,$7C
	.dc.b	$69,$7C,$C6,$2B,$9B,$63,$37,$DD,$B5,$F9,$CE,$37,$CC,$23,$F2,$F5
	.dc.b	$67,$31,$81,$A3,$E6,$3F,$CA,$B5,$FD,$2E,$B6,$FF,$4D,$9D,$7B,$BE
	.dc.b	$86,$C7,$DF,$6D,$FE,$B9,$AF,$EC,$DA,$F9,$5A,$F8,$AF,$C2,$F1,$6B
	.dc.b	$FB,$BB,$FF,$6E,$F9,$7E,$96,$DF,$F6,$1E,$9F,$A3,$E6,$67,$E1,$3F
	.dc.b	$B7,$4F,$AD,$B8,$7F,$56,$BE,$2F,$F5,$8B,$E3,$5B,$C3,$E3,$46,$27
	.dc.b	$F8,$4B,$5F,$F4,$BE,$AC,$7E,$DC,$E6,$7F,$C2,$73,$9B,$FD,$24,$E6
	.dc.b	$7F,$C1,$D3,$97,$AB,$5F,$E3,$5D,$CF,$98,$7C,$47,$D9,$8C,$3F,$0F
	.dc.b	$E5,$DF,$31,$9A,$DE,$5F,$6E,$72,$FE,$D7,$54,$63,$1F,$C5,$35,$FA
	.dc.b	$0D,$7F,$14,$D5,$1C,$AF,$F3,$6D,$AE,$D5,$BF,$2B,$FD,$69,$F3,$1F
	.dc.b	$91,$6C,$FF,$54,$F9,$8C,$79,$8A,$D7,$94,$9F,$D3,$DA,$31,$7F,$D8
	.dc.b	$E7,$31,$8C,$FF,$D9,$B8,$5E,$0F,$EF,$55,$E1,$F3,$3F,$CD,$38,$7F
	.dc.b	$7B,$8F,$09,$1F,$F0,$17,$AD,$70,$FE,$FD,$87,$39,$FE,$F9,$5C,$EB
	.dc.b	$58,$7F,$92,$30,$EB,$87,$F8,$46,$7D,$31,$88,$FF,$4C,$35,$FF,$03
	.dc.b	$7C,$C7,$C2,$3E,$37,$A0,$F9,$6C,$65,$FE,$23,$7E,$1C,$66,$3F,$D4
	.dc.b	$ED,$AF,$F7,$6B,$5F,$F5,$37,$A9,$FE,$5A,$9F,$F1,$FB,$63,$FC,$73
	.dc.b	$E2,$7E,$6F,$E3,$FE,$B4,$E7,$4F,$F2,$87,$AB,$83,$F8,$55,$6F,$0E
	.dc.b	$71,$8E,$A6,$14,$FC,$07,$AD,$E5,$C3,$FC,$A3,$3A,$BF,$F2,$8C,$F9
	.dc.b	$F2,$FF,$30,$69,$F8,$08,$C4,$7F,$A5,$5B,$C7,$FE,$FF,$F9,$8F,$C4
	.dc.b	$9F,$8D,$C3,$FD,$AF,$87,$F8,$DE,$32,$D3,$FC,$BF,$1A,$BF,$BB,$3F
	.dc.b	$95,$F8,$07,$CC,$53,$F3,$1E,$E7,$A7,$0F,$F6,$54,$65,$FF,$34,$CF
	.dc.b	$07,$FC,$D7,$3B,$BF,$F5,$64,$E5,$FC,$5F,$CC,$7E,$80,$FD,$27,$CC
	.dc.b	$3E,$5B,$18,$7F,$9D,$23,$0F,$FB,$91,$B5,$FE,$1C,$D1,$18,$8F,$F7
	.dc.b	$23,$5F,$F4,$13,$5F,$F5,$F3,$63,$8D,$7E,$BA,$31,$87,$6D,$FF,$D4
	.dc.b	$6D,$C3,$9F,$C6,$FC,$7B,$91,$F0,$9F,$5C,$63,$1F,$D4,$FC,$C7,$EF
	.dc.b	$8F,$E0,$5F,$0A,$F9,$1C,$61,$FE,$6D,$9B,$5F,$E6,$DA,$D4,$E4,$F9
	.dc.b	$FF,$31,$F5,$CF,$E1,$9F,$1B,$E5,$AF,$9F,$18,$CB,$F8,$4C,$FB,$C8
	.dc.b	$E0,$95,$FE,$19,$D9,$3A,$8C,$77,$BE,$63,$F8,$B9,$FC,$6F,$E3,$7F
	.dc.b	$C3,$13,$FE,$85,$8C,$B4,$FF,$56,$C6,$B2,$FE,$17,$3F,$8F,$F9,$89
	.dc.b	$5F,$EB,$2C,$3F,$C1,$5C,$3F,$42,$73,$7F,$E8,$FE,$14,$7C,$0F,$53
	.dc.b	$35,$FE,$97,$F9,$4F,$E5,$67,$C4,$F9,$3B,$97,$57,$0F,$F7,$FF,$93
	.dc.b	$E6,$E4,$99,$7F,$A7,$73,$57,$F7,$1A,$DC,$E6,$3F,$51,$DD,$2F,$FD
	.dc.b	$97,$C7,$FE,$B6,$7F,$4B,$19,$AF,$53,$F8,$2F,$88,$C6,$7F,$EA,$FE
	.dc.b	$DF,$E8,$67,$F4,$7D,$3F,$AD,$DF,$F3,$28,$CE,$1F,$EB,$37,$5E,$76
	.dc.b	$27,$F9,$BC,$FF,$4C,$3F,$0B,$E0,$D7,$E7,$3F,$84,$23,$F8,$1D,$BF
	.dc.b	$92,$4F,$FA,$2F,$97,$F0,$75,$A8,$FD,$84,$E9,$FF,$77,$35,$FF,$47
	.dc.b	$FC,$C7,$F8,$CF,$E5,$BF,$CB,$91,$FE,$23,$C3,$1E,$F7,$F9,$9E,$AC
	.dc.b	$E4,$57,$07,$FD,$39,$1F,$E3,$DF,$90,$C4,$7F,$8E,$AF,$17,$CE,$13
	.dc.b	$AF,$61,$5A,$9F,$29,$DA,$BF,$C7,$11,$E2,$8E,$E6,$4F,$26,$1E,$05
	.dc.b	$86,$DF,$11,$B3,$8C,$78,$67,$51,$C0,$F2,$D7,$81,$A7,$93,$0E,$0E
	.dc.b	$0F,$1B,$83,$C5,$7E,$B8,$FF,$26,$55,$7F,$CC,$51,$E5,$C3,$E0,$C7
	.dc.b	$94,$C9,$5C,$E1,$D2,$3C,$07,$B2,$F0,$78,$70,$F9,$CD,$8F,$44,$7A
	.dc.b	$E3,$94,$FA,$8B,$1E,$F7,$9B,$B3,$99,$67,$99,$F8,$39,$DB,$63,$EF
	.dc.b	$4F,$56,$9D,$4E,$71,$B3,$99,$F1,$8F,$69,$F4,$E7,$D4,$FC,$E8,$F1
	.dc.b	$3F,$69,$E8,$FE,$22,$7C,$0F,$43,$E3,$47,$81,$ED,$3E,$86,$C6,$5D
	.dc.b	$9C,$DE,$AF,$73,$F1,$EE,$EB,$EC,$CF,$39,$E2,$79,$CF,$71,$4E,$E7
	.dc.b	$53,$99,$4F,$85,$5C,$E7,$AD,$77,$C3,$F1,$55,$F9,$7A,$F7,$BE,$53
	.dc.b	$8C,$FD,$07,$A9,$93,$F4,$71,$F8,$07,$27,$E4,$1E,$6D,$8F,$82,$D3
	.dc.b	$67,$85,$CB,$4C,$E1,$F9,$DA,$FD,$6B,$63,$F3,$C7,$DD,$9F,$B6,$FE
	.dc.b	$83,$0F,$BE,$F7,$3D,$67,$C5,$3D,$31,$ED,$72,$7B,$70,$FD,$3C,$7F
	.dc.b	$0F,$8F,$F8,$00,$7E,$F4,$B5,$87,$F5,$E7,$FB,$54,$FF,$6B,$7A,$EF
	.dc.b	$98,$7B,$BC,$A3,$0D,$8C,$CF,$13,$8B,$87,$12,$F9,$87,$1F,$36,$CF
	.dc.b	$11,$3B,$9D,$12,$3F,$8D,$47,$F8,$55,$FE,$48,$D7,$33,$D3,$5F,$E6
	.dc.b	$D3,$CE,$78,$F8,$5E,$F9,$CB,$73,$F8,$55,$B9,$E3,$7E,$6D,$77,$87
	.dc.b	$F1,$DA,$F0,$46,$B1,$C5,$78,$70,$E7,$8F,$5C,$3E,$6B,$5F,$7C,$ED
	.dc.b	$FA,$2D,$7E,$1C,$6F,$0F,$76,$3D,$E7,$FA,$17,$0F,$F9,$F3,$E4,$3E
	.dc.b	$D6,$5F,$CB,$23,$12,$3E,$8F,$05,$8C,$C6,$03,$FE,$86,$F8,$FE,$73
	.dc.b	$BE,$86,$18,$BD,$C6,$1E,$D8,$78,$63,$11,$D7,$0E,$8E,$AF,$C5,$E0
	.dc.b	$BE,$8F,$A5,$23,$15,$B3,$CA,$E1,$FF,$60,$FD,$27,$08,$C6,$3F,$05
	.dc.b	$B8,$76,$C3,$55,$E9,$FC,$2F,$E2,$7A,$1F,$A6,$6E,$3F,$93,$6B,$F9
	.dc.b	$D6,$BF,$A7,$7F,$C0,$0F,$95,$AF,$CB,$6E,$3F,$D7,$9E,$6D,$87,$C7
	.dc.b	$78,$5A,$78,$4A,$78,$4A,$78,$4A,$7D,$D9,$FB,$D1,$CC,$91,$CF,$E0
	.dc.b	$BE,$02,$69,$88,$CF,$BB,$E3,$24,$F8,$DA,$F8,$DA,$F8,$DF,$57,$81
	.dc.b	$D8,$91,$F1,$44,$78,$BE,$7B,$E0,$12,$3E,$03,$5F,$80,$D7,$E0,$3F
	.dc.b	$1F,$97,$D6,$48,$ED,$1E,$EE,$6F,$98,$48,$F9,$8D,$7E,$63,$5F,$98
	.dc.b	$FD,$3E,$9F,$8E,$48,$F5,$47,$CA,$FA,$FF,$B0,$48,$FB,$0D,$7E,$C3
	.dc.b	$5F,$B0,$FD,$FF,$BB,$FA,$64,$8F,$8D,$1F,$5B,$DB,$FC,$21,$23,$F0
	.dc.b	$8D,$7F,$08,$D7,$F0,$8F,$E2,$FE,$F3,$EF,$92,$3E,$94,$7E,$0B,$EB
	.dc.b	$24,$7D,$57,$F0,$BF,$23,$F8,$6F,$87,$FC,$A1,$27,$F2,$8F,$E5,$FE
	.dc.b	$87,$F1,$64,$8F,$BD,$F2,$9F,$BA,$FD,$A5,$E4,$7F,$52,$ED,$46,$2B
	.dc.b	$FD,$78,$D7,$FD,$D0,$D7,$FC,$FC,$D7,$F8,$C3,$5F,$E2,$ED,$8E,$B3
	.dc.b	$FA,$BF,$9C,$FC,$F9,$2B,$F8,$54,$E9,$FC,$A7,$E0,$3F,$A4,$23,$FB
	.dc.b	$BF,$1F,$FA,$E7,$5F,$ED,$E7,$33,$FB,$D0,$FF,$0B,$9D,$F7,$FE,$CD
	.dc.b	$79,$39,$25,$75,$FC,$0E,$BF,$C1,$E7,$6F,$86,$7A,$4F,$CB,$F2,$24
	.dc.b	$FA,$B5,$FF,$49,$65,$FC,$FD,$7C,$B6,$A7,$F9,$84,$6B,$2F,$93,$7F
	.dc.b	$56,$BD,$1F,$45,$2F,$E0,$3C,$70,$FE,$31,$9B,$E8,$BD,$C4,$AF,$8D
	.dc.b	$75,$F5,$45,$AE,$33,$F1,$70,$95,$F3,$5F,$91,$F4,$38,$8F,$85,$16
	.dc.b	$BC,$33,$F8,$BD,$3F,$15,$A1,$A4,$BF,$C5,$B5,$FC,$5B,$5F,$E7,$19
	.dc.b	$7E,$3E,$FD,$BC,$3F,$64,$ED,$3F,$81,$F9,$9F,$B8,$47,$F0,$9F,$2B
	.dc.b	$5F,$B1,$16,$BE,$3C,$FE,$0F,$D6,$95,$F8,$B7,$FA,$EF,$3B,$9B,$33
	.dc.b	$F8,$19,$B7,$FD,$E6,$B5,$3F,$98,$5D,$5F,$AB,$6C,$7F,$8A,$79,$91
	.dc.b	$88,$FF,$1A,$46,$27,$81,$23,$84,$ED,$FF,$76,$35,$FC,$D7,$C0,$7E
	.dc.b	$74,$8F,$8E,$7F,$41,$EC,$C8,$FD,$43,$FD,$EB,$DE,$7C,$16,$67,$F3
	.dc.b	$D3,$72,$FF,$0B,$56,$A7,$F6,$1E,$BF,$EA,$C8,$FE,$E3,$E0,$DC,$18
	.dc.b	$8F,$D4,$4D,$BF,$F1,$05,$6A,$7F,$75,$F0,$FF,$DF,$78,$CD,$24,$7F
	.dc.b	$94,$7E,$A1,$F8,$13,$E8,$7F,$C3,$D1,$86,$C6,$67,$A9,$FC,$1F,$ED
	.dc.b	$7F,$81,$11,$FF,$0C,$57,$17,$44,$3F,$80,$47,$4A,$FE,$15,$F1,$3E
	.dc.b	$72,$4F,$F8,$4F,$14,$7A,$F6,$FE,$20,$49,$FE,$31,$1F,$E3,$57,$FC
	.dc.b	$E1,$EC,$FF,$9F,$FC,$B3,$49,$1F,$89,$F6,$1E,$B7,$F8,$C9,$B1,$FC
	.dc.b	$A1,$AF,$F9,$1C,$EA,$66,$38,$11,$FF,$1C,$1A,$3F,$C8,$51,$CB,$83
	.dc.b	$FC,$8D,$3B,$9E,$97,$E6,$7F,$C8,$19,$3F,$37,$FF,$0C,$47,$FA,$1E
	.dc.b	$31,$EA,$FF,$A7,$9B,$19,$9D,$BF,$E3,$98,$7C,$47,$F3,$72,$47,$69
	.dc.b	$DE,$3F,$C8,$3E,$17,$F1,$04,$9F,$A4,$D7,$A3,$5F,$E1,$AF,$E3,$F3
	.dc.b	$FE,$88,$48,$FE,$99,$1F,$E5,$EF,$DA,$5F,$15,$FD,$95,$B3,$FB,$B6
	.dc.b	$BF,$BA,$6B,$FE,$B7,$EE,$46,$23,$FC,$F8,$D7,$FC,$F6,$D8,$DC,$FF
	.dc.b	$97,$F0,$79,$5F,$E6,$36,$DF,$F9,$6D,$AF,$F9,$E6,$FC,$71,$FE,$5B
	.dc.b	$32,$59,$E4,$7B,$6B,$DD,$F1,$9E,$D2,$5F,$B5,$AF,$B5,$AF,$B6,$D1
	.dc.b	$E7,$FA,$9F,$31,$27,$FB,$DC,$E9,$F4,$7D,$4F,$B4,$8F,$F9,$EF,$14
	.dc.b	$7C,$1F,$9F,$93,$96,$BF,$ED,$3F,$9F,$7D,$03,$E0,$7F,$DD,$71,$86
	.dc.b	$C6,$67,$83,$5F,$F7,$24,$AB,$FB,$EC,$61,$AF,$FB,$C6,$35,$E9,$76
	.dc.b	$9C,$CB,$E0,$7F,$DD,$37,$F5,$23,$FD,$01,$97,$9B,$7F,$74,$E6,$78
	.dc.b	$4E,$5F,$F4,$D2,$FA,$4D,$5F,$FB,$D1,$B1,$9B,$F9,$11,$8F,$85,$FC
	.dc.b	$2C,$FE,$05,$B3,$FE,$CB,$F9,$07,$9D,$8C,$7C,$35,$79,$15,$AB,$FE
	.dc.b	$B0,$DA,$FE,$AE,$D7,$F3,$CD,$7B,$5F,$B6,$B9,$B6,$75,$3F,$ED,$A6
	.dc.b	$E4,$C7,$F9,$0E,$1F,$9D,$ED,$F2,$57,$F4,$C8,$C1,$2F,$FD,$75,$5C
	.dc.b	$1F,$F6,$D4,$ED,$CD,$72,$F1,$FC,$44,$BF,$6B,$67,$DA,$D7,$DA,$FD
	.dc.b	$0F,$A6,$F9,$18,$F9,$DB,$8F,$F5,$3B,$FD,$A4,$63,$C2,$FA,$E4,$8F
	.dc.b	$CF,$37,$0F,$BE,$D7,$FD,$B1,$5C,$DF,$C1,$FC,$DF,$63,$EC,$77,$3F
	.dc.b	$34,$FE,$2B,$87,$F0,$39,$CC,$62,$84,$15,$C7,$AF,$F2,$4D,$BF,$E7
	.dc.b	$15,$C6,$BF,$8D,$7C,$9B,$CC,$FC,$33,$6B,$FB,$4F,$DC,$BD,$DC,$D4
	.dc.b	$9F,$F3,$DD,$6E,$FE,$F4,$79,$3C,$A6,$1C,$63,$77,$C6,$B6,$39,$58
	.dc.b	$73,$DB,$F9,$1E,$3C,$B2,$E1,$18,$CB,$A6,$1F,$04,$56,$F2,$FF,$06
	.dc.b	$4E,$A7,$E3,$4E,$5F,$D9,$FC,$E7,$DE,$3E,$FF,$CD,$DB,$FE,$64,$FE
	.dc.b	$D0,$8F,$F1,$B7,$F6,$FE,$17,$E1,$8F,$D8,$5F,$F8,$6B,$E3,$31,$5F
	.dc.b	$CE,$E6,$E1,$F0,$2B,$53,$93,$FC,$2F,$E7,$2F,$9B,$63,$FA,$27,$A5
	.dc.b	$89,$FC,$AB,$FC,$20,$91,$FC,$BD,$FE,$17,$C7,$F9,$A3,$F8,$17,$A3
	.dc.b	$FD,$CA,$7F,$A7,$46,$5A,$7F,$8F,$FC,$AC,$65,$FB,$2E,$CF,$D1,$9F
	.dc.b	$C2,$FC,$FF,$F0,$BD,$7F,$57,$8C,$B4,$FF,$5F,$C6,$A7,$94,$7F,$5D
	.dc.b	$F9,$4F,$D6,$9F,$C5,$7F,$6B,$7C,$46,$27,$FD,$8F,$18,$7F,$CD,$CD
	.dc.b	$AF,$F3,$6B,$5F,$EE,$8D,$7F,$4B,$CE,$8C,$47,$E3,$DB,$19,$AF,$E3
	.dc.b	$91,$AE,$EF,$E3,$1F,$2B,$FC,$08,$F3,$7C,$5D,$C7,$FB,$B4,$5C,$7F
	.dc.b	$37,$8F,$7A,$C7,$47,$F7,$7E,$2F,$E1,$67,$F0,$CF,$4E,$E3,$FD,$FE
	.dc.b	$2E,$1F,$C8,$26,$D7,$F8,$13,$27,$D0,$7F,$63,$C3,$F8,$0C,$E6,$31
	.dc.b	$7E,$2D,$1F,$29,$FC,$72,$73,$5F,$E0,$ED,$3E,$24,$FC,$02,$61,$F3
	.dc.b	$9B,$87,$CE,$6B,$F3,$9F,$EB,$37,$DA,$7C,$67,$F1,$4C,$7C,$F3,$E1
	.dc.b	$F9,$1F,$13,$27,$77,$FD,$B6,$F9,$3F,$C1,$2F,$DF,$7B,$8C,$4B,$D5
	.dc.b	$F6,$1E,$DF,$90,$B8,$74,$C3,$E3,$CE,$2F,$8D,$5C,$FE,$47,$C8,$FC
	.dc.b	$53,$FB,$37,$ED,$6E,$2C,$65,$E6,$D9,$39,$8C,$37,$1F,$C0,$AE,$4E
	.dc.b	$77,$2B,$6F,$68,$C5,$7B,$38,$FE,$B1,$F5,$FE,$4A,$DF,$AF,$74,$E2
	.dc.b	$BE,$1D,$5D,$3F,$05,$ED,$7D,$F3,$F0,$1F,$25,$71,$E5,$16,$FE,$5C
	.dc.b	$DA,$FC,$57,$C9,$7E,$1C,$F8,$7D,$57,$0F,$7B,$FE,$01,$DB,$FC,$17
	.dc.b	$1F,$EE,$16,$E1,$FE,$E0,$6B,$FE,$67,$6B,$FB,$46,$BF,$D0,$5B,$BF
	.dc.b	$CE,$AF,$C6,$F4,$FE,$64,$FC,$D7,$C0,$DC,$3E,$27,$C9,$7E,$48,$8F
	.dc.b	$F4,$3C,$7F,$19,$1F,$96,$F9,$2F,$D1,$1F,$A3,$F9,$2B,$5F,$37,$E1
	.dc.b	$BF,$4C,$7E,$9F,$E4,$AC,$7D,$17,$F3,$E4,$7B,$75,$7E,$D2,$31,$8F
	.dc.b	$F5,$38,$C4,$EB,$95,$5A,$AF,$EF,$15,$FB,$CF,$89,$FD,$B1,$FB,$7F
	.dc.b	$32,$E1,$F8,$17,$F5,$E4,$7A,$36,$BC,$18,$7F,$5F,$7F,$64,$D7,$F5
	.dc.b	$D1,$FB,$8F,$92,$FE,$08,$7F,$88,$3D,$EF,$F0,$5D,$FE,$3E,$31,$5F
	.dc.b	$E5,$69,$CB,$ED,$6B,$EA,$6B,$FC,$05,$AF,$F9,$A5,$B8,$FF,$9D,$EB
	.dc.b	$E4,$C6,$07,$BB,$E3,$3C,$9F,$19,$67,$F7,$92,$8F,$21,$23,$9B,$6F
	.dc.b	$CC,$D7,$CC,$FA,$FE,$4B,$E0,$1F,$07,$E4,$AD,$76,$7E,$01,$1F,$3B
	.dc.b	$6B,$E1,$35,$F8,$4F,$C7,$F9,$2F,$98,$7C,$DF,$92,$B5,$EA,$7E,$61
	.dc.b	$1F,$86,$DA,$F9,$CD,$7E,$73,$F4,$FE,$4B,$EC,$1F,$67,$E4,$6D,$7C
	.dc.b	$67,$EC,$11,$F9,$EF,$DB,$F7,$BF,$00,$7E,$07,$E0,$2C,$FC,$E8,$B8
	.dc.b	$7F,$33,$D7,$FD,$0D,$DF,$FD,$4D,$FF,$39,$37,$E4,$7F,$14,$7E,$2F
	.dc.b	$E4,$6E,$1F,$62,$2D,$F3,$9B,$5F,$EA,$7F,$96,$FF,$2D,$FC,$AB,$6F
	.dc.b	$F1,$E7,$1F,$FB,$92,$31,$87,$F8,$FE,$B7,$5A,$CA,$CE,$45,$70,$7F
	.dc.b	$CC,$F7,$CE,$BF,$83,$5F,$09,$CD,$D7,$FC,$DB,$1F,$E6,$EF,$3B,$94
	.dc.b	$7E,$5A,$71,$5C,$A3,$FB,$7B,$E3,$69,$E3,$C3,$8E,$1E,$18,$C4,$EE
	.dc.b	$7A,$C3,$0F,$15,$6E,$3F,$CF,$D1,$8B,$F3,$46,$CF,$E4,$07,$A0,$A6
	.dc.b	$4C,$C6,$BE,$59,$E3,$7F,$A2,$27,$F9,$44,$63,$E0,$75,$39,$9F,$F3
	.dc.b	$74,$E5,$DF,$91,$FE,$BD,$6E,$BD,$63,$11,$C1,$AF,$FA,$5E,$51,$C7
	.dc.b	$3F,$AD,$0F,$27,$FD,$39,$18,$6E,$BC,$6B,$83,$70,$FB,$57,$C7,$0B
	.dc.b	$7C,$1F,$F5,$3C,$78,$27,$77,$E9,$DB,$B5,$7D,$C3,$77,$FE,$AD,$8F
	.dc.b	$2C,$EF,$0F,$05,$7E,$39,$F6,$57,$78,$61,$E3,$74,$78,$AB,$D2,$DC
	.dc.b	$3C,$EC,$36,$72,$8D,$F8,$EF,$8F,$7C,$3F,$FB,$06,$FF,$C7,$91,$8B
	.dc.b	$FF,$0C,$35,$FF,$2A,$AF,$87,$E1,$C3,$FE,$66,$8D,$57,$2B,$CD,$FF
	.dc.b	$AD,$67,$35,$BB,$CE,$0F,$96,$FF,$4C,$5F,$F8,$9D,$AF,$F9,$91,$B8
	.dc.b	$FF,$94,$57,$AD,$C9,$CE,$1F,$E7,$86,$D7,$FB,$15,$B7,$DE,$BC,$BE
	.dc.b	$07,$FB,$8B,$A3,$F4,$2D,$BF,$EE,$A2,$32,$D3,$3C,$3C,$9F,$F3,$39
	.dc.b	$FE,$49,$29,$FE,$53,$38,$17,$D5,$E6,$47,$BB,$5F,$F0,$10,$70,$FF
	.dc.b	$80,$0C,$1E,$6F,$57,$C4,$47,$C5,$1E,$5F,$57,$D2,$48,$F6,$B5,$EE
	.dc.b	$D7,$B3,$F0,$3D,$5F,$71,$23,$DD,$1F,$07,$D3,$F8,$A4,$8F,$92,$D7
	.dc.b	$D6,$D7,$D4,$FC,$CE,$5F,$94,$48,$F9,$51,$F3,$79,$3E,$89,$23,$EA
	.dc.b	$B5,$F8,$ED,$7E,$33,$F6,$3A,$3E,$B1,$23,$EB,$47,$D9,$F4,$FE,$E9
	.dc.b	$23,$F0,$2D,$7E,$9B,$5F,$A4,$FE,$13,$D1,$FC,$11,$23,$F0,$51,$F8
	.dc.b	$5F,$47,$F1,$24,$8F,$C6,$B5,$FB,$ED,$7E,$F3,$F9,$0F,$47,$F1,$C4
	.dc.b	$8F,$C7,$47,$E4,$7D,$19,$1F,$91,$7F,$29,$E8,$FE,$57,$D1,$FC,$C9
	.dc.b	$27,$F3,$AF,$E7,$3D,$1F,$CD,$12,$3F,$35,$E8,$FE,$DB,$E5,$9E,$57
	.dc.b	$F4,$6E,$A4,$62,$BF,$CD,$11,$89,$D4,$E6,$44,$13,$B7,$8C,$6D,$D4
	.dc.b	$65,$E0,$70,$8D,$3F,$EC,$71,$E8,$FE,$CC,$8F,$FA,$C6,$3C,$5D,$DF
	.dc.b	$EE,$38,$7A,$3F,$B9,$23,$FE,$C1,$F1,$BB,$CE,$E7,$33,$FB,$B7,$F8
	.dc.b	$4C,$EB,$C0,$FF,$6C,$3F,$E5,$29,$E5,$E4,$68,$DF,$A3,$63,$24,$3F
	.dc.b	$DB,$13,$FE,$55,$7D,$1E,$3F,$D1,$23,$EB,$F3,$EA,$CB,$FD,$E9,$9F
	.dc.b	$F8,$5D,$B5,$F6,$6B,$CF,$E8,$CB,$F7,$3C,$A9,$9F,$AB,$D1,$F6,$92
	.dc.b	$BE,$25,$D7,$CF,$16,$B8,$C3,$C4,$66,$BF,$8F,$94,$AF,$9C,$FC,$AF
	.dc.b	$47,$11,$F0,$A2,$E1,$D3,$0E,$75,$F9,$3D,$7F,$1A,$4B,$FC,$4B,$5F
	.dc.b	$C3,$F0,$FE,$5A,$31,$1F,$8C,$DB,$E7,$3F,$EA,$FC,$FF,$0B,$AF,$DA
	.dc.b	$E2,$FB,$A7,$79,$FC,$27,$A1,$F8,$02,$3F,$87,$F4,$6B,$F6,$A2,$EB
	.dc.b	$E9,$C1,$5F,$88,$F4,$65,$FE,$3D,$FE,$C7,$D0,$E7,$CC,$FE,$12,$6E
	.dc.b	$1F,$3B,$65,$EE,$BF,$5B,$39,$87,$5F,$ED,$DB,$3F,$B6,$6B,$F1,$9A
	.dc.b	$FE,$C1,$AF,$EE,$2B,$BC,$F3,$24,$7F,$B6,$B2,$7A,$3F,$A4,$24,$79
	.dc.b	$36,$E5,$87,$83,$1F,$89,$F0,$BF,$AA,$25,$6C,$FE,$0B,$ED,$3D,$EC
	.dc.b	$C7,$E9,$A6,$D7,$C3,$AD,$4F,$EE,$7E,$5B,$F6,$A4,$7A,$6A,$F7,$DC
	.dc.b	$59,$9F,$D8,$D5,$C3,$E7,$56,$A7,$F8,$27,$CB,$7F,$81,$B7,$34,$91
	.dc.b	$FC,$58,$7C,$13,$E0,$9F,$15,$FE,$24,$8C,$36,$33,$3D,$8F,$E2,$5F
	.dc.b	$2D,$FC,$38,$8F,$C5,$AE,$2E,$88,$7F,$0A,$8E,$95,$FC,$5B,$E5,$BD
	.dc.b	$04,$9F,$B0,$FA,$F7,$FE,$36,$48,$FE,$45,$1F,$E3,$A7,$FC,$DF,$EB
	.dc.b	$FF,$A1,$3C,$CF,$F2,$14,$7F,$A2,$1A,$FC,$16,$BF,$CE,$BD,$8F,$F1
	.dc.b	$BB,$FE,$4C,$F5,$FF,$C8,$31,$88,$FE,$35,$CF,$FD,$16,$B9,$F0,$FF
	.dc.b	$95,$2F,$64,$AF,$C3,$1E,$43,$A1,$0F,$F2,$BB,$E9,$9E,$75,$FD,$4B
	.dc.b	$81,$F2,$CF,$5B,$FD,$43,$18,$6C,$67,$0F,$B9,$39,$7F,$41,$F0,$BF
	.dc.b	$D0,$88,$FE,$66,$34,$FF,$3C,$F9,$6F,$C4,$11,$FD,$43,$5F,$E2,$4D
	.dc.b	$7F,$B3,$78,$BF,$CA,$A3,$F2,$1A,$FF,$50,$24,$FF,$57,$8F,$F3,$5F
	.dc.b	$CA,$BE,$23,$FB,$B3,$67,$FB,$AB,$5F,$E4,$AD,$7F,$B7,$B5,$FE,$00
	.dc.b	$D7,$FB,$BB,$7A,$CE,$70,$FB,$33,$FB,$96,$B8,$6F,$07,$F1,$99,$F1
	.dc.b	$3F,$CF,$4F,$E9,$33,$B6,$99,$DB,$DB,$87,$BB,$E2,$7D,$A4,$AF,$6B
	.dc.b	$5F,$6B,$5F,$6B,$FE,$7E,$A7,$CB,$7F,$A0,$9B,$3E,$96,$BE,$CC,$BD
	.dc.b	$F5,$BF,$96,$E4,$49,$FE,$01,$9F,$F7,$38,$ED,$F2,$B2,$73,$82,$31
	.dc.b	$1F,$C1,$FE,$59,$F2,$DF,$E8,$68,$FF,$13,$46,$3E,$0B,$FD,$EA,$D8
	.dc.b	$E1,$3B,$7F,$C5,$B1,$D7,$D3,$DD,$FF,$84,$E7,$FC,$53,$39,$7C,$11
	.dc.b	$97,$FC,$6B,$97,$F8,$2E,$FE,$9C,$63,$1F,$2E,$FE,$A9,$CC,$FF,$7F
	.dc.b	$9C,$BF,$E3,$6A,$FF,$0A,$7C,$B6,$9B,$7F,$0A,$31,$46,$FC,$2F,$F8
	.dc.b	$2A,$3F,$08,$D9,$FE,$A5,$F1,$5F,$DA,$9F,$AD,$19,$C3,$FC,$97,$39
	.dc.b	$7F,$C0,$2D,$7F,$C9,$6D,$79,$F9,$DE,$AE,$14,$E6,$79,$CE,$63,$13
	.dc.b	$A8,$F4,$F6,$F8,$73,$FF,$19,$3E,$39,$E7,$E3,$E7,$3F,$F7,$46,$7F
	.dc.b	$8D,$AC,$1B,$C7,$FA,$9C,$64,$FE,$AE,$E3,$2F,$5D,$F8,$C9,$3F,$E7
	.dc.b	$09,$D4,$FB,$A7,$2F,$09,$FC,$8F,$CB,$78,$CF,$27,$85,$63,$F4,$0F
	.dc.b	$D4,$23,$F9,$86,$CF,$DD,$C7,$FA,$DC,$E4,$91,$9F,$96,$F8,$07,$D5
	.dc.b	$EC,$7C,$B7,$F1,$5A,$FD,$74,$E6,$31,$5F,$E7,$EA,$DF,$57,$E3,$9B
	.dc.b	$7F,$CC,$AB,$75,$FE,$86,$F9,$67,$99,$F4,$5B,$5F,$D9,$FE,$5B,$FD
	.dc.b	$0A,$FF,$80,$A3,$35,$FB,$D6,$E7,$FE,$2D,$6C,$FF,$A6,$3C,$95,$F2
	.dc.b	$BE,$18,$77,$25,$77,$F7,$6C,$FF,$24,$BE,$31,$FA,$86,$CF,$F8,$1D
	.dc.b	$B3,$FE,$9E,$6B,$FD,$0B,$17,$CB,$7E,$2C,$FC,$67,$CB,$5B,$FE,$30
	.dc.b	$FE,$B0,$8F,$F1,$07,$F5,$FE,$07,$E4,$CF,$D4,$61,$FE,$15,$F9,$6C
	.dc.b	$57,$F2,$19,$B8,$FA,$7C,$E5,$63,$2F,$F3,$FE,$1C,$70,$FF,$0A,$56
	.dc.b	$9F,$E5,$5E,$97,$68,$C6,$1F,$CF,$63,$F8,$09,$27,$F9,$2B,$FC,$13
	.dc.b	$8B,$F5,$07,$EA,$FD,$17,$93,$8A,$FE,$6D,$36,$FF,$9B,$56,$A7,$27
	.dc.b	$F3,$BE,$9F,$D9,$1F,$B5,$F3,$DE,$76,$27,$44,$BF,$E8,$13,$A8,$C6
	.dc.b	$BF,$D5,$FE,$53,$F7,$87,$EF,$FE,$59,$F1,$18,$AF,$F2,$7C,$61,$FF
	.dc.b	$72,$B7,$0F,$F4,$AB,$5F,$F6,$FB,$5F,$F3,$83,$5F,$F7,$3B,$7C,$03
	.dc.b	$4D,$78,$3F,$EC,$3F,$96,$FE,$20,$79,$7E,$2E,$E1,$FD,$76,$2E,$1F
	.dc.b	$D7,$6D,$58,$CF,$FB,$A7,$2F,$F1,$93,$F8,$D6,$D7,$1F,$EC,$D1,$70
	.dc.b	$FE,$F1,$36,$BF,$D7,$1F,$2A,$F3,$3F,$93,$37,$0F,$8B,$F2,$38,$9F
	.dc.b	$94,$FF,$33,$CC,$D2,$4F,$CA,$6E,$3F,$29,$AF,$F8,$A2,$F8,$CF,$F8
	.dc.b	$4D,$B1,$FE,$13,$5F,$23,$E4,$A9,$5B,$7C,$F1,$8A,$E1,$EA,$FA,$8F
	.dc.b	$5F,$C8,$5B,$ED,$16,$7B,$4D,$CF,$E1,$FC,$8F,$C2,$3F,$A3,$7C,$B3
	.dc.b	$87,$18,$FF,$A5,$A3,$0D,$8D,$F2,$F0,$64,$64,$E1,$18,$C7,$DD,$C5
	.dc.b	$F4,$0F,$A3,$F2,$56,$BE,$24,$59,$F8,$93,$72,$FB,$3E,$D7,$D8,$3E
	.dc.b	$CF,$C9,$5C,$3E,$64,$5B,$F9,$93,$6B,$F0,$3F,$25,$F8,$03,$DF,$D7
	.dc.b	$70,$F7,$3F,$DC,$75,$FE,$E3,$1F,$EA,$9E,$94,$63,$1F,$F7,$DB,$51
	.dc.b	$FE,$D5,$70,$DA,$E8,$4A,$FF,$1A,$F2,$7E,$C7,$1F,$CA,$F5,$7E,$53
	.dc.b	$E0,$FF,$2A,$FE,$5B,$E4,$BF,$C3,$14,$69,$2F,$9F,$C9,$7E,$70,$FC
	.dc.b	$EF,$C9,$58,$FC,$87,$C2,$FE,$80,$FD,$0F,$C9,$58,$FC,$A3,$D8,$8F
	.dc.b	$99,$BA,$FF,$8A,$F1,$46,$23,$F3,$CF,$9B,$E2,$7F,$5A,$7E,$BB,$CC
	.dc.b	$B7,$FA,$17,$D6,$47,$E1,$B6,$BB,$35,$FD,$53,$F0,$BE,$4B,$F7,$27
	.dc.b	$F8,$7F,$C0,$FA,$35,$FA,$D8,$C5,$7F,$8F,$E7,$2F,$F9,$0D,$AF,$FA
	.dc.b	$71,$AF,$F9,$41,$AB,$C3,$FE,$10,$F3,$25,$7F,$9B,$A5,$EF,$78,$CF
	.dc.b	$27,$C6,$D9,$FF,$49,$4A,$3C,$84,$8E,$4D,$AF,$33,$5F,$33,$EB,$F9
	.dc.b	$2F,$80,$7C,$1F,$92,$B5,$D9,$F8,$04,$7C,$CD,$AF,$84,$D7,$E1,$3F
	.dc.b	$1F,$E4,$BE,$61,$F3,$7E,$4A,$D7,$A9,$F9,$84,$7E,$13,$6B,$E7,$35
	.dc.b	$F9,$CF,$D3,$F9,$2F,$B0,$7D,$9F,$91,$B5,$F1,$9F,$B0,$47,$E7,$3F
	.dc.b	$6F,$DE,$FC,$01,$F8,$1F,$80,$B3,$F3,$A2,$E1,$FE,$BF,$D7,$FC,$E5
	.dc.b	$D6,$AF,$BD,$5F,$89,$F9,$19,$1F,$5B,$E4,$6E,$1C,$66,$DF,$18,$FC
	.dc.b	$26,$0A,$CD,$FF,$02,$99,$F2,$DF,$C1,$3E,$59,$B6,$67,$F0,$F5,$98
	.dc.b	$C4,$2B,$0F,$3E,$3F,$05,$F2,$8E,$53,$A7,$F8,$8C,$EA,$B9,$CE,$67
	.dc.b	$B8,$AE,$4F,$98,$FE,$BB,$3E,$5A,$F7,$79,$DA,$9F,$E3,$B7,$FC,$7A
	.dc.b	$31,$3A,$AE,$F5,$D5,$E5,$87,$86,$3D,$86,$4F,$63,$9B,$F3,$57,$A6
	.dc.b	$7D,$A6,$CF,$6C,$ED,$F8,$E7,$90,$A7,$63,$33,$AF,$96,$78,$CF,$81
	.dc.b	$C3,$6B,$E8,$CE,$A7,$27,$1F,$21,$BD,$25,$70,$8C,$B6,$3E,$A5,$3C
	.dc.b	$9F,$C2,$6D,$F4,$23,$15,$AB,$E3,$39,$6E,$BF,$8B,$AD,$D7,$C3,$6D
	.dc.b	$6F,$0E,$47,$D5,$AE,$18,$7E,$4F,$6F,$03,$63,$A6,$3E,$89,$E5,$8F
	.dc.b	$5A,$F3,$D7,$28,$EA,$78,$6B,$CA,$FF,$60,$C7,$C7,$3B,$8D,$61,$F9
	.dc.b	$AF,$21,$F2,$DF,$E7,$89,$FE,$CF,$3F,$AB,$9C,$FA,$9F,$A9,$8C,$56
	.dc.b	$EB,$E8,$F0,$7F,$6E,$BE,$7E,$A6,$E3,$15,$BB,$E3,$86,$70,$61,$D9
	.dc.b	$F0,$DF,$0C,$35,$2F,$90,$7C,$36,$1B,$7C,$0E,$6E,$1F,$29,$8F,$86
	.dc.b	$E1,$32,$38,$57,$58,$E9,$3E,$5F,$79,$F8,$57,$F9,$E2,$FF,$D6,$4D
	.dc.b	$7F,$D6,$4D,$7F,$87,$47,$0F,$37,$F4,$7E,$27,$F9,$77,$E0,$FE,$03
	.dc.b	$6B,$FA,$8C,$77,$BE,$9B,$FF,$AA,$1F,$F0,$64,$F8,$3C,$EF,$D6,$E3
	.dc.b	$E4,$6C,$EA,$7C,$71,$82,$3C,$4F,$EC,$9F,$2D,$E2,$EE,$34,$F1,$4F
	.dc.b	$8F,$CA,$EA,$4A,$F8,$2D,$7E,$0B,$5F,$83,$68,$F8,$36,$F9,$6F,$01
	.dc.b	$2B,$EC,$63,$F3,$AB,$E8,$7E,$05,$FD,$E2,$7F,$DC,$0D,$7F,$DC,$0D
	.dc.b	$7E,$7B,$5F,$F7,$93,$54,$A3,$0D,$F0,$3E,$5C,$6F,$6F,$F4,$97,$CB
	.dc.b	$FC,$52,$5F,$9A,$7C,$5B,$7F,$B7,$17,$8F,$C8,$93,$E8,$6B,$E8,$6B
	.dc.b	$E8,$7E,$2F,$CB,$FD,$32,$47,$78,$F8,$7C,$3E,$62,$47,$C4,$6B,$F1
	.dc.b	$1A,$FC,$47,$E8,$FC,$07,$DF,$24,$7A,$E3,$E7,$FC,$1F,$C2,$24,$7D
	.dc.b	$06,$BF,$41,$AF,$D0,$7E,$E9,$F8,$B2,$3F,$1D,$FC,$77,$CB,$48,$FF
	.dc.b	$17,$BF,$91,$E9,$FD,$77,$C8,$3E,$4B,$F3,$AD,$AF,$EE,$3C,$28,$C4
	.dc.b	$7F,$77,$8C,$4F,$EB,$9F,$F0,$14,$7F,$00,$E0,$FC,$0F,$6F,$EC,$27
	.dc.b	$3C,$1F,$E0,$79,$FC,$2F,$CA,$FF,$13,$B7,$EB,$6D,$3F,$AC,$B6,$3F
	.dc.b	$AE,$B5,$FF,$02,$B5,$FF,$05,$35,$FF,$08,$35,$FF,$49,$B5,$EF,$1F
	.dc.b	$DA,$3B,$9E,$E7,$E9,$E7,$F8,$AB,$5F,$E3,$71,$AE,$8F,$45,$F8,$F2
	.dc.b	$F6,$66,$F9,$6F,$80,$4C,$3F,$93,$69,$FA,$4D,$B1,$1F,$1E,$B8,$3A
	.dc.b	$6B,$98,$F9,$5F,$2B,$23,$4F,$CD,$E0,$C4,$73,$8B,$8F,$92,$7F,$9D
	.dc.b	$EA,$E3,$A4,$BF,$7F,$75,$7D,$F3,$B7,$E1,$CF,$29,$FB,$9F,$2B,$27
	.dc.b	$D0,$F1,$C3,$E1,$FC,$AF,$E0,$49,$3F,$C4,$31,$7C,$8E,$23,$E1,$4D
	.dc.b	$AF,$F2,$66,$BF,$91,$AF,$C5,$78,$92,$FE,$33,$F8,$DF,$13,$11,$F2
	.dc.b	$E2,$DF,$D4,$9F,$C7,$7D,$5B,$E1,$7F,$C4,$AD,$BF,$F1,$4B,$5F,$F2
	.dc.b	$8B,$5F,$F2,$B3,$5F,$F2,$E3,$5F,$F6,$FB,$53,$47,$C4,$78,$89,$1F
	.dc.b	$DA,$4D,$CF,$E9,$FE,$5A,$47,$E0,$DF,$D5,$65,$88,$FC,$44,$5C,$7F
	.dc.b	$23,$3F,$DE,$3F,$01,$FE,$F1,$5F,$DB,$9A,$FE,$69,$B3,$FE,$79,$1F
	.dc.b	$BC,$FF,$62,$ED,$FF,$06,$4E,$9B,$DD,$D4,$FF,$05,$57,$0F,$2B,$E6
	.dc.b	$FC,$6F,$CC,$26,$3F,$31,$AF,$CC,$6B,$F3,$1F,$A7,$F1,$DF,$58,$91
	.dc.b	$F1,$77,$4F,$F0,$BF,$95,$93,$FA,$77,$F8,$DF,$C6,$E2,$3F,$59,$17
	.dc.b	$1F,$C5,$F3,$2B,$F8,$A7,$CA,$CB,$FD,$83,$FC,$93,$D1,$C4,$7E,$D6
	.dc.b	$2D,$F2,$9F,$E3,$1E,$AB,$A7,$FC,$DE,$DB,$FF,$39,$B5,$FF,$4D,$B5
	.dc.b	$FF,$52,$78,$9F,$E2,$DF,$4B,$FC,$58,$48,$FC,$24,$F1,$C3,$F8,$6F
	.dc.b	$CA,$7F,$2B,$24,$FF,$8A,$B8,$97,$88,$FE,$19,$36,$BF,$C9,$95,$A9
	.dc.b	$FF,$1C,$7B,$4F,$CF,$FF,$93,$37,$C9,$FF,$45,$B6,$7F,$D6,$6D,$7F
	.dc.b	$D7,$1E,$37,$F7,$EF,$95,$FE,$7C,$48,$EB,$7F,$C9,$7A,$11,$FE,$04
	.dc.b	$F9,$77,$CA,$7F,$8B,$1B,$1F,$D4,$1A,$FF,$4F,$6B,$FE,$CB,$C3,$FC
	.dc.b	$1B,$E3,$FF,$4D,$24,$7F,$87,$27,$8E,$BF,$CD,$27,$33,$FC,$BD,$FF
	.dc.b	$27,$4E,$B0,$F2,$DF,$4A,$FE,$65,$5C,$2F,$F8,$9D,$EE,$7D,$5C,$1E
	.dc.b	$92,$3E,$97,$FB,$27,$2F,$F6,$12,$47,$96,$3F,$CE,$5E,$A3,$E5,$1F
	.dc.b	$29,$FE,$BA,$8C,$36,$33,$5F,$EE,$49,$CB,$63,$2F,$FB,$5D,$AF,$FB
	.dc.b	$9D,$AF,$FB,$91,$AB,$3F,$F1,$BC,$7F,$77,$24,$7F,$9F,$67,$BB,$FE
	.dc.b	$7C,$F8,$F7,$ED,$CF,$1F,$FD,$A3,$18,$6C,$66,$7F,$DF,$73,$96,$C6
	.dc.b	$5F,$C2,$9F,$86,$8D,$38,$0E,$1D,$B5,$FB,$3E,$4F,$FA,$7F,$82,$F5
	.dc.b	$FF,$84,$4F,$F5,$05,$74,$E0,$FE,$EF,$19,$BF,$EE,$0E,$B0,$F2,$4A
	.dc.b	$FC,$91,$FE,$9D,$AF,$0B,$5C,$F9,$3F,$00,$8F,$B9,$AF,$AF,$0F,$93
	.dc.b	$1F,$E2,$9E,$AF,$F1,$31,$27,$FD,$59,$3A,$7E,$97,$E0,$6F,$4F,$F0
	.dc.b	$CD,$8F,$C2,$35,$F9,$1B,$23,$11,$F3,$A3,$1E,$8F,$22,$47,$27,$E8
	.dc.b	$3F,$53,$E7,$1E,$8B,$DC,$CC,$A7,$31,$86,$C7,$26,$BF,$6D,$FF,$28
	.dc.b	$5E,$FD,$3F,$F4,$44,$62,$78,$47,$FA,$36,$7D,$4F,$F9,$3A,$3F,$85
	.dc.b	$C6,$37,$F8,$33,$99,$4E,$5F,$87,$97,$AA,$BE,$3E,$3B,$AF,$8C,$D9
	.dc.b	$D5,$7F,$1E,$FA,$CF,$81,$7F,$E3,$A6,$C7,$E8,$FD,$A1,$FD,$AE,$7F
	.dc.b	$85,$CE,$B8,$BF,$15,$5F,$9A,$AD,$D7,$E9,$1B,$1F,$85,$6B,$F9,$E6
	.dc.b	$BF,$A5,$6B,$CD,$F8,$1B,$F8,$9A,$6B,$9D,$83,$E1,$3C,$D7,$FC,$06
	.dc.b	$31,$87,$E4,$2D,$EE,$EB,$0E,$38,$5B,$DE,$1F,$E7,$F6,$E5,$EE,$CB
	.dc.b	$E1,$E1,$BB,$EF,$18,$CB,$E0,$4F,$98,$8E,$09,$87,$C3,$8C,$BF,$4F
	.dc.b	$CD,$F1,$1E,$AD,$9F,$BC,$3E,$37,$38,$7F,$19,$8C,$3F,$CA,$5B,$7F
	.dc.b	$A6,$6B,$FA,$56,$BF,$8E,$6B,$F8,$C6,$D7,$82,$BF,$15,$1A,$D7,$E7
	.dc.b	$F9,$DF,$1C,$F9,$FD,$6F,$07,$F9,$FE,$5F,$8B,$9C,$C6,$32,$FE,$C1
	.dc.b	$DF,$F9,$BB,$E0,$D9,$D4,$FF,$63,$E1,$FE,$D7,$87,$E5,$7C,$3E,$7F
	.dc.b	$2F,$F5,$D9,$19,$3F,$B3,$46,$2B,$ED,$4D,$E3,$FE,$BF,$36,$BF,$BC
	.dc.b	$E5,$E9,$8F,$11,$23,$FB,$C3,$FB,$1D,$3F,$1F,$EB,$FE,$1C,$FC,$7F
	.dc.b	$90,$FC,$E1,$F2,$99,$C3,$FC,$27,$18,$7F,$C1,$ED,$D3,$FA,$15,$71
	.dc.b	$AF,$DF,$36,$BF,$8B,$E1,$F9,$EF,$67,$F2,$C7,$E7,$BC,$F7,$CF,$7F
	.dc.b	$A8,$6B,$F7,$53,$A6,$9F,$E1,$6D,$BF,$6F,$5C,$3D,$DD,$67,$FE,$C2
	.dc.b	$C3,$93,$63,$EC,$FC,$B7,$79,$CD,$FF,$82,$23,$EC,$13,$0F,$F0,$F3
	.dc.b	$F1,$F8,$9F,$17,$FA,$D3,$F0,$FF,$84,$7F,$8A,$2B,$F8,$84,$65,$A7
	.dc.b	$9C,$FA,$B1,$BC,$BF,$69,$3F,$B7,$F3,$65,$6E,$5C,$8B,$7C,$BE,$A7
	.dc.b	$F7,$97,$73,$FF,$59,$D6,$B0,$FE,$45,$3F,$C2,$AB,$FB,$AF,$C6,$5F
	.dc.b	$00,$DF,$95,$C3,$6F,$FC,$0F,$1F,$CD,$EB,$FB,$EB,$68,$D2,$4F,$02
	.dc.b	$4F,$F8,$17,$85,$FA,$83,$E0,$BF,$00,$DA,$F8,$1A,$3E,$B3,$FD,$A3
	.dc.b	$1F,$0E,$F8,$DF,$02,$47,$09,$DE,$0B,$FC,$2E,$1D,$63,$F9,$17,$C9
	.dc.b	$7F,$25,$3F,$93,$77,$DC,$3F,$A4,$45,$9F,$E5,$13,$6F,$FA,$2C,$7F
	.dc.b	$87,$BE,$67,$E5,$1F,$2F,$86,$E1,$FE,$C7,$DB,$E9,$57,$F8,$B2,$7E
	.dc.b	$41,$27,$FC,$66,$D8,$D4,$60,$FD,$3F,$23,$E0,$BF,$9E,$1F,$CF,$7E
	.dc.b	$76,$CF,$F5,$F8,$B5,$FC,$FE,$6E,$5F,$D7,$3A,$FF,$A6,$FC,$8F,$F4
	.dc.b	$53,$F3,$3E,$35,$C7,$FB,$44,$5C,$3E,$54,$7E,$62,$71,$7C,$F2,$78
	.dc.b	$5F,$D3,$CF,$DD,$7E,$DE,$F8,$2C,$61,$F6,$63,$0D,$8D,$FC,$0F,$06
	.dc.b	$46,$4E,$11,$8C,$BF,$AD,$7C,$F7,$F5,$D3,$FA,$F7,$CF,$5A,$FF,$00
	.dc.b	$45,$9F,$EC,$13,$73,$F4,$7C,$F7,$F6,$53,$FB,$37,$CA,$5C,$3F,$C1
	.dc.b	$11,$6F,$FB,$44,$DA,$FE,$1F,$9B,$E2,$3D,$E7,$EB,$7F,$5E,$FF,$67
	.dc.b	$63,$EF,$8C,$7C,$47,$EA,$67,$38,$FF,$8E,$1A,$FF,$9A,$5A,$FF,$9B
	.dc.b	$9A,$FF,$AD,$1A,$9A,$6B,$8F,$FC,$13,$E2,$FF,$80,$4F,$F0,$17,$95
	.dc.b	$71,$FE,$19,$E2,$FF,$0B,$F4,$FF,$AC,$78,$DF,$E0,$A9,$FF,$25,$CE
	.dc.b	$5B,$19,$CB,$FD,$97,$F8,$D7,$F1,$9D,$FF,$18,$DA,$FF,$5A,$BF,$E3
	.dc.b	$C7,$F8,$D6,$68,$D5,$7F,$A6,$DF,$D2,$D7,$2E,$CF,$EE,$3F,$17,$FE
	.dc.b	$7B,$C3,$8D,$69,$FE,$BB,$19,$BF,$F4,$86,$48,$C4,$7F,$A5,$A3,$13
	.dc.b	$FD,$D6,$73,$3C,$A3,$04,$3E,$0C,$6A,$3C,$DD,$FF,$E2,$DB,$3E,$6F
	.dc.b	$8F,$C3,$6F,$8C,$77,$23,$82,$57,$A2,$D3,$A8,$F9,$BF,$41,$DC,$F8
	.dc.b	$7C,$76,$FE,$CD,$FE,$45,$F3,$C7,$B4,$8E,$09,$5D,$E3,$2F,$DA,$F2
	.dc.b	$3D,$87,$D0,$F8,$6B,$5D,$9F,$90,$47,$E2,$B6,$7D,$4D,$7E,$85,$FF
	.dc.b	$B9,$E7,$FB,$FE,$7F,$D3,$9F,$B7,$E5,$7F,$8B,$87,$B1,$F3,$5C,$3F
	.dc.b	$E2,$B7,$31,$F9,$46,$DF,$E0,$1A,$FD,$F6,$BF,$E2,$26,$BF,$DA,$9B
	.dc.b	$B7,$F6,$0A,$FF,$11,$5B,$E5,$BE,$F1,$E5,$F5,$2E,$3F,$A8,$C1,$1E
	.dc.b	$52,$47,$95,$B8,$F9,$5A,$F9,$5E,$FE,$1F,$E2,$CF,$85,$E7,$5A,$EC
	.dc.b	$FC,$12,$3F,$05,$F3,$FC,$FF,$E4,$CF,$CA,$7D,$05,$9F,$D3,$C5,$C3
	.dc.b	$F9,$C9,$FA,$F9,$C7,$37,$F4,$3F,$82,$57,$FA,$25,$B7,$FB,$85,$59
	.dc.b	$99,$39,$78,$46,$2F,$B8,$27,$CF,$5F,$9E,$FF,$07,$C7,$FB,$76,$F1
	.dc.b	$AF,$FB,$6A,$AC,$E4,$57,$07,$A0,$8D,$BF,$EC,$EA,$AF,$80,$E3,$5F
	.dc.b	$EE,$E8,$CC,$71,$3F,$D5,$D1,$96,$9D,$8D,$7B,$FE,$17,$F8,$E1,$A3
	.dc.b	$C7,$5E,$0F,$7F,$FD,$93,$86,$E7,$B3,$C6,$33,$7E,$F9,$F2,$9E,$F9
	.dc.b	$E1,$5C,$F0,$F3,$BF,$EF,$A9,$F6,$1F,$EF,$C7,$0F,$AC,$F4,$7C,$F3
	.dc.b	$D2,$7B,$7E,$B9,$FA,$93,$98,$C4,$23,$0E,$DA,$F4,$E0,$FF,$16,$39
	.dc.b	$6B,$9E,$0E,$72,$9C,$CF,$D8,$8C,$4F,$D8,$7F,$CD,$71,$7D,$8F,$BB
	.dc.b	$9F,$DF,$9D,$CF,$C6,$3C,$D7,$E2,$9D,$57,$0D,$3B,$36,$BD,$F5,$B9
	.dc.b	$F1,$3E,$AD,$B8,$CF,$E5,$DA,$FD,$DC,$3C,$31,$E3,$9E,$53,$F2,$AF
	.dc.b	$B3,$F2,$5B,$5E,$8C,$38,$63,$ED,$A8,$7A,$E7,$58,$FE,$9F,$D8,$7C
	.dc.b	$F3,$E3,$3F,$3F,$87,$FA,$7A,$31,$87,$F9,$8D,$AF,$EF,$AB,$8F,$C5
	.dc.b	$70,$E1,$E9,$5A,$BE,$58,$67,$06,$1E,$08,$C3,$CA,$3F,$CE,$1F,$26
	.dc.b	$F8,$CC,$B6,$33,$18,$F8,$EC,$B2,$32,$72,$8F,$07,$73,$E7,$BF,$CB
	.dc.b	$93,$FE,$10,$6B,$FD,$F1,$AF,$F7,$B6,$BF,$C3,$7C,$87,$5B,$E0,$7D
	.dc.b	$91,$86,$C6,$67,$FC,$43,$5D,$B5,$6F,$FE,$E8,$87,$B3,$FA,$4C,$BC
	.dc.b	$CD,$8C,$DF,$9E,$30,$47,$04,$3F,$85,$65,$E6,$F9,$9F,$19,$23,$C7
	.dc.b	$3F,$12,$9E,$9F,$72,$4F,$C5,$6B,$F1,$5A,$FC,$57,$F7,$77,$F1,$3E
	.dc.b	$6F,$E0,$92,$7E,$0C,$FD,$2F,$9E,$77,$FB,$DB,$3F,$E4,$26,$BF,$E2
	.dc.b	$E6,$BF,$E2,$D6,$BF,$CF,$F3,$46,$23,$FD,$B2,$D8,$CC,$F9,$B6,$7A
	.dc.b	$1F,$58,$97,$FE,$56,$BF,$0E,$FF,$5A,$39,$F8,$BE,$72,$3E,$76,$BE
	.dc.b	$76,$BE,$77,$C1,$F1,$7F,$84,$24,$77,$8F,$89,$C5,$F0,$C9,$1F,$0D
	.dc.b	$AF,$C3,$6B,$F0,$DF,$67,$CF,$7E,$40,$91,$EB,$8F,$A1,$F3,$DF,$3C
	.dc.b	$91,$F3,$DA,$FC,$F6,$BF,$3D,$F9,$07,$E7,$08,$FC,$77,$F3,$FF,$33
	.dc.b	$23,$FB,$D3,$FA,$2E,$8F,$D8,$FC,$D3,$E4,$BF,$3C,$DA,$FF,$40,$B5
	.dc.b	$FF,$36,$B5,$FE,$E3,$C8,$E2,$FF,$71,$ED,$FE,$CD,$6F,$47,$ED,$5B
	.dc.b	$4A,$E1,$F2,$BF,$C4,$61,$F3,$CD,$3F,$AF,$36,$3F,$AA,$B5,$FE,$B4
	.dc.b	$D7,$FB,$B3,$5F,$F0,$2B,$5F,$F0,$93,$5E,$F1,$FD,$83,$B5,$F3,$2E
	.dc.b	$9F,$E4,$2D,$AF,$F0,$D6,$9F,$91,$AD,$D7,$A7,$0F,$24,$FC,$1F,$97
	.dc.b	$FE,$23,$A1,$A6,$37,$B8,$FF,$A9,$27,$4D,$9D,$39,$F9,$6E,$64,$78
	.dc.b	$70,$57,$FC,$7B,$BF,$D3,$9C,$63,$F4,$EB,$D7,$C7,$2F,$D7,$C3,$FD
	.dc.b	$3F,$7F,$E4,$91,$FD,$25,$77,$6E,$31,$57,$0F,$85,$5A,$9F,$57,$B1
	.dc.b	$F0,$08,$F6,$3F,$A6,$C7,$69,$FC,$37,$BF,$F1,$89,$3F,$ED,$4C,$9F
	.dc.b	$3D,$88,$F6,$EE,$AC,$5F,$F0,$38,$5F,$0A,$F9,$9E,$1F,$D2,$23,$F3
	.dc.b	$FC,$3A,$F4,$8B,$7F,$99,$9F,$CA,$7A,$CF,$91,$FF,$18,$36,$FF,$C4
	.dc.b	$ED,$7F,$C5,$8D,$7F,$C9,$0D,$7F,$CA,$ED,$7F,$CC,$8D,$7F,$DB,$11
	.dc.b	$DB,$D5,$F0,$92,$3E,$51,$A3,$F1,$1F,$21,$F8,$E2,$3F,$8B,$D6,$BF
	.dc.b	$22,$2E,$3F,$B5,$9F,$EF,$BF,$3A,$ED,$A4,$BF,$B1,$4F,$9E,$FE,$CD
	.dc.b	$CB,$FE,$15,$BE,$15,$F7,$7B,$11,$88,$FF,$09,$36,$FF,$9C,$1C,$FE
	.dc.b	$3B,$E6,$FC,$BF,$CC,$25,$7C,$C6,$BF,$31,$AF,$CC,$7E,$9F,$B9,$F5
	.dc.b	$89,$1F,$1B,$81,$3F,$B2,$F7,$E4,$FE,$BD,$FE,$43,$F3,$98,$8F,$C6
	.dc.b	$C5,$C7,$F1,$9C,$AA,$FD,$DF,$C6,$CB,$FD,$C3,$FC,$AB,$D0,$C4,$7E
	.dc.b	$5A,$2D,$F2,$9F,$E3,$DE,$BF,$E6,$6B,$FC,$DE,$D7,$FC,$E8,$D7,$FD
	.dc.b	$20,$D7,$FC,$89,$C4,$F4,$FF,$C6,$24,$8F,$C2,$CF,$3B,$FE,$65,$F1
	.dc.b	$BF,$C4,$89,$3F,$E3,$AA,$AF,$EB,$22,$CF,$F1,$DB,$ED,$3F,$B9,$F6
	.dc.b	$9F,$3B,$FC,$AD,$BE,$4F,$FA,$21,$B3,$FE,$8D,$6B,$FE,$20,$EE,$7C
	.dc.b	$F7,$F2,$42,$47,$5B,$FE,$53,$D1,$FE,$09,$F9,$A7,$CD,$7F,$8C,$5B
	.dc.b	$5F,$E9,$9C,$11,$88,$FF,$5B,$46,$27,$FC,$42,$FF,$3C,$8F,$E5,$DE
	.dc.b	$1F,$F3,$43,$FC,$A3,$5C,$75,$FE,$B1,$19,$9D,$9F,$C6,$6B,$A3,$83
	.dc.b	$F8,$9C,$6D,$FF,$76,$C6,$9B,$5F,$EA,$17,$D0,$D7,$CE,$D7,$FD,$B7
	.dc.b	$1E,$CE,$0F,$4F,$1F,$F4,$A2,$4F,$8E,$B9,$61,$FE,$23,$F5,$5F,$0A
	.dc.b	$F9,$6F,$F5,$DC,$61,$B1,$9B,$FF,$61,$4E,$5B,$19,$7F,$DC,$11,$96
	.dc.b	$C6,$5F,$F7,$8C,$65,$FF,$01,$E5,$E2,$7F,$AE,$4F,$F6,$23,$FD,$1D
	.dc.b	$3C,$EF,$FC,$75,$F2,$CF,$9E,$78,$DF,$ED,$28,$C3,$63,$38,$7F,$B6
	.dc.b	$67,$2D,$8C,$BF,$EF,$78,$CB,$63,$28,$91,$A7,$D3,$D1,$FD,$E0,$8F
	.dc.b	$F5,$98,$E7,$7F,$E1,$D9,$CC,$FF,$54,$7F,$D5,$53,$AE,$1E,$B7,$FE
	.dc.b	$A5,$AF,$17,$8F,$EB,$23,$E9,$6B,$A3,$07,$FB,$DF,$CD,$FF,$17,$75
	.dc.b	$7F,$83,$09,$1E,$5A,$ED,$97,$E0,$FE,$75,$E9,$FE,$09,$B5,$F8,$06
	.dc.b	$BF,$79,$AF,$DC,$7F,$C2,$6E,$7D,$0E,$44,$7E,$51,$07,$CF,$3D,$1F
	.dc.b	$F5,$9C,$64,$CB,$87,$84,$6F,$DB,$D4,$69,$B1,$FE,$55,$E2,$BE,$9F
	.dc.b	$FA,$2E,$31,$5D,$5A,$FA,$63,$FD,$0C,$DE,$0F,$81,$18,$8F,$C1,$53
	.dc.b	$2F,$F3,$15,$7C,$5C,$75,$5F,$EA,$49,$F6,$4E,$49,$1F,$EB,$CF,$98
	.dc.b	$FF,$68,$C7,$F8,$7F,$81,$18,$96,$DF,$A0,$F6,$87,$F6,$4B,$FF,$6D
	.dc.b	$56,$E7,$53,$E2,$E3,$DB,$4E,$35,$FA,$36,$DF,$E8,$5A,$FD,$69,$E3
	.dc.b	$3B,$9E,$47,$BB,$0F,$3F,$C2,$F8,$EF,$FC,$ED,$8A,$31,$87,$A3,$DD
	.dc.b	$D5,$F0,$C7,$E0,$5E,$DB,$7C,$31,$FF,$5E,$C6,$0F,$D6,$38,$C7,$D1
	.dc.b	$87,$B4,$93,$ED,$6C,$6A,$30,$7E,$96,$DE,$5F,$91,$B1,$FE,$51,$D9
	.dc.b	$F3,$CF,$8B,$D5,$FB,$E7,$31,$F8,$06,$D7,$F2,$6A,$E0,$D9,$D4,$63
	.dc.b	$0F,$E4,$ED,$7F,$92,$B5,$FC,$7B,$5F,$9B,$E6,$79,$CF,$A1,$DB,$FD
	.dc.b	$73,$BB,$18,$7F,$3F,$7F,$27,$19,$8F,$E9,$71,$8E,$FF,$E8,$11,$F0
	.dc.b	$FC,$3F,$84,$74,$87,$CC,$E2,$7E,$BC,$DC,$7F,$AB,$CD,$E4,$FF,$77
	.dc.b	$71,$FE,$6A,$B1,$7F,$DD,$6A,$DF,$FB,$07,$0E,$B3,$FD,$7A,$7C,$84
	.dc.b	$7E,$B7,$AB,$F5,$4F,$C7,$79,$0F,$9E,$7C,$A6,$6B,$FC,$21,$18,$7F
	.dc.b	$C1,$8D,$C7,$F8,$85,$71,$AF,$DB,$B6,$BF,$09,$EC,$7E,$0C,$FC,$E7
	.dc.b	$AA,$F9,$E7,$B9,$9A,$FC,$CD,$5C,$7F,$6F,$1F,$D1,$EF,$85,$67,$83
	.dc.b	$FD,$85,$87,$2A,$FF,$31,$36,$FF,$C0,$3F,$2D,$E0,$8C,$57,$FA,$5F
	.dc.b	$74,$FD,$92,$5F,$F8,$19,$FD,$A7,$4B,$E3,$3F,$38,$7E,$1B,$E5,$BF
	.dc.b	$AD,$57,$EC,$23,$2D,$3F,$C3,$11,$AC,$BF,$D8,$B5,$CE,$7F,$6F,$E7
	.dc.b	$7E,$9C,$E9,$F2,$D6,$32,$4A,$FF,$0E,$DA,$75,$1F,$18,$FE,$D7,$F1
	.dc.b	$9F,$C2,$57,$84,$8C,$2F,$94,$B1,$87,$F9,$7C,$F0,$6C,$EF,$0F,$F0
	.dc.b	$24,$60,$8E,$09,$3F,$D4,$F8,$5F,$3A,$F8,$1F,$C0,$B6,$BF,$D3,$BF
	.dc.b	$3D,$FE,$CE,$7F,$D3,$51,$9B,$F6,$5F,$2B,$E2,$7F,$25,$E2,$61,$C6
	.dc.b	$B7,$7C,$67,$F5,$3B,$BE,$4B,$F8,$A1,$FC,$57,$BA,$E7,$FC,$A2,$2E
	.dc.b	$1F,$5A,$6D,$FF,$5A,$D5,$F3,$1F,$28,$F9,$7C,$17,$1F,$F6,$2C,$FD
	.dc.b	$57,$FC,$55,$1F,$28,$8E,$09,$87,$F8,$AE,$32,$FF,$5B,$F8,$1F,$E5
	.dc.b	$47,$F2,$BF,$9E,$B5,$FC,$FA,$2C,$FE,$32,$6E,$3F,$03,$E4,$7F,$99
	.dc.b	$1F,$97,$F1,$AE,$1F,$D1,$A2,$DF,$E3,$26,$D7,$C8,$F0,$7F,$9D,$1F
	.dc.b	$B8,$F9,$B7,$C0,$E3,$1F,$F6,$84,$61,$B1,$BF,$81,$E0,$C8,$C9,$C2
	.dc.b	$31,$8F,$E9,$3E,$7B,$FA,$41,$FD,$27,$E7,$AD,$7F,$5C,$8B,$3F,$AA
	.dc.b	$9B,$97,$DE,$F9,$EF,$EA,$07,$F5,$1F,$94,$B8,$7F,$64,$8B,$7F,$B2
	.dc.b	$9B,$4A,$FE,$1D,$9B,$E1,$FE,$01,$FA,$BF,$9E,$7C,$6E,$2B,$FC,$B5
	.dc.b	$7F,$E1,$E9,$CC,$7F,$8A,$9B,$8F,$F8,$C5,$AF,$F8,$DD,$AF,$F8,$E9
	.dc.b	$AF,$F9,$95,$A9,$A6,$BC,$DF,$F0,$EF,$89,$FD,$BC,$FE,$E1,$E5,$5B
	.dc.b	$FE,$23,$E9,$FF,$58,$F1,$3F,$BF,$63,$FE,$9E,$6B,$FE,$73,$F3,$D1
	.dc.b	$88,$62,$F9,$DF,$E3,$5A,$7F,$54,$6E,$3F,$E7,$B9,$D7,$A7,$FE,$14
	.dc.b	$9C,$DF,$27,$F4,$8F,$F1,$79,$E9,$D9,$FD,$C3,$E2,$FE,$96,$1F,$E3
	.dc.b	$9B,$DE,$3F,$57,$06,$6B,$DC,$7F,$B4,$1A,$FF,$AE,$9A,$FB,$5A,$F2
	.dc.b	$6C,$73,$BF,$EF,$B1,$83,$78,$FC,$1E,$FF,$61,$E7,$F8,$FB,$5F,$ED
	.dc.b	$5A,$47,$22,$47,$FB,$9E,$32,$D8,$CE,$1E,$96,$BF,$E7,$CB,$FA,$1F
	.dc.b	$3D,$F2,$CF,$8D,$C9,$6F,$C0,$FA,$08,$F7,$6D,$7C,$56,$BF,$15,$FF
	.dc.b	$28,$69,$FC,$F6,$3E,$A3,$74,$FA,$1A,$7C,$FF,$86,$FB,$9E,$57,$D7
	.dc.b	$7E,$39,$31,$F7,$37,$0F,$A8,$D7,$EA,$3E,$0F,$2F,$F0,$C7,$F8,$C0
	.dc.b	$FF,$23,$C7,$89,$C7,$B1,$FE,$30,$73,$1F,$EF,$66,$BF,$99,$6B,$F8
	.dc.b	$66,$E1,$FD,$86,$79,$12,$FF,$D7,$99,$3E,$5B,$F3,$07,$8F,$D6,$B3
	.dc.b	$FB,$E3,$B3,$E3,$23,$FE,$AD,$BF,$ED,$D1,$8C,$7F,$D6,$91,$89,$D6
	.dc.b	$2A,$D5,$7D,$FC,$3F,$CB,$5A,$7E,$B7,$C3,$F8,$9E,$7F,$EA,$DB,$8F
	.dc.b	$28,$F8,$84,$7D,$7F,$3D,$FB,$13,$F6,$5F,$3D,$63,$F5,$91,$71,$FB
	.dc.b	$F3,$70,$FF,$27,$FC,$12,$BF,$D1,$AD,$BF,$F5,$62,$AC,$CC,$9C,$BC
	.dc.b	$23,$17,$FE,$15,$3F,$DA,$FF,$3D,$FF,$00,$7E,$7B,$FC,$29,$1F,$ED
	.dc.b	$AE,$BF,$DF,$46,$23,$61,$C2,$8D,$EF,$FE,$DD,$AE,$8F,$FB,$8E,$33
	.dc.b	$3F,$EE,$59,$CB,$DF,$1A,$F6,$F8,$0E,$B5,$FC,$3D,$EF,$0A,$3F,$EB
	.dc.b	$3B,$ED,$78,$AD,$D7,$4A,$F3,$4D,$79,$1E,$BA,$E2,$D3,$CC,$53,$FD
	.dc.b	$CB,$1C,$67,$A5,$FB,$27,$E2,$9F,$EF,$88,$FF,$7C,$BE,$66,$BE,$03
	.dc.b	$D2,$7A,$BE,$7B,$FC,$49,$EE,$7F,$C0,$06,$C6,$8D,$7A,$49,$D3,$63
	.dc.b	$42,$5C,$1F,$EE,$E8,$CC,$63,$83,$B4,$E6,$7B,$4E,$5D,$C7,$1D,$3F
	.dc.b	$05,$3F,$E7,$19,$F9,$DA,$7A,$A3,$11,$F4,$CF,$65,$71,$9D,$36,$3E
	.dc.b	$07,$B1,$F0,$AF,$AB,$6B,$78,$76,$AE,$6D,$7D,$D3,$B7,$99,$9C,$3D
	.dc.b	$57,$E8,$C3,$C1,$3C,$0D,$57,$17,$E4,$46,$AB,$8E,$1E,$DA,$D4,$7B
	.dc.b	$AF,$A6,$1E,$5E,$CF,$C9,$F1,$7D,$2C,$BF,$BB,$87,$11,$A8,$FA,$ED
	.dc.b	$7E,$F3,$5F,$EA,$07,$13,$65,$36,$53,$F1,$7C,$CF,$9D,$7A,$5E,$38
	.dc.b	$C3,$63,$33,$FC,$9D,$AF,$F2,$66,$BD,$1A,$FF,$85,$5A,$FF,$80,$9B
	.dc.b	$2F,$C3,$1F,$01,$E1,$6D,$7F,$35,$6B,$FC,$D1,$AF,$E7,$FC,$27,$BF
	.dc.b	$FA,$A6,$C7,$CD,$6B,$FA,$06,$BC,$6B,$81,$1E,$0D,$78,$35,$E0,$75
	.dc.b	$F8,$1F,$01,$1C,$1F,$D3,$7F,$2E,$FF,$73,$B7,$E0,$7E,$74,$62,$3F
	.dc.b	$A7,$B5,$FA,$AD,$7F,$C7,$CD,$7F,$C5,$6D,$C8,$E7,$5F,$F2,$07,$D1
	.dc.b	$7E,$2B,$BC,$D2,$5F,$97,$E1,$7C,$64,$70,$43,$CA,$E5,$A7,$F7,$2F
	.dc.b	$A2,$EC,$47,$07,$C1,$F9,$6F,$79,$1C,$10,$F8,$2E,$5A,$7F,$16,$E4
	.dc.b	$7D,$8B,$E2,$BF,$51,$18,$6C,$66,$78,$35,$FE,$F7,$D0,$F8,$EF,$CF
	.dc.b	$36,$3F,$C0,$4D,$77,$1A,$23,$A6,$BF,$B2,$D3,$E8,$13,$83,$E8,$1D
	.dc.b	$B0,$93,$D9,$AF,$91,$AF,$ED,$2F,$D0,$D8,$F3,$FC,$CC,$8F,$3B,$5F
	.dc.b	$6F,$9D,$EC,$23,$EC,$7D,$FE,$B4,$8F,$7B,$5F,$8B,$F2,$5F,$10,$8F
	.dc.b	$C4,$7F,$AA,$FC,$F3,$E8,$29,$27,$F8,$D4,$65,$B1,$97,$FA,$7C,$65
	.dc.b	$B1,$F6,$BE,$3F,$F2,$B1,$F7,$DA,$FF,$5B,$E0,$46,$23,$FB,$1C,$63
	.dc.b	$93,$F3,$DF,$3F,$27,$FC,$8A,$D7,$EB,$35,$FF,$1C,$35,$FF,$51,$70
	.dc.b	$3E,$83,$F5,$04,$8F,$F2,$7B,$5F,$C0,$35,$FF,$21,$3F,$DE,$BF,$7A
	.dc.b	$79,$F4,$93,$FC,$FE,$32,$D8,$CB,$FD,$D2,$32,$D8,$FE,$8D,$F1,$CF
	.dc.b	$D4,$1E,$B6,$1B,$5F,$DF,$27,$53,$93,$F8,$17,$B9,$27,$FC,$D2,$D7
	.dc.b	$F3,$4D,$7F,$CB,$8D,$7E,$2F,$67,$F0,$B2,$3F,$E7,$26,$BF,$A3,$6B
	.dc.b	$FE,$65,$7F,$C3,$BE,$E3,$E6,$69,$27,$FB,$14,$65,$B1,$97,$FC,$2F
	.dc.b	$19,$6C,$7F,$8A,$3E,$EB,$E0,$C7,$F8,$71,$AF,$F8,$8D,$AF,$F8,$A9
	.dc.b	$AF,$FB,$77,$E9,$3F,$93,$11,$FF,$46,$B5,$FD,$F3,$5F,$F4,$03,$F5
	.dc.b	$3F,$54,$7C,$15,$24,$FD,$38,$CB,$63,$2F,$F8,$DF,$44,$EA,$30,$D7
	.dc.b	$FC,$79,$8B,$E2,$7E,$59,$23,$E4,$B5,$F8,$ED,$78,$BF,$1F,$DA,$91
	.dc.b	$FE,$11,$CB,$FD,$10,$D8,$FC,$7F,$EA,$DF,$E9,$F8,$FC,$63,$5F,$43
	.dc.b	$5F,$F2,$AE,$28,$C4,$7F,$9C,$23,$1E,$E7,$42,$47,$47,$FA,$B7,$54
	.dc.b	$8F,$F5,$EB,$5F,$E5,$CD,$7F,$D6,$0D,$7F,$CC,$BF,$85,$7F,$A7,$5F
	.dc.b	$F1,$63,$5F,$F2,$FB,$5F,$F3,$33,$5F,$F4,$2F,$B0,$FB,$EF,$87,$1F
	.dc.b	$E6,$B8,$C6,$5F,$E3,$76,$C7,$F9,$A9,$AF,$FB,$23,$3F,$F1,$73,$FE
	.dc.b	$90,$F3,$11,$88,$46,$3D,$4F,$F1,$84,$E6,$31,$3F,$ED,$58,$C1,$27
	.dc.b	$7F,$01,$27,$4D,$7F,$AA,$79,$7F,$E0,$12,$3F,$EE,$87,$BF,$A1,$23
	.dc.b	$B3,$5F,$EB,$9F,$4B,$FE,$09,$23,$FE,$F0,$6B,$A3,$04,$30,$43,$FD
	.dc.b	$CC,$F0,$7F,$CA,$38,$7F,$A0,$9B,$EC,$7C,$22,$3F,$E6,$0F,$78,$7F
	.dc.b	$99,$0A,$E5,$C1,$FE,$6B,$F7,$9F,$4E,$F8,$6C,$36,$BF,$D4,$33,$A9
	.dc.b	$C9,$23,$33,$B7,$FD,$A8,$D7,$FD,$C0,$D7,$FD,$9B,$3F,$D5,$BE,$7F
	.dc.b	$C5,$E4,$9A,$49,$FF,$58,$FC,$A3,$C0,$6B,$FE,$83,$8C,$36,$33,$3C
	.dc.b	$1A,$FF,$AF,$E7,$6E,$67,$81,$1F,$E8,$9F,$40,$F9,$3F,$E9,$6D,$8C
	.dc.b	$C6,$08,$E0,$87,$F1,$78,$E0,$F8,$6F,$FB,$E3,$FE,$D6,$BF,$67,$4F
	.dc.b	$A4,$91,$E9,$6B,$C5,$AF,$F0,$AF,$BE,$78,$5F,$D8,$1B,$1F,$D7,$5A
	.dc.b	$FF,$29,$6B,$FE,$E1,$F3,$BF,$DD,$24,$7F,$C3,$27,$FB,$BB,$E8,$5F
	.dc.b	$4F,$49,$1F,$EB,$38,$CB,$7D,$5F,$DC,$47,$F9,$47,$AB,$04,$9F,$F7
	.dc.b	$5C,$3E,$89,$C7,$CD,$B3,$A6,$BF,$EF,$79,$E1,$3F,$EC,$3E,$57,$A1
	.dc.b	$F9,$EA,$FF,$7E,$46,$39,$F5,$39,$91,$27,$4F,$FB,$FB,$F6,$07,$CD
	.dc.b	$EE,$37,$18,$BF,$F6,$93,$63,$84,$EF,$C5,$FC,$B7,$B1,$79,$4D,$26
	.dc.b	$1F,$9C,$FD,$D3,$FC,$2A,$DF,$3F,$C1,$CE,$9C,$E5,$F9,$96,$BF,$89
	.dc.b	$6B,$EC,$8E,$13,$CA,$73,$29,$CD,$7E,$17,$E4,$BD,$1C,$C6,$92,$34
	.dc.b	$4A,$FC,$54,$65,$F2,$3D,$C9,$1E,$D6,$C6,$A3,$04,$7E,$99,$FC,$13
	.dc.b	$B7,$F0,$91,$FC,$41,$B1,$A3,$F0,$EE,$1E,$3C,$3F,$E6,$0E,$3F,$B0
	.dc.b	$FF,$1C,$F7,$EF,$79,$A4,$C3,$F8,$CB,$6F,$F8,$CB,$5F,$2F,$C0,$7F
	.dc.b	$2C,$F7,$FF,$08,$DA,$FE,$61,$18,$7F,$D5,$07,$F5,$3F,$8B,$FC,$57
	.dc.b	$91,$FD,$71,$BE,$29,$A7,$1F,$12,$57,$F3,$B6,$E3,$F0,$71,$FE,$7B
	.dc.b	$18,$8F,$2C,$FF,$29,$38,$11,$E0,$E5,$B1,$C5,$FE,$E1,$6F,$9B,$F2
	.dc.b	$92,$BF,$B6,$3E,$DD,$A4,$7F,$4F,$6D,$7F,$7B,$F1,$3D,$A4,$9F,$EE
	.dc.b	$8F,$C0,$F9,$99,$1F,$DD,$DB,$3F,$E0,$36,$BF,$D8,$1A,$FF,$84,$3F
	.dc.b	$00,$7C,$8F,$ED,$7C,$D4,$62,$FF,$C1,$CD,$AF,$F0,$A3,$5F,$C5,$7E
	.dc.b	$19,$FE,$6D,$9F,$C6,$47,$F0,$48,$C3,$86,$BC,$3D,$8F,$BD,$E2,$CA
	.dc.b	$FE,$E0,$D9,$F9,$ED,$7E,$AB,$5F,$F1,$F7,$A9,$F3,$E7,$FC,$8F,$B7
	.dc.b	$F0,$98,$C4,$FF,$92,$23,$04,$9D,$92,$3F,$49,$CE,$FD,$6B,$FC,$DB
	.dc.b	$3F,$DE,$1B,$3A,$6B,$FA,$86,$BF,$D4,$FD,$EF,$E4,$0F,$F0,$76,$F6
	.dc.b	$FF,$AF,$EB,$FB,$84,$60,$8E,$09,$3F,$E7,$7F,$D6,$7F,$C2,$71,$FE
	.dc.b	$AF,$6C,$66,$3F,$94,$46,$23,$FC,$CA,$D7,$FC,$C8,$D5,$E1,$A3,$0D
	.dc.b	$9F,$D9,$78,$51,$FF,$12,$36,$BF,$46,$D7,$F5,$2D,$7F,$A2,$FC,$13
	.dc.b	$F0,$AF,$81,$1F,$EE,$D6,$BF,$EE,$C6,$BC,$1A,$FF,$33,$E4,$F9,$24
	.dc.b	$BF,$F0,$C3,$63,$FC,$3A,$D7,$FC,$2C,$FF,$0B,$F8,$69,$3F,$E3,$D6
	.dc.b	$CF,$EF,$1A,$FF,$02,$6B,$FD,$7B,$AF,$FC,$87,$F0,$DF,$C5,$49,$5F
	.dc.b	$E4,$76,$CF,$F0,$66,$BF,$C3,$1F,$EB,$9F,$82,$FF,$5A,$9F,$F7,$43
	.dc.b	$67,$FD,$68,$D7,$FD,$3A,$D7,$F9,$EF,$13,$EA,$FF,$1A,$DA,$FD,$8C
	.dc.b	$FF,$11,$9C,$92,$33,$3B,$9F,$DB,$B5,$FE,$88,$D7,$E0,$43,$E2,$3F
	.dc.b	$9A,$11,$FF,$33,$36,$3F,$92,$B5,$FE,$56,$FF,$3B,$FA,$C9,$3F,$E6
	.dc.b	$D6,$CF,$F2,$C6,$BF,$CC,$9A,$FE,$0F,$C4,$FF,$05,$6F,$F8,$B6,$EF
	.dc.b	$FD,$4E,$FF,$88,$CE,$5F,$87,$F3,$DE,$96,$E3,$FB,$48,$C6,$1F,$C6
	.dc.b	$1A,$FF,$65,$6B,$FD,$B1,$AF,$F8,$81,$A9,$CF,$88,$F9,$17,$BD,$FD
	.dc.b	$61,$B8,$7F,$26,$9D,$4E,$4D,$71,$C9,$D5,$7F,$2F,$C3,$FD,$78,$D7
	.dc.b	$FD,$76,$BE,$2F,$B1,$27,$8B,$E1,$D6,$47,$46,$D7,$F8,$0F,$1F,$47
	.dc.b	$C6,$F9,$C9,$5E,$26,$CE,$F0,$46,$23,$FC,$23,$5C,$AB,$FD,$55,$F5
	.dc.b	$1F,$DE,$C8,$E0,$F6,$FC,$4F,$C5,$23,$82,$57,$78,$CB,$63,$EF,$FE
	.dc.b	$25,$FC,$C2,$1C,$08,$C3,$7E,$2B,$F1,$31,$BC,$7E,$B7,$BD,$F5,$49
	.dc.b	$87,$D0,$6E,$1E,$F6,$BE,$ED,$11,$FE,$17,$FB,$29,$5F,$EC,$96,$D7
	.dc.b	$CD,$FC,$DF,$F9,$86,$06,$92,$BF,$C3,$AD,$7F,$0C,$D9,$FF,$08,$FD
	.dc.b	$07,$F8,$94,$93,$FE,$D7,$78,$F9,$B2,$38,$B6,$7C,$4D,$7E,$BD,$7F
	.dc.b	$90,$5F,$F3,$DD,$F9,$67,$FA,$BF,$AF,$2F,$FD,$8C,$DA,$FF,$65,$FD
	.dc.b	$0F,$F5,$C2,$4F,$FB,$2D,$B1,$A8,$C1,$1F,$F6,$9F,$AD,$FD,$90,$91
	.dc.b	$FE,$D3,$7E,$1F,$85,$23,$E1,$B6,$7F,$31,$F3,$8F,$9D,$FF,$22,$B6
	.dc.b	$BF,$5C,$D9,$DC,$E8,$93,$A6,$C7,$F9,$17,$E0,$5F,$25,$FE,$56,$6D
	.dc.b	$7F,$87,$BC,$54,$62,$B8,$56,$C9,$5B,$9D,$4F,$F9,$89,$AF,$F8,$A3
	.dc.b	$E7,$BF,$D5,$7F,$3C,$CB,$FC,$27,$78,$75,$49,$D3,$FE,$78,$9D,$3F
	.dc.b	$E7,$DA,$AF,$FA,$DB,$CE,$4E,$23,$FD,$73,$36,$73,$1D,$6F,$BD,$F6
	.dc.b	$8E,$57,$C5,$5F,$18,$C3,$63,$91,$E3,$29,$C8,$A7,$88,$B7,$E2,$AF
	.dc.b	$04,$65,$B1,$9B,$F8,$73,$FD,$82,$BF,$B0,$BC,$CC,$1F,$E5,$4F,$97
	.dc.b	$FF,$43,$7A,$BF,$D9,$9B,$19,$F6,$93,$96,$C7,$1D,$BF,$86,$B7,$B3
	.dc.b	$94,$62,$3F,$D5,$97,$F9,$39,$C9,$FE,$E6,$F9,$FF,$95,$0F,$8C,$FF
	.dc.b	$5B,$C6,$1B,$9F,$FA,$EA,$79,$5A,$B8,$E1,$E2,$78,$47,$5A,$E1,$7C
	.dc.b	$A3,$FD,$EB,$7C,$AB,$8E,$1F,$84,$8F,$14,$DA,$FF,$64,$EB,$A8,$FA
	.dc.b	$F3,$C7,$0E,$B1,$DC,$EB,$5D,$D4,$ED,$F0,$E3,$E7,$AE,$31,$EB,$6E
	.dc.b	$3D,$CF,$CD,$F0,$FE,$8A,$7F,$C9,$3B,$FF,$8C,$5F,$D0,$18,$11,$B7
	.dc.b	$FA,$0C,$6F,$83,$FD,$C7,$A7,$E7,$2B,$4D,$AE,$67,$03,$99,$87,$53
	.dc.b	$CA,$BF,$C8,$4F,$F6,$09,$F1,$57,$F0,$2F,$07,$51,$E4,$9D,$3D,$7A
	.dc.b	$1F,$3C,$F9,$FD,$E1,$C6,$B9,$DF,$0C,$BD,$0D,$CB,$F9,$AE,$1C,$6B
	.dc.b	$E1,$0F,$9E,$FF,$08,$3F,$E4,$26,$BF,$EB,$46,$BF,$F0,$05,$A8,$8B
	.dc.b	$E5,$7D,$71,$D5,$B1,$98,$F0,$BD,$B0,$FE,$13,$1E,$3F,$9E,$FD,$9C
	.dc.b	$9A,$48,$F3,$35,$F3,$35,$F3,$7C,$A7,$80,$8F,$82,$3E,$4F,$CE,$7F
	.dc.b	$88,$63,$F9,$A3,$5F,$F7,$F3,$5F,$71,$F8,$87,$0F,$F9,$F5,$AE,$7C
	.dc.b	$AF,$F3,$5E,$A9,$CB,$63,$3A,$3E,$9F,$ED,$B6,$FF,$0A,$D7,$F1,$D5
	.dc.b	$C5,$FC,$4D,$BC,$BE,$ED,$8F,$5C,$60,$8E,$08,$7B,$9F,$1F,$CC,$F8
	.dc.b	$89,$1D,$A3,$F2,$BA,$BE,$79,$EF,$FF,$A2,$23,$0D,$8C,$D7,$F9,$B1
	.dc.b	$AF,$CA,$3F,$C5,$EE,$1F,$D7,$B5,$FF,$60,$B5,$DC,$7F,$A5,$BD,$A7
	.dc.b	$D9,$BD,$1F,$D4,$C6,$1B,$19,$AF,$E0,$18,$79,$9F,$93,$D1,$F2,$08
	.dc.b	$FE,$09,$AF,$E0,$9A,$FE,$0B,$5F,$B9,$1F,$6B,$E4,$FE,$C9,$27,$A9
	.dc.b	$FC,$3F,$A7,$F8,$EF,$8B,$FE,$AE,$6C,$7F,$9E,$FB,$93,$96,$C6,$63
	.dc.b	$04,$70,$7F,$82,$F8,$9F,$65,$F8,$A2,$47,$86,$B9,$BF,$C2,$D7,$B9
	.dc.b	$FC,$43,$0F,$E6,$2D,$8C,$CF,$1F,$77,$5F,$9C,$BE,$5F,$D4,$D9,$FF
	.dc.b	$4A,$35,$FF,$42,$35,$F9,$19,$3E,$CF,$F3,$E4,$8F,$37,$B7,$F8,$EF
	.dc.b	$8F,$FD,$5F,$11,$A4,$EC,$FC,$83,$6B,$FD,$72,$D7,$F1,$DC,$DF,$86
	.dc.b	$F6,$FF,$06,$48,$FC,$1C,$FE,$FE,$FF,$9F,$64,$FA,$6F,$D9,$92,$BF
	.dc.b	$5F,$F2,$F5,$FD,$9C,$ED,$FD,$BC,$FF,$0C,$9F,$DD,$FC,$4C,$9F,$DD
	.dc.b	$BF,$C4,$3E,$1B,$11,$CA,$2E,$1F,$D3,$FC,$05,$7E,$73,$E8,$9E,$B5
	.dc.b	$26,$1F,$CE,$BE,$8F,$FC,$9E,$FF,$5D,$DD,$5C,$23,$0D,$7F,$AC,$46
	.dc.b	$AB,$F9,$BD,$72,$9F,$6F,$C3,$7F,$14,$23,$CA,$7F,$A8,$FE,$4B,$FE
	.dc.b	$5A,$8F,$EB,$6D,$7F,$8B,$B7,$3F,$67,$C1,$70,$F3,$65,$70,$7F,$99
	.dc.b	$C7,$4F,$4E,$4F,$47,$F9,$BF,$CB,$E2,$3A,$C5,$BF,$91,$3F,$D4,$3E
	.dc.b	$4B,$DD,$5F,$C3,$DA,$FF,$3A,$6B,$FD,$65,$A9,$A4,$CB,$FC,$8A,$F0
	.dc.b	$D5,$F3,$DF,$CD,$70,$FF,$30,$C6,$3C,$CF,$E2,$73,$9E,$6F,$06,$1F
	.dc.b	$23,$CE,$99,$7C,$87,$FA,$9F,$76,$23,$E4,$C5,$CB,$F0,$93,$F4,$3C
	.dc.b	$E9,$5F,$41,$FE,$43,$F4,$8F,$86,$CC,$FD,$29,$B7,$E6,$AD,$4F,$F8
	.dc.b	$0B,$67,$6F,$B8,$91,$FC,$DC,$F8,$CE,$1C,$47,$07,$47,$C9,$8D,$3F
	.dc.b	$C9,$F9,$1F,$41,$F8,$82,$47,$86,$7F,$17,$E9,$C8,$FC,$5B,$FC,$F3
	.dc.b	$E4,$7F,$C1,$11,$F8,$C8,$D3,$4F,$D0,$46,$B2,$FF,$1C,$61,$FE,$0D
	.dc.b	$E0,$FF,$0B,$79,$A8,$C5,$FF,$90,$9B,$1F,$DD,$FE,$1B,$F3,$BF,$0F
	.dc.b	$2B,$F3,$AF,$F8,$22,$3F,$C6,$FE,$F3,$E8,$BE,$15,$7F,$5D,$8C,$6B
	.dc.b	$F1,$1B,$1F,$2D,$AF,$F6,$3F,$37,$F2,$1D,$BF,$CC,$F8,$7F,$51,$AF
	.dc.b	$FB,$B1,$B5,$FE,$8E,$C7,$BE,$DE,$EF,$7B,$F6,$64,$9F,$EA,$15,$C3
	.dc.b	$6F,$81,$39,$9F,$DA,$BD,$F1,$FF,$11,$FD,$17,$F8,$28,$93,$FE,$24
	.dc.b	$7F,$C3,$7C,$1F,$EE,$78,$FD,$FF,$BF,$FC,$08,$E2,$FF,$85,$3A,$FF
	.dc.b	$C2,$64,$7F,$C5,$AF,$2F,$80,$FE,$12,$48,$E0,$7F,$0A,$70,$FF,$0A
	.dc.b	$7F,$C6,$F1,$8D,$3E,$EC,$5D,$BF,$C6,$53,$FE,$2A,$7F,$C5,$A4,$AF
	.dc.b	$EC,$CF,$AB,$E7,$7F,$8C,$92,$3F,$C5,$F1,$FE,$72,$F8,$3F,$EA,$1F
	.dc.b	$33,$F9,$C6,$CF,$E8,$DA,$FF,$86,$DA,$FF,$91,$9A,$FF,$BA,$7C,$D4
	.dc.b	$F3,$6C,$73,$6B,$FE,$25,$E0,$FC,$B7,$C0,$FF,$2E,$24,$7F,$A6,$6F
	.dc.b	$C3,$A7,$F1,$0F,$A2,$7C,$D7,$F8,$95,$B5,$FB,$9F,$15,$18,$8F,$EC
	.dc.b	$AD,$7F,$CB,$B1,$A9,$E4,$48,$E5,$3B,$7F,$57,$C2,$FA,$DF,$E8,$64
	.dc.b	$8F,$EA,$53,$DF,$0F,$D9,$CE,$67,$FA,$2B,$FE,$38,$9D,$6D,$FE,$8B
	.dc.b	$C5,$F3,$DF,$E0,$9B,$C3,$54,$6B,$C4,$FF,$22,$4E,$63,$13,$F0,$63
	.dc.b	$04,$9F,$CD,$FC,$AF,$F5,$A2,$4F,$F6,$99,$FF,$0D,$65,$FC,$F3,$F7
	.dc.b	$E7,$D5,$7F,$94,$5B,$5F,$35,$AF,$F8,$41,$AF,$E0,$FF,$74,$FE,$42
	.dc.b	$FE,$25,$AF,$E2,$5A,$EF,$C5,$FE,$56,$BE,$97,$FC,$DF,$D9,$FC,$B9
	.dc.b	$B1,$FC,$D5,$AF,$F9,$B5,$AF,$FA,$61,$7C,$8B,$93,$93,$67,$93,$5E
	.dc.b	$4D,$7E,$A7,$EE,$CF,$A1,$FF,$5B,$C6,$1A,$FD,$48,$D7,$17,$E4,$5B
	.dc.b	$1F,$E3,$4F,$4D,$F2,$7C,$39,$11,$8B,$FC,$C3,$6B,$85,$76,$EA,$FE
	.dc.b	$75,$D2,$BF,$F0,$8B,$6B,$FC,$3B,$E9,$7F,$9B,$27,$27,$F8,$95,$C4
	.dc.b	$7F,$91,$9A,$FE,$89,$AF,$E4,$DA,$FE,$41,$B8,$9A,$4F,$84,$F3,$75
	.dc.b	$BE,$C3,$C9,$CF,$FD,$51,$B8,$FE,$73,$E5,$2E,$5F,$D4,$E6,$E9,$FD
	.dc.b	$5D,$B8,$73,$BF,$CB,$5E,$27,$8E,$7F,$A8,$8F,$99,$D2,$69,$2B,$F4
	.dc.b	$8D,$C7,$FC,$CF,$3B,$9C,$9F,$E0,$1C,$7F,$B4,$70,$F8,$8F,$3F,$1F
	.dc.b	$F9,$D7,$8F,$15,$FC,$57,$BF,$F1,$EF,$F1,$D8,$FC,$2C,$EE,$30,$4C
	.dc.b	$7F,$95,$57,$06,$D7,$08,$C4,$7F,$82,$FE,$C3,$E1,$9F,$1A,$9F,$49
	.dc.b	$87,$FD,$D2,$FE,$1D,$C9,$FE,$A7,$70,$7D,$6F,$B0,$F9,$A7,$3F,$0E
	.dc.b	$CF,$F9,$72,$2D,$FF,$A0,$26,$F1,$7F,$83,$A6,$D7,$F8,$38,$FE,$7E
	.dc.b	$71,$3F,$DF,$73,$E3,$23,$FE,$B9,$7F,$0F,$EA,$FD,$E3,$D9,$9D,$96
	.dc.b	$1F,$E9,$08,$F6,$12,$3F,$D8,$4F,$E4,$3B,$FF,$0E,$7E,$33,$CD,$FF
	.dc.b	$2E,$6F,$8A,$FF,$3B,$CD,$B6,$7F,$A9,$BD,$D6,$8F,$F0,$5D,$F3,$F3
	.dc.b	$3F,$77,$5F,$AF,$F3,$9F,$4B,$FE,$F3,$BD,$9E,$89,$D4,$67,$1F,$F1
	.dc.b	$33,$5F,$68,$AE,$7F,$1D,$FC,$31,$B7,$F9,$3F,$97,$F8,$6F,$53,$FC
	.dc.b	$7A,$E2,$BF,$D6,$BF,$21,$FA,$A8,$C6,$3F,$C1,$79,$91,$8B,$E3,$7F
	.dc.b	$8B,$BE,$0F,$1E,$69,$3F,$E4,$EF,$1E,$CF,$C8,$8F,$0C,$E3,$1F,$F4
	.dc.b	$1F,$AC,$BC,$57,$9B,$AB,$C8,$7E,$AB,$E9,$3F,$CF,$11,$EB,$C5,$A7
	.dc.b	$9A,$B5,$CB,$FA,$AB,$DD,$7F,$84,$5A,$FF,$84,$5A,$FE,$23,$34,$62
	.dc.b	$3F,$A1,$36,$33,$3D,$4F,$E5,$DA,$BE,$A7,$E5,$1F,$C3,$BE,$93,$F9
	.dc.b	$86,$7E,$7C,$FE,$5C,$EA,$31,$5C,$5F,$F6,$A5,$70,$CB,$E4,$1E,$E9
	.dc.b	$D3,$8B,$FA,$BE,$27,$D9,$3F,$7D,$E1,$5C,$3A,$C5,$AF,$F7,$9C,$DE
	.dc.b	$0F,$BD,$87,$FB,$A2,$FC,$7E,$F7,$F2,$E9,$7D,$17,$FB,$BA,$72,$7D
	.dc.b	$18,$CF,$C2,$FF,$5C,$6C,$7F,$A8,$DB,$8F,$E6,$0F,$D4,$D7,$37,$11
	.dc.b	$A7,$F3,$AD,$7F,$49,$1B,$72,$D7,$F3,$7F,$37,$F9,$93,$F8,$F7,$C7
	.dc.b	$DB,$F9,$B1,$67,$E4,$42,$71,$7F,$0B,$EB,$FF,$44,$7F,$27,$F9,$2B
	.dc.b	$5F,$5A,$2D,$F2,$9B,$5F,$2B,$E7,$7F,$54,$7F,$4E,$FA,$5F,$ED,$31
	.dc.b	$9C,$3F,$31,$E4,$A7,$31,$FE,$3B,$6E,$1F,$C6,$BA,$7F,$BF,$CF,$09
	.dc.b	$CD,$71,$E1,$EC,$7F,$64,$BF,$D5,$79,$D8,$BF,$C6,$FC,$07,$EF,$4F
	.dc.b	$E8,$5F,$23,$63,$24,$AF,$8B,$19,$7F,$1D,$C9,$FC,$4F,$4F,$C3,$F8
	.dc.b	$1F,$D2,$FE,$97,$F8,$43,$30,$FC,$DB,$96,$9F,$03,$E1,$9F,$15,$FD
	.dc.b	$C9,$B1,$F8,$BE,$67,$CD,$E6,$BF,$25,$18,$9F,$C6,$37,$3F,$C4,$B5
	.dc.b	$FC,$2D,$7E,$4E,$31,$1F,$8A,$8C,$12,$7F,$93,$7C,$9F,$F2,$13,$FA
	.dc.b	$17,$C4,$7F,$BF,$A3,$F6,$B1,$FD,$17,$F3,$EF,$F1,$75,$FF,$32,$F2
	.dc.b	$D1,$88,$FE,$A3,$18,$EE,$FF,$1F,$B6,$FF,$45,$28,$61,$F3,$67,$8B
	.dc.b	$67,$4B,$0F,$2E,$BF,$CD,$F2,$97,$FC,$8F,$0E,$71,$8C,$3F,$84,$C6
	.dc.b	$27,$C2,$D8,$FF,$76,$61,$F1,$9E,$F5,$E8,$8C,$4E,$E7,$E0,$46,$0E
	.dc.b	$1D,$FF,$D1,$BE,$6B,$FA,$51,$FE,$0F,$FA,$5B,$8F,$F2,$10,$FF,$83
	.dc.b	$C8,$FF,$8A,$4C,$95,$CD,$FF,$AF,$E3,$2D,$8F,$F3,$27,$03,$DC,$FC
	.dc.b	$61,$FC,$67,$E9,$7F,$A0,$F2,$7F,$97,$FA,$95,$BC,$7F,$8D,$57,$F3
	.dc.b	$FA,$34,$9E,$37,$F1,$D6,$CF,$F1,$D6,$E5,$FE,$D3,$E7,$7C,$B7,$F6
	.dc.b	$B3,$FC,$61,$F4,$B6,$FE,$13,$FE,$2F,$23,$F0,$5B,$5E,$F6,$BF,$E0
	.dc.b	$0D,$FF,$CF,$51,$FD,$43,$C5,$FE,$F0,$7F,$8F,$7E,$96,$DF,$CF,$7F
	.dc.b	$C7,$84,$7E,$73,$6F,$E6,$35,$ED,$08,$FF,$4B,$C3,$E9,$DE,$3F,$F0
	.dc.b	$F6,$D7,$F7,$39,$D7,$91,$9A,$D3,$CA,$BF,$A7,$4F,$E8,$9A,$FF,$45
	.dc.b	$6B,$FE,$1E,$6B,$FE,$3C,$6B,$DA,$BF,$56,$D3,$A7,$95,$E5,$3C,$DE
	.dc.b	$6D,$9F,$35,$7F,$9E,$6B,$C5,$3E,$52,$4F,$F9,$EF,$E6,$1E,$47,$FA
	.dc.b	$39,$B1,$FA,$FC,$9F,$1B,$9A,$93,$FD,$FE,$B7,$87,$EC,$AF,$73,$9A
	.dc.b	$FF,$4C,$F3,$BC,$BA,$62,$2C,$FF,$82,$67,$4D,$9D,$34,$FA,$57,$FE
	.dc.b	$62,$E3,$FF,$31,$FD,$2F,$D8,$8F,$F2,$4F,$9F,$FE,$C1,$8C,$47,$F9
	.dc.b	$32,$77,$97,$F7,$7A,$D3,$FC,$62,$33,$3C,$AE,$BC,$45,$F4,$7F,$DD
	.dc.b	$51,$A7,$C8,$7F,$42,$AE,$D1,$FE,$78,$F1,$78,$4F,$F9,$40,$D1,$DC
	.dc.b	$A7,$7A,$ED,$87,$9A,$3F,$C1,$D8,$77,$8C,$46,$F0,$EE,$F9,$9F,$63
	.dc.b	$4E,$D5,$E2,$AF,$C6,$C7,$A0,$FC,$7B,$87,$CC,$7F,$87,$7E,$79,$F1
	.dc.b	$FF,$E7,$58,$FE,$7F,$18,$F5,$7F,$2E,$D8,$CC,$ED,$E5,$C1,$FE,$8E
	.dc.b	$78,$B5,$CF,$67,$28,$C4,$7F,$A1,$2B,$F2,$73,$F2,$F2,$C7,$C8,$FF
	.dc.b	$A2,$35,$FB,$F4,$8C,$4F,$1C,$3C,$15,$A6,$C7,$2C,$3A,$1F,$EA,$EB
	.dc.b	$E4,$E0,$3E,$27,$EB,$B6,$BC,$37,$E0,$9E,$4D,$78,$DF,$81,$EC,$D7
	.dc.b	$AC,$6E,$BC,$31,$DA,$75,$1D,$EF,$D1,$3E,$2A,$E1,$5D,$7A,$BF,$D8
	.dc.b	$FF,$2D,$FC,$29,$B5,$FE,$FB,$11,$96,$99,$70,$47,$07,$F6,$EF,$A5
	.dc.b	$FF,$4E,$37,$D2,$FE,$2F,$E3,$27,$33,$FE,$65,$8C,$4E,$A7,$33,$FE
	.dc.b	$0F,$6B,$FE,$0E,$6B,$FE,$0B,$6A,$FA,$57,$C5,$FE,$69,$B1,$FC,$A1
	.dc.b	$AF,$F8,$51,$AF,$F9,$DD,$7C,$C7,$12,$47,$17,$F9,$8F,$63,$F1,$CF
	.dc.b	$D3,$B6,$73,$18,$23,$82,$1F,$D0,$DF,$D7,$F8,$EF,$A4,$77,$FF,$3E
	.dc.b	$8C,$36,$33,$5F,$E2,$C9,$CB,$FE,$1E,$FB,$07,$C6,$7D,$06,$C6,$63
	.dc.b	$04,$70,$43,$EA,$47,$07,$FD,$35,$C2,$F0,$7F,$CA,$F1,$FD,$1F,$CE
	.dc.b	$FC,$54,$62,$36,$D7,$FC,$91,$F4,$6E,$FF,$C1,$B6,$3F,$C6,$AD,$7F
	.dc.b	$08,$D7,$8C,$70,$23,$C1,$FC,$D7,$73,$E7,$A9,$27,$FC,$A3,$8F,$FB
	.dc.b	$F2,$3F,$27,$D6,$F8,$9A,$4A,$FF,$23,$C6,$5B,$19,$7F,$31,$19,$6C
	.dc.b	$7C,$2F,$42,$47,$F6,$B6,$BE,$CF,$4F,$E0,$91,$F8,$2D,$74,$60,$87
	.dc.b	$F8,$BF,$7F,$79,$1F,$6E,$5F,$5D,$FF,$3B,$60,$F7,$7D,$E4,$8F,$6B
	.dc.b	$5F,$5B,$5E,$2F,$D6,$F9,$49,$1F,$59,$AF,$D6,$F9,$AF,$B6,$47,$ED
	.dc.b	$BF,$C0,$3E,$03,$FC,$57,$DD,$FE,$D1,$F2,$30,$4A,$FF,$3A,$CB,$E2
	.dc.b	$B8,$12,$32,$FF,$54,$FB,$07,$D1,$52,$4F,$F8,$92,$32,$D8,$CB,$FE
	.dc.b	$92,$8C,$B6,$3F,$AF,$FC,$4B,$D1,$A4,$9F,$F1,$7C,$65,$F9,$FA,$7E
	.dc.b	$40,$8E,$08,$7F,$99,$5C,$B4,$FF,$17,$3F,$A8,$23,$82,$1F,$AA,$72
	.dc.b	$D3,$FD,$57,$E5,$3F,$0B,$7A,$5F,$E4,$48,$C3,$63,$33,$C1,$A8,$82
	.dc.b	$78,$BB,$9E,$44,$7F,$88,$7B,$2F,$8C,$FE,$2C,$D8,$CC,$60,$CF,$B5
	.dc.b	$23,$8B,$5F,$08,$F0,$7F,$D2,$F1,$98,$F2,$C6,$1E,$27,$E8,$3E,$7F
	.dc.b	$F8,$09,$1C,$10,$FF,$4B,$39,$69,$F8,$7F,$A9,$FC,$61,$1C,$10,$FC
	.dc.b	$43,$96,$9F,$84,$F8,$5F,$C4,$91,$C1,$FB,$4E,$6F,$E2,$A4,$70,$43
	.dc.b	$FD,$54,$E5,$A7,$F3,$DF,$8B,$7D,$0F,$F4,$36,$C6,$63,$04,$70,$43
	.dc.b	$E6,$E2,$9E,$31,$FA,$3F,$80,$FD,$61,$1F,$D5,$35,$D1,$82,$1F,$E6
	.dc.b	$7F,$95,$FF,$53,$F8,$86,$93,$AB,$FA,$2E,$B2,$BF,$D1,$CD,$7F,$A3
	.dc.b	$B5,$FF,$60,$78,$C8,$C4,$7E,$5F,$C5,$78,$0F,$A0,$FE,$82,$49,$FE
	.dc.b	$46,$D7,$FD,$AC,$D7,$FD,$99,$1C,$5F,$F5,$5F,$D8,$B8,$7F,$CC,$2D
	.dc.b	$8C,$C6,$0F,$F5,$8F,$B4,$FA,$2F,$F3,$CC,$A3,$1E,$CE,$A7,$31,$89
	.dc.b	$FE,$3F,$F0,$1F,$C9,$48,$E0,$87,$F8,$97,$C3,$FF,$6A,$38,$3F,$6F
	.dc.b	$E2,$23,$31,$BE,$8F,$E6,$04,$7F,$96,$BF,$E3,$FF,$CA,$7F,$DB,$BB
	.dc.b	$9A,$4E,$2F,$E7,$1F,$53,$2F,$F9,$AB,$5F,$F0,$1C,$FF,$61,$7F,$DE
	.dc.b	$CD,$7E,$B3,$55,$77,$F3,$3C,$44,$8F,$EE,$B3,$B9,$D4,$62,$74,$F8
	.dc.b	$1A,$FC,$2C,$5F,$1D,$FD,$3C,$91,$FD,$31,$F8,$1E,$44,$8E,$ED,$7F
	.dc.b	$DE,$5E,$5F,$E0,$DB,$CB,$FE,$3E,$8C,$47,$FB,$D5,$AF,$FB,$D5,$AB
	.dc.b	$D5,$FF,$18,$CF,$12,$47,$F8,$DF,$E8,$5F,$3B,$FE,$3D,$6C,$66,$30
	.dc.b	$78,$7A,$64,$7F,$26,$3B,$61,$FA,$57,$18,$7F,$90,$27,$71,$EC,$6B
	.dc.b	$E9,$7F,$C9,$9F,$38,$F2,$69,$27,$FC,$99,$19,$6C,$65,$FE,$17,$19
	.dc.b	$6C,$7F,$99,$3E,$69,$C7,$49,$3E,$68,$CB,$63,$2F,$F4,$2C,$13,$A8
	.dc.b	$C3,$5F,$E8,$77,$FE,$1F,$7F,$AD,$78,$48,$C4,$7F,$4B,$6C,$67,$2F
	.dc.b	$F6,$EF,$CF,$BE,$AA,$92,$FE,$70,$8D,$38,$7F,$5E,$FF,$58,$6B,$FD
	.dc.b	$A9,$AF,$F6,$D6,$BC,$FE,$65,$F1,$3C,$DB,$1F,$76,$51,$88,$EC,$D7
	.dc.b	$B7,$CE,$7F,$99,$5E,$CD,$7D,$0D,$93,$9F,$37,$FD,$0B,$B7,$B9,$B5
	.dc.b	$F8,$D6,$BF,$DD,$1A,$FF,$81,$97,$C7,$7D,$38,$FF,$05,$46,$27,$A4
	.dc.b	$E6,$7F,$C2,$AE,$F4,$FF,$4A,$F3,$DD,$7F,$C8,$95,$FD,$59,$AF,$F8
	.dc.b	$01,$AF,$F8,$5F,$81,$FB,$D7,$F8,$DF,$1F,$D0,$B6,$7F,$14,$D7,$FC
	.dc.b	$3B,$3C,$3D,$3F,$43,$6B,$8C,$62,$B8,$57,$76,$BF,$E1,$5B,$7C,$C7
	.dc.b	$A4,$97,$FE,$B3,$8C,$B6,$33,$3E,$D6,$BE,$4F,$B1,$EB,$89,$A4,$C7
	.dc.b	$CE,$DB,$FD,$FB,$6B,$F6,$7F,$03,$F3,$79,$3F,$D9,$35,$FB,$A8,$C1
	.dc.b	$C8,$8F,$26,$CF,$EC,$6B,$8B,$63,$2F,$9D,$FC,$07,$77,$F4,$9E,$1B
	.dc.b	$DF,$F8,$EC,$3F,$9D,$C6,$0F,$C3,$78,$4F,$CF,$DF,$19,$86,$D7,$F1
	.dc.b	$D9,$D4,$E4,$EB,$CF,$27,$FA,$3B,$73,$FE,$A6,$D7,$FA,$9B,$5F,$EA
	.dc.b	$6D,$7F,$A9,$E3,$FA,$E9,$E6,$78,$70,$FF,$17,$B7,$D1,$F3,$11,$FE
	.dc.b	$AF,$C3,$F0,$08,$E0,$E5,$F1,$9E,$D2,$38,$3B,$7C,$6F,$C1,$23,$82
	.dc.b	$63,$FE,$64,$D3,$E4,$4E,$A3,$18,$A3,$F6,$5E,$4B,$E7,$9F,$03,$86
	.dc.b	$E3,$FD,$26,$75,$39,$3F,$6D,$F6,$3F,$DA,$A7,$F9,$0B,$71,$FE,$A0
	.dc.b	$D7,$FC,$EE,$D7,$FB,$FF,$57,$E0,$89,$3F,$DF,$1B,$1F,$E0,$86,$BF
	.dc.b	$E0,$87,$F1,$71,$F4,$E3,$FD,$07,$C7,$FD,$2B,$87,$24,$9F,$F7,$D9
	.dc.b	$B9,$FF,$4C,$CE,$E3,$33,$B9,$FF,$46,$FA,$8F,$D7,$FF,$D4,$15,$FD
	.dc.b	$AE,$31,$3F,$C6,$DB,$19,$9D,$CF,$F6,$EF,$17,$F8,$E7,$5F,$F9,$56
	.dc.b	$BF,$A3,$B6,$75,$18,$9D,$91,$C1,$FE,$7D,$F8,$EF,$E4,$9E,$C1,$A4
	.dc.b	$8D,$12,$BF,$CF,$73,$B8,$C4,$7F,$8C,$BE,$0B,$F6,$64,$8F,$F1,$6B
	.dc.b	$63,$FC,$72,$D7,$84,$7C,$3F,$A0,$7B,$9F,$0D,$B1,$98,$C1,$F0,$BE
	.dc.b	$56,$47,$C2,$6D,$7F,$85,$1A,$FF,$7D,$6B,$FE,$46,$F9,$EF,$E1,$24
	.dc.b	$9F,$F2,$1B,$63,$FC,$9E,$D7,$FC,$9E,$FF,$4B,$FA,$0F,$F0,$7C,$FF
	.dc.b	$89,$7C,$64,$62,$BF,$D4,$2D,$7F,$D9,$6D,$7F,$C1,$2D,$7F,$DA,$4D
	.dc.b	$8E,$73,$DC,$FE,$B5,$F3,$0F,$80,$A4,$8F,$14,$65,$FF,$7B,$FE,$F9
	.dc.b	$FA,$7C,$3F,$DD,$CD,$F8,$43,$49,$EB,$7F,$6D,$8C,$B6,$33,$5F,$A2
	.dc.b	$F0,$7F,$52,$47,$04,$BF,$D5,$BF,$D5,$1B,$1B,$7F,$CF,$7F,$41,$FC
	.dc.b	$E8,$93,$FE,$74,$6C,$6A,$30,$47,$FD,$F3,$83,$E8,$1B,$FA,$5B,$5F
	.dc.b	$E5,$99,$D4,$E4,$FD,$5F,$CF,$49,$FD,$5B,$6F,$FC,$BA,$D7,$FC,$8C
	.dc.b	$D4,$3C,$3F,$E9,$DC,$5F,$D3,$49,$5F,$CF,$5F,$EA,$7F,$41,$FB,$CA
	.dc.b	$FF,$55,$B5,$E7,$48,$C4,$75,$6C,$67,$CA,$E6,$4B,$FE,$B5,$3F,$DA
	.dc.b	$AB,$AF,$D0,$78,$09,$3A,$7F,$B2,$FD,$03,$2A,$49,$FE,$C9,$19,$6C
	.dc.b	$66,$BF,$A3,$4E,$5B,$1F,$DD,$3E,$22,$57,$FA,$B9,$B3,$FD,$37,$07
	.dc.b	$C8,$7C,$42,$57,$A9,$B3,$B6,$BF,$21,$FE,$73,$52,$7F,$9A,$B6,$7C
	.dc.b	$BF,$23,$EE,$24,$FB,$1B,$1A,$8C,$11,$FF,$67,$F2,$FF,$83,$89,$1F
	.dc.b	$EC,$87,$FC,$1B,$FA,$17,$EE,$76,$34,$93,$A2,$1F,$E0,$F7,$2F,$F4
	.dc.b	$C9,$FE,$A6,$47,$07,$F8,$57,$EC,$FB,$11,$FF,$0E,$B5,$D1,$82,$1F
	.dc.b	$C5,$7E,$C1,$F2,$DF,$E6,$36,$C6,$63,$07,$F9,$5B,$E6,$73,$97,$FA
	.dc.b	$95,$AF,$F9,$71,$AF,$F9,$95,$AF,$FB,$4D,$BE,$21,$A4,$9F,$EB,$7E
	.dc.b	$BC,$BF,$EB,$6D,$BF,$F6,$6F,$1F,$F6,$02,$4F,$F6,$07,$FC,$7B,$F1
	.dc.b	$5F,$E2,$89,$FF,$0C,$35,$FD,$A9,$FB,$57,$0F,$EB,$9A,$E7,$D3,$E6
	.dc.b	$4A,$FF,$22,$4E,$BC,$BF,$F5,$4B,$FE,$52,$F7,$8D,$24,$FF,$B9,$E7
	.dc.b	$6D,$9D,$B6,$77,$7F,$EC,$01,$F1,$BF,$C3,$3D,$7E,$B1,$EA,$F3,$7F
	.dc.b	$DB,$31,$88,$F6,$CE,$EB,$FC,$1F,$76,$7E,$09,$D0,$FF,$30,$C6,$9A
	.dc.b	$6A,$78,$3F,$C7,$2F,$BC,$FF,$C0,$19,$CC,$A7,$2F,$13,$99,$C6,$B6
	.dc.b	$F8,$6F,$D0,$F1,$8C,$D7,$4A,$E7,$3E,$C3,$99,$B2,$9C,$EF,$D5,$7D
	.dc.b	$23,$DC,$7F,$BA,$AB,$E6,$38,$3D,$47,$C4,$1F,$2D,$FE,$5B,$F4,$FF
	.dc.b	$DE,$2D,$8D,$35,$F9,$6D,$7F,$1A,$BE,$7D,$E0,$F2,$6C,$F2,$6B,$C9
	.dc.b	$AF,$27,$78,$7D,$DA,$7B,$1F,$9A,$8C,$37,$2E,$18,$7D,$8A,$F1,$36
	.dc.b	$75,$87,$F5,$69,$E6,$EF,$0E,$F7,$CA,$3E,$CC,$E9,$B5,$C9,$F8,$F5
	.dc.b	$C1,$E8,$D7,$E5,$C6,$E7,$96,$1E,$BA,$D4,$79,$AB,$A9,$E7,$3C,$98
	.dc.b	$7E,$67,$6F,$ED,$BF,$3F,$F5,$E5,$18,$7F,$5C,$23,$51,$98,$FF,$3D
	.dc.b	$99,$37,$1A,$70,$69,$ED,$D4,$F9,$FF,$F3,$24,$FF,$B1,$5A,$FF,$A9
	.dc.b	$5A,$FF,$0E,$8E,$1E,$8F,$29,$CC,$FF,$7C,$6B,$FE,$07,$6C,$F0,$8F
	.dc.b	$77,$CF,$3E,$7F,$F8,$23,$63,$31,$83,$3F,$3F,$23,$8B,$5F,$E7,$07
	.dc.b	$FA,$1B,$AB,$F8,$77,$5F,$F3,$38,$C1,$0F,$D2,$39,$6C,$70,$3F,$8B
	.dc.b	$79,$EF,$9F,$79,$1F,$56,$30,$D8,$CD,$7F,$99,$5A,$FF,$B3,$7E,$7D
	.dc.b	$F0,$9F,$41,$B1,$FE,$F3,$6B,$B8,$D1,$1D,$35,$FF,$5C,$FC,$BF,$F5
	.dc.b	$E7,$F9,$93,$5C,$F8,$3F,$86,$8C,$47,$D9,$F9,$FF,$E3,$8F,$F0,$5E
	.dc.b	$04,$62,$3F,$DC,$8D,$78,$76,$FE,$22,$31,$1F,$E3,$A6,$BF,$E4,$56
	.dc.b	$CF,$49,$FC,$BF,$6B,$C7,$A4,$BF,$E9,$51,$97,$C9,$D2,$F9,$2F,$CB
	.dc.b	$B6,$33,$18,$23,$82,$1F,$EF,$D8,$E0,$FC,$8F,$80,$F8,$84,$7E,$4B
	.dc.b	$ED,$F2,$E4,$7C,$5C,$FF,$C7,$2D,$8C,$BF,$19,$AF,$CD,$7D,$FB,$48
	.dc.b	$F6,$B5,$F2,$7B,$5E,$C2,$3F,$01,$AE,$8C,$10,$FA,$BF,$1D,$F4,$C8
	.dc.b	$F9,$DF,$AD,$EC,$48,$F5,$35,$FE,$CF,$BB,$E0,$1E,$37,$F0,$46,$CF
	.dc.b	$E1,$5A,$F0,$6B,$FC,$0D,$7C,$97,$12,$47,$17,$FC,$01,$F3,$EF,$9F
	.dc.b	$A4,$9F,$E4,$11,$96,$C6,$5F,$EB,$51,$96,$C7,$F8,$2F,$E7,$DE,$75
	.dc.b	$24,$FF,$2B,$8C,$BF,$8E,$D3,$F2,$64,$70,$43,$FC,$17,$E2,$FF,$85
	.dc.b	$A3,$51,$8E,$54,$7E,$B4,$91,$FB,$26,$BF,$B6,$6B,$FE,$36,$97,$47
	.dc.b	$28,$E0,$7F,$AE,$7E,$7D,$F2,$1F,$D0,$FC,$84,$62,$75,$39,$8C,$47
	.dc.b	$FB,$19,$AF,$F9,$12,$78,$CE,$C9,$1F,$EF,$BF,$9F,$7C,$FF,$F1,$D6
	.dc.b	$C6,$63,$06,$7D,$B9,$1C,$5A,$FF,$8C,$BC,$35,$27,$33,$FE,$56,$9C
	.dc.b	$C6,$23,$FD,$9B,$9F,$F9,$13,$E1,$BF,$83,$92,$3F,$82,$B5,$FA,$4D
	.dc.b	$7E,$66,$C8,$FC,$7F,$CD,$49,$FC,$6B,$5F,$C5,$B5,$FF,$24,$F1,$A3
	.dc.b	$11,$FE,$6C,$E1,$7C,$EF,$E3,$C9,$1F,$8D,$7F,$8E,$F4,$C8,$FE,$36
	.dc.b	$D7,$F1,$0D,$7C,$ED,$7F,$DD,$BA,$3E,$21,$EB,$7F,$4F,$6C,$FF,$8E
	.dc.b	$E7,$53,$92,$46,$67,$6F,$D3,$6B,$F4,$5A,$FF,$A2,$32,$7B,$9F,$B6
	.dc.b	$24,$7E,$D1,$AF,$EC,$9A,$FC,$87,$FC,$31,$E2,$BE,$7F,$F8,$2C,$FF
	.dc.b	$9B,$5A,$FF,$9A,$DB,$1B,$ED,$FE,$AD,$BC,$AF,$EB,$6D,$7F,$AF,$B5
	.dc.b	$FD,$57,$47,$DF,$76,$7F,$4A,$F9,$BF,$E9,$E4,$70,$43,$F7,$4E,$5A
	.dc.b	$7F,$8C,$7E,$7F,$EF,$B7,$83,$FD,$31,$18,$8F,$F3,$63,$5F,$F5,$8F
	.dc.b	$B2,$F9,$FF,$F3,$B4,$23,$1E,$D7,$F8,$F5,$B1,$FE,$75,$6B,$FC,$B7
	.dc.b	$E7,$FF,$99,$11,$FE,$64,$D7,$46,$08,$7F,$35,$E6,$FE,$74,$47,$F9
	.dc.b	$D3,$FE,$48,$F9,$4F,$A7,$D2,$69,$37,$FE,$85,$F3,$F2,$FF,$A1,$35
	.dc.b	$FE,$4B,$3F,$DB,$1F,$E3,$8D,$7F,$76,$D7,$FC,$B6,$DA,$F0,$79,$BE
	.dc.b	$32,$4F,$FB,$2E,$74,$48,$CC,$78,$1F,$EA,$DF,$19,$23,$FA,$B3,$5F
	.dc.b	$2F,$8D,$EF,$23,$D5,$FF,$37,$7A,$BF,$E3,$4E,$2B,$E5,$9A,$4A,$FF
	.dc.b	$23,$C6,$5B,$19,$7F,$DA,$33,$CD,$FD,$B7,$CF,$BE,$7F,$FC,$90,$D8
	.dc.b	$CC,$60,$F0,$F3,$C8,$EE,$D7,$AE,$1F,$9F,$6C,$65,$FF,$80,$37,$FE
	.dc.b	$4E,$8F,$F2,$B7,$CD,$7F,$4D,$9F,$E6,$8D,$7F,$99,$35,$FE,$22,$D5
	.dc.b	$DE,$8C,$37,$C9,$EA,$49,$FF,$35,$FC,$FF,$EF,$B0,$34,$9C,$7F,$41
	.dc.b	$B3,$C2,$30,$7A,$32,$91,$FC,$7F,$1F,$F3,$2B,$E8,$8F,$F5,$32,$F9
	.dc.b	$E7,$CF,$E1,$B1,$FE,$AB,$6B,$FE,$E3,$6A,$21,$CF,$9F,$91,$1E,$4F
	.dc.b	$5F,$97,$7C,$37,$56,$CF,$56,$BD,$0F,$33,$87,$BD,$3E,$73,$FC,$D2
	.dc.b	$DF,$07,$D1,$18,$3F,$CD,$BE,$6F,$FA,$2B,$5F,$F6,$33,$6F,$FD,$FC
	.dc.b	$D7,$C0,$D7,$DA,$BE,$7F,$F9,$9B,$67,$9C,$60,$FE,$73,$C5,$FE,$98
	.dc.b	$F4,$EE,$FF,$E4,$7A,$FE,$A4,$DC,$3F,$C0,$0D,$7F,$27,$F3,$FF,$E3
	.dc.b	$49,$FF,$57,$B6,$3F,$4E,$D7,$7E,$9F,$F4,$8E,$44,$E6,$7E,$B4,$62
	.dc.b	$75,$39,$94,$E6,$BC,$BF,$2B,$E8,$23,$82,$4F,$A2,$32,$D8,$FF,$62
	.dc.b	$FC,$FF,$5A,$34,$98,$79,$DB,$3C,$23,$07,$CA,$E5,$FF,$64,$C7,$E6
	.dc.b	$FE,$07,$04,$7B,$BC,$C8,$E0,$98,$7F,$0E,$BE,$6D,$9D,$BF,$CD,$38
	.dc.b	$3F,$B8,$F8,$3F,$CF,$5B,$E0,$FE,$3F,$0F,$E5,$D1,$FD,$DB,$D8,$7C
	.dc.b	$FB,$E2,$F0,$DA,$FF,$63,$37,$1F,$F4,$4B,$5F,$E8,$3D,$1D,$89,$3F
	.dc.b	$CE,$DB,$1F,$D2,$1A,$F0,$8D,$11,$D3,$5F,$DA,$DF,$9B,$D1,$F2,$92
	.dc.b	$7C,$AF,$BB,$86,$47,$F4,$B6,$D7,$A3,$E3,$3D,$A4,$9F,$6B,$F0,$3E
	.dc.b	$7E,$47,$C0,$6C,$FC,$06,$BE,$4C,$91,$88,$FE,$19,$E5,$3E,$7D,$F0
	.dc.b	$3F,$E2,$48,$C3,$63,$38,$7F,$80,$DB,$5F,$C2,$47,$CF,$BC,$AF,$F1
	.dc.b	$5B,$6B,$53,$A9,$C9,$F8,$0E,$B9,$3F,$DB,$5B,$3F,$DE,$9A,$FF,$7A
	.dc.b	$6B,$FE,$74,$87,$07,$E3,$1A,$79,$79,$3F,$D9,$31,$92,$46,$4D,$CF
	.dc.b	$FA,$B1,$AF,$FA,$8D,$AF,$F3,$3E,$F7,$CF,$FF,$35,$9F,$85,$18,$9F
	.dc.b	$F2,$6B,$63,$33,$B9,$F9,$1D,$7F,$E8,$39,$FE,$FF,$E3,$FF,$55,$23
	.dc.b	$82,$4F,$08,$CB,$FE,$69,$F8,$EF,$E4,$7E,$C1,$A4,$AF,$F3,$83,$67
	.dc.b	$84,$60,$FF,$0E,$7C,$17,$EC,$88,$E0,$91,$FE,$2C,$8C,$B6,$3E,$0F
	.dc.b	$CF,$FF,$AA,$DB,$EF,$7C,$28,$C4,$FF,$4F,$D1,$18,$8F,$F0,$AF,$CD
	.dc.b	$7F,$75,$25,$FF,$88,$9B,$1F,$21,$AF,$08,$FF,$1A,$7C,$FF,$F0,$A2
	.dc.b	$38,$24,$7F,$91,$23,$2D,$8F,$F4,$C7,$CF,$9F,$E9,$AD,$72,$60,$86
	.dc.b	$09,$1F,$EC,$D8,$E0,$FF,$B3,$FE,$61,$F0,$1F,$D6,$1B,$5F,$EA,$76
	.dc.b	$CE,$E7,$FD,$2B,$F0,$AF,$9F,$FD,$54,$FF,$77,$8C,$73,$EA,$73,$3F
	.dc.b	$EE,$B6,$BF,$E0,$99,$DE,$1F,$B7,$6C,$73,$9D,$A8,$CB,$FA,$1E,$FF
	.dc.b	$D6,$12,$BF,$50,$D8,$FD,$43,$5E,$11,$FE,$67,$F9,$FF,$E7,$A4,$70
	.dc.b	$48,$FF,$3B,$C6,$5B,$1E,$7F,$9F,$6D,$F1,$1B,$19,$8C,$1F,$E3,$DF
	.dc.b	$9E,$91,$FE,$6A,$6D,$7E,$C1,$AF,$F9,$91,$AF,$F8,$BF,$C4,$FE,$9F
	.dc.b	$C7,$FD,$38,$95,$FD,$39,$FE,$A9,$F3,$DF,$BB,$AF,$E8,$2D,$79,$35
	.dc.b	$52,$30,$DF,$27,$99,$2F,$F9,$64,$65,$B2,$F9,$BF,$09,$23,$2F,$F6
	.dc.b	$7F,$9F,$65,$49,$3D,$23,$2D,$8C,$E3,$FE,$43,$C1,$5A,$8C,$35,$FF
	.dc.b	$55,$FC,$3F,$F7,$82,$5F,$FA,$95,$F8,$BF,$1F,$23,$D2,$D9,$ED,$92
	.dc.b	$31,$1F,$29,$AF,$F3,$DA,$FE,$78,$4A,$FE,$70,$FB,$FE,$4A,$47,$B5
	.dc.b	$B3,$E8,$6B,$FD,$DB,$04,$62,$3F,$D8,$7C,$DF,$E1,$22,$57,$FA,$FD
	.dc.b	$FF,$09,$7C,$FF,$EF,$27,$FB,$FB,$5F,$C4,$35,$FB,$ED,$CC,$D2,$57
	.dc.b	$43,$FA,$C5,$47,$FA,$A3,$6B,$EA,$7C,$FF,$32,$4F,$F8,$95,$AF,$E2
	.dc.b	$1B,$1C,$23,$FC,$CF,$F3,$EF,$96,$A4,$8F,$EF,$5E,$C2,$75,$1F,$D5
	.dc.b	$BE,$66,$DF,$FB,$29,$AF,$FA,$F1,$AF,$F9,$89,$AF,$F9,$81,$BE,$29
	.dc.b	$A4,$C3,$FD,$69,$EB,$CB,$FF,$70,$B6,$FF,$B3,$F2,$7F,$AE,$49,$3F
	.dc.b	$EE,$67,$FC,$8D,$F1,$2F,$4E,$92,$7F,$57,$19,$6C,$66,$BF,$41,$39
	.dc.b	$6C,$7F,$95,$BD,$E4,$D2,$57,$F8,$36,$32,$D8,$CD,$7F,$AB,$7C,$CF
	.dc.b	$F5,$9C,$62,$3F,$C4,$7F,$23,$FE,$BD,$F5,$BA,$C7,$FA,$FB,$C8,$FE
	.dc.b	$A4,$FF,$97,$EB,$FC,$5D,$5B,$7F,$B5,$61,$67,$FD,$75,$35,$FF,$0B
	.dc.b	$45,$7F,$D9,$51,$6B,$AD,$F8,$27,$FD,$9B,$39,$AE,$10,$8C,$3E,$33
	.dc.b	$B9,$4E,$E5,$3B,$96,$BB,$DF,$AA,$31,$5D,$AB,$AD,$F8,$E7,$89,$F1
	.dc.b	$8F,$F7,$5C,$7F,$BB,$1F,$21,$FE,$51,$F9,$7F,$F2,$EF,$A9,$FE,$4E
	.dc.b	$6C,$67,$D5,$4E,$5B,$1C,$7E,$7F,$FB,$03,$7C,$6E,$51,$88,$FF,$73
	.dc.b	$5A,$31,$1F,$EF,$48,$C4,$FF,$BD,$FE,$83,$E6,$43,$D9,$FF,$7C,$C6
	.dc.b	$1B,$AF,$1C,$3C,$33,$C1,$B1,$D6,$77,$80,$5F,$57,$C1,$3C,$F0,$B1
	.dc.b	$B9,$F2,$BF,$5D,$E0,$DA,$E8,$E6,$FE,$1E,$9E,$3B,$E7,$3E,$0A,$F1
	.dc.b	$B4,$CD,$F6,$C7,$84,$1F,$80,$EA,$78,$CE,$F1,$F6,$6F,$ED,$F9,$CF
	.dc.b	$1E,$3F,$A2,$8C,$61,$FE,$1B,$DF,$6A,$B7,$18,$FA,$0B,$2A,$E2,$FF
	.dc.b	$87,$67,$8D,$6A,$73,$7D,$E9,$79,$E2,$7C,$15,$24,$F8,$CF,$EF,$BF
	.dc.b	$41,$FE,$71,$71,$82,$B8,$C6,$1A,$F8,$CF,$33,$8C,$3F,$80,$F8,$0B
	.dc.b	$E9,$5B,$1F,$41,$F0,$5F,$F5,$6B,$5F,$F5,$A3,$5F,$BA,$70,$CF,$FD
	.dc.b	$66,$F2,$CF,$E3,$B4,$F6,$14,$F0,$4F,$5C,$3F,$8A,$F7,$3E,$7D,$E4
	.dc.b	$FF,$AF,$A3,$0D,$8C,$DF,$F4,$E6,$BE,$49,$DB,$99,$E0,$47,$FA,$AF
	.dc.b	$C7,$3C,$7F,$BB,$48,$C4,$FD,$38,$C4,$EA,$FB,$E1,$F8,$BF,$96,$FE
	.dc.b	$C5,$7F,$EC,$56,$BF,$CD,$DA,$FE,$93,$81,$18,$8F,$CB,$B5,$FC,$73
	.dc.b	$63,$7E,$07,$FB,$81,$B3,$FE,$CF,$F9,$3F,$C8,$72,$1A,$7E,$4E,$FC
	.dc.b	$78,$7E,$87,$B7,$FC,$2F,$7F,$ED,$B6,$BF,$D3,$5A,$F9,$DA,$FE,$B1
	.dc.b	$AF,$E8,$DB,$87,$F3,$DF,$23,$CD,$18,$F8,$47,$C5,$FE,$95,$B7,$F9
	.dc.b	$C6,$BF,$99,$BE,$4F,$E8,$31,$7C,$1F,$C1,$6C,$7C,$F8,$C1,$1C,$10
	.dc.b	$FA,$4F,$C9,$D1,$F1,$9F,$20,$93,$DE,$7F,$75,$93,$E2,$BD,$64,$9F
	.dc.b	$90,$D7,$E4,$35,$F9,$0F,$D7,$C9,$F5,$FF,$04,$93,$FB,$CA,$EC,$FF
	.dc.b	$01,$EA,$7C,$FB,$E4,$3F,$5F,$3F,$E8,$B8,$C3,$FE,$7C,$6B,$FE,$6F
	.dc.b	$6B,$FC,$CA,$79,$CF,$28,$FF,$44,$CE,$9E,$91,$99,$DC,$7F,$A2,$3C
	.dc.b	$07,$BD,$FC,$8A,$7F,$A3,$B6,$33,$3E,$9B,$F1,$F1,$FF,$10,$F8,$6F
	.dc.b	$F0,$8C,$FF,$15,$C1,$18,$9D,$70,$75,$BF,$F6,$ED,$FE,$17,$97,$E5
	.dc.b	$11,$F8,$7C,$6D,$FF,$41,$18,$7E,$EB,$F8,$DE,$2F,$C8,$1E,$9C,$9F
	.dc.b	$27,$FC,$5B,$E6,$DE,$77,$F1,$A6,$DF,$C7,$6B,$F4,$9A,$FF,$22,$6B
	.dc.b	$FD,$81,$AF,$F6,$D6,$A7,$3D,$9F,$C3,$F9,$3F,$88,$F6,$FE,$1A,$BF
	.dc.b	$88,$35,$FE,$6D,$5E,$2E,$0F,$EB,$5A,$3E,$5B,$F5,$84,$AF,$C6,$72
	.dc.b	$D7,$F5,$93,$B7,$F6,$13,$F9,$E9,$FE,$B1,$F5,$3F,$D5,$A7,$FA,$7B
	.dc.b	$5F,$DA,$36,$FF,$90,$4F,$0F,$0A,$57,$07,$F8,$84,$7F,$2C,$F9,$D7
	.dc.b	$CF,$52,$57,$F5,$FB,$7C,$FB,$E5,$73,$3F,$D6,$F8,$18,$72,$8F,$E6
	.dc.b	$8D,$7F,$B0,$5F,$2A,$FE,$1D,$F6,$52,$7F,$87,$3D,$27,$FC,$05,$A7
	.dc.b	$F5,$1E,$9F,$F0,$3B,$6B,$FB,$F3,$71,$F0,$FC,$07,$C6,$F2,$65,$7C
	.dc.b	$67,$DD,$1F,$C7,$7E,$CA,$4F,$F1,$D7,$F8,$5F,$C7,$62,$3E,$9C,$5C
	.dc.b	$BF,$9A,$4F,$E4,$FE,$7F,$F2,$44,$AF,$C9,$35,$FC,$93,$5F,$C9,$3F
	.dc.b	$D1,$72,$7D,$27,$DC,$24,$FF,$72,$E6,$57,$F3,$3F,$A3,$93,$FC,$CD
	.dc.b	$FE,$43,$F1,$98,$8F,$C4,$C5,$CB,$FA,$54,$FE,$23,$DB,$95,$F8,$87
	.dc.b	$FB,$5F,$C8,$FF,$AA,$63,$04,$3F,$17,$1A,$3F,$C5,$BE,$E3,$E8,$BF
	.dc.b	$83,$CF,$FA,$65,$AF,$F9,$5E,$35,$C7,$FC,$6D,$B1,$FC,$99,$AF,$F8
	.dc.b	$59,$AF,$38,$E1,$3D,$21,$C7,$FE,$01,$F1,$3F,$66,$48,$FF,$1A,$5F
	.dc.b	$4C,$FF,$13,$F0,$BF,$A8,$24,$FF,$57,$E2,$AF,$EA,$22,$CF,$F8,$1E
	.dc.b	$7F,$B1,$7C,$74,$AF,$EC,$4F,$F5,$0F,$03,$11,$FB,$88,$B5,$FD,$E6
	.dc.b	$7F,$CA,$7E,$FB,$E8,$7F,$D3,$D7,$FE,$7B,$8C,$7B,$7F,$CF,$DB,$1F
	.dc.b	$D2,$9A,$FF,$8E,$BC,$BF,$E6,$7E,$97,$F2,$52,$47,$F9,$4E,$77,$8F
	.dc.b	$F9,$A7,$5F,$F2,$94,$FF,$95,$1A,$FF,$79,$6C,$7F,$41,$AF,$F3,$24
	.dc.b	$63,$87,$FC,$EA,$DA,$FF,$36,$37,$0F,$1C,$7F,$81,$67,$FB,$51,$1F
	.dc.b	$E7,$CF,$F8,$8E,$5E,$F7,$62,$4F,$13,$FC,$1D,$F4,$3F,$DC,$C8,$FF
	.dc.b	$47,$79,$78,$BE,$62,$47,$03,$DC,$E1,$F7,$3F,$E1,$B8,$C6,$5F,$16
	.dc.b	$2E,$1F,$E9,$2F,$1D,$5F,$E2,$1F,$2B,$FC,$08,$4C,$3F,$AC,$3E,$BF
	.dc.b	$07,$E3,$92,$3F,$C4,$D1,$FE,$56,$FA,$17,$D0,$FF,$99,$5B,$3F,$E0
	.dc.b	$6E,$84,$62,$3F,$CC,$8D,$7F,$D6,$B1,$A9,$E4,$48,$E5,$3B,$7F,$92
	.dc.b	$FD,$7F,$F4,$62,$3F,$E6,$58,$ED,$AF,$F9,$EB,$E7,$FF,$C1,$93,$FA
	.dc.b	$C6,$E0,$69,$3C,$BF,$F3,$F3,$67,$D4,$F2,$F4,$5E,$2F,$FA,$63,$0F
	.dc.b	$E7,$3F,$31,$FD,$60,$FF,$3C,$D7,$F8,$43,$87,$F9,$DC,$67,$0F,$EB
	.dc.b	$6E,$B1,$DF,$AF,$E0,$24,$F8,$1A,$FF,$A2,$EF,$FC,$6A,$D8,$CB,$FE
	.dc.b	$7B,$E5,$FE,$5E,$FE,$8F,$C7,$FC,$F9,$D9,$FE,$13,$F5,$DA,$7F,$8A
	.dc.b	$35,$FF,$73,$CA,$31,$1F,$C7,$DA,$FF,$BD,$FE,$8B,$FC,$CF,$A7,$FA
	.dc.b	$A3,$F7,$BF,$F4,$BC,$BD,$AF,$F5,$D7,$0A,$73,$1F,$EB,$BB,$E3,$BF
	.dc.b	$FB,$D9,$7C,$EF,$EE,$B0,$C3,$5F,$03,$54,$70,$F8,$5F,$F2,$62,$FA
	.dc.b	$16,$B8,$6D,$7F,$96,$A7,$53,$93,$FC,$A9,$E9,$7F,$9B,$F1,$FF,$58
	.dc.b	$67,$FE,$68,$57,$ED,$EA,$AF,$F8,$1F,$85,$3C,$3D,$7B,$3B,$3F,$9E
	.dc.b	$FD,$7B,$EB,$7F,$A5,$F0,$A3,$17,$FA,$96,$D7,$FB,$2D,$AF,$DF,$F3
	.dc.b	$7A,$F1,$FF,$23,$6D,$79,$AF,$FD,$11,$ED,$E2,$FF,$8E,$D5,$9C,$D7
	.dc.b	$ED,$F3,$FF,$31,$B6,$FC,$B1,$8B,$FF,$6E,$35,$FD,$CB,$5F,$03,$5F
	.dc.b	$EB,$6D,$EB,$FD,$4F,$77,$8D,$B7,$FE,$DD,$97,$CD,$E7,$B7,$F8,$6C
	.dc.b	$FF,$90,$AF,$07,$A6,$B7,$19,$39,$57,$A0,$8F,$B5,$B8,$7A,$9A,$FA
	.dc.b	$9F,$EF,$B8,$7E,$FB,$CF,$F1,$1F,$1F,$B1,$F0,$F8,$C3,$FD,$E9,$18
	.dc.b	$7F,$83,$C7,$DC,$8C,$61,$FB,$98,$C1,$27,$64,$8F,$C5,$9F,$D4,$BC
	.dc.b	$FF,$8D,$CE,$7E,$47,$8B,$F9,$DA,$FE,$C1,$18,$EA,$F4,$CE,$6F,$9E
	.dc.b	$7F,$E9,$59,$C9,$EE,$C1,$F1,$7D,$8E,$BD,$B7,$2F,$ED,$91,$70,$FF
	.dc.b	$3C,$CD,$E5,$FE,$E5,$36,$BF,$A7,$4D,$A6,$3F,$EA,$58,$F2,$92,$7F
	.dc.b	$D4,$CF,$8F,$E3,$7D,$E7,$BB,$82,$CF,$FB,$D5,$F6,$91,$FF,$57,$3E
	.dc.b	$BF,$8E,$F8,$E7,$E7,$FC,$B7,$D1,$A3,$35,$FE,$07,$6E,$5F,$EC,$39
	.dc.b	$D4,$E4,$FF,$71,$38,$8D,$4F,$EA,$FC,$E7,$D1,$BE,$1F,$C1,$7B,$DF
	.dc.b	$F8,$0D,$6A,$7F,$C9,$2D,$C7,$FC,$45,$7C,$5B,$3F,$57,$EC,$BF,$02
	.dc.b	$7E,$03,$AE,$CF,$F8,$7E,$2E,$7F,$EA,$C9,$B8,$FF,$10,$E0,$FE,$65
	.dc.b	$3F,$AF,$F1,$BF,$8C,$E1,$FC,$6A,$B4,$FF,$30,$30,$7F,$1E,$AE,$2E
	.dc.b	$2B,$ED,$D7,$25,$5D,$E3,$0F,$4E,$1F,$04,$FE,$6B,$EC,$A5,$FF,$96
	.dc.b	$BE,$93,$6F,$F8,$CA,$3F,$8B,$72,$A7,$B5,$7F,$54,$6B,$FE,$71,$6B
	.dc.b	$FC,$F9,$B1,$C6,$3F,$CE,$ED,$AF,$F3,$A4,$E8,$E3,$97,$F5,$79,$F8
	.dc.b	$C7,$F2,$5F,$69,$E6,$E2,$FE,$15,$6D,$BA,$7F,$4E,$7E,$37,$82,$BE
	.dc.b	$31,$AC,$3F,$77,$E7,$C8,$E7,$D6,$FA,$4C,$6B,$FE,$59,$AB,$C1,$FE
	.dc.b	$EE,$9B,$7F,$57,$3F,$E1,$B3,$F4,$7B,$BE,$4F,$D2,$3E,$0B,$F0,$11
	.dc.b	$87,$E1,$C6,$6F,$FD,$2C,$DC,$BF,$D2,$CD,$7F,$C0,$6D,$7F,$D4,$0D
	.dc.b	$7F,$B6,$B7,$43,$9D,$FF,$B8,$F0,$7F,$1A,$3F,$8C,$FC,$65,$C3,$FC
	.dc.b	$F1,$17,$0F,$C5,$C7,$F9,$5E,$71,$9F,$E3,$7C,$DF,$CD,$1F,$D6,$7E
	.dc.b	$93,$FB,$7D,$7D,$28,$CB,$4F,$A7,$1A,$BE,$A7,$E6,$A7,$F9,$6F,$8F
	.dc.b	$2B,$EC,$FC,$E3,$E0,$71,$7F,$F0,$06,$31,$3A,$9F,$ED,$98,$BE,$91
	.dc.b	$F0,$1F,$DB,$6F,$78,$71,$BC,$1C,$2B,$78,$73,$3F,$17,$F3,$3F,$B7
	.dc.b	$3F,$B3,$3F,$D7,$AB,$D3,$BF,$F8,$B6,$75,$3C,$0F,$CA,$9B,$CB,$F6
	.dc.b	$91,$FD,$2B,$CB,$93,$F9,$0F,$A4,$B8,$7F,$A9,$E6,$DF,$F0,$9E,$95
	.dc.b	$62,$FF,$21,$F1,$DF,$C2,$CF,$83,$F4,$8F,$A4,$C4,$FF,$07,$C1,$39
	.dc.b	$8F,$EB,$B9,$7F,$08,$8C,$3F,$D8,$27,$75,$FE,$EE,$6D,$70,$6C,$EE
	.dc.b	$B8,$BF,$E3,$28,$CB,$63,$2F,$FA,$FB,$D4,$FE,$D6,$7F,$36,$E9,$B7
	.dc.b	$FE,$C9,$FB,$0F,$EC,$84,$7C,$FB,$FF,$92,$A3,$FE,$00,$F9,$6F,$A4
	.dc.b	$FE,$33,$8A,$30,$FF,$51,$8F,$DA,$DF,$F9,$22,$72,$FF,$2B,$7F,$BE
	.dc.b	$7C,$5F,$F5,$7F,$48,$F1,$E3,$C3,$6B,$E4,$D7,$86,$7C,$76,$69,$27
	.dc.b	$71,$96,$C6,$70,$FF,$02,$F7,$3E,$A7,$FC,$1E,$7F,$83,$BE,$6E,$DF
	.dc.b	$F1,$67,$FC,$1A,$47,$E2,$F2,$A3,$17,$FE,$10,$8C,$4E,$A7,$33,$FE
	.dc.b	$68,$DB,$FD,$0D,$E6,$7F,$55,$3F,$3F,$36,$BF,$36,$FF,$8D,$1E,$75
	.dc.b	$FE,$1C,$6B,$CD,$AF,$F5,$96,$BF,$E1,$26,$D7,$F9,$DA,$7A,$1F,$E2
	.dc.b	$3E,$0F,$F1,$79,$FE,$2E,$EA,$B5,$FC,$D9,$FF,$16,$91,$F8,$2D,$BF
	.dc.b	$80,$D7,$DD,$8A,$3F,$C9,$BE,$1B,$E7,$BF,$BA,$1F,$BE,$8B,$7F,$BA
	.dc.b	$7E,$01,$1F,$C9,$9C,$CA,$E6,$FE,$23,$5F,$F6,$27,$CA,$7F,$92,$8F
	.dc.b	$F4,$07,$D1,$3E,$73,$17,$FD,$62,$30,$FF,$BE,$5B,$5F,$E5,$4F,$2D
	.dc.b	$18,$8F,$F7,$CB,$5F,$F4,$83,$54,$70,$AF,$F4,$FC,$E6,$31,$5F,$C1
	.dc.b	$29,$F4,$4F,$8D,$FF,$4C,$46,$1B,$0E,$EB,$18,$7F,$D5,$F3,$FE,$6A
	.dc.b	$9D,$57,$56,$C7,$F1,$26,$BF,$93,$F9,$C7,$D6,$FF,$A5,$5B,$5F,$B7
	.dc.b	$C1,$F0,$19,$A9,$3F,$E0,$4A,$DD,$FF,$4E,$EA,$F5,$5E,$A3,$F9,$CF
	.dc.b	$C3,$78,$BC,$1F,$F1,$B3,$70,$FF,$18,$3F,$05,$FF,$07,$B9,$9F,$F3
	.dc.b	$5B,$79,$CD,$38,$47,$6D,$7E,$4B,$FE,$D8,$E4,$FF,$6C,$FD,$17,$CD
	.dc.b	$8F,$F7,$FE,$18,$DB,$F9,$AD,$59,$C8,$AE,$0F,$FB,$82,$34,$FF,$B9
	.dc.b	$3A,$FF,$DB,$F3,$A8,$F0,$1C,$6B,$FD,$3B,$3A,$7B,$C3,$E3,$F8,$4F
	.dc.b	$FB,$A1,$E5,$1F,$E5,$A7,$38,$7F,$BC,$9C,$E1,$E2,$9F,$F1,$91,$DC
	.dc.b	$FF,$55,$B9,$AF,$04,$60,$D1,$E9,$AE,$4F,$F9,$A3,$0E,$EE,$8F,$59
	.dc.b	$F9,$7A,$FC,$C3,$83,$D6,$7F,$AE,$BE,$7B,$FD,$21,$ED,$FE,$75,$B1
	.dc.b	$FD,$89,$AE,$FD,$D4,$EF,$DF,$FF,$36,$CE,$DA,$F5,$6A,$25,$23,$31
	.dc.b	$87,$9E,$7F,$6A,$7F,$D8,$33,$F2,$F4,$E4,$A3,$2F,$D5,$8C,$C7,$5E
	.dc.b	$0E,$13,$D5,$AF,$1F,$80,$E7,$7D,$DB,$7C,$B0,$F1,$B6,$F8,$E1,$E0
	.dc.b	$BE,$18,$62,$F6,$F7,$3C,$87,$C0,$C3,$AB,$4E,$17,$1D,$9E,$53,$BC
	.dc.b	$70,$7E,$04,$F6,$E1,$F9,$9F,$17,$FC,$55,$B3,$FE,$04,$11,$96,$99
	.dc.b	$70,$47,$07,$F7,$AF,$A2,$FE,$E0,$DF,$0F,$F9,$17,$32,$73,$3F,$50
	.dc.b	$FF,$07,$38,$7E,$BC,$6E,$78,$CE,$67,$FC,$20,$D7,$8C,$76,$FA,$DF
	.dc.b	$F6,$C4,$7F,$83,$1A,$F7,$A4,$62,$3B,$B5,$FF,$63,$74,$3E,$9F,$E3
	.dc.b	$36,$3F,$C5,$2D,$7D,$59,$23,$11,$D3,$E1,$FC,$84,$7C,$8D,$7C,$96
	.dc.b	$8C,$47,$13,$BF,$7F,$88,$8E,$0F,$F1,$4F,$DC,$FE,$95,$BB,$7D,$08
	.dc.b	$C4,$7C,$16,$BF,$39,$AF,$F8,$ED,$AF,$F9,$05,$B9,$9C,$F7,$BE,$5D
	.dc.b	$F2,$DF,$D7,$E3,$0D,$8C,$DF,$F9,$9A,$72,$FF,$90,$3C,$07,$AF,$F8
	.dc.b	$66,$C6,$63,$04,$70,$43,$EE,$47,$07,$FC,$E1,$F6,$FF,$CC,$5F,$F3
	.dc.b	$73,$5C,$F3,$FF,$75,$E7,$4E,$67,$F2,$35,$FE,$CF,$7C,$DE,$4F,$94
	.dc.b	$8E,$08,$78,$9C,$BF,$84,$C7,$F9,$5F,$1F,$F8,$22,$3F,$61,$18,$3F
	.dc.b	$97,$EC,$FD,$13,$FC,$C1,$3F,$E6,$46,$BF,$99,$6B,$FA,$66,$BF,$E7
	.dc.b	$36,$BF,$E7,$96,$C7,$5F,$37,$F6,$71,$AB,$FF,$0D,$F8,$5E,$42,$5F
	.dc.b	$FA,$1B,$04,$70,$FA,$B9,$3E,$56,$BE,$56,$BE,$56,$BE,$5F,$3B,$C2
	.dc.b	$47,$DC,$F9,$3D,$59,$1F,$05,$AF,$C1,$FA,$6F,$49,$1F,$8C,$FB,$3E
	.dc.b	$72,$47,$C9,$6B,$FE,$B6,$FB,$07,$CD,$7F,$5E,$6C,$7F,$8D,$DA,$EE
	.dc.b	$34,$47,$4D,$7F,$C9,$5F,$66,$F9,$1F,$D1,$36,$3F,$40,$D7,$71,$F8
	.dc.b	$2F,$86,$91,$F3,$DA,$FE,$25,$AF,$FA,$37,$64,$62,$3F,$B3,$6F,$FE
	.dc.b	$54,$F8,$2F,$C6,$92,$3E,$D3,$5F,$CA,$35,$E8,$FF,$80,$3E,$83,$FB
	.dc.b	$2C,$7F,$35,$6B,$FD,$A9,$AF,$F7,$56,$A6,$93,$C7,$E8,$7F,$05,$FD
	.dc.b	$75,$EF,$52,$57,$F0,$58,$CB,$F9,$1E,$6F,$D3,$91,$C1,$0F,$D7,$39
	.dc.b	$69,$F9,$9F,$47,$F5,$E4,$70,$43,$F6,$CE,$5A,$7F,$9E,$7F,$00,$73
	.dc.b	$7F,$89,$1B,$19,$8C,$11,$C1,$0F,$EB,$C2,$38,$B8,$FA,$3F,$EA,$0F
	.dc.b	$F8,$7D,$A9,$A6,$B5,$FF,$48,$F9,$DF,$C2,$89,$3F,$B7,$6B,$FC,$45
	.dc.b	$AF,$E2,$5F,$E6,$DF,$B0,$FF,$4F,$8F,$E6,$8D,$7F,$D4,$8D,$7F,$C5
	.dc.b	$3C,$28,$C4,$7F,$8D,$63,$1E,$47,$42,$47,$47,$FC,$97,$F2,$DF,$E0
	.dc.b	$A8,$F8,$6D,$7F,$C0,$4D,$7F,$C6,$ED,$7F,$C8,$4D,$4D,$27,$9F,$FE
	.dc.b	$58,$CB,$CB,$3F,$2E,$B0,$47,$FA,$97,$C6,$7E,$A9,$BC,$BF,$D6,$63
	.dc.b	$11,$FE,$C5,$6B,$F0,$9A,$FC,$16,$BF,$EC,$46,$E2,$73,$A7,$FB,$23
	.dc.b	$C2,$FE,$82,$4A,$FE,$5C,$D7,$FA,$43,$5F,$E0,$0F,$F8,$AB,$F4,$7F
	.dc.b	$ED,$71,$FE,$1B,$6B,$FE,$28,$6B,$FE,$69,$F0,$91,$88,$57,$FB,$6B
	.dc.b	$ED,$BF,$A9,$92,$3F,$A0,$3D,$FE,$AE,$47,$E6,$AF,$9C,$62,$3C,$4D
	.dc.b	$7C,$4D,$7C,$56,$FA,$9F,$51,$23,$A4,$65,$FE,$A1,$F0,$BF,$D9,$88
	.dc.b	$E0,$DF,$5F,$C3,$23,$83,$FA,$C7,$37,$F6,$E2,$38,$21,$FD,$D5,$CB
	.dc.b	$4C,$9F,$CB,$DC,$B4,$FF,$7E,$7E,$64,$F6,$7F,$D4,$ED,$8C,$C6,$0F
	.dc.b	$F1,$27,$DD,$B3,$A4,$9F,$F1,$24,$65,$FE,$0F,$E4,$3F,$09,$76,$7F
	.dc.b	$0A,$9F,$EE,$71,$87,$6D,$7F,$DA,$6D,$7F,$D4,$33,$CE,$79,$36,$38
	.dc.b	$C6,$27,$84,$7F,$4C,$E4,$FF,$2D,$B7,$D7,$FF,$0C,$91,$FE,$3B,$F0
	.dc.b	$2F,$BA,$7A,$FE,$78,$C3,$63,$33,$C1,$AF,$FA,$0E,$76,$E6,$78,$11
	.dc.b	$FF,$25,$7E,$F0,$FA,$4F,$F2,$83,$63,$31,$82,$38,$21,$FE,$91,$C5
	.dc.b	$3C,$63,$C5,$87,$F3,$67,$FD,$93,$87,$B7,$87,$D4,$48,$F5,$35,$E2
	.dc.b	$D7,$F9,$4D,$3E,$AF,$CE,$48,$F5,$BF,$46,$7F,$DA,$FF,$07,$F6,$35
	.dc.b	$FE,$8E,$D8,$CD,$7F,$B9,$77,$FC,$34,$62,$13,$FC,$DF,$EB,$B0,$D8
	.dc.b	$FF,$76,$3D,$BC,$34,$E9,$B3,$A3,$FD,$D3,$F5,$0F,$9F,$7C,$3F,$E4
	.dc.b	$E3,$0D,$8C,$CF,$F6,$09,$CB,$7C,$EF,$F2,$5B,$73,$E5,$18,$8F,$F6
	.dc.b	$24,$FF,$8A,$E7,$FB,$4E,$4F,$8B,$FF,$07,$3E,$5A,$F7,$C6,$1F,$82
	.dc.b	$7B,$9C,$47,$CC,$6C,$7C,$9D,$7F,$D7,$D3,$CE,$BF,$61,$CB,$CA,$3E
	.dc.b	$93,$6B,$83,$57,$1F,$E2,$BE,$53,$C3,$1F,$80,$6B,$FD,$9F,$89,$18
	.dc.b	$8E,$31,$8B,$FD,$5F,$CB,$3D,$6D,$72,$A3,$13,$A9,$CC,$62,$FF,$54
	.dc.b	$D7,$F3,$2D,$42,$BA,$4F,$23,$C9,$8F,$E9,$3D,$EB,$8F,$FA,$DB,$1F
	.dc.b	$D8,$C6,$08,$E0,$98,$7E,$B6,$78,$3E,$4F,$67,$CE,$49,$F7,$B6,$35
	.dc.b	$18,$23,$FE,$AB,$EA,$F9,$1E,$37,$D0,$6E,$3F,$C4,$2B,$FD,$31,$1F
	.dc.b	$37,$7F,$E5,$59,$DE,$0F,$85,$8F,$F2,$A8,$C1,$2F,$F9,$5B,$63,$F9
	.dc.b	$5B,$F1,$3E,$5B,$FD,$47,$3F,$ED,$B6,$CF,$F0,$B6,$BF,$E2,$86,$BF
	.dc.b	$40,$7E,$68,$F8,$FF,$C1,$46,$1B,$19,$AF,$E8,$0D,$9F,$F6,$66,$BF
	.dc.b	$00,$93,$FE,$B7,$6C,$6A,$30,$47,$04,$3F,$D1,$91,$C1,$F3,$79,$0F
	.dc.b	$D2,$5E,$87,$E5,$E3,$0D,$8C,$DF,$F8,$15,$B5,$FE,$E1,$F9,$1F,$AA
	.dc.b	$49,$FF,$66,$3C,$7E,$56,$47,$16,$CF,$C7,$18,$78,$FC,$7F,$01,$2B
	.dc.b	$C0,$D9,$E1,$3B,$9C,$92,$3F,$DE,$9E,$FF,$E2,$09,$1F,$ED,$F7,$D9
	.dc.b	$F4,$32,$3D,$8D,$9F,$F7,$E7,$07,$E4,$09,$3F,$EE,$76,$C6,$A3,$04
	.dc.b	$7F,$C9,$9E,$1F,$EB,$BD,$1F,$F4,$6C,$FF,$95,$23,$04,$70,$49,$E2
	.dc.b	$FF,$24,$FE,$00,$7B,$3F,$C2,$E1,$18,$75,$19,$8C,$4F,$FA,$53,$F0
	.dc.b	$6A,$FC,$F6,$BF,$97,$6B,$FC,$21,$A9,$A4,$9F,$CD,$BC,$3A,$7F,$C0
	.dc.b	$7E,$0F,$F0,$CE,$AF,$E7,$58,$FF,$A9,$27,$2D,$8C,$E1,$FD,$D5,$AF
	.dc.b	$FA,$49,$AF,$FA,$45,$AF,$F0,$3A,$FE,$07,$EB,$4A,$FF,$58,$F6,$7F
	.dc.b	$36,$8C,$57,$E3,$63,$13,$AC,$15,$AA,$FF,$78,$3F,$27,$D6,$95,$F4
	.dc.b	$5B,$3F,$EB,$EF,$1F,$FA,$93,$F9,$96,$BF,$C0,$5A,$FF,$32,$6B,$FD
	.dc.b	$B3,$D1,$FE,$07,$18,$23,$FC,$7B,$EB,$A5,$F9,$DB,$5E,$BC,$51,$88
	.dc.b	$FE,$21,$97,$F5,$68,$C4,$7F,$8F,$7E,$0F,$F9,$49,$2F,$E1,$36,$3A
	.dc.b	$35,$E7,$68,$FE,$8F,$FA,$FB,$E3,$69,$2F,$EF,$46,$5F,$F2,$D7,$ED
	.dc.b	$BF,$63,$03,$49,$BF,$F2,$46,$CF,$08,$C1,$FD,$03,$C8,$91,$F6,$5B
	.dc.b	$5F,$0D,$AF,$A1,$AF,$F9,$77,$E3,$3F,$A4,$92,$7F,$04,$D8,$F9,$8D
	.dc.b	$7E,$0B,$FD,$9B,$F6,$0F,$9F,$3F,$9B,$6C,$FF,$93,$5A,$FF,$BA,$9A
	.dc.b	$FE,$0F,$DF,$FB,$E4,$9F,$F3,$F7,$0F,$FB,$79,$B1,$99,$FC,$73,$5F
	.dc.b	$E6,$FF,$08,$F7,$FF,$9B,$36,$BF,$D7,$0D,$9F,$E4,$AD,$78,$FB,$FC
	.dc.b	$09,$3C,$1E,$7E,$84,$8E,$6D,$9E,$7E,$3F,$80,$93,$E0,$7C,$7E,$2C
	.dc.b	$8F,$1B,$67,$C6,$D7,$D6,$D7,$FC,$2F,$CD,$FE,$D8,$9F,$F0,$0B,$FE
	.dc.b	$5E,$7F,$7A,$D8,$FD,$E3,$5F,$8B,$BF,$C4,$24,$FC,$46,$C7,$C4,$6B
	.dc.b	$C2,$3A,$FD,$9F,$80,$8E,$0F,$F7,$27,$E9,$DF,$E6,$56,$FA,$BF,$D9
	.dc.b	$E3,$15,$FD,$51,$AF,$F6,$E6,$BF,$EF,$76,$BF,$D8,$1B,$E3,$1C,$FC
	.dc.b	$77,$FA,$BF,$E3,$FF,$C9,$4D,$C3,$FB,$45,$A3,$0F,$F6,$B6,$BE,$1F
	.dc.b	$9D,$F3,$12,$7D,$6D,$8F,$43,$5E,$11,$F4,$3E,$33,$E9,$91,$C1,$F1
	.dc.b	$3E,$C7,$E1,$11,$C1,$27,$E2,$46,$5B,$1F,$CE,$0F,$E7,$BB,$7F,$9F
	.dc.b	$9C,$74,$FF,$90,$9F,$E8,$3A,$3E,$FB,$F0,$17,$FE,$5E,$8C,$4F,$F8
	.dc.b	$66,$73,$3F,$E0,$C6,$CE,$A7,$E4,$B6,$7F,$C1,$73,$A3,$8D,$FF,$AB
	.dc.b	$BE,$4F,$C4,$4B,$F4,$3D,$FD,$C9,$1D,$29,$18,$BF,$F0,$BF,$C5,$7F
	.dc.b	$B1,$6B,$C2,$D8,$CF,$C0,$7F,$8C,$63,$11,$FE,$36,$6B,$EF,$6B,$FE
	.dc.b	$28,$6E,$47,$39,$7F,$AE,$BC,$AF,$F3,$5F,$6F,$F7,$E8,$C6,$3F,$DA
	.dc.b	$23,$04,$9D,$92,$3F,$1F,$87,$F9,$77,$E6,$FF,$D1,$BC,$7D,$E3,$FC
	.dc.b	$C1,$ED,$7F,$AD,$23,$11,$B0,$E1,$FF,$35,$B5,$4F,$4C,$BF,$C8,$95
	.dc.b	$CD,$FC,$FC,$66,$7B,$DD,$7B,$5F,$5B,$E1,$1F,$E4,$DA,$D3,$E0,$BE
	.dc.b	$6F,$91,$EF,$7A,$8F,$19,$93,$B9,$4E,$F1,$C1,$C1,$FE,$B0,$9E,$F5
	.dc.b	$E2,$8F,$84,$7C,$A8,$F9,$6E,$CA,$FC,$EF,$9A,$7A,$8F,$9B,$C3,$6B
	.dc.b	$F8,$0C,$EA,$72,$71,$F1,$FF,$82,$4E,$9A,$FF,$9B,$E7,$FB,$D7,$46
	.dc.b	$C9,$1E,$D8,$78,$3F,$E9,$BD,$7E,$DC,$62,$7E,$83,$C9,$E1,$7E,$09
	.dc.b	$DD,$71,$D3,$FD,$49,$7C,$6F,$78,$72,$C3,$85,$6A,$F8,$57,$FB,$6A
	.dc.b	$F8,$61,$E2,$BD,$CE,$5A,$A7,$77,$CE,$3B,$61,$1F,$76,$1D,$9E,$0D
	.dc.b	$AF,$66,$1F,$9D,$E0,$FF,$75,$FB,$1F,$A2,$6D,$FF,$94,$0F,$F5,$49
	.dc.b	$63,$4E,$0F,$F2,$91,$4F,$F1,$B9,$4F,$EC,$25,$3F,$D4,$7F,$36,$F9
	.dc.b	$BF,$D8,$F2,$23,$13,$FC,$69,$AF,$FA,$91,$AF,$FA,$F9,$AF,$F2,$76
	.dc.b	$BD,$A7,$AC,$62,$5F,$34,$F0,$FC,$2D,$BF,$E5,$2D,$7F,$D6,$6D,$7F
	.dc.b	$CA,$7F,$36,$F1,$7E,$43,$63,$FC,$EE,$D7,$FC,$EC,$D7,$FD,$AB,$3F
	.dc.b	$E7,$67,$BF,$A5,$E2,$23,$82,$1E,$27,$2F,$4F,$4F,$C2,$47,$C2,$FC
	.dc.b	$CF,$9B,$FF,$2E,$47,$F4,$46,$BF,$EC,$E6,$BF,$EE,$86,$BF,$D4,$5A
	.dc.b	$FF,$B6,$9B,$91,$A6,$BD,$17,$9C,$F0,$7F,$96,$C6,$1B,$19,$BF,$F5
	.dc.b	$53,$5F,$F7,$0F,$98,$F8,$3F,$C2,$B6,$3F,$AE,$B5,$DC,$68,$8E,$9A
	.dc.b	$FF,$82,$7E,$6F,$F9,$BB,$FE,$0E,$E8,$46,$23,$5C,$DF,$D1,$27,$2F
	.dc.b	$8B,$E6,$E4,$F8,$9A,$FA,$1A,$8F,$EF,$58,$23,$31,$FE,$44,$CF,$F9
	.dc.b	$3F,$39,$A4,$8F,$D2,$FC,$DF,$FB,$99,$FE,$FA,$D7,$B1,$FE,$07,$70
	.dc.b	$E3,$74,$62,$35,$E1,$74,$23,$D1,$FE,$05,$4F,$06,$92,$7F,$D9,$CF
	.dc.b	$93,$DB,$91,$E4,$6B,$E4,$18,$78,$A3,$31,$E5,$C1,$18,$8F,$57,$47
	.dc.b	$C0,$23,$F0,$1F,$87,$EF,$C8,$F8,$6D,$7C,$17,$F0,$3D,$EF,$92,$48
	.dc.b	$F9,$2F,$CC,$F6,$24,$7C,$C6,$BF,$E5,$2F,$9B,$79,$DF,$D7,$1B,$1F
	.dc.b	$E0,$86,$BB,$8D,$11,$D3,$5F,$89,$AB,$E6,$FF,$C5,$71,$FE,$60,$8C
	.dc.b	$79,$9F,$A8,$9C,$CF,$06,$BF,$ED,$AF,$77,$EB,$91,$FF,$34,$F1,$7E
	.dc.b	$2E,$31,$1F,$EE,$D6,$BC,$4D,$DB,$E6,$FF,$02,$48,$CB,$5F,$F7,$B3
	.dc.b	$5F,$F7,$5B,$FE,$00,$F7,$BF,$B1,$47,$E0,$5A,$FE,$59,$AF,$E9,$5A
	.dc.b	$9A,$4F,$1F,$A1,$FC,$2B,$E6,$9F,$37,$49,$5F,$26,$32,$FB,$39,$FF
	.dc.b	$36,$47,$04,$3F,$2D,$9F,$EC,$63,$51,$8C,$BF,$50,$A7,$F6,$DF,$27
	.dc.b	$FA,$62,$47,$CA,$6B,$C5,$AF,$ED,$A3,$8B,$FE,$50,$F9,$BF,$D0,$B7
	.dc.b	$9F,$FC,$53,$18,$8F,$F5,$1B,$5F,$D6,$35,$FE,$02,$D7,$F8,$83,$53
	.dc.b	$7E,$6F,$FA,$D4,$7F,$08,$6B,$FC,$5D,$A9,$CD,$7F,$15,$F3,$7F,$82
	.dc.b	$92,$7F,$12,$D7,$E4,$35,$F8,$18,$FF,$98,$DF,$E3,$DF,$25,$FD,$85
	.dc.b	$FD,$1B,$5F,$E3,$70,$8C,$47,$F2,$B8,$C7,$8F,$FD,$32,$73,$3D,$23
	.dc.b	$04,$7F,$CF,$BF,$01,$FE,$11,$4D,$27,$9D,$FE,$58,$6C,$70,$8C,$1F
	.dc.b	$E5,$CC,$BB,$C7,$C5,$AC,$11,$FD,$D7,$C3,$B9,$7F,$B0,$B6,$33,$18
	.dc.b	$23,$82,$1F,$C1,$A3,$83,$F9,$3F,$0B,$F9,$A9,$1F,$DC,$B5,$D1,$82
	.dc.b	$1F,$C9,$3E,$6B,$FB,$73,$FD,$15,$AF,$FB,$3D,$AF,$F3,$86,$ED,$FD
	.dc.b	$57,$E0,$BF,$A2,$92,$7F,$84,$BF,$E3,$A9,$7C,$C5,$24,$FF,$B5,$23
	.dc.b	$2D,$8C,$BE,$56,$BE,$5F,$9B,$E8,$47,$93,$FD,$5B,$E6,$E4,$7F,$1A
	.dc.b	$6B,$EF,$EA,$F2,$11,$F1,$3F,$D7,$B9,$64,$7F,$22,$6B,$FC,$05,$AF
	.dc.b	$E6,$9A,$FF,$79,$6B,$FC,$17,$8F,$F6,$59,$FF,$7B,$9F,$F5,$1F,$CD
	.dc.b	$FF,$A8,$2B,$FD,$9E,$D7,$FB,$30,$70,$FF,$69,$6B,$FA,$1F,$9B,$F9
	.dc.b	$2F,$F4,$A6,$BF,$A8,$6B,$BC,$FF,$DB,$BE,$43,$E5,$1D,$58,$6D,$78
	.dc.b	$A7,$53,$92,$46,$67,$6F,$FB,$69,$AF,$FB,$99,$AF,$F8,$77,$D5,$FF
	.dc.b	$05,$91,$FE,$9E,$FF,$BC,$3E,$59,$F3,$4F,$57,$0D,$AF,$8B,$3A,$9C
	.dc.b	$92,$33,$3B,$7F,$C2,$13,$B7,$33,$FE,$52,$F9,$77,$CD,$D2,$4F,$F7
	.dc.b	$18,$CB,$63,$2F,$F7,$88,$CB,$63,$CF,$BC,$8F,$1B,$5F,$1E,$1F,$CD
	.dc.b	$70,$F7,$CE,$67,$FC,$73,$18,$7C,$5F,$07,$E7,$23,$83,$E9,$4F,$F9
	.dc.b	$03,$E0,$BE,$83,$63,$FC,$A1,$B7,$DF,$8C,$47,$F8,$EF,$5F,$E9,$95
	.dc.b	$C7,$C0,$7C,$DE,$A3,$F9,$83,$63,$35,$FE,$5B,$9E,$13,$D2,$1F,$36
	.dc.b	$F8,$B7,$C3,$7E,$72,$73,$18,$7F,$3F,$3B,$9D,$36,$33,$E6,$FF,$94
	.dc.b	$5B,$A6,$A3,$11,$FE,$4D,$94,$62,$3C,$DF,$11,$FE,$0B,$CB,$D1,$18
	.dc.b	$EB,$46,$22,$CE,$DC,$9B,$8D,$61,$FA,$9E,$5C,$C7,$E7,$DB,$3F,$43
	.dc.b	$6F,$F6,$24,$F0,$6C,$74,$3F,$CF,$DC,$8F,$9B,$C4,$7F,$5D,$8C,$57
	.dc.b	$E4,$E7,$33,$C1,$BC,$6F,$8E,$7A,$98,$6D,$FE,$A1,$B7,$FA,$46,$BF
	.dc.b	$90,$6B,$FE,$97,$AE,$95,$CC,$95,$FA,$5F,$77,$C7,$DE,$CB,$FD,$6D
	.dc.b	$39,$26,$1F,$A7,$AD,$CE,$A3,$13,$A7,$C5,$EC,$7A,$49,$1E,$96,$C7
	.dc.b	$A5,$AF,$08,$FF,$2C,$F5,$FC,$6F,$03,$FD,$54,$DC,$7F,$86,$4E,$8F
	.dc.b	$E3,$59,$7C,$CE,$0F,$E4,$8D,$DF,$E1,$E3,$FA,$78,$C1,$2B,$89,$23
	.dc.b	$FC,$F7,$F3,$7F,$EA,$78,$FF,$3D,$36,$33,$7F,$22,$31,$1F,$EA,$5F
	.dc.b	$9A,$7C,$DF,$DE,$84,$62,$FF,$54,$D8,$FE,$8E,$D7,$FC,$5B,$AF,$BC
	.dc.b	$93,$FE,$7C,$6C,$7F,$96,$DA,$F0,$8D,$11,$D3,$5F,$F6,$57,$CD,$BC
	.dc.b	$EF,$C9,$FB,$68,$C5,$7F,$5A,$8C,$4E,$A7,$35,$FE,$6B,$F9,$3F,$A8
	.dc.b	$4B,$FF,$48,$BB,$F9,$B9,$1C,$E9,$18,$AF,$F5,$EC,$3E,$3B,$B1,$2B
	.dc.b	$83,$67,$6D,$7D,$6F,$E1,$7D,$F9,$3F,$EA,$26,$CF,$6F,$9B,$F5,$12
	.dc.b	$7C,$8F,$E3,$78,$64,$7F,$AA,$9B,$3F,$E8,$96,$BF,$E6,$26,$BF,$E5
	.dc.b	$BF,$47,$FD,$C3,$3F,$E4,$8F,$17,$F5,$B1,$89,$FF,$21,$46,$09,$3B
	.dc.b	$24,$7F,$BB,$D7,$CD,$BD,$9F,$E1,$31,$86,$C6,$63,$FC,$22,$64,$FE
	.dc.b	$71,$F3,$79,$AF,$F0,$AB,$5F,$F7,$E3,$51,$1C,$36,$BF,$39,$18,$83
	.dc.b	$4D,$F4,$7F,$80,$7C,$1F,$E1,$BD,$46,$93,$0F,$F7,$34,$65,$B1,$9B
	.dc.b	$FF,$76,$CE,$5F,$F5,$D7,$AD,$FC,$0C,$95,$FE,$E0,$6C,$6A,$30,$47
	.dc.b	$E6,$65,$FE,$7B,$8F,$8E,$D8,$CF,$AD,$F3,$0F,$DD,$78,$FF,$D2,$E3
	.dc.b	$E2,$9F,$19,$C3,$86,$B9,$F1,$78,$92,$FF,$B5,$4E,$A7,$FB,$6C,$FF
	.dc.b	$1D,$F9,$B9,$5F,$EF,$A6,$CF,$FB,$65,$AF,$FA,$C9,$AF,$D8,$F0,$1A
	.dc.b	$BE,$0F,$F9,$41,$2F,$FC,$AB,$3A,$8C,$57,$FB,$95,$AF,$FA,$F9,$FE
	.dc.b	$8B,$F3,$6F,$92,$A4,$BF,$F2,$FE,$09,$D4,$7F,$81,$FE,$6F,$EC,$61
	.dc.b	$F2,$03,$87,$E5,$B5,$FC,$8B,$5D,$F0,$FF,$2B,$8C,$11,$FE,$89,$E4
	.dc.b	$4A,$F6,$E1,$FD,$AD,$B1,$9B,$F1,$67,$F1,$A3,$11,$FE,$DC,$7F,$A7
	.dc.b	$7C,$64,$AE,$6D,$9E,$36,$8C,$47,$FB,$91,$AF,$FB,$77,$E6,$DB,$7F
	.dc.b	$6E,$6D,$FC,$86,$CF,$C5,$B4,$62,$3E,$47,$07,$FB,$7A,$BF,$13,$EF
	.dc.b	$FE,$50,$8E,$09,$3F,$E9,$29,$E0,$D9,$FE,$7F,$F0,$8F,$7E,$92,$FF
	.dc.b	$A0,$46,$5B,$EF,$F0,$23,$83,$7E,$87,$42,$38,$39,$F8,$FE,$02,$38
	.dc.b	$3B,$F8,$BE,$42,$38,$26,$5F,$89,$F1,$D8,$AB,$73,$98,$C5,$7E,$7B
	.dc.b	$8B,$FC,$8A,$D3,$FC,$18,$E2,$7F,$06,$D7,$F7,$B7,$DB,$8B,$E3,$92
	.dc.b	$BE,$2B,$67,$C0,$D7,$8D,$FE,$45,$EB,$F3,$7E,$02,$57,$95,$FE,$E3
	.dc.b	$F3,$7F,$E6,$F8,$FE,$E0,$D9,$F8,$0D,$7C,$AD,$7F,$9B,$E0,$8C,$42
	.dc.b	$31,$EA,$F6,$25,$7F,$B5,$BE,$3B,$FD,$6F,$F1,$DF,$E5,$46,$C6,$7D
	.dc.b	$B4,$E5,$B1,$E6,$F9,$BF,$49,$1C,$13,$1F,$4C,$65,$B1,$F5,$7E,$33
	.dc.b	$E9,$91,$C1,$F0,$BE,$6F,$E2,$91,$C1,$27,$E2,$C6,$5B,$1F,$C7,$8F
	.dc.b	$E1,$9C,$1F,$E8,$47,$1D,$9F,$E4,$D7,$F8,$86,$AF,$9B,$FC,$15,$FC
	.dc.b	$C6,$BF,$2D,$B1,$B9,$FF,$14,$C6,$08,$E0,$ED,$F2,$71,$F2,$36,$BF
	.dc.b	$39,$3F,$EA,$09,$FF,$18,$36,$33,$EE,$76,$3F,$C6,$FF,$14,$F8,$1A
	.dc.b	$4B,$FF,$1C,$46,$5B,$19,$AF,$D5,$4E,$5B,$1F,$E6,$FE,$DF,$EF,$FA
	.dc.b	$7F,$B0,$3C,$CC,$13,$0F,$E9,$33,$A2,$46,$6F,$FD,$8C,$3E,$6F,$FC
	.dc.b	$17,$C6,$CB,$FA,$17,$B5,$FE,$B4,$8C,$57,$F4,$5A,$DD,$6B,$1B,$39
	.dc.b	$3F,$27,$19,$69,$9A,$E6,$FF,$BA,$27,$8B,$FD,$E2,$AB,$FE,$7B,$9D
	.dc.b	$BF,$E2,$08,$CD,$F9,$6F,$FA,$C4,$FA,$CE,$E7,$52,$9C,$4B,$1C,$5F
	.dc.b	$24,$79,$EB,$BC,$73,$7C,$33,$B8,$F3,$1F,$2A,$BE,$5B,$E1,$3F,$DE
	.dc.b	$4F,$FB,$CF,$E6,$BF,$CD,$1F,$37,$7D,$63,$49,$5E,$A9,$FE,$F5,$D1
	.dc.b	$C7,$CA,$34,$93,$B8,$CA,$F9,$CF,$9B,$5F,$CC,$A3,$FC,$8F,$18,$EE
	.dc.b	$FF,$12,$BA,$AE,$2D,$BD,$EB,$FE,$2A,$C3,$96,$1D,$63,$18,$72,$75
	.dc.b	$87,$26,$DF,$DC,$BE,$3A,$F8,$70,$C5,$FA,$23,$C1,$1D,$4F,$F2,$EC
	.dc.b	$F8,$31,$E1,$19,$11,$F1,$9C,$36,$36,$4C,$3C,$73,$D7,$87,$FC,$D3
	.dc.b	$F3,$BF,$E4,$6C,$9C,$1F,$E3,$5A,$4E,$63,$0D,$85,$7F,$9C,$BD,$BC
	.dc.b	$3D,$23,$FC,$EB,$7D,$1D,$9D,$0C,$DF,$46,$9A,$AE,$7C,$6F,$66,$92
	.dc.b	$77,$D2,$F9,$C7,$CE,$E6,$7F,$0F,$8F,$F9,$3F,$0E,$51,$DB,$34,$62
	.dc.b	$36,$7F,$53,$BE,$6E,$FE,$75,$E4,$7E,$BA,$3E,$0C,$61,$FF,$2B,$B5
	.dc.b	$FF,$1C,$B5,$FF,$39,$DF,$CE,$EF,$F8,$F8,$9A,$7C,$AA,$E5,$B7,$C9
	.dc.b	$A4,$3D,$7F,$14,$61,$B1,$9C,$3C,$71,$82,$38,$21,$FE,$4E,$C1,$F6
	.dc.b	$FE,$42,$47,$C1,$9F,$C0,$79,$FF,$E4,$BF,$07,$E1,$36,$7F,$CD,$BE
	.dc.b	$8A,$72,$D8,$CC,$60,$8E,$0E,$3E,$1F,$69,$FF,$59,$47,$E1,$2B,$FC
	.dc.b	$B5,$39,$AF,$C9,$68,$F9,$CF,$F0,$0D,$FF,$81,$DA,$FF,$9E,$E3,$5D
	.dc.b	$5F,$65,$B1,$B8,$C4,$F0,$8F,$D2,$35,$FC,$DB,$5F,$F4,$CF,$DE,$3D
	.dc.b	$EF,$8A,$D8,$F4,$35,$FC,$6D,$F8,$1F,$C7,$F0,$7F,$A8,$9F,$B7,$9F
	.dc.b	$F5,$48,$C4,$7F,$A1,$8C,$90,$C1,$0F,$F2,$2F,$8B,$FB,$28,$34,$FD
	.dc.b	$AC,$FC,$3E,$0F,$F5,$92,$F0,$FC,$24,$BF,$F4,$AB,$5F,$F4,$5B,$5F
	.dc.b	$CD,$C7,$17,$FC,$B5,$E5,$78,$48,$FE,$BE,$34,$FE,$BB,$A5,$F3,$8F
	.dc.b	$43,$F5,$91,$86,$C6,$67,$FA,$EC,$E5,$B1,$97,$FC,$E0,$D7,$FC,$CA
	.dc.b	$D7,$FC,$A2,$D7,$F8,$CE,$6F,$4F,$E4,$12,$3F,$86,$CF,$5D,$7F,$8D
	.dc.b	$5F,$A3,$7F,$E4,$4D,$AD,$C6,$2B,$F8,$65,$7F,$21,$1F,$82,$3D,$1F
	.dc.b	$C9,$36,$FF,$CB,$CD,$7F,$30,$D7,$F8,$4E,$4F,$AF,$FB,$C4,$8F,$3F
	.dc.b	$B5,$FB,$6F,$9D,$FC,$0C,$7F,$B8,$1A,$FF,$AE,$5A,$89,$BA,$33,$18
	.dc.b	$6B,$FE,$F8,$6C,$6F,$8F,$F8,$1C,$6B,$67,$EB,$7F,$A1,$D7,$E4,$B7
	.dc.b	$93,$FD,$06,$DC,$7A,$61,$F9,$9F,$AD,$91,$FC,$DF,$54,$7E,$6B,$DF
	.dc.b	$CC,$FE,$AA,$72,$EE,$74,$F1,$FA,$CF,$D2,$11,$FC,$A7,$D0,$57,$94
	.dc.b	$5C,$FF,$98,$6C,$AF,$E4,$3F,$3E,$FA,$0A,$4C,$3F,$AC,$8F,$83,$FF
	.dc.b	$03,$E1,$FD,$87,$05,$F1,$AF,$E3,$8D,$7F,$A8,$C6,$AF,$FB,$74,$73
	.dc.b	$F4,$7F,$79,$F5,$52,$3E,$23,$E1,$7E,$17,$B5,$FC,$04,$8F,$EC,$7D
	.dc.b	$7A,$E8,$F9,$91,$FC,$12,$3F,$84,$78,$92,$7E,$83,$FC,$23,$E6,$71
	.dc.b	$1F,$3E,$2E,$9F,$1A,$7C,$FF,$41,$FA,$DA,$34,$9E,$07,$F9,$C1,$B1
	.dc.b	$FD,$AD,$AF,$FB,$7B,$C6,$FE,$34,$47,$E6,$C7,$EC,$FA,$3F,$31,$CB
	.dc.b	$FE,$12,$F2,$FF,$90,$12,$BF,$88,$7A,$75,$FC,$34,$5C,$FE,$EC,$FF
	.dc.b	$27,$F2,$E5,$7E,$3D,$FE,$BF,$F3,$2F,$89,$CC,$FE,$3E,$6D,$FD,$DA
	.dc.b	$D4,$F9,$AB,$DC,$47,$DC,$D7,$FC,$EF,$9F,$06,$C6,$5F,$DF,$FD,$EF
	.dc.b	$F3,$B2,$3D,$4F,$D2,$7C,$E7,$F3,$F2,$3F,$D9,$FE,$39,$DD,$88,$FD
	.dc.b	$34,$DC,$BF,$29,$5A,$9F,$D7,$7D,$07,$F4,$92,$3F,$CE,$7E,$22,$BF
	.dc.b	$AF,$8B,$5F,$A6,$9F,$E9,$BF,$11,$FD,$68,$95,$FE,$94,$E1,$76,$27
	.dc.b	$F9,$FF,$D0,$B3,$FE,$56,$DB,$FF,$1E,$B5,$FE,$CF,$5C,$5F,$E0,$9F
	.dc.b	$3B,$FD,$7C,$8F,$0A,$AF,$F0,$58,$B9,$7F,$3A,$BE,$8F,$EF,$BE,$87
	.dc.b	$FB,$89,$2B,$D1,$82,$3F,$C5,$F9,$7E,$F6,$7F,$8A,$78,$3F,$DB,$CF
	.dc.b	$ED,$B3,$FD,$BB,$E4,$BF,$BD,$92,$79,$BC,$BE,$B3,$FB,$B1,$23,$FB
	.dc.b	$34,$57,$F8,$E4,$5C,$7F,$89,$CD,$E9,$FD,$0E,$2A,$C5,$FF,$9D,$32
	.dc.b	$5F,$F8,$E3,$E8,$7E,$54,$FF,$3F,$3E,$9B,$87,$FA,$0B,$5C,$ED,$FE
	.dc.b	$4E,$FC,$F3,$FC,$CF,$1F,$E6,$96,$BF,$E1,$A6,$E7,$FC,$7E,$BE,$6E
	.dc.b	$7F,$E0,$EF,$73,$FC,$3A,$4B,$F9,$AD,$7C,$87,$A5,$C3,$FE,$0E,$7C
	.dc.b	$5E,$CF,$FA,$3B,$0F,$F1,$EE,$1F,$E6,$56,$DF,$7B,$FF,$24,$7C,$EB
	.dc.b	$E8,$BF,$CA,$ED,$8F,$F7,$3E,$A8,$C4,$78,$A3,$13,$C0,$91,$C2,$76
	.dc.b	$FC,$B9,$DB,$F9,$FB,$7C,$5F,$9C,$91,$E4,$BA,$FC,$98,$B8,$FF,$47
	.dc.b	$C3,$C6,$D8,$F4,$92,$7C,$4D,$7D,$13,$E0,$72,$43,$E6,$F0,$30,$EB
	.dc.b	$97,$F2,$A6,$CF,$FA,$F1,$B9,$7C,$5C,$3F,$D8,$1E,$D3,$E7,$BF,$1B
	.dc.b	$7F,$E0,$9F,$8C,$F9,$99,$75,$F8,$6F,$C7,$46,$32,$46,$36,$FF,$08
	.dc.b	$63,$CF,$8F,$18,$FF,$B4,$BC,$17,$4E,$33,$F4,$7C,$3D,$CF,$FD,$95
	.dc.b	$6D,$3E,$E7,$03,$1E,$33,$E5,$D3,$F4,$1F,$44,$F7,$FF,$33,$38,$6B
	.dc.b	$FE,$23,$8D,$7C,$37,$FA,$FD,$B1,$F9,$60,$FF,$9B,$B7,$D7,$AD,$FE
	.dc.b	$5E,$9E,$ED,$7A,$37,$27,$9B,$FE,$07,$7C,$D3,$F4,$7E,$73,$FD,$D1
	.dc.b	$AF,$B2,$B5,$3E,$FD,$3E,$C5,$69,$F9,$7C,$9F,$ED,$6C,$39,$F0,$FD
	.dc.b	$FF,$4B,$58,$7F,$9F,$EB,$4F,$33,$F0,$11,$A9,$FC,$37,$C6,$7F,$96
	.dc.b	$6F,$F0,$CD,$AF,$E7,$7D,$2F,$A2,$C5,$F0,$BF,$D8,$C7,$FC,$01,$AF
	.dc.b	$D3,$56,$E7,$F5,$53,$99,$FF,$28,$CE,$5F,$CD,$C6,$5E,$ED,$EA,$FD
	.dc.b	$C3,$67,$84,$FF,$BB,$B9,$5F,$1F,$E4,$6E,$1F,$C1,$3E,$87,$13,$23
	.dc.b	$F8,$E3,$ED,$C7,$F8,$24,$E6,$FF,$C6,$B8,$23,$C0,$4B,$FE,$08,$D9
	.dc.b	$FD,$9B,$5F,$82,$FA,$F2,$FD,$97,$D9,$F9,$4F,$93,$87,$FB,$F6,$FF
	.dc.b	$97,$46,$3E,$2B,$84,$E7,$0F,$DD,$B5,$FF,$59,$C6,$89,$5F,$D6,$9B
	.dc.b	$1F,$C0,$B3,$7C,$F7,$C5,$3E,$07,$D1,$7F,$B0,$EF,$F9,$F4,$63,$9F
	.dc.b	$E8,$4E,$6F,$F9,$33,$5F,$E4,$4D,$7F,$86,$7D,$8F,$D1,$3D,$BE,$BD
	.dc.b	$C3,$FA,$54,$59,$FE,$FB,$37,$C3,$FE,$79,$36,$BF,$DA,$F3,$6B,$FA
	.dc.b	$F3,$F1,$09,$1F,$DE,$1F,$9D,$93,$EC,$7F,$04,$7C,$EF,$9B,$B5,$FE
	.dc.b	$D5,$7E,$69,$1F,$F0,$0B,$CB,$87,$F1,$47,$E8,$7E,$11,$F4,$5F,$8C
	.dc.b	$CB,$FB,$14,$EA,$FF,$B3,$3B,$9D,$69,$FD,$5E,$31,$CB,$2B,$FD,$CD
	.dc.b	$F2,$97,$0F,$F0,$FB,$F8,$02,$47,$F8,$51,$F6,$F3,$FE,$74,$FC,$4F
	.dc.b	$C6,$D9,$FE,$ED,$16,$BF,$C5,$33,$71,$FE,$D1,$F0,$5F,$A5,$3F,$41
	.dc.b	$EB,$7F,$82,$30,$FE,$FD,$19,$69,$FE,$4D,$F2,$71,$87,$F8,$07,$EB
	.dc.b	$FF,$5C,$7F,$05,$97,$D1,$3E,$0B,$31,$FE,$CF,$C8,$FF,$B4,$4D,$C7
	.dc.b	$EF,$DA,$7F,$2F,$2D,$FF,$81,$9A,$FF,$86,$DA,$F3,$AF,$04,$71,$6C
	.dc.b	$71,$6B,$FC,$1F,$E5,$BF,$83,$1F,$C0,$7A,$FF,$9A,$D7,$F8,$7E,$32
	.dc.b	$D3,$D4,$7F,$BA,$23,$73,$FE,$23,$9F,$E2,$1F,$55,$2B,$73,$FE,$35
	.dc.b	$F7,$B0,$FF,$8A,$66,$E5,$FE,$D0,$9B,$5F,$DD,$7E,$9B,$F8,$D1,$F3
	.dc.b	$72,$FF,$3E,$C7,$F8,$CE,$31,$EF,$7C,$99,$CE,$3E,$F8,$C1,$2B,$E2
	.dc.b	$36,$35,$1F,$E7,$7F,$53,$EA,$43,$1F,$AE,$BC,$34,$EA,$7F,$DC,$FF
	.dc.b	$45,$FD,$0A,$31,$7F,$EB,$66,$DF,$16,$DF,$FA,$12,$B8,$4F,$1F,$06
	.dc.b	$5F,$C8,$AF,$F4,$4F,$C7,$E2,$BC,$93,$D6,$B1,$8F,$FB,$E2,$AC,$F8
	.dc.b	$72,$F2,$1F,$CF,$7E,$91,$E4,$62,$7E,$AF,$8C,$9C,$C6,$1B,$97,$80
	.dc.b	$0F,$A4,$FC,$0E,$5F,$EF,$F9,$DB,$70,$F0,$1F,$DA,$6B,$87,$9D,$8C
	.dc.b	$3D,$DB,$BE,$37,$E5,$1F,$D1,$3E,$52,$CE,$A2,$DF,$FA,$92,$6E,$3F
	.dc.b	$5B,$83,$EA,$F4,$7D,$33,$FA,$6F,$CC,$DC,$7B,$C5,$C3,$FD,$5B,$36
	.dc.b	$FE,$FF,$88,$FD,$71,$E6,$FF,$21,$6E,$3F,$23,$E9,$5C,$F9,$AF,$ED
	.dc.b	$51,$89,$FF,$37,$B7,$2F,$F3,$73,$5F,$F3,$53,$5F,$9B,$C8,$8C,$47
	.dc.b	$FB,$29,$B1,$9B,$FF,$02,$46,$B5,$FC,$AF,$C7,$7E,$4C,$FE,$71,$E8
	.dc.b	$BE,$1F,$19,$7E,$D3,$67,$D8,$FF,$04,$D3,$F8,$B3,$5F,$D1,$36,$37
	.dc.b	$C3,$FE,$15,$E0,$63,$FE,$DD,$C3,$F7,$F1,$8F,$88,$F0,$FC,$17,$89
	.dc.b	$FD,$37,$C6,$CA,$DD,$BD,$2C,$69,$F8,$FC,$7A,$C7,$79,$D5,$7F,$94
	.dc.b	$E7,$2D,$8C,$D7,$FA,$B2,$72,$D8,$CB,$FE,$C3,$8C,$B6,$32,$FE,$E7
	.dc.b	$5F,$DE,$1F,$E3,$ED,$FF,$9C,$DF,$E6,$23,$0F,$F8,$76,$33,$3F,$E9
	.dc.b	$06,$BF,$ED,$17,$F8,$54,$6C,$87,$F9,$B5,$CB,$5E,$31,$FC,$2F,$2F
	.dc.b	$E1,$C7,$8F,$E9,$AD,$7F,$07,$A4,$78,$C9,$1C,$DB,$5E,$56,$BE,$59
	.dc.b	$E3,$3F,$EC,$6F,$C3,$9F,$0F,$FE,$45,$6C,$7E,$27,$E9,$BF,$91,$BF
	.dc.b	$C2,$A3,$37,$FC,$1F,$85,$5A,$8F,$F6,$FC,$63,$B3,$FC,$BC,$DB,$FF
	.dc.b	$47,$36,$77,$5C,$5F,$F7,$D4,$65,$B1,$97,$FB,$73,$5F,$E6,$1F,$15
	.dc.b	$FC,$D4,$FB,$5F,$4D,$6F,$E4,$3F,$64,$8F,$D1,$6C,$FD,$B6,$BF,$1D
	.dc.b	$FF,$2E,$7C,$03,$D7,$FF,$41,$B6,$BF,$9D,$1F,$C9,$8F,$E5,$3A,$FF
	.dc.b	$AE,$67,$FD,$13,$1F,$DA,$AF,$F0,$CD,$76,$D7,$F1,$3E,$0F,$F9,$AE
	.dc.b	$31,$5F,$E0,$C8,$C1,$23,$F1,$96,$F5,$3C,$A7,$83,$E4,$EC,$F9,$9E
	.dc.b	$84,$7F,$28,$1F,$F3,$D7,$D0,$BA,$3F,$D4,$2D,$9F,$E0,$30,$F8,$BC
	.dc.b	$B2,$3F,$AE,$4E,$F3,$FD,$F4,$EE,$72,$FF,$AB,$3E,$1B,$FA,$24,$7F
	.dc.b	$63,$6C,$FF,$53,$D1,$1F,$E1,$EE,$1F,$E8,$4E,$9B,$E3,$73,$3A,$76
	.dc.b	$DC,$3F,$C2,$B8,$7F,$BE,$07,$D1,$FB,$5F,$EB,$E7,$FA,$05,$C3,$B6
	.dc.b	$B9,$CB,$FD,$CD,$BF,$F6,$69,$CF,$9B,$FE,$0A,$77,$5F,$EF,$93,$BD
	.dc.b	$76,$79,$16,$BF,$B7,$F3,$72,$C2,$D7,$6A,$FF,$09,$BC,$70,$FF,$0B
	.dc.b	$D7,$6B,$F1,$4E,$6E,$DF,$53,$C2,$7B,$AF,$CD,$3D,$70,$F0,$4E,$8F
	.dc.b	$81,$1C,$DC,$1F,$E0,$4B,$F5,$3D,$4A,$7B,$CA,$E3,$E9,$3F,$DA,$FF
	.dc.b	$4B,$7D,$C3,$49,$5D,$27,$FC,$33,$B7,$1F,$33,$FC,$6D,$39,$6B,$F2
	.dc.b	$E7,$6E,$67,$E4,$BF,$E1,$A8,$FE,$B2,$E0,$38,$7E,$69,$B9,$EF,$AF
	.dc.b	$E0,$A7,$AB,$7C,$0F,$C3,$B6,$BA,$4E,$B3,$FC,$5B,$63,$C5,$5C,$27
	.dc.b	$2D,$8F,$06,$1C,$B0,$F0,$D6,$A7,$AD,$6D,$F4,$CE,$9E,$27,$A8,$F6
	.dc.b	$C6,$EF,$DD,$87,$AE,$B8,$71,$7F,$8F,$BE,$1F,$EC,$B7,$0F,$EA,$C2
	.dc.b	$32,$D3,$2E,$08,$E0,$FE,$F1,$F8,$A3,$83,$F0,$BE,$1A,$31,$3A,$9C
	.dc.b	$C6,$23,$FB,$E3,$5F,$EF,$4D,$7F,$CA,$B3,$CE,$78,$9D,$BE,$C1,$EB
	.dc.b	$52,$57,$F8,$2E,$32,$FF,$73,$E8,$7B,$3F,$09,$B1,$98,$C1,$EA,$FC
	.dc.b	$0B,$C5,$0C,$51,$86,$E3,$B2,$4F,$F8,$6E,$74,$48,$FA,$3C,$AF,$7E
	.dc.b	$92,$BF,$9E,$BF,$2B,$C8,$7C,$0D,$24,$FF,$89,$E3,$2D,$8C,$BF,$E2
	.dc.b	$78,$CB,$63,$F7,$3C,$CF,$A7,$FF,$50,$CF,$F7,$66,$BF,$DF,$1A,$FF
	.dc.b	$76,$6B,$FE,$E1,$FC,$2D,$F2,$5D,$1B,$1F,$E0,$06,$BF,$E0,$66,$BF
	.dc.b	$E0,$06,$BC,$E3,$91,$1E,$5E,$57,$52,$38,$3F,$C9,$1F,$4D,$FE,$F5
	.dc.b	$6F,$81,$F9,$28,$C4,$7F,$92,$5A,$FF,$91,$DA,$FE,$39,$AF,$DF,$6F
	.dc.b	$01,$CF,$3B,$D4,$FC,$2B,$78,$7F,$5B,$18,$3F,$0D,$E7,$3E,$3B,$F4
	.dc.b	$AD,$C7,$FC,$C0,$D7,$FC,$BC,$D7,$FC,$B8,$D7,$F4,$6D,$7F,$28,$DC
	.dc.b	$4E,$70,$FF,$6E,$FC,$1F,$F0,$0E,$D3,$49,$7F,$2B,$93,$E4,$11,$C1
	.dc.b	$0F,$94,$E5,$A7,$F8,$FB,$E9,$9E,$DF,$E3,$9B,$19,$8C,$1F,$0B,$BA
	.dc.b	$47,$C2,$6B,$F0,$BC,$1F,$8E,$47,$E3,$BF,$2B,$C0,$91,$F2,$9A,$FF
	.dc.b	$58,$FB,$0F,$7B,$FD,$61,$AE,$72,$FE,$B6,$A3,$2F,$F7,$96,$BF,$E0
	.dc.b	$86,$BF,$EE,$86,$BC,$63,$9F,$CE,$FF,$A8,$63,$FB,$3B,$5F,$ED,$6D
	.dc.b	$7F,$C0,$CD,$7F,$C9,$7A,$3D,$6F,$B0,$48,$FF,$46,$46,$5B,$19,$7F
	.dc.b	$C2,$1B,$27,$51,$86,$BF,$E9,$9E,$2F,$F0,$CA,$9F,$CE,$78,$BF,$9A
	.dc.b	$24,$7F,$A6,$E7,$73,$A8,$C4,$E9,$FF,$41,$FC,$CB,$E8,$FF,$87,$B6
	.dc.b	$3F,$D0,$6D,$77,$1A,$23,$A6,$BF,$E2,$BB,$7E,$73,$FE,$53,$8F,$F0
	.dc.b	$AC,$63,$CF,$FF,$0A,$B6,$33,$3B,$79,$7C,$C7,$F4,$46,$F9,$1C,$23
	.dc.b	$11,$FE,$46,$A4,$62,$3F,$DF,$6D,$7F,$81,$36,$37,$3C,$CF,$D8,$78
	.dc.b	$D2,$BE,$CB,$5F,$A8,$D7,$FD,$99,$AF,$F5,$47,$27,$F8,$C3,$EA,$5F
	.dc.b	$37,$F9,$86,$C6,$63,$04,$70,$43,$FC,$77,$1C,$1F,$C9,$DF,$E6,$08
	.dc.b	$FE,$61,$FE,$11,$EE,$48,$FC,$8B,$5F,$C6,$35,$F4,$B5,$1F,$C9,$FE
	.dc.b	$39,$F0,$FF,$59,$B1,$F6,$9A,$FD,$F6,$BF,$4D,$AF,$4C,$D1,$88,$ED
	.dc.b	$3D,$4F,$DE,$F9,$CF,$A4,$A4,$9F,$F2,$DC,$65,$B1,$97,$FC,$C3,$19
	.dc.b	$6C,$7E,$3B,$EE,$5E,$E5,$24,$FE,$52,$32,$FE,$A7,$CD,$FE,$8A,$47
	.dc.b	$04,$3F,$73,$92,$35,$18,$7F,$9D,$7B,$92,$3F,$81,$B5,$FD,$F3,$5F
	.dc.b	$C5,$B5,$FF,$15,$FE,$A6,$F9,$1F,$F1,$BB,$63,$FC,$FF,$E3,$23,$11
	.dc.b	$FE,$9B,$8C,$4F,$F4,$BF,$76,$4F,$F1,$46,$BF,$D6,$A3,$FA,$F9,$1F
	.dc.b	$EB,$ED,$74,$60,$87,$7F,$C5,$3F,$9A,$D1,$A4,$8F,$EB,$9E,$3C,$7F
	.dc.b	$92,$B5,$E5,$C1,$FE,$54,$9F,$C3,$57,$38,$C1,$FD,$97,$8E,$57,$F2
	.dc.b	$E6,$BF,$CA,$5A,$FE,$D9,$AF,$F3,$86,$BF,$CA,$F9,$BF,$83,$77,$7F
	.dc.b	$33,$9F,$E9,$FF,$69,$FE,$6D,$AF,$F5,$6B,$5F,$F6,$43,$5F,$F6,$BB
	.dc.b	$5F,$F3,$77,$E0,$5F,$E9,$B7,$FD,$7C,$D7,$F5,$FE,$62,$31,$1C,$3C
	.dc.b	$7F,$F5,$CF,$DA,$BE,$13,$F8,$74,$FF,$B9,$23,$1D,$DA,$9C,$CF,$EE
	.dc.b	$A7,$2F,$FB,$11,$AF,$FB,$7D,$AF,$38,$E2,$A3,$2F,$EE,$7E,$77,$FC
	.dc.b	$35,$CE,$69,$3C,$FF,$17,$93,$2B,$C1,$87,$F7,$2C,$3D,$53,$99,$F1
	.dc.b	$4E,$49,$1F,$D1,$2F,$FA,$14,$FF,$3D,$6B,$E3,$F1,$34,$47,$FA,$6C
	.dc.b	$69,$FF,$20,$37,$1F,$44,$FC,$CF,$9E,$FF,$43,$E2,$69,$36,$D1,$30
	.dc.b	$F9,$51,$97,$FA,$DE,$CF,$86,$FE,$B6,$D8,$CC,$60,$FF,$79,$FC,$17
	.dc.b	$F5,$6F,$5B,$F9,$7C,$A2,$99,$7F,$99,$B8,$F1,$FF,$CE,$AD,$9D,$CE
	.dc.b	$A5,$F5,$98,$8E,$91,$87,$99,$B8,$D5,$7F,$68,$78,$43,$E8,$5F,$44
	.dc.b	$F8,$7F,$3C,$FA,$27,$31,$88,$FE,$93,$18,$70,$D7,$8A,$FA,$2F,$F5
	.dc.b	$BB,$76,$E5,$3E,$19,$C9,$FE,$CD,$C5,$F7,$6F,$A4,$F9,$15,$0F,$95
	.dc.b	$19,$9F,$7C,$62,$77,$1F,$3A,$75,$3C,$61,$F1,$8F,$A6,$DC,$FD,$28
	.dc.b	$C4,$F4,$37,$3A,$9F,$B5,$3A,$7C,$9C,$BF,$BD,$F6,$BC,$D1,$FD,$8F
	.dc.b	$8F,$FC,$E3,$87,$E2,$A7,$33,$FE,$97,$9C,$BF,$43,$6F,$DA,$7D,$4E
	.dc.b	$9B,$7F,$45,$B3,$BE,$65,$6D,$B5,$B6,$CE,$E7,$47,$8B,$1F,$DF,$7D
	.dc.b	$05,$E9,$34,$98,$7E,$C1,$B7,$FE,$41,$9D,$CE,$4F,$E2,$59,$FF,$BA
	.dc.b	$DB,$E3,$7C,$0C,$BF,$D2,$6D,$9D,$5F,$F1,$76,$BF,$C8,$5A,$FF,$1F
	.dc.b	$6B,$FC,$65,$B8,$FF,$17,$F8,$5F,$51,$30,$FE,$0E,$F7,$F4,$A4,$79
	.dc.b	$9B,$7E,$16,$BE,$16,$BF,$E1,$DF,$A5,$FF,$01,$4F,$F9,$79,$B1,$FE
	.dc.b	$63,$6B,$FE,$57,$6B,$FC,$0B,$E3,$FF,$C5,$F3,$FA,$26,$C7,$EB,$5A
	.dc.b	$FE,$C9,$AF,$FA,$4B,$E5,$BF,$13,$E3,$FF,$A2,$DB,$E4,$1A,$4B,$FF
	.dc.b	$4A,$B6,$BF,$D2,$0D,$7F,$CB,$AD,$7B,$47,$F8,$43,$CB,$7C,$27,$E6
	.dc.b	$3C,$CF,$F0,$03,$63,$33,$B8,$F1,$FC,$E4,$8F,$F5,$2B,$6F,$FC,$15
	.dc.b	$F0,$9C,$49,$3D,$DE,$DF,$AA,$FE,$F2,$7F,$CA,$F8,$A3,$10,$8C,$3C
	.dc.b	$09,$5F,$E2,$89,$DC,$EA,$31,$3F,$23,$8A,$47,$FA,$E9,$B1,$FE,$1F
	.dc.b	$E0,$F7,$12,$7E,$23,$F4,$3E,$6E,$47,$FB,$09,$B3,$FE,$BD,$6B,$FE
	.dc.b	$8A,$6B,$FE,$72,$DD,$C7,$F9,$C6,$F2,$7D,$E8,$C1,$1C,$12,$FF,$D8
	.dc.b	$4F,$EE,$FD,$67,$B7,$49,$3F,$E6,$17,$F8,$8E,$0F,$B6,$FE,$22,$FF
	.dc.b	$40,$6B,$FD,$61,$AF,$FA,$21,$AF,$1D,$39,$C6,$08,$FF,$0D,$F7,$1E
	.dc.b	$6D,$26,$1F,$E8,$D6,$DF,$FA,$35,$AF,$FA,$35,$FC,$FF,$E8,$1F,$DA
	.dc.b	$67,$FA,$96,$A8,$C5,$7F,$BB,$9A,$FF,$70,$8D,$4F,$22,$1C,$A7,$73
	.dc.b	$F4,$DA,$FF,$77,$FA,$0F,$CE,$BF,$D9,$25,$18,$8F,$EF,$6D,$7F,$BD
	.dc.b	$B5,$F9,$1C,$BF,$18,$93,$FD,$8B,$3D,$4F,$D3,$9C,$BF,$DB,$E3,$27
	.dc.b	$3C,$3E,$07,$CC,$FC,$42,$5F,$AD,$B5,$F0,$DA,$FC,$3C,$51,$FD,$07
	.dc.b	$D3,$FE,$C7,$5F,$EF,$86,$D7,$FB,$F5,$AF,$FB,$C5,$AF,$FA,$FD,$AF
	.dc.b	$E6,$9B,$C2,$69,$8F,$81,$7A,$BF,$D4,$5B,$7F,$E0,$DA,$46,$2F,$FC
	.dc.b	$27,$18,$AF,$ED,$3F,$1E,$FE,$00,$7D,$2E,$1B,$8F,$F8,$5E,$75,$39
	.dc.b	$24,$66,$77,$87,$F8,$82,$B7,$39,$AF,$D9,$7B,$52,$3F,$23,$8A,$31
	.dc.b	$7F,$EC,$66,$BF,$91,$8D,$4F,$09,$FC,$DF,$C0,$C8,$FC,$93,$6B,$F3
	.dc.b	$4D,$7E,$C3,$5F,$C6,$FB,$9F,$E0,$A9,$FC,$9B,$78,$FF,$A3,$CF,$E6
	.dc.b	$63,$F7,$FE,$57,$F8,$5E,$3F,$C1,$DE,$6E,$09,$7F,$E3,$49,$D4,$FF
	.dc.b	$96,$1A,$8F,$F2,$34,$F0,$7F,$DA,$5D,$EF,$03,$F9,$C9,$2F,$F9,$93
	.dc.b	$FC,$F3,$D7,$91,$FC,$D5,$B5,$CB,$A7,$FA,$09,$27,$F9,$D3,$F8,$0F
	.dc.b	$9C,$FD,$67,$09,$A4,$EA,$FE,$AD,$C9,$E3,$BF,$F7,$4B,$70,$FF,$74
	.dc.b	$35,$FF,$3A,$F8,$5F,$94,$CF,$89,$2B,$8B,$67,$F8,$77,$A5,$FC,$50
	.dc.b	$93,$FC,$2D,$FE,$B3,$E1,$C8,$FE,$A4,$D9,$F8,$ED,$7C,$3B,$23,$11
	.dc.b	$FD,$17,$B3,$F8,$98,$29,$2B,$DF,$DF,$1F,$7B,$67,$DE,$D7,$DF,$9F
	.dc.b	$F3,$CC,$FF,$67,$3F,$AC,$F6,$FF,$61,$5F,$E9,$8D,$FF,$DA,$97,$92
	.dc.b	$61,$FA,$DA,$FF,$36,$D6,$DA,$FE,$1E,$BF,$DB,$9E,$9F,$52,$57,$16
	.dc.b	$CF,$EC,$6B,$8A,$AF,$F7,$27,$93,$E5,$25,$F8,$5F,$F0,$A7,$DC,$3E
	.dc.b	$C6,$92,$7F,$D3,$1A,$27,$51,$86,$E1,$FE,$A0,$9D,$57,$18,$C4,$E9
	.dc.b	$FE,$99,$F7,$BF,$C8,$36,$34,$9E,$AF,$F8,$9F,$F0,$CF,$E9,$78,$9A
	.dc.b	$4C,$B4,$43,$CE,$E5,$FE,$BE,$77,$24,$FF,$90,$9F,$E6,$32,$FB,$AE
	.dc.b	$E4,$8F,$F1,$C3,$E2,$FC,$2D,$EF,$FF,$8E,$DB,$3F,$19,$B9,$FC,$36
	.dc.b	$BF,$A6,$6B,$FE,$34,$6B,$FE,$33,$6D,$9C,$E7,$FE,$CD,$EE,$FF,$2C
	.dc.b	$F7,$7F,$16,$8C,$61,$FD,$F2,$30,$49,$D9,$23,$FD,$93,$87,$F1,$3F
	.dc.b	$99,$FE,$29,$C2,$F8,$FF,$F6,$CF,$BB,$FE,$C8,$8C,$57,$FB,$72,$B7
	.dc.b	$5A,$BB,$39,$15,$C1,$FE,$41,$7C,$EB,$FD,$C5,$7C,$27,$37,$5F,$21
	.dc.b	$CA,$7B,$34,$ED,$7D,$55,$F5,$73,$3C,$27,$50,$BE,$D5,$B8,$FE,$68
	.dc.b	$64,$EC,$79,$0A,$79,$27,$C1,$1C,$63,$CE,$7F,$BD,$27,$E6,$BC,$CC
	.dc.b	$1F,$F0,$07,$E7,$5F,$3E,$F9,$FF,$F7,$DC,$61,$B1,$99,$E0,$D7,$CE
	.dc.b	$25,$F2,$DF,$BF,$8E,$CD,$77,$1A,$E8,$EF,$39,$9F,$83,$39,$7E,$0B
	.dc.b	$FE,$0E,$7E,$B3,$5C,$C7,$7C,$BE,$F4,$F1,$6F,$5F,$CC,$6D,$72,$AE
	.dc.b	$1A,$78,$1B,$1F,$D9,$6F,$B5,$70,$8E,$F8,$71,$3C,$78,$75,$C3,$D2
	.dc.b	$D9,$E5,$87,$47,$E3,$37,$0E,$B5,$C2,$7A,$1F,$8D,$E0,$FE,$ED,$F2
	.dc.b	$FF,$3A,$3F,$D1,$C1,$C3,$FD,$2C,$C1,$FE,$6C,$29,$F5,$A3,$9B,$98
	.dc.b	$E3,$18,$8D,$1C,$0F,$ED,$7F,$23,$F9,$C7,$F9,$8B,$5F,$E5,$CD,$77
	.dc.b	$BF,$93,$91,$39,$9F,$F6,$0B,$5F,$F6,$03,$63,$73,$C7,$E7,$FF,$C6
	.dc.b	$95,$FD,$01,$AF,$F3,$D6,$BF,$E0,$06,$BF,$EB,$3F,$9F,$FF,$2C,$3F
	.dc.b	$E8,$E6,$BF,$0D,$AF,$FA,$35,$AF,$FB,$5B,$E7,$FC,$84,$AF,$63,$63
	.dc.b	$FA,$9B,$5F,$EA,$0F,$B7,$1F,$DF,$7B,$3F,$3B,$90,$D2,$57,$F5,$CF
	.dc.b	$93,$7C,$0F,$C9,$6C,$66,$30,$47,$04,$3F,$B0,$47,$07,$FC,$C5,$F3
	.dc.b	$FF,$C8,$9F,$F3,$53,$5C,$F2,$FE,$2B,$95,$39,$9F,$F0,$47,$CF,$FF
	.dc.b	$4E,$7F,$CE,$4D,$7F,$CD,$CD,$77,$DB,$CE,$31,$1F,$E0,$96,$BF,$E0
	.dc.b	$36,$D1,$A3,$E7,$F9,$12,$7C,$2F,$E3,$7E,$7D,$E6,$D2,$4F,$F8,$1A
	.dc.b	$32,$D8,$CB,$FE,$05,$8C,$B6,$3F,$47,$EA,$FD,$92,$71,$60,$FC,$E7
	.dc.b	$9C,$F9,$2A,$4B,$FF,$0A,$C6,$5B,$19,$7F,$C2,$91,$96,$C7,$EC,$B6
	.dc.b	$FF,$76,$C7,$FB,$91,$AF,$EE,$5A,$FF,$B8,$DA,$FF,$9E,$F9,$D9,$7E
	.dc.b	$3A,$3E,$17,$17,$C7,$23,$82,$1F,$0D,$CB,$FE,$A9,$F9,$F7,$9B,$F8
	.dc.b	$56,$C7,$FA,$CD,$AE,$E3,$E2,$F5,$C8,$F8,$6D,$7E,$0F,$7F,$CA,$23
	.dc.b	$F2,$1F,$9B,$DF,$23,$E5,$B5,$FE,$45,$F3,$FF,$01,$FF,$4D,$35,$34
	.dc.b	$9E,$9F,$F8,$39,$B1,$FC,$A1,$E1,$F3,$FF,$E7,$78,$FE,$56,$D7,$F9
	.dc.b	$83,$5F,$EB,$ED,$7F,$D5,$5D,$9F,$E0,$EF,$9F,$FB,$44,$8F,$B4,$D7
	.dc.b	$ED,$35,$FB,$4F,$E4,$3C,$29,$1F,$ED,$46,$BF,$CE,$79,$7F,$C1,$2F
	.dc.b	$FA,$25,$B4,$69,$AE,$97,$CF,$BE,$7F,$F7,$CD,$BF,$F7,$63,$5F,$F7
	.dc.b	$43,$5F,$F4,$43,$5F,$F4,$13,$5F,$F3,$E3,$67,$BC,$FF,$54,$AF,$F8
	.dc.b	$03,$F3,$17,$E0,$0D,$26,$1F,$D7,$65,$F2,$FF,$CE,$A3,$D8,$D8,$CF
	.dc.b	$BF,$D6,$30,$47,$04,$9F,$D5,$7C,$17,$EB,$48,$E0,$87,$EB,$5C,$B4
	.dc.b	$FE,$DB,$F0,$EF,$9F,$FF,$2B,$B6,$33,$18,$23,$82,$1F,$DE,$63,$83
	.dc.b	$F8,$CA,$FF,$30,$F0,$FB,$E0,$D3,$E1,$6F,$FE,$5A,$F6,$FF,$83,$12
	.dc.b	$7F,$83,$35,$FE,$0C,$D7,$F8,$33,$F4,$FE,$21,$E9,$52,$4F,$D1,$8C
	.dc.b	$B6,$32,$FF,$8F,$B5,$4E,$A3,$0D,$7F,$C8,$7E,$73,$E7,$FF,$CF,$AD
	.dc.b	$8F,$F0,$BB,$5F,$F0,$E3,$5E,$31,$C0,$8F,$07,$F1,$9F,$14,$F8,$5A
	.dc.b	$49,$FC,$54,$65,$FF,$14,$67,$FA,$5F,$33,$F9,$E9,$23,$39,$23,$11
	.dc.b	$FA,$F6,$BF,$E7,$49,$E4,$FF,$35,$F9,$FF,$E7,$04,$70,$43,$F9,$C3
	.dc.b	$96,$9F,$ED,$AF,$9F,$7C,$F7,$F8,$9D,$B1,$98,$C1,$FD,$17,$E7,$E4
	.dc.b	$7F,$45,$6B,$FD,$32,$3F,$AA,$91,$FE,$A6,$D7,$46,$08,$7F,$B9,$FE
	.dc.b	$47,$FD,$67,$E4,$1A,$48,$FE,$AD,$E3,$C7,$FA,$B3,$5E,$5C,$5F,$E4
	.dc.b	$19,$FC,$24,$7F,$55,$6A,$AE,$D3,$FD,$87,$96,$4F,$F6,$16,$BF,$D8
	.dc.b	$5A,$FF,$61,$6B,$FC,$CB,$C9,$FD,$39,$23,$30,$1F,$3F,$FE,$5D,$7F
	.dc.b	$D5,$ED,$74,$D7,$7E,$3F,$F9,$BB,$CD,$7C,$D7,$FA,$46,$7F,$D6,$F1
	.dc.b	$8F,$27,$53,$99,$FF,$67,$B5,$7C,$BB,$E1,$BF,$C9,$F1,$86,$C6,$7C
	.dc.b	$0F,$F3,$EB,$67,$FD,$04,$D7,$FD,$62,$D7,$FD,$92,$D7,$E0,$35,$FE
	.dc.b	$8D,$F3,$1F,$E1,$17,$FD,$9F,$D0,$8C,$47,$C7,$8C,$7A,$1F,$CE,$E7
	.dc.b	$33,$E7,$F3,$3C,$B8,$FF,$73,$C3,$D9,$18,$24,$FA,$27,$44,$8E,$9E
	.dc.b	$2C,$9F,$E8,$57,$FD,$19,$B1,$97,$F8,$94,$65,$B1,$97,$E0,$FC,$FF
	.dc.b	$FA,$37,$03,$49,$B7,$CC,$6C,$FC,$C6,$BF,$C7,$F5,$7C,$6F,$F5,$B6
	.dc.b	$C7,$F5,$76,$BB,$8F,$F5,$9F,$C0,$FF,$5A,$F6,$3F,$DF,$6C,$F2,$7F
	.dc.b	$CD,$0D,$8C,$D7,$E5,$63,$83,$4C,$4B,$E7,$F7,$09,$FE,$CF,$18,$6C
	.dc.b	$71,$35,$3C,$5C,$0F,$9E,$7C,$FB,$E1,$7F,$D9,$71,$86,$C6,$67,$FD
	.dc.b	$A5,$39,$6F,$C7,$FF,$AE,$E0,$D2,$6D,$FE,$D9,$C5,$F3,$EF,$8F,$F5
	.dc.b	$61,$0F,$64,$66,$7D,$47,$09,$DB,$F3,$23,$53,$C6,$1F,$16,$F9,$2D
	.dc.b	$CF,$CD,$8C,$4F,$46,$C7,$16,$C7,$E9,$38,$7F,$CC,$2F,$DF,$F6,$FC
	.dc.b	$D5,$FA,$59,$CB,$F8,$78,$CB,$F2,$F9,$1F,$3F,$8E,$0F,$E0,$2D,$8C
	.dc.b	$E1,$FE,$FD,$9C,$B6,$33,$5F,$C6,$4F,$E3,$53,$A7,$07,$7A,$E8,$78
	.dc.b	$BA,$FF,$79,$F3,$F7,$A4,$D2,$63,$FB,$06,$DF,$EB,$9A,$FF,$0C,$CB
	.dc.b	$FD,$B3,$3F,$C4,$9B,$1F,$C4,$9A,$EF,$C6,$F8,$79,$FF,$A2,$9B,$3A
	.dc.b	$9F,$E2,$AD,$7F,$92,$B7,$1F,$E2,$9F,$0B,$E9,$26,$1F,$C1,$9E,$7E
	.dc.b	$9C,$8F,$23,$6F,$C8,$D7,$D4,$D7,$F8,$F7,$CF,$FF,$80,$67,$FC,$A4
	.dc.b	$D8,$FF,$10,$35,$FE,$4E,$D7,$EA,$7C,$FF,$F8,$B2,$7E,$9C,$7F,$4F
	.dc.b	$8C,$38,$6B,$F6,$1A,$FF,$59,$F9,$FF,$C4,$FA,$0F,$25,$18,$6D,$FF
	.dc.b	$B1,$63,$04,$70,$4A,$FF,$64,$46,$5B,$19,$7F,$19,$E5,$FF,$7F,$F8
	.dc.b	$5F,$CC,$B6,$7D,$6D,$9F,$F0,$34,$F8,$FE,$7E,$57,$FB,$5D,$B3,$FE
	.dc.b	$0B,$F8,$6E,$7E,$79,$A4,$AE,$BF,$2F,$FC,$52,$BF,$C0,$8D,$C4,$D2
	.dc.b	$47,$99,$B3,$CA,$3E,$37,$14,$8F,$F7,$53,$63,$FC,$3D,$C1,$E4,$24
	.dc.b	$FB,$B1,$47,$CF,$F9,$F9,$3F,$EF,$36,$D7,$FB,$CD,$AF,$FB,$CD,$AF
	.dc.b	$FB,$13,$8F,$F3,$9C,$1F,$EB,$5E,$4F,$BB,$18,$23,$82,$5F,$F9,$51
	.dc.b	$FD,$D7,$AE,$F7,$69,$27,$FC,$BE,$FF,$10,$C1,$F3,$FF,$C5,$1F,$E0
	.dc.b	$AD,$7F,$85,$35,$FF,$12,$35,$E3,$AF,$38,$C1,$1F,$E1,$9E,$EB,$CD
	.dc.b	$A4,$C3,$FD,$14,$DB,$FF,$42,$B5,$FF,$07,$CF,$19,$FF,$3A,$FC,$FF
	.dc.b	$F5,$B6,$9F,$A2,$71,$3F,$C7,$B5,$46,$23,$FC,$E8,$D7,$FD,$E6,$D5
	.dc.b	$1C,$27,$A1,$FC,$D8,$7C,$E3,$9B,$0D,$8F,$EB,$0D,$AF,$E5,$F1,$C2
	.dc.b	$7E,$31,$1F,$E0,$31,$F4,$A3,$0F,$F0,$4C,$F9,$3F,$52,$32,$D9,$F8
	.dc.b	$7F,$3F,$F0,$48,$E0,$91,$F0,$F1,$4E,$A3,$0F,$F4,$1F,$4F,$FB,$1D
	.dc.b	$7F,$BB,$5B,$3F,$EB,$B6,$BF,$E4,$26,$BF,$E4,$16,$AE,$14,$61,$BE
	.dc.b	$27,$73,$FA,$87,$C1,$BD,$5A,$4C,$3F,$DF,$D1,$97,$F9,$3F,$C2,$BE
	.dc.b	$7D,$F3,$FF,$DA,$A3,$0D,$8C,$CF,$06,$E7,$FD,$8A,$B7,$39,$AE,$04
	.dc.b	$7F,$D7,$FE,$DF,$EC,$88,$E0,$95,$F8,$FC,$D3,$A8,$C3,$F9,$5F,$82
	.dc.b	$95,$F9,$A6,$CF,$D5,$6B,$F9,$66,$BF,$E3,$FE,$2F,$E8,$2F,$F8,$ED
	.dc.b	$B3,$F9,$4F,$77,$F2,$67,$EF,$7C,$BF,$F0,$BD,$7F,$80,$1A,$FF,$81
	.dc.b	$5A,$FF,$81,$BC,$DE,$04,$AF,$CA,$CE,$A7,$FC,$81,$3F,$CA,$FC,$29
	.dc.b	$5F,$CD,$1B,$3F,$E0,$EF,$0B,$F9,$9F,$AF,$FC,$C4,$95,$FC,$E9,$FE
	.dc.b	$69,$D7,$23,$F9,$EB,$6B,$EB,$7C,$FF,$E8,$9F,$F5,$A7,$0A,$31,$08
	.dc.b	$C7,$4F,$E0,$E7,$38,$FE,$37,$2F,$EA,$FC,$BF,$D3,$72,$FE,$AE,$DC
	.dc.b	$7D,$0D,$7F,$AF,$B5,$FF,$62,$57,$39,$E3,$5D,$0F,$E2,$5E,$9C,$8F
	.dc.b	$E2,$ED,$AF,$89,$E2,$FF,$48,$9F,$EB,$7E,$27,$F5,$52,$38,$24,$F0
	.dc.b	$8C,$BF,$C7,$7B,$BF,$8C,$1F,$C6,$27,$FC,$06,$E0,$FE,$39,$8F,$EB
	.dc.b	$7C,$0F,$82,$49,$F8,$2D,$AF,$82,$D7,$E0,$BF,$AD,$F6,$FF,$65,$1F
	.dc.b	$A9,$6B,$FE,$4F,$DF,$FD,$AD,$AE,$CE,$1E,$9C,$C3,$F7,$37,$FE,$72
	.dc.b	$BE,$0D,$7E,$EB,$5F,$B8,$DB,$FF,$71,$57,$39,$F1,$79,$5E,$52,$5F
	.dc.b	$67,$FC,$2B,$F3,$CF,$9F,$A4,$9F,$F0,$BC,$65,$B1,$9A,$FC,$D4,$E5
	.dc.b	$B1,$FE,$30,$F9,$FF,$E3,$FB,$1A,$4F,$07,$FC,$4B,$F3,$FF,$48,$98
	.dc.b	$E0,$87,$F8,$A1,$CB,$4F,$F1,$E9,$D8,$91,$83,$FC,$63,$F3,$FF,$78
	.dc.b	$91,$83,$FC,$6B,$F3,$EF,$7E,$92,$3F,$C6,$F1,$96,$C6,$73,$FD,$D4
	.dc.b	$E5,$B1,$FE,$50,$ED,$FE,$27,$9F,$FB,$0F,$BB,$04,$C3,$FC,$BB,$19
	.dc.b	$6C,$F0,$24,$FF,$84,$EC,$3E,$66,$F0,$3E,$47,$FD,$33,$EE,$FF,$B0
	.dc.b	$A3,$15,$FE,$9C,$AD,$D6,$B1,$B3,$93,$FC,$CD,$19,$69,$9A,$E6,$F8
	.dc.b	$AF,$B4,$FF,$9D,$AF,$9C,$E6,$EB,$E8,$8B,$5C,$23,$A5,$77,$BE,$B1
	.dc.b	$B3,$C2,$73,$29,$97,$73,$CE,$3F,$D7,$95,$DA,$FC,$31,$F1,$CF,$F0
	.dc.b	$59,$4F,$F0,$6C,$61,$C1,$FE,$C5,$F9,$D7,$CF,$3E,$7F,$FC,$27,$18
	.dc.b	$6C,$66,$7F,$D2,$F3,$96,$FC,$BF,$F3,$16,$F4,$72,$8C,$47,$FB,$42
	.dc.b	$51,$88,$F1,$3C,$7E,$83,$E8,$CF,$F8,$0E,$7F,$07,$18,$EB,$FF,$6E
	.dc.b	$3A,$AE,$1B,$72,$C3,$6F,$B6,$B6,$DB,$C9,$F8,$EC,$3C,$75,$88,$46
	.dc.b	$1F,$AA,$F8,$DB,$5F,$EE,$DD,$BC,$55,$F8,$4A,$E1,$8F,$9E,$FA,$47
	.dc.b	$18,$DC,$1A,$6D,$C6,$1E,$76,$9E,$7C,$7D,$EF,$F9,$58,$F2,$5B,$5F
	.dc.b	$D4,$F0,$FE,$AB,$3F,$E6,$91,$8A,$DB,$7E,$2B,$31,$CE,$6D,$71,$8F
	.dc.b	$ED,$F5,$C6,$73,$0A,$DF,$13,$D6,$A4,$9E,$23,$BD,$F2,$4E,$4D,$4F
	.dc.b	$F7,$9C,$BD,$38,$73,$8E,$39,$A3,$11,$B3,$A5,$F3,$77,$F3,$FF,$E5
	.dc.b	$07,$F8,$8B,$5F,$F4,$23,$5F,$E1,$E7,$FA,$01,$C3,$F2,$AD,$F6,$9F
	.dc.b	$11,$B1,$E4,$6B,$E3,$C1,$5D,$F6,$FA,$93,$FE,$8C,$E6,$F4,$BE,$0A
	.dc.b	$FE,$3D,$AF,$FA,$09,$B1,$FC,$A2,$76,$FD,$C6,$BC,$E7,$84,$62,$78
	.dc.b	$CF,$FA,$87,$B5,$F4,$5F,$71,$B1,$F3,$1A,$FA,$6B,$E6,$E1,$E8,$1F
	.dc.b	$40,$E7,$FF,$4B,$C6,$1B,$19,$AF,$F3,$AB,$5F,$F3,$73,$5F,$F0,$33
	.dc.b	$5F,$F1,$DB,$5F,$F1,$D3,$7D,$17,$DD,$79,$36,$FE,$8B,$18,$D7,$F3
	.dc.b	$2D,$BF,$C7,$E3,$E5,$6F,$D0,$BE,$87,$D2,$D8,$FF,$45,$B5,$FF,$42
	.dc.b	$35,$FF,$0E,$35,$FF,$28,$B5,$FF,$28,$36,$4E,$6B,$D3,$E4,$F9,$B2
	.dc.b	$34,$FD,$8D,$70,$C3,$F6,$35,$F9,$BC,$3F,$D1,$9C,$BF,$85,$8C,$47
	.dc.b	$E2,$5A,$FF,$00,$6B,$FE,$DA,$6F,$0F,$FA,$CE,$79,$1F,$8F,$5F,$1B
	.dc.b	$FC,$01,$B5,$FB,$06,$BF,$9D,$AF,$26,$1F,$9C,$9F,$C2,$7A,$1F,$99
	.dc.b	$E0,$FF,$7F,$5F,$F8,$01,$AF,$FB,$CD,$AF,$E8,$5A,$FF,$18,$6B,$FE
	.dc.b	$F2,$6C,$F6,$8F,$F8,$03,$87,$F1,$6F,$6F,$CB,$87,$8D,$BC,$1C,$F1
	.dc.b	$F9,$BF,$40,$F3,$7F,$32,$DC,$7F,$D9,$6D,$7F,$D8,$8D,$7F,$CB,$8D
	.dc.b	$7F,$D2,$2D,$7F,$D2,$0D,$C0,$E7,$AD,$F4,$2F,$84,$FE,$5F,$18,$6C
	.dc.b	$66,$FF,$33,$5E,$3B,$FE,$13,$6F,$4B,$F9,$CD,$7F,$2B,$6C,$67,$C0
	.dc.b	$EB,$1F,$CD,$67,$9F,$CB,$FC,$02,$3F,$01,$FE,$21,$B3,$E0,$BE,$61
	.dc.b	$27,$F9,$CE,$1E,$1F,$0F,$FA,$3C,$3E,$73,$F3,$6F,$F2,$A3,$F9,$5B
	.dc.b	$87,$06,$FA,$3F,$C2,$B1,$98,$E0,$D7,$8C,$68,$8F,$F9,$2B,$93,$F8
	.dc.b	$BE,$BF,$91,$6F,$1F,$F6,$06,$CF,$F9,$2A,$FC,$55,$F8,$5F,$BA,$FC
	.dc.b	$41,$27,$FA,$6F,$33,$E8,$71,$1F,$CE,$F6,$56,$2F,$F4,$17,$BA,$FD
	.dc.b	$37,$DB,$FE,$A4,$8F,$F5,$8E,$9F,$F2,$5D,$79,$E7,$33,$FD,$1F,$34
	.dc.b	$7F,$00,$F2,$3F,$8D,$FC,$13,$E6,$29,$2F,$F8,$EF,$D1,$3E,$8B,$11
	.dc.b	$FD,$8F,$35,$F1,$8F,$EE,$31,$8B,$FE,$1F,$7C,$67,$FB,$6E,$CF,$99
	.dc.b	$FD,$61,$23,$8E,$1C,$DF,$47,$95,$FC,$1C,$8F,$6F,$99,$7C,$2E,$23
	.dc.b	$67,$F0,$B9,$FE,$6F,$3A,$8F,$0E,$9F,$BB,$23,$F5,$3E,$22,$BF,$4E
	.dc.b	$2F,$0F,$D2,$9F,$C9,$F2,$F3,$AF,$C7,$B5,$FC,$6E,$3F,$90,$23,$F9
	.dc.b	$43,$F3,$4E,$1E,$2F,$F2,$5F,$96,$FE,$48,$48,$FF,$0D,$9C,$0F,$93
	.dc.b	$F4,$3F,$C4,$C8,$FE,$33,$A6,$BF,$8B,$8B,$8F,$E2,$A7,$F9,$87,$8F
	.dc.b	$2B,$F9,$83,$FD,$73,$E3,$5E,$C6,$67,$F1,$13,$6F,$FC,$1B,$CA,$BD
	.dc.b	$D7,$E0,$3E,$39,$F1,$5E,$FA,$46,$27,$F6,$B1,$89,$D4,$E6,$7F,$79
	.dc.b	$39,$77,$19,$7F,$C5,$31,$97,$71,$97,$FC,$A7,$CB,$FE,$F1,$7F,$1F
	.dc.b	$F0,$1F,$CE,$4F,$EF,$33,$DF,$1F,$DE,$7D,$BF,$F5,$32,$57,$EA,$FE
	.dc.b	$8A,$BF,$A9,$8B,$3F,$D4,$F1,$57,$F4,$5F,$81,$97,$F9,$E7,$FC,$BB
	.dc.b	$DE,$FA,$2C,$CF,$EF,$E6,$E1,$FB,$E9,$DF,$8F,$FD,$74,$95,$FD,$75
	.dc.b	$FC,$57,$8D,$FC,$D7,$E8,$7F,$98,$92,$7F,$86,$CF,$F5,$7F,$62,$47
	.dc.b	$EC,$1D,$3F,$C3,$DC,$B4,$FE,$63,$E5,$2B,$D3,$EA,$FF,$6F,$24,$FF
	.dc.b	$6F,$7F,$9D,$F8,$BF,$D9,$49,$1F,$C7,$A3,$E0,$F5,$FF,$77,$24,$7F
	.dc.b	$77,$7F,$A1,$F8,$1F,$DB,$49,$1F,$C9,$A2,$BF,$C9,$62,$EB,$FC,$92
	.dc.b	$6F,$93,$F8,$09,$B5,$FD,$EE,$7F,$D1,$FE,$9B,$E7,$3F,$D5,$93,$FE
	.dc.b	$72,$8C,$79,$DF,$B5,$6C,$7F,$4A,$6B,$FD,$3F,$07,$88,$F8,$EF,$F4
	.dc.b	$DC,$61,$B1,$99,$FF,$49,$D7,$2C,$FF,$53,$F3,$8F,$A0,$FD,$33,$6B
	.dc.b	$F4,$AD,$7F,$46,$D7,$F4,$2D,$7F,$CA,$AD,$5D,$28,$C3,$67,$FD,$01
	.dc.b	$F6,$7F,$CB,$6F,$FC,$58,$DC,$FF,$D7,$6D,$C3,$9E,$1E,$3A,$F4,$F9
	.dc.b	$9E,$82,$4F,$A1,$AF,$A1,$AF,$A1,$F0,$E1,$FE,$C0,$BE,$98,$78,$23
	.dc.b	$FC,$DB,$87,$87,$CE,$56,$B0,$46,$2B,$91,$C3,$0F,$F3,$AE,$EF,$91
	.dc.b	$FE,$24,$DB,$FE,$22,$D7,$F8,$7B,$5F,$E1,$AD,$7F,$D0,$2D,$7E,$6B
	.dc.b	$74,$FC,$64,$FF,$87,$3C,$8F,$F2,$31,$2B,$FB,$1D,$77,$D7,$FD,$6B
	.dc.b	$AF,$E2,$BE,$1F,$FD,$87,$EF,$7E,$07,$5F,$7F,$C6,$63,$6F,$A6,$DC
	.dc.b	$BF,$DE,$1C,$4C,$31,$8F,$E5,$3D,$0D,$62,$F1,$BF,$D1,$D9,$F5,$6E
	.dc.b	$6C,$B9,$DF,$55,$F4,$1F,$E7,$6F,$84,$FE,$B0,$D8,$FF,$16,$B5,$FF
	.dc.b	$19,$35,$FF,$6E,$AF,$90,$FF,$42,$B7,$9F,$9C,$60,$FF,$45,$E8,$FA
	.dc.b	$07,$07,$FB,$6B,$5F,$47,$91,$75,$F4,$5E,$EB,$19,$FE,$13,$65,$E2
	.dc.b	$9E,$37,$ED,$DF,$81,$8F,$CF,$E9,$7C,$3E,$EB,$58,$AB,$53,$DB,$8B
	.dc.b	$E6,$71,$FF,$B5,$F1,$FE,$9E,$DB,$FC,$2F,$BF,$6B,$75,$F8,$28,$FF
	.dc.b	$7A,$CF,$E3,$9A,$FE,$39,$AF,$28,$FC,$BC,$63,$AB,$F8,$D3,$7A,$3F
	.dc.b	$D1,$D3,$A3,$FC,$C3,$7F,$A8,$F6,$FC,$5A,$9F,$36,$1F,$C1,$7E,$83
	.dc.b	$31,$2F,$F0,$9D,$5F,$C1,$23,$7D,$CA,$D6,$9F,$E4,$18,$C5,$FF,$18
	.dc.b	$9F,$34,$62,$7F,$8C,$C6,$09,$3B,$24,$7F,$9E,$F2,$FD,$FF,$C5,$FA
	.dc.b	$0F,$0F,$CF,$DA,$FF,$26,$E0,$8E,$E4,$8F,$E5,$2D,$BF,$E5,$2D,$7F
	.dc.b	$94,$BF,$3A,$3F,$0B,$F0,$9F,$65,$A7,$F9,$6B,$E7,$FF,$BE,$BF,$10
	.dc.b	$FC,$8F,$27,$FB,$69,$D7,$A1,$F5,$67,$78,$7F,$86,$9A,$FF,$86,$1A
	.dc.b	$FF,$85,$BC,$4F,$C3,$74,$7E,$43,$83,$FC,$FF,$F3,$D8,$99,$8F,$F9
	.dc.b	$A1,$FC,$67,$A2,$9C,$E1,$FE,$89,$9B,$7C,$AF,$81,$FE,$90,$AF,$C6
	.dc.b	$F0,$30,$E5,$3F,$EE,$99,$CC,$63,$0F,$A9,$8B,$DE,$F3,$9E,$2F,$85
	.dc.b	$B8,$7F,$A7,$1E,$24,$7F,$DE,$8F,$83,$E3,$5E,$27,$E9,$5B,$3F,$EC
	.dc.b	$0F,$95,$FD,$EE,$2D,$7F,$E0,$05,$6A,$31,$5F,$C6,$75,$4E,$6B,$78
	.dc.b	$7E,$CF,$D4,$63,$D2,$73,$87,$7C,$75,$39,$8E,$9A,$FE,$B2,$7B,$8B
	.dc.b	$E6,$FD,$2E,$CF,$8E,$4A,$FD,$8D,$71,$9F,$E5,$C7,$F5,$8C,$BE,$84
	.dc.b	$FE,$2E,$31,$3F,$73,$E7,$7E,$23,$27,$5A,$7B,$A3,$51,$8B,$E7,$7E
	.dc.b	$6A,$FE,$CF,$D3,$FC,$00,$FE,$25,$D6,$F4,$31,$87,$4E,$DF,$DB,$6F
	.dc.b	$FE,$02,$BD,$D6,$B1,$E9,$97,$FB,$17,$0F,$07,$9B,$FC,$2B,$D7,$FF
	.dc.b	$7D,$46,$0F,$F4,$17,$C3,$7F,$17,$8C,$DF,$F9,$DA,$30,$49,$FF,$3E
	.dc.b	$B6,$35,$1F,$C6,$B3,$FF,$29,$3E,$7F,$53,$F8,$CF,$B6,$F8,$9F,$E3
	.dc.b	$B1,$27,$47,$FB,$4A,$74,$E2,$FD,$7E,$57,$E4,$4F,$D8,$C5,$3E,$63
	.dc.b	$63,$FD,$E9,$17,$7F,$F1,$3F,$CD,$7F,$2C,$3F,$84,$7C,$D3,$E6,$B1
	.dc.b	$5F,$74,$46,$5F,$C1,$DF,$F5,$A8,$C3,$FE,$4D,$6B,$F8,$88,$E3,$3F
	.dc.b	$C5,$E7,$33,$C6,$30,$49,$FE,$83,$C0,$FA,$4F,$D7,$1F,$C3,$BE,$76
	.dc.b	$CF,$E2,$62,$DF,$E2,$7B,$D5,$8C,$FF,$C8,$BF,$63,$FD,$28,$FE,$93
	.dc.b	$F1,$1F,$DB,$31,$FC,$54,$65,$A7,$E7,$7A,$13,$BC,$7F,$2D,$C6,$BF
	.dc.b	$E0,$5E,$8C,$C3,$FE,$00,$FB,$56,$FF,$63,$8B,$E2,$93,$FD,$69,$B5
	.dc.b	$FA,$BC,$DF,$3C,$E3,$D5,$CA,$EA,$6E,$74,$E3,$1F,$2F,$D8,$FF,$68
	.dc.b	$3F,$53,$EF,$59,$FD,$44,$59,$FD,$86,$6A,$C6,$5F,$C6,$3D,$4F,$E3
	.dc.b	$C7,$F3,$EF,$86,$B8,$FE,$E6,$2E,$1F,$46,$6D,$7F,$9D,$3E,$C7,$FB
	.dc.b	$C9,$F8,$EF,$89,$B8,$7E,$37,$A7,$FD,$EE,$FE,$33,$03,$F9,$09,$FC
	.dc.b	$41,$1D,$19,$2C,$F2,$9F,$C9,$4E,$DF,$EE,$9D,$1F,$39,$FE,$07,$BF
	.dc.b	$F2,$8F,$84,$FE,$74,$FF,$88,$72,$7D,$07,$EC,$2B,$F3,$F1,$8E,$FF
	.dc.b	$F0,$EB,$73,$F4,$36,$7F,$4B,$E4,$FF,$38,$F8,$3F,$E6,$A7,$F3,$6B
	.dc.b	$B7,$FC,$A6,$7F,$07,$E9,$7F,$3F,$CB,$FB,$8E,$3F,$0B,$1F,$E9,$59
	.dc.b	$7E,$BA,$31,$7C,$30,$FC,$44,$FE,$B2,$B7,$DF,$E6,$8C,$5F,$B2,$30
	.dc.b	$4A,$F4,$CE,$8E,$5E,$47,$F5,$4F,$63,$FA,$F1,$FD,$FB,$DE,$B5,$FC
	.dc.b	$F3,$C1,$FF,$33,$3F,$E3,$86,$E5,$FD,$C2,$7F,$BB,$C6,$23,$FA,$5D
	.dc.b	$FF,$9F,$67,$24,$AE,$4F,$F7,$AD,$7F,$C0,$E7,$F9,$4E,$7F,$0B,$3F
	.dc.b	$D3,$3C,$BF,$8E,$FF,$6B,$8C,$DF,$F8,$CD,$AF,$81,$AE,$E3,$FC,$17
	.dc.b	$18,$8F,$F5,$AB,$4F,$F5,$BC,$68,$FF,$2D,$7D,$5F,$FA,$65,$FE,$0C
	.dc.b	$D9,$FE,$DF,$AB,$E8,$7F,$20,$FD,$BE,$B5,$6E,$3E,$F1,$BE,$1F,$C6
	.dc.b	$D7,$E5,$EF,$FB,$8C,$E7,$0F,$C7,$CE,$67,$FC,$BE,$D7,$FC,$20,$D9
	.dc.b	$E1,$1F,$9A,$8D,$61,$E2,$F8,$CF,$F3,$5E,$27,$F4,$53,$FD,$D2,$FD
	.dc.b	$13,$9C,$3F,$8B,$4F,$F7,$8C,$3C,$AD,$7C,$79,$FA,$89,$3E,$86,$C7
	.dc.b	$8D,$FF,$1D,$FC,$13,$BB,$FD,$20,$DA,$FF,$04,$FC,$CF,$F3,$1C,$BF
	.dc.b	$49,$B2,$B5,$8F,$F8,$26,$BF,$9B,$3F,$8E,$6B,$FE,$2F,$6B,$FE,$F2
	.dc.b	$6C,$9A,$4F,$6B,$FC,$9E,$EB,$E0,$5F,$19,$FE,$A5,$6E,$5F,$E1,$E3
	.dc.b	$CC,$7F,$8F,$5C,$E1,$FE,$1D,$8E,$E4,$BF,$F2,$DB,$E2,$F9,$A7,$D6
	.dc.b	$7F,$AC,$5B,$1F,$C1,$FE,$7F,$F1,$0F,$F9,$26,$33,$7F,$E2,$96,$EB
	.dc.b	$F1,$9B,$3F,$EE,$7E,$B5,$F2,$B6,$AB,$DC,$63,$DA,$F9,$D8,$7F,$8E
	.dc.b	$9B,$1F,$E3,$77,$F9,$86,$3F,$51,$CB,$77,$FF,$65,$46,$2B,$FC,$C0
	.dc.b	$F0,$6C,$7F,$B4,$6B,$FC,$9F,$AF,$FB,$F7,$C0,$7C,$AF,$F8,$42,$BF
	.dc.b	$92,$46,$2B,$FC,$27,$5B,$AD,$5D,$9C,$9F,$E6,$E8,$CB,$4C,$D7,$34
	.dc.b	$7B,$9C,$3F,$EF,$03,$FD,$21,$1C,$B8,$F3,$1F,$EE,$F3,$27,$F9,$A3
	.dc.b	$0F,$E7,$51,$AC,$3C,$93,$DA,$7D,$75,$FD,$86,$73,$3E,$18,$C3,$E7
	.dc.b	$38,$61,$A7,$C1,$87,$9A,$F9,$C7,$C0,$3F,$B6,$1E,$52,$DE,$1F,$61
	.dc.b	$FE,$3C,$F9,$E7,$C2,$BE,$7F,$8C,$61,$B1,$99,$FE,$F1,$39,$6F,$7F
	.dc.b	$FA,$85,$BB,$72,$8C,$47,$76,$BF,$E8,$E9,$FF,$7F,$4F,$D2,$D7,$EA
	.dc.b	$47,$F8,$A2,$71,$AE,$FB,$3F,$D2,$B5,$AC,$3C,$0D,$C3,$77,$F7,$30
	.dc.b	$E4,$FD,$60,$E2,$78,$46,$63,$C6,$DB,$EF,$87,$86,$3A,$9C,$B0,$F0
	.dc.b	$C7,$AD,$B1,$E1,$9E,$2D,$38,$D7,$6B,$F0,$9F,$02,$77,$3E,$66,$CE
	.dc.b	$EF,$FD,$95,$2B,$FB,$FC,$3F,$9A,$6C,$64,$FF,$17,$46,$5A,$64,$FF
	.dc.b	$3E,$B9,$69,$93,$FB,$C9,$4F,$E1,$9F,$3E,$EE,$FC,$7F,$2A,$31,$3F
	.dc.b	$E3,$D8,$C4,$EA,$73,$3F,$E2,$86,$BF,$ED,$09,$DB,$C6,$32,$F8,$27
	.dc.b	$BF,$D6,$3D,$FA,$49,$F0,$46,$5F,$F1,$BF,$CF,$BE,$1F,$F4,$BC,$08
	.dc.b	$C4,$EA,$73,$18,$8F,$F1,$EE,$DF,$E8,$E7,$D8,$D7,$3E,$BF,$C2,$8C
	.dc.b	$47,$AD,$AF,$F3,$36,$BF,$EF,$56,$E9,$F9,$49,$FB,$7D,$2F,$97,$A4
	.dc.b	$BF,$B7,$19,$7E,$76,$BF,$DA,$DB,$E0,$7D,$48,$C4,$7F,$95,$9A,$FC
	.dc.b	$F6,$BF,$D1,$DA,$B0,$FB,$D3,$CC,$FC,$7F,$89,$FC,$2F,$84,$D2,$57
	.dc.b	$E1,$3E,$7B,$FD,$9C,$FE,$0D,$AF,$F9,$A1,$AF,$E0,$5A,$FF,$59,$6B
	.dc.b	$E3,$6D,$9A,$6A,$BC,$7D,$5E,$62,$57,$E8,$AF,$FC,$14,$FF,$26,$FB
	.dc.b	$D7,$C4,$FF,$25,$6C,$66,$30,$47,$04,$3F,$D3,$51,$C1,$FF,$4E,$7C
	.dc.b	$DF,$F7,$C7,$F9,$9B,$5C,$F8,$5F,$CC,$E3,$11,$FE,$EE,$97,$CB,$78
	.dc.b	$C9,$1E,$76,$BE,$06,$BC,$5E,$9E,$7C,$8F,$0B,$5F,$3F,$2F,$BC,$8F
	.dc.b	$C3,$6B,$A3,$04,$3F,$D3,$FF,$39,$EF,$7F,$70,$D7,$FC,$20,$D7,$F1
	.dc.b	$4D,$DC,$BE,$2F,$F8,$4B,$67,$F8,$43,$5E,$F1,$F5,$7C,$D9,$1F,$65
	.dc.b	$AF,$FA,$CB,$F5,$A7,$8F,$FC,$51,$B1,$FC,$4D,$AE,$E3,$FB,$27,$DE
	.dc.b	$FE,$AE,$3F,$DA,$8D,$7F,$C0,$A7,$F8,$19,$C3,$86,$B9,$E8,$E6,$47
	.dc.b	$FB,$9C,$6B,$89,$F7,$BF,$C0,$9B,$D5,$FD,$E1,$B5,$FE,$E5,$6B,$FE
	.dc.b	$DC,$F1,$7F,$9F,$F9,$1F,$DE,$23,$F9,$EB,$5D,$18,$8D,$91,$C1,$C3
	.dc.b	$EC,$B9,$91,$C1,$0F,$F7,$5B,$96,$9F,$E0,$EF,$3F,$F9,$AB,$FE,$2A
	.dc.b	$6B,$FE,$A3,$ED,$FF,$08,$46,$08,$E0,$91,$EF,$FB,$CF,$E9,$CF,$F0
	.dc.b	$C6,$E2,$69,$AE,$C6,$4F,$BA,$F8,$A4,$BF,$8A,$D7,$FA,$FC,$6D,$C9
	.dc.b	$0C,$FD,$D7,$CE,$23,$F3,$9A,$E8,$C1,$0F,$C3,$9F,$E4,$CF,$C7,$9E
	.dc.b	$EF,$F8,$C5,$B1,$FE,$36,$6B,$C0,$38,$79,$B5,$FC,$9B,$5E,$D1,$D4
	.dc.b	$FF,$1F,$7B,$EF,$4E,$92,$7F,$C8,$B1,$97,$F1,$3F,$77,$F8,$C2,$38
	.dc.b	$21,$F2,$1C,$B4,$FC,$B7,$C7,$7F,$B2,$DB,$EC,$FE,$6E,$31,$1F,$13
	.dc.b	$74,$62,$3F,$5B,$18,$9E,$04,$8E,$67,$F9,$73,$E1,$BF,$92,$B7,$C1
	.dc.b	$FF,$2E,$46,$23,$E5,$35,$FA,$6D,$7F,$76,$D7,$FD,$02,$D4,$DF,$AB
	.dc.b	$72,$FF,$9A,$5B,$3F,$E6,$E6,$BF,$65,$AF,$B3,$F6,$67,$C2,$FE,$F5
	.dc.b	$B1,$FB,$D6,$BF,$15,$AF,$18,$E0,$47,$83,$FC,$0B,$E6,$64,$7F,$02
	.dc.b	$6B,$F9,$E6,$BF,$C6,$B9,$7F,$4F,$18,$8F,$F6,$37,$7B,$7F,$E5,$04
	.dc.b	$8F,$E5,$0F,$F1,$0F,$94,$91,$FC,$41,$AF,$F3,$17,$F9,$81,$1F,$E6
	.dc.b	$0D,$74,$60,$87,$F3,$9F,$89,$FE,$70,$47,$F9,$C3,$FC,$83,$E4,$64
	.dc.b	$7F,$20,$6B,$FD,$07,$CF,$FE,$80,$47,$FA,$03,$FC,$A2,$64,$7F,$28
	.dc.b	$6B,$FC,$11,$AF,$C7,$E9,$46,$23,$FD,$8F,$9F,$F1,$08,$C4,$7F,$BB
	.dc.b	$27,$6F,$53,$FC,$C5,$FA,$DF,$FA,$C7,$C8,$34,$91,$FC,$F3,$E3,$63
	.dc.b	$FC,$F1,$AF,$EA,$7F,$46,$FE,$7C,$FE,$A1,$AF,$FB,$6B,$9D,$18,$8F
	.dc.b	$F1,$AB,$63,$3E,$17,$F4,$F8,$C1,$1F,$E9,$BA,$4A,$FE,$9A,$D7,$F8
	.dc.b	$EF,$8E,$F8,$57,$D5,$FF,$B4,$23,$0D,$8C,$CF,$F1,$D9,$CB,$63,$2F
	.dc.b	$46,$BF,$DC,$23,$6E,$67,$FB,$37,$EF,$7F,$C2,$63,$E0,$35,$FF,$29
	.dc.b	$B7,$23,$49,$D9,$C8,$93,$E3,$F3,$FD,$04,$70,$43,$CC,$E5,$A7,$F7
	.dc.b	$79,$7D,$37,$98,$91,$97,$FD,$E5,$F4,$9F,$E9,$3C,$92,$F8,$47,$CB
	.dc.b	$7F,$84,$2B,$53,$98,$C7,$A7,$FE,$1C,$AD,$CF,$F9,$DF,$F4,$47,$1F
	.dc.b	$F8,$62,$73,$2C,$39,$C6,$1B,$87,$4C,$7C,$73,$D3,$95,$8F,$4C,$33
	.dc.b	$7B,$7F,$AC,$61,$E6,$7F,$B0,$B5,$D1,$8F,$4B,$FC,$E2,$DF,$67,$FB
	.dc.b	$03,$E5,$8C,$BF,$05,$AF,$C6,$94,$7F,$9E,$B1,$F5,$36,$FE,$5E,$2F
	.dc.b	$D2,$9F,$34,$F8,$9F,$4E,$5C,$F0,$E0,$DC,$F2,$F8,$DD,$E1,$CE,$F9
	.dc.b	$4F,$D3,$38,$17,$93,$F8,$3F,$8F,$A8,$FE,$BB,$DB,$FD,$33,$2F,$F1
	.dc.b	$E3,$63,$FB,$23,$5F,$D0,$7E,$92,$F7,$39,$47,$E1,$B8,$93,$9B,$FD
	.dc.b	$63,$67,$F6,$AD,$79,$D7,$27,$FC,$CB,$19,$7A,$35,$CC,$6F,$1F,$E4
	.dc.b	$3F,$35,$71,$34,$98,$7B,$1B,$7E,$C6,$BE,$CF,$94,$F1,$65,$EA,$6E
	.dc.b	$46,$92,$FD,$AD,$AF,$6B,$EB,$F0,$BF,$C8,$3E,$85,$F0,$FF,$C5,$F7
	.dc.b	$FE,$64,$8C,$1E,$BF,$0B,$F4,$B1,$FE,$6D,$6E,$1F,$CB,$7A,$74,$49
	.dc.b	$D3,$63,$51,$FE,$37,$C1,$FA,$4F,$EE,$A3,$0D,$9F,$F2,$1C,$E9,$F4
	.dc.b	$46,$5F,$F4,$6F,$1F,$F4,$4C,$FE,$B3,$6F,$FA,$4C,$63,$C4,$F8,$D1
	.dc.b	$DB,$97,$99,$1C,$12,$FC,$38,$27,$51,$86,$BF,$E0,$58,$D4,$F1,$8C
	.dc.b	$4F,$93,$F5,$DF,$C3,$62,$69,$39,$3D,$B9,$4C,$3E,$1B,$70,$FF,$32
	.dc.b	$DF,$1F,$8B,$E8,$4A,$FF,$31,$46,$5B,$83,$E5,$BC,$44,$8C,$B6,$FA
	.dc.b	$CE,$E7,$24,$8F,$87,$F7,$3F,$48,$91,$F6,$1F,$77,$CB,$C8,$F0,$B6
	.dc.b	$7E,$5F,$CB,$FD,$A2,$4F,$E0,$1B,$1A,$8C,$11,$FF,$31,$7E,$D1,$FC
	.dc.b	$6A,$3F,$D7,$CD,$8C,$F6,$FE,$C6,$31,$1F,$E6,$66,$BF,$DF,$DA,$BA
	.dc.b	$D1,$86,$CF,$F0,$1A,$7C,$CB,$C7,$C3,$63,$FD,$24,$6A,$32,$7F,$09
	.dc.b	$C1,$F8,$8B,$EC,$70,$DA,$FE,$76,$D9,$FF,$4E,$35,$FF,$52,$B5,$FF
	.dc.b	$76,$B5,$FE,$C6,$D7,$BC,$FF,$13,$F7,$7F,$D7,$D5,$FE,$E5,$6C,$FF
	.dc.b	$A3,$5A,$FF,$A3,$1A,$FF,$89,$5A,$FC,$76,$E4,$69,$AE,$8F,$DE,$F2
	.dc.b	$FE,$E0,$97,$FB,$E6,$DF,$EB,$9A,$FE,$AD,$FA,$1C,$5F,$C5,$23,$FD
	.dc.b	$8B,$3F,$9B,$EB,$C9,$1F,$EC,$63,$CF,$FA,$0D,$32,$E0,$8E,$09,$3F
	.dc.b	$6A,$3F,$87,$F3,$49,$FE,$2A,$D9,$FE,$0A,$D7,$F8,$13,$5F,$F2,$FF
	.dc.b	$4B,$E0,$3F,$8C,$92,$BF,$8F,$B6,$7F,$87,$B5,$FE,$18,$FF,$4B,$FB
	.dc.b	$2F,$F2,$3C,$FF,$B7,$F6,$46,$2B,$FD,$8C,$D7,$FD,$8A,$D7,$FC,$4E
	.dc.b	$3F,$43,$FD,$0B,$79,$3F,$2D,$18,$23,$82,$5F,$F8,$42,$3F,$53,$BF
	.dc.b	$F3,$AD,$BF,$9A,$92,$7F,$9D,$B6,$FF,$95,$B5,$FE,$50,$FF,$3C,$E3
	.dc.b	$93,$FD,$09,$B3,$FC,$C9,$AF,$F2,$F6,$BF,$E7,$9F,$C0,$DF,$29,$FD
	.dc.b	$C5,$B5,$F1,$DB,$3F,$EE,$B6,$BF,$10,$7C,$83,$D2,$FD,$C3,$6B,$FC
	.dc.b	$5E,$D9,$E0,$70,$9D,$92,$76,$D8,$FE,$73,$F8,$B3,$E8,$3F,$AB,$B6
	.dc.b	$BF,$9C,$B6,$77,$3F,$C2,$7C,$87,$A5,$49,$7F,$E8,$97,$9F,$04,$8E
	.dc.b	$6D,$BF,$0C,$F8,$09,$3E,$07,$B4,$C8,$FF,$4A,$36,$7F,$D2,$4D,$7F
	.dc.b	$D3,$6D,$7F,$CE,$7E,$23,$E9,$5E,$C7,$F8,$A6,$30,$D8,$CD,$FC,$A9
	.dc.b	$CB,$FE,$5C,$FB,$FF,$E3,$44,$EE,$FE,$54,$D8,$D4,$60,$8F,$D4,$F9
	.dc.b	$77,$D1,$7F,$25,$6C,$FF,$8E,$DB,$3B,$9E,$FE,$2F,$62,$47,$8D,$B9
	.dc.b	$FE,$03,$04,$62,$3F,$15,$18,$9F,$F0,$F7,$3F,$F9,$A5,$A7,$F8,$C5
	.dc.b	$C4,$FF,$87,$9A,$FF,$87,$3E,$73,$F2,$AF,$F9,$D1,$AB,$C9,$46,$1B
	.dc.b	$CF,$FE,$22,$8C,$11,$C1,$E1,$F8,$E9,$87,$99,$B7,$FE,$0E,$FB,$DF
	.dc.b	$C7,$CF,$F8,$31,$B1,$9E,$3F,$F1,$7C,$62,$3F,$C6,$6D,$7F,$D1,$6D
	.dc.b	$7D,$0D,$C4,$E7,$2F,$F4,$77,$6F,$F9,$57,$0F,$E3,$A7,$F9,$1C,$B7
	.dc.b	$FE,$44,$6B,$FE,$42,$F8,$A7,$D4,$FF,$90,$9B,$3F,$E2,$46,$CF,$F8
	.dc.b	$A9,$AF,$19,$E0,$49,$E0,$FE,$1F,$E4,$3F,$2F,$3F,$C1,$F8,$91,$88
	.dc.b	$FE,$5B,$18,$E6,$E0,$78,$FE,$B2,$61,$E3,$6D,$7F,$8F,$3E,$89,$F6
	.dc.b	$FF,$E5,$A6,$D7,$F8,$F1,$B3,$B9,$D1,$27,$4D,$8F,$F3,$4F,$5F,$F9
	.dc.b	$9D,$FF,$73,$EC,$8C,$42,$30,$FF,$A1,$63,$31,$8A,$FF,$3C,$46,$09
	.dc.b	$3F,$E1,$D3,$FD,$C2,$3E,$89,$E4,$7F,$2D,$EB,$FF,$60,$C6,$27,$F9
	.dc.b	$87,$BF,$FE,$FA,$8C,$47,$F0,$AC,$31,$7F,$CB,$AA,$CF,$F5,$AB,$E7
	.dc.b	$39,$8F,$03,$DE,$B7,$1F,$EF,$B3,$27,$17,$2F,$A2,$FC,$2F,$F9,$E8
	.dc.b	$D9,$C5,$CD,$78,$23,$D2,$68,$E1,$1C,$95,$F6,$9E,$11,$CE,$FC,$CF
	.dc.b	$13,$67,$FB,$F0,$A7,$FB,$FA,$70,$E0,$FE,$F7,$F3,$DF,$E1,$FF,$A3
	.dc.b	$FE,$A7,$08,$C3,$FD,$59,$AF,$F5,$F6,$BC,$BE,$8D,$C3,$CA,$51,$89
	.dc.b	$E4,$D7,$93,$5E,$53,$F4,$27,$F8,$AC,$5E,$BF,$F0,$8E,$7F,$7A,$7A
	.dc.b	$D6,$EF,$B4,$E6,$B7,$97,$B9,$B8,$0F,$75,$70,$9F,$0C,$7D,$8C,$3C
	.dc.b	$2D,$AE,$2F,$0C,$3C,$07,$95,$87,$1B,$E8,$F7,$AF,$3B,$CC,$DE,$1E
	.dc.b	$B9,$DE,$DE,$57,$E9,$B6,$7F,$DC,$07,$CC,$2C,$69,$C1,$FE,$E0,$29
	.dc.b	$FE,$38,$8E,$4E,$63,$EF,$F4,$3E,$8F,$FC,$9D,$3F,$EC,$26,$BF,$EC
	.dc.b	$86,$BF,$EE,$86,$BF,$2E,$39,$79,$9D,$27,$33,$FD,$3D,$B3,$A8,$FF
	.dc.b	$67,$FD,$1F,$F8,$C1,$BD,$DE,$48,$C4,$7F,$B6,$1A,$FF,$26,$FA,$3F
	.dc.b	$F0,$DB,$FD,$DD,$AF,$F2,$D6,$BB,$EE,$F7,$EA,$EF,$A4,$9F,$F7,$0C
	.dc.b	$65,$B1,$97,$FD,$D3,$19,$6C,$7D,$6F,$25,$F3,$B4,$93,$FE,$3E,$DF
	.dc.b	$FC,$8B,$1F,$43,$6F,$F0,$EC,$FF,$BC,$1A,$FF,$BD,$9A,$FF,$9C,$35
	.dc.b	$FF,$41,$76,$71,$24,$71,$6B,$FE,$53,$BF,$F6,$A3,$F8,$AF,$3F,$F8
	.dc.b	$24,$62,$3C,$91,$FC,$1E,$9F,$46,$F6,$7F,$0A,$DA,$FF,$2B,$CE,$A7
	.dc.b	$24,$8C,$CE,$DF,$F3,$24,$ED,$CC,$F8,$FA,$A4,$78,$9A,$FF,$99,$3E
	.dc.b	$85,$F4,$7F,$C6,$DB,$1F,$C5,$9A,$EE,$34,$47,$4D,$7F,$CD,$FF,$46
	.dc.b	$F6,$3F,$93,$36,$3F,$90,$B5,$DC,$78,$7E,$6A,$47,$85,$AF,$9D,$AF
	.dc.b	$FA,$0E,$D1,$88,$FF,$54,$D7,$FC,$01,$7B,$FA,$9E,$22,$3E,$DE,$5F
	.dc.b	$80,$47,$04,$3E,$33,$96,$9F,$5B,$E8,$FE,$2B,$F8,$96,$BF,$B7,$6B
	.dc.b	$FE,$43,$6E,$A5,$F4,$7F,$C1,$9B,$3D,$63,$07,$D5,$F4,$A4,$7D,$56
	.dc.b	$BF,$82,$F9,$7F,$F0,$2B,$7C,$6F,$E2,$31,$88,$FC,$F3,$5F,$F6,$7F
	.dc.b	$D1,$FE,$B1,$FE,$AA,$D7,$FD,$52,$D7,$FD,$64,$D7,$8F,$47,$F6,$38
	.dc.b	$C4,$7F,$41,$6A,$73,$CA,$FA,$3F,$DF,$4F,$FB,$82,$31,$D7,$FD,$99
	.dc.b	$B1,$FE,$EB,$9D,$BF,$E2,$6F,$89,$7B,$DF,$CE,$9B,$1F,$C0,$DA,$FE
	.dc.b	$DD,$AF,$F3,$76,$BC,$E3,$91,$1E,$5F,$3D,$D4,$8E,$08,$75,$72,$D3
	.dc.b	$FB,$FF,$9F,$FC,$D9,$FF,$29,$F7,$7F,$82,$63,$04,$70,$48,$E2,$FC
	.dc.b	$3F,$A3,$FE,$9F,$1F,$DB,$72,$46,$21,$18,$F1,$B8,$1F,$17,$E8,$E4
	.dc.b	$FC,$56,$BF,$15,$AF,$C5,$6B,$F3,$FE,$0F,$E7,$11,$F9,$CD,$7E,$73
	.dc.b	$5F,$9C,$FF,$89,$FE,$8D,$F0,$94,$93,$FE,$24,$8C,$B6,$32,$FF,$BE
	.dc.b	$A3,$2D,$8F,$F1,$97,$AC,$FA,$3A,$49,$FF,$17,$C6,$5F,$C2,$7C,$F7
	.dc.b	$E1,$C8,$E0,$87,$E1,$DC,$B4,$FF,$0E,$1F,$97,$FA,$37,$C1,$FE,$5D
	.dc.b	$B1,$A3,$EB,$38,$71,$AA,$31,$1C,$E7,$91,$1E,$47,$DC,$8D,$BE,$2F
	.dc.b	$94,$FE,$42,$FF,$93,$1A,$F0,$6B,$E3,$6B,$C7,$C2,$FF,$2D,$46,$23
	.dc.b	$FC,$70,$D4,$DF,$A3,$72,$7F,$98,$5B,$3F,$E5,$C6,$BE,$86,$BF,$BD
	.dc.b	$FA,$37,$D1,$FE,$F5,$B1,$EE,$CD,$18,$8F,$8E,$D7,$94,$F1,$23,$C6
	.dc.b	$34,$FF,$05,$FA,$3F,$E0,$84,$7F,$82,$35,$FE,$08,$D7,$F8,$23,$FC
	.dc.b	$A3,$59,$1F,$DD,$3B,$91,$88,$FE,$19,$F4,$7F,$C4,$48,$FF,$11,$7F
	.dc.b	$98,$91,$E4,$D7,$F9,$C3,$5F,$E6,$AD,$7F,$C8,$12,$F9,$9F,$E7,$64
	.dc.b	$8F,$23,$FC,$87,$E8,$E4,$7F,$21,$6B,$FD,$17,$D1,$FE,$86,$47,$D8
	.dc.b	$FF,$29,$99,$1F,$CA,$5A,$FF,$29,$6B,$FC,$A5,$AF,$F3,$36,$BF,$ED
	.dc.b	$7E,$EF,$43,$CC,$FF,$0E,$FC,$B7,$F8,$8C,$8F,$F9,$99,$FF,$14,$C3
	.dc.b	$E5,$FF,$9E,$12,$7F,$9E,$3F,$D1,$7E,$63,$F9,$CC,$7F,$20,$6F,$01
	.dc.b	$A4,$F0,$78,$12,$7F,$21,$1C,$BD,$46,$1F,$E4,$49,$FE,$9D,$9F,$F4
	.dc.b	$D2,$3F,$B4,$F1,$DF,$46,$FA,$3F,$F5,$EC,$61,$B1,$99,$E0,$D7,$F8
	.dc.b	$34,$ED,$CC,$F0,$23,$FA,$5F,$A3,$FE,$02,$FF,$3E,$6E,$86,$93,$B3
	.dc.b	$FB,$53,$67,$94,$79,$3D,$09,$1E,$46,$BE,$4C,$7F,$C5,$4D,$8C,$BF
	.dc.b	$C8,$B1,$F8,$31,$DB,$E8,$E4,$F9,$DA,$FF,$1A,$FA,$27,$A4,$BF,$F6
	.dc.b	$37,$C2,$BE,$8B,$F9,$6D,$6A,$73,$18,$AE,$17,$FE,$F8,$8F,$F5,$27
	.dc.b	$D1,$FE,$1E,$36,$7F,$B4,$6F,$93,$B3,$15,$CB,$0F,$04,$63,$0E,$3E
	.dc.b	$AB,$0E,$93,$C4,$61,$D1,$FE,$7C,$D7,$CD,$5F,$EF,$E8,$C4,$78,$FD
	.dc.b	$37,$B5,$96,$C7,$C5,$6B,$D1,$AF,$C4,$77,$8F,$A2,$3C,$71,$8C,$FF
	.dc.b	$D8,$DF,$46,$FA,$37,$C5,$7C,$4B,$DE,$37,$26,$3C,$67,$CD,$3F,$0B
	.dc.b	$0E,$57,$BA,$D7,$27,$F1,$3F,$1F,$AE,$1F,$D7,$3B,$BF,$A3,$DF,$F3
	.dc.b	$B6,$C6,$67,$6F,$E2,$7E,$8B,$FD,$0D,$5F,$AB,$6C,$7E,$91,$AE,$FD
	.dc.b	$DE,$28,$FC,$54,$69,$C7,$02,$31,$1A,$9E,$2D,$AF,$37,$D1,$E3,$2F
	.dc.b	$E5,$CD,$C8,$D2,$4F,$02,$57,$C1,$8E,$B8,$F8,$7E,$8E,$E4,$69,$2F
	.dc.b	$E0,$36,$FE,$03,$5F,$F2,$B7,$A1,$E8,$EF,$FE,$2A,$DE,$FF,$F1,$5D
	.dc.b	$FF,$97,$E3,$D5,$E0,$7F,$29,$8F,$E3,$BE,$16,$08,$E0,$98,$7F,$85
	.dc.b	$27,$83,$67,$C1,$F3,$DF,$3E,$BF,$DD,$CD,$9F,$F7,$13,$5F,$F1,$2E
	.dc.b	$28,$C4,$7D,$7E,$A7,$89,$F0,$EF,$FA,$1E,$9F,$52,$72,$D9,$D5,$76
	.dc.b	$E6,$F1,$92,$FC,$6D,$8D,$46,$08,$E0,$87,$9B,$0F,$D9,$C6,$A3,$FC
	.dc.b	$7B,$83,$E8,$FF,$12,$DE,$4F,$34,$60,$FC,$6E,$7E,$FC,$FE,$39,$30
	.dc.b	$F8,$EF,$0F,$8E,$91,$FE,$FE,$A4,$63,$1F,$F7,$DC,$3E,$7B,$B9,$2B
	.dc.b	$83,$67,$6D,$7D,$33,$C6,$7E,$BF,$D1,$FD,$A2,$38,$3C,$5F,$33,$EC
	.dc.b	$23,$83,$EF,$FC,$3F,$E0,$88,$E0,$97,$F8,$28,$CB,$63,$FC,$79,$F4
	.dc.b	$7F,$C3,$9B,$D9,$FA,$F8,$C4,$FF,$99,$1A,$FF,$B0,$DA,$FF,$A7,$DA
	.dc.b	$FF,$A7,$9B,$D6,$73,$4F,$A3,$78,$DF,$C0,$23,$0D,$8C,$C7,$F9,$B8
	.dc.b	$C9,$FE,$70,$B7,$D1,$BE,$7B,$F8,$4C,$61,$B1,$99,$E0,$D9,$FF,$3C
	.dc.b	$B5,$FF,$42,$35,$FF,$45,$B5,$FE,$38,$DC,$5E,$EB,$DE,$FE,$2C,$DB
	.dc.b	$FF,$34,$36,$FF,$D1,$AD,$7F,$DD,$8D,$7F,$D9,$0D,$7F,$D8,$ED,$AE
	.dc.b	$F5,$FC,$03,$96,$5F,$F0,$06,$DF,$F0,$06,$BF,$AB,$6B,$FC,$1F,$8B
	.dc.b	$F8,$AB,$FE,$A3,$9F,$CE,$76,$7D,$39,$CC,$FF,$AB,$27,$27,$D1,$3F
	.dc.b	$DF,$A5,$F4,$3E,$8B,$88,$FB,$11,$82,$46,$C9,$1F,$C5,$B9,$7F,$8D
	.dc.b	$92,$3F,$8D,$B6,$35,$18,$23,$FC,$7B,$E0,$7F,$91,$92,$3F,$91,$B6
	.dc.b	$35,$18,$23,$FE,$C0,$F9,$BF,$F2,$34,$7F,$A7,$DB,$19,$F4,$7F,$A9
	.dc.b	$46,$23,$FD,$95,$F4,$7F,$A0,$4D,$27,$2F,$E5,$1B,$7F,$99,$6C,$FF
	.dc.b	$36,$CF,$F9,$D9,$27,$F9,$DB,$63,$51,$82,$3F,$CF,$B9,$3F,$A1,$92
	.dc.b	$3F,$A1,$B6,$35,$18,$23,$FE,$E0,$FA,$37,$CA,$7F,$6F,$6C,$FF,$B0
	.dc.b	$5B,$3B,$9F,$D7,$FC,$8F,$FB,$BA,$BF,$68,$D9,$FF,$72,$B5,$FF,$78
	.dc.b	$35,$FF,$7B,$B5,$FF,$04,$35,$34,$D7,$7B,$E8,$DF,$47,$FD,$55,$B7
	.dc.b	$FE,$F4,$9D,$4E,$4F,$E0,$BE,$4F,$F9,$7A,$7F,$D0,$7E,$9E,$08,$F1
	.dc.b	$BF,$E3,$77,$DF,$7E,$C4,$9D,$57,$84,$8E,$0F,$F4,$7C,$78,$08,$E0
	.dc.b	$99,$7F,$A5,$E3,$2D,$8F,$F1,$53,$3C,$57,$D1,$BD,$9F,$F1,$3C,$61
	.dc.b	$B1,$9A,$E3,$5F,$E3,$3A,$D4,$7F,$29,$FA,$2F,$E3,$92,$69,$3B,$FF
	.dc.b	$95,$B6,$78,$46,$0F,$F1,$BF,$CD,$BE,$8F,$F9,$33,$63,$31,$83,$AF
	.dc.b	$8B,$D0,$95,$E3,$6E,$1E,$83,$C8,$E1,$F5,$E0,$8F,$F1,$87,$93,$FE
	.dc.b	$6A,$AE,$08,$C5,$7C,$27,$11,$C0,$FF,$0E,$7C,$EF,$E7,$27,$FC,$1A
	.dc.b	$D7,$FC,$48,$D5,$C0,$8C,$37,$97,$99,$2B,$F2,$53,$C6,$FC,$9F,$21
	.dc.b	$E8,$24,$FA,$21,$1F,$E2,$FF,$A3,$FC,$B6,$3F,$6B,$14,$62,$3F,$C4
	.dc.b	$6D,$7F,$C7,$4D,$7C,$ED,$8E,$3C,$7D,$49,$87,$C8,$1D,$9F,$E5,$B6
	.dc.b	$99,$70,$7F,$90,$FE,$2D,$F4,$74,$93,$FB,$18,$CB,$63,$38,$FA,$E7
	.dc.b	$2D,$8F,$C5,$FC,$8F,$E6,$63,$F3,$3C,$68,$C4,$7E,$92,$31,$CD,$C0
	.dc.b	$F2,$7D,$1C,$C7,$C4,$DA,$F9,$5F,$46,$FA,$3F,$F2,$E3,$6B,$F8,$43
	.dc.b	$67,$73,$A2,$4E,$9B,$1F,$8D,$EC,$FF,$35,$BF,$91,$6F,$17,$FB,$92
	.dc.b	$BF,$2B,$18,$24,$FF,$A2,$5B,$1F,$E1,$D8,$FF,$78,$8F,$A2,$FF,$29
	.dc.b	$BF,$C0,$E3,$FD,$5D,$18,$8F,$E0,$FF,$03,$8F,$1F,$F8,$0F,$67,$FB
	.dc.b	$0A,$B5,$5D,$1B,$19,$9F,$D8,$57,$1A,$EC,$78,$63,$2D,$33,$5D,$14
	.dc.b	$F1,$7F,$88,$5D,$7C,$15,$E4,$8C,$47,$03,$A1,$C0,$A7,$03,$BC,$78
	.dc.b	$AB,$D3,$3C,$AB,$E4,$57,$33,$FC,$8C,$E0,$FF,$24,$7C,$FB,$C9,$7D
	.dc.b	$1E,$1B,$5F,$C4,$67,$53,$93,$8F,$CB,$38,$A9,$2B,$84,$65,$B1,$FE
	.dc.b	$E3,$F7,$3E,$B5,$FF,$09,$9C,$69,$FC,$FA,$31,$09,$DD,$F3,$6D,$F0
	.dc.b	$9F,$F7,$75,$70,$9E,$BD,$9F,$D2,$1D,$CF,$3C,$3B,$61,$E5,$D7,$73
	.dc.b	$F5,$F0,$F0,$C7,$03,$8D,$EA,$79,$1E,$0C,$3B,$61,$D3,$1F,$24,$B0
	.dc.b	$63,$E9,$AF,$75,$F0,$AE,$47,$0A,$DD,$FB,$78,$13,$F5,$CB,$0D,$B3
	.dc.b	$3A,$D7,$F8,$24,$62,$B8,$1F,$5D,$C6,$5A,$8F,$E8,$C6,$AF,$A3,$8A
	.dc.b	$D5,$F5,$8F,$0D,$F2,$69,$FB,$FF,$90,$DF,$09,$A7,$7E,$B7,$D1,$BC
	.dc.b	$EC,$DF,$2C,$58,$71,$8C,$35,$F5,$1F,$D4,$AB,$93,$E2,$CD,$F4,$4E
	.dc.b	$EF,$F5,$34,$61,$B1,$99,$FE,$DD,$39,$7F,$B8,$7A,$6F,$BB,$FE,$DD
	.dc.b	$1E,$E8,$C7,$89,$F2,$5B,$1E,$1B,$E8,$F9,$B3,$FF,$1C,$FC,$EF,$FA
	.dc.b	$DD,$B1,$FD,$DB,$CD,$4E,$5B,$19,$8C,$11,$C1,$FD,$5F,$E2,$FE,$AF
	.dc.b	$11,$A7,$D8,$AE,$73,$E3,$C7,$F9,$66,$06,$92,$3C,$CD,$7F,$C1,$6D
	.dc.b	$7F,$C1,$50,$F3,$3F,$0A,$D8,$FB,$4D,$7C,$07,$E2,$A5,$DB,$D8,$91
	.dc.b	$D9,$AF,$F8,$5E,$36,$E4,$87,$8B,$7F,$09,$1F,$C5,$47,$07,$F0,$E3
	.dc.b	$E8,$1E,$B7,$F4,$D6,$C7,$FB,$AD,$AF,$F8,$C5,$AF,$F9,$29,$AF,$F9
	.dc.b	$51,$AF,$ED,$9A,$F7,$8F,$F7,$B4,$BE,$B5,$F7,$1F,$89,$6D,$7E,$35
	.dc.b	$AF,$A2,$FC,$7C,$7F,$B4,$8F,$57,$A9,$E9,$23,$E9,$6B,$E9,$6B,$E9
	.dc.b	$7F,$25,$ED,$BE,$1F,$F7,$AD,$9F,$D9,$35,$F1,$4F,$EE,$7E,$89,$F4
	.dc.b	$9F,$E5,$59,$FF,$21,$C6,$1F,$F4,$83,$5F,$F1,$C3,$5F,$F3,$DF,$62
	.dc.b	$31,$1F,$C7,$A3,$13,$D5,$B1,$D2,$3F,$55,$F0,$BF,$B4,$8F,$D7,$37
	.dc.b	$2F,$E3,$3C,$48,$C7,$37,$5C,$7C,$95,$F2,$FE,$93,$F2,$CD,$E5,$FE
	.dc.b	$49,$8A,$73,$7C,$7E,$93,$F8,$D7,$E1,$6F,$A8,$F3,$36,$FF,$D6,$0D
	.dc.b	$7F,$CB,$0D,$7F,$D3,$7A,$79,$7E,$03,$F9,$6E,$A6,$9F,$C9,$FE,$67
	.dc.b	$F9,$1D,$7F,$11,$FA,$5F,$F3,$2B,$FC,$59,$AF,$FA,$7B,$A1,$18,$8F
	.dc.b	$F5,$4C,$63,$D8,$FC,$9C,$E6,$7F,$A6,$CE,$5E,$93,$A3,$E7,$FD,$77
	.dc.b	$98,$91,$E3,$8F,$4F,$81,$23,$F9,$D6,$5F,$C4,$38,$BE,$9F,$9B,$FD
	.dc.b	$37,$87,$FA,$BC,$E5,$B7,$C6,$FF,$94,$56,$DF,$F4,$D5,$7C,$BB,$FB
	.dc.b	$AD,$BE,$7C,$1F,$D5,$AD,$F4,$3F,$32,$26,$1E,$47,$51,$FC,$E7,$E9
	.dc.b	$BF,$BA,$CF,$F3,$D6,$BF,$D3,$B5,$46,$32,$FD,$97,$AF,$FD,$42,$B5
	.dc.b	$39,$CB,$E5,$B6,$3F,$97,$35,$FF,$3D,$79,$5F,$80,$F7,$E4,$7C,$87
	.dc.b	$C9,$E7,$7E,$13,$D9,$93,$F3,$1F,$77,$C3,$E2,$3F,$43,$17,$3F,$9F
	.dc.b	$3F,$8C,$F3,$E5,$7D,$47,$E3,$7C,$46,$23,$F5,$31,$6B,$EB,$CF,$E8
	.dc.b	$BD,$C7,$D2,$FF,$9C,$DB,$7F,$C0,$5A,$FF,$B5,$DA,$FF,$BC,$1A,$FF
	.dc.b	$BE,$9A,$FF,$87,$1A,$9C,$F5,$FE,$E7,$EC,$7F,$4A,$49,$FF,$07,$9B
	.dc.b	$AF,$D4,$FB,$B2,$3F,$22,$FE,$0F,$E8,$71,$1F,$C1,$A2,$E5,$F9,$49
	.dc.b	$FD,$8F,$C1,$4A,$FC,$CB,$FD,$F7,$E9,$9F,$3F,$99,$FE,$1F,$36,$FF
	.dc.b	$C4,$FC,$EB,$DD,$79,$BE,$97,$F2,$6F,$F2,$76,$BF,$EF,$D6,$BF,$E3
	.dc.b	$E3,$FC,$7E,$E1,$C3,$5C,$F4,$FF,$A9,$63,$10,$75,$E5,$FF,$91,$7D
	.dc.b	$9F,$F1,$F3,$67,$FC,$6E,$D7,$FB,$2C,$F9,$31,$FD,$6F,$9F,$FC,$50
	.dc.b	$93,$FE,$21,$FA,$5F,$F1,$94,$EA,$3F,$7F,$3A,$7F,$51,$3A,$EC,$FF
	.dc.b	$2A,$F8,$5F,$D6,$6F,$F9,$AB,$63,$27,$EE,$7E,$97,$FC,$9F,$19,$B7
	.dc.b	$2A,$30,$DA,$E1,$3F,$E4,$9F,$05,$F4,$BF,$E2,$F6,$C7,$CD,$3F,$CC
	.dc.b	$EE,$1C,$57,$FA,$3E,$7F,$A8,$78,$F9,$CB,$FC,$C7,$E7,$A3,$18,$FF
	.dc.b	$57,$6E,$1E,$6C,$6E,$3E,$7C,$BD,$64,$9F,$F3,$3E,$1C,$F2,$C4,$7F
	.dc.b	$91,$E7,$F9,$DE,$5F,$19,$FE,$85,$3F,$C7,$7E,$97,$D8,$4A,$F6,$1E
	.dc.b	$17,$FC,$9D,$2F,$8E,$F9,$64,$9F,$8D,$3F,$C9,$FE,$77,$E3,$12,$39
	.dc.b	$E5,$FE,$B1,$8E,$7E,$87,$D4,$24,$FC,$F8,$AF,$A6,$2E,$5F,$D0,$26
	.dc.b	$ED,$F8,$29,$B5,$F1,$E7,$F9,$C7,$CD,$FD,$72,$47,$8B,$64,$7B,$3D
	.dc.b	$9F,$C2,$92,$7F,$01,$1F,$D0,$7E,$2B,$F0,$04,$8E,$4D,$7E,$FD,$23
	.dc.b	$11,$FD,$D2,$30,$FC,$2D,$9F,$43,$F9,$42,$47,$E3,$EB,$FD,$05,$F4
	.dc.b	$BF,$C0,$23,$FB,$93,$5F,$EE,$2D,$7F,$B7,$B5,$FF,$39,$37,$B0,$D2
	.dc.b	$77,$7D,$CF,$BB,$FD,$29,$2B,$FD,$29,$3E,$1E,$0F,$D6,$CE,$67,$F1
	.dc.b	$8F,$9A,$BC,$FF,$01,$FE,$54,$BF,$F4,$45,$7F,$A8,$1A,$FA,$A3,$44
	.dc.b	$74,$D7,$F5,$39,$BE,$E7,$F7,$24,$8F,$2C,$FF,$8B,$3E,$93,$FD,$69
	.dc.b	$DF,$FE,$DC,$F8,$D7,$D2,$FF,$21,$9C,$DA,$72,$FF,$B8,$A3,$31,$8E
	.dc.b	$B7,$D3,$7E,$1A,$BA,$5F,$3E,$5F,$F1,$4D,$F2,$96,$1C,$AB,$58,$F0
	.dc.b	$96,$3D,$39,$7F,$9A,$61,$C7,$0F,$2D,$7F,$0B,$6B,$C8,$FD,$9B,$87
	.dc.b	$D1,$CF,$FE,$93,$7D,$CD,$73,$F0,$9F,$E6,$B9,$FF,$56,$36,$7F,$D4
	.dc.b	$35,$E1,$6E,$5F,$E6,$7F,$A6,$FC,$AF,$57,$A7,$1F,$6D,$69,$EA,$7C
	.dc.b	$BA,$F0,$E1,$EF,$CD,$87,$2A,$D1,$C3,$1F,$A3,$D7,$FC,$C7,$DF,$67
	.dc.b	$FE,$F7,$E3,$FF,$3D,$3F,$EE,$5E,$C7,$D3,$6B,$1F,$E3,$97,$7A,$7F
	.dc.b	$B3,$D6,$AF,$FD,$DE,$DA,$FF,$2D,$35,$FF,$2C,$B5,$FF,$2B,$B5,$FF
	.dc.b	$6D,$37,$A8,$D3,$59,$7F,$49,$FB,$BF,$15,$FF,$54,$6D,$F2,$F8,$4B
	.dc.b	$5B,$AF,$75,$63,$43,$C3,$1E,$F7,$98,$C6,$3F,$EA,$DB,$F6,$CF,$FA
	.dc.b	$E7,$1F,$4B,$5F,$04,$68,$8F,$C2,$8D,$5F,$69,$FE,$29,$F2,$BE,$33
	.dc.b	$CF,$E0,$3E,$7B,$13,$FD,$8D,$ED,$A2,$75,$3F,$CB,$67,$31,$8C,$7F
	.dc.b	$92,$C6,$09,$3B,$3F,$92,$7A,$1F,$0C,$F8,$FE,$EA,$B1,$3F,$E8,$27
	.dc.b	$C3,$19,$8F,$E9,$AD,$8F,$EE,$2D,$7F,$D4,$BB,$3E,$57,$E8,$1D,$B9
	.dc.b	$AD,$FE,$46,$2D,$7F,$B0,$A6,$F0,$FF,$83,$A6,$D7,$FA,$42,$6D,$7F
	.dc.b	$A3,$1F,$29,$23,$CA,$FC,$8B,$FF,$7B,$C7,$E7,$FE,$27,$DF,$F4,$9F
	.dc.b	$83,$6C,$7B,$23,$DE,$47,$FA,$57,$D4,$FE,$24,$FC,$7F,$4B,$E9,$5F
	.dc.b	$41,$9A,$FD,$65,$5C,$DA,$FF,$84,$AF,$75,$A3,$FC,$16,$E2,$77,$1F
	.dc.b	$EE,$B8,$C4,$FF,$AA,$3E,$97,$C3,$18,$9F,$F5,$9C,$7D,$62,$3F,$59
	.dc.b	$F8,$5E,$27,$E8,$4F,$BF,$F4,$B6,$7F,$81,$45,$C3,$FD,$E7,$37,$0F
	.dc.b	$F1,$FF,$91,$FA,$93,$F5,$1F,$4A,$F0,$B1,$8F,$F0,$B9,$B8,$7F,$AC
	.dc.b	$6B,$53,$93,$FC,$B2,$7F,$A1,$CB,$EE,$7E,$E9,$7D,$2B,$E4,$3F,$7B
	.dc.b	$EC,$7F,$18,$AD,$4E,$6B,$FC,$52,$70,$AF,$DA,$34,$CC,$6F,$0F,$F2
	.dc.b	$54,$65,$B1,$97,$FC,$F5,$19,$6C,$65,$FF,$31,$FD,$CF,$F0,$B3,$F8
	.dc.b	$57,$C0,$3E,$0B,$17,$FC,$B6,$6D,$7F,$B4,$EB,$53,$93,$F9,$97,$DC
	.dc.b	$FF,$15,$3F,$67,$F2,$96,$32,$4A,$FC,$29,$FE,$EC,$8D,$3F,$E5,$1F
	.dc.b	$93,$FE,$3C,$7C,$BE,$7F,$E0,$9C,$5F,$E7,$88,$C7,$81,$F1,$A7,$F8
	.dc.b	$45,$7C,$06,$BF,$BF,$6B,$FD,$21,$B8,$7E,$82,$79,$11,$FF,$18,$7C
	.dc.b	$97,$F2,$F3,$F8,$6F,$CE,$D8,$FE,$A1,$16,$FD,$7E,$23,$C2,$5E,$73
	.dc.b	$FE,$A5,$3F,$E0,$6F,$47,$F9,$EF,$A6,$7F,$8C,$FC,$5E,$19,$9F,$F1
	.dc.b	$BB,$E3,$5B,$8C,$1F,$C7,$EF,$C1,$7C,$32,$F1,$D3,$E7,$BB,$9F,$D1
	.dc.b	$FC,$97,$C9,$E3,$1E,$76,$F2,$D5,$A9,$CC,$62,$BF,$AB,$D3,$E9,$FF
	.dc.b	$4A,$FD,$C9,$D5,$70,$BE,$53,$96,$E9,$FD,$D6,$39,$77,$F9,$71,$F8
	.dc.b	$BF,$6B,$30,$F2,$ED,$F3,$79,$31,$87,$BA,$AE,$3F,$3B,$8B,$F8,$64
	.dc.b	$FF,$AE,$37,$57,$D0,$F5,$66,$5D,$2F,$F5,$98,$7C,$E9,$CE,$1F,$29
	.dc.b	$FE,$9B,$DB,$89,$FA,$7E,$7F,$DA,$3F,$B9,$7D,$43,$E2,$71,$97,$C8
	.dc.b	$8C,$3F,$EA,$6E,$24,$63,$2F,$F6,$24,$62,$BF,$B7,$CE,$67,$FD,$E9
	.dc.b	$39,$AE,$53,$99,$FE,$C1,$39,$79,$C6,$5F,$85,$EA,$FE,$40,$FE,$83
	.dc.b	$F5,$1F,$EB,$5B,$F9,$BB,$FF,$88,$73,$FF,$67,$FE,$0C,$F0,$3F,$C0
	.dc.b	$DE,$6A,$31,$7F,$EC,$08,$C4,$EA,$73,$97,$FB,$13,$AD,$EE,$7E,$21
	.dc.b	$B8,$7F,$B4,$B3,$FC,$4E,$5F,$83,$9C,$DF,$E6,$BC,$6F,$D3,$9E,$8C
	.dc.b	$2E,$1C,$5F,$2E,$1C,$2B,$FC,$04,$7F,$9A,$9C,$3F,$E0,$5C,$D1,$88
	.dc.b	$D4,$F2,$25,$7B,$67,$5C,$FF,$E8,$3F,$3F,$FC,$6B,$EC,$9F,$F1,$CF
	.dc.b	$A6,$FA,$9F,$F5,$E6,$1F,$98,$AD,$3F,$AE,$9D,$7C,$0F,$F8,$C5,$B1
	.dc.b	$FE,$EA,$6E,$7E,$A0,$E1,$F6,$B5,$FF,$10,$35,$DD,$F8,$9E,$51,$9C
	.dc.b	$7F,$D3,$9E,$3F,$A8,$FF,$28,$F0,$3E,$A5,$F2,$99,$CB,$F8,$2C,$E6
	.dc.b	$3F,$C9,$0D,$8F,$F8,$02,$DC,$7D,$0D,$7E,$9B,$5F,$E3,$95,$DA,$78
	.dc.b	$B6,$38,$B5,$FB,$78,$7F,$1B,$F1,$7F,$CC,$BE,$29,$FF,$33,$C7,$FA
	.dc.b	$12,$FF,$83,$CE,$5F,$E1,$31,$9A,$F9,$88,$C3,$D9,$AF,$F2,$D6,$BB
	.dc.b	$69,$C6,$3F,$CB,$AE,$0C,$CF,$F1,$27,$F9,$97,$83,$E6,$F5,$FD,$2D
	.dc.b	$AF,$BD,$7D,$67,$99,$23,$35,$E7,$8C,$4F,$AA,$11,$88,$F8,$FC,$0F
	.dc.b	$98,$FC,$B9,$FE,$8C,$BF,$E7,$A7,$E5,$5C,$78,$DF,$E7,$68,$CD,$7C
	.dc.b	$66,$BF,$49,$AF,$F5,$16,$BF,$0E,$39,$12,$3F,$90,$7E,$2F,$FD,$66
	.dc.b	$7F,$A3,$37,$6F,$F5,$A7,$CC,$7E,$5D,$FF,$5C,$78,$09,$D5,$FF,$63
	.dc.b	$AF,$CE,$5F,$E3,$1A,$E7,$97,$F2,$B1,$88,$FB,$DF,$3E,$F5,$BF,$D5
	.dc.b	$AD,$BF,$EC,$3D,$6F,$A8,$FF,$40,$47,$F5,$09,$D5,$FF,$7E,$6E,$9F
	.dc.b	$DF,$1B,$3F,$C3,$6B,$8D,$6F,$0F,$F5,$DF,$D5,$7F,$86,$AB,$FB,$C3
	.dc.b	$67,$F2,$5D,$FF,$E9,$3C,$7F,$C0,$F3,$9E,$33,$4C,$4E,$EF,$9F,$A1
	.dc.b	$FD,$D3,$0F,$F5,$17,$1F,$FA,$8F,$E0,$3F,$C2,$71,$FC,$22,$B1,$E8
	.dc.b	$7E,$42,$AC,$E4,$FF,$5B,$46,$5A,$66,$B9,$BF,$CB,$A7,$8B,$FC,$4A
	.dc.b	$AB,$E2,$3F,$C2,$55,$FC,$4A,$3F,$D6,$DD,$1C,$A7,$4D,$ED,$FF,$7C
	.dc.b	$39,$85,$F4,$9F,$F7,$E6,$19,$8C,$5F,$82,$B6,$7A,$0F,$21,$4F,$24
	.dc.b	$F7,$8E,$13,$FB,$29,$F6,$14,$E6,$65,$D7,$D3,$BE,$9D,$F5,$38,$6D
	.dc.b	$7E,$F2,$75,$39,$38,$FC,$EF,$FA,$B2,$74,$D7,$FA,$04,$A3,$11,$FD
	.dc.b	$72,$31,$C3,$C8,$F9,$7A,$E7,$EA,$7F,$B0,$ED,$F5,$A9,$18,$9E,$58
	.dc.b	$78,$6B,$4D,$8E,$58,$74,$3F,$DD,$97,$C9,$C0,$7C,$6F,$DE,$37,$3C
	.dc.b	$DE,$37,$E1,$69,$CA,$78,$DF,$8B,$0C,$56,$EB,$B3,$6B,$78,$7A,$AB
	.dc.b	$75,$0F,$11,$E5,$C3,$C5,$C4,$87,$CC,$7D,$F8,$C3,$63,$27,$F3,$02
	.dc.b	$9F,$D2,$0A,$7F,$54,$29,$FE,$73,$29,$FD,$34,$BC,$EF,$A9,$FE,$69
	.dc.b	$1F,$E2,$96,$BF,$DF,$E3,$5D,$DF,$93,$6C,$6E,$31,$3C,$23,$FC,$3F
	.dc.b	$3A,$79,$46,$5F,$F1,$06,$7F,$D1,$9F,$F1,$C3,$5F,$03,$5E,$6D,$7F
	.dc.b	$D1,$DD,$3C,$DF,$E5,$77,$F2,$63,$18,$9D,$60,$FC,$1F,$FD,$1F,$7F
	.dc.b	$E4,$76,$BE,$F6,$BE,$A6,$BF,$E3,$76,$A8,$E9,$EB,$7C,$59,$CC,$62
	.dc.b	$7F,$B3,$F1,$BE,$7F,$EF,$36,$33,$18,$3E,$77,$D9,$7F,$45,$8F,$F2
	.dc.b	$E3,$5F,$A8,$D7,$E7,$B5,$FF,$26,$35,$F3,$36,$8D,$35,$97,$FA,$BF
	.dc.b	$A3,$F2,$5E,$0F,$F8,$FA,$FF,$BF,$B5,$FF,$0B,$B5,$FB,$FE,$23,$D6
	.dc.b	$FC,$33,$63,$FC,$E8,$D7,$F1,$8D,$78,$C7,$02,$3C,$1F,$D2,$73,$BE
	.dc.b	$AA,$92,$7F,$C2,$71,$97,$F9,$47,$D4,$3E,$27,$F9,$43,$63,$31,$82
	.dc.b	$38,$21,$FE,$22,$8E,$0F,$F8,$93,$EA,$5E,$D7,$F3,$36,$C7,$F3,$76
	.dc.b	$BB,$8F,$0F,$93,$23,$CA,$D7,$B3,$5F,$F3,$75,$7F,$B8,$1D,$9F,$E7
	.dc.b	$3F,$9A,$FF,$6B,$3F,$E7,$06,$BF,$E7,$06,$BF,$E7,$66,$A8,$7B,$DF
	.dc.b	$10,$91,$F2,$1A,$FC,$16,$BE,$47,$F8,$17,$D4,$BE,$9E,$92,$7F,$D6
	.dc.b	$51,$96,$FA,$3C,$08,$E0,$FA,$FF,$39,$F5,$88,$E0,$E7,$F4,$DE,$02
	.dc.b	$38,$3F,$B3,$7D,$8B,$CA,$FE,$CC,$D8,$CC,$60,$8E,$08,$7F,$98,$23
	.dc.b	$83,$FE,$62,$F6,$9F,$27,$FD,$C5,$B1,$FD,$D5,$AE,$E3,$F9,$D7,$BD
	.dc.b	$FD,$56,$3F,$CF,$AD,$7F,$D1,$2D,$7F,$A1,$75,$23,$11,$FE,$83,$6C
	.dc.b	$67,$C9,$EA,$49,$FF,$39,$F9,$AF,$A5,$FE,$94,$D8,$CC,$60,$8E,$08
	.dc.b	$7F,$A4,$A3,$83,$FE,$BF,$F4,$1F,$62,$F9,$6F,$F0,$A4,$61,$B1,$99
	.dc.b	$FF,$44,$CE,$5B,$19,$7A,$35,$FF,$72,$C6,$DC,$CF,$F8,$87,$BD,$F4
	.dc.b	$7F,$E2,$16,$CE,$63,$06,$7E,$99,$F2,$7C,$5B,$3F,$E9,$CA,$46,$23
	.dc.b	$FD,$B9,$18,$9E,$04,$8E,$13,$B7,$FD,$F7,$3B,$7F,$D4,$7B,$7F,$89
	.dc.b	$5B,$E7,$FE,$08,$8E,$08,$7E,$11,$CB,$FE,$B0,$F8,$5F,$E1,$EF,$FB
	.dc.b	$BD,$AF,$F8,$1C,$38,$7F,$C1,$4D,$73,$E0,$7F,$91,$E3,$04,$7F,$C9
	.dc.b	$5F,$4D,$FE,$04,$93,$49,$E8,$FE,$6B,$D5,$93,$EE,$6B,$F9,$C6,$BF
	.dc.b	$9C,$6B,$FE,$1D,$CB,$EC,$F9,$7F,$5C,$91,$F4,$F6,$FE,$C4,$FD,$26
	.dc.b	$BF,$39,$AF,$36,$A9,$EA,$FF,$99,$F2,$FE,$3E,$4F,$5B,$E1,$12,$34
	.dc.b	$D7,$E8,$3F,$E6,$FF,$93,$7B,$0F,$95,$C3,$6B,$FD,$CB,$3A,$9C,$9F
	.dc.b	$B0,$FA,$59,$3F,$B4,$6B,$FA,$D6,$BF,$6F,$A9,$D0,$9C,$CF,$F0,$4E
	.dc.b	$3F,$F1,$CD,$7F,$05,$E0,$FE,$1A,$49,$FC,$8B,$5F,$E1,$ED,$7F,$C6
	.dc.b	$7F,$37,$FD,$01,$FF,$7E,$FA,$28,$C4,$7F,$8F,$DA,$F0,$F1,$BF,$8D
	.dc.b	$FC,$AC,$8F,$CF,$B5,$E1,$F4,$7F,$E4,$5A,$34,$91,$E0,$8C,$B6,$32
	.dc.b	$F2,$35,$F4,$3C,$C8,$E0,$FD,$7F,$C6,$FF,$2E,$23,$83,$B5,$78,$C8
	.dc.b	$E0,$FD,$BF,$81,$FC,$D8,$8E,$08,$7F,$38,$72,$D3,$27,$F3,$A7,$2D
	.dc.b	$3E,$6F,$E4,$CF,$9F,$FF,$34,$36,$33,$18,$3F,$A3,$F0,$48,$FE,$16
	.dc.b	$D7,$F8,$01,$F9,$4F,$0B,$FC,$A3,$CB,$FC,$D3,$DA,$FF,$63,$36,$76
	.dc.b	$6E,$33,$3C,$08,$E0,$FE,$5D,$F4,$7F,$C7,$67,$F7,$0D,$7F,$5C,$DE
	.dc.b	$A3,$9B,$FE,$8B,$E9,$76,$25,$78,$A3,$2D,$C3,$FC,$4A,$FF,$1B,$8C
	.dc.b	$BE,$E6,$B9,$8D,$BE,$4E,$8F,$39,$1F,$E7,$97,$FB,$57,$F9,$DB,$5F
	.dc.b	$E6,$ED,$79,$1C,$23,$E4,$FC,$EB,$E6,$29,$27,$D5,$19,$6C,$65,$F4
	.dc.b	$46,$5B,$1F,$D6,$7C,$6F,$F0,$5C,$7F,$48,$6B,$FC,$95,$AF,$F1,$B6
	.dc.b	$AF,$A2,$7C,$97,$F1,$48,$7B,$DF,$E6,$B9,$CE,$9F,$EC,$B8,$C5,$6D
	.dc.b	$FF,$39,$75,$7F,$83,$7E,$8F,$FA,$54,$6A,$B8,$7B,$9F,$D3,$70,$E7
	.dc.b	$8F,$68,$C6,$3C,$DF,$F6,$C6,$3C,$E7,$89,$FE,$DB,$AE,$2F,$E6,$1A
	.dc.b	$F2,$BF,$CA,$BE,$DB,$F6,$3B,$EC,$FF,$46,$3C,$DA,$FC,$36,$D7,$F9
	.dc.b	$6B,$DE,$E6,$D9,$F4,$1C,$63,$CB,$83,$E8,$FF,$8A,$F8,$9E,$8A,$E9
	.dc.b	$56,$F9,$69,$F3,$71,$FF,$51,$3F,$2B,$0E,$37,$D8,$E2,$78,$39,$5F
	.dc.b	$A4,$BE,$03,$FC,$EB,$18,$6C,$67,$3F,$F0,$A3,$6B,$F4,$5F,$A7,$BE
	.dc.b	$AF,$FC,$FB,$C0,$8C,$5F,$EB,$5B,$5C,$2B,$71,$88,$FD,$73,$51,$FD
	.dc.b	$79,$C3,$63,$CF,$C7,$FE,$0E,$AF,$F1,$43,$67,$FC,$4E,$D7,$F3,$9E
	.dc.b	$4F,$02,$4F,$06,$C7,$EF,$7E,$D1,$F2,$BE,$C6,$D7,$F0,$76,$CE,$E7
	.dc.b	$44,$9D,$36,$3C,$7F,$39,$E8,$24,$FA,$1E,$DF,$31,$23,$B3,$67,$C8
	.dc.b	$D7,$F8,$A5,$A3,$11,$FD,$CF,$E0,$BF,$D5,$FE,$5F,$F9,$DD,$BE,$61
	.dc.b	$A7,$FA,$D3,$A1,$E2,$BE,$0F,$0D,$CF,$FA,$3C,$60,$E7,$F5,$32,$3F
	.dc.b	$C9,$AD,$CB,$FD,$16,$D7,$FD,$16,$D7,$FC,$10,$D7,$FD,$14,$DB,$FF
	.dc.b	$00,$7B,$1E,$5F,$2B,$FC,$CF,$7F,$DD,$63,$07,$B3,$8A,$47,$F9,$79
	.dc.b	$B8,$7F,$81,$FE,$8F,$91,$27,$FB,$F3,$D7,$E8,$A4,$7F,$80,$5B,$3F
	.dc.b	$E0,$F6,$BF,$E1,$66,$BF,$E2,$1F,$0B,$E6,$92,$7F,$CD,$EF,$CD,$F9
	.dc.b	$DF,$F1,$BC,$7F,$87,$1B,$3F,$DA,$33,$46,$23,$FC,$D5,$18,$E9,$FB
	.dc.b	$3F,$2F,$2F,$FC,$F6,$DA,$FF,$5C,$B5,$FF,$5C,$B5,$FF,$5C,$FD,$03
	.dc.b	$E5,$BF,$0A,$DA,$FE,$E7,$CA,$8C,$57,$F7,$C8,$C5,$70,$25,$70,$AD
	.dc.b	$D7,$F9,$1D,$AF,$EE,$BD,$77,$DB,$FF,$B0,$6B,$FC,$AA,$D9,$FF,$04
	.dc.b	$CE,$BD,$25,$69,$B3,$FE,$08,$F1,$1E,$A6,$23,$FC,$23,$18,$AF,$E7
	.dc.b	$2D,$8E,$13,$B7,$FA,$6B,$5F,$EA,$8D,$7F,$AC,$B6,$39,$47,$FA,$B3
	.dc.b	$E9,$9E,$3F,$F2,$86,$DF,$FA,$E5,$B3,$FE,$B8,$6B,$C6,$78,$12,$78
	.dc.b	$3F,$B6,$F8,$69,$1F,$EB,$86,$CF,$FB,$D9,$AF,$FB,$D9,$AF,$F2,$88
	.dc.b	$7D,$33,$D3,$F7,$46,$1B,$19,$8F,$F1,$99,$93,$FD,$1D,$F1,$5F,$0F
	.dc.b	$53,$49,$1A,$25,$FF,$A4,$23,$2F,$E2,$7E,$D7,$C8,$D3,$F8,$AB,$87
	.dc.b	$FD,$1C,$D7,$FA,$0F,$12,$31,$08,$C3,$C8,$95,$FE,$E9,$F6,$5E,$DF
	.dc.b	$F5,$C6,$C6,$63,$04,$70,$49,$FF,$77,$CF,$07,$FD,$81,$E3,$FF,$66
	.dc.b	$F7,$BF,$A0,$C7,$FB,$06,$30,$7E,$2F,$E1,$7F,$55,$DD,$79,$CD,$26
	.dc.b	$3F,$E6,$29,$DB,$67,$6D,$9D,$DF,$EC,$E7,$C2,$FE,$F2,$FF,$76,$4A
	.dc.b	$FD,$A3,$63,$83,$5F,$9F,$E9,$7D,$22,$4F,$F0,$2D,$51,$89,$FB,$4D
	.dc.b	$78,$CF,$E2,$3C,$F7,$1F,$E3,$9B,$19,$8C,$1F,$AD,$5F,$03,$48,$FF
	.dc.b	$8C,$0C,$95,$CD,$7E,$59,$AF,$F2,$6E,$F7,$BF,$FC,$BC,$95,$FC,$BD
	.dc.b	$FD,$E7,$7F,$F8,$92,$BF,$CF,$5E,$FE,$09,$5F,$CD,$78,$FF,$9C,$11
	.dc.b	$C1,$FC,$E7,$8F,$F9,$E1,$1C,$1F,$E9,$EF,$5B,$B9,$1C,$12,$FF,$D2
	.dc.b	$11,$96,$C7,$FA,$93,$67,$C2,$FE,$81,$B3,$FC,$5D,$B3,$B9,$FD,$2C
	.dc.b	$B9,$1F,$5D,$FC,$6D,$B5,$FE,$5C,$6C,$FE,$46,$74,$49,$D3,$63,$FD
	.dc.b	$7B,$6F,$A2,$78,$9E,$36,$DF,$A5,$B5,$C1,$AF,$FB,$1B,$0F,$E1,$73
	.dc.b	$E3,$EA,$F3,$11,$C1,$27,$F4,$D1,$97,$FA,$A7,$BF,$FC,$B1,$FE,$0E
	.dc.b	$D7,$F8,$33,$7A,$CE,$7C,$67,$DE,$3E,$63,$FC,$44,$DC,$7F,$DA,$B3
	.dc.b	$A9,$C9,$23,$33,$BC,$3F,$BD,$35,$FE,$F2,$D7,$F8,$DE,$1F,$D7,$27
	.dc.b	$FB,$7B,$63,$FB,$73,$5F,$F3,$09,$FC,$7D,$C3,$FE,$9C,$F8,$A7,$D3
	.dc.b	$7F,$8E,$5B,$5E,$86,$CF,$F8,$35,$AF,$F8,$31,$AF,$39,$E4,$49,$FF
	.dc.b	$01,$F9,$3F,$E4,$19,$FF,$01,$35,$F2,$DF,$F7,$CF,$47,$FC,$9F,$39
	.dc.b	$9F,$F5,$8F,$E2,$6F,$85,$FF,$27,$37,$0F,$9A,$D9,$FF,$11,$35,$FF
	.dc.b	$10,$B5,$E7,$3C,$89,$3F,$E1,$83,$C1,$1F,$D1,$5B,$87,$F8,$2B,$0F
	.dc.b	$F3,$1C,$6B,$57,$C7,$3E,$43,$FC,$C4,$D9,$FC,$0B,$70,$FF,$1B,$35
	.dc.b	$FF,$1A,$FD,$03,$EB,$BF,$CD,$4D,$AF,$C4,$B6,$7F,$C7,$CD,$78,$CF
	.dc.b	$02,$4F,$07,$FC,$F3,$C9,$FA,$89,$FE,$C6,$D7,$FB,$13,$73,$FE,$01
	.dc.b	$CD,$C0,$97,$FE,$3C,$9D,$1C,$6B,$FD,$DA,$3E,$61,$F4,$1F,$BE,$AF
	.dc.b	$F5,$8C,$62,$7F,$C5,$A6,$E7,$4D,$33,$1B,$7F,$D7,$3D,$3F,$E3,$88
	.dc.b	$C4,$7F,$C0,$13,$81,$AA,$EF,$5F,$E3,$D8,$C9,$FD,$4A,$32,$F6,$8F
	.dc.b	$0C,$F4,$87,$A1,$A7,$A1,$C8,$C9,$D2,$FC,$95,$DE,$BC,$B1,$89,$EC
	.dc.b	$7E,$C9,$C3,$F2,$0F,$F2,$69,$FB,$63,$E2,$98,$31,$F3,$DF,$D7,$BE
	.dc.b	$87,$51,$FE,$DC,$8C,$38,$6B,$C3,$C6,$4F,$08,$4B,$E5,$1F,$43,$98
	.dc.b	$C3,$63,$33,$D5,$AF,$91,$CE,$3F,$5E,$9D,$5F,$C4,$23,$0D,$CF,$F8
	.dc.b	$9C,$F6,$BF,$B2,$DA,$DC,$7D,$18,$E6,$DB,$EF,$5C,$9B,$1F,$0E,$FB
	.dc.b	$34,$E1,$84,$7B,$C6,$9F,$86,$DB,$F1,$5F,$5A,$F1,$4E,$4D,$5F,$83
	.dc.b	$81,$87,$F8,$20,$FF,$50,$94,$FF,$50,$14,$FE,$3E,$53,$FC,$AA,$53
	.dc.b	$FC,$A8,$57,$AF,$8F,$F6,$5B,$1F,$D3,$3E,$87,$FA,$83,$7B,$BE,$F7
	.dc.b	$42,$73,$3F,$EC,$56,$BF,$EC,$46,$BE,$A8,$E1,$3C,$A7,$32,$9C,$BF
	.dc.b	$D7,$34,$7D,$27,$81,$B1,$98,$C1,$E9,$FA,$2E,$78,$9A,$4C,$7D,$3F
	.dc.b	$45,$FC,$FE,$FF,$B4,$B5,$FF,$6F,$35,$FF,$6E,$B5,$F5,$35,$F4,$37
	.dc.b	$13,$4D,$5F,$F0,$0F,$97,$F9,$DC,$A6,$92,$FF,$BD,$FD,$17,$FB,$85
	.dc.b	$BE,$EF,$D0,$8C,$47,$FB,$CD,$AF,$FB,$C9,$AF,$CF,$6B,$F3,$5B,$89
	.dc.b	$CE,$0F,$8E,$FF,$35,$D7,$FB,$B9,$AF,$FB,$B5,$AF,$F6,$C6,$BF,$77
	.dc.b	$D2,$7C,$8F,$E1,$9B,$1F,$E1,$66,$BF,$C4,$F0,$46,$23,$FC,$17,$18
	.dc.b	$9E,$44,$8E,$4F,$E5,$F9,$DF,$45,$49,$3F,$C6,$B1,$FE,$91,$1F,$C5
	.dc.b	$61,$F4,$2F,$9C,$C3,$63,$F8,$D3,$5F,$E4,$6D,$7F,$95,$35,$FF,$10
	.dc.b	$B5,$FD,$BB,$5E,$F1,$E1,$FA,$2F,$F6,$F4,$7F,$27,$6B,$FC,$C1,$AF
	.dc.b	$F3,$76,$BD,$BC,$6F,$39,$1F,$3B,$5F,$3B,$5F,$3B,$FC,$D7,$BD,$F1
	.dc.b	$6F,$07,$0D,$AF,$F3,$E3,$5F,$F3,$5B,$5F,$83,$F0,$3F,$20,$8F,$C8
	.dc.b	$6B,$F2,$1A,$F0,$3F,$CA,$FF,$42,$FA,$2F,$E1,$6D,$8C,$C6,$0C,$FA
	.dc.b	$32,$38,$B5,$FA,$BF,$45,$F6,$88,$FD,$A7,$AF,$99,$23,$C2,$D7,$FA
	.dc.b	$FF,$D1,$7F,$B1,$1F,$EB,$4D,$73,$EE,$FE,$26,$31,$1F,$DA,$DA,$FF
	.dc.b	$98,$9A,$FF,$38,$6A,$6F,$9C,$FA,$2F,$CA,$36,$7F,$B5,$35,$FE,$E6
	.dc.b	$D7,$FC,$B7,$F2,$3F,$D6,$1F,$EF,$9C,$C8,$C4,$6B,$C8,$FE,$79,$18
	.dc.b	$8F,$E9,$73,$B9,$C9,$23,$F5,$13,$B5,$2F,$A1,$7C,$5F,$F4,$76,$CF
	.dc.b	$EB,$DA,$F9,$1A,$FF,$9C,$9A,$F8,$9A,$F6,$6C,$F7,$9F,$F6,$17,$B0
	.dc.b	$F8,$EA,$4B,$FF,$0B,$FA,$3F,$EA,$88,$C4,$7F,$AE,$67,$83,$A6,$B9
	.dc.b	$8F,$F0,$C7,$82,$F8,$7F,$E0,$CD,$9C,$C6,$0C,$FB,$AF,$8F,$E2,$D9
	.dc.b	$FF,$14,$D7,$FB,$92,$72,$48,$CC,$ED,$FF,$1A,$4E,$DC,$CF,$D8,$F4
	.dc.b	$24,$7D,$46,$BF,$80,$DB,$FC,$5C,$E4,$86,$63,$F9,$07,$CB,$7F,$13
	.dc.b	$27,$81,$82,$4F,$F8,$E6,$32,$D8,$FE,$51,$F4,$5F,$E0,$98,$34,$9E
	.dc.b	$97,$E3,$BD,$69,$3F,$8A,$6B,$E7,$6B,$F9,$A6,$BF,$E1,$EC,$FE,$97
	.dc.b	$99,$F4,$09,$1F,$33,$6F,$ED,$0D,$8C,$BF,$2A,$32,$D8,$F8,$3E,$0C
	.dc.b	$8F,$73,$5F,$F2,$FE,$7F,$C7,$5F,$F6,$DB,$5F,$F1,$8B,$53,$9F,$8B
	.dc.b	$7C,$33,$E7,$FF,$CC,$51,$86,$C6,$6B,$FC,$DB,$39,$7F,$51,$F4,$3F
	.dc.b	$AB,$23,$82,$1F,$AB,$72,$D3,$F6,$FA,$FE,$F8,$8E,$08,$7E,$1D,$CB
	.dc.b	$4F,$F3,$FF,$9A,$F9,$47,$8D,$FE,$3E,$6C,$E9,$AF,$F7,$96,$BF,$C5
	.dc.b	$3E,$23,$F8,$79,$1F,$E1,$6F,$3F,$89,$91,$CD,$AF,$36,$BF,$E0,$29
	.dc.b	$46,$23,$FC,$95,$0F,$92,$F3,$12,$3C,$CF,$F2,$0F,$8A,$91,$FC,$71
	.dc.b	$AF,$A6,$7D,$C4,$7D,$CF,$F2,$8E,$E9,$1F,$C9,$1A,$FE,$C5,$AF,$F3
	.dc.b	$46,$BF,$CB,$BA,$7F,$8D,$92,$33,$1F,$E5,$FF,$96,$7D,$17,$F9,$B1
	.dc.b	$B1,$F5,$1A,$FF,$89,$DA,$FF,$42,$DF,$F9,$E9,$1F,$E7,$2F,$FA,$CF
	.dc.b	$85,$DF,$FE,$57,$6F,$C2,$E6,$30,$47,$04,$AF,$F6,$1C,$F0,$7F,$8E
	.dc.b	$FD,$17,$F1,$97,$FD,$2A,$7F,$92,$5C,$3F,$BB,$6B,$9E,$FF,$07,$CA
	.dc.b	$48,$EB,$87,$F8,$79,$FD,$5B,$6F,$D3,$3B,$9F,$F6,$D4,$61,$E3,$1F
	.dc.b	$CE,$AF,$F5,$D1,$FC,$B1,$AF,$9B,$B1,$18,$8F,$37,$47,$12,$3E,$E6
	.dc.b	$BE,$0F,$A2,$7D,$17,$CD,$6C,$7C,$66,$BB,$8D,$11,$D3,$5F,$E8,$3E
	.dc.b	$27,$F8,$23,$0F,$EB,$DE,$67,$F5,$48,$C1,$FE,$0A,$FA,$27,$C8,$FF
	.dc.b	$10,$B7,$BF,$FE,$6C,$9C,$ED,$FE,$C8,$8C,$56,$DF,$F3,$9F,$6F,$F8
	.dc.b	$1F,$E6,$7E,$CC,$6A,$B8,$7B,$BF,$73,$0E,$78,$F6,$8C,$63,$CD,$FF
	.dc.b	$5E,$74,$B1,$E7,$1A,$C3,$F0,$B7,$4A,$30,$EA,$E9,$AF,$87,$BF,$FD
	.dc.b	$1A,$ED,$B1,$FC,$5B,$D8,$E7,$39,$6C,$66,$3D,$07,$FB,$A9,$C6,$7F
	.dc.b	$0B,$E8,$BF,$86,$F8,$BF,$02,$B5,$87,$FA,$8A,$BE,$63,$72,$F2,$63
	.dc.b	$BD,$98,$F2,$C3,$B9,$E3,$3C,$3C,$2F,$A2,$FF,$4F,$E7,$FE,$06,$6C
	.dc.b	$FF,$80,$9A,$FE,$5D,$A8,$FE,$9F,$F4,$4F,$A2,$FE,$B1,$BA,$31,$7F
	.dc.b	$B8,$6D,$7E,$A1,$AF,$EE,$27,$85,$72,$9C,$CF,$79,$CD,$73,$AF,$EC
	.dc.b	$D8,$7F,$8C,$FC,$8B,$C9,$FD,$92,$BF,$C4,$2D,$C3,$F3,$B3,$A2,$4E
	.dc.b	$9B,$1F,$BC,$FA,$17,$CD,$7B,$1B,$5F,$C4,$9B,$3B,$9D,$12,$74,$D8
	.dc.b	$FD,$6F,$D1,$79,$89,$3E,$67,$B7,$D1,$48,$EC,$D9,$F2,$35,$F9,$16
	.dc.b	$8C,$47,$F3,$FF,$80,$FF,$05,$7A,$DF,$E7,$36,$F4,$FF,$AC,$30,$FF
	.dc.b	$36,$C7,$F8,$3F,$8D,$E2,$BE,$07,$0D,$AF,$E8,$51,$83,$9F,$D1,$48
	.dc.b	$FF,$3B,$B7,$2F,$F2,$23,$5F,$F4,$3B,$5F,$EA,$8D,$7F,$CF,$B8,$A3
	.dc.b	$18,$7F,$80,$FC,$5F,$37,$95,$FE,$5F,$C3,$FC,$05,$18,$3D,$BC,$52
	.dc.b	$3F,$D1,$2D,$C7,$FC,$0D,$F3,$FC,$89,$3F,$E0,$B7,$AF,$D1,$48,$FF
	.dc.b	$07,$36,$7F,$C1,$CD,$7F,$C0,$8D,$7F,$B7,$78,$1F,$38,$93,$FE,$9A
	.dc.b	$7E,$77,$CF,$7F,$8D,$A3,$FC,$3E,$D9,$FE,$2B,$9A,$31,$1F,$D7,$23
	.dc.b	$1D,$3F,$6B,$E5,$A5,$FF,$A9,$9B,$5F,$E7,$36,$BF,$EB,$66,$BF,$EB
	.dc.b	$0F,$A0,$7C,$77,$E1,$9B,$5F,$C9,$39,$51,$8A,$FF,$5E,$B5,$FE,$DF
	.dc.b	$1A,$AE,$44,$AE,$55,$BA,$FE,$11,$EB,$3E,$6F,$FD,$73,$5F,$E5,$26
	.dc.b	$CF,$F3,$69,$D7,$A5,$FE,$63,$6D,$7F,$2E,$F1,$BF,$C6,$0D,$F5,$3F
	.dc.b	$CC,$D1,$8B,$FE,$7A,$D7,$FA,$0B,$5F,$E9,$F1,$C0,$93,$FE,$A8,$FA
	.dc.b	$2F,$F2,$8B,$7C,$8F,$E1,$F1,$8A,$FF,$53,$B5,$FE,$F8,$D7,$FD,$14
	.dc.b	$D5,$E1,$23,$0D,$9F,$DA,$F9,$72,$FF,$DC,$2D,$BF,$F5,$73,$5F,$F7
	.dc.b	$8B,$5F,$E6,$90,$FA,$27,$A3,$FE,$8B,$8C,$36,$33,$1F,$D8,$8C,$9F
	.dc.b	$E8,$8F,$89,$FF,$46,$11,$C1,$2B,$FD,$13,$19,$6C,$7E,$27,$E8,$BC
	.dc.b	$71,$FD,$C3,$8D,$18,$8F,$C4,$B5,$FE,$99,$1A,$7F,$8C,$C6,$5F,$F8
	.dc.b	$02,$D8,$D4,$F7,$9F,$E8,$7E,$C3,$DA,$A4,$9F,$F7,$1C,$65,$B1,$99
	.dc.b	$FF,$70,$CE,$5B,$1F,$D3,$BD,$DF,$E7,$F1,$FE,$39,$6B,$FE,$C7,$6B
	.dc.b	$F7,$9B,$F0,$7F,$AA,$EE,$FF,$5E,$37,$0F,$F0,$8B,$6B,$FC,$2A,$D7
	.dc.b	$FC,$B6,$D7,$FB,$83,$53,$4C,$63,$FB,$8F,$2B,$F7,$84,$AF,$D9,$B6
	.dc.b	$BF,$76,$D7,$F7,$7D,$1E,$AA,$FA,$FE,$87,$D2,$25,$FD,$D6,$D7,$DD
	.dc.b	$D5,$18,$8F,$BA,$FE,$47,$E4,$5C,$54,$97,$F9,$18,$CB,$FE,$2A,$1F
	.dc.b	$01,$FE,$E1,$69,$97,$04,$70,$4B,$FC,$E4,$7F,$30,$F7,$E4,$FF,$30
	.dc.b	$6C,$FF,$A3,$3D,$FF,$F7,$77,$81,$FE,$29,$8F,$F3,$EC,$60,$FE,$73
	.dc.b	$C5,$2F,$F9,$CB,$67,$F9,$CF,$84,$E4,$FE,$7E,$4A,$FE,$7E,$F6,$F5
	.dc.b	$E4,$7F,$A9,$5B,$5F,$E9,$E6,$BF,$E8,$E6,$BF,$E3,$1D,$9F,$09,$FA
	.dc.b	$56,$C2,$BF,$CB,$B1,$86,$CF,$EA,$E7,$F1,$3F,$45,$B7,$F2,$6D,$7F
	.dc.b	$26,$D7,$F2,$6D,$7F,$26,$D7,$9C,$FF,$98,$B9,$53,$96,$C7,$93,$E8
	.dc.b	$9E,$2F,$3C,$51,$89,$D4,$E6,$31,$87,$EA,$3A,$BA,$12,$BC,$0D,$8D
	.dc.b	$46,$08,$FF,$13,$F7,$FF,$9A,$73,$FF,$B5,$3B,$BF,$C4,$F1,$83,$F9
	.dc.b	$C7,$8E,$FA,$27,$CB,$FF,$89,$DB,$8F,$FA,$2A,$75,$39,$24,$66,$77
	.dc.b	$8F,$F7,$E6,$BF,$DF,$5A,$FF,$22,$C7,$ED,$4F,$F7,$36,$C7,$F2,$B6
	.dc.b	$BF,$DC,$9A,$FF,$9D,$6D,$F1,$4F,$A2,$FF,$1F,$36,$FD,$0D,$AF,$F0
	.dc.b	$8B,$5F,$F0,$83,$5E,$8D,$55,$DA,$BA,$9F,$E5,$0F,$2B,$FC,$91,$3F
	.dc.b	$D0,$5A,$FF,$82,$1A,$FF,$40,$6B,$D3,$E8,$9F,$0B,$FE,$55,$6E,$3F
	.dc.b	$3D,$B3,$FE,$28,$6B,$FE,$27,$6B,$F0,$DA,$F5,$9E,$87,$F9,$9A,$7C
	.dc.b	$39,$7E,$A7,$D5,$C1,$23,$EF,$7C,$6B,$E4,$3F,$CD,$0D,$8C,$C6,$0F
	.dc.b	$F3,$67,$D1,$3E,$8A,$92,$7F,$11,$19,$6C,$67,$2F,$F2,$1C,$E5,$B1
	.dc.b	$FE,$7C,$E8,$FD,$1C,$7F,$8C,$9A,$FF,$6A,$6A,$CB,$FA,$3F,$3F,$12
	.dc.b	$61,$FE,$3F,$AD,$CE,$A3,$FD,$B7,$5F,$EF,$91,$F3,$2F,$A1,$FF,$75
	.dc.b	$F5,$7F,$09,$8C,$4F,$F8,$B8,$DC,$EA,$BF,$D9,$13,$A9,$FF,$7F,$9A
	.dc.b	$3F,$D9,$87,$F4,$A8,$DC,$FF,$8E,$A3,$47,$F2,$18,$CB,$DE,$7A,$3C
	.dc.b	$4D,$9E,$97,$35,$E7,$AE,$91,$8A,$E7,$5E,$48,$C3,$C8,$7A,$3A,$7B
	.dc.b	$47,$81,$C9,$E0,$BE,$CF,$C3,$3F,$CB,$07,$CC,$9F,$9A,$F0,$3E,$08
	.dc.b	$F9,$FF,$EC,$3F,$45,$88,$FF,$78,$46,$1C,$35,$E2,$D7,$E5,$AF,$93
	.dc.b	$7D,$17,$39,$ED,$18,$79,$B5,$E6,$D7,$9B,$BF,$95,$FB,$55,$FC,$3A
	.dc.b	$42,$33,$D5,$F7,$1D,$D7,$1D,$79,$E1,$C2,$77,$76,$7D,$37,$C5,$DC
	.dc.b	$75,$6E,$9C,$23,$F0,$B8,$F8,$DB,$8F,$B6,$7C,$38,$F5,$9F,$97,$87
	.dc.b	$56,$99,$3E,$2C,$5B,$F9,$4D,$9E,$13,$F5,$DB,$3A,$8F,$31,$FB,$13
	.dc.b	$FC,$A8,$7C,$69,$FE,$E1,$B7,$C8,$F1,$BF,$43,$18,$9D,$9F,$63,$7F
	.dc.b	$F3,$2A,$F2,$77,$18,$AB,$3D,$DB,$7D,$63,$C1,$3C,$AB,$3E,$4B,$E1
	.dc.b	$E9,$2B,$67,$F3,$EF,$A2,$FE,$C2,$E3,$05,$71,$8C,$35,$F4,$1F,$C4
	.dc.b	$5C,$65,$FE,$0F,$EF,$5F,$4A,$DD,$BE,$87,$11,$FE,$CE,$6B,$FE,$86
	.dc.b	$6B,$FC,$75,$B1,$FE,$21,$F9,$1F,$83,$A1,$A7,$C4,$9F,$06,$3F,$4B
	.dc.b	$95,$F2,$BF,$A5,$C3,$F7,$ED,$8C,$E3,$FE,$E5,$6B,$FE,$9D,$6B,$FE
	.dc.b	$E4,$6B,$FE,$9C,$6B,$FC,$CD,$B3,$FE,$E3,$FA,$DF,$AF,$E1,$1A,$7D
	.dc.b	$A9,$EF,$87,$E2,$3D,$AF,$F3,$7E,$1F,$EF,$76,$BF,$EA,$F6,$BF,$EF
	.dc.b	$66,$BF,$EA,$E6,$BF,$D1,$5B,$87,$FB,$5B,$AB,$E1,$46,$21,$E6,$FE
	.dc.b	$31,$B5,$F8,$36,$BF,$6A,$7C,$77,$F9,$EE,$77,$D0,$7C,$D6,$D7,$F1
	.dc.b	$DB,$FE,$9B,$39,$24,$66,$76,$FF,$B2,$DA,$FF,$48,$9E,$2F,$3F,$27
	.dc.b	$C7,$7F,$AE,$6F,$0F,$FB,$39,$B3,$FD,$96,$FA,$D7,$F9,$87,$E7,$FF
	.dc.b	$DD,$33,$FC,$9D,$AF,$F3,$36,$BF,$CE,$1A,$FF,$5A,$6B,$FC,$91,$A9
	.dc.b	$A6,$B0,$7B,$EF,$8D,$FD,$CB,$6B,$F6,$0D,$7F,$5A,$73,$F7,$BF,$80
	.dc.b	$EE,$F8,$BF,$F3,$A3,$75,$FE,$80,$DA,$FF,$42,$CE,$89,$3A,$6C,$7F
	.dc.b	$51,$DD,$F6,$0F,$9B,$FE,$1F,$18,$6C,$66,$7F,$84,$E1,$DD,$FE,$07
	.dc.b	$D4,$FA,$07,$D0,$7F,$53,$8C,$36,$33,$3F,$E6,$D6,$BF,$D9,$A7,$6E
	.dc.b	$67,$81,$1F,$F3,$A7,$D8,$7F,$16,$7F,$8D,$CA,$31,$1E,$46,$F1,$7F
	.dc.b	$15,$BE,$EF,$FA,$8F,$E8,$9F,$45,$FC,$BA,$D1,$87,$F9,$39,$88,$FE
	.dc.b	$67,$87,$BF,$D7,$F8,$FF,$66,$F6,$3F,$B9,$70,$23,$18,$7C,$F6,$BF
	.dc.b	$E0,$48,$D4,$FF,$15,$F8,$27,$CA,$7F,$3E,$6D,$7F,$35,$6B,$F4,$35
	.dc.b	$FF,$2C,$3E,$6F,$51,$F4,$5F,$E3,$E6,$C7,$F8,$15,$AF,$F8,$45,$AF
	.dc.b	$F8,$59,$AF,$F8,$E1,$AF,$F8,$09,$A9,$CF,$53,$E1,$7F,$A1,$F4,$7F
	.dc.b	$82,$23,$1E,$4F,$F5,$89,$CD,$F3,$C3,$DD,$F3,$D2,$3D,$CF,$F5,$EF
	.dc.b	$95,$FF,$00,$72,$FF,$6C,$D5,$3B,$AF,$E9,$B1,$87,$FA,$D5,$70,$CF
	.dc.b	$C5,$8F,$FB,$62,$7F,$AF,$35,$FE,$AA,$DC,$3F,$BC,$FA,$4F,$8E,$F4
	.dc.b	$7C,$0C,$BE,$2F,$F7,$29,$D4,$7F,$6C,$E3,$7C,$5F,$F9,$03,$0F,$F7
	.dc.b	$94,$62,$7F,$5B,$39,$CB,$FC,$5B,$EB,$BC,$67,$C9,$6A,$E1,$D6,$F8
	.dc.b	$2A,$E1,$3F,$87,$F9,$09,$3F,$87,$78,$BE,$EF,$81,$FA,$44,$7D,$1F
	.dc.b	$0A,$F9,$3C,$46,$CF,$8F,$3F,$C1,$27,$51,$F9,$3F,$8D,$FC,$B1,$1F
	.dc.b	$A3,$ED,$D7,$EA,$C5,$DF,$FC,$4B,$C9,$FE,$59,$9F,$D6,$FD,$17,$F8
	.dc.b	$85,$FF,$2D,$35,$FF,$33,$35,$FF,$35,$B5,$FF,$46,$B5,$FF,$2B,$B5
	.dc.b	$34,$D7,$96,$E1,$FE,$0A,$7F,$97,$9C,$7B,$1F,$E3,$06,$D7,$9B,$D9
	.dc.b	$FD,$A7,$BB,$23,$F6,$8F,$E5,$FE,$33,$11,$F9,$48,$B9,$FC,$09,$FD
	.dc.b	$17,$99,$2B,$F2,$6F,$E1,$3D,$8C,$47,$E5,$22,$D7,$E2,$B9,$BF,$C0
	.dc.b	$FA,$79,$F4,$FF,$11,$5F,$FA,$2F,$4F,$F1,$CF,$B5,$FE,$83,$6C,$66
	.dc.b	$76,$FF,$A5,$23,$2D,$8C,$BF,$EB,$88,$CB,$63,$2F,$FA,$13,$A5,$F8
	.dc.b	$17,$FB,$F2,$3F,$44,$D7,$FC,$8F,$1A,$F8,$6F,$99,$5E,$2C,$FF,$CC
	.dc.b	$FF,$4F,$F4,$EF,$FC,$A4,$D7,$FC,$92,$D8,$FD,$AF,$D1,$FF,$98,$A3
	.dc.b	$1A,$7F,$0F,$6D,$7F,$B9,$5B,$E3,$7F,$A6,$6B,$F9,$67,$D5,$4A,$FE
	.dc.b	$58,$FF,$0F,$E4,$C4,$7F,$0C,$8B,$8F,$E6,$27,$F8,$A7,$C0,$FF,$92
	.dc.b	$F9,$FF,$CF,$77,$FE,$17,$6B,$FC,$85,$AB,$CF,$7C,$CB,$E4,$7F,$A5
	.dc.b	$B6,$7F,$AD,$B5,$FF,$34,$D7,$FA,$23,$1F,$E4,$5F,$25,$FD,$14,$93
	.dc.b	$C2,$EB,$FC,$92,$2C,$FF,$9D,$6F,$A3,$FB,$2F,$91,$F5,$12,$BF,$C7
	.dc.b	$18,$23,$FA,$A7,$B7,$FD,$4C,$93,$FC,$BA,$3F,$9B,$79,$3F,$0C,$91
	.dc.b	$C9,$FE,$B9,$EF,$FF,$5B,$24,$7F,$36,$8A,$FF,$3A,$8B,$87,$F0,$69
	.dc.b	$BC,$7F,$D2,$FC,$CF,$F1,$DB,$6B,$C3,$7F,$D0,$38,$BE,$A1,$1F,$5B
	.dc.b	$FD,$B7,$9B,$FB,$61,$23,$81,$FC,$E7,$E9,$3F,$8E,$BF,$DD,$5A,$FF
	.dc.b	$B8,$87,$42,$33,$1F,$A3,$6B,$A6,$C6,$FC,$6F,$F0,$2C,$60,$DF,$3B
	.dc.b	$A1,$F1,$7F,$9D,$26,$1E,$5B,$FF,$1A,$DF,$EC,$A3,$B7,$6F,$42,$3F
	.dc.b	$E3,$D8,$D6,$3E,$79,$CC,$FB,$5F,$F5,$26,$0F,$98,$FD,$61,$27,$E2
	.dc.b	$4D,$7F,$67,$17,$87,$FA,$EE,$1F,$BB,$DF,$F8,$27,$94,$F9,$5F,$F1
	.dc.b	$1B,$70,$FF,$28,$35,$FF,$11,$35,$F9,$D9,$FF,$62,$9E,$24,$8E,$2F
	.dc.b	$F8,$9F,$ED,$3F,$C4,$C4,$8F,$EE,$55,$D7,$1F,$E1,$5F,$49,$FE,$9E
	.dc.b	$F0,$14,$F8,$77,$D2,$7D,$DA,$D4,$E6,$31,$F1,$1F,$E4,$FA,$DC,$FE
	.dc.b	$EF,$B1,$F3,$9F,$85,$9C,$CB,$0E,$71,$86,$E1,$D3,$1F,$1C,$F4,$F0
	.dc.b	$58,$F4,$C3,$37,$B7,$F2,$47,$AB,$0F,$3B,$86,$BB,$31,$EE,$FF,$B1
	.dc.b	$23,$FC,$AD,$F0,$9C,$A7,$94,$E4,$91,$E0,$8C,$B7,$07,$D2,$7F,$19
	.dc.b	$E3,$D5,$7B,$EF,$FD,$B3,$9F,$BD,$B9,$7C,$D6,$CF,$F2,$3B,$E4,$FC
	.dc.b	$F3,$B9,$7A,$9D,$CF,$98,$FE,$8D,$C5,$FE,$43,$6E,$7F,$D4,$FE,$91
	.dc.b	$E2,$E3,$16,$3C,$AF,$1D,$0B,$DE,$1F,$EA,$36,$DF,$F9,$D5,$AF,$E5
	.dc.b	$1A,$F7,$C3,$FA,$B3,$67,$94,$62,$3F,$83,$CF,$07,$FA,$A7,$D3,$79
	.dc.b	$71,$FE,$00,$F9,$39,$DF,$49,$9E,$E4,$D2,$65,$F9,$E9,$B8,$7F,$6E
	.dc.b	$BE,$33,$FA,$E9,$F1,$67,$FE,$4C,$6C,$EA,$BF,$63,$18,$25,$71,$3F
	.dc.b	$DF,$58,$7F,$58,$7F,$BA,$F8,$FE,$7F,$8C,$F3,$36,$3F,$7B,$1D,$88
	.dc.b	$E0,$97,$FC,$02,$32,$FE,$73,$3F,$BF,$F1,$5F,$17,$87,$FD,$97,$7F
	.dc.b	$E7,$CF,$8C,$C4,$FC,$9E,$AF,$F6,$96,$1F,$DB,$E7,$E8,$61,$FA,$A6
	.dc.b	$CE,$A7,$F0,$78,$3E,$23,$B1,$CF,$E8,$AE,$5F,$EA,$F8,$B5,$FD,$CA
	.dc.b	$6F,$2F,$FA,$4E,$6D,$7F,$7F,$9B,$5F,$90,$C1,$1E,$22,$4F,$F6,$E7
	.dc.b	$CF,$83,$E6,$7E,$01,$ED,$E1,$B5,$FE,$9C,$7D,$64,$7F,$BB,$BC,$BD
	.dc.b	$6F,$90,$7E,$87,$B3,$FC,$1F,$F4,$58,$AF,$F6,$54,$DC,$FF,$DF,$3C
	.dc.b	$4A,$C6,$DF,$E0,$77,$94,$66,$FF,$D5,$BE,$0E,$27,$E9,$3F,$38,$91
	.dc.b	$FE,$0A,$7D,$BF,$0F,$F6,$CF,$B1,$F3,$36,$7F,$DB,$91,$71,$FF,$0F
	.dc.b	$CD,$C7,$FC,$8B,$E4,$FE,$8C,$FD,$47,$C3,$3E,$0B,$18,$FF,$92,$26
	.dc.b	$E1,$FE,$BC,$9B,$7F,$EB,$2F,$23,$FD,$3D,$7F,$EB,$76,$D2,$7F,$CA
	.dc.b	$5E,$CB,$E5,$3F,$D8,$57,$BF,$3B,$F8,$64,$E7,$D3,$FD,$F3,$67,$FC
	.dc.b	$E5,$5C,$30,$FF,$37,$35,$FF,$36,$35,$FF,$48,$35,$4F,$37,$8D,$7E
	.dc.b	$73,$BB,$CF,$8F,$F9,$A3,$1F,$E3,$7F,$44,$F1,$FF,$84,$44,$CF,$FC
	.dc.b	$D6,$EA,$31,$8F,$FB,$A3,$E2,$3F,$5C,$7E,$AF,$CF,$B7,$FD,$9B,$1F
	.dc.b	$F7,$84,$E2,$BF,$CA,$75,$67,$E3,$D3,$E2,$FE,$2B,$4E,$47,$F7,$01
	.dc.b	$F4,$4F,$67,$73,$FC,$DE,$73,$1F,$CA,$5A,$7F,$57,$2E,$1F,$D8,$9A
	.dc.b	$FF,$8D,$5A,$F7,$AE,$D1,$C5,$B1,$C5,$AF,$F8,$F7,$C3,$FE,$9F,$F0
	.dc.b	$FF,$C4,$4F,$E2,$51,$6B,$D6,$FF,$0A,$8C,$3F,$89,$8F,$F2,$B3,$8C
	.dc.b	$7F,$D3,$BF,$45,$FC,$EC,$FE,$A3,$ED,$BD,$DC,$5F,$FA,$82,$6D,$FF
	.dc.b	$AA,$2B,$53,$93,$C3,$8F,$F6,$4F,$82,$FE,$4C,$7F,$5C,$F8,$27,$C3
	.dc.b	$62,$BC,$FE,$72,$73,$18,$6E,$3E,$91,$F1,$2E,$E6,$9D,$71,$98,$71
	.dc.b	$7F,$B7,$56,$E7,$37,$FE,$B8,$FB,$1F,$EB,$47,$EE,$2A,$DF,$FA,$FA
	.dc.b	$2C,$F8,$74,$F0,$56,$37,$FF,$62,$43,$E9,$7F,$DF,$18,$7C,$26,$DF
	.dc.b	$1F,$29,$EC,$FF,$6D,$9C,$E1,$C9,$B7,$FE,$42,$C3,$A3,$FD,$3E,$B8
	.dc.b	$3F,$DC,$7C,$68,$FE,$00,$FF,$06,$F4,$63,$2F,$91,$18,$3F,$2D,$EA
	.dc.b	$FF,$98,$5B,$3F,$9B,$8C,$4F,$E6,$5A,$FE,$65,$AF,$E9,$DA,$A7,$B3
	.dc.b	$FD,$9F,$1F,$F4,$44,$7C,$3E,$EF,$E8,$1F,$48,$F8,$5F,$85,$1F,$E2
	.dc.b	$0F,$AF,$79,$3F,$E0,$46,$DF,$F9,$0F,$9D,$18,$CB,$FC,$9F,$18,$AF
	.dc.b	$F0,$E4,$3D,$EF,$F1,$CF,$0F,$FB,$DD,$B8,$7F,$9B,$BC,$E6,$9F,$9A
	.dc.b	$BF,$C7,$1C,$7E,$1B,$FA,$B7,$D7,$BE,$5A,$9E,$AE,$EB,$8F,$EC,$31
	.dc.b	$E7,$1F,$E2,$D6,$C7,$F9,$6D,$B8,$7F,$99,$9A,$FF,$AE,$5A,$F9,$71
	.dc.b	$FF,$6F,$4F,$16,$C7,$16,$BF,$C1,$79,$7D,$5A,$7E,$51,$B5,$FE,$8C
	.dc.b	$5F,$1F,$FC,$7E,$BF,$39,$1A,$7F,$2B,$19,$7F,$D9,$91,$C1,$FF,$24
	.dc.b	$36,$7F,$CE,$CD,$8F,$F4,$0B,$5E,$33,$CD,$F1,$35,$EE,$7F,$89,$3E
	.dc.b	$05,$F2,$BF,$E5,$16,$DF,$F1,$CB,$7D,$33,$D2,$D5,$FF,$9A,$E7,$31
	.dc.b	$E6,$9D,$57,$99,$B8,$79,$9A,$FB,$9A,$FC,$76,$BF,$B8,$6D,$76,$9F
	.dc.b	$F4,$B7,$5B,$E4,$DF,$25,$FE,$69,$6C,$FF,$24,$F9,$3F,$EC,$4F,$F2
	.dc.b	$68,$CE,$3F,$C3,$67,$2F,$F9,$3D,$AF,$F4,$16,$BF,$C1,$5A,$FF,$A2
	.dc.b	$9B,$66,$93,$CD,$FF,$28,$FC,$E3,$E4,$3F,$CF,$CD,$C7,$F9,$CD,$FF
	.dc.b	$12,$C3,$FD,$2B,$39,$7F,$91,$47,$F1,$C7,$FA,$A3,$5F,$E3,$0D,$7F
	.dc.b	$D3,$AD,$3F,$D3,$F3,$FE,$62,$70,$47,$FC,$CB,$9F,$FB,$61,$FE,$96
	.dc.b	$D9,$FE,$65,$EA,$3E,$93,$FD,$2D,$39,$E3,$FE,$67,$39,$BF,$AA,$DB
	.dc.b	$FF,$3D,$35,$FE,$D8,$D7,$F9,$83,$5F,$F5,$AB,$70,$FF,$18,$61,$FE
	.dc.b	$7A,$8C,$71,$7F,$B0,$6F,$FB,$0B,$6F,$FA,$44,$78,$70,$FE,$8F,$39
	.dc.b	$7F,$A4,$C7,$52,$5F,$FA,$19,$F0,$7C,$FB,$E5,$FF,$D6,$8D,$8F,$E2
	.dc.b	$1F,$48,$F9,$1C,$D4,$9F,$EE,$75,$BC,$BF,$B5,$74,$B0,$E1,$5A,$BF
	.dc.b	$F6,$07,$99,$F2,$E3,$FA,$93,$6F,$FD,$AE,$FE,$AB,$3F,$EC,$2E,$76
	.dc.b	$FE,$EB,$EE,$F2,$39,$F4,$5C,$3F,$C2,$B8,$09,$BD,$E6,$7C,$BF,$F0
	.dc.b	$AA,$C3,$BA,$95,$87,$F8,$AC,$D7,$27,$EC,$E7,$93,$4E,$42,$78,$3F
	.dc.b	$DD,$26,$B9,$1C,$9E,$47,$B9,$8A,$E6,$EE,$34,$FF,$2F,$9F,$74,$62
	.dc.b	$3D,$E7,$8C,$D1,$4C,$DE,$F9,$3F,$B6,$46,$4F,$24,$66,$BC,$F8,$7A
	.dc.b	$AB,$73,$DE,$BD,$47,$98,$A7,$A2,$70,$F9,$0F,$8B,$F4,$4F,$98,$7D
	.dc.b	$27,$18,$C3,$63,$33,$F0,$A7,$2D,$EB,$FF,$4C,$37,$87,$94,$62,$3B
	.dc.b	$B5,$FB,$D3,$FC,$AA,$7E,$4E,$B8,$FA,$4F,$C1,$6D,$F3,$A9,$18,$9E
	.dc.b	$38,$78,$2B,$4D,$8E,$58,$74,$3F,$25,$7C,$9C,$07,$C4,$FD,$F3,$73
	.dc.b	$CD,$E3,$7E,$16,$9C,$A7,$8D,$F8,$B0,$C5,$6E,$BB,$36,$B7,$87,$AA
	.dc.b	$B7,$50,$F1,$1E,$5C,$3C,$5C,$3F,$AB,$F0,$7E,$F3,$67,$FA,$A8,$8C
	.dc.b	$B4,$CB,$82,$38,$3F,$C3,$1F,$48,$F7,$3F,$15,$D0,$8C,$4E,$A7,$31
	.dc.b	$88,$FF,$0F,$35,$FF,$0F,$35,$FF,$19,$35,$4F,$59,$EF,$9B,$E5,$69
	.dc.b	$2B,$FC,$59,$19,$6F,$8F,$C0,$8E,$0D,$FC,$BF,$42,$38,$21,$FE,$35
	.dc.b	$94,$6A,$30,$FC,$2F,$AC,$FF,$2B,$C7,$F8,$F5,$AF,$F8,$F5,$AF,$F8
	.dc.b	$F5,$AF,$F9,$47,$EB,$1F,$39,$F3,$1B,$1F,$E4,$76,$BF,$E4,$76,$BF
	.dc.b	$E4,$76,$BC,$E3,$91,$1E,$5F,$2D,$D4,$8E,$0E,$DF,$1B,$E1,$23,$82
	.dc.b	$1E,$17,$2D,$3F,$90,$F1,$3F,$03,$7C,$0F,$F0,$F8,$C3,$63,$33,$C1
	.dc.b	$AF,$F2,$AF,$A3,$FB,$4D,$DF,$F8,$AC,$62,$3F,$9C,$B5,$FE,$88,$D7
	.dc.b	$FA,$63,$53,$4D,$70,$BE,$5D,$ED,$FE,$71,$B5,$FE,$77,$6B,$FE,$77
	.dc.b	$6B,$E2,$F2,$FC,$A4,$7F,$97,$5A,$31,$1E,$D6,$BC,$4F,$F4,$27,$D1
	.dc.b	$3E,$5B,$F5,$AD,$8C,$C6,$0F,$E6,$FC,$FF,$DF,$E3,$F9,$E3,$5F,$EA
	.dc.b	$6D,$7F,$AF,$B5,$FE,$D4,$D7,$FB,$93,$5F,$DE,$79,$5D,$8F,$DF,$FE
	.dc.b	$8C,$F1,$E9,$27,$FD,$39,$19,$6F,$8F,$C0,$8E,$0F,$3F,$57,$A8,$8E
	.dc.b	$0E,$7E,$F7,$80,$8E,$0F,$EB,$5F,$89,$BE,$4F,$FA,$CB,$63,$31,$82
	.dc.b	$38,$21,$FD,$E2,$38,$3F,$DE,$BF,$25,$FD,$5B,$FE,$15,$6A,$69,$AF
	.dc.b	$07,$FE,$00,$7B,$1F,$CC,$3E,$03,$FD,$8D,$5F,$E8,$46,$BF,$EE,$38
	.dc.b	$D1,$1D,$35,$FF,$74,$71,$73,$23,$CD,$AE,$8C,$10,$FF,$09,$FE,$0E
	.dc.b	$F8,$2F,$EF,$AD,$8F,$F0,$13,$5D,$C6,$88,$E9,$AF,$4F,$1F,$99,$1E
	.dc.b	$6F,$F4,$4F,$A7,$FF,$04,$47,$FB,$7D,$AF,$FB,$7D,$AF,$FB,$7D,$AF
	.dc.b	$FB,$CD,$AF,$F4,$26,$F5,$1A,$6B,$36,$7F,$C0,$5B,$D7,$F4,$C9,$7F
	.dc.b	$4D,$AF,$F9,$0A,$36,$E4,$FC,$6F,$A5,$F8,$C2,$38,$21,$F8,$97,$2D
	.dc.b	$3F,$C5,$DF,$A5,$FF,$6D,$7F,$C8,$AD,$7F,$CA,$2D,$7F,$CB,$8D,$53
	.dc.b	$EC,$E3,$FC,$61,$23,$F1,$8D,$7D,$58,$23,$11,$FE,$67,$9E,$4F,$F9
	.dc.b	$5B,$D3,$FE,$D5,$E6,$FF,$90,$DB,$19,$CB,$FC,$BE,$D9,$D4,$62,$3F
	.dc.b	$CD,$33,$C1,$B1,$97,$FC,$ED,$19,$7F,$CD,$FF,$A6,$3E,$23,$FC,$9E
	.dc.b	$D8,$FF,$2A,$35,$FF,$35,$B5,$E3,$1C,$08,$F0,$79,$FD,$D4,$8E,$6D
	.dc.b	$7F,$21,$D7,$F9,$52,$3F,$95,$6B,$A3,$04,$3F,$D0,$FC,$0F,$C4,$7F
	.dc.b	$CC,$ED,$8F,$F1,$1C,$60,$8E,$08,$7F,$A1,$A3,$83,$7E,$21,$F2,$FF
	.dc.b	$E2,$C6,$C6,$63,$07,$6E,$29,$1C,$EF,$FD,$88,$D8,$E5,$E0,$C8,$EC
	.dc.b	$D7,$D7,$7F,$D3,$5B,$19,$7F,$D5,$D8,$27,$51,$FE,$A8,$FC,$93,$FD
	.dc.b	$18,$FF,$AE,$5A,$9A,$4F,$8C,$FF,$46,$A6,$92,$3E,$17,$97,$F0,$48
	.dc.b	$E0,$F7,$7C,$07,$CE,$23,$82,$1F,$5B,$34,$6A,$30,$D7,$DB,$1A,$78
	.dc.b	$98,$8F,$F2,$EF,$EC,$9F,$E4,$08,$FB,$0D,$7F,$06,$D7,$E3,$B5,$FF
	.dc.b	$6E,$65,$F2,$7D,$4F,$B8,$48,$FC,$03,$FE,$B3,$F8,$A7,$E8,$DF,$9F
	.dc.b	$9F,$F7,$24,$63,$8F,$53,$99,$4E,$5F,$F6,$8B,$5F,$F6,$EB,$5F,$F7
	.dc.b	$43,$5E,$B1,$CE,$1F,$A5,$BE,$BF,$C8,$DA,$EA,$D7,$AB,$5F,$F7,$A7
	.dc.b	$C6,$3F,$57,$7D,$0F,$FB,$3E,$30,$D8,$CC,$FF,$BA,$27,$2D,$8C,$BF
	.dc.b	$EF,$68,$CB,$63,$2F,$81,$A8,$43,$C0,$7C,$5F,$8D,$B3,$E1,$6B,$DD
	.dc.b	$AF,$B1,$AF,$59,$E9,$3C,$C9,$1E,$9F,$A5,$F7,$11,$C1,$0F,$53,$96
	.dc.b	$9E,$33,$CC,$43,$07,$FB,$E7,$EA,$FF,$D4,$55,$EF,$F8,$47,$D5,$FC
	.dc.b	$1A,$56,$94,$E9,$F8,$D1,$98,$C7,$13,$E9,$BF,$C7,$55,$C2,$FA,$5E
	.dc.b	$DB,$E1,$7C,$E9,$D1,$D1,$C6,$1D,$2F,$9B,$E1,$9F,$2C,$F1,$7D,$06
	.dc.b	$31,$F7,$46,$63,$BC,$7C,$1F,$2D,$F1,$14,$93,$C8,$F8,$31,$A7,$19
	.dc.b	$7F,$9C,$E7,$E4,$65,$EC,$AD,$63,$F0,$BE,$AF,$FB,$57,$C5,$6B,$1F
	.dc.b	$8F,$5A,$BE,$18,$66,$7B,$BB,$D5,$87,$29,$F8,$A6,$8F,$8F,$9F,$F7
	.dc.b	$75,$8F,$E8,$FC,$9F,$CD,$36,$32,$7E,$8E,$32,$FE,$97,$C6,$7A,$9A
	.dc.b	$8F,$D9,$C6,$2B,$F5,$CD,$8E,$13,$B7,$F3,$91,$97,$8C,$65,$FD,$5C
	.dc.b	$65,$DC,$66,$DE,$BF,$FA,$16,$FF,$78,$DA,$FD,$D3,$5F,$E0,$ED,$7F
	.dc.b	$58,$D7,$F7,$2D,$9E,$BE,$3F,$78,$C6,$0F,$A7,$F1,$37,$0F,$7C,$60
	.dc.b	$8E,$09,$87,$C9,$8F,$77,$5B,$F5,$27,$D2,$61,$B7,$FC,$B9,$B7,$FC
	.dc.b	$B9,$AF,$F2,$EF,$D2,$BF,$B4,$4F,$F3,$46,$C7,$F3,$46,$BB,$F2,$3C
	.dc.b	$31,$88,$FE,$82,$D7,$F9,$FB,$70,$39,$E1,$FE,$ED,$F0,$3F,$5D,$B8
	.dc.b	$7F,$83,$63,$18,$7F,$BD,$23,$FA,$DF,$D4,$EF,$C5,$34,$91,$FE,$0D
	.dc.b	$F2,$FA,$79,$1F,$ED,$58,$C1,$34,$FE,$E1,$3A,$24,$66,$3F,$B7,$C6
	.dc.b	$1D,$C6,$5F,$EC,$FF,$34,$F9,$BF,$53,$67,$FC,$50,$D9,$FF,$19,$B5
	.dc.b	$FE,$13,$F8,$EF,$F2,$29,$FE,$05,$08,$C3,$FC,$15,$AF,$0E,$BF,$CF
	.dc.b	$3B,$FC,$06,$C9,$A4,$AF,$1F,$B7,$E6,$23,$82,$57,$9A,$32,$D8,$FF
	.dc.b	$1C,$7B,$DF,$30,$8E,$0F,$7F,$85,$F0,$88,$E0,$FF,$20,$79,$BF,$50
	.dc.b	$8E,$09,$5F,$E2,$38,$CB,$63,$FC,$A5,$F2,$DF,$C4,$1A,$7E,$25,$C4
	.dc.b	$7F,$98,$1A,$FF,$91,$1A,$FF,$87,$9A,$B9,$11,$86,$CF,$CE,$F0,$7F
	.dc.b	$DF,$0F,$D6,$6D,$7F,$39,$6B,$FD,$11,$AF,$F5,$D5,$F6,$3C,$C9,$3C
	.dc.b	$DB,$3C,$DA,$F3,$7F,$8E,$7A,$EF,$D6,$5F,$5B,$86,$DF,$FA,$9D,$B7
	.dc.b	$FE,$A0,$6B,$FE,$9E,$6B,$FE,$11,$AE,$95,$CC,$95,$FD,$57,$DB,$FF
	.dc.b	$6B,$FC,$2F,$F2,$D6,$C6,$7D,$D4,$E5,$B1,$98,$C1,$1C,$1F,$A0,$F9
	.dc.b	$AF,$F1,$8B,$7D,$2F,$CE,$C6,$32,$FE,$CC,$D7,$FB,$8A,$FB,$0D,$92
	.dc.b	$BA,$36,$78,$B5,$E2,$F9,$FE,$4A,$4F,$AD,$B3,$FE,$D6,$6B,$FE,$D5
	.dc.b	$C5,$18,$8F,$C6,$76,$FE,$4C,$95,$F9,$86,$C7,$E5,$5A,$FE,$55,$FD
	.dc.b	$DF,$B5,$F8,$88,$FF,$22,$35,$34,$C6,$BF,$D7,$63,$FC,$94,$D7,$F0
	.dc.b	$BE,$5F,$F0,$68,$C1,$1C,$12,$FE,$D7,$C5,$FD,$92,$38,$25,$7F,$BE
	.dc.b	$23,$2D,$8F,$C1,$61,$F8,$12,$38,$24,$7F,$C0,$08,$CB,$63,$F9,$0F
	.dc.b	$E5,$AF,$8D,$FE,$3A,$D8,$CC,$60,$FE,$43,$E9,$3E,$0A,$92,$7F,$95
	.dc.b	$C6,$5B,$19,$BF,$F1,$84,$E5,$B1,$FC,$EB,$DD,$95,$FC,$E9,$B3,$FE
	.dc.b	$37,$F9,$C7,$7F,$F3,$26,$D7,$F3,$A6,$CE,$E7,$FA,$2E,$32,$BF,$A2
	.dc.b	$B6,$7F,$A2,$FD,$FF,$F4,$A2,$4F,$F4,$A7,$B7,$8B,$23,$B3,$67,$C8
	.dc.b	$D7,$FC,$AB,$87,$F4,$F7,$FC,$DC,$D7,$F9,$07,$D0,$BE,$63,$FB,$93
	.dc.b	$6F,$FE,$00,$36,$7F,$CB,$FD,$E8,$C4,$7F,$A7,$FC,$2F,$EA,$7E,$A9
	.dc.b	$A4,$BF,$F5,$63,$67,$FD,$54,$D7,$FD,$F3,$F4,$5C,$67,$FC,$C9,$82
	.dc.b	$31,$3E,$26,$BC,$2B,$C3,$EC,$CB,$FF,$5D,$36,$7F,$D7,$0D,$7F,$D6
	.dc.b	$CD,$7F,$D6,$5F,$28,$FC,$0D,$F6,$3F,$E4,$B8,$C3,$63,$37,$FE,$79
	.dc.b	$9C,$BE,$8F,$9D,$7C,$7F,$F8,$75,$B1,$98,$C1,$1C,$13,$0F,$9F,$3C
	.dc.b	$1B,$F3,$1C,$C9,$3D,$DE,$DE,$8C,$8F,$13,$67,$9B,$5F,$F5,$86,$CB
	.dc.b	$FF,$13,$46,$AB,$E8,$FD,$EF,$BA,$3F,$02,$D7,$8B,$5F,$F6,$D5,$23
	.dc.b	$11,$FD,$B7,$EB,$DF,$37,$FE,$3F,$6D,$FD,$E6,$CF,$E3,$9A,$F9,$9A
	.dc.b	$F6,$6B,$D6,$7A,$1D,$7E,$42,$57,$56,$CF,$5C,$FC,$44,$9F,$13,$FE
	.dc.b	$53,$FA,$3F,$F0,$11,$23,$04,$3F,$2A,$E5,$A7,$F6,$5F,$33,$FA,$B5
	.dc.b	$7F,$BF,$7C,$C4,$62,$11,$8F,$2F,$81,$2B,$FB,$85,$6E,$75,$18,$9F
	.dc.b	$F2,$0D,$7F,$BA,$C7,$C9,$3E,$53,$FC,$25,$49,$CC,$7F,$20,$69,$FE
	.dc.b	$15,$8D,$4F,$18,$FF,$6E,$F8,$78,$AE,$71,$C0,$FF,$7C,$46,$5E,$46
	.dc.b	$4F,$24,$F3,$7C,$5E,$8F,$FB,$12,$31,$1A,$AE,$B1,$88,$FF,$80,$27
	.dc.b	$39,$E1,$3E,$99,$E3,$18,$9F,$39,$BB,$F2,$3E,$63,$07,$94,$EA,$7F
	.dc.b	$46,$3F,$CE,$26,$0F,$8C,$7F,$BF,$BE,$7D,$F4,$2F,$A2,$FF,$3B,$C6
	.dc.b	$1B,$19,$9F,$F7,$E4,$E5,$BF,$44,$E9,$E2,$D8,$CC,$60,$8F,$CF,$9F
	.dc.b	$DB,$3E,$C6,$B9,$9E,$59,$7C,$F9,$46,$33,$FC,$15,$27,$8D,$6F,$9F
	.dc.b	$2F,$1B,$E1,$5F,$89,$AE,$78,$78,$6F,$78,$74,$9E,$6D,$BE,$37,$DD
	.dc.b	$B3,$A8,$E9,$5C,$9E,$D3,$A3,$71,$98,$F4,$36,$7C,$71,$F9,$AD,$FF
	.dc.b	$3D,$E0,$FF,$AA,$9B,$1F,$E5,$93,$FA,$F1,$63,$4E,$0F,$8A,$7E,$D0
	.dc.b	$C1,$F2,$23,$9B,$A6,$BF,$CA,$BE,$1D,$EC,$7F,$AE,$BA,$91,$89,$FE
	.dc.b	$B6,$D7,$FB,$5B,$5E,$53,$C6,$31,$1F,$CB,$1A,$FF,$A5,$DB,$3F,$6F
	.dc.b	$C0,$F0,$63,$FD,$C6,$BF,$B4,$B5,$FE,$F2,$D7,$FA,$0A,$F8,$9E,$24
	.dc.b	$8E,$2F,$4F,$4A,$47,$46,$BD,$1A,$F4,$6B,$FD,$13,$E8,$9F,$4F,$F1
	.dc.b	$5B,$1F,$DF,$9A,$FF,$83,$9A,$FF,$56,$FA,$3F,$E9,$2F,$F8,$11,$AF
	.dc.b	$F8,$51,$AE,$FB,$BE,$94,$62,$3F,$AA,$B5,$FF,$60,$37,$33,$47,$6F
	.dc.b	$22,$4F,$85,$EB,$EA,$C8,$F2,$35,$F2,$35,$F5,$35,$FF,$07,$FD,$1B
	.dc.b	$E1,$BF,$E0,$0F,$3A,$31,$3F,$E0,$D6,$BF,$96,$8D,$4F,$F1,$8F,$A3
	.dc.b	$FB,$93,$FB,$E6,$BF,$D0,$5A,$9A,$4E,$1E,$44,$8F,$F1,$8F,$AD,$FE
	.dc.b	$73,$6F,$A3,$F9,$D8,$C4,$7F,$91,$1A,$FE,$AA,$9E,$5F,$94,$91,$FA
	.dc.b	$E3,$C8,$E1,$C4,$70,$74,$79,$E3,$4F,$F8,$63,$E8,$DE,$4F,$EB,$DB
	.dc.b	$1F,$E4,$86,$BF,$E5,$86,$BF,$BE,$E7,$FF,$6B,$BD,$9A,$F9,$5A,$EF
	.dc.b	$DD,$FE,$8B,$18,$8F,$E5,$6D,$7F,$B7,$35,$FE,$03,$1F,$C0,$FE,$8D
	.dc.b	$F2,$14,$93,$FE,$5E,$8C,$B7,$DB,$E0,$47,$07,$9F,$AB,$C6,$47,$07
	.dc.b	$3F,$A3,$F0,$11,$C1,$F3,$3E,$8D,$F0,$BF,$D8,$1B,$19,$8C,$11,$C1
	.dc.b	$0F,$B5,$1C,$1F,$B9,$F4,$7F,$AD,$4D,$27,$5F,$F3,$5F,$81,$FF,$54
	.dc.b	$CF,$E0,$FC,$24,$62,$3F,$97,$35,$FF,$78,$7B,$9C,$48,$F1,$8D,$3F
	.dc.b	$E2,$F6,$BF,$E2,$E7,$F1,$8F,$4E,$3F,$11,$1E,$4D,$74,$60,$FC,$9F
	.dc.b	$D1,$FF,$9F,$1B,$EA,$FF,$82,$23,$11,$F9,$D6,$BF,$D3,$1A,$FF,$72
	.dc.b	$6B,$FE,$35,$6A,$69,$FE,$23,$75,$E4,$F7,$23,$FE,$A6,$FA,$3F,$F0
	.dc.b	$AB,$7C,$CF,$EA,$51,$88,$FF,$5C,$B5,$FF,$40,$B5,$FF,$3C,$B5,$78
	.dc.b	$68,$C3,$67,$ED,$76,$CA,$FA,$4D,$7E,$CB,$5F,$ED,$59,$FF,$14,$D7
	.dc.b	$EF,$CF,$E4,$7E,$8E,$57,$E5,$5A,$FE,$39,$AF,$F7,$6E,$14,$62,$3F
	.dc.b	$C3,$BF,$39,$FD,$E1,$BE,$5F,$F8,$F2,$31,$1F,$E6,$36,$E1,$F9,$0E
	.dc.b	$6F,$CA,$12,$7F,$18,$D7,$F2,$6D,$7D,$6F,$F1,$6F,$53,$FD,$C1,$E7
	.dc.b	$DC,$FF,$CE,$75,$FC,$7E,$72,$D8,$CB,$FE,$10,$8C,$B6,$32,$FF,$8E
	.dc.b	$23,$2D,$8C,$BF,$E3,$CF,$97,$7C,$17,$F9,$69,$B1,$FC,$A9,$AF,$F3
	.dc.b	$16,$BF,$E2,$06,$BC,$E3,$91,$1E,$5F,$07,$D4,$8E,$0F,$C9,$77,$7E
	.dc.b	$40,$8E,$08,$7E,$49,$CB,$4F,$F7,$FF,$C8,$7F,$9C,$B8,$CD,$24,$7F
	.dc.b	$8B,$5A,$F0,$30,$7F,$8C,$7E,$49,$ED,$D2,$47,$F5,$1A,$4E,$A3,$C3
	.dc.b	$C5,$D4,$8E,$0E,$11,$FD,$29,$FF,$6F,$B5,$FD,$3B,$67,$AF,$8B,$E3
	.dc.b	$24,$F2,$6B,$FD,$8F,$E8,$FF,$7C,$9A,$4E,$9F,$F4,$AA,$FA,$3A,$49
	.dc.b	$FE,$CF,$19,$7C,$FE,$77,$C3,$23,$83,$E7,$FD,$1F,$CB,$23,$82,$1F
	.dc.b	$E6,$4C,$7E,$9C,$6A,$31,$AA,$31,$1F,$E8,$E9,$DB,$C8,$C4,$ED,$FA
	.dc.b	$5F,$2A,$F5,$7F,$CD,$0D,$8C,$C6,$0F,$B9,$EB,$C8,$FA,$6D,$7F,$CF
	.dc.b	$3F,$2E,$FA,$3F,$CD,$C2,$31,$CB,$FE,$0C,$6C,$7F,$84,$DA,$F2,$9E
	.dc.b	$31,$88,$FF,$54,$B5,$FF,$76,$35,$ED,$F4,$4F,$A3,$EC,$D8,$EC,$D7
	.dc.b	$CE,$D7,$FD,$6B,$F1,$2F,$A3,$7C,$27,$F8,$8A,$30,$D8,$CC,$FF,$8A
	.dc.b	$E7,$2D,$8C,$BF,$E9,$C8,$CB,$63,$2F,$81,$A8,$43,$C1,$7D,$1F,$8D
	.dc.b	$B3,$E1,$6B,$DD,$AF,$B1,$AF,$59,$E9,$3C,$C9,$1E,$9F,$A3,$F7,$11
	.dc.b	$C1,$0F,$53,$96,$9E,$33,$CC,$43,$07,$F9,$33,$E8,$9E,$07,$BA,$1F
	.dc.b	$0E,$F9,$CF,$F7,$3D,$6B,$1F,$89,$39,$AF,$8D,$39,$78,$D7,$C9,$E6
	.dc.b	$7D,$1F,$F8,$F2,$72,$70,$C7,$C1,$87,$38,$61,$CE,$30,$DC,$3A,$5F
	.dc.b	$39,$CD,$FA,$6F,$FD,$B3,$7E,$36,$C6,$5F,$5B,$5F,$4F,$9E,$F8,$9F
	.dc.b	$88,$D8,$FF,$3B,$57,$F9,$EA,$72,$48,$F5,$9E,$89,$E1,$7F,$17,$27
	.dc.b	$D1,$FF,$B3,$7E,$2B,$E4,$E9,$F3,$2F,$6D,$C7,$38,$F9,$1B,$87,$2F
	.dc.b	$29,$86,$AB,$C1,$5C,$36,$FF,$6D,$46,$8F,$D0,$57,$E6,$EB,$F4,$CD
	.dc.b	$8C,$9B,$3F,$63,$E4,$3D,$8C,$47,$E5,$E3,$15,$FA,$F6,$C7,$EB,$5A
	.dc.b	$F2,$9E,$2F,$EA,$A3,$2F,$38,$CD,$A7,$33,$FB,$0F,$6F,$FD,$0F,$7F
	.dc.b	$BD,$6C,$7E,$ED,$AF,$F0,$86,$BF,$AD,$6B,$CF,$C9,$EF,$18,$B3,$6D
	.dc.b	$F4,$7E,$46,$E1,$EF,$8C,$11,$C1,$30,$F8,$51,$F0,$3A,$DF,$44,$FA
	.dc.b	$3C,$36,$FF,$98,$36,$FF,$9C,$B5,$FE,$5F,$F4,$7F,$DD,$A7,$F9,$B3
	.dc.b	$63,$F9,$FB,$5D,$F9,$7E,$18,$C4,$7F,$36,$6B,$FD,$0D,$B8,$1C,$F0
	.dc.b	$FF,$81,$3E,$13,$EC,$B7,$0F,$F0,$0C,$63,$0F,$F7,$C4,$7F,$5A,$F9
	.dc.b	$AD,$F8,$C6,$92,$3F,$C3,$BE,$6F,$4F,$23,$FD,$B3,$18,$26,$9F,$DB
	.dc.b	$27,$44,$8C,$C7,$F6,$A8,$C3,$B8,$CB,$FC,$47,$E8,$DF,$33,$EA,$6C
	.dc.b	$FF,$88,$5B,$3F,$DF,$5A,$FF,$41,$3F,$A1,$7D,$1F,$F2,$39,$FF,$79
	.dc.b	$B6,$3F,$DF,$4D,$78,$07,$0F,$35,$F4,$7F,$A4,$8D,$36,$33,$3D,$A3
	.dc.b	$D1,$EF,$C8,$F4,$36,$BD,$0D,$7F,$C2,$16,$8C,$47,$F3,$3C,$BF,$DB
	.dc.b	$F5,$FE,$4E,$F8,$2F,$A4,$47,$E2,$78,$5F,$1C,$8E,$0F,$F2,$AF,$9B
	.dc.b	$F6,$48,$E0,$97,$FE,$24,$8C,$B6,$3F,$CA,$1F,$47,$FC,$69,$A7,$E3
	.dc.b	$9C,$47,$F8,$BD,$AF,$F8,$AD,$AF,$F4,$D6,$BF,$EB,$66,$F1,$9C,$C7
	.dc.b	$F3,$DC,$3E,$87,$7F,$C2,$8F,$AE,$66,$EB,$4C,$C3,$E8,$F8,$92,$77
	.dc.b	$19,$6C,$66,$78,$B5,$FF,$4D,$7D,$1B,$E8,$7F,$92,$FB,$68,$C5,$FF
	.dc.b	$A9,$DB,$5F,$EA,$56,$BC,$AB,$8C,$62,$3F,$C3,$15,$CE,$72,$7F,$32
	.dc.b	$F7,$5F,$00,$F8,$9C,$37,$1F,$87,$3A,$9C,$92,$33,$3B,$C3,$F3,$F5
	.dc.b	$B9,$CD,$7E,$77,$E8,$FF,$C7,$31,$F3,$1B,$5F,$41,$AF,$EA,$1A,$FE
	.dc.b	$C5,$7D,$1F,$32,$4F,$36,$CF,$36,$BC,$DF,$6F,$B1,$27,$D2,$D9,$F4
	.dc.b	$E3,$FE,$71,$9C,$92,$33,$3F,$95,$EB,$93,$F9,$A6,$CF,$E5,$DA,$FE
	.dc.b	$5D,$AF,$F0,$BF,$6B,$F2,$0F,$F0,$66,$B9,$E1,$FE,$09,$18,$8F,$F2
	.dc.b	$96,$BF,$D9,$1F,$C4,$79,$3F,$C2,$63,$04,$70,$4C,$3F,$37,$1F,$DB
	.dc.b	$36,$FC,$0F,$C7,$7E,$00,$93,$F7,$1B,$7F,$11,$AF,$F2,$39,$E3,$3F
	.dc.b	$82,$CB,$F1,$24,$70,$49,$F9,$B1,$96,$C7,$EB,$BE,$8D,$F2,$1F,$C8
	.dc.b	$DB,$19,$8C,$1F,$C9,$FD,$17,$C1,$D2,$4F,$EE,$23,$2D,$8C,$DF,$EF
	.dc.b	$A7,$2D,$8F,$E8,$9E,$EC,$AF,$E8,$8D,$9F,$E0,$BF,$45,$FC,$AE,$7F
	.dc.b	$82,$B6,$33,$DF,$FC,$EE,$31,$1F,$CC,$B1,$FE,$A0,$4A,$FE,$A0,$FF
	.dc.b	$53,$FA,$39,$1F,$D4,$DB,$3E,$0F,$13,$A1,$27,$C2,$D8,$D4,$60,$8F
	.dc.b	$FB,$5F,$E8,$DF,$11,$FD,$E9,$B3,$FF,$00,$5B,$3B,$9F,$EA,$9E,$97
	.dc.b	$FA,$12,$BF,$D6,$1E,$1E,$08,$E0,$95,$FE,$B1,$8C,$BF,$EE,$7F,$A3
	.dc.b	$E4,$D8,$FF,$05,$C6,$23,$FD,$CC,$D7,$FD,$83,$EB,$F8,$89,$5F,$EC
	.dc.b	$16,$C7,$FA,$FD,$AF,$FA,$F5,$FF,$27,$FC,$AB,$E8,$9F,$47,$86,$DF
	.dc.b	$F4,$09,$D4,$E4,$FF,$0F,$FC,$F3,$E4,$A9,$2B,$FA,$3F,$92,$9D,$46
	.dc.b	$1B,$97,$F8,$0A,$75,$5C,$63,$13,$A7,$FC,$1D,$3C,$FE,$6B,$C2,$4B
	.dc.b	$E8,$F9,$3D,$09,$1D,$DB,$5D,$1A,$F4,$BF,$F1,$6B,$99,$F4,$FD,$1F
	.dc.b	$C2,$8F,$33,$5F,$1B,$5F,$F5,$D3,$56,$0F,$A3,$7C,$B7,$F9,$19,$B7
	.dc.b	$F1,$9B,$5F,$0D,$AF,$C0,$6B,$FE,$C7,$6B,$D9,$B5,$DE,$BB,$7C,$8C
	.dc.b	$BE,$CD,$BE,$D9,$F8,$C9,$3E,$37,$FC,$B7,$F4,$7F,$E0,$62,$46,$08
	.dc.b	$7D,$E7,$2D,$3F,$B5,$79,$9F,$D5,$A9,$E1,$23,$0D,$F8,$AD,$92,$BF
	.dc.b	$B6,$CE,$89,$1F,$E4,$3A,$FF,$76,$8F,$93,$7C,$97,$F9,$33,$D2,$FF
	.dc.b	$5E,$46,$27,$FC,$9F,$E2,$1E,$15,$BC,$71,$5B,$9F,$F2,$E5,$62,$B5
	.dc.b	$1C,$0E,$65,$3A,$5F,$82,$3F,$95,$46,$4F,$F7,$45,$F3,$78,$DF,$77
	.dc.b	$26,$C5,$ED,$EB,$35,$F8,$33,$C6,$FC,$11,$D2,$FC,$F3,$D6,$39,$9E
	.dc.b	$39,$DB,$EE,$3F,$C7,$66,$8F,$E8,$C6,$2F,$FC,$7F,$F4,$2F,$96,$7D
	.dc.b	$1E,$1B,$5F,$E4,$69,$D4,$E4,$E3,$F4,$6E,$AA,$4A,$E1,$19,$6C,$7D
	.dc.b	$7F,$A4,$FA,$97,$FC,$2E,$71,$A7,$FA,$46,$31,$09,$DD,$F3,$6D,$F0
	.dc.b	$9F,$C1,$57,$09,$EB,$D1,$FE,$9B,$77,$3C,$F0,$ED,$87,$97,$5D,$CF
	.dc.b	$D5,$C3,$C3,$1C,$0E,$37,$A9,$E4,$78,$30,$ED,$87,$4C,$7C,$92,$C1
	.dc.b	$8F,$A6,$BD,$D7,$C2,$B9,$1C,$2B,$77,$ED,$E1,$FC,$F7,$C2,$7F,$AD
	.dc.b	$61,$C1,$FA,$19,$CD,$FE,$8E,$72,$FE,$D8,$FF,$79,$19,$9E,$0F,$F6
	.dc.b	$63,$06,$EB,$B4,$69,$B1,$93,$75,$FB,$1F,$0B,$2B,$C2,$7B,$58,$6D
	.dc.b	$70,$6B,$F8,$9C,$3F,$75,$D4,$F8,$7C,$E7,$EA,$C7,$A6,$3C,$C9,$87
	.dc.b	$A2,$F8,$47,$99,$FA,$1F,$13,$FE,$9C,$6C,$FF,$5A,$D5,$39,$6C,$7F
	.dc.b	$A0,$B2,$7D,$87,$B1,$B1,$F0,$1A,$FA,$4F,$E8,$B3,$C9,$C5,$BE,$91
	.dc.b	$F4,$BE,$36,$CF,$F6,$C6,$BF,$DA,$5A,$FF,$66,$6B,$F1,$9A,$FF,$54
	.dc.b	$6C,$F7,$9F,$67,$A7,$F0,$EC,$D3,$E3,$57,$83,$5F,$C3,$DB,$E9,$5D
	.dc.b	$FF,$25,$B9,$7F,$80,$9A,$FF,$7F,$6B,$FD,$F1,$AF,$DC,$6B,$FD,$A9
	.dc.b	$B3,$DE,$7F,$DA,$3B,$FF,$4C,$DB,$F1,$6F,$CB,$C3,$F0,$4D,$D7,$A5
	.dc.b	$7E,$52,$5F,$4A,$F8,$8F,$F6,$DC,$61,$B1,$9C,$FF,$C4,$4D,$7F,$A5
	.dc.b	$1F,$A9,$70,$FE,$BD,$AE,$63,$99,$F1,$78,$7F,$21,$A7,$E6,$31,$F8
	.dc.b	$11,$E5,$BE,$D8,$7C,$ED,$BF,$D0,$38,$FF,$8B,$1A,$FF,$8A,$5A,$FF
	.dc.b	$89,$9A,$FE,$99,$AF,$F8,$51,$BA,$FD,$EF,$43,$F4,$11,$AD,$5F,$57
	.dc.b	$FE,$18,$BF,$DC,$B5,$FC,$D4,$6B,$C5,$FD,$55,$79,$31,$FC,$2F,$D2
	.dc.b	$BC,$9F,$D6,$37,$0F,$F2,$2B,$5F,$F2,$13,$5F,$F1,$FB,$5F,$DF,$B5
	.dc.b	$FF,$16,$B7,$33,$7E,$EB,$F8,$64,$FE,$FD,$AF,$EB,$9B,$7E,$6B,$FE
	.dc.b	$05,$F4,$CF,$92,$FE,$04,$DB,$FF,$2D,$B5,$FF,$2C,$35,$FF,$2A,$B5
	.dc.b	$FE,$2A,$D7,$FC,$86,$DC,$4E,$7E,$3D,$F5,$CF,$7B,$F9,$4C,$61,$B1
	.dc.b	$9B,$FE,$35,$5E,$3C,$BF,$9C,$C7,$F2,$3E,$8F,$E6,$2D,$3F,$9B,$94
	.dc.b	$FE,$1F,$2F,$7F,$F9,$17,$D2,$7F,$8A,$E7,$FC,$F2,$D7,$FC,$EC,$D7
	.dc.b	$FC,$E6,$D7,$FA,$67,$C2,$3E,$2B,$F9,$EB,$63,$FA,$0B,$5F,$0F,$D3
	.dc.b	$F9,$BE,$9F,$F1,$91,$FA,$C0,$E1,$FD,$7B,$5F,$F1,$D3,$5D,$F2,$FF
	.dc.b	$86,$A3,$04,$7A,$47,$F9,$07,$77,$D7,$BE,$43,$E7,$B6,$7F,$AE,$CE
	.dc.b	$A7,$CD,$5E,$CF,$A1,$95,$EC,$7D,$FF,$07,$88,$FD,$2E,$69,$C7,$0F
	.dc.b	$F5,$58,$57,$C6,$F8,$59,$7F,$19,$F9,$3E,$F6,$23,$9C,$5B,$E5,$3F
	.dc.b	$7B,$EA,$3F,$CB,$11,$F8,$26,$F8,$5F,$C5,$9B,$7F,$E6,$17,$FD,$70
	.dc.b	$FE,$11,$CA,$69,$3D,$DF,$C9,$F8,$CF,$84,$A4,$AE,$5E,$8F,$E0,$3E
	.dc.b	$F6,$47,$E0,$1F,$C6,$FA,$F8,$8F,$8B,$17,$3F,$E9,$7D,$1F,$E4,$D9
	.dc.b	$FC,$3F,$DE,$7E,$28,$8F,$E2,$BE,$12,$BF,$3A,$2D,$7C,$D9,$FD,$0F
	.dc.b	$D1,$7F,$85,$EB,$F0,$2D,$7F,$1C,$D7,$F9,$DB,$5F,$F4,$33,$53,$49
	.dc.b	$EA,$7F,$9A,$3F,$5A,$7C,$77,$EC,$1B,$5F,$B0,$6B,$FE,$01,$39,$CF
	.dc.b	$E9,$BC,$39,$1F,$A6,$7F,$53,$E1,$62,$3F,$17,$17,$2F,$C5,$4F,$F0
	.dc.b	$6E,$6F,$E8,$FC,$BF,$9D,$6D,$BB,$11,$8F,$80,$6D,$FB,$DC,$7F,$C8
	.dc.b	$35,$AF,$47,$A1,$30,$FF,$1F,$DF,$0F,$73,$F8,$6F,$CE,$F4,$9F,$D6
	.dc.b	$87,$0F,$EC,$1A,$FE,$F5,$AF,$F6,$A6,$BF,$EB,$56,$BC,$F2,$FF,$4A
	.dc.b	$C6,$07,$C9,$F8,$89,$3F,$C7,$A3,$C3,$97,$B3,$A7,$FC,$62,$DF,$63
	.dc.b	$F3,$67,$F2,$0F,$2E,$DF,$F0,$0A,$DE,$FF,$C1,$DB,$1F,$EF,$CE,$7F
	.dc.b	$E6,$13,$FC,$82,$FF,$8F,$92,$7F,$C9,$DF,$2D,$FC,$6F,$1F,$E0,$13
	.dc.b	$99,$FE,$0A,$FF,$94,$67,$55,$FE,$5C,$5F,$55,$FE,$53,$6D,$7F,$1A
	.dc.b	$F3,$51,$88,$FE,$51,$18,$9F,$F2,$D7,$CD,$FF,$6D,$C7,$FC,$6E,$DC
	.dc.b	$3F,$CD,$2D,$B5,$F1,$7F,$9B,$FD,$8F,$F3,$A2,$3F,$E6,$99,$E1,$EA
	.dc.b	$3C,$65,$6A,$BF,$90,$BF,$E6,$DA,$ED,$0F,$AA,$7D,$5F,$FB,$41,$B1
	.dc.b	$F8,$A6,$BF,$CE,$1A,$FF,$45,$F3,$5F,$57,$FD,$24,$91,$FD,$7A,$78
	.dc.b	$EB,$FE,$56,$DF,$73,$F2,$5A,$FC,$86,$BF,$E3,$CD,$7E,$6D,$FF,$93
	.dc.b	$B4,$7C,$A7,$F5,$C2,$57,$DE,$9A,$FF,$40,$8B,$87,$86,$6E,$FF,$D9
	.dc.b	$A6,$D7,$D0,$C7,$C7,$7F,$5B,$DC,$FB,$24,$9F,$C0,$6D,$FE,$A2,$7C
	.dc.b	$D5,$FE,$33,$AF,$77,$9D,$F8,$B3,$E8,$5F,$F8,$2B,$EA,$BF,$A9,$5F
	.dc.b	$F6,$2E,$E4,$62,$3F,$B8,$46,$3C,$FF,$DD,$B6,$38,$4E,$DF,$F3,$0C
	.dc.b	$65,$E3,$19,$7F,$B7,$35,$72,$7E,$1F,$DC,$FF,$02,$92,$3F,$B7,$CF
	.dc.b	$8F,$2F,$F6,$9F,$C1,$FF,$30,$FA,$9F,$E1,$AD,$AF,$EF,$6D,$7F,$C1
	.dc.b	$0D,$7F,$CE,$AD,$47,$FB,$3F,$C9,$4F,$46,$C7,$46,$BF,$13,$F5,$97
	.dc.b	$CC,$FF,$8D,$1B,$1F,$E0,$66,$BF,$EB,$DB,$F2,$61,$FC,$B2,$72,$DA
	.dc.b	$E4,$7F,$B9,$A7,$7C,$7E,$67,$FA,$7D,$F9,$48,$79,$4A,$7C,$0C,$3F
	.dc.b	$98,$BE,$D3,$27,$C2,$F8,$3F,$F1,$F3,$FE,$F8,$9F,$F6,$77,$B5,$E1
	.dc.b	$9C,$CF,$A6,$BF,$89,$BF,$23,$EA,$3F,$A3,$DF,$CA,$F8,$D7,$D3,$FF
	.dc.b	$A8,$A7,$34,$9C,$BC,$7C,$BF,$F3,$9C,$63,$B1,$F5,$3F,$E6,$B8,$DD
	.dc.b	$2F,$9C,$61,$B8,$77,$C3,$9B,$F7,$30,$E7,$E3,$7F,$AD,$2F,$85,$7F
	.dc.b	$A0,$30,$E1,$97,$F5,$6C,$7C,$F1,$FE,$4F,$6B,$EB,$F6,$5F,$21,$B6
	.dc.b	$C7,$F9,$2A,$51,$88,$E4,$D7,$94,$FC,$7C,$EC,$7C,$57,$D1,$F5,$2F
	.dc.b	$41,$C5,$BA,$FA,$18,$78,$34,$F8,$4D,$C3,$9C,$7C,$BC,$39,$5F,$1B
	.dc.b	$F9,$57,$B5,$9B,$F2,$1F,$E9,$B5,$F8,$36,$CF,$E6,$72,$FD,$3C,$E7
	.dc.b	$C8,$F2,$CE,$4F,$F4,$79,$F9,$79,$DB,$FB,$26,$BF,$B2,$6B,$B3,$89
	.dc.b	$E4,$72,$D3,$37,$F1,$5D,$F6,$3E,$B3,$FC,$F3,$87,$F9,$D9,$B3,$FB
	.dc.b	$1F,$8E,$FF,$5D,$4E,$7E,$07,$D3,$D5,$FE,$1E,$BD,$B6,$33,$3F,$B1
	.dc.b	$AE,$36,$AD,$4F,$EA,$67,$33,$FC,$39,$AF,$1E,$4F,$4C,$62,$3E,$8B
	.dc.b	$72,$34,$93,$F1,$3B,$FF,$B6,$7B,$BE,$4B,$FE,$53,$E4,$7F,$9C,$71
	.dc.b	$FF,$4A,$FD,$53,$E4,$35,$87,$09,$CC,$7C,$89,$D4,$7F,$41,$6D,$FF
	.dc.b	$40,$6B,$FC,$F1,$AF,$FA,$9D,$AF,$F9,$2D,$BB,$9C,$F5,$7F,$85,$3E
	.dc.b	$9B,$D0,$DB,$FF,$04,$7B,$18,$F2,$55,$AF,$29,$3F,$84,$B4,$62,$FF
	.dc.b	$B0,$CE,$63,$19,$7F,$B1,$F4,$FE,$77,$C3,$F0,$BC,$5F,$CF,$E1,$FE
	.dc.b	$32,$F8,$9C,$4C,$8F,$E8,$7C,$9F,$9C,$BE,$31,$B3,$27,$F4,$87,$38
	.dc.b	$7E,$0E,$31,$87,$68,$C1,$FD,$79,$C5,$7F,$84,$70,$F0,$4E,$B3,$FF
	.dc.b	$1C,$C7,$BF,$53,$49,$87,$F9,$B3,$F3,$0F,$E5,$EF,$FB,$E9,$B7,$FE
	.dc.b	$8B,$FA,$9B,$8F,$ED,$A1,$38,$AF,$EC,$30,$AE,$2F,$FB,$BD,$AE,$9A
	.dc.b	$FF,$68,$6E,$7F,$DA,$3D,$AF,$9C,$7D,$EF,$A9,$FF,$6D,$DF,$C8,$BF
	.dc.b	$85,$3A,$8C,$57,$17,$5A,$2F,$8D,$FC,$AA,$D3,$F1,$E3,$39,$7F,$12
	.dc.b	$AE,$1F,$0F,$F8,$BF,$81,$F1,$37,$1F,$F1,$F4,$7C,$D2,$3F,$0B,$0F
	.dc.b	$B3,$F1,$1F,$87,$3F,$29,$E8,$D9,$F3,$F0,$A7,$15,$F6,$78,$57,$8C
	.dc.b	$FF,$D5,$3F,$48,$F9,$DF,$DC,$74,$A3,$19,$FC,$0F,$8C,$FE,$7E,$FC
	.dc.b	$79,$D5,$70,$AD,$E5,$F8,$C9,$CD,$71,$35,$5F,$C1,$67,$FD,$7D,$F4
	.dc.b	$76,$8E,$7C,$AF,$F6,$AD,$FD,$5A,$DB,$63,$31,$F1,$4F,$EC,$3B,$FF
	.dc.b	$36,$6C,$FF,$AF,$A7,$71,$F2,$DA,$FE,$39,$AF,$F8,$C1,$AF,$23,$FC
	.dc.b	$9F,$FB,$57,$F6,$A9,$F8,$4D,$8F,$D5,$FC,$23,$EA,$9F,$0D,$F3,$2B
	.dc.b	$5B,$FF,$B6,$1B,$87,$F8,$FB,$44,$63,$16,$3C,$F0,$E5,$E5,$FF,$07
	.dc.b	$F3,$A6,$3F,$96,$F1,$3F,$D9,$AF,$E0,$A3,$38,$7E,$57,$9B,$F5,$8F
	.dc.b	$F3,$4F,$85,$CE,$BF,$B1,$EE,$FE,$03,$87,$EC,$1B,$5F,$E5,$2F,$AA
	.dc.b	$7C,$FF,$D9,$9C,$CA,$73,$5F,$EE,$C6,$C7,$D5,$6B,$F9,$B6,$BF,$BE
	.dc.b	$6B,$DE,$BE,$D4,$62,$3B,$47,$F4,$FC,$5F,$79,$FC,$A4,$FE,$97,$A3
	.dc.b	$EA,$B1,$8F,$E9,$26,$E1,$FB,$7A,$D4,$E4,$FD,$4F,$D5,$FF,$33,$3F
	.dc.b	$A1,$F5,$7F,$9A,$A3,$24,$AF,$99,$3A,$F6,$F8,$4F,$E9,$7A,$FF,$64
	.dc.b	$DE,$EF,$D0,$7D,$5F,$E6,$38,$7F,$63,$18,$EB,$95,$F9,$FE,$67,$B9
	.dc.b	$8D,$3F,$9B,$8C,$79,$57,$F2,$66,$C7,$FA,$22,$B8,$CF,$EA,$67,$3E
	.dc.b	$27,$FA,$2B,$BF,$FA,$81,$F2,$3E,$6A,$D7,$F0,$A8,$B8,$7C,$79,$BA
	.dc.b	$7F,$51,$E4,$7D,$07,$F5,$93,$E8,$F0,$5C,$7F,$89,$C5,$C3,$E8,$4D
	.dc.b	$BF,$F1,$AF,$77,$F6,$CF,$92,$FE,$23,$E7,$1F,$E0,$DC,$6F,$82,$CC
	.dc.b	$61,$FD,$F4,$67,$3F,$E0,$59,$7F,$19,$9C,$92,$33,$3B,$9F,$E1,$0D
	.dc.b	$7F,$97,$35,$FE,$CF,$F4,$3F,$DD,$4F,$E9,$1F,$0D,$6B,$F9,$25,$FF
	.dc.b	$80,$7A,$38,$5B,$CA,$46,$1B,$3F,$C5,$7E,$AF,$FA,$2D,$FF,$5E,$6D
	.dc.b	$F1,$F3,$3C,$DB,$FF,$65,$9C,$F2,$FF,$B0,$BE,$57,$FC,$10,$7F,$5B
	.dc.b	$F5,$AE,$3F,$D5,$B0,$FF,$18,$63,$FD,$27,$3F,$EB,$15,$FE,$17,$9C
	.dc.b	$CF,$F6,$E9,$CD,$7F,$2B,$6C,$71,$9D,$BF,$E9,$96,$BE,$7A,$FF,$11
	.dc.b	$7D,$57,$AA,$8D,$FA,$B7,$CA,$FF,$8B,$A3,$0F,$F8,$A6,$33,$8F,$85
	.dc.b	$B8,$7A,$5A,$F9,$DA,$FF,$AA,$DA,$FD,$16,$D9,$CE,$7F,$EB,$AE,$AF
	.dc.b	$F2,$8F,$C2,$FF,$58,$DF,$FB,$B6,$6F,$85,$CD,$79,$DF,$69,$33,$F8
	.dc.b	$D3,$B6,$C6,$73,$FA,$13,$96,$C7,$F9,$1F,$A9,$F1,$9F,$E5,$06,$DF
	.dc.b	$F8,$4F,$17,$D5,$BE,$AF,$57,$FE,$6D,$9C,$C7,$FA,$51,$B5,$FD,$8D
	.dc.b	$B8,$FF,$82,$5A,$FF,$8B,$1A,$FF,$93,$2F,$B4,$F1,$6C,$71,$6B,$FE
	.dc.b	$34,$D5,$EB,$BE,$6B,$FC,$F6,$DB,$FF,$13,$E6,$F9,$6C,$D7,$BD,$F8
	.dc.b	$44,$9F,$C8,$4E,$DB,$19,$C3,$F3,$53,$96,$C7,$F9,$C7,$77,$DC,$FF
	.dc.b	$9E,$9B,$7F,$E3,$8F,$AC,$F5,$EB,$FE,$33,$9C,$E1,$FE,$5D,$9F,$6B
	.dc.b	$FE,$F5,$6B,$FB,$B6,$F2,$1A,$4D,$FF,$D0,$4D,$C3,$FC,$29,$B3,$E2
	.dc.b	$DF,$59,$FE,$A0,$6C,$FF,$93,$32,$B9,$7F,$92,$E6,$E7,$E7,$BF,$71
	.dc.b	$2F,$F6,$D7,$C2,$FE,$A7,$5B,$C5,$FF,$56,$B6,$BF,$CB,$3F,$59,$FE
	.dc.b	$3A,$AF,$F2,$C4,$E5,$B5,$C0,$FF,$51,$CE,$F1,$FF,$48,$56,$DF,$F4
	.dc.b	$97,$AE,$FA,$FF,$F5,$DB,$71,$FF,$34,$BF,$E0,$29,$DC,$E8,$93,$AF
	.dc.b	$4E,$E5,$FD,$78,$1F,$24,$7D,$67,$E3,$DF,$F1,$C4,$EA,$31,$1F,$E8
	.dc.b	$8D,$B1,$CD,$FE,$3F,$AB,$3F,$E6,$B9,$AF,$F8,$FE,$2B,$E0,$8B,$EE
	.dc.b	$27,$33,$E3,$E8,$EE,$A7,$4F,$E9,$2F,$FC,$A9,$7E,$AA,$DD,$69,$E3
	.dc.b	$5E,$DC,$33,$87,$6B,$F0,$B6,$7E,$0E,$1E,$B6,$9E,$B2,$9E,$B2,$9E
	.dc.b	$B3,$FB,$29,$F2,$CF,$EC,$F3,$87,$07,$FA,$C3,$EA,$DE,$1B,$EB,$7F
	.dc.b	$B6,$C6,$1B,$19,$9F,$F3,$94,$E5,$BE,$43,$7E,$2D,$8C,$C6,$08,$FA
	.dc.b	$A7,$FC,$1A,$FD,$66,$B9,$9E,$59,$FE,$02,$51,$8C,$BD,$5D,$9F,$EC
	.dc.b	$AA,$D3,$63,$DD,$3B,$BF,$C7,$DF,$2C,$3E,$C5,$72,$9A,$6A,$FF,$DA
	.dc.b	$BA,$F8,$70,$F6,$36,$FD,$13,$DE,$B8,$1E,$BA,$F0,$2C,$3C,$0F,$C1
	.dc.b	$9E,$0E,$5B,$1A,$C7,$F4,$5C,$FF,$A6,$FB,$FF,$E7,$9B,$FF,$C0,$1C
	.dc.b	$11,$85,$3B,$7F,$DB,$A3,$EC,$75,$19,$8F,$F6,$E1,$C8,$DC,$E9,$E0
	.dc.b	$7F,$B9,$BE,$1D,$F1,$3C,$9B,$19,$8C,$7D,$4E,$59,$19,$39,$47,$9F
	.dc.b	$BF,$E2,$7B,$FF,$E3,$C6,$CF,$F8,$57,$54,$E5,$B1,$9F,$77,$94,$7F
	.dc.b	$A0,$E3,$FD,$F2,$D7,$F2,$9A,$FF,$0E,$BE,$B1,$83,$FC,$75,$F2,$FE
	.dc.b	$A7,$FC,$54,$D7,$FC,$50,$D7,$93,$5E,$39,$7A,$63,$12,$69,$8C,$7F
	.dc.b	$84,$D7,$86,$7F,$D7,$6F,$F2,$2E,$2F,$E1,$B1,$88,$D5,$79,$67,$3F
	.dc.b	$5C,$FB,$DF,$90,$DA,$FF,$1E,$35,$FF,$22,$35,$FF,$21,$35,$F8,$8D
	.dc.b	$7F,$CB,$8D,$B3,$4F,$F4,$A4,$79,$F2,$E3,$EC,$FD,$B9,$CC,$FB,$5F
	.dc.b	$16,$1E,$CF,$A8,$78,$7F,$7D,$B5,$FE,$53,$6B,$FE,$5B,$6B,$FE,$59
	.dc.b	$6B,$F6,$DA,$FF,$9C,$DB,$89,$A7,$F0,$CA,$F1,$73,$7E,$46,$31,$97
	.dc.b	$3C,$3F,$29,$ED,$3D,$1F,$CA,$37,$2F,$F3,$43,$5F,$F3,$83,$5F,$F3
	.dc.b	$73,$5F,$C8,$35,$FF,$44,$37,$23,$98,$7D,$AB,$E1,$BF,$79,$18,$6C
	.dc.b	$66,$3E,$91,$D2,$FF,$59,$F5,$DF,$64,$9C,$1F,$AD,$6C,$FF,$A0,$5A
	.dc.b	$FF,$A2,$5F,$C1,$3E,$2F,$8D,$EE,$47,$BE,$49,$FB,$BA,$7E,$B7,$EB
	.dc.b	$7F,$DB,$75,$FE,$8D,$6B,$FE,$95,$6B,$FE,$93,$6B,$C7,$E0,$7F,$81
	.dc.b	$46,$23,$FD,$5C,$DB,$39,$87,$CA,$BE,$63,$E3,$B6,$3E,$29,$A8,$FE
	.dc.b	$21,$3E,$4E,$3F,$F6,$AF,$5B,$E9,$BF,$95,$7C,$12,$31,$87,$CE,$8C
	.dc.b	$4E,$A5,$F4,$DF,$C6,$FF,$66,$FF,$33,$DF,$FA,$B5,$AF,$FA,$D5,$AF
	.dc.b	$FA,$CD,$AF,$F8,$88,$F5,$FB,$3F,$CD,$BD,$27,$CB,$53,$DB,$1F,$84
	.dc.b	$FC,$09,$F5,$FF,$E0,$06,$DF,$FB,$99,$AF,$53,$FC,$5E,$E1,$C3,$5F
	.dc.b	$13,$5D,$C7,$53,$FC,$1B,$F6,$EF,$9F,$FC,$AB,$67,$E7,$B5,$FE,$5C
	.dc.b	$2B,$C7,$BF,$F8,$AB,$C9,$FF,$2D,$7C,$F7,$BF,$E0,$E5,$FE,$72,$72
	.dc.b	$DD,$55,$A8,$FE,$75,$3F,$E6,$3A,$F9,$13,$F4,$FC,$0F,$95,$F3,$B2
	.dc.b	$FF,$33,$C1,$FC,$3B,$2F,$E5,$ED,$EA,$FF,$29,$E5,$FE,$6D,$BE,$17
	.dc.b	$FC,$DE,$B4,$B9,$5F,$B5,$FF,$85,$23,$F0,$DF,$B5,$BE,$53,$F1,$6D
	.dc.b	$8F,$C5,$4E,$DF,$EA,$55,$CB,$D2,$DD,$C9,$DC,$7F,$7D,$F6,$DF,$AD
	.dc.b	$BE,$87,$0D,$AF,$E0,$73,$A9,$E5,$1D,$B2,$7C,$CF,$90,$91,$F1,$3D
	.dc.b	$9A,$EE,$2F,$0F,$C6,$9F,$CA,$7A,$B2,$BE,$A3,$F9,$7F,$5B,$11,$F5
	.dc.b	$39,$FF,$C1,$8D,$9D,$3F,$B4,$FA,$6F,$F2,$8B,$F9,$53,$FD,$06,$E1
	.dc.b	$C3,$5F,$CD,$35,$FD,$43,$5E,$37,$FE,$62,$8C,$41,$BF,$5E,$F9,$0F
	.dc.b	$E5,$AD,$9F,$E0,$ED,$7F,$33,$3E,$3C,$BF,$0F,$97,$ED,$7E,$06,$57
	.dc.b	$E3,$23,$F8,$11,$87,$E1,$46,$1F,$C0,$FC,$BF,$F0,$EE,$53,$4F,$BD
	.dc.b	$83,$E6,$31,$5C,$08,$7F,$86,$78,$BD,$91,$F8,$5C,$DF,$23,$F2,$E9
	.dc.b	$49,$CC,$7C,$F8,$C4,$ED,$B1,$99,$DB,$FC,$26,$32,$D8,$CB,$FC,$66
	.dc.b	$32,$FF,$0F,$87,$87,$F3,$89,$1F,$96,$8E,$BA,$FF,$03,$E1,$7C,$27
	.dc.b	$F2,$12,$57,$EE,$39,$6B,$F8,$D8,$B5,$FB,$A9,$FE,$50,$49,$FE,$10
	.dc.b	$FF,$93,$FF,$0F,$FE,$45,$3D,$E3,$53,$F9,$77,$4E,$FC,$4F,$F2,$A5
	.dc.b	$7F,$1B,$8F,$E6,$DE,$6A,$31,$1F,$E4,$D6,$A8,$7C,$B7,$F1,$88,$FE
	.dc.b	$D2,$D7,$FC,$84,$DC,$3F,$C8,$35,$FC,$A5,$FF,$21,$C6,$7C,$1F,$E2
	.dc.b	$B9,$FF,$59,$6F,$9F,$FE,$6B,$9D,$61,$E1,$6F,$84,$BE,$13,$FE,$86
	.dc.b	$6E,$5E,$6F,$13,$FC,$1F,$F5,$AF,$8C,$FF,$65,$B6,$FF,$18,$D7,$F7
	.dc.b	$0D,$7F,$A7,$3E,$9C,$FF,$AE,$36,$3F,$DF,$ED,$4F,$3D,$3E,$0F,$C8
	.dc.b	$7D,$02,$4F,$F0,$5C,$7E,$84,$7F,$A1,$F4,$FF,$4A,$CF,$CA,$EB,$F8
	.dc.b	$47,$DF,$9B,$1F,$42,$2B,$FE,$71,$DD,$38,$C3,$E7,$D5,$CB,$EE,$4F
	.dc.b	$E2,$7E,$B7,$F0,$64,$C3,$C9,$82,$3F,$0F,$8F,$E3,$C9,$3F,$E2,$89
	.dc.b	$D3,$FE,$0B,$FD,$6F,$F9,$9B,$FC,$87,$C6,$46,$23,$FB,$C4,$63,$C2
	.dc.b	$FD,$E3,$63,$84,$ED,$FF,$05,$46,$5E,$31,$94,$7F,$59,$70,$BE,$15
	.dc.b	$F5,$1F,$B2,$8F,$F2,$6C,$63,$A3,$FD,$9C,$D8,$F5,$61,$8F,$57,$F0
	.dc.b	$51,$DA,$7F,$C6,$9C,$7F,$C4,$DB,$1F,$CF,$DA,$FF,$85,$5A,$FF,$88
	.dc.b	$DA,$F3,$9E,$AF,$9D,$B1,$E8,$E0,$F2,$E3,$F8,$4E,$17,$0F,$8E,$30
	.dc.b	$F3,$C7,$75,$31,$F7,$D6,$9E,$D3,$ED,$E5,$F5,$65,$FA,$B9,$F5,$35
	.dc.b	$F5,$35,$F5,$35,$F5,$47,$22,$3F,$0F,$E1,$7F,$D5,$39,$7A,$23,$EA
	.dc.b	$46,$32,$F9,$8D,$8F,$85,$5F,$E8,$6F,$AB,$FE,$5F,$A7,$DE,$F8,$CF
	.dc.b	$C5,$46,$29,$18,$EE,$FC,$04,$63,$7F,$F7,$37,$CB,$7F,$82,$E3,$07
	.dc.b	$E2,$2F,$95,$F1,$C3,$B5,$FF,$96,$67,$2F,$2E,$EF,$F2,$E5,$6E,$FF
	.dc.b	$BF,$57,$8D,$FF,$2A,$35,$E5,$8F,$91,$CF,$C0,$3E,$3B,$FA,$E5,$7F
	.dc.b	$A9,$5B,$19,$9E,$AD,$78,$C7,$99,$F5,$46,$EF,$F6,$3D,$6F,$AB,$FF
	.dc.b	$45,$70,$6E,$D8,$F7,$7E,$19,$8C,$39,$69,$F0,$5B,$1F,$D0,$6F,$9B
	.dc.b	$C2,$B9,$63,$FC,$03,$57,$C4,$39,$BF,$01,$C7,$F8,$19,$CC,$63,$4F
	.dc.b	$C0,$FD,$4D,$CB,$F7,$18,$71,$9F,$DB,$F2,$2B,$F1,$55,$FA,$B8,$C4
	.dc.b	$FE,$3E,$73,$3D,$23,$04,$7F,$69,$3C,$1B,$3C,$1F,$E1,$1E,$23,$C3
	.dc.b	$F2,$65,$FE,$0F,$6D,$FF,$62,$C3,$F8,$5D,$6D,$B3,$C2,$31,$6C,$FE
	.dc.b	$15,$F1,$8C,$5F,$F0,$DA,$DD,$6B,$BB,$C9,$8B,$0E,$18,$7C,$19,$CC
	.dc.b	$62,$3F,$57,$18,$26,$1F,$B3,$79,$E0,$F2,$FC,$C7,$D1,$EE,$FF,$39
	.dc.b	$F2,$E2,$FF,$D8,$DE,$87,$E5,$DF,$E7,$8D,$C7,$F9,$DB,$5F,$E6,$ED
	.dc.b	$7F,$9A,$D7,$49,$E3,$5F,$3E,$31,$3C,$A3,$E9,$73,$7F,$84,$FE,$5B
	.dc.b	$F0,$8D,$BF,$EA,$1F,$29,$8F,$09,$56,$38,$C1,$A4,$9F,$EE,$1F,$B4
	.dc.b	$FA,$78,$7F,$8A,$60,$8C,$4F,$F8,$87,$E3,$3C,$31,$8C,$FF,$C4,$4D
	.dc.b	$AE,$19,$F8,$EB,$6C,$BF,$EA,$B8,$76,$BE,$19,$2F,$58,$FF,$79,$7D
	.dc.b	$24,$BF,$EE,$6F,$E5,$61,$F5,$5F,$E6,$B9,$EA,$7F,$BB,$DC,$47,$FB
	.dc.b	$59,$FF,$5C,$46,$A3,$FB,$3C,$6B,$0F,$D5,$BA,$8F,$F4,$C3,$5F,$F7
	.dc.b	$73,$5E,$DE,$BF,$D6,$C7,$FC,$59,$7F,$ED,$BF,$7B,$E3,$F8,$9F,$BB
	.dc.b	$F3,$3F,$C6,$D1,$89,$D1,$E3,$DB,$E8,$56,$FC,$4F,$F8,$03,$39,$C3
	.dc.b	$AE,$5F,$A6,$3E,$8D,$7E,$07,$EA,$7A,$47,$F8,$85,$FC,$09,$2B,$FC
	.dc.b	$36,$FE,$F7,$27,$EB,$5F,$E0,$89,$FF,$7A,$36,$BF,$C5,$5F,$07,$E9
	.dc.b	$8C,$6F,$FE,$FD,$6D,$7F,$BB,$1B,$5F,$E6,$5E,$15,$F1,$8E,$5C,$52
	.dc.b	$FE,$07,$E5,$6F,$89,$C5,$7F,$C0,$0C,$7F,$1D,$9F,$9B,$0F,$D0,$DE
	.dc.b	$F0,$F3,$C7,$F2,$7B,$FF,$7C,$78,$B7,$A0,$FF,$2B,$F4,$FF,$D6,$1E
	.dc.b	$CE,$63,$1E,$B7,$FB,$1F,$1F,$F5,$43,$70,$FF,$52,$35,$FF,$51,$35
	.dc.b	$FF,$17,$DF,$49,$E3,$7F,$CA,$63,$13,$CA,$3F,$7D,$EC,$FF,$1C,$AF
	.dc.b	$E0,$6D,$9F,$F5,$DD,$BE,$C7,$F8,$94,$E7,$23,$4E,$18,$FF,$B2,$E7
	.dc.b	$7E,$27,$C6,$F3,$FE,$C7,$45,$CF,$EE,$B7,$3F,$9C,$DE,$5F,$B8,$FC
	.dc.b	$BD,$FF,$9A,$47,$F4,$E6,$D7,$F9,$3A,$9F,$51,$F9,$67,$F2,$87,$F7
	.dc.b	$BF,$13,$EA,$46,$63,$81,$C3,$B3,$F7,$51,$98,$E8,$D7,$FC,$B6,$D7
	.dc.b	$85,$7F,$B8,$63,$06,$F9,$3F,$37,$EC,$7E,$5C,$FC,$C7,$5D,$AF,$CE
	.dc.b	$C5,$BF,$C7,$4D,$CF,$FC,$69,$C4,$F7,$9E,$8F,$F4,$36,$E3,$F9,$8E
	.dc.b	$2C,$4C,$8F,$E1,$95,$C6,$3F,$2F,$39,$C7,$FD,$F9,$E0,$3E,$37,$F2
	.dc.b	$8D,$C3,$8E,$3F,$EE,$67,$F8,$B7,$2A,$75,$7F,$C0,$72,$FC,$1F,$C1
	.dc.b	$79,$7D,$6F,$C3,$B7,$0F,$1F,$42,$31,$7C,$63,$13,$E0,$AF,$D6,$FD
	.dc.b	$BF,$EF,$FE,$D2,$CF,$F0,$46,$DF,$ED,$5B,$C1,$FC,$0D,$FE,$0D,$F3
	.dc.b	$5F,$C1,$FD,$CB,$7F,$C3,$1B,$7F,$BD,$6D,$FF,$0B,$7F,$0F,$DB,$2B
	.dc.b	$F9,$B5,$3E,$97,$E0,$DF,$F8,$83,$A3,$F9,$C5,$7E,$4E,$7E,$2D,$7E
	.dc.b	$39,$AE,$78,$FF,$A4,$46,$23,$FD,$4E,$D7,$B4,$68,$93,$F9,$88,$FF
	.dc.b	$56,$FC,$EF,$F0,$C8,$FE,$4F,$F2,$BF,$C4,$63,$13,$FC,$66,$3F,$BB
	.dc.b	$F6,$7F,$AC,$6B,$FD,$BC,$D9,$FD,$1F,$94,$8C,$47,$EB,$63,$1E,$9F
	.dc.b	$F3,$7F,$19,$F4,$1F,$CC,$CF,$E6,$9F,$01,$71,$F8,$F9,$FF,$25,$C7
	.dc.b	$F6,$6D,$DF,$FB,$EF,$D2,$3E,$13,$D1,$E5,$FF,$83,$23,$18,$7F,$BC
	.dc.b	$DB,$87,$99,$AF,$95,$AF,$FB,$43,$54,$62,$11,$8B,$FF,$0D,$CE,$63
	.dc.b	$15,$FE,$B0,$3E,$FE,$BF,$E6,$CC,$7F,$C2,$7F,$48,$F8,$5F,$E7,$D3
	.dc.b	$9A,$F0,$4E,$5F,$AD,$68,$C6,$5F,$39,$AF,$CD,$6B,$CA,$FF,$C5,$F1
	.dc.b	$88,$F4,$B6,$7A,$4F,$F8,$46,$DF,$03,$E4,$3F,$C6,$BF,$49,$F8,$2C
	.dc.b	$BF,$CE,$B3,$AC,$BF,$BF,$4F,$E1,$B2,$FB,$CD,$7E,$53,$5F,$92,$DC
	.dc.b	$4D,$27,$9D,$FE,$46,$75,$F8,$67,$E1,$AF,$FC,$18,$DB,$FF,$04,$5A
	.dc.b	$1C,$3F,$E4,$C7,$59,$7F,$90,$9C,$DF,$CC,$6D,$FE,$61,$AF,$E5,$DA
	.dc.b	$FE,$0D,$AF,$E0,$9B,$7E,$1B,$FF,$2E,$C6,$3E,$6D,$CD,$FE,$5E,$6E
	.dc.b	$1F,$E2,$57,$FA,$8E,$1F,$8C,$72,$FB,$1F,$F4,$03,$9E,$6F,$F3,$4C
	.dc.b	$63,$13,$49,$1A,$25,$7E,$2B,$E9,$3F,$C3,$74,$68,$F3,$BF,$CE,$6E
	.dc.b	$1F,$F3,$BB,$99,$E5,$18,$23,$82,$61,$FC,$0A,$78,$3F,$C1,$3E,$43
	.dc.b	$E7,$EB,$FE,$74,$6D,$7E,$BD,$F8,$78,$1C,$D7,$EC,$1F,$80,$4C,$7F
	.dc.b	$D1,$6F,$C6,$F3,$5D,$1F,$E3,$86,$D7,$F9,$5E,$3F,$C6,$26,$48,$E7
	.dc.b	$CF,$D3,$FE,$5D,$F2,$5F,$1D,$4C,$7C,$1D,$D3,$FC,$E1,$8A,$B7,$18
	.dc.b	$6D,$FF,$97,$67,$55,$C6,$3F,$C1,$A0,$7C,$25,$FA,$17,$8D,$FE,$0E
	.dc.b	$CD,$96,$63,$F9,$D5,$5B,$FF,$0D,$56,$95,$F2,$7F,$C2,$F3,$B7,$C1
	.dc.b	$3B,$9D,$47,$F8,$8E,$74,$FF,$B0,$E3,$2F,$9E,$3D,$1F,$45,$8A,$F0
	.dc.b	$3A,$9F,$58,$75,$87,$0C,$3D,$33,$E6,$7D,$D1,$E0,$8E,$38,$70,$74
	.dc.b	$7A,$8A,$7A,$A3,$86,$1E,$8C,$3D,$97,$DA,$BF,$B0,$1E,$42,$CE,$1F
	.dc.b	$61,$FD,$97,$E1,$1F,$20,$FA,$3E,$31,$86,$C6,$67,$83,$5F,$F2,$2C
	.dc.b	$BE,$49,$CB,$C9,$B3,$DD,$AF,$06,$BF,$E5,$59,$E5,$3F,$26,$3F,$11
	.dc.b	$1F,$C0,$9C,$07,$0F,$D1,$C7,$95,$78,$35,$FB,$F5,$D2,$B4,$74,$E8
	.dc.b	$FF,$32,$4E,$9B,$7C,$B0,$EF,$7C,$6B,$57,$8B,$DB,$E6,$C3,$C1,$87
	.dc.b	$18,$EE,$79,$E7,$6D,$3A,$CF,$07,$96,$1D,$58,$72,$C3,$C5,$7F,$AB
	.dc.b	$E0,$FD,$67,$C3,$7E,$01,$B5,$FE,$21,$6B,$FE,$7A,$6B,$FE,$83,$6B
	.dc.b	$CC,$47,$36,$9C,$CA,$7F,$88,$BD,$AF,$F5,$BB,$FD,$E9,$AE,$7B,$FF
	.dc.b	$27,$D2,$9C,$CF,$F8,$9D,$AF,$F8,$4D,$AF,$F8,$49,$B1,$C6,$79,$E2
	.dc.b	$FA,$3A,$4B,$FF,$05,$C6,$5F,$F0,$6F,$C5,$3E,$F7,$C6,$D8,$CC,$60
	.dc.b	$F3,$FD,$1F,$F0,$D8,$FF,$1F,$B5,$FF,$1F,$35,$FF,$15,$35,$FF,$14
	.dc.b	$B7,$13,$49,$CD,$F5,$FC,$F7,$A7,$49,$7F,$66,$32,$FF,$89,$BE,$8D
	.dc.b	$F1,$1F,$31,$B1,$98,$C1,$1C,$10,$FF,$2C,$C7,$07,$FD,$7B,$C6,$FB
	.dc.b	$D7,$D2,$FF,$0A,$8C,$36,$33,$3F,$E6,$D9,$CB,$FE,$3A,$F1,$BF,$66
	.dc.b	$FF,$98,$DA,$FF,$93,$DA,$FF,$93,$9B,$AF,$F8,$2B,$A3,$F0,$F1,$82
	.dc.b	$3E,$1F,$A3,$95,$E1,$C1,$18,$8F,$F4,$07,$D2,$7E,$29,$34,$9C,$BF
	.dc.b	$C8,$DB,$1F,$CA,$1A,$FF,$9F,$BE,$91,$F2,$5F,$9A,$6C,$7F,$96,$1A
	.dc.b	$EE,$3A,$7A,$92,$3A,$35,$F2,$B5,$FF,$76,$D2,$31,$1F,$D9,$FD,$F7
	.dc.b	$DC,$7E,$91,$B8,$FF,$3E,$6C,$FF,$A7,$DA,$FF,$A8,$7E,$97,$F6,$0F
	.dc.b	$F4,$46,$B9,$EC,$FE,$89,$18,$8F,$F7,$FB,$5F,$EF,$01,$C3,$FD,$ED
	.dc.b	$AE,$63,$F7,$F7,$FD,$8E,$3F,$CF,$AD,$7F,$D5,$2D,$7F,$D5,$0D,$7F
	.dc.b	$D2,$9F,$3D,$5F,$F4,$9B,$78,$4D,$35,$1E,$5F,$53,$D6,$47,$E0,$3F
	.dc.b	$C4,$BC,$E7,$CF,$52,$4F,$FA,$F9,$FE,$CD,$F4,$2F,$8A,$A4,$9F,$ED
	.dc.b	$B1,$96,$C6,$5F,$F6,$84,$65,$B1,$FD,$C7,$E7,$DF,$41,$49,$3F,$DE
	.dc.b	$23,$2F,$F3,$9F,$7F,$FD,$9C,$DC,$7F,$9F,$46,$23,$F9,$E3,$5F,$E7
	.dc.b	$8D,$7F,$A3,$35,$78,$E8,$C3,$67,$FA,$57,$D7,$7D,$E8,$34,$9C,$1A
	.dc.b	$25,$FF,$4F,$8C,$BF,$EF,$1F,$49,$F3,$2F,$8D,$FF,$0A,$C6,$1B,$19
	.dc.b	$9E,$0D,$7F,$22,$7F,$94,$5C,$3F,$94,$9E,$4F,$09,$E6,$7E,$97,$EE
	.dc.b	$5F,$0F,$49,$3F,$A7,$8C,$BE,$7E,$2F,$C8,$5F,$FB,$1B,$D0,$F3,$C6
	.dc.b	$08,$E0,$93,$C5,$FA,$BF,$33,$23,$ED,$35,$FA,$ED,$7E,$BB,$5F,$D7
	.dc.b	$FC,$DB,$E5,$3F,$C7,$0D,$8F,$F1,$EB,$5F,$8A,$1C,$3F,$21,$AF,$48
	.dc.b	$E6,$47,$9E,$3D,$88,$E0,$FC,$97,$CD,$FE,$40,$8E,$08,$7E,$49,$CB
	.dc.b	$4F,$E0,$3A,$7F,$04,$23,$82,$1F,$C1,$1C,$B4,$F7,$E5,$F0,$88,$E0
	.dc.b	$87,$C2,$C3,$FD,$FA,$E0,$F8,$51,$FC,$9F,$E1,$1F,$21,$49,$3F,$CA
	.dc.b	$63,$2F,$CC,$EB,$7D,$A7,$F9,$A5,$B1,$FE,$1C,$8C,$11,$C1,$0E,$4F
	.dc.b	$F9,$C7,$93,$F6,$1E,$1F,$EE,$89,$3F,$BF,$6B,$FB,$C6,$BF,$BC,$7F
	.dc.b	$8D,$5C,$8F,$37,$32,$31,$1F,$E3,$FF,$82,$7C,$B7,$F9,$15,$B1,$FE
	.dc.b	$45,$6B,$B8,$FE,$4D,$F2,$52,$3A,$B5,$FD,$9E,$6C,$13,$99,$FE,$8F
	.dc.b	$39,$8C,$47,$F4,$AF,$A3,$FE,$48,$FE,$F1,$AE,$7B,$BF,$D2,$D1,$88
	.dc.b	$7B,$EF,$C8,$3F,$D1,$D1,$FA,$08,$C7,$BB,$FE,$A0,$6C,$66,$76,$FA
	.dc.b	$32,$7C,$17,$F4,$32,$47,$CA,$7F,$96,$F9,$B2,$3F,$99,$35,$FE,$5E
	.dc.b	$D7,$F9,$7B,$5F,$E5,$ED,$7F,$97,$F8,$3F,$E1,$67,$99,$F8,$DF,$9B
	.dc.b	$FE,$69,$B1,$A4,$C3,$F9,$FF,$AB,$2B,$FA,$23,$5F,$E4,$BF,$16,$FA
	.dc.b	$37,$CA,$FF,$B0,$A3,$0D,$8C,$CF,$F0,$89,$CB,$63,$2F,$46,$BF,$E0
	.dc.b	$28,$DB,$99,$ED,$E3,$B9,$69,$27,$B4,$65,$ED,$B7,$8C,$8E,$08,$78
	.dc.b	$5C,$B4,$E6,$78,$08,$60,$87,$C4,$8F,$F7,$05,$9A,$63,$DF,$FF,$31
	.dc.b	$4F,$B5,$A8,$4F,$FB,$3F,$2F,$99,$3A,$9C,$B6,$75,$1F,$36,$30,$6F
	.dc.b	$CF,$7E,$6C,$FB,$1F,$F1,$D3,$6B,$FC,$74,$D7,$FB,$0C,$F0,$9F,$F7
	.dc.b	$D7,$D1,$3C,$0F,$2D,$3E,$21,$F4,$9F,$E3,$8A,$D6,$7E,$B9,$CD,$7B
	.dc.b	$67,$2F,$1A,$F8,$1C,$8F,$A5,$FE,$E5,$7C,$29,$7C,$23,$0D,$BF,$06
	.dc.b	$3E,$49,$EF,$87,$3F,$17,$FC,$97,$1C,$27,$D3,$87,$F7,$EC,$3C,$6D
	.dc.b	$8C,$BD,$32,$FE,$BE,$F8,$BD,$6F,$E0,$AF,$F6,$2A,$46,$23,$A3,$5F
	.dc.b	$88,$D7,$E1,$3C,$AF,$D6,$DC,$BE,$1F,$D2,$FF,$B6,$7E,$0F,$E5,$47
	.dc.b	$2E,$DB,$87,$26,$BE,$57,$3A,$AF,$94,$FC,$F3,$81,$78,$1F,$98,$BE
	.dc.b	$6F,$FD,$9D,$18,$6C,$67,$2F,$F4,$03,$6B,$F8,$1F,$C4,$7F,$1E,$9F
	.dc.b	$DF,$36,$33,$D1,$FE,$D9,$E0,$4E,$67,$F7,$CD,$7F,$45,$1A,$9E,$13
	.dc.b	$99,$FD,$2B,$67,$FD,$1F,$5F,$C3,$7D,$5F,$D5,$36,$BC,$11,$87,$F8
	.dc.b	$CB,$5F,$E1,$CD,$7C,$2D,$7F,$66,$D9,$FF,$10,$78,$DE,$BF,$33,$FA
	.dc.b	$FD,$FF,$1D,$6D,$FF,$26,$6B,$FC,$6D,$AF,$B9,$AF,$F0,$06,$D7,$69
	.dc.b	$FE,$51,$ED,$FF,$96,$DB,$CD,$F2,$E3,$17,$FC,$D1,$AF,$7F,$8A,$F3
	.dc.b	$12,$BC,$CD,$8F,$33,$5E,$11,$FE,$83,$FB,$33,$49,$1F,$E8,$28,$CB
	.dc.b	$FD,$DB,$C3,$7A,$18,$8F,$EE,$B1,$8A,$FE,$A5,$39,$9F,$E0,$33,$97
	.dc.b	$FD,$1B,$F1,$59,$7F,$AD,$B7,$CF,$FF,$45,$37,$C7,$34,$98,$7F,$A6
	.dc.b	$DB,$5D,$23,$04,$70,$7F,$83,$BE,$9B,$FC,$03,$1F,$E3,$36,$C6,$7A
	.dc.b	$7C,$9D,$29,$CC,$FF,$B6,$3A,$1F,$A4,$F9,$30,$FF,$73,$36,$7F,$96
	.dc.b	$B5,$E0,$D7,$FD,$C3,$F5,$1F,$AD,$7F,$D2,$6D,$73,$D5,$F1,$54,$65
	.dc.b	$63,$E5,$F1,$FC,$F9,$7F,$67,$C7,$D9,$18,$23,$82,$57,$C3,$8F,$A0
	.dc.b	$3E,$96,$92,$7F,$C7,$8F,$6F,$A5,$91,$F2,$5B,$5F,$EC,$8F,$A8,$FB
	.dc.b	$04,$9F,$F6,$83,$63,$51,$82,$3F,$E4,$2F,$59,$3F,$49,$B3,$FE,$60
	.dc.b	$8C,$11,$C1,$27,$94,$7F,$1E,$F3,$9F,$69,$49,$5F,$E9,$EE,$0F,$F2
	.dc.b	$84,$7C,$F7,$F9,$FC,$7F,$44,$6D,$EF,$C4,$D1,$27,$4D,$8D,$47,$F9
	.dc.b	$FF,$EC,$FF,$5A,$D3,$F1,$AE,$23,$FC,$C2,$D7,$FC,$BE,$D7,$FB,$4B
	.dc.b	$5F,$C0,$B7,$73,$9E,$6F,$F7,$97,$C7,$7E,$A4,$97,$FF,$00,$5B,$7F
	.dc.b	$EF,$E6,$BF,$EF,$E7,$F2,$1C,$AE,$AA,$4A,$FF,$32,$3A,$D6,$47,$F9
	.dc.b	$9D,$B7,$FE,$6B,$6B,$FE,$F0,$AF,$F3,$F4,$EC,$FF,$67,$E0,$E8,$7D
	.dc.b	$57,$F0,$12,$4E,$A3,$2D,$8C,$DF,$FB,$F2,$B9,$46,$23,$FD,$F9,$F6
	.dc.b	$FF,$E2,$5A,$FF,$6E,$FC,$0A,$31,$3F,$E3,$53,$FC,$6D,$19,$F3,$BF
	.dc.b	$8B,$36,$B8,$D7,$0A,$EB,$5A,$8C,$47,$26,$AE,$17,$BC,$F8,$0F,$EE
	.dc.b	$6D,$C3,$F2,$8D,$AF,$F1,$BB,$5F,$36,$3F,$ED,$0F,$A5,$F2,$FC,$01
	.dc.b	$A4,$BF,$7C,$65,$B1,$9A,$F8,$02,$BA,$4F,$D4,$F8,$57,$CA,$FF,$35
	.dc.b	$6C,$66,$30,$47,$04,$AF,$A9,$3C,$1F,$AD,$E7,$FE,$5C,$93,$FD,$05
	.dc.b	$AF,$E1,$1B,$1C,$23,$F9,$27,$CA,$7F,$A5,$1F,$F1,$CB,$63,$D6,$1C
	.dc.b	$3E,$E6,$BF,$7D,$FE,$58,$EF,$C2,$EA,$49,$EA,$D8,$EB,$3C,$0F,$E6
	.dc.b	$9D,$3F,$D1,$FE,$0F,$F9,$99,$2F,$FA,$73,$FC,$EF,$E7,$3F,$85,$DF
	.dc.b	$D0,$6B,$F9,$16,$BF,$E8,$76,$D7,$FB,$97,$CC,$FE,$AB,$D3,$FE,$8A
	.dc.b	$C3,$FD,$EF,$18,$3F,$86,$FD,$24,$8F,$E2,$4D,$C3,$E2,$E0,$F8,$FF
	.dc.b	$21,$2B,$FC,$C7,$19,$6C,$66,$78,$B5,$FE,$C9,$F4,$9F,$D7,$5F,$CF
	.dc.b	$37,$98,$D3,$51,$FE,$30,$AF,$F6,$67,$A5,$FD,$7F,$2F,$F2,$DC,$FF
	.dc.b	$8E,$A7,$24,$7F,$C8,$4D,$7F,$71,$A3,$E9,$1F,$1B,$FD,$E9,$B5,$F8
	.dc.b	$49,$D4,$E4,$F0,$7B,$72,$7F,$DB,$CD,$BF,$F6,$E3,$5F,$F7,$2B,$5F
	.dc.b	$F5,$E7,$CB,$BE,$D1,$F4,$9F,$E1,$D8,$C3,$63,$37,$F9,$69,$CB,$E5
	.dc.b	$FA,$47,$D3,$7F,$84,$9B,$19,$8C,$11,$C1,$30,$FD,$DE,$6A,$E3,$1F
	.dc.b	$BE,$BF,$37,$9B,$E3,$25,$F8,$F1,$47,$F8,$8B,$E6,$9F,$37,$49,$5F
	.dc.b	$2E,$32,$D8,$CE,$1E,$29,$CB,$63,$FC,$7D,$E4,$7E,$CA,$1E,$3A,$30
	.dc.b	$DF,$88,$D9,$FE,$34,$FA,$5F,$F3,$C6,$C6,$92,$F4,$4C,$BF,$01,$19
	.dc.b	$6E,$7C,$C8,$E0,$FC,$37,$D2,$FD,$6A,$F9,$6D,$7F,$1E,$D7,$FD,$BA
	.dc.b	$D7,$FC,$6B,$F3,$6F,$8A,$FF,$25,$B6,$BF,$24,$D9,$FA,$ED,$7F,$36
	.dc.b	$D7,$FD,$D2,$D7,$AC,$F4,$3F,$CD,$1E,$2B,$E4,$A9,$2F,$FB,$6C,$65
	.dc.b	$B1,$9A,$FE,$D1,$39,$6C,$7F,$94,$5F,$F7,$F0,$F9,$57,$D0,$FF,$7B
	.dc.b	$B4,$E6,$3F,$98,$FC,$4E,$2B,$73,$FD,$B7,$87,$FD,$17,$18,$8F,$F7
	.dc.b	$C9,$C4,$D5,$F1,$BE,$71,$93,$BD,$72,$8C,$D7,$39,$FF,$3D,$B5,$5F
	.dc.b	$37,$A9,$CC,$F6,$C7,$1A,$E9,$1F,$D9,$27,$83,$E0,$BA,$F0,$BF,$84
	.dc.b	$F3,$3B,$9D,$6B,$9B,$EE,$3F,$AA,$9A,$2D,$61,$F9,$E7,$F7,$4F,$8D
	.dc.b	$7D,$13,$E8,$F8,$C6,$1B,$19,$9F,$EC,$33,$96,$FC,$A3,$A7,$8B,$63
	.dc.b	$31,$82,$2F,$A4,$FA,$D4,$E9,$FF,$13,$46,$1B,$A7,$F8,$A6,$58,$75
	.dc.b	$AE,$2D,$9D,$57,$0B,$FC,$4D,$F5,$7F,$17,$7C,$30,$B7,$D7,$0F,$46
	.dc.b	$DE,$1B,$FA,$B8,$71,$7F,$0C,$F8,$AF,$35,$E1,$9E,$F8,$F8,$55,$71
	.dc.b	$C7,$C1,$1F,$07,$1F,$06,$1E,$28,$E2,$53,$8D,$F3,$BF,$04,$71,$3F
	.dc.b	$8E,$8E,$7F,$F2,$75,$7F,$BB,$3A,$3F,$53,$18,$E0,$FD,$5C,$62,$BF
	.dc.b	$A6,$C6,$17,$C4,$6D,$EB,$56,$B8,$47,$DF,$AE,$13,$98,$E2,$74,$3F
	.dc.b	$98,$F4,$FF,$BE,$F0,$F2,$36,$BC,$6D,$9D,$FD,$C7,$78,$C6,$1F,$ED
	.dc.b	$4A,$C1,$99,$F3,$4F,$CE,$FB,$4F,$D0,$B6,$FF,$BA,$66,$9C,$B6,$3F
	.dc.b	$D0,$B6,$FB,$2F,$73,$63,$D5,$8A,$31,$1F,$04,$FF,$73,$D7,$37,$18
	.dc.b	$7F,$97,$3C,$9F,$1F,$A5,$FE,$23,$9F,$F0,$33,$5F,$ED,$6D,$7F,$C1
	.dc.b	$ED,$7F,$90,$CF,$37,$8C,$F4,$3F,$7F,$87,$F9,$4F,$D8,$FA,$DE,$81
	.dc.b	$A7,$C6,$C3,$C1,$B7,$AF,$0F,$F7,$FF,$D2,$FF,$B9,$1F,$F0,$D3,$5F
	.dc.b	$EF,$ED,$7F,$C4,$6D,$7C,$AD,$79,$FC,$0F,$D2,$8C,$66,$75,$97,$F7
	.dc.b	$4C,$3F,$18,$FF,$A0,$AF,$E7,$B5,$F9,$73,$E3,$C3,$D1,$7F,$E0,$8E
	.dc.b	$6F,$BF,$E1,$FF,$04,$BF,$F1,$93,$5F,$F0,$BB,$5F,$F1,$DB,$5F,$9A
	.dc.b	$D5,$3D,$5E,$73,$F3,$BF,$05,$79,$BF,$A3,$6C,$FE,$19,$AF,$E0,$AF
	.dc.b	$C5,$87,$A3,$F0,$17,$9D,$F9,$66,$D7,$F9,$25,$AF,$F8,$B9,$AF,$F9
	.dc.b	$49,$AF,$E1,$1A,$FA,$1B,$91,$CE,$8F,$C0,$9E,$A7,$EE,$DB,$7F,$9E
	.dc.b	$6B,$F9,$AA,$F1,$E3,$F4,$3E,$9F,$FC,$21,$3F,$E5,$F6,$BF,$E4,$46
	.dc.b	$BF,$E6,$86,$BF,$9A,$6B,$F3,$9B,$89,$A6,$B1,$7E,$10,$F8,$9F,$E2
	.dc.b	$4D,$BF,$DA,$B5,$FD,$8D,$79,$30,$FC,$3F,$DC,$3E,$2B,$F8,$2B,$6B
	.dc.b	$FC,$F0,$D7,$FC,$BA,$D7,$FD,$02,$D7,$F6,$6D,$7F,$0E,$DC,$8E,$72
	.dc.b	$7B,$AF,$87,$FE,$54,$DB,$FE,$1A,$D7,$F8,$4D,$78,$F1,$FD,$B4,$3E
	.dc.b	$49,$EB,$7F,$33,$D9,$F6,$F4,$FE,$71,$5B,$9F,$E4,$BC,$4F,$A8,$FE
	.dc.b	$4A,$D9,$FF,$4C,$35,$FF,$3E,$B5,$FF,$18,$8F,$71,$F0,$EF,$A5,$FE
	.dc.b	$8D,$18,$6C,$66,$7F,$A1,$4F,$0F,$80,$F2,$FE,$04,$F8,$BF,$F0,$A3
	.dc.b	$6B,$FC,$32,$D7,$FD,$DE,$D7,$C6,$1C,$3E,$66,$BF,$E4,$86,$BB,$8F
	.dc.b	$F1,$17,$DC,$FF,$7B,$E0,$F9,$AD,$F1,$FD,$ED,$AE,$55,$FD,$87,$E1
	.dc.b	$9E,$5D,$24,$FF,$46,$8F,$E5,$B7,$FD,$B9,$DC,$66,$76,$FF,$49,$E7
	.dc.b	$4F,$08,$FF,$4B,$AF,$8E,$F0,$12,$3F,$0F,$E3,$D7,$94,$ED,$F1,$4F
	.dc.b	$E3,$A7,$EA,$FC,$B3,$6A,$4A,$FA,$BF,$16,$FA,$8C,$47,$F5,$DE,$15
	.dc.b	$F1,$8F,$B6,$D7,$55,$C5,$CD,$FC,$7E,$69,$18,$3B,$FC,$87,$CB,$F1
	.dc.b	$E4,$60,$87,$B1,$CB,$DA,$3E,$97,$95,$F4,$FE,$BE,$4F,$D1,$8C,$BF
	.dc.b	$4E,$32,$FC,$68,$FD,$2F,$CD,$7F,$4D,$8F,$C5,$B5,$FC,$FB,$5F,$F4
	.dc.b	$8B,$53,$49,$1F,$E6,$D7,$5F,$39,$FD,$C6,$7F,$C4,$41,$C3,$FE,$09
	.dc.b	$31,$BF,$92,$BF,$1D,$E9,$C9,$FC,$63,$F9,$1F,$94,$C4,$7E,$02,$2F
	.dc.b	$4F,$F8,$FF,$91,$5F,$80,$E9,$FC,$C7,$89,$30,$FB,$31,$97,$F3,$11
	.dc.b	$97,$F0,$1C,$69,$FF,$21,$7C,$03,$EE,$BF,$90,$37,$C8,$FF,$47,$B6
	.dc.b	$FF,$6A,$1C,$3F,$B8,$6B,$FC,$31,$AF,$FA,$F9,$AF,$19,$FE,$5D,$CF
	.dc.b	$F9,$2F,$BB,$FC,$D9,$27,$FC,$9D,$5D,$B5,$FF,$30,$79,$1E,$99,$FD
	.dc.b	$A7,$DC,$7E,$E4,$FE,$63,$9C,$8C,$90,$FD,$A4,$F1,$7F,$5B,$D5,$FC
	.dc.b	$07,$19,$1F,$BF,$8F,$E0,$86,$1F,$D9,$C6,$1F,$F3,$17,$95,$FD,$C3
	.dc.b	$E0,$FF,$98,$92,$3F,$98,$BF,$E3,$AE,$27,$D9,$7F,$0F,$24,$F2,$F1
	.dc.b	$FF,$DE,$4F,$F0,$FE,$7F,$E2,$84,$7F,$89,$4D,$7F,$82,$C5,$DB,$F0
	.dc.b	$D3,$FD,$1B,$F2,$97,$D0,$FF,$B1,$5B,$7F,$EC,$A6,$BF,$71,$AF,$F4
	.dc.b	$4E,$27,$E3,$2F,$AE,$FE,$86,$D9,$FE,$82,$D7,$F4,$15,$D3,$1F,$F1
	.dc.b	$DF,$ED,$5F,$D8,$67,$FA,$7E,$E8,$C4,$7F,$98,$9A,$FF,$8D,$9B,$E3
	.dc.b	$7F,$2A,$E4,$FE,$6C,$49,$FA,$33,$FD,$B6,$FF,$4D,$39,$9F,$E5,$2F
	.dc.b	$FA,$2A,$75,$E4,$7F,$39,$8D,$57,$0A,$E7,$D7,$FD,$B3,$E0,$7F,$B5
	.dc.b	$92,$3F,$B5,$BF,$6F,$E0,$3F,$A4,$12,$3E,$D4,$7F,$68,$FC,$0F,$F6
	.dc.b	$F1,$FA,$8E,$04,$62,$3F,$AF,$C6,$3B,$FF,$DF,$AD,$8E,$13,$B7,$EA
	.dc.b	$35,$1F,$DE,$5C,$36,$7F,$21,$E7,$FF,$5D,$24,$7F,$78,$9F,$16,$5F
	.dc.b	$E8,$7F,$67,$F8,$D7,$27,$EF,$AB,$FC,$B7,$E2,$7E,$11,$EF,$3A,$7F
	.dc.b	$81,$B5,$FE,$F2,$D7,$FC,$1A,$D7,$8C,$73,$7A,$CF,$E9,$7C,$AF,$EE
	.dc.b	$44,$8F,$F6,$75,$75,$BF,$E4,$53,$99,$FE,$CA,$FF,$B6,$E7,$5C,$1D
	.dc.b	$BE,$0B,$D5,$1F,$CD,$5B,$8F,$FB,$AE,$FD,$2D,$7D,$2D,$7D,$2D,$7A
	.dc.b	$1C,$23,$F7,$9F,$35,$FE,$09,$24,$7A,$27,$A6,$3F,$E3,$EF,$A9,$FF
	.dc.b	$1E,$E1,$F6,$3E,$2D,$F5,$3F,$E3,$29,$CE,$09,$CB,$F7,$A3,$31,$8E
	.dc.b	$87,$D5,$7F,$BD,$2B,$85,$F3,$F4,$3F,$D2,$37,$CA,$58,$72,$AD,$63
	.dc.b	$E3,$7F,$DA,$78,$73,$BE,$15,$E4,$BF,$F1,$AB,$5E,$27,$F4,$47,$0F
	.dc.b	$9F,$D2,$7C,$77,$E2,$E7,$FC,$40,$D8,$CC,$F1,$6B,$D9,$CD,$78,$9B
	.dc.b	$87,$FB,$A3,$C3,$FF,$5B,$F0,$FA,$6F,$C1,$85,$7D,$F7,$C1,$C9,$F1
	.dc.b	$8F,$F6,$1D,$72,$70,$70,$9E,$9D,$AF,$D7,$3F,$D7,$B8,$7E,$C3,$83
	.dc.b	$FD,$DD,$39,$DB,$FD,$E1,$1F,$AB,$FC,$11,$F1,$98,$BF,$DB,$DE,$DF
	.dc.b	$CE,$CE,$9F,$D5,$B7,$2F,$D5,$B5,$FD,$5B,$5F,$D9,$B5,$FF,$68,$B7
	.dc.b	$03,$4D,$57,$AF,$E3,$BC,$94,$8C,$5F,$93,$C4,$CC,$62,$7F,$CB,$99
	.dc.b	$AF,$2D,$BE,$07,$C5,$3D,$EE,$65,$9F,$C5,$C3,$FD,$43,$1E,$A6,$BE
	.dc.b	$A6,$E5,$EA,$6E,$5F,$D4,$F0,$F8,$F1,$82,$3F,$E1,$6C,$BE,$C7,$B9
	.dc.b	$E6,$3D,$FD,$CF,$AF,$7D,$9E,$6F,$FD,$37,$18,$7F,$9C,$37,$3F,$E6
	.dc.b	$ED,$7F,$B3,$35,$FE,$67,$5D,$6B,$D6,$DA,$E7,$18,$AE,$15,$F4,$75
	.dc.b	$FE,$07,$5F,$26,$31,$C7,$F5,$BA,$3F,$DF,$BD,$BF,$42,$31,$E5,$7F
	.dc.b	$7E,$BF,$F5,$FC,$FC,$AA,$DC,$60,$F8,$98,$3E,$8B,$EC,$9D,$7C,$77
	.dc.b	$8B,$8A,$F1,$70,$7E,$86,$B1,$A7,$DA,$9F,$EF,$15,$8E,$4F,$D1,$E3
	.dc.b	$D6,$78,$B6,$B8,$1F,$DF,$6B,$84,$F2,$3F,$C9,$B9,$7F,$81,$63,$D5
	.dc.b	$97,$92,$72,$DA,$FE,$D3,$E4,$BE,$73,$EE,$B6,$BF,$CE,$BF,$51,$FE
	.dc.b	$EA,$7F,$C2,$F3,$FA,$0A,$4E,$5F,$F2,$EB,$5F,$F6,$9B,$5F,$F7,$3B
	.dc.b	$5F,$B1,$F4,$7F,$A2,$3F,$47,$F3,$36,$FE,$0C,$59,$FD,$4C,$DD,$7E
	.dc.b	$3B,$C2,$BF,$C9,$7F,$50,$F8,$6C,$5F,$C8,$D9,$39,$AF,$8E,$D9,$FF
	.dc.b	$74,$CE,$AB,$F0,$53,$99,$FF,$7F,$F0,$FF,$1B,$F9,$5F,$DC,$FC,$87
	.dc.b	$EE,$9B,$8F,$D8,$6D,$7F,$00,$6E,$7F,$C8,$31,$75,$7F,$02,$F9,$7F
	.dc.b	$E0,$F1,$8C,$3F,$03,$18,$9D,$47,$E3,$27,$58,$7F,$0C,$AF,$DA,$FD
	.dc.b	$D9,$A7,$F1,$6F,$6D,$F2,$EF,$63,$38,$FF,$A1,$A3,$0F,$FA,$5A,$77
	.dc.b	$3F,$C3,$1B,$19,$9D,$E1,$FE,$99,$6B,$FE,$99,$6B,$FE,$99,$6B,$FE
	.dc.b	$A1,$6F,$7B,$17,$17,$F2,$4F,$8B,$F5,$B7,$0F,$DE,$56,$E3,$18,$7F
	.dc.b	$BC,$EF,$8D,$7E,$A2,$FF,$96,$F8,$12,$3F,$57,$F4,$36,$FF,$99,$CD
	.dc.b	$E0,$FD,$8C,$DA,$FF,$22,$D3,$EB,$5E,$A7,$E8,$5B,$3F,$ED,$AD,$1F
	.dc.b	$15,$9A,$FD,$34,$62,$7F,$CA,$B3,$FC,$DA,$D3,$97,$FC,$BD,$19,$8C
	.dc.b	$47,$F9,$7A,$30,$49,$FE,$27,$C1,$FD,$2B,$EE,$BF,$A5,$B6,$3F,$84
	.dc.b	$B6,$BF,$A8,$B7,$34,$FF,$30,$FC,$0D,$E7,$7F,$54,$FB,$0F,$E7,$F1
	.dc.b	$89,$FE,$27,$18,$9D,$47,$D4,$9D,$61,$FD,$7A,$BF,$27,$E3,$FF,$5E
	.dc.b	$DA,$DF,$F1,$C6,$D7,$D9,$9F,$E9,$BF,$02,$FA,$AC,$E7,$E6,$AF,$8F
	.dc.b	$C9,$FC,$F9,$E1,$8F,$E1,$2F,$83,$F8,$69,$D5,$7F,$2C,$F6,$3F,$B7
	.dc.b	$9F,$AE,$C6,$D7,$F2,$E8,$B8,$7F,$2E,$EE,$FE,$15,$58,$BF,$E6,$BE
	.dc.b	$E7,$F7,$83,$FB,$CF,$95,$71,$FE,$6F,$17,$0F,$EF,$F3,$6B,$D5,$0F
	.dc.b	$D4,$5F,$19,$FC,$AD,$B8,$7F,$59,$97,$E0,$9F,$E2,$87,$FA,$09,$C6
	.dc.b	$BF,$99,$46,$63,$F2,$E7,$0F,$4B,$F1,$D1,$98,$FC,$DB,$5F,$F5,$0B
	.dc.b	$5E,$15,$FE,$99,$8C,$1B,$D3,$FC,$75,$F1,$3F,$E1,$83,$FC,$3B,$E1
	.dc.b	$FF,$AA,$AB,$FA,$BC,$7F,$6F,$F1,$9F,$59,$FC,$A8,$93,$FC,$C1,$FD
	.dc.b	$17,$D6,$7E,$23,$6F,$F2,$1F,$D7,$3E,$77,$FA,$74,$E6,$FF,$C8,$33
	.dc.b	$97,$F9,$17,$95,$FC,$A0,$FE,$FD,$F5,$BF,$B4,$DF,$ED,$79,$49,$CF
	.dc.b	$87,$FD,$F6,$B5,$5F,$EE,$C6,$BF,$EC,$4F,$31,$18,$84,$62,$79,$13
	.dc.b	$0F,$EE,$35,$BE,$17,$C6,$7E,$F7,$6F,$F2,$FB,$6B,$B7,$D6,$FF,$69
	.dc.b	$9C,$F8,$BF,$E1,$78,$FE,$CD,$91,$A4,$F2,$7F,$C2,$CD,$BF,$F0,$0B
	.dc.b	$74,$FC,$1E,$5F,$E9,$EF,$3F,$FC,$A7,$F0,$27,$FC,$49,$F7,$CF,$A5
	.dc.b	$CC,$61,$FF,$11,$46,$71,$FF,$06,$B7,$0F,$EC,$ED,$7F,$93,$E6,$8C
	.dc.b	$47,$CB,$6B,$F2,$9B,$1B,$AF,$F2,$CE,$2F,$A6,$7A,$7F,$E6,$A6,$E3
	.dc.b	$FE,$3D,$B7,$D6,$FE,$0F,$2F,$DC,$56,$9F,$F0,$04,$65,$F2,$B5,$FB
	.dc.b	$CD,$7E,$EB,$67,$AF,$47,$F9,$0E,$30,$47,$CF,$F0,$0F,$BF,$FF,$3F
	.dc.b	$B7,$1F,$36,$4F,$9B,$CD,$7F,$94,$E3,$13,$FE,$38,$6E,$5F,$E1,$26
	.dc.b	$BF,$31,$AF,$E5,$5A,$FE,$51,$B9,$7C,$8B,$FF,$2D,$46,$B4,$FF,$65
	.dc.b	$F9,$7F,$E9,$4F,$84,$FE,$B9,$8F,$CB,$FC,$D3,$FC,$C5,$97,$F2,$19
	.dc.b	$CB,$FE,$29,$8C,$BF,$79,$AF,$E9,$DA,$FE,$99,$B9,$1A,$4F,$1F,$FC
	.dc.b	$DC,$F0,$C7,$FD,$F1,$F0,$3F,$EA,$8F,$87,$3F,$E7,$9F,$61,$F6,$19
	.dc.b	$8C,$3F,$E7,$78,$CE,$9F,$E6,$26,$E7,$FE,$3C,$6B,$FE,$3A,$FC,$AF
	.dc.b	$DF,$DB,$FE,$D2,$6D,$7F,$B3,$97,$8F,$FE,$BA,$73,$5F,$EB,$F6,$E6
	.dc.b	$69,$C2,$BF,$D0,$2E,$BD,$CF,$F5,$16,$1F,$EE,$06,$CF,$FB,$71,$7B
	.dc.b	$5F,$EC,$87,$38,$7F,$B3,$5B,$81,$A7,$0A,$FF,$47,$BC,$27,$FC,$33
	.dc.b	$A9,$FD,$67,$C4,$7F,$46,$E2,$FE,$D9,$D5,$FE,$EE,$9F,$E9,$35,$B8
	.dc.b	$C4,$71,$9D,$D7,$F4,$E8,$56,$E3,$A6,$16,$79,$9E,$03,$FC,$57,$5C
	.dc.b	$9C,$F4,$76,$51,$FD,$FE,$31,$1D,$8F,$F7,$45,$76,$69,$DB,$0E,$55
	.dc.b	$D6,$3E,$09,$E1,$3D,$57,$E5,$73,$3C,$23,$11,$F1,$F0,$B5,$CE,$30
	.dc.b	$7C,$12,$96,$30,$E0,$FF,$24,$FB,$BF,$E7,$3F,$AC,$E0,$D8,$F9,$4D
	.dc.b	$7F,$BC,$B5,$E3,$E3,$FF,$C0,$07,$6D,$76,$D7,$7B,$F6,$8E,$1A,$7D
	.dc.b	$AA,$FE,$57,$5F,$83,$8C,$74,$FF,$97,$9D,$57,$0D,$79,$D7,$CE,$3F
	.dc.b	$D1,$35,$B8,$F6,$B6,$3F,$1D,$7F,$DB,$F0,$ED,$5C,$B0,$EF,$58,$37
	.dc.b	$7C,$5F,$F4,$AE,$1E,$17,$FC,$CB,$FA,$47,$F2,$97,$F2,$1E,$F2,$31
	.dc.b	$1F,$E2,$08,$C7,$67,$D5,$6C,$70,$9D,$BF,$E7,$C8,$CB,$C6,$32,$FE
	.dc.b	$9E,$32,$EF,$F4,$8F,$F0,$A3,$E2,$6B,$E1,$6B,$BE,$FF,$14,$62,$3A
	.dc.b	$35,$FF,$40,$B6,$4E,$61,$F8,$5B,$E8,$7F,$1A,$D8,$FC,$93,$5F,$DA
	.dc.b	$35,$FF,$21,$B5,$FF,$4D,$B5,$FE,$08,$D7,$BC,$76,$FA,$E9,$1D,$9A
	.dc.b	$FF,$93,$B7,$7E,$09,$FC,$DA,$3F,$82,$46,$3B,$BF,$4C,$D8,$CC,$ED
	.dc.b	$FE,$15,$F6,$4F,$96,$FD,$4B,$63,$F5,$CD,$77,$1A,$23,$A6,$BF,$C5
	.dc.b	$7F,$19,$FE,$E8,$FE,$DD,$AE,$7C,$6F,$DB,$C6,$23,$FC,$D3,$5F,$E2
	.dc.b	$37,$C3,$E4,$79,$08,$E0,$87,$A9,$CB,$F9,$3F,$C1,$7F,$D5,$6F,$E6
	.dc.b	$1A,$E7,$E1,$7F,$2D,$18,$8F,$F3,$CF,$E6,$2F,$8C,$FE,$16,$D8,$FE
	.dc.b	$20,$D7,$F9,$93,$5E,$31,$C0,$8F,$07,$DF,$F4,$92,$3E,$23,$5F,$E7
	.dc.b	$9F,$92,$A3,$F8,$DB,$63,$F9,$03,$5D,$C6,$88,$E9,$AF,$C3,$F5,$3D
	.dc.b	$E4,$7E,$23,$FC,$A3,$F1,$EF,$ED,$F1,$FC,$B5,$AF,$F4,$F6,$BF,$E9
	.dc.b	$E6,$BF,$EF,$66,$BF,$D7,$DA,$9A,$63,$CA,$7D,$2F,$F3,$86,$CF,$F3
	.dc.b	$D6,$BF,$D7,$DA,$FF,$AB,$BF,$07,$7D,$67,$F4,$06,$C7,$F4,$56,$BF
	.dc.b	$D9,$DA,$F1,$8E,$04,$78,$3F,$BE,$CE,$47,$EF,$9A,$FF,$71,$FC,$37
	.dc.b	$DA,$4F,$3B,$FA,$8B,$F0,$2D,$F3,$5E,$52,$57,$95,$FE,$AF,$F5,$DE
	.dc.b	$48,$FF,$03,$07,$0F,$F8,$25,$AE,$71,$FE,$BF,$FA,$D3,$E5,$69,$27
	.dc.b	$FB,$34,$65,$B1,$97,$FC,$15,$19,$6C,$7F,$6A,$F1,$7F,$DC,$B1,$FD
	.dc.b	$BD,$AF,$F8,$45,$AF,$FB,$81,$AF,$AB,$F3,$77,$91,$FD,$D1,$B1,$FD
	.dc.b	$E5,$AF,$F8,$6D,$AF,$FB,$A9,$AF,$38,$E4,$47,$97,$B1,$D4,$8E,$08
	.dc.b	$75,$72,$D3,$FC,$15,$F8,$87,$F0,$F7,$FC,$58,$D7,$FD,$F6,$D7,$EA
	.dc.b	$B5,$FF,$1E,$35,$34,$9D,$5F,$E1,$4F,$D9,$7F,$71,$3F,$78,$38,$7F
	.dc.b	$02,$D7,$FC,$90,$D5,$F9,$43,$EC,$7F,$C3,$ED,$8F,$F1,$23,$5F,$F2
	.dc.b	$33,$5F,$03,$5F,$C7,$B5,$FF,$2B,$35,$EF,$1F,$E2,$CF,$B1,$7C,$75
	.dc.b	$24,$FF,$8C,$E3,$2D,$8C,$BF,$E5,$78,$CB,$63,$FC,$6F,$DE,$F9,$3A
	.dc.b	$49,$FF,$1F,$46,$5B,$19,$7F,$CC,$51,$96,$C7,$9F,$E4,$A4,$79,$5A
	.dc.b	$FF,$9A,$7E,$E7,$FB,$13,$F4,$DA,$FE,$C9,$AF,$F9,$E5,$A9,$A4,$F2
	.dc.b	$7F,$CA,$2E,$BF,$26,$F9,$44,$CF,$FC,$AA,$DA,$FF,$2D,$35,$FF,$3E
	.dc.b	$3F,$47,$F1,$77,$D7,$7F,$97,$1B,$3F,$E6,$16,$BF,$E8,$36,$BF,$8D
	.dc.b	$6B,$FC,$31,$AF,$FA,$4D,$A9,$BF,$0C,$F9,$3F,$F3,$5B,$67,$FC,$E0
	.dc.b	$D7,$FD,$24,$D7,$F3,$8F,$D3,$6E,$9C,$88,$F2,$6B,$C9,$AF,$E9,$7E
	.dc.b	$0D,$F3,$5F,$E7,$C6,$C7,$FA,$05,$AF,$FA,$8D,$AF,$18,$E0,$47,$83
	.dc.b	$FE,$86,$FC,$3B,$FD,$3B,$1F,$EB,$56,$A6,$93,$CE,$E0,$48,$FF,$49
	.dc.b	$47,$F8,$65,$FF,$1E,$FB,$3F,$E4,$62,$3F,$E4,$66,$BF,$E4,$66,$BF
	.dc.b	$E4,$67,$FD,$33,$F6,$2F,$A0,$A4,$9F,$F5,$04,$65,$FF,$39,$FC,$48
	.dc.b	$FF,$38,$15,$CB,$83,$AF,$81,$23,$9B,$5F,$F3,$D7,$E6,$EF,$B2,$FF
	.dc.b	$3D,$36,$3F,$D0,$2D,$77,$1A,$23,$A6,$BA,$30,$43,$07,$FB,$6F,$EB
	.dc.b	$7F,$C9,$CF,$F2,$56,$BF,$D9,$1A,$FF,$BA,$DB,$EE,$BF,$80,$1F,$87
	.dc.b	$FF,$A4,$5B,$5F,$E9,$86,$BF,$E9,$76,$BF,$E7,$63,$FC,$FA,$43,$FC
	.dc.b	$FA,$53,$FC,$FA,$53,$81,$FE,$A4,$FD,$07,$FC,$43,$46,$93,$CE,$FF
	.dc.b	$6A,$E0,$FB,$7A,$4B,$FF,$6E,$46,$5B,$19,$7E,$33,$5F,$F7,$DF,$E2
	.dc.b	$2F,$AF,$FF,$72,$7C,$5A,$31,$3F,$D3,$CF,$F7,$EB,$87,$FA,$94,$6E
	.dc.b	$78,$CE,$67,$AC,$60,$8F,$9B,$E9,$BD,$64,$70,$43,$EE,$39,$69,$9F
	.dc.b	$81,$7C,$6F,$FA,$FF,$F3,$CF,$E3,$7F,$1C,$FA,$6C,$78,$5F,$E2,$48
	.dc.b	$C6,$1F,$81,$C7,$FC,$4F,$8F,$28,$FC,$27,$DA,$7F,$52,$AD,$C3,$74
	.dc.b	$E7,$1F,$C6,$36,$5F,$2E,$F9,$A7,$95,$E1,$BD,$F9,$8B,$65,$DB,$2E
	.dc.b	$B3,$98,$C6,$5D,$72,$E9,$8F,$3C,$3C,$F7,$F8,$0A,$FE,$CB,$1D,$8F
	.dc.b	$D2,$38,$7F,$41,$E6,$38,$FF,$0E,$D8,$FD,$53,$5F,$D5,$35,$FD,$53
	.dc.b	$BC,$7D,$0D,$CF,$D9,$93,$ED,$1F,$7B,$F3,$6F,$78,$FB,$6F,$9B,$71
	.dc.b	$F2,$B7,$0E,$57,$C6,$B7,$5F,$21,$FF,$2F,$FB,$7F,$E7,$AE,$4F,$99
	.dc.b	$C3,$F6,$FA,$9C,$AF,$BB,$E5,$8F,$D1,$C3,$82,$BE,$15,$BC,$3C,$93
	.dc.b	$FE,$72,$F6,$FF,$0C,$FF,$8C,$6F,$8B,$63,$2F,$91,$AF,$76,$BC,$E7
	.dc.b	$F5,$8D,$3F,$5D,$3D,$1F,$C3,$62,$FC,$41,$EC,$74,$8C,$36,$33,$7F
	.dc.b	$B7,$9C,$B6,$33,$5F,$E4,$99,$CB,$63,$2F,$FA,$2A,$32,$FB,$FE,$9F
	.dc.b	$D3,$81,$A4,$BF,$5C,$65,$B1,$99,$F6,$BE,$9F,$B9,$7E,$13,$49,$5E
	.dc.b	$98,$CB,$F1,$BE,$AD,$F4,$9F,$29,$B1,$98,$C1,$1C,$13,$0F,$E3,$53
	.dc.b	$C1,$F8,$5E,$17,$FA,$3A,$7E,$43,$63,$3E,$F7,$CF,$8C,$47,$E0,$7E
	.dc.b	$17,$E1,$92,$BE,$0B,$F5,$7F,$32,$FD,$74,$7D,$36,$CF,$D1,$6B,$F3
	.dc.b	$DA,$FE,$01,$AB,$64,$61,$BD,$1D,$C9,$5D,$DB,$5D,$DF,$C8,$FD,$A3
	.dc.b	$1A,$4A,$FC,$9C,$65,$FC,$57,$E5,$2E,$DF,$C8,$36,$33,$18,$3A,$FD
	.dc.b	$94,$8E,$6D,$C7,$E8,$F5,$7D,$92,$4F,$DD,$6C,$6A,$30,$47,$FA,$96
	.dc.b	$1F,$CF,$63,$CF,$3E,$C2,$3E,$AF,$67,$DE,$47,$07,$F5,$BF,$C9,$1F
	.dc.b	$40,$FB,$9F,$DB,$C6,$1B,$19,$9E,$0D,$C7,$FD,$65,$5B,$9C,$D7,$02
	.dc.b	$3C,$3E,$3B,$99,$1C,$12,$B9,$DA,$75,$18,$7F,$85,$F7,$3E,$21,$F5
	.dc.b	$B8,$6E,$1F,$DD,$E7,$53,$93,$F8,$8F,$EB,$2F,$9B,$A4,$AF,$E2,$71
	.dc.b	$96,$C6,$71,$FF,$01,$4E,$5B,$1F,$1B,$E4,$E5,$7C,$66,$CF,$B1,$AF
	.dc.b	$F6,$11,$AF,$D2,$8C,$C7,$F6,$58,$FE,$CD,$F8,$6F,$D4,$83,$49,$5F
	.dc.b	$6E,$32,$D8,$CC,$7D,$63,$27,$FB,$A7,$E9,$9F,$A9,$7F,$88,$6B,$FC
	.dc.b	$3D,$18,$F4,$FF,$9C,$B6,$33,$3B,$AF,$F1,$27,$EB,$2F,$AA,$FE,$7A
	.dc.b	$DA,$FE,$7A,$D9,$DC,$E8,$93,$A6,$C7,$9F,$E9,$FE,$01,$27,$DC,$D8
	.dc.b	$D4,$60,$8F,$BF,$B3,$E4,$12,$3E,$33,$63,$51,$82,$3F,$E3,$CF,$D0
	.dc.b	$1F,$82,$7F,$5A,$A4,$62,$BF,$C8,$91,$89,$D4,$E6,$BF,$C9,$1D,$1F
	.dc.b	$E0,$2B,$FE,$BE,$D8,$CF,$9D,$FD,$92,$31,$1F,$85,$3F,$D9,$51,$92
	.dc.b	$4F,$27,$FB,$6A,$FC,$ED,$F7,$D8,$6C,$7F,$96,$4D,$46,$48,$E6,$37
	.dc.b	$5E,$18,$D9,$99,$E9,$F3,$92,$38,$36,$78,$52,$31,$1F,$9F,$9F,$34
	.dc.b	$62,$3F,$C4,$FF,$41,$FE,$B1,$9F,$F3,$77,$A6,$8C,$4F,$C3,$8C,$7C
	.dc.b	$2F,$F8,$31,$B5,$FE,$0A,$F5,$BF,$9C,$57,$F9,$DD,$AF,$CB,$6B,$FB
	.dc.b	$16,$BF,$E5,$36,$BF,$E3,$9F,$1B,$FC,$2F,$18,$33,$E5,$BD,$8F,$EE
	.dc.b	$44,$BF,$EE,$CF,$F8,$7B,$F1,$17,$D9,$D2,$4F,$F8,$82,$32,$FF,$88
	.dc.b	$BE,$F5,$E8,$FF,$89,$9B,$19,$8C,$19,$FA,$79,$1C,$5B,$A7,$FA,$5F
	.dc.b	$DB,$7C,$07,$F8,$C1,$B5,$FE,$30,$6C,$EE,$74,$49,$D3,$63,$F0,$9F
	.dc.b	$59,$F2,$24,$D2,$6F,$FE,$3E,$87,$C5,$52,$3F,$E4,$03,$25,$73,$5F
	.dc.b	$97,$6B,$FE,$33,$FB,$77,$D5,$FF,$90,$5B,$5F,$E3,$F6,$CF,$F8,$F5
	.dc.b	$AF,$EA,$FE,$EB,$FD,$E9,$3F,$E5,$06,$C7,$FA,$F5,$AE,$FE,$33,$FC
	.dc.b	$AF,$18,$8F,$E7,$0D,$7F,$D3,$4D,$4E,$65,$F9,$43,$EE,$BF,$6E,$D8
	.dc.b	$FD,$A3,$6B,$F6,$0D,$7F,$81,$72,$38,$3F,$CD,$2D,$AF,$F3,$4B,$67
	.dc.b	$FD,$AC,$D7,$8C,$F0,$24,$F0,$7E,$57,$E3,$12,$3E,$53,$67,$9F,$77
	.dc.b	$80,$93,$E0,$6C,$6A,$30,$47,$F8,$D7,$9E,$FC,$55,$F2,$1F,$DE,$9B
	.dc.b	$5F,$DD,$DB,$5C,$1A,$FF,$57,$F7,$7F,$D1,$3E,$59,$A4,$AF,$F4,$5C
	.dc.b	$65,$B1,$99,$FE,$BA,$F8,$7E,$5E,$4F,$85,$B3,$FE,$FF,$EA,$7D,$07
	.dc.b	$EC,$C9,$5F,$B7,$7F,$6D,$F7,$F2,$3F,$74,$DA,$E4,$D7,$F9,$E5,$23
	.dc.b	$11,$FE,$9E,$FD,$27,$F1,$9F,$00,$69,$39,$BF,$D4,$0D,$C3,$FD,$3C
	.dc.b	$DA,$FF,$4E,$7D,$AF,$ED,$27,$FD,$44,$D8,$CF,$67,$FA,$A6,$31,$1F
	.dc.b	$E9,$F6,$BF,$EB,$06,$BF,$E7,$16,$E0,$73,$1F,$E1,$BE,$EF,$F7,$E7
	.dc.b	$C1,$BC,$FF,$EB,$5E,$7B,$2C,$C7,$F1,$5B,$56,$DF,$F1,$25,$6D,$CC
	.dc.b	$EE,$75,$03,$FD,$16,$E1,$C5,$F6,$79,$DD,$7C,$C7,$43,$25,$33,$3C
	.dc.b	$EF,$C4,$F5,$85,$7F,$B6,$1E,$E5,$70,$E8,$FE,$F4,$3E,$DF,$F9,$1F
	.dc.b	$BB,$FE,$E0,$6C,$69,$AF,$F8,$05,$AF,$F9,$1D,$7D,$4F,$FB,$DE,$38
	.dc.b	$35,$E0,$D7,$83,$5F,$91,$1C,$BA,$3E,$22,$FB,$7F,$F2,$A6,$3F,$3A
	.dc.b	$31,$5A,$BF,$04,$E5,$BA,$79,$6F,$8B,$6B,$86,$11,$F9,$D8,$74,$6D
	.dc.b	$70,$7C,$B7,$D1,$FC,$87,$CE,$7C,$A6,$9F,$9C,$71,$BF,$E4,$DB,$B9
	.dc.b	$A4,$9E,$07,$D4,$97,$DB,$FF,$5A,$8C,$35,$FF,$19,$46,$B8,$7B,$CE
	.dc.b	$67,$FC,$F5,$39,$78,$CE,$88,$FA,$7E,$DF,$FA,$CB,$7C,$5F,$C5,$46
	.dc.b	$23,$F0,$ED,$7F,$C8,$4D,$7F,$D1,$0D,$7F,$44,$D4,$D3,$FD,$3C,$F8
	.dc.b	$7D,$EE,$E4,$7F,$D4,$DF,$6E,$FA,$9F,$9D,$C0,$8C,$4E,$A7,$31,$88
	.dc.b	$FC,$EF,$DB,$7F,$55,$7E,$B3,$5C,$F8,$9F,$A7,$8C,$47,$F9,$75,$AF
	.dc.b	$FA,$8D,$AF,$EF,$5A,$9B,$F6,$EF,$A4,$FD,$93,$67,$F0,$CD,$7F,$62
	.dc.b	$D7,$C1,$F5,$DE,$32,$3E,$86,$BE,$C6,$BC,$0F,$C4,$7D,$BB,$E3,$BF
	.dc.b	$22,$D8,$CC,$60,$FE,$0D,$EC,$FF,$57,$8F,$CD,$35,$FE,$0A,$D7,$FD
	.dc.b	$00,$D7,$FD,$8C,$D7,$F9,$03,$5F,$F3,$37,$8B,$D8,$F8,$7F,$6D,$23
	.dc.b	$E1,$B5,$FE,$27,$F6,$E8,$FE,$34,$D8,$FD,$73,$5D,$C6,$88,$E9,$AF
	.dc.b	$E9,$B0,$7C,$47,$C4,$24,$7C,$87,$F9,$47,$DB,$BD,$5A,$49,$FD,$F4
	.dc.b	$65,$B1,$97,$F9,$5C,$65,$B1,$FC,$CF,$DE,$7C,$6D,$24,$FF,$09,$8C
	.dc.b	$BF,$C2,$FE,$D9,$F3,$7F,$CE,$DB,$19,$8C,$11,$C1,$0F,$E7,$B1,$C1
	.dc.b	$FD,$CE,$BF,$BC,$23,$FC,$05,$FE,$91,$F6,$FE,$98,$FF,$5F,$35,$FE
	.dc.b	$B8,$1C,$3F,$D7,$DA,$A5,$F3,$5E,$92,$47,$B5,$FE,$AD,$F6,$FE,$58
	.dc.b	$FF,$66,$35,$EE,$D7,$FB,$63,$57,$DB,$FF,$9E,$9F,$E6,$4D,$7F,$AE
	.dc.b	$B5,$FF,$6C,$35,$F5,$B5,$FE,$F0,$D7,$AF,$BB,$FD,$AE,$31,$E2,$FF
	.dc.b	$89,$67,$FA,$03,$5F,$ED,$6D,$7F,$DC,$CD,$7E,$4F,$C9,$FF,$83,$1F
	.dc.b	$E9,$0D,$7F,$B9,$B5,$FF,$77,$35,$E3,$E5,$7F,$7D,$8C,$47,$F8,$3D
	.dc.b	$A9,$A3,$E7,$B8,$12,$3C,$2D,$7C,$2D,$7C,$2F,$F8,$2F,$ED,$DD,$54
	.dc.b	$93,$FD,$86,$32,$D8,$CB,$FE,$0E,$8C,$B6,$3F,$C2,$DF,$6F,$FB,$B8
	.dc.b	$FC,$68,$70,$FE,$41,$AF,$F8,$B5,$AB,$ED,$DF,$6D,$FE,$21,$6C,$7F
	.dc.b	$75,$6B,$FE,$20,$6B,$E0,$6B,$F9,$E6,$BF,$E3,$B6,$BD,$E3,$FC,$5D
	.dc.b	$EC,$7F,$BB,$E3,$FC,$04,$D7,$FC,$5A,$D7,$DA,$D7,$F5,$6D,$7F,$C9
	.dc.b	$0D,$7F,$CF,$BE,$FF,$63,$FC,$79,$DE,$FB,$3A,$49,$FF,$09,$46,$5B
	.dc.b	$19,$7F,$C7,$F1,$96,$C7,$AB,$E9,$E4,$7A,$9A,$FF,$91,$BE,$DF,$FB
	.dc.b	$63,$F5,$9A,$FF,$04,$6B,$FE,$61,$6A,$69,$3C,$8F,$F2,$B3,$AF,$B7
	.dc.b	$F9,$A4,$CF,$FC,$B4,$DA,$FF,$17,$35,$FF,$2C,$BF,$53,$ED,$DF,$5B
	.dc.b	$FE,$61,$6C,$FF,$8D,$5A,$FF,$98,$1A,$FE,$41,$AF,$F1,$E6,$BF,$E7
	.dc.b	$B6,$A6,$FD,$BB,$E5,$7F,$CE,$0D,$9F,$F2,$23,$5F,$F3,$7B,$5F,$CF
	.dc.b	$BF,$4D,$BA,$72,$23,$C9,$AF,$26,$BF,$A8,$FB,$77,$C1,$FF,$A0,$5B
	.dc.b	$1F,$E5,$56,$BF,$E8,$06,$BC,$63,$81,$1E,$0F,$FA,$2F,$ED,$FF,$D1
	.dc.b	$D1,$FE,$9D,$6A,$69,$3C,$DE,$04,$8F,$F3,$34,$7F,$87,$DF,$F2,$3F
	.dc.b	$B1,$FE,$4E,$23,$FE,$53,$6B,$FE,$50,$6B,$FE,$50,$7F,$D3,$DF,$6C
	.dc.b	$FB,$7A,$49,$FF,$3A,$46,$5F,$F3,$67,$CB,$8F,$F3,$A9,$5C,$B8,$3A
	.dc.b	$F8,$32,$39,$B5,$FF,$38,$FD,$BB,$E1,$BF,$CF,$ED,$8F,$F3,$D3,$5D
	.dc.b	$C6,$88,$E9,$AE,$8C,$10,$C1,$FE,$B8,$FB,$2F,$F2,$89,$3D,$6F,$F6
	.dc.b	$03,$FE,$56,$F7,$5F,$6E,$F8,$DF,$F4,$9B,$6F,$FD,$1C,$D7,$FD,$16
	.dc.b	$D7,$FC,$FA,$7F,$A0,$88,$7F,$A1,$CA,$7F,$A1,$0A,$70,$3F,$D3,$5F
	.dc.b	$6F,$FE,$06,$A3,$49,$E9,$7A,$3E,$DE,$57,$FA,$AF,$14,$62,$3E,$03
	.dc.b	$5F,$E8,$F1,$A7,$84,$7F,$AC,$FE,$25,$F6,$EF,$A5,$C3,$6B,$FD,$C3
	.dc.b	$3A,$9C,$92,$33,$3B,$7F,$AA,$CE,$DC,$CF,$D0,$F9,$89,$1E,$96,$BF
	.dc.b	$5D,$AF,$A1,$AF,$F8,$B3,$04,$7F,$8B,$BE,$05,$F1,$AA,$6A,$FB,$4F
	.dc.b	$E3,$7F,$1C,$FA,$CD,$4E,$7C,$3F,$F1,$BC,$E6,$F9,$67,$D9,$E9,$97
	.dc.b	$67,$3F,$6D,$FD,$6E,$B8,$42,$93,$9C,$7D,$0D,$97,$D0,$BE,$DD,$E5
	.dc.b	$F8,$AF,$7E,$6F,$AF,$2E,$B9,$74,$8C,$4E,$B1,$57,$A7,$1E,$98,$73
	.dc.b	$9F,$3D,$FE,$1F,$2F,$ED,$58,$F3,$9F,$D3,$46,$3C,$D7,$1F,$E3,$5B
	.dc.b	$3F,$AB,$6B,$F8,$C6,$BF,$55,$DE,$5E,$86,$E7,$F5,$73,$7D,$13,$E6
	.dc.b	$7E,$65,$EF,$1F,$6E,$3F,$4B,$1E,$2D,$8C,$E1,$CA,$F8,$E3,$E1,$9E
	.dc.b	$35,$FE,$6C,$F7,$FE,$C7,$07,$D9,$E2,$FC,$27,$DB,$F0,$AF,$C4,$7B
	.dc.b	$BF,$E8,$AE,$6F,$BE,$7E,$1B,$99,$5B,$AE,$93,$FE,$32,$78,$E3,$F8
	.dc.b	$57,$FC,$6B,$7C,$DD,$35,$CC,$6D,$FF,$43,$C6,$5B,$19,$7A,$CF,$7C
	.dc.b	$1F,$6F,$FE,$57,$8C,$36,$7F,$C8,$B3,$AF,$67,$93,$6B,$71,$8A,$E1
	.dc.b	$5F,$E9,$8A,$D3,$CA,$32,$F9,$BE,$DF,$C8,$FF,$A6,$1B,$89,$A4,$C3
	.dc.b	$E3,$4E,$DB,$3B,$6C,$F9,$BE,$89,$F6,$F4,$93,$F0,$63,$2F,$4F,$A6
	.dc.b	$7D,$BF,$91,$B1,$98,$C1,$1C,$13,$1F,$B5,$3C,$1F,$93,$E1,$3E,$03
	.dc.b	$E4,$B6,$BE,$93,$67,$73,$F0,$BE,$1A,$57,$B9,$B3,$F5,$3E,$DF,$F5
	.dc.b	$B3,$F6,$DB,$19,$E5,$FA,$F1,$88,$FA,$8D,$7F,$8C,$B5,$FE,$3E,$DE
	.dc.b	$03,$49,$3D,$09,$5F,$53,$ED,$7F,$DF,$4D,$CB,$EE,$C6,$30,$FC,$F3
	.dc.b	$5F,$C3,$7D,$BB,$AF,$F1,$AD,$BF,$CB,$36,$7F,$26,$D7,$FD,$33,$4F
	.dc.b	$B7,$EE,$4A,$F2,$3F,$6F,$AA,$47,$DF,$6D,$7D,$86,$BF,$EA,$4D,$FF
	.dc.b	$9F,$4E,$CF,$74,$FB,$08,$E0,$F8,$5E,$D7,$C0,$23,$83,$F7,$DF,$68
	.dc.b	$FA,$07,$DB,$FE,$D6,$30,$D8,$CC,$F0,$6E,$3F,$EB,$8A,$DC,$E6,$B8
	.dc.b	$11,$E1,$F1,$FC,$C8,$E0,$95,$CE,$D3,$A8,$C3,$FC,$27,$BF,$FD,$BD
	.dc.b	$F6,$F7,$E2,$8D,$26,$1F,$B6,$FB,$77,$CD,$FE,$F9,$B1,$98,$C1,$1C
	.dc.b	$13,$1F,$E4,$93,$C1,$F3,$7C,$A7,$C5,$24,$FC,$56,$C6,$A3,$04,$7E
	.dc.b	$40,$FB,$7F,$A0,$D8,$FA,$AD,$37,$1A,$24,$69,$B1,$FE,$EA,$FA,$87
	.dc.b	$DB,$FF,$88,$6B,$F9,$E4,$63,$D3,$FE,$64,$D8,$CC,$EE,$BF,$A1,$7D
	.dc.b	$BB,$ED,$FF,$92,$B6,$BF,$90,$B6,$77,$3A,$24,$E9,$B1,$F0,$3E,$A3
	.dc.b	$DE,$49,$F7,$B6,$35,$18,$23,$F2,$3B,$7E,$39,$23,$E3,$B6,$35,$18
	.dc.b	$23,$FE,$3A,$FB,$67,$DB,$FF,$53,$A4,$62,$BF,$C8,$AD,$AE,$15,$FD
	.dc.b	$1B,$A3,$FB,$F5,$7F,$42,$6C,$FF,$69,$6B,$FE,$39,$3F,$C7,$4E,$1C
	.dc.b	$35,$CF,$9D,$D4,$93,$FD,$CC,$7D,$B3,$ED,$FF,$B5,$46,$1B,$19,$8E
	.dc.b	$0D,$AF,$11,$FE,$4C,$70,$F9,$63,$91,$C2,$79,$9D,$3E,$76,$47,$46
	.dc.b	$CF,$48,$F2,$C6,$23,$A5,$23,$11,$FE,$27,$FA,$2F,$F5,$95,$7F,$83
	.dc.b	$FD,$34,$62,$7C,$F1,$8F,$86,$FF,$04,$36,$BF,$B8,$FA,$DF,$CE,$6B
	.dc.b	$E7,$35,$FF,$34,$35,$FF,$1A,$F8,$FF,$DF,$A3,$04,$70,$7F,$6F,$F6
	.dc.b	$65,$7F,$73,$6D,$7F,$89,$7E,$D9,$F6,$FF,$E0,$86,$D7,$F8,$11,$B3
	.dc.b	$B9,$FF,$06,$FD,$8B,$D1,$A4,$BF,$F0,$5C,$65,$BF,$0D,$C0,$8E,$0F
	.dc.b	$F0,$8F,$B7,$FE,$88,$6F,$C0,$FF,$86,$A3,$18,$FF,$8E,$DA,$FE,$09
	.dc.b	$AF,$FA,$1A,$38,$12,$7F,$21,$F5,$7F,$2D,$34,$9A,$FF,$8A,$21,$F1
	.dc.b	$54,$8F,$F8,$98,$C9,$5C,$DF,$E7,$5A,$FF,$8D,$FE,$DD,$F5,$9F,$E3
	.dc.b	$E6,$D7,$F9,$1D,$B3,$FE,$44,$6B,$F9,$EF,$B7,$FF,$7A,$4F,$F8,$D9
	.dc.b	$B1,$FE,$5D,$6B,$BF,$8C,$FF,$20,$C6,$23,$F7,$0D,$7F,$D6,$4D,$4E
	.dc.b	$65,$F6,$CF,$B7,$FD,$AB,$63,$F7,$4D,$AF,$D4,$35,$FD,$AF,$23,$83
	.dc.b	$FC,$A6,$DA,$FF,$27,$B6,$7F,$CE,$CD,$78,$CF,$02,$4F,$07,$E5,$7D
	.dc.b	$BC,$8F,$94,$D9,$E7,$DB,$E0,$24,$F8,$1B,$1A,$8C,$11,$FE,$1B,$E7
	.dc.b	$3E,$DD,$F2,$1F,$DD,$DB,$5F,$E8,$D9,$D4,$E4,$FF,$38,$F9,$2F,$76
	.dc.b	$92,$BF,$CD,$F1,$96,$C6,$70,$FE,$BA,$D7,$FD,$D5,$83,$E5,$FC,$44
	.dc.b	$AF,$14,$65,$FD,$8F,$D0,$7E,$D8,$8E,$0F,$DA,$FD,$BF,$EE,$88,$E0
	.dc.b	$97,$FE,$A9,$EB,$4E,$A3,$0F,$FA,$63,$ED,$FF,$19,$F0,$26,$93,$9B
	.dc.b	$44,$C3,$FD,$4D,$19,$7F,$D5,$1F,$6B,$FB,$46,$F6,$7F,$A9,$23,$18
	.dc.b	$FF,$AA,$1A,$FF,$A8,$DA,$86,$08,$CC,$61,$B1,$99,$FF,$53,$CF,$FB
	.dc.b	$EC,$FF,$46,$79,$BF,$D3,$BC,$D6,$1D,$7E,$0D,$3F,$C4,$EB,$67,$F6
	.dc.b	$5A,$DB,$83,$FB,$47,$83,$C6,$7A,$5D,$9F,$E3,$46,$CF,$F4,$7D,$F6
	.dc.b	$72,$D7,$DD,$7D,$A3,$11,$E9,$6C,$66,$BC,$D3,$FC,$B8,$FF,$5F,$8B
	.dc.b	$AD,$3F,$DD,$1F,$6F,$FC,$8B,$DB,$FF,$62,$B6,$33,$E0,$FF,$9B,$DB
	.dc.b	$3A,$FA,$97,$4D,$24,$F4,$8C,$B6,$3E,$C7,$DC,$7B,$70,$7D,$C7,$F8
	.dc.b	$3E,$30,$DC,$FF,$C2,$52,$C3,$9D,$71,$6C,$EA,$B8,$61,$F8,$5B,$EA
	.dc.b	$F8,$27,$9D,$70,$8F,$C3,$7C,$EF,$FB,$93,$87,$F1,$4D,$E2,$34,$D7
	.dc.b	$E9,$1F,$EE,$4A,$FC,$96,$9F,$82,$B4,$E6,$7E,$2F,$17,$E5,$E7,$3F
	.dc.b	$17,$FE,$2D,$BE,$18,$F6,$BE,$77,$BC,$39,$57,$56,$C6,$6F,$93,$D6
	.dc.b	$F9,$5F,$80,$7D,$D7,$E9,$FD,$0D,$C6,$03,$88,$D4,$78,$8F,$D6,$FD
	.dc.b	$D7,$F6,$A7,$11,$A7,$B3,$72,$FD,$1B,$70,$F8,$15,$C6,$BC,$99,$BA
	.dc.b	$7F,$97,$C7,$78,$C7,$83,$E6,$6C,$78,$C7,$DD,$7F,$6E,$78,$57,$27
	.dc.b	$AB,$72,$FA,$8D,$C7,$FC,$B9,$5C,$6B,$E3,$53,$EE,$9F,$59,$FE,$50
	.dc.b	$9F,$BB,$18,$7F,$A1,$35,$FE,$92,$D7,$FC,$D1,$8B,$F3,$37,$D3,$FA
	.dc.b	$DB,$3E,$F6,$BE,$2A,$E9,$87,$F0,$C8,$7D,$D3,$EA,$3F,$76,$D8,$F3
	.dc.b	$1A,$8C,$91,$CC,$6D,$F6,$35,$FF,$36,$B5,$FC,$B7,$E2,$0F,$9B,$FA
	.dc.b	$4D,$8F,$AC,$D7,$F1,$C7,$84,$FB,$5F,$72,$F9,$5F,$D6,$36,$3E,$66
	.dc.b	$A8,$C4,$7F,$9F,$A3,$13,$C0,$91,$C2,$76,$FC,$E6,$BF,$D0,$FD,$C7
	.dc.b	$E2,$7F,$ED,$58,$FC,$D4,$63,$E2,$7F,$42,$D8,$F3,$DF,$E9,$1F,$9F
	.dc.b	$F7,$7F,$38,$8F,$CA,$6B,$F2,$9A,$FC,$A7,$E9,$F4,$7E,$C3,$DB,$FD
	.dc.b	$62,$69,$DF,$64,$FE,$7F,$37,$C5,$FC,$82,$61,$F0,$5A,$FC,$78,$DB
	.dc.b	$92,$1F,$EA,$8C,$1F,$8A,$FF,$82,$E3,$F7,$2D,$7F,$72,$DE,$2F,$E3
	.dc.b	$55,$D2,$7E,$EF,$C1,$BE,$47,$FA,$5B,$63,$F2,$2D,$7F,$50,$D7,$FD
	.dc.b	$72,$D7,$FD,$98,$D7,$FB,$D3,$5F,$F5,$84,$7C,$AB,$7A,$FF,$C5,$5B
	.dc.b	$3E,$A8,$FB,$15,$E1,$C3,$F5,$1D,$5F,$E3,$3A,$FD,$2B,$5F,$DE,$35
	.dc.b	$FF,$66,$B5,$FE,$0B,$EA,$FF,$68,$8C,$47,$F8,$3D,$AF,$F1,$28,$FE
	.dc.b	$91,$F0,$CF,$87,$FE,$30,$D9,$FE,$38,$D7,$F4,$F5,$E3,$BF,$DF,$7D
	.dc.b	$D3,$E7,$7F,$24,$DA,$FE,$40,$D7,$F5,$FF,$76,$F8,$4F,$CC,$7E,$3C
	.dc.b	$F8,$EF,$EF,$ED,$AF,$E0,$AD,$7F,$8E,$35,$FF,$76,$F8,$BF,$D9,$9F
	.dc.b	$DD,$B7,$2F,$E2,$4D,$F1,$7F,$C6,$12,$FB,$C7,$D1,$FE,$A7,$CF,$FF
	.dc.b	$24,$CF,$F1,$46,$BF,$CA,$DA,$FF,$BF,$1A,$F4,$6B,$FE,$3D,$6B,$FE
	.dc.b	$01,$F6,$FF,$C3,$71,$AF,$09,$E1,$FF,$B5,$6B,$F8,$AF,$C4,$FF,$18
	.dc.b	$3C,$97,$FD,$77,$95,$EC,$D2,$57,$F4,$DF,$97,$7B,$78,$8F,$EB,$B5
	.dc.b	$C1,$FE,$C2,$D7,$F9,$0C,$F0,$73,$F0,$FD,$C8,$F5,$FA,$3A,$F4,$8B
	.dc.b	$C5,$FC,$FE,$95,$FC,$4F,$EE,$7F,$82,$DF,$F6,$16,$BF,$CC,$1B,$5F
	.dc.b	$D1,$FE,$ED,$F7,$7C,$7C,$69,$7C,$5F,$E3,$D3,$A8,$F5,$F6,$7C,$52
	.dc.b	$3B,$3D,$FE,$BF,$C9,$23,$FD,$CF,$EE,$DF,$75,$88,$F8,$F3,$72,$FE
	.dc.b	$B5,$5A,$9F,$93,$E8,$FD,$32,$3F,$DF,$7E,$19,$F1,$98,$8F,$A3,$36
	.dc.b	$FF,$B2,$56,$A7,$87,$D6,$7F,$97,$A0,$D2,$47,$F6,$68,$CB,$63,$2F
	.dc.b	$F7,$F8,$CB,$F9,$E1,$F8,$03,$F5,$57,$C7,$7F,$8A,$A3,$0D,$8C,$CF
	.dc.b	$90,$FC,$3F,$C1,$7E,$44,$8F,$F8,$6F,$EF,$7F,$BF,$F8,$5F,$E9,$BF
	.dc.b	$47,$F8,$55,$6A,$BF,$22,$FF,$55,$AD,$C7,$F8,$F3,$AB,$FC,$7F,$AF
	.dc.b	$F8,$C1,$B8,$73,$6C,$FF,$A9,$FE,$F7,$E9,$3F,$9C,$F8,$4F,$D2,$11
	.dc.b	$FF,$17,$46,$AF,$FC,$25,$F7,$9F,$C6,$67,$E2,$35,$FE,$06,$D7,$FD
	.dc.b	$68,$D4,$D2,$5F,$FA,$51,$E1,$3F,$E0,$3E,$0F,$F2,$AB,$6B,$FC,$A4
	.dc.b	$D7,$FC,$1F,$3F,$CA,$E3,$E0,$7C,$67,$EF,$49,$3F,$E4,$79,$7D,$E6
	.dc.b	$23,$F7,$73,$6B,$DB,$3F,$C1,$7E,$06,$57,$EF,$9F,$F0,$21,$FD,$97
	.dc.b	$4F,$98,$EA,$3F,$82,$BA,$F9,$1F,$F0,$57,$DE,$7F,$8A,$E7,$FD,$AC
	.dc.b	$D4,$D3,$1E,$97,$D6,$8F,$F2,$EE,$FF,$E1,$38,$C4,$FF,$8E,$72,$FF
	.dc.b	$47,$7E,$F9,$FE,$45,$6E,$5F,$E6,$78,$C3,$FE,$5E,$BC,$33,$0F,$F3
	.dc.b	$74,$E9,$FE,$B9,$3F,$E6,$FF,$88,$FF,$41,$C7,$F8,$D9,$AF,$F8,$D5
	.dc.b	$AF,$F9,$79,$AF,$E3,$FD,$6F,$E5,$C4,$7F,$CF,$91,$CB,$0F,$F5,$37
	.dc.b	$DD,$FF,$A0,$67,$FC,$86,$D7,$FC,$84,$D7,$FC,$D6,$D7,$F4,$DF,$9D
	.dc.b	$FF,$01,$7F,$D2,$2D,$9F,$F4,$8B,$72,$FF,$58,$57,$59,$CF,$6F,$FA
	.dc.b	$7E,$B5,$7F,$EA,$66,$F9,$CB,$E1,$87,$82,$F8,$12,$FF,$D3,$13,$A5
	.dc.b	$4F,$BC,$FF,$4A,$47,$FA,$11,$AF,$FB,$0D,$AF,$06,$F5,$7F,$82,$FE
	.dc.b	$27,$FA,$C1,$27,$FD,$4D,$3F,$E9,$B7,$BF,$DE,$7F,$80,$DF,$F3,$6B
	.dc.b	$5F,$F3,$63,$5F,$F4,$AB,$5F,$F6,$A3,$5E,$7E,$C7,$FB,$8E,$31,$81
	.dc.b	$D6,$1F,$E1,$FE,$4F,$ED,$84,$BF,$F5,$D4,$F8,$76,$6A,$9F,$EE,$3F
	.dc.b	$76,$FA,$8F,$F4,$CF,$A2,$8C,$4F,$FA,$C6,$31,$3A,$9C,$CF,$F3,$59
	.dc.b	$CB,$B8,$CB,$FE,$16,$8C,$BB,$8C,$BE,$3A,$62,$FA,$EF,$F7,$2C,$61
	.dc.b	$B1,$99,$FF,$72,$5F,$4C,$7D,$B1,$EE,$9D,$3F,$EF,$1B,$E4,$FF,$BA
	.dc.b	$DA,$FF,$BD,$6B,$94,$7F,$8B,$3E,$E7,$FD,$0D,$3F,$EE,$3F,$B8,$7D
	.dc.b	$DF,$F8,$C3,$E6,$BC,$5C,$09,$CD,$7C,$D9,$F0,$DF,$28,$FF,$5A,$B7
	.dc.b	$3F,$3D,$F1,$C3,$A6,$9F,$56,$31,$7C,$E3,$C9,$F1,$5F,$EF,$3C,$BB
	.dc.b	$CB,$2F,$1C,$61,$B8,$74,$9F,$C0,$65,$E1,$CF,$DD,$F7,$5F,$DE,$6F
	.dc.b	$CE,$DC,$B3,$CF,$E1,$C1,$F1,$AF,$BB,$FE,$EB,$87,$0A,$D5,$32,$63
	.dc.b	$CE,$3C,$2D,$C3,$97,$8B,$E5,$9E,$13,$FC,$2D,$F0,$9F,$F0,$06,$B9
	.dc.b	$DF,$FB,$B9,$B1,$9F,$8E,$FF,$21,$3F,$65,$AF,$C0,$6B,$E5,$6C,$7F
	.dc.b	$B9,$78,$BD,$2D,$9F,$99,$A7,$FB,$6F,$EA,$3F,$C1,$FB,$6B,$0F,$F2
	.dc.b	$96,$BF,$79,$B5,$E1,$C7,$E4,$6A,$C3,$8E,$3D,$0F,$24,$FC,$DE,$2F
	.dc.b	$C5,$7C,$97,$E6,$B2,$FB,$6D,$C7,$FC,$FF,$EA,$FF,$9F,$B4,$FC,$A4
	.dc.b	$E7,$1F,$DE,$4E,$57,$2F,$E3,$B9,$15,$B9,$FF,$4F,$7C,$17,$EA,$2B
	.dc.b	$53,$9B,$E5,$18,$C3,$EE,$C6,$1F,$CB,$61,$D1,$FC,$31,$FE,$D5,$FA
	.dc.b	$BE,$8F,$EF,$7E,$39,$18,$CB,$F9,$AF,$DD,$FF,$9B,$AB,$F6,$95,$A6
	.dc.b	$D6,$DE,$87,$F1,$9C,$7F,$53,$5C,$6F,$F7,$D7,$C5,$F8,$35,$FA,$B9
	.dc.b	$D4,$7F,$AA,$36,$46,$23,$5E,$F7,$F0,$A7,$51,$97,$CF,$19,$77,$5F
	.dc.b	$E5,$57,$2D,$AE,$B3,$B9,$FF,$23,$7B,$9F,$07,$C0,$FF,$6B,$3F,$0E
	.dc.b	$9F,$06,$F6,$F5,$E9,$2B,$53,$9C,$BB,$37,$1F,$F3,$36,$1C,$B0,$F0
	.dc.b	$36,$B7,$18,$AE,$15,$F0,$DA,$FF,$11,$6B,$FE,$5D,$FB,$BF,$19,$E7
	.dc.b	$F1,$7F,$D6,$B7,$FE,$B2,$8C,$7C,$1F,$82,$73,$5E,$46,$BF,$45,$AF
	.dc.b	$F8,$2B,$E3,$7E,$27,$96,$7E,$27,$DD,$7F,$B1,$B1,$FC,$01,$B2,$D7
	.dc.b	$F8,$1E,$78,$9C,$3A,$EC,$EB,$81,$87,$58,$C3,$6F,$F0,$E7,$F8,$23
	.dc.b	$0F,$03,$FE,$1C,$C3,$A5,$7D,$D9,$CC,$FE,$2D,$AF,$F3,$96,$BF,$E8
	.dc.b	$5F,$99,$FB,$67,$DF,$FB,$A7,$81,$8B,$F5,$F0,$A7,$31,$F8,$B6,$DF
	.dc.b	$DE,$6B,$F9,$B6,$BF,$D2,$7D,$2F,$E1,$FD,$1F,$E5,$6F,$77,$F6,$12
	.dc.b	$FB,$8F,$ED,$D8,$7F,$94,$A7,$F2,$57,$C2,$39,$47,$09,$DD,$7C,$6C
	.dc.b	$57,$B8,$FF,$06,$4E,$B0,$FD,$53,$70,$FF,$08,$4B,$EE,$B0,$F0,$F0
	.dc.b	$7C,$B1,$DA,$3F,$D8,$0D,$78,$B5,$FF,$0D,$37,$85,$D2,$F9,$9F,$D4
	.dc.b	$9F,$C3,$FE,$E5,$F4,$8F,$B9,$CE,$7F,$85,$BE,$0D,$BE,$15,$B8,$C1
	.dc.b	$BF,$17,$F6,$A6,$BC,$1B,$19,$26,$9F,$8C,$CB,$C3,$3A,$8F,$C7,$FD
	.dc.b	$53,$E9,$3F,$8D,$70,$FF,$86,$A3,$1E,$17,$F2,$78,$7C,$73,$EE,$7F
	.dc.b	$95,$56,$B3,$FF,$5A,$CE,$5F,$E2,$95,$C2,$BF,$89,$36,$33,$3B,$C7
	.dc.b	$FD,$DD,$39,$6C,$65,$FB,$31,$96,$C6,$5F,$F6,$F7,$CD,$FF,$18,$3E
	.dc.b	$4F,$CF,$5B,$F9,$D1,$67,$E2,$6E,$AC,$61,$FE,$E9,$FB,$9F,$E4,$27
	.dc.b	$D9,$AF,$AA,$7E,$DB,$93,$EF,$C6,$B1,$FB,$4E,$F8,$B1,$5F,$BC,$F8
	.dc.b	$B7,$EA,$CF,$1B,$FA,$94,$61,$F9,$7F,$73,$FE,$63,$C7,$F8,$6C,$E5
	.dc.b	$FE,$1F,$C2,$9D,$61,$F9,$C1,$19,$F7,$3F,$A0,$36,$BF,$A3,$36,$78
	.dc.b	$35,$FD,$2B,$5F,$F3,$B3,$5E,$B5,$E0,$7F,$57,$A7,$E8,$3E,$CF,$FA
	.dc.b	$39,$F1,$FE,$36,$E1,$F9,$D8,$B5,$F9,$31,$38,$DF,$FC,$D3,$7F,$AA
	.dc.b	$F9,$8F,$EA,$27,$D1,$F2,$6E,$3F,$A7,$8B,$87,$39,$B7,$CB,$EE,$7F
	.dc.b	$D1,$F7,$FC,$C5,$B5,$FE,$6F,$FB,$9F,$EE,$D5,$BA,$FF,$02,$5A,$73
	.dc.b	$1F,$E9,$78,$C7,$A5,$FC,$6B,$BF,$F8,$84,$62,$7F,$D0,$75,$C3,$04
	.dc.b	$E5,$FE,$7A,$7F,$06,$F1,$BB,$CE,$F4,$67,$F9,$3C,$3D,$1C,$FF,$E1
	.dc.b	$1F,$B9,$FE,$8D,$AF,$B6,$3F,$BD,$56,$F2,$FC,$A3,$67,$FD,$27,$6B
	.dc.b	$E2,$FF,$AC,$70,$EF,$87,$2F,$98,$93,$FE,$95,$A7,$D1,$E3,$2E,$B5
	.dc.b	$71,$EB,$5A,$BF,$9D,$3F,$E0,$BF,$AC,$FF,$04,$47,$F4,$C6,$DF,$F5
	.dc.b	$EF,$71,$F7,$2F,$8B,$E1,$EC,$7F,$8D,$67,$31,$FE,$19,$6E,$1F,$E1
	.dc.b	$F6,$DF,$E8,$5A,$FF,$3E,$6B,$FE,$CF,$6B,$FE,$8A,$BF,$04,$F3,$6C
	.dc.b	$7F,$8A,$FF,$0B,$7C,$FF,$F5,$86,$D7,$F8,$02,$7F,$AD,$7C,$16,$6B
	.dc.b	$FA,$EF,$DD,$3E,$53,$FC,$61,$48,$C5,$FF,$8D,$E3,$13,$A9,$CC,$7F
	.dc.b	$8E,$D7,$DD,$74,$3F,$B4,$FD,$AF,$F8,$DA,$FC,$53,$A7,$FB,$CF,$C8
	.dc.b	$7F,$98,$EF,$FB,$33,$78,$FF,$C9,$96,$FB,$7F,$F2,$04,$E6,$FF,$B2
	.dc.b	$CE,$5F,$33,$6F,$FC,$9C,$D7,$F8,$83,$5E,$35,$F0,$23,$11,$FE,$FF
	.dc.b	$6A,$73,$CF,$FC,$C6,$7F,$CA,$9C,$FD,$F1,$FC,$CF,$CC,$3D,$FC,$D7
	.dc.b	$F9,$CE,$31,$97,$F9,$84,$38,$CB,$F9,$24,$62,$BF,$CC,$4D,$38,$4E
	.dc.b	$E7,$FC,$0B,$19,$78,$C6,$50,$8D,$3F,$C8,$BE,$33,$F9,$9E,$1F,$C6
	.dc.b	$9B,$3F,$EE,$BF,$BA,$78,$9F,$E7,$19,$CD,$BD,$B5,$6A,$FF,$CF,$4D
	.dc.b	$9F,$F3,$6B,$5F,$F4,$53,$5E,$D7,$FE,$70,$8C,$47,$56,$D7,$69,$FF
	.dc.b	$42,$6B,$FE,$54,$CF,$FC,$C2,$7F,$A7,$FE,$E5,$F4,$B8,$CB,$FC,$BE
	.dc.b	$23,$2F,$D1,$6E,$5F,$41,$AF,$CC,$6B,$F2,$A3,$94,$FD,$49,$CC,$FC
	.dc.b	$D6,$CE,$8F,$E0,$1F,$84,$3E,$FB,$F8,$33,$63,$FC,$B3,$A3,$D3,$CD
	.dc.b	$7F,$A7,$63,$13,$F3,$27,$E3,$11,$C1,$30,$F9,$F1,$97,$FB,$76,$6F
	.dc.b	$8D,$7D,$1F,$FA,$AD,$B7,$FE,$84,$E1,$7D,$DB,$E1,$75,$7F,$E2,$B8
	.dc.b	$C5,$7F,$89,$9B,$8F,$F8,$89,$AF,$F8,$81,$AF,$FB,$45,$AF,$FB,$B9
	.dc.b	$B7,$E0,$BF,$F1,$7C,$6B,$89,$F9,$B7,$FA,$92,$BF,$35,$7F,$91,$8C
	.dc.b	$61,$D7,$E5,$FF,$D8,$91,$89,$ED,$1D,$09,$3E,$97,$FD,$7F,$FA,$53
	.dc.b	$E6,$3F,$D9,$8D,$9F,$E5,$1D,$7F,$CA,$AF,$FC,$F5,$39,$6D,$70,$3F
	.dc.b	$8F,$F8,$16,$FF,$D5,$5E,$A7,$DF,$AF,$F1,$93,$73,$F5,$E8,$FB,$D7
	.dc.b	$D7,$7F,$B7,$EB,$58,$7F,$61,$6E,$0E,$3F,$F1,$4F,$37,$FB,$8F,$0E
	.dc.b	$37,$E0,$8F,$F7,$37,$99,$FE,$0A,$CF,$BB,$67,$F8,$2E,$6F,$BA,$FE
	.dc.b	$27,$1F,$EA,$C9,$D6,$5F,$BF,$6D,$7F,$75,$9F,$F7,$95,$70,$AF,$F4
	.dc.b	$F5,$6A,$31,$1F,$E9,$AA,$ED,$DF,$FE,$EE,$9C,$CF,$89,$AF,$F9,$17
	.dc.b	$9F,$F8,$CF,$DE,$7F,$8C,$23,$FD,$47,$E9,$FE,$0A,$31,$1F,$EA,$69
	.dc.b	$DF,$AF,$FE,$4D,$AD,$3F,$E4,$F9,$FD,$10,$AE,$2F,$FB,$61,$AE,$9C
	.dc.b	$CF,$28,$F2,$1F,$D8,$CE,$67,$F6,$5A,$C3,$83,$FD,$73,$F0,$6F,$BD
	.dc.b	$FF,$55,$C7,$8E,$31,$F0,$3F,$DB,$5B,$19,$9D,$BF,$E6,$78,$7D,$83
	.dc.b	$83,$83,$67,$A3,$5E,$0D,$7C,$6E,$F2,$F9,$34,$FB,$DF,$F6,$7C,$61
	.dc.b	$BA,$70,$C3,$E6,$E1,$FD,$32,$FC,$71,$88,$FF,$5A,$57,$07,$4D,$7F
	.dc.b	$D7,$18,$78,$6F,$F2,$3F,$14,$EC,$FF,$07,$70,$FF,$B3,$DB,$19,$BF
	.dc.b	$B5,$39,$6F,$17,$FA,$8D,$FC,$C7,$95,$FE,$B0,$E4,$FC,$EC,$E6,$7F
	.dc.b	$DE,$57,$CD,$FC,$BD,$6D,$FC,$F3,$6F,$FC,$A4,$D9,$E1,$A7,$FA,$02
	.dc.b	$BF,$BF,$1F,$DB,$9C,$78,$FC,$FA,$3F,$B8,$B7,$8B,$FD,$8F,$9F,$F9
	.dc.b	$6E,$F8,$9F,$EB,$52,$DF,$27,$A4,$7F,$98,$67,$8C,$7F,$01,$F4,$9E
	.dc.b	$B7,$76,$C7,$03,$51,$9F,$90,$D3,$23,$59,$2B,$8C,$7F,$B1,$D5,$7A
	.dc.b	$FE,$FB,$FD,$85,$3F,$AB,$A4,$62,$3F,$A7,$B5,$FF,$57,$B5,$FF,$3D
	.dc.b	$63,$FE,$F3,$7C,$2D,$7F,$DA,$CD,$CB,$FD,$5B,$3D,$5F,$F5,$7F,$3F
	.dc.b	$F6,$7A,$F1,$35,$FE,$BE,$D7,$FD,$7E,$D7,$FD,$14,$D7,$F9,$53,$6F
	.dc.b	$FB,$37,$8F,$FE,$5F,$6C,$FB,$FD,$0F,$E6,$AF,$C8,$6B,$FE,$ED,$6D
	.dc.b	$F9,$AB,$F0,$5C,$3F,$E1,$4A,$F9,$0D,$7F,$BA,$B5,$FF,$6A,$B5,$FF
	.dc.b	$50,$35,$FE,$80,$DB,$FF,$19,$79,$9E,$78,$C4,$7F,$A0,$A7,$EE,$B5
	.dc.b	$FE,$11,$9F,$FA,$AE,$31,$87,$FB,$6A,$BC,$8F,$FB,$73,$E5,$5E,$8F
	.dc.b	$C3,$6E,$1F,$81,$6B,$FE,$0E,$6B,$FE,$EE,$6B,$FE,$B9,$6B,$FD,$65
	.dc.b	$B6,$6F,$E3,$AF,$B1,$FD,$53,$6B,$F2,$ED,$7D,$67,$3B,$FD,$87,$6F
	.dc.b	$F2,$F8,$FC,$B3,$5F,$F1,$1B,$5F,$F3,$29,$FD,$AD,$C3,$FE,$CF,$6B
	.dc.b	$9E,$EF,$C2,$46,$31,$FE,$95,$1F,$87,$FB,$97,$B7,$FC,$05,$B7,$FA
	.dc.b	$E6,$BE,$8A,$E2,$FE,$CB,$D4,$FE,$64,$FE,$B5,$AF,$F8,$E9,$AF,$F9
	.dc.b	$E7,$C0,$FC,$B4,$62,$3F,$DC,$CD,$7F,$C0,$CD,$93,$7F,$1D,$7C,$F7
	.dc.b	$F1,$76,$D7,$F0,$46,$BE,$79,$F1,$F0,$7F,$8C,$AD,$F8,$6F,$F0,$8A
	.dc.b	$F9,$71,$8E,$BF,$D5,$4E,$67,$FD,$F4,$D7,$FC,$34,$D9,$E1,$1F,$C3
	.dc.b	$63,$04,$F9,$1F,$E5,$89,$A6,$2B,$F9,$87,$33,$E6,$AF,$DC,$CC,$3F
	.dc.b	$8F,$CE,$5F,$E4,$3F,$87,$7E,$48,$9C,$9F,$C0,$5F,$CA,$FD,$1F,$F4
	.dc.b	$9D,$7E,$03,$7A,$BF,$D5,$BB,$23,$18,$7F,$8D,$5F,$E5,$75,$FE,$58
	.dc.b	$87,$C9,$FE,$7D,$FF,$76,$35,$FF,$0B,$35,$FC,$A9,$C3,$AB,$FC,$21
	.dc.b	$19,$8F,$CC,$B5,$FF,$5F,$B5,$E0,$7F,$37,$F8,$A7,$99,$FD,$55,$B1
	.dc.b	$FD,$09,$AF,$EB,$0E,$27,$EB,$BB,$5F,$AF,$3E,$0F,$FB,$33,$E9,$CB
	.dc.b	$FC,$B6,$DD,$7F,$C9,$AE,$EF,$FA,$A5,$F2,$BE,$39,$C9,$CD,$6D,$FD
	.dc.b	$37,$C4,$FF,$52,$7F,$AB,$B5,$CF,$8B,$FD,$BE,$1F,$4D,$62,$4F,$07
	.dc.b	$51,$FD,$FE,$1F,$96,$3E,$6B,$1F,$71,$D7,$6F,$B0,$DD,$BF,$83,$D6
	.dc.b	$EF,$FB,$71,$CA,$F8,$E3,$23,$33,$B7,$E6,$5F,$37,$FC,$F5,$93,$E3
	.dc.b	$FD,$47,$18,$F8,$3E,$37,$B4,$8F,$C2,$F6,$EB,$A3,$E2,$C7,$C3,$8F
	.dc.b	$89,$F7,$1F,$17,$E4,$65,$7C,$A8,$CB,$F3,$23,$2F,$D0,$8F,$8D,$F6
	.dc.b	$EF,$A6,$FF,$30,$36,$7F,$CD,$81,$C3,$FE,$6F,$6B,$F9,$A6,$BF,$CC
	.dc.b	$1A,$8E,$51,$D8,$FE,$E5,$93,$EB,$3E,$F1,$27,$E8,$D7,$83,$87,$F8
	.dc.b	$A5,$7F,$05,$9F,$C3,$F8,$5F,$85,$3F,$0D,$F0,$36,$32,$43,$F1,$EF
	.dc.b	$E2,$FE,$AA,$47,$E4,$1F,$F2,$27,$E2,$6F,$95,$CC,$FE,$5A,$6E,$3F
	.dc.b	$EB,$CA,$D4,$FE,$1B,$ED,$DF,$23,$FE,$89,$6C,$7C,$A6,$BF,$E9,$76
	.dc.b	$BF,$C0,$9A,$FF,$58,$6B,$F2,$5A,$9C,$F8,$4D,$1F,$8D,$FE,$9C,$95
	.dc.b	$E5,$BF,$D5,$7E,$13,$23,$F5,$EF,$EB,$7E,$23,$11,$FB,$48,$B9,$FD
	.dc.b	$09,$FD,$97,$BB,$2B,$F6,$EF,$ED,$7E,$8B,$11,$FB,$C8,$B5,$F6,$27
	.dc.b	$F7,$5F,$29,$F8,$02,$57,$F9,$3F,$34,$7E,$13,$EB,$3F,$80,$12,$7F
	.dc.b	$82,$C7,$F9,$C7,$F0,$8F,$E1,$58,$7F,$57,$6F,$83,$FC,$E1,$B5,$FB
	.dc.b	$2B,$E1,$E2,$C8,$FF,$32,$CE,$B8,$3E,$55,$75,$F1,$7F,$C8,$D3,$FC
	.dc.b	$C6,$4D,$27,$8F,$E3,$25,$F4,$AF,$37,$D4,$72,$24,$7F,$9B,$69,$1D
	.dc.b	$BE,$83,$D8,$49,$F2,$C5,$7F,$C1,$92,$9C,$69,$F0,$6A,$E9,$F0,$66
	.dc.b	$DB,$2F,$4D,$7C,$9E,$3F,$69,$27,$C8,$FC,$0E,$8F,$9C,$48,$F8,$F1
	.dc.b	$F4,$7D,$7F,$8A,$48,$F6,$35,$FE,$BB,$7F,$22,$31,$1F,$E0,$AC,$FF
	.dc.b	$C3,$07,$FB,$81,$FA,$7F,$11,$F7,$4E,$CF,$F4,$CF,$AE,$FF,$55,$BF
	.dc.b	$E3,$CF,$31,$18,$8F,$E1,$D1,$8F,$85,$FF,$67,$CE,$65,$39,$7F,$D9
	.dc.b	$F1,$98,$C4,$7F,$74,$6B,$E7,$F7,$7F,$20,$47,$C9,$55,$FA,$B1,$78
	.dc.b	$1C,$1E,$4C,$3F,$D8,$2D,$F9,$AF,$F2,$AB,$EB,$84,$62,$F4,$7F,$29
	.dc.b	$9D,$46,$63,$D7,$3C,$1D,$1C,$0F,$F0,$A6,$3F,$1A,$78,$36,$36,$72
	.dc.b	$8F,$F3,$3F,$C1,$FE,$52,$3F,$11,$A3,$E3,$B3,$C5,$FA,$D6,$CF,$13
	.dc.b	$F9,$F5,$71,$5F,$61,$FD,$E3,$A3,$37,$FE,$74,$8C,$4A,$B8,$46,$1F
	.dc.b	$17,$C3,$FF,$97,$3E,$C7,$BD,$74,$F4,$FF,$C2,$15,$C1,$B8,$75,$6C
	.dc.b	$FF,$B8,$7D,$F5,$F2,$C7,$C3,$3F,$E3,$26,$BC,$72,$FE,$C3,$18,$8F
	.dc.b	$EB,$4F,$FB,$A3,$D6,$CF,$A9,$FE,$2C,$9F,$63,$5D,$B5,$F8,$6F,$F8
	.dc.b	$D3,$3F,$7B,$73,$FE,$CD,$A7,$F9,$3F,$D2,$FC,$4D,$FC,$59,$EB,$87
	.dc.b	$CA,$8E,$AD,$7A,$B9,$AE,$5C,$79,$CB,$D1,$5C,$CF,$D1,$FE,$BD,$FE
	.dc.b	$B3,$9E,$9D,$5F,$C7,$BD,$EF,$0E,$BF,$83,$E5,$5E,$F0,$E5,$7B,$C3
	.dc.b	$C9,$5A,$F5,$3D,$31,$FB,$7B,$E0,$D8,$CD,$FE,$E2,$72,$D8,$CB,$E6
	.dc.b	$6D,$75,$AF,$CC,$B6,$7F,$CE,$57,$FC,$DB,$DE,$F9,$3D,$3F,$D1,$5B
	.dc.b	$7D,$F8,$2D,$FE,$7A,$6D,$F1,$AF,$E8,$B1,$AC,$3E,$0D,$EB,$9D,$39
	.dc.b	$CF,$FC,$7D,$7F,$E7,$CC,$7E,$6D,$FF,$A1,$5F,$04,$67,$0E,$6D,$AE
	.dc.b	$33,$FC,$45,$AF,$C9,$AE,$4F,$49,$FE,$95,$E6,$78,$CF,$27,$E1,$96
	.dc.b	$3F,$AE,$BC,$33,$FE,$B2,$DC,$CD,$24,$FF,$55,$6D,$FF,$3A,$9D,$CF
	.dc.b	$F5,$9F,$C2,$FE,$11,$F0,$FE,$B2,$C7,$F6,$C7,$B1,$1F,$EC,$0F,$87
	.dc.b	$B3,$E5,$1F,$2F,$A6,$CF,$F7,$48,$B8,$7F,$60,$9B,$C1,$FD,$FE,$6D
	.dc.b	$7F,$7F,$9B,$5F,$C6,$1F,$71,$23,$FB,$83,$F0,$7E,$8F,$ED,$1F,$6F
	.dc.b	$C0,$B3,$FE,$04,$7E,$31,$1F,$EF,$0F,$CB,$FD,$77,$F9,$7E,$3F,$EE
	.dc.b	$76,$CF,$FB,$21,$FF,$59,$BF,$E6,$18,$D6,$1F,$8E,$9D,$BF,$DD,$5B
	.dc.b	$B1,$E3,$09,$E3,$5C,$CE,$DD,$3C,$CE,$9F,$07,$63,$55,$D6,$3D,$04
	.dc.b	$8F,$CB,$61,$DE,$3D,$9F,$0F,$E5,$3C,$DE,$CD,$AE,$75,$E5,$9C,$5F
	.dc.b	$96,$7E,$05,$63,$1F,$89,$F3,$DE,$F3,$E0,$7C,$A5,$AE,$7A,$27,$18
	.dc.b	$73,$3E,$25,$62,$FE,$77,$DD,$FC,$93,$F7,$3F,$5E,$F9,$37,$C0,$E7
	.dc.b	$3F,$83,$39,$8F,$D4,$B6,$3F,$6C,$DC,$7F,$DE,$6D,$7F,$D3,$01,$C3
	.dc.b	$FE,$9B,$6B,$9A,$F0,$CF,$46,$C7,$F8,$8B,$37,$CC,$FE,$04,$FE,$19
	.dc.b	$FB,$7B,$E2,$F1,$87,$C4,$9B,$7F,$EF,$5A,$D4,$E4,$F9,$3F,$7F,$F8
	.dc.b	$B3,$F1,$9F,$7B,$63,$24,$AF,$99,$C7,$F8,$D9,$D4,$63,$2E,$AD,$CB
	.dc.b	$E6,$AF,$A3,$FC,$C4,$61,$F9,$71,$9C,$3E,$5B,$4F,$96,$53,$E5,$EB
	.dc.b	$FE,$C3,$9E,$24,$9E,$2F,$EB,$7E,$6F,$F4,$DF,$29,$FA,$76,$CF,$E2
	.dc.b	$1B,$7E,$96,$F1,$7E,$CF,$EB,$FF,$5C,$7E,$BF,$0B,$7F,$90,$8B,$7F
	.dc.b	$06,$6D,$7F,$70,$E9,$7D,$63,$E1,$FF,$71,$F6,$3F,$D2,$27,$31,$8C
	.dc.b	$BF,$5F,$8B,$DD,$7C,$46,$F1,$E5,$5A,$9C,$BF,$B6,$C3,$9B,$9A,$E1
	.dc.b	$87,$F5,$6C,$3A,$46,$27,$F5,$3F,$1F,$FC,$2C,$FE,$19,$F3,$56,$32
	.dc.b	$4A,$FB,$D1,$97,$F0,$1F,$65,$FC,$4F,$EA,$FF,$8A,$36,$BF,$6A,$D9
	.dc.b	$DC,$7E,$E2,$7F,$B3,$FC,$2B,$E5,$E9,$FD,$B2,$5F,$41,$FE,$EB,$7F
	.dc.b	$84,$CE,$DF,$E2,$A6,$BE,$0F,$FA,$EB,$4F,$ED,$45,$C3,$F6,$67,$F9
	.dc.b	$B2,$BA,$38,$8D,$3F,$E6,$C8,$CB,$67,$AC,$62,$07,$FA,$97,$F1,$AF
	.dc.b	$F2,$2D,$7F,$1E,$6C,$7E,$A1,$79,$3F,$E0,$67,$3C,$9F,$E3,$6E,$D7
	.dc.b	$CC,$7F,$A0,$30,$FE,$5A,$DC,$3F,$D4,$6D,$7F,$86,$CF,$09,$FE,$ED
	.dc.b	$F3,$2F,$1B,$FC,$26,$BC,$6B,$18,$7F,$C2,$FE,$53,$E1,$69,$FE,$1A
	.dc.b	$F9,$A7,$CB,$63,$1F,$E7,$31,$87,$FA,$0B,$73,$FF,$5B,$B5,$F2,$B5
	.dc.b	$F8,$F9,$A3,$11,$FE,$CD,$6C,$66,$BF,$C2,$11,$83,$E1,$6B,$F1,$DB
	.dc.b	$7D,$3E,$6F,$FA,$FC,$E6,$3F,$C6,$71,$FD,$8B,$0F,$8E,$D7,$87,$5F
	.dc.b	$F8,$76,$31,$1F,$ED,$F6,$E2,$69,$27,$FC,$49,$F0,$7F,$EF,$2B,$FE
	.dc.b	$BA,$DB,$FE,$4F,$F3,$8F,$27,$FC,$8B,$39,$CB,$F0,$13,$9B,$FE,$D6
	.dc.b	$DB,$FF,$70,$B5,$FB,$ED,$7E,$5B,$5F,$F7,$B3,$72,$34,$C7,$CE,$FE
	.dc.b	$82,$FF,$20,$DB,$FA,$BF,$3D,$FA,$DC,$3F,$9F,$CE,$56,$09,$D4,$7E
	.dc.b	$49,$AF,$2E,$AF,$F2,$44,$62,$38,$B6,$79,$4E,$89,$7F,$E0,$CF,$8E
	.dc.b	$7A,$5F,$E6,$C6,$C7,$F8,$2B,$DE,$7D,$86,$6B,$FC,$E3,$18,$9F,$F0
	.dc.b	$AB,$72,$FE,$03,$8F,$FB,$EA,$FF,$CB,$93,$99,$FD,$5B,$5F,$CB,$B5
	.dc.b	$F8,$2D,$9E,$55,$FE,$22,$E0,$7B,$5F,$E8,$26,$C7,$F8,$83,$E8,$3F
	.dc.b	$4D,$BF,$EF,$E7,$38,$FF,$6C,$9F,$D0,$93,$97,$FC,$E0,$DB,$FF,$17
	.dc.b	$B6,$7D,$6F,$EA,$3E,$0D,$E9,$7F,$A4,$9B,$5F,$E3,$03,$C2,$7E,$C9
	.dc.b	$CD,$7F,$7C,$8F,$D8,$92,$FF,$D0,$0F,$ED,$3E,$C9,$EE,$7F,$A7,$9B
	.dc.b	$1F,$E3,$C7,$F8,$DB,$FB,$E7,$35,$FE,$08,$6E,$7F,$C7,$DB,$3B,$F7
	.dc.b	$3A,$5F,$FA,$AB,$E9,$DF,$21,$4C,$7C,$85,$BF,$F2,$5C,$E9,$B3,$A6
	.dc.b	$9F,$E5,$AE,$2F,$E9,$7F,$2B,$FE,$B3,$F1,$CF,$FA,$D7,$E5,$9E,$AE
	.dc.b	$62,$5F,$1A,$DB,$CE,$B6,$E6,$79,$57,$FA,$BA,$D5,$AD,$FF,$A8,$C6
	.dc.b	$27,$53,$99,$FF,$60,$0E,$1F,$E9,$BD,$9F,$CE,$E3,$FC,$97,$CB,$FE
	.dc.b	$FC,$8C,$47,$F9,$3A,$77,$E2,$FF,$5E,$AD,$3F,$EC,$B8,$CC,$F2,$CE
	.dc.b	$BC,$45,$F4,$7F,$D1,$2D,$7C,$87,$98,$A7,$A2,$70,$E0,$FF,$30,$FD
	.dc.b	$2B,$83,$FC,$DD,$1F,$ED,$D8,$C7,$BD,$F0,$1B,$19,$9D,$BF,$DE,$23
	.dc.b	$F0,$7C,$BB,$E1,$FF,$4E,$4F,$36,$BB,$6B,$E2,$77,$8F,$C8,$AF,$EC
	.dc.b	$13,$F0,$9A,$FC,$2C,$F8,$5E,$6F,$8C,$7B,$3C,$5F,$F3,$EE,$1C,$DB
	.dc.b	$1C,$2F,$90,$E0,$FC,$1F,$B1,$F7,$9B,$1F,$EF,$BE,$04,$E5,$B1,$FE
	.dc.b	$FF,$E3,$7D,$2F,$FA,$33,$47,$0F,$F8,$4B,$1F,$F1,$3C,$E6,$31,$5C
	.dc.b	$2F,$F3,$18,$E5,$F2,$46,$5F,$F5,$45,$F2,$7F,$41,$7B,$8F,$3C,$7F
	.dc.b	$A0,$87,$C1,$F3,$EE,$FF,$11,$B6,$32,$7F,$B0,$38,$BF,$C6,$39,$FF
	.dc.b	$87,$E3,$27,$39,$E0,$73,$9D,$E7,$FE,$08,$F8,$EF,$F2,$35,$7F,$B2
	.dc.b	$9A,$FF,$A5,$1B,$7F,$E2,$B5,$F2,$1E,$57,$35,$CF,$C4,$C1,$99,$F5
	.dc.b	$7D,$2F,$F0,$2C,$3F,$DB,$AD,$7F,$A4,$B5,$FF,$18,$B5,$FF,$0E,$D7
	.dc.b	$F0,$C7,$F9,$87,$57,$C5,$8C,$47,$FA,$AA,$7A,$63,$FE,$41,$E8,$7A
	.dc.b	$DF,$C0,$B6,$FE,$04,$D8,$CD,$FF,$BC,$1A,$FF,$5C,$6B,$FE,$44,$6B
	.dc.b	$FC,$FE,$39,$1D,$7C,$27,$CD,$7D,$96,$CF,$F5,$78,$FB,$95,$DB,$0F
	.dc.b	$F2,$B7,$0B,$CE,$F3,$36,$FF,$B7,$C6,$08,$E0,$87,$F7,$28,$E0,$FF
	.dc.b	$74,$CB,$FD,$E2,$FF,$9B,$2B,$FC,$C9,$18,$8F,$F6,$8B,$63,$C7,$5F
	.dc.b	$91,$F2,$3F,$DC,$55,$FE,$09,$F8,$1F,$95,$18,$8F,$F0,$0B,$5F,$F3
	.dc.b	$4B,$5F,$F2,$AB,$5F,$F0,$FB,$6C,$DF,$29,$F1,$FF,$A1,$6D,$7A,$78
	.dc.b	$3F,$C5,$73,$E5,$C3,$F4,$DD,$BF,$C8,$EB,$FD,$26,$7F,$8B,$1C,$3F
	.dc.b	$E1,$66,$BF,$E7,$E6,$BF,$E6,$A6,$BC,$7C,$0F,$C4,$46,$32,$FE,$77
	.dc.b	$1F,$8B,$F3,$1F,$03,$FB,$76,$DF,$9C,$FF,$24,$C7,$9A,$7F,$79,$E7
	.dc.b	$7F,$24,$9F,$F4,$7B,$5F,$F3,$DB,$5F,$F2,$2B,$6C,$D2,$78,$3F,$9D
	.dc.b	$75,$E8,$3E,$23,$F8,$53,$6F,$C9,$7F,$E7,$19,$F1,$DF,$F0,$FF,$4B
	.dc.b	$F8,$0D,$7F,$A3,$BB,$3F,$51,$18,$8F,$F4,$7B,$5F,$F4,$74,$68,$8F
	.dc.b	$FA,$69,$E3,$EF,$75,$24,$7F,$A8,$23,$F8,$F7,$23,$E4,$2F,$C8,$4A
	.dc.b	$F3,$CE,$5F,$4F,$AF,$F8,$92,$71,$FE,$F5,$FC,$67,$B1,$FC,$C7,$A7
	.dc.b	$FD,$84,$DE,$7F,$F5,$3B,$6F,$FC,$E2,$FF,$1A,$7F,$DA,$3F,$4B,$F9
	.dc.b	$47,$FC,$12,$D7,$F1,$AD,$7F,$DC,$2D,$7F,$DB,$21,$C3,$D3,$AB,$FB
	.dc.b	$F4,$62,$35,$1F,$C5,$7E,$09,$E7,$7F,$41,$6C,$FD,$86,$AC,$7F,$AB
	.dc.b	$BD,$4F,$F1,$A7,$D2,$BE,$57,$FA,$73,$E9,$9D,$F7,$A3,$1A,$7F,$5F
	.dc.b	$78,$5F,$2B,$E3,$94,$8C,$1F,$8A,$F8,$0F,$E8,$6D,$F1,$FF,$AC,$46
	.dc.b	$23,$CD,$87,$FB,$5A,$1F,$2D,$AA,$95,$C4,$E9,$1F,$DA,$D7,$D2,$BE
	.dc.b	$17,$1F,$4B,$D7,$6F,$83,$48,$C6,$5F,$4A,$F8,$4F,$EC,$CE,$77,$CB
	.dc.b	$19,$19,$9D,$BE,$80,$75,$8F,$6F,$CA,$C8,$E9,$87,$81,$FF,$5D,$E7
	.dc.b	$FC,$61,$E5,$DC,$8C,$67,$F6,$23,$1E,$2F,$C8,$F9,$0F,$92,$CB,$F9
	.dc.b	$CE,$5F,$A1,$F4,$BF,$3B,$DE,$91,$82,$1E,$0C,$51,$A8,$C4,$FC,$DF
	.dc.b	$A4,$FF,$61,$C7,$F9,$99,$AF,$F2,$A6,$BF,$9C,$0E,$1F,$CF,$B5,$F8
	.dc.b	$ED,$77,$F0,$7F,$E5,$D8,$D6,$5F,$0F,$E9,$7F,$0E,$4A,$FA,$D3,$E2
	.dc.b	$D3,$E3,$64,$F0,$BF,$1A,$4A,$FC,$97,$C6,$57,$E0,$45,$AE,$33,$F9
	.dc.b	$4F,$A5,$95,$F9,$97,$FC,$65,$F3,$2F,$8C,$CC,$FC,$99,$B7,$FA,$8A
	.dc.b	$D4,$FE,$2F,$E9,$5F,$41,$FE,$79,$6C,$7F,$A4,$1A,$FF,$54,$6B,$F2
	.dc.b	$DA,$FF,$08,$6B,$F9,$06,$A7,$3B,$7E,$2F,$E9,$7F,$5E,$49,$F2,$57
	.dc.b	$EC,$BE,$96,$47,$ED,$DF,$DA,$FA,$D8,$8F,$C3,$45,$CB,$E7,$CF,$EE
	.dc.b	$BE,$06,$57,$EF,$DF,$DE,$FB,$78,$8F,$C7,$45,$AF,$AF,$3F,$C0,$BE
	.dc.b	$97,$FC,$61,$C4,$69,$2F,$FB,$BE,$9F,$A5,$FA,$5F,$E1,$24,$8F,$E1
	.dc.b	$D3,$FE,$63,$F9,$EF,$E2,$B1,$FA,$C6,$E3,$FA,$B6,$F8,$3F,$C3,$EF
	.dc.b	$87,$89,$2F,$FB,$0D,$69,$FD,$BC,$F5,$F1,$3F,$C8,$D3,$FC,$E3,$23
	.dc.b	$49,$E3,$F8,$C9,$7D,$2B,$CD,$F4,$5C,$89,$1E,$EA,$47,$6F,$A5,$F6
	.dc.b	$12,$7C,$B1,$5F,$74,$5D,$3D,$D3,$74,$FE,$2B,$97,$9A,$B1,$68,$F4
	.dc.b	$D7,$C9,$E3,$F6,$92,$7C,$8F,$C0,$E8,$F9,$C4,$8F,$8F,$1F,$47,$D4
	.dc.b	$F8,$A4,$8F,$63,$5F,$E0,$D7,$F2,$23,$11,$FD,$92,$30,$FC,$3F,$88
	.dc.b	$FB,$84,$7E,$AC,$FF,$A9,$7E,$97,$FD,$37,$1F,$DB,$3C,$C4,$62,$3F
	.dc.b	$D6,$CD,$7F,$D8,$8D,$8D,$FC,$2F,$FB,$16,$30,$47,$EF,$FB,$7F,$8D
	.dc.b	$25,$7F,$8E,$6B,$8B,$F8,$88,$CB,$F9,$17,$CD,$3E,$7F,$8D,$FF,$00
	.dc.b	$EF,$E7,$C1,$18,$BF,$3B,$5F,$3B,$70,$F6,$B6,$78,$CF,$F5,$2A,$E9
	.dc.b	$39,$24,$7F,$56,$F8,$1F,$C6,$BF,$81,$E2,$7D,$2F,$F8,$52,$73,$B7
	.dc.b	$E7,$23,$1C,$3E,$4B,$E4,$FF,$86,$BE,$97,$FB,$BF,$46,$27,$FB,$1C
	.dc.b	$61,$61,$E0,$51,$98,$F1,$FC,$3F,$F9,$63,$E4,$3B,$D7,$2F,$4F,$FC
	.dc.b	$1F,$5C,$6F,$86,$3E,$19,$CB,$FD,$CB,$0E,$57,$E3,$3B,$CF,$F0,$47
	.dc.b	$86,$3F,$E8,$27,$2D,$BF,$44,$7F,$5B,$7F,$DB,$7E,$AE,$7D,$2F,$F1
	.dc.b	$6C,$FC,$16,$BF,$01,$AF,$B9,$FF,$1A,$E5,$E9,$6E,$5E,$7F,$5D,$E9
	.dc.b	$FF,$8D,$AB,$58,$7A,$71,$F0,$E3,$F2,$B0,$F0,$4F,$CF,$D1,$8F,$2E
	.dc.b	$2E,$19,$F6,$C1,$C8,$FA,$5F,$F5,$4D,$7D,$7E,$2F,$E3,$3E,$EF,$87
	.dc.b	$4F,$F0,$DD,$EF,$D3,$C6,$3F,$87,$AB,$39,$C3,$F2,$13,$C3,$1E,$AF
	.dc.b	$F8,$C2,$74,$D8,$CC,$FE,$DD,$B1,$C6,$FF,$CA,$F3,$96,$C6,$63,$BB
	.dc.b	$F9,$9F,$77,$FC,$BB,$E0,$7F,$3D,$6C,$F8,$F8,$AD,$6E,$3F,$9E,$CE
	.dc.b	$34,$FE,$72,$EB,$2F,$8D,$79,$E7,$4E,$73,$F9,$D8,$FF,$87,$EF,$78
	.dc.b	$FF,$B3,$1B,$19,$9D,$BE,$FC,$3F,$CA,$6F,$96,$B8,$4E,$A3,$07,$F2
	.dc.b	$6C,$1E,$67,$8C,$F2,$7D,$2D,$BF,$EA,$EF,$4A,$FD,$E3,$73,$FE,$AA
	.dc.b	$DB,$34,$95,$FD,$45,$B5,$E1,$F9,$5F,$84,$7C,$3F,$A5,$B5,$FD,$99
	.dc.b	$E8,$47,$FA,$D3,$DF,$B3,$E5,$1F,$2F,$AE,$CF,$F6,$E8,$B7,$FD,$DE
	.dc.b	$6F,$07,$F7,$79,$B5,$FD,$6B,$25,$62,$FF,$8D,$3E,$D2,$4F,$F6,$B7
	.dc.b	$E0,$7C,$DF,$DB,$3E,$E7,$7D,$9F,$EF,$EF,$C5,$23,$FD,$CD,$F9,$5F
	.dc.b	$45,$F9,$EC,$BF,$CE,$8D,$9F,$ED,$AF,$FA,$D1,$FF,$74,$B9,$C3,$FB
	.dc.b	$73,$76,$34,$E1,$1F,$E8,$27,$1D,$5F,$95,$73,$7E,$0F,$82,$C4,$E5
	.dc.b	$F2,$93,$0F,$F1,$3D,$23,$D3,$F0,$BC,$CE,$9E,$CD,$AF,$3C,$5C,$7C
	.dc.b	$53,$71,$F7,$FD,$2F,$8C,$F2,$7C,$A5,$C3,$DF,$16,$FD,$53,$6B,$E3
	.dc.b	$FC,$E7,$AC,$FD,$9F,$D1,$BE,$4D,$F0,$39,$CB,$FD,$57,$7F,$ED,$8A
	.dc.b	$D4,$FE,$75,$B3,$FE,$8F,$C5,$18,$7F,$D2,$4D,$7F,$DD,$ED,$7F,$D6
	.dc.b	$ED,$78,$DF,$8A,$7A,$4E,$78,$3E,$BF,$D2,$FD,$33,$F8,$37,$D2,$BE
	.dc.b	$2B,$17,$F8,$09,$B8,$FF,$B1,$7B,$D7,$BA,$D1,$F8,$6F,$9F,$FB,$C7
	.dc.b	$DF,$F9,$EB,$19,$25,$FE,$16,$32,$FE,$1F,$1F,$C3,$1F,$31,$7D,$2E
	.dc.b	$2B,$E5,$9F,$02,$32,$FB,$9A,$7C,$82,$9F,$20,$A7,$18,$FA,$4E,$0F
	.dc.b	$A2,$5D,$3F,$DC,$71,$FA,$8F,$9B,$FC,$D7,$C9,$FE,$6D,$B7,$FA,$66
	.dc.b	$DF,$E8,$5B,$C3,$FA,$EF,$9D,$FD,$11,$FA,$3C,$2D,$FE,$BA,$2D,$FE
	.dc.b	$A6,$6D,$2B,$FD,$EB,$D4,$FA,$4F,$F6,$34,$7E,$DB,$E9,$7F,$A2,$CE
	.dc.b	$7E,$23,$F6,$11,$FD,$23,$17,$BA,$F8,$8D,$E3,$CA,$59,$77,$47,$09
	.dc.b	$DE,$1C,$EF,$96,$5F,$52,$99,$78,$63,$11,$C5,$D4,$66,$3F,$83,$7C
	.dc.b	$7F,$F0,$53,$F8,$37,$CD,$DA,$FE,$23,$16,$7F,$86,$4D,$CB,$F8,$CF
	.dc.b	$D2,$FF,$0D,$FA,$5F,$E1,$CD,$C3,$F8,$CB,$6F,$F8,$A3,$6B,$FD,$DF
	.dc.b	$F2,$DF,$DA,$7E,$18,$FF,$6E,$B7,$D0,$7F,$B9,$DF,$77,$6A,$B7,$1F
	.dc.b	$04,$DF,$C1,$FF,$59,$69,$FB,$D2,$E5,$F3,$0F,$F2,$F5,$F5,$70,$7D
	.dc.b	$F7,$0F,$F9,$B9,$AF,$09,$ED,$18,$8F,$F5,$17,$D2,$FF,$4C,$AF,$3B
	.dc.b	$63,$EE,$C7,$FA,$AA,$1E,$57,$F8,$25,$D6,$7F,$81,$CB,$FC,$67,$5F
	.dc.b	$DC,$7D,$9F,$C9,$3F,$ED,$A6,$E1,$FE,$A2,$6B,$FE,$A1,$F9,$87,$8D
	.dc.b	$FE,$15,$5E,$35,$8C,$3F,$E1,$9F,$29,$F0,$74,$FF,$0E,$7C,$CB,$E5
	.dc.b	$71,$9F,$D2,$8C,$3F,$EE,$C6,$E5,$FE,$B4,$6B,$FE,$B3,$6B,$FE,$CA
	.dc.b	$6B,$FE,$C9,$6E,$86,$92,$7C,$67,$F8,$BF,$E6,$BF,$AD,$BC,$CC,$74
	.dc.b	$FF,$85,$E3,$FB,$1D,$FF,$B0,$DA,$FF,$B6,$9A,$FF,$B6,$5B,$66,$92
	.dc.b	$7F,$C4,$5F,$37,$FE,$EC,$BF,$C9,$B6,$FE,$E9,$FE,$E3,$F9,$C7,$91
	.dc.b	$FE,$45,$9D,$71,$7D,$D9,$CD,$FF,$51,$C9,$18,$C3,$FD,$BA,$D7,$FD
	.dc.b	$B8,$D7,$FD,$E6,$D7,$C1,$87,$F8,$DE,$31,$3D,$23,$FC,$75,$F3,$BF
	.dc.b	$9F,$C3,$F1,$CD,$C3,$FD,$FB,$F3,$DF,$AC,$C7,$D7,$82,$B5,$1F,$EF
	.dc.b	$F9,$D3,$FE,$EF,$6B,$CB,$AB,$FC,$91,$18,$8E,$11,$82,$5F,$C1,$9D
	.dc.b	$5B,$E3,$9E,$97,$F9,$B5,$B5,$FC,$0B,$CE,$7D,$2E,$6B,$FC,$E5,$18
	.dc.b	$9F,$98,$DC,$BF,$7B,$87,$FA,$6A,$FF,$CB,$93,$99,$FE,$09,$39,$AE
	.dc.b	$35,$A2,$3F,$5F,$81,$ED,$FF,$A0,$5B,$1F,$C5,$BE,$83,$F4,$7B,$FE
	.dc.b	$EE,$73,$8F,$CC,$9F,$CF,$93,$97,$FC,$DC,$DB,$FC,$23,$67,$D2,$FE
	.dc.b	$9B,$E2,$1E,$8F,$FA,$45,$B5,$FC,$98,$F0,$9F,$D2,$B4,$FF,$74,$CF
	.dc.b	$FA,$5E,$3F,$5E,$4C,$3F,$CF,$CF,$EC,$BE,$95,$EE,$FF,$A7,$5B,$1F
	.dc.b	$CC,$9F,$E2,$EF,$57,$39,$7E,$21,$BA,$7F,$1B,$6C,$EF,$DC,$E9,$7F
	.dc.b	$EA,$9F,$A5,$7C,$85,$31,$F2,$16,$FF,$A0,$CE,$9B,$3A,$69,$FD,$8F
	.dc.b	$8F,$FA,$4F,$CA,$FF,$AC,$FC,$73,$FE,$B5,$F9,$67,$AB,$98,$97,$C6
	.dc.b	$B6,$F3,$AD,$B9,$9E,$55,$FE,$A7,$B5,$6B,$7F,$F5,$E3,$5F,$EA,$11
	.dc.b	$A9,$FF,$57,$47,$F5,$3E,$3F,$EF,$FD,$9F,$CE,$23,$FA,$B7,$37,$F8
	.dc.b	$0E,$31,$1B,$3F,$69,$E2,$FF,$58,$9D,$3F,$E5,$BA,$E6,$E7,$3B,$5F
	.dc.b	$D7,$A6,$BF,$E1,$C8,$AF,$F8,$42,$3F,$CC,$8F,$72,$B8,$74,$7F,$9A
	.dc.b	$7E,$95,$C1,$FE,$73,$8C,$35,$FE,$DB,$1A,$F7,$7A,$B6,$39,$70,$7F
	.dc.b	$88,$9B,$CB,$A8,$C4,$72,$6B,$FD,$EE,$7F,$13,$3E,$C7,$FA,$DC,$5F
	.dc.b	$17,$FC,$01,$AF,$CA,$9D,$CF,$CA,$3F,$C5,$EE,$23,$53,$C3,$4E,$CD
	.dc.b	$AD,$E1,$E7,$AF,$B1,$BB,$F5,$67,$C6,$FC,$A8,$C3,$63,$37,$FE,$FF
	.dc.b	$6B,$FE,$12,$97,$D9,$3E,$0B,$FC,$31,$AF,$F9,$16,$73,$18,$E1,$FF
	.dc.b	$23,$C6,$2F,$F4,$58,$74,$BE,$18,$73,$9F,$D3,$35,$FF,$47,$36,$3B
	.dc.b	$4F,$1B,$FE,$15,$BF,$F9,$3A,$FF,$C6,$6D,$7F,$A3,$B6,$BF,$C3,$4F
	.dc.b	$FB,$26,$37,$C5,$E7,$5D,$BF,$EB,$98,$C3,$63,$AF,$95,$8F,$90,$7C
	.dc.b	$9D,$24,$FF,$91,$21,$F1,$B8,$8F,$EB,$31,$C1,$D4,$F3,$9E,$FF,$4D
	.dc.b	$FE,$9E,$9F,$EE,$4D,$7F,$A3,$B5,$FE,$8A,$D7,$F9,$66,$8F,$9A,$F4
	.dc.b	$B6,$3F,$99,$47,$6F,$80,$EF,$ED,$7E,$B3,$46,$08,$C3,$6F,$C7,$18
	.dc.b	$23,$82,$1F,$D7,$3E,$0F,$FD,$1A,$FF,$99,$9A,$FF,$7A,$6F,$07,$F9
	.dc.b	$36,$BB,$4F,$F7,$EE,$37,$C4,$79,$5B,$1F,$DA,$A3,$04,$70,$43,$FB
	.dc.b	$6C,$70,$7F,$B7,$E2,$FA,$2F,$BA,$D8,$FF,$3D,$B5,$F1,$57,$E0,$AB
	.dc.b	$F8,$FC,$7F,$74,$6B,$FD,$6D,$AF,$E2,$9B,$26,$93,$AF,$D8,$7D,$18
	.dc.b	$7D,$07,$90,$95,$FE,$0D,$9F,$C9,$F5,$3E,$63,$EC,$36,$7F,$C7,$0D
	.dc.b	$7F,$C1,$AD,$7F,$C1,$6D,$7F,$B9,$35,$FD,$0B,$6F,$FC,$5D,$3F,$85
	.dc.b	$F4,$FF,$AA,$D7,$FA,$81,$AF,$F8,$D1,$BA,$FF,$8E,$A7,$CF,$F2,$3E
	.dc.b	$62,$57,$99,$AF,$99,$AF,$99,$F0,$7B,$FF,$B2,$B3,$4E,$B3,$FE,$81
	.dc.b	$F9,$6F,$CD,$7B,$26,$92,$3F,$D0,$8D,$7F,$D0,$8D,$7F,$D0,$9F,$0B
	.dc.b	$FC,$02,$7F,$DC,$91,$FE,$8D,$8C,$4F,$F0,$7E,$A7,$E5,$8F,$81,$FE
	.dc.b	$0F,$DA,$D2,$9B,$9D,$63,$FE,$97,$EC,$F3,$91,$FF,$4C,$3F,$B3,$F7
	.dc.b	$FD,$51,$FE,$CE,$6B,$FE,$62,$6F,$07,$FA,$D7,$57,$CF,$FF,$1B,$A7
	.dc.b	$D3,$BE,$53,$FB,$33,$63,$FB,$5B,$5F,$F6,$79,$C2,$36,$47,$6D,$7F
	.dc.b	$CB,$3F,$11,$FC,$AD,$FF,$3A,$B7,$0F,$F6,$43,$7C,$1F,$F6,$F5,$F7
	.dc.b	$9F,$F6,$67,$C3,$7A,$A7,$FC,$88,$DB,$FE,$66,$DE,$7F,$F7,$77,$87
	.dc.b	$FC,$D2,$30,$4E,$4C,$47,$F3,$BE,$C7,$D3,$FF,$0E,$C7,$FA,$9B,$FD
	.dc.b	$27,$1F,$DA,$36,$77,$A7,$18,$E1,$3A,$F2,$B0,$66,$7E,$77,$D3,$FF
	.dc.b	$98,$24,$7F,$AC,$07,$D3,$FD,$47,$3B,$32,$FF,$80,$13,$99,$E1,$5C
	.dc.b	$DB,$ED,$FD,$B9,$D5,$F1,$9E,$F1,$8B,$F0,$C2,$F9,$56,$E3,$C7,$C5
	.dc.b	$FD,$73,$E5,$7F,$D2,$F1,$E4,$BF,$F3,$A4,$62,$3D,$58,$75,$7F,$B2
	.dc.b	$FC,$1F,$3C,$3C,$5C,$48,$7D,$47,$F6,$86,$D7,$53,$8E,$39,$35,$32
	.dc.b	$35,$5F,$EB,$AB,$EB,$7D,$67,$2D,$7F,$D8,$6D,$8D,$F1,$78,$62,$4F
	.dc.b	$07,$C9,$1F,$A5,$FA,$77,$CA,$7F,$92,$5B,$5F,$E4,$E6,$BF,$A4,$6B
	.dc.b	$F3,$83,$87,$E8,$B5,$FC,$1B,$5D,$C7,$F6,$FE,$0F,$F3,$4E,$1F,$A6
	.dc.b	$F1,$7F,$C1,$F1,$8D,$39,$D7,$CB,$E5,$F9,$9D,$73,$1F,$87,$19,$7E
	.dc.b	$2C,$65,$F8,$B1,$FA,$3F,$1B,$F6,$51,$F6,$BE,$77,$EB,$19,$F0,$24
	.dc.b	$7C,$E8,$CB,$FE,$99,$9E,$2B,$E9,$FF,$D4,$AF,$F9,$D1,$AF,$F0,$A6
	.dc.b	$BF,$1D,$AF,$E5,$9A,$FE,$A1,$AF,$5F,$57,$FC,$F9,$1A,$E2,$FC,$77
	.dc.b	$CC,$FE,$34,$95,$F8,$D3,$81,$FE,$AF,$FD,$E3,$F0,$9B,$7F,$B3,$F9
	.dc.b	$3F,$BE,$71,$7E,$33,$CD,$FE,$0D,$5A,$BF,$EE,$CD,$C7,$F3,$93,$F9
	.dc.b	$B3,$77,$F9,$DF,$84,$FC,$F7,$D4,$CC,$3F,$2B,$19,$7F,$2F,$19,$7F
	.dc.b	$2F,$1F,$A5,$F9,$5F,$F1,$85,$1A,$49,$FE,$67,$87,$EB,$7C,$BF,$D6
	.dc.b	$12,$3F,$47,$3F,$AF,$E0,$91,$FA,$67,$4F,$E9,$9C,$B4,$FD,$5C,$FE
	.dc.b	$DF,$EA,$7C,$64,$8F,$0E,$08,$F3,$F7,$7E,$F4,$93,$FB,$48,$FE,$01
	.dc.b	$F5,$3E,$C2,$47,$27,$DF,$F0,$5F,$C1,$49,$1F,$BC,$8A,$FE,$F2,$2F
	.dc.b	$07,$F0,$29,$BC,$3F,$6E,$6D,$7F,$14,$9F,$E2,$1F,$07,$F3,$09,$1E
	.dc.b	$A7,$E7,$FB,$3F,$C5,$49,$1F,$C3,$23,$F8,$C7,$D4,$FD,$42,$47,$C2
	.dc.b	$6B,$FC,$07,$83,$F9,$04,$62,$3F,$98,$C6,$1F,$B3,$FB,$13,$E9,$FF
	.dc.b	$D1,$6D,$8F,$E3,$ED,$7F,$8D,$D7,$F5,$9F,$A4,$72,$FF,$AD,$DB,$3F
	.dc.b	$DD,$BC,$B4,$62,$3F,$D6,$CD,$7F,$DF,$AD,$7F,$D6,$6D,$7F,$CA,$8D
	.dc.b	$8E,$73,$FE,$83,$B7,$C8,$FA,$09,$7F,$CE,$EB,$6F,$AA,$32,$B8,$3F
	.dc.b	$A2,$D7,$28,$E1,$87,$9E,$7F,$D9,$0B,$9F,$FA,$CF,$C3,$FF,$B4,$B1
	.dc.b	$F6,$C6,$3C,$CF,$4B,$70,$F4,$F5,$5A,$DC,$7F,$72,$BE,$58,$75,$8C
	.dc.b	$61,$C9,$A7,$22,$D7,$25,$5C,$B0,$ED,$7C,$E7,$37,$F3,$30,$F2,$43
	.dc.b	$E9,$FF,$D8,$BB,$F6,$8C,$78,$5F,$81,$CB,$FD,$DB,$5A,$9E,$63,$D7
	.dc.b	$7C,$AB,$E3,$BA,$D6,$EF,$8E,$1C,$F3,$FF,$00,$CE,$5F,$F0,$9E,$1C
	.dc.b	$AF,$8D,$7E,$3F,$1D,$B5,$FF,$0C,$DF,$8D,$FE,$BC,$D7,$AF,$97,$FE
	.dc.b	$ED,$6F,$C1,$FF,$92,$1F,$81,$19,$79,$B5,$CC,$7C,$39,$FF,$3B,$3F
	.dc.b	$4B,$07,$AF,$FE,$E8,$EC,$D6,$1F,$EF,$DD,$3E,$3B,$74,$F0,$E5,$F2
	.dc.b	$30,$E3,$E4,$FF,$89,$6F,$79,$7D,$CF,$DF,$4E,$70,$FF,$1B,$F2,$27
	.dc.b	$39,$EF,$2D,$1F,$EB,$98,$CB,$F5,$A7,$85,$E9,$B5,$C4,$F0,$46,$6B
	.dc.b	$8E,$1F,$9D,$6E,$1E,$27,$FC,$93,$5C,$0F,$CA,$F5,$39,$DE,$67,$F9
	.dc.b	$07,$A3,$F5,$AD,$AF,$5F,$C3,$3E,$6F,$1B,$31,$E5,$78,$6D,$FF,$12
	.dc.b	$8D,$EC,$C3,$94,$E4,$FE,$00,$E2,$FF,$C4,$53,$BC,$3F,$C9,$D1,$E2
	.dc.b	$6B,$FE,$E1,$9E,$0F,$F9,$5A,$78,$57,$F1,$B9,$CC,$F1,$8C,$57,$2C
	.dc.b	$3E,$16,$9E,$9C,$CD,$FA,$BC,$36,$FF,$98,$C7,$19,$34,$91,$A2,$61
	.dc.b	$FC,$F6,$77,$1F,$CE,$BE,$AF,$C0,$78,$7D,$FB,$1F,$CE,$9E,$C4,$7F
	.dc.b	$A0,$BE,$1F,$8C,$F4,$1E,$9F,$3E,$CF,$F4,$28,$B8,$7F,$29,$8F,$F4
	.dc.b	$5E,$7F,$81,$AC,$F0,$7F,$2C,$BB,$5A,$C9,$58,$C3,$F0,$11,$89,$FE
	.dc.b	$A7,$3F,$0C,$8F,$F5,$6F,$8E,$F9,$87,$CD,$F8,$4B,$1F,$D5,$9F,$94
	.dc.b	$47,$FA,$EB,$F3,$7E,$AB,$EA,$1F,$57,$EA,$AC,$FF,$5E,$8B,$8F,$B4
	.dc.b	$F2,$CE,$33,$FE,$8E,$F1,$AF,$EC,$BF,$09,$6B,$86,$BF,$A3,$8F,$B8
	.dc.b	$49,$FE,$DE,$FD,$FF,$AA,$FC,$37,$CC,$7E,$1D,$B3,$FD,$C1,$B7,$D3
	.dc.b	$64,$63,$2F,$F5,$77,$33,$D4,$79,$7F,$91,$F8,$4F,$D3,$C6,$32,$FE
	.dc.b	$F7,$18,$9D,$47,$F9,$4D,$B9,$77,$F9,$6F,$CD,$1F,$B6,$F8,$D7,$C8
	.dc.b	$3C,$AC,$E1,$FE,$A9,$9C,$C7,$EB,$1B,$1F,$EF,$86,$DF,$FA,$55,$AF
	.dc.b	$FB,$11,$AF,$FB,$59,$A9,$A6,$33,$7A,$EF,$A4,$FD,$87,$95,$FC,$1A
	.dc.b	$31,$87,$F8,$6A,$31,$3A,$8F,$F4,$43,$70,$F9,$DD,$FF,$B9,$FA,$9F
	.dc.b	$DD,$36,$BF,$C4,$6D,$9D,$C7,$F8,$9A,$7F,$8C,$78,$4F,$8C,$A7,$C9
	.dc.b	$AF,$19,$F8,$68,$D5,$7C,$98,$C9,$F0,$CA,$7C,$32,$9F,$0C,$BA,$7F
	.dc.b	$BF,$31,$FA,$11,$82,$3F,$C5,$7F,$51,$FE,$73,$1F,$96,$6F,$27,$FB
	.dc.b	$99,$FE,$92,$D8,$FE,$65,$1A,$BF,$DF,$B6,$3F,$9C,$3C,$23,$31,$C3
	.dc.b	$E0,$38,$FC,$85,$9F,$F7,$4D,$23,$15,$E3,$6D,$F8,$DF,$E8,$BE,$4B
	.dc.b	$E9,$BE,$FE,$1C,$BE,$7B,$31,$8C,$3D,$91,$FD,$31,$FE,$A7,$EC,$EB
	.dc.b	$8F,$1B,$A3,$18,$7F,$AD,$63,$0A,$BF,$81,$63,$E1,$BE,$37,$C2,$7E
	.dc.b	$57,$CB,$FC,$6F,$86,$F8,$ED,$C3,$DE,$D9,$DC,$7C,$29,$FE,$BD,$F1
	.dc.b	$4F,$A6,$A6,$3E,$5A,$D7,$8F,$80,$FF,$30,$AE,$0E,$4C,$57,$17,$18
	.dc.b	$7F,$6D,$9D,$BF,$C5,$BE,$41,$F1,$9F,$DB,$DB,$5C,$07,$CD,$BC,$9D
	.dc.b	$57,$F7,$89,$CC,$7F,$6A,$69,$F2,$6E,$38,$70,$7A,$AB,$A4,$62,$31
	.dc.b	$3C,$E7,$31,$C2,$79,$3D,$23,$2F,$FA,$73,$E9,$5E,$0F,$F8,$25,$B5
	.dc.b	$F0,$BC,$97,$4F,$F7,$48,$C4,$EA,$FF,$9C,$FC,$A3,$E3,$7F,$BE,$36
	.dc.b	$BF,$DB,$ED,$9F,$F5,$73,$5F,$81,$EC,$BF,$33,$FF,$6A,$57,$87,$E9
	.dc.b	$9C,$DF,$E2,$0A,$D7,$C4,$FF,$87,$E7,$35,$E7,$D1,$EA,$FF,$88,$DB
	.dc.b	$1E,$BC,$FF,$99,$3B,$9F,$F1,$44,$63,$3F,$2F,$91,$EA,$25,$7A,$9B
	.dc.b	$1A,$8C,$1F,$EF,$CE,$3F,$E9,$11,$FD,$71,$BC,$FF,$92,$8F,$F1,$C7
	.dc.b	$CC,$7A,$1D,$79,$7F,$E3,$98,$C4,$7F,$B4,$1A,$FF,$82,$1A,$F2,$DF
	.dc.b	$FC,$55,$18,$E3,$34,$91,$FE,$2F,$F8,$2F,$F2,$7E,$5F,$75,$B7,$FD
	.dc.b	$CA,$9F,$57,$FE,$4B,$9C,$E5,$E9,$C7,$AC,$F8,$09,$87,$E4,$E3,$2D
	.dc.b	$BE,$6D,$9D,$E1,$FE,$EF,$9C,$BF,$EF,$1F,$C0,$3F,$CA,$58,$77,$6D
	.dc.b	$FF,$8A,$3C,$DF,$D9,$CE,$7C,$CF,$F3,$04,$7E,$DA,$FF,$DE,$ED,$78
	.dc.b	$77,$7F,$93,$E3,$11,$FE,$69,$6E,$66,$92,$7F,$CA,$DF,$11,$FC,$C7
	.dc.b	$1F,$D2,$36,$FF,$C3,$34,$FA,$9F,$F3,$94,$E7,$2F,$91,$39,$BF,$8C
	.dc.b	$DB,$FD,$EE,$3E,$1C,$3C,$C4,$AD,$B6,$77,$5C,$5F,$E1,$15,$FE,$DC
	.dc.b	$F1,$FF,$D0,$5F,$02,$7F,$D0,$98,$7F,$AC,$6B,$FC,$5F,$39,$8F,$DF
	.dc.b	$19,$8E,$6D,$7F,$CA,$ED,$7F,$D2,$0D,$AF,$E5,$BF,$05,$FE,$7B,$8C
	.dc.b	$11,$FF,$3F,$53,$E9,$FC,$A7,$93,$9A,$D7,$89,$E4,$47,$F1,$87,$ED
	.dc.b	$70,$F7,$FD,$23,$D9,$FF,$50,$36,$7F,$8C,$2F,$5B,$33,$27,$2F,$08
	.dc.b	$C5,$7F,$94,$BE,$93,$F8,$3B,$7B,$3F,$D5,$90,$F8,$0B,$13,$0F,$E4
	.dc.b	$D3,$C3,$8B,$F4,$CE,$E3,$33,$FE,$B2,$FA,$27,$D3,$53,$FD,$6B,$F0
	.dc.b	$CF,$84,$C5,$FF,$52,$E5,$5F,$08,$C3,$74,$FF,$38,$4F,$FA,$F6,$B8
	.dc.b	$C6,$2F,$A5,$FF,$AC,$E7,$FD,$73,$5B,$8C,$1F,$EF,$FA,$FE,$F5,$C5
	.dc.b	$FD,$EF,$E8,$5E,$87,$F5,$2F,$4F,$F8,$D4,$62,$BF,$AA,$D6,$EB,$57
	.dc.b	$67,$22,$B8,$3F,$E0,$7B,$E7,$5F,$D6,$EF,$84,$E6,$EB,$E4,$3C,$D1
	.dc.b	$E7,$4D,$3F,$B2,$2F,$A4,$7C,$2F,$F9,$CA,$3F,$B4,$46,$3D,$FD,$B6
	.dc.b	$33,$3B,$7F,$B8,$C7,$FA,$CB,$8B,$7F,$0C,$69,$27,$AC,$65,$F8,$19
	.dc.b	$7F,$7D,$AF,$C5,$47,$B9,$AE,$8F,$F1,$53,$DD,$C1,$EB,$9E,$9A,$7C
	.dc.b	$A9,$E2,$DE,$4F,$F0,$53,$99,$DD,$FF,$8D,$76,$70,$7F,$BB,$B1,$FF
	.dc.b	$7B,$7C,$47,$C8,$9C,$B6,$75,$3F,$E1,$29,$FC,$9F,$8D,$FE,$17,$D3
	.dc.b	$FD,$57,$C7,$FD,$DA,$BF,$2B,$39,$E1,$F0,$61,$CA,$F7,$39,$AF,$E6
	.dc.b	$17,$C5,$FC,$DB,$67,$FD,$6F,$5C,$6F,$FD,$D3,$C0,$E2,$E5,$DB,$FE
	.dc.b	$AE,$6C,$67,$1F,$F0,$D3,$FE,$1A,$6C,$68,$8E,$BC,$AE,$7E,$6B,$E3
	.dc.b	$69,$27,$FC,$49,$0F,$03,$11,$FE,$DA,$8E,$0E,$A7,$9C,$F9,$BE,$8D
	.dc.b	$F4,$9F,$96,$CD,$18,$BF,$ED,$4D,$7F,$A2,$B5,$FE,$6B,$ED,$7C,$0D
	.dc.b	$7F,$CE,$91,$FC,$EA,$3B,$5F,$7F,$6B,$B5,$9A,$4A,$EE,$D7,$BB,$5F
	.dc.b	$E9,$BF,$02,$F7,$7E,$7B,$63,$FC,$8C,$D7,$FD,$F7,$5D,$DF,$F2,$57
	.dc.b	$2F,$FA,$65,$FE,$C3,$EF,$FC,$18,$C4,$7F,$5C,$8C,$11,$D9,$0F,$EC
	.dc.b	$FA,$7F,$BA,$DF,$F2,$B3,$5F,$DA,$69,$FE,$93,$81,$9E,$4F,$EE,$75
	.dc.b	$F8,$8B,$FE,$27,$7F,$E0,$8E,$8F,$21,$1F,$EE,$ED,$7F,$B8,$B5,$FE
	.dc.b	$E3,$7F,$EE,$C7,$C7,$EF,$7E,$52,$31,$1F,$E0,$69,$FC,$BF,$5B,$E2
	.dc.b	$BF,$04,$D9,$FF,$15,$35,$FF,$02,$B5,$FE,$FE,$D7,$DC,$D7,$DE,$DA
	.dc.b	$FE,$61,$3F,$8F,$F4,$1F,$4B,$FA,$86,$DF,$F9,$FD,$AF,$CA,$39,$CF
	.dc.b	$A3,$E2,$BC,$E4,$8F,$3B,$5F,$3B,$5F,$3B,$E1,$F8,$5F,$01,$23,$FD
	.dc.b	$1D,$8A,$7F,$77,$EC,$3E,$3B,$F4,$CD,$AF,$F4,$5C,$60,$8E,$08,$7F
	.dc.b	$A3,$9E,$3F,$35,$D0,$91,$FE,$83,$8F,$E2,$5D,$6F,$A5,$FF,$44,$CF
	.dc.b	$FA,$2F,$0F,$E1,$BD,$BE,$58,$FF,$62,$D6,$FC,$BF,$3F,$6F,$98,$93
	.dc.b	$FE,$A0,$7F,$7D,$F2,$3E,$98,$F9,$9C,$48,$C7,$3F,$F9,$B5,$BD,$1F
	.dc.b	$99,$DD,$F3,$5F,$C9,$F0,$7D,$2F,$F8,$75,$FE,$BE,$D7,$FD,$88,$D7
	.dc.b	$FD,$92,$D7,$FD,$9C,$D7,$FD,$F0,$D7,$AF,$A1,$FE,$03,$8D,$65,$FC
	.dc.b	$07,$6F,$E5,$FD,$DF,$CA,$EF,$D6,$DD,$3F,$D3,$D1,$C8,$FF,$65,$7C
	.dc.b	$3F,$B1,$FD,$9B,$6C,$D3,$F9,$4C,$7F,$BB,$F8,$FF,$9E,$C6,$5B,$E1
	.dc.b	$FF,$49,$ED,$7D,$2F,$F1,$1C,$7F,$D5,$D1,$8F,$2B,$53,$FD,$27,$1F
	.dc.b	$F3,$F3,$BE,$1F,$EC,$8D,$8C,$CF,$03,$8E,$9F,$47,$E9,$1F,$4B,$F4
	.dc.b	$FE,$97,$EA,$D8,$FD,$40,$72,$E2,$78,$BB,$CF,$E3,$C6,$23,$8C,$F0
	.dc.b	$9F,$96,$DC,$BB,$4A,$B8,$56,$B8,$FF,$B1,$FC,$67,$F9,$3A,$BF,$DF
	.dc.b	$4D,$7A,$C7,$F6,$8F,$A0,$FA,$F5,$FE,$BD,$5F,$0B,$93,$58,$CB,$F4
	.dc.b	$3F,$CB,$67,$71,$83,$F9,$7B,$87,$F3,$EF,$68,$E0,$F8,$49,$1C,$78
	.dc.b	$D1,$F8,$52,$C6,$5A,$F4,$7D,$5C,$5F,$A5,$FA,$57,$D2,$FF,$96,$DB
	.dc.b	$5F,$E4,$46,$BF,$28,$FB,$0E,$1C,$35,$F9,$ED,$77,$1D,$4F,$EF,$5C
	.dc.b	$1F,$E5,$CC,$FF,$DC,$CD,$CB,$FA,$93,$6B,$89,$EB,$E4,$FA,$5D,$72
	.dc.b	$30,$43,$E4,$B9,$7E,$5C,$7E,$87,$C7,$FD,$85,$FD,$3F,$8B,$FB,$86
	.dc.b	$7A,$65,$FC,$98,$CB,$FE,$8D,$AE,$2F,$E6,$7E,$95,$EA,$FF,$A0,$DB
	.dc.b	$1F,$E6,$C6,$BF,$24,$38,$7E,$63,$5F,$CB,$B5,$ED,$1D,$4F,$F1,$17
	.dc.b	$03,$E9,$7F,$28,$49,$FC,$61,$DA,$7F,$80,$7C,$BF,$E0,$B7,$FC,$C7
	.dc.b	$C0,$FE,$6C,$CE,$12,$FF,$11,$19,$7F,$D5,$95,$C7,$5F,$CD,$7D,$07
	.dc.b	$E8,$FE,$86,$57,$E3,$E3,$F4,$C6,$1F,$CD,$C6,$1F,$CA,$7D,$2F,$A8
	.dc.b	$8F,$C5,$E7,$FC,$DC,$7B,$BC,$AF,$D8,$12,$78,$9F,$9D,$E0,$FD,$A1
	.dc.b	$1F,$D0,$45,$7F,$57,$17,$5F,$DC,$4F,$EE,$FE,$43,$E5,$92,$BC,$5B
	.dc.b	$23,$E7,$F5,$FF,$01,$24,$FE,$B6,$3F,$82,$7D,$27,$D4,$24,$72,$7E
	.dc.b	$BF,$D2,$FF,$09,$24,$7E,$D6,$2B,$FC,$02,$2E,$3F,$C3,$66,$EF,$FC
	.dc.b	$26,$6D,$7F,$08,$9F,$E2,$9F,$23,$F8,$42,$47,$A9,$FC,$3F,$B5,$FC
	.dc.b	$64,$91,$FC,$16,$3F,$8E,$7D,$2F,$E3,$09,$1F,$09,$AF,$F0,$1E,$0F
	.dc.b	$E1,$F1,$88,$FE,$5B,$18,$7E,$D7,$CF,$BE,$93,$FD,$1E,$D8,$FE,$30
	.dc.b	$D7,$F8,$FD,$7F,$55,$FA,$57,$27,$FA,$B5,$B3,$FA,$9F,$29,$18,$8F
	.dc.b	$F5,$73,$5F,$E7,$F1,$A9,$E4,$48,$E5,$3B,$7F,$CC,$16,$F9,$0F,$39
	.dc.b	$23,$F9,$B5,$F3,$7D,$31,$97,$FC,$2B,$3C,$5D,$D7,$97,$83,$FB,$6A
	.dc.b	$E7,$FE,$A7,$F1,$1F,$EC,$BC,$7D,$71,$8F,$2F,$FD,$A1,$5B,$C3,$FB
	.dc.b	$35,$EE,$3F,$CF,$0D,$AE,$12,$BE,$6F,$0B,$E7,$3A,$9F,$93,$D7,$70
	.dc.b	$ED,$6C,$3A,$3C,$4E,$91,$A8,$F1,$4F,$82,$5F,$4B,$FE,$C1,$DF,$84
	.dc.b	$63,$C2,$FC,$05,$FF,$B9,$2B,$53,$DF,$3F,$F0,$17,$A8,$F9,$67,$D0
	.dc.b	$F5,$BD,$E1,$C7,$1E,$78,$7F,$7C,$9C,$BC,$AF,$8E,$3E,$29,$EE,$FF
	.dc.b	$59,$6B,$C6,$BF,$24,$DC,$BF,$9B,$3E,$CF,$2B,$FD,$C8,$F1,$6B,$E1
	.dc.b	$6B,$FE,$1B,$AF,$F3,$97,$C2,$72,$6C,$7B,$CD,$4F,$FB,$BF,$26,$5F
	.dc.b	$0B,$D6,$FF,$73,$76,$7A,$6B,$53,$DF,$1C,$1F,$57,$6F,$F3,$27,$92
	.dc.b	$BC,$E5,$DF,$89,$F4,$BF,$F0,$06,$7E,$E4,$63,$0E,$15,$BA,$FF,$1A
	.dc.b	$56,$9E,$11,$99,$DE,$9F,$99,$9C,$B6,$73,$97,$F8,$F7,$0E,$D8,$73
	.dc.b	$C3,$35,$F9,$7C,$38,$DE,$DF,$0C,$77,$AF,$C9,$75,$39,$FF,$DD,$91
	.dc.b	$FC,$2B,$CB,$F4,$F4,$7F,$96,$BE,$1D,$F2,$D8,$AF,$DE,$D5,$C7,$FD
	.dc.b	$BB,$87,$27,$F7,$31,$9D,$58,$72,$BE,$C7,$F0,$27,$15,$FE,$23,$9D
	.dc.b	$E1,$FE,$7A,$8F,$81,$3A,$7F,$89,$B5,$CC,$6D,$FD,$C4,$65,$B1,$99
	.dc.b	$EB,$5F,$CB,$34,$F5,$67,$F0,$DB,$93,$E9,$6C,$61,$ED,$3F,$25,$B3
	.dc.b	$BA,$FE,$6F,$18,$23,$82,$63,$FD,$06,$5F,$1F,$E0,$3C,$3E,$E5,$AF
	.dc.b	$E7,$AF,$62,$3F,$D1,$1F,$0F,$C6,$FA,$0F,$4F,$9B,$67,$FA,$2C,$5C
	.dc.b	$3F,$96,$CD,$E0,$FC,$BE,$7F,$83,$AC,$5F,$F3,$AA,$B7,$F7,$9F,$79
	.dc.b	$23,$FA,$AB,$F0,$BE,$3F,$E4,$9F,$2B,$E0,$EC,$FF,$56,$7E,$39,$1F
	.dc.b	$EB,$AF,$CA,$F9,$FF,$A2,$7D,$2F,$A5,$B3,$FD,$7A,$2E,$1F,$D1,$A6
	.dc.b	$E7,$EA,$D9,$1C,$6F,$FB,$2F,$C1,$5B,$FE,$9A,$FD,$82,$47,$F6,$E7
	.dc.b	$ED,$FD,$2F,$E0,$BE,$57,$F0,$6D,$9F,$ED,$ED,$AF,$EA,$ED,$C7,$FD
	.dc.b	$83,$EA,$3C,$BF,$C5,$FC,$1F,$E8,$E3,$18,$7F,$78,$8C,$4E,$A3,$FC
	.dc.b	$8A,$DC,$3F,$B1,$FD,$07,$E5,$0F,$D8,$7C,$73,$E4,$5E,$56,$70,$FF
	.dc.b	$51,$4E,$63,$F4,$8D,$8F,$F3,$43,$6F,$FD,$24,$D7,$FD,$7A,$D7,$FD
	.dc.b	$98,$D4,$D3,$1D,$6F,$5D,$F4,$9F,$AA,$F3,$3F,$81,$46,$30,$FF,$0B
	.dc.b	$C6,$27,$51,$FE,$83,$6E,$1F,$E0,$2E,$FF,$D9,$FD,$2F,$ED,$1B,$5F
	.dc.b	$E2,$06,$CE,$E3,$FC,$47,$3F,$BC,$F0,$9F,$19,$4F,$93,$5E,$31,$C7
	.dc.b	$FE,$A8,$8C,$E3,$F2,$27,$47,$C4,$29,$F1,$0A,$7C,$42,$EB,$FE,$F3
	.dc.b	$8E,$44,$C7,$FC,$E9,$F4,$BF,$CD,$30,$FD,$8B,$63,$FD,$F5,$DE,$8F
	.dc.b	$E8,$F3,$AC,$3F,$8B,$C6,$D4,$ED,$CF,$8D,$A7,$7F,$01,$C3,$E4,$6E
	.dc.b	$3F,$ED,$FC,$11,$8A,$F8,$73,$E2,$9C,$BB,$AE,$4F,$FB,$BE,$DF,$4B
	.dc.b	$F8,$18,$E9,$F3,$BF,$D1,$67,$3C,$7E,$88,$FE,$91,$8B,$81,$EC,$EF
	.dc.b	$1E,$53,$E2,$06,$E3,$57,$CA,$B8,$E3,$E3,$BE,$3B,$7F,$72,$C3,$84
	.dc.b	$E6,$FE,$4F,$CC,$7C,$AF,$94,$B3,$F0,$DB,$5E,$1E,$0F,$A5,$18,$C3
	.dc.b	$EB,$BF,$D2,$3E,$25,$F3,$F4,$C7,$CB,$DC,$7D,$37,$FC,$BA,$B7,$18
	.dc.b	$3E,$FB,$8C,$7C,$0A,$B9,$47,$8F,$E4,$1F,$19,$FD,$61,$B7,$CE,$7F
	.dc.b	$C3,$1F,$4A,$F2,$75,$7F,$DD,$E9,$39,$7F,$DD,$46,$27,$FB,$53,$4C
	.dc.b	$C6,$EB,$E9,$9F,$E1,$C7,$0E,$23,$83,$A3,$C7,$3D,$27,$71,$F9,$EF
	.dc.b	$A5,$78,$5F,$DC,$9B,$7E,$BD,$DC,$FF,$E3,$C8,$C4,$EA,$FE,$B7,$CA
	.dc.b	$3E,$37,$FB,$AB,$6B,$FD,$C4,$D9,$FF,$58,$35,$F5,$FB,$2F,$A5,$FF
	.dc.b	$6B,$57,$87,$E9,$5C,$DF,$E1,$FA,$D7,$C4,$FF,$82,$27,$35,$E7,$D1
	.dc.b	$EA,$FF,$83,$1B,$1E,$BC,$FF,$9B,$BB,$9F,$F0,$8C,$63,$3F,$2F,$91
	.dc.b	$EA,$25,$7A,$9B,$1A,$8C,$1F,$EF,$FE,$3F,$E9,$B1,$F9,$96,$F3,$FF
	.dc.b	$7E,$F9,$9F,$3C,$E7,$CB,$FF,$10,$C6,$23,$FD,$A2,$D7,$F8,$9B,$5F
	.dc.b	$E1,$ED,$E3,$34,$9C,$1F,$E2,$67,$5F,$07,$FE,$45,$C3,$FC,$1E,$DC
	.dc.b	$BA,$FD,$2F,$F8,$CA,$31,$8F,$68,$E4,$49,$FC,$64,$65,$B8,$F8,$A7
	.dc.b	$53,$FE,$ED,$C1,$18,$8F,$F7,$6F,$D2,$FF,$94,$31,$FF,$10,$B7,$0F
	.dc.b	$E6,$3E,$6F,$EC,$E7,$3E,$67,$F9,$1A,$3F,$6D,$7F,$EF,$66,$BC,$3B
	.dc.b	$FF,$C8,$B1,$88,$FE,$6A,$DC,$FF,$D9,$93,$A2,$61,$EA,$F8,$8F,$E5
	.dc.b	$F7,$FC,$45,$B8,$FF,$8D,$B0,$7D,$2F,$F9,$62,$73,$7E,$88,$F0,$91
	.dc.b	$F8,$B3,$C1,$B1,$9B,$FD,$B6,$3F,$ED,$F8,$C4,$7E,$FA,$3F,$DB,$5E
	.dc.b	$3F,$F9,$9B,$E0,$4F,$F9,$A2,$FF,$D5,$B5,$FC,$9E,$73,$1F,$B9,$33
	.dc.b	$1C,$DA,$FF,$5E,$6B,$FD,$69,$B5,$FE,$B1,$F8,$2F,$F3,$7C,$60,$8F
	.dc.b	$F9,$CA,$9F,$4B,$E5,$3C,$9C,$D6,$BC,$4F,$22,$3F,$88,$3F,$61,$87
	.dc.b	$BF,$E9,$5E,$C7,$FA,$11,$B3,$FC,$4D,$7A,$D9,$99,$39,$78,$46,$2B
	.dc.b	$FA,$57,$D2,$FF,$02,$6F,$67,$FA,$42,$1F,$01,$62,$61,$FC,$86,$78
	.dc.b	$71,$FF,$9D,$9D,$C6,$67,$FD,$5D,$F4,$6F,$A5,$A7,$FA,$C7,$E1,$BF
	.dc.b	$D1,$B5,$FD,$17,$95,$5B,$8C,$37,$3F,$EE,$13,$AB,$E5,$F0,$9D,$2F
	.dc.b	$FD,$70,$FF,$AA,$F0,$EB,$7F,$EB,$96,$BF,$EB,$68,$FF,$5A,$3F,$D2
	.dc.b	$38,$BF,$BD,$FD,$0B,$D0,$FE,$8D,$EA,$FF,$BA,$23,$15,$FD,$26,$B7
	.dc.b	$5A,$C6,$CE,$45,$70,$7F,$C0,$F7,$CE,$BF,$83,$DF,$09,$CD,$D7,$FC
	.dc.b	$1B,$1F,$C3,$1F,$31,$43,$5A,$7F,$12,$FA,$57,$C2,$FF,$9C,$E3,$0D
	.dc.b	$7F,$B1,$46,$BD,$FE,$CD,$8E,$5F,$0D,$FE,$52,$6F,$1E,$A3,$11,$C9
	.dc.b	$AF,$F7,$29,$47,$1D,$3E,$14,$7E,$32,$7E,$03,$5F,$8F,$9F,$1A,$E0
	.dc.b	$DB,$DF,$27,$F7,$EC,$39,$B6,$75,$7C,$87,$23,$F5,$A7,$CE,$7C,$B8
	.dc.b	$C3,$63,$33,$FC,$EA,$72,$FF,$3C,$E4,$7D,$87,$F0,$E8,$FC,$C3,$5F
	.dc.b	$CA,$B7,$37,$0F,$F9,$1A,$F9,$D6,$BB,$FF,$C2,$F5,$A8,$61,$DA,$73
	.dc.b	$7F,$E4,$E8,$C4,$FF,$8A,$EF,$9D,$F1,$97,$E2,$DF,$8F,$8C,$36,$F9
	.dc.b	$36,$FB,$BF,$E0,$A7,$F6,$4E,$6B,$9E,$9E,$29,$DB,$63,$BF,$93,$7E
	.dc.b	$5D,$F5,$D4,$97,$FB,$D8,$7D,$3E,$23,$F7,$71,$C1,$D4,$F3,$9F,$04
	.dc.b	$3F,$07,$FE,$CE,$FE,$8D,$AF,$F7,$06,$BF,$E3,$D6,$BF,$27,$D5,$F7
	.dc.b	$3F,$C2,$9B,$AF,$C9,$6E,$BF,$E4,$AA,$EB,$3F,$54,$7E,$A3,$FE,$94
	.dc.b	$BF,$EF,$F1,$8F,$84,$FD,$B3,$63,$33,$B7,$FC,$A1,$19,$6C,$65,$FC
	.dc.b	$1B,$5F,$F3,$AB,$5F,$C2,$FD,$BB,$E2,$7E,$93,$63,$F9,$53,$5F,$F4
	.dc.b	$B5,$77,$3F,$D2,$1F,$0B,$FA,$F8,$FF,$31,$35,$FF,$34,$B5,$FF,$4E
	.dc.b	$B5,$FF,$51,$F4,$7F,$11,$8C,$11,$FE,$2D,$C2,$FC,$C7,$FA,$9D,$7F
	.dc.b	$3D,$6B,$FC,$E9,$B7,$E4,$AF,$C7,$2F,$9D,$79,$5F,$31,$B7,$F4,$1A
	.dc.b	$FC,$C6,$BF,$AC,$6B,$FE,$9E,$6B,$FE,$EC,$6C,$F7,$9F,$4F,$E9,$1F
	.dc.b	$EC,$1A,$FE,$A8,$D7,$FA,$8B,$67,$F0,$95,$E2,$7F,$D7,$9F,$82,$BE
	.dc.b	$CF,$F9,$B3,$6F,$F9,$D3,$5F,$F2,$3B,$5F,$F4,$A3,$5F,$F4,$DB,$5F
	.dc.b	$F6,$7B,$53,$9C,$FE,$87,$81,$FB,$5D,$3F,$98,$D7,$D1,$8F,$D5,$D7
	.dc.b	$87,$0F,$E2,$1D,$6F,$B2,$FF,$4F,$DF,$F2,$16,$BF,$EA,$78,$D7,$C1
	.dc.b	$FF,$29,$9C,$CF,$FA,$C2,$72,$F1,$9D,$11,$FE,$55,$F9,$EF,$E8,$B9
	.dc.b	$FF,$8E,$36,$3F,$90,$B6,$B9,$C7,$FB,$BC,$7E,$73,$FB,$C7,$FD,$E2
	.dc.b	$D7,$3E,$D7,$F6,$58,$C1,$1C,$12,$3F,$B6,$47,$F2,$37,$AF,$7F,$F1
	.dc.b	$08,$C1,$FC,$0B,$0F,$E6,$5F,$9B,$FC,$CA,$FF,$09,$FE,$70,$F9,$BF
	.dc.b	$E4,$8D,$8F,$F0,$E4,$EF,$6F,$9F,$F9,$C3,$E5,$BF,$A9,$36,$BF,$A8
	.dc.b	$35,$FA,$0D,$7F,$DB,$BC,$9F,$CB,$BA,$BF,$99,$78,$46,$9F,$CD,$7E
	.dc.b	$AB,$EF,$6B,$FE,$E4,$FC,$51,$F5,$7F,$E0,$96,$E3,$FE,$0D,$6B,$FE
	.dc.b	$9A,$6B,$FE,$E7,$6B,$CE,$39,$11,$FF,$0B,$7E,$2B,$FC,$ED,$FB,$0D
	.dc.b	$D7,$FA,$F4,$23,$17,$CE,$07,$DF,$FB,$17,$C1,$7F,$42,$FB,$27,$D2
	.dc.b	$D3,$FA,$2C,$EA,$B8,$E3,$DA,$31,$1F,$89,$AE,$2E,$B0,$EE,$A3,$2F
	.dc.b	$87,$0E,$BF,$67,$E1,$FA,$F9,$3C,$4F,$3E,$09,$DC,$64,$E9,$3C,$1F
	.dc.b	$64,$7A,$3C,$9F,$F1,$D5,$7F,$3D,$93,$49,$1A,$24,$7F,$90,$E3,$87
	.dc.b	$8B,$FE,$FC,$8F,$03,$5F,$D4,$37,$3F,$89,$5F,$DB,$23,$F1,$93,$F0
	.dc.b	$9A,$FD,$26,$BF,$CC,$B9,$7F,$CA,$11,$82,$38,$3F,$CA,$9A,$BE,$97
	.dc.b	$FB,$CF,$D9,$FF,$A8,$AF,$E2,$35,$FD,$E4,$ED,$F9,$0D,$7F,$80,$46
	.dc.b	$F1,$FC,$4B,$77,$FB,$35,$B9,$FC,$6A,$FA,$7F,$19,$2F,$FC,$D5,$19
	.dc.b	$6C,$65,$FF,$7C,$C6,$5B,$1F,$E6,$FB,$FF,$4D,$47,$AB,$DF,$F6,$9F
	.dc.b	$E1,$6A,$E1,$7F,$9E,$FC,$81,$F6,$3F,$E1,$AF,$C5,$5F,$67,$4F,$F0
	.dc.b	$EC,$EB,$1F,$C4,$35,$FE,$41,$3C,$1F,$C6,$35,$FE,$49,$1B,$BF,$D6
	.dc.b	$FD,$AF,$F5,$79,$FC,$E8,$70,$FE,$81,$AF,$EC,$1A,$EF,$E8,$FF,$D1
	.dc.b	$D1,$82,$3F,$E9,$1A,$FD,$3C,$F6,$6B,$FA,$C6,$BF,$BC,$6A,$69,$32
	.dc.b	$E8,$78,$3E,$A7,$FC,$83,$66,$9F,$22,$3F,$C8,$BF,$19,$FE,$3C,$BF
	.dc.b	$7B,$5F,$DE,$B5,$FE,$1A,$D7,$FD,$EB,$ED,$7F,$AA,$E3,$04,$71,$D3
	.dc.b	$27,$FD,$65,$3F,$C3,$1E,$F1,$97,$E7,$9F,$F0,$01,$C3,$A3,$FD,$19
	.dc.b	$FA,$0F,$E7,$5F,$ED,$0D,$7F,$D0,$CD,$7F,$D3,$2D,$F0,$9F,$BE,$FC
	.dc.b	$39,$FF,$74,$35,$FF,$72,$B7,$6F,$EF,$B5,$FA,$5E,$5F,$F4,$57,$67
	.dc.b	$E6,$AF,$FD,$D6,$DA,$FF,$45,$36,$7F,$D2,$17,$CD,$E3,$C9,$FE,$81
	.dc.b	$9D,$5F,$FA,$61,$B8,$F9,$B6,$FF,$44,$79,$6F,$C1,$5F,$6D,$FE,$B1
	.dc.b	$6E,$1F,$EA,$F6,$BF,$E0,$F6,$BF,$D8,$3E,$57,$FC,$27,$AF,$FA,$7D
	.dc.b	$BA,$7F,$A9,$5B,$3F,$EC,$0A,$E7,$5F,$E1,$EF,$DD,$1F,$73,$FE,$C1
	.dc.b	$6D,$7F,$AF,$DA,$FF,$87,$DA,$FF,$70,$F7,$9F,$41,$FE,$A1,$6C,$7F
	.dc.b	$43,$DB,$FD,$2D,$3F,$02,$B9,$E1,$FE,$B5,$BD,$DA,$31,$5C,$CE,$18
	.dc.b	$59,$E9,$5C,$2B,$71,$FE,$C1,$BE,$B9,$FF,$68,$FA,$17,$DF,$FF,$09
	.dc.b	$AF,$EB,$B1,$8E,$EF,$CA,$C2,$30,$FE,$61,$AF,$F0,$8D,$1F,$8B,$7F
	.dc.b	$76,$8F,$F6,$AB,$63,$3E,$77,$FB,$16,$B8,$E7,$FD,$FB,$F5,$5F,$E7
	.dc.b	$D8,$7F,$12,$6B,$FD,$C1,$AF,$F8,$01,$A9,$A4,$EA,$F4,$C6,$5F,$F4
	.dc.b	$7F,$E2,$EE,$3F,$F6,$A3,$67,$EA,$35,$FF,$25,$57,$F4,$97,$FC,$1F
	.dc.b	$FA,$CF,$F8,$DD,$FD,$CB,$5F,$E5,$2D,$7F,$C0,$ED,$78,$FA,$BE,$48
	.dc.b	$C4,$1A,$6F,$47,$DC,$7F,$85,$33,$FF,$7C,$37,$87,$FD,$FC,$DB,$EA
	.dc.b	$7E,$35,$C5,$72,$EE,$FF,$79,$46,$E7,$C1,$8F,$FB,$DF,$EE,$DF,$B8
	.dc.b	$7F,$85,$67,$FC,$5B,$8F,$F8,$82,$73,$E7,$FF,$88,$A3,$15,$BA,$FF
	.dc.b	$1B,$D6,$9E,$27,$F8,$D9,$C3,$A8,$E1,$3C,$AB,$A4,$EE,$72,$D8,$FF
	.dc.b	$18,$FC,$AF,$F7,$3C,$FE,$2F,$DA,$63,$CC,$F8,$6D,$B6,$5E,$0B,$FE
	.dc.b	$ED,$F7,$8F,$BD,$CD,$27,$31,$86,$D7,$F9,$39,$FF,$01,$73,$3E,$9F
	.dc.b	$FC,$E1,$1A,$BE,$59,$7F,$91,$B1,$ED,$87,$18,$C6,$3D,$5F,$D9,$61
	.dc.b	$C9,$CC,$FE,$C2,$B8,$DF,$F4,$66,$BF,$DD,$9A,$F7,$F2,$5F,$35,$C5
	.dc.b	$B1,$C5,$AF,$16,$BC,$63,$CC,$FF,$9B,$27,$FC,$D3,$18,$CD,$F7,$AF
	.dc.b	$A3,$71,$70,$C1,$87,$07,$E1,$9C,$F1,$F1,$ED,$F4,$9B,$A7,$F1,$3C
	.dc.b	$7A,$61,$CF,$0E,$57,$FE,$31,$D1,$F9,$0F,$F2,$CA,$F3,$F0,$BE,$D3
	.dc.b	$FC,$FF,$1E,$AA,$DF,$8B,$E6,$D7,$C7,$7B,$D7,$19,$75,$CB,$A5,$7E
	.dc.b	$69,$B8,$FE,$85,$B1,$F9,$B6,$BF,$8E,$6E,$3E,$09,$FD,$0C,$E7,$1F
	.dc.b	$14,$73,$6C,$7F,$A6,$B1,$F5,$7B,$9E,$7E,$AF,$44,$62,$FF,$D0,$DE
	.dc.b	$BF,$C6,$BF,$E8,$91,$C7,$5F,$F5,$BB,$FD,$35,$B1,$A1,$5C,$DF,$F7
	.dc.b	$35,$73,$BE,$F5,$C1,$FF,$58,$C7,$79,$FF,$22,$B5,$FF,$51,$B5,$FF
	.dc.b	$59,$B5,$FF,$6C,$70,$FF,$AE,$9D,$B5,$DB,$5D,$E1,$FE,$24,$C1,$F1
	.dc.b	$AF,$47,$E2,$4F,$C7,$8C,$3F,$EA,$9F,$C6,$5F,$84,$6B,$0F,$7D,$69
	.dc.b	$F8,$42,$77,$5F,$D7,$9B,$1F,$E3,$46,$BF,$E6,$46,$BF,$EC,$26,$BF
	.dc.b	$ED,$26,$BD,$6B,$D9,$1A,$C1,$F2,$7F,$E8,$FB,$FF,$01,$36,$FE,$B7
	.dc.b	$3B,$EA,$3F,$C4,$35,$AE,$8F,$64,$E7,$CF,$F9,$4D,$9F,$EE,$95,$C2
	.dc.b	$FF,$CF,$73,$97,$FB,$B7,$E3,$2F,$7F,$E3,$66,$8C,$61,$FE,$D7,$F4
	.dc.b	$9F,$72,$FA,$9D,$D6,$B1,$FC,$66,$1C,$71,$C9,$F4,$70,$E3,$78,$C7
	.dc.b	$FC,$41,$87,$1E,$5F,$CE,$B6,$3F,$DF,$75,$C1,$DD,$F3,$CB,$C5,$1F
	.dc.b	$E6,$89,$D6,$5F,$E9,$86,$BF,$EF,$66,$BF,$AD,$F9,$1F,$CF,$4D,$E5
	.dc.b	$F3,$7E,$30,$FB,$8F,$87,$5A,$B6,$0A,$D6,$1E,$16,$CF,$85,$AF,$85
	.dc.b	$AF,$EF,$7F,$19,$7D,$BF,$D0,$6D,$7F,$90,$FD,$B7,$D0,$E6,$A4,$FF
	.dc.b	$91,$EB,$77,$FE,$72,$AC,$B6,$79,$47,$D1,$E9,$79,$3F,$7B,$1E,$B5
	.dc.b	$F7,$67,$2F,$CB,$6D,$FC,$B6,$BF,$2D,$AF,$F8,$13,$F2,$C7,$D2,$FE
	.dc.b	$EE,$91,$87,$FD,$3A,$62,$35,$19,$9F,$F6,$7C,$E4,$DC,$66,$3F,$C1
	.dc.b	$CD,$7C,$13,$FC,$27,$C9,$7E,$AD,$FA,$99,$12,$33,$D9,$B9,$D4,$E5
	.dc.b	$AF,$F6,$86,$CF,$0C,$3F,$D6,$AD,$9F,$F7,$2B,$5E,$E7,$1F,$85,$FE
	.dc.b	$21,$1E,$47,$F8,$94,$3E,$F7,$FC,$21,$5E,$6E,$9F,$EA,$97,$B8,$C1
	.dc.b	$D2,$77,$5F,$EC,$66,$BF,$EF,$66,$BB,$F3,$BF,$95,$46,$23,$E1,$B5
	.dc.b	$F1,$46,$8F,$E6,$56,$FC,$25,$F6,$F8,$6E,$1F,$ED,$69,$D4,$E4,$91
	.dc.b	$99,$DD,$FC,$43,$FD,$30,$E1,$F8,$F5,$CA,$7F,$D1,$BF,$87,$FF,$30
	.dc.b	$8F,$87,$A7,$F8,$E2,$7E,$1D,$7D,$CA,$D4,$7D,$9F,$2F,$E6,$C6,$27
	.dc.b	$ED,$46,$38,$7E,$9D,$F3,$9F,$EC,$5A,$BF,$00,$7C,$AF,$D8,$6E,$5E
	.dc.b	$0A,$65,$FD,$27,$3F,$AF,$5A,$9F,$E9,$73,$FD,$35,$F8,$0D,$7F,$0A
	.dc.b	$D7,$D5,$E4,$FF,$5F,$8C,$47,$68,$C1,$2F,$FD,$63,$F8,$07,$FA,$5D
	.dc.b	$F8,$79,$7A,$7E,$07,$F1,$B3,$9B,$F6,$E2,$9C,$FB,$DE,$D6,$E1,$CA
	.dc.b	$74,$FF,$3F,$FC,$65,$ED,$7F,$75,$F0,$11,$89,$FE,$60,$DC,$7F,$98
	.dc.b	$35,$FE,$60,$D7,$F9,$DB,$5F,$E5,$6D,$9E,$B5,$FB,$BB,$75,$FD,$33
	.dc.b	$DB,$F6,$57,$1F,$A9,$A2,$3F,$C1,$0D,$8F,$8F,$18,$23,$82,$65,$F9
	.dc.b	$51,$3C,$5C,$74,$FD,$DE,$1F,$BF,$E2,$FE,$EA,$31,$7F,$C3,$72,$FE
	.dc.b	$CA,$FE,$1E,$33,$7F,$E7,$29,$CB,$FE,$10,$8C,$F0,$7E,$46,$FA,$7C
	.dc.b	$C7,$E4,$7E,$46,$E3,$F5,$5B,$87,$DC,$6E,$5F,$92,$7F,$31,$E7,$4A
	.dc.b	$F1,$79,$16,$FC,$0F,$F8,$B7,$3F,$F1,$A1,$1C,$12,$7F,$DF,$F1,$97
	.dc.b	$FC,$65,$F1,$5F,$E1,$5C,$7F,$CA,$6D,$8F,$F2,$C6,$5F,$B1,$F3,$53
	.dc.b	$9C,$7F,$B4,$F1,$7F,$8E,$AF,$FB,$C3,$63,$FB,$C3,$5F,$EF,$0D,$7F
	.dc.b	$CC,$9F,$34,$FC,$1B,$FC,$2F,$F3,$8F,$A1,$FF,$0C,$CE,$70,$FE,$0F
	.dc.b	$92,$B5,$1F,$BB,$F2,$1F,$09,$FC,$D1,$B1,$FB,$FE,$8F,$D5,$6F,$FE
	.dc.b	$87,$D3,$FB,$4D,$FF,$3A,$9F,$D7,$47,$F1,$D6,$BF,$D0,$9A,$FF,$52
	.dc.b	$6B,$F9,$CE,$CF,$F2,$F4,$60,$8F,$F9,$8F,$27,$D0,$FF,$0F,$C7,$FD
	.dc.b	$08,$DB,$FF,$09,$FD,$E7,$EE,$1F,$F1,$8C,$67,$B1,$39,$9F,$F1,$13
	.dc.b	$5F,$F1,$6B,$5E,$7D,$DF,$BF,$8C,$4F,$28,$C1,$2F,$FC,$E9,$F8,$57
	.dc.b	$F7,$9D,$7F,$D2,$AD,$8F,$EC,$7F,$A5,$7F,$AC,$B1,$FF,$22,$CE,$5F
	.dc.b	$EE,$B1,$97,$FA,$33,$5E,$3F,$05,$FE,$83,$8C,$47,$08,$C1,$2B,$F7
	.dc.b	$50,$9D,$E8,$FA,$97,$D5,$7F,$97,$9B,$7F,$D3,$32,$7E,$9D,$F8,$6C
	.dc.b	$7F,$CB,$75,$A7,$FD,$51,$3A,$E6,$FE,$CF,$5B,$9F,$EB,$6D,$7F,$BE
	.dc.b	$35,$4F,$1B,$FE,$D5,$39,$8C,$4F,$F6,$76,$BF,$E0,$CC,$5F,$9F,$BE
	.dc.b	$AB,$FC,$EE,$DC,$3F,$AB,$7E,$A8,$F0,$73,$5F,$E1,$08,$C4,$FC,$66
	.dc.b	$E1,$FC,$85,$AF,$F7,$76,$BF,$E1,$C6,$BF,$E2,$C6,$A6,$9A,$C3,$C5
	.dc.b	$F7,$FF,$E8,$CC,$0F,$FA,$CF,$2F,$F4,$77,$C1,$E6,$31,$EF,$7E,$FF
	.dc.b	$1F,$8E,$DA,$DE,$5F,$E1,$18,$C1,$2F,$FC,$32,$D8,$FF,$64,$7C,$47
	.dc.b	$F9,$E7,$23,$FE,$7B,$C3,$F8,$55,$FF,$A8,$23,$1D,$FE,$19,$CD,$FE
	.dc.b	$32,$30,$79,$6D,$F8,$BF,$F3,$CC,$BF,$DE,$AD,$AF,$F7,$D2,$F8,$9F
	.dc.b	$F5,$6B,$9C,$3F,$DE,$2D,$BE,$7E,$46,$AB,$6F,$FA,$D3,$F1,$7F,$F1
	.dc.b	$5E,$2C,$7F,$CB,$B6,$9C,$C6,$1F,$E0,$93,$BC,$7F,$DF,$F1,$A6,$FA
	.dc.b	$DD,$4E,$DD,$76,$FF,$94,$FD,$BF,$ED,$F1,$3F,$B9,$ED,$7D,$C7,$DC
	.dc.b	$AD,$FB,$9F,$EC,$AA,$E0,$6E,$75,$97,$FB,$1A,$31,$C9,$FE,$0D,$9C
	.dc.b	$CF,$FA,$DD,$AF,$FB,$09,$AF,$F9,$2F,$E3,$7F,$CF,$BF,$72,$FA,$FF
	.dc.b	$F3,$9E,$58,$7F,$BE,$4D,$AF,$F3,$AF,$A3,$F0,$EF,$84,$73,$0E,$1F
	.dc.b	$F4,$45,$F1,$7F,$CF,$B5,$55,$73,$BE,$F1,$FC,$48,$4D,$69,$97,$F8
	.dc.b	$BF,$DD,$3E,$F1,$F7,$98,$6D,$7F,$85,$67,$53,$93,$8F,$77,$F0,$C9
	.dc.b	$D3,$5F,$F3,$84,$FF,$07,$E5,$D9,$23,$DF,$8F,$FA,$62,$7F,$92,$C7
	.dc.b	$BD,$AE,$8F,$F5,$5B,$CD,$C0,$71,$A7,$C9,$78,$B7,$C3,$FF,$52,$B9
	.dc.b	$C3,$D0,$FD,$2E,$2F,$C5,$F1,$7F,$3A,$6E,$1F,$EF,$DE,$0F,$F6,$6B
	.dc.b	$67,$F9,$FF,$1F,$E5,$3E,$13,$FD,$17,$C5,$FA,$4E,$4F,$F7,$4D,$FE
	.dc.b	$6A,$72,$FF,$BE,$B0,$E6,$FE,$4E,$33,$7C,$E3,$81,$FE,$90,$C7,$C6
	.dc.b	$FF,$3A,$C3,$F6,$3F,$79,$F8,$7E,$0F,$D7,$5A,$31,$3A,$E0,$EB,$0F
	.dc.b	$6E,$C4,$C3,$FC,$2F,$3C,$15,$F3,$AF,$2F,$DE,$3E,$7E,$92,$BF,$C3
	.dc.b	$F0,$FB,$CC,$47,$9C,$E7,$1C,$5D,$CF,$49,$F8,$B0,$FA,$47,$DB,$7E
	.dc.b	$6D,$B1,$FD,$B9,$AF,$F9,$A5,$AF,$F5,$7F,$BC,$F7,$3F,$E3,$26,$BE
	.dc.b	$3C,$91,$8D,$BF,$C5,$D5,$D1,$FE,$30,$3E,$F1,$F2,$5F,$DF,$A3,$0D
	.dc.b	$8C,$E1,$FE,$D5,$6B,$FD,$AD,$AF,$FB,$41,$AF,$FB,$C5,$AF,$F7,$66
	.dc.b	$E0,$FB,$C7,$CB,$FE,$19,$B5,$FE,$4E,$6B,$FE,$06,$AF,$0E,$1F,$E4
	.dc.b	$9F,$87,$FD,$5C,$FF,$B8,$DA,$88,$E1,$BD,$7F,$C2,$E3,$11,$FE,$D6
	.dc.b	$8C,$11,$FF,$4D,$62,$FB,$B7,$E0,$1F,$9C,$6C,$FF,$99,$1A,$FE,$14
	.dc.b	$E8,$7F,$88,$FC,$C7,$DA,$FC,$F6,$C7,$FC,$01,$6B,$FE,$17,$6B,$FE
	.dc.b	$FC,$6B,$ED,$C9,$18,$8F,$F1,$1B,$63,$33,$FB,$CF,$51,$F0,$7F,$B4
	.dc.b	$6D,$FF,$9F,$1A,$FF,$8D,$2F,$C7,$7F,$E7,$7F,$96,$7D,$EF,$F3,$36
	.dc.b	$D7,$F9,$19,$AF,$FB,$05,$AF,$D3,$3E,$A3,$87,$0D,$73,$1D,$4F,$E2
	.dc.b	$79,$7F,$8D,$63,$E7,$78,$7F,$44,$FD,$6D,$78,$2F,$FB,$A7,$73,$EF
	.dc.b	$7F,$D3,$57,$FE,$D3,$8C,$7B,$1F,$C8,$E7,$33,$E5,$13,$A7,$84,$69
	.dc.b	$F5,$46,$5E,$0D,$73,$1B,$7F,$D1,$A7,$F2,$2F,$77,$FC,$A5,$E0,$E2
	.dc.b	$3F,$91,$4F,$8B,$0F,$F5,$27,$DE,$FF,$00,$9F,$90,$D7,$F3,$8D,$4D
	.dc.b	$27,$67,$F6,$76,$C7,$F9,$E9,$FE,$44,$F0,$F0,$BF,$8A,$C6,$0F,$E0
	.dc.b	$D8,$7F,$B1,$35,$FE,$73,$F0,$6F,$85,$A7,$F3,$CF,$85,$FA,$98,$FF
	.dc.b	$23,$D9,$F7,$AF,$A0,$FE,$A0,$D9,$FA,$4D,$7E,$4B,$5F,$F3,$A7,$88
	.dc.b	$ED,$FE,$92,$D8,$FF,$68,$B5,$FE,$41,$F1,$3F,$81,$D7,$F1,$DF,$7A
	.dc.b	$FA,$3F,$F0,$43,$6B,$FD,$36,$D7,$F0,$01,$C3,$F8,$36,$BD,$23,$99
	.dc.b	$1F,$F5,$2F,$DE,$FF,$41,$7F,$80,$37,$93,$FD,$BC,$DB,$E5,$3F,$B9
	.dc.b	$7F,$49,$F7,$9F,$EE,$88,$C1,$FC,$67,$EF,$5F,$0F,$FD,$29,$B1,$A1
	.dc.b	$3C,$5F,$E2,$0D,$C3,$FD,$DD,$3C,$27,$F8,$A3,$63,$FD,$E5,$3B,$BF
	.dc.b	$1F,$D5,$77,$F8,$C9,$87,$7A,$4F,$F5,$C6,$C7,$8F,$05,$70,$8F,$1D
	.dc.b	$6F,$CA,$FF,$1D,$5F,$F4,$49,$F4,$C6,$08,$FF,$B4,$23,$44,$8F,$F2
	.dc.b	$17,$88,$FC,$03,$FB,$6B,$6B,$E2,$35,$F8,$73,$AE,$4E,$24,$8F,$F6
	.dc.b	$EC,$FE,$45,$FE,$2C,$D7,$FA,$53,$5F,$E7,$31,$C5,$FC,$4E,$AF,$BE
	.dc.b	$FE,$FF,$F7,$CF,$B3,$A7,$F8,$0A,$75,$5F,$65,$AF,$F4,$79,$E0,$FD
	.dc.b	$D6,$BF,$D2,$E3,$7C,$1F,$E9,$18,$FC,$2F,$D0,$FF,$99,$64,$D2,$47
	.dc.b	$FB,$EA,$32,$D8,$CB,$EA,$7F,$CD,$D4,$FB,$9F,$59,$27,$D1,$3F,$E1
	.dc.b	$CF,$BD,$7C,$1F,$F8,$7B,$EF,$BF,$DC,$D5,$F9,$56,$BF,$D8,$E7,$6F
	.dc.b	$E6,$5A,$FF,$67,$8D,$E5,$ED,$6F,$2F,$F4,$9A,$DC,$FC,$1F,$BE,$FE
	.dc.b	$AF,$7F,$CE,$5A,$FF,$4F,$6B,$FE,$03,$6A,$69,$3C,$EF,$F4,$73,$AA
	.dc.b	$FD,$44,$FF,$45,$6B,$E3,$D7,$FD,$29,$18,$8F,$F0,$83,$53,$49,$1D
	.dc.b	$BE,$F3,$FC,$8B,$46,$9F,$36,$3F,$C9,$3B,$FF,$8F,$AF,$FA,$B3,$5F
	.dc.b	$ED,$8D,$7F,$C4,$0D,$7F,$25,$ED,$7F,$AA,$63,$04,$71,$F2,$F2,$7F
	.dc.b	$91,$9F,$99,$70,$FF,$26,$9F,$E1,$CF,$8A,$32,$D3,$FB,$77,$DF,$7E
	.dc.b	$71,$FE,$CE,$D7,$FC,$69,$E1,$A3,$11,$FD,$C1,$BE,$A3,$EF,$BE,$24
	.dc.b	$FF,$24,$8C,$67,$FE,$8C,$9C,$CF,$F5,$1A,$FD,$2F,$07,$FB,$EB,$B7
	.dc.b	$F3,$53,$C5,$AF,$F9,$9D,$BC,$3F,$E8,$EB,$E6,$FF,$A4,$23,$35,$CF
	.dc.b	$97,$FD,$37,$18,$BF,$F4,$5D,$F0,$AF,$6F,$CD,$BE,$77,$FD,$60,$DA
	.dc.b	$FF,$09,$35,$FF,$04,$B5,$FF,$13,$68,$F9,$BF,$F0,$25,$FF,$96,$E3
	.dc.b	$1E,$97,$FA,$B2,$73,$5F,$E1,$8B,$E9,$A7,$F8,$8B,$EF,$5F,$7D,$FE
	.dc.b	$C1,$6D,$7F,$88,$DA,$FF,$86,$9A,$F8,$BD,$F7,$C9,$7F,$AE,$9B,$1F
	.dc.b	$4D,$AF,$F8,$06,$79,$DF,$FA,$E2,$72,$DA,$E4,$7F,$B2,$38,$D5,$C3
	.dc.b	$5F,$75,$F0,$6D,$70,$C3,$FD,$91,$5B,$5F,$62,$FB,$EF,$E1,$71,$FE
	.dc.b	$4F,$8C,$77,$7E,$71,$B1,$FE,$00,$6B,$FE,$41,$E0,$7D,$F7,$F7,$78
	.dc.b	$FF,$63,$B6,$33,$E7,$7F,$B8,$2B,$8E,$DF,$CC,$FE,$F7,$F9,$BE,$1F
	.dc.b	$E4,$B6,$BF,$E6,$16,$BF,$E9,$56,$A6,$93,$AF,$F6,$71,$97,$FD,$EB
	.dc.b	$F7,$DF,$85,$8F,$D2,$35,$FF,$61,$36,$BC,$93,$FE,$D2,$FB,$EF,$F1
	.dc.b	$D5,$7E,$F9,$AF,$F8,$89,$AF,$F9,$8D,$AF,$F9,$E5,$AF,$FA,$B5,$A9
	.dc.b	$A6,$B2,$7D,$F7,$F8,$36,$7F,$DE,$2D,$8C,$F7,$79,$7D,$6F,$F4,$36
	.dc.b	$3F,$F0,$02,$B4,$49,$D5,$FF,$89,$BE,$B9,$F7,$DF,$E1,$28,$FE,$C9
	.dc.b	$18,$F4,$3F,$D2,$1C,$DF,$E9,$18,$C4,$FF,$B1,$63,$15,$C5,$B5,$C2
	.dc.b	$7F,$D9,$55,$C1,$EB,$5B,$72,$D5,$1B,$F9,$6F,$ED,$FA,$7C,$2F,$A7
	.dc.b	$FF,$51,$46,$33,$46,$3C,$BE,$79,$5C,$7F,$DB,$5F,$78,$FB,$2E,$B0
	.dc.b	$C7,$FB,$DA,$9D,$67,$FE,$E1,$9C,$C6,$27,$86,$1F,$EE,$5E,$57,$DF
	.dc.b	$7F,$9C,$EF,$78,$F6,$6E,$3C,$EF,$FD,$75,$8F,$39,$FF,$05,$E1,$CB
	.dc.b	$D1,$F4,$65,$E3,$C7,$CE,$FF,$AB,$6F,$C4,$D8,$CB,$D1,$AF,$C0,$F2
	.dc.b	$DF,$37,$FD,$52,$BF,$CC,$6D,$8C,$CF,$F9,$EA,$72,$D8,$CE,$3F,$09
	.dc.b	$B9,$7C,$1F,$BE,$7D,$63,$93,$DD,$87,$87,$2A,$FC,$DC,$38,$B9,$3E
	.dc.b	$59,$FC,$4A,$F9,$D7,$0C,$BC,$79,$7F,$8C,$36,$7D,$27,$F2,$CA,$EF
	.dc.b	$C8,$FB,$EF,$F3,$FC,$78,$AB,$7E,$1F,$97,$3E,$F7,$BD,$31,$97,$5C
	.dc.b	$BA,$57,$E8,$9B,$8F,$E8,$5B,$1F,$9C,$6B,$F9,$86,$E3,$E0,$9F,$D3
	.dc.b	$4E,$F1,$F1,$47,$36,$C7,$F5,$5C,$BF,$D6,$13,$FD,$DD,$B0,$E9,$FF
	.dc.b	$2C,$71,$7C,$68,$C5,$71,$F7,$7D,$D5,$EF,$D7,$E2,$46,$27,$47,$FA
	.dc.b	$86,$F9,$56,$F0,$EE,$F6,$E1,$FF,$33,$62,$25,$78,$5F,$F8,$00,$D7
	.dc.b	$47,$FB,$66,$38,$BC,$1A,$E6,$FF,$C4,$F3,$C6,$35,$1F,$EC,$CC,$5F
	.dc.b	$1A,$F4,$BD,$54,$9C,$C6,$21,$F7,$CF,$B5,$DD,$6B,$0F,$F6,$04,$E5
	.dc.b	$FF,$1B,$B6,$BF,$DB,$6D,$7D,$67,$4A,$F4,$C6,$63,$DE,$D7,$E3,$35
	.dc.b	$E1,$3E,$BF,$95,$FF,$46,$BF,$E3,$36,$C7,$DA,$3F,$B9,$FD,$F7,$F8
	.dc.b	$86,$75,$D1,$F0,$3A,$93,$9A,$F2,$C6,$3D,$1F,$9E,$DB,$FE,$ED,$5B
	.dc.b	$AF,$F1,$2F,$D4,$BD,$FF,$8A,$DA,$FF,$74,$F9,$EF,$A7,$7D,$F6,$AE
	.dc.b	$57,$F9,$2E,$F8,$61,$E9,$C3,$85,$EA,$3D,$9C,$0A,$DF,$2F,$E5,$5B
	.dc.b	$87,$F9,$73,$2E,$EF,$5C,$BE,$FE,$5D,$23,$E9,$E2,$8C,$3F,$59,$AF
	.dc.b	$E4,$9A,$FF,$90,$7E,$4B,$F3,$F3,$79,$BA,$FD,$EB,$EF,$BF,$0F,$5A
	.dc.b	$BE,$D3,$9B,$F0,$E0,$8C,$4F,$85,$AF,$85,$AF,$F9,$47,$EF,$5F,$7D
	.dc.b	$F9,$46,$DF,$F9,$B7,$E8,$B1,$32,$3F,$CF,$13,$AA,$FF,$38,$56,$5B
	.dc.b	$3C,$4F,$B3,$D1,$FD,$91,$C4,$7D,$58,$C3,$F1,$DA,$FC,$76,$BF,$E8
	.dc.b	$E6,$BF,$CF,$7E,$F9,$F4,$DF,$BC,$A4,$61,$FA,$A6,$23,$51,$99,$FD
	.dc.b	$5C,$E4,$DC,$66,$3F,$C1,$ED,$7F,$C1,$CD,$7F,$AD,$FD,$E7,$EA,$1F
	.dc.b	$E1,$FE,$42,$31,$1F,$C5,$E3,$1D,$5F,$C4,$9B,$7F,$8A,$6C,$FE,$F9
	.dc.b	$AF,$4A,$E6,$F8,$2B,$94,$3E,$1B,$F8,$9D,$78,$5F,$E2,$5F,$7D,$FE
	.dc.b	$0F,$C3,$F9,$47,$3F,$F5,$1A,$D4,$60,$FE,$4F,$3B,$C7,$F3,$AD,$7F
	.dc.b	$88,$B5,$DF,$97,$FC,$AA,$31,$1F,$CD,$5A,$9A,$6A,$DF,$7C,$FA,$FF
	.dc.b	$E6,$B1,$86,$C6,$70,$FD,$7C,$E5,$B1,$9B,$FE,$4B,$39,$6C,$65,$FF
	.dc.b	$29,$B5,$FE,$8B,$F7,$DF,$CB,$E7,$E2,$69,$FE,$40,$BF,$89,$5F,$D3
	.dc.b	$EB,$51,$F6,$E7,$53,$FC,$9C,$4E,$9F,$C2,$F9,$3F,$D0,$E3,$1A,$7E
	.dc.b	$16,$3E,$BF,$DD,$BE,$5A,$9B,$F1,$AE,$5E,$2C,$7A,$C7,$F5,$8C,$7F
	.dc.b	$A9,$4F,$FB,$F9,$B3,$FC,$51,$AF,$F4,$86,$BF,$02,$79,$12,$7F,$8D
	.dc.b	$C3,$EF,$3F,$DB,$11,$F0,$B2,$62,$F7,$FF,$AF,$56,$AF,$FA,$45,$6A
	.dc.b	$73,$F0,$BE,$E6,$E1,$CF,$0E,$4F,$F3,$EF,$BE,$7B,$7F,$DD,$BB,$91
	.dc.b	$89,$FE,$3E,$DC,$BF,$97,$35,$FE,$5C,$D7,$FC,$0D,$3C,$AB,$A4,$E6
	.dc.b	$7F,$9C,$37,$2F,$6F,$C4,$FD,$43,$DB,$F7,$D7,$1F,$A5,$AA,$3F,$C1
	.dc.b	$2D,$8F,$92,$DC,$3F,$A2,$B5,$FE,$F2,$D7,$FC,$2A,$D7,$FC,$38,$D7
	.dc.b	$E9,$4F,$F8,$5B,$A3,$F0,$D5,$F2,$7C,$5F,$F0,$74,$62,$BE,$8E,$BF
	.dc.b	$D9,$EB,$F1,$13,$9E,$5F,$E0,$13,$9A,$FC,$C4,$E6,$BF,$8B,$4E,$6B
	.dc.b	$FC,$AD,$5F,$C3,$7E,$67,$F2,$FF,$71,$E0,$6D,$F6,$6C,$F4,$6D,$78
	.dc.b	$3C,$DF,$CF,$9E,$8F,$1E,$DF,$99,$F7,$11,$FC,$9E,$5F,$E3,$3C,$3F
	.dc.b	$8F,$4E,$5F,$EE,$31,$93,$FC,$71,$E5,$3E,$2B,$FC,$37,$8F,$F9,$45
	.dc.b	$B8,$7F,$92,$65,$97,$ED,$DE,$1E,$07,$F7,$EE,$4F,$F1,$FD,$7F,$6E
	.dc.b	$6E,$5F,$DF,$5A,$FF,$7D,$6B,$D7,$E7,$1F,$71,$FE,$40,$F9,$E7,$C7
	.dc.b	$7F,$90,$A7,$38,$7F,$0E,$C9,$5A,$8F,$E0,$1F,$02,$F8,$4F,$F2,$63
	.dc.b	$63,$F8,$27,$47,$EC,$78,$3E,$BC,$E7,$4F,$E2,$93,$FB,$57,$FC,$92
	.dc.b	$D7,$FC,$CA,$D7,$FC,$D8,$D4,$D2,$76,$FF,$9A,$1E,$19,$3E,$93,$F8
	.dc.b	$8E,$1F,$E7,$F6,$E3,$FB,$2F,$BE,$FD,$FE,$3F,$E3,$29,$CB,$FE,$29
	.dc.b	$8C,$BF,$E2,$76,$BC,$7B,$FF,$82,$46,$23,$84,$60,$95,$FE,$24,$9D
	.dc.b	$60,$FB,$EF,$F0,$06,$3F,$E9,$36,$D7,$FA,$1B,$EF,$BF,$D6,$F8,$FF
	.dc.b	$91,$A7,$2F,$FA,$3E,$32,$FF,$9C,$9A,$F1,$F8,$2F,$F4,$5C,$62,$38
	.dc.b	$46,$09,$5F,$CB,$A7,$58,$3E,$A9,$F5,$9F,$E8,$66,$D7,$F5,$2C,$9F
	.dc.b	$7D,$F8,$5C,$BF,$CB,$75,$A7,$FC,$F1,$3A,$E5,$FE,$D7,$5B,$9F,$F5
	.dc.b	$1C,$E6,$FF,$B5,$CE,$67,$FC,$0A,$D7,$FD,$7C,$D9,$E1,$1F,$DB,$A3
	.dc.b	$58,$BE,$F9,$F4,$FF,$E9,$E6,$E1,$FD,$67,$EF,$9E,$0E,$6B,$FD,$4D
	.dc.b	$18,$9F,$EA,$2D,$C7,$FC,$DA,$D7,$FD,$42,$D7,$FD,$7A,$D7,$FD,$88
	.dc.b	$D4,$D3,$58,$78,$BE,$F3,$FD,$67,$81,$FF,$5D,$E7,$FE,$85,$F8,$3C
	.dc.b	$C6,$3D,$EF,$DE,$E3,$FE,$E1,$6D,$6F,$2F,$F0,$AC,$60,$97,$FE,$1C
	.dc.b	$6C,$7F,$B2,$FE,$23,$FC,$F9,$91,$FF,$3F,$61,$FC,$22,$FF,$D8,$F1
	.dc.b	$8E,$EF,$0C,$E6,$FF,$BB,$C6,$0F,$2D,$BE,$FB,$F9,$D4,$FF,$1F,$6D
	.dc.b	$7F,$B5,$D7,$C4,$FF,$B5,$1C,$D7,$1F,$27,$06,$66,$DF,$F1,$7F,$BE
	.dc.b	$FF,$18,$6B,$FE,$FD,$6C,$7F,$2D,$7F,$7F,$19,$F5,$BF,$DC,$31,$AC
	.dc.b	$BF,$DD,$0D,$C4,$F0,$BE,$55,$FE,$E8,$FB,$AF,$D9,$D1,$FD,$AF,$A2
	.dc.b	$FB,$AF,$B5,$7F,$CA,$E7,$64,$F7,$3F,$D7,$33,$AC,$BF,$D8,$73,$CA
	.dc.b	$31,$1F,$EB,$76,$BF,$E0,$A6,$BF,$EB,$F6,$A2,$7C,$6D,$FB,$B7,$D8
	.dc.b	$7F,$9E,$F1,$C3,$FE,$C8,$9B,$5F,$E8,$89,$AF,$31,$18,$5E,$AA,$71
	.dc.b	$1A,$17,$E0,$7A,$18,$BF,$03,$FE,$3E,$38,$1F,$E4,$0B,$C3,$83,$C5
	.dc.b	$F7,$8F,$BE,$7D,$F7,$F9,$1A,$30,$D8,$CC,$FF,$B9,$27,$2D,$ED,$71
	.dc.b	$F1,$6C,$66,$30,$47,$D0,$7C,$D9,$FE,$4A,$FC,$66,$BA,$0F,$F1,$28
	.dc.b	$D4,$F6,$CB,$E2,$56,$BC,$2F,$9D,$18,$BE,$6D,$AE,$19,$FF,$81,$F7
	.dc.b	$7C,$27,$F9,$9B,$8F,$F2,$4D,$8C,$E5,$F6,$4E,$11,$BF,$C2,$5F,$27
	.dc.b	$FC,$CB,$87,$FC,$25,$18,$AF,$F0,$13,$5F,$CA,$D6,$D6,$09,$D6,$DF
	.dc.b	$83,$9C,$C6,$2F,$AE,$1C,$6F,$CF,$5F,$E9,$CF,$8E,$FB,$7F,$1B,$D5
	.dc.b	$B1,$9E,$24,$F1,$FB,$FF,$93,$9F,$7B,$DD,$FD,$CC,$BF,$45,$18,$9D
	.dc.b	$61,$CA,$BF,$C3,$D7,$CA,$F7,$F2,$0F,$8C,$F1,$36,$BC,$CD,$7B,$1F
	.dc.b	$0B,$EF,$DF,$41,$C2,$A4,$F0,$7E,$04,$EA,$3F,$C5,$DF,$7C,$F8,$5F
	.dc.b	$B4,$D8,$FB,$07,$CE,$70,$FD,$BD,$D1,$88,$FD,$B6,$DF,$B3,$F1,$BE
	.dc.b	$56,$86,$9F,$46,$B9,$EF,$F5,$F0,$7D,$F3,$E5,$7F,$18,$DC,$BC,$73
	.dc.b	$A9,$C9,$23,$33,$B7,$CC,$D7,$F8,$54,$F1,$7D,$B9,$3E,$93,$FD,$75
	.dc.b	$1F,$83,$6B,$F3,$A3,$5E,$B7,$E2,$6F,$B6,$3E,$9F,$BB,$7D,$EF,$E8
	.dc.b	$DB,$87,$A1,$AF,$C9,$6B,$E4,$6B,$FE,$73,$6B,$F9,$E6,$E4,$73,$88
	.dc.b	$7C,$37,$E6,$5B,$5F,$9D,$6B,$F8,$8A,$E5,$7F,$A3,$F5,$5F,$0B,$FB
	.dc.b	$B6,$D7,$D0,$6B,$F8,$16,$BF,$31,$AF,$FA,$2D,$AF,$ED,$9B,$5F,$E3
	.dc.b	$D9,$FF,$76,$DB,$EF,$1F,$55,$FB,$16,$E1,$FB,$56,$BF,$E3,$EA,$FD
	.dc.b	$D6,$7F,$E2,$E8,$F9,$DE,$0F,$D1,$23,$82,$1F,$45,$CB,$FC,$57,$07
	.dc.b	$B8,$FC,$53,$F8,$13,$67,$F8,$33,$5F,$95,$7E,$17,$FC,$F4,$7F,$23
	.dc.b	$F7,$BF,$A6,$69,$FE,$83,$8C,$10,$FC,$DC,$68,$8F,$C4,$FC,$57,$F9
	.dc.b	$4B,$F9,$16,$F0,$FF,$0F,$6F,$47,$FA,$E6,$BF,$97,$4F,$FB,$5B,$EF
	.dc.b	$7F,$9E,$DF,$F3,$16,$BF,$DA,$DA,$FF,$77,$6A,$69,$3C,$6F,$EC,$CF
	.dc.b	$0C,$7F,$87,$F5,$7F,$2B,$EE,$FD,$F5,$FF,$AE,$5A,$FF,$16,$3A,$5F
	.dc.b	$F3,$9F,$3D,$F7,$AF,$A0,$C3,$6B,$FC,$15,$9F,$F3,$BA,$DD,$7F,$39
	.dc.b	$F8,$77,$C4,$FF,$3B,$F8,$B7,$DF,$FF,$53,$6E,$3F,$C2,$1A,$FF,$19
	.dc.b	$6B,$FC,$13,$47,$CA,$7F,$41,$FA,$47,$CD,$53,$DD,$1F,$D4,$69,$F7
	.dc.b	$BF,$CA,$DF,$E4,$ED,$7F,$B0,$35,$FF,$0D,$35,$FF,$13,$35,$E7,$7F
	.dc.b	$E1,$38,$C1,$BE,$AF,$DA,$F9,$BF,$D5,$7B,$7F,$A5,$5F,$F1,$56,$F0
	.dc.b	$FF,$61,$8E,$67,$CA,$E8,$FF,$15,$7D,$EB,$E2,$BF,$AF,$7C,$B3,$E5
	.dc.b	$29,$ED,$9D,$5F,$13,$FC,$78,$E1,$FE,$49,$3C,$5F,$ED,$4D,$7F,$01
	.dc.b	$1B,$C3,$C8,$FF,$BA,$A3,$FB,$87,$CE,$3E,$CF,$FB,$83,$63,$C0,$27
	.dc.b	$83,$D3,$E4,$37,$52,$37,$3A,$BF,$85,$F7,$6F,$BC,$FB,$0D,$AE,$AD
	.dc.b	$7F,$2D,$F7,$BF,$2B,$6F,$F2,$84,$E6,$F9,$DF,$23,$F0,$55,$BB,$FF
	.dc.b	$30,$56,$EF,$FC,$C5,$E0,$3F,$13,$FC,$BD,$7C,$B8,$C7,$5F,$38,$F9
	.dc.b	$DF,$82,$BF,$23,$F9,$5A,$7F,$65,$EC,$7D,$67,$E7,$CF,$F1,$15,$6F
	.dc.b	$92,$DF,$0D,$3C,$37,$C3,$1F,$03,$5F,$F0,$4D,$F1,$7F,$C4,$5F,$87
	.dc.b	$7E,$9B,$2F,$F0,$E3,$6F,$FC,$36,$D9,$FF,$3F,$7E,$00,$FB,$5F,$F1
	.dc.b	$44,$E7,$AC,$D3,$85,$2B,$83,$F7,$9F,$C0,$3F,$E1,$06,$BF,$E4,$96
	.dc.b	$BF,$E8,$36,$BE,$2E,$5F,$F3,$D4,$62,$3B,$46,$0F,$E9,$F2,$FC,$0F
	.dc.b	$CE,$49,$F1,$4F,$EF,$7F,$01,$FF,$7A,$C7,$C4,$C7,$DA,$67,$E5,$A4
	.dc.b	$64,$87,$F1,$F9,$E2,$FF,$8D,$B4,$FF,$6E,$78,$DF,$E5,$9F,$7D,$F2
	.dc.b	$BF,$A5,$DF,$FD,$51,$18,$EF,$FF,$2F,$5F,$0C,$3F,$64,$D7,$FC,$83
	.dc.b	$3B,$7F,$6C,$D7,$C1,$F2,$BF,$CA,$DF,$E6,$5F,$80,$3F,$00,$FF,$50
	.dc.b	$36,$3F,$C7,$E1,$C3,$FE,$42,$6B,$FE,$7B,$6B,$FE,$B5,$6B,$FE,$C1
	.dc.b	$6B,$CE,$3E,$8F,$DB,$7E,$18,$91,$FB,$E9,$EF,$87,$F3,$FF,$CE,$9F
	.dc.b	$69,$FD,$37,$A2,$5F,$F1,$3F,$3D,$5A,$8C,$35,$FF,$0F,$79,$4A,$E3
	.dc.b	$1F,$E6,$8A,$E3,$7F,$E8,$3F,$01,$F8,$9F,$E6,$BE,$56,$5F,$ED,$E7
	.dc.b	$2D,$9C,$BF,$E4,$5E,$A5,$73,$9F,$F4,$77,$E0,$3F,$13,$C3,$34,$9C
	.dc.b	$3F,$EA,$9F,$23,$F4,$3C,$1F,$E6,$36,$F3,$7F,$0A,$6E,$7F,$EA,$0B
	.dc.b	$F8,$F3,$9E,$CF,$D8,$FD,$94,$AF,$E1,$D7,$C3,$23,$CB,$C0,$FD,$AD
	.dc.b	$6E,$BF,$C8,$93,$BF,$A5,$7D,$1F,$FA,$F5,$B3,$FE,$72,$6B,$FE,$83
	.dc.b	$6B,$FE,$82,$E0,$FE,$C9,$AF,$F0,$AE,$5F,$E0,$A7,$FA,$C6,$B8,$C7
	.dc.b	$FA,$07,$F0,$17,$CE,$FF,$B2,$DB,$87,$FA,$01,$AF,$FA,$45,$AF,$7F
	.dc.b	$C1,$FF,$88,$91,$FE,$5D,$1C,$2B,$F8,$DC,$E6,$7E,$C3,$FE,$C3,$9D
	.dc.b	$70,$7F,$1E,$8D,$57,$0A,$E7,$7F,$EC,$9F,$C0,$7F,$BD,$70,$1A,$4C
	.dc.b	$FF,$87,$FC,$17,$F2,$42,$5F,$F4,$19,$F8,$B9,$7F,$46,$FC,$01,$AF
	.dc.b	$FB,$AD,$B5,$FE,$A0,$6B,$FE,$B1,$6B,$FE,$9C,$6B,$CE,$39,$11,$FF
	.dc.b	$4F,$EC,$FA,$AF,$E6,$E4,$8F,$EA,$95,$DB,$5F,$F5,$A7,$DF,$3E,$B3
	.dc.b	$FD,$FC,$DA,$FF,$59,$B5,$FF,$62,$35,$FF,$57,$B5,$FE,$C6,$D7,$FD
	.dc.b	$F2,$DD,$8E,$79,$7F,$A4,$D7,$F4,$C2,$57,$F6,$99,$F0,$5F,$F5,$69
	.dc.b	$CC,$FE,$CD,$FF,$7F,$4E,$B4,$4B,$ED,$1F,$7E,$F0,$F3,$2F,$A1,$7D
	.dc.b	$87,$A3,$7C,$C6,$27,$D3,$18,$9E,$B8,$74,$BC,$57,$7C,$3B,$3E,$2C
	.dc.b	$39,$6D,$F0,$31,$DB,$67,$58,$F4,$BE,$31,$E5,$8F,$23,$A3,$C6,$59
	.dc.b	$E4,$EC,$F1,$61,$E4,$5F,$7D,$FE,$5E,$CB,$94,$63,$E0,$3E,$33,$C4
	.dc.b	$C4,$77,$F2,$5F,$7F,$FD,$37,$5F,$89,$5A,$38,$61,$D5,$B8,$74,$9F
	.dc.b	$95,$7C,$BC,$DF,$25,$EE,$FE,$99,$CC,$AE,$5F,$AE,$D7,$EE,$BF,$E5
	.dc.b	$DF,$A5,$FE,$B6,$D8,$E6,$7C,$0F,$35,$3A,$6C,$6A,$78,$61,$EC,$6E
	.dc.b	$DF,$DD,$BE,$0F,$FC,$D1,$C5,$F2,$27,$37,$E2,$C3,$9E,$1C,$5B,$87
	.dc.b	$1F,$5D,$87,$19,$CD,$B8,$7F,$C7,$F1,$A8,$FB,$51,$89,$FC,$07,$03
	.dc.b	$CF,$FE,$53,$1F,$81,$F6,$3F,$CF,$CE,$A3,$37,$CA,$72,$FE,$16,$B9
	.dc.b	$78,$1D,$F4,$FC,$AD,$69,$FD,$04,$EB,$0E,$D9,$6F,$4F,$D0,$E5,$DA
	.dc.b	$F8,$CE,$72,$75,$7F,$91,$32,$F6,$75,$7F,$93,$9B,$97,$F9,$A1,$7C
	.dc.b	$F6,$2F,$8F,$6F,$F9,$8A,$BF,$CA,$F5,$9C,$3A,$47,$F9,$B6,$38,$BA
	.dc.b	$31,$7F,$EA,$5B,$EC,$D7,$E8,$4F,$07,$F2,$3B,$FF,$9C,$27,$4D,$AF
	.dc.b	$C5,$B6,$77,$5C,$5F,$E1,$71,$97,$FB,$27,$E2,$5E,$7C,$4F,$B6,$5F
	.dc.b	$14,$F9,$BD,$4E,$63,$18,$F8,$1B,$87,$99,$AF,$81,$AF,$FA,$49,$AF
	.dc.b	$E3,$5B,$7E,$0B,$F6,$C6,$B3,$FF,$3C,$FD,$6F,$9C,$F3,$7C,$DF,$F2
	.dc.b	$4C,$7F,$D0,$F1,$8E,$2F,$8B,$39,$CB,$E4,$B5,$F8,$6D,$7F,$D1,$7E
	.dc.b	$67,$C3,$3E,$17,$AE,$F8,$7C,$63,$FE,$94,$9B,$4E,$2F,$D2,$6D,$FE
	.dc.b	$6A,$BC,$C7,$FA,$62,$AD,$FF,$AC,$B0,$E8,$F3,$3F,$D6,$4E,$30,$FB
	.dc.b	$ED,$7E,$CB,$5E,$57,$F8,$7D,$1F,$90,$FE,$EA,$FF,$DC,$D1,$87,$EE
	.dc.b	$7D,$FF,$F0,$6B,$F4,$47,$F5,$29,$D3,$FE,$C3,$9D,$4F,$E2,$1A,$FF
	.dc.b	$B2,$9A,$FF,$05,$6E,$6F,$BE,$7D,$0F,$E8,$F1,$7E,$1D,$F8,$D9,$CE
	.dc.b	$3E,$69,$CB,$FE,$D3,$8C,$E3,$E6,$3F,$6F,$3A,$71,$87,$7C,$3B,$61
	.dc.b	$D5,$98,$7F,$94,$3E,$9A,$DF,$F9,$31,$FD,$4F,$A6,$69,$27,$F9,$AE
	.dc.b	$9F,$33,$A7,$D5,$D5,$EB,$6C,$FA,$5B,$87,$9E,$B8,$4F,$B6,$BF,$84
	.dc.b	$7B,$6F,$CA,$7F,$C0,$33,$FE,$F1,$FB,$9F,$F3,$FB,$FE,$8B,$8C,$E1
	.dc.b	$EE,$9C,$BF,$E8,$78,$CF,$2F,$E4,$27,$2F,$F3,$5B,$ED,$5F,$BB,$9C
	.dc.b	$DF,$DF,$E0,$79,$1F,$54,$FE,$39,$E3,$3E,$F9,$F2,$19,$BF,$F0,$BE
	.dc.b	$7F,$D0,$6B,$53,$FC,$5D,$B3,$FD,$71,$B9,$7F,$68,$6B,$FE,$1C,$6B
	.dc.b	$FE,$22,$6B,$FD,$8A,$FC,$33,$CD,$B1,$FD,$CB,$37,$D6,$3E,$87,$DD
	.dc.b	$F8,$17,$CE,$8C,$61,$F3,$63,$13,$A8,$FF,$5C,$4E,$AF,$E7,$57,$C2
	.dc.b	$C3,$FC,$69,$5F,$95,$F2,$71,$E4,$7E,$6D,$B1,$93,$E2,$9E,$B9,$E3
	.dc.b	$7F,$83,$DF,$F6,$1B,$7F,$69,$E3,$FD,$BF,$3F,$E9,$4F,$A3,$F7,$8F
	.dc.b	$96,$C6,$5D,$31,$FF,$02,$CE,$63,$E7,$36,$3E,$73,$5F,$9C,$D7,$E7
	.dc.b	$35,$E7,$3F,$66,$31,$C2,$75,$87,$F0,$CF,$A2,$FD,$D1,$E6,$FA,$6B
	.dc.b	$8F,$EE,$62,$E3,$FB,$7C,$39,$56,$33,$FF,$04,$7C,$7B,$F2,$27,$BF
	.dc.b	$FD,$3F,$F3,$4F,$8B,$9F,$F9,$92,$B5,$39,$E2,$FE,$E1,$18,$9F,$89
	.dc.b	$E0,$AF,$8E,$8A,$D3,$FC,$1A,$33,$A7,$B7,$C0,$FE,$29,$F2,$9F,$41
	.dc.b	$B8,$FC,$F9,$D3,$78,$7F,$DA,$EF,$D3,$C1,$F3,$2F,$BF,$E5,$3B,$C2
	.dc.b	$D7,$F1,$93,$73,$8E,$0F,$D5,$60,$FB,$7F,$EE,$5E,$27,$E0,$9B,$8F
	.dc.b	$85,$B8,$7C,$DA,$E1,$3F,$CF,$BD,$1F,$E3,$4F,$9F,$E4,$BF,$C2,$5F
	.dc.b	$37,$FE,$AD,$8C,$1F,$D1,$1C,$5F,$39,$DF,$BF,$F9,$26,$D7,$F4,$AD
	.dc.b	$BF,$A6,$57,$F1,$D6,$E9,$EA,$FA,$9F,$F1,$04,$62,$BF,$1F,$1F,$8D
	.dc.b	$23,$F9,$86,$D7,$E6,$1A,$FE,$61,$FF,$0E,$F8,$DF,$DE,$6F,$F5,$1F
	.dc.b	$8A,$FE,$DD,$FE,$41,$E9,$7F,$8B,$E3,$18,$7F,$AB,$39,$DF,$24,$F8
	.dc.b	$CF,$F1,$3B,$67,$FC,$4E,$DC,$FF,$C5,$AD,$7F,$D6,$3E,$43,$E6,$BF
	.dc.b	$C4,$75,$FB,$38,$C7,$AF,$FE,$27,$FA,$07,$C7,$7F,$3E,$BD,$E1,$FC
	.dc.b	$E7,$CD,$FE,$64,$7F,$92,$5F,$F7,$07,$C4,$63,$2F,$C6,$72,$FF,$B7
	.dc.b	$5F,$F4,$2C,$E9,$B1,$98,$FF,$44,$B5,$FF,$4A,$B5,$FF,$64,$B5,$FF
	.dc.b	$67,$B5,$E8,$7F,$A7,$3C,$67,$D2,$BD,$9F,$F2,$D3,$6F,$F5,$FF,$4B
	.dc.b	$FE,$3E,$E5,$FE,$35,$39,$C3,$FB,$9C,$FF,$8E,$89,$EE,$FF,$95,$9B
	.dc.b	$7F,$E5,$66,$CF,$F9,$71,$FF,$1F,$7C,$3B,$F1,$1F,$F3,$53,$6B,$F7
	.dc.b	$FF,$80,$3E,$5B,$35,$FE,$57,$8C,$4F,$F9,$89,$B8,$7F,$9A,$1A,$FF
	.dc.b	$98,$9A,$FE,$1D,$AF,$FA,$7D,$BD,$06,$9A,$D9,$F6,$BF,$C0,$6F,$FC
	.dc.b	$5E,$DB,$FF,$0A,$FD,$6B,$ED,$BF,$D0,$93,$9E,$0F,$F2,$3C,$E6,$FF
	.dc.b	$CF,$2D,$BF,$F3,$FB,$5F,$F3,$CB,$5F,$CF,$35,$FF,$5B,$37,$23,$4D
	.dc.b	$63,$F3,$7F,$14,$FF,$4B,$68,$7F,$D0,$1F,$7E,$FA,$FC,$C6,$1F,$F4
	.dc.b	$7C,$67,$2F,$F4,$7B,$72,$FF,$4A,$B5,$FF,$47,$B5,$FD,$A3,$5F,$F6
	.dc.b	$63,$72,$39,$C7,$FD,$C7,$F3,$7F,$EA,$BF,$34,$FF,$A5,$7C,$67,$C4
	.dc.b	$66,$30,$FF,$A9,$23,$39,$7F,$A9,$1B,$97,$FA,$A9,$AF,$FA,$91,$AF
	.dc.b	$F0,$96,$BF,$EE,$16,$E4,$73,$8B,$E2,$DF,$47,$FE,$C1,$6E,$1F,$D5
	.dc.b	$BE,$37,$FA,$AD,$6E,$BF,$D5,$D1,$FE,$68,$27,$AB,$FE,$BC,$7F,$CD
	.dc.b	$B3,$FD,$6F,$4F,$F6,$67,$AB,$F0,$B0,$FE,$C3,$F7,$BF,$D0,$B5,$FF
	.dc.b	$28,$4E,$5B,$5C,$0F,$F5,$E4,$EF,$97,$FC,$25,$5B,$7F,$C2,$9E,$37
	.dc.b	$E5,$70,$FF,$42,$37,$06,$EF,$BD,$7D,$EF,$FB,$7E,$B5,$A1,$A7,$1A
	.dc.b	$E1,$3B,$F5,$7F,$D9,$DF,$74,$F7,$29,$FE,$D1,$FB,$BF,$DC,$EB,$FE
	.dc.b	$AC,$9D,$6B,$FE,$00,$6E,$BF,$E8,$26,$CF,$FA,$CA,$B8,$D7,$F4,$39
	.dc.b	$CB,$FD,$16,$3F,$7A,$4F,$03,$FD,$D6,$FF,$02,$B1,$39,$6F,$73,$EC
	.dc.b	$7F,$CB,$98,$61,$DD,$4A,$C3,$FE,$5F,$9A,$E4,$FE,$13,$3C,$9A,$72
	.dc.b	$3F,$18,$60,$FF,$33,$47,$27,$C8,$7F,$46,$36,$7F,$48,$9C,$38,$3C
	.dc.b	$BF,$0C,$FB,$C7,$DE,$7F,$4E,$8C,$36,$33,$3E,$33,$FC,$DE,$E1,$F3
	.dc.b	$4F,$F0,$7D,$79,$77,$FF,$B8,$E7,$84,$E5,$B1,$97,$8B,$BC,$FE,$64
	.dc.b	$7D,$09,$FE,$50,$FC,$86,$BF,$0D,$E9,$1C,$23,$FD,$07,$3C,$F3,$F7
	.dc.b	$F8,$7F,$52,$B5,$7C,$9B,$5B,$BF,$CE,$EB,$FE,$FF,$7E,$07,$1A,$31
	.dc.b	$1F,$56,$31,$F2,$3F,$DD,$9B,$1C,$3F,$12,$F2,$68,$E0,$FF,$0A,$46
	.dc.b	$1B,$AF,$E2,$27,$F3,$18,$59,$CD,$FF,$97,$AF,$83,$FB,$26,$E1,$E0
	.dc.b	$8F,$F6,$86,$FF,$ED,$B7,$FE,$00,$B5,$CF,$BD,$D5,$74,$7E,$0B,$CA
	.dc.b	$C3,$F8,$49,$B5,$C7,$3F,$F4,$C3,$72,$FD,$2D,$71,$C3,$FD,$2C,$BE
	.dc.b	$6D,$F0,$BD,$1B,$3A,$6B,$FE,$FD,$5E,$0F,$F2,$D8,$D4,$F1,$23,$FE
	.dc.b	$A3,$9D,$F8,$7E,$8F,$C6,$DE,$87,$F8,$A3,$14,$62,$FF,$DF,$6D,$7F
	.dc.b	$C4,$0D,$7F,$73,$0F,$C5,$DF,$19,$F1,$1B,$3E,$E6,$BF,$EF,$FA,$E9
	.dc.b	$B7,$DF,$8F,$E4,$FF,$76,$FA,$3F,$C3,$36,$3F,$1A,$D7,$F1,$8D,$7F
	.dc.b	$85,$F1,$A3,$11,$FE,$5D,$6B,$DA,$7A,$9F,$07,$07,$BD,$E9,$24,$FF
	.dc.b	$AA,$2B,$C1,$87,$FB,$4F,$0F,$C3,$4F,$C2,$6B,$F7,$9A,$FE,$21,$B9
	.dc.b	$FF,$44,$E3,$FA,$11,$82,$3F,$4E,$3F,$C5,$53,$F8,$66,$AA,$FF,$72
	.dc.b	$46,$1F,$24,$FE,$5E,$9F,$87,$BD,$7F,$D8,$37,$0F,$DB,$B5,$FD,$B3
	.dc.b	$5F,$96,$D7,$FD,$0C,$D7,$FA,$EB,$67,$BC,$FA,$7E,$77,$F4,$B8,$9A
	.dc.b	$79,$6B,$C1,$8F,$FB,$F7,$B5,$F2,$3F,$93,$6E,$5F,$CA,$F0,$FF,$24
	.dc.b	$CE,$49,$19,$9D,$BF,$E7,$56,$BF,$96,$6B,$FB,$2F,$BB,$FC,$8F,$45
	.dc.b	$E4,$FD,$DC,$62,$7F,$7F,$5D,$31,$FC,$B6,$AF,$D1,$DF,$80,$7F,$1B
	.dc.b	$6D,$FF,$22,$6B,$FC,$85,$AF,$E5,$9A,$FF,$AC,$5A,$FF,$82,$1B,$3D
	.dc.b	$E7,$F6,$D8,$7F,$8E,$AB,$F8,$4B,$5F,$C2,$65,$ED,$F4,$BF,$05,$87
	.dc.b	$8B,$1F,$D7,$74,$BD,$1F,$EB,$98,$7F,$57,$8C,$78,$DF,$EA,$A6,$C6
	.dc.b	$67,$6F,$F6,$F8,$CB,$63,$2F,$F7,$C8,$CB,$63,$2F,$F2,$2F,$C5,$FF
	.dc.b	$DF,$4F,$F2,$26,$BE,$A3,$F9,$7B,$8B,$F4,$CF,$89,$FA,$7F,$03,$FB
	.dc.b	$8A,$FF,$08,$C6,$3A,$3F,$45,$39,$9F,$E5,$1F,$8E,$71,$D7,$FC,$2F
	.dc.b	$E8,$3E,$63,$FD,$55,$5F,$DE,$63,$1E,$D7,$FB,$31,$B1,$99,$DB,$FE
	.dc.b	$E2,$E0,$FE,$87,$F4,$9F,$CD,$23,$F9,$EB,$5F,$67,$27,$FB,$BA,$7F
	.dc.b	$A6,$43,$EE,$9F,$2F,$FE,$15,$6C,$7F,$B7,$1A,$FF,$83,$5A,$FF,$88
	.dc.b	$5A,$FF,$8B,$1A,$FF,$8D,$1A,$F7,$8F,$E8,$1F,$09,$FD,$56,$3F,$AE
	.dc.b	$B7,$AC,$E6,$38,$1F,$D6,$3E,$37,$C7,$C5,$FC,$E6,$91,$8D,$FF,$AF
	.dc.b	$B6,$BF,$B4,$BF,$D6,$AB,$83,$9A,$E3,$87,$F6,$C8,$C1,$BB,$FE,$D3
	.dc.b	$F3,$BF,$D1,$F0,$FC,$35,$A1,$F3,$DF,$DD,$1B,$3F,$DB,$9E,$18,$F9
	.dc.b	$DB,$7D,$2B,$81,$C5,$FA,$BF,$2A,$FA,$6F,$F2,$43,$70,$FF,$1F,$07
	.dc.b	$0F,$F9,$05,$AF,$F9,$55,$AF,$F9,$81,$AF,$F9,$A1,$AF,$38,$EB,$EE
	.dc.b	$BE,$E3,$F1,$59,$7F,$12,$F8,$E7,$D3,$7F,$3F,$6E,$1F,$A2,$CB,$FC
	.dc.b	$E7,$87,$1F,$43,$FA,$27,$D9,$3D,$FA,$4B,$E0,$7F,$A6,$21,$F7,$59
	.dc.b	$8F,$CC,$C7,$17,$7C,$E8,$C4,$6A,$7F,$C4,$BF,$3E,$F9,$6B,$F1,$92
	.dc.b	$BF,$C4,$33,$96,$CE,$5F,$BB,$1F,$D7,$74,$FF,$14,$BB,$8F,$E0,$C3
	.dc.b	$ED,$FF,$3A,$4E,$8F,$F3,$F3,$63,$EB,$B5,$FF,$3B,$3F,$A6,$E7,$FC
	.dc.b	$FF,$DE,$7F,$8F,$3A,$8D,$3F,$6F,$3F,$9B,$D9,$F3,$8F,$8E,$FC,$86
	.dc.b	$53,$2F,$8B,$39,$6C,$E5,$FC,$CF,$5F,$39,$D4,$7C,$0F,$67,$E4,$91
	.dc.b	$F9,$FF,$BB,$EB,$A7,$F6,$1B,$FF,$90,$67,$81,$F3,$2B,$FC,$05,$5B
	.dc.b	$E0,$FF,$57,$FD,$CB,$E5,$FF,$D4,$4D,$BF,$CD,$B5,$FF,$4D,$35,$FF
	.dc.b	$57,$35,$E7,$1C,$88,$FF,$77,$E1,$FF,$7E,$BF,$E5,$E6,$BF,$B8,$6E
	.dc.b	$5F,$BB,$AF,$03,$FD,$BB,$9B,$FB,$86,$1F,$EC,$CE,$0F,$C7,$7A,$7F
	.dc.b	$9A,$3F,$C3,$DF,$B2,$3E,$D7,$19,$FE,$53,$0E,$2F,$E6,$EB,$73,$F9
	.dc.b	$4E,$D9,$3F,$92,$7F,$C5,$1F,$BE,$7E,$BB,$2F,$C7,$C6,$A7,$F2,$6E
	.dc.b	$BB,$7F,$3B,$C7,$FE,$69,$E3,$FE,$B3,$5F,$E6,$46,$BF,$98,$8D,$7B
	.dc.b	$1F,$A4,$24,$7E,$8A,$2B,$FA,$0F,$A2,$FD,$49,$8E,$0F,$BF,$58,$9F
	.dc.b	$D4,$CE,$72,$FD,$2F,$AD,$E3,$24,$FF,$8A,$71,$FF,$06,$F6,$17,$C7
	.dc.b	$7E,$D8,$93,$FB,$4A,$FD,$CF,$C7,$FB,$49,$1C,$DF,$81,$F4,$DF,$BE
	.dc.b	$24,$7E,$F2,$2B,$FB,$A8,$BA,$7E,$62,$6E,$5F,$BC,$9B,$5F,$EA,$EC
	.dc.b	$55,$FC,$2B,$E7,$7E,$71,$27,$D8,$FF,$A6,$7E,$27,$FD,$1B,$8F,$F5
	.dc.b	$36,$E3,$FE,$B0,$6D,$75,$AF,$F5,$CC,$7F,$7A,$8F,$F4,$DB,$5E,$1E
	.dc.b	$97,$52,$3F,$E5,$38,$D3,$FE,$C4,$6E,$A6,$9F,$DA,$A3,$FD,$7E,$D7
	.dc.b	$FA,$CB,$70,$F4,$57,$F9,$E7,$C8,$FF,$5F,$D7,$FA,$71,$AF,$E8,$DA
	.dc.b	$BA,$7F,$A1,$FC,$37,$FB,$B6,$73,$18,$9F,$F7,$AC,$60,$E3,$8F,$9F
	.dc.b	$D6,$F8,$64,$9F,$E7,$B3,$FD,$C5,$F7,$46,$5F,$4B,$E7,$BF,$47,$A6
	.dc.b	$99,$AF,$A6,$31,$1F,$DE,$71,$FE,$B3,$18,$9D,$4F,$03,$DD,$19,$69
	.dc.b	$93,$E0,$61,$FE,$39,$71,$5D,$1B,$19,$5F,$1F,$FE,$47,$CB,$EB,$7C
	.dc.b	$EF,$92,$31,$B2,$31,$A7,$82,$F8,$BE,$AF,$B0,$FF,$24,$7C,$4F,$BF
	.dc.b	$0F,$F2,$74,$62,$B6,$7C,$37,$11,$FE,$57,$F7,$5F,$37,$FE,$08,$8E
	.dc.b	$3B,$7F,$BA,$1B,$87,$5C,$7C,$13,$97,$E7,$5F,$2F,$4F,$CF,$97,$7A
	.dc.b	$FF,$77,$BC,$2F,$FC,$2F,$97,$96,$3E,$98,$70,$FB,$7E,$AD,$F3,$5C
	.dc.b	$1B,$1C,$0E,$0E,$1E,$72,$8C,$47,$39,$F5,$3F,$0E,$39,$E7,$F0,$FD
	.dc.b	$B7,$80,$E4,$DD,$F6,$C2,$E3,$D5,$B8,$F2,$DB,$E1,$E3,$CB,$0C,$D7
	.dc.b	$83,$91,$FB,$F3,$E5,$7B,$67,$F5,$B9,$7F,$71,$EB,$63,$7F,$F4,$66
	.dc.b	$1C,$2F,$1A,$7F,$0B,$AB,$87,$EC,$71,$E9,$9F,$E4,$9B,$1F,$9F,$6C
	.dc.b	$7E,$8B,$0E,$8E,$F0,$F1,$B9,$9D,$D7,$F6,$16,$BD,$E7,$6F,$F1,$DF
	.dc.b	$5B,$D7,$D7,$FC,$85,$B7,$C5,$B3,$FC,$5D,$BA,$7F,$17,$6F,$47,$F9
	.dc.b	$02,$BF,$C7,$B8,$DB,$FE,$2F,$5C,$B2,$F9,$6F,$F9,$CE,$3B,$57,$E1
	.dc.b	$67,$2D,$AE,$53,$ED,$AE,$0F,$C9,$6D,$70,$8F,$CA,$B5,$FE,$6B,$97
	.dc.b	$98,$F3,$FC,$A5,$C3,$F9,$6B,$D3,$2F,$E6,$4D,$C8,$D2,$4F,$F3,$B6
	.dc.b	$DF,$E9,$2B,$8C,$FF,$3F,$F8,$4F,$8A,$7C,$6F,$C6,$2C,$7F,$3C,$7B
	.dc.b	$11,$FE,$8C,$F8,$7C,$BF,$9A,$7C,$EF,$BB,$B3,$FD,$12,$2E,$1F,$D1
	.dc.b	$26,$EF,$FD,$6E,$6D,$7F,$5B,$9B,$5F,$D4,$5F,$71,$23,$FA,$B3,$F0
	.dc.b	$7F,$0A,$FB,$A7,$DE,$F9,$DB,$3F,$D5,$1F,$8C,$47,$FA,$F3,$F2,$7E
	.dc.b	$F7,$F0,$BF,$5F,$F8,$66,$CF,$F5,$C6,$E1,$FD,$71,$B9,$7F,$73,$78
	.dc.b	$4F,$F6,$2F,$56,$D7,$F6,$27,$EB,$12,$3F,$B7,$3F,$67,$BF,$F2,$A7
	.dc.b	$E4,$7D,$07,$95,$8A,$FE,$DB,$36,$FF,$DD,$F5,$A9,$C9,$FD,$CB,$E5
	.dc.b	$3F,$3E,$7E,$6F,$EE,$FF,$8C,$46,$49,$5F,$DD,$A7,$5D,$FC,$27,$FC
	.dc.b	$B9,$F2,$5F,$AA,$F6,$0F,$EB,$BE,$55,$F2,$EF,$73,$53,$98,$C5,$FF
	.dc.b	$2B,$AD,$D7,$E9,$1B,$19,$9D,$E5,$FC,$E2,$72,$D8,$CB,$FD,$BA,$32
	.dc.b	$D8,$CB,$FE,$23,$FB,$5F,$DF,$9F,$BB,$F7,$7F,$D9,$77,$FE,$16,$8C
	.dc.b	$B4,$FF,$48,$F9,$D8,$A7,$27,$F2,$AF,$03,$FD,$47,$F2,$9F,$C3,$3F
	.dc.b	$1F,$FF,$74,$61,$FE,$A2,$8C,$72,$7F,$09,$F6,$91,$89,$D1,$FE,$D2
	.dc.b	$F0,$17,$CB,$2F,$F4,$46,$7F,$3C,$F9,$5F,$15,$6F,$FD,$B3,$C1,$F4
	.dc.b	$23,$E5,$70,$1A,$49,$F9,$8D,$BF,$94,$D7,$FC,$2F,$8F,$F6,$57,$F9
	.dc.b	$3F,$99,$FD,$92,$BC,$CF,$F8,$5F,$7F,$E9,$75,$FD,$AF,$CB,$DD,$F0
	.dc.b	$6D,$F0,$79,$FC,$14,$8E,$7E,$35,$AF,$9B,$1F,$E2,$BB,$FE,$ED,$59
	.dc.b	$C3,$BE,$3F,$E4,$FA,$F1,$CF,$F3,$EF,$88,$7D,$8B,$EC,$7F,$9C,$7D
	.dc.b	$8F,$F6,$5C,$FA,$8A,$DC,$EB,$D2,$FE,$8F,$F2,$2F,$C4,$3F,$C4,$99
	.dc.b	$FF,$80,$63,$1E,$CF,$F3,$A6,$C6,$67,$78,$FF,$83,$A7,$2D,$8C,$BF
	.dc.b	$E4,$48,$CB,$63,$2F,$C1,$FB,$1F,$F7,$75,$7E,$15,$F8,$11,$EA,$E0
	.dc.b	$46,$A3,$0E,$B8,$3E,$AB,$96,$F0,$FF,$62,$8C,$5F,$78,$FE,$CB,$F6
	.dc.b	$3F,$26,$3F,$28,$DB,$F0,$2F,$13,$7B,$4A,$DB,$CA,$31,$5F,$EE,$1F
	.dc.b	$AE,$FE,$10,$7F,$07,$7F,$18,$66,$BF,$36,$E0,$FD,$E1,$4F,$F3,$77
	.dc.b	$47,$F9,$BF,$9B,$F8,$44,$64,$46,$67,$F8,$74,$66,$30,$FF,$0B,$38
	.dc.b	$77,$39,$3F,$91,$C7,$CD,$FB,$07,$95,$F8,$AF,$1B,$FC,$13,$18,$C7
	.dc.b	$FD,$01,$C3,$FE,$2C,$E4,$7E,$8A,$F8,$6F,$C3,$36,$BF,$14,$DC,$BF
	.dc.b	$0C,$D7,$FC,$4D,$F3,$3F,$C0,$70,$FE,$1C,$DA,$FF,$30,$FE,$17,$FC
	.dc.b	$92,$73,$E2,$FF,$87,$3B,$9E,$D5,$3F,$C4,$3F,$00,$F9,$47,$CA,$67
	.dc.b	$2F,$49,$FE,$6C,$8C,$BF,$E5,$F9,$E1,$3F,$E1,$86,$C6,$67,$7A,$7F
	.dc.b	$9B,$27,$2D,$8C,$BF,$EA,$28,$CB,$63,$2F,$EF,$FC,$4F,$F7,$AD,$7F
	.dc.b	$23,$6C,$7E,$E3,$95,$EB,$3D,$7F,$F1,$DD,$6B,$6F,$E6,$93,$9C,$3F
	.dc.b	$C6,$8D,$C3,$FC,$5A,$D7,$FC,$A8,$D7,$FD,$75,$87,$6C,$3F,$C6,$B3
	.dc.b	$99,$FF,$68,$36,$75,$5F,$E8,$5F,$03,$FD,$3B,$F0,$3F,$E4,$EF,$83
	.dc.b	$3F,$E5,$0F,$9E,$7C,$AE,$63,$0F,$F8,$E6,$33,$7C,$DB,$97,$F8,$D5
	.dc.b	$AF,$F3,$66,$BF,$C9,$DA,$B0,$46,$1B,$5F,$E4,$7F,$C7,$BF,$19,$5B
	.dc.b	$6E,$4F,$B3,$FC,$76,$3F,$CC,$E7,$31,$FD,$B2,$BF,$DE,$4D,$AE,$1E
	.dc.b	$A7,$F9,$72,$30,$47,$04,$BF,$91,$F1,$3F,$C0,$B8,$3F,$A5,$B6,$3F
	.dc.b	$A1,$5B,$EE,$FF,$CF,$33,$9C,$7D,$D3,$9A,$FA,$4D,$8F,$F2,$E3,$5F
	.dc.b	$EB,$AD,$7A,$D7,$F9,$A6,$31,$1D,$23,$07,$F9,$B7,$E1,$FF,$96,$57
	.dc.b	$F5,$56,$CA,$9F,$AA,$FF,$AE,$EB,$FB,$E5,$6A,$73,$1F,$E8,$09,$D5
	.dc.b	$7F,$A0,$1A,$FF,$A9,$DA,$FE,$51,$AF,$76,$CF,$49,$FF,$49,$46,$B6
	.dc.b	$FF,$80,$3B,$7F,$A8,$30,$FC,$5E,$7F,$DE,$2D,$F1,$79,$AF,$F5,$2C
	.dc.b	$60,$FC,$E1,$72,$FF,$43,$35,$FF,$03,$B5,$FE,$F1,$E8,$3E,$0B,$FD
	.dc.b	$5E,$D9,$FE,$1F,$2E,$97,$E7,$2F,$B2,$DE,$12,$FF,$C4,$98,$71,$C3
	.dc.b	$FB,$CE,$1C,$2F,$51,$FD,$8F,$C4,$7C,$5F,$FA,$F9,$B8,$7F,$8B,$5F
	.dc.b	$EE,$BA,$FF,$86,$1C,$D7,$1A,$E0,$71,$F4,$ED,$FA,$38,$9F,$56,$FA
	.dc.b	$1F,$F6,$23,$67,$5F,$5A,$F9,$8C,$D4,$9E,$55,$BC,$BF,$8E,$5E,$E7
	.dc.b	$35,$FE,$CF,$8C,$4F,$FB,$11,$AF,$FB,$AD,$AF,$F1,$86,$A2,$0E,$4C
	.dc.b	$7D,$3B,$EB,$73,$3F,$DC,$A3,$0E,$27,$73,$AF,$49,$8E,$2B,$42,$F8
	.dc.b	$BF,$DB,$9A,$F8,$EF,$83,$9B,$F0,$BE,$52,$9B,$29,$93,$31,$AF,$AE
	.dc.b	$7C,$77,$FA,$42,$79,$C6,$3D,$FD,$4E,$67,$DF,$39,$77,$F1,$FF,$E7
	.dc.b	$C6,$F2,$75,$8C,$47,$06,$BF,$32,$7F,$DD,$B3,$EC,$5F,$21,$F4,$33
	.dc.b	$F8,$71,$8A,$D5,$F7,$9C,$B7,$3E,$97,$C5,$B5,$C3,$0F,$A1,$C4,$DD
	.dc.b	$EE,$7C,$88,$C3,$63,$35,$F3,$A7,$2D,$E2,$FA,$F7,$FE,$9B,$6F,$07
	.dc.b	$F6,$26,$CF,$F1,$DB,$E6,$FF,$73,$8C,$DF,$E9,$31,$FC,$7E,$1C,$DB
	.dc.b	$87,$E5,$71,$B8,$7F,$55,$DD,$F5,$78,$EF,$FF,$64,$B6,$33,$3F,$EA
	.dc.b	$55,$F5,$D9,$99,$39,$78,$47,$3E,$BF,$E3,$B3,$FE,$AD,$6B,$FE,$69
	.dc.b	$6E,$7F,$9C,$8F,$E2,$9F,$03,$FE,$D9,$78,$78,$3C,$C8,$FF,$1A,$A4
	.dc.b	$EE,$3F,$CE,$9F,$5D,$FE,$BE,$7F,$C0,$8D,$7F,$5F,$9A,$31,$1F,$EB
	.dc.b	$06,$BF,$EA,$1F,$73,$FC,$12,$FF,$B1,$5A,$FF,$A0,$5B,$8F,$FA,$CA
	.dc.b	$7B,$3F,$ED,$5F,$AE,$7C,$77,$E1,$5B,$7F,$83,$6B,$FC,$0B,$89,$18
	.dc.b	$8F,$E1,$31,$89,$E4,$48,$E4,$FE,$E3,$1F,$F3,$8F,$7F,$DB,$CB,$F9
	.dc.b	$9D,$7A,$63,$F0,$37,$CE,$3F,$DC,$0F,$DE,$7F,$BB,$EF,$CE,$31,$1F
	.dc.b	$DE,$23,$04,$7F,$BF,$B5,$F9,$B7,$FE,$5F,$7F,$DD,$8D,$7F,$D5,$0D
	.dc.b	$CB,$FD,$B3,$5E,$17,$FD,$F9,$F5,$CF,$8E,$FD,$63,$6F,$F5,$4D,$7E
	.dc.b	$43,$5F,$8E,$D7,$FC,$96,$D7,$FC,$0C,$DA,$39,$BF,$F1,$C7,$CF,$FE
	.dc.b	$8D,$34,$F9,$31,$E1,$C3,$FD,$E3,$D7,$FE,$63,$C3,$FC,$A4,$D7,$F1
	.dc.b	$ED,$7F,$DF,$AD,$7F,$DD,$CD,$7F,$C9,$0D,$BF,$E7,$9E,$9F,$DD,$8C
	.dc.b	$7A,$1F,$8E,$E8,$FF,$28,$7C,$67,$EE,$63,$18,$72,$BF,$BF,$F5,$CF
	.dc.b	$9F,$FE,$2C,$D9,$FE,$28,$D7,$F2,$4D,$7F,$22,$D7,$FC,$EA,$D7,$FC
	.dc.b	$60,$DD,$8E,$73,$FF,$30,$F8,$BE,$52,$57,$F7,$B9,$F0,$6B,$FC,$07
	.dc.b	$A1,$F5,$BF,$D3,$2B,$FD,$44,$D7,$FD,$3C,$D7,$FD,$38,$D7,$FB,$84
	.dc.b	$72,$F1,$3F,$AD,$B6,$39,$C6,$27,$84,$7F,$9D,$FE,$B9,$F2,$3F,$C8
	.dc.b	$DB,$1F,$E2,$8C,$D1,$88,$FF,$11,$D7,$97,$1F,$F6,$3F,$3F,$EB,$BE
	.dc.b	$07,$F6,$35,$FE,$A9,$6C,$FB,$FE,$BB,$87,$27,$F8,$1F,$CD,$7C,$FB
	.dc.b	$D6,$FE,$E7,$18,$6C,$67,$1F,$EF,$33,$97,$FA,$77,$D2,$7F,$89,$5F
	.dc.b	$47,$D0,$7F,$26,$6F,$87,$FC,$0A,$75,$0C,$FF,$99,$2F,$AE,$7D,$57
	.dc.b	$F8,$09,$B1,$FE,$06,$6B,$FE,$0A,$6B,$FE,$1C,$6B,$FE,$22,$6B,$FE
	.dc.b	$25,$6B,$DE,$3F,$8D,$7C,$D7,$F0,$3D,$7F,$C5,$5C,$68,$C5,$7F,$AB
	.dc.b	$5A,$F1,$3E,$B7,$D7,$79,$31,$FE,$99,$48,$C5,$F4,$6D,$7F,$AD,$DF
	.dc.b	$EA,$71,$9C,$3F,$B2,$46,$AB,$94,$63,$C8,$FE,$D1,$F5,$7F,$23,$1E
	.dc.b	$EB,$E1,$32,$6B,$19,$97,$E2,$E7,$4D,$9C,$BF,$4E,$7F,$C9,$3F,$5C
	.dc.b	$F8,$AF,$F1,$9B,$67,$FC,$6E,$D7,$FC,$76,$D7,$FC,$9A,$D7,$9C,$72
	.dc.b	$23,$FE,$42,$F6,$DF,$5D,$F8,$78,$FF,$80,$1C,$2F,$9B,$FC,$41,$FE
	.dc.b	$0A,$AD,$F1,$FF,$4C,$FA,$B7,$C6,$52,$61,$FE,$79,$87,$D3,$62,$3D
	.dc.b	$C7,$FA,$E6,$38,$BB,$9E,$93,$FE,$15,$FA,$A7,$C7,$5F,$8E,$97,$FB
	.dc.b	$59,$CB,$67,$2F,$6E,$64,$FF,$51,$E1,$FC,$6B,$C2,$7C,$63,$EB,$BF
	.dc.b	$2A,$4E,$7F,$F3,$1B,$63,$FC,$CE,$D7,$FC,$D6,$FE,$7A,$FF,$31,$F0
	.dc.b	$FF,$E2,$DE,$A3,$4F,$F1,$9D,$77,$BF,$EE,$BE,$2B,$E4,$3F,$1B,$12
	.dc.b	$7F,$D4,$53,$96,$CE,$5F,$77,$4F,$E7,$3C,$BF,$E3,$73,$F2,$7E,$2E
	.dc.b	$57,$C7,$7F,$C0,$7F,$45,$CE,$3E,$36,$DF,$E2,$BE,$DF,$D4,$B7,$83
	.dc.b	$E3,$DF,$0C,$3D,$3F,$5C,$FA,$6F,$F4,$3B,$71,$FF,$45,$B5,$FF,$47
	.dc.b	$B5,$FF,$50,$B5,$FF,$53,$B5,$FF,$55,$35,$39,$E1,$FE,$09,$F1,$5F
	.dc.b	$34,$93,$FE,$DE,$37,$5F,$95,$F4,$A4,$7E,$4D,$FF,$11,$FD,$77,$FB
	.dc.b	$1E,$30,$43,$F2,$BB,$BA,$BF,$BE,$6F,$FE,$57,$85,$71,$AF,$93,$DD
	.dc.b	$F9,$72,$3F,$ED,$9F,$A6,$FD,$65,$FC,$C9,$CC,$F0,$35,$DB,$F9,$FE
	.dc.b	$5F,$F1,$E7,$1F,$F4,$F9,$FF,$65,$EF,$F9,$A9,$C9,$AF,$67,$F4,$C4
	.dc.b	$9F,$D2,$45,$7E,$EF,$CC,$7E,$AC,$C6,$5F,$7A,$B1,$3F,$AB,$9C,$F2
	.dc.b	$FE,$9B,$D6,$F1,$92,$7F,$D9,$38,$FE,$B2,$3C,$FF,$5D,$FB,$72,$4F
	.dc.b	$ED,$63,$F7,$5F,$5D,$EC,$24,$72,$D1,$1F,$03,$EA,$7F,$80,$12,$7F
	.dc.b	$7D,$15,$FC,$BC,$5D,$3F,$2F,$37,$2F,$DE,$CD,$A6,$5F,$C0,$AB,$F8
	.dc.b	$67,$D7,$7C,$E2,$4F,$8D,$FF,$36,$7C,$3F,$F9,$FB,$0F,$EA,$6D,$9E
	.dc.b	$6D,$BF,$E1,$B5,$FE,$AB,$8F,$EE,$D3,$FE,$90,$6B,$F6,$9A,$F4,$F4
	.dc.b	$BB,$91,$F9,$B1,$A7,$BC,$79,$3C,$1E,$C4,$8F,$F0,$B4,$F8,$33,$FC
	.dc.b	$B7,$91,$FE,$B0,$9F,$F4,$5B,$50,$E7,$FE,$7D,$E6,$FF,$B6,$E7,$53
	.dc.b	$96,$CE,$A3,$FD,$B9,$18,$37,$9F,$FB,$93,$D5,$F8,$44,$9E,$F8,$79
	.dc.b	$30,$C4,$7F,$35,$C7,$FC,$4B,$AF,$CE,$BF,$F4,$BE,$3F,$DD,$2B,$C5
	.dc.b	$E9,$FA,$1F,$F0,$BE,$7F,$D5,$63,$17,$EE,$6C,$FB,$A7,$4F,$C0,$9D
	.dc.b	$34,$D5,$7F,$8C,$9C,$1E,$02,$CF,$27,$C3,$1B,$5F,$1D,$FE,$B0,$D1
	.dc.b	$93,$E4,$7D,$13,$98,$C5,$7D,$28,$FF,$5B,$E1,$D5,$FF,$24,$FD,$47
	.dc.b	$F9,$13,$E6,$75,$AF,$C1,$8C,$5F,$82,$31,$AF,$FB,$62,$FF,$D8,$3E
	.dc.b	$EB,$E6,$BF,$C0,$D1,$D2,$75,$5C,$31,$EB,$87,$E1,$31,$EB,$7C,$B0
	.dc.b	$E3,$7B,$AF,$A4,$6F,$0F,$F7,$2B,$93,$37,$E6,$6F,$CF,$BE,$53,$8C
	.dc.b	$77,$8C,$38,$94,$62,$38,$CF,$05,$3A,$8F,$4C,$62,$79,$E3,$F0,$BD
	.dc.b	$A7,$80,$E3,$DE,$3C,$70,$F9,$47,$8F,$1C,$E3,$CB,$3F,$F4,$6E,$3C
	.dc.b	$B0,$CD,$7C,$8D,$DF,$5C,$F9,$3E,$D9,$74,$E4,$FC,$1F,$AB,$8D,$FE
	.dc.b	$FD,$EF,$2F,$A3,$5A,$7A,$46,$6F,$F3,$ED,$8F,$D1,$63,$E0,$C7,$BC
	.dc.b	$62,$75,$5F,$A1,$6B,$FB,$F6,$D7,$19,$FC,$EC,$6A,$FF,$33,$EA,$FA
	.dc.b	$BA,$7F,$CD,$7C,$DF,$E3,$AB,$FD,$D9,$D6,$74,$FF,$26,$6E,$3F,$C9
	.dc.b	$9B,$7D,$27,$8C,$F8,$E7,$35,$D6,$3F,$B8,$E1,$5F,$64,$FF,$A5,$9F
	.dc.b	$05,$7E,$2B,$1F,$A3,$39,$9E,$93,$9A,$E5,$1C,$0F,$CC,$D7,$09,$FE
	.dc.b	$5F,$97,$88,$F1,$FC,$A5,$8F,$E5,$AF,$5C,$BF,$7A,$DC,$8D,$24,$FF
	.dc.b	$3A,$6E,$1F,$D0,$DE,$1F,$07,$F0,$0F,$83,$F5,$D6,$BF,$9D,$BD,$48
	.dc.b	$FF,$44,$7C,$1E,$67,$C8,$3E,$4F,$CF,$59,$FE,$87,$17,$0F,$E5,$D3
	.dc.b	$77,$FB,$73,$6B,$FA,$EC,$DA,$FE,$BA,$FB,$49,$1F,$D5,$1F,$81,$F3
	.dc.b	$3F,$60,$FB,$3F,$5D,$67,$FA,$9B,$F1,$48,$FF,$5C,$7E,$47,$D1,$FE
	.dc.b	$03,$EB,$BF,$02,$D9,$FE,$B6,$DC,$3F,$A3,$37,$2F,$C9,$BC,$27,$FB
	.dc.b	$0F,$A5,$6B,$FA,$63,$F5,$49,$1F,$DB,$1F,$B1,$F0,$1F,$90,$3F,$2F
	.dc.b	$C6,$F4,$5E,$56,$6F,$FB,$6D,$5C,$3F,$C8,$B5,$A9,$C9,$FD,$CB,$E4
	.dc.b	$FF,$36,$74,$FA,$7F,$E3,$71,$92,$57,$F6,$19,$D7,$7F,$09,$FD,$2F
	.dc.b	$3F,$FB,$87,$E4,$BF,$53,$E5,$1F,$D6,$FC,$A3,$E6,$1E,$E6,$A7,$31
	.dc.b	$8C,$3F,$97,$D6,$EB,$F4,$CD,$8C,$CE,$F3,$FE,$7B,$39,$6C,$65,$FE
	.dc.b	$E5,$19,$6C,$65,$FF,$13,$FD,$77,$EF,$0F,$9B,$EF,$7F,$B2,$EF,$FC
	.dc.b	$2F,$19,$69,$FC,$13,$89,$3B,$AF,$F0,$35,$7D,$6F,$61,$F2,$7F,$C3
	.dc.b	$3E,$BB,$FD,$CD,$8F,$FA,$9A,$31,$C9,$B6,$C6,$4F,$9B,$3C,$30,$FF
	.dc.b	$63,$6F,$DF,$A3,$F9,$CF,$0A,$31,$7F,$E2,$8D,$FE,$87,$C5,$7D,$4F
	.dc.b	$4B,$E7,$BF,$33,$84,$D2,$57,$CE,$6E,$5F,$31,$AF,$F9,$FB,$CC,$FF
	.dc.b	$0A,$67,$E2,$7F,$B4,$57,$E1,$1C,$BF,$E3,$AC,$FF,$B1,$F9,$5F,$CB
	.dc.b	$A7,$37,$FE,$60,$78,$BF,$EE,$37,$33,$C2,$BA,$7C,$0F,$13,$97,$8D
	.dc.b	$63,$24,$BE,$51,$97,$FD,$F7,$F1,$0F,$AE,$7D,$5F,$EB,$3E,$AB,$FB
	.dc.b	$45,$6A,$73,$87,$DF,$9C,$EB,$FD,$A3,$E4,$1F,$5D,$FE,$29,$AF,$F0
	.dc.b	$3C,$63,$D9,$FE,$7C,$D8,$CC,$EF,$1F,$F0,$AC,$E5,$B1,$97,$FC,$95
	.dc.b	$19,$6C,$65,$F9,$9F,$5D,$FE,$F3,$AF,$D9,$BF,$E3,$C6,$ED,$FD,$72
	.dc.b	$31,$2C,$7F,$13,$1C,$63,$15,$D3,$6E,$17,$C5,$FE,$CB,$F5,$DF,$12
	.dc.b	$B6,$7F,$A6,$1C,$61,$F9,$45,$E1,$F0,$DE,$63,$FC,$22,$76,$FE,$15
	.dc.b	$DB,$98,$FE,$17,$F5,$D2,$3F,$B9,$F4,$7F,$A1,$39,$71,$97,$8A,$31
	.dc.b	$3F,$C2,$A3,$F1,$67,$F0,$22,$9F,$A5,$28,$8E,$2F,$F0,$C3,$2E,$1E
	.dc.b	$53,$FD,$E7,$C0,$71,$7F,$28,$9F,$87,$CA,$FA,$E7,$93,$F8,$DF,$17
	.dc.b	$FC,$1B,$18,$CF,$FC,$1B,$CA,$FA,$E7,$C3,$7E,$21,$B3,$F8,$C6,$E5
	.dc.b	$F8,$86,$BF,$E2,$EF,$98,$FE,$0D,$C9,$FF,$00,$5B,$5F,$EA,$AF,$80
	.dc.b	$7D,$77,$F2,$9A,$D7,$8B,$FE,$20,$E9,$7B,$74,$FF,$12,$DB,$E5,$1F
	.dc.b	$29,$9C,$BF,$C4,$B3,$98,$FF,$0E,$36,$3F,$CD,$2D,$D3,$FC,$E6,$D7
	.dc.b	$FD,$4C,$D7,$CD,$5D,$A7,$8B,$63,$8B,$5F,$F5,$77,$1F,$F1,$AF,$8A
	.dc.b	$FF,$1C,$F8,$87,$FB,$3F,$AC,$F5,$F3,$18,$7E,$D4,$67,$1F,$F1,$B3
	.dc.b	$70,$FF,$18,$B5,$FF,$2D,$B5,$FF,$5C,$35,$FF,$6B,$37,$A4,$E7,$1F
	.dc.b	$F4,$F7,$C2,$7F,$93,$BE,$10,$FF,$94,$3E,$79,$F2,$99,$8C,$3F,$E3
	.dc.b	$B8,$CE,$5F,$E9,$4B,$46,$32,$FF,$4F,$C6,$2B,$81,$2B,$85,$6E,$BE
	.dc.b	$4B,$5F,$E6,$DF,$5D,$F8,$85,$68,$C6,$6F,$AE,$FC,$56,$1C,$39,$7F
	.dc.b	$C2,$B1,$FD,$B7,$0F,$F7,$C3,$6B,$87,$AB,$FA,$98,$C1,$1C,$12,$FD
	.dc.b	$DF,$15,$FD,$13,$A3,$FA,$FB,$63,$FA,$65,$BE,$BB,$FC,$F3,$39,$C7
	.dc.b	$D3,$39,$AF,$90,$D8,$F8,$ED,$7F,$16,$D7,$AD,$7F,$9A,$E3,$11,$D2
	.dc.b	$30,$7F,$9B,$FE,$1B,$F9,$6E,$1F,$DD,$1B,$7F,$EF,$DF,$AE,$FF,$5E
	.dc.b	$3F,$E2,$88,$CD,$B0,$56,$A3,$FD,$02,$D7,$FD,$56,$D7,$A5,$7F,$13
	.dc.b	$8C,$47,$36,$CF,$49,$D1,$E7,$E0,$FF,$4F,$DF,$E2,$70,$FF,$05,$CB
	.dc.b	$E3,$33,$5F,$EA,$48,$C1,$F9,$12,$E3,$F9,$06,$BF,$B4,$6B,$FE,$01
	.dc.b	$F4,$5F,$05,$FE,$AE,$6C,$FF,$0A,$FA,$E7,$CB,$66,$A4,$FF,$93,$AB
	.dc.b	$77,$FD,$43,$B5,$87,$0A,$D5,$FF,$AE,$BC,$57,$C6,$53,$1E,$A5,$AF
	.dc.b	$F2,$CC,$E9,$B3,$A6,$9F,$E3,$FF,$AB,$7D,$27,$F9,$C9,$B1,$FE,$43
	.dc.b	$FA,$C7,$D7,$66,$A4,$F1,$7F,$CF,$33,$A8,$C1,$C6,$7F,$A6,$46,$75
	.dc.b	$FF,$60,$B5,$FF,$75,$B5,$10,$BC,$78,$FA,$77,$D6,$E6,$7F,$C5,$B1
	.dc.b	$87,$13,$B9,$D7,$A4,$C7,$15,$A1,$7C,$5F,$EC,$ED,$7C,$26,$AB,$9C
	.dc.b	$F2,$8F,$35,$F8,$DE,$85,$0D,$79,$9E,$9F,$AE,$7C,$77,$FA,$2E,$39
	.dc.b	$46,$3D,$FE,$4D,$8C,$CE,$DE,$5F,$1F,$FE,$21,$78,$B5,$CF,$57,$28
	.dc.b	$C4,$7C,$DA,$FF,$74,$CF,$AD,$7C,$87,$D1,$D3,$E2,$46,$2B,$57,$DE
	.dc.b	$72,$DD,$3A,$5F,$16,$D7,$0C,$3E,$8F,$63,$F6,$27,$D6,$7C,$78,$C3
	.dc.b	$63,$35,$F3,$67,$2D,$FB,$17,$B1,$F8,$7C,$3E,$F4,$FF,$11,$9C,$BF
	.dc.b	$E9,$D6,$BF,$E3,$EA,$E6,$D8,$CC,$F2,$94,$63,$87,$F2,$B1,$8B,$F0
	.dc.b	$46,$30,$F3,$E1,$FE,$28,$E0,$FC,$66,$3F,$E9,$C6,$D7,$FA,$25,$B3
	.dc.b	$FD,$51,$FE,$CE,$FF,$B4,$9C,$F9,$1F,$EC,$FD,$3B,$CE,$DB,$1F,$ED
	.dc.b	$3F,$02,$F7,$3E,$9A,$92,$FF,$DB,$B1,$FE,$20,$6C,$65,$FF,$1F,$C7
	.dc.b	$17,$7E,$97,$6F,$AF,$FF,$71,$47,$F7,$76,$BF,$DD,$9A,$FF,$86,$9A
	.dc.b	$FF,$A9,$FD,$D7,$91,$E8,$6C,$7F,$BC,$1A,$FF,$94,$27,$AE,$DF,$D4
	.dc.b	$7C,$8F,$D6,$69,$FA,$C6,$F5,$7F,$67,$BF,$F0,$53,$5F,$F0,$4B,$5F
	.dc.b	$F1,$6B,$5F,$F5,$D3,$5F,$F0,$2B,$65,$F8,$7B,$E7,$7E,$7B,$6B,$F8
	.dc.b	$BE,$1F,$ED,$88,$C4,$7F,$9A,$2F,$C5,$3F,$ED,$CE,$57,$D8,$FA,$9B
	.dc.b	$5F,$DE,$A3,$04,$70,$43,$FB,$EB,$F0,$B1,$79,$9F,$82,$6C,$F9,$39
	.dc.b	$7F,$DA,$D1,$88,$FF,$3D,$DF,$81,$FF,$3E,$EF,$FE,$21,$7F,$C2,$BE
	.dc.b	$6F,$CE,$8C,$47,$F8,$C9,$AF,$F9,$59,$AF,$FB,$8D,$AF,$F8,$B9,$B5
	.dc.b	$F8,$09,$FB,$3F,$87,$FE,$36,$BF,$D1,$AD,$CF,$FE,$00,$B7,$1E,$95
	.dc.b	$F9,$EE,$EF,$F2,$8D,$FF,$91,$DA,$FF,$91,$9A,$FF,$9A,$9A,$FF,$BD
	.dc.b	$DA,$FF,$90,$9B,$87,$F4,$2F,$53,$F1,$B1,$8F,$4B,$F2,$38,$7F,$BF
	.dc.b	$DB,$7F,$E4,$A6,$DF,$18,$FA,$07,$F9,$3B,$D4,$78,$3F,$98,$6C,$7F
	.dc.b	$97,$DA,$FF,$97,$9A,$FF,$A0,$9A,$FF,$64,$94,$62,$3B,$4F,$53,$F4
	.dc.b	$BF,$73,$F7,$75,$FE,$47,$68,$C6,$3C,$DB,$7C,$8F,$4C,$BE,$C1,$F4
	.dc.b	$1F,$B0,$6C,$FF,$9D,$DA,$FF,$9D,$9A,$FF,$A6,$9A,$FA,$5A,$FF,$9C
	.dc.b	$9B,$3D,$E7,$B7,$D2,$7F,$1C,$93,$4C,$47,$F2,$0F,$79,$F3,$94,$98
	.dc.b	$7A,$A3,$FD,$AE,$FB,$3F,$22,$FC,$99,$38,$BF,$85,$BF,$96,$FA,$17
	.dc.b	$D5,$FF,$2C,$7F,$C9,$15,$F1,$1A,$FF,$B9,$1D,$BF,$80,$3F,$98,$7B
	.dc.b	$BF,$DF,$FC,$3F,$F5,$1C,$7F,$80,$DA,$FF,$81,$DA,$FF,$86,$DA,$FF
	.dc.b	$88,$5A,$FF,$89,$1A,$F6,$3F,$D5,$1F,$29,$F6,$B7,$34,$FB,$B1,$B3
	.dc.b	$FD,$F7,$F5,$DF,$AD,$7F,$D4,$6F,$F4,$37,$3E,$27,$F4,$C8,$D5,$1A
	.dc.b	$6F,$93,$F1,$AD,$E4,$FC,$B3,$C3,$EB,$07,$F5,$62,$E1,$F8,$E6,$AA
	.dc.b	$FC,$BC,$7F,$6F,$8F,$F5,$BB,$A8,$CB,$CC,$E9,$E5,$FF,$64,$9F,$F0
	.dc.b	$1F,$D1,$5F,$C2,$DF,$85,$D3,$54,$9A,$66,$BE,$D5,$70,$7E,$E4,$ED
	.dc.b	$FD,$2F,$CA,$FF,$76,$7D,$19,$FE,$D6,$31,$97,$EC,$1B,$1F,$E5,$BB
	.dc.b	$E9,$5C,$0F,$F6,$9F,$D4,$FC,$96,$FC,$4F,$F7,$C8,$37,$79,$3F,$B6
	.dc.b	$9C,$B6,$72,$FF,$87,$7E,$B7,$FC,$70,$DE,$EF,$F0,$4F,$8A,$9F,$E2
	.dc.b	$3A,$7F,$87,$DD,$B7,$D5,$DC,$6B,$4F,$F1,$0B,$70,$FF,$5F,$3C,$23
	.dc.b	$FC,$F1,$F5,$EF,$9F,$FF,$30,$36,$BF,$CC,$8D,$7F,$CD,$0D,$7F,$CF
	.dc.b	$8D,$79,$C7,$22,$3F,$E6,$FE,$57,$CE,$7F,$8C,$32,$FE,$3E,$DC,$3F
	.dc.b	$94,$36,$BF,$D1,$B8,$78,$6B,$7E,$5F,$F8,$AA,$BF,$54,$4F,$53,$11
	.dc.b	$2F,$F9,$2C,$E5,$FE,$7F,$AB,$F7,$AF,$BF,$0C,$FC,$BE,$AF,$53,$E8
	.dc.b	$CE,$AB,$84,$66,$BA,$BA,$3F,$39,$F6,$0F,$A2,$FF,$45,$36,$7F,$D1
	.dc.b	$CD,$7F,$D2,$4D,$7F,$D4,$CD,$7F,$D5,$8D,$7F,$D5,$ED,$4E,$70,$F9
	.dc.b	$BF,$13,$F1,$49,$3F,$C7,$E7,$C1,$AF,$D0,$EB,$F9,$5F,$82,$CA,$F8
	.dc.b	$F1,$F3,$4C,$3F,$A1,$C3,$EC,$CE,$63,$E6,$FD,$CF,$D3,$F1,$A4,$60
	.dc.b	$87,$1D,$FF,$C2,$B8,$FE,$02,$7E,$D5,$7D,$BF,$83,$FF,$23,$D7,$FB
	.dc.b	$6A,$FF,$10,$D7,$FD,$95,$8F,$E2,$A7,$7F,$39,$F8,$32,$4F,$E0,$22
	.dc.b	$BE,$1F,$9E,$FC,$41,$8C,$BF,$23,$58,$9F,$C3,$CE,$79,$BF,$27,$F4
	.dc.b	$9E,$42,$4F,$91,$F3,$FD,$37,$E4,$C9,$1F,$90,$8F,$CB,$7D,$27,$B0
	.dc.b	$91,$EC,$7D,$FE,$EF,$E6,$C9,$1F,$98,$8A,$FC,$98,$B9,$7E,$A2,$6E
	.dc.b	$1F,$A8,$9B,$5F,$9B,$D5,$5F,$A7,$F8,$9F,$9A,$49,$F9,$AF,$D0,$FA
	.dc.b	$0F,$D6,$12,$3F,$53,$1F,$E9,$8F,$CE,$3F,$D6,$F1,$F4,$9A,$FF,$C0
	.dc.b	$1E,$04,$62,$3F,$C4,$4D,$7F,$5C,$D8,$DE,$5F,$E8,$68,$C1,$3E,$67
	.dc.b	$C0,$4A,$FF,$0F,$C7,$8B,$4F,$F6,$07,$D6,$BE,$67,$FD,$A8,$DA,$FF
	.dc.b	$69,$35,$10,$76,$23,$51,$98,$C4,$F4,$24,$74,$7D,$1E,$67,$C5,$24
	.dc.b	$7F,$0E,$AF,$05,$FF,$8F,$30,$FE,$C9,$97,$C4,$BD,$DF,$FB,$76,$F7
	.dc.b	$3F,$ED,$87,$FA,$D4,$3E,$05,$F2,$19,$F3,$BD,$51,$89,$F5,$D7,$08
	.dc.b	$C6,$9E,$DF,$8D,$DE,$1C,$6F,$E1,$D6,$AF,$84,$E6,$7C,$0F,$8E,$B8
	.dc.b	$C7,$6C,$2E,$1F,$25,$B7,$CA,$7C,$75,$C1,$FE,$83,$EC,$FF,$8B,$B1
	.dc.b	$F4,$46,$3D,$4F,$1E,$5F,$57,$97,$FC,$B7,$18,$97,$C4,$7F,$9B,$3E
	.dc.b	$63,$A5,$6A,$7C,$E7,$DE,$71,$87,$66,$E1,$F8,$2B,$E5,$5D,$4F,$11
	.dc.b	$F8,$53,$96,$3F,$E3,$CA,$ED,$19,$6B,$E7,$F4,$DF,$1F,$C6,$3C,$51
	.dc.b	$87,$12,$8C,$47,$36,$BC,$E7,$C9,$3B,$7E,$15,$BD,$87,$D7,$B8,$B7
	.dc.b	$8F,$7D,$3E,$03,$74,$E9,$9F,$C6,$C7,$96,$1C,$63,$E4,$9F,$DD,$AB
	.dc.b	$83,$FA,$6F,$C7,$5D,$1B,$C7,$F1,$B1,$8C,$BE,$B7,$22,$3C,$91,$87
	.dc.b	$57,$F5,$A7,$57,$CF,$3F,$DA,$C6,$2F,$81,$9A,$F1,$55,$9D,$47,$3B
	.dc.b	$EB,$87,$92,$F5,$18,$C3,$F8,$9E,$3E,$FE,$7F,$DA,$36,$BC,$3F,$1C
	.dc.b	$EB,$C6,$5F,$C6,$2A,$F5,$FF,$1E,$6F,$57,$DB,$C3,$FC,$1B,$87,$17
	.dc.b	$F8,$BC,$F2,$D5,$39,$C3,$F1,$D1,$8C,$7F,$94,$36,$FF,$14,$D9,$DD
	.dc.b	$71,$7F,$8B,$35,$F4,$D7,$28,$C4,$7F,$30,$E4,$F3,$9E,$8E,$0B,$87
	.dc.b	$F2,$C7,$F9,$C6,$7F,$CE,$9B,$9F,$E3,$A7,$FC,$74,$DB,$FE,$85,$3A
	.dc.b	$24,$E9,$E3,$EE,$7C,$63,$E3,$FC,$CD,$9F,$E7,$6F,$62,$3D,$9F,$0F
	.dc.b	$BB,$F3,$8F,$9F,$EC,$D9,$FE,$87,$17,$0F,$F0,$B6,$2A,$C7,$07,$F3
	.dc.b	$1A,$B7,$FD,$52,$6D,$FF,$54,$7D,$E4,$8F,$7B,$F0,$BD,$8F,$BE,$7E
	.dc.b	$03,$E2,$AC,$FF,$54,$7E,$39,$1F,$8E,$FC,$AF,$B3,$FC,$3F,$AF,$F8
	.dc.b	$86,$CF,$F5,$C6,$E1,$EA,$6E,$5F,$D2,$5E,$13,$FD,$8B,$E0,$ED,$7C
	.dc.b	$07,$EB,$92,$3E,$BB,$F6,$BE,$8B,$F2,$FF,$67,$F9,$86,$CF,$F6,$C6
	.dc.b	$D7,$CA,$6E,$1F,$D6,$7E,$7F,$F3,$FF,$21,$FA,$06,$DF,$F7,$46,$DF
	.dc.b	$D2,$6D,$7F,$61,$FA,$FF,$D3,$9F,$AC,$F9,$37,$CA,$35,$CE,$3F,$DD
	.dc.b	$AF,$F8,$A5,$6A,$7F,$C1,$8D,$9F,$E5,$CD,$C7,$F9,$C3,$5F,$EF,$CD
	.dc.b	$7F,$C0,$2D,$4D,$31,$D7,$FE,$85,$F3,$3F,$7B,$F6,$7F,$B9,$F7,$51
	.dc.b	$8C,$3F,$D1,$71,$87,$FC,$2B,$3B,$9D,$61,$FE,$69,$FC,$C1,$E5,$7F
	.dc.b	$0A,$6D,$FB,$CF,$F0,$AF,$6A,$1F,$29,$BA,$95,$BB,$FF,$4C,$57,$27
	.dc.b	$B5,$69,$F0,$7C,$57,$73,$E8,$70,$7F,$BF,$AF,$96,$3D,$E7,$3E,$87
	.dc.b	$CE,$AD,$5F,$CC,$6B,$F3,$1A,$FC,$C6,$BF,$31,$BB,$FD,$E9,$E8,$7F
	.dc.b	$0E,$CB,$FC,$37,$8F,$89,$7C,$EF,$C5,$F1,$BE,$33,$67,$4D,$C7,$E6
	.dc.b	$3F,$CD,$3C,$DF,$F1,$EC,$FF,$B0,$3C,$DC,$56,$5B,$19,$3F,$D8,$B1
	.dc.b	$97,$FD,$ED,$E1,$7F,$3E,$F8,$83,$FD,$0A,$9F,$81,$BA,$B5,$3F,$D1
	.dc.b	$B0,$7C,$63,$EC,$78,$5E,$F2,$FB,$15,$A9,$CE,$1C,$6F,$85,$7D,$DC
	.dc.b	$7B,$CE,$6B,$73,$F6,$3E,$37,$DF,$87,$FA,$D5,$BD,$9F,$DE,$F9,$51
	.dc.b	$FE,$14,$AE,$15,$FD,$6A,$37,$1F,$90,$6C,$7F,$B8,$DE,$0F,$53,$FB
	.dc.b	$0F,$D6,$3E,$06,$FC,$06,$1B,$7F,$8C,$9D,$36,$74,$D3,$F2,$3D,$1F
	.dc.b	$D8,$F2,$F4,$36,$3F,$6F,$83,$E2,$7F,$B8,$CE,$63,$F7,$26,$67,$F8
	.dc.b	$1B,$4F,$E0,$65,$3F,$81,$94,$EB,$3F,$C3,$1C,$1D,$1C,$1F,$D9,$3E
	.dc.b	$A9,$F4,$5F,$E0,$25,$F2,$76,$30,$FF,$81,$BE,$4B,$F8,$25,$FF,$B6
	.dc.b	$1A,$FF,$A9,$DA,$FF,$00,$6A,$DD,$0F,$AD,$E8,$75,$F9,$FB,$9F,$EE
	.dc.b	$BE,$57,$F5,$2D,$EB,$FF,$0B,$FC,$B7,$E6,$A3,$07,$F0,$07,$19,$7F
	.dc.b	$BB,$1A,$FF,$AE,$DA,$FF,$14,$6B,$E5,$6E,$46,$93,$AB,$FC,$17,$F2
	.dc.b	$FF,$5E,$4D,$3F,$1E,$E3,$D9,$FF,$16,$C7,$E7,$2F,$FD,$98,$D7,$F9
	.dc.b	$2B,$5F,$8C,$D9,$EB,$1E,$02,$61,$FE,$73,$8C,$B7,$DD,$79,$1F,$E3
	.dc.b	$C6,$C7,$F2,$4F,$99,$FC,$57,$37,$EA,$E7,$38,$23,$F5,$D3,$F8,$8C
	.dc.b	$0D,$27,$2F,$F8,$B1,$B8,$78,$DB,$7F,$E8,$FA,$7C,$D7,$DA,$8E,$D5
	.dc.b	$FD,$12,$31,$87,$EF,$7E,$73,$FC,$D6,$FE,$FE,$33,$3F,$D2,$A7,$FC
	.dc.b	$DF,$5F,$EF,$56,$B9,$CB,$FC,$7F,$18,$8F,$C8,$37,$03,$49,$3F,$E4
	.dc.b	$6F,$27,$EC,$9F,$6B,$CD,$7C,$47,$F9,$9A,$30,$FF,$93,$23,$38,$FD
	.dc.b	$36,$E3,$F0,$5A,$FF,$BD,$71,$FF,$7A,$D7,$12,$57,$17,$FC,$AF,$E1
	.dc.b	$7E,$D2,$3F,$AD,$B6,$BE,$49,$FA,$BF,$9D,$7B,$5F,$E7,$A9,$D6,$DF
	.dc.b	$C9,$E7,$37,$F8,$76,$DF,$D5,$6B,$E5,$6B,$FD,$E1,$AF,$7B,$FF,$35
	.dc.b	$C6,$31,$37,$E0,$BF,$1B,$7F,$E0,$06,$DF,$F0,$18,$7B,$BF,$E9,$09
	.dc.b	$CD,$FF,$AF,$E3,$FC,$EF,$3A,$3F,$3C,$5A,$FC,$63,$5F,$99,$F4,$AF
	.dc.b	$8A,$FF,$4D,$B6,$7F,$C0,$6B,$E7,$B3,$32,$72,$F0,$8C,$57,$CE,$C1
	.dc.b	$F4,$9F,$E1,$8B,$FF,$06,$B6,$74,$BC,$4F,$F5,$5B,$9D,$BF,$AC,$36
	.dc.b	$BF,$C4,$6D,$AF,$E4,$8F,$18,$EB,$C5,$FE,$AD,$FA,$17,$B9,$4F,$F5
	.dc.b	$87,$C2,$FC,$CE,$54,$EA,$FF,$C5,$51,$8A,$E1,$1F,$EB,$AE,$AB,$97
	.dc.b	$E2,$AB,$FD,$73,$1F,$EA,$D2,$4F,$FA,$DB,$0F,$E9,$34,$27,$15,$E9
	.dc.b	$FE,$B5,$19,$3F,$B6,$B8,$71,$1B,$BD,$46,$9A,$A3,$4F,$3E,$6F,$F7
	.dc.b	$D5,$73,$8F,$09,$FC,$32,$B9,$CE,$9E,$75,$D2,$7C,$25,$3B,$15,$D7
	.dc.b	$D1,$3E,$33,$FC,$F3,$3E,$48,$C7,$B1,$A9,$CC,$FA,$A7,$2E,$FE,$7F
	.dc.b	$F9,$53,$78,$BA,$C6,$23,$83,$5F,$91,$28,$E3,$9F,$C9,$87,$CF,$FC
	.dc.b	$D8,$C3,$73,$E3,$5C,$1B,$87,$CA,$C1,$87,$27,$0D,$7E,$51,$F7,$5F
	.dc.b	$14,$7D,$2D,$FF,$DF,$BE,$3F,$CB,$6D,$7E,$16,$31,$B9,$D7,$13,$E2
	.dc.b	$BF,$09,$83,$8B,$F1,$58,$FE,$62,$73,$18,$AE,$18,$FE,$6B,$1E,$8F
	.dc.b	$E7,$23,$2F,$E6,$5B,$87,$58,$61,$D5,$E5,$8F,$67,$FC,$25,$C1,$F9
	.dc.b	$5A,$FF,$26,$37,$1F,$F4,$BB,$67,$7F,$39,$C7,$A3,$FD,$7B,$B5,$AD
	.dc.b	$46,$2B,$FD,$23,$0A,$E1,$F1,$BF,$CD,$A7,$FD,$2A,$D7,$F8,$73,$71
	.dc.b	$FF,$2F,$3F,$EF,$F8,$DF,$7F,$22,$3C,$BE,$27,$C9,$F4,$4F,$95,$FD
	.dc.b	$06,$68,$C5,$7F,$81,$DA,$FF,$75,$6B,$FD,$57,$E0,$3F,$D6,$AF,$FA
	.dc.b	$A9,$AF,$F2,$36,$E5,$FE,$73,$9E,$6F,$F9,$D7,$A5,$F2,$9E,$26,$DF
	.dc.b	$F8,$5D,$AF,$F8,$15,$AF,$F6,$56,$BF,$D2,$63,$91,$1F,$EB,$5F,$0F
	.dc.b	$F4,$FC,$53,$4F,$AD,$5D,$AF,$FD,$11,$C8,$F9,$EF,$8A,$DC,$3F,$BC
	.dc.b	$46,$08,$E0,$87,$F7,$B7,$E6,$E8,$F8,$CF,$C2,$36,$7F,$DA,$2D,$7F
	.dc.b	$A5,$57,$77,$FA,$67,$83,$FE,$8F,$7F,$C7,$4D,$7F,$C5,$0D,$7F,$C1
	.dc.b	$0D,$7F,$DD,$6D,$7F,$CD,$8D,$9F,$F7,$6F,$B5,$F8,$28,$C7,$97,$F8
	.dc.b	$DE,$8F,$F1,$57,$27,$E6,$E3,$18,$72,$BF,$CF,$78,$1F,$D9,$E3,$FC
	.dc.b	$9C,$D7,$FC,$78,$D7,$FC,$38,$D7,$FC,$79,$E3,$7E,$4A,$31,$1F,$E8
	.dc.b	$16,$E5,$FE,$44,$9F,$CB,$F9,$9F,$8A,$CB,$FA,$74,$FF,$94,$39,$BF
	.dc.b	$61,$39,$C3,$95,$FE,$CF,$E8,$BF,$94,$47,$F9,$8D,$AF,$F9,$45,$AF
	.dc.b	$F8,$C5,$AF,$66,$BF,$E9,$56,$D9,$A6,$3C,$FF,$B5,$C5,$FE,$D3,$6E
	.dc.b	$5F,$E7,$AE,$2F,$27,$87,$FC,$1F,$D2,$7C,$8F,$ED,$1B,$9F,$F9,$E5
	.dc.b	$AF,$FA,$5E,$34,$47,$4D,$7F,$D3,$9F,$1B,$D4,$8F,$FA,$86,$3F,$8B
	.dc.b	$7A,$DF,$EB,$A8,$FF,$55,$7B,$38,$33,$F5,$9F,$C6,$FD,$AF,$C4,$13
	.dc.b	$97,$F8,$4B,$F8,$BF,$94,$FF,$1B,$DF,$F8,$07,$83,$FD,$E5,$0F,$67
	.dc.b	$F9,$23,$67,$5C,$1F,$6A,$38,$C7,$F9,$60,$FF,$2B,$FC,$57,$F7,$3E
	.dc.b	$A3,$49,$1F,$DE,$63,$2D,$8C,$BF,$E0,$88,$CB,$FE,$8F,$F4,$FE,$9B
	.dc.b	$FE,$CF,$6F,$37,$F8,$D7,$93,$FA,$14,$F8,$8F,$39,$FE,$D5,$F8,$4F
	.dc.b	$D1,$C9,$FE,$65,$1F,$E4,$36,$E9,$FD,$3E,$75,$87,$F4,$59,$E2,$6B
	.dc.b	$3F,$CD,$B6,$FB,$4F,$F4,$BF,$8A,$14,$91,$E4,$BC,$F9,$F8,$8F,$F2
	.dc.b	$A4,$F0,$74,$D7,$31,$EE,$1F,$2F,$F0,$3E,$8D,$F3,$D4,$D4,$4A,$F8
	.dc.b	$62,$76,$E1,$AF,$DB,$AE,$AF,$9F,$E1,$BF,$B0,$37,$CE,$FE,$D1,$18
	.dc.b	$8F,$C1,$57,$2D,$BE,$57,$C5,$7C,$5D,$0B,$E5,$3F,$B8,$36,$FE,$23
	.dc.b	$B2,$46,$8D,$F9,$FF,$DF,$FC,$87,$D1,$FF,$87,$F1,$FF,$03,$46,$3C
	.dc.b	$1D,$4F,$F0,$89,$CB,$BE,$4F,$F0,$BB,$72,$FF,$06,$BB,$8C,$D7,$2F
	.dc.b	$63,$F1,$BF,$46,$F9,$6F,$F2,$83,$70,$FF,$2A,$35,$FF,$30,$B5,$FF
	.dc.b	$32,$B5,$FF,$36,$35,$FF,$39,$35,$39,$D9,$F1,$FF,$E2,$89,$FF,$4D
	.dc.b	$37,$3F,$CC,$37,$2E,$B8,$7F,$87,$63,$F4,$E4,$E9,$C7,$AF,$96,$61
	.dc.b	$F9,$59,$DC,$FC,$FF,$93,$FB,$D8,$F6,$E5,$E8,$64,$97,$C5,$CB,$FE
	.dc.b	$98,$9E,$39,$7F,$A2,$FE,$85,$F4,$7F,$E7,$96,$D7,$F9,$F5,$AF,$FA
	.dc.b	$29,$AF,$FA,$39,$AF,$38,$E4,$47,$FD,$23,$AF,$C8,$FA,$3F,$84,$48
	.dc.b	$FD,$1D,$76,$CB,$E4,$F5,$FC,$7F,$A3,$95,$F0,$E3,$E5,$18,$7E,$8C
	.dc.b	$61,$F9,$1F,$47,$F3,$FC,$69,$18,$21,$F6,$1C,$BF,$6A,$3E,$B7,$C8
	.dc.b	$FF,$8F,$B8,$4D,$24,$FB,$21,$F4,$7F,$74,$91,$F6,$66,$BF,$86,$FA
	.dc.b	$3F,$C1,$18,$CF,$F0,$F5,$89,$FC,$04,$E7,$83,$FB,$BE,$0F,$A3,$F2
	.dc.b	$12,$BC,$4F,$9F,$D5,$FC,$71,$23,$F1,$51,$F9,$1F,$8D,$F6,$12,$3D
	.dc.b	$4F,$BF,$DE,$FC,$B1,$23,$F2,$51,$5F,$CF,$C5,$CF,$F3,$F3,$71,$F3
	.dc.b	$CD,$A7,$1F,$EA,$2B,$F4,$5F,$21,$F3,$49,$3F,$2D,$FA,$1F,$37,$FA
	.dc.b	$72,$47,$E8,$E3,$FD,$0D,$F3,$FF,$E9,$38,$FF,$6F,$DA,$31,$1F,$EE
	.dc.b	$78,$C6,$3F,$E7,$89,$CC,$FF,$88,$9A,$F9,$9B,$3C,$23,$F3,$31,$81
	.dc.b	$F2,$3E,$22,$47,$F4,$B8,$F2,$67,$FC,$8B,$E8,$DF,$21,$FE,$BD,$6D
	.dc.b	$7F,$BD,$3C,$44,$62,$3F,$DF,$31,$89,$E0,$48,$E1,$3B,$47,$FB,$01
	.dc.b	$C3,$ED,$F3,$3E,$39,$1F,$EE,$75,$D5,$F4,$46,$55,$FF,$71,$C3,$FB
	.dc.b	$03,$EB,$C3,$FA,$06,$1F,$D5,$29,$F0,$4F,$9C,$F1,$79,$DE,$49,$CC
	.dc.b	$61,$F3,$CE,$DF,$F6,$5D,$3E,$3B,$84,$E7,$4F,$8B,$86,$30,$E5,$87
	.dc.b	$43,$8C,$65,$E7,$76,$B8,$C7,$82,$B8,$DF,$5C,$3D,$4D,$BE,$55,$FC
	.dc.b	$FB,$DA,$FF,$12,$CF,$9E,$31,$EA,$73,$7E,$8F,$47,$F9,$4A,$31,$F1
	.dc.b	$3F,$E5,$DF,$A3,$E7,$39,$AF,$41,$C8,$AE,$6F,$E2,$57,$19,$EC,$6C
	.dc.b	$F9,$85,$3E,$74,$F5,$72,$D7,$3E,$9B,$E4,$38,$C7,$A2,$30,$E2,$51
	.dc.b	$88,$E6,$D7,$9C,$F9,$27,$6F,$C0,$F6,$1F,$44,$E8,$D6,$1D,$B2,$F9
	.dc.b	$58,$70,$C7,$E0,$CE,$78,$3F,$37,$9B,$2E,$78,$F2,$CF,$E8,$70,$BE
	.dc.b	$8F,$F6,$73,$86,$E5,$F8,$59,$D4,$F8,$E3,$13,$AD,$3F,$1C,$DC,$BF
	.dc.b	$57,$7D,$7A,$38,$56,$B0,$B5,$AB,$E1,$83,$0E,$15,$E3,$C3,$73,$9A
	.dc.b	$F8,$93,$93,$F1,$5D,$3F,$C6,$75,$F3,$73,$FE,$C5,$B3,$E4,$F9,$17
	.dc.b	$5E,$31,$FE,$31,$57,$0E,$57,$FC,$4E,$35,$9B,$0E,$33,$93,$F8,$1B
	.dc.b	$8B,$FE,$22,$D7,$F7,$53,$C7,$0F,$E5,$2D,$AF,$E3,$55,$BA,$F3,$56
	.dc.b	$A3,$11,$F1,$5B,$87,$F3,$7E,$5F,$47,$57,$FA,$EB,$0F,$E5,$5C,$98
	.dc.b	$9C,$BF,$CE,$2B,$F9,$F3,$6B,$F0,$B8,$7F,$42,$6B,$FD,$05,$AF,$F3
	.dc.b	$F8,$E0,$4B,$FE,$7B,$F2,$DE,$23,$C7,$EE,$D8,$FE,$78,$F7,$23,$FD
	.dc.b	$11,$F1,$7B,$7E,$A3,$D7,$EB,$D9,$FE,$89,$17,$0F,$E5,$B1,$FD,$46
	.dc.b	$71,$C1,$FC,$BE,$AC,$E4,$E2,$7F,$4D,$9F,$80,$47,$FA,$AB,$F0,$FD
	.dc.b	$8F,$96,$7C,$CF,$8B,$B3,$FD,$55,$F9,$04,$7F,$AE,$BF,$2F,$E8,$7E
	.dc.b	$9F,$AD,$F5,$1B,$3F,$D7,$5B,$97,$A8,$FE,$D0,$E3,$2F,$E9,$0F,$19
	.dc.b	$FE,$C9,$F0,$96,$B8,$70,$7F,$77,$8F,$B4,$49,$FE,$DE,$FD,$DF,$A3
	.dc.b	$FC,$2F,$CB,$7E,$19,$B3,$FD,$BD,$B7,$D3,$89,$18,$CB,$FA,$D7,$D0
	.dc.b	$7E,$3B,$E4,$3F,$1E,$DC,$3F,$BC,$37,$0E,$4D,$AF,$EC,$5F,$47,$F9
	.dc.b	$63,$F6,$1F,$28,$F9,$56,$F9,$C7,$FD,$BF,$39,$8F,$F0,$7B,$63,$F9
	.dc.b	$8B,$71,$FE,$76,$D7,$FB,$43,$5F,$EF,$AD,$4D,$31,$D6,$F6,$DE,$6F
	.dc.b	$EB,$3E,$6F,$F8,$0C,$63,$0F,$F0,$B4,$62,$75,$1F,$E8,$56,$E1,$F5
	.dc.b	$BE,$8D,$E5,$7F,$0D,$6D,$7F,$BA,$57,$CA,$E6,$64,$E5,$E1,$18,$AE
	.dc.b	$9F,$15,$CC,$F9,$7B,$7F,$BD,$AF,$A4,$63,$D1,$F9,$33,$9A,$F8,$ED
	.dc.b	$7E,$3B,$5F,$8E,$D7,$E3,$B7,$6F,$B1,$3D,$0F,$E1,$99,$7F,$86,$70
	.dc.b	$F6,$7C,$EF,$BF,$C8,$F8,$0D,$8C,$9F,$06,$32,$FF,$AD,$F1,$FE,$65
	.dc.b	$E6,$FF,$6F,$8C,$1F,$EB,$BF,$2F,$33,$A6,$C6,$4F,$F7,$8F,$85,$FC
	.dc.b	$EF,$E2,$0F,$F3,$F9,$7D,$1B,$A7,$53,$FD,$0F,$07,$C6,$3E,$8F,$85
	.dc.b	$EF,$2F,$F2,$3F,$5F,$F9,$A6,$B5,$18,$3F,$C9,$79,$76,$CB,$A4,$EB
	.dc.b	$1F,$E0,$13,$9A,$E1,$1F,$E5,$29,$D5,$FF,$6F,$AD,$4F,$F0,$0F,$8D
	.dc.b	$FF,$2B,$DF,$0F,$09,$18,$9D,$57,$F5,$B8,$C5,$FF,$79,$AD,$D6,$4D
	.dc.b	$55,$BE,$9F,$46,$F8,$1F,$EC,$8F,$FA,$EA,$33,$18,$F8,$0C,$B6,$32
	.dc.b	$72,$8F,$ED,$9D,$1F,$D9,$72,$F1,$36,$7F,$6B,$8B,$E2,$37,$E8,$7F
	.dc.b	$06,$8C,$4F,$F0,$06,$9F,$C0,$0A,$7F,$00,$29,$D2,$7F,$84,$B8,$3F
	.dc.b	$80,$97,$A7,$FD,$03,$1F,$D9,$7E,$8D,$F4,$7F,$E0,$35,$F2,$76,$30
	.dc.b	$FF,$81,$FE,$4B,$F8,$26,$1F,$EA,$A6,$BF,$EA,$96,$BF,$E8,$EE,$14
	.dc.b	$62,$3F,$D7,$AB,$E8,$FA,$1D,$7E,$7E,$E9,$F7,$FE,$59,$DD,$FE,$17
	.dc.b	$6C,$7E,$0F,$E5,$FF,$3E,$EE,$7F,$C3,$91,$8C,$7F,$D7,$ED,$7F,$D7
	.dc.b	$CD,$7C,$CD,$7F,$DA,$2D,$C8,$D2,$76,$7F,$83,$7E,$63,$EE,$57,$FB
	.dc.b	$61,$B4,$7F,$27,$87,$C3,$72,$7E,$66,$7F,$DA,$0D,$74,$75,$24,$7F
	.dc.b	$9C,$A7,$A4,$60,$FF,$6D,$CF,$17,$FD,$D3,$EE,$3C,$8F,$F1,$F3,$6F
	.dc.b	$F5,$5F,$37,$F8,$CE,$7F,$D3,$4E,$6F,$FB,$7C,$FE,$27,$13,$49,$C9
	.dc.b	$FE,$2D,$6E,$1E,$56,$D7,$FB,$BB,$E7,$3E,$BB,$FD,$E9,$B8,$FE,$D7
	.dc.b	$E7,$7F,$CE,$2F,$ED,$E3,$37,$F8,$99,$FF,$3A,$D7,$FB,$D5,$AE,$72
	.dc.b	$FF,$1F,$46,$21,$92,$30,$D9,$D9,$2B,$E6,$79,$3F,$5A,$F8,$B6,$BF
	.dc.b	$CE,$DF,$11,$FE,$67,$8C,$79,$C8,$FF,$27,$36,$FE,$B3,$6F,$FD,$63
	.dc.b	$7C,$A7,$35,$C0,$8E,$0F,$BD,$E0,$7E,$D1,$FE,$40,$DA,$FF,$11,$FC
	.dc.b	$F3,$DB,$FF,$3D,$4E,$76,$FE,$41,$39,$BF,$C4,$36,$FC,$6D,$7F,$DE
	.dc.b	$18,$A3,$11,$FB,$96,$BF,$31,$B1,$BC,$3F,$CD,$D1,$8F,$82,$FC,$65
	.dc.b	$FF,$01,$6D,$FD,$48,$7B,$BF,$E9,$19,$CE,$1F,$EC,$18,$FF,$3C,$4E
	.dc.b	$8F,$D0,$16,$BE,$5B,$5F,$4F,$D1,$BE,$2B,$FD,$38,$D9,$FE,$90,$BE
	.dc.b	$7F,$33,$27,$2F,$08,$C5,$7F,$94,$3E,$8F,$FC,$2F,$5F,$D4,$5B,$19
	.dc.b	$F1,$7F,$D5,$51,$8D,$BF,$A8,$36,$BF,$91,$36,$BF,$2A,$AB,$8F,$17
	.dc.b	$FA,$C3,$E8,$9E,$ED,$3F,$D6,$7F,$0B,$F2,$B9,$53,$AC,$BF,$B5,$C6
	.dc.b	$2B,$84,$7F,$AD,$BA,$AE,$7F,$E7,$1A,$FF,$5E,$47,$FA,$BC,$93,$FE
	.dc.b	$B4,$7F,$DC,$13,$FF,$00,$07,$11,$74,$FF,$59,$70,$7F,$86,$4C,$CE
	.dc.b	$1E,$13,$C6,$78,$47,$F8,$22,$B1,$5B,$8E,$E7,$23,$AF,$37,$FB,$CE
	.dc.b	$B8,$C7,$43,$47,$32,$99,$3D,$57,$EB,$70,$77,$7D,$DF,$46,$F8,$CF
	.dc.b	$F3,$B4,$F8,$E3,$1E,$C6,$A7,$33,$ED,$9C,$BB,$F9,$E7,$1F,$46,$C6
	.dc.b	$63,$07,$BF,$57,$D2,$3E,$73,$E8,$46,$1B,$19,$AE,$0D,$7E,$74,$ED
	.dc.b	$CC,$F0,$23,$EE,$FA,$9F,$85,$89,$A4,$8F,$27,$EC,$EF,$91,$F1,$B6
	.dc.b	$33,$18,$23,$82,$1E,$48,$E0,$F9,$BF,$16,$79,$5E,$E6,$C7,$C3,$6B
	.dc.b	$B8,$D1,$1D,$35,$FC,$47,$E2,$2F,$B8,$FB,$CD,$8F,$CD,$35,$DC,$68
	.dc.b	$8E,$9A,$FE,$4F,$EA,$DF,$0F,$F8,$96,$C7,$E9,$5A,$EE,$3F,$1D,$F8
	.dc.b	$43,$E9,$29,$27,$F5,$91,$97,$97,$C6,$F5,$23,$82,$1E,$37,$2D,$39
	.dc.b	$FD,$0F,$80,$8E,$0F,$DD,$7E,$60,$FA,$2F,$D1,$36,$33,$18,$23,$82
	.dc.b	$1F,$B5,$8E,$0F,$ED,$ED,$EE,$FC,$02,$47,$B9,$FD,$87,$CA,$FF,$8E
	.dc.b	$A3,$F8,$63,$5F,$E0,$0D,$7F,$84,$35,$FE,$1A,$D7,$F8,$93,$53,$4C
	.dc.b	$4B,$E7,$FE,$99,$27,$E8,$BF,$C0,$FF,$3B,$7D,$35,$24,$FF,$21,$8C
	.dc.b	$B6,$32,$FF,$17,$8C,$B6,$3F,$87,$70,$3E,$5A,$92,$7F,$96,$46,$5F
	.dc.b	$E5,$DC,$ED,$FF,$8B,$36,$33,$18,$23,$82,$1F,$CA,$63,$83,$FC,$AF
	.dc.b	$F3,$67,$D0,$7F,$21,$6C,$7F,$3D,$6B,$B8,$FE,$4D,$F9,$BF,$F9,$D6
	.dc.b	$3F,$A1,$35,$FE,$6C,$D7,$F9,$F3,$5F,$E8,$6D,$7F,$A4,$35,$34,$C7
	.dc.b	$E6,$FF,$DD,$63,$FA,$83,$5F,$E8,$8D,$7F,$A6,$35,$FE,$A2,$FF,$39
	.dc.b	$78,$F8,$5F,$CF,$E3,$11,$CA,$3F,$A1,$7E,$8A,$F9,$8A,$49,$FE,$C1
	.dc.b	$19,$6F,$A9,$C0,$8E,$0F,$EC,$DF,$3E,$FB,$EF,$E9,$CD,$8C,$C6,$08
	.dc.b	$E0,$87,$F6,$18,$E0,$FF,$63,$F6,$7F,$94,$BF,$DA,$DA,$FF,$70,$6B
	.dc.b	$FD,$D5,$AB,$EB,$DF,$09,$FD,$85,$B1,$FD,$F5,$AF,$F7,$16,$BF,$DE
	.dc.b	$5A,$FF,$7E,$6B,$FE,$03,$6B,$DE,$3B,$7D,$54,$8E,$CD,$7B,$35,$EC
	.dc.b	$D7,$FC,$17,$6F,$96,$F9,$64,$8F,$9E,$D7,$E4,$35,$FF,$08,$47,$17
	.dc.b	$DD,$F3,$5F,$04,$8E,$0F,$F0,$F7,$D6,$FF,$95,$1B,$A7,$F8,$1E,$31
	.dc.b	$1F,$E1,$56,$BF,$E1,$D6,$BF,$E2,$26,$BF,$E2,$76,$A6,$8F,$A0,$EC
	.dc.b	$48,$F1,$B5,$F0,$35,$F0,$3F,$E1,$BF,$CA,$5F,$45,$49,$3F,$E3,$98
	.dc.b	$CB,$63,$2F,$F8,$BE,$32,$D8,$FF,$13,$7D,$6F,$F9,$56,$3F,$C8,$4D
	.dc.b	$7F,$C6,$CD,$7F,$C7,$CD,$7F,$C8,$6B,$D2,$E6,$48,$E6,$D7,$9B,$5E
	.dc.b	$6F,$8B,$86,$47,$89,$AF,$89,$AF,$89,$AF,$F9,$53,$F4,$EF,$F5,$B3
	.dc.b	$FE,$5B,$6B,$FE,$4F,$6B,$FE,$57,$6B,$C7,$E1,$7F,$C9,$11,$88,$FF
	.dc.b	$31,$35,$37,$C7,$FE,$F9,$1F,$E5,$B6,$BF,$E6,$36,$BF,$E6,$86,$BF
	.dc.b	$E6,$D5,$F2,$BF,$E5,$76,$C7,$F9,$CD,$AF,$58,$D1,$1D,$35,$F1,$7C
	.dc.b	$AF,$32,$3C,$DA,$E8,$C1,$0F,$F3,$A7,$CF,$7F,$39,$27,$C8,$FF,$9A
	.dc.b	$DF,$E7,$DE,$E3,$F3,$7F,$C0,$57,$F9,$EE,$31,$AF,$FA,$2D,$B1,$99
	.dc.b	$DB,$FE,$81,$FA,$57,$CF,$7F,$9D,$DB,$1F,$E9,$36,$BB,$8D,$11,$D3
	.dc.b	$5E,$9F,$2F,$CC,$8F,$37,$FC,$DF,$FB,$FF,$F3,$08,$FF,$4D,$FB,$88
	.dc.b	$C4,$7F,$A8,$63,$1C,$5C,$0F,$37,$93,$27,$CC,$D7,$FD,$45,$9F,$C5
	.dc.b	$FB,$2F,$85,$A1,$A4,$9F,$88,$D7,$8C,$6D,$C9,$0C,$C6,$DF,$F5,$3C
	.dc.b	$6D,$FF,$56,$7D,$17,$F5,$B6,$F8,$9F,$EA,$38,$C4,$7F,$AD,$DA,$FB
	.dc.b	$FD,$6F,$F1,$EB,$FE,$97,$6B,$FE,$90,$6B,$F2,$1A,$FF,$B1,$21,$F4
	.dc.b	$FD,$C9,$1F,$EC,$D8,$CB,$63,$2F,$A1,$AF,$B3,$E6,$7E,$49,$1F,$C5
	.dc.b	$3E,$BF,$A7,$91,$D5,$AF,$C2,$6B,$EF,$6B,$E7,$7F,$DB,$7F,$06,$F8
	.dc.b	$8F,$C0,$7D,$57,$F2,$6F,$87,$FE,$99,$C4,$8C,$78,$FC,$71,$FF,$73
	.dc.b	$61,$FE,$EA,$D1,$97,$4C,$B9,$F8,$3F,$EE,$8C,$F9,$E5,$CB,$1F,$F7
	.dc.b	$76,$1C,$1E,$77,$F2,$FC,$BF,$F7,$7B,$F5,$DB,$19,$3F,$DE,$D1,$96
	.dc.b	$99,$3F,$DF,$A2,$53,$CA,$3C,$2D,$74,$D8,$DC,$7F,$BF,$33,$FD,$27
	.dc.b	$ED,$0F,$8E,$F6,$46,$1B,$3D,$0E,$07,$BB,$EA,$DF,$53,$88,$F2,$46
	.dc.b	$16,$9F,$DE,$DC,$CF,$2F,$AD,$7D,$B7,$FB,$3A,$B5,$9B,$0E,$51,$86
	.dc.b	$E5,$F9,$56,$E2,$C3,$9F,$C4,$FA,$23,$B9,$FA,$A9,$F2,$3F,$A7,$6B
	.dc.b	$CF,$2F,$CE,$79,$EE,$BA,$49,$FD,$9C,$65,$B1,$F0,$73,$4E,$73,$FF
	.dc.b	$0D,$64,$FA,$EF,$F4,$FF,$0E,$EB,$5A,$FE,$BD,$B9,$7E,$29,$B7,$C7
	.dc.b	$E0,$7F,$39,$3F,$32,$B8,$67,$F4,$7C,$67,$A3,$AD,$3F,$3F,$C7,$FD
	.dc.b	$EF,$4D,$4E,$70,$FC,$0F,$8D,$72,$0C,$79,$DF,$E4,$A3,$31,$87,$F2
	.dc.b	$6D,$8E,$31,$AB,$F3,$C6,$23,$A4,$61,$DC,$F1,$6C,$65,$E5,$3E,$17
	.dc.b	$C9,$F0,$1F,$0B,$AB,$E1,$F5,$FE,$EF,$D0,$F5,$56,$AF,$F6,$2A,$75
	.dc.b	$1F,$1E,$F9,$3F,$C1,$AB,$8E,$3F,$ED,$18,$E3,$5C,$2F,$E8,$1F,$0A
	.dc.b	$FF,$66,$D9,$F7,$34,$B5,$CD,$B3,$C1,$F3,$18,$9E,$2E,$0F,$EB,$7F
	.dc.b	$9A,$3F,$05,$EF,$82,$31,$5E,$08,$C4,$EA,$73,$5E,$2F,$B5,$F0,$62
	.dc.b	$69,$30,$F0,$C6,$5B,$19,$9E,$51,$DF,$CD,$FC,$DC,$7D,$DC,$51,$88
	.dc.b	$FE,$E6,$D7,$FB,$93,$5F,$C5,$FA,$DE,$42,$57,$89,$F7,$FE,$2E,$F9
	.dc.b	$0A,$49,$F8,$11,$96,$C6,$6B,$C9,$39,$6C,$65,$FF,$03,$C6,$5B,$19
	.dc.b	$7F,$2B,$F1,$BF,$20,$95,$F1,$9F,$A7,$F6,$8F,$6A,$92,$7E,$A4,$65
	.dc.b	$FB,$BF,$61,$F5,$2F,$F4,$8D,$7F,$52,$D7,$F6,$6D,$4B,$E0,$BF,$14
	.dc.b	$D9,$EB,$18,$23,$82,$4F,$E5,$A7,$83,$F6,$7F,$4D,$7B,$5F,$8E,$6D
	.dc.b	$7E,$35,$B3,$B9,$F0,$FC,$2C,$AF,$6B,$67,$F8,$07,$D6,$3E,$59,$E8
	.dc.b	$7E,$A2,$30,$D8,$CD,$FE,$AA,$72,$D8,$CD,$74,$6B,$FC,$42,$76,$E6
	.dc.b	$BA,$7D,$84,$BE,$4D,$9E,$4D,$79,$5A,$31,$1F,$C9,$3C,$17,$E2,$DF
	.dc.b	$ED,$DB,$FF,$33,$36,$7F,$C8,$ED,$7F,$91,$CF,$0F,$03,$F8,$07,$D7
	.dc.b	$BE,$32,$93,$0F,$E0,$31,$96,$C6,$6F,$F8,$6C,$E5,$B1,$F2,$49,$3F
	.dc.b	$1D,$B1,$F1,$DA,$FB,$9A,$FF,$40,$D3,$EB,$79,$FF,$4C,$95,$F4,$5B
	.dc.b	$3F,$45,$AF,$C8,$7F,$91,$F8,$2F,$CA,$9E,$B6,$1B,$7F,$C9,$A7,$53
	.dc.b	$93,$F9,$27,$E5,$4E,$7A,$4A,$FE,$4B,$19,$6C,$67,$1F,$E6,$33,$96
	.dc.b	$C7,$E3,$7E,$D6,$57,$E2,$DB,$3F,$8B,$6B,$F8,$26,$BF,$17,$89,$F5
	.dc.b	$3F,$97,$25,$7E,$55,$B3,$F9,$56,$BF,$8C,$7F,$A3,$F0,$BE,$D7,$F3
	.dc.b	$33,$F8,$D6,$BF,$E1,$16,$C6,$F9,$7E,$2F,$CC,$4C,$3E,$1B,$6F,$E0
	.dc.b	$B5,$FE,$E1,$A2,$31,$1F,$7C,$7D,$CF,$EC,$DB,$5F,$B1,$69,$FB,$A2
	.dc.b	$9F,$AF,$29,$CE,$39,$12,$39,$7D,$D7,$52,$38,$24,$F5,$8C,$B6,$3F
	.dc.b	$B5,$FD,$3B,$F0,$67,$BD,$FD,$C2,$30,$D8,$CC,$F0,$6D,$7F,$74,$FB
	.dc.b	$87,$D5,$7F,$71,$6D,$7F,$70,$6C,$EE,$74,$49,$D3,$63,$FB,$87,$07
	.dc.b	$F7,$BF,$27,$F0,$B5,$FD,$E9,$B3,$A8,$FF,$75,$79,$1F,$E0,$1E,$13
	.dc.b	$49,$1F,$E0,$2F,$AE,$7C,$D7,$F8,$1D,$B1,$98,$C1,$9F,$87,$91,$C5
	.dc.b	$B8,$7F,$85,$29,$E1,$F3,$25,$77,$6C,$ED,$AF,$F8,$E6,$78,$CF,$F8
	.dc.b	$78,$7E,$72,$F8,$FF,$F1,$1C,$61,$B1,$98,$E0,$DA,$FE,$7D,$C7,$FE
	.dc.b	$23,$F0,$0D,$25,$7F,$89,$23,$2D,$8C,$CF,$28,$F2,$7D,$EF,$C6,$8F
	.dc.b	$F2,$84,$23,$11,$FE,$B2,$6B,$FE,$B0,$6B,$FE,$B2,$F9,$8F,$F1,$B4
	.dc.b	$9A,$47,$FC,$6E,$64,$AE,$63,$E3,$3F,$0B,$E5,$A4,$FC,$26,$CF,$F9
	.dc.b	$23,$EB,$D1,$FE,$45,$6C,$FF,$91,$1B,$3B,$9D,$12,$74,$D8,$F8,$39
	.dc.b	$3C,$FF,$A6,$4A,$FA,$6F,$F9,$4F,$E0,$BF,$88,$D7,$FA,$F3,$BB,$04
	.dc.b	$AF,$F2,$AC,$E8,$91,$99,$FF,$3F,$72,$3F,$51,$7C,$8F,$F2,$06,$E1
	.dc.b	$FE,$66,$9D,$4E,$4F,$F3,$27,$6F,$DE,$BF,$E0,$3C,$18,$25,$FF,$99
	.dc.b	$A7,$53,$FE,$8D,$6B,$FE,$92,$6B,$FE,$9C,$6F,$8C,$FC,$E2,$92,$FF
	.dc.b	$AC,$BC,$29,$4F,$A4,$F2,$93,$0F,$13,$E8,$DA,$47,$95,$B5,$FC,$69
	.dc.b	$AF,$9D,$AF,$FA,$C3,$AD,$FB,$1F,$FB,$CE,$BF,$CF,$8D,$9F,$F3,$73
	.dc.b	$5F,$F5,$7C,$F0,$F8,$BF,$F4,$2C,$E6,$7A,$35,$FF,$51,$36,$78,$57
	.dc.b	$FA,$27,$EA,$FF,$C7,$B5,$D1,$AF,$95,$AF,$FA,$AD,$B2,$69,$3D,$3F
	.dc.b	$F4,$BB,$C2,$7F,$D9,$35,$FE,$9E,$7F,$D9,$CD,$F7,$3F,$D4,$91,$83
	.dc.b	$FD,$19,$3F,$B9,$9F,$37,$4F,$42,$38,$24,$F0,$8C,$BD,$FE,$13,$C0
	.dc.b	$9A,$47,$FD,$73,$F5,$7F,$BF,$F3,$3F,$D6,$6D,$8C,$F6,$7F,$96,$DB
	.dc.b	$3A,$F8,$47,$2D,$24,$F4,$8C,$BE,$CF,$73,$FA,$36,$9F,$E9,$D8,$FA
	.dc.b	$51,$82,$38,$21,$C4,$F2,$F8,$5F,$2F,$2F,$E7,$B1,$82,$1F,$EB,$5F
	.dc.b	$AB,$7C,$6F,$F5,$56,$C7,$FB,$25,$AE,$E3,$44,$74,$D7,$FD,$81,$F5
	.dc.b	$7F,$C3,$5F,$F6,$93,$5C,$F3,$7F,$63,$8C,$47,$E6,$1A,$FE,$85,$AF
	.dc.b	$FB,$DD,$BC,$87,$30,$FA,$B7,$CB,$FE,$2D,$B1,$FE,$EA,$6B,$FE,$F4
	.dc.b	$6A,$21,$CE,$39,$11,$E4,$FE,$5F,$EA,$FF,$D3,$31,$FE,$F6,$6B,$FF
	.dc.b	$00,$1A,$F4,$0E,$1E,$FF,$53,$FE,$33,$7F,$DF,$ED,$78,$07,$0F,$36
	.dc.b	$BE,$88,$7D,$47,$94,$91,$E5,$6B,$E5,$6B,$E5,$8E,$2F,$8B,$C9,$FD
	.dc.b	$A5,$7F,$7B,$8C,$10,$F2,$FD,$5F,$F2,$37,$CE,$1C,$3A,$EE,$FD,$F4
	.dc.b	$62,$3E,$23,$5F,$A2,$D7,$EA,$35,$E9,$1E,$4F,$AB,$91,$F2,$1A,$FC
	.dc.b	$8F,$90,$FF,$22,$BE,$46,$B9,$F5,$7F,$88,$46,$23,$EB,$35,$FC,$1B
	.dc.b	$5F,$C3,$35,$37,$07,$C6,$FC,$82,$4F,$DC,$7F,$8F,$7D,$5F,$F8,$EA
	.dc.b	$3E,$3B,$5F,$BE,$D7,$F1,$0D,$7F,$2A,$D7,$F3,$0D,$4D,$31,$B7,$FB
	.dc.b	$36,$3E,$AB,$5F,$C6,$B5,$FC,$B3,$5F,$D1,$F3,$7E,$11,$FC,$9B,$5F
	.dc.b	$CE,$35,$DE,$FF,$CE,$E3,$11,$FA,$D6,$BF,$E5,$08,$D1,$FD,$07,$EA
	.dc.b	$DF,$13,$49,$3F,$8C,$8C,$BF,$8E,$FA,$B7,$87,$FD,$25,$B1,$98,$C1
	.dc.b	$1C,$10,$FD,$6C,$70,$7F,$5F,$F5,$7F,$E8,$47,$F3,$6D,$73,$EC,$FF
	.dc.b	$57,$8C,$47,$EF,$5F,$F2,$23,$C3,$C2,$D1,$1F,$D2,$FD,$5F,$FB,$89
	.dc.b	$BF,$07,$FD,$86,$31,$1F,$C0,$1A,$FF,$09,$5F,$1B,$B2,$47,$47,$FB
	.dc.b	$57,$CB,$BD,$AA,$49,$FD,$AC,$65,$B1,$97,$F8,$7C,$65,$B1,$FD,$D7
	.dc.b	$E1,$FF,$A1,$C7,$F1,$B6,$BF,$CA,$1A,$FF,$2B,$6A,$FA,$BF,$F2,$DB
	.dc.b	$FC,$19,$AF,$F1,$E6,$BF,$C9,$DA,$FF,$34,$6B,$FC,$DD,$AF,$5F,$67
	.dc.b	$FC,$15,$18,$3F,$D5,$71,$DF,$DE,$C9,$1C,$10,$F1,$39,$7E,$1F,$C5
	.dc.b	$FC,$C2,$3F,$29,$AE,$8C,$10,$F6,$7C,$B7,$90,$8F,$C1,$7F,$C4,$1F
	.dc.b	$57,$F7,$23,$F9,$4B,$5F,$E8,$2D,$7F,$A5,$35,$FE,$AE,$D7,$FA,$D3
	.dc.b	$53,$49,$F2,$FE,$02,$47,$81,$AF,$81,$AF,$81,$FF,$1B,$7D,$5B,$E8
	.dc.b	$29,$27,$FA,$0C,$65,$B1,$97,$FA,$DC,$65,$B1,$FE,$41,$FA,$B7,$D8
	.dc.b	$D2,$4F,$F4,$C8,$CB,$7D,$1E,$04,$70,$43,$FB,$4C,$23,$51,$87,$C1
	.dc.b	$C1,$23,$B3,$5E,$2D,$7F,$BD,$4F,$E9,$5B,$1F,$E5,$9F,$AC,$7B,$14
	.dc.b	$93,$FD,$92,$32,$D8,$CB,$FD,$FA,$32,$D8,$FF,$32,$79,$3F,$DF,$A3
	.dc.b	$FC,$2C,$D7,$FD,$DD,$F1,$9A,$37,$F5,$92,$3F,$B9,$CA,$31,$1F,$E0
	.dc.b	$C8,$C3,$A8,$CB,$FE,$17,$8C,$BB,$F9,$AF,$19,$1E,$AD,$74,$60,$87
	.dc.b	$F8,$6B,$EB,$3F,$97,$3F,$E2,$36,$BF,$E3,$06,$BF,$E3,$36,$FC,$1B
	.dc.b	$EA,$3E,$D4,$FF,$B1,$DA,$FF,$B3,$9A,$FF,$8E,$A3,$86,$DF,$E8,$DF
	.dc.b	$AC,$FF,$68,$CF,$F8,$59,$AF,$F8,$B9,$AF,$F8,$DD,$AF,$F9,$11,$AF
	.dc.b	$F9,$1D,$A9,$A4,$FA,$CF,$01,$23,$C0,$FF,$89,$BD,$57,$D6,$7E,$2E
	.dc.b	$7F,$D8,$0D,$51,$AE,$1E,$13,$99,$F0,$79,$7E,$D2,$3E,$D7,$FC,$67
	.dc.b	$97,$C9,$FA,$CB,$A1,$A4,$AF,$98,$D7,$E6,$35,$F9,$91,$C5,$D9,$0D
	.dc.b	$9F,$E4,$2F,$A7,$FE,$B2,$FF,$9B,$1A,$FF,$9B,$DA,$9C,$E4,$F8,$B7
	.dc.b	$A9,$F1,$DB,$5F,$E6,$49,$D4,$E4,$FF,$65,$53,$EB,$29,$2B,$C8,$D7
	.dc.b	$FC,$DB,$1B,$72,$43,$F1,$9E,$AF,$C8,$23,$F9,$17,$BF,$D6,$48,$F8
	.dc.b	$0D,$7A,$B5,$FF,$41,$4A,$31,$1F,$15,$FF,$48,$FC,$13,$E1,$BF,$D2
	.dc.b	$99,$3E,$B3,$F9,$47,$C3,$3C,$5C,$C6,$1E,$58,$FF,$A6,$F0,$E4,$FC
	.dc.b	$FC,$7A,$E3,$C3,$C1,$FF,$4F,$E3,$CE,$BF,$1F,$8F,$37,$81,$CE,$BE
	.dc.b	$D7,$93,$FE,$A6,$7D,$6D,$8C,$9F,$EA,$B8,$CB,$4C,$9E,$43,$F0,$51
	.dc.b	$C5,$A6,$CA,$6C,$FC,$33,$E8,$CB,$E2,$FD,$43,$E2,$B1,$5F,$A8,$8C
	.dc.b	$3F,$EB,$DC,$7D,$CF,$23,$15,$F9,$5F,$AB,$7C,$8E,$27,$F2,$D1,$85
	.dc.b	$5C,$B8,$13,$98,$F9,$5F,$54,$FA,$CF,$F7,$4C,$6E,$75,$8F,$7C,$7B
	.dc.b	$63,$FA,$19,$CA,$C3,$97,$C3,$FB,$23,$55,$FA,$DA,$F2,$47,$E3,$1A
	.dc.b	$F5,$3F,$3B,$E8,$BA,$FF,$68,$D8,$CC,$60,$8F,$BB,$1F,$DC,$4E,$F2
	.dc.b	$F6,$FD,$67,$FA,$83,$87,$D3,$B2,$B5,$7C,$31,$CC,$63,$1F,$66,$1C
	.dc.b	$5E,$C7,$F0,$38,$D3,$FE,$7C,$F3,$FF,$33,$97,$E5,$79,$DC,$3F,$DE
	.dc.b	$F4,$F9,$3E,$D2,$F7,$A7,$E1,$3C,$4B,$97,$FB,$F6,$B5,$7F,$8D,$8E
	.dc.b	$CD,$8C,$C7,$E3,$DB,$1F,$EF,$96,$C0,$E9,$7F,$AE,$9D,$36,$34,$76
	.dc.b	$3A,$7C,$07,$ED,$3A,$BF,$6B,$9F,$EF,$7D,$0F,$3D,$7F,$A7,$2B,$58
	.dc.b	$7E,$FA,$73,$8F,$FB,$3A,$3F,$69,$5F,$B8,$9C,$D7,$08,$C4,$FC,$68
	.dc.b	$FD,$54,$7C,$36,$BD,$E3,$FD,$F3,$5D,$63,$D2,$6E,$33,$3C,$27,$53
	.dc.b	$F1,$63,$AD,$BE,$AD,$F5,$98,$6E,$1F,$CE,$5B,$7D,$AB,$85,$75,$FA
	.dc.b	$C9,$7E,$46,$D7,$90,$FF,$60,$B8,$7E,$B3,$5F,$E5,$3E,$5F,$E3,$DF
	.dc.b	$A7,$68,$C4,$6B,$E1,$7D,$31,$88,$FC,$47,$AD,$EF,$26,$1F,$19,$F7
	.dc.b	$FD,$63,$E3,$E9,$27,$FA,$B4,$65,$B1,$9A,$F8,$13,$96,$C6,$5F,$3C
	.dc.b	$65,$B1,$97,$F1,$BB,$3E,$37,$EA,$12,$FE,$DB,$F5,$3E,$97,$FD,$93
	.dc.b	$1F,$DA,$5B,$5F,$45,$AF,$C6,$6B,$F9,$AF,$AC,$FA,$EF,$E1,$1A,$E7
	.dc.b	$7F,$C5,$C6,$23,$F4,$CF,$FB,$F2,$BE,$C4,$23,$0F,$DB,$6B,$D7,$E1
	.dc.b	$B9,$92,$78,$CE,$A7,$C9,$3F,$90,$FA,$CF,$F3,$8D,$7F,$81,$5B,$3F
	.dc.b	$8D,$6B,$F8,$26,$BE,$DF,$89,$F4,$92,$7B,$BF,$A8,$F9,$E7,$CA,$FF
	.dc.b	$85,$67,$F2,$51,$8F,$3F,$53,$9B,$FD,$4C,$E6,$B7,$39,$9F,$E0,$B3
	.dc.b	$97,$71,$97,$F8,$74,$65,$DC,$3E,$B3,$B9,$2B,$AB,$67,$9B,$5E,$2F
	.dc.b	$F0,$2F,$09,$F5,$9F,$E0,$BA,$FD,$64,$63,$BF,$53,$9B,$FE,$07,$39
	.dc.b	$AF,$B7,$F5,$8F,$AC,$FE,$16,$D8,$CC,$60,$8E,$09,$5F,$BB,$9E,$0F
	.dc.b	$C2,$9F,$96,$49,$F9,$2D,$8D,$46,$08,$FC,$AF,$37,$E9,$92,$3E,$8B
	.dc.b	$63,$51,$82,$3F,$C8,$3C,$07,$D6,$3E,$2B,$F9,$4C,$61,$B1,$9A,$FE
	.dc.b	$29,$39,$7F,$3F,$F5,$4E,$7F,$E5,$ED,$8C,$C6,$08,$E0,$98,$7F,$1F
	.dc.b	$9E,$0F,$E1,$BE,$B3,$F1,$E4,$9F,$C6,$B6,$35,$18,$23,$F8,$EF,$A8
	.dc.b	$FC,$B9,$23,$F2,$AD,$8D,$46,$08,$FE,$0B,$EB,$3F,$2C,$FF,$83,$5B
	.dc.b	$B9,$A6,$A1,$83,$E1,$3E,$19,$2F,$E3,$B6,$7E,$03,$5F,$EA,$73,$C6
	.dc.b	$7F,$63,$0F,$AC,$FD,$B3,$63,$31,$82,$38,$24,$7E,$D8,$E0,$5F,$A4
	.dc.b	$E6,$48,$EE,$D8,$D4,$60,$8F,$F6,$8F,$A3,$7D,$63,$DC,$FE,$E3,$18
	.dc.b	$6C,$66,$BF,$AE,$4E,$5F,$E4,$FF,$58,$F9,$6F,$EE,$ED,$8C,$C6,$08
	.dc.b	$E0,$98,$7F,$67,$9E,$0F,$FA,$A7,$7F,$F0,$0F,$94,$69,$2B,$FD,$93
	.dc.b	$E4,$7F,$81,$B8,$4D,$24,$7F,$3C,$FA,$B7,$D6,$7F,$83,$1B,$19,$8C
	.dc.b	$19,$F8,$79,$1C,$5B,$97,$F7,$DA,$78,$7C,$C9,$5D,$DB,$3B,$6B,$FE
	.dc.b	$18,$9E,$33,$FE,$07,$1F,$56,$F8,$CF,$F1,$1C,$61,$B1,$98,$E0,$DA
	.dc.b	$FE,$7F,$C7,$FE,$28,$F1,$0D,$25,$7F,$5E,$8C,$B6,$33,$3C,$A3,$C9
	.dc.b	$F4,$FF,$1A,$3F,$C5,$F0,$8C,$47,$FA,$BD,$AF,$FA,$B5,$AF,$FA,$BF
	.dc.b	$E6,$7F,$C7,$72,$69,$1F,$EE,$26,$4A,$E6,$3E,$33,$F0,$BE,$5E,$4F
	.dc.b	$C2,$6C,$FF,$8C,$7E,$B1,$1F,$E4,$A6,$CF,$F7,$E6,$CE,$E7,$44,$9D
	.dc.b	$36,$3E,$0E,$5F,$E1,$B9,$FA,$3E,$77,$D4,$23,$FE,$0A,$F8,$0F,$E1
	.dc.b	$84,$F2,$FF,$CB,$6D,$8D,$46,$08,$FF,$5E,$FA,$C7,$D6,$7F,$63,$6C
	.dc.b	$66,$30,$7F,$99,$BA,$FE,$DE,$5F,$B6,$DF,$04,$AF,$F0,$EC,$EA,$7F
	.dc.b	$81,$4E,$5E,$13,$FD,$1B,$EB,$35,$1F,$EF,$76,$AF,$15,$18,$6D,$78
	.dc.b	$FE,$92,$5F,$81,$B7,$FC,$2B,$5F,$41,$27,$CA,$D8,$D4,$60,$8F,$F9
	.dc.b	$CF,$EB,$3F,$DD,$91,$FE,$6F,$ED,$46,$27,$5F,$13,$FE,$80,$8C,$47
	.dc.b	$F9,$B9,$AF,$FA,$9E,$78,$CE,$C9,$5D,$7E,$B3,$FC,$67,$4A,$46,$1B
	.dc.b	$F3,$1F,$E8,$E8,$C1,$1C,$12,$BF,$C9,$F4,$F6,$3F,$D3,$ED,$8C,$C6
	.dc.b	$0E,$9D,$12,$3C,$8D,$9F,$F4,$B5,$FE,$D6,$72,$48,$CC,$F6,$5F,$09
	.dc.b	$E2,$69,$E4,$70,$7F,$99,$3E,$B3,$F7,$1E,$B5,$EB,$FF,$27,$D7,$FA
	.dc.b	$7E,$72,$D8,$CB,$FE,$B6,$87,$C2,$7E,$26,$38,$35,$D3,$5F,$F6,$AC
	.dc.b	$70,$E3,$FB,$DF,$82,$7E,$F2,$7F,$D9,$BB,$A3,$11,$F9,$5D,$3E,$E4
	.dc.b	$62,$BF,$DB,$7A,$FF,$A8,$EB,$8C,$E6,$B7,$D9,$FE,$E4,$BF,$F7,$0C
	.dc.b	$E6,$BF,$01,$7C,$5E,$C7,$E0,$FE,$B5,$F1,$9F,$E8,$AC,$91,$89,$FC
	.dc.b	$B3,$5F,$F6,$63,$5F,$F6,$0D,$F5,$C3,$C3,$EF,$F8,$F3,$F5,$B6,$BC
	.dc.b	$18,$DC,$7E,$14,$DC,$39,$5F,$0B,$F8,$F8,$F8,$71,$4A,$5F,$69,$FE
	.dc.b	$8A,$E2,$EC,$D8,$FF,$7D,$B5,$F8,$39,$3E,$BB,$FD,$7D,$1A,$E2,$FE
	.dc.b	$8C,$DB,$E7,$18,$8E,$33,$C1,$F9,$63,$EB,$BF,$B6,$BE,$5C,$BC,$EF
	.dc.b	$9E,$32,$F0,$6C,$6A,$39,$7C,$CF,$C7,$7A,$79,$16,$B7,$A7,$C1,$9C
	.dc.b	$9D,$DC,$42,$B9,$7D,$67,$F9,$16,$BF,$D8,$ED,$7F,$E0,$03,$5E,$81
	.dc.b	$C3,$FC,$3F,$E3,$1F,$19,$F6,$5B,$1F,$9C,$6B,$F1,$27,$B6,$5E,$D7
	.dc.b	$FC,$0D,$1E,$2F,$2B,$F8,$5C,$62,$3C,$71,$82,$3B,$21,$E6,$E0,$79
	.dc.b	$5F,$E4,$38,$FB,$ED,$7F,$3F,$1A,$F5,$3E,$1D,$7C,$4F,$98,$F0,$92
	.dc.b	$BE,$2B,$5F,$8A,$D7,$E2,$E0,$8F,$F7,$8D,$FE,$4B,$E5,$FF,$22,$9A
	.dc.b	$77,$AF,$AC,$7E,$66,$9F,$1D,$F0,$89,$1F,$59,$AF,$D6,$6B,$E2,$7C
	.dc.b	$34,$F9,$8F,$09,$27,$B9,$F6,$E7,$F1,$9B,$FF,$B3,$63,$EF,$9F,$AB
	.dc.b	$70,$E1,$AF,$E1,$1A,$FE,$45,$AF,$1F,$82,$FE,$8B,$18,$8F,$C9,$47
	.dc.b	$C6,$FB,$67,$89,$FB,$66,$CF,$ED,$9A,$FE,$B2,$BC,$7C,$1F,$A5,$FA
	.dc.b	$B7,$C5,$7F,$52,$6D,$7A,$83,$87,$DA,$D7,$F3,$AD,$7F,$54,$D7,$F8
	.dc.b	$63,$5E,$71,$FE,$51,$E6,$7A,$3F,$C6,$F9,$3F,$89,$B6,$33,$5F,$42
	.dc.b	$FC,$58,$7E,$F7,$EA,$BF,$D2,$13,$E7,$6B,$F4,$9A,$FE,$C9,$AF,$EF
	.dc.b	$DA,$FF,$21,$6A,$69,$8F,$CF,$BF,$1B,$1F,$82,$6B,$FC,$61,$B3,$E5
	.dc.b	$F7,$7F,$88,$7D,$5F,$FB,$D6,$BF,$12,$D7,$F8,$FB,$67,$D9,$E4,$F0
	.dc.b	$E6,$FF,$02,$CF,$E8,$6B,$EA,$35,$FC,$7B,$5F,$E1,$FF,$A5,$7E,$65
	.dc.b	$FC,$BB,$5F,$E5,$AD,$BF,$E4,$D2,$FA,$CF,$E4,$DF,$15,$FD,$4E,$FF
	.dc.b	$92,$35,$FE,$8C,$D4,$D2,$79,$1F,$E1,$C6,$C7,$F9,$CF,$F1,$27,$CA
	.dc.b	$FF,$59,$F7,$11,$89,$FD,$DC,$62,$7C,$77,$FD,$4B,$E0,$DF,$41,$49
	.dc.b	$7F,$B4,$FA,$DF,$EE,$DD,$EA,$DC,$FF,$58,$6B,$FC,$69,$BB,$7F,$47
	.dc.b	$AE,$3F,$56,$FB,$4F,$EC,$2D,$BF,$EC,$2D,$7F,$AB,$CF,$FB,$41,$CC
	.dc.b	$F0,$23,$E5,$F8,$5F,$EA,$52,$F1,$E9,$27,$F8,$6F,$CE,$BE,$53,$11
	.dc.b	$DA,$6E,$5F,$D7,$6B,$53,$FD,$D3,$F0,$1F,$ED,$93,$FD,$9B,$54,$63
	.dc.b	$B7,$FB,$F4,$63,$D0,$E7,$18,$DB,$FC,$0F,$87,$F8,$09,$B4,$73,$0F
	.dc.b	$AF,$FC,$43,$F1,$3D,$5F,$8E,$47,$F9,$3C,$69,$FE,$93,$F4,$3F,$58
	.dc.b	$9E,$C7,$F7,$59,$FE,$BF,$A7,$A6,$73,$5F,$E1,$07,$FC,$64,$F8,$7D
	.dc.b	$0F,$EF,$9F,$5D,$E5,$8F,$E3,$0D,$7F,$99,$B5,$FE,$E6,$D7,$FC,$00
	.dc.b	$D7,$FC,$44,$D4,$D3,$54,$FC,$57,$FC,$55,$AF,$F8,$55,$B5,$FE,$A5
	.dc.b	$6B,$CB,$4F,$F1,$5F,$CA,$3E,$FA,$92,$7F,$A6,$E3,$FE,$66,$7F,$C5
	.dc.b	$F3,$B6,$CE,$DF,$F1,$3C,$ED,$FE,$EB,$5F,$7D,$FB,$5C,$1F,$9A,$3E
	.dc.b	$3F,$C6,$D8,$FC,$DC,$57,$9C,$7F,$89,$FD,$4F,$F0,$BC,$7F,$91,$5A
	.dc.b	$9A,$4F,$94,$FF,$4D,$B6,$3F,$A5,$3C,$BE,$27,$FC,$79,$F8,$3F,$F9
	.dc.b	$65,$B3,$FD,$E9,$AF,$F9,$0A,$7F,$C3,$FB,$7E,$55,$FD,$99,$23,$FB
	.dc.b	$9A,$FB,$0C,$47,$C2,$8B,$C1,$FE,$43,$75,$1F,$E5,$AF,$B0,$FF,$20
	.dc.b	$BF,$E6,$06,$F3,$9C,$C7,$BF,$E8,$FF,$CC,$71,$A2,$7B,$BF,$1E,$3F
	.dc.b	$CE,$9F,$28,$FA,$CF,$A9,$7F,$E4,$D8,$C7,$C3,$FF,$6D,$6C,$7F,$82
	.dc.b	$1A,$FF,$82,$BE,$C3,$FE,$00,$BF,$E1,$06,$BF,$E6,$66,$E9,$FE,$64
	.dc.b	$9E,$2F,$F9,$9B,$E1,$1F,$A3,$FF,$B3,$6F,$FD,$4E,$D7,$FC,$F2,$D9
	.dc.b	$FF,$34,$5F,$4E,$0F,$F7,$24,$E5,$BF,$17,$FE,$8F,$8C,$5F,$FA,$26
	.dc.b	$B8,$3F,$EE,$5F,$B0,$73,$7F,$A8,$DB,$1F,$EF,$76,$BF,$E6,$6C,$91
	.dc.b	$88,$FF,$3A,$72,$3F,$1E,$DE,$9F,$E8,$28,$C6,$1F,$8B,$9C,$E9,$FE
	.dc.b	$FB,$C2,$E0,$D7,$EA,$7D,$8F,$D1,$E0,$34,$93,$FE,$3A,$B7,$E3,$2F
	.dc.b	$98,$FF,$58,$46,$1B,$19,$9F,$F2,$9E,$1D,$9F,$F2,$AD,$6D,$B3,$B7
	.dc.b	$FD,$93,$19,$7F,$D9,$73,$C5,$CC,$EF,$5F,$F5,$0D,$6D,$B1,$98,$FF
	.dc.b	$5D,$57,$19,$FC,$15,$FF,$9B,$6F,$FC,$AB,$B7,$AF,$E1,$FF,$DA,$D3
	.dc.b	$99,$FA,$4D,$7F,$D4,$9C,$FF,$A3,$F8,$1F,$D0,$92,$3F,$DE,$33,$C7
	.dc.b	$0F,$C9,$6B,$FE,$87,$9F,$D0,$35,$F2,$B5,$FA,$8D,$CB,$F5,$1F,$11
	.dc.b	$F5,$49,$3F,$A5,$9E,$1A,$7E,$FF,$CC,$FA,$7D,$9F,$D8,$AB,$E5,$B7
	.dc.b	$1F,$1E,$1F,$EC,$1F,$1B,$FC,$51,$1F,$A5,$FB,$2F,$F0,$1F,$C8,$FF
	.dc.b	$6A,$E1,$46,$32,$FF,$62,$5F,$0C,$EF,$B9,$FE,$EF,$AD,$68,$C3,$8F
	.dc.b	$A9,$FE,$CB,$CB,$C9,$8F,$1A,$D5,$F0,$BD,$C7,$FB,$C6,$BC,$53,$D2
	.dc.b	$15,$C6,$B7,$0A,$DD,$F7,$CB,$FD,$F5,$5A,$9F,$0E,$3F,$07,$D9,$D6
	.dc.b	$1E,$7F,$AE,$7C,$9E,$30,$46,$27,$C9,$8F,$C2,$BE,$31,$E4,$78,$4E
	.dc.b	$E3,$C9,$3E,$28,$C6,$3F,$0A,$B8,$E7,$FE,$48,$BE,$A7,$8F,$1E,$66
	.dc.b	$A7,$7F,$61,$FE,$82,$BF,$96,$DC,$7E,$AB,$6D,$F6,$2F,$B2,$7B,$7C
	.dc.b	$BE,$1F,$BD,$7C,$CB,$E0,$D8,$F1,$DF,$1A,$E1,$3F,$EB,$CD,$17,$FE
	.dc.b	$94,$8F,$F7,$EC,$62,$FF,$D5,$0D,$8E,$1E,$BF,$F1,$16,$F4,$6A,$31
	.dc.b	$1C,$9A,$FE,$41,$E1,$A7,$E2,$9B,$A2,$7D,$FF,$65,$FE,$6C,$E5,$F6
	.dc.b	$E2,$BD,$D6,$E3,$18,$EA,$73,$AB,$0E,$57,$F6,$8E,$13,$C9,$78,$DF
	.dc.b	$E5,$9C,$7E,$CF,$1F,$E2,$3C,$74,$7C,$E9,$D5,$73,$AD,$4E,$7D,$9D
	.dc.b	$CE,$79,$58,$63,$7F,$DE,$63,$C9,$FD,$B4,$EB,$0E,$47,$89,$C5,$FA
	.dc.b	$4F,$3E,$1C,$F7,$FC,$A6,$36,$7F,$67,$EA,$DC,$FF,$80,$B9,$8F,$7F
	.dc.b	$C7,$79,$BA,$FF,$84,$4E,$FA,$CD,$38,$6C,$AE,$1A,$FE,$D2,$B6,$FC
	.dc.b	$58,$CE,$1E,$28,$C1,$F7,$70,$EE,$E2,$11,$87,$DD,$7D,$E7,$81,$DF
	.dc.b	$0F,$72,$F8,$E7,$E1,$1E,$08,$C3,$E2,$CD,$F6,$6F,$B1,$D5,$F0,$BE
	.dc.b	$B9,$7F,$1F,$AD,$47,$F4,$C6,$C7,$F7,$B6,$BF,$97,$38,$57,$E0,$1C
	.dc.b	$5F,$82,$31,$1F,$C6,$A7,$8B,$63,$2F,$2A,$E8,$FF,$A7,$3D,$1F,$EF
	.dc.b	$71,$FE,$3D,$F2,$7D,$9E,$87,$E7,$36,$FF,$52,$56,$AF,$E5,$DF,$F2
	.dc.b	$8A,$FD,$0C,$79,$1A,$E7,$E0,$FC,$39,$A7,$33,$FA,$0D,$3F,$82,$DF
	.dc.b	$F8,$01,$B1,$D8,$FE,$8D,$F6,$4F,$7F,$E6,$F7,$CB,$E0,$FD,$39,$D4
	.dc.b	$60,$95,$FE,$08,$BE,$BA,$7F,$8F,$EB,$73,$9B,$E5,$87,$A2,$73,$3F
	.dc.b	$41,$AF,$F4,$86,$AF,$45,$E2,$BE,$D3,$F2,$F1,$87,$F1,$1F,$58,$F8
	.dc.b	$9C,$D6,$F5,$FC,$3E,$EA,$D4,$7F,$84,$9B,$8F,$F9,$05,$AF,$E1,$1A
	.dc.b	$FF,$94,$BE,$83,$F2,$DE,$B9,$FD,$4C,$3E,$CD,$F2,$9A,$99,$87,$F9
	.dc.b	$81,$DC,$EA,$30,$7E,$72,$B9,$56,$8F,$D2,$7D,$9B,$E5,$F1,$5E,$C8
	.dc.b	$C3,$FE,$2C,$6D,$7F,$96,$DA,$FE,$79,$AF,$E5,$FC,$0F,$E0,$3E,$AF
	.dc.b	$FA,$3D,$FE,$07,$F2,$CF,$9C,$FF,$28,$61,$FE,$97,$9D,$F5,$FF,$1E
	.dc.b	$E8,$FC,$DC,$70,$AD,$E3,$F8,$FC,$7A,$61,$FB,$87,$FD,$F5,$F5,$9E
	.dc.b	$0F,$84,$3C,$7E,$CF,$0C,$AE,$13,$A7,$F8,$0E,$9E,$5A,$D6,$1F,$9D
	.dc.b	$E8,$7C,$E7,$F5,$5B,$FF,$5E,$36,$7F,$CB,$BF,$66,$F2,$7F,$17,$39
	.dc.b	$DF,$FD,$35,$39,$BF,$F5,$23,$6F,$F7,$FE,$82,$31,$1F,$EA,$D6,$BF
	.dc.b	$E6,$26,$BF,$E8,$16,$C7,$1C,$3F,$85,$B6,$BF,$85,$D3,$66,$BF,$CB
	.dc.b	$63,$FD,$DB,$E4,$FC,$FF,$07,$E5,$32,$FE,$3B,$96,$9F,$47,$F1,$3F
	.dc.b	$F5,$1B,$7E,$0B,$EC,$7A,$FF,$82,$9C,$E1,$F6,$5B,$B7,$F6,$6C,$D1
	.dc.b	$8C,$3F,$C0,$5C,$5F,$87,$C3,$96,$3F,$B6,$FA,$4F,$E6,$78,$FF,$A9
	.dc.b	$DB,$7F,$D3,$7E,$C9,$F6,$7B,$F2,$FF,$A9,$DA,$73,$5F,$12,$31,$5F
	.dc.b	$CC,$9B,$19,$9D,$EB,$FC,$D2,$72,$D8,$CB,$FD,$0E,$32,$D8,$CB,$FC
	.dc.b	$EF,$83,$F3,$E7,$E7,$7E,$3A,$DF,$FB,$E6,$2C,$FF,$2D,$D7,$F5,$15
	.dc.b	$8C,$3E,$45,$7E,$D1,$BD,$BF,$A7,$FB,$0F,$D1,$33,$1F,$CF,$B9,$69
	.dc.b	$FA,$9F,$03,$F8,$BD,$7D,$36,$C7,$FA,$C3,$EC,$FF,$63,$39,$F1,$FF
	.dc.b	$BE,$F3,$A7,$35,$BF,$83,$FE,$33,$F6,$6F,$8E,$DD,$6B,$83,$FD,$B0
	.dc.b	$DC,$39,$5F,$0A,$DB,$FB,$BF,$22,$63,$FB,$6C,$3F,$70,$7F,$03,$DF
	.dc.b	$F9,$94,$7E,$F2,$77,$3F,$24,$DE,$9F,$22,$3F,$8A,$FD,$A3,$E6,$29
	.dc.b	$FC,$5A,$9F,$5B,$8C,$3F,$87,$56,$DB,$5B,$9D,$57,$F6,$3F,$9C,$7E
	.dc.b	$26,$FB,$6F,$F1,$1C,$61,$FE,$9F,$F6,$CF,$8A,$CD,$6F,$2F,$F0,$9F
	.dc.b	$8E,$AD,$47,$EE,$A3,$15,$FE,$08,$6C,$66,$77,$8F,$F8,$2A,$72,$D8
	.dc.b	$CB,$FE,$19,$8C,$B6,$32,$FF,$87,$7B,$7F,$43,$BF,$F8,$75,$B7,$FE
	.dc.b	$1E,$E9,$FE,$AF,$18,$AF,$F7,$BF,$87,$FE,$1C,$5C,$BF,$C3,$DF,$E3
	.dc.b	$F1,$98,$C4,$7F,$8A,$DF,$9C,$FF,$83,$4E,$3C,$1F,$D9,$3E,$D5,$F6
	.dc.b	$DF,$D9,$AB,$79,$FF,$8C,$7F,$33,$7C,$4F,$F6,$A6,$EB,$FD,$8F,$0F
	.dc.b	$E1,$D9,$7F,$A3,$E7,$2F,$78,$FE,$2E,$FF,$8C,$9A,$FF,$8F,$DA,$FF
	.dc.b	$A5,$DA,$9A,$4F,$13,$E1,$3C,$3C,$1F,$F4,$5F,$C8,$7F,$64,$C0,$FF
	.dc.b	$96,$7E,$DD,$F3,$19,$8C,$3F,$E4,$69,$4E,$9F,$E7,$A6,$23,$51,$9C
	.dc.b	$7F,$C9,$73,$93,$71,$98,$FF,$2D,$46,$5D,$C6,$5F,$F5,$A7,$94,$E3
	.dc.b	$FD,$D3,$7B,$BF,$CE,$1F,$35,$F5,$30,$FE,$E9,$5F,$BF,$51,$9B,$F1
	.dc.b	$B5,$F0,$B5,$FE,$DC,$D7,$FD,$62,$D8,$E7,$C5,$F0,$E3,$06,$F0,$7E
	.dc.b	$21,$F7,$9B,$BF,$CD,$FA,$0F,$F0,$AC,$FE,$0B,$0F,$97,$1F,$20,$8F
	.dc.b	$C6,$6E,$3F,$11,$AF,$C8,$D1,$1F,$E7,$AC,$3C,$39,$76,$3E,$E7,$DA
	.dc.b	$DC,$3F,$D1,$E7,$DA,$7E,$B9,$1F,$AA,$DC,$3E,$9B,$5F,$03,$DF,$E3
	.dc.b	$0F,$F8,$C1,$B3,$FE,$9C,$E6,$7D,$C3,$EE,$35,$7F,$CE,$A3,$15,$FE
	.dc.b	$06,$AD,$D7,$F8,$2A,$73,$87,$F8,$59,$AF,$ED,$DA,$FF,$BE,$67,$8D
	.dc.b	$73,$9C,$CF,$F0,$26,$E4,$C7,$FA,$35,$7F,$8E,$DB,$5F,$E8,$DE,$87
	.dc.b	$DC,$7F,$34,$AD,$78,$FF,$DF,$23,$15,$FD,$D5,$AF,$F7,$36,$BF,$ED
	.dc.b	$16,$A3,$FD,$09,$F1,$9F,$E7,$0F,$00,$DD,$B0,$CC,$3F,$CD,$93,$A6
	.dc.b	$CE,$6B,$FC,$AD,$3F,$C3,$31,$FF,$0F,$36,$BC,$1F,$02,$FB,$87,$CB
	.dc.b	$FF,$9A,$2B,$58,$7F,$8D,$DB,$B7,$FA,$67,$BF,$FD,$B9,$7C,$31,$FF
	.dc.b	$31,$57,$0C,$5F,$6F,$FD,$93,$1F,$F5,$B4,$63,$CB,$FF,$72,$7D,$C7
	.dc.b	$F8,$27,$0F,$1D,$EF,$54,$62,$BF,$DD,$53,$98,$F4,$D7,$1C,$18,$75
	.dc.b	$9F,$F5,$2C,$FF,$A1,$27,$2D,$93,$4C,$7E,$05,$E0,$AF,$F5,$6B,$6B
	.dc.b	$D3,$3F,$C2,$A7,$27,$FB,$F6,$75,$8F,$A2,$BC,$71,$89,$F3,$F6,$6D
	.dc.b	$FF,$56,$CF,$E0,$79,$3F,$CC,$AF,$FB,$ED,$AE,$4A,$EB,$D7,$7D,$87
	.dc.b	$FA,$6E,$7F,$CD,$B1,$8F,$81,$D4,$E6,$7F,$DF,$73,$97,$7F,$04,$CF
	.dc.b	$A3,$63,$31,$83,$FC,$F3,$AF,$DC,$E8,$FF,$27,$67,$F6,$1B,$5F,$4B
	.dc.b	$7F,$F1,$83,$FE,$D6,$FC,$E7,$FD,$28,$FF,$B6,$23,$2F,$E1,$1B,$87
	.dc.b	$C9,$C3,$B4,$F1,$F7,$FA,$C2,$FF,$2F,$3E,$46,$BF,$AB,$3F,$3A,$E1
	.dc.b	$C4,$71,$E7,$F9,$EB,$85,$F1,$BD,$9B,$5F,$82,$8F,$0D,$F6,$C7,$CD
	.dc.b	$5A,$AF,$CE,$B6,$33,$5C,$30,$F8,$BD,$1F,$ED,$1F,$1A,$F1,$7F,$6E
	.dc.b	$C7,$FD,$94,$D7,$FD,$9C,$FF,$BC,$DC,$C7,$02,$3C,$3E,$23,$AF,$CC
	.dc.b	$BE,$76,$92,$7F,$DA,$D0,$F8,$3C,$47,$FB,$6E,$38,$3A,$9E,$73,$E8
	.dc.b	$FC,$43,$FD,$ED,$3F,$EB,$AC,$91,$88,$FA,$07,$FA,$F5,$C3,$F4,$E3
	.dc.b	$7E,$FF,$E7,$97,$9D,$FE,$C2,$9F,$F7,$53,$5F,$F7,$73,$77,$FF,$5B
	.dc.b	$5F,$77,$FD,$71,$D4,$F9,$1E,$8D,$BF,$33,$5F,$2E,$1F,$CF,$E7,$64
	.dc.b	$8D,$B5,$DF,$A0,$F3,$FE,$CB,$63,$FE,$00,$35,$FD,$57,$57,$F8,$5B
	.dc.b	$0F,$06,$3F,$EC,$DE,$77,$D7,$7C,$16,$DF,$CB,$6B,$E6,$6B,$F2,$5A
	.dc.b	$FF,$88,$9A,$FF,$BA,$DB,$9F,$F2,$29,$F9,$CB,$E5,$FF,$22,$DA,$FE
	.dc.b	$A1,$7F,$EE,$18,$C4,$7A,$8E,$95,$F9,$8E,$67,$D1,$7D,$96,$CF,$E0
	.dc.b	$DA,$FC,$E6,$BF,$81,$6B,$FE,$3A,$6A,$AF,$EA,$33,$D4,$FC,$37,$E0
	.dc.b	$EF,$99,$FD,$4B,$6B,$D3,$AF,$F9,$12,$31,$1E,$9B,$E6,$E8,$ED,$F8
	.dc.b	$23,$E5,$7F,$24,$D8,$FC,$E3,$5F,$C3,$B5,$FC,$D3,$5F,$F2,$9B,$5E
	.dc.b	$B1,$D0,$FC,$F7,$DC,$BC,$8F,$E0,$2D,$9F,$25,$A3,$11,$E9,$BF,$16
	.dc.b	$DF,$D2,$E9,$F4,$7F,$D0,$2B,$F6,$4D,$7F,$3A,$D7,$F6,$0D,$7F,$CD
	.dc.b	$2D,$7E,$83,$63,$AF,$8F,$FB,$58,$C7,$E1,$3F,$94,$BF,$13,$75,$FF
	.dc.b	$39,$B7,$0E,$77,$FC,$7B,$F0,$A7,$CE,$7E,$FD,$B3,$FC,$25,$AF,$ED
	.dc.b	$1A,$FF,$07,$6B,$FE,$80,$6B,$F8,$76,$E4,$68,$FB,$CF,$11,$27,$E7
	.dc.b	$C7,$83,$2F,$EE,$DE,$03,$F1,$A7,$D7,$FF,$36,$F4,$DF,$07,$4F,$EB
	.dc.b	$B5,$B9,$FE,$77,$F8,$67,$E6,$B4,$34,$9D,$1F,$C7,$FE,$9D,$F2,$BF
	.dc.b	$D0,$9B,$B7,$E0,$9A,$FE,$65,$FC,$DE,$BF,$E0,$B8,$7D,$B3,$E1,$BF
	.dc.b	$C0,$AD,$8F,$F0,$4B,$5F,$E6,$67,$08,$D9,$1D,$B5,$FE,$4B,$F1,$FF
	.dc.b	$6F,$33,$4F,$BD,$1C,$CF,$CC,$FD,$BF,$F4,$B6,$F9,$5F,$D7,$A3,$11
	.dc.b	$FB,$35,$F4,$39,$99,$3C,$0E,$71,$FD,$9B,$E5,$BF,$B5,$75,$7F,$B0
	.dc.b	$1F,$B0,$DE,$8D,$7A,$32,$FF,$57,$5A,$7F,$86,$37,$33,$D2,$3F,$CE
	.dc.b	$1F,$81,$BE,$77,$FC,$33,$0F,$B7,$F4,$F0,$FD,$16,$E7,$FE,$0D,$BE
	.dc.b	$15,$FC,$34,$38,$7F,$95,$1C,$F5,$EB,$5C,$A3,$C3,$86,$27,$FC,$0B
	.dc.b	$F7,$1F,$E0,$1C,$7F,$80,$37,$1F,$F0,$E6,$1C,$7C,$AF,$E7,$9D,$7F
	.dc.b	$1F,$E6,$66,$7F,$C7,$E7,$2D,$F2,$FF,$D6,$AF,$C6,$9D,$BF,$E7,$48
	.dc.b	$CB,$FC,$A7,$EB,$7F,$C6,$78,$FF,$BC,$1B,$1F,$CD,$DB,$C0,$7F,$10
	.dc.b	$60,$9E,$E7,$F8,$81,$FC,$4F,$DB,$3D,$EF,$F2,$EB,$6F,$FC,$C2,$D7
	.dc.b	$F3,$2D,$7F,$C1,$AD,$7F,$C7,$CD,$7F,$D5,$4D,$4E,$6F,$F0,$3F,$6F
	.dc.b	$F8,$D2,$4F,$D9,$9F,$07,$07,$F9,$47,$F1,$1F,$E5,$8D,$FA,$7F,$F2
	.dc.b	$8C,$1B,$E4,$4B,$FA,$13,$96,$CE,$5F,$F5,$0F,$E7,$5F,$92,$7F,$2F
	.dc.b	$E0,$FE,$7C,$FF,$2E,$C6,$63,$7D,$39,$76,$7D,$D8,$D6,$3F,$C1,$7D
	.dc.b	$D7,$D5,$FF,$A2,$1B,$5F,$E8,$C6,$BF,$C1,$1A,$FF,$8E,$1A,$FF,$9A
	.dc.b	$5A,$FF,$B5,$5A,$9C,$EC,$FA,$8F,$8F,$C7,$FE,$DB,$8C,$6D,$FE,$60
	.dc.b	$9C,$DF,$3C,$3F,$7D,$EA,$7E,$FF,$DE,$93,$FB,$A8,$CB,$FB,$C8,$CB
	.dc.b	$FD,$F7,$5F,$E4,$B9,$77,$C3,$E8,$4F,$F8,$05,$7E,$1B,$E1,$3C,$5C
	.dc.b	$12,$7F,$B8,$CE,$9B,$39,$79,$7D,$FF,$F9,$5E,$4D,$24,$7F,$98,$E3
	.dc.b	$FB,$BD,$F8,$FD,$6F,$80,$7C,$1A,$F8,$5F,$87,$CB,$F0,$BA,$69,$0F
	.dc.b	$53,$97,$D9,$E6,$F8,$C8,$F8,$5F,$37,$9F,$F3,$C9,$1E,$E8,$FA,$5F
	.dc.b	$2F,$EB,$24,$7A,$5F,$77,$DA,$FD,$72,$47,$C6,$8A,$FC,$D8,$BC,$BF
	.dc.b	$E1,$5D,$15,$8D,$3E,$A5,$5B,$FF,$1F,$56,$9F,$A3,$F2,$9F,$34,$8F
	.dc.b	$CB,$7E,$87,$CA,$FE,$24,$91,$C0,$FE,$3D,$F9,$DF,$FA,$61,$FF,$48
	.dc.b	$F8,$68,$C4,$7F,$B0,$1A,$84,$6F,$93,$FD,$87,$3A,$9C,$B6,$75,$1F
	.dc.b	$EE,$28,$D7,$43,$EA,$BF,$D8,$9A,$7E,$85,$B5,$FE,$5D,$6D,$F2,$8F
	.dc.b	$F3,$73,$E7,$E7,$FF,$80,$1F,$19,$E5,$8C,$4F,$9F,$E0,$BF,$DD,$31
	.dc.b	$88,$46,$32,$FF,$04,$B6,$F9,$CF,$A9,$B5,$C6,$78,$77,$62,$FF,$CD
	.dc.b	$31,$C6,$FC,$18,$78,$E7,$2D,$AE,$18,$4A,$D9,$DC,$B8,$7D,$39,$F0
	.dc.b	$FD,$9B,$F0,$AA,$4A,$DC,$7F,$8E,$5B,$19,$7D,$F8,$7C,$59,$E5,$1C
	.dc.b	$3D,$CD,$67,$F1,$A7,$33,$C0,$FB,$EE,$23,$53,$F1,$A7,$FA,$D7,$A0
	.dc.b	$F9,$5F,$5C,$61,$BE,$42,$75,$3D,$7D,$BF,$F3,$AF,$3F,$3A,$D5,$F0
	.dc.b	$BD,$E2,$CF,$F2,$58,$F5,$39,$7B,$EC,$79,$5F,$69,$FF,$76,$E3,$D1
	.dc.b	$FF,$1E,$B5,$CC,$6D,$EB,$0F,$B7,$7C,$D7,$F9,$12,$7F,$BC,$46,$1E
	.dc.b	$6D,$7B,$B5,$E6,$EF,$2F,$54,$7E,$AE,$BB,$57,$E0,$B4,$7B,$9F,$E8
	.dc.b	$2E,$6F,$93,$8F,$1D,$79,$DE,$0E,$15,$BC,$7E,$C1,$E6,$8F,$C5,$E8
	.dc.b	$EA,$63,$86,$D7,$DF,$E4,$F9,$5D,$58,$DB,$F5,$38,$71,$C3,$81,$2F
	.dc.b	$84,$2B,$86,$7F,$B3,$8E,$13,$AC,$72,$FF,$0A,$8C,$E1,$DA,$F3,$8F
	.dc.b	$DB,$BB,$F9,$3F,$D2,$CB,$F4,$AD,$C7,$D0,$FF,$15,$8C,$E7,$C0,$8F
	.dc.b	$09,$A7,$08,$B8,$FF,$06,$AE,$53,$9A,$E6,$47,$B9,$C3,$3F,$E3,$1E
	.dc.b	$DF,$CA,$EF,$FF,$68,$75,$FF,$A0,$AF,$FC,$99,$F2,$5F,$0A,$73,$5B
	.dc.b	$9F,$E3,$13,$F1,$B0,$FC,$EE,$A8,$C4,$6B,$5F,$F4,$E3,$6B,$FD,$2F
	.dc.b	$5B,$9C,$CF,$F3,$6A,$E2,$D8,$CC,$EE,$BF,$D3,$FD,$BF,$94,$7F,$DD
	.dc.b	$B9,$FC,$2C,$BF,$DD,$4F,$CF,$6E,$1E,$91,$5F,$B5,$DB,$C2,$DB,$EB
	.dc.b	$7F,$5E,$BF,$73,$39,$7F,$76,$72,$87,$A9,$D6,$7F,$80,$36,$1F,$33
	.dc.b	$62,$5F,$A1,$D4,$62,$D8,$FA,$A7,$0F,$B2,$6D,$7F,$8C,$6F,$AB,$E8
	.dc.b	$8F,$5C,$62,$FC,$F8,$7F,$13,$9C,$FC,$E7,$94,$F3,$7A,$97,$0F,$96
	.dc.b	$F2,$23,$F1,$1E,$BF,$17,$ED,$3D,$DE,$75,$9F,$A7,$17,$1F,$5C,$79
	.dc.b	$E7,$19,$FD,$67,$8C,$7D,$AF,$97,$B5,$C0,$F9,$CF,$B0,$91,$F5,$9F
	.dc.b	$7F,$A3,$F3,$FE,$4F,$E8,$36,$7F,$0A,$DA,$E9,$AF,$F3,$58,$C6,$3F
	.dc.b	$A2,$F3,$3E,$D7,$33,$CB,$B1,$86,$E3,$FE,$0D,$AF,$D4,$C6,$31,$EB
	.dc.b	$7C,$6B,$F2,$BF,$05,$F8,$33,$F2,$5F,$24,$F9,$4F,$EC,$D3,$A3,$A1
	.dc.b	$FE,$19,$8D,$79,$1F,$A6,$6C,$7F,$C0,$06,$E1,$F9,$46,$BF,$D4,$1A
	.dc.b	$F3,$AE,$EE,$DA,$ED,$AF,$F8,$57,$D6,$7D,$77,$C4,$F4,$7F,$55,$D6
	.dc.b	$9C,$E3,$E8,$BF,$F1,$15,$6A,$73,$1F,$12,$31,$7C,$EB,$F7,$3F,$73
	.dc.b	$FA,$7F,$88,$FD,$43,$63,$27,$EF,$A3,$2D,$3C,$1C,$7F,$E3,$5F,$4F
	.dc.b	$EC,$9F,$3B,$47,$9F,$8A,$F3,$6D,$F4,$63,$E6,$4E,$A7,$E6,$37,$4F
	.dc.b	$8D,$3A,$24,$E9,$B1,$FA,$EF,$4F,$FC,$19,$3F,$E9,$0F,$3B,$F8,$24
	.dc.b	$3E,$46,$C4,$9D,$1F,$E9,$68,$C5,$7F,$BD,$FE,$CB,$F2,$7E,$57,$D0
	.dc.b	$F8,$04,$62,$FF,$E0,$03,$6B,$84,$7C,$4E,$7F,$F0,$73,$F5,$7C,$57
	.dc.b	$D8,$F9,$AB,$F4,$7F,$61,$FD,$CF,$B3,$FB,$6D,$6B,$5F,$E3,$93,$9C
	.dc.b	$3F,$C3,$7F,$60,$E6,$FE,$59,$7B,$9F,$C1,$62,$AD,$47,$87,$2F,$06
	.dc.b	$5F,$DE,$AB,$58,$72,$9C,$D7,$68,$C6,$5F,$B4,$F2,$7F,$6B,$E7,$5A
	.dc.b	$F7,$37,$0F,$D8,$B7,$1F,$EB,$AF,$F3,$5F,$AF,$FF,$6E,$57,$E2,$1B
	.dc.b	$5F,$C1,$17,$87,$BF,$0E,$56,$DE,$51,$8A,$FE,$17,$E6,$BE,$57,$FA
	.dc.b	$13,$6F,$F1,$F3,$F6,$1F,$CB,$C6,$63,$F8,$F7,$57,$FA,$5F,$93,$F7
	.dc.b	$D1,$B3,$F7,$85,$3F,$78,$53,$F8,$31,$44,$73,$AF,$E0,$D1,$98,$C3
	.dc.b	$FD,$13,$F1,$2F,$F7,$4C,$FE,$A9,$B1,$FC,$99,$FF,$71,$C6,$7D,$FF
	.dc.b	$EB,$DF,$22,$F9,$AF,$ED,$2D,$9F,$EE,$AD,$BF,$EA,$2D,$7F,$CC,$7E
	.dc.b	$BB,$F1,$6F,$C6,$47,$F7,$49,$FE,$D7,$5F,$C2,$EB,$5F,$21,$E0,$F5
	.dc.b	$3F,$B7,$1F,$E6,$D9,$FE,$02,$DB,$FE,$7B,$F2,$AF,$8C,$DF,$AD,$FD
	.dc.b	$D6,$31,$7F,$E0,$A6,$D7,$F6,$66,$BF,$E0,$06,$BF,$EC,$BF,$1D,$18
	.dc.b	$8F,$F0,$5B,$63,$37,$FE,$0C,$8C,$7C,$B7,$DF,$E9,$FF,$2C,$D2,$31
	.dc.b	$87,$5F,$7F,$FC,$A8,$FF,$1C,$8C,$E3,$FE,$18,$9C,$BF,$DF,$1A,$FF
	.dc.b	$85,$9A,$F3,$EA,$FF,$0E,$C6,$23,$94,$60,$95,$FE,$2C,$FA,$8F,$E6
	.dc.b	$F8,$F9,$9B,$1F,$07,$E6,$5E,$C7,$F8,$7E,$73,$87,$F8,$B2,$73,$5F
	.dc.b	$E3,$66,$C7,$F8,$49,$AF,$F8,$B1,$AF,$5A,$FF,$19,$C6,$23,$A4,$60
	.dc.b	$FF,$1B,$FC,$CF,$DF,$C3,$E5,$B6,$FE,$B7,$CD,$7F,$A2,$1F,$E8,$B1
	.dc.b	$9B,$FF,$20,$CE,$5F,$F1,$2B,$5F,$F1,$FB,$5E,$73,$FE,$46,$8C,$47
	.dc.b	$28,$C1,$0F,$F2,$AF,$D4,$7F,$9A,$5F,$C4,$36,$3F,$CA,$3E,$A3,$EA
	.dc.b	$FF,$C9,$13,$9E,$04,$E6,$FF,$CB,$ED,$BF,$F1,$DB,$5F,$F2,$AB,$5F
	.dc.b	$CB,$35,$EF,$7F,$E5,$E8,$C5,$76,$8F,$F3,$0F,$91,$FB,$EA,$FF,$12
	.dc.b	$B7,$3F,$CB,$FC,$EB,$E4,$FF,$CB,$D3,$9C,$BF,$49,$39,$7F,$CE,$AD
	.dc.b	$BF,$F2,$8B,$5F,$F3,$5B,$5E,$35,$FE,$72,$8C,$47,$F9,$D9,$B8,$9B
	.dc.b	$F0,$6F,$80,$FF,$38,$37,$0F,$F0,$19,$FC,$32,$91,$9B,$FF,$20,$CF
	.dc.b	$FB,$45,$FF,$3E,$B5,$FD,$F3,$5F,$F4,$3B,$65,$F6,$0F,$92,$FF,$40
	.dc.b	$36,$7F,$C2,$6B,$E4,$73,$32,$72,$F0,$8C,$5F,$F9,$53,$57,$CF,$FF
	.dc.b	$A2,$DB,$7F,$E1,$E5,$F8,$9E,$66,$4E,$5E,$11,$8A,$FF,$35,$70,$BE
	.dc.b	$8D,$F1,$DF,$E3,$A6,$E1,$FE,$C0,$FB,$9C,$4C,$8F,$F1,$E5,$71,$8F
	.dc.b	$F2,$0F,$C5,$2B,$58,$76,$8C,$5F,$FA,$F9,$B1,$FE,$A8,$BF,$F6,$35
	.dc.b	$7F,$AF,$27,$33,$FE,$CA,$FA,$57,$DD,$FF,$93,$9B,$87,$FB,$53,$F0
	.dc.b	$2C,$4C,$8F,$2C,$EA,$BC,$55,$97,$FD,$14,$09,$C7,$FE,$BA,$7F,$D7
	.dc.b	$6D,$74,$7E,$20,$C1,$4F,$F6,$0F,$D4,$7E,$07,$E0,$78,$36,$3A,$35
	.dc.b	$DF,$1F,$F9,$FD,$CF,$CE,$FF,$AF,$9B,$C7,$CA,$31,$1D,$9A,$FA,$77
	.dc.b	$67,$FC,$CA,$3E,$0F,$7F,$A2,$DA,$EE,$B5,$5F,$57,$C7,$D4,$E6,$EC
	.dc.b	$FF,$10,$AD,$DF,$28,$C4,$EB,$0F,$F7,$F6,$17,$83,$FC,$B9,$5F,$7F
	.dc.b	$E0,$3E,$55,$27,$33,$F8,$16,$BF,$3A,$3F,$29,$F2,$9D,$79,$4D,$3B
	.dc.b	$5D,$AF,$C9,$CE,$9B,$3A,$7F,$D0,$7C,$0F,$5F,$DB,$D7,$FE,$AF,$6C
	.dc.b	$66,$77,$EB,$E2,$64,$72,$77,$1C,$BC,$E7,$CC,$52,$57,$FA,$4A,$1F
	.dc.b	$41,$88,$FF,$67,$C7,$07,$53,$CE,$7C,$BF,$4E,$FA,$8F,$C9,$66,$8C
	.dc.b	$5F,$E5,$E3,$13,$A9,$CC,$FF,$BF,$BD,$87,$B5,$F0,$1B,$1F,$EA,$66
	.dc.b	$BF,$EE,$29,$E3,$CB,$FE,$14,$EB,$FF,$1E,$4F,$76,$BF,$A4,$18,$FF
	.dc.b	$2C,$F5,$FC,$F3,$A9,$FD,$4C,$E5,$E5,$3A,$23,$FD,$1B,$C2,$FA,$BE
	.dc.b	$37,$F2,$08,$FF,$5D,$B5,$FF,$7B,$DF,$8B,$0F,$F1,$4F,$33,$E4,$FE
	.dc.b	$3B,$6B,$FB,$3C,$60,$8E,$08,$7F,$6A,$7E,$7E,$4F,$85,$FC,$3B,$67
	.dc.b	$FD,$9A,$D5,$7E,$2A,$FC,$6F,$43,$E2,$BE,$CB,$6B,$EC,$35,$F9,$57
	.dc.b	$FE,$C3,$9D,$92,$36,$D7,$FC,$2B,$F3,$5F,$9A,$F3,$CD,$3C,$33,$F9
	.dc.b	$EF,$41,$F3,$BF,$8A,$6D,$FE,$25,$AF,$89,$AF,$D8,$6B,$FE,$5F,$6B
	.dc.b	$F8,$06,$E0,$6F,$A5,$FE,$A1,$9F,$F7,$9B,$5F,$8F,$E1,$FF,$92,$7E
	.dc.b	$03,$CD,$87,$ED,$3D,$6F,$ED,$55,$F9,$D6,$BF,$E5,$DF,$1F,$F4,$71
	.dc.b	$88,$FF,$2F,$46,$08,$F1,$7F,$53,$EB,$FE,$0E,$3C,$27,$E6,$1C,$70
	.dc.b	$FF,$9C,$9B,$8F,$F9,$C6,$BF,$83,$FB,$0F,$6F,$F6,$6D,$C3,$FD,$11
	.dc.b	$18,$23,$82,$1F,$E8,$B7,$8F,$9F,$D0,$91,$FD,$4F,$1F,$F4,$1D,$75
	.dc.b	$8F,$3F,$82,$FA,$87,$C9,$FF,$1A,$F7,$5F,$21,$4F,$E7,$35,$B9,$FE
	.dc.b	$3F,$EF,$FE,$3F,$0F,$D9,$B5,$FF,$4D,$35,$FF,$58,$73,$7F,$10,$F9
	.dc.b	$97,$C7,$FF,$28,$6C,$F9,$F8,$7F,$D4,$51,$88,$F9,$BD,$08,$FC,$AC
	.dc.b	$6C,$FE,$E1,$F5,$0F,$95,$FE,$D6,$D8,$FE,$E0,$D7,$FB,$A3,$5F,$E2
	.dc.b	$0D,$7F,$BA,$B5,$EB,$1D,$0F,$E5,$3E,$3F,$DC,$8F,$E2,$ED,$BF,$E4
	.dc.b	$BD,$FF,$D8,$25,$7E,$1E,$3F,$F5,$AF,$D4,$7F,$3A,$9F,$55,$A3,$11
	.dc.b	$F0,$1B,$3C,$3E,$A3,$FA,$7C,$63,$4F,$F6,$43,$6B,$F9,$4B,$6F,$A5
	.dc.b	$7F,$56,$FA,$2F,$EB,$95,$F3,$1F,$C2,$61,$FE,$C8,$6E,$1F,$ED,$37
	.dc.b	$71,$9F,$47,$97,$4C,$BF,$4D,$6A,$FF,$C3,$BF,$50,$FA,$8F,$85,$F4
	.dc.b	$FE,$5B,$36,$75,$1F,$6C,$38,$7F,$C5,$B1,$C4,$E1,$3D,$3D,$0F,$EE
	.dc.b	$3F,$51,$FD,$BE,$BF,$0E,$DD,$3F,$C1,$D1,$E0,$9F,$F0,$77,$92,$EC
	.dc.b	$F8,$9F,$15,$2B,$FC,$A7,$39,$6C,$E5,$F9,$13,$CA,$3F,$C0,$3F,$47
	.dc.b	$FE,$1C,$D7,$FD,$F4,$DA,$FE,$B6,$DE,$23,$F8,$73,$04,$F6,$F9,$F3
	.dc.b	$7E,$27,$EA,$1E,$F7,$F9,$19,$B8,$7F,$92,$DA,$FF,$93,$DA,$FF,$6B
	.dc.b	$6B,$FE,$50,$6B,$FE,$57,$6A,$73,$83,$8F,$FD,$69,$3F,$7B,$E1,$FF
	.dc.b	$14,$79,$6F,$FC,$71,$F5,$1F,$CB,$38,$7F,$C7,$9E,$8A,$3F,$D2,$D7
	.dc.b	$FE,$7E,$6B,$FE,$88,$78,$6D,$FE,$51,$6C,$7F,$B5,$9D,$D7,$46,$CE
	.dc.b	$FC,$8F,$F2,$47,$CF,$79,$EF,$F2,$7E,$0F,$E7,$CC,$E9,$23,$24,$3F
	.dc.b	$D3,$33,$C5,$FF,$3D,$FB,$8F,$A2,$FF,$3A,$B6,$3F,$CF,$4D,$7F,$CF
	.dc.b	$CD,$7F,$C3,$4D,$7F,$CF,$ED,$7F,$D0,$CD,$7F,$C9,$51,$FE,$34,$F9
	.dc.b	$DF,$83,$C0,$69,$F1,$27,$67,$EC,$3D,$3F,$DE,$7B,$B2,$30,$43,$F7
	.dc.b	$CE,$5F,$E0,$11,$FE,$58,$9F,$97,$1F,$E3,$05,$F2,$59,$35,$F1,$52
	.dc.b	$7F,$8B,$CE,$9B,$39,$7F,$9B,$C3,$EA,$3D,$84,$8F,$F2,$74,$65,$BE
	.dc.b	$A7,$8C,$8F,$69,$F3,$7D,$14,$8F,$0B,$A7,$BB,$96,$9E,$69,$F7,$79
	.dc.b	$DF,$34,$91,$E2,$7E,$87,$A1,$F0,$C9,$1E,$E8,$F8,$DF,$41,$F5,$49
	.dc.b	$1E,$A7,$EC,$7D,$07,$CB,$24,$7C,$68,$AF,$C4,$8B,$D1,$FE,$A1,$E0
	.dc.b	$FB,$D5,$8E,$7F,$F6,$54,$7E,$06,$B1,$7F,$56,$BE,$C7,$CD,$7E,$1C
	.dc.b	$91,$F3,$1F,$C5,$7C,$B7,$DD,$24,$7D,$88,$FF,$57,$7D,$47,$FA,$1E
	.dc.b	$3F,$DD,$0D,$7F,$D5,$0D,$7F,$D8,$8D,$5D,$88,$C3,$79,$7F,$D6,$B1
	.dc.b	$83,$86,$4F,$A1,$FF,$5B,$EB,$F0,$5B,$AF,$C0,$6E,$3D,$AF,$CD,$5F
	.dc.b	$C3,$1F,$F0,$0C,$3E,$05,$F1,$9E,$6E,$8F,$F7,$6C,$62,$79,$46,$27
	.dc.b	$D7,$DB,$67,$27,$FB,$E2,$39,$34,$C9,$DD,$E2,$E7,$5F,$85,$5C,$DB
	.dc.b	$19,$C3,$CF,$87,$8E,$7E,$46,$1B,$9C,$E1,$E0,$BE,$C7,$82,$F8,$4F
	.dc.b	$F8,$8D,$E5,$F5,$0F,$A4,$F0,$36,$BC,$8D,$79,$E2,$87,$BB,$C2,$A5
	.dc.b	$70,$79,$C6,$23,$8C,$F3,$BF,$0D,$6A,$7E,$31,$FD,$73,$D2,$7C,$FE
	.dc.b	$3C,$AF,$F7,$7B,$63,$33,$EE,$F7,$1F,$00,$E9,$FA,$93,$CE,$F8,$11
	.dc.b	$E1,$3B,$70,$7F,$B2,$70,$EE,$FF,$B2,$CE,$45,$72,$EF,$EA,$1F,$33
	.dc.b	$B8,$FE,$EB,$18,$78,$35,$E1,$1B,$94,$EF,$4F,$C8,$57,$2A,$FC,$56
	.dc.b	$5F,$E5,$9F,$5D,$CF,$F1,$70,$E1,$C6,$C2,$65,$D3,$1D,$1C,$BD,$DF
	.dc.b	$F4,$46,$18,$AF,$BB,$B3,$AF,$AC,$61,$BA,$FE,$46,$5D,$98,$CF,$C1
	.dc.b	$CB,$F4,$6F,$18,$7E,$FA,$FA,$63,$F9,$F6,$C7,$83,$0D,$DF,$1A,$D5
	.dc.b	$E2,$F7,$8F,$86,$F1,$3C,$3E,$A3,$E8,$DF,$EA,$DB,$8F,$E6,$1F,$89
	.dc.b	$19,$D3,$81,$1E,$13,$4E,$11,$70,$FD,$85,$72,$9C,$D7,$32,$3D,$CE
	.dc.b	$19,$FF,$13,$F7,$7E,$5F,$5F,$F1,$76,$D7,$C8,$F9,$2F,$87,$39,$8F
	.dc.b	$F4,$A6,$C9,$F8,$F8,$7D,$D3,$F9,$4B,$87,$96,$9F,$E5,$66,$D7,$E6
	.dc.b	$6B,$94,$E6,$78,$B6,$33,$3B,$AE,$94,$EE,$FE,$45,$3F,$E8,$EC,$BE
	.dc.b	$16,$3F,$D4,$5F,$9E,$DA,$F6,$9C,$63,$E9,$BE,$9A,$FE,$5F,$AF,$8D
	.dc.b	$B7,$FA,$F6,$CF,$BC,$E5,$EA,$74,$8F,$E6,$ED,$B7,$D4,$58,$97,$E7
	.dc.b	$75,$18,$BF,$F6,$1D,$7F,$9E,$6B,$47,$B6,$75,$8F,$B5,$F3,$4F,$69
	.dc.b	$F3,$36,$BC,$D8,$7F,$4A,$AD,$FC,$DF,$94,$F3,$7A,$D7,$0F,$94,$F4
	.dc.b	$23,$D1,$EF,$EB,$7B,$4F,$77,$9D,$67,$E9,$45,$C7,$D7,$37,$3F,$A8
	.dc.b	$F0,$AF,$B1,$F5,$16,$BE,$0B,$EC,$24,$7B,$1F,$7F,$A1,$F3,$7E,$4F
	.dc.b	$E7,$36,$7F,$04,$DA,$F9,$6D,$C3,$FB,$A7,$9B,$F5,$FC,$87,$97,$63
	.dc.b	$0D,$C3,$FD,$B3,$3A,$6C,$E9,$FC,$87,$C1,$FD,$F3,$F1,$BF,$24,$F9
	.dc.b	$4F,$EB,$53,$A3,$FD,$F9,$19,$F1,$FF,$40,$D8,$FE,$78,$DD,$3F,$1C
	.dc.b	$D7,$F5,$CD,$7F,$96,$35,$34,$C7,$4B,$D5,$7D,$47,$C1,$F4,$7F,$47
	.dc.b	$39,$8C,$61,$FB,$EA,$DB,$F0,$AB,$71,$8C,$1B,$FF,$87,$E7,$F4,$1F
	.dc.b	$4B,$FA,$1F,$88,$B8,$7E,$D5,$B7,$F9,$E6,$DF,$ED,$DF,$A9,$E9,$CA
	.dc.b	$F8,$FB,$3D,$1C,$61,$E0,$9F,$83,$39,$7E,$4B,$63,$E4,$B5,$F9,$2D
	.dc.b	$78,$CF,$D2,$8C,$47,$CA,$6E,$DF,$6E,$7F,$82,$7A,$DF,$E0,$2B,$FF
	.dc.b	$43,$F9,$DF,$C0,$61,$F2,$36,$24,$E8,$FF,$46,$C6,$31,$F5,$1F,$E5
	.dc.b	$0F,$81,$7D,$27,$E5,$7C,$BF,$AB,$18,$BF,$C9,$C6,$27,$51,$FE,$FC
	.dc.b	$6D,$74,$F1,$5F,$4D,$E7,$8E,$FE,$23,$EA,$3F,$B7,$C6,$E7,$5C,$7F
	.dc.b	$E7,$99,$CB,$FD,$C2,$5F,$50,$E5,$FE,$57,$8F,$6D,$FF,$96,$D6,$A7
	.dc.b	$38,$F4,$D3,$FA,$36,$1C,$5F,$F0,$26,$3F,$DF,$31,$E5,$18,$37,$7C
	.dc.b	$B1,$FD,$C7,$95,$FB,$9F,$3A,$D7,$ED,$9B,$9F,$BB,$BF,$FC,$07,$3F
	.dc.b	$BA,$AE,$0F,$59,$FE,$95,$F5,$0F,$0B,$F9,$D3,$67,$5C,$08,$78,$7A
	.dc.b	$A9,$5A,$9E,$17,$F3,$9D,$C6,$67,$FA,$37,$9A,$F9,$3A,$7F,$08,$BF
	.dc.b	$BC,$7D,$68,$D5,$7F,$08,$8C,$9F,$C0,$8A,$7F,$02,$29,$FC,$08,$BD
	.dc.b	$5F,$E3,$1E,$5F,$E1,$F1,$82,$3F,$D7,$FE,$A3,$FD,$CB,$5F,$01,$B3
	.dc.b	$F8,$87,$FD,$C1,$19,$F8,$0F,$EC,$1F,$22,$F9,$AF,$ED,$0D,$9F,$ED
	.dc.b	$0D,$BF,$EA,$6D,$7F,$C1,$1E,$CB,$EA,$3F,$1F,$5F,$EC,$6A,$FE,$C5
	.dc.b	$5F,$A6,$AD,$7B,$5E,$0F,$53,$FB,$81,$FE,$A6,$BF,$AE,$DA,$FC,$FF
	.dc.b	$CA,$BE,$33,$7E,$B7,$F7,$78,$C5,$FF,$7D,$6D,$7F,$68,$6B,$FE,$05
	.dc.b	$6B,$FE,$DA,$6B,$FE,$13,$6F,$31,$A6,$3E,$5B,$ED,$CF,$EC,$9B,$DD
	.dc.b	$FE,$E7,$F7,$BF,$2F,$39,$E6,$FF,$06,$46,$23,$FC,$2C,$D7,$FD,$D6
	.dc.b	$D7,$FC,$48,$DC,$3F,$D2,$3D,$9F,$E1,$B8,$C1,$1F,$F0,$F7,$D4,$7F
	.dc.b	$3B,$C3,$FD,$F0,$DC,$FF,$DE,$FF,$30,$F6,$3F,$C3,$D3,$9C,$BF,$C6
	.dc.b	$53,$9A,$FF,$13,$36,$3F,$C2,$2D,$7F,$C5,$CD,$7A,$D7,$F8,$BE,$31
	.dc.b	$1D,$23,$07,$F8,$CF,$E6,$7E,$E6,$1F,$E8,$2C,$D1,$8C,$3F,$D1,$D0
	.dc.b	$F9,$AF,$F4,$9C,$7E,$EA,$75,$87,$F9,$26,$72,$FF,$89,$9A,$FF,$90
	.dc.b	$DA,$F4,$6B,$C6,$7F,$C8,$D1,$8A,$F0,$46,$0F,$F2,$57,$D4,$7F,$9A
	.dc.b	$9F,$2B,$6B,$FC,$B5,$EA,$3E,$A3,$FC,$95,$39,$B4,$E6,$FF,$CA,$4D
	.dc.b	$BF,$F1,$EB,$5F,$F2,$D3,$5F,$86,$D7,$BD,$FF,$97,$63,$15,$DA,$3F
	.dc.b	$CC,$1E,$3F,$EF,$AB,$FA,$3B,$73,$FA,$1F,$3A,$F9,$3F,$F3,$04,$E7
	.dc.b	$2F,$A3,$39,$7F,$CD,$0D,$BF,$F2,$9B,$5F,$F3,$83,$5E,$35,$FE,$71
	.dc.b	$8C,$47,$F9,$FD,$B8,$9B,$F0,$6F,$80,$FF,$39,$37,$0F,$E7,$27,$F0
	.dc.b	$CB,$FF,$44,$46,$5F,$EC,$71,$FD,$89,$FF,$41,$35,$FC,$5B,$5F,$F4
	.dc.b	$83,$6D,$F5,$0F,$92,$FF,$41,$36,$7F,$A4,$AF,$91,$CC,$C9,$CB,$C2
	.dc.b	$31,$7F,$DD,$B5,$7C,$FF,$FA,$35,$B7,$FD,$51,$7D,$46,$66,$4E,$5E
	.dc.b	$11,$8A,$FF,$07,$F0,$BE,$8D,$F1,$FF,$E4,$26,$E1,$FE,$BE,$F8,$97
	.dc.b	$D4,$66,$A4,$FF,$B0,$AB,$78,$74,$BD,$CE,$6B,$FC,$9B,$18,$9F,$F5
	.dc.b	$46,$1F,$EC,$9C,$3F,$D7,$D3,$99,$FF,$65,$4E,$6F,$FD,$8D,$F4,$AF
	.dc.b	$A8,$FF,$2B,$B6,$3C,$DF,$51,$89,$91,$FE,$D6,$9D,$5F,$96,$B2,$FF
	.dc.b	$8C,$0F,$F6,$C7,$1F,$FA,$ED,$FF,$5E,$35,$C9,$43,$9F,$A8,$FC,$07
	.dc.b	$C0,$DE,$3F,$F4,$05,$73,$9C,$B6,$32,$FF,$B2,$61,$F3,$9F,$E6,$A8
	.dc.b	$E0,$D7,$4D,$7F,$DD,$D1,$C3,$8F,$E2,$FC,$77,$FB,$C5,$68,$AF,$E1
	.dc.b	$3A,$2B,$53,$9D,$93,$9A,$FF,$7E,$FD,$4F,$FB,$61,$F9,$57,$C1,$E6
	.dc.b	$DA,$E2,$D9,$E1,$3F,$8E,$FC,$FD,$ED,$7C,$BB,$46,$27,$E8,$46,$27
	.dc.b	$53,$99,$FC,$3E,$0F,$8F,$F0,$36,$3E,$F4,$78,$E7,$86,$7F,$0A,$7E
	.dc.b	$0C,$F0,$23,$FE,$41,$C3,$C0,$FF,$8F,$B7,$7C,$6F,$BF,$B3,$FD,$74
	.dc.b	$D8,$CC,$F1,$F8,$A7,$C2,$66,$A4,$F3,$7E,$34,$EA,$3C,$1F,$A1,$FE
	.dc.b	$6D,$BC,$FF,$2E,$31,$1F,$E8,$4D,$1F,$98,$BD,$9D,$5C,$AE,$27,$F9
	.dc.b	$6E,$B7,$3F,$57,$F2,$2F,$95,$3F,$AC,$6B,$F8,$16,$BF,$D3,$9B,$AF
	.dc.b	$FB,$2B,$D3,$F8,$5E,$E9,$A7,$5E,$4F,$C1,$7E,$C0,$ED,$FB,$CD,$C7
	.dc.b	$EE,$B5,$FE,$02,$D7,$F7,$ED,$7F,$28,$D7,$FB,$0B,$78,$CE,$71,$FF
	.dc.b	$6F,$D7,$98,$95,$F7,$67,$C1,$A7,$E0,$ED,$F8,$27,$F6,$FA,$FC,$8B
	.dc.b	$5F,$E2,$AD,$7F,$89,$B5,$FD,$43,$5F,$EE,$AD,$8E,$BE,$5F,$E7,$E3
	.dc.b	$58,$87,$C9,$7E,$35,$B7,$FE,$8E,$6B,$FC,$DA,$7C,$98,$7A,$7E,$CD
	.dc.b	$ED,$7E,$AD,$B5,$FA,$A6,$BF,$CB,$9A,$FF,$2C,$6B,$FC,$05,$AF,$F8
	.dc.b	$39,$B9,$1C,$E1,$E9,$FC,$5F,$D0,$4A,$FD,$54,$F8,$30,$F9,$DF,$91
	.dc.b	$AB,$F8,$03,$6B,$F7,$ED,$7F,$A1,$B5,$FE,$82,$D7,$F8,$B3,$5F,$F1
	.dc.b	$1B,$70,$39,$C1,$EA,$FF,$A4,$EB,$FD,$72,$D7,$FD,$C2,$DC,$3C,$F8
	.dc.b	$7F,$1B,$FC,$23,$ED,$13,$6F,$E3,$0D,$BF,$E2,$ED,$7F,$AD,$BF,$7A
	.dc.b	$FF,$73,$7F,$E5,$56,$F8,$5F,$E6,$AE,$7E,$D8,$7F,$2B,$FB,$47,$D0
	.dc.b	$3E,$4B,$FA,$84,$61,$B1,$9C,$FF,$AB,$4E,$5B,$19,$7F,$A0,$1F,$D0
	.dc.b	$63,$4E,$03,$87,$6D,$7F,$A2,$7D,$7F,$C5,$C6,$FC,$37,$F3,$F8,$C4
	.dc.b	$F2,$BF,$06,$9F,$AD,$FC,$E1,$F1,$5F,$D1,$5B,$5F,$CB,$DA,$A7,$F9
	.dc.b	$B6,$4F,$C4,$3F,$6F,$9F,$F5,$9F,$87,$7E,$19,$FC,$A7,$0F,$54,$63
	.dc.b	$C6,$FE,$EC,$D8,$FE,$F4,$D7,$FC,$1F,$F3,$5F,$8F,$CB,$FD,$11,$F4
	.dc.b	$4F,$A8,$FE,$7C,$DA,$FE,$89,$5B,$CB,$E5,$FD,$8F,$FB,$26,$7F,$C0
	.dc.b	$8D,$7F,$C1,$0D,$7E,$73,$5E,$3F,$17,$FE,$09,$8C,$47,$F7,$56,$A7
	.dc.b	$3D,$5F,$2B,$E8,$BF,$A9,$71,$1A,$7F,$73,$AE,$17,$FD,$5F,$EC,$DF
	.dc.b	$39,$49,$3F,$D7,$FE,$AD,$F7,$78,$8F,$A5,$A7,$FA,$A2,$F8,$C7,$F8
	.dc.b	$27,$83,$FD,$5D,$18,$8F,$F2,$4D,$F2,$7F,$B5,$61,$FC,$06,$FF,$D6
	.dc.b	$3F,$5D,$F1,$70,$FF,$5B,$36,$3F,$C2,$13,$BA,$FF,$4E,$B7,$55,$6E
	.dc.b	$7F,$94,$37,$C2,$F6,$F8,$9F,$07,$EC,$3F,$17,$5F,$9D,$6B,$FA,$26
	.dc.b	$BF,$E2,$36,$A6,$93,$83,$FC,$7C,$F0,$CF,$F1,$47,$F8,$2F,$95,$F8
	.dc.b	$3B,$EB,$3F,$C4,$91,$87,$FC,$4D,$3C,$36,$FF,$69,$FD,$13,$EC,$BF
	.dc.b	$C1,$8D,$CF,$E7,$4E,$A7,$FC,$59,$F9,$0F,$C3,$AD,$57,$D6,$3F,$A5
	.dc.b	$38,$BF,$F7,$13,$6F,$FC,$79,$5F,$37,$67,$E3,$9F,$89,$C3,$B9,$FD
	.dc.b	$4F,$EB,$F0,$F5,$E2,$46,$32,$FE,$D1,$57,$1F,$F2,$43,$6F,$97,$47
	.dc.b	$F8,$9E,$F1,$8F,$82,$3E,$B7,$D4,$FC,$C2,$57,$F9,$B7,$85,$18,$8F
	.dc.b	$F3,$94,$61,$D4,$65,$FE,$3B,$9F,$F0,$EF,$73,$F4,$A4,$8F,$69,$F7
	.dc.b	$67,$FC,$AD,$F5,$5F,$8C,$8F,$83,$5F,$8E,$FA,$AF,$F1,$F7,$39,$DB
	.dc.b	$27,$6E,$5B,$19,$7F,$D5,$7F,$69,$FE,$35,$7F,$C9,$8D,$73,$ED,$7F
	.dc.b	$99,$FF,$23,$78,$3F,$E6,$9A,$D7,$1F,$F8,$55,$BC,$7F,$E7,$56,$E0
	.dc.b	$E1,$61,$CA,$3D,$FF,$48,$FA,$9F,$F4,$B3,$63,$FD,$34,$D7,$FD,$3C
	.dc.b	$D7,$FA,$23,$5F,$E9,$2D,$7F,$D1,$4D,$4E,$76,$FE,$7E,$FF,$16,$24
	.dc.b	$79,$27,$F3,$FE,$64,$8F,$A3,$D0,$8F,$F4,$57,$E4,$6F,$A2,$D5,$7E
	.dc.b	$4E,$B4,$F0,$AD,$CF,$C9,$F0,$A4,$FD,$07,$FD,$29,$FB,$5F,$F9,$DA
	.dc.b	$3E,$AC,$69,$A7,$FA,$06,$35,$C1,$FE,$45,$FC,$C7,$F8,$E4,$FF,$B5
	.dc.b	$9A,$FF,$B5,$1A,$FF,$AA,$5A,$FD,$5F,$C3,$5F,$37,$FE,$A5,$6C,$7F
	.dc.b	$A3,$DA,$FF,$BC,$A7,$BD,$7F,$95,$67,$2F,$F9,$5F,$E4,$DF,$80,$7F
	.dc.b	$AB,$DB,$7C,$F0,$EC,$66,$64,$64,$FF,$4F,$7C,$13,$E6,$3F,$D6,$4D
	.dc.b	$8F,$F5,$8B,$5F,$F7,$AB,$5F,$D7,$FC,$23,$F0,$8F,$F5,$23,$63,$F5
	.dc.b	$6D,$7F,$8B,$73,$FF,$B0,$2F,$FD,$8D,$F9,$8F,$F6,$08,$34,$9E,$87
	.dc.b	$FB,$BF,$A3,$FD,$1D,$F8,$57,$62,$61,$FE,$B7,$AF,$E8,$93,$A8,$F0
	.dc.b	$CE,$9F,$0C,$EB,$14,$E7,$7F,$F6,$25,$F1,$BD,$E1,$FE,$DB,$FC,$C7
	.dc.b	$FD,$C5,$1F,$E2,$7C,$91,$88,$FF,$16,$C6,$3E,$4F,$FD,$B9,$39,$9F
	.dc.b	$F7,$85,$FF,$A9,$FE,$BB,$E3,$92,$3F,$DA,$13,$DB,$1F,$F4,$A7,$03
	.dc.b	$E8,$7F,$DC,$CD,$AF,$F7,$2B,$5F,$23,$5F,$1B,$5F,$F7,$E3,$5F,$E1
	.dc.b	$0D,$CB,$F1,$73,$F3,$FE,$E7,$FD,$79,$5F,$C9,$1A,$FF,$43,$6D,$F9
	.dc.b	$A7,$FE,$00,$5F,$F9,$62,$87,$FB,$BC,$EF,$EF,$FF,$94,$BC,$2F,$2B
	.dc.b	$63,$CA,$D7,$67,$28,$DB,$4D,$94,$D9,$DF,$1F,$D2,$CF,$F9,$83,$E6
	.dc.b	$BD,$7A,$B3,$7E,$15,$9E,$2F,$AB,$18,$C7,$D3,$7F,$0B,$1E,$71,$8A
	.dc.b	$F0,$A9,$EB,$5C,$63,$13,$D2,$FB,$47,$9A,$BA,$7E,$77,$FC,$01,$BD
	.dc.b	$1A,$C5,$39,$C3,$E5,$C6,$2F,$53,$97,$FA,$13,$63,$8D,$6A,$76,$C7
	.dc.b	$D4,$70,$74,$E0,$FE,$FF,$9B,$ED,$71,$E8,$7F,$A5,$1B,$19,$9F,$78
	.dc.b	$F5,$5F,$59,$FE,$75,$D3,$F1,$38,$71,$AD,$9D,$B0,$E6,$79,$2F,$8F
	.dc.b	$C0,$7C,$89,$D1,$FE,$97,$29,$FE,$99,$AE,$8D,$C7,$C8,$E7,$D7,$FE
	.dc.b	$5A,$FF,$A8,$A7,$FC,$6D,$9F,$38,$FF,$0E,$CE,$9F,$CC,$CE,$9F,$03
	.dc.b	$72,$F5,$CF,$F3,$A7,$F4,$F8,$BD,$FF,$EB,$1C,$7F,$22,$B5,$C1,$E1
	.dc.b	$6E,$3D,$9B,$7C,$7C,$55,$F1,$CF,$EB,$F0,$3F,$3C,$EB,$8F,$D0,$8C
	.dc.b	$5F,$08,$C4,$FF,$B2,$AF,$8B,$C3,$4F,$CE,$37,$3F,$C6,$B6,$BF,$C1
	.dc.b	$8F,$17,$9D,$F5,$AE,$2D,$AE,$2D,$9E,$3B,$5C,$7C,$B8,$FA,$BF,$25
	.dc.b	$7D,$16,$F9,$FF,$69,$F2,$9B,$9C,$D7,$E5,$A7,$2F,$1A,$E1,$D0,$BE
	.dc.b	$11,$86,$CF,$F0,$5C,$18,$74,$CB,$F8,$33,$67,$F9,$CB,$63,$F5,$ED
	.dc.b	$7F,$C5,$CD,$9E,$77,$F3,$EB,$8B,$6B,$F8,$5F,$C1,$7D,$3E,$CF,$E9
	.dc.b	$B9,$7F,$AB,$AF,$FC,$DF,$87,$92,$B5,$1C,$0C,$E3,$BC,$3F,$A8,$35
	.dc.b	$F3,$CF,$F2,$D7,$CE,$D7,$CE,$DA,$ED,$5F,$A9,$6C,$F4,$9E,$19,$B6
	.dc.b	$7C,$5F,$92,$93,$98,$C5,$BE,$B9,$EA,$EE,$B5,$97,$BA,$72,$F8,$5B
	.dc.b	$87,$F6,$B6,$BF,$DA,$9A,$FF,$68,$C7,$FC,$3F,$3F,$E0,$CE,$EF,$0F
	.dc.b	$C0,$7F,$9A,$23,$04,$AF,$F4,$C5,$70,$7F,$C2,$33,$B7,$35,$E7,$9D
	.dc.b	$6F,$FE,$02,$6D,$FF,$30,$BE,$13,$9A,$F0,$FA,$DF,$36,$7A,$61,$FE
	.dc.b	$02,$6B,$C6,$3E,$6E,$0F,$AF,$F0,$1F,$2B,$F3,$1F,$F7,$AE,$5F,$E2
	.dc.b	$88,$C7,$C2,$FD,$99,$CE,$1F,$E1,$16,$BF,$E1,$06,$BF,$D1,$BC,$FF
	.dc.b	$F4,$05,$FE,$3B,$F3,$17,$E0,$7F,$62,$73,$5F,$E3,$49,$CD,$7F,$94
	.dc.b	$FD,$8F,$BB,$18,$9F,$F1,$CD,$70,$DF,$F8,$77,$1A,$B5,$1F,$E5,$7B
	.dc.b	$EC,$7E,$1B,$D2,$7C,$FE,$32,$FD,$0C,$61,$FC,$A3,$72,$FF,$17,$35
	.dc.b	$FF,$16,$B5,$FC,$27,$C8,$7D,$93,$F2,$7F,$BF,$FF,$24,$C7,$ED,$46
	.dc.b	$6F,$FC,$A6,$EB,$BF,$17,$FE,$1B,$F1,$FF,$31,$DA,$7F,$3B,$ED,$BE
	.dc.b	$AF,$31,$31,$FF,$12,$3A,$8C,$5F,$FB,$9F,$E2,$9E,$8F,$E8,$DB,$7F
	.dc.b	$EE,$EF,$9D,$7D,$13,$F0,$7D,$5F,$EB,$E7,$31,$FB,$B6,$D7,$F6,$16
	.dc.b	$E1,$FE,$2E,$3F,$C5,$EE,$1C,$35,$CD,$F7,$9E,$4D,$8E,$42,$78,$3F
	.dc.b	$C0,$34,$7B,$7F,$18,$FE,$03,$EC,$3C,$EC,$63,$FE,$47,$9B,$7F,$E8
	.dc.b	$0A,$D4,$E4,$FE,$43,$F0,$CF,$C0,$3F,$84,$7D,$EB,$E9,$7F,$86,$CE
	.dc.b	$6B,$FD,$EB,$39,$78,$D7,$0A,$DB,$63,$FD,$F7,$C1,$FC,$69,$BE,$4F
	.dc.b	$CD,$DD,$F2,$1F,$56,$30,$FC,$F8,$CC,$FC,$68,$C1,$1C,$13,$4F,$93
	.dc.b	$C1,$F7,$27,$F9,$87,$C7,$BE,$6B,$F9,$27,$E3,$4F,$AF,$FE,$53,$39
	.dc.b	$C7,$FC,$CF,$D9,$FE,$92,$AD,$47,$F4,$69,$D6,$4A,$D5,$F2,$C3,$AB
	.dc.b	$F9,$AF,$32,$61,$FD,$3B,$E4,$FF,$8D,$E1,$F6,$E7,$2F,$F9,$ED,$FE
	.dc.b	$77,$1A,$F3,$BF,$CC,$D1,$FD,$07,$E6,$9F,$57,$F4,$6F,$FD,$77,$F9
	.dc.b	$97,$F3,$AD,$3F,$C2,$15,$AE,$5F,$E8,$B3,$9C,$FD,$BF,$8A,$FD,$BC
	.dc.b	$3F,$1F,$82,$31,$1F,$9B,$8C,$73,$7F,$5B,$6D,$7F,$90,$5B,$3C,$AB
	.dc.b	$8B,$DA,$B8,$46,$21,$E4,$BC,$DF,$D9,$C7,$CB,$C3,$FD,$6B,$D3,$FC
	.dc.b	$05,$FE,$EB,$3A,$C7,$FD,$8B,$39,$AF,$F6,$DC,$E6,$B8,$F8,$5F,$D9
	.dc.b	$23,$13,$FD,$A2,$B8,$5F,$FB,$3B,$EF,$DF,$2D,$FD,$9D,$B8,$FF,$81
	.dc.b	$3F,$41,$FE,$45,$5B,$AD,$12,$75,$E5,$DC,$7F,$88,$7A,$BF,$C3,$CF
	.dc.b	$CF,$7D,$DD,$AF,$E2,$1D,$C8,$FC,$F1,$23,$F8,$1B,$6B,$F3,$73,$B9
	.dc.b	$C9,$23,$FD,$C5,$C5,$FE,$15,$CB,$FD,$D3,$F8,$03,$E3,$7F,$03,$ED
	.dc.b	$7F,$80,$E3,$18,$7E,$78,$FF,$62,$FC,$13,$E7,$FF,$CB,$F1,$FE,$7D
	.dc.b	$8C,$3F,$E1,$B6,$B9,$8D,$E5,$FE,$F3,$FA,$2F,$E2,$B3,$FE,$2C,$6C
	.dc.b	$7D,$D1,$D5,$FE,$47,$F0,$FF,$C3,$33,$AD,$3F,$AD,$FE,$5D,$FC,$5A
	.dc.b	$F8,$74,$A3,$13,$F9,$EF,$CB,$9E,$AF,$F8,$8A,$73,$9F,$FB,$83,$35
	.dc.b	$6A,$3F,$D3,$11,$8A,$FF,$1B,$36,$33,$3B,$C7,$F9,$9C,$E5,$B1,$97
	.dc.b	$FA,$04,$65,$B1,$97,$FD,$E3,$E0,$FD,$FF,$A8,$FF,$1A,$72,$9F,$F1
	.dc.b	$CF,$EA,$DE,$8A,$FF,$1E,$CE,$71,$FF,$2A,$CF,$C3,$9F,$E9,$DA,$A3
	.dc.b	$11,$FE,$D4,$6B,$FA,$E6,$C6,$FC,$FF,$F2,$94,$60,$8F,$F9,$42,$FF
	.dc.b	$A9,$6D,$FE,$99,$CB,$F2,$DE,$8F,$F9,$3B,$F0,$DF,$E1,$59,$79,$A3
	.dc.b	$1F,$05,$FE,$5D,$9F,$F2,$3C,$FE,$F9,$B9,$9A,$49,$E0,$4D,$3F,$D9
	.dc.b	$98,$3E,$F7,$F8,$F6,$5E,$36,$CF,$FB,$CB,$EB,$5F,$82,$FF,$97,$E7
	.dc.b	$39,$FF,$0E,$9C,$D7,$F9,$BD,$B1,$FD,$BB,$34,$62,$3E,$6B,$5E,$D7
	.dc.b	$FE,$79,$8C,$47,$58,$C1,$FE,$78,$BF,$E1,$B7,$F8,$06,$CF,$F2,$0F
	.dc.b	$77,$FA,$DB,$FA,$3C,$BD,$DE,$37,$F9,$EA,$73,$3F,$79,$A8,$E5,$D1
	.dc.b	$FE,$8B,$8C,$C6,$30,$FF,$46,$46,$33,$34,$DE,$0F,$8A,$7C,$67,$FA
	.dc.b	$1D,$B7,$FE,$A4,$FB,$17,$CE,$66,$BF,$D1,$F1,$89,$FF,$4D,$37,$2F
	.dc.b	$98,$D7,$F2,$ED,$7D,$59,$23,$11,$FD,$11,$B1,$9B,$FF,$54,$46,$B1
	.dc.b	$F4,$7B,$7E,$73,$FD,$43,$F9,$77,$F1,$2E,$2F,$CC,$68,$9C,$E7,$FE
	.dc.b	$B1,$AF,$E2,$F9,$FE,$A9,$AF,$91,$AF,$F5,$6F,$D7,$7F,$90,$BF,$C4
	.dc.b	$9B,$87,$FB,$2F,$F6,$4F,$B6,$FF,$9C,$A3,$37,$EE,$6D,$73,$F0,$7F
	.dc.b	$D7,$54,$BE,$37,$B9,$FF,$5C,$4F,$F7,$53,$F6,$45,$8F,$93,$F8,$B7
	.dc.b	$F3,$F7,$F8,$6C,$66,$FF,$D8,$AD,$D3,$FC,$2D,$E8,$FF,$B1,$AF,$85
	.dc.b	$6A,$3F,$D9,$3F,$75,$FE,$14,$E1,$37,$F6,$9F,$F4,$76,$5F,$62,$34
	.dc.b	$4F,$33,$FD,$A3,$3A,$C3,$FB,$EC,$F1,$E5,$FF,$70,$4F,$FA,$7A,$7F
	.dc.b	$8C,$B5,$FC,$1B,$5E,$3F,$0D,$E4,$3F,$DB,$BF,$79,$FE,$C0,$C7,$C5
	.dc.b	$19,$7F,$DF,$38,$23,$7F,$1F,$9B,$FE,$A0,$7F,$9D,$79,$7F,$CD,$AF
	.dc.b	$F9,$B9,$AE,$4A,$7E,$94,$DF,$B0,$FB,$EF,$F4,$3C,$61,$AF,$F9,$E2
	.dc.b	$35,$E8,$F4,$6C,$72,$FB,$EF,$F2,$43,$79,$35,$18,$8E,$4D,$7E,$87
	.dc.b	$D0,$7F,$8C,$F5,$B1,$F0,$23,$0F,$F8,$B3,$6F,$D0,$39,$21,$FD,$AB
	.dc.b	$64,$F0,$E2,$57,$0A,$FE,$DF,$F7,$DF,$E8,$E7,$E3,$CA,$31,$1F,$CA
	.dc.b	$5A,$FF,$54,$6B,$F8,$7F,$5B,$B6,$1F,$E2,$E8,$FA,$F1,$E1,$9E,$79
	.dc.b	$FF,$B6,$AB,$55,$E7,$AE,$26,$FE,$23,$DF,$D1,$FE,$B8,$F5,$EF,$07
	.dc.b	$FB,$72,$FF,$C3,$82,$74,$E3,$3F,$F4,$E7,$87,$C2,$E4,$F0,$AD,$BF
	.dc.b	$E1,$0F,$77,$FD,$A4,$F1,$FB,$EF,$F4,$53,$7E,$0B,$E5,$46,$23,$FC
	.dc.b	$3D,$B3,$EF,$9E,$5E,$AE,$57,$13,$CF,$5B,$9F,$A9,$F7,$DF,$2A,$7F
	.dc.b	$CC,$ED,$D4,$D3,$59,$FF,$B1,$FE,$5B,$E2,$FC,$41,$A7,$4E,$1F,$C0
	.dc.b	$79,$0F,$9A,$FD,$0E,$3F,$CC,$DB,$19,$CB,$F9,$EC,$60,$8E,$08,$7F
	.dc.b	$B6,$31,$FF,$6C,$47,$88,$91,$FD,$2D,$FC,$87,$DE,$BE,$2F,$C2,$D9
	.dc.b	$FC,$7B,$5F,$EE,$CD,$7F,$C4,$2D,$7F,$A5,$B5,$FF,$46,$B7,$03,$9B
	.dc.b	$FC,$27,$DF,$7E,$29,$34,$F2,$DF,$E9,$7D,$27,$DB,$7C,$66,$E3,$FE
	.dc.b	$0B,$8C,$11,$C1,$0F,$F0,$74,$70,$7E,$35,$FF,$90,$5F,$C8,$35,$FF
	.dc.b	$02,$37,$85,$DD,$FE,$1F,$BF,$DC,$7D,$F6,$EF,$FC,$86,$D7,$FB,$6B
	.dc.b	$5F,$F5,$7B,$74,$34,$9E,$D7,$E0,$9E,$18,$BE,$F7,$F9,$65,$FE,$95
	.dc.b	$AF,$F8,$65,$B9,$79,$A3,$BD,$FF,$0D,$FB,$EF,$A4,$4D,$7F,$1E,$D9
	.dc.b	$FE,$1A,$D7,$FC,$68,$FD,$7B,$FD,$65,$FF,$94,$DB,$A7,$F9,$B1,$B5
	.dc.b	$FE,$CC,$AF,$C4,$57,$FB,$8B,$E8,$5F,$7D,$FC,$FB,$E0,$11,$8B,$FF
	.dc.b	$41,$B5,$FF,$3F,$B5,$FF,$3E,$B5,$FF,$46,$35,$FF,$4D,$B6,$3A,$CF
	.dc.b	$EB,$3E,$C7,$E2,$E5,$FE,$20,$6D,$FF,$B8,$9B,$5C,$23,$FD,$51,$F2
	.dc.b	$9F,$CD,$7E,$14,$D3,$F9,$FC,$BE,$FB,$F6,$D5,$FE,$AD,$F8,$77,$DF
	.dc.b	$7F,$18,$8F,$F2,$41,$FE,$BE,$70,$E2,$37,$E2,$FF,$62,$6C,$7F,$40
	.dc.b	$F4,$3F,$25,$A7,$FA,$6B,$E9,$FF,$D8,$D3,$FC,$9D,$AF,$FA,$DA,$76
	.dc.b	$FF,$95,$FE,$F9,$F6,$1F,$DB,$1B,$1F,$DC,$1A,$FF,$74,$6B,$E8,$0E
	.dc.b	$1F,$5B,$5E,$D1,$D4,$FE,$F7,$CF,$F2,$3E,$F7,$FA,$3E,$E6,$9F,$DA
	.dc.b	$2B,$AD,$7F,$4C,$FB,$E7,$CA,$D2,$4F,$F4,$9F,$82,$7C,$86,$23,$FC
	.dc.b	$F5,$5C,$1F,$E9,$4D,$7F,$DB,$33,$C1,$FE,$A5,$5F,$C0,$FA,$3F,$D5
	.dc.b	$DF,$79,$F0,$9B,$8F,$F7,$59,$D7,$07,$F9,$A1,$B7,$FD,$76,$F8,$4F
	.dc.b	$FB,$39,$B0,$EF,$F5,$F8,$1E,$FF,$BE,$FC,$5C,$FF,$88,$1A,$FF,$89
	.dc.b	$1A,$FD,$16,$BF,$8A,$6B,$FE,$93,$6A,$69,$AC,$BF,$17,$1F,$95,$FA
	.dc.b	$A7,$C7,$7F,$86,$39,$91,$89,$FE,$C7,$D1,$F9,$9B,$FF,$69,$7D,$CB
	.dc.b	$E2,$BF,$C0,$2D,$AF,$EF,$0D,$53,$FC,$4F,$47,$CE,$7C,$8E,$75,$EE
	.dc.b	$BA,$E1,$D0,$FF,$12,$57,$CE,$B7,$CF,$7E,$27,$1F,$F0,$37,$D7,$D7
	.dc.b	$FD,$74,$DC,$3C,$13,$78,$BF,$D3,$ED,$AF,$F0,$A5,$5C,$CE,$70,$EF
	.dc.b	$D1,$F5,$FE,$FB,$FC,$A7,$C4,$69,$30,$FB,$4D,$7E,$D3,$5F,$E0,$38
	.dc.b	$7F,$0A,$F6,$BF,$46,$48,$FF,$13,$9F,$66,$7F,$C8,$9F,$44,$E6,$FF
	.dc.b	$17,$8F,$BE,$A6,$99,$5F,$E2,$97,$4D,$8C,$BF,$53,$6F,$F6,$0F,$DF
	.dc.b	$7F,$89,$23,$FC,$68,$D7,$47,$E6,$5C,$70,$7D,$BF,$BE,$78,$3F,$E5
	.dc.b	$99,$CD,$F4,$F6,$30,$7F,$97,$6B,$78,$7B,$BE,$FB,$FD,$C6,$FF,$A0
	.dc.b	$1A,$FF,$A1,$1A,$FF,$11,$6B,$FC,$A9,$AF,$FB,$CD,$A9,$A6,$B7,$FE
	.dc.b	$72,$FF,$12,$24,$78,$FE,$2B,$F3,$7E,$54,$8F,$CD,$BF,$E7,$5F,$AA
	.dc.b	$7D,$5E,$67,$E2,$E1,$F9,$AA,$C6,$DF,$9F,$BD,$D7,$CC,$F0,$FE,$31
	.dc.b	$1F,$F3,$17,$DF,$7F,$99,$DF,$9F,$D0,$E0,$FF,$7F,$C7,$12,$1A,$9E
	.dc.b	$18,$7E,$B7,$EF,$BF,$8E,$4F,$F3,$7B,$46,$23,$FA,$8C,$63,$C7,$FF
	.dc.b	$27,$CE,$64,$7F,$17,$FB,$07,$CC,$7F,$A5,$5B,$1F,$E7,$96,$BE,$89
	.dc.b	$F0,$E1,$FB,$F9,$CB,$FC,$07,$E8,$1F,$25,$FE,$9F,$6D,$F3,$C3,$B1
	.dc.b	$99,$91,$93,$FD,$11,$EF,$BE,$8F,$FC,$CE,$D8,$FF,$52,$B5,$F8,$4D
	.dc.b	$7F,$63,$F0,$2F,$BE,$FF,$4C,$B6,$3F,$D2,$0D,$7E,$9C,$F3,$CB,$F8
	.dc.b	$D7,$DF,$7F,$5F,$DC,$D2,$79,$FF,$D8,$30,$7D,$77,$62,$63,$FE,$A1
	.dc.b	$AF,$E8,$BC,$FF,$E8,$59,$CC,$FE,$32,$B8,$BB,$BC,$E3,$FE,$3D,$BE
	.dc.b	$13,$E8,$BF,$F6,$56,$1F,$D4,$31,$7D,$F7,$FB,$3A,$BF,$D9,$2D,$7F
	.dc.b	$0C,$D7,$F0,$AD,$7F,$07,$7F,$E9,$9F,$B0,$F8,$E4,$8F,$F5,$D4,$F3
	.dc.b	$C7,$F9,$F6,$EF,$BE,$FF,$4D,$B6,$BF,$DB,$2D,$7F,$2C,$D7,$F2,$AD
	.dc.b	$7F,$27,$8F,$E2,$27,$99,$23,$FC,$28,$7F,$B3,$7E,$83,$FD,$5C,$FF
	.dc.b	$B2,$5A,$FE,$D9,$B7,$E6,$9F,$F7,$85,$FF,$94,$AB,$F4,$3E,$0F,$8B
	.dc.b	$DE,$FF,$1B,$B6,$32,$79,$27,$83,$FE,$6F,$3F,$BF,$38,$8E,$11,$9C
	.dc.b	$7F,$43,$5D,$23,$93,$5E,$2F,$F9,$1B,$E6,$3D,$7A,$B3,$7D,$F6,$78
	.dc.b	$BE,$AC,$63,$0F,$DC,$DF,$C2,$C7,$BC,$62,$BB,$B4,$F9,$17,$CD,$A6
	.dc.b	$67,$8A,$AE,$9F,$7D,$FC,$01,$BC,$FC,$69,$39,$C3,$EC,$BF,$E4,$9A
	.dc.b	$DD,$F5,$8D,$3F,$E2,$56,$D7,$28,$E9,$18,$25,$E2,$3D,$07,$F7,$8C
	.dc.b	$DF,$7B,$ED,$8C,$37,$CF,$FF,$47,$BF,$07,$D6,$7D,$F7,$F9,$D7,$3F
	.dc.b	$C3,$D6,$AF,$E0,$69,$FE,$78,$8C,$4E,$B1,$EF,$5A,$18,$74,$F7,$FC
	.dc.b	$B7,$CE,$7F,$CD,$4D,$7A,$57,$27,$2D,$7C,$1E,$CF,$F2,$E6,$E9,$FA
	.dc.b	$07,$FC,$2F,$19,$7F,$9C,$D7,$F8,$F2,$74,$D8,$F3,$1C,$23,$E1,$7B
	.dc.b	$FF,$D7,$38,$FE,$35,$FF,$3E,$CF,$E1,$B7,$4F,$0E,$1E,$FD,$18,$71
	.dc.b	$8E,$B9,$77,$C3,$F2,$5F,$7D,$96,$F9,$3F,$ED,$57,$3A,$DC,$BF,$7F
	.dc.b	$5B,$6D,$67,$0F,$F3,$54,$E5,$B5,$9E,$24,$E5,$B5,$98,$C1,$FE,$D0
	.dc.b	$D7,$F3,$CE,$FE,$F5,$F7,$DD,$F9,$FF,$61,$F1,$FD,$67,$37,$FA,$99
	.dc.b	$CB,$FA,$78,$CE,$7F,$83,$9C,$BF,$85,$C9,$3A,$D7,$F7,$BD,$0C,$7B
	.dc.b	$65,$FE,$F9,$CB,$FB,$BD,$F4,$6C,$67,$1F,$E4,$B3,$96,$C6,$5F,$F4
	.dc.b	$84,$65,$B1,$9C,$7C,$6E,$F2,$FE,$1D,$E3,$B0,$FE,$54,$DB,$FF,$0A
	.dc.b	$E1,$FE,$B0,$9C,$9F,$E0,$8E,$CF,$F3,$F4,$66,$7F,$C2,$4D,$7D,$75
	.dc.b	$FC,$9D,$F4,$B5,$F3,$B7,$1F,$8E,$D9,$E5,$3C,$2F,$FA,$0E,$0E,$07
	.dc.b	$C5,$79,$F2,$FE,$CD,$18,$AF,$31,$FD,$33,$ED,$9E,$96,$E3,$FA,$FC
	.dc.b	$EF,$1F,$F1,$9D,$6A,$7F,$A7,$E8,$8C,$4F,$F8,$BD,$AF,$F8,$B9,$AF
	.dc.b	$4E,$EE,$CB,$3F,$44,$62,$BF,$D6,$3E,$9E,$A7,$38,$7F,$85,$2B,$18
	.dc.b	$FF,$8A,$A3,$D7,$87,$FB,$37,$E0,$3F,$CE,$71,$B2,$47,$F8,$2E,$BB
	.dc.b	$4E,$A3,$13,$CB,$1F,$0B,$EE,$9D,$4F,$F9,$21,$AF,$F7,$46,$BE,$BB
	.dc.b	$7D,$F7,$98,$F6,$7D,$F7,$FB,$EA,$FF,$C5,$B1,$8F,$84,$FB,$D3,$9B
	.dc.b	$FF,$2C,$35,$FF,$2B,$B5,$FF,$01,$F9,$CF,$57,$E4,$7D,$EB,$EF,$BE
	.dc.b	$54,$E7,$2F,$64,$E5,$FF,$2C,$57,$0A,$DB,$63,$24,$BF,$F3,$9F,$A2
	.dc.b	$F9,$CD,$CF,$C7,$8C,$3F,$84,$6C,$FF,$9C,$5A,$FF,$9C,$1A,$FE,$3B
	.dc.b	$E9,$7E,$D9,$F8,$1F,$BE,$FE,$55,$CF,$F9,$88,$CD,$FF,$95,$1D,$78
	.dc.b	$18,$BF,$F0,$E7,$8F,$F8,$7E,$F3,$F8,$BF,$65,$F7,$D9,$89,$8F,$F8
	.dc.b	$95,$D4,$62,$FF,$D3,$7F,$18,$F4,$7F,$24,$DB,$FC,$7F,$CD,$3E,$81
	.dc.b	$F6,$5A,$BF,$CE,$CE,$63,$F7,$8D,$AF,$F1,$BB,$70,$FF,$47,$9F,$E9
	.dc.b	$07,$0E,$1A,$E6,$FB,$CF,$26,$C7,$21,$3C,$1F,$F2,$27,$B1,$F1,$4F
	.dc.b	$D7,$7A,$EF,$3F,$18,$7F,$91,$E6,$D7,$FA,$CA,$B5,$39,$3F,$4D,$F0
	.dc.b	$8F,$AA,$FD,$B7,$DD,$BE,$8F,$F7,$53,$9A,$FD,$AF,$8A,$AD,$47,$2B
	.dc.b	$E3,$7C,$1B,$1C,$38,$3F,$89,$73,$FF,$9F,$DB,$1F,$0F,$E3,$BE,$AF
	.dc.b	$95,$F4,$38,$51,$F4,$1B,$3F,$1E,$30,$47,$04,$DB,$E5,$70,$FD,$D9
	.dc.b	$FE,$65,$F1,$AF,$A1,$FE,$29,$F7,$CF,$BE,$FE,$31,$39,$C7,$DF,$39
	.dc.b	$7F,$86,$46,$71,$FF,$73,$CE,$5F,$F6,$FC,$67,$5F,$F4,$E4,$E5,$FF
	.dc.b	$4F,$75,$FF,$2F,$BF,$F3,$E7,$9B,$FC,$9F,$E4,$7F,$91,$63,$FE,$7D
	.dc.b	$8F,$D1,$12,$7F,$8E,$4E,$AF,$9C,$FF,$32,$F9,$87,$D5,$FD,$4C,$3F
	.dc.b	$DE,$3F,$77,$FC,$F2,$FF,$B0,$56,$B9,$7F,$9B,$4E,$72,$FE,$C1,$F7
	.dc.b	$DF,$7F,$0F,$C9,$60,$8C,$47,$E7,$A3,$1C,$DF,$D7,$DB,$5F,$E9,$A6
	.dc.b	$CF,$2A,$E2,$F6,$AE,$11,$88,$7A,$0F,$37,$F6,$D0,$BF,$F6,$37,$4F
	.dc.b	$F0,$4C,$7F,$9F,$56,$A7,$3E,$17,$F5,$18,$C4,$FF,$B0,$2B,$85,$F3
	.dc.b	$8F,$F5,$9C,$EA,$3F,$AD,$4E,$AF,$FD,$A9,$F7,$CF,$94,$FE,$B4,$DC
	.dc.b	$3F,$A7,$7D,$F7,$F2,$6A,$DD,$68,$93,$AF,$2E,$E3,$FC,$53,$D4,$FE
	.dc.b	$26,$7E,$7F,$EE,$AD,$7F,$14,$EE,$47,$E8,$49,$1F,$C1,$DB,$5F,$9E
	.dc.b	$9D,$CE,$49,$1F,$C2,$F7,$71,$7F,$86,$F3,$FF,$78,$7D,$F3,$E3,$7F
	.dc.b	$07,$ED,$FF,$77,$8C,$63,$FD,$D3,$E0,$DF,$6B,$FE,$2B,$8F,$F6,$BC
	.dc.b	$61,$FF,$10,$35,$CC,$6F,$3F,$17,$D0,$BC,$AF,$F0,$33,$6B,$FC,$03
	.dc.b	$D7,$FE,$49,$F1,$33,$5F,$E0,$BF,$BE,$FE,$33,$87,$D9,$3F,$0E,$E3
	.dc.b	$6F,$CE,$FD,$F3,$D5,$E1,$E7,$7E,$96,$31,$9F,$FB,$BB,$34,$63,$1F
	.dc.b	$E6,$11,$8B,$FF,$1D,$36,$B8,$56,$EB,$F9,$FC,$E5,$E3,$19,$42,$34
	.dc.b	$FF,$46,$F3,$3E,$EF,$D3,$FF,$89,$79,$CF,$F8,$A7,$EF,$BD,$98,$F2
	.dc.b	$AD,$BF,$E1,$D8,$F8,$77,$FD,$43,$9D,$18,$8F,$0C,$63,$4F,$F2,$94
	.dc.b	$E6,$7F,$5E,$D9,$D5,$7F,$91,$23,$06,$EF,$F2,$7E,$8F,$9C,$FF,$1F
	.dc.b	$7D,$F7,$F0,$9C,$3F,$C8,$59,$FF,$5A,$DF,$FC,$DD,$92,$BF,$C9,$35
	.dc.b	$FB,$E6,$C6,$7E,$0B,$FC,$CD,$18,$23,$82,$67,$FE,$66,$FB,$DF,$E3
	.dc.b	$D8,$7F,$9E,$F3,$46,$27,$E0,$FD,$63,$EF,$BF,$CA,$53,$9C,$FF,$87
	.dc.b	$CE,$6B,$FC,$C6,$D8,$FD,$43,$5F,$9C,$D7,$AD,$7F,$9F,$63,$11,$D2
	.dc.b	$30,$7F,$A0,$EF,$CB,$E3,$78,$8F,$F3,$37,$BB,$FD,$72,$BF,$39,$3A
	.dc.b	$C3,$FC,$B5,$39,$9F,$BC,$D7,$EE,$B5,$FE,$52,$DD,$7F,$C9,$9D,$1F
	.dc.b	$E9,$18,$C1,$1F,$27,$C4,$BE,$3B,$FC,$EC,$DC,$3F,$CD,$5F,$60,$F9
	.dc.b	$CC,$D7,$F9,$F6,$31,$3F,$E8,$A6,$E5,$FC,$3D,$AF,$E5,$5A,$FE,$51
	.dc.b	$AF,$F4,$06,$E4,$69,$AC,$D8,$BD,$2F,$39,$FE,$8F,$FB,$EF,$E1,$DC
	.dc.b	$5F,$83,$9D,$65,$FE,$83,$9F,$E2,$59,$7E,$9D,$AF,$E9,$9A,$FF,$52
	.dc.b	$6E,$4F,$BE,$FE,$3D,$3F,$C4,$63,$1E,$0F,$FA,$8B,$17,$DF,$7D,$B7
	.dc.b	$F4,$37,$C1,$B3,$FE,$A5,$8C,$63,$FE,$CF,$AE,$0F,$EE,$23,$38,$7F
	.dc.b	$AB,$6F,$FB,$A1,$F9,$F2,$CF,$EF,$BE,$FB,$F9,$E3,$FC,$1E,$33,$3C
	.dc.b	$7D,$2F,$F5,$94,$63,$25,$F0,$AD,$5F,$FA,$A7,$EE,$3F,$C1,$CF,$F0
	.dc.b	$A6,$C7,$F7,$6F,$BE,$FF,$4A,$3F,$0E,$33,$66,$98,$9D,$E1,$FC,$12
	.dc.b	$31,$CB,$FE,$E5,$9E,$77,$FC,$6D,$AF,$F1,$A6,$BF,$E0,$69,$7C,$37
	.dc.b	$FB,$3D,$B8,$FF,$B0,$7E,$EF,$FD,$89,$8F,$7A,$D1,$27,$5F,$21,$70
	.dc.b	$FE,$98,$7F,$9E,$B8,$BF,$CE,$4F,$F9,$CD,$AE,$4A,$7E,$50,$DF,$B2
	.dc.b	$FB,$EF,$F4,$64,$61,$AF,$F9,$F6,$35,$E8,$F4,$6C,$72,$FB,$EF,$F1
	.dc.b	$03,$73,$D4,$62,$39,$35,$FE,$7F,$F9,$5F,$FA,$A1,$FF,$7E,$B5,$FF
	.dc.b	$01,$35,$FE,$90,$D7,$8F,$77,$BB,$89,$39,$9F,$F1,$AB,$63,$33,$F1
	.dc.b	$3C,$97,$E8,$3F,$E1,$8B,$FE,$DC,$D7,$FB,$6B,$5F,$ED,$2D,$7F,$AB
	.dc.b	$FD,$FF,$F4,$77,$E8,$DA,$31,$1F,$E1,$B6,$BF,$D8,$5A,$F2,$F1,$BD
	.dc.b	$51,$88,$FF,$26,$B6,$7A,$4F,$BF,$DE,$C1,$1C,$12,$BC,$71,$96,$C7
	.dc.b	$E0,$BF,$2B,$7C,$AD,$24,$FE,$22,$32,$D8,$CB,$F8,$E8,$CB,$63,$F1
	.dc.b	$BF,$98,$BE,$F6,$92,$7F,$27,$19,$7F,$2B,$F9,$13,$EB,$BF,$2A,$D8
	.dc.b	$CC,$60,$EB,$E1,$48,$E6,$D7,$F4,$35,$FE,$40,$72,$43,$31,$FA,$4F
	.dc.b	$1D,$F8,$A3,$AF,$0D,$AF,$D5,$CE,$A7,$27,$EA,$BE,$51,$F4,$D4,$95
	.dc.b	$FB,$08,$CB,$63,$2F,$ED,$A3,$2D,$8F,$83,$F1,$F2,$3E,$2B,$5F,$DE
	.dc.b	$FE,$A6,$F9,$8F,$DD,$B6,$3F,$7E,$D7,$71,$A2,$3A,$6B,$FE,$6F,$FC
	.dc.b	$01,$F6,$9F,$AB,$6C,$7E,$71,$AE,$E3,$F8,$5F,$D1,$7D,$B8,$FE,$22
	.dc.b	$D7,$FC,$96,$7F,$93,$1C,$38,$6B,$9E,$2E,$64,$7F,$8C,$C6,$97,$E5
	.dc.b	$BF,$E8,$17,$F8,$EB,$5F,$E4,$2D,$7C,$27,$0F,$85,$F3,$ED,$23,$AB
	.dc.b	$5F,$4B,$5E,$35,$F8,$37,$27,$F2,$EF,$95,$7D,$4F,$F9,$E6,$30,$D7
	.dc.b	$FD,$F6,$D7,$E1,$C7,$0F,$81,$FE,$67,$F9,$5B,$D0,$A4,$AF,$E7,$31
	.dc.b	$96,$C6,$5F,$E7,$F1,$96,$C7,$4F,$B4,$91,$C9,$AF,$FA,$22,$5F,$3B
	.dc.b	$FC,$8D,$B1,$FE,$8F,$6B,$B8,$F3,$FD,$44,$8F,$2B,$5F,$F4,$9F,$73
	.dc.b	$E5,$7F,$97,$36,$3F,$D3,$4D,$77,$1F,$D5,$3F,$1C,$FD,$C4,$7F,$5A
	.dc.b	$6B,$F6,$9A,$FE,$31,$AF,$FA,$61,$A9,$A4,$EC,$FE,$C3,$F7,$2F,$9D
	.dc.b	$A4,$AF,$EC,$F1,$96,$F0,$F0,$23,$82,$1F,$DC,$21,$1A,$8C,$3E,$0F
	.dc.b	$2E,$47,$66,$BC,$5A,$FE,$62,$51,$88,$F8,$1F,$0F,$F1,$C8,$FF,$7E
	.dc.b	$C5,$18,$8F,$94,$D7,$E5,$3F,$0B,$F1,$89,$1F,$01,$AF,$AF,$EE,$BE
	.dc.b	$39,$1F,$8A,$D7,$46,$08,$7C,$3D,$3E,$79,$1F,$9A,$D7,$46,$08,$7F
	.dc.b	$86,$BF,$2F,$7C,$A7,$F8,$61,$B1,$FE,$1C,$6B,$B8,$D1,$1D,$35,$FF
	.dc.b	$12,$F8,$CF,$C4,$BF,$C4,$6D,$8F,$F1,$3B,$5D,$C6,$BE,$F6,$46,$9A
	.dc.b	$F4,$6B,$FC,$42,$51,$88,$FE,$47,$0F,$77,$C2,$48,$F0,$B5,$F3,$B5
	.dc.b	$E8,$FC,$3F,$97,$91,$F0,$DA,$FC,$96,$BF,$C8,$B1,$46,$23,$C7,$E2
	.dc.b	$FC,$22,$3F,$09,$F8,$BE,$DC,$8F,$8A,$D7,$E5,$B5,$F8,$0D,$7F,$A0
	.dc.b	$FE,$D3,$FD,$21,$FE,$66,$D7,$3E,$27,$F9,$52,$31,$00,$E1,$B1,$B2
	.dc.b	$3F,$E5,$EF,$A0,$7A,$3F,$E5,$E6,$C6,$63,$07,$F9,$1F,$F2,$CF,$E2
	.dc.b	$51,$FD,$5B,$D3,$46,$23,$C5,$18,$E2,$E0,$7F,$92,$7F,$07,$7C,$B5
	.dc.b	$25,$7F,$9C,$BD,$84,$EA,$30,$D7,$FA,$AC,$6A,$78,$C6,$27,$4F,$F6
	.dc.b	$A9,$FF,$67,$7C,$67,$F9,$67,$E2,$CD,$24,$FF,$94,$B2,$FF,$0E,$B7
	.dc.b	$D4,$F8,$04,$7E,$47,$95,$F1,$C8,$E0,$FF,$2D,$FD,$CF,$F6,$17,$E3
	.dc.b	$35,$F8,$AD,$5A,$A3,$0D,$F3,$3F,$CE,$D1,$82,$38,$24,$F8,$1A,$FF
	.dc.b	$7A,$A7,$DC,$FF,$91,$63,$0D,$7F,$C1,$F1,$AE,$9F,$F5,$0B,$63,$FC
	.dc.b	$EB,$E4,$7E,$89,$BD,$1F,$E8,$98,$C4,$7F,$82,$32,$FE,$CB,$3F,$EA
	.dc.b	$8F,$B5,$79,$74,$95,$FE,$AE,$8C,$B6,$32,$FF,$AD,$23,$2D,$8C,$FC
	.dc.b	$07,$F9,$3E,$DF,$8E,$FF,$62,$F8,$5F,$F3,$8C,$63,$CB,$FF,$5F,$7B
	.dc.b	$BF,$E2,$CC,$B9,$E1,$FE,$15,$5F,$80,$FF,$8C,$38,$7F,$C4,$31,$8D
	.dc.b	$3F,$59,$18,$9D,$47,$36,$BF,$B1,$9E,$0F,$E0,$5B,$3C,$63,$99,$C4
	.dc.b	$A7,$19,$F3,$BF,$84,$FC,$AD,$F8,$4E,$33,$46,$27,$F8,$06,$9F,$E8
	.dc.b	$CC,$7C,$2D,$C3,$85,$6E,$93,$F1,$A3,$0E,$E3,$9E,$7F,$C2,$F8,$5F
	.dc.b	$96,$FF,$B3,$74,$FC,$3B,$6B,$4D,$7F,$C7,$AD,$9E,$3F,$0B,$F1,$9B
	.dc.b	$3F,$01,$AF,$F9,$1D,$B5,$DA,$7E,$07,$F0,$0F,$FB,$D7,$0F,$E1,$AD
	.dc.b	$BE,$F3,$F2,$16,$8F,$53,$FC,$9F,$1F,$02,$FF,$05,$18,$CF,$E4,$36
	.dc.b	$3F,$95,$7D,$0B,$EF,$7F,$DD,$95,$AD,$3E,$2E,$1F,$7F,$0E,$0D,$8C
	.dc.b	$DF,$1C,$BC,$99,$78,$E7,$F4,$17,$E4,$BF,$C7,$B5,$F2,$C7,$E9,$7D
	.dc.b	$07,$C1,$D2,$4F,$E9,$63,$2D,$8F,$2B,$F0,$A3,$39,$7F,$B5,$32,$7D
	.dc.b	$EB,$E1,$3F,$D7,$13,$F8,$9A,$DE,$DF,$A7,$6E,$1F,$9D,$6E,$1C,$AF
	.dc.b	$8D,$6F,$1F,$8B,$93,$F0,$2F,$E7,$FC,$3F,$23,$95,$C4,$F9,$7E,$3A
	.dc.b	$FF,$A1,$2F,$6F,$D7,$3F,$92,$CE,$EF,$AC,$E9,$FE,$32,$D9,$FC,$C3
	.dc.b	$5E,$9E,$3F,$FB,$56,$76,$D8,$DB,$5D,$C7,$E5,$67,$94,$79,$FE,$C7
	.dc.b	$CF,$8F,$EC,$BD,$8F,$4C,$E6,$FD,$DD,$1F,$E5,$27,$E9,$E6,$C3,$8C
	.dc.b	$E6,$BD,$C7,$13,$E0,$61,$D6,$F8,$D7,$96,$75,$1F,$EB,$6B,$E8,$FF
	.dc.b	$0A,$8C,$9F,$EC,$A8,$CB,$FD,$4D,$B1,$C6,$7A,$46,$27,$87,$A1,$FE
	.dc.b	$FB,$EC,$FE,$5D,$DC,$FA,$DF,$7D,$AB,$58,$FF,$80,$2B,$4F,$C0,$8F
	.dc.b	$EB,$D3,$B8,$FF,$66,$E1,$FE,$17,$8C,$38,$6B,$ED,$9E,$24,$9E,$31
	.dc.b	$F8,$98,$FE,$D9,$93,$D1,$7E,$01,$86,$DF,$82,$75,$39,$3F,$BA,$71
	.dc.b	$3E,$CB,$95,$79,$A3,$17,$A9,$CE,$3F,$DB,$5A,$FF,$88,$FE,$F3,$FC
	.dc.b	$E3,$7D,$DB,$1F,$D9,$5A,$EF,$D9,$F4,$C6,$23,$FC,$52,$D7,$FC,$0A
	.dc.b	$DC,$8D,$24,$F2,$25,$7C,$6D,$1F,$5B,$F4,$89,$19,$7E,$4F,$DE,$7F
	.dc.b	$62,$8F,$8C,$DC,$7F,$BE,$35,$FF,$20,$35,$FF,$28,$7A,$2F,$CE,$DF
	.dc.b	$53,$FE,$1F,$8C,$36,$33,$7F,$E3,$F6,$D7,$F8,$F3,$CA,$7C,$4F,$16
	.dc.b	$D7,$F9,$15,$B3,$B9,$D1,$27,$4D,$8F,$4F,$83,$EE,$24,$FB,$9F,$F1
	.dc.b	$97,$E0,$EF,$A3,$A4,$9F,$C7,$C6,$5F,$CC,$FC,$5B,$A9,$CD,$FE,$3D
	.dc.b	$8C,$36,$33,$3C,$1B,$8F,$F9,$DC,$FF,$3C,$57,$09,$D5,$71,$23,$E0
	.dc.b	$F9,$97,$A3,$DD,$B1,$98,$C1,$1C,$12,$FC,$02,$D5,$CA,$33,$2F,$D1
	.dc.b	$7F,$D4,$2F,$EE,$0C,$7A,$5F,$E5,$B6,$99,$8D,$CF,$EF,$BE,$CD,$F1
	.dc.b	$BF,$E5,$F6,$D7,$EE,$9B,$3B,$9D,$12,$74,$D8,$F4,$E3,$E8,$FA,$BF
	.dc.b	$84,$4A,$F8,$CD,$9D,$E4,$8C,$47,$D1,$AE,$55,$F3,$7E,$47,$E6,$11
	.dc.b	$C1,$27,$9C,$65,$B1,$FC,$73,$C9,$7E,$40,$F2,$FF,$D0,$11,$86,$C6
	.dc.b	$67,$83,$6B,$E4,$7B,$BF,$1C,$93,$F1,$DB,$1A,$8C,$11,$F7,$F3,$FE
	.dc.b	$14,$91,$F8,$96,$C6,$A3,$04,$7E,$3C,$FE,$34,$91,$F9,$16,$C6,$A3
	.dc.b	$04,$7F,$1B,$87,$C2,$F8,$EF,$79,$27,$DE,$FE,$0F,$D2,$91,$F8,$36
	.dc.b	$D7,$E2,$5A,$FE,$63,$83,$FC,$79,$3B,$3F,$49,$EB,$FE,$8C,$8E,$09
	.dc.b	$3F,$82,$8C,$B6,$3F,$59,$E3,$7E,$AC,$8E,$09,$1F,$8A,$8C,$B6,$3F
	.dc.b	$AD,$F3,$BE,$DB,$F3,$73,$FB,$D8,$C7,$27,$FB,$05,$B5,$C2,$B7,$5F
	.dc.b	$C0,$FF,$20,$78,$3F,$EC,$56,$D7,$F6,$06,$CF,$F6,$A6,$BF,$BC,$6B
	.dc.b	$CE,$79,$12,$7F,$DD,$DF,$3F,$FC,$6E,$7F,$DD,$6D,$7F,$DD,$0D,$7F
	.dc.b	$9E,$B5,$FF,$7E,$F7,$7F,$B5,$89,$3F,$DD,$3C,$74,$7F,$41,$F9,$1F
	.dc.b	$6D,$3C,$44,$61,$BA,$EC,$FF,$72,$71,$FF,$80,$AB,$F9,$33,$51,$FE
	.dc.b	$64,$F2,$F6,$70,$EE,$99,$FF,$81,$AB,$F8,$9D,$7E,$52,$72,$FE,$DD
	.dc.b	$AE,$67,$F9,$EF,$E3,$5E,$FE,$C3,$49,$E1,$FF,$BE,$7D,$7F,$F0,$6D
	.dc.b	$FF,$87,$1A,$FF,$2E,$6B,$FC,$FD,$AF,$FA,$E3,$97,$91,$30,$FF,$0E
	.dc.b	$CE,$97,$E5,$2F,$80,$FE,$60,$D8,$FE,$68,$DB,$FE,$72,$D7,$F9,$3F
	.dc.b	$DE,$A7,$FC,$53,$9A,$31,$3F,$D7,$1A,$F0,$9D,$C6,$23,$F9,$43,$5F
	.dc.b	$EB,$2D,$7F,$A6,$3F,$E3,$4F,$99,$FF,$47,$4F,$F8,$C5,$B3,$FD,$4B
	.dc.b	$91,$18,$8F,$EC,$91,$8F,$6B,$A7,$E3,$5F,$E2,$E9,$FE,$D0,$D7,$F0
	.dc.b	$6D,$4D,$27,$A7,$C8,$98,$7C,$8F,$C6,$FE,$59,$1C,$1F,$E4,$E9,$47
	.dc.b	$88,$91,$96,$D7,$F9,$72,$77,$39,$24,$7F,$24,$C7,$FC,$EC,$FF,$AF
	.dc.b	$1B,$C7,$F0,$A3,$FA,$3C,$69,$FD,$47,$57,$CD,$E7,$FD,$EC,$7F,$8C
	.dc.b	$23,$04,$BF,$F3,$2C,$EA,$78,$C6,$27,$FA,$9F,$D8,$7F,$AB,$A7,$F5
	.dc.b	$AD,$7F,$5B,$82,$31,$1F,$ED,$26,$BF,$BB,$DB,$C8,$4B,$F0,$BE,$7F
	.dc.b	$94,$91,$E4,$6C,$FF,$A0,$B1,$46,$23,$FC,$2B,$18,$9E,$07,$F9,$E7
	.dc.b	$EA,$DF,$65,$FE,$80,$6C,$66,$30,$47,$04,$AF,$C8,$4F,$07,$F2,$5F
	.dc.b	$59,$FE,$6D,$27,$AF,$FE,$8C,$6D,$7F,$A2,$5B,$3C,$27,$FC,$BD,$F8
	.dc.b	$73,$E6,$69,$2F,$DD,$19,$7F,$D2,$5F,$59,$FE,$7B,$25,$7F,$A6,$9B
	.dc.b	$1A,$8C,$11,$FF,$50,$DF,$F2,$2E,$AA,$9A,$47,$FD,$5F,$E1,$BE,$AF
	.dc.b	$FC,$7F,$18,$6B,$FE,$44,$8D,$7A,$5F,$EA,$E6,$C7,$2F,$82,$FF,$30
	.dc.b	$37,$C9,$D4,$62,$39,$35,$FE,$6F,$D8,$FA,$B7,$C5,$7F,$47,$6C,$FF
	.dc.b	$AD,$5A,$FF,$05,$6B,$FC,$11,$AF,$46,$A9,$ED,$3D,$4F,$D1,$FD,$5B
	.dc.b	$ED,$29,$2B,$F2,$51,$97,$FD,$8D,$F5,$8F,$AC,$FE,$B8,$D8,$CC,$60
	.dc.b	$8E,$08,$7F,$1D,$8E,$0F,$F1,$FF,$9A,$AF,$F1,$E6,$BF,$19,$AF,$AD
	.dc.b	$B8,$9A,$4A,$FE,$E0,$F0,$97,$D2,$7F,$94,$5F,$F7,$3B,$5F,$F7,$8B
	.dc.b	$5F,$F4,$1B,$5F,$F4,$A3,$5F,$ED,$ED,$7A,$FB,$3F,$92,$8C,$7D,$5B
	.dc.b	$DB,$FC,$AB,$6B,$FD,$EE,$D7,$FE,$00,$B5,$FF,$49,$FD,$67,$F5,$D7
	.dc.b	$FD,$FA,$D7,$FD,$2E,$1C,$3C,$3C,$BE,$FF,$2D,$23,$BB,$5E,$ED,$7F
	.dc.b	$D4,$34,$8C,$47,$F2,$0F,$99,$FF,$08,$93,$A7,$FC,$14,$FF,$5D,$87
	.dc.b	$96,$FA,$CF,$D9,$B6,$3C,$E6,$A3,$24,$73,$1B,$7D,$AD,$7F,$D6,$AD
	.dc.b	$7F,$D8,$D9,$3D,$3F,$E0,$2D,$8F,$8F,$18,$3F,$82,$FD,$63,$EB,$29
	.dc.b	$27,$CB,$19,$6C,$65,$F9,$F1,$96,$C7,$F8,$F7,$CC,$7D,$15,$24,$FF
	.dc.b	$7D,$7F,$8A,$FB,$1F,$76,$3F,$DB,$CD,$7F,$C7,$2D,$7F,$C8,$7C,$3C
	.dc.b	$08,$FC,$A8,$D1,$BF,$AC,$7C,$4F,$F2,$26,$CF,$D0,$6B,$F8,$86,$BF
	.dc.b	$EE,$09,$FB,$44,$E0,$F3,$92,$3C,$0D,$7F,$23,$1F,$E1,$DF,$53,$FD
	.dc.b	$FF,$F5,$9F,$E2,$AF,$89,$34,$95,$F8,$2F,$AC,$FF,$70,$B7,$DD,$FE
	.dc.b	$73,$18,$8F,$CE,$B5,$FF,$7A,$B5,$FF,$2F,$87,$0F,$F9,$89,$AE,$63
	.dc.b	$AF,$C3,$48,$F1,$35,$FE,$DB,$2F,$8A,$FF,$03,$B6,$3F,$BA,$35,$DC
	.dc.b	$7A,$3E,$16,$47,$B5,$AF,$F7,$8E,$C7,$D2,$7F,$84,$9B,$1F,$DF,$DA
	.dc.b	$EE,$3F,$AD,$FD,$5F,$EC,$E3,$F5,$2D,$7F,$80,$B5,$F8,$61,$C3,$F1
	.dc.b	$9A,$FF,$A1,$9A,$EF,$B3,$B9,$FD,$AB,$EB,$1F,$13,$49,$3F,$BA,$8C
	.dc.b	$B7,$87,$81,$1C,$10,$FE,$2F,$1F,$E9,$E7,$07,$A1,$C3,$AF,$2B,$C4
	.dc.b	$47,$C0,$D7,$46,$08,$7C,$4F,$85,$F9,$24,$7E,$4B,$5D,$18,$21,$E6
	.dc.b	$FA,$CF,$84,$47,$E0,$3F,$17,$D5,$91,$F0,$DA,$FB,$9A,$FB,$5A,$FE
	.dc.b	$62,$FE,$FB,$F0,$74,$FA,$04,$7E,$73,$5F,$92,$D7,$EF,$FD,$53,$D4
	.dc.b	$FF,$10,$B6,$3F,$97,$B5,$FE,$8E,$D7,$8C,$70,$23,$C1,$FF,$15,$F8
	.dc.b	$DF,$E5,$28,$FE,$72,$D7,$FA,$8B,$5F,$C8,$B5,$FD,$AA,$FA,$CE,$64
	.dc.b	$8E,$6D,$79,$B5,$E6,$F8,$BE,$B2,$47,$89,$AF,$89,$AF,$89,$AF,$F0
	.dc.b	$6B,$7C,$4F,$C6,$24,$7C,$66,$BF,$19,$AF,$CE,$7E,$1F,$89,$23,$E1
	.dc.b	$B5,$F8,$7F,$0B,$F2,$08,$FC,$86,$BA,$30,$43,$FB,$DF,$D6,$7F,$4A
	.dc.b	$7F,$B3,$B5,$CF,$87,$FE,$5E,$8C,$47,$F0,$06,$BF,$C9,$1A,$83,$86
	.dc.b	$C7,$F9,$A3,$EB,$3F,$C5,$51,$FD,$D9,$AF,$F8,$39,$AF,$F0,$E6,$BF
	.dc.b	$E5,$AF,$AA,$FC,$DB,$77,$FE,$FF,$E8,$A7,$33,$FE,$53,$6F,$C0,$BE
	.dc.b	$B1,$F4,$5F,$EB,$66,$D7,$FA,$8D,$AF,$FA,$89,$AF,$F4,$29,$E5,$3C
	.dc.b	$49,$1C,$5F,$F6,$0F,$D6,$3E,$52,$92,$7F,$D6,$11,$97,$FC,$5B,$8F
	.dc.b	$A7,$D2,$F5,$E4,$69,$27,$E4,$78,$FF,$2C,$8E,$0E,$7F,$58,$F2,$BE
	.dc.b	$2B,$63,$31,$82,$38,$21,$FC,$FB,$04,$F1,$8C,$35,$FE,$CB,$1A,$9E
	.dc.b	$71,$A7,$FC,$67,$4F,$AC,$FF,$24,$CF,$E1,$E3,$1D,$3A,$9C,$CA,$72
	.dc.b	$FC,$6F,$1B,$F5,$AD,$E7,$FC,$0C,$62,$3D,$0D,$7F,$AB,$FD,$4B,$C9
	.dc.b	$FF,$5C,$B6,$3F,$C8,$ED,$7F,$CB,$ED,$78,$C7,$02,$3C,$1A,$BD,$FF
	.dc.b	$9D,$A3,$E8,$BF,$B5,$7C,$27,$FA,$0E,$31,$E4,$7F,$83,$23,$FD,$09
	.dc.b	$8F,$86,$FE,$72,$F9,$7F,$F0,$8F,$07,$A9,$B9,$7F,$85,$27,$4F,$EC
	.dc.b	$67,$4D,$35,$3C,$1F,$DB,$4F,$07,$ED,$36,$78,$C6,$27,$CD,$ED,$FF
	.dc.b	$B2,$BE,$95,$F5,$98,$AF,$F6,$04,$61,$FF,$72,$4F,$81,$A7,$82,$B8
	.dc.b	$68,$AE,$13,$FE,$E9,$AE,$D7,$D6,$18,$F9,$F3,$FF,$56,$70,$BE,$B3
	.dc.b	$FD,$AF,$A7,$F8,$DA,$31,$F0,$9E,$F6,$E1,$F0,$CF,$C6,$57,$07,$F2
	.dc.b	$4D,$73,$3C,$63,$13,$CA,$7F,$44,$DB,$FB,$9F,$53,$FE,$24,$6C,$FC
	.dc.b	$59,$D6,$7F,$5F,$D2,$FC,$E4,$FB,$CF,$A6,$E2,$BF,$06,$D9,$FE,$59
	.dc.b	$F5,$4F,$AC,$FF,$7A,$D6,$B0,$F9,$B8,$7C,$6C,$38,$3D,$23,$E8,$DF
	.dc.b	$1C,$FE,$66,$AB,$F1,$DF,$E9,$9A,$F9,$EF,$F4,$B1,$88,$FC,$4F,$9E
	.dc.b	$F9,$4F,$D4,$B6,$3F,$52,$D7,$F6,$2D,$7F,$30,$EF,$2F,$D0,$B7,$4F
	.dc.b	$7F,$D6,$7F,$AF,$B8,$3D,$B9,$AB,$58,$78,$B1,$CC,$63,$56,$1C,$6B
	.dc.b	$E7,$1F,$8A,$9E,$3A,$3E,$B3,$FA,$1E,$5F,$96,$E4,$70,$BE,$A3,$F9
	.dc.b	$46,$FF,$90,$BD,$E5,$F6,$6F,$75,$AD,$FF,$8E,$56,$B1,$FC,$A3,$6F
	.dc.b	$3E,$27,$F3,$39,$CB,$63,$31,$F9,$7C,$FC,$D3,$99,$FE,$49,$5C,$DB
	.dc.b	$19,$7D,$1F,$59,$E3,$7F,$6B,$EA,$D7,$38,$7B,$FA,$13,$9C,$D7,$C7
	.dc.b	$4F,$DB,$4E,$63,$15,$C2,$FF,$23,$7C,$1F,$D7,$61,$D8,$FE,$8D,$19
	.dc.b	$6D,$79,$E3,$13,$FE,$8C,$6C,$FF,$AC,$1A,$FF,$83,$5A,$FD,$5F,$3B
	.dc.b	$FA,$E7,$5F,$F3,$0E,$D7,$D6,$7A,$E7,$38,$7F,$85,$74,$FE,$69,$39
	.dc.b	$8F,$F4,$DE,$1D,$6B,$FA,$E3,$63,$2F,$76,$E2,$9D,$CE,$4E,$71,$99
	.dc.b	$DC,$F4,$B7,$9E,$F9,$EC,$36,$BF,$C0,$6D,$C3,$FB,$93,$5F,$E3,$DF
	.dc.b	$59,$CA,$7D,$3C,$48,$C4,$FF,$AD,$A3,$15,$EC,$6D,$7F,$84,$3E,$B3
	.dc.b	$FC,$F8,$DF,$5B,$CD,$18,$BF,$F0,$13,$5F,$EF,$ED,$7E,$63,$56,$48
	.dc.b	$C3,$67,$A9,$2B,$C0,$4A,$F9,$3F,$59,$F3,$35,$34,$93,$FE,$28,$F9
	.dc.b	$EF,$EC,$CD,$F1,$3E,$74,$62,$FF,$C3,$4D,$7F,$C7,$5F,$56,$FA,$CF
	.dc.b	$BF,$E9,$23,$18,$7E,$12,$31,$3A,$9C,$D7,$FB,$8A,$7F,$C1,$9F,$0F
	.dc.b	$CB,$CA,$34,$98,$73,$8C,$B6,$33,$5C,$E7,$2F,$A3,$C0,$F5,$92,$BE
	.dc.b	$03,$F9,$0F,$AC,$7D,$2D,$24,$FE,$62,$32,$FE,$7B,$E2,$3F,$C2,$DC
	.dc.b	$9F,$9E,$6C,$67,$AF,$FC,$60,$D9,$D4,$63,$1F,$F8,$03,$5C,$1B,$19
	.dc.b	$7F,$D0,$42,$34,$E1,$7C,$CB,$CF,$F1,$37,$0F,$33,$6B,$BB,$5E,$4D
	.dc.b	$7F,$D2,$15,$D2,$B9,$92,$BF,$7A,$3E,$B3,$FD,$6A,$E1,$B5,$E8,$3F
	.dc.b	$CD,$0E,$1F,$F1,$8B,$5F,$6F,$D5,$3E,$2B,$F8,$0B,$6B,$F8,$3B,$67
	.dc.b	$F8,$6B,$5E,$CD,$79,$CF,$22,$4F,$C2,$FA,$CF,$8A,$47,$04,$9F,$8B
	.dc.b	$19,$6C,$7C,$AF,$8E,$F9,$A4,$70,$48,$F9,$B1,$96,$C7,$F2,$4F,$35
	.dc.b	$F5,$8F,$2B,$F9,$2C,$61,$B1,$99,$E0,$DA,$F8,$3E,$EF,$C6,$24,$FC
	.dc.b	$66,$C6,$A3,$04,$7F,$03,$CB,$F8,$62,$47,$E1,$9B,$1A,$8C,$11,$FC
	.dc.b	$4C,$FE,$38,$91,$F8,$E6,$C6,$A3,$04,$7F,$09,$87,$C2,$F8,$AF,$79
	.dc.b	$27,$DE,$FE,$0B,$D2,$91,$F8,$26,$D7,$E0,$9A,$FE,$09,$AF,$E6,$B8
	.dc.b	$BF,$8B,$37,$B7,$F4,$A4,$AF,$D2,$B6,$3F,$4A,$D7,$F4,$BE,$2F,$EB
	.dc.b	$09,$3F,$AC,$6C,$7E,$B1,$AF,$08,$FE,$C7,$CA,$FA,$CF,$CB,$4F,$FA
	.dc.b	$75,$B1,$9E,$2F,$EC,$ED,$AF,$EC,$FF,$58,$F0,$3F,$B2,$B6,$33,$18
	.dc.b	$23,$82,$61,$FD,$D2,$78,$3F,$CB,$FE,$7F,$F8,$DB,$7B,$3F,$9C,$C6
	.dc.b	$23,$FD,$EC,$D7,$FD,$AB,$DB,$FD,$DB,$C5,$FF,$25,$61,$FD,$F1,$B3
	.dc.b	$FE,$F2,$9F,$E8,$7F,$1B,$F0,$3C,$13,$49,$AF,$F8,$03,$93,$FB,$F5
	.dc.b	$FE,$9F,$CA,$C1,$32,$FF,$04,$CF,$F0,$99,$1F,$E6,$9F,$73,$91,$27
	.dc.b	$6D,$8F,$D9,$35,$FE,$55,$F5,$9E,$D7,$F9,$73,$7C,$13,$4D,$43,$D4
	.dc.b	$FF,$05,$3F,$ED,$1E,$5F,$F0,$CC,$62,$3C,$AD,$78,$46,$89,$87,$F8
	.dc.b	$92,$78,$2F,$AC,$7C,$07,$F8,$AA,$30,$D8,$CC,$7F,$34,$6D,$FF,$55
	.dc.b	$FA,$9F,$F1,$5E,$46,$92,$BF,$C6,$71,$96,$C6,$67,$94,$7F,$8F,$38
	.dc.b	$9F,$2D,$FE,$83,$9F,$EA,$B1,$8F,$9A,$FF,$1A,$4E,$67,$FC,$8C,$D7
	.dc.b	$EB,$FD,$5F,$F8,$B1,$FE,$8C,$D7,$FA,$FB,$5D,$FA,$5F,$E4,$48,$C4
	.dc.b	$1A,$48,$F8,$9F,$59,$31,$F8,$8D,$AF,$F2,$AD,$23,$A1,$27,$FC,$AB
	.dc.b	$19,$6C,$66,$78,$B5,$FE,$EB,$C5,$F1,$1F,$E7,$78,$7F,$9E,$1F,$F5
	.dc.b	$E3,$63,$55,$FE,$62,$E6,$FD,$9D,$7F,$88,$7A,$FF,$CC,$31,$88,$FF
	.dc.b	$00,$B5,$FC,$FB,$53,$49,$1F,$D3,$3E,$B3,$FD,$5B,$3E,$66,$BF,$A8
	.dc.b	$B4,$62,$3F,$DA,$4D,$7A,$ED,$E4,$26,$1E,$97,$CF,$F5,$92,$3D,$8D
	.dc.b	$9F,$F2,$03,$5F,$F1,$F3,$5F,$F4,$57,$D5,$BE,$B3,$FC,$F8,$DA,$FC
	.dc.b	$83,$67,$FC,$19,$C2,$8C,$47,$E4,$9A,$F4,$AE,$64,$AF,$CD,$7D,$67
	.dc.b	$F9,$A8,$9E,$BF,$FA,$2D,$B1,$A8,$C1,$1F,$F2,$D7,$D6,$3E,$67,$FC
	.dc.b	$BE,$D8,$CC,$60,$FF,$4B,$7D,$67,$F9,$E0,$95,$82,$4F,$EB,$63,$2D
	.dc.b	$8F,$F5,$1E,$3F,$C2,$8F,$F5,$49,$FC,$44,$BD,$58,$70,$7F,$8C,$7C
	.dc.b	$27,$D6,$7F,$91,$A3,$FD,$6B,$18,$F2,$7F,$D6,$6D,$8C,$CE,$DE,$5F
	.dc.b	$04,$F2,$36,$D8,$E6,$D7,$71,$FC,$1F,$EB,$9F,$37,$FE,$5A,$86,$9F
	.dc.b	$EC,$99,$CD,$7F,$3A,$9C,$BF,$EA,$08,$CB,$E1,$9F,$F6,$87,$D6,$BE
	.dc.b	$33,$A4,$73,$C2,$D7,$16,$F2,$B0,$E7,$8F,$68,$63,$CE,$FC,$58,$7F
	.dc.b	$B7,$7E,$AD,$F5,$7F,$D7,$5B,$3F,$C8,$DA,$FE,$51,$AF,$F2,$4F,$C6
	.dc.b	$DE,$1F,$89,$B1,$D8,$FF,$5E,$B8,$7D,$C7,$FB,$02,$7A,$E5,$E7,$9D
	.dc.b	$36,$BA,$1E,$E9,$DE,$1F,$D9,$3E,$CF,$FD,$0B,$F4,$1F,$11,$B8,$F8
	.dc.b	$5A,$F8,$34,$7D,$6F,$28,$4E,$B2,$FA,$AD,$C3,$B7,$23,$1F,$9F,$87
	.dc.b	$1C,$1F,$4A,$F5,$FE,$AC,$61,$B1,$98,$F8,$87,$0F,$6F,$33,$23,$33
	.dc.b	$B7,$E3,$9B,$38,$FD,$5F,$F6,$87,$FD,$24,$7F,$81,$9C,$38,$6B,$FE
	.dc.b	$97,$6B,$F2,$7C,$B7,$D6,$7C,$96,$C7,$C9,$6B,$E2,$9E,$BB,$7E,$6B
	.dc.b	$C6,$7C,$E7,$F8,$74,$9B,$FF,$86,$9B,$7F,$E8,$56,$BF,$E6,$67,$FC
	.dc.b	$5D,$F1,$EF,$80,$FC,$DB,$67,$F1,$5C,$68,$C4,$7E,$2B,$0F,$0E,$1F
	.dc.b	$43,$27,$AB,$FC,$29,$B5,$F1,$83,$87,$E3,$46,$C8,$EE,$34,$FC,$98
	.dc.b	$E5,$C9,$F8,$8F,$A9,$FC,$82,$69,$E3,$8F,$DA,$F9,$CF,$A3,$FF,$25
	.dc.b	$36,$FF,$C2,$0D,$7F,$A0,$B5,$FF,$08,$B5,$FF,$07,$C7,$22,$3F,$82
	.dc.b	$B7,$E3,$8F,$9D,$FD,$23,$67,$F4,$8D,$7F,$49,$7E,$1E,$1F,$E2,$B0
	.dc.b	$FA,$B7,$C6,$7F,$34,$6C,$7C,$B6,$BF,$92,$6B,$FE,$E5,$6B,$FE,$FE
	.dc.b	$6B,$FE,$5C,$6B,$DE,$3F,$C6,$FB,$3E,$2F,$F8,$FF,$2F,$F2,$C6,$C6
	.dc.b	$6B,$F5,$77,$E2,$C3,$F9,$6D,$F1,$DF,$FC,$83,$87,$D5,$BF,$D3,$CE
	.dc.b	$5B,$19,$7F,$DF,$91,$96,$C6,$5F,$F3,$98,$8D,$38,$6B,$FE,$78,$E3
	.dc.b	$7E,$2E,$F7,$7F,$9B,$B6,$7F,$3A,$D7,$F9,$2D,$78,$71,$FA,$7F,$58
	.dc.b	$F8,$8F,$EC,$AD,$AF,$CE,$35,$FE,$06,$D7,$CE,$1C,$3E,$A6,$BF,$E8
	.dc.b	$E6,$BB,$8E,$FF,$87,$7F,$1C,$93,$4C,$46,$7E,$22,$57,$F1,$EA,$47
	.dc.b	$F5,$8F,$6B,$C5,$F1,$DF,$D6,$3D,$6F,$E1,$B7,$FE,$6E,$6B,$F2,$5A
	.dc.b	$B6,$FF,$15,$77,$FF,$AC,$FE,$CD,$F2,$FF,$D8,$9B,$5F,$C2,$1A,$FF
	.dc.b	$50,$C7,$F9,$BF,$D4,$7F,$36,$FA,$CF,$E5,$D3,$F1,$9A,$FD,$E6,$BF
	.dc.b	$EB,$26,$A6,$93,$C5,$FF,$0C,$BC,$38,$5F,$89,$7F,$71,$AF,$EC,$31
	.dc.b	$8F,$4B,$FB,$DB,$63,$FA,$2D,$7F,$41,$8F,$EA,$7F,$8A,$7F,$20,$6F
	.dc.b	$8D,$FD,$7E,$31,$D9,$E7,$F8,$AC,$4C,$C3,$FB,$15,$69,$F5,$4F,$F6
	.dc.b	$7F,$0D,$F1,$14,$93,$E2,$F5,$AB,$FD,$E3,$C5,$FE,$35,$38,$CF,$F8
	.dc.b	$3E,$8B,$F9,$9F,$5F,$30,$F5,$BF,$D9,$3E,$C1,$F1,$B9,$9F,$EB,$B6
	.dc.b	$AC,$61,$FC,$4A,$F7,$5F,$DE,$FF,$25,$7D,$D7,$F7,$F6,$C6,$63,$8F
	.dc.b	$0F,$F8,$17,$91,$F3,$9F,$E0,$56,$D7,$F8,$15,$AF,$8B,$E0,$B5,$32
	.dc.b	$34,$FE,$1E,$3F,$C1,$DF,$4F,$C3,$3F,$F0,$3B,$70,$FF,$0A,$36,$FF
	.dc.b	$11,$D1,$C3,$09,$1F,$E1,$59,$D3,$F8,$B9,$FF,$04,$FC,$7B,$E2,$3F
	.dc.b	$CC,$6D,$9F,$EB,$4D,$7F,$C0,$2D,$7F,$7E,$D7,$F8,$BB,$51,$FE,$F0
	.dc.b	$8E,$A7,$FA,$DF,$D5,$FE,$F7,$1F,$E3,$9E,$CF,$F0,$EF,$AB,$FB,$99
	.dc.b	$CC,$FF,$85,$2F,$C5,$A7,$F8,$A3,$F2,$3F,$F3,$9C,$FF,$8C,$5A,$FE
	.dc.b	$25,$B1,$FD,$3A,$7F,$CA,$FA,$FF,$8E,$E7,$2D,$F8,$2F,$F1,$E4,$62
	.dc.b	$FD,$35,$C1,$FF,$1C,$F0,$7F,$85,$2F,$F0,$13,$F7,$CF,$97,$ED,$5A
	.dc.b	$FC,$04,$57,$D5,$1F,$43,$D5,$F9,$E4,$8F,$9E,$D7,$E7,$B5,$F9,$EF
	.dc.b	$F8,$37,$C5,$7E,$2B,$FE,$4D,$6F,$C0,$7F,$9D,$A3,$FC,$91,$87,$F8
	.dc.b	$47,$5F,$91,$E0,$7B,$89,$8F,$F9,$1A,$5F,$21,$88,$F9,$53,$6B,$E5
	.dc.b	$4F,$F9,$87,$DC,$FF,$13,$57,$F9,$85,$AF,$EE,$5B,$3F,$DF,$BE,$CF
	.dc.b	$FC,$01,$B7,$F9,$AE,$72,$DF,$77,$A1,$FE,$5C,$9D,$DF,$F3,$3F,$B0
	.dc.b	$74,$7C,$16,$C7,$F9,$53,$74,$62,$3F,$DB,$AD,$7F,$E0,$0F,$D9,$3E
	.dc.b	$A3,$FA,$33,$63,$F2,$4D,$7F,$CF,$53,$CF,$4F,$F4,$F7,$E9,$B7,$4F
	.dc.b	$EE,$0D,$AF,$E9,$0D,$EA,$61,$CE,$F8,$E7,$FE,$AA,$AE,$19,$9E,$15
	.dc.b	$BC,$7F,$BE,$65,$FA,$2C,$BF,$C0,$77,$EE,$EA,$F8,$44,$8F,$F3,$AE
	.dc.b	$7E,$0F,$43,$DA,$48,$FF,$2E,$DF,$EB,$34,$FF,$5C,$7D,$97,$F8,$CA
	.dc.b	$7A,$B5,$FF,$06,$35,$FE,$C8,$DF,$25,$FA,$FF,$F8,$EA,$BF,$D3,$6D
	.dc.b	$D3,$FD,$50,$DB,$E3,$1F,$EA,$98,$C5,$72,$8C,$1F,$EC,$EB,$CC,$C8
	.dc.b	$CD,$7F,$9D,$A7,$A6,$8F,$4F,$FC,$39,$3F,$19,$AF,$F8,$95,$AF,$F7
	.dc.b	$D6,$E1,$EA,$FB,$BF,$21,$27,$FD,$07,$3F,$E2,$7C,$BF,$CA,$FE,$17
	.dc.b	$F8,$85,$FF,$70,$7D,$5F,$F9,$77,$E3,$DD,$1F,$E3,$38,$C3,$FE,$3A
	.dc.b	$AE,$1A,$DF,$67,$FD,$D7,$5A,$CB,$FC,$20,$D5,$5C,$1C,$FC,$27,$F9
	.dc.b	$13,$1E,$F1,$83,$F9,$7B,$8D,$97,$CE,$F9,$57,$4F,$A0,$E5,$9F,$F9
	.dc.b	$27,$EB,$1F,$67,$8A,$FF,$11,$46,$1E,$B3,$EB,$AD,$47,$33,$27,$0A
	.dc.b	$E5,$8A,$F9,$4B,$2F,$4B,$C4,$C6,$1D,$5E,$75,$C6,$F9,$DF,$2B,$E3
	.dc.b	$8B,$1F,$1B,$70,$E3,$5B,$C7,$DD,$86,$AF,$DD,$F0,$4F,$B3,$FF,$14
	.dc.b	$65,$F2,$7E,$C3,$FC,$B6,$DC,$B7,$3A,$C7,$FA,$ED,$F1,$F6,$BD,$F8
	.dc.b	$70,$9F,$F3,$A3,$6F,$FD,$1D,$7C,$5B,$5C,$18,$7C,$AB,$E8,$FC,$C3
	.dc.b	$F0,$7E,$FF,$FA,$4F,$1F,$82,$DC,$7F,$1C,$D9,$7D,$93,$EC,$FC,$11
	.dc.b	$CB,$0E,$9B,$FF,$9A,$1B,$7C,$6F,$86,$5F,$32,$3F,$D4,$57,$FE,$7E
	.dc.b	$BE,$0F,$FA,$00,$E8,$57,$2E,$FD,$87,$C6,$FF,$B9,$67,$FC,$6A,$D8
	.dc.b	$CC,$F1,$6B,$C6,$3C,$F1,$AD,$3F,$43,$8B,$EC,$DF,$24,$E8,$DE,$13
	.dc.b	$1F,$C6,$9F,$E1,$4C,$BC,$2E,$30,$E5,$7C,$63,$67,$EC,$B2,$F0,$6F
	.dc.b	$F8,$0F,$B3,$FE,$95,$97,$E5,$B8,$9F,$63,$C7,$6F,$DA,$F0,$AF,$79
	.dc.b	$78,$B9,$D8,$70,$C7,$F0,$15,$AC,$3F,$25,$8F,$3C,$BC,$1E,$45,$CB
	.dc.b	$FD,$2D,$8F,$17,$DB,$9F,$79,$D6,$7D,$B0,$E1,$39,$CF,$D1,$1C,$EF
	.dc.b	$D3,$E5,$70,$E9,$FF,$37,$57,$E4,$69,$39,$F7,$B8,$46,$27,$F5,$F1
	.dc.b	$89,$D4,$71,$6D,$FC,$CA,$E0,$FC,$3B,$EE,$E6,$77,$AF,$F0,$6A,$DB
	.dc.b	$FC,$76,$B6,$FF,$1F,$EB,$FF,$5A,$E1,$FC,$21,$B8,$77,$F4,$7E,$E7
	.dc.b	$7F,$F3,$8F,$85,$F4,$C6,$2F,$F9,$C4,$61,$FF,$2C,$CE,$DF,$85,$3B
	.dc.b	$AF,$E6,$F5,$B5,$5B,$7F,$1F,$E1,$FF,$66,$F8,$1E,$0D,$C3,$A9,$FE
	.dc.b	$F9,$F3,$FF,$D4,$AE,$29,$3B,$7A,$C6,$FE,$07,$FB,$B7,$17,$F8,$E1
	.dc.b	$B7,$CA,$BF,$DE,$73,$9A,$FD,$B5,$FF,$78,$E7,$F1,$F9,$1F,$EA,$69
	.dc.b	$F0,$53,$EC,$9F,$37,$AB,$95,$E5,$C3,$93,$C2,$B6,$E7,$0E,$98,$FF
	.dc.b	$AE,$DB,$AF,$F8,$1B,$0F,$95,$97,$7C,$BA,$DF,$B8,$FF,$79,$63,$D1
	.dc.b	$F5,$B5,$FF,$11,$B5,$FF,$47,$CF,$3C,$7F,$C4,$7E,$3F,$EC,$31,$FD
	.dc.b	$E7,$9A,$8C,$4F,$0F,$A2,$FE,$94,$FF,$A4,$E3,$38,$30,$56,$A3,$E6
	.dc.b	$B5,$FF,$1A,$B5,$E9,$E3,$FD,$6F,$C4,$1F,$8B,$7C,$06,$E3,$DF,$E4
	.dc.b	$F1,$32,$3E,$9C,$EA,$FE,$26,$7E,$9B,$D4,$63,$34,$7E,$56,$75,$7F
	.dc.b	$73,$2F,$A2,$FF,$AD,$23,$37,$F3,$F3,$F6,$EF,$F8,$CB,$FB,$33,$EC
	.dc.b	$8C,$1F,$E4,$5E,$1F,$F2,$0F,$8F,$F8,$A3,$F4,$9F,$1F,$FA,$2E,$FC
	.dc.b	$67,$F8,$39,$B8,$7D,$DA,$D4,$E4,$FF,$5D,$77,$FE,$9B,$DE,$FE,$07
	.dc.b	$C7,$FC,$CE,$9F,$31,$FE,$3F,$BD,$93,$0F,$F6,$06,$1E,$29,$D4,$62
	.dc.b	$7F,$31,$F6,$4E,$BF,$D9,$B6,$3F,$60,$3E,$C9,$F2,$9A,$AF,$E0,$13
	.dc.b	$98,$FF,$33,$34,$FE,$C4,$5B,$FF,$70,$9F,$EE,$27,$0E,$1A,$E6,$BB
	.dc.b	$C7,$26,$C7,$23,$FA,$E3,$87,$AE,$EF,$82,$7E,$1D,$FC,$39,$B7,$FC
	.dc.b	$2F,$5F,$93,$F6,$59,$A9,$3F,$BB,$AD,$E1,$F5,$75,$5E,$A3,$FD,$79
	.dc.b	$9F,$F1,$AF,$2B,$CB,$C1,$F5,$6D,$F2,$19,$A9,$97,$FB,$15,$FA,$91
	.dc.b	$8C,$3F,$37,$F3,$7F,$15,$A7,$F2,$DF,$45,$F6,$5F,$E6,$CC,$3F,$92
	.dc.b	$75,$AF,$86,$7F,$DD,$EF,$7A,$7F,$A6,$5B,$7F,$E1,$56,$E3,$F2,$DA
	.dc.b	$FD,$66,$A3,$FC,$2F,$8F,$85,$E0,$D7,$83,$5E,$1C,$2F,$B9,$FC,$8E
	.dc.b	$3F,$E0,$B8,$C7,$B5,$F9,$5E,$FF,$E8,$75,$AC,$B8,$B7,$0F,$E8,$2D
	.dc.b	$E4,$FD,$0D,$70,$BF,$BD,$E4,$CC,$3E,$F7,$13,$EC,$F1,$97,$F2,$BC
	.dc.b	$FF,$0D,$78,$D3,$EC,$55,$C3,$D3,$D2,$EC,$FD,$29,$F4,$F7,$FC,$74
	.dc.b	$FE,$AA,$31,$CF,$23,$FA,$C6,$6F,$B4,$E5,$1F,$D6,$F8,$BE,$DD,$6E
	.dc.b	$BF,$9F,$46,$27,$8D,$70,$9C,$E1,$E0,$7D,$5F,$85,$BE,$43,$FB,$5B
	.dc.b	$6F,$E0,$F8,$F8,$99,$1F,$D5,$E7,$55,$F9,$DA,$CB,$F1,$70,$7D,$37
	.dc.b	$F7,$66,$DF,$CC,$F3,$71,$32,$3F,$B1,$4E,$AB,$F5,$15,$97,$F9,$BF
	.dc.b	$D3,$BE,$8F,$F9,$A3,$6F,$FB,$F7,$13,$ED,$5F,$67,$AB,$FF,$07,$4E
	.dc.b	$63,$FD,$E8,$DA,$FF,$38,$37,$0F,$E0,$CD,$7F,$88,$B5,$FF,$0D,$DF
	.dc.b	$69,$E2,$D8,$E2,$D7,$F8,$CF,$94,$FC,$29,$F2,$7F,$DE,$1B,$7F,$B4
	.dc.b	$CD,$F6,$B9,$AF,$EF,$BE,$23,$EF,$7F,$C2,$ED,$BF,$F4,$46,$3F,$CB
	.dc.b	$A7,$27,$F9,$C7,$2F,$F0,$CF,$9B,$FE,$2F,$7F,$B9,$3F,$D7,$7E,$47
	.dc.b	$D1,$B6,$EF,$FA,$ED,$7F,$80,$BE,$27,$84,$1E,$5D,$2E,$EC,$C6,$1F
	.dc.b	$05,$7F,$03,$6B,$FC,$0E,$78,$67,$FE,$9F,$8F,$F4,$DC,$E9,$B1,$99
	.dc.b	$DC,$7F,$11,$9D,$46,$23,$E6,$FC,$AB,$EE,$7F,$C4,$6D,$BF,$E7,$1F
	.dc.b	$6F,$F3,$1F,$F1,$AC,$66,$BF,$B5,$4E,$5F,$E8,$DE,$32,$31,$1F,$D4
	.dc.b	$A3,$1C,$5E,$C6,$DF,$FA,$E5,$B3,$CA,$B8,$BF,$C8,$E3,$2F,$38,$CA
	.dc.b	$8F,$F5,$87,$E1,$5F,$E0,$2A,$3F,$DA,$3E,$E1,$F2,$F9,$8C,$3F,$DA
	.dc.b	$A3,$33,$FD,$A9,$B7,$FD,$86,$11,$88,$FE,$D4,$D7,$FB,$73,$5F,$F6
	.dc.b	$13,$63,$8C,$FF,$86,$7F,$21,$FE,$19,$C3,$FC,$D5,$B3,$FE,$4A,$FA
	.dc.b	$CF,$F6,$5B,$FE,$5F,$8C,$E1,$FE,$43,$9C,$BF,$E6,$E6,$BF,$A8,$6B
	.dc.b	$CF,$E2,$7F,$CE,$F1,$88,$E5,$18,$25,$7F,$9D,$BF,$06,$FE,$2F,$87
	.dc.b	$F4,$96,$C7,$BB,$EE,$7F,$D1,$71,$8B,$F7,$C7,$CB,$24,$FC,$B6,$C7
	.dc.b	$CB,$CD,$18,$8F,$96,$FF,$8F,$78,$9F,$2D,$FE,$98,$6D,$FF,$3B,$CF
	.dc.b	$FB,$8B,$FD,$82,$33,$7F,$CF,$27,$FB,$E3,$FD,$B5,$AF,$EA,$DA,$A1
	.dc.b	$18,$6C,$7F,$7B,$8C,$11,$FF,$29,$7C,$33,$F1,$9F,$F5,$6B,$6F,$FA
	.dc.b	$87,$DD,$7F,$9E,$BC,$F4,$E6,$FF,$A8,$CF,$FB,$D1,$FF,$01,$35,$FE
	.dc.b	$00,$D7,$CC,$D5,$EB,$3F,$20,$FE,$2C,$DC,$3F,$B1,$AF,$A2,$CC,$C9
	.dc.b	$CB,$C2,$31,$5F,$DD,$F8,$7F,$C6,$35,$FE,$AF,$6C,$7F,$9D,$FE,$EB
	.dc.b	$E7,$C7,$F8,$34,$C9,$7D,$BF,$F6,$7B,$8A,$FF,$52,$CF,$07,$FC,$2F
	.dc.b	$19,$AF,$F4,$F7,$DA,$FF,$8E,$A7,$FD,$86,$D8,$FF,$43,$7D,$D7,$FA
	.dc.b	$CE,$95,$FE,$B0,$AD,$46,$2B,$9F,$BB,$FE,$BE,$9C,$D7,$FA,$DA,$F9
	.dc.b	$3C,$67,$FD,$87,$1F,$EB,$BA,$FF,$58,$37,$C2,$34,$C7,$DC,$7F,$AD
	.dc.b	$2F,$FD,$B2,$DA,$FF,$4E,$7D,$CF,$67,$FD,$75,$19,$CB,$CD,$82,$31
	.dc.b	$7F,$EC,$A6,$D7,$FB,$06,$F9,$56,$FE,$1B,$FA,$A4,$FF,$91,$F8,$BF
	.dc.b	$A2,$34,$AE,$1F,$C8,$9F,$E7,$B1,$E5,$FF,$6C,$F6,$FF,$A6,$36,$34
	.dc.b	$D7,$FD,$04,$D7,$FD,$1A,$BE,$11,$A7,$26,$CF,$26,$BC,$9A,$FB,$FE
	.dc.b	$A3,$FC,$0F,$AD,$8F,$97,$18,$7F,$DC,$5F,$70,$F9,$5C,$6B,$F2,$6A
	.dc.b	$EE,$69,$AA,$E1,$7F,$E6,$DC,$BE,$8F,$E3,$2F,$B2,$FE,$D9,$8A,$31
	.dc.b	$5F,$E0,$66,$BF,$EA,$38,$D4,$F0,$C7,$FA,$6C,$FE,$0B,$C5,$F1,$9E
	.dc.b	$6B,$E3,$87,$E0,$2B,$75,$FE,$E1,$6C,$66,$B8,$5F,$C1,$EC,$FF,$68
	.dc.b	$B7,$83,$9E,$DF,$E8,$06,$E2,$CD,$F2,$1F,$EA,$48,$F8,$15,$BB,$E4
	.dc.b	$48,$C9,$C6,$78,$3F,$E7,$9F,$9A,$7C,$2F,$8D,$B1,$C3,$24,$62,$3F
	.dc.b	$07,$87,$FA,$6E,$1E,$3F,$FA,$F2,$78,$46,$F5,$FF,$4B,$36,$BF,$D3
	.dc.b	$6D,$9E,$B5,$C5,$EA,$7B,$FF,$02,$7A,$5F,$69,$B1,$FE,$3E,$6B,$F6
	.dc.b	$5A,$FF,$B6,$BC,$CC,$E3,$FE,$A7,$6E,$3F,$EA,$C6,$CF,$7B,$FA,$9F
	.dc.b	$95,$3A,$BF,$C6,$9A,$23,$15,$FA,$F8,$C4,$EA,$73,$3F,$EE,$D9,$CB
	.dc.b	$B8,$CB,$FE,$EE,$6B,$FE,$BE,$6A,$8F,$F1,$17,$9B,$F8,$7C,$3F,$C4
	.dc.b	$33,$F5,$5A,$FF,$AC,$EB,$C3,$97,$F1,$2F,$17,$FA,$D4,$FF,$99,$1A
	.dc.b	$FE,$7D,$AF,$F9,$D1,$AF,$FB,$5F,$89,$18,$8F,$C9,$36,$33,$E3,$7E
	.dc.b	$BA,$35,$97,$A7,$D1,$FC,$F3,$FE,$53,$BF,$C6,$35,$FB,$F7,$C9,$FC
	.dc.b	$83,$FE,$A8,$7E,$73,$5F,$98,$D7,$FD,$D8,$7C,$77,$0F,$FB,$A1,$AE
	.dc.b	$7D,$EF,$B7,$18,$9F,$D6,$47,$DE,$7F,$25,$8F,$FB,$B5,$BB,$FF,$BC
	.dc.b	$9B,$7C,$CF,$D6,$DB,$E9,$FF,$D0,$51,$FE,$92,$6B,$FC,$35,$AF,$FA
	.dc.b	$89,$AF,$CF,$6B,$FB,$F6,$C7,$5F,$53,$F8,$DC,$6A,$9F,$8F,$3E,$73
	.dc.b	$F8,$63,$70,$FD,$D3,$5F,$9F,$5E,$4C,$BF,$CA,$DF,$65,$F9,$EF,$50
	.dc.b	$D2,$57,$E7,$E3,$2D,$8C,$BF,$9F,$8C,$BF,$A1,$F6,$3F,$D6,$15,$FD
	.dc.b	$E7,$2F,$F5,$46,$BF,$C8,$67,$2A,$BF,$41,$7C,$AF,$FC,$01,$D2,$FB
	.dc.b	$57,$D1,$FF,$A9,$E3,$0D,$8C,$DF,$F6,$39,$CB,$63,$2F,$E2,$0F,$E3
	.dc.b	$AE,$1F,$C5,$C7,$07,$4D,$7F,$DB,$3D,$DF,$99,$BF,$EF,$ED,$C7,$F5
	.dc.b	$0D,$9E,$B8,$7F,$3F,$E5,$7C,$3D,$24,$FF,$36,$C5,$F9,$57,$9B,$2F
	.dc.b	$83,$F9,$F7,$F8,$0A,$4D,$25,$7F,$16,$FC,$9B,$C1,$C0,$FA,$67,$C8
	.dc.b	$7F,$4F,$6D,$7F,$3D,$AF,$32,$FB,$7F,$2B,$FE,$D3,$6B,$FE,$0D,$6B
	.dc.b	$F2,$5A,$FE,$C9,$AF,$DB,$6B,$D7,$4F,$F0,$DC,$6B,$07,$C7,$FF,$7D
	.dc.b	$AF,$EA,$0D,$7F,$A1,$C6,$BC,$CF,$F0,$44,$73,$3F,$AB,$7D,$C3,$EF
	.dc.b	$7F,$B6,$36,$33,$18,$EF,$CB,$23,$27,$28,$FF,$08,$FE,$2E,$F9,$BF
	.dc.b	$EE,$C3,$E4,$69,$A6,$57,$F6,$77,$4D,$8C,$BF,$D7,$0F,$F6,$BF,$B9
	.dc.b	$FE,$3C,$F3,$FF,$CF,$51,$FF,$00,$5A,$FF,$8E,$DA,$FE,$5D,$AF,$F2
	.dc.b	$06,$BF,$A9,$6B,$D8,$FF,$22,$7B,$CF,$B7,$7D,$97,$F8,$BB,$B3,$FB
	.dc.b	$97,$8D,$FE,$27,$F4,$3F,$C6,$97,$BA,$FF,$19,$FD,$B7,$FA,$E3,$1F
	.dc.b	$E3,$ED,$7F,$9C,$B7,$B3,$FC,$C3,$F8,$17,$F8,$63,$14,$E7,$C1,$FE
	.dc.b	$98,$DB,$FF,$0E,$36,$55,$FE,$20,$E6,$7D,$77,$F8,$AB,$0F,$E0,$D1
	.dc.b	$8F,$B9,$FF,$15,$49,$BC,$13,$2F,$F0,$D4,$E5,$B3,$97,$EA,$4B,$E3
	.dc.b	$7E,$A1,$23,$EA,$35,$FA,$8D,$7E,$A3,$FE,$06,$EA,$F3,$47,$F7,$96
	.dc.b	$E1,$F9,$EA,$FC,$C9,$DE,$7F,$11,$F1,$FF,$8E,$FB,$39,$18,$21,$FA
	.dc.b	$2E,$74,$6A,$31,$3F,$E4,$BF,$9C,$FF,$16,$60,$69,$F7,$BF,$18,$FF
	.dc.b	$1B,$D7,$F8,$F6,$33,$CB,$C4,$FF,$38,$CE,$A7,$E2,$FD,$5F,$E5,$C8
	.dc.b	$FE,$5D,$AF,$E5,$DA,$FE,$5D,$FF,$14,$FD,$2F,$F8,$BB,$03,$4F,$F1
	.dc.b	$8C,$B4,$57,$E4,$3E,$4E,$4F,$E4,$1F,$F4,$2F,$E3,$EF,$96,$CC,$FE
	.dc.b	$4A,$6F,$47,$FA,$CA,$B5,$3F,$E5,$BF,$AC,$F6,$4F,$F5,$36,$F3,$7E
	.dc.b	$31,$B3,$FE,$31,$5F,$80,$7F,$9D,$DC,$EA,$69,$23,$5F,$3D,$FE,$48
	.dc.b	$83,$49,$5F,$E4,$98,$FE,$0E,$FF,$1E,$6E,$1E,$5F,$07,$9C,$FA,$5F
	.dc.b	$F5,$2F,$D6,$3E,$7A,$9E,$58,$93,$FE,$14,$9D,$36,$72,$F7,$E5,$F3
	.dc.b	$11,$F2,$3E,$9E,$6F,$80,$48,$F9,$51,$FE,$B6,$F8,$43,$FE,$AE,$6C
	.dc.b	$7F,$B4,$FB,$BF,$D2,$F3,$93,$E2,$79,$DF,$2C,$93,$FE,$2A,$AE,$2F
	.dc.b	$C3,$8C,$B4,$D4,$E3,$3F,$B5,$37,$83,$FC,$B1,$7C,$A7,$EF,$74,$7D
	.dc.b	$32,$61,$F4,$5F,$AD,$F5,$DF,$74,$91,$F8,$78,$F7,$7D,$27,$F9,$3B
	.dc.b	$0F,$F2,$7E,$7F,$EB,$28,$C5,$78,$1A,$FF,$AB,$DB,$1B,$8F,$F0,$F4
	.dc.b	$F2,$C5,$5A,$6C,$E8,$E8,$53,$A1,$FE,$27,$F8,$E7,$D3,$FF,$24,$FB
	.dc.b	$17,$DF,$FF,$99,$27,$35,$FE,$8B,$9C,$BC,$6B,$AA,$9D,$E5,$FE,$E7
	.dc.b	$F9,$AE,$55,$FE,$83,$F8,$E4,$62,$FF,$DA,$91,$87,$A1,$83,$55,$CD
	.dc.b	$CB,$5F,$05,$75,$AD,$D7,$F8,$D6,$FC,$6F,$5A,$D9,$E0,$9D,$9C,$1A
	.dc.b	$76,$8E,$13,$CA,$75,$3E,$2F,$B2,$7D,$17,$99,$B1,$98,$C7,$D7,$E5
	.dc.b	$91,$93,$94,$61,$F2,$E4,$AE,$71,$86,$E1,$F3,$95,$7C,$B9,$FE,$BF
	.dc.b	$B3,$EB,$FE,$5C,$61,$BE,$57,$E2,$1F,$9F,$EE,$3E,$D7,$FC,$BB,$97
	.dc.b	$E0,$2F,$81,$F2,$0A,$E6,$F9,$47,$CC,$AE,$5F,$01,$E3,$9C,$CF,$F9
	.dc.b	$9D,$AF,$FA,$C1,$AF,$F9,$E5,$AF,$F7,$8F,$A5,$FB,$0E,$DA,$ED,$AE
	.dc.b	$F7,$FF,$3A,$39,$7D,$F1,$F8,$DC,$3F,$37,$87,$F9,$77,$E4,$9C,$DA
	.dc.b	$B9,$87,$BF,$16,$3C,$78,$FF,$D1,$D8,$F1,$C3,$79,$7E,$83,$B5,$DD
	.dc.b	$D3,$2F,$B7,$18,$AF,$CC,$AF,$C1,$75,$7C,$6E,$E3,$FE,$E9,$AE,$18
	.dc.b	$72,$9C,$AE,$4F,$E7,$58,$61,$FE,$79,$36,$B8,$E1,$D7,$1F,$0D,$E6
	.dc.b	$B8,$7E,$3E,$DE,$98,$E3,$B7,$FB,$09,$D3,$67,$4D,$26,$7F,$9A,$9D
	.dc.b	$B6,$B9,$B9,$AB,$8F,$F0,$8A,$FE,$37,$7C,$A3,$F8,$5F,$E3,$CF,$C7
	.dc.b	$FE,$0F,$8E,$EB,$7C,$AF,$C1,$AD,$61,$FC,$2A,$B5,$39,$BF,$A9,$3F
	.dc.b	$CC,$AB,$79,$FD,$73,$E6,$38,$8D,$4F,$09,$FF,$67,$70,$2B,$51,$86
	.dc.b	$BC,$A7,$83,$6B,$F9,$96,$1F,$E0,$1F,$9F,$FC,$37,$85,$FE,$0B,$6C
	.dc.b	$FF,$82,$FD,$FB,$7F,$E0,$B9,$B7,$FE,$09,$9F,$44,$EE,$BF,$C0,$91
	.dc.b	$88,$FF,$09,$36,$7F,$C2,$1E,$FF,$2F,$24,$F6,$D3,$D9,$E1,$66,$26
	.dc.b	$3F,$E2,$17,$5A,$FF,$46,$8C,$13,$2F,$F0,$94,$EA,$B8,$C3,$4F,$D2
	.dc.b	$61,$DA,$FF,$BF,$61,$CE,$BF,$DD,$93,$96,$C6,$5F,$F7,$F1,$FE,$2D
	.dc.b	$FA,$2F,$51,$EB,$E7,$B7,$FE,$37,$7E,$06,$7F,$E1,$F8,$FF,$14,$46
	.dc.b	$0C,$FD,$87,$C1,$F8,$E7,$DD,$FC,$58,$C3,$69,$8F,$AD,$FC,$1D,$70
	.dc.b	$7F,$C8,$D5,$C1,$A4,$9F,$F1,$B7,$E1,$9C,$DB,$1F,$E3,$B8,$F3,$91
	.dc.b	$FE,$37,$C0,$FC,$3B,$F9,$E4,$78,$B7,$FE,$1D,$39,$F1,$3F,$0B,$5E
	.dc.b	$3C,$3F,$DC,$2F,$06,$C6,$8D,$E7,$70,$E8,$7F,$A9,$BB,$BF,$CD,$6E
	.dc.b	$29,$87,$F8,$23,$D5,$FC,$47,$E3,$BF,$BE,$CF,$C7,$87,$0B,$FF,$78
	.dc.b	$B6,$33,$5C,$3B,$7F,$37,$39,$C7,$FD,$6B,$EC,$3E,$8B,$F3,$ED,$8F
	.dc.b	$17,$CD,$BE,$7D,$C5,$AB,$FD,$3C,$E6,$3F,$44,$DA,$FE,$8A,$DD,$3F
	.dc.b	$16,$1C,$3F,$8E,$6B,$F8,$06,$BB,$BF,$0C,$F4,$6C,$74,$DB,$F8,$B7
	.dc.b	$93,$FB,$8F,$00,$FE,$BF,$DC,$DF,$07,$DA,$9C,$E3,$F3,$EF,$8B,$63
	.dc.b	$38,$7E,$44,$FE,$85,$7C,$A7,$FC,$75,$87,$F0,$6F,$B3,$73,$DF,$85
	.dc.b	$C3,$6B,$B4,$E9,$B3,$A6,$9E,$3F,$81,$7D,$BF,$F1,$26,$C7,$D4,$D9
	.dc.b	$DB,$9A,$FA,$71,$89,$F8,$D3,$F5,$23,$1A,$7D,$2E,$0F,$B9,$39,$25
	.dc.b	$70,$24,$FC,$DF,$B6,$E3,$C3,$FA,$27,$F7,$D1,$F1,$78,$51,$AA,$FF
	.dc.b	$80,$2E,$AB,$85,$FF,$2C,$8C,$4E,$A3,$1D,$BF,$67,$E9,$64,$FE,$8F
	.dc.b	$C1,$FE,$F5,$E7,$63,$3F,$07,$2A,$F1,$9F,$F2,$0F,$37,$FC,$17,$8F
	.dc.b	$DB,$9F,$E7,$BF,$12,$FB,$7F,$F3,$C5,$FF,$26,$FB,$37,$E0,$9A,$9C
	.dc.b	$E1,$FD,$4A,$72,$FE,$D7,$D4,$7D,$C7,$F9,$F2,$3F,$C4,$71,$8F,$03
	.dc.b	$FA,$23,$63,$33,$BD,$BF,$47,$39,$6C,$65,$FE,$63,$19,$6C,$65,$FE
	.dc.b	$CB,$F2,$3F,$29,$FE,$0F,$D6,$FB,$7F,$E9,$D3,$9E,$74,$E7,$1F,$99
	.dc.b	$39,$8C,$5F,$E7,$F9,$13,$9C,$3F,$B1,$FD,$C3,$E4,$E9,$8F,$B1,$B9
	.dc.b	$7E,$AE,$74,$D9,$D3,$4F,$CF,$7D,$27,$E7,$4F,$CC,$FC,$25,$8F,$CF
	.dc.b	$3F,$99,$23,$FC,$05,$B9,$FE,$6F,$B9,$18,$8F,$E1,$91,$89,$FE,$7B
	.dc.b	$F3,$CF,$07,$FB,$EF,$E3,$5F,$E8,$C9,$FE,$05,$39,$7F,$82,$46,$6F
	.dc.b	$F8,$17,$BA,$F9,$E7,$D5,$7F,$80,$63,$0D,$8C,$DF,$F9,$9A,$72,$FF
	.dc.b	$4E,$FC,$9F,$FD,$6A,$FF,$11,$6D,$F6,$FC,$A1,$F2,$BF,$E1,$29,$CF
	.dc.b	$77,$E5,$FE,$B3,$FC,$15,$DA,$7F,$C3,$39,$3E,$3B,$FC,$DF,$9F,$EA
	.dc.b	$6B,$4F,$F4,$D9,$D3,$E9,$9D,$E7,$FC,$DA,$73,$3E,$99,$CC,$FF,$81
	.dc.b	$27,$2E,$E3,$2F,$F9,$E1,$AA,$36,$FF,$83,$3A,$DF,$4C,$F7,$BF,$C5
	.dc.b	$ED,$C3,$F9,$A7,$D4,$3E,$DF,$35,$FE,$34,$8C,$4F,$F7,$86,$E1,$FD
	.dc.b	$0D,$AF,$76,$BF,$E0,$CC,$11,$88,$FE,$DE,$D8,$CD,$FF,$8E,$E3,$18
	.dc.b	$7F,$8E,$AF,$FA,$43,$6F,$F5,$1F,$8C,$FF,$10,$7F,$92,$46,$76,$FF
	.dc.b	$19,$CE,$67,$FC,$5B,$DE,$8C,$47,$F9,$81,$AF,$FA,$45,$B1,$BF,$2B
	.dc.b	$FC,$A0,$DC,$3F,$D6,$6D,$AF,$F1,$ED,$72,$CD,$F7,$4F,$47,$FC,$BA
	.dc.b	$DC,$3F,$B1,$61,$FD,$3B,$0F,$E7,$93,$97,$F9,$F4,$65,$FB,$4D,$7E
	.dc.b	$63,$5F,$F0,$CB,$6F,$FD,$55,$E8,$FF,$99,$63,$04,$7F,$CD,$15,$FC
	.dc.b	$4B,$3F,$F1,$7B,$70,$FF,$41,$FD,$73,$E7,$BD,$3E,$27,$F9,$AE,$31
	.dc.b	$3F,$09,$B9,$FF,$97,$9A,$FD,$06,$BD,$2B,$E5,$C6,$23,$FD,$80,$DC
	.dc.b	$8E,$73,$7C,$47,$F4,$6A,$FF,$1B,$B6,$3F,$DB,$FF,$15,$FE,$8A,$8C
	.dc.b	$63,$FE,$B0,$AF,$F0,$C3,$5F,$D2,$35,$FC,$B3,$5F,$F2,$1B,$6B,$AB
	.dc.b	$FE,$58,$30,$66,$7F,$83,$BF,$9F,$6C,$FF,$B6,$FE,$C1,$F3,$7E,$5F
	.dc.b	$33,$FD,$29,$18,$BF,$C5,$B6,$BF,$D0,$ED,$7F,$32,$D7,$F5,$9F,$88
	.dc.b	$7F,$92,$5F,$F0,$23,$78,$3F,$DE,$DD,$2F,$C5,$5E,$27,$FA,$BA,$B5
	.dc.b	$8F,$2F,$7B,$07,$F1,$CA,$DC,$7F,$8D,$E7,$78,$FF,$99,$B1,$FD,$74
	.dc.b	$FE,$F5,$B1,$FE,$F9,$F9,$67,$C3,$FF,$AF,$27,$3C,$1F,$EA,$7F,$8B
	.dc.b	$C1,$FE,$B5,$AD,$E1,$FE,$45,$AE,$7C,$8F,$B3,$79,$BF,$EB,$B6,$D7
	.dc.b	$F8,$5F,$E5,$DE,$D6,$6A,$4F,$F8,$6E,$B7,$AF,$F6,$FB,$DC,$E6,$BF
	.dc.b	$D8,$D1,$89,$FF,$5B,$B5,$FE,$12,$D7,$FD,$CE,$D5,$87,$9F,$2F,$31
	.dc.b	$FE,$D1,$FB,$77,$D1,$E3,$1F,$24,$DB,$FF,$01,$56,$A7,$27,$FA,$56
	.dc.b	$3F,$D8,$BD,$1F,$E7,$C7,$FC,$FA,$D7,$25,$3F,$4E,$6F,$ED,$9F,$35
	.dc.b	$FE,$8D,$8C,$35,$FF,$43,$C6,$BD,$2E,$8D,$8E,$5C,$DF,$ED,$F6,$F0
	.dc.b	$EA,$31,$1C,$9A,$FF,$B3,$35,$FB,$5F,$55,$FE,$DE,$AF,$F7,$3F,$0A
	.dc.b	$31,$A6,$E3,$17,$FE,$53,$FB,$6F,$F2,$83,$F6,$EB,$6F,$D0,$9D,$BF
	.dc.b	$E2,$B6,$DF,$FB,$D2,$B8,$CF,$E0,$7E,$A7,$E4,$62,$E2,$46,$1B,$3F
	.dc.b	$CF,$9A,$FF,$BE,$B0,$7D,$37,$8A,$FF,$2C,$D8,$CC,$FE,$56,$F1,$32
	.dc.b	$3F,$28,$F1,$BE,$B8,$FF,$B8,$36,$BA,$7F,$69,$6D,$FF,$B5,$9B,$93
	.dc.b	$89,$F2,$DF,$13,$A3,$FD,$95,$A5,$BE,$31,$8C,$3F,$CF,$D7,$C1,$FF
	.dc.b	$5A,$76,$BE,$7F,$93,$63,$31,$F4,$21,$F1,$FF,$EA,$49,$C9,$0F,$F4
	.dc.b	$B4,$71,$7F,$C7,$71,$B7,$FC,$C1,$E0,$3E,$67,$E9,$B6,$3F,$91,$C6
	.dc.b	$0F,$CF,$63,$FE,$DB,$7D,$DF,$17,$B9,$FF,$52,$37,$81,$C1,$F9,$1C
	.dc.b	$30,$FF,$84,$FE,$DB,$FC,$97,$7F,$AB,$C5,$18,$8F,$F6,$F3,$5F,$F1
	.dc.b	$53,$5E,$5E,$D7,$92,$31,$09,$E7,$39,$3D,$38,$FF,$92,$67,$FD,$C4
	.dc.b	$D7,$D3,$EB,$7D,$9B,$F1,$65,$EA,$9F,$EC,$BE,$2F,$E6,$3C,$9F,$E7
	.dc.b	$71,$FD,$C6,$30,$47,$04,$3F,$BA,$3F,$2B,$27,$87,$F8,$76,$CF,$FB
	.dc.b	$D5,$AF,$FA,$56,$BB,$BF,$E1,$4F,$B6,$FA,$7C,$E6,$92,$3E,$64,$65
	.dc.b	$B1,$94,$7F,$87,$1F,$27,$C6,$78,$C9,$1F,$E2,$68,$FD,$77,$DA,$FF
	.dc.b	$42,$8F,$F1,$8B,$5E,$67,$F0,$F7,$0F,$F9,$01,$AE,$7C,$5F,$E0,$31
	.dc.b	$88,$FE,$72,$DB,$39,$C7,$FC,$B3,$F6,$DF,$AD,$F5,$8D,$3C,$B8,$7E
	.dc.b	$C7,$ED,$9F,$23,$F9,$86,$E5,$F8,$F6,$BE,$83,$F1,$0E,$1F,$F3,$C3
	.dc.b	$5F,$F3,$53,$5D,$C7,$53,$F4,$BE,$C7,$E1,$E3,$FC,$C7,$CD,$FC,$0E
	.dc.b	$31,$C5,$F8,$BC,$D8,$78,$A3,$FC,$0F,$F6,$CF,$93,$FE,$99,$D8,$8C
	.dc.b	$65,$FE,$AD,$6B,$FE,$99,$6B,$FE,$97,$6B,$F7,$A3,$94,$F4,$9C,$CF
	.dc.b	$FA,$5D,$BC,$1F,$CC,$B9,$3F,$96,$78,$1F,$92,$C7,$F4,$53,$9B,$FF
	.dc.b	$56,$5F,$8A,$7D,$9C,$CF,$B3,$FE,$44,$DA,$FF,$67,$C7,$F1,$9F,$B6
	.dc.b	$E1,$C1,$E5,$F8,$4F,$ED,$1B,$7F,$B3,$EF,$E0,$C6,$0F,$ED,$9C,$0F
	.dc.b	$B5,$F2,$5F,$FA,$77,$ED,$BF,$DC,$77,$F9,$CD,$3E,$74,$62,$11,$F4
	.dc.b	$6B,$8C,$7E,$8B,$ED,$BC,$CF,$F7,$D6,$BF,$85,$6B,$F9,$E6,$BC,$74
	.dc.b	$FF,$00,$46,$20,$D3,$78,$FF,$BC,$FC,$CF,$E8,$DF,$13,$FD,$9B,$3F
	.dc.b	$F5,$B3,$6F,$A1,$EB,$8F,$F5,$D7,$DB,$7F,$B2,$63,$07,$F1,$F7,$0F
	.dc.b	$BA,$37,$F0,$9F,$D4,$E7,$DF,$81,$DE,$9F,$93,$6D,$70,$BE,$B5,$BE
	.dc.b	$5F,$ED,$DF,$1F,$FE,$F6,$9C,$FC,$4B,$EB,$BF,$B0,$36,$3C,$BC,$EA
	.dc.b	$E1,$1E,$53,$FC,$6E,$E3,$3F,$D2,$57,$19,$D6,$28,$C4,$FE,$4A,$B8
	.dc.b	$65,$FC,$43,$D0,$7C,$77,$F8,$C5,$B8,$FF,$8D,$5A,$FF,$8E,$5A,$FF
	.dc.b	$01,$6B,$CE,$39,$11,$FF,$20,$7C,$B3,$ED,$BF,$D9,$91,$EE,$EC,$FE
	.dc.b	$DD,$97,$E9,$9B,$8F,$F2,$EB,$E1,$F0,$FF,$E1,$7F,$B6,$7D,$DD,$26
	.dc.b	$3F,$23,$ED,$FF,$C1,$B3,$A8,$F9,$6D,$7E,$57,$0A,$31,$9F,$ED,$AB
	.dc.b	$FC,$1F,$F5,$BF,$E2,$2B,$F9,$6F,$FB,$2D,$FB,$AD,$73,$BF,$F8,$66
	.dc.b	$38,$57,$F1,$77,$4D,$F8,$8D,$4F,$F9,$0A,$BF,$C6,$7F,$03,$F4,$48
	.dc.b	$FD,$16,$BF,$45,$AF,$D1,$7F,$C1,$BD,$3E,$0D,$FF,$47,$5F,$A6,$3B
	.dc.b	$1F,$E7,$AA,$FC,$8F,$D9,$7E,$43,$E4,$24,$60,$87,$E2,$B9,$91,$A8
	.dc.b	$C4,$FF,$8D,$FE,$DF,$EB,$C7,$F8,$25,$BD,$07,$3C,$2F,$A2,$FF,$18
	.dc.b	$D7,$E9,$A7,$64,$F6,$BF,$CA,$93,$A9,$FD,$01,$FE,$58,$FA,$5F,$CB
	.dc.b	$11,$FC,$B3,$5F,$CB,$35,$FC,$B3,$FE,$32,$FA,$5F,$F1,$5C,$7F,$87
	.dc.b	$DB,$89,$CC,$77,$3F,$13,$F1,$BF,$91,$23,$FB,$6F,$B6,$7D,$BE,$23
	.dc.b	$F1,$BD,$1F,$EE,$87,$89,$B8,$FF,$2F,$7D,$67,$B7,$9F,$FA,$E3,$79
	.dc.b	$3F,$C1,$B9,$A3,$17,$FE,$35,$5F,$6F,$CB,$E0,$64,$FF,$10,$9D,$D7
	.dc.b	$1A,$EF,$F6,$FF,$E5,$29,$34,$95,$FE,$4B,$8F,$E0,$F8,$78,$9B,$1E
	.dc.b	$1F,$13,$9B,$DD,$FF,$45,$7D,$9B,$EA,$69,$EE,$89,$3F,$C7,$27,$4D
	.dc.b	$9C,$BF,$C6,$29,$CF,$E7,$24,$79,$5F,$57,$47,$CA,$24,$7C,$18,$FF
	.dc.b	$4E,$7C,$2F,$28,$FF,$68,$74,$FF,$A3,$E7,$FD,$53,$39,$9F,$F0,$D4
	.dc.b	$E5,$F8,$5F,$41,$F5,$88,$FF,$A4,$2A,$BE,$78,$BC,$3F,$DF,$AF,$A3
	.dc.b	$F7,$23,$3A,$FC,$B7,$CB,$7F,$82,$E9,$FA,$84,$C3,$E9,$3F,$5F,$ED
	.dc.b	$FF,$0E,$48,$FB,$D1,$EE,$F9,$7F,$F2,$4E,$1F,$E4,$AC,$FF,$D5,$91
	.dc.b	$8A,$FF,$6D,$66,$8C,$47,$FC,$00,$8C,$3F,$E5,$CA,$E5,$18,$8F,$F1
	.dc.b	$A5,$74,$74,$74,$AE,$71,$98,$DC,$F1,$F9,$17,$D3,$7E,$73,$EA,$9F
	.dc.b	$65,$FE,$60,$9C,$DF,$FA,$46,$72,$FF,$9B,$A3,$38,$75,$6F,$C7,$BE
	.dc.b	$6B,$96,$3F,$EF,$78,$C5,$F1,$8C,$4E,$A1,$7F,$EE,$97,$0F,$FB,$AE
	.dc.b	$2D,$72,$95,$F8,$27,$31,$A6,$3D,$CE,$B3,$C6,$7C,$F1,$A3,$9D,$F9
	.dc.b	$1E,$17,$E1,$6D,$7F,$9B,$6B,$E0,$FD,$BB,$E5,$BC,$ED,$AE,$2D,$7A
	.dc.b	$9E,$E5,$F6,$FC,$26,$4F,$07,$9C,$62,$38,$E0,$AE,$97,$E2,$BD,$D7
	.dc.b	$BC,$79,$7F,$8D,$FB,$2C,$36,$3C,$66,$B8,$13,$F1,$FD,$E7,$DB,$FF
	.dc.b	$9D,$31,$FC,$1D,$EE,$B9,$DF,$08,$E7,$87,$5A,$FC,$4E,$1C,$BE,$0F
	.dc.b	$EF,$57,$13,$FD,$2C,$53,$FD,$31,$5E,$56,$DF,$57,$3F,$47,$F8,$17
	.dc.b	$FD,$6B,$28,$C4,$7F,$B0,$A3,$1C,$3D,$A3,$E1,$CE,$9B,$1F,$12,$3F
	.dc.b	$2B,$8F,$A3,$E0,$1E,$27,$FB,$9E,$3E,$0E,$1D,$30,$F7,$DF,$0E,$3F
	.dc.b	$1B,$61,$A3,$1E,$7C,$FD,$F3,$FC,$6F,$7B,$C0,$ED,$97,$FB,$6B,$93
	.dc.b	$F9,$C4,$63,$2F,$D2,$D3,$E7,$31,$5A,$C7,$9D,$E2,$BA,$47,$6A,$C4
	.dc.b	$C8,$D9,$92,$DF,$E3,$BE,$DF,$CE,$DE,$0E,$F1,$FC,$5E,$7F,$4A,$EB
	.dc.b	$83,$F8,$1B,$73,$F5,$36,$FF,$7A,$F1,$7A,$57,$2B,$EA,$76,$E1,$B1
	.dc.b	$92,$4F,$96,$32,$FE,$93,$ED,$9F,$6F,$F1,$FC,$87,$63,$E6,$3E,$3D
	.dc.b	$6B,$1F,$E0,$95,$A9,$CD,$A7,$F9,$5E,$1F,$C5,$2F,$83,$FC,$C1,$B1
	.dc.b	$FC,$F5,$B1,$F6,$5A,$F1,$9E,$8F,$16,$CF,$16,$C7,$F7,$DF,$9A,$DF
	.dc.b	$8B,$FE,$12,$6D,$7F,$80,$7E,$0A,$D7,$F8,$46,$6E,$5F,$E1,$59,$FF
	.dc.b	$7C,$4E,$EB,$FC,$15,$18,$8F,$F0,$C3,$67,$FC,$39,$F0,$3C,$BC,$93
	.dc.b	$DB,$99,$A7,$F4,$6C,$7F,$C3,$93,$B2,$3F,$EA,$68,$D6,$3F,$E0,$99
	.dc.b	$CB,$67,$5E,$1E,$34,$FF,$79,$DF,$0A,$F9,$93,$E7,$C3,$C2,$D9,$D4
	.dc.b	$FF,$88,$FE,$4F,$D0,$7A,$7A,$2E,$3F,$E3,$17,$DD,$9F,$F8,$8E,$3F
	.dc.b	$C5,$91,$83,$3F,$69,$EF,$F8,$F7,$DB,$7C,$38,$C3,$6B,$FD,$87,$3A
	.dc.b	$7F,$C7,$D3,$A6,$93,$1F,$F1,$57,$D1,$71,$6C,$7F,$8F,$63,$CA,$47
	.dc.b	$F9,$2F,$03,$ED,$FF,$9A,$C7,$FB,$69,$B5,$FE,$CC,$7C,$4D,$F1,$3F
	.dc.b	$07,$1A,$E1,$FC,$8D,$F3,$9C,$E7,$A7,$F1,$3D,$6F,$62,$9F,$81,$FB
	.dc.b	$7F,$E0,$78,$72,$9D,$67,$FE,$EC,$6E,$7F,$ED,$E6,$CF,$E8,$39,$17
	.dc.b	$CA,$FF,$CD,$75,$A7,$FC,$DB,$EC,$BE,$8B,$F3,$6D,$86,$3F,$EE,$FF
	.dc.b	$9D,$7D,$03,$8B,$78,$7E,$92,$73,$1F,$9F,$6D,$FF,$50,$6D,$F4,$3A
	.dc.b	$B8,$70,$D7,$37,$DE,$79,$36,$39,$09,$E0,$FF,$8A,$7C,$AF,$DA,$F7
	.dc.b	$9F,$D6,$7B,$BC,$77,$FA,$B3,$9C,$7D,$F1,$8A,$E2,$70,$CA,$CE,$8F
	.dc.b	$F4,$AF,$D9,$39,$FF,$81,$AF,$87,$B1,$86,$CF,$1C,$DC,$CB,$E1,$39
	.dc.b	$8C,$5F,$18,$C7,$C1,$BE,$AB,$F8,$8B,$63,$A6,$FF,$5B,$B7,$35,$F5
	.dc.b	$78,$51,$F5,$5B,$5F,$32,$30,$47,$04,$CF,$E7,$DF,$58,$FD,$3F,$DB
	.dc.b	$F2,$97,$07,$E8,$23,$F8,$04,$62,$7F,$D7,$71,$B9,$D7,$6F,$F2,$A8
	.dc.b	$C4,$EA,$31,$86,$5F,$B7,$F6,$F2,$7E,$57,$4F,$F7,$5F,$43,$19,$77
	.dc.b	$3F,$CD,$55,$8C,$BF,$49,$E6,$FF,$86,$70,$F9,$DC,$CA,$FE,$7B,$F1
	.dc.b	$6F,$B6,$FF,$3E,$E1,$FA,$BF,$B0,$7D,$BE,$A7,$38,$FF,$82,$67,$2F
	.dc.b	$DD,$F4,$9F,$6F,$FE,$80,$8F,$F1,$8C,$63,$C0,$FE,$88,$D8,$CC,$EF
	.dc.b	$6F,$CE,$CE,$5B,$19,$7F,$55,$19,$6C,$65,$FF,$0E,$7C,$9F,$D2,$CF
	.dc.b	$F8,$9F,$5B,$EB,$FF,$A7,$4E,$71,$FC,$DC,$63,$93,$F3,$6D,$8F,$CA
	.dc.b	$57,$17,$F2,$DF,$6C,$F9,$5F,$EC,$6D,$8C,$FD,$BE,$26,$47,$07,$71
	.dc.b	$8C,$FE,$C7,$7B,$B9,$F4,$5F,$9E,$3F,$35,$F0,$D7,$1F,$CF,$F7,$A3
	.dc.b	$F3,$84,$8F,$E0,$ED,$C3,$F3,$33,$B9,$C9,$23,$F8,$5E,$0F,$9F,$F1
	.dc.b	$4F,$EC,$FE,$DF,$FD,$20,$FE,$D6,$33,$5E,$89,$CB,$F9,$1F,$79,$F3
	.dc.b	$EF,$AB,$FF,$01,$46,$1B,$19,$C3,$FC,$EB,$39,$7F,$A9,$FD,$B7,$FA
	.dc.b	$F5,$FE,$08,$DB,$ED,$F6,$EF,$96,$FF,$09,$CE,$7A,$3F,$31,$F5,$BF
	.dc.b	$E0,$BE,$D3,$FE,$1A,$C9,$F1,$FF,$E3,$1C,$FF,$57,$5A,$7F,$8B,$4E
	.dc.b	$9F,$4C,$EF,$3F,$E6,$33,$99,$F4,$CE,$67,$F9,$F4,$E5,$DC,$65,$FE
	.dc.b	$9B,$19,$77,$19,$7F,$92,$FD,$3B,$DD,$FF,$17,$B6,$FF,$93,$7D,$4B
	.dc.b	$ED,$F3,$5F,$E3,$48,$C4,$FF,$7C,$6D,$FD,$E6,$BF,$D4,$F9,$11,$88
	.dc.b	$FE,$C3,$18,$BF,$F1,$C4,$E6,$7F,$9C,$B6,$75,$5F,$E0,$AC,$3F,$C7
	.dc.b	$6F,$F3,$B6,$D7,$F1,$5F,$29,$F6,$FF,$C4,$E3,$F4,$33,$AD,$FF,$BD
	.dc.b	$4E,$67,$FB,$45,$70,$F2,$FF,$C6,$D3,$99,$FF,$4D,$35,$FF,$7E,$B6
	.dc.b	$78,$61,$FE,$C2,$6C,$FF,$6A,$6B,$F0,$FE,$DF,$FA,$9B,$FB,$46,$E1
	.dc.b	$FE,$1C,$9F,$EA,$13,$9F,$4B,$FC,$C7,$18,$8F,$98,$D7,$E5,$B5,$FE
	.dc.b	$C2,$DC,$CD,$27,$A3,$FE,$67,$75,$87,$F1,$1A,$F2,$78,$7F,$E6,$7F
	.dc.b	$AF,$7D,$07,$F9,$D2,$B5,$9F,$F8,$8E,$73,$A7,$C1,$6E,$1F,$05,$AF
	.dc.b	$CE,$6B,$FE,$C5,$6B,$E6,$6E,$26,$9A,$CD,$F1,$5F,$D1,$AB,$FB,$FB
	.dc.b	$63,$FD,$41,$F1,$1F,$E8,$A8,$C6,$1F,$EB,$7A,$FC,$A3,$5F,$C9,$B5
	.dc.b	$FF,$04,$B6,$B9,$BF,$E5,$A3,$04,$AF,$F4,$85,$7F,$05,$8F,$CE,$B6
	.dc.b	$BE,$8F,$D8,$BE,$7B,$C9,$E6,$FF,$A4,$E3,$17,$F8,$96,$D7,$E2,$5A
	.dc.b	$FE,$59,$AF,$EA,$BE,$DB,$FC,$98,$FF,$3E,$6E,$BF,$EB,$6E,$A7,$DB
	.dc.b	$FF,$31,$C3,$FC,$45,$5A,$79,$7B,$DF,$EB,$08,$C6,$1F,$F0,$02,$B8
	.dc.b	$3D,$67,$FD,$6B,$87,$EA,$EF,$F7,$6D,$9F,$C7,$FC,$BB,$E1,$B7,$F1
	.dc.b	$52,$FF,$C6,$57,$BE,$35,$EE,$73,$5F,$EB,$7F,$B4,$FE,$AF,$1F,$E1
	.dc.b	$46,$E3,$FB,$FF,$98,$7B,$9B,$F3,$A5,$ED,$FF,$60,$4E,$A3,$07,$F8
	.dc.b	$79,$C5,$7E,$DD,$AF,$F0,$46,$BF,$E7,$8C,$91,$88,$FD,$6E,$1E,$AA
	.dc.b	$F3,$65,$FE,$CF,$FB,$77,$D2,$FA,$E2,$4E,$8F,$F1,$7C,$E9,$C5,$FF
	.dc.b	$8B,$CF,$F3,$E7,$37,$F9,$D1,$FF,$3A,$B5,$C9,$4F,$D2,$9B,$FB,$77
	.dc.b	$CC,$FF,$A3,$A3,$0D,$7F,$CF,$F1,$AF,$47,$A3,$63,$97,$1F,$F9,$49
	.dc.b	$BC,$5A,$8C,$47,$26,$BF,$1F,$57,$D3,$7F,$A5,$6B,$FD,$55,$18,$9F
	.dc.b	$B9,$39,$CF,$E7,$6E,$FA,$9C,$57,$CE,$AB,$C2,$C3,$94,$FF,$41,$6E
	.dc.b	$3F,$EB,$7C,$39,$4F,$F9,$73,$DA,$75,$FC,$9C,$11,$89,$FE,$74,$D7
	.dc.b	$F9,$D3,$5F,$E8,$FF,$72,$FA,$4F,$1B,$63,$FC,$D2,$D7,$FA,$A4,$F3
	.dc.b	$C7,$FD,$95,$39,$6D,$72,$3D,$B3,$BD,$7F,$B3,$F0,$3E,$6B,$E1,$F8
	.dc.b	$3F,$EC,$86,$C6,$72,$FC,$E6,$F8,$99,$1F,$EA,$19,$D3,$FA,$99,$E7
	.dc.b	$DC,$FA,$1A,$49,$FF,$0B,$FC,$D7,$FB,$55,$FF,$55,$4E,$DB,$3B,$7F
	.dc.b	$C6,$D3,$B7,$FC,$A3,$F6,$FF,$E3,$77,$F3,$59,$23,$11,$FE,$46,$6B
	.dc.b	$FE,$09,$6B,$FE,$08,$F8,$BF,$F1,$7B,$FE,$BB,$6B,$FE,$5E,$6E,$ED
	.dc.b	$7F,$C1,$F5,$DA,$31,$5F,$D2,$3E,$D7,$F5,$58,$7F,$4C,$6F,$51,$A4
	.dc.b	$9F,$4C,$65,$B1,$97,$FC,$BF,$F7,$BF,$C8,$5F,$F6,$83,$5F,$F3,$9B
	.dc.b	$6F,$C7,$7E,$19,$FF,$6B,$72,$3E,$E3,$CA,$D8,$F9,$2D,$7F,$CD,$AD
	.dc.b	$7F,$C0,$B1,$C0,$8F,$07,$DF,$3F,$ED,$88,$FF,$72,$35,$FF,$43,$B6
	.dc.b	$BF,$C2,$93,$F9,$0F,$97,$FC,$4D,$7F,$8F,$5A,$FF,$8F,$1A,$FF,$BC
	.dc.b	$9B,$46,$93,$8F,$E8,$BA,$FC,$43,$FD,$21,$5F,$EF,$76,$BF,$E9,$A6
	.dc.b	$E1,$F2,$27,$F3,$6F,$CA,$FB,$87,$C9,$7D,$C6,$DF,$E4,$1A,$FF,$A5
	.dc.b	$1A,$FF,$95,$DA,$FF,$95,$9A,$B1,$46,$1B,$3F,$8A,$F5,$7F,$1B,$5F
	.dc.b	$E5,$0E,$1F,$DB,$C6,$30,$F3,$57,$EE,$BE,$91,$F4,$DF,$92,$6E,$1F
	.dc.b	$A6,$6B,$FE,$AA,$6B,$FE,$6D,$6B,$FE,$6C,$6B,$E7,$6E,$26,$FE,$51
	.dc.b	$F8,$9C,$3F,$22,$DC,$BF,$CE,$AD,$BE,$58,$7F,$32,$E8,$FF,$58,$C7
	.dc.b	$F5,$16,$BF,$EA,$EF,$82,$FE,$A5,$18,$8F,$F4,$DC,$60,$8F,$17,$FA
	.dc.b	$C6,$4F,$B7,$79,$9F,$C7,$9B,$43,$FD,$B6,$E1,$FF,$62,$56,$5F,$F6
	.dc.b	$57,$CC,$7F,$26,$F8,$73,$4F,$57,$E8,$9B,$9F,$F7,$2F,$C2,$7E,$58
	.dc.b	$9C,$DF,$C1,$1F,$CC,$FD,$3B,$EA,$BF,$98,$BF,$C8,$6B,$FD,$DA,$DE
	.dc.b	$2F,$DD,$BB,$8F,$44,$7E,$13,$C2,$FF,$01,$C7,$F7,$96,$BF,$DF,$5A
	.dc.b	$F9,$5A,$FB,$5A,$FC,$36,$BF,$DB,$BD,$0F,$F0,$14,$6B,$83,$FE,$00
	.dc.b	$3F,$D2,$3E,$33,$FB,$2F,$2F,$FA,$C9,$B7,$CE,$FF,$A6,$FD,$AF,$F0
	.dc.b	$1A,$FF,$5B,$B6,$7F,$44,$DC,$0F,$F6,$A8,$F7,$CF,$0F,$9A,$E6,$48
	.dc.b	$C9,$CB,$C8,$FE,$B7,$F1,$4F,$AC,$BF,$13,$2B,$79,$7F,$BC,$E7,$66
	.dc.b	$61,$B7,$F2,$AB,$FF,$78,$C7,$07,$A3,$E0,$3F,$C2,$7E,$BF,$C6,$23
	.dc.b	$F4,$DA,$FD,$36,$BF,$4D,$FF,$04,$7D,$83,$E6,$7F,$B9,$FE,$1C,$FB
	.dc.b	$1A,$7F,$85,$A7,$57,$EA,$FB,$3F,$C2,$3E,$8E,$0F,$F7,$1C,$63,$3F
	.dc.b	$E8,$6D,$8F,$F2,$87,$D1,$3D,$0F,$F0,$0C,$E6,$FF,$8A,$37,$73,$9F
	.dc.b	$2D,$7C,$6B,$E5,$FD,$7F,$9D,$BC,$DF,$E0,$C8,$FC,$9E,$7F,$D5,$1D
	.dc.b	$4F,$9C,$FE,$44,$E2,$35,$3C,$4E,$5A,$FF,$8E,$2B,$FC,$83,$F4,$DF
	.dc.b	$94,$23,$FA,$16,$BF,$A1,$6B,$FA,$17,$FB,$AF,$D2,$BE,$BB,$FC,$42
	.dc.b	$D9,$FC,$2B,$5F,$F3,$47,$5A,$B7,$E3,$F0,$FB,$59,$5C,$23,$2F,$28
	.dc.b	$CB,$EF,$E0,$61,$FC,$CB,$4F,$F0,$4E,$EE,$3F,$F5,$E5,$FE,$72,$31
	.dc.b	$F1,$1F,$81,$3F,$CA,$FF,$69,$F1,$F9,$7F,$D1,$57,$C2,$FE,$65,$F0
	.dc.b	$79,$46,$70,$EF,$7C,$67,$F9,$17,$E1,$CE,$AF,$F4,$03,$63,$FD,$08
	.dc.b	$D7,$FD,$10,$D7,$F9,$03,$5F,$E5,$0D,$7A,$C7,$43,$FC,$4B,$F4,$7F
	.dc.b	$E1,$F8,$F8,$8D,$BF,$F0,$CB,$72,$E7,$5F,$E5,$6F,$AD,$FC,$85,$FC
	.dc.b	$B6,$DF,$F8,$81,$B5,$FC,$F7,$64,$7F,$99,$63,$0E,$AB,$FB,$93,$67
	.dc.b	$7D,$3D,$FD,$DF,$D8,$7A,$32,$FE,$5B,$FB,$3F,$93,$C4,$7D,$0D,$FF
	.dc.b	$C0,$79,$FD,$C6,$E7,$F5,$AB,$FC,$DF,$F5,$AF,$1B,$FC,$8E,$DC,$3F
	.dc.b	$CB,$6D,$7F,$A6,$68,$8C,$47,$F5,$58,$C6,$FD,$09,$1F,$C7,$A7,$96
	.dc.b	$5F,$E4,$45,$3F,$EA,$0F,$C6,$1F,$29,$4E,$37,$F9,$7A,$DB,$F9,$39
	.dc.b	$DB,$67,$77,$D7,$EC,$3D,$64,$9F,$5B,$EE,$EC,$F8,$24,$8F,$34,$7F
	.dc.b	$AC,$BE,$15,$F5,$7F,$EA,$B6,$CF,$FB,$5F,$A3,$FD,$CF,$39,$3E,$4F
	.dc.b	$CF,$7C,$C2,$4F,$FB,$CE,$2B,$EB,$8B,$8F,$FA,$EE,$6F,$0F,$E1,$F5
	.dc.b	$FF,$3F,$B6,$BC,$93,$F6,$3E,$CB,$EB,$93,$0F,$AE,$FD,$AF,$8B,$FB
	.dc.b	$A4,$8F,$A5,$1E,$DF,$89,$FF,$78,$F0,$FF,$B5,$1B,$2D,$1A,$FF,$B1
	.dc.b	$FE,$13,$E8,$C6,$27,$6E,$AA,$D6,$EF,$A5,$7F,$87,$6F,$8C,$FF,$BF
	.dc.b	$A7,$32,$9C,$D7,$23,$9C,$76,$FB,$4F,$F7,$FB,$CE,$BF,$C5,$18,$27
	.dc.b	$33,$CE,$73,$18,$BF,$4B,$FE,$50,$6F,$C7,$75,$25,$75,$3E,$14,$EE
	.dc.b	$B9,$CE,$A7,$B1,$A7,$CD,$E1,$E6,$FA,$CF,$3C,$3C,$D5,$A8,$C7,$A0
	.dc.b	$F9,$AF,$96,$BC,$FA,$62,$24,$F0,$8C,$BC,$A3,$2F,$92,$FF,$E0,$0C
	.dc.b	$EA,$31,$5D,$DF,$70,$F2,$DF,$2E,$FB,$1C,$C6,$1B,$19,$9F,$83,$F0
	.dc.b	$1F,$E7,$1F,$73,$F1,$31,$CE,$75,$AF,$EE,$23,$17,$C8,$F9,$D7,$C6
	.dc.b	$30,$7F,$96,$4D,$97,$0F,$13,$6F,$C4,$FF,$97,$E1,$F3,$AE,$2E,$53
	.dc.b	$ED,$8C,$3C,$DA,$F1,$6B,$FE,$6F,$9E,$73,$F0,$A7,$8E,$3F,$C4,$B0
	.dc.b	$7B,$3F,$E1,$9E,$9D,$5C,$C7,$F0,$A7,$85,$F7,$CE,$E3,$38,$F6,$7E
	.dc.b	$C7,$ED,$0F,$97,$FF,$7D,$E5,$FB,$0F,$C3,$BE,$A6,$5F,$8B,$AD,$3F
	.dc.b	$8D,$8C,$DF,$2B,$E3,$CD,$FA,$69,$CE,$1F,$BA,$C3,$8D,$FE,$1E,$73
	.dc.b	$5F,$AC,$9C,$D7,$87,$7E,$77,$FC,$16,$73,$5D,$67,$3C,$6E,$27,$BF
	.dc.b	$86,$E9,$FA,$0F,$98,$C4,$C8,$FD,$1B,$B8,$C6,$5F,$98,$8B,$9F,$F1
	.dc.b	$2B,$56,$32,$FE,$07,$1D,$AF,$86,$1D,$CE,$A7,$EC,$FE,$DB,$E9,$61
	.dc.b	$FC,$71,$B7,$FC,$31,$FD,$1C,$EF,$CC,$F6,$C6,$AF,$F9,$33,$73,$3C
	.dc.b	$23,$F6,$4E,$49,$3C,$B9,$79,$BF,$33,$C7,$FE,$D9,$D4,$C7,$FB,$43
	.dc.b	$70,$FE,$C5,$D7,$7C,$3F,$ED,$53,$6B,$FA,$44,$FF,$BD,$AA,$C7,$F7
	.dc.b	$A8,$C4,$7F,$BF,$FC,$3F,$F7,$AC,$FF,$7E,$F4,$F9,$4E,$5B,$AF,$B3
	.dc.b	$AF,$39,$FF,$80,$66,$E0,$69,$C3,$6F,$D9,$E3,$FE,$01,$8C,$B4,$FF
	.dc.b	$26,$46,$B1,$FE,$FD,$3C,$70,$F3,$E3,$E1,$BF,$EE,$17,$BB,$FF,$72
	.dc.b	$4E,$5F,$F0,$AE,$1F,$E9,$9A,$FF,$08,$7B,$7E,$6F,$4F,$CE,$D8,$FF
	.dc.b	$0E,$C7,$C4,$CF,$FC,$38,$DA,$37,$CE,$F8,$7C,$AF,$B3,$F8,$31,$86
	.dc.b	$E1,$FE,$67,$9D,$3F,$E2,$19,$D3,$49,$97,$F8,$77,$CE,$E2,$D8,$FF
	.dc.b	$18,$C7,$8C,$8F,$F8,$D3,$CC,$FA,$27,$D2,$FB,$DB,$1F,$E2,$F8,$B5
	.dc.b	$FE,$29,$9B,$C9,$FE,$31,$EB,$FB,$27,$DA,$F8,$DB,$87,$F8,$DE,$2D
	.dc.b	$FF,$8C,$66,$D7,$F9,$77,$A3,$F0,$7E,$9F,$E7,$E3,$F2,$DF,$27,$FE
	.dc.b	$D8,$C3,$35,$F8,$C8,$C6,$3F,$CF,$5B,$88,$81,$CB,$86,$BB,$6B,$BB
	.dc.b	$FE,$42,$ED,$AE,$DA,$B8,$BF,$CA,$3E,$CF,$E7,$3B,$3F,$D5,$98,$7F
	.dc.b	$A7,$FE,$D7,$F9,$B6,$1F,$E8,$2A,$D1,$27,$5D,$B7,$1F,$F4,$4F,$B7
	.dc.b	$CF,$E8,$BA,$36,$BF,$DF,$F9,$23,$15,$C2,$73,$5F,$AE,$F7,$1E,$8D
	.dc.b	$3E,$57,$CB,$FE,$32,$FF,$D7,$53,$AB,$F9,$5A,$7D,$2A,$D5,$7C,$96
	.dc.b	$BF,$25,$AF,$CF,$6E,$E6,$93,$A7,$F8,$4C,$65,$FD,$47,$DA,$BE,$9F
	.dc.b	$F8,$2A,$F8,$FB,$18,$6E,$5E,$63,$FD,$49,$3B,$71,$8F,$D5,$8E,$91
	.dc.b	$8E,$34,$6E,$7F,$DB,$5E,$B3,$D1,$FE,$2A,$DB,$FF,$6C,$7C,$1F,$C7
	.dc.b	$DB,$F9,$56,$3F,$E7,$CB,$D9,$2B,$7A,$5C,$7F,$DC,$DF,$09,$FE,$D3
	.dc.b	$9F,$13,$FC,$5F,$0E,$6E,$67,$7E,$37,$F2,$AA,$7E,$0E,$F8,$4F,$E6
	.dc.b	$31,$86,$C6,$63,$FC,$40,$64,$AE,$6B,$F4,$A7,$F4,$57,$0F,$EA,$23
	.dc.b	$83,$A6,$BF,$EB,$5F,$B1,$78,$7F,$CE,$97,$87,$63,$0D,$AF,$A1,$3A
	.dc.b	$6C,$E9,$FA,$9F,$63,$FE,$A6,$6F,$0F,$F4,$68,$7D,$55,$89,$3A,$38
	.dc.b	$C6,$31,$6D,$F5,$BE,$0B,$FC,$C9,$5F,$81,$6C,$EA,$3F,$35,$F1,$9F
	.dc.b	$D5,$27,$35,$FA,$18,$FC,$BF,$49,$A4,$AF,$E2,$51,$96,$C6,$6F,$E7
	.dc.b	$F4,$FE,$8E,$3F,$A3,$7D,$73,$E0,$3F,$B1,$BF,$B1,$AF,$21,$FC,$35
	.dc.b	$D5,$FE,$65,$D6,$3F,$92,$F9,$07,$D7,$7F,$51,$6D,$7F,$98,$7A,$51
	.dc.b	$8C,$3F,$DE,$2D,$7F,$DE,$0F,$F2,$99,$FD,$33,$63,$C7,$EF,$3E,$49
	.dc.b	$F3,$BF,$DE,$AB,$59,$7E,$93,$E5,$1C,$3F,$DF,$9B,$97,$ED,$31,$7C
	.dc.b	$AF,$8E,$F8,$5F,$F8,$22,$73,$AF,$F8,$39,$B7,$FE,$7B,$6E,$9F,$9A
	.dc.b	$6B,$FD,$01,$AF,$F4,$36,$A6,$98,$D5,$F2,$EF,$A5,$FF,$0A,$B6,$FF
	.dc.b	$84,$7C,$CB,$ED,$B3,$5F,$E1,$A8,$C4,$FC,$F6,$E3,$FA,$EC,$11,$88
	.dc.b	$FE,$AB,$18,$BF,$F0,$44,$E6,$7F,$AD,$4E,$6B,$A5,$68,$FF,$07,$7C
	.dc.b	$DF,$F1,$CA,$FC,$3D,$23,$1A,$FF,$16,$FA,$DF,$D4,$E1,$FC,$86,$72
	.dc.b	$F6,$8C,$DF,$9D,$AF,$99,$AE,$FA,$7F,$C3,$D1,$88,$FE,$C4,$DC,$0D
	.dc.b	$31,$F3,$AF,$57,$FC,$82,$DC,$3F,$9A,$47,$E3,$1F,$E6,$D1,$9B,$F6
	.dc.b	$CE,$5F,$98,$D7,$E5,$B5,$F9,$2D,$7F,$BB,$36,$FC,$5C,$BF,$E3,$78
	.dc.b	$C1,$9F,$AC,$FE,$2D,$5F,$31,$BA,$7F,$40,$F7,$7F,$82,$E5,$F5,$A7
	.dc.b	$2F,$D8,$8F,$E1,$B5,$F7,$9A,$E7,$E3,$3F,$C8,$D1,$88,$FF,$05,$37
	.dc.b	$23,$49,$3F,$E4,$DF,$8B,$FF,$37,$D7,$E0,$1B,$8F,$F5,$4F,$A1,$7C
	.dc.b	$AF,$F9,$BA,$73,$97,$E3,$27,$2F,$F0,$56,$DF,$E5,$DA,$FE,$59,$AF
	.dc.b	$1A,$FF,$2F,$46,$23,$FC,$44,$DC,$8D,$F8,$6F,$E0,$B7,$F9,$C6,$D7
	.dc.b	$F8,$A6,$3F,$C8,$06,$5F,$F4,$2B,$FE,$A1,$7F,$4E,$D7,$F4,$AD,$79
	.dc.b	$7B,$5F,$E7,$1F,$AA,$FE,$4F,$1F,$A4,$6E,$7F,$EA,$1E,$97,$D5,$78
	.dc.b	$B3,$56,$B1,$FD,$8B,$70,$FF,$1F,$36,$F8,$C7,$FA,$6A,$35,$5F,$D4
	.dc.b	$3D,$4F,$F4,$EE,$3F,$AE,$AF,$B7,$F2,$5F,$C7,$F1,$FD,$EC,$E6,$F8
	.dc.b	$47,$F8,$56,$38,$70,$AB,$85,$FF,$93,$E7,$3F,$4C,$FA,$6F,$F5,$6B
	.dc.b	$70,$FF,$10,$E1,$FE,$E9,$F2,$B3,$52,$73,$5B,$C3,$FA,$7D,$EE,$73
	.dc.b	$5F,$EB,$8F,$8C,$FE,$E9,$5F,$E0,$8A,$D3,$63,$2F,$FA,$A9,$AF,$F9
	.dc.b	$CE,$1F,$4E,$F9,$4F,$F6,$1B,$70,$F3,$E9,$FE,$F2,$7D,$31,$9B,$F2
	.dc.b	$B7,$4F,$F6,$D6,$1D,$30,$FF,$66,$E1,$C6,$F7,$1F,$ED,$27,$FD,$07
	.dc.b	$C7,$FE,$20,$7F,$C4,$2D,$72,$50,$E7,$EA,$5F,$1F,$FE,$5B,$9E,$51
	.dc.b	$8F,$3F,$53,$99,$FF,$16,$CE,$5D,$FB,$DF,$E1,$46,$E7,$D6,$31,$1C
	.dc.b	$1A,$FC,$9D,$3F,$CF,$CF,$C4,$F8,$0F,$F4,$ED,$7C,$29,$CC,$FD,$AD
	.dc.b	$1F,$53,$AD,$F9,$61,$6B,$F8,$6D,$F0,$C3,$94,$E7,$7F,$C8,$61,$78
	.dc.b	$1D,$9F,$1E,$51,$8C,$7E,$64,$62,$75,$39,$9F,$C4,$71,$FF,$14,$7F
	.dc.b	$23,$D9,$DE,$31,$1F,$92,$9E,$19,$7D,$E9,$CB,$6B,$81,$E5,$BF,$67
	.dc.b	$37,$FA,$3F,$E4,$6F,$19,$A7,$B7,$83,$E0,$70,$7F,$6E,$9D,$69,$FE
	.dc.b	$85,$6B,$FD,$B1,$B9,$7E,$42,$B8,$CF,$FB,$07,$AB,$F6,$4F,$FA,$35
	.dc.b	$AF,$F7,$46,$CF,$F1,$1F,$84,$FF,$80,$13,$BE,$8F,$4D,$6B,$1F,$F0
	.dc.b	$0B,$67,$FC,$31,$5C,$1F,$EF,$FF,$53,$FE,$40,$7F,$C7,$CD,$7F,$C2
	.dc.b	$0D,$7F,$C1,$ED,$7F,$C1,$BE,$AB,$E6,$7E,$4B,$63,$FD,$4C,$D7,$BF
	.dc.b	$9D,$F3,$BA,$FF,$C8,$31,$FE,$4D,$6B,$FE,$1F,$6B,$FE,$1E,$6B,$FD
	.dc.b	$07,$C6,$FD,$87,$0F,$EB,$9B,$3A,$9F,$F7,$63,$67,$FA,$5F,$CB,$7D
	.dc.b	$EF,$10,$D3,$C9,$7F,$82,$E8,$7D,$4F,$66,$E3,$FE,$63,$6B,$FE,$35
	.dc.b	$6B,$FD,$DE,$38,$11,$E0,$FA,$30,$FF,$41,$47,$FB,$2D,$AF,$F8,$C9
	.dc.b	$BC,$5F,$E0,$79,$FC,$AF,$91,$F8,$CA,$FF,$0B,$F3,$7C,$28,$C4,$7F
	.dc.b	$86,$23,$04,$7F,$CF,$51,$A7,$FC,$37,$1F,$2B,$EA,$5F,$61,$FA,$26
	.dc.b	$CF,$FB,$7D,$AF,$F8,$AA,$7F,$4F,$E5,$FF,$AE,$A3,$FD,$0C,$D7,$FC
	.dc.b	$B0,$D7,$FC,$AE,$D7,$FC,$AA,$D5,$92,$30,$DF,$73,$EF,$C6,$3C,$F7
	.dc.b	$47,$EE,$1B,$7F,$E5,$58,$F2,$D7,$EF,$3D,$17,$CB,$FE,$21,$B5,$FE
	.dc.b	$6F,$8C,$11,$C1,$0F,$F3,$94,$70,$7F,$CE,$9E,$A7,$E0,$B8,$3F,$DF
	.dc.b	$6D,$C3,$FC,$E2,$D9,$F6,$D7,$1A,$FF,$05,$7A,$DF,$D1,$FB,$3F,$D3
	.dc.b	$75,$FD,$2D,$AF,$FA,$5E,$34,$47,$4D,$7F,$D6,$DF,$54,$F8,$9F,$E3
	.dc.b	$6D,$81,$FE,$7E,$70,$FF,$AD,$27,$FD,$F7,$7F,$EB,$1F,$87,$FE,$47
	.dc.b	$ED,$1A,$7F,$41,$C7,$FD,$75,$7F,$CA,$3D,$BF,$C6,$71,$7F,$B1,$B8
	.dc.b	$FF,$6B,$18,$33,$F3,$8F,$AA,$FE,$5E,$FF,$15,$A3,$4D,$5F,$A3,$3F
	.dc.b	$BA,$BC,$C7,$D3,$7F,$6F,$6C,$7F,$73,$6B,$FD,$DD,$AF,$C2,$3E,$1B
	.dc.b	$87,$0D,$73,$1D,$4F,$E1,$FE,$F7,$F3,$8C,$3F,$DA,$B5,$FD,$22,$3F
	.dc.b	$9D,$D7,$96,$BF,$A5,$FD,$57,$EE,$B9,$7F,$D6,$6D,$9F,$CF,$B7,$03
	.dc.b	$FD,$92,$3E,$04,$F0,$F8,$3E,$64,$8C,$9C,$BC,$FF,$EB,$3F,$4A,$FA
	.dc.b	$9B,$F5,$52,$B7,$92,$B5,$18,$6B,$FE,$C2,$53,$C6,$31,$5F,$E0,$17
	.dc.b	$E9,$7B,$1F,$44,$8F,$D1,$6B,$F4,$5A,$FD,$13,$89,$FE,$00,$FA,$77
	.dc.b	$CF,$7F,$72,$FA,$A7,$C7,$53,$FC,$19,$3A,$AF,$5F,$C8,$7E,$19,$FF
	.dc.b	$6C,$B7,$93,$F8,$1E,$A8,$C7,$07,$F9,$2F,$E4,$9E,$9F,$F8,$02,$73
	.dc.b	$87,$F1,$26,$DF,$1E,$8F,$F0,$E5,$F0,$AF,$97,$F4,$9E,$66,$F3,$7F
	.dc.b	$82,$A3,$F2,$39,$F8,$9D,$4F,$FB,$B9,$B1,$99,$E0,$71,$D7,$E7,$74
	.dc.b	$3E,$63,$F3,$C4,$C3,$F3,$CD,$7F,$3C,$D7,$E7,$47,$17,$FC,$B1,$9B
	.dc.b	$E5,$9F,$4F,$FE,$20,$6C,$FE,$19,$AB,$2F,$D8,$DF,$38,$C7,$07,$F0
	.dc.b	$CC,$3A,$78,$FC,$FE,$8E,$57,$38,$EC,$61,$E2,$7F,$40,$C5,$3A,$8F
	.dc.b	$F4,$8F,$1B,$E8,$7F,$02,$48,$FF,$4A,$FD,$57,$CA,$CB,$FD,$0D,$5B
	.dc.b	$9E,$35,$B7,$FD,$2B,$5B,$7F,$91,$FD,$57,$E8,$DF,$F3,$D3,$5F,$F3
	.dc.b	$F3,$5F,$CC,$35,$FE,$14,$D7,$9F,$67,$FA,$06,$31,$07,$5A,$3E,$77
	.dc.b	$FC,$3F,$BF,$E5,$9B,$9F,$CB,$6D,$F4,$3F,$73,$F4,$1F,$91,$7E,$4D
	.dc.b	$A3,$17,$FA,$06,$DF,$D5,$7F,$CC,$2D,$E9,$FF,$2F,$C6,$AF,$FB,$8B
	.dc.b	$6F,$A7,$C1,$7E,$C3,$CF,$95,$F5,$1F,$D9,$FC,$CE,$23,$FD,$83,$C1
	.dc.b	$FE,$01,$CF,$FD,$5A,$DC,$FE,$AD,$7F,$9B,$BE,$91,$E3,$7F,$8E,$9B
	.dc.b	$87,$E5,$9A,$FF,$07,$CD,$18,$8F,$E7,$51,$8D,$BA,$12,$3F,$8F,$CF
	.dc.b	$2C,$BF,$81,$BF,$46,$3F,$DB,$BF,$54,$FA,$4F,$F4,$C3,$63,$8D,$F3
	.dc.b	$7F,$DB,$B1,$96,$CE,$6F,$AF,$D5,$7A,$C9,$3E,$B7,$DD,$D5,$F0,$49
	.dc.b	$1E,$68,$FF,$46,$7C,$77,$FA,$97,$E1,$8D,$24,$FC,$5F,$84,$F9,$44
	.dc.b	$7F,$DC,$31,$5F,$C2,$E9,$FE,$71,$C7,$E0,$D7,$69,$F8,$AE,$E3,$31
	.dc.b	$C8,$F9,$3F,$37,$F5,$48,$FD,$57,$EC,$7C,$6F,$DB,$24,$70,$3F,$DE
	.dc.b	$7F,$15,$FE,$E8,$DF,$FD,$F5,$9F,$FA,$EF,$E1,$7E,$97,$C1,$23,$13
	.dc.b	$AB,$E0,$27,$57,$86,$C7,$F8,$72,$F9,$DF,$28,$E0,$FF,$BD,$AF,$C3
	.dc.b	$7C,$63,$57,$CA,$BF,$DE,$F5,$E1,$7A,$3F,$E2,$88,$7D,$57,$FB,$EA
	.dc.b	$72,$79,$5C,$3C,$8C,$7C,$87,$9E,$3F,$C9,$18,$FC,$07,$6D,$8D,$1B
	.dc.b	$F0,$ED,$76,$BF,$0C,$E2,$3B,$47,$92,$71,$5C,$E7,$85,$F6,$C3,$D4
	.dc.b	$F8,$9F,$67,$A2,$F9,$5F,$98,$BD,$0A,$62,$25,$F8,$E7,$4D,$9C,$BE
	.dc.b	$48,$E8,$E4,$86,$67,$EE,$AF,$2D,$F2,$AF,$A5,$F9,$B3,$98,$C3,$F3
	.dc.b	$4C,$47,$D2,$F6,$9F,$01,$FE,$57,$9E,$37,$D2,$AD,$F2,$D7,$F8,$15
	.dc.b	$72,$3E,$C6,$1C,$BD,$CF,$2E,$1C,$67,$FC,$30,$6C,$FF,$0D,$39,$6D
	.dc.b	$F9,$9C,$FD,$0B,$93,$8C,$7B,$E3,$0F,$4A,$46,$23,$8C,$F0,$9C,$BE
	.dc.b	$C9,$E3,$8F,$F1,$BC,$7F,$CD,$3E,$3F,$FA,$8A,$3D,$B7,$C3,$0F,$90
	.dc.b	$D8,$CD,$F2,$E0,$FF,$42,$E1,$CB,$C3,$ED,$97,$E4,$7E,$A7,$FD,$CF
	.dc.b	$1F,$93,$FA,$AF,$AC,$E7,$E3,$3C,$18,$7E,$9E,$73,$A7,$E5,$AF,$8B
	.dc.b	$F9,$88,$CE,$7F,$C0,$A7,$2F,$F0,$48,$CD,$7E,$C6,$72,$F5,$E0,$EF
	.dc.b	$7F,$C1,$67,$2F,$F0,$7A,$E2,$FF,$09,$E3,$7C,$07,$0E,$84,$63,$4F
	.dc.b	$CE,$7D,$4E,$63,$17,$F9,$BB,$C1,$99,$B5,$FA,$38,$B3,$FA,$19,$BA
	.dc.b	$7F,$1E,$BE,$6D,$AE,$57,$C4,$FD,$F7,$D5,$7F,$95,$1B,$C7,$E7,$8F
	.dc.b	$D2,$C6,$09,$8F,$F0,$28,$E0,$DE,$4E,$15,$2B,$78,$7F,$30,$BE,$77
	.dc.b	$C6,$BE,$5F,$8F,$FD,$BB,$B3,$FD,$D7,$E0,$7F,$68,$6E,$3F,$E3,$FF
	.dc.b	$0B,$FD,$D6,$FF,$6E,$8C,$E1,$FE,$FA,$9C,$BF,$DE,$5A,$FF,$78,$6B
	.dc.b	$CF,$BB,$FD,$7A,$F5,$F4,$B9,$E3,$FE,$00,$7A,$F0,$B4,$FD,$3E,$5F
	.dc.b	$DF,$A7,$78,$FF,$92,$A3,$07,$FA,$9A,$B8,$74,$EA,$A4,$EB,$1F,$F2
	.dc.b	$9D,$6A,$31,$A1,$FE,$85,$8F,$82,$BF,$D3,$55,$B6,$C6,$5F,$F6,$74
	.dc.b	$65,$F8,$1E,$E7,$98,$F3,$FA,$76,$FF,$C2,$EF,$C2,$9F,$94,$DD,$3F
	.dc.b	$C4,$2D,$D3,$FC,$C9,$5F,$E2,$7F,$3B,$E2,$F2,$BE,$9B,$E1,$C6,$1B
	.dc.b	$3F,$0E,$74,$FF,$89,$A7,$4D,$26,$3F,$E0,$AF,$3B,$8B,$63,$FC,$6D
	.dc.b	$1E,$32,$3F,$E3,$9F,$33,$E9,$9F,$53,$EA,$AC,$7F,$8D,$22,$D7,$F8
	.dc.b	$5A,$6F,$07,$F8,$DB,$B3,$ED,$9F,$73,$E3,$AE,$1F,$E3,$C8,$B7,$FE
	.dc.b	$1E,$9B,$5F,$E6,$1E,$9F,$C4,$FA,$7F,$9E,$8F,$B5,$4F,$93,$FF,$4E
	.dc.b	$61,$AB,$FC,$CC,$63,$1F,$E7,$ED,$AF,$F6,$C3,$51,$1C,$36,$FF,$90
	.dc.b	$BA,$3C,$0E,$1D,$5F,$DF,$9E,$30,$E2,$FF,$43,$7B,$5F,$A2,$F0,$7F
	.dc.b	$D8,$58,$FE,$6F,$EA,$BF,$7F,$97,$FA,$26,$B4,$49,$D7,$6D,$C7,$FD
	.dc.b	$0F,$83,$DB,$E9,$F5,$5D,$5B,$7E,$CC,$7F,$D2,$91,$8B,$E2,$D8,$FF
	.dc.b	$76,$FB,$8F,$43,$F6,$ED,$BE,$47,$C5,$F9,$8F,$C7,$E1,$FE,$98,$9D
	.dc.b	$3F,$EC,$2D,$D3,$A8,$F9,$11,$8E,$9F,$8D,$39,$9E,$91,$82,$5F,$D8
	.dc.b	$9D,$4F,$DF,$6F,$03,$EA,$9F,$3D,$FC,$39,$7C,$7D,$8C,$37,$0F,$5E
	.dc.b	$1F,$56,$B7,$18,$BE,$97,$C2,$31,$EB,$BD,$0F,$E3,$6D,$8F,$E2,$9F
	.dc.b	$07,$F1,$38,$3F,$2B,$87,$FA,$02,$F6,$4A,$DE,$B7,$3F,$F6,$0F,$C2
	.dc.b	$BE,$07,$F9,$3B,$FC,$57,$1E,$6E,$6F,$CC,$E5,$FC,$C8,$FA,$A7,$C2
	.dc.b	$7F,$11,$8C,$36,$33,$1F,$E2,$23,$25,$73,$87,$E7,$CF,$E5,$0E,$1F
	.dc.b	$D1,$47,$07,$4D,$7F,$D7,$7F,$54,$F0,$FF,$A0,$AF,$0E,$C6,$1B,$5F
	.dc.b	$46,$74,$D9,$D3,$F5,$7E,$AB,$FD,$56,$DE,$1F,$E9,$B0,$FA,$6B,$12
	.dc.b	$74,$71,$8C,$63,$FE,$F0,$F8,$3F,$F3,$44,$FE,$05,$B1,$F9,$BE,$77
	.dc.b	$C6,$FF,$5A,$9C,$F0,$7F,$12,$E9,$FC,$FC,$FE,$70,$98,$7E,$65,$B8
	.dc.b	$7F,$10,$C3,$9C,$E4,$91,$FE,$E5,$FA,$A7,$C0,$FF,$68,$7F,$5B,$8F
	.dc.b	$E3,$1C,$D7,$E5,$DC,$BF,$C6,$B9,$DF,$22,$F9,$DF,$EA,$0D,$9F,$F3
	.dc.b	$0C,$EA,$72,$6B,$1F,$EE,$F5,$F8,$C5,$87,$F3,$0F,$75,$F5,$4F,$9D
	.dc.b	$DD,$7F,$7E,$F9,$57,$0D,$3F,$C1,$76,$F9,$6F,$15,$74,$D7,$D1,$7C
	.dc.b	$1F,$DA,$D6,$F4,$FE,$D4,$D8,$FF,$3C,$B6,$BF,$2C,$D7,$F9,$B3,$5F
	.dc.b	$F3,$1D,$F8,$1C,$B5,$CD,$7F,$75,$F9,$87,$D5,$7F,$86,$1B,$87,$F0
	.dc.b	$5F,$85,$7D,$56,$6B,$FC,$3B,$18,$9F,$9A,$DC,$3F,$55,$C3,$FE,$9C
	.dc.b	$BF,$F0,$14,$E6,$7F,$A6,$4E,$6B,$8D,$68,$8F,$D8,$F9,$BF,$E3,$EF
	.dc.b	$F1,$76,$D7,$CB,$3F,$A8,$7D,$57,$E8,$E7,$5E,$37,$F8,$CA,$31,$3E
	.dc.b	$46,$BE,$26,$BF,$8B,$6B,$D3,$AB,$FC,$35,$18,$C4,$D3,$1F,$3A,$F5
	.dc.b	$7F,$C8,$2D,$C7,$F9,$74,$7E,$25,$FE,$61,$19,$C3,$D5,$7C,$79,$3F
	.dc.b	$C5,$17,$B9,$F9,$6D,$7E,$4B,$5F,$D1,$35,$FE,$E4,$D9,$E9,$7F,$A2
	.dc.b	$8C,$7D,$57,$F1,$0A,$F0,$B7,$9F,$F9,$DF,$BB,$FC,$02,$FF,$5D,$39
	.dc.b	$7F,$61,$1F,$C2,$2B,$ED,$B5,$CF,$C5,$FF,$90,$63,$11,$FE,$05,$6E
	.dc.b	$26,$92,$7F,$C9,$1F,$15,$FE,$72,$C7,$FA,$6B,$70,$FE,$01,$F4,$4F
	.dc.b	$95,$FF,$37,$CE,$72,$FF,$09,$4E,$6B,$F8,$0B,$63,$F2,$AD,$7F,$26
	.dc.b	$D7,$AD,$7F,$96,$63,$11,$D2,$30,$7F,$97,$7E,$1F,$F8,$46,$1F,$D8
	.dc.b	$9B,$7F,$C5,$4F,$F1,$D1,$FC,$5D,$CD,$7F,$89,$E3,$FD,$36,$FE,$91
	.dc.b	$AF,$1F,$6F,$FC,$D5,$18,$8F,$F1,$77,$D4,$FF,$24,$7F,$12,$DB,$FE
	.dc.b	$31,$F5,$5D,$B6,$4E,$6F,$F9,$43,$6F,$9F,$87,$FC,$62,$F8,$4E,$6F
	.dc.b	$F9,$1F,$37,$FA,$73,$1F,$D8,$9F,$E5,$4E,$07,$C9,$7F,$1E,$8F,$F0
	.dc.b	$19,$AC,$FF,$9B,$37,$3F,$F2,$33,$6B,$8F,$6F,$5C,$3F,$D4,$FF,$4E
	.dc.b	$FA,$7A,$7F,$AA,$BE,$3B,$FB,$1E,$3F,$EA,$09,$D6,$5F,$EA,$76,$C6
	.dc.b	$6B,$87,$97,$CA,$BF,$D5,$97,$CA,$BF,$D3,$ED,$7F,$CC,$B9,$7F,$BB
	.dc.b	$AB,$FD,$38,$DA,$FF,$03,$4F,$E6,$7E,$A5,$F2,$9F,$EB,$E6,$CF,$FA
	.dc.b	$E7,$67,$C3,$E6,$A4,$FF,$AF,$6B,$77,$FE,$BA,$BD,$CE,$6B,$FD,$AB
	.dc.b	$3F,$E7,$3E,$5F,$0B,$4A,$1A,$D3,$CF,$F5,$4F,$90,$FF,$2B,$C6,$1A
	.dc.b	$FF,$88,$63,$5E,$77,$66,$C7,$2F,$7F,$FB,$2B,$79,$35,$18,$8E,$4D
	.dc.b	$7E,$5F,$E6,$6F,$82,$F4,$F0,$23,$13,$F3,$23,$13,$A9,$CC,$FF,$B6
	.dc.b	$1A,$FF,$B5,$9A,$F6,$8E,$07,$C6,$F8,$77,$E5,$AE,$DC,$36,$BE,$DC
	.dc.b	$EA,$72,$78,$7E,$FE,$4F,$97,$04,$62,$3C,$8D,$7C,$8D,$7F,$C2,$DA
	.dc.b	$FE,$09,$FF,$09,$B5,$FE,$CE,$D7,$FB,$2B,$5E,$3C,$9F,$80,$EB,$FF
	.dc.b	$47,$B6,$75,$3F,$EB,$16,$BF,$E9,$DF,$BD,$FC,$2E,$C6,$92,$3F,$17
	.dc.b	$19,$6C,$65,$FC,$FC,$65,$FD,$0F,$C2,$FD,$07,$FD,$96,$D4,$D3,$54
	.dc.b	$FC,$03,$FD,$75,$3F,$AA,$8C,$7C,$3F,$E7,$1B,$19,$9D,$BE,$09,$FC
	.dc.b	$FD,$62,$3F,$CF,$0D,$7F,$DB,$0D,$7C,$B1,$C0,$8F,$EA,$A3,$55,$E5
	.dc.b	$FB,$6F,$15,$9A,$4A,$FF,$19,$FE,$AC,$FA,$6F,$D9,$FA,$C8,$C4,$EA
	.dc.b	$73,$18,$8F,$F1,$CB,$5F,$F1,$CB,$5F,$F7,$AC,$F3,$9E,$27,$E0,$79
	.dc.b	$7F,$23,$3F,$E5,$C6,$BF,$E3,$06,$BF,$DD,$38,$3C,$BF,$B3,$3E,$4A
	.dc.b	$92,$7F,$C9,$B1,$96,$C6,$5F,$F2,$7C,$65,$B1,$F9,$3F,$3B,$F2,$3C
	.dc.b	$7F,$E2,$1E,$3C,$12,$BF,$CB,$3F,$98,$3E,$57,$E3,$B6,$33,$18,$23
	.dc.b	$82,$1F,$E6,$08,$E0,$FF,$93,$BD,$2F,$F5,$73,$FE,$7E,$6B,$9F,$85
	.dc.b	$FD,$44,$62,$3F,$CC,$FF,$54,$F9,$AF,$B2,$D8,$FF,$37,$B5,$FF,$37
	.dc.b	$B5,$E3,$1C,$08,$F0,$7F,$6D,$ED,$3E,$6A,$92,$7F,$CC,$EF,$E2,$3E
	.dc.b	$6D,$F5,$34,$93,$FE,$80,$8C,$B6,$32,$FF,$A0,$A3,$2D,$8F,$C9,$FC
	.dc.b	$03,$CC,$A4,$9F,$F4,$5C,$65,$FF,$46,$FC,$1D,$7F,$D1,$AD,$7F,$D1
	.dc.b	$AD,$7F,$2A,$D5,$D1,$FE,$F2,$C7,$F4,$7F,$19,$F0,$67,$FD,$32,$D7
	.dc.b	$FD,$32,$D7,$FD,$32,$FD,$C6,$F0,$F2,$23,$FB,$7E,$BF,$E0,$24,$70
	.dc.b	$43,$F8,$AB,$96,$9F,$BF,$FB,$AF,$E1,$24,$70,$7F,$B1,$BE,$B5,$F2
	.dc.b	$3F,$CB,$DB,$19,$8C,$11,$C1,$0F,$F5,$0C,$70,$7F,$D4,$7F,$22,$FB
	.dc.b	$1F,$E7,$4D,$8F,$F6,$AB,$5D,$C6,$88,$E9,$A8,$83,$EF,$9F,$8C,$38
	.dc.b	$FF,$C0,$F1,$86,$C6,$67,$FC,$37,$39,$6C,$65,$EA,$75,$70,$F8,$23
	.dc.b	$83,$A6,$BE,$4F,$01,$F8,$93,$E4,$BF,$C0,$2D,$9F,$F0,$03,$5E,$F3
	.dc.b	$C2,$7F,$B0,$7C,$AF,$F9,$0E,$7F,$DF,$6D,$7F,$D9,$ED,$7F,$D9,$AD
	.dc.b	$7E,$37,$7A,$31,$08,$C7,$B5,$D8,$91,$FE,$3E,$87,$D3,$BE,$DF,$FB
	.dc.b	$93,$67,$4D,$7F,$DC,$2D,$7F,$DB,$ED,$7C,$6D,$7C,$4D,$9E,$F3,$FD
	.dc.b	$7B,$AA,$57,$F5,$76,$BF,$DC,$9A,$FF,$62,$6B,$F8,$5F,$85,$7E,$6E
	.dc.b	$F3,$FF,$C9,$31,$86,$C6,$67,$FC,$BB,$39,$7E,$17,$9C,$FA,$4F,$F0
	.dc.b	$93,$63,$31,$82,$38,$21,$F8,$AC,$BF,$C0,$B1,$F8,$2F,$AB,$93,$F8
	.dc.b	$26,$BF,$82,$E1,$46,$23,$F1,$91,$87,$81,$FE,$64,$F9,$DF,$EA,$CF
	.dc.b	$E6,$1A,$FE,$99,$AF,$EA,$5A,$9A,$4F,$2F,$FC,$DA,$F0,$E3,$7D,$43
	.dc.b	$F1,$7F,$F2,$DB,$6F,$FC,$B4,$D7,$F3,$6D,$7F,$6F,$9B,$E9,$7F,$C5
	.dc.b	$91,$86,$BF,$98,$6B,$F9,$76,$C7,$1F,$5B,$FC,$89,$18,$23,$FE,$47
	.dc.b	$FA,$77,$CC,$52,$5F,$E7,$63,$2D,$8C,$BF,$5E,$32,$D8,$FF,$0F,$FD
	.dc.b	$73,$E5,$E9,$27,$FD,$4B,$E4,$27,$51,$FE,$7D,$FB,$0F,$94,$4E,$0F
	.dc.b	$F4,$1B,$F5,$29,$F8,$EF,$FA,$71,$B5,$FE,$2E,$8C,$11,$C1,$0F,$0B
	.dc.b	$AF,$9B,$E0,$48,$CB,$F1,$7E,$8E,$47,$C3,$6B,$FC,$4B,$F5,$BF,$EA
	.dc.b	$CF,$FB,$11,$AB,$8D,$18,$6F,$9B,$F4,$7E,$72,$57,$CF,$6B,$F3,$5A
	.dc.b	$FE,$13,$81,$18,$8F,$E3,$31,$87,$91,$0E,$47,$FA,$A6,$FF,$C3,$CF
	.dc.b	$9F,$D6,$FF,$57,$C6,$23,$D0,$DE,$45,$7E,$56,$7B,$F3,$7E,$24,$8F
	.dc.b	$FB,$43,$C1,$FF,$55,$7C,$1F,$F5,$1F,$89,$7D,$3F,$F5,$29,$CD,$7F
	.dc.b	$B4,$E3,$1F,$0A,$BE,$11,$FE,$5E,$F5,$DC,$3F,$ED,$D8,$DC,$EA,$FF
	.dc.b	$AF,$61,$D3,$1E,$F1,$8C,$7A,$3D,$CF,$F0,$9E,$3D,$1E,$47,$F3,$A8
	.dc.b	$DC,$F2,$8C,$5E,$AF,$B6,$1E,$5C,$7C,$95,$DE,$B5,$87,$2A,$F3,$CF
	.dc.b	$87,$91,$97,$86,$3C,$87,$62,$E3,$D9,$E8,$71,$8E,$19,$79,$67,$53
	.dc.b	$C6,$31,$87,$86,$3D,$F9,$79,$9D,$63,$E4,$AD,$E5,$F2,$F2,$F2,$DE
	.dc.b	$F1,$EA,$4C,$7A,$4E,$B1,$F7,$DC,$7D,$33,$AC,$BC,$F9,$77,$24,$FB
	.dc.b	$32,$FA,$F8,$73,$8E,$38,$73,$9F,$2C,$7C,$1C,$BC,$98,$FB,$63,$D6
	.dc.b	$73,$C3,$EA,$3E,$33,$FC,$2B,$97,$E4,$1B,$3F,$84,$6E,$1F,$99,$F6
	.dc.b	$1E,$1D,$24,$FF,$87,$8F,$EF,$F1,$A7,$17,$FE,$87,$9F,$C3,$4E,$72
	.dc.b	$F6,$FD,$BB,$EA,$5F,$1D,$AB,$98,$FE,$03,$0E,$1A,$7F,$AF,$B0,$E1
	.dc.b	$7A,$C7,$F0,$15,$AB,$FD,$0B,$76,$5F,$09,$E9,$8F,$1C,$7F,$65,$7B
	.dc.b	$AD,$61,$D2,$B5,$7D,$1F,$C2,$4E,$AF,$8C,$6A,$7E,$1B,$AC,$BD,$B1
	.dc.b	$E0,$25,$F5,$7F,$45,$7C,$DE,$C6,$AF,$AB,$AA,$E7,$5C,$49,$3A,$7E
	.dc.b	$6C,$6A,$30,$49,$EB,$3A,$CB,$D1,$9F,$D7,$9F,$01,$FE,$96,$C3,$C8
	.dc.b	$FB,$0C,$96,$B9,$3C,$38,$9F,$8D,$7F,$9B,$F0,$FF,$22,$36,$7F,$C3
	.dc.b	$8D,$7F,$BE,$B5,$FE,$91,$C1,$F2,$1F,$F3,$27,$27,$F2,$0A,$FE,$71
	.dc.b	$39,$9D,$C7,$6D,$71,$1F,$E5,$7B,$FC,$CC,$E4,$91,$E2,$9D,$E1,$E8
	.dc.b	$6D,$70,$9D,$57,$07,$CF,$3A,$24,$7E,$8F,$A1,$FB,$2F,$1A,$FF,$3E
	.dc.b	$35,$FF,$49,$46,$AB,$CE,$DA,$F8,$FD,$5E,$22,$38,$3F,$BB,$FD,$07
	.dc.b	$BB,$8F,$FC,$02,$DE,$C3,$4C,$7B,$32,$BF,$BF,$B7,$2F,$F4,$BD,$FF
	.dc.b	$82,$A3,$11,$FE,$A6,$8C,$4F,$03,$ED,$F9,$BF,$18,$8E,$0F,$F0,$5F
	.dc.b	$C2,$F8,$C8,$E0,$FF,$15,$76,$FD,$0E,$D3,$49,$1A,$26,$1F,$E2,$D8
	.dc.b	$CB,$FE,$65,$F1,$9F,$65,$F6,$9B,$5F,$E9,$36,$CE,$E7,$44,$9D,$36
	.dc.b	$3F,$D5,$7F,$30,$EC,$FA,$AD,$AF,$BC,$D9,$DC,$FE,$41,$FF,$54,$7B
	.dc.b	$B7,$D4,$34,$97,$FC,$A6,$76,$D9,$DB,$67,$73,$FE,$BE,$F9,$87,$BF
	.dc.b	$F8,$86,$D7,$E4,$5B,$3B,$9D,$12,$74,$D8,$FF,$47,$77,$B7,$FD,$1B
	.dc.b	$6B,$FD,$78,$D9,$DC,$E8,$93,$A6,$C7,$FB,$53,$E3,$BF,$CC,$13,$FA
	.dc.b	$56,$C6,$7C,$BF,$D0,$C6,$23,$FD,$D3,$9F,$F9,$8E,$BE,$27,$97,$F2
	.dc.b	$C8,$E0,$93,$FD,$4E,$78,$36,$7F,$7B,$E7,$BF,$08,$A4,$BF,$F6,$C4
	.dc.b	$65,$B1,$9A,$FF,$59,$4E,$5B,$1F,$C2,$39,$5F,$61,$49,$7F,$EE,$38
	.dc.b	$CB,$63,$35,$FE,$BD,$9C,$B6,$3E,$E7,$4C,$AF,$B8,$D9,$10,$7B,$6F
	.dc.b	$85,$7D,$07,$F2,$28,$C3,$63,$37,$FD,$E6,$72,$D8,$CD,$75,$3A,$B8
	.dc.b	$7C,$13,$C1,$D3,$5F,$F5,$BF,$D5,$FE,$69,$FF,$6D,$35,$CF,$2F,$F3
	.dc.b	$08,$C4,$7B,$3C,$44,$62,$11,$89,$E0,$4C,$3F,$C2,$50,$FA,$B7,$83
	.dc.b	$FC,$F1,$B3,$A6,$DF,$FB,$B5,$AF,$FB,$39,$AF,$8D,$AF,$89,$B5,$DE
	.dc.b	$BF,$7D,$CD,$2F,$F7,$4D,$BF,$CA,$7A,$9F,$C0,$C9,$3F,$BF,$6C,$6A
	.dc.b	$30,$47,$F9,$AF,$EA,$FE,$2F,$04,$D2,$61,$FD,$0D,$FE,$83,$5F,$EC
	.dc.b	$C6,$D7,$2E,$0D,$1F,$D2,$7E,$21,$DD,$49,$7F,$ED,$48,$CB,$FE,$D6
	.dc.b	$F8,$97,$CD,$7F,$51,$6C,$66,$30,$7F,$55,$FD,$45,$E9,$D2,$4F,$FB
	.dc.b	$82,$32,$D8,$CE,$5F,$D5,$27,$2D,$8F,$EE,$FE,$A7,$F8,$0E,$3F,$2D
	.dc.b	$C3,$FE,$76,$BF,$C1,$CE,$67,$F4,$D3,$97,$FA,$FF,$EA,$5F,$EA,$DA
	.dc.b	$FF,$78,$B6,$33,$EA,$FF,$6D,$8C,$47,$F6,$36,$BF,$B2,$E1,$46,$21
	.dc.b	$18,$9E,$47,$F8,$3B,$EA,$1D,$94,$98,$7E,$6A,$32,$FF,$9B,$69,$F9
	.dc.b	$83,$E6,$FF,$C0,$0D,$9D,$37,$1F,$EF,$6D,$7F,$BD,$35,$F4,$35,$F3
	.dc.b	$B6,$BB,$D7,$F8,$93,$E5,$1F,$67,$49,$87,$EC,$23,$2F,$9F,$F6,$67
	.dc.b	$C4,$FF,$85,$5B,$19,$8C,$11,$C1,$32,$FF,$09,$CF,$07,$EE,$FA,$6F
	.dc.b	$96,$FF,$1B,$B6,$BF,$80,$B6,$77,$3F,$E2,$5F,$D9,$DF,$03,$49,$7F
	.dc.b	$36,$32,$D8,$CD,$7F,$89,$27,$2D,$8F,$F2,$6F,$CE,$3F,$1A,$F3,$4F
	.dc.b	$F3,$38,$C6,$7A,$9C,$CF,$F3,$89,$CE,$1F,$E0,$1F,$C0,$3F,$C2,$B3
	.dc.b	$FC,$BD,$AF,$F3,$D6,$BF,$D0,$1A,$9A,$4F,$53,$FC,$24,$F0,$AF,$37
	.dc.b	$07,$8C,$97,$FE,$B9,$8C,$BE,$3F,$96,$E2,$47,$07,$E4,$FF,$44,$7C
	.dc.b	$37,$F9,$45,$B1,$98,$C1,$1C,$13,$2F,$F2,$7C,$F0,$7F,$73,$F4,$5F
	.dc.b	$DB,$5F,$DB,$B5,$FE,$88,$D7,$FA,$1B,$7A,$CD,$27,$9F,$FE,$7C,$78
	.dc.b	$71,$BD,$3F,$C3,$5F,$F6,$98,$C7,$27,$F8,$48,$99,$7F,$85,$DB,$3F
	.dc.b	$E1,$B6,$BF,$DC,$7B,$1F,$9C,$7F,$01,$69,$FE,$92,$2D,$7F,$B9,$DB
	.dc.b	$3F,$E2,$5E,$4F,$F1,$09,$27,$FC,$4C,$D8,$D4,$60,$8F,$F9,$24,$7C
	.dc.b	$25,$69,$97,$07,$F9,$3F,$E9,$1F,$4A,$E3,$C3,$6B,$FD,$F9,$3A,$9C
	.dc.b	$9C,$7E,$53,$E7,$CE,$9A,$FF,$81,$E5,$18,$8F,$F0,$B4,$63,$7F,$85
	.dc.b	$F9,$1F,$F8,$A2,$7F,$C3,$71,$FE,$E6,$D1,$39,$9F,$BB,$9F,$F8,$82
	.dc.b	$73,$EB,$7F,$88,$6F,$FD,$DB,$C1,$F7,$79,$F1,$7B,$C7,$95,$75,$6C
	.dc.b	$66,$F7,$8B,$0E,$4F,$92,$7F,$D6,$FC,$5F,$E6,$58,$FF,$2D,$7B,$7F
	.dc.b	$E7,$9A,$FC,$63,$67,$53,$FE,$29,$CD,$F7,$EF,$8F,$EF,$18,$6C,$66
	.dc.b	$7C,$07,$8E,$B9,$E9,$F1,$67,$53,$F1,$05,$F3,$78,$B5,$D6,$3E,$67
	.dc.b	$AF,$86,$FD,$11,$F4,$1F,$0A,$30,$D8,$CC,$F9,$3A,$BF,$11,$D3,$9A
	.dc.b	$93,$D2,$B6,$F9,$6B,$6F,$B3,$07,$D9,$FF,$AE,$23,$0D,$7C,$ED,$D1
	.dc.b	$C1,$FE,$5F,$E5,$7C,$9F,$D8,$C3,$9F,$83,$70,$E1,$7F,$EE,$E6,$E3
	.dc.b	$FE,$68,$C3,$8B,$F3,$BA,$1F,$4E,$FA,$8F,$F5,$A4,$61,$B1,$99,$FE
	.dc.b	$50,$D7,$FD,$D3,$F8,$97,$D2,$7E,$6B,$5F,$AD,$C4,$8C,$6D,$EC,$9F
	.dc.b	$D9,$57,$38,$C6,$1E,$4F,$8B,$F0,$92,$BC,$2D,$7F,$CE,$D1,$B7,$24
	.dc.b	$3F,$DF,$BF,$64,$FA,$9F,$BC,$D8,$FC,$1B,$5F,$C6,$4F,$37,$ED,$70
	.dc.b	$7E,$71,$FF,$15,$B5,$FF,$3C,$35,$FF,$3B,$B7,$8B,$F5,$9D,$5F,$ED
	.dc.b	$F6,$CF,$F7,$88,$C4,$F1,$D5,$F5,$4F,$BD,$FC,$A3,$6F,$F3,$2D,$7C
	.dc.b	$B7,$FB,$AC,$1E,$77,$F8,$1E,$7F,$C2,$DD,$3E,$18,$C4,$7F,$92,$1A
	.dc.b	$FF,$91,$9A,$FF,$A4,$DA,$FF,$A4,$9B,$3D,$A7,$FD,$F5,$8F,$FB,$A7
	.dc.b	$B7,$F5,$58,$7E,$DB,$83,$EF,$4E,$71,$FD,$56,$3E,$29,$FC,$16,$2F
	.dc.b	$A9,$7C,$7F,$C7,$6D,$FF,$A7,$9A,$FF,$98,$1A,$FF,$97,$9A,$FF,$AA
	.dc.b	$DA,$FF,$AA,$9B,$3D,$E7,$D3,$F5,$5D,$2F,$CE,$DE,$9F,$F2,$F3,$70
	.dc.b	$E5,$5F,$3F,$D7,$FF,$16,$C7,$F9,$C3,$D0,$FB,$D1,$88,$FF,$3A,$B5
	.dc.b	$FF,$39,$B5,$FF,$60,$35,$FF,$5F,$B7,$13,$9D,$3F,$8B,$FC,$D7,$F1
	.dc.b	$7F,$00,$D3,$C9,$87,$F1,$BF,$73,$FC,$DF,$8F,$FA,$67,$D8,$FA,$F1
	.dc.b	$88,$FF,$6F,$60,$8C,$47,$FB,$B6,$31,$3C,$09,$1C,$CF,$F3,$B7,$D3
	.dc.b	$3F,$00,$FE,$54,$D8,$FE,$5E,$D7,$F8,$1D,$F8,$78,$3F,$0B,$F2,$1F
	.dc.b	$1D,$B9,$7F,$33,$8C,$70,$FF,$A3,$FE,$03,$83,$FE,$E7,$F8,$4F,$F1
	.dc.b	$9E,$3F,$EE,$26,$BF,$EA,$46,$BF,$EA,$26,$BF,$EE,$FF,$9B,$7A,$9F
	.dc.b	$A9,$6C,$7F,$3F,$6B,$FC,$6B,$07,$C4,$FF,$68,$87,$D1,$BE,$AF,$FC
	.dc.b	$28,$D8,$10,$8E,$5C,$1A,$70,$E1,$AF,$08,$EA,$47,$89,$DB,$F5,$57
	.dc.b	$E1,$5E,$56,$C7,$F6,$C0,$E1,$FE,$D4,$7A,$A7,$F8,$57,$D6,$FF,$48
	.dc.b	$E9,$34,$9E,$EE,$BC,$E9,$7F,$D9,$67,$2F,$F6,$7F,$AD,$FE,$59,$C9
	.dc.b	$FB,$E6,$F1,$7C,$4C,$51,$8A,$46,$3E,$B7,$F8,$E4,$E6,$37,$E5,$60
	.dc.b	$CC,$FF,$63,$FA,$77,$D6,$52,$61,$FE,$0C,$F9,$47,$D6,$BB,$33,$3F
	.dc.b	$E0,$6B,$E2,$FF,$77,$6B,$FE,$0F,$AE,$33,$FD,$C7,$E7,$3F,$BD,$CF
	.dc.b	$F7,$D6,$BF,$E1,$06,$ED,$FE,$36,$BE,$6F,$F2,$0F,$9B,$FC,$6B,$7A
	.dc.b	$BF,$9D,$46,$38,$3E,$32,$FA,$DC,$CC,$C7,$FC,$29,$3B,$7E,$4B,$B7
	.dc.b	$FC,$31,$F4,$FF,$B1,$C3,$F9,$7B,$70,$FA,$4D,$9F,$A2,$FF,$B6,$9E
	.dc.b	$3E,$0F,$F4,$D8,$D7,$11,$A6,$E7,$FC,$69,$F4,$6F,$AD,$FF,$31,$C7
	.dc.b	$E5,$A3,$1E,$CF,$E6,$21,$18,$75,$3C,$1F,$D2,$C6,$5B,$19,$7F,$59
	.dc.b	$19,$6C,$65,$FD,$87,$CF,$7F,$04,$CF,$FC,$21,$08,$C6,$7A,$F8,$AF
	.dc.b	$F3,$1E,$1D,$EF,$FA,$AF,$7F,$F0,$0C,$3F,$C8,$4D,$7F,$C7,$CD,$8D
	.dc.b	$FD,$6F,$F7,$38,$C6,$1F,$9F,$6D,$7E,$79,$BA,$FF,$0A,$AF,$F1,$E7
	.dc.b	$E4,$1F,$12,$BF,$C9,$0D,$7F,$CB,$6D,$C3,$F1,$9F,$5D,$C3,$29,$5C
	.dc.b	$1F,$D2,$47,$F9,$1F,$EB,$9F,$5B,$FE,$79,$6D,$7F,$0A,$F6,$91,$88
	.dc.b	$FE,$2B,$18,$9E,$04,$8E,$13,$B7,$FD,$98,$D7,$F9,$0F,$53,$F1,$BF
	.dc.b	$9F,$08,$C7,$07,$FA,$49,$B1,$FE,$67,$AF,$16,$7F,$E0,$6F,$A8,$7E
	.dc.b	$2F,$FE,$1F,$6D,$7F,$9D,$5A,$FF,$9C,$5F,$F2,$FB,$63,$81,$1E,$1F
	.dc.b	$0B,$FE,$29,$FA,$A7,$D1,$D2,$4F,$FA,$06,$1F,$47,$88,$FF,$3F,$47
	.dc.b	$07,$53,$CE,$7F,$CE,$DF,$61,$FE,$98,$9F,$F7,$83,$5F,$E8,$BC,$08
	.dc.b	$C4,$7F,$AC,$DB,$EB,$3E,$B3,$FA,$3E,$BF,$E2,$C6,$F4,$FF,$4E,$9D
	.dc.b	$5F,$83,$E0,$3C,$3F,$5F,$23,$FD,$25,$8A,$3F,$D4,$18,$72,$9F,$04
	.dc.b	$6D,$FF,$54,$CE,$DF,$F5,$4C,$EF,$A9,$B6,$67,$FA,$DC,$62,$FF,$CC
	.dc.b	$13,$99,$FE,$C5,$39,$7F,$AC,$E3,$ED,$FA,$9F,$19,$23,$E8,$5F,$D5
	.dc.b	$E5,$E6,$48,$FE,$C3,$7E,$28,$FF,$6A,$7D,$73,$EB,$7F,$D4,$CD,$AF
	.dc.b	$F5,$93,$5F,$B3,$3A,$99,$1A,$7C,$0D,$21,$FE,$CC,$9E,$5B,$FF,$96
	.dc.b	$B3,$FC,$EE,$1F,$13,$A7,$E0,$92,$FC,$F3,$FA,$57,$FD,$E1,$EA,$3E
	.dc.b	$F3,$FD,$78,$D8,$FF,$64,$35,$FC,$7D,$7F,$84,$7C,$47,$D2,$FF,$AC
	.dc.b	$7E,$6B,$F2,$55,$FC,$D6,$31,$D1,$FE,$11,$AD,$FC,$9F,$F8,$52,$FF
	.dc.b	$31,$C7,$FE,$4E,$E2,$FF,$0B,$47,$2A,$DE,$28,$C6,$3D,$23,$FC,$3B
	.dc.b	$3A,$9E,$7F,$07,$F8,$37,$78,$F7,$CB,$C5,$87,$1C,$3C,$D3,$95,$97
	.dc.b	$8D,$C1,$A3,$C5,$5C,$A7,$38,$F8,$63,$F4,$58,$F8,$5D,$57,$03,$CF
	.dc.b	$87,$86,$B8,$47,$82,$B8,$3A,$BE,$B5,$B3,$E0,$61,$E5,$9E,$87,$97
	.dc.b	$1F,$7B,$8C,$3A,$CE,$E7,$24,$C7,$CA,$FC,$8C,$BE,$8C,$60,$97,$E3
	.dc.b	$3D,$55,$CE,$33,$1E,$17,$71,$98,$E1,$5D,$DB,$3F,$1A,$BB,$C6,$08
	.dc.b	$73,$3E,$23,$A7,$E9,$1F,$E5,$8B,$FC,$2B,$70,$FF,$80,$0F,$D0,$35
	.dc.b	$F5,$FF,$D6,$5F,$C9,$35,$FC,$4B,$5F,$F0,$63,$64,$FF,$63,$C3,$F2
	.dc.b	$8D,$C3,$F9,$07,$BB,$FE,$A2,$EA,$FC,$1C,$E7,$0E,$BE,$26,$0F,$C7
	.dc.b	$68,$BE,$13,$F5,$EB,$83,$E2,$AE,$0F,$D0,$9D,$E3,$E2,$9C,$B6,$75
	.dc.b	$F0,$6C,$73,$97,$C0,$BB,$5F,$9E,$C3,$9C,$E6,$3C,$53,$AC,$BC,$D8
	.dc.b	$73,$9C,$93,$1F,$01,$FB,$AC,$39,$C7,$C0,$35,$87,$59,$D5,$73,$78
	.dc.b	$C6,$49,$3E,$59,$D4,$62,$3E,$21,$D0,$D6,$3E,$C7,$35,$D1,$C9,$B3
	.dc.b	$33,$DB,$0F,$6D,$7C,$4E,$DF,$F4,$CF,$AF,$EF,$E7,$F8,$11,$89,$FF
	.dc.b	$55,$C7,$C3,$E7,$7D,$87,$FA,$BA,$7F,$93,$DF,$17,$FD,$1D,$19,$CB
	.dc.b	$FA,$AB,$5F,$EA,$6D,$76,$7F,$91,$8F,$EE,$BC,$CF,$07,$A4,$7F,$38
	.dc.b	$6E,$1E,$2E,$0C,$E7,$FE,$B2,$9B,$7C,$71,$FE,$77,$1C,$31,$FF,$30
	.dc.b	$C6,$09,$8B,$3F,$83,$7C,$63,$13,$EC,$C7,$AE,$1C,$DF,$44,$FF,$AA
	.dc.b	$B0,$EA,$FC,$08,$CE,$5E,$7B,$FE,$E3,$DD,$FE,$F9,$7F,$BE,$F9,$DE
	.dc.b	$5F,$80,$F7,$EF,$FE,$11,$8C,$4F,$F7,$0C,$95,$D1,$B1,$98,$C7,$D5
	.dc.b	$FB,$CE,$78,$3E,$47,$15,$F7,$0F,$AF,$38,$D3,$DE,$7F,$44,$AE,$13
	.dc.b	$B8,$F7,$4F,$4A,$93,$9C,$FF,$0B,$F0,$D8,$BE,$71,$89,$F5,$4F,$A4
	.dc.b	$8F,$F7,$6F,$6B,$EE,$F8,$1F,$EC,$C8,$FF,$21,$7C,$46,$26,$47,$F9
	.dc.b	$15,$DC,$63,$0F,$F2,$97,$C0,$7E,$1B,$CB,$37,$E0,$30,$CC,$3F,$CB
	.dc.b	$4E,$5A,$7F,$B6,$7D,$FF,$F6,$C3,$7D,$DF,$C7,$FD,$5F,$F2,$68,$C1
	.dc.b	$F1,$7A,$91,$9C,$BF,$CF,$6D,$47,$F9,$49,$C3,$7C,$0F,$C5,$CF,$26
	.dc.b	$CE,$A5,$39,$AF,$F4,$54,$E6,$1E,$A3,$CC,$FD,$2B,$6B,$FD,$D1,$F0
	.dc.b	$B8,$99,$1F,$EA,$07,$F5,$7E,$FA,$73,$8F,$DA,$F0,$5D,$FF,$B2,$6E
	.dc.b	$1F,$81,$5F,$1B,$99,$93,$97,$84,$62,$BF,$1E,$3D,$DF,$9E,$E1,$BE
	.dc.b	$17,$F1,$4F,$99,$FF,$27,$61,$F9,$3A,$E1,$C3,$FC,$0A,$73,$1F,$EA
	.dc.b	$A6,$BF,$40,$F9,$2E,$1F,$F4,$1B,$5C,$FA,$1F,$4A,$31,$C0,$69,$BE
	.dc.b	$BF,$F6,$E7,$27,$F1,$4F,$70,$DF,$7F,$0C,$CF,$F4,$D3,$A6,$CE,$72
	.dc.b	$FD,$3F,$6B,$AF,$F8,$FB,$6B,$F5,$CB,$E6,$B3,$32,$72,$F0,$8C,$57
	.dc.b	$EF,$34,$7E,$6A,$F9,$5F,$E5,$2D,$C3,$F3,$5C,$EF,$0B,$35,$F8,$46
	.dc.b	$4F,$F0,$3F,$73,$F1,$DF,$61,$8C,$BB,$61,$C6,$B9,$DF,$F3,$AB,$DC
	.dc.b	$60,$FE,$11,$F6,$0F,$81,$FE,$7A,$D8,$CF,$C6,$62,$64,$70,$77,$18
	.dc.b	$C7,$F8,$7F,$D8,$7F,$32,$E9,$FE,$B0,$DB,$FD,$8F,$32,$3F,$94,$D7
	.dc.b	$0F,$3B,$FA,$6C,$6E,$13,$B9,$CF,$97,$A7,$FA,$AC,$BF,$3D,$FF,$6D
	.dc.b	$E1,$F9,$56,$C7,$E6,$FC,$0D,$C6,$32,$FF,$07,$4F,$E8,$67,$FA,$5B
	.dc.b	$6B,$F3,$0F,$E6,$7B,$B9,$B8,$6B,$99,$E3,$18,$8F,$E1,$ED,$7A,$1F
	.dc.b	$EC,$8F,$C0,$5F,$0B,$FD,$B1,$B7,$FC,$FF,$D1,$7D,$97,$C6,$CB,$FC
	.dc.b	$05,$5A,$C3,$F9,$EE,$AF,$77,$FA,$FB,$71,$FF,$63,$B6,$7F,$DA,$ED
	.dc.b	$7F,$69,$8F,$6E,$5F,$EC,$39,$7E,$26,$7F,$A9,$CB,$E4,$7F,$96,$E1
	.dc.b	$E5,$AD,$67,$FE,$37,$A7,$D8,$BD,$7F,$ED,$6D,$BF,$67,$CA,$3E,$CB
	.dc.b	$35,$FD,$C2,$31,$3F,$EE,$66,$E7,$F4,$B0,$46,$23,$D2,$D7,$F9,$EC
	.dc.b	$6A,$FF,$C0,$AD,$AE,$91,$8A,$E1,$5F,$EB,$2F,$96,$FE,$AF,$5F,$45
	.dc.b	$B1,$FD,$AB,$E0,$9F,$03,$FE,$06,$9C,$E9,$FE,$35,$9C,$DF,$FB,$D9
	.dc.b	$B7,$F8,$AE,$8F,$F6,$84,$E7,$0F,$F0,$AD,$6A,$BF,$93,$56,$9E,$11
	.dc.b	$97,$F5,$2D,$7F,$D9,$3F,$61,$FD,$02,$FF,$10,$D8,$FF,$00,$FD,$8B
	.dc.b	$C7,$FF,$0E,$CE,$72,$FF,$26,$4E,$6F,$F3,$F7,$FB,$08,$C3,$E9,$6B
	.dc.b	$E6,$6B,$FB,$C6,$BC,$6F,$FC,$63,$18,$C4,$EB,$67,$E0,$9F,$E5,$7C
	.dc.b	$7F,$14,$DC,$7F,$83,$FC,$E3,$EC,$BF,$C6,$B3,$9C,$FF,$82,$CE,$6B
	.dc.b	$D3,$92,$31,$3F,$45,$AF,$CE,$6B,$DA,$FF,$C8,$F1,$88,$EB,$18,$3F
	.dc.b	$17,$F6,$2F,$AC,$FF,$23,$B7,$0F,$F1,$97,$94,$FA,$0C,$D7,$F9,$A2
	.dc.b	$31,$3E,$66,$E1,$F8,$56,$BF,$82,$6B,$FC,$A1,$AF,$F2,$46,$E5,$FE
	.dc.b	$14,$BF,$F2,$FC,$6B,$57,$EB,$3E,$5B,$FD,$DB,$76,$FF,$57,$7D,$23
	.dc.b	$EC,$9F,$5F,$FE,$62,$AD,$69,$FE,$28,$AD,$4F,$F9,$C1,$B1,$F8,$86
	.dc.b	$E1,$FC,$A1,$AF,$F8,$05,$AF,$3B,$EF,$3C,$5B,$1C,$5A,$FF,$16,$FA
	.dc.b	$87,$CA,$7F,$9D,$9B,$5F,$E6,$5B,$F5,$7D,$76,$6B,$FD,$33,$18,$9E
	.dc.b	$D3,$E9,$C0,$D3,$FC,$83,$F7,$5E,$2D,$3F,$D6,$6D,$AF,$E1,$BF,$64
	.dc.b	$F9,$6F,$F5,$0C,$E7,$4F,$F2,$2B,$78,$0E,$78,$17,$C7,$0F,$F2,$C4
	.dc.b	$E7,$EB,$BF,$85,$C7,$F1,$A6,$E1,$FD,$8B,$27,$B3,$FE,$96,$9C,$E2
	.dc.b	$69,$C3,$93,$FD,$59,$3B,$F8,$6F,$F5,$DF,$CA,$BE,$ED,$F5,$F8,$6E
	.dc.b	$9F,$E6,$BF,$80,$FC,$AE,$9F,$EF,$3B,$E1,$7B,$25,$6F,$A6,$ED,$FE
	.dc.b	$1F,$C3,$F2,$FC,$3F,$E3,$99,$FF,$1E,$12,$BF,$87,$E1,$FE,$68,$F0
	.dc.b	$3C,$BF,$0F,$FC,$43,$3F,$F2,$87,$56,$26,$5F,$FA,$8E,$F8,$3E,$79
	.dc.b	$CE,$7F,$E5,$C3,$FC,$FF,$F2,$7F,$C7,$1F,$F2,$93,$5C,$95,$D7,$D9
	.dc.b	$7E,$F7,$AA,$F7,$1A,$4A,$FE,$4F,$2F,$A2,$FF,$6D,$46,$1A,$FF,$82
	.dc.b	$A3,$5A,$76,$9C,$CF,$FA,$53,$7F,$F4,$4F,$C8,$FF,$99,$E9,$A6,$77
	.dc.b	$FF,$65,$5F,$FA,$7D,$AF,$FB,$23,$67,$D9,$7F,$95,$63,$58,$73,$BE
	.dc.b	$4D,$8F,$F4,$DD,$F2,$C3,$EF,$5F,$19,$E9,$F6,$5F,$B0,$7E,$BB,$5F
	.dc.b	$F1,$2E,$28,$C4,$7F,$9F,$E3,$1E,$07,$D1,$CF,$FD,$4B,$F5,$5E,$16
	.dc.b	$D7,$AE,$3C,$95,$C2,$30,$43,$FC,$D5,$92,$77,$87,$8F,$2F,$8F,$5F
	.dc.b	$E4,$8D,$DF,$05,$EF,$EB,$FB,$4D,$8C,$CF,$F8,$BF,$B1,$E6,$E6,$A4
	.dc.b	$FF,$BB,$6B,$6F,$F8,$92,$B6,$FF,$BC,$7B,$BF,$6A,$DE,$EF,$04,$62
	.dc.b	$3F,$D0,$5D,$8F,$91,$CD,$49,$E0,$7F,$B6,$E7,$4F,$F9,$0F,$C5,$70
	.dc.b	$FE,$6F,$04,$62,$7F,$BC,$C6,$27,$53,$99,$FF,$09,$7E,$0A,$FC,$03
	.dc.b	$E2,$36,$3F,$DC,$BC,$1F,$2B,$81,$39,$9F,$F2,$1D,$F3,$AF,$F2,$07
	.dc.b	$C3,$7F,$AC,$27,$C2,$D7,$FC,$70,$D7,$FC,$9E,$D7,$FC,$6E,$D5,$97
	.dc.b	$FA,$97,$E2,$FD,$D3,$98,$C4,$FF,$7E,$F8,$1F,$17,$5F,$FA,$31,$B1
	.dc.b	$FE,$88,$6C,$F0,$8F,$F2,$DF,$D8,$BE,$CB,$E5,$36,$3E,$4B,$5F,$F2
	.dc.b	$83,$5F,$F3,$3B,$5F,$F2,$7B,$5F,$4B,$64,$E7,$5F,$F4,$DF,$8D,$F9
	.dc.b	$37,$F9,$95,$7D,$0C,$7F,$D4,$B1,$88,$F3,$E1,$FB,$7C,$3F,$DF,$BD
	.dc.b	$9F,$81,$F9,$0F,$D7,$D7,$F9,$81,$AF,$CA,$6C,$F0,$8F,$C2,$C6,$09
	.dc.b	$3F,$C9,$6D,$F8,$87,$D9,$DF,$F0,$4D,$CB,$FD,$72,$DC,$3B,$5F,$97
	.dc.b	$EC,$BF,$DD,$11,$F9,$26,$BF,$E7,$96,$BF,$E9,$06,$BF,$E7,$86,$BF
	.dc.b	$82,$6E,$46,$9A,$C5,$F4,$72,$BE,$51,$FE,$C8,$70,$B2,$FC,$E4,$61
	.dc.b	$FF,$4C,$57,$99,$FF,$4D,$76,$3E,$83,$F5,$CD,$BF,$D6,$B5,$FF,$4A
	.dc.b	$35,$FF,$55,$B5,$FF,$49,$B5,$FF,$47,$47,$43,$E8,$7E,$08,$DB,$F8
	.dc.b	$B3,$67,$D7,$92,$31,$1F,$EB,$4B,$F1,$61,$FE,$B1,$F9,$57,$D0,$7F
	.dc.b	$05,$6D,$7F,$04,$6B,$FE,$AB,$6B,$FE,$C2,$6B,$FE,$AA,$6B,$FB,$46
	.dc.b	$EA,$73,$8B,$F1,$07,$E0,$FF,$CB,$DB,$7E,$A6,$BF,$EC,$CA,$F1,$E3
	.dc.b	$FE,$C9,$F9,$A7,$DA,$FF,$37,$F8,$9F,$F1,$75,$FC,$26,$BF,$ED,$49
	.dc.b	$DB,$FE,$E7,$5F,$23,$FC,$95,$B1,$FC,$91,$AF,$FB,$2D,$AF,$FA,$C3
	.dc.b	$E9,$5F,$2D,$FD,$1D,$B1,$F4,$9A,$FE,$2F,$A0,$B6,$4F,$F3,$FC,$FF
	.dc.b	$A9,$CB,$ED,$5F,$3B,$FE,$11,$6C,$7F,$85,$5A,$FF,$AC,$9A,$F6,$3B
	.dc.b	$B8,$70,$D7,$31,$D4,$FE,$E1,$F3,$3E,$2A,$FE,$CB,$E7,$A3,$19,$7F
	.dc.b	$84,$5B,$7C,$8F,$C8,$FE,$59,$FC,$B3,$33,$4C,$7D,$B7,$F5,$69,$F8
	.dc.b	$87,$C3,$8D,$79,$1F,$DC,$5D,$F9,$7F,$DC,$FE,$35,$F6,$74,$97,$FA
	.dc.b	$08,$FE,$1D,$89,$DC,$66,$76,$F8,$27,$5E,$47,$F8,$37,$8B,$FC,$BB
	.dc.b	$3F,$C1,$36,$FB,$BE,$66,$4F,$F0,$97,$DA,$3E,$9B,$15,$BB,$E2,$FB
	.dc.b	$2F,$8B,$CE,$B8,$3B,$BF,$F0,$77,$DD,$BE,$92,$9D,$27,$CD,$F3,$5F
	.dc.b	$85,$E4,$FC,$4E,$A8,$FC,$6F,$2F,$F7,$16,$CE,$9E,$15,$F8,$F6,$D7
	.dc.b	$1D,$3F,$C3,$D1,$C2,$75,$E9,$7F,$88,$7E,$B3,$F5,$F8,$78,$1F,$F6
	.dc.b	$B4,$FF,$0E,$6B,$F2,$DD,$BF,$9D,$6E,$07,$78,$FF,$81,$DB,$5D,$7C
	.dc.b	$3F,$F2,$75,$BE,$93,$F7,$2F,$E6,$CC,$7C,$47,$F9,$CE,$33,$1F,$9F
	.dc.b	$6B,$FA,$E6,$BC,$0F,$F3,$E3,$83,$7C,$3F,$E8,$B8,$FF,$33,$7C,$1F
	.dc.b	$F9,$3F,$DC,$3E,$4A,$FE,$17,$9F,$F8,$7F,$ED,$3E,$DD,$FE,$2D,$B1
	.dc.b	$9E,$5F,$F2,$54,$7F,$0A,$8C,$4F,$F3,$68,$DE,$9F,$71,$B1,$FC,$15
	.dc.b	$E1,$1F,$E7,$BF,$53,$F2,$DF,$4D,$2B,$FC,$F9,$1E,$73,$0F,$E4,$78
	.dc.b	$D3,$98,$FE,$BF,$F4,$8F,$B4,$F8,$ED,$E0,$FF,$50,$36,$7E,$80,$70
	.dc.b	$FD,$36,$BF,$C3,$DA,$FF,$1E,$6B,$C6,$7E,$7F,$C0,$BE,$7B,$FC,$F8
	.dc.b	$D9,$FD,$83,$5F,$E0,$B5,$E0,$D7,$FC,$9D,$F5,$4F,$B3,$FF,$42,$72
	.dc.b	$FF,$88,$67,$F8,$BB,$77,$FD,$3D,$6F,$0F,$EC,$AD,$7F,$6D,$3B,$5C
	.dc.b	$BF,$D9,$3D,$AF,$F4,$15,$6F,$0F,$F2,$F7,$D6,$3E,$93,$FD,$29,$F8
	.dc.b	$77,$FA,$2F,$3F,$F5,$A3,$72,$FE,$7D,$5B,$7F,$BA,$B5,$FE,$0D,$1B
	.dc.b	$C8,$F1,$EB,$56,$F1,$FF,$4B,$FA,$8F,$AE,$FF,$6A,$37,$1F,$F3,$F4
	.dc.b	$60,$FF,$6C,$62,$FB,$37,$CF,$FF,$AA,$DB,$5F,$C7,$DA,$FF,$34,$BE
	.dc.b	$8E,$7E,$B3,$81,$1E,$15,$5F,$E4,$15,$FE,$C3,$C7,$FA,$4B,$75,$FF
	.dc.b	$6C,$3D,$A3,$D7,$97,$F2,$1A,$FF,$49,$B5,$FE,$01,$8F,$BF,$C3,$F6
	.dc.b	$9F,$E5,$1F,$B0,$FF,$5E,$4F,$F8,$81,$AF,$F3,$D6,$DF,$69,$FF,$65
	.dc.b	$7D,$9F,$E9,$2B,$F3,$BA,$A3,$11,$FC,$3E,$31,$D7,$FE,$D3,$9C,$CF
	.dc.b	$F6,$DE,$06,$3F,$EE,$3D,$7F,$9A,$B6,$7F,$BF,$37,$0E,$71,$E7,$F8
	.dc.b	$EF,$F6,$74,$EE,$F9,$C7,$9B,$1C,$B2,$B8,$1F,$D5,$71,$F3,$E7,$FD
	.dc.b	$DB,$D2,$FF,$2C,$CF,$F8,$07,$2F,$F0,$4F,$C0,$FF,$BA,$E7,$33,$FE
	.dc.b	$12,$9C,$BF,$BE,$CB,$D7,$3E,$42,$47,$F6,$37,$FC,$C5,$A7,$F8,$37
	.dc.b	$C6,$7D,$9F,$FB,$2F,$E4,$FF,$39,$18,$F4,$FF,$C3,$F1,$89,$D7,$4F
	.dc.b	$F8,$8B,$8D,$F4,$7F,$90,$AD,$5F,$0A,$ED,$7C,$33,$FF,$14,$E3,$D5
	.dc.b	$5F,$EB,$71,$E7,$CF,$FE,$21,$8E,$73,$AC,$39,$E3,$D6,$39,$E3,$D7
	.dc.b	$0E,$31,$93,$FA,$C4,$78,$5B,$39,$C7,$C5,$87,$87,$2F,$56,$3D,$AF
	.dc.b	$C0,$76,$BE,$18,$77,$AD,$CE,$4D,$9D,$8B,$87,$6B,$F9,$13,$C2,$75
	.dc.b	$87,$B8,$93,$AC,$BE,$13,$E3,$8C,$CF,$77,$D1,$3A,$8C,$4E,$C8,$FF
	.dc.b	$BB,$AB,$C1,$7E,$38,$D9,$32,$F9,$2F,$A3,$1F,$91,$3B,$8F,$3D,$F9
	.dc.b	$5F,$87,$94,$78,$3F,$56,$F9,$E5,$F6,$A7,$FC,$05,$7F,$81,$6E,$0B
	.dc.b	$FE,$AF,$1F,$24,$DF,$D9,$3C,$6C,$DA,$31,$3F,$D7,$A3,$13,$A9,$CC
	.dc.b	$F2,$8F,$C0,$47,$0C,$3F,$21,$EC,$FF,$B0,$76,$CD,$4B,$FB,$37,$BD
	.dc.b	$3F,$41,$7B,$9C,$D6,$F0,$FD,$4D,$EE,$5A,$7F,$A1,$BE,$11,$96,$F2
	.dc.b	$FD,$95,$EE,$FF,$C5,$CF,$1B,$E1,$5B,$9F,$D4,$CE,$F3,$F7,$C7,$E5
	.dc.b	$AB,$6F,$90,$98,$F9,$32,$FA,$99,$FA,$63,$13,$AC,$BC,$75,$AC,$FC
	.dc.b	$F3,$93,$C3,$1E,$3C,$3B,$DF,$22,$5F,$2A,$F2,$61,$DE,$31,$3A,$9E
	.dc.b	$66,$B0,$EF,$19,$8F,$87,$87,$7B,$E5,$1C,$27,$58,$FC,$27,$ED,$DF
	.dc.b	$2C,$3B,$9F,$CF,$3E,$13,$D7,$CB,$EC,$E3,$FE,$83,$18,$DB,$F9,$67
	.dc.b	$1F,$FA,$45,$FE,$A9,$19,$CF,$FC,$1B,$39,$7F,$A7,$35,$FE,$70,$D7
	.dc.b	$9F,$AD,$FE,$C4,$7F,$A6,$FD,$DF,$3C,$3F,$A8,$BD,$79,$BF,$D8,$FF
	.dc.b	$69,$9E,$C4,$D2,$67,$FC,$EE,$6E,$1F,$D9,$E6,$E9,$FD,$2E,$6D,$7B
	.dc.b	$1F,$15,$7A,$9B,$87,$82,$7F,$B5,$E6,$9D,$E5,$FD,$9B,$E7,$BC,$87
	.dc.b	$C3,$E4,$FF,$0D,$78,$D8,$C3,$FD,$CB,$9F,$F8,$12,$3B,$CE,$A7,$FC
	.dc.b	$0B,$18,$F0,$0F,$6D,$9E,$FF,$C1,$3E,$17,$D6,$5B,$FF,$03,$45,$DF
	.dc.b	$FC,$0D,$37,$2F,$B9,$68,$E3,$7E,$43,$FA,$C6,$2F,$03,$38,$7F,$89
	.dc.b	$E3,$13,$FE,$1A,$9F,$31,$1F,$ED,$DE,$2F,$FB,$25,$FF,$67,$7C,$DB
	.dc.b	$E7,$7E,$E4,$E7,$0F,$F6,$54,$E6,$BF,$C9,$3E,$9F,$DF,$8C,$4F,$FB
	.dc.b	$6E,$B8,$69,$FE,$F8,$9C,$BF,$E6,$8C,$3F,$0D,$6F,$A4,$FC,$57,$97
	.dc.b	$AC,$BF,$C0,$93,$97,$FC,$CF,$5C,$2B,$6D,$8F,$F7,$37,$AF,$F4,$9B
	.dc.b	$C3,$F9,$FF,$3D,$EE,$BE,$1F,$F2,$D7,$BD,$FF,$CF,$B3,$9C,$3F,$D1
	.dc.b	$F7,$C2,$FF,$1E,$D8,$CC,$EE,$BF,$D2,$4D,$7F,$D2,$4D,$7F,$D2,$4D
	.dc.b	$7F,$D3,$4D,$EA,$61,$FE,$F1,$AF,$F8,$01,$F0,$6F,$8A,$F3,$F9,$DF
	.dc.b	$AD,$8C,$4F,$C0,$8C,$4E,$A3,$8B,$FB,$5F,$07,$F0,$79,$38,$BF,$84
	.dc.b	$46,$1B,$DD,$BE,$69,$87,$E3,$ED,$DB,$FE,$8F,$9E,$95,$A9,$E9,$5F
	.dc.b	$E3,$BD,$1D,$BF,$38,$FE,$21,$2F,$77,$E4,$65,$FC,$D3,$7F,$83,$5A
	.dc.b	$9F,$8D,$5B,$E2,$FD,$FB,$71,$FE,$DB,$3A,$24,$E9,$B1,$FE,$4B,$F6
	.dc.b	$FF,$02,$FA,$F5,$FF,$7E,$7C,$0F,$F1,$99,$CE,$1E,$A3,$DE,$A3,$5C
	.dc.b	$7B,$6D,$F3,$9D,$3D,$7D,$D7,$D9,$FF,$25,$5F,$1D,$63,$0D,$9E,$0D
	.dc.b	$C7,$C5,$97,$FC,$00,$8F,$C0,$7C,$97,$F2,$FD,$1E,$9E,$30,$F6,$46
	.dc.b	$1F,$C4,$CE,$70,$FF,$79,$78,$57,$8B,$E8,$74,$AB,$17,$FE,$73,$F7
	.dc.b	$FF,$88,$3C,$8F,$91,$F6,$6F,$8C,$FE,$80,$D8,$CF,$D9,$E2,$64,$70
	.dc.b	$77,$18,$CF,$DB,$F6,$67,$FA,$53,$6B,$EB,$3F,$E2,$76,$C6,$88,$EB
	.dc.b	$D3,$B1,$F6,$7A,$5A,$7F,$87,$75,$F7,$FA,$1F,$92,$FB,$3F,$F7,$46
	.dc.b	$3F,$AE,$8C,$77,$FE,$3A,$73,$8F,$94,$8E,$1C,$07,$0E,$A3,$91,$2B
	.dc.b	$FC,$87,$3A,$9F,$D5,$FD,$93,$EF,$FF,$B5,$36,$BF,$2C,$BE,$D3,$33
	.dc.b	$FD,$5F,$EC,$DF,$27,$FD,$51,$B7,$E4,$C3,$FC,$6F,$18,$7F,$DB,$2D
	.dc.b	$7D,$9E,$92,$3E,$FC,$7E,$42,$9F,$0F,$DF,$A3,$FB,$F4,$63,$0F,$D5
	.dc.b	$3F,$CD,$6B,$87,$99,$FD,$AF,$EC,$FF,$C1,$9C,$3F,$E1,$5A,$D3,$FE
	.dc.b	$32,$CD,$3F,$E0,$DB,$FF,$26,$B6,$33,$EB,$FF,$70,$8C,$11,$C1,$31
	.dc.b	$FF,$64,$7C,$BB,$EC,$FF,$C2,$8D,$8F,$E3,$57,$FC,$AF,$D0,$CD,$7F
	.dc.b	$80,$7C,$DF,$EB,$39,$7C,$46,$BF,$EC,$56,$BF,$E6,$26,$D7,$3B,$FA
	.dc.b	$0D,$9F,$C2,$B5,$E9,$7F,$E1,$2F,$05,$F1,$DF,$E2,$96,$DF,$9B,$EC
	.dc.b	$9E,$E6,$6B,$FC,$2D,$18,$9E,$ED,$C3,$BB,$5F,$BA,$D7,$FD,$BC,$D7
	.dc.b	$FC,$F6,$DC,$4D,$35,$E3,$3E,$EF,$FD,$39,$7F,$CA,$5B,$7F,$C9,$ED
	.dc.b	$F6,$5F,$B9,$8F,$E3,$33,$AE,$2F,$C0,$4E,$67,$F2,$2D,$7F,$DE,$4D
	.dc.b	$7F,$D1,$ED,$AE,$B1,$FE,$32,$8C,$11,$FF,$1E,$FE,$11,$FE,$10,$C3
	.dc.b	$F8,$66,$28,$C4,$FF,$A6,$BE,$C9,$F5,$9F,$E5,$99,$CE,$29,$CE,$1F
	.dc.b	$71,$B7,$F7,$1A,$FE,$A9,$AF,$F1,$B9,$E5,$7F,$E4,$89,$CC,$F8,$63
	.dc.b	$07,$F9,$53,$C3,$EB,$9F,$F9,$ED,$B8,$73,$FB,$2B,$7F,$DC,$A6,$E2
	.dc.b	$9E,$AD,$C3,$FC,$AF,$18,$23,$82,$5F,$E6,$23,$CB,$9B,$E9,$7F,$7D
	.dc.b	$8F,$F5,$26,$E5,$FB,$9F,$A5,$FE,$B7,$87,$F6,$B9,$CB,$F0,$63,$35
	.dc.b	$FA,$26,$BF,$C4,$1A,$EF,$CE,$FF,$3A,$FE,$0D,$FC,$7E,$BF,$BB,$36
	.dc.b	$7F,$9E,$7D,$8B,$EC,$3F,$D2,$13,$9C,$BF,$D4,$4D,$BF,$C0,$B6,$DE
	.dc.b	$53,$0E,$55,$DB,$C5,$FF,$4D,$61,$FD,$EA,$FF,$B8,$B4,$FE,$31,$F5
	.dc.b	$5F,$E5,$58,$FF,$11,$C6,$6B,$F1,$ED,$8C,$FA,$DF,$E8,$DA,$E1,$BF
	.dc.b	$F8,$C6,$3F,$CA,$3D,$53,$0F,$F1,$77,$C6,$BE,$59,$F3,$B9,$CB,$FA
	.dc.b	$E7,$BB,$FB,$FD,$3F,$C5,$0D,$C3,$FB,$56,$1C,$6F,$85,$7F,$B8,$23
	.dc.b	$13,$FD,$15,$B7,$FD,$69,$AF,$F5,$F6,$BF,$E3,$BF,$7B,$FC,$CE,$7F
	.dc.b	$B2,$63,$FC,$AF,$ED,$62,$BF,$CE,$33,$6B,$FC,$55,$5A,$9C,$9F,$E4
	.dc.b	$B8,$FE,$D9,$F1,$9F,$D5,$DA,$50,$D6,$9F,$D7,$7E,$C3,$F7,$3F,$59
	.dc.b	$C5,$B1,$9E,$B4,$E5,$B1,$C7,$EC,$3F,$CC,$4D,$D3,$94,$62,$3F,$BE
	.dc.b	$DA,$31,$1F,$E5,$DF,$83,$FF,$02,$57,$D2,$6E,$59,$9F,$B7,$F3,$8F
	.dc.b	$9A,$CC,$7E,$4B,$87,$FD,$70,$FF,$A6,$2F,$3D,$3F,$E8,$27,$7B,$B0
	.dc.b	$E5,$5F,$86,$FA,$FF,$F2,$5B,$F6,$DA,$FD,$B6,$BF,$6D,$AF,$F9,$D7
	.dc.b	$EC,$3B,$E2,$69,$E0,$9E,$7A,$7E,$3A,$33,$5F,$EA,$E7,$5D,$DE,$CE
	.dc.b	$9F,$9F,$F6,$17,$8C,$D3,$E0,$F8,$6E,$CF,$F4,$E6,$3F,$E7,$3A,$DB
	.dc.b	$6B,$83,$FD,$E2,$76,$FF,$99,$2B,$FB,$4B,$FE,$7B,$6B,$9F,$2B,$E5
	.dc.b	$F5,$3E,$97,$CD,$5A,$D5,$CB,$FD,$1A,$30,$DB,$E3,$5E,$8F,$0D,$C7
	.dc.b	$F9,$7B,$46,$32,$FF,$00,$B5,$FE,$41,$3F,$9D,$DB,$F9,$67,$D8,$7C
	.dc.b	$7F,$14,$D3,$B4,$FC,$6F,$84,$FF,$6B,$DF,$F8,$31,$AF,$F8,$31,$AF
	.dc.b	$F8,$E9,$AF,$F5,$16,$AC,$7F,$CF,$DF,$11,$E9,$9C,$C6,$27,$FC,$91
	.dc.b	$8B,$C9,$F5,$6F,$FD,$4F,$1F,$BD,$18,$DF,$B6,$5F,$73,$EC,$1F,$61
	.dc.b	$F1,$5B,$97,$F8,$91,$AF,$F8,$91,$AF,$F9,$49,$AF,$F6,$96,$BE,$96
	.dc.b	$E6,$73,$9F,$FB,$53,$C4,$FC,$93,$FC,$A6,$BF,$D7,$CD,$7F,$C2,$D3
	.dc.b	$E2,$D3,$FC,$A5,$E2,$FE,$92,$7E,$33,$70,$FF,$18,$F4,$7D,$F8,$C1
	.dc.b	$1C,$12,$7E,$4F,$D8,$7D,$6B,$FC,$6C,$7F,$BA,$23,$1A,$FF,$85,$1B
	.dc.b	$1E,$0A,$FD,$2F,$D8,$7F,$B2,$23,$FC,$96,$D7,$FC,$96,$D7,$FC,$F2
	.dc.b	$D7,$FC,$20,$D7,$F0,$0D,$C4,$D3,$59,$FF,$B2,$BD,$3F,$DD,$1F,$EC
	.dc.b	$CA,$FF,$73,$35,$FF,$1F,$4F,$8F,$2F,$F3,$D7,$AA,$F8,$3F,$D2,$36
	.dc.b	$BF,$CC,$6D,$7F,$CC,$6D,$7F,$CD,$D1,$C0,$8F,$07,$F5,$1E,$C7,$F9
	.dc.b	$8A,$3F,$DF,$0D,$7F,$DC,$AD,$AF,$1C,$FF,$0E,$F8,$D7,$C8,$FE,$DD
	.dc.b	$B7,$FE,$76,$6B,$FE,$76,$6B,$FE,$A4,$6B,$FE,$3B,$6B,$FA,$B6,$E8
	.dc.b	$73,$A3,$EB,$DF,$61,$FC,$95,$B7,$FE,$DF,$CB,$F0,$B5,$BA,$F4,$63
	.dc.b	$FC,$9B,$E6,$BF,$DD,$D7,$FE,$AE,$FA,$9F,$E5,$DF,$02,$75,$F1,$BB
	.dc.b	$FB,$0F,$E2,$96,$69,$2F,$F7,$BF,$40,$F9,$1F,$E7,$0D,$8F,$27,$47
	.dc.b	$F9,$A2,$31,$0D,$BC,$F1,$FC,$9A,$3F,$DF,$1F,$60,$F5,$BF,$BC,$36
	.dc.b	$3F,$BE,$35,$FF,$6A,$35,$FF,$74,$B5,$FF,$77,$35,$EB,$1D,$0F,$F0
	.dc.b	$37,$A3,$FE,$FC,$C3,$FB,$3F,$C7,$FE,$BA,$BF,$E0,$03,$71,$C6,$1E
	.dc.b	$17,$E7,$FD,$7F,$F1,$9A,$71,$A3,$0D,$F1,$3F,$AE,$7D,$87,$F4,$9A
	.dc.b	$F8,$A7,$0A,$D9,$27,$7E,$57,$F6,$3F,$88,$7B,$94,$97,$CB,$14,$FF
	.dc.b	$02,$6C,$EA,$7E,$86,$BF,$E9,$AA,$E7,$1C,$5F,$E1,$3B,$7D,$BF,$3B
	.dc.b	$FC,$13,$C5,$FE,$48,$AF,$3C,$E5,$BF,$60,$F9,$CD,$DC,$9D,$DF,$17
	.dc.b	$C2,$7F,$9A,$2B,$83,$A9,$FF,$00,$7D,$83,$E7,$29,$CE,$78,$EA,$F9
	.dc.b	$9F,$C4,$37,$1F,$F0,$44,$7E,$3B,$8F,$F4,$6E,$89,$1A,$37,$E9,$FF
	.dc.b	$83,$FE,$97,$F7,$35,$FA,$A5,$F4,$19,$35,$E0,$4C,$BE,$84,$E9,$B3
	.dc.b	$97,$DE,$3E,$77,$F5,$CF,$F9,$55,$A8,$85,$72,$E0,$E5,$E0,$7F,$96
	.dc.b	$E3,$31,$C1,$AF,$18,$D1,$F9,$EF,$71,$F0,$9F,$E3,$62,$4F,$F9,$A6
	.dc.b	$BC,$79,$7F,$0A,$FB,$0F,$AB,$3F,$41,$B1,$9E,$4F,$F1,$B4,$7F,$05
	.dc.b	$9F,$C9,$3A,$E7,$FF,$73,$B6,$3F,$4A,$DE,$13,$C6,$3F,$D2,$7E,$AF
	.dc.b	$E4,$3E,$B2,$47,$E4,$34,$FF,$05,$36,$72,$FE,$33,$A3,$F8,$95,$7F
	.dc.b	$94,$A7,$FC,$CB,$F4,$2F,$B0,$F8,$AD,$E0,$FF,$43,$37,$0F,$9C,$1C
	.dc.b	$3F,$45,$AF,$D6,$6B,$F6,$5A,$F1,$9F,$A7,$87,$F9,$BF,$E6,$3F,$CE
	.dc.b	$7F,$06,$69,$E6,$9F,$A3,$B7,$E0,$F6,$7D,$1B,$EC,$3F,$CF,$7C,$FF
	.dc.b	$E1,$1E,$3F,$F0,$1E,$7F,$D5,$3D,$AF,$F3,$45,$EF,$5F,$E0,$8D,$7E
	.dc.b	$35,$F1,$CB,$F2,$17,$C6,$77,$8F,$F9,$C7,$E9,$5F,$41,$FE,$8A,$FB
	.dc.b	$0F,$F3,$CE,$1C,$1B,$7F,$C8,$2B,$6F,$F1,$36,$BF,$56,$37,$81,$E3
	.dc.b	$AA,$B7,$AF,$F8,$7B,$D4,$7D,$3F,$FA,$E5,$B8,$FF,$99,$E3,$07,$FA
	.dc.b	$F7,$17,$D8,$3E,$BF,$FD,$42,$DA,$FE,$32,$D7,$F8,$9D,$F4,$73,$F1
	.dc.b	$BC,$08,$F0,$AA,$FE,$3E,$BF,$D5,$2D,$9E,$F3,$FE,$94,$C3,$F8,$D4
	.dc.b	$FE,$32,$FD,$7E,$27,$A4,$91,$FE,$66,$FA,$AF,$F5,$43,$FC,$ED,$AF
	.dc.b	$F2,$06,$E7,$F1,$6F,$AC,$FE,$87,$EC,$1D,$9F,$D7,$1B,$1F,$9C,$D5
	.dc.b	$18,$8F,$E1,$91,$89,$FF,$46,$61,$FE,$CA,$9F,$07,$C7,$FF,$AC,$E3
	.dc.b	$18,$FE,$43,$0E,$4F,$F8,$42,$33,$C5,$F9,$96,$DF,$F6,$46,$E3,$DE
	.dc.b	$3D,$D8,$75,$23,$FD,$3B,$0E,$B7,$F9,$6F,$4F,$FC,$91,$3F,$C9,$AF
	.dc.b	$FC,$0D,$F0,$1F,$DE,$A7,$33,$B8,$F5,$C7,$A4,$91,$FD,$69,$FF,$29
	.dc.b	$65,$FD,$FF,$C6,$7D,$7F,$FA,$C7,$E5,$3F,$33,$18,$F4,$BF,$9B,$46
	.dc.b	$27,$5D,$3F,$CE,$78,$DF,$61,$F8,$DA,$D5,$F0,$AF,$E7,$B8,$73,$CD
	.dc.b	$5A,$6C,$F3,$BE,$4F,$0C,$7B,$CF,$F4,$7B,$E5,$7C,$4F,$D8,$E3,$E2
	.dc.b	$E5,$FE,$97,$19,$C3,$96,$5E,$3C,$39,$63,$D3,$0E,$11,$91,$8F,$6B
	.dc.b	$E3,$1E,$63,$8D,$F3,$AE,$31,$93,$8D,$FA,$1B,$3D,$6F,$A4,$7F,$B2
	.dc.b	$F2,$F7,$BD,$0E,$51,$A7,$89,$E1,$BE,$F8,$FB,$63,$83,$63,$38,$FA
	.dc.b	$23,$53,$C9,$F8,$F7,$E0,$8C,$4E,$C8,$F9,$72,$F9,$71,$88,$D9,$E0
	.dc.b	$7E,$C4,$6A,$30,$43,$77,$E4,$7D,$C7,$F7,$DC,$3F,$08,$DC,$BF,$E0
	.dc.b	$03,$F4,$0D,$7D,$87,$F6,$26,$F8,$BF,$D6,$1F,$AF,$19,$7F,$1E,$D7
	.dc.b	$F1,$0F,$F6,$AC,$FF,$3A,$DC,$BF,$35,$EC,$FF,$AC,$78,$BE,$B4,$E6
	.dc.b	$F8,$E1,$83,$F0,$15,$B8,$FA,$F3,$B7,$C7,$3B,$7E,$C4,$EF,$0F,$0C
	.dc.b	$E5,$B3,$AF,$81,$FF,$6F,$C6,$09,$8F,$B2,$BC,$33,$A8,$C5,$78,$A7
	.dc.b	$A4,$7B,$27,$53,$EA,$34,$4A,$FD,$4D,$F9,$6B,$9C,$78,$27,$83,$AA
	.dc.b	$E7,$1E,$E3,$94,$F0,$3C,$18,$F8,$A3,$8C,$65,$68,$C3,$C1,$1A,$21
	.dc.b	$D7,$0F,$05,$F3,$8C,$5F,$3B,$E5,$87,$93,$46,$7F,$67,$B5,$CE,$F8
	.dc.b	$76,$7F,$D3,$DB,$9F,$F8,$AF,$93,$FD,$17,$39,$3F,$94,$7A,$BF,$EA
	.dc.b	$59,$FE,$68,$D5,$CF,$FE,$B9,$8C,$36,$7F,$AF,$7D,$87,$6E,$7F,$E9
	.dc.b	$0D,$8F,$E7,$DF,$61,$70,$FE,$91,$37,$4F,$4B,$74,$3B,$C7,$FD,$91
	.dc.b	$79,$89,$87,$F6,$B7,$59,$A7,$38,$FB,$DF,$3C,$66,$BF,$B4,$36,$7F
	.dc.b	$B3,$B5,$FF,$3B,$E3,$F3,$3E,$C3,$C8,$7A,$78,$7F,$C3,$3E,$2E,$30
	.dc.b	$FF,$5F,$C6,$1F,$F0,$27,$7A,$7E,$54,$E7,$0F,$F4,$03,$5E,$FA,$BE
	.dc.b	$03,$E1,$1F,$0F,$EA,$2E,$3F,$E0,$98,$B5,$FE,$09,$9B,$C1,$FD,$DD
	.dc.b	$E1,$5E,$4C,$9E,$06,$2F,$FC,$4B,$18,$9F,$3C,$F9,$A3,$13,$E6,$6B
	.dc.b	$F1,$3C,$5F,$F6,$25,$7F,$B2,$3D,$47,$D2,$FD,$B9,$CD,$7F,$9E,$67
	.dc.b	$35,$FE,$49,$F4,$FE,$E4,$62,$7F,$D8,$35,$C3,$3F,$F3,$0C,$E5,$FC
	.dc.b	$CF,$0F,$F9,$A3,$1F,$C2,$60,$FA,$8F,$C5,$79,$9B,$CF,$FC,$C3,$39
	.dc.b	$7F,$CC,$F5,$C2,$B6,$D8,$FF,$49,$71,$7E,$3F,$D3,$FB,$11,$FE,$FA
	.dc.b	$F4,$9F,$20,$F8,$9D,$5F,$E6,$E7,$31,$FA,$76,$D7,$FA,$41,$BC,$1F
	.dc.b	$E8,$FB,$E5,$39,$6C,$66,$B9,$46,$23,$FD,$28,$D7,$FC,$FB,$87,$FB
	.dc.b	$AE,$BF,$DF,$3F,$08,$F8,$BF,$57,$9D,$FA,$98,$C4,$FF,$AB,$E3,$13
	.dc.b	$A8,$E2,$FE,$CF,$C4,$FC,$0F,$17,$D0,$6C,$FF,$A6,$57,$66,$F9,$E4
	.dc.b	$FF,$B7,$EB,$84,$73,$9C,$E7,$FE,$BB,$CD,$D9,$F3,$CF,$E0,$91,$FF
	.dc.b	$00,$76,$C6,$3C,$CF,$F7,$4F,$BB,$F0,$67,$FD,$D8,$D7,$84,$7F,$67
	.dc.b	$8C,$12,$7F,$8D,$4E,$8F,$F6,$1C,$73,$E6,$FF,$7D,$70,$7F,$13,$F6
	.dc.b	$9F,$03,$63,$0D,$C7,$C9,$82,$D7,$C2,$73,$1D,$1B,$19,$96,$DF,$9C
	.dc.b	$F7,$1F,$61,$FC,$8D,$7C,$75,$8C,$37,$0E,$B3,$A6,$CE,$9F,$C1,$FC
	.dc.b	$97,$F2,$DD,$5E,$9F,$E0,$E3,$0D,$9D,$9E,$CF,$1D,$F0,$2F,$0F,$5A
	.dc.b	$FF,$BC,$FA,$98,$E3,$6F,$F6,$BD,$EF,$1E,$77,$EE,$7F,$DC,$1F,$60
	.dc.b	$F8,$BF,$E8,$0D,$8F,$8E,$BE,$C3,$33,$27,$2F,$08,$C5,$75,$3E,$97
	.dc.b	$D8,$1F,$E9,$8D,$AF,$A6,$BD,$4C,$CC,$9C,$BC,$23,$0F,$BF,$D2,$FD
	.dc.b	$51,$F9,$6F,$B0,$FF,$74,$D7,$EC,$23,$1D,$FF,$8E,$9C,$D7,$8C,$8E
	.dc.b	$1C,$07,$0E,$A3,$91,$2B,$FC,$85,$3A,$9F,$C8,$7D,$83,$EC,$3F,$B5
	.dc.b	$36,$BF,$22,$BE,$93,$33,$FD,$67,$EC,$1F,$27,$FD,$51,$B7,$FE,$DB
	.dc.b	$6E,$1F,$D5,$9A,$FA,$FD,$24,$7D,$F8,$FA,$5D,$DD,$3E,$1F,$B7,$47
	.dc.b	$F7,$DC,$51,$7C,$C9,$97,$F6,$8F,$B0,$FF,$09,$E3,$CA,$FF,$9C,$3F
	.dc.b	$EF,$A8,$CB,$E1,$6B,$FE,$B5,$6B,$FE,$30,$6E,$66,$93,$D7,$FE,$E6
	.dc.b	$F0,$CD,$F2,$EF,$B0,$FF,$0A,$37,$2F,$EF,$BE,$67,$F4,$DC,$BF,$80
	.dc.b	$4E,$5F,$C9,$65,$FC,$C7,$D0,$FE,$FF,$5F,$D7,$67,$E4,$35,$FF,$66
	.dc.b	$35,$FF,$23,$B6,$79,$5F,$E1,$A7,$35,$D0,$FF,$1A,$63,$FB,$CF,$8E
	.dc.b	$FF,$15,$78,$47,$FC,$2D,$F6,$1F,$E4,$3C,$3F,$89,$CE,$5F,$F0,$BC
	.dc.b	$67,$0F,$F0,$9B,$5F,$BC,$D7,$FD,$C4,$D7,$FC,$BA,$DC,$4D,$27,$B9
	.dc.b	$E6,$FB,$07,$C2,$FF,$8F,$DB,$87,$F8,$A3,$EC,$3F,$6D,$8F,$E2,$A7
	.dc.b	$2F,$CC,$8C,$BF,$91,$6B,$FE,$F5,$6B,$FE,$70,$6E,$26,$92,$3F,$C6
	.dc.b	$6F,$0D,$1F,$61,$FE,$11,$C3,$F8,$54,$63,$D8,$FF,$20,$7D,$83,$EB
	.dc.b	$3F,$CB,$17,$BC,$7F,$CF,$F3,$9C,$3E,$C6,$48,$C4,$FF,$90,$1A,$FE
	.dc.b	$AD,$AF,$F1,$98,$E1,$7F,$E4,$89,$CC,$F6,$8C,$1F,$82,$F0,$BA,$F9
	.dc.b	$1F,$E2,$66,$E1,$CF,$EC,$2D,$FF,$57,$9B,$8A,$7A,$B7,$0F,$F2,$BC
	.dc.b	$60,$8E,$09,$7E,$E8,$F2,$E8,$FA,$5F,$DF,$CF,$F6,$26,$E3,$FA,$6F
	.dc.b	$A6,$FE,$B0,$FF,$08,$8C,$E7,$FE,$96,$C9,$5A,$AF,$E2,$0D,$7E,$33
	.dc.b	$5F,$F5,$7B,$70,$7D,$7F,$F2,$1C,$7F,$B9,$B6,$BF,$A4,$79,$EF,$AF
	.dc.b	$7D,$87,$FA,$4E,$B5,$9F,$F9,$41,$B8,$7E,$11,$B8,$7F,$A9,$2F,$8C
	.dc.b	$F6,$F1,$BF,$D3,$97,$FD,$FC,$FF,$2C,$96,$3E,$F7,$D5,$7F,$0B,$CF
	.dc.b	$FC,$07,$EC,$70,$24,$FF,$A3,$27,$55,$C7,$1F,$F5,$9F,$B9,$FB,$98
	.dc.b	$FF,$5E,$EA,$8C,$5F,$F6,$3E,$97,$CB,$BE,$77,$FC,$BD,$D9,$2B,$FC
	.dc.b	$01,$7C,$32,$FD,$BD,$EE,$73,$5F,$EB,$9C,$39,$E1,$FE,$C5,$6C,$FF
	.dc.b	$17,$6D,$FF,$5E,$6B,$FD,$8F,$1F,$37,$BF,$FE,$6F,$3A,$64,$F6,$B1
	.dc.b	$87,$92,$6D,$FF,$7F,$C3,$DD,$58,$D3,$FD,$AD,$F1,$DF,$C9,$EB,$FB
	.dc.b	$83,$5F,$F0,$23,$53,$BF,$B0,$FD,$E7,$D6,$DE,$B3,$49,$5F,$CC,$E5
	.dc.b	$F6,$1F,$E3,$A8,$C3,$5F,$EF,$B1,$AC,$3B,$4E,$67,$FC,$2D,$B3,$F3
	.dc.b	$0F,$F3,$1C,$7F,$3F,$8C,$7C,$27,$AD,$B1,$99,$DB,$FE,$0A,$8C,$B6
	.dc.b	$32,$FC,$F6,$BE,$06,$BF,$CD,$3F,$38,$FF,$1D,$BF,$65,$AF,$D9,$6B
	.dc.b	$FE,$1B,$D9,$18,$8E,$9C,$7E,$7A,$34,$91,$E9,$6B,$E7,$6B,$FE,$F0
	.dc.b	$D3,$FB,$93,$FD,$87,$9F,$E9,$F9,$9F,$75,$B1,$99,$DB,$FE,$2C,$8C
	.dc.b	$B6,$32,$F5,$9E,$51,$88,$FF,$24,$2F,$D9,$1F,$6F,$F8,$B8,$C3,$63
	.dc.b	$33,$FD,$D6,$72,$D8,$CB,$FE,$3E,$8C,$B6,$32,$FF,$91,$63,$2F,$F1
	.dc.b	$EF,$BC,$FC,$D7,$83,$FC,$89,$BE,$09,$A7,$C3,$FC,$57,$9D,$7F,$58
	.dc.b	$6B,$FE,$95,$6E,$67,$37,$FE,$7A,$F3,$7B,$FC,$17,$F8,$26,$FF,$AC
	.dc.b	$B5,$FF,$2B,$CE,$DC,$CF,$02,$38,$3E,$C7,$E1,$9F,$EB,$0C,$0D,$27
	.dc.b	$67,$BB,$EC,$9F,$5F,$49,$7F,$DA,$A3,$2D,$8C,$BE,$48,$CB,$63,$F0
	.dc.b	$DC,$CF,$98,$A4,$9F,$C6,$46,$5F,$EE,$FF,$82,$FF,$07,$D0,$D2,$47
	.dc.b	$D4,$6B,$C0,$C1,$F9,$5F,$69,$F6,$D4,$91,$F9,$B8,$CB,$FE,$06,$FB
	.dc.b	$37,$D2,$7E,$01,$B1,$98,$C1,$1C,$10,$FA,$B1,$C1,$FD,$17,$AB,$FE
	.dc.b	$CA,$7F,$54,$D7,$3F,$1D,$FA,$B8,$C4,$7F,$95,$BE,$73,$F2,$3E,$D9
	.dc.b	$A4,$8F,$F0,$EC,$65,$B1,$97,$F0,$F1,$97,$F6,$9F,$45,$FE,$BD,$7F
	.dc.b	$78,$D7,$3F,$15,$FB,$D8,$C4,$7F,$A9,$BE,$AD,$F2,$BF,$CE,$DB,$1F
	.dc.b	$CF,$DA,$FF,$99,$DA,$F1,$8E,$04,$78,$3C,$FE,$DE,$47,$36,$BC,$FE
	.dc.b	$1B,$C0,$47,$C0,$F8,$FC,$D9,$1E,$36,$BF,$E7,$9F,$94,$7D,$67,$F5
	.dc.b	$36,$C7,$F5,$76,$BB,$8D,$7B,$F2,$34,$D7,$A3,$5F,$F6,$5C,$A3,$11
	.dc.b	$FD,$3A,$7A,$BF,$DC,$5A,$E5,$FE,$9C,$61,$76,$BE,$C1,$CB,$FC,$C5
	.dc.b	$B3,$FC,$D5,$AF,$F3,$16,$BF,$EE,$19,$E5,$3C,$49,$1C,$5F,$E7,$9F
	.dc.b	$0F,$FC,$3E,$38,$35,$FF,$39,$B5,$F2,$B6,$4D,$27,$81,$D0,$FF,$02
	.dc.b	$7C,$D3,$F1,$A7,$26,$1B,$87,$FA,$BE,$75,$39,$24,$66,$76,$FF,$BE
	.dc.b	$A7,$6E,$67,$E5,$F4,$C8,$FE,$F9,$A7,$ED,$5F,$A5,$1F,$D6,$FE,$5D
	.dc.b	$F6,$F4,$93,$FD,$86,$32,$D8,$CB,$FD,$7E,$32,$D8,$FE,$CD,$F4,$BF
	.dc.b	$C3,$A3,$FB,$23,$5F,$A8,$D7,$FD,$3E,$D7,$F0,$2D,$D4,$D2,$76,$FE
	.dc.b	$6F,$ED,$65,$7E,$6D,$AF,$E8,$5A,$FE,$59,$AF,$F8,$8B,$F0,$1F,$88
	.dc.b	$DC,$BF,$C4,$11,$88,$78,$48,$C6,$2E,$A7,$CE,$7F,$81,$5B,$7F,$E0
	.dc.b	$96,$BF,$E0,$56,$BF,$95,$6B,$FE,$C0,$6B,$FA,$26,$CF,$79,$FD,$3F
	.dc.b	$CF,$4A,$FB,$AD,$7F,$C6,$9B,$A3,$11,$FB,$16,$A1,$C2,$FC,$61,$F8
	.dc.b	$2F,$E6,$1B,$3F,$95,$6B,$F9,$16,$BC,$CE,$6E,$1E,$D3,$D6,$7A,$12
	.dc.b	$3F,$CA,$19,$FC,$0D,$3E,$76,$5F,$48,$93,$FE,$33,$FA,$3F,$D0,$BF
	.dc.b	$BD,$6B,$FE,$E8,$6B,$FC,$29,$BA,$1A,$4E,$4F,$F1,$E3,$AF,$A5,$7E
	.dc.b	$29,$FE,$3F,$6D,$FF,$90,$DA,$FF,$8F,$DA,$FF,$0F,$6B,$FE,$FA,$6B
	.dc.b	$FC,$71,$B6,$73,$E4,$BE,$9D,$F2,$FF,$E3,$26,$DF,$FB,$F1,$AF,$F8
	.dc.b	$91,$AF,$87,$EB,$7D,$04,$7D,$0F,$3F,$A6,$91,$CD,$AF,$36,$BF,$6E
	.dc.b	$FC,$91,$88,$FF,$64,$FA,$3F,$E4,$6F,$89,$34,$91,$FE,$8B,$FB,$07
	.dc.b	$CD,$FF,$A2,$DB,$19,$8C,$1F,$E8,$FF,$88,$7D,$BD,$24,$FF,$A5,$63
	.dc.b	$2F,$F9,$D7,$EB,$19,$7F,$9D,$5B,$19,$8C,$11,$C1,$0F,$F3,$C4,$70
	.dc.b	$6C,$6D,$A6,$CF,$F6,$57,$EB,$EF,$A3,$FF,$53,$36,$3F,$D5,$4D,$7F
	.dc.b	$D4,$CD,$7F,$AA,$57,$FC,$00,$7D,$5F,$29,$E9,$23,$FE,$CB,$F0,$DF
	.dc.b	$60,$F8,$8F,$F6,$5E,$3F,$AF,$9C,$C6,$2B,$F7,$91,$87,$FD,$9F,$3F
	.dc.b	$BF,$E1,$7A,$6E,$3F,$F6,$BC,$F3,$C7,$B5,$F0,$39,$B8,$F8,$EF,$66
	.dc.b	$3D,$5F,$81,$8F,$4C,$39,$DF,$0C,$BC,$8F,$F3,$AC,$39,$63,$DE,$7D
	.dc.b	$17,$C6,$B7,$8F,$5A,$D4,$78,$CE,$55,$C3,$2F,$54,$70,$C3,$AB,$E1
	.dc.b	$38,$5F,$35,$87,$77,$33,$C1,$EA,$62,$FB,$63,$EC,$8C,$E5,$F0,$AB
	.dc.b	$86,$1D,$70,$E9,$39,$C3,$C8,$4B,$F9,$93,$AC,$BD,$B5,$B9,$F0,$1A
	.dc.b	$CB,$DA,$E2,$34,$49,$FB,$93,$A8,$C4,$FC,$16,$D7,$78,$C5,$78,$08
	.dc.b	$FC,$4A,$E5,$1F,$42,$3F,$9E,$1F,$10,$B8,$7E,$35,$B8,$FE,$21,$CF
	.dc.b	$D7,$BE,$67,$F3,$4D,$8E,$67,$F7,$C7,$0F,$08,$D9,$F8,$28,$E1,$97
	.dc.b	$E3,$BE,$C1,$F4,$FF,$C4,$2F,$EE,$D6,$B2,$FD,$23,$63,$F4,$6D,$D1
	.dc.b	$7C,$B9,$FB,$65,$2F,$9B,$F9,$49,$D4,$60,$99,$7E,$8B,$0E,$93,$A8
	.dc.b	$C5,$73,$6D,$7E,$02,$75,$8F,$9A,$B4,$D9,$D6,$5E,$DB,$E2,$49,$D3
	.dc.b	$F0,$E3,$51,$82,$4F,$C0,$AF,$54,$EA,$30,$7A,$A7,$79,$7C,$49,$D4
	.dc.b	$FD,$5A,$F2,$BC,$2F,$C1,$96,$23,$47,$33,$F5,$B8,$FB,$9E,$18,$F9
	.dc.b	$1D,$D7,$27,$87,$63,$89,$F5,$DF,$76,$7F,$93,$B5,$FF,$0C,$4E,$E3
	.dc.b	$11,$FE,$88,$9E,$38,$FF,$32,$F9,$8C,$F8,$6E,$64,$61,$B3,$B3,$FA
	.dc.b	$67,$D7,$F4,$CB,$C8,$DC,$0D,$24,$F0,$26,$3F,$D3,$23,$2D,$8C,$BF
	.dc.b	$D0,$A3,$C5,$E4,$C9,$FE,$92,$D9,$FE,$AB,$E7,$78,$49,$3F,$D8,$5F
	.dc.b	$83,$C5,$FE,$09,$F8,$CB,$E0,$9A,$4A,$FF,$05,$6F,$A2,$38,$26,$1F
	.dc.b	$E0,$D8,$CB,$63,$FB,$87,$85,$F0,$C8,$E0,$FF,$0B,$7B,$5E,$22,$38
	.dc.b	$3F,$C5,$5E,$87,$CD,$F1,$4D,$24,$68,$97,$FE,$2F,$8C,$BF,$C5,$38
	.dc.b	$3F,$C7,$9F,$31,$F5,$FC,$83,$49,$7F,$E4,$96,$D7,$19,$DC,$E4,$FE
	.dc.b	$2B,$CA,$FC,$15,$F6,$BF,$8A,$6C,$E9,$B7,$FC,$9A,$B8,$57,$E5,$3C
	.dc.b	$FF,$F1,$E7,$A9,$78,$BF,$D5,$F8,$FF,$AF,$DB,$87,$FA,$ED,$AF,$F9
	.dc.b	$AE,$B8,$CE,$EB,$91,$1F,$3F,$C9,$BE,$6F,$F3,$ED,$8C,$C6,$08,$E0
	.dc.b	$97,$E9,$9E,$0F,$CD,$ED,$7C,$07,$E8,$DB,$5F,$CE,$B8,$3F,$C4,$53
	.dc.b	$92,$46,$67,$75,$CE,$72,$FC,$7D,$9F,$19,$F3,$49,$7F,$35,$FD,$EC
	.dc.b	$BD,$97,$C3,$61,$B5,$FD,$32,$75,$39,$24,$66,$77,$3F,$D4,$9A,$FF
	.dc.b	$5E,$6B,$FC,$3B,$E0,$7F,$D4,$73,$EF,$6C,$7E,$45,$AF,$FA,$87,$34
	.dc.b	$62,$39,$7A,$FF,$C2,$E3,$10,$9D,$CE,$4F,$E2,$1F,$04,$F8,$2A,$4C
	.dc.b	$3E,$84,$65,$B1,$9B,$FD,$24,$E5,$B1,$F8,$4F,$22,$57,$E2,$1B,$3F
	.dc.b	$DF,$B8,$DF,$56,$FA,$DF,$CD,$36,$FF,$BC,$CE,$A7,$24,$8C,$CE,$EB
	.dc.b	$A3,$5F,$CE,$35,$FE,$77,$F5,$DF,$ED,$79,$FC,$63,$63,$F7,$CD,$7E
	.dc.b	$73,$5E,$3E,$27,$F3,$78,$C4,$7C,$D6,$EA,$6F,$C2,$FF,$A5,$2F,$F3
	.dc.b	$4D,$AF,$E1,$8D,$7E,$E3,$5F,$EA,$2D,$7E,$D3,$68,$D3,$5D,$6E,$77
	.dc.b	$9B,$FB,$62,$61,$FB,$A7,$F3,$9E,$E4,$8F,$CC,$37,$1F,$CC,$35,$FC
	.dc.b	$C3,$5F,$F7,$97,$27,$C2,$7F,$51,$A7,$E9,$C9,$5F,$A3,$FA,$CF,$D5
	.dc.b	$11,$C1,$FA,$7F,$73,$F5,$C4,$70,$7E,$AF,$AF,$F6,$44,$70,$7E,$BF
	.dc.b	$E4,$BF,$6C,$47,$04,$C3,$F6,$51,$96,$C7,$F0,$DF,$51,$F0,$7F,$DC
	.dc.b	$DB,$19,$8C,$1F,$D8,$BD,$F7,$C7,$D2,$4F,$F9,$43,$CD,$4E,$A3,$0D
	.dc.b	$C3,$FC,$BF,$3A,$AE,$31,$89,$D3,$FB,$FF,$C0,$FF,$DE,$F7,$FC,$8D
	.dc.b	$B1,$FD,$55,$AE,$FC,$CF,$EE,$FF,$50,$F8,$BA,$4B,$F3,$46,$5B,$19
	.dc.b	$AF,$2C,$E5,$B1,$FE,$17,$F8,$BF,$EB,$15,$FC,$D5,$B3,$FD,$9D,$AF
	.dc.b	$F1,$76,$BF,$EC,$5F,$90,$7C,$CF,$F8,$31,$B5,$F2,$5B,$3F,$19,$A8
	.dc.b	$E5,$3C,$49,$3C,$5F,$F1,$67,$4F,$F8,$DA,$7F,$BD,$35,$FE,$54,$D7
	.dc.b	$FC,$7E,$D7,$FD,$73,$F4,$DF,$E3,$87,$EB,$B5,$F1,$1A,$F6,$7F,$C4
	.dc.b	$B1,$98,$FF,$71,$35,$D5,$FE,$16,$31,$C0,$75,$DA,$F9,$5F,$F4,$1D
	.dc.b	$FF,$3F,$6B,$FE,$54,$6B,$FC,$F5,$B4,$FA,$67,$DF,$FF,$8E,$1B,$8F
	.dc.b	$E5,$5B,$7F,$92,$6B,$EC,$AE,$55,$C4,$95,$C5,$FF,$2C,$3F,$C3,$8F
	.dc.b	$EE,$C5,$9F,$F1,$43,$5F,$EB,$0D,$7E,$D4,$BE,$67,$FA,$54,$78,$5A
	.dc.b	$E9,$AF,$FB,$B1,$BF,$34,$DF,$F9,$33,$7E,$2B,$FC,$EE,$DC,$3F,$D8
	.dc.b	$6D,$BF,$F6,$7B,$5F,$B6,$D7,$F1,$CD,$4D,$D1,$F4,$0F,$B3,$FF,$3D
	.dc.b	$36,$FF,$C3,$0D,$BF,$F2,$53,$5F,$EF,$4D,$7F,$D1,$ED,$7F,$BC,$36
	.dc.b	$BB,$D7,$FA,$3B,$E7,$FF,$D0,$B7,$FE,$4A,$6D,$FF,$91,$5A,$FE,$5B
	.dc.b	$89,$18,$8F,$D5,$46,$3E,$03,$A1,$2B,$A3,$FE,$A0,$F8,$87,$E8,$1F
	.dc.b	$9B,$6F,$2E,$63,$07,$E7,$EF,$BF,$C5,$F9,$C9,$87,$9D,$B8,$F9,$DA
	.dc.b	$F9,$F2,$47,$F8,$D2,$3F,$80,$7C,$37,$F8,$C5,$A5,$70,$9F,$F1,$B7
	.dc.b	$D1,$BE,$93,$EC,$CF,$EE,$0F,$E1,$51,$9E,$1D,$CE,$A7,$53,$97,$80
	.dc.b	$5F,$49,$F8,$28,$CC,$63,$8B,$C3,$39,$9E,$6D,$7E,$8F,$D3,$BE,$97
	.dc.b	$F8,$5D,$35,$CC,$61,$B5,$FC,$86,$74,$B6,$FB,$B5,$B8,$C5,$F5,$78
	.dc.b	$5F,$27,$F3,$2D,$BE,$77,$CA,$76,$75,$DB,$F9,$BB,$F8,$FE,$DF,$F3
	.dc.b	$14,$FE,$31,$B3,$A9,$FF,$4B,$7E,$20,$F9,$FF,$25,$A3,$13,$DA,$31
	.dc.b	$3A,$AE,$58,$79,$2F,$77,$E8,$3E,$03,$88,$D5,$71,$C3,$E1,$73,$FE
	.dc.b	$E3,$EE,$2F,$09,$A6,$3E,$9E,$E5,$F0,$2B,$83,$6B,$83,$67,$AF,$6B
	.dc.b	$ED,$69,$27,$E4,$D3,$EA,$3F,$DC,$D1,$D6,$B8,$36,$B8,$3E,$DA,$E1
	.dc.b	$F1,$DE,$AF,$A8,$E3,$3F,$ED,$36,$BF,$E8,$36,$BF,$D9,$72,$46,$23
	.dc.b	$FD,$05,$F5,$9F,$98,$7D,$CD,$7E,$64,$A3,$1B,$FE,$3A,$39,$DF,$CC
	.dc.b	$F1,$7F,$D4,$F1,$FD,$5F,$93,$FD,$48,$DC,$3F,$D7,$13,$A2,$4E,$9B
	.dc.b	$1F,$EB,$DF,$95,$79,$1F,$61,$B1,$F4,$DA,$F9,$A7,$F4,$2F,$A3,$99
	.dc.b	$F5,$1E,$26,$C7,$FB,$25,$AF,$F8,$0A,$34,$47,$4D,$7F,$E0,$06,$2F
	.dc.b	$B0,$FC,$5B,$63,$F0,$CD,$7C,$53,$F9,$0C,$5F,$29,$F1,$1B,$3F,$E2
	.dc.b	$78,$C1,$1C,$10,$FF,$15,$3F,$33,$F0,$1F,$F7,$84,$7E,$55,$AF,$D8
	.dc.b	$6E,$DF,$5E,$BB,$CF,$DC,$FA,$9F,$F6,$DB,$FE,$EB,$6B,$FC,$8F,$14
	.dc.b	$62,$3F,$D9,$CD,$7F,$18,$D7,$FD,$98,$D8,$E3,$F1,$3F,$82,$8C,$7D
	.dc.b	$8B,$EA,$7F,$66,$DB,$FD,$73,$5F,$35,$7E,$E3,$D8,$7D,$57,$E3,$1B
	.dc.b	$3F,$E7,$78,$C1,$1C,$10,$F6,$47,$07,$DD,$F6,$5F,$82,$D3,$F3,$8D
	.dc.b	$E4,$FD,$43,$67,$FD,$D1,$5C,$6B,$F1,$FF,$52,$F9,$1F,$EA,$4D,$AF
	.dc.b	$E8,$A1,$C3,$FD,$21,$AF,$F5,$16,$BF,$E4,$A6,$BF,$E5,$D6,$BC,$E3
	.dc.b	$FB,$07,$E0,$0F,$A4,$FE,$7D,$28,$C3,$FC,$E0,$C4,$6A,$BC,$77,$EA
	.dc.b	$F8,$FC,$12,$B8,$4B,$E7,$3F,$A5,$75,$3F,$08,$FE,$0D,$5F,$E6,$08
	.dc.b	$C7,$97,$F0,$1B,$1F,$D9,$9A,$FF,$6F,$EB,$7D,$83,$2F,$E8,$6D,$9F
	.dc.b	$E8,$4D,$7F,$9F,$0F,$94,$7D,$67,$F4,$A2,$C7,$F7,$4B,$CC,$7F,$54
	.dc.b	$33,$1E,$16,$BF,$E2,$36,$BF,$E2,$26,$BF,$DA,$DB,$1D,$23,$FC,$03
	.dc.b	$D2,$F8,$5F,$ED,$91,$FC,$A1,$BB,$FF,$27,$6F,$2F,$5C,$3F,$A9,$FC
	.dc.b	$77,$1B,$FE,$9C,$D7,$F9,$8B,$6F,$F9,$83,$FC,$81,$B7,$C8,$93,$CB
	.dc.b	$B3,$FA,$FF,$5F,$F4,$C8,$FE,$76,$DA,$FE,$74,$DC,$7F,$A3,$2F,$AE
	.dc.b	$E3,$E3,$CB,$FE,$C5,$3B,$5A,$7E,$EE,$BE,$97,$C3,$C9,$FE,$F5,$AF
	.dc.b	$FB,$A9,$FF,$04,$FC,$9B,$EB,$BF,$DC,$71,$AA,$E0,$FF,$7B,$E0,$FB
	.dc.b	$B1,$FE,$19,$AE,$1F,$11,$FD,$EB,$EF,$FF,$C8,$77,$FD,$D5,$AF,$F5
	.dc.b	$86,$EF,$FE,$1A,$CB,$C5,$3F,$D7,$3E,$3F,$FD,$EE,$FF,$7D,$6B,$9F
	.dc.b	$8B,$FF,$03,$C3,$E7,$2C,$49,$E0,$EA,$3F,$C1,$BF,$50,$F4,$E9,$2B
	.dc.b	$FC,$15,$1F,$CC,$B1,$FE,$D4,$EE,$33,$3B,$7F,$B8,$C6,$F9,$3F,$CE
	.dc.b	$BF,$5D,$EA,$9F,$F3,$D3,$5C,$EF,$FE,$32,$8C,$47,$F9,$0D,$BC,$06
	.dc.b	$92,$7F,$8B,$F0,$3E,$FB,$E3,$E7,$FE,$4C,$8C,$4F,$D3,$9C,$E3,$E0
	.dc.b	$C7,$FC,$4D,$E0,$FF,$4A,$C3,$FC,$0C,$D9,$FF,$09,$36,$7F,$12,$BE
	.dc.b	$53,$FC,$6A,$E7,$9B,$F9,$47,$93,$83,$33,$FE,$3A,$F9,$17,$D3,$52
	.dc.b	$61,$F9,$0F,$95,$AF,$F8,$6A,$76,$FF,$8E,$DA,$FE,$3F,$99,$3F,$E5
	.dc.b	$4E,$0F,$24,$FF,$0B,$6B,$FC,$65,$AF,$D2,$C3,$E8,$11,$E0,$D7,$FD
	.dc.b	$1D,$C1,$F3,$7E,$6B,$E5,$12,$3F,$CC,$15,$DB,$7F,$F1,$77,$D1,$BF
	.dc.b	$19,$FF,$2C,$B6,$BF,$CA,$8D,$7F,$C6,$EF,$EB,$1B,$1C,$08,$F0,$EB
	.dc.b	$FF,$31,$FD,$3B,$E6,$69,$27,$FC,$C1,$0F,$99,$C4,$7F,$91,$23,$83
	.dc.b	$A9,$E7,$3D,$3E,$B3,$FC,$83,$46,$92,$BF,$D5,$3E,$FB,$E9,$5F,$5B
	.dc.b	$FE,$97,$8C,$36,$33,$3E,$4A,$FF,$3D,$FD,$57,$F5,$AE,$AF,$F2,$F3
	.dc.b	$67,$F8,$AB,$70,$E7,$1F,$E7,$B8,$D7,$49,$A6,$F0,$EF,$3F,$C4,$FE
	.dc.b	$BD,$F2,$BF,$E9,$E6,$C7,$F4,$F6,$BF,$D7,$5A,$FF,$BB,$DF,$EA,$3C
	.dc.b	$5F,$EA,$C6,$E5,$FE,$F0,$6C,$FF,$AE,$5E,$73,$FE,$6F,$C3,$C7,$23
	.dc.b	$DF,$A7,$F9,$BA,$33,$18,$3F,$AA,$BF,$C3,$5F,$F3,$DB,$67,$A7,$83
	.dc.b	$F2,$AB,$E5,$B2,$BE,$33,$96,$90,$E6,$E5,$EB,$F2,$1E,$32,$3E,$27
	.dc.b	$CD,$F0,$FF,$60,$91,$F4,$E3,$FC,$33,$E1,$BE,$B9,$F2,$9F,$EF,$EE
	.dc.b	$FF,$81,$6A,$D4,$E6,$3E,$24,$61,$FF,$80,$13,$F1,$F8,$9E,$DF,$E4
	.dc.b	$E7,$81,$CF,$0E,$EF,$03,$9B,$8F,$92,$5F,$8B,$0E,$D1,$8A,$61,$DB
	.dc.b	$97,$CF,$39,$9E,$D9,$78,$2F,$8E,$1C,$F1,$EC,$F9,$F2,$F3,$5F,$8C
	.dc.b	$E2,$5A,$E2,$F9,$AF,$E1,$65,$E1,$8E,$D8,$74,$74,$71,$71,$1B,$8C
	.dc.b	$4F,$02,$3F,$0E,$35,$8F,$AE,$FC,$33,$E4,$8C,$5F,$4C,$3D,$31,$8C
	.dc.b	$39,$B6,$BA,$4E,$B1,$F8,$55,$A6,$CE,$B2,$F9,$71,$8B,$E4,$47,$E4
	.dc.b	$46,$A7,$DF,$39,$6C,$E9,$EC,$60,$99,$7B,$A3,$C3,$1F,$90,$3E,$89
	.dc.b	$72,$FC,$73,$72,$FC,$4B,$9F,$A0,$7A,$FF,$E8,$2A,$46,$27,$A1,$F9
	.dc.b	$B7,$0F,$19,$E1,$1B,$C7,$E8,$37,$0F,$CE,$7D,$73,$E9,$5C,$DF,$8D
	.dc.b	$AD,$63,$FA,$46,$BD,$F1,$FC,$3D,$F1,$AE,$11,$27,$F1,$D5,$CB,$45
	.dc.b	$EE,$73,$5B,$CB,$CA,$F4,$C7,$C7,$3A,$24,$E9,$FC,$65,$73,$8C,$13
	.dc.b	$2F,$D7,$63,$E1,$9D,$46,$30,$F0,$36,$F9,$E1,$E0,$C3,$94,$7C,$1A
	.dc.b	$DB,$FB,$6B,$E6,$E4,$F2,$DF,$3A,$FD,$E6,$3F,$03,$0E,$93,$92,$57
	.dc.b	$1A,$DD,$F4,$8F,$9C,$74,$35,$3F,$62,$B9,$C6,$08,$70,$AE,$71,$A8
	.dc.b	$C4,$FF,$5E,$FA,$9F,$9B,$CB,$FD,$15,$BA,$7F,$62,$F1,$ED,$FF,$62
	.dc.b	$9B,$D7,$FD,$7E,$7E,$AC,$FF,$2C,$8C,$57,$F4,$89,$CC,$F3,$8F,$E9
	.dc.b	$7F,$15,$CB,$A4,$DF,$99,$C3,$27,$FB,$73,$96,$92,$BF,$A9,$C6,$5B
	.dc.b	$1F,$DD,$67,$C3,$1F,$DE,$9B,$87,$F7,$76,$D9,$BF,$69,$E5,$3C,$DF
	.dc.b	$13,$70,$FF,$00,$3C,$08,$FF,$66,$79,$FC,$B7,$B4,$F9,$3C,$5F,$E1
	.dc.b	$5D,$31,$5F,$E0,$99,$B9,$B6,$FC,$FD,$EE,$B4,$7F,$77,$8C,$F8,$1F
	.dc.b	$E5,$88,$C1,$FD,$55,$C5,$7F,$76,$FA,$3F,$0C,$62,$BF,$C0,$51,$F0
	.dc.b	$09,$5F,$DF,$9F,$8B,$E1,$3E,$23,$EF,$B6,$7F,$C7,$7F,$5D,$89,$91
	.dc.b	$FE,$3E,$77,$18,$C7,$FC,$97,$EB,$7E,$23,$CA,$37,$EB,$B0,$CC,$3F
	.dc.b	$CA,$6E,$5A,$7F,$B0,$3E,$7B,$FD,$EE,$DF,$4B,$F2,$BE,$BB,$A5,$F5
	.dc.b	$DF,$98,$F4,$98,$70,$D3,$FD,$89,$5A,$C7,$FD,$FC,$D8,$FE,$6D,$87
	.dc.b	$3C,$3F,$12,$DA,$DC,$62,$B8,$57,$F8,$32,$B4,$F2,$8C,$BF,$E2,$FE
	.dc.b	$AF,$F0,$0C,$7E,$AB,$B3,$D1,$87,$DA,$A5,$BE,$A1,$F5,$DB,$C2,$5F
	.dc.b	$FA,$45,$F8,$33,$A8,$C1,$FE,$95,$F8,$AE,$C7,$7F,$8D,$B1,$92,$63
	.dc.b	$FE,$9D,$8C,$BF,$93,$8F,$F2,$36,$DF,$2B,$8F,$E6,$6E,$7E,$8E,$4F
	.dc.b	$AE,$CC,$61,$F8,$C4,$BF,$9E,$DC,$3E,$33,$5F,$95,$5C,$E7,$FA,$EE
	.dc.b	$EF,$88,$FF,$4D,$47,$C7,$8C,$7A,$DF,$C5,$25,$F5,$D6,$25,$6E,$78
	.dc.b	$E1,$ED,$17,$D6,$F7,$19,$87,$C5,$FB,$EF,$C0,$D9,$D2,$FA,$EF,$E4
	.dc.b	$AE,$7B,$7F,$AC,$DF,$2D,$72,$6A,$AE,$1E,$BF,$83,$FE,$E2,$D3,$EE
	.dc.b	$F9,$1F,$CC,$23,$39,$7A,$79,$BF,$0D,$3A,$AF,$C5,$71,$3E,$B9,$EE
	.dc.b	$E7,$0E,$1C,$FF,$DC,$76,$FE,$DB,$97,$2C,$78,$E1,$C1,$FC,$A4,$66
	.dc.b	$FE,$27,$D7,$3E,$93,$FA,$1B,$6B,$EA,$3F,$EE,$26,$C6,$88,$EB,$E0
	.dc.b	$6D,$7C,$FF,$8E,$FD,$B6,$5F,$21,$B5,$FA,$97,$FD,$52,$FE,$05,$CC
	.dc.b	$FD,$A6,$F1,$1A,$70,$8E,$DD,$FF,$D6,$BE,$3B,$FA,$05,$FE,$2D,$B3
	.dc.b	$F1,$BD,$3D,$79,$DF,$AE,$7F,$81,$F0,$7F,$8C,$AF,$FC,$63,$BF,$ED
	.dc.b	$E7,$24,$8C,$CE,$EF,$F8,$6B,$5F,$E1,$0D,$7F,$D6,$94,$F8,$F7,$99
	.dc.b	$FC,$79,$B7,$FC,$1E,$9C,$DF,$3D,$FC,$24,$E9,$FC,$37,$E0,$AF,$73
	.dc.b	$F8,$DB,$6F,$D2,$D9,$FE,$42,$D7,$FC,$B1,$7F,$EF,$D9,$F0,$36,$3F
	.dc.b	$54,$78,$BE,$07,$FC,$0E,$F9,$3E,$0A,$F3,$FF,$A8,$2F,$D0,$D9,$E3
	.dc.b	$77,$93,$FC,$17,$36,$BF,$D6,$F3,$BA,$F1,$F8,$FF,$DF,$1F,$82,$D8
	.dc.b	$D4,$70,$8C,$12,$7F,$CC,$F1,$C1,$FE,$1B,$F2,$AF,$4F,$FC,$3A,$D8
	.dc.b	$FE,$1D,$EF,$3E,$2B,$35,$FE,$22,$8C,$4F,$F1,$66,$DF,$F9,$FF,$A7
	.dc.b	$FD,$8F,$7F,$E1,$09,$CC,$FF,$89,$DA,$FF,$45,$6B,$FC,$51,$B3,$CA
	.dc.b	$BF,$91,$FC,$B7,$F3,$FA,$FE,$2C,$DD,$7F,$87,$FD,$7B,$D4,$FF,$1B
	.dc.b	$CE,$6F,$FA,$94,$E5,$FE,$5A,$DB,$F4,$35,$FF,$1E,$35,$E3,$5F,$E2
	.dc.b	$A8,$C4,$7F,$2A,$6E,$06,$FC,$CB,$EC,$7F,$C9,$2D,$C3,$F9,$D7,$CA
	.dc.b	$FF,$62,$C3,$F9,$84,$E5,$FE,$65,$19,$7F,$C9,$CD,$7F,$B6,$35,$FE
	.dc.b	$76,$DC,$0D,$27,$99,$FE,$42,$75,$F6,$FF,$CD,$F0,$FE,$EC,$DB,$FE
	.dc.b	$8B,$D7,$FE,$29,$9C,$F9,$3F,$E6,$28,$FF,$17,$CF,$F9,$85,$AF,$F7
	.dc.b	$F6,$BC,$BD,$BF,$F2,$8C,$63,$2F,$EF,$F1,$82,$5F,$F6,$0F,$81,$7C
	.dc.b	$BF,$F9,$C1,$B1,$FD,$87,$E7,$BF,$9D,$3F,$E0,$D8,$CE,$3F,$D6,$67
	.dc.b	$FD,$5C,$FF,$9D,$5A,$FF,$84,$DA,$FF,$62,$6E,$06,$93,$CC,$FF,$34
	.dc.b	$BA,$F9,$FF,$E5,$E7,$F3,$0C,$3F,$8F,$72,$FF,$60,$8C,$78,$9F,$E7
	.dc.b	$09,$FE,$3D,$8F,$F7,$06,$E3,$FE,$D0,$9F,$F3,$AF,$DB,$FF,$0D,$C3
	.dc.b	$FB,$4B,$69,$6F,$A2,$7D,$67,$F3,$DB,$DF,$83,$FE,$96,$C3,$85,$FF
	.dc.b	$7E,$6D,$F0,$F9,$0D,$57,$FA,$72,$FF,$C1,$38,$FF,$83,$5B,$87,$F8
	.dc.b	$0D,$FF,$25,$C6,$7D,$4E,$04,$78,$7B,$16,$BF,$BB,$7D,$3F,$06,$C7
	.dc.b	$F9,$63,$EA,$3F,$A7,$E1,$FD,$CA,$B4,$D9,$FF,$2E,$C6,$2F,$FA,$2D
	.dc.b	$70,$7F,$A3,$C6,$6F,$FC,$93,$3F,$D6,$9F,$F2,$A3,$5E,$3E,$6F,$F9
	.dc.b	$72,$31,$1F,$E5,$0F,$A9,$FE,$40,$FF,$83,$5B,$8F,$F8,$1A,$5F,$21
	.dc.b	$FE,$70,$9C,$E3,$FD,$61,$B7,$CF,$C5,$F7,$56,$DF,$E8,$B1,$F6,$BE
	.dc.b	$2F,$F9,$9B,$4A,$E1,$3F,$CE,$3E,$AB,$FC,$21,$C3,$7D,$6F,$C1,$57
	.dc.b	$E1,$09,$3A,$70,$7E,$21,$C3,$87,$F5,$BA,$75,$EB,$FF,$59,$4F,$01
	.dc.b	$3A,$70,$D7,$FD,$01,$FA,$C3,$EB,$FF,$A6,$F7,$A3,$13,$FE,$06,$6B
	.dc.b	$FE,$8E,$6B,$CA,$78,$C6,$23,$FC,$06,$D7,$FD,$0A,$DF,$93,$7E,$DA
	.dc.b	$F8,$4F,$EB,$B1,$86,$C6,$6B,$D9,$39,$7C,$9D,$1E,$02,$38,$21,$E7
	.dc.b	$72,$D3,$FB,$5E,$6F,$AD,$FB,$6D,$8C,$C6,$08,$E0,$87,$CF,$8E,$0F
	.dc.b	$D8,$F8,$E7,$E1,$0F,$57,$FB,$CC,$61,$B1,$99,$FB,$13,$96,$C6,$5E
	.dc.b	$8D,$7F,$11,$1B,$73,$3D,$BE,$C7,$51,$FC,$42,$51,$88,$46,$27,$81
	.dc.b	$FE,$0E,$F0,$9F,$5A,$F9,$2C,$36,$FF,$D3,$AD,$7F,$C1,$0D,$7C,$BF
	.dc.b	$12,$FA,$2F,$82,$D8,$F7,$B5,$F9,$0D,$78,$C7,$02,$3C,$1F,$37,$90
	.dc.b	$F5,$29,$27,$FD,$63,$19,$7E,$6F,$DD,$3E,$BB,$E9,$B6,$33,$18,$23
	.dc.b	$82,$1F,$76,$38,$3F,$E2,$7E,$8F,$F2,$FB,$FE,$C0,$6B,$9F,$80,$F9
	.dc.b	$91,$88,$FB,$3E,$57,$F8,$7D,$FC,$0B,$5F,$C5,$35,$DF,$AF,$F8,$A8
	.dc.b	$C4,$7D,$D6,$BF,$71,$BC,$3F,$D3,$A7,$EE,$7A,$6F,$9C,$A4,$BF,$F6
	.dc.b	$BC,$65,$FC,$9F,$DD,$3E,$17,$F3,$2D,$8C,$C6,$08,$E0,$87,$E8,$E3
	.dc.b	$83,$FE,$4D,$FA,$EF,$F2,$A3,$FE,$E6,$6B,$9F,$77,$F2,$51,$88,$FC
	.dc.b	$FF,$D0,$8F,$D5,$95,$FD,$4B,$5F,$D9,$35,$E2,$F0,$23,$C1,$FD,$07
	.dc.b	$BC,$F6,$69,$27,$FC,$CC,$FE,$EB,$E2,$7F,$CE,$11,$FB,$86,$BF,$C0
	.dc.b	$5A,$FE,$D9,$AF,$EC,$DA,$FE,$C9,$BB,$1A,$63,$F0,$A7,$D7,$7F,$0A
	.dc.b	$6D,$7F,$08,$6B,$FC,$45,$AF,$F0,$6F,$C2,$EB,$FC,$0D,$AF,$F0,$26
	.dc.b	$D9,$BF,$81,$BE,$9F,$F8,$CB,$6B,$F8,$BB,$5F,$E4,$0D,$7F,$8A,$FE
	.dc.b	$BA,$F9,$6F,$E4,$2D,$8F,$E3,$ED,$7F,$94,$35,$FE,$C3,$F2,$DF,$EF
	.dc.b	$77,$FD,$42,$D7,$FA,$F3,$5D,$FA,$1F,$DD,$23,$11,$FD,$B5,$AF,$F8
	.dc.b	$C9,$A9,$CF,$4B,$F3,$1F,$E2,$53,$FE,$DC,$6A,$8D,$78,$1F,$DA,$27
	.dc.b	$3E,$27,$FB,$E7,$E2,$BF,$95,$7C,$91,$A4,$C3,$FB,$B4,$ED,$B3,$B6
	.dc.b	$CE,$DF,$F5,$A7,$D8,$76,$26,$1F,$CD,$DB,$1D,$1A,$F0,$8F,$F6,$4F
	.dc.b	$CB,$3F,$59,$FE,$AF,$3F,$E1,$18,$C7,$9D,$A9,$CC,$FF,$96,$27,$2F
	.dc.b	$F8,$35,$AF,$F8,$59,$AF,$38,$E2,$A3,$2F,$CB,$FB,$1F,$90,$47,$FC
	.dc.b	$31,$AF,$D1,$8F,$EA,$5F,$20,$F9,$FA,$4A,$F9,$31,$96,$C6,$5F,$F6
	.dc.b	$D4,$65,$B1,$FD,$77,$E9,$DF,$2B,$49,$3F,$52,$32,$D8,$CB,$FE,$E5
	.dc.b	$8C,$B6,$3F,$21,$F7,$32,$3F,$2E,$D7,$F2,$ED,$7F,$3A,$D7,$FD,$0B
	.dc.b	$F1,$AF,$B4,$FF,$14,$37,$DC,$FF,$80,$2D,$9F,$F1,$F3,$5F,$F7,$C7
	.dc.b	$D8,$7F,$27,$7F,$D1,$FC,$A8,$C4,$23,$1E,$37,$F7,$F9,$CC,$62,$7F
	.dc.b	$2B,$18,$24,$FE,$7F,$D2,$93,$F6,$3A,$11,$88,$FF,$2C,$F0,$FE,$AE
	.dc.b	$31,$1F,$E5,$F8,$C3,$C6,$3F,$71,$ED,$48,$FD,$FB,$5F,$DF,$B5,$FE
	.dc.b	$0A,$D7,$F6,$DF,$0B,$FB,$32,$3E,$57,$FC,$BB,$BF,$F4,$38,$FF,$59
	.dc.b	$DF,$E0,$BC,$8C,$92,$3F,$27,$3B,$7F,$80,$36,$78,$46,$A7,$FC,$D3
	.dc.b	$F8,$B3,$E8,$29,$27,$F3,$B1,$96,$C6,$5F,$8F,$19,$6C,$7F,$9D,$3F
	.dc.b	$14,$7D,$05,$24,$FE,$A2,$32,$FF,$0C,$FA,$8F,$ED,$EF,$FA,$37,$85
	.dc.b	$18,$8F,$C2,$B5,$FC,$23,$72,$7E,$13,$F9,$E9,$FB,$0D,$7E,$4B,$5F
	.dc.b	$E4,$8D,$8E,$3A,$7F,$A3,$23,$04,$7F,$D1,$FE,$5B,$E6,$A9,$2F,$F8
	.dc.b	$1C,$65,$FE,$4F,$F5,$6F,$B0,$FF,$2B,$36,$33,$18,$3F,$4F,$FB,$23
	.dc.b	$8E,$92,$7F,$53,$19,$7F,$88,$FC,$7B,$A7,$FD,$46,$D8,$CC,$60,$8E
	.dc.b	$08,$7E,$82,$38,$36,$36,$D3,$67,$E8,$FD,$CF,$F6,$5E,$C6,$92,$3F
	.dc.b	$CE,$9F,$2E,$E7,$A4,$8F,$E9,$71,$97,$FD,$84,$3B,$DF,$61,$EF,$F8
	.dc.b	$FF,$E9,$D3,$98,$C7,$95,$FE,$D3,$AD,$ED,$FE,$DF,$8F,$F4,$57,$AE
	.dc.b	$FB,$1F,$F6,$C6,$3C,$AF,$78,$FF,$B8,$2F,$96,$3D,$20,$77,$C7,$A3
	.dc.b	$FE,$92,$C7,$A3,$83,$47,$33,$8D,$F6,$C3,$C1,$7D,$7D,$55,$F2,$BE
	.dc.b	$07,$2A,$DE,$1D,$70,$E5,$8F,$B3,$2F,$5B,$E8,$9D,$65,$E9,$C3,$94
	.dc.b	$ED,$CC,$C7,$BD,$F4,$7D,$97,$D1,$F4,$DF,$39,$D4,$74,$36,$73,$C3
	.dc.b	$CE,$F4,$9F,$76,$3E,$D9,$C9,$31,$F4,$57,$C7,$CB,$E1,$E3,$E1,$8E
	.dc.b	$C4,$C7,$CD,$3A,$8F,$05,$72,$BE,$91,$E4,$3C,$59,$69,$F5,$BF,$6A
	.dc.b	$B9,$46,$8F,$BE,$79,$0A,$7D,$C2,$DF,$E2,$1B,$9F,$EA,$1F,$C9,$35
	.dc.b	$FC,$0B,$5F,$CE,$36,$0F,$ED,$EF,$F2,$8D,$C3,$FD,$C5,$F6,$2F,$9C
	.dc.b	$FD,$35,$6B,$87,$F0,$97,$BC,$7B,$C7,$B6,$F8,$6F,$A8,$93,$F8,$0A
	.dc.b	$E1,$9A,$F7,$39,$AD,$E1,$DE,$3E,$B4,$EA,$30,$4C,$BE,$0D,$78,$72
	.dc.b	$F3,$C6,$2F,$A1,$2B,$FC,$91,$87,$86,$B7,$3E,$A3,$55,$C6,$B7,$87
	.dc.b	$86,$30,$49,$E1,$3A,$AE,$71,$F4,$30,$F5,$D6,$F0,$F0,$65,$8A,$E3
	.dc.b	$39,$C7,$B4,$60,$87,$33,$E1,$E1,$DB,$3F,$B7,$8F,$39,$D6,$5E,$3C
	.dc.b	$7B,$75,$FE,$03,$E1,$7F,$90,$57,$F5,$FD,$51,$88,$FF,$11,$C6,$2B
	.dc.b	$F9,$3C,$E6,$7F,$CF,$9C,$5F,$DE,$67,$FC,$A6,$D8,$FF,$0D,$B5,$FE
	.dc.b	$71,$DF,$C7,$0F,$C8,$FA,$17,$13,$49,$87,$F3,$66,$DF,$E9,$2B,$8C
	.dc.b	$E4,$91,$9A,$E7,$18,$8F,$75,$7A,$79,$7F,$A4,$75,$5C,$BF,$D8,$F7
	.dc.b	$EA,$8C,$1E,$7E,$A7,$CB,$52,$4F,$F6,$E7,$A7,$E0,$5F,$ED,$38,$FF
	.dc.b	$3A,$B7,$2F,$F1,$EB,$5F,$F1,$E3,$5F,$F1,$6D,$BD,$1F,$33,$6B,$FB
	.dc.b	$EC,$60,$8E,$09,$3F,$E4,$A9,$E0,$FC,$7C,$3F,$BB,$4F,$F8,$13,$87
	.dc.b	$D8,$47,$FD,$17,$E0,$6A,$BF,$CA,$6D,$7F,$C3,$73,$F2,$63,$11,$FE
	.dc.b	$44,$FC,$73,$F8,$5D,$7F,$A5,$1B,$1F,$E5,$C6,$BF,$E5,$B6,$BC,$77
	.dc.b	$FA,$91,$88,$FF,$64,$37,$63,$7F,$26,$7C,$67,$DB,$6E,$1F,$EA,$46
	.dc.b	$DF,$F9,$B1,$AF,$F9,$AD,$AF,$F9,$71,$AF,$FB,$5D,$B6,$6F,$E6,$1F
	.dc.b	$EA,$2A,$FF,$58,$B6,$BF,$CF,$2D,$7F,$CF,$0D,$7F,$9E,$F2,$3F,$00
	.dc.b	$FD,$54,$7F,$98,$5B,$19,$AF,$F0,$9B,$67,$FA,$19,$B8,$CC,$F0,$23
	.dc.b	$E4,$F9,$AF,$CF,$59,$A4,$8D,$12,$BD,$31,$97,$F5,$DD,$9F,$B9,$7F
	.dc.b	$D4,$AD,$7F,$D4,$8D,$E7,$FF,$7C,$63,$FB,$D8,$C1,$1C,$1E,$4E,$29
	.dc.b	$87,$CA,$6D,$FB,$A5,$F1,$3F,$10,$95,$F1,$1B,$3B,$6B,$FD,$A2,$78
	.dc.b	$CF,$C8,$C1,$ED,$BC,$DF,$E2,$4D,$9D,$36,$FF,$87,$35,$FF,$42,$35
	.dc.b	$FF,$63,$B5,$FF,$63,$36,$BB,$D7,$F1,$FB,$7D,$A5,$26,$1E,$58,$CB
	.dc.b	$63,$38,$7F,$1F,$9C,$B6,$3E,$EC,$49,$FC,$9B,$67,$FC,$13,$EC,$BE
	.dc.b	$C5,$F0,$5F,$CE,$63,$0D,$8C,$DF,$E3,$A7,$2D,$8C,$D7,$46,$BF,$E1
	.dc.b	$89,$DB,$9A,$FE,$7D,$F5,$EF,$95,$A4,$C3,$EA,$46,$5B,$19,$AF,$E7
	.dc.b	$D3,$96,$C7,$F4,$9F,$80,$FF,$36,$57,$DD,$6C,$FF,$45,$6B,$FE,$BB
	.dc.b	$6B,$FE,$FE,$6B,$FE,$FD,$6E,$A6,$9A,$E6,$72,$7F,$20,$A4,$63,$AF
	.dc.b	$F5,$E4,$C3,$FD,$17,$EE,$BE,$8B,$F7,$ED,$8C,$C6,$08,$E0,$97,$FC
	.dc.b	$02,$78,$3F,$E4,$4F,$25,$F0,$6F,$B3,$FE,$E9,$18,$6C,$66,$FF,$84
	.dc.b	$CE,$5B,$F8,$57,$AE,$0D,$26,$5D,$3E,$21,$F4,$14,$97,$D2,$32,$F6
	.dc.b	$F8,$A7,$D3,$F9,$1B,$19,$8C,$1E,$8F,$C2,$BF,$C1,$DA,$FB,$5A,$FF
	.dc.b	$99,$33,$46,$23,$FD,$4B,$18,$F3,$79,$12,$FD,$F5,$BA,$56,$E7,$F8
	.dc.b	$27,$A8,$FB,$BF,$F0,$A3,$63,$31,$83,$E6,$F7,$3E,$2A,$92,$7E,$6C
	.dc.b	$65,$B1,$9C,$3E,$C4,$E5,$B1,$FE,$23,$FA,$DF,$C6,$74,$9A,$4E,$1F
	.dc.b	$BB,$F8,$3B,$ED,$A9,$32,$FB,$B1,$96,$C6,$71,$FC,$44,$E5,$B1,$FE
	.dc.b	$33,$F9,$6F,$EA,$D0,$C5,$18,$6F,$C0,$6C,$FC,$7F,$CC,$3E,$8A,$93
	.dc.b	$2F,$C7,$C6,$5B,$19,$C3,$F3,$53,$96,$C7,$F9,$17,$B7,$FA,$C5,$7F
	.dc.b	$39,$6C,$FF,$90,$5A,$FF,$AA,$30,$FD,$9F,$97,$FA,$2F,$C1,$BF,$BD
	.dc.b	$F2,$1A,$4F,$3F,$44,$C7,$F4,$B1,$97,$FA,$47,$CE,$BE,$03,$FC,$B0
	.dc.b	$D8,$CC,$60,$FD,$8F,$E0,$CF,$A3,$A4,$9F,$D8,$C6,$5B,$19,$D3,$F7
	.dc.b	$93,$96,$C7,$F9,$9E,$7F,$CB,$71,$FE,$61,$6B,$F0,$5A,$8F,$B4,$E1
	.dc.b	$FD,$0F,$D1,$3E,$6B,$F8,$3B,$67,$F8,$2B,$67,$F8,$73,$5F,$D9,$7D
	.dc.b	$1B,$EE,$3F,$87,$B6,$BF,$86,$B6,$7F,$8B,$35,$E3,$3C,$09,$3C,$1E
	.dc.b	$7F,$3F,$23,$9B,$67,$9B,$5F,$1B,$5F,$81,$C4,$FA,$6F,$EC,$F1,$FE
	.dc.b	$EC,$8C,$7A,$7F,$E8,$19,$CC,$F0,$6B,$FC,$F1,$B3,$C2,$FF,$C6,$2D
	.dc.b	$9F,$F7,$AB,$5F,$F7,$CF,$C3,$BF,$6D,$FF,$2D,$37,$DD,$FF,$54,$36
	.dc.b	$FF,$C7,$6D,$AE,$78,$7F,$A0,$DF,$17,$CF,$F9,$C9,$5E,$29,$D6,$3E
	.dc.b	$F9,$CD,$72,$7F,$C9,$AF,$E0,$FE,$1B,$FC,$92,$FE,$14,$38,$74,$57
	.dc.b	$7F,$50,$F9,$E7,$D5,$61,$B5,$FE,$54,$9D,$4E,$4E,$3E,$B3,$EA,$A9
	.dc.b	$2B,$84,$65,$F9,$7F,$1E,$FA,$2F,$35,$36,$FB,$B3,$9A,$FC,$F4,$E5
	.dc.b	$FE,$E7,$19,$7C,$13,$F7,$7E,$73,$9B,$CA,$EC,$F0,$6E,$CB,$E5,$87
	.dc.b	$58,$FC,$66,$1A,$DF,$F7,$D8,$7D,$3E,$9F,$9B,$5F,$91,$6C,$EA,$7F
	.dc.b	$0F,$F5,$8F,$BA,$EE,$D8,$F5,$E9,$F9,$A8,$C4,$7A,$8F,$CA,$5F,$0A
	.dc.b	$91,$C0,$FF,$09,$CF,$0C,$BD,$18,$78,$5F,$57,$D5,$BE,$5F,$DE,$DB
	.dc.b	$E5,$9A,$31,$1E,$13,$FC,$75,$F1,$BC,$2A,$47,$09,$DD,$FD,$38,$D9
	.dc.b	$E5,$F1,$BF,$B8,$F9,$FF,$41,$B1,$9F,$1B,$FB,$63,$EB,$5D,$FF,$EB
	.dc.b	$E8,$D5,$71,$9D,$C7,$C5,$AE,$6F,$C7,$8E,$3E,$37,$7F,$AB,$F6,$4F
	.dc.b	$FA,$0D,$AF,$E8,$CF,$EC,$0E,$1F,$E9,$8D,$7F,$96,$53,$EC,$7F,$49
	.dc.b	$1E,$D6,$BE,$36,$E5,$F9,$59,$57,$82,$31,$E3,$BD,$6F,$DA,$D2,$31
	.dc.b	$87,$F5,$29,$CC,$62,$3D,$13,$B9,$C9,$23,$81,$FE,$A1,$F4,$5F,$75
	.dc.b	$F4,$5B,$1F,$55,$AF,$CB,$BE,$D8,$7A,$78,$9F,$59,$E6,$6D,$7F,$75
	.dc.b	$8C,$11,$C1,$0F,$EF,$0F,$BB,$2F,$C3,$F4,$7C,$C6,$CF,$F8,$45,$A7
	.dc.b	$F9,$BA,$7F,$11,$E8,$BE,$B3,$E3,$B6,$BF,$D7,$CD,$7E,$2B,$5F,$F0
	.dc.b	$33,$5F,$B6,$D7,$FC,$40,$DA,$FF,$10,$CF,$D4,$FB,$57,$D4,$FE,$65
	.dc.b	$B7,$F9,$D6,$BF,$93,$BF,$16,$7F,$81,$FA,$C7,$D3,$7D,$E6,$D7,$FB
	.dc.b	$55,$AF,$DC,$6B,$FE,$1D,$6B,$F9,$26,$BF,$E3,$76,$E4,$6F,$0B,$C9
	.dc.b	$FD,$83,$6B,$F6,$8D,$7F,$55,$3E,$2C,$3F,$31,$EE,$7F,$81,$E7,$FC
	.dc.b	$CF,$E8,$7E,$5A,$31,$1F,$92,$6B,$FE,$33,$6B,$FA,$86,$BF,$E4,$D6
	.dc.b	$E2,$6F,$DB,$7E,$32,$7F,$78,$DB,$FD,$53,$78,$7A,$61,$FC,$EB,$A5
	.dc.b	$F0,$0F,$91,$FE,$BF,$18,$6C,$67,$1F,$F4,$B4,$60,$8E,$08,$7E,$BB
	.dc.b	$F1,$37,$D6,$7F,$19,$6C,$7F,$1D,$6B,$FC,$46,$BB,$3F,$C4,$FC,$9F
	.dc.b	$F7,$8B,$FC,$75,$AE,$7E,$3B,$D5,$F0,$BC,$36,$FE,$11,$F8,$EB,$E7
	.dc.b	$3F,$82,$B6,$FF,$8C,$1F,$E7,$77,$0F,$F0,$46,$BF,$E6,$5F,$A1,$FE
	.dc.b	$4F,$5F,$C4,$DB,$C1,$FC,$71,$B3,$FE,$A6,$F8,$B7,$D7,$FF,$64,$FA
	.dc.b	$BF,$E5,$F3,$FE,$2B,$6B,$FE,$30,$6B,$FE,$E4,$6F,$80,$69,$3C,$7F
	.dc.b	$EC,$EF,$0E,$87,$DB,$BD,$9F,$E8,$CD,$C3,$FA,$53,$5F,$EC,$98,$75
	.dc.b	$C3,$FA,$67,$C9,$FF,$47,$9F,$E8,$8D,$DB,$F9,$93,$76,$FE,$76,$BE
	.dc.b	$6B,$8F,$97,$2F,$FA,$C4,$ED,$FF,$52,$3B,$7F,$AE,$7D,$17,$FB,$C5
	.dc.b	$BF,$25,$FD,$76,$31,$1F,$D5,$5F,$9B,$B1,$CE,$5F,$D1,$5B,$97,$F4
	.dc.b	$E6,$D7,$FA,$CD,$FA,$BA,$7C,$BF,$1B,$EA,$12,$7F,$C0,$5F,$2E,$FB
	.dc.b	$07,$A7,$99,$F8,$FC,$7F,$C5,$B3,$FF,$0E,$37,$3F,$F4,$46,$1C,$6F
	.dc.b	$FB,$B7,$E0,$7F,$EF,$AC,$3F,$BD,$35,$FE,$E0,$DC,$39,$E5,$FD,$F7
	.dc.b	$E5,$9F,$67,$49,$7F,$E0,$38,$FD,$3D,$7F,$65,$77,$19,$9D,$BF,$DF
	.dc.b	$63,$79,$7F,$72,$FA,$67,$DC,$FF,$82,$DB,$5F,$E0,$F6,$BF,$E0,$55
	.dc.b	$F2,$1B,$99,$3B,$79,$47,$F9,$13,$D3,$7D,$93,$DF,$FF,$34,$C6,$1B
	.dc.b	$19,$BF,$D9,$CE,$5B,$19,$7F,$C1,$D1,$96,$C6,$5F,$F3,$A4,$65,$FF
	.dc.b	$3B,$62,$FC,$5F,$E2,$C7,$F9,$0E,$31,$A7,$F9,$69,$B1,$FE,$43,$AE
	.dc.b	$DA,$FF,$84,$7E,$CD,$F5,$9F,$E3,$16,$D7,$F8,$D5,$AF,$F8,$AD,$FE
	.dc.b	$85,$7F,$E2,$07,$35,$CE,$B9,$1C,$FD,$3F,$F1,$E7,$6F,$F1,$29,$FF
	.dc.b	$10,$36,$FF,$C5,$2D,$C3,$F8,$37,$D9,$F0,$F1,$A4,$7F,$91,$E7,$5D
	.dc.b	$5F,$C2,$6B,$F8,$95,$FF,$08,$25,$7F,$AD,$3A,$7F,$CA,$0F,$F1,$58
	.dc.b	$CB,$FE,$B6,$8C,$BF,$C5,$E3,$FC,$5F,$F7,$FF,$E5,$7C,$BF,$C6,$CD
	.dc.b	$8F,$F1,$FB,$67,$F8,$E4,$7F,$96,$61,$F7,$54,$93,$FE,$6A,$E4,$7C
	.dc.b	$DE,$23,$FC,$AD,$5C,$1F,$F3,$23,$5F,$E3,$73,$C1,$FF,$34,$7D,$93
	.dc.b	$ED,$BF,$CD,$0D,$8C,$C7,$FA,$63,$ED,$9F,$51,$A8,$93,$A3,$FC,$FB
	.dc.b	$3A,$7A,$FC,$A7,$F9,$42,$CD,$24,$7F,$B8,$3E,$BB,$F5,$2F,$FA,$19
	.dc.b	$AF,$F3,$26,$ED,$FE,$5F,$AE,$73,$FE,$81,$FA,$F7,$E1,$7F,$E8,$16
	.dc.b	$C6,$63,$FD,$63,$7C,$B8,$3F,$CD,$F5,$AA,$FF,$3F,$37,$0F,$F3,$63
	.dc.b	$67,$FE,$00,$D7,$1A,$78,$EA,$7D,$AF,$F9,$B9,$B3,$FD,$AA,$75,$39
	.dc.b	$35,$F0,$7E,$72,$4F,$AE,$3F,$D8,$FF,$20,$ED,$FF,$60,$B6,$7F,$BA
	.dc.b	$F4,$7F,$B2,$E7,$26,$A3,$89,$0F,$F5,$8C,$F1,$7F,$AE,$BF,$E9,$5D
	.dc.b	$3F,$DA,$4D,$D3,$FD,$C7,$3E,$2A,$E4,$F0,$2D,$70,$8F,$F5,$87,$C1
	.dc.b	$7E,$8A,$7E,$2B,$73,$FF,$23,$5F,$FA,$4A,$30,$7F,$91,$6F,$FD,$34
	.dc.b	$D8,$FF,$51,$B6,$B9,$C7,$FB,$B1,$78,$8F,$9C,$FE,$F5,$F2,$DF,$DA
	.dc.b	$67,$31,$8E,$D4,$4A,$E1,$E9,$E2,$FF,$B8,$7B,$FF,$E4,$7F,$0B,$76
	.dc.b	$9E,$97,$C2,$31,$86,$A3,$13,$AC,$7C,$14,$C3,$A3,$DE,$F8,$4E,$A3
	.dc.b	$CF,$8E,$5F,$34,$72,$AE,$31,$E0,$33,$3B,$78,$99,$2D,$FB,$3E,$0B
	.dc.b	$F9,$8B,$E5,$9E,$18,$FA,$31,$F1,$C7,$87,$0E,$F5,$BC,$3A,$5E,$DF
	.dc.b	$09,$E3,$9D,$DF,$28,$DE,$3E,$DA,$E1,$5A,$9E,$37,$D4,$97,$F5,$A7
	.dc.b	$53,$EA,$8E,$93,$A8,$F9,$98,$FB,$27,$24,$8E,$AF,$38,$D4,$60,$87
	.dc.b	$53,$9E,$5F,$80,$AE,$D1,$93,$57,$E4,$C3,$F0,$51,$FE,$FD,$6B,$F0
	.dc.b	$4C,$63,$FE,$4B,$7F,$0D,$A7,$EC,$5B,$BF,$F9,$B9,$FC,$DC,$65,$FC
	.dc.b	$B3,$5C,$C7,$D8,$8F,$E0,$58,$7E,$63,$D8,$7D,$A3,$7D,$5C,$C3,$EC
	.dc.b	$E1,$C3,$6F,$9D,$87,$0B,$D6,$3F,$9E,$AD,$5F,$8B,$0F,$C2,$5E,$E3
	.dc.b	$9C,$70,$C3,$D7,$1F,$83,$9D,$5F,$3A,$D4,$F3,$35,$8F,$9D,$E6,$49
	.dc.b	$E0,$EB,$1E,$2F,$C2,$35,$87,$AA,$B9,$57,$02,$47,$91,$F8,$D1,$A8
	.dc.b	$C1,$27,$C9,$3A,$AE,$F9,$FF,$B7,$70,$F4,$E8,$CB,$DB,$8F,$1C,$7C
	.dc.b	$98,$F8,$1F,$E0,$13,$A7,$AC,$EF,$0E,$91,$ED,$3D,$67,$F2,$8D,$3E
	.dc.b	$27,$07,$F2,$46,$E7,$FC,$93,$E1,$6D,$6E,$3F,$8E,$D7,$F2,$A9,$FF
	.dc.b	$09,$F1,$7F,$BD,$5B,$B9,$A6,$37,$E7,$3F,$EF,$B6,$CF,$FA,$2E,$FA
	.dc.b	$9F,$CD,$23,$24,$73,$87,$F4,$0F,$4F,$0B,$1F,$E7,$B5,$87,$F9,$FC
	.dc.b	$DA,$FE,$7E,$F9,$88,$FF,$69,$BF,$15,$FF,$34,$3F,$B8,$F4,$3F,$02
	.dc.b	$F4,$B6,$BF,$C0,$9E,$3E,$27,$2F,$C4,$9F,$F7,$04,$FF,$75,$6D,$75
	.dc.b	$CF,$FB,$E5,$6F,$F0,$9E,$A7,$D0,$E9,$7C,$27,$FA,$8B,$1F,$F0,$94
	.dc.b	$E9,$A7,$C4,$9D,$FA,$99,$C7,$FD,$9B,$7C,$5B,$1E,$AE,$7E,$D3,$9D
	.dc.b	$FF,$C2,$71,$E9,$26,$1E,$97,$EF,$F8,$0F,$B9,$FB,$6D,$9F,$17,$DA
	.dc.b	$FF,$72,$CB,$EA,$47,$F6,$4D,$95,$B9,$C9,$FE,$1E,$71,$8F,$F8,$C7
	.dc.b	$0F,$85,$87,$38,$C6,$1F,$E5,$0F,$73,$F1,$7E,$49,$F9,$1F,$6A,$F7
	.dc.b	$73,$12,$B8,$1F,$1E,$74,$E3,$3F,$F2,$EF,$2F,$43,$E9,$7D,$87,$FA
	.dc.b	$C2,$7F,$CB,$F1,$8F,$33,$FC,$E7,$C6,$AD,$5F,$E2,$63,$13,$A9,$CC
	.dc.b	$FF,$80,$27,$37,$FE,$2E,$9C,$CF,$F8,$2A,$72,$F3,$8C,$BF,$E4,$4E
	.dc.b	$DE,$FA,$FF,$A5,$B0,$F3,$FD,$2F,$EB,$A7,$38,$FF,$A6,$27,$06,$9F
	.dc.b	$F4,$B3,$BF,$B4,$91,$FE,$99,$F3,$AE,$3F,$5B,$91,$82,$F3,$97,$FA
	.dc.b	$9A,$EE,$1F,$E2,$23,$FD,$1B,$DF,$FC,$29,$A7,$E5,$F4,$FC,$EB,$FB
	.dc.b	$C3,$39,$7E,$13,$5F,$99,$CF,$F2,$49,$3F,$1D,$B3,$F1,$5A,$FC,$DA
	.dc.b	$EA,$FE,$BF,$ED,$1F,$37,$FC,$49,$B7,$FC,$33,$64,$7F,$A8,$23,$0E
	.dc.b	$AB,$E1,$37,$4F,$E9,$95,$C2,$FA,$7A,$3D,$5F,$E3,$FA,$FE,$73,$1F
	.dc.b	$D1,$37,$1F,$F3,$EA,$FA,$3D,$F4,$CA,$F2,$4E,$F1,$FF,$52,$D6,$6F
	.dc.b	$EA,$7B,$EF,$99,$FE,$5A,$DC,$3E,$76,$18,$9F,$CD,$B2,$3F,$95,$FC
	.dc.b	$0B,$ED,$1F,$17,$9C,$7E,$97,$0F,$FA,$D2,$F9,$9C,$2B,$6E,$27,$F0
	.dc.b	$DF,$68,$F8,$1F,$E7,$ED,$9F,$9C,$BE,$83,$33,$27,$2F,$08,$C5,$7E
	.dc.b	$47,$E2,$1E,$FF,$F4,$A6,$DF,$D9,$5E,$96,$66,$4E,$5E,$11,$8A,$FC
	.dc.b	$E7,$C4,$FC,$D3,$EA,$75,$7E,$EF,$C7,$FE,$AF,$18,$DB,$FD,$4D,$8F
	.dc.b	$ED,$E7,$F7,$84,$70,$4B,$FD,$C4,$65,$FE,$81,$F8,$7F,$F1,$07,$F5
	.dc.b	$8D,$EA,$FF,$0B,$7A,$8F,$8C,$7C,$5F,$F6,$8A,$D6,$3F,$CA,$FE,$37
	.dc.b	$F8,$1F,$59,$A4,$F1,$3F,$C4,$19,$FF,$A3,$B2,$7D,$A7,$AA,$3F,$98
	.dc.b	$C6,$35,$FE,$FE,$BE,$3D,$F6,$98,$BE,$1B,$FE,$83,$C6,$7B,$9F,$DF
	.dc.b	$1B,$B7,$F7,$6C,$5F,$24,$FB,$4D,$5F,$68,$C5,$7F,$6F,$D7,$FD,$35
	.dc.b	$18,$7F,$B9,$5F,$2B,$F5,$C7,$F9,$E6,$74,$FB,$5A,$E6,$36,$F4,$6C
	.dc.b	$F3,$AF,$F7,$47,$CB,$FF,$74,$CB,$F7,$6D,$8F,$EB,$1E,$FB,$EC,$FF
	.dc.b	$C3,$33,$9C,$7F,$DE,$B3,$9A,$FF,$79,$36,$3F,$BF,$35,$FF,$5F,$F1
	.dc.b	$A3,$11,$DA,$FF,$C5,$71,$88,$EB,$18,$3F,$C5,$FE,$EB,$EF,$BF,$C5
	.dc.b	$6D,$C3,$F8,$D7,$DA,$7D,$67,$FB,$84,$66,$FF,$9D,$4F,$E0,$5F,$1B
	.dc.b	$5F,$EE,$CD,$58,$A3,$0D,$E4,$FF,$20,$C6,$08,$FF,$91,$3E,$75,$F2
	.dc.b	$1F,$E4,$06,$E3,$FC,$BF,$E7,$7F,$AA,$E3,$DA,$72,$FF,$4F,$8F,$CE
	.dc.b	$93,$A3,$FC,$9E,$DA,$FC,$46,$48,$C5,$7F,$8B,$5F,$D2,$F7,$7F,$2D
	.dc.b	$AF,$E7,$CD,$9F,$EB,$9D,$EF,$53,$F7,$FE,$6F,$F9,$5A,$31,$7D,$9B
	.dc.b	$5F,$E3,$96,$BF,$71,$AF,$F8,$7D,$AF,$CA,$6E,$26,$98,$D7,$F9,$34
	.dc.b	$FF,$6F,$6F,$27,$F4,$EF,$9F,$7A,$9F,$E6,$A9,$CE,$3F,$E2,$D9,$CB
	.dc.b	$F1,$1B,$7F,$E4,$F6,$BF,$92,$6B,$C6,$BF,$CF,$71,$88,$FC,$0B,$71
	.dc.b	$37,$E8,$BF,$97,$DF,$F6,$96,$D7,$F8,$1E,$3F,$CD,$78,$FF,$66,$9C
	.dc.b	$BF,$DA,$23,$FC,$DF,$3F,$A4,$6B,$FE,$47,$6B,$F9,$26,$DB,$EF,$FF
	.dc.b	$91,$CF,$E5,$9B,$D2,$FA,$47,$D7,$7F,$A4,$27,$77,$C7,$DD,$C1,$FC
	.dc.b	$82,$B7,$8F,$F8,$0E,$BA,$EE,$F8,$BF,$C0,$E9,$FE,$CB,$8C,$71,$71
	.dc.b	$F9,$0F,$F3,$C5,$6B,$0F,$F1,$D3,$76,$FC,$03,$76,$F3,$57,$0B,$FF
	.dc.b	$40,$7D,$43,$E5,$29,$FE,$83,$FA,$8F,$CC,$69,$E1,$9D,$67,$FE,$81
	.dc.b	$6E,$7F,$E2,$96,$CF,$9E,$B8,$D7,$F9,$36,$72,$FF,$93,$E3,$F4,$95
	.dc.b	$FE,$55,$6B,$9E,$6F,$F2,$FC,$62,$3D,$7F,$11,$F0,$0F,$37,$D6,$DB
	.dc.b	$FF,$4A,$6D,$FC,$9F,$7F,$95,$5A,$9C,$BF,$EC,$EB,$13,$E4,$3F,$A2
	.dc.b	$38,$0E,$1D,$15,$DF,$D6,$3E,$B5,$F5,$B8,$6D,$7F,$4F,$9D,$4E,$4E
	.dc.b	$3E,$BF,$FB,$2E,$74,$D7,$FD,$89,$3F,$E1,$9E,$0D,$9F,$D6,$73,$FC
	.dc.b	$37,$D2,$79,$E3,$F0,$DA,$A3,$1A,$FE,$11,$B8,$7F,$03,$DF,$FC,$50
	.dc.b	$FC,$88,$CE,$1D,$23,$07,$FA,$FE,$76,$FF,$99,$D7,$D6,$FC,$89,$46
	.dc.b	$27,$CC,$D7,$FD,$2C,$D7,$FD,$29,$C5,$FE,$66,$7F,$19,$F0,$7E,$28
	.dc.b	$C4,$77,$BC,$4C,$8D,$9E,$58,$F4,$FC,$0B,$EB,$69,$27,$FD,$F6,$7F
	.dc.b	$90,$FE,$3F,$0F,$FA,$06,$78,$3F,$ED,$D6,$BF,$EC,$83,$81,$FE,$79
	.dc.b	$EA,$FF,$61,$37,$E3,$3E,$6F,$7A,$73,$3F,$E2,$0E,$EF,$F6,$33,$B2
	.dc.b	$1F,$E9,$F8,$DF,$17,$F8,$D2,$77,$3F,$EA,$3F,$AD,$FF,$38,$3F,$D4
	.dc.b	$9A,$FF,$BF,$5A,$FF,$80,$1A,$FF,$B8,$FC,$2F,$CB,$3F,$CA,$7A,$3C
	.dc.b	$F1,$88,$ED,$9F,$AB,$C3,$79,$9F,$A4,$D3,$F9,$9B,$63,$35,$FD,$06
	.dc.b	$30,$47,$04,$3F,$58,$7F,$7E,$F1,$9F,$5B,$F2,$1B,$1F,$EA,$56,$BF
	.dc.b	$EC,$3A,$EE,$FF,$6A,$FA,$DF,$27,$39,$A4,$8F,$ED,$4D,$7F,$B5,$35
	.dc.b	$FF,$12,$E3,$F7,$B7,$FF,$56,$B7,$A7,$F8,$2B,$FE,$14,$E4,$FB,$B8
	.dc.b	$77,$BF,$F0,$67,$CD,$FC,$5F,$2C,$D2,$47,$F8,$22,$32,$D8,$CB,$FE
	.dc.b	$09,$7E,$77,$9A,$FA,$7F,$C5,$B6,$7F,$D9,$0D,$7F,$C3,$13,$F9,$0F
	.dc.b	$AD,$FF,$4B,$C7,$F8,$59,$AF,$D0,$C9,$18,$8F,$F2,$2B,$5F,$7B,$5F
	.dc.b	$F1,$9B,$63,$8F,$C1,$FE,$0E,$31,$C4,$F9,$1F,$D0,$B6,$FF,$DB,$CD
	.dc.b	$7F,$DF,$93,$E4,$DF,$FD,$B7,$EB,$3E,$B7,$F1,$CD,$AF,$F2,$EC,$60
	.dc.b	$8E,$08,$7F,$98,$1F,$CB,$7B,$1F,$56,$15,$FE,$6E,$E3,$FD,$AC,$E6
	.dc.b	$3C,$35,$FC,$93,$B3,$FD,$13,$F2,$3F,$D0,$9B,$87,$FA,$1D,$AF,$FA
	.dc.b	$1B,$DB,$4E,$DB,$1B,$6B,$B8,$D1,$FA,$BF,$AC,$7D,$6F,$F0,$86,$CF
	.dc.b	$F5,$4C,$7F,$DF,$F1,$88,$FF,$54,$DF,$49,$FF,$5C,$F8,$FF,$EE,$A9
	.dc.b	$FF,$51,$35,$FF,$57,$B6,$3F,$D7,$FF,$05,$C7,$AB,$FD,$79,$E7,$7E
	.dc.b	$F7,$E0,$CD,$26,$1F,$E6,$0F,$9F,$FE,$2B,$5F,$A4,$6C,$7F,$9E,$1B
	.dc.b	$B7,$FB,$2F,$EB,$7F,$A5,$FD,$37,$F2,$2A,$FE,$D9,$DC,$8C,$47,$F7
	.dc.b	$28,$C7,$91,$FD,$2E,$73,$3F,$E5,$96,$CE,$A3,$FB,$14,$60,$DF,$5B
	.dc.b	$DC,$FF,$63,$CF,$F9,$CB,$E0,$FF,$9A,$C6,$23,$F0,$BA,$78,$6F,$D7
	.dc.b	$6F,$8C,$FE,$63,$C2,$AF,$E0,$71,$86,$FB,$9F,$D0,$21,$F2,$F6,$26
	.dc.b	$3F,$6E,$B6,$D3,$FD,$3F,$5C,$1C,$60,$F9,$FF,$F7,$9C,$61,$AF,$A0
	.dc.b	$FF,$53,$B8,$C7,$CD,$18,$DB,$FA,$9C,$E7,$0F,$E3,$6D,$9E,$3F,$17
	.dc.b	$AA,$FA,$FE,$2C,$9F,$37,$07,$D7,$8F,$F0,$4F,$C9,$BE,$B5,$E9,$6E
	.dc.b	$FC,$97,$B7,$8F,$0F,$F1,$7D,$7C,$B8,$71,$BF,$EC,$FF,$2D,$FE,$F7
	.dc.b	$BF,$82,$D7,$DD,$BA,$31,$7F,$E6,$3C,$3A,$6F,$FC,$DB,$EB,$7F,$D8
	.dc.b	$38,$7C,$A6,$BC,$9B,$5F,$1B,$04,$7E,$B6,$30,$EA,$BA,$7C,$3F,$F7
	.dc.b	$C8,$C5,$FF,$B5,$9E,$13,$FD,$1B,$E8,$DF,$5B,$FE,$02,$6D,$7D,$C6
	.dc.b	$BE,$C5,$F1,$3B,$99,$3B,$79,$47,$F8,$E3,$D2,$7D,$6F,$F8,$CA,$BF
	.dc.b	$CA,$D1,$8F,$96,$FF,$2C,$B6,$3F,$B7,$87,$0F,$F7,$26,$BC,$A7,$A3
	.dc.b	$FE,$64,$8C,$B6,$32,$FE,$8F,$EB,$7E,$1E,$46,$9F,$1A,$BB,$EB,$F3
	.dc.b	$FE,$B5,$F3,$1F,$E2,$66,$E1,$FA,$06,$BF,$80,$7F,$A3,$61,$FD,$C9
	.dc.b	$CD,$73,$AE,$47,$3F,$4B,$FC,$63,$DD,$FC,$6B,$0F,$D5,$B5,$FC,$5B
	.dc.b	$6B,$F1,$5F,$5B,$C3,$C5,$95,$C1,$FE,$33,$1F,$C6,$B2,$FE,$28,$49
	.dc.b	$FE,$22,$D7,$FC,$3B,$D3,$FC,$76,$31,$1F,$C8,$39,$BF,$6A,$FF,$73
	.dc.b	$FA,$2B,$8F,$E4,$DB,$5F,$E1,$06,$CF,$F2,$18,$FF,$26,$C7,$FA,$DE
	.dc.b	$3F,$83,$B5,$FD,$3B,$67,$FD,$91,$C4,$FA,$0E,$3F,$19,$2F,$8B,$FE
	.dc.b	$80,$9D,$47,$F1,$2F,$A4,$7D,$6F,$F9,$71,$B1,$98,$FF,$43,$FD,$6B
	.dc.b	$E8,$B5,$12,$74,$7F,$0A,$9D,$3D,$7E,$4B,$FC,$99,$66,$92,$3F,$D3
	.dc.b	$DF,$59,$FA,$67,$F8,$9B,$5F,$E1,$ED,$D3,$F7,$95,$CE,$7F,$95,$FD
	.dc.b	$4B,$EB,$7F,$CE,$AD,$8C,$C7,$FA,$82,$F9,$70,$7F,$90,$2B,$55,$FC
	.dc.b	$49,$B8,$7F,$09,$6C,$FF,$59,$AE,$34,$F0,$54,$FA,$DF,$F3,$53,$67
	.dc.b	$F6,$B3,$A9,$C9,$AF,$80,$F3,$92,$7D,$71,$FE,$B6,$F8,$FF,$DE,$47
	.dc.b	$FB,$0D,$AF,$E4,$9A,$FF,$70,$6B,$FD,$BD,$FF,$39,$E5,$FE,$C8,$8E
	.dc.b	$87,$5E,$DA,$FF,$BB,$0A,$7F,$3E,$9E,$F3,$C6,$34,$43,$53,$FE,$A4
	.dc.b	$F8,$1F,$D0,$C7,$FB,$B7,$1F,$F2,$46,$9F,$E8,$D9,$CC,$FF,$9C,$1A
	.dc.b	$FC,$B7,$FC,$95,$7F,$CE,$5B,$3B,$7B,$3F,$ED,$E9,$78,$6F,$99,$FF
	.dc.b	$01,$FC,$A7,$F6,$F9,$CC,$63,$BB,$FE,$00,$56,$FD,$35,$2F,$FB,$B7
	.dc.b	$AA,$F8,$1F,$E2,$F1,$CE,$F8,$56,$F0,$E7,$8F,$58,$E7,$3A,$AE,$77
	.dc.b	$CA,$F7,$8F,$82,$7A,$61,$C7,$0F,$3C,$E6,$BC,$B8,$F8,$FC,$3F,$3D
	.dc.b	$78,$4E,$F7,$C7,$2F,$05,$F6,$9D,$5F,$17,$A1,$C3,$0A,$F8,$0E,$85
	.dc.b	$C7,$D3,$8F,$81,$F3,$CF,$0C,$3B,$D7,$33,$BE,$5E,$F7,$15,$C1,$EB
	.dc.b	$3C,$30,$F2,$46,$32,$F5,$13,$0F,$3C,$EA,$F9,$13,$2F,$73,$E6,$BF
	.dc.b	$25,$F1,$7A,$5F,$2C,$7C,$91,$ED,$BE,$24,$7F,$DF,$91,$AC,$7D,$B1
	.dc.b	$B3,$D7,$3F,$52,$75,$7F,$E4,$F9,$FA,$07,$FB,$8A,$BF,$0E,$D8,$D3
	.dc.b	$F8,$6D,$3F,$6E,$FE,$69,$B3,$FE,$AF,$DF,$FB,$EC,$E5,$B1,$98,$F9
	.dc.b	$73,$FC,$1B,$2F,$CC,$FA,$EF,$AD,$FF,$23,$C7,$E6,$67,$59,$FE,$89
	.dc.b	$B7,$E3,$C7,$E7,$E1,$C7,$7F,$BD,$5A,$AF,$11,$E0,$AD,$E3,$DB,$0E
	.dc.b	$91,$C0,$93,$F9,$2C,$3A,$3C,$49,$8F,$96,$79,$E3,$C2,$31,$3A,$BF
	.dc.b	$D1,$DF,$29,$C9,$23,$C3,$3D,$F0,$F2,$3E,$D3,$91,$67,$E3,$C6,$A3
	.dc.b	$DB,$1C,$DD,$4F,$82,$75,$97,$D1,$AD,$57,$0A,$D6,$1E,$97,$C7,$3C
	.dc.b	$F6,$63,$E9,$9C,$92,$FB,$89,$DE,$3E,$0C,$FE,$BF,$0F,$FB,$CF,$5F
	.dc.b	$F4,$1F,$0F,$F2,$76,$E9,$FC,$B7,$E1,$2E,$5F,$CB,$66,$DF,$FB,$EE
	.dc.b	$7F,$99,$B6,$BF,$AB,$E0,$9C,$B6,$33,$8F,$28,$FB,$78,$FF,$37,$F5
	.dc.b	$71,$32,$BF,$D3,$F5,$A7,$FA,$04,$E7,$2F,$F4,$0C,$E5,$FE,$85,$5D
	.dc.b	$67,$57,$D6,$FC,$B1,$E4,$23,$FD,$BB,$9D,$F5,$BE,$E6,$C7,$F8,$1B
	.dc.b	$C7,$C4,$E5,$F8,$33,$FE,$E4,$9E,$51,$83,$3F,$5B,$CC,$F9,$7E,$3B
	.dc.b	$E0,$BF,$D4,$B5,$FE,$12,$9D,$34,$FF,$56,$C6,$B4,$FF,$07,$4F,$28
	.dc.b	$C6,$7F,$EB,$CB,$E9,$E9,$FA,$79,$F4,$DC,$7D,$D1,$E8,$23,$FE,$0E
	.dc.b	$F0,$1F,$55,$F7,$9B,$1F,$EB,$EF,$AD,$FE,$E7,$A7,$FA,$C2,$72,$DA
	.dc.b	$E0,$7B,$A7,$79,$7F,$8C,$23,$F1,$3E,$53,$D9,$A7,$C4,$FA,$D7,$BB
	.dc.b	$8C,$FF,$CB,$33,$72,$FF,$64,$D6,$A7,$27,$F9,$7B,$9B,$A9,$F3,$BE
	.dc.b	$A7,$FD,$53,$19,$3F,$DA,$EE,$2B,$FC,$E6,$D7,$FB,$2F,$02,$31,$1F
	.dc.b	$DF,$A3,$1E,$67,$F8,$49,$B5,$FE,$AA,$6C,$F4,$9D,$57,$28,$FF,$20
	.dc.b	$76,$F8,$76,$FF,$67,$37,$0E,$D4,$F9,$EF,$D7,$CE,$6F,$8E,$18,$3D
	.dc.b	$D3,$6F,$FD,$37,$F4,$BC,$0E,$3E,$75,$8C,$92,$FF,$D3,$F1,$95,$92
	.dc.b	$7F,$C3,$A3,$CF,$FF,$13,$3F,$E2,$08,$D4,$EE,$72,$FF,$BF,$B3,$FC
	.dc.b	$F5,$6C,$E7,$3A,$AF,$CC,$4E,$AB,$E9,$F3,$FC,$82,$5F,$CD,$9D,$4F
	.dc.b	$D6,$6B,$F2,$27,$6F,$F8,$78,$FF,$37,$7D,$6B,$E6,$7F,$8C,$B6,$FC
	.dc.b	$93,$AF,$4B,$35,$27,$3E,$12,$F8,$5F,$C6,$9F,$F2,$0C,$E6,$F8,$4E
	.dc.b	$4F,$F3,$F6,$9F,$A6,$CF,$E3,$D7,$FA,$8F,$9F,$F9,$5C,$BE,$82,$C4
	.dc.b	$C7,$C3,$3C,$5E,$EE,$DC,$C7,$F3,$4F,$79,$F2,$B4,$FD,$15,$5A,$FD
	.dc.b	$0B,$87,$E0,$FC,$03,$EB,$1F,$5B,$9D,$7E,$D5,$DD,$7E,$FD,$EE,$B1
	.dc.b	$87,$C2,$FA,$D7,$C0,$7F,$46,$6C,$7C,$A5,$F4,$39,$99,$39,$78,$46
	.dc.b	$32,$F9,$5C,$0F,$88,$7B,$FF,$D4,$9B,$87,$D6,$5E,$AE,$66,$4E,$5E
	.dc.b	$11,$8B,$EB,$F1,$3F,$35,$FF,$5A,$37,$BF,$F8,$17,$8F,$FD,$82,$31
	.dc.b	$D6,$8F,$DE,$93,$1F,$CB,$36,$FF,$86,$B5,$FC,$C3,$FD,$A3,$EB,$3F
	.dc.b	$8D,$47,$B9,$BC,$5F,$AC,$F4,$9F,$5A,$F8,$BF,$EE,$35,$AC,$7F,$23
	.dc.b	$F1,$BF,$C3,$78,$4D,$27,$89,$FE,$A8,$C5,$F5,$9E,$9B,$FD,$0C,$63
	.dc.b	$5F,$EF,$EB,$E3,$DF,$5B,$8B,$E1,$BF,$D4,$F1,$9E,$EF,$F8,$11,$B9
	.dc.b	$74,$F9,$27,$D6,$E6,$BA,$C6,$27,$F6,$D7,$FE,$17,$8C,$3F,$B9,$CD
	.dc.b	$18,$8F,$F4,$CC,$62,$FD,$93,$99,$FF,$4F,$B5,$E9,$1A,$3F,$C2,$FF
	.dc.b	$2F,$FD,$DE,$3F,$1A,$DE,$4F,$DA,$FC,$03,$EB,$7F,$C3,$F3,$9C,$BF
	.dc.b	$7F,$39,$7F,$DE,$AD,$BF,$E2,$1C,$7F,$EC,$1B,$FF,$12,$CE,$67,$FD
	.dc.b	$B8,$D7,$9C,$68,$95,$FF,$00,$3D,$D7,$D6,$FF,$8C,$1B,$1F,$C5,$3E
	.dc.b	$B7,$E9,$E7,$FC,$32,$73,$87,$EA,$27,$EB,$BF,$EE,$B6,$BF,$C6,$1A
	.dc.b	$B2,$46,$1B,$C9,$FE,$3E,$8C,$11,$FF,$20,$FC,$EB,$E5,$3F,$C8,$AD
	.dc.b	$CB,$F9,$4F,$CF,$7F,$59,$C1,$39,$C7,$F7,$B3,$F9,$92,$74,$7F,$93
	.dc.b	$1B,$7F,$86,$B4,$62,$BE,$7B,$FA,$1E,$EF,$E4,$91,$FB,$CF,$37,$FC
	.dc.b	$AF,$E0,$7F,$6B,$C3,$F9,$DD,$6B,$7F,$E6,$73,$9B,$FB,$2D,$7F,$DC
	.dc.b	$F8,$A3,$11,$FD,$01,$AF,$C9,$6C,$6F,$D3,$FF,$32,$C6,$0C,$ED,$FC
	.dc.b	$92,$BF,$8E,$B6,$FF,$A2,$FD,$07,$F6,$5C,$BF,$A6,$CE,$5F,$E9,$F1
	.dc.b	$FD,$BA,$BC,$AD,$73,$F0,$9F,$E7,$48,$C4,$7D,$D6,$E4,$69,$27,$FC
	.dc.b	$F7,$F4,$5F,$CA,$AF,$FB,$0B,$6F,$F9,$9C,$3E,$13,$FD,$07,$39,$CB
	.dc.b	$FA,$EC,$E4,$FB,$05,$AF,$D2,$B5,$F9,$AD,$7F,$B5,$FD,$67,$F1,$D7
	.dc.b	$F2,$8D,$EC,$7D,$33,$EB,$7F,$D2,$73,$BD,$4D,$38,$4E,$F0,$FE,$F7
	.dc.b	$1D,$77,$7C,$67,$E0,$73,$FF,$35,$36,$99,$FF,$AA,$F0,$7C,$8F,$F9
	.dc.b	$EA,$B5,$3F,$85,$6E,$1D,$38,$BF,$95,$DF,$0C,$3F,$C1,$F3,$96,$FD
	.dc.b	$4F,$F8,$B5,$BE,$EF,$FA,$1B,$EA,$BF,$2D,$1F,$E5,$19,$D6,$5F,$E0
	.dc.b	$DB,$46,$34,$FF,$41,$36,$BF,$CB,$17,$CA,$FF,$35,$39,$7F,$39,$1F
	.dc.b	$A8,$7F,$CC,$2D,$78,$F3,$FF,$97,$E3,$11,$EF,$F8,$9F,$84,$79,$FE
	.dc.b	$B6,$E1,$EE,$DB,$F9,$45,$6E,$3E,$0D,$AB,$73,$FE,$27,$3E,$78,$F8
	.dc.b	$FF,$E8,$AE,$1A,$FD,$46,$BB,$3F,$A5,$FD,$63,$EA,$9F,$5B,$CA,$30
	.dc.b	$D8,$CC,$F0,$6B,$FE,$49,$9F,$F0,$BE,$DC,$7D,$7F,$F3,$2C,$F6,$6B
	.dc.b	$B6,$BF,$E8,$7D,$1F,$49,$EC,$AF,$C6,$B7,$5D,$36,$BE,$9F,$B3,$78
	.dc.b	$BF,$2B,$37,$0E,$75,$F3,$63,$5B,$FF,$8C,$23,$E6,$FD,$6F,$C8,$B4
	.dc.b	$A3,$0D,$9F,$F4,$C3,$5F,$F3,$44,$FE,$33,$E1,$7C,$5C,$9F,$E6,$99
	.dc.b	$FC,$BC,$78,$2F,$2C,$8C,$9E,$23,$F3,$FF,$02,$FA,$BF,$63,$63,$31
	.dc.b	$F0,$5F,$F2,$2F,$C8,$6A,$24,$68,$FF,$3B,$9A,$3F,$CE,$BD,$FF,$EC
	.dc.b	$B6,$FC,$67,$CB,$EF,$4E,$67,$FC,$E8,$FF,$0A,$F8,$DD,$39,$E3,$FF
	.dc.b	$3A,$B7,$8F,$FC,$FA,$DA,$FE,$05,$5D,$7E,$B7,$FC,$DF,$1F,$D5,$5A
	.dc.b	$FF,$80,$9A,$FF,$4A,$6B,$FE,$88,$F0,$FF,$2C,$FF,$2E,$E8,$F4,$46
	.dc.b	$23,$FC,$C3,$3D,$31,$FE,$8B,$E7,$7E,$93,$3F,$E7,$4D,$F1,$CD,$25
	.dc.b	$FF,$69,$8C,$B6,$32,$FF,$86,$3C,$87,$D6,$FC,$96,$C7,$FA,$89,$AF
	.dc.b	$F9,$D2,$B9,$BF,$E7,$6E,$67,$D6,$79,$5B,$1F,$DE,$5A,$FF,$6E,$8D
	.dc.b	$11,$D3,$5F,$F1,$5E,$3F,$87,$7F,$D0,$6D,$CB,$FD,$56,$DE,$0F,$EE
	.dc.b	$95,$F8,$1F,$31,$F3,$7F,$1D,$B8,$7F,$82,$DA,$FF,$82,$A3,$44,$74
	.dc.b	$D7,$FC,$7B,$E7,$BE,$A3,$F1,$CD,$8F,$F6,$3B,$5F,$F0,$EC,$FE,$4F
	.dc.b	$EB,$7F,$D3,$31,$FE,$1B,$6B,$FE,$48,$6B,$FE,$10,$6B,$FE,$CE,$6B
	.dc.b	$FE,$37,$6E,$46,$98,$E0,$7C,$97,$E8,$DB,$5F,$ED,$C6,$BF,$EA,$E9
	.dc.b	$F1,$EB,$FE,$62,$FA,$DF,$C7,$7B,$06,$92,$BF,$CC,$2D,$7F,$CC,$31
	.dc.b	$B7,$27,$F8,$E7,$D9,$FC,$0D,$7F,$B8,$9B,$C1,$FE,$EA,$6C,$FF,$94
	.dc.b	$EB,$93,$FB,$8F,$6D,$F2,$5F,$D1,$3A,$FF,$D0,$AD,$8C,$CF,$FA,$32
	.dc.b	$72,$D8,$CB,$FE,$8D,$6B,$FE,$8C,$6B,$FE,$8A,$9F,$D7,$7D,$63,$EB
	.dc.b	$7F,$86,$36,$7F,$C0,$B9,$FF,$BF,$23,$11,$E7,$AF,$E7,$98,$7F,$42
	.dc.b	$F8,$DB,$E3,$FF,$BB,$AB,$FD,$62,$D7,$F8,$AF,$C2,$79,$B1,$6D,$FE
	.dc.b	$BE,$F4,$3F,$80,$7C,$29,$A4,$BF,$F3,$37,$CF,$7F,$17,$E6,$34,$FE
	.dc.b	$35,$A7,$FB,$33,$EB,$7F,$A7,$7D,$37,$F2,$5B,$FE,$E3,$DC,$8C,$47
	.dc.b	$F8,$8A,$31,$E3,$7F,$4D,$9C,$CF,$F9,$95,$B3,$A8,$FF,$5E,$B5,$FE
	.dc.b	$9B,$1C,$31,$7B,$DF,$EC,$A9,$FF,$3A,$FC,$2F,$F3,$98,$C4,$7F,$87
	.dc.b	$EB,$D9,$F1,$EF,$27,$F9,$F3,$67,$E9,$F0,$7F,$8E,$31,$4E,$65,$C3
	.dc.b	$F3,$1F,$F5,$04,$3E,$63,$9E,$12,$79,$9A,$7B,$C7,$F4,$DF,$80,$FF
	.dc.b	$7A,$4E,$B0,$FF,$32,$7C,$66,$4D,$6F,$67,$55,$C1,$F1,$E7,$E7,$8C
	.dc.b	$67,$E5,$C3,$FC,$DD,$7F,$5B,$C4,$98,$7F,$92,$B5,$FE,$1D,$7E,$BF
	.dc.b	$92,$7D,$6B,$D0,$DD,$FC,$9C,$2D,$EE,$B5,$3E,$9F,$96,$FF,$7B,$B7
	.dc.b	$E2,$FF,$B5,$46,$23,$91,$EF,$FA,$DF,$F6,$03,$7E,$27,$FB,$84,$62
	.dc.b	$3D,$A7,$FB,$35,$FD,$2E,$1E,$D7,$4D,$9E,$27,$2F,$0B,$FB,$D7,$D1
	.dc.b	$BE,$B6,$93,$0F,$A1,$0F,$8B,$C4,$78,$38,$93,$C6,$37,$5D,$2B,$FC
	.dc.b	$61,$E9,$3E,$B7,$FC,$55,$5F,$E5,$21,$19,$F7,$FF,$CA,$6D,$8D,$1A
	.dc.b	$9E,$2F,$F9,$6A,$34,$F3,$6B,$98,$DB,$FE,$8E,$8C,$BF,$E9,$0E,$77
	.dc.b	$D6,$FC,$67,$0D,$8F,$E0,$AD,$9E,$FB,$7E,$27,$EB,$5F,$31,$FE,$23
	.dc.b	$6D,$FE,$5D,$AF,$C7,$7F,$A0,$E1,$FA,$97,$35,$CE,$B9,$1C,$FE,$67
	.dc.b	$FC,$5D,$DB,$FC,$4F,$0F,$D1,$35,$FA,$4D,$AF,$D8,$78,$5F,$C3,$7E
	.dc.b	$B7,$8F,$8F,$2F,$8B,$FC,$52,$3F,$8A,$E7,$FC,$3C,$93,$FC,$33,$B7
	.dc.b	$F7,$4F,$F0,$C8,$CB,$4D,$3F,$E2,$7F,$A2,$DE,$3F,$AD,$6C,$7F,$07
	.dc.b	$6C,$FF,$1B,$8F,$F2,$54,$3E,$32,$92,$7F,$D7,$7C,$6F,$A0,$C4,$7E
	.dc.b	$46,$B8,$3F,$C1,$5A,$FF,$AE,$67,$83,$FC,$27,$E9,$1F,$5B,$FE,$59
	.dc.b	$6C,$66,$3F,$D0,$9F,$5A,$FA,$1D,$44,$9D,$1F,$EC,$E9,$D3,$D7,$E5
	.dc.b	$3F,$C9,$17,$FE,$DA,$8F,$CC,$F6,$23,$11,$FA,$18,$C4,$FF,$A2,$FE
	.dc.b	$B3,$F5,$D3,$FE,$DA,$6B,$FE,$D3,$6E,$FF,$C6,$6F,$9B,$FC,$6F,$EA
	.dc.b	$5F,$5B,$FE,$75,$6D,$FF,$29,$6B,$CF,$5D,$4C,$8D,$3F,$EA,$5C,$3B
	.dc.b	$F0,$FF,$3F,$B7,$D6,$FF,$99,$5B,$7E,$E6,$BE,$D6,$BF,$ED,$0D,$11
	.dc.b	$FE,$44,$CB,$D1,$F0,$3E,$73,$D1,$7F,$EB,$3E,$CF,$F5,$77,$C8,$7F
	.dc.b	$06,$C7,$FA,$FB,$5F,$F5,$5B,$5F,$ED,$AF,$F0,$DB,$FD,$1B,$72,$FF
	.dc.b	$64,$4F,$53,$FD,$D4,$7F,$B0,$48,$7F,$BB,$6A,$BF,$CB,$BA,$55,$CE
	.dc.b	$11,$87,$A1,$FC,$97,$E0,$BF,$52,$FC,$16,$BF,$DA,$5A,$B4,$FF,$25
	.dc.b	$E1,$FE,$91,$9F,$F2,$4E,$3F,$D5,$DB,$1C,$5F,$F5,$91,$FA,$CA,$FE
	.dc.b	$67,$F3,$7F,$DF,$FE,$57,$FB,$84,$0C,$FA,$BF,$EF,$DE,$EC,$46,$04
	.dc.b	$F0,$F5,$DF,$01,$FC,$6A,$39,$CE,$DB,$E4,$31,$F0,$61,$D2,$31,$8F
	.dc.b	$7B,$FF,$63,$E1,$CE,$B5,$7C,$27,$CD,$7C,$1F,$1C,$EA,$F9,$63,$E4
	.dc.b	$F0,$FD,$39,$78,$B1,$E5,$87,$97,$0E,$0F,$23,$C3,$5C,$23,$35,$C7
	.dc.b	$0F,$35,$71,$36,$63,$1F,$24,$EE,$3A,$1B,$C7,$C8,$EF,$1F,$1D,$70
	.dc.b	$C3,$B9,$F4,$EB,$75,$CA,$74,$FC,$2A,$E2,$7C,$59,$DD,$F8,$49,$7D
	.dc.b	$2B,$CE,$73,$AE,$38,$78,$CF,$93,$97,$D1,$9E,$D1,$97,$83,$AC,$BE
	.dc.b	$34,$68,$C9,$30,$FC,$35,$73,$8F,$C5,$1F,$DF,$72,$FB,$4D,$C3,$FC
	.dc.b	$9E,$FD,$43,$59,$FE,$E5,$FC,$0B,$5F,$F0,$6B,$67,$FD,$5F,$C3,$FE
	.dc.b	$04,$9C,$BF,$56,$7F,$83,$61,$F9,$9F,$65,$F5,$BF,$E5,$38,$FC,$BC
	.dc.b	$EB,$5F,$CB,$37,$0F,$0E,$1F,$8E,$C3,$8E,$FD,$62,$4F,$D7,$AE,$1B
	.dc.b	$FF,$A0,$B2,$ED,$8F,$4B,$DE,$5E,$07,$A4,$EB,$0E,$E7,$38,$DD,$7E
	.dc.b	$17,$2F,$0C,$E4,$97,$CD,$FC,$E6,$1E,$1C,$B8,$C7,$49,$D1,$F0,$F1
	.dc.b	$F0,$47,$C8,$39,$E1,$E7,$79,$E1,$E0,$AE,$33,$96,$CB,$0F,$03,$C8
	.dc.b	$F0,$4E,$CF,$55,$F8,$33,$FB,$33,$B3,$C5,$7E,$5C,$7D,$31,$8B,$E4
	.dc.b	$F7,$C7,$C9,$18,$E1,$FF,$78,$EB,$FE,$8F,$E1,$FE,$4C,$DC,$3F,$97
	.dc.b	$7C,$2D,$C7,$F9,$76,$FF,$C5,$E7,$FD,$18,$A7,$F9,$AE,$1F,$EF,$C6
	.dc.b	$C6,$67,$FB,$DB,$6B,$FB,$D3,$F5,$B0,$FE,$6D,$AF,$7C,$12,$F5,$B0
	.dc.b	$EF,$0F,$E8,$55,$83,$33,$63,$CB,$17,$0F,$E9,$93,$70,$FE,$88,$F9
	.dc.b	$89,$5E,$67,$D7,$D0,$FA,$DF,$23,$67,$FC,$09,$E3,$E2,$72,$F0,$24
	.dc.b	$7F,$7D,$79,$FD,$6F,$23,$E6,$74,$3E,$0B,$FD,$49,$5F,$E0,$F9,$D3
	.dc.b	$4F,$F5,$FC,$EF,$D7,$CE,$9F,$DB,$6F,$8B,$63,$CF,$CF,$DA,$73,$AF
	.dc.b	$F8,$7E,$3D,$A4,$C3,$FC,$24,$FD,$AF,$05,$F5,$5F,$4D,$B3,$E2,$FA
	.dc.b	$DF,$ED,$F9,$7F,$A6,$E7,$2D,$AE,$07,$BA,$77,$87,$F8,$C6,$3E,$F7
	.dc.b	$90,$F6,$E9,$F1,$3E,$B7,$FD,$D5,$97,$F9,$56,$32,$D3,$E1,$C6,$B1
	.dc.b	$FF,$2E,$4F,$4E,$59,$5F,$23,$EA,$BF,$D6,$4F,$F9,$7E,$3E,$8C,$62
	.dc.b	$BF,$D5,$9C,$08,$C4,$7F,$7E,$8C,$78,$FF,$E1,$26,$DF,$FA,$B9,$B3
	.dc.b	$CA,$B8,$BF,$E0,$98,$CB,$CE,$32,$FF,$86,$7A,$FC,$5A,$FF,$BB,$5B
	.dc.b	$7C,$69,$F4,$1F,$A5,$9C,$DF,$1C,$70,$7B,$26,$DF,$FA,$83,$E9,$B8
	.dc.b	$1C,$7C,$DB,$19,$25,$FF,$A9,$23,$2F,$FB,$9A,$3F,$D1,$DD,$DF,$C0
	.dc.b	$DA,$7C,$DE,$5C,$79,$FF,$C1,$6B,$FC,$81,$A7,$E7,$6B,$E3,$36,$3E
	.dc.b	$35,$7C,$B8,$C4,$7C,$58,$C1,$27,$8B,$FC,$3B,$EB,$5F,$33,$FC,$1D
	.dc.b	$B3,$FE,$37,$DC,$FF,$8E,$7C,$04,$FF,$A8,$E3,$46,$5E,$38,$7F,$BD
	.dc.b	$5B,$5C,$A3,$FC,$9D,$18,$9E,$1E,$8F,$87,$0F,$E3,$1A,$FE,$8B,$0E
	.dc.b	$18,$FF,$A8,$F9,$F3,$F4,$3A,$66,$3E,$2C,$FF,$CB,$58,$71,$8F,$42
	.dc.b	$AD,$CB,$0E,$BE,$F3,$E6,$BF,$94,$37,$0F,$D0,$61,$89,$FC,$FB,$23
	.dc.b	$C1,$F0,$6F,$AD,$7C,$6E,$72,$FB,$37,$72,$F8,$57,$BA,$C6,$1F,$37
	.dc.b	$EB,$5E,$FF,$F3,$96,$C7,$B7,$74,$3E,$83,$55,$2B,$53,$C3,$3F,$74
	.dc.b	$75,$AD,$46,$0F,$F4,$07,$C5,$3D,$EF,$E9,$0D,$AF,$02,$F4,$F3,$32
	.dc.b	$72,$F0,$8C,$57,$5C,$9F,$17,$F4,$E3,$FD,$1C,$DF,$03,$F8,$37,$8F
	.dc.b	$FD,$5E,$31,$D6,$8F,$CE,$13,$2F,$D0,$B6,$FF,$83,$35,$FE,$08,$FF
	.dc.b	$6A,$FA,$CF,$E3,$9C,$5F,$8A,$6D,$7F,$98,$7E,$B5,$F1,$9F,$D9,$E7
	.dc.b	$38,$7F,$A8,$FE,$3F,$F8,$6E,$A6,$93,$C4,$FF,$55,$FA,$EF,$AC,$F4
	.dc.b	$67,$F9,$38,$C6,$BF,$E0,$15,$F2,$2F,$AD,$C5,$F0,$CF,$E4,$78,$EF
	.dc.b	$73,$FB,$D3,$72,$E9,$F2,$AF,$8F,$CD,$75,$8C,$4F,$FA,$B2,$FF,$CF
	.dc.b	$D1,$87,$19,$A3,$11,$FE,$B2,$8C,$5F,$B2,$73,$3C,$A3,$04,$AE,$D1
	.dc.b	$FE,$C9,$F9,$67,$AD,$FE,$11,$6C,$FE,$67,$DF,$7D,$6E,$6B,$FC,$2D
	.dc.b	$18,$9E,$7C,$7F,$EC,$0A,$FF,$11,$CE,$70,$F6,$4E,$6B,$73,$99,$FF
	.dc.b	$7F,$35,$FE,$30,$D7,$FD,$BC,$D9,$FF,$18,$7D,$6F,$F8,$9F,$DE,$3F
	.dc.b	$E2,$9F,$AD,$FA,$9D,$5F,$C4,$A7,$38,$FE,$CE,$7E,$C6,$3F,$ED,$AC
	.dc.b	$11,$88,$FE,$50,$D7,$FD,$DC,$D8,$DF,$2F,$F9,$02,$30,$47,$FC,$87
	.dc.b	$F3,$8F,$93,$FF,$1E,$37,$0F,$E0,$9F,$3B,$FD,$6B,$04,$E7,$1F,$D9
	.dc.b	$4F,$E6,$89,$D3,$FE,$4E,$6D,$FD,$06,$CF,$CE,$7F,$43,$DF,$FC,$A6
	.dc.b	$7F,$89,$36,$7F,$7D,$E0,$7F,$6C,$9C,$F9,$BF,$E5,$18,$C4,$FD,$86
	.dc.b	$BF,$1D,$AF,$F4,$D6,$AC,$91,$86,$FA,$BF,$E6,$68,$C1,$9D,$7F,$90
	.dc.b	$70,$7F,$48,$6E,$3F,$D6,$7E,$7F,$FB,$64,$E7,$C9,$FF,$33,$47,$F6
	.dc.b	$FC,$BE,$D3,$5E,$1E,$B7,$F9,$D2,$31,$1E,$56,$E6,$69,$27,$FC,$F7
	.dc.b	$F4,$5F,$CA,$71,$FE,$74,$DB,$FE,$4D,$1F,$E7,$53,$3E,$3F,$F9,$E5
	.dc.b	$C4,$7E,$81,$AF,$E3,$DA,$FF,$75,$6B,$F2,$7E,$B3,$F8,$DE,$27,$F8
	.dc.b	$EF,$73,$EB,$1F,$5B,$FE,$87,$9D,$93,$E7,$7F,$8C,$4E,$B1,$FE,$A5
	.dc.b	$3F,$C9,$B8,$5F,$19,$F7,$B8,$7F,$BF,$46,$38,$F8,$FC,$8F,$F9,$DA
	.dc.b	$B5,$87,$F8,$ED,$BB,$7D,$E6,$EB,$E5,$AE,$17,$FE,$7E,$FA,$97,$BD
	.dc.b	$4F,$F4,$0F,$D5,$7E,$4F,$4F,$0C,$EB,$3F,$F3,$FB,$73,$FE,$C1,$A2
	.dc.b	$31,$5E,$8B,$E5,$7F,$E4,$F9,$CB,$FE,$51,$8F,$D0,$5F,$F9,$6D,$AE
	.dc.b	$7A,$3F,$CB,$91,$88,$F6,$78,$7F,$04,$F3,$FD,$6D,$BF,$F4,$9E,$FF
	.dc.b	$C9,$AB,$CA,$DC,$BC,$95,$B9,$FE,$F0,$09,$F1,$FF,$CC,$9C,$07,$0E
	.dc.b	$8A,$EF,$EB,$1F,$54,$FA,$DC,$36,$BF,$9E,$CE,$A7,$27,$1F,$63,$FC
	.dc.b	$55,$3A,$6B,$FD,$46,$7F,$C2,$D8,$EC,$FF,$3F,$7E,$DA,$F9,$CF,$47
	.dc.b	$12,$31,$7A,$9C,$C6,$23,$FD,$1A,$D7,$FD,$18,$D7,$FD,$0C,$D7,$F8
	.dc.b	$4B,$79,$BF,$C1,$9D,$5F,$1D,$B3,$FD,$69,$AF,$F6,$6F,$07,$F4,$AF
	.dc.b	$A3,$BF,$C6,$47,$04,$3C,$AE,$5A,$7F,$66,$D5,$F9,$13,$E2,$3E,$C4
	.dc.b	$61,$B1,$99,$E0,$D7,$FB,$E4,$ED,$CC,$F0,$23,$EB,$ED,$FF,$23,$37
	.dc.b	$83,$E0,$46,$23,$FC,$06,$D7,$FC,$04,$D7,$F1,$D9,$7F,$07,$9E,$24
	.dc.b	$8F,$91,$F6,$3F,$E5,$86,$F8,$9F,$36,$31,$1F,$E1,$06,$BF,$C6,$79
	.dc.b	$FF,$31,$E0,$FF,$A3,$9A,$7F,$A1,$E3,$F9,$03,$F5,$F8,$5F,$5F,$49
	.dc.b	$3F,$66,$32,$D8,$CB,$FE,$20,$8C,$B6,$3C,$3E,$2B,$E5,$69,$27,$FC
	.dc.b	$27,$19,$7F,$5D,$F1,$5F,$A8,$6F,$4F,$EB,$A3,$11,$FE,$24,$6B,$FE
	.dc.b	$33,$6B,$FB,$66,$BF,$BB,$6B,$FE,$AE,$8F,$D2,$FC,$FF,$FB,$F6,$3F
	.dc.b	$DC,$CD,$7F,$DD,$ED,$7F,$DD,$CD,$7B,$7C,$FF,$94,$8F,$A9,$AF,$9D
	.dc.b	$AF,$9D,$FE,$19,$CC,$FA,$2F,$B3,$3F,$C4,$63,$1C,$1A,$9C,$CF,$F9
	.dc.b	$E5,$BC,$4F,$D4,$BF,$89,$B6,$7F,$65,$18,$23,$82,$1F,$B5,$7F,$8D
	.dc.b	$F0,$BE,$B7,$FD,$29,$3F,$E4,$86,$BF,$E4,$66,$BF,$E3,$56,$BF,$9C
	.dc.b	$FC,$3B,$FD,$C2,$FF,$01,$6B,$FE,$66,$6B,$FE,$65,$6B,$C7,$E3,$FF
	.dc.b	$85,$46,$21,$82,$30,$D9,$FC,$FF,$AF,$FC,$32,$BF,$CB,$6D,$7F,$CB
	.dc.b	$3D,$BA,$3F,$8B,$7E,$CC,$F9,$9A,$49,$FE,$31,$19,$6C,$65,$FF,$3F
	.dc.b	$C6,$5B,$1F,$C8,$7C,$57,$D3,$D2,$4F,$F2,$38,$CB,$FC,$97,$E0,$DF
	.dc.b	$61,$FC,$A5,$B1,$98,$C1,$9F,$AE,$91,$C5,$AF,$FA,$49,$FA,$6F,$C8
	.dc.b	$6E,$A6,$9A,$87,$D1,$7F,$98,$DF,$EC,$0D,$7F,$D2,$6D,$7F,$D4,$4D
	.dc.b	$7F,$B2,$B5,$FE,$D4,$D7,$AF,$B7,$FD,$A2,$35,$E4,$3F,$3B,$FF,$4E
	.dc.b	$D7,$FA,$82,$31,$E7,$7F,$5A,$9C,$CF,$FA,$C6,$72,$FF,$A7,$3E,$39
	.dc.b	$F6,$9F,$C8,$1B,$19,$8C,$11,$C1,$0F,$F5,$0C,$70,$7F,$D4,$7E,$0B
	.dc.b	$EB,$3F,$95,$B6,$3F,$D5,$6D,$77,$1A,$23,$A6,$BF,$EC,$6F,$0B,$FD
	.dc.b	$CA,$FF,$65,$6B,$9F,$8D,$FE,$D5,$18,$8F,$F6,$53,$5F,$E9,$2D,$7F
	.dc.b	$4A,$DD,$FF,$CD,$13,$F2,$7D,$79,$5F,$37,$44,$62,$3F,$D8,$3F,$5B
	.dc.b	$FC,$E9,$FF,$58,$35,$FF,$12,$76,$A3,$11,$FE,$8B,$6C,$67,$C6,$FE
	.dc.b	$A3,$18,$23,$FD,$53,$85,$F4,$74,$97,$FE,$D6,$8C,$B6,$32,$FF,$B5
	.dc.b	$A3,$2D,$8F,$C6,$6B,$23,$F3,$2D,$7F,$32,$D7,$F3,$2D,$7F,$C8,$1C
	.dc.b	$3F,$EC,$3C,$BF,$C5,$3C,$7F,$1A,$7F,$28,$D7,$F2,$4D,$7F,$DA,$DF
	.dc.b	$58,$F6,$FF,$B8,$B6,$3F,$DE,$2D,$7F,$DD,$ED,$78,$C7,$02,$3C,$1F
	.dc.b	$D1,$7D,$5C,$8F,$B4,$D7,$F4,$9B,$A3,$11,$FE,$58,$C7,$F5,$D1,$88
	.dc.b	$FF,$2F,$7C,$67,$EC,$7E,$97,$FC,$BC,$D8,$FF,$2F,$9A,$B4,$E9,$B1
	.dc.b	$A1,$3C,$5E,$2D,$7F,$CD,$2D,$7A,$4F,$CE,$E9,$91,$F3,$9A,$F8,$B8
	.dc.b	$9F,$46,$F8,$AF,$F0,$EB,$67,$C5,$3A,$9C,$92,$33,$3B,$7A,$35,$FF
	.dc.b	$3E,$CF,$17,$9F,$D2,$FF,$70,$83,$49,$E5,$7F,$8B,$9B,$3C,$23,$07
	.dc.b	$F9,$CB,$F0,$27,$CD,$52,$4F,$F1,$F8,$CB,$E0,$FB,$EE,$84,$70,$7F
	.dc.b	$25,$FA,$37,$C7,$FF,$9F,$9B,$19,$8C,$11,$C1,$0F,$E5,$51,$C1,$FE
	.dc.b	$59,$F8,$07,$F9,$8D,$FC,$43,$5F,$D2,$B5,$71,$A3,$1A,$3E,$A7,$F4
	.dc.b	$95,$FC,$21,$AA,$35,$D1,$FE,$56,$9C,$DB,$F0,$1F,$F5,$6B,$6F,$FD
	.dc.b	$37,$18,$3F,$C2,$FF,$4F,$FE,$5B,$8F,$B2,$D7,$F8,$BB,$5F,$03,$70
	.dc.b	$34,$9E,$D7,$F9,$A1,$B3,$F8,$C6,$BF,$93,$7A,$7A,$AF,$C0,$BF,$C6
	.dc.b	$AD,$E2,$FF,$1E,$C6,$0F,$F1,$CE,$5F,$EF,$CF,$A3,$FF,$5A,$FC,$41
	.dc.b	$A4,$C7,$FB,$8F,$C2,$7E,$B1,$FF,$68,$7C,$17,$F6,$E8,$C7,$C4,$7F
	.dc.b	$A6,$7C,$54,$62,$7F,$D1,$BE,$B7,$F8,$E2,$B5,$39,$BF,$EE,$AD,$7F
	.dc.b	$C0,$2E,$7C,$A5,$71,$BF,$07,$81,$FE,$E4,$6D,$F9,$0E,$71,$97,$FC
	.dc.b	$6F,$3C,$5C,$9B,$3B,$C6,$DA,$73,$AF,$0F,$C7,$FF,$BC,$31,$F4,$D6
	.dc.b	$F2,$F7,$B7,$2F,$46,$3D,$F1,$E7,$7C,$1F,$05,$F0,$7B,$4E,$E3,$13
	.dc.b	$B2,$65,$F2,$B1,$F1,$3F,$02,$33,$7E,$C7,$FD,$99,$97,$A2,$3C,$99
	.dc.b	$47,$C3,$97,$9B,$0E,$71,$C7,$0E,$84,$7E,$4C,$69,$F5,$C6,$5A,$68
	.dc.b	$ED,$87,$91,$CE,$1E,$37,$C3,$87,$E3,$08,$FF,$8F,$30,$FC,$73,$FE
	.dc.b	$3C,$E1,$70,$FF,$90,$9B,$1F,$E4,$26,$BB,$8D,$1F,$56,$37,$87,$C4
	.dc.b	$FA,$D7,$D6,$B9,$35,$73,$0F,$AB,$87,$0D,$3F,$DC,$B8,$70,$BD,$63
	.dc.b	$F9,$FA,$D5,$FF,$93,$9B,$A2,$F8,$4F,$4C,$78,$E3,$FE,$56,$BD,$D6
	.dc.b	$B0,$E9,$5A,$BE,$8F,$E6,$F0,$E9,$97,$A6,$7A,$12,$B8,$BF,$99,$BE
	.dc.b	$98,$F7,$C3,$A4,$FF,$9B,$B1,$F0,$E1,$CF,$0E,$24,$BE,$15,$CE,$75
	.dc.b	$18,$9D,$4F,$98,$D6,$5D,$B6,$FE,$3B,$8F,$AF,$3F,$4C,$E7,$2F,$93
	.dc.b	$7E,$3A,$F7,$67,$E8,$8E,$98,$F6,$74,$76,$3E,$06,$5F,$CB,$BE,$0F
	.dc.b	$F9,$87,$37,$F2,$C6,$E2,$69,$FC,$D7,$91,$F7,$8F,$47,$0D,$D7,$FD
	.dc.b	$04,$DC,$BF,$D0,$0D,$7F,$9F,$F7,$76,$24,$FF,$37,$6C,$7F,$37,$6B
	.dc.b	$C2,$34,$47,$4D,$7F,$C5,$1D,$5F,$CD,$63,$FD,$28,$D8,$FF,$49,$35
	.dc.b	$DF,$3F,$C2,$EB,$7C,$DD,$25,$FF,$85,$A3,$2F,$FA,$7F,$B9,$F0,$5F
	.dc.b	$15,$B1,$98,$C1,$F2,$7C,$0F,$85,$8F,$F9,$61,$AF,$F7,$46,$EC,$69
	.dc.b	$36,$E4,$4B,$F2,$57,$A4,$8E,$0F,$F0,$B7,$5F,$D9,$F0,$8D,$24,$7F
	.dc.b	$8A,$3D,$9F,$C1,$F8,$A6,$92,$34,$4B,$FF,$1F,$4E,$E3,$11,$FE,$34
	.dc.b	$F1,$9F,$87,$FE,$39,$B3,$FE,$3C,$8C,$11,$C1,$27,$FC,$7F,$1F,$91
	.dc.b	$F4,$7F,$E0,$04,$FF,$B3,$DB,$3F,$EC,$D6,$BF,$E6,$96,$BF,$E6,$8F
	.dc.b	$4B,$FD,$33,$3F,$E6,$1F,$09,$18,$9F,$F3,$0B,$5F,$F3,$0B,$5E,$55
	.dc.b	$FE,$17,$9E,$2D,$9D,$4F,$F9,$B1,$BB,$BC,$87,$D5,$FE,$AD,$B8,$FF
	.dc.b	$BA,$5B,$7F,$EE,$76,$BF,$E8,$36,$BF,$EA,$06,$BF,$E3,$C6,$D7,$7A
	.dc.b	$FD,$EF,$6F,$F4,$4B,$FF,$37,$B6,$FF,$CE,$CD,$7F,$CD,$6D,$7F,$D9
	.dc.b	$2D,$7F,$C8,$4D,$ED,$34,$D5,$FF,$B1,$BC,$3F,$B0,$4B,$FB,$0F,$F8
	.dc.b	$F7,$A5,$E6,$52,$4F,$FA,$57,$1F,$F1,$74,$61,$FF,$4C,$4F,$09,$D3
	.dc.b	$63,$33,$FC,$73,$E1,$FF,$03,$1F,$E5,$96,$E4,$69,$3D,$EE,$04,$C3
	.dc.b	$FD,$2B,$1C,$30,$FC,$F5,$FF,$A7,$5A,$FF,$A4,$9A,$FF,$BC,$1A,$FF
	.dc.b	$9A,$5B,$7F,$E4,$9E,$4F,$E6,$31,$83,$79,$FE,$1F,$4F,$C3,$13,$1F
	.dc.b	$C3,$3F,$E6,$7F,$17,$FC,$E9,$F5,$97,$13,$49,$5F,$EB,$F9,$DB,$67
	.dc.b	$6D,$9D,$E5,$FE,$B9,$FA,$8F,$E3,$6D,$EC,$FE,$8F,$18,$8F,$F5,$8B
	.dc.b	$51,$FE,$87,$70,$D8,$E2,$4A,$FF,$60,$FB,$6F,$9E,$FE,$A6,$D8,$CC
	.dc.b	$60,$8E,$09,$5F,$EC,$F9,$E0,$FE,$A7,$E6,$FF,$5C,$49,$FD,$73,$F9
	.dc.b	$1F,$80,$91,$F8,$A6,$CF,$E3,$9A,$FF,$BB,$37,$46,$23,$DF,$AF,$E6
	.dc.b	$BD,$8F,$A9,$C8,$69,$30,$FA,$5F,$02,$F5,$3E,$D3,$63,$31,$83,$3F
	.dc.b	$61,$23,$8B,$71,$FF,$7F,$E3,$FC,$DF,$B3,$F0,$9C,$DF,$C4,$A3,$17
	.dc.b	$F8,$18,$C7,$C2,$7E,$02,$F8,$B5,$FB,$22,$91,$98,$F7,$B5,$D3,$63
	.dc.b	$7C,$9F,$95,$8C,$1B,$E5,$FF,$4A,$FA,$FF,$E1,$0F,$53,$FD,$DB,$97
	.dc.b	$83,$1F,$C4,$D6,$A3,$11,$FE,$D1,$9D,$BF,$03,$E1,$DF,$6F,$FA,$36
	.dc.b	$D7,$E7,$DB,$3B,$9D,$12,$74,$D8,$F1,$5B,$E8,$3F,$37,$5E,$28,$C4
	.dc.b	$FF,$8A,$27,$2F,$06,$BF,$2B,$47,$C5,$BE,$8F,$F7,$0D,$C3,$F6,$8D
	.dc.b	$AF,$B6,$D7,$93,$5F,$C7,$B5,$FA,$CD,$AE,$F5,$FE,$37,$FA,$9F,$F0
	.dc.b	$55,$FC,$AE,$F4,$63,$0F,$A1,$18,$F3,$78,$1F,$CD,$6E,$5F,$F2,$E6
	.dc.b	$D7,$F3,$26,$BF,$A5,$6B,$FC,$0A,$BB,$FB,$1E,$32,$57,$8F,$0F,$E9
	.dc.b	$94,$F8,$5F,$EC,$D3,$FD,$05,$B1,$F8,$26,$BF,$87,$6B,$F9,$86,$BF
	.dc.b	$C4,$9A,$F5,$F1,$FF,$D0,$4F,$17,$FC,$C3,$F0,$EF,$62,$92,$FF,$01
	.dc.b	$19,$7F,$95,$7D,$EB,$C1,$FE,$64,$D8,$CC,$60,$8E,$09,$87,$EE,$27
	.dc.b	$83,$F9,$08,$7A,$FF,$E7,$46,$9F,$91,$2C,$6E,$3F,$A1,$7C,$93,$E4
	.dc.b	$A9,$2B,$FA,$04,$65,$FE,$85,$F2,$9F,$D5,$38,$CD,$27,$6F,$F4,$D6
	.dc.b	$CF,$08,$C1,$F2,$B1,$FE,$7F,$F2,$AF,$DB,$3F,$C7,$B1,$FB,$B8,$C7
	.dc.b	$A7,$B8,$46,$2F,$F7,$AD,$8E,$15,$C9,$FE,$27,$19,$6C,$65,$F0,$B5
	.dc.b	$72,$BE,$61,$F3,$DF,$E9,$96,$E1,$FA,$36,$D7,$E8,$9A,$FF,$22,$AE
	.dc.b	$55,$C4,$95,$C5,$FE,$E5,$1F,$E8,$49,$FE,$66,$D7,$F8,$13,$5F,$E8
	.dc.b	$ED,$7C,$F5,$FC,$AE,$3D,$1E,$17,$9C,$97,$E7,$6D,$79,$DA,$FF,$89
	.dc.b	$FE,$07,$FC,$50,$FF,$72,$6A,$69,$FE,$2D,$F9,$BF,$C4,$67,$78,$7E
	.dc.b	$AD,$7F,$30,$9C,$B6,$32,$FF,$37,$87,$CC,$BE,$6F,$A3,$67,$93,$5E
	.dc.b	$0D,$7F,$83,$7C,$43,$C9,$FE,$A3,$5F,$EF,$7D,$53,$9A,$FF,$06,$CE
	.dc.b	$63,$11,$E1,$D7,$F8,$86,$DF,$ED,$98,$DC,$76,$C3,$AB,$63,$27,$E2
	.dc.b	$2B,$9D,$F8,$6F,$94,$F5,$3E,$BF,$C4,$7F,$95,$71,$FC,$73,$64,$D3
	.dc.b	$5F,$3D,$FE,$34,$AE,$F4,$8C,$47,$58,$C7,$07,$9A,$3B,$9C,$27,$78
	.dc.b	$7C,$56,$E5,$FC,$BD,$B3,$C0,$F4,$78,$BF,$B0,$F9,$6E,$8D,$8F,$82
	.dc.b	$D7,$DD,$C2,$F9,$CF,$F3,$EC,$6A,$15,$C9,$B1,$98,$F4,$47,$1F,$63
	.dc.b	$BF,$53,$E5,$69,$2F,$E8,$C7,$F1,$2E,$3C,$47,$A2,$38,$3E,$66,$BF
	.dc.b	$4D,$E0,$FA,$B4,$FE,$1C,$B3,$46,$1A,$FF,$A3,$5B,$1F,$DF,$7E,$87
	.dc.b	$FD,$C0,$FC,$46,$BF,$29,$BB,$74,$8E,$D5,$FD,$32,$31,$4F,$05,$F0
	.dc.b	$BD,$DB,$7F,$D3,$23,$04,$70,$43,$FC,$27,$1C,$1F,$F0,$AF,$8F,$F5
	.dc.b	$1F,$BA,$DB,$F0,$B7,$0F,$F1,$5D,$71,$9F,$D6,$7C,$EB,$EC,$7F,$34
	.dc.b	$DC,$3F,$C0,$CD,$7F,$C0,$AD,$7F,$92,$35,$FE,$FC,$D7,$F9,$8B,$79
	.dc.b	$4E,$73,$7D,$1B,$E1,$3F,$26,$DB,$FC,$9B,$5F,$D2,$57,$8F,$1F,$A1
	.dc.b	$A3,$E7,$DF,$27,$FE,$C7,$8C,$36,$33,$7F,$E4,$A9,$CB,$63,$2F,$F9
	.dc.b	$46,$32,$D8,$CB,$FE,$64,$8C,$BF,$E6,$6D,$9F,$6E,$F9,$AF,$DC,$46
	.dc.b	$1B,$19,$9F,$E1,$F5,$DB,$1F,$C3,$FD,$03,$E3,$3F,$28,$DA,$FF,$28
	.dc.b	$35,$FF,$2A,$B5,$FF,$3B,$EE,$8C,$47,$F9,$A9,$AF,$E8,$DB,$1B,$9F
	.dc.b	$D1,$79,$7F,$C6,$2B,$F8,$1B,$5F,$E1,$2D,$BF,$3C,$FF,$0A,$F6,$1F
	.dc.b	$3F,$FA,$B6,$DF,$F9,$9D,$AF,$F9,$B1,$AF,$A1,$AF,$FA,$01,$AF,$F9
	.dc.b	$FE,$3A,$1F,$B7,$FA,$77,$C0,$FF,$1A,$6C,$FF,$1A,$6B,$FC,$82,$BC
	.dc.b	$38,$7E,$DF,$A9,$E6,$FF,$18,$AF,$F5,$4B,$5F,$F3,$C3,$5F,$F5,$EB
	.dc.b	$5F,$F5,$23,$63,$9F,$81,$FE,$96,$9C,$B6,$33,$1F,$EF,$1F,$A8,$FF
	.dc.b	$64,$BF,$C8,$1A,$FF,$1B,$6D,$F9,$EF,$CC,$FA,$7E,$07,$DA,$DC,$FF
	.dc.b	$A3,$60,$9C,$F2,$7F,$B5,$1E,$6F,$FA,$B3,$EA,$DF,$57,$FC,$85,$B9
	.dc.b	$FF,$A7,$5A,$FF,$A8,$9A,$FE,$3B,$E6,$7F,$98,$E5,$FC,$79,$B7,$FE
	.dc.b	$B8,$6C,$FF,$B2,$72,$FC,$C5,$7F,$7B,$F9,$E7,$D1,$FF,$69,$6D,$7F
	.dc.b	$A5,$5A,$FF,$A5,$1A,$FF,$8C,$DA,$FF,$A3,$5A,$FF,$90,$1B,$DE,$73
	.dc.b	$CC,$FB,$CF,$8B,$5F,$D1,$5B,$19,$C7,$FB,$85,$F1,$C3,$FB,$3F,$C5
	.dc.b	$BE,$36,$93,$1F,$ED,$30,$FA,$4C,$47,$F7,$08,$E0,$EA,$79,$CF,$F7
	.dc.b	$2F,$1B,$F8,$24,$FF,$76,$6E,$BF,$D2,$9B,$BF,$FB,$91,$7D,$27,$1F
	.dc.b	$02,$5F,$F7,$A9,$DB,$FD,$85,$DB,$FE,$10,$E0,$7C,$5F,$B9,$B1,$98
	.dc.b	$FC,$5F,$91,$F6,$B7,$F9,$51,$B2,$4F,$F7,$89,$FB,$FC,$2F,$B3,$FB
	.dc.b	$3A,$FF,$62,$6E,$BF,$EF,$C6,$DF,$D7,$C3,$FC,$27,$F1,$FF,$CF,$EF
	.dc.b	$FB,$4B,$6D,$5F,$DE,$E3,$0F,$F7,$DF,$83,$7C,$B7,$2F,$2E,$4F,$F8
	.dc.b	$72,$B7,$9F,$FB,$36,$BF,$C4,$3F,$40,$FB,$3A,$4B,$FF,$11,$C3,$E4
	.dc.b	$31,$1F,$E2,$A8,$E0,$EA,$79,$CF,$E2,$38,$FE,$5C,$FF,$BB,$1A,$FF
	.dc.b	$A2,$DA,$FF,$B9,$5A,$FF,$A6,$1B,$C3,$FD,$2F,$7F,$F2,$74,$60,$DE
	.dc.b	$4F,$A6,$FC,$6D,$FF,$8F,$9B,$C1,$FE,$12,$6F,$07,$7C,$3F,$C8,$30
	.dc.b	$F8,$8A,$4B,$FF,$21,$C3,$E4,$B1,$1D,$66,$E1,$D6,$7F,$C9,$7E,$B7
	.dc.b	$F0,$88,$FF,$10,$B7,$1F,$B3,$92,$31,$87,$ED,$FE,$9B,$8F,$85,$23
	.dc.b	$FC,$A7,$3A,$DB,$F7,$55,$F3,$7B,$FE,$19,$2B,$F0,$A7,$CD,$70,$FD
	.dc.b	$16,$B9,$36,$79,$3E,$6F,$FD,$DE,$FF,$98,$9A,$FF,$9A,$1B,$9F,$FA
	.dc.b	$4F,$83,$F8,$3C,$3E,$CF,$FC,$D6,$DA,$E5,$18,$FA,$6C,$B2,$32,$77
	.dc.b	$8F,$F3,$9F,$CF,$7F,$AA,$A3,$FC,$E6,$D7,$FC,$F2,$DC,$3F,$C9,$AF
	.dc.b	$F8,$61,$BE,$D7,$22,$4F,$2F,$5B,$AF,$1F,$F9,$7E,$FF,$01,$3F,$08
	.dc.b	$FF,$33,$38,$7E,$2F,$0A,$31,$1F,$EE,$3F,$91,$F8,$7C,$7F,$B4,$E1
	.dc.b	$46,$27,$FD,$E2,$DB,$E5,$87,$FA,$43,$E8,$77,$5F,$E9,$06,$BF,$E9
	.dc.b	$66,$E9,$FE,$C1,$BE,$51,$8C,$7F,$16,$DA,$FF,$3F,$37,$1F,$1F,$A9
	.dc.b	$DF,$A3,$FC,$FD,$3F,$A8,$23,$8F,$AD,$F2,$12,$FC,$91,$FE,$A6,$F9
	.dc.b	$1F,$F4,$41,$31,$FF,$44,$38,$34,$49,$FF,$55,$BF,$E8,$38,$FD,$0A
	.dc.b	$35,$1F,$EB,$47,$0F,$08,$CE,$7F,$E8,$A9,$CC,$ED,$F2,$4F,$F5,$6F
	.dc.b	$86,$78,$1F,$ED,$36,$D7,$DD,$D3,$FD,$D3,$39,$35,$F2,$7C,$49,$3F
	.dc.b	$ED,$4A,$E2,$FF,$B7,$C7,$81,$FE,$DC,$F0,$BF,$99,$FC,$5F,$FA,$46
	.dc.b	$31,$28,$C7,$C3,$FF,$9E,$DF,$F6,$4F,$A6,$F5,$FF,$CB,$51,$AC,$39
	.dc.b	$F8,$1F,$BA,$C7,$9C,$E7,$1E,$93,$FD,$0B,$0E,$5E,$D7,$F0,$0A,$D1
	.dc.b	$27,$96,$5E,$17,$86,$5E,$07,$39,$78,$31,$FE,$9B,$97,$82,$73,$1E
	.dc.b	$01,$97,$81,$E0,$7C,$4C,$3B,$3C,$9A,$E6,$FB,$4F,$18,$E8,$64,$B7
	.dc.b	$E1,$79,$4F,$0C,$3D,$11,$83,$77,$DA,$B8,$5F,$5C,$7E,$15,$6D,$B1
	.dc.b	$9C,$7D,$B1,$A9,$F1,$3F,$5B,$0E,$F1,$89,$D9,$1F,$87,$1A,$BF,$44
	.dc.b	$6C,$F0,$4F,$D0,$CB,$EB,$3C,$63,$46,$F0,$F2,$3E,$D3,$F9,$AD,$7E
	.dc.b	$31,$B8,$FF,$B0,$DF,$AA,$6B,$81,$EE,$7F,$9C,$6B,$F5,$CD,$8C,$CF
	.dc.b	$F1,$F9,$CB,$63,$39,$FE,$6D,$B8,$FE,$3F,$D4,$7C,$FF,$FB,$FE,$B5
	.dc.b	$8F,$E0,$B1,$E7,$18,$C7,$F1,$19,$B0,$E3,$C7,$BC,$FA,$D7,$E4,$1A
	.dc.b	$FE,$B9,$CD,$71,$CB,$C1,$9F,$E9,$67,$79,$78,$23,$19,$FE,$3B,$2E
	.dc.b	$F8,$FE,$32,$F6,$49,$D6,$1D,$27,$C5,$1E,$52,$3E,$AA,$E7,$9F,$B4
	.dc.b	$98,$78,$27,$57,$CD,$F5,$5F,$2A,$DC,$7E,$AA,$F9,$DF,$19,$CC,$F0
	.dc.b	$39,$E7,$C6,$74,$48,$E5,$87,$7C,$BE,$14,$7A,$8F,$81,$8F,$BA,$F9
	.dc.b	$46,$70,$F1,$E7,$F3,$67,$3D,$7F,$CD,$7E,$7D,$1F,$CE,$9B,$7F,$CD
	.dc.b	$FE,$9A,$F3,$FF,$35,$9B,$5F,$CF,$A7,$75,$FE,$F4,$DD,$18,$8D,$4F
	.dc.b	$F7,$1C,$3F,$A2,$4E,$65,$1F,$D1,$F8,$3A,$C6,$2D,$18,$9F,$E9,$9D
	.dc.b	$76,$73,$87,$F5,$2A,$C1,$99,$B1,$92,$5F,$F5,$38,$CB,$63,$CD,$49
	.dc.b	$F3,$5F,$82,$73,$18,$EB,$7B,$9E,$56,$CF,$F7,$FF,$A4,$C4,$E5,$E0
	.dc.b	$48,$F8,$92,$8E,$9F,$82,$73,$3E,$1F,$67,$F8,$43,$D4,$C5,$FF,$82
	.dc.b	$E6,$E8,$E0,$FF,$05,$56,$1B,$5F,$E8,$C6,$CF,$F8,$44,$EB,$E0,$2B
	.dc.b	$D5,$1E,$D2,$57,$99,$F8,$7E,$13,$E4,$3E,$A3,$67,$FC,$77,$EF,$62
	.dc.b	$64,$7F,$8F,$9D,$C6,$2B,$FC,$A1,$F7,$7F,$77,$CA,$37,$E5,$30,$CC
	.dc.b	$3F,$CB,$0E,$5A,$7F,$96,$F2,$E4,$7C,$8F,$49,$D1,$89,$F8,$47,$F8
	.dc.b	$EE,$32,$FF,$82,$5B,$8F,$F9,$A3,$83,$FD,$49,$5F,$97,$6D,$6E,$31
	.dc.b	$5C,$2B,$FC,$EB,$5A,$79,$46,$5F,$F4,$2F,$73,$EE,$3F,$38,$DC,$3C
	.dc.b	$B2,$BF,$F4,$C7,$D2,$EA,$E5,$7F,$A2,$1F,$C2,$F9,$49,$CD,$FE,$0E
	.dc.b	$BB,$BD,$8E,$FD,$F6,$3F,$D2,$31,$4E,$07,$ED,$31,$FE,$9F,$1A,$3F
	.dc.b	$D4,$FE,$5F,$EC,$FD,$43,$F3,$FC,$3C,$36,$7E,$64,$7C,$62,$38,$25
	.dc.b	$FC,$68,$CB,$FE,$74,$FA,$5F,$E1,$73,$F1,$3B,$91,$89,$F3,$C6,$38
	.dc.b	$3F,$83,$4E,$7A,$7F,$8D,$CF,$0F,$07,$26,$AA,$E1,$FE,$E9,$E8,$FE
	.dc.b	$1D,$E5,$FE,$6F,$2E,$AB,$A7,$33,$2B,$81,$FE,$F1,$8C,$57,$93,$D9
	.dc.b	$FF,$76,$4F,$A5,$B1,$F3,$B9,$7F,$DA,$9C,$3F,$C7,$67,$3A,$FF,$32
	.dc.b	$9C,$DF,$E1,$1B,$7F,$5B,$C4,$FE,$53,$F2,$5F,$82,$C6,$5E,$F9,$BF
	.dc.b	$DA,$4F,$E3,$2B,$6E,$67,$F9,$67,$D2,$BD,$FA,$63,$E0,$2D,$7D,$19
	.dc.b	$D3,$67,$4D,$3E,$B7,$C0,$BE,$8B,$F9,$C3,$63,$3F,$4B,$89,$91,$C1
	.dc.b	$DC,$63,$5F,$BB,$F0,$BF,$C9,$6B,$F0,$4D,$8F,$95,$AB,$E7,$BF,$A3
	.dc.b	$4E,$78,$7E,$54,$E6,$FF,$D0,$CD,$BF,$DF,$35,$FD,$CB,$5F,$F6,$03
	.dc.b	$5F,$DA,$37,$84,$D3,$5D,$0F,$B3,$FA,$35,$F9,$66,$E1,$FD,$BB,$89
	.dc.b	$EE,$BE,$03,$FA,$F5,$6B,$36,$1F,$D9,$BE,$23,$F8,$33,$FE,$D8,$6B
	.dc.b	$A3,$5E,$1F,$F5,$88,$CC,$7F,$A2,$FE,$25,$ED,$FF,$6E,$CD,$18,$C3
	.dc.b	$F0,$CB,$E6,$74,$E6,$FF,$21,$F1,$7E,$8A,$FC,$63,$63,$F2,$9D,$BF
	.dc.b	$9A,$9F,$F4,$D4,$E7,$7F,$81,$3F,$A2,$AF,$46,$3F,$AD,$F2,$3F,$C0
	.dc.b	$8D,$C3,$FD,$F0,$D9,$F7,$1F,$E9,$6A,$E4,$FF,$A9,$DA,$E6,$B9,$C6
	.dc.b	$38,$5F,$19,$F8,$5C,$3E,$6C,$63,$D1,$FF,$07,$7B,$7F,$97,$C7,$F8
	.dc.b	$6D,$ED,$FF,$0B,$C6,$5F,$A1,$5F,$EB,$FE,$AF,$F0,$3C,$E6,$7A,$B5
	.dc.b	$E3,$1A,$26,$1F,$E2,$59,$D5,$7F,$84,$7E,$6F,$FC,$41,$EB,$1F,$F1
	.dc.b	$17,$D1,$7E,$12,$FF,$92,$4E,$5F,$E4,$D1,$9C,$7E,$3D,$A3,$11,$E8
	.dc.b	$6B,$F0,$9A,$FF,$8D,$1B,$1C,$63,$FC,$47,$18,$37,$3E,$5F,$67,$FC
	.dc.b	$72,$D3,$F8,$C7,$D0,$BD,$6C,$E5,$FE,$38,$C3,$F9,$14,$E6,$BF,$DF
	.dc.b	$4D,$8F,$EE,$4D,$7F,$B6,$B5,$FD,$2F,$6A,$31,$1F,$DA,$5B,$19,$BF
	.dc.b	$EF,$51,$AC,$9F,$2B,$FD,$A7,$0F,$E9,$CD,$C3,$CB,$E8,$FF,$94,$63
	.dc.b	$1C,$BF,$CD,$70,$FE,$CB,$5E,$6A,$EA,$4B,$F5,$B6,$35,$18,$3F,$CC
	.dc.b	$90,$F1,$BF,$CB,$CD,$8F,$EC,$1F,$1A,$F7,$33,$5F,$E6,$38,$C4,$FF
	.dc.b	$9B,$5B,$7F,$80,$6B,$F9,$66,$BF,$A0,$6B,$FE,$72,$6E,$66,$98,$F9
	.dc.b	$77,$DA,$FF,$9C,$DB,$7F,$DD,$4F,$F3,$41,$FD,$DD,$CD,$FF,$76,$8C
	.dc.b	$BF,$94,$6B,$FA,$A6,$BF,$B4,$6B,$FE,$86,$FB,$1F,$E4,$8D,$F3,$3F
	.dc.b	$8D,$7D,$0F,$F7,$5A,$FF,$03,$56,$B2,$34,$E3,$DA,$AE,$18,$7F,$82
	.dc.b	$A7,$87,$03,$E9,$7F,$D2,$0D,$C7,$FC,$6B,$F3,$AF,$A2,$7B,$DA,$B9
	.dc.b	$5F,$E1,$7B,$E1,$8F,$F8,$B3,$0E,$17,$A8,$FF,$01,$7D,$06,$AB,$FC
	.dc.b	$00,$DB,$FC,$EF,$D0,$FE,$F6,$B7,$8F,$FA,$12,$B5,$1F,$E4,$76,$F4
	.dc.b	$AB,$86,$1D,$2F,$FD,$17,$1C,$EF,$FC,$AE,$D7,$BF,$57,$F9,$7E,$31
	.dc.b	$1F,$E5,$3F,$8B,$FE,$45,$81,$B2,$F5,$7F,$D2,$F1,$A2,$78,$7E,$08
	.dc.b	$CB,$FE,$BE,$BF,$EB,$1F,$0F,$FC,$F1,$FE,$7A,$D7,$25,$0E,$7E,$8B
	.dc.b	$FD,$25,$C1,$7D,$93,$49,$5D,$65,$F1,$FF,$EA,$78,$C3,$5F,$E9,$B1
	.dc.b	$AD,$FB,$4E,$67,$FB,$73,$FE,$8C,$7C,$CD,$7D,$4D,$7F,$AB,$B5,$E3
	.dc.b	$EE,$7A,$E3,$11,$FD,$F9,$AF,$9A,$3E,$07,$77,$EC,$23,$E2,$35,$F9
	.dc.b	$0D,$7F,$B2,$B5,$FF,$02,$7D,$7B,$E3,$7E,$53,$63,$E7,$35,$FA,$4D
	.dc.b	$7F,$B7,$B5,$E7,$1C,$88,$FF,$BF,$72,$79,$9F,$67,$B3,$E9,$B6,$33
	.dc.b	$3B,$8C,$11,$C1,$0F,$EF,$FE,$13,$E6,$BF,$CE,$51,$F8,$78,$C7,$9B
	.dc.b	$F8,$76,$C6,$67,$6F,$26,$BF,$E2,$78,$DB,$99,$E5,$3A,$8F,$F1,$8C
	.dc.b	$BF,$48,$7C,$07,$76,$CF,$8D,$AF,$A1,$AF,$7F,$C6,$BF,$DD,$CF,$A5
	.dc.b	$AF,$B9,$AE,$FE,$23,$F4,$5F,$71,$FE,$C7,$F9,$AB,$EC,$9A,$4A,$FD
	.dc.b	$54,$ED,$B3,$B6,$CE,$DF,$F6,$F7,$E9,$AF,$A0,$FD,$87,$7A,$31,$3F
	.dc.b	$B5,$8C,$4E,$A7,$33,$F5,$BC,$27,$E0,$BF,$4D,$B1,$F5,$DA,$EE,$34
	.dc.b	$47,$4D,$7F,$C3,$37,$FE,$66,$7F,$83,$72,$7F,$B9,$7D,$FF,$C7,$4E
	.dc.b	$67,$71,$F2,$3F,$0E,$FF,$2D,$47,$C4,$6B,$FE,$4A,$D5,$18,$8F,$F3
	.dc.b	$54,$63,$DC,$E4,$48,$E4,$D7,$FC,$A7,$6F,$DA,$DE,$77,$CA,$6C,$FC
	.dc.b	$96,$BC,$1A,$FF,$9E,$67,$FC,$AB,$E0,$FD,$2F,$6C,$D2,$4F,$D3,$8C
	.dc.b	$B6,$32,$F9,$23,$2F,$F9,$0B,$E9,$3F,$D1,$CF,$CA,$6B,$9F,$82,$FD
	.dc.b	$C4,$62,$3F,$D1,$9F,$32,$FB,$AF,$9D,$18,$6B,$FE,$96,$8D,$6F,$FC
	.dc.b	$E5,$B1,$FC,$F5,$AF,$F9,$E9,$AF,$38,$E1,$3D,$21,$93,$F6,$97,$D5
	.dc.b	$7E,$59,$B3,$F9,$56,$BF,$51,$AF,$CD,$F9,$7F,$E1,$4F,$E6,$DA,$FD
	.dc.b	$A6,$BB,$EF,$FD,$1C,$62,$3F,$C9,$ED,$7C,$AD,$B3,$7F,$5B,$FF,$3C
	.dc.b	$4F,$EA,$5A,$FE,$21,$AF,$E0,$5A,$FF,$97,$3F,$69,$FC,$2B,$78,$BF
	.dc.b	$6B,$18,$8F,$8A,$DD,$5F,$B3,$7F,$41,$94,$63,$DB,$FA,$93,$98,$C4
	.dc.b	$FE,$F7,$F6,$7F,$EF,$4D,$F4,$7F,$81,$C6,$23,$F3,$ED,$7F,$D6,$7F
	.dc.b	$8D,$7F,$75,$4D,$23,$FD,$E5,$AF,$F7,$D6,$BF,$AA,$FD,$7B,$FD,$D6
	.dc.b	$FF,$0F,$6B,$9F,$8E,$FE,$2B,$18,$8F,$F5,$8F,$C9,$3E,$67,$FA,$83
	.dc.b	$63,$F8,$3B,$5F,$F5,$1B,$5E,$31,$C0,$8F,$07,$FA,$D7,$8C,$FA,$AA
	.dc.b	$49,$FE,$27,$19,$6C,$65,$FF,$58,$C6,$5B,$1F,$E1,$BF,$C1,$5F,$7F
	.dc.b	$49,$3F,$E2,$18,$CB,$63,$2F,$F8,$9E,$32,$D8,$F8,$BD,$92,$3E,$53
	.dc.b	$5F,$F6,$1F,$98,$F9,$5F,$EE,$CD,$8F,$E5,$AD,$77,$1A,$23,$A6,$BF
	.dc.b	$EC,$EF,$9B,$7C,$07,$F8,$05,$B1,$FC,$E1,$AE,$E3,$44,$74,$D7,$FC
	.dc.b	$89,$F4,$9F,$E6,$E7,$FC,$88,$D7,$3F,$05,$FE,$45,$8C,$40,$81,$CB
	.dc.b	$86,$BB,$6B,$B8,$FF,$06,$7D,$8B,$F0,$2A,$49,$FF,$09,$46,$5F,$EC
	.dc.b	$DF,$95,$FF,$A2,$5B,$F0,$BF,$DB,$23,$11,$FD,$21,$AF,$F4,$16,$BF
	.dc.b	$ED,$C6,$BF,$C8,$9B,$E0,$1B,$F5,$5F,$E4,$A8,$34,$9E,$F7,$F7,$C6
	.dc.b	$D7,$F7,$B6,$BF,$3B,$EA,$FE,$51,$1F,$9E,$D7,$46,$08,$60,$7C,$0F
	.dc.b	$E0,$2D,$FA,$07,$C0,$F8,$4F,$EC,$71,$8C,$FF,$D0,$3E,$27,$D3,$C7
	.dc.b	$EB,$E1,$C7,$27,$C9,$FD,$AE,$2F,$EC,$D1,$8D,$3E,$F4,$62,$75,$1F
	.dc.b	$81,$CB,$E1,$C6,$23,$FA,$5D,$71,$7F,$27,$3A,$79,$9C,$63,$55,$D6
	.dc.b	$32,$7F,$54,$FC,$25,$EF,$7A,$A3,$0D,$B4,$EA,$7F,$29,$AF,$C2,$8F
	.dc.b	$F4,$FE,$9E,$77,$53,$FE,$9E,$8C,$5F,$39,$CD,$71,$AF,$01,$F9,$DC
	.dc.b	$BF,$9F,$BC,$4F,$CE,$38,$8F,$F0,$8F,$27,$CC,$F1,$3C,$78,$FE,$9F
	.dc.b	$E3,$9F,$2F,$FE,$83,$BD,$E1,$C6,$BF,$50,$DC,$7A,$E1,$C9,$61,$CB
	.dc.b	$E1,$3C,$B9,$77,$9F,$21,$E7,$AE,$0D,$3D,$15,$C5,$CC,$F1,$C7,$C7
	.dc.b	$58,$C3,$C5,$7B,$BE,$B7,$BC,$7C,$F7,$B3,$CD,$5B,$C3,$9D,$F0,$C3
	.dc.b	$D8,$F8,$E3,$FC,$73,$3B,$BE,$73,$BB,$E8,$FA,$32,$F9,$13,$AB,$EA
	.dc.b	$47,$E7,$46,$A7,$E4,$9C,$C9,$3E,$08,$D4,$78,$E7,$A1,$1F,$9F,$1A
	.dc.b	$C7,$E0,$47,$4A,$F0,$3A,$CB,$E9,$E5,$EE,$AD,$CF,$53,$47,$B2,$B8
	.dc.b	$B9,$CB,$E1,$3D,$71,$F1,$39,$C7,$C5,$87,$52,$65,$F4,$23,$2F,$5A
	.dc.b	$F5,$5F,$4B,$F1,$13,$1F,$5C,$65,$F3,$93,$0F,$C0,$E1,$D2,$BE,$4D
	.dc.b	$F4,$8C,$63,$EA,$BE,$71,$89,$F4,$C6,$32,$F8,$B9,$79,$70,$F7,$1F
	.dc.b	$10,$EA,$60,$F8,$11,$C1,$C9,$83,$F0,$A7,$F5,$23,$F3,$C5,$9F,$D2
	.dc.b	$37,$1F,$DC,$B9,$F8,$6F,$E0,$6F,$F5,$76,$BF,$D5,$DA,$FF,$5D,$6C
	.dc.b	$9F,$E2,$19,$7F,$00,$6E,$3F,$BF,$FA,$C7,$D2,$FF,$76,$AD,$6F,$FB
	.dc.b	$0B,$DE,$3D,$E3,$F5,$37,$C3,$87,$51,$27,$F6,$35,$C3,$75,$EE,$73
	.dc.b	$5B,$C3,$BC,$7F,$00,$9D,$46,$09,$97,$C1,$AF,$0E,$5E,$78,$C5,$F4
	.dc.b	$25,$7E,$E2,$F9,$CE,$A3,$9C,$FA,$CF,$05,$73,$71,$1A,$7E,$65,$F8
	.dc.b	$32,$F4,$5F,$32,$4F,$F2,$0C,$3B,$BC,$A3,$2F,$23,$BC,$F2,$74,$4C
	.dc.b	$7E,$3E,$3E,$38,$E1,$8F,$8F,$3F,$B9,$97,$8E,$7D,$F8,$F6,$9C,$C7
	.dc.b	$63,$27,$63,$9C,$7E,$06,$FB,$4F,$0B,$ED,$9F,$D9,$CF,$D7,$87,$03
	.dc.b	$95,$6F,$0E,$59,$FA,$E7,$46,$F3,$F5,$C6,$6B,$84,$FB,$EF,$AC,$62
	.dc.b	$38,$BA,$AE,$31,$EA,$33,$87,$97,$2F,$4B,$F9,$68,$CE,$7F,$17,$3F
	.dc.b	$5B,$F3,$EF,$9C,$62,$7E,$2C,$63,$3F,$95,$7C,$F1,$F1,$65,$E6,$8F
	.dc.b	$D1,$F0,$BC,$1F,$BE,$DD,$3F,$14,$E7,$6B,$3A,$24,$E2,$32,$F0,$FC
	.dc.b	$23,$FD,$35,$BF,$F8,$3A,$D1,$8D,$7F,$C2,$ED,$9F,$F2,$EC,$EB,$C5
	.dc.b	$FF,$37,$C6,$27,$55,$FE,$23,$6B,$FE,$EC,$6D,$71,$9F,$F0,$EC,$BC
	.dc.b	$17,$C1,$FF,$9E,$BA,$3F,$C7,$9D,$6F,$AF,$FC,$B5,$6B,$0F,$F3,$85
	.dc.b	$6A,$73,$7F,$E8,$0C,$3F,$CE,$77,$B7,$F9,$21,$DE,$34,$FF,$9E,$63
	.dc.b	$4F,$F2,$96,$BC,$A3,$6F,$1B,$E6,$D3,$F9,$86,$3F,$2B,$E5,$BA,$B6
	.dc.b	$7C,$2D,$9D,$E6,$AD,$B6,$B6,$D9,$DC,$E8,$EB,$F9,$53,$C6,$A4,$BE
	.dc.b	$F1,$96,$C6,$6B,$CF,$39,$6C,$7A,$3F,$55,$7D,$2D,$25,$FA,$A3,$2F
	.dc.b	$C2,$D1,$F2,$3F,$19,$B1,$98,$C1,$1C,$13,$0F,$AB,$3C,$1F,$F5,$F7
	.dc.b	$17,$D6,$F8,$47,$93,$FC,$F6,$7F,$D5,$F1,$89,$D4,$E6,$7F,$D6,$93
	.dc.b	$9B,$FF,$62,$56,$88,$FF,$B4,$BE,$43,$B6,$06,$92,$34,$4B,$FF,$6E
	.dc.b	$C6,$5F,$C0,$F9,$AF,$86,$FC,$0B,$6B,$F0,$2D,$9D,$CF,$E3,$39,$9F
	.dc.b	$09,$49,$7F,$8E,$8C,$B6,$33,$5F,$9C,$9C,$B6,$3E,$47,$D3,$FF,$B2
	.dc.b	$EB,$F4,$3A,$23,$15,$FE,$07,$8C,$79,$BC,$09,$5C,$2B,$75,$F4,$1A
	.dc.b	$FC,$E6,$AA,$7A,$4F,$8A,$FB,$4D,$BF,$1B,$6B,$C4,$D7,$EE,$B5,$FB
	.dc.b	$6D,$7B,$36,$BB,$D7,$D0,$FC,$46,$5F,$D0,$6D,$FF,$02,$FC,$61,$F2
	.dc.b	$1F,$BC,$6D,$7E,$F1,$B3,$B9,$D1,$27,$4D,$8F,$E0,$3E,$E3,$EB,$7F
	.dc.b	$80,$B6,$BF,$80,$B6,$77,$3F,$C4,$BC,$C7,$E8,$9F,$AC,$9F,$F1,$D4
	.dc.b	$63,$AB,$53,$99,$12,$75,$87,$F0,$EF,$D8,$3F,$D8,$ED,$F3,$7F,$89
	.dc.b	$C6,$2B,$F8,$BB,$5F,$E7,$BF,$B2,$3F,$13,$FE,$32,$DB,$FE,$32,$D9
	.dc.b	$FE,$3E,$D7,$FA,$2F,$DA,$FC,$86,$ED,$FC,$8A,$31,$1F,$EE,$76,$AF
	.dc.b	$11,$18,$6C,$F1,$25,$FF,$2A,$F8,$3F,$C7,$C9,$A4,$E7,$F9,$9F,$89
	.dc.b	$BE,$E2,$92,$FF,$A1,$EE,$9D,$46,$1B,$9F,$D2,$6B,$F4,$04,$F1,$AE
	.dc.b	$91,$FA,$BF,$8E,$7C,$1F,$F1,$96,$E1,$FB,$26,$CE,$E7,$FA,$3F,$C7
	.dc.b	$BD,$7A,$4B,$FE,$93,$19,$6C,$66,$BF,$A8,$CE,$5B,$1F,$D4,$BE,$47
	.dc.b	$F5,$1B,$1A,$4E,$BF,$EC,$7F,$24,$F9,$97,$D6,$E1,$B9,$FF,$6A,$9D
	.dc.b	$4E,$49,$19,$9D,$E5,$FE,$93,$AD,$CE,$6B,$F9,$E6,$3F,$C6,$2B,$F9
	.dc.b	$DB,$5F,$E7,$2D,$7F,$86,$37,$A1,$F2,$CF,$BB,$FE,$F7,$18,$6C,$67
	.dc.b	$0F,$F0,$0C,$E5,$B1,$9B,$FF,$54,$4E,$5B,$19,$7F,$01,$19,$7F,$C4
	.dc.b	$7F,$9D,$FE,$11,$FD,$9E,$08,$C4,$7E,$75,$B1,$9E,$2F,$9F,$FB,$4E
	.dc.b	$B4,$34,$91,$A2,$65,$F4,$63,$2F,$B3,$E7,$3F,$D6,$4D,$F6,$FE,$1C
	.dc.b	$63,$2F,$6B,$5F,$5F,$CE,$FC,$B2,$79,$3F,$29,$F9,$DF,$9B,$3E,$EF
	.dc.b	$E6,$B7,$1F,$96,$DB,$F9,$2D,$7E,$3B,$5F,$E2,$FB,$A3,$10,$8C,$57
	.dc.b	$63,$FC,$01,$ED,$BE,$4E,$93,$0F,$EB,$71,$96,$C6,$6F,$FA,$E4,$E5
	.dc.b	$B1,$F9,$EF,$D5,$FF,$AB,$F0,$1A,$4E,$FD,$13,$0F,$D0,$46,$5F,$F2
	.dc.b	$0F,$E5,$EF,$C5,$3F,$C8,$AD,$8C,$C6,$0F,$F2,$7F,$C3,$3E,$AF,$FC
	.dc.b	$61,$9F,$F0,$B8,$C7,$B9,$A9,$CC,$FF,$3E,$9C,$DF,$F9,$69,$B1,$FE
	.dc.b	$FF,$6B,$CE,$78,$A8,$CB,$FE,$62,$F8,$8F,$F2,$CB,$FD,$29,$A9,$A7
	.dc.b	$F9,$7B,$E9,$9F,$42,$FA,$5C,$36,$BF,$CD,$B3,$A9,$C9,$C7,$DB,$71
	.dc.b	$52,$57,$08,$CB,$FE,$76,$5F,$23,$E6,$6C,$7F,$9C,$1A,$EE,$34,$47
	.dc.b	$4D,$7F,$D0,$3D,$5F,$D1,$DF,$F3,$CB,$5C,$F9,$7F,$12,$31,$1F,$CF
	.dc.b	$C7,$D3,$7F,$35,$7F,$D0,$2D,$7F,$D1,$AD,$78,$35,$FE,$90,$D7,$FB
	.dc.b	$CB,$5E,$BE,$3F,$D4,$8D,$7B,$1F,$3F,$C6,$FF,$47,$B6,$BF,$C7,$7A
	.dc.b	$7F,$37,$6C,$EA,$31,$1F,$E3,$D6,$BF,$E6,$86,$BF,$E6,$66,$BF,$B3
	.dc.b	$F9,$FF,$F1,$2B,$FC,$A1,$AF,$F2,$66,$BF,$C9,$1A,$FE,$F9,$AF,$3F
	.dc.b	$3F,$FD,$4F,$DA,$9C,$B6,$33,$DF,$FE,$C3,$AF,$E6,$ED,$7F,$9A,$B5
	.dc.b	$FE,$33,$EF,$F4,$F0,$9F,$33,$49,$3F,$C9,$5F,$F5,$BF,$7F,$FB,$3B
	.dc.b	$E9,$AF,$C4,$9A,$4A,$FE,$BD,$19,$6C,$65,$FE,$B9,$19,$7F,$D8,$3F
	.dc.b	$4D,$FE,$5F,$7F,$D7,$ED,$73,$E7,$7F,$B2,$3C,$A4,$E6,$7F,$A6,$F9
	.dc.b	$9F,$E8,$97,$FA,$AF,$C1,$7C,$88,$C4,$7F,$54,$8C,$11,$D9,$0F,$F3
	.dc.b	$E5,$3E,$97,$F1,$2D,$8F,$F4,$1B,$5D,$C7,$C6,$FA,$6F,$F4,$9C,$7F
	.dc.b	$7F,$6B,$FD,$ED,$AF,$F8,$8F,$44,$62,$3F,$CA,$B1,$8F,$33,$A1,$23
	.dc.b	$A3,$F0,$FD,$27,$C9,$D2,$4F,$FB,$A2,$32,$FF,$88,$BC,$2F,$95,$EA
	.dc.b	$7F,$14,$8C,$10,$FF,$11,$C6,$88,$E4,$FF,$64,$FA,$DF,$EE,$97,$FC
	.dc.b	$69,$E4,$FE,$BA,$31,$1F,$E3,$B8,$FF,$7F,$FA,$EF,$A6,$F9,$13,$FE
	.dc.b	$43,$8C,$6D,$A9,$CC,$FF,$9D,$5A,$FF,$92,$1B,$3C,$23,$FC,$A7,$18
	.dc.b	$37,$BB,$E9,$5F,$4D,$F8,$96,$D7,$F1,$53,$FC,$D4,$E1,$FE,$31,$3C
	.dc.b	$67,$F2,$1F,$4D,$FB,$19,$E1,$28,$C4,$70,$6B,$FE,$5F,$8D,$77,$F2
	.dc.b	$24,$72,$9D,$BF,$E2,$26,$BF,$E7,$DF,$A6,$7D,$4F,$E8,$5B,$1E,$26
	.dc.b	$BE,$86,$BE,$1F,$A7,$FC,$0B,$EC,$6B,$9E,$3F,$D6,$46,$23,$FD,$49
	.dc.b	$F0,$FF,$CE,$9B,$E5,$7D,$A8,$C4,$7F,$9E,$9B,$9B,$E9,$FF,$C2,$13
	.dc.b	$F1,$9A,$FC,$C6,$BF,$15,$AF,$FA,$17,$E9,$FF,$B5,$91,$FE,$DE,$D7
	.dc.b	$EC,$07,$0F,$DB,$7F,$BA,$CB,$E9,$FF,$D1,$73,$F6,$1A,$FD,$E6,$BF
	.dc.b	$5D,$AF,$FB,$A3,$E6,$7F,$A4,$FC,$71,$A4,$8F,$F7,$9B,$5F,$F7,$9B
	.dc.b	$5F,$F5,$07,$99,$FC,$BB,$E4,$7F,$A9,$F8,$C6,$92,$7F,$29,$3B,$6C
	.dc.b	$ED,$B3,$B7,$F3,$5F,$36,$FA,$7F,$F0,$AB,$63,$E3,$B5,$DC,$68,$8E
	.dc.b	$9A,$FC,$5E,$BF,$E6,$19,$FE,$F2,$3E,$34,$60,$FE,$DB,$E5,$3E,$9E
	.dc.b	$92,$7D,$71,$96,$C6,$5F,$F0,$36,$69,$D4,$61,$AF,$CA,$F9,$67,$B9
	.dc.b	$FD,$F1,$B1,$F2,$1A,$F4,$6B,$C6,$38,$11,$E0,$FF,$8F,$3E,$7D,$F4
	.dc.b	$F4,$93,$F9,$08,$CB,$63,$2F,$F0,$B8,$CB,$63,$FB,$CF,$D3,$BE,$0E
	.dc.b	$92,$7F,$D5,$B1,$97,$F8,$6F,$D3,$BE,$9F,$FB,$03,$63,$31,$82,$38
	.dc.b	$21,$FC,$6A,$38,$3F,$C7,$3E,$9B,$FC,$82,$FF,$15,$6B,$FE,$B2,$6A
	.dc.b	$E7,$46,$1B,$EE,$FF,$73,$8C,$11,$F9,$5F,$4F,$2B,$E4,$35,$F8,$CD
	.dc.b	$7F,$C8,$DA,$23,$10,$24,$65,$C7,$C0,$7E,$03,$07,$CE,$7B,$BE,$0F
	.dc.b	$FA,$FC,$63,$4E,$57,$FE,$75,$F0,$AE,$3C,$E3,$FC,$DB,$C3,$F0,$7E
	.dc.b	$93,$6F,$C5,$6C,$64,$F2,$46,$5A,$64,$FC,$8E,$1F,$22,$78,$C7,$CB
	.dc.b	$D9,$18,$9D,$4F,$23,$E8,$7D,$3F,$FA,$F5,$FF,$47,$63,$FE,$E6,$8C
	.dc.b	$47,$AC,$DF,$BD,$EE,$D3,$E0,$CF,$F9,$C5,$B1,$9A,$E4,$DA,$E2,$76
	.dc.b	$3C,$26,$46,$7E,$8F,$37,$51,$8C,$91,$89,$D4,$F9,$7E,$19,$F4,$FF
	.dc.b	$E7,$FC,$BF,$A9,$D6,$9B,$87,$2B,$F6,$1D,$70,$F2,$61,$CB,$E1,$3C
	.dc.b	$98,$71,$CF,$ED,$E1,$C7,$1E,$78,$70,$CB,$C1,$5A,$8F,$35,$6E,$31
	.dc.b	$97,$9E,$B6,$FE,$42,$74,$DA,$E2,$75,$C7,$C7,$3A,$9F,$41,$EA,$C7
	.dc.b	$CC,$FF,$B2,$EF,$AB,$AC,$7C,$EE,$EB,$BE,$5E,$F8,$EE,$67,$1B,$5F
	.dc.b	$E3,$C9,$D4,$E4,$91,$D0,$D5,$F8,$30,$F1,$CF,$87,$2F,$83,$5A,$CB
	.dc.b	$D7,$8F,$69,$C9,$FE,$EB,$C3,$96,$5E,$FC,$79,$9E,$EC,$BC,$78,$F8
	.dc.b	$6F,$8E,$5F,$24,$F9,$D9,$76,$7C,$F1,$AA,$E2,$F0,$8F,$31,$C6,$BC
	.dc.b	$CF,$94,$C4,$71,$C7,$CB,$96,$67,$8D,$F4,$76,$43,$33,$C1,$F1,$9A
	.dc.b	$8E,$0F,$88,$98,$FB,$E7,$86,$1E,$F9,$E1,$18,$C3,$D5,$5B,$C7,$E0
	.dc.b	$5E,$EF,$C3,$18,$AD,$E5,$F8,$2C,$3C,$95,$FC,$9A,$3F,$4C,$DC,$BF
	.dc.b	$AA,$46,$23,$F8,$EF,$C2,$FF,$06,$6E,$9F,$D6,$63,$11,$FC,$B1,$AF
	.dc.b	$EF,$9F,$E2,$19,$7E,$D9,$B8,$FF,$33,$D1,$F4,$EF,$A1,$FE,$E3,$7B
	.dc.b	$C7,$F8,$16,$1C,$31,$ED,$1F,$BA,$BE,$1C,$5A,$89,$3F,$A7,$AE,$18
	.dc.b	$F6,$C7,$AE,$5E,$AA,$D5,$F0,$CB,$F4,$59,$74,$9C,$92,$3C,$8F,$F1
	.dc.b	$38,$D4,$60,$95,$D5,$E7,$7E,$0A,$F4,$9C,$E7,$47,$C3,$AE,$6E,$23
	.dc.b	$75,$D1,$61,$EB,$CB,$D1,$5B,$9F,$A1,$8F,$C5,$C7,$C1,$18,$3C,$07
	.dc.b	$49,$F0,$46,$B7,$67,$F3,$EB,$86,$5E,$1C,$BC,$11,$C8,$CE,$1E,$78
	.dc.b	$CC,$F9,$E7,$27,$73,$D3,$9F,$B9,$C1,$A3,$73,$CA,$BD,$98,$79,$32
	.dc.b	$F7,$D7,$2C,$BD,$19,$79,$EB,$53,$EE,$C3,$93,$DA,$33,$97,$AF,$2F
	.dc.b	$14,$F8,$67,$39,$7A,$E7,$31,$89,$E5,$87,$E3,$27,$2F,$84,$99,$7D
	.dc.b	$1C,$FC,$F5,$D6,$78,$67,$E8,$AC,$E1,$D7,$3F,$3D,$EE,$7D,$F1,$89
	.dc.b	$F1,$4E,$4F,$C9,$57,$37,$FB,$D7,$0B,$C3,$F9,$CD,$9F,$C9,$39,$DA
	.dc.b	$CE,$89,$38,$8C,$BC,$3E,$9B,$FD,$31,$A7,$F8,$36,$D1,$8C,$3F,$C2
	.dc.b	$ED,$9F,$F6,$C3,$63,$FC,$65,$3C,$3C,$7F,$F3,$7D,$69,$B3,$AA,$E0
	.dc.b	$FF,$92,$5A,$FF,$9F,$7E,$07,$FC,$EF,$D3,$FE,$39,$E9,$7D,$3F,$E8
	.dc.b	$3C,$25,$6B,$2F,$0D,$69,$FF,$08,$CE,$9F,$F6,$1B,$70,$FF,$7E,$35
	.dc.b	$FF,$20,$9E,$18,$DD,$7F,$A1,$23,$FC,$9D,$3B,$7F,$C8,$ED,$73,$1B
	.dc.b	$CF,$E4,$FD,$3D,$CC,$D2,$61,$E0,$BF,$85,$18,$7C,$75,$CA,$74,$D8
	.dc.b	$CF,$D3,$BC,$7F,$03,$6B,$6D,$9D,$CE,$89,$3A,$6C,$78,$3E,$9D,$F4
	.dc.b	$5E,$B6,$D7,$89,$B3,$B9,$F8,$CF,$FA,$D7,$E9,$EE,$A6,$92,$FE,$9B
	.dc.b	$67,$E9,$B5,$FF,$5D,$71,$7D,$1F,$25,$E4,$FF,$39,$9C,$DF,$FA,$EE
	.dc.b	$B5,$39,$9F,$F4,$25,$70,$9C,$92,$BF,$D9,$FF,$1F,$D7,$13,$49,$1A
	.dc.b	$25,$FF,$B3,$E3,$2F,$D6,$F3,$9F,$0B,$F8,$16,$D7,$D8,$6C,$EE,$7E
	.dc.b	$3F,$23,$E0,$E9,$2F,$F0,$95,$FE,$98,$8C,$3F,$87,$9E,$13,$A6,$C6
	.dc.b	$67,$E4,$7D,$3F,$FB,$2E,$BF,$1D,$A2,$31,$5F,$EE,$C6,$BF,$EA,$28
	.dc.b	$D7,$9B,$C8,$95,$CA,$B7,$5F,$39,$AA,$9E,$9B,$E2,$7E,$D3,$6F,$C6
	.dc.b	$DA,$FF,$7E,$B5,$F1,$35,$FB,$6D,$7B,$36,$BB,$D7,$D1,$FA,$79,$7F
	.dc.b	$45,$B7,$FA,$FF,$A7,$7C,$7F,$E6,$1B,$5F,$41,$B3,$B9,$D1,$27,$4D
	.dc.b	$8F,$DB,$FB,$4F,$A7,$FE,$04,$DA,$FD,$CB,$67,$73,$FC,$37,$CB,$7D
	.dc.b	$3F,$EB,$27,$FD,$A7,$18,$E9,$D4,$E6,$7F,$DC,$13,$9C,$3F,$82,$FD
	.dc.b	$3B,$C9,$FE,$24,$D8,$CC,$60,$FE,$2D,$F4,$CF,$A7,$A4,$9F,$E1,$F1
	.dc.b	$97,$F8,$8F,$D3,$FC,$4C,$7F,$DC,$FE,$12,$31,$1F,$D0,$9A,$A3,$5B
	.dc.b	$FF,$25,$9C,$C6,$2B,$AC,$E4,$FE,$55,$F0,$3F,$8C,$93,$49,$CF,$F2
	.dc.b	$7E,$9D,$F4,$74,$98,$7F,$2D,$DD,$3A,$8C,$37,$3F,$E8,$33,$AA,$E3
	.dc.b	$18,$9D,$3F,$A5,$F8,$D7,$D3,$FF,$15,$6E,$1F,$AE,$6C,$FF,$4C,$E9
	.dc.b	$46,$23,$E4,$FC,$83,$D8,$FE,$90,$DB,$FE,$7E,$D9,$FE,$86,$D7,$8C
	.dc.b	$F0,$24,$F0,$7F,$A9,$7C,$8F,$E9,$F7,$34,$9D,$9F,$D7,$7E,$49,$F3
	.dc.b	$4F,$A7,$C3,$71,$FE,$AD,$3A,$9C,$92,$33,$3B,$CB,$FB,$5D,$6E,$73
	.dc.b	$5F,$CE,$F1,$FE,$2F,$5F,$C3,$DA,$FF,$38,$6B,$FC,$2D,$BD,$0F,$96
	.dc.b	$7D,$3F,$F7,$58,$C3,$63,$38,$7F,$6D,$9C,$B6,$33,$7F,$E0,$09,$CB
	.dc.b	$63,$28,$91,$A7,$F4,$3F,$4F,$F8,$47,$FC,$42,$D7,$F4,$0D,$C0,$E7
	.dc.b	$27,$D3,$D2,$72,$FD,$36,$E5,$F4,$1B,$87,$CC,$7B,$62,$F9,$CF,$F3
	.dc.b	$ED,$7B,$DB,$5E,$D6,$BE,$B3,$F8,$33,$87,$DB,$F3,$BF,$38,$9E,$47
	.dc.b	$CD,$7E,$8F,$D3,$BE,$9F,$E8,$37,$0F,$9C,$DB,$F9,$8D,$7C,$18,$A3
	.dc.b	$11,$F3,$1A,$FF,$19,$6C,$6E,$BF,$C0,$5E,$D3,$E9,$E9,$30,$FE,$BB
	.dc.b	$19,$6C,$67,$0F,$D9,$4E,$5B,$1F,$A1,$FA,$7F,$EB,$3B,$9A,$4E,$FD
	.dc.b	$13,$0F,$D1,$C6,$5F,$EE,$FF,$4E,$FA,$7F,$F1,$1B,$63,$31,$83,$FC
	.dc.b	$99,$F0,$6F,$A7,$FF,$1B,$67,$F5,$63,$1E,$DE,$A7,$33,$FE,$ED,$6B
	.dc.b	$FE,$C8,$6C,$F0,$BF,$F2,$63,$67,$FC,$BA,$D7,$FC,$1B,$F1,$1F,$E0
	.dc.b	$E7,$F3,$CD,$4D,$3F,$C2,$7F,$4E,$FA,$27,$D3,$61,$B5,$FE,$5B,$9D
	.dc.b	$4E,$4E,$3E,$D3,$8A,$92,$B8,$46,$5F,$F3,$0F,$D3,$3D,$BF,$CC,$53
	.dc.b	$4F,$F4,$7C,$E6,$BF,$6D,$39,$7F,$A8,$46,$5F,$04,$FE,$86,$DF,$25
	.dc.b	$D2,$39,$E1,$6B,$8B,$7B,$D8,$73,$C7,$B5,$7F,$92,$30,$E0,$BE,$9F
	.dc.b	$F9,$E3,$FE,$3A,$6B,$FE,$7F,$6B,$FE,$90,$6B,$FC,$3F,$EA,$DF,$A4
	.dc.b	$3E,$07,$C9,$18,$6C,$66,$7E,$0E,$9F,$BD,$BE,$F8,$FC,$3C,$38,$56
	.dc.b	$83,$87,$E2,$98,$BE,$8E,$0F,$67,$95,$F1,$7C,$DF,$C0,$E9,$78,$3E
	.dc.b	$2D,$75,$9C,$B6,$3F,$A6,$B6,$78,$11,$D7,$93,$D7,$95,$F4,$74,$93
	.dc.b	$E9,$F0,$33,$E9,$E2,$3D,$39,$2F,$8C,$78,$DA,$FB,$AB,$8B,$CF,$C5
	.dc.b	$7C,$EF,$F9,$C2,$7F,$96,$36,$33,$3F,$CB,$63,$FC,$EB,$6F,$AD,$FF
	.dc.b	$20,$4F,$C1,$6B,$E4,$6F,$0F,$E5,$70,$EB,$3F,$4F,$99,$F5,$5F,$E6
	.dc.b	$78,$FF,$5F,$B5,$FF,$61,$B5,$FF,$18,$B5,$FF,$78,$B5,$FF,$18,$36
	.dc.b	$3A,$F9,$FF,$E8,$C6,$CF,$FA,$13,$0F,$E3,$0F,$A7,$C0,$FB,$D1,$88
	.dc.b	$FC,$2C,$F0,$7E,$7A,$FA,$9F,$86,$D8,$FF,$47,$B5,$FF,$11,$35,$FF
	.dc.b	$00,$47,$02,$3C,$1F,$9B,$F6,$0F,$9F,$FC,$83,$67,$F1,$CD,$7C,$73
	.dc.b	$FB,$7F,$4D,$F2,$FE,$F6,$CF,$FB,$A9,$AF,$FB,$B9,$AF,$F9,$71,$AF
	.dc.b	$F8,$07,$0F,$F1,$4C,$F3,$24,$7F,$AB,$2D,$F7,$3E,$62,$47,$DD,$BE
	.dc.b	$FB,$FF,$15,$E0,$FF,$21,$F8,$DF,$8A,$F5,$FF,$DF,$57,$FE,$49,$8C
	.dc.b	$11,$C1,$0F,$F2,$64,$70,$7F,$C9,$DF,$69,$FB,$8B,$EC,$DF,$03,$FD
	.dc.b	$C6,$D9,$FF,$5B,$57,$1A,$FE,$51,$AB,$EA,$DF,$57,$F8,$76,$DF,$F8
	.dc.b	$24,$F8,$8E,$1C,$4F,$19,$E0,$48,$E0,$7F,$A0,$E7,$83,$FE,$8A,$6B
	.dc.b	$E6,$FB,$CF,$E1,$D8,$9A,$7F,$33,$9F,$06,$1E,$8F,$AA,$7D,$5F,$E7
	.dc.b	$5B,$87,$83,$24,$62,$3C,$ED,$7F,$D3,$6D,$7F,$00,$D7,$FD,$34,$D8
	.dc.b	$E7,$3F,$C0,$BD,$F7,$E0,$3F,$C9,$5B,$7F,$C8,$DA,$FE,$AA,$FC,$58
	.dc.b	$7F,$57,$F8,$17,$C5,$7F,$A7,$AB,$F9,$B4,$63,$E0,$FF,$AA,$36,$3F
	.dc.b	$84,$7D,$67,$E3,$38,$BF,$D8,$7F,$AA,$7F,$A8,$A7,$FD,$40,$D7,$FD
	.dc.b	$3A,$D7,$FD,$9A,$D7,$FC,$BD,$CA,$FA,$1F,$E4,$F9,$FE,$5D,$B7,$A6
	.dc.b	$D7,$FA,$F7,$EB,$7F,$C6,$8F,$F1,$BF,$AB,$FF,$52,$BF,$45,$AF,$E1
	.dc.b	$9A,$FF,$B2,$9A,$FE,$BD,$AF,$FB,$25,$BC,$46,$9A,$C7,$F6,$3F,$01
	.dc.b	$FD,$57,$AC,$D3,$FB,$D4,$F8,$B3,$FD,$97,$C9,$3D,$5F,$EB,$8D,$CB
	.dc.b	$FA,$D3,$5F,$E9,$AB,$E7,$F7,$32,$76,$F2,$8F,$EC,$DF,$28,$E2,$A4
	.dc.b	$AF,$EC,$DE,$4F,$F8,$2F,$87,$E8,$D6,$E7,$FB,$53,$5F,$EB,$AD,$CB
	.dc.b	$F8,$55,$71,$9F,$8D,$DF,$EE,$23,$FD,$FC,$7D,$0B,$EB,$33,$1F,$CF
	.dc.b	$36,$FE,$05,$97,$D1,$BF,$F0,$DD,$EE,$BF,$BE,$7D,$F7,$FA,$C1,$BF
	.dc.b	$17,$FE,$03,$8C,$47,$3A,$FF,$03,$7C,$D7,$F2,$48,$FF,$02,$35,$FE
	.dc.b	$EE,$DD,$3F,$92,$3F,$CE,$DB,$C7,$C8,$93,$CB,$C9,$FF,$09,$7C,$DB
	.dc.b	$C4,$A4,$9F,$F0,$94,$7F,$8D,$AF,$F9,$83,$B8,$CC,$ED,$FF,$04,$C6
	.dc.b	$F5,$FF,$5F,$FD,$63,$E1,$7F,$C0,$8D,$AF,$F6,$63,$5F,$F6,$53,$5F
	.dc.b	$F4,$FB,$5E,$71,$C8,$8F,$F2,$49,$FC,$BF,$97,$FE,$51,$6C,$7F,$8D
	.dc.b	$DA,$FF,$86,$6B,$C1,$8F,$F4,$BE,$4F,$91,$5F,$8A,$CF,$F2,$27,$F9
	.dc.b	$67,$EB,$1F,$57,$8C,$7F,$1B,$35,$FE,$FB,$3A,$7E,$BE,$5F,$68,$8F
	.dc.b	$F8,$5F,$EA,$FF,$CD,$2E,$88,$7D,$C8,$FF,$34,$46,$A3,$FC,$95,$F5
	.dc.b	$0F,$AC,$FF,$36,$7D,$07,$79,$FF,$47,$DF,$F6,$F8,$C7,$C4,$7F,$76
	.dc.b	$BE,$15,$F9,$B6,$BF,$E1,$49,$DD,$71,$CF,$FD,$BD,$7C,$70,$EF,$18
	.dc.b	$9D,$5F,$FA,$81,$AF,$88,$FF,$80,$18,$76,$DB,$FB,$2E,$4F,$A9,$E5
	.dc.b	$4F,$09,$CD,$F8,$3D,$3F,$C1,$F1,$8E,$3D,$F8,$6C,$79,$61,$E1,$11
	.dc.b	$9C,$7A,$DF,$19,$D3,$5F,$34,$ED,$E4,$BE,$B1,$ED,$78,$1B,$3A,$6B
	.dc.b	$C9,$FE,$F9,$97,$F8,$9A,$B9,$57,$12,$57,$13,$E0,$8D,$7E,$27,$83
	.dc.b	$EB,$8C,$65,$FC,$6F,$0F,$B4,$DD,$D3,$FE,$77,$FA,$CF,$F5,$2D,$6A
	.dc.b	$F9,$61,$D2,$BE,$A1,$D2,$3E,$47,$BA,$BE,$38,$74,$AD,$1C,$2F,$96
	.dc.b	$1D,$63,$EF,$5F,$1F,$83,$F5,$D6,$AF,$93,$6B,$F0,$79,$7A,$5F,$45
	.dc.b	$70,$8C,$47,$83,$0E,$C7,$F9,$43,$0E,$D1,$EC,$9D,$DF,$58,$E1,$87
	.dc.b	$C4,$AD,$C7,$A6,$B6,$4C,$BE,$0E,$5F,$02,$7C,$B1,$D4,$8F,$BA,$34
	.dc.b	$F6,$36,$75,$8E,$79,$7A,$DE,$39,$7A,$B1,$F2,$5F,$2C,$BD,$04,$C7
	.dc.b	$E2,$65,$E8,$BE,$19,$76,$6C,$67,$0F,$9B,$5E,$BA,$F1,$1E,$7C,$BC
	.dc.b	$4E,$2F,$8E,$3E,$1A,$E1,$18,$AD,$BE,$0A,$F7,$61,$DD,$F1,$61,$E0
	.dc.b	$CB,$DA,$FB,$2F,$8B,$99,$F0,$39,$76,$72,$25,$75,$7E,$61,$32,$FC
	.dc.b	$44,$65,$F8,$D7,$CA,$31,$8F,$AA,$FC,$18,$FE,$1F,$2F,$75,$F8,$31
	.dc.b	$F5,$C6,$1F,$29,$E2,$70,$FF,$9D,$71,$FD,$43,$70,$FF,$05,$3E,$D3
	.dc.b	$5F,$59,$FC,$09,$FC,$FB,$5F,$E4,$2D,$7F,$7C,$D9,$3F,$C3,$30,$FE
	.dc.b	$6C,$DC,$7F,$35,$E5,$BE,$21,$D7,$FB,$CA,$D6,$3D,$B7,$C1,$FB,$CB
	.dc.b	$E1,$58,$9F,$DD,$57,$07,$C7,$5B,$BE,$B1,$A8,$E1,$8F,$91,$D5,$E3
	.dc.b	$1F,$7C,$5A,$6D,$FE,$61,$C3,$9E,$1C,$67,$A1,$1F,$DE,$65,$E6,$8C
	.dc.b	$61,$D1,$A7,$48,$D4,$F3,$24,$ED,$E7,$5E,$08,$C1,$2B,$E3,$1F,$C6
	.dc.b	$2B,$9C,$66,$3E,$19,$E0,$BF,$56,$3E,$D8,$CE,$1E,$07,$C9,$87,$77
	.dc.b	$AC,$65,$DD,$FD,$A9,$D4,$60,$DC,$79,$6F,$BC,$7C,$03,$46,$DC,$D7
	.dc.b	$5C,$FF,$13,$97,$D1,$9C,$DF,$AE,$73,$7E,$BC,$BD,$EF,$80,$87,$57
	.dc.b	$2F,$82,$38,$E1,$EF,$BE,$B3,$C2,$73,$7E,$9B,$E3,$18,$9D,$67,$F9
	.dc.b	$6B,$E3,$1D,$63,$F2,$53,$B9,$CB,$C6,$75,$8F,$E4,$1F,$89,$19,$38
	.dc.b	$5F,$82,$F9,$92,$B8,$63,$F5,$F8,$3F,$CF,$39,$FD,$76,$E9,$FE,$00
	.dc.b	$F1,$3F,$C2,$8D,$9D,$47,$C8,$9D,$E7,$FE,$08,$8D,$B6,$7A,$7D,$57
	.dc.b	$FA,$76,$FF,$C1,$FC,$EF,$BE,$CF,$95,$F8,$7E,$7F,$F0,$FC,$62,$B6
	.dc.b	$DC,$7F,$C5,$53,$A9,$FF,$0A,$F6,$BE,$B7,$D4,$D8,$FF,$04,$62,$FA
	.dc.b	$DF,$DD,$F6,$FE,$F3,$75,$EF,$2F,$F0,$1D,$EF,$0F,$F0,$5D,$EE,$B5
	.dc.b	$C0,$E6,$FF,$02,$5E,$F0,$FF,$24,$67,$F9,$C8,$C6,$1C,$63,$19,$F8
	.dc.b	$B2,$FF,$2A,$47,$C1,$A7,$D6,$E7,$0F,$F2,$14,$62,$BC,$67,$F8,$7B
	.dc.b	$47,$06,$6B,$FD,$43,$7C,$76,$FF,$24,$B7,$0D,$67,$F0,$6E,$E3,$FE
	.dc.b	$F6,$9B,$87,$F9,$47,$C3,$78,$FC,$DB,$5E,$29,$FF,$26,$FD,$6B,$C9
	.dc.b	$D5,$CA,$FF,$67,$BE,$BD,$55,$B9,$C9,$FE,$93,$F6,$9F,$2F,$F0,$61
	.dc.b	$18,$7C,$14,$FA,$D7,$D4,$6E,$B5,$9B,$7F,$BD,$5A,$C3,$FC,$75,$3F
	.dc.b	$26,$30,$E3,$0E,$98,$7F,$40,$8F,$F2,$04,$62,$B8,$C7,$F9,$B2,$B8
	.dc.b	$BA,$6B,$D6,$B9,$36,$78,$F0,$BC,$CF,$D3,$E3,$FE,$CA,$F0,$7E,$C4
	.dc.b	$7F,$41,$8C,$57,$FB,$42,$72,$6B,$AE,$C7,$FB,$8B,$AF,$EF,$1F,$7F
	.dc.b	$D3,$B1,$92,$65,$FE,$E6,$8C,$BF,$1E,$3F,$CD,$39,$3C,$F7,$E0,$5F
	.dc.b	$8B,$6C,$FD,$2F,$94,$7D,$6E,$6B,$EF,$46,$27,$53,$F9,$BC,$3F,$5F
	.dc.b	$EC,$6A,$24,$E8,$FC,$64,$E9,$C3,$E5,$3F,$29,$DD,$F9,$33,$F2,$9E
	.dc.b	$6D,$9F,$CA,$66,$EA,$FD,$B4,$61,$B7,$CE,$75,$5C,$63,$1E,$C3,$EA
	.dc.b	$3F,$54,$DC,$3A,$AF,$8B,$CC,$C9,$CB,$C2,$31,$5D,$FB,$1F,$49,$FB
	.dc.b	$56,$DF,$CB,$FA,$DC,$4F,$CA,$7F,$13,$5F,$EB,$AE,$04,$63,$AD,$18
	.dc.b	$BF,$CA,$C6,$27,$53,$9A,$FC,$F3,$5F,$CE,$B5,$FE,$18,$D7,$FD,$A5
	.dc.b	$F7,$FF,$C0,$1F,$90,$DE,$9F,$E5,$5F,$04,$FA,$DF,$E2,$13,$9C,$BF
	.dc.b	$D6,$7E,$93,$F4,$C7,$A9,$EC,$8C,$36,$33,$3F,$AE,$9C,$B7,$EB,$BF
	.dc.b	$6F,$97,$F8,$D2,$51,$89,$FE,$49,$F5,$DF,$ED,$67,$F8,$B4,$EB,$C2
	.dc.b	$FE,$49,$F1,$1F,$22,$7F,$0E,$DD,$BF,$1D,$C2,$FA,$EF,$C0,$47,$E1
	.dc.b	$E7,$59,$FF,$49,$9C,$CF,$F0,$66,$BF,$CD,$5A,$FF,$BC,$9B,$5D,$79
	.dc.b	$BE,$D4,$60,$8F,$E0,$3A,$5F,$01,$FC,$73,$0F,$F2,$9C,$63,$C8,$FE
	.dc.b	$59,$F1,$0E,$CF,$E8,$17,$BD,$7F,$A3,$4E,$5F,$C2,$B7,$4F,$E4,$0D
	.dc.b	$7F,$8F,$B5,$E3,$5F,$8F,$8C,$43,$44,$61,$B3,$E9,$B7,$BB,$E4,$3F
	.dc.b	$AA,$7C,$73,$EB,$71,$97,$EE,$23,$0F,$F2,$66,$E5,$FC,$E3,$44,$62
	.dc.b	$3F,$DE,$91,$8A,$F9,$13,$99,$FE,$36,$D5,$3B,$AE,$D5,$F3,$FE,$45
	.dc.b	$F4,$3F,$DA,$5B,$3F,$D4,$3E,$4B,$F8,$03,$FC,$3A,$33,$87,$E2,$27
	.dc.b	$2F,$F4,$B6,$BF,$D2,$9A,$FF,$79,$6B,$F2,$DB,$A9,$A4,$F0,$7C,$DF
	.dc.b	$31,$E5,$3F,$BF,$4B,$C7,$C6,$3F,$C8,$A3,$0F,$F5,$06,$E1,$F2,$DA
	.dc.b	$FC,$A6,$BF,$D0,$9A,$FC,$86,$E9,$F8,$1A,$FC,$B4,$63,$E5,$5F,$7B
	.dc.b	$FE,$11,$6E,$1F,$DE,$3E,$B1,$F5,$99,$AF,$F0,$24,$62,$7F,$8C,$B7
	.dc.b	$2F,$EE,$8D,$7F,$B9,$B5,$FF,$11,$35,$FC,$A3,$70,$34,$D6,$4C,$7F
	.dc.b	$84,$57,$E1,$1B,$F0,$0F,$9B,$FF,$69,$E3,$FD,$32,$73,$18,$AF,$F0
	.dc.b	$5D,$6F,$E0,$BF,$9B,$B6,$FF,$C0,$ED,$9F,$CD,$7D,$0F,$F2,$57,$F2
	.dc.b	$6D,$87,$D5,$3E,$AB,$FC,$75,$3B,$CC,$D3,$84,$EF,$0F,$EC,$31,$BE
	.dc.b	$D7,$D2,$7F,$90,$DB,$3F,$D9,$9F,$E0,$D9,$7E,$91,$CB,$FD,$95,$A7
	.dc.b	$F1,$02,$C7,$F8,$19,$E3,$87,$F8,$BE,$BF,$CB,$16,$FD,$61,$F6,$58
	.dc.b	$6E,$1F,$E4,$9F,$AC,$7D,$67,$F1,$3C,$BF,$DD,$F8,$71,$C3,$81,$2F
	.dc.b	$85,$6D,$67,$F1,$31,$EA,$FF,$81,$F1,$EB,$E0,$7F,$6A,$6D,$7F,$8E
	.dc.b	$9B,$3F,$E4,$46,$BF,$E4,$0F,$B8,$FE,$EA,$FF,$44,$6E,$6A,$F0,$FC
	.dc.b	$73,$EB,$3F,$CB,$97,$BF,$53,$FC,$C1,$87,$0B,$FF,$3C,$36,$57,$C7
	.dc.b	$E1,$37,$7F,$E2,$79,$F6,$FC,$57,$F9,$41,$F7,$87,$0E,$8A,$EF,$EB
	.dc.b	$3F,$8D,$FD,$55,$F1,$0D,$25,$7F,$97,$25,$F0,$1F,$E6,$98,$C3,$5F
	.dc.b	$F3,$4C,$6B,$6E,$D3,$99,$F2,$F2,$FF,$7E,$7D,$0D,$7F,$CD,$2D,$7F
	.dc.b	$CE,$6D,$78,$FA,$9E,$C8,$C4,$7D,$E6,$BF,$9C,$8F,$83,$F6,$5F,$D3
	.dc.b	$A3,$E3,$35,$FF,$3B,$B5,$FF,$40,$B5,$F9,$7F,$9B,$7F,$A7,$9F,$9C
	.dc.b	$D7,$FC,$FE,$D7,$FD,$12,$D7,$8F,$B9,$F4,$A3,$11,$F9,$46,$A7,$3E
	.dc.b	$13,$F6,$D7,$DB,$7D,$D8,$C3,$63,$35,$FE,$3C,$6B,$FE,$97,$6B,$FE
	.dc.b	$96,$6B,$FE,$C5,$6B,$FD,$DD,$BC,$EF,$AB,$FF,$57,$CF,$E8,$DA,$FE
	.dc.b	$8D,$AF,$E4,$23,$87,$A5,$F9,$36,$C7,$18,$C4,$F0,$8F,$DA,$CE,$9F
	.dc.b	$D7,$7C,$8B,$EB,$B9,$36,$3F,$68,$D7,$71,$F1,$6B,$FC,$E1,$1F,$E5
	.dc.b	$E6,$BF,$EB,$96,$BF,$EB,$86,$BE,$29,$FF,$4D,$BD,$BE,$1B,$C2,$47
	.dc.b	$04,$3C,$EE,$5F,$EA,$7F,$35,$F5,$FC,$A3,$49,$1F,$C3,$7E,$EB,$FB
	.dc.b	$A3,$7C,$EF,$91,$18,$8F,$E2,$0D,$7E,$73,$5F,$92,$D5,$9A,$30,$D9
	.dc.b	$FC,$57,$A2,$F8,$7A,$4B,$FF,$42,$C6,$5F,$37,$E2,$9E,$42,$38,$21
	.dc.b	$E6,$72,$D3,$9F,$27,$80,$8E,$0F,$F3,$D7,$DD,$BE,$BB,$F8,$D7,$C0
	.dc.b	$23,$13,$A9,$CC,$62,$3F,$DE,$F3,$B9,$C9,$23,$81,$FE,$A0,$FA,$C7
	.dc.b	$D7,$7E,$AD,$B1,$98,$C1,$DB,$E9,$E4,$79,$9A,$F8,$30,$46,$23,$F9
	.dc.b	$8E,$1E,$98,$C4,$7F,$52,$E9,$61,$FB,$06,$F4,$F8,$C9,$3F,$E9,$DF
	.dc.b	$89,$FF,$34,$B7,$E0,$3E,$2C,$62,$3E,$0B,$5F,$0B,$5F,$F5,$6B,$56
	.dc.b	$28,$C3,$67,$E8,$FD,$57,$F8,$CA,$BF,$D5,$CD,$7E,$7B,$5F,$7B,$5F
	.dc.b	$F5,$DF,$C9,$7F,$B4,$9F,$F5,$B3,$5F,$AE,$D7,$E3,$B5,$F3,$FB,$5E
	.dc.b	$E2,$3F,$0D,$FC,$2F,$EB,$4F,$2E,$92,$7F,$D8,$91,$97,$FB,$27,$D0
	.dc.b	$FF,$1E,$4D,$26,$9F,$D9,$9B,$1C,$23,$07,$F6,$DF,$65,$F4,$FF,$EA
	.dc.b	$B9,$FF,$1C,$B5,$FF,$1F,$B5,$FE,$E7,$1C,$3D,$3F,$E7,$5F,$31,$FE
	.dc.b	$9D,$9F,$F7,$9B,$5F,$EF,$27,$FA,$A5,$C3,$FD,$F3,$85,$18,$8F,$C6
	.dc.b	$35,$E1,$F0,$7D,$88,$FF,$C0,$0F,$A6,$FE,$B8,$FF,$AD,$5B,$46,$93
	.dc.b	$E1,$BF,$A8,$36,$79,$47,$F8,$33,$E9,$7D,$D1,$FE,$10,$6B,$FE,$30
	.dc.b	$6A,$69,$31,$E4,$48,$F8,$27,$FB,$27,$F5,$A7,$8D,$FD,$85,$B1,$AC
	.dc.b	$51,$88,$F9,$31,$89,$FE,$D1,$E2,$3E,$D6,$92,$B9,$46,$5B,$19,$79
	.dc.b	$C6,$5B,$1F,$DC,$FE,$81,$F2,$94,$93,$E5,$8C,$B6,$32,$F9,$A3,$2D
	.dc.b	$8F,$C9,$7D,$44,$8F,$CC,$B5,$F8,$BC,$88,$C4,$7E,$22,$30,$F0,$20
	.dc.b	$F5,$3F,$CA,$8F,$FB,$D7,$E9,$3F,$C4,$5E,$CF,$F3,$FD,$51,$8F,$23
	.dc.b	$FC,$97,$87,$4B,$E5,$EE,$FF,$1B,$E4,$FF,$01,$E3,$C5,$B7,$9F,$47
	.dc.b	$FC,$A7,$39,$F5,$AC,$E7,$0F,$F2,$CB,$5F,$F2,$C3,$67,$84,$68,$67
	.dc.b	$FD,$B3,$E7,$BF,$CB,$F9,$7E,$97,$89,$18,$AF,$E6,$11,$87,$FD,$07
	.dc.b	$3B,$9D,$47,$49,$D3,$F1,$8C,$46,$AB,$C1,$1C,$8E,$B1,$C9,$FE,$1B
	.dc.b	$8D,$D3,$FD,$5B,$1B,$3F,$D2,$6E,$1E,$C7,$03,$45,$32,$EF,$0F,$BD
	.dc.b	$3D,$49,$3F,$16,$DF,$9C,$3E,$AB,$E7,$CF,$C8,$8C,$3F,$2C,$ED,$9F
	.dc.b	$E0,$7A,$1D,$9C,$30,$FD,$C3,$63,$35,$DD,$FE,$2F,$5C,$67,$44,$9D
	.dc.b	$1C,$ED,$EE,$D2,$47,$A6,$1F,$1B,$88,$FC,$19,$E9,$9C,$5F,$E5,$EB
	.dc.b	$73,$EB,$D7,$F0,$3E,$3D,$DF,$FA,$FC,$F2,$3F,$A0,$B8,$BF,$F0,$F7
	.dc.b	$C0,$3E,$AB,$FD,$25,$1F,$E2,$1A,$DD,$F3,$23,$E4,$AE,$1E,$EF,$CE
	.dc.b	$8D,$9D,$F1,$F1,$B7,$1F,$07,$27,$FB,$13,$2E,$EF,$EB,$A3,$33,$F8
	.dc.b	$66,$BD,$32,$F3,$D7,$08,$D6,$5E,$49,$F4,$61,$E0,$77,$8F,$96,$31
	.dc.b	$8F,$7A,$E0,$F8,$E7,$58,$F8,$2B,$78,$76,$3E,$06,$3E,$17,$C9,$7C
	.dc.b	$DF,$76,$1F,$B7,$CB,$E2,$C7,$63,$96,$5E,$6C,$7D,$F3,$A9,$E2,$49
	.dc.b	$D5,$73,$C3,$C1,$8F,$92,$BC,$27,$4C,$7D,$4F,$73,$26,$B0,$ED,$87
	.dc.b	$48,$ED,$3E,$98,$F8,$F1,$C6,$BB,$1F,$0A,$B8,$D6,$C9,$3E,$13,$FD
	.dc.b	$EF,$7E,$6C,$7E,$51,$31,$F5,$65,$F3,$9E,$D1,$F0,$A7,$84,$60,$DD
	.dc.b	$53,$CE,$76,$AE,$4E,$0F,$74,$F2,$71,$8F,$A6,$31,$97,$C1,$26,$3F
	.dc.b	$22,$B7,$8F,$9D,$FB,$F8,$76,$AD,$C7,$BF,$2F,$86,$FE,$14,$FE,$A5
	.dc.b	$5F,$B0,$6E,$5F,$E3,$28,$C4,$7F,$4C,$F8,$BF,$E1,$6F,$F5,$96,$B9
	.dc.b	$DF,$FC,$73,$18,$8F,$E5,$2F,$F1,$2C,$3F,$3E,$DC,$92,$C5,$F5,$4F
	.dc.b	$2B,$FB,$0E,$1C,$31,$FE,$11,$8F,$1C,$3A,$C7,$E6,$6F,$87,$36,$A2
	.dc.b	$4F,$EF,$6B,$86,$3D,$B1,$EB,$97,$A6,$B5,$7C,$32,$FE,$1F,$97,$49
	.dc.b	$C9,$23,$C8,$FF,$0D,$8D,$46,$09,$5D,$5E,$77,$E0,$AF,$49,$CE,$74
	.dc.b	$7C,$4C,$3C,$D9,$78,$E3,$33,$C2,$F9,$CE,$63,$13,$AA,$E3,$1D,$4C
	.dc.b	$91,$CD,$79,$A3,$27,$F9,$C3,$2F,$8F,$8F,$B2,$FC,$33,$9C,$BC,$18
	.dc.b	$72,$9C,$EE,$C3,$EB,$5E,$CF,$8D,$9F,$C1,$71,$97,$85,$B7,$F0,$72
	.dc.b	$F0,$47,$93,$0E,$99,$78,$27,$33,$C0,$D6,$3E,$8C,$7B,$5F,$09,$F8
	.dc.b	$46,$F2,$F4,$3C,$30,$E9,$5C,$2F,$D3,$7C,$72,$F8,$D8,$70,$C3,$C0
	.dc.b	$49,$D6,$5E,$9C,$39,$63,$F5,$6B,$C1,$97,$A2,$F7,$1E,$13,$E2,$E7
	.dc.b	$B8,$CC,$70,$BE,$6F,$AA,$38,$E3,$E3,$79,$F2,$3E,$D1,$F1,$3F,$8F
	.dc.b	$8C,$36,$1F,$87,$58,$99,$7F,$85,$67,$8B,$67,$38,$FF,$7E,$B7,$D6
	.dc.b	$7D,$66,$E7,$DE,$3F,$C7,$B7,$FE,$1F,$74,$48,$D7,$0E,$DF,$C2,$7D
	.dc.b	$4B,$E0,$7D,$1C,$EF,$85,$CC,$61,$B8,$FF,$8C,$AB,$83,$67,$FC,$D2
	.dc.b	$FF,$AD,$A3,$39,$FF,$8E,$9B,$19,$9D,$CF,$F1,$41,$3A,$70,$D7,$FC
	.dc.b	$BB,$E2,$BE,$9B,$8C,$61,$BA,$A7,$58,$7F,$93,$63,$13,$A9,$CD,$7F
	.dc.b	$96,$E7,$35,$B9,$CB,$FE,$CD,$6C,$FF,$2A,$9F,$F5,$6F,$27,$DF,$F2
	.dc.b	$7F,$C8,$B7,$FE,$4F,$7F,$CC,$5C,$FF,$E8,$97,$37,$F6,$DB,$19,$AE
	.dc.b	$53,$CD,$FA,$BE,$4F,$FA,$53,$C6,$FF,$4A,$E3,$FE,$92,$FA,$BC,$E5
	.dc.b	$FC,$F2,$73,$E3,$A7,$38,$FF,$B2,$36,$FF,$63,$DF,$FB,$2A,$3E,$7E
	.dc.b	$3F,$E6,$E7,$FA,$2C,$E9,$E1,$19,$77,$3C,$67,$51,$F7,$E7,$96,$2F
	.dc.b	$D7,$9F,$6F,$86,$E1,$FE,$D5,$6D,$FF,$B5,$1A,$FF,$BE,$DA,$FC,$F6
	.dc.b	$BF,$E9,$E6,$D7,$7A,$F0,$EE,$FA,$DA,$4C,$3F,$DA,$0D,$C3,$FD,$9C
	.dc.b	$D7,$EF,$3E,$AF,$21,$F6,$34,$95,$FE,$E4,$7E,$27,$87,$FD,$5E,$FF
	.dc.b	$34,$1C,$3F,$9D,$6B,$FD,$C9,$AE,$FA,$39,$92,$BE,$B4,$F0,$CD,$F5
	.dc.b	$CF,$5F,$E7,$B6,$BF,$E0,$03,$70,$FF,$7F,$B5,$FD,$75,$72,$AE,$24
	.dc.b	$AE,$2F,$DA,$F3,$7C,$04,$C7,$04,$9F,$7D,$FF,$B4,$27,$51,$8B,$47
	.dc.b	$A7,$D4,$7D,$17,$E1,$9B,$5B,$6D,$FC,$AB,$46,$23,$F8,$67,$85,$F9
	.dc.b	$3E,$4F,$E4,$F1,$F9,$D8,$C1,$2F,$F2,$CD,$8D,$47,$A3,$C8,$7E,$37
	.dc.b	$FC,$42,$30,$D9,$FB,$73,$AB,$FD,$33,$6B,$97,$D2,$FC,$DA,$FF,$16
	.dc.b	$CA,$31,$08,$C6,$DC,$A7,$30,$F9,$07,$C2,$76,$6E,$1E,$76,$E1,$E4
	.dc.b	$6B,$FC,$A6,$9F,$4D,$F9,$1A,$F7,$36,$7D,$6D,$78,$35,$F4,$E1,$F3
	.dc.b	$DB,$BF,$9C,$95,$F0,$1B,$1F,$1A,$78,$3E,$F6,$EE,$F8,$07,$D8,$7F
	.dc.b	$06,$6E,$1F,$87,$6D,$FE,$05,$AF,$B7,$E9,$9F,$5B,$F4,$9B,$5F,$6D
	.dc.b	$B3,$F6,$1A,$F1,$9E,$04,$9E,$0F,$8F,$DC,$91,$E3,$6C,$F3,$FA,$1F
	.dc.b	$01,$27,$C0,$D8,$D4,$60,$8F,$E6,$FE,$31,$F0,$9F,$C9,$9B,$3F,$A7
	.dc.b	$6C,$EE,$7F,$9A,$7C,$77,$F2,$AF,$B1,$BE,$11,$A4,$C3,$F6,$73,$B6
	.dc.b	$CE,$DB,$3B,$BF,$27,$D1,$F8,$C9,$3E,$37,$BF,$CC,$C8,$E6,$D9,$FE
	.dc.b	$3F,$87,$F3,$E8,$C4,$7F,$7B,$8C,$4F,$03,$F8,$27,$8D,$FC,$6C,$9D
	.dc.b	$FF,$D2,$1F,$E4,$B2,$FB,$CF,$B3,$7F,$E9,$88,$C7,$07,$C4,$9C,$CA
	.dc.b	$73,$87,$C2,$9C,$C6,$27,$CF,$18,$38,$F2,$3E,$51,$C3,$FD,$69,$B8
	.dc.b	$7F,$1B,$6D,$FF,$15,$6B,$EF,$BA,$F9,$DA,$F9,$5B,$87,$B6,$FE,$EE
	.dc.b	$5F,$56,$BE,$5B,$5F,$94,$D7,$E1,$35,$F8,$0D,$93,$49,$B7,$F7,$4F
	.dc.b	$90,$7C,$7D,$26,$7F,$CD,$E3,$2D,$8C,$E1,$FC,$CE,$72,$D8,$FF,$04
	.dc.b	$79,$AF,$CE,$BF,$CC,$57,$FD,$12,$31,$EC,$EA,$73,$7F,$D9,$A7,$35
	.dc.b	$B9,$CC,$FF,$90,$E7,$2E,$E3,$2F,$F8,$7D,$AF,$FB,$73,$EB,$DF,$2F
	.dc.b	$FE,$17,$6D,$FF,$87,$DB,$3F,$D6,$1A,$FE,$5F,$E8,$1D,$5F,$B0,$6D
	.dc.b	$7E,$B1,$B3,$FA,$06,$BC,$67,$81,$27,$83,$FE,$29,$8F,$F7,$8F,$A9
	.dc.b	$FE,$31,$69,$5C,$27,$FC,$6D,$E1,$BE,$0D,$F3,$B8,$6D,$7F,$72,$9D
	.dc.b	$4E,$4E,$3E,$D3,$6A,$4A,$E1,$19,$7F,$BD,$FC,$47,$F8,$65,$FF,$23
	.dc.b	$35,$CF,$A1,$E9,$8C,$47,$F8,$81,$AF,$F9,$ED,$A8,$38,$6C,$7C,$3F
	.dc.b	$9D,$7C,$75,$24,$FF,$95,$E3,$2F,$F9,$67,$E7,$5F,$31,$F2,$DB,$19
	.dc.b	$8C,$11,$C1,$0F,$F0,$CC,$70,$7F,$C2,$1F,$3A,$F9,$4F,$F3,$37,$80
	.dc.b	$8C,$4F,$F8,$75,$AF,$09,$DC,$62,$3F,$C5,$ED,$7F,$D1,$0D,$53,$D6
	.dc.b	$7F,$CE,$9D,$AF,$9D,$7C,$86,$1B,$7F,$CF,$9A,$FF,$32,$6B,$FC,$C5
	.dc.b	$AF,$F2,$96,$BF,$C9,$DB,$3D,$E7,$C3,$DF,$FE,$C3,$AF,$E9,$6D,$7F
	.dc.b	$8A,$FC,$1E,$8F,$8B,$6F,$1E,$92,$7F,$C8,$51,$97,$9F,$67,$62,$38
	.dc.b	$21,$FC,$89,$CB,$4F,$E6,$5E,$DF,$D4,$F1,$4D,$24,$7F,$62,$F9,$D7
	.dc.b	$CE,$FC,$C6,$C6,$63,$04,$70,$43,$FA,$EC,$70,$7F,$BD,$7C,$6F,$E1
	.dc.b	$3C,$E3,$49,$1E,$1F,$9D,$F0,$11,$C1,$0E,$AE,$5A,$7F,$80,$F9,$7B
	.dc.b	$11,$C1,$FE,$59,$F9,$DF,$F2,$3B,$7D,$8F,$F5,$E7,$B0,$9C,$CF,$F9
	.dc.b	$69,$AF,$F9,$1D,$AF,$F9,$61,$AF,$F9,$5D,$B1,$CE,$7F,$47,$F3,$AF
	.dc.b	$2E,$92,$FF,$D0,$11,$97,$CD,$E5,$F9,$30,$FE,$A3,$18,$21,$E5,$8D
	.dc.b	$11,$C9,$FE,$98,$9F,$D4,$74,$9A,$4E,$9F,$83,$F3,$BF,$E4,$FA,$FF
	.dc.b	$3F,$35,$FF,$3D,$B5,$FF,$34,$35,$FF,$3C,$B5,$FF,$3C,$37,$33,$4C
	.dc.b	$7C,$EF,$F8,$22,$7F,$D1,$8D,$7F,$D1,$2D,$7F,$CE,$CD,$7F,$D0,$FE
	.dc.b	$93,$EB,$FE,$CB,$63,$FD,$2C,$D7,$FD,$26,$D7,$D3,$EB,$7C,$02,$3F
	.dc.b	$01,$FC,$1F,$CF,$7F,$B1,$E3,$FD,$40,$D7,$FD,$3A,$D7,$FD,$14,$D7
	.dc.b	$F9,$C0,$F9,$2F,$E3,$0D,$CF,$FA,$D4,$60,$8E,$09,$1F,$EC,$63,$81
	.dc.b	$FE,$C8,$F8,$C7,$CF,$7F,$70,$8C,$35,$FF,$75,$C6,$BC,$FE,$4D,$8F
	.dc.b	$F6,$47,$CF,$7F,$36,$F8,$63,$49,$1A,$21,$FE,$CD,$8D,$B8,$3F,$DC
	.dc.b	$BF,$0F,$FD,$39,$FF,$69,$35,$CF,$95,$FD,$12,$31,$1F,$D7,$79,$3F
	.dc.b	$DB,$F3,$92,$47,$FB,$8E,$76,$FF,$63,$F9,$AF,$73,$F3,$83,$87,$E8
	.dc.b	$B5,$E1,$97,$F8,$26,$31,$1C,$23,$04,$7F,$DE,$5F,$21,$FE,$44,$6F
	.dc.b	$8F,$FD,$72,$31,$1F,$EE,$F6,$BF,$80,$C7,$F9,$6F,$CF,$7F,$62,$F0
	.dc.b	$CD,$24,$FF,$BF,$E3,$2D,$8C,$BF,$EF,$B8,$CB,$FE,$FD,$F8,$E7,$CF
	.dc.b	$7F,$6E,$6C,$7C,$33,$FB,$2B,$87,$84,$6C,$8E,$ED,$3B,$8F,$53,$5F
	.dc.b	$CD,$F1,$BE,$7B,$F3,$24,$8F,$CB,$46,$5B,$19,$7C,$91,$96,$C6,$7D
	.dc.b	$47,$B3,$F8,$3F,$93,$FC,$57,$B5,$FD,$1A,$FF,$BF,$C6,$3C,$7F,$1E
	.dc.b	$1F,$E5,$FB,$FC,$57,$1F,$F7,$CC,$38,$68,$8C,$5E,$BD,$0F,$DA,$D6
	.dc.b	$A7,$FC,$5B,$97,$F9,$3A,$B4,$D3,$FC,$1F,$1C,$9B,$5C,$A7,$C3,$EB
	.dc.b	$59,$FF,$0A,$F0,$3E,$5B,$F8,$1E,$7F,$85,$8C,$4F,$F8,$F6,$31,$3A
	.dc.b	$8F,$D2,$4E,$9F,$8A,$62,$34,$F6,$6E,$7F,$DF,$1B,$3C,$6B,$F4,$DB
	.dc.b	$FB,$AF,$99,$FE,$4D,$71,$1A,$C3,$AC,$FF,$19,$6C,$EA,$3F,$07,$18
	.dc.b	$23,$EB,$32,$7C,$9C,$1F,$2C,$F9,$EF,$8D,$18,$7E,$44,$F7,$38,$74
	.dc.b	$3B,$36,$D8,$CC,$61,$CB,$1C,$9C,$8E,$F7,$FE,$83,$7C,$8D,$7F,$D1
	.dc.b	$1A,$FF,$13,$8C,$70,$7E,$29,$7C,$F7,$2F,$7A,$5F,$27,$AC,$7A,$B5
	.dc.b	$FC,$07,$8F,$F8,$9A,$FE,$85,$C2,$8C,$4E,$A7,$C2,$3E,$6D,$F3,$DF
	.dc.b	$E8,$9C,$FF,$45,$3C,$AF,$8E,$1C,$E3,$89,$93,$9D,$F2,$F7,$7C,$93
	.dc.b	$BB,$E7,$8F,$85,$B1,$9D,$7F,$D2,$17,$C7,$0E,$F8,$C7,$BE,$1C,$6B
	.dc.b	$79,$79,$DF,$E0,$F8,$73,$AF,$C7,$35,$F3,$63,$FB,$DC,$7B,$47,$49
	.dc.b	$D6,$1D,$63,$8E,$3F,$B3,$AD,$DF,$A2,$3C,$8D,$9F,$83,$8F,$8D,$E2
	.dc.b	$75,$9D,$E3,$E8,$CB,$E2,$C6,$27,$77,$E5,$BE,$51,$E8,$BF,$6D,$6F
	.dc.b	$0F,$1C,$62,$7B,$12,$F9,$61,$DA,$31,$3B,$9C,$C7,$84,$EC,$78,$A3
	.dc.b	$CC,$E7,$1F,$88,$EA,$FC,$2E,$C8,$67,$1F,$4C,$F0,$26,$5F,$56,$B8
	.dc.b	$D7,$80,$93,$F4,$63,$2F,$0C,$BE,$05,$F4,$BF,$AE,$48,$FA,$D1,$97
	.dc.b	$D1,$1D,$E3,$17,$F0,$8F,$45,$F6,$9E,$51,$98,$C5,$72,$3A,$E5,$F2
	.dc.b	$F2,$FC,$01,$32,$F8,$F7,$F4,$8F,$E4,$38,$7E,$9D,$B7,$FD,$62,$31
	.dc.b	$1F,$D7,$BE,$7B,$F8,$73,$FC,$A1,$AE,$77,$FE,$BF,$18,$8F,$E5,$4F
	.dc.b	$F1,$5C,$3F,$5A,$DC,$3F,$57,$F3,$CF,$27,$F9,$DD,$6B,$0F,$E0,$77
	.dc.b	$BC,$39,$C7,$E7,$EF,$87,$36,$A2,$4F,$EC,$AB,$86,$AB,$DC,$E6,$B7
	.dc.b	$87,$78,$FE,$0B,$3A,$8C,$13,$2F,$74,$EB,$2F,$3C,$62,$75,$3E,$63
	.dc.b	$59,$76,$9D,$12,$3F,$8E,$5F,$78,$C4,$7F,$99,$72,$F8,$0D,$C7,$E0
	.dc.b	$63,$E0,$8E,$D8,$F8,$1D,$1E,$03,$DC,$47,$CB,$97,$C3,$9E,$53,$C2
	.dc.b	$79,$6E,$C3,$E4,$E5,$E5,$AE,$F9,$79,$AB,$59,$79,$33,$F7,$56,$89
	.dc.b	$9F,$78,$CC,$77,$BF,$85,$39,$8F,$34,$75,$C7,$C3,$9F,$9B,$1E,$F5
	.dc.b	$A3,$AE,$5E,$57,$A1,$E1,$2E,$1D,$73,$F6,$E1,$F3,$2F,$9C,$7F,$2F
	.dc.b	$BE,$51,$8C,$7C,$59,$FB,$71,$F0,$1D,$8E,$B5,$C3,$0E,$84,$9E,$18
	.dc.b	$78,$1F,$8B,$8F,$8F,$0E,$B8,$7A,$E3,$0F,$A7,$1F,$13,$EE,$CB,$E1
	.dc.b	$63,$FA,$3E,$57,$C3,$3E,$5B,$F0,$91,$86,$C3,$E7,$AC,$4D,$FF,$C2
	.dc.b	$53,$C5,$B3,$9C,$7F,$B8,$DB,$E7,$BF,$04,$DC,$BB,$C7,$F8,$EE,$FF
	.dc.b	$C3,$8E,$89,$1A,$E1,$DB,$F9,$0F,$97,$7C,$07,$A3,$A5,$F0,$B9,$8C
	.dc.b	$37,$1F,$F1,$56,$61,$7C,$A3,$31,$8A,$FF,$22,$C7,$FA,$8E,$75,$3F
	.dc.b	$E3,$E3,$51,$C2,$32,$EE,$78,$CF,$FB,$62,$72,$FF,$AD,$7C,$47,$CF
	.dc.b	$6A,$3F,$D1,$AD,$D7,$FD,$12,$D9,$DE,$1E,$A9,$CB,$63,$35,$FE,$52
	.dc.b	$9C,$B6,$33,$3F,$EA,$BF,$23,$FC,$84,$FF,$92,$DB,$3F,$61,$FF,$28
	.dc.b	$73,$FF,$A0,$1C,$B7,$9F,$F0,$F1,$8E,$7F,$F6,$83,$C1,$B1,$F5,$FC
	.dc.b	$9F,$F3,$7F,$8F,$FE,$70,$CD,$E3,$BE,$7B,$5A,$7D,$5A,$D4,$E6,$3F
	.dc.b	$C8,$1B,$B0,$FA,$5A,$FD,$B8,$C4,$FF,$BF,$5B,$1F,$EB,$96,$BF,$6E
	.dc.b	$78,$BF,$82,$6C,$6A,$3E,$C4,$76,$C5,$F3,$CF,$93,$C3,$70,$FF,$60
	.dc.b	$36,$FF,$DD,$55,$C2,$B6,$4A,$DB,$67,$FD,$A5,$F3,$DD,$F8,$4D,$25
	.dc.b	$7F,$B3,$63,$2D,$8C,$CF,$28,$FF,$71,$7C,$F7,$A7,$C9,$34,$91,$FE
	.dc.b	$DD,$97,$87,$FD,$79,$FD,$31,$FA,$77,$0E,$1A,$E7,$A7,$E5,$C6,$23
	.dc.b	$9C,$60,$93,$FE,$E5,$F9,$E7,$CE,$FC,$D6,$C6,$63,$04,$70,$4A,$F9
	.dc.b	$B3,$C1,$F9,$FE,$6F,$93,$6F,$F6,$B6,$1F,$66,$30,$47,$04,$BF,$F7
	.dc.b	$C4,$71,$F4,$DF,$39,$F8,$16,$CF,$FB,$F5,$B8,$7D,$E6,$BF,$C1,$FC
	.dc.b	$3F,$C5,$92,$7F,$28,$D8,$FC,$A3,$5E,$11,$F4,$1F,$E2,$DE,$43,$E7
	.dc.b	$BF,$8B,$CE,$67,$FD,$41,$18,$78,$4E,$F9,$BF,$4A,$DA,$E0,$BE,$77
	.dc.b	$E4,$57,$DE,$8C,$6F,$C2,$73,$3E,$26,$BF,$EA,$FF,$92,$7C,$27,$66
	.dc.b	$E1,$D9,$B3,$E4,$6B,$DF,$E7,$BF,$1F,$3E,$46,$C7,$A9,$AE,$F8,$FC
	.dc.b	$9B,$CA,$F5,$E3,$F4,$2B,$CD,$39,$7E,$1C,$65,$FF,$63,$37,$3F,$85
	.dc.b	$3F,$BD,$F9,$D7,$CF,$52,$61,$F3,$F9,$13,$A8,$F9,$7F,$3C,$F9,$DF
	.dc.b	$9E,$D8,$CC,$60,$8E,$09,$9F,$D8,$9E,$0F,$F0,$8A,$7B,$9E,$52,$57
	.dc.b	$8E,$32,$F8,$BE,$73,$C2,$47,$04,$AE,$D1,$96,$C7,$B7,$E3,$5F,$09
	.dc.b	$FC,$7D,$B1,$98,$C1,$FC,$C3,$E5,$3F,$91,$7C,$E5,$F0,$8D,$25,$7E
	.dc.b	$5E,$76,$D9,$DB,$67,$78,$7F,$28,$C1,$F2,$BE,$32,$57,$9D,$F0,$7C
	.dc.b	$F4,$8E,$8D,$AF,$D5,$63,$FC,$DE,$31,$1F,$C7,$23,$13,$C0,$FB,$DE
	.dc.b	$37,$F0,$92,$78,$1F,$D0,$5F,$E3,$F2,$F9,$EF,$AF,$7C,$DA,$FF,$83
	.dc.b	$63,$5C,$5F,$1E,$73,$29,$CE,$1F,$16,$73,$18,$9F,$4C,$70,$E3,$7C
	.dc.b	$AB,$87,$FA,$A3,$70,$FC,$B3,$6F,$FA,$8B,$5F,$34,$F8,$C9,$3E,$17
	.dc.b	$EE,$65,$F5,$27,$E4,$B5,$F9,$0D,$7D,$CD,$7D,$8D,$C7,$FB,$B6,$FD
	.dc.b	$09,$7F,$AB,$F9,$17,$C8,$7F,$6D,$6C,$66,$30,$47,$04,$C3,$FB,$6C
	.dc.b	$F0,$7F,$94,$79,$8F,$9E,$FF,$2E,$D7,$F2,$78,$C7,$B5,$FE,$08,$6C
	.dc.b	$66,$77,$5C,$9A,$FF,$62,$9D,$B9,$AE,$53,$A8,$FE,$CF,$F3,$CF,$98
	.dc.b	$FF,$06,$B7,$0F,$ED,$CD,$9F,$E7,$4D,$7F,$29,$F3,$DF,$BC,$7F,$32
	.dc.b	$D7,$3D,$7F,$B1,$8C,$47,$E3,$DA,$FE,$35,$BB,$1A,$6A,$3F,$BE,$FA
	.dc.b	$DF,$E0,$17,$FC,$C8,$D4,$D3,$FC,$0D,$E2,$3E,$11,$F3,$D8,$6D,$7F
	.dc.b	$55,$9D,$4E,$4E,$3E,$DB,$82,$92,$B8,$46,$5F,$EC,$5F,$44,$FB,$BF
	.dc.b	$F7,$F4,$7F,$91,$77,$FF,$3D,$4E,$7B,$3E,$B4,$62,$B7,$5D,$E3,$FC
	.dc.b	$37,$F1,$5F,$E1,$E8,$E4,$2F,$93,$86,$BC,$4D,$57,$4C,$57,$D3,$C9
	.dc.b	$F0,$D7,$FA,$37,$E7,$DF,$35,$F3,$5B,$1F,$E2,$F6,$BF,$DE,$DA,$FE
	.dc.b	$3F,$F4,$9F,$33,$E1,$C1,$18,$7F,$CD,$6D,$C3,$FC,$6F,$5D,$27,$38
	.dc.b	$79,$1E,$2D,$AE,$17,$E7,$C3,$D3,$F4,$BE,$F8,$F8,$0D,$BF,$7C,$23
	.dc.b	$18,$FF,$AA,$FC,$07,$D0,$BE,$5F,$9E,$1C,$30,$E6,$D3,$9E,$1C,$A1
	.dc.b	$7C,$AB,$C7,$F1,$6F,$BA,$A4,$AF,$B3,$D2,$FA,$27,$C9,$E6,$7D,$37
	.dc.b	$C5,$FB,$F9,$23,$11,$F7,$6F,$95,$7D,$AC,$BA,$79,$5F,$E6,$56,$D7
	.dc.b	$F8,$0F,$1F,$EA,$AD,$9F,$F3,$56,$2F,$AA,$7D,$E7,$CC,$6D,$78,$9A
	.dc.b	$FE,$37,$0E,$AD,$F2,$5F,$01,$E1,$6C,$7F,$3E,$8C,$11,$C1,$0F,$E8
	.dc.b	$2F,$92,$9F,$5F,$E3,$24,$FE,$8A,$FF,$A6,$D7,$C1,$F3,$DF,$11,$F0
	.dc.b	$5B,$5F,$E0,$78,$C1,$1C,$10,$FE,$F4,$FC,$4F,$AD,$FC,$A7,$37,$FB
	.dc.b	$C7,$54,$62,$BF,$18,$DC,$58,$78,$27,$E3,$3F,$A6,$7F,$20,$D7,$F1
	.dc.b	$2D,$8E,$3C,$FE,$92,$4F,$F8,$7D,$AF,$F9,$BE,$FC,$78,$FE,$8B,$0F
	.dc.b	$07,$37,$E5,$9B,$97,$83,$0F,$C3,$7D,$13,$CB,$FC,$33,$72,$FF,$24
	.dc.b	$35,$FF,$61,$B5,$FF,$56,$E6,$8C,$47,$EA,$DA,$FE,$91,$B1,$B9,$FC
	.dc.b	$7F,$D7,$BE,$F3,$F7,$2D,$BF,$CF,$B5,$FC,$75,$F8,$F3,$FF,$79,$3F
	.dc.b	$B7,$8F,$3F,$57,$A8,$8E,$08,$7A,$9C,$BF,$26,$9F,$3B,$E2,$24,$7E
	.dc.b	$36,$BB,$3F,$CA,$FD,$C7,$D1,$FF,$86,$E3,$FD,$2A,$D7,$FD,$36,$D7
	.dc.b	$FD,$2C,$D7,$FD,$4A,$D5,$1D,$3D,$4F,$F6,$E3,$63,$AC,$7E,$EF,$EE
	.dc.b	$BF,$83,$C7,$E8,$5A,$FE,$CD,$B9,$FF,$20,$AE,$53,$E9,$FA,$5F,$F3
	.dc.b	$3B,$FD,$2A,$11,$88,$D7,$B7,$FD,$12,$5E,$4B,$E9,$71,$9F,$F7,$19
	.dc.b	$F8,$77,$E8,$8C,$1F,$DD,$79,$DF,$37,$FC,$63,$3F,$E5,$2D,$E6,$FD
	.dc.b	$63,$6F,$E1,$FD,$2F,$FB,$B9,$FE,$BF,$F2,$6F,$A5,$FF,$46,$C7,$FB
	.dc.b	$2E,$31,$EA,$FF,$B3,$61,$18,$75,$3C,$1F,$F6,$B4,$65,$B1,$94,$48
	.dc.b	$D3,$83,$AC,$6D,$DE,$0F,$A9,$FD,$7C,$7F,$6E,$6C,$67,$B7,$FB,$7D
	.dc.b	$F8,$73,$FE,$97,$EB,$FF,$85,$70,$FE,$7E,$D7,$FA,$DB,$63,$FA,$DF
	.dc.b	$63,$E3,$5F,$4D,$FD,$8A,$B5,$97,$F1,$76,$E1,$FD,$85,$B7,$C6,$BF
	.dc.b	$B4,$FD,$B3,$E9,$29,$2F,$F9,$77,$C8,$BE,$9F,$11,$FD,$BA,$B8,$3F
	.dc.b	$D5,$9A,$FF,$45,$9E,$0F,$F3,$7F,$9A,$7C,$27,$F5,$76,$C6,$63,$FC
	.dc.b	$26,$BE,$97,$EC,$6F,$F0,$2B,$73,$FD,$45,$B1,$9A,$E3,$5C,$36,$FE
	.dc.b	$71,$F6,$FA,$6C,$FF,$82,$23,$18,$7F,$7A,$C2,$F4,$3E,$4F,$ED,$B7
	.dc.b	$93,$FC,$1F,$1F,$7B,$87,$FA,$FF,$87,$89,$98,$7F,$78,$AD,$79,$7F
	.dc.b	$C3,$A3,$FB,$DF,$DC,$3E,$63,$FC,$32,$D8,$CC,$7F,$7B,$F9,$A7,$D4
	.dc.b	$EA,$24,$E8,$FE,$C5,$3A,$7F,$C8,$A3,$E9,$9F,$53,$FE,$05,$F9,$FF
	.dc.b	$A7,$3F,$E6,$6D,$3F,$C4,$51,$8F,$88,$FE,$B1,$7C,$2B,$9B,$5F,$F0
	.dc.b	$04,$EE,$B8,$F4,$FF,$9D,$1F,$D3,$56,$F9,$FF,$2F,$5B,$BF,$F1,$3B
	.dc.b	$78,$4E,$63,$D1,$97,$F8,$6F,$A3,$F8,$EE,$7F,$DE,$AD,$18,$CB,$FC
	.dc.b	$74,$DB,$FF,$19,$AF,$84,$F2,$46,$3D,$BB,$7C,$9B,$5C,$63,$AD,$6B
	.dc.b	$89,$F4,$9C,$B9,$BF,$09,$E2,$5E,$6F,$E1,$9A,$F9,$6F,$7A,$7F,$A5
	.dc.b	$35,$FF,$27,$46,$23,$E8,$1F,$E5,$3A,$E3,$3F,$E0,$D6,$C7,$F9,$32
	.dc.b	$78,$57,$D4,$CB,$D3,$F4,$9F,$AE,$8F,$88,$DC,$3E,$8C,$FC,$0B,$FC
	.dc.b	$13,$C3,$BE,$47,$94,$FE,$3F,$C7,$F1,$DC,$1C,$0D,$6C,$C3,$B4,$7F
	.dc.b	$98,$1A,$EC,$F6,$C7,$EE,$7C,$3F,$9F,$DF,$FC,$D2,$79,$B5,$DB,$FC
	.dc.b	$B5,$B1,$C0,$8F,$03,$E0,$FC,$54,$7C,$B7,$FC,$9F,$4F,$87,$F5,$39
	.dc.b	$9F,$1C,$E7,$5F,$F4,$03,$63,$D5,$3F,$1F,$3F,$BF,$E5,$FC,$26,$CF
	.dc.b	$D8,$3F,$59,$C7,$FD,$E1,$FB,$7F,$02,$FA,$9F,$F5,$0E,$7F,$AA,$AD
	.dc.b	$37,$0E,$91,$8C,$FF,$A7,$DF,$1F,$86,$F1,$4E,$71,$FA,$AD,$7C,$37
	.dc.b	$F6,$6B,$84,$74,$88,$F9,$F2,$F4,$E3,$FE,$62,$AD,$CE,$6F,$A5,$6B
	.dc.b	$1F,$DC,$C7,$86,$B7,$7E,$9B,$E1,$87,$68,$E9,$97,$C0,$C3,$8C,$E5
	.dc.b	$B5,$F1,$A7,$4F,$B2,$74,$D3,$59,$7C,$28,$CC,$7C,$03,$26,$32,$F4
	.dc.b	$3C,$70,$F1,$D7,$18,$F3,$92,$32,$F1,$CB,$E0,$9E,$FC,$3C,$B1,$DE
	.dc.b	$7E,$01,$31,$FA,$F3,$AC,$3C,$35,$B9,$F7,$D7,$68,$F0,$46,$08,$FF
	.dc.b	$AE,$EB,$95,$F8,$B1,$F8,$79,$7D,$0B,$DD,$F5,$3D,$CF,$8B,$2F,$76
	.dc.b	$3E,$59,$C9,$D6,$33,$3C,$B2,$F8,$F7,$D6,$B6,$75,$3B,$39,$71,$1E
	.dc.b	$E3,$8E,$5E,$8C,$7E,$51,$F9,$37,$13,$F0,$0F,$1D,$73,$AE,$59,$7B
	.dc.b	$E7,$ED,$9F,$58,$FE,$05,$1F,$B0,$6C,$7E,$BD,$B7,$FE,$3F,$97,$D4
	.dc.b	$FF,$0B,$9F,$EA,$2D,$7F,$8E,$B5,$FE,$C8,$D8,$E3,$BF,$EF,$E3,$F8
	.dc.b	$B6,$9F,$A7,$F3,$BF,$CC,$3E,$3E,$6A,$61,$FA,$DB,$DE,$BF,$BA,$BD
	.dc.b	$CE,$6B,$78,$7F,$6C,$BD,$CF,$FB,$EE,$B7,$39,$AD,$DF,$A9,$FD,$F4
	.dc.b	$EA,$31,$5D,$49,$5E,$7F,$45,$8F,$29,$F8,$38,$5D,$7F,$CE,$75,$A9
	.dc.b	$E3,$87,$4C,$3C,$51,$FC,$73,$1F,$03,$C8,$D9,$63,$C1,$8F,$78,$F9
	.dc.b	$B8,$F7,$AE,$31,$83,$89,$6B,$8B,$8C,$FE,$9B,$DB,$2F,$3E,$1D,$67
	.dc.b	$35,$E6,$38,$97,$1F,$51,$2F,$E7,$65,$EF,$9D,$47,$A6,$75,$87,$68
	.dc.b	$F1,$67,$F5,$9F,$34,$FD,$DB,$E4,$F5,$8C,$9C,$CE,$F8,$F9,$EB,$96
	.dc.b	$3E,$8A,$E5,$39,$CF,$E3,$61,$D3,$0E,$47,$43,$AE,$3E,$5C,$BD,$31
	.dc.b	$9C,$7A,$3E,$03,$D3,$7D,$30,$ED,$1F,$2F,$3F,$A7,$5C,$23,$96,$7F
	.dc.b	$94,$C7,$B6,$7E,$DA,$D1,$DA,$33,$5E,$FE,$47,$C8,$BE,$87,$D5,$18
	.dc.b	$6C,$3E,$A6,$C4,$D3,$FC,$1B,$3C,$5B,$39,$BF,$F0,$5F,$D4,$FE,$4B
	.dc.b	$03,$4F,$F1,$AD,$7F,$85,$9C,$70,$78,$9D,$71,$58,$F9,$D7,$FE,$7D
	.dc.b	$FA,$9F,$3F,$3B,$E4,$33,$DC,$9C,$C6,$1B,$97,$F8,$9A,$74,$FF,$84
	.dc.b	$A7,$4D,$35,$3C,$2F,$FC,$4B,$E3,$BE,$A7,$F0,$1E,$17,$FA,$3F,$E7
	.dc.b	$FE,$4C,$E7,$0F,$F1,$B4,$E5,$FF,$19,$C6,$6F,$FC,$97,$39,$7A,$DF
	.dc.b	$4C,$98,$7D,$4F,$3B,$FD,$CB,$8F,$F8,$F1,$B7,$FE,$38,$7F,$DB,$6D
	.dc.b	$E9,$E0,$49,$E1,$14,$FF,$1F,$F9,$0F,$A9,$FC,$0B,$67,$E3,$7A,$3F
	.dc.b	$C3,$FC,$BF,$F4,$1D,$7C,$E8,$C4,$FF,$A4,$A3,$0F,$FB,$CA,$76,$F2
	.dc.b	$30,$6C,$C5,$FF,$A5,$27,$E5,$F9,$FF,$D1,$27,$FD,$5F,$DB,$DA,$9C
	.dc.b	$EF,$A9,$7C,$FE,$78,$98,$F1,$B6,$7F,$C9,$E7,$38,$EE,$B5,$3F,$EF
	.dc.b	$4C,$39,$E5,$E0,$3F,$D8,$5F,$53,$D7,$88,$F8,$F4,$72,$62,$B6,$4D
	.dc.b	$3C,$2F,$FA,$16,$B8,$4E,$E7,$27,$FA,$2E,$3D,$5E,$A3,$CC,$A7,$73
	.dc.b	$FD,$A1,$7F,$2A,$9B,$2B,$84,$E7,$1E,$5E,$97,$13,$E5,$5B,$C8,$FE
	.dc.b	$BF,$B7,$F4,$BA,$DD,$F1,$CB,$BD,$F9,$A7,$2F,$A0,$E5,$D9,$D6,$BC
	.dc.b	$99,$BD,$57,$C3,$FD,$56,$CF,$97,$EA,$5F,$51,$9A,$93,$CA,$B7,$8F
	.dc.b	$86,$F7,$39,$AF,$C0,$7B,$4F,$8F,$A7,$D5,$F5,$7D,$B4,$C3,$E3,$D6
	.dc.b	$E3,$CE,$D8,$FA,$79,$A3,$1B,$FC,$5B,$E7,$5C,$B0,$FC,$A6,$0F,$B9
	.dc.b	$7A,$F8,$6C,$7D,$3F,$75,$F4,$D9,$AF,$C5,$46,$27,$FD,$55,$87,$E7
	.dc.b	$A3,$0F,$FA,$B7,$89,$18,$8F,$61,$BB,$FF,$02,$38,$9D,$C6,$23,$E1
	.dc.b	$46,$2B,$BB,$63,$B5,$75,$7E,$46,$AF,$B8,$7B,$1F,$AF,$6E,$1F,$93
	.dc.b	$F5,$1F,$56,$ED,$D5,$FE,$E6,$5C,$FF,$C5,$B2,$F4,$C6,$1F,$30,$71
	.dc.b	$87,$C4,$6C,$7F,$B3,$D7,$D4,$FE,$12,$3F,$57,$18,$E8,$F0,$7C,$FB
	.dc.b	$EA,$FC,$57,$BE,$6F,$D0,$7C,$53,$E0,$BF,$78,$DC,$FD,$9F,$57,$FA
	.dc.b	$77,$F5,$91,$9D,$3B,$CE,$5F,$23,$5F,$6B,$5F,$C9,$68,$8C,$47,$BD
	.dc.b	$B1,$9E,$FF,$86,$DB,$F8,$53,$C3,$8D,$F1,$EF,$81,$FE,$44,$DA,$FE
	.dc.b	$05,$F5,$5F,$D2,$9F,$DA,$C6,$72,$FD,$E4,$E5,$FC,$73,$5F,$C5,$35
	.dc.b	$FC,$4B,$5F,$E6,$6D,$E8,$34,$9E,$87,$DC,$C7,$CF,$D7,$E6,$3F,$95
	.dc.b	$4B,$E0,$F1,$97,$A1,$FC,$1E,$1D,$1B,$3B,$9F,$CC,$B7,$1F,$4C,$E8
	.dc.b	$93,$A7,$C5,$F2,$9F,$E0,$78,$FE,$2E,$D9,$FE,$3D,$F2,$AF,$8F,$FE
	.dc.b	$99,$39,$E4,$FE,$4D,$39,$BF,$D4,$B6,$FF,$6C,$D7,$F6,$4D,$7F,$62
	.dc.b	$D7,$FA,$DB,$72,$34,$D6,$8C,$1F,$0B,$E4,$3F,$A9,$4B,$E6,$71,$8F
	.dc.b	$F4,$18,$C3,$E6,$8F,$97,$18,$C7,$CF,$18,$24,$EC,$91,$FC,$38,$FE
	.dc.b	$9F,$F3,$1C,$73,$FE,$D8,$D9,$FE,$1B,$EA,$7F,$26,$CD,$39,$BF,$EF
	.dc.b	$33,$FE,$AE,$7F,$8B,$35,$E3,$DB,$F8,$08,$C4,$70,$8C,$12,$FF,$8C
	.dc.b	$CE,$AF,$BE,$DF,$E0,$3C,$3E,$A6,$3F,$D7,$7E,$65,$F2,$99,$AF,$EF
	.dc.b	$51,$8C,$7F,$91,$B7,$3F,$E6,$0D,$7F,$96,$35,$FE,$4B,$F4,$3F,$C8
	.dc.b	$9F,$E9,$2D,$C3,$FD,$6D,$F3,$8D,$1F,$4F,$FE,$17,$AD,$63,$CB,$C3
	.dc.b	$C1,$FE,$1F,$BE,$15,$89,$FE,$DB,$83,$E4,$3F,$C4,$0D,$8F,$F0,$07
	.dc.b	$CD,$BE,$9F,$35,$27,$8B,$FE,$19,$9D,$46,$0F,$F0,$66,$0F,$17,$FA
	.dc.b	$8E,$1E,$A6,$CF,$F8,$F7,$E9,$FF,$D1,$2F,$08,$CE,$CB,$E5,$E6,$7A
	.dc.b	$E7,$38,$30,$E3,$7F,$E2,$2B,$DE,$1F,$E1,$DA,$D7,$C4,$7F,$44,$6C
	.dc.b	$7F,$66,$6D,$7E,$55,$AF,$F6,$4D,$DF,$65,$FE,$95,$8F,$F1,$7C,$63
	.dc.b	$E2,$BE,$77,$D3,$FF,$09,$BF,$F7,$64,$E7,$0F,$F1,$73,$6F,$8E,$1C
	.dc.b	$EB,$EE,$DF,$1B,$FF,$07,$5E,$E7,$FA,$DF,$C8,$7F,$7C,$C3,$FB,$BB
	.dc.b	$5F,$F0,$A3,$53,$BD,$7F,$7F,$F4,$F7,$B8,$D2,$57,$F8,$22,$7F,$D4
	.dc.b	$57,$C7,$E2,$7F,$CC,$33,$96,$BF,$BB,$9D,$BF,$BD,$C5,$F4,$FF,$DC
	.dc.b	$E3,$E3,$B6,$FE,$2B,$73,$FF,$08,$F9,$1F,$E4,$BA,$F8,$F3,$95,$7D
	.dc.b	$3B,$FF,$C2,$F8,$72,$6E,$1C,$BC,$BB,$7F,$E3,$7F,$A7,$FF,$24,$BF
	.dc.b	$25,$AF,$F8,$99,$AF,$F1,$56,$BF,$E9,$1F,$A7,$F2,$3F,$E6,$EC,$11
	.dc.b	$87,$FD,$38,$DB,$FF,$15,$D7,$69,$CD,$F9,$1E,$2D,$AE,$17,$E7,$F6
	.dc.b	$FD,$3E,$1F,$F9,$96,$FF,$D2,$07,$F8,$D0,$B8,$FB,$BC,$77,$D3,$BE
	.dc.b	$8F,$8D,$6E,$BF,$C6,$2D,$C7,$FC,$FE,$DB,$E3,$5E,$0F,$93,$7D,$3D
	.dc.b	$25,$FE,$F3,$E9,$FF,$CA,$AF,$F9,$DA,$76,$D9,$DB,$FB,$39,$DB,$FE
	.dc.b	$81,$7F,$A7,$3F,$E4,$86,$BF,$E5,$56,$BF,$E5,$16,$BF,$E4,$AF,$27
	.dc.b	$FC,$EC,$FF,$A1,$9A,$FF,$9C,$1B,$8F,$FA,$4E,$7A,$BF,$CE,$7E,$0F
	.dc.b	$F5,$59,$3C,$8F,$F5,$36,$1F,$CE,$27,$31,$88,$FE,$73,$18,$23,$B3
	.dc.b	$FC,$DF,$B7,$F5,$6F,$A7,$FB,$39,$1A,$7D,$7B,$E9,$8F,$F8,$0B,$E6
	.dc.b	$3C,$7E,$71,$A4,$C7,$FC,$0A,$D7,$FC,$0A,$D7,$FC,$F3,$8F,$F7,$DE
	.dc.b	$2F,$C4,$F9,$FF,$91,$E2,$FF,$38,$B6,$BF,$C0,$F7,$F8,$9C,$3F,$3F
	.dc.b	$1F,$E0,$AB,$F1,$91,$FF,$0A,$B5,$FF,$0A,$B5,$F0,$D7,$A5,$FF,$11
	.dc.b	$D7,$88,$ED,$3F,$97,$FA,$67,$91,$F4,$9B,$5F,$E8,$A6,$BF,$E9,$76
	.dc.b	$BF,$E9,$36,$BF,$E9,$26,$BF,$EF,$C6,$F0,$1C,$E7,$FE,$5E,$FA,$7F
	.dc.b	$D6,$FA,$C6,$9E,$5C,$3C,$DC,$DE,$52,$63,$E5,$6B,$E5,$6B,$E5,$7F
	.dc.b	$9B,$5F,$F9,$B3,$CB,$EE,$49,$FF,$76,$63,$FB,$77,$FD,$41,$F4,$BF
	.dc.b	$E0,$C7,$FD,$2C,$D7,$FD,$2C,$D7,$FD,$3E,$D7,$FD,$06,$D7,$FD,$40
	.dc.b	$DF,$64,$D3,$5A,$3E,$9F,$F6,$BC,$7F,$EF,$96,$F1,$FF,$AF,$9B,$87
	.dc.b	$4F,$53,$F9,$6E,$2F,$6F,$FC,$BD,$3F,$EB,$96,$BF,$EA,$A6,$E0,$C3
	.dc.b	$FD,$33,$E0,$BE,$9F,$9E,$9F,$D6,$F1,$F3,$E3,$FD,$7D,$AF,$C4,$3E
	.dc.b	$6F,$CF,$7F,$10,$7F,$DD,$ED,$C7,$B4,$38,$13,$9C,$BF,$D5,$3F,$4E
	.dc.b	$F8,$2F,$E8,$FF,$2E,$FA,$7F,$F4,$56,$3F,$EE,$A8,$C7,$BB,$FD,$C5
	.dc.b	$B1,$99,$DB,$FE,$5D,$8C,$B6,$32,$FF,$BB,$A3,$2D,$8C,$A3,$E3,$7D
	.dc.b	$3F,$EA,$D6,$9E,$CC,$93,$98,$C3,$63,$95,$F8,$A7,$E1,$E8,$F8,$DF
	.dc.b	$F4,$0E,$1D,$E7,$31,$88,$EF,$7F,$EF,$88,$C4,$FC,$CE,$57,$C5,$3E
	.dc.b	$9F,$FA,$BD,$6A,$F8,$7A,$78,$3F,$AB,$56,$F1,$F0,$4F,$BF,$27,$C8
	.dc.b	$3D,$FF,$EC,$2D,$AD,$9F,$6D,$C3,$E6,$19,$BE,$3D,$F4,$FC,$AE,$57
	.dc.b	$2C,$3B,$C7,$F6,$9A,$DC,$F6,$B7,$CE,$7F,$9B,$E3,$0D,$7C,$11,$AF
	.dc.b	$83,$FF,$07,$C3,$E9,$BF,$07,$C3,$F1,$EF,$83,$6B,$FB,$C4,$63,$5F
	.dc.b	$C0,$5F,$1E,$5F,$F0,$0F,$CA,$E6,$3F,$34,$DD,$4E,$73,$F1,$4F,$B7
	.dc.b	$E3,$FF,$04,$FE,$85,$B4,$73,$2F,$3F,$FD,$05,$5C,$7C,$AF,$F0,$8D
	.dc.b	$70,$37,$3A,$9F,$A1,$6F,$99,$7C,$B7,$F8,$61,$B3,$F3,$DA,$FC,$8F
	.dc.b	$98,$7C,$86,$EA,$4E,$DF,$F0,$E4,$EA,$3F,$02,$3E,$91,$F4,$FF,$E2
	.dc.b	$1F,$9F,$FB,$33,$FE,$60,$D7,$F5,$5B,$27,$3F,$09,$FA,$BC,$38,$DF
	.dc.b	$46,$BF,$72,$77,$7C,$9B,$1F,$B5,$9D,$F2,$FE,$8E,$73,$87,$7E,$9F
	.dc.b	$F3,$AE,$1C,$5B,$87,$1C,$BF,$21,$F2,$38,$BF,$CC,$37,$A7,$FC,$4F
	.dc.b	$A7,$F6,$B9,$FC,$E7,$C1,$66,$A6,$1F,$E8,$0C,$BD,$F8,$70,$6D,$75
	.dc.b	$E3,$7D,$1F,$8B,$93,$F0,$5C,$F7,$9B,$F8,$CE,$BD,$EF,$79,$F5,$3F
	.dc.b	$5A,$E1,$F9,$C7,$E3,$6B,$91,$DF,$0F,$7C,$7D,$7C,$7D,$DF,$47,$FB
	.dc.b	$8C,$3D,$8D,$CF,$F4,$3C,$6A,$E2,$DB,$F1,$6E,$BE,$11,$FC,$01,$AF
	.dc.b	$73,$D0,$53,$CF,$1C,$7B,$7C,$67,$6B,$FA,$5D,$AF,$0E,$92,$FC,$31
	.dc.b	$FB,$67,$9B,$B7,$31,$B7,$27,$C0,$F0,$63,$EA,$C7,$F9,$76,$9F,$0B
	.dc.b	$E9,$F3,$3E,$69,$CE,$BF,$9A,$6C,$7B,$E7,$E2,$EC,$F2,$BD,$FB,$FD
	.dc.b	$E6,$C6,$6B,$EB,$FB,$EF,$A7,$FF,$4D,$E7,$FA,$3A,$D3,$6F,$89,$F8
	.dc.b	$1A,$E1,$F0,$7E,$08,$CE,$1F,$82,$3F,$69,$7C,$B4,$FE,$B5,$7C,$27
	.dc.b	$78,$75,$78,$65,$E7,$7C,$B9,$79,$2F,$F0,$78,$73,$9F,$D8,$35,$EF
	.dc.b	$8F,$96,$B8,$46,$A3,$BD,$72,$75,$97,$A6,$7C,$37,$CF,$0F,$16,$3F
	.dc.b	$E5,$9C,$BD,$15,$AC,$39,$BD,$4F,$3E,$1C,$EB,$D8,$4C,$7E,$0E,$5E
	.dc.b	$4A,$F3,$9C,$C9,$3D,$E3,$58,$F8,$AB,$D1,$5A,$79,$DF,$77,$37,$DE
	.dc.b	$58,$FA,$23,$11,$E0,$75,$19,$9E,$24,$9F,$64,$F8,$EB,$B4,$62,$75
	.dc.b	$5C,$63,$04,$3B,$47,$99,$F8,$E6,$0D,$C7,$17,$04,$BF,$80,$FA,$B0
	.dc.b	$F8,$79,$7D,$52,$65,$F6,$A7,$AD,$70,$7B,$5F,$58,$C1,$2B,$F1,$51
	.dc.b	$97,$9C,$FC,$D3,$C1,$8F,$C0,$7C,$91,$C3,$0F,$1E,$1E,$08,$C5,$7E
	.dc.b	$2E,$BF,$82,$61,$F9,$B6,$CF,$EC,$5B,$5F,$C2,$5F,$E9,$5F,$43,$FC
	.dc.b	$25,$E2,$D7,$F9,$1B,$5D,$EF,$C5,$FE,$C5,$1F,$C4,$74,$FE,$B8,$DD
	.dc.b	$3F,$63,$E7,$3D,$1F,$F3,$1D,$6B,$0F,$E0,$F7,$BC,$3A,$9F,$C1,$AF
	.dc.b	$87,$8B,$AA,$93,$FB,$8A,$E1,$BF,$F9,$B3,$1E,$B9,$7A,$2F,$79,$77
	.dc.b	$C3,$A5,$69,$EF,$19,$25,$7A,$27,$53,$EE,$AE,$73,$D6,$39,$E3,$87
	.dc.b	$9C,$66,$FF,$75,$97,$AA,$7D,$17,$B7,$A1,$C8,$C9,$27,$C6,$FE,$B2
	.dc.b	$79,$46,$23,$8B,$AC,$7E,$16,$1D,$75,$67,$F3,$6F,$79,$7B,$6B,$53
	.dc.b	$9A,$DE,$5F,$CB,$F2,$EF,$7E,$B2,$61,$EE,$3C,$99,$FC,$6B,$E3,$1E
	.dc.b	$99,$DE,$5E,$48,$C1,$C4,$F1,$9C,$B2,$F1,$E3,$F2,$E7,$86,$7E,$FC
	.dc.b	$3A,$BE,$FC,$BC,$B8,$73,$9C,$E3,$E0,$9C,$CF,$62,$4E,$9F,$01,$E4
	.dc.b	$33,$9F,$E3,$27,$8C,$EE,$73,$8F,$C5,$CF,$E0,$5F,$C9,$24,$67,$3F
	.dc.b	$5D,$7C,$0C,$FD,$79,$7C,$A8,$E6,$7E,$07,$3F,$BF,$CC,$F8,$E7,$C7
	.dc.b	$7E,$5E,$30,$D8,$7D,$3D,$89,$97,$F8,$5A,$78,$B6,$73,$87,$F8,$4B
	.dc.b	$E9,$FE,$4E,$06,$9F,$E3,$6A,$FF,$0F,$B8,$DB,$C4,$EB,$8A,$C7,$E9
	.dc.b	$BB,$5F,$4F,$E7,$E8,$7C,$76,$63,$0D,$C7,$FC,$24,$DC,$FF,$C2,$0D
	.dc.b	$DF,$FC,$6A,$DE,$0F,$F0,$93,$F0,$FC,$77,$D3,$FE,$EF,$C2,$FF,$46
	.dc.b	$7C,$DE,$63,$1A,$A3,$19,$FF,$92,$EB,$6F,$F8,$F2,$F9,$B9,$AE,$0F
	.dc.b	$F9,$03,$CE,$FF,$1A,$37,$9F,$EA,$47,$FB,$9E,$30,$4D,$7F,$C7,$F1
	.dc.b	$C1,$A7,$09,$FB,$9E,$4B,$E9,$E9,$F2,$FD,$06,$1F,$EC,$39,$D3,$FE
	.dc.b	$4C,$C5,$3A,$C3,$FD,$11,$7C,$6B,$83,$63,$31,$C6,$78,$79,$9F,$E8
	.dc.b	$9A,$F9,$5E,$7F,$F0,$8A,$FF,$54,$76,$E9,$FF,$4D,$FD,$33,$E9,$F1
	.dc.b	$93,$83,$FD,$65,$78,$6D,$F2,$3F,$DB,$97,$C5,$C5,$7F,$AE,$B4,$7D
	.dc.b	$3F,$16,$CF,$FB,$8F,$86,$D6,$EB,$44,$9D,$36,$3F,$D8,$BE,$9F,$FB
	.dc.b	$36,$7F,$DE,$0D,$8F,$F7,$4F,$9B,$FE,$C8,$9C,$BE,$CE,$EF,$F6,$A7
	.dc.b	$A5,$63,$55,$FE,$FF,$1E,$3F,$F3,$DE,$2F,$A5,$7C,$30,$FF,$7D,$5F
	.dc.b	$09,$CB,$63,$FD,$DF,$5C,$5F,$F6,$3C,$66,$FE,$67,$A8,$F8,$7F,$98
	.dc.b	$DA,$4B,$0F,$0F,$D3,$BD,$CD,$E1,$2F,$75,$F8,$7E,$26,$3C,$F1,$E5
	.dc.b	$5F,$ED,$8C,$79,$3C,$0E,$F5,$C3,$0F,$BD,$EC,$BE,$9E,$9F,$27,$D5
	.dc.b	$F6,$63,$F8,$59,$D6,$1C,$E3,$15,$C2,$31,$BA,$F8,$E2,$C3,$8E,$1E
	.dc.b	$63,$F4,$BF,$4E,$F5,$FF,$27,$8A,$30,$F4,$F6,$DF,$3B,$A9,$CD,$F3
	.dc.b	$CD,$5A,$BF,$F1,$34,$62,$FF,$5D,$5F,$40,$46,$67,$75,$C6,$7F,$D6
	.dc.b	$D3,$96,$C6,$5E,$51,$96,$C6,$5F,$9F,$F4,$EF,$67,$F5,$8D,$BF,$95
	.dc.b	$EA,$3E,$9D,$D9,$AB,$FD,$AC,$FF,$8C,$70,$F4,$9F,$88,$71,$7F,$E4
	.dc.b	$36,$C7,$FB,$35,$AF,$BD,$7D,$37,$E1,$23,$E9,$46,$39,$FB,$FC,$DB
	.dc.b	$E9,$FC,$37,$BE,$9F,$D2,$FC,$43,$E0,$BF,$5C,$DC,$FD,$7F,$4F,$FA
	.dc.b	$C7,$F0,$71,$9D,$3B,$CE,$5F,$33,$5F,$33,$5F,$73,$5F,$F6,$B3,$74
	.dc.b	$34,$9D,$DF,$1A,$7F,$8E,$7C,$8B,$E0,$7F,$8E,$37,$1F,$CB,$FD,$3F
	.dc.b	$F4,$6A,$FC,$A7,$5A,$B5,$1F,$8D,$9D,$3F,$80,$6B,$FB,$AE,$74,$62
	.dc.b	$3F,$1A,$D8,$CF,$9F,$E5,$6E,$1F,$96,$9D,$1C,$70,$F1,$75,$FA,$0F
	.dc.b	$E3,$90,$F8,$3C,$57,$A9,$E9,$19,$8E,$91,$82,$38,$26,$3F,$9E,$9E
	.dc.b	$0F,$E9,$4F,$E8,$9F,$27,$FE,$02,$9F,$E0,$18,$23,$13,$FB,$1F,$94
	.dc.b	$7D,$37,$F4,$99,$CF,$27,$F2,$C9,$CE,$1F,$B4,$6D,$FE,$B9,$AF,$EB
	.dc.b	$1A,$FC,$D6,$BF,$B8,$6E,$86,$9A,$C7,$CD,$F0,$BE,$33,$FA,$24,$3E
	.dc.b	$9F,$18,$7F,$18,$8C,$3E,$58,$FA,$D1,$8C,$3C,$D1,$82,$4E,$C9,$1F
	.dc.b	$26,$DF,$33,$C7,$2F,$E7,$0D,$AF,$C4,$7A,$7F,$C9,$A3,$17,$FC,$52
	.dc.b	$3F,$D2,$0F,$F0,$E6,$BC,$3D,$DF,$C5,$46,$23,$84,$60,$95,$FC,$8A
	.dc.b	$75,$87,$93,$6F,$EF,$D8,$7D,$2C,$7F,$B0,$FC,$C3,$E4,$B3,$5F,$DA
	.dc.b	$E3,$18,$FF,$2C,$6E,$7F,$C9,$9A,$FF,$24,$6B,$FD,$43,$E7,$FF,$90
	.dc.b	$3F,$AE,$6E,$1F,$E2,$1F,$A4,$FF,$11,$FD,$3F,$F8,$46,$73,$87,$1F
	.dc.b	$0F,$07,$F8,$62,$B7,$1F,$94,$9D,$E1,$FE,$1B,$C1,$F3,$1F,$E1,$B6
	.dc.b	$D7,$F5,$6F,$9A,$7D,$3E,$6A,$4F,$E7,$AB,$77,$FD,$36,$F7,$39,$AF
	.dc.b	$17,$8B,$FD,$1E,$FD,$4D,$9F,$EE,$FF,$4F,$FE,$83,$7F,$B4,$66,$9D
	.dc.b	$62,$8C,$79,$9D,$0F,$F1,$AE,$0C,$39,$57,$F6,$9A,$DE,$1F,$DC,$EB
	.dc.b	$5F,$15,$FC,$85,$B1,$FB,$BA,$FF,$07,$5F,$27,$FB,$14,$E9,$FE,$01
	.dc.b	$AB,$E9,$FF,$D2,$31,$FD,$DE,$31,$F1,$9F,$47,$E9,$FF,$83,$DF,$F9
	.dc.b	$66,$73,$8F,$F8,$15,$B7,$C7,$0E,$75,$F7,$EF,$8D,$FF,$88,$AF,$73
	.dc.b	$FC,$E7,$E2,$3D,$18,$7F,$77,$6B,$FE,$28,$6A,$77,$B3,$B5,$F4,$F8
	.dc.b	$6D,$7C,$39,$D4,$E4,$E3,$F1,$1F,$E3,$39,$D3,$5F,$E4,$93,$FE,$A4
	.dc.b	$C3,$67,$9B,$E9,$BF,$C7,$D3,$7D,$9F,$3C,$62,$3E,$93,$5F,$E5,$CD
	.dc.b	$7F,$C2,$8D,$7F,$C8,$AD,$4D,$FC,$0D,$EE,$FC,$46,$CF,$C8,$6B,$F8
	.dc.b	$16,$BF,$CF,$7A,$1F,$57,$F3,$1B,$1F,$41,$AF,$E2,$5A,$F1,$8E,$04
	.dc.b	$78,$3F,$63,$C5,$7D,$3F,$FA,$96,$7F,$C6,$6D,$7F,$C5,$AD,$7F,$C5
	.dc.b	$6D,$7F,$B6,$35,$FF,$1B,$B6,$3A,$FC,$47,$78,$D7,$4B,$E9,$BF,$C0
	.dc.b	$17,$FA,$18,$C7,$9F,$F8,$C6,$C6,$67,$6F,$F6,$18,$CB,$63,$2F,$F9
	.dc.b	$26,$32,$D8,$CB,$FE,$7C,$8F,$F5,$EF,$D5,$77,$F1,$CD,$24,$FF,$2B
	.dc.b	$DB,$E4,$78,$7F,$5A,$3F,$C1,$8D,$7F,$CA,$ED,$BF,$F1,$24,$7F,$09
	.dc.b	$F4,$7F,$C4,$0D,$EA,$F7,$46,$27,$64,$70,$70,$D9,$7F,$ED,$4B,$FE
	.dc.b	$7D,$97,$90,$8F,$CB,$FB,$07,$D8,$7C,$B6,$C6,$63,$04,$70,$43,$E7
	.dc.b	$47,$07,$FC,$05,$CB,$F8,$9F,$37,$F4,$58,$FF,$7D,$8C,$1F,$67,$F0
	.dc.b	$07,$C7,$D2,$5F,$DF,$8C,$B6,$32,$FD,$D8,$CB,$63,$F3,$1F,$45,$FD
	.dc.b	$12,$3F,$D1,$2D,$7F,$D0,$6D,$7F,$D0,$2D,$7F,$CE,$9F,$63,$57,$B2
	.dc.b	$8C,$37,$A7,$F8,$EB,$67,$FC,$F5,$3A,$24,$E9,$E9,$93,$EA,$1F,$53
	.dc.b	$FA,$F6,$DF,$FA,$81,$AF,$FA,$69,$AF,$FA,$0F,$E0,$5F,$77,$E2,$6C
	.dc.b	$7F,$A2,$DA,$FF,$A2,$DA,$F1,$8E,$04,$78,$3F,$C0,$7D,$87,$D2,$D2
	.dc.b	$4F,$F9,$7D,$F8,$3E,$FF,$F9,$22,$3F,$D3,$4D,$7F,$D3,$4D,$7F,$D3
	.dc.b	$4D,$7F,$D4,$4D,$7E,$7B,$75,$34,$C7,$D4,$7F,$4A,$83,$49,$3F,$53
	.dc.b	$D9,$FF,$11,$3F,$EA,$D6,$BF,$EA,$D6,$BF,$EA,$D6,$BF,$EB,$5F,$19
	.dc.b	$EA,$FD,$E6,$C7,$FA,$DD,$AF,$FA,$DD,$AF,$FA,$DF,$EA,$3F,$D1,$6F
	.dc.b	$FA,$F5,$AF,$FA,$F5,$AE,$F3,$FE,$CB,$F4,$2F,$A6,$A4,$9F,$ED,$71
	.dc.b	$96,$C6,$5F,$F0,$34,$65,$B1,$9F,$29,$E0,$FF,$A7,$FE,$65,$F4,$FF
	.dc.b	$DD,$7E,$77,$0D,$C0,$31,$E9,$87,$28,$FF,$78,$FC,$6F,$B3,$1F,$F7
	.dc.b	$4E,$7F,$EF,$56,$CF,$26,$BF,$EF,$AE,$8F,$F7,$FC,$EE,$3F,$D6,$55
	.dc.b	$CD,$B1,$97,$BB,$65,$F2,$0F,$85,$F4,$46,$1B,$19,$AF,$F0,$86,$9F
	.dc.b	$EB,$F8,$FE,$8F,$EA,$23,$15,$C0,$95,$C0,$5F,$83,$CC,$F8,$DC,$CF
	.dc.b	$D8,$BF,$91,$6B,$FE,$F7,$6D,$FF,$BF,$DA,$FF,$0C,$6C,$F1,$F1,$BD
	.dc.b	$CD,$9F,$90,$71,$8C,$CF,$03,$F9,$86,$7E,$99,$F2,$FC,$EF,$E5,$23
	.dc.b	$19,$23,$1A,$73,$AF,$64,$FF,$84,$BE,$87,$E6,$79,$A6,$92,$7C,$67
	.dc.b	$C2,$8D,$38,$9E,$4D,$8C,$BF,$7D,$BA,$7E,$32,$3E,$5F,$CD,$3E,$9F
	.dc.b	$EC,$B6,$3C,$AD,$7F,$C8,$D8,$FF,$BD,$E3,$FC,$8D,$1A,$9E,$24,$78
	.dc.b	$FB,$3F,$9B,$7F,$35,$C1,$F8,$B6,$9E,$C2,$9F,$9D,$3F,$43,$CD,$FE
	.dc.b	$0C,$C7,$FD,$D8,$D8,$F5,$35,$DC,$F5,$8C,$47,$4D,$BE,$9F,$D1,$3E
	.dc.b	$13,$F2,$ED,$9F,$C7,$1B,$D3,$F6,$DF,$87,$FE,$73,$AC,$D3,$1E,$3D
	.dc.b	$AF,$B7,$3A,$6C,$E9,$FB,$FF,$53,$FD,$A3,$2F,$C2,$B7,$0F,$DF,$F1
	.dc.b	$23,$15,$FA,$07,$EC,$3F,$48,$C7,$99,$FB,$08,$DC,$F1,$8C,$4F,$E9
	.dc.b	$BE,$C5,$EA,$7E,$79,$B1,$F8,$CE,$A4,$62,$3E,$94,$62,$78,$12,$38
	.dc.b	$4E,$F8,$BF,$19,$5B,$9C,$D7,$F0,$4F,$D4,$3D,$3A,$7E,$91,$BA,$7F
	.dc.b	$1E,$BE,$17,$ED,$CD,$C1,$F8,$E8,$C3,$70,$FB,$7C,$9F,$6B,$97,$ED
	.dc.b	$B2,$BF,$5B,$5B,$C7,$E6,$BF,$9B,$BF,$DA,$FC,$2F,$ED,$2B,$F8,$6B
	.dc.b	$FE,$39,$7C,$4D,$8D,$F8,$DF,$C7,$A3,$86,$47,$58,$7D,$A6,$DF,$4A
	.dc.b	$FD,$17,$37,$EF,$74,$FC,$7B,$67,$EE,$1F,$CF,$FA,$BF,$C4,$2F,$DF
	.dc.b	$FA,$97,$D5,$FD,$5D,$3F,$88,$4E,$F0,$E7,$7C,$27,$38,$75,$7F,$8B
	.dc.b	$61,$CB,$87,$C7,$5B,$D3,$F5,$6D,$C7,$F6,$2D,$AE,$11,$DB,$18,$FA
	.dc.b	$31,$ED,$7F,$D6,$2B,$73,$9B,$E8,$F9,$CC,$5F,$F3,$D9,$DE,$3E,$8A
	.dc.b	$DE,$1D,$2F,$C7,$7B,$CB,$E1,$63,$E8,$8F,$1E,$1C,$5D,$65,$EE,$73
	.dc.b	$3B,$BF,$99,$7C,$9E,$A7,$90,$97,$F2,$23,$51,$E9,$8E,$CE,$A7,$E0
	.dc.b	$4E,$B0,$F4,$4E,$E7,$C1,$5C,$F0,$F2,$4F,$13,$C3,$19,$7C,$E6,$27
	.dc.b	$89,$DA,$78,$CF,$88,$91,$99,$E2,$78,$4E,$43,$2F,$9F,$1D,$23,$26
	.dc.b	$E3,$2E,$2F,$D6,$7D,$38,$CC,$75,$24,$67,$0F,$85,$97,$DA,$26,$5F
	.dc.b	$52,$BC,$B5,$D2,$7F,$1D,$5C,$2B,$31,$8C,$BD,$F8,$F9,$F0,$F3,$5E
	.dc.b	$DF,$94,$62,$76,$FA,$6B,$93,$F9,$83,$FA,$5B,$FA,$E6,$DF,$F8,$FA
	.dc.b	$31,$1F,$D3,$BE,$93,$FD,$3E,$DE,$5F,$F2,$1C,$62,$3F,$AB,$B5,$FF
	.dc.b	$1C,$3F,$EA,$AD,$3F,$66,$DC,$BF,$4D,$F5,$6F,$2F,$FC,$79,$5A,$C3
	.dc.b	$F8,$2D,$EF,$0E,$B1,$FA,$7B,$E1,$CD,$A8,$93,$FB,$3A,$E1,$BA,$F7
	.dc.b	$39,$AD,$E1,$DE,$3F,$87,$4E,$A3,$04,$CB,$E0,$4E,$B2,$F3,$C6,$27
	.dc.b	$53,$E6,$35,$97,$69,$D1,$23,$F2,$57,$DE,$31,$1F,$23,$0E,$0E,$B0
	.dc.b	$EE,$F5,$C3,$BB,$B3,$B9,$ED,$23,$E4,$C7,$E1,$4F,$29,$E1,$3C,$B7
	.dc.b	$67,$F2,$32,$F2,$57,$7C,$BC,$B5,$AC,$BE,$96,$3C,$E3,$15,$E0,$AF
	.dc.b	$76,$5E,$0C,$FD,$73,$E0,$9C,$E3,$DA,$30,$77,$25,$71,$7A,$19,$36
	.dc.b	$E7,$3F,$56,$7F,$87,$9C,$D7,$5C,$7C,$B5,$AC,$FD,$39,$79,$27,$39
	.dc.b	$79,$33,$F6,$65,$CE,$73,$97,$93,$2E,$93,$92,$63,$D2,$7C,$E7,$B0
	.dc.b	$F1,$DF,$48,$F0,$4F,$CF,$8C,$4F,$0B,$EF,$9F,$B7,$1F,$C4,$71,$3E
	.dc.b	$E7,$FD,$E3,$A7,$F9,$AB,$F3,$37,$07,$E1,$AB,$5E,$8D,$9C,$D7,$F9
	.dc.b	$C2,$F8,$E7,$FE,$0F,$9C,$BF,$E1,$18,$CF,$07,$F9,$EA,$72,$FF,$82
	.dc.b	$3E,$2B,$B7,$31,$BF,$5B,$86,$63,$FE,$7E,$9D,$36,$73,$8F,$F8,$EF
	.dc.b	$47,$E7,$AF,$9E,$F5,$76,$BD,$6F,$D0,$D6,$A7,$31,$8A,$E1,$8F,$F8
	.dc.b	$ED,$B8,$7F,$8E,$9B,$1F,$E3,$66,$C7,$F8,$C5,$AF,$CA,$6D,$F7,$9E
	.dc.b	$97,$FE,$D1,$7F,$D1,$5C,$0F,$AE,$FD,$3B,$79,$7E,$94,$7F,$A6,$E7
	.dc.b	$FD,$3E,$E8,$91,$AE,$9D,$BF,$67,$AB,$D3,$C3,$FE,$AA,$6F,$2F,$F9
	.dc.b	$99,$7A,$FB,$CE,$65,$FE,$AC,$9D,$ED,$FE,$C8,$77,$19,$9F,$E0,$5B
	.dc.b	$BE,$BB,$F1,$1D,$EF,$8B,$CC,$61,$B7,$FE,$97,$AE,$0D,$9F,$89,$8F
	.dc.b	$FA,$B6,$72,$FF,$A6,$CF,$B5,$1A,$7F,$DF,$B1,$A9,$FF,$3E,$B6,$39
	.dc.b	$4E,$F8,$1E,$33,$E7,$3F,$3C,$DC,$3F,$D2,$26,$22,$3F,$E9,$47,$6E
	.dc.b	$2F,$FC,$DD,$E5,$BE,$BB,$B3,$67,$FD,$98,$D9,$FF,$5A,$60,$8C,$4F
	.dc.b	$E0,$9B,$3F,$EC,$5A,$E6,$49,$FF,$47,$7D,$AF,$FA,$A9,$FF,$70,$F9
	.dc.b	$DF,$CC,$1F,$F7,$F7,$3A,$75,$7F,$E8,$3A,$D3,$FA,$78,$CD,$7F,$A4
	.dc.b	$1B,$1F,$E8,$B6,$BF,$C1,$1B,$5D,$7C,$BF,$A4,$DA,$F2,$C6,$27,$8E
	.dc.b	$5F,$EF,$9F,$13,$F8,$67,$9E,$6E,$38,$66,$1F,$EF,$C9,$D3,$67,$27
	.dc.b	$FA,$77,$D3,$7D,$9F,$DD,$6C,$FF,$BE,$DB,$3F,$41,$B1,$FB,$06,$C7
	.dc.b	$FB,$21,$AB,$04,$61,$B5,$F9,$19,$FE,$19,$F0,$CB,$43,$D8,$C3,$CF
	.dc.b	$89,$F1,$9F,$A8,$BC,$D6,$B1,$FC,$AC,$E7,$0F,$F6,$BE,$1F,$9E,$BF
	.dc.b	$D1,$36,$32,$F8,$9B,$CA,$9D,$CE,$4E,$71,$99,$DC,$F4,$E0,$7D,$23
	.dc.b	$F1,$4C,$36,$BC,$73,$A9,$C9,$C7,$DE,$7C,$D5,$25,$70,$8C,$BC,$BE
	.dc.b	$03,$F9,$DE,$7F,$EF,$4C,$11,$88,$FF,$71,$B5,$FF,$6F,$35,$FE,$AF
	.dc.b	$E1,$7D,$0A,$F2,$B6,$3A,$35,$FF,$76,$B5,$FF,$74,$7D,$27,$F3,$E7
	.dc.b	$FD,$E2,$D7,$3D,$FF,$16,$31,$1F,$D9,$9B,$81,$A4,$9F,$AF,$C0,$FC
	.dc.b	$8D,$F7,$D8,$6E,$5F,$7A,$75,$39,$3E,$A7,$E2,$AE,$CA,$4A,$FA,$71
	.dc.b	$96,$C6,$73,$F9,$93,$96,$C7,$93,$E0,$5F,$4B,$49,$7E,$38,$CB,$F8
	.dc.b	$0F,$BC,$7C,$B7,$E1,$DB,$19,$8C,$11,$C1,$30,$FB,$D3,$C1,$FD,$39
	.dc.b	$FA,$8C,$DF,$1A,$FB,$4F,$92,$DB,$F8,$CD,$AE,$27,$ED,$DC,$3D,$38
	.dc.b	$5F,$28,$F8,$2F,$CF,$36,$FF,$36,$DA,$FC,$93,$5E,$4D,$7F,$7B,$5D
	.dc.b	$2B,$99,$2B,$ED,$7C,$97,$E7,$AB,$CB,$92,$31,$1F,$E0,$06,$BF,$05
	.dc.b	$AF,$A3,$27,$BF,$FC,$03,$2F,$88,$D8,$CE,$1F,$C1,$A7,$2D,$8C,$D7
	.dc.b	$C6,$9C,$B6,$32,$FD,$28,$CB,$FE,$14,$EC,$FE,$52,$DF,$07,$F3,$71
	.dc.b	$88,$FB,$0D,$7E,$7F,$A5,$F7,$5E,$8D,$7F,$92,$D7,$F8,$7F,$83,$D9
	.dc.b	$39,$9E,$31,$82,$63,$ED,$9D,$69,$FC,$77,$A7,$C6,$49,$FD,$56,$3F
	.dc.b	$EB,$6C,$3D,$9F,$77,$D7,$23,$49,$97,$CD,$AF,$E4,$95,$F1,$DA,$FC
	.dc.b	$56,$BE,$A6,$BF,$8E,$6D,$7F,$31,$7A,$12,$FF,$7D,$F4,$4F,$89,$FE
	.dc.b	$0A,$D8,$CC,$60,$8E,$09,$7F,$8C,$9E,$0F,$E8,$BC,$87,$CB,$3D,$4F
	.dc.b	$E8,$F1,$86,$C6,$6F,$FA,$6C,$E5,$B1,$9A,$E8,$D7,$F6,$F3,$B7,$35
	.dc.b	$FD,$3B,$F3,$B7,$D7,$52,$61,$FD,$5A,$32,$FF,$5C,$F2,$5F,$44,$F7
	.dc.b	$7F,$AE,$C6,$1B,$19,$9E,$0D,$C7,$F8,$0D,$6E,$73,$5C,$08,$FF,$66
	.dc.b	$9F,$EE,$BE,$CD,$4D,$23,$FD,$C3,$C6,$7C,$3B,$E7,$FF,$B7,$C6,$1B
	.dc.b	$19,$9E,$0D,$7F,$8A,$CF,$F9,$8F,$1E,$3E,$C7,$F8,$16,$7B,$35,$DB
	.dc.b	$5F,$E4,$BF,$3E,$F9,$AF,$53,$63,$FC,$02,$D7,$FC,$0E,$D7,$F9,$2B
	.dc.b	$5E,$71,$C8,$8F,$F8,$2B,$E7,$FF,$A7,$37,$C7,$F9,$11,$88,$FF,$09
	.dc.b	$B5,$FE,$63,$F3,$5F,$D1,$1F,$F0,$9B,$5F,$F0,$CB,$5D,$F5,$7D,$38
	.dc.b	$C4,$7F,$42,$6B,$FE,$2A,$6A,$73,$DA,$F9,$FF,$F1,$5C,$FF,$69,$6B
	.dc.b	$FC,$C1,$AF,$FA,$B9,$AF,$F4,$F6,$BF,$E3,$F6,$C7,$5F,$53,$FC,$5A
	.dc.b	$D9,$FE,$5D,$F3,$EF,$93,$FF,$18,$F7,$A3,$13,$A9,$CC,$62,$3F,$99
	.dc.b	$35,$FE,$60,$D7,$F9,$2B,$5F,$EB,$2D,$9F,$F1,$E7,$CF,$F8,$3C,$03
	.dc.b	$49,$5F,$E0,$BE,$EF,$AC,$FF,$6D,$6B,$FE,$5D,$6E,$3F,$C2,$37,$F5
	.dc.b	$FC,$17,$F9,$23,$13,$49,$D3,$A2,$61,$FD,$46,$32,$F8,$72,$F9,$38
	.dc.b	$FF,$94,$E3,$04,$3F,$AA,$7C,$FF,$F2,$37,$FA,$D3,$5C,$F8,$5F,$3A
	.dc.b	$31,$1F,$D5,$DA,$FF,$4C,$6B,$FE,$00,$6E,$27,$39,$FF,$73,$E4,$FB
	.dc.b	$BE,$4F,$E9,$6F,$FD,$00,$D9,$D4,$7F,$6D,$F9,$FF,$F4,$C3,$7C,$3F
	.dc.b	$C0,$C6,$23,$FB,$73,$5F,$ED,$8D,$7F,$B0,$B5,$FF,$0A,$B7,$13,$9A
	.dc.b	$FF,$11,$F2,$FE,$4B,$CD,$FD,$15,$FF,$88,$A3,$07,$FA,$0B,$CD,$7C
	.dc.b	$FE,$A7,$FC,$D8,$D7,$FC,$48,$D7,$FC,$36,$D5,$1C,$A7,$A4,$E6,$31
	.dc.b	$3F,$E8,$28,$E1,$6F,$9F,$FF,$17,$4F,$F9,$6D,$AF,$F8,$89,$AF,$F9
	.dc.b	$02,$38,$79,$FE,$9F,$71,$F1,$F4,$95,$FE,$90,$8C,$B6,$32,$FF,$A2
	.dc.b	$63,$2D,$8F,$DC,$FA,$6F,$88,$A4,$9F,$F2,$F3,$F2,$7E,$7F,$FD,$33
	.dc.b	$1F,$E9,$E6,$BF,$E9,$56,$BF,$E4,$D6,$BF,$E4,$46,$BF,$29,$BA,$1A
	.dc.b	$63,$E7,$FF,$AA,$41,$A4,$9F,$B9,$E8,$BE,$96,$92,$3F,$D3,$EF,$E0
	.dc.b	$7C,$8F,$F2,$3C,$7F,$AB,$7D,$6C,$11,$FF,$59,$67,$F8,$7F,$78,$D2
	.dc.b	$47,$FB,$33,$E5,$1F,$41,$FD,$A5,$B1,$98,$C1,$1C,$10,$FF,$6C,$C7
	.dc.b	$06,$C7,$FA,$03,$CC,$FF,$70,$7C,$23,$E8,$3F,$D0,$7E,$8F,$F8,$66
	.dc.b	$FF,$DD,$11,$8F,$5B,$FD,$0F,$8F,$3C,$3F,$DD,$1F,$41,$EA,$BF,$F5
	.dc.b	$F7,$42,$31,$1F,$EF,$23,$FD,$F8,$E3,$C7,$FC,$93,$63,$FD,$24,$D7
	.dc.b	$7A,$FF,$A5,$A7,$6D,$8D,$B5,$53,$E1,$DF,$41,$E9,$9C,$C6,$1F,$F0
	.dc.b	$9E,$7F,$EA,$18,$C4,$7F,$85,$23,$1E,$6F,$1B,$E9,$76,$F7,$B7,$C7
	.dc.b	$9C,$E1,$D0,$EF,$7E,$1E,$77,$D0,$7F,$2D,$D3,$FD,$4C,$DA,$FF,$4E
	.dc.b	$B5,$FE,$06,$D9,$E3,$E4,$7B,$DB,$3F,$EE,$D6,$BF,$3D,$B5,$DA,$7D
	.dc.b	$5F,$15,$F9,$6A,$F7,$1F,$21,$C6,$5F,$81,$BE,$6F,$B7,$CD,$7C,$97
	.dc.b	$CF,$D3,$FD,$C8,$D8,$CE,$1F,$82,$6B,$F8,$19,$DB,$99,$E0,$47,$F1
	.dc.b	$3F,$1E,$F9,$DF,$A0,$D8,$F9,$94,$8C,$47,$AA,$31,$EC,$F0,$A9,$3C
	.dc.b	$2B,$78,$FD,$77,$FC,$9B,$1F,$5C,$73,$7F,$81,$F6,$FF,$6E,$B6,$3F
	.dc.b	$43,$C1,$F8,$F7,$6D,$8D,$40,$C9,$67,$9C,$EC,$91,$EC,$F9,$37,$D0
	.dc.b	$7E,$3D,$B1,$E5,$BF,$F0,$D6,$3F,$8F,$EA,$7C,$8F,$E5,$5B,$5F,$19
	.dc.b	$AF,$C5,$7F,$A1,$CF,$19,$E0,$49,$E1,$E3,$FE,$7F,$E8,$3F,$B6,$57
	.dc.b	$E7,$5B,$D3,$ED,$E1,$46,$2F,$F6,$2F,$C1,$7E,$BB,$98,$DF,$9F,$D4
	.dc.b	$ED,$3F,$A5,$FA,$0F,$F0,$FB,$F9,$B6,$BF,$9B,$6F,$07,$5E,$64,$62
	.dc.b	$7E,$84,$63,$D6,$FD,$34,$E6,$7F,$B6,$B6,$75,$1D,$E7,$F7,$DF,$31
	.dc.b	$E9,$8F,$C6,$37,$4F,$D2,$5F,$91,$FE,$01,$C3,$F8,$AA,$FC,$93,$63
	.dc.b	$3C,$DF,$6F,$9B,$EE,$32,$BF,$00,$E5,$FC,$EF,$C0,$FE,$CB,$83,$E7
	.dc.b	$3F,$E3,$98,$FC,$F3,$6F,$F8,$83,$B8,$FE,$42,$DD,$FF,$D7,$9E,$C7
	.dc.b	$F1,$98,$E1,$7D,$3C,$6F,$E4,$97,$FA,$0E,$AF,$F0,$8B,$FC,$10,$C7
	.dc.b	$07,$E4,$63,$39,$7F,$45,$F9,$D7,$D0,$7C,$3C,$7F,$A2,$D6,$9B,$7C
	.dc.b	$4F,$E4,$95,$C3,$8B,$C3,$D3,$FC,$32,$75,$97,$86,$30,$FF,$16,$AE
	.dc.b	$0E,$B2,$FE,$AC,$F9,$8E,$18,$F8,$54,$F0,$73,$7C,$71,$F2,$C7,$5C
	.dc.b	$BD,$4F,$EE,$71,$F0,$E3,$D9,$B5,$C2,$3E,$15,$70,$AF,$14,$6F,$1F
	.dc.b	$14,$62,$BF,$82,$63,$E3,$AD,$63,$E3,$AF,$05,$FA,$6B,$57,$DC,$93
	.dc.b	$AA,$F9,$78,$78,$71,$F2,$CE,$4F,$79,$ED,$CB,$E7,$E3,$E4,$9D,$CF
	.dc.b	$94,$F8,$F5,$CE,$74,$4B,$EF,$7C,$9F,$45,$71,$C7,$CE,$F8,$72,$F6
	.dc.b	$BE,$27,$E3,$D7,$18,$C6,$5E,$CA,$E3,$1E,$32,$65,$EC,$8C,$BE,$82
	.dc.b	$63,$F1,$F2,$F4,$65,$E4,$9E,$11,$8B,$EE,$79,$70,$FA,$71,$9C,$BC
	.dc.b	$E4,$8C,$E5,$E7,$8C,$57,$13,$E3,$C6,$67,$EA,$C7,$C3,$AE,$91,$8B
	.dc.b	$F1,$D6,$EB,$91,$B7,$E2,$3F,$D4,$31,$FD,$9B,$71,$FF,$22,$46,$23
	.dc.b	$FC,$67,$F3,$FF,$E9,$BB,$34,$9C,$DF,$E4,$B7,$FD,$53,$9F,$E6,$DB
	.dc.b	$97,$E5,$3E,$81,$E6,$BA,$3F,$A8,$56,$B2,$FD,$23,$5F,$EC,$EE,$77
	.dc.b	$5F,$2A,$E3,$94,$BF,$DE,$5E,$EB,$51,$89,$D6,$5E,$57,$F8,$3C,$EA
	.dc.b	$31,$3B,$23,$FC,$4E,$35,$97,$C4,$AE,$6D,$AF,$E1,$53,$A7,$D7,$3A
	.dc.b	$69,$AC,$BE,$13,$DA,$B9,$5F,$88,$F0,$4E,$9C,$46,$89,$3F,$C4,$AF
	.dc.b	$C0,$FC,$8B,$F0,$34,$F0,$63,$8C,$FE,$DC,$F9,$73,$F9,$B9,$77,$C7
	.dc.b	$AE,$1C,$72,$F0,$BD,$E7,$4F,$B2,$32,$EE,$7F,$01,$3A,$8C,$1B,$8F
	.dc.b	$4D,$78,$31,$F2,$67,$F8,$1C,$BC,$B7,$E8,$3E,$31,$C4,$95,$C5,$C9
	.dc.b	$D8,$EE,$E7,$1F,$2C,$F4,$7D,$71,$9C,$3D,$F3,$97,$CB,$9F,$CC,$8C
	.dc.b	$5F,$84,$EB,$19,$C3,$D0,$4C,$FF,$1B,$9F,$C1,$BE,$37,$E5,$79,$E3
	.dc.b	$E1,$8E,$91,$A3,$38,$75,$BF,$8A,$74,$CF,$F1,$9C,$6F,$A0,$FF,$76
	.dc.b	$C7,$F9,$6B,$E8,$3F,$DF,$EF,$F9,$72,$33,$8F,$F9,$6E,$72,$FF,$97
	.dc.b	$63,$3B,$7F,$A3,$E7,$2F,$FA,$46,$33,$B7,$F8,$D6,$72,$FF,$82,$BE
	.dc.b	$83,$F0,$78,$B9,$51,$85,$F4,$16,$26,$7F,$E1,$47,$51,$8C,$FF,$D1
	.dc.b	$3C,$6F,$A0,$7C,$C7,$6E,$E7,$B1,$F9,$2A,$D4,$E6,$31,$5C,$32,$FF
	.dc.b	$4D,$37,$0F,$F4,$D3,$63,$FC,$80,$D8,$FF,$1F,$B5,$5F,$69,$E7,$7F
	.dc.b	$BA,$78,$B6,$7F,$CB,$5F,$23,$FC,$0F,$53,$4F,$F4,$FD,$7F,$A4,$1C
	.dc.b	$73,$7C,$A7,$5D,$16,$3E,$6F,$4F,$A7,$93,$FD,$88,$DE,$2F,$F3,$42
	.dc.b	$F6,$37,$A4,$CB,$FD,$3B,$3B,$D3,$FD,$B6,$EE,$33,$3F,$B7,$DD,$F4
	.dc.b	$1F,$73,$C1,$7C,$66,$63,$0D,$BF,$F4,$96,$3F,$E7,$5B,$E3,$18,$8F
	.dc.b	$F4,$C4,$61,$FF,$7C,$1F,$6A,$BE,$43,$A6,$B9,$9E,$4D,$38,$CF,$38
	.dc.b	$FF,$42,$70,$3C,$77,$D0,$7E,$51,$B3,$FE,$99,$31,$11,$FF,$4E,$3B
	.dc.b	$71,$7F,$EC,$1F,$2D,$F4,$1E,$6C,$3F,$D8,$CD,$8C,$D7,$FB,$3A,$30
	.dc.b	$47,$04,$AF,$F7,$57,$D0,$7F,$AB,$A3,$FD,$A3,$E7,$7F,$6E,$7F,$DF
	.dc.b	$DC,$E9,$D5,$7F,$B4,$EB,$4F,$E9,$A3,$35,$FE,$99,$6C,$7F,$A6,$1A
	.dc.b	$FE,$69,$B5,$D7,$CC,$FA,$2D,$AF,$F8,$01,$3A,$38,$E4,$F3,$DE,$0F
	.dc.b	$F0,$06,$D7,$FB,$A9,$FC,$1F,$3B,$9F,$FD,$79,$1A,$CF,$FD,$D8,$D9
	.dc.b	$D4,$30,$ED,$87,$58,$EC,$FF,$0E,$F4,$9F,$2D,$F4,$DB,$97,$E3,$ED
	.dc.b	$18,$BF,$ED,$E7,$57,$13,$F8,$86,$C7,$FB,$45,$AE,$EB,$B1,$F9,$4E
	.dc.b	$37,$95,$FD,$EB,$D7,$FE,$F7,$8F,$E4,$7E,$33,$F5,$53,$9C,$BF,$DB
	.dc.b	$9C,$C9,$CC,$7F,$B0,$F2,$FD,$2D,$7E,$91,$B1,$97,$F8,$FC,$65,$B1
	.dc.b	$9C,$7C,$6D,$9D,$46,$27,$FD,$A1,$C2,$FA,$07,$D0,$77,$8C,$36,$33
	.dc.b	$87,$96,$72,$DF,$75,$F4,$1C,$5B,$19,$8C,$1D,$7D,$FF,$E8,$71,$CD
	.dc.b	$B8,$F9,$1A,$FF,$BC,$5A,$FF,$BC,$3C,$2F,$E0,$2F,$F3,$26,$E1,$FE
	.dc.b	$4C,$8F,$54,$63,$E8,$3F,$9F,$47,$FB,$ED,$AF,$FB,$E9,$AF,$F3,$C6
	.dc.b	$C9,$A4,$F0,$7E,$33,$C3,$85,$F4,$0F,$A0,$FB,$11,$86,$C6,$73,$FB
	.dc.b	$D3,$97,$E6,$7D,$03,$BB,$E9,$36,$33,$18,$23,$82,$67,$F7,$27,$82
	.dc.b	$3F,$91,$FC,$03,$E8,$3C,$AD,$AF,$1B,$67,$81,$C7,$04,$66,$33,$AB
	.dc.b	$E8,$1F,$2F,$F8,$96,$E1,$F8,$76,$D7,$E4,$1A,$F2,$6B,$E7,$AE,$95
	.dc.b	$CC,$95,$F2,$3E,$39,$F4,$1F,$39,$B1,$98,$C1,$F9,$BF,$91,$7C,$0D
	.dc.b	$24,$FE,$6E,$32,$D8,$CE,$3F,$A6,$9C,$B6,$3E,$EF,$C9,$7E,$7A,$3E
	.dc.b	$39,$FC,$5D,$C3,$FD,$49,$AF,$CB,$6A,$9F,$5F,$BD,$FB,$F6,$DF,$F0
	.dc.b	$36,$CF,$66,$BF,$59,$AF,$DF,$6B,$D7,$05,$75,$6D,$7E,$4B,$B7,$F9
	.dc.b	$25,$7F,$68,$6B,$F8,$16,$BF,$CC,$DB,$7F,$E1,$AC,$FD,$1E,$97,$DA
	.dc.b	$AF,$63,$67,$57,$EB,$8C,$12,$B8,$9F,$A8,$BF,$F5,$BD,$F8,$FA,$30
	.dc.b	$4B,$F8,$9F,$41,$D4,$97,$F2,$5F,$05,$FF,$1F,$BF,$8C,$D7,$E3,$35
	.dc.b	$FA,$4D,$C7,$F9,$DB,$F4,$CC,$10,$FD,$EF,$D0,$3E,$1E,$93,$1F,$DE
	.dc.b	$C6,$5B,$19,$BF,$C3,$F2,$2B,$51,$86,$BF,$B2,$F1,$DF,$2F,$FD,$A7
	.dc.b	$0F,$E8,$D1,$8F,$2F,$FA,$33,$63,$33,$BA,$FC,$AC,$E5,$B1,$97,$B3
	.dc.b	$5F,$E1,$11,$B7,$F3,$BF,$3E,$FA,$0F,$EA,$0D,$BF,$EA,$8D,$9F,$EB
	.dc.b	$4D,$7F,$85,$F9,$2F,$A0,$FF,$42,$57,$F5,$F8,$C7,$BD,$FD,$7D,$B1
	.dc.b	$99,$DD,$7E,$BA,$72,$D8,$CB,$D9,$AF,$F2,$18,$DB,$FD,$A7,$CE,$FE
	.dc.b	$D4,$FF,$25,$6A,$69,$FD,$BF,$C5,$7C,$33,$E8,$30,$DA,$FE,$F1,$3A
	.dc.b	$9C,$9C,$7D,$86,$14,$95,$C2,$32,$FB,$7D,$37,$B5,$F2,$34,$F5,$46
	.dc.b	$27,$D9,$18,$9D,$47,$F3,$FF,$A0,$7C,$DE,$31,$FF,$24,$55,$C3,$95
	.dc.b	$F9,$30,$E4,$F4,$39,$5F,$45,$F3,$5F,$CF,$9F,$F0,$93,$5F,$F0,$C3
	.dc.b	$5F,$E4,$6D,$7F,$8A,$7E,$95,$FE,$06,$7B,$75,$A3,$11,$E3,$3F,$92
	.dc.b	$38,$CB,$FC,$6F,$5D,$D6,$1D,$3C,$FF,$83,$87,$36,$DF,$19,$FE,$B7
	.dc.b	$7C,$5F,$EC,$9F,$51,$F5,$3D,$DF,$92,$DC,$BE,$13,$5F,$27,$63,$E8
	.dc.b	$7F,$C7,$11,$AD,$3F,$D1,$AD,$BE,$71,$88,$E3,$3C,$1F,$75,$BE,$2D
	.dc.b	$F3,$7E,$26,$CF,$89,$AF,$C5,$D9,$F4,$2F,$94,$E1,$72,$B8,$3F,$12
	.dc.b	$75,$1F,$77,$4F,$E1,$7E,$97,$F9,$2D,$B1,$FE,$14,$F0,$3F,$C2,$4D
	.dc.b	$9F,$F2,$86,$6F,$A8,$7D,$E7,$C8,$6D,$7C,$86,$BF,$22,$F9,$B7,$C6
	.dc.b	$F0,$47,$F9,$A2,$3C,$11,$82,$38,$21,$FC,$A2,$38,$3F,$DC,$2F,$FD
	.dc.b	$23,$F5,$9F,$6B,$C4,$34,$FD,$2E,$7F,$CA,$B0,$BE,$77,$E9,$B0,$FF
	.dc.b	$19,$B5,$FE,$6A,$D7,$FB,$D7,$4F,$C7,$8C,$47,$F7,$B8,$C1,$1F,$F3
	.dc.b	$D6,$2F,$AC,$FF,$4D,$C7,$E0,$1B,$CB,$F9,$A8,$4E,$FC,$7F,$2D,$FA
	.dc.b	$3D,$1F,$D8,$D7,$F8,$87,$AB,$EC,$46,$23,$FA,$5B,$5F,$F7,$0B,$5F
	.dc.b	$F1,$D3,$5F,$F5,$63,$74,$39,$C1,$F4,$5F,$E2,$8A,$FD,$1B,$5F,$CC
	.dc.b	$B7,$2F,$F4,$FD,$78,$DF,$F5,$0F,$D1,$6E,$BF,$DE,$6D,$7F,$C9,$8D
	.dc.b	$7F,$D7,$CD,$C0,$D2,$7A,$7F,$97,$78,$63,$FE,$FC,$FA,$BF,$DC,$F6
	.dc.b	$9A,$7E,$02,$7C,$DD,$DF,$E5,$BF,$5D,$F1,$9F,$A1,$6E,$7F,$E6,$28
	.dc.b	$C1,$1C,$10,$FE,$F3,$1C,$1F,$C9,$7E,$16,$FB,$BF,$E1,$6D,$8F,$E1
	.dc.b	$6D,$7F,$85,$CF,$1D,$BF,$05,$EC,$BE,$93,$FA,$AB,$6B,$FD,$80,$D7
	.dc.b	$FD,$65,$1A,$23,$A6,$BF,$EB,$5F,$81,$FE,$8B,$2F,$82,$46,$1B,$3F
	.dc.b	$D0,$67,$B3,$FC,$CD,$79,$DF,$E4,$A8,$D3,$5D,$37,$0F,$5F,$B2,$FA
	.dc.b	$4F,$D8,$FB,$3F,$74,$9C,$BF,$C4,$5F,$B7,$F3,$1F,$C7,$AF,$F4,$9C
	.dc.b	$C8,$C7,$1F,$F3,$86,$E6,$D3,$D9,$C1,$FD,$5F,$E6,$5F,$49,$FC,$03
	.dc.b	$E2,$DF,$49,$FE,$9C,$F6,$1E,$0F,$FB,$CA,$30,$FF,$A8,$2F,$8E,$5F
	.dc.b	$D8,$C3,$87,$FA,$1D,$F4,$7F,$C1,$53,$CD,$A7,$FA,$C6,$78,$6D,$FB
	.dc.b	$FE,$D5,$F1,$AF,$F0,$85,$F1,$68,$9D,$E1,$E6,$8E,$99,$7F,$1C,$FA
	.dc.b	$4F,$F5,$AC,$78,$38,$1F,$66,$FA,$0E,$11,$87,$96,$2B,$E3,$7E,$08
	.dc.b	$C6,$FC,$30,$E5,$3D,$5A,$EE,$B8,$EF,$F3,$23,$D1,$F2,$2F,$A5,$EE
	.dc.b	$D8,$F8,$B8,$A3,$11,$E1,$7F,$AB,$B6,$78,$91,$E3,$1A,$AF,$07,$13
	.dc.b	$E9,$7F,$CE,$95,$DB,$E5,$3F,$2D,$5E,$5C,$51,$8E,$1F,$F6,$85,$F0
	.dc.b	$BF,$A3,$87,$3F,$85,$F8,$F7,$B9,$FA,$8D,$9D,$47,$7F,$89,$D9,$1F
	.dc.b	$C1,$C6,$9F,$9D,$E9,$BE,$6B,$EA,$FC,$9F,$F5,$89,$FB,$4D,$77,$1F
	.dc.b	$62,$78,$46,$BD,$1E,$26,$67,$EE,$FC,$C7,$D6,$DF,$E0,$37,$6F,$F7
	.dc.b	$E3,$6F,$F1,$CF,$F9,$79,$FA,$AE,$77,$34,$9F,$09,$F9,$39,$FD,$2F
	.dc.b	$B5,$F8,$CF,$9A,$7D,$2D,$35,$DF,$F5,$F2,$FD,$75,$66,$FF,$07,$7C
	.dc.b	$1F,$D4,$9D,$0B,$5C,$5F,$D2,$46,$DF,$CC,$37,$3F,$ED,$F3,$BC,$FF
	.dc.b	$5F,$F3,$3F,$A0,$EF,$34,$C7,$89,$71,$FC,$AC,$E9,$B3,$A7,$F3,$3F
	.dc.b	$4A,$F6,$7F,$56,$D8,$CC,$61,$CB,$1C,$9C,$8F,$DE,$72,$BE,$97,$FC
	.dc.b	$EB,$A7,$E5,$FE,$7D,$5F,$B8,$6B,$F8,$C9,$DF,$06,$AF,$97,$C3,$EE
	.dc.b	$A1,$B7,$BD,$70,$8F,$E1,$D7,$D1,$FE,$04,$DC,$3F,$5B,$3B,$D9,$C1
	.dc.b	$F8,$7B,$FA,$3C,$7F,$81,$3F,$05,$E1,$58,$FB,$B1,$5F,$AF,$1F,$C4
	.dc.b	$BE,$33,$F5,$96,$7F,$C7,$B8,$FE,$DD,$BA,$FC,$29,$D7,$5F,$13,$97
	.dc.b	$BB,$31,$FE,$1B,$7F,$B7,$7F,$9C,$F4,$FF,$47,$7F,$96,$98,$E0,$FC
	.dc.b	$6C,$67,$2F,$E3,$5F,$08,$FA,$5F,$8D,$7B,$38,$F4,$7E,$E7,$0E,$27
	.dc.b	$F4,$0A,$E1,$C7,$FA,$0C,$7A,$6D,$FD,$35,$B8,$7F,$45,$6D,$70,$8E
	.dc.b	$67,$68,$DB,$E1,$CB,$CB,$97,$93,$2F,$14,$EF,$2E,$35,$AC,$BC,$51
	.dc.b	$8C,$BC,$4F,$9F,$0E,$58,$F6,$C7,$9E,$1F,$AA,$8F,$2C,$EA,$B9,$63
	.dc.b	$E9,$79,$63,$DA,$3C,$44,$BF,$13,$C0,$F5,$63,$E2,$7F,$AF,$E5,$F0
	.dc.b	$5E,$11,$99,$F1,$5F,$69,$E5,$97,$9A,$B4,$7E,$9A,$F9,$5F,$0A,$E6
	.dc.b	$79,$CE,$F8,$FA,$DC,$9E,$43,$FD,$68,$E6,$39,$12,$BB,$46,$5F,$0C
	.dc.b	$F5,$9F,$7B,$D2,$FC,$CE,$67,$8E,$3F,$1E,$BA,$4E,$4F,$0C,$66,$7A
	.dc.b	$C7,$97,$0F,$6C,$62,$FA,$CE,$70,$F2,$3C,$CC,$5F,$37,$13,$E4,$BE
	.dc.b	$71,$8C,$BF,$04,$7E,$02,$B8,$47,$C5,$39,$9F,$96,$3F,$84,$57,$E4
	.dc.b	$9B,$3F,$AF,$6E,$5F,$C3,$5F,$F1,$27,$D2,$FF,$27,$78,$B5,$FF,$19
	.dc.b	$35,$DF,$2F,$17,$FC,$67,$1F,$CB,$F4,$FE,$7E,$DD,$3F,$05,$E6,$3D
	.dc.b	$0F,$F5,$75,$6B,$0F,$E9,$77,$BC,$3A,$9F,$92,$BE,$1E,$76,$AA,$4F
	.dc.b	$F0,$0A,$E1,$B7,$F9,$57,$1E,$B9,$7A,$2F,$79,$77,$C3,$A5,$69,$EF
	.dc.b	$19,$25,$7A,$27,$53,$EE,$AE,$73,$D6,$39,$E3,$87,$9C,$66,$FF,$8C
	.dc.b	$E5,$EA,$9F,$45,$ED,$E8,$72,$32,$49,$F1,$AC,$7D,$D8,$71,$D5,$9F
	.dc.b	$C9,$AE,$77,$E0,$CF,$D3,$5A,$C7,$85,$FC,$8C,$BC,$97,$DB,$0E,$39
	.dc.b	$FA,$4F,$85,$97,$0C,$3D,$D9,$F8,$AB,$CD,$5B,$7E,$09,$E7,$C3,$AC
	.dc.b	$66,$3A,$9A,$9E,$58,$78,$72,$F3,$4F,$43,$96,$7E,$AC,$FD,$D3,$AA
	.dc.b	$F8,$B7,$CB,$0E,$D3,$97,$C9,$19,$C3,$CA,$78,$8F,$1B,$97,$13,$C5
	.dc.b	$F3,$63,$EA,$78,$C6,$5E,$31,$97,$8E,$1E,$63,$E7,$63,$E9,$CB,$E5
	.dc.b	$4F,$DF,$CB,$E0,$56,$E7,$FA,$C5,$F3,$C3,$C5,$C6,$FA,$57,$C3,$7E
	.dc.b	$66,$1F,$4B,$FE,$B1,$DF,$FC,$F7,$5A,$7F,$CF,$B1,$9D,$3F,$D1,$93
	.dc.b	$97,$FD,$1D,$19,$D3,$FC,$E9,$39,$7F,$CE,$D1,$9D,$FF,$C0,$1C,$8F
	.dc.b	$A5,$F1,$37,$1F,$EF,$FF,$4B,$89,$91,$FE,$02,$77,$18,$CB,$FC,$FB
	.dc.b	$F4,$BF,$EE,$2A,$FF,$49,$F7,$3C,$8F,$DB,$73,$2B,$57,$FE,$24,$9C
	.dc.b	$C6,$35,$FF,$1B,$46,$3D,$9F,$8D,$87,$3C,$7F,$D0,$37,$BD,$7F,$C3
	.dc.b	$53,$9C,$7C,$77,$C3,$0F,$9D,$F3,$1F,$81,$E1,$34,$FF,$44,$65,$FE
	.dc.b	$AA,$71,$CF,$C0,$E3,$D3,$67,$FD,$5F,$D5,$E1,$CB,$FD,$98,$DC,$0B
	.dc.b	$DE,$FB,$CE,$49,$9E,$23,$2F,$F1,$6C,$DF,$4B,$F8,$5F,$09,$F2,$79
	.dc.b	$8C,$37,$3F,$F3,$FB,$73,$FF,$3C,$B7,$8B,$FD,$6D,$7C,$AB,$FD,$07
	.dc.b	$87,$67,$89,$FE,$76,$F8,$3F,$CA,$79,$06,$C6,$18,$EC,$FF,$3E,$98
	.dc.b	$C7,$FD,$C1,$E6,$3E,$97,$F9,$B7,$B1,$F8,$3F,$9D,$FD,$04,$E6,$FF
	.dc.b	$D6,$B1,$8A,$DE,$48,$C4,$E8,$FF,$6F,$FA,$3F,$CA,$1B,$E2,$7E,$3B
	.dc.b	$E5,$3F,$D2,$71,$83,$FD,$8E,$E3,$3F,$F6,$3C,$EF,$CE,$FD,$95,$F3
	.dc.b	$9F,$F6,$BD,$6F,$83,$FD,$BD,$39,$7F,$DC,$B1,$9E,$BF,$F3,$54,$E5
	.dc.b	$FF,$36,$7A,$7F,$E9,$BC,$BF,$DD,$1E,$3F,$EF,$E3,$FD,$23,$18,$26
	.dc.b	$2C,$7F,$DD,$93,$C6,$31,$7D,$1E,$57,$FC,$27,$D6,$7D,$2F,$CE,$F6
	.dc.b	$9F,$63,$98,$C3,$73,$FF,$50,$E6,$CB,$FD,$47,$C8,$C3,$8D,$6A,$7F
	.dc.b	$D0,$D8,$F6,$8C,$66,$BE,$18,$FE,$23,$E9,$A5,$7E,$3B,$CF,$3F,$EB
	.dc.b	$0F,$80,$B8,$3C,$07,$53,$0F,$D1,$61,$AB,$DD,$6A,$73,$87,$33,$FD
	.dc.b	$35,$7C,$5F,$01,$9C,$3B,$46,$5F,$F4,$BF,$3B,$1E,$D5,$CA,$FF,$2F
	.dc.b	$F6,$7F,$CB,$F2,$F3,$E8,$8C,$57,$FB,$17,$E9,$9F,$4F,$CB,$3F,$89
	.dc.b	$5A,$7D,$DF,$01,$F9,$BB,$FF,$1C,$6C,$8C,$4F,$FB,$1E,$31,$D7,$F1
	.dc.b	$A3,$FD,$AB,$18,$AF,$F2,$07,$94,$F8,$77,$CD,$78,$63,$0D,$F8,$8F
	.dc.b	$E7,$98,$37,$FA,$35,$AC,$3E,$32,$DB,$F9,$54,$E9,$B3,$FE,$E3,$F2
	.dc.b	$BF,$33,$7C,$E3,$E1,$71,$F8,$67,$33,$FE,$EC,$C3,$FC,$90,$FA,$CE
	.dc.b	$45,$0E,$AB,$DB,$3C,$A7,$71,$87,$84,$ED,$FF,$77,$70,$BE,$29,$F7
	.dc.b	$FF,$4E,$30,$FD,$5D,$DF,$4C,$FA,$7D,$5F,$0C,$BF,$11,$5A,$9F,$33
	.dc.b	$63,$EC,$D7,$16,$CE,$A3,$18,$FD,$B6,$BF,$69,$AF,$FB,$C5,$AF,$EA
	.dc.b	$FC,$77,$C1,$7A,$5B,$7E,$4F,$88,$FC,$3F,$0F,$E5,$27,$37,$F4,$CF
	.dc.b	$D6,$4F,$E3,$1F,$C4,$B5,$FF,$80,$0D,$77,$CB,$EE,$8C,$66,$69,$23
	.dc.b	$EC,$6C,$F9,$27,$AF,$F3,$DB,$9F,$CA,$F9,$27,$D2,$E6,$BE,$94,$62
	.dc.b	$7F,$46,$DC,$FF,$44,$D7,$F4,$2D,$47,$F0,$97,$0D,$AF,$AF,$18,$CC
	.dc.b	$EB,$17,$CB,$7F,$63,$BF,$91,$18,$F4,$3F,$0D,$F2,$FF,$AC,$CB,$F8
	.dc.b	$54,$E5,$FC,$2C,$66,$BF,$70,$D7,$F6,$ED,$7C,$F8,$A3,$11,$FC,$85
	.dc.b	$B1,$9E,$BF,$C7,$C6,$0D,$DB,$C9,$FC,$0E,$1F,$CE,$1B,$87,$F2,$BF
	.dc.b	$0B,$F9,$B3,$EC,$8C,$E3,$FC,$7F,$C9,$FC,$DD,$6A,$BF,$86,$B5,$F3
	.dc.b	$B5,$E9,$1C,$09,$7F,$C6,$A7,$58,$3E,$65,$F6,$5F,$B1,$6D,$7E,$B2
	.dc.b	$BF,$D5,$5C,$DF,$CB,$27,$38,$FE,$67,$2F,$E6,$9D,$BE,$62,$61,$DD
	.dc.b	$FC,$8F,$D7,$FF,$2D,$CC,$FF,$2F,$F9,$EF,$E7,$99,$FF,$01,$8D,$13
	.dc.b	$BF,$F9,$B4,$EB,$0F,$BB,$3C,$F4,$F9,$27,$F4,$02,$C7,$DA,$FA,$67
	.dc.b	$C8,$7F,$0D,$9C,$EB,$FD,$8D,$B9,$FF,$2F,$E0,$FE,$21,$7C,$30,$FC
	.dc.b	$5D,$70,$C9,$F8,$F7,$F2,$C8,$FE,$CB,$18,$C7,$FA,$67,$13,$E7,$BF
	.dc.b	$BC,$61,$FD,$76,$B4,$DC,$39,$3F,$A6,$8C,$DF,$F0,$5B,$E5,$E7,$FC
	.dc.b	$2B,$DC,$FE,$37,$0E,$AF,$EB,$8F,$D7,$D7,$0C,$31,$18,$9D,$BB,$9D
	.dc.b	$3A,$C1,$E3,$BE,$CF,$E8,$B7,$0F,$EE,$D9,$BE,$9B,$F8,$4E,$1F,$C6
	.dc.b	$F0,$E3,$87,$02,$5F,$0A,$DD,$FC,$8C,$3A,$61,$FC,$8C,$F5,$7A,$7E
	.dc.b	$87,$D2,$D7,$25,$3F,$C6,$46,$E1,$F1,$EF,$A6,$E1,$18,$6C,$66,$78
	.dc.b	$35,$F8,$F3,$FE,$7A,$BE,$3D,$BF,$DC,$E7,$B3,$5D,$B5,$FE,$3F,$6F
	.dc.b	$A6,$7C,$F7,$A6,$30,$D8,$CC,$FB,$E7,$2D,$8C,$BF,$6A,$32,$D8,$CB
	.dc.b	$F8,$18,$CB,$F8,$0F,$C6,$3F,$42,$FC,$96,$BF,$35,$AE,$F9,$FC,$9F
	.dc.b	$45,$23,$C4,$D7,$C0,$D7,$C1,$68,$C4,$7F,$95,$B2,$7D,$D7,$D9,$E9
	.dc.b	$FF,$0E,$36,$33,$3F,$7E,$72,$D8,$CB,$FE,$08,$6B,$F9,$F8,$DB,$99
	.dc.b	$FC,$37,$23,$F0,$5F,$F5,$24,$FE,$2E,$31,$E5,$EA,$73,$3F,$A6,$9C
	.dc.b	$BB,$8C,$BF,$A7,$6B,$FA,$E6,$BF,$CF,$DA,$BF,$0F,$FC,$B3,$4F,$CC
	.dc.b	$38,$3F,$35,$E0,$BF,$0A,$FF,$69,$4F,$F9,$41,$AF,$F9,$3D,$AF,$F9
	.dc.b	$39,$AF,$F0,$1F,$93,$7C,$DF,$E9,$5B,$1F,$A9,$6B,$FA,$D6,$BF,$C1
	.dc.b	$9A,$F3,$8E,$44,$7E,$57,$94,$F9,$DF,$9E,$D8,$CC,$60,$FA,$7F,$67
	.dc.b	$F9,$0E,$CF,$E2,$0D,$9F,$E2,$ED,$7F,$B4,$37,$53,$49,$C9,$FC,$7E
	.dc.b	$B9,$B6,$B9,$BF,$C2,$BC,$1F,$F5,$3F,$99,$E5,$9F,$F5,$A9,$FE,$A3
	.dc.b	$71,$DB,$FE,$AD,$8F,$0D,$71,$7F,$D4,$7F,$8D,$FF,$8A,$1F,$F3,$3E
	.dc.b	$08,$C4,$7F,$A0,$63,$1E,$7F,$8E,$38,$B6,$33,$1C,$A7,$8C,$7A,$DB
	.dc.b	$1F,$ED,$06,$BC,$8F,$F5,$DF,$37,$F8,$93,$D1,$F3,$36,$3F,$D7,$F5
	.dc.b	$D9,$F7,$8E,$17,$C7,$FF,$9E,$A3,$31,$8F,$2B,$FD,$8F,$1B,$9D,$69
	.dc.b	$F1,$7E,$65,$F3,$4F,$91,$F8,$D7,$BC,$38,$E3,$CE,$3A,$E3,$CB,$87
	.dc.b	$FC,$87,$8F,$2C,$38,$DE,$F2,$FA,$58,$73,$38,$E3,$FE,$DA,$BE,$11
	.dc.b	$DB,$1E,$8F,$F9,$3A,$F9,$63,$FA,$39,$CD,$70,$AE,$64,$C3,$BD,$F0
	.dc.b	$84,$EE,$73,$5D,$A7,$C9,$18,$C3,$C9,$97,$C3,$9F,$24,$47,$B6,$3F
	.dc.b	$02,$75,$18,$3D,$17,$D7,$1F,$16,$3E,$1B,$F5,$61,$CF,$1F,$06,$1F
	.dc.b	$9A,$AD,$3C,$A7,$55,$C2,$35,$8F,$82,$F8,$92,$74,$FD,$D8,$D4,$60
	.dc.b	$93,$D4,$E7,$97,$D8,$CB,$D0,$4C,$3E,$1C,$65,$F1,$CF,$1C,$3B,$4E
	.dc.b	$63,$E1,$1F,$32,$FE,$34,$66,$7C,$D3,$C6,$72,$F4,$38,$19,$3E,$CE
	.dc.b	$1E,$A7,$06,$8D,$E3,$F2,$1C,$65,$F8,$17,$33,$C5,$CE,$5E,$D7,$E0
	.dc.b	$9E,$19,$E0,$7D,$89,$E0,$E2,$78,$61,$F0,$2B,$84,$63,$0F,$5C,$7C
	.dc.b	$2A,$E1,$97,$CA,$8D,$BF,$2B,$1F,$44,$EA,$7F,$15,$18,$BE,$91,$F0
	.dc.b	$E3,$15,$CA,$3D,$27,$79,$DC,$EA,$31,$3E,$16,$9D,$B1,$F6,$D7,$68
	.dc.b	$F8,$F9,$7D,$7C,$7F,$0F,$18,$3C,$19,$7D,$AC,$7B,$3F,$72,$34,$76
	.dc.b	$75,$87,$6C,$78,$DF,$C6,$9E,$19,$7C,$17,$E6,$5F,$19,$FC,$BE,$1D
	.dc.b	$9F,$1D,$7D,$5C,$7C,$D5,$F8,$48,$C3,$D8,$F9,$18,$F8,$5E,$87,$E3
	.dc.b	$27,$8B,$F8,$4C,$7E,$36,$3E,$4C,$BE,$5D,$F1,$8C,$4E,$B1,$F1,$D6
	.dc.b	$B0,$E7,$8F,$AE,$7D,$C6,$AB,$F0,$18,$73,$8C,$63,$F6,$B1,$EA,$FD
	.dc.b	$5C,$39,$E3,$D6,$3D,$87,$CE,$9D,$9F,$3A,$77,$87,$66,$CF,$33,$8E
	.dc.b	$3F,$19,$F8,$67,$3C,$3C,$77,$CE,$37,$87,$F3,$0C,$7D,$11,$F3,$E3
	.dc.b	$18,$76,$AE,$13,$D3,$0E,$D1,$E6,$C3,$ED,$E1,$F4,$EB,$FC,$2D,$87
	.dc.b	$F5,$96,$E3,$FE,$1D,$8C,$47,$F8,$8F,$E2,$FF,$DE,$4F,$F8,$7D,$AF
	.dc.b	$F8,$7D,$AE,$F9,$3F,$C4,$D1,$C3,$1F,$E8,$AD,$C7,$FA,$CF,$CD,$3D
	.dc.b	$16,$FF,$D8,$2B,$58,$FF,$53,$6B,$FD,$C5,$CE,$EB,$E5,$5C,$72,$97
	.dc.b	$FD,$D2,$F7,$5A,$8C,$4E,$B2,$F2,$BF,$D9,$67,$51,$89,$D9,$1F,$EF
	.dc.b	$51,$AC,$BE,$04,$E5,$B3,$AC,$BD,$B3,$93,$D9,$7E,$1A,$F8,$73,$AA
	.dc.b	$F0,$12,$39,$9A,$C7,$85,$F2,$3A,$65,$F0,$EF,$95,$70,$9D,$5F,$37
	.dc.b	$FB,$AE,$3E,$97,$33,$B7,$E7,$63,$F0,$EF,$9C,$68,$86,$8F,$ED,$55
	.dc.b	$CD,$C5,$F7,$6D,$FA,$5F,$84,$60,$E9,$87,$A7,$1F,$6C,$64,$D4,$65
	.dc.b	$E3,$B3,$3F,$C2,$B7,$3F,$45,$F2,$8C,$57,$C6,$CB,$CD,$18,$CB,$CC
	.dc.b	$DC,$BC,$D1,$8C,$7A,$9F,$8A,$CF,$E1,$E1,$E5,$BE,$0F,$AB,$0E,$B5
	.dc.b	$C3,$2F,$AF,$9F,$A7,$1E,$91,$8A,$DB,$E7,$AE,$D9,$D8,$D6,$5E,$0C
	.dc.b	$BB,$E5,$E1,$BD,$BF,$11,$F0,$3F,$52,$78,$46,$31,$FC,$16,$5E,$3C
	.dc.b	$FF,$01,$87,$43,$D7,$19,$C3,$EF,$C6,$6F,$BD,$FA,$A7,$E6,$61,$DB
	.dc.b	$1F,$44,$78,$72,$B5,$E0,$CF,$DF,$3E,$7A,$FA,$ED,$39,$D7,$0C,$FF
	.dc.b	$35,$9F,$8E,$7E,$05,$7C,$4A,$D4,$7D,$1C,$7C,$4F,$3C,$7F,$65,$3A
	.dc.b	$8F,$C5,$E3,$E2,$9F,$51,$A8,$E1,$8F,$82,$74,$FC,$6C,$BE,$0B,$71
	.dc.b	$F1,$57,$03,$C5,$87,$7C,$FE,$5E,$1C,$A7,$E5,$E1,$C9,$B7,$CB,$1F
	.dc.b	$1E,$3E,$1C,$BD,$71,$8A,$FD,$CC,$62,$7C,$51,$F3,$CF,$11,$EE,$8D
	.dc.b	$3D,$8D,$4F,$37,$59,$79,$F0,$F2,$47,$CE,$3E,$99,$D3,$2F,$5D,$F4
	.dc.b	$8D,$5F,$C9,$3A,$4E,$AF,$A6,$3F,$7A,$73,$87,$6A,$FC,$64,$FC,$1C
	.dc.b	$7D,$75,$F8,$7C,$7B,$56,$B1,$F4,$D6,$8F,$8D,$8F,$F2,$41,$F5,$EF
	.dc.b	$C1,$7F,$8C,$F1,$39,$DF,$80,$FF,$1C,$BD,$D6,$A7,$31,$FE,$C2,$BE
	.dc.b	$4F,$FB,$61,$B8,$FF,$B5,$DB,$1F,$EE,$B6,$C0,$FE,$BA,$E1,$B5,$FD
	.dc.b	$12,$78,$36,$38,$47,$90,$AF,$FB,$FB,$EF,$7C,$DF,$23,$FC,$D3,$7F
	.dc.b	$F7,$EC,$62,$BC,$1B,$FF,$BC,$E7,$31,$CA,$D1,$87,$FA,$A9,$8E,$B5
	.dc.b	$F8,$E3,$0D,$9E,$15,$ED,$9C,$BD,$63,$31,$E7,$8F,$07,$80,$F8,$67
	.dc.b	$A1,$E0,$AF,$8D,$39,$8C,$42,$31,$5D,$E3,$75,$CA,$31,$1F,$3F,$E1
	.dc.b	$1F,$37,$CA,$11,$89,$DC,$62,$75,$39,$8E,$6D,$7C,$0D,$7E,$03,$5F
	.dc.b	$04,$7A,$2D,$F3,$CF,$C1,$30,$DC,$3C,$F3,$A9,$C9,$E7,$FC,$41,$F4
	.dc.b	$54,$95,$E8,$8C,$BF,$03,$57,$94,$EE,$F9,$31,$86,$C6,$67,$83,$74
	.dc.b	$F9,$95,$B9,$CD,$70,$23,$F3,$BA,$1E,$6F,$F8,$AE,$FF,$05,$18,$EC
	.dc.b	$FB,$2D,$AE,$15,$BA,$FD,$24,$E5,$FB,$8D,$79,$C6,$D4,$65,$FD,$67
	.dc.b	$A2,$F5,$9E,$47,$E2,$23,$0D,$8C,$E1,$F8,$69,$CB,$63,$35,$F8,$D9
	.dc.b	$CB,$63,$2F,$E7,$E3,$2F,$E4,$FF,$03,$FE,$06,$DE,$BF,$C8,$46,$23
	.dc.b	$F4,$CD,$53,$E9,$FC,$57,$CC,$4C,$3C,$8D,$9F,$13,$5F,$16,$08,$FC
	.dc.b	$F4,$7F,$00,$3F,$3A,$3A,$EB,$4C,$BF,$A2,$F5,$DE,$C7,$92,$7F,$57
	.dc.b	$18,$C7,$53,$99,$FD,$94,$E5,$DF,$B1,$F8,$46,$EF,$D6,$31,$1C,$1A
	.dc.b	$FD,$AF,$59,$EB,$7E,$D3,$14,$62,$7E,$C3,$5F,$AC,$D7,$EA,$35,$E9
	.dc.b	$3C,$E3,$11,$F4,$1A,$BD,$7A,$49,$F9,$2F,$0A,$F1,$FC,$7F,$32,$57
	.dc.b	$91,$AF,$91,$AF,$91,$FE,$1F,$C7,$FA,$8F,$26,$E7,$FE,$59,$AF,$D2
	.dc.b	$C6,$08,$E0,$87,$E9,$A3,$83,$FB,$8E,$87,$B1,$FC,$73,$A7,$F8,$1B
	.dc.b	$63,$33,$FC,$02,$30,$47,$04,$3F,$95,$F9,$4F,$57,$F1,$6D,$8F,$E5
	.dc.b	$2D,$77,$1E,$1F,$55,$EA,$D2,$4F,$F1,$78,$CB,$CB,$EA,$BA,$91,$C1
	.dc.b	$0F,$1B,$96,$99,$CF,$FD,$5B,$E1,$3E,$49,$F5,$7F,$D5,$7A,$5E,$87
	.dc.b	$F4,$E9,$CE,$1F,$D9,$A3,$FA,$D6,$AC,$BF,$AE,$65,$FD,$D7,$D7,$7B
	.dc.b	$1F,$D7,$FB,$E4,$63,$1F,$ED,$90,$8C,$47,$F6,$C8,$C4,$F1,$8F,$EF
	.dc.b	$B3,$A9,$E7,$39,$9F,$F7,$1C,$E5,$DC,$66,$79,$C7,$A4,$FE,$4B,$E8
	.dc.b	$BD,$8F,$81,$9F,$F2,$E8,$C4,$EE,$B8,$BF,$C1,$67,$C7,$F6,$1D,$71
	.dc.b	$34,$93,$EE,$8C,$B6,$32,$FB,$E3,$2D,$8F,$1B,$FC,$EB,$E2,$9F,$05
	.dc.b	$8E,$A4,$62,$7C,$D7,$CB,$87,$EC,$C7,$F8,$AF,$83,$FA,$F7,$B3,$F0
	.dc.b	$61,$18,$7D,$86,$FC,$A4,$FC,$7F,$65,$ED,$3D,$BF,$2D,$F2,$BD,$BC
	.dc.b	$8C,$5F,$7C,$3B,$63,$E4,$C7,$9E,$1C,$A7,$3B,$FE,$33,$1E,$77,$EA
	.dc.b	$C7,$99,$E6,$C3,$8B,$D6,$77,$5C,$71,$C1,$E9,$CB,$D1,$58,$3C,$D8
	.dc.b	$F8,$1F,$81,$3B,$CB,$C9,$87,$49,$C9,$C1,$C6,$3E,$28,$C5,$75,$CB
	.dc.b	$D1,$87,$5A,$D4,$62,$FA,$5F,$B2,$31,$7E,$18,$E5,$97,$AD,$EA,$7C
	.dc.b	$1C,$BD,$6E,$2B,$8B,$67,$FC,$D5,$87,$81,$E5,$3A,$23,$AA,$E7,$18
	.dc.b	$AF,$7B,$6B,$94,$62,$B9,$91,$46,$9F,$64,$65,$A6,$AB,$BC,$67,$1F
	.dc.b	$A5,$3A,$AE,$51,$82,$63,$F4,$2F,$8D,$78,$C8,$71,$72,$F8,$09,$3F
	.dc.b	$4E,$32,$F4,$C7,$E7,$CF,$97,$2F,$C0,$C6,$23,$A4,$61,$F9,$C7,$87
	.dc.b	$0F,$3E,$1F,$72,$FA,$4F,$D7,$AE,$19,$7C,$D8,$E6,$75,$9D,$C7,$E9
	.dc.b	$71,$F2,$DF,$32,$63,$F8,$5B,$F0,$9E,$C8,$F8,$98,$76,$CB,$E6,$46
	.dc.b	$27,$B3,$C9,$F2,$18,$AF,$83,$97,$A9,$C4,$68,$EF,$97,$9A,$FF,$11
	.dc.b	$87,$79,$FA,$33,$F3,$63,$11,$E5,$AF,$46,$5F,$06,$B9,$4F,$CF,$3B
	.dc.b	$C7,$E3,$F2,$F8,$35,$8C,$3B,$D5,$3E,$84,$E9,$F2,$E5,$EF,$6D,$7B
	.dc.b	$27,$57,$E1,$AD,$36,$75,$8F,$A3,$1F,$2E,$5F,$0F,$1E,$67,$C3,$BE
	.dc.b	$0F,$E0,$A3,$51,$83,$ED,$57,$3C,$3E,$84,$62,$BD,$47,$E0,$8E,$AF
	.dc.b	$C1,$9E,$18,$79,$27,$78,$7C,$D6,$DF,$AB,$2F,$BF,$5B,$9E,$71,$FB
	.dc.b	$5C,$7C,$AE,$A3,$D4,$DC,$7C,$11,$D2,$BE,$BC,$E4,$E5,$97,$CA,$71
	.dc.b	$3F,$64,$EE,$7D,$4C,$BD,$2F,$C5,$9E,$0F,$E1,$B1,$F4,$36,$FD,$18
	.dc.b	$7C,$9C,$3B,$E1,$CA,$3C,$B5,$FC,$88,$FF,$6E,$65,$FD,$65,$B8,$7F
	.dc.b	$52,$6E,$1F,$E3,$0E,$FF,$F7,$CC,$7F,$8D,$9A,$F2,$6B,$FE,$1C,$6B
	.dc.b	$FD,$A5,$E3,$7F,$CF,$1B,$8F,$F2,$EF,$6D,$E7,$38,$FF,$B0,$56,$B0
	.dc.b	$E5,$E0,$E0,$FE,$D9,$7C,$2B,$13,$FD,$52,$B8,$3F,$DD,$9E,$5B,$AB
	.dc.b	$84,$EE,$F8,$E5,$E5,$7A,$E3,$E3,$9D,$12,$74,$FF,$6B,$AE,$71,$82
	.dc.b	$65,$F0,$EF,$9E,$5D,$A3,$17,$CD,$F8,$58,$77,$BE,$33,$F0,$0D,$1F
	.dc.b	$D1,$B0,$EF,$5C,$A7,$73,$EF,$C3,$BB,$C0,$C1,$CB,$1F,$85,$87,$7B
	.dc.b	$E4,$49,$F0,$5F,$F8,$3F,$0E,$F1,$89,$D3,$D2,$75,$5C,$E3,$44,$35
	.dc.b	$3E,$EC,$7E,$14,$67,$56,$7F,$5B,$1E,$F3,$C0,$D4,$67,$0E,$D1,$92
	.dc.b	$57,$2C,$FE,$0E,$5F,$12,$31,$3F,$07,$0E,$B1,$89,$D6,$1D,$71,$F2
	.dc.b	$DF,$48,$C3,$F2,$EB,$86,$5F,$1F,$3F,$97,$7C,$27,$35,$F0,$A7,$2F
	.dc.b	$1B,$FC,$AC,$EA,$73,$87,$8B,$2F,$8B,$18,$C7,$E7,$D7,$4C,$3E,$24
	.dc.b	$E5,$F8,$B5,$C3,$2F,$6E,$3D,$89,$1F,$13,$0E,$71,$8A,$FC,$D6,$7F
	.dc.b	$33,$1E,$C4,$9F,$83,$97,$96,$BE,$55,$6E,$79,$4F,$E6,$32,$F2,$57
	.dc.b	$C9,$C7,$C9,$7B,$7E,$35,$F4,$C7,$C7,$1A,$CF,$2D,$BF,$90,$DC,$7C
	.dc.b	$47,$7C,$FD,$F1,$C2,$7D,$75,$6B,$58,$F8,$67,$99,$A2,$C6,$A7,$86
	.dc.b	$1F,$77,$0E,$B5,$A8,$FC,$D3,$4F,$BD,$1A,$AF,$2C,$E9,$B3,$AB,$F1
	.dc.b	$46,$32,$F8,$F9,$FC,$FC,$BD,$D8,$7C,$4C,$3E,$7E,$5E,$47,$93,$5F
	.dc.b	$C1,$E5,$E1,$9E,$18,$FE,$D2,$B5,$87,$E3,$EB,$57,$F8,$83,$E1,$57
	.dc.b	$99,$E7,$97,$9B,$0E,$67,$AC,$B1,$E2,$3E,$11,$C3,$0E,$F9,$FD,$D8
	.dc.b	$ED,$7F,$35,$DE,$3E,$3C,$3B,$46,$B0,$EC,$D3,$B5,$79,$6B,$86,$1D
	.dc.b	$B2,$FC,$E5,$7F,$10,$87,$EA,$9F,$ED,$57,$FD,$B1,$C7,$FE,$C0,$FC
	.dc.b	$7B,$F8,$B4,$6B,$A3,$F8,$C4,$E6,$3F,$DB,$15,$C6,$FF,$DD,$8D,$8F
	.dc.b	$F7,$A3,$60,$7F,$59,$70,$DA,$FE,$95,$3C,$1B,$1C,$23,$C4,$7F,$BF
	.dc.b	$27,$8A,$F7,$1E,$87,$9B,$A9,$EE,$FA,$A7,$35,$EE,$9C,$BF,$EF,$18
	.dc.b	$CE,$1F,$D6,$CF,$ED,$CE,$23,$D3,$18,$BF,$2B,$63,$CC,$F9,$E7,$83
	.dc.b	$67,$73,$E7,$6C,$EA,$7E,$2F,$B0,$F8,$7F,$ED,$9D,$FF,$DB,$7C,$37
	.dc.b	$84,$F9,$6F,$E7,$B7,$BC,$FE,$6D,$EE,$B5,$1F,$EF,$39,$D6,$3E,$A3
	.dc.b	$BE,$3E,$F7,$DB,$1A,$74,$D7,$31,$B7,$8C,$F3,$69,$67,$82,$C5,$ED
	.dc.b	$BD,$CE,$51,$86,$C6,$6F,$A4,$E5,$ED,$4B,$7B,$9E,$38,$C3,$63,$33
	.dc.b	$C1,$B3,$F0,$79,$5E,$67,$F6,$CA,$F5,$46,$3A,$FD,$CD,$8C,$CE,$EB
	.dc.b	$93,$5F,$C2,$CE,$DC,$D7,$29,$D4,$7E,$8B,$99,$BB,$BF,$E6,$46,$1B
	.dc.b	$19,$C7,$E7,$4E,$5B,$19,$AF,$B9,$39,$6C,$65,$FC,$CC,$65,$FC,$DF
	.dc.b	$9C,$F5,$9F,$27,$F7,$63,$0D,$8C,$E1,$F7,$E7,$2D,$8C,$D7,$E3,$67
	.dc.b	$2D,$8C,$BE,$06,$BF,$AC,$F6,$1E,$EF,$DB,$6D,$FE,$01,$B3,$F8,$F6
	.dc.b	$BC,$FD,$AF,$31,$27,$C8,$D8,$F1,$35,$E1,$1F,$8C,$7F,$8A,$F8,$15
	.dc.b	$34,$8F,$E7,$FD,$87,$B2,$CF,$F2,$B1,$86,$C6,$67,$83,$5F,$D9,$4E
	.dc.b	$D7,$A8,$F4,$7F,$87,$CE,$63,$13,$FC,$AA,$31,$3A,$AE,$D5,$AA,$78
	.dc.b	$9F,$E9,$97,$F8,$CB,$5F,$E6,$6D,$77,$C5,$FE,$9D,$76,$D7,$6D,$77
	.dc.b	$E4,$FE,$D2,$79,$12,$74,$D8,$D4,$60,$8E,$09,$3F,$C2,$E7,$83,$67
	.dc.b	$6D,$9D,$BF,$EB,$23,$04,$30,$43,$D6,$F0,$69,$B2,$9B,$3D,$EE,$DA
	.dc.b	$6C,$A6,$CF,$86,$64,$A6,$67,$D6,$D9,$F5,$B6,$7B,$3C,$0F,$19,$0E
	.dc.b	$E5,$34,$60,$86,$09,$3E,$07,$83,$5D,$B5,$DC,$F9,$1D,$B5,$DB,$5D
	.dc.b	$CF,$A2,$79,$12,$74,$D8,$D4,$60,$8E,$09,$3F,$26,$78,$36,$76,$D9
	.dc.b	$DB,$F2,$4C,$10,$C1,$0F,$5B,$C1,$A6,$CA,$6C,$F7,$BB,$69,$B2,$9B
	.dc.b	$3E,$19,$92,$99,$9F,$5B,$67,$D6,$D9,$EC,$F0,$3C,$64,$3B,$94,$D1
	.dc.b	$82,$18,$24,$F8,$1E,$0D,$76,$D7,$73,$E4,$76,$D7,$6D,$77,$3E,$89
	.dc.b	$E4,$49,$D3,$63,$51,$82,$38,$24,$FC,$99,$E0,$D9,$DB,$67,$6F,$C9
	.dc.b	$30,$43,$04,$3D,$6F,$06,$9B,$29,$B3,$DE,$ED,$A6,$CA,$6C,$F8,$66
	.dc.b	$4A,$66,$7D,$6D,$9F,$8B,$CA,$E6,$FE,$54,$D9,$FE,$44,$D7,$FC,$71
	.dc.b	$5F,$EF,$66,$C6,$5F,$E7,$AD,$7F,$9E,$B5,$FE,$86,$D8,$E5,$1C,$27
	.dc.b	$C2,$47,$C2,$D7,$BB,$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$3B,$D5
	.dc.b	$A7,$22,$9C,$4A,$71,$3D,$AF,$16,$9C,$4A,$71,$29,$C6,$3E,$1B,$D0
	.dc.b	$86,$8A,$E9,$C1,$1C,$12,$7E,$71,$27,$C9,$19,$6C,$66,$7C,$D3,$96
	.dc.b	$C6,$5F,$44,$65,$B1,$97,$D4,$D7,$CE,$D7,$CC,$D8,$E5,$1C,$27,$C6
	.dc.b	$47,$C2,$D7,$BB,$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$3B,$D5,$A7
	.dc.b	$22,$9C,$4A,$71,$3D,$AF,$16,$9C,$4A,$71,$29,$C6,$3E,$1B,$D0,$86
	.dc.b	$8A,$E9,$C1,$1C,$12,$7E,$71,$27,$C9,$19,$6C,$66,$7C,$D3,$96,$C6
	.dc.b	$5F,$44,$65,$B1,$97,$D4,$D7,$CE,$D7,$CC,$D8,$E5,$1C,$27,$C6,$47
	.dc.b	$C2,$D7,$BB,$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$3B,$D5,$A7,$22
	.dc.b	$9C,$4A,$71,$3D,$AF,$16,$9C,$4A,$71,$29,$C6,$3E,$1B,$D0,$86,$8A
	.dc.b	$E9,$67,$FE,$72,$7F,$CE,$99,$7F,$9C,$B3,$FF,$3A,$68,$9F,$F3,$EC
	.dc.b	$61,$A7,$08,$FF,$43,$91,$FF,$43,$B7,$4F,$F3,$E5,$F7,$9F,$0D,$F7
	.dc.b	$79,$1D,$0F,$41,$30,$F4,$13,$2F,$C1,$13,$2F,$91,$19,$6C,$66,$7E
	.dc.b	$54,$E5,$B1,$97,$E6,$46,$5B,$19,$7E,$73,$5F,$96,$D7,$CC,$D8,$E5
	.dc.b	$1C,$27,$C6,$47,$C2,$D7,$BB,$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F
	.dc.b	$3B,$D5,$A7,$22,$9C,$4A,$71,$3D,$AF,$16,$9C,$4A,$71,$29,$C6,$3E
	.dc.b	$1B,$D0,$86,$8A,$E9,$C1,$1C,$12,$7E,$71,$27,$C9,$19,$6C,$66,$7C
	.dc.b	$D3,$96,$C6,$5F,$44,$65,$B1,$97,$D4,$D7,$CE,$D7,$FB,$56,$7F,$E3
	.dc.b	$E7,$FC,$4D,$A7,$F8,$C5,$FF,$15,$9F,$E3,$62,$1F,$E3,$22,$9A,$30
	.dc.b	$43,$04,$AF,$F1,$9B,$C1,$AE,$DA,$EE,$7C,$8E,$DA,$ED,$AE,$E7,$D1
	.dc.b	$3C,$89,$3A,$6C,$6A,$30,$47,$04,$9F,$F2,$64,$F0,$6C,$ED,$B3,$B7
	.dc.b	$E4,$98,$21,$82,$1E,$B7,$83,$4D,$94,$D9,$EF,$76,$D3,$65,$36,$7C
	.dc.b	$33,$25,$33,$3E,$B6,$CF,$AD,$B3,$D9,$E0,$78,$C8,$77,$29,$A3,$04
	.dc.b	$30,$49,$F0,$3C,$1A,$ED,$AE,$E7,$C8,$ED,$AF,$5C,$7F,$9A,$13,$1F
	.dc.b	$E6,$91,$FC,$DB,$3F,$E8,$59,$FF,$3A,$94,$7F,$3F,$9F,$E8,$C7,$F3
	.dc.b	$A2,$9C,$4A,$6B,$5A,$D2,$86,$B4,$E7,$0E,$27,$03,$83,$AC,$E6,$31
	.dc.b	$5C,$2B,$75,$A6,$C6,$67,$6F,$AA,$B9,$56,$E7,$33,$FD,$8E,$72,$FB
	.dc.b	$38,$5C,$5A,$8F,$ED,$11,$8B,$F7,$B6,$38,$4E,$EB,$D9,$C7,$71,$34
	.dc.b	$98,$7A,$5B,$7E,$D6,$BF,$4B,$91,$CA,$E0,$F2,$46,$1B,$19,$BF,$AB
	.dc.b	$39,$6C,$66,$BE,$DC,$E5,$B1,$97,$C0,$D7,$FC,$11,$C8,$E6,$72,$FC
	.dc.b	$58,$C3,$63,$38,$7E,$1A,$72,$D8,$CD,$7E,$32,$72,$D8,$CB,$F9,$28
	.dc.b	$CB,$F9,$2E,$77,$47,$D5,$84,$62,$FF,$26,$D8,$FC,$BB,$5F,$CE,$F4
	.dc.b	$BA,$9D,$7F,$7E,$30,$D8,$CD,$FE,$82,$72,$F6,$EC,$F1,$60,$69,$23
	.dc.b	$44,$C3,$C7,$19,$6C,$7F,$8D,$75,$FC,$FD,$BB,$5D,$DF,$A2,$E8,$78
	.dc.b	$DF,$B6,$CF,$C1,$8F,$13,$FC,$7D,$87,$1B,$E1,$5F,$E4,$19,$78,$0F
	.dc.b	$07,$F5,$B1,$FC,$00,$CB,$8D,$38,$4E,$E7,$94,$62,$74,$FE,$E6,$74
	.dc.b	$FE,$CA,$79,$9C,$1D,$4F,$08,$CC,$6E,$78,$E5,$EE,$87,$23,$C2,$78
	.dc.b	$7E,$C9,$E7,$87,$F9,$76,$F8,$5E,$EB,$52,$F1,$1E,$2F,$85,$B1,$E5
	.dc.b	$35,$19,$23,$98,$DE,$1E,$D6,$C7,$C9,$6C,$7C,$F7,$78,$79,$3C,$D7
	.dc.b	$8F,$E4,$9C,$ED,$FE,$75,$CD,$5A,$C7,$D9,$3F,$CD,$3C,$5F,$D8,$F4
	.dc.b	$E3,$3F,$F3,$E4,$61,$FE,$61,$3F,$E7,$FF,$29,$E5,$3C,$AF,$91,$4B
	.dc.b	$E1,$B7,$CF,$8C,$77,$32,$ED,$18,$7C,$18,$F4,$C3,$9E,$1C,$B2,$F3
	.dc.b	$61,$C7,$0F,$19,$E6,$71,$8F,$DD,$C3,$96,$7D,$B0,$E5,$7B,$CB,$18
	.dc.b	$70,$C7,$D9,$9F,$AF,$0C,$D4,$BE,$E7,$9C,$F0,$E3,$D9,$CE,$3D,$9D
	.dc.b	$65,$F0,$63,$C3,$3B,$8C,$65,$E1,$AF,$21,$C9,$C4,$EF,$2F,$83,$87
	.dc.b	$8E,$F8,$BF,$23,$2F,$46,$1C,$63,$55,$C2,$3C,$F3,$AA,$E3,$1B,$23
	.dc.b	$F2,$63,$53,$C6,$36,$47,$E3,$E7,$F5,$B1,$F5,$5F,$C0,$9E,$06,$B3
	.dc.b	$F1,$C7,$BA,$F9,$2C,$BD,$CE,$6B,$8B,$4D,$63,$E4,$70,$77,$35,$3C
	.dc.b	$9C,$13,$3F,$A3,$3C,$63,$18,$FA,$27,$39,$7A,$F3,$F3,$D7,$5C,$3A
	.dc.b	$57,$DD,$25,$FD,$6C,$3A,$46,$2F,$DF,$5C,$2B,$B1,$23,$EC,$61,$DF
	.dc.b	$0F,$A3,$8F,$C0,$BE,$15,$E2,$C3,$CB,$9D,$C7,$96,$1F,$73,$1E,$59
	.dc.b	$FC,$C8,$F4,$67,$E4,$AD,$CF,$D0,$CB,$AD,$78,$0F,$93,$5B,$CB,$B1
	.dc.b	$E6,$C6,$CE,$EB,$8D,$F7,$CB,$CB,$3C,$CF,$9B,$7C,$5C,$47,$C5,$7D
	.dc.b	$17,$D2,$BC,$98,$78,$67,$78,$FC,$DC,$BD,$B9,$FE,$12,$7D,$B1,$87
	.dc.b	$F0,$78,$F8,$B0,$F2,$5F,$C0,$BF,$C9,$57,$4C,$BE,$AE,$3D,$EF,$71
	.dc.b	$E7,$6C,$FB,$E7,$55,$E1,$C3,$B3,$70,$ED,$5F,$96,$C7,$C8,$FC,$23
	.dc.b	$DE,$78,$4E,$8E,$AB,$E2,$61,$FB,$18,$C5,$74,$CF,$E7,$E1,$DF,$2F
	.dc.b	$33,$71,$E7,$9F,$CB,$8F,$C5,$67,$E8,$9F,$8B,$87,$CB,$AD,$47,$E0
	.dc.b	$31,$EF,$87,$C4,$6C,$74,$3D,$19,$7D,$68,$CB,$6B,$9C,$FD,$46,$E7
	.dc.b	$F1,$A3,$C3,$7D,$67,$38,$78,$E7,$AE,$5E,$8C,$7A,$B6,$B7,$7F,$02
	.dc.b	$B9,$9E,$5C,$7E,$B9,$E3,$7E,$71,$FE,$8E,$C7,$FA,$D3,$71,$FE,$94
	.dc.b	$DA,$FF,$59,$4F,$F9,$8F,$C9,$FE,$91,$E0,$1A,$7F,$56,$F3,$1E,$67
	.dc.b	$FA,$5F,$4F,$ED,$55,$BC,$7F,$A1,$B7,$3F,$E8,$4D,$9F,$ED,$57,$CA
	.dc.b	$BA,$9C,$30,$B7,$FD,$6E,$75,$39,$24,$74,$9F,$26,$1E,$39,$FE,$F3
	.dc.b	$87,$76,$CF,$4C,$3C,$35,$B9,$FE,$F1,$5B,$6D,$6D,$FE,$6D,$87,$8B
	.dc.b	$0F,$04,$78,$B0,$F8,$35,$E2,$8E,$07,$23,$47,$2C,$3C,$18,$FC,$F8
	.dc.b	$C9,$BB,$F8,$13,$AC,$7C,$95,$B9,$F0,$9A,$AE,$F1,$9C,$3C,$71,$FD
	.dc.b	$DB,$1F,$44,$62,$B8,$B4,$F8,$73,$CA,$FB,$9F,$27,$75,$F2,$9F,$96
	.dc.b	$7A,$30,$F0,$3E,$D3,$96,$5F,$26,$B8,$E5,$EB,$8C,$CF,$96,$30,$F3
	.dc.b	$C3,$C1,$87,$78,$F8,$98,$F9,$DF,$0E,$5F,$05,$E1,$87,$7A,$E1,$97
	.dc.b	$92,$78,$9C,$2F,$DE,$F0,$3E,$AE,$5E,$67,$CB,$87,$C6,$7D,$86,$DC
	.dc.b	$46,$E3,$13,$E5,$8C,$4F,$19,$F0,$9E,$28,$CC,$FA,$E3,$35,$F2,$A7
	.dc.b	$31,$E7,$21,$C6,$FD,$B8,$FD,$6C,$7E,$06,$5F,$2B,$2F,$06,$3C,$EF
	.dc.b	$75,$EC,$CB,$E0,$D6,$B0,$E5,$8F,$D3,$34,$78,$72,$F1,$BE,$68,$DC
	.dc.b	$FD,$13,$95,$F8,$A7,$89,$E7,$CB,$C9,$87,$BF,$2F,$24,$62,$7D,$B1
	.dc.b	$87,$8B,$5F,$9A,$78,$F2,$F5,$B8,$CB,$CB,$3F,$7C,$99,$7E,$2A,$FB
	.dc.b	$57,$A7,$1F,$DB,$3D,$0F,$31,$E0,$9E,$0D,$8D,$E5,$EE,$8F,$C3,$E3
	.dc.b	$E2,$8C,$1F,$0F,$2F,$75,$FC,$67,$E7,$98,$BF,$8D,$7E,$1C,$7D,$D7
	.dc.b	$CA,$BF,$15,$87,$57,$85,$7D,$53,$86,$1E,$66,$E1,$E6,$9D,$DF,$E5
	.dc.b	$AF,$95,$7C,$EB,$E1,$3D,$E3,$E6,$63,$F1,$B1,$F9,$38,$F3,$9E,$0F
	.dc.b	$A2,$B8,$47,$C8,$3C,$27,$95,$CC,$FB,$31,$F1,$9E,$CB,$ED,$87,$C5
	.dc.b	$BF,$D6,$DF,$FB,$57,$57,$9A,$F3,$FF,$9E,$71,$FF,$AE,$BC,$FB,$C8
	.dc.b	$7F,$A0,$5F,$0A,$B7,$C7,$1F,$F6,$93,$5F,$F7,$53,$67,$FD,$EC,$D8
	.dc.b	$FF,$7E,$35,$47,$57,$9C,$FF,$BF,$5D,$B5,$DC,$BD,$0F,$4F,$57,$FB
	.dc.b	$9A,$7F,$D9,$D9,$7F,$BC,$7A,$7C,$F5,$27,$CE,$F3,$C7,$E1,$B7,$0E
	.dc.b	$93,$A2,$1E,$C8,$D1,$E3,$7F,$DA,$9E,$89,$9E,$17,$B1,$78,$0F,$09
	.dc.b	$E8,$FF,$6B,$8C,$3F,$D6,$2F,$95,$2F,$93,$D8,$C9,$FE,$ED,$8A,$D2
	.dc.b	$9B,$33,$CA,$8C,$C6,$4F,$F0,$D3,$88,$D4,$74,$E9,$7A,$4F,$4F,$AC
	.dc.b	$61,$B1,$99,$E0,$D7,$C3,$EA,$75,$23,$E6,$B4,$74,$F2,$DE,$9D,$25
	.dc.b	$78,$E3,$2D,$8C,$BF,$13,$14,$EA,$30,$D7,$E6,$F1,$FC,$DE,$FF,$8E
	.dc.b	$D8,$F9,$6D,$7E,$7B,$5E,$3A,$A7,$8B,$63,$8B,$5E,$2D,$7E,$EF,$9E
	.dc.b	$F6,$1E,$B7,$D8,$8C,$36,$33,$3F,$7A,$72,$D8,$CB,$F8,$48,$CB,$63
	.dc.b	$2F,$81,$AF,$E3,$BD,$77,$B1,$F8,$46,$C7,$E2,$9A,$FE,$39,$AF,$38
	.dc.b	$F3,$11,$F2,$35,$F1,$35,$E0,$7E,$5B,$D8,$7B,$3F,$9F,$8F,$CD,$C6
	.dc.b	$32,$D4,$E6,$7F,$43,$39,$77,$EC,$B2,$E8,$D8,$CC,$60,$FD,$0A,$F0
	.dc.b	$69,$1F,$D2,$BF,$3F,$D6,$7B,$5F,$B2,$9F,$DF,$52,$73,$DD,$BA,$D5
	.dc.b	$78,$E7,$2F,$D5,$6B,$F8,$26,$BD,$1A,$F1,$8E,$6B,$D9,$7A,$38,$6D
	.dc.b	$FF,$00,$6C,$FF,$04,$6B,$DF,$DA,$F2,$E3,$E9,$6E,$26,$92,$FE,$13
	.dc.b	$6B,$D4,$FB,$B9,$3F,$8B,$79,$B7,$80,$D2,$5F,$E9,$1B,$7F,$A4,$6B
	.dc.b	$FC,$83,$9F,$F7,$3E,$47,$CA,$6D,$7F,$01,$8C,$74,$1A,$48,$D1,$2B
	.dc.b	$F9,$2F,$B6,$F7,$3E,$5B,$67,$F9,$3B,$67,$73,$F6,$BC,$27,$B8,$F7
	.dc.b	$70,$DC,$3F,$98,$CE,$A7,$27,$D8,$F6,$1E,$E5,$25,$7F,$36,$8C,$B6
	.dc.b	$33,$8F,$F3,$C9,$CB,$63,$C9,$ED,$BD,$CA,$4B,$F1,$C6,$5F,$E7,$5E
	.dc.b	$FB,$DF,$FE,$A9,$1F,$3E,$30,$FF,$59,$31,$1A,$C3,$FA,$E7,$0A,$B7
	.dc.b	$1F,$D5,$23,$13,$B8,$46,$2B,$A4,$72,$AE,$31,$D8,$CC,$70,$8E,$BD
	.dc.b	$6F,$53,$FB,$67,$33,$D0,$7C,$07,$F6,$DA,$D6,$1C,$73,$C6,$1C,$2F
	.dc.b	$79,$7F,$75,$BF,$F1,$17,$BB,$FA,$68,$FE,$DE,$1C,$61,$F0,$DA,$FF
	.dc.b	$85,$9B,$1B,$EE,$F8,$0D,$3D,$65,$9F,$EF,$2D,$3B,$CF,$ED,$AB,$D3
	.dc.b	$9F,$B7,$DD,$F4,$DF,$7C,$2C,$E7,$0F,$56,$5F,$E0,$6F,$2F,$FA,$96
	.dc.b	$F8,$C3,$FC,$25,$18,$7F,$8F,$4F,$F8,$57,$DA,$7B,$7F,$B6,$C3,$85
	.dc.b	$2F,$86,$9F,$32,$31,$E5,$FD,$C7,$99,$D9,$C4,$7D,$2C,$3A,$F7,$7A
	.dc.b	$AB,$8E,$7D,$B0,$EB,$9F,$A7,$1E,$2F,$97,$3F,$25,$F6,$38,$96,$B8
	.dc.b	$BE,$1B,$F8,$18,$F8,$33,$ED,$8F,$4A,$D1,$C5,$C4,$EE,$31,$5C,$08
	.dc.b	$FB,$A3,$59,$FC,$3C,$3C,$33,$E4,$8C,$5F,$4C,$7D,$31,$8C,$39,$B6
	.dc.b	$BA,$4E,$B2,$F8,$55,$A6,$CE,$B3,$F9,$58,$F8,$09,$8F,$78,$CB,$EF
	.dc.b	$9F,$25,$72,$C7,$EA,$63,$DE,$7E,$1C,$78,$C9,$3F,$46,$F9,$E3,$E3
	.dc.b	$AD,$CE,$73,$C4,$76,$32,$33,$F8,$F3,$E1,$24,$6A,$32,$F0,$39,$1C
	.dc.b	$DC,$B8,$CF,$DF,$7F,$7B,$2F,$54,$E7,$3F,$81,$5F,$84,$BE,$2F,$09
	.dc.b	$FB,$A7,$CB,$C7,$D3,$19,$CF,$E2,$E7,$F0,$2B,$71,$F1,$1F,$66,$3E
	.dc.b	$0A,$E1,$9F,$B3,$1E,$F9,$78,$89,$9F,$9F,$2E,$99,$73,$CB,$DD,$97
	.dc.b	$3B,$DC,$78,$31,$F9,$CD,$3E,$36,$5E,$DA,$E1,$8F,$AA,$FF,$05,$9D
	.dc.b	$8F,$B1,$9F,$91,$C6,$1F,$04,$E8,$7C,$97,$CD,$3A,$CB,$E0,$57,$1A
	.dc.b	$DE,$3E,$1C,$FD,$79,$FD,$C9,$F5,$C6,$1F,$3E,$5E,$47,$8B,$EE,$C3
	.dc.b	$E7,$E3,$DE,$FE,$56,$7E,$CB,$DE,$7E,$46,$9F,$0F,$3F,$26,$1F,$2F
	.dc.b	$3F,$26,$3C,$EB,$F1,$D8,$73,$AD,$65,$E3,$6E,$3F,$1A,$75,$97,$AF
	.dc.b	$1E,$78,$78,$2F,$E1,$D6,$A3,$DB,$7E,$59,$E7,$1E,$59,$D3,$F3,$4F
	.dc.b	$A1,$1A,$39,$E7,$F1,$E3,$51,$F1,$1F,$06,$1F,$12,$75,$87,$E4,$27
	.dc.b	$F2,$71,$8C,$7E,$65,$7E,$22,$FB,$3C,$72,$F8,$97,$F5,$9B,$87,$BA
	.dc.b	$B7,$3F,$23,$0E,$99,$7C,$56,$C7,$B5,$B3,$EA,$BE,$78,$FC,$DC,$7E
	.dc.b	$53,$E4,$BF,$3C,$EB,$0F,$91,$5A,$7B,$46,$67,$E3,$CE,$5F,$F6,$CD
	.dc.b	$7F,$51,$7F,$DC,$38,$FF,$5D,$6C,$65,$FE,$C9,$DA,$F6,$BF,$D9,$CD
	.dc.b	$8F,$F7,$1B,$5E,$6D,$78,$C7,$F5,$78,$E1,$97,$F4,$16,$E1,$FD,$AF
	.dc.b	$DB,$7B,$7F,$EF,$8A,$D5,$F1,$26,$1F,$D2,$EB,$6E,$27,$6F,$F5,$DC
	.dc.b	$3A,$BF,$CA,$E7,$6E,$6B,$86,$1E,$68,$E9,$87,$86,$31,$87,$42,$3E
	.dc.b	$78,$D4,$FF,$81,$B0,$EE,$FF,$80,$5B,$97,$F9,$B7,$0F,$ED,$B8,$F8
	.dc.b	$B1,$E7,$7C,$67,$C6,$F9,$F0,$EA,$F3,$8C,$B4,$F3,$98,$BE,$CF,$B4
	.dc.b	$C9,$6B,$BB,$C5,$FF,$07,$5F,$8F,$1F,$79,$27,$C8,$72,$9E,$35,$D9
	.dc.b	$F4,$EE,$BE,$55,$C2,$77,$97,$D2,$AD,$65,$EC,$C3,$96,$3D,$F0,$F2
	.dc.b	$3F,$35,$F6,$5F,$CA,$39,$63,$DA,$36,$77,$9F,$46,$1D,$A3,$8E,$5F
	.dc.b	$32,$31,$1F,$0E,$7C,$18,$78,$A3,$E8,$1C,$0D,$18,$29,$E0,$3C,$33
	.dc.b	$D3,$2F,$7D,$F7,$C3,$CD,$7D,$6F,$83,$DE,$F8,$61,$E2,$C7,$DD,$87
	.dc.b	$E1,$B2,$F8,$D1,$8C,$3C,$77,$C1,$F8,$04,$3C,$78,$F8,$27,$86,$5F
	.dc.b	$5A,$31,$87,$B6,$3F,$03,$87,$68,$C6,$3F,$47,$0F,$7E,$5E,$93,$D1
	.dc.b	$87,$28,$CE,$3F,$53,$0E,$51,$8C,$3D,$4F,$43,$CF,$97,$A5,$F8,$75
	.dc.b	$E6,$C7,$C1,$97,$CA,$9F,$7E,$3D,$E3,$07,$A2,$3A,$1F,$03,$1F,$0C
	.dc.b	$6A,$B9,$9F,$37,$1F,$A9,$3A,$8C,$57,$E7,$5B,$5F,$1E,$75,$3F,$37
	.dc.b	$2F,$83,$18,$8F,$4D,$FB,$9F,$80,$7A,$23,$53,$F2,$71,$F8,$D3,$E8
	.dc.b	$8E,$6D,$8C,$BC,$AF,$C4,$E6,$FB,$47,$B6,$3E,$5E,$3F,$17,$1F,$49
	.dc.b	$F0,$23,$D1,$7F,$35,$FC,$D5,$71,$7E,$41,$D4,$F8,$47,$0B,$F8,$EE
	.dc.b	$63,$E8,$3E,$5A,$FD,$7D,$F8,$63,$81,$C8,$F8,$A7,$C3,$39,$47,$95
	.dc.b	$CC,$FB,$9E,$87,$B3,$1F,$93,$87,$A7,$2F,$C4,$65,$F2,$F2,$F2,$DF
	.dc.b	$03,$D6,$7F,$AB,$34,$7A,$EF,$5F,$F9,$BF,$17,$FB,$87,$D8,$BC,$E7
	.dc.b	$F9,$DD,$F0,$AB,$7C,$70,$FF,$69,$B5,$FF,$69,$B6,$7F,$DC,$2D,$80
	.dc.b	$8E,$8F,$29,$FF,$73,$BA,$8C,$BD,$23,$2F,$FB,$EF,$C6,$F3,$FA,$B2
	.dc.b	$3F,$DA,$DC,$8E,$C5,$7F,$54,$8C,$3C,$9B,$3D,$CF,$2B,$87,$CC,$62
	.dc.b	$79,$BB,$3C,$45,$8E,$18,$7B,$A3,$27,$22,$B9,$87,$AA,$C3,$85,$66
	.dc.b	$30,$E0,$4F,$09,$C4,$62,$76,$EF,$77,$AB,$A8,$FF,$79,$36,$33,$1C
	.dc.b	$9A,$F1,$8D,$11,$D3,$E0,$F1,$1F,$02,$FA,$2C,$36,$BD,$13,$A9,$C9
	.dc.b	$E3,$F8,$3F,$24,$9A,$47,$DB,$9B,$CB,$7C,$26,$1B,$3F,$16,$75,$39
	.dc.b	$24,$66,$76,$FC,$E9,$DB,$99,$F9,$7C,$AF,$35,$EC,$E1,$B5,$F4,$E7
	.dc.b	$53,$92,$46,$67,$6F,$DC,$9D,$B9,$9F,$B3,$E8,$3D,$87,$AF,$86,$D7
	.dc.b	$E0,$67,$53,$92,$46,$67,$6F,$E2,$A7,$6E,$67,$F0,$5E,$C7,$E6,$A3
	.dc.b	$F1,$8D,$7F,$22,$D7,$F3,$4D,$53,$E7,$F8,$5F,$31,$23,$C4,$D7,$C5
	.dc.b	$68,$C4,$7E,$96,$79,$3F,$9A,$F6,$1E,$CF,$D1,$8F,$D0,$46,$34,$D4
	.dc.b	$E6,$7F,$4F,$39,$77,$EC,$F9,$9B,$7D,$63,$11,$C1,$AF,$EB,$7C,$4F
	.dc.b	$C9,$91,$FD,$43,$FB,$1F,$86,$5B,$D1,$FE,$0B,$97,$F0,$69,$CC,$63
	.dc.b	$B3,$F8,$FC,$BD,$8F,$E6,$11,$83,$F9,$1D,$70,$71,$8A,$30,$FE,$4E
	.dc.b	$37,$87,$85,$DD,$F5,$F0,$3C,$15,$C6,$B4,$F1,$3F,$84,$FC,$3B,$C9
	.dc.b	$FE,$29,$1F,$C8,$23,$18,$7E,$AD,$B1,$FA,$76,$BF,$C7,$71,$7C,$47
	.dc.b	$F2,$9C,$BE,$13,$6F,$A6,$1E,$4A,$DD,$DA,$D3,$EF,$6F,$43,$1F,$0E
	.dc.b	$3D,$2B,$DB,$CF,$FB,$5F,$89,$B2,$7E,$0F,$6B,$E0,$FF,$A0,$EB,$F3
	.dc.b	$2F,$83,$6F,$83,$6F,$8D,$70,$C3,$E4,$5F,$06,$FC,$57,$FB,$06,$FA
	.dc.b	$36,$3E,$76,$AF,$89,$95,$AF,$2F,$D3,$7B,$CF,$FA,$6E,$28,$C5,$7E
	.dc.b	$12,$F8,$E3,$E0,$C3,$DB,$0F,$47,$EC,$CB,$5F,$04,$61,$B7,$FC,$CA
	.dc.b	$B8,$3F,$C6,$7E,$2D,$F1,$9F,$31,$B1,$F9,$2A,$E3,$87,$C2,$9C,$BF
	.dc.b	$8B,$8F,$76,$FD,$F0,$F9,$73,$FD,$6A,$73,$7E,$1A,$D1,$1C,$12,$7F
	.dc.b	$93,$70,$FF,$84,$63,$EC,$36,$3C,$17,$DB,$6F,$93,$39,$7F,$BB,$E0
	.dc.b	$F8,$D7,$C7,$7C,$36,$E3,$FD,$51,$B8,$F9,$DA,$FF,$79,$AE,$55,$C4
	.dc.b	$95,$C5,$FA,$3C,$0F,$43,$F8,$CC,$FE,$3D,$B7,$F1,$F0,$EF,$8F,$E0
	.dc.b	$27,$2F,$A3,$CB,$7C,$7F,$D9,$6E,$1F,$DA,$9B,$87,$CD,$6B,$E7,$6B
	.dc.b	$FE,$2E,$CF,$FC,$1B,$5C,$C9,$5F,$D6,$7E,$2F,$F8,$9D,$7E,$79,$B8
	.dc.b	$78,$EB,$F5,$F1,$8A,$E9,$F2,$1F,$4E,$FE,$33,$5C,$FA,$7F,$91,$8C
	.dc.b	$47,$E5,$5B,$89,$A4,$9F,$97,$E4,$BE,$47,$F7,$ED,$C3,$F2,$95,$F7
	.dc.b	$DF,$F2,$5E,$FF,$E0,$4A,$FE,$03,$F2,$4F,$93,$FD,$13,$6B,$FC,$26
	.dc.b	$DC,$FF,$2C,$D7,$F0,$0D,$7E,$53,$5F,$D7,$37,$33,$9C,$1F,$28,$F9
	.dc.b	$4F,$E2,$ED,$C3,$F5,$95,$E1,$E0,$FE,$03,$39,$C7,$F9,$B7,$5F,$EF
	.dc.b	$5F,$DE,$B6,$FF,$C8,$4D,$8F,$F2,$3B,$5F,$F2,$83,$5F,$F3,$7B,$53
	.dc.b	$49,$F2,$AF,$87,$FE,$93,$08,$C5,$7F,$54,$8F,$06,$1F,$E6,$47,$55
	.dc.b	$C3,$E5,$BF,$3B,$5E,$06,$E1,$E0,$F1,$5F,$11,$BE,$AF,$E6,$FF,$11
	.dc.b	$F8,$42,$73,$7F,$19,$7F,$13,$C2,$F9,$7F,$E4,$B8,$A3,$1A,$FA,$37
	.dc.b	$7C,$87,$F4,$6B,$DD,$FB,$7B,$FF,$A2,$FC,$9B,$E5,$FF,$CF,$BE,$7F
	.dc.b	$F6,$3A,$FF,$4A,$F1,$FF,$30,$E4,$FE,$17,$7B,$E4,$FF,$4C,$DE,$EF
	.dc.b	$A4,$7F,$27,$9D,$42,$75,$7C,$EB,$F6,$95,$A9,$F0,$74,$23,$1E,$66
	.dc.b	$B2,$F2,$BF,$E8,$BC,$39,$63,$FC,$FE,$B6,$FF,$41,$F9,$9F,$E6,$38
	.dc.b	$FA,$F8,$5F,$34,$F9,$AE,$77,$BF,$1E,$D6,$A7,$3A,$B1,$E7,$A7,$EB
	.dc.b	$B1,$E5,$B7,$C2,$8F,$E1,$D3,$AC,$BE,$A5,$69,$F8,$53,$FC,$4F,$E6
	.dc.b	$FF,$D3,$37,$C6,$7F,$AC,$46,$1D,$98,$3C,$6E,$74,$F8,$ED,$AE,$9E
	.dc.b	$46,$AB,$6F,$BB,$99,$F3,$AF,$9E,$F6,$FC,$F3,$E3,$75,$39,$D3,$C9
	.dc.b	$BA,$B5,$1E,$7C,$39,$E1,$CB,$0E,$2D,$9E,$35,$C3,$2F,$5E,$1C,$EF
	.dc.b	$75,$A7,$E7,$FC,$FB,$E5,$3E,$9F,$37,$F5,$FE,$0F,$C2,$CE,$6F,$E4
	.dc.b	$72,$7E,$DF,$BB,$8E,$1C,$1B,$3A,$C3,$97,$0F,$E4,$3C,$9F,$BB,$D1
	.dc.b	$FE,$AA,$AF,$7B,$FE,$F3,$6C,$68,$8E,$8B,$B7,$D6,$3F,$84,$7D,$0F
	.dc.b	$E2,$3E,$87,$FA,$FE,$DF,$A0,$8C,$7B,$DF,$90,$AD,$DF,$CE,$BD,$E5
	.dc.b	$F9,$DB,$DC,$FE,$16,$B6,$7C,$AA,$DB,$FB,$69,$DB,$F3,$BE,$41,$F4
	.dc.b	$5F,$9A,$6D,$FD,$C7,$FD,$AC,$D8,$D1,$1D,$7C,$45,$AF,$B3,$F4,$7F
	.dc.b	$E2,$86,$FB,$1F,$A6,$87,$C9,$58,$97,$F8,$A9,$E1,$B7,$EF,$9D,$C6
	.dc.b	$67,$F2,$DF,$4B,$E6,$C3,$F0,$FF,$4B,$FC,$C1,$FD,$1F,$62,$75,$8F
	.dc.b	$F0,$8A,$D3,$FC,$66,$75,$A7,$EE,$A7,$3C,$1F,$D0,$AF,$8B,$FE,$05
	.dc.b	$8C,$E1,$FD,$12,$72,$FF,$92,$E3,$37,$D3,$2F,$D2,$C7,$0E,$A7,$C0
	.dc.b	$FF,$3C,$9F,$D4,$B6,$FF,$42,$BE,$9B,$F8,$6B,$9D,$3F,$90,$B7,$3F
	.dc.b	$F0,$C3,$6B,$FC,$B8,$F1,$BE,$BE,$77,$F1,$6F,$A8,$7C,$F7,$F1,$8F
	.dc.b	$A8,$FF,$39,$65,$FC,$B2,$73,$7F,$BC,$9C,$E5,$FE,$25,$9C,$E5,$FB
	.dc.b	$69,$CB,$FE,$6B,$8C,$DF,$F3,$69,$CC,$BE,$4F,$F8,$57,$8B,$F9,$76
	.dc.b	$C7,$F1,$AF,$98,$7C,$D7,$E0,$B2,$4C,$AD,$E3,$FD,$2D,$B3,$A9,$ED
	.dc.b	$E2,$31,$FE,$51,$8F,$4E,$6F,$D6,$61,$C2,$FA,$5E,$DF,$DB,$46,$6B
	.dc.b	$86,$BE,$7A,$D6,$5E,$16,$E5,$E1,$BE,$19,$FA,$71,$EB,$9F,$12,$3E
	.dc.b	$F8,$D3,$E8,$8C,$E1,$D5,$F4,$CE,$AB,$D1,$5B,$3C,$F8,$78,$96,$5E
	.dc.b	$87,$37,$D3,$1F,$35,$6E,$B5,$8F,$7B,$D9,$1F,$EF,$D1,$AA,$E3,$7C
	.dc.b	$EB,$DA,$7A,$F3,$F9,$B8,$78,$E3,$CD,$87,$53,$9E,$5E,$B9,$D6,$5E
	.dc.b	$C3,$E4,$DF,$2C,$3B,$61,$CB,$1F,$C1,$E5,$E3,$8F,$0E,$5D,$67,$CB
	.dc.b	$1E,$3C,$F8,$E3,$D7,$2F,$81,$9F,$87,$2C,$C6,$33,$F0,$61,$F0,$71
	.dc.b	$E7,$97,$A2,$72,$F7,$9F,$0C,$F9,$30,$E9,$1C,$4C,$9F,$51,$CC,$F1
	.dc.b	$8F,$61,$9A,$EB,$5E,$02,$46,$72,$F3,$E3,$F4,$73,$F8,$53,$C8,$99
	.dc.b	$FC,$AA,$DC,$78,$E7,$73,$F0,$EB,$F5,$13,$F2,$0B,$1F,$B4,$6E,$7F
	.dc.b	$A5,$73,$D1,$FE,$64,$78,$9F,$BD,$70,$EB,$97,$2E,$5B,$5F,$D0,$9F
	.dc.b	$E0,$53,$CF,$4F,$DE,$37,$3F,$BF,$F3,$4E,$BF,$DF,$56,$AF,$89,$31
	.dc.b	$FE,$15,$5B,$71,$3B,$7F,$7B,$87,$57,$F4,$F3,$B7,$35,$C3,$0F,$DE
	.dc.b	$56,$F4,$5E,$F0,$E8,$4A,$DD,$FF,$10,$9D,$46,$30,$EA,$4A,$FD,$6C
	.dc.b	$EA,$3F,$8F,$61,$E5,$7E,$41,$BA,$F8,$78,$FC,$27,$A5,$F8,$67,$8E
	.dc.b	$1E,$1A,$D4,$60,$E4,$59,$F0,$C6,$A3,$BC,$73,$6D,$78,$1F,$88,$71
	.dc.b	$7C,$26,$5E,$A7,$03,$47,$CE,$9E,$D9,$7E,$06,$36,$DA,$F1,$C7,$63
	.dc.b	$91,$CF,$2F,$A3,$5C,$B1,$F8,$D7,$C7,$0F,$1D,$F7,$39,$46,$5F,$1D
	.dc.b	$F2,$73,$5C,$8F,$BD,$97,$D3,$8F,$0D,$71,$7C,$51,$98,$C1,$27,$8E
	.dc.b	$1E,$5C,$BA,$9F,$4E,$E9,$F4,$B0,$F1,$3F,$0A,$BC,$53,$B8,$C3,$F4
	.dc.b	$27,$79,$7D,$5C,$3C,$13,$B8,$FF,$02,$FD,$4F,$E0,$39,$FF,$C1,$FE
	.dc.b	$8D,$E7,$36,$66,$5F,$E1,$28,$D1,$87,$FC,$95,$F1,$BC,$FC,$03,$4C
	.dc.b	$7D,$0D,$9F,$F2,$B4,$57,$FC,$A5,$1F,$92,$CD,$F5,$5F,$9F,$CD,$DB
	.dc.b	$FC,$4B,$5F,$F1,$B6,$1B,$AF,$D2,$5F,$0E,$1F,$65,$EF,$8F,$FC,$69
	.dc.b	$7B,$98,$FF,$8C,$0D,$98,$3F,$D1,$BF,$57,$FC,$6F,$1F,$F6,$5B,$6B
	.dc.b	$FC,$1F,$92,$31,$48,$C3,$F3,$63,$31,$BE,$6C,$19,$9F,$A5,$E3,$7F
	.dc.b	$AA,$DF,$F4,$4F,$9B,$83,$3F,$35,$63,$24,$3F,$D4,$72,$8F,$D4,$FC
	.dc.b	$17,$D3,$FA,$C9,$5F,$E2,$B8,$CB,$FE,$87,$8C,$BF,$E1,$C8,$CB,$FE
	.dc.b	$85,$8C,$E9,$FE,$2B,$9F,$C4,$79,$EE,$2A,$4B,$FF,$5F,$C7,$E0,$9B
	.dc.b	$19,$7F,$D7,$31,$C5,$DF,$1F,$E4,$FD,$0F,$F4,$0C,$7F,$AB,$FE,$73
	.dc.b	$F8,$AF,$A2,$F3,$D1,$BE,$E9,$5B,$21,$FE,$C6,$BE,$6F,$FB,$23,$D3
	.dc.b	$76,$7E,$85,$B1,$FE,$D6,$6B,$FE,$86,$FA,$D7,$CF,$FE,$55,$B1,$BA
	.dc.b	$E4,$66,$64,$64,$87,$37,$F6,$1E,$BB,$EB,$A9,$27,$FD,$CF,$0E,$1C
	.dc.b	$47,$FB,$82,$38,$3A,$9E,$73,$FB,$BF,$80,$7C,$47,$C8,$F6,$BE,$75
	.dc.b	$B9,$5D,$BF,$E6,$8E,$35,$6A,$73,$5F,$9B,$C3,$9B,$FE,$FA,$6B,$CE
	.dc.b	$F9,$57,$16,$F3,$78,$F1,$FC,$EF,$D5,$CB,$FF,$72,$BF,$CF,$BE,$AF
	.dc.b	$FC,$E9,$F0,$0F,$33,$FC,$CD,$18,$7F,$5F,$DA,$AE,$57,$FE,$72,$AD
	.dc.b	$3F,$E8,$16,$F8,$0A,$E1,$3E,$03,$FD,$19,$3C,$37,$FF,$45,$FA,$37
	.dc.b	$CF,$FF,$50,$D7,$FA,$33,$3F,$22,$AD,$B9,$CF,$FD,$23,$E3,$3E,$1B
	.dc.b	$E2,$DF,$0D,$1F,$5F,$FB,$5C,$3F,$C4,$71,$8F,$0F,$FC,$42,$D8,$CC
	.dc.b	$ED,$FF,$18,$E0,$F8,$7F,$CA,$E1,$E4,$8C,$75,$FC,$26,$D7,$94,$2F
	.dc.b	$49,$F6,$1B,$D3,$FD,$FB,$5B,$AF,$57,$B3,$93,$A7,$D8,$FF,$90,$B2
	.dc.b	$F0,$3F,$E3,$8A,$EA,$DA,$34,$91,$FE,$3E,$8C,$B6,$32,$FF,$91,$B8
	.dc.b	$1F,$14,$F9,$DF,$AC,$D9,$F6,$4E,$A7,$EF,$F4,$3D,$EA,$7C,$C7,$15
	.dc.b	$F4,$1A,$FF,$97,$38,$BF,$4F,$3B,$24,$6D,$AF,$F9,$8A,$DF,$02,$FA
	.dc.b	$FF,$C5,$36,$7C,$AD,$5A,$FE,$0F,$EC,$BE,$E3,$79,$BF,$03,$1F,$7E
	.dc.b	$7F,$DA,$D9,$23,$11,$FD,$41,$AF,$F4,$F6,$C6,$E7,$99,$27,$E6,$7C
	.dc.b	$8F,$83,$A3,$F8,$03,$63,$F3,$8D,$A7,$C9,$BE,$8B,$F5,$6D,$AF,$CC
	.dc.b	$B5,$F9,$4D,$7D,$2D,$7F,$B3,$B5,$FE,$CC,$D9,$EF,$3F,$A2,$8D,$6A
	.dc.b	$F9,$5F,$E2,$77,$F2,$B1,$FE,$31,$18,$D3,$F6,$B5,$BF,$0B,$F8,$0F
	.dc.b	$CB,$3E,$CE,$9F,$AE,$71,$8F,$EC,$9A,$FE,$0D,$AF,$D2,$6B,$FE,$00
	.dc.b	$6B,$FD,$FD,$BA,$1C,$DB,$D6,$FE,$AF,$5F,$98,$6B,$EB,$3F,$86,$38
	.dc.b	$C7,$F8,$47,$CE,$FF,$8C,$A7,$F7,$58,$23,$11,$F9,$96,$BF,$88,$6B
	.dc.b	$FE,$16,$6B,$FE,$15,$6C,$73,$F7,$7F,$88,$C6,$0D,$E0,$F9,$97,$DA
	.dc.b	$7F,$27,$6E,$1F,$B2,$6B,$EB,$C3,$CD,$E9,$FF,$32,$C0,$F6,$24,$FF
	.dc.b	$22,$8C,$BF,$C0,$7E,$0B,$F9,$8B,$F1,$71,$46,$33,$FB,$ED,$E1,$53
	.dc.b	$E7,$5F,$6B,$FD,$A6,$3F,$83,$D6,$DF,$E7,$4D,$C3,$AF,$93,$FD,$2A
	.dc.b	$F8,$FA,$3F,$D1,$72,$FD,$FE,$1F,$C4,$1A,$F9,$5A,$7F,$3F,$FB,$67
	.dc.b	$CA,$F1,$ED,$95,$C5,$FE,$AB,$3A,$8F,$E6,$D8,$3E,$DD,$F3,$5E,$36
	.dc.b	$CE,$9A,$FE,$24,$FD,$B7,$DC,$EE,$64,$6D,$FE,$CA,$E8,$FE,$88,$53
	.dc.b	$FD,$EB,$DA,$9F,$E7,$8D,$8F,$F7,$AB,$5F,$E8,$5F,$2D,$FE,$E3,$D3
	.dc.b	$FB,$28,$71,$C5,$FD,$A5,$B3,$F5,$BE,$9F,$F5,$F5,$C2,$72,$F1,$F8
	.dc.b	$6F,$EF,$6D,$AF,$3D,$70,$BF,$BF,$1F,$D8,$FE,$57,$F9,$6B,$83,$F0
	.dc.b	$7E,$8A,$33,$19,$8C,$5E,$E3,$73,$A7,$AC,$FF,$A6,$BE,$9F,$51,$FD
	.dc.b	$C7,$E1,$11,$89,$D4,$F3,$6B,$C3,$DC,$FE,$CC,$FF,$8C,$4F,$ED,$0E
	.dc.b	$0F,$16,$0F,$A6,$7C,$6F,$F8,$42,$BF,$C4,$B1,$8C,$F6,$D8,$FE,$E8
	.dc.b	$D7,$73,$C5,$FF,$0D,$46,$5B,$19,$7F,$AA,$C6,$5F,$D5,$FD,$CB,$EC
	.dc.b	$B5,$08,$C5,$FC,$46,$C7,$CB,$6B,$F5,$BC,$4F,$E3,$1E,$3F,$76,$D7
	.dc.b	$D1,$6C,$FD,$96,$BC,$7C,$0F,$F2,$BC,$ED,$B1,$B6,$BB,$AF,$67,$84
	.dc.b	$FA,$7F,$F3,$CD,$FF,$8C,$1B,$5F,$E0,$16,$BF,$DC,$1A,$FF,$AF,$9A
	.dc.b	$FF,$AF,$5B,$3D,$7E,$13,$BC,$6B,$AF,$FD,$D3,$E3,$FC,$D6,$E1,$F8
	.dc.b	$F6,$E1,$F9,$76,$BF,$9C,$6B,$FC,$71,$AF,$F1,$D9,$FF,$7A,$CF,$36
	.dc.b	$C7,$3F,$6F,$F7,$53,$DD,$B1,$9E,$BF,$24,$62,$3F,$C1,$F5,$FE,$45
	.dc.b	$AF,$F2,$AB,$63,$FC,$4C,$D7,$7E,$D7,$B7,$E6,$79,$47,$9D,$AF,$F3
	.dc.b	$36,$E2,$69,$31,$E4,$4C,$3C,$3F,$3D,$F6,$B1,$34,$91,$F2,$FE,$97
	.dc.b	$FD,$76,$DF,$7B,$E8,$46,$32,$F9,$CD,$7E,$73,$5F,$9C,$D7,$FA,$53
	.dc.b	$72,$39,$DF,$FB,$F7,$83,$F8,$EF,$3B,$F9,$BE,$3F,$E0,$B8,$C1,$FA
	.dc.b	$DF,$6F,$FA,$A7,$A9,$71,$FF,$7B,$57,$F1,$19,$CB,$63,$39,$7F,$A0
	.dc.b	$1A,$FF,$20,$9D,$B9,$AE,$87,$E8,$3D,$3F,$3D,$FF,$8D,$B3,$C1,$1F
	.dc.b	$07,$DD,$7D,$6C,$0D,$27,$07,$A5,$B3,$C2,$30,$7E,$BB,$D9,$7C,$F5
	.dc.b	$24,$FF,$9B,$DF,$85,$F7,$7F,$E0,$28,$F8,$0D,$D3,$FD,$3A,$D7,$DC
	.dc.b	$D7,$FD,$5A,$D7,$FD,$36,$DC,$4D,$35,$C0,$FA,$57,$D3,$FF,$04,$6E
	.dc.b	$1F,$EB,$26,$E1,$FE,$83,$6B,$F3,$FE,$F1,$F0,$5F,$6D,$B5,$F6,$5B
	.dc.b	$3F,$EB,$66,$BC,$67,$81,$27,$83,$F8,$5E,$9F,$F0,$CC,$7E,$0D,$B3
	.dc.b	$FE,$C0,$6B,$F8,$16,$BF,$ED,$0F,$19,$F3,$FF,$8F,$6D,$7E,$35,$B3
	.dc.b	$FE,$CA,$6B,$F8,$BF,$84,$FC,$CB,$FE,$CE,$6B,$9E,$7F,$CD,$46,$23
	.dc.b	$FD,$BF,$F4,$EF,$A7,$FC,$F3,$70,$FC,$E3,$67,$FD,$B2,$D7,$FB,$B7
	.dc.b	$CA,$FF,$18,$6F,$77,$F4,$E8,$C4,$7F,$7E,$6B,$FE,$3E,$6A,$69,$23
	.dc.b	$91,$FE,$14,$AF,$F2,$2F,$0B,$E2,$FF,$C6,$B1,$FE,$48,$F3,$3F,$D1
	.dc.b	$B7,$FC,$AB,$CC,$FF,$7C,$56,$B1,$53,$C6,$E7,$F2,$3F,$C5,$31,$FA
	.dc.b	$B9,$23,$17,$FD,$E6,$31,$E1,$AB,$75,$A7,$FD,$7E,$DA,$EB,$3B,$79
	.dc.b	$46,$55,$74,$AF,$0F,$C9,$FF,$A3,$DF,$66,$1F,$02,$31,$5E,$A8,$C3
	.dc.b	$D7,$1F,$63,$D2,$77,$18,$B4,$62,$BA,$79,$5C,$5E,$78,$78,$30,$E5
	.dc.b	$39,$5D,$7F,$E4,$5F,$98,$F6,$57,$F1,$BF,$55,$18,$AF,$F3,$DD,$6E
	.dc.b	$BF,$42,$D8,$FF,$12,$36,$39,$4F,$19,$EF,$5D,$27,$8B,$63,$8B,$73
	.dc.b	$F7,$7C,$03,$E7,$FD,$18,$F7,$9C,$C7,$AF,$57,$CC,$3E,$97,$C1,$18
	.dc.b	$6C,$67,$3F,$C2,$B7,$3F,$C2,$36,$3F,$DF,$ED,$8F,$CB,$B5,$FB,$ED
	.dc.b	$AF,$CD,$7B,$BE,$3E,$33,$6F,$0C,$7E,$A4,$F0,$6C,$FE,$4D,$B3,$F5
	.dc.b	$EF,$F5,$9F,$4B,$F2,$DB,$3F,$C4,$39,$93,$96,$C6,$63,$04,$70,$49
	.dc.b	$FC,$91,$F9,$AF,$9F,$FF,$3B,$47,$E2,$63,$B9,$AC,$53,$A8,$C7,$B9
	.dc.b	$FC,$C3,$6F,$CB,$37,$93,$F7,$EB,$E3,$7E,$C3,$9A,$FD,$4B,$70,$34
	.dc.b	$E1,$8F,$C8,$75,$E8,$FF,$88,$F1,$FD,$23,$71,$FE,$1A,$BE,$6B,$F0
	.dc.b	$CE,$71,$FD,$BB,$71,$FF,$11,$78,$7A,$AD,$BF,$90,$FA,$5F,$E9,$79
	.dc.b	$FF,$7F,$6E,$1F,$A4,$6B,$FA,$C6,$F2,$9A,$4F,$17,$F4,$91,$96,$C7
	.dc.b	$C1,$F9,$FF,$E5,$57,$FB,$9A,$D3,$F8,$88,$CD,$FD,$3F,$73,$FA,$5E
	.dc.b	$DF,$C1,$5B,$B7,$F2,$E5,$F4,$BF,$A7,$73,$5F,$C6,$1B,$B7,$FA,$7B
	.dc.b	$C8,$D5,$6D,$FD,$77,$D2,$BE,$23,$F6,$1F,$4B,$FC,$2B,$5F,$CD,$CE
	.dc.b	$5F,$CE,$C6,$72,$FC,$44,$E5,$FC,$54,$67,$4F,$E3,$93,$97,$F8,$F4
	.dc.b	$67,$4F,$F6,$6F,$D2,$7F,$03,$F0,$BF,$28,$DC,$3F,$33,$F1,$8F,$99
	.dc.b	$FB,$D0,$9D,$63,$FC,$B2,$31,$E0,$B0,$E6,$D8,$E7,$7C,$AF,$8E,$3E
	.dc.b	$2A,$D6,$1C,$E7,$36,$C7,$C5,$CD,$FB,$99,$CC,$F8,$32,$EE,$7F,$35
	.dc.b	$AD,$EB,$89,$FD,$99,$B2,$D7,$69,$DC,$63,$1F,$1E,$5E,$5C,$39,$65
	.dc.b	$FD,$2A,$F6,$F9,$EB,$76,$3B,$46,$6B,$C9,$9F,$C1,$AD,$D6,$B0,$EC
	.dc.b	$FC,$6C,$FD,$D3,$AC,$BD,$E7,$9E,$F9,$61,$E4,$C3,$8E,$3E,$59,$CE
	.dc.b	$3E,$BA,$D3,$F2,$E3,$33,$E5,$8D,$1E,$1C,$F6,$F9,$4F,$66,$1E,$1C
	.dc.b	$BC,$B8,$6A,$31,$57,$2F,$41,$F1,$33,$F8,$F9,$F9,$89,$8F,$49,$D6
	.dc.b	$7E,$49,$CC,$F8,$6B,$8B,$BC,$7B,$1F,$33,$2F,$1C,$7A,$B3,$F6,$1F
	.dc.b	$0E,$32,$E2,$FC,$37,$D2,$BD,$C4,$CF,$D6,$F3,$9F,$1C,$62,$B9,$4F
	.dc.b	$8C,$D1,$F3,$6F,$BD,$70,$CB,$DB,$18,$9E,$07,$E1,$CA,$7E,$08,$C1
	.dc.b	$F9,$72,$DF,$F3,$1E,$4C,$C7,$D4,$E7,$FF,$08,$B8,$9D,$36,$32,$F2
	.dc.b	$77,$8F,$EB,$DB,$8F,$F0,$46,$CF,$EC,$7E,$97,$FA,$4B,$FC,$1A,$32
	.dc.b	$FF,$03,$6D,$F8,$2F,$F4,$D5,$C2,$BF,$5F,$5A,$26,$1F,$B7,$9D,$57
	.dc.b	$8E,$3B,$12,$BF,$89,$CE,$B0,$F3,$4E,$6B,$85,$6B,$AB,$E6,$46,$09
	.dc.b	$5C,$27,$47,$17,$11,$A8,$F8,$13,$CA,$72,$48,$FA,$15,$E9,$9D,$46
	.dc.b	$0F,$7C,$75,$79,$94,$F9,$0E,$74,$61,$ED,$9E,$58,$78,$DD,$9E,$83
	.dc.b	$A9,$1F,$44,$F2,$BF,$1C,$F0,$9E,$59,$7D,$5C,$7F,$8B,$5E,$EB,$C9
	.dc.b	$39,$7C,$33,$AC,$3C,$91,$CE,$3E,$13,$4F,$69,$E3,$73,$3C,$EB,$B4
	.dc.b	$E4,$EA,$47,$37,$E6,$CB,$F0,$B3,$EF,$8C,$65,$F1,$F1,$F1,$E1,$E7
	.dc.b	$AF,$11,$E3,$C3,$9E,$3E,$2C,$79,$E3,$F9,$4C,$BE,$2E,$1C,$49,$3F
	.dc.b	$A0,$BE,$38,$79,$8E,$B3,$EB,$C7,$C3,$5D,$31,$FC,$25,$F0,$E5,$73
	.dc.b	$3D,$DF,$C2,$C6,$3C,$BB,$26,$CC,$CF,$FC,$B7,$39,$6C,$E5,$FF,$24
	.dc.b	$77,$3E,$1B,$AB,$63,$FC,$C6,$D7,$7F,$3B,$A9,$91,$A7,$8C,$7E,$6B
	.dc.b	$85,$F4,$BF,$93,$CD,$DB,$FC,$53,$8F,$FD,$50,$DD,$FF,$D3,$75,$F9
	.dc.b	$9B,$E3,$C7,$FC,$77,$A3,$F8,$F4,$63,$4F,$F4,$FD,$7F,$05,$9C,$B6
	.dc.b	$39,$D7,$16,$9C,$1C,$3F,$EB,$08,$E0,$FF,$B2,$3E,$97,$FC,$20,$FF
	.dc.b	$9B,$5B,$97,$FA,$FD,$B8,$FF,$BA,$1F,$54,$F1,$E8,$FA,$F1,$A8,$C1
	.dc.b	$FE,$8E,$77,$A7,$FB,$BF,$CB,$FB,$BE,$3F,$FA,$FA,$BF,$D2,$31,$8F
	.dc.b	$7F,$2C,$8C,$9F,$E9,$68,$FC,$9F,$C8,$7D,$AF,$A5,$93,$FE,$34,$8C
	.dc.b	$BF,$E0,$DD,$FF,$A1,$C7,$C2,$8F,$F1,$A3,$7A,$CD,$24,$71,$3F,$C8
	.dc.b	$2E,$1F,$F5,$BF,$1F,$E5,$BC,$B3,$4C,$72,$D9,$FF,$57,$CE,$9B,$3A
	.dc.b	$7F,$D6,$7E,$6F,$FA,$1D,$F0,$1F,$EF,$C7,$0F,$FC,$01,$31,$D9,$F7
	.dc.b	$37,$4E,$B3,$FE,$CD,$18,$9D,$4E,$6B,$FC,$61,$39,$7B,$B5,$FF,$75
	.dc.b	$FD,$23,$E1,$FF,$67,$E8,$FE,$D6,$7F,$D9,$2D,$7F,$D7,$3D,$2A,$E1
	.dc.b	$1F,$E8,$96,$BD,$2B,$F2,$73,$B6,$CE,$B2,$FD,$0B,$6B,$FE,$00,$B5
	.dc.b	$57,$17,$E3,$B5,$F8,$D8,$3D,$27,$D2,$7F,$0C,$6C,$FF,$B8,$9A,$FF
	.dc.b	$B3,$5F,$C1,$B9,$8E,$04,$78,$70,$7F,$16,$FA,$47,$C0,$FD,$36,$CF
	.dc.b	$CC,$6B,$E4,$6E,$7F,$59,$B1,$F5,$1A,$B1,$46,$1B,$5F,$99,$FA,$2F
	.dc.b	$F1,$27,$B0,$C3,$EE,$F4,$BE,$8F,$F5,$E7,$F3,$AA,$DC,$FF,$A7,$1F
	.dc.b	$F1,$2E,$5F,$E5,$86,$BF,$E9,$96,$F3,$9A,$48,$E3,$19,$6C,$74,$DD
	.dc.b	$EE,$7C,$29,$FE,$3A,$D7,$FD,$05,$1A,$CF,$84,$E6,$7E,$FF,$BD,$F8
	.dc.b	$F7,$F9,$2B,$5E,$2D,$77,$E3,$F8,$7C,$A7,$D1,$D2,$4F,$F9,$13,$14
	.dc.b	$EA,$31,$F4,$7F,$0A,$4D,$24,$7B,$7E,$F5,$F0,$34,$91,$D6,$32,$D8
	.dc.b	$CB,$FE,$58,$A4,$EA,$30,$D7,$F7,$7A,$3D,$9F,$AF,$C9,$F9,$06,$C6
	.dc.b	$67,$FC,$BF,$39,$7D,$DE,$DF,$FB,$A5,$BE,$AF,$D5,$8C,$47,$8D,$AF
	.dc.b	$F5,$26,$BF,$59,$AF,$EC,$9B,$89,$CD,$3E,$85,$E2,$FA,$1B,$7E,$66
	.dc.b	$BF,$E7,$A9,$E1,$3F,$89,$FA,$17,$CC,$52,$57,$D9,$8C,$B6,$32,$FC
	.dc.b	$D8,$CB,$63,$E4,$FC,$6E,$23,$F9,$3B,$5F,$E4,$AD,$7F,$64,$DF,$8E
	.dc.b	$7D,$9F,$F6,$89,$FF,$4A,$46,$3C,$DF,$D5,$B6,$3F,$D2,$2D,$7F,$D2
	.dc.b	$ED,$78,$CF,$37,$FD,$C9,$19,$6C,$65,$FE,$09,$F2,$0F,$A1,$FC,$1B
	.dc.b	$63,$F0,$2D,$7E,$63,$5F,$3F,$DF,$7F,$34,$6F,$8F,$E8,$8C,$47,$89
	.dc.b	$AB,$04,$61,$B3,$C4,$93,$F9,$2F,$95,$FF,$1A,$3F,$75,$AF,$F5,$66
	.dc.b	$BF,$D5,$1A,$F7,$F5,$7F,$B0,$3F,$39,$AF,$C6,$6B,$BF,$07,$E8,$C6
	.dc.b	$2C,$D2,$47,$EA,$AB,$FB,$4D,$79,$FC,$BC,$11,$FB,$3E,$C3,$EF,$FE
	.dc.b	$EB,$63,$31,$83,$F0,$3F,$43,$FD,$2A,$3E,$FB,$5F,$C3,$B5,$FB,$ED
	.dc.b	$7E,$73,$72,$34,$9E,$57,$F0,$2F,$9F,$7D,$05,$25,$FF,$02,$8C,$B6
	.dc.b	$32,$FE,$BE,$32,$D8,$FE,$83,$CA,$F9,$5A,$49,$FE,$8F,$19,$6C,$65
	.dc.b	$FF,$77,$C6,$5B,$1F,$D3,$3F,$00,$F9,$F1,$FE,$F4,$6B,$FE,$FA,$6B
	.dc.b	$FE,$FF,$6A,$1F,$41,$F9,$B7,$06,$39,$7F,$AC,$C6,$63,$83,$5D,$1F
	.dc.b	$D8,$1C,$12,$3F,$B0,$F9,$0F,$A0,$79,$FF,$D9,$A3,$0D,$8C,$DF,$88
	.dc.b	$F3,$4E,$9C,$07,$FD,$25,$F4,$1C,$FD,$93,$49,$3F,$DC,$BD,$EF,$EE
	.dc.b	$69,$A4,$7C,$18,$3E,$83,$FD,$57,$D1,$E0,$8C,$7C,$1F,$EA,$1B,$19
	.dc.b	$9D,$D7,$26,$BF,$E0,$C6,$BF,$46,$78,$BB,$AF,$D8,$FC,$FB,$E8,$29
	.dc.b	$30,$FF,$06,$C6,$5F,$87,$E0,$BB,$1F,$05,$FB,$C8,$C3,$63,$33,$C1
	.dc.b	$B8,$FC,$FA,$DC,$E6,$B8,$11,$FE,$C7,$F4,$1F,$E5,$B6,$F8,$1F,$C2
	.dc.b	$BC,$74,$E7,$0F,$EA,$4D,$7F,$CD,$6D,$7F,$CD,$2D,$7F,$C6,$CD,$8E
	.dc.b	$73,$FC,$5B,$B7,$FD,$B1,$DF,$4F,$F6,$D6,$1F,$C3,$1C,$11,$C1,$30
	.dc.b	$FE,$8F,$1C,$1F,$E9,$3F,$41,$FB,$59,$FE,$9A,$D8,$CF,$57,$86,$31
	.dc.b	$1F,$E8,$4A,$FF,$0F,$D7,$F8,$19,$B1,$FD,$C5,$AE,$FD,$2F,$57,$C9
	.dc.b	$72,$8F,$EA,$AD,$7F,$A2,$37,$13,$49,$8F,$22,$61,$E1,$F9,$7F,$A5
	.dc.b	$89,$A4,$8F,$ED,$3F,$40,$F5,$FE,$63,$63,$31,$82,$38,$26,$5F,$DC
	.dc.b	$27,$83,$FD,$93,$D5,$FC,$17,$95,$FE,$3A,$AF,$ED,$31,$83,$FA,$7F
	.dc.b	$A3,$FD,$87,$CF,$BB,$7F,$87,$2B,$FC,$94,$DC,$3F,$C7,$8D,$7F,$DF
	.dc.b	$6D,$7F,$CB,$ED,$7F,$D3,$6D,$AE,$D3,$FE,$3A,$F3,$99,$7E,$5D,$B1
	.dc.b	$98,$C1,$E6,$FA,$0F,$A9,$46,$93,$7D,$12,$BF,$CC,$51,$97,$FC,$93
	.dc.b	$EA,$3E,$83,$F4,$CD,$8C,$C6,$0F,$8B,$F4,$1F,$ED,$28,$FF,$36,$37
	.dc.b	$4F,$F2,$F3,$5F,$E6,$F8,$A3,$11,$FE,$82,$6B,$FE,$BE,$6C,$6F,$CF
	.dc.b	$EE,$7E,$E7,$E7,$DF,$41,$49,$87,$F9,$8A,$32,$FF,$9F,$FE,$81,$F3
	.dc.b	$5F,$71,$B1,$98,$C1,$1C,$13,$2F,$F3,$C4,$F0,$7F,$CF,$5D,$3F,$EF
	.dc.b	$87,$FD,$36,$D7,$FD,$AE,$DC,$BF,$D1,$F1,$F8,$CF,$25,$F3,$14,$98
	.dc.b	$7F,$A5,$63,$2F,$FA,$5F,$E0,$3F,$2D,$5F,$E8,$96,$BF,$EA,$CE,$7F
	.dc.b	$CB,$C6,$23,$FD,$5B,$F3,$EF,$A0,$FC,$DB,$6F,$FD,$44,$D9,$FF,$4A
	.dc.b	$35,$FE,$B9,$EE,$FF,$11,$6F,$77,$F8,$C2,$31,$1F,$DC,$9A,$FF,$77
	.dc.b	$6B,$FE,$85,$8E,$04,$AF,$F6,$8C,$FF,$1D,$F2,$BF,$DB,$55,$FE,$E1
	.dc.b	$F8,$6F,$F0,$ED,$7F,$80,$7C,$DF,$E8,$F3,$D7,$0D,$DF,$0F,$15,$87
	.dc.b	$0A,$FF,$7E,$78,$5D,$31,$FF,$03,$D3,$E2,$7F,$88,$47,$F8,$34,$CB
	.dc.b	$8A,$F1,$9F,$EF,$37,$11,$AA,$E9,$3F,$E1,$5A,$E9,$18,$8E,$0D,$8C
	.dc.b	$CF,$79,$E1,$F1,$6F,$2B,$E0,$E9,$EE,$B4,$E6,$FE,$17,$95,$C3,$0D
	.dc.b	$BC,$B0,$B0,$BE,$31,$86,$D7,$29,$D5,$F8,$F0,$F0,$3D,$E3,$FC,$21
	.dc.b	$87,$FB,$DB,$E8,$1F,$23,$ED,$E7,$46,$31,$F6,$5A,$73,$5B,$AD,$4F
	.dc.b	$CC,$9C,$D7,$06,$C6,$67,$73,$FC,$21,$B1,$E3,$AF,$67,$6B,$E6,$3E
	.dc.b	$0E,$1C,$B3,$F9,$53,$9C,$7E,$06,$8F,$91,$7D,$07,$D5,$6C,$FA,$A7
	.dc.b	$53,$92,$46,$67,$79,$FE,$1A,$B7,$39,$AF,$C0,$71,$3E,$83,$E9,$36
	.dc.b	$7C,$69,$CC,$47,$2F,$07,$17,$F5,$0F,$D7,$7D,$07,$E3,$DF,$F7,$93
	.dc.b	$79,$CD,$24,$FF,$17,$8C,$B6,$33,$3F,$EB,$09,$CB,$DF,$17,$CC,$FF
	.dc.b	$99,$6F,$F1,$53,$E0,$7E,$6C,$66,$38,$7C,$27,$F2,$7D,$BF,$38,$DE
	.dc.b	$3F,$D4,$2F,$8A,$FC,$B3,$9A,$FC,$D3,$71,$FF,$5D,$F8,$7A,$AD,$BF
	.dc.b	$9D,$F4,$7F,$C3,$78,$FE,$A5,$B8,$FE,$D5,$7C,$A6,$FC,$49,$AF,$E1
	.dc.b	$27,$7A,$7E,$A5,$DC,$66,$7F,$1B,$F4,$1F,$D0,$63,$FA,$DB,$6B,$FD
	.dc.b	$2D,$CC,$8C,$47,$E8,$DB,$19,$CF,$F5,$51,$82,$38,$3F,$71,$F3,$4E
	.dc.b	$1F,$0C,$61,$FD,$E7,$B7,$FD,$1B,$0F,$AC,$DD,$BF,$85,$2F,$A0,$DF
	.dc.b	$91,$34,$FD,$1C,$EF,$83,$F8,$D3,$B8,$CC,$FF,$09,$FA,$07,$C7,$FF
	.dc.b	$0B,$FA,$0F,$DF,$57,$EB,$27,$2F,$EB,$A3,$39,$FE,$26,$72,$FE,$2E
	.dc.b	$33,$9F,$F1,$F9,$CB,$FC,$86,$33,$97,$FA,$FB,$E5,$7F,$7D,$E1,$FE
	.dc.b	$41,$B8,$7E,$8F,$E7,$9F,$41,$F6,$A1,$3A,$C7,$FA,$4C,$63,$C2,$FE
	.dc.b	$60,$FE,$AC,$C6,$1F,$C9,$70,$E7,$CD,$FC,$B2,$72,$6F,$0E,$99,$79
	.dc.b	$70,$E7,$8F,$F5,$6C,$38,$EB,$E7,$9C,$D6,$F2,$C1,$DF,$3F,$46,$3D
	.dc.b	$32,$FF,$70,$35,$F0,$E5,$D6,$77,$87,$49,$DB,$5D,$E1,$FE,$E9,$CB
	.dc.b	$C7,$1E,$99,$D6,$1D,$4F,$04,$EB,$0F,$8F,$19,$9D,$9D,$E7,$57,$EB
	.dc.b	$BE,$98,$FB,$1F,$A9,$8F,$9A,$31,$8F,$4C,$39,$4E,$6B,$89,$27,$DD
	.dc.b	$9F,$D1,$CB,$C5,$87,$2C,$7D,$98,$71,$BE,$93,$9C,$19,$79,$9F,$8B
	.dc.b	$19,$7E,$29,$AA,$E5,$7E,$BB,$E7,$3E,$AC,$7C,$59,$7C,$2C,$3B,$63
	.dc.b	$D6,$B9,$93,$2F,$97,$97,$48,$F0,$9E,$DC,$BD,$38,$F4,$C3,$C9,$7C
	.dc.b	$23,$15,$B3,$8E,$5F,$13,$3D,$E7,$E3,$9F,$34,$63,$0F,$2E,$5E,$19
	.dc.b	$F2,$57,$D0,$C3,$AE,$1E,$E2,$65,$E4,$C7,$95,$88,$7A,$8A,$7E,$90
	.dc.b	$B9,$7E,$D1,$BD,$1F,$E0,$37,$67,$ED,$1C,$3A,$E7,$CB,$96,$CF,$F3
	.dc.b	$87,$F8,$04,$F3,$C3,$F2,$AD,$C3,$F3,$DF,$40,$EC,$F3,$D6,$AF,$89
	.dc.b	$31,$FE,$11,$5B,$71,$3B,$7F,$7D,$87,$57,$F1,$73,$B7,$35,$C3,$0F
	.dc.b	$E2,$D5,$BD,$D7,$BC,$3A,$12,$B7,$7F,$C3,$E7,$51,$8C,$3A,$92,$BF
	.dc.b	$86,$4E,$A3,$F8,$2E,$1E,$57,$E4,$1B,$AF,$87,$8F,$C2,$7A,$5F,$86
	.dc.b	$78,$E1,$E1,$AD,$46,$0E,$45,$9F,$0C,$6A,$3B,$C7,$36,$D7,$81,$F8
	.dc.b	$87,$17,$C2,$65,$EA,$70,$34,$7C,$E9,$ED,$97,$E0,$63,$6D,$AF,$1C
	.dc.b	$76,$39,$1C,$F2,$FA,$35,$CB,$1F,$8D,$7C,$70,$F1,$DF,$73,$94,$65
	.dc.b	$F1,$DF,$27,$35,$C8,$FB,$D9,$7D,$38,$F0,$D7,$17,$C5,$19,$8C,$12
	.dc.b	$78,$E1,$E5,$CB,$A9,$F4,$EE,$9F,$4B,$0F,$13,$F0,$AB,$C5,$3B,$8C
	.dc.b	$3F,$42,$77,$97,$D5,$C3,$C1,$3B,$8F,$F0,$2F,$BB,$F8,$0E,$D7,$33
	.dc.b	$CB,$C3,$72,$FF,$03,$D6,$0C,$CC,$8C,$9F,$E4,$2F,$86,$E9,$E0,$1A
	.dc.b	$63,$E7,$6C,$EC,$87,$F9,$3E,$51,$F9,$4C,$DF,$41,$FA,$1C,$DD,$FF
	.dc.b	$C4,$B5,$FF,$54,$37,$8F,$FC,$67,$5F,$A7,$BE,$3C,$5F,$C6,$B9,$3F
	.dc.b	$8D,$C6,$34,$FF,$49,$57,$EF,$67,$2D,$8E,$75,$C5,$A7,$07,$0F,$FA
	.dc.b	$86,$38,$3F,$EB,$3F,$A0,$FF,$07,$BF,$E6,$B6,$E5,$FE,$BF,$6E,$3F
	.dc.b	$EC,$B7,$D5,$3C,$7A,$7E,$B4,$6A,$30,$7F,$A7,$1D,$E9,$FE,$E2,$F2
	.dc.b	$BE,$E7,$8F,$FE,$B8,$AF,$F4,$7C,$63,$E0,$32,$C8,$C9,$FE,$94,$8F
	.dc.b	$D8,$FC,$8F,$DA,$FA,$09,$3F,$E3,$18,$CB,$FE,$C0,$3F,$C5,$EF,$C2
	.dc.b	$69,$9E,$84,$65,$B1,$98,$C1,$1F,$F5,$77,$1F,$E5,$7C,$CF,$C1,$C7
	.dc.b	$FA,$E9,$AF,$FA,$55,$FC,$73,$98,$E0,$47,$87,$37,$E7,$FC,$DF,$F4
	.dc.b	$1C,$7F,$AF,$DA,$FF,$AE,$DB,$C9,$FD,$83,$64,$62,$7F,$E0,$03,$5F
	.dc.b	$F7,$83,$63,$7D,$9F,$84,$6C,$FF,$B7,$FE,$81,$F0,$DF,$B1,$F4,$5D
	.dc.b	$B4,$FD,$04,$EA,$BF,$DF,$1D,$48,$C4,$7F,$BA,$DB,$9F,$E7,$1B,$1F
	.dc.b	$E9,$46,$BF,$EF,$56,$CF,$3C,$3F,$32,$D9,$F3,$47,$A7,$D3,$7D,$07
	.dc.b	$F0,$76,$CF,$FB,$B1,$AF,$FB,$1D,$FB,$AD,$8E,$04,$78,$70,$FF,$11
	.dc.b	$FA,$07,$C0,$FC,$76,$CF,$FB,$CD,$AF,$FB,$BD,$B9,$FD,$56,$C7,$F8
	.dc.b	$5F,$64,$62,$11,$8A,$EC,$7E,$4F,$E7,$FF,$C3,$BE,$C7,$FB,$CE,$B7
	.dc.b	$D4,$FA,$0F,$D5,$9F,$D5,$2B,$79,$7F,$A5,$5F,$F0,$CD,$7F,$A4,$1B
	.dc.b	$D0,$69,$23,$FC,$3A,$D7,$FC,$4B,$1B,$7F,$C4,$FC,$0F,$73,$E0,$C7
	.dc.b	$F1,$86,$BF,$ED,$06,$A8,$E1,$9F,$F8,$B1,$7B,$DF,$8E,$9E,$AD,$7F
	.dc.b	$C7,$71,$AE,$4E,$13,$99,$FE,$4F,$E5,$3E,$83,$A3,$63,$CE,$D7,$9B
	.dc.b	$5F,$E5,$FF,$41,$F1,$B0,$34,$91,$E7,$FA,$07,$C0,$D2,$47,$C4,$8C
	.dc.b	$B6,$32,$FA,$A3,$2D,$8F,$F2,$6F,$1F,$E3,$BD,$AB,$A9,$A4,$AF,$A3
	.dc.b	$EB,$FF,$B8,$1B,$EA,$FC,$B8,$C4,$7C,$86,$BF,$D3,$1A,$FF,$9A,$73
	.dc.b	$FF,$7D,$4F,$12,$47,$F9,$93,$07,$D0,$3C,$5F,$3B,$67,$4D,$7F,$CE
	.dc.b	$33,$C2,$7E,$F7,$D0,$3E,$6A,$92,$BF,$17,$19,$6C,$65,$FC,$0C,$65
	.dc.b	$B1,$F2,$3E,$37,$11,$FC,$8D,$AF,$EB,$DA,$FE,$B9,$BF,$1C,$FA,$0F
	.dc.b	$EC,$33,$FE,$8A,$8C,$79,$BF,$E8,$06,$C7,$FA,$19,$AF,$FA,$39,$AF
	.dc.b	$19,$E6,$FF,$A5,$63,$2D,$8C,$BF,$C0,$3E,$41,$F3,$3F,$82,$6C,$7D
	.dc.b	$16,$BF,$31,$AF,$9F,$E8,$3F,$97,$B7,$C6,$F4,$46,$23,$C4,$D5,$82
	.dc.b	$30,$D9,$E2,$49,$FC,$17,$CA,$FF,$8B,$DF,$BA,$D7,$FA,$93,$5F,$E3
	.dc.b	$CD,$7B,$FA,$BF,$C6,$5F,$90,$D7,$E8,$35,$F9,$0D,$78,$F8,$3F,$4E
	.dc.b	$31,$7F,$EA,$B8,$C1,$FA,$AA,$FE,$C7,$5E,$8F,$33,$04,$7E,$D7,$B0
	.dc.b	$FA,$0F,$BC,$D8,$CC,$60,$FC,$17,$D0,$7F,$48,$8F,$C3,$35,$FB,$CD
	.dc.b	$7E,$6B,$71,$34,$9E,$57,$43,$F8,$0F,$D0,$3E,$82,$92,$FF,$65,$19
	.dc.b	$6C,$65,$FD,$7C,$65,$B1,$FD,$03,$95,$F2,$D4,$93,$FE,$DC,$8C,$B6
	.dc.b	$32,$FF,$B9,$23,$2D,$8F,$E9,$7F,$41,$F3,$A3,$FD,$D2,$D7,$FD,$E2
	.dc.b	$D7,$FD,$EA,$D7,$FD,$F1,$F4,$1F,$9A,$7F,$DF,$2D,$73,$CB,$FD,$62
	.dc.b	$31,$06,$92,$34,$7F,$5E,$F1,$DF,$40,$F3,$FF,$DF,$31,$86,$C6,$6F
	.dc.b	$FE,$00,$CE,$5B,$EB,$3E,$83,$8B,$63,$31,$83,$FB,$6A,$F7,$A9,$1F
	.dc.b	$EE,$29,$85,$FA,$F7,$E0,$4F,$8B,$FE,$FD,$9F,$F7,$F9,$CC,$63,$AB
	.dc.b	$F8,$67,$E0,$4F,$AD,$C6,$2B,$DD,$77,$C3,$96,$3D,$63,$07,$85,$C5
	.dc.b	$7C,$78,$EB,$7D,$2F,$9C,$E4,$F9,$9F,$49,$FE,$B4,$AF,$5C,$77,$8C
	.dc.b	$38,$C1,$18,$8F,$F0,$DB,$5F,$E4,$DB,$3E,$CF,$D7,$D5,$FE,$04,$7D
	.dc.b	$B7,$CB,$E2,$73,$5B,$C7,$D7,$87,$F1,$C8,$C3,$E7,$C3,$94,$EA,$F9
	.dc.b	$3C,$23,$33,$ED,$FA,$7F,$C4,$7D,$5F,$CC,$6D,$76,$F2,$DF,$4B,$FD
	.dc.b	$6E,$3C,$33,$AB,$E5,$8B,$0E,$4D,$AE,$57,$C6,$B7,$5D,$F6,$7C,$35
	.dc.b	$3E,$1F,$85,$FE,$72,$C3,$BC,$E9,$B3,$B7,$D9,$3B,$BF,$65,$6D,$FB
	.dc.b	$59,$BC,$4F,$F1,$45,$FF,$3E,$F6,$3F,$99,$4E,$5B,$3A,$AF,$BF,$AB
	.dc.b	$EA,$7F,$C4,$78,$7C,$26,$CF,$6A,$E1,$B7,$F9,$66,$F8,$7A,$DF,$98
	.dc.b	$CB,$FD,$F9,$EC,$F8,$35,$34,$99,$FF,$84,$E3,$2D,$8C,$E5,$FE,$10
	.dc.b	$9C,$BE,$6C,$DF,$65,$F6,$DB,$7F,$6E,$B7,$9F,$96,$73,$7F,$1B,$CD
	.dc.b	$AF,$FA,$BD,$AF,$F9,$ED,$AF,$F7,$BC,$3E,$4C,$63,$23,$49,$1F,$D7
	.dc.b	$F4,$7D,$63,$BF,$F2,$0D,$D3,$F2,$17,$E4,$CF,$E5,$4E,$72,$F3,$6D
	.dc.b	$E0,$25,$FF,$82,$5B,$1F,$E0,$96,$BE,$B7,$C5,$3F,$CA,$27,$F3,$6D
	.dc.b	$9E,$B1,$8F,$03,$EC,$7B,$3F,$C4,$63,$FC,$8C,$D9,$FF,$20,$B5,$FF
	.dc.b	$6B,$35,$FF,$51,$35,$FF,$50,$B7,$5F,$F5,$7F,$67,$E4,$A3,$5A,$BE
	.dc.b	$6F,$FD,$7F,$87,$EC,$1B,$87,$E7,$2B,$C5,$A7,$E3,$E7,$2F,$FA,$A7
	.dc.b	$1F,$F5,$DF,$D2,$F5,$26,$1E,$A6,$E1,$EA,$6B,$EA,$7E,$C5,$BC,$BF
	.dc.b	$E0,$CD,$AF,$E0,$D8,$AB,$35,$FC,$DB,$D9,$7E,$07,$FE,$19,$C7,$FD
	.dc.b	$3A,$D7,$FA,$2C,$6B,$D1,$FF,$12,$37,$1F,$F5,$8D,$6E,$BF,$CF,$4D
	.dc.b	$7F,$D5,$CD,$9E,$11,$FE,$74,$F7,$5F,$63,$FB,$16,$DF,$F2,$AA,$DE
	.dc.b	$DF,$1E,$73,$7F,$D0,$65,$F4,$8F,$9C,$F6,$37,$0E,$5E,$1B,$EF,$73
	.dc.b	$5F,$07,$B7,$FB,$15,$7F,$AD,$31,$F6,$C6,$1F,$F5,$A9,$FA,$5F,$B3
	.dc.b	$FE,$31,$A7,$F0,$96,$D7,$FA,$67,$E9,$5F,$05,$FC,$DE,$73,$7F,$EE
	.dc.b	$CF,$8B,$FE,$86,$FF,$BC,$FC,$AF,$E3,$B1,$88,$FF,$7A,$35,$E1,$1A
	.dc.b	$26,$1F,$EA,$39,$D2,$FA,$EF,$C6,$BE,$C6,$D7,$28,$FE,$AA,$6F,$E3
	.dc.b	$7F,$A6,$7D,$9B,$F0,$4A,$7F,$23,$F9,$0F,$E2,$75,$FE,$07,$EE,$56
	.dc.b	$E3,$FA,$7B,$70,$FE,$B0,$DD,$FF,$3D,$7C,$A5,$7C,$B0,$FE,$B9,$F8
	.dc.b	$37,$FA,$43,$1F,$EB,$2D,$AF,$E8,$BF,$83,$7D,$B7,$FB,$24,$67,$1F
	.dc.b	$E7,$4D,$AF,$33,$65,$E0,$2F,$95,$F0,$F6,$3F,$A0,$7C,$0B,$EE,$A9
	.dc.b	$FE,$1A,$C1,$DF,$E9,$C5,$5B,$CF,$FB,$93,$6B,$FA,$DB,$71,$FF,$47
	.dc.b	$5F,$18,$C5,$7C,$BF,$C0,$F9,$BF,$DE,$1B,$7C,$B1,$FF,$00,$46,$30
	.dc.b	$FE,$B1,$CC,$FC,$07,$FC,$08,$DC,$3F,$C0,$9F,$35,$71,$FF,$05,$4D
	.dc.b	$BE,$37,$C0,$FE,$C1,$59,$BF,$BD,$56,$FE,$BF,$E1,$0F,$A4,$FF,$09
	.dc.b	$36,$BF,$C2,$5F,$84,$3E,$D7,$35,$27,$FC,$27,$5B,$BF,$EF,$F7,$B9
	.dc.b	$CD,$7C,$FC,$1E,$9F,$F7,$3F,$9A,$FF,$49,$57,$F9,$5F,$E0,$5F,$43
	.dc.b	$FE,$58,$BD,$CF,$F8,$8E,$B7,$97,$F9,$86,$B5,$87,$3F,$88,$5F,$39
	.dc.b	$CB,$C6,$33,$5C,$4F,$C4,$DF,$4C,$7C,$31,$89,$DC,$62,$3F,$80,$E1
	.dc.b	$DE,$FB,$47,$E4,$BE,$75,$F2,$1C,$32,$FF,$39,$73,$A7,$37,$C2,$31
	.dc.b	$7B,$39,$5D,$8E,$0D,$AE,$33,$E9,$9E,$35,$CE,$73,$8F,$AA,$5C,$2F
	.dc.b	$AE,$ED,$DB,$FC,$6B,$D0,$ED,$39,$C7,$D7,$39,$7C,$D1,$9C,$BE,$44
	.dc.b	$E5,$F1,$46,$71,$F9,$33,$97,$E4,$70,$3E,$BB,$D9,$F4,$DF,$DF,$E7
	.dc.b	$31,$8F,$5B,$E1,$63,$FC,$BA,$31,$97,$CE,$8C,$3F,$26,$F9,$70,$7E
	.dc.b	$1E,$B7,$5F,$74,$D9,$F1,$E7,$6F,$F0,$8F,$AE,$FF,$1C,$E3,$F1,$DB
	.dc.b	$7F,$61,$7D,$17,$D3,$73,$5C,$7C,$CC,$19,$9B,$97,$DD,$F8,$6F,$F4
	.dc.b	$FE,$3F,$4D,$B8,$FE,$11,$7D,$6F,$DF,$73,$8F,$F1,$C6,$F1,$9A,$70
	.dc.b	$C3,$B7,$C5,$FC,$8F,$A0,$FD,$D7,$3F,$C7,$6E,$1F,$9F,$FA,$97,$D5
	.dc.b	$FA,$A1,$3A,$D7,$F3,$CD,$D7,$FD,$92,$D9,$5F,$2D,$FC,$1D,$BF,$A0
	.dc.b	$9F,$A1,$5B,$7A,$C6,$70,$FD,$6D,$F1,$D3,$CD,$39,$C7,$BB,$6F,$FD
	.dc.b	$B5,$87,$86,$F8,$47,$69,$D3,$DB,$3E,$67,$ED,$27,$78,$FB,$2B,$77
	.dc.b	$E5,$3A,$67,$F0,$DC,$46,$8F,$15,$72,$C7,$CB,$87,$6C,$7C,$4F,$7C
	.dc.b	$39,$5F,$18,$D6,$3F,$02,$73,$18,$BE,$77,$F5,$67,$2F,$42,$57,$1C
	.dc.b	$FE,$7D,$F1,$9F,$8B,$87,$9A,$F8,$DF,$BE,$B5,$8F,$A6,$F7,$1C,$5F
	.dc.b	$90,$78,$70,$EA,$F8,$31,$F2,$DF,$87,$3E,$19,$FC,$6C,$3E,$06,$1C
	.dc.b	$2B,$C1,$87,$13,$F0,$F9,$79,$31,$E6,$48,$CD,$71,$8C,$47,$92,$BC
	.dc.b	$0F,$33,$11,$C7,$2F,$64,$60,$F6,$9E,$17,$38,$F9,$E3,$18,$78,$DD
	.dc.b	$DF,$A0,$EE,$5B,$F1,$D7,$E1,$B0,$F3,$57,$60,$7D,$A2,$99,$30,$7E
	.dc.b	$AB,$81,$EB,$EA,$30,$D7,$F3,$F1,$A7,$15,$FC,$4A,$75,$1F,$B8,$9D
	.dc.b	$61,$FB,$86,$E1,$D2,$7F,$80,$7D,$5F,$F8,$79,$FE,$09,$3A,$7B,$DE
	.dc.b	$0F,$D4,$4E,$E3,$F7,$73,$B7,$C1,$3B,$7A,$F9,$1F,$03,$0F,$14,$E5
	.dc.b	$B3,$A3,$18,$79,$63,$27,$F0,$4C,$3C,$0E,$23,$44,$9F,$E2,$33,$A8
	.dc.b	$C4,$79,$0F,$3D,$7A,$AF,$C9,$3A,$C7,$E5,$56,$8F,$21,$70,$F0,$3F
	.dc.b	$10,$CB,$87,$B3,$AD,$18,$F8,$E7,$84,$EE,$BB,$BE,$DC,$3C,$73,$AC
	.dc.b	$3C,$99,$7D,$3C,$79,$D6,$AB,$BD,$6B,$2F,$64,$7C,$A3,$91,$2B,$8B
	.dc.b	$93,$89,$9C,$BC,$CE,$09,$97,$9A,$FF,$37,$5C,$A7,$E2,$47,$85,$F8
	.dc.b	$6D,$CB,$E1,$5F,$4B,$E5,$5E,$43,$C3,$7C,$EB,$E2,$65,$F8,$BA,$DC
	.dc.b	$E7,$0E,$87,$DF,$71,$7C,$E7,$AB,$D3,$0F,$05,$F3,$26,$3F,$1A,$7D
	.dc.b	$27,$18,$F9,$D3,$F5,$F8,$9C,$AF,$83,$FA,$B1,$86,$C6,$6F,$FC,$21
	.dc.b	$F4,$F8,$99,$1F,$E1,$37,$71,$CB,$EA,$7F,$B2,$4F,$F8,$DD,$BB,$7F
	.dc.b	$8D,$5B,$B7,$F9,$19,$FE,$09,$19,$E2,$E4,$4A,$FF,$24,$4E,$AB,$FB
	.dc.b	$AF,$BF,$E6,$F8,$B9,$5F,$E0,$18,$FC,$57,$01,$CC,$7F,$88,$E7,$31
	.dc.b	$23,$FC,$59,$1A,$C7,$FC,$07,$3F,$13,$CD,$FF,$11,$D7,$F9,$1B,$B9
	.dc.b	$F0,$7F,$9C,$F0,$3F,$DC,$AE,$F8,$E5,$6E,$35,$5C,$23,$FC,$C5,$7D
	.dc.b	$1F,$F2,$E7,$A6,$F5,$7E,$9B,$63,$FC,$D0,$D7,$79,$6A,$64,$69,$E3
	.dc.b	$1F,$73,$C9,$7D,$3D,$25,$7F,$89,$FE,$9F,$F9,$CF,$A1,$FE,$86,$8F
	.dc.b	$F1,$2C,$63,$BF,$FC,$F7,$7C,$A7,$FC,$F4,$D7,$FC,$43,$3B,$7F,$C3
	.dc.b	$2D,$E9,$4E,$F8,$7F,$CE,$CD,$E4,$FF,$17,$D7,$1E,$27,$9D,$FE,$D9
	.dc.b	$C7,$FD,$5F,$E9,$66,$31,$0D,$3F,$D3,$B9,$FF,$A0,$3E,$0B,$57,$2B
	.dc.b	$89,$FE,$CA,$8E,$DC,$7F,$E9,$DB,$FF,$68,$E7,$83,$D1,$C6,$F8,$EC
	.dc.b	$61,$FD,$82,$B6,$DE,$0F,$F6,$6D,$F0,$AF,$57,$A9,$FE,$C0,$AF,$F7
	.dc.b	$17,$A1,$83,$E2,$7A,$19,$8F,$56,$1F,$EA,$88,$C6,$3E,$C6,$C6,$6B
	.dc.b	$E7,$53,$B3,$F1,$8F,$23,$FD,$69,$1D,$88,$F9,$CF,$EF,$31,$C2,$FF
	.dc.b	$D8,$BC,$2F,$5D,$D9,$F4,$DB,$7C,$67,$53,$F7,$7A,$BF,$05,$7C,$AF
	.dc.b	$77,$E3,$BD,$BC,$63,$33,$F6,$2B,$8E,$5F,$E0,$CB,$7B,$3F,$C4,$70
	.dc.b	$E4,$DB,$F3,$B6,$BF,$D9,$FE,$DB,$E9,$3F,$15,$39,$B6,$1D,$3C,$3C
	.dc.b	$9F,$95,$BE,$13,$F1,$70,$7D,$D7,$E1,$5F,$F0,$F9,$8E,$3F,$CF,$36
	.dc.b	$7E,$67,$86,$FA,$1F,$99,$3F,$E2,$68,$C7,$85,$FE,$1A,$6D,$7F,$88
	.dc.b	$5A,$FF,$89,$5A,$FF,$8B,$1A,$F3,$AF,$BE,$E5,$AE,$67,$FC,$4F,$9F
	.dc.b	$FA,$83,$F0,$3F,$D6,$79,$A6,$9F,$B7,$F2,$9F,$52,$F8,$4C,$7D,$47
	.dc.b	$F8,$C3,$2F,$F1,$CB,$68,$D3,$1F,$12,$F8,$EE,$8D,$AF,$C8,$52,$31
	.dc.b	$1F,$97,$87,$A4,$FC,$13,$F8,$47,$E1,$3F,$C7,$FA,$7F,$7A,$DE,$3F
	.dc.b	$F2,$8F,$8C,$8C,$63,$E6,$F9,$5F,$49,$FE,$48,$CB,$CE,$DD,$4D,$24
	.dc.b	$FF,$94,$E7,$6D,$9D,$B6,$7E,$6F,$C8,$7F,$05,$8F,$D2,$35,$FD,$03
	.dc.b	$72,$FE,$2B,$F5,$2F,$A9,$FE,$48,$DA,$FF,$31,$66,$8C,$47,$F9,$9E
	.dc.b	$31,$3F,$E6,$26,$C7,$09,$DB,$FE,$72,$8C,$BC,$63,$2F,$C8,$6B,$FA
	.dc.b	$4B,$7D,$A7,$F3,$16,$C7,$B2,$3C,$FF,$35,$23,$C0,$F8,$E8,$D2,$46
	.dc.b	$88,$78,$DF,$CA,$FC,$AF,$EB,$63,$F6,$AD,$E8,$39,$C5,$F2,$CF,$8B
	.dc.b	$FE,$98,$DB,$F9,$2D,$7E,$43,$5F,$88,$D7,$F1,$2D,$53,$DA,$7E,$74
	.dc.b	$E6,$31,$3F,$E9,$7B,$F2,$D7,$F5,$7B,$F0,$6B,$FC,$A9,$BC,$5D,$BB
	.dc.b	$BC,$E7,$42,$61,$E8,$6B,$E8,$6B,$E8,$7F,$77,$F3,$0F,$BD,$FE,$CC
	.dc.b	$D9,$FE,$68,$D7,$FA,$16,$6F,$9A,$79,$FF,$DC,$DB,$3F,$90,$6B,$F8
	.dc.b	$F6,$BF,$8B,$6B,$FB,$06,$BF,$21,$B3,$DE,$7F,$2D,$1A,$D5,$6F,$90
	.dc.b	$F4,$93,$0F,$EA,$98,$3E,$99,$F3,$7F,$E0,$96,$CF,$E9,$9A,$FE,$95
	.dc.b	$AF,$E8,$5A,$FF,$4B,$F8,$2F,$E4,$0F,$F5,$96,$BF,$D5,$5B,$C0,$DF
	.dc.b	$FD,$BB,$F3,$BE,$3F,$7B,$FC,$29,$5A,$9E,$9E,$26,$0F,$F1,$05,$6F
	.dc.b	$0F,$EC,$D8,$7F,$8A,$CF,$ED,$A5,$AF,$EB,$CD,$8F,$C8,$FD,$37,$F0
	.dc.b	$08,$DC,$FF,$88,$DD,$6F,$FD,$F6,$BE,$ED,$6E,$BF,$C2,$9F,$6B,$FE
	.dc.b	$D9,$BF,$55,$7F,$89,$63,$13,$AF,$86,$FF,$04,$FD,$27,$E5,$AB,$51
	.dc.b	$FE,$39,$B5,$F0,$9C,$9F,$E0,$91,$5C,$5B,$D1,$FD,$1E,$3F,$59,$18
	.dc.b	$3F,$DF,$2F,$EB,$FE,$81,$F6,$3F,$D0,$27,$FC,$29,$18,$F1,$BF,$C9
	.dc.b	$4D,$8F,$07,$C2,$78,$E3,$11,$FE,$1F,$6E,$5F,$01,$BA,$26,$57,$AD
	.dc.b	$7C,$F3,$E4,$7F,$C9,$51,$B2,$46,$70,$C4,$62,$76,$E2,$74,$E8,$EB
	.dc.b	$EA,$3E,$97,$11,$86,$CF,$FA,$C5,$B3,$B9,$FE,$E1,$F0,$DF,$DC,$53
	.dc.b	$48,$F9,$33,$7D,$23,$E3,$FF,$03,$9F,$C3,$8C,$6B,$FE,$41,$E6,$7D
	.dc.b	$2F,$FA,$6A,$3E,$8C,$EB,$1F,$98,$DB,$E5,$87,$4A,$FD,$2D,$F1,$BD
	.dc.b	$CE,$5F,$F3,$A7,$D2,$BE,$4B,$FD,$19,$1F,$E6,$98,$C5,$FC,$7C,$11
	.dc.b	$89,$F9,$F1,$8A,$E3,$0C,$BE,$CF,$CF,$F7,$6E,$1F,$E5,$DB,$E0,$FB
	.dc.b	$F8,$15,$B6,$C6,$63,$D1,$3C,$2F,$FC,$D5,$7D,$1F,$F2,$D4,$7F,$A0
	.dc.b	$EB,$F7,$BF,$0D,$EF,$F0,$7F,$B0,$FD,$47,$C1,$ED,$FF,$4A,$4D,$BE
	.dc.b	$18,$72,$8E,$1E,$5A,$F5,$1C,$78,$BF,$CE,$39,$7F,$9B,$DB,$7E,$4F
	.dc.b	$85,$7D,$7E,$FE,$16,$5E,$DF,$89,$3A,$8C,$1F,$CB,$BC,$9F,$F1,$0B
	.dc.b	$FD,$07,$D3,$F9,$BB,$FF,$47,$6C,$EB,$1F,$D0,$EF,$FC,$EB,$D9,$F7
	.dc.b	$79,$7F,$E3,$FC,$7A,$61,$89,$F8,$99,$7F,$8F,$27,$FC,$3A,$DC,$3F
	.dc.b	$C3,$8D,$7F,$C3,$0D,$7F,$C7,$AD,$7F,$BA,$B7,$23,$4D,$63,$FD,$3B
	.dc.b	$CA,$F2,$9E,$2F,$02,$E1,$FD,$3B,$D0,$FE,$9A,$FF,$82,$27,$E0,$C6
	.dc.b	$31,$FF,$04,$C6,$09,$3B,$24,$7C,$4F,$41,$DF,$F8,$16,$CF,$7A,$C4
	.dc.b	$FE,$1E,$FE,$45,$FF,$85,$B4,$F0,$92,$FF,$C3,$6D,$8D,$46,$0F,$F0
	.dc.b	$96,$1F,$C6,$5F,$F5,$A3,$63,$FC,$03,$5D,$B9,$BF,$C2,$9E,$0F,$E5
	.dc.b	$FE,$9F,$F7,$15,$FE,$47,$6B,$FE,$6C,$6B,$FE,$24,$6E,$E6,$93,$DE
	.dc.b	$FC,$4B,$C3,$77,$AA,$F3,$BF,$48,$DD,$7F,$BF,$D7,$F0,$9C,$FF,$DA
	.dc.b	$33,$9C,$BF,$C3,$32,$F9,$BE,$A4,$BF,$F3,$3C,$65,$B1,$99,$FF,$34
	.dc.b	$4E,$5B,$1F,$C2,$B1,$F2,$F8,$FE,$43,$E2,$45,$BF,$F4,$BF,$D3,$BC
	.dc.b	$DF,$E2,$3E,$D2,$31,$3F,$E8,$76,$E3,$FD,$1A,$75,$5C,$27,$33,$FD
	.dc.b	$36,$73,$5C,$6B,$47,$EC,$BD,$D7,$C6,$FF,$0D,$6C,$F9,$27,$13,$FC
	.dc.b	$BE,$9F,$4F,$FE,$B9,$AF,$F5,$7B,$70,$FF,$57,$78,$CF,$A7,$E1,$39
	.dc.b	$E1,$FF,$03,$E1,$FD,$CF,$0F,$F5,$17,$5F,$FB,$83,$1F,$63,$5F,$53
	.dc.b	$5F,$F7,$C7,$D3,$7F,$16,$BF,$F0,$8F,$0A,$31,$7F,$DC,$F0,$7D,$3B
	.dc.b	$E4,$7F,$94,$56,$B1,$FF,$39,$FC,$57,$F3,$67,$FD,$44,$D7,$FD,$3C
	.dc.b	$D7,$FE,$00,$F9,$7F,$C6,$A3,$11,$CA,$30,$4C,$7F,$BB,$0F,$A7,$FC
	.dc.b	$86,$9F,$E2,$88,$C7,$4F,$F3,$DC,$3C,$53,$9A,$DD,$7F,$83,$75,$7C
	.dc.b	$67,$F9,$76,$3F,$C2,$15,$F0,$A3,$19,$7F,$9D,$3E,$3B,$F9,$14,$E7
	.dc.b	$DE,$FE,$91,$5A,$8C,$5F,$F9,$92,$B6,$F7,$9D,$3E,$1F,$8F,$FF,$52
	.dc.b	$3E,$03,$1E,$D7,$F5,$5F,$A7,$FB,$D1,$83,$59,$33,$5F,$28,$65,$F6
	.dc.b	$A3,$0F,$F9,$33,$C1,$61,$D2,$B9,$56,$B0,$FF,$7A,$FC,$0B,$E5,$3F
	.dc.b	$B3,$36,$3F,$C9,$5D,$9E,$CB,$FF,$58,$CE,$5F,$F4,$CB,$72,$F0,$36
	.dc.b	$BF,$C6,$D5,$C7,$2F,$F4,$EE,$1F,$DC,$7E,$53,$FD,$3F,$5E,$ED,$11
	.dc.b	$8C,$BA,$D6,$F1,$FE,$F1,$CC,$FA,$7A,$7F,$7D,$FA,$3C,$65,$CE,$6E
	.dc.b	$1F,$1A,$3F,$05,$18,$D7,$F4,$2D,$8F,$F2,$96,$8B,$C5,$FD,$1A,$BA
	.dc.b	$3E,$5D,$F4,$1F,$E0,$B6,$E3,$FE,$A6,$FA,$67,$D3,$E6,$A4,$FB,$EB
	.dc.b	$78,$FF,$B0,$AF,$73,$9A,$FF,$1B,$4F,$EC,$7D,$1F,$F0,$D7,$CE,$3E
	.dc.b	$9F,$FD,$13,$F0,$6F,$8D,$FF,$32,$6D,$FD,$C3,$2F,$F5,$BD,$ED,$FF
	.dc.b	$37,$56,$F1,$FF,$38,$56,$DF,$F6,$9D,$6D,$FF,$2D,$DF,$4C,$7C,$38
	.dc.b	$F8,$23,$E3,$D6,$F0,$E9,$C0,$8C,$4F,$1C,$BC,$77,$C7,$1F,$F7,$DF
	.dc.b	$CE,$3E,$67,$18,$A3,$17,$E8,$C3,$EA,$C7,$F1,$FC,$BB,$BA,$C3,$CE
	.dc.b	$74,$C3,$9C,$66,$77,$38,$9E,$FC,$4F,$A7,$F2,$71,$FF,$1F,$F3,$B3
	.dc.b	$18,$E0,$FE,$AD,$8F,$BA,$F8,$69,$E5,$CB,$F0,$D8,$70,$8C,$63,$CD
	.dc.b	$D4,$FF,$2B,$FA,$7F,$7F,$D2,$7F,$80,$F1,$FA,$93,$9A,$E9,$BA,$B5
	.dc.b	$BF,$DE,$C3,$8F,$AB,$F2,$F0,$E0,$7C,$2B,$E1,$19,$D5,$5C,$31,$F1
	.dc.b	$C6,$1E,$F8,$F6,$C1,$83,$E9,$FF,$C8,$99,$FD,$E6,$E7,$FA,$85,$F4
	.dc.b	$7F,$69,$CD,$F2,$6D,$F3,$F3,$75,$5B,$7F,$03,$F0,$FF,$E9,$DC,$7F
	.dc.b	$16,$DC,$7F,$6A,$BE,$9F,$7F,$15,$35,$F8,$F3,$BD,$3D,$2E,$E3,$33
	.dc.b	$F1,$FE,$97,$F7,$3C,$DF,$51,$B7,$F8,$AF,$94,$7C,$FF,$AA,$13,$AD
	.dc.b	$3E,$BB,$74,$FF,$6A,$36,$57,$CB,$5F,$07,$6F,$E7,$E7,$F2,$E7,$47
	.dc.b	$18,$74,$6E,$1F,$EE,$5B,$E5,$A7,$9E,$73,$7F,$EE,$9C,$7C,$0F,$3C
	.dc.b	$FF,$29,$7C,$23,$BC,$E8,$95,$CA,$75,$8F,$86,$FB,$12,$FE,$26,$7F
	.dc.b	$0F,$0E,$33,$9A,$FE,$19,$7C,$B2,$F8,$78,$F9,$B1,$F2,$1E,$27,$C1
	.dc.b	$8F,$37,$58,$F4,$9E,$F9,$FC,$8C,$BB,$61,$D1,$F6,$E1,$C5,$CD,$7C
	.dc.b	$3C,$3A,$DF,$18,$F0,$4E,$B1,$F3,$46,$E3,$8B,$E6,$CB,$D7,$87,$58
	.dc.b	$E2,$48,$CD,$F5,$7E,$44,$67,$1F,$B1,$5C,$67,$24,$8C,$D7,$17,$D9
	.dc.b	$5C,$4F,$C4,$D7,$18,$CD,$71,$9C,$D7,$17,$D4,$71,$C3,$DA,$FA,$EB
	.dc.b	$89,$1E,$39,$FC,$0C,$FE,$86,$1E,$C8,$C6,$7F,$50,$F9,$6E,$33,$F9
	.dc.b	$59,$F9,$70,$F0,$DF,$97,$2F,$0E,$3F,$6E,$47,$B8,$A7,$A0,$A7,$E1
	.dc.b	$8A,$FF,$0C,$F4,$F3,$1F,$9F,$E0,$46,$23,$F6,$B1,$87,$73,$C5,$B1
	.dc.b	$9B,$FD,$CB,$74,$FD,$CB,$67,$F7,$3F,$4F,$FE,$8D,$7F,$83,$46,$5E
	.dc.b	$F8,$60,$FD,$7C,$EE,$3F,$5F,$3B,$C3,$C5,$87,$82,$F7,$E4,$7C,$28
	.dc.b	$C5,$F0,$9E,$8F,$F1,$3C,$3A,$4E,$89,$8F,$A2,$B9,$D7,$09,$CB,$F1
	.dc.b	$E3,$2F,$52,$46,$B2,$5F,$6B,$E9,$5A,$6D,$7E,$FA,$75,$3D,$4D,$16
	.dc.b	$35,$87,$B6,$B8,$92,$74,$FF,$1D,$AE,$71,$F3,$0E,$65,$AE,$6F,$B6
	.dc.b	$79,$39,$79,$57,$78,$EA,$43,$47,$DC,$74,$E0,$98,$FD,$69,$E7,$1D
	.dc.b	$CE,$51,$97,$91,$93,$8E,$5F,$89,$C7,$E5,$DF,$6A,$E1,$5A,$CB,$E9
	.dc.b	$E3,$F5,$A3,$07,$DF,$AE,$18,$7E,$24,$95,$C3,$1F,$83,$87,$43,$C9
	.dc.b	$8F,$83,$0F,$24,$66,$FC,$99,$7C,$E8,$C6,$1E,$DB,$E1,$7F,$42,$72
	.dc.b	$F8,$CE,$77,$E2,$70,$7F,$83,$F9,$5F,$05,$F4,$79,$7F,$C3,$0D,$8C
	.dc.b	$CF,$F8,$3B,$E9,$F1,$32,$3F,$C2,$2E,$E3,$97,$D2,$FF,$67,$9F,$F0
	.dc.b	$E3,$77,$FF,$27,$B7,$6F,$F3,$CB,$FC,$2A,$33,$C3,$C8,$95,$FE,$2C
	.dc.b	$9D,$57,$F7,$8F,$7B,$CD,$F1,$32,$BF,$C0,$91,$F6,$4C,$3E,$F8,$C3
	.dc.b	$A3,$FC,$52,$EB,$0F,$F0,$2C,$FC,$3F,$2F,$FC,$4B,$5F,$E4,$4E,$C7
	.dc.b	$C1,$7E,$5F,$C0,$FF,$77,$3B,$E1,$95,$B8,$D5,$70,$8F,$F2,$2D,$F4
	.dc.b	$7F,$C8,$1E,$9B,$D7,$FA,$4D,$8F,$F2,$5B,$5D,$E5,$A9,$91,$A7,$8C
	.dc.b	$7D,$BF,$21,$F4,$F4,$95,$FE,$29,$FA,$7F,$E7,$7E,$7B,$BB,$FC,$47
	.dc.b	$18,$7F,$D0,$F5,$C2,$BF,$C2,$73,$97,$FC,$34,$DE,$8F,$F4,$54,$EF
	.dc.b	$87,$FC,$5F,$3B,$E1,$FF,$59,$7A,$37,$CF,$FF,$6D,$63,$FE,$AD,$8E
	.dc.b	$5E,$53,$E0,$35,$12,$74,$7F,$BB,$B3,$FF,$60,$4F,$5E,$3F,$F3,$9E
	.dc.b	$DF,$EC,$6E,$7C,$1E,$5E,$47,$C7,$63,$1F,$F6,$84,$D7,$FD,$CB,$1E
	.dc.b	$7F,$51,$E8,$D2,$4F,$87,$D0,$CB,$63,$E1,$46,$23,$FD,$B4,$7F,$AA
	.dc.b	$23,$E4,$2E,$CF,$C2,$3A,$D7,$FD,$55,$3C,$89,$1F,$DC,$B3,$FA,$CF
	.dc.b	$F7,$4A,$E3,$18,$E1,$7A,$AE,$DF,$A0,$DB,$FF,$7C,$35,$F0,$75,$FE
	.dc.b	$19,$CC,$F0,$23,$F3,$E3,$4F,$B2,$9E,$C7,$F0,$68,$C3,$70,$F2,$37
	.dc.b	$A7,$FD,$EF,$ED,$BE,$9F,$F0,$93,$9B,$61,$D3,$C3,$C9,$F8,$DB,$E1
	.dc.b	$3F,$0F,$0F,$F0,$A8,$FA,$7F,$C0,$B8,$6C,$7E,$75,$BE,$1B,$E9,$FE
	.dc.b	$4D,$7F,$87,$A3,$1E,$1F,$F8,$55,$B7,$FE,$1A,$6B,$FE,$20,$6B,$FE
	.dc.b	$24,$6B,$CE,$BE,$EB,$96,$B9,$9F,$F1,$16,$5F,$E9,$8F,$9D,$FD,$37
	.dc.b	$9A,$69,$FA,$FF,$2D,$F4,$EF,$A7,$C7,$D2,$FF,$8A,$F2,$FF,$16,$B6
	.dc.b	$8D,$31,$F1,$2F,$8F,$E8,$DA,$F9,$87,$F8,$D9,$C3,$F9,$AA,$7A,$8F
	.dc.b	$A7,$FE,$01,$F4,$FF,$C5,$7A,$3F,$2B,$DC,$8C,$72,$7F,$92,$A3,$1E
	.dc.b	$77,$A3,$E4,$E6,$7F,$E4,$6C,$FC,$F1,$FE,$45,$6C,$67,$64,$E5,$B1
	.dc.b	$98,$C1,$1F,$F2,$8F,$1B,$E4,$3F,$80,$C7,$BA,$31,$E2,$7F,$15,$9C
	.dc.b	$CF,$EC,$FE,$95,$F4,$FF,$C7,$5B,$1F,$E5,$CD,$11,$88,$FF,$31,$C6
	.dc.b	$27,$FC,$BC,$D8,$E1,$3B,$7F,$CD,$71,$97,$8C,$65,$F9,$0D,$7F,$4B
	.dc.b	$4F,$A7,$F4,$12,$3F,$92,$4B,$E7,$3D,$07,$89,$C4,$78,$A3,$04,$70
	.dc.b	$43,$C6,$FE,$5B,$E5,$BF,$57,$99,$A7,$EB,$B1,$7C,$B3,$E3,$3F,$A2
	.dc.b	$37,$0F,$90,$D7,$E3,$B5,$F8,$6D,$7F,$10,$D5,$3D,$A7,$E6,$CE,$63
	.dc.b	$13,$F8,$0A,$ED,$C9,$FC,$95,$BC,$7F,$E9,$CC,$7D,$58,$F9,$BC,$C9
	.dc.b	$87,$99,$EA,$47,$04,$3D,$2E,$5F,$F5,$1C,$BE,$61,$F4,$FF,$D8,$5B
	.dc.b	$3F,$91,$6B,$FD,$07,$E6,$9E,$87,$F6,$B6,$C7,$E3,$DA,$FE,$39,$AF
	.dc.b	$E2,$9A,$FE,$BD,$AF,$C8,$6E,$A6,$9A,$C3,$D1,$F2,$1E,$72,$57,$F6
	.dc.b	$FF,$A7,$7C,$DF,$F7,$F6,$C7,$E8,$DA,$FE,$89,$AF,$E7,$9A,$FD,$1F
	.dc.b	$82,$FE,$3C,$FE,$CD,$AF,$F4,$D6,$F0,$FF,$60,$F9,$F5,$7F,$0C,$AF
	.dc.b	$E8,$F5,$AE,$1E,$9E,$2F,$F8,$3E,$31,$2B,$E3,$DD,$FE,$13,$C7,$FB
	.dc.b	$5C,$FF,$5B,$69,$FD,$AC,$B8,$7F,$19,$FA,$7F,$DF,$46,$63,$91,$27
	.dc.b	$FC,$29,$3A,$8F,$EF,$1F,$4F,$FE,$CD,$AF,$EF,$21,$C4,$6B,$E1,$FF
	.dc.b	$C0,$BF,$4F,$F9,$6C,$17,$B9,$FF,$18,$36,$3F,$9B,$B5,$FF,$07,$89
	.dc.b	$E4,$DE,$7F,$E8,$D1,$FA,$B8,$C1,$FE,$F3,$7F,$5D,$F4,$2F,$A7,$FE
	.dc.b	$7F,$3F,$D1,$23,$1E,$3F,$F9,$05,$B1,$FD,$20,$7C,$2F,$B1,$46,$9F
	.dc.b	$F1,$1B,$71,$FF,$0D,$B7,$34,$CA,$E4,$BE,$7D,$F2,$5F,$12,$36,$48
	.dc.b	$CE,$18,$8C,$4E,$DC,$4E,$9D,$1D,$7D,$57,$D3,$E2,$30,$D9,$FF,$53
	.dc.b	$B6,$77,$3F,$DC,$7E,$1B,$FB,$92,$69,$1F,$27,$23,$E9,$BF,$D1,$7A
	.dc.b	$78,$A3,$1E,$A7,$E5,$2B,$E1,$C6,$27,$55,$C2,$79,$B5,$F8,$73,$B7
	.dc.b	$35,$CD,$FD,$14,$ED,$F9,$3E,$CB,$EA,$B5,$1F,$E3,$C6,$C6,$70,$F8
	.dc.b	$8D,$9F,$9D,$B3,$E9,$DF,$27,$DE,$30,$D8,$CD,$FD,$09,$CB,$63,$35
	.dc.b	$D1,$AF,$FA,$59,$AF,$F9,$8D,$AF,$F4,$4F,$25,$F4,$FE,$FF,$13,$FB
	.dc.b	$13,$63,$37,$FD,$E5,$B5,$FD,$E1,$AF,$F7,$46,$BF,$E1,$96,$BD,$E7
	.dc.b	$E6,$6D,$FE,$86,$FA,$7B,$F0,$86,$93,$0F,$BD,$19,$6C,$66,$FF,$0B
	.dc.b	$39,$7F,$0F,$F7,$7F,$E2,$0A,$FC,$33,$63,$3E,$9F,$8A,$31,$1F,$D0
	.dc.b	$E3,$E1,$79,$18,$8F,$E8,$91,$83,$87,$D5,$CB,$EA,$D9,$F0,$B5,$FC
	.dc.b	$6D,$A3,$11,$FE,$3D,$C3,$FC,$8B,$5F,$E2,$56,$C7,$F8,$91,$AF,$F8
	.dc.b	$81,$AF,$E7,$BE,$7D,$F7,$FF,$31,$B5,$FA,$56,$CF,$EA,$9A,$FE,$81
	.dc.b	$AF,$39,$E4,$49,$FF,$08,$F9,$3F,$85,$F4,$7F,$AD,$C6,$09,$5D,$75
	.dc.b	$F1,$12,$3C,$4D,$8D,$46,$08,$FF,$87,$3D,$CE,$C4,$8F,$F1,$0B,$FB
	.dc.b	$3F,$65,$F7,$9F,$56,$91,$8D,$F5,$39,$8C,$5F,$F9,$9E,$30,$49,$D9
	.dc.b	$FA,$1F,$AB,$FD,$F5,$FF,$8A,$1B,$CA,$69,$AC,$3C,$7C,$1E,$02,$63
	.dc.b	$DD,$B7,$FC,$B6,$77,$39,$24,$7F,$98,$A9,$F3,$FE,$32,$4F,$F9,$A6
	.dc.b	$32,$FF,$A0,$7E,$A5,$F4,$FF,$15,$B1,$98,$C1,$1C,$12,$FC,$93,$C1
	.dc.b	$F3,$7D,$5B,$CC,$FA,$0D,$AF,$F4,$83,$67,$73,$F5,$7C,$57,$D5,$D2
	.dc.b	$5F,$FA,$5E,$32,$FA,$7E,$13,$D0,$47,$07,$FA,$77,$EA,$DF,$27,$F8
	.dc.b	$16,$C6,$63,$07,$F3,$21,$F4,$34,$91,$FD,$56,$32,$D8,$CE,$5F,$D7
	.dc.b	$27,$2D,$8F,$E4,$9F,$1B,$F1,$63,$FD,$22,$D7,$FD,$EB,$B6,$8F,$E5
	.dc.b	$5F,$1F,$FB,$68,$FF,$38,$B7,$90,$D2,$7A,$1C,$09,$97,$FA,$9E,$38
	.dc.b	$7D,$4F,$FB,$66,$3F,$D5,$4D,$BF,$F5,$4B,$5F,$F5,$2B,$5F,$F6,$53
	.dc.b	$5F,$F4,$1B,$6C,$D3,$51,$F8,$1B,$FE,$BB,$C1,$EB,$8C,$5F,$F7,$F8
	.dc.b	$C1,$27,$64,$8F,$F6,$07,$BA,$F9,$7F,$E9,$CD,$8C,$C6,$0F,$EA,$5F
	.dc.b	$30,$FA,$9A,$49,$FF,$65,$46,$5B,$19,$C7,$FD,$97,$39,$6C,$7F,$5D
	.dc.b	$FA,$97,$87,$49,$7F,$ED,$58,$CB,$63,$35,$FE,$D6,$9C,$B6,$3F,$21
	.dc.b	$E2,$4A,$FC,$73,$67,$F1,$8D,$7F,$C5,$DC,$68,$C4,$7F,$8D,$23,$13
	.dc.b	$FE,$3C,$F2,$CF,$F8,$FB,$81,$F3,$CF,$A9,$CE,$1F,$E2,$9F,$86,$E5
	.dc.b	$97,$F9,$A3,$0E,$17,$B7,$FC,$8F,$EA,$2B,$A7,$CD,$3E,$57,$9C,$E7
	.dc.b	$1F,$F6,$4C,$E6,$F8,$DF,$09,$DB,$CA,$EC,$70,$6E,$6B,$E5,$3E,$73
	.dc.b	$A9,$A7,$38,$78,$65,$C2,$FA,$9F,$7E,$5F,$E5,$A9,$CC,$63,$0F,$8D
	.dc.b	$5B,$AF,$F3,$29,$B7,$11,$FE,$9A,$AE,$33,$A8,$E3,$5C,$DF,$F7,$4D
	.dc.b	$70,$9E,$53,$99,$FE,$2B,$F5,$37,$80,$D2,$57,$59,$FF,$3B,$CE,$A3
	.dc.b	$11,$C9,$B1,$98,$FF,$2E,$36,$3F,$13,$3F,$33,$E9,$3F,$CF,$B1,$F5
	.dc.b	$2B,$AC,$E7,$DC,$FA,$3F,$53,$FE,$0C,$D3,$E6,$37,$7F,$C6,$2F,$A4
	.dc.b	$DF,$54,$CB,$A6,$EA,$E1,$9A,$31,$7B,$8F,$B9,$1F,$E5,$CF,$03,$F0
	.dc.b	$3E,$C7,$F2,$CC,$FE,$E2,$FA,$9C,$CC,$CB,$BC,$EF,$4F,$D0,$3B,$8C
	.dc.b	$CF,$E3,$B1,$7D,$3B,$EA,$7F,$35,$1F,$8D,$32,$E3,$2F,$CE,$7C,$53
	.dc.b	$E7,$3D,$75,$AE,$8F,$94,$D3,$96,$1D,$63,$E7,$57,$1E,$1F,$14,$E6
	.dc.b	$B7,$8F,$7D,$7F,$D5,$F3,$9C,$BC,$39,$78,$23,$13,$AA,$E1,$7F,$EB
	.dc.b	$C9,$CA,$C7,$C1,$A7,$9A,$73,$3E,$7C,$7F,$70,$DC,$3F,$D8,$B7,$C2
	.dc.b	$3C,$B9,$F1,$C7,$BC,$6C,$ED,$97,$96,$33,$8F,$63,$C7,$7C,$F1,$EF
	.dc.b	$19,$9D,$9E,$A9,$D6,$3E,$DC,$7C,$04,$C7,$F8,$14,$E9,$F8,$15,$CA
	.dc.b	$75,$87,$C8,$CF,$E7,$CE,$6B,$D6,$72,$CB,$D1,$7C,$31,$F0,$4E,$6F
	.dc.b	$8C,$E6,$3B,$BF,$72,$B8,$C7,$4C,$3D,$B8,$7A,$E7,$31,$8A,$E5,$8F
	.dc.b	$A6,$FE,$51,$32,$F8,$F7,$C3,$2F,$1E,$1C,$6B,$C9,$9F,$8D,$E1,$97
	.dc.b	$7C,$FC,$F1,$E3,$BE,$27,$DC,$C3,$A4,$7A,$8D,$1F,$6A,$78,$E1,$DC
	.dc.b	$95,$C6,$32,$F4,$9E,$53,$F3,$F0,$F2,$63,$F7,$73,$F8,$D8,$70,$BD
	.dc.b	$E1,$C7,$3F,$85,$7B,$91,$0F,$31,$4F,$D7,$15,$FD,$53,$78,$3F,$CF
	.dc.b	$AE,$CF,$DF,$B8,$75,$E7,$E5,$CB,$67,$F8,$E3,$FA,$79,$E7,$9F,$E6
	.dc.b	$1B,$87,$E9,$BE,$A7,$FC,$C1,$86,$EE,$3F,$C0,$AB,$6F,$EC,$E7,$6E
	.dc.b	$67,$77,$D5,$FE,$0F,$3A,$8C,$4E,$C8,$FF,$14,$8D,$5F,$F0,$8B,$EF
	.dc.b	$5E,$C3,$71,$E7,$F1,$3E,$65,$F4,$24,$71,$C3,$C3,$3A,$8C,$57,$23
	.dc.b	$F8,$75,$FB,$32,$63,$E2,$9C,$E1,$EC,$BE,$38,$F9,$F1,$E9,$5A,$3A
	.dc.b	$1E,$78,$D3,$C4,$D3,$E6,$7E,$54,$72,$78,$4F,$62,$3A,$39,$1A,$39
	.dc.b	$BF,$0D,$CE,$5F,$80,$7D,$D8,$FB,$2B,$B4,$E6,$FC,$55,$CE,$B4,$F2
	.dc.b	$CB,$E8,$D6,$F0,$ED,$3E,$E8,$C3,$E2,$3C,$C6,$67,$89,$F3,$72,$F9
	.dc.b	$91,$E9,$24,$F2,$BF,$0C,$71,$CB,$E4,$46,$31,$F3,$D7,$82,$BD,$17
	.dc.b	$D2,$30,$4C,$3C,$19,$7C,$07,$C7,$8F,$4A,$E1,$18,$BF,$8F,$18,$9F
	.dc.b	$84,$7E,$2A,$3F,$C0,$7C,$EF,$83,$FC,$0F,$2A,$31,$3A,$9C,$C3,$EA
	.dc.b	$6C,$49,$D1,$D2,$39,$FD,$27,$F7,$A8,$FF,$05,$B7,$5F,$F2,$23,$77
	.dc.b	$FF,$32,$3F,$C2,$E3,$3C,$3C,$89,$5F,$E1,$A9,$D5,$7F,$74,$F8,$7F
	.dc.b	$2F,$D1,$CA,$FE,$D9,$1F,$86,$3F,$8C,$72,$7F,$15,$6C,$71,$3F,$C7
	.dc.b	$3C,$48,$CC,$7F,$79,$6B,$FE,$16,$6C,$6E,$38,$C7,$0D,$9E,$4F,$F8
	.dc.b	$CE,$7F,$C9,$11,$8E,$2D,$4E,$67,$FC,$B4,$DC,$3F,$CC,$8D,$8F,$F4
	.dc.b	$43,$5F,$F1,$C3,$6B,$A4,$76,$F2,$DF,$1F,$F7,$1B,$1F,$E4,$26,$BF
	.dc.b	$E3,$E7,$D4,$D8,$E0,$47,$86,$3F,$B0,$FA,$8F,$E8,$5E,$47,$FA,$63
	.dc.b	$0F,$F4,$DF,$D4,$FE,$2B,$C7,$3B,$C3,$FD,$45,$A7,$F8,$AB,$0F,$E7
	.dc.b	$D1,$FE,$40,$6E,$C6,$92,$3F,$A1,$35,$FE,$95,$1B,$7F,$D9,$7F,$51
	.dc.b	$FD,$27,$1F,$F4,$7B,$7C,$73,$49,$3C,$08,$7F,$B4,$1E,$5B,$BE,$7A
	.dc.b	$92,$7F,$DC,$6D,$7F,$B1,$5F,$E0,$23,$11,$F6,$9F,$27,$93,$FE,$C7
	.dc.b	$8F,$EC,$53,$FE,$AB,$F2,$B2,$48,$FE,$C5,$CC,$F8,$CF,$8E,$D8,$CC
	.dc.b	$60,$8E,$08,$7F,$6F,$8E,$0F,$9B,$93,$EA,$BE,$26,$B9,$E0,$F9,$51
	.dc.b	$88,$FF,$5B,$35,$FF,$6D,$35,$FF,$45,$37,$03,$7D,$37,$D3,$7D,$56
	.dc.b	$D7,$C2,$6B,$F2,$9A,$FF,$B1,$5A,$FF,$BA,$5A,$FF,$A6,$5B,$47,$3D
	.dc.b	$4F,$A5,$F8,$35,$FE,$19,$8C,$67,$FE,$0F,$6C,$7F,$85,$5A,$FE,$3B
	.dc.b	$E8,$3F,$2A,$47,$FC,$28,$D7,$46,$08,$7F,$89,$7C,$17,$D2,$BD,$9F
	.dc.b	$F0,$F4,$61,$B1,$99,$FF,$13,$4E,$5B,$F4,$AF,$84,$E2,$D8,$CC,$60
	.dc.b	$EB,$E4,$BE,$96,$92,$7A,$C6,$5E,$FF,$48,$F8,$5F,$2B,$63,$31,$83
	.dc.b	$C3,$F0,$72,$3C,$AD,$7E,$06,$28,$C4,$7F,$92,$72,$4E,$63,$89,$FA
	.dc.b	$4F,$A2,$6D,$FB,$B6,$C6,$63,$07,$CA,$FB,$17,$C8,$52,$4F,$CA,$8C
	.dc.b	$B6,$32,$FD,$58,$CB,$63,$C1,$F2,$32,$3C,$8D,$78,$7C,$2F,$84,$8F
	.dc.b	$76,$BA,$30,$43,$F8,$3F,$C6,$3E,$6B,$F8,$C3,$63,$F8,$2B,$5D,$C7
	.dc.b	$F9,$97,$E4,$9F,$60,$FB,$0C,$36,$BF,$CD,$D3,$A9,$C9,$23,$33,$B7
	.dc.b	$FC,$F5,$3B,$73,$3E,$4F,$42,$47,$91,$AF,$0A,$7D,$4F,$88,$91,$D9
	.dc.b	$AE,$DA,$FF,$A2,$EF,$D7,$1F,$CE,$7E,$57,$FB,$4C,$FF,$26,$6B,$FC
	.dc.b	$B9,$AF,$FA,$4F,$BD,$18,$8F,$F4,$BD,$3C,$47,$CC,$7C,$96,$CF,$C6
	.dc.b	$6B,$F0,$9A,$F9,$5A,$F4,$6A,$9E,$D3,$D4,$FE,$A5,$F4,$2F,$95,$A4
	.dc.b	$AF,$E8,$31,$97,$C5,$F0,$9E,$52,$38,$3E,$DC,$7F,$78,$7E,$43,$5F
	.dc.b	$7B,$5F,$23,$6C,$D2,$7A,$9F,$8B,$78,$66,$FA,$0F,$D7,$DF,$F5,$C6
	.dc.b	$BF,$3D,$AF,$FB,$12,$38,$73,$FF,$71,$9C,$CA,$72,$FF,$60,$8C,$C6
	.dc.b	$23,$FD,$89,$F3,$FF,$9F,$7F,$D9,$8D,$7F,$DA,$0D,$7F,$DA,$6D,$4D
	.dc.b	$27,$1F,$F8,$39,$D7,$E1,$5F,$D5,$E4,$D2,$79,$7F,$E1,$4F,$5F,$FD
	.dc.b	$87,$3F,$ED,$D6,$BF,$EE,$16,$BF,$EE,$46,$BF,$E4,$9F,$7F,$91,$1F
	.dc.b	$F6,$E4,$6B,$E4,$9F,$10,$F6,$BF,$DC,$11,$86,$C6,$6B,$FD,$D1,$39
	.dc.b	$6F,$BE,$F9,$EE,$2D,$8C,$C6,$0F,$F1,$7A,$F7,$E9,$1F,$F1,$9B,$FC
	.dc.b	$23,$E7,$BF,$C2,$1E,$ED,$F4,$4D,$25,$7F,$08,$9D,$B6,$76,$D9,$DA
	.dc.b	$20,$F9,$EF,$DA,$C7,$FC,$01,$6C,$68,$E0,$E1,$C3,$5E,$9A,$3E,$79
	.dc.b	$F1,$5E,$29,$FF,$2B,$C6,$2F,$A4,$7D,$18,$C3,$DE,$B9,$57,$18,$EB
	.dc.b	$3A,$79,$C6,$5E,$4D,$9E,$35,$FE,$3E,$F1,$BF,$AA,$B7,$C7,$E1,$CE
	.dc.b	$9C,$E1,$FE,$4A,$6B,$FC,$75,$AF,$EB,$DA,$FE,$B5,$B1,$CE,$7C,$9D
	.dc.b	$9F,$C8,$7E,$72,$F8,$26,$93,$1F,$E4,$4D,$C7,$F9,$13,$5F,$E6,$DF
	.dc.b	$3D,$FE,$05,$9F,$E9,$AD,$8F,$E9,$2D,$77,$E7,$FB,$BC,$4C,$47,$F0
	.dc.b	$76,$BF,$C1,$5B,$9F,$F3,$A9,$F4,$D3,$E7,$29,$30,$F0,$B7,$0F,$0B
	.dc.b	$5F,$3B,$F3,$B2,$7A,$54,$95,$FE,$72,$8C,$BF,$DA,$3E,$4D,$F3,$DF
	.dc.b	$29,$B1,$98,$C1,$1C,$13,$0F,$ED,$73,$C1,$FE,$FF,$E3,$FE,$03,$CD
	.dc.b	$FE,$89,$5F,$DF,$A3,$07,$7D,$E4,$78,$DB,$5D,$1A,$FF,$69,$BF,$B9
	.dc.b	$3B,$3F,$C1,$3E,$B7,$72,$38,$3F,$C9,$BF,$3D,$F4,$BD,$43,$49,$C1
	.dc.b	$F9,$D8,$CB,$63,$33,$CA,$3F,$37,$EA,$FF,$02,$C7,$FB,$A3,$5F,$E8
	.dc.b	$CD,$7F,$A2,$36,$FC,$3C,$1D,$49,$7E,$66,$D7,$F9,$AF,$0F,$C2,$4E
	.dc.b	$49,$1F,$EB,$2B,$7C,$F7,$90,$93,$FE,$B0,$8C,$BF,$E6,$EF,$9E,$7C
	.dc.b	$E7,$BD,$B1,$98,$C1,$1C,$12,$FF,$CF,$D3,$C1,$FF,$40,$7C,$F3,$C7
	.dc.b	$F9,$4D,$AF,$F3,$EB,$67,$73,$F3,$FC,$57,$AD,$49,$7F,$E8,$48,$CB
	.dc.b	$E9,$F7,$7D,$04,$70,$7F,$A2,$BE,$79,$F3,$1F,$6D,$B1,$98,$C1,$FC
	.dc.b	$6F,$1D,$FC,$B5,$83,$49,$3F,$CA,$A7,$6D,$9D,$B6,$77,$3F,$EB,$DD
	.dc.b	$7F,$90,$FC,$1F,$C7,$8F,$F5,$F4,$60,$FE,$4B,$F0,$BF,$B4,$8F,$F6
	.dc.b	$C7,$6E,$09,$87,$FB,$65,$B1,$A8,$C1,$FE,$EE,$F9,$EF,$F5,$83,$74
	.dc.b	$FE,$63,$18,$9F,$F7,$CB,$5F,$F2,$63,$5F,$F0,$F3,$5F,$F0,$E3,$7C
	.dc.b	$03,$98,$FC,$05,$FC,$BE,$0F,$54,$7F,$5B,$6B,$FD,$89,$AF,$F6,$56
	.dc.b	$BC,$B0,$F9,$D1,$83,$76,$F8,$A7,$CF,$7F,$4C,$6D,$7F,$5E,$9D,$4E
	.dc.b	$4F,$EA,$3F,$1A,$F9,$EA,$4A,$E1,$19,$6C,$67,$1F,$EE,$96,$AD,$46
	.dc.b	$1A,$FF,$9D,$7E,$79,$DF,$FD,$81,$B8,$79,$5B,$3D,$1A,$F1,$9E,$04
	.dc.b	$9E,$0F,$E5,$7C,$49,$1F,$91,$6C,$FF,$7E,$E4,$FF,$41,$4E,$49,$19
	.dc.b	$9D,$F9,$5F,$E1,$8C,$9F,$24,$F9,$EF,$F0,$DF,$A4,$F8,$3F,$F0,$CE
	.dc.b	$7D,$32,$FF,$1D,$67,$FE,$63,$C3,$83,$FE,$1F,$BE,$13,$FE,$63,$F8
	.dc.b	$F7,$C6,$74,$C7,$FC,$D5,$18,$BE,$11,$8A,$D1,$C6,$AE,$6B,$E1,$3F
	.dc.b	$E3,$DA,$E7,$5A,$9F,$F1,$85,$F4,$AE,$4D,$BF,$B5,$F3,$DF,$E3,$88
	.dc.b	$FF,$56,$70,$A3,$15,$FE,$7F,$8C,$4F,$48,$C4,$E9,$FF,$40,$4E,$9F
	.dc.b	$F1,$95,$76,$AE,$11,$AA,$E1,$39,$8F,$F1,$86,$8F,$83,$E9,$1F,$C6
	.dc.b	$9B,$19,$BE,$73,$96,$C6,$63,$9C,$F0,$24,$E9,$F9,$7F,$31,$FE,$74
	.dc.b	$8F,$8D,$3C,$CF,$98,$7F,$A6,$E3,$4E,$2D,$F3,$DF,$C5,$F6,$FB,$F1
	.dc.b	$8E,$9F,$AF,$2F,$97,$B1,$31,$FA,$53,$C7,$83,$FD,$58,$EE,$33,$3F
	.dc.b	$80,$F5,$BF,$96,$E5,$E1,$6E,$5E,$05,$F3,$DB,$F0,$A6,$1F,$7E,$77
	.dc.b	$97,$FA,$E5,$DC,$66,$7F,$1F,$83,$E5,$DF,$3D,$F9,$88,$F8,$66,$5C
	.dc.b	$65,$F9,$EF,$9D,$7C,$F7,$AE,$BF,$D0,$17,$CB,$9F,$E6,$B6,$39,$B6
	.dc.b	$F9,$57,$19,$ED,$5C,$30,$EF,$A7,$FA,$2B,$0E,$0F,$FA,$37,$1F,$0E
	.dc.b	$DE,$A9,$CE,$1E,$2B,$E2,$DC,$3F,$49,$7C,$34,$ED,$12,$7C,$B5,$C2
	.dc.b	$B5,$18,$9D,$67,$FB,$E8,$F3,$9F,$88,$9D,$C6,$70,$F1,$CE,$A3,$17
	.dc.b	$D7,$3F,$81,$5A,$24,$67,$2F,$63,$EA,$BE,$59,$FC,$38,$D6,$7E,$48
	.dc.b	$C6,$76,$99,$F9,$27,$31,$EB,$26,$7F,$4B,$1E,$F7,$C6,$7B,$9C,$B3
	.dc.b	$F3,$E5,$E3,$CF,$CF,$5F,$32,$72,$F1,$8F,$A7,$97,$86,$FE,$91,$E0
	.dc.b	$CB,$C7,$97,$87,$3E,$19,$75,$3D,$B5,$B3,$8C,$FC,$AC,$7C,$19,$FA
	.dc.b	$27,$47,$A3,$1F,$03,$EA,$26,$5E,$4C,$FC,$F1,$C2,$31,$5C,$F0,$ED
	.dc.b	$3E,$1B,$E7,$5F,$02,$31,$7C,$AB,$64,$CF,$F2,$D1,$97,$DE,$49,$F9
	.dc.b	$98,$77,$8F,$AB,$1F,$EE,$3B,$FD,$43,$72,$FD,$A3,$72,$FF,$1A,$4F
	.dc.b	$F0,$FF,$42,$F0,$9A,$4A,$FF,$71,$CE,$DF,$D9,$CE,$DB,$3F,$A6,$F9
	.dc.b	$EF,$F2,$E4,$7E,$EA,$75,$A7,$E7,$5B,$97,$ED,$5B,$3F,$90,$AE,$2F
	.dc.b	$42,$57,$1B,$ED,$E1,$A7,$B4,$47,$78,$7E,$FE,$F9,$DF,$19,$C9,$FC
	.dc.b	$43,$0F,$0E,$CA,$DD,$F9,$70,$E9,$87,$13,$F8,$D4,$EB,$0E,$53,$AA
	.dc.b	$E1,$1F,$BD,$9D,$3F,$C9,$F0,$F7,$B6,$B8,$3F,$07,$1F,$63,$E6,$34
	.dc.b	$7A,$31,$F3,$46,$B1,$F2,$63,$E0,$6D,$73,$3A,$9C,$9C,$D7,$28,$CE
	.dc.b	$5F,$76,$73,$97,$A2,$31,$7D,$49,$5D,$4F,$34,$69,$F8,$67,$C1,$C3
	.dc.b	$CD,$3C,$23,$C4,$F2,$C3,$C5,$1C,$B2,$F8,$B7,$CE,$73,$8F,$B2,$B8
	.dc.b	$9D,$9F,$A7,$87,$68,$C6,$1D,$EF,$96,$3E,$BC,$7A,$93,$2F,$C5,$DF
	.dc.b	$C2,$8C,$3E,$93,$11,$F5,$B2,$F1,$E1,$D6,$B7,$97,$8A,$F6,$7D,$7C
	.dc.b	$3C,$18,$62,$75,$3F,$E0,$5E,$57,$C1,$7A,$79,$91,$89,$FF,$19,$35
	.dc.b	$E1,$F3,$DB,$8C,$5F,$F8,$31,$B9,$FE,$96,$F9,$E1,$BB,$ED,$F3,$5F
	.dc.b	$C1,$30,$FF,$09,$B6,$F8,$37,$16,$08,$FC,$AC,$7F,$87,$5D,$46,$78
	.dc.b	$3A,$9D,$A7,$EA,$1F,$E0,$7F,$86,$F4,$7C,$F4,$8F,$EF,$F1,$AE,$13
	.dc.b	$FA,$2E,$6F,$E3,$8D,$9F,$EF,$8D,$7F,$C1,$2D,$7B,$1F,$E4,$09,$E6
	.dc.b	$D8,$D4,$62,$3F,$CA,$3E,$47,$F8,$C5,$FF,$28,$E3,$FE,$A9,$E2,$EC
	.dc.b	$D8,$CC,$ED,$E3,$3A,$8C,$61,$FB,$AA,$E1,$5A,$3A,$F9,$6F,$8F,$FC
	.dc.b	$03,$63,$FC,$E0,$D7,$FC,$94,$F9,$F7,$FF,$24,$B9,$6C,$F2,$39,$E7
	.dc.b	$F2,$FE,$7B,$F8,$5F,$8D,$FE,$99,$F1,$5F,$15,$F8,$F6,$E5,$FE,$30
	.dc.b	$AD,$E9,$FE,$29,$7F,$DF,$51,$FE,$3F,$6E,$E6,$92,$3F,$A6,$B5,$FE
	.dc.b	$92,$FE,$21,$FF,$43,$7C,$F5,$7F,$D0,$6D,$DB,$FD,$90,$DF,$18,$D2
	.dc.b	$4F,$FB,$49,$AF,$FB,$43,$81,$F3,$DC,$DB,$1F,$ED,$F6,$BB,$8D,$11
	.dc.b	$D3,$5F,$F6,$DF,$90,$F2,$BC,$CD,$8F,$F7,$53,$5D,$C7,$C8,$E7,$7C
	.dc.b	$4D,$24,$FC,$18,$CB,$63,$2F,$C7,$8C,$B6,$3E,$37,$2F,$D8,$8F,$F7
	.dc.b	$7B,$5F,$C8,$35,$FE,$04,$D7,$F7,$6D,$D0,$D2,$70,$FD,$3F,$4D,$F3
	.dc.b	$D4,$97,$F6,$ED,$3A,$8C,$35,$FE,$0F,$1A,$9E,$31,$89,$D3,$FD,$83
	.dc.b	$A9,$F3,$DF,$02,$30,$D7,$FC,$01,$1A,$CB,$F1,$4D,$8F,$F0,$EF,$C9
	.dc.b	$FE,$58,$8E,$08,$7F,$87,$1C,$B4,$FC,$B7,$82,$F9,$E7,$AD,$F9,$78
	.dc.b	$C3,$63,$33,$C1,$AF,$F2,$69,$7C,$E3,$E0,$F9,$36,$7B,$B5,$E0,$D7
	.dc.b	$FC,$07,$2F,$29,$F3,$DE,$36,$CF,$A1,$AF,$85,$AF,$2F,$9E,$7C,$2F
	.dc.b	$AD,$B1,$F0,$1A,$FA,$5A,$F8,$B0,$7C,$17,$A8,$91,$EA,$6B,$EA,$6B
	.dc.b	$C5,$FD,$F7,$CF,$36,$A4,$9F,$C6,$46,$5F,$A7,$F3,$CF,$8F,$FA,$4D
	.dc.b	$8C,$C6,$08,$E0,$87,$D3,$8E,$0F,$E5,$25,$F2,$1E,$42,$47,$89,$EF
	.dc.b	$F0,$92,$3C,$6D,$7A,$35,$FA,$F4,$8C,$47,$F7,$3F,$8C,$7C,$C7,$F1
	.dc.b	$F6,$C7,$EA,$DA,$FE,$A1,$AF,$F9,$13,$E4,$5F,$3A,$F9,$EF,$E5,$51
	.dc.b	$86,$C6,$67,$F9,$84,$E5,$B1,$97,$FB,$DC,$65,$B1,$97,$C0,$D7,$D5
	.dc.b	$E8,$FA,$48,$FA,$5F,$0F,$CF,$48,$EE,$D7,$AB,$5E,$AD,$7F,$C3,$1F
	.dc.b	$28,$F1,$FF,$A2,$36,$3F,$88,$35,$FE,$1A,$D7,$E3,$78,$AF,$9E,$F8
	.dc.b	$AD,$8F,$86,$D7,$E0,$B5,$E3,$1C,$08,$F0,$7F,$AB,$7C,$F3,$E5,$29
	.dc.b	$27,$F9,$24,$65,$EF,$F0,$7E,$22,$38,$3E,$AC,$7F,$7F,$7E,$F3,$5F
	.dc.b	$F4,$F6,$A8,$C4,$23,$1E,$A7,$E0,$A7,$31,$89,$EB,$39,$3F,$B7,$7C
	.dc.b	$F7,$EB,$63,$F9,$AB,$5F,$F0,$C3,$5F,$F2,$53,$5F,$F2,$1B,$7C,$83
	.dc.b	$49,$D3,$FE,$08,$F9,$EF,$CC,$D7,$FA,$71,$AF,$FA,$E5,$A9,$A4,$E5
	.dc.b	$E4,$48,$FF,$04,$7C,$F7,$F5,$B4,$D2,$79,$7F,$E1,$8F,$5F,$FD,$97
	.dc.b	$3F,$E5,$86,$BF,$EA,$E6,$BF,$EC,$B6,$BF,$E5,$2F,$80,$E4,$47,$FC
	.dc.b	$31,$1A,$F9,$17,$CF,$3D,$AF,$F0,$E4,$61,$B1,$9A,$FF,$11,$4E,$5B
	.dc.b	$EF,$BE,$7B,$8B,$63,$31,$83,$FC,$66,$BD,$DA,$47,$FC,$6E,$FF,$75
	.dc.b	$FC,$2D,$F3,$6F,$7B,$FD,$D1,$9D,$BF,$F3,$ED,$6A,$72,$7F,$BB,$7E
	.dc.b	$15,$F8,$67,$F6,$28,$CD,$70,$C3,$96,$1C,$F9,$98,$F3,$AF,$89,$7F
	.dc.b	$E9,$38,$C5,$78,$71,$EB,$E9,$F2,$AD,$63,$70,$FC,$F7,$D0,$7F,$8A
	.dc.b	$A7,$F8,$DB,$63,$F8,$CB,$5F,$E4,$CD,$43,$E8,$5F,$64,$F8,$CE,$D1
	.dc.b	$87,$C1,$DA,$BE,$DA,$7C,$7A,$D6,$3F,$09,$B8,$FB,$8E,$D8,$F5,$9F
	.dc.b	$8F,$87,$3F,$67,$BC,$E6,$B8,$7D,$3F,$E2,$79,$FE,$03,$70,$F9,$EF
	.dc.b	$E3,$7B,$F3,$9F,$FA,$9A,$72,$DB,$E6,$D3,$EF,$57,$19,$DD,$F6,$CF
	.dc.b	$FA,$9C,$76,$6D,$FA,$FD,$A7,$DD,$3E,$FB,$87,$7C,$C3,$83,$F5,$EB
	.dc.b	$73,$9B,$E5,$F8,$2B,$E3,$FF,$21,$1F,$E3,$F8,$C6,$1F,$E8,$5B,$46
	.dc.b	$27,$FD,$9F,$18,$AE,$31,$F7,$34,$7D,$5B,$E7,$7E,$4B,$71,$FC,$BE
	.dc.b	$1C,$5F,$97,$19,$BF,$B1,$D4,$F3,$3F,$C4,$B5,$FD,$3B,$DB,$EF,$5F
	.dc.b	$EE,$29,$CB,$FD,$42,$33,$5C,$5B,$3A,$9F,$F6,$13,$5F,$EB,$B1,$D3
	.dc.b	$2F,$4F,$BF,$E8,$3F,$41,$7C,$F0,$F3,$C7,$E1,$21,$E8,$71,$6C,$FF
	.dc.b	$AB,$6B,$FD,$EF,$39,$24,$66,$77,$87,$FB,$EA,$B7,$39,$AF,$27,$D6
	.dc.b	$BE,$43,$F2,$6D,$9F,$C9,$E1,$CB,$9B,$F1,$B3,$9C,$7E,$5F,$AD,$FD
	.dc.b	$96,$FF,$C3,$9E,$87,$C6,$8C,$4F,$F8,$76,$30,$49,$D9,$23,$CD,$4F
	.dc.b	$1B,$C6,$7C,$6B,$E7,$87,$8A,$3F,$53,$F8,$6F,$FA,$63,$0F,$88,$DC
	.dc.b	$3E,$03,$5F,$E0,$A3,$54,$66,$3F,$1C,$D7,$4D,$8D,$FC,$37,$DB,$8C
	.dc.b	$7D,$87,$FB,$BA,$FF,$72,$DB,$FD,$FC,$F8,$F1,$FD,$AC,$E5,$FE,$55
	.dc.b	$EF,$FF,$19,$A6,$DF,$E7,$0F,$03,$F1,$F3,$98,$C6,$3F,$EB,$38,$C1
	.dc.b	$27,$67,$E3,$F0,$77,$FF,$0E,$6D,$7F,$3C,$BE,$78,$7F,$14,$9C,$BF
	.dc.b	$A6,$FC,$39,$F6,$1F,$A0,$6D,$FF,$B0,$DB,$3F,$EC,$06,$BF,$3A,$D1
	.dc.b	$88,$FE,$6B,$18,$AE,$84,$AF,$CE,$7D,$3B,$EE,$BF,$92,$B6,$3D,$18
	.dc.b	$78,$35,$FE,$3F,$39,$C7,$F9,$A7,$C4,$3E,$9F,$F8,$05,$7F,$61,$8C
	.dc.b	$77,$7E,$C3,$F0,$F7,$7F,$F5,$19,$CE,$7F,$D8,$17,$D4,$3D,$5F,$E0
	.dc.b	$B1,$86,$C6,$67,$FD,$96,$DC,$FF,$61,$F7,$3F,$CB,$B3,$FE,$6A,$DA
	.dc.b	$F8,$7F,$12,$F9,$4F,$E8,$73,$9B,$FC,$1F,$0F,$FB,$0A,$FF,$DE,$4D
	.dc.b	$8F,$F7,$7B,$5F,$A0,$D7,$E7,$B5,$E7,$F1,$BF,$C6,$A3,$1C,$9F,$3E
	.dc.b	$3F,$8B,$79,$3F,$DE,$2F,$F9,$83,$6F,$8C,$7F,$56,$8C,$76,$3E,$BF
	.dc.b	$EA,$4F,$F3,$76,$D7,$EB,$DF,$DC,$C6,$79,$FF,$B0,$C6,$AB,$94,$63
	.dc.b	$C1,$CB,$FD,$9B,$ED,$9F,$0D,$4C,$76,$DC,$7F,$B4,$CE,$9B,$3A,$69
	.dc.b	$FA,$1D,$9F,$5B,$FC,$C9,$B1,$FE,$20,$D1,$F3,$7F,$C8,$63,$E5,$D6
	.dc.b	$C8,$EE,$35,$3F,$62,$7F,$BE,$FD,$BF,$FA,$56,$BF,$BE,$B6,$7F,$A9
	.dc.b	$DF,$81,$FF,$02,$C6,$79,$7F,$BD,$FD,$2B,$E9,$BF,$C0,$ED,$C3,$FC
	.dc.b	$0E,$FF,$95,$1B,$1A,$23,$AF,$81,$B8,$7F,$84,$BE,$E1,$F4,$1F,$E1
	.dc.b	$26,$D7,$F8,$49,$7C,$B6,$66,$4E,$5E,$11,$8A,$FE,$DB,$F3,$3F,$DB
	.dc.b	$DF,$E0,$4D,$7F,$80,$B5,$FE,$3E,$DE,$93,$49,$E1,$7F,$70,$78,$60
	.dc.b	$FB,$9F,$E2,$FB,$7F,$97,$63,$1E,$47,$F6,$CA,$FE,$03,$1F,$E4,$43
	.dc.b	$27,$E1,$BA,$FF,$18,$7F,$72,$FA,$EF,$F3,$06,$7F,$E3,$6E,$7F,$D9
	.dc.b	$75,$9A,$7E,$D7,$55,$70,$CF,$FC,$63,$3C,$38,$DF,$5B,$F6,$0F,$EF
	.dc.b	$FF,$43,$8D,$3E,$CC,$65,$A7,$F3,$58,$D6,$7F,$E4,$8A,$B8,$7F,$8B
	.dc.b	$17,$C9,$7F,$82,$BA,$9F,$57,$FE,$7F,$F8,$9F,$D3,$E1,$FE,$64,$BD
	.dc.b	$E5,$FD,$06,$F6,$FF,$9A,$2B,$74,$AD,$BF,$E2,$8A,$DB,$FE,$93,$ED
	.dc.b	$FE,$D7,$3A,$CB,$FD,$09,$5A,$BF,$0E,$FF,$AA,$CB,$C9,$87,$F4,$2A
	.dc.b	$D3,$CB,$3F,$54,$EB,$2F,$EC,$7A,$7E,$7D,$F0,$1F,$DE,$73,$7D,$86
	.dc.b	$F2,$56,$F3,$F8,$3C,$3E,$BA,$D4,$E7,$3F,$F2,$66,$1C,$B6,$F8,$51
	.dc.b	$FD,$5A,$75,$97,$7B,$ED,$F5,$9F,$A4,$BD,$CF,$F7,$88,$C3,$E6,$CD
	.dc.b	$1F,$AC,$8C,$3A,$CF,$E2,$B7,$1F,$EC,$AD,$9E,$9D,$3E,$0E,$9F,$89
	.dc.b	$0F,$A6,$FF,$37,$79,$9E,$28,$C7,$BD,$FD,$71,$F8,$FF,$14,$FA,$AE
	.dc.b	$91,$AA,$E1,$7C,$A3,$E6,$99,$38,$CF,$08,$D6,$9F,$86,$6E,$9F,$E2
	.dc.b	$C6,$D3,$0F,$06,$5E,$89,$CC,$F5,$BE,$59,$79,$DF,$CB,$DF,$0D,$BC
	.dc.b	$B3,$9C,$7B,$36,$FF,$D7,$F8,$78,$2F,$84,$7A,$A3,$B9,$27,$84,$EB
	.dc.b	$0F,$0E,$7D,$89,$7F,$03,$2F,$45,$F7,$8F,$46,$3E,$2A,$FC,$0E,$1C
	.dc.b	$E3,$13,$E1,$AE,$18,$F9,$73,$F9,$51,$82,$61,$CA,$32,$F8,$B2,$F2
	.dc.b	$DF,$8F,$3C,$63,$DA,$72,$7C,$8B,$E3,$5E,$1B,$DB,$F4,$E3,$33,$C2
	.dc.b	$B9,$1F,$F0,$06,$32,$E2,$7E,$6E,$1F,$22,$73,$18,$BE,$B9,$7C,$18
	.dc.b	$C1,$E1,$26,$1E,$5C,$FE,$5D,$79,$27,$E5,$9E,$58,$CB,$C3,$1F,$15
	.dc.b	$F5,$8D,$93,$3F,$99,$5C,$30,$F6,$4F,$89,$F0,$67,$DE,$BF,$05,$8F
	.dc.b	$CF,$BE,$18,$FB,$F0,$F1,$E3,$DE,$F7,$1E,$33,$CD,$87,$3C,$FF,$5D
	.dc.b	$1F,$E8,$C3,$F5,$05,$9F,$D6,$36,$36,$FE,$9F,$47,$9F,$C6,$3F,$37
	.dc.b	$18,$71,$5F,$DD,$A7,$51,$FC,$06,$74,$F7,$78,$69,$FB,$56,$E5,$DE
	.dc.b	$BF,$2F,$F5,$5F,$E5,$BB,$FE,$03,$5A,$7C,$33,$83,$7E,$87,$FA,$A3
	.dc.b	$0E,$91,$99,$FD,$75,$71,$D1,$87,$59,$CD,$6E,$F8,$C7,$F1,$49,$D4
	.dc.b	$60,$98,$F9,$E7,$58,$F8,$63,$13,$A9,$F3,$9A,$3B,$46,$88,$FF,$13
	.dc.b	$AE,$F1,$88,$F9,$57,$C1,$D5,$F7,$7A,$DF,$77,$67,$73,$DA,$47,$C9
	.dc.b	$87,$C2,$9E,$53,$C2,$79,$65,$F6,$E3,$94,$63,$2F,$A5,$1B,$C4,$F9
	.dc.b	$0F,$31,$93,$CA,$EB,$0F,$04,$6A,$FB,$5F,$93,$2F,$83,$3C,$08,$73
	.dc.b	$AE,$2F,$AE,$FE,$35,$F2,$7B,$57,$29,$F1,$12,$74,$F9,$CE,$47,$67
	.dc.b	$35,$DA,$FE,$06,$5F,$46,$31,$87,$83,$1F,$87,$7C,$31,$F0,$3D,$B1
	.dc.b	$F0,$4F,$E4,$67,$2F,$02,$61,$F9,$A9,$F8,$91,$87,$E8,$DF,$79,$F1
	.dc.b	$57,$C6,$E1,$7D,$2F,$EF,$B3,$FF,$0D,$B6,$FF,$DB,$6D,$C3,$FD,$F2
	.dc.b	$FF,$03,$9D,$F0,$7E,$2E,$F9,$B6,$78,$47,$6A,$D7,$CD,$7F,$9D,$1F
	.dc.b	$F0,$6F,$3F,$F8,$CF,$CA,$FC,$A3,$63,$FC,$23,$4A,$E1,$1F,$E1,$16
	.dc.b	$E3,$FE,$22,$E8,$61,$C9,$B5,$CA,$B7,$9F,$F8,$D7,$E4,$FD,$1D,$86
	.dc.b	$9F,$9C,$E1,$7D,$47,$E5,$B0,$77,$39,$FF,$C0,$93,$98,$FF,$22,$DF
	.dc.b	$2A,$FF,$08,$CE,$6B,$71,$FE,$55,$9D,$3F,$DE,$E7,$4F,$F9,$07,$1B
	.dc.b	$BF,$F9,$5D,$BA,$7F,$96,$B1,$F1,$53,$7F,$8D,$F5,$1A,$B9,$3F,$EA
	.dc.b	$A7,$AC,$EA,$3F,$D9,$7F,$53,$DB,$C2,$34,$DF,$D4,$7C,$59,$D9,$0F
	.dc.b	$F4,$54,$6B,$5F,$4E,$CF,$84,$F3,$3F,$E9,$56,$D7,$C9,$6F,$33,$EA
	.dc.b	$3E,$BC,$FC,$EA,$FE,$0B,$39,$E5,$FE,$8A,$D9,$FE,$98,$D7,$FA,$B3
	.dc.b	$5F,$B0,$D7,$9D,$76,$72,$D7,$33,$FD,$4B,$85,$E6,$3D,$EF,$A6,$DA
	.dc.b	$FF,$65,$46,$31,$CB,$23,$27,$FB,$3E,$3E,$CF,$93,$23,$FD,$BA,$FD
	.dc.b	$CD,$71,$1F,$EA,$38,$BC,$7F,$EE,$69,$FC,$1F,$A0,$FB,$1A,$4B,$FE
	.dc.b	$57,$5F,$84,$E3,$7E,$14,$F9,$6F,$C0,$B6,$B7,$9F,$F3,$2A,$DD,$68
	.dc.b	$FC,$9C,$FF,$9F,$BE,$A3,$0F,$03,$F2,$17,$CD,$FB,$06,$27,$8F,$90
	.dc.b	$AF,$C7,$78,$52,$B9,$70,$FF,$B4,$63,$F3,$FF,$62,$F5,$B5,$5F,$ED
	.dc.b	$79,$BA,$FE,$82,$D7,$BA,$E7,$ED,$3D,$6F,$D4,$B6,$33,$18,$F5,$F2
	.dc.b	$C8,$C9,$CA,$3F,$63,$C9,$F9,$B8,$F3,$B5,$F7,$E0,$8C,$63,$F4,$77
	.dc.b	$7C,$37,$D0,$3F,$C3,$BC,$3F,$51,$B1,$AC,$3F,$C4,$D3,$96,$C6,$5F
	.dc.b	$F1,$94,$65,$B1,$97,$F8,$B4,$65,$FE,$31,$87,$FA,$97,$EB,$BF,$85
	.dc.b	$F9,$E6,$9F,$C1,$3E,$E9,$EC,$7F,$12,$E1,$7D,$63,$E3,$FF,$C6,$B1
	.dc.b	$86,$C6,$72,$FF,$1F,$4E,$5B,$F0,$6F,$89,$E8,$D8,$FC,$54,$A3,$11
	.dc.b	$F8,$AD,$5F,$88,$7F,$08,$D9,$F1,$2F,$B9,$FE,$4A,$DB,$FC,$5B,$5E
	.dc.b	$5F,$88,$BE,$07,$F9,$0B,$63,$E7,$B5,$F9,$AD,$7E,$5B,$5F,$D3,$35
	.dc.b	$FE,$60,$DC,$4D,$35,$87,$F0,$9F,$BC,$FE,$77,$E2,$1A,$7F,$34,$FB
	.dc.b	$07,$D8,$53,$F0,$EE,$31,$FC,$2B,$5F,$C1,$B5,$FC,$0B,$5F,$DE,$35
	.dc.b	$FE,$8C,$DC,$4E,$70,$FE,$3D,$EE,$F9,$89,$5F,$B6,$F9,$0F,$DF,$75
	.dc.b	$FE,$C5,$FE,$01,$1F,$8E,$6B,$FE,$3C,$A4,$62,$11,$8F,$13,$F9,$74
	.dc.b	$E6,$31,$3F,$E8,$76,$BF,$E9,$16,$BF,$9C,$F9,$3F,$D8,$3F,$C2,$9A
	.dc.b	$FE,$D1,$B7,$FC,$8F,$E4,$FF,$A8,$E6,$5E,$9F,$ED,$1B,$7F,$B2,$6B
	.dc.b	$FB,$06,$BF,$AE,$6B,$FC,$A5,$AF,$F7,$36,$CF,$79,$FE,$99,$7F,$DB
	.dc.b	$AB,$D1,$E5,$FF,$7E,$8C,$47,$F0,$CA,$7D,$7F,$F8,$09,$B1,$F1,$5A
	.dc.b	$FC,$56,$BF,$C0,$1A,$FD,$36,$BF,$49,$B9,$7F,$82,$67,$FB,$04,$63
	.dc.b	$E6,$1F,$79,$FE,$12,$6D,$FF,$38,$6B,$FB,$7F,$AC,$7C,$7F,$F8,$39
	.dc.b	$B1,$FC,$71,$AF,$F1,$A6,$BF,$C6,$1A,$FF,$53,$6B,$FE,$1D,$6D,$9A
	.dc.b	$6B,$15,$BE,$4B,$D0,$4B,$FE,$B1,$EA,$BE,$6F,$FA,$4C,$61,$AF,$F8
	.dc.b	$29,$AF,$F8,$29,$AF,$F8,$51,$AF,$E7,$3E,$17,$F9,$63,$FD,$71,$AF
	.dc.b	$F4,$66,$F1,$7F,$8E,$2B,$F9,$97,$CC,$FF,$8F,$A7,$3E,$9F,$EA,$5B
	.dc.b	$7D,$7C,$8F,$F2,$2D,$F0,$8F,$ED,$3F,$4E,$FA,$DF,$F2,$3B,$63,$31
	.dc.b	$8F,$B6,$CB,$23,$27,$28,$FF,$28,$FD,$0B,$EE,$DD,$B8,$6D,$7F,$80
	.dc.b	$67,$53,$FE,$06,$FA,$3F,$F7,$26,$FF,$E5,$49,$CD,$6E,$F8,$BF,$E0
	.dc.b	$68,$CB,$FE,$08,$15,$CB,$6F,$F3,$83,$FE,$EF,$8F,$F3,$91,$AF,$8D
	.dc.b	$AE,$9A,$FF,$B3,$23,$83,$FE,$1D,$CF,$FC,$0C,$FF,$4D,$6E,$3F,$E2
	.dc.b	$76,$DF,$F9,$FF,$D5,$7C,$27,$F9,$A2,$73,$7F,$E6,$F6,$F4,$7D,$1F
	.dc.b	$67,$FC,$4F,$7C,$2B,$FC,$4C,$3E,$AF,$FC,$29,$97,$99,$B1,$EA,$6B
	.dc.b	$EE,$6B,$FE,$91,$5F,$40,$E2,$E4,$D9,$E4,$D7,$93,$5F,$8F,$F1,$9F
	.dc.b	$19,$FF,$78,$B5,$34,$F9,$BC,$CF,$AB,$FF,$25,$6F,$86,$D7,$D4,$9D
	.dc.b	$78,$DF,$E9,$48,$FF,$54,$46,$2B,$84,$E6,$7F,$DF,$95,$C9,$DB,$5C
	.dc.b	$C6,$D0,$FC,$4B,$F8,$AD,$7E,$01,$B1,$F8,$46,$BF,$88,$6B,$CB,$EA
	.dc.b	$1F,$53,$FE,$C8,$AF,$C5,$C6,$3E,$1B,$C0,$D8,$CC,$EE,$BF,$27,$39
	.dc.b	$6C,$65,$FF,$64,$46,$5B,$19,$7F,$B3,$6F,$FE,$30,$AF,$EC,$0D,$8F
	.dc.b	$EC,$0D,$7F,$C7,$6D,$78,$FB,$3F,$17,$C2,$FF,$6A,$B6,$75,$3F,$DC
	.dc.b	$1B,$DA,$FA,$87,$D4,$7C,$D8,$C3,$63,$38,$FF,$AB,$9B,$7F,$EA,$E6
	.dc.b	$BF,$EA,$E6,$BF,$EB,$86,$AA,$EB,$3F,$EB,$8F,$C4,$FF,$D8,$CD,$F8
	.dc.b	$BF,$14,$62,$BF,$D7,$6D,$7F,$99,$7C,$BF,$9F,$C5,$34,$97,$F6,$BE
	.dc.b	$F3,$FC,$B0,$DF,$7F,$ED,$C6,$2F,$EC,$B5,$FA,$ED,$7F,$0C,$D5,$C2
	.dc.b	$8C,$36,$7F,$2D,$DA,$FA,$AE,$0D,$8C,$C6,$0E,$9A,$BF,$00,$FF,$5E
	.dc.b	$65,$FE,$DC,$6E,$5F,$ED,$C6,$BF,$ED,$C6,$BF,$EE,$66,$BF,$3D,$B3
	.dc.b	$D7,$E3,$3E,$F4,$6B,$91,$97,$CD,$BF,$F0,$37,$07,$E2,$E3,$11,$FE
	.dc.b	$07,$8F,$D1,$F9,$0F,$AE,$A4,$CF,$F5,$31,$96,$C6,$6B,$F4,$F3,$96
	.dc.b	$C7,$C8,$F5,$7F,$C1,$35,$FE,$31,$F4,$30,$47,$FD,$4B,$F8,$5B,$C9
	.dc.b	$F9,$ED,$8C,$C6,$08,$E0,$98,$7F,$AA,$A7,$83,$FE,$4D,$F5,$FF,$D6
	.dc.b	$F3,$FE,$51,$6C,$67,$E2,$3F,$5D,$18,$8F,$E2,$63,$E0,$FF,$D2,$BC
	.dc.b	$9F,$C5,$DB,$7F,$EA,$4F,$80,$FC,$54,$62,$BF,$AD,$B5,$E6,$D7,$F8
	.dc.b	$FB,$5E,$D1,$F8,$CF,$84,$FE,$15,$1F,$69,$BC,$3F,$EA,$FF,$07,$F2
	.dc.b	$71,$8F,$B1,$7A,$9F,$96,$6E,$5F,$ED,$06,$E1,$FE,$CE,$6B,$FE,$CC
	.dc.b	$6B,$FC,$CB,$04,$62,$3F,$16,$D8,$CD,$7E,$93,$EB,$5F,$51,$49,$87
	.dc.b	$FB,$82,$32,$FF,$B8,$BE,$B3,$F8,$43,$7A,$FF,$5B,$18,$CB,$FD,$C6
	.dc.b	$D7,$E4,$FD,$6B,$EA,$7E,$93,$6F,$FD,$AA,$D9,$FA,$2D,$7F,$DD,$3F
	.dc.b	$15,$FE,$49,$9F,$F7,$9B,$63,$FD,$E2,$D7,$7E,$DF,$F5,$55,$F8,$87
	.dc.b	$C3,$6E,$B9,$8C,$11,$C1,$2F,$FC,$33,$1F,$27,$D5,$7D,$67,$F0,$C6
	.dc.b	$CF,$F8,$83,$81,$18,$C3,$FC,$4B,$18,$AF,$E8,$9F,$27,$FA,$C9,$FF
	.dc.b	$55,$35,$FC,$CF,$82,$8C,$47,$FA,$E5,$AF,$FA,$E1,$B1,$BE,$DE,$A4
	.dc.b	$C7,$FD,$71,$F3,$9F,$94,$93,$49,$CF,$FD,$5D,$B3,$C2,$30,$7C,$BF
	.dc.b	$A0,$FE,$23,$89,$A4,$F0,$F4,$4A,$FF,$20,$C6,$5F,$8D,$AB,$EB,$3E
	.dc.b	$51,$23,$2F,$F6,$9F,$9F,$7B,$74,$93,$FE,$D3,$8C,$B6,$33,$8F,$FB
	.dc.b	$56,$72,$D8,$FE,$E9,$F3,$9F,$E6,$AA,$FF,$6E,$36,$7F,$DB,$8D,$7E
	.dc.b	$9B,$5F,$F7,$53,$5F,$F7,$4B,$75,$34,$D4,$2D,$F6,$3F,$95,$6D,$FF
	.dc.b	$99,$9B,$3F,$E6,$A6,$BF,$E6,$E6,$BF,$C7,$9A,$FF,$8C,$9A,$F7,$9F
	.dc.b	$EC,$3F,$5B,$FA,$B8,$FE,$0C,$D7,$FA,$7B,$78,$CD,$FA,$DF,$F2,$04
	.dc.b	$FF,$54,$6C,$9A,$4F,$67,$FB,$7B,$71,$FD,$23,$C3,$EA,$9F,$5B,$FD
	.dc.b	$D1,$B3,$FA,$96,$E1,$FA,$86,$BF,$CF,$30,$46,$23,$F9,$03,$5F,$ED
	.dc.b	$2D,$8D,$D7,$CF,$FA,$D9,$7F,$3D,$B7,$F3,$DA,$FC,$F6,$BF,$3D,$47
	.dc.b	$F9,$83,$D9,$B3,$FD,$E3,$8D,$F5,$B8,$F8,$07,$C8,$F1,$C3,$81,$FC
	.dc.b	$DE,$B6,$FF,$A0,$23,$D7,$F3,$FD,$23,$E0,$6A,$8C,$63,$A9,$FF,$38
	.dc.b	$9A,$8C,$C7,$C2,$6E,$BF,$21,$B5,$CE,$B8,$D7,$33,$FD,$0D,$4F,$A2
	.dc.b	$FD,$26,$5E,$A9,$FF,$0A,$C6,$31,$FE,$A0,$BE,$B3,$E0,$46,$3E,$36
	.dc.b	$D7,$06,$E1,$D5,$B7,$C7,$1F,$85,$5B,$6A,$FA,$9F,$F2,$6F,$93,$86
	.dc.b	$C7,$FA,$9B,$1F,$F4,$9F,$C3,$3E,$83,$CD,$3F,$E9,$6A,$DE,$BE,$B8
	.dc.b	$E9,$EF,$FF,$98,$DE,$46,$DC,$47,$C1,$BE,$9A,$7B,$2B,$8E,$5E,$3B
	.dc.b	$DE,$5F,$E1,$2C,$7B,$BE,$4C,$BC,$77,$C0,$EE,$5C,$BC,$95,$C2,$19
	.dc.b	$78,$33,$ED,$94,$9E,$35,$BA,$D4,$63,$2F,$1B,$EB,$AE,$18,$7A,$F1
	.dc.b	$F1,$E1,$DE,$BC,$59,$7A,$70,$E3,$8F,$4C,$3C,$B1,$D2,$B6,$75,$3E
	.dc.b	$4C,$EE,$B9,$63,$E0,$6C,$EB,$1F,$0D,$F3,$C7,$C4,$7A,$E3,$37,$C6
	.dc.b	$BB,$63,$E9,$CF,$E9,$5F,$0C,$39,$1E,$8C,$38,$DF,$C3,$C3,$96,$5E
	.dc.b	$58,$E7,$97,$93,$2F,$6D,$7E,$4A,$B5,$5E,$E2,$63,$E5,$CB,$C3,$97
	.dc.b	$82,$FA,$E5,$E0,$C7,$9E,$1E,$13,$E0,$5F,$0A,$F6,$47,$CA,$C7,$BE
	.dc.b	$3E,$B9,$E7,$5B,$AE,$57,$D0,$F8,$59,$F2,$BE,$13,$DA,$31,$97,$96
	.dc.b	$30,$72,$CF,$E0,$11,$FC,$7E,$7E,$98,$C4,$79,$63,$0F,$DF,$3F,$CD
	.dc.b	$E7,$E6,$CB,$7F,$B4,$6D,$7E,$B5,$CE,$DF,$E7,$57,$89,$FB,$D7,$0E
	.dc.b	$BD,$2C,$B9,$6D,$7F,$0F,$7F,$81,$4F,$3C,$BF,$52,$DC,$7F,$81,$FD
	.dc.b	$67,$FA,$57,$0D,$DC,$7F,$53,$5B,$7F,$5F,$3B,$73,$3B,$BE,$AF,$E1
	.dc.b	$67,$51,$89,$D9,$1F,$E1,$11,$AB,$FE,$09,$7D,$EB,$D8,$6E,$3C,$FE
	.dc.b	$C7,$FA,$6A,$FA,$12,$38,$E1,$E1,$9D,$46,$2B,$91,$FC,$7E,$FD,$9B
	.dc.b	$B1,$F1,$4E,$70,$F6,$5F,$1C,$7C,$F8,$F4,$AD,$1D,$0F,$3C,$69,$E2
	.dc.b	$69,$F3,$3F,$2A,$39,$3C,$27,$B1,$1D,$1C,$8D,$1C,$DF,$86,$E7,$2F
	.dc.b	$C0,$3E,$EC,$7D,$95,$DA,$73,$7E,$2A,$E7,$5A,$79,$65,$F4,$6B,$78
	.dc.b	$76,$9F,$74,$61,$F1,$1E,$63,$33,$C4,$F9,$B9,$7C,$C8,$F4,$92,$79
	.dc.b	$5F,$86,$38,$E5,$F2,$23,$18,$F9,$EB,$C1,$5E,$8B,$E9,$18,$26,$1E
	.dc.b	$0C,$BE,$03,$E3,$C7,$A5,$70,$8C,$5F,$C7,$8C,$4F,$C2,$3F,$15,$1F
	.dc.b	$DC,$38,$5F,$59,$FB,$7C,$BF,$C0,$11,$8E,$1F,$C2,$36,$B8,$57,$F0
	.dc.b	$2A,$DB,$C7,$5F,$F6,$A3,$65,$5C,$5F,$ED,$EF,$81,$EB,$F5,$9F,$E0
	.dc.b	$B7,$FC,$2A,$70,$70,$FF,$85,$9B,$C3,$FE,$45,$6C,$7F,$B8,$DA,$F1
	.dc.b	$D0,$F4,$A4,$E9,$B3,$A9,$FF,$1E,$9F,$E0,$9F,$AC,$C7,$4F,$F9,$02
	.dc.b	$BF,$D4,$5F,$57,$F1,$3A,$4F,$E4,$EB,$79,$7F,$80,$DF,$E3,$78,$7F
	.dc.b	$86,$DB,$C4,$69,$23,$F8,$FB,$5F,$E4,$F1,$B7,$FD,$A1,$9B,$E4,$7F
	.dc.b	$93,$E5,$EC,$6C,$66,$77,$18,$23,$82,$1F,$ED,$8F,$AC,$F5,$BF,$ED
	.dc.b	$D6,$B9,$CB,$BC,$62,$3F,$CE,$CD,$7E,$7D,$A3,$11,$F8,$26,$C6,$67
	.dc.b	$F9,$FF,$0B,$EB,$1E,$5E,$1B,$87,$F4,$89,$D4,$E4,$F9,$7F,$69,$27
	.dc.b	$FA,$3B,$5F,$92,$D7,$FA,$BB,$5F,$F5,$EE,$DF,$D4,$3C,$8F,$F5,$DB
	.dc.b	$7C,$8F,$C1,$CF,$FB,$12,$30,$47,$FD,$96,$D7,$FD,$A5,$F3,$DC,$FC
	.dc.b	$93,$49,$1F,$ED,$B6,$BC,$0C,$1F,$D9,$B5,$7D,$4B,$C8,$FB,$ED,$9D
	.dc.b	$35,$FE,$EB,$3C,$27,$F2,$3E,$13,$EB,$29,$2B,$FB,$7C,$65,$B1,$97
	.dc.b	$FB,$BC,$65,$B1,$F1,$FE,$B1,$E7,$D2,$4F,$F7,$ED,$93,$A8,$C3,$5F
	.dc.b	$F0,$5C,$6A,$78,$C6,$27,$4F,$ED,$65,$E9,$FF,$A9,$63,$B3,$5E,$6D
	.dc.b	$78,$35,$FE,$07,$1C,$BC,$8F,$C0,$4E,$65,$39,$7A,$CF,$D0,$F8,$39
	.dc.b	$1F,$41,$AF,$F8,$73,$F0,$FF,$DF,$A6,$91,$FE,$06,$D7,$FC,$3E,$D7
	.dc.b	$EC,$FD,$63,$E4,$FF,$7A,$D8,$FC,$23,$5D,$C7,$F8,$A3,$C5,$7B,$8F
	.dc.b	$83,$C3,$6B,$FC,$63,$3A,$9C,$9F,$C2,$BE,$AF,$FD,$4D,$3F,$EF,$16
	.dc.b	$BF,$BC,$6B,$FC,$F1,$BC,$6F,$A9,$7D,$57,$F1,$46,$D7,$E6,$DA,$FE
	.dc.b	$59,$AF,$FB,$FF,$F0,$37,$D1,$7F,$1C,$6C,$7E,$8D,$AF,$E7,$9A,$F1
	.dc.b	$8E,$04,$78,$3F,$C9,$BE,$A1,$F5,$34,$93,$FA,$F8,$CB,$F1,$BE,$ED
	.dc.b	$F5,$3F,$35,$B1,$98,$C1,$1C,$10,$FF,$33,$EC,$9E,$31,$FA,$0F,$8F
	.dc.b	$FD,$C3,$F9,$D6,$BF,$96,$6B,$F9,$46,$EA,$F9,$07,$D0,$7F,$42,$6D
	.dc.b	$7F,$09,$6B,$FC,$11,$AF,$F5,$F9,$E5,$3C,$49,$1C,$5F,$E9,$BF,$25
	.dc.b	$F7,$89,$E9,$E0,$FF,$3B,$FC,$C3,$EE,$9F,$0D,$86,$E1,$FE,$80,$9D
	.dc.b	$4E,$49,$19,$9D,$BF,$E8,$E9,$DB,$99,$FE,$6D,$8F,$CA,$8F,$E0,$8D
	.dc.b	$7F,$7A,$D7,$F7,$6D,$F2,$DF,$2E,$F8,$7F,$EA,$2D,$AF,$E9,$CD,$7F
	.dc.b	$D3,$F3,$C2,$76,$48,$DB,$5F,$E6,$3F,$52,$F8,$AF,$EE,$AD,$8F,$E7
	.dc.b	$8D,$77,$1F,$2B,$E1,$1F,$4B,$49,$3F,$26,$32,$D8,$CB,$F1,$E3,$2D
	.dc.b	$8F,$ED,$BF,$3B,$FC,$CA,$3F,$B5,$35,$F6,$B5,$FF,$11,$62,$8C,$47
	.dc.b	$F8,$AA,$1F,$3C,$F3,$FF,$BC,$B6,$7F,$BB,$35,$F9,$2D,$7A,$F5,$BE
	.dc.b	$43,$FC,$02,$D8,$FE,$FC,$D7,$E8,$B5,$E3,$1C,$08,$F0,$7F,$BD,$7D
	.dc.b	$23,$E7,$29,$27,$FA,$A4,$65,$B1,$97,$FA,$9C,$65,$B1,$FE,$3A,$FA
	.dc.b	$6F,$F6,$34,$7F,$B8,$9A,$FF,$92,$5A,$FF,$94,$1A,$9A,$4F,$85,$E8
	.dc.b	$7F,$91,$BE,$1F,$F9,$B4,$FF,$B9,$9A,$FF,$BC,$1A,$FF,$95,$DA,$FF
	.dc.b	$97,$BE,$43,$FC,$80,$FF,$BB,$9A,$FF,$BE,$1A,$FF,$97,$DA,$F1,$F7
	.dc.b	$BF,$CA,$F1,$88,$34,$C7,$D3,$63,$E1,$3F,$DE,$ED,$9F,$F7,$EB,$5F
	.dc.b	$F3,$60,$70,$FF,$9B,$DF,$E3,$39,$F2,$EE,$34,$93,$FE,$6A,$F8,$7F
	.dc.b	$ED,$A4,$70,$76,$F8,$97,$D3,$7F,$89,$FD,$DF,$0C,$A7,$3E,$DF,$F8
	.dc.b	$F2,$31,$5B,$79,$35,$FF,$42,$4F,$0A,$E6,$D8,$DD,$73,$8C,$47,$C0
	.dc.b	$FA,$67,$D5,$77,$84,$62,$FC,$B1,$89,$D4,$E6,$3F,$91,$7D,$3F,$F9
	.dc.b	$92,$BF,$95,$36,$33,$F0,$7E,$BF,$11,$39,$9F,$F4,$E4,$62,$75,$39
	.dc.b	$9F,$F5,$2C,$E6,$BA,$D7,$C5,$F6,$9C,$3F,$E1,$39,$FF,$1E,$36,$FF
	.dc.b	$C7,$2D,$7F,$C6,$CD,$7F,$C5,$ED,$7F,$C5,$AD,$9E,$BE,$B7,$78,$D7
	.dc.b	$6B,$E9,$9F,$4F,$F5,$E3,$0D,$8C,$E3,$FE,$85,$6E,$3F,$CE,$1A,$FF
	.dc.b	$1C,$6B,$FC,$6D,$AA,$BA,$CF,$F1,$AF,$23,$C1,$DF,$FD,$72,$30,$4A
	.dc.b	$FE,$53,$D5,$E5,$F0,$4D,$24,$FF,$61,$FA,$77,$D2,$7D,$76,$C6,$63
	.dc.b	$04,$70,$4B,$FE,$C9,$3C,$1F,$F3,$FF,$D3,$D7,$FB,$E3,$5F,$F5,$9B
	.dc.b	$5F,$F7,$9B,$73,$FE,$C7,$F4,$FF,$37,$CB,$34,$98,$7F,$64,$6D,$FF
	.dc.b	$64,$9D,$CE,$4F,$F4,$07,$99,$F1,$E7,$FC,$FA,$D7,$FB,$BE,$DF,$8D
	.dc.b	$8C,$47,$E4,$BC,$47,$D3,$FE,$6D,$B7,$FE,$09,$6C,$FF,$81,$5A,$FE
	.dc.b	$79,$AF,$39,$E4,$49,$FF,$56,$7A,$1F,$DD,$5B,$E6,$7F,$03,$C1,$39
	.dc.b	$BF,$F5,$5B,$5F,$F1,$47,$A2,$F1,$F8,$37,$0F,$F1,$84,$60,$8E,$09
	.dc.b	$3F,$E3,$38,$FD,$77,$A4,$F8,$0A,$4A,$FF,$55,$46,$5F,$F3,$CF,$D3
	.dc.b	$FD,$BF,$58,$D2,$72,$FF,$1D,$8C,$B6,$33,$3C,$A3,$E0,$7B,$DF,$C0
	.dc.b	$F2,$6B,$FE,$89,$EF,$C9,$2F,$FD,$A9,$F4,$FF,$DE,$5B,$E7,$FC,$98
	.dc.b	$C5,$FF,$B4,$9A,$FF,$93,$DA,$8E,$EE,$1B,$1C,$CF,$A9,$F4,$CF,$A7
	.dc.b	$A4,$BF,$F7,$34,$65,$FF,$74,$7D,$3F,$EE,$DB,$D5,$F7,$63,$18,$7F
	.dc.b	$B9,$9A,$F8,$C6,$EF,$A6,$7D,$3F,$CB,$6E,$1F,$EA,$A6,$D7,$F9,$F9
	.dc.b	$AF,$F9,$B7,$E1,$5F,$4F,$F8,$D6,$D7,$FB,$F5,$B3,$FE,$F8,$6B,$FD
	.dc.b	$A3,$E9,$FE,$22,$69,$36,$FE,$98,$DA,$FE,$96,$D9,$FE,$E3,$C3,$FE
	.dc.b	$F6,$FA,$7F,$CF,$7C,$71,$A4,$BF,$F7,$5F,$67,$F4,$1F,$8D,$FD,$06
	.dc.b	$1F,$EE,$86,$BF,$EE,$66,$F9,$9F,$EF,$49,$E0,$4A,$E6,$7F,$72,$87
	.dc.b	$D3,$FE,$3D,$BC,$FF,$D3,$E3,$04,$70,$49,$E5,$1F,$97,$C1,$F3,$9F
	.dc.b	$C2,$9F,$C6,$35,$CF,$87,$FD,$C6,$30,$47,$04,$AF,$F0,$3E,$AD,$7F
	.dc.b	$D0,$D1,$86,$DF,$CC,$6B,$FE,$AB,$6B,$F9,$DC,$BF,$CC,$DE,$D7,$F6
	.dc.b	$BF,$7B,$FC,$57,$7F,$0D,$B5,$FE,$26,$AD,$CE,$6B,$81,$1C,$1F,$2F
	.dc.b	$E5,$5F,$43,$FD,$DD,$B1,$98,$C1,$1C,$12,$FC,$13,$C1,$FF,$13,$E0
	.dc.b	$FA,$7F,$F0,$C3,$6B,$FC,$2E,$D9,$DC,$E8,$93,$A6,$C7,$E4,$BE,$9F
	.dc.b	$F4,$AF,$F0,$1D,$11,$88,$FE,$22,$D8,$CF,$4F,$F3,$5F,$A7,$FF,$1A
	.dc.b	$57,$16,$BF,$C6,$1B,$26,$93,$DB,$E4,$4C,$7E,$FF,$D2,$BE,$9F,$FA
	.dc.b	$13,$63,$31,$82,$38,$26,$1F,$AE,$9E,$0F,$CA,$F8,$7F,$92,$49,$F9
	.dc.b	$2D,$8D,$46,$08,$E0,$FF,$18,$BF,$1F,$91,$F4,$E8,$FE,$55,$F0,$2F
	.dc.b	$8C,$FE,$35,$C2,$C7,$58,$F2,$3F,$9E,$E3,$CB,$0E,$37,$F4,$F2,$7C
	.dc.b	$BF,$09,$FE,$49,$BA,$73,$1F,$CE,$4C,$B8,$C3,$B9,$FD,$29,$C4,$6A
	.dc.b	$BA,$4F,$76,$BB,$6D,$77,$6F,$CD,$7E,$57,$4F,$75,$7F,$56,$8C,$63
	.dc.b	$F8,$B5,$F4,$9F,$02,$B7,$82,$77,$7C,$79,$B5,$87,$16,$C7,$5C,$2D
	.dc.b	$F1,$9C,$CB,$E8,$FF,$C7,$FE,$47,$CA,$8F,$EB,$91,$85,$97,$A3,$E1
	.dc.b	$5F,$4F,$E2,$9E,$5E,$E7,$F9,$45,$DE,$9F,$D4,$F0,$E5,$87,$C0,$9D
	.dc.b	$E3,$DF,$1E,$79,$79,$2F,$78,$71,$C7,$A6,$3C,$67,$38,$FF,$A8,$F2
	.dc.b	$F1,$46,$23,$A9,$E2,$3D,$75,$CA,$FF,$BC,$63,$E0,$DF,$E0,$57,$0C
	.dc.b	$7B,$5E,$E3,$8E,$7E,$7C,$BB,$CE,$67,$86,$3E,$7A,$D6,$7E,$BA,$D6
	.dc.b	$7E,$53,$D3,$9F,$87,$0F,$31,$C8,$EE,$62,$37,$8F,$83,$0F,$3C,$73
	.dc.b	$BF,$35,$F3,$CB,$D3,$5A,$9F,$09,$AB,$E5,$87,$58,$CC,$FA,$33,$E0
	.dc.b	$C9,$D3,$96,$92,$FB,$BF,$3F,$0F,$16,$7F,$0E,$34,$D9,$FC,$56,$3E
	.dc.b	$27,$89,$23,$EA,$C6,$5E,$04,$9F,$C0,$67,$F3,$70,$F3,$5F,$19,$CE
	.dc.b	$38,$BE,$11,$F3,$71,$F2,$BF,$3A,$B8,$39,$C7,$C0,$E6,$BC,$15,$C2
	.dc.b	$3E,$B1,$33,$FA,$B9,$78,$E3,$DE,$71,$3C,$66,$0E,$19,$FD,$38,$E1
	.dc.b	$8F,$8A,$B9,$E5,$E9,$BF,$C1,$9F,$E3,$D3,$F3,$A5,$8F,$DB,$36,$FF
	.dc.b	$64,$E7,$5F,$F2,$1B,$C4,$FE,$00,$E1,$D7,$95,$FE,$FA,$AF,$E2,$2D
	.dc.b	$9E,$38,$7E,$C5,$B9,$FE,$F5,$B6,$CD,$F4,$FF,$E8,$D8,$FC,$BC,$E9
	.dc.b	$F0,$E1,$83,$F8,$14,$EE,$3F,$41,$3B,$C7,$C7,$18,$AD,$FA,$BF,$E5
	.dc.b	$FC,$BC,$91,$CE,$74,$7E,$4F,$0E,$93,$A8,$CB,$4F,$E3,$91,$AC,$BD
	.dc.b	$95,$CE,$75,$1F,$C2,$27,$55,$FC,$9B,$1F,$31,$2F,$CB,$1C,$6B,$9B
	.dc.b	$E0,$79,$61,$E9,$73,$87,$69,$E9,$5C,$9E,$06,$0B,$97,$CE,$BE,$51
	.dc.b	$9A,$E4,$FC,$E3,$E1,$46,$9C,$46,$B2,$F9,$AF,$C2,$9D,$46,$30,$F2
	.dc.b	$92,$BE,$F6,$5F,$09,$E7,$19,$69,$AA,$E7,$97,$C5,$8F,$4D,$71,$BE
	.dc.b	$D1,$E9,$BE,$D8,$FB,$63,$E5,$19,$23,$97,$E2,$9E,$03,$C2,$E7,$1F
	.dc.b	$76,$1E,$17,$DD,$3D,$0E,$17,$E2,$AE,$17,$EB,$8F,$2C,$F4,$C3,$D1
	.dc.b	$87,$3A,$F3,$C6,$32,$E9,$7C,$E3,$17,$E6,$8C,$65,$F6,$38,$1F,$4F
	.dc.b	$FB,$8D,$3F,$BF,$B7,$2F,$F2,$73,$70,$FF,$6E,$BF,$C0,$E7,$7C,$3C
	.dc.b	$E3,$57,$D2,$31,$5D,$FE,$9B,$FC,$1A,$FF,$88,$4F,$F2,$D3,$87,$FC
	.dc.b	$43,$28,$C7,$07,$F8,$A9,$B3,$C3,$6F,$F6,$84,$E5,$B1,$93,$B1,$FE
	.dc.b	$F4,$9E,$2F,$F8,$3B,$E9,$FF,$DB,$1F,$0B,$FE,$3E,$C3,$FD,$47,$F4
	.dc.b	$DF,$17,$A4,$FE,$52,$B7,$97,$F8,$35,$FE,$39,$87,$33,$F8,$D3,$87
	.dc.b	$FC,$3C,$DE,$03,$49,$1C,$0F,$F5,$4E,$4F,$90,$EB,$97,$C3,$6C,$66
	.dc.b	$77,$18,$23,$82,$1F,$ED,$AF,$A7,$F5,$BF,$EB,$66,$B9,$CB,$BC,$62
	.dc.b	$3F,$CB,$2D,$7F,$D2,$0D,$56,$8C,$36,$7E,$2E,$CF,$A7,$79,$98,$6E
	.dc.b	$1F,$D1,$67,$53,$93,$E4,$7D,$3C,$9F,$E8,$CD,$7F,$A6,$6A,$8C,$47
	.dc.b	$F5,$56,$BF,$EC,$BF,$23,$FC,$DA,$8C,$6F,$F6,$BC,$A4,$EA,$72,$7F
	.dc.b	$B1,$E7,$51,$98,$4F,$09,$C9,$CF,$C2,$7D,$3D,$25,$7F,$B6,$5A,$FF
	.dc.b	$B6,$63,$6E,$48,$7F,$71,$D9,$F4,$EF,$1F,$F0,$AD,$9F,$F7,$8B,$5E
	.dc.b	$0D,$7F,$BD,$78,$4F,$A7,$FC,$B3,$63,$F3,$6D,$7F,$BC,$35,$E3,$1C
	.dc.b	$08,$F0,$7E,$57,$D3,$FF,$C0,$18,$FF,$00,$68,$8C,$47,$FB,$DD,$AF
	.dc.b	$D3,$3E,$5B,$87,$EB,$47,$1F,$3F,$A9,$23,$AC,$ED,$FA,$7E,$9F,$FA
	.dc.b	$C9,$F0,$B5,$FF,$0D,$52,$31,$1E,$18,$C7,$8F,$F8,$39,$CC,$F8,$9A
	.dc.b	$F6,$8D,$1F,$4B,$E0,$A4,$7D,$26,$BF,$E2,$1F,$A7,$FE,$08,$9A,$47
	.dc.b	$F8,$43,$5F,$E1,$AD,$7F,$05,$F4,$EF,$91,$FE,$0A,$D8,$FF,$17,$70
	.dc.b	$23,$11,$C2,$7F,$8B,$F8,$EF,$6D,$F0,$58,$6D,$7F,$8D,$E7,$53,$93
	.dc.b	$F8,$9F,$D3,$FF,$AA,$E7,$FD,$BA,$D7,$F5,$98,$A3,$11,$FC,$19,$B0
	.dc.b	$FA,$67,$D3,$FF,$1E,$6D,$7A,$9A,$FE,$85,$AF,$FB,$B3,$E9,$DF,$3F
	.dc.b	$FC,$99,$B1,$F0,$9A,$FE,$A5,$AF,$18,$E0,$47,$83,$FC,$C3,$E9,$9F
	.dc.b	$4F,$49,$3F,$3A,$32,$FC,$7F,$A6,$7D,$3F,$CE,$6C,$66,$30,$47,$04
	.dc.b	$3F,$CD,$DB,$27,$8C,$7E,$63,$E3,$BF,$74,$FE,$91,$AF,$F1,$26,$F3
	.dc.b	$9C,$EC,$F8,$F7,$D0,$FF,$4C,$6D,$FE,$21,$AF,$F0,$F6,$BF,$B5,$9E
	.dc.b	$53,$C4,$91,$C5,$FE,$AF,$F2,$3F,$78,$9E,$A6,$0F,$EC,$FF,$38,$FA
	.dc.b	$77,$C3,$61,$B8,$7F,$A1,$E7,$53,$92,$46,$67,$6F,$FA,$56,$76,$E6
	.dc.b	$7F,$9E,$63,$F2,$A3,$F7,$CD,$7F,$86,$B5,$FE,$76,$DF,$31,$F2,$CF
	.dc.b	$A7,$FE,$B6,$DA,$FE,$B0,$D7,$FD,$4D,$3C,$27,$64,$8D,$B5,$FE,$79
	.dc.b	$F4,$EF,$8B,$FE,$FE,$D8,$FD,$DB,$5D,$C7,$CA,$F8,$37,$D3,$D2,$4F
	.dc.b	$C9,$8C,$B6,$32,$FC,$78,$CB,$63,$FB,$BF,$CE,$FF,$37,$8F,$EE,$8D
	.dc.b	$7D,$AD,$7F,$9C,$62,$8C,$47,$F6,$D8,$7C,$F3,$CE,$FF,$01,$B6,$7F
	.dc.b	$C0,$0D,$7E,$4B,$5E,$BD,$8F,$91,$FF,$05,$B6,$3F,$C1,$0D,$7E,$8B
	.dc.b	$5E,$31,$C0,$8F,$07,$FC,$01,$F4,$AF,$9C,$A4,$9F,$EB,$31,$96,$C6
	.dc.b	$5F,$E9,$71,$96,$C7,$F9,$07,$E9,$FF,$D9,$91,$FE,$E9,$6B,$FD,$7D
	.dc.b	$AF,$F8,$4D,$A9,$A4,$F8,$4E,$87,$F9,$37,$E1,$FF,$9B,$CF,$FB,$B5
	.dc.b	$AF,$FB,$DD,$AF,$F6,$D6,$BF,$E2,$1F,$91,$FF,$25,$3F,$EF,$56,$BF
	.dc.b	$EF,$F6,$BF,$DD,$5A,$F1,$F7,$FF,$CB,$D1,$88,$34,$C7,$D3,$E3,$E1
	.dc.b	$3F,$CC,$ED,$9F,$F0,$01,$FE,$33,$70,$E1,$AF,$F8,$15,$FE,$33,$9F
	.dc.b	$2E,$E3,$49,$3F,$E7,$1F,$87,$FE,$DE,$47,$07,$F8,$AF,$ED,$DF,$26
	.dc.b	$F7,$7F,$C8,$99,$7F,$A0,$E7,$31,$83,$FA,$EF,$E2,$9F,$DC,$5C,$5F
	.dc.b	$F9,$8A,$76,$AB,$95,$F5,$8C,$57,$F6,$6A,$DD,$7F,$A4,$6F,$A5,$68
	.dc.b	$E9,$7C,$67,$5E,$77,$F9,$57,$BD,$7F,$DA,$B1,$FF,$00,$B7,$0F,$F2
	.dc.b	$43,$5F,$E7,$5D,$FF,$E9,$C9,$F4,$D7,$F1,$EF,$84,$ED,$1E,$EC,$2D
	.dc.b	$FF,$9F,$AF,$83,$E5,$BE,$15,$CA,$73,$8F,$9A,$31,$87,$2F,$23,$E2
	.dc.b	$36,$77,$3F,$E7,$3D,$9F,$8A,$FF,$9E,$E3,$E6,$DF,$06,$D7,$0C,$BF
	.dc.b	$CE,$2D,$BE,$3F,$05,$D7,$1F,$07,$D5,$BC,$8A,$7B,$3E,$39,$F5,$7F
	.dc.b	$C3,$73,$EF,$5B,$6D,$70,$7E,$FD,$70,$ED,$C6,$9F,$5F,$5F,$C5,$7D
	.dc.b	$87,$FA,$25,$B7,$FE,$62,$11,$96,$9F,$E6,$6F,$AE,$FF,$19,$BF,$01
	.dc.b	$B5,$D6,$36,$F7,$8C,$EF,$F8,$DB,$7E,$2C,$ED,$F1,$37,$0F,$F3,$9B
	.dc.b	$70,$FF,$31,$B5,$FF,$31,$35,$FF,$38,$B5,$FF,$4F,$B6,$BB,$D7,$BF
	.dc.b	$C4,$7C,$BF,$DF,$6E,$1F,$6A,$FC,$18,$F9,$E7,$38,$FE,$2F,$F1,$74
	.dc.b	$7E,$75,$B7,$FE,$5C,$6C,$FF,$A8,$DA,$FF,$85,$5A,$FF,$94,$9A,$FF
	.dc.b	$B5,$9B,$CE,$73,$9B,$EC,$5F,$41,$F9,$76,$E1,$FA,$6A,$F0,$E5,$F3
	.dc.b	$E7,$38,$FA,$7F,$15,$F4,$12,$FD,$0D,$8F,$43,$5F,$43,$FE,$CF,$8F
	.dc.b	$F3,$1F,$CA,$FE,$C1,$A7,$86,$31,$D8,$9F,$D6,$7E,$31,$FE,$DB,$AF
	.dc.b	$F5,$7B,$6F,$FD,$3E,$D7,$FD,$3C,$D7,$FD,$5A,$D7,$FD,$B6,$DD,$0D
	.dc.b	$31,$C7,$FC,$37,$3F,$C8,$36,$F9,$DF,$EF,$E3,$0F,$FB,$9B,$E6,$1F
	.dc.b	$79,$FA,$56,$D7,$FB,$19,$B3,$FE,$B6,$6B,$FE,$B5,$6B,$FE,$C4,$6B
	.dc.b	$D6,$7A,$1F,$B1,$F6,$3F,$96,$D7,$FB,$29,$B3,$B8,$FE,$1B,$18,$E8
	.dc.b	$FF,$07,$79,$7F,$F0,$05,$B3,$FD,$5A,$91,$87,$FA,$CB,$5F,$F6,$D3
	.dc.b	$51,$C3,$A7,$FC,$1D,$3C,$E7,$53,$9A,$F0,$7E,$33,$FE,$8A,$AF,$E1
	.dc.b	$ED,$9F,$E8,$95,$E1,$D3,$FA,$62,$9D,$46,$3F,$1A,$7D,$6F,$F1,$F6
	.dc.b	$E1,$FC,$69,$FF,$38,$B6,$3F,$9B,$7E,$36,$F5,$7F,$8D,$36,$7F,$DF
	.dc.b	$6D,$CB,$FD,$D6,$D7,$FD,$D5,$F7,$DF,$CA,$5F,$E4,$8D,$C7,$CF,$F0
	.dc.b	$2F,$94,$FE,$89,$39,$C3,$FD,$93,$F1,$AF,$95,$FE,$BC,$D8,$FF,$6E
	.dc.b	$B7,$1F,$F0,$4F,$72,$31,$1F,$E9,$76,$BD,$2B,$99,$2B,$EE,$74,$3E
	.dc.b	$F3,$F0,$AD,$F3,$BF,$B0,$E3,$D6,$FF,$AD,$56,$BE,$47,$F2,$FB,$7F
	.dc.b	$31,$6E,$3F,$80,$C5,$0F,$A5,$FE,$C4,$EB,$34,$EB,$A7,$26,$AA,$DF
	.dc.b	$F6,$2F,$A7,$7D,$F7,$F6,$B6,$C6,$7C,$7C,$4C,$8E,$0E,$E3,$18,$FF
	.dc.b	$79,$E0,$FA,$0B,$7F,$95,$E5,$7F,$65,$D3,$FC,$27,$B7,$F8,$66,$73
	.dc.b	$97,$F8,$23,$0E,$31,$8C,$3F,$2B,$7C,$2B,$ED,$E1,$FE,$01,$FB,$EF
	.dc.b	$D0,$D7,$F6,$76,$E3,$D7,$1F,$1E,$09,$CE,$1F,$2F,$DE,$FF,$6C,$BF
	.dc.b	$E0,$46,$E1,$FD,$C5,$FE,$F4,$DF,$5F,$FC,$21,$1A,$C7,$FB,$33,$6B
	.dc.b	$A7,$7E,$AB,$FC,$2D,$F3,$6F,$BC,$A6,$3D,$CB,$7F,$E1,$49,$D3,$67
	.dc.b	$4D,$3F,$C5,$BE,$13,$E9,$7F,$C7,$0D,$8F,$E1,$9C,$3F,$E9,$19,$C9
	.dc.b	$23,$33,$BD,$7F,$3E,$D7,$FD,$E8,$D7,$F2,$1F,$8C,$3E,$67,$FC,$60
	.dc.b	$DA,$F1,$DE,$27,$FC,$69,$C7,$FD,$1F,$83,$FC,$14,$DA,$FC,$D3,$FD
	.dc.b	$81,$B8,$70,$23,$FE,$36,$8D,$47,$E7,$1E,$4D,$F0,$7B,$9F,$AC,$E4
	.dc.b	$FE,$D3,$7E,$08,$CD,$F3,$75,$E7,$E2,$FF,$CA,$1A,$38,$3E,$59,$2F
	.dc.b	$E3,$B6,$7E,$3B,$5F,$8E,$FF,$8D,$7F,$15,$7C,$DF,$F9,$55,$B5,$F6
	.dc.b	$27,$13,$9F,$C4,$24,$FF,$95,$2B,$FC,$AD,$7F,$E4,$15,$3A,$8C,$3D
	.dc.b	$7D,$AC,$69,$FE,$37,$F9,$5F,$C9,$57,$F0,$26,$F5,$1F,$CD,$CE,$F9
	.dc.b	$FF,$CD,$51,$A2,$74,$62,$3F,$CD,$DF,$69,$FD,$E4,$98,$7F,$94,$DF
	.dc.b	$EF,$DC,$EF,$4F,$DF,$3F,$E7,$B6,$E3,$FE,$4B,$AF,$B3,$5B,$DF,$AF
	.dc.b	$DD,$CE,$0F,$F4,$4E,$1C,$1F,$E4,$16,$9D,$63,$FE,$65,$8C,$4F,$0B
	.dc.b	$CE,$3D,$FC,$6F,$31,$27,$F9,$AD,$7F,$AB,$2F,$FB,$59,$FC,$A1,$7E
	.dc.b	$0D,$F0,$0F,$83,$56,$BF,$D1,$3E,$D7,$F9,$A2,$F8,$F0,$7F,$99,$7D
	.dc.b	$3F,$F3,$B4,$E6,$7F,$9C,$35,$FE,$6A,$FF,$8B,$EB,$9F,$AB,$FE,$9B
	.dc.b	$8F,$F3,$9E,$7F,$E7,$17,$55,$FE,$83,$6C,$66,$78,$5F,$16,$5F,$43
	.dc.b	$CF,$18,$BE,$5D,$FE,$62,$4F,$F5,$8C,$3F,$88,$CF,$FB,$AB,$1F,$F4
	.dc.b	$07,$3F,$94,$F3,$5D,$BF,$F6,$BD,$3E,$93,$FD,$0D,$ED,$7F,$85,$BE
	.dc.b	$17,$FD,$4D,$3F,$E1,$89,$CA,$8C,$FA,$AF,$AC,$FE,$29,$1C,$27,$A6
	.dc.b	$9F,$E7,$A7,$87,$83,$F6,$30,$E7,$5B,$6C,$7F,$89,$B0,$E9,$E1,$FF
	.dc.b	$B9,$EB,$58,$73,$6E,$5E,$7B,$E9,$39,$79,$65,$E4,$73,$3B,$CB,$D9
	.dc.b	$5B,$6C,$67,$1F,$0E,$7B,$9F,$5C,$DC,$FC,$F1,$C0,$F6,$9F,$E4,$4C
	.dc.b	$7C,$0E,$72,$F3,$BA,$39,$9C,$30,$EE,$E6,$BA,$BA,$9E,$27,$D0,$AE
	.dc.b	$B1,$9C,$BD,$AF,$1C,$BD,$98,$74,$BE,$35,$A6,$D7,$9F,$3F,$9D,$39
	.dc.b	$24,$7A,$5F,$97,$1A,$8C,$10,$F1,$1C,$F3,$F8,$D8,$7C,$D8,$C9,$AC
	.dc.b	$3C,$97,$DA,$3E,$19,$FB,$C2,$9E,$C3,$19,$7E,$49,$BC,$0F,$98,$FF
	.dc.b	$3A,$47,$E6,$E3,$0F,$F9,$DE,$7F,$D4,$D3,$A6,$C6,$67,$79,$FE,$4D
	.dc.b	$B7,$F9,$76,$CF,$FA,$03,$D9,$7C,$57,$F1,$2B,$E1,$8F,$E1,$6F,$77
	.dc.b	$E3,$8F,$C6,$5F,$08,$D4,$47,$F0,$13,$DB,$25,$6E,$73,$5B,$BE,$F1
	.dc.b	$F8,$99,$D4,$60,$98,$FA,$6B,$C3,$8F,$9E,$31,$7D,$09,$5F,$BC,$BE
	.dc.b	$73,$A8,$E7,$3F,$C0,$F1,$F3,$56,$A3,$18,$79,$6F,$81,$FB,$B9,$D4
	.dc.b	$FC,$33,$BC,$69,$F9,$98,$79,$5E,$66,$8B,$87,$96,$35,$1C,$E3,$CA
	.dc.b	$D9,$F0,$4E,$B2,$FB,$F5,$A6,$CE,$AB,$F8,$27,$5B,$E2,$1F,$6F,$FD
	.dc.b	$3B,$9F,$FC,$C1,$18,$AF,$F5,$DC,$62,$75,$C1,$FE,$BE,$BE,$5E,$5E
	.dc.b	$2F,$FC,$D9,$82,$3F,$2F,$3A,$9F,$F2,$0C,$E5,$B3,$F9,$CE,$77,$14
	.dc.b	$9F,$EB,$6F,$58,$CB,$46,$15,$93,$C0,$FE,$C7,$18,$7F,$AB,$45,$CB
	.dc.b	$FA,$D4,$F3,$3C,$51,$D2,$7F,$A8,$46,$0F,$37,$5B,$EA,$BC,$8D,$9F
	.dc.b	$8F,$3F,$E0,$49,$CF,$B5,$27,$FB,$EB,$C8,$8F,$F8,$03,$C0,$E6,$47
	.dc.b	$FC,$1F,$D9,$5F,$F0,$64,$5C,$BF,$D2,$CA,$7E,$96,$DF,$E1,$F6,$C6
	.dc.b	$6B,$99,$E6,$3F,$C6,$5E,$57,$CD,$F1,$3F,$D9,$91,$FE,$C6,$6B,$FD
	.dc.b	$BD,$AF,$F8,$31,$AF,$FB,$05,$7C,$67,$32,$47,$7D,$FF,$C7,$DE,$66
	.dc.b	$21,$9F,$F9,$16,$B0,$FF,$91,$E7,$F0,$1E,$4B,$EB,$E9,$2F,$FC,$A9
	.dc.b	$F5,$95,$FF,$2A,$45,$BF,$B5,$8A,$D7,$D3,$59,$87,$F9,$75,$F9,$3F
	.dc.b	$80,$BD,$DA,$7E,$69,$C4,$FF,$A6,$9A,$FF,$BD,$1A,$FF,$8E,$38,$11
	.dc.b	$88,$FD,$43,$5F,$F6,$93,$63,$73,$FA,$3E,$97,$DD,$7E,$85,$B7,$FE
	.dc.b	$91,$8F,$BF,$F8,$EB,$F1,$5D,$44,$9D,$1F,$7F,$05,$6E,$3E,$A7,$A9
	.dc.b	$F7,$B7,$F8,$ED,$E8,$FF,$56,$F2,$FE,$C6,$1F,$01,$62,$5F,$27,$51
	.dc.b	$BF,$62,$4F,$67,$E6,$F4,$C8,$F9,$4F,$CD,$23,$82,$1F,$35,$CB,$FE
	.dc.b	$41,$F0,$1F,$2F,$FC,$15,$B1,$F8,$F6,$BF,$8E,$C8,$B9,$BF,$C8,$9E
	.dc.b	$AF,$1B,$93,$C5,$E9,$1F,$C4,$7C,$0F,$ED,$F3,$FE,$F1,$F1,$B0,$7F
	.dc.b	$15,$FC,$75,$F5,$78,$9F,$F7,$C4,$D7,$C4,$78,$67,$6F,$E9,$7D,$B7
	.dc.b	$E1,$DF,$C7,$DB,$19,$8F,$BF,$B4,$8F,$B9,$C8,$8C,$47,$6F,$31,$F1
	.dc.b	$AF,$C7,$73,$CF,$FC,$77,$A3,$F0,$78,$63,$8F,$FC,$A9,$7B,$C7,$F7
	.dc.b	$1E,$B7,$E6,$9B,$DD,$FC,$CE,$31,$1E,$CE,$5F,$F6,$6F,$CC,$E6,$A4
	.dc.b	$F0,$39,$C7,$F3,$BF,$84,$74,$D2,$4F,$F0,$78,$7C,$0E,$23,$B4,$70
	.dc.b	$75,$3C,$E7,$FA,$37,$C2,$FE,$D3,$7F,$E0,$2D,$D7,$F0,$39,$A3,$18
	.dc.b	$37,$FD,$2F,$1F,$EB,$0F,$EB,$3C,$7F,$4A,$7F,$C7,$13,$A7,$FC,$74
	.dc.b	$71,$8D,$3F,$C0,$5A,$FF,$67,$6B,$B9,$EA,$7F,$61,$F5,$7E,$84,$7C
	.dc.b	$B6,$E5,$FC,$69,$B9,$7F,$A2,$7C,$3F,$F1,$44,$FF,$1E,$F8,$8F,$E0
	.dc.b	$B7,$FE,$A5,$6B,$FE,$C2,$6B,$F0,$9A,$B9,$9F,$13,$F1,$1B,$7F,$DF
	.dc.b	$5B,$3F,$D2,$5D,$FE,$19,$9F,$8B,$34,$93,$BF,$81,$FF,$6C,$B7,$DF
	.dc.b	$FF,$2D,$DA,$73,$3F,$E6,$58,$C4,$EA,$73,$3F,$EE,$49,$CB,$F4,$9A
	.dc.b	$F1,$8D,$BF,$EA,$F6,$BF,$C0,$7E,$33,$F0,$3C,$66,$98,$F6,$9F,$3B
	.dc.b	$FE,$11,$6D,$7E,$02,$BF,$CB,$33,$FE,$19,$9C,$CF,$FB,$55,$AF,$FB
	.dc.b	$C9,$AF,$E4,$9A,$FF,$13,$6C,$F4,$8E,$DE,$6B,$DC,$FF,$11,$36,$3F
	.dc.b	$B6,$35,$FD,$97,$D2,$7E,$13,$8F,$F3,$EF,$F9,$B6,$3F,$90,$9F,$C8
	.dc.b	$5C,$3A,$F7,$7F,$C6,$51,$88,$D4,$60,$8F,$C1,$8D,$4B,$F1,$CF,$E0
	.dc.b	$7C,$BF,$E2,$98,$C7,$7F,$F9,$15,$B7,$FE,$11,$F8,$07,$DA,$3D,$7F
	.dc.b	$F2,$54,$61,$FF,$4F,$B6,$7F,$5E,$D7,$FD,$54,$D7,$DA,$D7,$F9,$FB
	.dc.b	$53,$4C,$6C,$F9,$3F,$E7,$B9,$7F,$82,$1B,$8F,$F0,$86,$DF,$F0,$7F
	.dc.b	$C7,$9F,$55,$FE,$30,$6C,$7F,$04,$F0,$11,$88,$FE,$12,$D7,$FC,$0C
	.dc.b	$D7,$FC,$D8,$D7,$B4,$FF,$9A,$A3,$13,$D2,$3F,$CD,$19,$3D,$7F,$F3
	.dc.b	$33,$6F,$FC,$60,$D7,$F9,$B7,$90,$F2,$3F,$CD,$8D,$8F,$CE,$35,$FF
	.dc.b	$6B,$6B,$FD,$2A,$7F,$CE,$D3,$99,$FC,$5C,$E5,$E3,$3A,$23,$FA,$7F
	.dc.b	$97,$FD,$F3,$FE,$43,$6B,$FD,$25,$BB,$7F,$8E,$36,$FF,$39,$4F,$F4
	.dc.b	$D6,$BE,$06,$BF,$CF,$1A,$FF,$7D,$6D,$FF,$1D,$F2,$FF,$D1,$5D,$9F
	.dc.b	$CA,$AB,$FC,$04,$D1,$F3,$3F,$E5,$A8,$FF,$2C,$46,$EB,$94,$62,$BF
	.dc.b	$92,$56,$E3,$FB,$17,$CD,$FF,$1A,$C7,$FC,$A2,$DC,$7F,$A8,$B6,$4B
	.dc.b	$F1,$FF,$AD,$5C,$3C,$0F,$F5,$04,$EF,$2D,$C7,$FA,$BE,$38,$70,$7F
	.dc.b	$A0,$3E,$79,$F6,$14,$99,$7F,$90,$7E,$83,$F8,$36,$5F,$E0,$1A,$DC
	.dc.b	$FF,$91,$5A,$FF,$85,$B9,$51,$89,$FF,$04,$57,$2D,$7F,$D8,$4F,$F1
	.dc.b	$1B,$FF,$30,$35,$CF,$7F,$FB,$0A,$31,$1F,$EC,$6F,$A7,$FF,$4B,$FC
	.dc.b	$49,$A7,$FA,$6A,$5F,$1B,$8A,$FF,$2D,$56,$DB,$5B,$9F,$F6,$94,$3E
	.dc.b	$8B,$FD,$27,$97,$F7,$68,$C7,$C3,$7F,$A4,$1F,$EF,$33,$A6,$C6,$5F
	.dc.b	$E4,$E7,$3F,$92,$FA,$EF,$06,$BA,$94,$62,$3F,$97,$C6,$38,$7C,$51
	.dc.b	$FE,$52,$F8,$EA,$9A,$47,$FC,$1F,$A7,$DC,$DB,$F8,$63,$63,$3C,$3F
	.dc.b	$A2,$E1,$4E,$67,$E9,$CE,$63,$11,$F6,$23,$07,$EA,$BA,$3F,$C4,$5F
	.dc.b	$23,$6C,$D2,$61,$E2,$F9,$EE,$84,$70,$4C,$3D,$31,$96,$C7,$F8,$A3
	.dc.b	$37,$D1,$3E,$57,$F7,$71,$86,$C6,$67,$83,$6B,$EA,$D6,$E7,$35,$C0
	.dc.b	$8F,$F8,$17,$AF,$FC,$1C,$DF,$3B,$E0,$46,$2F,$FC,$20,$D7,$FC,$1C
	.dc.b	$D7,$FD,$13,$A7,$F0,$69,$E2,$4A,$FF,$09,$7E,$3D,$F8,$26,$F1,$7C
	.dc.b	$D8,$C5,$FF,$87,$5A,$FF,$1B,$F3,$3F,$8D,$F6,$AF,$FD,$F8,$7F,$83
	.dc.b	$CB,$3F,$CD,$1A,$FF,$88,$FF,$20,$7C,$FF,$DB,$6C,$FF,$89,$5B,$3F
	.dc.b	$E2,$D6,$BC,$67,$81,$27,$83,$E4,$F1,$1F,$29,$49,$3F,$D2,$1F,$E6
	.dc.b	$3F,$04,$F8,$AA,$49,$FD,$94,$65,$B1,$9C,$3F,$CD,$73,$96,$C7,$E7
	.dc.b	$7E,$8B,$FB,$CD,$7F,$6D,$6C,$FF,$9D,$1A,$FF,$BC,$5A,$FF,$88,$3E
	.dc.b	$1D,$F2,$DF,$D1,$9B,$5F,$C1,$5B,$3F,$E7,$C6,$BF,$C0,$9A,$F3,$9E
	.dc.b	$44,$9F,$27,$CA,$79,$C8,$E0,$FF,$2B,$7E,$43,$CA,$BF,$CE,$0D,$E4
	.dc.b	$34,$97,$FB,$56,$D7,$28,$FC,$3F,$D1,$FF,$B7,$23,$FC,$A8,$DB,$FE
	.dc.b	$F2,$D7,$FB,$BB,$5F,$CD,$FE,$44,$FA,$6F,$E0,$4D,$AF,$F3,$63,$67
	.dc.b	$FC,$E8,$D7,$FC,$E4,$D7,$9C,$F2,$24,$FF,$99,$FD,$A7,$D2,$7E,$79
	.dc.b	$B1,$98,$C1,$FC,$4F,$F2,$3F,$E2,$F7,$FE,$95,$6E,$86,$93,$C2,$E0
	.dc.b	$4B,$FF,$42,$47,$0E,$B7,$C8,$7F,$1F,$6C,$FF,$A2,$5B,$87,$FA,$26
	.dc.b	$7F,$92,$7C,$1F,$F5,$5A,$FF,$47,$36,$7F,$D2,$CD,$7F,$D2,$8D,$7F
	.dc.b	$83,$BF,$EF,$58,$E3,$F1,$BD,$08,$FF,$A6,$7E,$11,$F2,$7F,$CC,$DB
	.dc.b	$19,$8C,$1F,$99,$F8,$49,$1F,$99,$6E,$1F,$99,$6B,$F9,$96,$A3,$FD
	.dc.b	$C1,$D1,$FC,$09,$C7,$91,$FB,$9E,$DF,$E9,$93,$99,$DC,$7F,$16,$F8
	.dc.b	$D7,$D2,$D2,$63,$FE,$A5,$8C,$B6,$33,$5F,$E4,$C9,$CB,$63,$F9,$0F
	.dc.b	$D2,$BE,$3E,$92,$FF,$D6,$11,$96,$C6,$6B,$FC,$AF,$39,$6C,$7F,$5D
	.dc.b	$F0,$FF,$12,$4E,$4C,$12,$FF,$D9,$51,$96,$C7,$C2,$F5,$E4,$7F,$9F
	.dc.b	$74,$FF,$38,$36,$3F,$9C,$F7,$FF,$47,$C5,$82,$30,$DE,$5D,$92,$FF
	.dc.b	$D9,$93,$A2,$47,$F4,$4F,$A5,$7D,$0D,$25,$7F,$B4,$E3,$2D,$8C,$E1
	.dc.b	$FE,$7A,$9C,$B6,$3F,$91,$72,$4A,$FE,$44,$D9,$FE,$44,$D7,$F9,$13
	.dc.b	$5F,$C3,$F3,$BE,$87,$FC,$24,$DA,$FF,$51,$36,$7F,$C5,$7C,$9F,$B2
	.dc.b	$AF,$EB,$FA,$FF,$36,$9F,$8E,$DD,$7F,$D3,$9F,$01,$A2,$63,$FE,$F1
	.dc.b	$9D,$12,$3F,$2F,$F4,$92,$7E,$B3,$6F,$C3,$C0,$8C,$47,$FB,$93,$2F
	.dc.b	$EB,$11,$88,$FD,$37,$CB,$3E,$5F,$F8,$2C,$7F,$BA,$A3,$1A,$FF,$A4
	.dc.b	$9B,$8F,$F9,$22,$51,$87,$FC,$96,$D7,$95,$74,$7B,$B5,$E2,$DC,$3F
	.dc.b	$8C,$EF,$F3,$89,$7F,$39,$FF,$04,$FC,$F7,$F8,$96,$FF,$D6,$AD,$7F
	.dc.b	$80,$6C,$8C,$47,$E5,$9B,$19,$F4,$79,$92,$FD,$91,$96,$FD,$07,$E7
	.dc.b	$09,$C9,$FE,$1A,$6C,$6A,$30,$47,$FD,$CD,$F9,$23,$E3,$FF,$CB,$4D
	.dc.b	$8C,$C6,$0F,$F1,$2F,$D1,$FF,$59,$CF,$F5,$AD,$C4,$D3,$5A,$BC,$CF
	.dc.b	$F7,$06,$1F,$EF,$56,$DF,$E5,$1A,$FE,$0D,$AF,$F2,$46,$A9,$E9,$F1
	.dc.b	$1F,$E7,$19,$CC,$62,$BF,$04,$BF,$25,$FE,$29,$7F,$8F,$6B,$FC,$C1
	.dc.b	$AF,$56,$E3,$EC,$F6,$FE,$09,$2F,$E9,$E5,$FE,$76,$7B,$FE,$37,$FC
	.dc.b	$3D,$BD,$BF,$E8,$58,$C6,$3F,$B2,$6B,$FE,$FA,$F4,$FF,$C9,$6F,$E1
	.dc.b	$9A,$FF,$1A,$6E,$3F,$E4,$2F,$5B,$FC,$AF,$18,$23,$82,$61,$F8,$88
	.dc.b	$CB,$63,$FD,$31,$F9,$37,$C8,$26,$F8,$3A,$FC,$1C,$AE,$4D,$C7,$F5
	.dc.b	$DF,$41,$FE,$54,$EE,$FF,$56,$77,$FF,$5D,$94,$E5,$FF,$57,$47,$F9
	.dc.b	$EB,$E6,$7F,$8A,$7C,$12,$38,$56,$E3,$15,$FD,$9E,$31,$3A,$D5,$87
	.dc.b	$82,$9E,$83,$1E,$B7,$BA,$D5,$79,$2F,$8B,$E4,$3B,$38,$C7,$C3,$87
	.dc.b	$3F,$0F,$FD,$09,$5B,$39,$5F,$4C,$BC,$B1,$CA,$3C,$95,$C7,$4D,$63
	.dc.b	$D7,$2F,$25,$F3,$C3,$8D,$EE,$31,$5B,$7A,$18,$C3,$AE,$7E,$E8,$CE
	.dc.b	$7F,$21,$F2,$9B,$3D,$D7,$D6,$35,$87,$A2,$77,$18,$C3,$C1,$5B,$C7
	.dc.b	$CF,$39,$C7,$D6,$78,$B3,$F8,$B1,$9C,$3C,$0D,$3E,$5E,$5E,$78,$EF
	.dc.b	$7D,$08,$FD,$A8,$D6,$5E,$52,$65,$E4,$8C,$BF,$16,$7D,$E7,$E4,$4A
	.dc.b	$7C,$E3,$13,$F9,$36,$FC,$47,$F4,$67,$F3,$6D,$73,$E0,$FF,$4B,$7F
	.dc.b	$37,$19,$7F,$2E,$F1,$C7,$E5,$B7,$1C,$D7,$E0,$BD,$3F,$E5,$D7,$F8
	.dc.b	$7A,$D3,$F9,$D6,$BD,$A7,$87,$5F,$EA,$F0,$EB,$1C,$49,$3F,$72,$FA
	.dc.b	$BF,$8B,$BE,$78,$9A,$78,$36,$57,$0C,$3A,$DF,$1C,$3A,$C7,$E8,$A7
	.dc.b	$59,$7B,$2B,$44,$9D,$65,$E7,$9D,$47,$42,$3F,$EF,$AA,$E7,$5F,$06
	.dc.b	$FB,$CF,$03,$57,$CE,$74,$77,$8D,$3F,$C0,$2B,$9B,$F2,$4D,$1C,$2F
	.dc.b	$CA,$F6,$31,$3C,$B0,$F8,$77,$DE,$79,$57,$A1,$D9,$DE,$3C,$9D,$4F
	.dc.b	$88,$FF,$27,$4F,$F5,$06,$BF,$E3,$06,$BF,$E1,$F6,$BF,$E8,$3E,$07
	.dc.b	$E4,$6F,$92,$FC,$84,$61,$B1,$99,$FF,$0F,$4E,$5F,$E7,$DC,$7F,$97
	.dc.b	$E8,$71,$7F,$18,$9F,$D2,$46,$27,$53,$99,$FE,$A5,$18,$23,$D0,$87
	.dc.b	$28,$E0,$FF,$57,$8D,$BF,$D7,$4F,$D4,$64,$E9,$FD,$54,$7F,$79,$6C
	.dc.b	$66,$3D,$1D,$6F,$4E,$92,$3F,$B7,$BD,$3B,$BF,$8F,$47,$F7,$BE,$2C
	.dc.b	$11,$FF,$07,$76,$78,$BB,$CD,$24,$68,$87,$F8,$4D,$CB,$F2,$3C,$27
	.dc.b	$93,$F3,$5B,$1F,$E3,$28,$C1,$F0,$3C,$2F,$F7,$7C,$7F,$8B,$BD,$6C
	.dc.b	$11,$FF,$1C,$F2,$73,$23,$82,$1E,$17,$2D,$3F,$C8,$7E,$33,$E9,$7F
	.dc.b	$00,$D8,$CC,$60,$8E,$08,$7E,$12,$38,$3F,$E8,$6F,$95,$79,$BF,$69
	.dc.b	$B1,$FE,$2C,$6B,$B8,$FE,$4D,$DC,$F9,$D7,$DE,$E1,$B5,$FE,$75,$9D
	.dc.b	$4E,$49,$19,$9D,$BF,$E7,$86,$BF,$E8,$16,$BF,$EE,$85,$F4,$DF,$90
	.dc.b	$6C,$7F,$90,$9A,$FF,$A8,$9A,$FF,$A4,$5A,$F3,$8E,$44,$7F,$CE,$3F
	.dc.b	$4A,$F6,$BF,$5C,$D8,$CC,$60,$8E,$08,$7E,$CA,$38,$3F,$3F,$9B,$E7
	.dc.b	$11,$F9,$CF,$EA,$3C,$57,$D1,$D2,$4F,$F5,$58,$CB,$63,$2F,$F8,$D9
	.dc.b	$AF,$F8,$D7,$BB,$FC,$6B,$E9,$7F,$07,$A3,$49,$3F,$E9,$08,$CB,$63
	.dc.b	$2F,$F0,$C8,$CB,$FC,$3B,$E1,$3F,$D7,$0F,$FA,$61,$AE,$7D,$3F,$E2
	.dc.b	$D1,$88,$F2,$35,$FF,$53,$B5,$FF,$7D,$B6,$CE,$7B,$BF,$21,$97,$E0
	.dc.b	$49,$5F,$81,$7F,$BD,$78,$EF,$85,$7C,$AE,$1B,$5F,$EC,$99,$D4,$E4
	.dc.b	$91,$99,$DB,$FE,$0F,$9D,$B9,$9F,$E6,$5E,$7F,$D4,$8F,$E6,$2D,$7E
	.dc.b	$93,$5F,$F6,$2B,$56,$48,$C3,$76,$EA,$49,$FF,$62,$FB,$8F,$A3,$FE
	.dc.b	$84,$D8,$CC,$60,$8E,$08,$7F,$46,$8E,$0F,$E9,$FE,$8B,$F5,$84,$7F
	.dc.b	$58,$FE,$57,$A6,$47,$E4,$DA,$FF,$A1,$9A,$FF,$8C,$9A,$8F,$F3,$CF
	.dc.b	$4F,$F8,$69,$FF,$45,$69,$FC,$4B,$CC,$FB,$33,$99,$E1,$39,$7E,$C7
	.dc.b	$C0,$FF,$14,$78,$B5,$DB,$5F,$F4,$3B,$60,$BE,$07,$F0,$ED,$9E,$B1
	.dc.b	$83,$F1,$7E,$C3,$E5,$A9,$27,$F1,$51,$97,$F1,$9F,$06,$F7,$3F,$26
	.dc.b	$D8,$CC,$60,$8E,$08,$7E,$3E,$38,$3F,$DF,$3D,$47,$CF,$FF,$80,$5B
	.dc.b	$1F,$E6,$33,$F1,$6E,$1E,$11,$FA,$4F,$86,$7D,$75,$24,$FE,$8E,$32
	.dc.b	$D8,$CB,$FA,$18,$CB,$63,$FC,$25,$F4,$1F,$5E,$3F,$36,$DE,$63,$4D
	.dc.b	$6E,$F8,$BF,$F0,$3B,$6B,$F7,$31,$82,$38,$21,$FB,$97,$FC,$1F,$EB
	.dc.b	$3E,$7F,$FC,$51,$3F,$E8,$58,$C7,$97,$FE,$2E,$9C,$CF,$FB,$66,$72
	.dc.b	$FF,$A0,$7E,$11,$F0,$DF,$E1,$A6,$C6,$63,$04,$70,$43,$E9,$47,$07
	.dc.b	$FD,$2D,$E0,$BE,$79,$F3,$DF,$E2,$86,$CF,$D9,$9D,$4E,$4F,$17,$C9
	.dc.b	$49,$F3,$35,$EE,$D7,$FD,$49,$68,$C4,$7E,$C7,$7F,$E4,$9E,$0F,$F9
	.dc.b	$27,$C3,$FE,$21,$3F,$89,$9C,$B6,$32,$FF,$21,$F8,$3F,$E6,$DC,$26
	.dc.b	$93,$C7,$FE,$62,$D9,$FE,$5C,$D7,$FC,$AB,$EE,$7F,$93,$1F,$E4,$2D
	.dc.b	$C0,$D3,$56,$F9,$2F,$B3,$5F,$C1,$1A,$A3,$5C,$7F,$D0,$A7,$30,$F9
	.dc.b	$37,$CC,$FF,$46,$6D,$FF,$43,$6B,$FC,$FD,$AF,$F1,$66,$BF,$C5,$1A
	.dc.b	$F8,$5B,$3D,$E7,$FD,$99,$F2,$AF,$9D,$FF,$2D,$DF,$F2,$38,$C7,$BD
	.dc.b	$A9,$CC,$A7,$2F,$EF,$5A,$FF,$B8,$5A,$FF,$2D,$38,$47,$28,$CC,$7F
	.dc.b	$31,$E4,$7C,$CF,$FB,$1E,$3F,$20,$D7,$F2,$ED,$7A,$47,$0F,$87,$FF
	.dc.b	$3F,$CE,$67,$A3,$54,$ED,$EB,$3F,$DD,$3A,$7F,$8A,$47,$F4,$06,$BF
	.dc.b	$CF,$9A,$FD,$B6,$CF,$A3,$87,$C2,$49,$F4,$B5,$F4,$B5,$F4,$DF,$E5
	.dc.b	$DF,$F7,$F7,$CD,$7E,$0B,$E6,$EE,$A6,$92,$7F,$96,$4F,$EA,$F3,$E3
	.dc.b	$CF,$F5,$67,$2D,$7E,$90,$9E,$0F,$FA,$93,$E2,$7F,$D4,$A9,$A4,$7F
	.dc.b	$D9,$5E,$6F,$E8,$9F,$27,$2F,$D9,$F8,$94,$E7,$2D,$6C,$FC,$9D,$E4
	.dc.b	$E7,$0E,$17,$71,$FA,$B8,$F2,$9F,$EE,$98,$F2,$8C,$4F,$80,$FB,$78
	.dc.b	$74,$AD,$9D,$0B,$1A,$FC,$53,$FD,$8F,$97,$8D,$B4,$69,$AF,$9F,$7C
	.dc.b	$F7,$96,$91,$8C,$3A,$C7,$36,$C6,$70,$F3,$4F,$0A,$E5,$39,$CB,$E2
	.dc.b	$E3,$E1,$C3,$8D,$77,$CB,$FD,$A5,$1D,$9B,$8F,$C1,$E8,$FE,$11,$EE
	.dc.b	$6F,$D8,$97,$B7,$E6,$4E,$A3,$07,$C4,$EE,$7D,$83,$F0,$2F,$A5,$18
	.dc.b	$78,$F7,$7F,$7C,$AE,$15,$B6,$CF,$08,$C6,$5E,$5A,$E0,$FD,$7B,$E9
	.dc.b	$7E,$13,$FC,$98,$7F,$3D,$2C,$7F,$32,$F0,$FF,$DF,$B3,$FC,$C1,$BF
	.dc.b	$44,$F5,$BC,$ED,$BE,$93,$89,$F5,$F8,$AF,$03,$9B,$6B,$FA,$54,$60
	.dc.b	$8E,$09,$8F,$EA,$31,$FF,$3C,$4F,$93,$C0,$78,$9F,$21,$B7,$F3,$70
	.dc.b	$ED,$87,$C1,$9C,$BF,$0F,$9F,$ED,$BF,$EA,$36,$BF,$D8,$DA,$FF,$9C
	.dc.b	$9B,$8F,$F1,$AE,$1F,$8B,$18,$23,$F2,$3E,$8F,$F1,$95,$F4,$5B,$CF
	.dc.b	$FB,$9B,$E5,$97,$E0,$E7,$2F,$F7,$3F,$9C,$7D,$6F,$D2,$6C,$7F,$73
	.dc.b	$6E,$1F,$15,$AF,$FA,$F5,$AF,$39,$E4,$49,$FE,$F9,$F7,$2F,$6F,$F2
	.dc.b	$2D,$8F,$CB,$5F,$6C,$BD,$13,$9C,$3F,$19,$F4,$0D,$BF,$0E,$DC,$3F
	.dc.b	$C1,$0D,$9F,$B4,$D7,$FD,$A4,$D7,$FC,$24,$D7,$FD,$42,$DD,$0D,$F8
	.dc.b	$F7,$C5,$FE,$A1,$B7,$FA,$D9,$F0,$65,$F3,$E7,$38,$7E,$8F,$D8,$7C
	.dc.b	$FF,$E7,$9B,$87,$F8,$7D,$B3,$FE,$6D,$9D,$12,$74,$D8,$FC,$B7,$D3
	.dc.b	$FF,$26,$C3,$EF,$36,$BF,$C5,$D7,$D2,$73,$E2,$7F,$06,$FA,$87,$CE
	.dc.b	$FF,$3D,$C7,$FD,$43,$18,$C3,$FD,$A8,$D8,$FF,$68,$35,$DC,$F1,$7F
	.dc.b	$D4,$B1,$96,$C6,$5F,$F1,$24,$65,$FF,$67,$FD,$1B,$E3,$FF,$89,$B6
	.dc.b	$FF,$9D,$F8,$4B,$F1,$65,$F9,$BA,$D6,$1F,$C9,$7F,$28,$FA,$38,$7F
	.dc.b	$AD,$9B,$1F,$EB,$0F,$AE,$FE,$4B,$18,$E3,$7D,$5F,$E6,$2B,$FD,$C2
	.dc.b	$D8,$CF,$4F,$F1,$48,$C7,$D9,$FF,$20,$CF,$F5,$8D,$CB,$FD,$8B,$F9
	.dc.b	$4F,$FB,$F9,$FE,$61,$19,$C7,$FC,$27,$F5,$3F,$8F,$EF,$34,$9D,$5F
	.dc.b	$D6,$1B,$7F,$E7,$36,$D7,$F1,$5E,$E7,$D3,$7F,$3A,$6D,$7F,$56,$AF
	.dc.b	$74,$62,$BF,$9A,$7C,$63,$EF,$FF,$A0,$B6,$FF,$A3,$3F,$CB,$67,$73
	.dc.b	$A2,$4E,$BC,$6B,$3F,$CF,$FE,$A7,$F8,$FE,$FF,$C3,$DB,$5F,$E7,$F7
	.dc.b	$FC,$BE,$FF,$53,$73,$2E,$24,$61,$BE,$37,$F5,$68,$E3,$3D,$E3,$FB
	.dc.b	$B6,$8F,$A7,$A7,$F8,$37,$EE,$9F,$66,$F9,$DC,$E7,$F1,$6A,$E3,$FE
	.dc.b	$0D,$BD,$D6,$34,$FE,$51,$F6,$4F,$B3,$FE,$CE,$DC,$79,$1F,$DB,$5C
	.dc.b	$61,$FD,$87,$1F,$A1,$7F,$CC,$5B,$5F,$EA,$17,$F9,$8C,$FC,$17,$2D
	.dc.b	$E3,$FE,$EB,$18,$DF,$DD,$EC,$62,$7F,$BC,$FC,$2F,$FB,$AB,$0F,$EF
	.dc.b	$CD,$C3,$FB,$92,$FA,$DD,$FC,$6C,$AD,$BC,$A3,$15,$FE,$1A,$E1,$7D
	.dc.b	$27,$E4,$C9,$7F,$93,$6C,$FF,$B3,$27,$73,$92,$47,$F9,$49,$FF,$16
	.dc.b	$FA,$5F,$E4,$16,$9F,$97,$AC,$7B,$5F,$DA,$B7,$FE,$B1,$F0,$CF,$13
	.dc.b	$FC,$2E,$DB,$FF,$0E,$3F,$81,$6C,$68,$8E,$B8,$6E,$5F,$E1,$5A,$FE
	.dc.b	$9D,$3F,$ED,$36,$C7,$ED,$BD,$3E,$FE,$1C,$AE,$EF,$E4,$23,$13,$F1
	.dc.b	$E3,$F7,$04,$9F,$DC,$36,$3E,$33,$5F,$F4,$DE,$BF,$C3,$FE,$85,$F5
	.dc.b	$1F,$E3,$46,$DF,$EE,$EF,$13,$9F,$8B,$93,$FE,$54,$FC,$23,$F8,$77
	.dc.b	$3B,$0F,$F7,$EC,$FF,$90,$5B,$97,$F8,$D5,$BB,$FF,$9E,$2F,$8C,$62
	.dc.b	$BF,$DE,$DF,$4D,$F9,$13,$F2,$BE,$2B,$F0,$4F,$2E,$3F,$E3,$F9,$D7
	.dc.b	$4F,$F9,$31,$DC,$EB,$2F,$F0,$94,$7E,$6F,$EA,$7F,$B7,$13,$C7,$FF
	.dc.b	$20,$BF,$C9,$79,$DF,$31,$FE,$73,$6E,$9F,$E5,$B9,$FE,$7B,$A7,$E5
	.dc.b	$AB,$FC,$B7,$0F,$A6,$A7,$18,$B5,$C6,$2E,$BF,$E7,$8C,$3A,$BF,$E5
	.dc.b	$9B,$E5,$5D,$FD,$3F,$29,$1F,$F5,$6C,$FF,$19,$9F,$F4,$BB,$FE,$E1
	.dc.b	$E1,$43,$E8,$FC,$A7,$9A,$ED,$FF,$9D,$79,$3F,$CC,$78,$71,$6A,$3F
	.dc.b	$93,$38,$6F,$BF,$FE,$68,$87,$D6,$FF,$9F,$1B,$5D,$67,$13,$23,$C0
	.dc.b	$EE,$31,$5F,$E7,$9A,$E2,$D9,$D5,$70,$3D,$1E,$D7,$84,$8F,$C4,$CF
	.dc.b	$F3,$71,$E5,$EF,$F2,$1E,$56,$CF,$FA,$E2,$DF,$49,$FA,$2E,$2F,$81
	.dc.b	$F1,$6E,$EF,$85,$5A,$D3,$FC,$84,$FF,$B6,$63,$E3,$7A,$EF,$AD,$F4
	.dc.b	$DF,$1A,$F9,$58,$F5,$78,$18,$C3,$9C,$EE,$3B,$63,$E2,$C7,$AE,$1D
	.dc.b	$1F,$1E,$1C,$EF,$83,$63,$38,$74,$78,$ED,$FE,$EB,$AD,$CE,$71,$F1
	.dc.b	$63,$DD,$A7,$9B,$2F,$65,$F1,$C3,$AD,$6E,$BC,$E7,$B0,$B8,$7C,$1B
	.dc.b	$E3,$9E,$F2,$F3,$57,$6A,$F1,$DF,$86,$F7,$3C,$EB,$6F,$4C,$3C,$2F
	.dc.b	$97,$3F,$88,$F3,$8C,$E1,$D2,$35,$97,$A5,$D1,$0E,$83,$0F,$0C,$F1
	.dc.b	$25,$74,$3E,$5E,$5E,$BC,$FE,$55,$F2,$39,$DF,$2A,$E1,$1D,$F2,$F8
	.dc.b	$33,$98,$C5,$7E,$31,$FC,$79,$F3,$0B,$3F,$89,$6C,$6D,$FC,$1E,$BF
	.dc.b	$E5,$B7,$F8,$2C,$FF,$B4,$FD,$9F,$E0,$F1,$F9,$E9,$D3,$D1,$AF,$28
	.dc.b	$FA,$D1,$CF,$2F,$C6,$B7,$1F,$CF,$FB,$7F,$EE,$5B,$DD,$C7,$EC,$56
	.dc.b	$DF,$99,$3B,$73,$3B,$BE,$AF,$E0,$67,$51,$89,$D9,$1F,$C6,$46,$AF
	.dc.b	$D1,$39,$6C,$EB,$D1,$FD,$CB,$CC,$91,$A3,$D7,$1A,$71,$3F,$B8,$C3
	.dc.b	$D4,$FC,$9C,$3C,$D5,$B8,$D4,$7A,$EB,$8E,$6C,$7C,$D5,$A3,$CC,$5C
	.dc.b	$7C,$CF,$13,$15,$CA,$78,$C7,$84,$E2,$59,$E2,$E3,$2F,$C0,$3D,$AF
	.dc.b	$C5,$8F,$9F,$87,$FB,$FE,$9F,$C7,$3A,$FF,$DE,$CF,$F8,$0A,$FF,$A2
	.dc.b	$39,$DB,$FA,$15,$7F,$70,$9F,$E6,$D1,$F8,$C8,$C4,$7F,$49,$8C,$4F
	.dc.b	$FA,$C6,$1D,$4F,$89,$FC,$EE,$0B,$FE,$9D,$39,$AF,$01,$C2,$B7,$3A
	.dc.b	$7F,$4D,$5C,$67,$47,$0A,$E4,$FF,$59,$8D,$46,$21,$18,$9E,$75,$D6
	.dc.b	$79,$47,$81,$EA,$73,$8F,$17,$1E,$8F,$D6,$DF,$B5,$B5,$FB,$5A,$D9
	.dc.b	$FD,$57,$A1,$F7,$1E,$EC,$FF,$BB,$B6,$33,$3F,$E0,$FF,$B4,$C1,$F3
	.dc.b	$E0,$D2,$47,$5F,$A9,$E4,$47,$D0,$7D,$2F,$CA,$9E,$36,$63,$FC,$21
	.dc.b	$37,$6F,$8F,$5A,$9E,$3A,$A3,$11,$F2,$B6,$FB,$31,$8A,$E3,$87,$B7
	.dc.b	$C1,$7C,$6D,$3F,$10,$E3,$0F,$EF,$0D,$7F,$90,$EC,$8C,$47,$F8,$75
	.dc.b	$7C,$DF,$52,$47,$9B,$F0,$9F,$F5,$A3,$F0,$2B,$94,$EA,$3F,$C9,$0F
	.dc.b	$E1,$E7,$71,$FE,$55,$F3,$BE,$A7,$90,$69,$F8,$DC,$BF,$D3,$93,$FE
	.dc.b	$5D,$8C,$B4,$F8,$F1,$A7,$FC,$C1,$AF,$42,$3F,$33,$E7,$3F,$34,$FC
	.dc.b	$B3,$F4,$EF,$EA,$77,$34,$93,$F9,$06,$CF,$F9,$12,$77,$39,$3F,$6D
	.dc.b	$D8,$F0,$FF,$20,$DA,$FF,$4D,$C7,$EA,$6D,$F3,$3A,$89,$3A,$3D,$11
	.dc.b	$F9,$9F,$AB,$FE,$79,$C5,$FB,$3E,$04,$63,$2F,$E1,$2D,$BF,$F7,$82
	.dc.b	$F7,$79,$76,$49,$FB,$13,$BB,$E7,$5F,$C1,$2D,$F6,$EF,$8E,$C3,$70
	.dc.b	$FC,$FC,$EA,$7F,$85,$F6,$BE,$7D,$EF,$E1,$B7,$FE,$6C,$9D,$4F,$F6
	.dc.b	$86,$C6,$67,$6F,$F7,$68,$CB,$63,$2F,$F9,$22,$32,$D8,$CB,$FC,$B3
	.dc.b	$C2,$7D,$2F,$F0,$36,$C7,$E6,$5A,$FE,$17,$4F,$F2,$BF,$2E,$EA,$4E
	.dc.b	$DE,$51,$FC,$37,$EB,$3E,$24,$F1,$6F,$0F,$D7,$6E,$BF,$CF,$97,$D5
	.dc.b	$F1,$DA,$5F,$EA,$E7,$6B,$24,$70,$8F,$D8,$7B,$CF,$8C,$FE,$3C,$D8
	.dc.b	$CC,$7E,$3F,$EE,$A4,$7E,$8D,$AF,$E8,$FE,$0D,$F9,$5F,$FA,$D2,$34
	.dc.b	$43,$F5,$1D,$9F,$A6,$AF,$E9,$35,$C3,$D6,$FE,$5D,$E9,$BE,$BA,$93
	.dc.b	$0F,$E0,$D0,$F6,$F1,$1F,$3E,$38,$3A,$9E,$73,$FC,$E7,$E2,$1F,$35
	.dc.b	$49,$5F,$C4,$63,$FC,$F9,$C9,$FA,$57,$71,$99,$DB,$F7,$63,$7C,$9F
	.dc.b	$DC,$3E,$2B,$FC,$6F,$7F,$AD,$6C,$67,$D4,$FD,$C6,$33,$0F,$E0,$5E
	.dc.b	$1F,$AD,$B1,$FC,$1C,$8E,$08,$7F,$94,$38,$3F,$87,$C7,$F7,$DF,$5F
	.dc.b	$94,$FE,$45,$B9,$7E,$FD,$B8,$FF,$A8,$BE,$31,$F6,$FF,$D7,$7E,$3B
	.dc.b	$F8,$76,$1F,$CF,$1A,$FF,$B5,$DA,$FF,$83,$9A,$FF,$9C,$FE,$B9,$F1
	.dc.b	$9F,$E0,$4F,$61,$18,$9F,$EB,$B1,$89,$D7,$D7,$3E,$03,$FB,$77,$E2
	.dc.b	$7F,$95,$BF,$D5,$37,$6F,$E3,$0D,$D7,$FD,$77,$F2,$9C,$6B,$AB,$5F
	.dc.b	$F3,$AE,$48,$C4,$7F,$B1,$A3,$18,$7F,$9C,$1B,$1F,$EF,$86,$BC,$A7
	.dc.b	$8B,$E0,$9E,$11,$88,$FD,$77,$CA,$BE,$BF,$FC,$10,$D8,$FE,$BC,$D7
	.dc.b	$F8,$37,$C0,$3F,$13,$FF,$07,$36,$3F,$DC,$CD,$7F,$B7,$B5,$FF,$6F
	.dc.b	$F8,$9F,$E5,$99,$FF,$1F,$4E,$67,$FC,$86,$D7,$F8,$13,$67,$84,$7F
	.dc.b	$BC,$FE,$13,$FC,$98,$FF,$19,$6E,$DF,$D1,$DB,$AF,$F9,$3B,$EB,$7F
	.dc.b	$A2,$57,$F7,$96,$BF,$E0,$46,$BE,$66,$BC,$7D,$0F,$F2,$A4,$62,$3F
	.dc.b	$8B,$37,$03,$4C,$7C,$E3,$C4,$FF,$1A,$36,$BF,$C1,$4D,$7F,$99,$7C
	.dc.b	$EF,$F8,$D9,$BF,$05,$FE,$63,$8C,$47,$CA,$6B,$FE,$6A,$6B,$FC,$A5
	.dc.b	$B8,$1A,$49,$FF,$37,$43,$EB,$3F,$C9,$0D,$AF,$F0,$F3,$5F,$E8,$DD
	.dc.b	$DF,$83,$DB,$FB,$A3,$F8,$58,$FB,$AD,$7F,$CF,$AD,$7F,$9E,$36,$7F
	.dc.b	$CF,$3C,$5F,$E7,$F8,$C1,$1F,$F4,$17,$A4,$FB,$CF,$F2,$E3,$6F,$FC
	.dc.b	$64,$D7,$FA,$D7,$D0,$38,$BF,$CC,$4D,$8F,$F7,$09,$FD,$35,$C3,$FE
	.dc.b	$44,$6B,$F9,$66,$BF,$E9,$66,$BC,$63,$FD,$2D,$18,$C4,$DF,$A1,$7D
	.dc.b	$47,$F9,$CD,$B7,$FE,$49,$6B,$FD,$C7,$F2,$C7,$CD,$FF,$9E,$5B,$1E
	.dc.b	$8C,$11,$88,$FF,$2D,$B5,$FD,$6B,$5F,$F5,$6F,$67,$F2,$A7,$FB,$63
	.dc.b	$64,$E4,$7D,$27,$F4,$A9,$FF,$02,$46,$C9,$E3,$7F,$27,$9D,$70,$FF
	.dc.b	$A8,$B1,$FC,$33,$78,$BF,$D5,$11,$8E,$96,$33,$EA,$FF,$B8,$3F,$DB
	.dc.b	$A7,$4F,$FA,$F1,$BC,$9F,$12,$FF,$D7,$32,$AE,$38,$7F,$AC,$BE,$A1
	.dc.b	$F3,$9F,$EB,$76,$D7,$FA,$39,$AF,$F7,$CF,$A8,$7D,$9E,$EA,$4E,$DF
	.dc.b	$F6,$04,$EA,$3F,$D2,$ED,$7F,$AE,$35,$F6,$9F,$ED,$47,$0F,$FB,$49
	.dc.b	$AF,$FB,$47,$C7,$FF,$61,$91,$EB,$EF,$D7,$FC,$23,$F2,$1F,$D5,$1F
	.dc.b	$F5,$3B,$5F,$54,$7C,$4F,$AB,$F8,$99,$FE,$7A,$51,$87,$F4,$2D,$7C
	.dc.b	$E6,$B4,$FF,$6C,$BA,$E7,$7D,$67,$3A,$FC,$DB,$63,$33,$B8,$F0,$3F
	.dc.b	$7B,$E5,$2A,$69,$1F,$F0,$26,$5F,$E8,$9E,$3F,$6F,$C4,$7F,$BC,$73
	.dc.b	$FF,$76,$EC,$FA,$CC,$57,$FB,$DE,$AE,$1F,$EF,$B6,$F2,$31,$E9,$3F
	.dc.b	$7E,$F3,$18,$C7,$AF,$D1,$FF,$A9,$AB,$EB,$35,$FB,$4D,$B6,$48,$C7
	.dc.b	$27,$FA,$EF,$3F,$C7,$7C,$67,$8A,$B5,$18,$26,$1F,$E4,$CA,$E0,$FE
	.dc.b	$4E,$FD,$31,$8C,$55,$FA,$8C,$7D,$9F,$3F,$EA,$3E,$0F,$5F,$EE,$3E
	.dc.b	$4B,$15,$F8,$69,$B9,$7F,$B0,$6B,$53,$93,$FD,$B5,$F5,$9F,$E4,$E6
	.dc.b	$FB,$9D,$3B,$5F,$65,$F3,$BC,$05,$EF,$1F,$F4,$43,$75,$FF,$5B,$37
	.dc.b	$1F,$DF,$E5,$FC,$33,$0E,$53,$9B,$FF,$3F,$53,$DF,$FA,$18,$FC,$06
	.dc.b	$C6,$4F,$E4,$45,$9F,$F7,$DF,$86,$F8,$EF,$33,$67,$FD,$EC,$6F,$3F
	.dc.b	$F6,$54,$66,$BF,$A3,$FC,$D7,$83,$C6,$34,$97,$FC,$C6,$32,$D8,$CC
	.dc.b	$FF,$49,$7D,$3E,$47,$FB,$0A,$7F,$AA,$F8,$1F,$26,$F8,$DF,$FA,$BE
	.dc.b	$B4,$FF,$6C,$E2,$F8,$3C,$86,$93,$0F,$E8,$51,$96,$C6,$67,$F7,$7B
	.dc.b	$2B,$51,$8F,$7F,$F2,$6F,$FA,$FD,$BA,$FF,$7A,$BF,$BD,$18,$C3,$FC
	.dc.b	$19,$F1,$5F,$53,$CE,$34,$91,$FD,$56,$32,$D8,$CE,$1E,$49,$CB,$E6
	.dc.b	$F3,$FF,$DD,$D5,$FE,$22,$F8,$9F,$C7,$56,$B2,$FF,$80,$15,$AB,$FF
	.dc.b	$6B,$7D,$63,$6F,$C3,$B7,$1F,$EC,$CD,$9F,$86,$D7,$FC,$A2,$D7,$FA
	.dc.b	$9E,$C8,$C4,$7F,$B4,$5B,$19,$AF,$CA,$F9,$CF,$AD,$FD,$23,$70,$FE
	.dc.b	$BB,$97,$EE,$31,$E5,$7A,$BF,$D5,$7A,$CF,$AD,$FC,$FB,$70,$FF,$34
	.dc.b	$C6,$08,$E0,$97,$FE,$6C,$8F,$D3,$FB,$3F,$C6,$E7,$D6,$D9,$FF,$34
	.dc.b	$CF,$EE,$63,$1D,$9F,$E8,$DF,$8A,$FE,$3C,$DB,$FF,$45,$7B,$7F,$EE
	.dc.b	$D6,$CE,$A3,$18,$7F,$A3,$63,$04,$9F,$F0,$B7,$D3,$3E,$1B,$F8,$43
	.dc.b	$63,$FD,$3F,$38,$9F,$E3,$3F,$59,$F5,$F7,$F9,$19,$23,$17,$FC,$2B
	.dc.b	$E0,$5F,$5B,$C3,$87,$F8,$A7,$C2,$7E,$33,$5F,$F6,$07,$4F,$F0,$A8
	.dc.b	$C4,$7F,$B0,$63,$F8,$6F,$D6,$FF,$11,$E6,$3F,$C4,$F4,$7C,$3F,$FB
	.dc.b	$EF,$4F,$B6,$7E,$1A,$7F,$8C,$77,$3E,$A3,$F0,$ED,$D3,$FA,$34,$60
	.dc.b	$8E,$09,$5F,$E5,$29,$E0,$FF,$8F,$7B,$9F,$4F,$FB,$23,$FA,$DD,$FC
	.dc.b	$2D,$5A,$2B,$77,$E6,$9C,$FC,$63,$E1,$7F,$9C,$36,$75,$68,$FE,$45
	.dc.b	$C9,$F9,$78,$C5,$70,$25,$70,$AD,$BF,$A8,$7A,$5A,$F9,$7D,$6F,$F1
	.dc.b	$C6,$F8,$BF,$D2,$63,$FC,$71,$7E,$27,$4F,$EC,$9B,$5E,$E6,$E2,$78
	.dc.b	$C7,$79,$E1,$5F,$1B,$EA,$A4,$7C,$3F,$91,$7D,$6F,$FB,$FA,$3E,$37
	.dc.b	$07,$D9,$AD,$E1,$CC,$FE,$51,$BA,$3F,$C1,$B8,$72,$F8,$DD,$EB,$F1
	.dc.b	$FE,$53,$FC,$EF,$97,$CC,$6E,$3C,$F2,$ED,$39,$F8,$4F,$ED,$59,$7D
	.dc.b	$8A,$F0,$6F,$DF,$8F,$3E,$B6,$99,$87,$D8,$BE,$19,$FD,$77,$F9,$B6
	.dc.b	$1F,$84,$73,$3F,$C7,$3E,$B7,$FC,$C6,$FE,$01,$B3,$E0,$5F,$5B,$FD
	.dc.b	$F5,$CD,$FE,$0D,$B5,$CF,$E1,$35,$5B,$7F,$2B,$F1,$FF,$A0,$24,$FE
	.dc.b	$77,$89,$18,$C7,$FC,$9F,$BF,$EA,$63,$11,$FE,$68,$8C,$57,$FA,$D8
	.dc.b	$7B,$9F,$DA,$33,$FE,$59,$18,$F4,$BC,$2F,$E8,$4C,$1F,$9E,$F9,$F7
	.dc.b	$8F,$FE,$19,$6C,$67,$7C,$4C,$8E,$0E,$E3,$19,$7D,$6A,$FE,$A9,$5F
	.dc.b	$D2,$1B,$1F,$88,$F4,$BB,$78,$52,$BC,$4F,$EE,$A3,$13,$F8,$5D,$BF
	.dc.b	$8A,$BF,$90,$9F,$E1,$24,$AF,$94,$D8,$F9,$33,$C3,$5F,$E2,$53,$FE
	.dc.b	$28,$FA,$07,$D4,$7F,$8D,$DB,$87,$E8,$2F,$CF,$CC,$9C,$E3,$F9,$25
	.dc.b	$F1,$7C,$4F,$F3,$5F,$D6,$FF,$27,$8C,$E1,$FC,$29,$B5,$C6,$B7,$EF
	.dc.b	$71,$AF,$E1,$FF,$4B,$F9,$4A,$FE,$32,$DD,$5F,$5B,$E2,$E4,$FE,$25
	.dc.b	$3B,$D0,$D3,$84,$EF,$0F,$EF,$31,$BF,$AA,$FE,$EE,$4F,$1F,$FC,$96
	.dc.b	$FF,$36,$E8,$7C,$E7,$FA,$19,$B7,$FC,$92,$7F,$97,$67,$F9,$8A,$FF
	.dc.b	$30,$43,$EB,$69,$C6,$2D,$71,$8B,$AF,$F0,$1C,$3A,$BF,$BE,$BE,$55
	.dc.b	$DF,$E2,$7C,$A4,$7F,$DF,$D1,$FE,$98,$9F,$F6,$5E,$A8,$7D,$6F,$90
	.dc.b	$F2,$DD,$BF,$E7,$BC,$BF,$E6,$5C,$38,$B5,$FF,$0F,$B5,$FE,$96,$DC
	.dc.b	$3F,$13,$F5,$BF,$E7,$F8,$37,$E0,$F0,$CB,$EE,$E5,$A4,$BF,$E3,$B5
	.dc.b	$CA,$3C,$B5,$C6,$73,$EE,$7F,$A0,$28,$D2,$57,$FB,$5A,$7C,$FD,$DE
	.dc.b	$53,$CD,$36,$7F,$B1,$4B,$E9,$3F,$47,$C5,$FE,$D3,$ED,$FF,$7A,$C6
	.dc.b	$27,$FD,$A9,$1F,$E9,$6F,$9E,$F3,$46,$01,$CF,$2F,$F2,$83,$93,$17
	.dc.b	$D5,$DD,$F5,$F8,$E6,$1D,$75,$FF,$22,$46,$2B,$78,$74,$8C,$47,$43
	.dc.b	$B5,$F3,$BE,$59,$77,$9D,$57,$0A,$D4,$74,$C7,$C9,$EA,$33,$F6,$DE
	.dc.b	$F2,$F0,$65,$F1,$1B,$87,$13,$E1,$5F,$1C,$FD,$35,$BC,$7C,$59,$78
	.dc.b	$2F,$C7,$3E,$67,$D3,$87,$47,$27,$5C,$7E,$16,$3D,$73,$E7,$5B,$C3
	.dc.b	$A1,$EC,$9D,$5F,$4C,$76,$47,$CF,$9F,$BB,$2F,$34,$6C,$99,$78,$27
	.dc.b	$51,$E3,$9D,$CF,$C5,$AF,$93,$5A,$AF,$49,$30,$F2,$46,$5F,$3E,$3F
	.dc.b	$86,$9F,$C4,$9F,$2C,$A7,$DD,$2C,$7E,$4D,$BA,$FF,$9A,$5F,$D0,$47
	.dc.b	$FB,$53,$D9,$FD,$14,$7E,$76,$74,$F0,$8C,$CE,$E3,$F1,$33,$CF,$1F
	.dc.b	$C3,$37,$2F,$F7,$AF,$BD,$FE,$E9,$AF,$B7,$39,$7F,$36,$D7,$C1,$87
	.dc.b	$E3,$27,$84,$76,$8E,$24,$7F,$13,$1A,$BF,$0C,$E5,$B3,$AF,$86,$F8
	.dc.b	$F1,$82,$61,$FB,$8C,$3C,$9A,$B1,$E9,$39,$C7,$A3,$71,$E8,$60,$F5
	.dc.b	$C5,$AE,$58,$F8,$2B,$86,$3E,$0B,$D9,$C8,$F7,$11,$D5,$72,$8E,$C7
	.dc.b	$C6,$35,$3F,$3A,$BB,$47,$4A,$EC,$ED,$FD,$1E,$1E,$C8,$D5,$79,$08
	.dc.b	$AF,$C0,$FF,$17,$E2,$FF,$00,$E9,$F9,$0E,$CF,$F3,$E3,$FE,$04,$9F
	.dc.b	$E8,$AE,$78,$3F,$A1,$96,$3F,$A1,$BF,$8E,$AF,$F0,$2C,$7E,$61,$AF
	.dc.b	$F4,$C8,$FC,$9F,$53,$E3,$7F,$3B,$87,$FB,$86,$31,$5D,$5A,$FF,$9C
	.dc.b	$A7,$F4,$D3,$B7,$8D,$7F,$30,$AD,$D7,$F9,$AE,$72,$A3,$C1,$C5,$D6
	.dc.b	$73,$1C,$2F,$A5,$66,$A3,$FE,$B1,$AF,$DB,$3F,$E5,$16,$BE,$16,$BF
	.dc.b	$EB,$0E,$87,$CC,$FC,$7C,$BF,$BA,$36,$33,$3F,$E0,$BF,$AD,$C1,$F2
	.dc.b	$E0,$D2,$47,$F8,$03,$EB,$79,$11,$FF,$4B,$FD,$6B,$C7,$C4,$7F,$82
	.dc.b	$E6,$E9,$F3,$EB,$53,$C2,$BE,$84,$62,$3F,$D2,$8D,$9F,$F1,$7F,$C7
	.dc.b	$FA,$FC,$23,$F8,$16,$D7,$F7,$1D,$3F,$C6,$73,$93,$5F,$5B,$C4,$93
	.dc.b	$FE,$AE,$FA,$DF,$F5,$53,$FE,$42,$8C,$BF,$EA,$0B,$47,$E1,$67,$73
	.dc.b	$FE,$53,$F3,$BE,$77,$93,$FE,$82,$8F,$F2,$84,$7E,$2F,$4F,$F4,$CC
	.dc.b	$60,$87,$13,$FD,$7E,$E9,$FF,$2F,$72,$74,$23,$FE,$C4,$FA,$44,$7E
	.dc.b	$5D,$B1,$FE,$F7,$6B,$FE,$9A,$6B,$FE,$99,$6B,$F9,$96,$BF,$97,$6F
	.dc.b	$29,$A6,$B5,$FF,$03,$F8,$7F,$8D,$E9,$34,$F2,$7D,$6D,$C3,$FD,$3D
	.dc.b	$15,$FF,$68,$5A,$7F,$2D,$F5,$BF,$CD,$B8,$BF,$DF,$CD,$CF,$FC,$3B
	.dc.b	$92,$31,$82,$31,$EF,$F2,$EF,$93,$FE,$24,$C3,$9C,$76,$9F,$E1,$58
	.dc.b	$3E,$B5,$F2,$18,$6E,$1F,$EB,$76,$BF,$EF,$1E,$D7,$D6,$BE,$03,$F8
	.dc.b	$A4,$61,$FE,$C2,$D9,$FF,$22,$35,$FE,$DA,$D7,$FB,$8B,$5F,$F1,$DB
	.dc.b	$53,$4C,$69,$FE,$49,$FA,$DF,$DF,$78,$E6,$98,$E5,$B7,$F0,$F4,$F9
	.dc.b	$75,$B8,$C1,$F2,$DC,$3F,$F0,$01,$FD,$94,$7D,$2F,$AC,$F8,$7C,$1F
	.dc.b	$90,$C5,$18,$9D,$6D,$FC,$46,$1F,$5B,$62,$63,$E0,$BE,$31,$B9,$FE
	.dc.b	$31,$F0,$0F,$8D,$A4,$BF,$D1,$7D,$3F,$E8,$48,$E0,$F5,$FC,$23,$EB
	.dc.b	$7F,$D5,$71,$93,$F2,$33,$C1,$6B,$FA,$AA,$E1,$EB,$7F,$29,$F4,$DF
	.dc.b	$5B,$49,$7E,$98,$FF,$69,$B6,$32,$F4,$E0,$FC,$CC,$F2,$8E,$1E,$D7
	.dc.b	$F3,$5F,$89,$7B,$74,$93,$F1,$63,$FC,$DF,$9F,$CF,$77,$19,$9D,$BD
	.dc.b	$CF,$D2,$47,$0C,$3F,$53,$F1,$5F,$E3,$2B,$FB,$0D,$8C,$FA,$9F,$C2
	.dc.b	$F1,$98,$7F,$0B,$7F,$B4,$F8,$3E,$E8,$FE,$19,$18,$23,$B2,$1F,$DC
	.dc.b	$FD,$6E,$18,$FE,$1D,$B8,$7E,$F1,$B3,$FE,$07,$F8,$C7,$D6,$FF,$59
	.dc.b	$F8,$EF,$E1,$B3,$FC,$F5,$AF,$F7,$F6,$BF,$ED,$26,$BF,$EA,$4F,$59
	.dc.b	$F4,$AF,$8C,$FF,$05,$46,$1B,$19,$9F,$D1,$CB,$E4,$5F,$01,$FD,$AF
	.dc.b	$EB,$7F,$37,$5F,$99,$6B,$F4,$7B,$D1,$8D,$FF,$9A,$E0,$F9,$4E,$77
	.dc.b	$D9,$AF,$F9,$C2,$35,$87,$F9,$A9,$B1,$FE,$CD,$6B,$B9,$E2,$FF,$9E
	.dc.b	$A3,$2D,$8C,$BF,$0D,$AF,$F2,$9F,$95,$7D,$5F,$F8,$1D,$B1,$FB,$56
	.dc.b	$BE,$DF,$7D,$F5,$BF,$E0,$D6,$C7,$F7,$16,$BF,$E3,$36,$BF,$E8,$ED
	.dc.b	$3F,$CC,$93,$FE,$39,$9C,$CF,$F3,$E6,$BF,$CF,$5B,$3C,$23,$FB,$F7
	.dc.b	$C2,$7F,$91,$DF,$A0,$DD,$3F,$8A,$37,$97,$FA,$4F,$D6,$BC,$FF,$F1
	.dc.b	$23,$6F,$FC,$10,$D7,$FC,$94,$D7,$D0,$D7,$9C,$7F,$94,$63,$11,$FD
	.dc.b	$4D,$B8,$1B,$F3,$8F,$0F,$FC,$68,$DA,$FE,$44,$D7,$F1,$5F,$3B,$FE
	.dc.b	$36,$6F,$C1,$7F,$97,$E3,$11,$F2,$DA,$FF,$66,$6B,$FD,$95,$B8,$1A
	.dc.b	$49,$FF,$35,$43,$EA,$FF,$C9,$0D,$AF,$E6,$CD,$7F,$3D,$EC,$FE,$13
	.dc.b	$6F,$E8,$4F,$E1,$A3,$EF,$35,$FE,$F2,$D7,$FB,$C3,$67,$FC,$CD,$C5
	.dc.b	$FE,$7B,$8C,$11,$FF,$3F,$7A,$2F,$AD,$FF,$2E,$36,$FF,$A5,$35,$FD
	.dc.b	$97,$D0,$38,$BF,$CC,$4D,$8F,$F2,$33,$5F,$F4,$43,$5F,$CB,$35,$FF
	.dc.b	$08,$B5,$FF,$08,$37,$13,$4C,$7D,$0B,$EA,$7F,$CE,$4D,$AF,$EC,$0D
	.dc.b	$7F,$84,$7D,$6B,$E6,$FF,$CF,$0D,$8F,$F2,$E3,$5F,$F4,$F3,$5F,$D6
	.dc.b	$35,$FF,$13,$F5,$FF,$2C,$C7,$FD,$65,$E2,$FF,$A0,$E3,$F9,$78,$FA
	.dc.b	$6F,$E9,$75,$FC,$66,$38,$12,$7F,$93,$CE,$B8,$7F,$D5,$18,$FE,$0D
	.dc.b	$BC,$5F,$EA,$98,$C7,$4B,$19,$F5,$9F,$DC,$9F,$EA,$DC,$9E,$FB,$E0
	.dc.b	$48,$FF,$09,$89,$E1,$87,$FA,$DF,$EA,$1F,$35,$FE,$B7,$6D,$7F,$86
	.dc.b	$B0,$46,$23,$BF,$D4,$BE,$B7,$85,$C9,$E1,$5B,$C3,$FD,$89,$7B,$9E
	.dc.b	$8D,$7D,$87,$FB,$05,$C3,$8C,$11,$88,$D6,$1F,$ED,$2F,$90,$FE,$A7
	.dc.b	$1F,$EC,$DF,$1F,$FD,$88,$7F,$B4,$6B,$FC,$99,$19,$21,$FE,$CD,$8D
	.dc.b	$7B,$FF,$EE,$38,$7D,$6F,$C7,$CB,$CB,$1E,$68,$C3,$83,$7A,$FF,$88
	.dc.b	$1C,$46,$9E,$2F,$E6,$7E,$B7,$84,$76,$E8,$46,$27,$53,$CA,$72,$F0
	.dc.b	$1F,$29,$5A,$65,$C1,$FE,$DF,$F8,$8F,$F4,$DE,$3F,$EE,$F6,$F1,$FC
	.dc.b	$FF,$AD,$B7,$FE,$ED,$DF,$FD,$23,$3F,$EF,$7A,$CF,$0F,$FB,$FD,$BC
	.dc.b	$0C,$33,$7C,$30,$FC,$2F,$D1,$FF,$A9,$30,$FB,$2D,$B3,$4D,$71,$BE
	.dc.b	$33,$BB,$73,$FC,$3C,$E2,$64,$7E,$39,$F1,$C6,$30,$FD,$1E,$2F,$9E
	.dc.b	$F2,$9E,$EE,$BF,$DB,$7C,$96,$31,$FC,$F5,$6D,$B5,$BE,$BF,$F7,$2C
	.dc.b	$60,$FF,$37,$7D,$67,$F9,$39,$BE,$E6,$FE,$B7,$E5,$F8,$0A,$D6,$9F
	.dc.b	$E1,$B6,$F2,$7F,$9F,$DB,$87,$F8,$B7,$16,$1C,$A7,$37,$FE,$B5,$A7
	.dc.b	$BD,$FE,$D0,$C7,$FD,$B2,$D8,$C9,$FE,$F9,$2C,$FE,$6A,$5E,$1B,$E3
	.dc.b	$7C,$ED,$AF,$F4,$51,$C3,$3F,$F4,$A4,$66,$FF,$A1,$FC,$D7,$87,$C6
	.dc.b	$34,$97,$FD,$11,$B3,$FD,$11,$AF,$F8,$67,$C8,$FF,$60,$4F,$F5,$0F
	.dc.b	$03,$E4,$DF,$48,$C1,$FE,$9D,$E9,$FA,$D7,$EB,$6B,$FD,$9B,$8B,$E2
	.dc.b	$46,$23,$FB,$3C,$60,$93,$FB,$9D,$55,$BA,$FF,$71,$7B,$FF,$96,$AF
	.dc.b	$F5,$63,$63,$FB,$BD,$FD,$E8,$C6,$BF,$EB,$9F,$35,$F1,$7F,$55,$B8
	.dc.b	$77,$6E,$1F,$4D,$AF,$F8,$46,$78,$12,$78,$3F,$7F,$CF,$FF,$76,$C7
	.dc.b	$F8,$7B,$E2,$7F,$1F,$7D,$30,$FF,$65,$56,$B0,$FF,$35,$7D,$6F,$FB
	.dc.b	$83,$0F,$80,$D8,$FC,$23,$5F,$F2,$63,$5F,$EA,$1A,$A3,$10,$8C,$70
	.dc.b	$7E,$4E,$73,$18,$AF,$F2,$6F,$D6,$FE,$93,$CF,$3F,$A6,$AF,$16,$5F
	.dc.b	$E4,$E8,$FD,$4F,$AC,$FA,$DF,$CF,$37,$0F,$F3,$24,$60,$8E,$09,$9F
	.dc.b	$F9,$9E,$3F,$4D,$EC,$FF,$15,$C3,$FD,$DC,$DA,$FF,$33,$4F,$EE,$23
	.dc.b	$14,$F0,$3F,$D1,$DF,$14,$BF,$F4,$47,$B7,$FE,$EB,$6C,$66,$BF,$D1
	.dc.b	$4D,$7F,$D1,$0D,$7F,$D1,$0D,$7B,$4F,$37,$8C,$FE,$57,$E9,$5F,$0D
	.dc.b	$FC,$31,$B1,$FE,$9F,$BC,$4F,$F1,$DF,$AD,$F9,$73,$FE,$B7,$6D,$FA
	.dc.b	$FE,$05,$F5,$BB,$E4,$FE,$2F,$F0,$9F,$92,$AF,$BE,$D7,$EF,$37,$94
	.dc.b	$DF,$AD,$FE,$2B,$07,$F8,$BE,$EF,$87,$FF,$7F,$E9,$FE,$19,$E3,$F8
	.dc.b	$D7,$BF,$23,$FA,$6F,$D3,$FE,$33,$C0,$34,$9A,$FF,$4B,$6E,$1F,$DC
	.dc.b	$67,$73,$93,$F9,$27,$7B,$EA,$FF,$6C,$76,$8B,$5E,$33,$F3,$BF,$19
	.dc.b	$FE,$9B,$71,$C1,$FF,$00,$3E,$17,$F9,$DC,$FF,$21,$D7,$D2,$EE,$7F
	.dc.b	$DA,$CD,$8C,$CF,$0F,$23,$8B,$FD,$13,$EB,$7F,$8F,$DF,$FB,$75,$B5
	.dc.b	$F1,$1F,$F1,$B6,$1E,$2D,$D1,$A8,$FD,$9B,$6B,$F6,$4D,$9D,$FB,$5D
	.dc.b	$FC,$8F,$9B,$F5,$53,$0F,$6F,$C8,$BE,$B7,$FD,$F5,$1F,$E3,$FE,$35
	.dc.b	$6F,$3F,$F2,$6C,$62,$78,$D7,$09,$CD,$7F,$60,$F9,$4F,$F3,$85,$7A
	.dc.b	$5B,$3A,$8C,$7C,$1F,$F6,$6A,$FA,$D1,$EF,$6C,$FF,$C0,$17,$F9,$6C
	.dc.b	$7F,$12,$73,$B7,$E2,$9B,$19,$E3,$E9,$EB,$75,$7F,$B9,$FD,$6F,$F9
	.dc.b	$7A,$7E,$63,$70,$61,$F9,$48,$7D,$6F,$0F,$84,$97,$C1,$E7,$18,$BF
	.dc.b	$8B,$F1,$DF,$95,$25,$7E,$57,$74,$62,$7F,$C9,$7D,$5F,$9E,$8C,$47
	.dc.b	$F9,$7E,$77,$5F,$E4,$D3,$FC,$49,$E9,$FF,$8D,$9A,$7D,$7A,$C7,$B7
	.dc.b	$FD,$9E,$FD,$73,$FC,$E3,$E7,$DE,$47,$F8,$51,$B1,$F8,$D7,$EE,$4F
	.dc.b	$03,$F9,$E3,$A7,$1D,$08,$C6,$FF,$BA,$6D,$77,$E2,$ED,$87,$7F,$4E
	.dc.b	$4F,$E3,$7D,$4B,$97,$4C,$3F,$AC,$4F,$9E,$BF,$81,$D6,$B1,$FE,$59
	.dc.b	$BF,$F0,$49,$FE,$04,$4A,$F9,$CD,$C3,$E6,$EF,$FC,$66,$B8,$56,$8F
	.dc.b	$F0,$07,$D0,$3E,$A3,$FC,$6E,$D8,$FD,$16,$18,$9C,$FC,$54,$9F,$F5
	.dc.b	$27,$D6,$FF,$20,$B4,$E7,$2F,$8F,$18,$AD,$F8,$B9,$D3,$F9,$0F,$D2
	.dc.b	$FE,$62,$FF,$B5,$37,$47,$D6,$F8,$E7,$E8,$CE,$F8,$CD,$38,$4E,$EF
	.dc.b	$AF,$57,$F9,$4B,$EA,$BF,$B9,$93,$C7,$FF,$24,$3F,$CB,$FA,$1F,$39
	.dc.b	$FE,$7E,$6E,$DF,$C4,$67,$FA,$4E,$BF,$9A,$AE,$9F,$5B,$2B,$FD,$8D
	.dc.b	$7C,$36,$FC,$3E,$0A,$D4,$61,$ED,$78,$CF,$FA,$1D,$7F,$A2,$6B,$FD
	.dc.b	$87,$B2,$31,$1F,$E1,$26,$BF,$E1,$16,$C6,$FE,$B7,$E0,$1F,$07,$E2
	.dc.b	$AE,$7F,$CA,$F9,$7F,$CC,$78,$76,$6B,$FE,$1E,$6B,$FE,$1D,$6D,$FE
	.dc.b	$1F,$EB,$7F,$CF,$B0,$6F,$C1,$E1,$97,$DD,$CB,$49,$7F,$AC,$AE,$51
	.dc.b	$E5,$AE,$33,$9F,$7B,$FC,$FD,$46,$92,$BF,$CB,$51,$FC,$07,$4F,$ED
	.dc.b	$AD,$8F,$64,$FA,$3C,$0F,$F5,$FD,$BE,$8F,$F4,$5C,$68,$FF,$66,$F7
	.dc.b	$7F,$BD,$27,$31,$83,$FC,$49,$F3,$DE,$87,$07,$3C,$97,$CF,$E3,$BF
	.dc.b	$84,$4F,$5B,$E7,$18,$8E,$13,$BC,$3C,$5E,$A3,$1E,$D7,$BC,$38,$D7
	.dc.b	$97,$1E,$F8,$F8,$A7,$2F,$9B,$7F,$F1,$A5,$F1,$AD,$B7,$2F,$05,$79
	.dc.b	$67,$38,$78,$B1,$F0,$BD,$EB,$75,$E8,$3C,$15,$C1,$F0,$6B,$AC,$BC
	.dc.b	$71,$BC,$7C,$0F,$CB,$CF,$DD,$9F,$9B,$2E,$D3,$D3,$2E,$59,$75,$BD
	.dc.b	$E5,$89,$95,$B3,$A6,$7E,$87,$D9,$1A,$CB,$C2,$7D,$4B,$E7,$3C,$63
	.dc.b	$38,$F8,$9D,$5F,$39,$F5,$4E,$48,$FC,$03,$93,$87,$66,$67,$AB,$E4
	.dc.b	$3F,$20,$7C,$02,$9F,$34,$B9,$FE,$49,$BA,$7F,$99,$1D,$9F,$9B,$70
	.dc.b	$EB,$DA,$CB,$96,$CF,$FB,$61,$FA,$D3,$CE,$BF,$1C,$DB,$FC,$9F,$BF
	.dc.b	$FE,$E7,$C3,$77,$1F,$C8,$56,$DF,$F4,$44,$ED,$CC,$EE,$FA,$BF,$90
	.dc.b	$C9,$5B,$9C,$D7,$02,$3F,$5E,$35,$87,$ED,$70,$F0,$57,$B4,$DC,$7A
	.dc.b	$3D,$3F,$9B,$7D,$09,$1C,$71,$F1,$61,$D2,$FF,$6D,$87,$4C,$BE,$34
	.dc.b	$74,$3D,$B8,$F4,$C3,$94,$64,$F1,$63,$E7,$9D,$47,$4C,$7E,$15,$70
	.dc.b	$AD,$4E,$0D,$1D,$F2,$F8,$B7,$CA,$3A,$D7,$12,$3E,$43,$91,$A3,$9D
	.dc.b	$76,$7F,$8D,$71,$FF,$80,$77,$FC,$8F,$17,$F2,$86,$F0,$7F,$43,$6D
	.dc.b	$7F,$43,$8B,$5F,$DD,$63,$F2,$ED,$7F,$A6,$47,$E4,$23,$11,$FC,$B7
	.dc.b	$A9,$F1,$5F,$9E,$C3,$FC,$3A,$DB,$D3,$69,$5C,$38,$BC,$6F,$28,$B3
	.dc.b	$B8,$C4,$7F,$A6,$9E,$0D,$8D,$11,$E3,$31,$FF,$56,$47,$F2,$C7,$FC
	.dc.b	$AE,$D7,$F6,$F3,$D1,$F8,$3D,$0F,$98,$F8,$AD,$8F,$F0,$54,$63,$EB
	.dc.b	$64,$7E,$E3,$3F,$EF,$D1,$F1,$E8,$D3,$1F,$5B,$D0,$95,$ED,$FA,$D7
	.dc.b	$8F,$88,$FF,$08,$CD,$DB,$FC,$93,$5A,$9E,$15,$F2,$A3,$11,$FE,$61
	.dc.b	$CD,$18,$AF,$F1,$BF,$C7,$FE,$03,$C4,$3F,$75,$B7,$EB,$6B,$F8,$46
	.dc.b	$BF,$E1,$C5,$F5,$BC,$C9,$1F,$5F,$EB,$7F,$D6,$2F,$FB,$06,$76,$D9
	.dc.b	$DB,$E7,$9D,$BF,$E5,$0F,$3F,$F2,$BE,$49,$A7,$E2,$34,$FF,$4E,$CF
	.dc.b	$FB,$36,$74,$D9,$DB,$F8,$99,$DB,$FE,$5C,$E4,$EA,$47,$F1,$DF,$48
	.dc.b	$8F,$CA,$B6,$3F,$DE,$CD,$7F,$D4,$2D,$7F,$16,$D7,$F4,$CD,$7F,$42
	.dc.b	$DD,$8D,$35,$C5,$FE,$95,$F0,$FF,$6B,$D2,$69,$E4,$FA,$DB,$87,$FA
	.dc.b	$86,$2B,$F8,$1B,$4F,$EF,$BE,$B7,$F9,$D7,$17,$AF,$14,$63,$3F,$F1
	.dc.b	$44,$63,$B3,$87,$BD,$C5,$99,$7F,$BE,$EF,$82,$AE,$75,$FC,$3A,$DF
	.dc.b	$5A,$F9,$0C,$36,$FF,$C6,$D9,$7C,$DA,$DD,$7F,$16,$ED,$7D,$6B,$DF
	.dc.b	$C3,$70,$FE,$CF,$3A,$9F,$EC,$AD,$8C,$CE,$DF,$EE,$71,$96,$C6,$5F
	.dc.b	$EF,$31,$96,$C6,$5F,$8B,$E2,$3E,$B7,$F9,$4B,$63,$D7,$8E,$A7,$3C
	.dc.b	$BA,$64,$E8,$F9,$15,$C5,$FF,$6A,$BF,$B7,$AF,$A7,$F5,$BF,$1E,$7F
	.dc.b	$DB,$2D,$D3,$ED,$36,$77,$F5,$BF,$CF,$23,$15,$F5,$5B,$C0,$68,$9F
	.dc.b	$E8,$7F,$00,$F8,$DA,$4B,$FD,$2F,$D3,$7E,$90,$8E,$0F,$67,$C2,$3E
	.dc.b	$B7,$FD,$67,$19,$3F,$29,$3C,$1F,$F7,$B4,$EF,$D4,$FE,$A7,$E9,$BE
	.dc.b	$B6,$92,$BE,$46,$A9,$FF,$6B,$C6,$1E,$0D,$9E,$0F,$FC,$00,$8D,$AD
	.dc.b	$FF,$45,$F1,$2F,$6F,$FB,$1B,$67,$BB,$5E,$1C,$3F,$A6,$8F,$F4,$05
	.dc.b	$72,$AE,$24,$AE,$3E,$07,$F6,$EF,$8A,$FF,$1D,$D7,$EC,$5B,$87,$EC
	.dc.b	$1B,$5F,$C1,$BD,$DF,$E2,$27,$F9,$03,$C3,$FE,$0E,$DF,$0B,$D7,$3F
	.dc.b	$C4,$23,$04,$7F,$8B,$B5,$FE,$FB,$EB,$7C,$2C,$7F,$81,$37,$1F,$E0
	.dc.b	$2D,$97,$A6,$F8,$B7,$D6,$FF,$82,$3E,$3B,$F8,$A5,$7F,$61,$6B,$FE
	.dc.b	$0A,$6B,$FE,$E2,$6B,$FE,$A4,$FA,$47,$D6,$FF,$85,$30,$46,$27,$F4
	.dc.b	$D1,$89,$D7,$C9,$3E,$07,$FC,$33,$F5,$BF,$99,$BF,$C2,$B5,$FA,$E7
	.dc.b	$F1,$57,$1B,$7F,$3B,$C1,$F2,$BC,$EB,$BB,$5F,$F3,$CC,$6A,$FF,$CE
	.dc.b	$2D,$8F,$F3,$BB,$5D,$CF,$17,$FD,$07,$19,$6C,$65,$FF,$62,$B5,$FE
	.dc.b	$55,$F2,$CF,$AB,$FF,$1A,$B6,$3F,$42,$D7,$DB,$D7,$FE,$70,$FA,$DF
	.dc.b	$F1,$DF,$C1,$1F,$F1,$C3,$67,$FC,$2C,$D7,$FC,$7E,$D7,$FD,$33,$3C
	.dc.b	$DD,$CF,$22,$3F,$E1,$FF,$84,$7C,$07,$F9,$29,$B1,$FB,$66,$BE,$6F
	.dc.b	$AD,$7A,$3F,$E5,$06,$C7,$F8,$A5,$AF,$F9,$49,$AF,$91,$AF,$F5,$66
	.dc.b	$BF,$D4,$5B,$81,$A6,$3E,$75,$E2,$FF,$98,$1B,$5F,$C3,$1A,$FC,$CF
	.dc.b	$9E,$FF,$20,$37,$E0,$BF,$CC,$51,$88,$F8,$ED,$7F,$B4,$B5,$FE,$C8
	.dc.b	$DC,$0D,$24,$FF,$9B,$61,$F5,$7F,$E7,$46,$D7,$F2,$06,$BF,$80,$F0
	.dc.b	$3F,$05,$CB,$FD,$ED,$FC,$24,$7D,$A6,$BF,$DF,$1A,$FF,$77,$6C,$FF
	.dc.b	$9D,$38,$BF,$CF,$D1,$82,$3F,$E8,$1F,$49,$F5,$BF,$E8,$86,$DF,$F3
	.dc.b	$46,$BF,$94,$FA,$17,$0F,$FA,$39,$B1,$FE,$67,$6B,$FE,$90,$6B,$F9
	.dc.b	$26,$BF,$E1,$46,$BF,$E0,$F6,$E2,$69,$8F,$A2,$7D,$4F,$FA,$79,$B5
	.dc.b	$FD,$29,$AF,$EA,$BE,$B5,$F3,$7F,$EA,$46,$C7,$F9,$F5,$AF,$FA,$99
	.dc.b	$AF,$EA,$1A,$FF,$8A,$BA,$FF,$95,$63,$FC,$AB,$C6,$FF,$57,$C7,$F2
	.dc.b	$D1,$F4,$FF,$D3,$78,$7F,$C6,$91,$C0,$93,$FC,$96,$75,$E3,$FF,$AC
	.dc.b	$72,$FC,$1D,$FF,$95,$78,$BD,$3C,$7F,$EA,$C9,$CD,$33,$9F,$3D,$FD
	.dc.b	$D1,$FF,$43,$4E,$9F,$F5,$AB,$6F,$9C,$7F,$85,$C4,$F1,$C3,$C1,$F5
	.dc.b	$0F,$9C,$FF,$5D,$B6,$BF,$D7,$0D,$7F,$D2,$9F,$52,$FA,$DD,$D4,$9D
	.dc.b	$BF,$EC,$49,$D4,$7F,$B0,$5A,$FA,$CF,$F6,$93,$87,$FD,$76,$D7,$33
	.dc.b	$FD,$A7,$E4,$3F,$A6,$C7,$FB,$27,$C9,$FF,$67,$9D,$FE,$02,$C6,$48
	.dc.b	$7A,$1F,$F7,$4A,$FA,$CF,$87,$9F,$8E,$3C,$91,$87,$06,$F4,$FF,$08
	.dc.b	$38,$8D,$3C,$5F,$CB,$FD,$6F,$08,$ED,$D0,$8C,$4E,$A7,$94,$E5,$E0
	.dc.b	$3E,$4E,$B4,$CB,$83,$CB,$C0,$D9,$F5,$BF,$E9,$98,$C3,$63,$3C,$7C
	.dc.b	$1B,$7F,$06,$B7,$39,$AE,$04,$7E,$1F,$CA,$BF,$2D,$FE,$93,$18,$6D
	.dc.b	$FC,$B9,$D7,$91,$FE,$AB,$6D,$78,$7E,$93,$C8,$47,$04,$AF,$26,$49
	.dc.b	$D4,$61,$FF,$5B,$75,$FE,$CB,$E4,$EF,$5F,$FB,$93,$0F,$EF,$AD,$BF
	.dc.b	$F0,$1B,$5F,$EF,$AD,$7F,$5D,$5C,$E7,$8D,$74,$3A,$FD,$74,$8E,$AD
	.dc.b	$AE,$AD,$7A,$B5,$EB,$F6,$3F,$DF,$67,$FC,$22,$D8,$FF,$0B,$B5,$FF
	.dc.b	$08,$B5,$FF,$5E,$C7,$F9,$A4,$FC,$AF,$7F,$FB,$7F,$DE,$FF,$6C,$36
	.dc.b	$32,$6E,$7F,$C4,$7F,$7B,$FB,$98,$FF,$10,$B6,$33,$D9,$F3,$23,$11
	.dc.b	$FE,$22,$6B,$F1,$9A,$FF,$93,$9B,$99,$CD,$FF,$3D,$F9,$BF,$2F,$8A
	.dc.b	$69,$31,$FF,$19,$FD,$E7,$F9,$ED,$BF,$01,$F7,$A3,$19,$7F,$8F,$1A
	.dc.b	$FF,$8D,$9A,$FD,$A6,$BF,$E6,$26,$E4,$73,$8B,$9D,$F4,$9F,$29,$B8
	.dc.b	$FC,$E6,$E1,$FC,$1E,$B8,$57,$E4,$BF,$2E,$7E,$25,$49,$87,$F9,$3A
	.dc.b	$32,$D8,$CD,$FF,$95,$A7,$2D,$8F,$B7,$E8,$3E,$3E,$92,$FF,$03,$19
	.dc.b	$7F,$CB,$7F,$97,$BF,$06,$FD,$2B,$63,$31,$82,$38,$26,$1F,$E6,$89
	.dc.b	$E0,$FB,$FE,$D1,$E8,$FE,$39,$B5,$F9,$26,$CE,$E7,$F6,$BF,$58,$F3
	.dc.b	$E9,$2F,$FC,$E3,$19,$6C,$66,$BF,$CF,$33,$96,$C7,$E7,$BD,$B7,$D9
	.dc.b	$D2,$5F,$F9,$99,$FE,$0B,$F9,$87,$F8,$6A,$3F,$CF,$ED,$BF,$F4,$2B
	.dc.b	$5F,$F3,$FB,$5F,$E0,$6D,$7F,$D5,$6D,$E5,$34,$C7,$BF,$FE,$F1,$AF
	.dc.b	$F4,$6B,$6B,$FD,$1B,$F2,$1F,$C6,$23,$11,$FE,$8E,$F8,$37,$D9,$FF
	.dc.b	$1C,$6D,$FF,$A5,$9B,$3F,$E9,$C6,$BF,$E9,$6F,$B4,$FA,$2D,$DF,$F9
	.dc.b	$2C,$62,$3F,$8E,$B5,$FF,$62,$FC,$4F,$FB,$DE,$FF,$D4,$6D,$8F,$F5
	.dc.b	$4B,$5F,$F5,$1B,$5F,$EF,$5F,$2A,$FB,$E5,$7F,$61,$8C,$36,$33,$3F
	.dc.b	$DB,$27,$2D,$8C,$BF,$DC,$E3,$2D,$8C,$BE,$07,$FA,$F7,$D7,$FA,$FB
	.dc.b	$4D,$26,$5F,$C9,$3E,$45,$F3,$D4,$99,$7F,$2A,$8C,$B6,$33,$8F,$EF
	.dc.b	$E7,$2D,$8F,$E5,$FF,$69,$FE,$2C,$AF,$E6,$6D,$9F,$E0,$CD,$7F,$D6
	.dc.b	$AD,$7F,$C7,$9E,$02,$31,$08,$C7,$AD,$D8,$95,$FE,$14,$F9,$67,$DA
	.dc.b	$7F,$BC,$A3,$0D,$BF,$F0,$E4,$EB,$E3,$FE,$4C,$23,$15,$C2,$73,$18
	.dc.b	$9E,$35,$CD,$FF,$1C,$46,$5F,$F1,$D6,$DF,$43,$EC,$FE,$99,$30,$FF
	.dc.b	$0A,$EE,$8F,$EA,$A7,$F7,$B8,$FE,$B2,$DA,$FE,$58,$D7,$FD,$B6,$D7
	.dc.b	$E3,$B5,$F8,$CD,$CF,$B7,$A3,$D8,$FE,$C7,$F6,$8E,$FA,$4C,$3F,$B4
	.dc.b	$C6,$5B,$19,$C3,$F9,$F4,$E5,$B1,$F2,$9F,$67,$DA,$5C,$4D,$25,$FF
	.dc.b	$90,$E3,$2D,$8C,$CF,$F9,$3A,$72,$FF,$8C,$F2,$7C,$8F,$E7,$89,$7F
	.dc.b	$31,$FE,$FF,$F3,$1F,$36,$3F,$C0,$6D,$AF,$EA,$CD,$7F,$CE,$1A,$23
	.dc.b	$11,$F9,$26,$BF,$91,$6C,$6F,$0E,$E7,$EC,$3C,$A9,$7F,$79,$B7,$E9
	.dc.b	$6B,$F8,$6C,$7F,$B4,$B6,$3E,$47,$E0,$92,$BE,$33,$67,$E2,$35,$FF
	.dc.b	$3B,$5A,$31,$1F,$EA,$5D,$FF,$80,$F9,$3F,$C1,$09,$7F,$92,$7F,$CC
	.dc.b	$9F,$6B,$FE,$04,$93,$49,$E7,$68,$95,$FE,$6E,$8C,$BF,$E7,$0F,$B5
	.dc.b	$FF,$2F,$E1,$FA,$26,$BE,$36,$BF,$E0,$D6,$D9,$A4,$F6,$FF,$CF,$0E
	.dc.b	$BE,$BB,$F8,$E1,$3B,$7F,$CF,$8F,$F3,$9E,$07,$D8,$D2,$4F,$F9,$09
	.dc.b	$FE,$49,$C6,$F9,$FF,$F4,$23,$74,$FF,$44,$B7,$4F,$F4,$13,$5F,$DE
	.dc.b	$B5,$F9,$CD,$7F,$C5,$0D,$D3,$F2,$F5,$FE,$91,$FC,$03,$FD,$1B,$5F
	.dc.b	$C1,$9A,$FD,$76,$BF,$E3,$16,$E5,$EC,$FA,$BF,$51,$31,$FF,$2F,$E4
	.dc.b	$8F,$F4,$F7,$E0,$3F,$A6,$9F,$F5,$1B,$70,$FF,$4E,$35,$FE,$2E,$D7
	.dc.b	$F9,$8E,$FF,$E6,$57,$F4,$AD,$73,$ED,$FF,$9B,$23,$11,$FC,$D5,$B8
	.dc.b	$9A,$49,$E0,$4B,$FF,$3B,$C6,$5F,$3F,$2F,$A8,$8E,$0D,$FC,$97,$84
	.dc.b	$8E,$0F,$F6,$77,$D3,$78,$78,$59,$7F,$B2,$7A,$BF,$DD,$D3,$98,$C0
	.dc.b	$F3,$DF,$40,$FB,$4F,$EE,$BF,$11,$FC,$16,$B9,$DF,$18,$C4,$23,$18
	.dc.b	$76,$D7,$FC,$8B,$3D,$B1,$EB,$87,$4A,$D6,$5E,$6C,$79,$E3,$CB,$0E
	.dc.b	$F8,$F2,$9F,$16,$3C,$B3,$E1,$8F,$2B,$F5,$9E,$3B,$E1,$19,$BE,$39
	.dc.b	$FB,$6F,$8D,$F6,$CA,$BC,$5B,$97,$8F,$2F,$04,$F6,$8F,$33,$B3,$19
	.dc.b	$78,$1F,$99,$87,$77,$37,$D2,$75,$1C,$71,$F2,$E7,$F1,$63,$C8,$70
	.dc.b	$24,$6B,$2E,$57,$27,$39,$FF,$B6,$31,$F0,$5F,$CC,$C7,$BE,$5E,$83
	.dc.b	$D9,$3A,$AF,$8C,$4B,$F0,$DF,$77,$A9,$88,$D6,$7F,$12,$7B,$63,$E3
	.dc.b	$AF,$C8,$9E,$A2,$9F,$1C,$B8,$7E,$4D,$BE,$EB,$E1,$BF,$36,$D8,$E2
	.dc.b	$7E,$3D,$C3,$C2,$36,$7D,$88,$E1,$7F,$8C,$6E,$3F,$EF,$CF,$5B,$FC
	.dc.b	$83,$7B,$B8,$FD,$DA,$DB,$FA,$59,$DB,$99,$DD,$F5,$7F,$17,$92,$B7
	.dc.b	$39,$AE,$04,$7F,$37,$1A,$C3,$F6,$B8,$78,$2B,$DD,$87,$69,$F4,$79
	.dc.b	$7F,$33,$0E,$84,$8E,$33,$E2,$C3,$A5,$FE,$E3,$0E,$99,$7F,$BB,$EB
	.dc.b	$A4,$FB,$71,$E9,$87,$28,$F3,$37,$1F,$3C,$EA,$7A,$57,$27,$CE,$7E
	.dc.b	$FF,$1F,$0B,$8A,$F3,$3C,$0D,$57,$37,$57,$DC,$8F,$94,$ED,$87,$A5
	.dc.b	$F2,$99,$3F,$8D,$F1,$FF,$6E,$7F,$9A,$6F,$F8,$86,$9F,$CC,$1C,$74
	.dc.b	$B9,$75,$3F,$F0,$05,$B1,$98,$DC,$63,$E3,$FF,$8E,$C7,$E7,$25,$18
	.dc.b	$8F,$F3,$1B,$5F,$F0,$A3,$54,$70,$E0,$F1,$CE,$63,$13,$FA,$06,$BD
	.dc.b	$E3,$44,$9E,$27,$BC,$7E,$05,$FA,$69,$FD,$9C,$63,$7D,$4E,$67,$FA
	.dc.b	$3C,$E5,$FE,$DB,$F6,$BF,$27,$3F,$F0,$6B,$7A,$8D,$3E,$5F,$6B,$E5
	.dc.b	$69,$2F,$FC,$1A,$FD,$0E,$F7,$87,$49,$3E,$F8,$CB,$63,$2F,$CF,$8C
	.dc.b	$B6,$3C,$DF,$68,$F8,$9A,$49,$FE,$13,$9A,$75,$1F,$37,$F0,$27,$DC
	.dc.b	$FE,$11,$B1,$98,$C1,$F8,$8F,$B5,$FF,$4D,$47,$D5,$6B,$F8,$36,$BF
	.dc.b	$EA,$46,$BF,$E5,$56,$BF,$82,$6F,$19,$A6,$32,$FF,$4C,$4F,$E0,$DA
	.dc.b	$FE,$45,$AF,$FA,$C9,$AF,$F9,$8D,$AF,$E4,$1B,$3F,$4B,$CD,$FD,$1C
	.dc.b	$63,$EA,$3F,$D3,$92,$69,$39,$FF,$4F,$E7,$FE,$27,$0F,$F0,$7B,$53
	.dc.b	$49,$CD,$C0,$91,$FB,$58,$E5,$7D,$FE,$CF,$F6,$B8,$1A,$4A,$FD,$24
	.dc.b	$65,$B1,$97,$F6,$91,$97,$FA,$8E,$EF,$B3,$7D,$2F,$E7,$5B,$3F,$86
	.dc.b	$6B,$C1,$AF,$FA,$A1,$AF,$FB,$E1,$AF,$FB,$DD,$B3,$DE,$7F,$69,$F6
	.dc.b	$2F,$9A,$A4,$BF,$EB,$BD,$49,$D4,$7C,$5F,$B6,$F9,$E4,$70,$43,$FB
	.dc.b	$2E,$A8,$D4,$61,$FD,$D7,$7F,$EA,$23,$FD,$82,$D7,$FC,$8F,$97,$F8
	.dc.b	$EF,$87,$64,$8F,$D4,$CE,$D4,$ED,$CC,$FF,$07,$F7,$9F,$63,$49,$3F
	.dc.b	$B2,$8C,$B6,$32,$FD,$68,$CB,$63,$F0,$7F,$4F,$23,$F0,$2D,$7F,$C1
	.dc.b	$BE,$D3,$EC,$3F,$D3,$D1,$FE,$08,$8C,$7A,$7F,$D0,$DB,$19,$9D,$BC
	.dc.b	$9A,$FF,$86,$A3,$6E,$67,$94,$EA,$3E,$4F,$A2,$F8,$AF,$E5,$0D,$8F
	.dc.b	$E2,$2D,$7F,$26,$D7,$FD,$E6,$D7,$9C,$72,$23,$FC,$6B,$E0,$7F,$CE
	.dc.b	$2F,$FB,$E9,$AF,$DA,$6B,$F5,$9B,$A9,$A4,$F2,$7F,$9F,$3C,$3A,$7F
	.dc.b	$80,$5F,$EC,$89,$7F,$B0,$7F,$80,$FC,$DC,$8F,$DD,$35,$FC,$9B,$5F
	.dc.b	$F2,$0F,$1A,$31,$1F,$E9,$6F,$3D,$F0,$8F,$AE,$FE,$F1,$18,$6C,$66
	.dc.b	$7F,$C8,$73,$96,$FD,$03,$E0,$F8,$B6,$33,$18,$3A,$FC,$43,$CB,$A4
	.dc.b	$9E,$B1,$97,$BF,$C5,$57,$FC,$B7,$82,$31,$1F,$E6,$08,$C5,$7A,$A7
	.dc.b	$33,$EF,$F9,$D7,$C5,$FB,$9B,$1E,$C6,$BD,$9A,$F5,$6B,$CE,$39,$11
	.dc.b	$FE,$BF,$EB,$3E,$AF,$FB,$D3,$63,$31,$83,$E7,$F9,$3F,$D1,$B2,$34
	.dc.b	$86,$89,$3F,$49,$CB,$FD,$BB,$EA,$FF,$90,$93,$C5,$FF,$05,$3F,$C9
	.dc.b	$A9,$F9,$8F,$F7,$2A,$F9,$6D,$7E,$53,$5F,$C3,$35,$FF,$6C,$36,$39
	.dc.b	$FA,$1F,$88,$8C,$1B,$EA,$7C,$AB,$EA,$FF,$C3,$8D,$C3,$FC,$0A,$D7
	.dc.b	$FA,$3B,$5F,$B9,$F3,$0F,$8B,$FC,$AB,$63,$F2,$6D,$7F,$0A,$D7,$8C
	.dc.b	$70,$23,$C1,$FF,$17,$76,$3E,$AE,$92,$7F,$C3,$31,$97,$F4,$3E,$E3
	.dc.b	$ED,$3F,$4E,$D8,$CC,$60,$8E,$08,$7E,$6A,$38,$3F,$D9,$BE,$71,$F0
	.dc.b	$BF,$E4,$16,$C7,$F8,$A9,$AE,$E3,$F6,$FE,$9F,$F1,$98,$70,$FF,$79
	.dc.b	$F0,$B2,$48,$FD,$CC,$EE,$75,$18,$9D,$3F,$DE,$3E,$A9,$F5,$5F,$E5
	.dc.b	$36,$C7,$F8,$F1,$AE,$E3,$F8,$4F,$D0,$BE,$3A,$92,$7F,$84,$C6,$5F
	.dc.b	$E1,$7F,$47,$5F,$DE,$B5,$FD,$E3,$5F,$E3,$0D,$7D,$ED,$C4,$D2,$57
	.dc.b	$FA,$3B,$E5,$7F,$A4,$50,$38,$6F,$9B,$B2,$3F,$E9,$B8,$D1,$23,$FD
	.dc.b	$37,$F4,$8F,$A3,$A4,$AF,$F4,$54,$65,$B1,$97,$FD,$35,$19,$6C,$7F
	.dc.b	$AA,$7D,$C7,$E6,$4F,$A2,$C3,$6B,$FD,$EB,$3A,$9C,$9E,$4F,$72,$4F
	.dc.b	$89,$AF,$81,$AF,$C3,$3F,$D4,$B8,$7F,$88,$9D,$1F,$EB,$BF,$A7,$F8
	.dc.b	$FF,$53,$74,$34,$93,$CC,$F9,$D3,$FD,$47,$EA,$79,$7A,$26,$92,$7F
	.dc.b	$D8,$7F,$1D,$FE,$C4,$4D,$23,$FE,$B3,$F2,$9F,$07,$F6,$F1,$B3,$FE
	.dc.b	$BE,$F8,$DF,$E7,$F3,$93,$F0,$1D,$EF,$CC,$DF,$35,$C3,$0E,$33,$FE
	.dc.b	$CC,$C7,$17,$FE,$CA,$6E,$3D,$5B,$3C,$EE,$E3,$E0,$AF,$8B,$7E,$0C
	.dc.b	$5F,$2D,$E6,$9F,$A1,$87,$8E,$31,$1B,$6B,$FE,$D8,$9E,$35,$FE,$D5
	.dc.b	$FA,$C7,$DF,$78,$9B,$1E,$FA,$E1,$79,$A9,$39,$C1,$7C,$2F,$CE,$70
	.dc.b	$9D,$46,$74,$F9,$BE,$57,$E0,$3E,$E7,$E1,$B6,$BC,$0F,$E1,$31,$CD
	.dc.b	$2A,$92,$B7,$1D,$5F,$F3,$1B,$B7,$9F,$DF,$FE,$B5,$BD,$5F,$4B,$89
	.dc.b	$E5,$3E,$8F,$C7,$7B,$D7,$D7,$A2,$31,$97,$A9,$B9,$7F,$03,$C7,$9E
	.dc.b	$1C,$B2,$F6,$CB,$E6,$7E,$9E,$3F,$90,$6C,$64,$FF,$6F,$97,$2F,$F5
	.dc.b	$0F,$D8,$7F,$B9,$63,$E2,$B6,$3C,$B1,$BC,$FE,$6C,$E5,$F9,$FE,$4B
	.dc.b	$E6,$FC,$2D,$AF,$E8,$91,$82,$38,$25,$FF,$4D,$8F,$4F,$C6,$FF,$20
	.dc.b	$9F,$AE,$D9,$F8,$F5,$D5,$F9,$31,$9B,$FC,$0F,$1B,$E9,$BE,$1B,$70
	.dc.b	$FE,$CF,$18,$23,$82,$5F,$FB,$12,$78,$3F,$EC,$6F,$B6,$FF,$14,$4F
	.dc.b	$E2,$9B,$1F,$6A,$77,$7F,$92,$9C,$BF,$94,$F4,$9F,$73,$F5,$9B,$87
	.dc.b	$F8,$3A,$30,$47,$04,$C3,$FC,$23,$1F,$7B,$ED,$1F,$33,$F9,$E6,$D7
	.dc.b	$E7,$AF,$AD,$FE,$3A,$72,$FE,$43,$EB,$3F,$C3,$15,$FC,$8E,$3F,$D7
	.dc.b	$31,$87,$FD,$CA,$D7,$F9,$97,$12,$31,$1F,$EC,$66,$BC,$BD,$6F,$C9
	.dc.b	$C6,$27,$84,$7E,$5B,$D3,$7D,$47,$ED,$9B,$87,$ED,$AF,$C5,$AF,$EF
	.dc.b	$27,$38,$FE,$A3,$DF,$FF,$7B,$DF,$F9,$73,$DD,$FD,$0C,$62,$7F,$DF
	.dc.b	$AD,$7D,$2D,$7F,$DB,$AD,$7F,$DB,$2D,$CC,$DF,$B5,$FE,$51,$8F,$F0
	.dc.b	$E6,$E1,$CB,$0F,$E1,$F1,$8C,$3F,$BD,$7C,$8B,$E3,$3F,$93,$C2,$31
	.dc.b	$7F,$46,$31,$3A,$9C,$C7,$D5,$6B,$FD,$65,$AF,$F6,$56,$BE,$08,$EF
	.dc.b	$F8,$C7,$2C,$7F,$9E,$36,$F8,$4F,$F4,$8A,$4E,$5F,$4F,$66,$70,$FE
	.dc.b	$27,$1E,$CF,$4F,$13,$FD,$3B,$F3,$47,$CA,$7F,$44,$E9,$46,$33,$FC
	.dc.b	$74,$62,$75,$39,$C3,$FB,$A4,$BF,$35,$FF,$30,$5F,$F4,$36,$CF,$F4
	.dc.b	$23,$E5,$7C,$3F,$F5,$E8,$CE,$DF,$D9,$BF,$0E,$FE,$3D,$5F,$D6,$9A
	.dc.b	$FF,$6E,$6A,$C1,$18,$6F,$85,$EB,$8C,$11,$FE,$5F,$F7,$3F,$D8,$2B
	.dc.b	$F9,$DB,$6F,$C5,$3F,$D4,$A3,$19,$7F,$BC,$FE,$1B,$FA,$AF,$C9,$7F
	.dc.b	$2B,$8F,$F7,$B7,$8D,$89,$91,$FD,$62,$75,$8F,$FB,$E6,$B2,$FF,$4A
	.dc.b	$FA,$FF,$F7,$55,$7F,$5E,$6C,$67,$6F,$EC,$71,$8C,$FF,$B3,$B6,$BF
	.dc.b	$CA,$1D,$28,$C5,$A3,$17,$CB,$D4,$FF,$06,$ED,$FC,$CF,$1F,$A1,$E3
	.dc.b	$FF,$84,$BE,$5B,$F1,$7F,$3F,$A8,$9A,$7F,$7D,$BE,$0F,$F9,$7E,$B6
	.dc.b	$E6,$7F,$BC,$F4,$7F,$8A,$A3,$FB,$CB,$74,$FF,$01,$F1,$FD,$8A,$DD
	.dc.b	$7F,$80,$3E,$A7,$F9,$8B,$EC,$6D,$7A,$D7,$DB,$EF,$E3,$E5,$7F,$81
	.dc.b	$A7,$7C,$07,$71,$99,$FF,$06,$FC,$BB,$E5,$E9,$8F,$4E,$D7,$F8,$42
	.dc.b	$74,$D9,$D3,$4F,$F1,$07,$96,$F8,$AF,$D5,$36,$3F,$52,$DD,$3F,$4E
	.dc.b	$D7,$FC,$6E,$D7,$FC,$96,$D7,$F3,$CD,$DB,$FA,$9D,$7F,$8E,$B8,$3F
	.dc.b	$C6,$F9,$FD,$9F,$57,$E3,$D6,$B1,$FF,$1C,$56,$9F,$F1,$5F,$87,$FE
	.dc.b	$81,$6F,$C0,$FF,$8C,$21,$F6,$B6,$24,$E8,$E3,$18,$D3,$FC,$47,$5F
	.dc.b	$D6,$39,$FF,$C7,$6D,$AF,$DC,$79,$7D,$76,$91,$E2,$7F,$27,$18,$BF
	.dc.b	$C2,$C3,$EE,$7E,$59,$2B,$E5,$B6,$7E,$5B,$5F,$96,$FC,$5D,$7F,$63
	.dc.b	$A7,$F9,$73,$BF,$FC,$DB,$87,$5C,$3E,$2D,$7E,$DB,$95,$F7,$EF,$BD
	.dc.b	$FF,$29,$B6,$BF,$CA,$6B,$E7,$F3,$32,$72,$F0,$8C,$69,$FE,$48,$FA
	.dc.b	$77,$DA,$7F,$97,$9B,$7F,$E5,$E7,$F3,$26,$27,$84,$EC,$E1,$CB,$6B
	.dc.b	$FC,$CB,$F5,$7F,$DD,$C9,$E9,$FF,$93,$1F,$EF,$BF,$1C,$FC,$DB,$FC
	.dc.b	$3F,$5F,$ED,$78,$C7,$B5,$FE,$70,$AF,$EF,$33,$9A,$DE,$5F,$C2,$E7
	.dc.b	$F9,$BC,$FF,$10,$6C,$2F,$9C,$F9,$3E,$5B,$A1,$1F,$F4,$54,$6A,$7F
	.dc.b	$CD,$75,$CB,$80,$FF,$A7,$E3,$FC,$32,$D7,$C1,$F0,$FE,$D3,$DD,$F2
	.dc.b	$5E,$93,$D4,$DC,$3E,$2F,$C2,$FC,$32,$3F,$E2,$9C,$3B,$C7,$C7,$FC
	.dc.b	$0B,$E1,$1F,$0D,$B3,$F3,$A2,$E1,$E8,$9B,$D9,$FE,$9A,$C9,$58,$BF
	.dc.b	$8D,$56,$FE,$56,$DF,$54,$93,$E3,$BF,$F2,$CC,$FC,$0E,$7F,$A8,$7D
	.dc.b	$5A,$B5,$FE,$B3,$97,$D2,$FF,$A2,$FE,$05,$1F,$ED,$EE,$EF,$F7,$94
	.dc.b	$E6,$30,$7D,$7F,$B1,$FF,$3A,$7C,$67,$F8,$7E,$38,$CE,$E3,$11,$DB
	.dc.b	$D4,$61,$DA,$73,$5B,$C3,$F1,$98,$74,$7C,$37,$C3,$0F,$09,$CF,$0E
	.dc.b	$53,$A6,$C7,$97,$CB,$FF,$78,$E3,$DA,$F9,$36,$B7,$5C,$CF,$6F,$23
	.dc.b	$1F,$14,$7B,$2F,$8E,$3D,$E3,$C7,$97,$A5,$F0,$67,$D2,$31,$1D,$72
	.dc.b	$F0,$3D,$B0,$EA,$E6,$FA,$4E,$8F,$99,$8F,$8F,$3F,$5C,$E8,$95,$E5
	.dc.b	$CF,$E0,$D6,$F2,$F1,$67,$E8,$AD,$1E,$87,$13,$B8,$C5,$70,$23,$E7
	.dc.b	$C3,$C2,$F2,$C7,$CD,$8F,$8A,$F9,$46,$27,$53,$9C,$BD,$F9,$FC,$43
	.dc.b	$F1,$93,$F9,$03,$CA,$53,$EA,$15,$DB,$F9,$4F,$B1,$7C,$97,$F9,$AE
	.dc.b	$51,$89,$FF,$4F,$46,$27,$53,$99,$E8,$EF,$0F,$C9,$B7,$0E,$D5,$F3
	.dc.b	$3D,$F7,$D8,$F9,$EF,$73,$F7,$EF,$86,$1F,$A2,$6E,$7F,$E8,$CB,$E1
	.dc.b	$8F,$4A,$FC,$4D,$70,$7F,$66,$F0,$C7,$BC,$75,$C2,$3C,$1F,$25,$F4
	.dc.b	$8F,$04,$6B,$1F,$2C,$7E,$72,$B9,$BC,$49,$1A,$3F,$1F,$1A,$71,$87
	.dc.b	$9D,$B3,$FC,$DE,$75,$39,$24,$78,$8F,$05,$F9,$5F,$6D,$F9,$67,$78
	.dc.b	$FC,$59,$C1,$C7,$0E,$F1,$98,$F3,$CF,$67,$99,$D8,$E2,$62,$78,$E1
	.dc.b	$E6,$8F,$EA,$5E,$DF,$E3,$F9,$BF,$98,$37,$87,$FA,$A3,$67,$FA,$C7
	.dc.b	$C7,$DE,$8F,$E9,$AF,$FA,$E5,$B1,$FC,$C6,$31,$CC,$6F,$CE,$71,$83
	.dc.b	$4C,$7D,$85,$AF,$EE,$71,$5F,$EB,$B1,$72,$FE,$8D,$37,$2F,$ED,$53
	.dc.b	$CA,$BF,$A1,$3E,$22,$3F,$EB,$2F,$CD,$DE,$47,$B7,$CE,$46,$27,$FA
	.dc.b	$0E,$2C,$3F,$A1,$D6,$AF,$85,$BD,$BA,$7F,$B6,$E7,$9B,$63,$58,$78
	.dc.b	$63,$E2,$F9,$DF,$0C,$91,$FD,$F3,$C3,$67,$88,$FE,$FB,$37,$3F,$24
	.dc.b	$7F,$81,$EB,$6F,$1A,$FB,$1C,$9F,$E1,$76,$CE,$A7,$FC,$55,$CD,$F6
	.dc.b	$FC,$7F,$D1,$C7,$F1,$1E,$24,$62,$3F,$D4,$AD,$7F,$D4,$8D,$8D,$FC
	.dc.b	$07,$32,$4F,$F8,$CF,$E0,$6B,$FE,$39,$8B,$9F,$AA,$BB,$D7,$E3,$7C
	.dc.b	$CF,$F4,$65,$FF,$91,$BD,$3C,$19,$FA,$1B,$19,$20,$C3,$A5,$27,$F3
	.dc.b	$5E,$83,$F3,$87,$B9,$86,$E1,$FE,$7E,$6B,$FE,$AE,$FC,$D1,$F4,$FF
	.dc.b	$87,$6C,$7F,$BF,$1A,$FF,$BE,$9A,$FF,$3B,$E7,$46,$23,$F1,$6D,$7F
	.dc.b	$DE,$4D,$8D,$CF,$E2,$E3,$5B,$3C,$07,$91,$FB,$36,$E1,$FE,$94,$8F
	.dc.b	$CE,$75,$3E,$03,$51,$27,$47,$E7,$BB,$3F,$3B,$5B,$8C,$1F,$EE,$1F
	.dc.b	$AF,$7D,$E7,$F0,$16,$C6,$B2,$46,$23,$E9,$5F,$FB,$B2,$3E,$2F,$1F
	.dc.b	$DF,$74,$DB,$E8,$75,$D7,$EA,$7C,$54,$C3,$EA,$3F,$44,$8F,$D1,$7E
	.dc.b	$31,$1C,$10,$FA,$8E,$5F,$ED,$98,$FF,$90,$3C,$0F,$E3,$9E,$47,$F1
	.dc.b	$49,$F6,$35,$F5,$89,$FF,$23,$FA,$1C,$2A,$4F,$07,$9C,$7F,$27,$F8
	.dc.b	$1F,$F5,$14,$FF,$BE,$BD,$8C,$19,$FA,$2B,$19,$21,$E0,$C1,$49,$FE
	.dc.b	$65,$F8,$1F,$F2,$DA,$FA,$4D,$78,$37,$6F,$E0,$FC,$DF,$C8,$35,$FC
	.dc.b	$29,$2B,$EF,$C6,$5F,$F6,$FF,$E7,$3F,$A9,$73,$C6,$AE,$11,$F1,$2B
	.dc.b	$83,$F1,$BA,$97,$C6,$3D,$5F,$5C,$FC,$3F,$FA,$53,$63,$F1,$4D,$77
	.dc.b	$F6,$3A,$99,$1A,$78,$C7,$F5,$3F,$8B,$7C,$65,$25,$7E,$56,$1F,$29
	.dc.b	$88,$F9,$91,$C1,$D4,$F3,$9F,$EB,$CB,$DD,$A4,$9F,$87,$B7,$EB,$8F
	.dc.b	$E8,$07,$F4,$97,$FC,$71,$AF,$F4,$E8,$C9,$23,$FC,$78,$23,$4E,$1A
	.dc.b	$F2,$8E,$8F,$F0,$BF,$8E,$7C,$57,$F5,$96,$C7,$DA,$6B,$F9,$81,$F1
	.dc.b	$3E,$5E,$2F,$E9,$3E,$DB,$EC,$DF,$43,$E4,$BF,$F4,$AC,$61,$F3,$CE
	.dc.b	$E7,$4D,$8E,$1E,$07,$82,$FF,$AA,$C7,$F0,$4C,$BF,$C1,$4D,$3F,$D6
	.dc.b	$12,$F8,$AC,$FE,$34,$FA,$FA,$4A,$DF,$AB,$CA,$CF,$9E,$3F,$88,$B6
	.dc.b	$33,$1E,$48,$C1,$1C,$10,$FE,$07,$F2,$AF,$C2,$7F,$C3,$CD,$8F,$E3
	.dc.b	$CD,$7F,$69,$EE,$BE,$2B,$FC,$48,$D8,$FF,$72,$35,$FE,$1E,$D7,$FB
	.dc.b	$73,$5F,$F6,$7F,$84,$8C,$47,$F0,$F6,$C6,$67,$FC,$61,$18,$F9,$77
	.dc.b	$6F,$F8,$D5,$B7,$FC,$D5,$AF,$F0,$BF,$AF,$FE,$63,$E5,$F8,$1F,$E7
	.dc.b	$11,$FE,$01,$6B,$9F,$2F,$FC,$7D,$18,$8F,$E4,$4D,$C8,$D2,$4F,$F9
	.dc.b	$17,$E6,$9F,$23,$FE,$49,$6D,$FF,$49,$6B,$FC,$7F,$BD,$E5,$7F,$93
	.dc.b	$DB,$1F,$A8,$C7,$F7,$73,$FE,$51,$9C,$CF,$F3,$C6,$BF,$E1,$B6,$BF
	.dc.b	$39,$AF,$F3,$B6,$CF,$48,$EC,$BE,$BB,$FC,$C0,$D8,$FE,$C0,$D7,$F9
	.dc.b	$B7,$7F,$E1,$7A,$7F,$C3,$8F,$E1,$E3,$FC,$58,$D7,$EF,$B5,$FE,$98
	.dc.b	$D9,$FE,$09,$C7,$FE,$6D,$8C,$11,$FF,$38,$75,$7F,$1C,$AF,$EE,$4D
	.dc.b	$7F,$A4,$36,$FF,$C5,$DF,$3D,$FB,$46,$F0,$7F,$9E,$63,$F6,$F5,$FE
	.dc.b	$41,$6B,$F9,$46,$BF,$D8,$5B,$89,$A4,$9F,$F4,$27,$CF,$FF,$09,$AF
	.dc.b	$EB,$4D,$BF,$F2,$03,$67,$F4,$3F,$5A,$F8,$FF,$F4,$73,$6F,$F0,$0D
	.dc.b	$7F,$BC,$B5,$FF,$2C,$35,$FD,$5F,$C1,$7F,$4B,$C3,$FC,$9C,$DB,$CF
	.dc.b	$93,$FE,$9D,$FA,$17,$E1,$DD,$6F,$73,$FE,$9F,$BE,$18,$7F,$80,$5B
	.dc.b	$7C,$32,$FF,$52,$2F,$89,$A4,$AF,$F1,$34,$7F,$82,$F0,$FF,$30,$3B
	.dc.b	$8C,$CE,$DF,$F0,$04,$6F,$2F,$F4,$8F,$D2,$BE,$B3,$FD,$30,$DA,$FE
	.dc.b	$F2,$D7,$FB,$A7,$D3,$BE,$FF,$75,$27,$6F,$FA,$86,$75,$1F,$E0,$26
	.dc.b	$BF,$D6,$9A,$FF,$96,$DA,$FF,$A9,$1A,$FF,$5B,$FA,$9F,$ED,$0F,$F8
	.dc.b	$31,$AF,$F8,$1D,$B9,$FF,$85,$26,$E3,$FE,$AE,$9B,$5C,$FC,$EF,$F5
	.dc.b	$B4,$63,$9D,$7A,$A7,$D5,$7F,$5A,$C7,$FA,$D3,$63,$FB,$7B,$51,$0E
	.dc.b	$3E,$5F,$FA,$25,$DF,$D1,$37,$E6,$D8,$CC,$60,$FE,$A4,$BE,$46,$91
	.dc.b	$FE,$AC,$FF,$7D,$C7,$FD,$C9,$AD,$F9,$1F,$D7,$E9,$EC,$6C,$EA,$FE
	.dc.b	$94,$60,$95,$C4,$FF,$05,$FC,$93,$F3,$A7,$DC,$7E,$EA,$30,$D8,$CC
	.dc.b	$F0,$6D,$FD,$FF,$9A,$F0,$92,$7D,$0D,$8D,$46,$08,$FF,$9B,$38,$1F
	.dc.b	$3F,$FC,$2B,$BF,$FD,$68,$D8,$CD,$7F,$3B,$6D,$71,$9D,$12,$74,$FF
	.dc.b	$14,$DD,$F5,$9F,$EF,$19,$FF,$15,$C6,$3E,$57,$53,$9B,$FB,$C7,$DB
	.dc.b	$70,$FE,$06,$78,$57,$29,$CC,$FE,$1A,$72,$F2,$6B,$F8,$9E,$C7,$D8
	.dc.b	$BE,$97,$F9,$4C,$61,$B1,$9C,$3F,$C8,$33,$97,$FA,$E0,$F9,$DF,$90
	.dc.b	$D3,$2E,$0E,$BF,$83,$3E,$C2,$92,$7A,$C6,$5B,$19,$C7,$E1,$DA,$B5
	.dc.b	$18,$6B,$FD,$0F,$E7,$FE,$D7,$8E,$69,$30,$F5,$FE,$0D,$F8,$AA,$F8
	.dc.b	$8D,$7F,$5F,$82,$31,$08,$C7,$2F,$C8,$9C,$C6,$2B,$AC,$E4,$FC,$67
	.dc.b	$4F,$FB,$8A,$3F,$C0,$8D,$BF,$F5,$EB,$5F,$EE,$3F,$1D,$F5,$38,$7F
	.dc.b	$92,$57,$D2,$6C,$FC,$46,$BF,$75,$AF,$A9,$AF,$A5,$B8,$7E,$53,$C2
	.dc.b	$EC,$7E,$87,$CD,$7D,$0D,$26,$1F,$E1,$B8,$CB,$F8,$4F,$C1,$9F,$85
	.dc.b	$FE,$29,$B1,$98,$C1,$1C,$13,$2F,$BB,$3C,$1F,$F6,$FF,$D7,$BE,$9B
	.dc.b	$F6,$8D,$AF,$F1,$63,$67,$73,$F9,$AF,$99,$7C,$8D,$25,$FE,$6A,$32
	.dc.b	$D8,$CD,$7E,$3A,$72,$D8,$FE,$09,$EE,$7F,$B2,$EB,$FC,$CF,$F1,$D8
	.dc.b	$23,$FE,$93,$F9,$1F,$F5,$EB,$7E,$2F,$FC,$27,$18,$C3,$FD,$56,$D7
	.dc.b	$F7,$ED,$7F,$9F,$35,$FE,$82,$D7,$FD,$DB,$1D,$3E,$5A,$57,$85,$B3
	.dc.b	$DB,$E7,$3C,$24,$9F,$33,$E4,$F9,$D9,$1E,$96,$CF,$9B,$E6,$3D,$24
	.dc.b	$9F,$73,$EC,$F9,$89,$1F,$05,$B3,$EE,$6B,$EE,$6B,$FD,$73,$A5,$F5
	.dc.b	$9F,$C2,$63,$FD,$A6,$D7,$FB,$04,$6B,$C1,$FE,$A3,$39,$9F,$F4,$9F
	.dc.b	$C9,$7F,$22,$5E,$4A,$30,$DF,$1F,$F9,$F4,$60,$8E,$09,$97,$F9,$E7
	.dc.b	$EB,$5A,$7F,$45,$6C,$66,$30,$47,$04,$C3,$F9,$DE,$0A,$E3,$1F,$CF
	.dc.b	$BE,$59,$F5,$EF,$9E,$FF,$2D,$C6,$1B,$19,$C3,$FD,$E7,$39,$6C,$66
	.dc.b	$BA,$35,$FF,$04,$4E,$DC,$D7,$CA,$F1,$65,$FF,$BE,$F5,$FE,$50,$D8
	.dc.b	$FE,$C9,$F2,$5F,$D0,$EF,$D6,$D7,$FC,$29,$9F,$FB,$FB,$CB,$D9,$2F
	.dc.b	$FD,$45,$5B,$AF,$F6,$4B,$5F,$AE,$DA,$FF,$80,$3F,$5D,$FD,$CF,$E5
	.dc.b	$4D,$25,$FF,$AB,$63,$2D,$8C,$DF,$D0,$9C,$BF,$B0,$F8,$AF,$D9,$92
	.dc.b	$BF,$70,$D8,$D4,$60,$8F,$F7,$4E,$4F,$E4,$BF,$5D,$FE,$22,$F1,$4D
	.dc.b	$25,$7F,$B0,$7A,$BF,$E0,$0B,$7E,$47,$FC,$1D,$18,$C3,$F0,$AD,$7E
	.dc.b	$5B,$5F,$F7,$7B,$5F,$CE,$B7,$43,$9C,$7F,$57,$F1,$9F,$A6,$26,$1F
	.dc.b	$75,$B7,$FA,$E6,$BF,$AE,$E1,$FF,$70,$CF,$E3,$BE,$95,$F3,$BF,$94
	.dc.b	$6C,$66,$30,$47,$04,$AF,$CF,$4F,$07,$E5,$6A,$FA,$EF,$94,$4A,$FF
	.dc.b	$10,$CB,$2F,$EA,$73,$FE,$35,$FA,$37,$D7,$52,$5F,$FB,$BE,$32,$D8
	.dc.b	$CD,$FE,$CA,$72,$D8,$FF,$20,$7D,$33,$EB,$A9,$2F,$FD,$F5,$19,$6C
	.dc.b	$66,$BF,$7B,$39,$6C,$7F,$8B,$33,$7C,$E5,$25,$FF,$8B,$23,$2F,$C0
	.dc.b	$E1,$7D,$5F,$F5,$DC,$3F,$5A,$D7,$FD,$07,$1A,$F3,$3F,$CB,$13,$99
	.dc.b	$FF,$0B,$5F,$F8,$2D,$FF,$0D,$B5,$CF,$A9,$E5,$26,$1E,$36,$C7,$F8
	.dc.b	$C2,$78,$7A,$67,$FD,$C1,$5E,$9F,$8C,$F7,$11,$F5,$60,$F9,$4F,$29
	.dc.b	$23,$2F,$FA,$2B,$F0,$2F,$E4,$78,$FF,$3E,$6B,$FD,$15,$AF,$E0,$DB
	.dc.b	$A7,$C6,$F8,$1F,$82,$4B,$F0,$B6,$BF,$98,$E0,$8C,$47,$CC,$6B,$F3
	.dc.b	$23,$53,$C8,$91,$FC,$06,$FF,$C8,$7C,$86,$93,$D9,$FA,$7B,$4B,$F8
	.dc.b	$6D,$BF,$CE,$FD,$37,$F9,$A7,$E5,$3F,$D7,$BF,$11,$FB,$A9,$FF,$5F
	.dc.b	$CE,$5F,$F5,$B7,$6F,$F9,$53,$EB,$BF,$9A,$C6,$0D,$F9,$CC,$39,$D7
	.dc.b	$FB,$1E,$F9,$61,$D6,$D1,$89,$E9,$E3,$B3,$E3,$5D,$30,$AF,$0C,$7B
	.dc.b	$3E,$13,$11,$C3,$95,$9F,$9B,$3F,$06,$7D,$EF,$A1,$C5,$C5,$70,$CB
	.dc.b	$C5,$39,$C7,$BD,$F2,$6E,$7E,$D6,$CF,$F9,$D7,$1F,$04,$E7,$2F,$4E
	.dc.b	$5E,$1A,$F3,$9F,$27,$2F,$06,$7A,$C7,$9C,$EB,$2F,$06,$1E,$DC,$BA
	.dc.b	$CF,$AA,$72,$F9,$E7,$59,$FA,$A7,$24,$CF,$C9,$5E,$1B,$E6,$F7,$8C
	.dc.b	$CE,$E3,$D9,$5C,$E7,$24,$8F,$13,$F6,$B1,$F5,$61,$27,$C8,$77,$AE
	.dc.b	$4F,$03,$E9,$9F,$ED,$0C,$FF,$20,$DC,$BE,$FB,$6B,$F9,$AB,$F8,$AF
	.dc.b	$AE,$7C,$46,$2D,$18,$9F,$CF,$B5,$EC,$D7,$F2,$CE,$F3,$FC,$3B,$73
	.dc.b	$CD,$FE,$37,$D7,$7D,$63,$0F,$BB,$7B,$7F,$52,$D7,$CD,$5C,$67,$84
	.dc.b	$7E,$46,$75,$8F,$A2,$B9,$9B,$9A,$6D,$B8,$FE,$DF,$76,$3E,$4A,$CE
	.dc.b	$5E,$87,$81,$F8,$F8,$D3,$8A,$F7,$DF,$08,$FF,$67,$4E,$A7,$31,$BA
	.dc.b	$E9,$1E,$19,$D4,$60,$93,$F0,$AB,$D4,$F1,$30,$5B,$F4,$E5,$F1,$30
	.dc.b	$EE,$4B,$E7,$3A,$AF,$31,$F3,$0E,$58,$78,$9E,$B5,$C5,$B5,$FE,$EC
	.dc.b	$9D,$73,$7F,$5E,$F8,$AF,$C5,$B6,$BF,$A4,$C6,$39,$8F,$E4,$39,$BF
	.dc.b	$98,$61,$79,$BF,$8E,$5F,$F4,$98,$C1,$C7,$EB,$64,$7F,$54,$6B,$FD
	.dc.b	$11,$AF,$F6,$E6,$BF,$D0,$DA,$FF,$A3,$30,$46,$2D,$F4,$4F,$B1,$F3
	.dc.b	$36,$FF,$C3,$73,$A9,$C9,$EB,$FA,$F7,$AD,$49,$5C,$23,$2F,$F8,$93
	.dc.b	$81,$F9,$DF,$F2,$68,$FC,$D8,$8C,$F4,$FE,$D9,$B1,$C6,$78,$3F,$A2
	.dc.b	$8F,$EF,$99,$EF,$C0,$F6,$CF,$F8,$56,$30,$47,$04,$3F,$C0,$CF,$DA
	.dc.b	$F1,$3F,$3B,$1F,$E8,$46,$BF,$E1,$86,$BF,$EB,$66,$D7,$F8,$37,$C2
	.dc.b	$F1,$12,$7F,$C4,$CF,$5F,$6E,$47,$F8,$D1,$AF,$F8,$C9,$AF,$85,$AF
	.dc.b	$F8,$A3,$C7,$7C,$07,$E4,$1B,$1F,$E4,$98,$C1,$1C,$10,$FF,$25,$3F
	.dc.b	$93,$FA,$E7,$B9,$49,$3F,$80,$8C,$BF,$EB,$8F,$CF,$3F,$D3,$6D,$F5
	.dc.b	$7F,$95,$FB,$29,$CC,$FF,$B1,$5A,$FF,$AE,$9A,$FF,$21,$8E,$13,$CA
	.dc.b	$73,$29,$CB,$F9,$2F,$AB,$FF,$6D,$B7,$DD,$FD,$64,$62,$3F,$D3,$CD
	.dc.b	$7F,$DE,$2D,$7F,$DF,$0D,$7F,$CE,$CD,$E7,$39,$ED,$7D,$47,$C2,$AF
	.dc.b	$F7,$1B,$5F,$F6,$4B,$5F,$F2,$EB,$5F,$F7,$DB,$63,$9E,$9F,$C2,$23
	.dc.b	$06,$ED,$A3,$E0,$FE,$B1,$30,$FA,$8F,$F5,$A5,$ED,$7F,$A0,$5F,$EB
	.dc.b	$AD,$7F,$D2,$6D,$7F,$5E,$D7,$FA,$5B,$5F,$EA,$8D,$7A,$FA,$3D,$E3
	.dc.b	$59,$7F,$8C,$BC,$FF,$E4,$3E,$3F,$F9,$9E,$FF,$D0,$4D,$7F,$7B,$3B
	.dc.b	$73,$3C,$08,$E0,$FF,$45,$7C,$1F,$FA,$E1,$BE,$B7,$F2,$E8,$C4,$74
	.dc.b	$6B,$FE,$FA,$6B,$FE,$8E,$6A,$A4,$61,$B3,$F8,$AC,$A5,$7E,$3D,$AF
	.dc.b	$F1,$6E,$17,$D6,$FE,$C2,$30,$D7,$FB,$AC,$6B,$97,$F4,$B3,$99,$FD
	.dc.b	$03,$5F,$CF,$36,$78,$47,$FB,$29,$AF,$CA,$FA,$77,$D6,$FF,$50,$6C
	.dc.b	$7F,$AC,$5A,$EE,$34,$47,$4D,$7E,$AF,$C2,$3E,$2F,$FA,$DB,$63,$FD
	.dc.b	$78,$D7,$71,$A2,$3A,$6B,$FB,$2E,$6F,$DA,$91,$FD,$D3,$F8,$BF,$82
	.dc.b	$91,$F9,$86,$BF,$98,$6B,$F9,$86,$BF,$98,$8F,$CA,$11,$FD,$0B,$FA
	.dc.b	$5E,$67,$C1,$52,$4F,$F9,$0A,$32,$DF,$6B,$81,$1C,$1B,$F8,$0E,$84
	.dc.b	$70,$73,$F8,$BF,$01,$1C,$10,$FE,$BE,$7F,$0C,$74,$E2,$91,$FE,$0C
	.dc.b	$F4,$1F,$53,$49,$5F,$EE,$F8,$CB,$FD,$B7,$89,$F3,$DF,$DA,$5B,$19
	.dc.b	$8C,$11,$C1,$0F,$EE,$7E,$2F,$F4,$E8,$FF,$0E,$FD,$43,$E1,$A9,$2B
	.dc.b	$FE,$00,$46,$5F,$EF,$DF,$75,$FC,$65,$BD,$1F,$DE,$E3,$11,$D9,$AF
	.dc.b	$F7,$B6,$BF,$E6,$5C,$11,$88,$FF,$03,$36,$33,$3F,$E3,$2E,$3F,$E2
	.dc.b	$15,$FE,$66,$EA,$FC,$E7,$66,$49,$1F,$E1,$3F,$91,$FE,$55,$81,$A4
	.dc.b	$8F,$F0,$93,$5E,$06,$0F,$F2,$0F,$84,$FA,$7A,$48,$FB,$F8,$27,$51
	.dc.b	$FE,$23,$FB,$A7,$C8,$7F,$87,$DB,$19,$8C,$11,$C1,$0F,$AF,$1C,$1F
	.dc.b	$E5,$FF,$2B,$FE,$26,$7F,$D1,$D6,$8C,$47,$F2,$36,$C6,$7D,$2F,$F1
	.dc.b	$87,$C6,$7F,$0A,$AF,$F1,$73,$5F,$A8,$D7,$FC,$70,$DD,$0D,$27,$7F
	.dc.b	$43,$FC,$CF,$F4,$EF,$88,$A4,$BF,$8F,$19,$7F,$C9,$1F,$32,$F9,$3F
	.dc.b	$F2,$23,$63,$31,$83,$FD,$7B,$F3,$4F,$A7,$E5,$3F,$E4,$E8,$C6,$1A
	.dc.b	$9C,$CF,$FA,$B2,$72,$FE,$D9,$AF,$18,$DB,$FE,$B9,$8C,$BC,$63,$2F
	.dc.b	$F4,$F8,$7D,$8F,$FA,$19,$B1,$F8,$46,$BF,$D9,$9A,$FE,$65,$AF,$DA
	.dc.b	$6B,$D6,$3A,$1F,$E8,$DF,$A0,$7C,$D5,$24,$FF,$9E,$A3,$2D,$8C,$BF
	.dc.b	$E7,$2E,$44,$EA,$30,$D7,$FD,$A3,$F0,$EF,$93,$FB,$F1,$FC,$72,$31
	.dc.b	$C1,$FE,$EE,$6C,$66,$76,$F4,$FA,$5F,$41,$1F,$33,$5D,$18,$21,$FC
	.dc.b	$AB,$E9,$1F,$48,$FA,$2F,$F7,$EC,$61,$B1,$99,$E0,$D7,$FD,$23,$2F
	.dc.b	$A4,$6F,$C9,$B3,$DD,$AF,$06,$BF,$CF,$3E,$33,$FC,$68,$9A,$47,$E7
	.dc.b	$7D,$87,$FA,$C3,$A5,$1E,$96,$F0,$27,$53,$91,$B6,$7C,$E5,$F2,$D6
	.dc.b	$D7,$F0,$FB,$E3,$18,$C3,$A7,$C4,$EA,$31,$3A,$BF,$1E,$3D,$5E,$87
	.dc.b	$92,$35,$87,$8B,$E5,$DF,$69,$FC,$3E,$D1,$8B,$FC,$23,$6B,$FD,$7F
	.dc.b	$3A,$AF,$E2,$D9,$7B,$F9,$7C,$98,$77,$6D,$F7,$AB,$5B,$AF,$2C,$FB
	.dc.b	$EB,$38,$FE,$8A,$73,$87,$77,$77,$C6,$FE,$67,$77,$E2,$7E,$6F,$BB
	.dc.b	$6B,$AF,$7F,$FB,$63,$EB,$33,$B2,$E9,$2B,$7F,$14,$AC,$63,$DB,$ED
	.dc.b	$DF,$35,$D9,$B7,$E4,$FA,$97,$CA,$66,$A4,$FB,$B5,$5F,$0C,$3F,$09
	.dc.b	$B3,$1E,$37,$BA,$D6,$BF,$84,$D9,$E1,$FF,$7B,$AF,$E5,$BE,$87,$F4
	.dc.b	$79,$CB,$67,$59,$7F,$9E,$BE,$B7,$FD,$FF,$7E,$36,$C7,$E5,$6F,$8C
	.dc.b	$E7,$E4,$7E,$CC,$3E,$9F,$F5,$D3,$FE,$22,$8C,$76,$78,$67,$37,$FE
	.dc.b	$12,$6B,$FE,$FF,$6B,$FE,$05,$6A,$9E,$75,$D6,$73,$BB,$E6,$7A,$EB
	.dc.b	$FA,$68,$C6,$3E,$6A,$CD,$FF,$68,$BB,$8F,$5F,$A7,$EF,$91,$A4,$C3
	.dc.b	$E5,$36,$FD,$CD,$AF,$F5,$1F,$DB,$BE,$03,$F1,$ED,$AF,$C7,$D7,$83
	.dc.b	$34,$E6,$FE,$77,$E7,$BF,$C2,$2F,$E5,$36,$3F,$C7,$2D,$7F,$86,$9F
	.dc.b	$15,$C3,$FE,$2D,$6B,$93,$F0,$4E,$32,$3A,$B7,$B1,$FC,$4A,$FF,$28
	.dc.b	$DB,$F2,$9F,$AA,$71,$B7,$E4,$7E,$A7,$F6,$93,$FE,$4E,$6B,$E6,$C5
	.dc.b	$18,$8F,$F2,$13,$5F,$BC,$D8,$DF,$77,$E6,$63,$06,$F1,$7D,$8F,$FA
	.dc.b	$A3,$2F,$D5,$37,$0F,$CC,$57,$93,$1F,$C2,$4E,$5F,$D4,$7B,$9F,$BB
	.dc.b	$BF,$F3,$77,$67,$E9,$E3,$18,$7F,$9A,$1A,$FA,$DA,$FF,$96,$DA,$FE
	.dc.b	$5D,$B9,$9C,$E4,$FB,$2F,$F0,$4E,$1F,$C0,$5B,$7F,$AF,$AF,$0E,$3F
	.dc.b	$9B,$9C,$BF,$BF,$FC,$29,$F0,$FF,$BD,$6E,$1F,$B1,$6E,$1F,$E8,$06
	.dc.b	$BF,$4D,$AF,$F9,$CD,$AF,$EB,$DB,$91,$CE,$4F,$96,$FF,$66,$5F,$F1
	.dc.b	$96,$DF,$F0,$7A,$F0,$E3,$FB,$39,$CB,$FC,$5F,$F0,$A7,$C3,$FF,$34
	.dc.b	$9F,$E1,$F3,$98,$C7,$57,$F0,$EA,$F0,$63,$FC,$35,$B8,$FF,$A7,$1A
	.dc.b	$FE,$3D,$AF,$F9,$9B,$F0,$F7,$D6,$FF,$42,$6D,$7F,$31,$7F,$95,$BB
	.dc.b	$8F,$EA,$B2,$FA,$C7,$D0,$FF,$9B,$9B,$5F,$EC,$09,$D4,$E4,$91,$99
	.dc.b	$DC,$FF,$B1,$9A,$FE,$E8,$E3,$1F,$EC,$AF,$11,$F9,$F3,$E9,$FF,$B2
	.dc.b	$B6,$7F,$B2,$D7,$03,$FA,$3C,$65,$FE,$C3,$F9,$E7,$F0,$9D,$FF,$92
	.dc.b	$36,$BF,$D1,$2F,$F2,$97,$FA,$93,$99,$71,$23,$0D,$8F,$ED,$71,$C5
	.dc.b	$EE,$7F,$6E,$FA,$F7,$D7,$D3,$1F,$1B,$70,$FE,$C1,$3A,$6C,$E9,$A7
	.dc.b	$F7,$CC,$BF,$D2,$2D,$F6,$FF,$C0,$FF,$6D,$F8,$A8,$C1,$31,$F8,$73
	.dc.b	$AE,$5F,$92,$C9,$FE,$E1,$F2,$FF,$C2,$9E,$1B,$7F,$49,$8C,$5F,$F8
	.dc.b	$2B,$D7,$7C,$E7,$F8,$29,$FE,$D3,$5F,$D3,$DC,$DF,$F7,$B7,$2F,$FA
	.dc.b	$D9,$CD,$73,$F3,$79,$5F,$37,$FC,$2B,$F3,$4F,$B3,$BE,$AE,$1B,$8F
	.dc.b	$F8,$1E,$74,$D9,$D3,$4F,$C2,$E6,$F9,$EF,$F6,$2B,$63,$F3,$0D,$CB
	.dc.b	$F2,$41,$C3,$F9,$56,$BF,$D4,$1A,$FF,$97,$9A,$F1,$9F,$F2,$0F,$B0
	.dc.b	$F8,$0F,$F1,$C9,$FA,$CA,$F0,$6D,$FE,$56,$8F,$F1,$AF,$5B,$E5,$29
	.dc.b	$8F,$76,$E7,$FE,$27,$9D,$36,$74,$D3,$FC,$87,$8B,$E9,$3F,$53,$E4
	.dc.b	$7F,$92,$9B,$3A,$6E,$BF,$E3,$0A,$E1,$3F,$91,$AF,$F2,$BF,$CD,$7D
	.dc.b	$5D,$BF,$DE,$4D,$9F,$92,$DB,$FD,$E3,$5F,$EE,$ED,$7F,$D0,$AD,$4D
	.dc.b	$31,$DF,$FE,$87,$FA,$EF,$F3,$09,$E2,$E3,$F9,$37,$FE,$79,$8F,$F2
	.dc.b	$EF,$D1,$FF,$48,$8F,$F4,$7F,$A9,$FC,$C1,$FF,$49,$F8,$29,$D6,$4C
	.dc.b	$7F,$C3,$1F,$29,$CB,$CE,$FF,$39,$E3,$C6,$BF,$81,$DF,$0B,$DE,$3F
	.dc.b	$E5,$9F,$A5,$FF,$0D,$73,$7F,$A4,$FE,$BF,$AE,$89,$CD,$7F,$A8,$A7
	.dc.b	$31,$8C,$7C,$51,$8C,$BF,$D0,$6F,$9B,$CE,$7D,$3F,$E0,$2B,$F9,$44
	.dc.b	$63,$9B,$FC,$BD,$39,$9F,$F0,$83,$5F,$83,$F5,$5F,$C1,$DF,$F4,$2B
	.dc.b	$74,$FF,$35,$4F,$57,$FC,$95,$19,$CF,$8F,$E0,$7F,$E8,$A8,$37,$C3
	.dc.b	$C3,$27,$FC,$F9,$3A,$6C,$E7,$0F,$F3,$66,$0E,$3F,$49,$2B,$FC,$D5
	.dc.b	$19,$7F,$6B,$BB,$0F,$D6,$47,$6E,$CF,$F3,$9D,$6F,$0F,$6C,$7F,$AB
	.dc.b	$3B,$1F,$39,$FE,$B4,$6D,$FF,$4F,$6D,$7F,$53,$E3,$46,$23,$FC,$9F
	.dc.b	$97,$87,$E6,$3E,$11,$D7,$0B,$87,$C0,$8B,$5F,$E9,$AB,$EB,$BF,$CC
	.dc.b	$9C,$B6,$BF,$D7,$33,$A7,$EB,$CF,$A7,$D4,$F4,$12,$3D,$0F,$FB,$92
	.dc.b	$FC,$BA,$FF,$8F,$E7,$EF,$FB,$1C,$5F,$A6,$FF,$7C,$FA,$47,$D7,$FF
	.dc.b	$A2,$6F,$FD,$F9,$E1,$7F,$9B,$27,$3F,$01,$FE,$FA,$F8,$CF,$F0,$0D
	.dc.b	$7F,$A4,$FE,$C3,$FB,$74,$62,$15,$C1,$E6,$0C,$E1,$DE,$78,$4E,$DE
	.dc.b	$BD,$2B,$EB,$8F,$8E,$F8,$37,$1F,$23,$71,$EF,$7C,$7D,$36,$9A,$9F
	.dc.b	$16,$7D,$EF,$86,$7E,$2C,$BA,$D7,$B7,$0E,$2F,$A8,$F4,$E5,$CD,$A6
	.dc.b	$70,$E8,$F1,$C7,$51,$88,$DE,$1D,$EB,$C9,$97,$9E,$F7,$8F,$8A,$F7
	.dc.b	$97,$98,$F9,$55,$B8,$F0,$4E,$F1,$EC,$78,$A7,$58,$F7,$AD,$E3,$D8
	.dc.b	$F0,$E7,$F3,$A3,$E2,$E1,$CC,$93,$AC,$FE,$1D,$F3,$26,$5E,$5B,$F8
	.dc.b	$07,$62,$46,$9C,$B5,$F9,$27,$F9,$07,$2F,$C4,$37,$0F,$CA,$37,$0F
	.dc.b	$F2,$74,$BE,$C1,$F1,$98,$6D,$7F,$7A,$9D,$4E,$49,$1F,$5E,$77,$87
	.dc.b	$E4,$9B,$8F,$CE,$FA,$17,$D8,$7F,$B3,$AB,$58,$7D,$CB,$DE,$1F,$8C
	.dc.b	$6E,$CC,$38,$CE,$EA,$3E,$17,$F0,$31,$A8,$C1,$2F,$BB,$F9,$C8,$D4
	.dc.b	$60,$98,$FE,$DF,$0F,$0C,$EA,$31,$7E,$0A,$DD,$B0,$F3,$63,$ED,$BD
	.dc.b	$CF,$85,$B5,$FB,$4C,$3A,$4F,$BF,$0E,$95,$A3,$E4,$5F,$39,$D1,$CC
	.dc.b	$B7,$CE,$35,$1F,$C2,$31,$F9,$6D,$BE,$AF,$9F,$0F,$0B,$93,$A1,$6F
	.dc.b	$A3,$C0,$C6,$1E,$57,$96,$1E,$5E,$7F,$E9,$5C,$BF,$C5,$67,$FB,$1B
	.dc.b	$79,$BF,$B0,$B7,$8B,$F8,$D7,$47,$F4,$4F,$80,$C5,$FF,$82,$6B,$FD
	.dc.b	$E6,$FF,$4B,$8F,$F5,$24,$EE,$30,$7F,$5A,$E6,$FF,$1F,$3F,$D9,$BB
	.dc.b	$F9,$E5,$FD,$75,$EB,$4F,$C2,$5F,$71,$AC,$38,$E1,$FD,$EA,$F0,$70
	.dc.b	$AD,$C7,$52,$1B,$3F,$9A,$69,$FC,$7B,$0E,$D8,$71,$3D,$35,$CA,$BF
	.dc.b	$B2,$35,$F9,$13,$C1,$F6,$7D,$37,$94,$8F,$6D,$7F,$8C,$F7,$FF,$94
	.dc.b	$E3,$07,$E4,$DC,$3F,$E1,$C6,$BF,$E1,$AF,$3B,$FD,$9D,$CD,$FE,$22
	.dc.b	$7E,$6F,$D8,$3E,$97,$E3,$B6,$BA,$E4,$8C,$43,$5F,$81,$1F,$B1,$6D
	.dc.b	$72,$24,$F2,$9D,$65,$F8,$26,$C7,$F9,$25,$B7,$DF,$8F,$F0,$B9,$7F
	.dc.b	$8E,$2A,$E7,$FE,$E7,$7E,$01,$EC,$71,$1B,$8F,$F0,$FF,$91,$FD,$B7
	.dc.b	$3F,$AD,$F9,$FB,$D5,$FC,$2B,$6B,$FC,$A5,$CF,$F8,$AB,$E1,$AF,$E6
	.dc.b	$2B,$55,$F0,$63,$17,$BB,$FC,$8F,$96,$FB,$4F,$C8,$F9,$0F,$97,$A6
	.dc.b	$A2,$63,$F3,$A7,$4D,$9C,$BD,$CF,$F3,$77,$E3,$4F,$86,$FC,$EB,$63
	.dc.b	$EB,$B5,$E1,$E9,$BE,$FD,$F5,$9F,$B0,$8C,$3F,$EB,$89,$FF,$44,$CE
	.dc.b	$9F,$F4,$9B,$5F,$F4,$9B,$5F,$F4,$9B,$5F,$F4,$9B,$5E,$71,$D6,$7B
	.dc.b	$C7,$FB,$1B,$95,$EC,$7F,$56,$E3,$FD,$57,$BF,$FA,$76,$F6,$FE,$EA
	.dc.b	$75,$C7,$FC,$29,$B8,$7F,$A4,$6F,$84,$FF,$0E,$6C,$7F,$A5,$67,$7E
	.dc.b	$03,$EB,$1E,$AF,$F0,$86,$D7,$EB,$27,$53,$9F,$B0,$D3,$27,$55,$C1
	.dc.b	$FA,$D5,$D6,$BD,$1F,$01,$27,$FD,$96,$FF,$20,$5E,$FF,$CD,$D7,$FC
	.dc.b	$A4,$FC,$98,$CF,$07,$E7,$E3,$04,$70,$48,$FB,$51,$F4,$70,$FF,$25
	.dc.b	$F2,$FF,$24,$F3,$1E,$C5,$31,$12,$FE,$04,$E9,$B3,$97,$DB,$EF,$3E
	.dc.b	$9F,$F9,$7A,$E1,$A6,$22,$4F,$CB,$9D,$36,$72,$FE,$77,$93,$97,$CE
	.dc.b	$FF,$3A,$FB,$FF,$E3,$7B,$FF,$92,$B5,$46,$39,$7E,$F5,$7F,$3E,$C4
	.dc.b	$D2,$63,$FC,$6B,$F3,$DF,$AD,$8F,$E3,$A7,$33,$E1,$AE,$2E,$F9,$7F
	.dc.b	$A5,$FB,$0F,$83,$A4,$9F,$C3,$43,$E0,$71,$1F,$82,$8E,$0E,$A7,$9C
	.dc.b	$FF,$56,$FB,$0F,$F1,$7C,$FE,$21,$B9,$7B,$DB,$A7,$FB,$C3,$F2,$F7
	.dc.b	$DC,$71,$F5,$E6,$1F,$97,$BD,$AE,$15,$F0,$8F,$E7,$7F,$13,$FE,$9F
	.dc.b	$CB,$FA,$C3,$72,$FE,$EA,$D9,$FF,$48,$7E,$01,$FE,$BA,$7F,$A7,$87
	.dc.b	$0F,$F8,$04,$DF,$AF,$FC,$D6,$33,$19,$6B,$FE,$07,$6B,$C0,$FE,$24
	.dc.b	$E1,$ED,$1F,$C5,$71,$FF,$4E,$70,$FF,$7E,$F7,$BF,$17,$5F,$B0,$6E
	.dc.b	$5F,$87,$7F,$2D,$1F,$D9,$BC,$EF,$F7,$7F,$81,$FD,$8B,$E1,$0D,$24
	.dc.b	$FA,$CF,$F1,$F4,$3E,$2F,$FC,$1B,$F2,$7F,$B7,$DF,$FD,$69,$A2,$31
	.dc.b	$7C,$2B,$FC,$27,$F8,$9B,$EC,$69,$27,$C9,$F1,$7F,$EF,$96,$F8,$7F
	.dc.b	$DD,$A3,$11,$F3,$75,$46,$23,$FC,$61,$18,$9E,$04,$8F,$54,$65,$B3
	.dc.b	$C5,$F4,$FC,$BB,$E1,$BF,$C5,$2D,$8F,$E4,$4D,$7F,$A9,$53,$E3,$5F
	.dc.b	$0F,$FD,$D9,$B3,$E0,$6B,$E9,$6B,$E8,$6B,$FE,$B4,$6B,$FE,$6D,$6C
	.dc.b	$F7,$9F,$F0,$CC,$6B,$67,$C1,$7F,$92,$2F,$F9,$8B,$79,$3F,$84,$37
	.dc.b	$87,$FD,$83,$F6,$8F,$8F,$FF,$06,$B6,$FE,$3B,$5F,$A0,$D7,$E7,$B5
	.dc.b	$E7,$1F,$E2,$C8,$C4,$7F,$A1,$5B,$99,$CE,$2F,$C0,$7F,$84,$57,$F4
	.dc.b	$F6,$BF,$D2,$9B,$97,$FB,$AA,$FE,$B7,$B5,$FE,$24,$DF,$FC,$27,$7F
	.dc.b	$82,$6B,$FE,$DE,$6B,$FE,$97,$6C,$F2,$8F,$C2,$C6,$09,$3F,$AF,$C3
	.dc.b	$FC,$27,$3F,$D7,$1B,$7F,$CC,$DB,$9F,$FB,$B3,$EC,$DF,$6F,$FE,$38
	.dc.b	$6D,$FE,$49,$AF,$E6,$DA,$FE,$69,$AF,$38,$FF,$2B,$C6,$23,$FD,$60
	.dc.b	$DC,$8E,$6D,$F1,$7F,$C8,$9B,$E1,$7F,$A1,$23,$F9,$2D,$3E,$4F,$E6
	.dc.b	$E4,$C7,$FD,$79,$1F,$49,$FD,$6B,$5D,$4F,$1F,$17,$FC,$D7,$39,$9E
	.dc.b	$31,$82,$3F,$B2,$8D,$62,$FA,$37,$DF,$FF,$A6,$1B,$5F,$E1,$16,$BF
	.dc.b	$B0,$FA,$5F,$49,$FD,$21,$FE,$02,$D7,$CA,$D7,$FD,$AC,$DA,$FF,$6C
	.dc.b	$7A,$3F,$EA,$6C,$BF,$92,$D7,$F8,$61,$BC,$1F,$E9,$BF,$37,$FA,$44
	.dc.b	$E6,$B8,$11,$FF,$42,$C6,$A7,$FB,$B7,$D5,$BE,$DF,$FD,$66,$DA,$FF
	.dc.b	$16,$35,$FE,$17,$F5,$9F,$C2,$DC,$CF,$02,$3F,$CE,$63,$4F,$F8,$1F
	.dc.b	$DB,$7C,$67,$F9,$14,$91,$FE,$A4,$E4,$7E,$71,$FE,$78,$CB,$FD,$6D
	.dc.b	$7C,$2B,$FD,$3A,$D7,$A5,$FF,$B5,$2F,$8F,$45,$C3,$F5,$71,$8A,$FE
	.dc.b	$8C,$D7,$FB,$13,$5F,$F5,$ED,$FF,$B3,$B0,$F0,$FC,$7F,$F9,$6C,$ED
	.dc.b	$F0,$5F,$EA,$C8,$CD,$79,$27,$4F,$AE,$75,$CF,$FE,$EF,$FA,$E7,$D7
	.dc.b	$FF,$AC,$EB,$FA,$C4,$63,$0F,$F3,$E3,$63,$33,$B7,$FB,$C4,$3E,$A9
	.dc.b	$C3,$C1,$B3,$D1,$AF,$06,$BF,$D9,$FE,$53,$FD,$68,$9A,$47,$FC,$0D
	.dc.b	$EE,$7F,$A0,$3D,$26,$7F,$49,$B3,$FE,$9F,$DF,$FD,$0B,$F0,$58,$D7
	.dc.b	$EC,$D5,$BF,$F4,$13,$6F,$FC,$FC,$DB,$E3,$3D,$4F,$C2,$7C,$47,$F9
	.dc.b	$19,$FB,$2D,$AF,$B2,$D7,$F8,$20,$9F,$F5,$D7,$3F,$9A,$3E,$EF,$87
	.dc.b	$E1,$CF,$FD,$36,$F8,$6F,$A6,$5F,$90,$8D,$71,$FF,$6F,$CB,$F2,$31
	.dc.b	$BC,$72,$FA,$FA,$FF,$DD,$DE,$D5,$C3,$FC,$A5,$7F,$A3,$F9,$CC,$4C
	.dc.b	$8F,$F4,$C4,$EB,$4F,$D6,$D6,$5F,$F4,$AF,$D6,$3E,$BB,$B3,$6F,$FD
	.dc.b	$FF,$F5,$CF,$AF,$CD,$49,$FF,$52,$56,$EF,$FC,$F1,$7B,$9C,$D7,$D3
	.dc.b	$F0,$5E,$17,$FB,$BE,$7F,$C2,$6D,$BF,$F0,$E3,$5F,$E3,$7D,$DF,$35
	.dc.b	$FF,$16,$F8,$7F,$ED,$FA,$FE,$49,$F1,$BD,$6F,$96,$DF,$E8,$5A,$D5
	.dc.b	$FF,$85,$BE,$BF,$F3,$98,$7F,$88,$9B,$1F,$E2,$B6,$BF,$E2,$A6,$BF
	.dc.b	$CF,$9A,$B2,$46,$1B,$D7,$E7,$9D,$4F,$0E,$2F,$EA,$73,$F6,$FC,$9F
	.dc.b	$0E,$5F,$DE,$EA,$E3,$FE,$4A,$9B,$87,$7F,$AF,$EC,$4C,$3B,$64,$8C
	.dc.b	$5F,$F8,$F9,$B5,$FE,$47,$7C,$5F,$5F,$FE,$DE,$9F,$F6,$E3,$67,$FC
	.dc.b	$77,$3E,$0C,$D3,$9C,$FF,$C2,$1F,$5F,$FD,$97,$0F,$F2,$5B,$70,$FF
	.dc.b	$2A,$35,$FF,$29,$B5,$FE,$CE,$D7,$E5,$37,$13,$4D,$60,$FA,$FF,$D8
	.dc.b	$5F,$FB,$CD,$B7,$FE,$52,$AF,$16,$1F,$EC,$59,$CB,$FE,$27,$FA,$FF
	.dc.b	$D7,$57,$F9,$A5,$AF,$F9,$A1,$AF,$F7,$C6,$BF,$7D,$B8,$9A,$4E,$9F
	.dc.b	$C0,$65,$FE,$52,$F2,$FF,$71,$EA,$7F,$B9,$74,$FF,$48,$5F,$8B,$2F
	.dc.b	$F2,$9D,$7E,$97,$D5,$75,$FD,$C6,$DF,$F9,$A6,$30,$47,$04,$D3,$FC
	.dc.b	$D9,$1E,$2F,$5D,$F6,$5F,$C2,$5B,$5F,$EA,$1B,$EB,$6C,$7F,$DD,$95
	.dc.b	$A8,$FF,$3E,$FD,$83,$E2,$7F,$1E,$DC,$3F,$D1,$6D,$9F,$F4,$A3,$5F
	.dc.b	$F4,$9B,$5F,$F1,$3B,$5F,$D5,$37,$80,$E7,$7F,$F5,$57,$D4,$FF,$21
	.dc.b	$F8,$13,$FC,$2E,$7C,$19,$FF,$46,$AD,$E3,$FE,$BD,$FB,$07,$D5,$71
	.dc.b	$EE,$FE,$25,$18,$9F,$F1,$F5,$F1,$CB,$F8,$3B,$FE,$6D,$9F,$F5,$D4
	.dc.b	$FF,$AA,$1B,$A7,$FA,$C5,$AF,$FA,$C1,$AF,$FB,$0B,$EC,$1F,$61,$FC
	.dc.b	$E5,$B5,$E0,$35,$E1,$66,$7F,$AC,$53,$EC,$1F,$3D,$FC,$2E,$30,$D8
	.dc.b	$CD,$FF,$B5,$67,$2D,$8C,$C7,$FB,$65,$AF,$FB,$D5,$AF,$F5,$C6,$AF
	.dc.b	$31,$F1,$BF,$EB,$CA,$F4,$4E,$63,$0F,$E2,$63,$95,$FF,$61,$9C,$E9
	.dc.b	$F5,$7E,$C3,$F6,$B5,$FE,$E6,$6D,$B8,$51,$FE,$78,$9E,$F1,$89,$FE
	.dc.b	$C5,$3C,$23,$13,$A8,$E1,$3E,$33,$A3,$BF,$AF,$FF,$7C,$4F,$46,$CE
	.dc.b	$E9,$0F,$B0,$FE,$E0,$EB,$0F,$C7,$36,$FF,$DF,$ED,$BF,$53,$C6,$3B
	.dc.b	$7C,$5F,$C3,$F8,$09,$1F,$2B,$57,$C7,$FE,$4F,$0F,$44,$7D,$0A,$DC
	.dc.b	$63,$5F,$F4,$B5,$F1,$BE,$19,$7F,$89,$B0,$E1,$3D,$6B,$77,$ED,$F9
	.dc.b	$7F,$F7,$2C,$E7,$1F,$F0,$64,$62,$7E,$17,$AF,$FE,$E3,$AF,$A2,$FF
	.dc.b	$6C,$7F,$D4,$9B,$23,$53,$C3,$CE,$FF,$0B,$46,$33,$FD,$A3,$C6,$7E
	.dc.b	$F3,$9B,$E6,$E6,$DF,$34,$F8,$4F,$F1,$03,$FE,$19,$C3,$F7,$CE,$6F
	.dc.b	$F0,$AE,$5F,$2B,$9C,$3E,$E3,$97,$F1,$CE,$67,$9F,$A9,$FB,$0C,$DF
	.dc.b	$17,$FA,$F6,$E1,$FA,$D6,$E1,$FE,$42,$3F,$D8,$CE,$1C,$35,$FF,$23
	.dc.b	$B5,$DC,$F5,$3F,$C0,$DB,$BD,$77,$D8,$7F,$91,$0F,$24,$DC,$3F,$47
	.dc.b	$CE,$F8,$1F,$F2,$23,$63,$3E,$E6,$26,$47,$07,$71,$8C,$BF,$D5,$1F
	.dc.b	$53,$FB,$7F,$23,$FC,$A6,$DB,$FF,$51,$B6,$77,$1F,$8D,$9F,$F3,$77
	.dc.b	$D0,$7D,$7C,$BF,$C4,$0D,$9F,$8D,$48,$C3,$F2,$5A,$FF,$9B,$1A,$FF
	.dc.b	$BC,$5A,$F1,$E2,$F9,$51,$8D,$BE,$47,$C7,$FF,$9A,$8F,$1C,$FD,$6E
	.dc.b	$2F,$F5,$EC,$7F,$9A,$7E,$8F,$FA,$FC,$7E,$9F,$E6,$BF,$A5,$E5,$FC
	.dc.b	$3A,$72,$FF,$C0,$1C,$13,$AC,$BF,$83,$56,$9F,$E4,$D1,$9B,$FE,$A5
	.dc.b	$39,$AE,$3E,$4F,$F9,$E7,$E9,$7F,$C5,$58,$F5,$F9,$EE,$5D,$29,$CE
	.dc.b	$7F,$CC,$67,$31,$8C,$11,$8C,$FF,$D1,$CF,$8F,$CC,$7D,$37,$E0,$F0
	.dc.b	$FC,$B4,$63,$97,$FC,$8F,$39,$9F,$F4,$DC,$E7,$3F,$AF,$9B,$EA,$FF
	.dc.b	$86,$C7,$F0,$06,$D7,$F9,$86,$FB,$6F,$FD,$C6,$73,$7C,$7E,$63,$FD
	.dc.b	$31,$06,$F9,$D8,$66,$3F,$CF,$A7,$4D,$9C,$E3,$FC,$EF,$2F,$6F,$17
	.dc.b	$A8,$95,$F8,$E8,$CB,$EA,$BF,$D5,$BD,$7A,$BF,$D2,$5A,$2F,$CB,$1F
	.dc.b	$EB,$7E,$97,$D8,$7E,$D1,$B7,$F9,$AE,$24,$63,$3F,$F6,$CB,$5F,$F6
	.dc.b	$2E,$DF,$27,$EC,$3E,$09,$D7,$0B,$87,$BE,$2D,$7F,$40,$BE,$9C,$1F
	.dc.b	$2E,$72,$DA,$FE,$91,$3A,$7E,$A4,$FA,$7D,$2F,$41,$23,$D1,$3F,$E2
	.dc.b	$BD,$BF,$CA,$11,$F6,$FD,$9E,$0F,$D1,$7F,$C2,$3F,$46,$FB,$07,$C0
	.dc.b	$B4,$FF,$7D,$78,$3F,$EE,$2A,$D4,$E7,$E2,$FF,$C2,$F5,$FE,$A5,$F9
	.dc.b	$EF,$F0,$2C,$62,$15,$C1,$E6,$75,$71,$1B,$6D,$F8,$79,$97,$D2,$3C
	.dc.b	$58,$F9,$B0,$E6,$76,$71,$87,$36,$CF,$3A,$E5,$5E,$19,$CB,$71,$F6
	.dc.b	$57,$66,$C7,$B7,$E6,$DA,$62,$3C,$13,$73,$F4,$C7,$33,$DF,$9F,$AA
	.dc.b	$F8,$C7,$A2,$76,$F2,$31,$8F,$92,$B8,$61,$D6,$3C,$F3,$BC,$FD,$D8
	.dc.b	$F9,$AF,$73,$CA,$38,$11,$F1,$E7,$F0,$9E,$59,$FB,$DE,$06,$B3,$F3
	.dc.b	$E1,$DA,$32,$D3,$F0,$38,$78,$1C,$57,$36,$CF,$C9,$CB,$C5,$18,$8F
	.dc.b	$09,$93,$AC,$F8,$DF,$EE,$19,$FE,$25,$B9,$FE,$05,$B9,$FF,$98,$E5
	.dc.b	$F6,$0F,$8E,$C3,$6B,$FA,$AC,$EA,$72,$48,$FA,$93,$BC,$7F,$28,$DC
	.dc.b	$7F,$29,$F6,$0F,$AD,$FD,$35,$F0,$C3,$F0,$57,$B8,$FC,$F3,$75,$5F
	.dc.b	$19,$DD,$47,$91,$F8,$88,$D3,$8C,$7C,$73,$B7,$E8,$CE,$9C,$CE,$F1
	.dc.b	$FD,$C6,$1E,$1A,$D4,$62,$F9,$36,$BC,$18,$FB,$30,$E6,$49,$D3,$E0
	.dc.b	$8D,$46,$09,$3F,$02,$BF,$29,$7E,$08,$C1,$E5,$2D,$F3,$7C,$66,$2B
	.dc.b	$9E,$3F,$23,$0E,$F3,$92,$1F,$C3,$B1,$F7,$46,$A3,$18,$F4,$69,$EE
	.dc.b	$3F,$7B,$CF,$FD,$87,$8F,$ED,$F3,$7F,$5D,$6F,$07,$F6,$36,$D7,$FA
	.dc.b	$9D,$7C,$1E,$3A,$7F,$A3,$BF,$EB,$56,$CF,$F3,$48,$C4,$F0,$8F,$EB
	.dc.b	$1D,$FD,$31,$FE,$E4,$DE,$7F,$F2,$04,$FF,$A2,$7E,$C1,$F5,$79,$A9
	.dc.b	$3F,$DA,$DF,$14,$EA,$30,$43,$FD,$41,$5D,$1F,$F4,$8C,$6D,$B5,$E9
	.dc.b	$8C,$47,$E7,$B2,$46,$23,$FC,$50,$D7,$FD,$79,$4F,$98,$F5,$12,$3F
	.dc.b	$D9,$7C,$3F,$E1,$8F,$33,$E5,$B6,$3F,$BE,$78,$29,$FF,$6F,$4E,$67
	.dc.b	$FC,$3B,$39,$7F,$BB,$7D,$7B,$EC,$3E,$93,$63,$FD,$60,$D7,$FD,$5E
	.dc.b	$FE,$B3,$8B,$FD,$54,$E5,$B3,$C8,$E7,$C9,$2B,$9B,$5F,$F0,$B5,$FF
	.dc.b	$A1,$6B,$F9,$EE,$1F,$E3,$49,$F4,$CF,$FB,$DD,$FF,$15,$18,$3F,$DF
	.dc.b	$5C,$1F,$E4,$3E,$6F,$C5,$7D,$83,$D5,$A7,$E1,$7C,$AF,$ED,$99,$7F
	.dc.b	$4A,$BE,$18,$7F,$B3,$6F,$83,$6F,$86,$1F,$8D,$F3,$1F,$61,$F9,$7F
	.dc.b	$21,$F2,$F4,$D4,$4C,$3F,$DB,$13,$A6,$CE,$5F,$F6,$CF,$CD,$BE,$1B
	.dc.b	$F4,$4D,$8F,$F7,$03,$5F,$F7,$B7,$A4,$FB,$07,$D7,$FE,$B2,$30,$FF
	.dc.b	$A2,$A7,$FD,$15,$3A,$7F,$D1,$B3,$A6,$C6,$5F,$F4,$73,$5F,$F4,$73
	.dc.b	$5F,$F4,$73,$5E,$B3,$FB,$7F,$5F,$FA,$67,$0F,$FB,$B7,$DE,$FD,$23
	.dc.b	$7B,$BF,$7D,$3A,$F0,$BF,$D1,$0D,$C1,$DF,$FE,$FE,$AF,$E0,$98,$72
	.dc.b	$BD,$F9,$1F,$BF,$FB,$0F,$F4,$7E,$1F,$E2,$FC,$BE,$E4,$62,$11,$8F
	.dc.b	$57,$87,$D5,$71,$64,$F1,$C3,$FD,$EF,$7C,$6F,$9D,$FA,$3D,$E9,$7B
	.dc.b	$3C,$FA,$FF,$BF,$5C,$1F,$EF,$4D,$BF,$8F,$B4,$FA,$9E,$F7,$C9,$70
	.dc.b	$7F,$B0,$1C,$12,$3F,$40,$D8,$FF,$47,$F9,$BF,$E7,$37,$C5,$D9,$FE
	.dc.b	$46,$F6,$3F,$91,$4F,$FA,$82,$72,$ED,$F8,$7A,$FF,$95,$6B,$FC,$9B
	.dc.b	$D1,$C2,$E5,$70,$8F,$E5,$9E,$EB,$E3,$EF,$1C,$AF,$0C,$E5,$B3,$97
	.dc.b	$B8,$C1,$2A,$EB,$5F,$7B,$E7,$3F,$9D,$7D,$87,$F1,$5C,$38,$E0,$8C
	.dc.b	$71,$7E,$36,$B8,$57,$D3,$E7,$E9,$18,$8F,$0C,$3E,$C3,$EE,$6D,$FE
	.dc.b	$D7,$CD,$7B,$C3,$A6,$1C,$EF,$6F,$A7,$D8,$7C,$17,$F4,$F6,$C7,$D2
	.dc.b	$6B,$E8,$5F,$01,$B9,$93,$B7,$94,$7F,$59,$FA,$FF,$F1,$74,$FC,$06
	.dc.b	$E1,$FC,$5D,$B9,$FD,$9F,$AF,$7D,$87,$1F,$5A,$61,$F8,$4B,$DB,$F4
	.dc.b	$6B,$6F,$C7,$F8,$7F,$F3,$5B,$F1,$9B,$8F,$F9,$BB,$81,$18,$CB,$FC
	.dc.b	$6F,$F6,$1F,$EB,$CB,$FC,$FB,$5F,$F1,$D9,$FE,$A1,$70,$E1,$AF,$2F
	.dc.b	$5F,$F9,$E4,$62,$38,$46,$08,$FF,$BB,$BD,$EF,$C4,$3F,$A0,$6B,$F8
	.dc.b	$66,$E3,$FC,$F5,$FC,$A4,$7F,$B8,$FE,$11,$E0,$7F,$64,$6C,$7F,$B8
	.dc.b	$DA,$FA,$BD,$1F,$F7,$94,$EF,$E2,$7F,$C1,$DF,$25,$FC,$9F,$1F,$E9
	.dc.b	$4D,$D3,$FD,$61,$8A,$BF,$C2,$9F,$60,$FB,$0A,$4C,$FF,$BD,$FC,$57
	.dc.b	$ED,$72,$FE,$C2,$FE,$DE,$3F,$CC,$BA,$23,$11,$FE,$BD,$6A,$8D,$78
	.dc.b	$1F,$E0,$09,$CC,$62,$7F,$E0,$04,$60,$E3,$DC,$F9,$67,$C3,$7F,$8A
	.dc.b	$DB,$5F,$C3,$9A,$FE,$F2,$DF,$18,$F8,$7F,$F0,$9B,$67,$F0,$67,$95
	.dc.b	$C3,$F3,$5A,$FB,$9A,$FF,$B6,$5A,$F1,$8F,$F0,$DC,$62,$7C,$31,$FE
	.dc.b	$02,$F8,$3F,$F2,$25,$7F,$08,$6E,$9F,$DF,$DB,$C3,$FE,$E5,$FA,$F7
	.dc.b	$D8,$7F,$8A,$9B,$7E,$8C,$51,$88,$FC,$23,$5F,$AA,$D7,$A4,$FF,$8C
	.dc.b	$23,$11,$F5,$5B,$3D,$27,$FC,$33,$F6,$1F,$C1,$AB,$F9,$DB,$5F,$E4
	.dc.b	$2D,$C7,$FC,$4B,$AF,$F8,$5A,$7E,$93,$5F,$C3,$B5,$1F,$85,$70,$DF
	.dc.b	$6F,$FC,$8B,$18,$C3,$F0,$31,$82,$4F,$DA,$DF,$F9,$D7,$B1,$FE,$6F
	.dc.b	$7F,$C3,$91,$FC,$D9,$B8,$7F,$8E,$BE,$BD,$F6,$1F,$E5,$66,$D7,$CD
	.dc.b	$6B,$FA,$F6,$BF,$A0,$6B,$EA,$C5,$18,$8F,$D0,$36,$33,$3F,$E6,$08
	.dc.b	$D5,$BD,$5F,$E4,$6D,$F0,$FF,$D1,$11,$FC,$9A,$DF,$25,$F3,$F8,$7C
	.dc.b	$0F,$D3,$CB,$F6,$6D,$73,$E3,$7F,$9B,$A3,$11,$FB,$36,$D7,$49,$D1
	.dc.b	$27,$E6,$7D,$2B,$EC,$3F,$D3,$4D,$8F,$EF,$AD,$7F,$61,$F4,$FE,$73
	.dc.b	$FA,$3B,$FC,$19,$AF,$C7,$6B,$FC,$15,$BA,$7F,$07,$F4,$3F,$D5,$18
	.dc.b	$FF,$26,$E0,$FF,$32,$37,$4F,$C4,$F8,$FF,$D9,$BC,$AD,$DC,$CB,$FC
	.dc.b	$23,$7B,$7F,$C7,$3E,$77,$F4,$E7,$FB,$CF,$D5,$BE,$C3,$FD,$6C,$D8
	.dc.b	$FF,$0C,$B5,$FE,$19,$F6,$1F,$C3,$1C,$CF,$02,$3F,$EA,$28,$D3,$FD
	.dc.b	$FB,$D9,$FF,$54,$3F,$E0,$36,$B9,$F8,$CF,$F6,$8F,$33,$EC,$3F,$55
	.dc.b	$C1,$FD,$7E,$F6,$DA,$FF,$2B,$9C,$27,$7C,$1F,$AC,$BE,$5D,$7F,$E6
	.dc.b	$F6,$E3,$FC,$81,$AF,$F6,$46,$BF,$EC,$3C,$7F,$D9,$58,$F8,$7E,$43
	.dc.b	$FC,$CA,$79,$BE,$0F,$FD,$59,$8F,$F9,$9A,$33,$7E,$D7,$5F,$19,$FE
	.dc.b	$4E,$FA,$F7,$D8,$7F,$98,$2B,$F9,$8C,$63,$1F,$F4,$23,$63,$33,$B7
	.dc.b	$FB,$CC,$3E,$B1,$C5,$C1,$B3,$D1,$AF,$06,$BF,$D0,$3E,$57,$FD,$70
	.dc.b	$9A,$47,$E2,$6A,$F8,$36,$BF,$E6,$DD,$D3,$98,$FF,$1F,$7E,$0A,$FB
	.dc.b	$1D,$E1,$FE,$4E,$BD,$E1,$8D,$B3,$18,$C3,$A6,$3C,$EF,$FC,$8B,$8F
	.dc.b	$39,$CE,$1C,$5F,$D0,$C3,$E0,$BF,$C4,$F5,$86,$BF,$E9,$88,$D4,$FF
	.dc.b	$60,$EC,$46,$27,$F0,$5E,$37,$F9,$42,$BD,$CD,$8F,$1D,$F8,$56,$5F
	.dc.b	$EC,$EE,$AC,$36,$F8,$C6,$F1,$EF,$7A,$C3,$A6,$9F,$EE,$CE,$97,$D4
	.dc.b	$F5,$EC,$46,$32,$FF,$45,$D7,$F9,$FD,$FD,$75,$6E,$75,$A7,$ED,$A3
	.dc.b	$CD,$DF,$FE,$AE,$EE,$D4,$78,$FE,$C3,$E6,$57,$F9,$C9,$B8,$FF,$BF
	.dc.b	$BF,$08,$FF,$6C,$E5,$FE,$79,$9C,$BF,$C4,$1B,$5F,$EC,$46,$E6,$AE
	.dc.b	$33,$D7,$93,$E4,$73,$BF,$3A,$7D,$07,$D6,$8C,$36,$33,$97,$F8,$5A
	.dc.b	$72,$FF,$80,$7C,$47,$81,$F1,$DB,$1F,$EC,$19,$DE,$3F,$CD,$2B,$C3
	.dc.b	$7F,$02,$3F,$B8,$9F,$AA,$70,$FF,$74,$9F,$36,$3E,$32,$5F,$8E,$B7
	.dc.b	$5C,$67,$2F,$F9,$57,$EC,$BF,$6D,$5F,$ED,$3D,$11,$89,$FF,$58,$E1
	.dc.b	$DA,$B5,$39,$DB,$F5,$3C,$3F,$EF,$0E,$4F,$CD,$B6,$BF,$C0,$1C,$BF
	.dc.b	$99,$C7,$D7,$39,$7F,$CF,$ED,$73,$1B,$7F,$D9,$CD,$7F,$D8,$4D,$C9
	.dc.b	$F6,$5E,$3B,$FE,$FC,$DA,$DE,$1F,$97,$8C,$60,$FA,$47,$DB,$FA,$1B
	.dc.b	$87,$F8,$5D,$B8,$79,$9A,$FA,$9A,$FF,$B8,$DA,$FF,$B5,$DB,$5D,$EB
	.dc.b	$F7,$F8,$BE,$CB,$FC,$5B,$87,$D1,$6E,$1F,$ED,$8C,$3C,$3A,$7F,$16
	.dc.b	$BE,$13,$FA,$8F,$C0,$FF,$89,$3F,$31,$AF,$FB,$D5,$AF,$FB,$A5,$B9
	.dc.b	$1A,$4E,$EF,$B2,$F0,$A7,$D9,$7F,$9F,$33,$FC,$33,$70,$FF,$76,$D7
	.dc.b	$F2,$3C,$7C,$D3,$97,$F0,$BF,$48,$F9,$FF,$C2,$B7,$0F,$F2,$23,$70
	.dc.b	$FC,$1B,$5F,$C3,$B5,$FF,$29,$E6,$8C,$42,$31,$5D,$8F,$9F,$ED,$FF
	.dc.b	$10,$C3,$E5,$37,$0E,$78,$78,$67,$3D,$9F,$C7,$7F,$0A,$7D,$9F,$D8
	.dc.b	$6D,$FF,$9F,$5B,$9F,$FA,$FD,$AF,$FA,$C1,$AF,$F9,$45,$AF,$FA,$BD
	.dc.b	$BB,$1B,$F6,$6F,$8F,$FE,$5C,$DB,$FD,$B4,$F8,$32,$FC,$24,$E7,$0F
	.dc.b	$F5,$67,$D9,$BE,$B3,$F9,$B3,$70,$FD,$F7,$E1,$6F,$C1,$73,$5F,$CC
	.dc.b	$BE,$2D,$F4,$1F,$BA,$6D,$FF,$9E,$5B,$8F,$EE,$1A,$FE,$F3,$ED,$5F
	.dc.b	$6B,$FC,$F1,$B5,$FC,$37,$35,$E6,$BF,$AB,$5B,$E6,$9F,$4B,$FD,$92
	.dc.b	$30,$D8,$CE,$1F,$EC,$49,$CB,$63,$31,$FE,$C8,$6B,$FC,$A8,$D9,$99
	.dc.b	$ED,$F6,$9C,$38,$3F,$C0,$72,$8C,$5F,$F3,$2B,$B3,$FE,$A4,$9B,$D2
	.dc.b	$FB,$4F,$F4,$35,$FF,$AA,$30,$FE,$4B,$18,$B4,$63,$3F,$EC,$33,$98
	.dc.b	$DF,$C1,$E0,$CC,$DB,$E8,$7E,$8B,$EC,$DF,$69,$FD,$A9,$B1,$AC,$7F
	.dc.b	$49,$0F,$95,$D5,$4A,$D6,$6A,$E3,$7E,$69,$ED,$7B,$8C,$0F,$1D,$F3
	.dc.b	$DF,$8D,$8C,$C6,$39,$7F,$C2,$5F,$4E,$FA,$EF,$F3,$D6,$3E,$5A,$D3
	.dc.b	$70,$E5,$BF,$F8,$D2,$73,$8F,$83,$0E,$5F,$09,$FE,$17,$BD,$E1,$FE
	.dc.b	$8E,$F9,$CF,$F5,$8B,$D5,$B7,$E4,$9D,$BE,$98,$CE,$3F,$D6,$BC,$7F
	.dc.b	$F7,$DD,$FC,$46,$D7,$C0,$DF,$FC,$9D,$F0,$3F,$E1,$89,$CD,$F1,$F8
	.dc.b	$FC,$19,$9B,$87,$CE,$FA,$37,$C5,$FF,$88,$1B,$19,$F6,$B1,$32,$38
	.dc.b	$3B,$8C,$67,$FE,$50,$FA,$7F,$F2,$4F,$0F,$F9,$19,$BF,$00,$69,$30
	.dc.b	$FF,$26,$46,$5B,$19,$AF,$D4,$7C,$B7,$ED,$3B,$4F,$F9,$16,$F8,$4F
	.dc.b	$E0,$8F,$DC,$47,$F8,$EB,$B9,$F3,$B4,$C7,$C9,$5C,$BF,$1F,$3A,$6C
	.dc.b	$E9,$A7,$FA,$C7,$A3,$FC,$A9,$F0,$9F,$89,$8F,$C3,$7C,$E6,$26,$46
	.dc.b	$CC,$97,$2F,$C9,$F8,$EF,$B2,$F8,$ED,$9F,$F3,$73,$67,$F8,$37,$07
	.dc.b	$F2,$5A,$D9,$2B,$6D,$9F,$E1,$9D,$9F,$73,$6F,$CF,$F1,$FC,$8B,$FC
	.dc.b	$F6,$3D,$79,$FF,$D1,$F7,$B9,$FD,$8F,$D5,$FF,$53,$C7,$FA,$D3,$7A
	.dc.b	$8B,$E4,$3F,$CC,$EE,$49,$E8,$E2,$32,$FF,$9B,$3E,$B5,$F5,$54,$C7
	.dc.b	$6D,$CB,$F8,$0C,$E9,$B3,$A6,$9F,$C1,$BE,$C3,$FC,$12,$4E,$FF,$F1
	.dc.b	$FB,$FE,$07,$FA,$F7,$C8,$FF,$A1,$5B,$5C,$6B,$13,$FE,$80,$AF,$D9
	.dc.b	$EA,$7F,$6B,$EA,$B0,$E3,$BF,$EC,$67,$64,$F1,$B9,$E5,$E5,$7A,$79
	.dc.b	$5E,$42,$47,$F4,$1A,$FF,$03,$57,$F5,$28,$C6,$1D,$FE,$7F,$D0,$72
	.dc.b	$8F,$F3,$26,$28,$F4,$79,$9F,$00,$93,$D7,$3F,$54,$7E,$D3,$B9,$F6
	.dc.b	$9F,$EA,$D6,$D7,$CB,$9C,$4C,$8F,$E4,$93,$AC,$FE,$05,$65,$FE,$5B
	.dc.b	$19,$E7,$FE,$4C,$DA,$FF,$12,$E9,$FD,$2D,$57,$F3,$79,$D3,$F0,$8F
	.dc.b	$B2,$48,$F5,$4F,$F9,$4A,$3F,$D2,$3D,$9F,$BC,$F5,$FD,$47,$D9,$8B
	.dc.b	$5F,$D3,$D7,$D5,$FE,$BF,$2F,$0F,$C8,$FF,$7C,$9F,$1C,$E5,$FF,$7E
	.dc.b	$78,$7E,$6E,$17,$D2,$BE,$87,$FD,$EB,$87,$1B,$EB,$E1,$32,$EB,$8F
	.dc.b	$FA,$9B,$2E,$97,$B6,$C7,$5C,$7A,$67,$C3,$1F,$1E,$3C,$DF,$19,$8C
	.dc.b	$74,$F8,$1E,$19,$F9,$B3,$F2,$E3,$E1,$BE,$59,$F8,$5B,$19,$BE,$39
	.dc.b	$78,$B1,$EE,$75,$2E,$7E,$87,$C4,$79,$32,$F0,$67,$E4,$9D,$4F,$1C
	.dc.b	$BC,$18,$76,$CF,$CB,$3C,$A7,$2F,$6C,$FC,$AB,$1E,$AE,$6F,$B4,$EA
	.dc.b	$3B,$67,$F2,$F0,$EB,$18,$3B,$47,$BD,$F4,$67,$EF,$C7,$C2,$49,$D3
	.dc.b	$F8,$5C,$7C,$2E,$6B,$99,$23,$33,$C0,$91,$9A,$ED,$18,$8F,$98,$7F
	.dc.b	$90,$72,$FC,$83,$71,$FC,$03,$72,$FF,$26,$4B,$EC,$DF,$29,$86,$D7
	.dc.b	$F0,$B9,$D4,$E4,$91,$F6,$A7,$79,$FE,$4D,$B8,$FE,$7F,$DF,$FF,$52
	.dc.b	$57,$DB,$9C,$BF,$A6,$6B,$FA,$C7,$3A,$7F,$A7,$A3,$86,$11,$F0,$3F
	.dc.b	$96,$AE,$B1,$82,$61,$E3,$9D,$7C,$0F,$F0,$69,$CE,$3F,$AB,$C7,$C3
	.dc.b	$97,$BA,$7F,$25,$8F,$4C,$79,$47,$03,$D1,$8F,$47,$C9,$1A,$26,$1F
	.dc.b	$07,$1F,$04,$62,$78,$9C,$F2,$F7,$3D,$4F,$85,$7D,$2F,$94,$53,$8D
	.dc.b	$73,$AE,$2F,$58,$E2,$75,$C3,$D5,$5E,$FB,$E5,$3A,$AF,$CA,$56,$A3
	.dc.b	$F9,$57,$D1,$7C,$7E,$5F,$E6,$2D,$E1,$FE,$5F,$C7,$FE,$22,$8C,$57
	.dc.b	$FB,$57,$BF,$FD,$5A,$FF,$70,$F7,$7F,$D3,$5D,$29,$CC,$FF,$A5,$E3
	.dc.b	$13,$A9,$CC,$FF,$34,$FB,$3F,$F1,$F3,$FE,$6B,$6B,$FE,$72,$6E,$4C
	.dc.b	$BF,$DD,$34,$FA,$EF,$2E,$6B,$DD,$7A,$9B,$87,$F9,$FD,$B9,$79,$EF
	.dc.b	$8C,$6A,$F9,$7A,$38,$3E,$0D,$6E,$FF,$C1,$F7,$EB,$23,$FD,$A7,$1F
	.dc.b	$23,$FD,$D3,$BB,$FD,$10,$FF,$72,$F3,$3E,$5F,$1F,$F8,$15,$EB,$5F
	.dc.b	$9E,$D7,$DD,$5F,$3F,$93,$E4,$11,$EC,$FF,$B2,$74,$7D,$9F,$1D,$FE
	.dc.b	$CB,$6B,$FD,$3C,$DC,$BF,$C2,$91,$F2,$23,$55,$CA,$F0,$79,$B3,$FF
	.dc.b	$68,$63,$E4,$C3,$8E,$1F,$E0,$C3,$47,$3F,$91,$8F,$F8,$5F,$0F,$2C
	.dc.b	$79,$F1,$34,$FC,$27,$37,$C2,$25,$7F,$8F,$3E,$F2,$BF,$E9,$9D,$7F
	.dc.b	$B5,$65,$F1,$B8,$3F,$21,$7F,$2A,$FE,$BF,$96,$FB,$3A,$4C,$7F,$CA
	.dc.b	$FE,$1D,$7F,$D4,$93,$B7,$FC,$B7,$3C,$A7,$F4,$3E,$6B,$EC,$FE,$ED
	.dc.b	$7F,$AE,$DB,$9F,$F9,$91,$BC,$7F,$AA,$FC,$07,$F4,$73,$FE,$C3,$7F
	.dc.b	$43,$59,$E9,$FB,$51,$88,$FF,$6A,$37,$88,$D2,$4F,$03,$F0,$5E,$A7
	.dc.b	$F4,$2B,$FF,$68,$37,$0F,$DD,$B7,$0F,$F7,$07,$D9,$3E,$13,$F5,$33
	.dc.b	$9C,$8D,$31,$3B,$AF,$F7,$7F,$DA,$7F,$BC,$DB,$F0,$3F,$A6,$8C,$47
	.dc.b	$FB,$3D,$FF,$7E,$BA,$21,$98,$E1,$F0,$7F,$C3,$A1,$EF,$7C,$A7,$FC
	.dc.b	$82,$D7,$3D,$DF,$C5,$3C,$77,$E1,$0F,$A2,$C3,$6B,$FC,$D1,$3A,$9F
	.dc.b	$F3,$54,$7F,$8C,$67,$4F,$F4,$93,$11,$A9,$E1,$1C,$A7,$6F,$F4,$F8
	.dc.b	$CB,$C5,$AF,$F5,$3E,$AF,$F2,$5F,$A1,$FC,$2B,$C9,$FF,$39,$4F,$F9
	.dc.b	$73,$8F,$F3,$51,$88,$FF,$7B,$2F,$98,$E1,$52,$78,$72,$7F,$99,$30
	.dc.b	$EB,$18,$0F,$F9,$6A,$BF,$8C,$FC,$03,$E0,$69,$2B,$85,$25,$F6,$99
	.dc.b	$96,$7F,$AF,$9E,$51,$C1,$AE,$B1,$F2,$47,$28,$F2,$73,$3E,$15,$F4
	.dc.b	$3F,$CE,$1B,$3A,$6B,$CC,$FC,$BF,$B7,$FC,$FB,$A3,$FD,$87,$1D,$1A
	.dc.b	$EE,$D1,$8C,$BF,$DA,$FF,$8B,$BE,$0B,$96,$13,$0E,$4F,$EF,$A7,$51
	.dc.b	$F1,$3D,$77,$C1,$7F,$42,$6C,$66,$31,$F8,$2E,$59,$19,$39,$47,$F4
	.dc.b	$AF,$B5,$FF,$71,$47,$D0,$6B,$EF,$E2,$46,$33,$FE,$3C,$BE,$0B,$97
	.dc.b	$C1,$CB,$E4,$F5,$8F,$D5,$7D,$AB,$DF,$A4,$AF,$F4,$F7,$84,$FA,$AF
	.dc.b	$D8,$36,$3F,$4A,$D7,$FD,$A0,$D7,$F4,$CD,$7F,$C6,$0D,$7F,$D6,$ED
	.dc.b	$4D,$35,$9F,$FA,$9B,$A7,$FB,$6F,$BD,$F8,$AD,$7F,$9E,$37,$A7,$FD
	.dc.b	$59,$F0,$FF,$C3,$27,$FA,$07,$C1,$7F,$32,$CB,$F5,$CD,$7F,$48,$D7
	.dc.b	$F4,$6D,$F5,$54,$FB,$5F,$EE,$6D,$AF,$D5,$4E,$A7,$FB,$BF,$C9,$BE
	.dc.b	$1F,$FB,$CF,$57,$ED,$F2,$FE,$C8,$DC,$FF,$D8,$4D,$C7,$FD,$29,$F2
	.dc.b	$BC,$5E,$AD,$7F,$88,$64,$8C,$47,$F9,$6E,$31,$87,$F1,$86,$C7,$FC
	.dc.b	$00,$9D,$BF,$ED,$16,$BE,$09,$E1,$18,$97,$CB,$FF,$B5,$6B,$DE,$D7
	.dc.b	$F6,$11,$AF,$86,$FF,$10,$FC,$17,$C5,$9F,$CB,$35,$FE,$3C,$D7,$F8
	.dc.b	$A3,$5F,$E2,$6D,$E2,$34,$9B,$7F,$92,$9E,$16,$FB,$27,$C5,$FF,$8A
	.dc.b	$1B,$87,$D6,$6B,$FA,$9F,$BB,$BC,$1F,$E0,$57,$31,$FC,$CD,$AF,$F2
	.dc.b	$96,$BF,$CA,$1B,$91,$A4,$BF,$F2,$FB,$C3,$07,$C7,$7F,$87,$2F,$F1
	.dc.b	$2D,$7E,$5B,$70,$FF,$0C,$7C,$E3,$E8,$3F,$C9,$AD,$AF,$EA,$47,$F3
	.dc.b	$A7,$0F,$F9,$AD,$AF,$F4,$D6,$BF,$D0,$DA,$F1,$8F,$F3,$AC,$63,$23
	.dc.b	$AC,$1F,$62,$FB,$2F,$F2,$AB,$70,$FD,$1B,$5F,$E2,$DE,$47,$C0,$7F
	.dc.b	$A4,$B5,$FF,$40,$35,$FE,$C4,$D7,$FA,$C3,$5F,$EA,$ED,$C0,$D2,$67
	.dc.b	$FA,$1F,$B2,$FE,$63,$5F,$96,$6E,$9F,$E5,$56,$F3,$A5,$F4,$AF,$CD
	.dc.b	$9F,$29,$FE,$9C,$8C,$3F,$E0,$46,$CF,$D2,$6B,$FE,$05,$6B,$FE,$CF
	.dc.b	$6B,$DD,$AF,$82,$7B,$BC,$DA,$FF,$00,$FA,$8F,$F7,$9B,$FB,$06,$E3
	.dc.b	$FE,$6F,$6E,$1D,$69,$F7,$7E,$96,$DF,$58,$F5,$46,$23,$D0,$D7,$FD
	.dc.b	$CD,$A7,$F7,$2F,$1F,$FD,$53,$E8,$7F,$4F,$C3,$FD,$82,$DA,$F9,$5E
	.dc.b	$C3,$E0,$F7,$72,$7F,$BE,$D6,$DF,$F1,$94,$FF,$A7,$BE,$B7,$F8,$7C
	.dc.b	$7F,$2C,$6B,$FC,$39,$BA,$FF,$9E,$F6,$FC,$FB,$FE,$8F,$8C,$E5,$F6
	.dc.b	$F9,$51,$8B,$46,$3B,$BF,$D6,$F8,$72,$BF,$C1,$7C,$9B,$EC,$BF,$65
	.dc.b	$18,$6E,$09,$D7,$C6,$7F,$B3,$3E,$6B,$F3,$9A,$7F,$AD,$27,$35,$FE
	.dc.b	$2F,$6B,$FE,$17,$6C,$7F,$AE,$B8,$55,$CB,$1F,$F7,$6B,$FA,$18,$CF
	.dc.b	$3F,$FA,$EE,$30,$7E,$96,$DF,$64,$F9,$CF,$F7,$9C,$61,$B1,$9C,$BF
	.dc.b	$C5,$F4,$FA,$DC,$D4,$9F,$45,$6D,$FF,$14,$56,$DF,$F2,$50,$FB,$2F
	.dc.b	$F3,$66,$9F,$E6,$86,$C7,$F1,$66,$BF,$E6,$96,$BF,$B1,$39,$7D,$97
	.dc.b	$EC,$9E,$0D,$78,$35,$E1,$08,$C4,$7F,$9D,$7E,$6B,$FC,$F2,$FE,$F5
	.dc.b	$A9,$A7,$FB,$46,$FF,$CD,$5B,$5F,$80,$FF,$76,$D7,$E8,$A7,$2F,$FA
	.dc.b	$07,$7F,$AF,$3A,$8F,$AE,$D7,$F8,$AC,$F0,$79,$46,$67,$8D,$7F,$3A
	.dc.b	$D9,$F9,$57,$E3,$63,$0D,$BF,$F4,$84,$E8,$EC,$59,$F5,$78,$9D,$71
	.dc.b	$FE,$91,$18,$25,$7A,$63,$2D,$9E,$04,$9F,$E9,$BE,$0F,$F9,$F3,$D2
	.dc.b	$F6,$36,$33,$BF,$FA,$1D,$B3,$A8,$C5,$FF,$A9,$EB,$83,$63,$2F,$FA
	.dc.b	$85,$AF,$EE,$7E,$C5,$F0,$FF,$2F,$A9,$18,$C7,$FB,$0C,$62,$75,$39
	.dc.b	$AF,$F1,$3B,$5F,$F1,$33,$5E,$D1,$C0,$FB,$0B,$EE,$FF,$CD,$57,$FE
	.dc.b	$C2,$6C,$7F,$70,$6B,$FE,$C3,$6B,$FC,$F7,$EC,$BF,$59,$3F,$D1,$5B
	.dc.b	$1F,$D1,$5A,$FF,$45,$6B,$FA,$39,$78,$1E,$46,$D7,$8E,$30,$47,$04
	.dc.b	$9F,$34,$7E,$4F,$B1,$7F,$93,$9F,$37,$1E,$63,$11,$FD,$89,$B9,$BE
	.dc.b	$C7,$FC,$C5,$8F,$D0,$6D,$7F,$95,$1A,$FF,$94,$DA,$FF,$33,$6B,$FC
	.dc.b	$C9,$B3,$D7,$CF,$FB,$D1,$AD,$FF,$BE,$F0,$FE,$43,$97,$F2,$F8,$7F
	.dc.b	$81,$A3,$07,$E1,$7F,$05,$F3,$93,$4C,$13,$3F,$C3,$C6,$5B,$1F,$9C
	.dc.b	$F3,$1F,$65,$49,$3F,$A2,$8C,$BF,$90,$FB,$0F,$AF,$8F,$F4,$C6,$EC
	.dc.b	$69,$3B,$FF,$34,$DA,$E5,$1F,$AC,$FB,$1F,$F4,$44,$7E,$C1,$B7,$FA
	.dc.b	$96,$BF,$C9,$F6,$46,$23,$FB,$07,$E8,$1F,$C1,$DF,$F4,$23,$5C,$F5
	.dc.b	$FE,$AE,$31,$1F,$D9,$1B,$91,$A4,$9F,$E0,$3E,$A3,$E8,$29,$32,$FE
	.dc.b	$0F,$19,$7F,$6B,$F6,$2F,$A6,$FD,$D3,$63,$31,$82,$38,$26,$5F,$EA
	.dc.b	$09,$E0,$FF,$A8,$BD,$97,$D9,$7F,$05,$6D,$7F,$A7,$E3,$07,$F0,$9F
	.dc.b	$75,$F5,$74,$93,$FC,$2A,$32,$FF,$0C,$FB,$17,$D6,$7F,$11,$6C,$66
	.dc.b	$30,$67,$E0,$A4,$71,$6E,$5F,$E7,$6F,$91,$7D,$17,$F0,$E6,$D7,$F9
	.dc.b	$EA,$51,$87,$84,$EC,$93,$B9,$D4,$FF,$69,$38,$CF,$5F,$A3,$95,$B6
	.dc.b	$CF,$F3,$3F,$B0,$61,$FC,$F1,$B5,$FD,$01,$B3,$B9,$D1,$27,$4D,$8F
	.dc.b	$E7,$DF,$60,$F9,$DF,$E8,$ED,$AF,$E9,$6D,$9D,$CE,$89,$3A,$6C,$7B
	.dc.b	$7E,$8D,$F5,$2F,$B0,$FE,$DD,$18,$6C,$66,$FF,$D4,$B3,$96,$C6,$6B
	.dc.b	$A3,$5F,$F0,$94,$ED,$CD,$7F,$77,$F9,$9F,$C0,$C7,$F8,$05,$AF,$F9
	.dc.b	$99,$AF,$FA,$2F,$C5,$46,$21,$DC,$F8,$AF,$F1,$B6,$5F,$DB,$9B,$3F
	.dc.b	$D9,$DA,$FE,$FD,$AF,$FA,$5E,$79,$7A,$FF,$DD,$27,$33,$D9,$B3,$AA
	.dc.b	$FE,$F7,$F3,$FF,$DB,$AB,$C6,$D9,$34,$9E,$97,$F8,$01,$B5,$CA,$3F
	.dc.b	$35,$F7,$F2,$3F,$28,$DB,$FD,$03,$5F,$CB,$72,$A3,$11,$FE,$5E,$FA
	.dc.b	$CF,$F0,$C1,$2B,$FC,$68,$FF,$84,$39,$7F,$DD,$31,$FE,$14,$6C,$FF
	.dc.b	$83,$5A,$FF,$25,$6B,$F1,$DA,$FD,$26,$E4,$69,$AD,$BE,$CF,$C7,$FE
	.dc.b	$04,$97,$FE,$F0,$8C,$B7,$74,$EB,$0F,$C3,$35,$F8,$3C,$7F,$C0,$49
	.dc.b	$5F,$BB,$6C,$7F,$07,$6B,$C2,$3C,$1D,$6F,$AF,$FF,$25,$36,$33,$18
	.dc.b	$3F,$C6,$FF,$55,$FE,$08,$AF,$F1,$93,$5F,$E9,$4D,$7F,$18,$D7,$F2
	.dc.b	$ED,$D0,$D2,$7A,$7F,$E4,$4F,$AB,$7C,$7D,$26,$3F,$E4,$A8,$CB,$63
	.dc.b	$38,$7F,$91,$67,$2D,$8F,$F2,$07,$D6,$FD,$8F,$28,$D2,$7C,$0F,$F9
	.dc.b	$53,$EC,$3F,$B1,$5F,$EB,$9B,$A1,$A6,$AF,$9F,$C6,$F6,$26,$7F,$E4
	.dc.b	$CC,$7F,$E0,$0B,$63,$38,$7F,$2F,$AE,$33,$93,$87,$87,$E8,$23,$83
	.dc.b	$FC,$D7,$87,$F2,$CA,$FE,$F4,$D7,$F7,$8D,$7D,$5E,$0F,$FA,$1F,$EB
	.dc.b	$DF,$4D,$49,$7F,$93,$8C,$BD,$A3,$E2,$11,$C1,$2B,$FC,$E9,$7C,$1B
	.dc.b	$3B,$6C,$ED,$F3,$C6,$5F,$F5,$C7,$3B,$F1,$4F,$F0,$04,$7F,$A9,$5A
	.dc.b	$E9,$AF,$F2,$56,$BF,$5C,$FB,$44,$AF,$8E,$FF,$9D,$3E,$9F,$F4,$17
	.dc.b	$FE,$BB,$EA,$FF,$64,$46,$27,$FD,$79,$1F,$F0,$07,$E6,$5F,$30,$C3
	.dc.b	$85,$05,$F1,$F8,$6F,$E0,$51,$98,$C6,$1D,$BA,$7E,$8C,$F6,$1C,$2C
	.dc.b	$BB,$E3,$DA,$7A,$D7,$0C,$7A,$C6,$33,$D6,$5D,$A3,$9E,$1C,$9E,$C6
	.dc.b	$CB,$1B,$71,$5E,$8C,$7C,$59,$FB,$E7,$78,$F8,$73,$F5,$65,$E0,$8C
	.dc.b	$4F,$82,$B8,$61,$E6,$C7,$B1,$2B,$78,$F6,$CF,$96,$33,$1E,$EF,$CE
	.dc.b	$9D,$46,$09,$87,$5B,$E7,$9F,$BE,$31,$7C,$DB,$5F,$3E,$F9,$C6,$09
	.dc.b	$9F,$B2,$75,$7C,$E3,$13,$A9,$FB,$98,$78,$33,$F7,$D6,$89,$3A,$C7
	.dc.b	$D5,$87,$38,$C1,$EB,$3F,$C6,$19,$FE,$45,$B8,$7E,$19,$B5,$FE,$3D
	.dc.b	$9F,$F3,$17,$A7,$7D,$13,$49,$5F,$E3,$B3,$F2,$B1,$A7,$F0,$2D,$CF
	.dc.b	$EB,$37,$2F,$CE,$68,$F4,$BF,$CE,$D7,$F8,$EA,$D3,$F9,$D6,$BD,$B0
	.dc.b	$FB,$13,$C2,$3B,$C7,$12,$3F,$4E,$35,$87,$8A,$72,$D9,$D7,$A7,$FA
	.dc.b	$F7,$99,$23,$47,$E6,$6B,$9E,$3F,$BC,$C3,$CB,$97,$C8,$BE,$13,$EE
	.dc.b	$AD,$C7,$AF,$1F,$0C,$F1,$7D,$78,$F8,$71,$EF,$39,$AE,$67,$EE,$AF
	.dc.b	$9B,$F0,$CC,$96,$7C,$D5,$CE,$B9,$47,$49,$F1,$1F,$00,$F1,$98,$75
	.dc.b	$97,$D5,$AE,$53,$9A,$E6,$47,$C9,$1A,$E8,$75,$BE,$B3,$FD,$11,$18
	.dc.b	$6C,$66,$BF,$B8,$47,$FA,$37,$93,$F9,$B7,$65,$E5,$FE,$EF,$5F,$E1
	.dc.b	$B6,$BF,$CE,$63,$9F,$C1,$C8,$FE,$F4,$D7,$FB,$C3,$5F,$F6,$A6,$08
	.dc.b	$C4,$7F,$95,$E3,$0F,$22,$1C,$8F,$17,$C0,$79,$48,$E0,$FF,$05,$67
	.dc.b	$EE,$23,$83,$FB,$C7,$57,$FA,$AF,$6F,$F0,$3B,$7B,$7F,$D3,$F3,$A3
	.dc.b	$FB,$F1,$92,$18,$21,$FD,$FD,$FF,$5B,$3F,$D0,$1B,$3F,$E1,$48,$F9
	.dc.b	$DF,$60,$F8,$CA,$48,$FF,$24,$46,$5F,$DF,$4F,$F9,$8F,$C7,$F4,$79
	.dc.b	$06,$92,$7F,$C5,$5E,$9F,$DE,$23,$82,$1F,$E2,$B7,$2D,$3F,$86,$71
	.dc.b	$FF,$8E,$BE,$43,$F1,$3E,$31,$A4,$9D,$C6,$5B,$19,$7F,$C9,$6F,$E3
	.dc.b	$7E,$BB,$F2,$11,$FE,$D3,$E5,$46,$23,$FC,$0E,$D7,$FD,$4C,$D7,$FB
	.dc.b	$A7,$9D,$F1,$5F,$E4,$4D,$7F,$DB,$CD,$7F,$86,$35,$E3,$B7,$D6,$8C
	.dc.b	$47,$FA,$72,$30,$7E,$8F,$EA,$1F,$5D,$49,$3F,$E7,$D8,$CB,$63,$2F
	.dc.b	$B6,$32,$D8,$FD,$9F,$CF,$BE,$B6,$92,$7E,$8C,$65,$B1,$97,$FD,$F5
	.dc.b	$AA,$75,$18,$6B,$FC,$AF,$C2,$7D,$6B,$E6,$3F,$A4,$4F,$F9,$1A,$30
	.dc.b	$FF,$4B,$9D,$CE,$9B,$1C,$2B,$FD,$A8,$DF,$57,$F3,$F1,$88,$FE,$A8
	.dc.b	$D7,$F9,$E1,$E7,$8D,$91,$E4,$7E,$43,$C4,$7C,$3F,$F1,$86,$C6,$63
	.dc.b	$04,$70,$43,$E2,$C7,$07,$E3,$FA,$AF,$AC,$FE,$46,$D8,$FC,$FB,$5D
	.dc.b	$C6,$88,$E9,$AF,$DD,$FA,$CF,$C3,$11,$FC,$23,$F3,$BD,$B7,$D6,$3E
	.dc.b	$93,$0D,$AF,$EF,$53,$A9,$C9,$23,$33,$B7,$E5,$CE,$DC,$CF,$F3,$EF
	.dc.b	$AC,$7D,$5D,$24,$FE,$F2,$32,$D8,$CB,$F8,$C8,$CB,$63,$FA,$4F,$D5
	.dc.b	$BE,$AE,$92,$7F,$83,$C6,$5B,$19,$7F,$2F,$19,$6C,$7E,$23,$EA,$64
	.dc.b	$7E,$35,$AF,$E2,$BD,$9F,$C6,$91,$FC,$93,$5D,$18,$21,$FE,$B1,$F7
	.dc.b	$9E,$F7,$F3,$F6,$C7,$F2,$C6,$BB,$8D,$79,$52,$34,$D7,$FD,$79,$F0
	.dc.b	$FF,$9D,$7F,$C5,$6D,$7F,$CF,$ED,$F1,$CD,$F8,$99,$CF,$FD,$45,$B5
	.dc.b	$FC,$FB,$81,$F0,$BF,$D5,$5B,$19,$8C,$11,$C1,$0F,$F6,$84,$70,$7F
	.dc.b	$83,$FA,$AC,$BF,$C0,$6D,$8F,$E7,$ED,$77,$1F,$D9,$7E,$7D,$EC,$D2
	.dc.b	$4F,$F5,$48,$CB,$63,$2F,$FB,$86,$32,$D8,$FF,$09,$FD,$47,$F8,$2A
	.dc.b	$3F,$C8,$1E,$0A,$31,$1F,$EB,$18,$C7,$A9,$C0,$FE,$F1,$F5,$3F,$62
	.dc.b	$7F,$C9,$1C,$08,$C4,$7F,$97,$9A,$FF,$AC,$9B,$1B,$E1,$E6,$49,$FE
	.dc.b	$DF,$3C,$2D,$F2,$4F,$95,$FF,$15,$B6,$BF,$B3,$B5,$FE,$6A,$D7,$CB
	.dc.b	$FA,$0F,$F5,$1C,$4D,$27,$95,$FE,$13,$6C,$FF,$80,$1A,$FF,$40,$F3
	.dc.b	$5F,$0B,$FE,$3C,$6C,$7F,$78,$6B,$B8,$FF,$0F,$7D,$4B,$EB,$69,$27
	.dc.b	$FC,$1B,$19,$6C,$65,$FE,$73,$8A,$75,$18,$6B,$FE,$79,$F9,$B7,$D3
	.dc.b	$BB,$7F,$A7,$C6,$1B,$19,$9F,$F4,$D4,$E5,$FE,$9D,$F5,$1F,$D3,$08
	.dc.b	$E0,$87,$F4,$E7,$2D,$39,$CA,$3B,$92,$32,$FF,$58,$F9,$89,$1F,$D7
	.dc.b	$5A,$FF,$AB,$7C,$EF,$F0,$E3,$7D,$5F,$F1,$EC,$62,$3F,$AE,$B5,$FF
	.dc.b	$4D,$35,$1F,$DC,$63,$89,$1F,$F1,$7E,$1F,$87,$7F,$40,$D5,$C6,$8C
	.dc.b	$37,$8F,$FC,$FB,$18,$23,$83,$FC,$F9,$F4,$3F,$84,$AF,$F5,$77,$0A
	.dc.b	$31,$1E,$68,$C7,$0F,$02,$47,$FB,$3E,$32,$D9,$E2,$A3,$2F,$F8,$03
	.dc.b	$E1,$5F,$95,$BE,$1F,$FC,$23,$18,$6C,$66,$7F,$DB,$93,$97,$B6,$DE
	.dc.b	$A2,$38,$21,$FE,$4B,$C1,$1A,$8C,$3F,$E7,$4F,$A5,$FF,$11,$7D,$1F
	.dc.b	$F8,$89,$B3,$FE,$EB,$6B,$FE,$18,$6B,$FE,$D5,$5F,$4B,$F4,$E3,$83
	.dc.b	$5E,$0D,$78,$35,$FF,$11,$FC,$57,$FA,$19,$34,$8F,$F9,$97,$EE,$5F
	.dc.b	$36,$F8,$FF,$C5,$E3,$6B,$E8,$56,$A7,$27,$DB,$FC,$5B,$F8,$C3,$89
	.dc.b	$DD,$F1,$BE,$5C,$0C,$39,$52,$73,$18,$77,$87,$49,$D6,$1F,$89,$AD
	.dc.b	$3F,$8B,$FB,$B7,$99,$E7,$AF,$C8,$C6,$27,$F1,$8D,$8F,$F4,$23,$5F
	.dc.b	$CB,$FE,$21,$FE,$39,$9F,$56,$48,$C4,$F8,$8F,$C9,$57,$3C,$BD,$B3
	.dc.b	$A6,$FB,$1F,$07,$0F,$D0,$4E,$5F,$46,$39,$C3,$F3,$93,$97,$F6,$76
	.dc.b	$F9,$8D,$74,$FC,$77,$86,$6D,$5F,$31,$AB,$E1,$9F,$D3,$C3,$83,$63
	.dc.b	$38,$FA,$F0,$E5,$7C,$71,$EF,$F8,$33,$EF,$E9,$E7,$F0,$DF,$A1,$7F
	.dc.b	$10,$CF,$B5,$6D,$B5,$C1,$FC,$25,$70,$CB,$07,$CB,$D3,$F2,$1D,$7F
	.dc.b	$66,$3F,$BA,$57,$EE,$CF,$F6,$39,$5F,$F4,$AB,$5F,$EF,$BC,$BE,$37
	.dc.b	$A3,$76,$FC,$D5,$F6,$AF,$89,$39,$7F,$C0,$5F,$86,$3E,$CF,$C5,$87
	.dc.b	$F2,$48,$C4,$FF,$B6,$1B,$1F,$EA,$66,$BF,$E0,$E6,$BD,$27,$9D,$A7
	.dc.b	$33,$D6,$73,$F0,$9F,$E7,$9B,$FB,$0D,$9F,$8D,$5E,$0C,$7D,$73,$97
	.dc.b	$DB,$C2,$FA,$0E,$ED,$C3,$FB,$C4,$60,$8E,$09,$87,$F8,$06,$3C,$FF
	.dc.b	$84,$3D,$0F,$C8,$B6,$BF,$19,$7D,$6F,$ED,$4E,$5F,$B9,$E7,$3E,$CF
	.dc.b	$E0,$B6,$FF,$C3,$71,$82,$38,$24,$FF,$87,$A3,$E4,$FA,$5F,$E2,$69
	.dc.b	$FC,$E3,$67,$BC,$FE,$9A,$31,$9F,$E0,$3E,$A1,$F4,$1F,$49,$B8,$7F
	.dc.b	$B2,$30,$FE,$C3,$18,$7F,$DC,$2D,$7F,$C9,$2D,$7F,$C9,$0D,$78,$CF
	.dc.b	$53,$EF,$FD,$E7,$EB,$AF,$F2,$2D,$BE,$33,$FB,$C8,$C5,$7F,$72,$F6
	.dc.b	$5F,$35,$F8,$86,$DF,$F9,$AA,$30,$47,$04,$9F,$F3,$9C,$7E,$43,$F0
	.dc.b	$47,$CA,$FF,$0A,$6D,$7F,$05,$BE,$B7,$FB,$39,$CB,$FC,$7F,$A5,$EF
	.dc.b	$FF,$0E,$BF,$91,$18,$EF,$FD,$3B,$63,$33,$BA,$FF,$1E,$B5,$FF,$35
	.dc.b	$B5,$FF,$0C,$CF,$17,$75,$FA,$8F,$99,$7C,$B7,$F2,$26,$E1,$FC,$33
	.dc.b	$0E,$DB,$FE,$FA,$73,$87,$F0,$DF,$C7,$7F,$7C,$FF,$87,$9B,$97,$FA
	.dc.b	$A3,$8F,$F9,$6C,$6B,$B0,$F9,$6B,$86,$9F,$BA,$8C,$4E,$A3,$F7,$BF
	.dc.b	$80,$FF,$21,$AF,$E3,$8B,$D5,$FE,$72,$E7,$3F,$E2,$6E,$75,$FE,$D1
	.dc.b	$F8,$27,$F8,$C7,$F0,$6F,$E2,$2D,$8F,$CB,$37,$0F,$F1,$F9,$FE,$40
	.dc.b	$70,$E1,$AE,$7D,$65,$74,$6D,$74,$6C,$F4,$3F,$C8,$7C,$3F,$BD,$C7
	.dc.b	$F9,$4B,$6B,$A5,$FF,$54,$8C,$57,$F9,$2B,$EA,$FF,$B2,$57,$FA,$B5
	.dc.b	$B1,$9E,$8F,$EB,$71,$8D,$8D,$24,$6B,$CB,$FE,$C3,$F5,$6F,$91,$BE
	.dc.b	$0E,$1B,$97,$F5,$F9,$D3,$67,$4D,$3F,$C2,$C3,$F0,$47,$C5,$FF,$40
	.dc.b	$8C,$36,$33,$1F,$C1,$CC,$95,$CE,$9F,$26,$32,$D8,$CB,$F6,$23,$2F
	.dc.b	$FB,$53,$E4,$DF,$23,$FD,$F9,$FF,$71,$CF,$F7,$97,$2D,$F8,$8F,$F0
	.dc.b	$1E,$8F,$A1,$A7,$F8,$13,$F4,$37,$C8,$E2,$FF,$C0,$55,$B6,$D6,$E7
	.dc.b	$59,$7F,$73,$F9,$DF,$E2,$7C,$65,$F7,$BF,$E1,$27,$26,$30,$FF,$07
	.dc.b	$38,$8D,$7A,$BB,$7F,$C3,$DD,$5F,$E8,$DC,$7F,$24,$4B,$FC,$93,$71
	.dc.b	$FC,$94,$EE,$72,$48,$FC,$B6,$2F,$9A,$FD,$54,$7F,$82,$DB,$D5,$FE
	.dc.b	$59,$BE,$53,$9E,$DF,$F1,$5F,$D3,$FF,$AE,$71,$FF,$13,$3F,$E7,$C6
	.dc.b	$FC,$3E,$BE,$07,$6D,$8C,$9C,$23,$9F,$D9,$C8,$FF,$24,$FE,$88,$FA
	.dc.b	$BC,$65,$FE,$1F,$C5,$78,$D3,$FC,$CB,$7B,$C3,$C7,$3F,$22,$5F,$0D
	.dc.b	$FC,$11,$B3,$FC,$11,$B8,$7C,$86,$BF,$C0,$DA,$FF,$07,$6A,$D7,$FC
	.dc.b	$E5,$5D,$4F,$F0,$BF,$D2,$7F,$8B,$EB,$FC,$52,$DB,$E5,$3F,$E5,$48
	.dc.b	$C5,$FF,$B2,$BE,$A3,$C7,$7F,$E3,$06,$DF,$FB,$5B,$85,$1F,$DD,$A3
	.dc.b	$0E,$AB,$96,$1F,$E6,$08,$C5,$70,$F5,$31,$39,$F9,$29,$87,$F9,$7E
	.dc.b	$DF,$61,$8C,$7F,$C9,$D0,$AC,$65,$E5,$AB,$7E,$2F,$07,$FC,$6F,$7F
	.dc.b	$E1,$BC,$3F,$97,$E7,$E8,$9C,$9A,$F0,$FC,$C7,$9D,$B7,$E3,$8F,$F3
	.dc.b	$BF,$C6,$FE,$D2,$30,$7F,$A3,$9C,$5F,$9A,$77,$D1,$CA,$3F,$CE,$D3
	.dc.b	$AC,$FE,$67,$CA,$F9,$89,$87,$F8,$6F,$2F,$9B,$1E,$AF,$A3,$F9,$47
	.dc.b	$CB,$6C,$FF,$A7,$BF,$1C,$FE,$1F,$8F,$F4,$1F,$71,$18,$8F,$F7,$DB
	.dc.b	$5F,$F7,$D3,$5F,$C0,$7C,$CF,$D7,$3F,$D2,$B1,$D3,$2F,$C1,$46,$6F
	.dc.b	$EB,$3A,$F1,$31,$7F,$5A,$2D,$7F,$B2,$2F,$95,$FE,$1A,$7F,$09,$F3
	.dc.b	$5F,$58,$91,$F0,$78,$7F,$CD,$18,$7D,$87,$FD,$D7,$1F,$85,$F8,$EE
	.dc.b	$2D,$3E,$17,$67,$FB,$7F,$E1,$1D,$BE,$E8,$C3,$F0,$3E,$16,$E2,$DF
	.dc.b	$E3,$D6,$1B,$82,$7F,$DE,$37,$CE,$31,$87,$69,$E5,$AF,$FA,$76,$31
	.dc.b	$7C,$A7,$89,$AC,$3B,$57,$F0,$78,$CC,$FD,$8C,$73,$5E,$3F,$8C,$7E
	.dc.b	$05,$4C,$72,$DC,$FC,$7A,$7F,$BA,$E7,$19,$74,$BF,$F7,$96,$9E,$47
	.dc.b	$9D,$F0,$9F,$F7,$1D,$70,$AD,$61,$70,$E9,$7B,$CD,$8F,$2C,$7B,$5E
	.dc.b	$DF,$99,$F9,$B3,$EC,$BE,$67,$D5,$7F,$1E,$CF,$E2,$CE,$72,$E9,$39
	.dc.b	$CF,$E8,$66,$AD,$63,$F1,$AB,$4F,$6C,$7A,$F7,$7D,$9A,$DE,$5F,$4A
	.dc.b	$F7,$87,$3B,$DE,$5F,$07,$17,$D3,$39,$DF,$13,$98,$C3,$F1,$FD,$E7
	.dc.b	$D8,$7D,$9D,$3F,$05,$87,$2C,$38,$BF,$80,$31,$7C,$E3,$33,$F8,$9B
	.dc.b	$E7,$C7,$F9,$4B,$D9,$C7,$1F,$0B,$FE,$9F,$6B,$FE,$3F,$6B,$FE,$4C
	.dc.b	$6B,$F9,$95,$C9,$FD,$4A,$39,$35,$E4,$D7,$93,$5E,$4F,$1B,$F7,$CF
	.dc.b	$4E,$0F,$8D,$39,$E1,$FF,$41,$FC,$1F,$F8,$77,$0C,$D4,$7D,$95,$B5
	.dc.b	$92,$B8,$47,$E6,$7E,$2B,$FC,$CF,$97,$E6,$7E,$52,$E5,$F8,$6E,$64
	.dc.b	$62,$FF,$79,$1F,$EE,$B7,$F3,$0D,$7F,$49,$1B,$E3,$FC,$DB,$67,$57
	.dc.b	$FA,$7C,$31,$32,$F9,$C7,$1F,$BF,$75,$52,$4F,$E9,$E1,$D9,$88,$FD
	.dc.b	$FC,$70,$75,$3C,$E7,$58,$7F,$0A,$AD,$57,$0A,$D3,$C0,$86,$4F,$05
	.dc.b	$7B,$7E,$4B,$F2,$71,$FC,$39,$B7,$FC,$59,$BA,$7E,$C1,$FF,$45,$46
	.dc.b	$79,$39,$12,$BF,$81,$4E,$AB,$84,$E6,$7B,$3E,$17,$EA,$78,$5F,$E0
	.dc.b	$8F,$3F,$FD,$43,$E0,$FF,$83,$5B,$D3,$FE,$0A,$6D,$7F,$84,$FE,$42
	.dc.b	$DF,$F8,$41,$FF,$80,$11,$93,$FA,$43,$88,$D4,$7F,$4B,$FA,$BE,$3E
	.dc.b	$11,$A7,$6E,$46,$AF,$5F,$37,$FE,$13,$AA,$B4,$69,$F1,$EB,$31,$FE
	.dc.b	$27,$AB,$39,$AF,$E9,$53,$C1,$FE,$A1,$7D,$DF,$F5,$C1,$8B,$F6,$E1
	.dc.b	$E4,$6C,$EE,$7F,$CD,$38,$BE,$B7,$DA,$48,$FF,$15,$7E,$29,$F0,$0F
	.dc.b	$89,$26,$92,$3F,$B8,$70,$7C,$3F,$C3,$BF,$C5,$D1,$E8,$D3,$F9,$9C
	.dc.b	$62,$3F,$C6,$73,$F3,$0E,$1E,$74,$8F,$F1,$D4,$65,$A4,$3F,$C7,$7F
	.dc.b	$07,$F5,$0F,$29,$1F,$F0,$13,$F0,$BE,$C3,$FD,$91,$1E,$33,$FB,$C3
	.dc.b	$87,$EA,$61,$FD,$22,$31,$BF,$CF,$DD,$0F,$3F,$F0,$0E,$AB,$87,$9F
	.dc.b	$83,$33,$F9,$0F,$4B,$FD,$05,$8F,$87,$91,$18,$8F,$08,$8C,$F8,$9C
	.dc.b	$7E,$57,$93,$27,$95,$71,$D3,$FC,$EB,$5F,$9E,$D9,$F1,$74,$97,$F0
	.dc.b	$47,$E4,$5F,$ED,$07,$0D,$7F,$16,$D7,$FC,$34,$7F,$86,$DC,$38,$6B
	.dc.b	$9F,$82,$FD,$6B,$D9,$B1,$A8,$C1,$F9,$8E,$0F,$9B,$E9,$7C,$C2,$47
	.dc.b	$D7,$87,$E8,$BF,$EF,$68,$C7,$12,$B7,$1F,$2A,$B9,$BC,$6B,$83,$FE
	.dc.b	$C5,$F0,$3F,$81,$FB,$AF,$52,$98,$89,$3F,$5E,$74,$D9,$CB,$FE,$CF
	.dc.b	$EB,$E6,$47,$E7,$7B,$BF,$EB,$E6,$FB,$7F,$12,$31,$1F,$B8,$6B,$F3
	.dc.b	$83,$87,$E8,$B5,$FF,$26,$35,$DC,$75,$3F,$8D,$79,$DF,$03,$8F,$F8
	.dc.b	$47,$37,$F8,$5B,$EA,$7F,$91,$C9,$BF,$0D,$30,$FC,$B4,$E5,$B3,$97
	.dc.b	$F2,$F4,$F8,$57,$D3,$7F,$2E,$6C,$FE,$7D,$AF,$D5,$F8,$6F,$CC,$77
	.dc.b	$7F,$B0,$E7,$39,$F9,$9B,$19,$BF,$E5,$D5,$C2,$FF,$73,$EE,$7F,$3C
	.dc.b	$FC,$03,$DD,$7F,$EC,$D6,$CE,$8F,$A7,$E8,$7D,$12,$3F,$D0,$B8,$3F
	.dc.b	$C0,$2D,$8C,$BF,$55,$AF,$FB,$0B,$A9,$F5,$5F,$D3,$17,$E3,$34,$C4
	.dc.b	$49,$FE,$01,$3A,$6C,$E5,$F9,$7F,$5C,$F8,$FF,$EA,$ED,$8F,$E1,$0D
	.dc.b	$77,$EA,$EA,$64,$69,$E3,$1F,$D8,$3C,$7F,$C5,$CF,$F1,$16,$BF,$69
	.dc.b	$BC,$5F,$D3,$63,$F2,$3C,$9F,$DA,$9A,$7C,$77,$F3,$44,$70,$43,$F3
	.dc.b	$8E,$5F,$CE,$72,$7F,$98,$38,$FF,$B9,$FB,$6F,$A1,$A6,$23,$FB,$BF
	.dc.b	$CA,$7F,$A6,$2F,$FD,$B8,$D7,$9B,$5F,$F7,$AB,$5F,$F5,$85,$BE,$57
	.dc.b	$FC,$87,$97,$E8,$9B,$CB,$A6,$D7,$F9,$2B,$EB,$7F,$65,$5F,$A8,$FC
	.dc.b	$41,$F1,$BF,$E0,$A6,$D7,$F3,$B6,$BF,$CE,$BD,$E7,$CC,$72,$CB,$FD
	.dc.b	$85,$18,$9F,$F0,$8B,$5F,$EF,$2D,$7F,$81,$35,$FE,$62,$D7,$9C,$FF
	.dc.b	$2E,$8C,$47,$F9,$8A,$3F,$98,$E6,$F9,$97,$E1,$1F,$E2,$26,$D7,$F5
	.dc.b	$36,$BF,$BC,$F8,$1F,$F2,$7B,$FD,$5D,$AF,$DA,$6B,$FD,$12,$BF,$CE
	.dc.b	$1E,$3F,$F8,$A6,$73,$3F,$49,$AF,$F5,$D6,$CF,$08,$FF,$0C,$B5,$FC
	.dc.b	$3F,$CE,$FE,$4F,$6F,$E3,$2D,$D3,$EE,$B6,$7F,$B4,$7C,$FF,$F7,$36
	.dc.b	$FA,$BF,$E3,$B8,$FE,$FD,$1E,$76,$BF,$86,$6B,$C0,$F1,$B8,$C0,$D3
	.dc.b	$1F,$88,$3E,$CB,$FC,$9A,$DA,$FE,$FC,$D7,$F9,$3F,$C8,$FE,$67,$3F
	.dc.b	$E6,$6F,$E5,$E3,$FC,$00,$DC,$0D,$27,$C7,$7F,$95,$5B,$3F,$21,$E1
	.dc.b	$3F,$C1,$6B,$F9,$FB,$74,$FD,$1B,$6B,$FC,$1D,$E8,$7E,$02,$BF,$A2
	.dc.b	$BF,$82,$8F,$B8,$D7,$F5,$6D,$78,$F0,$FF,$99,$E3,$14,$69,$23,$FC
	.dc.b	$D9,$F4,$7F,$EF,$BA,$FF,$11,$35,$FE,$96,$DB,$FE,$AF,$F0,$BF,$E1
	.dc.b	$A9,$FF,$14,$35,$FE,$16,$D7,$F2,$4D,$7F,$80,$B5,$FF,$16,$37,$03
	.dc.b	$49,$EB,$7F,$A0,$7E,$4F,$FC,$FD,$5F,$E3,$76,$BF,$D8,$1B,$7F,$DB
	.dc.b	$3F,$15,$7D,$07,$FA,$39,$B5,$F9,$C6,$BF,$C8,$9A,$FE,$A5,$AF,$F1
	.dc.b	$8E,$9F,$E4,$8F,$F9,$1D,$AF,$F7,$06,$D7,$FA,$0D,$7B,$AF,$94,$FF
	.dc.b	$4F,$C6,$B8,$3A,$78,$58,$3F,$D4,$B1,$C2,$7F,$D4,$BF,$3F,$F9,$5A
	.dc.b	$FE,$52,$DC,$7F,$CA,$8D,$E2,$FF,$4F,$F9,$DC,$39,$65,$7F,$97,$6B
	.dc.b	$5F,$0B,$FD,$33,$04,$FE,$17,$E9,$DF,$01,$FE,$A9,$6C,$7F,$8F,$DA
	.dc.b	$FF,$8E,$FC,$BF,$F6,$8F,$07,$F9,$CE,$72,$DA,$E4,$7F,$AC,$E7,$77
	.dc.b	$FE,$FD,$F8,$BF,$F5,$1C,$7F,$B6,$B0,$FF,$2D,$EA,$8C,$46,$CF,$F2
	.dc.b	$A7,$D4,$7A,$5F,$CE,$36,$4D,$CD,$F6,$DF,$EC,$38,$D1,$37,$FF,$63
	.dc.b	$4E,$AF,$FC,$E5,$F5,$4F,$8E,$FF,$62,$47,$FA,$EA,$31,$B7,$FB,$59
	.dc.b	$B1,$99,$DB,$FE,$47,$87,$C7,$B8,$78,$36,$7A,$35,$E0,$D7,$FD,$93
	.dc.b	$D7,$FE,$78,$7F,$4A,$7F,$9E,$9C,$1F,$13,$D7,$FF,$69,$63,$FE,$A5
	.dc.b	$D1,$18,$E1,$FB,$FB,$FF,$B9,$3E,$63,$18,$FC,$AA,$B3,$F2,$9B,$7C
	.dc.b	$AF,$FD,$01,$B3,$0C,$D7,$0E,$1F,$F7,$75,$7F,$A2,$5B,$8F,$F9,$A7
	.dc.b	$1F,$B9,$F1,$DF,$E1,$B8,$F1,$CE,$9F,$C0,$F2,$FF,$84,$6F,$F2,$1D
	.dc.b	$3D,$6F,$CB,$39,$AF,$F4,$63,$75,$FC,$95,$6E,$F8,$9F,$EB,$D8,$7C
	.dc.b	$EF,$C0,$F0,$3F,$E0,$03,$F0,$36,$FE,$0F,$5B,$8D,$57,$08,$C1,$F0
	.dc.b	$BB,$B3,$3C,$FE,$A5,$F2,$54,$F9,$BD,$CF,$AA,$FD,$7E,$1F,$EC,$FA
	.dc.b	$DB,$6B,$83,$FE,$E3,$AE,$1B,$60,$FF,$3E,$C3,$E5,$BE,$54,$7F,$B5
	.dc.b	$63,$07,$F7,$82,$BF,$E3,$E6,$BF,$CD,$0F,$85,$CD,$E4,$7A,$CE,$1F
	.dc.b	$8D,$E2,$FF,$04,$AF,$F0,$1B,$63,$FC,$92,$D7,$FC,$20,$D7,$FC,$7E
	.dc.b	$D7,$FD,$D4,$DD,$CD,$31,$DC,$F3,$BE,$B3,$6F,$FA,$AF,$81,$89,$FB
	.dc.b	$5C,$EF,$96,$F8,$78,$FF,$5F,$6C,$66,$FF,$C3,$6D,$CB,$FC,$C2,$D7
	.dc.b	$FB,$6C,$F0,$24,$FF,$88,$BE,$A9,$F2,$3F,$89,$6C,$7F,$2A,$E5,$FE
	.dc.b	$EF,$59,$AF,$C7,$79,$7F,$F0,$02,$BF,$C0,$9E,$57,$9A,$31,$8F,$F9
	.dc.b	$C9,$AF,$F8,$E5,$AF,$F9,$A1,$AF,$F8,$52,$39,$1F,$03,$CE,$FF,$0A
	.dc.b	$D7,$8C,$FF,$15,$57,$8A,$7F,$41,$3A,$AF,$F6,$3F,$D5,$3E,$9B,$E4
	.dc.b	$B7,$0F,$F2,$0B,$67,$FD,$0E,$D7,$FC,$9C,$D7,$FC,$FA,$D5,$B2,$30
	.dc.b	$DA,$FA,$FE,$97,$EB,$6B,$FD,$B2,$DD,$7F,$DC,$B7,$E1,$7F,$C9,$FD
	.dc.b	$5F,$B9,$F4,$DF,$55,$F8,$06,$E7,$FE,$58,$6E,$3F,$E6,$59,$D1,$27
	.dc.b	$4D,$8F,$F5,$1F,$AA,$FA,$AF,$E0,$CD,$AF,$F3,$85,$62,$7F,$7B,$EB
	.dc.b	$FF,$08,$BF,$F4,$2B,$5F,$F4,$2B,$5F,$F4,$2F,$81,$FB,$8E,$DF,$F4
	.dc.b	$9B,$67,$53,$FE,$87,$8C,$1F,$9C,$F6,$DF,$01,$FC,$65,$B9,$7F,$07
	.dc.b	$BF,$16,$DF,$EA,$BA,$FE,$3D,$EE,$FE,$F6,$15,$FE,$B2,$8C,$3F,$EA
	.dc.b	$2F,$77,$F2,$73,$AC,$BE,$CD,$69,$B9,$7E,$BA,$3C,$FC,$86,$98,$FA
	.dc.b	$AF,$E3,$B4,$BE,$94,$9F,$ED,$D8,$FE,$5E,$BF,$0F,$1C,$2B,$6E,$A3
	.dc.b	$94,$FF,$86,$3D,$8F,$ED,$9F,$05,$FC,$89,$B1,$FE,$1C,$99,$18,$70
	.dc.b	$1C,$3A,$F4,$7F,$DF,$D3,$A8,$D4,$61,$E3,$3C,$1F,$14,$65,$FF,$58
	.dc.b	$6B,$FC,$37,$1F,$7B,$6F,$78,$7F,$51,$8C,$57,$E9,$BE,$AB,$FB,$3D
	.dc.b	$FC,$66,$DF,$FB,$A1,$FD,$13,$7C,$DF,$EB,$51,$AA,$E5,$18,$EA,$CB
	.dc.b	$FD,$7F,$EA,$9E,$FD,$F0,$30,$DC,$7E,$97,$02,$B7,$18,$7F,$8F,$E3
	.dc.b	$F4,$8F,$DE,$BA,$71,$3C,$6F,$FC,$1B,$4F,$AA,$7C,$7F,$F5,$88,$C3
	.dc.b	$63,$31,$F4,$4C,$95,$CD,$FD,$A8,$CB,$63,$2F,$E0,$63,$2F,$FB,$8F
	.dc.b	$E3,$9F,$15,$FE,$01,$5F,$43,$63,$0F,$F8,$17,$57,$D5,$53,$FC,$4B
	.dc.b	$F2,$1F,$E7,$BC,$7E,$3C,$E9,$B3,$B7,$F3,$9E,$52,$B8,$65,$F1,$AF
	.dc.b	$83,$F2,$3E,$AB,$F9,$05,$FF,$9D,$37,$43,$E2,$7F,$C3,$0E,$AB,$F9
	.dc.b	$6B,$9A,$FB,$2E,$6B,$8D,$70,$F1,$78,$BF,$E4,$0E,$AF,$F2,$A7,$C1
	.dc.b	$FE,$A7,$AF,$CB,$97,$E8,$E3,$04,$CB,$F4,$95,$C1,$B5,$C2,$BF,$47
	.dc.b	$5F,$DC,$BE,$5B,$F7,$1D,$5F,$C0,$1B,$7A,$3F,$1E,$E2,$BF,$A3,$FC
	.dc.b	$DB,$E0,$FF,$C6,$CB,$EA,$AC,$61,$B7,$F9,$89,$D3,$67,$4E,$7E,$5B
	.dc.b	$99,$FE,$5E,$FA,$AF,$F5,$AC,$64,$98,$7E,$6A,$D5,$BF,$89,$E3,$5F
	.dc.b	$C5,$F6,$FE,$39,$F4,$BF,$C4,$75,$34,$93,$F2,$23,$2D,$8C,$E3,$FC
	.dc.b	$32,$72,$FF,$16,$E2,$7C,$FF,$F8,$EA,$7F,$94,$B6,$FF,$91,$47,$03
	.dc.b	$3E,$DF,$F9,$77,$E8,$F1,$3F,$8B,$6E,$1F,$DD,$1F,$F0,$15,$7F,$34
	.dc.b	$73,$81,$A4,$F5,$BA,$65,$D7,$E8,$66,$3F,$C2,$29,$F5,$58,$CB,$F8
	.dc.b	$1C,$2B,$19,$77,$AB,$7E,$5F,$0F,$FC,$73,$7F,$C3,$B0,$FE,$75,$9F
	.dc.b	$8E,$72,$6B,$C5,$F7,$1E,$F6,$DF,$8E,$3F,$D3,$7F,$11,$FB,$0C,$BD
	.dc.b	$D1,$8E,$8D,$E3,$FC,$6F,$BF,$FD,$39,$18,$AF,$EC,$51,$8C,$3D,$DF
	.dc.b	$41,$E6,$23,$FD,$1F,$3F,$91,$1E,$AF,$9C,$FA,$47,$D3,$6C,$FF,$04
	.dc.b	$FA,$AF,$E2,$2F,$F4,$FB,$46,$32,$FF,$54,$35,$FF,$72,$B5,$FB,$9F
	.dc.b	$41,$F7,$CF,$F5,$CC,$71,$CB,$EE,$46,$6F,$EB,$3A,$F1,$71,$7F,$5A
	.dc.b	$2D,$7C,$0B,$E5,$7D,$8F,$BD,$F0,$1F,$58,$8F,$C1,$E1,$FF,$3B,$61
	.dc.b	$F9,$97,$FD,$C3,$1F,$85,$F8,$EF,$CA,$57,$E5,$5F,$3F,$57,$F8,$3F
	.dc.b	$E1,$5D,$BF,$ED,$6C,$BF,$6B,$92,$72,$B6,$F8,$FD,$5C,$6F,$86,$16
	.dc.b	$B5,$87,$2C,$38,$E1,$F3,$AB,$4F,$2C,$D7,$DF,$0E,$71,$8B,$E3,$87
	.dc.b	$39,$CD,$6C,$EF,$8F,$49,$D4,$79,$FE,$31,$F2,$5E,$06,$E3,$D7,$0F
	.dc.b	$F7,$1F,$3E,$6A,$4E,$6F,$94,$71,$9C,$C7,$87,$36,$1D,$23,$0D,$CF
	.dc.b	$E0,$BF,$EF,$9C,$3A,$57,$0C,$39,$D6,$9E,$D3,$AA,$E5,$8F,$C6,$B7
	.dc.b	$D5,$3E,$A7,$E5,$FD,$57,$F1,$AC,$BE,$7C,$E7,$1E,$F3,$9C,$39,$4E
	.dc.b	$73,$F5,$4E,$5F,$BD,$19,$BF,$B3,$39,$C7,$C3,$8F,$83,$BB,$ED,$DE
	.dc.b	$F1,$F3,$5F,$3C,$5F,$45,$FE,$00,$F8,$AF,$99,$18,$E8,$47,$CF,$F7
	.dc.b	$DF,$3B,$F6,$AB,$53,$C2,$F9,$47,$E0,$4C,$9F,$83,$AE,$3C,$BF,$93
	.dc.b	$8C,$EF,$F9,$A7,$67,$F9,$52,$B9,$B7,$1F,$1B,$FE,$53,$8C,$BF,$E5
	.dc.b	$D5,$CB,$FD,$42,$34,$D7,$A3,$5E,$0D,$7F,$3F,$1C,$B6,$F6,$4F,$3B
	.dc.b	$F3,$4E,$6B,$D9,$F0,$9F,$E2,$47,$77,$1F,$85,$3B,$AF,$DB,$E4,$BE
	.dc.b	$15,$F9,$4F,$89,$FF,$3B,$57,$E5,$3E,$4E,$F0,$7E,$37,$95,$18,$BF
	.dc.b	$D5,$5D,$E4,$FC,$FB,$6B,$73,$F9,$7A,$C4,$C8,$C1,$0F,$11,$FA,$2F
	.dc.b	$AA,$75,$75,$6C,$66,$31,$D7,$96,$46,$4E,$51,$87,$F6,$93,$C9,$B3
	.dc.b	$C9,$E0,$60,$B9,$FF,$16,$95,$7B,$FE,$4B,$F0,$75,$FB,$F6,$D7,$F1
	.dc.b	$C6,$E9,$FB,$57,$FD,$3F,$19,$E4,$E4,$4A,$FE,$03,$3A,$AE,$13,$99
	.dc.b	$EC,$F8,$5F,$A5,$E1,$7F,$81,$7E,$5F,$FD,$5D,$E0,$7F,$82,$5B,$D9
	.dc.b	$FE,$0F,$6D,$7F,$83,$FE,$46,$DF,$F8,$35,$FF,$57,$46,$4F,$E9,$4E
	.dc.b	$23,$51,$FD,$37,$EA,$B8,$F8,$46,$9D,$B9,$1A,$BD,$7C,$DF,$F8,$56
	.dc.b	$AB,$FD,$FE,$74,$E4,$87,$F8,$62,$3A,$46,$A3,$84,$FA,$2B,$BB,$67
	.dc.b	$53,$FE,$D1,$9C,$BF,$D9,$7E,$97,$D2,$47,$FC,$51,$F5,$5E,$C3,$E0
	.dc.b	$62,$69,$23,$FB,$77,$07,$BF,$EA,$BF,$C5,$91,$FE,$8E,$6B,$FE,$30
	.dc.b	$8F,$8E,$6F,$CE,$91,$FE,$37,$8C,$B4,$87,$F8,$E3,$E1,$3E,$79,$E4
	.dc.b	$23,$FD,$FD,$F8,$5F,$53,$FE,$CC,$8F,$F4,$FB,$5F,$F1,$63,$78,$FF
	.dc.b	$C5,$6B,$CF,$FB,$0E,$79,$3F,$A6,$36,$BF,$BF,$37,$63,$C6,$3F,$DA
	.dc.b	$FE,$93,$C4,$FC,$6B,$63,$31,$8F,$95,$CB,$23,$27,$28,$FC,$AE,$8F
	.dc.b	$94,$A4,$9F,$F0,$2E,$8C,$FF,$C0,$F0,$FA,$AF,$F6,$6B,$A8,$CC,$62
	.dc.b	$36,$7F,$83,$9C,$38,$6B,$C3,$E0,$BF,$4A,$F5,$6C,$6A,$39,$46,$0F
	.dc.b	$A7,$E8,$FC,$C2,$47,$FB,$4B,$EA,$BF,$DE,$2F,$D6,$8C,$BF,$59,$FD
	.dc.b	$54,$6A,$3F,$DF,$7E,$EF,$F1,$27,$FD,$6F,$E9,$FE,$EA,$0D,$F0,$64
	.dc.b	$E8,$87,$27,$87,$4C,$8F,$F6,$0B,$FC,$19,$FF,$60,$7B,$57,$DC,$FF
	.dc.b	$58,$C7,$F0,$66,$BF,$28,$D9,$98,$F8,$F1,$A8,$F9,$ED,$7E,$BB,$5F
	.dc.b	$F1,$FB,$7B,$FF,$C0,$BF,$0D,$FC,$5B,$CC,$7D,$35,$31,$12,$FE,$F0
	.dc.b	$AF,$F0,$6C,$EE,$33,$1D,$1B,$F0,$CF,$A6,$FE,$46,$D8,$CC,$7F,$1B
	.dc.b	$F8,$6F,$CD,$43,$97,$F2,$0F,$7A,$E0,$E7,$97,$FC,$B3,$82,$73,$5C
	.dc.b	$23,$18,$75,$E9,$FD,$8F,$B7,$FC,$D3,$EA,$BF,$01,$7B,$6E,$3F,$B4
	.dc.b	$9F,$A9,$E7,$FC,$E2,$67,$F5,$5A,$FD,$5E,$0F,$EE,$6D,$8C,$BF,$EB
	.dc.b	$CE,$87,$D5,$7F,$44,$5F,$55,$4C,$44,$9F,$97,$3A,$6C,$E5,$F8,$47
	.dc.b	$D3,$FA,$A7,$C8,$7F,$50,$6C,$7D,$46,$BC,$38,$11,$FE,$6A,$8D,$4F
	.dc.b	$12,$3C,$7D,$4F,$EB,$7E,$33,$4A,$49,$FC,$29,$9E,$5F,$CE,$1F,$96
	.dc.b	$23,$F9,$D6,$BF,$25,$AF,$E6,$9F,$EC,$9E,$D7,$DC,$E0,$FB,$EB,$E8
	.dc.b	$7F,$B6,$B6,$BF,$22,$ED,$FC,$B7,$C9,$FF,$A4,$9F,$F6,$C3,$5F,$E2
	.dc.b	$BA,$A3,$11,$FE,$F2,$6B,$F8,$EF,$97,$FF,$1D,$F2,$FE,$35,$B8,$E7
	.dc.b	$E3,$FF,$C0,$3E,$DB,$E5,$BF,$90,$DF,$E4,$FE,$A9,$F1,$FF,$E0,$46
	.dc.b	$E5,$FA,$86,$BF,$72,$DF,$31,$CB,$2F,$F5,$BC,$62,$7F,$C1,$6D,$7F
	.dc.b	$83,$35,$FE,$62,$D7,$F9,$CB,$5E,$73,$FC,$C2,$31,$06,$EA,$F9,$B7
	.dc.b	$D4,$7F,$87,$1B,$5F,$C0,$1A,$FE,$37,$E0,$7F,$95,$37,$C7,$FF,$0F
	.dc.b	$46,$23,$FA,$5D,$FF,$99,$E7,$24,$8F,$E4,$11,$96,$CF,$17,$F8,$0B
	.dc.b	$5F,$E2,$9F,$3D,$FC,$21,$FC,$F3,$76,$FE,$08,$DD,$3F,$84,$FC,$FF
	.dc.b	$F7,$1C,$0F,$F7,$BB,$F2,$B5,$DF,$C7,$7F,$8D,$E3,$11,$FC,$8D,$B3
	.dc.b	$CA,$74,$49,$F5,$FD,$4B,$EA,$BF,$C8,$AD,$8F,$E5,$2D,$7F,$6D,$F2
	.dc.b	$5F,$97,$CF,$F8,$E3,$F9,$48,$FE,$64,$DC,$0D,$27,$C7,$7F,$93,$5B
	.dc.b	$3F,$31,$E1,$3F,$C1,$2B,$F9,$43,$73,$FE,$5A,$DA,$FE,$81,$EC,$FD
	.dc.b	$EA,$FE,$5C,$FE,$02,$3E,$DB,$5F,$E6,$AD,$78,$F0,$FF,$97,$E3,$14
	.dc.b	$69,$23,$FC,$C9,$F4,$7F,$EE,$FA,$FE,$9A,$D7,$F8,$D3,$6F,$FA,$07
	.dc.b	$C3,$7F,$86,$E7,$FA,$9B,$5F,$DF,$B5,$FC,$8B,$5F,$E9,$6D,$7F,$AF
	.dc.b	$37,$03,$49,$EB,$FF,$9E,$7E,$53,$FC,$ED,$5F,$D8,$9A,$FF,$2F,$6D
	.dc.b	$FF,$55,$FA,$A7,$D0,$7F,$A1,$9B,$5F,$A0,$6B,$FC,$59,$AF,$EA,$1A
	.dc.b	$FF,$62,$E8,$FE,$44,$FF,$3D,$6D,$7F,$63,$6E,$DF,$C2,$BC,$E7,$CA
	.dc.b	$FF,$A5,$A7,$37,$FD,$F5,$BA,$FF,$03,$F0,$FF,$8E,$DF,$0A,$FE,$FF
	.dc.b	$F3,$3F,$95,$C7,$FB,$9B,$63,$FC,$0C,$DE,$2F,$F4,$BF,$8D,$FE,$A4
	.dc.b	$8D,$62,$69,$BF,$86,$FE,$9D,$7F,$E1,$0B,$7D,$3B,$E0,$3F,$D4,$4D
	.dc.b	$9D,$35,$FE,$CB,$2F,$2F,$FD,$97,$C1,$FE,$0C,$8E,$77,$C8,$95,$DE
	.dc.b	$B7,$A7,$FB,$D7,$E3,$3F,$D4,$51,$FE,$D4,$9F,$F5,$03,$5D,$1F,$E3
	.dc.b	$7B,$7D,$47,$AA,$3C,$6D,$C7,$C4,$DC,$3F,$DB,$5F,$55,$FE,$BD,$8C
	.dc.b	$57,$E9,$5B,$7D,$37,$FF,$5B,$D7,$F9,$2F,$EA,$9F,$1F,$FE,$29,$C3
	.dc.b	$FC,$91,$18,$CB,$FD,$A2,$D8,$CC,$ED,$FF,$3B,$C3,$E4,$7E,$A4,$72
	.dc.b	$6B,$A6,$BF,$E8,$48,$E1,$C3,$FE,$7A,$8F,$D3,$9F,$E7,$CE,$BC,$11
	.dc.b	$FF,$42,$E7,$F1,$BE,$97,$FC,$7F,$97,$C4,$6E,$CA,$FE,$0D,$FA,$37
	.dc.b	$F9,$5A,$3F,$09,$BA,$B7,$5F,$1D,$B8,$75,$AF,$C5,$5F,$1A,$FA,$75
	.dc.b	$AE,$47,$CD,$FF,$49,$9F,$06,$1F,$EE,$E8,$C4,$7F,$17,$9D,$E1,$F7
	.dc.b	$DB,$1E,$CC,$5E,$D7,$89,$B5,$F9,$BA,$DE,$7F,$82,$AD,$12,$75,$85
	.dc.b	$C3,$FC,$71,$5F,$E0,$A9,$FF,$37,$B6,$3F,$3F,$93,$E6,$FE,$1F,$47
	.dc.b	$F9,$AF,$CF,$B7,$B3,$FC,$CC,$5C,$3D,$F7,$C5,$EB,$3C,$BE,$AB,$F1
	.dc.b	$B1,$FE,$7E,$6E,$3F,$E1,$7F,$AB,$7A,$9B,$E6,$97,$B7,$EA,$F7,$AB
	.dc.b	$73,$93,$F7,$DF,$8C,$3E,$2F,$E7,$F2,$23,$13,$FA,$D6,$E5,$F8,$A6
	.dc.b	$BF,$CF,$3B,$7E,$0F,$39,$FC,$1C,$F0,$E0,$FE,$8F,$1F,$1A,$DF,$37
	.dc.b	$F7,$B8,$BF,$DC,$0D,$8C,$9F,$E1,$42,$E3,$FE,$4A,$6B,$FE,$1C,$6B
	.dc.b	$FE,$96,$6B,$FE,$A9,$6C,$FE,$99,$FE,$AF,$1E,$39,$F4,$BF,$9A,$FD
	.dc.b	$1D,$F4,$7F,$11,$B5,$FC,$3D,$B3,$FA,$C6,$BF,$EF,$8D,$11,$88,$FC
	.dc.b	$EB,$5F,$EE,$4D,$8D,$D7,$C3,$FA,$D7,$CA,$FE,$31,$B8,$7F,$35,$BF
	.dc.b	$0E,$9F,$EB,$19,$CE,$3F,$E8,$FC,$5F,$3B,$F3,$5B,$87,$F8,$F9,$B3
	.dc.b	$FE,$74,$6B,$FE,$0C,$9E,$04,$9E,$0F,$97,$E5,$BC,$87,$E8,$2B,$9E
	.dc.b	$1E,$38,$FE,$07,$FA,$43,$E8,$3F,$1A,$DC,$3F,$9C,$37,$1F,$E2,$AD
	.dc.b	$7E,$53,$5F,$E0,$AD,$7F,$C3,$4D,$C0,$E7,$77,$B1,$FA,$FA,$FF,$2B
	.dc.b	$F5,$79,$31,$FD,$BC,$E6,$9F,$5C,$FA,$BF,$CA,$37,$3F,$F3,$2B,$71
	.dc.b	$FF,$4D,$B5,$FF,$36,$B5,$FF,$73,$B5,$FF,$7D,$36,$BB,$D7,$E8,$6D
	.dc.b	$F5,$1E,$43,$F8,$5D,$77,$DB,$FC,$D7,$1F,$BC,$F8,$96,$5F,$C9,$BA
	.dc.b	$FF,$DB,$AD,$8C,$BF,$E9,$66,$E7,$FE,$96,$6B,$FE,$96,$6B,$FE,$A0
	.dc.b	$6B,$FE,$95,$6E,$EF,$C6,$9F,$43,$FD,$01,$B7,$FD,$27,$2F,$2E,$1C
	.dc.b	$63,$55,$FC,$9B,$E1,$BF,$83,$46,$CF,$EA,$1B,$FF,$B1,$7A,$BF,$8C
	.dc.b	$FE,$92,$F8,$DE,$11,$FC,$73,$E1,$7F,$C7,$77,$FE,$B2,$6E,$5F,$EB
	.dc.b	$26,$BF,$EB,$26,$BF,$EC,$DF,$8B,$FE,$6F,$F5,$FF,$DC,$6B,$C5,$2C
	.dc.b	$BF,$96,$D6,$BD,$0F,$EB,$FE,$DF,$F8,$A7,$E0,$BF,$AA,$B7,$0F,$ED
	.dc.b	$18,$48,$C3,$80,$E1,$D7,$A3,$FE,$95,$9D,$46,$A3,$0F,$19,$E0,$FF
	.dc.b	$75,$6B,$E7,$C3,$FA,$FA,$E1,$FA,$39,$27,$31,$F5,$2B,$CA,$EA,$33
	.dc.b	$87,$D2,$FA,$F7,$DD,$7F,$5C,$6C,$7C,$6C,$10,$FB,$7D,$54,$AD,$4F
	.dc.b	$0C,$BC,$91,$D6,$B5,$1F,$E4,$8F,$D2,$BF,$C9,$13,$D7,$EC,$BF,$8D
	.dc.b	$E7,$FE,$49,$9C,$D7,$23,$EB,$4E,$AB,$FC,$13,$3A,$AF,$83,$39,$AF
	.dc.b	$F5,$04,$E6,$BF,$1F,$39,$E2,$7E,$76,$FA,$3F,$F0,$8C,$61,$B1,$9A
	.dc.b	$FA,$F3,$96,$C6,$70,$FC,$04,$E5,$B1,$97,$F0,$F1,$97,$F1,$3F,$7A
	.dc.b	$FC,$4B,$FC,$65,$7F,$E7,$3E,$B7,$CC,$FF,$8C,$2B,$57,$FB,$18,$FF
	.dc.b	$08,$FD,$5B,$ED,$A9,$FE,$34,$F6,$7F,$97,$70,$7E,$12,$75,$97,$DB
	.dc.b	$6E,$1F,$37,$91,$18,$AF,$F3,$D5,$F2,$BF,$DB,$CE,$5F,$DC,$FD,$AB
	.dc.b	$E6,$BF,$C4,$6F,$CF,$C7,$F5,$8E,$67,$F1,$4E,$71,$39,$9F,$13,$93
	.dc.b	$18,$FE,$71,$E9,$D1,$FE,$9B,$C3,$F6,$7D,$3C,$6F,$F5,$B1,$82,$63
	.dc.b	$FA,$EA,$E0,$DA,$E1,$5F,$AD,$6B,$FA,$DF,$37,$FC,$7B,$D5,$FD,$7A
	.dc.b	$B5,$86,$27,$FC,$83,$CC,$F9,$1B,$E7,$61,$B7,$F9,$D6,$E3,$F3,$EB
	.dc.b	$84,$FE,$86,$BE,$DF,$CC,$48,$FF,$5A,$7D,$07,$FB,$29,$FB,$11,$9B
	.dc.b	$FB,$1C,$28,$FF,$39,$D6,$E7,$31,$FC,$6B,$5C,$47,$F1,$2F,$8B,$F9
	.dc.b	$24,$C3,$E2,$36,$3F,$89,$4F,$07,$E3,$47,$F8,$E7,$E9,$BF,$CA,$BC
	.dc.b	$5F,$C4,$1B,$7A,$3F,$84,$38,$AF,$E5,$DD,$8F,$A4,$FF,$34,$B6,$BF
	.dc.b	$DA,$5F,$6A,$F8,$DC,$D4,$9F,$E2,$55,$BB,$FF,$68,$DE,$E7,$35,$FE
	.dc.b	$74,$FA,$E7,$D8,$D3,$1F,$2D,$6B,$F8,$E4,$E9,$B3,$A6,$9F,$C8,$7E
	.dc.b	$B3,$FC,$F1,$8F,$F3,$DE,$B4,$62,$3F,$A1,$C6,$3E,$1F,$FC,$BD,$39
	.dc.b	$9F,$EB,$EF,$D5,$DF,$F9,$BB,$6F,$35,$FE,$8C,$8C,$57,$F5,$56,$FA
	.dc.b	$3F,$8F,$CF,$5C,$3F,$EB,$2E,$84,$63,$1F,$F7,$E4,$61,$F5,$4E,$E7
	.dc.b	$B7,$C7,$F9,$89,$3F,$D5,$34,$F6,$C7,$AB,$AF,$C8,$70,$9F,$88,$FF
	.dc.b	$B0,$BF,$26,$7D,$87,$FA,$31,$B3,$FE,$F7,$6E,$BF,$ED,$46,$BF,$23
	.dc.b	$5F,$F1,$A1,$FE,$39,$D5,$3C,$37,$F8,$FE,$E7,$C5,$26,$1F,$DE,$AB
	.dc.b	$78,$E3,$1F,$E3,$98,$F4,$BF,$34,$E7,$0F,$A3,$EC,$7D,$62,$63,$F0
	.dc.b	$6F,$FC,$6A,$DF,$C2,$3E,$A1,$F5,$66,$C7,$87,$37,$D9,$7F,$88,$A3
	.dc.b	$FD,$09,$18,$7F,$9E,$E7,$FE,$E0,$8C,$3F,$E2,$81,$A7,$C0,$AE,$51
	.dc.b	$B6,$BA,$9E,$35,$DA,$73,$5C,$63,$BD,$6E,$7F,$C6,$BF,$0D,$FD,$7E
	.dc.b	$3F,$C4,$1D,$EF,$AD,$CF,$5F,$7C,$14,$BE,$1C,$7F,$D1,$AF,$73,$E8
	.dc.b	$FB,$6F,$D9,$3B,$C3,$FD,$FB,$18,$BF,$06,$68,$FF,$5E,$C6,$1D,$4F
	.dc.b	$0E,$DE,$B5,$B6,$CF,$09,$D6,$5E,$6B,$E5,$5A,$37,$F0,$F6,$75,$5C
	.dc.b	$31,$F3,$1E,$9A,$DC,$EA,$30,$DC,$5C,$CF,$9E,$F2,$B6,$74,$BC,$AC
	.dc.b	$CC,$9C,$BC,$23,$17,$E2,$BE,$B3,$8C,$FE,$3E,$1C,$F0,$E2,$79,$6D
	.dc.b	$F5,$DF,$EA,$FF,$47,$0D,$8F,$6F,$C0,$3E,$9F,$E6,$C7,$DC,$9D,$63
	.dc.b	$F8,$38,$E7,$F0,$BF,$EF,$B7,$E2,$98,$C3,$E2,$61,$CF,$6F,$C0,$CE
	.dc.b	$4D,$E1,$E1,$7F,$CC,$4D,$7F,$CC,$ED,$7C,$57,$DD,$DF,$07,$F3,$87
	.dc.b	$FC,$CB,$28,$C4,$6B,$C2,$FF,$34,$BF,$0A,$32,$F6,$79,$65,$ED,$AE
	.dc.b	$55,$F1,$A7,$3C,$3F,$9E,$C9,$F4,$7F,$DA,$F1,$D5,$C7,$DB,$7C,$15
	.dc.b	$70,$F2,$7F,$D0,$79,$7E,$7B,$ED,$1E,$EE,$33,$FD,$5F,$22,$73,$8F
	.dc.b	$E3,$5A,$FE,$0D,$7D,$2F,$18,$FE,$0B,$1A,$C3,$B6,$19,$BE,$8D,$9E
	.dc.b	$35,$D6,$30,$47,$C1,$7C,$7E,$5B,$E9,$61,$FB,$66,$E3,$E8,$6D,$7F
	.dc.b	$16,$7F,$D3,$D3,$BE,$2E,$44,$9F,$D4,$4E,$AB,$94,$E6,$7F,$41,$3D
	.dc.b	$32,$3B,$8E,$D3,$BB,$ED,$1D,$0F,$81,$F2,$BF,$85,$7F,$71,$D7,$FE
	.dc.b	$5A,$F0,$FE,$43,$63,$F7,$53,$B7,$F6,$0D,$E7,$FF,$5D,$E4,$8C,$4F
	.dc.b	$D8,$6B,$FC,$D9,$B1,$BE,$1F,$F6,$1D,$72,$9C,$92,$33,$5F,$E0,$2F
	.dc.b	$9E,$FF,$64,$F8,$9F,$E0,$26,$DE,$67,$FC,$21,$F5,$B8,$7F,$C0,$4F
	.dc.b	$E8,$32,$FF,$08,$B7,$C5,$34,$C7,$C1,$71,$9F,$F0,$BB,$64,$F5,$D3
	.dc.b	$E0,$7D,$66,$62,$61,$FE,$21,$75,$AA,$73,$3F,$DB,$A6,$E8,$CF,$E4
	.dc.b	$57,$03,$F6,$F8,$76,$BF,$F7,$0B,$5F,$F7,$43,$5F,$F0,$27,$C5,$79
	.dc.b	$C8,$FF,$8C,$E7,$D4,$7C,$12,$3F,$DE,$5F,$89,$F8,$BF,$B7,$C1,$FE
	.dc.b	$73,$3F,$E0,$76,$BF,$E0,$68,$F8,$86,$FC,$19,$1F,$E4,$18,$CB,$48
	.dc.b	$7F,$83,$BE,$F7,$E6,$1F,$4C,$8F,$F8,$2D,$FA,$DE,$0F,$D1,$24,$7F
	.dc.b	$93,$7E,$86,$BF,$E1,$88,$BA,$FF,$91,$A7,$EB,$7D,$AB,$E6,$29,$2F
	.dc.b	$FC,$45,$F6,$8F,$B3,$C4,$7F,$BA,$78,$17,$C6,$3F,$C5,$0D,$75,$5C
	.dc.b	$5F,$F3,$CF,$95,$F8,$9F,$67,$FC,$0D,$18,$3F,$C0,$B9,$FF,$86,$7E
	.dc.b	$93,$FC,$F9,$18,$E4,$FF,$04,$A8,$CB,$83,$FC,$1A,$E1,$C3,$5E,$2D
	.dc.b	$77,$EC,$FF,$40,$7A,$36,$35,$18,$3F,$CB,$5F,$67,$FA,$47,$C8,$3A
	.dc.b	$91,$98,$FF,$71,$37,$0F,$CD,$47,$F8,$D3,$DD,$FF,$4A,$3C,$AB,$8F
	.dc.b	$67,$EB,$6F,$86,$5F,$E2,$89,$CD,$7F,$AE,$B8,$FF,$65,$F2,$1F,$A9
	.dc.b	$8E,$4B,$AB,$26,$BA,$A4,$EA,$32,$F0,$8F,$DE,$7C,$17,$CD,$8F,$23
	.dc.b	$5F,$F7,$E3,$75,$FA,$DF,$63,$FC,$2E,$7E,$A3,$F2,$48,$FF,$6C,$DD
	.dc.b	$18,$8F,$B0,$70,$34,$E5,$FF,$50,$F9,$DD,$F5,$F8,$4B,$E7,$3F,$89
	.dc.b	$36,$7E,$53,$B7,$EA,$37,$6F,$F0,$EC,$F0,$38,$E1,$FA,$2F,$B4,$97
	.dc.b	$F9,$A7,$F8,$D7,$E9,$6F,$A8,$CC,$FE,$87,$A5,$58,$D7,$EE,$F5,$30
	.dc.b	$E1,$7F,$86,$F8,$C7,$E9,$4F,$87,$FE,$4D,$9F,$DE,$AF,$F3,$BB,$6B
	.dc.b	$73,$F3,$74,$7C,$A7,$E5,$49,$7F,$35,$AF,$CC,$8D,$B9,$21,$FC,$93
	.dc.b	$E4,$9C,$FF,$8A,$F1,$24,$FF,$A8,$67,$2D,$9C,$AD,$BF,$37,$C5,$CA
	.dc.b	$FF,$A0,$5F,$0A,$D7,$27,$E6,$7E,$D5,$F0,$1F,$D2,$5B,$3F,$A1,$6B
	.dc.b	$C0,$FD,$62,$F8,$2E,$13,$27,$83,$CE,$3F,$33,$F7,$F2,$7F,$32,$FE
	.dc.b	$57,$F0,$07,$DE,$D3,$F8,$03,$89,$FF,$20,$B5,$FC,$AB,$5F,$E1,$CD
	.dc.b	$7F,$89,$B5,$FF,$4E,$B5,$39,$D1,$F6,$FE,$B9,$FE,$01,$F0,$1F,$DA
	.dc.b	$1B,$E7,$FF,$6D,$9D,$47,$C3,$F7,$5F,$37,$FC,$5D,$B1,$FE,$58,$6B
	.dc.b	$FA,$C6,$BF,$C9,$70,$7D,$4F,$B7,$8F,$EC,$B7,$87,$F8,$73,$6B,$FC
	.dc.b	$A7,$FA,$67,$F9,$65,$FE,$81,$D5,$E4,$BF,$CF,$B6,$FF,$CA,$6D,$9F
	.dc.b	$F2,$DF,$E9,$3C,$6B,$F9,$0E,$48,$E9,$3F,$CF,$23,$1F,$03,$EB,$9C
	.dc.b	$CF,$F9,$AD,$AF,$48,$D1,$1F,$81,$1A,$CD,$F4,$7F,$EE,$3C,$3F,$2E
	.dc.b	$DC,$BF,$96,$36,$FF,$9D,$7D,$C3,$E4,$3F,$C3,$CD,$8F,$E2,$F8,$23
	.dc.b	$11,$F3,$5A,$FF,$56,$8D,$4F,$F8,$B6,$73,$3F,$CE,$DA,$F8,$23,$47
	.dc.b	$F8,$CF,$E9,$1F,$6B,$FE,$2F,$6C,$FF,$47,$6B,$FD,$CF,$F1,$B7,$D8
	.dc.b	$7F,$8C,$5B,$1F,$E2,$46,$BF,$E2,$46,$BF,$E5,$66,$BF,$E8,$C6,$AF
	.dc.b	$19,$18,$6C,$FF,$8C,$A3,$79,$7F,$63,$F9,$2F,$F2,$32,$69,$FE,$49
	.dc.b	$A7,$DC,$53,$FC,$A4,$E3,$0F,$9E,$D7,$F3,$4D,$7E,$73,$5F,$EC,$CD
	.dc.b	$7F,$BA,$36,$BD,$53,$FE,$5F,$7F,$2C,$7F,$79,$ED,$F3,$9F,$E5,$9F
	.dc.b	$AC,$FD,$3D,$FF,$7E,$B4,$62,$3F,$C9,$AD,$7F,$CF,$ED,$7F,$D6,$AD
	.dc.b	$7E,$53,$63,$9F,$3F,$F9,$6A,$30,$6E,$DF,$3A,$FA,$9F,$F3,$83,$70
	.dc.b	$FF,$04,$35,$FE,$51,$2F,$B6,$7D,$BF,$F9,$DD,$B3,$F9,$A6,$BF,$C1
	.dc.b	$5A,$FE,$65,$AF,$F8,$45,$AF,$F8,$7D,$B3,$DE,$7F,$D1,$51,$8A,$FF
	.dc.b	$1E,$D7,$91,$BC,$DF,$D2,$1B,$CD,$FE,$DA,$FB,$0B,$5F,$8F,$78,$BF
	.dc.b	$EA,$A6,$BB,$F7,$7F,$D0,$31,$88,$FF,$19,$75,$FF,$6B,$CF,$FD,$D7
	.dc.b	$ED,$7F,$A5,$E3,$FB,$70,$F6,$7F,$29,$5F,$E3,$18,$E0,$49,$FF,$53
	.dc.b	$3B,$E4,$FF,$50,$7C,$FF,$F7,$CA,$F9,$ED,$D7,$FB,$1B,$74,$FF,$5F
	.dc.b	$7C,$D7,$0F,$02,$5F,$F8,$FE,$B4,$FF,$B0,$67,$FD,$51,$F6,$9F,$E1
	.dc.b	$9E,$63,$4C,$7C,$7F,$08,$CF,$A1,$C0,$97,$F9,$79,$D1,$E7,$7F,$DA
	.dc.b	$6D,$7F,$DE,$58,$FF,$AA,$A3,$11,$FB,$FB,$79,$7E,$42,$47,$E8,$FF
	.dc.b	$4D,$7D,$2E,$23,$FD,$09,$5C,$1F,$D4,$35,$FF,$44,$4F,$07,$FC,$D2
	.dc.b	$3E,$DD,$F6,$BF,$E0,$D8,$FF,$22,$C6,$38,$3F,$C8,$F1,$B9,$D3,$FE
	.dc.b	$81,$9D,$3C,$3E,$59,$C1,$C2,$51,$89,$D4,$E6,$31,$1D,$BE,$4F,$FD
	.dc.b	$08,$9A,$47,$FD,$9B,$97,$FB,$F7,$8B,$DF,$F1,$FF,$E7,$EC,$FF,$C8
	.dc.b	$7F,$6F,$FE,$60,$9C,$92,$FE,$3D,$6E,$11,$FC,$BE,$78,$3A,$8F,$B9
	.dc.b	$F3,$9F,$E8,$A8,$FB,$6D,$78,$F2,$38,$93,$99,$FE,$75,$18,$AF,$F7
	.dc.b	$3E,$5F,$D2,$1F,$AF,$E3,$78,$EB,$77,$70,$FC,$9C,$DB,$E1,$5F,$80
	.dc.b	$8D,$61,$D6,$31,$87,$FA,$77,$3F,$F1,$D3,$7C,$4E,$9A,$3E,$8F,$C1
	.dc.b	$DB,$FE,$4D,$9B,$7C,$30,$FE,$3A,$D3,$FC,$DF,$7C,$70,$E8,$7F,$AA
	.dc.b	$FE,$DF,$F1,$6D,$E4,$F6,$76,$3E,$DF,$FC,$47,$39,$26,$3F,$CA,$AF
	.dc.b	$8B,$FE,$31,$9D,$B9,$BF,$89,$E0,$7F,$BA,$6F,$FD,$37,$C2,$8C,$57
	.dc.b	$FB,$56,$31,$F2,$1F,$4D,$B5,$FE,$0C,$E5,$7C,$CF,$C2,$6C,$7F,$51
	.dc.b	$C3,$9E,$DF,$E7,$09,$CE,$3F,$D0,$29,$F3,$1F,$7F,$85,$18,$3F,$C3
	.dc.b	$25,$9F,$F2,$23,$5F,$F0,$C3,$5F,$F6,$83,$5F,$F6,$F3,$63,$AC,$F7
	.dc.b	$AF,$CC,$CF,$FB,$6D,$B5,$E5,$9F,$63,$F8,$BF,$27,$FD,$75,$5F,$21
	.dc.b	$B1,$FD,$B3,$E1,$FE,$34,$62,$3F,$B6,$C6,$09,$3C,$5F,$81,$F6,$CF
	.dc.b	$B7,$FC,$5B,$67,$FB,$D5,$F6,$CF,$FD,$33,$39,$C3,$FB,$4E,$CF,$AB
	.dc.b	$F9,$4D,$C3,$FC,$27,$18,$23,$82,$4F,$F8,$5A,$3C,$7F,$0B,$E2,$3F
	.dc.b	$3B,$5C,$E7,$FB,$BD,$FF,$B0,$EF,$FC,$1D,$F6,$EF,$A3,$FC,$33,$63
	.dc.b	$F1,$AD,$AF,$C6,$35,$FE,$12,$7F,$8D,$DC,$3F,$C1,$1A,$E6,$7A,$9F
	.dc.b	$87,$F6,$3F,$59,$5F,$E6,$4E,$AF,$E0,$17,$DF,$4F,$F3,$4C,$7E,$DB
	.dc.b	$D5,$7D,$BF,$E3,$DB,$7F,$E6,$66,$E5,$FE,$91,$6B,$FE,$65,$6B,$FC
	.dc.b	$83,$91,$18,$84,$62,$BB,$1E,$7F,$83,$FE,$0F,$9F,$FB,$33,$0F,$F3
	.dc.b	$24,$E2,$7F,$7F,$F2,$5F,$A2,$9F,$F5,$23,$70,$F8,$E3,$B5,$19,$8F
	.dc.b	$F4,$A3,$5F,$E6,$B1,$A9,$FE,$58,$D7,$A9,$89,$E1,$3F,$E6,$9F,$B0
	.dc.b	$FF,$1D,$4F,$F9,$2D,$B1,$FE,$BD,$9F,$26,$BF,$B8,$8E,$1E,$B7,$F2
	.dc.b	$4F,$B7,$FD,$FC,$FF,$BC,$1B,$87,$93,$ED,$FF,$DF,$6F,$F4,$A9,$D7
	.dc.b	$6F,$F2,$0F,$82,$FF,$21,$CF,$FB,$15,$B8,$FC,$2D,$3F,$D9,$BE,$CF
	.dc.b	$B2,$73,$3F,$E9,$BF,$69,$F2,$9F,$CD,$BE,$DF,$FB,$5E,$3D,$CF,$F6
	.dc.b	$4C,$67,$CE,$FE,$95,$ED,$FF,$72,$F8,$6F,$EA,$4D,$AF,$A3,$8F,$F9
	.dc.b	$7B,$D1,$56,$9B,$3A,$AE,$15,$F6,$1A,$FE,$81,$AF,$F6,$A6,$BF,$8A
	.dc.b	$C3,$FA,$BE,$FF,$37,$A3,$D7,$E7,$79,$B1,$EF,$7F,$E6,$AB,$DE,$3F
	.dc.b	$95,$FB,$77,$C7,$FF,$56,$6C,$7C,$1A,$FC,$BC,$3E,$DF,$55,$2B,$59
	.dc.b	$2B,$8D,$FF,$9E,$1E,$13,$A7,$39,$FE,$0F,$C2,$7C,$73,$E8,$7F,$B3
	.dc.b	$FD,$9F,$F1,$9A,$D4,$E7,$0E,$F3,$97,$FC,$0B,$19,$C7,$FD,$0F,$39
	.dc.b	$7F,$D7,$D1,$9A,$F8,$D3,$9E,$87,$DB,$BE,$A3,$FC,$19,$18,$6C,$67
	.dc.b	$1F,$BB,$39,$6C,$67,$1F,$D8,$CE,$5B,$19,$7F,$8D,$C6,$5F,$E3,$BF
	.dc.b	$2A,$FB,$7F,$E9,$B9,$7C,$5F,$05,$F6,$FF,$E1,$9A,$D5,$FF,$86,$67
	.dc.b	$39,$7D,$2F,$B0,$FF,$2C,$B7,$E0,$3F,$C3,$DE,$EF,$F1,$F8,$FE,$29
	.dc.b	$3A,$BF,$F5,$6B,$72,$FC,$53,$67,$FC,$63,$5C,$6B,$F8,$CC,$E5,$FE
	.dc.b	$37,$F6,$FF,$EE,$3A,$FC,$1B,$F3,$DF,$E4,$4E,$63,$FD,$76,$E5,$E3
	.dc.b	$F0,$DF,$E2,$68,$C7,$31,$E1,$8F,$E7,$5E,$1D,$9F,$E6,$2A,$FD,$61
	.dc.b	$30,$FD,$67,$5F,$7C,$7F,$54,$D7,$F5,$4D,$7F,$6D,$3C,$5F,$E5,$38
	.dc.b	$FF,$04,$F4,$7F,$C7,$3D,$9F,$D7,$73,$FE,$35,$97,$87,$2F,$D0,$CE
	.dc.b	$5F,$E8,$1C,$CF,$8E,$FC,$DF,$A5,$F9,$C6,$CF,$E6,$9B,$9F,$E6,$1B
	.dc.b	$8F,$D2,$F9,$1F,$F2,$4F,$01,$FF,$2C,$7D,$13,$ED,$73,$13,$1F,$C1
	.dc.b	$3A,$8C,$62,$C3,$D3,$C3,$C6,$FE,$3D,$7F,$15,$23,$82,$57,$F1,$59
	.dc.b	$E0,$D9,$FF,$1A,$7D,$2F,$F9,$32,$FF,$89,$36,$FE,$7C,$FF,$98,$A3
	.dc.b	$15,$FD,$8B,$B1,$F4,$BF,$E6,$66,$DF,$F1,$1F,$01,$F6,$FF,$ED,$D8
	.dc.b	$FE,$1F,$3A,$BF,$64,$7F,$6F,$8C,$65,$FC,$8D,$B1,$FE,$8A,$AE,$55
	.dc.b	$C0,$F1,$61,$FE,$76,$FB,$77,$D9,$53,$1F,$6F,$70,$FE,$3F,$3A,$6C
	.dc.b	$E9,$A7,$F2,$4F,$AF,$FF,$3D,$61,$F8,$C6,$BF,$E1,$CE,$94,$62,$3F
	.dc.b	$C1,$0D,$7F,$C0,$EF,$D0,$D3,$F9,$F3,$6B,$7F,$11,$FE,$8E,$8C,$57
	.dc.b	$0E,$0F,$F4,$F7,$2B,$4D,$46,$1B,$87,$CF,$9D,$3E,$99,$D3,$EB,$F1
	.dc.b	$BF,$CE,$58,$7F,$2F,$C3,$D4,$D7,$D3,$1A,$EC,$F2,$1E,$56,$D7,$DD
	.dc.b	$FB,$77,$D9,$7F,$A2,$5B,$1F,$E9,$06,$DF,$FA,$3D,$AF,$83,$DC,$F8
	.dc.b	$07,$FA,$EE,$37,$B7,$C6,$8C,$D7,$D3,$75,$78,$BF,$C1,$6F,$F4,$A7
	.dc.b	$17,$FD,$57,$0E,$78,$7D,$49,$F9,$DE,$CF,$C2,$24,$7F,$80,$2F,$ED
	.dc.b	$C7,$C6,$F9,$3F,$A8,$70,$9F,$C2,$3E,$7F,$B7,$FF,$12,$47,$F9,$7B
	.dc.b	$A1,$8A,$73,$9F,$FB,$8A,$31,$5F,$E9,$28,$57,$0A,$DC,$62,$B6,$FF
	.dc.b	$BB,$67,$55,$C5,$B1,$99,$DC,$F8,$DE,$3F,$0D,$FD,$96,$7F,$A0,$FC
	.dc.b	$BF,$F8,$EE,$31,$7C,$9D,$66,$EF,$56,$FA,$FF,$DF,$7F,$6F,$FB,$2B
	.dc.b	$E2,$D8,$CF,$C4,$6F,$2F,$6C,$E5,$BF,$31,$9C,$D3,$96,$D6,$63,$04
	.dc.b	$9C,$46,$59,$AA,$D8,$F9,$27,$73,$98,$C3,$71,$E1,$8F,$F8,$67,$EC
	.dc.b	$FC,$9C,$A6,$F9,$58,$65,$F1,$CB,$EC,$D6,$E3,$17,$FE,$23,$BD,$D7
	.dc.b	$93,$0E,$F7,$B8,$C1,$CB,$0E,$F8,$72,$9D,$63,$F6,$69,$F5,$9F,$EB
	.dc.b	$3F,$47,$C5,$18,$F8,$7F,$F7,$BB,$D3,$DF,$7D,$1F,$D1,$8D,$57,$0B
	.dc.b	$E5,$1E,$23,$27,$19,$E1,$1A,$C3,$C4,$F0,$9F,$F4,$BB,$5F,$F5,$FB
	.dc.b	$5F,$A8,$D7,$D3,$C1,$FC,$D9,$FF,$37,$C2,$31,$1D,$1A,$F4,$6B,$D1
	.dc.b	$E5,$3E,$46,$EB,$BF,$0B,$EB,$64,$FB,$7F,$EE,$58,$7C,$1B,$DB,$CE
	.dc.b	$B0,$70,$9D,$C7,$D6,$E1,$7D,$BB,$DE,$FF,$6E,$E7,$FA,$A8,$C5,$7E
	.dc.b	$09,$AF,$15,$F6,$FC,$34,$FC,$8C,$6A,$B9,$5E,$6F,$F8,$2C,$E6,$B7
	.dc.b	$79,$C3,$C7,$76,$B5,$7C,$23,$15,$BF,$96,$FC,$2C,$FA,$DB,$8F,$F0
	.dc.b	$26,$E3,$FA,$47,$FD,$A7,$19,$E0,$E4,$4A,$FC,$FC,$EA,$B8,$4E,$67
	.dc.b	$B3,$E1,$71,$5C,$DE,$31,$97,$F5,$3F,$2D,$F8,$37,$F5,$DD,$7F,$C4
	.dc.b	$FC,$2F,$88,$D8,$DF,$07,$C3,$AE,$11,$FC,$9D,$BC,$DF,$CD,$5B,$1F
	.dc.b	$C0,$1A,$FF,$32,$6E,$E6,$92,$BF,$87,$4F,$FB,$8C,$FE,$FB,$F6,$F8
	.dc.b	$F0,$FF,$BE,$B7,$0F,$F0,$1B,$67,$FC,$07,$F3,$F7,$2F,$F0,$4B,$FE
	.dc.b	$D0,$8C,$9F,$D4,$FC,$64,$66,$77,$39,$F8,$3E,$55,$FE,$D6,$8F,$F0
	.dc.b	$C4,$77,$D3,$D5,$F6,$5A,$89,$3A,$3F,$C6,$DA,$A7,$31,$FD,$B2,$FA
	.dc.b	$36,$BA,$56,$E3,$8D,$7A,$6F,$BC,$7F,$BA,$9A,$FC,$83,$E7,$47,$AE
	.dc.b	$DF,$19,$EA,$24,$FF,$8C,$67,$DA,$6C,$8F,$C5,$79,$7D,$67,$BF,$BF
	.dc.b	$FD,$A5,$3F,$E5,$16,$BE,$B8,$F8,$E6,$FB,$E4,$7F,$8F,$E3,$2D,$21
	.dc.b	$F0,$FE,$DF,$E7,$9E,$92,$3F,$5D,$F6,$78,$5F,$50,$91,$FE,$4B,$FB
	.dc.b	$7A,$FC,$D8,$BC,$7F,$E4,$49,$FB,$1F,$66,$F9,$AA,$4B,$FF,$36,$7D
	.dc.b	$B3,$ED,$F1,$1F,$E6,$3E,$15,$F1,$8F,$F3,$93,$5D,$57,$17,$FE,$00
	.dc.b	$FB,$1F,$DF,$9B,$E5,$7E,$56,$31,$1F,$E6,$F1,$F4,$DF,$E8,$07,$FB
	.dc.b	$86,$9F,$A3,$8C,$47,$DF,$37,$EC,$7E,$75,$A7,$E7,$E3,$6F,$E1,$5A
	.dc.b	$FE,$B9,$AF,$F7,$D6,$BC,$8F,$F5,$FF,$DB,$FE,$99,$F1,$F6,$7F,$80
	.dc.b	$FA,$93,$99,$FF,$23,$B7,$05,$7F,$80,$BD,$D7,$17,$ED,$6B,$53,$CD
	.dc.b	$B6,$C3,$AF,$67,$EB,$F0,$E3,$3E,$23,$5F,$67,$FA,$77,$FC,$AE,$FE
	.dc.b	$81,$D5,$23,$11,$AE,$9F,$E0,$D1,$C7,$3F,$3C,$F4,$EA,$C9,$AA,$FE
	.dc.b	$19,$F0,$3F,$5A,$7F,$DB,$BB,$60,$FE,$24,$FD,$AF,$B7,$B0,$7E,$B3
	.dc.b	$63,$EB,$46,$08,$E0,$87,$DF,$8E,$0F,$FB,$CB,$CD,$F1,$3F,$E7,$65
	.dc.b	$F3,$FF,$C8,$5B,$1F,$09,$DE,$5E,$ED,$3F,$C3,$F7,$BB,$97,$B8,$FD
	.dc.b	$3F,$DB,$C8,$FC,$FB,$FC,$87,$ED,$DF,$57,$99,$FD,$37,$4A,$B1,$97
	.dc.b	$CC,$EA,$61,$C2,$FE,$D7,$C6,$3E,$DD,$F0,$FF,$CB,$73,$FC,$1E,$1D
	.dc.b	$B9,$BF,$9F,$5E,$EB,$E6,$FC,$9F,$CC,$26,$3F,$A1,$6B,$F2,$5A,$FE
	.dc.b	$8D,$FE,$59,$F2,$5F,$85,$8F,$C3,$F3,$BC,$4F,$C2,$9F,$CF,$E7,$53
	.dc.b	$FC,$0D,$BB,$7C,$BA,$E1,$3C,$AB,$84,$7F,$53,$FB,$77,$C0,$52,$57
	.dc.b	$E5,$A1,$F0,$38,$8F,$F5,$8C,$70,$75,$3C,$E7,$F2,$7F,$5B,$27,$F3
	.dc.b	$FD,$48,$FC,$9F,$DB,$3E,$DE,$9F,$C1,$DC,$57,$F6,$56,$BF,$C3,$5A
	.dc.b	$FF,$4D,$6B,$FD,$D9,$AF,$FA,$15,$A9,$CF,$53,$ED,$FD,$B3,$FB,$4F
	.dc.b	$7B,$FB,$0E,$FF,$93,$6F,$87,$FC,$36,$B7,$4A,$FD,$AF,$BC,$F9,$DF
	.dc.b	$E4,$0D,$9F,$EF,$CD,$7F,$93,$35,$FE,$CD,$8B,$E7,$7D,$F7,$FA,$36
	.dc.b	$E3,$FC,$75,$B4,$B7,$DB,$FF,$96,$2F,$F3,$9D,$7E,$66,$F0,$FE,$64
	.dc.b	$FF,$0A,$53,$ED,$F8,$65,$FE,$AB,$8C,$7C,$97,$A2,$7F,$CC,$79,$7F
	.dc.b	$88,$A7,$2F,$AA,$32,$F1,$6B,$E5,$6C,$70,$8F,$F0,$23,$5F,$F1,$6E
	.dc.b	$8F,$A5,$FF,$17,$65,$FC,$0D,$B8,$FF,$3D,$6D,$7F,$33,$FB,$77,$C7
	.dc.b	$FF,$8B,$5B,$1F,$EF,$D6,$BF,$39,$AF,$81,$AF,$F4,$96,$BD,$63,$FC
	.dc.b	$65,$18,$C0,$DF,$A7,$7D,$BF,$F8,$F5,$B7,$FC,$F9,$AF,$F7,$7F,$B7
	.dc.b	$7D,$1F,$F8,$DD,$B1,$FE,$29,$6B,$FE,$29,$F1,$11,$88,$FF,$29,$C6
	.dc.b	$27,$FC,$65,$39,$9F,$F6,$EB,$54,$ED,$ED,$3F,$E4,$DC,$5F,$25,$FE
	.dc.b	$55,$6C,$FF,$56,$6B,$F9,$FD,$7F,$B7,$5B,$ED,$FF,$CB,$8D,$9F,$F0
	.dc.b	$15,$78,$27,$FC,$B5,$39,$9F,$CF,$B5,$FC,$0B,$5E,$B1,$C0,$8F,$49
	.dc.b	$FC,$E5,$7F,$7F,$6C,$7F,$75,$6D,$3E,$B7,$F5,$93,$FD,$D6,$31,$CF
	.dc.b	$FE,$59,$9C,$CF,$CF,$6B,$F3,$9A,$8F,$96,$E1,$B1,$CC,$8F,$C2,$F9
	.dc.b	$C7,$D5,$7F,$A0,$5B,$1F,$E0,$26,$BF,$34,$FF,$07,$7D,$BB,$EA,$7F
	.dc.b	$D1,$0D,$8F,$9A,$D7,$F8,$4B,$5F,$D5,$B5,$FF,$12,$35,$EB,$1F,$E8
	.dc.b	$D8,$C6,$46,$D7,$F9,$0B,$97,$FB,$A3,$74,$FF,$17,$37,$0F,$C6,$FD
	.dc.b	$85,$D4,$FF,$92,$A3,$F2,$ED,$7E,$54,$6B,$DC,$FF,$45,$4E,$67,$FC
	.dc.b	$5D,$D1,$FE,$30,$CF,$F8,$F7,$9B,$FE,$A9,$8F,$F1,$99,$FE,$38,$E2
	.dc.b	$C4,$4A,$FF,$16,$CE,$BD,$AF,$CC,$EB,$FE,$0D,$F9,$FF,$EF,$D5,$F3
	.dc.b	$F1,$46,$27,$5D,$FF,$EA,$0F,$9A,$FF,$5C,$4E,$78,$3F,$C9,$CD,$C3
	.dc.b	$FC,$88,$DD,$FF,$D9,$55,$FE,$B3,$FB,$4F,$F0,$ED,$FC,$06,$DF,$ED
	.dc.b	$1B,$4F,$90,$E5,$18,$27,$AF,$FE,$BD,$8E,$C7,$AA,$BF,$81,$B5,$FF
	.dc.b	$7D,$65,$FE,$B5,$8C,$47,$7F,$2B,$C8,$47,$F4,$9F,$6E,$FA,$5C,$47
	.dc.b	$FA,$37,$1F,$EF,$0F,$EA,$5A,$FF,$A4,$67,$83,$FE,$61,$1F,$6E,$FB
	.dc.b	$4F,$EC,$D1,$FE,$85,$8C,$70,$7F,$B8,$5B,$19,$9D,$BF,$EC,$98,$7C
	.dc.b	$B7,$D6,$8E,$4D,$74,$D7,$FD,$EF,$1C,$38,$7F,$D2,$50,$F9,$4A,$47
	.dc.b	$FD,$2C,$FF,$B6,$72,$FD,$87,$0F,$BF,$B3,$FC,$E3,$9F,$F9,$1F,$F4
	.dc.b	$E7,$C8,$62,$BE,$E5,$5C,$7E,$E3,$79,$18,$F4,$9F,$F5,$C6,$1C,$AF
	.dc.b	$B1,$FC,$57,$EE,$31,$3F,$D0,$5A,$FF,$40,$6B,$FA,$A6,$BF,$CF,$BE
	.dc.b	$E5,$CF,$E1,$6D,$75,$CF,$13,$23,$F9,$24,$EA,$BF,$D9,$97,$E2,$D3
	.dc.b	$FC,$CD,$39,$7F,$CE,$C7,$F4,$8A,$FF,$4D,$B6,$BF,$DA,$9F,$A7,$FF
	.dc.b	$9B,$E7,$3D,$BF,$17,$AB,$FC,$50,$DB,$E0,$7F,$9C,$EB,$85,$FE,$D6
	.dc.b	$33,$7F,$CC,$61,$EF,$3E,$C7,$E8,$46,$1B,$0F,$7A,$C4,$AF,$CB,$4F
	.dc.b	$16,$CE,$6B,$F9,$97,$CC,$FF,$49,$EA,$FF,$58,$C7,$86,$31,$5F,$ED
	.dc.b	$F6,$C7,$94,$38,$7F,$C3,$BC,$BF,$E3,$D9,$FF,$60,$B6,$3B,$F9,$1F
	.dc.b	$B1,$9C,$F9,$1F,$8B,$B7,$EA,$1F,$D4,$67,$C0,$DB,$F0,$4E,$9F,$39
	.dc.b	$FD,$5D,$C4,$F0,$8C,$9E,$69,$E4,$FB,$5A,$F1,$6B,$FE,$E1,$9F,$AB
	.dc.b	$AF,$FA,$DD,$B5,$B9,$FB,$D1,$87,$FD,$DF,$FA,$8B,$ED,$FC,$34,$8C
	.dc.b	$3F,$26,$77,$3A,$9C,$C7,$CD,$9E,$13,$92,$3C,$4C,$9F,$E1,$2F,$B9
	.dc.b	$FF,$47,$3F,$EF,$06,$CF,$F6,$F8,$DB,$FD,$0E,$33,$87,$FB,$17,$17
	.dc.b	$8B,$F2,$3A,$FE,$9B,$63,$38,$7A,$67,$2D,$8C,$E1,$FE,$3B,$6B,$FE
	.dc.b	$3A,$6B,$FD,$91,$AF,$F9,$77,$DE,$FD,$47,$A6,$7F,$86,$4F,$4C,$7F
	.dc.b	$DD,$13,$9B,$FF,$0B,$41,$FC,$46,$8F,$1F,$E6,$12,$3E,$44,$65,$B1
	.dc.b	$9C,$7E,$84,$E5,$B1,$F8,$AF,$33,$B7,$9D,$FC,$85,$B7,$E2,$3C,$33
	.dc.b	$97,$FC,$46,$66,$92,$FB,$D7,$CA,$F5,$6D,$7E,$16,$75,$39,$24,$66
	.dc.b	$77,$5F,$B3,$AD,$CE,$6B,$ED,$7C,$6F,$FB,$72,$3F,$C8,$8D,$AF,$F1
	.dc.b	$95,$70,$7F,$C6,$9B,$A7,$5A,$7F,$04,$E3,$FF,$3D,$E1,$F9,$9B,$FE
	.dc.b	$03,$E9,$FE,$B9,$B1,$99,$DC,$7F,$09,$9D,$46,$23,$F7,$71,$83,$96
	.dc.b	$1F,$EA,$5A,$FE,$DC,$D9,$FF,$24,$1D,$8A,$E6,$FF,$8C,$AF,$A4,$FE
	.dc.b	$4E,$DA,$FA,$5D,$6F,$49,$F3,$3A,$BF,$EB,$5F,$2E,$F9,$1F,$D1,$37
	.dc.b	$0F,$D0,$37,$3F,$F4,$C3,$5F,$F4,$BF,$D9,$7F,$7A,$9F,$F3,$93,$63
	.dc.b	$FC,$C1,$97,$F8,$23,$87,$FB,$44,$E7,$5F,$EE,$3F,$8D,$79,$30,$FE
	.dc.b	$0D,$08,$C6,$1F,$CC,$63,$1A,$7F,$1E,$6D,$70,$AD,$D7,$F4,$B9,$CB
	.dc.b	$C6,$32,$FE,$A9,$AF,$F5,$0F,$39,$F7,$9F,$DF,$F3,$FF,$0F,$BF,$E0
	.dc.b	$1B,$E6,$7E,$C6,$B8,$47,$F6,$1F,$B8,$FF,$54,$63,$FA,$1F,$21,$18
	.dc.b	$AF,$C8,$FA,$0F,$9D,$E1,$EE,$4C,$38,$5E,$F2,$ED,$D3,$FE,$5A,$BD
	.dc.b	$63,$FE,$5D,$FB,$3F,$E3,$78,$7F,$A6,$73,$46,$27,$F6,$39,$BE,$3F
	.dc.b	$FB,$B4,$E6,$B7,$F1,$98,$3F,$7D,$5B,$C3,$D5,$5C,$FA,$DF,$71,$FC
	.dc.b	$4F,$2F,$ED,$6D,$D7,$FB,$5E,$CF,$A4,$FE,$BD,$39,$C7,$FA,$C3,$6F
	.dc.b	$FB,$5B,$5F,$8F,$3C,$3C,$5F,$F0,$67,$CE,$73,$BF,$98,$DA,$F0,$CE
	.dc.b	$AB,$FC,$24,$FE,$CB,$1F,$AD,$B7,$F2,$DE,$FF,$6F,$D1,$E9,$99,$7D
	.dc.b	$7B,$E0,$FF,$B0,$F8,$FF,$D1,$35,$AC,$BF,$D1,$9F,$6C,$FB,$1F,$F1
	.dc.b	$03,$6F,$F9,$05,$FF,$A4,$3E,$0F,$35,$27,$F0,$95,$BB,$FC,$1D,$65
	.dc.b	$F9,$7F,$51,$FD,$EB,$8B,$E7,$B7,$84,$D2,$61,$FE,$17,$8C,$B6,$33
	.dc.b	$5F,$E1,$9E,$A7,$DC,$7F,$83,$78,$7F,$35,$18,$9F,$AF,$5F,$A1,$AD
	.dc.b	$5F,$E3,$61,$E0,$FF,$8F,$DB,$1F,$03,$9F,$FC,$94,$EE,$74,$47,$5E
	.dc.b	$D5,$C3,$FD,$6D,$C4,$FB,$7F,$B2,$7E,$4B,$D9,$B7,$F5,$62,$D7,$EA
	.dc.b	$A6,$EB,$FB,$3A,$7E,$0F,$E4,$AF,$F0,$63,$5F,$F2,$74,$6B,$2F,$F0
	.dc.b	$BC,$E6,$7F,$C6,$4D,$7F,$B7,$36,$78,$1F,$E1,$37,$5C,$EE,$97,$CC
	.dc.b	$FE,$60,$FF,$44,$57,$0B,$FF,$6A,$4E,$A3,$F8,$37,$D2,$FF,$AA,$DB
	.dc.b	$F0,$9F,$B2,$E8,$FF,$36,$4E,$6F,$FD,$B8,$DD,$3F,$3A,$DB,$FC,$C7
	.dc.b	$0F,$E3,$67,$38,$7F,$9D,$7E,$A7,$F5,$11,$FA,$46,$D7,$E7,$E1,$F5
	.dc.b	$FB,$EE,$93,$FC,$6E,$B7,$97,$F7,$CB,$DC,$E6,$BF,$D4,$9F,$0B,$FD
	.dc.b	$46,$BF,$CD,$1E,$7A,$31,$1F,$E7,$F8,$C7,$9D,$FE,$31,$9C,$C8,$FF
	.dc.b	$41,$F6,$BE,$B7,$F9,$A3,$71,$FF,$07,$61,$C9,$FE,$BF,$19,$AF,$EF
	.dc.b	$BF,$90,$3E,$B3,$FD,$70,$DA,$DE,$18,$99,$1D,$27,$55,$FD,$6A,$FA
	.dc.b	$69,$FD,$07,$15,$6A,$31,$97,$F2,$9B,$F8,$2D,$7E,$03,$5F,$75,$7C
	.dc.b	$5F,$23,$FC,$F3,$DE,$C7,$FD,$11,$5E,$1F,$80,$F9,$B5,$F0,$9F,$F6
	.dc.b	$97,$E6,$1F,$EC,$6C,$0D,$27,$C3,$FF,$A1,$3A,$9C,$7F,$57,$5F,$C1
	.dc.b	$71,$72,$AF,$C3,$33,$83,$F0,$0E,$5A,$4D,$15,$CA,$3E,$9E,$3F,$EC
	.dc.b	$D7,$FA,$A6,$1D,$DF,$7F,$D5,$FF,$61,$7F,$CD,$8D,$BF,$C3,$B5,$FC
	.dc.b	$33,$5E,$7D,$7F,$9B,$3F,$DE,$51,$C7,$0F,$CE,$47,$FB,$23,$ED,$3F
	.dc.b	$9D,$D3,$57,$E8,$0F,$B0,$FE,$F5,$F0,$DF,$CB,$2B,$F2,$17,$C2,$F7
	.dc.b	$5A,$F1,$BF,$DA,$D3,$9C,$39,$69,$F9,$7C,$79,$62,$AD,$61,$C6,$3C
	.dc.b	$57,$C1,$57,$87,$1E,$F8,$73,$C7,$BF,$9B,$FE,$0F,$D3,$F2,$18,$F1
	.dc.b	$C3,$BE,$6F,$98,$ED,$93,$EC,$BF,$84,$56,$A7,$31,$8C,$BD,$71,$8E
	.dc.b	$FF,$36,$BE,$98,$C6,$BE,$38,$F8,$B1,$87,$D3,$5C,$A7,$58,$F8,$6F
	.dc.b	$9E,$5E,$59,$CE,$3E,$1C,$39,$C6,$2B,$E3,$77,$BF,$12,$FD,$0E,$9F
	.dc.b	$17,$F5,$0F,$99,$7E,$3A,$D7,$4F,$C0,$9C,$E3,$E4,$3F,$8D,$E1,$CB
	.dc.b	$5F,$D7,$CE,$B1,$FC,$06,$1C,$B1,$F4,$CE,$6B,$F8,$94,$E7,$0E,$9A
	.dc.b	$AB,$51,$DB,$2F,$16,$BF,$2B,$EB,$FF,$CE,$9E,$7F,$C9,$A4,$61,$F3
	.dc.b	$7C,$C3,$E9,$7F,$D9,$7B,$7E,$2A,$B4,$DE,$B6,$1C,$E7,$3B,$7F,$27
	.dc.b	$C3,$97,$C4,$F8,$6F,$79,$7C,$33,$C2,$E1,$FE,$5C,$7F,$2F,$BF,$23
	.dc.b	$83,$78,$7F,$A8,$63,$31,$C1,$FF,$1F,$6F,$9F,$1B,$FA,$4C,$FF,$BF
	.dc.b	$E7,$2D,$8C,$BD,$63,$39,$FB,$E7,$8B,$6B,$F2,$19,$3E,$D5,$F4,$5E
	.dc.b	$8A,$D6,$3E,$DC,$79,$EB,$E5,$CB,$9B,$F1,$70,$E3,$5F,$89,$FD,$22
	.dc.b	$BF,$73,$8B,$E0,$63,$F8,$F8,$ED,$7F,$82,$AD,$CF,$E2,$1B,$1F,$8C
	.dc.b	$9D,$9F,$C3,$EB,$9C,$71,$CB,$3A,$2F,$8D,$EC,$E3,$C7,$F8,$D7,$71
	.dc.b	$A7,$F1,$3E,$1E,$BA,$1F,$77,$4F,$0C,$E8,$FC,$34,$E5,$FD,$8C,$6B
	.dc.b	$7F,$F4,$24,$68,$C9,$0F,$99,$3B,$39,$3F,$A2,$AF,$0B,$C4,$FD,$2F
	.dc.b	$5B,$EC,$FD,$6D,$8F,$E0,$AD,$7E,$C7,$8D,$FE,$B7,$EA,$FC,$63,$63
	.dc.b	$F9,$3D,$72,$3D,$11,$98,$FD,$5B,$BD,$3F,$9E,$CE,$E7,$F4,$D3,$CF
	.dc.b	$C6,$FE,$C3,$F4,$97,$C5,$34,$E5,$E7,$FF,$97,$AF,$FC,$1A,$D7,$FC
	.dc.b	$36,$D7,$FC,$34,$D7,$FB,$DB,$55,$7F,$70,$CB,$C0,$48,$FF,$0C,$FE
	.dc.b	$65,$FE,$CB,$7F,$B5,$C7,$F6,$83,$4E,$0E,$78,$6F,$C3,$FF,$09,$55
	.dc.b	$AE,$15,$FE,$37,$78,$CE,$9F,$8D,$3D,$E7,$FC,$64,$D7,$FC,$26,$D7
	.dc.b	$FB,$43,$7C,$A7,$D1,$BE,$CB,$EB,$46,$1B,$19,$AF,$F1,$AF,$E1,$AF
	.dc.b	$B1,$FA,$CD,$8F,$E1,$C7,$F7,$57,$0F,$F9,$21,$AF,$F9,$1D,$AF,$F8
	.dc.b	$A3,$F2,$37,$E3,$3F,$81,$6C,$7F,$91,$1A,$FC,$9C,$7D,$3E,$3E,$EA
	.dc.b	$4E,$DF,$4C,$48,$FF,$29,$6B,$FE,$A6,$C9,$39,$9F,$A9,$C8,$CB,$BD
	.dc.b	$FF,$91,$5F,$0C,$FA,$70,$F7,$BE,$F6,$D7,$17,$3C,$1F,$EB,$EF,$2B
	.dc.b	$F3,$1D,$DF,$DF,$6F,$C7,$C3,$FE,$B3,$9F,$D1,$7E,$1D,$FE,$90,$9C
	.dc.b	$9F,$90,$AE,$0F,$E2,$6F,$A3,$A6,$BF,$8D,$13,$CD,$FF,$47,$FC,$87
	.dc.b	$EA,$7D,$83,$4F,$D9,$7C,$9B,$D2,$C5,$7F,$A7,$26,$BD,$67,$4F,$F9
	.dc.b	$E3,$ED,$DF,$82,$7E,$E5,$B1,$98,$EB,$E9,$48,$F9,$34,$8F,$97,$08
	.dc.b	$C3,$A8,$CC,$62,$3F,$1E,$D7,$FA,$F3,$5F,$EC,$6D,$7F,$73,$F0,$3D
	.dc.b	$5F,$90,$D7,$FD,$B3,$1F,$0B,$E5,$3F,$87,$4E,$79,$1C,$3F,$E9,$BD
	.dc.b	$D3,$98,$C3,$70,$FF,$50,$60,$C3,$E8,$7D,$2C,$9F,$F6,$3B,$F4,$3E
	.dc.b	$3B,$11,$FE,$C4,$8B,$97,$F0,$BB,$57,$ED,$BE,$DF,$F5,$B3,$FF,$00
	.dc.b	$1B,$97,$FA,$BF,$A9,$18,$C7,$C7,$F4,$1F,$78,$F9,$64,$7E,$9B,$5F
	.dc.b	$9A,$D7,$E6,$BF,$9C,$F2,$1F,$51,$FC,$CD,$B3,$F9,$36,$BF,$93,$FB
	.dc.b	$4F,$D5,$E3,$F2,$67,$39,$7C,$CA,$FE,$5B,$18,$9D,$5F,$F4,$5B,$E3
	.dc.b	$5F,$6A,$3E,$A9,$1F,$D3,$88,$7C,$8E,$63,$F5,$1C,$1D,$5E,$47,$09
	.dc.b	$DB,$FA,$BF,$C3,$9F,$75,$FD,$3D,$B1,$98,$F2,$7D,$8C,$8E,$4D,$7F
	.dc.b	$65,$F8,$D3,$DA,$C6,$DF,$D1,$71,$5E,$35,$FF,$1C,$DE,$F0,$FD,$E7
	.dc.b	$C0,$BE,$5B,$FB,$03,$63,$31,$F1,$7E,$07,$2C,$8C,$9E,$D8,$FE,$D1
	.dc.b	$F7,$2F,$90,$A4,$9F,$85,$DF,$5F,$E1,$D3,$B7,$F8,$73,$5F,$B9,$1F
	.dc.b	$DF,$75,$7D,$A7,$F8,$C9,$FF,$4F,$37,$4F,$E6,$2D,$A1,$F5,$F8,$BE
	.dc.b	$F1,$FC,$0B,$7F,$D6,$C7,$F2,$E6,$BF,$D3,$9A,$FF,$97,$5A,$FF,$98
	.dc.b	$9A,$F6,$6B,$DC,$FE,$AF,$E3,$BE,$C7,$FD,$95,$1F,$CC,$23,$1F,$0D
	.dc.b	$FE,$13,$9C,$CF,$F5,$1E,$AF,$C9,$D7,$F6,$9F,$BF,$FE,$83,$E8,$1A
	.dc.b	$4A,$FE,$8F,$FA,$23,$E0,$BF,$C4,$0D,$8F,$CC,$B5,$FE,$BC,$BD,$5D
	.dc.b	$FE,$01,$27,$6F,$2F,$62,$47,$67,$F9,$5D,$1A,$48,$D1,$0F,$EF,$31
	.dc.b	$B7,$E2,$FE,$31,$FE,$27,$E3,$FD,$B6,$88,$C7,$1F,$DA,$8C,$7C,$A7
	.dc.b	$F8,$FF,$E6,$9F,$04,$FB,$DC,$37,$2F,$F0,$3C,$EA,$7F,$B5,$B6,$33
	.dc.b	$3B,$7F,$D3,$51,$96,$C6,$5F,$F5,$0C,$65,$B1,$97,$F1,$F1,$FC,$EA
	.dc.b	$3C,$D8,$7F,$A6,$B2,$FE,$25,$39,$AF,$E1,$7F,$A9,$3E,$53,$FC,$AC
	.dc.b	$D8,$FF,$00,$B5,$FF,$10,$B5,$FF,$58,$35,$FF,$5A,$B5,$FD,$0B,$53
	.dc.b	$4D,$79,$CF,$C4,$3F,$CD,$D8,$7F,$27,$8C,$7C,$0F,$F9,$8E,$73,$97
	.dc.b	$D3,$FA,$CF,$F0,$05,$FD,$87,$FC,$05,$8F,$FB,$1B,$74,$62,$38,$78
	.dc.b	$7F,$E1,$A8,$C4,$27,$73,$92,$47,$EA,$A7,$79,$7F,$99,$B7,$FF,$3B
	.dc.b	$3F,$67,$0F,$43,$71,$FF,$6B,$FC,$E7,$F7,$B9,$FF,$34,$3F,$DD,$F0
	.dc.b	$34,$9E,$9F,$F8,$19,$B1,$FE,$32,$6B,$FE,$DF,$CD,$ED,$7F,$17,$BF
	.dc.b	$CC,$B7,$0F,$CC,$36,$BF,$DE,$7F,$01,$FE,$71,$6C,$FF,$8F,$BE,$3B
	.dc.b	$FC,$C0,$D9,$FF,$36,$46,$27,$FC,$C2,$D7,$FD,$E0,$D7,$FD,$EA,$D7
	.dc.b	$F9,$0B,$5F,$F4,$6F,$1F,$F0,$97,$FC,$78,$D7,$FC,$80,$DC,$3E,$3F
	.dc.b	$A9,$FE,$A9,$CF,$FC,$35,$5F,$EF,$F6,$BF,$3A,$3E,$41,$FE,$2E,$EF
	.dc.b	$FE,$57,$5F,$E9,$76,$E5,$FC,$D2,$BF,$C1,$3E,$1F,$FA,$D6,$5E,$E5
	.dc.b	$89,$3F,$DC,$27,$8B,$FC,$F1,$DB,$FE,$55,$F9,$8F,$F1,$13,$7D,$8F
	.dc.b	$F5,$D4,$62,$3F,$CE,$3F,$45,$FE,$48,$7F,$CB,$D1,$9C,$BF,$40,$DC
	.dc.b	$BF,$C6,$4D,$9F,$CB,$57,$1A,$FE,$2F,$F6,$6F,$92,$FF,$66,$36,$BF
	.dc.b	$9F,$35,$FF,$40,$FD,$BF,$1C,$7F,$8E,$4E,$6F,$9D,$F2,$3F,$DA,$D5
	.dc.b	$BB,$FD,$33,$E9,$7F,$5C,$D7,$F9,$33,$5D,$FB,$BF,$E8,$8F,$B7,$FF
	.dc.b	$1E,$C7,$F5,$16,$BF,$E9,$46,$DF,$F3,$2F,$D4,$FF,$B5,$73,$A7,$39
	.dc.b	$FF,$33,$6D,$FE,$E5,$B2,$AF,$EE,$52,$FA,$77,$A6,$C7,$FD,$3B,$18
	.dc.b	$6C,$66,$7F,$D7,$F3,$96,$FC,$03,$83,$8B,$63,$31,$83,$FC,$FE,$BE
	.dc.b	$52,$91,$FF,$41,$BF,$DD,$BE,$E1,$F3,$9F,$ED,$DD,$11,$8D,$B5,$39
	.dc.b	$8C,$5F,$F0,$36,$BF,$C8,$5A,$FF,$59,$6B,$FC,$7D,$BF,$04,$FC,$0B
	.dc.b	$FC,$97,$7F,$ED,$E8,$C7,$C0,$7F,$BB,$DB,$19,$9D,$DF,$F1,$3F,$B9
	.dc.b	$EB,$91,$A4,$CB,$D8,$DB,$FF,$0A,$B6,$BF,$8E,$7D,$CF,$FB,$DE,$7F
	.dc.b	$C3,$2D,$8C,$FA,$BF,$16,$31,$1F,$CF,$DA,$FF,$77,$6B,$FC,$F5,$B8
	.dc.b	$1C,$C3,$27,$C2,$FF,$9D,$E3,$F8,$31,$97,$0F,$F0,$98,$DE,$1F,$EE
	.dc.b	$66,$C7,$C9,$9D,$BF,$E9,$76,$BC,$E7,$84,$62,$5F,$82,$D2,$4F,$57
	.dc.b	$9F,$86,$FB,$9F,$F7,$EC,$FF,$34,$6E,$1F,$CD,$1A,$FF,$34,$6B,$E9
	.dc.b	$FA,$2F,$01,$27,$D4,$D8,$FC,$13,$5F,$87,$82,$3C,$7E,$5F,$F9,$7A
	.dc.b	$BF,$A2,$B6,$BF,$A2,$B5,$FE,$8A,$D7,$F9,$97,$E2,$2F,$B9,$F8,$AD
	.dc.b	$AF,$A3,$3F,$E6,$A8,$C3,$86,$BF,$59,$AF,$49,$E6,$49,$E6,$3E,$DF
	.dc.b	$73,$FE,$E0,$6C,$66,$7E,$CB,$4F,$1C,$A9,$F1,$DE,$82,$4E,$A3,$2D
	.dc.b	$8C,$CF,$A5,$AF,$F8,$5F,$0F,$E7,$53,$FE,$0F,$F0,$7F,$81,$46,$27
	.dc.b	$FC,$00,$D7,$FC,$8F,$B7,$E3,$3C,$AF,$C4,$DF,$F9,$3A,$30,$47,$04
	.dc.b	$9F,$F2,$AC,$7B,$B3,$79,$34,$95,$F2,$23,$2F,$C2,$EA,$FF,$5C,$37
	.dc.b	$E1,$7F,$3B,$18,$C3,$EC,$B5,$FD,$5B,$5F,$EE,$4D,$7F,$D4,$ED,$7F
	.dc.b	$94,$47,$EA,$B1,$7D,$B3,$ED,$70,$DC,$3F,$CD,$ED,$BF,$F3,$5B,$5F
	.dc.b	$F0,$BF,$B3,$FE,$1C,$24,$FF,$88,$5B,$1F,$E1,$E6,$BC,$23,$FC,$4D
	.dc.b	$E8,$FF,$89,$08,$E0,$FE,$51,$F0,$AF,$B4,$FE,$50,$D8,$CC,$60,$FE
	.dc.b	$07,$D7,$FE,$E9,$8F,$C6,$B6,$FF,$D7,$3F,$17,$A3,$F9,$97,$8B,$EB
	.dc.b	$8F,$F6,$6F,$4F,$F5,$DD,$72,$4C,$3F,$2F,$F9,$67,$ED,$DB,$D5,$FC
	.dc.b	$4A,$31,$7F,$AE,$6B,$FC,$81,$AF,$F8,$9D,$AF,$FB,$71,$A9,$CF,$90
	.dc.b	$FB,$5F,$81,$3F,$CA,$A3,$1A,$FF,$24,$6E,$1F,$BE,$6C,$FF,$A8,$BC
	.dc.b	$7F,$C2,$71,$FF,$6F,$E4,$FC,$14,$60,$8E,$09,$7F,$EB,$08,$D7,$D0
	.dc.b	$70,$24,$65,$B7,$DD,$AF,$FA,$BA,$51,$FC,$E3,$ED,$3F,$DB,$95,$FC
	.dc.b	$09,$B5,$FC,$59,$AF,$F4,$86,$BF,$E4,$F6,$A0,$E1,$BF,$0D,$DC,$FE
	.dc.b	$7F,$F6,$4F,$5A,$92,$FF,$CE,$71,$97,$E0,$7D,$A7,$C6,$23,$82,$5F
	.dc.b	$CD,$8C,$B6,$3E,$47,$C3,$7C,$E2,$38,$24,$7D,$58,$CB,$63,$E2,$4F
	.dc.b	$C7,$23,$82,$47,$C7,$8C,$B6,$3D,$7E,$CF,$C0,$23,$83,$EE,$7C,$7F
	.dc.b	$E0,$88,$E0,$93,$F8,$98,$CB,$63,$F0,$9F,$65,$F8,$A2,$38,$24,$7E
	.dc.b	$46,$32,$D8,$FC,$07,$17,$E0,$C8,$E0,$91,$F8,$38,$CB,$63,$FA,$8F
	.dc.b	$D9,$3B,$BF,$A6,$B6,$33,$18,$3F,$3F,$F1,$72,$3F,$38,$DA,$FD,$33
	.dc.b	$5F,$CC,$35,$FC,$77,$27,$F9,$7F,$E2,$3F,$95,$73,$1A,$4B,$FE,$DF
	.dc.b	$3B,$6C,$ED,$B3,$BA,$FE,$EB,$F6,$5F,$C4,$DF,$CE,$35,$34,$D7,$7F
	.dc.b	$F8,$27,$BF,$FC,$BD,$D9,$FC,$7E,$BF,$DF,$0D,$7F,$4D,$3B,$C3,$FC
	.dc.b	$A9,$C6,$F8,$9F,$F2,$A3,$63,$31,$82,$38,$25,$7F,$86,$A7,$83,$FE
	.dc.b	$1C,$FC,$75,$F8,$17,$F9,$7D,$B5,$FE,$63,$6C,$EE,$7D,$3D,$72,$BF
	.dc.b	$C2,$F8,$FF,$89,$63,$15,$FE,$1A,$FB,$07,$CD,$7F,$90,$9B,$7F,$E0
	.dc.b	$C6,$CE,$E7,$FD,$05,$F6,$5F,$E6,$C2,$7B,$18,$25,$F8,$23,$2D,$8C
	.dc.b	$D7,$F9,$6B,$15,$6A,$30,$FF,$A1,$38,$5F,$86,$7F,$86,$27,$FC,$C3
	.dc.b	$18,$F4,$7F,$CB,$6D,$C7,$FC,$6A,$D9,$ED,$3F,$DE,$F0,$FF,$12,$F8
	.dc.b	$7F,$E9,$B8,$C1,$FD,$F7,$47,$95,$F0,$C9,$8F,$C3,$6E,$1F,$25,$AF
	.dc.b	$F9,$D2,$78,$CE,$CF,$E5,$9C,$C7,$F1,$3F,$54,$F9,$6C,$5F,$FA,$CF
	.dc.b	$C0,$FE,$64,$B0,$E6,$FF,$B1,$6F,$8D,$70,$F3,$BF,$DF,$F3,$BF,$8B
	.dc.b	$E7,$5E,$36,$C7,$4B,$46,$23,$FA,$04,$62,$7F,$CC,$4D,$8F,$F4,$6B
	.dc.b	$63,$A1,$A9,$E5,$3E,$18,$FF,$13,$E1,$FA,$6F,$AB,$FF,$6E,$C6,$27
	.dc.b	$BD,$7A,$6B,$5F,$05,$67,$35,$BE,$06,$1C,$A7,$E3,$1A,$2C,$6A,$B9
	.dc.b	$70,$BE,$CF,$FD,$23,$87,$CC,$8C,$7D,$3F,$C1,$6C,$7F,$A0,$DB,$87
	.dc.b	$FA,$BD,$B1,$C6,$B9,$CF,$F6,$79,$CB,$63,$2F,$F1,$B7,$FD,$43,$AB
	.dc.b	$3F,$E1,$3F,$61,$F7,$70,$F8,$98,$78,$2B,$58,$33,$5E,$E3,$E0,$8F
	.dc.b	$A2,$78,$0F,$7B,$E7,$4E,$63,$0F,$C8,$FB,$27,$D9,$FF,$9E,$72,$FC
	.dc.b	$4E,$1C,$B0,$E2,$DB,$EE,$D9,$F2,$DF,$2F,$77,$EC,$57,$7B,$E3,$8F
	.dc.b	$82,$3F,$DA,$F8,$73,$6C,$65,$FC,$FC,$65,$BE,$87,$F4,$37,$FD,$5F
	.dc.b	$08,$C4,$74,$6B,$C3,$5F,$F5,$1C,$ED,$F5,$CF,$3C,$BD,$ED,$DF,$FC
	.dc.b	$41,$9B,$EC,$DF,$67,$FE,$82,$9E,$99,$7C,$2B,$E0,$FF,$89,$F2,$EC
	.dc.b	$FC,$DB,$E2,$FD,$CF,$B3,$AF,$E5,$B8,$3F,$D1,$F8,$7E,$0A,$3B,$1F
	.dc.b	$D8,$67,$79,$FE,$0B,$44,$72,$E2,$FC,$D3,$6B,$F2,$4D,$79,$69,$FC
	.dc.b	$2A,$76,$D8,$DB,$5D,$9F,$A1,$9E,$4F,$E3,$7C,$3F,$2F,$1F,$EC,$BE
	.dc.b	$CE,$F4,$1F,$4D,$6C,$FC,$4D,$6A,$24,$60,$87,$F2,$03,$E5,$47,$8D
	.dc.b	$E4,$64,$B8,$FE,$2A,$35,$BF,$F9,$8A,$7F,$DF,$0D,$78,$36,$7F,$23
	.dc.b	$E3,$FF,$2C,$EE,$FE,$5F,$93,$B9,$F3,$7E,$E6,$E3,$FA,$2B,$E1,$87
	.dc.b	$E5,$9F,$F7,$7C,$7E,$C5,$B8,$9A,$48,$FC,$9B,$5F,$CD,$BF,$EA,$CF
	.dc.b	$37,$FC,$61,$1F,$DD,$9A,$FF,$65,$6B,$FD,$13,$14,$62,$3F,$89,$7E
	.dc.b	$77,$FE,$E6,$7F,$B0,$B5,$FE,$CE,$D7,$F9,$13,$5F,$F0,$DF,$D9,$7F
	.dc.b	$B0,$09,$8F,$FB,$01,$B1,$FE,$0A,$6B,$FD,$E5,$AF,$95,$AF,$48,$F8
	.dc.b	$FF,$71,$FB,$68,$FE,$BA,$D7,$FD,$20,$D4,$E7,$07,$D8,$BE,$6F,$FD
	.dc.b	$A6,$DA,$FF,$0E,$B5,$FF,$06,$35,$FE,$D7,$F0,$0F,$C0,$9E,$D7,$F6
	.dc.b	$E9,$CC,$61,$C0,$9E,$13,$88,$C4,$ED,$E5,$2F,$B3,$78,$BB,$8C,$36
	.dc.b	$33,$3D,$A7,$2D,$8C,$BD,$1A,$FA,$9A,$FF,$BB,$9A,$FF,$47,$C1,$F6
	.dc.b	$7F,$18,$91,$F1,$9F,$A7,$F3,$72,$3D,$9A,$A3,$11,$F0,$BE,$CD,$F4
	.dc.b	$5F,$A0,$6C,$7C,$36,$BB,$8D,$11,$D3,$5F,$9B,$F1,$6F,$43,$F5,$6D
	.dc.b	$8F,$9C,$D7,$71,$A2,$3A,$6B,$D3,$D0,$E6,$47,$9B,$FB,$8F,$C3,$BF
	.dc.b	$51,$1F,$ED,$F6,$A6,$93,$DD,$E0,$48,$FC,$14,$70,$B7,$CF,$FA,$89
	.dc.b	$3E,$A6,$BF,$01,$AF,$E4,$63,$8B,$F0,$3E,$CB,$E1,$91,$C1,$0F,$92
	.dc.b	$E5,$A7,$C8,$F9,$4F,$96,$47,$07,$A3,$D9,$F8,$44,$70,$43,$F4,$38
	.dc.b	$23,$51,$87,$F9,$0F,$D6,$7E,$3A,$3F,$48,$D7,$F5,$2D,$7F,$5C,$D7
	.dc.b	$FC,$BA,$D4,$1C,$37,$87,$B9,$E0,$F9,$A9,$1E,$06,$BE,$06,$BE,$06
	.dc.b	$BE,$0F,$B0,$F4,$91,$F4,$BF,$CF,$3E,$C3,$FC,$57,$1F,$BF,$6B,$FC
	.dc.b	$11,$AF,$F0,$96,$BF,$E7,$86,$BE,$C6,$A6,$98,$B7,$AF,$F3,$09,$3F
	.dc.b	$31,$AF,$CC,$6B,$F3,$1F,$A7,$F2,$32,$3E,$9B,$5F,$A6,$D7,$E9,$B5
	.dc.b	$FA,$7F,$17,$F7,$08,$FD,$C7,$E2,$78,$F2,$3E,$23,$5F,$88,$D7,$E2
	.dc.b	$35,$FB,$BD,$6F,$A4,$FC,$03,$75,$FE,$4D,$18,$3F,$03,$F5,$7F,$E3
	.dc.b	$FA,$FE,$60,$D7,$F9,$AB,$5F,$E7,$4D,$7B,$BF,$D9,$9F,$E5,$CD,$7F
	.dc.b	$9B,$B5,$FE,$8A,$D7,$FD,$62,$FF,$7B,$7C,$3E,$3F,$F8,$02,$31,$1E
	.dc.b	$28,$C7,$C5,$48,$F3,$B5,$F3,$B5,$F3,$B5,$FD,$0F,$C4,$BD,$27,$D4
	.dc.b	$7F,$4F,$8C,$36,$33,$3F,$D5,$A7,$2F,$F5,$5D,$BF,$C5,$0F,$FB,$29
	.dc.b	$AF,$EB,$9A,$9A,$4F,$23,$FC,$3A,$D8,$FE,$C1,$F5,$4F,$89,$FF,$10
	.dc.b	$B6,$33,$18,$38,$FC,$74,$8F,$23,$5F,$ED,$B4,$8C,$47,$F7,$38,$C3
	.dc.b	$C0,$FE,$E7,$E5,$BE,$67,$FA,$7C,$7F,$88,$DA,$FF,$0A,$8D,$78,$7F
	.dc.b	$DD,$DB,$1F,$DE,$7F,$4C,$7D,$F7,$F8,$E9,$B1,$98,$C1,$1C,$10,$FF
	.dc.b	$02,$C7,$07,$FC,$0F,$F5,$4F,$17,$FC,$8A,$D8,$FF,$04,$35,$DC,$7F
	.dc.b	$92,$BE,$0D,$F5,$F4,$93,$FE,$0F,$8C,$B6,$32,$FF,$85,$A3,$2D,$8F
	.dc.b	$F0,$B7,$C2,$7F,$89,$F8,$6F,$A5,$F8,$28,$C5,$7F,$87,$9A,$FF,$88
	.dc.b	$3F,$15,$FF,$30,$B7,$B7,$FC,$BF,$18,$8F,$F1,$43,$5F,$E7,$21,$C3
	.dc.b	$FC,$F1,$AE,$63,$99,$F1,$BC,$89,$1F,$19,$AF,$8F,$8F,$C2,$47,$FC
	.dc.b	$65,$5F,$86,$7F,$C6,$FD,$9F,$EE,$86,$FC,$07,$F9,$DA,$31,$1F,$E3
	.dc.b	$F6,$BE,$86,$BF,$D4,$9A,$FF,$17,$8E,$47,$E2,$FC,$E7,$D5,$52,$4F
	.dc.b	$E5,$A3,$2D,$8C,$BF,$E4,$FE,$4F,$F1,$74,$62,$3F,$CF,$9F,$57,$C1
	.dc.b	$D3,$5C,$C7,$E6,$E3,$11,$F9,$F6,$A9,$46,$1B,$3C,$4F,$F2,$A7,$E3
	.dc.b	$0F,$BB,$A4,$BF,$F2,$A4,$65,$FD,$8F,$D6,$7F,$B6,$9F,$F4,$93,$5F
	.dc.b	$D8,$35,$F1,$36,$CD,$27,$C2,$7E,$D5,$E1,$2F,$A0,$FE,$C7,$F0,$3F
	.dc.b	$E6,$F6,$C7,$F9,$D1,$AF,$F9,$E5,$AF,$E4,$97,$D6,$3E,$7B,$93,$67
	.dc.b	$93,$5E,$4D,$7F,$D0,$3F,$11,$FE,$BE,$4D,$23,$F6,$FE,$B3,$FC,$4D
	.dc.b	$D3,$FC,$59,$B1,$9F,$4F,$F9,$9E,$A9,$CC,$FF,$AE,$DA,$FF,$19,$6B
	.dc.b	$F6,$DB,$1C,$67,$F9,$EF,$9A,$FA,$C7,$A9,$86,$E3,$FE,$8D,$6E,$3F
	.dc.b	$E9,$66,$BF,$EC,$BF,$AC,$EA,$4C,$3C,$ED,$AF,$E6,$2D,$9F,$CD,$3E
	.dc.b	$4F,$AB,$7D,$67,$B9,$B5,$FC,$E1,$B5,$FA,$16,$BF,$ED,$B6,$BF,$CE
	.dc.b	$5A,$FE,$6D,$B9,$9C,$C7,$FA,$D7,$D9,$FE,$BB,$A1,$A4,$BF,$F5,$93
	.dc.b	$67,$FD,$60,$D7,$FD,$73,$F5,$3C,$89,$3F,$EB,$97,$FB,$5F,$82,$FA
	.dc.b	$CF,$F5,$EC,$FF,$29,$6D,$7F,$B2,$E7,$5F,$09,$C2,$73,$3F,$EC,$6F
	.dc.b	$A0,$F3,$92,$FF,$D8,$CD,$8F,$F6,$23,$5E,$11,$FC,$DF,$CD,$FF,$33
	.dc.b	$B6,$3C,$B1,$87,$C4,$D7,$FD,$BD,$F5,$6F,$AC,$FF,$01,$B6,$BF,$DB
	.dc.b	$8D,$9F,$F7,$0B,$5E,$33,$C0,$93,$C1,$FF,$07,$AF,$AC,$E1,$3E,$66
	.dc.b	$A8,$D4,$F0,$9C,$CA,$BF,$D8,$95,$E4,$F8,$4F,$39,$1C,$12,$7C,$F1
	.dc.b	$97,$F9,$76,$0F,$33,$F6,$2D,$AF,$F0,$5B,$67,$73,$FE,$26,$F5,$5B
	.dc.b	$D2,$5F,$F8,$CD,$B5,$FE,$31,$6B,$FD,$88,$4F,$29,$C4,$7C,$1C,$9E
	.dc.b	$57,$C1,$24,$E8,$F9,$4F,$F8,$E3,$AD,$F3,$D4,$93,$D2,$FF,$8A,$C6
	.dc.b	$1F,$E3,$2D,$7C,$93,$CA,$7F,$C9,$1F,$58,$FA,$CF,$F2,$3E,$48,$C4
	.dc.b	$EA,$73,$18,$BF,$F0,$47,$C3,$FF,$82,$C9,$5F,$E1,$06,$C6,$A3,$04
	.dc.b	$7F,$C2,$1E,$97,$F8,$4C,$91,$FE,$18,$7F,$8E,$FA,$EF,$72,$92,$7E
	.dc.b	$5F,$57,$F9,$66,$3F,$CB,$BD,$8F,$9B,$A4,$AF,$F2,$13,$FC,$A3,$C6
	.dc.b	$F5,$65,$FE,$18,$C3,$FA,$86,$B9,$25,$FC,$DF,$AC,$75,$7F,$9A,$9B
	.dc.b	$19,$8C,$11,$C1,$2F,$F0,$33,$C1,$FC,$37,$90,$FA,$CF,$75,$7E,$1A
	.dc.b	$31,$AF,$F9,$E9,$B1,$99,$DD,$7E,$3F,$8B,$FA,$F7,$27,$DA,$F2,$7E
	.dc.b	$F4,$E5,$AF,$F8,$DA,$77,$7F,$E8,$99,$CC,$FF,$A4,$9A,$FE,$69,$B3
	.dc.b	$C2,$3A,$7C,$B7,$72,$57,$36,$C6,$A3,$04,$7F,$47,$F5,$8F,$9D,$FF
	.dc.b	$4A,$B6,$7F,$38,$D9,$DC,$E8,$93,$A6,$C7,$F3,$1F,$AC,$7C,$3F,$F3
	.dc.b	$86,$D7,$F9,$7D,$B3,$B9,$F9,$1F,$29,$2B,$E2,$36,$7E,$53,$5F,$95
	.dc.b	$9A,$31,$1F,$27,$E1,$BE,$91,$2B,$E7,$36,$3E,$A3,$5E,$6F,$C2,$99
	.dc.b	$3F,$09,$B3,$F2,$1A,$FF,$0F,$CB,$C5,$3B,$3D,$DE,$C7,$C1,$23,$83
	.dc.b	$F0,$1E,$DF,$E1,$C8,$E0,$95,$F8,$98,$CB,$63,$F1,$1F,$59,$F8,$F2
	.dc.b	$38,$24,$7E,$46,$32,$D8,$FB,$BC,$7F,$80,$23,$82,$47,$E1,$A3,$2D
	.dc.b	$8F,$E4,$DF,$58,$EE,$FE,$A2,$D8,$CC,$60,$FD,$27,$CC,$C8,$FD,$03
	.dc.b	$6B,$F4,$CD,$7E,$EB,$5F,$F7,$AF,$37,$FB,$D6,$7F,$DE,$0D,$8F,$D9
	.dc.b	$35,$FD,$73,$5E,$3F,$11,$FC,$BA,$31,$1F,$D1,$9B,$CA,$73,$0F,$AC
	.dc.b	$FE,$2E,$FF,$AB,$9A,$8D,$75,$FF,$BF,$A3,$3D,$FF,$C6,$3B,$FF,$CA
	.dc.b	$3D,$BF,$C9,$73,$FF,$75,$B7,$0F,$F4,$A3,$5F,$E4,$9C,$BF,$EF,$F9
	.dc.b	$FE,$5E,$D8,$FE,$BD,$DC,$8C,$47,$0F,$88,$FF,$2A,$C6,$23,$CD,$3B
	.dc.b	$9C,$92,$BF,$9D,$FD,$63,$EB,$3F,$CB,$AD,$8C,$C6,$0F,$5F,$74,$8F
	.dc.b	$5B,$70,$FE,$DF,$F5,$0F,$AC,$FF,$08,$46,$1A,$FF,$77,$8D,$57,$F6
	.dc.b	$B6,$C7,$FA,$DF,$EB,$3F,$CE,$24,$F6,$7F,$CF,$6D,$AF,$EF,$BC,$08
	.dc.b	$C5,$71,$AD,$92,$B7,$3A,$9F,$F7,$4F,$2B,$EB,$3F,$C4,$31,$FE,$0B
	.dc.b	$8C,$7A,$3E,$88,$FC,$04,$62,$FF,$C1,$AD,$8F,$F6,$74,$FF,$3A,$C4
	.dc.b	$D2,$78,$BC,$E7,$F1,$59,$BC,$AC,$42,$31,$8F,$C5,$9C,$C6,$27,$E7
	.dc.b	$C6,$09,$3B,$24,$7E,$75,$FC,$C6,$0F,$A2,$7D,$67,$D6,$F3,$BF,$DF
	.dc.b	$58,$F5,$EB,$FB,$37,$BA,$D7,$87,$FA,$1C,$BC,$DF,$59,$F7,$B3,$FB
	.dc.b	$8D,$8D,$62,$8C,$47,$C0,$6B,$CA,$7F,$CD,$0D,$3E,$84,$F1,$8C,$B6
	.dc.b	$33,$3E,$47,$FD,$E5,$F4,$CF,$97,$C4,$73,$D1,$8F,$C2,$AD,$63,$FB
	.dc.b	$B9,$F8,$38,$73,$CD,$5A,$C3,$FD,$A7,$87,$B2,$B5,$39,$BE,$B4,$E1
	.dc.b	$7D,$67,$FA,$5B,$2F,$C7,$B6,$7F,$16,$D7,$83,$67,$8F,$B7,$F2,$1B
	.dc.b	$5F,$E8,$D6,$D7,$EB,$1B,$1D,$A7,$D5,$F5,$3F,$8B,$8F,$76,$7E,$09
	.dc.b	$CE,$DF,$1B,$E9,$FF,$C7,$5E,$F6,$B2,$FE,$3D,$18,$7E,$08,$FA,$B7
	.dc.b	$D6,$7F,$9E,$F6,$F8,$32,$AE,$11,$86,$E5,$F8,$96,$CF,$3A,$E5,$39
	.dc.b	$F1,$18,$F8,$EB,$A4,$61,$FF,$3E,$DF,$95,$FF,$41,$35,$E5,$3F,$E7
	.dc.b	$78,$FF,$10,$6D,$7D,$3F,$AD,$5C,$9A,$F2,$6B,$E7,$77,$97,$AD,$B8
	.dc.b	$E6,$BE,$46,$6F,$AC,$7C,$5F,$E9,$F0,$E1,$5E,$BC,$BC,$79,$7C,$C3
	.dc.b	$E4,$65,$D1,$F9,$D8,$72,$7E,$A5,$3E,$B3,$11,$F5,$39,$9C,$7F,$E9
	.dc.b	$0D,$7F,$7E,$D9,$FC,$A5,$70,$CF,$F2,$F8,$73,$D3,$F7,$9C,$7F,$C9
	.dc.b	$2B,$AB,$63,$37,$FC,$8A,$72,$D8,$CB,$D9,$B5,$D6,$3F,$31,$3C,$1F
	.dc.b	$C6,$F8,$5F,$17,$2F,$DA,$7D,$65,$9F,$DD,$B5,$FC,$64,$DD,$CE,$F5
	.dc.b	$FD,$BD,$5E,$73,$B9,$97,$FC,$3A,$F8,$B6,$3D,$2F,$63,$04,$3F,$A1
	.dc.b	$47,$06,$C7,$08,$FE,$E9,$1F,$E0,$5E,$BF,$E5,$3D,$8F,$95,$F3,$36
	.dc.b	$7F,$05,$E4,$FF,$2B,$AF,$EC,$B7,$CA,$7F,$3B,$3F,$EC,$F7,$F5,$8D
	.dc.b	$8D,$47,$E8,$23,$04,$7F,$40,$FF,$B5,$F3,$79,$DF,$E3,$99,$FE,$D2
	.dc.b	$D7,$FD,$D5,$1A,$F3,$F8,$4E,$67,$FB,$3F,$D6,$7F,$BA,$9F,$EC,$2D
	.dc.b	$7F,$B4,$35,$DF,$BF,$E1,$FA,$CF,$F6,$39,$39,$70,$49,$FE,$F7,$19
	.dc.b	$6C,$65,$F2,$C6,$5B,$1F,$17,$EA,$7F,$6F,$1F,$D5,$5A,$FF,$8D,$DA
	.dc.b	$9C,$E4,$FA,$A7,$CD,$7C,$76,$D7,$F8,$39,$AF,$C6,$6B,$FE,$DB,$FA
	.dc.b	$CF,$F4,$A3,$FE,$E6,$6B,$FE,$38,$6B,$BF,$73,$FB,$53,$FE,$E3,$FA
	.dc.b	$C7,$8B,$88,$FF,$1C,$C6,$1F,$F7,$8B,$5F,$F2,$1B,$5E,$51,$C5,$7A
	.dc.b	$89,$D4,$74,$9D,$4E,$5F,$81,$F4,$B2,$7E,$03,$5F,$26,$8F,$9B,$FA
	.dc.b	$04,$8F,$04,$65,$F1,$7D,$63,$E8,$FF,$34,$D8,$CC,$60,$8E,$08,$7C
	.dc.b	$38,$E0,$FC,$5E,$2F,$F1,$6B,$EF,$6B,$9F,$47,$F5,$11,$88,$F9,$ED
	.dc.b	$7F,$C3,$4D,$7F,$CF,$8D,$4D,$3F,$C6,$CF,$7F,$43,$A9,$1F,$A1,$F5
	.dc.b	$5F,$A7,$7F,$C4,$CD,$7F,$D1,$0D,$4D,$27,$47,$EF,$DB,$1E,$DF,$AC
	.dc.b	$F5,$91,$C1,$0F,$88,$E5,$A7,$C5,$FA,$CF,$86,$47,$04,$3E,$63,$96
	.dc.b	$9F,$37,$E5,$3E,$59,$1C,$1F,$03,$D8,$F4,$11,$C1,$0F,$CE,$5A,$35
	.dc.b	$18,$7F,$8F,$7D,$53,$86,$92,$7F,$39,$19,$6C,$65,$FD,$44,$65,$B1
	.dc.b	$D3,$E6,$A4,$72,$6B,$C1,$AF,$ED,$25,$18,$8F,$F3,$04,$3E,$B3,$CE
	.dc.b	$48,$F2,$BF,$CD,$FE,$B1,$E9,$D2,$4F,$ED,$63,$2D,$8C,$BF,$C0,$63
	.dc.b	$2D,$8F,$8B,$EC,$48,$F8,$6D,$7E,$0B,$5F,$53,$5F,$4F,$C8,$7C,$E2
	.dc.b	$3F,$31,$AF,$CA,$6B,$F0,$DF,$AB,$F1,$92,$3E,$9B,$5F,$47,$93,$F0
	.dc.b	$48,$FB,$DA,$E8,$C1,$0F,$F6,$EF,$D2,$7D,$97,$FD,$1D,$D8,$8C,$47
	.dc.b	$FB,$71,$AF,$F1,$E6,$AF,$AC,$7A,$DF,$DA,$5B,$1F,$C9,$5A,$FF,$2F
	.dc.b	$6B,$E0,$7F,$B3,$BF,$CC,$EB,$A7,$91,$FC,$A2,$73,$3F,$EF,$46,$BF
	.dc.b	$CF,$1F,$EE,$AF,$FA,$9D,$A8,$F3,$79,$3E,$53,$CD,$F1,$52,$3C,$CD
	.dc.b	$7C,$CD,$7B,$35,$E9,$F1,$3F,$E9,$1F,$AC,$FE,$84,$D8,$FE,$98,$D7
	.dc.b	$FA,$8B,$5F,$EA,$7B,$7F,$B4,$93,$49,$E5,$7F,$84,$DB,$1F,$D3,$DA
	.dc.b	$FF,$59,$FA,$C7,$C4,$7F,$86,$5B,$1F,$D5,$5A,$EE,$3A,$7C,$74,$8E
	.dc.b	$8D,$7F,$B2,$52,$31,$1E,$86,$BF,$ED,$EF,$99,$FE,$8E,$FF,$B5,$FC
	.dc.b	$E4,$62,$35,$E3,$7F,$89,$63,$11,$F5,$7E,$AD,$F5,$9F,$E3,$16,$C7
	.dc.b	$F6,$D6,$BF,$DD,$DA,$FF,$7C,$6B,$FE,$E7,$6B,$D6,$3A,$1F,$E3,$DF
	.dc.b	$AC,$FE,$9B,$1F,$E0,$16,$BF,$E0,$66,$BF,$EF,$16,$BF,$8B,$F9,$C7
	.dc.b	$D6,$7F,$91,$DB,$1F,$E0,$36,$BF,$E0,$D6,$BF,$E1,$26,$BF,$EF,$D6
	.dc.b	$BD,$63,$A1,$FE,$0D,$F8,$4F,$F1,$3F,$05,$F5,$0D,$25,$7F,$84,$FE
	.dc.b	$B3,$FC,$C0,$FF,$87,$5A,$FF,$88,$9A,$FE,$94,$38,$7F,$50,$D7,$3D
	.dc.b	$9F,$E6,$58,$C1,$9C,$BF,$CA,$53,$F1,$FC,$CF,$9C,$47,$C9,$C3,$C4
	.dc.b	$8E,$0F,$F1,$57,$6F,$FB,$71,$FF,$1B,$B5,$F3,$35,$FD,$E3,$5F,$E3
	.dc.b	$1E,$F7,$F9,$E6,$30,$47,$FC,$E9,$E8,$BE,$AE,$92,$BF,$C8,$5C,$BF
	.dc.b	$E2,$F8,$C4,$7F,$9D,$A7,$83,$A6,$B9,$8F,$F4,$07,$D6,$62,$3F,$44
	.dc.b	$D7,$F8,$95,$23,$10,$8C,$4F,$22,$47,$92,$77,$3A,$8F,$E4,$DF,$56
	.dc.b	$FA,$CF,$F2,$6B,$63,$31,$83,$FD,$29,$F5,$9F,$ED,$A8,$FD,$B3,$5F
	.dc.b	$2B,$5F,$23,$68,$D2,$7C,$37,$43,$FD,$51,$F4,$0F,$21,$F0,$3F,$E6
	.dc.b	$58,$C3,$63,$38,$7F,$9C,$67,$2D,$FA,$C7,$CF,$71,$6C,$66,$30,$7F
	.dc.b	$AD,$D7,$C5,$52,$3F,$EB,$B7,$EE,$FD,$6B,$DB,$FD,$4C,$F9,$BA,$BE
	.dc.b	$AC,$E5,$FF,$66,$1F,$63,$EB,$5F,$67,$C3,$5E,$3C,$0B,$E1,$87,$33
	.dc.b	$EE,$38,$9D,$B7,$0E,$97,$CD,$F0,$E3,$6F,$F0,$9F,$5A,$F7,$3C,$F1
	.dc.b	$FC,$B6,$31,$5F,$98,$6C,$7F,$B3,$1A,$FF,$33,$C3,$FC,$85,$3E,$C8
	.dc.b	$F2,$FB,$3D,$F3,$5F,$2C,$FC,$F3,$9C,$38,$DF,$B6,$FE,$04,$F1,$BE
	.dc.b	$18,$7C,$A8,$F5,$7D,$87,$E0,$BC,$EB,$91,$F0,$7D,$6B,$F0,$1C,$C4
	.dc.b	$C3,$CD,$7C,$5F,$46,$4B,$E1,$1A,$C3,$A3,$AF,$D5,$1E,$27,$B5,$B1
	.dc.b	$F0,$65,$F9,$8B,$E2,$35,$72,$B9,$3F,$4E,$B7,$39,$DF,$E1,$7E,$A9
	.dc.b	$FC,$CA,$FF,$84,$B6,$3F,$DF,$AD,$7F,$B2,$35,$FF,$0A,$69,$F9,$8F
	.dc.b	$99,$F8,$FC,$E7,$EA,$57,$3C,$7F,$2F,$5B,$C3,$E1,$FC,$5B,$E9,$7F
	.dc.b	$0B,$97,$F0,$E8,$C4,$FF,$1C,$6C,$7F,$B8,$B8,$11,$88,$FE,$F0,$D7
	.dc.b	$95,$74,$7F,$8B,$46,$5B,$19,$BF,$D8,$F9,$9E,$A3,$F1,$95,$E0,$C7
	.dc.b	$D3,$A2,$7F,$21,$CF,$F4,$73,$FE,$62,$DC,$BC,$CD,$7F,$C1,$8D,$7F
	.dc.b	$95,$B5,$FE,$54,$DC,$FF,$51,$B7,$A5,$B5,$FE,$FF,$CD,$F8,$33,$C0
	.dc.b	$FC,$BB,$67,$F1,$F8,$78,$F4,$FA,$33,$9C,$BF,$45,$F6,$4F,$BB,$EC
	.dc.b	$DC,$3F,$A5,$B6,$7E,$83,$5F,$F1,$33,$5F,$E8,$AD,$7F,$A2,$37,$33
	.dc.b	$9C,$FD,$3E,$47,$A0,$FD,$D4,$77,$CB,$CF,$1F,$BD,$F2,$1F,$69,$F0
	.dc.b	$DB,$97,$F6,$06,$E5,$F8,$56,$BF,$E3,$B6,$BF,$D6,$9A,$FF,$59,$6E
	.dc.b	$5F,$DB,$6B,$F0,$9E,$3B,$EA,$FF,$81,$B7,$0F,$DD,$DF,$83,$2F,$CE
	.dc.b	$CE,$71,$FE,$1D,$F8,$CF,$FA,$F2,$FF,$BA,$B6,$3F,$34,$D7,$FC,$A0
	.dc.b	$D7,$FB,$7B,$5F,$ED,$CD,$C8,$D3,$5A,$FF,$6B,$FA,$8F,$39,$FC,$96
	.dc.b	$3C,$18,$F9,$A3,$F9,$B7,$C1,$3F,$55,$7C,$07,$E9,$E3,$0D,$8C,$E7
	.dc.b	$FE,$9D,$6E,$7F,$E3,$11,$5C,$27,$55,$C4,$8F,$F8,$5B,$F5,$6F,$1B
	.dc.b	$FE,$0C,$6D,$72,$BE,$B8,$FF,$56,$9C,$E8,$FD,$5D,$F2,$BF,$D0,$5B
	.dc.b	$8F,$F4,$5C,$FF,$D2,$7D,$39,$AF,$EA,$9F,$47,$FC,$AA,$BD,$1A,$23
	.dc.b	$0F,$AD,$AF,$F9,$35,$AF,$F8,$8B,$F0,$9F,$E7,$2F,$F6,$7E,$17,$E3
	.dc.b	$5F,$E9,$AB,$FE,$7F,$5A,$DB,$F9,$0C,$E6,$FF,$B7,$7C,$37,$F8,$1B
	.dc.b	$93,$F8,$B3,$70,$FF,$58,$FA,$1F,$E2,$56,$CE,$A3,$15,$FE,$B4,$6B
	.dc.b	$FE,$63,$AE,$4E,$67,$F8,$F7,$C5,$FF,$62,$C7,$F9,$9B,$70,$FF,$61
	.dc.b	$E1,$D2,$73,$E6,$7F,$62,$FB,$27,$D7,$53,$FB,$87,$EB,$0F,$98,$C6
	.dc.b	$3F,$D8,$2B,$6D,$AD,$CE,$B4,$FE,$E5,$F5,$8F,$C2,$BF,$B7,$FE,$17
	.dc.b	$FC,$57,$4F,$F0,$AC,$E7,$1F,$ED,$73,$97,$FC,$A7,$19,$BF,$A5,$39
	.dc.b	$7F,$CA,$91,$9A,$FE,$A3,$39,$5F,$6F,$FD,$C2,$BF,$35,$18,$F4,$FF
	.dc.b	$A4,$37,$0F,$C3,$B6,$77,$5C,$5F,$F4,$94,$65,$B1,$97,$FD,$F5,$19
	.dc.b	$7F,$DF,$9D,$2F,$6D,$E2,$FF,$86,$F9,$FF,$86,$63,$A9,$CB,$F8,$CF
	.dc.b	$17,$FC,$2B,$E8,$3E,$66,$9F,$E2,$3F,$D6,$7E,$E6,$6E,$CF,$EE,$93
	.dc.b	$AC,$3F,$C2,$B7,$C5,$FF,$0A,$DF,$0A,$DC,$7F,$86,$3F,$5A,$7D,$8F
	.dc.b	$F8,$93,$E8,$9F,$67,$FE,$31,$9C,$D7,$F8,$82,$72,$F1,$AE,$15,$B6
	.dc.b	$C7,$F8,$EF,$BF,$FC,$CF,$E1,$FE,$74,$91,$FA,$26,$EB,$F9,$F6,$BF
	.dc.b	$9F,$EE,$FE,$D5,$9B,$BD,$F4,$5F,$E3,$A6,$E1,$DE,$BF,$B2,$3F,$E1
	.dc.b	$A8,$CD,$FF,$91,$7F,$17,$FC,$27,$3F,$E2,$39,$BF,$C9,$D1,$8B,$FF
	.dc.b	$10,$F5,$7D,$4B,$E1,$D5,$F9,$AB,$DB,$6B,$E7,$F8,$BF,$60,$FF,$2C
	.dc.b	$7D,$F7,$F2,$0D,$BF,$C5,$11,$99,$D9,$FE,$48,$9D,$73,$7F,$1E,$C7
	.dc.b	$F8,$4E,$DF,$D3,$DF,$85,$E7,$FF,$05,$25,$FF,$0A,$6C,$7F,$07,$6B
	.dc.b	$FE,$E9,$C7,$E4,$77,$FC,$73,$FC,$F5,$C7,$FE,$9A,$AC,$5F,$E2,$3D
	.dc.b	$4F,$C8,$9F,$E7,$1F,$C6,$1F,$39,$89,$D1,$32,$FF,$38,$4E,$A3,$1C
	.dc.b	$7F,$E4,$4F,$8B,$FF,$46,$57,$E5,$DB,$E6,$BF,$1D,$7C,$A7,$F9,$D2
	.dc.b	$77,$B9,$A7,$09,$DE,$3F,$E7,$28,$E7,$E9,$7F,$17,$8F,$F7,$3B,$71
	.dc.b	$FF,$26,$35,$1F,$EE,$07,$0F,$FB,$8B,$F0,$FF,$F4,$3B,$FE,$E1,$6E
	.dc.b	$5F,$DA,$BC,$9F,$F4,$A4,$60,$CF,$E6,$AF,$97,$CF,$B9,$FE,$99,$8C
	.dc.b	$65,$FE,$86,$AD,$D6,$A3,$FD,$13,$3A,$C7,$FD,$35,$F9,$C3,$CD,$FF
	.dc.b	$23,$36,$BF,$AE,$E6,$8C,$57,$F9,$A5,$AF,$E0,$75,$7D,$EF,$CD,$3F
	.dc.b	$D5,$93,$C3,$8B,$EB,$47,$FA,$BF,$EF,$3F,$D9,$78,$F6,$6E,$1F,$E7
	.dc.b	$56,$BF,$28,$62,$8C,$C7,$FA,$2B,$1F,$E2,$07,$F1,$89,$D7,$C6,$FD
	.dc.b	$FF,$52,$61,$F6,$DF,$14,$62,$FF,$B9,$5F,$29,$CD,$6C,$FF,$58,$F9
	.dc.b	$9F,$E3,$E2,$79,$BF,$E6,$D7,$FC,$87,$CD,$F8,$BF,$1F,$F1,$C7,$FB
	.dc.b	$36,$3F,$99,$E3,$F9,$58,$FF,$68,$7E,$3D,$FE,$02,$8F,$76,$88,$C4
	.dc.b	$7E,$2A,$31,$F1,$7F,$E7,$B9,$CC,$FC,$46,$CE,$B2,$FB,$31,$83,$78
	.dc.b	$FF,$AB,$6F,$FA,$09,$E2,$2C,$FF,$B3,$A5,$F8,$FA,$3F,$36,$79,$A3
	.dc.b	$2E,$30,$F3,$46,$1F,$F7,$BE,$1E,$BA,$E5,$18,$9D,$57,$39,$CB,$4E
	.dc.b	$93,$ED,$8C,$BD,$9E,$7F,$0B,$FB,$98,$FE,$AD,$E2,$3E,$83,$3C,$1B
	.dc.b	$C3,$F6,$77,$BD,$3F,$D9,$D5,$AC,$3B,$56,$AF,$FD,$AB,$87,$38,$C4
	.dc.b	$E8,$61,$CE,$75,$18,$AF,$B1,$F8,$FF,$F0,$A8,$C1,$F9,$A7,$15,$E9
	.dc.b	$D1,$0F,$A6,$F2,$46,$3C,$0B,$3C,$30,$EB,$18,$AC,$B3,$2F,$3D,$F1
	.dc.b	$6D,$F5,$6C,$FC,$1C,$5F,$68,$FA,$8F,$F6,$B4,$7A,$0C,$B8,$CF,$E8
	.dc.b	$7C,$03,$EF,$7C,$95,$FC,$82,$F9,$7C,$5F,$C8,$8E,$2D,$8C,$C7,$C4
	.dc.b	$AE,$58,$F8,$E7,$38,$7D,$DA,$E2,$7F,$9C,$23,$8B,$FE,$89,$6B,$FC
	.dc.b	$1D,$AF,$8F,$57,$DD,$FF,$9E,$23,$DF,$18,$7A,$35,$E8,$D7,$A3,$BA
	.dc.b	$F2,$57,$D7,$AF,$64,$E7,$7F,$E8,$38,$BE,$F7,$FD,$DD,$8E,$AE,$3F
	.dc.b	$37,$0E,$4F,$C3,$9D,$BF,$88,$E0,$7E,$33,$F5,$23,$0D,$CF,$F1,$4D
	.dc.b	$CB,$F8,$43,$FE,$AD,$8C,$EB,$FC,$AD,$B5,$F8,$69,$E0,$FF,$2E,$78
	.dc.b	$57,$2A,$E0,$D9,$E1,$18,$78,$16,$DD,$7F,$9B,$8F,$D4,$F2,$B9,$1F
	.dc.b	$4B,$C6,$30,$F1,$AE,$15,$F0,$5B,$9F,$EB,$E7,$85,$F2,$AF,$CF,$D6
	.dc.b	$B2,$FE,$50,$F2,$E4,$FE,$5A,$D9,$CF,$0F,$F4,$D9,$CB,$63,$27,$F1
	.dc.b	$2C,$FD,$71,$98,$FE,$A7,$5C,$DB,$19,$CF,$DD,$87,$F9,$F3,$0F,$E1
	.dc.b	$6D,$AF,$E2,$ED,$C3,$E3,$FE,$01,$71,$3F,$97,$8C,$D7,$F7,$49,$CB
	.dc.b	$67,$51,$88,$FF,$76,$4F,$06,$C6,$5F,$87,$3F,$A3,$C3,$FC,$45,$E8
	.dc.b	$79,$5B,$3F,$E2,$18,$F3,$7D,$CB,$F1,$9A,$79,$5C,$4F,$F8,$21,$AF
	.dc.b	$89,$AF,$F8,$1F,$D4,$ED,$99,$A7,$C3,$FB,$A7,$E1,$F8,$AF,$F1,$94
	.dc.b	$D5,$E3,$FF,$4C,$D9,$7C,$27,$64,$3F,$BE,$D7,$5B,$91,$FE,$1E,$9E
	.dc.b	$0F,$C8,$6B,$FD,$D1,$AF,$F8,$3F,$F1,$7F,$49,$1F,$F5,$AF,$07,$F7
	.dc.b	$7F,$61,$F5,$FF,$75,$B3,$FE,$2C,$6B,$F5,$9A,$FF,$80,$9A,$FF,$91
	.dc.b	$F9,$FF,$0D,$E9,$7E,$B2,$3B,$9F,$E2,$47,$0F,$C4,$BE,$0F,$F9,$0E
	.dc.b	$78,$F2,$48,$C1,$0E,$07,$F9,$4E,$3F,$29,$E8,$7F,$0F,$DF,$E7,$7E
	.dc.b	$06,$F1,$FF,$30,$DB,$FF,$23,$B5,$FC,$8B,$5F,$F0,$EB,$5F,$F0,$E7
	.dc.b	$E2,$8F,$9A,$FD,$13,$63,$DB,$C6,$8C,$47,$9A,$1F,$8F,$BD,$6E,$37
	.dc.b	$2B,$8D,$F0,$CB,$F0,$D7,$B9,$F9,$3D,$4F,$C1,$DF,$6B,$FA,$16,$CE
	.dc.b	$9A,$FF,$A4,$BC,$7D,$4C,$8D,$3F,$BB,$F6,$BF,$D1,$F1,$DB,$E1,$1F
	.dc.b	$5D,$FA,$D6,$C6,$63,$F8,$23,$AF,$6E,$C1,$FE,$B1,$19,$3F,$AD,$46
	.dc.b	$5B,$1A,$8C,$47,$F9,$A4,$FF,$49,$B8,$70,$D7,$FC,$DC,$D7,$FC,$DF
	.dc.b	$AB,$D4,$FA,$A4,$8F,$AB,$F8,$7B,$D6,$C4,$7D,$5A,$E0,$FD,$2A,$FE
	.dc.b	$3D,$5B,$8F,$BB,$F0,$CF,$B5,$FE,$20,$3F,$03,$A7,$F2,$58,$D4,$F0
	.dc.b	$9C,$BF,$69,$E0,$E4,$FA,$1F,$27,$F3,$C8,$FE,$A0,$7D,$DB,$EF,$3E
	.dc.b	$8B,$63,$FC,$0C,$D7,$E5,$1C,$23,$E9,$C6,$63,$E7,$35,$FF,$4C,$35
	.dc.b	$FF,$61,$B5,$E4,$7F,$AB,$BE,$41,$F7,$FF,$CC,$7D,$4B,$3F,$96,$6B
	.dc.b	$FA,$19,$BD,$3F,$C8,$5B,$E1,$7F,$B4,$3A,$7F,$C7,$78,$73,$3F,$A7
	.dc.b	$4E,$AD,$3F,$A1,$F8,$EF,$D0,$12,$3F,$63,$F2,$4F,$C6,$B1,$1F,$C1
	.dc.b	$AB,$83,$FA,$5A,$FE,$B5,$5B,$8F,$E2,$B9,$3F,$14,$FF,$0D,$BF,$E8
	.dc.b	$2F,$6B,$FD,$75,$EE,$7F,$63,$6D,$70,$AD,$EB,$F2,$BA,$FE,$79,$27
	.dc.b	$EA,$B5,$FA,$BC,$08,$C4,$7D,$47,$FB,$57,$C3,$79,$63,$D5,$C9,$FD
	.dc.b	$03,$3F,$F4,$AE,$DF,$93,$F3,$FF,$BA,$5E,$E7,$FA,$73,$72,$F9,$D7
	.dc.b	$C5,$B5,$C3,$83,$FB,$57,$E2,$BF,$EE,$DC,$BF,$87,$B5,$F9,$4D,$C7
	.dc.b	$FA,$D7,$89,$FC,$A5,$F1,$C6,$70,$FB,$8D,$AF,$B2,$DF,$1B,$F9,$57
	.dc.b	$91,$FC,$37,$9A,$57,$F1,$17,$F8,$6F,$E6,$1E,$78,$07,$FA,$EB,$86
	.dc.b	$B9,$D3,$F8,$8C,$60,$8E,$09,$1F,$A6,$BF,$E0,$9F,$45,$FE,$0B,$F8
	.dc.b	$07,$D5,$53,$FC,$41,$E1,$A7,$F0,$5C,$DF,$E6,$0C,$3F,$E0,$07,$3F
	.dc.b	$E2,$7E,$3F,$FC,$F0,$D8,$FD,$9B,$5F,$F0,$FF,$D0,$FE,$05,$FE,$92
	.dc.b	$D7,$FA,$43,$78,$BF,$DA,$7F,$38,$FD,$53,$FD,$91,$5F,$D4,$3C,$B7
	.dc.b	$D8,$FF,$8D,$5B,$19,$8F,$F1,$D5,$7F,$AD,$FF,$09,$BF,$02,$7F,$D1
	.dc.b	$D1,$9A,$FE,$12,$D7,$F9,$A3,$5F,$A6,$1C,$3F,$59,$AE,$67,$B3,$D1
	.dc.b	$AF,$F3,$FE,$D7,$E3,$CF,$9C,$FF,$1E,$36,$7F,$B0,$B5,$FE,$D9,$F8
	.dc.b	$43,$E3,$FF,$CA,$8D,$8F,$B8,$D7,$FC,$9A,$D7,$D3,$CE,$8C,$47,$E3
	.dc.b	$A3,$13,$FE,$5B,$9C,$CA,$72,$F6,$9F,$F2,$A5,$7D,$08,$FE,$20,$DD
	.dc.b	$FF,$6E,$DE,$DE,$DF,$80,$FE,$2E,$BF,$C1,$2D,$7F,$CC,$6D,$7C,$ED
	.dc.b	$78,$F3,$7F,$9B,$E3,$11,$E5,$6E,$26,$9A,$D3,$F9,$5D,$7F,$9A,$5F
	.dc.b	$06,$3F,$C3,$1B,$87,$C3,$FA,$CF,$E1,$58,$FC,$17,$F8,$A3,$F2,$9A
	.dc.b	$EF,$C1,$FF,$41,$46,$23,$84,$60,$8F,$EC,$63,$55,$FC,$FB,$EF,$7C
	.dc.b	$84,$9F,$F1,$6F,$E3,$1F,$6F,$1F,$F3,$53,$F5,$34,$34,$9B,$7F,$A4
	.dc.b	$DB,$5F,$BF,$6B,$F9,$1F,$9D,$FD,$5B,$FD,$1D,$B7,$FC,$B9,$BA,$7E
	.dc.b	$17,$EC,$6D,$7E,$0D,$E2,$FE,$31,$AE,$FE,$4F,$FD,$49,$18,$8E,$11
	.dc.b	$82,$3F,$C4,$E3,$5A,$2D,$F1,$1F,$E9,$F6,$D7,$F9,$11,$AF,$F9,$3F
	.dc.b	$F5,$B7,$C2,$FF,$AD,$DB,$1F,$C7,$DA,$FF,$AC,$1A,$FE,$91,$AF,$E8
	.dc.b	$FF,$5C,$FD,$86,$9F,$E9,$96,$E9,$9E,$7F,$F6,$34,$BF,$49,$FF,$15
	.dc.b	$DF,$F9,$86,$F6,$FE,$A5,$BC,$27,$35,$C7,$DA,$FF,$5E,$7D,$17,$F9
	.dc.b	$86,$FF,$BD,$36,$7F,$B0,$B7,$2F,$DC,$7D,$27,$F7,$2A,$FF,$03,$CE
	.dc.b	$5B,$F0,$9D,$0F,$F6,$D4,$EF,$27,$DD,$3D,$BF,$F6,$F3,$6F,$FC,$FA
	.dc.b	$D7,$FC,$F9,$F8,$07,$CA,$CF,$FD,$79,$39,$6D,$72,$3F,$DD,$13,$BC
	.dc.b	$FF,$87,$3F,$3A,$FF,$86,$37,$03,$4C,$61,$FE,$D6,$C3,$BF,$C8,$F9
	.dc.b	$8F,$F7,$A7,$DE,$3E,$57,$18,$FA,$26,$BF,$E9,$39,$D3,$FE,$4D,$FA
	.dc.b	$B7,$DF,$7F,$B5,$A3,$FC,$93,$18,$C3,$FC,$97,$08,$C3,$A9,$E0,$FF
	.dc.b	$9E,$61,$F7,$CE,$1E,$31,$86,$C6,$67,$BB,$5E,$FF,$35,$FE,$8A,$4D
	.dc.b	$23,$FE,$87,$DB,$EC,$F9,$4E,$0F,$AE,$D9,$F8,$1E,$8D,$D3,$EB,$CD
	.dc.b	$AF,$E8,$1C,$1F,$DA,$9E,$27,$FA,$F3,$6F,$F4,$43,$8C,$3F,$D7,$9F
	.dc.b	$7D,$FE,$9F,$8F,$F0,$53,$63,$C8,$D7,$FA,$7B,$5E,$FF,$25,$E1,$7F
	.dc.b	$A6,$5A,$31,$9A,$BA,$65,$F8,$E9,$CD,$6F,$0F,$CA,$C7,$0A,$DC,$76
	.dc.b	$3F,$D9,$34,$FA,$BF,$07,$80,$8C,$4F,$FB,$6B,$EF,$BF,$DD,$0F,$0A
	.dc.b	$DC,$EB,$3F,$F3,$63,$4F,$9D,$9A,$F8,$CE,$AB,$D7,$F7,$DF,$E2,$CC
	.dc.b	$3F,$DC,$ED,$9F,$F7,$DE,$0F,$BE,$7C,$37,$0F,$66,$61,$C1,$F9,$15
	.dc.b	$B9,$CD,$FF,$BB,$FE,$FB,$F0,$57,$FE,$3D,$6C,$7E,$1D,$AF,$F8,$09
	.dc.b	$AF,$FA,$4F,$93,$F5,$53,$FE,$F0,$6C,$75,$8F,$93,$18,$BF,$83,$E3
	.dc.b	$BE,$07,$F2,$17,$FD,$0D,$AF,$F4,$96,$BF,$D2,$5B,$BA,$79,$61,$FD
	.dc.b	$0E,$73,$18,$AF,$D0,$35,$FF,$19,$F0,$7E,$9B,$E5,$FC,$A7,$E2,$A7
	.dc.b	$86,$3F,$E8,$8C,$93,$F8,$EF,$91,$F9,$F9,$7F,$98,$DB,$8F,$EB,$5A
	.dc.b	$FF,$8A,$1A,$FF,$89,$DA,$FF,$89,$5B,$99,$A6,$B2,$7D,$F3,$DD,$FC
	.dc.b	$9B,$70,$FF,$3F,$E7,$F8,$CB,$F1,$BF,$EA,$09,$D6,$3F,$99,$AF,$F0
	.dc.b	$B7,$DF,$7C,$AF,$48,$D2,$47,$F8,$3D,$B3,$FC,$26,$77,$39,$3F,$DD
	.dc.b	$39,$79,$BC,$5F,$29,$FB,$49,$DD,$7F,$B3,$70,$4F,$EE,$3E,$F9,$E4
	.dc.b	$7D,$C6,$E3,$FE,$8F,$6D,$FF,$19,$6B,$FE,$54,$6B,$FE,$53,$6B,$FE
	.dc.b	$51,$6E,$C7,$39,$3E,$E9,$F6,$7F,$BE,$6E,$1F,$30,$E9,$8E,$67,$F8
	.dc.b	$7F,$DF,$3E,$3B,$F2,$6D,$AF,$F5,$33,$73,$FE,$60,$D7,$FC,$D6,$D7
	.dc.b	$FC,$D4,$D7,$FC,$D0,$DC,$CE,$72,$F4,$7B,$3E,$73,$C7,$85,$C3,$D1
	.dc.b	$87,$FA,$3F,$EF,$9F,$3F,$FA,$AF,$80,$46,$2B,$FD,$32,$DA,$FF,$4C
	.dc.b	$B5,$FF,$4C,$CF,$0A,$E5,$39,$9F,$F4,$D3,$5F,$04,$7E,$DB,$EF,$BE
	.dc.b	$17,$2F,$F2,$06,$E1,$C2,$7F,$9C,$C6,$2B,$FD,$11,$F0,$BF,$E3,$4A
	.dc.b	$FF,$5F,$FA,$FF,$D0,$72,$7B,$5F,$9E,$9F,$F4,$95,$6E,$FF,$3C,$3E
	.dc.b	$1B,$F9,$54,$FF,$B3,$3C,$2F,$E3,$33,$96,$CE,$A3,$FB,$37,$DD,$7F
	.dc.b	$34,$D7,$FB,$27,$0B,$EF,$9F,$47,$FD,$42,$B5,$97,$FB,$C2,$B5,$3F
	.dc.b	$D5,$FD,$DF,$EF,$9D,$3F,$C6,$9B,$5F,$EC,$4C,$BF,$DE,$F1,$87,$0D
	.dc.b	$7F,$D9,$2D,$77,$E9,$7E,$66,$79,$47,$F9,$86,$30,$F2,$6B,$FC,$83
	.dc.b	$E4,$FF,$B1,$3F,$1B,$14,$63,$4F,$F0,$16,$1E,$37,$E0,$C6,$70,$F7
	.dc.b	$7D,$EF,$F9,$05,$FC,$7D,$A3,$18,$7D,$2E,$A7,$DF,$3D,$CF,$ED,$75
	.dc.b	$AC,$BE,$6B,$71,$EB,$EE,$7F,$7C,$BE,$11,$DA,$75,$8F,$FB,$37,$E5
	.dc.b	$5F,$55,$FD,$DB,$EE,$FF,$8F,$E5,$FD,$BA,$73,$87,$9A,$72,$FF,$B6
	.dc.b	$63,$37,$F0,$E7,$2F,$5B,$E9,$E4,$FF,$6A,$FB,$EF,$EF,$93,$FE,$FC
	.dc.b	$C5,$18,$8F,$E6,$6D,$7F,$C1,$0D,$8D,$FA,$7F,$D7,$1B,$97,$DA,$6D
	.dc.b	$7E,$15,$AF,$EC,$BD,$CF,$E5,$33,$FB,$EE,$07,$3F,$F2,$6B,$FF,$1D
	.dc.b	$76,$FF,$94,$7C,$8F,$CB,$78,$FF,$E1,$AF,$7D,$F6,$B4,$FF,$14,$FC
	.dc.b	$DF,$DD,$D5,$D1,$FE,$65,$9D,$6B,$F7,$AF,$8B,$E1,$BE,$15,$B8,$F2
	.dc.b	$7C,$E3,$E1,$BF,$C5,$3F,$7C,$FB,$EF,$F1,$BC,$E6,$BF,$13,$39,$78
	.dc.b	$D7,$0A,$DB,$63,$F8,$EF,$77,$F9,$9F,$D9,$FD,$67,$81,$FD,$B2,$31
	.dc.b	$3F,$A4,$6E,$DF,$A3,$AD,$92,$75,$5F,$DB,$FC,$07,$C6,$7F,$90,$1B
	.dc.b	$3E,$09,$FE,$21,$7F,$1A,$72,$FE,$7F,$E8,$7F,$C7,$D5,$FC,$0F,$B7
	.dc.b	$EB,$CE,$7A,$BF,$2D,$CB,$FE,$58,$BD,$CF,$D0,$AD,$F3,$FF,$91,$A7
	.dc.b	$3C,$49,$CE,$3F,$C3,$F9,$3F,$90,$5F,$F1,$66,$C3,$07,$4B,$EB,$BF
	.dc.b	$97,$63,$FE,$A1,$F2,$FF,$08,$4B,$FC,$64,$EB,$0E,$78,$7F,$0D,$F4
	.dc.b	$BE,$79,$2B,$E4,$B6,$3F,$88,$35,$FE,$33,$C1,$FD,$B1,$FE,$41,$8F
	.dc.b	$F1,$0F,$03,$E7,$1F,$C9,$E7,$5A,$7F,$10,$8F,$CC,$78,$13,$1F,$D0
	.dc.b	$7D,$3B,$EF,$B1,$97,$F1,$BD,$3F,$73,$9F,$FA,$F7,$0F,$F4,$2D,$EF
	.dc.b	$0F,$CD,$47,$F9,$F3,$E3,$FF,$D3,$58,$7F,$23,$6E,$7F,$E7,$6F,$BE
	.dc.b	$7D,$56,$FE,$26,$63,$FC,$82,$F6,$FC,$FE,$35,$F0,$8D,$57,$FA,$5B
	.dc.b	$D4,$FE,$37,$97,$FA,$CD,$AF,$D7,$3E,$BB,$87,$A3,$5F,$B7,$F5,$DF
	.dc.b	$E9,$86,$FB,$DD,$7C,$EF,$F5,$1C,$E7,$23,$7E,$F5,$F7,$DA,$BF,$F5
	.dc.b	$4C,$63,$3F,$E1,$95,$BA,$D4,$7F,$0F,$9D,$63,$FC,$F7,$EF,$9E,$77
	.dc.b	$F9,$D1,$B5,$FD,$53,$24,$62,$BF,$DA,$4D,$7F,$CE,$7B,$BE,$FB,$EA
	.dc.b	$1F,$EB,$C9,$E1,$C5,$FD,$CA,$75,$87,$F2,$1F,$BE,$FF,$67,$E1,$FE
	.dc.b	$DD,$6B,$F2,$F1,$46,$23,$EB,$35,$FE,$3F,$5F,$C6,$8F,$E4,$53,$C3
	.dc.b	$E4,$3F,$0D,$F1,$D3,$0F,$C0,$3F,$ED,$0C,$3C,$38,$F9,$63,$58,$FE
	.dc.b	$2D,$D5,$E2,$7F,$D8,$3E,$6F,$F9,$50,$9E,$77,$FA,$71,$FF,$2A,$E7
	.dc.b	$FC,$32,$7E,$07,$8F,$FE,$DE,$9F,$E7,$D9,$FF,$83,$EB,$4F,$E4,$7E
	.dc.b	$FB,$FD,$97,$9F,$C1,$6B,$9F,$A8,$FF,$54,$C6,$23,$F8,$2B,$76,$34
	.dc.b	$93,$FE,$B2,$ED,$77,$7F,$88,$2F,$FA,$3B,$6F,$8D,$7F,$4F,$3F,$DE
	.dc.b	$B6,$FB,$E6,$3F,$C5,$4F,$33,$8C,$3F,$D8,$77,$E2,$8C,$4E,$8E,$13
	.dc.b	$EF,$8C,$B6,$32,$FB,$23,$2D,$8C,$CF,$57,$F9,$37,$C5,$33,$FE,$59
	.dc.b	$F0,$1E,$38,$C5,$F1,$0E,$34,$F8,$73,$BA,$F1,$56,$9E,$77,$D9,$CD
	.dc.b	$F1,$D3,$FC,$B7,$39,$AF,$E6,$FF,$7D,$FC,$25,$E4,$63,$C0,$EC,$EB
	.dc.b	$E8,$AE,$8A,$E1,$8F,$A2,$31,$7A,$9C,$E5,$FC,$5D,$5E,$66,$63,$E3
	.dc.b	$BE,$2D,$AD,$B6,$7C,$78,$BE,$BB,$FD,$49,$F1,$BE,$B8,$C7,$C3,$FF
	.dc.b	$BA,$9F,$5F,$C1,$BE,$A3,$C7,$1F,$D5,$EF,$8B,$F0,$0E,$37,$CE,$33
	.dc.b	$B7,$D4,$BE,$58,$F8,$A7,$3B,$7F,$2B,$79,$E1,$FE,$6F,$6B,$9C,$BC
	.dc.b	$8E,$5A,$F1,$D9,$F7,$DF,$E6,$E8,$F7,$46,$1F,$F3,$8D,$7C,$A9,$D3
	.dc.b	$63,$33,$BC,$BF,$17,$1C,$AB,$D9,$39,$C3,$FB,$C6,$4F,$BE,$FF,$80
	.dc.b	$17,$AB,$8F,$C5,$C7,$B6,$3F,$0E,$B7,$3F,$88,$E0,$7D,$F7,$BA,$BF
	.dc.b	$08,$DC,$BF,$AC,$36,$BF,$C0,$EF,$FA,$56,$77,$B7,$F5,$56,$CF,$E1
	.dc.b	$67,$85,$F6,$8D,$5F,$4B,$B1,$D1,$B3,$E4,$6D,$77,$AF,$ED,$5D,$BF
	.dc.b	$94,$BF,$D2,$72,$FF,$10,$FB,$EE,$6D,$8F,$CB,$4E,$DF,$80,$DC,$BF
	.dc.b	$5D,$5C,$39,$4F,$8A,$75,$87,$83,$83,$FB,$97,$27,$F4,$99,$E8,$D8
	.dc.b	$CE,$3B,$3F,$85,$CE,$9C,$1F,$C6,$9C,$3F,$E3,$26,$E5,$FC,$AA,$FF
	.dc.b	$D1,$57,$FE,$3B,$6D,$7F,$2C,$6E,$AB,$7D,$F6,$1B,$5F,$DB,$DA,$FF
	.dc.b	$B3,$67,$84,$FE,$65,$B1,$B8,$C4,$F0,$8F,$81,$18,$3F,$41,$87,$F2
	.dc.b	$5F,$3F,$CA,$D9,$FF,$11,$47,$9B,$EF,$9F,$41,$4F,$03,$89,$FF,$02
	.dc.b	$35,$F3,$35,$FF,$01,$FA,$9D,$B3,$FF,$2C,$47,$F8,$BE,$3E,$1F,$DF
	.dc.b	$3E,$C3,$51,$27,$47,$FA,$96,$74,$E0,$FF,$1F,$46,$DF,$F6,$B4,$69
	.dc.b	$B3,$F1,$E3,$11,$FE,$0B,$6B,$FE,$3F,$6B,$FE,$0A,$FB,$EF,$41,$1F
	.dc.b	$F5,$B7,$B0,$F8,$DF,$B6,$D8,$FF,$13,$35,$FB,$0D,$7F,$C3,$4D,$7F
	.dc.b	$C9,$DE,$7F,$F1,$8F,$1F,$FB,$47,$13,$D2,$FC,$35,$7F,$B1,$5A,$FF
	.dc.b	$BC,$EB,$87,$27,$2E,$AF,$C6,$D6,$B0,$E6,$D9,$D5,$FF,$91,$2B,$F2
	.dc.b	$77,$CA,$3F,$CA,$71,$F9,$6F,$BE,$78,$F4,$FC,$83,$89,$FF,$22,$35
	.dc.b	$FC,$B3,$5F,$F1,$C7,$DD,$BE,$73,$F4,$2D,$8F,$F7,$03,$56,$7F,$0B
	.dc.b	$EF,$9E,$B7,$0B,$95,$C1,$FD,$5C,$EA,$3F,$D4,$D5,$FE,$11,$E9,$7D
	.dc.b	$F3,$EC,$FF,$3E,$DA,$DE,$FF,$EA,$DA,$DD,$7E,$DB,$DC,$FF,$43,$D6
	.dc.b	$8F,$F7,$5D,$F1,$7C,$F7,$C6,$31,$87,$85,$B1,$FE,$8B,$AE,$75,$C0
	.dc.b	$FF,$68,$7D,$F7,$EC,$BE,$14,$D3,$F8,$4C,$63,$DB,$B0,$7F,$AD,$46
	.dc.b	$4F,$EB,$91,$96,$C6,$A3,$11,$FE,$95,$3F,$D9,$8E,$1C,$35,$FF,$4D
	.dc.b	$35,$FF,$68,$EE,$F5,$3E,$D9,$23,$E3,$73,$3E,$F9,$EB,$66,$7E,$04
	.dc.b	$AF,$94,$7E,$0A,$FF,$92,$DF,$08,$E7,$F0,$8F,$BB,$FE,$2E,$3E,$FA
	.dc.b	$9F,$CB,$23,$53,$C2,$72,$FD,$9E,$3E,$55,$B9,$FA,$1F,$2D,$2B,$5B
	.dc.b	$A3,$E8,$FD,$EB,$EF,$A9,$F4,$1C,$57,$F8,$2D,$AF,$CD,$0E,$1F,$A0
	.dc.b	$D7,$FD,$8C,$D7,$FD,$FC,$D7,$8C,$7F,$3F,$F9,$17,$DE,$FF,$38,$F4
	.dc.b	$ED,$7B,$5A,$FC,$E9,$BD,$1F,$C3,$1B,$CF,$F9,$FA,$B4,$DB,$FC,$81
	.dc.b	$87,$F5,$6B,$CC,$7E,$AB,$E3,$BF,$52,$4A,$FB,$FF,$26,$FB,$6C,$47
	.dc.b	$E7,$6B,$83,$F4,$7A,$D7,$FD,$8A,$F8,$4F,$EB,$74,$7D,$E7,$F8,$81
	.dc.b	$FF,$45,$7B,$3F,$EC,$0F,$6B,$FB,$5B,$6B,$85,$6F,$7F,$97,$C7,$F5
	.dc.b	$49,$3F,$55,$AF,$D5,$E2,$46,$23,$EA,$BF,$DB,$7E,$1B,$CB,$1E,$BE
	.dc.b	$5F,$E9,$99,$FF,$0D,$6E,$09,$75,$FF,$BA,$B8,$7F,$BF,$E1,$CB,$0E
	.dc.b	$1E,$6F,$E6,$8C,$CF,$F7,$2F,$BE,$FF,$78,$E1,$F9,$76,$BF,$2D,$B5
	.dc.b	$FD,$1B,$C4,$FE,$60,$FF,$1F,$8C,$E5,$FC,$6D,$B5,$F7,$5B,$97,$F3
	.dc.b	$3F,$1F,$F8,$87,$3C,$AF,$D0,$BF,$C4,$3E,$FB,$DB,$1E,$D7,$FB,$2A
	.dc.b	$69,$34,$FE,$26,$D8,$FF,$37,$35,$FF,$0D,$78,$2F,$A4,$FF,$08,$FC
	.dc.b	$03,$E8,$69,$FE,$26,$9D,$57,$F3,$1E,$6F,$C2,$FC,$7F,$F9,$E5,$B1
	.dc.b	$FA,$F6,$BF,$C7,$5A,$FD,$53,$EA,$FD,$17,$E0,$1F,$E2,$ED,$7F,$8A
	.dc.b	$B7,$87,$FD,$B5,$F3,$6F,$BE,$7C,$3F,$F8,$EB,$CD,$7C,$8D,$26,$3F
	.dc.b	$D1,$EF,$FD,$67,$F7,$DF,$E4,$26,$C7,$F0,$56,$BF,$CC,$DA,$F9,$DA
	.dc.b	$FE,$35,$A9,$A4,$F8,$14,$FF,$8F,$BE,$F5,$F3,$9F,$E3,$E6,$DF,$F2
	.dc.b	$36,$BF,$CD,$3E,$F9,$F2,$1F,$E5,$66,$C7,$F1,$BE,$A4,$62,$3F,$CA
	.dc.b	$8D,$7E,$74,$6A,$7F,$CA,$D3,$99,$FD,$34,$E5,$E9,$19,$7F,$52,$DF
	.dc.b	$21,$7F,$3A,$7D,$1A,$7A,$CF,$F3,$5F,$DF,$7E,$1A,$3F,$A6,$B5,$F1
	.dc.b	$B5,$F0,$B5,$FD,$BB,$70,$34,$9D,$1F,$E6,$F7,$86,$1F,$C9,$E3,$FC
	.dc.b	$D2,$F8,$78,$BF,$C2,$0D,$EE,$F7,$FD,$6F,$F0,$16,$F8,$3F,$E7,$98
	.dc.b	$FE,$1B,$97,$C8,$6B,$A3,$E8,$38,$7F,$86,$B7,$23,$4D,$5F,$F3,$BF
	.dc.b	$BE,$F2,$92,$BF,$BA,$FD,$F7,$D9,$CB,$FC,$9E,$FD,$1C,$0D,$27,$07
	.dc.b	$FA,$3D,$B5,$F8,$36,$BF,$E0,$BC,$1F,$43,$FA,$68,$FE,$E0,$D7,$FA
	.dc.b	$23,$70,$FF,$2F,$FD,$8D,$D7,$F0,$AE,$EB,$F1,$6D,$73,$D1,$FE,$A1
	.dc.b	$8C,$47,$F3,$06,$E6,$69,$27,$FD,$43,$87,$FB,$E3,$0F,$F2,$33,$70
	.dc.b	$FF,$21,$37,$0F,$E7,$7F,$7C,$F8,$2F,$F5,$CB,$63,$F4,$8D,$7F,$D6
	.dc.b	$CD,$7F,$46,$D7,$F4,$3F,$7B,$FB,$0C,$3F,$CF,$ED,$C3,$3C,$FF,$EC
	.dc.b	$89,$7D,$F7,$F8,$9A,$FF,$C5,$37,$B7,$FA,$7B,$75,$39,$AE,$3E,$D7
	.dc.b	$FA,$FB,$E8,$7F,$CC,$D7,$FD,$D5,$B3,$FE,$67,$6E,$5F,$B6,$FA,$4F
	.dc.b	$ED,$95,$FE,$05,$9C,$B7,$E0,$BA,$1F,$ED,$C9,$DE,$4F,$BA,$FF,$15
	.dc.b	$57,$F9,$11,$AF,$F9,$09,$B8,$FF,$A8,$3E,$FB,$E5,$3F,$EA,$28,$CE
	.dc.b	$5F,$C2,$1B,$5F,$DB,$9B,$6B,$E3,$5F,$C3,$9F,$A1,$5F,$DD,$1B,$3B
	.dc.b	$E0,$FF,$75,$46,$2F,$FD,$B5,$87,$87,$E3,$FF,$DE,$F1,$8C,$7F,$8C
	.dc.b	$FD,$E3,$E5,$73,$53,$3F,$73,$FE,$FE,$9D,$47,$FA,$24,$7D,$63,$EF
	.dc.b	$BF,$D3,$F1,$FE,$47,$8C,$65,$FE,$49,$8D,$CE,$9F,$F3,$C4,$E9,$E1
	.dc.b	$F7,$CE,$1E,$12,$8C,$4E,$A7,$31,$88,$ED,$F3,$3F,$E8,$A4,$D2,$3F
	.dc.b	$E8,$7E,$0F,$B9,$CC,$DB,$EE,$36,$7F,$CF,$BF,$9C,$7F,$66,$9C,$92
	.dc.b	$3F,$01,$3B,$D3,$FD,$BF,$5B,$9C,$D7,$FA,$2A,$1F,$41,$D6,$7F,$A6
	.dc.b	$C6,$27,$FA,$A3,$63,$CC,$D7,$FA,$9F,$E3,$7F,$EE,$B9,$FF,$4A,$B6
	.dc.b	$3C,$F5,$E3,$9F,$EA,$B3,$9B,$FF,$79,$B6,$33,$F2,$5B,$BF,$F1,$FE
	.dc.b	$1E,$1B,$DC,$E7,$BF,$FD,$B7,$F5,$56,$4D,$FA,$EC,$33,$1F,$F4,$B4
	.dc.b	$E9,$B3,$9C,$7F,$CD,$9F,$8D,$7D,$D9,$FF,$4E,$36,$33,$F2,$DE,$18
	.dc.b	$C5,$FF,$B8,$DB,$5F,$ED,$F6,$D7,$FA,$05,$57,$1E,$1F,$55,$3E,$8F
	.dc.b	$E9,$5F,$FB,$9E,$30,$7F,$A7,$F3,$FC,$24,$EA,$7F,$D1,$93,$A7,$FA
	.dc.b	$37,$13,$F1,$BF,$8C,$D9,$FF,$5B,$1B,$CF,$FD,$F5,$19,$AF,$F4,$67
	.dc.b	$E3,$4F,$A0,$ED,$92,$31,$87,$8D,$B5,$FD,$55,$AF,$F9,$0E,$78,$57
	.dc.b	$29,$CC,$FF,$9E,$E7,$35,$CE,$BB,$FC,$CB,$EF,$3E,$F3,$67,$FD,$9D
	.dc.b	$87,$83,$2F,$D2,$60,$AD,$63,$FE,$A5,$E7,$FF,$17,$E1,$F2,$5B,$1F
	.dc.b	$DA,$DA,$F9,$DA,$F9,$5A,$FF,$AE,$5B,$5F,$E9,$2F,$2F,$E2,$C6,$3E
	.dc.b	$F7,$FA,$D5,$FF,$B8,$DB,$7E,$3A,$FC,$C4,$63,$1F,$4F,$E3,$9F,$BF
	.dc.b	$BF,$BC,$DB,$FF,$00,$35,$F9,$8D,$7E,$53,$5F,$F6,$73,$71,$34,$C7
	.dc.b	$E4,$2F,$A2,$FD,$53,$6F,$FD,$F5,$3E,$1C,$3E,$6C,$E7,$0F,$F6,$87
	.dc.b	$AC,$FC,$4B,$F0,$AD,$C3,$F2,$CD,$9F,$F0,$C3,$5F,$C2,$35,$FC,$13
	.dc.b	$5F,$F7,$33,$72,$FF,$2A,$D7,$E4,$BE,$F1,$F6,$FF,$C0,$5B,$87,$F9
	.dc.b	$BE,$71,$3F,$C1,$7E,$AD,$F5,$FF,$9A,$6D,$7E,$AD,$B9,$7F,$8B,$5A
	.dc.b	$FE,$65,$AF,$E5,$DA,$FF,$BE,$DB,$89,$BF,$94,$3E,$CF,$F8,$B3,$6F
	.dc.b	$FD,$B5,$B7,$E6,$2B,$C2,$FE,$6A,$75,$87,$FB,$F7,$F2,$67,$C6,$7E
	.dc.b	$C9,$B8,$7F,$03,$6C,$FF,$91,$5A,$FE,$C5,$AF,$EB,$DA,$B2,$46,$1B
	.dc.b	$5F,$C0,$3D,$CF,$C0,$63,$FA,$E6,$E1,$CE,$7F,$99,$46,$30,$7C,$3B
	.dc.b	$F2,$7F,$E1,$D7,$FB,$FF,$CA,$17,$FC,$F6,$B5,$E0,$7F,$3B,$9C,$D7
	.dc.b	$C6,$38,$7C,$17,$F1,$28,$FF,$61,$C7,$F2,$36,$E3,$FE,$65,$6B,$FC
	.dc.b	$47,$F0,$17,$B1,$FD,$1D,$B5,$D6,$7E,$14,$7F,$08,$D7,$F8,$5B,$A8
	.dc.b	$47,$F7,$1F,$71,$0F,$AF,$FE,$65,$18,$6C,$66,$FF,$DC,$8F,$09,$D4
	.dc.b	$62,$3F,$CF,$66,$CD,$C6,$9A,$64,$FF,$77,$39,$7F,$97,$7C,$BF,$E2
	.dc.b	$2F,$F9,$1B,$6B,$C1,$3D,$67,$3C,$BF,$DA,$7F,$00,$7D,$3D,$31,$F8
	.dc.b	$2D,$C3,$E3,$72,$AB,$71,$86,$E3,$F1,$A3,$F9,$E4,$2B,$94,$E4,$FC
	.dc.b	$87,$DF,$7F,$21,$AF,$E8,$4F,$E8,$9B,$EE,$FF,$7D,$8C,$4F,$08,$C7
	.dc.b	$5E,$5B,$5F,$E4,$1F,$C8,$1F,$57,$FE,$09,$F2,$1F,$90,$3E,$E3,$FC
	.dc.b	$19,$92,$F7,$87,$E0,$EF,$75,$A9,$CD,$7A,$67,$35,$DF,$0E,$D8,$75
	.dc.b	$8E,$4D,$AF,$81,$B3,$E9,$7F,$47,$1F,$D5,$E7,$5C,$5F,$D6,$A7,$3E
	.dc.b	$5B,$E7,$FE,$C6,$3F,$1B,$E5,$FF,$C4,$93,$9E,$8F,$8D,$18,$AD,$E5
	.dc.b	$F3,$A3,$13,$AB,$FB,$7F,$91,$F8,$1F,$EE,$4D,$AF,$E0,$8B,$E4,$FF
	.dc.b	$C5,$CE,$75,$34,$91,$AE,$1F,$F1,$C7,$57,$F9,$9F,$EA,$FF,$4B,$F0
	.dc.b	$06,$93,$2F,$D1,$4E,$DB,$3B,$6C,$EF,$2F,$F7,$17,$3B,$EA,$BF,$0C
	.dc.b	$78,$2F,$84,$FD,$09,$D5,$FE,$6B,$D0,$FB,$7C,$7F,$E0,$85,$F3,$7F
	.dc.b	$E4,$B7,$33,$FE,$15,$72,$F1,$E6,$D4,$ED,$B1,$FC,$2F,$91,$F5,$9F
	.dc.b	$D5,$32,$F1,$7C,$5F,$D8,$8C,$65,$FE,$18,$CD,$39,$F3,$BF,$40,$DC
	.dc.b	$7F,$63,$3A,$7F,$8C,$70,$7F,$A4,$FE,$4F,$E6,$92,$3E,$4B,$72,$FE
	.dc.b	$17,$BA,$31,$1F,$C2,$A3,$13,$FE,$4D,$DD,$F1,$BF,$0C,$F9,$97,$67
	.dc.b	$C7,$F8,$C7,$ED,$7F,$14,$FD,$B3,$63,$27,$A2,$32,$D3,$E9,$78,$9F
	.dc.b	$3F,$C3,$FA,$0D,$8C,$9F,$5E,$32,$D3,$ED,$7D,$77,$F9,$53,$23,$49
	.dc.b	$1E,$BF,$A3,$FB,$87,$DD,$8B,$1F,$83,$F8,$EF,$C0,$18,$DB,$F6,$F6
	.dc.b	$8C,$CE,$CF,$C5,$5E,$2B,$7E,$B7,$8C,$93,$F2,$DF,$37,$C0,$FE,$34
	.dc.b	$FC,$73,$63,$F2,$9F,$1F,$EB,$23,$F4,$DF,$77,$A7,$F9,$53,$F2,$CD
	.dc.b	$8F,$CE,$45,$BF,$4C,$DE,$7F,$CF,$4D,$AF,$CE,$4D,$AF,$F0,$16,$7F
	.dc.b	$E6,$99,$FC,$17,$8F,$F3,$48,$FE,$16,$7F,$4C,$7E,$9E,$2C,$7F,$B2
	.dc.b	$FC,$9F,$F4,$A7,$5B,$E2,$7F,$CC,$37,$FE,$CC,$9C,$C6,$23,$FB,$5C
	.dc.b	$EE,$72,$4A,$FF,$68,$BF,$E9,$2E,$97,$9D,$FD,$14,$FE,$47,$87,$F4
	.dc.b	$BC,$3F,$A6,$C7,$F2,$6F,$D6,$BD,$91,$FE,$F0,$CF,$BC,$63,$E1,$3F
	.dc.b	$CC,$52,$CD,$3F,$DF,$B3,$FE,$FF,$5A,$9C,$E9,$FE,$F3,$8C,$57,$3C
	.dc.b	$3A,$CE,$5E,$F8,$53,$B4,$ED,$B3,$9B,$EB,$87,$82,$F8,$B6,$B6,$D9
	.dc.b	$DC,$7B,$4C,$95,$F5,$BD,$6F,$C1,$0F,$27,$FD,$2B,$8F,$B5,$B8,$7D
	.dc.b	$47,$FC,$CB,$19,$D7,$CD,$1A,$C3,$E4,$36,$FA,$79,$DA,$A9,$97,$9E
	.dc.b	$B4,$74,$C7,$B4,$FC,$5C,$1F,$5E,$FA,$CF,$F5,$E4,$7B,$4C,$B8,$CF
	.dc.b	$EC,$FB,$0F,$A5,$FF,$71,$47,$DC,$AD,$EF,$F2,$5A,$73,$C3,$B4,$7C
	.dc.b	$DB,$E5,$EC,$F8,$E7,$35,$BC,$7C,$6F,$F8,$F2,$32,$D8,$CB,$FE,$46
	.dc.b	$3B,$9B,$FA,$0F,$F0,$93,$FE,$41,$94,$62,$35,$E6,$7F,$91,$5F,$87
	.dc.b	$19,$7A,$4F,$09,$F6,$36,$3D,$93,$DB,$2F,$8B,$EF,$FD,$F9,$DC,$E7
	.dc.b	$6E,$6D,$8F,$F2,$ED,$F3,$BE,$3B,$FA,$DF,$D1,$7E,$3B,$FD,$2E,$BF
	.dc.b	$24,$D7,$F2,$0D,$8D,$FC,$7F,$1C,$7F,$7F,$39,$6F,$8B,$9B,$FD,$94
	.dc.b	$E5,$B5,$9E,$54,$E5,$B5,$98,$C0,$9E,$1F,$96,$BF,$04,$A4,$C3,$F5
	.dc.b	$51,$FE,$15,$6C,$65,$FD,$54,$71,$77,$E8,$E7,$3F,$E0,$93,$99,$DD
	.dc.b	$78,$2B,$59,$FF,$1E,$AD,$5F,$0A,$D5,$FE,$F7,$D4,$76,$FF,$8C,$63
	.dc.b	$FC,$03,$E9,$FE,$2F,$C1,$FF,$04,$37,$AF,$A7,$D4,$5A,$FE,$35,$9A
	.dc.b	$3F,$15,$C2,$8C,$4F,$FA,$46,$31,$3A,$9C,$CF,$FA,$62,$72,$EE,$32
	.dc.b	$FE,$76,$32,$EE,$32,$FF,$93,$3B,$7C,$99,$FF,$92,$1B,$EA,$FF,$BD
	.dc.b	$27,$FC,$A7,$DD,$83,$F4,$15,$FE,$4E,$6C,$9A,$63,$97,$99,$2B,$FC
	.dc.b	$BB,$EA,$D7,$FC,$E3,$5F,$A1,$6C,$F8,$A7,$5B,$35,$FF,$31,$CE,$1F
	.dc.b	$F3,$2C,$EE,$BF,$D7,$18,$FF,$98,$A3,$CA,$48,$CE,$1F,$E3,$C7,$FC
	.dc.b	$DF,$E8,$FC,$03,$E0,$F9,$B2,$36,$FB,$88,$F2,$F3,$BE,$39,$1F,$F3
	.dc.b	$CF,$D6,$57,$DB,$17,$4F,$54,$7F,$9E,$A7,$E7,$91,$E2,$7F,$A0,$BE
	.dc.b	$37,$E9,$9F,$38,$8F,$A9,$FA,$3F,$21,$F6,$3E,$06,$4F,$FA,$2A,$3E
	.dc.b	$D9,$87,$E7,$C6,$1F,$F6,$4F,$D6,$FE,$97,$0F,$2E,$C8,$FC,$7C,$61
	.dc.b	$BB,$FF,$BB,$A7,$5D,$1F,$8A,$9E,$53,$97,$76,$AE,$9C,$DB,$BF,$C7
	.dc.b	$FC,$8C,$AF,$F5,$0B,$FA,$77,$FD,$C3,$E8,$5F,$50,$FF,$31,$8C,$9F
	.dc.b	$CC,$E3,$2D,$8D,$46,$23,$FD,$F8,$D7,$FE,$00,$B5,$FE,$88,$D7,$FC
	.dc.b	$21,$EC,$FC,$37,$E3,$F7,$FB,$CF,$DE,$75,$FF,$A3,$34,$B5,$BA,$FF
	.dc.b	$66,$C7,$F1,$96,$D6,$A3,$8F,$DA,$32,$BC,$32,$FF,$D3,$13,$96,$CE
	.dc.b	$5F,$F7,$C6,$29,$57,$4A,$FA,$9F,$25,$2F,$EF,$BF,$37,$EB,$DF,$37
	.dc.b	$4F,$98,$E2,$7F,$AF,$35,$FE,$D0,$03,$97,$2D,$7F,$B7,$35,$E0,$D7
	.dc.b	$8C,$7F,$0C,$F1,$DE,$AF,$F1,$46,$CF,$7E,$5F,$CA,$C6,$23,$FD,$F2
	.dc.b	$FF,$A0,$DD,$72,$FE,$56,$35,$3C,$2B,$A1,$D7,$D7,$FB,$3F,$02,$F5
	.dc.b	$FF,$35,$F1,$73,$2F,$F6,$44,$E5,$B3,$97,$EE,$CF,$28,$FC,$C7,$D6
	.dc.b	$BE,$CB,$F9,$6F,$E4,$AE,$8A,$6A,$3F,$3D,$F8,$E3,$F0,$8C,$65,$EA
	.dc.b	$3E,$76,$1C,$DC,$09,$E0,$E1,$AB,$6F,$E1,$EA,$79,$39,$E0,$FF,$11
	.dc.b	$46,$27,$75,$C9,$FE,$01,$E9,$7E,$F7,$D6,$98,$7F,$08,$8F,$E0,$06
	.dc.b	$1F,$CF,$C6,$1F,$57,$80,$FB,$3E,$D9,$7F,$09,$6C,$FF,$05,$6C,$EF
	.dc.b	$EC,$3F,$AA,$C6,$09,$AE,$23,$FA,$CF,$BB,$F8,$1B,$F6,$1F,$C3,$DC
	.dc.b	$67,$FC,$71,$B8,$FF,$90,$7F,$30,$FF,$75,$D7,$CC,$6B,$F9,$E0,$E1
	.dc.b	$FD,$0B,$5E,$5F,$09,$FC,$4E,$31,$1F,$A5,$6B,$CA,$34,$7F,$6A,$F7
	.dc.b	$BF,$82,$61,$FC,$C1,$B7,$FC,$B5,$B5,$FC,$A5,$FE,$5A,$FF,$2B,$F9
	.dc.b	$17,$C5,$7F,$74,$6C,$7F,$62,$6B,$FB,$8D,$91,$88,$FF,$31,$7B,$EF
	.dc.b	$CA,$9E,$7F,$F8,$13,$A7,$FB,$35,$7C,$CC,$3F,$4D,$E8,$7F,$86,$F0
	.dc.b	$E1,$7F,$E0,$AF,$95,$FF,$1E,$5F,$E0,$1A,$F1,$6E,$DF,$D3,$97,$2F
	.dc.b	$13,$FC,$1A,$E2,$7F,$BD,$B5,$F5,$35,$FF,$3D,$62,$8C,$47,$F5,$56
	.dc.b	$BF,$D4,$5B,$1B,$9F,$F0,$DF,$CB,$7E,$12,$BF,$AB,$36,$4D,$F8,$0F
	.dc.b	$E6,$53,$F0,$DA,$F8,$9A,$FF,$65,$6B,$FD,$85,$B2,$69,$3C,$6F,$F1
	.dc.b	$6B,$AF,$98,$7C,$A7,$F8,$D9,$B7,$FA,$C6,$BF,$57,$E6,$5E,$AF,$F8
	.dc.b	$F1,$B1,$FE,$24,$6B,$F8,$36,$BF,$29,$AF,$F7,$D6,$BF,$DE,$5B,$89
	.dc.b	$A6,$3E,$6B,$F2,$B0,$69,$8F,$7B,$F2,$7B,$FF,$83,$9F,$C9,$5F,$F8
	.dc.b	$2D,$AF,$F8,$1D,$B4,$69,$39,$BF,$CA,$6D,$9F,$F2,$0C,$BC,$6F,$F2
	.dc.b	$EB,$63,$F8,$9B,$5F,$CE,$79,$FF,$7C,$FE,$14,$FE,$31,$AF,$F8,$7D
	.dc.b	$AF,$F8,$6D,$B0,$69,$3C,$8F,$F3,$3B,$AF,$9D,$7D,$07,$F9,$C1,$B7
	.dc.b	$FC,$99,$AF,$EB,$FE,$09,$F3,$5F,$E7,$66,$C7,$F9,$79,$AF,$EF,$5A
	.dc.b	$FE,$91,$AF,$F8,$DD,$AF,$F8,$CD,$B8,$9A,$63,$E4,$9E,$1F,$FA,$19
	.dc.b	$B5,$FC,$F1,$AF,$F0,$5F,$CA,$9F,$81,$FF,$A3,$1B,$1F,$E7,$46,$BF
	.dc.b	$C4,$DA,$FE,$E9,$AF,$F9,$37,$B7,$FC,$93,$89,$A4,$F1,$FF,$D3,$2E
	.dc.b	$3D,$CA,$62,$3F,$D3,$BF,$29,$FD,$C6,$FF,$A9,$B5,$FE,$38,$DE,$5F
	.dc.b	$F3,$0A,$F9,$4E,$3E,$84,$AE,$2F,$48,$FF,$20,$FD,$1B,$CE,$A4,$AF
	.dc.b	$EF,$5C,$AF,$92,$FF,$64,$47,$F8,$E2,$B8,$3A,$6B,$9A,$FF,$26,$CF
	.dc.b	$0C,$BF,$D1,$37,$CF,$A7,$FD,$93,$3D,$2F,$FD,$80,$D7,$FC,$6D,$3B
	.dc.b	$7F,$D9,$9F,$51,$FE,$24,$F3,$32,$D8,$FC,$D3,$67,$D1,$F1,$9F,$E5
	.dc.b	$E8,$C1,$3D,$7F,$2C,$7F,$9D,$FE,$9F,$FC,$97,$9F,$F6,$9C,$11,$87
	.dc.b	$FB,$63,$5F,$F2,$5B,$5F,$F2,$61,$C3,$C6,$70,$F3,$9F,$F3,$33,$63
	.dc.b	$33,$D1,$AF,$87,$AB,$FD,$0C,$FE,$74,$FF,$44,$38,$3E,$1F,$D3,$FF
	.dc.b	$AD,$30,$F9,$B9,$A3,$1A,$FC,$AD,$9F,$5B,$8C,$7E,$3D,$5D,$D8,$71
	.dc.b	$C3,$F0,$2D,$AF,$F3,$BE,$1C,$A7,$37,$C6,$1F,$51,$FE,$38,$AF,$03
	.dc.b	$5F,$E1,$D1,$AA,$FF,$0B,$61,$F8,$DA,$D2,$E3,$7C,$8F,$95,$B5,$F9
	.dc.b	$4B,$E5,$87,$E4,$EB,$44,$9D,$61,$67,$FC,$05,$A7,$F1,$A9,$FF,$1E
	.dc.b	$B6,$33,$F5,$1F,$17,$AF,$FC,$A5,$E1,$5A,$D1,$FE,$CE,$2D,$FF,$B2
	.dc.b	$D5,$6E,$79,$7D,$47,$FA,$96,$BF,$C9,$6D,$C3,$FC,$28,$FF,$A8,$23
	.dc.b	$3C,$5C,$08,$F0,$F8,$3B,$5F,$E1,$B8,$FF,$08,$9F,$E3,$FC,$8E,$23
	.dc.b	$FC,$50,$D7,$FC,$1E,$D7,$F9,$67,$CD,$FB,$F9,$0F,$C1,$9E,$5A,$7F
	.dc.b	$BB,$27,$31,$FC,$FF,$E1,$3E,$E6,$0F,$19,$18,$6C,$F9,$5B,$8F,$F8
	.dc.b	$ED,$AF,$F8,$89,$AF,$FB,$D5,$AF,$FB,$C5,$B4,$F9,$97,$C6,$FD,$76
	.dc.b	$DF,$94,$FF,$3C,$B8,$BF,$EC,$FF,$51,$E8,$E6,$34,$95,$F2,$E3,$2D
	.dc.b	$8C,$CF,$F9,$4E,$72,$FF,$95,$7E,$A1,$F2,$9F,$89,$6D,$FF,$A2,$AB
	.dc.b	$7B,$7D,$6B,$66,$BD,$C7,$F9,$5F,$CF,$FB,$78,$7D,$D6,$C7,$F9,$8D
	.dc.b	$AF,$F9,$29,$AF,$FB,$76,$BF,$C3,$53,$F5,$A7,$33,$E0,$6C,$EA,$7F
	.dc.b	$C5,$DD,$FF,$A1,$F4,$0F,$E8,$EB,$C3,$8F,$A6,$FF,$C5,$53,$FA,$7F
	.dc.b	$4B,$FD,$C3,$87,$E4,$DB,$8F,$F9,$E5,$AF,$F9,$81,$AF,$9D,$AF,$C4
	.dc.b	$6D,$1A,$63,$EA,$1F,$41,$FB,$76,$DF,$F9,$37,$C4,$C4,$FE,$EF,$D6
	.dc.b	$7C,$BF,$E6,$1B,$5F,$A9,$6E,$5F,$E8,$F6,$BF,$E6,$59,$E0,$49,$E0
	.dc.b	$FE,$8B,$D8,$FF,$2F,$C7,$FB,$21,$B3,$FE,$6F,$9F,$E1,$91,$8D,$7F
	.dc.b	$D1,$5F,$13,$FA,$AF,$6C,$D2,$61,$FE,$8A,$6E,$1F,$E8,$A6,$BF,$E8
	.dc.b	$AF,$7B,$EB,$CF,$FB,$61,$B1,$FE,$99,$9F,$E3,$91,$8B,$FC,$97,$A2
	.dc.b	$FA,$8F,$DF,$E1,$F8,$9F,$81,$5F,$F4,$7A,$D7,$BD,$FC,$96,$73,$5F
	.dc.b	$EE,$67,$FC,$7E,$FF,$AE,$7D,$9F,$E0,$B1,$8C,$FF,$D7,$4D,$7F,$D4
	.dc.b	$3F,$50,$F9,$5F,$E6,$0D,$BF,$F7,$8B,$FB,$9D,$3F,$20,$E6,$BF,$DC
	.dc.b	$FE,$D1,$7D,$27,$F4,$F8,$C3,$63,$37,$FE,$E5,$9C,$B6,$32,$7F,$7A
	.dc.b	$11,$A7,$0D,$7F,$BF,$B5,$FD,$87,$B9,$F7,$66,$FC,$0F,$F6,$FA,$E5
	.dc.b	$8F,$FB,$EE,$3F,$A7,$7D,$43,$DA,$A6,$3D,$AB,$87,$B6,$FF,$61,$5B
	.dc.b	$8C,$1F,$BF,$71,$97,$FB,$7D,$FF,$21,$46,$A3,$F8,$7C,$6A,$7F,$75
	.dc.b	$F3,$BF,$C4,$1B,$D1,$FD,$86,$31,$7F,$C2,$DF,$D2,$CF,$0F,$13,$55
	.dc.b	$B6,$D7,$7C,$1F,$50,$FA,$6F,$F0,$87,$90,$FA,$87,$D1,$FF,$84,$EF
	.dc.b	$75,$A9,$CE,$1C,$F0,$E4,$FA,$F0,$E5,$87,$19,$CE,$0C,$FF,$97,$56
	.dc.b	$AF,$86,$1D,$2B,$51,$FC,$9F,$EA,$3F,$43,$1F,$D3,$67,$57,$FC,$7E
	.dc.b	$73,$E5,$7F,$AD,$79,$7E,$37,$D0,$FF,$8A,$7E,$1F,$F1,$37,$F3,$23
	.dc.b	$13,$AC,$BF,$CD,$75,$A7,$E6,$4F,$F8,$3F,$E8,$3B,$C7,$F3,$96,$E7
	.dc.b	$FD,$19,$FF,$65,$47,$DB,$F2,$51,$A6,$EF,$CC,$E9,$F0,$56,$BF,$05
	.dc.b	$D3,$FE,$54,$F9,$CF,$D3,$7C,$11,$A4,$9F,$D3,$46,$5B,$19,$CB,$F4
	.dc.b	$73,$97,$F5,$18,$B9,$DF,$43,$F8,$B3,$F4,$55,$BE,$0F,$C8,$C7,$F8
	.dc.b	$BF,$CE,$FB,$31,$FD,$35,$7C,$5E,$7A,$34,$DD,$3E,$14,$E9,$B3,$A7
	.dc.b	$FD,$33,$F3,$DF,$D1,$34,$FB,$E7,$F7,$0C,$DF,$1F,$FE,$61,$9D,$65
	.dc.b	$FA,$48,$C7,$9D,$FA,$36,$E1,$F4,$6F,$8B,$F3,$35,$7D,$27,$CC,$24
	.dc.b	$7C,$A6,$E5,$F2,$1A,$FF,$0A,$9E,$33,$FE,$3B,$D1,$F2,$7F,$14,$F9
	.dc.b	$37,$67,$C5,$F4,$5F,$B7,$F9,$2F,$DC,$36,$32,$79,$E3,$2D,$3E,$4F
	.dc.b	$89,$F4,$7C,$4F,$A4,$D8,$C9,$F3,$A3,$2D,$3E,$8F,$D4,$7F,$90,$B2
	.dc.b	$34,$91,$EB,$FA,$2F,$BC,$7D,$F8,$B1,$F6,$BE,$4F,$F0,$46,$36,$FA
	.dc.b	$FC,$FF,$E1,$28,$CC,$EC,$FC,$0D,$E2,$B7,$87,$F8,$20,$9E,$B7,$94
	.dc.b	$F3,$7C,$0F,$E3,$CF,$C8,$36,$FF,$17,$F5,$1E,$B2,$3F,$4D,$F7,$7A
	.dc.b	$7F,$97,$3F,$30,$D8,$FC,$9C,$5C,$3C,$73,$79,$FF,$39,$36,$BF,$C6
	.dc.b	$D8,$AB,$17,$FE,$44,$8F,$F2,$F4,$FE,$0F,$C7,$F9,$C4,$7F,$0D,$3F
	.dc.b	$A9,$3F,$55,$16,$3F,$D4,$DE,$4F,$FA,$B3,$E2,$BF,$C9,$B5,$FE,$EE
	.dc.b	$8C,$12,$7F,$D6,$73,$A9,$FF,$64,$4E,$5F,$DE,$F4,$3C,$EF,$E8,$C7
	.dc.b	$FB,$06,$6D,$FF,$4F,$F2,$1F,$51,$EA,$C1,$F0,$9F,$E4,$C5,$3A,$CF
	.dc.b	$FC,$C9,$39,$8E,$33,$8D,$3F,$BF,$46,$F0,$FF,$00,$56,$F0,$EC,$DC
	.dc.b	$38,$61,$DA,$7F,$C0,$D5,$C3,$0E,$71,$C1,$B5,$C3,$0E,$D7,$E8,$9D
	.dc.b	$B9,$AE,$0F,$5A,$C1,$96,$DF,$77,$DF,$E4,$7F,$9F,$B1,$F7,$37,$0F
	.dc.b	$6B,$FE,$52,$8C,$E5,$E3,$8D,$57,$28,$C7,$9D,$96,$61,$DF,$35,$F0
	.dc.b	$6D,$6A,$FD,$87,$F8,$A7,$E8,$1F,$51,$FE,$AD,$8C,$88,$CE,$9E,$BF
	.dc.b	$61,$F3,$5F,$EC,$FD,$7E,$E5,$EF,$0E,$51,$8C,$38,$CF,$46,$C7,$E0
	.dc.b	$EF,$9F,$B3,$F6,$2B,$96,$3F,$25,$A7,$F8,$D4,$A7,$F8,$DE,$FA,$B7
	.dc.b	$0F,$53,$9F,$94,$FF,$07,$BC,$65,$18,$8D,$79,$DD,$5F,$81,$19,$7F
	.dc.b	$C9,$C6,$23,$C9,$3C,$9B,$3D,$B2,$F6,$7B,$FF,$4A,$F3,$51,$F8,$55
	.dc.b	$BE,$2F,$F2,$D6,$8B,$E1,$5F,$5B,$77,$D4,$7F,$4E,$8F,$BC,$D7,$4D
	.dc.b	$CB,$F5,$4B,$E4,$3D,$31,$8F,$1A,$E5,$FA,$CA,$E0,$D8,$CD,$FE,$DA
	.dc.b	$79,$3E,$08,$CE,$1E,$46,$CF,$29,$E1,$86,$BE,$A1,$F5,$14,$93,$F9
	.dc.b	$78,$FF,$09,$36,$32,$FC,$18,$E2,$EF,$D1,$CE,$7F,$C1,$E7,$33,$BA
	.dc.b	$F0,$56,$B3,$FE,$13,$5A,$BE,$15,$AB,$FD,$B7,$89,$FE,$19,$F4,$FE
	.dc.b	$FF,$83,$FE,$66,$6F,$57,$F9,$8D,$B5,$FC,$6B,$17,$CC,$E3,$81,$18
	.dc.b	$9F,$CC,$46,$27,$F1,$CD,$8C,$CE,$DF,$F4,$94,$65,$B1,$97,$F2,$71
	.dc.b	$96,$C6,$5F,$F2,$57,$57,$9B,$2F,$F2,$3B,$7D,$5F,$F7,$1C,$FF,$92
	.dc.b	$D7,$6E,$4F,$D1,$57,$F9,$39,$B1,$A9,$FD,$94,$E5,$B1,$F9,$8D,$1A
	.dc.b	$7F,$96,$FB,$7C,$07,$87,$D4,$B7,$E1,$8A,$FF,$80,$2F,$FC,$92,$E4
	.dc.b	$86,$63,$93,$E9,$9F,$29,$1F,$1B,$E8,$F4,$3E,$01,$23,$E0,$79,$BF
	.dc.b	$0C,$F6,$91,$F5,$BF,$03,$CE,$F9,$04,$8F,$91,$F3,$95,$F4,$45,$CF
	.dc.b	$E0,$C7,$F9,$E6,$7E,$81,$1E,$27,$D1,$F8,$CF,$A8,$7C,$E2,$3F,$31
	.dc.b	$FA,$3F,$1F,$F6,$7E,$06,$4F,$D9,$8F,$B8,$61,$F9,$51,$87,$FD,$11
	.dc.b	$F5,$1F,$A6,$C3,$CB,$C0,$8F,$C8,$46,$1B,$C1,$FE,$ED,$9D,$73,$FE
	.dc.b	$2E,$79,$4E,$5D,$DA,$BA,$73,$6E,$FF,$21,$F2,$32,$BF,$20,$FE,$A9
	.dc.b	$FF,$71,$7A,$17,$D4,$3F,$CC,$E3,$27,$F3,$58,$CB,$63,$51,$88,$FE
	.dc.b	$8C,$D7,$FD,$F8,$D7,$F9,$DB,$5F,$F0,$9F,$B1,$F0,$DF,$8F,$DD,$EF
	.dc.b	$3F,$79,$19,$8F,$F4,$5F,$5A,$31,$B7,$C9,$AE,$15,$FE,$92,$D7,$F8
	.dc.b	$27,$D4,$33,$BC,$32,$BF,$C5,$13,$96,$CE,$5E,$F0,$9E,$73,$F4,$FE
	.dc.b	$4A,$57,$DE,$7E,$67,$D4,$3E,$A2,$9F,$2D,$C4,$FF,$5F,$6B,$FD,$B5
	.dc.b	$AF,$F6,$00,$E1,$FE,$C6,$D4,$6A,$3B,$1F,$C2,$BC,$A7,$AB,$FC,$4D
	.dc.b	$B3,$FE,$DB,$6B,$FE,$80,$5E,$BE,$E6,$4E,$DE,$51,$FC,$BF,$E0,$1E
	.dc.b	$BF,$E5,$7E,$2E,$5F,$FB,$C2,$34,$61,$FF,$46,$EE,$9C,$C7,$AB,$87
	.dc.b	$F7,$2B,$E9,$9F,$3F,$FC,$AF,$EA,$1C,$F4,$FC,$D4,$EA,$B9,$76,$7F
	.dc.b	$82,$3E,$9D,$F5,$1A,$C6,$56,$8F,$93,$3A,$70,$1C,$3F,$ED,$A3,$C9
	.dc.b	$1C,$9C,$46,$B2,$FD,$D7,$A7,$FB,$7F,$5E,$57,$F0,$28,$FD,$D1,$87
	.dc.b	$E1,$C6,$1F,$27,$80,$FA,$8F,$64,$7F,$AD,$1B,$7F,$C2,$9B,$9F,$F0
	.dc.b	$75,$F5,$1F,$D4,$5C,$D7,$F0,$56,$DF,$5C,$F5,$5F,$D5,$FD,$CF,$C0
	.dc.b	$DF,$89,$AF,$FB,$11,$B8,$FD,$6F,$A8,$FF,$73,$CF,$C8,$6B,$FE,$47
	.dc.b	$6B,$F9,$B6,$BF,$E3,$C6,$BF,$9A,$6A,$69,$3E,$13,$FB,$2F,$BB,$FE
	.dc.b	$18,$9F,$F6,$BB,$6B,$F9,$6B,$74,$FE,$54,$FF,$28,$9F,$E4,$DF,$22
	.dc.b	$F8,$9F,$ED,$ED,$8F,$ED,$AD,$7F,$65,$AA,$31,$1F,$E5,$4F,$71,$F5
	.dc.b	$0F,$3F,$FC,$01,$D6,$F4,$29,$FE,$18,$AD,$CF,$F8,$17,$E5,$7F,$C7
	.dc.b	$B8,$7F,$BC,$1B,$B7,$F4,$66,$ED,$FD,$0D,$72,$F1,$3F,$C0,$EE,$2B
	.dc.b	$FC,$0A,$D7,$D2,$D7,$FC,$D7,$AA,$31,$1D,$67,$99,$1F,$F0,$7F,$CB
	.dc.b	$7E,$02,$FF,$AA,$36,$3F,$A8,$36,$FF,$CD,$5E,$37,$F8,$87,$DF,$FE
	.dc.b	$59,$5F,$29,$AF,$89,$AF,$F5,$F6,$BF,$D6,$DB,$3D,$23,$FC,$43,$18
	.dc.b	$33,$F3,$0F,$94,$FF,$18,$36,$7F,$56,$D5,$A7,$F7,$2F,$9A,$7A,$BF
	.dc.b	$E3,$76,$CF,$F8,$B9,$AF,$E0,$9A,$FC,$A6,$BF,$DD,$DA,$FF,$73,$6E
	.dc.b	$26,$98,$F9,$AF,$CA,$41,$A6,$3D,$EF,$C9,$61,$FE,$08,$7F,$23,$7F
	.dc.b	$E0,$56,$BF,$E0,$16,$D1,$A4,$E7,$FF,$26,$B6,$7F,$C9,$72,$F1,$FF
	.dc.b	$CB,$0D,$8F,$E2,$6D,$7E,$4F,$07,$F8,$57,$D2,$FC,$01,$FC,$1A,$3F
	.dc.b	$1A,$D7,$5E,$4F,$F9,$72,$31,$1F,$E1,$76,$D1,$A4,$9F,$F3,$1F,$CE
	.dc.b	$BE,$7F,$FC,$D8,$DB,$FE,$4E,$D7,$CA,$7F,$88,$FE,$05,$F3,$5F,$E7
	.dc.b	$36,$C7,$F9,$A9,$AF,$EF,$9A,$FE,$99,$AF,$F8,$CD,$AF,$58,$FF,$3B
	.dc.b	$C6,$31,$37,$E5,$1E,$1F,$FA,$0D,$B7,$FC,$F9,$AF,$A2,$9F,$4E,$FA
	.dc.b	$8F,$F4,$53,$67,$FD,$02,$D7,$F8,$AB,$5F,$DE,$B5,$FF,$25,$77,$7F
	.dc.b	$91,$A8,$D2,$78,$FF,$E9,$77,$1E,$E5,$31,$1F,$E9,$BF,$94,$FE,$E3
	.dc.b	$7F,$D5,$9A,$FA,$5B,$CD,$FE,$5E,$5F,$29,$C7,$D0,$95,$C5,$E9,$1F
	.dc.b	$EB,$9F,$A4,$79,$D4,$95,$FD,$DF,$99,$F2,$5F,$9D,$8C,$57,$06,$D7
	.dc.b	$07,$FC,$8F,$3B,$6F,$57,$FB,$0A,$31,$1F,$EB,$C6,$BF,$E2,$CB,$EF
	.dc.b	$3F,$EC,$8F,$A6,$FF,$13,$E9,$FE,$8E,$94,$63,$AB,$83,$6F,$F3,$98
	.dc.b	$3E,$2F,$FD,$AD,$39,$BE,$3E,$C6,$0F,$F6,$D5,$6E,$FF,$C2,$7F,$51
	.dc.b	$FE,$4B,$CB,$FC,$89,$1F,$E4,$08,$C3,$86,$BF,$E4,$96,$BF,$E4,$60
	.dc.b	$FF,$98,$F8,$39,$C6,$3C,$64,$62,$75,$3D,$67,$2F,$FA,$17,$AB,$FD
	.dc.b	$0C,$FE,$80,$C1,$0F,$F7,$3F,$07,$FB,$53,$6F,$AE,$7F,$AD,$B1,$F9
	.dc.b	$5F,$52,$F9,$8C,$4F,$D1,$AE,$17,$FE,$80,$6D,$FD,$16,$CF,$D7,$BE
	.dc.b	$4F,$63,$F1,$0F,$EA,$DF,$93,$B2,$31,$7F,$F0,$01,$AF,$FB,$F9,$AF
	.dc.b	$DC,$F1,$FC,$AF,$F1,$7E,$0F,$EA,$17,$DB,$AF,$38,$7E,$12,$ED,$6A
	.dc.b	$FF,$D3,$2D,$CB,$E0,$DE,$AB,$AF,$53,$ED,$1F,$27,$86,$E7,$FA,$3F
	.dc.b	$C5,$BF,$9F,$3F,$E4,$88,$CE,$7F,$C2,$DB,$97,$FB,$41,$B3,$F9,$DA
	.dc.b	$E3,$5B,$BF,$6B,$E1,$F9,$9B,$66,$FA,$58,$66,$1F,$E6,$89,$D3,$67
	.dc.b	$35,$FC,$D3,$17,$E8,$0F,$C1,$3F,$CF,$F3,$FE,$81,$8C,$5F,$E7,$AB
	.dc.b	$75,$AA,$FE,$F3,$5A,$7F,$9C,$74,$FC,$8E,$5F,$F4,$FE,$1F,$E7,$DC
	.dc.b	$38,$BF,$E6,$C8,$CD,$7F,$46,$F9,$97,$DA,$3E,$53,$B4,$61,$B1,$9B
	.dc.b	$FD,$8C,$E5,$B1,$9A,$FF,$5C,$4E,$5B,$19,$7F,$65,$19,$7F,$A7,$69
	.dc.b	$E9,$3F,$1B,$1C,$71,$5F,$F5,$19,$FC,$8A,$FC,$A7,$FD,$62,$FF,$04
	.dc.b	$31,$1F,$1D,$A6,$63,$77,$FE,$CD,$8C,$B6,$32,$FF,$07,$8C,$B6,$32
	.dc.b	$FF,$B4,$38,$9F,$52,$F9,$7F,$CC,$B6,$FF,$D5,$C7,$6C,$7D,$76,$BE
	.dc.b	$37,$EC,$C1,$F3,$FE,$96,$E1,$E9,$8C,$11,$C1,$2F,$D5,$1F,$87,$BF
	.dc.b	$2F,$CB,$79,$0E,$D1,$6B,$CB,$DB,$E7,$BF,$88,$D7,$E1,$35,$FF,$77
	.dc.b	$C7,$F1,$7C,$7E,$74,$E6,$7E,$BB,$67,$47,$D5,$75,$83,$EA,$7F,$C3
	.dc.b	$93,$FE,$C0,$6D,$7F,$86,$A3,$8B,$FE,$05,$8C,$E7,$F2,$A7,$F0,$5E
	.dc.b	$3F,$C7,$23,$F7,$DB,$5F,$75,$AF,$DB,$7F,$59,$E6,$7E,$3B,$EE,$FF
	.dc.b	$8E,$B6,$BE,$CE,$DF,$EF,$07,$FD,$D1,$19,$BF,$F7,$27,$A0,$DF,$F6
	.dc.b	$D0,$8C,$57,$F4,$E6,$C7,$F5,$A6,$BF,$D5,$4E,$13,$CA,$33,$1F,$D6
	.dc.b	$5A,$8D,$CF,$F5,$AF,$CF,$7F,$8E,$C6,$5A,$E8,$FF,$7D,$38,$E3,$7C
	.dc.b	$1F,$F2,$06,$D7,$5D,$7F,$2F,$7E,$BE,$1D,$5E,$FE,$B3,$DF,$81,$A4
	.dc.b	$C3,$FB,$6F,$C0,$BE,$93,$F9,$6B,$63,$B1,$FD,$D7,$C5,$FC,$23,$8C
	.dc.b	$BD,$D3,$B7,$FC,$BB,$DE,$FD,$5D,$F8,$4F,$F2,$58,$C3,$63,$37,$FE
	.dc.b	$08,$9C,$B6,$33,$5F,$46,$72,$D8,$CB,$FE,$09,$8C,$BF,$E0,$BF,$AF
	.dc.b	$7E,$B0,$FC,$23,$FB,$2C,$61,$FE,$CF,$5D,$38,$7F,$80,$CE,$71,$F7
	.dc.b	$75,$3E,$BF,$E8,$E5,$FE,$63,$6D,$FF,$08,$E7,$FC,$07,$8B,$FD,$9A
	.dc.b	$73,$87,$F1,$36,$DF,$F0,$D6,$D7,$09,$EF,$1A,$C3,$FC,$EB,$E4,$FF
	.dc.b	$70,$F6,$BF,$BA,$57,$E3,$6F,$F1,$DF,$59,$9A,$93,$E8,$AD,$E2,$BD
	.dc.b	$CE,$6B,$F9,$DF,$E9,$BF,$B8,$8F,$CC,$73,$A3,$17,$F9,$BF,$B0,$FE
	.dc.b	$AB,$B7,$F2,$6A,$D7,$BF,$C8,$98,$7E,$16,$D7,$C2,$3A,$E1,$D3,$57
	.dc.b	$D7,$BE,$5F,$FC,$1A,$DA,$FC,$76,$5D,$9F,$A7,$19,$C7,$FD,$29,$B7
	.dc.b	$F8,$9B,$4F,$E7,$38,$FF,$A7,$3D,$1F,$E0,$FE,$BF,$E6,$2F,$71,$F2
	.dc.b	$C9,$8F,$F0,$89,$D5,$F2,$C3,$A4,$FF,$4B,$FA,$7F,$F5,$3B,$F8,$46
	.dc.b	$E1,$FE,$BC,$E1,$FF,$2A,$FC,$CF,$F8,$9A,$73,$87,$91,$B7,$CF,$0F
	.dc.b	$2D,$6D,$FF,$1C,$D3,$F1,$5F,$D6,$63,$FE,$65,$8C,$77,$FF,$83,$DB
	.dc.b	$87,$F8,$4E,$B7,$5F,$C1,$5A,$A7,$6F,$F1,$B6,$BD,$63,$7D,$3F,$E7
	.dc.b	$2E,$EF,$D2,$FB,$47,$CB,$3F,$BF,$BF,$EC,$F5,$BC,$3F,$4F,$F5,$FF
	.dc.b	$CA,$A7,$E9,$B6,$3F,$31,$BB,$E8,$BF,$C9,$73,$9C,$FF,$B8,$37,$88
	.dc.b	$E7,$87,$FD,$03,$5C,$3C,$8F,$D2,$78,$93,$2F,$B5,$F0,$57,$2F,$C2
	.dc.b	$CD,$D3,$F6,$DA,$2B,$17,$FA,$0F,$B1,$FF,$51,$3F,$E2,$A6,$E1,$FE
	.dc.b	$28,$6B,$FE,$26,$6B,$C7,$E0,$3F,$C6,$B1,$88,$74,$A3,$0D,$9F,$E4
	.dc.b	$9F,$90,$FF,$78,$AF,$E5,$2D,$CB,$F5,$D7,$E2,$7E,$7C,$67,$0F,$F0
	.dc.b	$F7,$33,$C0,$FE,$0C,$7E,$83,$E7,$2C,$FF,$03,$8B,$8F,$E5,$F5,$56
	.dc.b	$31,$FC,$B7,$D0,$7E,$64,$FD,$4F,$83,$71,$FD,$04,$5C,$39,$4D,$AF
	.dc.b	$F6,$87,$E6,$7F,$E9,$49,$79,$68,$C3,$7B,$FF,$CD,$0D,$C7,$FD,$7F
	.dc.b	$E0,$3E,$93,$FC,$FE,$D8,$FF,$77,$5F,$ED,$30,$FE,$DD,$39,$7F,$3D
	.dc.b	$3F,$EE,$5F,$C3,$3F,$DF,$AF,$FB,$D5,$B1,$A3,$8D,$FF,$A9,$23,$39
	.dc.b	$7F,$BA,$5B,$5C,$FE,$27,$C7,$5B,$7F,$D5,$7E,$3F,$F1,$3C,$3F,$B3
	.dc.b	$35,$FF,$7B,$35,$F8,$6D,$EE,$7C,$F3,$E5,$3F,$A9,$B7,$1F,$ED,$DB
	.dc.b	$2F,$F9,$06,$9F,$B0,$AD,$63,$EB,$A7,$E2,$4F,$A9,$F9,$AD,$C7,$FD
	.dc.b	$9A,$DA,$FF,$0E,$57,$0A,$FF,$3A,$76,$3E,$37,$FC,$F6,$DC,$3C,$78
	.dc.b	$71,$CB,$FC,$7D,$39,$C7,$FC,$88,$D8,$CD,$FF,$B4,$2B,$85,$FF,$95
	.dc.b	$AB,$4D,$9E,$E6,$F0,$B1,$F0,$BE,$45,$F3,$5F,$88,$6C,$7F,$BA,$1B
	.dc.b	$97,$B1,$AF,$F8,$CC,$FF,$3E,$79,$6F,$B9,$FE,$6C,$7F,$B9,$27,$8E
	.dc.b	$9F,$CD,$E3,$FD,$CF,$FA,$AF,$F2,$1C,$3F,$DE,$AD,$C3,$E4,$B5,$FF
	.dc.b	$13,$35,$FF,$21,$6C,$8C,$47,$E9,$1B,$19,$F2,$BF,$47,$1A,$AF,$E2
	.dc.b	$DF,$59,$F4,$8F,$1C,$D8,$FE,$DD,$F1,$5F,$E3,$D7,$F0,$FF,$66,$F9
	.dc.b	$8F,$F4,$DF,$C5,$3C,$7F,$85,$7B,$D3,$F1,$17,$BC,$3F,$CB,$17,$B9
	.dc.b	$FF,$2D,$56,$F6,$FF,$52,$EA,$C3,$85,$2F,$83,$FE,$F3,$F2,$7F,$D5
	.dc.b	$77,$C2,$B7,$5D,$6F,$96,$7E,$6C,$FE,$3D,$69,$F3,$E1,$CB,$3C,$D6
	.dc.b	$F2,$FE,$1B,$7B,$6C,$65,$F1,$4F,$C4,$8C,$66,$FB,$BE,$59,$3F,$2B
	.dc.b	$78,$9A,$9C,$C6,$30,$F8,$BA,$7C,$A9,$CD,$6E,$73,$97,$F2,$0B,$E0
	.dc.b	$FC,$5B,$E1,$3A,$C3,$C1,$7C,$6B,$47,$17,$D5,$F5,$6F,$B7,$FE,$97
	.dc.b	$3A,$11,$9F,$37,$C3,$F3,$4F,$A7,$FD,$1F,$17,$D4,$AD,$5F,$16,$C6
	.dc.b	$70,$EA,$FE,$06,$F9,$75,$78,$A7,$3A,$FE,$AE,$F9,$E1,$FE,$A5,$6B
	.dc.b	$99,$F3,$B9,$6B,$C7,$B1,$F2,$7F,$EA,$D8,$FF,$54,$57,$F1,$18,$C4
	.dc.b	$7B,$A7,$6E,$E7,$4D,$8C,$E5,$EF,$6E,$1D,$EF,$E2,$FD,$13,$ED,$3F
	.dc.b	$82,$5E,$E7,$EC,$E1,$DD,$B8,$F6,$9F,$89,$7C,$6B,$3E,$A7,$2C,$7F
	.dc.b	$7F,$C6,$F9,$CC,$57,$E0,$DA,$FD,$8A,$B8,$FF,$88,$E3,$F0,$95,$C5
	.dc.b	$B5,$AA,$91,$82,$1F,$6D,$EF,$C7,$FB,$5F,$9D,$BD,$26,$9E,$0F,$17
	.dc.b	$FC,$99,$CF,$FB,$08,$CE,$1F,$95,$AE,$0F,$C2,$AD,$BF,$59,$B7,$DE
	.dc.b	$38,$57,$57,$FD,$05,$9F,$F5,$E7,$F8,$AC,$76,$7F,$25,$5D,$30,$7D
	.dc.b	$CF,$F0,$AE,$D4,$7F,$BA,$CC,$ED,$B9,$D4,$7F,$A6,$3F,$37,$FF,$67
	.dc.b	$E1,$FE,$88,$6B,$FE,$F9,$E9,$46,$23,$FA,$24,$63,$E2,$7F,$1E,$D8
	.dc.b	$FC,$A3,$5E,$53,$C5,$F0,$4F,$08,$C4,$7F,$4E,$F8,$FF,$8F,$EE,$BC
	.dc.b	$74,$61,$B3,$FD,$FB,$D4,$7D,$17,$CC,$6C,$7E,$97,$89,$18,$8F,$EF
	.dc.b	$2D,$7F,$DE,$CD,$7E,$47,$E0,$9C,$1F,$F7,$C6,$A8,$C4,$7F,$88,$E3
	.dc.b	$13,$F7,$B5,$F8,$FE,$4F,$D5,$BD,$F0,$FD,$4C,$3E,$AC,$63,$0E,$F8
	.dc.b	$71,$DB,$FB,$75,$78,$2B,$4F,$09,$E8,$6F,$09,$1F,$42,$B8,$BF,$42
	.dc.b	$3E,$71,$FC,$13,$D0,$7E,$0E,$F7,$BF,$2B,$18,$6C,$66,$BF,$C5,$7F
	.dc.b	$96,$57,$FC,$50,$D7,$FC,$0D,$B2,$31,$1F,$DE,$9A,$FF,$97,$1B,$0F
	.dc.b	$59,$E4,$FE,$91,$B5,$FE,$51,$8F,$D4,$AF,$C6,$9F,$39,$B9,$93,$B3
	.dc.b	$FD,$D5,$3A,$6C,$F2,$8C,$47,$F9,$6C,$E9,$1F,$47,$CA,$91,$F2,$DF
	.dc.b	$CB,$11,$FE,$61,$CF,$F9,$5F,$C0,$BF,$28,$48,$FF,$7D,$7C,$BB,$D3
	.dc.b	$C4,$7F,$9A,$6B,$83,$FE,$E3,$AF,$6D,$6E,$07,$FB,$D9,$FD,$1F,$C0
	.dc.b	$9A,$70,$ED,$B3,$C3,$A3,$FD,$E5,$38,$8F,$73,$FC,$19,$D7,$47,$6E
	.dc.b	$47,$E4,$CF,$AE,$FC,$1B,$6B,$62,$77,$3E,$8F,$A0,$93,$FD,$43,$04
	.dc.b	$7F,$52,$8F,$43,$63,$31,$B8,$C1,$1C,$1F,$3B,$37,$CE,$7C,$22,$47
	.dc.b	$90,$7D,$B7,$E5,$74,$FD,$76,$EB,$E3,$5E,$A8,$C5,$F2,$8F,$C1,$5F
	.dc.b	$1D,$7E,$8D,$3E,$39,$F6,$FF,$C5,$1B,$7C,$DA,$F2,$E3,$D4,$C8,$D3
	.dc.b	$F9,$AC,$3F,$C4,$F1,$F3,$33,$7E,$1B,$FD,$96,$3F,$D5,$D1,$8F,$47
	.dc.b	$FA,$4B,$63,$E0,$57,$B5,$BA,$FC,$F8,$C4,$7F,$81,$5A,$FF,$9E,$CF
	.dc.b	$8F,$1B,$23,$FE,$08,$8D,$BF,$E8,$16,$AC,$FF,$C5,$3F,$41,$FC,$DB
	.dc.b	$CB,$34,$FD,$BF,$DB,$FF,$9A,$EB,$E8,$4E,$9B,$3B,$7E,$D4,$ED,$FD
	.dc.b	$36,$1F,$EB,$37,$EC,$35,$CF,$BF,$DB,$E4,$FF,$4A,$C8,$DB,$97,$F0
	.dc.b	$1F,$71,$F8,$27,$F6,$4D,$D3,$F0,$ED,$E2,$FF,$50,$7D,$B7,$F9,$7E
	.dc.b	$BF,$1C,$D7,$F8,$43,$6B,$F9,$E4,$23,$13,$FE,$A9,$F6,$7F,$B7,$CF
	.dc.b	$EF,$27,$2D,$F7,$7F,$8E,$C6,$23,$FC,$B1,$E8,$7F,$00,$FA,$F9,$1F
	.dc.b	$09,$DF,$CF,$E2,$3E,$3C,$5C,$BC,$76,$AF,$ED,$7F,$6F,$FE,$14,$BF
	.dc.b	$E2,$4D,$AF,$D3,$B7,$3F,$83,$C3,$C3,$D5,$97,$FA,$9A,$D3,$EB,$9F
	.dc.b	$F0,$EF,$CE,$BF,$2D,$FF,$67,$CF,$F0,$88,$C7,$C0,$6A,$73,$3F,$CC
	.dc.b	$7D,$AF,$D0,$67,$F2,$9F,$D1,$47,$E5,$DA,$FF,$B4,$DA,$FF,$26,$6A
	.dc.b	$69,$3A,$BF,$89,$3A,$FD,$69,$F8,$0F,$F4,$C6,$D7,$F0,$B6,$BF,$D1
	.dc.b	$7F,$2C,$74,$7F,$52,$7F,$C6,$91,$FE,$71,$6B,$FA,$D6,$BF,$EE,$96
	.dc.b	$BF,$CF,$17,$CE,$F3,$24,$7F,$1B,$FC,$3D,$F6,$7F,$D8,$3F,$5D,$7C
	.dc.b	$5D,$25,$6F,$F0,$2F,$EC,$B6,$7F,$98,$36,$3F,$D0,$6D,$7F,$D1,$0D
	.dc.b	$7F,$82,$35,$FF,$80,$0D,$7F,$A9,$B5,$39,$F9,$EF,$F2,$7F,$C4,$FF
	.dc.b	$97,$9B,$3F,$D2,$1A,$FF,$6F,$FC,$B5,$F2,$FF,$DF,$1B,$1F,$E9,$86
	.dc.b	$BF,$E9,$D6,$BF,$C6,$5A,$FF,$66,$0E,$1F,$ED,$2D,$73,$1F,$D4,$E3
	.dc.b	$5E,$33,$EE,$DF,$37,$FE,$50,$6D,$FF,$47,$6B,$FE,$07,$F8,$67,$CA
	.dc.b	$FF,$85,$1B,$1F,$EB,$06,$BF,$EB,$56,$BF,$CC,$5A,$FA,$5A,$FF,$80
	.dc.b	$1A,$FF,$98,$23,$FB,$54,$6A,$FF,$99,$72,$FF,$99,$9F,$CB,$F4,$7F
	.dc.b	$60,$6F,$7F,$C5,$F9,$FF,$C4,$E3,$FD,$9D,$AF,$FB,$25,$AF,$F4,$56
	.dc.b	$BC,$78,$BF,$BD,$46,$23,$FC,$2A,$D4,$D3,$55,$FD,$1F,$E2,$BC,$C4
	.dc.b	$9F,$17,$C4,$7F,$8F,$2B,$FC,$E1,$1D,$23,$B9,$1F,$F6,$DC,$69,$EF
	.dc.b	$19,$7D,$4F,$FB,$82,$9C,$3F,$E9,$AC,$BF,$C0,$6D,$CB,$FB,$EB,$6F
	.dc.b	$D7,$EA,$F8,$CF,$79,$1F,$0B,$5F,$90,$D7,$E4,$3F,$EF,$0F,$80,$79
	.dc.b	$3F,$D6,$F0,$FF,$40,$46,$3E,$4F,$FD,$83,$39,$C7,$EB,$7A,$FF,$0C
	.dc.b	$F9,$E4,$C7,$E0,$BF,$E2,$BF,$69,$EE,$7F,$99,$1B,$3F,$E2,$36,$BF
	.dc.b	$E5,$D7,$FD,$4E,$D8,$E0,$47,$87,$1F,$FA,$C3,$E7,$BF,$B1,$47,$F7
	.dc.b	$F6,$EE,$73,$A3,$F5,$E7,$D9,$71,$F7,$26,$1F,$E2,$FB,$DD,$6A,$3F
	.dc.b	$C9,$76,$FB,$6F,$EF,$9C,$3F,$19,$BB,$FF,$98,$A7,$5E,$AF,$FB,$A3
	.dc.b	$F2,$BF,$E8,$F3,$9A,$DE,$0B,$DB,$DA,$33,$0A,$E3,$3B,$7F,$A5,$C6
	.dc.b	$7C,$AF,$B5,$18,$33,$5F,$E1,$9A,$63,$E7,$F3,$BF,$DF,$73,$98,$FB
	.dc.b	$FF,$71,$E0,$8C,$65,$E7,$AD,$9F,$C6,$6B,$79,$FF,$A7,$BE,$A1,$EF
	.dc.b	$F6,$CB,$FC,$70,$23,$35,$88,$C4,$ED,$FF,$1F,$4E,$9E,$2F,$F4,$CD
	.dc.b	$F8,$36,$33,$E4,$FD,$D6,$CF,$F9,$E5,$7C,$BD,$23,$FE,$7D,$7F,$DD
	.dc.b	$DA,$FF,$9D,$B4,$BC,$5F,$E2,$4A,$FB,$53,$97,$FC,$A5,$C2,$8C,$47
	.dc.b	$FB,$29,$AF,$09,$E4,$FD,$D8,$CB,$63,$2F,$FA,$A3,$E2,$2B,$FE,$CC
	.dc.b	$BF,$F4,$94,$60,$E7,$7E,$1F,$16,$E1,$FD,$0A,$BD,$71,$82,$38,$25
	.dc.b	$FA,$A7,$83,$EC,$FB,$87,$CB,$76,$6D,$7D,$AC,$11,$8A,$E1,$5B,$25
	.dc.b	$6E,$75,$3F,$EE,$E6,$BF,$D8,$7E,$1B,$F6,$33,$FE,$FD,$E8,$F8,$71
	.dc.b	$89,$FC,$4B,$5F,$F4,$13,$5E,$51,$C0,$93,$F8,$D9,$7D,$17,$42,$46
	.dc.b	$5F,$AB,$C6,$FB,$87,$D7,$61,$B5,$F9,$A9,$D4,$E4,$F3,$92,$3C,$0D
	.dc.b	$C7,$D2,$D7,$D2,$D7,$FD,$CF,$F2,$CF,$B8,$7C,$77,$E2,$23,$0D,$8C
	.dc.b	$DF,$EB,$27,$2F,$EA,$FD,$17,$CB,$7E,$31,$B1,$98,$C1,$1C,$13,$0F
	.dc.b	$DB,$CF,$07,$FD,$61,$0F,$B6,$FA,$10,$8C,$6F,$F9,$E6,$9F,$C0,$0B
	.dc.b	$1F,$C0,$BD,$EF,$F5,$7C,$7F,$01,$6C,$67,$BB,$F4,$71,$88,$FF,$60
	.dc.b	$35,$FF,$14,$35,$07,$0D,$8F,$25,$EE,$BD,$51,$82,$3F,$C4,$CF,$F2
	.dc.b	$C9,$7A,$FF,$6A,$E2,$BD,$91,$82,$46,$C9,$1F,$23,$A5,$F6,$9F,$15
	.dc.b	$B3,$F1,$5B,$3B,$9F,$E0,$7E,$67,$E4,$23,$F9,$4B,$5F,$F7,$13,$5F
	.dc.b	$F2,$53,$5F,$86,$D7,$FC,$17,$C3,$D0,$97,$F5,$72,$79,$9F,$18,$91
	.dc.b	$97,$F8,$AF,$DB,$7F,$78,$8F,$E7,$0D,$BF,$E7,$AD,$7F,$DE,$CD,$7F
	.dc.b	$CB,$CD,$7E,$BB,$53,$49,$F2,$FE,$02,$4F,$81,$F1,$E5,$23,$C6,$DA
	.dc.b	$FD,$27,$C2,$7F,$2C,$E3,$FF,$65,$57,$F4,$F6,$CE,$A3,$FC,$41,$F6
	.dc.b	$FF,$CC,$BB,$4D,$24,$7F,$57,$F8,$4F,$F4,$B3,$7D,$CF,$E7,$31,$8B
	.dc.b	$FE,$BE,$D7,$FC,$FC,$1C,$3F,$E8,$16,$BF,$9A,$6B,$B8,$FE,$93,$DA
	.dc.b	$FB,$67,$CF,$61,$B8,$7F,$26,$6D,$FF,$26,$6B,$F8,$63,$F0,$DC,$7F
	.dc.b	$A8,$E3,$FE,$AA,$DA,$FC,$7C,$E8,$93,$A6,$C6,$BE,$BB,$04,$70,$48
	.dc.b	$FC,$B8,$8D,$38,$7F,$B0,$FD,$9B,$D4,$A4,$AF,$F0,$34,$65,$B1,$9A
	.dc.b	$FF,$06,$CE,$5B,$1F,$CA,$7E,$CF,$F8,$D4,$7F,$AB,$F8,$51,$88,$FE
	.dc.b	$62,$D5,$1A,$EE,$F8,$BE,$57,$E1,$31,$FD,$9B,$6B,$F6,$6D,$7F,$66
	.dc.b	$D7,$F6,$6D,$7F,$CF,$AD,$D3,$FC,$A9,$C5,$D8,$EF,$F2,$D2,$FB,$B6
	.dc.b	$FB,$B5,$EE,$D7,$FD,$99,$A3,$EB,$3E,$69,$2B,$E5,$B6,$7E,$4B,$5F
	.dc.b	$CB,$CF,$19,$FE,$C1,$F6,$7F,$F0,$01,$BF,$0D,$FD,$A2,$31,$5F,$9E
	.dc.b	$F1,$91,$88,$F8,$7F,$15,$F2,$89,$7F,$29,$B1,$F2,$9A,$F0,$8F,$9F
	.dc.b	$F0,$5F,$48,$8E,$09,$1F,$4A,$32,$D8,$FF,$27,$7D,$A7,$FB,$81,$B3
	.dc.b	$FE,$30,$8C,$4F,$F9,$59,$AF,$EE,$1A,$FF,$BD,$DA,$FF,$50,$6A,$73
	.dc.b	$C2,$FB,$37,$C9,$7F,$7B,$6E,$1F,$DE,$5B,$7F,$C1,$AB,$85,$7E,$43
	.dc.b	$EC,$E5,$FE,$41,$B5,$F9,$06,$BF,$90,$6B,$F9,$0E,$4F,$CC,$92,$7F
	.dc.b	$32,$D8,$FC,$CB,$5F,$CC,$BF,$E5,$4F,$B2,$FE,$6C,$4F,$7F,$07,$F9
	.dc.b	$4B,$EC,$FF,$C9,$35,$FD,$C8,$FF,$4D,$38,$7F,$D4,$4D,$73,$EA,$F5
	.dc.b	$F8,$B9,$8F,$FA,$0F,$04,$62,$BF,$D1,$31,$89,$D4,$E6,$7F,$96,$CE
	.dc.b	$5D,$FD,$7F,$A4,$95,$D5,$FF,$34,$F3,$3E,$42,$92,$7F,$CB,$D1,$97
	.dc.b	$FD,$2D,$EC,$FF,$88,$2C,$D2,$7A,$3F,$E7,$4F,$B2,$7B,$D4,$97,$FE
	.dc.b	$69,$8C,$B6,$33,$7F,$E8,$C9,$CB,$63,$2F,$CD,$F2,$53,$A8,$C3,$5F
	.dc.b	$F6,$83,$5F,$F4,$67,$95,$FE,$8E,$C7,$FB,$0C,$7F,$54,$6B,$F6,$5A
	.dc.b	$FF,$B4,$FE,$89,$F2,$BF,$E9,$06,$E5,$FE,$83,$6C,$FF,$A9,$1A,$F5
	.dc.b	$FB,$5F,$01,$27,$E5,$B6,$3F,$D8,$B6,$8C,$47,$19,$FF,$48,$7D,$9B
	.dc.b	$83,$FD,$42,$D8,$CC,$60,$F7,$7C,$7C,$8F,$B2,$DA,$EE,$D7,$E1,$35
	.dc.b	$F8,$4F,$F9,$75,$FF,$73,$C7,$FB,$EF,$EB,$5F,$1B,$9F,$83,$FD,$ED
	.dc.b	$7F,$AE,$6F,$87,$87,$C7,$85,$87,$0C,$7F,$CA,$F8,$70,$C3,$F1,$78
	.dc.b	$70,$9E,$D8,$F5,$ED,$F1,$C7,$F9,$7B,$2E,$98,$F3,$C3,$FD,$BB,$5F
	.dc.b	$67,$E3,$7D,$17,$F7,$5B,$39,$8F,$F3,$39,$92,$B9,$3F,$DC,$1B,$A7
	.dc.b	$8C,$7F,$B5,$5A,$FE,$89,$B1,$B9,$F6,$D7,$F9,$D3,$47,$D7,$3E,$1F
	.dc.b	$D9,$8F,$3B,$DA,$AD,$CE,$A1,$F4,$2F,$9E,$FE,$EB,$1F,$00,$FF,$40
	.dc.b	$C6,$7C,$3E,$7F,$62,$FB,$2F,$CA,$F1,$FC,$FA,$D5,$F1,$3A,$E1,$D2
	.dc.b	$74,$DD,$7F,$D1,$B8,$73,$E8,$F2,$61,$CA,$F7,$97,$F8,$FB,$1F,$1E
	.dc.b	$3E,$67,$FC,$83,$5E,$06,$C6,$5F,$E5,$10,$FB,$27,$C2,$73,$9F,$83
	.dc.b	$18,$7B,$4E,$E7,$4D,$8C,$CE,$F0,$F8,$0D,$CF,$BB,$65,$9B,$EC,$9F
	.dc.b	$21,$FC,$EF,$0E,$39,$7D,$6B,$F1,$B7,$2F,$04,$FE,$8B,$0E,$57,$C7
	.dc.b	$2F,$6F,$9B,$71,$FB,$DC,$6F,$8C,$C7,$1F,$E1,$9D,$4F,$E0,$2A,$E6
	.dc.b	$7F,$A7,$72,$7C,$C6,$CF,$63,$F3,$67,$E1,$1C,$B4,$C9,$D4,$B3,$D5
	.dc.b	$B1,$C1,$FC,$8F,$CB,$FF,$5D,$E9,$FD,$8F,$D0,$7A,$3A,$8F,$9A,$75
	.dc.b	$8F,$DC,$6D,$FE,$76,$B6,$FE,$83,$3F,$E0,$B7,$C6,$3F,$7D,$7C,$5F
	.dc.b	$BB,$7F,$83,$7C,$66,$08,$7E,$11,$D6,$46,$9F,$EA,$1F,$B0,$FF,$44
	.dc.b	$CF,$E5,$F9,$91,$88,$FF,$09,$35,$FF,$34,$35,$FC,$BF,$CD,$76,$23
	.dc.b	$F9,$26,$BF,$96,$6B,$FA,$27,$F3,$99,$3C,$E7,$CC,$E1,$B5,$FA,$B9
	.dc.b	$D4,$E4,$F6,$FA,$B2,$7F,$42,$D7,$D3,$F6,$1E,$F2,$3E,$FC,$51,$88
	.dc.b	$D9,$97,$0F,$C0,$6A,$3F,$D0,$B1,$C9,$E0,$7E,$DF,$EB,$5F,$35,$F5
	.dc.b	$DB,$19,$8C,$1F,$0B,$EB,$DF,$13,$49,$3F,$06,$32,$F5,$FB,$4E,$84
	.dc.b	$70,$78,$7E,$4F,$C0,$47,$04,$3F,$8B,$60,$8D,$46,$1F,$C3,$78,$92
	.dc.b	$3E,$43,$5F,$E3,$DF,$67,$FE,$22,$7F,$8B,$35,$CF,$8D,$F9,$78,$C4
	.dc.b	$7F,$B8,$BE,$39,$E7,$7E,$6D,$B1,$FC,$7D,$AF,$F2,$B6,$BC,$63,$81
	.dc.b	$1E,$0F,$3F,$B0,$91,$CD,$AF,$36,$BE,$36,$BF,$E5,$99,$FF,$69,$3E
	.dc.b	$3F,$86,$F3,$11,$FE,$69,$F9,$7B,$DA,$FD,$A3,$63,$31,$82,$38,$21
	.dc.b	$FD,$1E,$38,$3F,$D2,$7E,$BB,$EF,$B7,$87,$F8,$04,$62,$3F,$CD,$C1
	.dc.b	$C3,$FE,$71,$6B,$F9,$76,$BB,$8E,$67,$4D,$64,$7F,$4F,$9F,$EE,$CD
	.dc.b	$8C,$BE,$56,$BE,$28,$DB,$E9,$FA,$DF,$39,$1E,$8F,$F1,$9F,$C0,$FF
	.dc.b	$90,$C7,$F6,$76,$BE,$C6,$BF,$E8,$B6,$BF,$B1,$6A,$69,$3C,$0F,$E4
	.dc.b	$DF,$5F,$FE,$36,$9F,$ED,$4D,$7F,$BB,$35,$F9,$6D,$7F,$D3,$4D,$7F
	.dc.b	$80,$35,$34,$9E,$E7,$80,$91,$E0,$6B,$E0,$6B,$E0,$7F,$9E,$7D,$6F
	.dc.b	$F8,$BE,$3F,$C0,$0D,$7F,$C1,$8D,$7F,$02,$D7,$FD,$61,$E8,$7F,$92
	.dc.b	$DF,$C2,$B5,$CF,$95,$FD,$26,$31,$1F,$C6,$DA,$FF,$8B,$63,$44,$7C
	.dc.b	$7E,$CF,$98,$8E,$08,$79,$9C,$B4,$F5,$FC,$CF,$B8,$8E,$08,$7B,$9C
	.dc.b	$B4,$FF,$11,$79,$5F,$E1,$46,$F9,$BF,$D8,$E3,$11,$FE,$31,$6B,$F0
	.dc.b	$B8,$7E,$01,$1F,$F1,$66,$3F,$A4,$24,$66,$3F,$C7,$51,$FA,$8F,$AB
	.dc.b	$FF,$63,$47,$F8,$CD,$AF,$F8,$FD,$AF,$ED,$1A,$FF,$B8,$D7,$85,$FD
	.dc.b	$E5,$B1,$CE,$30,$7F,$7D,$FA,$A7,$CD,$D2,$4F,$F9,$0E,$32,$DF,$96
	.dc.b	$E0,$47,$07,$F9,$27,$F5,$FF,$E6,$2F,$FB,$D5,$AF,$F5,$B6,$A6,$93
	.dc.b	$BF,$FC,$20,$D8,$FF,$2A,$7D,$5F,$FB,$D9,$BF,$0D,$FE,$15,$8C,$47
	.dc.b	$F9,$8D,$AF,$F1,$8F,$88,$7E,$15,$FE,$1B,$6C,$7F,$97,$DA,$FF,$9A
	.dc.b	$DA,$F1,$8E,$04,$78,$3F,$E2,$5F,$21,$F1,$B4,$93,$FE,$6B,$8C,$BF
	.dc.b	$E6,$DF,$D1,$1F,$39,$FE,$2D,$6C,$66,$30,$47,$04,$3F,$D0,$11,$C1
	.dc.b	$FF,$41,$7D,$63,$E1,$3F,$C7,$0D,$8F,$F3,$E3,$5D,$C7,$F8,$F7,$E0
	.dc.b	$7F,$C5,$D1,$FE,$0F,$6B,$FE,$F2,$F9,$BD,$11,$FF,$43,$46,$89,$1E
	.dc.b	$0F,$4A,$4F,$81,$AF,$FA,$5F,$F4,$47,$D2,$FF,$93,$5B,$1F,$E9,$16
	.dc.b	$BB,$8D,$11,$D3,$5F,$83,$E5,$FC,$02,$3F,$01,$FF,$2D,$7E,$63,$F7
	.dc.b	$23,$FC,$58,$1C,$3F,$E2,$F6,$B9,$E0,$E2,$47,$FD,$4D,$1A,$37,$A7
	.dc.b	$F7,$A8,$FA,$7E,$67,$D2,$23,$C3,$D7,$F1,$91,$C1,$FE,$B1,$FA,$EF
	.dc.b	$F1,$CB,$FE,$BD,$6B,$FD,$C9,$AF,$A9,$AF,$F9,$1D,$A9,$A4,$E2,$FF
	.dc.b	$31,$FD,$6B,$E4,$29,$2B,$FC,$C5,$19,$6C,$65,$FF,$01,$FB,$49,$D4
	.dc.b	$61,$AF,$CE,$87,$D6,$FF,$26,$8F,$F7,$B3,$5F,$F7,$93,$63,$7E,$2F
	.dc.b	$F9,$E2,$30,$47,$07,$FA,$57,$EB,$1D,$54,$97,$FE,$DB,$8C,$BF,$E7
	.dc.b	$DF,$AC,$7C,$9F,$FA,$11,$B1,$98,$C1,$1C,$10,$F8,$31,$C1,$FF,$79
	.dc.b	$4B,$D3,$7C,$2F,$85,$B3,$A6,$BF,$E2,$C9,$E1,$3E,$7F,$15,$F3,$F4
	.dc.b	$95,$DA,$32,$FF,$AD,$7E,$2B,$FD,$6C,$9A,$47,$CF,$F5,$0F,$87,$F8
	.dc.b	$4D,$8C,$C6,$08,$E0,$87,$DB,$CD,$3C,$63,$FC,$8A,$3C,$7C,$3F,$E4
	.dc.b	$86,$BF,$B0,$6B,$FE,$9D,$73,$C2,$AF,$F2,$F5,$FD,$7F,$6B,$31,$8B
	.dc.b	$FA,$31,$82,$4E,$C9,$1F,$96,$FA,$CF,$F8,$01,$1F,$EB,$D6,$C6,$7E
	.dc.b	$2B,$C3,$18,$8F,$CC,$35,$FF,$59,$60,$8C,$42,$31,$3C,$8F,$6F,$D3
	.dc.b	$7E,$B6,$FE,$2D,$23,$0F,$C8,$6B,$FE,$6E,$6B,$FC,$59,$AF,$FB,$0D
	.dc.b	$AF,$3E,$7E,$E7,$83,$C0,$95,$E0,$AF,$E5,$ED,$8F,$A7,$C2,$FA,$C7
	.dc.b	$D3,$61,$B5,$F7,$27,$53,$93,$DA,$48,$EC,$DC,$7D,$8D,$7D,$8D,$7F
	.dc.b	$DB,$9F,$14,$FA,$CF,$F6,$25,$7E,$2A,$31,$F0,$BF,$88,$6C,$66,$77
	.dc.b	$5F,$90,$F9,$1F,$F7,$E4,$FC,$E6,$C6,$7C,$1F,$C7,$C6,$23,$FD,$36
	.dc.b	$D7,$FA,$1B,$5F,$F7,$83,$53,$99,$7D,$5F,$D4,$6E,$FF,$9E,$69,$FA
	.dc.b	$42,$D7,$E9,$BE,$B1,$ED,$FE,$89,$B3,$F8,$66,$CE,$E7,$44,$9D,$36
	.dc.b	$3F,$D6,$B1,$E4,$8F,$15,$F2,$9F,$EB,$F3,$93,$C9,$19,$8C,$F5,$7E
	.dc.b	$D2,$7F,$C8,$67,$E6,$0B,$1E,$B8,$D1,$23,$4D,$8F,$1F,$4F,$F8,$7A
	.dc.b	$7E,$13,$63,$3E,$87,$C7,$8C,$47,$FB,$3B,$E9,$FF,$22,$FF,$0C,$6B
	.dc.b	$9E,$1F,$E1,$31,$88,$FE,$F6,$D4,$1C,$36,$38,$9F,$1B,$CB,$98,$7C
	.dc.b	$46,$CF,$F1,$CF,$AC,$FE,$9B,$3F,$BA,$6C,$67,$C3,$FE,$37,$18,$8F
	.dc.b	$F7,$3B,$5F,$F0,$7B,$5F,$2B,$53,$47,$B9,$D0,$95,$E1,$7B,$E7,$23
	.dc.b	$C8,$DA,$FD,$57,$BD,$FC,$BB,$94,$D2,$57,$F1,$4E,$C7,$D4,$7F,$34
	.dc.b	$6C,$66,$30,$7F,$38,$FA,$C7,$B7,$49,$3F,$C7,$A3,$2D,$8C,$E3,$FD
	.dc.b	$1A,$72,$D8,$FE,$8B,$DC,$FA,$C7,$AB,$86,$E1,$FC,$9D,$B7,$FC,$9D
	.dc.b	$AF,$E0,$8F,$C1,$71,$FE,$0B,$8F,$FA,$8B,$6B,$F1,$93,$A2,$4E,$9B
	.dc.b	$1A,$FA,$CC,$11,$C1,$23,$F3,$02,$34,$E1,$FE,$BF,$F5,$4F,$4A,$92
	.dc.b	$BF,$A2,$C6,$5B,$19,$AF,$ED,$B3,$96,$C7,$F2,$5F,$AC,$FE,$31,$C4
	.dc.b	$69,$3B,$FE,$0F,$97,$F7,$2F,$F5,$CD,$7F,$52,$D7,$F5,$3A,$7F,$84
	.dc.b	$B8,$B8,$93,$3F,$DA,$56,$EB,$F5,$CD,$AF,$F0,$CF,$D6,$78,$09,$5C
	.dc.b	$1B,$5E,$26,$BE,$27,$E3,$FD,$34,$9F,$8A,$D9,$F8,$6D,$7F,$CF,$63
	.dc.b	$54,$66,$3F,$81,$F8,$AF,$AC,$FF,$72,$DF,$F5,$D6,$CF,$29,$D7,$C2
	.dc.b	$7F,$66,$9C,$CF,$C6,$F8,$7F,$96,$4B,$F8,$6D,$8F,$9A,$D7,$84,$7C
	.dc.b	$8F,$81,$FA,$64,$70,$48,$FA,$B1,$96,$C7,$F8,$2B,$EB,$3F,$D9,$ED
	.dc.b	$9F,$F1,$5C,$62,$7F,$C6,$6D,$7E,$2B,$5F,$F5,$0B,$5F,$E4,$ED,$4E
	.dc.b	$78,$9F,$58,$F8,$AF,$EE,$8D,$C3,$FB,$C3,$6F,$E9,$D7,$0A,$FC,$8F
	.dc.b	$D6,$4B,$FC,$4B,$6B,$F2,$6D,$7F,$26,$D7,$F2,$7C,$5F,$DD,$78,$BF
	.dc.b	$CC,$BC,$9F,$9C,$25,$FE,$51,$B5,$F9,$E6,$BF,$86,$FA,$BF,$E6,$04
	.dc.b	$F1,$3F,$CA,$6F,$F3,$5F,$01,$F5,$9F,$E4,$FA,$FF,$4D,$35,$FF,$21
	.dc.b	$B5,$FF,$21,$35,$4B,$E2,$7A,$93,$1F,$F2,$1C,$65,$B1,$99,$FF,$35
	.dc.b	$CE,$5B,$1E,$6F,$A9,$95,$C9,$B3,$FE,$B4,$E7,$7C,$87,$F9,$B1,$B5
	.dc.b	$FE,$76,$CB,$F4,$B3,$93,$FC,$E7,$EC,$FF,$93,$09,$E8,$60,$FF,$3A
	.dc.b	$FD,$67,$FA,$A2,$F9,$37,$1F,$F6,$1B,$5F,$F6,$03,$5F,$F3,$4B,$5F
	.dc.b	$F3,$43,$70,$34,$9E,$F7,$80,$95,$E0,$7C,$FE,$4C,$8F,$F3,$6E,$3F
	.dc.b	$D7,$9B,$1F,$E8,$CF,$99,$7C,$55,$24,$FC,$38,$CB,$CB,$EB,$3E,$19
	.dc.b	$1C,$13,$3F,$F4,$FE,$09,$D4,$61,$FF,$4D,$FD,$63,$82,$92,$FE,$94
	.dc.b	$65,$F4,$FC,$6F,$D2,$23,$82,$5F,$82,$32,$D8,$CF,$85,$FE,$BB,$DD
	.dc.b	$F4,$AF,$AC,$FF,$5F,$7C,$1F,$EE,$70,$FF,$58,$63,$FE,$C1,$C3,$85
	.dc.b	$EF,$1F,$F1,$4D,$EE,$FF,$D7,$97,$B9,$FF,$51,$D6,$E7,$FD,$A1,$DD
	.dc.b	$FE,$BD,$AD,$47,$FA,$96,$75,$7C,$F6,$7D,$67,$FB,$47,$1F,$F6,$33
	.dc.b	$6F,$FD,$8C,$D7,$FD,$8C,$D7,$FD,$9C,$D9,$E7,$3F,$E4,$03,$99,$67
	.dc.b	$9B,$5E,$6F,$BB,$1F,$F5,$EF,$D4,$7F,$BE,$63,$13,$A3,$DA,$7F,$AF
	.dc.b	$E3,$4E,$07,$D4,$3E,$61,$E1,$F2,$9C,$C6,$1E,$7F,$52,$FA,$CF,$C8
	.dc.b	$C7,$3B,$E5,$7C,$5B,$19,$BE,$5A,$FD,$BC,$39,$73,$FD,$ED,$3F,$CC
	.dc.b	$37,$E0,$C7,$96,$1E,$18,$FE,$CF,$8F,$46,$C6,$5E,$31,$96,$FD,$67
	.dc.b	$FA,$89,$FF,$7F,$C2,$31,$1D,$1A,$F0,$F8,$2F,$F7,$0C,$ED,$F6,$4F
	.dc.b	$3C,$FF,$1D,$39,$AF,$8B,$F5,$8F,$8E,$F5,$D6,$AF,$86,$7D,$6F,$E4
	.dc.b	$46,$16,$AC,$39,$47,$CB,$F3,$79,$DF,$E0,$3E,$26,$CF,$DD,$6B,$F7
	.dc.b	$26,$E6,$7F,$D5,$71,$F1,$9A,$FF,$A4,$23,$4F,$E3,$CE,$45,$72,$B8
	.dc.b	$BE,$9C,$F2,$8D,$99,$3B,$1C,$7E,$B3,$EE,$F4,$7E,$E3,$E8,$3D,$3D
	.dc.b	$7F,$C0,$1F,$3C,$EB,$0F,$B6,$DB,$FD,$A4,$F0,$E9,$3C,$23,$FC,$4F
	.dc.b	$5C,$1F,$B3,$5F,$E2,$E7,$07,$E0,$9C,$3F,$05,$AF,$E9,$1B,$91,$A7
	.dc.b	$0E,$47,$D5,$7F,$A0,$2B,$B4,$63,$DC,$D4,$E6,$7F,$C1,$AD,$7F,$37
	.dc.b	$F5,$9D,$88,$FE,$41,$AF,$E6,$1A,$F0,$3F,$B2,$79,$AF,$9B,$FF,$52
	.dc.b	$36,$33,$18,$3D,$7E,$B4,$8F,$CE,$B5,$F2,$FD,$67,$A0,$8F,$B3,$04
	.dc.b	$62,$36,$65,$C3,$FB,$AC,$BE,$2C,$62,$3F,$DB,$B1,$87,$81,$F4,$BE
	.dc.b	$A9,$F3,$1F,$49,$B1,$98,$C1,$FE,$C1,$FA,$CF,$F6,$7C,$7F,$81,$5A
	.dc.b	$FF,$9B,$38,$D1,$88,$FE,$1B,$18,$F8,$7E,$DF,$59,$27,$C4,$D7,$F1
	.dc.b	$18,$3E,$47,$C8,$48,$F4,$35,$DB,$5F,$F3,$EC,$71,$7E,$57,$CC,$7E
	.dc.b	$10,$8E,$0F,$C9,$FD,$63,$C7,$FC,$9B,$63,$31,$83,$F2,$FF,$1A,$F3
	.dc.b	$E9,$27,$F3,$B1,$96,$C6,$5F,$D2,$46,$5B,$1D,$3E,$B2,$47,$26,$BC
	.dc.b	$1A,$FF,$A6,$A5,$18,$8F,$E7,$71,$FE,$C9,$7C,$7F,$0D,$E8,$23,$FB
	.dc.b	$2F,$AC,$7C,$CF,$EC,$9B,$19,$8C,$11,$C1,$0F,$DF,$47,$07,$F8,$07
	.dc.b	$D6,$7E,$05,$FE,$9A,$D4,$1C,$37,$83,$F8,$1C,$60,$8E,$0C,$EB,$27
	.dc.b	$F8,$5C,$A3,$11,$E2,$6B,$FE,$C0,$8D,$3C,$23,$BF,$D6,$48,$E6,$D7
	.dc.b	$F8,$D7,$D6,$7F,$20,$7F,$8C,$B5,$CF,$83,$FC,$6A,$31,$1F,$EC,$E6
	.dc.b	$BF,$DB,$9A,$FC,$26,$A6,$FD,$67,$F8,$CE,$3F,$93,$B5,$FE,$58,$D7
	.dc.b	$FD,$B6,$D7,$FB,$DB,$5F,$9E,$D4,$D2,$7B,$9F,$CC,$9F,$F4,$1C,$78
	.dc.b	$23,$04,$70,$43,$C6,$FF,$3A,$FA,$CF,$F1,$BC,$7F,$3E,$6B,$FD,$0D
	.dc.b	$AF,$FB,$C9,$AF,$F8,$4B,$D0,$FF,$25,$3F,$EF,$66,$B9,$F2,$BF,$A3
	.dc.b	$C6,$23,$F1,$6D,$7F,$C5,$91,$A2,$3E,$3F,$6B,$CC,$47,$04,$3C,$CE
	.dc.b	$5A,$7A,$FE,$67,$DC,$47,$04,$3D,$CE,$5A,$7F,$60,$F2,$FF,$C1,$AF
	.dc.b	$F6,$56,$BF,$E3,$23,$FC,$11,$E8,$FF,$64,$8C,$C7,$C5,$E1,$F8,$24
	.dc.b	$7F,$B5,$E3,$FA,$A2,$46,$63,$FB,$94,$7E,$BB,$EB,$3E,$EC,$7F,$75
	.dc.b	$6B,$E1,$6B,$FE,$44,$6B,$FB,$85,$E2,$7F,$78,$6C,$7F,$7C,$6B,$D6
	.dc.b	$3F,$BE,$FD,$63,$E6,$A9,$27,$FC,$07,$19,$6F,$CB,$70,$23,$83,$FC
	.dc.b	$13,$F5,$9F,$CD,$5F,$F0,$73,$5F,$94,$D7,$FC,$BA,$D7,$F8,$8B,$53
	.dc.b	$49,$DD,$FE,$14,$FA,$CF,$F7,$A4,$FF,$86,$1A,$FF,$87,$5A,$FD,$96
	.dc.b	$BF,$E6,$CF,$89,$7D,$57,$F8,$79,$B1,$FE,$22,$6B,$FE,$27,$6B,$F8
	.dc.b	$36,$BC,$E3,$91,$1F,$F1,$5F,$90,$F9,$0F,$F1,$5B,$63,$31,$83,$FC
	.dc.b	$5F,$F5,$8F,$9C,$A4,$9F,$F1,$AC,$65,$B1,$97,$FC,$79,$19,$6C,$7F
	.dc.b	$8E,$BE,$A9,$F0,$74,$93,$FE,$40,$8C,$BF,$E4,$2F,$80,$FF,$1A,$3F
	.dc.b	$E9,$26,$BF,$D1,$9A,$FF,$BD,$3D,$5F,$F2,$3C,$60,$8E,$0F,$17,$A7
	.dc.b	$27,$C0,$D7,$FC,$A9,$F5,$8F,$A5,$FF,$28,$B6,$3F,$CA,$AD,$77,$1A
	.dc.b	$23,$A6,$BF,$E5,$FC,$1E,$57,$C3,$24,$7C,$17,$FC,$BF,$F5,$9F,$72
	.dc.b	$3F,$B1,$B5,$34,$9C,$1C,$09,$1F,$E6,$A8,$E1,$5F,$DF,$E3,$C5,$E5
	.dc.b	$FD,$32,$3B,$F5,$FC,$24,$70,$7F,$9D,$3E,$B3,$FC,$82,$FF,$9E,$5A
	.dc.b	$FF,$15,$6B,$FE,$C8,$6B,$FD,$F1,$A9,$A4,$E3,$FF,$40,$FA,$8F,$AC
	.dc.b	$7C,$7E,$1B,$7F,$E8,$99,$D4,$E4,$91,$99,$DB,$FE,$51,$6B,$FE,$50
	.dc.b	$6B,$FE,$10,$97,$D6,$7F,$2C,$84,$63,$C4,$F2,$CE,$63,$13,$E3,$8C
	.dc.b	$12,$76,$7F,$A6,$BE,$B1,$D5,$49,$5F,$EA,$08,$CB,$ED,$FA,$C7,$C8
	.dc.b	$FC,$46,$C6,$63,$04,$70,$43,$FC,$E1,$1C,$1F,$F5,$84,$FF,$88,$FE
	.dc.b	$13,$C1,$E9,$1A,$49,$DF,$88,$F9,$8F,$33,$63,$31,$83,$FD,$6E,$BE
	.dc.b	$2A,$91,$FF,$5D,$B7,$F0,$47,$DA,$7F,$5D,$F8,$1F,$F6,$B4,$62,$BF
	.dc.b	$B0,$57,$F9,$12,$31,$1F,$E9,$8F,$87,$C5,$FF,$B2,$EA,$CF,$FA,$7E
	.dc.b	$6B,$FF,$00,$05,$72,$57,$E0,$BF,$C1,$7E,$BD,$E4,$BF,$C7,$36,$BF
	.dc.b	$D3,$4D,$7F,$D3,$2D,$7F,$15,$F8,$13,$F1,$8F,$F7,$86,$BF,$EF,$1E
	.dc.b	$54,$E7,$0F,$F7,$2D,$FE,$3B,$0E,$D7,$BA,$D4,$FD,$5A,$EB,$87,$46
	.dc.b	$9C,$B0,$61,$CA,$33,$5C,$47,$D9,$BE,$A9,$F1,$DC,$E7,$31,$8A,$F2
	.dc.b	$D6,$FD,$ED,$3C,$6E,$D6,$EF,$B4,$E6,$F8,$CA,$F7,$7C,$BE,$BD,$F8
	.dc.b	$6D,$31,$E9,$5C,$3C,$33,$A6,$CE,$9A,$79,$87,$BA,$7F,$C5,$19,$FC
	.dc.b	$07,$26,$33,$D4,$EE,$7F,$C5,$D1,$89,$D7,$27,$CB,$9F,$92,$D9,$D4
	.dc.b	$7C,$08,$C7,$D7,$3D,$AE,$93,$FE,$9C,$3E,$8C,$66,$FC,$4D,$8F,$0B
	.dc.b	$5E,$16,$8C,$47,$FA,$9A,$31,$5D,$A3,$13,$A7,$BC,$65,$FC,$0F,$2B
	.dc.b	$E2,$BE,$DB,$70,$FB,$F7,$E3,$E5,$F8,$98,$FE,$9A,$FF,$03,$C7,$FE
	.dc.b	$C4,$BF,$94,$DC,$7E,$4B,$5F,$4B,$5F,$43,$5F,$C9,$B6,$BF,$97,$FC
	.dc.b	$2F,$9D,$B5,$F5,$F8,$7F,$D8,$AF,$E5,$1B,$7F,$90,$BF,$16,$3E,$89
	.dc.b	$CB,$EA,$FC,$1D,$F5,$5D,$5B,$87,$E0,$DB,$87,$E0,$9A,$FD,$26,$BF
	.dc.b	$45,$AF,$EB,$1B,$89,$BD,$4F,$47,$F5,$2D,$BF,$D6,$CF,$83,$1F,$9F
	.dc.b	$39,$C3,$F5,$7F,$6D,$FC,$35,$FD,$CB,$71,$34,$9D,$9F,$25,$B8,$FE
	.dc.b	$7D,$E1,$9B,$F1,$3F,$DD,$B6,$7F,$7F,$3F,$9D,$8C,$61,$FB,$CF,$C2
	.dc.b	$7E,$7B,$F9,$46,$BF,$93,$6B,$FC,$39,$B8,$9A,$48,$FB,$CE,$3E,$33
	.dc.b	$F8,$6A,$6C,$61,$FE,$23,$F8,$53,$EF,$FF,$0C,$DC,$BF,$78,$DA,$FD
	.dc.b	$DB,$5F,$D5,$B5,$FD,$53,$5F,$E4,$AD,$C0,$DF,$33,$EA,$6F,$FA,$46
	.dc.b	$DF,$2C,$3F,$92,$C6,$30,$FE,$5F,$F8,$63,$DF,$FE,$52,$DB,$FE,$5A
	.dc.b	$BE,$1B,$33,$AF,$A0,$FD,$14,$9A,$4A,$FE,$35,$F8,$97,$EE,$70,$FE
	.dc.b	$62,$BF,$10,$FE,$6E,$E7,$4F,$DD,$B9,$AF,$EA,$3F,$58,$BF,$C7,$37
	.dc.b	$97,$F8,$43,$67,$F9,$5B,$67,$FD,$A8,$D7,$F8,$F3,$5E,$8D,$4D,$E0
	.dc.b	$FE,$89,$C3,$FD,$33,$C5,$FC,$05,$F9,$2F,$F9,$74,$61,$FE,$9B,$F5
	.dc.b	$EF,$94,$FE,$A6,$BE,$02,$C6,$1B,$7F,$D6,$A7,$4D,$9D,$3F,$D7,$BF
	.dc.b	$01,$FE,$0E,$FF,$2D,$6D,$FF,$68,$7F,$C0,$F1,$B8,$FE,$C7,$1A,$C8
	.dc.b	$D3,$71,$FD,$91,$E6,$BE,$E1,$F3,$3F,$DC,$7C,$57,$DA,$BE,$C7,$FB
	.dc.b	$A5,$EE,$B5,$39,$CB,$FC,$13,$39,$7F,$C1,$77,$CA,$F8,$E1,$DA,$73
	.dc.b	$7C,$6F,$FB,$BF,$D8,$7E,$9F,$93,$F9,$FB,$6B,$F0,$98,$F3,$7F,$C0
	.dc.b	$71,$9A,$FF,$05,$FD,$17,$C1,$3D,$F7,$FD,$21,$C5,$FF,$82,$1B,$7F
	.dc.b	$E0,$C6,$D7,$F8,$23,$E7,$1F,$2D,$FE,$0E,$7F,$55,$B9,$CD,$FF,$85
	.dc.b	$1C,$98,$AE,$5E,$37,$1B,$E4,$FF,$87,$79,$BF,$CB,$5F,$6B,$F9,$CF
	.dc.b	$5C,$D2,$63,$F9,$E8,$CB,$63,$38,$7E,$82,$72,$FF,$82,$31,$FF,$0F
	.dc.b	$57,$97,$9F,$C1,$5A,$DB,$FB,$5D,$6A,$FE,$07,$D4,$7B,$CF,$81,$E5
	.dc.b	$5C,$7D,$D1,$67,$E3,$4D,$BF,$F2,$67,$E3,$9F,$C1,$6F,$FC,$6F,$F4
	.dc.b	$1F,$CE,$27,$3E,$07,$F8,$E2,$31,$3F,$E9,$FA,$E1,$87,$F8,$E2,$72
	.dc.b	$FF,$A8,$23,$37,$FE,$02,$9C,$ED,$FE,$83,$FA,$5F,$94,$4C,$3E,$43
	.dc.b	$70,$F8,$CD,$7F,$82,$EA,$8F,$EF,$D6,$F8,$7F,$31,$F2,$22,$DF,$B3
	.dc.b	$A7,$D5,$F2,$FE,$B6,$C6,$4F,$83,$19,$69,$F1,$3E,$97,$E1,$7B,$BF
	.dc.b	$0D,$B1,$93,$E4,$C6,$5A,$7C,$CF,$8D,$F3,$91,$FF,$25,$64,$8F,$5E
	.dc.b	$3F,$3C,$FA,$0D,$9F,$AB,$1F,$4F,$EB,$31,$AF,$D1,$6E,$FF,$3D,$B5
	.dc.b	$FE,$3F,$C1,$1F,$6F,$E5,$BE,$49,$27,$C6,$FC,$CF,$A1,$FC,$01,$B8
	.dc.b	$FC,$2B,$F8,$3F,$C2,$3E,$89,$23,$D6,$FD,$4F,$6F,$F1,$06,$E3,$F1
	.dc.b	$AE,$1A,$7E,$42,$31,$97,$E2,$E6,$E5,$F8,$99,$B5,$EB,$F4,$3F,$02
	.dc.b	$47,$E5,$BF,$84,$F7,$7F,$30,$7E,$65,$B1,$F9,$EF,$95,$FC,$49,$1F
	.dc.b	$A6,$DC,$7F,$CF,$7A,$A3,$11,$FE,$6C,$8C,$4F,$DA,$FA,$BF,$D3,$1F
	.dc.b	$A7,$AB,$1F,$EA,$DF,$84,$FE,$A5,$7F,$EC,$0F,$37,$FC,$D9,$EE,$7F
	.dc.b	$AC,$1B,$3A,$9F,$F6,$5C,$60,$95,$FE,$CC,$7F,$43,$F3,$DF,$B5,$3F
	.dc.b	$D6,$57,$CE,$7F,$77,$19,$69,$E3,$9C,$57,$FA,$F6,$9F,$53,$FE,$5A
	.dc.b	$F8,$2F,$E5,$72,$9C,$FB,$7F,$E9,$F9,$CC,$70,$C3,$1E,$3F,$F8,$0E
	.dc.b	$37,$1F,$E0,$59,$DE,$1D,$6B,$4D,$C3,$B3,$FE,$B1,$AE,$13,$AC,$3A
	.dc.b	$D6,$9B,$87,$68,$61,$DA,$F3,$87,$59,$79,$2C,$FF,$DF,$31,$C3,$2F
	.dc.b	$F5,$74,$E7,$CA,$F0,$CB,$EA,$DF,$59,$F8,$08,$E0,$7F,$88,$6B,$83
	.dc.b	$83,$BE,$1D,$B3,$F9,$71,$8C,$3A,$E1,$D3,$83,$DB,$39,$39,$61,$D1
	.dc.b	$FF,$61,$19,$2B,$97,$D3,$87,$B1,$CF,$0B,$C7,$FF,$6D,$C7,$C2,$8C
	.dc.b	$3F,$ED,$EA,$46,$23,$8C,$F0,$9C,$BF,$12,$39,$36,$33,$1F,$13,$07
	.dc.b	$D6,$3E,$47,$FD,$B5,$A7,$C2,$BD,$B7,$1F,$1B,$71,$F8,$B7,$C7,$1F
	.dc.b	$A7,$C2,$FC,$95,$F6,$9D,$38,$7F,$21,$18,$AD,$B5,$FA,$F5,$C7,$E3
	.dc.b	$B8,$47,$E7,$67,$59,$7E,$71,$B7,$D3,$0D,$E1,$C6,$FA,$C6,$2B,$71
	.dc.b	$CE,$31,$87,$6F,$45,$D3,$FE,$08,$AF,$ED,$FE,$87,$FB,$F7,$BB,$FB
	.dc.b	$33,$7A,$3D,$7F,$15,$69,$87,$2A,$C3,$63,$99,$FE,$73,$B5,$F2,$8C
	.dc.b	$35,$FC,$5E,$9F,$2F,$0F,$C2,$B6,$3F,$1E,$7C,$58,$CE,$86,$92,$37
	.dc.b	$3C,$5F,$E5,$3D,$5E,$6C,$FF,$DD,$31,$FE,$0E,$8C,$61,$23,$FB,$1B
	.dc.b	$F9,$6C,$BE,$3B,$6B,$D7,$3A,$21,$FC,$DA,$34,$71,$E3,$FF,$50,$47
	.dc.b	$F9,$1F,$CE,$F8,$F1,$88,$69,$FE,$4F,$EC,$C9,$C4,$91,$F0,$B1,$FE
	.dc.b	$13,$1D,$7E,$D3,$99,$27,$FC,$A9,$F1,$35,$EB,$97,$FA,$96,$FF,$88
	.dc.b	$D6,$70,$FF,$2F,$5D,$AD,$62,$AC,$61,$FE,$66,$AF,$49,$ED,$23,$E5
	.dc.b	$94,$7C,$1F,$9F,$F7,$92,$7F,$CD,$1F,$73,$F0,$8F,$8C,$47,$93,$F2
	.dc.b	$7E,$D3,$E3,$FD,$D4,$9F,$F3,$84,$7C,$A3,$0F,$B8,$FF,$3D,$C6,$5D
	.dc.b	$90,$D6,$9F,$CC,$A7,$E9,$FC,$04,$9F,$53,$F6,$48,$FB,$7E,$AB,$EC
	.dc.b	$FC,$04,$8F,$F4,$14,$7D,$C3,$0F,$6D,$3F,$D7,$CE,$8F,$F4,$47,$D0
	.dc.b	$FE,$0B,$E8,$A4,$60,$87,$17,$2F,$38,$FC,$4F,$DE,$48,$FF,$49,$3F
	.dc.b	$B5,$5F,$5E,$FA,$6C,$36,$3F,$98,$B5,$FD,$6B,$5F,$E7,$AD,$7F,$DD
	.dc.b	$87,$48,$FE,$72,$D3,$9B,$88,$E1,$1F,$E7,$1F,$9D,$FC,$F9,$1F,$E0
	.dc.b	$26,$4B,$D4,$F7,$53,$B9,$F9,$4F,$F1,$19,$D3,$FA,$AF,$A3,$93,$FE
	.dc.b	$B3,$7F,$5B,$F1,$38,$8F,$A5,$17,$A7,$FD,$79,$3F,$B2,$F8,$E9,$5F
	.dc.b	$EB,$B7,$E5,$7D,$77,$F9,$5A,$3F,$DA,$AD,$7F,$89,$B5,$FE,$C6,$D7
	.dc.b	$E3,$B5,$07,$0D,$F6,$3E,$5C,$60,$E1,$93,$F1,$17,$D9,$FC,$15,$F3
	.dc.b	$F4,$C4,$4C,$3E,$5C,$E9,$B3,$97,$FD,$CF,$EC,$7F,$38,$F8,$97,$B7
	.dc.b	$4C,$44,$9F,$F7,$3C,$65,$FF,$74,$C7,$F4,$0F,$8D,$FA,$F1,$F7,$3B
	.dc.b	$51,$88,$FD,$7B,$78,$3F,$94,$C3,$EC,$1D,$1F,$99,$F5,$7F,$B3,$56
	.dc.b	$A7,$35,$F8,$83,$85,$6F,$CE,$91,$99,$DB,$FC,$0C,$D9,$C6,$FA,$F8
	.dc.b	$DF,$C8,$FC,$DF,$EC,$53,$FC,$11,$B8,$FF,$3E,$CF,$FD,$DB,$1F,$D7
	.dc.b	$8F,$F1,$C4,$FB,$DD,$D7,$37,$46,$EF,$FB,$2F,$C6,$BD,$EB,$B4,$C3
	.dc.b	$E4,$4E,$5B,$39,$7F,$8B,$C2,$79,$CF,$F7,$0F,$8B,$75,$52,$5F,$C7
	.dc.b	$9F,$F1,$56,$9F,$41,$FF,$16,$47,$F8,$E9,$AF,$EF,$9A,$FE,$61,$AF
	.dc.b	$A7,$D5,$FE,$03,$18,$23,$FD,$9F,$DB,$7B,$FF,$E0,$45,$EF,$D3,$11
	.dc.b	$FE,$08,$F9,$2F,$E4,$B8,$7F,$93,$38,$51,$88,$FE,$23,$18,$F2,$3F
	.dc.b	$BA,$4E,$67,$FA,$E7,$CA,$3E,$4F,$FC,$26,$D8,$FC,$A3,$5F,$E8,$2B
	.dc.b	$F0,$7F,$F0,$CF,$CB,$7F,$B8,$67,$F3,$4D,$7F,$A3,$B7,$7F,$C2,$3F
	.dc.b	$E5,$68,$DF,$BD,$FE,$07,$8C,$47,$AD,$AF,$A9,$AF,$F7,$36,$BF,$E2
	.dc.b	$C6,$E6,$69,$8F,$97,$FB,$F3,$FA,$F6,$BF,$D7,$1B,$7F,$99,$F7,$FF
	.dc.b	$81,$4F,$F6,$06,$BF,$31,$AF,$F8,$15,$AF,$F8,$F9,$B7,$FE,$6A,$F0
	.dc.b	$3F,$C4,$91,$82,$3F,$E2,$7F,$98,$7A,$BF,$E4,$46,$DF,$F0,$56,$BF
	.dc.b	$DC,$BE,$69,$F3,$FF,$E4,$A6,$C7,$F8,$7D,$AF,$E1,$5A,$FE,$11,$AF
	.dc.b	$F8,$75,$AF,$F9,$59,$B8,$9A,$63,$E7,$3F,$61,$3F,$E0,$46,$E7,$FC
	.dc.b	$05,$BD,$FF,$E2,$9F,$94,$FC,$BC,$FF,$05,$7F,$25,$5F,$E6,$46,$E0
	.dc.b	$69,$39,$7F,$C9,$8D,$9F,$F1,$E3,$C2,$5F,$17,$FE,$6C,$6C,$FF,$2D
	.dc.b	$6B,$FE,$1A,$F2,$BE,$FE,$3F,$E3,$D7,$F0,$71,$FA,$86,$B9,$ED,$FF
	.dc.b	$2E,$46,$23,$FC,$F6,$DA,$FF,$77,$4E,$89,$3F,$E5,$1F,$9E,$7C,$8F
	.dc.b	$FA,$01,$B1,$FD,$05,$AF,$F8,$BB,$E1,$FF,$C9,$4F,$F4,$56,$BF,$BF
	.dc.b	$6B,$FB,$E6,$BF,$E5,$76,$BF,$E9,$06,$DF,$FB,$BB,$DF,$FF,$3B,$46
	.dc.b	$0C,$FC,$AF,$D1,$AF,$EA,$8D,$7F,$C8,$2D,$DF,$F9,$F7,$E2,$1F,$94
	.dc.b	$6F,$2F,$FA,$06,$3F,$2F,$5F,$C4,$9A,$FF,$9A,$DA,$FF,$A9,$3E,$2F
	.dc.b	$F0,$ED,$F1,$3F,$D4,$B1,$8C,$FF,$DF,$5E,$36,$26,$57,$F6,$9A,$D1
	.dc.b	$FC,$FD,$FE,$D9,$F2,$CF,$5B,$FD,$5E,$D8,$CC,$63,$D7,$CB,$23,$27
	.dc.b	$28,$FF,$15,$7D,$17,$F5,$A8,$FF,$34,$72,$23,$1B,$7F,$A7,$A3,$1E
	.dc.b	$67,$F8,$C7,$E4,$5E,$CF,$28,$98,$7F,$8E,$B0,$E3,$D1,$FE,$20,$C1
	.dc.b	$87,$0A,$FF,$1F,$B5,$FF,$60,$B5,$FF,$27,$C7,$0C,$3F,$D8,$73,$99
	.dc.b	$FF,$29,$FD,$1F,$F8,$8B,$2F,$F1,$6F,$77,$C3,$94,$FF,$8A,$1C,$EB
	.dc.b	$C3,$D8,$FF,$2D,$D6,$EB,$95,$7F,$05,$AE,$BE,$17,$FA,$D2,$30,$7F
	.dc.b	$8E,$A7,$A3,$83,$BD,$F4,$AE,$38,$78,$27,$85,$F3,$7F,$CD,$67,$99
	.dc.b	$C3,$C0,$FF,$6F,$DF,$91,$DA,$F8,$D7,$0F,$FB,$92,$7D,$F1,$87,$8B
	.dc.b	$5E,$2D,$7F,$DD,$7D,$3F,$E7,$A4,$D2,$3F,$E7,$E7,$D4,$FF,$96,$BE
	.dc.b	$47,$FD,$C3,$A7,$CF,$8F,$F2,$EC,$E9,$4E,$B3,$EF,$1F,$1C,$CF,$06
	.dc.b	$DF,$09,$99,$E0,$B8,$3F,$13,$5F,$E4,$88,$FC,$66,$FF,$A3,$BF,$F7
	.dc.b	$FB,$63,$F1,$73,$B7,$97,$0B,$E5,$7F,$DF,$3C,$1F,$9D,$8C,$61,$F9
	.dc.b	$4B,$CB,$F9,$98,$CD,$FE,$99,$B5,$C6,$F7,$3E,$7C,$3C,$53,$9B,$FF
	.dc.b	$32,$7D,$1F,$F4,$36,$F8,$5C,$3A,$92,$F4,$F0,$EF,$1F,$F6,$BB,$70
	.dc.b	$FF,$2A,$37,$0F,$F7,$E3,$C5,$EA,$7F,$9C,$FE,$8D,$F2,$1D,$DB,$19
	.dc.b	$F2,$31,$32,$38,$3B,$8C,$61,$FE,$E8,$9F,$F1,$E9,$F9,$A5,$92,$33
	.dc.b	$1F,$E3,$B6,$BF,$E3,$A6,$BF,$CF,$77,$F8,$5C,$5F,$22,$7F,$D4,$F5
	.dc.b	$CF,$D2,$CD,$7C,$7F,$13,$FC,$D5,$5F,$D4,$FB,$FE,$F5,$A7,$38,$FF
	.dc.b	$92,$DA,$FF,$92,$9A,$F6,$6B,$EA,$6C,$73,$9E,$BC,$AF,$A3,$FB,$0D
	.dc.b	$C3,$FB,$04,$E2,$7E,$E7,$33,$E3,$7C,$AD,$AF,$91,$87,$E7,$67,$24
	.dc.b	$8C,$CE,$F2,$FF,$32,$35,$FE,$F9,$3F,$03,$93,$FD,$9B,$5F,$EB,$16
	.dc.b	$CF,$F8,$16,$7F,$1D,$18,$E3,$FF,$0E,$7C,$8F,$C8,$F3,$0D,$26,$1E
	.dc.b	$68,$CB,$63,$38,$7F,$9D,$E7,$2F,$F8,$D7,$E7,$7F,$39,$D6,$7F,$3D
	.dc.b	$3A,$C3,$FD,$59,$39,$C3,$FC,$89,$C3,$F6,$BC,$EF,$DD,$63,$FE,$80
	.dc.b	$6B,$F8,$46,$BF,$90,$6E,$C6,$92,$7F,$0B,$9B,$E8,$FF,$60,$DC,$3C
	.dc.b	$27,$FB,$65,$C6,$1F,$E6,$B3,$F1,$BE,$8F,$CC,$7F,$D2,$6D,$7F,$2E
	.dc.b	$D7,$F4,$6D,$C4,$D2,$47,$4F,$91,$FE,$05,$06,$C6,$1F,$E0,$DE,$9F
	.dc.b	$FA,$AB,$0F,$C4,$B6,$7F,$D1,$DE,$CF,$E8,$A3,$11,$FE,$8F,$8C,$12
	.dc.b	$78,$BF,$A8,$F5,$FE,$94,$FF,$A9,$38,$7B,$CF,$F1,$59,$CF,$07,$FB
	.dc.b	$8F,$D9,$7D,$1F,$F1,$96,$E3,$FF,$00,$5F,$F0,$2B,$63,$5F,$27,$FB
	.dc.b	$69,$34,$93,$FA,$4F,$A3,$FD,$66,$6C,$3F,$DE,$51,$FE,$BD,$94,$E5
	.dc.b	$FE,$0F,$1C,$27,$7E,$E7,$F3,$4F,$53,$7C,$9F,$D0,$DB,$7F,$C6,$5B
	.dc.b	$87,$F7,$46,$BD,$83,$87,$C2,$D7,$27,$E5,$1C,$71,$7F,$3B,$E1,$FE
	.dc.b	$7D,$E2,$FD,$DB,$F4,$E5,$FC,$9F,$0E,$38,$EA,$FF,$A2,$7D,$1B,$DA
	.dc.b	$BF,$25,$86,$E3,$C8,$FE,$4F,$3B,$71,$87,$F3,$48,$57,$19,$CE,$69
	.dc.b	$CC,$63,$E1,$BF,$83,$3F,$CC,$9B,$7F,$CF,$9F,$EF,$2F,$D1,$30,$7F
	.dc.b	$5B,$77,$46,$9C,$33,$F8,$11,$99,$FE,$D2,$3E,$09,$F4,$7F,$DB,$3C
	.dc.b	$47,$D1,$BE,$87,$75,$AC,$FD,$F5,$A9,$CD,$F1,$BE,$15,$F1,$30,$ED
	.dc.b	$39,$AF,$EE,$5F,$17,$FA,$BC,$FC,$AD,$BF,$F1,$06,$1C,$B0,$F8,$19
	.dc.b	$AB,$51,$FD,$8B,$E5,$3E,$09,$EF,$BF,$EB,$6E,$33,$FF,$00,$36,$3E
	.dc.b	$9B,$67,$AF,$D1,$BE,$73,$FC,$0C,$FE,$CA,$BF,$B2,$B9,$7E,$DB,$99
	.dc.b	$FE,$E6,$E5,$E7,$DD,$CA,$B9,$BF,$A5,$F5,$9F,$15,$FA,$56,$E1,$FA
	.dc.b	$56,$DF,$E9,$5A,$FE,$97,$A7,$FC,$A7,$5C,$49,$5C,$5F,$EE,$B8,$FF
	.dc.b	$85,$6B,$DB,$D3,$E1,$9F,$6B,$EE,$F8,$A9,$7E,$CF,$26,$DF,$C4,$9B
	.dc.b	$C5,$F8,$AC,$7F,$C4,$53,$F0,$27,$FC,$67,$E5,$BE,$4B,$F8,$7E,$3F
	.dc.b	$A0,$F9,$BF,$E7,$B8,$FF,$A7,$67,$38,$7D,$C9,$CD,$7F,$A6,$67,$35
	.dc.b	$F9,$89,$CD,$7F,$95,$67,$35,$FE,$18,$9C,$EB,$FE,$76,$FA,$0F,$98
	.dc.b	$49,$F9,$8D,$BF,$98,$D7,$E6,$6E,$8F,$F0,$ED,$BE,$8F,$FC,$90,$DB
	.dc.b	$FD,$FE,$1D,$DF,$74,$EB,$A3,$7F,$11,$67,$E1,$37,$2F,$80,$DD,$3D
	.dc.b	$CF,$6F,$A3,$EF,$EE,$5B,$F9,$4D,$BF,$90,$DB,$F8,$CF,$9B,$E8,$FD
	.dc.b	$24,$AF,$F1,$0E,$88,$F6,$EB,$EC,$38,$4F,$D7,$70,$FD,$9F,$97,$B8
	.dc.b	$FD,$76,$E3,$F5,$5B,$3F,$4D,$F8,$FF,$1B,$F2,$89,$1E,$27,$E6,$FC
	.dc.b	$AF,$CC,$37,$1F,$87,$7E,$7F,$C5,$7D,$22,$47,$A9,$FA,$BE,$CF,$D4
	.dc.b	$37,$1F,$8F,$70,$D3,$F2,$51,$8C,$7F,$21,$37,$1F,$C7,$4D,$AF,$C6
	.dc.b	$7A,$1F,$82,$23,$F2,$9F,$C2,$FB,$7F,$84,$3F,$0A,$D8,$FD,$07,$D1
	.dc.b	$7E,$28,$8F,$D2,$6E,$3F,$ED,$4F,$0D,$C0,$9C,$CF,$F6,$39,$CC,$7E
	.dc.b	$53,$E4,$7F,$26,$7E,$51,$B7,$FE,$FE,$F3,$3F,$D6,$5D,$6F,$83,$FE
	.dc.b	$C5,$8F,$FC,$00,$9C,$C6,$23,$FE,$00,$B5,$FF,$5F,$C6,$89,$5F,$8A
	.dc.b	$F8,$AF,$D1,$9F,$C0,$2B,$AE,$FF,$B7,$8C,$CF,$F0,$17,$55,$8B,$FE
	.dc.b	$B5,$83,$E8,$FF,$CB,$7F,$03,$FC,$C6,$7F,$DA,$51,$8F,$1B,$FC,$23
	.dc.b	$5B,$C7,$FB,$D5,$EE,$39,$7B,$58,$AD,$DF,$1C,$39,$CE,$5E,$B8,$5B
	.dc.b	$E2,$FF,$BB,$2F,$8D,$66,$F9,$57,$86,$B5,$7D,$1A,$BC,$86,$7F,$EF
	.dc.b	$48,$FF,$7F,$E3,$FE,$1D,$9C,$F9,$3C,$BE,$35,$F2,$9F,$81,$C7,$9D
	.dc.b	$6B,$0E,$27,$F8,$C9,$C6,$BF,$1E,$3C,$19,$77,$8F,$95,$87,$4E,$1F
	.dc.b	$81,$5A,$CB,$AD,$F9,$63,$FC,$60,$64,$FF,$67,$61,$E1,$72,$D7,$D5
	.dc.b	$C4,$F1,$BF,$DA,$91,$F0,$63,$0F,$26,$BC,$9A,$F2,$8F,$7C,$70,$CB
	.dc.b	$D9,$48,$C6,$BF,$E5,$2C,$DF,$26,$FA,$3F,$F2,$46,$1C,$2F,$79,$7C
	.dc.b	$76,$BE,$5C,$BE,$85,$F1,$8F,$B7,$C0,$FA,$37,$CC,$74,$D3,$F2,$31
	.dc.b	$8A,$DB,$5F,$AB,$5C,$7E,$37,$84,$23,$A5,$E1,$B1,$CA,$31,$4A,$B1
	.dc.b	$27,$47,$82,$3A,$F4,$FF,$08,$F3,$FF,$CD,$1D,$3F,$D9,$DB,$CB,$FD
	.dc.b	$95,$B7,$E3,$C5,$F1,$58,$C3,$94,$FF,$77,$AC,$76,$7F,$85,$A3,$FC
	.dc.b	$BB,$A7,$C5,$BE,$6D,$8F,$C4,$9D,$63,$F0,$71,$98,$FC,$6B,$5F,$F3
	.dc.b	$F3,$5F,$9B,$1C,$30,$F2,$C7,$49,$D2,$C3,$F9,$3F,$57,$98,$91,$FE
	.dc.b	$0F,$CB,$D2,$7E,$52,$7F,$96,$B7,$1F,$F4,$14,$FF,$31,$6B,$FC,$BD
	.dc.b	$AF,$F8,$36,$38,$11,$F7,$CF,$EF,$FD,$DF,$95,$CD,$FE,$AE,$9F,$F2
	.dc.b	$6C,$63,$B2,$47,$09,$FF,$68,$36,$3C,$D1,$88,$FE,$88,$BE,$57,$A9
	.dc.b	$23,$FC,$B3,$F1,$15,$FF,$32,$CE,$DF,$F2,$EC,$F9,$27,$5C,$1F,$E8
	.dc.b	$2A,$D5,$70,$AD,$3D,$0F,$35,$79,$08,$FA,$7E,$57,$DE,$47,$FC,$D7
	.dc.b	$F4,$7F,$08,$F5,$91,$F8,$0F,$BB,$E8,$BE,$3F,$D1,$C9,$FF,$39,$C7
	.dc.b	$CA,$30,$FB,$23,$0E,$88,$7C,$08,$FA,$3E,$FC,$8F,$9C,$FC,$C2,$3F
	.dc.b	$47,$E8,$FE,$BF,$C0,$C8,$FF,$41,$C7,$DA,$30,$FC,$C8,$C3,$FE,$88
	.dc.b	$F9,$DF,$BF,$F4,$72,$30,$43,$E9,$B9,$7E,$AC,$7E,$1B,$E8,$E4,$7F
	.dc.b	$A4,$9F,$DA,$AF,$A2,$7D,$1E,$1B,$1F,$CB,$9A,$FE,$C1,$AF,$F3,$A6
	.dc.b	$BF,$EE,$B6,$BF,$EF,$A6,$BD,$E3,$F9,$CB,$4F,$D8,$7B,$8F,$9D,$FC
	.dc.b	$E9,$23,$F8,$1C,$65,$B1,$97,$F2,$6F,$F0,$68,$D4,$7E,$5B,$E7,$FF
	.dc.b	$4E,$47,$FD,$67,$F1,$35,$D1,$FE,$B6,$8F,$D6,$C7,$EB,$BE,$7A,$4F
	.dc.b	$FA,$E9,$F9,$3F,$46,$F6,$29,$EC,$71,$3F,$D5,$9A,$FF,$18,$6B,$FD
	.dc.b	$89,$A8,$8E,$1B,$1D,$4F,$E3,$FF,$46,$FA,$0F,$E4,$AB,$E8,$29,$88
	.dc.b	$95,$FC,$B2,$74,$D9,$CB,$FE,$E0,$F8,$97,$B3,$FC,$DD,$7B,$74,$C4
	.dc.b	$49,$F4,$77,$FF,$BB,$27,$51,$8F,$0F,$D9,$E1,$FF,$BA,$EF,$FA,$17
	.dc.b	$C6,$FD,$6C,$15,$A8,$C3,$78,$FF,$2F,$F4,$6E,$8F,$CC,$FA,$BF,$D8
	.dc.b	$F0,$FE,$5B,$39,$CF,$C0,$7C,$B7,$0A,$3F,$99,$CF,$1C,$3E,$84,$FF
	.dc.b	$8A,$BC,$EE,$F8,$48,$E6,$64,$EB,$A7,$FB,$B3,$CD,$FE,$B8,$7F,$59
	.dc.b	$8E,$D7,$F0,$23,$18,$FB,$A7,$FA,$64,$7F,$8E,$1E,$6E,$A7,$83,$D0
	.dc.b	$EB,$8F,$F6,$9F,$8E,$7B,$D7,$79,$87,$92,$72,$D9,$CB,$E4,$3F,$AE
	.dc.b	$47,$DE,$74,$7B,$3E,$2D,$D5,$FD,$DD,$B1,$98,$FE,$13,$5F,$E2,$EC
	.dc.b	$FF,$AE,$BF,$E2,$E9,$F9,$6D,$7F,$38,$D7,$D3,$EA,$7F,$02,$8C,$11
	.dc.b	$C1,$FD,$AF,$DB,$7B,$FF,$E0,$95,$EF,$D3,$11,$FE,$0D,$F9,$27,$8B
	.dc.b	$FD,$CD,$B8,$FF,$6A,$6B,$FE,$A3,$6B,$FE,$A1,$F9,$47,$C9,$7F,$85
	.dc.b	$DB,$1F,$80,$6B,$F2,$F9,$10,$FA,$3F,$F0,$F7,$CB,$7F,$B8,$6B,$F0
	.dc.b	$ED,$78,$37,$7F,$C2,$5A,$3F,$CA,$53,$C3,$DD,$FF,$06,$46,$23,$FD
	.dc.b	$74,$D7,$FD,$70,$D7,$F5,$ED,$7F,$C3,$4D,$9E,$D3,$D0,$FF,$0C,$7C
	.dc.b	$BF,$DE,$AF,$63,$6F,$CA,$DE,$8F,$DD,$FB,$DF,$C2,$AB,$E1,$35,$FF
	.dc.b	$65,$35,$DF,$83,$FE,$25,$8C,$47,$F8,$B5,$B8,$FF,$9B,$A7,$44,$9F
	.dc.b	$F1,$0F,$CC,$BD,$3F,$F2,$53,$63,$F6,$6D,$7E,$A7,$CD,$3E,$7B,$FC
	.dc.b	$A0,$D8,$FF,$15,$B5,$FF,$72,$35,$FF,$71,$35,$FE,$32,$D7,$FC,$88
	.dc.b	$DC,$0D,$31,$F3,$9F,$B2,$9F,$C3,$37,$5F,$BE,$DD,$3F,$92,$FC,$AF
	.dc.b	$E5,$27,$F0,$6F,$E5,$2B,$3C,$BF,$E4,$D8,$C4,$7F,$95,$5B,$89,$A4
	.dc.b	$9E,$07,$F9,$56,$1F,$17,$FE,$73,$6D,$7F,$1C,$6B,$F9,$DF,$80,$FC
	.dc.b	$03,$FE,$63,$7F,$0C,$FF,$28,$8F,$F4,$87,$6F,$F9,$86,$73,$3F,$D0
	.dc.b	$1A,$F1,$8D,$11,$FF,$2E,$C6,$A5,$F3,$CF,$91,$FF,$44,$36,$BF,$98
	.dc.b	$B5,$FD,$97,$C3,$FF,$93,$B4,$FD,$5B,$FE,$50,$8F,$33,$5F,$EA,$6D
	.dc.b	$7F,$CF,$CD,$B3,$49,$EF,$7F,$9F,$5D,$7C,$A7,$CF,$AF,$E8,$6D,$7F
	.dc.b	$82,$36,$FF,$7B,$F4,$7F,$95,$6F,$2F,$FA,$1E,$3F,$31,$5F,$2D,$AF
	.dc.b	$F6,$86,$BF,$E9,$3F,$8B,$FC,$43,$7C,$4F,$F5,$64,$63,$33,$FC,$F6
	.dc.b	$7F,$8B,$39,$24,$F0,$F1,$FF,$D6,$9F,$2C,$F5,$E9,$2F,$FA,$FC,$7F
	.dc.b	$74,$7F,$D4,$8E,$DC,$C6,$DF,$E4,$91,$BE,$17,$D1,$7F,$5F,$9F,$EA
	.dc.b	$51,$8F,$2F,$FD,$7B,$39,$DB,$FB,$34,$BE,$45,$EC,$EA,$E6,$1F,$EC
	.dc.b	$6C,$38,$75,$7F,$8A,$70,$E1,$7F,$EC,$8C,$38,$DF,$FA,$F5,$AF,$FB
	.dc.b	$19,$AF,$FB,$19,$AF,$F8,$13,$E8,$FF,$C5,$99,$7F,$8D,$7B,$7E,$04
	.dc.b	$A7,$FC,$64,$E7,$5E,$1E,$C7,$F8,$32,$B7,$5C,$AB,$F8,$3D,$75,$F0
	.dc.b	$7F,$C2,$91,$83,$FC,$85,$3D,$1C,$1D,$EF,$A5,$71,$C3,$C1,$3C,$2F
	.dc.b	$9B,$FE,$1C,$3C,$CE,$1E,$07,$FB,$AA,$FC,$8E,$D7,$C6,$B8,$3F,$DD
	.dc.b	$F3,$EF,$8C,$3C,$5A,$F1,$6B,$FE,$F6,$E8,$FF,$3F,$BF,$AC,$3F,$D0
	.dc.b	$2E,$08,$FB,$87,$C8,$7F,$89,$F4,$F9,$F1,$FE,$65,$9F,$7B,$FE,$67
	.dc.b	$9D,$67,$DE,$39,$46,$1F,$85,$3C,$1E,$BC,$1F,$23,$81,$F4,$8F,$84
	.dc.b	$FF,$79,$C6,$1B,$19,$9F,$99,$4A,$D4,$61,$7E,$08,$F8,$BF,$F2,$07
	.dc.b	$0F,$E7,$23,$15,$FE,$56,$AE,$37,$FA,$49,$CB,$6F,$8D,$7E,$76,$73
	.dc.b	$7F,$A6,$C3,$C9,$7B,$AF,$F2,$AF,$D2,$BE,$4B,$8C,$BB,$13,$9A,$DF
	.dc.b	$CE,$EA,$73,$76,$B5,$7F,$E1,$F6,$E3,$FE,$18,$57,$BA,$E5,$F4,$AF
	.dc.b	$A5,$A7,$8B,$BD,$F5,$8F,$8A,$CE,$BF,$E6,$CB,$E0,$DB,$E1,$5B,$8C
	.dc.b	$1F,$E7,$23,$FC,$86,$7D,$4D,$7F,$D1,$7E,$C7,$F9,$BF,$6F,$EA,$D3
	.dc.b	$A7,$F4,$93,$A8,$E1,$C9,$FE,$D1,$7F,$CF,$AD,$9F,$F1,$F4,$71,$DF
	.dc.b	$FA,$6F,$C0,$FC,$CE,$47,$E5,$DF,$E9,$AA,$FC,$13,$6F,$F6,$53,$A9
	.dc.b	$FC,$6C,$E6,$7F,$D2,$33,$99,$E2,$68,$87,$9B,$E9,$BE,$B6,$27,$C5
	.dc.b	$7D,$F1,$FE,$C3,$1F,$6E,$DE,$4F,$C3,$C7,$FB,$A3,$5F,$EF,$8D,$CF
	.dc.b	$F3,$8D,$9E,$39,$7F,$76,$8C,$11,$FA,$1E,$7F,$EB,$2F,$FC,$C8,$DC
	.dc.b	$7F,$CB,$B7,$CA,$73,$D5,$F2,$7C,$FF,$F0,$9E,$1F,$A9,$6C,$FC,$86
	.dc.b	$BF,$19,$AF,$F3,$F6,$BF,$B3,$6D,$7F,$A0,$BC,$FF,$DD,$C6,$B3,$7E
	.dc.b	$20,$F9,$AF,$D0,$37,$1F,$DA,$57,$8B,$3F,$F3,$AC,$E7,$2F,$F4,$07
	.dc.b	$A9,$F8,$57,$FC,$44,$D7,$FD,$BE,$D7,$F6,$0D,$C4,$D2,$71,$7D,$17
	.dc.b	$5E,$AB,$DF,$FD,$A3,$73,$FF,$2F,$D6,$27,$F2,$5E,$B3,$E6,$BF,$8C
	.dc.b	$B6,$BF,$8A,$37,$2F,$C9,$B5,$FC,$8B,$5F,$ED,$4D,$7F,$8F,$37,$83
	.dc.b	$FC,$D9,$5F,$87,$F9,$97,$D0,$FF,$09,$6E,$1F,$C7,$6F,$C1,$BF,$FA
	.dc.b	$7A,$73,$8F,$FB,$4F,$B9,$F1,$4F,$4B,$F2,$91,$86,$C6,$71,$FE,$63
	.dc.b	$39,$6C,$66,$BF,$B0,$4E,$5B,$19,$7F,$DF,$F1,$97,$FD,$81,$F5,$3F
	.dc.b	$C8,$7E,$00,$FF,$17,$AE,$99,$7F,$AD,$A7,$38,$7F,$39,$FD,$83,$F0
	.dc.b	$93,$FE,$18,$6C,$30,$FF,$19,$FE,$2F,$F8,$B8,$FF,$0B,$4E,$A7,$F9
	.dc.b	$9F,$E5,$CF,$A5,$FD,$A7,$92,$8C,$5F,$F5,$08,$C4,$EA,$73,$5F,$C3
	.dc.b	$BF,$2E,$7B,$3F,$CE,$FE,$B3,$FD,$A9,$8F,$F4,$F9,$CB,$E5,$D3,$FD
	.dc.b	$A3,$3A,$B4,$27,$D5,$EF,$7F,$6F,$7F,$B0,$B6,$7F,$C1,$CD,$7E,$00
	.dc.b	$70,$F2,$3F,$83,$1A,$72,$EC,$C9,$BF,$07,$F8,$A7,$C5,$BE,$83,$FA
	.dc.b	$B3,$6F,$D4,$6C,$B5,$FD,$7E,$34,$FF,$BD,$FE,$B1,$F3,$BF,$D7,$DB
	.dc.b	$5C,$3C,$14,$3E,$7B,$55,$2B,$53,$C2,$FA,$56,$5F,$93,$3F,$81,$B7
	.dc.b	$CA,$B9,$BF,$B7,$3F,$CE,$67,$C0,$E5,$BE,$56,$A7,$6D,$8C,$9E,$EE
	.dc.b	$57,$E0,$DF,$DC,$A3,$F0,$CD,$C7,$FC,$85,$3A,$F5,$3F,$1B,$39,$9F
	.dc.b	$AF,$39,$AE,$35,$A2,$3F,$E2,$0F,$49,$F8,$9F,$C1,$E4,$FF,$3B,$F3
	.dc.b	$3E,$9B,$FC,$1F,$5A,$E2,$FE,$CD,$39,$C3,$E7,$7D,$27,$BF,$4F,$8E
	.dc.b	$7F,$85,$5C,$63,$F9,$18,$C6,$DF,$E1,$59,$CD,$7C,$D6,$CF,$1D,$F5
	.dc.b	$5B,$78,$47,$1E,$6B,$7E,$26,$C7,$8B,$89,$87,$F5,$9A,$ED,$3C,$5F
	.dc.b	$C6,$F3,$FF,$99,$7C,$2F,$ED,$4D,$8F,$D2,$FB,$E9,$CB,$63,$31,$82
	.dc.b	$38,$3F,$C3,$5D,$0F,$A4,$FC,$F9,$FE,$3A,$BE,$D8,$79,$E7,$55,$FE
	.dc.b	$4B,$F9,$D7,$D2,$7F,$8E,$FF,$13,$FE,$1D,$E5,$A7,$35,$F9,$F9,$CC
	.dc.b	$63,$0F,$85,$18,$BE,$5E,$97,$F9,$17,$89,$F4,$54,$FF,$11,$7E,$43
	.dc.b	$E2,$C3,$E5,$4E,$B2,$FD,$63,$63,$35,$C3,$6E,$55,$F3,$38,$3F,$BD
	.dc.b	$EF,$FE,$91,$F5,$7E,$49,$27,$E6,$B7,$0F,$E2,$35,$C0,$FE,$FD,$C0
	.dc.b	$F8,$EF,$E0,$E7,$C9,$8F,$E6,$53,$FC,$16,$3D,$9F,$61,$30,$FF,$3A
	.dc.b	$78,$5F,$C9,$AB,$D7,$39,$6C,$EB,$C6,$C7,$57,$E5,$FA,$3F,$C6,$9C
	.dc.b	$5F,$1C,$FF,$23,$7E,$46,$FA,$FC,$61,$F1,$26,$E3,$FE,$E4,$AD,$4E
	.dc.b	$4F,$93,$F7,$DF,$E3,$5E,$43,$49,$1F,$E6,$18,$FE,$75,$97,$8D,$B1
	.dc.b	$FC,$4D,$E2,$FD,$4F,$9A,$FB,$87,$DD,$FC,$16,$C7,$D8,$6E,$3F,$71
	.dc.b	$B7,$D5,$FC,$27,$CE,$79,$49,$5E,$37,$D1,$F0,$9F,$89,$37,$1F,$84
	.dc.b	$7F,$19,$EB,$7B,$49,$1E,$B7,$E0,$7B,$1F,$91,$37,$1F,$8C,$70,$D3
	.dc.b	$F1,$F1,$8C,$7F,$27,$37,$1F,$89,$36,$BE,$27,$D1,$FC,$D2,$3F,$2D
	.dc.b	$FA,$1F,$1D,$FA,$13,$F4,$4D,$8F,$CE,$FC,$87,$D5,$23,$F4,$DB,$8F
	.dc.b	$FA,$43,$64,$62,$3F,$B3,$F1,$FF,$9E,$E3,$EF,$7C,$8F,$EB,$CF,$D8
	.dc.b	$36,$FF,$D9,$1F,$5A,$FB,$7F,$F6,$BB,$63,$FD,$A4,$DC,$3F,$D2,$2D
	.dc.b	$7F,$D1,$ED,$7F,$C2,$DE,$2A,$31,$1F,$ED,$D6,$C6,$6B,$F2,$47,$F0
	.dc.b	$33,$FD,$45,$1D,$F6,$FC,$84,$66,$FF,$80,$BA,$31,$7F,$D0,$2F,$C1
	.dc.b	$8F,$F3,$FA,$F2,$B6,$7C,$B8,$23,$11,$E5,$6D,$72,$AF,$F1,$81,$C8
	.dc.b	$B3,$CA,$77,$5C,$E7,$33,$CD,$FF,$1C,$7C,$47,$FB,$A6,$3F,$DD,$1A
	.dc.b	$7F,$90,$14,$67,$0F,$04,$63,$E3,$3F,$5D,$EE,$FF,$59,$AE,$53,$FD
	.dc.b	$33,$AD,$E5,$F6,$BF,$F1,$3F,$97,$FE,$38,$8C,$53,$DE,$7D,$17,$F6
	.dc.b	$98,$EC,$7F,$93,$6B,$83,$83,$B5,$72,$69,$C2,$B9,$FA,$5E,$5C,$3C
	.dc.b	$38,$F9,$2B,$96,$1D,$1F,$F1,$D9,$C8,$AE,$5E,$38,$69,$F2,$7C,$FF
	.dc.b	$FA,$65,$C4,$A3,$11,$C9,$AF,$C2,$6B,$CB,$E0,$FD,$53,$C5,$B3,$D7
	.dc.b	$4F,$AD,$9B,$E1,$7F,$D7,$98,$6A,$E3,$EC,$BE,$19,$FF,$9D,$6F,$87
	.dc.b	$85,$FE,$58,$C7,$EE,$79,$7E,$7C,$3E,$9F,$4F,$FB,$BF,$D7,$D6,$3F
	.dc.b	$CA,$DB,$2E,$3F,$EF,$19,$E1,$B1,$F8,$E6,$BF,$E1,$D3,$86,$1D,$DC
	.dc.b	$46,$B9,$7F,$CD,$6D,$8F,$DB,$CE,$E7,$2D,$8C,$D7,$F9,$CA,$7F,$86
	.dc.b	$F2,$FE,$4B,$A7,$FA,$26,$5F,$E7,$66,$DF,$F9,$A3,$1F,$F0,$87,$C6
	.dc.b	$67,$0F,$E9,$17,$6B,$F4,$6D,$D1,$8F,$5A,$F2,$DE,$F0,$E7,$87,$E1
	.dc.b	$5A,$FE,$31,$AF,$AA,$3F,$14,$FC,$46,$C6,$A3,$A9,$CF,$34,$F0,$C3
	.dc.b	$C7,$E5,$F9,$71,$34,$C7,$C5,$4C,$3F,$B4,$3C,$64,$D2,$46,$88,$7F
	.dc.b	$3B,$78,$71,$77,$24,$7F,$A2,$FE,$9D,$F6,$BF,$1D,$B1,$FA,$7C,$51
	.dc.b	$88,$FD,$7C,$62,$7D,$93,$99,$FE,$8E,$FF,$82,$DE,$DE,$57,$53,$B7
	.dc.b	$CF,$58,$C9,$0F,$F2,$F4,$A3,$51,$89,$BC,$3F,$E1,$BA,$D4,$FA,$EB
	.dc.b	$C6,$47,$D5,$1A,$7D,$2B,$E2,$7E,$01,$23,$E2,$7D,$9F,$C3,$3D,$A4
	.dc.b	$79,$BF,$03,$EC,$FE,$47,$C3,$C9,$F9,$91,$F2,$CC,$3E,$C8,$C3,$A2
	.dc.b	$1F,$02,$3E,$97,$B5,$23,$D2,$FC,$E2,$3E,$CF,$B3,$FB,$04,$7F,$4D
	.dc.b	$E8,$FF,$A7,$DF,$BD,$19,$7E,$6B,$FA,$18,$D4,$7F,$5F,$FA,$C7,$9B
	.dc.b	$FA,$56,$C7,$EC,$5A,$FF,$65,$FC,$8D,$F3,$5B,$A9,$3B,$7F,$5D,$3A
	.dc.b	$8F,$F6,$D7,$3B,$E5,$FF,$6C,$D8,$CC,$7E,$ED,$7D,$87,$F9,$31,$FF
	.dc.b	$74,$98,$F5,$BF,$66,$D3,$F6,$B1,$B7,$FA,$03,$5F,$E8,$ED,$44,$70
	.dc.b	$D8,$EB,$18,$E7,$FD,$1F,$B7,$FA,$72,$57,$F1,$5F,$7B,$FD,$D4,$FE
	.dc.b	$B6,$32,$FE,$71,$FE,$21,$1A,$8F,$F0,$77,$05,$E7,$FF,$3B,$7D,$63
	.dc.b	$EA,$FF,$8D,$36,$BF,$D5,$5F,$09,$FC,$66,$F8,$78,$7F,$EF,$BA,$D5
	.dc.b	$7F,$B1,$DB,$87,$FB,$5F,$C3,$FF,$57,$61,$F4,$3E,$52,$5F,$D4,$7E
	.dc.b	$87,$D7,$F0,$C7,$E2,$CA,$3A,$BF,$DC,$1A,$E9,$AF,$CC,$6B,$C7,$3F
	.dc.b	$AF,$18,$23,$FC,$2B,$E5,$7F,$89,$7C,$AC,$AF,$E2,$D1,$FC,$54,$C3
	.dc.b	$FC,$1A,$30,$FE,$57,$D4,$FC,$C1,$1F,$CC,$7A,$F5,$FC,$F4,$5D,$7E
	.dc.b	$27,$57,$F9,$5E,$7F,$07,$1F,$EF,$DF,$8B,$7E,$25,$FA,$68,$F9,$5D
	.dc.b	$DF,$EE,$1A,$FE,$A8,$D5,$1C,$67,$86,$DF,$2B,$E9,$BF,$C0,$35,$FE
	.dc.b	$13,$6B,$FE,$1A,$6B,$BF,$47,$F3,$B1,$88,$FC,$BB,$53,$4D,$74,$39
	.dc.b	$7F,$BD,$63,$F8,$CF,$8B,$FC,$59,$F8,$CC,$66,$1F,$01,$FC,$54,$6A
	.dc.b	$30,$2F,$A4,$FF,$6D,$FA,$EF,$C3,$4E,$5B,$26,$9F,$D8,$A3,$9C,$67
	.dc.b	$8F,$91,$23,$33,$B7,$E5,$D7,$ED,$AB,$FB,$3F,$D3,$7F,$08,$24,$FF
	.dc.b	$04,$FD,$4B,$C6,$9F,$C8,$93,$0F,$E0,$ED,$8F,$EF,$6D,$7F,$C8,$CF
	.dc.b	$E6,$BA,$5C,$0F,$A3,$FF,$06,$7B,$BF,$72,$FE,$23,$6B,$73,$FE,$13
	.dc.b	$F9,$3F,$F2,$67,$A3,$FE,$08,$6D,$FF,$94,$DA,$FF,$96,$9A,$FF,$8B
	.dc.b	$3E,$6B,$FC,$30,$FF,$8E,$5B,$67,$3E,$D3,$F1,$A7,$C2,$FF,$88,$7E
	.dc.b	$BD,$E1,$D2,$63,$F7,$2D,$F8,$AF,$08,$C3,$5F,$F3,$6C,$6A,$FF,$C4
	.dc.b	$2D,$8F,$F3,$63,$5D,$CF,$17,$F8,$4C,$65,$B1,$97,$FC,$BC,$D7,$F1
	.dc.b	$DF,$3F,$F3,$DF,$C3,$B5,$FC,$19,$AD,$BF,$C7,$5F,$91,$BD,$3A,$7F
	.dc.b	$8D,$1C,$4F,$F8,$F5,$AF,$F8,$F5,$AF,$F9,$69,$AF,$FB,$41,$AF,$F9
	.dc.b	$A5,$BB,$1C,$F3,$30,$7C,$AF,$F9,$45,$B7,$EA,$8F,$F2,$A7,$D4,$FF
	.dc.b	$6A,$EC,$75,$7E,$76,$3F,$B6,$D7,$FB,$C1,$AE,$7D,$1F,$F2,$54,$62
	.dc.b	$3F,$68,$DC,$4D,$24,$FF,$94,$61,$EA,$7F,$99,$5B,$5F,$B4,$6B,$FE
	.dc.b	$E1,$FC,$65,$F2,$9F,$E6,$76,$C7,$F9,$9D,$AF,$F9,$9D,$AF,$FA,$11
	.dc.b	$AF,$FB,$E9,$AF,$FA,$4D,$BB,$1A,$6A,$BF,$88,$79,$3E,$72,$57,$F0
	.dc.b	$AA,$77,$FF,$9F,$DB,$1F,$E5,$96,$BF,$E5,$86,$BF,$D7,$2B,$FD,$49
	.dc.b	$3F,$E7,$59,$CC,$FF,$0F,$6B,$FC,$85,$B3,$C2,$3F,$CC,$DF,$2D,$FB
	.dc.b	$68,$34,$FD,$BF,$E4,$4F,$95,$FF,$48,$36,$BF,$D2,$0D,$7F,$D2,$0D
	.dc.b	$7F,$D5,$AD,$7F,$C2,$F6,$8C,$47,$FA,$F5,$B1,$99,$FF,$4A,$C6,$AF
	.dc.b	$CB,$F0,$1F,$EA,$5C,$3F,$DB,$FC,$3F,$92,$C7,$F9,$F4,$63,$87,$F9
	.dc.b	$87,$D7,$FE,$6A,$7F,$63,$82,$31,$1F,$2D,$AF,$0E,$3F,$F5,$04,$62
	.dc.b	$3F,$A5,$74,$FF,$3D,$C7,$C1,$68,$C6,$1F,$CF,$DB,$59,$F8,$D7,$AB
	.dc.b	$FE,$BD,$9C,$DF,$1F,$84,$C1,$FE,$C3,$85,$70,$BF,$EA,$3F,$31,$F8
	.dc.b	$4C,$BC,$2D,$CB,$FA,$73,$67,$FB,$02,$FB,$1F,$F6,$63,$9C,$8D,$24
	.dc.b	$7F,$B3,$7E,$B1,$F5,$B4,$97,$FE,$DB,$FA,$DF,$F6,$AB,$FE,$DD,$9D
	.dc.b	$B6,$76,$FF,$20,$E5,$57,$08,$C3,$5F,$F7,$9B,$5F,$C3,$B5,$47,$2F
	.dc.b	$89,$FF,$34,$FD,$73,$E5,$BF,$C3,$35,$FC,$2B,$0F,$F2,$0C,$BE,$5B
	.dc.b	$FC,$87,$E0,$79,$A9,$39,$7F,$CD,$71,$98,$E9,$C1,$8B,$E7,$87,$6A
	.dc.b	$D6,$1D,$27,$BB,$70,$FF,$3A,$E1,$C9,$F2,$18,$BF,$3B,$FE,$56,$3C
	.dc.b	$6E,$1E,$F8,$7F,$BE,$27,$94,$6D,$FE,$CF,$B6,$7C,$6F,$B7,$3C,$CF
	.dc.b	$6B,$87,$83,$5D,$7C,$8F,$FA,$49,$34,$8F,$FA,$59,$F8,$4F,$F9,$B3
	.dc.b	$E5,$3F,$61,$A7,$CB,$8F,$F4,$9C,$E9,$FF,$4A,$CE,$B4,$EF,$1F,$08
	.dc.b	$CF,$1E,$DE,$26,$44,$F1,$7F,$03,$C1,$FE,$B8,$9F,$77,$07,$E3,$3E
	.dc.b	$1F,$F8,$15,$7E,$29,$AF,$E2,$9A,$FE,$2B,$8B,$FC,$49,$5F,$99,$F2
	.dc.b	$3F,$85,$70,$7F,$A3,$F1,$ED,$7F,$99,$BD,$B6,$B7,$7F,$C7,$27,$37
	.dc.b	$C7,$0C,$DF,$7B,$BC,$AF,$39,$F2,$18,$6E,$5F,$E4,$E5,$F2,$59,$99
	.dc.b	$39,$78,$46,$30,$FE,$3B,$EE,$3E,$67,$BB,$6F,$FC,$51,$F5,$DF,$ED
	.dc.b	$97,$FC,$B7,$19,$AE,$35,$C0,$F4,$F6,$AF,$84,$EA,$FF,$CC,$63,$BB
	.dc.b	$E6,$E9,$FE,$04,$6C,$64,$FC,$A4,$65,$BC,$AF,$84,$F8,$AD,$8E,$B5
	.dc.b	$89,$F9,$7C,$8F,$AE,$FF,$08,$D7,$E0,$9B,$8F,$EA,$E7,$53,$F8,$06
	.dc.b	$BB,$31,$3C,$27,$FD,$8F,$3A,$79,$46,$5F,$F3,$D6,$0E,$EF,$39,$E4
	.dc.b	$9F,$0F,$27,$F4,$49,$D6,$1F,$DF,$FE,$BB,$E2,$E2,$69,$23,$FC,$00
	.dc.b	$DC,$7F,$C0,$0D,$7F,$C0,$1E,$BF,$E9,$27,$FD,$26,$D8,$FE,$BB,$5D
	.dc.b	$67,$3D,$3F,$23,$EB,$BF,$A8,$C7,$EA,$1B,$5F,$1D,$AF,$C2,$6B,$FD
	.dc.b	$E1,$AF,$F3,$26,$D1,$A6,$B7,$7D,$73,$E0,$BF,$42,$DC,$3F,$D6,$B7
	.dc.b	$C7,$C6,$CD,$7E,$9B,$D3,$FC,$16,$5F,$C9,$71,$FF,$2B,$71,$7E,$3A
	.dc.b	$73,$3F,$E2,$36,$BF,$EA,$F6,$BF,$E7,$48,$E0,$4C,$7F,$27,$F3,$2F
	.dc.b	$33,$F6,$AD,$8F,$F3,$CC,$E2,$7F,$23,$F3,$CF,$57,$F1,$8D,$AF,$E2
	.dc.b	$6D,$C3,$F2,$4D,$7F,$1A,$D7,$FC,$46,$D7,$FA,$EB,$75,$39,$CD,$F0
	.dc.b	$4F,$94,$FE,$16,$DC,$3E,$5D,$78,$79,$BF,$C2,$B3,$9C,$7F,$D2,$3D
	.dc.b	$6F,$51,$F1,$5F,$C9,$23,$0D,$8C,$E3,$FE,$A6,$6D,$7F,$AB,$1A,$FF
	.dc.b	$A9,$9A,$8F,$F1,$23,$86,$CF,$F0,$6F,$8C,$7C,$57,$F2,$46,$DF,$FA
	.dc.b	$F6,$FB,$E5,$FE,$2E,$9C,$E1,$FE,$0B,$F8,$1F,$C5,$37,$97,$F8,$BF
	.dc.b	$D7,$7E,$72,$FF,$C8,$9C,$3F,$EC,$F9,$FE,$67,$E3,$3E,$B9,$F2,$DF
	.dc.b	$D1,$23,$0D,$8C,$EB,$FE,$DF,$6E,$5F,$C3,$7E,$9D,$F3,$FF,$CF,$3E
	.dc.b	$B9,$F4,$BF,$C5,$A7,$38,$7C,$3D,$7F,$C8,$F5,$A8,$C1,$F1,$FD,$CF
	.dc.b	$EC,$78,$FF,$80,$DA,$FC,$B3,$FC,$EA,$7E,$FC,$C3,$A7,$91,$92,$F8
	.dc.b	$5F,$D8,$23,$51,$CA,$3F,$9B,$FC,$C3,$EB,$BF,$AA,$B6,$FC,$B5,$B6
	.dc.b	$F6,$27,$53,$FE,$6C,$FA,$E7,$D5,$FF,$5E,$6D,$F0,$5F,$53,$99,$93
	.dc.b	$97,$84,$62,$BF,$CE,$FF,$5C,$E9,$FE,$D2,$FF,$2C,$8C,$C6,$3C,$DC
	.dc.b	$B7,$0F,$3D,$71,$6C,$FE,$0B,$EB,$BF,$B6,$C7,$F8,$9F,$9D,$18,$C3
	.dc.b	$FC,$69,$18,$F4,$FF,$0D,$39,$9F,$A9,$39,$AE,$53,$99,$FD,$14,$E5
	.dc.b	$E7,$19,$5E,$9B,$EB,$BE,$3C,$7F,$99,$F9,$BF,$D0,$38,$7D,$2A,$D6
	.dc.b	$DF,$8D,$F8,$2F,$F0,$5F,$D4,$7B,$70,$FF,$4E,$B7,$A3,$FD,$2A,$FC
	.dc.b	$76,$F0,$70,$26,$3F,$72,$74,$FE,$E7,$91,$1C,$E3,$0F,$EF,$5E,$1C
	.dc.b	$DE,$2C,$3C,$6D,$8C,$9D,$63,$2F,$E3,$39,$FF,$D2,$3E,$2F,$E5,$7D
	.dc.b	$FF,$E3,$B3,$FE,$59,$9C,$B6,$33,$5F,$A5,$6B,$FA,$26,$BF,$AB,$6B
	.dc.b	$FE,$DC,$6E,$3F,$99,$FA,$7F,$CF,$9F,$E3,$6B,$EF,$5F,$EC,$6E,$94
	.dc.b	$FF,$84,$3E,$41,$F0,$FF,$E3,$AF,$AE,$FE,$23,$C1,$FC,$3A,$73,$A7
	.dc.b	$E9,$A7,$2F,$FB,$82,$33,$8F,$FB,$4A,$72,$FF,$A7,$23,$35,$FE,$CC
	.dc.b	$9C,$E7,$FE,$0F,$FA,$EF,$F2,$4F,$51,$FF,$11,$FD,$07,$9B,$2F,$DA
	.dc.b	$CE,$5B,$3C,$0D,$E9,$63,$E9,$6F,$FD,$F3,$7F,$F6,$1F,$A9,$F4,$C9
	.dc.b	$3F,$45,$BB,$7D,$16,$BF,$C5,$73,$FE,$09,$F1,$FF,$C2,$8F,$99,$87
	.dc.b	$F3,$3B,$FC,$67,$0A,$7D,$7F,$1B,$27,$FC,$07,$E0,$7F,$28,$C7,$FD
	.dc.b	$EF,$B2,$B5,$18,$7F,$C8,$73,$BE,$1F,$F7,$9D,$6E,$FE,$1F,$91,$F2
	.dc.b	$0F,$F2,$4F,$D2,$BE,$A3,$13,$A2,$65,$FE,$D7,$9D,$46,$30,$EB,$F4
	.dc.b	$9F,$E5,$EC,$4D,$24,$FF,$97,$A3,$F9,$F5,$FA,$BC,$8F,$B3,$5F,$69
	.dc.b	$95,$F6,$3E,$BA,$D7,$C0,$6E,$1E,$C6,$CF,$ED,$70,$47,$E1,$3E,$BB
	.dc.b	$C8,$49,$F1,$3E,$7F,$84,$FC,$49,$B8,$FC,$3B,$F8,$CF,$5B,$D8,$48
	.dc.b	$F5,$3E,$FF,$63,$F2,$26,$E3,$F1,$EE,$1A,$7E,$4A,$31,$97,$DA,$9B
	.dc.b	$97,$D6,$9B,$5E,$BF,$AE,$F9,$84,$7E,$53,$F3,$FE,$37,$F4,$27,$E8
	.dc.b	$9B,$1F,$A0,$F8,$FF,$A8,$47,$E9,$37,$1F,$F2,$56,$A8,$C4,$7E,$66
	.dc.b	$31,$3F,$67,$E4,$3F,$5E,$7E,$C2,$AC,$7F,$96,$3E,$B9,$F4,$DF,$E8
	.dc.b	$D6,$C7,$FB,$51,$B8,$7F,$A4,$DA,$FF,$A3,$9A,$FF,$89,$FD,$74,$62
	.dc.b	$11,$8A,$EC,$7E,$44,$FE,$06,$7F,$A9,$27,$BE,$BF,$A2,$8C,$DF,$F0
	.dc.b	$47,$46,$2D,$3F,$E2,$6A,$62,$BF,$E7,$D5,$E5,$9C,$C6,$23,$CB,$18
	.dc.b	$7F,$DE,$38,$7A,$AB,$94,$62,$75,$5F,$06,$72,$D3,$A5,$71,$6C,$F5
	.dc.b	$F8,$7F,$F7,$3D,$BB,$7F,$D6,$5E,$E7,$69,$CE,$1C,$69,$1C,$2B,$78
	.dc.b	$7F,$95,$63,$9C,$FF,$5E,$F5,$DE,$4E,$7E,$2F,$FC,$61,$E5,$FF,$8F
	.dc.b	$A3,$14,$F7,$DF,$3D,$FD,$DA,$3B,$DF,$0C,$39,$C6,$0F,$F3,$26,$1C
	.dc.b	$DC,$1E,$1C,$3A,$E1,$D3,$CF,$F6,$61,$CE,$7C,$35,$B3,$FC,$88,$74
	.dc.b	$C3,$BB,$8C,$3B,$36,$32,$FB,$3E,$BB,$FD,$40,$F2,$3F,$BF,$B8,$79
	.dc.b	$35,$E1,$F0,$7B,$76,$F9,$29,$3D,$34,$F9,$B1,$89,$F6,$FC,$2F,$FB
	.dc.b	$0E,$F8,$5E,$F2,$EB,$1F,$12,$F8,$3F,$27,$45,$F1,$8F,$99,$C1,$FD
	.dc.b	$DB,$0F,$F3,$36,$5F,$7F,$9D,$EC,$E3,$3F,$E5,$0D,$C3,$F5,$BE,$67
	.dc.b	$A1,$FB,$B8,$F3,$7F,$27,$19,$7F,$3A,$DE,$2F,$F3,$A4,$EE,$73,$97
	.dc.b	$F9,$DE,$73,$5B,$9F,$DF,$35,$F0,$CE,$DF,$E7,$98,$7F,$9F,$F9,$FF
	.dc.b	$AD,$B6,$F3,$3F,$EE,$8B,$7C,$66,$63,$F4,$99,$FE,$73,$0E,$31,$FE
	.dc.b	$04,$C3,$59,$7F,$87,$2E,$E1,$CA,$BF,$05,$39,$6C,$65,$FC,$7C,$65
	.dc.b	$B1,$97,$DF,$3F,$85,$9D,$46,$0F,$EC,$9E,$57,$9F,$99,$97,$FA,$2E
	.dc.b	$7F,$C3,$33,$98,$7C,$55,$3F,$21,$5C,$1B,$1A,$AF,$EE,$11,$82,$38
	.dc.b	$21,$FD,$0B,$8F,$B1,$1F,$F4,$EF,$D6,$3E,$BB,$CC,$D8,$FD,$36,$08
	.dc.b	$C4,$7E,$CA,$31,$3E,$69,$CC,$FF,$89,$5F,$F0,$83,$DB,$C9,$EA,$76
	.dc.b	$FA,$EB,$19,$21,$FC,$DB,$87,$FC,$E9,$49,$DC,$E4,$FE,$49,$1D,$9C
	.dc.b	$61,$E7,$BE,$D1,$EE,$AF,$31,$1F,$6D,$75,$6F,$C4,$FC,$22,$3F,$1B
	.dc.b	$E7,$FE,$29,$EF,$23,$C9,$F8,$5F,$3D,$F2,$BE,$1E,$4F,$CE,$8F,$9A
	.dc.b	$61,$F3,$DF,$FB,$D1,$D1,$0C,$C7,$07,$EB,$FB,$3F,$58,$FA,$04,$7D
	.dc.b	$6F,$D3,$FA,$3F,$B6,$48,$FF,$6E,$F1,$BD,$07,$83,$99,$F1,$D5,$E9
	.dc.b	$FE,$B7,$7B,$AF,$D5,$FD,$53,$CC,$A4,$C3,$AF,$D7,$3E,$5F,$11,$F5
	.dc.b	$F0,$5F,$18,$F1,$B5,$D5,$71,$7F,$DE,$3D,$4F,$AE,$FD,$53,$63,$31
	.dc.b	$FC,$0D,$7D,$77,$F9,$59,$FF,$7D,$98,$F5,$7F,$9D,$34,$FE,$79,$1B
	.dc.b	$7F,$A2,$35,$FE,$9E,$D4,$47,$0D,$8E,$B1,$8E,$8F,$D4,$7B,$7F,$AC
	.dc.b	$25,$7D,$2F,$7B,$FD,$FC,$FE,$CA,$32,$FE,$11,$FE,$33,$1A,$8F,$95
	.dc.b	$C7,$78,$7F,$C1,$F6,$FA,$B7,$D5,$7F,$0F,$6D,$FE,$02,$F8,$76,$FF
	.dc.b	$AD,$7E,$13,$57,$2B,$51,$F3,$BE,$52,$47,$D2,$7E,$77,$D7,$74,$8F
	.dc.b	$D9,$07,$0F,$F7,$66,$BF,$20,$D5,$7D,$28,$CC,$7C,$C6,$BA,$3F,$B8
	.dc.b	$38,$37,$97,$F1,$3F,$95,$FE,$2F,$F2,$B2,$BF,$8D,$C7,$F1,$A3,$0F
	.dc.b	$EB,$A3,$0F,$E4,$3D,$3F,$CD,$91,$FC,$9F,$AD,$5F,$CC,$73,$A7,$1C
	.dc.b	$1F,$A4,$8F,$F3,$1C,$FE,$12,$3F,$C8,$3F,$14,$FA,$EF,$D6,$CF,$EC
	.dc.b	$3B,$BF,$DD,$AD,$F3,$3F,$AD,$56,$FB,$FF,$35,$F5,$DF,$DF,$EB,$FC
	.dc.b	$3E,$D7,$E5,$35,$FC,$9B,$53,$49,$F0,$5F,$A0,$78,$70,$7E,$1F,$97
	.dc.b	$F0,$DF,$17,$2F,$D5,$1F,$9E,$9D,$46,$0F,$F2,$97,$47,$F9,$79,$FE
	.dc.b	$5E,$AB,$94,$62,$BF,$B0,$7D,$77,$E1,$F7,$FC,$BB,$63,$B3,$67,$FC
	.dc.b	$BC,$FF,$71,$9D,$F3,$72,$24,$F8,$F1,$56,$EF,$FD,$DF,$F5,$1F,$C1
	.dc.b	$89,$5F,$C0,$BE,$BB,$B2,$7F,$1E,$4C,$3F,$82,$B6,$3F,$C8,$0D,$7F
	.dc.b	$C9,$0F,$E6,$3C,$37,$13,$E8,$7F,$BF,$7B,$BF,$80,$BF,$EA,$6D,$D7
	.dc.b	$F9,$5D,$6E,$7E,$67,$CA,$7F,$92,$BD,$3F,$F2,$73,$63,$FC,$AC,$D7
	.dc.b	$FC,$C2,$D7,$FC,$67,$F3,$7F,$E2,$1D,$BF,$D5,$8D,$8F,$F5,$5B,$7D
	.dc.b	$C7,$D7,$3E,$17,$FC,$29,$F5,$CF,$16,$93,$0F,$C0,$5B,$EB,$B8,$46
	.dc.b	$1A,$FF,$9D,$23,$57,$FE,$67,$6C,$7F,$9B,$DA,$EE,$78,$BF,$C1,$63
	.dc.b	$2D,$8C,$BF,$CC,$23,$2F,$E5,$7E,$7F,$E5,$3F,$E0,$06,$EF,$FC,$91
	.dc.b	$BA,$7F,$B3,$FE,$B9,$EB,$7F,$91,$1B,$7F,$E4,$26,$BF,$E4,$06,$BF
	.dc.b	$E6,$06,$BC,$E3,$FC,$81,$18,$8F,$F6,$CB,$79,$0E,$7A,$58,$3E,$5B
	.dc.b	$FC,$88,$DB,$FC,$E3,$5F,$F0,$DF,$D4,$FF,$65,$CF,$F9,$FB,$FD,$9E
	.dc.b	$3F,$96,$5A,$31,$1F,$B9,$6B,$FE,$C9,$6C,$6F,$D4,$FF,$26,$C6,$08
	.dc.b	$FF,$94,$21,$EB,$7F,$96,$5B,$5F,$B1,$6B,$FD,$F7,$EB,$9F,$29,$FE
	.dc.b	$70,$6C,$7F,$9B,$9A,$FF,$9B,$1A,$FF,$A3,$1A,$FF,$6E,$D5,$18,$84
	.dc.b	$62,$7F,$CE,$13,$98,$C4,$FF,$96,$29,$E5,$FA,$49,$1F,$C2,$65,$DF
	.dc.b	$FE,$75,$6C,$7F,$98,$1A,$FF,$97,$DA,$FD,$06,$BF,$C9,$DA,$FF,$BD
	.dc.b	$DB,$5E,$89,$FF,$3D,$C6,$3E,$5B,$F6,$D2,$69,$FB,$98,$7D,$73,$E5
	.dc.b	$BF,$D3,$6D,$C3,$FD,$32,$D7,$FD,$2E,$D7,$FD,$6E,$D7,$E6,$35,$F9
	.dc.b	$6D,$9E,$F3,$FE,$9B,$8D,$5F,$97,$E0,$BF,$D3,$58,$7F,$B9,$F8,$3F
	.dc.b	$C2,$2D,$E5,$FE,$91,$F5,$DF,$99,$9F,$F0,$AB,$F9,$BC,$FF,$2E,$D7
	.dc.b	$7C,$9F,$E9,$E8,$C4,$31,$73,$FF,$3D,$8C,$36,$33,$F0,$BF,$EC,$38
	.dc.b	$FF,$2D,$FA,$FF,$EA,$F7,$44,$8D,$79,$89,$FF,$5A,$7C,$C7,$DE,$CF
	.dc.b	$FC,$56,$DD,$3F,$B1,$37,$2F,$22,$FA,$EE,$3C,$72,$7F,$AA,$4E,$F4
	.dc.b	$3B,$8F,$F3,$8F,$D5,$BE,$B7,$FD,$AA,$D8,$CC,$7F,$B6,$3E,$B7,$FD
	.dc.b	$8F,$18,$21,$FE,$6B,$9E,$0F,$F9,$B6,$36,$DF,$8A,$FF,$6A,$46,$23
	.dc.b	$FD,$F2,$D7,$F5,$CD,$7F,$CB,$BF,$5C,$F9,$66,$DF,$E6,$0E,$FE,$F2
	.dc.b	$F9,$6F,$F1,$E4,$61,$B9,$FF,$8F,$E7,$95,$71,$8C,$5F,$4C,$3F,$8C
	.dc.b	$46,$EF,$8C,$78,$9A,$AF,$9B,$C0,$F3,$D7,$37,$FC,$9C,$6D,$C3,$E0
	.dc.b	$BC,$BC,$1A,$F5,$F1,$7E,$BB,$AA,$FE,$CD,$AE,$9B,$19,$9D,$BC,$BE
	.dc.b	$47,$FD,$1A,$9A,$47,$FD,$22,$FC,$37,$FC,$C9,$F2,$AE,$3C,$47,$F9
	.dc.b	$9E,$6A,$9D,$3F,$04,$D1,$4D,$4F,$07,$C4,$66,$38,$EB,$F8,$FE,$E6
	.dc.b	$EF,$B9,$FF,$1E,$5F,$E1,$1A,$FF,$30,$8D,$61,$F2,$AB,$8D,$69,$7D
	.dc.b	$8B,$E3,$BC,$FB,$FF,$B2,$A3,$17,$FE,$D2,$C3,$9E,$3F,$EC,$59,$CB
	.dc.b	$6F,$8D,$7F,$43,$9C,$DF,$FB,$52,$F8,$BD,$6B,$77,$FC,$27,$CF,$FF
	.dc.b	$3C,$61,$FE,$47,$6C,$7F,$B4,$FF,$61,$7C,$87,$0E,$85,$6B,$1F,$F5
	.dc.b	$6B,$70,$FF,$32,$37,$0F,$EA,$78,$72,$9E,$DF,$03,$CF,$DE,$7C,$ED
	.dc.b	$3C,$BB,$BF,$62,$FF,$18,$5F,$F9,$6E,$B6,$DA,$E0,$FF,$BE,$27,$78
	.dc.b	$FE,$D2,$3F,$CE,$D5,$C2,$3F,$DD,$DE,$8F,$FB,$F2,$7E,$B3,$63,$F4
	.dc.b	$B8,$7F,$8E,$FC,$1F,$B1,$39,$9F,$F0,$5F,$37,$FB,$BE,$FF,$CD,$CD
	.dc.b	$8E,$B5,$F1,$E3,$1B,$7E,$53,$0F,$E5,$9E,$CF,$E0,$FC,$2F,$F3,$26
	.dc.b	$3F,$87,$6E,$3E,$26,$BF,$EC,$5A,$E3,$3B,$AE,$44,$7F,$73,$C3,$FA
	.dc.b	$3D,$BE,$C7,$47,$C1,$CB,$C5,$97,$67,$FD,$CB,$19,$BF,$F0,$67,$E7
	.dc.b	$DF,$C7,$5F,$CD,$37,$3F,$8A,$D7,$C8,$D7,$8F,$6F,$E7,$E3,$11,$FE
	.dc.b	$14,$6E,$87,$39,$BD,$87,$C4,$FE,$3D,$B8,$FF,$AB,$6F,$86,$39,$AF
	.dc.b	$C3,$FE,$77,$FC,$BE,$BF,$60,$DC,$BE,$DB,$5F,$94,$D7,$FC,$54,$D7
	.dc.b	$FC,$52,$DB,$34,$D6,$3F,$E4,$BE,$5F,$D2,$76,$FE,$7E,$B7,$E0,$E2
	.dc.b	$7F,$37,$E7,$FE,$EB,$1F,$B4,$D7,$FC,$74,$D7,$FC,$72,$DA,$FF,$1A
	.dc.b	$F7,$FF,$09,$8C,$11,$FE,$23,$8B,$E6,$BE,$56,$9F,$EC,$16,$E1,$FA
	.dc.b	$33,$AD,$7E,$EE,$33,$C5,$F9,$5F,$C7,$BF,$BC,$5F,$F1,$96,$DF,$E9
	.dc.b	$1A,$FE,$4D,$AF,$F9,$4D,$AF,$F9,$49,$B9,$1A,$6B,$17,$DC,$BE,$BB
	.dc.b	$F8,$73,$70,$FC,$05,$78,$B1,$FE,$7B,$7C,$2F,$F8,$DE,$3F,$E8,$7F
	.dc.b	$57,$F9,$9B,$70,$FE,$B3,$E8,$7F,$9C,$1B,$3A,$8C,$4F,$F6,$06,$BF
	.dc.b	$D9,$9A,$88,$E1,$FF,$37,$FE,$64,$F9,$4F,$E3,$2D,$BE,$19,$7F,$54
	.dc.b	$AF,$21,$F4,$F5,$56,$E3,$EB,$78,$6F,$9E,$7C,$E7,$F3,$88,$C3,$FC
	.dc.b	$9B,$E0,$9F,$2B,$9A,$DE,$7F,$A7,$F8,$47,$DA,$7F,$51,$6C,$7F,$55
	.dc.b	$6E,$5F,$C8,$1A,$FF,$0F,$CD,$F3,$DF,$D0,$7E,$C5,$F4,$FE,$E8,$C3
	.dc.b	$FD,$33,$C3,$7E,$06,$F9,$6F,$EC,$B1,$86,$C6,$73,$FF,$07,$CE,$5B
	.dc.b	$19,$CB,$FC,$2F,$39,$6C,$65,$FF,$02,$B5,$FA,$DD,$0F,$8E,$FF,$04
	.dc.b	$5F,$D2,$32,$E2,$7F,$49,$7D,$79,$7F,$1B,$39,$7F,$C1,$71,$EC,$F1
	.dc.b	$FF,$B1,$7D,$97,$FB,$57,$0E,$98,$7F,$84,$72,$FF,$34,$46,$32,$56
	.dc.b	$EB,$BD,$69,$FD,$A4,$EB,$A7,$E5,$4E,$72,$FC,$FD,$F1,$7C,$71,$FD
	.dc.b	$79,$FD,$1F,$D9,$7F,$AD,$B1,$F1,$E7,$F4,$BE,$2F,$FB,$C4,$E7,$3F
	.dc.b	$0C,$63,$DD,$EE,$DB,$E3,$8B,$E8,$BC,$67,$EB,$B7,$FE,$F9,$97,$DE
	.dc.b	$8C,$D7,$07,$5E,$2E,$32,$F7,$C7,$F8,$27,$E8,$9F,$2B,$9D,$BF,$77
	.dc.b	$18,$C3,$E5,$FA,$8F,$B1,$7D,$87,$F8,$13,$EC,$7F,$BB,$61,$FE,$3E
	.dc.b	$AD,$4E,$7D,$6F,$F0,$EC,$63,$0F,$E9,$17,$C1,$FD,$D4,$67,$4F,$D9
	.dc.b	$4E,$5F,$F5,$66,$4F,$85,$F4,$1F,$53,$DC,$B8,$7D,$D9,$D3,$67,$4D
	.dc.b	$3E,$5F,$D3,$7E,$80,$8F,$E7,$9B,$7F,$9C,$6B,$F9,$C7,$F9,$ED,$3E
	.dc.b	$5F,$E5,$1F,$86,$8B,$5F,$80,$FC,$59,$F1,$9F,$3F,$5F,$DA,$46,$27
	.dc.b	$D1,$18,$9D,$47,$16,$E3,$FC,$4B,$E1,$3F,$7F,$3F,$DC,$7C,$37,$C8
	.dc.b	$FF,$13,$9C,$D7,$F1,$B9,$CD,$7E,$93,$CB,$FF,$4A,$78,$7F,$C9,$30
	.dc.b	$E0,$D9,$F4,$79,$BF,$C1,$B4,$FE,$45,$8F,$FA,$01,$AF,$F3,$1E,$A4
	.dc.b	$62,$3F,$D0,$0D,$8C,$CF,$32,$57,$F8,$93,$81,$F5,$1F,$93,$3E,$2C
	.dc.b	$D9,$F4,$7E,$11,$FC,$FD,$F8,$1F,$5D,$FC,$C6,$31,$97,$BE,$30,$FF
	.dc.b	$40,$6C,$F0,$3F,$0D,$F3,$BF,$A1,$FA,$CF,$D1,$36,$32,$7E,$2E,$32
	.dc.b	$D3,$FC,$41,$DD,$FB,$9B,$FF,$21,$35,$FE,$A9,$E9,$7E,$DF,$C4,$FF
	.dc.b	$47,$4E,$65,$2F,$CC,$7F,$15,$7F,$E2,$B6,$CF,$F5,$3A,$FE,$0D,$DF
	.dc.b	$FE,$57,$9C,$B7,$AB,$F7,$5F,$3F,$FE,$76,$9C,$E1,$FC,$C3,$85,$39
	.dc.b	$AD,$CF,$76,$E3,$FD,$87,$F3,$1F,$F3,$63,$EA,$6B,$E8,$6E,$67,$3E
	.dc.b	$5B,$1F,$E6,$D8,$78,$4F,$F4,$4C,$FE,$EE,$75,$1F,$E8,$CF,$C5,$DF
	.dc.b	$23,$FE,$B3,$6C,$7F,$AD,$DB,$8F,$FA,$09,$AF,$F8,$04,$53,$CB,$FF
	.dc.b	$0B,$47,$F3,$A6,$ED,$FE,$7F,$AE,$6F,$F3,$6F,$67,$F9,$99,$30,$FF
	.dc.b	$2E,$4E,$9F,$EB,$35,$C9,$B3,$BB,$FF,$04,$56,$DF,$1D,$6D,$FB,$FF
	.dc.b	$6B,$FD,$E2,$FF,$DA,$4D,$8F,$F4,$EB,$5F,$F4,$63,$5F,$0F,$C4,$7F
	.dc.b	$48,$3F,$D6,$71,$C7,$1F,$E9,$F1,$FE,$B6,$F6,$BF,$C7,$A4,$F9,$4F
	.dc.b	$F6,$F3,$70,$FF,$71,$37,$0F,$F5,$6B,$FE,$41,$C5,$F7,$7F,$D5,$8E
	.dc.b	$97,$6F,$FD,$01,$F4,$AE,$9F,$F7,$53,$63,$FD,$DC,$DC,$7F,$D7,$2D
	.dc.b	$7F,$D5,$0D,$7F,$4E,$D7,$F4,$AD,$E3,$39,$E0,$FB,$CF,$F6,$A3,$FD
	.dc.b	$AA,$77,$D7,$FB,$8B,$9B,$FF,$00,$67,$C9,$F4,$E3,$9C,$7F,$CE,$5E
	.dc.b	$2B,$E3,$5F,$3A,$BF,$F2,$77,$C7,$7F,$8C,$2F,$CB,$DB,$95,$19,$C3
	.dc.b	$95,$E6,$38,$6D,$EA,$BC,$7B,$5F,$E5,$9B,$DE,$3D,$A3,$E1,$CE,$B1
	.dc.b	$FE,$93,$97,$92,$B5,$1E,$26,$E1,$D5,$BA,$AF,$96,$3D,$AF,$CF,$8F
	.dc.b	$6B,$E7,$0F,$B1,$E1,$E2,$61,$B1,$C0,$7C,$CB,$E9,$3F,$DE,$DB,$7C
	.dc.b	$59,$E5,$9B,$0E,$4D,$8E,$46,$A3,$D7,$7C,$FC,$DF,$81,$39,$37,$8F
	.dc.b	$89,$FE,$00,$64,$AE,$5F,$E0,$8D,$7F,$D0,$CF,$F4,$ED,$38,$C6,$3C
	.dc.b	$7F,$EE,$AE,$27,$4D,$8C,$BD,$9D,$E5,$EF,$9D,$B6,$33,$AF,$FA,$57
	.dc.b	$1F,$F6,$17,$D3,$7F,$17,$AD,$63,$F3,$6F,$78,$F8,$E3,$E1,$FC,$42
	.dc.b	$F8,$63,$FD,$33,$F2,$67,$BD,$8C,$3F,$A5,$C6,$1F,$C1,$B5,$FB,$18
	.dc.b	$3F,$62,$7E,$0F,$C6,$30,$F3,$E5,$5F,$1D,$7F,$32,$DC,$BF,$31,$87
	.dc.b	$2C,$7F,$27,$8F,$27,$C3,$97,$3B,$E1,$1F,$E2,$2A,$E0,$FE,$2F,$AF
	.dc.b	$F9,$C7,$C4,$FF,$3B,$EC,$FF,$34,$B7,$9F,$F9,$8B,$6B,$DE,$DA,$E3
	.dc.b	$1F,$BE,$8C,$76,$1E,$1B,$7B,$1B,$96,$6B,$F0,$73,$96,$C6,$5F,$C7
	.dc.b	$C6,$5B,$19,$7F,$A9,$46,$5B,$19,$7E,$C4,$FE,$41,$FE,$8B,$9F,$98
	.dc.b	$8F,$FA,$62,$9F,$AF,$3E,$8F,$D2,$D9,$F3,$B5,$F2,$B5,$F1,$B5,$FF
	.dc.b	$09,$B5,$FF,$08,$B6,$7B,$CF,$95,$B3,$FA,$6C,$BF,$B6,$7D,$D7,$CC
	.dc.b	$24,$79,$FF,$31,$7D,$B7,$CE,$6C,$7C,$C6,$BF,$29,$AF,$C8,$6B,$FD
	.dc.b	$BB,$DF,$FB,$24,$7E,$27,$E6,$3F,$CD,$1F,$E7,$71,$97,$FB,$F3,$F7
	.dc.b	$A3,$51,$83,$64,$3F,$DF,$CE,$B8,$7F,$9D,$64,$8E,$33,$F7,$1A,$FF
	.dc.b	$8F,$1A,$FF,$8E,$9B,$97,$F8,$47,$1F,$C8,$92,$7E,$E6,$6F,$8E,$FC
	.dc.b	$9B,$63,$F2,$2D,$7F,$1E,$D7,$F1,$AD,$7F,$C9,$BE,$93,$CC,$FC,$FF
	.dc.b	$C6,$C9,$FF,$2A,$C7,$E8,$27,$51,$FD,$8A,$74,$FF,$93,$67,$44,$78
	.dc.b	$9F,$93,$FC,$C5,$F3,$9F,$AA,$6C,$7E,$A1,$AF,$E9,$9A,$FE,$91,$AF
	.dc.b	$F9,$93,$EC,$5E,$C7,$ED,$BE,$56,$4F,$E8,$67,$2D,$9C,$BF,$E3,$B9
	.dc.b	$FD,$DC,$EA,$1C,$3F,$DC,$FD,$F7,$E2,$2F,$BE,$FE,$03,$C2,$F9,$3A
	.dc.b	$7F,$0B,$AD,$CD,$AE,$18,$7E,$CE,$F8,$36,$F5,$5F,$C3,$BC,$39,$3F
	.dc.b	$E0,$E7,$F8,$B2,$FB,$8F,$8B,$97,$F8,$1E,$31,$BF,$F0,$36,$9F,$C1
	.dc.b	$67,$E3,$A3,$27,$F5,$87,$2E,$0F,$EC,$6E,$0E,$33,$FE,$1B,$F1,$7F
	.dc.b	$94,$79,$0F,$5F,$FC,$09,$7D,$6B,$FC,$35,$39,$F3,$7F,$97,$47,$19
	.dc.b	$D7,$89,$BF,$43,$FA,$A4,$60,$FF,$01,$E1,$D1,$FA,$73,$FD,$86,$31
	.dc.b	$5F,$E2,$73,$A4,$E8,$F7,$7D,$B3,$E4,$FF,$9C,$FE,$15,$27,$B0,$9D
	.dc.b	$38,$6B,$FE,$7F,$8E,$0F,$53,$84,$FF,$44,$FB,$A9,$1F,$E4,$47,$EB
	.dc.b	$FE,$17,$FA,$68,$E0,$D7,$FC,$1E,$D7,$FD,$20,$D7,$EB,$35,$34,$9D
	.dc.b	$7F,$7D,$E1,$E0,$7F,$A5,$FE,$EB,$FA,$B9,$2B,$DF,$2C,$71,$1F,$E9
	.dc.b	$B9,$B9,$FE,$2E,$B5,$3F,$4F,$E4,$1F,$72,$F5,$7F,$B0,$F4,$FF,$9A
	.dc.b	$AB,$FA,$73,$5F,$A5,$5C,$27,$FB,$73,$63,$7E,$C7,$23,$37,$D6,$3E
	.dc.b	$AF,$C9,$BE,$EB,$F6,$BF,$39,$27,$31,$FA,$49,$D4,$60,$87,$D7,$FD
	.dc.b	$6F,$FD,$86,$E3,$5F,$D4,$4F,$08,$FB,$8D,$8F,$D6,$4F,$09,$FC,$07
	.dc.b	$DD,$BC,$8F,$F0,$37,$CE,$7D,$DB,$FF,$3C,$71,$A3,$1D,$3F,$EC,$BB
	.dc.b	$E1,$87,$E4,$1A,$FE,$5E,$B8,$72,$F4,$3F,$7F,$3A,$C3,$F7,$CF,$E8
	.dc.b	$2F,$F7,$54,$F9,$AF,$91,$87,$36,$DF,$0E,$CE,$67,$F8,$76,$33,$7C
	.dc.b	$9B,$5F,$C1,$70,$EC,$FF,$B8,$BE,$3F,$E1,$61,$EE,$6D,$F8,$9B,$3F
	.dc.b	$C3,$7F,$63,$FE,$EB,$5F,$C5,$5F,$EE,$71,$9F,$17,$F8,$B4,$62,$0D
	.dc.b	$24,$68,$91,$FE,$2C,$5F,$2D,$FE,$37,$F8,$27,$C0,$53,$F9,$AC,$7E
	.dc.b	$5B,$B3,$FC,$25,$D6,$69,$30,$FF,$3C,$F9,$4F,$D8,$3F,$89,$C7,$F0
	.dc.b	$B8,$C7,$8D,$FC,$F6,$73,$AF,$F7,$EF,$A1,$75,$FF,$44,$F8,$99,$8F
	.dc.b	$F3,$F7,$F9,$CF,$E6,$2F,$B4,$A7,$F3,$77,$13,$FE,$3B,$6B,$FE,$8C
	.dc.b	$6B,$FE,$96,$6B,$FC,$EC,$38,$7F,$9F,$35,$CC,$7F,$91,$FA,$1F,$4A
	.dc.b	$F9,$5F,$F3,$03,$6B,$F9,$94,$EA,$7F,$CB,$FF,$4B,$FE,$A6,$D3,$FC
	.dc.b	$3A,$FF,$4F,$F6,$CD,$25,$FF,$90,$5A,$FF,$99,$1A,$FF,$AC,$2D,$F7
	.dc.b	$4F,$7F,$FC,$E4,$D9,$F1,$B5,$FE,$7D,$6F,$D1,$1F,$5B,$FE,$74,$6C
	.dc.b	$FF,$9C,$9A,$FF,$9C,$1A,$FF,$9B,$9A,$FF,$65,$9E,$93,$FE,$59,$9C
	.dc.b	$CA,$72,$F5,$9F,$99,$F8,$7B,$E8,$7F,$D1,$4D,$9F,$96,$D7,$D3,$F5
	.dc.b	$AE,$FF,$F4,$5B,$63,$FD,$12,$D7,$FD,$0E,$D7,$FD,$0A,$D7,$D0,$D7
	.dc.b	$CC,$DC,$0D,$35,$A0,$7D,$4F,$FA,$85,$B5,$F8,$26,$BF,$43,$37,$D0
	.dc.b	$7C,$F3,$E3,$12,$3E,$83,$5F,$A0,$D7,$E8,$3F,$93,$FC,$3D,$F4,$BF
	.dc.b	$EB,$16,$CF,$E4,$9A,$FE,$03,$EF,$1F,$37,$FE,$B2,$6C,$7F,$AC,$1A
	.dc.b	$FF,$AB,$9A,$FF,$AB,$1A,$FD,$F6,$BF,$75,$BA,$1A,$6B,$27,$8C,$EA
	.dc.b	$FF,$64,$36,$FF,$52,$D7,$F2,$FF,$85,$BE,$6B,$FD,$92,$D8,$FF,$63
	.dc.b	$B5,$FF,$62,$B5,$FF,$61,$B5,$FE,$83,$EE,$BE,$83,$FD,$B0,$D8,$FD
	.dc.b	$CB,$5F,$D3,$FE,$01,$FC,$D3,$0E,$35,$C0,$93,$FE,$C7,$9D,$3F,$85
	.dc.b	$FB,$BF,$E7,$B9,$9A,$7F,$40,$D5,$F8,$B7,$F1,$8E,$DF,$EC,$73,$AF
	.dc.b	$A1,$FF,$76,$3B,$9D,$67,$FA,$FF,$BB,$7D,$07,$FB,$B1,$B9,$FF,$B6
	.dc.b	$5A,$FF,$00,$FB,$BF,$E2,$CF,$F0,$28,$CE,$5F,$EB,$06,$DF,$5B,$FF
	.dc.b	$7B,$DF,$0A,$FF,$5A,$BC,$AF,$FD,$DE,$D7,$F8,$13,$5F,$E0,$0D,$5F
	.dc.b	$59,$FC,$42,$7F,$CD,$9F,$6B,$FC,$E2,$1F,$6B,$FC,$D6,$30,$DF,$1F
	.dc.b	$F9,$C4,$F1,$B5,$71,$F0,$BF,$D2,$31,$8A,$DD,$F1,$8F,$F3,$D5,$F6
	.dc.b	$7C,$33,$D2,$B8,$DF,$3C,$3F,$A3,$56,$5A,$F2,$9E,$EE,$AF,$C8,$DF
	.dc.b	$BC,$FC,$03,$C2,$11,$88,$D6,$FD,$1F,$8B,$19,$7B,$FD,$0F,$FA,$55
	.dc.b	$34,$8F,$FA,$65,$F5,$BF,$D7,$7E,$93,$F5,$EF,$C3,$3D,$71,$A7,$FD
	.dc.b	$39,$1A,$CB,$FB,$2C,$E6,$53,$9C,$FC,$71,$CC,$E3,$BF,$83,$D0,$7E
	.dc.b	$C8,$FA,$8F,$F5,$75,$7F,$AA,$67,$31,$88,$46,$2B,$FC,$17,$5C,$5F
	.dc.b	$9F,$CC,$E2,$7C,$A7,$FA,$66,$7F,$1E,$DC,$39,$DF,$17,$F4,$11,$9B
	.dc.b	$E4,$DB,$E7,$7F,$EA,$9C,$3C,$57,$C2,$3F,$D5,$96,$FB,$C7,$DE,$F3
	.dc.b	$6C,$E9,$FE,$66,$D8,$D1,$1D,$78,$96,$FF,$B9,$FD,$EB,$EE,$3C,$CD
	.dc.b	$AF,$0A,$FA,$BC,$CC,$9C,$BC,$23,$15,$FE,$AF,$FC,$2B,$F9,$75,$7E
	.dc.b	$35,$B1,$F9,$A6,$BF,$9E,$6B,$FD,$EF,$E5,$3F,$D3,$13,$F0,$9B,$1D
	.dc.b	$7C,$6F,$99,$18,$EB,$FF,$66,$7C,$A7,$2A,$47,$FB,$7E,$BF,$00,$DB
	.dc.b	$FD,$5B,$67,$FC,$4D,$5C,$63,$11,$FE,$37,$6B,$FE,$E7,$8E,$DB,$FF
	.dc.b	$B5,$F9,$3E,$F7,$C9,$FE,$06,$FC,$78,$E2,$7F,$0D,$83,$F6,$5F,$F6
	.dc.b	$F6,$5F,$B1,$6C,$7E,$F9,$AF,$F0,$26,$BF,$E4,$96,$A3,$A6,$1F,$9D
	.dc.b	$6C,$F5,$8F,$D0,$62,$FB,$FF,$F1,$8C,$FE,$39,$B5,$FE,$E7,$BF,$1E
	.dc.b	$5F,$BA,$BE,$1D,$FF,$9F,$87,$EC,$CF,$A2,$F9,$AD,$CB,$E6,$B7,$1F
	.dc.b	$F2,$93,$5F,$98,$D7,$D5,$5D,$2B,$99,$2B,$FC,$5B,$E1,$7E,$AF,$D3
	.dc.b	$FE,$6F,$1C,$0E,$38,$77,$BF,$44,$EB,$0F,$C1,$D7,$F9,$C7,$F4,$17
	.dc.b	$C5,$FC,$76,$C7,$F2,$36,$CF,$F3,$06,$BF,$CD,$1A,$F3,$9E,$44,$9F
	.dc.b	$E1,$3F,$7F,$FB,$E9,$FF,$30,$B7,$0F,$E4,$B0,$BE,$F1,$F3,$E7,$55
	.dc.b	$FC,$67,$DA,$FD,$ED,$FF,$44,$ED,$FF,$38,$F5,$FF,$1B,$6D,$7F,$A9
	.dc.b	$2B,$75,$FE,$8F,$9C,$BC,$63,$2F,$5A,$E5,$18,$8F,$F3,$47,$B9,$FE
	.dc.b	$21,$BF,$E2,$6D,$8F,$F5,$74,$F9,$31,$FF,$46,$FA,$5F,$9E,$F8,$F7
	.dc.b	$D2,$7E,$4D,$B8,$FF,$4F,$6E,$1F,$D7,$1A,$FF,$60,$6B,$FE,$A7,$6B
	.dc.b	$F9,$F6,$BF,$EF,$D9,$FD,$07,$D4,$BE,$AB,$F5,$ED,$BF,$C3,$57,$1A
	.dc.b	$CD,$7F,$35,$FC,$E1,$DF,$FC,$DF,$E1,$BF,$15,$87,$FA,$36,$72,$FF
	.dc.b	$34,$8C,$E1,$FD,$2B,$E1,$BF,$88,$70,$FF,$B3,$3B,$FE,$A4,$60,$FE
	.dc.b	$0F,$F7,$6F,$B1,$FE,$88,$DC,$7F,$9F,$FC,$63,$EA,$73,$5F,$C2,$3E
	.dc.b	$13,$F9,$0E,$7F,$E0,$46,$BF,$EC,$B6,$BF,$C0,$DA,$FF,$A5,$FE,$77
	.dc.b	$F8,$24,$60,$8F,$F6,$3F,$45,$F7,$FF,$E8,$CC,$BE,$3C,$63,$E1,$BC
	.dc.b	$55,$FD,$7E,$31,$3F,$33,$0F,$F5,$26,$79,$C3,$F8,$4C,$F9,$63,$1E
	.dc.b	$3F,$AD,$B1,$C7,$24,$62,$FF,$6E,$7A,$FE,$43,$A7,$9B,$F0,$DB,$7C
	.dc.b	$5B,$3C,$0F,$F5,$B3,$8E,$BF,$D7,$FC,$B3,$EF,$FF,$93,$36,$FF,$C4
	.dc.b	$4D,$BF,$F1,$7B,$5F,$F1,$9B,$5F,$F7,$D3,$5F,$E6,$4D,$4E,$75,$F9
	.dc.b	$BD,$9F,$14,$FE,$87,$1D,$F0,$FF,$13,$CE,$AF,$FC,$09,$F8,$83,$97
	.dc.b	$F7,$CD,$8F,$F2,$1F,$3A,$31,$97,$F0,$56,$BF,$C0,$5A,$FF,$88,$DA
	.dc.b	$FF,$88,$9B,$1C,$EB,$E2,$F6,$7D,$BF,$07,$CE,$DC,$3F,$01,$9F,$EE
	.dc.b	$B3,$F4,$F1,$FF,$03,$9F,$8B,$5B,$CF,$EA,$7E,$55,$FC,$25,$FF,$2E
	.dc.b	$B5,$FE,$B4,$1C,$3F,$D7,$5A,$E7,$C0,$FE,$BD,$18,$23,$F9,$BB,$70
	.dc.b	$FE,$4C,$FE,$E3,$3F,$C6,$71,$F3,$C6,$3A,$7F,$39,$F7,$CF,$B6,$C6
	.dc.b	$9F,$A2,$8C,$3F,$9F,$C3,$F6,$CE,$67,$79,$FE,$63,$89,$7C,$23,$F3
	.dc.b	$79,$3F,$2A,$FF,$1C,$C7,$F9,$FA,$31,$EB,$7F,$85,$A7,$33,$FE,$11
	.dc.b	$6B,$FD,$99,$AF,$F8,$79,$B3,$CB,$1F,$F0,$EC,$6A,$5F,$3C,$F9,$9F
	.dc.b	$F0,$33,$6B,$FD,$04,$D9,$FF,$47,$B5,$FF,$49,$B5,$FA,$ED,$7F,$C2
	.dc.b	$4D,$7B,$CF,$F0,$DC,$5D,$7F,$38,$FF,$11,$D7,$7C,$7F,$43,$3A,$C3
	.dc.b	$E8,$77,$F1,$23,$FE,$A2,$D3,$E6,$C6,$32,$FF,$53,$C6,$27,$55,$FC
	.dc.b	$B6,$72,$D7,$F1,$F0,$FD,$17,$FB,$B4,$61,$AF,$F8,$B6,$35,$EE,$7F
	.dc.b	$A2,$5A,$7F,$AD,$0B,$1F,$EB,$86,$BC,$E3,$84,$F4,$8F,$F4,$8F,$0B
	.dc.b	$C8,$FE,$F4,$DB,$FD,$EB,$67,$FC,$00,$D7,$F4,$FE,$9B,$B5,$F2,$BF
	.dc.b	$E8,$F6,$DF,$FA,$36,$F9,$6D,$FE,$3B,$9C,$E1,$FC,$BB,$07,$DF,$BE
	.dc.b	$33,$FD,$30,$DC,$7F,$D2,$77,$C3,$0F,$F2,$24,$E7,$0E,$4D,$8C,$DF
	.dc.b	$8F,$2E,$6F,$FA,$9F,$BD,$F9,$9B,$F0,$2F,$F0,$D3,$73,$FE,$3A,$DC
	.dc.b	$BF,$C4,$0D,$7E,$0F,$DF,$FF,$A0,$67,$FC,$C5,$AA,$31,$3F,$EB,$EA
	.dc.b	$E5,$9F,$F9,$6E,$73,$8F,$FA,$87,$EC,$7F,$C5,$5E,$31,$A4,$CB,$FD
	.dc.b	$87,$EF,$3E,$3B,$FD,$78,$D8,$FF,$5D,$57,$17,$FD,$73,$19,$CF,$E7
	.dc.b	$37,$2F,$F3,$63,$67,$FD,$AD,$5C,$6B,$77,$D8,$93,$DB,$0B,$3F,$D9
	.dc.b	$BE,$45,$F2,$1F,$DE,$63,$F9,$58,$8C,$F9,$DF,$EE,$16,$DE,$A7,$55
	.dc.b	$C5,$FC,$6F,$CA,$7E,$A7,$4F,$F4,$13,$6B,$FD,$21,$7C,$AB,$F1,$53
	.dc.b	$97,$FD,$29,$F8,$EF,$F2,$67,$B3,$5F,$2B,$5F,$E8,$2D,$78,$F3,$FF
	.dc.b	$A6,$E3,$10,$69,$BE,$27,$A5,$FC,$F7,$1F,$C1,$B6,$B9,$C7,$FB,$B2
	.dc.b	$31,$0F,$9F,$FF,$13,$93,$D2,$FF,$55,$36,$BE,$34,$EE,$72,$48,$F9
	.dc.b	$56,$FA,$A7,$C9,$7F,$C0,$16,$CF,$F8,$9F,$0F,$CA,$D6,$EB,$44,$9D
	.dc.b	$5D,$C3,$FC,$47,$E4,$3E,$1F,$FD,$E9,$5F,$C5,$67,$FD,$85,$39,$F8
	.dc.b	$84,$E6,$7F,$AE,$57,$07,$FD,$23,$5B,$9D,$36,$33,$3B,$9F,$EE,$B3
	.dc.b	$96,$C6,$67,$C2,$F5,$F8,$EF,$F2,$F4,$32,$F8,$9F,$03,$E4,$84,$EB
	.dc.b	$0F,$03,$C2,$B7,$87,$FB,$76,$39,$D3,$EF,$FF,$D7,$3E,$3E,$7C,$8F
	.dc.b	$F0,$4C,$62,$7A,$C7,$E5,$BE,$8D,$B7,$87,$0E,$77,$B6,$E8,$C7,$A6
	.dc.b	$3F,$EE,$DA,$D6,$3C,$F0,$E4,$F8,$2F,$85,$78,$0E,$0E,$1F,$D9,$D7
	.dc.b	$95,$B1,$97,$95,$FE,$3A,$7F,$BB,$69,$7C,$CF,$F0,$A5,$7E,$DE,$72
	.dc.b	$D8,$CB,$D9,$DE,$5E,$E6,$E3,$F1,$9B,$3F,$C1,$7E,$FF,$F4,$4F,$C0
	.dc.b	$3E,$74,$69,$B8,$67,$0F,$0E,$1F,$7B,$A1,$87,$2B,$F8,$56,$FB,$F7
	.dc.b	$DA,$FF,$8F,$B0,$FB,$DF,$7D,$FD,$86,$FE,$7F,$12,$31,$1C,$AF,$8B
	.dc.b	$F5,$30,$E9,$87,$F7,$A8,$FD,$0C,$E9,$FC,$EB,$78,$BF,$32,$D8,$DC
	.dc.b	$F1,$7F,$C8,$11,$96,$C6,$6B,$BC,$7E,$AF,$0F,$37,$C1,$FF,$96,$3B
	.dc.b	$7F,$39,$18,$9F,$F6,$7B,$5F,$F4,$A7,$E7,$7F,$E7,$B7,$39,$FF,$B5
	.dc.b	$DB,$7C,$F3,$FE,$B7,$39,$9E,$3E,$E6,$50,$BE,$8F,$F3,$B6,$BF,$D0
	.dc.b	$9A,$FF,$9B,$9A,$F2,$6B,$CE,$3F,$8F,$B4,$FE,$43,$1C,$30,$7D,$FF
	.dc.b	$FB,$FA,$7D,$ED,$7F,$C1,$8D,$CB,$F8,$04,$3F,$3C,$7D,$AF,$AE,$30
	.dc.b	$F8,$23,$31,$FC,$C1,$AF,$F5,$76,$BF,$D7,$9A,$FF,$A2,$9A,$FC,$66
	.dc.b	$BD,$8F,$E6,$FF,$80,$3E,$B7,$E6,$62,$8C,$4F,$DA,$6B,$FD,$BF,$E9
	.dc.b	$1F,$05,$F4,$F9,$91,$89,$FE,$D9,$18,$9F,$E9,$2D,$8C,$CE,$DF,$EE
	.dc.b	$F6,$FB,$F7,$81,$E4,$6C,$FE,$31,$AF,$FB,$9B,$CE,$7E,$B2,$FA,$AE
	.dc.b	$17,$2B,$83,$F8,$9A,$DC,$E6,$7F,$C8,$57,$D3,$83,$F3,$6D,$E5,$FC
	.dc.b	$7D,$F0,$AF,$C8,$D7,$F3,$B8,$C4,$7F,$84,$1A,$FF,$AF,$1A,$F2,$C7
	.dc.b	$CC,$FF,$68,$F8,$47,$CA,$FD,$26,$C7,$EA,$9A,$FB,$FE,$2D,$E7,$FA
	.dc.b	$DB,$1F,$EE,$46,$BE,$36,$BE,$36,$BF,$E4,$8F,$69,$F6,$5F,$82,$6C
	.dc.b	$7F,$98,$23,$F0,$BF,$80,$3E,$F7,$51,$27,$47,$F0,$09,$D3,$67,$89
	.dc.b	$1F,$93,$B7,$CC,$FC,$3D,$F8,$BF,$D5,$6C,$FF,$BF,$9A,$FC,$D6,$BF
	.dc.b	$35,$AF,$FB,$AB,$AF,$F8,$43,$FE,$42,$6B,$9F,$03,$F7,$B1,$88,$FE
	.dc.b	$16,$D7,$F6,$B1,$A2,$3F,$E2,$28,$7C,$5F,$22,$46,$5A,$FF,$93,$63
	.dc.b	$6E,$48,$7F,$94,$7E,$FF,$E3,$AE,$4F,$E0,$F1,$86,$E7,$FC,$1F,$F2
	.dc.b	$97,$9B,$4F,$E2,$EE,$2F,$FC,$6E,$D7,$FC,$BC,$D7,$FC,$CA,$D7,$F9
	.dc.b	$58,$70,$FF,$2E,$6B,$98,$FE,$4B,$7F,$ED,$C8,$FF,$22,$B5,$F6,$7C
	.dc.b	$46,$88,$E9,$AE,$8F,$3F,$77,$F2,$3F,$C0,$1F,$27,$4F,$81,$12,$7F
	.dc.b	$12,$7F,$42,$9D,$B8,$0E,$1F,$99,$C3,$EE,$CB,$F8,$86,$9F,$CD,$3E
	.dc.b	$FF,$F3,$F5,$A3,$FA,$6B,$88,$D7,$27,$F3,$7D,$7F,$C6,$71,$FC,$F1
	.dc.b	$FF,$1A,$D7,$FA,$41,$AF,$CC,$6B,$FD,$61,$AF,$DA,$F4,$3F,$A7,$C6
	.dc.b	$08,$E3,$93,$FA,$36,$1F,$D0,$AF,$D1,$C3,$E3,$99,$87,$C1,$AD,$67
	.dc.b	$F7,$1B,$1E,$19,$FE,$A1,$F7,$9F,$E1,$78,$FF,$53,$B5,$FC,$1B,$5F
	.dc.b	$ED,$ED,$4D,$27,$AF,$FD,$A1,$C7,$DE,$F6,$24,$FD,$5F,$BD,$FF,$65
	.dc.b	$37,$E0,$7F,$B7,$C6,$23,$FD,$68,$D7,$FD,$76,$D7,$F2,$ED,$7F,$C0
	.dc.b	$6D,$4D,$27,$A1,$FD,$BF,$EF,$5D,$B4,$FE,$E5,$3F,$E2,$9D,$FF,$BB
	.dc.b	$B7,$6F,$D2,$D7,$0C,$3F,$1C,$D7,$F1,$07,$83,$0E,$7D,$7D,$F8,$A4
	.dc.b	$FC,$18,$CB,$CA,$32,$F2,$C3,$FC,$41,$8F,$F8,$23,$EF,$7F,$B8,$B7
	.dc.b	$CC,$FF,$04,$46,$33,$FD,$CF,$EC,$9F,$EB,$0C,$3F,$44,$D7,$FD,$BC
	.dc.b	$D7,$FD,$CA,$D7,$8F,$BB,$FE,$20,$8C,$47,$F8,$B9,$A9,$A6,$BC,$BF
	.dc.b	$F0,$57,$CA,$FF,$82,$B1,$34,$FF,$06,$FC,$D3,$EF,$BF,$B9,$7D,$F7
	.dc.b	$E3,$7C,$C3,$49,$D1,$FE,$D0,$C5,$F3,$CF,$A7,$FF,$0B,$FC,$FF,$F9
	.dc.b	$03,$8B,$FC,$A2,$DE,$1F,$F2,$2D,$F0,$AF,$F1,$77,$DF,$3E,$C2,$93
	.dc.b	$8B,$F8,$44,$3F,$01,$D9,$FE,$42,$71,$1F,$EE,$26,$BF,$CA,$0F,$F2
	.dc.b	$9B,$87,$0D,$7F,$96,$B5,$DC,$75,$3D,$DF,$4B,$CF,$3F,$F3,$5B,$72
	.dc.b	$F5,$B6,$9B,$BE,$F7,$FC,$0B,$7F,$E1,$37,$FC,$0F,$3F,$E1,$16,$BF
	.dc.b	$E1,$A6,$BC,$9B,$3C,$FD,$3F,$F7,$FB,$67,$DC,$6E,$33,$3C,$B5,$7D
	.dc.b	$EF,$D5,$AF,$F4,$0B,$5F,$E5,$91,$AD,$BF,$CC,$1F,$29,$F8,$39,$FE
	.dc.b	$70,$D7,$BB,$5F,$9C,$D7,$FA,$BB,$5F,$F4,$1B,$53,$49,$C9,$F8,$BF
	.dc.b	$AA,$FE,$AD,$A7,$FA,$61,$BB,$7E,$21,$B7,$FD,$CF,$E3,$BF,$8F,$BF
	.dc.b	$D4,$29,$18,$8F,$B0,$D7,$FD,$48,$D7,$FC,$8E,$D7,$A7,$8B,$F3,$1A
	.dc.b	$7C,$E9,$EB,$19,$6C,$7D,$2C,$FE,$7F,$DE,$FF,$9E,$35,$34,$FF,$46
	.dc.b	$75,$3E,$B6,$9F,$E9,$C7,$17,$E1,$6B,$F7,$9A,$FE,$75,$AF,$F8,$19
	.dc.b	$AF,$FA,$C9,$AF,$F0,$08,$FD,$C7,$DE,$BE,$53,$FD,$1A,$D9,$FF,$62
	.dc.b	$B5,$FE,$E1,$93,$F5,$8F,$E8,$4B,$7F,$F3,$7C,$7F,$3F,$8D,$79,$3F
	.dc.b	$16,$30,$47,$04,$AF,$99,$1F,$D1,$F5,$7D,$F7,$F8,$D2,$7F,$2A,$DB
	.dc.b	$FF,$6D,$B7,$0F,$F1,$2F,$BF,$FC,$02,$7F,$D7,$6F,$EE,$EB,$FD,$31
	.dc.b	$D5,$FE,$C8,$8C,$1F,$BF,$C1,$F7,$DF,$CD,$E9,$18,$F0,$FF,$D6,$33
	.dc.b	$98,$DF,$E0,$D8,$99,$5F,$EE,$CA,$D0,$9F,$F5,$CF,$DF,$FF,$1D,$9F
	.dc.b	$F7,$8B,$5F,$D8,$B7,$2F,$F7,$BB,$ED,$8F,$1B,$9C,$FF,$94,$B6,$33
	.dc.b	$E0,$76,$F7,$3F,$D8,$FF,$72,$FA,$2A,$4A,$FF,$1A,$F8,$2F,$BA,$7D
	.dc.b	$3E,$67,$F8,$1D,$F1,$7F,$C7,$7C,$08,$C4,$7F,$AF,$6B,$8B,$D1,$AF
	.dc.b	$FA,$BD,$AF,$FA,$BD,$AF,$FA,$E1,$A8,$7D,$F7,$F4,$BD,$B1,$F7,$9F
	.dc.b	$1E,$3F,$46,$DE,$AF,$F3,$2B,$9F,$0F,$FC,$ED,$68,$C5,$F2,$DF,$FC
	.dc.b	$D7,$87,$26,$E1,$C9,$CE,$1C,$AF,$8E,$1F,$CB,$AF,$C8,$D9,$EE,$FF
	.dc.b	$4F,$8C,$B7,$EF,$5F,$7D,$FC,$2E,$51,$89,$E8,$D7,$84,$FF,$11,$7E
	.dc.b	$27,$CB,$D4,$D2,$3B,$7D,$CF,$FA,$23,$E6,$FF,$6C,$FC,$3C,$BF,$A3
	.dc.b	$CE,$A3,$E8,$4E,$B2,$FC,$A4,$78,$0E,$07,$3E,$5E,$0E,$63,$73,$C6
	.dc.b	$FE,$E7,$DE,$78,$2B,$F0,$0D,$73,$7F,$D6,$38,$FF,$25,$39,$9F,$D8
	.dc.b	$F0,$FE,$52,$BF,$22,$D8,$FC,$84,$FB,$38,$3D,$BC,$5F,$95,$E3,$C6
	.dc.b	$1F,$E0,$EC,$38,$56,$9B,$1E,$2C,$3F,$3F,$5C,$1F,$E0,$3F,$7C,$FB
	.dc.b	$1C,$62,$8C,$5F,$E9,$17,$C2,$E9,$CD,$FF,$6C,$6E,$5E,$66,$D7,$F8
	.dc.b	$89,$E2,$F3,$C3,$BF,$DF,$3E,$A2,$98,$F9,$3B,$5F,$B4,$9D,$36,$74
	.dc.b	$D3,$FD,$61,$F7,$1F,$D9,$1F,$EB,$8D,$CB,$FC,$12,$D7,$FC,$50,$D7
	.dc.b	$FC,$6F,$F4,$8F,$A3,$F8,$6D,$AE,$BE,$76,$27,$E5,$6A,$FA,$AE,$58
	.dc.b	$7F,$BA,$5B,$19,$AF,$C0,$36,$FF,$B7,$4E,$89,$3A,$79,$F0,$7F,$B4
	.dc.b	$B8,$FE,$AF,$C6,$9F,$0C,$FB,$70,$FC,$2C,$E7,$0F,$F6,$97,$DE,$FF
	.dc.b	$B6,$9F,$EF,$CD,$CB,$FC,$56,$D7,$FC,$94,$D7,$FD,$E6,$D7,$F8,$C3
	.dc.b	$53,$4D,$62,$FB,$EF,$F3,$35,$7F,$2C,$6D,$7F,$A5,$AB,$C5,$87,$E7
	.dc.b	$27,$2F,$E7,$FE,$F9,$EC,$7C,$86,$E1,$FE,$15,$6E,$1F,$E5,$76,$BF
	.dc.b	$CB,$45,$23,$31,$FC,$C6,$31,$5D,$49,$5F,$E1,$DF,$57,$FC,$30,$FF
	.dc.b	$44,$6D,$7F,$81,$AF,$C3,$7F,$E3,$9F,$0F,$F5,$92,$FB,$E7,$C2,$FE
	.dc.b	$FE,$30,$D8,$CE,$3F,$E3,$16,$CF,$F9,$79,$AF,$F9,$F5,$AF,$B5,$AF
	.dc.b	$F4,$B6,$EA,$FB,$EF,$E0,$33,$FD,$F9,$AF,$8A,$BF,$AF,$C6,$BB,$9F
	.dc.b	$7D,$FB,$6A,$FF,$46,$46,$3B,$7F,$8C,$36,$FF,$8E,$56,$EB,$E8,$4E
	.dc.b	$5E,$31,$97,$FB,$04,$65,$DC,$65,$FF,$09,$FB,$AF,$2F,$F8,$9B,$6F
	.dc.b	$FB,$DD,$79,$76,$FF,$4A,$D7,$D0,$F8,$B7,$CD,$FC,$F6,$DF,$F9,$75
	.dc.b	$B8,$FF,$A2,$9A,$FF,$AA,$5A,$FE,$25,$AF,$F7,$76,$BF,$EF,$89,$FB
	.dc.b	$9F,$6A,$F8,$BF,$C0,$B6,$FF,$1C,$73,$9C,$CF,$F2,$EF,$84,$77,$FF
	.dc.b	$30,$F8,$4F,$C5,$DF,$FB,$22,$72,$FF,$85,$23,$38,$FF,$AC,$3E,$97
	.dc.b	$F8,$7F,$01,$A4,$EF,$FE,$9D,$C8,$F9,$C7,$CA,$7F,$3E,$6E,$5F,$E2
	.dc.b	$0F,$8A,$7D,$7E,$6B,$F3,$7F,$33,$FC,$87,$4F,$F5,$8B,$5F,$F6,$83
	.dc.b	$5F,$D8,$35,$FF,$0F,$35,$34,$9E,$2F,$F6,$3F,$3D,$F7,$6F,$9F,$FE
	.dc.b	$AD,$18,$7C,$73,$89,$D6,$7F,$3B,$0E,$F9,$E7,$C6,$F4,$B6,$75,$3E
	.dc.b	$28,$C3,$FC,$1A,$3D,$31,$8E,$AF,$EF,$99,$BE,$17,$C1,$E5,$FC,$46
	.dc.b	$DF,$49,$D3,$E3,$D7,$D7,$5B,$8C,$67,$FD,$6B,$2F,$01,$FE,$43,$FB
	.dc.b	$C7,$DF,$7F,$07,$6D,$7F,$AD,$5B,$3F,$EE,$26,$BF,$F0,$05,$AF,$F2
	.dc.b	$16,$BD,$67,$A1,$F3,$E9,$DB,$F2,$CF,$E3,$35,$DB,$4F,$F1,$5E,$5F
	.dc.b	$52,$7F,$C1,$1C,$EF,$BE,$FC,$86,$7F,$E1,$D6,$E7,$FC,$21,$AF,$F0
	.dc.b	$46,$BF,$C0,$9A,$A7,$A7,$47,$F0,$E9,$CC,$62,$BE,$17,$6F,$E0,$7C
	.dc.b	$2F,$3B,$63,$EE,$E9,$FB,$ED,$3D,$3A,$FF,$07,$9F,$89,$5B,$CF,$E8
	.dc.b	$8F,$BE,$FE,$1C,$FF,$49,$3F,$CE,$CE,$1C,$35,$FE,$9C,$D7,$7E,$0F
	.dc.b	$F3,$F8,$C4,$75,$8C,$1F,$47,$8F,$F3,$47,$F4,$89,$F0,$E3,$E8,$8C
	.dc.b	$74,$FE,$8B,$EF,$1F,$5D,$8C,$FF,$1D,$18,$7F,$3B,$87,$EE,$9C,$CE
	.dc.b	$F2,$FC,$7F,$1A,$F8,$47,$FB,$83,$EF,$BF,$C7,$6F,$F6,$9C,$D1,$88
	.dc.b	$FE,$D0,$D7,$FB,$6B,$63,$7E,$AF,$F8,$7A,$30,$47,$FB,$17,$DF,$3E
	.dc.b	$D2,$93,$3F,$A1,$87,$FA,$96,$75,$18,$84,$61,$DB,$5F,$B0,$D7,$FC
	.dc.b	$04,$D5,$97,$C1,$EA,$F9,$87,$F7,$89,$E9,$97,$E8,$33,$FE,$2F,$87
	.dc.b	$CE,$EF,$FE,$2D,$1F,$CA,$31,$FC,$36,$3F,$2E,$72,$4A,$F5,$C6,$5B
	.dc.b	$5C,$6B,$FD,$30,$D7,$FD,$89,$0F,$BE,$FF,$41,$47,$F8,$6E,$31,$ED
	.dc.b	$7F,$A2,$5A,$7C,$82,$C7,$E6,$5A,$F1,$9E,$6F,$66,$A8,$DB,$FD,$DB
	.dc.b	$85,$E4,$FF,$80,$1B,$3F,$E9,$B6,$CF,$F8,$19,$AF,$F8,$C7,$C4,$FE
	.dc.b	$1B,$F0,$9F,$E8,$CE,$EF,$F7,$B5,$FF,$BD,$2F,$8E,$DF,$E9,$C9,$CE
	.dc.b	$1F,$C9,$30,$7D,$F3,$EF,$BF,$D2,$AD,$C7,$FD,$EF,$7C,$30,$FF,$52
	.dc.b	$CE,$70,$E4,$D8,$CD,$F8,$F2,$E6,$FF,$1C,$FB,$D7,$DF,$7F,$87,$9B
	.dc.b	$97,$FA,$E5,$B8,$FF,$89,$1A,$FB,$FE,$63,$FB,$36,$3F,$EB,$66,$D7
	.dc.b	$F8,$82,$B8,$CE,$7C,$8F,$E5,$7E,$33,$EF,$BF,$C5,$AD,$BF,$F6,$43
	.dc.b	$71,$FF,$19,$B5,$F9,$DE,$E3,$E4,$FF,$D7,$2D,$AF,$EC,$9E,$37,$FA
	.dc.b	$B7,$0C,$D4,$9F,$ED,$1C,$5F,$E9,$98,$F1,$CE,$70,$FA,$58,$73,$6C
	.dc.b	$EA,$F8,$CE,$73,$FF,$78,$FD,$93,$EF,$BF,$BA,$E1,$FC,$AE,$31,$E7
	.dc.b	$7F,$B8,$5B,$87,$F0,$06,$CF,$F3,$1C,$DF,$2F,$FA,$FC,$7F,$DC,$2D
	.dc.b	$B5,$2F,$8D,$6B,$B7,$FC,$E7,$F7,$DF,$C8,$EB,$F8,$6B,$5F,$E7,$CD
	.dc.b	$7F,$D1,$0D,$7A,$B5,$34,$9D,$3F,$E2,$EF,$4B,$F9,$CE,$3E,$4C,$D1
	.dc.b	$8B,$E7,$5F,$EE,$E8,$C4,$3E,$7F,$FC,$4C,$4E,$4F,$F4,$43,$70,$FE
	.dc.b	$44,$D9,$FE,$4E,$FF,$8B,$F0,$7D,$F3,$E8,$FF,$DF,$6D,$BF,$F1,$0D
	.dc.b	$7E,$52,$D7,$C2,$F6,$4A,$DE,$16,$BF,$C2,$FE,$43,$E1,$9F,$11,$FC
	.dc.b	$26,$61,$FD,$5A,$33,$7F,$C7,$23,$EB,$D7,$1A,$E1,$3F,$EB,$78,$C3
	.dc.b	$B9,$E5,$3A,$78,$D7,$47,$BD,$70,$6C,$66,$BA,$3B,$F8,$DF,$F3,$0C
	.dc.b	$FC,$0D,$3F,$D5,$F1,$8B,$E4,$EB,$3F,$97,$F0,$3E,$D8,$E3,$6F,$BD
	.dc.b	$FF,$5E,$F9,$1F,$92,$9C,$C6,$25,$1E,$0E,$57,$DF,$7F,$9A,$E3,$78
	.dc.b	$7F,$54,$AD,$F2,$FF,$B6,$70,$E5,$5A,$C7,$B4,$FE,$83,$1E,$7C,$1E
	.dc.b	$58,$E1,$8F,$3B,$E1,$8F,$F6,$CC,$7C,$8F,$F8,$75,$AF,$0C,$7F,$B7
	.dc.b	$47,$F7,$5D,$2F,$99,$FE,$10,$AE,$4D,$79,$35,$F4,$3B,$BF,$86,$DC
	.dc.b	$73,$5E,$AA,$FD,$C7,$DF,$7E,$85,$C1,$F3,$6B,$83,$FB,$EC,$3C,$8F
	.dc.b	$C6,$31,$1A,$D1,$5C,$E7,$56,$FB,$E7,$D5,$7F,$08,$9E,$5C,$0F,$BE
	.dc.b	$FE,$BF,$1E,$58,$C7,$97,$BD,$BF,$0D,$87,$2A,$F5,$63,$D7,$1F,$EE
	.dc.b	$D1,$F9,$E9,$D3,$F9,$66,$F2,$FE,$7D,$B1,$B9,$E2,$FF,$8F,$A3,$2D
	.dc.b	$8C,$D7,$78,$FD,$36,$3E,$9F,$83,$FF,$1C,$61,$FC,$47,$3F,$F6,$A5
	.dc.b	$7F,$3E,$6B,$FE,$60,$FB,$EF,$F6,$43,$9A,$E2,$FF,$04,$9D,$F7,$7E
	.dc.b	$F7,$0E,$D3,$FE,$AA,$BE,$2F,$3F,$77,$FC,$6C,$A3,$2F,$F9,$AD,$AF
	.dc.b	$D1,$6B,$FC,$E9,$AF,$18,$E7,$18,$37,$2F,$BE,$FF,$7F,$D7,$F5,$A6
	.dc.b	$BF,$E0,$96,$DF,$F1,$4F,$BE,$7D,$9F,$BA,$91,$87,$FC,$9C,$D7,$FC
	.dc.b	$A6,$D7,$FD,$08,$D7,$F1,$4D,$7F,$A9,$B5,$EB,$1F,$12,$35,$6F,$BE
	.dc.b	$7D,$0F,$D0,$8C,$36,$33,$7F,$ED,$DE,$57,$DF,$3E,$0B,$EB,$C6,$1F
	.dc.b	$F3,$33,$67,$FB,$64,$ED,$B3,$FE,$6B,$C1,$F4,$7F,$CE,$A7,$FC,$06
	.dc.b	$D7,$FC,$64,$DD,$3F,$9F,$79,$CF,$BD,$7D,$F7,$A6,$B5,$CB,$FD,$85
	.dc.b	$B8,$7F,$4E,$6D,$F1,$AE,$1E,$2C,$AF,$0B,$F9,$29,$D4,$7F,$86,$B0
	.dc.b	$EE,$FF,$A1,$A3,$2F,$FA,$DD,$AF,$EF,$5A,$FF,$73,$F8,$37,$CA,$FC
	.dc.b	$F6,$C7,$F8,$89,$AF,$FB,$D2,$5F,$16,$F3,$FC,$8D,$9F,$13,$5F,$13
	.dc.b	$5F,$13,$5F,$F0,$27,$B6,$FB,$8F,$C0,$36,$3F,$CB,$71,$F8,$3F,$BE
	.dc.b	$7D,$EE,$A2,$4E,$8F,$F1,$CC,$E9,$B3,$C4,$8F,$F9,$0B,$EF,$9F,$75
	.dc.b	$F3,$5B,$1F,$2D,$AF,$CB,$6B,$F2,$DA,$FF,$29,$EB,$FE,$16,$FF,$B9
	.dc.b	$DA,$E7,$C0,$FE,$07,$18,$8F,$F1,$BB,$5F,$D9,$C6,$88,$FF,$AF,$E1
	.dc.b	$F1,$5C,$89,$19,$6B,$FE,$C8,$8D,$B9,$21,$FE,$CC,$FB,$EF,$8C,$FF
	.dc.b	$C0,$06,$F3,$FF,$86,$1B,$CF,$FD,$17,$EF,$5E,$77,$F0,$76,$DF,$FB
	.dc.b	$4D,$AF,$FB,$61,$AF,$F4,$D0,$E1,$E9,$1F,$C9,$A3,$11,$FE,$5A,$6B
	.dc.b	$C2,$3F,$95,$D3,$E2,$29,$27,$FD,$C9,$19,$6C,$65,$F8,$0D,$7C,$95
	.dc.b	$FC,$0D,$61,$FE,$38,$FB,$E7,$C9,$FF,$25,$6C,$FF,$9B,$A7,$FC,$7B
	.dc.b	$5C,$5B,$DD,$BB,$95,$BC,$FE,$54,$FF,$2F,$FB,$EF,$CF,$4F,$82,$7F
	.dc.b	$B9,$46,$2F,$FC,$92,$D8,$FE,$ED,$A7,$F8,$C2,$7F,$C9,$AF,$F8,$CA
	.dc.b	$BE,$73,$5D,$FA,$3F,$D4,$A3,$11,$FE,$86,$6B,$F7,$23,$44,$7E,$C1
	.dc.b	$BE,$2F,$46,$F2,$3D,$86,$4A,$7E,$1A,$34,$FB,$4F,$F4,$AE,$3F,$D2
	.dc.b	$F0,$F4,$CF,$F5,$1F,$B5,$FF,$0B,$57,$E1,$DA,$FF,$85,$5A,$FF,$A7
	.dc.b	$9A,$9A,$4F,$67,$FB,$8B,$8F,$BE,$EC,$49,$F3,$FD,$EF,$FB,$1D,$BF
	.dc.b	$05,$FD,$E2,$31,$1F,$09,$AF,$E6,$DA,$FF,$8A,$5A,$FF,$AC,$9A,$9A
	.dc.b	$4F,$43,$FB,$87,$DF,$3B,$69,$FD,$CE,$7F,$C4,$FA,$7F,$A2,$9B,$C3
	.dc.b	$FE,$37,$AE,$18,$7D,$06,$BF,$3C,$F0,$61,$CF,$A7,$BF,$14,$9F,$83
	.dc.b	$19,$79,$46,$5E,$58,$7F,$88,$30,$FF,$23,$7D,$F7,$F7,$29,$EC,$DC
	.dc.b	$BF,$D3,$ED,$9F,$F2,$67,$DF,$7F,$AC,$6B,$F0,$AD,$7F,$1A,$D7,$F8
	.dc.b	$33,$5E,$3E,$FF,$F8,$AA,$31,$1F,$ED,$E6,$A6,$9A,$F2,$BF,$C1,$7F
	.dc.b	$2B,$FE,$0B,$F3,$CD,$3F,$C1,$DF,$34,$FB,$BF,$ED,$DF,$7D,$F8,$EF
	.dc.b	$24,$D2,$74,$7F,$81,$FE,$79,$F6,$5F,$E1,$7F,$9F,$FF,$20,$F1,$7F
	.dc.b	$91,$5B,$C3,$FE,$E1,$AD,$F0,$7E,$9B,$EF,$9F,$5F,$FE,$31,$6C,$66
	.dc.b	$31,$F7,$12,$3F,$66,$25,$1F,$E4,$E6,$C6,$9A,$FE,$E9,$AF,$F3,$36
	.dc.b	$BF,$E8,$16,$A3,$B4,$7F,$89,$7E,$97,$F9,$D6,$5F,$E6,$16,$E3,$FE
	.dc.b	$FD,$6D,$36,$7D,$F7,$F8,$2A,$FF,$C4,$EF,$F8,$32,$7F,$C2,$8D,$7C
	.dc.b	$13,$C3,$D3,$F7,$B6,$3F,$C3,$D3,$B7,$B3,$63,$51,$FE,$21,$8D,$72
	.dc.b	$BE,$FB,$E8,$CF,$F1,$38,$C7,$07,$F9,$3E,$73,$8F,$FB,$E3,$E5,$BF
	.dc.b	$09,$7F,$C7,$1A,$FF,$24,$6B,$FD,$75,$AF,$FA,$AD,$AF,$CB,$6A,$69
	.dc.b	$39,$3F,$CA,$5F,$59,$FD,$66,$7C,$CD,$AF,$F4,$83,$76,$FE,$A5,$F1
	.dc.b	$FF,$C8,$6B,$F9,$9D,$23,$11,$F6,$1A,$FF,$AF,$E3,$5E,$2F,$E3,$1B
	.dc.b	$1F,$D0,$1A,$F1,$8D,$4F,$28,$F0,$C7,$F9,$A7,$27,$DF,$7C,$B2,$4F
	.dc.b	$F4,$4E,$A7,$D6,$7F,$A3,$DB,$1F,$9C,$6B,$FD,$41,$AF,$F8,$25,$AF
	.dc.b	$FB,$5D,$AF,$E5,$DA,$FE,$FA,$3F,$D6,$31,$AD,$DF,$7C,$F9,$5F,$F4
	.dc.b	$73,$6F,$FA,$DC,$EA,$7F,$D8,$3A,$3E,$FB,$FA,$1A,$DB,$FC,$05,$1F
	.dc.b	$D0,$23,$5E,$4F,$CB,$8C,$11,$C1,$30,$F9,$D1,$FD,$23,$07,$DF,$7F
	.dc.b	$8E,$27,$F2,$CD,$BF,$F6,$83,$70,$FF,$7C,$7C,$0F,$EF,$A7,$C9,$5F
	.dc.b	$B9,$8F,$DD,$BF,$E9,$CE,$AF,$F6,$9C,$60,$FD,$FE,$0F,$BE,$FE,$71
	.dc.b	$48,$C7,$87,$FE,$B3,$9C,$C6,$FE,$FB,$13,$2B,$E4,$D6,$84,$FF,$AE
	.dc.b	$FE,$FB,$F8,$EC,$FE,$C5,$B9,$7F,$BA,$DB,$9F,$F1,$1C,$51,$EF,$AF
	.dc.b	$23,$A6,$F8,$1D,$4E,$DE,$D7,$FB,$27,$EE,$5F,$47,$49,$5F,$E1,$8F
	.dc.b	$01,$F7,$CF,$A7,$CC,$FF,$05,$BE,$2F,$E1,$1A,$FF,$A8,$E7,$83,$CD
	.dc.b	$AF,$FA,$C5,$AF,$FA,$C5,$AF,$F1,$E6,$BF,$CB,$FE,$FB,$FA,$5E,$DF
	.dc.b	$CC,$7E,$F3,$E3,$C7,$E9,$1B,$D7,$FC,$D6,$91,$AF,$0F,$FC,$B3,$18
	.dc.b	$DB,$57,$D2,$30,$77,$BE,$57,$D2,$3C,$15,$C2,$F9,$CF,$5C,$3F,$A8
	.dc.b	$E1,$CE,$FF,$CF,$8D,$78,$46,$BE,$F9,$F7,$D4,$93,$FE,$4C,$A4,$EA
	.dc.b	$3F,$AE,$8F,$97,$AD,$32,$E0,$8F,$B0,$FE,$B1,$F3,$9F,$A2,$C7,$F1
	.dc.b	$F1,$FD,$6E,$7D,$AF,$C9,$9D,$67,$FE,$5E,$9C,$CF,$2E,$5A,$F1,$3B
	.dc.b	$0A,$E2,$FE,$9F,$F3,$97,$DA,$FE,$CA,$3E,$B7,$3A,$73,$7F,$E6,$C8
	.dc.b	$C5,$6E,$BF,$41,$5A,$7F,$D5,$12,$FC,$4B,$FD,$CD,$7F,$C5,$9B,$3F
	.dc.b	$D6,$EB,$86,$FF,$C0,$AF,$87,$C5,$FF,$04,$8C,$5F,$06,$C6,$70,$F0
	.dc.b	$E3,$71,$FE,$05,$7B,$9F,$F0,$27,$D3,$7F,$7D,$F7,$F4,$D8,$FF,$41
	.dc.b	$BF,$E0,$1E,$4C,$F4,$2A,$92,$B7,$1F,$D1,$1B,$1D,$3F,$05,$7D,$4D
	.dc.b	$31,$E4,$DA,$EF,$3A,$6C,$E9,$A7,$90,$FF,$43,$1F,$E9,$CC,$7F,$9D
	.dc.b	$79,$7F,$47,$44,$62,$74,$70,$8F,$F6,$D4,$7E,$57,$E1,$3E,$0F,$21
	.dc.b	$F8,$75,$BE,$1F,$E6,$75,$B8,$F1,$7D,$1F,$78,$FE,$94,$D8,$34,$8F
	.dc.b	$89,$B8,$78,$A3,$67,$AF,$0F,$E3,$EF,$D2,$6C,$7F,$AC,$6B,$BC,$E7
	.dc.b	$BF,$EC,$F2,$3E,$FF,$C8,$D9,$F8,$CD,$BF,$82,$D7,$F8,$95,$FF,$47
	.dc.b	$AE,$24,$AE,$2F,$BF,$F1,$27,$97,$F8,$86,$CF,$E2,$30,$ED,$A7,$C8
	.dc.b	$9C,$E3,$E9,$97,$C1,$79,$89,$7E,$66,$CF,$99,$AF,$9A,$78,$CF,$E0
	.dc.b	$7F,$1E,$7C,$AF,$E6,$9B,$1E,$3B,$C4,$FE,$7B,$EF,$3F,$CD,$F3,$F8
	.dc.b	$96,$E1,$F8,$46,$BF,$11,$AF,$EA,$1A,$FE,$39,$BA,$9A,$63,$E9,$5F
	.dc.b	$23,$FA,$E6,$DF,$EB,$A7,$C3,$AF,$E3,$A7,$38,$7D,$F9,$79,$3F,$8F
	.dc.b	$6D,$FA,$63,$04,$70,$49,$F6,$47,$E5,$FF,$0E,$7D,$BF,$F0,$06,$D7
	.dc.b	$8E,$71,$3F,$C1,$3F,$17,$FD,$45,$FE,$19,$AF,$F0,$A6,$BF,$B0,$6E
	.dc.b	$86,$93,$A3,$F5,$0E,$BF,$14,$FE,$19,$8F,$F1,$06,$DF,$E7,$E7,$F6
	.dc.b	$53,$FC,$62,$73,$AF,$F8,$9B,$E8,$3F,$C5,$FF,$3D,$FC,$6B,$F2,$2F
	.dc.b	$8B,$8A,$B5,$EB,$AB,$5B,$7B,$7E,$DB,$C6,$4C,$FF,$86,$68,$8F,$DF
	.dc.b	$7E,$2A,$F8,$4F,$E4,$ED,$AF,$47,$CB,$62,$7F,$98,$43,$F0,$97,$DC
	.dc.b	$7F,$15,$8C,$36,$33,$1F,$E0,$3D,$BF,$D7,$0E,$8C,$3F,$E5,$78,$DB
	.dc.b	$B8,$D3,$63,$2F,$F8,$36,$32,$FE,$FB,$F0,$07,$37,$F4,$06,$CF,$FB
	.dc.b	$1F,$83,$FC,$A2,$F4,$7F,$90,$C6,$AB,$F8,$97,$DE,$FE,$C9,$BD,$9F
	.dc.b	$D2,$A1,$F6,$96,$26,$1F,$D3,$27,$86,$1F,$CB,$1D,$C6,$67,$FA,$9F
	.dc.b	$E2,$1F,$6A,$BF,$98,$BF,$EE,$E9,$FE,$AA,$E5,$FE,$B6,$E7,$0F,$E4
	.dc.b	$CE,$5F,$E8,$6E,$67,$9F,$17,$6F,$4F,$FB,$35,$7E,$DB,$E1,$7F,$B4
	.dc.b	$7A,$8F,$C0,$5F,$77,$AF,$09,$5A,$9C,$E9,$F2,$4F,$F1,$E4,$E9,$C6
	.dc.b	$1F,$E0,$F9,$DB,$F9,$58,$DD,$73,$B9,$1E,$EF,$26,$D7,$B1,$F8,$BD
	.dc.b	$DF,$1B,$C8,$B5,$F1,$1B,$5F,$09,$BB,$7C,$A7,$E6,$C4,$9F,$97,$A5
	.dc.b	$8F,$92,$5A,$FA,$31,$65,$2F,$C5,$EF,$DA,$FF,$71,$6C,$FE,$61,$B3
	.dc.b	$F9,$D6,$BF,$E6,$16,$BD,$1A,$F5,$13,$D5,$B3,$FE,$01,$D9,$F8,$97
	.dc.b	$F8,$69,$B1,$F4,$2F,$C3,$8F,$F7,$A9,$CD,$FF,$6B,$FC,$7B,$FA,$3B
	.dc.b	$7E,$2B,$FC,$47,$18,$AF,$F0,$0B,$FE,$CF,$C3,$FC,$4A,$E5,$FF,$15
	.dc.b	$B9,$AE,$9E,$57,$2B,$FF,$19,$79,$9F,$8A,$D3,$AF,$CB,$63,$8B,$D7
	.dc.b	$87,$8E,$31,$3A,$CB,$FD,$B7,$5A,$CF,$C7,$5F,$E3,$DD,$7F,$CE,$BE
	.dc.b	$47,$D1,$F8,$0F,$E6,$71,$8C,$3E,$94,$60,$93,$B2,$47,$F8,$AF,$6F
	.dc.b	$E3,$99,$FC,$5F,$07,$B5,$FF,$32,$8F,$F1,$6C,$EB,$2F,$F2,$B7,$E7
	.dc.b	$5F,$D2,$A7,$E6,$7A,$FF,$02,$31,$3F,$2A,$31,$E6,$FC,$E6,$E5,$F3
	.dc.b	$1F,$A3,$F5,$5F,$4B,$D7,$B3,$F4,$1B,$8F,$D9,$6E,$1F,$59,$FB,$3F
	.dc.b	$83,$FF,$8D,$73,$34,$97,$EF,$FC,$53,$EF,$9C,$1F,$BA,$FE,$0B,$C8
	.dc.b	$91,$83,$F0,$2E,$31,$FC,$53,$CA,$32,$7E,$1B,$E9,$7C,$44,$7E,$73
	.dc.b	$E5,$F8,$4F,$C7,$9F,$90,$6C,$7E,$3B,$E1,$BD,$44,$7E,$B3,$ED,$F8
	.dc.b	$0F,$CB,$9F,$98,$6C,$7E,$5A,$2E,$1F,$9F,$9B,$AF,$E7,$26,$D7,$E6
	.dc.b	$A6,$D7,$DE,$F7,$FE,$51,$27,$F0,$CF,$CD,$F8,$2F,$D3,$9F,$A8,$6C
	.dc.b	$7E,$9B,$ED,$FE,$91,$1F,$C7,$36,$BF,$CE,$7B,$A3,$11,$FE,$54,$8C
	.dc.b	$4F,$E5,$7D,$0F,$DA,$9F,$B6,$AB,$1F,$CB,$BB,$7F,$D6,$DD,$FF,$E8
	.dc.b	$36,$C7,$FA,$DF,$E1,$7F,$7F,$8F,$FA,$F1,$AF,$FA,$E2,$B8,$46,$23
	.dc.b	$FD,$89,$18,$3F,$19,$E0,$FF,$06,$3F,$7B,$3F,$C0,$E7,$F4,$B1,$9A
	.dc.b	$FE,$0E,$EA,$B1,$3F,$D0,$45,$BE,$E3,$FD,$A9,$7F,$ED,$FC,$BF,$97
	.dc.b	$CE,$7C,$44,$E6,$F8,$DF,$F8,$21,$FF,$78,$5F,$08,$E7,$18,$37,$5C
	.dc.b	$9C,$5F,$5B,$E8,$F8,$E3,$5F,$00,$C3,$A5,$E1,$B7,$C3,$0E,$F5,$BB
	.dc.b	$F2,$9B,$3C,$33,$E9,$5F,$78,$FA,$5F,$F7,$A4,$79,$CC,$B8,$AE,$9E
	.dc.b	$E3,$EF,$7F,$D5,$35,$BB,$E5,$9F,$56,$C6,$70,$E5,$9F,$F5,$AC,$39
	.dc.b	$7C,$27,$96,$BB,$E1,$C6,$FC,$71,$FE,$35,$C3,$9B,$63,$2F,$F6,$43
	.dc.b	$89,$FE,$38,$87,$DE,$BE,$F3,$FC,$77,$3F,$02,$30,$ED,$AF,$76,$BD
	.dc.b	$67,$9C,$FB,$67,$8E,$9F,$3E,$BA,$4F,$E1,$72,$7C,$1B,$C9,$F3,$5E
	.dc.b	$F1,$E7,$87,$C9,$C7,$9E,$BD,$DB,$0C,$79,$E1,$CA,$BF,$2B,$F9,$DB
	.dc.b	$ED,$B1,$7F,$9C,$8C,$3F,$8C,$6B,$F9,$05,$E5,$70,$E4,$46,$B4,$FC
	.dc.b	$54,$62,$F5,$39,$C3,$C3,$78,$99,$87,$77,$71,$E4,$F8,$47,$53,$E1
	.dc.b	$FE,$C7,$81,$FD,$F9,$BC,$FF,$DF,$1B,$8F,$AF,$F0,$AB,$87,$29,$B8
	.dc.b	$73,$AF,$F0,$2C,$70,$D7,$D7,$77,$53,$F8,$28,$CD,$7F,$9F,$9A,$FF
	.dc.b	$A0,$DA,$FE,$35,$AF,$D0,$8E,$53,$D2,$39,$35,$FE,$5B,$C9,$E6,$23
	.dc.b	$FE,$54,$C5,$F5,$2F,$9E,$F6,$B6,$7F,$D2,$33,$A9,$FC,$CB,$63,$33
	.dc.b	$B7,$FD,$35,$19,$6C,$65,$FD,$34,$65,$B1,$97,$FC,$9B,$1E,$4F,$63
	.dc.b	$C6,$48,$F4,$4F,$FB,$69,$BE,$6F,$92,$31,$1E,$08,$C1,$9F,$B0,$F6
	.dc.b	$92,$39,$FC,$CD,$7F,$9A,$61,$F1,$27,$1C,$3F,$C9,$E3,$E3,$D6,$38
	.dc.b	$7C,$F8,$AF,$18,$2F,$6F,$63,$E3,$DF,$9C,$8F,$CA,$FA,$FF,$A2,$47
	.dc.b	$DF,$F1,$3F,$50,$F7,$11,$FA,$2F,$C1,$F5,$BE,$C9,$23,$E3,$FD,$55
	.dc.b	$7D,$B9,$7F,$A7,$1F,$97,$3F,$80,$99,$1C,$08,$70,$FB,$1F,$C2,$9F
	.dc.b	$3C,$8F,$E0,$9F,$A5,$D1,$F8,$AE,$49,$3F,$62,$3F,$1A,$61,$F1,$C6
	.dc.b	$1F,$EF,$5E,$AF,$E5,$1F,$BF,$D3,$F9,$43,$F5,$91,$98,$FE,$D5,$CE
	.dc.b	$8C,$71,$7D,$7D,$3F,$86,$D7,$23,$AB,$F9,$EF,$B9,$91,$F8,$C7,$F7
	.dc.b	$EB,$ED,$5E,$76,$1B,$1F,$CE,$9A,$FF,$BD,$9A,$FF,$BE,$DA,$FF,$45
	.dc.b	$6B,$FE,$AA,$6B,$DE,$3F,$A2,$B4,$FE,$09,$BB,$DF,$FD,$AF,$C6,$49
	.dc.b	$FF,$42,$C7,$F0,$F9,$D4,$7E,$9A,$74,$FC,$68,$CB,$F8,$2F,$0F,$F7
	.dc.b	$FE,$44,$8F,$D6,$C7,$F0,$23,$0F,$E2,$23,$0F,$F0,$9F,$31,$F0,$7F
	.dc.b	$C9,$1B,$19,$8F,$97,$F6,$6F,$9D,$A7,$B9,$C4,$FF,$64,$6B,$FD,$AC
	.dc.b	$FF,$67,$B8,$70,$D7,$FB,$83,$5D,$C7,$53,$F8,$26,$FF,$C7,$09,$1F
	.dc.b	$C2,$BC,$BA,$FE,$92,$2F,$0F,$E2,$67,$F9,$1F,$77,$F2,$4F,$8C,$97
	.dc.b	$FC,$4A,$32,$FF,$14,$8C,$BF,$AE,$8F,$C5,$FD,$9B,$B7,$FA,$4F,$C5
	.dc.b	$BE,$6A,$9A,$8F,$C8,$FD,$93,$EC,$F1,$8F,$FB,$C3,$4F,$F1,$C2,$C7
	.dc.b	$A3,$9C,$BF,$C2,$57,$C6,$30,$7F,$BE,$1C,$3F,$E5,$13,$96,$1D,$9C
	.dc.b	$46,$AF,$FC,$8D,$D6,$FC,$55,$C7,$FD,$75,$FF,$08,$4E,$9B,$8F,$F9
	.dc.b	$61,$DE,$1F,$E7,$56,$BF,$B8,$76,$FF,$98,$9B,$7D,$EF,$FB,$7F,$C8
	.dc.b	$FF,$91,$6B,$FA,$C3,$5F,$E6,$ED,$C3,$F9,$B3,$FE,$30,$8D,$F0,$72
	.dc.b	$23,$CB,$9B,$F3,$1D,$32,$3F,$38,$FF,$02,$FC,$25,$F0,$B4,$FD,$AB
	.dc.b	$89,$FF,$1D,$B5,$FC,$C0,$70,$FE,$6D,$AF,$F9,$1D,$A8,$E1,$1D,$8F
	.dc.b	$ED,$FE,$BF,$B7,$6F,$EC,$AD,$9F,$E9,$CD,$F2,$1F,$2B,$F8,$49,$FE
	.dc.b	$B6,$BB,$3F,$BC,$B6,$3F,$BE,$B5,$FE,$F2,$D7,$FC,$B9,$EA,$BF,$1C
	.dc.b	$7B,$DF,$E2,$2D,$7F,$DD,$B5,$FE,$0B,$6B,$FE,$9D,$AE,$13,$FE,$28
	.dc.b	$B7,$95,$49,$5F,$E1,$38,$F5,$1F,$75,$FF,$03,$B5,$FF,$02,$B5,$FE
	.dc.b	$DC,$DC,$BE,$0F,$07,$B8,$93,$EE,$7F,$C3,$3F,$32,$F8,$EF,$F1,$DB
	.dc.b	$67,$FC,$48,$D7,$FC,$07,$F4,$1E,$CE,$5F,$F0,$13,$EE,$8F,$F0,$DB
	.dc.b	$5F,$F0,$CB,$5F,$F0,$13,$70,$34,$9F,$11,$FE,$2C,$75,$6F,$A5,$F2
	.dc.b	$92,$FF,$C7,$12,$F8,$6F,$F2,$9B,$63,$FC,$6E,$D7,$FC,$66,$D7,$CF
	.dc.b	$1D,$C8,$F0,$6B,$FE,$3A,$C3,$FC,$0C,$FF,$89,$1B,$8F,$F8,$8D,$BC
	.dc.b	$9F,$E3,$8E,$AF,$E8,$35,$FE,$2A,$6B,$FE,$40,$6B,$BF,$1B,$FC,$97
	.dc.b	$18,$8F,$F1,$43,$6B,$FD,$CB,$3A,$24,$FF,$95,$23,$FB,$EB,$E5,$F3
	.dc.b	$7F,$CE,$B1,$8C,$3F,$C7,$16,$F9,$1F,$F3,$CB,$6F,$FC,$BA,$D7,$FC
	.dc.b	$B2,$D7,$D1,$1E,$62,$3C,$1A,$FF,$98,$3E,$3F,$F8,$ED,$7F,$91,$DB
	.dc.b	$97,$F9,$71,$B3,$FE,$45,$FC,$6D,$F3,$1F,$E8,$C6,$C7,$F9,$C1,$AF
	.dc.b	$F9,$B1,$AF,$F9,$A9,$AF,$F9,$A1,$AF,$58,$FF,$3B,$C6,$32,$36,$DE
	.dc.b	$CF,$FA,$71,$B7,$FE,$85,$6B,$FE,$64,$97,$E2,$9E,$73,$A9,$23,$D0
	.dc.b	$FF,$A1,$BA,$7F,$99,$47,$F9,$B1,$B9,$FF,$9A,$DB,$7F,$E8,$07,$F8
	.dc.b	$94,$67,$C0,$FF,$57,$C6,$AB,$FC,$D8,$DC,$BF,$DA,$9D,$BF,$EB,$4F
	.dc.b	$91,$FF,$7B,$E1,$FE,$77,$6E,$3F,$E8,$C6,$E1,$FE,$75,$7F,$C4,$73
	.dc.b	$BF,$86,$E4,$49,$FF,$53,$4E,$AB,$F9,$B7,$0B,$E8,$DF,$31,$FE,$BF
	.dc.b	$6D,$FF,$70,$9D,$4F,$FB,$0F,$E2,$FF,$AE,$D6,$AB,$FC,$B2,$DC,$BF
	.dc.b	$B7,$B7,$0E,$F5,$FD,$90,$FF,$7F,$57,$FB,$0D,$AE,$BE,$17,$FD,$99
	.dc.b	$18,$8F,$E9,$9F,$49,$FE,$5A,$C3,$FA,$77,$57,$ED,$A1,$F1,$FF,$D4
	.dc.b	$6B,$F8,$AF,$55,$9C,$C7,$08,$C7,$B0,$C3,$13,$DA,$30,$F2,$3C,$37
	.dc.b	$D7,$0F,$03,$6F,$9D,$F2,$A6,$1E,$77,$FB,$5B,$5F,$F2,$23,$5E,$26
	.dc.b	$BE,$41,$C1,$FD,$CA,$39,$C6,$1D,$46,$65,$39,$9E,$1F,$13,$FE,$72
	.dc.b	$4D,$23,$FE,$76,$7D,$0B,$89,$BE,$23,$89,$F0,$CF,$EF,$D1,$A7,$07
	.dc.b	$F8,$06,$79,$C6,$74,$FE,$4F,$18,$AF,$9F,$C4,$DD,$F2,$1F,$E4,$4C
	.dc.b	$7F,$0E,$D5,$3B,$8C,$36,$BF,$C2,$EF,$E6,$FD,$4F,$26,$DF,$92,$6C
	.dc.b	$F4,$3F,$39,$C7,$FE,$59,$6C,$6B,$0E,$F3,$97,$FD,$6B,$8F,$9A,$B6
	.dc.b	$7F,$95,$39,$9F,$49,$FE,$0C,$8C,$36,$BF,$DE,$6F,$E5,$A3,$3E,$4F
	.dc.b	$7A,$E2,$D9,$E1,$97,$FB,$DD,$E1,$7C,$A7,$87,$D2,$3E,$56,$98,$F3
	.dc.b	$EE,$5E,$59,$D3,$67,$4D,$3F,$93,$78,$9F,$3F,$3F,$C0,$B6,$0D,$35
	.dc.b	$C2,$FA,$4F,$53,$67,$99,$F1,$5C,$67,$FA,$F8,$FF,$23,$5F,$E3,$BB
	.dc.b	$7B,$12,$3F,$CA,$31,$96,$9F,$E4,$12,$C6,$E7,$FD,$5D,$3B,$6D,$3E
	.dc.b	$6F,$CA,$7C,$FB,$E5,$3E,$31,$5F,$A6,$9F,$A9,$C6,$FA,$4F,$88,$DB
	.dc.b	$FE,$DF,$18,$23,$82,$5F,$F7,$28,$F9,$9F,$1F,$FD,$E6,$7F,$10,$D9
	.dc.b	$FF,$07,$4F,$E0,$E3,$1C,$5F,$E0,$0F,$9F,$F0,$92,$FF,$BF,$B6,$FF
	.dc.b	$BF,$B5,$FF,$0B,$3F,$6F,$E9,$1F,$2F,$F8,$F6,$D7,$E6,$72,$56,$6B
	.dc.b	$F2,$9E,$73,$EA,$BF,$08,$DB,$FA,$AD,$C7,$E8,$B5,$FE,$65,$B2,$31
	.dc.b	$1F,$80,$6B,$F4,$1B,$1B,$AF,$C9,$7C,$BF,$FA,$DA,$FF,$5C,$DB,$FB
	.dc.b	$15,$B9,$CF,$B7,$E8,$F1,$FC,$E4,$8F,$3B,$6B,$CE,$D7,$CF,$8F,$F6
	.dc.b	$29,$FC,$8F,$C1,$3E,$47,$F6,$CD,$8F,$F3,$6D,$E2,$7F,$75,$EB,$FE
	.dc.b	$9A,$7F,$D2,$7D,$1F,$AB,$8C,$63,$FE,$94,$8C,$12,$76,$48,$FC,$DF
	.dc.b	$B3,$FC,$0A,$3F,$4E,$D8,$FF,$43,$CF,$F0,$98,$C7,$98,$FA,$47,$C8
	.dc.b	$7F,$0D,$FA,$5F,$66,$1F,$E9,$B9,$CE,$3F,$C6,$67,$3C,$FF,$CE,$F8
	.dc.b	$7F,$D6,$3F,$17,$E9,$25,$7F,$AB,$9F,$E0,$DF,$4A,$FA,$0F,$E3,$4D
	.dc.b	$9F,$F5,$57,$A5,$89,$FE,$59,$0F,$A5,$7C,$FF,$F3,$D8,$C3,$63,$31
	.dc.b	$FE,$D2,$32,$57,$3A,$FF,$7F,$3F,$AB,$B8,$7F,$C0,$51,$C1,$D3,$5F
	.dc.b	$CA,$FC,$2B,$9B,$F9,$93,$67,$FD,$B5,$AF,$F5,$D7,$93,$FE,$F4,$9F
	.dc.b	$E7,$1F,$17,$FA,$89,$FE,$84,$DB,$FC,$AB,$FE,$64,$8C,$F6,$F0,$23
	.dc.b	$C3,$DB,$B5,$F9,$CF,$A5,$FA,$73,$FC,$99,$7A,$DF,$D1,$DC,$CF,$F1
	.dc.b	$87,$33,$FD,$4D,$CF,$11,$CC,$F5,$EB,$E5,$7F,$D7,$E1,$DB,$FE,$6A
	.dc.b	$C3,$E0,$9F,$DF,$BE,$95,$F2,$3B,$9D,$57,$F2,$9F,$55,$5A,$BE,$32
	.dc.b	$3A,$FB,$DF,$DB,$6B,$78,$73,$8C,$5F,$1A,$E8,$DE,$1F,$CA,$69,$FC
	.dc.b	$EF,$1E,$B8,$F2,$7E,$07,$93,$F1,$7B,$BE,$37,$C0,$59,$F8,$8D,$CB
	.dc.b	$E7,$37,$1E,$AF,$CD,$89,$3F,$2F,$0B,$1F,$54,$B5,$E4,$8B,$3E,$FF
	.dc.b	$80,$F8,$1F,$49,$FE,$19,$6C,$7E,$5E,$D1,$87,$F3,$4D,$7F,$CB,$4D
	.dc.b	$7A,$09,$E8,$D9,$E9,$3B,$8F,$EB,$7B,$BE,$17,$FC,$1C,$DA,$FA,$B5
	.dc.b	$E1,$C7,$F8,$04,$E6,$FF,$96,$FC,$F7,$F4,$66,$FD,$2F,$F8,$5E,$31
	.dc.b	$5F,$C1,$D4,$67,$CB,$D3,$71,$FF,$13,$4E,$9F,$F4,$67,$1F,$A7,$0E
	.dc.b	$FF,$41,$FE,$31,$F3,$7F,$1F,$B7,$78,$C5,$6F,$3F,$04,$62,$75,$87
	.dc.b	$B3,$E0,$3F,$97,$E3,$F2,$9A,$FD,$56,$BF,$49,$BB,$7F,$9C,$3C,$9F
	.dc.b	$B1,$18,$23,$FE,$25,$D3,$F8,$F5,$7C,$6F,$0B,$FC,$91,$7C,$70,$F7
	.dc.b	$C7,$AF,$E6,$3F,$A4,$57,$CC,$F6,$31,$E6,$FC,$B6,$DF,$C6,$6C,$EE
	.dc.b	$3E,$4F,$D2,$FD,$1F,$63,$E9,$36,$32,$7C,$E8,$CB,$4F,$A3,$F1,$5E
	.dc.b	$92,$3F,$E2,$ED,$11,$ED,$FA,$5F,$BC,$7D,$F6,$CF,$DA,$F2,$7F,$04
	.dc.b	$63,$1F,$C3,$B8,$8D,$1F,$80,$AC,$4F,$E3,$3E,$2B,$E5,$12,$7C,$4F
	.dc.b	$CD,$F8,$4F,$C7,$1F,$8F,$6C,$7E,$2B,$E1,$FE,$91,$1F,$53,$F5,$7E
	.dc.b	$1F,$F2,$C7,$E5,$DB,$1F,$92,$8B,$7F,$9E,$9B,$CF,$F9,$B9,$B5,$FE
	.dc.b	$56,$D1,$58,$BF,$F2,$3C,$7F,$9F,$5F,$F0,$FB,$5E,$1F,$01,$F8,$73
	.dc.b	$F1,$1F,$0B,$FA,$93,$F5,$4D,$BF,$D2,$FD,$2F,$E3,$48,$FD,$56,$D7
	.dc.b	$F9,$7A,$D1,$88,$F0,$BF,$95,$F4,$7F,$6C,$7E,$DD,$B5,$FE,$FB,$F8
	.dc.b	$4F,$DE,$57,$FB,$C7,$C0,$FF,$55,$C6,$23,$FD,$EF,$18,$24,$FF,$AB
	.dc.b	$A7,$53,$FE,$B4,$6F,$98,$C7,$FC,$6D,$B7,$F0,$5F,$0B,$F7,$95,$FC
	.dc.b	$23,$0F,$99,$1F,$C2,$C8,$F9,$71,$B8,$FF,$9D,$7C,$17,$D2,$FF,$B5
	.dc.b	$6B,$FD,$9D,$AF,$F4,$D9,$CF,$8B,$FE,$D3,$F8,$14,$E6,$FF,$A9,$5F
	.dc.b	$0C,$3F,$D9,$71,$CD,$FF,$72,$CE,$EF,$93,$6B,$86,$1E,$2A,$C9,$CA
	.dc.b	$AC,$78,$A7,$77,$D7,$0F,$1D,$F8,$0D,$C6,$70,$FE,$CD,$0F,$A5,$7C
	.dc.b	$7F,$FB,$CE,$3C,$A7,$F6,$B8,$CE,$5E,$BF,$71,$F2,$5F,$E9,$CC,$BB
	.dc.b	$DF,$3C,$39,$46,$30,$E3,$3D,$0D,$61,$CE,$F9,$63,$E4,$AD,$DF,$1C
	.dc.b	$3C,$98,$7F,$8C,$67,$9B,$63,$2F,$16,$BE,$9F,$8D,$7D,$27,$F6,$38
	.dc.b	$F8,$11,$87,$FB,$35,$23,$11,$DD,$AF,$95,$E1,$7E,$D6,$E1,$9B,$F8
	.dc.b	$DF,$04,$F2,$BE,$7D,$EE,$7F,$09,$87,$89,$B8,$F8,$5E,$35,$C3,$0F
	.dc.b	$C8,$7D,$2B,$E8,$71,$83,$89,$39,$8F,$C4,$B5,$FB,$71,$AF,$2B,$8B
	.dc.b	$C2,$AD,$6F,$3F,$C4,$CE,$6F,$9D,$EA,$E4,$76,$78,$C7,$93,$A7,$F6
	.dc.b	$9F,$0F,$F8,$EE,$9F,$EF,$0D,$E6,$FE,$3C,$DC,$3D,$59,$BE,$87,$18
	.dc.b	$72,$9F,$EF,$75,$8E,$B3,$C7,$4F,$83,$77,$0E,$57,$FE,$21,$9C,$B6
	.dc.b	$32,$FF,$9E,$E3,$2D,$8C,$BF,$A6,$8C,$B6,$32,$FD,$49,$FC,$66,$1F
	.dc.b	$C4,$F8,$3C,$C4,$9F,$F2,$AC,$3E,$75,$F4,$BE,$E6,$CF,$F8,$DE,$75
	.dc.b	$3D,$27,$33,$FE,$91,$9C,$BB,$8C,$BF,$EE,$26,$BF,$E6,$38,$DB,$C6
	.dc.b	$32,$A3,$C9,$EC,$F8,$C9,$1E,$89,$FF,$6A,$37,$CC,$F2,$11,$F1,$3E
	.dc.b	$7F,$A5,$F6,$12,$39,$7D,$2D,$7F,$DD,$74,$9C,$70,$FF,$B0,$E1,$58
	.dc.b	$E1,$F5,$46,$EF,$C9,$7F,$26,$A4,$7C,$37,$E4,$11,$F8,$DF,$3D,$F3
	.dc.b	$C8,$FB,$7E,$27,$E9,$1F,$3C,$8F,$CD,$7E,$97,$AD,$F5,$C9,$1F,$17
	.dc.b	$E6,$AB,$EC,$8B,$9F,$AE,$7E,$F1,$27,$81,$F3,$7E,$97,$F0,$47,$DF
	.dc.b	$23,$F7,$5F,$C1,$74,$7E,$1F,$8E,$4F,$D5,$8F,$C5,$18,$7E,$6C,$61
	.dc.b	$FF,$52,$F2,$BD,$4F,$C9,$47,$DD,$E9,$FC,$91,$FA,$B8,$CC,$7F,$63
	.dc.b	$6E,$5F,$56,$77,$3D,$27,$3C,$3F,$5B,$E9,$7F,$38,$49,$FF,$5A,$8F
	.dc.b	$A5,$FF,$0E,$B8,$6B,$FE,$97,$6B,$FE,$EF,$6B,$FD,$61,$AF,$FA,$FD
	.dc.b	$AF,$5F,$37,$F9,$F3,$4F,$E8,$11,$C3,$2F,$F6,$6F,$BD,$FB,$1F,$8B
	.dc.b	$95,$FE,$86,$8F,$E1,$9C,$27,$31,$F9,$89,$CC,$7E,$E7,$C4,$FD,$D7
	.dc.b	$C9,$CA,$FD,$44,$65,$FD,$54,$65,$FD,$0C,$7F,$1E,$F3,$3F,$DD,$F1
	.dc.b	$FC,$19,$AB,$83,$FD,$4F,$F0,$9F,$C9,$E1,$F4,$AF,$A5,$C3,$63,$FA
	.dc.b	$F3,$5F,$F6,$13,$5F,$F0,$09,$CA,$3E,$3C,$66,$3F,$C0,$AD,$7F,$DD
	.dc.b	$8D,$78,$1F,$C4,$7B,$3F,$8D,$11,$FE,$11,$E5,$57,$F7,$F1,$70,$FD
	.dc.b	$F4,$FF,$21,$E3,$FE,$45,$F1,$72,$FF,$88,$46,$5F,$E2,$51,$97,$F8
	.dc.b	$4C,$7C,$3F,$A4,$76,$FF,$48,$F8,$C7,$CD,$D3,$51,$F2,$7E,$91,$F4
	.dc.b	$B8,$C7,$F2,$B8,$72,$7F,$2F,$3A,$6C,$ED,$FA,$98,$FF,$8A,$AB,$84
	.dc.b	$60,$0A,$E4,$FF,$93,$7E,$95,$C5,$FD,$65,$FF,$14,$EC,$C3,$FC,$9D
	.dc.b	$18,$6C,$EC,$E1,$E7,$4B,$FD,$ED,$74,$7F,$A7,$FC,$8F,$F8,$F1,$BC
	.dc.b	$1F,$DB,$23,$11,$E7,$F2,$51,$FE,$21,$8D,$90,$D3,$C3,$2F,$34,$7F
	.dc.b	$49,$8E,$0F,$3C,$7F,$50,$FE,$6F,$A3,$F3,$C7,$E6,$7E,$95,$F0,$B4
	.dc.b	$FD,$BB,$89,$FF,$1B,$35,$FC,$B8,$70,$FE,$69,$AF,$F9,$71,$AF,$D2
	.dc.b	$6B,$C6,$3F,$BB,$FA,$BE,$BD,$BF,$B3,$36,$7F,$B2,$37,$37,$CB,$FD
	.dc.b	$ED,$3E,$0C,$63,$B3,$FC,$03,$26,$92,$FD,$3E,$C3,$E9,$5E,$F7,$F8
	.dc.b	$7B,$6F,$F7,$1D,$7F,$81,$9A,$FE,$2E,$B8,$4F,$F8,$93,$07,$97,$49
	.dc.b	$5F,$E0,$E8,$F4,$1F,$79,$FF,$3F,$E5,$EA,$E2,$F3,$92,$3F,$C2,$13
	.dc.b	$B7,$FC,$10,$D7,$FC,$0C,$D9,$E5,$1F,$E1,$7F,$99,$7C,$77,$F8,$E1
	.dc.b	$B1,$FE,$20,$6B,$F8,$FF,$A1,$F6,$EF,$F2,$9F,$7C,$79,$9A,$FF,$86
	.dc.b	$5A,$FF,$85,$DB,$66,$93,$E2,$3F,$C6,$6E,$AD,$F4,$BE,$52,$5F,$F8
	.dc.b	$D2,$5F,$0D,$FE,$50,$6C,$7F,$8C,$DA,$FF,$8C,$5A,$F9,$E3,$B9,$1E
	.dc.b	$0D,$7F,$C7,$18,$3C,$8F,$F2,$EB,$63,$FC,$8A,$D7,$F6,$9E,$4F,$F4
	.dc.b	$2C,$FF,$1E,$FF,$44,$8F,$B6,$D7,$FC,$7C,$D7,$FC,$78,$D9,$FF,$43
	.dc.b	$78,$FF,$E5,$68,$C1,$1F,$F2,$CC,$3C,$EF,$29,$2B,$FC,$B3,$4F,$91
	.dc.b	$FF,$3A,$B6,$3F,$CB,$0D,$7F,$CA,$CD,$7C,$F1,$E1,$23,$C1,$AF,$F9
	.dc.b	$77,$E3,$BF,$8D,$3F,$C5,$5B,$8F,$EC,$1B,$D5,$FE,$5D,$FA,$57,$CB
	.dc.b	$FF,$A2,$1B,$7F,$E6,$B6,$BF,$E6,$96,$BF,$A4,$6B,$CE,$3F,$CF,$71
	.dc.b	$88,$FF,$33,$37,$33,$69,$ED,$7F,$A6,$1B,$5F,$E8,$06,$BF,$CB,$E5
	.dc.b	$F4,$BE,$73,$B9,$23,$D0,$FF,$A3,$3A,$BF,$96,$E7,$FC,$41,$B7,$FE
	.dc.b	$7D,$6D,$7F,$9E,$DF,$E2,$4F,$FA,$41,$CE,$26,$93,$B7,$FD,$5F,$F2
	.dc.b	$5F,$EE,$EC,$3F,$8E,$B7,$0F,$F4,$53,$70,$FF,$44,$3F,$CF,$9F,$E9
	.dc.b	$4E,$70,$34,$9D,$BF,$EB,$7F,$A4,$7C,$CD,$25,$FF,$95,$BE,$2F,$FB
	.dc.b	$03,$FE,$DD,$9D,$BF,$E5,$A6,$BF,$DB,$F8,$D1,$8C,$FF,$AA,$C2,$B9
	.dc.b	$B7,$C8,$FE,$97,$7F,$EE,$99,$46,$27,$53,$FE,$C9,$9C,$BC,$3E,$95
	.dc.b	$EB,$FF,$8D,$63,$F8,$97,$47,$F8,$01,$4B,$E3,$F1,$D9,$FE,$3C,$8C
	.dc.b	$4F,$F8,$1A,$31,$7C,$DB,$9E,$70,$E9,$7C,$2F,$B9,$FE,$0E,$71,$5C
	.dc.b	$23,$27,$8A,$FC,$CF,$F9,$02,$7D,$37,$EB,$9C,$CF,$66,$BD,$7E,$43
	.dc.b	$ED,$BF,$DC,$A9,$18,$8E,$8D,$7C,$ED,$7F,$C9,$FF,$11,$FE,$74,$4D
	.dc.b	$23,$FE,$78,$7E,$03,$F2,$FE,$33,$FC,$1F,$8F,$F1,$F9,$CC,$FF,$8B
	.dc.b	$A7,$39,$7F,$81,$A7,$32,$9C,$BE,$23,$E2,$99,$E4,$EF,$B7,$FB,$4F
	.dc.b	$C0,$FF,$0E,$CF,$FC,$01,$F8,$3F,$E9,$35,$FE,$12,$AD,$36,$33,$3F
	.dc.b	$E6,$C9,$7C,$E3,$E1,$7F,$C4,$FC,$1F,$94,$8C,$61,$FE,$B4,$C3,$8C
	.dc.b	$E6,$B7,$3F,$E4,$AB,$E7,$18,$C1,$18,$9E,$58,$E6,$31,$3F,$E3,$DF
	.dc.b	$A5,$FC,$BD,$7F,$95,$30,$FF,$0C,$C6,$27,$5B,$F5,$E9,$4E,$6B,$7C
	.dc.b	$59,$8C,$61,$FE,$5E,$AD,$DE,$6B,$97,$D2,$BE,$62,$98,$F8,$EB,$87
	.dc.b	$8C,$E5,$3B,$71,$9B,$14,$E6,$3F,$CB,$0A,$B9,$C7,$F9,$6E,$9F,$0D
	.dc.b	$F6,$A3,$CD,$18,$3F,$C9,$77,$FE,$6F,$9D,$4F,$79,$D3,$FE,$A6,$9F
	.dc.b	$CB,$FB,$3F,$17,$93,$FD,$13,$5F,$CF,$6F,$13,$F2,$7E,$FF,$8E,$1F
	.dc.b	$E3,$06,$BF,$1D,$AF,$B5,$B8,$1A,$4A,$F2,$C6,$5B,$CA,$F9,$8F,$B0
	.dc.b	$DC,$7E,$4D,$6D,$B1,$9C,$3F,$C1,$BE,$EF,$4E,$3F,$F5,$3D,$7F,$7A
	.dc.b	$8C,$11,$C1,$27,$FB,$EC,$79,$3E,$3D,$F3,$9F,$89,$6D,$7F,$82,$27
	.dc.b	$13,$F8,$DF,$B2,$7C,$1F,$B1,$B5,$F5,$1B,$87,$D1,$6B,$FE,$5A,$6B
	.dc.b	$F9,$16,$BF,$87,$6E,$66,$F6,$FF,$9C,$2B,$FC,$81,$EB,$F8,$F0,$FD
	.dc.b	$04,$E7,$0F,$F5,$27,$A4,$FA,$7F,$A2,$DC,$3F,$CC,$11,$82,$38,$24
	.dc.b	$FE,$26,$78,$3F,$8B,$F9,$5F,$E9,$F3,$F5,$9B,$1E,$0A,$FD,$94,$63
	.dc.b	$B3,$FD,$1F,$F8,$AF,$F1,$7A,$FC,$A3,$5F,$F3,$EB,$5F,$DA,$35,$FD
	.dc.b	$53,$75,$34,$9C,$5F,$8E,$F6,$3F,$DA,$59,$7F,$A2,$CF,$E0,$93,$E1
	.dc.b	$C3,$FD,$43,$3A,$C3,$FD,$94,$FE,$68,$9C,$BC,$C9,$7F,$E8,$E6,$CF
	.dc.b	$EB,$67,$F4,$3E,$57,$F6,$7A,$FF,$51,$78,$DF,$C4,$E7,$F6,$78,$7F
	.dc.b	$AC,$2B,$57,$FE,$D9,$FA,$47,$8D,$FC,$65,$7D,$95,$8C,$3F,$C7,$7D
	.dc.b	$EF,$45,$7F,$A9,$1A,$FF,$B0,$B3,$FD,$7C,$62,$3F,$89,$7D,$2F,$F0
	.dc.b	$F9,$FD,$6B,$63,$FD,$D4,$BE,$6F,$F9,$53,$9C,$FF,$DC,$F5,$FD,$13
	.dc.b	$9F,$F8,$7B,$4F,$F6,$E9,$6F,$FC,$14,$1C,$3F,$E0,$D6,$A3,$5E,$57
	.dc.b	$E6,$5D,$35,$D4,$70,$7F,$DF,$1D,$1F,$CF,$FE,$13,$F5,$F8,$79,$AF
	.dc.b	$F4,$4F,$FB,$7A,$33,$7F,$CF,$FE,$27,$FD,$7E,$FF,$0E,$5F,$53,$FD
	.dc.b	$25,$CE,$2B,$FF,$73,$C6,$AB,$87,$C0,$6E,$F8,$36,$3A,$7D,$37,$33
	.dc.b	$A7,$B7,$63,$24,$C7,$B3,$E5,$8D,$46,$0F,$8D,$F0,$FF,$EE,$89,$FE
	.dc.b	$9B,$D4,$FA,$67,$CB,$7F,$6D,$B5,$ED,$E5,$5B,$9D,$4E,$75,$FA,$D5
	.dc.b	$FE,$0F,$BE,$78,$7F,$5A,$8E,$7F,$0D,$FD,$E7,$1E,$53,$E2,$AD,$D7
	.dc.b	$9F,$B9,$BB,$E9,$38,$56,$B2,$F9,$71,$F3,$FD,$39,$8F,$D3,$F7,$AE
	.dc.b	$5C,$F1,$FF,$06,$CF,$D9,$AF,$AD,$5A,$C3,$EA,$FE,$0F,$F6,$7E,$BB
	.dc.b	$ED,$36,$FE,$F3,$71,$E8,$DA,$F8,$3F,$07,$F9,$3C,$3F,$B5,$B6,$FF
	.dc.b	$37,$68,$C3,$F9,$E6,$BF,$E6,$96,$BF,$9D,$6B,$C4,$FF,$54,$C7,$48
	.dc.b	$CE,$BF,$23,$E4,$FF,$24,$7F,$89,$E3,$BE,$1F,$92,$8F,$CB,$78,$F2
	.dc.b	$FF,$35,$E8,$DC,$7E,$44,$DB,$FC,$CC,$DA,$FD,$17,$D5,$7E,$7C,$FF
	.dc.b	$18,$7E,$82,$C7,$13,$FA,$59,$B8,$FF,$A5,$EB,$53,$93,$FC,$3F,$97
	.dc.b	$FA,$A3,$E6,$7E,$4F,$C1,$9A,$49,$F8,$11,$96,$C6,$6F,$E6,$4E,$5F
	.dc.b	$F0,$D7,$17,$F2,$4A,$FD,$C7,$89,$F1,$6B,$5A,$FE,$DA,$3F,$75,$F0
	.dc.b	$72,$BF,$7D,$E3,$5C,$7F,$19,$36,$FF,$7B,$36,$BF,$83,$7C,$47,$F0
	.dc.b	$33,$F8,$27,$A5,$63,$F8,$44,$5C,$3F,$2D,$36,$FF,$83,$F9,$5F,$E2
	.dc.b	$0D,$0D,$26,$1F,$E3,$A8,$FE,$A8,$FF,$15,$EB,$FE,$25,$3F,$C4,$DF
	.dc.b	$E2,$98,$49,$FE,$31,$16,$3F,$4C,$5B,$FE,$2F,$16,$7F,$CB,$DF,$87
	.dc.b	$BE,$AB,$FC,$66,$D8,$FF,$25,$B6,$FF,$C9,$0D,$7F,$D8,$9C,$E8,$FE
	.dc.b	$8D,$3F,$E3,$66,$CF,$F0,$17,$96,$1F,$C3,$2B,$87,$95,$FE,$75,$9F
	.dc.b	$EA,$15,$D1,$AF,$9B,$DB,$F2,$9E,$6F,$77,$81,$C5,$B8,$F8,$62,$CF
	.dc.b	$86,$6E,$3E,$1B,$FF,$38,$36,$B9,$5F,$13,$D5,$F2,$3F,$10,$8F,$C4
	.dc.b	$7E,$3F,$B7,$E9,$3D,$4D,$8F,$85,$F5,$5F,$30,$8F,$CC,$6E,$5F,$31
	.dc.b	$AF,$C1,$7E,$9F,$C8,$7C,$53,$E3,$36,$7F,$D1,$7E,$EB,$E8,$DF,$17
	.dc.b	$FE,$80,$8C,$36,$33,$3F,$EB,$F6,$E1,$FE,$BD,$6B,$FE,$FF,$AE,$33
	.dc.b	$BA,$E4,$7E,$1B,$EA,$7E,$C1,$FE,$B5,$BE,$7A,$FC,$D8,$CD,$7E,$0D
	.dc.b	$D5,$63,$0F,$EB,$58,$7A,$7E,$77,$FC,$17,$1F,$EC,$5C,$51,$88,$FF
	.dc.b	$62,$C6,$30,$FF,$00,$C6,$E7,$55,$CE,$73,$3C,$27,$35,$C6,$3F,$DC
	.dc.b	$75,$B6,$C6,$7E,$0D,$F1,$9F,$A6,$FA,$AF,$EC,$71,$8A,$E9,$8F,$F8
	.dc.b	$5E,$31,$CB,$FA,$3A,$DD,$FF,$86,$ED,$F4,$EF,$96,$FF,$7D,$43,$0E
	.dc.b	$8E,$A3,$31,$DF,$E4,$5E,$E7,$D8,$8D,$DF,$3E,$BF,$16,$3D,$E3,$18
	.dc.b	$F5,$8C,$63,$D1,$FF,$17,$63,$D3,$83,$CF,$5F,$E3,$1B,$E3,$8F,$6C
	.dc.b	$3C,$F3,$FE,$3A,$9E,$AD,$8C,$BD,$B1,$FE,$CA,$FC,$0F,$9F,$7D,$36
	.dc.b	$2D,$18,$9F,$ED,$71,$89,$D4,$E6,$7C,$0E,$F0,$F8,$2D,$CF,$E0,$36
	.dc.b	$59,$BE,$13,$FD,$D9,$1F,$18,$F7,$4E,$DF,$36,$19,$69,$F2,$4E,$73
	.dc.b	$C9,$C7,$13,$F3,$47,$5F,$DF,$CF,$F0,$7D,$8F,$99,$C6,$BF,$CC,$1B
	.dc.b	$8B,$1E,$58,$61,$B1,$F9,$A6,$BF,$98,$C7,$BB,$96,$B9,$9F,$F3,$63
	.dc.b	$FE,$6B,$9E,$3C,$BF,$BD,$AD,$B6,$B7,$4E,$5F,$E7,$9F,$1F,$F9,$9E
	.dc.b	$AF,$E5,$0D,$BF,$E9,$2D,$AF,$56,$1F,$D2,$DF,$E8,$51,$9B,$E7,$18
	.dc.b	$3F,$8C,$CE,$DF,$E0,$6D,$7C,$DE,$0F,$E2,$B2,$F9,$CD,$9D,$4F,$B5
	.dc.b	$AF,$8A,$34,$47,$FA,$DF,$C7,$F8,$F6,$34,$C7,$C3,$49,$FE,$C2,$FE
	.dc.b	$6B,$1F,$EB,$CD,$A3,$49,$3C,$08,$7F,$68,$F8,$3F,$01,$1F,$EC,$DF
	.dc.b	$57,$E3,$3A,$11,$FE,$E2,$F7,$FA,$BF,$41,$23,$FB,$77,$15,$7F,$B7
	.dc.b	$45,$E0,$FE,$BD,$37,$83,$FB,$FC,$DA,$FB,$D9,$7F,$10,$BD,$C7,$F8
	.dc.b	$02,$FD,$C4,$7F,$C0,$7E,$EF,$C9,$23,$FD,$FF,$E3,$FE,$61,$F1,$48
	.dc.b	$FF,$82,$9F,$91,$CF,$F4,$49,$1F,$EB,$FB,$FF,$5A,$7D,$4E,$23,$FC
	.dc.b	$11,$36,$73,$1B,$CF,$FD,$8D,$BB,$E1,$58,$7E,$3D,$B5,$B3,$F3,$8E
	.dc.b	$1F,$F4,$A2,$F9,$AE,$07,$DF,$71,$1F,$E7,$EA,$ED,$39,$35,$F2,$DD
	.dc.b	$C9,$3D,$E7,$C9,$F2,$D8,$8E,$F3,$72,$FF,$69,$4F,$9F,$E8,$A5,$71
	.dc.b	$C5,$1E,$BF,$A9,$CC,$FA,$22,$DF,$FB,$6E,$7E,$07,$D4,$CA,$E4,$FE
	.dc.b	$D5,$7D,$3B,$97,$0D,$8F,$E6,$ED,$7F,$D3,$AD,$7F,$DF,$2D,$7F,$AE
	.dc.b	$35,$FF,$7C,$35,$EF,$1F,$D0,$9A,$7C,$BC,$5F,$41,$F5,$C9,$1F,$E1
	.dc.b	$CF,$39,$D1,$88,$FA,$F3,$73,$FF,$6A,$F6,$AF,$75,$F2,$7E,$27,$F0
	.dc.b	$44,$7E,$67,$C4,$D7,$F0,$31,$6F,$FC,$A9,$92,$BF,$15,$F5,$32,$FE
	.dc.b	$A3,$F2,$FE,$A1,$EC,$D3,$DC,$E2,$7F,$B0,$B5,$FF,$63,$B5,$FF,$02
	.dc.b	$07,$0F,$F8,$1D,$A8,$D4,$76,$3F,$2D,$F4,$5F,$9F,$F9,$69,$3F,$89
	.dc.b	$8F,$D1,$18,$7F,$3F,$18,$7F,$1D,$E9,$FE,$9C,$8F,$E4,$7E,$06,$BF
	.dc.b	$A6,$8B,$A7,$C6,$9F,$C6,$7D,$3F,$F8,$6E,$BF,$31,$E9,$7E,$C4,$CF
	.dc.b	$A5,$F9,$1F,$A7,$7C,$8E,$2B,$F6,$82,$F9,$2C,$5E,$0F,$F8,$A6,$F8
	.dc.b	$CF,$06,$BB,$8F,$F1,$75,$72,$AE,$33,$F2,$63,$13,$CE,$FF,$83,$61
	.dc.b	$2B,$F6,$0F,$F0,$9F,$47,$11,$FC,$1A,$2E,$1F,$86,$9F,$EC,$1C,$CF
	.dc.b	$A7,$FF,$0B,$4F,$E5,$17,$8B,$A6,$E1,$FE,$78,$77,$97,$F7,$56,$C7
	.dc.b	$F8,$0D,$DB,$FE,$66,$6D,$75,$F1,$7F,$B6,$78,$DF,$E3,$48,$FF,$56
	.dc.b	$36,$FF,$92,$B6,$FF,$15,$F4,$FC,$A7,$FC,$A8,$F6,$8F,$C9,$B5,$CE
	.dc.b	$3F,$B9,$8C,$47,$28,$C1,$1F,$F1,$FC,$6B,$C5,$FF,$2B,$FD,$47,$F8
	.dc.b	$07,$D8,$34,$C7,$CC,$7F,$81,$7E,$69,$F1,$BF,$DB,$DB,$8F,$F3,$16
	.dc.b	$BF,$A8,$C1,$18,$8F,$F6,$9F,$C2,$3D,$DF,$F0,$82,$F7,$E9,$88,$FF
	.dc.b	$0A,$7C,$DB,$C4,$A4,$BF,$E7,$50,$CF,$67,$F8,$0D,$C4,$7F,$45,$6B
	.dc.b	$E5,$6B,$FE,$E6,$6B,$FD,$61,$AF,$F8,$BD,$B9,$1B,$F3,$8F,$88,$FF
	.dc.b	$15,$36,$BF,$A7,$35,$FD,$11,$FE,$39,$F7,$BF,$C2,$AF,$F5,$36,$BF
	.dc.b	$25,$AF,$FB,$E9,$AF,$F6,$C6,$BC,$FD,$7F,$F1,$14,$63,$0F,$F3,$64
	.dc.b	$60,$FF,$13,$7C,$EB,$E5,$7F,$C8,$0D,$BF,$EC,$8D,$7C,$F4,$F9,$E7
	.dc.b	$BD,$FE,$47,$6C,$FF,$71,$6B,$F8,$16,$BF,$EA,$D9,$E5,$3F,$E3,$79
	.dc.b	$CC,$FF,$81,$9A,$FF,$97,$5B,$3C,$23,$FB,$F7,$CF,$FF,$B4,$1F,$EF
	.dc.b	$4D,$7C,$ED,$EB,$FE,$1B,$3F,$BE,$E5,$FF,$2F,$74,$FE,$4E,$FF,$C2
	.dc.b	$AD,$7F,$CD,$8D,$9E,$11,$FE,$0D,$8C,$12,$7F,$CA,$0E,$E5,$E2,$FF
	.dc.b	$9A,$DB,$5F,$E0,$F6,$BF,$33,$BD,$CF,$FE,$71,$6C,$7F,$86,$DA,$FE
	.dc.b	$AD,$AF,$CB,$6B,$FE,$2E,$6B,$FE,$85,$6C,$FF,$A6,$67,$FC,$CD,$18
	.dc.b	$FA,$17,$AB,$FE,$81,$6D,$FF,$89,$5A,$FE,$0B,$E0,$DF,$39,$FE,$87
	.dc.b	$6C,$7F,$8C,$5A,$FF,$01,$6B,$F8,$16,$BF,$E4,$46,$BF,$E9,$B6,$E3
	.dc.b	$FE,$EF,$9F,$F3,$E4,$63,$C0,$7B,$3F,$E9,$76,$DF,$F8,$ED,$AF,$E6
	.dc.b	$3E,$9D,$F3,$7F,$E9,$D6,$C7,$F9,$1D,$AF,$F1,$76,$BF,$97,$6B,$FE
	.dc.b	$5A,$F8,$7F,$E1,$3C,$5F,$8C,$6D,$E7,$B7,$FD,$55,$1F,$D6,$AB,$FD
	.dc.b	$6C,$E9,$B1,$A3,$7E,$4F,$FA,$C6,$3F,$D5,$D7,$F9,$86,$DF,$F9,$A9
	.dc.b	$A7,$FB,$07,$E5,$3F,$C2,$B3,$BF,$85,$E4,$49,$FF,$2F,$4E,$AB,$FD
	.dc.b	$9D,$F4,$1F,$CF,$A7,$FC,$E1,$C6,$8C,$47,$F1,$68,$C7,$8D,$FD,$AB
	.dc.b	$E0,$1F,$39,$FD,$B2,$F7,$9F,$F8,$F9,$B8,$F6,$BE,$15,$31,$FF,$15
	.dc.b	$9F,$EE,$BC,$39,$B8,$C5,$18,$8F,$F1,$8C,$62,$7F,$CA,$DF,$4B,$FE
	.dc.b	$C4,$CB,$FC,$61,$DA,$F9,$5F,$F1,$96,$0A,$FE,$57,$39,$8F,$F1,$B4
	.dc.b	$62,$76,$F3,$8F,$E3,$D1,$8F,$89,$FF,$0F,$46,$0E,$3E,$1F,$6B,$E9
	.dc.b	$1D,$E3,$15,$C7,$0F,$2B,$67,$E0,$E1,$DD,$FF,$26,$B5,$D1,$8F,$05
	.dc.b	$C1,$F2,$23,$D9,$18,$79,$B5,$E1,$1F,$E7,$B5,$D1,$48,$FF,$9F,$DB
	.dc.b	$97,$AA,$1F,$1D,$FE,$34,$8C,$47,$23,$E1,$1F,$E5,$F8,$D6,$3F,$72
	.dc.b	$BD,$F3,$F4,$A7,$2F,$F9,$AA,$32,$F1,$E1,$DE,$FF,$80,$E8,$FF,$21
	.dc.b	$CF,$DA,$8C,$4F,$E1,$78,$7F,$76,$DA,$DD,$7E,$73,$A7,$FC,$97,$BF
	.dc.b	$E4,$DB,$5F,$E7,$5C,$39,$D7,$E7,$79,$3F,$85,$CF,$E8,$9B,$19,$AE
	.dc.b	$2F,$E5,$F0,$B7,$FA,$79,$CF,$5B,$E9,$5C,$98,$6E,$5F,$EC,$F7,$FB
	.dc.b	$4B,$63,$44,$75,$C5,$72,$FE,$E3,$F4,$AF,$8E,$EE,$DA,$FF,$6D,$2F
	.dc.b	$95,$CC,$C9,$CB,$C2,$31,$5F,$DF,$25,$EE,$7C,$A8,$67,$FE,$7C,$8C
	.dc.b	$36,$7F,$DD,$C7,$03,$FA,$07,$CE,$FC,$2E,$33,$F1,$23,$57,$FD,$26
	.dc.b	$3E,$3F,$85,$88,$FF,$02,$35,$FF,$55,$5A,$31,$1E,$46,$C6,$70,$F2
	.dc.b	$C6,$08,$FA,$39,$5F,$2D,$F5,$9B,$8F,$F5,$D9,$C4,$FD,$AE,$27,$D2
	.dc.b	$FC,$06,$D7,$F7,$38,$C1,$1C,$13,$2F,$EE,$D1,$F2,$3C,$8F,$F6,$8C
	.dc.b	$FF,$B1,$79,$FF,$C2,$71,$8A,$FF,$01,$57,$E2,$A3,$13,$FE,$18,$F2
	.dc.b	$BF,$CB,$97,$FE,$20,$F0,$FE,$9C,$63,$0F,$F3,$73,$5F,$F1,$73,$5F
	.dc.b	$96,$D7,$EE,$B7,$63,$7B,$3F,$DF,$77,$E7,$3F,$C6,$9F,$11,$F9,$E9
	.dc.b	$F0,$63,$FE,$2C,$AD,$E1,$FE,$46,$F3,$1F,$3F,$F8,$86,$E3,$FD,$03
	.dc.b	$24,$62,$BF,$D0,$CD,$7F,$C8,$CD,$7F,$06,$D7,$F2,$AD,$8E,$75,$F9
	.dc.b	$6F,$3D,$F2,$1F,$B4,$6E,$1F,$E5,$C9,$C4,$FE,$E3,$D1,$FE,$3F,$3E
	.dc.b	$56,$E3,$FE,$99,$6B,$FE,$59,$6B,$F9,$66,$BF,$A8,$6E,$06,$98,$F4
	.dc.b	$FF,$DC,$F5,$FE,$98,$3C,$5C,$BF,$C2,$A3,$37,$FE,$70,$7F,$3E,$4E
	.dc.b	$5F,$D9,$B6,$BF,$D3,$91,$82,$3F,$3F,$CF,$78,$FF,$C5,$9B,$1F,$EA
	.dc.b	$89,$C4,$FF,$1A,$F5,$FF,$B5,$4F,$E0,$70,$FE,$07,$0F,$9C,$D7,$8F
	.dc.b	$B7,$F9,$0F,$B3,$E8,$AF,$F3,$F3,$5F,$DD,$B5,$FE,$1A,$DD,$9F,$4B
	.dc.b	$FC,$67,$1F,$04,$63,$C0,$FE,$5B,$2F,$9B,$B1,$FC,$D2,$BF,$C3,$1D
	.dc.b	$3F,$C6,$5A,$7F,$83,$8B,$8F,$F8,$10,$38,$7F,$C0,$ED,$46,$BC,$9F
	.dc.b	$CE,$BA,$6B,$A8,$E0,$FF,$28,$F8,$BF,$D2,$C7,$F1,$A6,$C7,$3A,$FE
	.dc.b	$8F,$18,$C9,$F3,$8F,$A4,$FE,$96,$BE,$96,$C6,$1B,$8F,$C6,$9D,$36
	.dc.b	$74,$E7,$E1,$79,$9D,$3D,$8B,$19,$26,$7F,$EB,$EB,$FF,$3E,$C6,$3B
	.dc.b	$FF,$B9,$FC,$13,$E3,$BF,$B1,$F5,$BE,$97,$FD,$E9,$7F,$87,$C5,$7B
	.dc.b	$8F,$C5,$56,$FE,$07,$FB,$5D,$6B,$0E,$78,$72,$8F,$F7,$29,$9B,$E3
	.dc.b	$19,$C7,$EC,$CF,$F8,$5A,$31,$3F,$D2,$32,$7D,$2F,$4C,$7E,$5B,$F3
	.dc.b	$3D,$49,$5F,$27,$E9,$6E,$3F,$06,$6E,$5F,$EC,$8C,$55,$8B,$ED,$EC
	.dc.b	$FD,$6F,$82,$FA,$ED,$8F,$A8,$DC,$7E,$63,$70,$F4,$FD,$1F,$E3,$B0
	.dc.b	$FE,$E0,$DB,$FF,$32,$B6,$FF,$30,$D7,$FC,$BA,$D7,$F2,$ED,$4D,$31
	.dc.b	$CB,$F8,$EF,$83,$FC,$69,$FE,$1C,$8E,$D8,$7B,$E3,$F2,$3E,$3C,$C3
	.dc.b	$F1,$DE,$95,$C7,$F0,$D3,$6F,$E3,$4D,$AF,$A5,$F4,$BF,$9A,$3F,$C4
	.dc.b	$FF,$4A,$C7,$13,$F9,$99,$B8,$FF,$A3,$EB,$53,$93,$F8,$96,$5F,$E6
	.dc.b	$2F,$85,$F8,$FE,$F7,$FB,$BA,$7E,$3B,$6F,$E3,$B5,$F8,$ED,$7E,$3B
	.dc.b	$5F,$8E,$DA,$ED,$3F,$D5,$F8,$7F,$90,$3F,$AF,$F1,$3E,$45,$77,$D7
	.dc.b	$F0,$D1,$FB,$8F,$A5,$95,$FB,$5F,$1A,$E5,$FA,$D9,$B7,$F8,$D9,$B5
	.dc.b	$F9,$9F,$87,$FE,$02,$7F,$02,$F9,$3B,$1F,$BF,$8B,$87,$ED,$E6,$DF
	.dc.b	$E5,$FC,$CF,$F1,$BE,$86,$93,$0F,$67,$6F,$F0,$E3,$F8,$7C,$58,$FE
	.dc.b	$19,$87,$F1,$23,$07,$F0,$53,$0E,$8F,$E2,$53,$88,$5F,$ED,$B9,$5F
	.dc.b	$3C,$F8,$5F,$F2,$5B,$70,$FF,$02,$36,$FF,$9B,$57,$0A,$DF,$93,$FE
	.dc.b	$6E,$AF,$E8,$57,$FD,$19,$E1,$87,$77,$38,$78,$A7,$FC,$97,$1F,$D3
	.dc.b	$63,$AB,$5F,$F6,$3E,$1E,$6F,$27,$CA,$79,$BD,$DF,$01,$E1,$6D,$F8
	.dc.b	$62,$D7,$73,$FC,$D3,$7D,$F0,$F2,$4E,$9B,$5E,$9B,$C3,$E6,$FA,$5F
	.dc.b	$8A,$47,$E2,$BF,$23,$DC,$F8,$47,$C3,$6C,$7C,$3F,$A5,$F9,$A4,$7E
	.dc.b	$6B,$72,$F8,$2D,$7E,$0E,$48,$FA,$BF,$33,$F3,$CF,$A0,$DA,$FF,$02
	.dc.b	$7B,$CF,$A5,$7C,$B7,$FA,$26,$30,$D8,$CC,$FF,$BF,$5B,$8F,$FB,$F1
	.dc.b	$AF,$FA,$FD,$AF,$FA,$DD,$AA,$BA,$CF,$E2,$3E,$43,$F0,$87,$FA,$D2
	.dc.b	$BA,$ED,$F8,$38,$CC,$FE,$1D,$D6,$18,$C3,$EC,$61,$EB,$F9,$7F,$E9
	.dc.b	$B7,$E3,$6C,$F8,$DA,$F8,$F1,$46,$2B,$95,$7F,$81,$6B,$83,$5E,$13
	.dc.b	$BA,$E3,$39,$AE,$35,$FE,$F8,$F6,$FF,$33,$F2,$3F,$D8,$3E,$17,$FC
	.dc.b	$BB,$3C,$AF,$FC,$2F,$18,$E4,$FD,$7D,$6E,$FF,$C3,$76,$FA,$57,$C7
	.dc.b	$FF,$BF,$A3,$B1,$FE,$20,$8C,$F4,$F4,$AF,$F1,$1F,$C8,$3D,$BF,$B3
	.dc.b	$9F,$8B,$0E,$98,$F3,$8C,$63,$CA,$7A,$9A,$C7,$A2,$C3,$A7,$07,$9E
	.dc.b	$3C,$58,$73,$CB,$D3,$87,$39,$FF,$1D,$35,$CC,$FA,$DC,$B5,$E3,$F2
	.dc.b	$CF,$95,$FF,$21,$47,$BE,$30,$E2,$91,$88,$EE,$D7,$9C,$FA,$9F,$99
	.dc.b	$1C,$B3,$F8,$2D,$C3,$E4,$FC,$1F,$FB,$D6,$B8,$61,$DB,$0B,$7B,$9F
	.dc.b	$24,$7C,$6C,$D7,$C6,$7E,$BF,$0B,$E9,$1D,$BF,$7B,$4F,$C1,$75,$3E
	.dc.b	$97,$1B,$7F,$22,$6E,$3F,$AB,$C3,$8B,$F5,$31,$E9,$8F,$F9,$92,$3F
	.dc.b	$CC,$71,$87,$F2,$D3,$C1,$FF,$19,$9C,$E7,$83,$4D,$B8,$9E,$51,$C1
	.dc.b	$BC,$BF,$C4,$BE,$3B,$F2,$FD,$5F,$CE,$DB,$1F,$CE,$9B,$5E,$9F,$06
	.dc.b	$F4,$7F,$4E,$9B,$A9,$A7,$0C,$7E,$45,$5C,$FF,$AA,$61,$F8,$49,$CB
	.dc.b	$63,$2F,$FA,$26,$32,$D8,$CB,$FA,$E8,$CB,$63,$2F,$CA,$6B,$FD,$1B
	.dc.b	$E3,$FC,$E4,$7F,$AE,$7C,$27,$A8,$FC,$BB,$FD,$31,$B9,$7F,$6C,$6E
	.dc.b	$5F,$EB,$E9,$FE,$D7,$18,$23,$83,$FB,$7F,$C0,$FC,$62,$3F,$DA,$FE
	.dc.b	$7F,$E4,$9D,$08,$FF,$76,$7B,$FD,$2F,$CE,$24,$7F,$73,$E2,$AF,$F6
	.dc.b	$3C,$7F,$C0,$33,$8E,$0F,$EE,$D5,$78,$3F,$BA,$CD,$BF,$ED,$33,$F6
	.dc.b	$8F,$69,$1F,$F0,$3B,$F0,$3D,$EF,$BE,$48,$FF,$02,$7C,$77,$E0,$8F
	.dc.b	$8A,$47,$FC,$1E,$FC,$8E,$8F,$C3,$92,$3F,$DA,$3F,$08,$FA,$5C,$47
	.dc.b	$AF,$83,$F8,$26,$9F,$E0,$FB,$FC,$8D,$EE,$B8,$DE,$08,$F4,$FA,$5D
	.dc.b	$9F,$6A,$4D,$24,$7F,$16,$D9,$2F,$96,$EE,$49,$F3,$FC,$B5,$7E,$84
	.dc.b	$ED,$F5,$CF,$9A,$7C,$FF,$4B,$27,$DE,$FA,$BE,$6F,$11,$CA,$2E,$9F
	.dc.b	$E3,$BC,$15,$F0,$3E,$96,$5F,$C8,$7F,$6A,$BE,$95,$CB,$86,$C7,$E7
	.dc.b	$1A,$FF,$34,$6B,$FE,$F7,$6B,$FD,$69,$AF,$FB,$D9,$AF,$78,$FD,$3B
	.dc.b	$4F,$95,$6F,$A5,$FA,$E4,$8F,$F0,$DF,$9C,$E8,$C4,$7C,$A9,$B9,$FF
	.dc.b	$B4,$3B,$17,$BA,$FC,$3F,$C3,$FE,$08,$8F,$E2,$BE,$26,$BF,$52,$2D
	.dc.b	$FC,$B9,$FC,$4F,$D2,$CA,$FC,$8B,$F2,$BE,$95,$EC,$D3,$DA,$E2,$7F
	.dc.b	$82,$B5,$FE,$BA,$D7,$FC,$02,$1C,$3F,$E0,$46,$A3,$51,$D8,$FC,$B7
	.dc.b	$D2,$FE,$7B,$E5,$24,$FE,$9A,$3F,$42,$61,$FC,$8C,$61,$FD,$5F,$A7
	.dc.b	$FA,$62,$3F,$AE,$F8,$2A,$FE,$62,$2E,$7F,$8F,$9F,$C5,$FD,$27,$F8
	.dc.b	$62,$BF,$6B,$E8,$FE,$C0,$CF,$A3,$F9,$0F,$A5,$7C,$8E,$2B,$D9,$8F
	.dc.b	$F5,$6C,$39,$47,$F9,$DC,$57,$07,$7E,$3F,$F8,$A2,$31,$1B,$9E,$35
	.dc.b	$CF,$0F,$06,$1F,$C0,$99,$5F,$C2,$9F,$E0,$BE,$96,$23,$F6,$B1,$6B
	.dc.b	$F5,$93,$FD,$57,$E9,$7F,$86,$E5,$FD,$C1,$78,$D9,$35,$E2,$CC,$7F
	.dc.b	$CE,$73,$A6,$CE,$5F,$E6,$5E,$3F,$F8,$97,$1F,$F5,$A3,$70,$FC,$07
	.dc.b	$88,$8C,$57,$E0,$FE,$97,$A3,$F1,$DA,$FE,$31,$AF,$F9,$39,$AF,$1C
	.dc.b	$3F,$67,$18,$83,$49,$1F,$D7,$3D,$77,$D2,$7F,$7C,$6D,$7D,$41,$87
	.dc.b	$F9,$4E,$75,$F3,$FF,$E0,$2F,$9A,$7C,$6F,$F6,$66,$DF,$F2,$B6,$BF
	.dc.b	$EC,$E6,$BF,$EC,$2F,$84,$7B,$BF,$E0,$C5,$EF,$D3,$11,$FE,$10,$F9
	.dc.b	$B7,$85,$49,$7E,$7D,$12,$CF,$81,$FD,$E9,$C4,$FF,$3F,$6B,$FE,$E2
	.dc.b	$6B,$FE,$D6,$6B,$F6,$1A,$FC,$46,$CF,$82,$7F,$C0,$FF,$38,$F8,$8F
	.dc.b	$F1,$3B,$6F,$CC,$D7,$FB,$C7,$BD,$FE,$0B,$6F,$AF,$FE,$10,$8C,$47
	.dc.b	$FB,$A1,$AF,$E2,$9A,$FD,$66,$E1,$FE,$5A,$9E,$44,$9F,$EB,$DF,$3A
	.dc.b	$F9,$5F,$F1,$DB,$63,$E5,$35,$FF,$06,$7C,$F7,$F8,$F5,$F9,$CD,$7F
	.dc.b	$8F,$DA,$31,$1F,$EF,$E6,$BF,$A0,$6B,$F8,$C6,$C7,$1F,$7B,$FC,$5F
	.dc.b	$18,$33,$F3,$FF,$EC,$0A,$FB,$CD,$7F,$C4,$0D,$EB,$75,$FE,$BB,$A7
	.dc.b	$F2,$31,$8A,$FC,$DC,$7E,$49,$D7,$3F,$F8,$FA,$31,$1A,$8C,$11,$FF
	.dc.b	$03,$46,$B2,$FE,$0F,$3F,$E4,$65,$E2,$7F,$99,$9B,$5F,$95,$6B,$FE
	.dc.b	$37,$F3,$5D,$1F,$E6,$C6,$C7,$F8,$4D,$AF,$CE,$6B,$FE,$CE,$9F,$F7
	.dc.b	$B4,$FF,$95,$A7,$33,$FC,$39,$AF,$EE,$9B,$3C,$23,$FC,$41,$F4,$3F
	.dc.b	$E0,$E7,$F5,$4D,$7F,$C9,$8D,$CB,$E5,$4B,$E0,$DF,$39,$FE,$84,$6D
	.dc.b	$FF,$8A,$DA,$FE,$19,$AF,$A5,$AF,$F2,$46,$BF,$C4,$1B,$3D,$E7,$FC
	.dc.b	$E7,$18,$D5,$EC,$FF,$A5,$1B,$7F,$C0,$DA,$FF,$99,$7E,$95,$F3,$7F
	.dc.b	$E9,$A6,$C7,$F9,$05,$AF,$E7,$1A,$FC,$F6,$BF,$CE,$BE,$1F,$F8,$4E
	.dc.b	$7F,$11,$B7,$9E,$BF,$F5,$3C,$7F,$5A,$8C,$57,$F1,$78,$DB,$FC,$B9
	.dc.b	$BA,$7F,$BD,$FC,$8F,$F5,$7C,$FF,$A9,$30,$FA,$2D,$BF,$E9,$6D,$3F
	.dc.b	$9B,$FC,$A7,$F8,$4A,$77,$F0,$DC,$89,$3F,$C9,$E7,$55,$FE,$D0,$FA
	.dc.b	$0F,$E8,$32,$BF,$F1,$94,$61,$BE,$2F,$FA,$F7,$E0,$1F,$39,$FE,$BF
	.dc.b	$AD,$61,$CE,$B0,$6E,$66,$1F,$EC,$1B,$E2,$D8,$FF,$58,$F2,$FF,$9C
	.dc.b	$5F,$F2,$6F,$D2,$BE,$23,$C9,$5F,$C3,$FE,$57,$FC,$54,$A7,$F9,$57
	.dc.b	$72,$73,$18,$6D,$72,$9D,$57,$F8,$CE,$BB,$6B,$FE,$35,$9C,$CF,$5F
	.dc.b	$0B,$BC,$62,$BA,$4F,$1B,$E9,$87,$6C,$3F,$D9,$75,$D2,$BF,$C8,$A7
	.dc.b	$9E,$BD,$6F,$F9,$22,$34,$DF,$01,$C1,$C9,$B1,$CA,$7F,$DE,$13,$93
	.dc.b	$FC,$E7,$1F,$B5,$3F,$CE,$BD,$38,$23,$83,$E5,$1F,$2B,$E3,$BF,$93
	.dc.b	$67,$F0,$23,$FC,$AD,$3F,$0F,$3F,$F2,$ED,$69,$F5,$C6,$67,$97,$0D
	.dc.b	$7E,$AC,$55,$5C,$56,$EF,$B0,$7C,$DF,$F3,$79,$FF,$7F,$46,$30,$F9
	.dc.b	$35,$BA,$D5,$7F,$9A,$E3,$13,$AE,$4F,$E8,$6F,$F8,$6E,$11,$89,$FC
	.dc.b	$F4,$F1,$BE,$D1,$83,$CB,$3B,$E0,$FD,$1D,$7F,$4B,$6B,$D6,$B8,$C6
	.dc.b	$2B,$FD,$E9,$F4,$CF,$90,$E1,$7F,$E4,$28,$C3,$E5,$FC,$5F,$FD,$7E
	.dc.b	$FF,$86,$E3,$35,$CA,$B8,$BF,$E3,$48,$CD,$FF,$A9,$EF,$8C,$EE,$FE
	.dc.b	$07,$E2,$EF,$82,$E3,$A2,$31,$3F,$E3,$FA,$7E,$0B,$FD,$6A,$38,$DF
	.dc.b	$0A,$DD,$7C,$5B,$E5,$3E,$48,$E3,$E3,$6E,$3C,$F9,$3F,$38,$FE,$0F
	.dc.b	$8F,$E8,$A3,$1D,$5F,$6B,$0F,$E2,$15,$A7,$F8,$D3,$5F,$DF,$E2,$FB
	.dc.b	$8F,$F3,$BE,$1F,$75,$B3,$F6,$70,$EB,$39,$F2,$BC,$78,$3E,$93,$C1
	.dc.b	$5E,$56,$C6,$6B,$CC,$DB,$F1,$CE,$89,$3A,$7D,$5E,$17,$F6,$5F,$9F
	.dc.b	$FA,$ED,$9F,$F3,$66,$1D,$DF,$F2,$C7,$42,$BE,$AF,$EA,$3F,$DB,$27
	.dc.b	$F3,$ED,$C7,$FD,$9B,$AA,$31,$1F,$11,$AF,$E8,$DA,$FE,$B9,$B1,$C7
	.dc.b	$CA,$FA,$0D,$AF,$E2,$BE,$33,$F5,$2F,$C6,$C7,$FB,$5E,$31,$D1,$E6
	.dc.b	$BF,$C9,$C6,$27,$FD,$10,$3F,$39,$FF,$0E,$BF,$C8,$0C,$79,$DF,$7D
	.dc.b	$A6,$63,$79,$FF,$BA,$A3,$2D,$8C,$BF,$D0,$23,$2D,$8C,$BF,$EA,$EF
	.dc.b	$B7,$79,$3F,$49,$B5,$E6,$AE,$FC,$1F,$E1,$C9,$CD,$FE,$77,$67,$D6
	.dc.b	$7C,$66,$E1,$F5,$23,$04,$70,$49,$FA,$F3,$C1,$FF,$33,$FA,$AF,$9A
	.dc.b	$FC,$2B,$6B,$F8,$B5,$6F,$1F,$F1,$7C,$E6,$FF,$8A,$57,$F2,$6C,$FE
	.dc.b	$57,$85,$E8,$9F,$88,$D4,$7F,$99,$1C,$36,$38,$92,$7E,$B4,$EB,$97
	.dc.b	$FD,$8B,$F2,$FF,$9A,$F3,$4F,$F1,$2A,$F1,$1F,$EB,$18,$CB,$FC,$2F
	.dc.b	$C0,$B8,$FF,$60,$6B,$FD,$E0,$FE,$1F,$5F,$C0,$5B,$3F,$D9,$9A,$71
	.dc.b	$8D,$4F,$28,$FF,$00,$B5,$FE,$97,$EA,$3E,$43,$F8,$FB,$6F,$C9,$78
	.dc.b	$9F,$E3,$1E,$AB,$E4,$E9,$FC,$6F,$17,$CC,$63,$0F,$E4,$D4,$F9,$CF
	.dc.b	$43,$63,$D0,$DC,$FF,$BF,$B5,$F8,$1F,$9F,$FF,$2D,$7F,$A7,$07,$1C
	.dc.b	$5F,$D8,$7F,$14,$7C,$4F,$F2,$D9,$CE,$1F,$B3,$FC,$C1,$F4,$DF,$CA
	.dc.b	$7B,$51,$89,$FD,$DB,$72,$FA,$13,$AA,$E1,$39,$9F,$F1,$5C,$E6,$B8
	.dc.b	$D6,$8F,$E0,$FF,$69,$FE,$C9,$8F,$ED,$8D,$AF,$ED,$15,$C5,$FE,$8D
	.dc.b	$19,$CF,$FD,$13,$F3,$5F,$EF,$5B,$FF,$7C,$36,$BF,$3B,$E8,$BE,$6B
	.dc.b	$F9,$4C,$E6,$F8,$FB,$B8,$3F,$25,$5B,$C3,$9C,$70,$E8,$91,$FC,$52
	.dc.b	$5F,$11,$8C,$BF,$BF,$75,$32,$61,$9C,$BF,$D9,$18,$FF,$82,$30,$C3
	.dc.b	$FA,$CF,$55,$F5,$EF,$96,$FE,$EB,$18,$7F,$B2,$7C,$7B,$E5,$73,$5B
	.dc.b	$BF,$DA,$E3,$FE,$D5,$C3,$84,$60,$ED,$3B,$D7,$F0,$F8,$73,$9D,$47
	.dc.b	$F8,$B2,$75,$7F,$3A,$7E,$49,$E0,$C6,$BC,$35,$60,$FB,$2F,$EA,$ED
	.dc.b	$9E,$BF,$59,$EF,$78,$46,$72,$F9,$2D,$C7,$FD,$8A,$D9,$FE,$23,$5C
	.dc.b	$6B,$77,$DB,$C7,$95,$DA,$7F,$CF,$0F,$68,$CC,$76,$7F,$C4,$35,$FC
	.dc.b	$73,$DC,$CE,$3F,$EF,$DF,$6D,$F8,$CF,$F2,$7C,$3F,$D0,$B1,$8F,$2B
	.dc.b	$FC,$1C,$DC,$BF,$CE,$6D,$9F,$E2,$8D,$78,$D7,$37,$F9,$5C,$65,$B1
	.dc.b	$95,$F6,$8F,$C8,$FF,$D9,$F7,$F5,$A3,$1F,$13,$FE,$3A,$C3,$C1,$92
	.dc.b	$BF,$9B,$D6,$9F,$9F,$F6,$9F,$50,$F8,$38,$BE,$D3,$13,$A2,$47,$F9
	.dc.b	$67,$6F,$E8,$33,$A8,$FA,$FE,$9F,$DE,$3E,$5F,$C2,$7E,$82,$FE,$EC
	.dc.b	$65,$A7,$E0,$A3,$53,$CA,$3F,$C3,$7F,$2F,$FE,$F2,$7F,$C7,$8D,$8F
	.dc.b	$F6,$93,$5F,$F5,$BE,$68,$C4,$7E,$89,$AF,$FB,$45,$B1,$BF,$8D,$FF
	.dc.b	$1B,$C6,$B5,$FF,$60,$7D,$6F,$CB,$F8,$83,$FE,$50,$9F,$15,$FE,$16
	.dc.b	$73,$8F,$E6,$BE,$6F,$F3,$E7,$C2,$F9,$6B,$19,$26,$7F,$2F,$44,$EA
	.dc.b	$3F,$43,$B7,$EA,$0F,$97,$F1,$D6,$FF,$4F,$16,$79,$4D,$CB,$FD,$3D
	.dc.b	$F9,$0B,$EA,$BF,$B1,$37,$0F,$EC,$4D,$BF,$F6,$4B,$5F,$ED,$5E,$76
	.dc.b	$6F,$FA,$45,$FF,$9C,$EB,$17,$FE,$E6,$AB,$7F,$EB,$0F,$C6,$7F,$85
	.dc.b	$3F,$E7,$B6,$E1,$CA,$BF,$D5,$11,$87,$FD,$5C,$D9,$E5,$E1,$FF,$A0
	.dc.b	$6F,$AF,$C0,$AB,$4B,$F0,$57,$C9,$FF,$94,$1B,$8F,$FA,$F1,$B8,$FF
	.dc.b	$97,$9A,$FF,$BF,$BC,$3F,$E2,$FF,$0D,$FC,$58,$FF,$4D,$CF,$0D,$3F
	.dc.b	$88,$C7,$FA,$7B,$F1,$9F,$F6,$7E,$1F,$EC,$86,$E1,$FE,$6A,$6B,$FE
	.dc.b	$8C,$6B,$FE,$49,$F2,$7F,$C9,$07,$3F,$26,$D7,$F2,$9F,$81,$FE,$96
	.dc.b	$D9,$D3,$49,$5F,$D3,$A7,$6F,$FA,$A6,$FC,$78,$FF,$94,$6F,$FD,$6B
	.dc.b	$EF,$FF,$A6,$AB,$FD,$3E,$D7,$FD,$BC,$D7,$FD,$D2,$DD,$3F,$9D,$F8
	.dc.b	$5F,$CE,$8E,$7F,$13,$73,$FF,$71,$7B,$4F,$A5,$FF,$40,$B6,$3F,$DD
	.dc.b	$AD,$C7,$FD,$1E,$D7,$FD,$62,$D7,$FD,$E0,$D7,$FD,$F6,$DE,$2F,$F1
	.dc.b	$4D,$7F,$A9,$3C,$CF,$F0,$15,$FF,$95,$32,$FE,$A3,$F3,$3F,$ED,$4C
	.dc.b	$3C,$1B,$7F,$4A,$8F,$F6,$BF,$D7,$7E,$E3,$3F,$F5,$0B,$5F,$F6,$1B
	.dc.b	$5F,$F2,$FE,$68,$C4,$23,$1E,$17,$FA,$82,$73,$18,$CB,$FC,$A9,$5B
	.dc.b	$9E,$31,$FC,$03,$CF,$FF,$30,$77,$FF,$6E,$8E,$F3,$E5,$CB,$FC,$0F
	.dc.b	$39,$BF,$E7,$8D,$AE,$47,$F7,$8C,$3C,$13,$FE,$7C,$9E,$0B,$E7,$1E
	.dc.b	$17,$FB,$E3,$F1,$9F,$F2,$D4,$BE,$37,$F8,$43,$FE,$7E,$AD,$E3,$D6
	.dc.b	$B7,$5A,$AE,$97,$CF,$E0,$FF,$A8,$47,$3F,$0B,$FD,$0E,$EB,$0E,$53
	.dc.b	$AB,$EF,$9B,$1E,$B8,$7D,$06,$E1,$E3,$C7,$B6,$3F,$E9,$2C,$3C,$5F
	.dc.b	$61,$FC,$0B,$B3,$18,$F8,$23,$18,$78,$C7,$B8,$FA,$2F,$E2,$F8,$72
	.dc.b	$F0,$58,$F2,$BF,$F4,$FE,$1C,$AB,$53,$9C,$39,$78,$9E,$BC,$BC,$18
	.dc.b	$F2,$C7,$C0,$31,$F1,$BF,$E9,$76,$BC,$30,$FF,$4B,$46,$23,$C3,$F4
	.dc.b	$9F,$E7,$87,$F1,$70,$8C,$47,$36,$BF,$05,$AF,$2F,$82,$F6,$3E,$38
	.dc.b	$E9,$9F,$CA,$6E,$3F,$EF,$5F,$8A,$7D,$A7,$DA,$AD,$61,$CB,$1F,$13
	.dc.b	$70,$E3,$8F,$CE,$C3,$8B,$F2,$F8,$BE,$F7,$D5,$F3,$CF,$EF,$36,$7F
	.dc.b	$59,$C6,$8C,$47,$C9,$F9,$6F,$EF,$FC,$7F,$75,$B1,$C2,$79,$61,$DA
	.dc.b	$3C,$D3,$A7,$91,$F9,$B9,$DD,$F7,$8D,$9F,$E6,$2B,$7C,$B7,$E7,$FA
	.dc.b	$7F,$CF,$2D,$8F,$F7,$19,$AC,$BC,$7C,$2F,$CE,$3F,$DE,$98,$73,$BE
	.dc.b	$0F,$F8,$D5,$B7,$C6,$3F,$CE,$35,$C7,$DB,$E7,$7C,$A7,$FD,$BD,$39
	.dc.b	$7F,$86,$B5,$CC,$6D,$FD,$8C,$65,$B1,$97,$FD,$20,$D7,$F4,$2D,$F0
	.dc.b	$7F,$CF,$78,$7C,$E6,$CF,$F9,$E2,$3E,$8E,$CF,$C5,$DE,$3E,$1B,$5F
	.dc.b	$C8,$A7,$53,$FE,$F3,$6C,$66,$76,$FF,$BE,$E3,$2D,$8C,$BF,$C3,$23
	.dc.b	$2D,$8C,$BF,$D0,$FC,$47,$C7,$FA,$5B,$1F,$E9,$08,$F4,$FE,$2C,$F3
	.dc.b	$69,$E5,$71,$3E,$96,$BE,$56,$BF,$E0,$BF,$37,$F6,$7C,$5F,$C8,$B5
	.dc.b	$46,$38,$7F,$AB,$36,$BF,$A5,$7E,$0E,$FB,$3F,$9D,$39,$BE,$3D,$58
	.dc.b	$3E,$7D,$6E,$FF,$C1,$11,$96,$9F,$D0,$E3,$4B,$1F,$E6,$BB,$A7,$33
	.dc.b	$F2,$DA,$FF,$60,$6B,$E6,$C7,$FA,$DA,$C3,$FD,$7F,$F7,$9F,$34,$91
	.dc.b	$FD,$4F,$E3,$07,$E1,$8A,$FE,$0D,$AF,$DF,$6B,$F7,$5A,$FF,$91,$B4
	.dc.b	$FC,$7F,$A7,$FD,$3A,$3F,$C9,$D1,$FB,$DF,$D5,$5F,$79,$A8,$93,$FE
	.dc.b	$50,$75,$12,$3F,$DB,$E2,$78,$B8,$6B,$F2,$67,$FA,$DF,$E6,$2F,$9D
	.dc.b	$FE,$18,$D9,$F8,$AD,$7F,$C0,$2D,$7F,$B1,$87,$0F,$E5,$FE,$FF,$F2
	.dc.b	$C4,$7C,$AF,$F1,$27,$FD,$CE,$E5,$FF,$73,$BF,$C7,$4D,$3F,$EE,$AF
	.dc.b	$99,$FD,$01,$1F,$7F,$DB,$BE,$5F,$11,$B3,$FD,$E7,$3F,$C9,$E7,$50
	.dc.b	$3E,$11,$F8,$BF,$C7,$3F,$9E,$B8,$E4,$46,$1F,$F0,$19,$8F,$27,$A7
	.dc.b	$5C,$AF,$C5,$DA,$3F,$1B,$08,$C3,$A8,$CC,$62,$3F,$C5,$AD,$7E,$6B
	.dc.b	$5F,$F0,$8B,$5F,$CC,$47,$DC,$AF,$DB,$37,$5F,$E1,$1B,$FF,$47,$FC
	.dc.b	$59,$F6,$DF,$C2,$62,$61,$F0,$F0,$E3,$1B,$9F,$E1,$DF,$1D,$2B,$58
	.dc.b	$7F,$13,$8F,$EA,$5F,$71,$FC,$4A,$7E,$5F,$3F,$C1,$9E,$11,$8C,$D1
	.dc.b	$C3,$BF,$FA,$FF,$C6,$BF,$23,$7A,$38,$6E,$1E,$01,$3B,$9F,$EC,$5F
	.dc.b	$73,$F5,$74,$FF,$1A,$BF,$6A,$BF,$2A,$D7,$3C,$5F,$52,$31,$1F,$83
	.dc.b	$6A,$69,$23,$F8,$3E,$9F,$77,$A3,$F9,$91,$2B,$F6,$35,$FC,$7F,$5F
	.dc.b	$D4,$70,$FD,$4A,$D5,$7E,$40,$ED,$E0,$7F,$3B,$F7,$A4,$7B,$F1,$FC
	.dc.b	$54,$7F,$7F,$FB,$8F,$B3,$39,$C7,$FD,$2C,$D9,$E2,$F5,$9D,$BF,$47
	.dc.b	$D9,$FE,$90,$47,$F8,$3F,$E2,$CF,$7F,$FC,$18,$D8,$CC,$60,$FF,$08
	.dc.b	$7E,$2C,$F9,$BC,$4E,$88,$7B,$63,$4F,$9F,$E9,$BF,$AC,$11,$FE,$2F
	.dc.b	$F1,$EF,$83,$C4,$7E,$1E,$6F,$0F,$F2,$5A,$D4,$FC,$1F,$4B,$FB,$11
	.dc.b	$1E,$7F,$0B,$5F,$C8,$45,$AE,$53,$FD,$FB,$E6,$1F,$0D,$49,$7F,$CA
	.dc.b	$BE,$C5,$F5,$9F,$E2,$F6,$C7,$F4,$86,$BF,$EA,$46,$BF,$C9,$1A,$FF
	.dc.b	$A2,$1A,$FF,$11,$6A,$69,$8F,$39,$F6,$5F,$E3,$F6,$CF,$F3,$E6,$BF
	.dc.b	$8B,$FC,$FB,$FD,$46,$FF,$42,$FC,$DD,$F7,$5F,$C4,$9B,$1F,$D8,$1A
	.dc.b	$FF,$AF,$1A,$FF,$3C,$FC,$DF,$FD,$5C,$FF,$4C,$6B,$F9,$96,$ED,$FD
	.dc.b	$67,$F3,$7F,$F0,$FB,$FD,$D3,$E4,$3F,$8E,$37,$B3,$FC,$5B,$18,$F0
	.dc.b	$BF,$D5,$7E,$93,$D5,$FE,$C0,$7F,$28,$71,$97,$F7,$66,$BF,$ED,$26
	.dc.b	$BF,$D4,$9A,$F4,$8E,$44,$7F,$DB,$3E,$03,$EE,$9F,$39,$FE,$43,$6C
	.dc.b	$FE,$C5,$AF,$ED,$FE,$99,$F2,$1F,$E7,$16,$C7,$F8,$29,$AF,$FB,$99
	.dc.b	$AF,$F6,$66,$BF,$EC,$A6,$BF,$D4,$DA,$9A,$6A,$FD,$1F,$03,$E7,$24
	.dc.b	$FF,$B9,$FD,$47,$D2,$7F,$96,$9B,$1F,$E5,$76,$BF,$E5,$36,$BF,$E5
	.dc.b	$16,$BF,$D9,$7D,$DF,$EC,$33,$FE,$5D,$9C,$CA,$72,$F5,$9F,$E4,$B6
	.dc.b	$F8,$8F,$F3,$73,$67,$F8,$E3,$5F,$E4,$5F,$94,$FF,$0F,$6F,$8D,$FD
	.dc.b	$A2,$31,$1F,$E0,$E6,$BF,$EF,$26,$BF,$DE,$DA,$9A,$48,$FF,$1F,$FC
	.dc.b	$7B,$F6,$77,$CE,$7F,$AC,$23,$0D,$8C,$DF,$CD,$F9,$1F,$98,$7C,$32
	.dc.b	$3F,$35,$AF,$CD,$6B,$F3,$5F,$F4,$37,$86,$F9,$4F,$E2,$44,$9F,$F1
	.dc.b	$CE,$5F,$D2,$39,$FD,$DE,$8F,$C3,$24,$FF,$96,$27,$FC,$5E,$FC,$18
	.dc.b	$CB,$FE,$34,$3E,$13,$87,$FD,$1D,$FA,$13,$E0,$7F,$D3,$ED,$8F,$ED
	.dc.b	$DC,$48,$C4,$7F,$88,$FF,$4D,$FE,$FD,$DD,$FE,$22,$7F,$C0,$11,$FE
	.dc.b	$2B,$3F,$CE,$8E,$1F,$F0,$63,$5C,$FB,$5F,$EA,$7F,$17,$FD,$23,$5F
	.dc.b	$E7,$A6,$E9,$F9,$16,$D7,$F8,$DF,$B9,$F4,$FF,$EB,$69,$CE,$7F,$E8
	.dc.b	$FF,$8A,$C1,$FE,$69,$9F,$F5,$F5,$FF,$9D,$31,$FF,$4D,$37,$1F,$F2
	.dc.b	$17,$3F,$F9,$0F,$E1,$FF,$D8,$3F,$8B,$3E,$17,$8C,$4B,$FE,$E8,$71
	.dc.b	$3F,$45,$3B,$7E,$4F,$D6,$7F,$74,$7F,$C9,$BA,$23,$19,$7F,$A6,$1B
	.dc.b	$97,$EB,$FE,$EF,$FC,$71,$7F,$AD,$9C,$B7,$DB,$FF,$6A,$C6,$30,$57
	.dc.b	$06,$FC,$2F,$FB,$4E,$3F,$CF,$F7,$FE,$CD,$6B,$DE,$3F,$CA,$BF,$49
	.dc.b	$FD,$47,$2F,$F3,$27,$DE,$7F,$8D,$AB,$FC,$CD,$3F,$D3,$A3,$10,$8C
	.dc.b	$7C,$2F,$E0,$5E,$50,$9D,$C6,$1B,$E9,$7E,$1D,$B3,$E0,$AE,$5E,$9F
	.dc.b	$FA,$F6,$39,$46,$30,$FF,$5B,$E1,$DA,$7F,$D2,$2D,$74,$7E,$85,$7D
	.dc.b	$EB,$E9,$B8,$B6,$74,$D7,$9B,$5F,$D5,$7C,$E7,$F9,$E9,$34,$8F,$F9
	.dc.b	$F9,$FF,$46,$7B,$3F,$1A,$1E,$D3,$8F,$39,$F4,$95,$6E,$B5,$18,$9D
	.dc.b	$4E,$6B,$8C,$FF,$AC,$2E,$BF,$E9,$BB,$E3,$5E,$18,$F3,$9C,$CF,$DB
	.dc.b	$46,$AF,$A9,$FE,$BA,$71,$1F,$CF,$78,$3F,$45,$87,$F0,$BE,$1F,$C7
	.dc.b	$F3,$7F,$B5,$EB,$F3,$0D,$78,$36,$37,$3C,$6B,$9D,$7F,$B2,$27,$2D
	.dc.b	$8C,$BF,$C2,$E1,$F7,$AE,$BE,$8D,$BF,$23,$6B,$A3,$5E,$4D,$7F,$AE
	.dc.b	$35,$FE,$B4,$DA,$EF,$5E,$B9,$7D,$A7,$F9,$0A,$FF,$96,$46,$3D,$BD
	.dc.b	$4E,$63,$F9,$83,$5F,$F7,$2B,$5F,$E4,$CD,$7F,$DA,$4D,$7A,$4F,$68
	.dc.b	$FF,$49,$37,$E1,$BB,$91,$FE,$7D,$C6,$FB,$B7,$A3,$F4,$E3,$0D,$8C
	.dc.b	$CF,$06,$E1,$FD,$1F,$EA,$BF,$C6,$53,$EE,$6C,$67,$D7,$F8,$91,$88
	.dc.b	$FF,$80,$2D,$7F,$A1,$35,$FF,$77,$35,$39,$C3,$D1,$E1,$79,$49,$7E
	.dc.b	$57,$E9,$7C,$47,$FB,$12,$3E,$8B,$6F,$FA,$F3,$5F,$EA,$41,$C3,$FD
	.dc.b	$55,$AF,$FC,$01,$6B,$BF,$8A,$EE,$7E,$67,$EE,$3F,$C8,$B5,$FA,$16
	.dc.b	$CF,$E6,$5A,$FE,$5D,$AF,$7F,$B8,$F4,$92,$7D,$2D,$8F,$4B,$5F,$4D
	.dc.b	$A3,$F5,$92,$FB,$97,$C3,$61,$B1,$FD,$FC,$D4,$64,$FC,$D7,$DF,$7E
	.dc.b	$7F,$03,$49,$C9,$A2,$57,$E7,$A3,$2F,$C1,$F2,$7E,$21,$1C,$1F,$A4
	.dc.b	$3F,$97,$97,$BB,$F5,$4E,$2F,$CD,$18,$24,$6C,$91,$F6,$9F,$EE,$5F
	.dc.b	$39,$F0,$BA,$CD,$24,$7E,$D3,$BF,$FC,$A0,$DF,$67,$F7,$51,$8B,$FE
	.dc.b	$77,$18,$24,$EC,$91,$FB,$AD,$BF,$7A,$48,$FD,$A3,$FC,$03,$E1,$E4
	.dc.b	$7E,$DD,$B3,$FE,$31,$FB,$97,$DD,$7F,$0B,$6D,$7F,$0A,$6C,$EE,$7F
	.dc.b	$88,$7D,$77,$E6,$6B,$F8,$7B,$67,$FC,$78,$D7,$F3,$ED,$7F,$C6,$5F
	.dc.b	$51,$F1,$A7,$F8,$BB,$63,$FC,$88,$D7,$F4,$ED,$78,$E9,$FC,$86,$31
	.dc.b	$00,$E1,$B1,$FD,$07,$DD,$7D,$CB,$D1,$C3,$70,$FF,$2A,$4E,$A7,$27
	.dc.b	$E7,$BE,$E7,$FD,$41,$3F,$9B,$6E,$1F,$83,$6B,$FE,$4F,$84,$62,$3C
	.dc.b	$D1,$8F,$83,$E8,$47,$A3,$DF,$E1,$E4,$77,$6C,$F7,$6B,$DE,$91,$88
	.dc.b	$FF,$30,$7C,$47,$F0,$A2,$76,$FF,$49,$6D,$FF,$48,$6C,$FF,$9D,$5F
	.dc.b	$E2,$7E,$53,$EE,$1F,$5D,$FD,$AE,$30,$D8,$CD,$FF,$9F,$67,$2F,$F9
	.dc.b	$FF,$ED,$7F,$C0,$4D,$C3,$FC,$22,$D9,$FE,$ED,$3A,$9F,$E4,$ED,$7F
	.dc.b	$CF,$4D,$7F,$00,$D4,$D3,$F0,$9E,$77,$48,$34,$95,$E3,$6C,$78,$0D
	.dc.b	$C6,$4F,$ED,$BF,$72,$F9,$0F,$EE,$4D,$8C,$C6,$08,$E0,$93,$FE,$A1
	.dc.b	$9E,$0F,$F5,$6F,$B9,$7D,$4F,$F8,$5F,$C9,$46,$2F,$FB,$13,$6B,$85
	.dc.b	$7B,$7E,$26,$57,$CD,$6C,$FA,$DA,$FC,$A3,$FD,$9B,$1B,$3E,$B7,$AB
	.dc.b	$FE,$F6,$6F,$C5,$FD,$48,$C4,$FD,$46,$BF,$51,$AF,$D4,$6B,$F5,$1B
	.dc.b	$AF,$F9,$7E,$7F,$C3,$5F,$84,$3E,$E6,$93,$0F,$F0,$DC,$65,$B1,$9C
	.dc.b	$3F,$D9,$53,$96,$C7,$F8,$37,$E5,$9F,$73,$49,$7F,$E1,$48,$CB,$F0
	.dc.b	$3E,$47,$DC,$47,$04,$BF,$81,$19,$6C,$7C,$8F,$B9,$F8,$C4,$70,$48
	.dc.b	$F9,$11,$96,$C7,$F8,$E3,$EE,$5D,$1F,$E3,$D6,$C6,$63,$06,$7E,$BE
	.dc.b	$47,$16,$D7,$F3,$27,$F9,$B9,$23,$F9,$BB,$63,$51,$82,$3F,$EF,$5A
	.dc.b	$7E,$17,$E3,$24,$F9,$DF,$F3,$57,$DD,$3E,$4E,$92,$7F,$CE,$91,$97
	.dc.b	$DD,$E2,$7B,$48,$E0,$FF,$3D,$7D,$D7,$F0,$4C,$7F,$CE,$4D,$7F,$CE
	.dc.b	$0D,$7F,$A0,$7B,$E8,$C4,$23,$1D,$BF,$E7,$F9,$CC,$62,$BB,$CE,$48
	.dc.b	$E0,$95,$FE,$38,$1F,$45,$FE,$47,$27,$B7,$FE,$73,$7F,$C9,$97,$FE
	.dc.b	$02,$F8,$3F,$F4,$76,$5F,$EA,$AC,$3F,$D2,$CD,$C3,$FD,$1C,$D7,$FB
	.dc.b	$3F,$DE,$FF,$6E,$24,$FF,$6E,$6C,$7F,$6E,$6B,$C2,$3F,$D4,$5E,$7B
	.dc.b	$E2,$FF,$D3,$CD,$8C,$C6,$0F,$47,$DD,$C8,$FF,$44,$E6,$8C,$5F,$C1
	.dc.b	$8C,$4E,$A7,$33,$FE,$5C,$9C,$BF,$EA,$9F,$BB,$7D,$17,$FA,$BD,$B7
	.dc.b	$FE,$B7,$6C,$EE,$74,$49,$D3,$63,$C5,$F7,$7F,$3C,$93,$E8,$6C,$6A
	.dc.b	$30,$47,$03,$E0,$BF,$DB,$7B,$BE,$DD,$F3,$5F,$ED,$EF,$69,$D5,$FE
	.dc.b	$0D,$C3,$F0,$39,$7F,$B8,$B1,$E3,$87,$0C,$BF,$C2,$58,$70,$BF,$F0
	.dc.b	$A7,$DD,$FE,$97,$9B,$39,$73,$8C,$4F,$FB,$E2,$3F,$C3,$5E,$A7,$F8
	.dc.b	$6B,$D0,$F1,$46,$FA,$FF,$DE,$F5,$B3,$AB,$8C,$7C,$10,$C3,$AC,$78
	.dc.b	$F0,$E9,$97,$8A,$B8,$E5,$5E,$18,$76,$7F,$D2,$F8,$79,$9B,$19,$7C
	.dc.b	$93,$FE,$D0,$97,$DD,$FF,$A7,$67,$F5,$11,$8F,$84,$D4,$E6,$7A,$B5
	.dc.b	$F8,$8F,$1C,$BD,$75,$C6,$7E,$01,$ED,$9D,$67,$F2,$3E,$ED,$F5,$9F
	.dc.b	$5F,$0E,$35,$AB,$FA,$6D,$79,$63,$F0,$F5,$61,$CA,$7E,$26,$4F,$7F
	.dc.b	$F0,$5C,$3F,$80,$FB,$BE,$DA,$FF,$12,$E2,$46,$30,$FC,$0D,$F0,$E1
	.dc.b	$FC,$03,$FE,$F6,$8F,$AC,$DC,$7B,$47,$CB,$8C,$11,$C1,$0F,$F6,$5F
	.dc.b	$03,$EE,$FF,$C1,$31,$FE,$31,$6B,$A6,$BF,$9B,$8E,$1E,$87,$F1,$D8
	.dc.b	$46,$1E,$51,$98,$C4,$73,$8F,$47,$E1,$32,$3F,$04,$D7,$F0,$2D,$7F
	.dc.b	$16,$D7,$F9,$CF,$DE,$7F,$AA,$1F,$E5,$DD,$28,$C4,$7F,$BA,$23,$1E
	.dc.b	$0F,$9D,$B1,$E9,$F9,$CF,$98,$9A,$47,$E8,$35,$E0,$60,$E9,$F7,$8F
	.dc.b	$8E,$A4,$8F,$2C,$65,$FD,$07,$C3,$7D,$92,$38,$3D,$7F,$76,$F8,$CF
	.dc.b	$43,$63,$31,$82,$38,$21,$F0,$63,$83,$63,$6D,$36,$7F,$17,$1F,$9C
	.dc.b	$3E,$2F,$F2,$0D,$8F,$C0,$35,$FE,$BC,$D7,$F9,$06,$7F,$EA,$7F,$85
	.dc.b	$FA,$07,$FA,$A6,$3E,$BB,$5F,$AC,$D7,$C5,$2F,$C3,$3C,$24,$8F,$0B
	.dc.b	$F5,$3E,$76,$47,$DB,$6B,$FD,$C7,$64,$62,$3D,$D1,$87,$81,$F9,$BF
	.dc.b	$BC,$7D,$37,$EB,$DB,$19,$8C,$19,$F8,$49,$1C,$5A,$FF,$80,$7E,$99
	.dc.b	$F2,$FF,$B9,$6C,$7E,$A5,$AE,$E3,$44,$74,$D7,$FC,$19,$F7,$6F,$BC
	.dc.b	$FE,$08,$D8,$FD,$A3,$5D,$C6,$88,$E9,$AF,$4F,$6B,$99,$1E,$6F,$F1
	.dc.b	$1F,$C8,$3F,$8B,$C7,$F0,$16,$BF,$E1,$F6,$BF,$8B,$6B,$FD,$8D,$AF
	.dc.b	$FB,$C5,$A9,$A6,$3E,$F5,$F7,$BF,$C8,$5B,$3F,$C4,$5A,$FF,$8B,$DA
	.dc.b	$FE,$6D,$AF,$F7,$36,$BF,$F0,$05,$A9,$BF,$7A,$FA,$AF,$E6,$0D,$9F
	.dc.b	$E4,$0D,$7F,$C7,$CD,$7F,$54,$D7,$FC,$04,$D4,$1C,$36,$3F,$9F,$7C
	.dc.b	$FF,$F4,$B8,$FF,$25,$35,$FD,$B3,$5F,$F0,$73,$5E,$BE,$87,$E0,$5F
	.dc.b	$DE,$35,$FF,$0A,$35,$DE,$7F,$D3,$63,$11,$FE,$75,$8C,$11,$FE,$81
	.dc.b	$3F,$DC,$5F,$0F,$9B,$E8,$23,$A6,$BE,$A6,$BF,$C3,$FE,$EF,$E8,$93
	.dc.b	$3F,$EB,$CD,$8F,$E9,$CD,$7F,$CD,$8F,$D8,$FB,$B7,$D9,$7F,$65,$6C
	.dc.b	$FF,$55,$6B,$FE,$73,$6B,$FC,$7F,$EE,$9F,$77,$FD,$B5,$B1,$FD,$75
	.dc.b	$AF,$F9,$ED,$AF,$18,$E0,$47,$83,$FD,$DF,$E2,$3F,$CB,$71,$FD,$9D
	.dc.b	$AF,$FA,$15,$AF,$F3,$16,$BF,$E4,$96,$BF,$88,$6A,$69,$3E,$07,$C0
	.dc.b	$48,$F0,$35,$F0,$35,$F0,$3F,$E0,$DF,$AA,$7D,$3D,$24,$FF,$7F,$8C
	.dc.b	$BF,$E0,$1F,$DA,$3F,$A0,$3F,$E5,$F6,$BF,$9E,$6A,$69,$3C,$3F,$F0
	.dc.b	$E3,$63,$FC,$19,$F7,$8F,$77,$FC,$40,$D8,$CC,$60,$FF,$12,$7D,$DF
	.dc.b	$FA,$BA,$3F,$C2,$8D,$7F,$D6,$4D,$7F,$AF,$35,$FF,$3B,$7D,$DF,$F8
	.dc.b	$21,$FE,$C4,$D7,$3E,$6F,$F8,$C6,$31,$1F,$BA,$F0,$DF,$11,$FE,$06
	.dc.b	$24,$7F,$83,$5F,$2E,$12,$3C,$AD,$7C,$AD,$7C,$AD,$7F,$D1,$5F,$3C
	.dc.b	$F8,$FF,$F2,$13,$63,$FC,$64,$D7,$FD,$A0,$D7,$FB,$E7,$E7,$7E,$84
	.dc.b	$D2,$65,$FE,$4D,$6C,$7F,$8F,$5A,$FF,$8A,$FE,$E9,$F2,$BF,$E2,$B6
	.dc.b	$C7,$F8,$C9,$AE,$E3,$FC,$B1,$FB,$4F,$EC,$23,$FD,$48,$D7,$F9,$1B
	.dc.b	$53,$49,$D3,$C8,$91,$FE,$51,$FB,$9E,$04,$8F,$F3,$33,$CF,$E8,$FE
	.dc.b	$AC,$FF,$95,$DA,$FF,$BD,$5A,$FF,$88,$5A,$FF,$AE,$5A,$FF,$38,$6A
	.dc.b	$69,$AF,$45,$8B,$C2,$F8,$C4,$AF,$8C,$FF,$9E,$FF,$39,$FE,$17,$1F
	.dc.b	$EC,$76,$BF,$D1,$9A,$9A,$4E,$CE,$44,$8F,$89,$E6,$FC,$22,$38,$3F
	.dc.b	$CF,$12,$F5,$7C,$44,$8C,$B5,$FF,$22,$1E,$17,$0F,$F9,$1E,$39,$3A
	.dc.b	$F8,$37,$DC,$FF,$5F,$6F,$91,$FE,$87,$6C,$EE,$7F,$B0,$F4,$3E,$67
	.dc.b	$FC,$DE,$49,$FF,$3B,$3F,$E7,$2F,$AB,$91,$FE,$79,$6B,$FE,$3F,$6B
	.dc.b	$FE,$3E,$6B,$FD,$C3,$77,$DC,$3E,$B3,$FD,$08,$D9,$FF,$45,$B5,$FF
	.dc.b	$45,$35,$FF,$2E,$B5,$E8,$D5,$3D,$A7,$A9,$FE,$91,$FB,$4F,$F2,$84
	.dc.b	$FF,$A5,$5A,$FF,$9A,$1A,$F9,$DA,$F4,$6C,$9A,$4F,$5F,$FD,$8D,$F6
	.dc.b	$AF,$15,$B7,$F8,$43,$C7,$7A,$FF,$E1,$2C,$96,$FB,$7C,$56,$A7,$FC
	.dc.b	$2F,$39,$79,$57,$18,$ED,$78,$9D,$E1,$E0,$57,$CD,$EF,$87,$6C,$39
	.dc.b	$36,$38,$DF,$37,$CB,$7C,$EF,$96,$1E,$39,$EB,$87,$27,$84,$79,$1C
	.dc.b	$9D,$0F,$CD,$B9,$7F,$D4,$0D,$7D,$95,$E9,$73,$E4,$3E,$8B,$FD,$C1
	.dc.b	$5F,$EA,$56,$C6,$67,$AB,$5F,$F7,$37,$C8,$7F,$B3,$88,$FF,$BE,$DB
	.dc.b	$8F,$C5,$8F,$F7,$5F,$C5,$7C,$17,$FC,$8F,$37,$E1,$D3,$99,$4E,$63
	.dc.b	$0D,$8F,$82,$7D,$F7,$15,$DE,$38,$9A,$3C,$27,$C8,$E0,$FC,$77,$0F
	.dc.b	$82,$3F,$30,$D7,$F3,$0D,$7F,$30,$D7,$F3,$0D,$8E,$77,$FE,$F8,$E1
	.dc.b	$FD,$93,$67,$5F,$6D,$FE,$E1,$8F,$F7,$EB,$6F,$FD,$F8,$D7,$FD,$9E
	.dc.b	$D7,$F7,$CD,$7F,$66,$DB,$34,$D4,$3E,$DB,$FC,$43,$5F,$C8,$23,$1E
	.dc.b	$7F,$AD,$B1,$99,$DC,$7F,$26,$6B,$FE,$EB,$6B,$FA,$F6,$BF,$E8,$26
	.dc.b	$BE,$1F,$5B,$B4,$1A,$49,$FE,$5D,$F6,$DF,$EC,$C6,$FB,$DF,$3F,$95
	.dc.b	$39,$C3,$FD,$F9,$18,$9F,$57,$D4,$7F,$86,$2F,$DC,$DA,$FE,$7E,$D7
	.dc.b	$F8,$30,$70,$FF,$09,$6B,$FE,$9C,$6B,$BF,$3F,$B9,$E5,$C1,$F1,$14
	.dc.b	$97,$EC,$7E,$77,$B4,$F9,$6A,$49,$FA,$91,$96,$C6,$70,$FE,$AF,$39
	.dc.b	$6C,$7E,$47,$ED,$9F,$7D,$49,$7F,$D7,$B8,$93,$A8,$EF,$E3,$79,$08
	.dc.b	$E0,$98,$78,$23,$2D,$8F,$ED,$23,$ED,$DF,$6F,$FA,$58,$C3,$63,$31
	.dc.b	$C1,$B3,$FD,$CB,$ED,$FF,$2E,$4F,$8C,$FC,$A3,$6B,$F3,$0D,$9E,$13
	.dc.b	$F0,$7C,$59,$5E,$FC,$91,$8A,$F3,$F7,$7E,$7E,$8D,$23,$FA,$53,$25
	.dc.b	$73,$1C,$A3,$E0,$F5,$3D,$5F,$80,$D8,$CC,$60,$FD,$77,$73,$E4,$E9
	.dc.b	$27,$F9,$5B,$6F,$FC,$2B,$3B,$9C,$92,$3F,$69,$BF,$EB,$09,$1F,$B9
	.dc.b	$7F,$5F,$F4,$52,$3F,$78,$D9,$FF,$10,$FD,$BB,$ED,$FF,$81,$36,$BF
	.dc.b	$83,$36,$77,$3F,$C1,$FE,$DD,$CF,$49,$7F,$C3,$63,$2F,$F0,$FF,$B7
	.dc.b	$F8,$97,$FE,$30,$6B,$FB,$26,$BF,$DE,$5A,$83,$86,$EB,$FC,$66,$30
	.dc.b	$70,$F9,$EF,$F2,$D7,$C9,$FF,$37,$6D,$FF,$90,$1B,$7F,$C0,$9A,$FF
	.dc.b	$82,$61,$F6,$EF,$7F,$F3,$8D,$8F,$D0,$B6,$7F,$85,$C7,$09,$D9,$1D
	.dc.b	$B6,$37,$2F,$B7,$EA,$4A,$EA,$D9,$DB,$5F,$34,$F1,$9F,$E7,$9F,$21
	.dc.b	$FC,$16,$7F,$8E,$B5,$FF,$12,$35,$F9,$8D,$4D,$27,$77,$F4,$57,$87
	.dc.b	$96,$FB,$67,$DB,$FF,$63,$8C,$36,$33,$87,$F9,$BA,$72,$FF,$9C,$3E
	.dc.b	$CF,$EE,$47,$F6,$B6,$C6,$71,$FE,$EE,$7F,$8E,$E3,$2F,$F3,$86,$B9
	.dc.b	$36,$7E,$1D,$CB,$4C,$87,$F1,$9E,$97,$86,$3F,$C9,$33,$E1,$8C,$11
	.dc.b	$C1,$23,$C4,$70,$3F,$D1,$7F,$6C,$FB,$7F,$ED,$8D,$9F,$EE,$0D,$9D
	.dc.b	$CE,$89,$3A,$6C,$7F,$56,$F2,$5F,$6E,$FA,$0F,$F0,$94,$61,$B1,$9B
	.dc.b	$FE,$BE,$DA,$F3,$FD,$BF,$BC,$93,$F3,$DB,$1A,$8C,$11,$FA,$1E,$AF
	.dc.b	$FB,$9A,$3E,$D3,$63,$3F,$11,$F6,$E3,$11,$F6,$5A,$FD,$56,$BF,$4D
	.dc.b	$BA,$FF,$94,$E7,$FC,$29,$F6,$FF,$EA,$9B,$FF,$0C,$36,$FF,$D7,$2D
	.dc.b	$7F,$BD,$B5,$FF,$41,$35,$FE,$14,$D4,$D3,$5E,$8B,$E4,$5F,$6F,$FE
	.dc.b	$11,$6D,$FF,$B0,$DB,$7F,$E1,$26,$BE,$FF,$8D,$F8,$44,$9F,$8E,$D8
	.dc.b	$F9,$6D,$78,$47,$CC,$FB,$7F,$94,$47,$04,$8F,$A7,$19,$6C,$7F,$8F
	.dc.b	$7E,$DF,$F7,$6D,$E8,$FF,$1F,$46,$27,$FD,$C2,$D7,$FC,$46,$BE,$AB
	.dc.b	$64,$AE,$8F,$F4,$12,$3F,$CF,$DB,$3F,$EE,$CF,$0B,$94,$E4,$91,$99
	.dc.b	$F1,$FD,$BC,$9F,$43,$67,$FC,$6B,$E9,$BE,$DD,$F1,$9F,$E6,$F6,$DF
	.dc.b	$F8,$75,$B5,$C1,$AF,$B3,$C6,$F8,$24,$9F,$8C,$FF,$9E,$7E,$DF,$F8
	.dc.b	$2E,$06,$93,$BB,$44,$AF,$F1,$64,$65,$B1,$9B,$F4,$CE,$5F,$F2,$78
	.dc.b	$FB,$4F,$F2,$59,$3D,$BF,$F3,$93,$FE,$50,$C3,$FC,$19,$F0,$5F,$E8
	.dc.b	$DC,$BF,$D9,$F9,$7F,$8E,$DB,$87,$C6,$6B,$FE,$5C,$DD,$F6,$FF,$DD
	.dc.b	$C9,$5F,$DD,$5B,$3F,$5D,$AF,$17,$FD,$39,$E7,$BE,$DE,$92,$BF,$C9
	.dc.b	$91,$97,$FD,$19,$9B,$ED,$FD,$64,$8C,$B7,$0F,$87,$3B,$9C,$92,$3F
	.dc.b	$03,$F6,$EF,$B7,$FF,$57,$36,$7F,$CB,$8D,$9D,$CE,$89,$3A,$6C,$78
	.dc.b	$3E,$6B,$E7,$92,$7D,$2D,$8D,$46,$08,$E0,$76,$FF,$B2,$78,$1F,$66
	.dc.b	$FB,$7F,$F6,$67,$B1,$FA,$BB,$FF,$0D,$68,$CB,$FD,$A1,$8F,$1C,$38
	.dc.b	$65,$FE,$E4,$EA,$FF,$10,$60,$FB,$3F,$D0,$F4,$66,$3F,$C4,$91,$87
	.dc.b	$FD,$CF,$3F,$E2,$7F,$4F,$FC,$4F,$E8,$78,$A3,$9D,$6F,$5F,$25,$F5
	.dc.b	$EC,$63,$D6,$30,$F7,$C7,$9E,$59,$C3,$96,$3D,$30,$E3,$87,$E7,$A7
	.dc.b	$C8,$FF,$AB,$9A,$EF,$0F,$F6,$94,$3E,$DB,$FC,$97,$3C,$5A,$F1,$6B
	.dc.b	$EA,$6B,$F0,$9E,$39,$7B,$27,$A7,$AF,$E7,$9C,$E3,$F1,$7E,$D9,$F6
	.dc.b	$FF,$3B,$0E,$15,$AC,$7A,$61,$C4,$F8,$57,$C2,$3E,$17,$BB,$F5,$B8
	.dc.b	$7E,$D7,$DB,$F3,$CF,$FC,$B1,$CD,$F8,$48,$C5,$FD,$8B,$E1,$C1,$F5
	.dc.b	$1F,$F7,$A4,$7D,$F6,$ED,$FE,$87,$8F,$8F,$18,$23,$82,$1F,$ED,$CD
	.dc.b	$DF,$6F,$FE,$07,$8F,$F1,$93,$5D,$35,$FC,$7C,$70,$F4,$7F,$8B,$C7
	.dc.b	$E1,$27,$4F,$F8,$E1,$AE,$63,$73,$C6,$3E,$EF,$DB,$FA,$88,$E0,$87
	.dc.b	$E1,$DC,$B4,$FF,$6C,$7D,$BF,$FA,$B5,$BE,$17,$93,$54,$E6,$7F,$A3
	.dc.b	$B5,$FF,$7E,$FC,$CF,$CC,$4D,$23,$F4,$1A,$FC,$06,$BF,$D4,$BE,$DF
	.dc.b	$FD,$FA,$FF,$BA,$DA,$E7,$E3,$7C,$91,$88,$FD,$07,$C2,$7D,$A2,$3F
	.dc.b	$A0,$7D,$5F,$6E,$F8,$AA,$49,$FF,$7C,$46,$5B,$19,$7F,$B2,$C6,$5B
	.dc.b	$19,$7F,$84,$E4,$9D,$46,$1A,$FE,$59,$AF,$FA,$BF,$ED,$FF,$DC,$AF
	.dc.b	$E0,$9A,$FF,$5A,$6B,$FC,$58,$38,$7F,$43,$F0,$83,$EC,$15,$FE,$C7
	.dc.b	$59,$8C,$9F,$A2,$97,$DB,$D2,$57,$8D,$FB,$1F,$37,$23,$EC,$35,$FE
	.dc.b	$DF,$AA,$31,$1F,$84,$6B,$FC,$BB,$ED,$DF,$45,$FB,$26,$C7,$E8,$9A
	.dc.b	$FF,$7A,$6B,$F0,$57,$C1,$F1,$24,$71,$7F,$7B,$F6,$EF,$96,$A4,$9F
	.dc.b	$D7,$C6,$5B,$19,$7F,$C1,$31,$96,$C7,$F0,$8F,$B7,$7D,$A5,$24,$FE
	.dc.b	$EA,$32,$D8,$CB,$FE,$14,$8C,$B6,$3A,$7B,$52,$39,$35,$FF,$0D,$FD
	.dc.b	$BB,$C0,$FE,$32,$D8,$FE,$12,$D7,$71,$A2,$3A,$6B,$FE,$26,$FB,$77
	.dc.b	$DB,$7F,$24,$6C,$7F,$14,$6B,$B8,$D1,$1D,$35,$FF,$18,$FD,$BB,$EA
	.dc.b	$3F,$97,$B6,$3F,$8F,$B5,$DC,$68,$8E,$9A,$FF,$8F,$3E,$7B,$FA,$23
	.dc.b	$7C,$CF,$E7,$71,$88,$FD,$33,$5F,$EF,$BE,$6F,$D0,$7F,$54,$D7,$3A
	.dc.b	$FF,$44,$8C,$40,$FF,$31,$C6,$48,$FF,$38,$8D,$A9,$FE,$CE,$F1,$6B
	.dc.b	$9F,$2F,$CE,$47,$FA,$0C,$68,$91,$FD,$5B,$ED,$FE,$8C,$FF,$01,$6B
	.dc.b	$FE,$18,$6B,$E7,$6A,$69,$33,$E8,$7F,$62,$FB,$77,$D8,$52,$57,$F5
	.dc.b	$48,$CB,$FD,$5B,$ED,$DF,$5D,$FD,$A9,$B1,$98,$C1,$1C,$10,$FF,$3A
	.dc.b	$C7,$07,$FC,$EF,$F1,$1F,$E5,$17,$FB,$1B,$5C,$FB,$5F,$DE,$23,$11
	.dc.b	$FC,$8D,$AF,$F8,$E9,$AF,$D9,$6B,$FE,$CF,$8E,$5F,$03,$23,$BB,$5F
	.dc.b	$1B,$5E,$CD,$7F,$C8,$FF,$52,$FA,$6F,$F0,$5B,$63,$FB,$DB,$5F,$F4
	.dc.b	$9B,$5F,$E7,$5F,$6F,$FC,$E5,$FC,$73,$53,$49,$F3,$FF,$E1,$86,$C7
	.dc.b	$F8,$21,$E1,$F6,$FF,$E5,$F8,$FF,$05,$B5,$FF,$51,$B5,$FE,$94,$D7
	.dc.b	$FC,$C3,$F6,$FF,$EA,$77,$FC,$26,$D7,$FD,$56,$D7,$FA,$93,$5F,$F3
	.dc.b	$4F,$DB,$FF,$80,$DB,$E8,$7F,$8B,$A3,$11,$FA,$B6,$BF,$E0,$6F,$89
	.dc.b	$FF,$02,$11,$FF,$02,$3F,$E3,$6A,$F3,$47,$EC,$DA,$9A,$4F,$81,$E0
	.dc.b	$48,$F5,$BF,$E4,$0F,$9E,$7C,$85,$24,$FF,$8C,$A3,$2F,$F8,$D3,$ED
	.dc.b	$FC,$E4,$CB,$FC,$92,$D8,$D4,$60,$8F,$F7,$0F,$B4,$7C,$AF,$F8,$AD
	.dc.b	$B1,$98,$C1,$FE,$54,$FB,$7F,$D7,$C7,$FB,$69,$AF,$F8,$05,$AF,$FA
	.dc.b	$61,$AF,$F1,$76,$A6,$93,$A7,$FC,$C3,$F6,$FA,$23,$83,$FC,$CB,$F6
	.dc.b	$ED,$E9,$2F,$FC,$AF,$19,$6C,$65,$FF,$78,$C6,$5B,$1F,$07,$C0,$91
	.dc.b	$F1,$5A,$FF,$BE,$3E,$DF,$F8,$23,$FE,$1F,$6B,$FE,$B6,$6B,$FC,$D5
	.dc.b	$A9,$A4,$EC,$FF,$3F,$BA,$C7,$FC,$17,$3F,$EC,$06,$BF,$CF,$DA,$9B
	.dc.b	$0F,$63,$99,$27,$FC,$F3,$19,$6C,$65,$F6,$35,$FF,$4D,$7D,$BF,$F6
	.dc.b	$16,$F8,$DF,$E8,$2F,$85,$4E,$67,$FA,$94,$62,$7F,$CD,$5F,$6F,$27
	.dc.b	$FC,$D4,$D7,$FC,$67,$F5,$7F,$E6,$F2,$3F,$E6,$F6,$BA,$30,$43,$FC
	.dc.b	$9D,$ED,$3E,$DD,$F5,$9F,$E8,$36,$CF,$F9,$31,$AF,$06,$BF,$E6,$46
	.dc.b	$BF,$EB,$66,$A9,$ED,$3D,$4F,$F4,$87,$DA,$FF,$92,$E7,$FC,$B2,$D7
	.dc.b	$CA,$D7,$FC,$E2,$D7,$FD,$86,$D7,$B3,$64,$D3,$50,$FB,$67,$84,$D3
	.dc.b	$FC,$05,$E2,$7F,$94,$2B,$FC,$09,$4F,$B7,$FF,$03,$42,$74,$F3,$8C
	.dc.b	$C6,$27,$AD,$74,$F6,$38,$C6,$2F,$94,$62,$F8,$B6,$38,$1E,$2C,$3C
	.dc.b	$CF,$03,$17,$A7,$C4,$5B,$F4,$36,$78,$E1,$E5,$79,$C7,$FA,$3E,$7C
	.dc.b	$8D,$8C,$BE,$86,$BF,$13,$C5,$7D,$1F,$FB,$72,$91,$89,$FF,$4D,$1F
	.dc.b	$3D,$C3,$DD,$AE,$BE,$43,$FD,$98,$47,$FD,$F2,$DF,$B8,$F9,$F0,$F8
	.dc.b	$97,$C7,$7F,$8E,$67,$32,$AC,$6B,$FE,$F9,$AB,$3F,$EA,$D9,$AF,$F9
	.dc.b	$06,$F9,$AB,$E7,$3A,$8F,$05,$57,$B1,$F4,$CE,$4F,$D3,$7F,$C9,$91
	.dc.b	$6B,$AE,$9D,$27,$B6,$1E,$56,$9E,$52,$9D,$0A,$74,$2C,$FA,$30,$E8
	.dc.b	$7C,$67,$13,$FE,$5D,$FB,$A7,$D4,$3E,$BB,$DF,$C1,$F0,$27,$31,$8D
	.dc.b	$38,$53,$EA,$B1,$9A,$F7,$5F,$C6,$0D,$DE,$A3,$33,$FA,$D9,$E7,$59
	.dc.b	$BF,$14,$71,$9E,$9F,$80,$7F,$B7,$63,$16,$8C,$4F,$33,$7E,$3F,$78
	.dc.b	$CE,$5E,$B6,$DF,$C1,$6D,$70,$8E,$B1,$8F,$97,$79,$4F,$83,$FC,$1D
	.dc.b	$FF,$00,$9C,$C6,$23,$FD,$11,$18,$BF,$CF,$B6,$3F,$DF,$59,$7E,$4F
	.dc.b	$BB,$E2,$71,$FF,$A9,$EF,$E0,$5E,$F3,$FA,$53,$9C,$3D,$D2,$FD,$67
	.dc.b	$E0,$C7,$8D,$AF,$89,$AA,$38,$7C,$5F,$9E,$31,$3A,$9C,$E1,$E8,$8C
	.dc.b	$1C,$B2,$FD,$77,$AB,$F6,$F9,$0F,$E0,$AF,$8E,$1F,$4E,$B7,$87,$B3
	.dc.b	$AD,$F1,$1E,$1B,$FE,$54,$D8,$CC,$FC,$86,$DF,$F2,$E3,$D4,$E1,$FA
	.dc.b	$4D,$7E,$8B,$5D,$CF,$BE,$FF,$9B,$7C,$47,$E4,$FA,$4F,$A6,$FB,$EF
	.dc.b	$FA,$19,$CE,$3F,$53,$F2,$CF,$EC,$6F,$DC,$6E,$5E,$5C,$D1,$88,$FC
	.dc.b	$33,$5F,$C2,$B5,$FA,$8D,$8E,$3E,$0F,$C5,$8D,$67,$FC,$47,$2F,$D5
	.dc.b	$75,$FF,$51,$C7,$C9,$78,$9F,$A5,$F0,$FF,$EB,$69,$FC,$8B,$72,$F3
	.dc.b	$35,$FC,$D3,$5F,$CC,$B5,$FC,$43,$72,$34,$D6,$2F,$98,$7D,$9F,$EF
	.dc.b	$DB,$87,$E7,$2B,$C5,$B7,$F0,$D9,$CE,$3F,$9C,$FC,$21,$F4,$9F,$85
	.dc.b	$6E,$1F,$AA,$6C,$FC,$F6,$BF,$B3,$6B,$FB,$26,$BF,$A1,$6E,$67,$39
	.dc.b	$BF,$0D,$FF,$3B,$DF,$F0,$E6,$DF,$97,$0F,$E3,$51,$8C,$BF,$A2,$E2
	.dc.b	$FB,$77,$D6,$FE,$75,$B8,$FF,$02,$6E,$3F,$87,$6B,$FC,$2D,$AF,$F0
	.dc.b	$A6,$BF,$B9,$6D,$77,$AF,$CE,$7D,$D7,$F8,$62,$FF,$92,$B6,$FF,$83
	.dc.b	$DF,$83,$0F,$E7,$93,$97,$F8,$DF,$DC,$BD,$4F,$E6,$DF,$87,$7E,$72
	.dc.b	$73,$18,$E4,$FD,$F7,$EA,$EF,$A9,$FD,$D3,$74,$FE,$3E,$DD,$3F,$48
	.dc.b	$D7,$FC,$4F,$C2,$FC,$5D,$E3,$7F,$45,$6D,$FF,$3E,$F9,$45,$E6,$BF
	.dc.b	$A7,$AF,$AD,$7C,$77,$F1,$B8,$C3,$63,$37,$FE,$10,$9C,$B6,$33,$1F
	.dc.b	$E1,$46,$BF,$C3,$DA,$FF,$8F,$9A,$F9,$3F,$0C,$7C,$57,$91,$B5,$FD
	.dc.b	$66,$7B,$9F,$D5,$44,$69,$FE,$4D,$93,$E4,$BE,$11,$F0,$FC,$7B,$7F
	.dc.b	$D8,$27,$4D,$9D,$34,$D7,$C9,$7C,$B3,$FB,$87,$33,$E9,$B1,$3A,$26
	.dc.b	$3F,$CD,$E7,$51,$8E,$5F,$F1,$0F,$B0,$FC,$07,$37,$FD,$BE,$31,$87
	.dc.b	$F8,$09,$7D,$33,$E6,$B1,$7C,$30,$E6,$DC,$7E,$ED,$F2,$7F,$C0,$0F
	.dc.b	$2A,$E3,$84,$9F,$85,$E5,$DC,$BE,$4B,$E1,$FC,$D3,$F3,$53,$F1,$FE
	.dc.b	$CF,$1D,$9F,$8D,$6C,$64,$EF,$3C,$2F,$E9,$CF,$E4,$FF,$14,$FC,$A7
	.dc.b	$D0,$5B,$FA,$4D,$AF,$B2,$DD,$BF,$C5,$1C,$68,$FC,$BE,$DF,$C2,$AB
	.dc.b	$FD,$5B,$F1,$9F,$97,$8C,$47,$FA,$29,$AF,$C4,$8D,$13,$0F,$CD,$4E
	.dc.b	$A7,$FB,$8F,$87,$FA,$4D,$BF,$7D,$5C,$EF,$13,$FA,$DF,$6A,$4F,$E2
	.dc.b	$FD,$AB,$87,$E4,$A6,$E9,$EA,$9B,$5F,$5B,$8F,$FC,$46,$FE,$4F,$C8
	.dc.b	$FB,$BE,$C7,$EE,$5B,$19,$3E,$DC,$F0,$C7,$F3,$F3,$FE,$48,$F6,$BF
	.dc.b	$D6,$1C,$3F,$E1,$36,$DF,$FA,$19,$B5,$FE,$01,$6B,$FE,$B0,$6B,$FC
	.dc.b	$19,$AF,$F9,$73,$DD,$57,$56,$D7,$FA,$4B,$83,$F3,$9E,$EF,$E5,$8F
	.dc.b	$24,$EA,$FF,$3D,$97,$FA,$6A,$FF,$96,$F8,$7F,$C5,$EF,$F5,$AD,$8C
	.dc.b	$9F,$B7,$8C,$B4,$FD,$D7,$A3,$FC,$7F,$DF,$FE,$40,$D8,$C9,$FB,$F8
	.dc.b	$CB,$4F,$F2,$2F,$8E,$EE,$FE,$62,$48,$FE,$62,$FF,$87,$FA,$3E,$87
	.dc.b	$1F,$E5,$38,$BF,$9C,$CF,$F2,$F7,$F9,$87,$A5,$2F,$F8,$74,$5D,$BF
	.dc.b	$8A,$CD,$BE,$55,$FC,$4E,$3F,$99,$FD,$AF,$F6,$EA,$34,$9C,$9F,$E4
	.dc.b	$FE,$27,$E8,$EF,$8B,$FE,$10,$DC,$7F,$94,$5F,$F7,$EC,$BF,$88,$CE
	.dc.b	$71,$FF,$30,$FE,$BA,$FB,$8F,$F2,$53,$70,$FF,$37,$B6,$7F,$C6,$8D
	.dc.b	$7F,$A8,$EE,$FA,$CF,$EA,$07,$F9,$66,$78,$63,$FB,$A8,$CD,$FF,$31
	.dc.b	$75,$86,$30,$FE,$73,$16,$BF,$C0,$77,$CA,$FF,$D5,$D8,$77,$7F,$CF
	.dc.b	$BF,$03,$FE,$6F,$27,$93,$FE,$63,$7F,$CD,$D3,$FC,$97,$1F,$F5,$63
	.dc.b	$7D,$54,$FF,$A6,$3D,$0F,$ED,$5E,$3F,$F3,$1F,$D9,$BF,$D5,$38,$7F
	.dc.b	$99,$F8,$91,$88,$FE,$1C,$D7,$F9,$8B,$54,$70,$F0,$7E,$C4,$E6,$31
	.dc.b	$A7,$EA,$63,$86,$1F,$DE,$FB,$3F,$BD,$1E,$3B,$B5,$FE,$BB,$F4,$1F
	.dc.b	$AB,$3E,$D3,$FD,$25,$18,$6C,$66,$7F,$D0,$19,$FF,$7E,$8C,$3F,$E8
	.dc.b	$3A,$E1,$3A,$6C,$66,$76,$FF,$82,$8E,$55,$FA,$5F,$7F,$FC,$20,$7F
	.dc.b	$B6,$AB,$8F,$1F,$F3,$A8,$CE,$1F,$DC,$5D,$61,$E2,$9D,$4F,$B6,$1F
	.dc.b	$0F,$EA,$24,$79,$63,$2D,$8C,$E3,$E7,$6B,$F9,$6C,$3F,$C4,$87,$82
	.dc.b	$2C,$FF,$83,$7E,$C7,$FC,$71,$F0,$9F,$E2,$6F,$86,$FE,$FD,$87,$F8
	.dc.b	$A2,$72,$A3,$3C,$8F,$C0,$DF,$68,$C7,$97,$2F,$C7,$C7,$8E,$1D,$31
	.dc.b	$46,$27,$5E,$93,$3E,$35,$FE,$CF,$CB,$AE,$38,$8E,$35,$B7,$C9,$7C
	.dc.b	$AF,$FC,$6D,$87,$A2,$7F,$D8,$0D,$74,$7E,$11,$7D,$A7,$F9,$0A,$30
	.dc.b	$D7,$8B,$5F,$63,$5F,$8C,$F1,$CB,$DD,$3D,$36,$F4,$36,$BC,$E7,$FB
	.dc.b	$AF,$E0,$FF,$C7,$0E,$AF,$96,$3D,$AF,$07,$C4,$DB,$FD,$4D,$EC,$7E
	.dc.b	$4F,$2F,$AF,$CF,$F4,$FE,$E3,$F2,$DC,$9F,$4F,$34,$E5,$FF,$31,$C6
	.dc.b	$63,$10,$8E,$18,$BF,$4A,$7B,$5D,$38,$BF,$CD,$73,$98,$FF,$32,$36
	.dc.b	$FF,$55,$5C,$1B,$3A,$8C,$47,$EC,$5A,$FE,$B6,$79,$39,$9F,$F3,$86
	.dc.b	$1F,$8F,$F9,$4F,$14,$FB,$39,$11,$89,$D4,$FF,$07,$C1,$E0,$66,$F7
	.dc.b	$3F,$0B,$2F,$F4,$2D,$F0,$8F,$85,$18,$9D,$C6,$0F,$F4,$3D,$F4,$7B
	.dc.b	$5F,$B4,$8F,$C2,$6B,$EF,$6B,$EF,$8E,$4F,$F0,$D3,$D9,$EA,$79,$C8
	.dc.b	$FF,$97,$3F,$67,$7E,$09,$F0,$BC,$14,$62,$7F,$C2,$8D,$7F,$9D,$35
	.dc.b	$FF,$23,$35,$FF,$34,$B5,$FE,$72,$D8,$EB,$3F,$E4,$88,$D6,$7F,$C0
	.dc.b	$7C,$6F,$98,$4B,$FE,$AB,$C7,$F5,$B9,$3F,$A7,$BF,$EA,$96,$C7,$F3
	.dc.b	$46,$BF,$05,$AF,$FA,$FB,$57,$D9,$FD,$92,$47,$D9,$12,$F1,$9F,$5D
	.dc.b	$A9,$FB,$19,$7F,$17,$BC,$72,$FF,$6E,$AB,$C5,$F7,$E6,$DF,$4C,$39
	.dc.b	$BD,$F1,$ED,$7E,$B9,$CC,$FF,$81,$A3,$07,$F9,$BB,$C9,$FE,$AD,$1F
	.dc.b	$E9,$46,$BF,$EA,$16,$E3,$FB,$78,$7D,$97,$E1,$23,$0F,$E1,$A3,$F0
	.dc.b	$CD,$7F,$AA,$35,$FF,$6F,$B7,$3F,$E6,$DF,$6F,$FA,$12,$4F,$C9,$8F
	.dc.b	$E7,$5F,$8D,$E2,$3E,$3E,$CA,$C6,$1F,$C1,$E1,$5F,$AB,$25,$73,$8C
	.dc.b	$BF,$E1,$6F,$67,$F0,$5E,$5F,$F9,$AA,$3F,$00,$D7,$F0,$0D,$7F,$B7
	.dc.b	$6D,$FE,$16,$9F,$E0,$35,$FE,$C4,$9E,$DF,$77,$FB,$B3,$F7,$9F,$07
	.dc.b	$23,$24,$3D,$6F,$F0,$5F,$8D,$FC,$5C,$79,$FE,$97,$F8,$21,$9E,$39
	.dc.b	$19,$21,$F8,$A9,$E2,$FF,$B6,$7E,$89,$87,$EC,$DB,$1F,$ED,$26,$BF
	.dc.b	$C2,$7D,$F7,$E7,$4F,$8B,$FE,$37,$18,$7F,$C2,$F2,$8C,$3F,$E9,$B9
	.dc.b	$E0,$D9,$DC,$EA,$3F,$D3,$D3,$A6,$CE,$A3,$11,$FE,$C7,$AF,$5F,$0F
	.dc.b	$DD,$E6,$FE,$4C,$7A,$F3,$95,$F7,$63,$2F,$B6,$3F,$96,$7D,$8C,$8F
	.dc.b	$C1,$3F,$CB,$FB,$71,$1F,$8C,$8B,$7F,$8B,$9F,$E7,$73,$FD,$3F,$96
	.dc.b	$F8,$1F,$91,$8F,$DF,$19,$2F,$E8,$AF,$B1,$E2,$D8,$F9,$2D,$7F,$C8
	.dc.b	$2D,$7F,$97,$35,$FF,$1E,$35,$07,$0D,$8F,$AA,$D3,$FC,$9B,$D3,$FE
	.dc.b	$23,$8F,$E9,$DF,$1B,$27,$F4,$91,$FD,$44,$C3,$FA,$E8,$C3,$FC,$6F
	.dc.b	$C2,$73,$3E,$83,$FA,$C7,$A3,$2B,$53,$96,$CE,$5F,$DC,$CF,$F4,$59
	.dc.b	$D4,$7E,$5F,$C7,$FE,$C5,$E6,$C8,$FC,$B4,$7E,$64,$8E,$0E,$FF,$35
	.dc.b	$87,$91,$93,$FB,$26,$3E,$88,$FE,$DD,$E9,$49,$FE,$08,$FE,$C7,$83
	.dc.b	$11,$FC,$3A,$2F,$17,$F8,$4F,$AF,$F0,$13,$FC,$4B,$E2,$7F,$BC,$11
	.dc.b	$FE,$1D,$F4,$B5,$D1,$FC,$46,$3F,$BF,$47,$E3,$39,$DF,$0D,$49,$5F
	.dc.b	$8C,$FA,$D7,$D5,$FE,$49,$B1,$FB,$26,$BF,$E8,$A6,$BF,$DD,$DA,$FF
	.dc.b	$A1,$5A,$FE,$29,$A9,$A6,$B6,$7E,$20,$F3,$FF,$BE,$36,$BF,$08,$D7
	.dc.b	$E7,$BF,$E1,$4D,$7F,$C7,$DE,$37,$FA,$F3,$DE,$FE,$DD,$E4,$1A,$4B
	.dc.b	$FE,$1B,$FA,$9F,$F9,$81,$FE,$85,$F1,$0F,$86,$FF,$15,$36,$3F,$63
	.dc.b	$3B,$7F,$91,$78,$7F,$CA,$C8,$FF,$65,$FD,$AB,$F8,$86,$7E,$47,$F8
	.dc.b	$7C,$7E,$B3,$34,$62,$3F,$BC,$B6,$33,$E3,$FF,$81,$E3,$04,$70,$7F
	.dc.b	$91,$3F,$1D,$65,$FE,$41,$6D,$FF,$5D,$6B,$FE,$39,$B7,$E2,$8F,$C2
	.dc.b	$7F,$C2,$CD,$9F,$F0,$A3,$5F,$F0,$A3,$5F,$EA,$AD,$7D,$8D,$7F,$C2
	.dc.b	$6D,$9E,$F3,$FE,$21,$8D,$78,$8F,$C5,$7F,$DE,$77,$FD,$CD,$AF,$F8
	.dc.b	$61,$BA,$7D,$EF,$BA,$FF,$4A,$37,$DC,$FF,$16,$47,$FA,$4E,$BF,$C4
	.dc.b	$6D,$C8,$D2,$4F,$02,$4F,$F1,$AF,$C7,$7F,$D7,$6F,$F8,$15,$AF,$F8
	.dc.b	$A9,$B7,$F9,$0F,$9E,$7C,$97,$F8,$FD,$B5,$FE,$3D,$6B,$FE,$3D,$6B
	.dc.b	$FD,$F9,$AF,$E1,$5A,$FF,$8F,$1B,$89,$A6,$B3,$FF,$4E,$F9,$5F,$E8
	.dc.b	$17,$FA,$06,$1F,$E9,$D6,$FB,$5F,$91,$FD,$7D,$F5,$DF,$E5,$56,$E5
	.dc.b	$FE,$53,$6B,$FE,$53,$6B,$FE,$14,$6B,$CE,$3F,$CB,$91,$88,$FF,$29
	.dc.b	$B7,$23,$9B,$7E,$23,$FC,$7B,$1F,$F5,$93,$71,$FF,$49,$B7,$0F,$F2
	.dc.b	$EE,$4F,$0F,$FC,$D8,$D8,$FF,$35,$35,$FF,$35,$35,$FF,$15,$B5,$FD
	.dc.b	$93,$5E,$B1,$FE,$75,$8C,$65,$FD,$B6,$3F,$D6,$36,$F7,$FF,$D5,$CD
	.dc.b	$BF,$F2,$53,$5F,$D7,$7C,$C3,$D4,$FF,$41,$36,$3F,$D0,$0D,$7F,$D0
	.dc.b	$0D,$7F,$C8,$2D,$7F,$D6,$B8,$7F,$99,$73,$FF,$6C,$36,$F2,$7F,$B0
	.dc.b	$57,$8F,$62,$57,$F9,$7E,$78,$3F,$EE,$47,$6F,$F9,$8C,$FF,$6A,$D7
	.dc.b	$FB,$95,$B3,$FE,$C6,$69,$FE,$EC,$3F,$DF,$27,$FA,$38,$CE,$BF,$E0
	.dc.b	$AF,$83,$FF,$6A,$46,$0D,$D7,$F9,$FF,$1F,$D4,$BF,$E1,$0F,$A7,$7C
	.dc.b	$5F,$F8,$4A,$95,$FD,$69,$D4,$66,$31,$09,$E2,$FF,$86,$A3,$AB,$99
	.dc.b	$E1,$C9,$FE,$FF,$8C,$1B,$F2,$7C,$17,$CA,$3C,$31,$8A,$E3,$87,$8A
	.dc.b	$7B,$5F,$3F,$8D,$E1,$8F,$91,$B8,$7A,$DC,$CE,$F0,$E7,$7E,$23,$B9
	.dc.b	$D4,$A6,$6F,$D7,$8F,$2C,$75,$1F,$0A,$3B,$CE,$2F,$DD,$1B,$8F,$F7
	.dc.b	$71,$F3,$23,$E1,$3F,$EF,$19,$F8,$ED,$8C,$BF,$3F,$E6,$1F,$5B,$FE
	.dc.b	$76,$A4,$62,$7A,$C6,$27,$53,$99,$FF,$3E,$FC,$57,$F9,$29,$34,$8F
	.dc.b	$F9,$39,$FF,$29,$E7,$F4,$E1,$F1,$CE,$0C,$EB,$EC,$BB,$5F,$89,$AD
	.dc.b	$2B,$E4,$F8,$0C,$57,$E3,$DE,$B7,$C9,$FC,$95,$D7,$27,$03,$83,$9B
	.dc.b	$EF,$7D,$A3,$15,$C6,$BC,$13,$EB,$AD,$DF,$A6,$7A,$1F,$E7,$28,$CB
	.dc.b	$D8,$C9,$E2,$9E,$CE,$23,$7C,$1F,$EA,$9F,$83,$FF,$76,$4F,$F9,$F2
	.dc.b	$31,$87,$F9,$E3,$E0,$15,$AA,$56,$9F,$88,$D7,$F6,$8D,$7A,$4F,$2A
	.dc.b	$E9,$5C,$F8,$DF,$63,$FE,$0F,$9F,$C8,$C6,$3C,$3D,$4E,$70,$FA,$53
	.dc.b	$9A,$DC,$E6,$7F,$DC,$AD,$7F,$C5,$8D,$53,$C5,$E7,$F3,$2F,$98,$F2
	.dc.b	$36,$FC,$6D,$9D,$CF,$A7,$F5,$B7,$E1,$54,$97,$EA,$8C,$B6,$B8,$11
	.dc.b	$C1,$1F,$6C,$7F,$AB,$A3,$11,$FF,$00,$23,$0F,$D0,$D9,$D3,$F5,$77
	.dc.b	$FD,$0B,$63,$37,$FE,$B2,$6D,$7C,$AE,$9F,$F0,$44,$FD,$06,$C6,$7C
	.dc.b	$2F,$AD,$18,$8F,$E6,$1A,$FF,$A2,$27,$24,$AE,$4F,$E1,$FA,$9F,$8A
	.dc.b	$FC,$2B,$FD,$54,$63,$5D,$4E,$67,$F9,$E4,$E6,$FF,$07,$F5,$AF,$B0
	.dc.b	$FC,$43,$63,$31,$82,$38,$25,$7E,$22,$78,$3E,$2B,$FB,$8F,$81,$AF
	.dc.b	$FB,$7A,$FD,$3C,$1E,$63,$F4,$50,$F1,$DE,$FE,$1B,$8F,$F2,$A9,$D4
	.dc.b	$E4,$91,$99,$DC,$FF,$30,$6B,$FC,$3D,$AF,$FA,$17,$EB,$5F,$45,$FA
	.dc.b	$96,$D7,$EA,$1B,$3F,$A6,$6B,$CF,$E0,$FC,$C4,$9F,$13,$63,$C4,$D7
	.dc.b	$84,$7F,$44,$F6,$9F,$5A,$FC,$57,$F8,$24,$61,$B1,$99,$E0,$DA,$FE
	.dc.b	$9B,$F5,$DF,$E3,$79,$F8,$CD,$8C,$F9,$5F,$3A,$31,$1F,$CC,$5A,$FF
	.dc.b	$80,$5A,$83,$86,$C7,$F0,$FF,$5D,$F0,$B4,$97,$FC,$42,$32,$FF,$12
	.dc.b	$F7,$3E,$F3,$77,$FE,$31,$18,$C3,$F8,$9B,$5F,$37,$C1,$FA,$49,$5E
	.dc.b	$E7,$F9,$07,$D2,$3E,$B6,$92,$7F,$90,$C6,$5F,$C5,$7A,$4F,$A5,$FC
	.dc.b	$8B,$63,$31,$82,$38,$26,$1F,$DF,$27,$83,$7C,$0E,$64,$9E,$EF,$F0
	.dc.b	$BF,$93,$FE,$D5,$1F,$E0,$4F,$09,$18,$AF,$E8,$2D,$7F,$A0,$B5,$FE
	.dc.b	$98,$D7,$FA,$4B,$63,$9C,$F7,$3F,$8D,$7C,$3B,$D5,$A4,$8F,$14,$65
	.dc.b	$B1,$9B,$FE,$A7,$39,$6C,$7F,$59,$F2,$9F,$48,$F9,$CC,$37,$0F,$F1
	.dc.b	$0C,$EA,$72,$48,$CC,$EE,$FF,$C0,$95,$B9,$CD,$7E,$93,$E3,$1F,$57
	.dc.b	$49,$3F,$AB,$8C,$BE,$0F,$D1,$5F,$1B,$E7,$6C,$66,$30,$47,$04,$C7
	.dc.b	$C7,$3C,$1F,$EE,$BF,$5E,$F8,$AF,$E8,$CD,$AF,$B8,$D9,$DC,$E8,$93
	.dc.b	$A6,$C7,$F9,$1B,$CD,$74,$7F,$24,$6D,$7F,$21,$6C,$EE,$74,$49,$D3
	.dc.b	$63,$A7,$C6,$F3,$24,$F3,$7D,$3F,$21,$23,$D2,$D9,$F7,$B5,$FF,$09
	.dc.b	$E2,$8C,$47,$F0,$7F,$39,$DB,$F9,$56,$DF,$F0,$72,$1F,$C1,$4B,$3F
	.dc.b	$E1,$29,$E0,$7E,$75,$78,$6E,$DF,$E8,$ED,$AF,$E8,$8D,$9F,$E5,$93
	.dc.b	$A2,$4E,$9B,$1F,$E5,$19,$F9,$FF,$29,$FC,$C3,$43,$49,$7F,$E2,$AF
	.dc.b	$CF,$FF,$DF,$6D,$F8,$3F,$F1,$84,$63,$0F,$F1,$5B,$5F,$EF,$AD,$7E
	.dc.b	$33,$5F,$F1,$4B,$73,$39,$EA,$6D,$49,$7F,$90,$26,$1F,$90,$6E,$1F
	.dc.b	$96,$7F,$92,$FA,$12,$7F,$91,$B6,$7F,$C9,$FF,$12,$F8,$AF,$F1,$03
	.dc.b	$6B,$F8,$B3,$67,$73,$FC,$C2,$24,$FF,$2E,$6C,$FF,$93,$3C,$24,$62
	.dc.b	$3F,$C3,$F1,$89,$E0,$7F,$7D,$FA,$9F,$E5,$73,$FE,$85,$6B,$FE,$D5
	.dc.b	$6B,$FD,$55,$BC,$0F,$D3,$7F,$B2,$4F,$FB,$6E,$31,$E5,$FF,$6C,$9C
	.dc.b	$CF,$F5,$D9,$CD,$FF,$46,$E7,$FE,$96,$47,$07,$EF,$2F,$F7,$E4,$70
	.dc.b	$4A,$FE,$0B,$19,$6C,$65,$FD,$4C,$65,$B1,$FE,$76,$FC,$C5,$F3,$3F
	.dc.b	$E7,$B6,$C6,$63,$07,$F1,$0E,$99,$1F,$C4,$1B,$5F,$E1,$FF,$98,$7D
	.dc.b	$6F,$FA,$3F,$CB,$46,$2F,$FC,$4C,$DA,$E1,$5B,$8C,$47,$F8,$0C,$FF
	.dc.b	$1B,$38,$7F,$C0,$D5,$D2,$78,$1D,$FE,$49,$F4,$54,$98,$7F,$8D,$63
	.dc.b	$2F,$8B,$E9,$3C,$04,$70,$4C,$3D,$91,$96,$C7,$F8,$B9,$F7,$7A,$7F
	.dc.b	$E7,$C7,$D1,$8A,$31,$1E,$86,$BF,$E5,$16,$C6,$E3,$99,$27,$FC,$99
	.dc.b	$F3,$8F,$AF,$F9,$8D,$8C,$C6,$08,$E0,$93,$FE,$53,$9E,$0F,$F1,$3F
	.dc.b	$C4,$3F,$36,$4F,$47,$FC,$86,$FE,$82,$1F,$53,$FC,$8F,$93,$FC,$6D
	.dc.b	$F1,$2F,$A8,$FA,$D3,$9A,$FF,$7F,$C7,$F3,$5F,$4B,$E1,$67,$C2,$37
	.dc.b	$E3,$7F,$BF,$AB,$67,$F9,$06,$F9,$61,$D6,$31,$3B,$8C,$78,$9F,$E1
	.dc.b	$AE,$0F,$F8,$03,$1B,$AF,$3D,$6F,$0E,$F9,$FF,$9E,$F1,$F3,$57,$E3
	.dc.b	$4D,$C6,$6B,$C9,$19,$9F,$F4,$2F,$07,$1F,$73,$FD,$11,$3C,$4F,$8A
	.dc.b	$E1,$E0,$D7,$E2,$BC,$31,$F3,$57,$49,$F8,$27,$2A,$F8,$CF,$FB,$7B
	.dc.b	$EA,$7F,$36,$E0,$D1,$C2,$79,$57,$4B,$ED,$97,$F6,$5B,$E3,$7F,$3B
	.dc.b	$B1,$FB,$4F,$F8,$5A,$3F,$D0,$8D,$7F,$AB,$35,$FE,$38,$D7,$F6,$3F
	.dc.b	$1E,$F8,$AF,$F5,$13,$63,$FD,$44,$D7,$FD,$44,$D7,$F4,$AD,$79,$C7
	.dc.b	$22,$3F,$EA,$BF,$C7,$BF,$D6,$4F,$FA,$DD,$BA,$9A,$48,$FF,$5A,$35
	.dc.b	$E4,$7F,$AD,$FE,$BF,$F5,$4F,$FA,$DD,$AF,$FA,$DD,$AF,$EE,$9A,$FF
	.dc.b	$10,$6B,$FE,$C9,$6C,$9A,$62,$7F,$A9,$CF,$AF,$C5,$F2,$91,$F1,$7B
	.dc.b	$5E,$52,$38,$21,$E5,$72,$D3,$FC,$83,$87,$FB,$5F,$CB,$FE,$F2,$DF
	.dc.b	$2F,$E7,$4A,$72,$FF,$41,$F3,$BF,$DB,$DC,$EF,$17,$FC,$87,$39,$21
	.dc.b	$FE,$52,$FC,$3F,$A1,$1F,$F2,$8B,$5D,$18,$21,$82,$1F,$E3,$C7,$83
	.dc.b	$FC,$FE,$7F,$CA,$7D,$FE,$4B,$F9,$F1,$89,$FF,$5E,$4F,$F9,$57,$F1
	.dc.b	$17,$CA,$52,$4F,$F9,$76,$32,$FF,$BE,$FF,$0D,$7D,$8F,$E3,$1B,$19
	.dc.b	$8C,$1E,$1F,$92,$91,$FE,$CB,$6B,$FE,$C8,$6B,$E9,$C9,$18,$8F,$F6
	.dc.b	$E6,$BF,$E1,$4F,$17,$EB,$71,$1A,$49,$FF,$6C,$7D,$97,$C1,$23,$83
	.dc.b	$FC,$FF,$F0,$7F,$74,$8E,$08,$7F,$A2,$1C,$B4,$FF,$26,$1F,$80,$F5
	.dc.b	$7F,$5B,$C1,$FE,$42,$8C,$10,$D9,$0F,$F2,$79,$FE,$62,$AF,$F5,$9B
	.dc.b	$FC,$8F,$AD,$18,$8F,$F3,$03,$5F,$F1,$33,$5F,$EF,$DE,$63,$BF,$F7
	.dc.b	$ED,$8F,$E5,$C7,$F8,$2A,$32,$47,$31,$B7,$F9,$A3,$5F,$E7,$CD,$7F
	.dc.b	$D1,$1F,$81,$FF,$24,$7F,$20,$D7,$3E,$67,$E9,$23,$10,$C5,$18,$6C
	.dc.b	$EC,$93,$FE,$B4,$F3,$FF,$64,$47,$04,$3F,$D7,$AE,$5A,$7B,$3D,$DF
	.dc.b	$86,$47,$07,$C0,$FA,$6F,$8E,$47,$04,$3E,$1B,$96,$9F,$EC,$8F,$7F
	.dc.b	$F8,$31,$1C,$10,$FF,$68,$B9,$69,$FE,$EE,$CD,$F2,$DF,$5B,$E0,$BE
	.dc.b	$D3,$63,$33,$B8,$C1,$1C,$10,$FF,$53,$60,$FA,$9F,$F5,$7B,$63,$FD
	.dc.b	$E1,$18,$3C,$1F,$9C,$7D,$6C,$0D,$26,$9A,$25,$7F,$79,$9D,$C6,$23
	.dc.b	$FC,$09,$CC,$F8,$7F,$F1,$A4,$61,$AF,$F9,$56,$35,$E7,$FF,$31,$9C
	.dc.b	$CF,$F0,$06,$BF,$CC,$5B,$3C,$23,$FC,$14,$D7,$FD,$FF,$D5,$FC,$E4
	.dc.b	$8F,$FB,$F9,$AE,$8C,$10,$FF,$0A,$78,$AF,$8E,$FE,$D2,$D8,$FD,$5B
	.dc.b	$5D,$C7,$F6,$EF,$C1,$BF,$91,$C7,$F8,$75,$AF,$F8,$9D,$AF,$F9,$D5
	.dc.b	$A8,$38,$6F,$77,$52,$3F,$E2,$9E,$56,$0F,$A9,$FE,$80,$49,$D4,$65
	.dc.b	$FE,$07,$EC,$7E,$ED,$F0,$35,$34,$D5,$3F,$2C,$FC,$9C,$FF,$90,$63
	.dc.b	$1C,$9F,$D8,$A7,$33,$FD,$4D,$AF,$F0,$88,$FE,$08,$47,$D5,$D3,$A7
	.dc.b	$C7,$C1,$FE,$18,$C3,$F0,$31,$E4,$8C,$11,$F5,$B5,$FF,$26,$47,$E2
	.dc.b	$3B,$3F,$87,$FA,$7F,$C8,$9B,$3F,$CB,$ED,$39,$6C,$7F,$28,$F9,$B7
	.dc.b	$C7,$FF,$8A,$9B,$19,$8C,$11,$C1,$0F,$F2,$F4,$70,$7E,$47,$D3,$BC
	.dc.b	$2F,$F0,$4B,$63,$EE,$35,$DC,$78,$70,$91,$E5,$6B,$D9,$AF,$F9,$D6
	.dc.b	$BD,$6E,$4F,$F3,$87,$CA,$BF,$03,$79,$9F,$D1,$A3,$0D,$8C,$CF,$06
	.dc.b	$BF,$81,$F9,$A7,$CD,$FF,$88,$9B,$1F,$95,$6B,$B8,$D1,$1D,$35,$FF
	.dc.b	$46,$5F,$97,$E3,$3C,$44,$8F,$13,$FE,$30,$FA,$5F,$F1,$AF,$29,$A4
	.dc.b	$F4,$F4,$4A,$FD,$44,$65,$FE,$13,$F2,$8F,$B2,$FF,$21,$36,$33,$18
	.dc.b	$3F,$C7,$FF,$2A,$F8,$1A,$49,$FD,$A4,$65,$B1,$97,$F5,$11,$96,$C7
	.dc.b	$F9,$2B,$E1,$5F,$11,$49,$3F,$BF,$8C,$BF,$C0,$7D,$D7,$D7,$FF,$94
	.dc.b	$9B,$19,$8C,$1F,$E5,$6E,$CF,$C1,$C7,$F9,$1D,$AF,$F1,$1D,$BF,$DE
	.dc.b	$1C,$DB,$24,$7F,$0F,$9D,$BF,$DE,$DB,$3C,$23,$F9,$98,$F8,$8F,$F3
	.dc.b	$63,$4C,$B8,$23,$82,$1F,$D2,$8E,$07,$FB,$1F,$0F,$F2,$BF,$17,$F2
	.dc.b	$7E,$E7,$D0,$FF,$29,$A5,$7E,$42,$73,$1F,$CB,$63,$13,$B7,$9C,$7F
	.dc.b	$72,$8C,$7B,$4B,$ED,$1B,$E6,$ED,$87,$4B,$EF,$18,$9E,$35,$C2,$7C
	.dc.b	$17,$CD,$EC,$78,$EB,$83,$4E,$07,$AA,$38,$5F,$7C,$3C,$ED,$AE,$35
	.dc.b	$E2,$38,$CE,$B0,$F1,$C7,$80,$D1,$63,$E1,$BE,$6C,$38,$E1,$E7,$8F
	.dc.b	$F7,$F4,$FC,$66,$C6,$5E,$A6,$BE,$89,$F4,$7F,$E5,$0A,$46,$27,$ED
	.dc.b	$35,$E1,$3F,$ED,$65,$F0,$D4,$8F,$FB,$69,$BF,$49,$F4,$61,$F1,$3F
	.dc.b	$AF,$8C,$47,$20,$EE,$35,$9F,$E1,$E7,$FB,$85,$7F,$71,$AD,$72,$DA
	.dc.b	$D5,$73,$9C,$BD,$23,$35,$F6,$23,$C7,$55,$F1,$6D,$C0,$F2,$CF,$58
	.dc.b	$FC,$C6,$1E,$47,$C9,$19,$7F,$C0,$91,$97,$F3,$D1,$97,$B9,$E3,$38
	.dc.b	$94,$E3,$5D,$FC,$D7,$A0,$FB,$AF,$E2,$DC,$1F,$C5,$E7,$31,$8A,$FF
	.dc.b	$58,$D3,$E0,$71,$87,$FB,$8E,$39,$4F,$F8,$67,$0E,$51,$89,$FD,$54
	.dc.b	$F2,$BE,$38,$78,$B0,$B7,$BC,$3C,$5F,$3B,$FE,$23,$C3,$A1,$FE,$C2
	.dc.b	$71,$8F,$43,$FD,$E8,$FC,$29,$E3,$1D,$A3,$71,$89,$E7,$3C,$33,$D3
	.dc.b	$EC,$E5,$7D,$2F,$F8,$B2,$BF,$C5,$CD,$9F,$F1,$03,$5F,$F7,$83,$5F
	.dc.b	$E5,$4B,$E8,$7A,$36,$7C,$35,$DB,$CE,$CD,$78,$7E,$95,$F0,$5F,$D1
	.dc.b	$71,$FE,$69,$18,$C3,$FC,$80,$D8,$FF,$19,$35,$FE,$45,$82,$31,$1C
	.dc.b	$AB,$A3,$FE,$B3,$8C,$B6,$33,$3C,$63,$C8,$79,$62,$D7,$AB,$A7,$E0
	.dc.b	$4F,$FA,$F5,$B5,$FE,$04,$C7,$9C,$60,$8E,$09,$5F,$E5,$B8,$78,$DF
	.dc.b	$4D,$B1,$DE,$71,$3F,$5F,$E9,$3E,$66,$1F,$E5,$36,$BF,$09,$AF,$F5
	.dc.b	$D6,$BF,$ED,$36,$E0,$69,$33,$F8,$6B,$E9,$7C,$07,$86,$2E,$3F,$73
	.dc.b	$EF,$BF,$C9,$6F,$F9,$D5,$B8,$7F,$99,$DA,$FD,$66,$BF,$DC,$5A,$FF
	.dc.b	$B9,$DB,$81,$A6,$A1,$F1,$BF,$F0,$06,$BF,$3B,$18,$F7,$FF,$39,$5E
	.dc.b	$2E,$2F,$F0,$85,$EE,$BF,$55,$83,$E9,$5F,$0B,$F8,$26,$E5,$FE,$8F
	.dc.b	$6D,$7F,$A0,$5A,$FE,$41,$AF,$F8,$2D,$AF,$FC,$01,$6D,$77,$AF,$C9
	.dc.b	$CF,$AB,$C3,$FE,$1A,$DB,$FD,$3D,$F8,$31,$F4,$D7,$ED,$FE,$95,$F4
	.dc.b	$BF,$98,$6D,$FF,$A9,$5B,$87,$FA,$5D,$AF,$E9,$9A,$FF,$88,$5A,$B2
	.dc.b	$46,$1B,$5E,$8E,$8F,$F4,$2D,$FD,$06,$DF,$CC,$BF,$0E,$3F,$E3,$F9
	.dc.b	$CB,$FC,$BF,$E3,$BF,$19,$7F,$C5,$3B,$11,$8C,$3F,$8F,$46,$38,$7F
	.dc.b	$94,$CE,$6F,$FC,$9D,$9F,$FB,$0B,$7F,$DC,$F2,$FE,$13,$2F,$F6,$0C
	.dc.b	$60,$F4,$78,$FC,$B9,$BE,$7E,$BF,$85,$AF,$F6,$96,$9E,$29,$FA,$7F
	.dc.b	$76,$FA,$5F,$AE,$DA,$FE,$CB,$BA,$30,$FF,$68,$6B,$FD,$BD,$AF,$F8
	.dc.b	$DD,$AF,$FB,$85,$AF,$39,$FC,$2F,$9E,$F9,$9F,$C2,$B6,$FF,$3F,$7E
	.dc.b	$0D,$FF,$CD,$93,$9C,$3E,$4F,$07,$F9,$1A,$FE,$72,$F8,$6F,$97,$E3
	.dc.b	$7C,$C6,$CE,$9B,$5C,$9F,$E8,$DF,$98,$FF,$A1,$E7,$9F,$E0,$18,$F4
	.dc.b	$FE,$AB,$63,$27,$F4,$09,$E1,$7E,$49,$FE,$D7,$EF,$3F,$37,$7C,$8E
	.dc.b	$3F,$38,$7C,$B6,$A7,$39,$7F,$5A,$B5,$6A,$31,$3A,$C7,$FB,$35,$6A
	.dc.b	$F9,$61,$D5,$FE,$F7,$E5,$3E,$AB,$E5,$62,$C3,$FC,$0F,$18,$58,$73
	.dc.b	$9F,$ED,$75,$C7,$97,$D1,$F1,$92,$7E,$07,$DC,$DC,$3C,$7A,$AB,$1C
	.dc.b	$BF,$9A,$E0,$5E,$30,$FC,$9F,$23,$E2,$7F,$42,$7E,$8B,$E9,$6C,$FD
	.dc.b	$C8,$B9,$FE,$02,$6E,$3F,$84,$FA,$5F,$C5,$92,$FF,$33,$D0,$8C,$5F
	.dc.b	$F9,$02,$31,$3A,$9C,$CF,$F9,$27,$D0,$FF,$02,$5F,$FA,$ED,$B5,$FE
	.dc.b	$93,$BE,$35,$FE,$17,$3F,$CA,$FE,$0F,$E2,$BC,$6F,$DD,$F9,$D6,$FF
	.dc.b	$2A,$DA,$FC,$CB,$71,$FE,$10,$FE,$0F,$E5,$E5,$7D,$2E,$FF,$92,$FE
	.dc.b	$0A,$33,$7F,$82,$EA,$47,$E0,$2B,$73,$98,$C7,$CA,$7F,$81,$B9,$FF
	.dc.b	$60,$DA,$34,$98,$7F,$99,$A7,$6D,$9D,$B6,$7F,$9F,$F4,$7F,$0F,$C3
	.dc.b	$F1,$CD,$BE,$31,$F1,$E3,$5E,$17,$F2,$4E,$0F,$E4,$27,$EC,$BE,$26
	.dc.b	$E3,$FB,$C8,$B7,$FD,$73,$9F,$F9,$7C,$FE,$2A,$7F,$94,$CC,$8F,$F1
	.dc.b	$9F,$CD,$3E,$A7,$13,$FC,$20,$BC,$3F,$CB,$A7,$55,$FC,$2A,$7F,$C9
	.dc.b	$10,$FA,$7F,$F1,$4B,$67,$FD,$13,$E5,$23,$15,$FE,$91,$8C,$57,$F8
	.dc.b	$C3,$E9,$DD,$3F,$E5,$36,$E1,$E6,$9C,$4C,$8F,$E9,$1C,$BF,$8E,$CB
	.dc.b	$FA,$1D,$F0,$C7,$FA,$4C,$E6,$BF,$DF,$D3,$9A,$FF,$40,$4E,$6B,$FB
	.dc.b	$04,$E6,$BF,$A1,$CE,$65,$F4,$FF,$C1,$E7,$FA,$0B,$63,$FD,$4E,$D7
	.dc.b	$FD,$5C,$D7,$FD,$EC,$FF,$35,$B4,$E3,$93,$FC,$03,$3C,$E7,$35,$FC
	.dc.b	$4E,$38,$57,$F1,$5C,$FE,$67,$C4,$7D,$02,$47,$D0,$D5,$1F,$37,$7F
	.dc.b	$98,$7D,$86,$CF,$F9,$D3,$4F,$F4,$53,$F1,$A7,$4E,$5D,$E3,$FE,$28
	.dc.b	$9D,$BF,$E0,$7B,$EF,$5B,$3F,$1B,$F3,$3F,$84,$23,$E5,$7F,$0F,$F3
	.dc.b	$5F,$8A,$3F,$16,$D8,$FE,$7D,$F3,$3F,$EB,$EB,$FE,$68,$7F,$32,$70
	.dc.b	$FE,$A9,$AE,$71,$FF,$3F,$F3,$27,$2D,$8C,$C6,$0E,$3E,$83,$E6,$DF
	.dc.b	$0F,$FE,$C0,$8C,$3F,$D6,$27,$E0,$C6,$31,$FF,$56,$AF,$86,$FF,$42
	.dc.b	$B7,$C0,$FF,$48,$B6,$7F,$DD,$AD,$CF,$FD,$E8,$D7,$DE,$D7,$FA,$A3
	.dc.b	$5F,$F3,$0D,$76,$C7,$F8,$ED,$78,$9B,$3F,$EC,$7A,$ED,$C7,$F8,$C7
	.dc.b	$DD,$DD,$ED,$AD,$CE,$5B,$7E,$0C,$3F,$9E,$E1,$FE,$70,$9C,$BE,$2F
	.dc.b	$91,$F6,$C9,$A4,$AF,$7C,$65,$B1,$99,$F8,$33,$97,$E1,$FC,$57,$FB
	.dc.b	$C6,$BE,$6B,$63,$D9,$3B,$C3,$D9,$39,$7F,$53,$2F,$A3,$7A,$CE,$0F
	.dc.b	$EF,$DF,$0A,$FA,$3F,$C6,$D6,$A7,$31,$FA,$FF,$A4,$78,$FF,$B4,$BE
	.dc.b	$18,$FF,$81,$F0,$E0,$A3,$2F,$87,$1F,$03,$9C,$3B,$C6,$27,$9F,$5B
	.dc.b	$1E,$FB,$78,$B0,$ED,$7B,$C3,$8E,$7F,$E5,$2C,$BC,$35,$FA,$63,$C1
	.dc.b	$19,$AF,$0C,$66,$5F,$01,$FE,$AA,$8E,$07,$C7,$70,$ED,$AF,$C6,$6B
	.dc.b	$F0,$5E,$59,$FC,$09,$EB,$C7,$E9,$AE,$8F,$AF,$47,$D2,$7F,$88,$E7
	.dc.b	$55,$C1,$F8,$C7,$5A,$CE,$1D,$B3,$FE,$45,$7C,$6B,$75,$FB,$1E,$DF
	.dc.b	$E0,$FC,$1F,$9F,$E2,$FE,$47,$5F,$B6,$6B,$FA,$5E,$17,$81,$C3,$87
	.dc.b	$F5,$F1,$89,$FD,$BB,$5F,$DB,$B5,$FD,$BB,$5E,$33,$FE,$86,$C7,$F1
	.dc.b	$2D,$9D,$4F,$FB,$06,$35,$F4,$6F,$88,$F1,$4F,$FB,$7A,$30,$FF,$0E
	.dc.b	$6B,$E9,$F0,$F7,$82,$73,$9F,$EF,$DB,$97,$F0,$A6,$DF,$FB,$46,$F8
	.dc.b	$E5,$FD,$CA,$3D,$F1,$E7,$23,$82,$1F,$09,$CB,$FC,$6A,$7F,$CF,$2F
	.dc.b	$9F,$D5,$F2,$C6,$23,$FC,$EB,$FB,$4F,$E9,$B1,$FE,$BF,$DE,$8F,$D4
	.dc.b	$CF,$F4,$08,$C7,$0F,$FA,$C1,$B1,$F9,$86,$BF,$A1,$6B,$C6,$79,$BF
	.dc.b	$E3,$06,$BF,$E7,$C6,$F2,$7F,$85,$FB,$7E,$8B,$67,$FC,$4F,$18,$F0
	.dc.b	$9E,$FD,$3F,$D8,$75,$FE,$1F,$6C,$6A,$BF,$BA,$46,$AB,$E8,$D1,$A6
	.dc.b	$2B,$FD,$6D,$5F,$E9,$3C,$BF,$DB,$F3,$AE,$1F,$2B,$BF,$6E,$C4,$AE
	.dc.b	$1C,$3F,$E0,$7A,$D4,$7F,$90,$2B,$CA,$47,$FC,$10,$FF,$B7,$17,$D1
	.dc.b	$FF,$A4,$5E,$67,$FA,$95,$C3,$FE,$7D,$6E,$1F,$CC,$1F,$C5,$46,$78
	.dc.b	$7A,$46,$23,$F8,$B3,$5F,$EF,$AD,$59,$FC,$5F,$84,$FD,$1F,$9D,$FB
	.dc.b	$6F,$0E,$F3,$FF,$9F,$2F,$CB,$7B,$6E,$8C,$37,$72,$73,$1F,$28,$8E
	.dc.b	$0F,$F3,$2F,$A9,$F7,$CF,$B6,$47,$FC,$5C,$FF,$28,$58,$FF,$0A,$F5
	.dc.b	$FC,$A4,$8F,$F3,$79,$4F,$F2,$AB,$75,$FF,$37,$47,$B3,$9E,$3F,$E5
	.dc.b	$E7,$F2,$7F,$39,$FC,$7E,$7F,$CB,$31,$A9,$FF,$2F,$BA,$AF,$E1,$31
	.dc.b	$FE,$1B,$EE,$BE,$37,$F4,$08,$FB,$66,$4F,$F6,$DA,$F8,$4F,$E2,$3E
	.dc.b	$9F,$F8,$8D,$F7,$36,$7F,$D4,$13,$B6,$CE,$A3,$11,$FD,$4E,$78,$36
	.dc.b	$32,$FF,$8F,$23,$2F,$EF,$FD,$57,$CE,$FF,$22,$EC,$46,$27,$FD,$FF
	.dc.b	$3E,$6F,$A0,$7A,$5B,$AD,$57,$FB,$06,$B0,$6E,$7E,$5F,$5F,$E2,$BC
	.dc.b	$E9,$5F,$8A,$E1,$FD,$A9,$FC,$9B,$8B,$FA,$2C,$EA,$5C,$5F,$E7,$6E
	.dc.b	$75,$FF,$2C,$9F,$F2,$57,$AF,$F4,$08,$FD,$0D,$5F,$25,$F5,$9B,$1F
	.dc.b	$8B,$6B,$FD,$D1,$AF,$F7,$D6,$BF,$E6,$46,$A0,$E1,$B1,$FB,$D8,$DE
	.dc.b	$5F,$D8,$3E,$3F,$F5,$64,$AF,$F5,$6E,$95,$FF,$55,$C5,$CF,$FD,$53
	.dc.b	$3F,$57,$DE,$FE,$75,$E0,$7F,$B7,$1B,$6E,$24,$63,$6F,$8F,$A7,$1F
	.dc.b	$12,$5F,$F0,$EB,$DC,$7E,$82,$7F,$27,$E9,$7E,$87,$1F,$77,$2F,$F9
	.dc.b	$F7,$A7,$27,$8F,$E0,$3F,$05,$F2,$77,$3E,$9A,$A7,$31,$DF,$A1,$5B
	.dc.b	$8F,$BB,$8F,$AE,$B7,$18,$3F,$8A,$F2,$BE,$43,$F8,$9F,$C1,$C9,$FC
	.dc.b	$54,$7E,$E0,$C3,$FE,$D4,$8C,$3F,$ED,$AF,$03,$F8,$DF,$D0,$48,$C1
	.dc.b	$0F,$F6,$DB,$97,$FD,$BB,$1F,$BE,$FA,$0F,$E6,$39,$7F,$A9,$9B,$3F
	.dc.b	$83,$6A,$FC,$31,$F0,$7F,$C1,$9B,$1F,$C6,$DA,$FF,$91,$1A,$FF,$93
	.dc.b	$5A,$FF,$AD,$1A,$FE,$45,$AF,$78,$FE,$99,$18,$FA,$1F,$E2,$53,$FE
	.dc.b	$F6,$FB,$7F,$E7,$67,$F1,$88,$CF,$57,$EB,$76,$7C,$B7,$F6,$E6,$DF
	.dc.b	$F9,$63,$AB,$F6,$33,$93,$FC,$0D,$F0,$6F,$B9,$FE,$14,$DA,$FC,$73
	.dc.b	$5F,$B9,$D8,$87,$D9,$F8,$FD,$59,$5F,$CA,$34,$FF,$41,$BF,$D4,$3E
	.dc.b	$E3,$FA,$93,$F8,$5C,$D1,$88,$FF,$6C,$C6,$3C,$DF,$F0,$3C,$E6,$7F
	.dc.b	$DF,$8D,$7F,$B0,$B6,$78,$47,$BE,$30,$6D,$7D,$1B,$E2,$7F,$A0,$B6
	.dc.b	$3F,$9F,$B5,$F3,$7D,$23,$E9,$3F,$C3,$ED,$8F,$98,$D7,$FD,$19,$82
	.dc.b	$31,$1E,$C6,$BF,$C4,$63,$53,$FE,$24,$9C,$CF,$F7,$C6,$CE,$A3,$FB
	.dc.b	$55,$7F,$AD,$9F,$3F,$C0,$FA,$8F,$F1,$9F,$DA,$3E,$42,$9F,$E3,$07
	.dc.b	$13,$F8,$36,$BE,$86,$BF,$49,$AF,$9D,$AF,$F8,$41,$B6,$6F,$CA,$BE
	.dc.b	$73,$FC,$8A,$DA,$FF,$08,$B5,$FC,$13,$FC,$15,$CF,$81,$FE,$46,$8C
	.dc.b	$47,$CC,$6B,$F8,$66,$BF,$2D,$AF,$F8,$91,$B6,$69,$8F,$A3,$FF,$4E
	.dc.b	$F4,$FF,$55,$6F,$77,$D5,$6D,$FF,$8B,$9F,$BD,$3B,$DF,$FC,$B7,$1F
	.dc.b	$58,$8E,$09,$3F,$B1,$8C,$BF,$B3,$FA,$3F,$F7,$8D,$7F,$65,$6D,$FE
	.dc.b	$15,$B3,$FE,$3E,$F8,$A7,$D2,$7F,$9C,$1B,$1F,$BD,$6B,$F9,$26,$BF
	.dc.b	$AC,$6B,$F9,$16,$BD,$63,$FC,$EB,$18,$CC,$DF,$85,$7D,$1F,$FA,$01
	.dc.b	$B7,$FE,$4C,$6B,$FB,$D5,$86,$CF,$F4,$0B,$88,$FE,$28,$D7,$F5,$2D
	.dc.b	$7F,$BB,$1F,$E4,$8F,$86,$F6,$E6,$CF,$FB,$D7,$A6,$9C,$C6,$1B,$5C
	.dc.b	$7B,$7F,$CC,$B1,$8A,$E1,$AE,$0C,$CF,$F9,$AA,$9E,$9D,$23,$FE,$7E
	.dc.b	$AF,$F0,$57,$A7,$88,$FE,$C1,$1F,$E0,$D9,$C6,$5E,$79,$FF,$30,$9F
	.dc.b	$E2,$1C,$FF,$B5,$C7,$F8,$4F,$E8,$9F,$35,$F0,$27,$F4,$D1,$8A,$FF
	.dc.b	$0C,$46,$27,$5E,$77,$F8,$BA,$55,$B5,$5D,$9C,$D7,$5E,$7F,$9B,$5B
	.dc.b	$AE,$95,$DA,$FA,$4E,$5E,$DE,$2E,$EF,$EA,$B6,$BF,$C5,$8E,$F1,$B8
	.dc.b	$F9,$9F,$73,$E1,$9A,$7B,$2B,$C7,$1B,$8F,$39,$E3,$9D,$BE,$C9,$DB
	.dc.b	$C4,$CE,$3E,$FC,$7C,$51,$F8,$A3,$91,$4F,$8B,$8F,$C9,$75,$3F,$1D
	.dc.b	$BE,$A3,$E8,$FF,$D3,$53,$FE,$54,$6C,$66,$7A,$4E,$5F,$F6,$37,$C6
	.dc.b	$FF,$B1,$D3,$48,$E8,$F7,$9F,$E5,$5F,$8E,$FC,$46,$3F,$6E,$3F,$CB
	.dc.b	$13,$F0,$94,$EB,$4F,$64,$E6,$31,$3F,$06,$31,$C9,$C6,$3C,$67,$F8
	.dc.b	$22,$BE,$35,$7E,$5F,$6F,$04,$E9,$F1,$46,$67,$BC,$FA,$0F,$1D,$75
	.dc.b	$BE,$F1,$93,$BE,$1C,$A7,$51,$C6,$FC,$FC,$0F,$59,$ED,$FE,$A7,$87
	.dc.b	$F5,$53,$98,$C6,$1F,$C0,$A9,$F0,$38,$C7,$FD,$9F,$1C,$A7,$F7,$0D
	.dc.b	$8F,$DB,$B7,$0E,$AF,$29,$DD,$F3,$C1,$F3,$AF,$85,$C3,$6B,$FD,$DB
	.dc.b	$F7,$5F,$C5,$5F,$E0,$D1,$9C,$BF,$82,$B7,$2F,$5B,$67,$FC,$09,$5C
	.dc.b	$6B,$77,$EB,$EE,$FF,$35,$7C,$9F,$F0,$B9,$FF,$13,$F8,$68,$C4,$7F
	.dc.b	$83,$63,$18,$7F,$90,$5B,$1F,$A3,$CF,$F2,$75,$ED,$E1,$CE,$9F,$EB
	.dc.b	$7B,$B8,$7F,$2E,$9B,$83,$17,$D2,$3E,$2B,$C3,$39,$8C,$5F,$F9,$45
	.dc.b	$B1,$FE,$26,$6B,$FE,$40,$6B,$FE,$25,$6B,$CE,$7A,$D7,$81,$B8,$3E
	.dc.b	$8F,$E3,$37,$0F,$F5,$F6,$5F,$ED,$1C,$7C,$2F,$B6,$BE,$67,$5B,$E5
	.dc.b	$BA,$B6,$FF,$CD,$0D,$CF,$FC,$7A,$D7,$FC,$AE,$D7,$FC,$78,$D7,$E3
	.dc.b	$37,$3F,$F1,$25,$7C,$6D,$BF,$61,$7F,$D2,$FA,$3F,$A0,$BE,$83,$9C
	.dc.b	$FE,$46,$33,$87,$FA,$B7,$27,$DA,$7C,$66,$ED,$F4,$9B,$87,$F1,$F6
	.dc.b	$D7,$F2,$36,$BF,$DB,$9A,$FC,$8A,$E8,$7D,$2C,$BF,$C7,$9F,$1D,$F2
	.dc.b	$0F,$D7,$57,$43,$FD,$FD,$A2,$3F,$31,$F4,$8F,$8C,$FB,$2D,$CB,$FD
	.dc.b	$28,$DC,$BF,$CD,$2D,$7F,$D0,$6D,$7F,$CD,$0D,$7F,$24,$DC,$4E,$76
	.dc.b	$7C,$AF,$FB,$46,$FF,$BC,$36,$FF,$D2,$55,$E1,$CF,$F8,$6C,$E5,$FE
	.dc.b	$F1,$83,$E9,$1F,$03,$F9,$26,$E3,$FE,$AD,$6E,$3F,$E8,$26,$BF,$E9
	.dc.b	$C6,$BF,$E8,$16,$BF,$AB,$6D,$77,$AF,$8F,$EA,$7C,$A3,$F8,$14,$F6
	.dc.b	$C3,$CB,$97,$FA,$6F,$1F,$C4,$63,$FB,$5F,$A4,$FF,$41,$47,$FB,$05
	.dc.b	$B1,$FE,$96,$6B,$FE,$B0,$6B,$FE,$95,$6B,$CF,$D2,$FD,$8C,$63,$33
	.dc.b	$AD,$7F,$92,$78,$9F,$C9,$FA,$CF,$F1,$99,$F0,$E7,$FC,$EA,$73,$97
	.dc.b	$A3,$2F,$E5,$BA,$FF,$39,$6C,$7B,$32,$EF,$F1,$F9,$EE,$F8,$15,$FE
	.dc.b	$CB,$FB,$24,$FF,$06,$6D,$FF,$B8,$5B,$3F,$EB,$66,$BF,$E4,$ED,$7F
	.dc.b	$9E,$79,$1E,$93,$D1,$F1,$B6,$FC,$DF,$4B,$C2,$FF,$CA,$AD,$41,$C3
	.dc.b	$71,$FE,$4A,$D9,$FF,$09,$34,$FF,$0A,$4F,$03,$A7,$BE,$F9,$37,$DD
	.dc.b	$FF,$5D,$8C,$3F,$1E,$7D,$BC,$3F,$E4,$F9,$CE,$1F,$E5,$7E,$1F,$F5
	.dc.b	$A3,$7A,$FF,$AE,$7D,$5B,$EC,$FE,$C4,$49,$D1,$FE,$5F,$9D,$38,$D3
	.dc.b	$FC,$BF,$7F,$A7,$8F,$F3,$9B,$61,$C0,$AF,$F6,$35,$78,$FC,$4F,$C3
	.dc.b	$11,$ED,$ED,$DB,$69,$FD,$7F,$E7,$1F,$65,$F8,$56,$DF,$E2,$7E,$D5
	.dc.b	$F0,$79,$A9,$3F,$8D,$AD,$E1,$D2,$3F,$D1,$B7,$C1,$D5,$7C,$DE,$17
	.dc.b	$CB,$D3,$9D,$F1,$1C,$5F,$7E,$3F,$41,$F2,$F3,$0F,$CC,$74,$FC,$8B
	.dc.b	$F8,$F3,$96,$CE,$B9,$31,$B7,$E9,$7E,$47,$F5,$47,$D6,$F7,$AD,$7D
	.dc.b	$B8,$B7,$FE,$90,$E3,$FF,$5A,$CE,$4F,$D6,$7D,$37,$F8,$6E,$7F,$CC
	.dc.b	$0D,$7F,$CC,$8D,$7F,$D5,$ED,$7F,$36,$D4,$D2,$7A,$9F,$9C,$FC,$13
	.dc.b	$FD,$65,$5F,$BB,$6D,$7C,$73,$8C,$FE,$F6,$33,$97,$F0,$3F,$98,$FE
	.dc.b	$10,$7E,$53,$E0,$6E,$1F,$99,$8B,$87,$BE,$6D,$7E,$7F,$F0,$37,$C3
	.dc.b	$FE,$95,$B8,$7F,$24,$F6,$5F,$6E,$E8,$D5,$CA,$FE,$53,$CC,$C3,$8E
	.dc.b	$7F,$7F,$1E,$38,$6E,$3D,$9E,$DF,$E3,$1F,$F3,$4F,$4F,$EF,$A3,$11
	.dc.b	$FE,$DA,$6B,$FE,$7D,$8D,$13,$1F,$E0,$73,$A9,$FC,$6F,$DC,$7F,$23
	.dc.b	$C8,$FF,$25,$9F,$CD,$EB,$FE,$43,$9C,$DF,$AB,$AB,$F6,$51,$FC,$4F
	.dc.b	$BB,$FC,$95,$D5,$FE,$4D,$BF,$E0,$D5,$A2,$4E,$AF,$F4,$4F,$F3,$BE
	.dc.b	$B9,$1F,$CC,$FE,$D1,$F5,$58,$D7,$F8,$1D,$5C,$7F,$B4,$D6,$AF,$F8
	.dc.b	$94,$FF,$96,$E3,$E1,$F9,$46,$93,$7F,$E9,$FE,$5F,$F4,$93,$83,$6F
	.dc.b	$F8,$E7,$AF,$FD,$3C,$C6,$5F,$C5,$1C,$46,$8F,$E9,$5E,$6F,$F4,$8C
	.dc.b	$FF,$A9,$4F,$F4,$FF,$AA,$75,$7F,$36,$6D,$FF,$36,$6E,$1F,$EC,$96
	.dc.b	$BF,$EC,$DE,$BC,$E1,$FE,$75,$6D,$7F,$44,$95,$61,$FE,$8F,$37,$5F
	.dc.b	$85,$F5,$3F,$E5,$ED,$8D,$26,$1F,$ED,$9F,$1F,$FD,$25,$F1,$3F,$DB
	.dc.b	$CF,$99,$36,$7F,$A0,$45,$C7,$F8,$C6,$1D,$A7,$24,$8F,$E7,$73,$CA
	.dc.b	$31,$9F,$F8,$8B,$E5,$9F,$15,$FE,$72,$6C,$FF,$BA,$F0,$46,$30,$FF
	.dc.b	$78,$46,$2B,$D9,$E7,$7E,$04,$FF,$4E,$56,$F0,$FC,$3C,$7F,$5D,$FC
	.dc.b	$3D,$F3,$9F,$E8,$CE,$B4,$63,$3F,$F7,$D4,$62,$75,$39,$CF,$FE,$00
	.dc.b	$4E,$6B,$73,$99,$FA,$93,$97,$71,$97,$FB,$DE,$DF,$D6,$7D,$6F,$EC
	.dc.b	$E7,$97,$0B,$87,$F8,$C3,$67,$CA,$3E,$7F,$FD,$47,$18,$6C,$66,$7F
	.dc.b	$09,$8F,$F8,$22,$B5,$18,$84,$61,$DB,$5F,$C6,$35,$E5,$3D,$5F,$83
	.dc.b	$F5,$0F,$9E,$FF,$04,$36,$FD,$15,$BD,$FF,$C1,$13,$9B,$FE,$BD,$B2
	.dc.b	$31,$3A,$C3,$FD,$DD,$8F,$96,$31,$8F,$C3,$F9,$8F,$F1,$1B,$EB,$94
	.dc.b	$63,$1F,$5B,$5F,$5B,$5F,$88,$D7,$E1,$B6,$39,$C7,$CE,$8C,$7E,$06
	.dc.b	$FA,$5F,$2B,$6F,$C9,$58,$9F,$F2,$77,$D5,$7F,$48,$F1,$7F,$C7,$3F
	.dc.b	$15,$FC,$0A,$BF,$C7,$73,$95,$19,$E8,$7D,$5F,$B3,$1E,$3D,$DF,$BF
	.dc.b	$C3,$85,$F3,$C1,$18,$9D,$79,$8C,$B8,$D7,$FB,$B7,$1E,$8F,$82,$B7
	.dc.b	$3A,$8E,$75,$C5,$FC,$45,$F9,$DF,$C6,$B5,$E2,$6B,$E1,$7F,$1A,$DD
	.dc.b	$FF,$CA,$73,$FE,$B7,$6C,$EA,$79,$BC,$32,$F6,$56,$E7,$CC,$7C,$89
	.dc.b	$D6,$3F,$E6,$4D,$3F,$D6,$7C,$BF,$C8,$F0,$E3,$5A,$BE,$18,$FB,$71
	.dc.b	$E9,$5F,$3F,$E1,$7F,$59,$5F,$1A,$9F,$2D,$FE,$40,$8F,$D9,$35,$FF
	.dc.b	$1B,$46,$BD,$AF,$F6,$F7,$07,$EB,$FE,$9B,$FD,$C4,$DA,$FF,$39,$D2
	.dc.b	$7F,$CE,$B1,$FB,$26,$C6,$63,$71,$82,$38,$3F,$CF,$BD,$0F,$C0,$FF
	.dc.b	$CD,$D1,$FE,$0E,$6B,$FA,$E6,$F3,$B0,$F3,$7B,$DE,$8A,$FF,$27,$D6
	.dc.b	$A7,$FD,$1F,$7D,$7D,$BD,$36,$BF,$8A,$62,$C3,$96,$5E,$D7,$D2,$49
	.dc.b	$FD,$B4,$ED,$B1,$97,$F8,$8C,$65,$B1,$FE,$9A,$EF,$FF,$3F,$7C,$0F
	.dc.b	$9D,$B3,$FE,$89,$8F,$2F,$E5,$8F,$B1,$A7,$C6,$71,$3F,$CF,$1A,$FF
	.dc.b	$3A,$6B,$FC,$E1,$AF,$C2,$6B,$FE,$74,$6E,$47,$3C,$DF,$E0,$FF,$AD
	.dc.b	$F9,$8D,$AF,$F0,$E4,$63,$C6,$FF,$0E,$C7,$FB,$45,$AF,$F5,$C6,$BF
	.dc.b	$DF,$67,$FB,$0F,$A3,$FE,$BC,$6C,$7F,$77,$F7,$5F,$0B,$F1,$5B,$1F
	.dc.b	$EA,$D8,$F1,$F9,$8F,$8C,$D4,$49,$D1,$FE,$31,$8C,$3F,$E8,$18,$BC
	.dc.b	$DF,$E3,$09,$E6,$78,$EB,$A5,$7F,$77,$6B,$FD,$D9,$AF,$F5,$4F,$C5
	.dc.b	$73,$5F,$93,$8F,$F6,$EC,$63,$9B,$FC,$38,$D8,$FF,$70,$FD,$6B,$EC
	.dc.b	$FF,$31,$08,$C3,$FE,$E0,$6B,$FD,$7D,$AF,$F6,$46,$BF,$DC,$B3,$7D
	.dc.b	$2F,$C9,$24,$7D,$5F,$46,$BF,$EA,$38,$B9,$7F,$86,$67,$EE,$12,$78
	.dc.b	$1F,$E5,$BF,$1F,$E8,$9F,$1C,$8F,$C7,$7F,$3D,$6F,$A6,$FB,$44,$9F
	.dc.b	$F3,$21,$4F,$F5,$A3,$6F,$FC,$C7,$1F,$B6,$F9,$4E,$4F,$FA,$41,$AF
	.dc.b	$F7,$06,$C7,$D1,$E1,$7D,$C7,$F2,$2A,$D5,$F3,$22,$3F,$C7,$F1,$BA
	.dc.b	$FB,$50,$FA,$F7,$C2,$FD,$66,$C7,$19,$DC,$62,$3F,$D7,$4B,$EE,$78
	.dc.b	$1F,$C9,$9C,$47,$F8,$64,$ED,$19,$23,$98,$DB,$FE,$1F,$6B,$FE,$24
	.dc.b	$6B,$FC,$7A,$3F,$CA,$2F,$FA,$D9,$AF,$F8,$59,$B3,$F9,$7F,$84,$79
	.dc.b	$2F,$6F,$F9,$2D,$6B,$97,$A7,$B3,$83,$F9,$75,$F0,$AF,$DB,$7C,$24
	.dc.b	$CB,$E2,$5F,$E4,$E3,$F7,$5E,$9E,$67,$FD,$E5,$17,$3F,$D4,$78,$6E
	.dc.b	$3F,$F0,$BD,$7F,$B8,$FF,$02,$7B,$FF,$A6,$6C,$76,$3E,$D3,$87,$FD
	.dc.b	$D5,$E3,$BE,$5B,$F5,$AD,$8F,$A0,$D7,$FC,$94,$D7,$FC,$9E,$D7,$FD
	.dc.b	$28,$D7,$F1,$0D,$7F,$B1,$47,$F6,$28,$D6,$1E,$BF,$B7,$FE,$34,$4A
	.dc.b	$F4,$47,$BF,$8B,$10,$CF,$E1,$E6,$BC,$C6,$2B,$F9,$27,$C9,$4B,$E7
	.dc.b	$68,$FE,$6F,$CC,$F9,$5D,$57,$5E,$1F,$E0,$D5,$8C,$7F,$51,$7B,$AF
	.dc.b	$D7,$7B,$DF,$B1,$23,$F3,$C7,$E7,$2F,$A1,$FE,$32,$D8,$D3,$5F,$01
	.dc.b	$FA,$F7,$17,$FA,$E5,$F5,$3C,$AA,$57,$27,$AC,$7F,$43,$E9,$93,$F0
	.dc.b	$1F,$E8,$DE,$D6,$23,$BF,$0F,$E2,$E7,$18,$FC,$CA,$FE,$99,$E7,$4B
	.dc.b	$F9,$2F,$F4,$FF,$87,$C4,$72,$E1,$FE,$DB,$7F,$52,$B5,$3F,$91,$F3
	.dc.b	$FF,$D1,$1F,$11,$FE,$0C,$F8,$03,$4F,$E7,$BE,$53,$D0,$A7,$F8,$A1
	.dc.b	$C5,$FF,$03,$6B,$FE,$9D,$6B,$FD,$3E,$38,$11,$E0,$FF,$30,$F8,$BF
	.dc.b	$ED,$04,$8F,$D3,$78,$CF,$AC,$7D,$7F,$E1,$BC,$3F,$EE,$5D,$1F,$C0
	.dc.b	$6F,$FD,$AC,$D7,$FD,$AA,$D7,$FD,$67,$F0,$BF,$B6,$E0,$FE,$4E,$DE
	.dc.b	$BF,$F7,$D3,$65,$97,$87,$DB,$FF,$18,$FD,$4B,$C5,$A4,$AF,$3E,$BF
	.dc.b	$CC,$F1,$FF,$32,$3F,$CE,$E3,$FC,$14,$D7,$3F,$03,$FE,$0B,$8C,$47
	.dc.b	$81,$B8,$FC,$19,$D1,$27,$FD,$E9,$F8,$0F,$F9,$E7,$4F,$E9,$4D,$C3
	.dc.b	$FC,$E8,$D9,$F5,$FD,$73,$D1,$FF,$12,$B6,$3F,$DF,$CD,$7F,$DF,$AD
	.dc.b	$7F,$C4,$4D,$7F,$D0,$0D,$7A,$C7,$F8,$96,$31,$89,$D7,$6B,$F0,$2F
	.dc.b	$C3,$E5,$FC,$A9,$B9,$7F,$BD,$1B,$8F,$D2,$F9,$B7,$BB,$FE,$3C,$6C
	.dc.b	$7F,$BA,$F2,$46,$23,$FD,$E3,$18,$9F,$F1,$C4,$E6,$7F,$C7,$8D,$7F
	.dc.b	$D3,$4D,$7C,$11,$C0,$FF,$20,$7D,$6F,$F9,$1B,$1F,$EF,$0D,$C7,$FD
	.dc.b	$44,$DA,$FE,$49,$F7,$BF,$D4,$DB,$E8,$FF,$69,$B4,$FF,$5E,$9F,$EF
	.dc.b	$86,$8A,$7F,$77,$74,$D9,$E5,$18,$8E,$07,$09,$EB,$3F,$E7,$9D,$3F
	.dc.b	$38,$0A,$7E,$7F,$36,$0F,$57,$FC,$D2,$DC,$3E,$73,$5F,$9A,$D7,$FC
	.dc.b	$CC,$D7,$E1,$C7,$C8,$23,$C9,$FF,$35,$7D,$5F,$D8,$8F,$DF,$B5,$FE
	.dc.b	$63,$48,$C6,$3F,$E2,$7F,$85,$7D,$EF,$F9,$F9,$B7,$F8,$36,$BF,$82
	.dc.b	$6B,$FE,$7B,$6B,$FE,$D9,$6B,$FC,$01,$B7,$FE,$D2,$9F,$F4,$04,$6B
	.dc.b	$2F,$F5,$D7,$B1,$FE,$A3,$DB,$FC,$76,$7F,$8D,$63,$38,$7B,$3D,$5F
	.dc.b	$F4,$9E,$7F,$E0,$68,$FF,$46,$35,$FF,$75,$35,$FE,$22,$D9,$7C,$27
	.dc.b	$E3,$5B,$F0,$3F,$EB,$38,$FC,$86,$8E,$DD,$44,$9D,$1F,$EA,$48,$E5
	.dc.b	$83,$E3,$FF,$D7,$8D,$3F,$BF,$94,$FE,$AA,$7F,$84,$77,$FF,$10,$B9
	.dc.b	$6C,$72,$39,$FA,$DF,$E8,$6C,$DF,$43,$FE,$07,$B7,$D6,$7F,$6F,$97
	.dc.b	$D1,$7E,$8A,$30,$DF,$4B,$F4,$B3,$CB,$2F,$F3,$15,$71,$C3,$A3,$C3
	.dc.b	$0C,$DF,$0C,$39,$47,$F9,$A7,$0E,$4F,$8A,$7A,$DF,$97,$C5,$E5,$3F
	.dc.b	$E4,$D9,$E4,$FF,$9C,$63,$C1,$8D,$9E,$2F,$82,$7B,$61,$E0,$7F,$34
	.dc.b	$DC,$3C,$38,$F2,$9F,$11,$C0,$E4,$70,$2C,$70,$BE,$F1,$FC,$18,$FD
	.dc.b	$36,$3F,$35,$FE,$13,$3E,$E6,$C6,$5F,$97,$EB,$3E,$B7,$FC,$F9,$5F
	.dc.b	$E3,$16,$C6,$67,$B4,$E5,$FF,$41,$0F,$8D,$AD,$32,$E0,$8F,$CB,$3F
	.dc.b	$8F,$FC,$87,$FB,$47,$5F,$F1,$84,$7F,$A7,$67,$E6,$3A,$E6,$AF,$F2
	.dc.b	$68,$AA,$B9,$67,$F0,$E7,$31,$89,$FE,$57,$18,$AE,$31,$E7,$BF,$0D
	.dc.b	$6E,$3B,$1E,$73,$FD,$60,$E5,$F3,$19,$3C,$53,$D2,$BB,$46,$4E,$D7
	.dc.b	$FC,$F2,$BC,$9C,$AF,$7D,$F3,$DF,$EB,$AB,$FE,$15,$18,$C7,$EF,$56
	.dc.b	$EB,$55,$FC,$EE,$B4,$FF,$48,$9D,$57,$09,$CC,$FF,$4E,$9C,$BB,$8C
	.dc.b	$BF,$EC,$DE,$77,$D6,$2F,$BC,$61,$B1,$97,$FB,$93,$67,$FB,$8B,$5F
	.dc.b	$EB,$B3,$C6,$76,$72,$3A,$FC,$DF,$F5,$17,$FB,$E3,$67,$FB,$DB,$5E
	.dc.b	$6D,$7D,$0B,$EB,$39,$92,$79,$BD,$BD,$19,$1D,$9B,$5D,$9A,$F6,$6B
	.dc.b	$F1,$7E,$B1,$F5,$7F,$11,$B5,$FE,$13,$6C,$FF,$84,$9A,$FC,$2E,$7F
	.dc.b	$F1,$94,$FF,$86,$5B,$1F,$E1,$86,$BB,$F2,$BE,$8C,$62,$3E,$B3,$5F
	.dc.b	$F2,$BB,$73,$FE,$57,$3C,$3C,$39,$7D,$DB,$7D,$FD,$DE,$A4,$9F,$23
	.dc.b	$63,$51,$82,$3F,$E0,$FF,$AB,$F6,$A3,$FD,$88,$E1,$B9,$FE,$63,$B1
	.dc.b	$39,$9F,$EB,$BE,$EF,$E7,$1B,$7F,$C7,$5A,$7F,$1F,$F0,$93,$B6,$CE
	.dc.b	$DB,$3B,$9D,$1F,$9A,$FA,$A7,$C1,$D2,$5F,$F9,$2E,$32,$F3,$8E,$C4
	.dc.b	$70,$4A,$F2,$46,$5B,$1F,$E5,$5F,$AB,$FF,$7E,$37,$C2,$FD,$5C,$62
	.dc.b	$7F,$CA,$ED,$7F,$C8,$3F,$56,$FA,$4F,$88,$DB,$F8,$8D,$9F,$E8,$4D
	.dc.b	$78,$CF,$02,$4F,$07,$F7,$7F,$46,$F8,$6A,$49,$FF,$36,$46,$5B,$E8
	.dc.b	$F0,$23,$83,$C7,$F0,$9E,$62,$38,$39,$FC,$A7,$80,$8E,$0F,$BD,$E1
	.dc.b	$BE,$1B,$F0,$4D,$8C,$C6,$08,$E0,$99,$FF,$41,$8F,$C5,$7D,$27,$FA
	.dc.b	$22,$7F,$14,$D9,$FE,$D0,$D7,$FC,$DE,$D7,$FC,$75,$F0,$EF,$99,$FE
	.dc.b	$5F,$E2,$23,$17,$FC,$D1,$B5,$FC,$C5,$AF,$F2,$F6,$BD,$2B,$9C,$62
	.dc.b	$3F,$CE,$8D,$AE,$DE,$C6,$89,$7D,$DB,$5D,$DA,$F7,$7F,$55,$EF,$3E
	.dc.b	$02,$92,$BF,$59,$19,$6C,$66,$BF,$C1,$93,$96,$C7,$ED,$7E,$AD,$F1
	.dc.b	$F4,$97,$FB,$78,$CB,$FE,$B5,$F8,$5F,$77,$86,$69,$35,$FE,$9A,$D9
	.dc.b	$E1,$18,$3E,$3F,$D5,$48,$FA,$0D,$CB,$E8,$35,$FA,$0D,$7F,$D5,$BF
	.dc.b	$52,$FA,$BF,$E1,$CD,$AF,$E1,$CD,$9F,$F1,$63,$5F,$F5,$9B,$5E,$73
	.dc.b	$C8,$93,$CB,$CE,$EA,$47,$07,$E1,$FE,$53,$F0,$24,$70,$4A,$FC,$54
	.dc.b	$65,$B1,$FE,$3F,$F3,$DF,$54,$E1,$FF,$01,$C6,$1B,$19,$9E,$0D,$AF
	.dc.b	$F6,$6D,$6E,$73,$5C,$08,$FF,$31,$F4,$5F,$4F,$FC,$D5,$B1,$98,$C1
	.dc.b	$1C,$12,$FF,$CA,$D3,$C1,$BC,$5C,$C9,$3D,$DF,$C1,$B0,$FC,$51,$63
	.dc.b	$FD,$FB,$C1,$FE,$0D,$7F,$C3,$0D,$5C,$3F,$8F,$9E,$11,$FE,$12,$F9
	.dc.b	$B7,$BB,$FE,$1C,$6C,$FF,$9A,$E9,$18,$AF,$F7,$9C,$62,$B8,$12,$B8
	.dc.b	$56,$EB,$B4,$FF,$2E,$F1,$65,$7F,$2C,$6C,$F4,$AF,$D3,$B7,$1F,$21
	.dc.b	$2B,$F9,$9F,$77,$F3,$82,$38,$24,$F9,$BC,$04,$EA,$30,$FF,$6D,$FA
	.dc.b	$97,$C7,$D2,$5F,$F6,$F8,$CB,$7E,$93,$81,$1C,$1F,$D1,$BC,$5F,$E9
	.dc.b	$64,$70,$7C,$4F,$47,$E7,$11,$C1,$31,$F8,$51,$96,$C6,$5F,$B1,$19
	.dc.b	$6C,$7D,$7F,$D9,$BF,$86,$93,$B3,$FC,$AE,$FF,$10,$FA,$07,$C4,$7F
	.dc.b	$96,$9B,$5F,$7D,$B7,$F7,$9A,$FF,$86,$C7,$CA,$3F,$17,$4F,$F9,$CA
	.dc.b	$30,$D8,$CD,$7F,$AD,$67,$2D,$8C,$D7,$53,$AB,$87,$C1,$3C,$1D,$35
	.dc.b	$CF,$D5,$7F,$BE,$6B,$B3,$63,$6D,$77,$F0,$BF,$E8,$8F,$4D,$F3,$AF
	.dc.b	$AF,$C3,$70,$FF,$17,$36,$FC,$D5,$C2,$B6,$4A,$DB,$67,$7F,$10,$F9
	.dc.b	$2F,$03,$6B,$FC,$74,$D9,$DC,$E8,$93,$A6,$C7,$8B,$0F,$EA,$8F,$D5
	.dc.b	$C3,$D7,$E4,$FA,$89,$7E,$A6,$CF,$16,$BC,$7E,$2B,$FC,$B1,$3F,$E4
	.dc.b	$D6,$C7,$C0,$6B,$B3,$FC,$AD,$F5,$1F,$E5,$2F,$07,$FD,$B1,$CF,$F8
	.dc.b	$78,$C4,$FF,$B9,$63,$FD,$59,$F5,$2B,$FF,$56,$4E,$4F,$F3,$45,$70
	.dc.b	$F4,$97,$CF,$DA,$FF,$6F,$56,$C5,$6E,$7A,$DF,$4B,$E3,$5D,$2F,$8C
	.dc.b	$62,$38,$C7,$47,$F2,$57,$E8,$6C,$65,$F1,$63,$FE,$73,$9F,$F3,$4E
	.dc.b	$97,$C1,$34,$95,$F9,$D3,$CF,$1A,$7D,$53,$C9,$B1,$9C,$3D,$3E,$1B
	.dc.b	$EA,$7F,$15,$5F,$38,$F8,$F5,$BD,$3E,$83,$63,$38,$EE,$F9,$70,$FE
	.dc.b	$EF,$1F,$8B,$2F,$A0,$FF,$47,$5F,$A9,$B1,$98,$FF,$46,$46,$3E,$AD
	.dc.b	$F6,$3F,$E8,$F6,$CF,$FA,$55,$AF,$EA,$1A,$FE,$5D,$AF,$FA,$9E,$91
	.dc.b	$88,$F9,$6D,$8C,$CF,$8B,$EA,$A5,$7F,$A8,$ED,$18,$8F,$DB,$C6,$1D
	.dc.b	$46,$5F,$D4,$46,$5F,$37,$9F,$EB,$23,$EB,$6B,$A3,$04,$3F,$8A,$FD
	.dc.b	$53,$E8,$7F,$D7,$1D,$A8,$C4,$FF,$23,$6B,$C2,$77,$FA,$73,$DC,$A4
	.dc.b	$9F,$E5,$51,$96,$C6,$5F,$E4,$B1,$96,$C7,$FB,$27,$83,$F9,$BF,$9B
	.dc.b	$7C,$2F,$F0,$0D,$7F,$41,$8C,$1B,$F1,$1E,$B5,$24,$FF,$4F,$7A,$79
	.dc.b	$CF,$AB,$A4,$9F,$F0,$1B,$5F,$EF,$CD,$7F,$C2,$4D,$7F,$DD,$18,$FF
	.dc.b	$84,$23,$11,$FE,$1A,$9E,$CF,$F5,$7F,$63,$D4,$47,$07,$FB,$9F,$EA
	.dc.b	$9F,$57,$F9,$C6,$C6,$63,$06,$7E,$46,$47,$16,$BF,$D8,$FC,$DF,$90
	.dc.b	$47,$FB,$43,$5D,$18,$21,$FE,$1D,$FA,$AF,$8E,$47,$FC,$4E,$FD,$3F
	.dc.b	$90,$91,$FD,$CD,$AF,$F8,$94,$F9,$E4,$3F,$C5,$E5,$34,$60,$87,$F8
	.dc.b	$E3,$E8,$FF,$84,$B7,$BB,$F0,$91,$88,$FE,$FD,$B2,$31,$1F,$51,$AF
	.dc.b	$F9,$31,$B1,$B9,$E6,$7F,$0A,$F2,$7F,$DB,$55,$FC,$84,$FE,$DF,$F0
	.dc.b	$79,$23,$98,$FC,$C3,$FE,$03,$6B,$FE,$11,$6A,$3F,$BC,$F2,$F7,$23
	.dc.b	$CD,$AF,$F9,$FB,$C1,$7C,$A7,$8F,$DC,$46,$27,$FD,$18,$D7,$84,$EE
	.dc.b	$31,$1F,$E2,$39,$E3,$39,$24,$7F,$B8,$FC,$9F,$81,$F0,$5F,$EE,$76
	.dc.b	$FC,$1F,$F6,$19,$D1,$0F,$F7,$44,$68,$8F,$C9,$F0,$5F,$13,$FC,$D9
	.dc.b	$B1,$98,$C1,$FC,$EB,$ED,$DC,$74,$93,$F4,$E3,$2D,$8C,$BF,$D5,$E3
	.dc.b	$2D,$8F,$4E,$92,$3F,$E0,$03,$5F,$F8,$00,$D7,$FE,$00,$35,$FC,$57
	.dc.b	$CB,$FE,$40,$8F,$FA,$3F,$8D,$18,$8F,$CC,$35,$FC,$AC,$69,$FC,$67
	.dc.b	$D4,$39,$FF,$AC,$B6,$33,$18,$33,$F2,$32,$38,$B5,$FE,$ED,$28,$C4
	.dc.b	$7F,$95,$63,$0F,$03,$FB,$E6,$ED,$DE,$87,$DE,$24,$EA,$32,$D8,$CB
	.dc.b	$FE,$61,$9E,$51,$88,$FF,$38,$47,$5F,$7B,$F1,$24,$8E,$8D,$79,$57
	.dc.b	$FA,$81,$C9,$0F,$F0,$97,$D3,$BE,$5B,$FB,$FB,$63,$F6,$CD,$77,$1F
	.dc.b	$8B,$F8,$29,$1F,$89,$6B,$F8,$76,$BF,$E7,$BE,$14,$62,$3F,$9E,$FD
	.dc.b	$2B,$E9,$FF,$AA,$36,$3F,$DA,$0D,$7F,$D9,$ED,$7F,$D0,$DE,$D7,$EA
	.dc.b	$5F,$F1,$63,$5C,$F3,$FF,$86,$A3,$11,$E0,$F6,$BB,$91,$EE,$F8,$BB
	.dc.b	$A4,$78,$9A,$F6,$F8,$9E,$84,$7F,$8E,$52,$31,$1B,$32,$E1,$FE,$5B
	.dc.b	$F4,$AF,$0F,$FB,$A3,$63,$FD,$E0,$D7,$FD,$DE,$D7,$FD,$4D,$F6,$BF
	.dc.b	$8A,$7F,$C9,$8D,$73,$C9,$FE,$3A,$8C,$47,$FA,$E5,$AF,$F2,$96,$A6
	.dc.b	$9A,$F0,$DF,$1C,$F9,$0F,$F0,$53,$6B,$FC,$B9,$3A,$9C,$9E,$1E,$D9
	.dc.b	$3F,$D0,$6F,$F9,$7B,$63,$2F,$A9,$AF,$B9,$FE,$91,$F2,$CF,$A5,$FF
	.dc.b	$6C,$CF,$F9,$AE,$31,$F0,$DA,$9C,$CF,$FB,$2E,$72,$F9,$BE,$95,$E6
	.dc.b	$FF,$88,$DB,$19,$8C,$11,$C1,$0F,$F6,$B6,$69,$E3,$18,$5F,$35,$D4
	.dc.b	$91,$E1,$7C,$1D,$52,$3C,$8D,$79,$35,$FE,$E7,$3F,$B3,$6C,$7F,$8E
	.dc.b	$FE,$91,$F4,$B4,$93,$F5,$23,$2F,$D6,$FA,$1F,$F1,$DB,$7D,$4F,$F2
	.dc.b	$24,$62,$3D,$AD,$7F,$C0,$6D,$7F,$C6,$AD,$59,$23,$0D,$9E,$BD,$52
	.dc.b	$BC,$4D,$7B,$F0,$78,$88,$F9,$DF,$F2,$C7,$47,$F9,$5A,$3E,$8B,$5F
	.dc.b	$F0,$BB,$5F,$F2,$4B,$5F,$2B,$6C,$D2,$7A,$DF,$DD,$A7,$FB,$59,$FC
	.dc.b	$CC,$A7,$F3,$22,$9F,$E1,$1F,$0F,$FC,$25,$E5,$71,$24,$7F,$44,$36
	.dc.b	$23,$6F,$F8,$8F,$37,$D1,$B8,$FF,$C3,$9E,$6F,$F5,$9A,$F9,$9F,$3F
	.dc.b	$FE,$1F,$7F,$C3,$3E,$55,$C1,$39,$AE,$11,$D2,$F9,$D6,$F0,$ED,$87
	.dc.b	$23,$FC,$55,$5C,$9D,$98,$C3,$A5,$71,$BC,$CF,$DF,$BE,$78,$78,$2F
	.dc.b	$73,$C4,$DE,$1E,$37,$C0,$71,$BC,$CF,$08,$F7,$9C,$2F,$AE,$1E,$18
	.dc.b	$E3,$7E,$76,$D7,$18,$ED,$87,$C2,$BF,$7D,$7C,$93,$C0,$7F,$BC,$E7
	.dc.b	$C8,$E5,$AF,$CD,$F3,$1F,$47,$F5,$DB,$1C,$5A,$F1,$6B,$FE,$FF,$F8
	.dc.b	$4F,$F7,$5A,$69,$1F,$F7,$7B,$F3,$17,$C4,$3E,$2B,$11,$C4,$F9,$26
	.dc.b	$9C,$CF,$19,$E1,$3A,$38,$65,$F8,$B7,$81,$99,$E1,$5C,$A3,$AD,$72
	.dc.b	$AE,$11,$CB,$4F,$95,$39,$7D,$D1,$C2,$B9,$57,$7A,$D4,$FD,$1B,$EE
	.dc.b	$FE,$47,$67,$A6,$F8,$2F,$4C,$FF,$88,$23,$18,$FF,$8B,$9B,$1F,$E2
	.dc.b	$76,$BB,$9E,$2F,$ED,$63,$2D,$8C,$BF,$B3,$8C,$BF,$C6,$BE,$8F,$E8
	.dc.b	$57,$F5,$C6,$C7,$F4,$06,$BF,$EA,$66,$BF,$ED,$86,$BC,$E3,$BF,$2A
	.dc.b	$72,$D8,$CF,$D1,$FF,$48,$7F,$B5,$B6,$BF,$A7,$B5,$FF,$5A,$B5,$FF
	.dc.b	$73,$AF,$6B,$99,$27,$9B,$DB,$B6,$47,$66,$D7,$66,$BD,$9A,$FF,$BD
	.dc.b	$3E,$8D,$F4,$7E,$C6,$D7,$F8,$09,$B3,$FD,$99,$AF,$FB,$33,$9B,$FC
	.dc.b	$41,$3F,$E0,$86,$C7,$F6,$C6,$BB,$F2,$BE,$4C,$62,$3F,$D8,$19,$FF
	.dc.b	$20,$9C,$92,$BA,$78,$5C,$08,$E0,$F0,$7A,$DD,$08,$E0,$97,$E3,$8C
	.dc.b	$B6,$3F,$C2,$1F,$47,$EC,$8F,$F6,$97,$5A,$31,$1F,$CE,$DB,$19,$D3
	.dc.b	$F2,$32,$F8,$AF,$F4,$3D,$FF,$B9,$5B,$1F,$C5,$CD,$79,$9C,$23,$31
	.dc.b	$FC,$6D,$AF,$F4,$56,$BF,$EE,$56,$BC,$A7,$FC,$6F,$F4,$7F,$EE,$06
	.dc.b	$F8,$5F,$96,$8C,$4F,$F8,$79,$AF,$78,$F1,$12,$7D,$0D,$8F,$45,$70
	.dc.b	$7C,$91,$97,$FA,$D6,$CF,$A3,$FF,$6D,$DF,$F9,$29,$B3,$FE,$2E,$6B
	.dc.b	$C1,$AF,$F6,$3F,$A3,$7C,$C7,$CC,$6D,$7C,$99,$46,$1F,$98,$D7,$FA
	.dc.b	$23,$5E,$93,$CC,$93,$FE,$5E,$F9,$B7,$D1,$FE,$F1,$B1,$98,$C1,$9F
	.dc.b	$46,$47,$16,$DF,$8B,$DD,$F4,$12,$7D,$18,$A3,$B7,$C1,$49,$F1,$36
	.dc.b	$BF,$A7,$7C,$27,$E3,$3C,$13,$49,$5F,$0E,$32,$D8,$CC,$F2,$8F,$91
	.dc.b	$F4,$7F,$E7,$F6,$E5,$F9,$88,$C4,$FE,$55,$AF,$FB,$5B,$E0,$1F,$37
	.dc.b	$FC,$BB,$C4,$FF,$72,$36,$33,$87,$FB,$8E,$30,$47,$04,$AF,$F6,$DF
	.dc.b	$B1,$C4,$91,$FE,$EB,$6C,$6A,$30,$47,$F5,$5F,$40,$F7,$7F,$62,$D9
	.dc.b	$FC,$33,$67,$73,$A2,$4E,$9B,$1F,$B5,$F5,$FF,$DB,$B3,$F8,$F6,$C6
	.dc.b	$7E,$03,$F7,$F1,$88,$FD,$0F,$CF,$FC,$27,$FD,$16,$D7,$39,$7F,$48
	.dc.b	$8C,$47,$F2,$06,$BF,$E7,$06,$F4,$1A,$6B,$56,$AF,$5B,$E9,$13,$3F
	.dc.b	$E1,$D3,$A8,$C6,$1F,$E2,$78,$C4,$EA,$B9,$36,$BF,$8D,$FD,$1B,$E8
	.dc.b	$E9,$27,$F6,$51,$96,$C6,$70,$FE,$3D,$39,$6C,$74,$F3,$65,$72,$6C
	.dc.b	$F9,$70,$FF,$78,$4F,$97,$E0,$3F,$1C,$47,$04,$9E,$11,$97,$FC,$89
	.dc.b	$E7,$3E,$8D,$C7,$FE,$06,$8C,$36,$33,$7F,$CD,$67,$2D,$8C,$D7,$46
	.dc.b	$BF,$E7,$99,$DB,$9A,$FE,$87,$E8,$3E,$8E,$93,$0F,$E3,$51,$96,$C6
	.dc.b	$6B,$FA,$34,$E5,$B1,$D3,$92,$57,$26,$CF,$E4,$23,$F1,$C4,$3F,$DF
	.dc.b	$7C,$3F,$E1,$62,$39,$8E,$0F,$FC,$00,$F8,$F7,$B5,$FE,$1E,$6C,$66
	.dc.b	$30,$47,$04,$9F,$F7,$5C,$F0,$7F,$DD,$BE,$0F,$F4,$97,$FD,$9A,$DE
	.dc.b	$53,$4C,$4F,$EB,$67,$FC,$0A,$D7,$FD,$AA,$D9,$36,$7F,$97,$57,$F3
	.dc.b	$DE,$FF,$E7,$84,$70,$4B,$F2,$C6,$5F,$EE,$DF,$44,$FA,$3F,$EF,$AD
	.dc.b	$AF,$EA,$4D,$9D,$CE,$BE,$72,$56,$9B,$3F,$0B,$C5,$FE,$9C,$49,$FE
	.dc.b	$A4,$FC,$FF,$42,$47,$D3,$6C,$FC,$C6,$BF,$DC,$33,$FF,$15,$3F,$12
	.dc.b	$32,$FC,$86,$B9,$9F,$F0,$96,$8F,$94,$FE,$1E,$4E,$DF,$F2,$EB,$FC
	.dc.b	$4B,$E5,$5F,$0D,$FE,$60,$6D,$F2,$6E,$1F,$EB,$3C,$11,$88,$FF,$6F
	.dc.b	$7C,$CB,$E8,$F3,$7F,$E2,$38,$C5,$7F,$9E,$9B,$19,$9D,$D7,$FA,$FE
	.dc.b	$72,$D8,$CB,$D9,$A8,$47,$07,$C7,$F4,$6F,$A3,$F0,$36,$FC,$8D,$9F
	.dc.b	$03,$5F,$F1,$A7,$CB,$3E,$8F,$FD,$23,$E9,$23,$17,$FE,$28,$6D,$7F
	.dc.b	$8F,$5A,$F2,$AE,$31,$88,$FF,$1F,$35,$FF,$26,$36,$9F,$0E,$F8,$FF
	.dc.b	$0B,$70,$FF,$1B,$B6,$FF,$C9,$8D,$7F,$C9,$8D,$7F,$C9,$8D,$7F,$CB
	.dc.b	$4D,$AE,$F5,$E6,$F2,$A5,$FB,$5B,$7E,$D6,$BF,$19,$AF,$F9,$67,$E2
	.dc.b	$7E,$A4,$FF,$93,$1B,$1F,$E6,$16,$BF,$E6,$16,$BF,$6C,$7D,$13,$D8
	.dc.b	$FF,$76,$C7,$FB,$CF,$A7,$53,$9F,$0B,$F0,$5E,$92,$7F,$D1,$FF,$46
	.dc.b	$C7,$FD,$DB,$39,$C3,$F1,$18,$71,$7A,$18,$C7,$BC,$EE,$3C,$71,$8C
	.dc.b	$3A,$DF,$47,$C7,$19,$9E,$58,$FF,$96,$F0,$ED,$1E,$D3,$9C,$66,$7D
	.dc.b	$B1,$99,$7B,$1F,$E8,$F8,$FF,$46,$46,$36,$E9,$39,$9E,$0D,$7E,$2B
	.dc.b	$C7,$2F,$55,$71,$9F,$79,$F3,$A7,$58,$FF,$AD,$BE,$8F,$F2,$13,$F3
	.dc.b	$23,$9E,$16,$3E,$4D,$70,$6C,$67,$A1,$7C,$A7,$E6,$CB,$E6,$BF,$CF
	.dc.b	$D1,$FE,$AB,$6C,$66,$3F,$D5,$91,$8F,$A3,$7D,$1F,$FA,$BD,$B3,$FE
	.dc.b	$98,$6B,$FE,$AD,$6B,$FE,$AD,$6B,$FE,$AD,$6B,$FE,$BB,$6D,$1B,$4F
	.dc.b	$A3,$F1,$92,$BF,$D4,$F1,$96,$C6,$5F,$F5,$F4,$65,$B1,$E9,$F4,$24
	.dc.b	$7A,$5A,$FA,$5A,$FA,$5A,$FF,$1F,$EB,$7D,$1B,$E8,$FF,$D5,$11,$86
	.dc.b	$C6,$67,$F9,$93,$5F,$E5,$DF,$46,$F7,$78,$B6,$3F,$9C,$35,$DC,$68
	.dc.b	$8E,$9A,$FF,$45,$F0,$7F,$C1,$AF,$F4,$9F,$33,$FD,$79,$C3,$FD,$21
	.dc.b	$B3,$A9,$FE,$99,$E2,$7F,$97,$DF,$EB,$2D,$73,$EB,$F2,$8C,$47,$F6
	.dc.b	$5F,$A3,$EB,$E4,$9A,$48,$FF,$10,$35,$FF,$0C,$B5,$E4,$68,$86,$8F
	.dc.b	$EC,$5E,$D7,$A0,$8E,$0F,$F6,$C7,$D1,$BE,$8F,$F3,$AD,$8C,$C6,$0C
	.dc.b	$FC,$8C,$8E,$2D,$7E,$57,$9B,$F1,$88,$FF,$73,$6B,$A3,$04,$3F,$C6
	.dc.b	$BF,$47,$F1,$08,$FF,$90,$5F,$A3,$F1,$F2,$3F,$BF,$35,$FF,$1E,$9F
	.dc.b	$30,$87,$F9,$24,$A6,$8C,$10,$FF,$03,$7D,$1F,$F0,$17,$F9,$E7,$0A
	.dc.b	$31,$1F,$DE,$63,$1E,$47,$E1,$A7,$32,$9C,$BF,$E1,$C8,$CC,$62,$3B
	.dc.b	$CF,$F0,$CF,$2B,$FD,$D7,$1F,$CB,$CF,$F0,$33,$87,$0D,$7F,$9A,$35
	.dc.b	$FE,$A4,$FE,$85,$CF,$27,$52,$3D,$4F,$F0,$23,$87,$AB,$5F,$F4,$57
	.dc.b	$80,$F9,$3E,$1E,$DF,$F3,$E6,$C6,$67,$FD,$2C,$D7,$FD,$3E,$D7,$9C
	.dc.b	$70,$23,$FE,$BF,$F2,$3C,$9F,$03,$FE,$CA,$6F,$BF,$FD,$8E,$74,$43
	.dc.b	$FD,$69,$1A,$23,$ED,$F0,$5F,$31,$FC,$E1,$B1,$98,$C1,$FC,$F3,$E8
	.dc.b	$5C,$54,$93,$F1,$E3,$2D,$8C,$BF,$5E,$32,$D8,$F4,$E7,$23,$FD,$B8
	.dc.b	$D7,$FD,$98,$D7,$FC,$15,$97,$FB,$AA,$31,$1F,$8F,$E1,$7C,$B7,$E3
	.dc.b	$C9,$1F,$8D,$6B,$F9,$A6,$BF,$9A,$8E,$2F,$E0,$7E,$8D,$CD,$FD,$71
	.dc.b	$B1,$98,$C1,$9F,$91,$91,$C5,$AF,$E5,$E5,$18,$8F,$F0,$44,$61,$E0
	.dc.b	$7F,$BF,$7C,$EF,$BA,$47,$04,$3F,$DE,$0E,$5A,$7E,$97,$27,$03,$DD
	.dc.b	$FC,$39,$27,$51,$96,$C6,$5F,$F7,$E3,$5F,$F0,$F7,$D1,$BE,$5B,$FB
	.dc.b	$FB,$63,$F5,$0D,$7F,$6C,$D7,$FC,$4D,$C0,$F8,$2F,$C6,$12,$3F,$14
	.dc.b	$D7,$F1,$0D,$78,$BF,$D4,$3E,$89,$F4,$74,$93,$FE,$A6,$8C,$BF,$BB
	.dc.b	$F6,$7F,$4C,$FF,$09,$6B,$FE,$39,$6B,$FE,$57,$6B,$FC,$5E,$1E,$CF
	.dc.b	$84,$91,$DD,$F2,$76,$C8,$F1,$35,$F0,$7C,$47,$F8,$9A,$CD,$24,$79
	.dc.b	$DA,$F0,$30,$7F,$B0,$BE,$8D,$E1,$7F,$72,$6C,$66,$30,$7F,$8D,$3E
	.dc.b	$8F,$FC,$E1,$1F,$C7,$5A,$FF,$2B,$6B,$FE,$5B,$6B,$FE,$79,$6B,$FE
	.dc.b	$B5,$6A,$69,$AF,$09,$F1,$AF,$8F,$FF,$03,$B6,$BF,$DB,$6D,$7F,$DA
	.dc.b	$4D,$7C,$5D,$9E,$92,$3E,$96,$BE,$96,$BC,$0F,$E8,$DF,$2C,$FA,$3F
	.dc.b	$F7,$1C,$61,$AF,$F9,$E6,$35,$F0,$BF,$D0,$5B,$1F,$EF,$0F,$A3,$79
	.dc.b	$9F,$E2,$06,$C6,$63,$04,$70,$43,$FD,$DD,$1C,$1B,$F3,$5C,$C8,$F7
	.dc.b	$7B,$74,$C8,$F1,$35,$E6,$D7,$FB,$9C,$FE,$BD,$FE,$E4,$D7,$FB,$87
	.dc.b	$D1,$3E,$8F,$FC,$72,$D8,$FE,$D9,$BA,$31,$1F,$E9,$D8,$C4,$FF,$90
	.dc.b	$3E,$8B,$FC,$6F,$3F,$DF,$DA,$FF,$7D,$6C,$9A,$4F,$53,$91,$27,$C3
	.dc.b	$1F,$B0,$9E,$1E,$57,$72,$4E,$B8,$3C,$44,$70,$7A,$B9,$FF,$CA,$4F
	.dc.b	$F8,$59,$AF,$F8,$55,$AF,$F8,$4D,$B5,$FE,$F6,$F5,$BF,$CB,$B1,$82
	.dc.b	$3F,$DC,$A7,$FB,$39,$FE,$C1,$F0,$8E,$0F,$EA,$E5,$32,$EB,$CA,$E4
	.dc.b	$48,$FE,$70,$6C,$46,$DF,$F0,$A6,$4F,$A3,$71,$F8,$3C,$DF,$EA,$B5
	.dc.b	$F3,$FE,$7F,$FD,$E1,$18,$B4,$63,$CA,$FF,$7A,$47,$3B,$E5,$18,$C3
	.dc.b	$AE,$1F,$E1,$D8,$DD,$F1,$8F,$7D,$F4,$78,$98,$AE,$38,$78,$5B,$7C
	.dc.b	$A7,$67,$9C,$B5,$E3,$7C,$07,$3C,$33,$87,$86,$38,$1A,$3F,$1D,$7D
	.dc.b	$9E,$06,$6B,$CC,$D9,$E5,$1D,$F0,$F3,$5F,$CE,$BF,$83,$7F,$EF,$DA
	.dc.b	$F2,$36,$32,$FC,$AF,$31,$F4,$7F,$59,$B1,$FE,$57,$AF,$F0,$AC,$E4
	.dc.b	$FF,$73,$C3,$E1,$A9,$1F,$F7,$5B,$7E,$93,$EE,$C3,$E2,$1F,$15,$F9
	.dc.b	$09,$CC,$F3,$0F,$09,$DC,$63,$6F,$F3,$55,$F1,$7F,$33,$1F,$9A,$9D
	.dc.b	$D7,$4A,$D4,$EA,$73,$5C,$A3,$E9,$1E,$5B,$FB,$AF,$E9,$2A,$D7,$8B
	.dc.b	$5E,$93,$E3,$3A,$57,$67,$C7,$55,$E8,$6C,$C9,$4C,$E1,$D0,$F8,$13
	.dc.b	$E8,$DD,$EA,$3C,$6F,$87,$C1,$F1,$27,$31,$8A,$FD,$BD,$7F,$80,$1F
	.dc.b	$DE,$D6,$E3,$94,$60,$DC,$F1,$C3,$F7,$33,$A7,$F6,$F3,$D3,$C7,$D6
	.dc.b	$1E,$08,$E5,$3F,$C0,$3E,$73,$FC,$DF,$18,$94,$62,$79,$9F,$EA,$BF
	.dc.b	$A4,$FF,$05,$D6,$E7,$59,$7B,$5B,$1F,$C4,$9B,$5C,$7C,$9F,$1C,$63
	.dc.b	$0F,$C6,$FC,$B7,$E2,$2B,$FC,$E3,$68,$C4,$7F,$1C,$8C,$1F,$E5,$12
	.dc.b	$BF,$C9,$76,$78,$5F,$01,$B3,$ED,$37,$97,$D3,$8C,$D7,$BB,$07,$CD
	.dc.b	$F4,$8F,$F4,$1B,$5F,$E5,$AD,$7F,$94,$B6,$38,$E3,$E8,$9C,$B6,$33
	.dc.b	$87,$F9,$A6,$B8,$71,$7D,$9C,$7F,$0C,$D9,$EF,$7F,$6A,$31,$87,$F3
	.dc.b	$8C,$3F,$6B,$5F,$E7,$36,$C7,$F4,$76,$BF,$E9,$C6,$BF,$D1,$9A,$F3
	.dc.b	$8F,$34,$60,$FE,$8A,$FA,$78,$9E,$F7,$E3,$1B,$1F,$89,$9F,$0E,$3F
	.dc.b	$9F,$9C,$DF,$CD,$FC,$15,$EA,$FC,$36,$E1,$FE,$8B,$6C,$FF,$5F,$6B
	.dc.b	$FE,$B4,$6B,$FD,$79,$AF,$F5,$A6,$E0,$6F,$2F,$F5,$BA,$FC,$F3,$6B
	.dc.b	$C9,$7F,$A6,$8C,$63,$FC,$6B,$37,$EA,$3F,$B5,$BF,$E3,$27,$F2,$57
	.dc.b	$0F,$F1,$76,$BF,$C7,$5A,$EF,$93,$F0,$93,$D1,$B3,$A8,$C6,$3F,$EA
	.dc.b	$DE,$07,$C4,$FE,$6C,$FE,$01,$3A,$C3,$F3,$13,$AA,$FD,$EC,$BF,$47
	.dc.b	$FF,$03,$C7,$A2,$BF,$9E,$46,$1F,$5C,$ED,$F2,$C6,$5F,$E6,$4D,$7F
	.dc.b	$9C,$35,$E6,$70,$3F,$3D,$0F,$AC,$FC,$C1,$E8,$9F,$06,$7F,$9E,$8F
	.dc.b	$E3,$5E,$1B,$E8,$DE,$17,$EA,$63,$0D,$8C,$E5,$FE,$4A,$6E,$5F,$E4
	.dc.b	$96,$BF,$D3,$AB,$8C,$EE,$B9,$1D,$7E,$93,$FD,$B1,$1F,$C9,$25,$18
	.dc.b	$AF,$E4,$31,$C9,$B1,$9C,$7F,$8C,$7C,$7B,$C8,$FE,$69,$48,$C6,$1D
	.dc.b	$72,$F8,$8F,$F3,$29,$D5,$7F,$3B,$F8,$D7,$CE,$FF,$9A,$6B,$EC,$46
	.dc.b	$3C,$CF,$E1,$8D,$8C,$CE,$EF,$FC,$8D,$EF,$3E,$6F,$FA,$0B,$6B,$C5
	.dc.b	$F1,$98,$9F,$E9,$50,$F8,$2F,$F6,$FB,$F8,$73,$1E,$EF,$F1,$B6,$99
	.dc.b	$8D,$E1,$FA,$48,$CB,$63,$2F,$F8,$02,$32,$D8,$CB,$FE,$58,$AF,$F5
	.dc.b	$D4,$FF,$53,$6C,$7F,$52,$8E,$BB,$FF,$6B,$9C,$BF,$DB,$71,$FE,$23
	.dc.b	$3F,$E6,$26,$D7,$F9,$7B,$F0,$8F,$EF,$B5,$BA,$F8,$A4,$3F,$B1,$46
	.dc.b	$9E,$47,$F6,$AB,$7C,$3D,$3F,$B6,$FD,$5B,$E2,$F1,$5E,$99,$B8,$FF
	.dc.b	$4D,$AD,$4E,$4F,$EE,$7B,$BF,$3D,$CD,$7F,$70,$8C,$61,$FD,$FF,$CB
	.dc.b	$60,$FB,$2F,$C4,$61,$C3,$5F,$F1,$8E,$1C,$32,$EB,$39,$7F,$D1,$38
	.dc.b	$72,$BE,$38,$FF,$81,$B5,$7D,$7D,$33,$7F,$11,$F8,$55,$23,$D3,$D5
	.dc.b	$F9,$CD,$7F,$B5,$6F,$E5,$9F,$B5,$5B,$9F,$65,$6F,$5F,$B5,$3E,$1F
	.dc.b	$76,$5F,$D2,$BA,$7F,$75,$C9,$38,$BF,$F1,$64,$7F,$0C,$AF,$F5,$0B
	.dc.b	$A7,$FC,$3E,$3F,$28,$7D,$CF,$E6,$DB,$1F,$9F,$6E,$1F,$9B,$6B,$F9
	.dc.b	$96,$BF,$CB,$1A,$FF,$99,$9B,$1D,$E7,$FC,$05,$F9,$27,$F9,$AB,$EC
	.dc.b	$3F,$C7,$0D,$9F,$2D,$78,$31,$FF,$34,$CE,$70,$F8,$1E,$D7,$D5,$3F
	.dc.b	$C7,$7F,$39,$FB,$0A,$F8,$31,$9C,$3F,$32,$EB,$A7,$17,$F5,$3E,$C7
	.dc.b	$F3,$47,$F0,$FF,$D5,$BF,$B7,$E1,$FA,$B8,$CE,$1F,$B1,$75,$E6,$62
	.dc.b	$FF,$C8,$79,$FF,$89,$BE,$93,$DB,$E9,$9A,$4C,$3F,$A8,$4E,$DB,$3B
	.dc.b	$6C,$EE,$BF,$B6,$F5,$3E,$59,$97,$F2,$56,$DF,$F9,$6B,$1E,$97,$FA
	.dc.b	$79,$CB,$F9,$2F,$AA,$FD,$C1,$FB,$6F,$B0,$B7,$FC,$36,$2C,$FE,$D6
	.dc.b	$6E,$BF,$8C,$F1,$3F,$48,$7D,$0F,$42,$E1,$FC,$1A,$2D,$FF,$92,$78
	.dc.b	$FF,$C0,$D7,$C6,$BF,$CD,$91,$FE,$75,$9F,$F5,$0F,$A6,$8C,$47,$FB
	.dc.b	$E1,$AF,$FB,$D1,$BE,$5B,$EA,$3F,$8F,$DA,$93,$9F,$8C,$FC,$2F,$BB
	.dc.b	$FD,$B2,$F7,$5F,$C8,$7E,$47,$F7,$C6,$6A,$E1,$FC,$82,$2C,$FF,$60
	.dc.b	$AE,$39,$3B,$55,$F9,$2F,$A0,$FF,$2E,$68,$69,$31,$FF,$08,$5F,$1F
	.dc.b	$8C,$FE,$1C,$71,$8F,$E5,$EF,$E7,$3E,$9B,$99,$23,$8D,$A3,$FC,$D5
	.dc.b	$8B,$EF,$7F,$8C,$1C,$6B,$F9,$CB,$86,$9F,$4A,$71,$A7,$F0,$B9,$BC
	.dc.b	$1F,$EB,$5C,$3B,$5F,$0A,$FB,$9F,$49,$EE,$24,$79,$1F,$1F,$C9,$FF
	.dc.b	$29,$3F,$6A,$D8,$FE,$7B,$AB,$EF,$3F,$BC,$B6,$3F,$BE,$B6,$FF,$42
	.dc.b	$D7,$FD,$CF,$D7,$F3,$6B,$89,$2B,$8B,$FE,$8D,$F2,$3F,$0D,$1F,$DA
	.dc.b	$78,$BF,$C1,$98,$75,$D3,$FB,$44,$63,$EB,$FC,$64,$AF,$1B,$72,$F1
	.dc.b	$B5,$F1,$DA,$3C,$1F,$29,$FD,$94,$E9,$5F,$DE,$5C,$34,$F8,$71,$8A
	.dc.b	$FE,$BD,$FB,$27,$F8,$46,$7F,$C3,$ED,$8F,$E0,$6D,$7E,$1B,$5F,$D1
	.dc.b	$68,$8C,$42,$31,$E9,$FF,$8A,$27,$31,$8A,$FC,$AF,$1F,$FB,$0B,$4F
	.dc.b	$E9,$7F,$0B,$FE,$33,$BF,$14,$62,$BF,$C7,$1F,$53,$FE,$9A,$F7,$BF
	.dc.b	$65,$EF,$BE,$A7,$F6,$9A,$AB,$4A,$74,$FE,$E3,$EA,$BF,$C8,$AE,$6B
	.dc.b	$8D,$70,$AD,$CF,$3F,$25,$87,$82,$F7,$18,$9E,$47,$5F,$71,$96,$E3
	.dc.b	$F8,$46,$3D,$6F,$8E,$1F,$E8,$59,$EC,$D9,$F4,$C7,$FA,$81,$AF,$9B
	.dc.b	$E4,$FF,$CB,$CF,$FA,$36,$11,$88,$E0,$D7,$87,$BF,$D6,$38,$E7,$EE
	.dc.b	$AE,$33,$E7,$AF,$95,$5A,$C3,$F9,$07,$87,$FC,$A7,$C3,$F9,$F8,$F1
	.dc.b	$AD,$5F,$B9,$AF,$2C,$3E,$15,$F1,$8F,$89,$E0,$FF,$1D,$F2,$3E,$CF
	.dc.b	$0F,$F8,$31,$BB,$7F,$83,$1B,$5F,$8D,$E9,$79,$D8,$E7,$FE,$44,$FC
	.dc.b	$B6,$CF,$EF,$A3,$13,$C2,$30,$43,$F8,$0C,$68,$FA,$98,$FF,$9E,$FE
	.dc.b	$53,$C2,$D9,$FF,$33,$47,$87,$F6,$97,$97,$AE,$F4,$D2,$57,$F9,$42
	.dc.b	$B6,$DA,$F4,$46,$23,$FB,$9B,$5F,$E2,$6D,$7F,$62,$D7,$F6,$0D,$C8
	.dc.b	$D3,$58,$7F,$A5,$7E,$37,$CC,$4A,$F2,$79,$DE,$B3,$C6,$47,$FD,$47
	.dc.b	$5F,$E0,$46,$C6,$5F,$3B,$5F,$ED,$FD,$AF,$B0,$F9,$2D,$8F,$F3,$FC
	.dc.b	$7C,$CF,$19,$F7,$F4,$FA,$2E,$27,$FA,$9C,$60,$FA,$B4,$F9,$AE,$A4
	.dc.b	$9F,$F0,$0F,$A5,$5F,$F2,$D4,$5E,$5F,$EA,$13,$79,$7F,$CB,$33,$6B
	.dc.b	$FC,$05,$3E,$53,$C6,$47,$FB,$2B,$E2,$F6,$3D,$A4,$8F,$F0,$8F,$B7
	.dc.b	$ED,$3D,$64,$7F,$B6,$BE,$AF,$17,$E2,$92,$3F,$DC,$9E,$47,$FB,$75
	.dc.b	$FF,$52,$C6,$5F,$F3,$A3,$FA,$4E,$AF,$F0,$B7,$85,$6B,$85,$7F,$82
	.dc.b	$E3,$07,$CF,$F8,$79,$1F,$DF,$9F,$9E,$47,$F9,$AF,$C7,$7D,$82,$3F
	.dc.b	$E2,$CF,$85,$AF,$FA,$16,$2E,$FF,$DE,$E7,$F4,$5B,$CA,$FF,$66,$3F
	.dc.b	$AA,$F8,$CF,$B9,$1F,$E9,$88,$D3,$4F,$F1,$75,$7F,$2A,$5D,$0A,$E3
	.dc.b	$3F,$EB,$0F,$D0,$7F,$19,$5F,$EB,$96,$C6,$79,$BF,$84,$FB,$2F,$0B
	.dc.b	$F5,$D3,$FE,$80,$7F,$55,$19,$ED,$FE,$68,$DC,$3F,$CD,$73,$A2,$4E
	.dc.b	$9B,$1F,$E3,$0F,$85,$FF,$5D,$BE,$AD,$3F,$C9,$1A,$7F,$B2,$BE,$0B
	.dc.b	$F8,$5D,$6A,$BF,$DA,$1F,$0B,$FE,$3E,$78,$1F,$02,$35,$9F,$C3,$BE
	.dc.b	$53,$A6,$BC,$BD,$8F,$D7,$FC,$9C,$8F,$F7,$6B,$FA,$FF,$96,$C4,$7F
	.dc.b	$AF,$62,$DF,$F8,$CE,$7E,$3E,$7F,$D2,$E3,$FD,$4A,$D3,$FC,$C5,$B7
	.dc.b	$F5,$ED,$7F,$9A,$E0,$FC,$09,$C7,$86,$C7,$F5,$F6,$BF,$DE,$0F,$F3
	.dc.b	$63,$87,$0D,$7F,$BE,$B5,$DC,$7F,$69,$7C,$0D,$8D,$69,$FE,$69,$F8
	.dc.b	$7F,$E2,$24,$AF,$F4,$2F,$CC,$57,$FD,$BD,$17,$3F,$F2,$A4,$FD,$6F
	.dc.b	$87,$95,$F3,$F8,$3F,$DC,$6F,$D6,$F1,$DF,$1F,$FC,$96,$18,$7D,$7A
	.dc.b	$DC,$63,$7F,$C1,$C6,$27,$58,$7F,$5B,$BE,$51,$FE,$EE,$F8,$3F,$E5
	.dc.b	$24,$C3,$C1,$C9,$F2,$67,$F4,$D9,$23,$19,$7F,$8D,$9B,$5F,$E3,$56
	.dc.b	$FC,$53,$E4,$BF,$D1,$53,$D0,$FE,$D9,$3C,$2F,$A4,$7F,$72,$AE,$1E
	.dc.b	$7F,$E6,$BE,$22,$4F,$C5,$7F,$7D,$E3,$62,$3E,$01,$F9,$B8,$C6,$7F
	.dc.b	$9B,$EE,$FF,$6D,$4F,$E6,$FC,$87,$D3,$7F,$48,$24,$7F,$AF,$3E,$6A
	.dc.b	$BD,$39,$BF,$D4,$D3,$8C,$3F,$D1,$95,$FA,$EF,$87,$97,$F6,$5F,$CB
	.dc.b	$FC,$63,$EB,$A9,$F8,$B7,$13,$FE,$35,$6B,$F9,$70,$E1,$FC,$D3,$5F
	.dc.b	$F2,$9B,$5F,$F6,$43,$5E,$31,$FD,$6B,$E5,$3F,$B2,$92,$3F,$80,$F9
	.dc.b	$4F,$A1,$7C,$57,$F8,$7F,$E9,$DF,$03,$FE,$1F,$6D,$F9,$79,$91,$88
	.dc.b	$FF,$7E,$B5,$FD,$87,$D1,$3E,$9B,$F7,$2D,$8F,$E2,$21,$C3,$F6,$B2
	.dc.b	$F2,$7C,$67,$F5,$1F,$BA,$FE,$09,$3F,$EE,$46,$C9,$BA,$3E,$37,$FA
	.dc.b	$C3,$6B,$E6,$57,$BE,$7F,$CC,$8D,$8C,$CE,$DF,$E1,$11,$96,$C6,$5F
	.dc.b	$F4,$1C,$65,$B1,$97,$FD,$33,$E6,$3E,$F3,$FC,$94,$D8,$FD,$E3,$5F
	.dc.b	$CC,$FC,$3B,$E3,$BF,$AA,$46,$1A,$FF,$A4,$A3,$5E,$17,$F9,$F5,$B1
	.dc.b	$F8,$96,$BF,$C8,$5A,$F1,$9E,$68,$FF,$48,$B8,$6C,$FF,$95,$BC,$CF
	.dc.b	$F6,$F4,$7F,$43,$6B,$FA,$56,$F4,$FA,$3F,$19,$7B,$5F,$CE,$1B,$5F
	.dc.b	$E9,$36,$BF,$9F,$6B,$FC,$DD,$AF,$FA,$C1,$AF,$9D,$A9,$A6,$BB,$7F
	.dc.b	$88,$7C,$2F,$F9,$B3,$C6,$FF,$4C,$D7,$F5,$B6,$BF,$BD,$F9,$DF,$F1
	.dc.b	$F3,$FD,$7D,$AF,$F4,$9A,$46,$23,$FD,$87,$18,$F7,$7F,$CF,$73,$99
	.dc.b	$FE,$E6,$D7,$FC,$1A,$D9,$E1,$1F,$BF,$6B,$DD,$F2,$E5,$FD,$CD,$BA
	.dc.b	$7F,$7E,$6C,$FF,$86,$3A,$7F,$D8,$CF,$F1,$76,$BE,$66,$BF,$E0,$66
	.dc.b	$BC,$77,$FF,$49,$C6,$27,$FD,$DD,$18,$24,$F9,$FE,$51,$F7,$BF,$E9
	.dc.b	$86,$C7,$F8,$1D,$AF,$F2,$9F,$89,$FF,$08,$37,$D7,$FF,$50,$C6,$23
	.dc.b	$E5,$B5,$FF,$0D,$B5,$FF,$1A,$B7,$0F,$F7,$84,$F2,$24,$FC,$BC,$BF
	.dc.b	$C6,$27,$F8,$90,$AF,$F8,$F9,$B2,$FA,$07,$C8,$7F,$AC,$9B,$5F,$55
	.dc.b	$AF,$E0,$DA,$FE,$09,$AF,$F8,$CB,$D7,$78,$DF,$EB,$C6,$C7,$F8,$AD
	.dc.b	$AF,$F1,$FF,$A4,$EB,$C1,$FE,$31,$9C,$B6,$B9,$1F,$E7,$39,$DE,$5F
	.dc.b	$E4,$0B,$7B,$FF,$EC,$A6,$9F,$E3,$92,$9F,$D6,$CF,$F1,$2D,$FF,$8F
	.dc.b	$DC,$B6,$39,$1C,$FD,$9F,$F1,$8F,$D2,$3D,$8F,$F0,$8D,$FF,$99,$FD
	.dc.b	$CF,$E1,$4A,$5E,$6E,$3D,$05,$6A,$73,$5F,$C4,$67,$2F,$F1,$38,$ED
	.dc.b	$96,$6B,$86,$1E,$0B,$DD,$F7,$C3,$B5,$75,$6C,$73,$AE,$13,$CA,$FF
	.dc.b	$90,$74,$72,$C7,$C7,$7F,$BE,$AE,$4E,$9E,$0E,$4E,$95,$E5,$7D,$18
	.dc.b	$78,$1F,$21,$93,$E2,$61,$E2,$78,$9F,$24,$E2,$6E,$FC,$78,$F9,$67
	.dc.b	$84,$F2,$BF,$F0,$33,$4F,$F0,$45,$7C,$86,$CF,$D6,$7F,$CC,$50,$F7
	.dc.b	$1F,$49,$CA,$30,$D8,$CC,$F4,$6B,$E0,$E6,$FF,$32,$26,$91,$FF,$34
	.dc.b	$3F,$25,$FE,$B1,$BD,$7F,$0F,$15,$74,$FF,$AC,$F6,$FB,$B1,$FD,$56
	.dc.b	$75,$39,$DB,$FC,$4D,$18,$BE,$55,$E3,$3E,$B1,$FD,$76,$B8,$3F,$E8
	.dc.b	$39,$E0,$E8,$EC,$58,$DB,$E6,$31,$3D,$A7,$8D,$F9,$A7,$C0,$FA,$2B
	.dc.b	$BF,$33,$D6,$7C,$CF,$F7,$08,$C3,$63,$38,$FF,$43,$6E,$1F,$D4,$5A
	.dc.b	$FF,$50,$6B,$FE,$C0,$6B,$FE,$DA,$6F,$8E,$FA,$47,$C6,$7F,$80,$23
	.dc.b	$0D,$8C,$DF,$F8,$99,$B7,$FE,$22,$6B,$FE,$20,$6B,$FE,$09,$AE,$73
	.dc.b	$C6,$BB,$7A,$9F,$E4,$C8,$FF,$18,$B6,$BF,$C5,$AD,$7F,$C5,$6D,$7F
	.dc.b	$9C,$7D,$27,$C7,$93,$48,$FF,$8F,$3E,$F5,$F0,$9E,$D6,$C6,$63,$04
	.dc.b	$70,$4A,$FF,$1F,$CF,$07,$FC,$01,$E0,$F8,$27,$FB,$05,$FD,$98,$C4
	.dc.b	$EA,$3E,$67,$4B,$E0,$29,$2F,$FC,$A3,$19,$6C,$66,$BF,$C9,$F3,$96
	.dc.b	$C7,$E1,$F8,$15,$48,$7F,$82,$5C,$BF,$E5,$EF,$C3,$3F,$AD,$B7,$C2
	.dc.b	$FC,$04,$62,$FF,$CB,$AD,$7F,$CB,$6D,$7E,$E3,$56,$E8,$C3,$67,$F3
	.dc.b	$9E,$3F,$FA,$22,$FF,$C3,$8D,$BF,$F1,$53,$5F,$F1,$4B,$5F,$E9,$3C
	.dc.b	$4F,$4B,$FC,$A4,$DA,$FE,$3E,$D9,$FF,$53,$35,$FF,$7F,$35,$E7,$3C
	.dc.b	$89,$3F,$E6,$5F,$A6,$FD,$3F,$31,$A4,$E6,$F0,$FB,$B2,$BC,$2D,$C3
	.dc.b	$C2,$D7,$C2,$D7,$F8,$87,$C1,$7F,$9B,$67,$FC,$E8,$D8,$FF,$3A,$35
	.dc.b	$FF,$3E,$35,$FE,$D7,$E3,$3C,$CF,$F3,$A3,$6B,$FA,$0B,$67,$FD,$90
	.dc.b	$D7,$FB,$98,$70,$F4,$9E,$64,$9F,$F5,$2F,$D2,$BE,$37,$F8,$33,$63
	.dc.b	$31,$83,$3E,$3C,$8E,$2D,$BF,$1F,$C5,$FA,$09,$3E,$87,$AF,$91,$23
	.dc.b	$C2,$D9,$FF,$71,$47,$F5,$73,$FA,$F9,$5C,$F8,$7F,$E9,$78,$C4,$7C
	.dc.b	$06,$BF,$E1,$46,$BF,$4E,$38,$1F,$EA,$28,$7C,$6B,$E5,$30,$D8,$F1
	.dc.b	$B6,$BB,$35,$FF,$41,$FC,$1F,$F5,$A9,$FF,$3E,$36,$3F,$D1,$4D,$77
	.dc.b	$C7,$FD,$0E,$31,$1F,$EB,$3E,$A4,$62,$11,$89,$E0,$7F,$4B,$F8,$87
	.dc.b	$C4,$D2,$61,$FE,$8D,$8C,$B6,$33,$7F,$E9,$D9,$CB,$63,$FD,$8B,$D0
	.dc.b	$97,$D1,$61,$B7,$FB,$26,$DF,$EC,$9A,$FE,$F1,$AF,$FB,$3E,$BA,$57
	.dc.b	$32,$57,$E6,$25,$EA,$3E,$27,$CB,$1F,$E4,$E8,$C6,$3F,$ED,$A6,$D7
	.dc.b	$16,$CF,$E8,$9F,$CE,$71,$6F,$8F,$FC,$CF,$5B,$6D,$6D,$AF,$76,$BF
	.dc.b	$EE,$7A,$E7,$3C,$6B,$A1,$F6,$E6,$47,$E0,$9A,$7E,$08,$A7,$E2,$0A
	.dc.b	$7F,$9A,$BD,$9F,$E3,$4F,$FB,$D5,$A9,$A4,$EC,$FF,$7C,$36,$7F,$C8
	.dc.b	$AF,$25,$F4,$5F,$EF,$47,$F5,$AD,$AF,$D4,$35,$FD,$43,$5F,$CA,$7C
	.dc.b	$97,$E7,$DB,$C9,$FE,$1B,$8C,$47,$E5,$9B,$C4,$69,$27,$81,$F6,$7E
	.dc.b	$8B,$FA,$15,$7E,$01,$BA,$1A,$4F,$86,$E0,$4C,$BF,$01,$1C,$AB,$FC
	.dc.b	$A9,$F2,$DF,$DA,$1F,$F2,$3B,$5F,$F2,$13,$5F,$F2,$9B,$5F,$F2,$83
	.dc.b	$5E,$73,$FE,$61,$8C,$7A,$07,$86,$AF,$29,$F1,$98,$6D,$FE,$76,$75
	.dc.b	$39,$3F,$75,$F1,$D2,$7F,$70,$DC,$7F,$6C,$D7,$FD,$5F,$BF,$ED,$63
	.dc.b	$11,$F5,$BC,$4F,$A8,$4A,$FD,$65,$7D,$38,$C4,$FF,$46,$8C,$4E,$B8
	.dc.b	$3F,$C9,$71,$89,$FE,$97,$E4,$3C,$1F,$E9,$97,$FD,$23,$C3,$FE,$8E
	.dc.b	$47,$97,$CC,$FE,$14,$8E,$0F,$DF,$53,$E5,$1F,$45,$FE,$55,$6C,$E9
	.dc.b	$B8,$F8,$9A,$FF,$B4,$2B,$95,$71,$25,$71,$7A,$76,$C8,$E8,$DA,$F8
	.dc.b	$5A,$3E,$47,$F3,$84,$AF,$88,$F9,$7E,$5E,$47,$95,$B5,$CB,$E8,$BF
	.dc.b	$48,$49,$F9,$8D,$8D,$46,$08,$E0,$87,$DE,$B4,$F1,$8F,$81,$FB,$27
	.dc.b	$F9,$49,$BE,$9F,$F9,$F6,$31,$1F,$E7,$56,$BF,$B5,$F8,$8F,$DC,$13
	.dc.b	$0F,$C0,$3F,$A1,$E0,$7D,$2B,$E4,$F0,$DA,$FF,$07,$B6,$FF,$C1,$F5
	.dc.b	$C2,$B6,$4A,$DB,$67,$F9,$A6,$AF,$A4,$FE,$74,$DA,$FE,$74,$D9,$DC
	.dc.b	$F8,$BC,$59,$5E,$26,$CF,$8B,$1F,$E8,$E8,$FF,$11,$38,$6C,$70,$7F
	.dc.b	$A6,$5B,$E8,$FC,$64,$BF,$1B,$67,$D2,$D7,$8B,$F2,$7F,$5F,$7C,$3D
	.dc.b	$25,$7F,$8D,$A3,$2D,$8C,$D7,$CA,$9C,$B6,$3F,$AF,$7D,$27,$F8,$5A
	.dc.b	$BF,$AF,$36,$7F,$AF,$35,$FE,$CC,$D7,$E1,$0F,$A3,$78,$9F,$EF,$08
	.dc.b	$FF,$7D,$7D,$1E,$A7,$3E,$7F,$FB,$C6,$3F,$D7,$DF,$48,$F5,$7F,$96
	.dc.b	$78,$E9,$E1,$8F,$38,$C6,$3F,$47,$1E,$6F,$63,$13,$BA,$EB,$7D,$1E
	.dc.b	$78,$F8,$6B,$6F,$F8,$15,$F5,$57,$A5,$FF,$2F,$DF,$66,$C6,$5F,$67
	.dc.b	$AD,$FE,$63,$7F,$C1,$D3,$FE,$6D,$E5,$E8,$D8,$CC,$ED,$EA,$F1,$CB
	.dc.b	$DB,$1F,$2E,$31,$8F,$6B,$F4,$77,$7F,$61,$F0,$3E,$8E,$1C,$31,$EC
	.dc.b	$DC,$38,$E3,$F1,$DB,$7C,$63,$E1,$79,$5F,$C5,$7C,$BF,$84,$D8,$FD
	.dc.b	$3C,$63,$97,$FC,$B5,$1F,$E8,$4F,$9C,$7D,$FF,$FA,$21,$B1,$98,$C1
	.dc.b	$1C,$10,$FD,$EC,$70,$7F,$C5,$9F,$46,$FA,$0F,$1B,$63,$FC,$64,$D7
	.dc.b	$71,$A2,$3A,$6B,$FC,$2B,$DE,$7C,$AF,$FA,$79,$B1,$FE,$9E,$6B,$B8
	.dc.b	$D1,$1D,$35,$F3,$F8,$BE,$62,$3E,$67,$B7,$C2,$C8,$EC,$D7,$C8,$D7
	.dc.b	$FD,$6B,$48,$C4,$7F,$89,$BE,$8F,$F2,$DE,$67,$F8,$85,$BE,$41,$A7
	.dc.b	$E6,$F8,$5E,$33,$E9,$B0,$DC,$3F,$B3,$B5,$FE,$E6,$D7,$FA,$37,$A1
	.dc.b	$D8,$8F,$F7,$B6,$BF,$D7,$9A,$F0,$34,$43,$45,$3F,$B7,$79,$BF,$DD
	.dc.b	$DF,$EE,$8F,$A4,$C3,$FD,$D5,$F7,$FB,$12,$3F,$C1,$8D,$7F,$BD,$7A
	.dc.b	$9C,$48,$FF,$7F,$7A,$78,$12,$3F,$C0,$4D,$7F,$C0,$4D,$7C,$AD,$7F
	.dc.b	$CA,$FE,$C7,$D0,$23,$FE,$1D,$7D,$1E,$04,$8F,$F0,$73,$5F,$F0,$1F
	.dc.b	$BF,$F6,$08,$FF,$89,$5A,$E8,$C1,$0F,$F1,$67,$C8,$FD,$F7,$FC,$62
	.dc.b	$D7,$3B,$FE,$1A,$31,$1F,$E3,$C6,$BF,$EC,$36,$BF,$EC,$26,$F0,$9C
	.dc.b	$FA,$DF,$EF,$FE,$7F,$F4,$0F,$85,$FB,$1B,$FF,$46,$35,$FF,$46,$35
	.dc.b	$FE,$CD,$E5,$7F,$97,$DF,$EC,$5E,$47,$F2,$F6,$9F,$CC,$63,$44,$8D
	.dc.b	$36,$3F,$D6,$5E,$DB,$E5,$3F,$8B,$B6,$33,$18,$23,$82,$1F,$EB,$98
	.dc.b	$E0,$FF,$8D,$BE,$07,$F5,$64,$7F,$CB,$0D,$74,$60,$87,$FB,$37,$E8
	.dc.b	$9B,$FB,$5B,$1F,$D3,$5A,$EE,$3D,$BF,$35,$23,$E1,$35,$FF,$6C,$62
	.dc.b	$8C,$47,$F7,$03,$FC,$18,$E1,$FF,$55,$FC,$9F,$F0,$42,$3F,$E7,$16
	.dc.b	$BF,$E5,$46,$BF,$E4,$37,$FA,$0F,$C2,$BC,$0A,$49,$FF,$71,$46,$5B
	.dc.b	$19,$7F,$DC,$91,$96,$C7,$D6,$61,$F6,$8A,$7A,$FE,$09,$E0,$FF,$99
	.dc.b	$E3,$FD,$F1,$18,$F1,$BF,$BB,$36,$33,$3B,$79,$35,$FF,$0A,$C6,$DC
	.dc.b	$CF,$29,$D4,$7E,$8F,$B7,$F5,$04,$7F,$5A,$D7,$F6,$2D,$7F,$58,$FF
	.dc.b	$32,$F8,$B9,$1F,$E9,$F6,$BF,$E8,$26,$BF,$E7,$16,$BF,$E2,$AF,$8D
	.dc.b	$7D,$17,$E9,$5B,$1F,$E1,$56,$BF,$81,$0E,$1F,$C2,$7E,$3D,$FD,$D9
	.dc.b	$D3,$5C,$F9,$7F,$B1,$8C,$47,$F8,$E9,$AF,$F9,$75,$A9,$A6,$BC,$57
	.dc.b	$CE,$BE,$8F,$FA,$83,$6B,$FD,$66,$D7,$FD,$7E,$D7,$F3,$5F,$42,$FA
	.dc.b	$2F,$E0,$8D,$8F,$F1,$A3,$5F,$80,$D7,$FD,$6D,$2F,$8A,$F0,$92,$3C
	.dc.b	$2F,$93,$E8,$64,$79,$1A,$F4,$F1,$3C,$24,$7C,$2D,$74,$60,$87,$EC
	.dc.b	$FD,$87,$C0,$FF,$BF,$5B,$E0,$7F,$93,$5B,$3F,$65,$AF,$B3,$E0,$BE
	.dc.b	$09,$1F,$82,$D7,$46,$08,$7E,$0F,$E8,$3F,$C3,$2F,$F0,$56,$BF,$E7
	.dc.b	$66,$BF,$E9,$96,$AD,$3F,$87,$3F,$43,$B3,$E9,$91,$C1,$0F,$A6,$E5
	.dc.b	$FF,$80,$1F,$30,$FA,$0F,$F0,$BB,$63,$F3,$1E,$4A,$31,$1C,$27,$FC
	.dc.b	$31,$F2,$3F,$E6,$28,$FF,$79,$35,$FF,$58,$72,$23,$10,$8C,$7B,$DC
	.dc.b	$89,$1F,$A7,$8C,$B6,$78,$E0,$F6,$BF,$CF,$93,$FE,$BD,$8C,$6D,$FE
	.dc.b	$30,$9C,$CA,$72,$F5,$FA,$C7,$C6,$FF,$8C,$5B,$19,$8C,$11,$C1,$0F
	.dc.b	$E7,$98,$A7,$8C,$7F,$3F,$A7,$BD,$FD,$62,$3F,$D6,$31,$8F,$2F,$FC
	.dc.b	$8B,$39,$9F,$53,$5F,$8D,$F5,$9F,$E4,$D7,$E2,$35,$F0,$B5,$FF,$5E
	.dc.b	$E0,$8C,$47,$D2,$6B,$D3,$DA,$FF,$28,$46,$27,$84,$7F,$96,$3E,$85
	.dc.b	$F2,$F4,$97,$F8,$48,$CB,$F5,$FE,$7D,$F5,$1F,$E5,$B6,$C6,$63,$07
	.dc.b	$F9,$7F,$E8,$B8,$C7,$CA,$6B,$EA,$6B,$F8,$A6,$BF,$89,$6E,$26,$93
	.dc.b	$0F,$F3,$8C,$FF,$7A,$3E,$89,$4F,$C4,$14,$FC,$31,$75,$34,$9E,$7F
	.dc.b	$43,$FC,$49,$8B,$CE,$7D,$17,$0E,$87,$D0,$FF,$BC,$29,$E5,$7F,$BC
	.dc.b	$A7,$38,$F6,$9C,$BC,$6B,$86,$1D,$23,$F6,$51,$8C,$35,$1C,$70,$E9
	.dc.b	$3D,$DB,$1D,$6B,$96,$1E,$08,$61,$E9,$C3,$9C,$62,$F8,$B4,$CC,$ED
	.dc.b	$F1,$9D,$67,$33,$C2,$7D,$73,$A3,$BD,$F8,$B0,$F5,$46,$67,$78,$79
	.dc.b	$5B,$3A,$8C,$57,$C3,$AF,$F1,$65,$78,$DB,$19,$7A,$9A,$F4,$DF,$45
	.dc.b	$FD,$CE,$3E,$0C,$61,$FE,$ED,$48,$C4,$71,$9F,$F3,$13,$FB,$C3,$FC
	.dc.b	$C9,$C9,$82,$38,$3E,$00,$E9,$78,$3F,$3A,$3F,$1B,$36,$73,$1C,$0E
	.dc.b	$D9,$7D,$77,$95,$71,$73,$9F,$F5,$B8,$C5,$F3,$6C,$F8,$AF,$AD,$F3
	.dc.b	$8E,$27,$E7,$8E,$A7,$86,$B9,$34,$E5,$5D,$27,$C9,$5D,$2F,$F9,$97
	.dc.b	$A0,$F9,$6F,$5E,$1F,$C8,$23,$18,$7F,$4C,$6C,$7F,$27,$6B,$B9,$E2
	.dc.b	$F4,$6B,$FC,$EE,$36,$E6,$7F,$A4,$7D,$13,$E8,$B7,$CC,$8C,$4E,$A7
	.dc.b	$31,$8A,$FE,$76,$D7,$F9,$5B,$5F,$EB,$6D,$7F,$AD,$36,$BF,$C7,$BE
	.dc.b	$67,$83,$CF,$FF,$17,$61,$FD,$6D,$B7,$FD,$21,$AF,$F4,$48,$78,$FE
	.dc.b	$EE,$7F,$C3,$B6,$32,$7F,$74,$2C,$FF,$52,$FA,$27,$6F,$C0,$6C,$FF
	.dc.b	$6B,$6C,$EE,$74,$49,$D3,$63,$FA,$D7,$CD,$76,$7F,$AC,$B5,$FE,$E0
	.dc.b	$D7,$FB,$73,$73,$FE,$3F,$F4,$5F,$4B,$A4,$D2,$61,$FD,$FE,$32,$D8
	.dc.b	$CD,$FF,$6E,$9C,$BF,$DB,$73,$FD,$5D,$7F,$85,$9B,$19,$8F,$95,$18
	.dc.b	$8F,$EC,$BF,$45,$FD,$72,$7F,$C1,$CD,$8F,$EF,$6D,$77,$DB,$F8,$A8
	.dc.b	$C4,$7F,$87,$1A,$FF,$86,$DB,$89,$CD,$FF,$A9,$BC,$DF,$B9,$EB,$1A
	.dc.b	$4C,$7F,$D4,$FE,$87,$E7,$B1,$72,$FF,$99,$E7,$24,$8F,$F5,$7C,$EE
	.dc.b	$75,$18,$9D,$3F,$BD,$E8,$7D,$17,$EA,$23,$F8,$A3,$63,$3D,$1F,$AE
	.dc.b	$6D,$7F,$AE,$BE,$8B,$B1,$1C,$12,$FF,$D8,$31,$96,$C7,$F9,$A7,$DB
	.dc.b	$FE,$D6,$D8,$ED,$18,$7F,$CE,$8D,$7F,$DD,$1E,$FB,$E4,$3E,$23,$6B
	.dc.b	$F9,$D3,$67,$FD,$5C,$D7,$8C,$F0,$24,$F0,$7F,$82,$FC,$F3,$E2,$69
	.dc.b	$27,$FC,$A5,$19,$6F,$C3,$F0,$23,$83,$C7,$F1,$1E,$62,$38,$39,$F9
	.dc.b	$3E,$02,$38,$3F,$AA,$1F,$D4,$4B,$E3,$7E,$0D,$C6,$5F,$EC,$B6,$BF
	.dc.b	$CE,$4F,$E9,$B1,$B2,$47,$23,$87,$A4,$F7,$7F,$95,$C7,$F9,$D5,$B1
	.dc.b	$98,$DC,$7F,$13,$FA,$2F,$E8,$B7,$FE,$5A,$6B,$FE,$71,$6B,$FE,$6F
	.dc.b	$6F,$41,$A4,$E2,$E8,$7F,$1F,$FA,$27,$CD,$52,$63,$FE,$92,$8C,$B6
	.dc.b	$33,$87,$F9,$BE,$72,$D8,$FE,$8F,$CE,$97,$CF,$E1,$B7,$FE,$F5,$9D
	.dc.b	$4E,$49,$19,$9D,$DF,$CF,$AD,$CE,$6B,$FA,$AF,$AE,$F8,$8F,$1D,$FF
	.dc.b	$6D,$8C,$63,$A9,$CC,$FF,$93,$A7,$37,$F6,$E3,$FD,$3D,$C8,$E3,$EA
	.dc.b	$D9,$D5,$7E,$66,$72,$48,$CC,$EE,$BF,$0D,$5B,$9C,$D7,$DD,$99,$1F
	.dc.b	$EC,$0D,$7F,$65,$1F,$74,$C9,$FE,$9D,$72,$FD,$F3,$27,$E4,$FE,$2B
	.dc.b	$F8,$9B,$7B,$3F,$8E,$46,$23,$FC,$20,$D7,$FC,$DE,$D4,$D2,$47,$E6
	.dc.b	$3E,$89,$F2,$74,$98,$7E,$96,$32,$FF,$B6,$BE,$4F,$F3,$98,$7F,$AD
	.dc.b	$5A,$FF,$AD,$1A,$FF,$B2,$DA,$FF,$B2,$5B,$C6,$69,$39,$3F,$03,$F4
	.dc.b	$5F,$CF,$6F,$EA,$B6,$FE,$8B,$5F,$C2,$35,$FC,$13,$5F,$F6,$F3,$76
	.dc.b	$34,$D4,$7F,$B3,$BE,$47,$FA,$47,$A6,$69,$2F,$FC,$C1,$19,$6C,$66
	.dc.b	$7F,$21,$39,$7E,$D7,$94,$F8,$AF,$F1,$8F,$02,$31,$87,$E5,$E3,$13
	.dc.b	$A9,$CD,$7E,$D7,$DF,$FD,$F1,$2F,$F7,$8D,$8D,$46,$08,$FD,$7F,$13
	.dc.b	$FC,$49,$B7,$F5,$9D,$FE,$94,$60,$95,$F7,$DB,$5C,$63,$F9,$DF,$85
	.dc.b	$FD,$00,$8E,$0F,$C3,$7C,$CF,$62,$38,$3F,$69,$E5,$3E,$4D,$F4,$5F
	.dc.b	$E4,$E6,$CE,$9B,$87,$F3,$5A,$E1,$5B,$25,$6D,$B3,$BA,$FE,$1E,$FF
	.dc.b	$BF,$1A,$FF,$B5,$2B,$C5,$DB,$E1,$3E,$2F,$C7,$4C,$3F,$40,$DA,$E9
	.dc.b	$F1,$FE,$72,$4F,$9D,$F9,$BF,$45,$23,$F5,$0D,$9F,$D4,$35,$FD,$43
	.dc.b	$5F,$D4,$35,$FD,$43,$70,$7C,$B7,$F9,$14,$9E,$B7,$F9,$D9,$FF,$25
	.dc.b	$70,$7E,$B7,$E3,$3E,$F9,$31,$FD,$D3,$FE,$82,$E0,$7D,$03,$E8,$B0
	.dc.b	$DA,$FF,$01,$B7,$4F,$F0,$5B,$5F,$F5,$05,$72,$AE,$24,$AE,$2F,$F3
	.dc.b	$CD,$9F,$45,$49,$3F,$CC,$A3,$2F,$83,$C6,$F1,$91,$C1,$30,$F1,$C6
	.dc.b	$5B,$1C,$FE,$57,$C0,$47,$04,$8F,$E7,$74,$9D,$46,$1F,$8D,$F4,$4F
	.dc.b	$A2,$A4,$BF,$F1,$3C,$65,$B1,$9A,$FF,$18,$CE,$5B,$1F,$D7,$BE,$89
	.dc.b	$E7,$52,$5F,$F5,$48,$CB,$FE,$EF,$FA,$1F,$F3,$07,$9F,$FE,$E0,$F9
	.dc.b	$0F,$F6,$84,$63,$17,$86,$9F,$CF,$7D,$13,$D4,$FF,$72,$CE,$68,$2F
	.dc.b	$94,$66,$31,$ED,$7F,$BC,$5E,$B8,$78,$23,$13,$CF,$A9,$8F,$7C,$FC
	.dc.b	$D8,$76,$C7,$AE,$3C,$73,$F6,$65,$D6,$BF,$C6,$55,$EB,$6C,$65,$F7
	.dc.b	$FA,$EE,$6F,$88,$D8,$FF,$00,$57,$F9,$7E,$72,$48,$F4,$9E,$39,$E1
	.dc.b	$A7,$BE,$B8,$4E,$B4,$FE,$AD,$E0,$FF,$BF,$6B,$57,$C3,$3F,$1E,$3E
	.dc.b	$EC,$3E,$54,$E5,$61,$C9,$F8,$7E,$53,$97,$FC,$95,$1F,$A1,$F2,$FF
	.dc.b	$C3,$0A,$75,$18,$6C,$7F,$9D,$BE,$85,$F4,$5F,$E7,$A6,$C6,$63,$04
	.dc.b	$70,$43,$FD,$0B,$1C,$1F,$F0,$EF,$D1,$3E,$7F,$C0,$D8,$FF,$13,$B5
	.dc.b	$DC,$68,$8E,$9A,$FF,$A4,$BD,$FF,$F0,$7B,$FE,$8E,$6B,$9F,$4B,$FD
	.dc.b	$27,$18,$8F,$E1,$AD,$7F,$2F,$B2,$31,$1F,$83,$6C,$66,$7D,$3E,$34
	.dc.b	$AF,$4B,$5F,$27,$C2,$78,$08,$F9,$5A,$E8,$C1,$0F,$F1,$17,$D0,$FD
	.dc.b	$8F,$17,$FC,$38,$DF,$34,$D3,$EE,$F1,$7F,$3A,$FA,$2B,$E4,$1A,$4C
	.dc.b	$3F,$C6,$1E,$8F,$32,$38,$21,$FE,$33,$72,$D3,$27,$F8,$D5,$CB,$4F
	.dc.b	$ED,$DE,$6F,$90,$FE,$E8,$E0,$87,$F8,$FF,$D9,$F6,$11,$FF,$22,$3C
	.dc.b	$7D,$49,$1F,$DE,$5A,$FF,$73,$F0,$BB,$11,$FE,$FC,$D7,$46,$08,$7F
	.dc.b	$93,$FD,$9F,$98,$47,$FC,$A8,$FA,$7C,$09,$1F,$E0,$96,$BF,$E5,$7F
	.dc.b	$7B,$EA,$11,$FF,$2E,$35,$D1,$82,$1F,$EC,$5E,$07,$D1,$7D,$C8,$FF
	.dc.b	$47,$46,$38,$3F,$05,$39,$9E,$0D,$7F,$97,$89,$E0,$F2,$8D,$3F,$CD
	.dc.b	$23,$2F,$29,$FE,$0B,$EB,$FF,$21,$E8,$BC,$FF,$B3,$AF,$F4,$1B,$5F
	.dc.b	$F4,$1B,$5F,$E7,$FE,$5B,$C6,$FE,$1F,$1F,$D8,$5B,$19,$8F,$E2,$6D
	.dc.b	$7F,$D5,$8D,$7F,$DD,$0D,$7D,$AD,$7B,$C7,$F1,$9F,$71,$F2,$F4,$93
	.dc.b	$FE,$B2,$8C,$B6,$32,$FF,$AC,$23,$2D,$8F,$D7,$7C,$04,$8F,$F4,$B3
	.dc.b	$5F,$F4,$B3,$5F,$F4,$B3,$5F,$A7,$F4,$4D,$FE,$1B,$63,$F9,$9B,$5F
	.dc.b	$F6,$63,$5F,$83,$F3,$7F,$20,$8F,$C8,$6B,$F2,$1A,$F0,$3F,$D5,$9F
	.dc.b	$29,$FC,$1C,$8E,$08,$7F,$AB,$DC,$B4,$FF,$6C,$FC,$2B,$BF,$FA,$33
	.dc.b	$63,$31,$82,$38,$21,$FE,$DC,$8E,$0F,$CF,$3E,$C1,$0F,$AC,$7F,$56
	.dc.b	$F8,$3F,$E6,$5E,$45,$F0,$FF,$AE,$4F,$FB,$C6,$72,$D8,$CB,$FF,$00
	.dc.b	$5A,$FE,$92,$36,$E6,$7A,$1F,$AC,$ED,$91,$FA,$76,$BF,$AE,$6B,$FA
	.dc.b	$EE,$6F,$DB,$C6,$23,$FD,$A9,$F1,$5F,$CD,$48,$FF,$B7,$5A,$FF,$B7
	.dc.b	$5A,$FF,$B7,$5F,$D3,$FC,$6B,$E8,$A9,$27,$FB,$CC,$65,$FE,$F7,$F4
	.dc.b	$5F,$DF,$9F,$E0,$27,$F8,$C1,$C3,$86,$BF,$C1,$9A,$EF,$CC,$FD,$BC
	.dc.b	$62,$0D,$37,$E2,$BE,$85,$F2,$3F,$D4,$1B,$5F,$EB,$26,$BF,$EA,$D6
	.dc.b	$BE,$6F,$9F,$7D,$17,$F0,$66,$C7,$F8,$59,$AF,$B0,$38,$7F,$8F,$DF
	.dc.b	$8F,$E2,$BC,$44,$8F,$13,$E5,$FA,$29,$1E,$56,$BD,$69,$E2,$78,$C9
	.dc.b	$1E,$36,$BB,$6B,$EB,$7F,$C5,$BE,$C3,$E0,$BF,$8B,$37,$C0,$CC,$60
	.dc.b	$FE,$35,$5E,$CF,$83,$F8,$44,$7E,$13,$5F,$33,$5F,$94,$FF,$2E,$FA
	.dc.b	$2F,$F0,$BC,$1A,$4F,$4B,$E7,$77,$49,$F9,$CD,$7E,$13,$5F,$AC,$D7
	.dc.b	$FA,$47,$CC,$BE,$8B,$FC,$28,$D8,$FE,$AE,$D7,$FA,$E3,$5F,$AF,$F0
	.dc.b	$1F,$E8,$16,$FB,$7F,$E1,$88,$C4,$7F,$BA,$5A,$FF,$97,$9B,$C5,$FE
	.dc.b	$9C,$9E,$04,$9F,$EC,$FE,$D7,$F9,$ED,$FE,$BC,$D7,$FB,$23,$7B,$7F
	.dc.b	$CC,$5C,$1F,$E2,$8F,$A2,$7C,$65,$25,$FE,$47,$14,$EA,$30,$D7,$F6
	.dc.b	$31,$A9,$E3,$18,$9D,$3F,$C2,$7D,$EF,$E9,$ED,$F2,$BF,$C7,$D1,$88
	.dc.b	$FE,$EC,$D7,$FC,$FF,$F4,$4F,$6F,$FC,$7E,$D8,$F3,$35,$F1,$B5,$FD
	.dc.b	$3B,$5F,$F5,$AE,$28,$C4,$76,$9E,$A7,$F9,$47,$E8,$9F,$31,$49,$3F
	.dc.b	$E1,$68,$CB,$F1,$FE,$7D,$F4,$5F,$E5,$46,$C6,$63,$07,$F9,$63,$E8
	.dc.b	$B8,$C7,$CA,$6B,$FB,$E6,$BE,$86,$BF,$EA,$26,$E4,$69,$30,$FF,$35
	.dc.b	$4F,$F6,$D3,$FC,$44,$53,$FC,$50,$5E,$0F,$F6,$CF,$A1,$C0,$91,$FE
	.dc.b	$34,$34,$71,$18,$BC,$F7,$D1,$7F,$B8,$B9,$DF,$43,$FE,$36,$A7,$97
	.dc.b	$FE,$38,$9C,$E1,$FE,$E8,$9C,$BC,$6B,$86,$1D,$21,$5C,$1F,$0C,$66
	.dc.b	$31,$5C,$A7,$75,$E1,$BF,$03,$D3,$0E,$D3,$E1,$C3,$94,$EE,$32,$7B
	.dc.b	$AB,$8B,$BA,$CD,$FC,$16,$CE,$A7,$81,$AB,$EA,$F1,$3C,$A7,$C1,$8E
	.dc.b	$15,$E9,$C3,$D7,$5C,$2B,$DF,$7F,$ED,$CA,$F0,$47,$F4,$76,$BF,$D9
	.dc.b	$1A,$FD,$1F,$4D,$F4,$5C,$5B,$1C,$5A,$F1,$6B,$FD,$4F,$93,$FC,$BA
	.dc.b	$9A,$47,$FC,$C2,$9F,$F3,$EE,$7F,$DD,$BA,$3F,$CC,$5E,$DF,$D9,$9D
	.dc.b	$56,$E3,$13,$AA,$F0,$57,$23,$FC,$FF,$19,$7F,$C0,$35,$C6,$7B,$78
	.dc.b	$76,$72,$7C,$E3,$C9,$5F,$E7,$3D,$7C,$51,$EB,$34,$23,$55,$E2,$C3
	.dc.b	$CB,$7D,$6B,$FD,$23,$85,$78,$45,$7A,$4F,$07,$D4,$70,$29,$C0,$F7
	.dc.b	$70,$3D,$17,$9B,$FE,$89,$DF,$FD,$15,$39,$8C,$61,$FE,$9D,$F6,$3E
	.dc.b	$83,$F4,$6B,$77,$CA,$30,$7F,$A8,$DC,$61,$FB,$88,$C5,$F4,$8F,$F5
	.dc.b	$65,$F4,$74,$7F,$AB,$EB,$8D,$F9,$3A,$3F,$A5,$E1,$DA,$BF,$D6,$91
	.dc.b	$87,$8F,$3B,$E8,$5F,$80,$77,$D1,$7B,$C7,$A7,$7E,$63,$19,$7F,$3A
	.dc.b	$C3,$8B,$DE,$F8,$D7,$A6,$DF,$94,$BE,$63,$FD,$59,$18,$6C,$66,$FF
	.dc.b	$C6,$13,$97,$FC,$65,$E1,$3F,$36,$70,$7C,$58,$C3,$DE,$F9,$6F,$F1
	.dc.b	$A7,$37,$E5,$1E,$57,$FA,$F7,$4F,$F6,$43,$63,$27,$FB,$88,$B5,$FE
	.dc.b	$E7,$6B,$FE,$73,$6B,$FE,$DD,$6B,$FE,$E3,$6C,$FE,$DF,$BB,$E7,$72
	.dc.b	$FF,$5C,$9F,$25,$E2,$7E,$95,$7F,$5A,$9F,$F7,$9B,$6F,$FD,$F0,$D7
	.dc.b	$FD,$08,$D7,$FD,$DC,$D7,$FD,$E8,$D3,$FD,$ED,$E0,$7C,$17,$1D,$2F
	.dc.b	$1B,$F0,$4D,$BF,$B9,$3E,$1D,$FF,$43,$39,$C3,$F2,$73,$EF,$AF,$39
	.dc.b	$D4,$86,$09,$1E,$A7,$2F,$F8,$E3,$EA,$FC,$27,$C2,$8A,$7E,$6E,$DF
	.dc.b	$81,$BD,$1F,$C5,$B6,$7F,$12,$DB,$FF,$4A,$D7,$0A,$D9,$2B,$6D,$9F
	.dc.b	$F4,$DE,$1F,$CB,$7E,$57,$EB,$9F,$9B,$C3,$A6,$1F,$C0,$B7,$4F,$ED
	.dc.b	$FB,$9F,$47,$F9,$2C,$7F,$93,$B5,$FE,$42,$D7,$F8,$F3,$63,$8F,$33
	.dc.b	$4F,$F4,$94,$61,$B3,$B3,$96,$7F,$D0,$FE,$63,$F1,$A7,$8E,$3F,$47
	.dc.b	$9F,$9E,$3F,$86,$76,$3E,$C1,$F2,$5D,$F4,$FF,$25,$C6,$32,$FF,$29
	.dc.b	$36,$33,$3B,$7F,$D4,$71,$96,$C6,$5F,$F1,$FC,$65,$B1,$FB,$1F,$8F
	.dc.b	$7B,$9F,$C4,$5B,$7F,$C6,$70,$F3,$56,$6B,$F9,$25,$FF,$02,$8B,$D7
	.dc.b	$E5,$CF,$D5,$7E,$DF,$5B,$35,$F0,$3E,$8D,$F1,$BE,$5C,$56,$9C,$E7
	.dc.b	$FE,$71,$6C,$FF,$9B,$9A,$FF,$BC,$F4,$7C,$3F,$A4,$F9,$BB,$7F,$B3
	.dc.b	$AF,$F9,$24,$E9,$FE,$69,$DC,$FB,$07,$D3,$F8,$E3,$0D,$8C,$CF,$F2
	.dc.b	$26,$E3,$FE,$F0,$AD,$CE,$6B,$81,$1F,$F6,$D7,$23,$07,$CE,$7C,$F3
	.dc.b	$F9,$56,$3D,$38,$7F,$AE,$CE,$AF,$D3,$EB,$FD,$53,$FA,$37,$D2,$BD
	.dc.b	$BC,$4E,$89,$AF,$F0,$49,$D4,$63,$0E,$BF,$80,$7D,$F3,$9F,$85,$67
	.dc.b	$E3,$C5,$BE,$53,$6B,$F2,$DB,$FF,$3A,$F5,$FF,$A1,$78,$DF,$E9,$5C
	.dc.b	$7F,$B0,$60,$FD,$2B,$FD,$DB,$1F,$42,$75,$3F,$D1,$5B,$87,$46,$CF
	.dc.b	$F8,$A6,$79,$74,$F6,$C3,$F2,$DE,$2C,$AE,$91,$6B,$F9,$1F,$07,$F5
	.dc.b	$69,$FC,$0F,$5F,$C8,$F1,$7F,$3C,$C8,$C9,$FD,$A7,$81,$39,$BF,$E5
	.dc.b	$BE,$33,$E6,$7F,$BB,$36,$F8,$7A,$58,$99,$1F,$DC,$E7,$55,$F4,$EB
	.dc.b	$2F,$F8,$2A,$7F,$A0,$7A,$3F,$A3,$69,$F0,$63,$FC,$37,$C7,$FE,$16
	.dc.b	$7E,$24,$69,$D3,$5C,$C6,$DE,$33,$D6,$31,$1F,$D4,$BE,$8F,$D6,$7F
	.dc.b	$87,$EB,$A7,$17,$F8,$BB,$14,$FC,$0F,$B2,$95,$F1,$FC,$8B,$87,$C6
	.dc.b	$9B,$87,$F8,$DB,$6F,$E6,$B7,$F0,$E7,$E4,$FD,$24,$8F,$D9,$F8,$6F
	.dc.b	$95,$C6,$3F,$42,$AE,$3F,$90,$AD,$5F,$CD,$9F,$47,$D3,$FE,$7A,$FE
	.dc.b	$1B,$5F,$8E,$D7,$FC,$8A,$D7,$F9,$F3,$64,$D2,$76,$7F,$17,$F9,$5F
	.dc.b	$E8,$9E,$47,$F9,$15,$B5,$FE,$3F,$38,$CE,$5F,$E5,$77,$F8,$AF,$92
	.dc.b	$FC,$49,$FE,$29,$FA,$7F,$D2,$C6,$67,$F1,$13,$A7,$F8,$BC,$EB,$4F
	.dc.b	$A9,$3F,$92,$F9,$69,$5F,$99,$F5,$AE,$1F,$98,$9B,$7F,$4E,$6D,$7F
	.dc.b	$A5,$7C,$67,$13,$E7,$BF,$C6,$8D,$9F,$E3,$2D,$C7,$FB,$AB,$5F,$DB
	.dc.b	$79,$8F,$C9,$BF,$C2,$99,$7F,$42,$6D,$F7,$BE,$53,$9E,$9F,$D5,$FC
	.dc.b	$4C,$AF,$EA,$17,$C3,$1F,$F1,$3C,$E5,$FD,$9B,$FD,$AA,$35,$9B,$A9
	.dc.b	$3F,$C9,$BD,$0F,$59,$30,$F5,$BF,$E5,$19,$7D,$77,$EF,$0E,$13,$FB
	.dc.b	$97,$F9,$7F,$A7,$F1,$09,$1F,$11,$E5,$EE,$FF,$04,$37,$1F,$C0,$5C
	.dc.b	$34,$FE,$1D,$18,$CF,$F8,$6C,$DD,$7F,$51,$36,$BE,$45,$7D,$42,$3F
	.dc.b	$51,$F5,$7C,$9F,$F1,$43,$FA,$2B,$63,$FD,$09,$F0,$1F,$C0,$35,$34
	.dc.b	$9E,$0F,$F9,$CD,$B3,$FD,$61,$B5,$FE,$34,$D3,$FC,$E7,$97,$F0,$5F
	.dc.b	$2B,$FD,$31,$5F,$EA,$2B,$FC,$65,$69,$FF,$71,$79,$8F,$95,$FE,$B9
	.dc.b	$8F,$F6,$96,$BF,$CF,$E3,$5E,$8F,$F7,$39,$CC,$F2,$6A,$9D,$D7,$FA
	.dc.b	$5A,$73,$18,$9F,$47,$CE,$FF,$38,$3F,$D3,$13,$E2,$CF,$F8,$F4,$66
	.dc.b	$FF,$9C,$3A,$C3,$17,$FE,$9B,$BF,$F0,$53,$E9,$F5,$7C,$C4,$BF,$13
	.dc.b	$63,$51,$83,$FC,$07,$93,$E7,$BF,$A5,$9E,$4A,$B3,$FD,$AB,$E8,$9E
	.dc.b	$8F,$FB,$CF,$07,$B8,$F5,$7F,$B3,$56,$B9,$55,$A9,$BD,$AF,$A3,$7C
	.dc.b	$FF,$F9,$3F,$1E,$57,$C3,$1E,$4F,$13,$19,$78,$67,$70,$9D,$C7,$2C
	.dc.b	$3B,$5F,$5C,$7C,$B8,$71,$6E,$5F,$D7,$71,$EA,$D7,$A4,$FB,$DF,$EC
	.dc.b	$58,$F9,$1B,$EE,$7F,$B4,$9E,$51,$FE,$6A,$F4,$F9,$36,$33,$3B,$78
	.dc.b	$CF,$C6,$CF,$D7,$5B,$6D,$75,$7D,$9E,$A3,$E8,$FF,$C2,$91,$AB,$E5
	.dc.b	$9F,$D2,$BE,$0D,$8C,$E3,$FE,$5E,$C3,$93,$8E,$BF,$E3,$1F,$43,$FE
	.dc.b	$1F,$E8,$FE,$E0,$DE,$2F,$D5,$B6,$BF,$41,$F4,$6F,$23,$1D,$68,$C4
	.dc.b	$FF,$BE,$1A,$FF,$95,$DA,$FE,$BD,$AF,$ED,$9A,$F6,$9F,$A9,$3D,$5B
	.dc.b	$3A,$C7,$F9,$87,$D7,$79,$5B,$7F,$E2,$18,$F2,$CB,$CA,$D7,$52,$69
	.dc.b	$2B,$FC,$EB,$5B,$6D,$78,$08,$FB,$5A,$FB,$5A,$FB,$5F,$C6,$5F,$F4
	.dc.b	$5E,$6F,$09,$27,$F9,$2F,$C1,$7A,$7B,$CF,$06,$C7,$F2,$38,$C1,$1C
	.dc.b	$10,$FF,$6C,$C7,$07,$F9,$47,$1F,$C0,$23,$FC,$D3,$97,$C0,$77,$23
	.dc.b	$FC,$C5,$F4,$E9,$FC,$0A,$3F,$C9,$9C,$FF,$28,$F4,$76,$D8,$C9,$07
	.dc.b	$1F,$FB,$37,$4F,$E8,$33,$B9,$FF,$2E,$D5,$AD,$65,$FC,$DE,$7F,$A4
	.dc.b	$D7,$C3,$23,$FD,$15,$F2,$FC,$D7,$DC,$24,$7F,$50,$FA,$9F,$98,$7C
	.dc.b	$B2,$3F,$D3,$5F,$6E,$5F,$84,$24,$7F,$B4,$7E,$C1,$F3,$58,$8F,$F3
	.dc.b	$9C,$DC,$BF,$D9,$F5,$A9,$E1,$5F,$A7,$EE,$4E,$67,$E1,$F4,$7F,$24
	.dc.b	$8C,$5F,$C0,$BF,$C1,$1F,$81,$25,$7F,$63,$7E,$9D,$7E,$68,$91,$FD
	.dc.b	$BB,$E9,$6B,$FE,$88,$8B,$5F,$32,$7F,$4D,$E7,$3E,$C2,$92,$FF,$DA
	.dc.b	$BF,$4C,$F9,$AC,$47,$FA,$EE,$6E,$1F,$EB,$2A,$D4,$FF,$40,$86,$3F
	.dc.b	$D0,$BD,$6B,$E0,$7F,$09,$6D,$6E,$74,$7F,$25,$AE,$31,$FD,$2B,$E9
	.dc.b	$9F,$33,$FC,$46,$11,$87,$F9,$C3,$5F,$E7,$8D,$7F,$D3,$4D,$7F,$D7
	.dc.b	$8D,$7B,$47,$F4,$17,$AB,$63,$5C,$FF,$E5,$6F,$89,$F9,$24,$AF,$F7
	.dc.b	$DF,$D3,$7F,$07,$7F,$DC,$11,$96,$9A,$EE,$FA,$3F,$5D,$23,$FD,$80
	.dc.b	$FF,$01,$FA,$6C,$47,$FB,$A6,$2E,$7F,$A4,$9F,$B3,$F5,$52,$BF,$D9
	.dc.b	$0F,$C5,$F7,$9E,$A5,$3D,$8E,$27,$FA,$F3,$5F,$EC,$4D,$7F,$D8,$0D
	.dc.b	$7F,$DC,$4D,$7F,$BE,$B5,$39,$D3,$FC,$F7,$C9,$F8,$E2,$4F,$F9,$0B
	.dc.b	$E0,$6B,$FA,$21,$3C,$1F,$F7,$14,$FF,$23,$9F,$CB,$FD,$34,$9F,$F7
	.dc.b	$33,$FC,$9F,$E0,$F1,$1C,$EF,$FD,$17,$38,$D3,$F8,$5D,$7E,$83,$E4
	.dc.b	$E5,$FF,$BB,$DF,$EB,$3F,$87,$BE,$7A,$92,$7F,$2C,$7F,$8F,$E3,$4E
	.dc.b	$07,$CB,$FF,$BE,$1D,$F9,$C9,$DC,$7B,$E3,$15,$D6,$3E,$11,$CB,$E3
	.dc.b	$7F,$6B,$F4,$D2,$3C,$3A,$FF,$A9,$A3,$FA,$37,$4E,$67,$E0,$77,$A7
	.dc.b	$1B,$7F,$2A,$AF,$DF,$FC,$C4,$BE,$76,$8F,$EA,$1F,$3B,$99,$E7,$16
	.dc.b	$FF,$99,$CF,$F0,$8F,$B4,$95,$C9,$FF,$00,$FD,$9F,$FA,$86,$3E,$0B
	.dc.b	$5F,$F2,$19,$FE,$72,$70,$FF,$9B,$DA,$E7,$C8,$FE,$FF,$18,$8F,$F3
	.dc.b	$73,$77,$34,$D6,$BF,$EC,$AF,$93,$FE,$3A,$4A,$FE,$75,$E1,$FF,$33
	.dc.b	$FA,$9F,$8D,$1F,$E4,$1F,$29,$18,$8F,$38,$8C,$F0,$FF,$91,$5B,$1F
	.dc.b	$E4,$7F,$A9,$FE,$C8,$FD,$A6,$EF,$FC,$6A,$11,$8F,$17,$FC,$C7,$F0
	.dc.b	$0F,$95,$FE,$67,$F5,$12,$FE,$AB,$FD,$26,$5F,$2B,$4F,$13,$89,$F5
	.dc.b	$35,$F4,$E5,$FE,$7D,$9D,$92,$36,$D7,$DF,$BB,$CA,$7C,$27,$F8,$AD
	.dc.b	$B3,$F9,$26,$BF,$98,$FA,$97,$C9,$7F,$35,$6C,$7F,$9A,$DA,$FF,$9B
	.dc.b	$DA,$FC,$E6,$BF,$86,$6B,$FE,$8A,$6A,$69,$AE,$5F,$EB,$5F,$17,$FE
	.dc.b	$52,$F4,$3F,$DD,$75,$FB,$26,$BE,$7F,$96,$F3,$1F,$CF,$EC,$D2,$47
	.dc.b	$A5,$AF,$A5,$AF,$E7,$33,$FE,$33,$ED,$FF,$94,$53,$4F,$F2,$A6,$3F
	.dc.b	$E8,$BE,$4F,$F1,$13,$FE,$8D,$AE,$4D,$7F,$D5,$35,$FE,$CD,$F8,$6F
	.dc.b	$F2,$4C,$E6,$7F,$93,$B6,$75,$1F,$E9,$B6,$BF,$72,$35,$1F,$91,$8F
	.dc.b	$E5,$ED,$3F,$98,$FB,$7E,$68,$87,$39,$FE,$C9,$3F,$ED,$F8,$F1,$90
	.dc.b	$D1,$4F,$01,$FE,$60,$E5,$FF,$40,$E1,$FE,$41,$69,$EE,$2C,$7F,$42
	.dc.b	$FA,$2F,$E2,$8F,$F2,$06,$BF,$96,$6B,$F4,$5A,$FC,$86,$BC,$FC,$0F
	.dc.b	$F3,$8C,$62,$CD,$37,$83,$F3,$37,$3F,$FA,$71,$B8,$7F,$3C,$6B,$FD
	.dc.b	$67,$27,$E3,$5F,$80,$8F,$E8,$F1,$8D,$BF,$D9,$8D,$8F,$F6,$83,$5F
	.dc.b	$E1,$ED,$7F,$19,$DB,$FE,$2C,$7F,$AE,$36,$3F,$CD,$4D,$D7,$FD,$DB
	.dc.b	$F5,$3F,$D4,$6B,$7E,$77,$FA,$8E,$B5,$1F,$DB,$9B,$D4,$AE,$18,$7F
	.dc.b	$9D,$69,$E2,$7F,$AA,$DA,$7F,$52,$29,$FD,$BC,$FE,$BB,$1C,$63,$81
	.dc.b	$23,$87,$91,$FE,$39,$FA,$8F,$ED,$D4,$97,$C0,$3E,$73,$FA,$3D,$3E
	.dc.b	$A7,$FA,$4C,$E6,$B8,$4E,$5E,$51,$95,$1D,$9E,$B5,$D3,$0F,$F4,$9C
	.dc.b	$60,$E7,$7D,$A7,$C2,$D8,$EF,$3C,$2F,$B5,$F5,$F9,$EB,$7F,$D7,$9B
	.dc.b	$7E,$D7,$77,$E0,$6E,$3E,$8C,$7B,$D5,$3C,$65,$3D,$75,$E1,$8F,$54
	.dc.b	$E7,$0F,$85,$5E,$B3,$E0,$E1,$E9,$78,$CE,$EF,$E0,$C7,$F5,$EA,$F8
	.dc.b	$CD,$8C,$BD,$8F,$EC,$67,$CC,$C9,$F5,$38,$8C,$36,$BF,$12,$D9,$DC
	.dc.b	$FF,$9B,$7E,$3B,$FC,$DC,$9A,$47,$43,$3F,$B8,$BA,$FF,$D0,$B1,$88
	.dc.b	$E4,$6F,$4F,$EF,$33,$A7,$D3,$5F,$4D,$CF,$15,$AF,$EF,$F3,$5F,$CA
	.dc.b	$57,$17,$27,$53,$FC,$36,$E5,$FF,$0E,$C7,$17,$CD,$35,$F2,$DF,$AA
	.dc.b	$B8,$47,$E8,$6B,$84,$62,$37,$7E,$67,$C0,$76,$29,$D8,$F2,$74,$B8
	.dc.b	$BF,$59,$8F,$F4,$36,$CF,$F7,$E6,$BF,$C3,$8F,$F0,$B1,$FD,$41,$CB
	.dc.b	$4C,$9F,$C2,$4B,$3D,$DC,$F1,$3E,$A1,$F2,$D8,$6E,$1F,$E2,$F6,$D7
	.dc.b	$F8,$59,$AF,$F9,$1A,$B9,$57,$12,$57,$17,$BF,$B0,$F7,$29,$27,$FC
	.dc.b	$79,$19,$7F,$97,$FA,$AF,$89,$F5,$F2,$FF,$9E,$1B,$19,$9D,$C7,$6F
	.dc.b	$66,$47,$46,$E7,$FE,$2E,$C1,$18,$8F,$54,$62,$78,$1F,$DD,$47,$CA
	.dc.b	$FC,$A6,$99,$70,$7C,$CF,$DA,$FF,$C1,$71,$FE,$57,$6D,$7F,$8F,$9A
	.dc.b	$FC,$96,$BF,$CB,$75,$46,$23,$FC,$B8,$D8,$CF,$6F,$73,$EE,$F0,$3D
	.dc.b	$4A,$4C,$3F,$C1,$D1,$97,$FC,$D5,$D6,$F8,$1F,$C1,$B6,$33,$18,$23
	.dc.b	$82,$65,$FE,$58,$9E,$0F,$F9,$52,$93,$E4,$21,$E2,$3F,$2B,$0F,$AA
	.dc.b	$FE,$F3,$1F,$C5,$5B,$1F,$C6,$9A,$FF,$9F,$9A,$FF,$A5,$9A,$FF,$55
	.dc.b	$6B,$D7,$C3,$EF,$1A,$E2,$7D,$4B,$DD,$FC,$6B,$70,$FE,$FE,$DC,$3F
	.dc.b	$BF,$B5,$F1,$4F,$42,$4F,$A5,$B1,$E9,$6B,$C2,$3F,$9A,$F5,$3E,$A7
	.dc.b	$F3,$53,$FC,$89,$AF,$F1,$86,$BF,$C7,$9A,$FF,$73,$EA,$79,$5F,$31
	.dc.b	$B7,$FC,$FD,$B3,$FD,$11,$AF,$FA,$A1,$AF,$FB,$01,$AF,$59,$E8,$7F
	.dc.b	$08,$FA,$8F,$C9,$13,$8B,$06,$62,$5F,$FA,$DE,$79,$F1,$9E,$B7,$FC
	.dc.b	$86,$FD,$04,$AF,$43,$DF,$E3,$64,$73,$6C,$FF,$5E,$F2,$DF,$0D,$F8
	.dc.b	$C6,$D7,$F5,$F6,$CE,$E7,$44,$9D,$36,$3F,$D9,$7E,$63,$EA,$3F,$3B
	.dc.b	$1F,$EE,$48,$C7,$3F,$F2,$E6,$DF,$F7,$06,$CF,$FA,$DB,$DF,$FF,$57
	.dc.b	$CF,$F9,$FD,$B1,$9F,$53,$F9,$AC,$62,$3F,$D2,$CD,$7F,$D1,$AD,$5D
	.dc.b	$08,$C3,$67,$9F,$D0,$4B,$F0,$36,$FC,$8D,$7B,$B5,$FF,$80,$3F,$4F
	.dc.b	$F5,$89,$C5,$FB,$86,$D7,$F8,$2D,$B3,$FE,$10,$7F,$13,$97,$F2,$8F
	.dc.b	$7F,$E3,$77,$9A,$4B,$F8,$9F,$07,$F0,$88,$E0,$98,$7C,$B8,$CB,$63
	.dc.b	$E6,$79,$1F,$28,$8E,$09,$1F,$4E,$32,$D8,$FF,$12,$1F,$E6,$A2,$F9
	.dc.b	$7F,$C5,$9C,$47,$F8,$B1,$AF,$F9,$8C,$FE,$F3,$1B,$24,$72,$38,$68
	.dc.b	$F9,$EF,$EE,$B0,$8C,$3A,$8C,$C6,$2B,$F1,$5E,$B7,$E1,$89,$5F,$8B
	.dc.b	$6C,$6A,$30,$47,$F3,$7F,$0E,$F8,$7F,$CC,$36,$7F,$30,$D9,$DC,$E8
	.dc.b	$93,$A6,$C7,$F9,$3B,$DC,$FE,$48,$FC,$36,$BF,$E8,$76,$A6,$93,$C0
	.dc.b	$FE,$86,$DA,$F9,$DF,$4B,$F9,$F4,$D2,$72,$7E,$C7,$E9,$7F,$1E,$4E
	.dc.b	$5C,$12,$BF,$69,$19,$6C,$7E,$C3,$E9,$64,$7E,$D5,$B9,$7E,$E5,$AF
	.dc.b	$ED,$1A,$FF,$77,$FA,$4F,$EA,$3E,$09,$A4,$F2,$3F,$7B,$F4,$B3,$0F
	.dc.b	$E0,$4D,$BF,$C5,$0B,$E5,$C5,$FE,$36,$C3,$F0,$D7,$FE,$7C,$9C,$B6
	.dc.b	$33,$5C,$A7,$2D,$8C,$BF,$E3,$76,$BE,$9A,$ED,$CB,$2F,$A3,$67,$F8
	.dc.b	$D7,$D2,$7F,$16,$24,$FF,$1B,$7C,$DC,$92,$3C,$8D,$9F,$E4,$5E,$3F
	.dc.b	$F1,$E2,$4F,$F2,$36,$C6,$A3,$04,$70,$43,$F9,$2C,$70,$7F,$CA,$5F
	.dc.b	$94,$2F,$FC,$BA,$DA,$F9,$0D,$9F,$8C,$D7,$F9,$79,$FC,$C8,$91,$FC
	.dc.b	$DD,$FE,$7D,$F6,$AF,$9B,$A4,$9F,$E8,$11,$96,$C6,$67,$FA,$5C,$E5
	.dc.b	$B1,$FE,$74,$F7,$7F,$7D,$1F,$C2,$FC,$1F,$F7,$6F,$5E,$49,$87,$F7
	.dc.b	$9F,$B4,$F4,$57,$F8,$05,$AF,$FA,$2D,$BE,$59,$A4,$D3,$FD,$0B,$19
	.dc.b	$6C,$72,$F0,$9F,$38,$F8,$8F,$EB,$CD,$C7,$FA,$F3,$71,$FE,$CE,$D7
	.dc.b	$FB,$33,$50,$AE,$B5,$D0,$95,$ED,$E1,$F1,$91,$C1,$2B,$FC,$31,$82
	.dc.b	$75,$18,$7C,$FF,$4D,$2B,$CE,$D9,$FF,$72,$FD,$2F,$F8,$A3,$8F,$FD
	.dc.b	$79,$F4,$BF,$CE,$E3,$13,$FE,$DF,$8F,$F3,$A7,$D3,$3E,$87,$FC,$BB
	.dc.b	$39,$3D,$B5,$C3,$E0,$57,$CF,$BF,$FD,$C5,$5B,$15,$B9,$EB,$7D,$2F
	.dc.b	$8D,$74,$BE,$31,$88,$E3,$1D,$1F,$E1,$37,$E8,$6C,$65,$F1,$63,$FF
	.dc.b	$00,$65,$E9,$7F,$B0,$E7,$8B,$5F,$84,$D7,$DA,$D7,$D2,$FF,$B2,$EF
	.dc.b	$C9,$3D,$38,$FE,$5D,$70,$6F,$6B,$E9,$BF,$55,$3A,$39,$CD,$AE,$51
	.dc.b	$F1,$DB,$7F,$12,$F9,$74,$FF,$BE,$B5,$FC,$6F,$23,$E8,$DF,$4B,$FC
	.dc.b	$B3,$2C,$C6,$1F,$F1,$B3,$5F,$F1,$FB,$5F,$F3,$17,$D8,$BC,$6F,$F3
	.dc.b	$53,$63,$FC,$D4,$D7,$FC,$D4,$D7,$8C,$70,$23,$C1,$F2,$3F,$A3,$9F
	.dc.b	$F2,$CB,$E4,$F5,$33,$87,$AD,$B3,$A9,$F3,$C6,$08,$F1,$3C,$BE,$7F
	.dc.b	$A0,$8E,$08,$7A,$1C,$B4,$F9,$1D,$CE,$FF,$F7,$94,$61,$AF,$F5,$88
	.dc.b	$D7,$C0,$FB,$1B,$1F,$C9,$3C,$FF,$5F,$84,$69,$23,$44,$3F,$93,$B9
	.dc.b	$7B,$5F,$E2,$1B,$C1,$DE,$31,$1F,$CA,$5A,$FF,$BA,$FC,$3F,$9E,$FF
	.dc.b	$6D,$6F,$11,$A6,$B1,$F8,$1E,$0F,$52,$57,$C1,$6B,$E4,$6B,$E1,$6B
	.dc.b	$FC,$82,$39,$BC,$48,$7F,$77,$FA,$4F,$96,$FF,$92,$ED,$18,$8F,$E5
	.dc.b	$31,$8D,$3F,$12,$D8,$FA,$9F,$4B,$F0,$C8,$E0,$FF,$04,$57,$F3,$C6
	.dc.b	$E1,$CA,$31,$1E,$38,$C7,$D2,$C8,$E2,$D7,$D0,$D7,$D1,$28,$C4,$7C
	.dc.b	$1C,$F0,$FE,$26,$3E,$C1,$1F,$2F,$D0,$F8,$C8,$E0,$FC,$87,$D2,$FD
	.dc.b	$F2,$38,$21,$F8,$67,$2D,$3F,$D6,$9F,$4A,$F9,$1F,$D6,$B6,$33,$18
	.dc.b	$23,$82,$1F,$EC,$48,$E0,$FF,$0D,$FA,$57,$D0,$FE,$39,$B1,$FB,$E6
	.dc.b	$BB,$8F,$C9,$F8,$CF,$A9,$A4,$9F,$E0,$B1,$96,$C6,$5F,$E3,$91,$96
	.dc.b	$C7,$F0,$8F,$A6,$FE,$0D,$1F,$ED,$76,$BF,$C9,$9A,$FF,$73,$6F,$59
	.dc.b	$A4,$EA,$E8,$7E,$BF,$59,$5F,$C0,$DA,$FE,$D9,$AF,$EC,$DA,$FF,$A9
	.dc.b	$FE,$2D,$F4,$BF,$B4,$6C,$7F,$24,$6B,$FC,$E5,$AF,$F9,$9B,$EC,$7F
	.dc.b	$D5,$4F,$F2,$A6,$BF,$CF,$DA,$EF,$D1,$FE,$03,$18,$8F,$F5,$C3,$5F
	.dc.b	$F0,$AB,$53,$9E,$9F,$E4,$5E,$EF,$F0,$F2,$4F,$F2,$36,$BF,$C5,$9A
	.dc.b	$FF,$13,$7F,$8B,$7B,$B2,$3F,$95,$35,$FE,$3A,$D7,$F8,$D3,$5F,$EB
	.dc.b	$BF,$4C,$F9,$7F,$E8,$AD,$8F,$EB,$DB,$A3,$11,$FE,$8C,$8C,$4F,$F2
	.dc.b	$2F,$91,$7C,$0D,$25,$7F,$54,$8C,$B6,$32,$FF,$6B,$8C,$B6,$3F,$99
	.dc.b	$72,$48,$FE,$8C,$D7,$F9,$CB,$5F,$E6,$CD,$7F,$83,$78,$5F,$CE,$C8
	.dc.b	$FF,$4F,$6B,$FD,$09,$AF,$F3,$F7,$FB,$3F,$D2,$BE,$42,$92,$7E,$2C
	.dc.b	$65,$FE,$E3,$F9,$2B,$E7,$7F,$A2,$B6,$33,$18,$23,$82,$1F,$E0,$D8
	.dc.b	$E0,$FF,$3B,$F4,$3F,$96,$11,$FE,$62,$FF,$52,$FA,$47,$AB,$49,$3F
	.dc.b	$E0,$48,$CB,$FE,$BF,$FA,$37,$D2,$7F,$40,$6C,$66,$30,$67,$E9,$24
	.dc.b	$71,$6B,$FE,$23,$AF,$F2,$0B,$DF,$D7,$F2,$11,$C1,$0E,$0E,$5F,$F1
	.dc.b	$67,$D1,$7F,$87,$DF,$F3,$E8,$70,$FF,$A0,$1A,$E7,$D2,$F8,$5E,$0C
	.dc.b	$8F,$80,$D7,$CC,$D7,$FD,$CB,$7F,$C7,$DB,$1F,$DF,$2B,$FD,$05,$1F
	.dc.b	$E2,$C6,$BE,$6F,$77,$47,$D1,$C6,$47,$CF,$6B,$F1,$5A,$F8,$1A,$FF
	.dc.b	$7B,$E7,$FF,$36,$37,$DA,$FF,$10,$46,$21,$D9,$FE,$9F,$9F,$EE,$FF
	.dc.b	$64,$F8,$7A,$4A,$FF,$7E,$46,$5B,$19,$7F,$DF,$91,$96,$C7,$81,$87
	.dc.b	$F8,$E2,$FF,$C3,$47,$F8,$7B,$A7,$FC,$15,$0F,$43,$F8,$0F,$A3,$FE
	.dc.b	$0A,$9C,$C6,$27,$FC,$D0,$D7,$EE,$9C,$63,$64,$9F,$F2,$07,$C8,$FE
	.dc.b	$8A,$7F,$DB,$4D,$7F,$C4,$4D,$7E,$2B,$69,$F4,$1F,$B5,$9E,$F1,$8E
	.dc.b	$BF,$F0,$FC,$E6,$7F,$C4,$CD,$7F,$C4,$AD,$7D,$CD,$9E,$51,$EC,$8D
	.dc.b	$7C,$BB,$E8,$BF,$47,$3E,$E8,$C7,$37,$F9,$D9,$B1,$FE,$8B,$6B,$E6
	.dc.b	$FD,$B1,$F2,$9F,$E3,$46,$C7,$CA,$6B,$B8,$FF,$1C,$FD,$1F,$F2,$A8
	.dc.b	$FF,$20,$35,$FF,$1F,$B5,$FB,$0D,$C8,$D2,$6D,$D0,$FF,$39,$4F,$F4
	.dc.b	$23,$F6,$BC,$7F,$EF,$2F,$13,$24,$8F,$F3,$F9,$B3,$FC,$B0,$53,$F3
	.dc.b	$65,$8E,$4F,$FA,$6B,$5F,$EA,$99,$FF,$BF,$BE,$89,$F1,$7F,$C1,$25
	.dc.b	$F3,$BF,$C1,$A7,$36,$3F,$03,$19,$AE,$18,$B0,$E9,$18,$6D,$71,$C3
	.dc.b	$AB,$83,$99,$5F,$2A,$C3,$C2,$FE,$42,$FB,$39,$BF,$46,$1E,$38,$EC
	.dc.b	$D7,$86,$1D,$6B,$8D,$F4,$C3,$C7,$5C,$CE,$26,$EA,$39,$3C,$E5,$8E
	.dc.b	$33,$CE,$BA,$E1,$F0,$6B,$FD,$B9,$5E,$C7,$F6,$AD,$74,$63,$C5,$7D
	.dc.b	$1F,$D7,$6C,$72,$6B,$B8,$FF,$53,$AF,$85,$A4,$7F,$D5,$6D,$C7,$E5
	.dc.b	$C3,$A9,$F1,$5F,$4E,$3E,$F5,$59,$CC,$7C,$13,$25,$73,$5D,$23,$C4
	.dc.b	$66,$BA,$38,$8F,$0C,$F0,$CB,$FE,$00,$5F,$7B,$DB,$6F,$C0,$71,$9F
	.dc.b	$E3,$B5,$C6,$3F,$8F,$CE,$EF,$C4,$DA,$FF,$5E,$78,$BF,$99,$C3,$FD
	.dc.b	$CE,$DB,$FF,$72,$35,$FD,$8B,$5E,$27,$C2,$32,$53,$31,$D1,$C1,$FC
	.dc.b	$E7,$E8,$DF,$47,$8E,$44,$62,$BF,$9C,$B6,$BF,$95,$B5,$FF,$67,$35
	.dc.b	$FF,$66,$B5,$ED,$5D,$63,$14,$F4,$9F,$47,$E1,$6E,$1F,$D2,$1B,$87
	.dc.b	$F3,$A6,$BF,$E0,$6E,$27,$9C,$EE,$F6,$C6,$1B,$19,$BF,$E6,$0D,$AE
	.dc.b	$9F,$47,$E1,$24,$F8,$5B,$1A,$8C,$11,$FE,$B7,$CF,$F9,$6F,$9A,$F9
	.dc.b	$AD,$9F,$EB,$F3,$FD,$E9,$4E,$9A,$7F,$66,$FA,$37,$77,$D3,$6C,$66
	.dc.b	$30,$47,$04,$AF,$EB,$F3,$C1,$F2,$DF,$F9,$6E,$7C,$6F,$F8,$33,$C0
	.dc.b	$F3,$B9,$9D,$C7,$E0,$FA,$FF,$A9,$D7,$F7,$A6,$CF,$F6,$C6,$BF,$E3
	.dc.b	$8E,$0F,$E6,$BD,$DC,$49,$5C,$5B,$3E,$48,$F0,$90,$F2,$E1,$FC,$01
	.dc.b	$F9,$30,$FA,$37,$CC,$7E,$6A,$30,$D8,$CC,$F0,$6C,$FD,$26,$BF,$E1
	.dc.b	$26,$BF,$E5,$06,$BF,$D2,$DB,$C2,$FA,$37,$CE,$FE,$39,$B7,$FD,$ED
	.dc.b	$B7,$FE,$01,$6B,$DE,$7C,$C4,$9F,$33,$63,$CC,$D7,$84,$7E,$2B,$B1
	.dc.b	$F4,$7F,$99,$9F,$F3,$3C,$63,$97,$F7,$AD,$AE,$15,$BA,$FF,$36,$F5
	.dc.b	$7F,$A2,$27,$E6,$B6,$3F,$2E,$D7,$FC,$6E,$D7,$8F,$A5,$F4,$63,$11
	.dc.b	$FD,$C5,$A9,$CF,$03,$E8,$FF,$29,$3F,$87,$6B,$FD,$E6,$35,$CB,$FC
	.dc.b	$4E,$73,$3F,$E4,$F7,$99,$2F,$47,$8E,$66,$1E,$C6,$D7,$F9,$0A,$9F
	.dc.b	$47,$D4,$95,$FE,$56,$8C,$BF,$90,$F2,$5F,$01,$F8,$F6,$C6,$63,$04
	.dc.b	$70,$4B,$FE,$03,$3C,$1F,$F7,$57,$D1,$FE,$A1,$FF,$0C,$79,$88,$C4
	.dc.b	$23,$1C,$FF,$CE,$1B,$7F,$BA,$FA,$37,$AB,$FC,$41,$B1,$98,$C1,$1C
	.dc.b	$12,$FF,$8C,$CF,$06,$FC,$EF,$32,$4F,$76,$C6,$A3,$04,$7F,$8F,$FD
	.dc.b	$1F,$E0,$89,$C1,$FB,$76,$CF,$EF,$1B,$3C,$27,$DD,$DE,$F9,$3A,$4B
	.dc.b	$FE,$57,$9F,$F2,$A8,$F8,$1E,$74,$9F,$88,$DC,$3D,$AD,$7F,$CB,$39
	.dc.b	$A3,$11,$FE,$49,$EE,$4F,$F9,$42,$BE,$77,$91,$F3,$48,$E0,$95,$F2
	.dc.b	$A3,$2F,$F4,$08,$79,$9F,$C5,$DA,$7F,$1C,$2C,$6E,$34,$48,$D3,$63
	.dc.b	$FD,$53,$0D,$9F,$07,$FD,$F2,$30,$D8,$CC,$78,$DB,$3F,$80,$F5,$BF
	.dc.b	$10,$49,$FC,$63,$63,$51,$82,$3F,$EB,$AD,$5F,$04,$F8,$2F,$CB,$36
	.dc.b	$BE,$BB,$6B,$83,$5F,$F7,$FD,$72,$AE,$24,$AE,$2F,$F3,$FF,$A3,$FE
	.dc.b	$57,$5F,$E7,$F0,$E1,$FF,$41,$35,$CF,$83,$C4,$95,$FD,$32,$74,$71
	.dc.b	$5F,$47,$FA,$3A,$ED,$18,$E6,$F9,$D3,$99,$FF,$1C,$35,$FD,$C7,$D1
	.dc.b	$FF,$24,$C8,$D2,$73,$7E,$ED,$B8,$7E,$71,$B5,$FB,$4F,$67,$F7,$84
	.dc.b	$9F,$E0,$0D,$8D,$46,$08,$FF,$B8,$3E,$8B,$FA,$DF,$82,$69,$3C,$9F
	.dc.b	$E1,$1F,$47,$2F,$F8,$63,$6F,$97,$1F,$F9,$16,$FF,$21,$5F,$EF,$C6
	.dc.b	$CE,$A3,$13,$B2,$38,$38,$73,$73,$23,$83,$F8,$E7,$D1,$FF,$1B,$23
	.dc.b	$83,$C1,$CB,$E3,$23,$83,$F9,$27,$8F,$FC,$8C,$8E,$09,$87,$F2,$08
	.dc.b	$CB,$63,$2F,$F0,$68,$CB,$63,$FC,$77,$A3,$E5,$D8,$7F,$9A,$5B,$3A
	.dc.b	$6D,$7E,$5C,$57,$1A,$FE,$71,$12,$BF,$9E,$34,$FD,$77,$CC,$BE,$67
	.dc.b	$FA,$13,$67,$F9,$1B,$67,$73,$A2,$4E,$9B,$1F,$E5,$8F,$96,$FE,$06
	.dc.b	$FF,$BF,$1A,$FF,$23,$6A,$EE,$46,$3C,$A7,$D1,$7B,$AB,$FC,$1A,$D7
	.dc.b	$FC,$08,$D7,$FC,$08,$D8,$E3,$97,$FA,$51,$B8,$FF,$84,$A7,$47,$1E
	.dc.b	$77,$CC,$BE,$8F,$FB,$23,$6B,$FA,$2B,$71,$FE,$2C,$D7,$F8,$7B,$5F
	.dc.b	$EE,$2D,$7F,$B8,$36,$BB,$D7,$96,$DC,$34,$98,$79,$DB,$87,$F8,$A2
	.dc.b	$77,$39,$24,$79,$3E,$53,$D2,$48,$F4,$BF,$E5,$DF,$A0,$75,$FF,$B8
	.dc.b	$EB,$FD,$D5,$F4,$1F,$C8,$67,$3C,$BF,$EE,$48,$FF,$15,$7D,$1B,$E8
	.dc.b	$FF,$DC,$BF,$02,$9E,$18,$F3,$8C,$62,$18,$F4,$7B,$99,$71,$1C,$39
	.dc.b	$D8,$F7,$D3,$C5,$7B,$C3,$B6,$3E,$0C,$FF,$CF,$79,$75,$AF,$D2,$35
	.dc.b	$D1,$FE,$A8,$7F,$D4,$7C,$5C,$3D,$3F,$F2,$64,$E5,$AF,$26,$BF,$15
	.dc.b	$DE,$9E,$E8,$F6,$C6,$30,$D5,$7A,$34,$7D,$1F,$EB,$EF,$D6,$70,$9B
	.dc.b	$8F,$92,$3E,$29,$93,$E3,$D7,$2E,$8F,$EA,$D9,$7C,$EE,$47,$D0,$BE
	.dc.b	$8F,$FD,$3B,$18,$6C,$67,$0F,$F0,$FB,$5F,$F3,$9F,$D1,$FF,$90,$1F
	.dc.b	$F3,$BB,$5C,$F8,$DF,$E7,$C8,$C4,$7F,$A0,$1A,$FF,$9D,$DA,$FF,$9D
	.dc.b	$DB,$A1,$CE,$1E,$8F,$57,$CA,$4A,$F2,$B5,$FF,$02,$61,$FE,$68,$72
	.dc.b	$43,$D3,$E8,$79,$48,$FA,$9A,$E8,$C1,$0F,$E3,$1E,$CF,$FB,$13,$BF
	.dc.b	$F8,$9B,$7B,$CD,$3E,$6F,$81,$FE,$32,$AF,$E4,$FE,$7E,$08,$E0,$87
	.dc.b	$F3,$07,$2F,$F2,$7E,$1E,$C4,$7F,$DD,$77,$F8,$77,$F9,$F7,$17,$F4
	.dc.b	$57,$FD,$B2,$D7,$FA,$21,$8F,$0B,$FD,$BA,$FF,$2D,$9F,$E3,$AF,$4F
	.dc.b	$07,$D8,$47,$04,$3F,$9D,$39,$69,$93,$F9,$E3,$97,$FD,$ED,$F4,$5F
	.dc.b	$19,$FE,$B6,$D7,$D2,$D7,$79,$FC,$DF,$A3,$91,$F0,$1A,$F0,$AF,$7D
	.dc.b	$7A,$A3,$07,$68,$CF,$D1,$C8,$F0,$B5,$F0,$B5,$F3,$B5,$FE,$A8,$FD
	.dc.b	$F2,$3F,$49,$F0,$FC,$F4,$8F,$6B,$5F,$EC,$1F,$47,$F8,$72,$3F,$69
	.dc.b	$AE,$8C,$10,$FF,$30,$7D,$13,$E8,$FF,$72,$D8,$FF,$4A,$B5,$DC,$68
	.dc.b	$8E,$9A,$FF,$81,$47,$D1,$BE,$7F,$F0,$AD,$8D,$1A,$8C,$9F,$8A,$F1
	.dc.b	$DF,$33,$49,$3C,$23,$2D,$8C,$BE,$18,$CB,$63,$F8,$AF,$D1,$FE,$EE
	.dc.b	$3F,$D0,$0D,$7F,$CF,$2D,$7F,$DB,$6D,$7F,$DB,$4D,$EB,$34,$9D,$7F
	.dc.b	$B9,$D6,$57,$EB,$1A,$FF,$85,$DA,$FF,$85,$DA,$FF,$A0,$FE,$35,$F4
	.dc.b	$7F,$AC,$6C,$7C,$86,$BF,$3D,$AF,$F8,$D7,$E8,$FF,$D5,$6F,$CE,$6B
	.dc.b	$F5,$5A,$EF,$DD,$FD,$CC,$62,$3F,$D2,$8D,$7F,$B6,$35,$39,$ED,$FF
	.dc.b	$15,$FB,$BF,$C6,$49,$3F,$C3,$1A,$FF,$8D,$DA,$FF,$8D,$DF,$E4,$1E
	.dc.b	$EC,$8F,$E2,$6D,$7F,$C7,$CD,$7F,$C7,$CD,$7F,$C0,$1A,$BE,$8D,$F2
	.dc.b	$FF,$D5,$1B,$3F,$EF,$06,$BF,$EC,$16,$BF,$E5,$DF,$91,$FF,$76,$3F
	.dc.b	$96,$6B,$F9,$D6,$BB,$F8,$2F,$E4,$31,$88,$FF,$5E,$35,$FF,$0A,$35
	.dc.b	$39,$CB,$FC,$A7,$C9,$FD,$04,$93,$FC,$D1,$AF,$F9,$7D,$AF,$F9,$7D
	.dc.b	$FE,$91,$E1,$48,$FE,$76,$D7,$FC,$CC,$D7,$FC,$CC,$D7,$E4,$FD,$1B
	.dc.b	$E4,$3F,$BD,$B6,$3F,$79,$A2,$31,$1F,$EE,$96,$BF,$E8,$5F,$A2,$7D
	.dc.b	$1F,$F4,$46,$C7,$F0,$06,$BF,$C1,$DA,$F1,$8E,$04,$78,$3F,$CD,$3C
	.dc.b	$B9,$1F,$D0,$1A,$FF,$11,$FA,$1F,$F2,$7B,$FC,$39,$AE,$7D,$5F,$EA
	.dc.b	$F1,$88,$FF,$79,$79,$0F,$A3,$7D,$1F,$F6,$06,$CF,$FA,$A1,$AF,$F6
	.dc.b	$66,$BC,$97,$D1,$71,$24,$71,$7C,$3E,$BC,$8F,$33,$5F,$53,$5F,$F5
	.dc.b	$86,$08,$C4,$7F,$97,$C7,$D1,$FF,$88,$9B,$E9,$FF,$74,$8C,$19,$AF
	.dc.b	$E4,$B3,$D3,$C2,$F8,$64,$7E,$43,$5F,$98,$D7,$FD,$85,$3F,$E8,$A7
	.dc.b	$F9,$DB,$5F,$E8,$4D,$77,$EF,$7D,$0C,$64,$7D,$56,$BF,$65,$AF,$C0
	.dc.b	$6B,$FE,$7D,$E6,$71,$FF,$A2,$E3,$FD,$A9,$18,$F6,$BF,$C5,$93,$99
	.dc.b	$F8,$D3,$97,$FD,$C3,$F4,$6F,$87,$FF,$0E,$B6,$33,$18,$23,$82,$1F
	.dc.b	$E2,$68,$E0,$F8,$AB,$FC,$47,$5E,$68,$F1,$90,$FF,$78,$F5,$FF,$85
	.dc.b	$5F,$F4,$C7,$12,$31,$1F,$EE,$E6,$C6,$7D,$0F,$F1,$9C,$60,$8E,$0F
	.dc.b	$F2,$67,$C8,$FE,$B2,$BE,$43,$5F,$F7,$13,$6C,$E6,$DF,$41,$FB,$CA
	.dc.b	$F2,$B5,$FF,$80,$2D,$8D,$F5,$FF,$91,$A3,$04,$70,$7F,$80,$3E,$5D
	.dc.b	$F4,$7F,$9B,$BF,$F0,$44,$63,$9B,$53,$99,$FC,$F0,$9D,$3F,$E0,$FC
	.dc.b	$1F,$46,$F9,$4F,$F2,$BB,$67,$4D,$7F,$CB,$AD,$7F,$80,$FD,$1F,$F3
	.dc.b	$7A,$34,$9B,$FF,$98,$9B,$3E,$16,$BF,$92,$3F,$A5,$9F,$97,$28,$DB
	.dc.b	$FD,$AD,$E2,$FF,$A0,$63,$31,$87,$FC,$54,$D7,$A4,$6C,$FF,$37,$ED
	.dc.b	$FD,$7B,$3F,$F7,$07,$D1,$3E,$2F,$F8,$24,$BE,$77,$F8,$34,$E7,$0F
	.dc.b	$F7,$34,$E5,$E3,$5C,$30,$F9,$D0,$AE,$0F,$E1,$23,$31,$8A,$E5,$3B
	.dc.b	$AF,$0D,$F8,$1E,$98,$76,$9F,$0E,$1C,$A7,$71,$93,$F2,$35,$C5,$DD
	.dc.b	$66,$FE,$04,$77,$3D,$73,$B7,$89,$BB,$EC,$F2,$38,$1A,$2C,$FA,$B0
	.dc.b	$F6,$4F,$09,$F8,$17,$FE,$58,$AF,$0C,$7E,$51,$AF,$B8,$FF,$31,$2F
	.dc.b	$15,$F4,$7C,$5B,$3A,$6B,$CD,$AF,$F9,$AF,$E1,$3F,$D5,$A9,$A4,$7F
	.dc.b	$D6,$29,$7B,$BF,$61,$75,$BE,$23,$53,$99,$E8,$70,$DB,$F5,$F5,$B8
	.dc.b	$C1,$FB,$2B,$E3,$3A,$BB,$39,$36,$78,$EB,$95,$F7,$BF,$BD,$3E,$4A
	.dc.b	$AF,$8B,$5F,$45,$F8,$5E,$11,$D0,$E0,$53,$D3,$7C,$2B,$C7,$1F,$D9
	.dc.b	$6E,$BC,$27,$99,$B7,$DB,$BB,$C6,$EB,$B7,$C1,$8C,$65,$F0,$E3,$12
	.dc.b	$F8,$DC,$65,$FD,$56,$38,$B6,$B3,$87,$F0,$19,$CB,$6B,$38,$76,$BF
	.dc.b	$B5,$87,$87,$A9,$E0,$69,$B3,$A3,$FA,$4F,$D1,$3E,$BF,$55,$27,$59
	.dc.b	$7F,$40,$8F,$1F,$42,$F8,$56,$B2,$FE,$87,$5E,$48,$FF,$49,$9F,$C5
	.dc.b	$0B,$7F,$DC,$1A,$FF,$6E,$6B,$FD,$DF,$77,$D4,$7B,$9B,$3D,$CE,$7A
	.dc.b	$7C,$F8,$CD,$7F,$3F,$9F,$F0,$07,$DB,$F4,$8F,$E5,$4D,$C4,$D2,$7A
	.dc.b	$FF,$99,$9F,$F0,$0C,$62,$BF,$96,$BC,$78,$5E,$0F,$D2,$6C,$FC,$39
	.dc.b	$F2,$BF,$E1,$18,$CD,$79,$4F,$EA,$9E,$8F,$F2,$79,$FE,$B6,$D8,$FE
	.dc.b	$B3,$9F,$FA,$03,$97,$C9,$1F,$7A,$74,$FF,$8E,$E7,$55,$CA,$B7,$3D
	.dc.b	$9B,$3A,$8F,$EC,$BE,$1B,$F5,$E7,$B5,$F9,$08,$C3,$F9,$2A,$E0,$7E
	.dc.b	$F6,$33,$87,$E1,$A3,$FD,$2B,$DD,$DB,$2F,$E9,$53,$EC,$8C,$11,$C1
	.dc.b	$31,$F7,$CF,$07,$CB,$8F,$E3,$9F,$91,$87,$9F,$97,$C9,$5B,$C7,$C9
	.dc.b	$7B,$7F,$65,$C1,$FC,$67,$1F,$EC,$AD,$C7,$FC,$82,$D7,$FC,$7E,$D7
	.dc.b	$FC,$90,$D7,$FB,$93,$4F,$EE,$91,$F3,$9A,$FF,$72,$D3,$FC,$68,$FE
	.dc.b	$95,$B3,$F8,$19,$DB,$63,$35,$FC,$27,$E2,$9E,$87,$56,$E1,$FD,$F9
	.dc.b	$B7,$FE,$5A,$6B,$FE,$58,$6B,$FE,$61,$6B,$FE,$0B,$6E,$06,$F5,$FF
	.dc.b	$B8,$AB,$F7,$8D,$AF,$44,$F8,$30,$FE,$3D,$39,$7F,$8F,$7D,$17,$F9
	.dc.b	$96,$FF,$C6,$AD,$BF,$CD,$B5,$FC,$C3,$5F,$E3,$2D,$7F,$C7,$0D,$93
	.dc.b	$4D,$74,$FF,$85,$FC,$3F,$E2,$BC,$3F,$2B,$0E,$7E,$56,$27,$F8,$CF
	.dc.b	$DC,$BD,$0A,$63,$E6,$2D,$FE,$6F,$24,$7E,$A1,$B1,$EC,$8C,$1F,$81
	.dc.b	$D7,$F2,$51,$FE,$8F,$6E,$AD,$18,$BE,$5B,$F9,$CD,$6A,$DF,$35,$FD
	.dc.b	$15,$FF,$2C,$E1,$FD,$C6,$30,$FF,$97,$23,$7E,$17,$EF,$9A,$70,$2C
	.dc.b	$6E,$78,$BF,$EC,$18,$CB,$63,$2F,$FA,$9F,$F3,$97,$BF,$FD,$39,$B5
	.dc.b	$FD,$10,$EB,$C5,$FD,$7E,$33,$5F,$9D,$C5,$F0,$1E,$E3,$FA,$BF,$23
	.dc.b	$E8,$F1,$8F,$E3,$A6,$DF,$E3,$A6,$E3,$D7,$C0,$F1,$1F,$D8,$3B,$5F
	.dc.b	$0D,$8C,$7F,$A9,$69,$FC,$BA,$F8,$47,$F5,$5A,$DB,$AB,$FE,$3D,$FA
	.dc.b	$3F,$D6,$93,$F6,$29,$FA,$73,$E3,$37,$D7,$31,$FE,$CB,$87,$08,$FE
	.dc.b	$4D,$5B,$9C,$E1,$F8,$2E,$CF,$F1,$6F,$C4,$7E,$00,$E5,$BF,$C7,$7E
	.dc.b	$77,$C8,$4B,$EB,$F6,$1F,$D2,$AB,$FD,$05,$BA,$B5,$1F,$C2,$E3,$15
	.dc.b	$C3,$C8,$CE,$FF,$B6,$F9,$CF,$C7,$9F,$4B,$DA,$B5,$FB,$E8,$B8,$7E
	.dc.b	$FA,$6D,$7F,$8E,$EB,$F2,$BE,$17,$E3,$FA,$BF,$6D,$18,$3F,$20,$E0
	.dc.b	$91,$B2,$47,$FA,$5B,$2F,$8D,$87,$A7,$8F,$FC,$3F,$3C,$E3,$0F,$F1
	.dc.b	$2F,$99,$FD,$39,$F8,$A6,$99,$24,$7F,$16,$72,$FC,$5F,$92,$FC,$D1
	.dc.b	$FD,$77,$E9,$FF,$77,$19,$9F,$C4,$CE,$9F,$BB,$3A,$CF,$F3,$D3,$FD
	.dc.b	$97,$E3,$9F,$39,$FD,$BB,$A1,$18,$C3,$F7,$31,$89,$D4,$E7,$1F,$EE
	.dc.b	$2D,$7F,$CE,$42,$78,$57,$29,$D4,$FF,$9D,$E7,$2F,$FB,$7B,$97,$FC
	.dc.b	$F3,$F6,$DF,$C3,$0F,$36,$17,$0F,$D2,$FC,$A7,$EE,$CF,$D5,$7D,$5D
	.dc.b	$8C,$93,$0F,$EA,$D8,$A7,$51,$FB,$DF,$33,$F8,$C9,$FE,$52,$1F,$5F
	.dc.b	$8B,$FE,$93,$16,$7F,$AF,$4D,$DB,$FC,$CD,$3F,$EF,$1A,$FE,$D1,$F0
	.dc.b	$3F,$DA,$E3,$17,$ED,$6B,$F1,$BC,$9F,$E5,$6F,$F3,$C3,$C5,$36,$BF
	.dc.b	$AC,$FB,$5F,$CB,$8C,$4F,$F1,$57,$11,$A3,$FC,$BF,$5D,$1C,$4F,$F7
	.dc.b	$5F,$AD,$F2,$12,$3F,$BF,$3F,$DD,$7E,$2F,$F8,$E9,$F0,$DB,$1F,$E0
	.dc.b	$1F,$B2,$F6,$11,$FF,$03,$3F,$DF,$7E,$23,$F9,$29,$F2,$DB,$1F,$E6
	.dc.b	$DF,$25,$38,$C3,$FC,$11,$57,$5F,$F0,$44,$DB,$FF,$5E,$72,$7F,$BC
	.dc.b	$EB,$17,$F8,$DF,$CC,$BF,$C2,$6F,$E4,$38,$91,$88,$FC,$73,$5E,$1E
	.dc.b	$9F,$F4,$FF,$05,$F1,$1F,$B8,$6E,$5F,$C9,$2F,$FB,$95,$FF,$B2,$67
	.dc.b	$2F,$F8,$EB,$F6,$B7,$CD,$7E,$A5,$B7,$FA,$86,$CF,$F9,$59,$AF,$9D
	.dc.b	$AF,$E8,$DA,$FE,$A5,$B9,$1C,$E8,$F0,$3F,$DD,$D7,$FD,$23,$E1,$3F
	.dc.b	$CC,$15,$E0,$E0,$FE,$B9,$18,$EA,$F3,$12,$BC,$CD,$CB,$CC,$D7,$CC
	.dc.b	$FF,$91,$70,$FF,$13,$4F,$F6,$DF,$87,$E8,$FF,$58,$7F,$B7,$11,$FE
	.dc.b	$9B,$8A,$B1,$7F,$E6,$2F,$C8,$3F,$46,$FF,$03,$6B,$FC,$25,$BA,$1A
	.dc.b	$4E,$9F,$E3,$0D,$C3,$FC,$99,$EE,$FF,$76,$3F,$D0,$51,$FB,$2D,$7F
	.dc.b	$CA,$91,$F7,$F9,$7F,$DF,$DF,$02,$F1,$3F,$C1,$1E,$9E,$16,$9F,$81
	.dc.b	$AC,$3F,$D6,$A6,$E4,$FD,$55,$E7,$6B,$93,$FC,$A9,$F7,$5A,$9C,$D7
	.dc.b	$4C,$FC,$18,$FF,$96,$EB,$AA,$FA,$4F,$F3,$4F,$C1,$FF,$8A,$A3,$1F
	.dc.b	$00,$8C,$4E,$A7,$FC,$5E,$D7,$F9,$99,$CF,$B3,$FD,$FC,$E9,$AE,$A3
	.dc.b	$F4,$7E,$1F,$49,$CC,$FC,$2C,$9F,$81,$3E,$EB,$F9,$FE,$9E,$AD,$BF
	.dc.b	$CE,$5F,$0B,$8E,$3F,$F3,$C3,$71,$63,$CB,$0C,$3D,$39,$35,$87,$E7
	.dc.b	$78,$AF,$67,$CC,$C3,$E4,$65,$E3,$D3,$FE,$00,$46,$0F,$F4,$27,$D3
	.dc.b	$7F,$B2,$F2,$FF,$44,$36,$32,$7F,$88,$3A,$3F,$CD,$4D,$F8,$4F,$9D
	.dc.b	$D7,$F8,$F6,$CE,$A7,$FD,$9E,$D7,$F8,$83,$5F,$CD,$47,$02,$3F,$E0
	.dc.b	$7C,$7C,$5A,$7E,$E5,$B9,$FE,$0A,$7F,$86,$79,$B8,$99,$1F,$E4,$47
	.dc.b	$D5,$8F,$F0,$39,$BD,$C7,$D2,$DA,$FE,$25,$18,$23,$82,$1F,$C5,$5F
	.dc.b	$C7,$61,$FC,$35,$F2,$92,$3F,$77,$7E,$03,$B9,$1F,$E4,$0D,$7F,$90
	.dc.b	$35,$FE,$40,$FE,$6F,$4F,$C5,$F5,$FD,$56,$CF,$F9,$D7,$6F,$E8,$EF
	.dc.b	$EC,$BB,$3F,$12,$FE,$87,$67,$B5,$F7,$7A,$7F,$83,$3F,$A7,$6C,$FF
	.dc.b	$A4,$9A,$FE,$91,$AF,$FB,$3F,$E5,$3E,$AF,$90,$69,$F6,$3C,$CF,$F3
	.dc.b	$BC,$FF,$5B,$8C,$B4,$FF,$73,$7B,$B8,$7F,$D9,$F1,$72,$FF,$03,$CF
	.dc.b	$33,$E7,$D7,$4A,$FF,$03,$46,$0F,$A7,$6F,$B8,$F2,$92,$7F,$DE,$9E
	.dc.b	$87,$94,$E6,$47,$FD,$F4,$F2,$FB,$7F,$69,$23,$FC,$0F,$F0,$8F,$57
	.dc.b	$11,$FE,$E6,$9B,$9F,$D0,$F4,$BF,$D8,$15,$FA,$B6,$D7,$28,$C5,$F3
	.dc.b	$8F,$B1,$C5,$FD,$EB,$57,$D8,$7C,$B3,$E0,$92,$7F,$D8,$FC,$E9,$DC
	.dc.b	$63,$67,$9D,$F5,$89,$1D,$85,$71,$F8,$CC,$C7,$FC,$00,$9B,$AF,$F8
	.dc.b	$E2,$7F,$67,$4F,$22,$92,$FF,$D6,$1E,$8D,$57,$F9,$2A,$EF,$59,$8C
	.dc.b	$76,$AF,$F1,$7F,$65,$30,$FF,$11,$3F,$BA,$F6,$3F,$D9,$51,$FE,$B2
	.dc.b	$6B,$FE,$6C,$F8,$4F,$E2,$71,$88,$FF,$0F,$35,$FF,$35,$35,$FF,$38
	.dc.b	$36,$FF,$CF,$73,$C8,$FD,$EE,$4F,$1F,$E0,$92,$FF,$D0,$FF,$0F,$5F
	.dc.b	$7F,$17,$F8,$CA,$B8,$47,$CF,$7E,$7D,$7F,$14,$F6,$D7,$FA,$AF,$BB
	.dc.b	$97,$FE,$41,$7F,$55,$F3,$98,$8E,$A7,$D8,$E3,$4E,$78,$3F,$CD,$B5
	.dc.b	$F5,$BC,$89,$7E,$1C,$D1,$FC,$F6,$1F,$35,$FA,$BB,$FE,$0F,$18,$EE
	.dc.b	$FE,$D4,$D3,$FB,$64,$FE,$C5,$FE,$CE,$D7,$FD,$36,$D7,$FC,$06,$D7
	.dc.b	$91,$FD,$B8,$A7,$FA,$8A,$D9,$3E,$73,$F8,$69,$27,$E5,$74,$3E,$1B
	.dc.b	$11,$F1,$2F,$8D,$63,$2F,$F4,$7D,$7E,$63,$C1,$97,$F0,$DF,$E8,$BE
	.dc.b	$2F,$C1,$E3,$FB,$11,$A9,$E9,$8F,$F1,$D7,$FE,$00,$D6,$DF,$A5,$E7
	.dc.b	$7F,$22,$23,$FA,$DE,$67,$C3,$FF,$4D,$8F,$B3,$18,$F2,$BF,$AA,$CE
	.dc.b	$74,$FE,$6F,$39,$AF,$DB,$43,$DF,$F8,$91,$D6,$B8,$69,$FE,$C5,$6C
	.dc.b	$7F,$B2,$DB,$5C,$76,$FD,$CF,$8D,$2B,$F0,$8F,$F6,$EC,$5F,$2B,$99
	.dc.b	$F8,$33,$6F,$F6,$F3,$FD,$03,$EE,$E5,$7F,$A5,$DF,$E8,$1F,$57,$88
	.dc.b	$FA,$9C,$89,$C5,$FF,$B0,$AB,$F8,$5F,$91,$2F,$F2,$CF,$E5,$79,$9D
	.dc.b	$34,$FC,$9B,$89,$FC,$9C,$60,$8E,$08,$7E,$3E,$38,$3F,$EE,$9C,$7F
	.dc.b	$DF,$6F,$FA,$BF,$E1,$3F,$AD,$9F,$CD,$3D,$EF,$F1,$16,$4F,$C0,$FF
	.dc.b	$96,$D7,$F9,$1A,$31,$E4,$7F,$93,$DB,$1F,$9D,$6B,$FE,$4C,$F2,$DE
	.dc.b	$57,$F7,$28,$FE,$97,$E7,$7F,$42,$3F,$A1,$FB,$BF,$D1,$FE,$79,$F0
	.dc.b	$B4,$95,$F8,$4C,$DF,$01,$E6,$6C,$78,$5A,$F9,$DA,$FF,$9D,$8F,$E6
	.dc.b	$AE,$1F,$F0,$43,$5C,$C7,$B6,$31,$9F,$C3,$8F,$ED,$FE,$E7,$F8,$08
	.dc.b	$95,$FB,$6F,$92,$FF,$01,$1F,$DE,$08,$FF,$58,$F1,$11,$88,$FF,$80
	.dc.b	$0D,$7F,$82,$46,$9F,$EB,$FE,$17,$FB,$49,$FE,$39,$5F,$D5,$0F,$CD
	.dc.b	$7A,$92,$3F,$B2,$F6,$7F,$A0,$5B,$1F,$E3,$88,$C4,$7F,$A0,$29,$18
	.dc.b	$8F,$E2,$8D,$7F,$DD,$AD,$7F,$B4,$B4,$7D,$DF,$E3,$B3,$FD,$DB,$7C
	.dc.b	$0F,$F5,$2C,$EB,$A7,$F1,$BF,$3F,$FD,$1B,$0F,$E7,$B8,$23,$11,$F5
	.dc.b	$E3,$1E,$6F,$E3,$E7,$33,$FE,$3F,$6B,$FD,$C1,$B3,$C2,$3F,$29,$18
	.dc.b	$37,$7E,$8F,$5B,$FC,$CF,$91,$A7,$E8,$BF,$07,$FF,$07,$5F,$F1,$46
	.dc.b	$BF,$A3,$6B,$E7,$6B,$FE,$53,$6B,$FE,$04,$6E,$06,$93,$7F,$F3,$6F
	.dc.b	$CC,$BB,$FF,$D0,$6D,$BF,$ED,$8D,$7F,$A1,$7C,$0F,$A5,$FE,$E0,$D7
	.dc.b	$EF,$35,$FE,$A0,$D7,$FD,$B3,$C0,$8C,$47,$F8,$99,$B1,$9C,$BF,$D0
	.dc.b	$31,$83,$3C,$1F,$D1,$EB,$FA,$A3,$73,$FF,$11,$34,$EF,$F3,$9F,$D6
	.dc.b	$E7,$FC,$0C,$D7,$F2,$8D,$7F,$B3,$35,$E3,$E6,$FF,$A4,$BE,$0F,$DD
	.dc.b	$7F,$E2,$F6,$DF,$B1,$B4,$EE,$7C,$FF,$E8,$AF,$86,$7F,$EB,$0B,$DE
	.dc.b	$87,$8D,$F0,$E6,$FF,$5A,$E3,$FC,$6C,$FE,$E4,$5C,$3F,$C8,$0D,$CB
	.dc.b	$E6,$AE,$DE,$3D,$D2,$3F,$C4,$C6,$CF,$F2,$A9,$B3,$FC,$B0,$3E,$7D
	.dc.b	$E1,$7F,$93,$A1,$F2,$FF,$D6,$91,$BA,$FE,$41,$19,$8C,$1A,$71,$5D
	.dc.b	$31,$C7,$93,$FC,$9E,$31,$5D,$A3,$17,$D1,$B1,$C0,$F0,$CF,$0B,$EB
	.dc.b	$87,$9F,$9B,$BE,$1F,$CC,$DB,$7E,$87,$78,$DC,$7C,$57,$EF,$C3,$77
	.dc.b	$E3,$3B,$E1,$E3,$7C,$33,$BA,$E3,$7E,$59,$F8,$27,$12,$DF,$81,$C6
	.dc.b	$3F,$01,$FE,$90,$7D,$1B,$FA,$0E,$23,$4F,$63,$FC,$7C,$BE,$2D,$F3
	.dc.b	$FF,$E4,$19,$FB,$51,$87,$6D,$7A,$35,$FF,$24,$73,$7F,$95,$1F,$DB
	.dc.b	$1F,$E5,$67,$04,$7F,$D4,$99,$7F,$B6,$76,$F9,$51,$EA,$9A,$FE,$1B
	.dc.b	$7F,$F7,$3E,$DF,$85,$8F,$C4,$CE,$94,$EA,$FC,$71,$D8,$FE,$D9,$19
	.dc.b	$7B,$19,$3B,$5F,$37,$D3,$7D,$E7,$84,$7F,$75,$BF,$04,$62,$3C,$67
	.dc.b	$33,$C7,$5C,$E7,$6D,$8D,$4F,$8F,$8B,$F7,$5E,$BF,$F9,$FA,$7F,$82
	.dc.b	$35,$FF,$61,$F1,$FF,$A5,$2F,$FD,$9A,$D8,$FE,$CC,$D7,$FC,$F0,$D7
	.dc.b	$9C,$70,$9E,$9C,$6F,$9F,$7C,$4E,$1B,$5F,$E8,$39,$D4,$E4,$91,$99
	.dc.b	$DD,$FF,$8A,$DA,$FF,$86,$EB,$8C,$F3,$FA,$A3,$E2,$6B,$93,$07,$C1
	.dc.b	$CB,$7F,$05,$63,$FD,$D2,$7F,$BB,$8B,$7F,$E0,$76,$BE,$1F,$D9,$7A
	.dc.b	$BF,$E3,$F6,$B9,$8F,$87,$18,$8F,$E2,$B8,$23,$11,$FC,$D5,$B1,$99
	.dc.b	$E2,$79,$FB,$7C,$35,$FD,$61,$AF,$FB,$F1,$B8,$1C,$D3,$EF,$5F,$03
	.dc.b	$F5,$9B,$87,$D3,$6E,$1F,$E5,$96,$BF,$3D,$AF,$A9,$AF,$F4,$D6,$D7
	.dc.b	$7A,$FC,$77,$4B,$E8,$5C,$98,$6E,$3F,$E6,$96,$E3,$FE,$6F,$6B,$FE
	.dc.b	$23,$5F,$1B,$C4,$93,$C5,$B3,$C5,$AF,$17,$C1,$D9,$27,$C0,$D9,$FF
	.dc.b	$3C,$3F,$E6,$E8,$FC,$B3,$63,$3E,$5F,$E7,$A3,$11,$F9,$36,$BF,$79
	.dc.b	$AF,$F7,$16,$E6,$73,$B0,$FC,$19,$6F,$F2,$CE,$0F,$C2,$FC,$FF,$FA
	.dc.b	$9A,$3F,$54,$DB,$FF,$48,$35,$FD,$33,$5F,$C8,$B5,$FF,$02,$36,$CD
	.dc.b	$35,$5F,$EF,$6F,$90,$FD,$47,$A8,$69,$30,$F8,$FE,$97,$C5,$23,$82
	.dc.b	$63,$F3,$A3,$2D,$8F,$F5,$57,$D0,$3E,$67,$F1,$2D,$8C,$C6,$0F,$E1
	.dc.b	$F9,$7F,$A3,$23,$E1,$FA,$38,$23,$F1,$BE,$47,$64,$70,$7C,$AC,$79
	.dc.b	$91,$C1,$FD,$67,$B5,$F1,$DF,$CA,$73,$FF,$13,$36,$33,$3B,$8C,$11
	.dc.b	$C1,$27,$FB,$6A,$F9,$AF,$E5,$91,$FC,$69,$B1,$9E,$FF,$E7,$0D,$3F
	.dc.b	$BA,$CF,$E2,$77,$7C,$4F,$D6,$24,$EB,$85,$3A,$8C,$36,$BF,$DC,$D5
	.dc.b	$C1,$B3,$FE,$AC,$F8,$6F,$E5,$1E,$91,$A4,$8D,$12,$7F,$D5,$D3,$B8
	.dc.b	$C4,$7F,$A7,$7E,$79,$F3,$7F,$C1,$9B,$3F,$EE,$B6,$CF,$FB,$C5,$AF
	.dc.b	$19,$E0,$49,$E0,$FF,$5D,$A7,$C1,$FE,$BE,$7E,$3C,$63,$A3,$53,$9B
	.dc.b	$FF,$0F,$CE,$6B,$7E,$8F,$F2,$E8,$34,$9E,$1F,$76,$CF,$08,$C1,$FA
	.dc.b	$8F,$9C,$91,$F4,$1B,$7F,$ED,$3D,$91,$88,$FF,$7C,$6F,$F6,$23,$11
	.dc.b	$FE,$FB,$FC,$AF,$F4,$8F,$E0,$9A,$FF,$8E,$1A,$EF,$A7,$F3,$11,$88
	.dc.b	$FF,$33,$B5,$34,$D6,$1F,$95,$9F,$C8,$F1,$E0,$8F,$F0,$48,$87,$E3
	.dc.b	$0B,$5E,$D6,$BE,$CC,$7F,$8F,$9F,$D8,$B4,$7C,$6F,$EE,$E7,$E9,$B5
	.dc.b	$FF,$3E,$9F,$EF,$87,$0F,$FB,$B9,$AE,$7A,$7F,$05,$18,$9E,$B1,$83
	.dc.b	$E1,$F7,$4C,$3F,$CA,$D9,$FF,$38,$C3,$E1,$CE,$5F,$9C,$D7,$E7,$3F
	.dc.b	$8F,$F8,$E7,$C1,$52,$5F,$E4,$23,$2F,$EC,$F1,$FF,$78,$37,$D5,$FD
	.dc.b	$69,$1C,$13,$0F,$D5,$C6,$5F,$F3,$AF,$07,$F6,$CF,$9D,$FF,$0E,$78
	.dc.b	$A6,$92,$FF,$81,$46,$5B,$19,$AE,$73,$97,$FD,$03,$E1,$BF,$66,$FE
	.dc.b	$8C,$4F,$7F,$FC,$7E,$FF,$46,$AF,$EB,$64,$DB,$C6,$7F,$36,$B9,$8F
	.dc.b	$F5,$86,$9F,$D6,$0A,$7F,$58,$29,$FE,$28,$A7,$83,$E0,$24,$FF,$92
	.dc.b	$E3,$2F,$BB,$E0,$BE,$19,$1C,$1B,$E5,$F4,$91,$C1,$F1,$BE,$7B,$E5
	.dc.b	$91,$C1,$30,$F8,$F1,$96,$C6,$5F,$DC,$E4,$9D,$46,$1F,$E9,$5F,$71
	.dc.b	$FD,$02,$0D,$27,$85,$FC,$6F,$E7,$DF,$29,$49,$8F,$F1,$C8,$CB,$FD
	.dc.b	$13,$EF,$DB,$FF,$7F,$6F,$07,$FA,$05,$B3,$B9,$D1,$27,$4D,$8F,$E5
	.dc.b	$3F,$36,$FA,$0F,$E5,$CD,$8C,$C6,$0E,$3C,$D2,$3B,$37,$5F,$F6,$6E
	.dc.b	$3E,$68,$C4,$7E,$FA,$31,$3C,$0F,$0F,$D0,$78,$08,$E0,$93,$FC,$EB
	.dc.b	$04,$EA,$30,$FF,$A8,$7C,$A5,$BE,$5F,$0D,$C3,$FB,$6C,$EA,$72,$48
	.dc.b	$CC,$EE,$FF,$DC,$F5,$B9,$CD,$7F,$53,$F9,$CF,$F2,$1F,$59,$A4,$D7
	.dc.b	$FB,$EF,$89,$FE,$0A,$C7,$FD,$36,$0F,$F4,$BF,$99,$8C,$7F,$91,$89
	.dc.b	$E0,$E1,$B3,$C6,$1F,$41,$FE,$6A,$71,$19,$8C,$3E,$03,$FD,$56,$E2
	.dc.b	$3B,$B6,$3F,$99,$FD,$07,$F9,$0F,$DD,$FF,$7F,$36,$3F,$C2,$AD,$7F
	.dc.b	$9C,$87,$0F,$2F,$8D,$6D,$C9,$B1,$C9,$AF,$29,$46,$23,$DD,$F3,$2F
	.dc.b	$99,$F8,$B1,$86,$BF,$EC,$68,$D3,$F0,$8C,$46,$B2,$4F,$FB,$6A,$75
	.dc.b	$18,$F8,$0D,$4E,$67,$E4,$4E,$5E,$9C,$1F,$7F,$D4,$FF,$1B,$C2,$30
	.dc.b	$FF,$85,$38,$53,$A9,$CC,$FE,$9B,$F3,$B6,$DC,$B9,$11,$89,$FE,$00
	.dc.b	$D7,$84,$EE,$31,$1F,$AD,$6B,$F8,$D9,$E7,$3B,$3C,$5C,$CF,$9A,$A4
	.dc.b	$AF,$E0,$6D,$7F,$60,$D7,$FB,$BE,$09,$E5,$1F,$C5,$7D,$B7,$99,$EB
	.dc.b	$6C,$66,$30,$47,$04,$3F,$84,$C7,$07,$F8,$07,$D0,$7A,$08,$FF,$18
	.dc.b	$7B,$7A,$52,$3C,$4D,$79,$B5,$FE,$25,$3F,$E4,$07,$67,$F9,$C3,$2F
	.dc.b	$DD,$7A,$9F,$71,$B1,$9D,$FF,$D7,$CD,$9F,$C0,$AF,$CA,$5F,$19,$86
	.dc.b	$C7,$E1,$9A,$FF,$9F,$9A,$FF,$4C,$F6,$7F,$17,$E7,$FF,$BB,$1B,$E4
	.dc.b	$FF,$9D,$67,$FB,$FC,$60,$8E,$08,$7F,$80,$1C,$B4,$C9,$FE,$30,$F3
	.dc.b	$BF,$C0,$5E,$47,$E6,$9B,$1F,$DC,$63,$1E,$69,$F1,$F8,$B2,$7F,$C2
	.dc.b	$2D,$7F,$C1,$B3,$F1,$2B,$FC,$1D,$1C,$88,$7F,$A7,$9F,$F0,$0F,$D0
	.dc.b	$F8,$7C,$F3,$49,$3B,$8C,$B6,$32,$FF,$87,$23,$2F,$F8,$83,$E5,$FE
	.dc.b	$49,$1F,$F1,$3B,$EF,$FD,$9B,$71,$F0,$DA,$FF,$35,$B4,$62,$11,$8B
	.dc.b	$FA,$BF,$45,$27,$FC,$66,$D7,$FC,$62,$D7,$FC,$12,$D7,$FA,$16,$DF
	.dc.b	$C2,$7C,$EF,$C0,$F5,$7F,$81,$27,$FC,$99,$18,$23,$82,$1F,$EB,$47
	.dc.b	$F8,$77,$A9,$FE,$C8,$8F,$F2,$B7,$BD,$82,$3F,$C6,$3E,$45,$F4,$3F
	.dc.b	$9A,$6C,$66,$30,$47,$04,$3F,$DA,$F1,$C1,$FF,$3E,$F1,$7F,$1F,$F6
	.dc.b	$BF,$11,$1F,$EB,$66,$BA,$31,$1B,$23,$83,$FC,$BB,$E9,$7E,$88,$8E
	.dc.b	$08,$7F,$97,$9C,$B4,$FE,$6B,$F3,$7F,$A9,$7F,$DD,$AD,$7F,$AC,$B5
	.dc.b	$FF,$49,$B5,$FE,$ED,$ED,$3E,$93,$F7,$ED,$8F,$E8,$2D,$7F,$DF,$2D
	.dc.b	$7F,$B2,$B5,$FF,$4F,$B5,$EB,$1D,$0F,$DF,$F8,$12,$3F,$CF,$AD,$7F
	.dc.b	$CF,$6D,$7F,$CA,$6D,$7F,$C1,$9B,$FF,$5E,$E9,$FE,$A1,$E4,$FA,$E3
	.dc.b	$13,$FE,$E8,$8C,$12,$76,$48,$FF,$76,$FC,$07,$F3,$E7,$FD,$D6,$D7
	.dc.b	$FA,$1B,$5F,$0B,$57,$C3,$BE,$85,$F4,$3F,$DA,$23,$0D,$8C,$CF,$F8
	.dc.b	$30,$E2,$E1,$FF,$07,$CF,$07,$4D,$73,$1B,$7F,$D9,$B1,$97,$FD,$E3
	.dc.b	$8B,$E4,$7F,$5E,$48,$FF,$5D,$35,$FF,$59,$C6,$DC,$90,$FF,$56,$FC
	.dc.b	$FF,$F3,$62,$3F,$EB,$46,$BA,$30,$43,$F2,$7C,$1F,$D2,$B8,$0D,$24
	.dc.b	$7F,$83,$BF,$31,$FE,$C8,$FF,$8C,$1A,$FF,$BA,$9A,$FF,$93,$5A,$9A
	.dc.b	$4F,$2B,$FA,$CB,$C3,$B7,$FD,$67,$F2,$FF,$CC,$C9,$5F,$CC,$5F,$E6
	.dc.b	$DE,$3C,$8F,$E6,$8D,$7F,$99,$78,$FF,$CE,$C8,$FF,$39,$7F,$01,$F3
	.dc.b	$FF,$E9,$A8,$FB,$2D,$7E,$EB,$5F,$B2,$D7,$F9,$A7,$D0,$3C,$7F,$EB
	.dc.b	$0D,$8F,$E9,$8D,$7F,$DB,$ED,$7F,$B0,$35,$E7,$1C,$88,$FF,$BB,$33
	.dc.b	$FE,$E2,$47,$07,$F5,$3F,$37,$FA,$C1,$1C,$10,$FE,$A8,$E5,$A7,$E7
	.dc.b	$7C,$2F,$F2,$47,$99,$FD,$00,$91,$9E,$2F,$F1,$DB,$FD,$43,$17,$C1
	.dc.b	$FC,$02,$47,$C0,$6B,$F0,$1A,$FC,$07,$F6,$1F,$3E,$FA,$0A,$49,$FD
	.dc.b	$5C,$65,$BE,$BF,$02,$38,$21,$FB,$48,$FF,$34,$3F,$B0,$6B,$FE,$9F
	.dc.b	$E8,$46,$23,$F8,$2F,$CF,$BC,$FF,$E0,$8D,$8F,$DE,$B5,$FE,$02,$D7
	.dc.b	$FC,$35,$E4,$FF,$B6,$5F,$F0,$63,$5F,$F4,$AF,$37,$FB,$BB,$E0,$FF
	.dc.b	$C3,$D3,$99,$E3,$18,$23,$C9,$FF,$19,$FC,$CF,$E1,$E1,$C2,$8C,$37
	.dc.b	$93,$67,$F8,$AF,$F0,$7F,$B3,$89,$A4,$E3,$D1,$30,$FF,$12,$46,$5F
	.dc.b	$C3,$7C,$FF,$F9,$5D,$FF,$57,$E0,$8C,$47,$FB,$2C,$F2,$B8,$7F,$C1
	.dc.b	$AD,$7F,$C3,$3F,$59,$FD,$77,$03,$49,$E2,$7F,$90,$FC,$BF,$49,$32
	.dc.b	$C1,$2F,$FC,$75,$19,$6C,$7F,$96,$9F,$F2,$E9,$FE,$1E,$29,$F2,$8B
	.dc.b	$8F,$FB,$8F,$AF,$81,$23,$F4,$C6,$8E,$23,$E7,$DD,$DF,$D3,$A3,$F1
	.dc.b	$5E,$1F,$C3,$9F,$F3,$BC,$60,$FF,$78,$38,$E4,$FE,$79,$1C,$EB,$94
	.dc.b	$62,$7A,$F3,$7F,$BD,$E3,$1D,$FF,$3D,$B7,$9C,$3C,$18,$72,$6D,$E6
	.dc.b	$7C,$37,$C6,$7C,$CD,$8F,$2D,$71,$6D,$71,$7E,$04,$ED,$A6,$70,$F3
	.dc.b	$BE,$F3,$D9,$39,$9E,$13,$F0,$6F,$CB,$5B,$8E,$2F,$FB,$F1,$AE,$8C
	.dc.b	$5F,$CF,$75,$F3,$CF,$9F,$FF,$71,$D2,$31,$3F,$DE,$E3,$13,$A9,$CC
	.dc.b	$FF,$BA,$B9,$7F,$D5,$6F,$EB,$4F,$F5,$6B,$82,$3E,$D3,$FD,$D3,$F0
	.dc.b	$1F,$15,$1C,$74,$FB,$AE,$8E,$06,$72,$FD,$43,$98,$DB,$93,$F5,$9B
	.dc.b	$AB,$B4,$64,$EF,$19,$7A,$DF,$8E,$31,$5D,$23,$15,$C9,$B5,$C0,$F6
	.dc.b	$57,$0B,$FF,$11,$4E,$AF,$FC,$25,$BB,$E1,$3F,$D6,$98,$FF,$4B,$6E
	.dc.b	$3F,$C1,$9A,$FF,$05,$9E,$1E,$7F,$9E,$31,$3A,$7F,$D2,$6D,$7F,$8F
	.dc.b	$B6,$78,$D7,$FA,$E7,$E7,$DD,$DC,$EF,$FD,$51,$18,$9D,$4E,$67,$FD
	.dc.b	$53,$39,$7F,$9A,$B5,$FF,$6A,$35,$E7,$1C,$79,$D3,$99,$FF,$59,$FC
	.dc.b	$FF,$D5,$BF,$F6,$63,$63,$FA,$0B,$5F,$F7,$03,$5F,$F5,$8C,$3E,$23
	.dc.b	$CD,$8F,$DF,$6C,$64,$FE,$F0,$59,$FE,$9B,$F3,$FC,$DF,$F6,$2B,$71
	.dc.b	$34,$93,$EB,$6B,$FE,$E3,$79,$5B,$E7,$FB,$5F,$D9,$6B,$FE,$C8,$6B
	.dc.b	$FD,$51,$B6,$F9,$F7,$CF,$FC,$96,$E3,$FE,$F1,$6D,$FF,$66,$6B,$FD
	.dc.b	$2E,$B9,$57,$12,$57,$17,$EE,$74,$BE,$0F,$FB,$D4,$FF,$9D,$1B,$7F
	.dc.b	$E7,$86,$BF,$E7,$46,$BF,$E0,$55,$F3,$FD,$1B,$3C,$E3,$04,$70,$49
	.dc.b	$ED,$1C,$7A,$E4,$F8,$DB,$3F,$E0,$45,$F1,$FF,$8D,$6C,$FF,$A4,$72
	.dc.b	$46,$2B,$85,$6C,$95,$B9,$D4,$FC,$96,$BF,$EF,$ED,$7E,$E9,$27,$F1
	.dc.b	$EF,$E1,$FE,$77,$FD,$03,$1E,$16,$CF,$F8,$69,$AF,$D5,$6B,$FC,$83
	.dc.b	$04,$62,$11,$8F,$47,$F5,$53,$98,$C5,$7F,$94,$FE,$65,$F3,$FF,$A2
	.dc.b	$6C,$66,$30,$7C,$7F,$85,$91,$F0,$5B,$8F,$CD,$6B,$F0,$DA,$FF,$90
	.dc.b	$FE,$7F,$FC,$B7,$3F,$EB,$A6,$C7,$FA,$F9,$AF,$FA,$E9,$AF,$BF,$17
	.dc.b	$A7,$EC,$25,$7C,$97,$DF,$F1,$D2,$3E,$63,$6B,$E6,$61,$F1,$09,$3F
	.dc.b	$45,$FE,$31,$1F,$CB,$FE,$4E,$F6,$1A,$4A,$FF,$6C,$4E,$DB,$3B,$6C
	.dc.b	$EE,$7F,$A8,$C3,$E7,$3F,$98,$37,$BB,$F9,$43,$4F,$F7,$21,$67,$FC
	.dc.b	$87,$B3,$E1,$3E,$C1,$27,$EE,$36,$76,$D7,$FA,$24,$F1,$9F,$F4,$87
	.dc.b	$95,$FC,$8B,$D6,$34,$91,$A2,$57,$FB,$9A,$32,$FF,$BD,$BE,$7D,$F3
	.dc.b	$FF,$C1,$9B,$5F,$EF,$86,$CE,$E7,$44,$9D,$36,$3F,$BB,$DB,$E0,$3F
	.dc.b	$6B,$5F,$DE,$A5,$39,$E8,$FE,$A0,$D9,$D3,$67,$9F,$C1,$7F,$2A,$83
	.dc.b	$49,$E1,$72,$6D,$72,$6D,$7D,$2F,$9F,$FD,$41,$27,$F5,$8D,$8D,$46
	.dc.b	$08,$FA,$7E,$7F,$F3,$AF,$F8,$69,$A9,$A4,$F6,$FF,$2A,$D9,$FC,$8B
	.dc.b	$C9,$FC,$64,$7E,$2B,$8B,$04,$7F,$7D,$10,$FE,$04,$5A,$FC,$6B,$5F
	.dc.b	$C3,$B5,$FF,$17,$E4,$F8,$7F,$D9,$CF,$CE,$6C,$FD,$86,$BF,$E9,$79
	.dc.b	$E1,$D1,$F5,$A7,$33,$D1,$AF,$DC,$6C,$F0,$AF,$7F,$7F,$C2,$23,$82
	.dc.b	$57,$CD,$8C,$B6,$3F,$07,$F1,$2F,$88,$FC,$3B,$63,$31,$83,$F4,$9E
	.dc.b	$CC,$8F,$F7,$B3,$6B,$FD,$6B,$82,$31,$1F,$E3,$18,$C4,$F0,$3F,$31
	.dc.b	$E1,$7F,$B8,$7E,$7F,$FC,$23,$E7,$1A,$49,$DC,$65,$B1,$9B,$FF,$62
	.dc.b	$CE,$5F,$F7,$0F,$C7,$FF,$3F,$27,$8B,$FE,$2B,$7F,$A3,$61,$FE,$F1
	.dc.b	$D7,$C2,$4C,$7B,$BF,$D2,$2E,$47,$F5,$56,$9F,$D5,$4A,$7F,$55,$29
	.dc.b	$FD,$56,$9E,$07,$84,$93,$FD,$BE,$32,$FB,$BE,$7F,$E1,$91,$C1,$BE
	.dc.b	$2F,$51,$1C,$1F,$1B,$E7,$FE,$59,$1C,$13,$0F,$8F,$19,$6C,$65,$FF
	.dc.b	$34,$68,$9D,$46,$1F,$E8,$BF,$3F,$FE,$67,$83,$49,$E0,$FF,$12,$F9
	.dc.b	$F7,$C0,$D2,$63,$FC,$52,$32,$FF,$1B,$F9,$8F,$F0,$AF,$0F,$F9,$A5
	.dc.b	$B1,$9E,$0F,$EF,$D1,$8C,$BF,$D3,$6D,$7F,$D4,$61,$C3,$FE,$A6,$6B
	.dc.b	$98,$FF,$04,$71,$BE,$5D,$F0,$7F,$CA,$9B,$87,$F2,$A6,$DF,$F3,$46
	.dc.b	$BE,$0C,$3F,$C5,$BC,$DE,$9C,$7F,$43,$3E,$66,$BB,$6B,$E5,$6B,$E5
	.dc.b	$8E,$44,$BF,$17,$CF,$F9,$48,$E0,$93,$E5,$8C,$B6,$3F,$A4,$79,$3F
	.dc.b	$17,$E7,$FF,$D2,$CD,$8C,$E2,$9C,$B6,$33,$18,$23,$83,$FA,$7F,$CC
	.dc.b	$FF,$90,$AF,$FA,$AB,$75,$34,$D4,$FF,$A4,$FC,$BF,$F6,$63,$F6,$8C
	.dc.b	$1F,$ED,$E3,$1E,$27,$F8,$36,$FF,$D7,$91,$A6,$FC,$FF,$F9,$A9,$BE
	.dc.b	$97,$47,$94,$66,$7B,$36,$3F,$13,$F3,$FF,$E4,$4F,$7F,$FC,$3A,$D8
	.dc.b	$FE,$E0,$D7,$FD,$D6,$D7,$8F,$CF,$FF,$B5,$DD,$B5,$DB,$5D,$EF,$FE
	.dc.b	$1E,$F9,$57,$CF,$FC,$19,$F8,$11,$87,$E1,$18,$8D,$3F,$EF,$78,$D3
	.dc.b	$F1,$1F,$F6,$BB,$C0,$D7,$C0,$70,$CD,$3A,$9E,$13,$D3,$83,$EB,$FA
	.dc.b	$9F,$E3,$E6,$CF,$DE,$8C,$70,$9F,$F2,$1F,$1B,$E7,$DB,$E1,$B7,$FB
	.dc.b	$F6,$BF,$BA,$6B,$FA,$B6,$BF,$B9,$6B,$FB,$D6,$CF,$79,$F3,$74,$3E
	.dc.b	$7E,$92,$FF,$74,$D7,$F7,$4D,$7F,$BB,$47,$17,$F8,$A7,$B7,$FE,$70
	.dc.b	$6F,$99,$EE,$8C,$47,$F1,$26,$BF,$C0,$DA,$FF,$11,$6B,$FC,$55,$B8
	.dc.b	$7F,$B2,$E7,$D3,$F2,$52,$BF,$88,$35,$ED,$E8,$F8,$48,$F9,$9A,$E8
	.dc.b	$C1,$0F,$E4,$A7,$D4,$D7,$FD,$7E,$FF,$92,$F3,$FD,$AF,$A5,$FE,$7C
	.dc.b	$6C,$66,$76,$FF,$94,$7E,$7D,$F1,$7F,$E8,$28,$46,$1F,$F4,$2B,$5E
	.dc.b	$11,$FE,$88,$F3,$5E,$4F,$F9,$C2,$7F,$BF,$FB,$19,$8C,$47,$F7,$E8
	.dc.b	$C1,$1D,$90,$D1,$4F,$EB,$6E,$DF,$F0,$2F,$91,$FE,$98,$F7,$9E,$72
	.dc.b	$30,$D9,$FE,$ED,$1E,$3F,$16,$47,$F8,$51,$AF,$F8,$40,$E2,$43,$FB
	.dc.b	$F1,$DF,$CF,$7C,$ED,$24,$7F,$BC,$63,$2D,$8C,$BF,$63,$0F,$F0,$5C
	.dc.b	$62,$3F,$BB,$FC,$FF,$C7,$23,$FE,$28,$7D,$DF,$3F,$D2,$39,$35,$FE
	.dc.b	$CF,$48,$C4,$7F,$BB,$5A,$B5,$FA,$FF,$3F,$F5,$49,$1F,$E3,$66,$BF
	.dc.b	$E3,$46,$BF,$E0,$C7,$F0,$1D,$BF,$E0,$48,$FF,$21,$B5,$FF,$25,$F9
	.dc.b	$FA,$23,$A6,$BA,$3F,$CB,$7E,$FF,$F1,$5F,$5F,$FD,$89,$18,$21,$FE
	.dc.b	$A7,$F9,$07,$CF,$FE,$5D,$B1,$FE,$DA,$6B,$B8,$D1,$1D,$35,$FF,$3F
	.dc.b	$70,$FF,$25,$F6,$BF,$11,$18,$3F,$D0,$91,$92,$38,$21,$FE,$5A,$F4
	.dc.b	$FF,$40,$47,$FC,$C2,$D7,$46,$08,$7F,$B0,$FE,$6F,$F4,$EF,$EF,$9A
	.dc.b	$FF,$89,$1A,$83,$87,$FD,$09,$F3,$FF,$BB,$F7,$0D,$24,$7F,$C0,$08
	.dc.b	$CB,$63,$2F,$FB,$56,$32,$FF,$9D,$BB,$FF,$78,$47,$FC,$FA,$D7,$46
	.dc.b	$08,$7F,$B8,$BC,$9F,$4F,$07,$F5,$8E,$AF,$EA,$91,$82,$38,$24,$FF
	.dc.b	$B8,$A3,$8F,$C0,$7F,$3D,$8F,$F7,$6B,$5F,$F5,$B3,$5F,$F5,$FB,$57
	.dc.b	$C3,$BE,$79,$F3,$FF,$EF,$39,$CC,$61,$FF,$7B,$4E,$E7,$4D,$8C,$CE
	.dc.b	$DF,$E6,$B1,$96,$C6,$5F,$F2,$EC,$65,$FF,$57,$7C,$8F,$EB,$08,$FE
	.dc.b	$B1,$AF,$FA,$D9,$AF,$03,$FD,$5B,$F3,$FF,$CB,$C8,$E0,$87,$FA,$B9
	.dc.b	$CB,$4F,$81,$BB,$83,$FA,$13,$63,$31,$83,$FA,$37,$CF,$FF,$62,$83
	.dc.b	$49,$E4,$E8,$91,$F3,$63,$2F,$F2,$7F,$97,$FE,$58,$47,$07,$F2,$BF
	.dc.b	$27,$F9,$81,$1C,$1F,$CB,$FC,$6F,$E6,$84,$70,$7D,$8F,$9F,$7B,$DF
	.dc.b	$61,$B1,$98,$C1,$FD,$2F,$E7,$FF,$DC,$51,$FD,$B8,$78,$68,$CC,$7F
	.dc.b	$3B,$6B,$FC,$C9,$AF,$ED,$63,$87,$8F,$D4,$91,$D6,$32,$FF,$B9,$B4
	.dc.b	$FE,$D0,$47,$07,$F4,$DF,$37,$FA,$91,$1C,$10,$FE,$9D,$E4,$23,$51
	.dc.b	$87,$F9,$CF,$9B,$23,$F9,$CB,$5F,$67,$C2,$7C,$32,$3E,$7C,$91,$88
	.dc.b	$D9,$97,$0F,$E7,$BE,$79,$F3,$FF,$A8,$6C,$7E,$B1,$AF,$EA,$1A,$F1
	.dc.b	$F5,$F8,$11,$E0,$D7,$83,$5E,$07,$EE,$3E,$7F,$FC,$2A,$DF,$43,$F7
	.dc.b	$31,$88,$FD,$BB,$5F,$ED,$7E,$57,$FB,$29,$FA,$CD,$7F,$BD,$35,$DF
	.dc.b	$C2,$7F,$84,$E3,$11,$F8,$16,$BF,$8A,$6F,$07,$FB,$B2,$7F,$C5,$5F
	.dc.b	$33,$F8,$8A,$FE,$EC,$DF,$38,$D3,$54,$F9,$FF,$F3,$5B,$78,$3F,$C4
	.dc.b	$91,$82,$38,$25,$FE,$46,$3F,$CE,$76,$F9,$F7,$8F,$FE,$3D,$6D,$7E
	.dc.b	$8E,$75,$39,$3F,$C6,$5F,$3F,$FD,$53,$13,$49,$E3,$FF,$8D,$FC,$EF
	.dc.b	$39,$33,$C1,$31,$FD,$24,$65,$B1,$FE,$4F,$7F,$CA,$47,$F6,$13,$FC
	.dc.b	$38,$60,$FE,$C9,$8F,$F6,$5E,$CE,$04,$8F,$51,$B3,$8B,$97,$F8,$07
	.dc.b	$CF,$BB,$FF,$A3,$C7,$F9,$7B,$C4,$F0,$4F,$F9,$AB,$91,$18,$F0,$3C
	.dc.b	$91,$8A,$FE,$ED,$5C,$FC,$5F,$EF,$13,$99,$4E,$5E,$C7,$0A,$E7,$7D
	.dc.b	$63,$8B,$E2,$C3,$CB,$7C,$EF,$93,$6F,$8D,$F3,$38,$9E,$4A,$E3,$7C
	.dc.b	$F0,$CC,$F9,$2B,$8B,$4C,$E1,$DA,$B8,$47,$A0,$D1,$63,$E3,$46,$B0
	.dc.b	$F3,$61,$F3,$2B,$FD,$37,$5F,$05,$B1,$97,$A9,$AF,$9E,$7C,$FF,$F7
	.dc.b	$7A,$46,$27,$EC,$B5,$E1,$3F,$E9,$D7,$F5,$67,$FA,$7F,$97,$04,$70
	.dc.b	$3E,$83,$EB,$3F,$EE,$2F,$82,$FF,$18,$46,$23,$9B,$AE,$04,$EB,$4F
	.dc.b	$E9,$57,$6F,$F3,$15,$A7,$B4,$E9,$A6,$AB,$83,$F9,$FA,$FA,$AF,$8C
	.dc.b	$FA,$07,$F9,$0D,$CD,$75,$D3,$CD,$1C,$A7,$86,$18,$7B,$9F,$E4,$BA
	.dc.b	$F2,$46,$4F,$65,$F1,$9F,$7E,$1F,$ED,$2A,$DB,$FB,$0F,$84,$7C,$77
	.dc.b	$B7,$83,$DD,$C4,$9C,$E3,$FE,$59,$9C,$C7,$F4,$C8,$C1,$CE,$B9,$34
	.dc.b	$CE,$3F,$E9,$B9,$E6,$FF,$8F,$23,$C5,$E0,$70,$8D,$55,$C3,$F8,$F6
	.dc.b	$37,$A7,$F7,$38,$74,$E8,$FD,$DF,$C1,$66,$A4,$F8,$2F,$8B,$B8,$4E
	.dc.b	$B1,$F1,$1F,$EA,$F9,$7D,$0F,$DF,$8F,$E3,$17,$F8,$68,$C4,$7F,$18
	.dc.b	$9D,$9F,$CB,$CA,$FF,$C0,$09,$FE,$87,$EE,$7A,$F8,$4F,$B6,$B7,$BF
	.dc.b	$D0,$9C,$C7,$A4,$7C,$5F,$83,$0F,$33,$70,$FC,$CB,$67,$67,$8D,$C1
	.dc.b	$1C,$1F,$D1,$69,$EE,$79,$4F,$24,$D9,$F9,$54,$F7,$3C,$0D,$8F,$E9
	.dc.b	$CD,$C3,$E0,$35,$F9,$2D,$7F,$9F,$B5,$FF,$6F,$34,$FF,$70,$4F,$B7
	.dc.b	$95,$F0,$DF,$83,$6D,$F9,$27,$13,$F8,$9C,$DF,$08,$BF,$91,$18,$6C
	.dc.b	$66,$FF,$D0,$ED,$C7,$FD,$06,$D7,$FD,$02,$D7,$F4,$4D,$7F,$D2,$AD
	.dc.b	$F6,$1F,$57,$FE,$2A,$AF,$D0,$C6,$3C,$AF,$35,$77,$C7,$F8,$64,$E6
	.dc.b	$BE,$7F,$CB,$7C,$E2,$3F,$39,$B8,$FC,$E6,$BF,$39,$F9,$1E,$77,$C7
	.dc.b	$3F,$55,$B2,$31,$5F,$70,$FB,$64,$3E,$D9,$6B,$ED,$B5,$FB,$6F,$D5
	.dc.b	$F2,$7E,$A1,$D9,$A7,$ED,$7C,$2F,$F4,$B3,$FE,$69,$6C,$FF,$80,$CF
	.dc.b	$F3,$BB,$87,$FC,$0D,$D0,$8C,$47,$F9,$D5,$AF,$0F,$53,$F4,$93,$B6
	.dc.b	$CE,$B1,$FF,$01,$5F,$F0,$3E,$0F,$E1,$38,$7F,$00,$C3,$84,$E7,$A7
	.dc.b	$F8,$94,$78,$74,$B8,$9B,$EB,$61,$FC,$87,$CD,$FF,$23,$C7,$FD,$96
	.dc.b	$D7,$FD,$92,$D7,$F9,$53,$56,$7F,$97,$FA,$CF,$CF,$1E,$0D,$AE,$9F
	.dc.b	$8A,$FC,$0D,$F2,$BF,$C0,$F8,$91,$89,$FF,$51,$46,$27,$53,$9C,$7E
	.dc.b	$84,$E6,$B7,$39,$9F,$F1,$B4,$E5,$DC,$65,$FF,$53,$7E,$0C,$FB,$2F
	.dc.b	$8E,$DB,$F3,$61,$89,$FD,$C7,$87,$27,$F9,$A6,$2F,$99,$C6,$3F,$CC
	.dc.b	$B5,$FD,$C5,$E3,$9F,$FA,$9D,$EF,$0F,$B5,$3F,$C0,$FC,$49,$1D,$3D
	.dc.b	$AB,$8F,$3E,$7F,$E9,$F8,$FF,$5C,$6D,$7F,$94,$31,$FF,$67,$BF,$85
	.dc.b	$C3,$FA,$94,$3F,$12,$FF,$26,$4F,$EE,$9B,$87,$F4,$81,$FA,$A7,$F3
	.dc.b	$07,$FA,$5D,$71,$F3,$B9,$D7,$0C,$BF,$7E,$DC,$3F,$CC,$98,$71,$9C
	.dc.b	$D7,$F1,$FF,$86,$95,$E1,$8B,$8F,$B7,$CB,$FE,$48,$7F,$67,$C9,$F3
	.dc.b	$78,$BF,$D0,$4D,$AF,$9D,$5A,$9C,$9F,$A5,$8F,$E5,$C7,$3E,$3B,$19
	.dc.b	$24,$72,$8C,$BF,$DA,$A1,$F3,$8C,$BF,$AC,$46,$1B,$19,$8F,$F7,$C1
	.dc.b	$92,$B9,$9F,$E0,$D1,$96,$C6,$5F,$F4,$64,$65,$FE,$EF,$F5,$CF,$B1
	.dc.b	$FF,$05,$36,$7F,$C2,$67,$3D,$7F,$C3,$F1,$9A,$FC,$92,$FB,$1E,$27
	.dc.b	$CD,$F5,$ED,$FF,$83,$67,$4D,$9D,$34,$D7,$D8,$F8,$8F,$C6,$FB,$5F
	.dc.b	$7E,$3F,$D4,$34,$9D,$47,$03,$FA,$4C,$EF,$0F,$E0,$F5,$B9,$F5,$E7
	.dc.b	$FC,$2D,$FA,$3C,$FF,$38,$8F,$CC,$6D,$7C,$A6,$BF,$23,$E8,$BE,$39
	.dc.b	$FE,$33,$9E,$67,$BD,$F9,$7F,$25,$23,$FA,$EF,$C7,$BE,$17,$17,$F2
	.dc.b	$AA,$EF,$FD,$7A,$B5,$7F,$26,$7E,$AF,$D0,$48,$FE,$0B,$F0,$F7,$0F
	.dc.b	$A7,$37,$0F,$E0,$73,$6B,$F7,$5E,$AF,$F8,$8F,$9B,$FC,$53,$87,$F8
	.dc.b	$C5,$AF,$F8,$E9,$A8,$79,$4F,$D9,$FF,$D2,$B7,$FE,$4D,$6C,$FE,$A2
	.dc.b	$FF,$A2,$C6,$31,$FC,$4F,$C1,$FE,$38,$CF,$BD,$70,$FC,$34,$5C,$3F
	.dc.b	$A8,$5F,$19,$FC,$0C,$FE,$E7,$D7,$F5,$12,$3F,$DF,$59,$FF,$8D,$9F
	.dc.b	$F7,$E7,$23,$EE,$BF,$3A,$71,$8F,$CA,$3F,$C0,$FD,$1F,$88,$48,$E6
	.dc.b	$FF,$92,$65,$F7,$3F,$A6,$38,$4F,$E7,$5C,$34,$FE,$73,$38,$CF,$F4
	.dc.b	$F3,$75,$FE,$6F,$36,$BE,$76,$2F,$A4,$FF,$51,$61,$FD,$39,$AE,$9A
	.dc.b	$A3,$87,$81,$E7,$FB,$9F,$DD,$9F,$CD,$AB,$FD,$59,$8F,$EE,$A3,$F9
	.dc.b	$17,$5F,$E0,$C9,$8F,$CA,$6E,$3F,$85,$6B,$F8,$57,$E2,$79,$7F,$C2
	.dc.b	$0E,$73,$FB,$F7,$FA,$2F,$CE,$B8,$7F,$B9,$36,$7F,$B8,$B6,$FF,$B7
	.dc.b	$B5,$FE,$DA,$D7,$E2,$B5,$F8,$0D,$D0,$E7,$CB,$70,$7E,$26,$BF,$81
	.dc.b	$F2,$79,$31,$FE,$3E,$CC,$7F,$8B,$39,$69,$FE,$54,$C1,$D1,$EA,$24
	.dc.b	$7A,$9B,$7E,$A6,$BE,$A7,$F1,$5E,$37,$F2,$E3,$E2,$B6,$7F,$D8,$3C
	.dc.b	$2F,$9E,$7A,$9F,$EE,$BF,$79,$F1,$DF,$C8,$BD,$7C,$3F,$ED,$DA,$DE
	.dc.b	$1F,$EE,$78,$D3,$73,$FF,$50,$FD,$27,$07,$F9,$2F,$C4,$59,$FF,$54
	.dc.b	$CE,$78,$BE,$4C,$EE,$9F,$44,$FA,$A6,$DF,$E5,$59,$FE,$0D,$39,$8C
	.dc.b	$47,$09,$DB,$B9,$D3,$7E,$8F,$FD,$D4,$DF,$88,$E3,$29,$CC,$FF,$98
	.dc.b	$E3,$13,$FD,$93,$E6,$5F,$23,$FC,$5E,$9C,$3F,$21,$FE,$87,$F4,$B9
	.dc.b	$8C,$6B,$F8,$68,$C7,$07,$F1,$FE,$06,$3D,$30,$F9,$5B,$7D,$5F,$2B
	.dc.b	$71,$FD,$3B,$E7,$2E,$5F,$90,$73,$5B,$9F,$A5,$3F,$94,$C7,$CD,$E0
	.dc.b	$7F,$6C,$6E,$5F,$D7,$9B,$5F,$9E,$F8,$BB,$DD,$F9,$E7,$E6,$B6,$3F
	.dc.b	$57,$18,$EB,$FF,$7C,$C6,$08,$7E,$BA,$34,$7D,$6C,$7F,$4E,$F8,$1B
	.dc.b	$1F,$DE,$63,$C3,$3F,$2E,$3F,$C0,$31,$BE,$EE,$F1,$83,$F5,$18,$70
	.dc.b	$3C,$84,$7F,$80,$B5,$FE,$02,$D7,$F8,$0B,$E5,$8F,$F6,$B4,$7F,$82
	.dc.b	$FD,$AF,$01,$E7,$C7,$F4,$F1,$FE,$68,$EB,$F6,$46,$23,$FA,$F4,$60
	.dc.b	$8F,$F4,$86,$BA,$3F,$A4,$F2,$7B,$48,$FF,$B5,$BF,$37,$7C,$3F,$D4
	.dc.b	$6C,$7F,$A3,$1A,$FF,$A2,$9A,$FF,$A2,$1A,$FF,$BF,$BE,$6B,$E8,$6C
	.dc.b	$69,$ED,$E3,$B3,$FD,$CF,$85,$C3,$C6,$30,$D7,$FC,$B3,$3C,$1F,$F4
	.dc.b	$45,$F5,$76,$7F,$AB,$1C,$3C,$0F,$F4,$8C,$7C,$A2,$3F,$DB,$9F,$C4
	.dc.b	$43,$B3,$D2,$49,$FF,$4C,$FC,$57,$D5,$3E,$99,$1F,$EE,$EF,$2E,$9F
	.dc.b	$82,$48,$F2,$F9,$CF,$C8,$7E,$5E,$7F,$02,$7D,$15,$B7,$DA,$FE,$A2
	.dc.b	$B8,$6D,$6F,$8E,$9F,$E1,$B9,$CD,$F6,$9C,$BF,$EC,$EE,$1F,$F6,$D3
	.dc.b	$FE,$B2,$F8,$7F,$C5,$91,$FF,$07,$BF,$0B,$ED,$FE,$C9,$23,$FD,$73
	.dc.b	$F0,$D5,$FA,$31,$74,$FF,$5B,$CF,$E0,$33,$95,$F5,$37,$FF,$6F,$C7
	.dc.b	$F0,$0F,$57,$FD,$0D,$39,$9F,$39,$A7,$13,$BF,$77,$F1,$7F,$53,$27
	.dc.b	$83,$FC,$31,$7C,$EB,$DE,$C3,$63,$F9,$CB,$5F,$E7,$CD,$7F,$97,$1C
	.dc.b	$A3,$F9,$FB,$4E,$2E,$23,$84,$7F,$35,$6B,$F9,$1F,$B7,$FD,$09,$1F
	.dc.b	$CA,$EF,$FC,$CA,$B1,$1F,$9A,$9B,$AF,$E0,$EB,$53,$F1,$7E,$DF,$F5
	.dc.b	$64,$7F,$DD,$7F,$1B,$5F,$D1,$C5,$AF,$F7,$4C,$FE,$CB,$E4,$E5,$7C
	.dc.b	$F7,$E2,$FA,$D2,$3E,$2B,$ED,$23,$82,$1F,$21,$CB,$F2,$37,$0F,$93
	.dc.b	$FE,$02,$48,$FF,$80,$3F,$1D,$5F,$DC,$C5,$C7,$FD,$FF,$3F,$C2,$3E
	.dc.b	$C2,$57,$E1,$5F,$E4,$9F,$05,$88,$FE,$03,$16,$BF,$55,$87,$F9,$26
	.dc.b	$BF,$88,$FD,$24,$BF,$C7,$3F,$D4,$3E,$ED,$F0,$74,$93,$F5,$A3,$2F
	.dc.b	$F0,$C3,$F8,$27,$AB,$A7,$FD,$FF,$1A,$6C,$FE,$4F,$C1,$57,$F2,$2F
	.dc.b	$A3,$95,$F9,$F7,$F9,$F7,$C7,$62,$3F,$8D,$C5,$D3,$D7,$C8,$AF,$E5
	.dc.b	$7F,$59,$2F,$F5,$0F,$F4,$7E,$AC,$47,$F2,$48,$B7,$CA,$7F,$85,$7E
	.dc.b	$16,$F8,$7A,$4B,$FC,$AF,$DE,$7F,$88,$DB,$EC,$FF,$74,$8C,$47,$E9
	.dc.b	$B3,$46,$23,$FC,$2D,$18,$9E,$04,$8F,$F1,$64,$65,$B3,$C5,$FF,$5A
	.dc.b	$71,$FE,$2B,$EE,$BF,$A4,$92,$3F,$94,$FD,$2B,$E4,$7F,$C1,$7A,$BE
	.dc.b	$99,$F2,$BF,$E3,$68,$C3,$63,$37,$FE,$3D,$9C,$BF,$E3,$FF,$29,$87
	.dc.b	$C5,$8E,$FB,$7F,$31,$8C,$42,$BF,$9C,$57,$0F,$8D,$FE,$BF,$F5,$32
	.dc.b	$BF,$8E,$3C,$7B,$E4,$77,$C1,$1F,$E4,$CB,$F0,$B6,$33,$3B,$8C,$11
	.dc.b	$C1,$FE,$57,$DB,$F6,$3F,$01,$FD,$C0,$91,$FC,$C7,$87,$FC,$AE,$FA
	.dc.b	$7C,$EF,$9F,$18,$8F,$F7,$37,$B7,$FE,$0E,$23,$FE,$D2,$6B,$FE,$ED
	.dc.b	$6B,$FE,$55,$BF,$E0,$3C,$9F,$E0,$02,$47,$84,$FF,$0E,$1F,$E1,$A2
	.dc.b	$1F,$ED,$E2,$9F,$EF,$92,$9F,$EF,$93,$FC,$8B,$F7,$5F,$4B,$7F,$E4
	.dc.b	$CD,$C7,$FD,$9E,$D9,$FA,$DF,$4F,$FE,$2E,$7E,$01,$8F,$67,$F0,$D1
	.dc.b	$98,$FF,$2C,$B5,$FF,$0A,$35,$E0,$7E,$25,$C5,$9A,$6F,$37,$CC,$FF
	.dc.b	$23,$AF,$E8,$2D,$7C,$71,$AF,$1B,$FC,$C7,$2F,$1E,$9F,$E6,$47,$15
	.dc.b	$FA,$A6,$BF,$B3,$6B,$FB,$26,$BF,$EF,$26,$BF,$EE,$D6,$D7,$9A,$7F
	.dc.b	$CD,$19,$BC,$AF,$F3,$43,$6F,$FB,$D3,$5F,$97,$B3,$E8,$5F,$19,$FE
	.dc.b	$6F,$6C,$FE,$55,$AF,$F4,$76,$BF,$89,$6B,$FC,$09,$AF,$F4,$C6,$CF
	.dc.b	$79,$FF,$3F,$46,$29,$E6,$7A,$09,$7F,$E1,$19,$FF,$34,$6D,$FD,$19
	.dc.b	$FF,$35,$47,$87,$DC,$E6,$47,$F5,$9E,$E3,$AB,$FD,$14,$D8,$FF,$0E
	.dc.b	$35,$FC,$6F,$4B,$EF,$1F,$2D,$C2,$E5,$70,$7F,$C7,$93,$A8,$FF,$13
	.dc.b	$0F,$85,$FF,$4C,$34,$CB,$8F,$84,$CB,$23,$27,$28,$FF,$12,$7D,$2B
	.dc.b	$E3,$3F,$8C,$4B,$E9,$1E,$AF,$FB,$36,$BF,$AB,$C8,$D4,$67,$CE,$FF
	.dc.b	$69,$57,$09,$D5,$75,$E1,$E1,$39,$94,$E5,$E6,$70,$AE,$77,$D6,$38
	.dc.b	$BE,$3C,$3C,$F7,$CE,$F9,$76,$5C,$3F,$DC,$8D,$C3,$D0,$EE,$FD,$ED
	.dc.b	$C7,$AC,$F8,$4A,$79,$AB,$C6,$FB,$2B,$C3,$8F,$8E,$B7,$87,$C6,$AF
	.dc.b	$59,$C4,$CD,$70,$79,$4F,$F9,$31,$A5,$9F,$88,$FF,$94,$1A,$FD,$3F
	.dc.b	$8E,$7D,$2F,$18,$F9,$71,$87,$93,$5E,$4D,$7C,$1F,$17,$FE,$5A,$4D
	.dc.b	$23,$FE,$5E,$7E,$4E,$3F,$E6,$2F,$0B,$FD,$29,$A7,$DB,$AF,$C2,$56
	.dc.b	$9F,$C3,$09,$DD,$78,$21,$5C,$67,$31,$FE,$6E,$AE,$4F,$E3,$E3,$2F
	.dc.b	$48,$CB,$D7,$A2,$BE,$A3,$AD,$F9,$E3,$33,$C2,$B9,$D7,$A0,$D9,$E8
	.dc.b	$72,$F3,$BF,$35,$72,$8D,$9E,$57,$E0,$75,$39,$9E,$B7,$FA,$E6,$BF
	.dc.b	$B1,$46,$2B,$F8,$4C,$E6,$7F,$C3,$B3,$97,$FC,$D0,$D7,$8C,$6D,$FF
	.dc.b	$35,$C6,$5E,$31,$9E,$7F,$DC,$FC,$4D,$D0,$D2,$61,$FE,$1D,$8C,$B6
	.dc.b	$33,$7F,$EA,$0A,$E7,$3F,$E2,$3F,$BA,$78,$BD,$DB,$19,$8C,$1E,$7C
	.dc.b	$FF,$DF,$3E,$05,$9F,$EB,$47,$F8,$54,$B8,$7F,$A1,$DA,$FE,$2B,$06
	.dc.b	$09,$F0,$10,$F2,$16,$7D,$11,$B8,$C9,$23,$B7,$87,$E3,$24,$78,$DF
	.dc.b	$95,$F7,$1F,$D2,$23,$FC,$84,$D9,$F8,$8D,$7C,$CD,$7F,$DE,$0D,$56
	.dc.b	$8C,$37,$C0,$EE,$7C,$7C,$9F,$4F,$49,$87,$F8,$EE,$32,$F9,$3D,$BF
	.dc.b	$39,$1C,$13,$1F,$3C,$65,$B1,$FE,$5F,$FA,$6F,$D9,$B7,$8F,$F1,$DD
	.dc.b	$29,$CD,$7F,$B8,$A3,$13,$F1,$BC,$A9,$7F,$19,$B5,$F1,$9A,$FC,$66
	.dc.b	$BF,$E7,$CC,$5F,$4D,$F0,$49,$5E,$F7,$F3,$9D,$BF,$E7,$A8,$FF,$3A
	.dc.b	$36,$BF,$2C,$D7,$F1,$0D,$7F,$AC,$6D,$FE,$B0,$F4,$79,$92,$B9,$BF
	.dc.b	$45,$87,$CF,$2C,$F6,$F9,$BF,$19,$23,$C6,$D8,$D4,$60,$8F,$9B,$CE
	.dc.b	$F5,$92,$3D,$6F,$BB,$C7,$91,$EE,$6C,$FB,$BE,$1F,$E1,$92,$7E,$1B
	.dc.b	$F1,$BE,$1A,$47,$C6,$6C,$FF,$4F,$ED,$7A,$5F,$C2,$E3,$F8,$C3,$63
	.dc.b	$35,$FD,$35,$B3,$C6,$74,$49,$D3,$FC,$57,$CE,$7D,$2B,$E8,$B0,$DA
	.dc.b	$FF,$5F,$B6,$FF,$D9,$0D,$7F,$D1,$15,$FE,$66,$9F,$F4,$D3,$63,$FA
	.dc.b	$68,$E2,$FF,$2C,$4E,$BD,$1F,$E2,$15,$BA,$E7,$18,$24,$F4,$E0,$D9
	.dc.b	$1C,$12,$7B,$DA,$FF,$19,$1E,$16,$C6,$6B,$E6,$7D,$1B,$B6,$93,$0F
	.dc.b	$EF,$11,$96,$C6,$70,$FF,$02,$4E,$5B,$1F,$37,$E9,$3F,$C7,$15,$FE
	.dc.b	$E8,$6C,$FF,$BC,$5A,$FF,$BC,$1A,$FF,$AF,$3D,$9F,$8A,$49,$F9,$AD
	.dc.b	$8F,$7E,$5F,$EE,$B9,$C9,$23,$3A,$FC,$E2,$47,$D6,$6C,$6A,$30,$47
	.dc.b	$F2,$5F,$03,$F8,$F2,$47,$E3,$DB,$1A,$8C,$11,$FC,$0A,$FA,$27,$BF
	.dc.b	$FD,$96,$30,$D8,$CC,$7F,$8B,$CC,$9F,$8C,$F8,$7F,$C3,$91,$C1,$2B
	.dc.b	$E1,$C6,$5B,$1F,$E3,$9C,$7F,$D3,$BA,$7E,$87,$D7,$FB,$15,$FE,$22
	.dc.b	$9C,$D6,$E7,$33,$FE,$6B,$9C,$BF,$9C,$6B,$C6,$36,$FE,$45,$AB,$ED
	.dc.b	$7F,$C0,$4F,$F9,$D5,$AE,$7C,$BF,$DF,$46,$08,$E0,$97,$F1,$72,$E6
	.dc.b	$FC,$D8,$F9,$44,$7E,$07,$C6,$3E,$47,$F8,$03,$63,$31,$82,$38,$26
	.dc.b	$1E,$F9,$E0,$FE,$43,$DF,$FD,$C9,$27,$F8,$0B,$63,$51,$82,$3F,$E1
	.dc.b	$7F,$A2,$7C,$FF,$F8,$8D,$B3,$FE,$6F,$F0,$51,$8A,$E1,$5F,$C2,$3D
	.dc.b	$B9,$7F,$C4,$1B,$3D,$7D,$4E,$84,$9E,$96,$8C,$4E,$E7,$31,$88,$FF
	.dc.b	$0E,$4B,$E6,$FC,$C4,$AE,$6F,$F1,$AF,$12,$47,$F2,$26,$D7,$BF,$DA
	.dc.b	$F7,$12,$7C,$6F,$F2,$2F,$16,$47,$F2,$A6,$CF,$EF,$9A,$FF,$30,$6B
	.dc.b	$FE,$BF,$EE,$FE,$41,$18,$8F,$E4,$0D,$5F,$8E,$BE,$9B,$FA,$63,$6F
	.dc.b	$FD,$2C,$D9,$FF,$50,$35,$FD,$37,$D1,$7F,$9E,$DF,$D1,$35,$CF,$B1
	.dc.b	$FD,$36,$31,$1F,$E8,$5F,$9F,$71,$FF,$99,$69,$18,$C3,$FD,$5D,$18
	.dc.b	$9D,$4E,$6B,$FD,$38,$D7,$FD,$04,$D7,$B4,$70,$3F,$A6,$F4,$CB,$FE
	.dc.b	$AA,$D9,$FE,$AB,$F6,$AF,$11,$AF,$F9,$E6,$30,$D8,$CD,$FF,$B0,$E7
	.dc.b	$2D,$8C,$D7,$46,$BF,$EA,$E9,$DB,$9A,$FE,$E5,$F3,$CF,$A2,$A4,$C3
	.dc.b	$F8,$C4,$65,$B1,$9A,$FE,$25,$39,$6C,$79,$F2,$95,$E5,$6C,$F9,$5A
	.dc.b	$F9,$71,$F9,$4B,$1F,$89,$3E,$6F,$A2,$97,$E4,$6D,$7D,$4F,$3B,$FD
	.dc.b	$00,$FF,$AA,$F6,$FB,$B1,$8A,$FA,$ED,$9F,$C2,$35,$FF,$44,$78,$2F
	.dc.b	$13,$15,$FD,$D5,$B1,$98,$4F,$FA,$E5,$D1,$83,$FC,$23,$F4,$5F,$EA
	.dc.b	$2C,$3F,$DE,$4D,$9F,$F7,$E3,$5D,$FB,$9F,$EE,$37,$3E,$9F,$FB,$1D
	.dc.b	$BE,$7F,$28,$C4,$76,$6B,$E1,$FA,$27,$CE,$FF,$BB,$B2,$91,$89,$F3
	.dc.b	$1B,$9D,$34,$C9,$FE,$F5,$E6,$FE,$74,$E3,$3F,$D4,$F8,$9C,$23,$4D
	.dc.b	$8C,$C7,$2F,$21,$CD,$FE,$CE,$8F,$C6,$FB,$FF,$E3,$67,$FA,$F4,$65
	.dc.b	$FC,$6C,$65,$FC,$27,$B3,$E2,$7F,$52,$D7,$F3,$CD,$7E,$AF,$37,$FA
	.dc.b	$57,$1E,$4D,$8F,$D5,$4E,$DE,$B3,$A8,$C6,$1E,$8F,$A2,$F1,$36,$BF
	.dc.b	$4B,$18,$C4,$D2,$46,$88,$7A,$3A,$FB,$11,$F1,$B5,$D1,$82,$1F,$E3
	.dc.b	$0E,$6F,$89,$DF,$FE,$4F,$6E,$06,$9F,$27,$5F,$E3,$FF,$33,$77,$34
	.dc.b	$98,$7F,$22,$8C,$B6,$32,$FF,$22,$7E,$1F,$D0,$BE,$76,$92,$7F,$C4
	.dc.b	$11,$97,$ED,$7D,$A7,$F6,$16,$F9,$1F,$75,$46,$5F,$C2,$35,$FF,$3F
	.dc.b	$F0,$FB,$88,$FF,$A0,$1A,$FF,$95,$1A,$F0,$34,$43,$45,$3F,$76,$7F
	.dc.b	$BC,$BE,$83,$F8,$13,$FE,$03,$6B,$FE,$8C,$F3,$11,$88,$E3,$18,$EE
	.dc.b	$FA,$7F,$43,$27,$FD,$2A,$D7,$95,$3E,$87,$F3,$D7,$D1,$B1,$99,$F1
	.dc.b	$46,$34,$91,$E5,$6B,$D1,$AF,$92,$7E,$93,$93,$FD,$4F,$EE,$FE,$20
	.dc.b	$8E,$0F,$81,$F0,$FE,$22,$38,$3F,$D5,$FE,$F7,$E4,$08,$E0,$87,$FA
	.dc.b	$19,$CB,$4F,$F2,$2F,$AF,$FA,$CF,$4F,$FC,$EB,$18,$21,$FE,$41,$8D
	.dc.b	$11,$C9,$FE,$BE,$FA,$07,$B1,$F8,$B6,$C7,$F0,$96,$BF,$C3,$DA,$FF
	.dc.b	$B1,$BD,$F7,$A3,$F9,$36,$C7,$F1,$16,$BF,$C5,$DA,$F1,$8E,$04,$78
	.dc.b	$3F,$C2,$3C,$3E,$F1,$FE,$7B,$6B,$FE,$C7,$6F,$59,$A4,$C3,$91,$27
	.dc.b	$FD,$BF,$F1,$3F,$C0,$08,$E0,$87,$FA,$D9,$CB,$4F,$89,$E7,$FC,$72
	.dc.b	$38,$3F,$98,$7D,$87,$F2,$54,$D2,$73,$FE,$E5,$B1,$C2,$30,$7F,$14
	.dc.b	$EB,$91,$FE,$F6,$6B,$FE,$CA,$6B,$FE,$FC,$6B,$FE,$4D,$C9,$EC,$FF
	.dc.b	$1E,$24,$7F,$C0,$06,$BF,$ED,$46,$BF,$BC,$F0,$3F,$DE,$91,$FC,$EB
	.dc.b	$E8,$1F,$41,$49,$5F,$EB,$78,$CB,$FD,$47,$B1,$F0,$9F,$C5,$DB,$19
	.dc.b	$8C,$11,$C1,$0F,$EB,$D1,$C1,$FF,$74,$FC,$FF,$F3,$82,$3F,$D8,$F1
	.dc.b	$64,$9C,$CF,$03,$51,$96,$C7,$F4,$2E,$59,$1A,$6B,$FE,$F8,$6B,$F1
	.dc.b	$67,$F6,$2E,$4F,$F6,$C7,$CE,$BD,$0F,$EC,$2D,$8C,$C6,$0F,$E7,$5F
	.dc.b	$7A,$F8,$CA,$49,$FE,$F5,$19,$6C,$65,$FF,$01,$46,$5B,$1F,$C9,$3E
	.dc.b	$71,$F0,$B4,$93,$FE,$B7,$8C,$B7,$D8,$E0,$47,$06,$FE,$07,$A1,$1C
	.dc.b	$1C,$F9,$FC,$04,$70,$77,$E6,$F2,$11,$C1,$0F,$E8,$D7,$FE,$0D,$F1
	.dc.b	$4D,$3F,$C2,$1C,$3F,$7B,$CB,$FE,$F2,$4A,$FB,$AF,$C1,$E7,$91,$EF
	.dc.b	$6B,$D9,$AF,$F5,$2A,$FD,$3B,$B3,$F0,$E7,$F8,$18,$86,$0F,$F7,$07
	.dc.b	$82,$FA,$AF,$EA,$6D,$8C,$C6,$08,$E0,$87,$F6,$58,$E0,$FF,$6C,$1F
	.dc.b	$1A,$F9,$C7,$C7,$7F,$8F,$63,$0D,$8C,$CF,$F9,$16,$72,$F9,$BD,$FF
	.dc.b	$29,$1C,$10,$F1,$39,$69,$FE,$4F,$E7,$7C,$63,$C8,$FF,$1E,$B6,$74
	.dc.b	$D7,$D5,$3C,$27,$FB,$BF,$E2,$6F,$97,$A4,$AF,$F2,$E7,$12,$75,$18
	.dc.b	$6B,$FE,$0B,$6B,$FE,$0A,$6B,$FE,$8E,$8E,$53,$FE,$40,$F9,$C7,$AD
	.dc.b	$49,$5F,$C2,$63,$2F,$8B,$EC,$3F,$A7,$D9,$A4,$F2,$3F,$C2,$2D,$9E
	.dc.b	$11,$83,$C1,$EB,$C8,$EC,$D7,$FC,$49,$3F,$ED,$A7,$C1,$D3,$D0,$8F
	.dc.b	$C4,$F9,$C7,$CE,$FF,$88,$1B,$19,$8C,$11,$C1,$0F,$F1,$94,$70,$7F
	.dc.b	$1D,$D5,$FE,$B4,$F3,$7F,$CC,$B1,$FD,$D8,$FF,$5B,$C6,$5F,$E5,$C6
	.dc.b	$43,$97,$99,$97,$0F,$F4,$B3,$80,$F9,$BF,$C1,$72,$FF,$B1,$39,$DE
	.dc.b	$6F,$F9,$32,$7F,$A0,$46,$2B,$FD,$93,$18,$9D,$71,$7E,$9A,$7F,$4B
	.dc.b	$18,$BE,$8F,$3F,$27,$FC,$AF,$87,$3A,$F1,$46,$27,$84,$72,$3B,$DF
	.dc.b	$5A,$E1,$87,$86,$7C,$77,$C5,$DE,$1E,$06,$9C,$CE,$E6,$4A,$66,$7E
	.dc.b	$1B,$6F,$C1,$87,$A6,$38,$4F,$85,$F9,$53,$A7,$B9,$FC,$26,$39,$B6
	.dc.b	$3E,$33,$6B,$FD,$8D,$2F,$61,$F3,$78,$6D,$7F,$B3,$E7,$53,$93,$FD
	.dc.b	$25,$0F,$9B,$A4,$7F,$D2,$CD,$CF,$E8,$43,$C1,$7C,$1F,$C3,$81,$CC
	.dc.b	$F9,$06,$DD,$39,$9E,$73,$CA,$36,$67,$4F,$E5,$51,$99,$F0,$CE,$4F
	.dc.b	$F4,$CC,$66,$7F,$DD,$97,$E3,$8C,$4F,$F3,$38,$C5,$F8,$9B,$1E,$06
	.dc.b	$DF,$66,$DF,$EA,$BE,$17,$FC,$67,$7F,$E1,$76,$DF,$EC,$9A,$FF,$8C
	.dc.b	$5A,$F1,$F0,$FD,$CE,$5A,$E6,$7A,$46,$21,$D1,$FA,$CF,$83,$C3,$6B
	.dc.b	$F9,$7E,$09,$CB,$63,$31,$82,$38,$3F,$9B,$FC,$DF,$F8,$7D,$BE,$67
	.dc.b	$68,$C6,$1F,$EC,$16,$BF,$E4,$D9,$FE,$94,$7F,$8E,$4A,$FF,$14,$6B
	.dc.b	$BE,$FE,$91,$0F,$09,$63,$B3,$5F,$F6,$6C,$FE,$3A,$36,$77,$F8,$8F
	.dc.b	$8D,$CD,$F9,$26,$C1,$A7,$C8,$F9,$B7,$CD,$D2,$4F,$F5,$88,$CB,$63
	.dc.b	$38,$7F,$B8,$67,$2D,$8F,$0D,$FF,$90,$6B,$FC,$B4,$D9,$FE,$6A,$D7
	.dc.b	$FC,$E8,$D7,$FA,$F5,$3E,$6F,$CC,$4A,$F2,$36,$7C,$4D,$7F,$DE,$33
	.dc.b	$C6,$7F,$C1,$1F,$0F,$FE,$0E,$6F,$1F,$E2,$FA,$13,$9B,$FF,$0B,$46
	.dc.b	$27,$E3,$7C,$54,$BF,$88,$DA,$F8,$4D,$7C,$2D,$7F,$A7,$E4,$F9,$BF
	.dc.b	$8A,$4A,$F8,$AF,$E6,$BB,$1F,$1D,$49,$3F,$E0,$A8,$CB,$63,$37,$FE
	.dc.b	$2D,$E5,$FF,$4A,$C6,$23,$E6,$CF,$D4,$21,$F5,$0F,$07,$C9,$47,$B3
	.dc.b	$63,$A5,$A3,$11,$FC,$70,$FD,$9C,$66,$78,$9E,$7F,$0B,$D8,$47,$07
	.dc.b	$AF,$C4,$F8,$04,$70,$7B,$FE,$0B,$E2,$11,$C1,$F0,$FC,$EF,$90,$47
	.dc.b	$07,$F9,$3F,$0F,$E6,$3D,$8F,$80,$FE,$15,$08,$C4,$EE,$31,$3A,$9C
	.dc.b	$C7,$36,$BF,$E5,$B6,$BF,$0D,$AF,$82,$3F,$8C,$79,$8F,$9B,$7C,$CE
	.dc.b	$1B,$87,$F9,$4D,$B7,$FE,$4E,$6B,$FE,$B5,$F3,$FF,$7B,$7F,$E5,$8A
	.dc.b	$FF,$60,$C6,$08,$E0,$93,$FE,$24,$9E,$0D,$E0,$E6,$49,$FF,$65,$B6
	.dc.b	$35,$18,$23,$F3,$7E,$6F,$F8,$64,$7C,$96,$C6,$7B,$3E,$7C,$62,$3F
	.dc.b	$D0,$4D,$7F,$14,$D7,$FB,$3B,$53,$9B,$7C,$DF,$F9,$BE,$FF,$CF,$4D
	.dc.b	$AF,$F3,$AB,$5F,$CB,$4F,$0F,$27,$E4,$FB,$32,$FF,$DD,$2D,$AF,$8C
	.dc.b	$D7,$FD,$C8,$D7,$FC,$93,$5F,$51,$BA,$FD,$42,$57,$FB,$D1,$B1,$F4
	.dc.b	$5A,$FE,$27,$DF,$FC,$71,$27,$F2,$5C,$08,$C4,$FE,$45,$AF,$19,$FC
	.dc.b	$F0,$F9,$A7,$BF,$FD,$9E,$30,$D8,$CC,$70,$6C,$FE,$17,$D2,$FC,$69
	.dc.b	$27,$F5,$5D,$8C,$D3,$9A,$E1,$5A,$9C,$B6,$7F,$43,$EA,$7D,$92,$70
	.dc.b	$E0,$93,$FE,$B3,$8C,$B6,$3F,$6F,$F3,$7F,$E0,$8C,$3F,$C5,$6D,$4D
	.dc.b	$27,$97,$C0,$95,$FB,$D8,$E4,$FC,$78,$F8,$84,$7E,$23,$FB,$8F,$88
	.dc.b	$7B,$94,$93,$FE,$C9,$8C,$B6,$33,$7F,$EC,$69,$CB,$63,$F7,$3E,$F4
	.dc.b	$AF,$90,$D9,$FD,$B3,$5F,$E5,$DA,$FD,$18,$C4,$7F,$B0,$FE,$69,$F3
	.dc.b	$7F,$E2,$26,$DF,$FB,$41,$B3,$FD,$15,$AF,$4D,$DF,$01,$FC,$30,$95
	.dc.b	$F6,$DE,$9E,$B4,$8E,$8D,$AE,$8D,$7F,$A2,$78,$AC,$3F,$CE,$55,$C0
	.dc.b	$F2,$7C,$DF,$9C,$8E,$0F,$C5,$78,$DF,$C7,$88,$E0,$F6,$7B,$1E,$F2
	.dc.b	$38,$3F,$25,$E3,$7F,$26,$23,$82,$63,$FC,$8E,$32,$D8,$CB,$F1,$63
	.dc.b	$2D,$8F,$E8,$DF,$34,$F9,$BF,$E9,$6D,$8C,$C6,$0F,$E8,$3F,$37,$FE
	.dc.b	$7F,$AF,$ED,$07,$EA,$DC,$3F,$C5,$9A,$FF,$A2,$37,$46,$23,$FD,$1D
	.dc.b	$F2,$AE,$2F,$F3,$27,$B4,$8C,$63,$FD,$86,$31,$3A,$9C,$D7,$FA,$29
	.dc.b	$AF,$FA,$21,$AF,$FA,$62,$78,$D7,$39,$CC,$FF,$B9,$B0,$75,$7F,$54
	.dc.b	$25,$7F,$03,$7F,$CF,$1F,$31,$FE,$1C,$D6,$F8,$C6,$92,$BF,$BC,$4E
	.dc.b	$DB,$3B,$6C,$EF,$0F,$97,$83,$E6,$1F,$09,$FD,$C1,$B7,$FE,$03,$9D
	.dc.b	$4E,$49,$19,$9D,$D7,$46,$BF,$D2,$9A,$FF,$AF,$31,$FF,$1B,$EB,$EC
	.dc.b	$25,$7A,$1B,$3F,$80,$9D,$CE,$49,$1D,$FE,$6F,$D0,$48,$F2,$3F,$67
	.dc.b	$CF,$FF,$62,$47,$D7,$6C,$FD,$7D,$91,$88,$FC,$3B,$67,$FD,$05,$E2
	.dc.b	$5F,$07,$FA,$74,$7F,$B5,$8F,$F1,$E1,$A3,$15,$98,$DB,$FE,$B6,$F9
	.dc.b	$B7,$B9,$FE,$E1,$8F,$F1,$84,$63,$8F,$FC,$58,$D8,$CC,$ED,$FD,$1C
	.dc.b	$3D,$4F,$F1,$A4,$72,$6B,$A6,$BF,$E5,$18,$E1,$E4,$7F,$62,$F9,$27
	.dc.b	$CD,$FA,$AB,$D9,$18,$7D,$67,$FB,$D9,$C3,$FE,$D2,$CD,$18,$7F,$49
	.dc.b	$E2,$73,$9F,$D8,$B7,$3F,$D8,$37,$7F,$E8,$F3,$DB,$97,$F2,$9E,$FF
	.dc.b	$F9,$35,$B1,$F8,$F8,$C7,$2F,$FB,$5E,$3F,$43,$ED,$72,$7F,$10,$D7
	.dc.b	$FC,$E1,$CB,$FE,$98,$C7,$FC,$B0,$D8,$E3,$18,$9E,$11,$F9,$76,$A9
	.dc.b	$E0,$FE,$7F,$E6,$FC,$64,$7F,$48,$D7,$46,$08,$79,$BA,$FC,$04,$7D
	.dc.b	$6D,$74,$60,$87,$F9,$57,$9B,$FC,$E5,$E3,$FF,$83,$DB,$A7,$E7,$27
	.dc.b	$FC,$1D,$1F,$E7,$7D,$BF,$90,$7C,$9D,$DC,$D2,$57,$F1,$96,$BF,$C8
	.dc.b	$5A,$FF,$91,$BE,$6D,$F3,$1F,$5D,$B1,$FC,$25,$AF,$F1,$06,$BF,$CF
	.dc.b	$3E,$6D,$E4,$FF,$A2,$E1,$18,$7F,$D1,$CD,$7F,$D2,$CD,$7F,$BA,$70
	.dc.b	$FC,$02,$3F,$E2,$D6,$BF,$E7,$B6,$BC,$0D,$10,$D1,$4F,$EE,$BF,$33
	.dc.b	$FE,$4D,$7F,$D4,$DE,$52,$31,$1F,$EC,$B6,$BC,$3B,$BE,$9F,$CD,$C8
	.dc.b	$FF,$20,$35,$E1,$F3,$7F,$EA,$BB,$E8,$DB,$34,$C6,$92,$BC,$73,$F3
	.dc.b	$10,$AF,$34,$EE,$3C,$CD,$7D,$AF,$E2,$3D,$E9,$1F,$E5,$56,$BE,$DF
	.dc.b	$87,$F2,$11,$F0,$BF,$90,$F7,$E4,$7F,$97,$5A,$FF,$AA,$DA,$FF,$8A
	.dc.b	$DA,$FF,$79,$F4,$BF,$CE,$6F,$F9,$1F,$D7,$FE,$05,$18,$8F,$F2,$34
	.dc.b	$60,$8E,$C8,$7F,$B6,$FE,$6F,$FC,$E2,$FF,$B5,$DA,$E7,$D7,$FC,$84
	.dc.b	$62,$3F,$C1,$FF,$00,$F4,$7F,$28,$D8,$FF,$70,$35,$FF,$74,$35,$E3
	.dc.b	$1C,$08,$F0,$7F,$8E,$78,$7D,$E3,$FC,$30,$D7,$FD,$10,$DE,$BF,$E5
	.dc.b	$B8,$70,$24,$FF,$75,$9D,$1C,$7B,$3F,$D2,$1F,$13,$FC,$08,$93,$FE
	.dc.b	$8F,$6B,$FE,$DD,$6B,$FE,$5D,$7E,$1F,$9D,$23,$E4,$35,$FF,$19,$8F
	.dc.b	$9B,$FE,$4A,$DE,$8F,$DD,$46,$08,$E0,$91,$C8,$E0,$7F,$A8,$FB,$3F
	.dc.b	$8C,$91,$C1,$0F,$F7,$A3,$96,$9F,$EA,$BF,$67,$F9,$09,$1C,$10,$FF
	.dc.b	$7E,$39,$69,$FE,$1F,$F9,$A7,$CD,$FF,$51,$6C,$66,$30,$7F,$17,$EC
	.dc.b	$7C,$2D,$24,$FB,$63,$2D,$8C,$BF,$2A,$32,$D8,$FE,$73,$F3,$72,$3F
	.dc.b	$D8,$CD,$7F,$CC,$7B,$7F,$B3,$63,$11,$FE,$95,$6B,$FE,$F0,$E5,$FE
	.dc.b	$86,$47,$FD,$A0,$D7,$9F,$17,$EB,$5F,$F4,$F4,$65,$FE,$16,$7F,$B6
	.dc.b	$FE,$6F,$FC,$32,$FF,$91,$5A,$FF,$93,$1A,$F0,$94,$62,$39,$FC,$DB
	.dc.b	$E6,$FF,$A0,$36,$3F,$08,$D7,$F1,$ED,$7F,$D0,$6D,$79,$C7,$22,$3F
	.dc.b	$CD,$BE,$6D,$F0,$BF,$CB,$5B,$19,$8C,$19,$F6,$24,$71,$6B,$FE,$92
	.dc.b	$F0,$55,$FE,$C2,$8E,$DF,$03,$D4,$8F,$87,$9F,$C0,$47,$07,$97,$9B
	.dc.b	$C8,$47,$04,$3C,$15,$FE,$12,$BF,$0C,$EE,$7E,$CF,$95,$FE,$00,$23
	.dc.b	$FB,$6E,$16,$FF,$E4,$98,$FF,$67,$E0,$E6,$F8,$84,$8F,$90,$D7,$C2
	.dc.b	$D7,$8B,$FE,$0E,$61,$DE,$7F,$BD,$57,$82,$3F,$AE,$78,$0F,$9B,$A4
	.dc.b	$8F,$ED,$71,$96,$C6,$5F,$A1,$19,$6C,$7F,$09,$F8,$E7,$CD,$BE,$3F
	.dc.b	$0D,$AF,$E2,$F3,$A9,$C9,$E4,$F7,$A4,$F8,$9A,$F8,$1A,$FF,$BD,$B0
	.dc.b	$FF,$53,$BB,$3F,$D4,$DF,$1A,$F1,$FF,$CB,$0D,$8C,$C6,$0F,$EF,$DF
	.dc.b	$36,$F8,$CA,$49,$FF,$06,$C6,$5B,$19,$7F,$25,$19,$6C,$7F,$90,$BE
	.dc.b	$6F,$FD,$9B,$1F,$81,$6B,$FC,$21,$AF,$EF,$36,$46,$23,$FC,$13,$C8
	.dc.b	$F9,$BF,$EA,$8D,$F1,$FF,$C2,$D1,$82,$38,$24,$FF,$89,$A3,$FA,$A5
	.dc.b	$7F,$94,$E3,$C5,$EA,$F9,$C8,$EF,$A3,$A9,$1C,$1F,$E3,$0F,$9A,$7C
	.dc.b	$DF,$F8,$9D,$B1,$98,$C1,$1C,$10,$FD,$D4,$70,$7F,$95,$9F,$DE,$4F
	.dc.b	$F1,$4E,$7F,$E2,$AF,$2F,$FC,$DB,$19,$11,$97,$F5,$C6,$5C,$3D,$23
	.dc.b	$89,$FE,$5D,$F9,$BF,$C0,$72,$FF,$8D,$B9,$5E,$5F,$F2,$E9,$FE,$87
	.dc.b	$18,$BF,$F1,$DC,$62,$75,$C2,$BE,$51,$86,$D7,$0F,$1B,$F9,$CE,$1C
	.dc.b	$E1,$87,$37,$AE,$1C,$DF,$D5,$61,$CD,$A7,$33,$A9,$CA,$FA,$C6,$EF
	.dc.b	$B5,$FB,$AB,$89,$B3,$8B,$E1,$7C,$77,$E1,$9E,$2D,$AF,$89,$7D,$AB
	.dc.b	$C4,$74,$BE,$D7,$C3,$0F,$F0,$64,$78,$67,$FA,$C9,$DA,$33,$87,$BE
	.dc.b	$33,$3F,$E9,$2F,$9B,$E3,$EB,$9A,$49,$FF,$4B,$7C,$DF,$FA,$5D,$34
	.dc.b	$8E,$87,$CE,$7D,$15,$E0,$BE,$0B,$FC,$EF,$39,$9E,$67,$0D,$BD,$35
	.dc.b	$B8,$C1,$EC,$BE,$33,$AB,$B3,$93,$67,$FA,$12,$B9,$5F,$7B,$FB,$73
	.dc.b	$E4,$AA,$F8,$B5,$F4,$5F,$85,$E1,$1D,$0E,$05,$3D,$37,$C2,$BC,$71
	.dc.b	$FE,$0A,$BA,$F0,$9E,$66,$DF,$6F,$03,$E1,$3F,$DA,$B7,$FE,$9C,$DF
	.dc.b	$E1,$4E,$7D,$0F,$89,$18,$AF,$D5,$D6,$CF,$DD,$BC,$AF,$F5,$11,$8A
	.dc.b	$E2,$D3,$84,$7E,$FA,$79,$BC,$0C,$4F,$07,$F8,$1F,$3B,$BB,$18,$FF
	.dc.b	$2C,$75,$D7,$FB,$EF,$9D,$CD,$6F,$2E,$F7,$B6,$C6,$70,$F0,$5F,$19
	.dc.b	$CE,$1E,$19,$7C,$0F,$B5,$FC,$FB,$5F,$F1,$CB,$5F,$F6,$2B,$5E,$3C
	.dc.b	$3C,$E3,$DF,$AF,$E3,$E7,$AE,$99,$AF,$81,$E2,$3F,$2A,$7C,$0E,$EF
	.dc.b	$FD,$A3,$39,$8C,$47,$F8,$4A,$77,$9F,$FA,$2A,$B4,$D8,$CC,$FE,$E2
	.dc.b	$72,$D8,$CB,$FE,$4D,$6B,$FC,$E3,$91,$F1,$DF,$49,$B7,$F7,$2F,$9E
	.dc.b	$7F,$8A,$9C,$DF,$DD,$E0,$FD,$A7,$C5,$FA,$1B,$87,$F8,$AC,$46,$5A
	.dc.b	$65,$C1,$1C,$1F,$E3,$3E,$57,$C8,$7E,$19,$B1,$F8,$23,$9D,$F8,$E3
	.dc.b	$2F,$97,$EF,$72,$FF,$79,$6B,$FD,$61,$AF,$F5,$66,$E0,$69,$30,$F7
	.dc.b	$3A,$C7,$FC,$6D,$87,$E5,$1B,$7E,$1A,$F9,$9E,$57,$E7,$A9,$F3,$BF
	.dc.b	$0C,$8F,$C7,$6D,$7F,$A6,$23,$66,$48,$FF,$A7,$32,$78,$7F,$14,$ED
	.dc.b	$56,$BF,$4F,$97,$2B,$FF,$51,$B5,$F8,$93,$EC,$24,$FF,$98,$E7,$53
	.dc.b	$F1,$E7,$2F,$FA,$B3,$2F,$F0,$6D,$7C,$1F,$0F,$F7,$77,$DF,$6F,$C4
	.dc.b	$47,$EF,$3F,$13,$FF,$35,$57,$FB,$93,$A1,$18,$C7,$FD,$10,$D7,$FD
	.dc.b	$57,$1A,$F3,$7F,$4A,$DA,$E5,$5B,$AF,$F1,$9B,$54,$ED,$FF,$4A,$7C
	.dc.b	$A3,$E1,$7F,$87,$B6,$FF,$8D,$D6,$F2,$FE,$94,$A7,$57,$FC,$3B,$03
	.dc.b	$ED,$3D,$DE,$C5,$C3,$F0,$9F,$5F,$FA,$5D,$8D,$24,$7F,$A4,$63,$FD
	.dc.b	$D6,$FE,$2F,$DC,$FE,$47,$F9,$97,$E3,$34,$FD,$94,$7F,$2B,$5F,$64
	.dc.b	$F5,$FF,$84,$C6,$1B,$19,$BF,$F3,$EB,$73,$FE,$FF,$5B,$9C,$D7,$02
	.dc.b	$3F,$B3,$EF,$FD,$91,$E1,$BE,$55,$FD,$3B,$9B,$FA,$43,$FC,$1F,$E5
	.dc.b	$3F,$86,$1F,$84,$FA,$BB,$87,$E9,$E2,$E5,$FB,$79,$B5,$CF,$93,$F8
	.dc.b	$01,$FA,$9F,$84,$B8,$7F,$42,$C7,$F1,$15,$B8,$FE,$09,$3A,$C3,$FA
	.dc.b	$97,$77,$F1,$2F,$96,$FE,$67,$DA,$7E,$57,$CB,$BE,$B7,$31,$31,$FE
	.dc.b	$45,$5B,$7F,$AF,$E4,$BE,$11,$AA,$FE,$55,$ED,$48,$C1,$FC,$15,$FE
	.dc.b	$35,$F0,$72,$3F,$97,$7D,$E7,$F8,$75,$F6,$46,$74,$FE,$38,$FF,$4A
	.dc.b	$9D,$46,$27,$F8,$87,$D9,$7F,$39,$3F,$63,$F5,$D6,$32,$49,$FE,$49
	.dc.b	$19,$7F,$BB,$74,$BE,$33,$F8,$CC,$7E,$70,$FF,$18,$B8,$7F,$C4,$8D
	.dc.b	$73,$E2,$FF,$58,$9E,$4D,$9D,$46,$2F,$FC,$9B,$5C,$1F,$E7,$DD,$BF
	.dc.b	$CF,$4F,$F0,$34,$EE,$7F,$9E,$47,$F5,$5F,$B2,$95,$FC,$73,$EB,$ED
	.dc.b	$7F,$53,$B5,$63,$87,$FC,$43,$E2,$AF,$18,$7F,$A6,$39,$FF,$C6,$75
	.dc.b	$F9,$9C,$3F,$BB,$FE,$C0,$DF,$FA,$84,$4B,$FC,$83,$A8,$C6,$7F,$2B
	.dc.b	$EE,$BF,$81,$D7,$BA,$91,$89,$FF,$2A,$B5,$F9,$D1,$A7,$E5,$46,$5F
	.dc.b	$F2,$93,$5F,$F3,$1B,$63,$84,$FF,$09,$FA,$2F,$EF,$47,$CA,$9F,$06
	.dc.b	$3F,$DE,$63,$FB,$55,$CA,$FE,$D5,$F3,$F6,$3F,$B5,$16,$BF,$C9,$32
	.dc.b	$9C,$3F,$E0,$9F,$B4,$FF,$04,$1F,$D0,$3E,$C6,$C7,$F6,$38,$B7,$FD
	.dc.b	$CE,$6D,$FF,$9D,$7D,$B7,$E6,$FF,$D6,$B0,$FC,$2C,$63,$CF,$FF,$11
	.dc.b	$B6,$33,$3B,$C3,$FD,$2F,$E6,$3C,$97,$BB,$FC,$A0,$FF,$28,$61,$C6
	.dc.b	$BF,$C8,$B3,$AC,$3F,$BC,$57,$F8,$84,$E0,$E0,$93,$F8,$D8,$CB,$63
	.dc.b	$FB,$6F,$87,$FE,$04,$23,$FD,$A5,$FE,$81,$EF,$FF,$8B,$4F,$EC,$AD
	.dc.b	$8F,$EE,$BB,$FF,$83,$08,$FF,$71,$7F,$A4,$7C,$07,$F8,$D4,$FE,$DA
	.dc.b	$D8,$FF,$08,$45,$BF,$F1,$04,$DD,$BF,$C7,$13,$6B,$FC,$41,$36,$BF
	.dc.b	$C4,$1F,$47,$FE,$1C,$24,$FF,$80,$5F,$EB,$1E,$87,$F9,$14,$FE,$FA
	.dc.b	$D8,$FF,$23,$7E,$AC,$ED,$FF,$2C,$F9,$08,$C4,$FF,$96,$5B,$87,$F9
	.dc.b	$5D,$AF,$F0,$F9,$E1,$5C,$A7,$33,$FE,$66,$6A,$9D,$D7,$F0,$CF,$55
	.dc.b	$F8,$83,$CD,$FF,$46,$B6,$7C,$F8,$E2,$7F,$CE,$BF,$10,$E0,$F4,$B6
	.dc.b	$BF,$CE,$0D,$CB,$FC,$DC,$D7,$D0,$D7,$FC,$F6,$D7,$CC,$DC,$8E,$72
	.dc.b	$F3,$F9,$1E,$63,$C7,$85,$C3,$FD,$3F,$9A,$71,$8F,$F5,$CC,$3C,$9C
	.dc.b	$3F,$EB,$EA,$D5,$FF,$9E,$65,$D3,$EC,$24,$7B,$1B,$7E,$C6,$BE,$C7
	.dc.b	$FC,$31,$F0,$1F,$E8,$53,$E3,$B6,$7F,$D0,$DF,$98,$FF,$86,$2B,$FD
	.dc.b	$24,$D7,$EE,$35,$FF,$4F,$B5,$FF,$14,$F9,$5F,$12,$7F,$D2,$1F,$9E
	.dc.b	$3E,$CB,$FD,$F5,$E2,$BE,$8F,$FA,$A5,$EF,$0F,$EF,$B7,$BC,$3F,$C6
	.dc.b	$77,$B9,$FF,$1A,$FD,$43,$9D,$E2,$FF,$52,$94,$E6,$30,$D8,$FE,$AE
	.dc.b	$D7,$FB,$21,$C3,$AB,$FD,$E0,$E9,$AE,$A3,$F9,$47,$C3,$74,$9C,$CF
	.dc.b	$F6,$D9,$7C,$7F,$F6,$FE,$46,$7F,$EC,$4D,$3F,$D0,$F7,$EC,$DB,$1E
	.dc.b	$FF,$F7,$5D,$BD,$D3,$EA,$C7,$96,$1F,$E5,$4D,$5F,$81,$3E,$33,$E0
	.dc.b	$46,$1B,$19,$BF,$EB,$0D,$7F,$CF,$D1,$D1,$75,$7C,$A8,$C3,$63,$F8
	.dc.b	$1F,$3E,$FA,$FF,$81,$3F,$3D,$B1,$FC,$0E,$31,$DD,$FA,$48,$C1,$0F
	.dc.b	$E0,$D1,$A3,$EC,$57,$EF,$BC,$BF,$0F,$5F,$F7,$7A,$FD,$3C,$78,$6D
	.dc.b	$F5,$DA,$89,$3A,$3F,$8C,$C6,$CF,$19,$1F,$E3,$0D,$7F,$8C,$35,$FE
	.dc.b	$31,$3D,$9F,$E2,$55,$FE,$12,$7F,$73,$E8,$79,$8F,$89,$FA,$07,$EE
	.dc.b	$E3,$FD,$DC,$D7,$FD,$F4,$D7,$F9,$23,$5F,$F4,$7E,$68,$C4,$7F,$23
	.dc.b	$6C,$67,$BF,$E8,$46,$0D,$E7,$FA,$2E,$4F,$B3,$C1,$F8,$AE,$6F,$F7
	.dc.b	$F4,$7D,$CF,$45,$F3,$34,$FA,$2E,$33,$FF,$63,$46,$0F,$AB,$6F,$AA
	.dc.b	$EA,$49,$FF,$69,$FD,$9D,$7F,$C7,$11,$71,$FF,$1B,$CD,$E6,$FF,$56
	.dc.b	$E7,$FB,$8A,$C5,$FF,$BD,$EB,$CC,$79,$08,$FF,$B6,$5F,$1F,$C8,$7B
	.dc.b	$89,$1F,$EE,$4B,$F7,$1E,$C2,$3F,$EE,$17,$D7,$DD,$F1,$89,$1F,$EC
	.dc.b	$6F,$0F,$F7,$2F,$FC,$01,$8C,$BF,$E5,$37,$F5,$1E,$77,$F3,$2E,$BB
	.dc.b	$5C,$29,$A7,$FB,$3A,$30,$D9,$FD,$87,$9C,$FC,$59,$F1,$B8,$6E,$1F
	.dc.b	$E2,$86,$BF,$C5,$A7,$84,$FF,$1F,$6C,$7F,$BF,$7D,$FF,$BC,$47,$FD
	.dc.b	$59,$F2,$F5,$F2,$EB,$F4,$B9,$3F,$D4,$D1,$FE,$E0,$CF,$FC,$DB,$7F
	.dc.b	$C0,$BC,$16,$1F,$B4,$F3,$A5,$FF,$5A,$7F,$83,$FC,$8B,$E7,$33,$3D
	.dc.b	$8F,$E1,$BD,$DF,$EE,$F7,$81,$DF,$4F,$C1,$D7,$18,$C1,$FE,$03,$C1
	.dc.b	$F6,$FF,$C4,$5B,$1A,$EE,$F2,$CF,$F1,$BE,$AF,$F0,$24,$F2,$6B,$FC
	.dc.b	$F9,$AF,$F9,$DB,$D0,$FE,$6A,$FF,$9F,$5A,$ED,$AE,$F1,$FE,$5F,$18
	.dc.b	$8F,$E6,$8D,$7D,$FE,$7F,$F1,$A2,$3E,$7F,$6D,$18,$8F,$8D,$DB,$FE
	.dc.b	$1E,$8C,$71,$79,$33,$5F,$0B,$F8,$B6,$E3,$7D,$07,$EF,$49,$3F,$EE
	.dc.b	$BF,$82,$AF,$69,$DB,$C9,$AF,$8B,$93,$FD,$D4,$FF,$85,$BC,$9F,$E6
	.dc.b	$64,$7E,$A8,$FA,$47,$C7,$7F,$4B,$8C,$3F,$DC,$DA,$7F,$75,$8D,$B6
	.dc.b	$35,$18,$8F,$F6,$5B,$5F,$EE,$AD,$7F,$DB,$4D,$7F,$AC,$DB,$EB,$BF
	.dc.b	$8F,$12,$3E,$3E,$1F,$8D,$7E,$14,$65,$A6,$B9,$FF,$A8,$F8,$B2,$3F
	.dc.b	$C6,$EF,$F6,$1F,$C5,$DF,$09,$99,$FB,$3D,$5F,$CB,$F9,$BF,$8E,$B6
	.dc.b	$3F,$CB,$95,$C1,$FA,$1F,$55,$FC,$CC,$8F,$E6,$E1,$F2,$9D,$C9,$19
	.dc.b	$7F,$BC,$73,$3F,$51,$7B,$1A,$AF,$D8,$74,$30,$E7,$1C,$4F,$D6,$DF
	.dc.b	$1A,$FE,$DF,$E5,$49,$FF,$2A,$BF,$BC,$F4,$31,$1E,$E8,$B5,$FA,$8E
	.dc.b	$97,$81,$FD,$9E,$BF,$1D,$F5,$5F,$D3,$C8,$FD,$2C,$2B,$A3,$F2,$51
	.dc.b	$FC,$3E,$3F,$7D,$ED,$49,$FD,$F7,$85,$FE,$49,$7F,$23,$D2,$F9,$0F
	.dc.b	$C8,$36,$3F,$1D,$1F,$8C,$23,$82,$1F,$8E,$72,$FF,$90,$B2,$FD,$67
	.dc.b	$DE,$FF,$67,$24,$7F,$45,$F4,$3F,$B5,$78,$BF,$DA,$09,$1F,$E9,$7E
	.dc.b	$DF,$E2,$AF,$FB,$6B,$CC,$FE,$17,$C1,$FD,$0F,$A7,$FA,$39,$FD,$2B
	.dc.b	$F0,$2F,$D7,$D7,$F2,$7F,$39,$F3,$CF,$8D,$FE,$9E,$D9,$FE,$4B,$3A
	.dc.b	$9E,$5C,$72,$7F,$CB,$62,$57,$F5,$60,$48,$CC,$FF,$9A,$5B,$1C,$9A
	.dc.b	$FF,$BD,$27,$93,$C5,$AF,$76,$EF,$FC,$57,$8F,$FC,$20,$49,$FD,$9F
	.dc.b	$B3,$FD,$34,$FF,$20,$F8,$7F,$E4,$78,$FA,$4D,$7C,$E7,$F9,$59,$C3
	.dc.b	$FE,$0A,$6B,$98,$E6,$47,$FD,$95,$A3,$8B,$FC,$BA,$D8,$FE,$56,$D7
	.dc.b	$F9,$9F,$3B,$E2,$3F,$AE,$9F,$E5,$67,$11,$FE,$D5,$6B,$F8,$96,$BE
	.dc.b	$8C,$BF,$4B,$3C,$49,$1C,$5F,$F1,$8F,$A5,$FE,$35,$24,$7F,$18,$FA
	.dc.b	$6F,$EE,$47,$C6,$23,$FE,$E5,$CD,$18,$8F,$92,$D7,$C5,$1A,$7E,$1F
	.dc.b	$2F,$C2,$23,$FD,$33,$17,$CE,$7F,$7C,$3E,$01,$23,$E0,$35,$F8,$0D
	.dc.b	$7E,$04,$72,$7F,$27,$DF,$FE,$0A,$F7,$BF,$BD,$12,$3F,$BD,$7D,$E3
	.dc.b	$E1,$BF,$D1,$CD,$8E,$2D,$7F,$D6,$F4,$8C,$47,$E1,$A3,$13,$FE,$95
	.dc.b	$9C,$CF,$F9,$B9,$AA,$76,$F6,$9F,$EB,$1F,$85,$FF,$97,$E3,$FA,$FB
	.dc.b	$5F,$F0,$E0,$71,$D3,$FE,$08,$7F,$A0,$46,$7C,$9E,$64,$7F,$DA,$D1
	.dc.b	$FE,$BE,$7F,$0D,$EC,$BE,$77,$FD,$62,$D8,$FE,$FE,$D7,$F8,$37,$D1
	.dc.b	$7F,$A9,$5F,$F0,$4C,$67,$0F,$E6,$8D,$E5,$FF,$37,$B6,$7F,$82,$57
	.dc.b	$1A,$FE,$75,$1F,$D5,$F9,$7F,$82,$E5,$FE,$C2,$8C,$1E,$EC,$3F,$CF
	.dc.b	$2E,$23,$46,$FC,$5F,$F0,$17,$73,$E2,$FF,$CC,$74,$FA,$2F,$F2,$DC
	.dc.b	$BE,$33,$F7,$B1,$86,$FB,$BF,$C0,$27,$96,$9F,$C6,$2B,$8E,$1D,$1E
	.dc.b	$18,$66,$F8,$61,$CA,$3F,$8F,$E1,$C9,$F1,$4F,$5B,$F2,$FB,$FC,$A4
	.dc.b	$D3,$BE,$3F,$CA,$1B,$1C,$8E,$B8,$77,$7F,$9C,$B7,$0F,$06,$3C,$0E
	.dc.b	$C4,$C7,$C5,$87,$A6,$FC,$F8,$FC,$87,$F9,$09,$FD,$11,$C3,$C0,$FE
	.dc.b	$7F,$7E,$D7,$6B,$E3,$5E,$FF,$F4,$49,$FE,$9B,$18,$78,$B5,$E2,$D7
	.dc.b	$FA,$5F,$5F,$F8,$ED,$34,$8F,$F8,$FD,$F8,$D8,$FF,$9B,$F7,$B3,$EA
	.dc.b	$9A,$FF,$52,$F7,$7F,$CE,$3B,$7F,$A7,$A3,$FA,$AC,$E9,$4E,$AF,$C7
	.dc.b	$1F,$59,$EC,$D3,$B1,$4E,$D7,$CD,$F4,$99,$3A,$B9,$AF,$05,$77,$9E
	.dc.b	$77,$E5,$8C,$47,$F5,$49,$DD,$71,$8E,$93,$C7,$85,$C7,$FE,$BE,$AF
	.dc.b	$EE,$B1,$8C,$7F,$D1,$4D,$8C,$CE,$EF,$FC,$28,$D7,$F8,$3D,$70,$8C
	.dc.b	$47,$EF,$A3,$FD,$29,$E2,$7F,$B3,$3D,$CB,$1F,$E9,$0B,$FE,$F8,$DA
	.dc.b	$FC,$6D,$6E,$73,$5C,$08,$E0,$FC,$57,$D0,$BE,$67,$FD,$40,$D8,$CC
	.dc.b	$60,$CF,$9B,$23,$8B,$70,$FC,$ED,$3F,$6D,$FE,$A3,$5F,$E1,$A8,$C7
	.dc.b	$83,$FE,$AF,$6C,$66,$77,$5C,$9A,$FF,$82,$27,$6E,$6B,$94,$EA,$3F
	.dc.b	$C4,$7D,$CF,$A2,$C4,$7F,$45,$6C,$66,$BF,$D8,$0D,$C3,$FC,$4D,$F7
	.dc.b	$C7,$E8,$B5,$C9,$82,$18,$25,$FD,$28,$E0,$FF,$94,$E1,$F3,$9F,$86
	.dc.b	$69,$FE,$1C,$70,$7E,$03,$F0,$CE,$33,$FE,$28,$DD,$18,$8F,$F2,$03
	.dc.b	$5F,$F2,$33,$63,$75,$CC,$87,$F9,$0E,$38,$71,$7F,$B4,$6F,$F3,$3E
	.dc.b	$8F,$93,$0F,$F3,$53,$5F,$F1,$13,$5F,$F2,$87,$C2,$FF,$86,$30,$FF
	.dc.b	$27,$36,$3F,$26,$D7,$D4,$D7,$8F,$93,$F9,$E8,$C4,$7F,$97,$DB,$91
	.dc.b	$CE,$3F,$57,$C7,$F9,$44,$C3,$FC,$8A,$FE,$A3,$E8,$DF,$4D,$49,$3F
	.dc.b	$E6,$38,$CB,$63,$38,$FE,$AA,$72,$D8,$F8,$B5,$2B,$FD,$0C,$D9,$F7
	.dc.b	$FD,$5F,$AC,$93,$F1,$1B,$1A,$8C,$11,$F0,$73,$7A,$49,$1E,$77,$D9
	.dc.b	$E4,$C8,$F5,$36,$7D,$07,$C1,$24,$7B,$DF,$E1,$7F,$31,$FE,$E0,$8F
	.dc.b	$F4,$23,$63,$F8,$4B,$5F,$C8,$35,$FD,$95,$FF,$06,$24,$FF,$0A,$6C
	.dc.b	$7E,$FD,$AF,$F0,$47,$F5,$8F,$F6,$2F,$0D,$F4,$4F,$A3,$CC,$FF,$A1
	.dc.b	$A7,$31,$88,$FF,$4F,$C6,$2F,$FD,$4A,$D8,$FC,$BF,$D1,$FE,$68,$93
	.dc.b	$FE,$BB,$6C,$6A,$30,$47,$F6,$DF,$37,$FB,$52,$47,$FA,$41,$B1,$A8
	.dc.b	$C1,$1F,$67,$D1,$BE,$3F,$D1,$0C,$13,$9A,$F4,$B6,$78,$D7,$0A,$DB
	.dc.b	$67,$53,$F1,$5A,$FC,$46,$BF,$E6,$3E,$77,$C6,$7F,$1E,$6E,$1F,$E7
	.dc.b	$A6,$CF,$F9,$C9,$AF,$F9,$C7,$E8,$FF,$86,$76,$1A,$4A,$FE,$52,$D7
	.dc.b	$FD,$0A,$D9,$FE,$13,$EE,$FF,$0C,$24,$FF,$BB,$5B,$1A,$8C,$11,$FE
	.dc.b	$91,$B7,$F5,$42,$47,$F5,$86,$C6,$A3,$04,$7F,$C5,$FE,$33,$E8,$FF
	.dc.b	$35,$1F,$E1,$26,$C6,$7E,$2B,$EB,$B6,$BF,$BE,$F7,$3E,$8F,$F9,$29
	.dc.b	$23,$2D,$C3,$F9,$3B,$5F,$E4,$EF,$F4,$EF,$91,$7C,$CD,$25,$7F,$AB
	.dc.b	$23,$2D,$8C,$D7,$FA,$9A,$72,$D8,$FE,$EF,$F1,$D2,$BF,$BD,$B6,$7F
	.dc.b	$B8,$35,$FE,$EA,$D7,$FC,$71,$F4,$67,$E0,$B6,$7F,$CA,$F1,$FE,$38
	.dc.b	$8C,$3F,$31,$AF,$FB,$96,$1F,$3D,$FD,$74,$8F,$FB,$E1,$B1,$FD,$8D
	.dc.b	$AF,$17,$FA,$8F,$C3,$49,$FA,$2D,$9F,$EA,$6D,$7F,$A9,$B5,$F5,$0F
	.dc.b	$D4,$7F,$46,$DE,$7F,$D9,$B6,$BF,$CD,$53,$FA,$8F,$51,$F4,$4F,$83
	.dc.b	$FD,$CC,$61,$B1,$9A,$F8,$B3,$97,$F9,$5F,$AD,$FC,$B8,$8E,$09,$87
	.dc.b	$F2,$E8,$CB,$63,$F9,$9F,$B5,$FC,$D8,$8E,$0F,$C9,$F7,$7F,$76,$23
	.dc.b	$83,$F9,$DF,$A7,$FC,$F8,$8E,$0F,$CD,$ED,$FD,$F8,$8E,$09,$7F,$E0
	.dc.b	$08,$CB,$63,$2F,$F8,$0A,$32,$D8,$FE,$97,$EB,$FF,$4E,$23,$83,$F5
	.dc.b	$7F,$43,$FE,$0E,$23,$83,$FC,$8D,$F9,$B3,$C7,$FF,$23,$36,$33,$18
	.dc.b	$23,$82,$5F,$F9,$4A,$78,$3F,$EB,$0E,$F7,$CF,$E2,$3F,$0F,$18,$BF
	.dc.b	$F2,$1C,$E6,$7F,$C9,$6D,$5E,$B2,$DE,$BF,$98,$98,$FA,$9B,$5E,$36
	.dc.b	$BF,$E5,$B9,$E3,$3F,$AE,$C5,$F1,$0D,$FF,$CA,$AD,$9D,$36,$FD,$8D
	.dc.b	$7D,$6D,$7F,$CC,$6D,$55,$DA,$BA,$9F,$E7,$3F,$D5,$7C,$73,$FE,$7A
	.dc.b	$6B,$FE,$71,$6B,$FC,$1F,$24,$62,$3E,$2B,$63,$39,$75,$26,$3E,$AE
	.dc.b	$8F,$69,$1C,$1F,$ED,$DF,$91,$7C,$27,$EE,$1B,$19,$8C,$19,$D2,$47
	.dc.b	$16,$E5,$FC,$6B,$E8,$DC,$3F,$ED,$28,$C3,$5F,$F7,$AC,$6B,$C2,$FF
	.dc.b	$77,$36,$39,$72,$7F,$B0,$9B,$EF,$6A,$31,$1C,$9A,$FF,$14,$97,$C5
	.dc.b	$7F,$BE,$61,$18,$7B,$C6,$63,$11,$E9,$F8,$BC,$11,$F2,$BF,$CD,$9F
	.dc.b	$BB,$CB,$F8,$A7,$F1,$8D,$7F,$44,$D7,$F4,$2D,$7E,$E7,$D4,$FF,$B0
	.dc.b	$89,$AF,$FB,$19,$B1,$FE,$BB,$6B,$FE,$C0,$7F,$96,$F2,$FE,$BB,$E3
	.dc.b	$FC,$8D,$9F,$1C,$62,$CD,$24,$68,$87,$FB,$33,$77,$87,$FE,$D8,$6C
	.dc.b	$7F,$B3,$5A,$EE,$34,$47,$4D,$7F,$82,$7D,$03,$E8,$7C,$ED,$8F,$DA
	.dc.b	$35,$DC,$78,$7E,$52,$47,$85,$AF,$9D,$AF,$FB,$66,$D1,$88,$FE,$E1
	.dc.b	$EE,$3C,$B7,$C9,$FD,$B8,$C3,$63,$33,$FE,$14,$8F,$C0,$7B,$EE,$AF
	.dc.b	$F2,$84,$FF,$78,$8C,$79,$3A,$9C,$CF,$F6,$61,$3A,$7F,$C3,$FF,$43
	.dc.b	$DC,$8E,$08,$7F,$B1,$1C,$B4,$C9,$FE,$C6,$72,$D3,$C9,$2F,$96,$F2
	.dc.b	$12,$32,$FB,$F9,$24,$7F,$8C,$9A,$FA,$6F,$E0,$7C,$9F,$B4,$91,$D9
	.dc.b	$F8,$1D,$12,$3C,$2D,$79,$61,$FA,$E7,$FD,$DC,$D7,$FC,$49,$5F,$36
	.dc.b	$78,$47,$F8,$3B,$C9,$7D,$0B,$E2,$BF,$65,$18,$6C,$66,$7F,$C0,$ED
	.dc.b	$7F,$C5,$BF,$92,$3E,$7F,$F8,$23,$63,$F7,$8D,$77,$1F,$85,$F8,$DF
	.dc.b	$C5,$C7,$F8,$DB,$8D,$18,$8F,$BD,$18,$E4,$E0,$48,$E1,$3B,$7F,$C5
	.dc.b	$33,$B7,$33,$F9,$3F,$96,$FD,$6C,$69,$AF,$E6,$9A,$A5,$18,$6F,$4F
	.dc.b	$32,$4F,$A6,$78,$F8,$4F,$A2,$FF,$1B,$B7,$8F,$F8,$F3,$6B,$FC,$AA
	.dc.b	$D7,$FC,$AF,$F6,$0F,$87,$FE,$5C,$D8,$FE,$4C,$D7,$71,$A2,$3A,$6B
	.dc.b	$FE,$71,$F4,$3F,$8D,$FB,$06,$92,$3F,$CF,$4D,$78,$18,$3E,$7F,$C0
	.dc.b	$7D,$92,$38,$21,$F6,$37,$FF,$5C,$B8,$3E,$AB,$87,$FC,$EB,$F3,$7F
	.dc.b	$85,$7F,$78,$D7,$FC,$C6,$D7,$FC,$FC,$D5,$8B,$A7,$E3,$92,$3E,$3B
	.dc.b	$5F,$9A,$D7,$F8,$54,$71,$7F,$17,$BF,$E5,$48,$E0,$87,$B5,$CB,$4F
	.dc.b	$F1,$17,$CE,$3D,$DF,$E9,$2D,$8C,$C6,$08,$E0,$87,$F8,$FE,$38,$37
	.dc.b	$E7,$B9,$91,$EE,$FF,$69,$F8,$57,$CF,$3D,$1C,$36,$BF,$D4,$73,$A9
	.dc.b	$C9,$23,$33,$B7,$F9,$74,$ED,$CC,$FE,$AF,$C8,$91,$FA,$16,$BF,$B8
	.dc.b	$E6,$FE,$CA,$E4,$86,$63,$F8,$11,$1F,$DD,$B5,$FC,$53,$5F,$C4,$B5
	.dc.b	$FC,$2F,$A5,$FE,$13,$DF,$FC,$B7,$1F,$E0,$B6,$BF,$8E,$8F,$D5,$FD
	.dc.b	$1C,$8F,$D5,$B5,$FD,$8B,$5F,$C2,$B5,$FE,$57,$E1,$7F,$5A,$7F,$98
	.dc.b	$F8,$DF,$D5,$48,$FF,$46,$E9,$FE,$B0,$47,$07,$F5,$DF,$9F,$FD,$13
	.dc.b	$FE,$C0,$F1,$11,$88,$FE,$72,$D7,$FA,$FB,$63,$79,$FF,$6A,$8F,$9F
	.dc.b	$3F,$D3,$5A,$F5,$6B,$FD,$13,$E5,$DF,$07,$FD,$C9,$B1,$FD,$C5,$AF
	.dc.b	$F5,$76,$BE,$76,$BC,$E3,$91,$1F,$E6,$BE,$5F,$F3,$F2,$38,$3E,$17
	.dc.b	$C7,$FA,$48,$E0,$87,$AD,$CB,$4F,$E6,$1F,$03,$FC,$F4,$8E,$0F,$57
	.dc.b	$97,$ED,$23,$82,$1E,$D7,$2D,$3F,$A6,$FC,$97,$F5,$72,$38,$3E,$D7
	.dc.b	$85,$F4,$48,$E0,$87,$D3,$72,$D3,$FA,$47,$CA,$FF,$55,$23,$83,$FC
	.dc.b	$45,$F5,$BF,$E5,$D7,$FC,$3E,$DD,$CD,$26,$7F,$E2,$C6,$CF,$28,$FF
	.dc.b	$2D,$FC,$FF,$FB,$56,$3F,$D7,$0D,$7F,$D9,$CD,$7F,$D9,$AD,$7F,$D5
	.dc.b	$1F,$33,$FE,$81,$F7,$8D,$24,$7F,$9E,$E3,$2D,$8C,$BF,$E3,$93,$FD
	.dc.b	$A6,$FF,$A2,$D7,$CF,$BE,$7B,$FD,$1C,$D8,$FF,$43,$35,$E4,$D7,$FC
	.dc.b	$88,$BE,$93,$89,$23,$8B,$FE,$50,$FA,$07,$C6,$D2,$4F,$F9,$46,$32
	.dc.b	$D8,$CB,$FE,$39,$8C,$B6,$3F,$CC,$0B,$E8,$29,$1F,$F0,$F1,$92,$B9
	.dc.b	$7F,$C3,$E6,$4A,$FF,$AB,$F9,$9F,$3E,$F0,$94,$F8,$AF,$F1,$A5,$3E
	.dc.b	$2F,$33,$9F,$5E,$CE,$63,$8D,$70,$C3,$A4,$6C,$EB,$3E,$0C,$37,$7C
	.dc.b	$B0,$EB,$87,$81,$B1,$DB,$0E,$73,$D6,$B8,$E1,$D2,$F6,$D3,$23,$0F
	.dc.b	$03,$E3,$31,$7D,$30,$F1,$47,$09,$F6,$3B,$78,$E1,$E1,$8C,$5F,$81
	.dc.b	$EC,$D9,$DD,$7A,$70,$FF,$6A,$47,$8A,$BF,$D2,$0D,$74,$7E,$6D,$7C
	.dc.b	$63,$D9,$E2,$D9,$D3,$5E,$6D,$7F,$49,$E7,$FE,$8C,$8F,$FB,$BD,$B9
	.dc.b	$7F,$9C,$61,$D2,$F0,$7E,$4C,$7E,$12,$AC,$E6,$3E,$51,$92,$B9,$AE
	.dc.b	$91,$FE,$72,$33,$5D,$1C,$47,$86,$78,$67,$E8,$9D,$9D,$8D,$BF,$EB
	.dc.b	$78,$DC,$7F,$AE,$6F,$CB,$7D,$5B,$1C,$5B,$7F,$E9,$1F,$55,$F0,$DF
	.dc.b	$E9,$7B,$FF,$44,$36,$FF,$D0,$ED,$7F,$D0,$6D,$7F,$DA,$ED,$7F,$CF
	.dc.b	$CD,$9E,$B8,$FC,$79,$E2,$D7,$FD,$56,$FF,$9A,$6B,$FD,$AC,$D8,$FF
	.dc.b	$5E,$35,$FE,$F8,$D7,$8F,$A7,$CD,$FF,$6C,$C6,$5B,$19,$EC,$FF,$5A
	.dc.b	$3B,$87,$4B,$E6,$B0,$DC,$3F,$98,$46,$0C,$F9,$32,$3F,$9A,$36,$BF
	.dc.b	$DD,$55,$FC,$EB,$BF,$C8,$DA,$FE,$6C,$D9,$DF,$62,$B6,$DA,$DB,$67
	.dc.b	$73,$A3,$E0,$7C,$1B,$E7,$F5,$3F,$C5,$A3,$13,$A9,$CC,$FF,$1E,$9C
	.dc.b	$DF,$F4,$9D,$1F,$3F,$FA,$18,$C3,$6B,$FD,$DC,$D7,$FC,$5C,$D7,$FD
	.dc.b	$CA,$D5,$3D,$27,$E9,$4E,$63,$11,$FE,$D7,$1F,$2F,$F6,$F8,$3F,$64
	.dc.b	$D8,$C9,$B7,$EE,$7C,$FF,$32,$46,$08,$7B,$5C,$B4,$ED,$F0,$9D,$31
	.dc.b	$FF,$72,$B5,$FF,$70,$B5,$FE,$8B,$68,$C4,$7B,$BE,$7D,$E3,$FE,$3D
	.dc.b	$B7,$F3,$5B,$3F,$D2,$8E,$13,$B2,$4E,$DB,$1F,$EF,$88,$FF,$6D,$1F
	.dc.b	$1F,$DD,$F8,$84,$7E,$D7,$CF,$BC,$CF,$D0,$36,$33,$18,$23,$82,$57
	.dc.b	$97,$55,$71,$8F,$8B,$7F,$20,$95,$FE,$11,$CF,$E2,$47,$C1,$F9,$E9
	.dc.b	$3F,$05,$B5,$D7,$34,$62,$3F,$CF,$11,$89,$E0,$7B,$39,$BD,$64,$70
	.dc.b	$7C,$0F,$94,$F7,$91,$C1,$F1,$2B,$E1,$91,$C1,$FA,$1F,$9F,$FF,$67
	.dc.b	$37,$DC,$FE,$11,$18,$9F,$88,$D7,$F7,$F8,$FF,$08,$25,$7F,$8D,$DB
	.dc.b	$1F,$C0,$1A,$F0,$8F,$F6,$4F,$CF,$BE,$6F,$F8,$D7,$8A,$8C,$4E,$A7
	.dc.b	$31,$8A,$FC,$BF,$CF,$FE,$68,$95,$F4,$F9,$11,$89,$DC,$E6,$31,$1F
	.dc.b	$E1,$CE,$0F,$EC,$1F,$2F,$FA,$F2,$57,$ED,$5B,$3F,$93,$6B,$C6,$11
	.dc.b	$EA,$F9,$F7,$C7,$52,$4F,$FB,$A2,$32,$D8,$CD,$7F,$B9,$A7,$2D,$8F
	.dc.b	$E3,$FD,$2F,$86,$A4,$BF,$F3,$17,$67,$F0,$A8,$FE,$4B,$F3,$FF,$C4
	.dc.b	$E0,$D2,$4E,$88,$72,$72,$FC,$BF,$9F,$FE,$0E,$47,$04,$C7,$F8,$4C
	.dc.b	$65,$B1,$FE,$79,$E0,$FE,$9A,$47,$04,$8F,$E8,$B1,$96,$C7,$F9,$F7
	.dc.b	$C9,$7C,$FF,$E4,$67,$FC,$AF,$18,$E6,$FE,$C4,$DA,$E1,$5B,$A7,$07
	.dc.b	$F1,$9F,$6F,$F9,$09,$2B,$F1,$0D,$9F,$E4,$6D,$7F,$91,$BF,$D4,$3E
	.dc.b	$2B,$FC,$FF,$3F,$41,$B3,$FC,$77,$54,$62,$3F,$AA,$35,$FE,$A4,$D5
	.dc.b	$1C,$3D,$3E,$C4,$AF,$F5,$17,$C4,$7F,$76,$23,$82,$57,$F6,$E8,$CB
	.dc.b	$63,$E5,$7C,$FF,$F9,$F9,$B1,$F2,$23,$0F,$F8,$B5,$AF,$E7,$7E,$67
	.dc.b	$FA,$91,$27,$FA,$C3,$63,$F9,$B3,$5E,$11,$FB,$3F,$83,$FE,$9E,$47
	.dc.b	$04,$8F,$EA,$31,$96,$C7,$FA,$F5,$7C,$FF,$E9,$1F,$E9,$4D,$7F,$D2
	.dc.b	$AD,$7F,$D3,$6D,$F5,$1F,$3E,$F8,$2F,$F1,$44,$61,$B1,$9B,$FF,$66
	.dc.b	$4E,$5F,$E4,$DE,$B7,$F2,$42,$38,$26,$1F,$C9,$A3,$2D,$8F,$E5,$DE
	.dc.b	$C7,$F2,$C2,$38,$3F,$8C,$76,$FF,$74,$23,$83,$F9,$B7,$A7,$FC,$D0
	.dc.b	$8E,$0F,$E4,$1A,$FF,$7C,$23,$82,$5F,$F7,$E8,$CB,$63,$2F,$F8,$02
	.dc.b	$32,$D8,$FE,$8D,$EB,$7F,$44,$23,$83,$F9,$87,$CE,$7F,$83,$08,$E0
	.dc.b	$FF,$20,$FC,$FB,$C8,$FF,$22,$B6,$33,$18,$23,$82,$5F,$F9,$3A,$78
	.dc.b	$3F,$AF,$F9,$AA,$FF,$90,$5A,$FF,$8F,$DA,$FE,$B9,$BE,$B2,$5E,$BF
	.dc.b	$90,$97,$E8,$6D,$F8,$5A,$FF,$B9,$2B,$94,$7F,$95,$3E,$0F,$FA,$1D
	.dc.b	$7F,$34,$6D,$7F,$05,$6B,$FE,$57,$6B,$FE,$55,$6B,$FC,$11,$BD,$26
	.dc.b	$9A,$B7,$CF,$FA,$AF,$FC,$DE,$DB,$FF,$3A,$B5,$F4,$35,$FF,$80,$33
	.dc.b	$CB,$0F,$F3,$CC,$E6,$7F,$D0,$ED,$9D,$57,$B7,$9F,$E0,$11,$C1,$FB
	.dc.b	$9F,$8D,$7C,$FF,$FA,$45,$B1,$98,$C1,$9C,$A4,$71,$6E,$3F,$E4,$91
	.dc.b	$F3,$EE,$0F,$F6,$CC,$66,$31,$E1,$7F,$09,$6C,$66,$76,$FF,$94,$E1
	.dc.b	$C7,$FE,$C4,$8E,$4D,$74,$D4,$48,$E3,$EF,$77,$F9,$AF,$E3,$13,$E6
	.dc.b	$6C,$9A,$63,$E7,$E4,$F9,$DA,$FE,$77,$2F,$E4,$5C,$4E,$4F,$C1,$4F
	.dc.b	$F8,$01,$AE,$CF,$F5,$7F,$CF,$FF,$AF,$09,$B7,$F2,$BE,$24,$62,$77
	.dc.b	$39,$8C,$47,$EB,$7E,$3B,$C3,$66,$92,$3C,$4D,$78,$18,$3F,$D8,$38
	.dc.b	$BC,$2F,$13,$63,$31,$82,$38,$21,$FE,$CE,$8E,$0F,$F6,$FF,$9F,$7C
	.dc.b	$FF,$C6,$6C,$7F,$89,$1A,$EE,$3C,$3F,$37,$23,$CA,$D7,$B3,$5F,$DF
	.dc.b	$52,$31,$1F,$EA,$4F,$51,$F2,$5F,$D5,$FC,$BF,$F5,$33,$63,$33,$FE
	.dc.b	$55,$8F,$EB,$9E,$F3,$AF,$FC,$8F,$3F,$ED,$36,$BF,$E9,$36,$BF,$90
	.dc.b	$3F,$D9,$AE,$1F,$17,$C8,$F7,$23,$FE,$63,$6B,$FE,$BE,$6B,$FE,$BE
	.dc.b	$6B,$C8,$E0,$43,$C5,$F3,$FE,$89,$34,$91,$FE,$71,$E2,$F8,$04,$70
	.dc.b	$7B,$7E,$6F,$91,$1C,$1F,$07,$9F,$B1,$1C,$10,$F8,$8E,$5A,$7F,$78
	.dc.b	$F9,$F7,$C5,$7F,$82,$7C,$A4,$62,$75,$39,$8C,$47,$F0,$DF,$9F,$7C
	.dc.b	$FF,$EE,$9B,$1F,$E1,$36,$BB,$8F,$C0,$7C,$6F,$E1,$63,$FD,$88,$D7
	.dc.b	$FD,$C1,$C4,$8C,$47,$FA,$45,$AF,$FB,$B9,$AF,$F9,$61,$B1,$C7,$8F
	.dc.b	$B9,$F8,$FF,$9A,$FD,$1D,$7F,$A6,$DA,$FF,$BE,$1A,$FF,$98,$9B,$26
	.dc.b	$93,$A3,$A1,$FE,$31,$F0,$DF,$3E,$E1,$C3,$70,$FF,$3B,$35,$FF,$7F
	.dc.b	$4F,$09,$FE,$47,$F3,$EF,$88,$A4,$AF,$F1,$F4,$65,$B1,$97,$FA,$04
	.dc.b	$65,$B1,$FC,$93,$D7,$79,$F4,$93,$FE,$75,$6B,$FE,$75,$8D,$B9,$21
	.dc.b	$F5,$B6,$FF,$57,$3F,$6B,$DE,$FB,$A4,$70,$43,$83,$97,$FA,$B7,$CE
	.dc.b	$7E,$19,$FD,$C0,$70,$FE,$ED,$AF,$F4,$C6,$AC,$5C,$FF,$28,$91,$F2
	.dc.b	$1A,$FC,$66,$BE,$08,$E2,$FE,$53,$6F,$CB,$91,$C1,$0F,$84,$E5,$A7
	.dc.b	$FB,$3B,$E7,$DE,$EF,$F5,$46,$C6,$63,$04,$70,$43,$FD,$4D,$1C,$1B
	.dc.b	$F3,$FC,$C8,$F7,$7F,$D0,$5F,$0E,$F9,$F7,$A3,$86,$D7,$F8,$16,$75
	.dc.b	$39,$24,$66,$76,$FD,$49,$DB,$99,$FD,$67,$8F,$23,$FC,$21,$CF,$FA
	.dc.b	$78,$C4,$7E,$F3,$9F,$FB,$23,$A6,$B9,$8D,$BF,$BF,$7F,$83,$91,$FE
	.dc.b	$0E,$D7,$F2,$4D,$78,$1F,$EA,$3D,$FF,$CA,$ED,$F4,$FF,$C1,$F1,$88
	.dc.b	$FC,$8C,$7E,$C3,$E7,$E4,$7E,$B9,$AF,$EB,$1A,$FE,$1D,$AF,$F8,$7B
	.dc.b	$C4,$FE,$AD,$E2,$FF,$41,$24,$7F,$57,$7F,$A2,$74,$C8,$FE,$B4,$D7
	.dc.b	$F4,$32,$F9,$FF,$D6,$C6,$1A,$FF,$5D,$8D,$73,$7F,$66,$9C,$CF,$8D
	.dc.b	$F9,$AE,$DA,$E7,$3E,$07,$F6,$EF,$97,$7C,$15,$24,$FF,$92,$F0,$4E
	.dc.b	$A3,$0D,$7D,$2D,$7C,$CD,$7E,$33,$5F,$E7,$1E,$57,$F4,$42,$3F,$D1
	.dc.b	$1F,$67,$C8,$C8,$F1,$62,$8C,$47,$C5,$6B,$EF,$6B,$FC,$F7,$E0,$3F
	.dc.b	$A1,$11,$FC,$93,$F0,$7C,$B9,$1E,$C6,$BE,$C6,$BE,$C6,$BF,$D4,$FE
	.dc.b	$4B,$FA,$E9,$1F,$EB,$AF,$D5,$F0,$A4,$7A,$9A,$FD,$D6,$BF,$5D,$AF
	.dc.b	$F5,$5F,$96,$FE,$B4,$47,$F5,$4F,$F8,$9B,$E7,$FF,$CB,$B1,$F8,$56
	.dc.b	$BF,$95,$6A,$D5,$18,$6E,$9C,$C9,$3F,$81,$9E,$3E,$AB,$E7,$FF,$DA
	.dc.b	$D3,$FE,$8F,$8C,$7B,$FF,$E7,$29,$CC,$FF,$3C,$9C,$BF,$E9,$5F,$5D
	.dc.b	$F3,$3F,$E8,$26,$C7,$F9,$C6,$30,$47,$04,$39,$3F,$E8,$A5,$F3,$CF
	.dc.b	$9F,$FF,$46,$B6,$3F,$D4,$86,$A3,$26,$BE,$7E,$46,$A5,$18,$8F,$DA
	.dc.b	$FC,$FB,$E3,$3F,$CA,$AD,$8F,$D3,$35,$DC,$68,$8E,$9A,$FF,$88,$07
	.dc.b	$CF,$FF,$99,$5A,$7F,$2C,$29,$B7,$44,$74,$D7,$FD,$85,$D0,$F9,$F7
	.dc.b	$8B,$FE,$B4,$F8,$8F,$F1,$DC,$BE,$2B,$FC,$D1,$18,$6F,$B1,$FE,$6B
	.dc.b	$9E,$55,$C6,$31,$7D,$30,$61,$C5,$D1,$FE,$C2,$BE,$CE,$0F,$E1,$73
	.dc.b	$C6,$7C,$6D,$AF,$06,$1E,$2A,$E4,$D7,$8E,$1C,$F0,$F2,$C7,$8A,$B8
	.dc.b	$B9,$AE,$13,$AC,$3B,$DE,$6A,$1C,$4F,$46,$1E,$36,$C7,$E2,$AB,$BC
	.dc.b	$7F,$86,$1A,$FC,$48,$D7,$C5,$BD,$AA,$49,$FA,$B1,$97,$FD,$C7,$E9
	.dc.b	$7E,$6C,$8E,$09,$F4,$1F,$E7,$25,$D9,$FE,$4D,$8C,$47,$F9,$A4,$3B
	.dc.b	$8D,$6B,$FE,$2E,$9D,$4E,$7C,$5F,$F7,$BB,$CA,$ED,$7F,$95,$A6,$D7
	.dc.b	$F8,$DA,$7E,$D3,$E2,$8A,$F8,$75,$E8,$FC,$B3,$FD,$69,$3C,$91,$C4
	.dc.b	$F3,$D7,$96,$35,$19,$BF,$23,$A3,$FD,$7D,$87,$B1,$F2,$19,$3A,$B9
	.dc.b	$F2,$DF,$0C,$FA,$DF,$F3,$A6,$DF,$E7,$59,$CC,$63,$4F,$2C,$70,$FB
	.dc.b	$5C,$62,$8D,$CF,$AA,$F9,$61,$D6,$30,$72,$71,$5F,$B2,$8C,$5F,$4B
	.dc.b	$E7,$39,$3F,$D2,$51,$9F,$38,$F8,$BC,$1F,$F5,$CF,$1E,$62,$61,$E0
	.dc.b	$BE,$2D,$AD,$B9,$C3,$FD,$39,$5E,$CA,$76,$3B,$FF,$D2,$D1,$FE,$EB
	.dc.b	$8C,$5F,$F2,$A6,$C7,$F3,$B9,$DB,$F7,$23,$FC,$A3,$3E,$38,$EB,$E7
	.dc.b	$F7,$9E,$3A,$FF,$A8,$DD,$CF,$C1,$3C,$7E,$77,$46,$DF,$86,$31,$26
	.dc.b	$92,$34,$49,$FE,$A1,$C7,$E1,$3C,$53,$67,$E2,$0F,$7F,$D7,$A7,$FA
	.dc.b	$ED,$B1,$93,$FC,$BE,$5B,$FF,$6C,$B5,$FC,$E9,$F8,$77,$0F,$E3,$1A
	.dc.b	$E6,$3B,$F2,$3E,$4F,$F0,$8D,$AF,$2C,$E2,$7F,$11,$0F,$96,$F2,$91
	.dc.b	$F2,$B6,$FC,$AD,$7C,$B8,$7F,$B5,$E3,$C3,$F1,$BE,$03,$AB,$6B,$F1
	.dc.b	$9A,$BE,$EB,$FD,$3B,$1F,$E8,$B6,$E1,$FE,$87,$6B,$F0,$A7,$87,$B1
	.dc.b	$F8,$66,$D7,$18,$C5,$70,$AF,$D6,$35,$78,$CF,$63,$F7,$77,$F9,$F8
	.dc.b	$C7,$47,$ED,$B1,$EB,$7F,$AB,$9C,$D7,$DE,$F7,$BE,$D9,$1F,$B6,$D9
	.dc.b	$FB,$6D,$7E,$3C,$7F,$BD,$4F,$F4,$FF,$03,$CF,$FA,$E7,$D4,$AB,$3F
	.dc.b	$B1,$F9,$F7,$9D,$FA,$6E,$A4,$62,$7F,$CD,$CD,$C7,$FD,$06,$D7,$FD
	.dc.b	$00,$D7,$FD,$20,$D7,$FC,$78,$D9,$EB,$5F,$8C,$3F,$16,$7F,$09,$8E
	.dc.b	$D9,$FF,$21,$CA,$E1,$D3,$E7,$74,$6F,$D2,$B1,$FC,$2F,$E8,$1F,$41
	.dc.b	$EB,$6C,$7F,$A1,$DB,$A7,$FA,$71,$AF,$E4,$30,$FF,$3D,$4F,$F1,$36
	.dc.b	$C7,$F2,$1D,$3E,$2F,$C2,$FE,$8B,$1F,$B9,$F3,$FE,$FD,$0D,$26,$DF
	.dc.b	$8D,$6E,$5F,$ED,$F6,$E3,$FE,$EC,$F1,$BF,$9D,$F1,$7E,$2D,$EB,$78
	.dc.b	$E4,$FF,$0D,$E1,$E5,$E0,$FD,$99,$FB,$0F,$9F,$B1,$92,$63,$FC,$E7
	.dc.b	$A9,$09,$DC,$E7,$81,$BF,$EF,$8F,$DD,$FC,$E5,$BF,$89,$16,$BA,$42
	.dc.b	$71,$BF,$69,$FD,$2F,$D9,$BE,$67,$DA,$DB,$FE,$65,$E8,$3E,$7F,$35
	.dc.b	$27,$F5,$55,$BB,$FE,$65,$7B,$9C,$D7,$F1,$4F,$1E,$46,$0F,$B4,$FF
	.dc.b	$18,$F2,$A4,$7F,$2C,$F7,$6E,$1F,$06,$6E,$7F,$DC,$6B,$FD,$A2,$7E
	.dc.b	$0E,$3F,$90,$F8,$D2,$3F,$9A,$FC,$7D,$BF,$95,$36,$F9,$4D,$AF,$EE
	.dc.b	$9D,$1F,$CA,$6F,$F9,$99,$FE,$62,$70,$FF,$15,$6B,$C3,$C5,$FC,$51
	.dc.b	$27,$F1,$4D,$8E,$7C,$8E,$3F,$EA,$55,$F9,$BF,$1F,$F2,$F7,$FC,$E9
	.dc.b	$FE,$77,$DD,$27,$FA,$7F,$05,$C7,$F0,$93,$74,$F8,$93,$6B,$FC,$27
	.dc.b	$8F,$F6,$69,$FD,$C7,$93,$F9,$AF,$9F,$FC,$7D,$7E,$AE,$3F,$A4,$91
	.dc.b	$FE,$4D,$1A,$C7,$F2,$53,$FE,$1C,$F3,$9F,$3F,$FC,$BD,$BE,$47,$F7
	.dc.b	$C6,$E1,$FE,$49,$6D,$FC,$C6,$BF,$E3,$4A,$E4,$4A,$FE,$BB,$E6,$FF
	.dc.b	$5C,$3C,$11,$FD,$23,$2F,$E2,$51,$FD,$82,$E6,$1F,$CF,$7E,$2F,$FD
	.dc.b	$60,$FF,$07,$8C,$E3,$FA,$B7,$F7,$73,$A8,$C4,$FF,$0D,$7F,$B5,$9F
	.dc.b	$DA,$3E,$66,$C6,$49,$1F,$AE,$8C,$BF,$E3,$3F,$51,$F3,$AF,$07,$FC
	.dc.b	$3F,$18,$6C,$66,$FF,$CD,$D3,$97,$FC,$E1,$E3,$BE,$21,$F5,$FF,$E3
	.dc.b	$B6,$CF,$F8,$E6,$B8,$63,$FC,$8E,$72,$FF,$0D,$FA,$87,$C6,$7F,$92
	.dc.b	$5B,$7F,$E3,$FA,$DF,$07,$F9,$76,$73,$7C,$6F,$81,$E1,$C6,$E1,$FE
	.dc.b	$22,$F9,$DF,$F0,$F9,$1F,$F0,$FB,$FE,$24,$BF,$B1,$1F,$E6,$66,$DF
	.dc.b	$F8,$7E,$7C,$9C,$7F,$E1,$6F,$8B,$FF,$14,$92,$FF,$C5,$2F,$3F,$9B
	.dc.b	$F5,$1B,$8F,$F2,$B6,$08,$CC,$61,$D6,$1F,$0D,$CB,$63,$FC,$77,$58
	.dc.b	$CF,$F8,$77,$C4,$FF,$8E,$89,$3F,$E3,$A7,$D9,$F6,$1F,$2C,$FF,$16
	.dc.b	$B6,$3F,$CD,$9C,$AF,$C1,$1A,$FE,$95,$B3,$FA,$56,$E5,$FA,$56,$BF
	.dc.b	$E9,$86,$BF,$E9,$66,$BF,$EB,$C6,$D7,$7A,$FA,$1B,$7D,$F3,$F7,$13
	.dc.b	$DB,$2F,$BD,$1F,$D7,$3E,$77,$FC,$EB,$87,$F9,$4B,$D1,$46,$23,$FD
	.dc.b	$63,$18,$E9,$FF,$3C,$B7,$3F,$F6,$33,$67,$F9,$4B,$5E,$95,$CD,$FF
	.dc.b	$1A,$B7,$C2,$70,$BE,$5B,$F2,$C7,$97,$1B,$8F,$DE,$8B,$87,$F8,$DF
	.dc.b	$0F,$F4,$2C,$FE,$36,$7F,$BB,$FC,$67,$F5,$5A,$F6,$35,$F5,$B5,$FF
	.dc.b	$23,$B7,$23,$49,$E0,$FC,$07,$86,$7E,$8F,$9C,$FD,$61,$E0,$9B,$8F
	.dc.b	$F9,$3F,$CE,$FF,$0D,$E3,$F2,$9A,$FF,$95,$9B,$8F,$F6,$FF,$13,$E6
	.dc.b	$FD,$7F,$EE,$AB,$EA,$35,$FA,$6D,$7F,$CB,$ED,$5F,$41,$F9,$0E,$5F
	.dc.b	$F7,$D4,$63,$95,$18,$9D,$4F,$FB,$FD,$AF,$F6,$25,$E0,$FF,$B9,$23
	.dc.b	$4D,$74,$D7,$FD,$37,$1C,$3D,$CF,$E8,$5F,$5C,$F8,$BF,$1E,$08,$C5
	.dc.b	$EA,$73,$18,$8F,$C5,$FE,$83,$FC,$69,$FC,$7B,$5C,$F8,$1E,$ED,$93
	.dc.b	$99,$FF,$4C,$F4,$3D,$3F,$F4,$77,$17,$F1,$36,$CF,$FB,$DE,$74,$FF
	.dc.b	$A1,$FA,$1F,$29,$AC,$7F,$4A,$F7,$8C,$CF,$E8,$9A,$FE,$89,$AF,$E8
	.dc.b	$A3,$81,$1E,$0F,$D6,$E7,$FD,$1F,$D7,$F9,$1B,$3F,$C7,$A3,$C5,$F4
	.dc.b	$4E,$9D,$64,$9A,$4A,$FF,$50,$56,$DB,$5F,$1E,$31,$1F,$B8,$6B,$FC
	.dc.b	$0D,$AF,$F0,$26,$BF,$EA,$E6,$E4,$69,$AC,$3F,$7B,$F6,$5E,$62,$57
	.dc.b	$F2,$6F,$01,$E1,$FD,$2E,$4F,$E5,$0F,$E1,$A3,$33,$FC,$9E,$30,$47
	.dc.b	$04,$3F,$95,$3F,$2F,$47,$D9,$7C,$62,$4F,$FB,$EB,$87,$E7,$1F,$E6
	.dc.b	$57,$FD,$EA,$FD,$33,$0F,$FA,$F1,$E9,$F1,$BF,$40,$90,$3F,$CF,$9F
	.dc.b	$18,$FA,$0C,$C7,$F7,$39,$BE,$2F,$F9,$CF,$65,$EE,$BF,$AD,$D6,$B0
	.dc.b	$FE,$D0,$FE,$0E,$75,$3D,$B0,$FC,$44,$62,$3F,$DE,$ED,$7C,$97,$DA
	.dc.b	$7E,$27,$C6,$FE,$20,$93,$E9,$C7,$C9,$6F,$9A,$F0,$1D,$89,$3F,$E6
	.dc.b	$61,$87,$AA,$3E,$87,$C8,$FE,$50,$93,$FE,$0F,$ED,$AF,$F8,$56,$2E
	.dc.b	$1F,$E9,$5F,$3F,$FC,$41,$5F,$B4,$6D,$71,$21,$FE,$2F,$FA,$2F,$F5
	.dc.b	$2B,$FE,$23,$6B,$FE,$7A,$F4,$11,$88,$FA,$C2,$33,$F0,$5F,$C9,$9B
	.dc.b	$1E,$DC,$FF,$AA,$FC,$4F,$EB,$89,$1F,$1F,$87,$E6,$EC,$F9,$EC,$CF
	.dc.b	$F8,$C6,$AE,$BF,$32,$7F,$6F,$F4,$32,$BF,$C3,$2F,$F0,$DF,$5C,$E6
	.dc.b	$3F,$C7,$4D,$AF,$F5,$D7,$62,$B7,$5F,$E2,$2F,$17,$F8,$19,$1F,$31
	.dc.b	$F9,$5F,$57,$E2,$DF,$E7,$DF,$89,$1F,$11,$AE,$F7,$FE,$92,$FF,$94
	.dc.b	$DB,$1A,$8C,$1F,$D3,$9C,$1C,$B6,$FE,$FB,$F3,$1F,$C5,$89,$3E,$4C
	.dc.b	$9F,$10,$F5,$F3,$3F,$E5,$9A,$B9,$7C,$0A,$D4,$FF,$90,$3E,$5B,$F9
	.dc.b	$09,$1F,$7E,$3F,$AC,$F1,$71,$1F,$E6,$39,$B7,$F9,$E9,$FE,$79,$D6
	.dc.b	$FA,$FA,$4B,$F9,$27,$F9,$CF,$E0,$3F,$A3,$B4,$F8,$0F,$C4,$23,$82
	.dc.b	$1F,$A7,$D9,$1A,$8F,$F6,$2F,$80,$F8,$CF,$E7,$64,$8F,$5F,$3F,$F8
	.dc.b	$51,$FF,$2F,$C6,$5F,$F3,$F3,$FC,$17,$D5,$FE,$87,$DB,$27,$FC,$CA
	.dc.b	$FF,$09,$F9,$DC,$47,$FA,$16,$2E,$DF,$C0,$27,$F9,$1F,$AB,$2B,$E5
	.dc.b	$F6,$23,$FB,$17,$D7,$FA,$A7,$ED,$35,$5F,$E5,$A3,$0D,$D7,$E4,$FD
	.dc.b	$DB,$E2,$F9,$C4,$BE,$6E,$A3,$F9,$AF,$D0,$C9,$F1,$BF,$C8,$FE,$1F
	.dc.b	$11,$E0,$CF,$F8,$F8,$9C,$E3,$FC,$72,$BF,$9F,$FD,$94,$BF,$6B,$FC
	.dc.b	$B7,$E6,$31,$1C,$CF,$E3,$B8,$27,$37,$FC,$9A,$BF,$C0,$9F,$41,$FA
	.dc.b	$CA,$FF,$33,$B6,$DD,$C8,$C5,$FF,$A1,$7E,$53,$FC,$22,$DA,$FE,$58
	.dc.b	$D7,$F8,$B9,$F3,$3C,$F4,$FF,$8F,$27,$53,$FC,$71,$AF,$F3,$B9,$DB
	.dc.b	$D1,$AF,$48,$DB,$EE,$FB,$2F,$EC,$64,$7F,$84,$F9,$4F,$9E,$7C,$A7
	.dc.b	$F3,$5F,$A3,$FF,$0E,$D7,$F6,$1A,$46,$21,$18,$F0,$3F,$CA,$6D,$8F
	.dc.b	$F6,$FD,$BE,$53,$FA,$03,$63,$C2,$D7,$EA,$4F,$8F,$E4,$BF,$BC,$FD
	.dc.b	$44,$9F,$CB,$BC,$BE,$1A,$47,$67,$C3,$66,$92,$34,$43,$E8,$DF,$C1
	.dc.b	$8F,$F0,$27,$D5,$7F,$82,$09,$3F,$B6,$F9,$CF,$EE,$C7,$F7,$42,$3F
	.dc.b	$D2,$FC,$24,$62,$3E,$EF,$91,$FE,$02,$8C,$47,$F7,$A7,$FC,$B9,$C8
	.dc.b	$F9,$0F,$C7,$92,$7F,$9E,$F1,$BF,$0B,$FF,$3B,$C6,$1A,$FD,$F6,$BF
	.dc.b	$87,$6B,$FE,$01,$6B,$FB,$A6,$C7,$5F,$87,$FC,$84,$60,$DE,$2E,$97
	.dc.b	$D4,$FF,$8B,$09,$87,$F1,$AF,$9C,$FF,$0D,$1F,$E1,$82,$3F,$15,$AF
	.dc.b	$C4,$D5,$18,$8F,$F0,$EB,$FE,$35,$F2,$FF,$C7,$24,$8F,$E4,$FD,$7F
	.dc.b	$E2,$73,$FC,$4A,$47,$E9,$35,$EC,$D7,$FC,$52,$FF,$60,$FA,$57,$9D
	.dc.b	$FD,$01,$B3,$FD,$11,$AF,$F0,$70,$BD,$0E,$07,$32,$47,$FA,$D2,$34
	.dc.b	$D8,$CB,$FE,$2D,$8C,$B6,$3C,$9E,$AF,$8C,$91,$E3,$F5,$FC,$C7,$A4
	.dc.b	$8F,$77,$FD,$1D,$E6,$7E,$4A,$3F,$B3,$B5,$FE,$41,$BA,$31,$C5,$FE
	.dc.b	$35,$FA,$57,$C1,$7F,$A5,$67,$3C,$5F,$E4,$E6,$E1,$FE,$44,$6F,$52
	.dc.b	$F8,$CF,$F7,$63,$FB,$01,$7D,$0F,$F5,$03,$83,$C6,$7F,$B3,$F1,$FF
	.dc.b	$34,$39,$6C,$70,$38,$DF,$F7,$6E,$C7,$BD,$FD,$52,$DF,$4A,$F9,$CF
	.dc.b	$F2,$1D,$3D,$FF,$F2,$2C,$E6,$BF,$CF,$93,$97,$FC,$FD,$19,$7F,$92
	.dc.b	$5F,$58,$C4,$F8,$32,$DE,$1D,$B0,$EA,$DC,$32,$76,$C3,$94,$79,$1F
	.dc.b	$46,$5D,$2B,$FC,$B2,$DA,$FF,$48,$3E,$09,$DD,$71,$BE,$B5,$FC,$E6
	.dc.b	$FC,$2F,$B1,$AF,$AA,$37,$87,$A6,$3E,$3D,$79,$DA,$71,$37,$7E,$87
	.dc.b	$D5,$5F,$E8,$57,$A1,$FE,$87,$AE,$8F,$F5,$C6,$BF,$4B,$E0,$7F,$CE
	.dc.b	$8E,$DA,$ED,$AE,$FD,$FF,$F2,$AA,$F9,$BA,$47,$FC,$B2,$DC,$FE,$C4
	.dc.b	$3E,$17,$F8,$C3,$FE,$31,$CF,$FD,$39,$5F,$31,$FE,$E3,$3A,$AE,$35
	.dc.b	$DC,$E0,$7F,$A3,$23,$53,$CA,$32,$7F,$B1,$63,$2F,$F7,$CA,$EC,$F1
	.dc.b	$E1,$AF,$FA,$5A,$2B,$E7,$8A,$F8,$6F,$D7,$5D,$23,$81,$E0,$38,$14
	.dc.b	$E1,$7E,$CA,$E7,$7D,$AB,$CB,$C7,$FC,$07,$D2,$FF,$70,$E1,$FB,$B6
	.dc.b	$C6,$6F,$FD,$4D,$C3,$FE,$3B,$6C,$EA,$31,$5F,$D7,$9A,$FF,$AC,$AB
	.dc.b	$93,$FD,$BB,$1F,$DB,$DF,$FB,$37,$8F,$F6,$DC,$1E,$26,$C6,$67,$75
	.dc.b	$C6,$B5,$18,$8F,$F5,$FD,$70,$75,$5D,$7E,$CF,$FC,$04,$DF,$1B,$B4
	.dc.b	$62,$7F,$C0,$2D,$7F,$C6,$7F,$4C,$F9,$5F,$8B,$08,$C5,$FE,$85,$B1
	.dc.b	$FE,$E7,$6B,$CB,$E5,$38,$C1,$A4,$9E,$4D,$9E,$4D,$7C,$7E,$BF,$22
	.dc.b	$4F,$85,$F5,$FC,$9C,$8F,$80,$D9,$F6,$B5,$F8,$F8,$A3,$11,$FE,$6C
	.dc.b	$FA,$6F,$C0,$D1,$A4,$7E,$61,$F2,$88,$60,$91,$F3,$1C,$B5,$FF,$78
	.dc.b	$0F,$A6,$EE,$47,$2F,$E0,$FB,$BF,$DE,$11,$F8,$76,$CF,$F8,$E1,$AF
	.dc.b	$E0,$9A,$FF,$95,$DA,$FF,$95,$9B,$B1,$A6,$BA,$9F,$4D,$FE,$D7,$BF
	.dc.b	$F3,$BB,$6F,$FC,$D8,$D7,$FC,$E8,$D7,$C1,$EF,$FA,$49,$3E,$96,$C7
	.dc.b	$A5,$AF,$A5,$FA,$3F,$4D,$27,$E3,$B6,$7E,$07,$9F,$F1,$49,$3F,$15
	.dc.b	$B1,$A8,$C1,$1F,$91,$E4,$FC,$D2,$47,$CD,$7E,$87,$91,$23,$E8,$36
	.dc.b	$7E,$87,$9D,$F5,$49,$3F,$55,$FB,$1F,$0F,$23,$EC,$36,$7F,$D3,$E3
	.dc.b	$F2,$DF,$F4,$7C,$FF,$4B,$6C,$69,$AF,$FA,$99,$A8,$8E,$1B,$EC,$7E
	.dc.b	$E2,$31,$1D,$E3,$F8,$FE,$CF,$A3,$FF,$28,$DF,$FA,$D1,$B5,$FE,$A9
	.dc.b	$6B,$FE,$B1,$6B,$FE,$D2,$FA,$6F,$D3,$92,$7F,$3A,$D8,$FF,$40,$C2
	.dc.b	$31,$1F,$C9,$63,$07,$E9,$2F,$F6,$24,$70,$48,$E7,$19,$6C,$7F,$73
	.dc.b	$F9,$6F,$F7,$D3,$7E,$17,$F8,$CC,$62,$7E,$8B,$5F,$F6,$8B,$5F,$98
	.dc.b	$1C,$3F,$3D,$AE,$63,$FA,$4F,$BE,$FA,$6F,$E5,$53,$F5,$63,$1D,$FA
	.dc.b	$9C,$CF,$FB,$A1,$BE,$17,$FA,$37,$DC,$FE,$22,$4C,$BF,$83,$B6,$BE
	.dc.b	$5B,$5F,$94,$FF,$18,$F7,$24,$FF,$0E,$6C,$FD,$26,$BF,$45,$AF,$C0
	.dc.b	$F0,$7F,$D6,$53,$FC,$F7,$8F,$F9,$C9,$1C,$12,$7F,$A0,$46,$5F,$C4
	.dc.b	$7C,$1F,$E3,$B8,$4D,$25,$7F,$22,$D3,$F9,$79,$1C,$12,$BF,$13,$19
	.dc.b	$6C,$7E,$AB,$57,$D2,$FF,$AD,$63,$0D,$9F,$B6,$27,$73,$F6,$27,$53
	.dc.b	$F8,$09,$CC,$F2,$34,$43,$FB,$05,$FF,$5C,$23,$82,$4F,$F6,$38,$CB
	.dc.b	$63,$F8,$0F,$92,$FA,$4F,$E5,$13,$FE,$37,$6C,$67,$E3,$BF,$C2,$AD
	.dc.b	$AF,$81,$F2,$1F,$95,$6F,$27,$E4,$A3,$18,$7C,$0C,$91,$88,$F7,$35
	.dc.b	$FC,$B3,$5F,$8C,$D8,$E3,$3F,$D9,$3D,$E9,$7F,$D6,$9B,$7F,$BE,$6B
	.dc.b	$FB,$D6,$BF,$AC,$FB,$07,$C4,$FF,$83,$1B,$5F,$E6,$B6,$CF,$F2,$66
	.dc.b	$BF,$B4,$F4,$9F,$4B,$F9,$68,$FB,$B1,$8E,$3F,$F1,$DB,$6F,$F9,$7B
	.dc.b	$67,$F0,$7E,$33,$E3,$DF,$45,$FE,$31,$6D,$FF,$8D,$5B,$5C,$1A,$FF
	.dc.b	$8C,$1A,$FF,$33,$AE,$95,$CC,$95,$CF,$E6,$BB,$11,$C1,$FE,$04,$F3
	.dc.b	$3F,$C2,$24,$70,$78,$F7,$F3,$11,$C1,$FE,$0C,$F9,$9F,$F0,$C9,$1C
	.dc.b	$13,$1F,$E6,$71,$96,$C6,$5F,$E6,$91,$96,$C7,$C3,$F6,$3E,$31,$1C
	.dc.b	$1F,$E1,$CF,$A3,$FF,$14,$91,$C1,$FA,$0F,$B7,$FE,$93,$5F,$E9,$56
	.dc.b	$F3,$1A,$4F,$27,$F4,$2D,$AE,$51,$FA,$4F,$8E,$FF,$1D,$47,$EB,$1B
	.dc.b	$7F,$E9,$C6,$BF,$E9,$96,$BF,$DA,$7B,$DF,$63,$FE,$58,$8C,$35,$FF
	.dc.b	$30,$36,$37,$EB,$7F,$A2,$1B,$7F,$E6,$19,$D1,$27,$FA,$2F,$CF,$F3
	.dc.b	$EB,$34,$91,$A2,$57,$F1,$D8,$CB,$D7,$D6,$F0,$59,$A4,$8D,$13,$1F
	.dc.b	$F3,$DC,$EE,$31,$1F,$C7,$FE,$57,$FA,$9E,$26,$93,$C3,$FF,$40,$FE
	.dc.b	$60,$F9,$EA,$4C,$3F,$DC,$51,$97,$FD,$C9,$F5,$2F,$95,$FF,$44,$B6
	.dc.b	$33,$18,$3F,$D1,$BF,$4B,$FE,$73,$E4,$FF,$76,$37,$4F,$F1,$3B,$5F
	.dc.b	$E7,$6D,$7F,$BB,$AF,$0D,$F0,$1C,$9B,$3C,$9A,$F2,$6B,$CB,$E4,$BF
	.dc.b	$DF,$8A,$91,$86,$C7,$9A,$37,$F9,$2F,$FB,$D8,$91,$83,$FC,$7E,$7F
	.dc.b	$51,$E3,$7C,$AE,$38,$BF,$06,$D8,$CC,$FE,$06,$3F,$D5,$DC,$4F,$9F
	.dc.b	$A4,$9F,$F5,$84,$65,$B1,$97,$FD,$77,$19,$6C,$78,$7E,$89,$E0,$D2
	.dc.b	$4F,$F6,$D8,$CB,$63,$2F,$F4,$D8,$CB,$63,$CF,$F3,$D2,$3C,$AD,$7C
	.dc.b	$6D,$7F,$C2,$16,$8C,$47,$F9,$77,$81,$D6,$F0,$3E,$5C,$61,$B1,$99
	.dc.b	$FE,$05,$1A,$EE,$73,$52,$4F,$F1,$36,$BF,$E0,$6B,$FE,$25,$18,$8F
	.dc.b	$F2,$1C,$61,$FF,$5C,$F3,$7D,$CF,$4F,$FC,$13,$1F,$E5,$B8,$C1,$FE
	.dc.b	$E8,$F1,$7F,$A7,$7C,$5F,$E0,$9B,$3F,$D3,$63,$1E,$39,$E9,$F2,$9F
	.dc.b	$0E,$7F,$D0,$F8,$23,$11,$FE,$1C,$8C,$6D,$C0,$91,$C2,$76,$FF,$97
	.dc.b	$27,$6E,$67,$6F,$39,$DB,$FE,$8D,$9D,$B9,$9F,$3F,$D0,$48,$FE,$CC
	.dc.b	$D7,$97,$D1,$7A,$88,$F8,$DF,$7F,$CA,$C8,$FE,$DC,$D7,$FB,$77,$6F
	.dc.b	$C4,$23,$FD,$CD,$AE,$8C,$10,$FF,$01,$FD,$13,$E0,$3F,$63,$E7,$23
	.dc.b	$13,$FE,$04,$6B,$C2,$7F,$80,$FD,$8B,$D0,$A4,$9F,$DD,$C6,$5F,$9D
	.dc.b	$D7,$F6,$08,$E0,$87,$CD,$72,$D3,$FC,$77,$F2,$BF,$C0,$FD,$53,$49
	.dc.b	$1A,$21,$FE,$3D,$8D,$B8,$3F,$CB,$FC,$7E,$C6,$E3,$FC,$72,$31,$1F
	.dc.b	$E9,$E6,$BF,$EC,$BF,$B1,$7C,$D7,$F2,$26,$C7,$F1,$A6,$BF,$ED,$A6
	.dc.b	$BF,$E6,$E6,$BF,$EA,$A6,$BD,$63,$A1,$FC,$87,$E9,$1E,$CD,$24,$FF
	.dc.b	$9F,$63,$2D,$8C,$BF,$E5,$D8,$CB,$63,$F5,$3E,$14,$8F,$D1,$35,$FC
	.dc.b	$F3,$5F,$D2,$73,$FF,$30,$72,$7F,$38,$C7,$FB,$73,$7C,$EF,$94,$47
	.dc.b	$E2,$7B,$3F,$1C,$8E,$08,$7C,$77,$2D,$3F,$6D,$E1,$7F,$00,$23,$82
	.dc.b	$1F,$B5,$72,$D3,$F7,$DF,$13,$FC,$20,$8E,$08,$7E,$F5,$CB,$4F,$F5
	.dc.b	$57,$D1,$FB,$9F,$F4,$43,$5F,$F5,$23,$5F,$F6,$E3,$50,$FA,$37,$CA
	.dc.b	$7F,$6A,$6C,$7F,$62,$6B,$FE,$B4,$38,$46,$C8,$ED,$AF,$F1,$5F,$7F
	.dc.b	$F9,$29,$1F,$E3,$CD,$74,$60,$87,$F1,$DF,$43,$F9,$69,$1F,$E4,$CD
	.dc.b	$74,$60,$87,$FB,$2B,$E2,$3F,$A9,$FC,$57,$F7,$C6,$C7,$AD,$AE,$FC
	.dc.b	$A4,$FF,$81,$BA,$64,$FF,$81,$3C,$FF,$EE,$8F,$FB,$59,$AF,$CC,$0E
	.dc.b	$1F,$9E,$D7,$31,$D1,$FF,$4E,$FC,$BB,$E2,$7F,$A8,$36,$3E,$87,$90
	.dc.b	$8C,$47,$09,$DF,$C3,$C8,$DC,$A3,$11,$DB,$E1,$FA,$91,$DB,$E0,$F6
	.dc.b	$64,$72,$6B,$E6,$F5,$FC,$A4,$7B,$B5,$D1,$82,$1F,$D5,$FE,$33,$FB
	.dc.b	$39,$1F,$EC,$0F,$C0,$F7,$E4,$7A,$9A,$FC,$66,$BF,$3F,$3F,$89,$18
	.dc.b	$8F,$CE,$F0,$3E,$7B,$FC,$9D,$1F,$7D,$AF,$E1,$5A,$FD,$98,$E1,$87
	.dc.b	$F8,$07,$EC,$1E,$65,$25,$7D,$48,$CB,$63,$2F,$FB,$62,$32,$D8,$FF
	.dc.b	$03,$FC,$0C,$8F,$F0,$03,$5E,$1D,$7E,$12,$3D,$DA,$E8,$C1,$0F,$F0
	.dc.b	$27,$27,$F8,$50,$8F,$F8,$35,$F6,$6B,$23,$D4,$D7,$D0,$D7,$E7,$60
	.dc.b	$8C,$47,$E7,$FE,$79,$F3,$DF,$E7,$26,$C7,$F9,$A9,$AF,$F0,$96,$BF
	.dc.b	$A7,$0B,$D0,$E4,$48,$E4,$D7,$93,$5E,$51,$C5,$F0,$7C,$9F,$8C,$8E
	.dc.b	$0F,$D5,$FD,$CE,$DF,$17,$12,$31,$1F,$8D,$6B,$FC,$A2,$35,$7F,$E4
	.dc.b	$5F,$9E,$7C,$FD,$25,$7E,$E2,$32,$D8,$CB,$E0,$8C,$B6,$3F,$CA,$CF
	.dc.b	$FA,$78,$FF,$1E,$14,$7A,$DF,$CE,$BD,$4D,$92,$3F,$CA,$26,$E3,$4E
	.dc.b	$23,$FD,$4F,$D4,$F9,$C7,$37,$F4,$4F,$9E,$7C,$EF,$F1,$BA,$FF,$2F
	.dc.b	$61,$98,$C7,$B5,$FC,$7E,$B7,$3F,$D3,$2B,$A5,$EE,$31,$87,$65,$5C
	.dc.b	$1F,$E3,$97,$CF,$0E,$B1,$C4,$FE,$41,$5C,$6B,$86,$1D,$A7,$9D,$70
	.dc.b	$69,$E3,$BE,$8F,$8B,$0E,$95,$C6,$F9,$D6,$AB,$C6,$6B,$0F,$06,$1F
	.dc.b	$0D,$B1,$D4,$E7,$3C,$6F,$A5,$6F,$0F,$F6,$5C,$79,$EB,$F4,$0D,$74
	.dc.b	$7F,$6D,$5F,$16,$F6,$78,$B6,$74,$D7,$9B,$5F,$EE,$DF,$3D,$FE,$E4
	.dc.b	$4D,$23,$FE,$E8,$7F,$C0,$7F,$3F,$F4,$E1,$EF,$BD,$FC,$E7,$D2,$55
	.dc.b	$BA,$D4,$62,$75,$39,$AE,$33,$FB,$FB,$AF,$FB,$1E,$2B,$FE,$14,$8F
	.dc.b	$C0,$3E,$28,$AF,$A8,$FB,$A7,$12,$9C,$4F,$11,$FE,$C2,$7A,$BC,$67
	.dc.b	$C8,$E8,$FE,$F1,$3E,$2B,$F5,$CF,$C3,$9E,$17,$FE,$CF,$BF,$5F,$03
	.dc.b	$83,$FD,$31,$1F,$D4,$0F,$F5,$93,$8A,$FD,$6C,$63,$DC,$F8,$FB,$7C
	.dc.b	$8B,$E5,$EC,$E2,$B8,$B6,$FF,$7F,$87,$3B,$E0,$FF,$A7,$EB,$C1,$3A
	.dc.b	$AE,$17,$FE,$D1,$6F,$A9,$C6,$BF,$D8,$CF,$1A,$F2,$9F,$C3,$1D,$36
	.dc.b	$78,$9C,$B1,$B5,$E0,$E7,$FE,$0F,$C7,$F8,$2B,$FE,$E9,$A7,$F0,$DA
	.dc.b	$FF,$0C,$B6,$3F,$97,$B5,$FF,$3D,$F9,$DE,$63,$E0,$5F,$4A,$F6,$E7
	.dc.b	$FE,$EB,$9C,$DF,$FA,$C2,$7A,$FB,$7E,$1C,$3F,$9B,$4E,$5E,$F1,$95
	.dc.b	$87,$F1,$E8,$C3,$6B,$99,$27,$FD,$DD,$F5,$FF,$28,$F0,$D5,$9F,$85
	.dc.b	$6F,$7F,$9B,$63,$E0,$C6,$08,$E0,$97,$F1,$23,$CD,$93,$6F,$E6,$75
	.dc.b	$F0,$9B,$5F,$63,$2F,$E6,$F7,$F7,$B0,$F9,$3D,$CF,$A6,$F6,$B7,$1F
	.dc.b	$F2,$5B,$70,$FE,$C2,$D7,$FC,$58,$D7,$FB,$03,$5F,$92,$DC,$3F,$BC
	.dc.b	$D7,$A3,$17,$AF,$F9,$03,$C7,$17,$0F,$C4,$5B,$E7,$FC,$A4,$8F,$2B
	.dc.b	$71,$F2,$B5,$F2,$BD,$FE,$87,$F3,$A7,$C0,$6C,$F6,$F3,$BE,$09,$1F
	.dc.b	$82,$DA,$F8,$2D,$7E,$0B,$EB,$F8,$4F,$D4,$9F,$31,$B3,$EA,$3E,$69
	.dc.b	$0F,$9A,$59,$F9,$AD,$7E,$6B,$F1,$FE,$43,$F6,$A7,$D8,$69,$F7,$E3
	.dc.b	$F7,$27,$EE,$BA,$AC,$7E,$C7,$E0,$5E,$DF,$E6,$9B,$1F,$E8,$86,$D7
	.dc.b	$F8,$79,$AF,$FA,$5F,$8D,$F8,$27,$FB,$F2,$BF,$5E,$D9,$FE,$0B,$3A
	.dc.b	$FA,$BF,$E1,$B3,$9E,$DF,$E0,$1F,$3F,$FE,$8F,$C3,$FD,$8D,$82,$30
	.dc.b	$FF,$B2,$5A,$FF,$90,$5A,$FF,$AA,$5A,$FF,$B5,$9A,$F3,$F5,$BF,$74
	.dc.b	$D9,$FF,$6A,$7C,$73,$F4,$AF,$C6,$47,$F2,$48,$C7,$17,$F2,$BA,$E0
	.dc.b	$7F,$20,$8C,$CF,$C0,$FB,$0F,$69,$EE,$FB,$1B,$19,$25,$7C,$4D,$13
	.dc.b	$A8,$C7,$D8,$FC,$63,$E3,$FC,$DD,$BF,$9B,$16,$79,$4D,$DB,$FB,$5E
	.dc.b	$5E,$BF,$A8,$FE,$84,$DC,$3D,$9A,$BE,$2F,$2A,$A9,$27,$73,$71,$F4
	.dc.b	$7C,$97,$DA,$31,$8F,$F5,$3D,$51,$F7,$BE,$6E,$61,$F8,$FF,$16,$E1
	.dc.b	$F0,$66,$E1,$FE,$02,$F0,$5C,$EB,$CE,$0C,$3F,$CB,$39,$7F,$9A,$23
	.dc.b	$F1,$FE,$8F,$E3,$3E,$BB,$F1,$AC,$8C,$9F,$94,$8C,$57,$F5,$81,$F3
	.dc.b	$8E,$DF,$C2,$B6,$BF,$26,$D9,$FF,$48,$1C,$27,$64,$9D,$B6,$37,$2E
	.dc.b	$FF,$C5,$36,$BF,$B8,$CF,$36,$C6,$63,$F0,$DE,$7F,$EA,$0F,$D4,$FC
	.dc.b	$DD,$AF,$D7,$C5,$9F,$C6,$4D,$DF,$F0,$7F,$87,$3E,$9F,$F0,$6D,$C3
	.dc.b	$FC,$03,$2F,$D9,$1F,$55,$AB,$95,$FA,$3B,$E1,$8F,$F8,$03,$0E,$17
	.dc.b	$A8,$F9,$99,$FF,$3A,$BF,$93,$E4,$FC,$C2,$38,$24,$FC,$68,$CB,$F2
	.dc.b	$3C,$BF,$8E,$7C,$88,$B3,$FC,$3B,$CF,$FE,$16,$7F,$80,$39,$7E,$B6
	.dc.b	$5F,$93,$8C,$CE,$CF,$E4,$1C,$38,$C7,$F6,$5E,$BF,$F1,$43,$93,$67
	.dc.b	$F8,$DC,$58,$FE,$D1,$87,$E4,$E7,$27,$F8,$6F,$F0,$0F,$E4,$13,$F9
	.dc.b	$73,$FC,$0E,$E1,$FF,$61,$35,$CF,$CA,$7F,$83,$7B,$DF,$A4,$7F,$B9
	.dc.b	$EF,$F8,$93,$6B,$FB,$C5,$FF,$29,$9F,$D2,$57,$0F,$81,$FE,$5B,$F0
	.dc.b	$12,$78,$C5,$BF,$E2,$13,$76,$FE,$0F,$8F,$57,$F7,$51,$FE,$49,$F6
	.dc.b	$DF,$8A,$3C,$CF,$F1,$14,$61,$B1,$9C,$BF,$4B,$39,$7F,$4F,$E6,$FF
	.dc.b	$46,$F5,$BF,$A0,$9D,$6B,$86,$7F,$C8,$E3,$FC,$73,$B7,$F0,$AB,$FD
	.dc.b	$7F,$84,$8C,$57,$FB,$C1,$AF,$CF,$8D,$78,$9F,$E2,$2F,$09,$F0,$1F
	.dc.b	$D4,$0E,$B1,$70,$FE,$AF,$16,$FE,$6C,$DD,$FF,$8E,$4D,$AF,$B3,$36
	.dc.b	$BE,$EF,$7F,$F9,$0B,$23,$49,$5F,$0B,$17,$C1,$7F,$62,$3D,$F3,$67
	.dc.b	$FC,$E7,$E9,$3F,$68,$FE,$45,$1F,$C4,$23,$1E,$3F,$F9,$4D,$B1,$99
	.dc.b	$DE,$5F,$91,$3F,$D6,$6E,$1F,$CA,$4F,$07,$4D,$73,$1B,$7F,$98,$70
	.dc.b	$FF,$7E,$F6,$7F,$BC,$9F,$E8,$DA,$E5,$BF,$F4,$18,$FF,$48,$78,$0F
	.dc.b	$86,$FF,$03,$E1,$FC,$E6,$31,$EA,$FF,$9A,$DB,$9F,$F2,$6A,$DD,$7F
	.dc.b	$34,$6C,$EA,$3D,$8D,$7F,$5C,$D7,$CF,$BB,$E0,$BF,$C1,$E7,$92,$BF
	.dc.b	$C3,$79,$7F,$57,$8C,$E1,$FE,$18,$75,$8E,$30,$FF,$4E,$E3,$EB,$C3
	.dc.b	$D0,$7B,$6F,$D0,$E0,$8E,$09,$3E,$88,$E0,$FD,$5D,$3F,$C5,$3E,$FF
	.dc.b	$F8,$90,$F3,$4D,$AF,$C1,$FC,$83,$E2,$BF,$00,$D8,$FC,$23,$70,$FB
	.dc.b	$6D,$7E,$FA,$FA,$5E,$24,$9E,$2F,$F8,$93,$E8,$F5,$C3,$FE,$32,$6D
	.dc.b	$FF,$60,$6B,$F8,$D6,$BF,$C7,$DF,$36,$5C,$71,$FF,$6F,$4F,$16,$BC
	.dc.b	$5A,$FF,$7D,$A6,$2F,$A0,$FF,$21,$C6,$1B,$19,$9F,$83,$18,$F8,$F9
	.dc.b	$1F,$11,$AF,$CF,$EB,$7D,$67,$F3,$5E,$3F,$F7,$54,$62,$7F,$D8,$ED
	.dc.b	$7F,$03,$F4,$4F,$93,$E1,$C1,$F9,$B8,$C4,$FF,$7E,$6B,$F9,$66,$BF
	.dc.b	$96,$6B,$C6,$7F,$C1,$B1,$88,$FF,$09,$B7,$8C,$E7,$8F,$FD,$A3,$F3
	.dc.b	$5E,$2A,$46,$1F,$4C,$66,$B8,$FD,$6E,$EB,$55,$FE,$74,$C9,$79,$6C
	.dc.b	$7F,$8B,$F3,$67,$EF,$C3,$A6,$47,$D2,$4A,$F8,$4D,$7E,$13,$5F,$84
	.dc.b	$FF,$9F,$B1,$FF,$75,$FC,$37,$84,$93,$FC,$89,$FE,$56,$DE,$DF,$C1
	.dc.b	$C6,$23,$F8,$CB,$5F,$E3,$AD,$7F,$18,$D5,$C5,$FE,$DD,$9E,$64,$8F
	.dc.b	$EA,$F9,$BE,$CF,$E2,$12,$3F,$BF,$79,$FF,$46,$7F,$AA,$3F,$E9,$86
	.dc.b	$C7,$F5,$8F,$05,$39,$6C,$76,$FB,$2F,$A0,$47,$FC,$15,$E2,$57,$EB
	.dc.b	$45,$E0,$FF,$33,$57,$F5,$C9,$FC,$BC,$E7,$83,$F9,$BE,$28,$F3,$56
	.dc.b	$EB,$FC,$13,$87,$8C,$8F,$9D,$FA,$94,$FA,$7F,$C2,$12,$7F,$09,$87
	.dc.b	$F9,$CF,$C8,$7E,$4C,$F9,$7F,$CA,$46,$1F,$F5,$DB,$6B,$FC,$0B,$3B
	.dc.b	$6C,$F8,$3D,$FF,$C9,$12,$3F,$D6,$5F,$4D,$FB,$3D,$BF,$6F,$39,$9F
	.dc.b	$CD,$3F,$A3,$F4,$55,$BC,$BF,$C4,$AD,$A4,$EB,$AF,$84,$E6,$7E,$BF
	.dc.b	$C4,$7E,$B5,$A7,$FB,$53,$89,$1F,$ED,$68,$C9,$FE,$84,$8C,$BF,$8A
	.dc.b	$F8,$2F,$D6,$11,$F9,$23,$81,$F3,$5F,$B0,$8F,$DA,$4E,$DA,$70,$9D
	.dc.b	$F6,$7E,$DF,$E2,$64,$FF,$18,$7F,$29,$7F,$99,$8C,$10,$FE,$01,$1A
	.dc.b	$E7,$F7,$7C,$6B,$CC,$A4,$AF,$F4,$AC,$7F,$86,$1B,$E7,$FC,$18,$C4
	.dc.b	$7F,$A9,$1A,$FF,$84,$5A,$FF,$98,$DA,$FF,$AE,$DA,$9A,$6B,$81,$F6
	.dc.b	$7F,$72,$7F,$DE,$8D,$7F,$DE,$4D,$E4,$FE,$2F,$05,$F5,$0F,$86,$FC
	.dc.b	$35,$6F,$97,$FC,$D2,$DB,$FC,$2B,$6F,$8D,$7F,$25,$F8,$29,$5F,$D1
	.dc.b	$9F,$E5,$1F,$03,$88,$FE,$59,$16,$BF,$A6,$CF,$CE,$F5,$A5,$7C,$C7
	.dc.b	$E8,$FB,$32,$3F,$D9,$A7,$C3,$7E,$71,$1C,$EA,$8C,$47,$D1,$6B,$FE
	.dc.b	$91,$F1,$7F,$1D,$E4,$FF,$3D,$24,$7F,$5E,$F1,$EB,$FD,$0E,$2E,$5F
	.dc.b	$D9,$27,$FA,$27,$54,$AF,$EC,$AF,$E3,$B3,$C4,$7F,$4B,$8B,$5F,$EF
	.dc.b	$BE,$6F,$F6,$5C,$FF,$6C,$F8,$EF,$E9,$C4,$7F,$47,$C7,$FC,$5F,$DC
	.dc.b	$FD,$D7,$8C,$69,$27,$70,$F9,$FC,$47,$DF,$E5,$FF,$51,$CF,$18,$DD
	.dc.b	$74,$AF,$EB,$DE,$84,$9F,$EF,$EF,$F6,$1C,$B1,$1F,$D9,$62,$E7,$FE
	.dc.b	$06,$9F,$EC,$DE,$7C,$AF,$F0,$3B,$FD,$A7,$D3,$C4,$7F,$6D,$8B,$5F
	.dc.b	$E0,$E9,$F6,$FD,$04,$AF,$5B,$FC,$3F,$E8,$3F,$92,$47,$AC,$07,$2F
	.dc.b	$F9,$F5,$AF,$FB,$3D,$AF,$06,$BC,$7C,$7F,$E3,$11,$82,$3F,$D0,$36
	.dc.b	$FF,$01,$12,$7F,$C4,$7E,$97,$F0,$CC,$3F,$AF,$67,$FE,$4B,$9F,$F5
	.dc.b	$AB,$5F,$F4,$87,$87,$FC,$72,$7F,$BF,$7A,$9E,$CD,$3F,$6B,$D0,$8C
	.dc.b	$65,$FE,$8E,$6E,$1F,$E1,$C7,$F6,$93,$BF,$A0,$FF,$0D,$41,$A7,$F7
	.dc.b	$3F,$C5,$31,$5F,$C9,$DA,$F9,$1A,$FF,$B1,$DA,$FC,$EC,$91,$88,$FE
	.dc.b	$6C,$D8,$CC,$FF,$7F,$8C,$1B,$BF,$67,$D2,$F9,$89,$7E,$3E,$4F,$31
	.dc.b	$C4,$8F,$FB,$7A,$BF,$30,$F9,$E3,$2D,$34,$FC,$CF,$67,$E1,$92,$3E
	.dc.b	$07,$CB,$FC,$33,$CC,$47,$B3,$5F,$88,$D7,$E2,$3F,$63,$DB,$F9,$E4
	.dc.b	$8F,$99,$EB,$7C,$F3,$E1,$11,$F5,$35,$FA,$0D,$7E,$83,$F8,$4E,$5F
	.dc.b	$B6,$48,$FE,$BD,$F3,$AF,$94,$FF,$17,$36,$3F,$DB,$87,$F6,$97,$0F
	.dc.b	$E4,$1A,$FE,$A0,$D4,$7F,$8D,$63,$31,$F7,$DA,$E8,$EB,$BA,$33,$B7
	.dc.b	$E5,$FA,$BF,$22,$4A,$F3,$FC,$8F,$98,$EC,$47,$9D,$23,$11,$E7,$6B
	.dc.b	$E7,$BF,$04,$7E,$A3,$6F,$CF,$92,$7E,$2F,$8F,$F1,$0F,$61,$1E,$EF
	.dc.b	$F4,$AE,$EF,$F1,$1F,$87,$FE,$1E,$24,$FF,$8B,$3C,$3A,$FF,$0F,$9D
	.dc.b	$BF,$E3,$09,$FF,$12,$CF,$F8,$B7,$F0,$9F,$F0,$B7,$4F,$F6,$EB,$46
	.dc.b	$3A,$7F,$8B,$C6,$3C,$FE,$1E,$9F,$16,$65,$FD,$02,$F8,$3F,$E3,$47
	.dc.b	$FC,$65,$F3,$DF,$E6,$8C,$7F,$90,$FC,$EB,$E3,$FE,$E4,$BE,$37,$EF
	.dc.b	$4E,$70,$FE,$4F,$39,$78,$D7,$0F,$5B,$FC,$09,$0A,$E0,$AB,$C0,$E6
	.dc.b	$BB,$DF,$66,$C7,$43,$BD,$F4,$BE,$77,$CB,$E1,$2E,$1F,$93,$6D,$FF
	.dc.b	$3C,$77,$7E,$46,$E3,$D6,$7C,$95,$E1,$7D,$AD,$7D,$15,$E1,$C7,$DD
	.dc.b	$5C,$70,$F8,$55,$EB,$38,$99,$AE,$0F,$AA,$7F,$C2,$33,$F0,$A3,$F5
	.dc.b	$47,$68,$CC,$FC,$A8,$CC,$BE,$49,$EE,$72,$6C,$FE,$E2,$75,$39,$3F
	.dc.b	$CD,$50,$F9,$EA,$47,$FC,$DC,$DD,$3E,$BC,$7F,$AE,$B8,$7E,$33,$FD
	.dc.b	$9A,$6F,$6F,$F9,$0A,$B7,$3D,$4F,$84,$7F,$6C,$8D,$69,$FC,$32,$B8
	.dc.b	$DF,$4A,$E1,$B7,$82,$FF,$C9,$AF,$4A,$FB,$57,$F1,$A7,$2F,$3A,$EE
	.dc.b	$79,$CA,$78,$F0,$E9,$87,$7B,$F0,$CE,$9B,$3F,$9D,$EB,$FE,$D1,$F3
	.dc.b	$DF,$E5,$ED,$FF,$B5,$F8,$A8,$C6,$3F,$DA,$FE,$7A,$E5,$FD,$AE,$6E
	.dc.b	$1C,$AB,$FD,$26,$D9,$FF,$08,$3F,$DD,$6B,$77,$D6,$B7,$F3,$D6,$F6
	.dc.b	$DC,$7F,$BE,$AB,$EE,$E6,$BF,$C3,$78,$23,$16,$8C,$63,$C2,$F8,$BE
	.dc.b	$AE,$DF,$E0,$5C,$7F,$E7,$EB,$FE,$51,$B7,$FB,$BB,$2F,$E7,$2D,$8F
	.dc.b	$F0,$9B,$5F,$F3,$26,$4F,$8C,$F2,$9D,$67,$86,$9E,$A8,$C7,$AD,$DE
	.dc.b	$FF,$C0,$D4,$8C,$47,$FA,$EE,$31,$87,$8A,$73,$97,$F1,$96,$CE,$AB
	.dc.b	$CB,$18,$39,$EB,$F1,$7C,$5F,$92,$78,$B0,$FA,$2F,$9F,$DB,$E6,$48
	.dc.b	$F8,$2D,$BF,$82,$D7,$E3,$BD,$B4,$FE,$59,$3F,$73,$1E,$8D,$3F,$DA
	.dc.b	$1E,$BB,$E1,$BD,$6D,$8F,$EC,$ED,$C3,$FC,$82,$D7,$FA,$E3,$5F,$F1
	.dc.b	$73,$5F,$F7,$13,$74,$FF,$61,$57,$97,$E5,$3F,$12,$7E,$06,$7B,$6F
	.dc.b	$F8,$E8,$C7,$BF,$E4,$26,$3E,$46,$E3,$E4,$6B,$E4,$7E,$17,$CF,$7E
	.dc.b	$58,$E7,$3F,$9E,$7C,$1E,$C7,$C0,$24,$7C,$06,$D7,$C0,$6B,$F0,$1F
	.dc.b	$9D,$F3,$1F,$A2,$39,$CF,$EA,$9F,$63,$F3,$08,$7C,$C2,$CF,$CC,$6B
	.dc.b	$F3,$1F,$B5,$F2,$DF,$AE,$39,$BF,$B6,$7F,$63,$12,$3F,$73,$D7,$63
	.dc.b	$FD,$1D,$B3,$E5,$9E,$7F,$E6,$1B,$7F,$E2,$46,$D7,$F9,$F9,$AF,$FB
	.dc.b	$3F,$E4,$1F,$2D,$FC,$39,$B5,$FC,$39,$7C,$AE,$67,$F8,$AD,$3E,$79
	.dc.b	$F0,$1F,$AC,$8C,$36,$33,$87,$FB,$99,$B3,$FE,$44,$6B,$FD,$61,$AF
	.dc.b	$F9,$0D,$AF,$FA,$B1,$BE,$5B,$E6,$DC,$7F,$CA,$23,$0F,$9A,$7B,$F2
	.dc.b	$FF,$88,$6B,$78,$7B,$7E,$7B,$E0,$1F,$07,$D5,$B1,$92,$65,$FC,$BB
	.dc.b	$6F,$F2,$F4,$63,$0F,$F5,$2C,$3E,$7B,$E5,$1F,$2F,$E7,$AD,$7C,$B8
	.dc.b	$B7,$CE,$6D,$F9,$7E,$33,$D5,$99,$BF,$1D,$86,$63,$FE,$F2,$9D,$36
	.dc.b	$73,$87,$9F,$E6,$7E,$E1,$8C,$3C,$EF,$DF,$F6,$25,$FF,$B6,$24,$FF
	.dc.b	$AB,$79,$7F,$98,$3F,$83,$9D,$DF,$C3,$7F,$9F,$4E,$A3,$13,$F8,$9F
	.dc.b	$5F,$F1,$87,$E3,$7E,$7A,$C6,$49,$3F,$36,$32,$FF,$BA,$C7,$CB,$BB
	.dc.b	$3F,$1C,$DA,$FF,$46,$CE,$A7,$24,$8C,$CE,$E7,$FD,$26,$D7,$FC,$4E
	.dc.b	$D7,$FD,$24,$BB,$7F,$3E,$D9,$F8,$C6,$AB,$33,$FA,$9F,$4E,$4F,$EA
	.dc.b	$7E,$5A,$E1,$F8,$99,$BA,$FE,$22,$6D,$7F,$9C,$BE,$75,$F1,$DF,$9B
	.dc.b	$6C,$78,$30,$5B,$E7,$9F,$1D,$BC,$25,$EF,$B5,$8F,$2C,$78,$CF,$DF
	.dc.b	$C3,$8B,$B9,$D6,$3F,$DD,$F6,$78,$DF,$40,$95,$F3,$1B,$87,$CC,$6B
	.dc.b	$F3,$1F,$AF,$E4,$FC,$A3,$E5,$B6,$7F,$87,$7A,$1F,$C4,$4F,$C5,$F0
	.dc.b	$D8,$C9,$2F,$F9,$0F,$27,$DB,$CB,$F6,$37,$FC,$67,$D9,$97,$F1,$FC
	.dc.b	$6F,$E4,$D8,$60,$FD,$D4,$DB,$FC,$FC,$27,$75,$FC,$86,$BF,$C3,$5F
	.dc.b	$31,$FC,$92,$BF,$CD,$B4,$8C,$47,$FB,$0D,$A8,$E8,$E1,$70,$3E,$79
	.dc.b	$F0,$3F,$C6,$5B,$8F,$A6,$FC,$19,$7F,$AE,$67,$38,$7F,$13,$F9,$1F
	.dc.b	$E7,$07,$1C,$2E,$1F,$CE,$62,$CF,$FB,$1E,$B8,$DF,$F1,$89,$FF,$27
	.dc.b	$FB,$8F,$9E,$FE,$7F,$87,$FB,$E2,$31,$E5,$FF,$8A,$1B,$19,$9D,$E3
	.dc.b	$FE,$90,$F5,$5E,$37,$F8,$9F,$2F,$E4,$BE,$B7,$6B,$DC,$63,$0F,$77
	.dc.b	$85,$FE,$30,$DC,$D2,$47,$FA,$73,$2F,$F2,$86,$2F,$9A,$FE,$B0,$7D
	.dc.b	$6A,$B5,$FD,$66,$2D,$7F,$68,$BE,$B3,$92,$47,$F3,$D9,$E5,$18,$C3
	.dc.b	$F0,$BC,$5F,$7F,$BB,$CC,$49,$F2,$46,$5F,$36,$BF,$0B,$E5,$BF,$B6
	.dc.b	$1F,$8A,$AB,$3F,$BB,$F4,$5F,$3D,$FC,$A7,$3F,$C4,$08,$CF,$B5,$FE
	.dc.b	$58,$6D,$7C,$36,$CF,$FA,$6E,$B9,$3F,$EA,$76,$BC,$DB,$1C,$23,$FC
	.dc.b	$05,$EC,$FF,$80,$4F,$F4,$8C,$74,$D3,$FA,$EC,$7F,$A5,$3B,$DF,$15
	.dc.b	$FE,$0A,$BF,$CF,$46,$3D,$5F,$F3,$6B,$72,$FE,$5F,$5B,$AF,$E7,$6D
	.dc.b	$9D,$47,$F3,$28,$C1,$BD,$BF,$17,$F0,$3F,$E1,$53,$C5,$1F,$E1,$CC
	.dc.b	$BF,$B8,$C6,$70,$FF,$0C,$3A,$C7,$18,$7F,$20,$3C,$37,$E6,$3D,$75
	.dc.b	$E6,$70,$47,$04,$8F,$34,$7F,$A0,$39,$1E,$FF,$F8,$A4,$F2,$C5,$AF
	.dc.b	$BD,$F3,$AF,$95,$FB,$4D,$8F,$B4,$DC,$3F,$DA,$8D,$7F,$07,$BA,$1F
	.dc.b	$21,$C8,$95,$C9,$FF,$19,$7C,$F6,$AB,$FB,$2B,$6F,$F5,$4D,$7F,$1E
	.dc.b	$1C,$3F,$92,$7D,$19,$72,$C7,$DF,$3C,$9A,$F2,$9D,$BE,$AF,$96,$FE
	.dc.b	$EB,$7E,$96,$E4,$69,$8F,$94,$95,$F1,$1A,$FF,$0D,$E8,$7C,$5F,$E9
	.dc.b	$B8,$BF,$9C,$F1,$FF,$39,$8C,$6F,$FE,$D8,$AD,$DF,$F6,$9E,$47,$CF
	.dc.b	$3E,$3B,$73,$FE,$0E,$8C,$3F,$E0,$66,$BF,$E0,$66,$BF,$E0,$66,$BF
	.dc.b	$A2,$6B,$CE,$3F,$C2,$31,$89,$F0,$C7,$F8,$63,$1F,$EA,$1F,$1B,$E2
	.dc.b	$6D,$FF,$BF,$B0,$4F,$9A,$FD,$3F,$1D,$B9,$CD,$6E,$B0,$6F,$15,$53
	.dc.b	$C6,$49,$F7,$35,$F7,$35,$F7,$3E,$7A,$FF,$66,$47,$97,$E1,$BC,$47
	.dc.b	$CA,$E1,$FF,$68,$76,$D8,$3F,$82,$6C,$FF,$19,$6B,$F9,$29,$E2,$E6
	.dc.b	$78,$11,$C1,$FC,$BA,$9F,$37,$F0,$C9,$1F,$DA,$7C,$CF,$BB,$3F,$DD
	.dc.b	$DF,$F2,$3B,$63,$F9,$CF,$84,$9C,$B6,$3B,$7C,$CF,$CF,$23,$FD,$D7
	.dc.b	$C3,$AF,$F0,$DE,$0F,$F0,$54,$E3,$93,$FA,$4D,$FF,$57,$AF,$B9,$39
	.dc.b	$E3,$FC,$B3,$E6,$9D,$4F,$F5,$7C,$3C,$64,$7C,$EF,$D3,$F9,$EF,$C1
	.dc.b	$12,$3F,$C8,$DE,$43,$E7,$9F,$35,$F9,$E8,$C3,$FE,$68,$6C,$FF,$7F
	.dc.b	$9D,$B6,$7F,$01,$EF,$7E,$3C,$91,$FE,$4D,$F9,$EF,$D6,$BF,$07,$81
	.dc.b	$3A,$8C,$3F,$AE,$F4,$55,$C3,$7F,$F1,$8C,$63,$AF,$53,$9E,$0F,$F1
	.dc.b	$54,$E7,$0F,$F3,$11,$F5,$FE,$23,$F7,$2D,$3F,$D0,$7C,$7F,$ED,$46
	.dc.b	$C7,$FA,$19,$D1,$FC,$2F,$E0,$BF,$54,$47,$FC,$91,$F3,$7F,$AC,$7D
	.dc.b	$91,$96,$9A,$EC,$FD,$97,$C4,$C8,$FD,$C3,$F9,$0A,$FC,$AC,$60,$87
	.dc.b	$C7,$8D,$74,$7A,$BE,$35,$E6,$52,$57,$C0,$87,$9F,$B3,$D8,$E2,$3F
	.dc.b	$D4,$6D,$7F,$C5,$0D,$7F,$9F,$B5,$FF,$07,$B5,$FF,$2C,$35,$39,$D5
	.dc.b	$F3,$BF,$6A,$7F,$D1,$CD,$7F,$B7,$46,$B8,$3F,$96,$F1,$3E,$79,$F0
	.dc.b	$BF,$83,$C3,$84,$FE,$09,$BC,$7F,$80,$6E,$3F,$EA,$0B,$E2,$FF,$25
	.dc.b	$F8,$1F,$E4,$44,$7F,$94,$7C,$15,$7F,$29,$17,$0F,$E8,$FD,$1F,$EA
	.dc.b	$89,$F9,$9E,$BF,$D1,$23,$F3,$FD,$9F,$A8,$7C,$E2,$3F,$39,$AF,$CE
	.dc.b	$6B,$F3,$9F,$E5,$1E,$4F,$F3,$72,$47,$F3,$DF,$1E,$BF,$A9,$8B,$87
	.dc.b	$9A,$7F,$9F,$74,$CA,$FE,$88,$FE,$37,$3C,$47,$EC,$62,$D7,$F9,$63
	.dc.b	$9F,$DD,$3F,$B5,$F8,$EF,$E9,$04,$7F,$D8,$DE,$27,$FA,$49,$BE,$E7
	.dc.b	$ED,$E3,$11,$F7,$1B,$7F,$B8,$5F,$33,$BA,$95,$C8,$EB,$1F,$D5,$FD
	.dc.b	$19,$1F,$D7,$5F,$EB,$58,$E2,$3F,$85,$45,$C3,$EB,$CF,$F5,$FF,$3E
	.dc.b	$57,$F6,$57,$FB,$17,$A7,$88,$FE,$25,$16,$BE,$FC,$FA,$FE,$6A,$57
	.dc.b	$C0,$7F,$84,$FC,$F7,$F2,$08,$FF,$23,$07,$0F,$F8,$99,$AF,$F9,$D9
	.dc.b	$AF,$FA,$FD,$AF,$1F,$1B,$F8,$84,$60,$8F,$F6,$BE,$9F,$EF,$44,$9F
	.dc.b	$F0,$07,$A5,$FC,$27,$0F,$EA,$F9,$FF,$90,$27,$FD,$1E,$D7,$FD,$A6
	.dc.b	$D7,$FD,$13,$E6,$FF,$95,$30,$FF,$06,$7A,$BE,$DD,$3F,$C4,$ED,$CB
	.dc.b	$FD,$AD,$5F,$E5,$AE,$ED,$FC,$F4,$C7,$FD,$B5,$87,$08,$FE,$D7,$F3
	.dc.b	$A6,$9F,$DB,$1C,$4F,$FA,$91,$AF,$FB,$A1,$AF,$FA,$7B,$B1,$18,$8F
	.dc.b	$F7,$44,$62,$7A,$12,$3F,$D5,$B9,$3E,$7B,$DA,$48,$FF,$80,$17,$F9
	.dc.b	$6C,$F9,$BF,$9A,$8F,$03,$5C,$F2,$FA,$48,$FA,$5A,$FA,$5F,$F1,$C5
	.dc.b	$FF,$A8,$A3,$E2,$D7,$CA,$8C,$47,$F8,$0D,$B9,$7F,$B2,$AD,$F2,$FF
	.dc.b	$28,$F9,$04,$AF,$90,$D7,$E4,$35,$F4,$BE,$2F,$73,$EE,$92,$3E,$0F
	.dc.b	$AB,$F5,$8F,$A8,$47,$EA,$35,$FA,$8D,$7E,$2B,$EE,$E5,$FC,$49,$23
	.dc.b	$DE,$36,$7C,$F3,$E4,$FF,$C5,$ED,$9F,$F7,$2B,$5F,$F4,$7C,$F0,$9F
	.dc.b	$F1,$9C,$E6,$7F,$08,$D7,$F0,$6D,$7C,$0D,$9E,$51,$F8,$4E,$AF,$CF
	.dc.b	$91,$F3,$C7,$9B,$E4,$3C,$A7,$80,$91,$E6,$6B,$E6,$A4,$62,$3C,$AF
	.dc.b	$E0,$B8,$3F,$5A,$48,$EF,$3F,$D5,$9F,$F6,$85,$7C,$6F,$1F,$DE,$48
	.dc.b	$F8,$AF,$FA,$C9,$FF,$0F,$F2,$7F,$A4,$38,$3F,$C4,$AD,$E2,$CF,$85
	.dc.b	$FE,$26,$F0,$BF,$C4,$EC,$BF,$F1,$A3,$97,$EB,$F9,$0F,$9E,$FF,$0E
	.dc.b	$63,$FE,$30,$6E,$1F,$E2,$E6,$D2,$B8,$7A,$5F,$E3,$99,$CC,$73,$F4
	.dc.b	$30,$66,$7F,$C6,$7F,$3D,$FE,$6B,$C5,$83,$E7,$5F,$1D,$88,$53,$E2
	.dc.b	$F1,$5A,$9E,$33,$97,$9C,$65,$E3,$EB,$30,$CE,$1D,$6F,$C5,$5E,$18
	.dc.b	$C6,$1D,$9B,$1C,$0E,$F7,$CD,$E0,$62,$FD,$7F,$0B,$E1,$C3,$C9,$28
	.dc.b	$C6,$1E,$D7,$57,$CD,$B7,$DE,$BE,$0D,$78,$DF,$7B,$5F,$6C,$6E,$FD
	.dc.b	$78,$F6,$9C,$56,$E7,$38,$FC,$1C,$7D,$55,$FE,$FE,$9F,$86,$D8,$CB
	.dc.b	$BC,$3F,$24,$FD,$4F,$91,$7B,$78,$B4,$62,$7F,$31,$18,$9D,$4E,$67
	.dc.b	$9F,$CF,$7F,$9C,$53,$48,$FF,$9D,$5F,$88,$BB,$5C,$58,$8E,$27,$D4
	.dc.b	$3F,$1B,$1A,$70,$7E,$42,$79,$C6,$75,$FC,$44,$E4,$DC,$67,$4F,$D9
	.dc.b	$57,$DE,$7F,$DA,$97,$6B,$F3,$07,$8B,$6E,$77,$E1,$7A,$C7,$9B,$0C
	.dc.b	$D7,$A7,$0F,$1B,$67,$C2,$D9,$DC,$7E,$43,$C4,$78,$4F,$A1,$F7,$6F
	.dc.b	$FD,$E9,$B8,$7F,$51,$FA,$2B,$83,$0E,$35,$86,$D7,$F3,$A6,$CF,$F5
	.dc.b	$7A,$D6,$1D,$EB,$4B,$E7,$F2,$6B,$E8,$2C,$EA,$2D,$FF,$61,$9B,$97
	.dc.b	$F8,$43,$E3,$7E,$B4,$FF,$85,$B0,$46,$23,$FB,$53,$5F,$54,$6B,$0F
	.dc.b	$F0,$0F,$8D,$E6,$3C,$F5,$63,$CF,$C3,$C7,$4F,$92,$DB,$FE,$23,$3F
	.dc.b	$E0,$A8,$C1,$1C,$13,$0F,$BF,$5F,$4B,$E5,$3E,$29,$F1,$AA,$D7,$C6
	.dc.b	$F9,$FE,$C4,$7F,$C2,$ED,$BE,$78,$23,$11,$F0,$DF,$37,$CC,$7D,$23
	.dc.b	$E9,$B6,$BE,$9F,$CB,$7A,$C8,$FF,$89,$1B,$5D,$9A,$FC,$F7,$E1,$72
	.dc.b	$FD,$E3,$EF,$B6,$7E,$FC,$FC,$72,$3F,$E3,$06,$9E,$A2,$9F,$6C,$FA
	.dc.b	$1F,$18,$F9,$4F,$C8,$36,$7A,$9F,$91,$71,$3F,$CE,$63,$FC,$A7,$0F
	.dc.b	$9F,$E6,$48,$FE,$CD,$B2,$75,$18,$6D,$7D,$86,$BF,$E6,$46,$BD,$AB
	.dc.b	$CF,$E5,$F9,$8F,$3B,$6B,$CD,$F3,$DE,$C2,$3D,$DB,$5D,$DA,$F7,$7E
	.dc.b	$1F,$CE,$FC,$23,$E1,$B6,$7E,$17,$CB,$7C,$63,$E3,$FC,$F5,$8F,$8D
	.dc.b	$E7,$7C,$C2,$3E,$F7,$E7,$F9,$7F,$38,$F9,$FF,$3B,$67,$FC,$03,$FA
	.dc.b	$4B,$DE,$FD,$67,$94,$8C,$4F,$F8,$46,$31,$3A,$9C,$EB,$FC,$82,$73
	.dc.b	$5B,$9C,$CF,$F7,$B9,$CB,$B8,$CB,$F7,$3E,$4B,$F0,$67,$F1,$69,$DF
	.dc.b	$6F,$F3,$9E,$3F,$B9,$3F,$88,$F2,$26,$3F,$87,$F1,$AE,$3F,$87,$3F
	.dc.b	$8F,$72,$AB,$38,$F3,$8F,$EA,$F7,$F8,$F9,$FC,$9F,$C1,$C8,$FC,$97
	.dc.b	$95,$70,$E9,$9F,$E3,$6B,$18,$F3,$AB,$7F,$E1,$CD,$3E,$57,$C9,$FA
	.dc.b	$F6,$37,$E3,$30,$CA,$FA,$6E,$5A,$7E,$8F,$E1,$BF,$4E,$6F,$4F,$1E
	.dc.b	$C8,$FD,$5F,$C5,$CA,$FD,$57,$C9,$5B,$DC,$DD,$7E,$14,$DA,$FD,$6F
	.dc.b	$9B,$FB,$43,$F6,$BF,$1B,$63,$F6,$91,$70,$F0,$CD,$BF,$ED,$DB,$3F
	.dc.b	$63,$FF,$10,$61,$F8,$96,$DF,$E1,$DA,$FE,$15,$AF,$C9,$6B,$FE,$4F
	.dc.b	$6C,$F5,$C3,$F2,$F1,$AE,$2F,$E0,$1E,$2F,$F0,$C3,$F1,$51,$E0,$C7
	.dc.b	$F0,$11,$FD,$7D,$7A,$74,$C7,$A7,$73,$FE,$C5,$3A,$6C,$E9,$A6,$B8
	.dc.b	$7A,$1E,$0E,$5B,$19,$26,$BF,$95,$D7,$E3,$4E,$A3,$19,$BA,$FD,$A4
	.dc.b	$BF,$94,$D9,$F9,$4D,$7E,$53,$FD,$77,$27,$27,$BC,$F8,$71,$6B,$DB
	.dc.b	$D5,$F0,$CF,$D8,$7B,$ED,$B1,$3A,$26,$3E,$99,$D4,$63,$1F,$DB,$0F
	.dc.b	$93,$F9,$C7,$EE,$FF,$59,$7C,$16,$2B,$E6,$CD,$C3,$E2,$56,$A7,$27
	.dc.b	$D2,$F9,$3F,$61,$1F,$E7,$F8,$23,$FC,$09,$EE,$BF,$1D,$FB,$18,$7F
	.dc.b	$AB,$9B,$5E,$6A,$FF,$11,$D6,$F9,$3F,$09,$F2,$93,$1F,$F6,$2E,$1C
	.dc.b	$1F,$F4,$BF,$3F,$FA,$2E,$75,$97,$F0,$4C,$91,$D2,$F8,$4F,$F8,$BE
	.dc.b	$73,$7F,$8C,$F9,$4F,$AC,$47,$D6,$FD,$9F,$90,$FC,$B1,$F9,$76,$C7
	.dc.b	$FB,$63,$E5,$BF,$AB,$DF,$FB,$BD,$BA,$1A,$6B,$A1,$F3,$CF,$85,$FE
	.dc.b	$5C,$DC,$3A,$79,$D8,$99,$1F,$EC,$69,$D6,$3F,$ED,$8C,$3F,$B5,$E9
	.dc.b	$FC,$02,$72,$DB,$E3,$5B,$C2,$CF,$F9,$CF,$DF,$7E,$D9,$F8,$7C,$BF
	.dc.b	$DB,$D1,$8E,$5F,$F1,$F3,$63,$33,$BC,$3F,$C6,$36,$F9,$4F,$DE,$1F
	.dc.b	$E5,$EB,$E1,$3F,$56,$3F,$80,$7C,$2F,$E9,$49,$87,$E2,$9B,$3D,$1A
	.dc.b	$FF,$B4,$F9,$3F,$D9,$51,$FA,$DF,$85,$FE,$14,$74,$AF,$E0,$EF,$F5
	.dc.b	$9E,$CF,$F3,$24,$7F,$69,$6D,$FF,$67,$6B,$FD,$95,$AF,$F4,$96,$BF
	.dc.b	$EE,$F6,$F0,$FF,$77,$F7,$FF,$BB,$C6,$B1,$FF,$6F,$7C,$A7,$F1,$D3
	.dc.b	$C5,$37,$0F,$E3,$71,$70,$FE,$25,$7E,$4D,$7F,$73,$3E,$9F,$80,$EA
	.dc.b	$4B,$F5,$36,$FD,$4D,$7D,$4F,$F1,$1F,$94,$FE,$5C,$73,$9F,$E5,$6F
	.dc.b	$F3,$1F,$8F,$FE,$2C,$48,$FD,$DB,$FC,$67,$E5,$24,$7E,$F9,$B7,$FE
	.dc.b	$AE,$1F,$36,$F9,$2F,$F6,$24,$66,$31,$EF,$FF,$B6,$1B,$19,$9D,$BF
	.dc.b	$EB,$88,$57,$FA,$B2,$39,$35,$D3,$51,$23,$8F,$97,$FC,$52,$BF,$DC
	.dc.b	$FC,$0D,$BC,$D7,$92,$1F,$9A,$F9,$5D,$C1,$A4,$8F,$E3,$FF,$29,$E0
	.dc.b	$7F,$88,$36,$7F,$98,$B7,$4F,$F0,$97,$ED,$BF,$C5,$2B,$CF,$5F,$DD
	.dc.b	$A3,$11,$FE,$39,$6B,$FE,$37,$6B,$CB,$C3,$FF,$80,$3C,$A9,$CC,$A3
	.dc.b	$13,$A9,$CC,$62,$7F,$9B,$DF,$95,$FE,$48,$DC,$BF,$C7,$11,$FD,$96
	.dc.b	$BF,$A5,$3F,$DC,$A7,$58,$7F,$B1,$8F,$E9,$EF,$3B,$F2,$75,$DB,$E3
	.dc.b	$7E,$D8,$F2,$11,$FF,$2E,$46,$9F,$81,$19,$78,$1F,$E8,$CE,$A7,$C7
	.dc.b	$F9,$DB,$1F,$EE,$76,$BF,$E4,$65,$F2,$BB,$3F,$06,$E2,$3A,$46,$08
	.dc.b	$E0,$87,$FB,$3A,$38,$37,$C9,$F1,$11,$FB,$B9,$7E,$52,$5F,$B4,$BD
	.dc.b	$5F,$A1,$18,$7F,$55,$48,$C3,$FC,$CA,$78,$36,$7F,$03,$8F,$F8,$16
	.dc.b	$3B,$9F,$81,$70,$FF,$87,$5B,$AF,$FB,$F3,$DA,$75,$FD,$79,$CE,$5F
	.dc.b	$E6,$76,$DF,$3F,$3B,$F0,$CF,$F9,$E6,$B8,$44,$8F,$0F,$07,$D4,$8D
	.dc.b	$E1,$DA,$7C,$A7,$13,$DE,$E5,$FE,$14,$D7,$F5,$86,$CF,$A1,$E7,$7D
	.dc.b	$F2,$3F,$7F,$E7,$BF,$04,$7D,$C2,$3F,$87,$7E,$FF,$CB,$7E,$1C,$91
	.dc.b	$FE,$0C,$F0,$EB,$F8,$78,$BA,$B5,$FE,$1B,$E5,$7F,$86,$EF,$F5,$0D
	.dc.b	$BE,$44,$3F,$21,$F2,$DF,$95,$3F,$1E,$47,$F3,$4F,$E4,$BE,$5B,$F3
	.dc.b	$44,$8F,$97,$E9,$3C,$6C,$47,$E6,$E6,$E5,$FE,$7E,$AD,$4F,$E8,$7C
	.dc.b	$BF,$D2,$11,$FA,$BB,$7F,$8D,$5D,$10,$FD,$34,$7E,$E6,$35,$1F,$A9
	.dc.b	$F9,$0F,$D7,$11,$FB,$FF,$0E,$F8,$3F,$E0,$0D,$8F,$F6,$B3,$5F,$F6
	.dc.b	$A3,$5F,$F6,$93,$5F,$F5,$B3,$5F,$E3,$4D,$E4,$34,$D7,$77,$E9,$FE
	.dc.b	$47,$F8,$11,$2B,$CB,$F1,$35,$FE,$05,$17,$5F,$F5,$2D,$AB,$F8,$56
	.dc.b	$F2,$FF,$85,$3F,$C3,$7C,$BC,$47,$F0,$B8,$B7,$FE,$E3,$ED,$57,$F1
	.dc.b	$3F,$90,$97,$FC,$4D,$FE,$49,$FA,$43,$E3,$69,$FC,$E9,$C4,$FF,$BF
	.dc.b	$5A,$FF,$BE,$DA,$FF,$BE,$5A,$FF,$B7,$5A,$FF,$41,$6E,$C7,$3B,$7F
	.dc.b	$0D,$F9,$0F,$E5,$04,$AF,$E5,$1F,$25,$5F,$E5,$11,$74,$F7,$4F,$F2
	.dc.b	$FF,$90,$95,$FC,$BD,$FE,$65,$F2,$58,$8F,$E6,$11,$6B,$E3,$4F,$F3
	.dc.b	$7F,$92,$95,$FC,$DD,$F8,$B8,$48,$FE,$17,$92,$31,$1F,$EB,$78,$56
	.dc.b	$2F,$FD,$BF,$1B,$9E,$35,$F4,$2B,$FA,$37,$A3,$27,$FA,$33,$FD,$27
	.dc.b	$5C,$47,$F4,$78,$B8,$7D,$D9,$FE,$C9,$2F,$92,$A4,$7F,$A8,$14,$FF
	.dc.b	$07,$77,$FF,$A5,$9E,$07,$F3,$6D,$BF,$C2,$F3,$FD,$66,$3A,$FB,$31
	.dc.b	$D0,$7F,$13,$C3,$23,$F0,$CF,$E2,$48,$E0,$87,$E2,$5C,$BF,$CB,$2B
	.dc.b	$FC,$4F,$F2,$5E,$C2,$47,$83,$9B,$DF,$CF,$FE,$28,$8F,$D8,$9F,$E3
	.dc.b	$FA,$FF,$17,$C6,$4D,$7C,$97,$C6,$24,$FA,$BE,$4B,$E4,$FC,$94,$8F
	.dc.b	$6B,$F3,$3B,$24,$75,$3F,$7A,$47,$C4,$47,$4D,$7D,$87,$FB,$DC,$F9
	.dc.b	$3F,$0B,$F6,$0F,$B3,$F2,$52,$3B,$3E,$D2,$38,$21,$D1,$CB,$DB,$E4
	.dc.b	$BF,$06,$47,$EB,$F9,$5F,$86,$3E,$41,$1F,$63,$5F,$3B,$5F,$F4,$E0
	.dc.b	$7E,$87,$C9,$7E,$40,$91,$F8,$4F,$8B,$FC,$99,$F5,$48,$FC,$96,$BF
	.dc.b	$11,$AF,$67,$EE,$7C,$87,$E7,$09,$1F,$90,$F9,$0F,$CF,$9F,$81,$23
	.dc.b	$F5,$5A,$FD,$06,$BE,$A7,$F1,$1E,$97,$EA,$09,$1F,$9C,$F9,$0F,$D5
	.dc.b	$9F,$8D,$23,$F8,$16,$BF,$71,$AF,$C6,$7F,$29,$F2,$1F,$B4,$24,$7E
	.dc.b	$A3,$DE,$FD,$B9,$F9,$92,$3F,$8D,$7F,$CD,$3F,$2C,$F8,$DF,$F2,$C3
	.dc.b	$67,$FC,$36,$D7,$F7,$67,$F9,$B5,$FF,$12,$F8,$C7,$5D,$FF,$E0,$79
	.dc.b	$D5,$74,$3A,$FA,$BF,$E2,$FF,$84,$93,$FE,$28,$F6,$11,$FE,$33,$F6
	.dc.b	$33,$3F,$E3,$98,$B8,$75,$CB,$FC,$C3,$5F,$E3,$4F,$96,$FC,$44,$FF
	.dc.b	$A8,$BE,$57,$F9,$16,$09,$72,$7F,$A5,$E7,$31,$89,$E1,$E4,$7F,$A6
	.dc.b	$E3,$9C,$27,$7C,$DF,$EA,$98,$C5,$6E,$FA,$36,$39,$9F,$EB,$4B,$E7
	.dc.b	$7C,$B0,$FF,$5B,$F0,$F8,$B0,$FF,$55,$36,$F3,$87,$8F,$1F,$04,$F1
	.dc.b	$9F,$09,$D6,$BC,$2D,$3C,$27,$C1,$9F,$46,$1C,$4E,$D1,$C8,$F8,$D3
	.dc.b	$C9,$E1,$87,$2C,$7E,$1C,$7F,$95,$1A,$F9,$63,$FC,$F3,$18,$8F,$81
	.dc.b	$CF,$FE,$7B,$7F,$DC,$D0,$8C,$47,$36,$BF,$45,$AF,$F9,$8B,$E5,$BF
	.dc.b	$D0,$AF,$FB,$79,$A9,$A4,$7D,$83,$C8,$FC,$A6,$3F,$84,$8E,$46,$A3
	.dc.b	$3A,$7F,$BE,$EB,$E4,$CF,$1F,$1E,$BD,$CE,$0E,$23,$EB,$C7,$FB,$8E
	.dc.b	$BF,$45,$AF,$78,$E0,$F8,$63,$37,$DE,$79,$9E,$83,$B6,$1B,$C3,$B4
	.dc.b	$6A,$3A,$61,$E6,$BF,$4C,$BC,$27,$33,$D1,$FD,$A6,$DF,$B5,$9C,$C6
	.dc.b	$32,$5F,$1B,$7A,$59,$8C,$55,$BE,$37,$FD,$EE,$FB,$36,$F5,$18,$BF
	.dc.b	$E1,$35,$E4,$F4,$B2,$6B,$D1,$B8,$73,$8B,$5C,$2F,$FC,$9B,$39,$3C
	.dc.b	$1E,$4F,$F0,$F9,$F4,$9E,$97,$0F,$B3,$0F,$87,$18,$8D,$4F,$F7,$BA
	.dc.b	$7A,$3E,$83,$D3,$56,$7D,$DC,$1C,$70,$F9,$AD,$BF,$F1,$B4,$F3,$8C
	.dc.b	$11,$C1,$30,$FF,$09,$4F,$F8,$EA,$3E,$57,$9D,$F2,$2B,$E4,$BF,$2B
	.dc.b	$D1,$EC,$49,$F4,$36,$F9,$B5,$F8,$78,$7E,$0A,$3C,$DE,$8F,$D4,$3A
	.dc.b	$57,$D9,$7E,$BF,$A1,$EB,$24,$7C,$56,$DF,$A1,$AF,$6C,$BF,$CF,$93
	.dc.b	$F2,$F8,$BF,$02,$74,$7F,$0C,$FE,$12,$7E,$41,$23,$E9,$34,$F8,$A5
	.dc.b	$3B,$1F,$47,$E5,$7F,$CA,$51,$FE,$4B,$6C,$7E,$1C,$FC,$B3,$8E,$3F
	.dc.b	$D0,$3F,$E4,$B9,$FF,$1D,$36,$7F,$C8,$8D,$7E,$4B,$5F,$C5,$35,$FD
	.dc.b	$73,$73,$34,$9E,$77,$88,$F1,$FA,$16,$FD,$1E,$87,$AC,$8F,$91,$B8
	.dc.b	$79,$1A,$F9,$1F,$85,$E8,$7B,$8F,$7B,$67,$E2,$7A,$1F,$08,$F8,$7E
	.dc.b	$85,$8F,$91,$E8,$7C,$B2,$3F,$11,$F9,$DE,$7F,$CA,$3E,$5F,$A1,$67
	.dc.b	$F8,$3E,$FF,$DC,$FE,$59,$F1,$5F,$AD,$F1,$D1,$8A,$FD,$AE,$DF,$C3
	.dc.b	$A7,$35,$BA,$D4,$EA,$73,$5C,$21,$18,$75,$3C,$1F,$E0,$31,$97,$EC
	.dc.b	$FA,$5F,$82,$3F,$8D,$47,$1C,$3F,$07,$1F,$86,$F1,$E6,$1F,$8B,$F1
	.dc.b	$6E,$1F,$C6,$39,$3F,$CA,$93,$F8,$FA,$FC,$6D,$6B,$0F,$C6,$FA,$5F
	.dc.b	$90,$3F,$23,$E4,$5B,$FC,$B4,$5C,$7A,$4D,$AF,$95,$F2,$DE,$6D,$4D
	.dc.b	$F9,$5C,$33,$1F,$A2,$7C,$B9,$DB,$8C,$3E,$3C,$2B,$84,$7A,$FD,$4F
	.dc.b	$D2,$98,$C3,$F4,$AF,$EA,$3D,$39,$87,$EB,$7D,$4B,$7F,$0E,$6D,$FE
	.dc.b	$B2,$6D,$7E,$9B,$CC,$FD,$89,$FB,$2F,$52,$C7,$ED,$E2,$E1,$F3,$66
	.dc.b	$DF,$F8,$F7,$E5,$95,$F8,$B6,$DF,$E0,$F9,$11,$8C,$3F,$10,$D7,$F0
	.dc.b	$6D,$7F,$20,$D7,$F1,$6D,$8E,$75,$FC,$37,$C1,$FE,$12,$7E,$22,$7B
	.dc.b	$61,$FC,$2A,$3F,$B1,$3F,$E6,$9C,$3F,$DB,$8D,$C3,$FA,$FA,$F4,$F7
	.dc.b	$EA,$4A,$DB,$CA,$31,$5C,$38,$7A,$9E,$1E,$3B,$7D,$A2,$CF,$F9,$D3
	.dc.b	$4F,$87,$58,$C7,$E5,$75,$7C,$92,$61,$F2,$74,$46,$30,$F9,$2D,$7E
	.dc.b	$4B,$FD,$7E,$BF,$15,$5E,$EE,$2F,$89,$3F,$09,$F8,$7D,$32,$7F,$25
	.dc.b	$93,$D9,$F8,$51,$F1,$A7,$58,$FC,$67,$F6,$B3,$A8,$C4,$FC,$CF,$85
	.dc.b	$F9,$E7,$EE,$BD,$C7,$C2,$62,$74,$4A,$F8,$93,$A8,$C6,$9F,$E2,$11
	.dc.b	$F2,$DE,$D2,$47,$FB,$F6,$32,$FF,$11,$F8,$17,$CB,$7D,$D9,$FF,$4F
	.dc.b	$46,$39,$3F,$07,$7F,$E2,$78,$C5,$FE,$13,$E2,$7F,$0E,$67,$0E,$51
	.dc.b	$FE,$15,$C1,$3A,$8E,$07,$EE,$EB,$9E,$5F,$8C,$9F,$C8,$FC,$AF,$D5
	.dc.b	$25,$7A,$5F,$B1,$F2,$DF,$96,$37,$1F,$96,$7C,$DF,$2D,$FD,$6F,$03
	.dc.b	$49,$E5,$FF,$3B,$F9,$67,$C2,$FF,$2A,$6E,$1E,$5A,$FE,$D5,$A7,$FA
	.dc.b	$D6,$73,$A7,$E5,$5B,$19,$BF,$0E,$1C,$D9,$87,$E2,$2B,$6F,$F9,$53
	.dc.b	$1F,$2D,$6B,$0F,$F6,$9F,$C2,$BE,$59,$C9,$FE,$3D,$8C,$36,$33,$87
	.dc.b	$F4,$B9,$CB,$F7,$FE,$5B,$F7,$67,$F9,$7E,$33,$7F,$BA,$8F,$DF,$FC
	.dc.b	$57,$E8,$C9,$5F,$87,$6C,$FF,$5E,$F4,$DB,$27,$33,$FE,$34,$8C,$57
	.dc.b	$E3,$FE,$5B,$F8,$59,$D9,$FE,$16,$FF,$58,$EA,$7B,$BF,$DB,$5B,$3F
	.dc.b	$C8,$DB,$7F,$DA,$5A,$FF,$64,$6B,$FD,$C5,$AF,$F6,$C6,$F6,$7F,$7A
	.dc.b	$AF,$E0,$1F,$09,$FC,$78,$F1,$36,$FF,$8F,$C5,$BF,$E2,$57,$E2,$C7
	.dc.b	$F5,$D3,$E9,$F7,$FD,$04,$BF,$43,$6F,$D0,$D7,$D0,$FF,$11,$F8,$5F
	.dc.b	$E5,$E7,$39,$FE,$5E,$FF,$32,$F9,$6F,$E2,$C4,$8F,$DD,$3F,$C6,$7E
	.dc.b	$5A,$47,$EF,$5B,$7F,$E0,$DF,$96,$7C,$B7,$FB,$1E,$30,$D7,$FC,$33
	.dc.b	$1A,$F7,$7F,$C2,$90,$8C,$3C,$EB,$FD,$62,$DF,$2F,$72,$9C,$CF,$46
	.dc.b	$BF,$C0,$EF,$AF,$1F,$E2,$9B,$B7,$9A,$7F,$8B,$46,$3E,$5A,$47,$F1
	.dc.b	$96,$BF,$93,$97,$CB,$7F,$A9,$9F,$E4,$4D,$D3,$F8,$93,$67,$FC,$5F
	.dc.b	$F2,$DF,$C5,$1F,$E7,$6D,$7F,$C6,$CD,$7F,$C6,$2D,$7F,$C7,$CD,$79
	.dc.b	$F8,$5F,$23,$95,$39,$6C,$66,$38,$67,$FA,$AC,$3C,$8D,$BF,$ED,$F1
	.dc.b	$E7,$BE,$7D,$7A,$CB,$FD,$88,$7F,$3F,$9C,$1B,$C3,$F8,$3D,$FB,$2B
	.dc.b	$C0,$47,$FC,$1D,$1A,$7D,$F1,$96,$9A,$7E,$BF,$53,$E3,$7C,$ED,$9F
	.dc.b	$F1,$A6,$48,$C4,$7F,$B9,$17,$CB,$70,$39,$92,$3F,$CC,$53,$C6,$31
	.dc.b	$1F,$C3,$E7,$8B,$A6,$B9,$8F,$2F,$93,$E4,$24,$7F,$91,$E5,$F2,$CF
	.dc.b	$57,$F0,$51,$87,$F3,$ED,$9F,$E5,$53,$B6,$CF,$DD,$C7,$FC,$41,$1F
	.dc.b	$E0,$26,$BF,$E1,$76,$EB,$FE,$F9,$F6,$9D,$7F,$6E,$73,$BF,$FA,$1D
	.dc.b	$B7,$FE,$82,$6A,$AE,$2F,$7F,$4F,$91,$1F,$F3,$21,$BB,$FC,$F3,$4F
	.dc.b	$DD,$CF,$27,$FD,$20,$D7,$FD,$32,$D4,$36,$FB,$9E,$77,$E0,$09,$1F
	.dc.b	$80,$D5,$F2,$DF,$84,$3F,$0E,$48,$FB,$0F,$E2,$BE,$5B,$F1,$24,$8F
	.dc.b	$F0,$17,$9B,$FD,$2D,$E9,$19,$7F,$14,$FE,$E7,$C8,$FF,$0B,$F8,$B6
	.dc.b	$B8,$57,$EA,$67,$2F,$49,$FC,$BF,$CB,$48,$FC,$53,$F9,$C2,$3F,$8D
	.dc.b	$F9,$6F,$CE,$91,$FF,$1F,$FA,$5F,$D5,$9F,$81,$19,$7F,$3C,$FE,$DE
	.dc.b	$35,$1F,$A1,$F3,$7F,$4E,$47,$FC,$93,$AF,$F8,$D9,$F9,$51,$97,$81
	.dc.b	$AF,$47,$F5,$FF,$1F,$23,$E8,$3F,$C1,$7E,$1D,$F0,$74,$FE,$24,$E2
	.dc.b	$7F,$D1,$8D,$7F,$DA,$4D,$7F,$D9,$AD,$7F,$DB,$4D,$7F,$DA,$AD,$EF
	.dc.b	$39,$E3,$FE,$01,$F2,$1F,$C1,$09,$5F,$E6,$69,$7C,$4E,$23,$F8,$24
	.dc.b	$DE,$B7,$5F,$82,$BF,$86,$F9,$D2,$FF,$1C,$FF,$10,$F3,$71,$1F,$C3
	.dc.b	$62,$DF,$28,$FC,$C7,$4A,$BF,$8B,$FC,$7C,$9F,$CC,$3F,$CB,$7E,$59
	.dc.b	$F1,$94,$FE,$74,$E2,$7F,$D6,$8D,$7F,$DF,$4D,$7F,$DE,$AD,$7F,$53
	.dc.b	$BA,$31,$1F,$EF,$E6,$C6,$67,$F9,$6F,$C8,$7F,$2C,$25,$7E,$B7,$E4
	.dc.b	$6B,$FC,$AE,$2E,$7E,$FE,$1F,$E6,$15,$FC,$D3,$E4,$25,$FE,$D5,$FE
	.dc.b	$6F,$F2,$38,$8F,$E6,$91,$6F,$96,$CA,$FE,$7B,$F2,$32,$FF,$7C,$FC
	.dc.b	$7C,$A4,$7F,$A8,$B5,$46,$23,$FD,$71,$1C,$DD,$46,$70,$FA,$DA,$FC
	.dc.b	$58,$FA,$58,$7F,$4C,$F4,$E5,$FF,$0D,$7F,$A7,$ED,$88,$FE,$99,$17
	.dc.b	$0F,$4C,$FF,$6A,$97,$CB,$52,$3F,$C5,$CA,$7F,$6F,$DF,$F8,$FC,$62
	.dc.b	$BE,$1C,$F3,$F6,$E3,$FD,$78,$3F,$89,$E1,$91,$F8,$67,$F1,$24,$70
	.dc.b	$43,$F1,$2E,$5F,$F3,$60,$F9,$2F,$51,$1F,$07,$2F,$B7,$A3,$FC,$53
	.dc.b	$1F,$B2,$3F,$B4,$94,$E6,$EB,$E4,$BE,$21,$23,$D5,$F2,$5F,$1F,$E4
	.dc.b	$A4,$7B,$5F,$95,$D9,$23,$AB,$E0,$23,$82,$1E,$A7,$2F,$B3,$E1,$BE
	.dc.b	$A1,$1F,$95,$F2,$5F,$5C,$F5,$91,$F5,$B5,$F8,$AD,$7F,$AA,$9F,$D5
	.dc.b	$43,$F1,$3E,$47,$F0,$44,$8F,$B1,$E6,$7E,$14,$F9,$24,$7E,$4B,$5F
	.dc.b	$A6,$D7,$BF,$17,$F6,$48,$FA,$3F,$23,$F8,$F2,$4F,$E1,$7E,$2F,$F2
	.dc.b	$47,$D6,$23,$F5,$9A,$FE,$01,$AF,$60,$FD,$EF,$90,$FC,$E1,$23,$F2
	.dc.b	$5F,$21,$F9,$F3,$F0,$64,$7F,$06,$D7,$F1,$AD,$7B,$3F,$8A,$F5,$3F
	.dc.b	$50,$48,$FC,$F7,$C8,$7E,$AC,$FC,$79,$1F,$C7,$B5,$FC,$CB,$5F,$53
	.dc.b	$F9,$6F,$90,$FD,$A1,$23,$F5,$5F,$05,$FB,$73,$F3,$64,$7F,$36,$FF
	.dc.b	$9A,$FE,$59,$F1,$DF,$E6,$36,$CF,$F8,$5B,$AF,$F8,$5C,$62,$3F,$CD
	.dc.b	$5E,$0A,$3F,$C5,$AD,$AE,$44,$9E,$5E,$4F,$F3,$8A,$DF,$B2,$F8,$6F
	.dc.b	$F1,$C1,$2B,$FC,$81,$EC,$D7,$FC,$81,$17,$2F,$E3,$B5,$FE,$2F,$7F
	.dc.b	$AA,$4B,$E5,$BF,$C7,$91,$FE,$2E,$F9,$6F,$E5,$33,$FE,$30,$97,$27
	.dc.b	$E7,$A7,$33,$C6,$31,$E3,$FE,$8A,$B9,$CA,$B7,$18,$6D,$70,$9E,$CD
	.dc.b	$7C,$9C,$DF,$EF,$C7,$95,$F5,$8C,$70,$F2,$9F,$F2,$2B,$53,$E0,$8D
	.dc.b	$CF,$1C,$7D,$CF,$9E,$FC,$18,$FB,$B0,$C9,$F0,$6F,$C5,$8F,$3C,$33
	.dc.b	$3F,$04,$B5,$C5,$F3,$9C,$63,$35,$EB,$8F,$E0,$E7,$BA,$3F,$DB,$B3
	.dc.b	$F2,$23,$2D,$7E,$67,$47,$FA,$6D,$E2,$7D,$97,$0F,$06,$BC,$3C,$AF
	.dc.b	$F7,$4B,$FE,$6A,$F9,$6A,$9A,$47,$6F,$B9,$79,$1F,$E1,$BC,$BF,$03
	.dc.b	$1C,$CF,$79,$FE,$76,$8D,$69,$E6,$AE,$51,$89,$F8,$31,$8E,$2E,$D1
	.dc.b	$E3,$3E,$CD,$7C,$6A,$FF,$15,$6D,$E0,$9D,$3E,$28,$CC,$F7,$9F,$19
	.dc.b	$E8,$AE,$B7,$DE,$32,$77,$C3,$94,$EA,$38,$DF,$9D,$FE,$7B,$5C,$6B
	.dc.b	$C4,$DB,$F0,$47,$29,$EB,$39,$7D,$E7,$4B,$ED,$3B,$39,$C6,$4E,$25
	.dc.b	$73,$87,$52,$DF,$C5,$AF,$D0,$BE,$23,$15,$CA,$BB,$47,$C5,$AE,$4D
	.dc.b	$36,$F4,$9F,$EB,$78,$7C,$C7,$D3,$1C,$1F,$5D,$74,$6D,$78,$23,$94
	.dc.b	$7E,$0E,$FC,$AF,$29,$EE,$78,$5E,$F1,$A7,$E7,$9F,$82,$32,$53,$33
	.dc.b	$E1,$6C,$F8,$4E,$8F,$84,$E0,$7E,$5E,$BD,$4F,$66,$B9,$8F,$2B,$EC
	.dc.b	$3C,$33,$F5,$8F,$03,$B3,$CB,$3F,$01,$FC,$81,$E0,$3F,$2F,$3D,$1F
	.dc.b	$03,$5F,$D5,$4F,$47,$F2,$91,$EB,$7E,$41,$D9,$F6,$BD,$A7,$C1,$7F
	.dc.b	$5E,$3A,$9B,$3D,$47,$C3,$AF,$55,$75,$8D,$9F,$1E,$B9,$DF,$D3,$8F
	.dc.b	$71,$99,$F0,$DF,$78,$F4,$C7,$36,$9F,$21,$FC,$A3,$4F,$19,$F1,$AF
	.dc.b	$BB,$F1,$F0,$F0,$57,$58,$F8,$67,$13,$EF,$9C,$F9,$1F,$A4,$3C,$6F
	.dc.b	$EB,$B3,$FE,$A0,$8C,$65,$FC,$A2,$B7,$5A,$A4,$62,$75,$5C,$27,$F9
	.dc.b	$BC,$E5,$B1,$97,$FC,$83,$19,$6F,$2F,$FB,$2A,$BF,$BD,$C7,$68,$C3
	.dc.b	$FE,$04,$AE,$DC,$1F,$E0,$8A,$FF,$64,$3F,$E0,$AF,$95,$B1,$FE,$03
	.dc.b	$6D,$7A,$3E,$51,$EC,$E3,$2F,$F0,$75,$5C,$8D,$35,$5C,$30,$F2,$D7
	.dc.b	$06,$FC,$8B,$9F,$A3,$63,$7F,$AB,$3C,$0C,$D4,$9E,$2F,$8E,$75,$18
	.dc.b	$3B,$EC,$F5,$BD,$0D,$8F,$8F,$F3,$0F,$D6,$9F,$0F,$AB,$95,$C9,$F8
	.dc.b	$75,$B9,$CF,$0F,$F8,$C3,$81,$BF,$C3,$33,$C5,$FE,$36,$7D,$B5,$31
	.dc.b	$F2,$FB,$97,$0F,$F1,$CF,$07,$8A,$7E,$2D,$7A,$2B,$58,$7F,$90,$3E
	.dc.b	$47,$E8,$1F,$4F,$F6,$2F,$F6,$9D,$FC,$EC,$D3,$A8,$C3,$BE,$AC,$67
	.dc.b	$F8,$5F,$7F,$F2,$DC,$BF,$9E,$6F,$11,$A4,$C3,$F3,$2D,$AF,$CB,$BC
	.dc.b	$B6,$7C,$4F,$DC,$3E,$8C,$5A,$F3,$7C,$57,$E2,$4F,$B5,$E1,$D8,$C9
	.dc.b	$30,$FF,$2E,$F3,$7F,$91,$A3,$1C,$1F,$EE,$CC,$BC,$D1,$F8,$CC,$7F
	.dc.b	$2D,$FB,$4B,$E0,$7F,$1B,$12,$74,$7E,$52,$74,$E3,$1F,$F3,$5C,$74
	.dc.b	$7F,$6A,$D7,$F6,$6D,$47,$EF,$9C,$36,$7E,$6C,$63,$63,$4C,$70,$BE
	.dc.b	$0B,$F5,$87,$AA,$AE,$1F,$A8,$F8,$9F,$D3,$9F,$B3,$F8,$47,$C2,$E2
	.dc.b	$74,$4C,$FF,$D0,$B3,$A8,$C6,$DF,$E8,$AF,$8D,$FD,$D1,$FB,$DF,$01
	.dc.b	$F0,$58,$AF,$C6,$4D,$C3,$FD,$1F,$5A,$9C,$9F,$BE,$FD,$27,$B4,$FF
	.dc.b	$1C,$6B,$FC,$69,$AF,$C8,$E0,$46,$21,$E8,$7F,$0C,$F0,$BF,$89,$B6
	.dc.b	$FB,$6B,$8C,$7F,$84,$5F,$FA,$8B,$EC,$7F,$8F,$D7,$56,$C7,$49,$DE
	.dc.b	$AA,$D7,$8D,$FC,$76,$F7,$A1,$E3,$97,$87,$1E,$D5,$FC,$8F,$EB,$3D
	.dc.b	$16,$69,$33,$FD,$6F,$23,$E7,$3F,$94,$1C,$62,$E7,$FC,$9B,$F0,$B7
	.dc.b	$B9,$FB,$46,$C7,$F4,$D6,$E7,$FD,$21,$AF,$F4,$4F,$9D,$FE,$68,$7F
	.dc.b	$37,$8D,$EF,$FE,$C8,$8C,$D7,$F0,$77,$57,$8B,$FE,$39,$16,$BE,$34
	.dc.b	$7F,$5A,$BE,$77,$FE,$C9,$AF,$E8,$9F,$B6,$9F,$E1,$6D,$8F,$EC,$8D
	.dc.b	$AF,$EB,$ED,$7F,$AD,$FC,$E7,$F4,$B3,$9D,$5A,$FE,$6F,$FA,$93,$DC
	.dc.b	$FE,$D9,$E5,$A3,$13,$FE,$98,$8C,$4E,$A7,$38,$7F,$A9,$27,$35,$B9
	.dc.b	$CC,$FE,$E8,$4E,$9E,$11,$A7,$F7,$FF,$56,$FD,$5B,$FA,$F4,$35,$F8
	.dc.b	$F3,$9F,$47,$FA,$4D,$F8,$A7,$37,$C1,$FE,$E7,$F8,$07,$F6,$AC,$BF
	.dc.b	$D6,$30,$8C,$63,$FE,$BC,$8C,$74,$FF,$83,$5B,$5C,$2B,$75,$EB,$9C
	.dc.b	$BC,$63,$2F,$F1,$88,$CB,$F0,$73,$FF,$7C,$7C,$3F,$F6,$0F,$68,$FF
	.dc.b	$81,$2B,$99,$F8,$49,$DE,$1F,$DB,$2B,$6D,$AE,$18,$7A,$AF,$85,$6B
	.dc.b	$0F,$91,$7F,$EB,$DF,$BE,$7C,$0F,$F8,$A7,$D7,$46,$27,$FD,$AB,$18
	.dc.b	$9D,$4E,$6B,$FD,$78,$D7,$FD,$1E,$D7,$B4,$70,$3F,$17,$F7,$5F,$CE
	.dc.b	$AF,$FC,$0A,$D9,$F2,$9F,$ED,$6A,$ED,$8F,$C1,$9D,$3F,$A6,$F7,$3E
	.dc.b	$C7,$C9,$FF,$90,$5B,$87,$FB,$B1,$B8,$7F,$BD,$9A,$FE,$45,$FC,$07
	.dc.b	$AB,$C7,$53,$49,$5F,$EF,$6A,$7D,$03,$E7,$FF,$DD,$71,$FE,$FA,$8C
	.dc.b	$6D,$B6,$C6,$67,$6F,$E8,$4E,$7F,$41,$FC,$C9,$E4,$D7,$52,$8C,$47
	.dc.b	$EA,$63,$1E,$27,$BF,$EF,$BF,$DD,$92,$D1,$18,$6F,$AF,$EB,$6C,$F1
	.dc.b	$F9,$8C,$51,$A4,$8F,$E0,$1F,$5A,$F8,$FF,$D0,$6D,$FD,$D3,$A1,$39
	.dc.b	$AF,$F2,$0B,$5F,$BA,$75,$F9,$6E,$2F,$F7,$CC,$11,$F3,$E3,$47,$D1
	.dc.b	$8C,$B6,$35,$18,$8F,$98,$70,$2B,$97,$E8,$35,$FD,$A7,$E0,$AF,$93
	.dc.b	$F1,$F3,$23,$13,$FE,$80,$8C,$4F,$A2,$3E,$6F,$C6,$F0,$AD,$C7,$E6
	.dc.b	$AB,$6F,$18,$CC,$7F,$3E,$8E,$27,$FA,$30,$A7,$AB,$55,$7C,$36,$C7
	.dc.b	$E4,$DA,$EE,$34,$47,$4F,$E6,$7C,$17,$EA,$AF,$A5,$FA,$31,$86,$C6
	.dc.b	$6B,$C9,$F4,$DE,$8E,$0F,$E7,$0F,$BA,$3F,$34,$D7,$3A,$7D,$98,$C4
	.dc.b	$7F,$8E,$1A,$9A,$48,$FD,$77,$9E,$F3,$9F,$3D,$F8,$68,$C3,$63,$37
	.dc.b	$FE,$A0,$F2,$FF,$4C,$DE,$AE,$71,$88,$FE,$C7,$18,$3B,$CF,$8E,$3F
	.dc.b	$CF,$98,$F5,$3C,$1E,$67,$FA,$96,$32,$43,$FD,$29,$1A,$F8,$9E,$11
	.dc.b	$9E,$2F,$EF,$79,$FA,$EB,$53,$FE,$11,$AF,$7D,$F5,$8F,$F0,$74,$60
	.dc.b	$F2,$FC,$87,$BC,$91,$FE,$BE,$F4,$BF,$DC,$CD,$F4,$FE,$4C,$62,$3F
	.dc.b	$DB,$ED,$7F,$DB,$AD,$7F,$C7,$1D,$1F,$37,$D4,$FD,$56,$9E,$86,$F1
	.dc.b	$7F,$9C,$3B,$DF,$B4,$7F,$56,$C3,$FD,$9F,$5A,$7F,$DE,$2D,$DF,$FC
	.dc.b	$79,$8F,$EF,$AF,$8F,$97,$70,$E5,$87,$F4,$09,$CB,$FE,$EE,$6B,$FD
	.dc.b	$12,$FF,$CB,$3A,$7E,$0F,$D7,$F5,$4F,$FC,$00,$6C,$EA,$3F,$81,$B5
	.dc.b	$FF,$80,$3C,$4F,$91,$FC,$51,$23,$FD,$D3,$E7,$7F,$36,$DF,$EC,$D6
	.dc.b	$E7,$FD,$81,$5B,$7F,$CE,$BE,$3F,$E5,$7E,$07,$F9,$44,$7F,$BC,$3C
	.dc.b	$1C,$1F,$99,$F7,$3F,$A6,$C6,$48,$7F,$B5,$23,$5E,$57,$E7,$7B,$E4
	.dc.b	$FF,$B3,$DF,$D1,$7E,$A3,$FE,$58,$8F,$F7,$FB,$5F,$E9,$9B,$23,$11
	.dc.b	$FE,$2F,$8C,$7B,$1F,$AE,$9C,$CF,$FA,$15,$AA,$76,$FC,$A8,$CC,$62
	.dc.b	$3B,$E5,$F8,$FD,$BF,$D8,$DC,$3F,$0C,$F3,$FC,$DD,$AF,$F1,$17,$47
	.dc.b	$FA,$4E,$F8,$36,$B7,$3F,$EE,$DE,$EF,$DF,$11,$E2,$79,$7E,$11,$F2
	.dc.b	$39,$8F,$C3,$72,$7E,$FF,$3F,$B3,$7F,$C2,$6F,$75,$CB,$46,$6F,$A0
	.dc.b	$FB,$44,$9D,$7E,$CF,$FC,$85,$F8,$8F,$F1,$CA,$34,$9E,$2F,$F1,$A6
	.dc.b	$CF,$E4,$9A,$FE,$4B,$D1,$FE,$34,$47,$C9,$F1,$AF,$A1,$C4,$7B,$26
	.dc.b	$EB,$FC,$32,$B5,$3E,$BC,$DE,$97,$E4,$C9,$1E,$FF,$8D,$FF,$67,$BB
	.dc.b	$8C,$BF,$AD,$9E,$2E,$FE,$17,$F9,$77,$D4,$C8,$F9,$1C,$48,$FE,$69
	.dc.b	$87,$8A,$7F,$6C,$DC,$BF,$90,$B7,$97,$FD,$A4,$D9,$FF,$40,$FC,$9B
	.dc.b	$E4,$F9,$EF,$27,$D0,$FF,$81,$BD,$55,$6E,$7E,$07,$8F,$FB,$12,$3F
	.dc.b	$77,$9B,$FC,$BA,$E8,$87,$BE,$3F,$09,$EE,$FF,$4B,$FA,$99,$3E,$E7
	.dc.b	$FA,$87,$CA,$3E,$3F,$33,$F8,$E9,$BB,$FF,$3E,$AD,$4F,$D3,$D3,$F8
	.dc.b	$21,$1F,$A5,$D5,$F2,$0F,$EB,$FB,$FE,$46,$3F,$8B,$B5,$FD,$DB,$5F
	.dc.b	$DD,$35,$F9,$71,$C8,$8F,$F9,$5F,$E0,$5F,$31,$FE,$A7,$87,$CB,$FD
	.dc.b	$E5,$2F,$92,$FF,$5A,$4E,$63,$12,$38,$F4,$FF,$AE,$63,$AE,$99,$AE
	.dc.b	$18,$77,$F0,$39,$63,$FE,$BF,$6E,$3C,$DD,$DD,$C7,$A5,$F9,$F0,$DD
	.dc.b	$F8,$0F,$16,$3E,$47,$07,$AA,$78,$57,$2B,$F2,$3E,$7C,$7D,$AD,$AC
	.dc.b	$9D,$2B,$8B,$6B,$C8,$F8,$2B,$D5,$5C,$63,$F8,$8B,$DD,$A7,$12,$9F
	.dc.b	$3A,$38,$BC,$CE,$A7,$A6,$6D,$FC,$C8,$F6,$BE,$FC,$3E,$19,$F2,$23
	.dc.b	$38,$7C,$FC,$3C,$98,$FC,$83,$44,$72,$49,$F3,$47,$3A,$EB,$3E,$D7
	.dc.b	$0F,$B6,$3A,$3C,$67,$D0,$7B,$AF,$D1,$3A,$8F,$71,$9B,$F3,$3F,$16
	.dc.b	$FB,$BA,$7B,$BE,$4B,$ED,$7F,$7E,$FD,$53,$93,$AB,$DD,$F1,$BC,$9F
	.dc.b	$A0,$7D,$E2,$9F,$23,$0C,$4F,$A2,$BB,$E3,$F5,$AB,$9C,$7C,$13,$2F
	.dc.b	$D6,$C7,$D6,$FC,$2C,$7D,$6F,$67,$C5,$7D,$63,$CF,$5E,$CA,$E3,$39
	.dc.b	$9F,$2E,$3E,$B8,$C1,$F8,$3B,$EA,$FE,$0E,$FF,$21,$1F,$12,$B8,$C7
	.dc.b	$5C,$7C,$C7,$BB,$1F,$D0,$CF,$CF,$9E,$24,$70,$43,$D8,$E5,$F6,$3F
	.dc.b	$85,$31,$7E,$57,$11,$C6,$30,$FD,$F8,$CB,$C8,$C1,$0C,$12,$38,$BF
	.dc.b	$48,$F9,$24,$30,$79,$8F,$04,$FB,$9C,$12,$3E,$43,$E5,$C7,$F7,$EF
	.dc.b	$E6,$8F,$F7,$A4,$7E,$C5,$B1,$FC,$A5,$B7,$FC,$B7,$E6,$3F,$DD,$D1
	.dc.b	$FE,$D0,$A4,$62,$39,$B5,$FE,$50,$D7,$F9,$61,$FC,$B7,$E6,$3F,$DC
	.dc.b	$09,$A4,$7F,$DC,$8F,$E6,$D7,$96,$E9,$C4,$71,$3F,$92,$9A,$73,$3C
	.dc.b	$67,$84,$E8,$E1,$A7,$F4,$57,$81,$99,$FE,$2E,$E3,$5D,$3D,$DE,$15
	.dc.b	$D1,$CC,$F7,$BF,$0D,$79,$2B,$C3,$7A,$8C,$4E,$A3,$9B,$75,$F8,$F3
	.dc.b	$E4,$9E,$06,$9F,$31,$C8,$AE,$4D,$C7,$2A,$F0,$BD,$27,$E0,$B9,$AF
	.dc.b	$04,$7D,$69,$DB,$E2,$BF,$63,$E0,$C3,$C4,$7A,$CB,$7D,$67,$78,$78
	.dc.b	$AB,$A6,$1F,$12,$F7,$1D,$4F,$E9,$27,$CF,$BE,$4E,$30,$F9,$D1,$89
	.dc.b	$E3,$3F,$8C,$C3,$BC,$F0,$C3,$E1,$3B,$3D,$D3,$D8,$FB,$4E,$2B,$D4
	.dc.b	$D3,$84,$7A,$0E,$15,$D9,$F4,$57,$67,$D5,$5E,$47,$F3,$C7,$A7,$0F
	.dc.b	$C5,$BE,$03,$A4,$78,$1B,$7E,$93,$CF,$5F,$26,$3F,$C2,$98,$7D,$47
	.dc.b	$E0,$1D,$4F,$29,$6B,$A3,$C2,$FC,$38,$7E,$72,$BE,$5C,$FA,$B0,$EE
	.dc.b	$78,$5F,$81,$87,$27,$EA,$1E,$E8,$F8,$CD,$3D,$25,$3C,$07,$92,$3E
	.dc.b	$33,$E8,$6B,$F1,$B0,$F9,$0F,$C5,$38,$BE,$1A,$F3,$C7,$C0,$8D,$DF
	.dc.b	$D3,$8E,$97,$EA,$7D,$87,$A9,$F6,$3E,$4B,$E9,$87,$C4,$8F,$91,$3A
	.dc.b	$7F,$60,$DC,$3F,$47,$E1,$38,$1E,$27,$F0,$CB,$FF,$2B,$B5,$FF,$40
	.dc.b	$B5,$FF,$67,$47,$0C,$3F,$CC,$33,$9A,$E8,$D9,$D5,$7F,$99,$23,$5D
	.dc.b	$BF,$C8,$6F,$FD,$2A,$7F,$1C,$DF,$FD,$23,$36,$3F,$CB,$58,$2B,$85
	.dc.b	$FF,$B2,$3C,$DD,$3F,$EE,$18,$46,$25,$CA,$F8,$9C,$56,$B0,$FF,$52
	.dc.b	$DE,$0E,$15,$B8,$C9,$FE,$B7,$F8,$FE,$9E,$11,$ED,$DA,$F5,$F3,$12
	.dc.b	$B8,$1F,$E9,$D8,$C6,$9F,$EB,$0F,$07,$CD,$C1,$FE,$AB,$AF,$F7,$3F
	.dc.b	$83,$FE,$A8,$F9,$8C,$D4,$9F,$F6,$23,$F0,$27,$51,$83,$FD,$7F,$A7
	.dc.b	$B8,$E1,$87,$FB,$B9,$F4,$DC,$9F,$F6,$77,$9D,$6B,$FD,$99,$37,$8F
	.dc.b	$FD,$87,$36,$BE,$9F,$03,$E6,$3E,$59,$F3,$FC,$7F,$F6,$DC,$FF,$B4
	.dc.b	$23,$2D,$3F,$DC,$FC,$F8,$CB,$FD,$A3,$F2,$9F,$87,$24,$FF,$AB,$1B
	.dc.b	$1F,$EA,$C6,$BF,$EA,$C7,$F2,$1C,$6F,$98,$FA,$A7,$CC,$8B,$5F,$EE
	.dc.b	$2F,$98,$FC,$09,$F6,$3C,$1B,$19,$26,$0D,$BF,$B7,$F4,$B9,$FF,$D6
	.dc.b	$15,$A9,$CE,$6C,$BF,$0A,$7E,$47,$E6,$3F,$DA,$55,$FE,$FF,$8C,$B4
	.dc.b	$FF,$5B,$57,$69,$DE,$3F,$EF,$DA,$E4,$7E,$B6,$3B,$57,$F6,$1C,$51
	.dc.b	$8A,$FE,$DF,$18,$AE,$04,$AE,$15,$BA,$FF,$0D,$D6,$E7,$A7,$03,$E6
	.dc.b	$3F,$4E,$78,$A3,$CE,$7E,$09,$FD,$1F,$CC,$48,$CD,$AD,$EB,$BE,$5B
	.dc.b	$53,$F3,$6B,$F7,$BA,$B0,$CE,$BF,$43,$0E,$15,$D3,$15,$EE,$31,$9B
	.dc.b	$E6,$3F,$7A,$7F,$01,$EE,$7A,$18,$AD,$CD,$D3,$C4,$79,$EB,$73,$A3
	.dc.b	$C1,$ED,$F8,$A7,$E6,$B5,$F9,$2D,$7F,$8B,$B5,$7B,$9F,$C5,$3C,$1F
	.dc.b	$E2,$ED,$AE,$BC,$18,$C3,$F8,$64,$7B,$FD,$FF,$E5,$3C,$1F,$92,$6C
	.dc.b	$7F,$BF,$6F,$96,$3E,$AD,$D5,$A8,$FA,$4D,$8C,$F8,$9D,$F3,$FE,$4D
	.dc.b	$39,$F8,$0F,$45,$1A,$4D,$3F,$9C,$FC,$C7,$F2,$A3,$F9,$74,$7C,$5D
	.dc.b	$3F,$4F,$C2,$9F,$E6,$5F,$06,$F7,$7F,$A3,$37,$7F,$CD,$37,$6F,$CC
	.dc.b	$B5,$FC,$AD,$BE,$5F,$F9,$C9,$D2,$ED,$FE,$0A,$2D,$7C,$3B,$E9,$8F
	.dc.b	$92,$72,$DA,$FC,$64,$E9,$FE,$3B,$3F,$D2,$7E,$1B,$C1,$1F,$AF,$6D
	.dc.b	$FE,$B9,$AF,$EA,$9A,$FF,$26,$F9,$8F,$E9,$E7,$3A,$B5,$FC,$FB,$E5
	.dc.b	$DF,$31,$FD,$B7,$C9,$46,$27,$FC,$BD,$18,$9D,$4E,$70,$FF,$36,$CE
	.dc.b	$6B,$73,$99,$FF,$4B,$CE,$5D,$C6,$50,$F8,$C7,$CC,$7F,$61,$8F,$D2
	.dc.b	$C6,$3C,$FF,$EA,$15,$E1,$DB,$FD,$27,$39,$C7,$F8,$0C,$3E,$57,$FB
	.dc.b	$76,$1F,$E8,$28,$C7,$57,$F8,$2D,$B5,$FE,$7B,$6C,$EE,$B8,$BF,$EB
	.dc.b	$08,$CB,$63,$2F,$F2,$A6,$AC,$DE,$B3,$E2,$BF,$B4,$37,$0F,$3D,$F9
	.dc.b	$B0,$FF,$56,$DF,$0B,$D9,$2B,$78,$DC,$7F,$84,$FA,$6F,$98,$7B,$DF
	.dc.b	$E2,$78,$C3,$63,$37,$FE,$99,$9C,$B6,$33,$5F,$E9,$F6,$BF,$E9,$B6
	.dc.b	$BF,$E6,$A6,$BE,$FF,$98,$FB,$73,$FE,$0A,$6C,$79,$A7,$AE,$7E,$E9
	.dc.b	$CB,$FE,$CC,$F9,$6F,$B1,$F3,$1F,$E3,$F6,$E1,$FE,$AF,$6E,$1F,$EB
	.dc.b	$C6,$BF,$EE,$47,$F0,$1E,$9F,$1D,$CD,$25,$7F,$AE,$69,$F3,$0F,$86
	.dc.b	$FF,$72,$C7,$FA,$F6,$31,$AE,$DB,$19,$9D,$BF,$EC,$E8,$7C,$4F,$F3
	.dc.b	$08,$E4,$D7,$4D,$7F,$DF,$71,$C3,$C6,$F7,$7C,$AF,$FB,$9A,$5A,$23
	.dc.b	$0D,$F5,$7D,$4D,$9E,$3F,$1D,$8A,$34,$91,$FB,$CF,$97,$7C,$C7,$E7
	.dc.b	$76,$FD,$9F,$42,$73,$5F,$E4,$46,$BF,$D6,$0E,$BF,$13,$C5,$FD,$DE
	.dc.b	$08,$F9,$D1,$A3,$E8,$46,$5B,$1A,$8C,$47,$FB,$B5,$AF,$FB,$E9,$AF
	.dc.b	$F0,$83,$A1,$FB,$DF,$98,$7C,$77,$8F,$99,$18,$9F,$E0,$AD,$7D,$5A
	.dc.b	$3E,$63,$84,$E6,$BF,$D0,$8D,$D3,$FC,$4E,$DB,$FE,$33,$7C,$5F,$F0
	.dc.b	$89,$F5,$EF,$E0,$C6,$63,$D0,$70,$8C,$91,$FC,$A4,$6D,$DB,$5F,$80
	.dc.b	$D7,$F9,$47,$CC,$3E,$4F,$E9,$78,$08,$C4,$FF,$89,$23,$13,$F5,$E1
	.dc.b	$F2,$FE,$B6,$E9,$F6,$23,$E3,$47,$A1,$AF,$D1,$6B,$FC,$E9,$AF,$23
	.dc.b	$F5,$CE,$0D,$F9,$FF,$EA,$BF,$98,$FC,$43,$67,$FC,$57,$18,$F2,$BF
	.dc.b	$57,$1F,$E3,$16,$BF,$D8,$7A,$7F,$75,$49,$CC,$FF,$63,$9C,$F9,$BE
	.dc.b	$E8,$FF,$1B,$E1,$E0,$8C,$47,$F7,$7F,$87,$C4,$C8,$D9,$FD,$DA,$30
	.dc.b	$FF,$87,$E7,$8F,$17,$F8,$42,$BD,$7E,$6F,$FA,$CE,$B8,$1F,$E0,$EB
	.dc.b	$F3,$C6,$63,$FC,$1F,$1E,$6F,$98,$F2,$12,$3F,$CA,$1E,$8F,$FB,$99
	.dc.b	$BE,$AF,$C0,$8C,$47,$F9,$C1,$AF,$F9,$B1,$AF,$FB,$6B,$D4,$FD,$6E
	.dc.b	$5F,$E3,$BE,$9F,$A3,$18,$E8,$FF,$1E,$7C,$C7,$F5,$7B,$FF,$21,$56
	.dc.b	$88,$FF,$9B,$BC,$04,$EE,$3F,$CB,$71,$8F,$2F,$F8,$23,$63,$84,$EF
	.dc.b	$1F,$F4,$EF,$AF,$EA,$6E,$9F,$72,$3D,$35,$FE,$7B,$6F,$21,$BC,$4F
	.dc.b	$98,$FB,$84,$AF,$F3,$C7,$9B,$FC,$DB,$1F,$F2,$2C,$E6,$7E,$D3,$F9
	.dc.b	$29,$D6,$DF,$E8,$4E,$FF,$CA,$FB,$7F,$CA,$AF,$FD,$04,$D7,$FC,$C1
	.dc.b	$EE,$7F,$4E,$73,$3C,$08,$FF,$A2,$63,$4F,$E2,$7B,$3F,$24,$47,$FD
	.dc.b	$1D,$F3,$1F,$E5,$76,$FB,$1F,$9E,$8C,$47,$FA,$B9,$AF,$FA,$A9,$AF
	.dc.b	$F8,$B3,$74,$62,$3F,$D4,$8D,$8C,$CF,$53,$F5,$78,$FD,$3C,$BF,$D4
	.dc.b	$BC,$1F,$08,$FD,$BD,$BE,$63,$18,$FC,$49,$AF,$FA,$0E,$74,$FF,$A9
	.dc.b	$BB,$3F,$60,$47,$FD,$59,$F0,$6F,$8F,$C4,$7E,$B6,$6F,$17,$FB,$8A
	.dc.b	$B5,$3F,$4B,$E4,$BE,$A9,$1F,$F5,$B7,$CC,$7F,$35,$DC,$FF,$10,$27
	.dc.b	$87,$FC,$21,$B3,$FE,$DB,$6B,$FE,$DA,$7F,$8B,$F3,$7F,$81,$63,$FD
	.dc.b	$7F,$E7,$FF,$09,$3F,$8C,$FC,$5B,$E5,$31,$3A,$21,$FE,$A5,$8D,$3F
	.dc.b	$8A,$F4,$3F,$1A,$47,$FD,$9D,$F2,$1F,$EC,$87,$44,$3F,$D5,$DC,$A9
	.dc.b	$FE,$53,$3B,$9F,$E2,$9F,$31,$FC,$74,$8F,$4C,$36,$47,$07,$FB,$A3
	.dc.b	$D0,$7C,$C3,$E4,$B5,$3F,$EB,$DC,$3D,$2D,$BF,$F0,$05,$F0,$AF,$D3
	.dc.b	$F8,$92,$BF,$4E,$FF,$3E,$E3,$FF,$2A,$C6,$08,$7B,$23,$5E,$E7,$F3
	.dc.b	$5F,$97,$93,$FE,$EE,$7F,$A3,$FC,$9B,$E6,$33,$3F,$CC,$E6,$F4,$3B
	.dc.b	$BF,$45,$7B,$AF,$DC,$EB,$FB,$C2,$3F,$43,$A7,$E5,$9F,$D3,$37,$34
	.dc.b	$91,$FD,$E8,$74,$A7,$71,$98,$C4,$F0,$8F,$F0,$0C,$61,$FF,$29,$7C
	.dc.b	$03,$E6,$19,$FF,$A9,$3E,$5B,$EF,$57,$FA,$9A,$9F,$25,$FE,$21,$9C
	.dc.b	$CF,$FA,$3A,$31,$CF,$FE,$25,$AE,$97,$CE,$31,$87,$6B,$FD,$84,$6E
	.dc.b	$FF,$C5,$9D,$FD,$30,$EF,$5F,$EB,$68,$C5,$F1,$6E,$1D,$67,$CD,$7E
	.dc.b	$26,$99,$3F,$D4,$75,$E1,$C7,$8C,$1F,$04,$F1,$8F,$61,$C7,$0F,$5B
	.dc.b	$BC,$7D,$78,$79,$5F,$54,$53,$C2,$7B,$8F,$95,$87,$A6,$38,$C6,$4E
	.dc.b	$E5,$AF,$82,$D9,$E0,$F2,$AF,$25,$F9,$EB,$78,$7B,$0F,$6E,$3E,$CB
	.dc.b	$F2,$1D,$9C,$E3,$E2,$75,$1E,$1B,$FA,$73,$93,$CF,$5A,$23,$83,$53
	.dc.b	$F0,$2B,$99,$D9,$F4,$1D,$4F,$1D,$FB,$A3,$51,$F5,$B0,$F6,$B0,$FA
	.dc.b	$07,$2B,$DD,$71,$BF,$A3,$3F,$1E,$BA,$E3,$F2,$23,$E0,$4E,$B0,$E3
	.dc.b	$3F,$83,$8E,$38,$7B,$2B,$B4,$7B,$1E,$27,$B5,$C6,$1F,$84,$BF,$7C
	.dc.b	$75,$33,$7D,$6B,$F0,$77,$D5,$F7,$1E,$C2,$D7,$27,$B1,$F2,$1F,$7E
	.dc.b	$3F,$52,$32,$70,$BF,$A4,$F6,$7D,$66,$1D,$1C,$30,$F9,$0F,$58,$D4
	.dc.b	$7E,$29,$E3,$1F,$08,$87,$E5,$AF,$D5,$3F,$A4,$7F,$21,$3C,$27,$4F
	.dc.b	$4C,$3E,$14,$60,$F3,$1D,$2B,$94,$F1,$7C,$51,$99,$E3,$18,$8E,$47
	.dc.b	$CA,$AF,$8A,$FE,$32,$FE,$33,$F0,$0D,$5F,$58,$FD,$45,$F4,$3A,$BF
	.dc.b	$AB,$33,$8F,$F0,$18,$FE,$31,$5F,$EA,$AC,$3F,$81,$B6,$32,$FE,$F0
	.dc.b	$D7,$CC,$7F,$BA,$9F,$E7,$12,$8C,$46,$BD,$DF,$E6,$71,$88,$FE,$7D
	.dc.b	$F3,$1F,$ED,$A4,$D2,$3F,$ED,$E7,$F7,$AB,$E0,$3F,$A9,$3F,$C3,$41
	.dc.b	$C8,$C3,$AD,$7F,$9D,$C6,$A7,$9C,$E7,$4F,$1C,$F0,$E3,$EE,$F3,$BE
	.dc.b	$35,$B8,$FF,$67,$DF,$2A,$DB,$63,$C8,$71,$BF,$F6,$AD,$71,$8F,$F7
	.dc.b	$74,$EE,$FC,$B1,$B3,$FD,$E5,$3B,$7D,$07,$94,$EB,$3C,$6B,$D3,$5E
	.dc.b	$6B,$F1,$D6,$E3,$AD,$6E,$3D,$26,$6B,$89,$E6,$AE,$35,$CE,$7B,$9B
	.dc.b	$38,$B9,$3A,$B9,$3D,$A5,$3D,$2E,$E7,$BC,$7B,$A7,$BC,$78,$63,$BC
	.dc.b	$ED,$EF,$19,$8E,$E6,$63,$BB,$5E,$F1,$B3,$BB,$F8,$6B,$F9,$8F,$CF
	.dc.b	$31,$5E,$27,$C8,$78,$A7,$93,$E2,$33,$5E,$28,$CC,$78,$9A,$F8,$8D
	.dc.b	$BE,$27,$F3,$53,$DD,$F0,$9D,$CD,$9D,$CC,$CF,$77,$27,$72,$9D,$CD
	.dc.b	$CF,$7A,$F7,$DF,$E0,$67,$81,$A3,$B1,$E4,$3C,$46,$4D,$99,$AF,$13
	.dc.b	$93,$C0,$66,$BC,$17,$ED,$8C,$1D,$AF,$C6,$FE,$14,$FA,$46,$8E,$EE
	.dc.b	$AB,$EE,$46,$A3,$15,$F3,$A3,$15,$DE,$73,$1F,$10,$C9,$EC,$3B,$1E
	.dc.b	$87,$F7,$9B,$FF,$93,$FC,$5F,$E1,$B3,$FC,$30,$FF,$29,$F1,$23,$31
	.dc.b	$FE,$C3,$8C,$5F,$F9,$55,$B1,$FE,$54,$E4,$7C,$3F,$08,$FF,$2E,$C6
	.dc.b	$2B,$FC,$B8,$D8,$E1,$3B,$C1,$39,$9F,$24,$62,$75,$39,$9F,$F3,$AC
	.dc.b	$71,$EA,$7A,$B4,$98,$7F,$9A,$E3,$2D,$8C,$D7,$F9,$B2,$72,$D8,$F5
	.dc.b	$76,$FF,$A7,$3E,$62,$F7,$9A,$4C,$3F,$CF,$93,$B6,$CE,$DB,$3B,$BF
	.dc.b	$F3,$DE,$9F,$EA,$79,$F9,$9D,$FF,$0E,$31,$3A,$8C,$7C,$1C,$AF,$A4
	.dc.b	$D9,$E6,$D7,$E4,$D2,$31,$1F,$EF,$DF,$0F,$FD,$85,$5F,$E8,$F6,$C7
	.dc.b	$FA,$39,$AF,$D2,$6B,$C7,$8B,$EC,$46,$23,$FD,$36,$DE,$03,$98,$FE
	.dc.b	$B3,$F2,$3F,$85,$6D,$FF,$A8,$DA,$58,$D2,$69,$23,$84,$65,$FC,$A7
	.dc.b	$CA,$FC,$92,$38,$3F,$05,$D9,$F8,$72,$38,$24,$FE,$0E,$32,$D8,$FF
	.dc.b	$5A,$7C,$AF,$FB,$25,$BE,$17,$E6,$63,$13,$FE,$B6,$6B,$F9,$A6,$BF
	.dc.b	$DE,$B2,$46,$23,$FD,$86,$D8,$CC,$FE,$4B,$F6,$47,$D0,$D2,$61,$F9
	.dc.b	$08,$CB,$63,$38,$7E,$F4,$71,$AB,$71,$98,$C4,$FA,$F3,$97,$F0,$1B
	.dc.b	$5F,$ED,$2F,$49,$F2,$7F,$B9,$6D,$7F,$B6,$5B,$3B,$9D,$12,$74,$D8
	.dc.b	$FF,$6E,$FA,$BF,$6D,$BB,$FF,$05,$8C,$47,$F0,$46,$BF,$2D,$AF,$FB
	.dc.b	$B1,$BA,$9A,$6B,$D9,$74,$9F,$E3,$06,$0A,$E7,$1F,$F7,$C4,$65,$FF
	.dc.b	$7B,$7B,$1E,$EC,$3F,$8B,$B5,$FB,$CD,$7F,$DF,$ED,$9F,$27,$CA,$78
	.dc.b	$30,$34,$97,$E0,$8C,$B6,$33,$7F,$76,$B9,$CF,$9B,$F0,$A7,$CE,$FB
	.dc.b	$1B,$19,$8C,$1F,$CC,$7D,$FF,$36,$86,$93,$6F,$8B,$FB,$6B,$D3,$A4
	.dc.b	$C3,$E2,$C6,$5F,$C8,$66,$C1,$F0,$4F,$6F,$FA,$03,$6B,$73,$A9,$C9
	.dc.b	$23,$33,$BA,$FC,$E5,$6E,$73,$5B,$9C,$D6,$DF,$E9,$6D,$7F,$64,$D5
	.dc.b	$F4,$FE,$7F,$13,$FA,$D3,$6F,$D6,$D9,$FD,$63,$5F,$EA,$FF,$0A,$F9
	.dc.b	$6F,$EB,$8D,$AF,$6B,$67,$D2,$D7,$F6,$DF,$AC,$F8,$D3,$49,$97,$F2
	.dc.b	$96,$D7,$F2,$96,$CF,$EE,$FE,$55,$F2,$FF,$D6,$9B,$5F,$D5,$9B,$3B
	.dc.b	$9F,$E6,$9F,$AC,$BE,$7E,$92,$FF,$9B,$46,$5B,$19,$AF,$BF,$39,$6C
	.dc.b	$7F,$80,$FD,$57,$CA,$FE,$CE,$FF,$0B,$18,$EA,$D4,$E6,$FF,$8C,$4E
	.dc.b	$6B,$73,$99,$FF,$0C,$4E,$5D,$C6,$5F,$F5,$3C,$65,$FE,$45,$FA,$EF
	.dc.b	$CF,$7F,$C4,$0D,$7F,$D5,$AD,$4D,$26,$FF,$D6,$5B,$7F,$9E,$F4,$5F
	.dc.b	$32,$F7,$7F,$C4,$51,$86,$C6,$67,$83,$70,$FD,$3F,$E1,$8F,$A3,$FE
	.dc.b	$CC,$DA,$FE,$CC,$D9,$DC,$EB,$EB,$A5,$69,$B3,$FB,$2F,$9A,$FD,$C6
	.dc.b	$5F,$DB,$DB,$19,$EC,$FF,$68,$36,$75,$F1,$1F,$C9,$63,$93,$5F,$E9
	.dc.b	$94,$8C,$47,$F9,$2E,$31,$E1,$7F,$7D,$FC,$29,$F1,$BF,$DE,$1B,$59
	.dc.b	$8C,$19,$F9,$C9,$1C,$5A,$FF,$0C,$87,$C3,$7F,$80,$9B,$1F,$E0,$B6
	.dc.b	$BB,$8F,$F0,$3F,$E8,$CF,$36,$92,$7F,$C2,$51,$96,$C6,$5F,$E3,$71
	.dc.b	$96,$C7,$F8,$4B,$E6,$5F,$35,$49,$3F,$E1,$D8,$CB,$63,$2F,$F2,$68
	.dc.b	$CB,$63,$FC,$3B,$FA,$53,$E8,$69,$27,$FC,$51,$19,$6C,$65,$FE,$63
	.dc.b	$19,$6C,$7F,$8A,$3E,$65,$F2,$34,$93,$FE,$33,$8C,$BF,$E3,$5E,$3F
	.dc.b	$F0,$8B,$7D,$4F,$F1,$8C,$62,$3F,$9F,$B5,$FF,$08,$35,$FF,$47,$35
	.dc.b	$1F,$D2,$E3,$99,$FE,$43,$F2,$9E,$3F,$E3,$67,$F9,$4F,$2E,$63,$11
	.dc.b	$FC,$AA,$3F,$C9,$5E,$43,$9D,$F5,$98,$6D,$7F,$3A,$6B,$FC,$E9,$AF
	.dc.b	$F4,$6F,$82,$F1,$79,$9F,$E8,$C8,$FE,$A3,$18,$23,$82,$1F,$D4,$1C
	.dc.b	$B4,$C9,$D4,$FF,$05,$FE,$DA,$F8,$EF,$F3,$17,$5A,$31,$3A,$9C,$C6
	.dc.b	$23,$FB,$0F,$A3,$FE,$B7,$7F,$B4,$B5,$CF,$C3,$7C,$B8,$C4,$7F,$AF
	.dc.b	$ED,$F8,$93,$E6,$3C,$2D,$9F,$F0,$FB,$5F,$F0,$73,$5E,$5F,$32,$F9
	.dc.b	$AF,$4B,$63,$FC,$4E,$D7,$FC,$2C,$D7,$8C,$70,$23,$C1,$FC,$17,$B5
	.dc.b	$E8,$8F,$F1,$5E,$58,$23,$FE,$34,$FD,$7B,$F0,$6F,$B5,$AF,$E3,$B2
	.dc.b	$46,$23,$FC,$80,$DE,$5F,$F2,$37,$0F,$E4,$7C,$FF,$E4,$75,$FE,$54
	.dc.b	$8C,$11,$FF,$0B,$C6,$9F,$F2,$A4,$65,$FF,$33,$7C,$37,$E6,$FD,$FF
	.dc.b	$F5,$94,$7F,$99,$63,$04,$70,$43,$89,$FE,$24,$F9,$9F,$F1,$43,$7D
	.dc.b	$8F,$D3,$46,$23,$FC,$46,$D7,$FD,$F9,$F0,$2F,$D0,$DF,$2D,$FE,$AF
	.dc.b	$8C,$36,$33,$3F,$ED,$A6,$BF,$E8,$56,$BD,$1A,$FE,$EA,$79,$BC,$67
	.dc.b	$CB,$F3,$72,$3D,$4D,$7C,$ED,$7F,$DF,$F6,$8C,$47,$FA,$9F,$E0,$DF
	.dc.b	$21,$FC,$15,$B1,$FE,$FD,$8C,$11,$C1,$0F,$F7,$F3,$FC,$37,$E6,$FF
	.dc.b	$8E,$C7,$F9,$29,$AF,$F1,$43,$FC,$B0,$E1,$FF,$5C,$35,$FE,$33,$F2
	.dc.b	$BF,$C1,$5F,$F2,$FB,$5F,$F5,$83,$5E,$8D,$78,$F6,$FF,$B6,$A3,$10
	.dc.b	$69,$BE,$2F,$F7,$5F,$6F,$F2,$5F,$8C,$FE,$05,$5F,$EF,$58,$C1,$1F
	.dc.b	$F7,$E3,$5F,$F7,$E3,$FC,$B7,$E3,$BF,$DD,$91,$FE,$70,$6B,$FE,$6C
	.dc.b	$6B,$F0,$E9,$18,$8F,$F6,$D3,$5F,$F3,$F3,$63,$7F,$13,$DC,$F0,$54
	.dc.b	$AF,$03,$5F,$F6,$97,$C8,$BE,$43,$F2,$CD,$8F,$F4,$5F,$12,$31,$1C
	.dc.b	$27,$64,$77,$1A,$7F,$46,$D7,$F1,$9E,$E7,$E2,$C8,$FF,$A4,$B8,$3F
	.dc.b	$C7,$4D,$8C,$BF,$E9,$08,$CB,$63,$FA,$C7,$C9,$3E,$3A,$92,$7F,$D3
	.dc.b	$B1,$96,$C6,$5F,$F4,$DC,$65,$B1,$F5,$FE,$62,$47,$D7,$6B,$F9,$0F
	.dc.b	$97,$7C,$1F,$F6,$CE,$57,$C9,$FE,$43,$E4,$FF,$39,$18,$BF,$CC,$C6
	.dc.b	$27,$F2,$B7,$CA,$7F,$5F,$18,$BE,$8F,$38,$A7,$22,$9F,$99,$9E,$57
	.dc.b	$DE,$31,$5A,$E5,$61,$E1,$9E,$47,$80,$FD,$6D,$F2,$C3,$C5,$19,$3F
	.dc.b	$4D,$3C,$5A,$75,$9E,$0F,$2B,$F0,$4E,$EF,$C3,$3B,$98,$F5,$8C,$B6
	.dc.b	$78,$57,$48,$F5,$99,$3A,$9C,$B0,$F8,$14,$9E,$0F,$48,$CD,$74,$BE
	.dc.b	$18,$7C,$DC,$3A,$1E,$3C,$38,$9C,$AB,$D4,$71,$8D,$57,$89,$E3,$7E
	.dc.b	$2A,$E3,$39,$8F,$0C,$F8,$0F,$8C,$70,$BF,$14,$71,$9D,$4F,$CE,$34
	.dc.b	$70,$8D,$38,$34,$E0,$93,$D0,$91,$C5,$FB,$B1,$E4,$8E,$58,$7C,$9A
	.dc.b	$F0,$BF,$26,$B8,$CF,$D2,$AE,$AF,$E2,$9E,$46,$1D,$47,$09,$E5,$1E
	.dc.b	$12,$1E,$88,$D5,$FC,$88,$F1,$E1,$F3,$EF,$A5,$F2,$9E,$07,$27,$D0
	.dc.b	$F0,$3D,$35,$F6,$5F,$A5,$7F,$00,$8F,$A5,$F0,$E1,$F0,$30,$FB,$F3
	.dc.b	$CA,$38,$BB,$3C,$15,$D5,$F9,$2F,$A7,$0F,$87,$7D,$27,$F0,$E7,$94
	.dc.b	$B1,$EC,$BE,$F1,$89,$F2,$BE,$E2,$61,$EE,$AE,$11,$E4,$3E,$79,$F8
	.dc.b	$C3,$FB,$1D,$FF,$0D,$6E,$1F,$DB,$DF,$DE,$9F,$DA,$3E,$59,$F2,$F9
	.dc.b	$8F,$ED,$F1,$87,$FB,$7B,$5E,$11,$B7,$FB,$D7,$CB,$D7,$FD,$1E,$D4
	.dc.b	$D2,$2F,$98,$FE,$30,$FF,$A2,$FE,$13,$FD,$2F,$18,$8E,$26,$F4,$F5
	.dc.b	$78,$8A,$DC,$7A,$6E,$D7,$F8,$36,$B4,$F1,$3E,$14,$69,$FF,$52,$C6
	.dc.b	$AF,$C1,$3F,$CB,$A7,$C8,$7F,$23,$3D,$A5,$9E,$BA,$70,$8E,$C7,$4C
	.dc.b	$3B,$3C,$2F,$DF,$5D,$F0,$F4,$4F,$03,$BC,$F3,$37,$5C,$A3,$E1,$C6
	.dc.b	$1E,$06,$27,$E2,$CE,$F0,$F0,$DF,$7A,$F8,$27,$77,$0F,$DA,$31,$3C
	.dc.b	$5A,$76,$9E,$71,$9B,$F1,$1F,$07,$0F,$21,$70,$F1,$DF,$C9,$3C,$C7
	.dc.b	$D5,$AE,$51,$AA,$E1,$39,$38,$5F,$D3,$C3,$E7,$D7,$C8,$C3,$BB,$4E
	.dc.b	$EF,$C9,$7A,$1F,$3A,$FC,$EE,$EF,$B4,$F8,$1F,$B0,$78,$0F,$BD,$87
	.dc.b	$E0,$DF,$C9,$C7,$81,$F9,$75,$D1,$B5,$E0,$7E,$7C,$ED,$F9,$CD,$7F
	.dc.b	$16,$D7,$AC,$7A,$8F,$41,$CD,$FA,$D3,$DD,$E7,$1B,$7B,$B5,$C9,$EE
	.dc.b	$7E,$03,$EE,$9F,$8C,$FC,$F9,$AF,$C2,$7E,$23,$FA,$28,$B7,$F4,$A3
	.dc.b	$B9,$D4,$A7,$5A,$F4,$BD,$5A,$F5,$7C,$6F,$56,$C7,$58,$F4,$9C,$CE
	.dc.b	$A6,$CF,$9E,$75,$29,$D4,$A7,$EB,$23,$C2,$D3,$9C,$77,$7E,$9C,$7C
	.dc.b	$97,$F7,$0F,$D1,$69,$DC,$F2,$9E,$32,$C7,$37,$ED,$9F,$14,$CD,$FE
	.dc.b	$2F,$0F,$D5,$C7,$EB,$4F,$D7,$E1,$F4,$9F,$31,$F8,$0B,$F1,$3F,$AA
	.dc.b	$3D,$C6,$4F,$21,$E6,$32,$53,$99,$6F,$F8,$0B,$6F,$F6,$6F,$23,$B1
	.dc.b	$C0,$B1,$C1,$B1,$C1,$EE,$60,$F6,$57,$E5,$DE,$C6,$2B,$D7,$7F,$0A
	.dc.b	$7E,$05,$F5,$8C,$4F,$D8,$8C,$57,$C6,$8F,$F5,$2F,$91,$FC,$FA,$3F
	.dc.b	$D5,$5D,$68,$C4,$6B,$1F,$F5,$94,$62,$3F,$D5,$6D,$EB,$FF,$58,$B6
	.dc.b	$B9,$57,$03,$FB,$34,$BE,$6F,$FA,$94,$2F,$FD,$0A,$D9,$D3,$FD,$4B
	.dc.b	$7F,$EA,$71,$DA,$38,$46,$2B,$9D,$72,$14,$F9,$B7,$CB,$F3,$C3,$B4
	.dc.b	$61,$E6,$7F,$BE,$FC,$0D,$5F,$29,$D3,$D6,$FA,$5E,$9B,$7C,$A7,$C3
	.dc.b	$D4,$C3,$58,$72,$CB,$D6,$BE,$6B,$27,$3D,$1E,$46,$32,$E9,$9F,$AE
	.dc.b	$B1,$9F,$8E,$F7,$5A,$C3,$A3,$C2,$1F,$03,$F5,$49,$1F,$EE,$D6,$D7
	.dc.b	$FB,$B5,$AF,$E1,$1F,$BD,$4F,$8F,$F5,$9E,$78,$B5,$E1,$7D,$E7,$C5
	.dc.b	$F9,$AF,$F6,$84,$64,$97,$F2,$EF,$CB,$1E,$E9,$DC,$E6,$BE,$4F,$D4
	.dc.b	$CA,$F8,$7E,$6B,$E6,$31,$7C,$EA,$EF,$F8,$8A,$D5,$FC,$59,$FB,$1E
	.dc.b	$97,$C8,$BF,$F1,$F6,$E8,$C4,$7E,$54,$FC,$5B,$87,$F2,$CD,$5E,$93
	.dc.b	$E7,$9E,$57,$E2,$A3,$0F,$83,$8F,$13,$AC,$BE,$EA,$F0,$FF,$18,$D8
	.dc.b	$FC,$65,$F2,$77,$3A,$23,$A9,$E9,$87,$E5,$DE,$4F,$E6,$7D,$8F,$46
	.dc.b	$66,$93,$1F,$BF,$F6,$BF,$98,$3F,$3D,$1F,$1F,$7F,$CC,$C7,$E8,$7D
	.dc.b	$B7,$D7,$FF,$03,$6E,$7F,$C0,$1B,$97,$CE,$6B,$F3,$72,$FD,$B7,$17
	.dc.b	$E9,$8E,$97,$6F,$F3,$91,$6B,$F6,$57,$D3,$1F,$D5,$CE,$5B,$5F,$A8
	.dc.b	$9D,$3F,$C0,$27,$F7,$1F,$00,$FA,$2F,$E3,$4D,$9F,$E2,$ED,$C3,$F0
	.dc.b	$AD,$7D,$1C,$EF,$9B,$FD,$E9,$D2,$ED,$FF,$03,$F8,$17,$DC,$BD,$6F
	.dc.b	$E5,$51,$86,$C6,$67,$FD,$19,$39,$6C,$67,$1F,$F4,$A4,$E5,$B1,$97
	.dc.b	$F3,$11,$97,$CF,$C5,$FC,$48,$FE,$3B,$1C,$76,$FE,$27,$1F,$C8,$3D
	.dc.b	$37,$CD,$BE,$6F,$FA,$04,$61,$B1,$9C,$BF,$D4,$53,$96,$C6,$71,$FF
	.dc.b	$55,$4E,$5B,$19,$7F,$5D,$19,$7F,$61,$F4,$8F,$AA,$F9,$D7,$FC,$B6
	.dc.b	$31,$BF,$9F,$0E,$D9,$FF,$A6,$27,$35,$BB,$E3,$87,$F4,$59,$CB,$FD
	.dc.b	$1F,$1F,$25,$F1,$AD,$BF,$D4,$7E,$71,$F0,$1F,$D8,$61,$18,$C7,$FD
	.dc.b	$85,$18,$9D,$4E,$70,$FF,$65,$4E,$6B,$73,$99,$FE,$1D,$39,$77,$19
	.dc.b	$7F,$B3,$60,$F4,$DF,$05,$F5,$DB,$87,$F5,$1C,$39,$1F,$EB,$AA,$E1
	.dc.b	$5F,$D7,$3D,$2F,$4F,$77,$F7,$B6,$DF,$FB,$6D,$B8,$FF,$B8,$5A,$FF
	.dc.b	$23,$7E,$0F,$CA,$71,$CC,$D2,$57,$FB,$7F,$E5,$9F,$18,$F4,$B0,$DC
	.dc.b	$7F,$DD,$53,$A9,$C9,$C7,$E2,$FF,$96,$CE,$9A,$FF,$BD,$65,$18,$8F
	.dc.b	$E7,$11,$8F,$13,$D9,$F2,$4F,$95,$FF,$79,$6A,$8C,$5E,$A7,$31,$88
	.dc.b	$F8,$4B,$DC,$C9,$23,$AB,$FE,$FD,$E2,$7D,$3B,$E4,$7E,$EE,$FF,$E1
	.dc.b	$88,$C5,$FD,$76,$BF,$E1,$93,$C9,$F2,$9C,$1F,$F1,$36,$28,$FE,$A9
	.dc.b	$12,$30,$E0,$DC,$56,$96,$36,$ED,$AF,$38,$E0,$F8,$8E,$9D,$2F,$A7
	.dc.b	$7D,$37,$96,$30,$D8,$CC,$FF,$88,$C7,$CA,$7B,$B4,$FD,$B4,$EA,$7F
	.dc.b	$68,$D8,$CD,$7C,$19,$E3,$A5,$BF,$8F,$1F,$09,$AF,$F7,$76,$BF,$E4
	.dc.b	$D6,$BC,$8F,$88,$E0,$91,$F1,$FC,$AF,$F2,$3F,$CD,$7D,$16,$D7,$F8
	.dc.b	$CB,$DA,$4F,$97,$E0,$9F,$4F,$4F,$92,$E2,$BE,$93,$5F,$AC,$D7,$ED
	.dc.b	$B5,$FF,$08,$35,$FF,$33,$35,$39,$BF,$F2,$F7,$17,$9F,$CA,$FC,$B5
	.dc.b	$7F,$90,$E3,$D5,$E5,$BE,$4E,$9E,$F7,$13,$FD,$82,$30,$7C,$3A,$7C
	.dc.b	$4F,$52,$4F,$F6,$DF,$86,$AF,$F8,$EE,$2F,$0F,$F9,$E3,$D0,$56,$3A
	.dc.b	$3F,$CE,$35,$6F,$FC,$EF,$5A,$7A,$1F,$E0,$9A,$FA,$91,$98,$FF,$05
	.dc.b	$47,$97,$C0,$F8,$04,$8F,$F3,$3F,$A7,$FC,$51,$B1,$F7,$63,$11,$FE
	.dc.b	$40,$6B,$FE,$13,$6B,$FE,$93,$F5,$1F,$05,$F8,$36,$C7,$F8,$E2,$3F
	.dc.b	$6D,$DE,$FC,$0D,$F2,$FB,$99,$5B,$3F,$C7,$73,$A6,$CF,$28,$C4,$7F
	.dc.b	$A6,$EB,$A4,$FE,$4F,$D9,$F4,$EF,$FE,$A4,$7C,$F5,$FE,$56,$CF,$F3
	.dc.b	$51,$8E,$4F,$F2,$E4,$7D,$BF,$07,$F0,$C4,$AF,$F3,$5F,$A3,$5F,$F3
	.dc.b	$A4,$5C,$3F,$CD,$53,$F8,$DF,$84,$95,$FE,$8F,$7F,$4D,$F1,$FF,$E1
	.dc.b	$D8,$C1,$0F,$F3,$F4,$6B,$D6,$FC,$B7,$99,$27,$FD,$30,$FE,$BF,$E3
	.dc.b	$FF,$83,$73,$FF,$B0,$DF,$DD,$EA,$69,$3C,$0F,$DC,$36,$FF,$D9,$ED
	.dc.b	$7F,$CF,$BD,$3F,$61,$FF,$48,$72,$7E,$9C,$FD,$F7,$C1,$7F,$9E,$23
	.dc.b	$FD,$55,$19,$69,$FE,$AB,$F7,$7F,$5D,$F3,$72,$3F,$D6,$0F,$F0,$7F
	.dc.b	$86,$7C,$6E,$67,$FD,$3D,$37,$8F,$FD,$B7,$5A,$9F,$F5,$C7,$C7,$FE
	.dc.b	$E4,$8F,$FA,$E7,$83,$CA,$7F,$15,$CF,$FD,$0D,$1F,$EE,$C6,$BF,$E9
	.dc.b	$A6,$BF,$EF,$C8,$E0,$47,$83,$FC,$FB,$D6,$FE,$17,$87,$FB,$CF,$E2
	.dc.b	$FE,$34,$7E,$BD,$BD,$FF,$EA,$37,$3E,$2A,$B8,$6F,$FE,$A7,$BE,$78
	.dc.b	$AF,$8B,$A3,$FE,$00,$47,$59,$E1,$7C,$E7,$84,$62,$F8,$47,$F8,$87
	.dc.b	$0F,$14,$F7,$6C,$74,$AE,$58,$79,$AF,$8E,$1D,$EB,$C3,$7C,$1F,$F1
	.dc.b	$65,$F0,$C3,$C1,$1E,$23,$67,$29,$E1,$5E,$EC,$3B,$4E,$CF,$31,$B3
	.dc.b	$C3,$7D,$5F,$36,$19,$78,$57,$95,$F1,$C7,$03,$55,$F2,$A3,$75,$CD
	.dc.b	$D9,$CA,$FC,$A7,$9A,$35,$7E,$7C,$3E,$45,$74,$7D,$F7,$EA,$3B,$11
	.dc.b	$C9,$23,$E1,$9D,$CF,$31,$E4,$C0,$BE,$33,$D0,$49,$F9,$AF,$C8,$31
	.dc.b	$1D,$1C,$47,$78,$F9,$67,$73,$B1,$23,$8B,$A7,$8B,$F3,$8E,$15,$EB
	.dc.b	$70,$FB,$E7,$E1,$C6,$9F,$05,$FE,$6A,$BA,$BF,$D4,$5E,$B3,$F4,$1C
	.dc.b	$E3,$F0,$23,$E0,$57,$B8,$E7,$1F,$52,$3E,$3D,$F2,$BE,$95,$E2,$BC
	.dc.b	$63,$FA,$37,$AD,$FC,$C7,$38,$FD,$98,$FA,$78,$7A,$DF,$81,$3C,$1F
	.dc.b	$19,$8B,$F2,$3F,$3D,$F0,$18,$7F,$17,$3F,$1D,$FA,$2F,$57,$E0,$47
	.dc.b	$07,$91,$D4,$87,$D3,$74,$F0,$3E,$8B,$F4,$23,$B6,$1F,$95,$9E,$2F
	.dc.b	$C4,$7A,$9C,$48,$E5,$EC,$70,$C3,$F3,$F8,$FC,$F8,$FC,$3D,$7C,$F8
	.dc.b	$FA,$D8,$7D,$CA,$F9,$D8,$7C,$1C,$3A,$12,$33,$87,$A0,$F9,$F3,$E9
	.dc.b	$3F,$0F,$1A,$8F,$19,$27,$F1,$53,$F8,$28,$FB,$07,$CE,$70,$FE,$44
	.dc.b	$EC,$64,$EE,$43,$04,$8F,$E1,$04,$9F,$DF,$9F,$0A,$3B,$C7,$CD,$78
	.dc.b	$5F,$C5,$C3,$EE,$4F,$9A,$7E,$7B,$F0,$71,$FC,$86,$1E,$6A,$FA,$F8
	.dc.b	$76,$9E,$B8,$7F,$0F,$7F,$81,$E1,$FE,$96,$C3,$FA,$1B,$6B,$F9,$F3
	.dc.b	$67,$FD,$2D,$2F,$8D,$7A,$58,$6D,$7F,$A7,$67,$53,$93,$FD,$2F,$0F
	.dc.b	$8D,$A4,$7F,$D3,$6D,$CB,$FC,$05,$1F,$ED,$3E,$8F,$E9,$AF,$F1,$89
	.dc.b	$BE,$B2,$73,$3D,$4F,$E7,$C7,$12,$99,$AE,$B3,$C6,$33,$5F,$29,$FE
	.dc.b	$0D,$19,$DB,$C7,$1D,$5E,$37,$B8,$FE,$19,$5E,$48,$F1,$B6,$3A,$9E
	.dc.b	$4A,$F0,$5F,$49,$D3,$67,$55,$C9,$A7,$28,$F0,$B7,$6F,$F6,$7D,$F4
	.dc.b	$7F,$D9,$D7,$D1,$B3,$CA,$BB,$1C,$8B,$1C,$AF,$C1,$87,$63,$C8,$58
	.dc.b	$E7,$3C,$5A,$71,$AF,$7C,$ED,$FF,$71,$DF,$C7,$7E,$A4,$74,$9E,$0D
	.dc.b	$8F,$A3,$7B,$7E,$C4,$F5,$BF,$3C,$68,$EF,$1C,$8F,$7D,$78,$63,$EA
	.dc.b	$9F,$CB,$8F,$EC,$38,$7D,$D9,$D9,$C6,$3C,$E7,$C7,$7D,$AF,$7A,$F3
	.dc.b	$3C,$AF,$CD,$87,$B6,$FA,$4F,$2C,$3C,$4F,$E6,$4F,$11,$F0,$B0,$EA
	.dc.b	$F3,$3E,$E1,$DD,$F2,$BE,$F8,$F0,$BC,$4E,$D1,$E9,$7D,$F1,$E0,$7E
	.dc.b	$49,$CC,$F6,$E1,$F6,$5E,$27,$59,$EE,$F8,$E7,$BD,$7B,$23,$AC,$F7
	.dc.b	$7D,$95,$CD,$F9,$ED,$7E,$05,$78,$EF,$E3,$61,$F2,$27,$E1,$E1,$D5
	.dc.b	$DE,$1D,$5F,$49,$F1,$70,$FC,$0B,$63,$E4,$BF,$BF,$8C,$B6,$3D,$6F
	.dc.b	$C7,$3D,$27,$BC,$F8,$6F,$E3,$4E,$73,$FA,$B8,$F8,$A7,$C3,$8E,$71
	.dc.b	$E7,$8F,$E0,$91,$E7,$BF,$6C,$7D,$E3,$D2,$78,$CE,$E7,$D3,$3E,$3C
	.dc.b	$78,$5C,$9E,$D2,$9C,$4D,$9C,$4F,$A2,$F1,$72,$71,$29,$C4,$D9,$C4
	.dc.b	$F2,$CF,$BD,$DB,$5D,$9C,$CA,$73,$38,$DF,$F1,$7B,$FC,$04,$60,$FC
	.dc.b	$D9,$C4,$C3,$F5,$A7,$B4,$62,$BE,$8D,$73,$7D,$91,$E0,$9F,$A9,$D2
	.dc.b	$E1,$75,$FF,$3B,$AF,$F1,$F4,$E6,$31,$1F,$E6,$08,$C6,$1F,$EB,$59
	.dc.b	$CC,$A7,$2F,$FA,$DE,$33,$18,$8E,$B5,$E0,$AF,$F2,$64,$75,$F8,$DF
	.dc.b	$EA,$91,$FE,$6E,$E0,$FE,$9D,$69,$CD,$7F,$AD,$6B,$C9,$7C,$E3,$07
	.dc.b	$8E,$77,$7C,$BA,$5E,$77,$F5,$98,$C3,$70,$F7,$73,$BE,$37,$18,$FB
	.dc.b	$70,$E3,$5E,$53,$E2,$38,$8F,$0E,$68,$C6,$3F,$EE,$4B,$E7,$39,$C6
	.dc.b	$D3,$E3,$69,$AD,$1E,$0E,$33,$F0,$67,$E3,$AC,$3E,$0A,$DC,$EB,$1F
	.dc.b	$15,$73,$F8,$BF,$A7,$46,$92,$BF,$DE,$0D,$BF,$F7,$83,$5F,$C2,$E2
	.dc.b	$F8,$DF,$49,$D6,$BC,$11,$96,$D7,$B8,$F8,$3F,$1A,$F4,$71,$3A,$26
	.dc.b	$1F,$2A,$75,$18,$DF,$ED,$7C,$67,$C8,$3E,$5F,$C6,$BE,$2F,$15,$E2
	.dc.b	$9B,$87,$D4,$AD,$4E,$4F,$8D,$E5,$FC,$89,$FC,$9F,$02,$31,$1F,$A4
	.dc.b	$6B,$FA,$36,$AF,$35,$E7,$BC,$7F,$C4,$C6,$1F,$C0,$D7,$7D,$3F,$C5
	.dc.b	$53,$9C,$7E,$C5,$BC,$4F,$C3,$B6,$3F,$03,$3B,$C7,$F1,$F3,$96,$D7
	.dc.b	$03,$C3,$1E,$3A,$EB,$1F,$92,$F4,$3D,$39,$9A,$4E,$6F,$DC,$F1,$BE
	.dc.b	$37,$F2,$E7,$18,$B8,$7E,$3F,$D2,$7C,$6F,$F0,$06,$C7,$EF,$B4,$FE
	.dc.b	$0D,$18,$7E,$7B,$5F,$C6,$F1,$FE,$90,$FD,$3C,$6F,$7F,$DD,$46,$6B
	.dc.b	$F1,$EE,$AF,$17,$FB,$2C,$BE,$FB,$67,$55,$C0,$FD,$37,$AC,$F8,$DF
	.dc.b	$E2,$ED,$8F,$46,$C8,$C5,$FE,$19,$AF,$E9,$3E,$2B,$F7,$67,$EF,$A3
	.dc.b	$79,$FF,$11,$8F,$E1,$7F,$18,$F8,$D7,$A7,$FC,$AA,$30,$D8,$CE,$5F
	.dc.b	$26,$72,$D8,$CE,$3F,$CA,$A7,$2D,$8C,$BF,$EF,$98,$CB,$FB,$AE,$0F
	.dc.b	$E2,$27,$92,$79,$E3,$FC,$9E,$3F,$90,$FC,$4B,$E3,$1F,$1B,$FD,$02
	.dc.b	$30,$D8,$CE,$5F,$7E,$72,$D8,$CE,$3F,$D0,$27,$2D,$8C,$A2,$46,$9C
	.dc.b	$0F,$83,$7C,$6B,$6F,$E6,$F1,$87,$D3,$7E,$0C,$FF,$D3,$13,$9C,$7F
	.dc.b	$95,$B7,$0F,$E9,$2D,$9F,$F4,$F6,$3E,$67,$8E,$3D,$EF,$F9,$9C,$3E
	.dc.b	$2D,$F1,$BF,$D9,$23,$0D,$8C,$CF,$E8,$E7,$2D,$8C,$E3,$FD,$92,$72
	.dc.b	$D8,$CB,$F0,$A3,$2F,$C4,$C1,$E9,$3E,$37,$EC,$B7,$0F,$3E,$1D,$B0
	.dc.b	$F6,$CE,$70,$FE,$D3,$E9,$FA,$FB,$7F,$BE,$B7,$0F,$DD,$36,$7F,$BC
	.dc.b	$B5,$FA,$AF,$C3,$F7,$B8,$E2,$69,$2B,$F8,$25,$3E,$2D,$F1,$AF,$4B
	.dc.b	$31,$86,$C6,$67,$F8,$64,$E5,$BF,$1B,$FC,$CD,$BE,$27,$28,$C4,$7F
	.dc.b	$84,$25,$18,$8F,$C8,$66,$F8,$C7,$C1,$79,$A3,$0D,$8C,$CF,$C4,$6B
	.dc.b	$F1,$1F,$EB,$CE,$7C,$CE,$04,$7F,$05,$F1,$AF,$82,$FB,$9B,$7F,$88
	.dc.b	$63,$13,$F8,$E6,$BF,$E1,$E3,$CD,$F1,$BC,$1F,$D4,$60,$8F,$9B,$1A
	.dc.b	$3E,$7C,$65,$B1,$A8,$C4,$7F,$37,$6B,$FE,$36,$6B,$FB,$26,$BF,$B3
	.dc.b	$F8,$D7,$C6,$F8,$FA,$D1,$89,$FD,$7B,$5F,$57,$C6,$FB,$3A,$3F,$6B
	.dc.b	$39,$6D,$F3,$8C,$67,$FC,$0A,$B8,$37,$4F,$7C,$7C,$6C,$3F,$C9,$AD
	.dc.b	$74,$7E,$41,$C3,$FC,$3D,$A9,$A6,$BC,$5F,$F2,$4F,$C6,$FC,$F6,$CF
	.dc.b	$F8,$BF,$D8,$4F,$D3,$87,$C1,$3E,$37,$0D,$8F,$D0,$B5,$FE,$CC,$D7
	.dc.b	$FC,$CA,$D7,$F9,$0B,$5F,$E4,$CD,$7B,$C7,$E0,$23,$57,$FE,$53,$E2
	.dc.b	$F4,$79,$9F,$94,$BF,$EC,$D1,$EB,$F2,$9F,$1B,$4F,$DA,$DF,$F6,$16
	.dc.b	$C6,$AB,$FB,$2B,$5F,$F1,$4D,$BE,$37,$B9,$23,$FC,$0D,$F0,$D5,$FE
	.dc.b	$ED,$17,$8F,$FC,$FF,$E8,$AB,$1C,$5F,$E6,$EA,$B7,$FE,$0A,$9F,$A5
	.dc.b	$5B,$8F,$EE,$ED,$7F,$C1,$B1,$EA,$8C,$47,$F9,$6B,$C0,$F8,$64,$7F
	.dc.b	$CD,$1E,$9F,$F1,$76,$C7,$C2,$8C,$47,$F9,$19,$AF,$FA,$A1,$AF,$FA
	.dc.b	$9F,$D4,$7C,$17,$E1,$DB,$1F,$E3,$E8,$FD,$C7,$82,$F8,$D7,$C6,$EE
	.dc.b	$65,$6C,$FF,$20,$CE,$9B,$3C,$A3,$11,$FE,$A2,$AE,$93,$F9,$7F,$63
	.dc.b	$D3,$BF,$FA,$B1,$F3,$D7,$FA,$E5,$BC,$7F,$E6,$1D,$3F,$03,$18,$E1
	.dc.b	$FE,$3B,$7F,$EA,$BF,$07,$F1,$67,$E8,$FD,$AF,$F4,$BE,$3F,$E8,$28
	.dc.b	$CB,$4F,$F4,$A4,$69,$FF,$1B,$7C,$27,$E5,$08,$FF,$A6,$7E,$37,$FC
	.dc.b	$4A,$FF,$A3,$23,$2F,$03,$5E,$BF,$E7,$3C,$C9,$1F,$E9,$27,$F6,$DF
	.dc.b	$1B,$FC,$1A,$3F,$DA,$CF,$EE,$C9,$E0,$FE,$79,$B1,$FE,$40,$6B,$FE
	.dc.b	$84,$7F,$81,$69,$F6,$A3,$FC,$9D,$C7,$FA,$C3,$F8,$27,$C1,$7F,$A0
	.dc.b	$63,$FD,$47,$19,$69,$FE,$B5,$8D,$3F,$E5,$7F,$8C,$FD,$A9,$1F,$F5
	.dc.b	$D7,$C3,$3E,$37,$11,$B3,$FD,$5F,$3F,$C4,$E7,$51,$FE,$B2,$F8,$DF
	.dc.b	$DF,$91,$F0,$69,$E6,$3F,$7B,$87,$FA,$3E,$3F,$CB,$ED,$7F,$D4,$6D
	.dc.b	$7F,$DF,$31,$C0,$8F,$07,$FA,$7F,$A7,$FC,$83,$8F,$FC,$75,$F1,$5F
	.dc.b	$1E,$3F,$68,$DE,$EF,$F1,$DB,$9F,$11,$5C,$36,$FF,$1F,$DF,$3C,$57
	.dc.b	$C5,$D1,$FE,$47,$8E,$B3,$C2,$F9,$CF,$08,$C5,$F0,$8F,$F1,$3E,$1E
	.dc.b	$29,$EE,$D8,$E9,$5C,$B0,$F3,$5F,$1C,$3B,$D7,$86,$F8,$3F,$E3,$2B
	.dc.b	$E1,$87,$82,$3C,$46,$CE,$53,$C2,$BD,$D8,$76,$9D,$9E,$63,$67,$86
	.dc.b	$FA,$BE,$6C,$32,$F0,$AF,$2B,$E3,$8E,$06,$AB,$E5,$46,$EB,$9B,$B3
	.dc.b	$95,$F9,$4F,$34,$6A,$FC,$F8,$7C,$8A,$E8,$FB,$EF,$D4,$76,$23,$92
	.dc.b	$47,$C3,$3B,$9E,$63,$C9,$87,$FA,$0F,$E3,$3C,$E4,$9F,$98,$FC,$73
	.dc.b	$13,$CD,$C4,$76,$8F,$94,$76,$3A,$92,$38,$BA,$78,$BF,$34,$E1,$5E
	.dc.b	$A7,$0F,$BA,$7E,$14,$69,$F0,$5F,$E6,$6B,$AB,$FD,$55,$EB,$3F,$3D
	.dc.b	$CE,$3F,$E9,$E7,$33,$F3,$A7,$E3,$5F,$85,$FA,$CE,$AB,$85,$F8,$AF
	.dc.b	$8E,$3F,$A2,$7A,$DF,$CB,$9F,$92,$79,$AB,$EF,$47,$D1,$BE,$8E,$CF
	.dc.b	$99,$3C,$1F,$89,$5D,$13,$E0,$C7,$F0,$71,$E8,$9F,$87,$87,$8D,$F3
	.dc.b	$12,$39,$BD,$5F,$A4,$E5,$F8,$CF,$9D,$EA,$7C,$27,$86,$3F,$35,$E0
	.dc.b	$FD,$17,$38,$FC,$69,$F5,$9E,$03,$ED,$C7,$DA,$C7,$E7,$57,$D1,$9F
	.dc.b	$09,$C0,$F2,$9C,$4C,$98,$23,$E1,$7A,$9F,$0C,$87,$A4,$E4,$FA,$5E
	.dc.b	$CF,$17,$47,$E6,$A3,$CB,$7F,$B3,$3F,$69,$7F,$6A,$7F,$3B,$3F,$95
	.dc.b	$7B,$D7,$F0,$43,$D8,$49,$E1,$3B,$9E,$F3,$B9,$FE,$03,$87,$E1,$70
	.dc.b	$F8,$C4,$AF,$69,$F0,$AB,$D4,$FC,$2C,$3F,$00,$EA,$3A,$3C,$CC,$5F
	.dc.b	$8A,$BF,$0F,$7F,$EC,$6A,$FE,$85,$43,$F9,$61,$4F,$DB,$14,$FF,$71
	.dc.b	$FC,$6B,$D1,$E4,$D8,$D1,$A8,$C9,$FE,$A8,$5F,$1B,$48,$FF,$AB,$1A
	.dc.b	$FF,$B6,$9A,$F4,$6B,$E3,$6B,$E7,$6B,$EB,$6B,$E0,$3B,$10,$EC,$53
	.dc.b	$B1,$4E,$C5,$3B,$14,$EC,$53,$B1,$D0,$87,$42,$9D,$0A,$74,$29,$D0
	.dc.b	$A7,$43,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70
	.dc.b	$34,$43,$45,$34,$60,$86,$08,$FD,$42,$3F,$25,$AF,$C0,$6B,$E9,$6B
	.dc.b	$E4,$6B,$E1,$6B,$E0,$6B,$E0,$3B,$10,$EC,$53,$B1,$4E,$C5,$3B,$14
	.dc.b	$EC,$53,$B1,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0E,$45
	.dc.b	$39,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60
	.dc.b	$86,$08,$FD,$42,$3F,$25,$AF,$C0,$6B,$E9,$6B,$E4,$6B,$E1,$6B,$E0
	.dc.b	$6B,$E0,$3B,$10,$EC,$53,$B1,$4E,$C5,$3B,$14,$EC,$53,$B1,$D0,$87
	.dc.b	$42,$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0E,$45,$39,$14,$E4,$53,$91
	.dc.b	$FC,$02,$95,$FC,$E7,$EA,$1F,$35,$FE,$86,$8C,$37,$0F,$E8,$D3,$AB
	.dc.b	$FF,$43,$56,$9B,$3A,$AE,$10,$AE,$0F,$F4,$53,$11,$A8,$CD,$FF,$54
	.dc.b	$9F,$EA,$64,$3F,$A9,$94,$FF,$36,$1F,$D5,$5E,$0E,$88,$68,$E0,$E9
	.dc.b	$C1,$A7,$04,$D3,$FD,$3F,$1D,$5A,$F2,$6B,$C5,$AF,$42,$47,$3C,$BF
	.dc.b	$B8,$4E,$A3,$13,$FD,$A1,$AF,$79,$D9,$1D,$4F,$12,$46,$63,$81,$A7
	.dc.b	$0F,$9D,$AF,$9D,$AF,$9A,$39,$BC,$48,$74,$21,$CC,$C9,$4C,$9D,$4C
	.dc.b	$94,$C9,$DE,$3C,$33,$F0,$23,$91,$0E,$05,$38,$1D,$0A,$74,$79,$3C
	.dc.b	$08,$68,$E4,$E9,$C1,$A7,$04,$9F,$87,$1D,$5A,$F2,$6B,$C5,$AF,$42
	.dc.b	$47,$39,$F7,$CE,$A3,$13,$EB,$6B,$DE,$76,$47,$53,$C4,$91,$98,$E0
	.dc.b	$69,$C3,$E7,$6B,$E7,$6B,$E6,$8E,$6F,$12,$1D,$08,$73,$32,$53,$27
	.dc.b	$53,$25,$32,$77,$8F,$0C,$FC,$08,$E4,$43,$81,$4E,$07,$42,$9D,$1E
	.dc.b	$4F,$02,$1A,$39,$3A,$70,$69,$C1,$27,$E1,$C7,$56,$BC,$9A,$F1,$6B
	.dc.b	$D0,$91,$CE,$7D,$F3,$A8,$C4,$FA,$DA,$F7,$9D,$91,$D4,$F1,$24,$66
	.dc.b	$38,$1A,$70,$F9,$DA,$FF,$36,$3F,$88,$63,$FE,$79,$7F,$95,$35,$FE
	.dc.b	$72,$63,$1F,$F3,$DB,$C1,$AE,$4C,$10,$C1,$0E,$64,$3C,$46,$4A,$64
	.dc.b	$FE,$8C,$64,$A6,$6F,$FA,$CC,$78,$A3,$13,$E9,$79,$90,$E2,$59,$E2
	.dc.b	$F9,$CE,$33,$CD,$D4,$F3,$78,$92,$78,$BE,$56,$BC,$8E,$87,$29,$E8
	.dc.b	$F1,$9E,$93,$C2,$34,$70,$35,$3C,$23,$51,$83,$4E,$09,$3D,$09,$3C
	.dc.b	$5B,$3C,$1A,$FA,$5A,$FA,$4C,$47,$12,$3C,$5A,$E8,$C1,$0C,$10,$E8
	.dc.b	$43,$DE,$64,$A6,$4F,$71,$92,$99,$9F,$7C,$78,$A3,$13,$EA,$79,$90
	.dc.b	$E2,$59,$E2,$F9,$CE,$33,$CD,$D4,$F3,$78,$92,$78,$BE,$56,$BC,$8E
	.dc.b	$87,$29,$E8,$F1,$9E,$93,$C2,$34,$70,$35,$3C,$23,$51,$83,$4E,$09
	.dc.b	$3D,$09,$3C,$5B,$3C,$1A,$FA,$5A,$FA,$4C,$47,$12,$3C,$5A,$E8,$C1
	.dc.b	$0C,$10,$E8,$43,$DE,$64,$A6,$4F,$71,$92,$99,$9F,$7C,$78,$A3,$13
	.dc.b	$EA,$79,$90,$E2,$59,$E2,$F9,$CE,$33,$CD,$D4,$F3,$78,$92,$78,$BE
	.dc.b	$56,$BC,$8E,$87,$29,$E8,$F1,$9E,$93,$C2,$34,$70,$35,$3C,$23,$51
	.dc.b	$83,$4E,$09,$3D,$09,$3C,$5B,$3C,$1A,$FA,$5A,$FA,$4C,$47,$12,$3C
	.dc.b	$5A,$E8,$C1,$0C,$10,$E8,$43,$DE,$64,$A6,$4F,$71,$92,$99,$9F,$7C
	.dc.b	$78,$A3,$13,$EA,$79,$90,$E2,$59,$E2,$F9,$CE,$33,$CD,$D4,$F3,$78
	.dc.b	$92,$78,$BE,$56,$BC,$8E,$87,$29,$E8,$F1,$9E,$93,$C2,$34,$70,$35
	.dc.b	$3C,$23,$51,$83,$4E,$09,$3D,$09,$3C,$5B,$3C,$1A,$FA,$5A,$FA,$4C
	.dc.b	$47,$CD,$F8,$F1,$00

;--------------------------------------------------------------------------------
;	FSGLMUL.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsglmul88x_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#11,-(sp)
	peamsg	'FSGLMUL.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSGLMUL.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsglmul88x_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsglmul.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSGLMUL.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsglmul.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSGLMUL.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsglmul88x_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3A,$F9,$AF,$EC,$ED,$2D,$FC,$E8,$CC,$76,$B4,$62
	.dc.b	$75,$2F,$A0,$F5,$BE,$9C,$FF,$BA,$46,$23,$CF,$18,$CF,$D7,$87,$49
	.dc.b	$E5,$18,$C3,$FB,$ED,$F1,$C3,$DD,$5A,$3B,$1A,$AE,$0E,$4F,$45,$7F
	.dc.b	$81,$BE,$7F,$96,$3F,$E0,$76,$CE,$6B,$8C,$77,$C7,$C9,$8E,$AB,$A6
	.dc.b	$3E,$28,$D6,$3E,$19,$F9,$51,$B2,$61,$E4,$7E,$2F,$CA,$3E,$17,$F5
	.dc.b	$75,$FE,$1A,$BF,$E0,$D3,$9F,$17,$FC,$2B,$F0,$9F,$E1,$E8,$C5,$6E
	.dc.b	$31,$7F,$E0,$E8,$C0,$9D,$BF,$62,$77,$5D,$23,$55,$E2,$74,$63,$E0
	.dc.b	$75,$7E,$7F,$81,$B4,$69,$8F,$4E,$DF,$8A,$B8,$36,$B8,$36,$7C,$5F
	.dc.b	$37,$FC,$13,$C0,$FC,$7B,$67,$F2,$AD,$7F,$C7,$AD,$7F,$C7,$FC,$DF
	.dc.b	$DB,$F5,$FD,$C3,$75,$FD,$9F,$37,$BE,$93,$FA,$9A,$FE,$69,$F3,$8F
	.dc.b	$C5,$BF,$3F,$48,$C3,$FA,$26,$BF,$A8,$6B,$FE,$50,$6B,$FE,$49,$8E
	.dc.b	$71,$88,$FF,$2B,$B5,$FA,$3E,$B7,$CC,$23,$F2,$7F,$27,$FE,$0D,$7F
	.dc.b	$D2,$73,$FE,$53,$CD,$F1,$BC,$CF,$E4,$77,$F1,$9A,$FC,$66,$BF,$E5
	.dc.b	$06,$BF,$E6,$89,$E6,$F1,$9E,$87,$F0,$0E,$1F,$CB,$7A,$FF,$80,$24
	.dc.b	$FD,$BF,$07,$F1,$58,$BF,$6F,$FF,$2B,$4F,$F0,$C8,$C7,$81,$FC,$29
	.dc.b	$B1,$99,$DB,$FE,$78,$8C,$B6,$32,$FF,$9D,$9A,$FE,$9D,$AF,$E7,$F4
	.dc.b	$7B,$DF,$98,$6C,$7C,$B8,$F3,$CB,$E1,$BF,$09,$E2,$BC,$3F,$C0,$B6
	.dc.b	$FF,$90,$B5,$FE,$4E,$D7,$FD,$1E,$D7,$FA,$11,$C2,$39,$9F,$2F,$A7
	.dc.b	$FA,$A4,$7E,$8F,$CB,$FD,$59,$F9,$5F,$BB,$7C,$87,$93,$97,$C6,$4A
	.dc.b	$F1,$B5,$F1,$B5,$F1,$BD,$FC,$8F,$DB,$92,$3F,$82,$FC,$DB,$F4,$DF
	.dc.b	$EB,$DC,$3F,$D7,$FD,$0F,$E0,$55,$FE,$AA,$6B,$FE,$AD,$6B,$FD,$BB
	.dc.b	$C3,$FD,$3C,$62,$0D,$24,$7C,$1F,$2F,$F8,$49,$27,$F8,$BF,$53,$F2
	.dc.b	$07,$B3,$FC,$3B,$92,$57,$F0,$97,$F8,$8F,$C5,$FE,$EF,$27,$C6,$78
	.dc.b	$89,$5F,$D6,$23,$2F,$F5,$EE,$DF,$E6,$1D,$7F,$C6,$89,$1F,$C5,$79
	.dc.b	$FF,$83,$7D,$9F,$F8,$6E,$3F,$B2,$B5,$FF,$66,$B5,$FF,$68,$35,$FF
	.dc.b	$69,$B5,$34,$9E,$87,$E0,$FE,$0B,$F9,$59,$27,$F9,$C9,$FE,$1E,$F2
	.dc.b	$FF,$98,$FF,$00,$7F,$A3,$63,$F9,$62,$F4,$32,$6B,$E6,$E4,$EA,$32
	.dc.b	$F0,$8E,$7F,$07,$D3,$E0,$E4,$F4,$8C,$BD,$A3,$2F,$F4,$4D,$98,$7F
	.dc.b	$93,$63,$CF,$F6,$7F,$E3,$D7,$FD,$E2,$D7,$FD,$E8,$D7,$FD,$EE,$D4
	.dc.b	$D2,$7C,$CF,$F3,$27,$5E,$97,$F3,$7F,$43,$FB,$15,$FF,$89,$2F,$C8
	.dc.b	$DE,$CF,$F1,$C5,$6E,$7F,$B1,$4F,$F4,$36,$FA,$1F,$30,$F9,$BF,$21
	.dc.b	$63,$E6,$C5,$7A,$C7,$D2,$F5,$FE,$9F,$9D,$27,$E9,$C6,$5F,$AB,$19
	.dc.b	$7C,$B1,$FD,$47,$E2,$1E,$2F,$F4,$E6,$CF,$F8,$B9,$AF,$F8,$D1,$A8
	.dc.b	$FF,$41,$9B,$8E,$44,$7F,$98,$7E,$47,$FA,$A7,$FC,$07,$A2,$31,$1F
	.dc.b	$DF,$A3,$1D,$FF,$D9,$7F,$18,$FB,$93,$F3,$3A,$7F,$23,$F3,$B2,$3F
	.dc.b	$23,$1F,$94,$30,$FD,$28,$C3,$F9,$6F,$A1,$FC,$C9,$1F,$F0,$27,$E9
	.dc.b	$7F,$CD,$1F,$B3,$19,$69,$AF,$47,$FB,$C7,$C9,$FF,$9F,$63,$FC,$AE
	.dc.b	$D7,$E2,$1F,$EB,$7F,$9D,$FE,$F3,$19,$8F,$F5,$C4,$60,$8F,$F9,$72
	.dc.b	$34,$FE,$1B,$E2,$BF,$5E,$47,$F6,$7F,$28,$F9,$3F,$D9,$FC,$84,$9F
	.dc.b	$D9,$BF,$E2,$6F,$D3,$FC,$DC,$3F,$E3,$AE,$94,$EE,$BF,$2E,$EF,$3F
	.dc.b	$DE,$FE,$8F,$27,$F7,$AF,$F0,$0F,$47,$11,$F9,$E8,$B8,$75,$9F,$F1
	.dc.b	$4F,$5B,$E5,$7F,$C5,$2D,$BF,$F1,$4B,$5F,$F1,$4B,$5F,$93,$E8,$3F
	.dc.b	$33,$FE,$E5,$E0,$7F,$81,$5B,$29,$D3,$FC,$53,$E3,$3F,$8A,$FC,$A4
	.dc.b	$8F,$E2,$AF,$F1,$8E,$BC,$47,$ED,$A2,$E9,$F1,$67,$FC,$83,$F2,$3F
	.dc.b	$EA,$FA,$F1,$B5,$FF,$0B,$35,$F1,$37,$C6,$FE,$27,$F1,$DF,$C9,$89
	.dc.b	$3F,$C9,$BE,$47,$FC,$BD,$EC,$BF,$61,$FE,$55,$3F,$88,$11,$9F,$37
	.dc.b	$FD,$3E,$D8,$FF,$52,$B5,$E1,$83,$F8,$07,$E2,$B6,$3F,$CC,$F1,$83
	.dc.b	$AB,$FE,$63,$70,$E8,$DC,$D7,$45,$3F,$98,$47,$27,$FC,$EB,$1E,$BF
	.dc.b	$1F,$CF,$87,$FA,$92,$7F,$47,$39,$7D,$31,$97,$FD,$15,$5F,$E4,$CE
	.dc.b	$2B,$F4,$9E,$83,$D3,$E6,$FF,$99,$FE,$83,$FB,$6E,$9F,$C4,$7E,$AB
	.dc.b	$FD,$AB,$8A,$72,$A3,$31,$87,$FD,$B7,$CA,$FB,$3F,$E2,$31,$AF,$6B
	.dc.b	$FA,$E3,$70,$ED,$E9,$7E,$F3,$0E,$4F,$F9,$5E,$F8,$B9,$BE,$2D,$AE
	.dc.b	$38,$FF,$27,$BE,$2F,$E6,$6F,$8B,$9B,$E2,$DA,$E2,$FB,$63,$38,$FA
	.dc.b	$79,$7F,$87,$E5,$E7,$AF,$81,$97,$85,$A7,$32,$D7,$17,$AE,$5E,$47
	.dc.b	$B6,$38,$C7,$BE,$5E,$69,$C9,$D0,$B8,$F8,$4F,$9B,$1B,$BE,$84,$AE
	.dc.b	$79,$7A,$32,$F3,$5F,$16,$9F,$36,$35,$87,$A6,$74,$D9,$D3,$E7,$C6
	.dc.b	$3E,$5C,$3E,$7E,$5E,$F8,$C6,$3F,$39,$F4,$DF,$81,$FA,$98,$F8,$1F
	.dc.b	$79,$FB,$5B,$FC,$6B,$6B,$F7,$2F,$CF,$31,$87,$EE,$DF,$A0,$60,$C4
	.dc.b	$F2,$7C,$07,$88,$F6,$1E,$23,$D7,$8F,$C0,$75,$8F,$AE,$31,$3F,$36
	.dc.b	$BF,$1F,$39,$C7,$AC,$E6,$FF,$03,$97,$D5,$C7,$E1,$4E,$71,$FB,$39
	.dc.b	$7C,$18,$C6,$1F,$17,$1E,$98,$F9,$8F,$41,$E4,$C7,$A6,$5E,$8A,$E1
	.dc.b	$18,$9F,$A3,$39,$7D,$C6,$09,$8F,$DE,$BE,$F1,$DB,$0F,$1B,$C2,$7E
	.dc.b	$0E,$5E,$A3,$F1,$B1,$FB,$0E,$2F,$E4,$9A,$FF,$0B,$8F,$E4,$0F,$F0
	.dc.b	$C7,$FD,$23,$C1,$FC,$4D,$FE,$62,$DE,$1F,$E3,$11,$FC,$D5,$AF,$F7
	.dc.b	$EA,$FE,$5C,$FF,$26,$21,$FC,$9A,$3F,$94,$73,$38,$7F,$8F,$CE,$1F
	.dc.b	$E4,$33,$61,$5D,$1F,$E5,$E7,$F3,$43,$CF,$AF,$F2,$E2,$3F,$CB,$1F
	.dc.b	$F7,$37,$80,$E6,$FE,$A3,$3F,$E5,$47,$FC,$F3,$CD,$FD,$7B,$1F,$CB
	.dc.b	$4F,$F5,$A6,$F0,$39,$FF,$AF,$36,$3A,$F4,$7F,$68,$7F,$CC,$2F,$F6
	.dc.b	$DA,$FE,$DE,$43,$FB,$71,$FE,$86,$E5,$FF,$12,$ED,$F0,$E3,$FB,$D7
	.dc.b	$57,$F7,$CC,$3F,$BD,$C7,$FA,$CA,$3F,$AD,$E8,$EA,$FF,$06,$F6,$7F
	.dc.b	$83,$09,$5F,$E0,$C7,$E0,$F7,$7F,$85,$34,$FF,$3A,$BF,$E1,$99,$FF
	.dc.b	$0D,$90,$FF,$0D,$1F,$E9,$AE,$EF,$F1,$26,$7F,$E2,$7F,$02,$DF,$F9
	.dc.b	$7F,$8D,$E0,$FF,$8B,$23,$FD,$03,$3A,$C3,$FC,$FA,$FF,$8A,$DF,$F5
	.dc.b	$2F,$89,$FE,$39,$7F,$D0,$AF,$C2,$9F,$F2,$3F,$89,$FE,$3B,$9F,$F5
	.dc.b	$7F,$8B,$FE,$42,$4D,$27,$89,$FE,$48,$8F,$F2,$8B,$5F,$F0,$54,$7F
	.dc.b	$92,$B2,$FF,$52,$F8,$FF,$E5,$BD,$3F,$C8,$4F,$86,$BF,$1F,$E4,$7F
	.dc.b	$98,$67,$31,$FE,$67,$21,$FE,$AF,$8D,$65,$FE,$A0,$F2,$BF,$CD,$F5
	.dc.b	$FE,$4D,$F2,$BF,$CD,$F7,$FE,$77,$3F,$D8,$AE,$AB,$84,$7F,$A0,$BC
	.dc.b	$CF,$F4,$0D,$7E,$0C,$FF,$40,$C6,$7A,$3F,$D0,$5E,$9F,$FA,$5F,$CE
	.dc.b	$FF,$46,$4F,$FA,$4D,$AF,$F8,$A7,$CF,$FF,$4A,$BF,$E9,$43,$47,$FA
	.dc.b	$54,$FF,$78,$74,$BC,$BF,$F5,$26,$3F,$E6,$5F,$43,$FD,$51,$5F,$EA
	.dc.b	$63,$FD,$52,$43,$07,$FA,$FB,$CD,$FF,$5A,$4F,$F9,$B5,$FF,$5C,$C7
	.dc.b	$FA,$E0,$86,$CF,$F7,$3F,$A3,$F4,$57,$A5,$FE,$C5,$E4,$FF,$6D,$FC
	.dc.b	$77,$FB,$B8,$D4,$7F,$B2,$67,$FD,$AB,$E7,$FF,$B4,$C8,$FE,$EB,$D4
	.dc.b	$FF,$6B,$B6,$32,$FF,$53,$8F,$F7,$05,$FF,$BA,$38,$3F,$DD,$79,$FF
	.dc.b	$A1,$7D,$0F,$F7,$7D,$FF,$B6,$A7,$3C,$9F,$ED,$E9,$CD,$7F,$A2,$FC
	.dc.b	$9F,$F7,$DD,$FF,$B8,$3D,$5F,$F7,$FB,$FE,$E2,$8C,$CF,$FA,$46,$72
	.dc.b	$FF,$01,$C5,$F0,$AF,$63,$F2,$91,$86,$C6,$70,$FF,$68,$57,$F0,$95
	.dc.b	$C4,$F6,$3E,$67,$B2,$EC,$FE,$3B,$48,$C3,$FC,$81,$AF,$F7,$76,$BF
	.dc.b	$E7,$36,$BF,$EB,$C6,$BF,$EE,$16,$BD,$63,$F9,$5E,$5E,$CF,$6F,$D2
	.dc.b	$49,$F4,$EF,$FC,$2F,$37,$B7,$F8,$3F,$71,$EA,$FA,$1B,$7F,$CD,$9A
	.dc.b	$FF,$83,$5A,$FF,$53,$A7,$BD,$F2,$09,$1E,$03,$FA,$6C,$3D,$EF,$9B
	.dc.b	$F0,$12,$75,$A7,$D4,$8F,$E0,$DE,$8B,$E0,$3F,$6F,$AF,$FB,$7B,$37
	.dc.b	$C0,$F9,$C9,$8F,$9D,$AF,$9D,$AF,$FA,$7E,$38,$BF,$E2,$7C,$7F,$07
	.dc.b	$F0,$5F,$80,$24,$7A,$7D,$97,$A3,$F8,$1E,$7F,$C1,$7C,$37,$F0,$0A
	.dc.b	$07,$0D,$F1,$FF,$07,$18,$23,$82,$47,$8F,$85,$F0,$DF,$92,$24,$7C
	.dc.b	$B1,$F0,$CF,$86,$FD,$E6,$1F,$72,$7F,$B6,$B5,$FF,$22,$35,$FF,$60
	.dc.b	$35,$F4,$9E,$9E,$7F,$39,$1E,$6F,$E3,$BE,$23,$F9,$9C,$7F,$82,$1B
	.dc.b	$B9,$CE,$8F,$88,$FC,$2D,$7E,$0E,$31,$D5,$FE,$02,$6C,$7F,$96,$5A
	.dc.b	$FF,$B5,$5A,$FF,$7A,$94,$F1,$6C,$71,$6B,$C5,$AF,$E5,$FE,$2B,$F7
	.dc.b	$A4,$7F,$23,$B3,$E2,$BF,$23,$F1,$7F,$BE,$9F,$F3,$67,$4A,$31,$1F
	.dc.b	$EE,$98,$C7,$83,$FE,$1A,$9C,$CF,$CC,$9C,$BC,$A3,$2F,$E2,$23,$2F
	.dc.b	$38,$CB,$FA,$8E,$3F,$E2,$64,$7F,$4F,$AF,$FB,$65,$7C,$63,$E3,$7F
	.dc.b	$88,$B6,$3F,$C5,$ED,$7F,$D0,$CD,$7F,$02,$72,$8E,$24,$78,$BF,$C7
	.dc.b	$17,$C7,$3C,$FF,$E9,$8D,$8F,$CB,$35,$70,$FF,$29,$87,$9F,$F3,$BC
	.dc.b	$FF,$90,$4A,$F8,$CF,$F1,$CE,$0F,$EA,$F1,$F5,$BC,$FF,$E6,$A7,$F3
	.dc.b	$6F,$8F,$B1,$92,$1E,$0F,$07,$F9,$04,$7F,$3C,$F0,$64,$FE,$01,$FE
	.dc.b	$7F,$E4,$E2,$38,$1F,$C6,$23,$1A,$7F,$49,$C7,$FA,$04,$FE,$CF,$E4
	.dc.b	$BF,$A1,$CF,$F0,$2F,$91,$7C,$97,$EB,$27,$FC,$01,$59,$8C,$7A,$9F
	.dc.b	$B0,$AD,$CF,$F0,$AA,$E6,$A3,$2A,$33,$18,$7A,$C6,$EB,$AB,$DC,$E9
	.dc.b	$D1,$F4,$9D,$18,$C3,$F8,$CF,$9F,$CF,$1F,$E0,$6D,$9C,$DF,$AE,$F9
	.dc.b	$4F,$85,$E9,$8E,$EB,$A9,$E7,$BF,$75,$F1,$C3,$C9,$3E,$BB,$E3,$3E
	.dc.b	$88,$E3,$8F,$B3,$0E,$D7,$C2,$3E,$2C,$FE,$61,$AF,$53,$FD,$C4,$E1
	.dc.b	$F9,$9E,$6B,$EB,$FF,$D2,$33,$FD,$DD,$B1,$99,$E2,$D7,$B6,$3F,$25
	.dc.b	$B7,$CA,$31,$1E,$13,$27,$89,$CB,$F3,$CC,$47,$12,$38,$21,$C8,$87
	.dc.b	$52,$1C,$CC,$94,$C9,$F3,$5E,$AE,$8C,$1D,$8D,$DF,$C7,$72,$6D,$C1
	.dc.b	$27,$D7,$1A,$38,$B8,$8D,$4F,$82,$72,$D7,$CC,$43,$91,$63,$8B,$E5
	.dc.b	$31,$3D,$49,$3C,$E3,$89,$27,$8B,$E9,$9D,$BC,$CD,$38,$75,$3C,$E7
	.dc.b	$8C,$7A,$0E,$0E,$1F,$1C,$6E,$79,$46,$CE,$67,$07,$2F,$03,$11,$C4
	.dc.b	$8E,$08,$72,$21,$D4,$87,$33,$25,$32,$7B,$9E,$AE,$8C,$1D,$8D,$CF
	.dc.b	$B5,$C9,$B7,$04,$9F,$5C,$68,$E2,$E2,$35,$3E,$09,$CB,$5F,$31,$0E
	.dc.b	$45,$8E,$2F,$94,$C4,$F5,$24,$F3,$8E,$24,$9E,$2F,$A6,$76,$F3,$34
	.dc.b	$E1,$D4,$F3,$9E,$31,$E8,$38,$38,$7C,$71,$B9,$E5,$1B,$39,$9C,$1C
	.dc.b	$BC,$0C,$47,$12,$38,$21,$C8,$87,$52,$1C,$CC,$94,$C9,$EE,$7A,$BA
	.dc.b	$30,$76,$37,$3E,$D7,$26,$DC,$12,$7D,$71,$A3,$8B,$88,$D4,$F8,$27
	.dc.b	$2D,$7C,$C4,$39,$1F,$CD,$AF,$FC,$D7,$F4,$4E,$AF,$ED,$58,$FF,$35
	.dc.b	$8C,$56,$C3,$87,$FB,$34,$6E,$78,$D7,$46,$CF,$06,$D7,$1C,$3F,$A2
	.dc.b	$4E,$4F,$F1,$2F,$C7,$7F,$88,$48,$FF,$AF,$9A,$FF,$8A,$8F,$F4,$1B
	.dc.b	$87,$0D,$79,$9C,$48,$71,$0E,$DF,$F3,$AB,$5E,$A7,$03,$A9,$4E,$A7
	.dc.b	$37,$83,$D0,$D1,$D0,$A7,$43,$81,$83,$81,$82,$1E,$02,$9E,$03,$04
	.dc.b	$30,$43,$C0,$70,$29,$B2,$9B,$3C,$86,$4A,$64,$F2,$99,$29,$97,$E0
	.dc.b	$91,$F6,$35,$F1,$1D,$0F,$09,$0C,$10,$EA,$64,$A6,$5E,$8F,$8D,$AF
	.dc.b	$8C,$F0,$90,$F0,$9C,$88,$F0,$7B,$35,$EC,$71,$3B,$14,$EC,$74,$78
	.dc.b	$3D,$0D,$1D,$0A,$74,$38,$18,$38,$18,$21,$E0,$29,$E0,$30,$43,$04
	.dc.b	$3C,$07,$02,$9B,$29,$B3,$C8,$64,$A6,$4F,$29,$92,$99,$7E,$09,$1F
	.dc.b	$63,$5F,$11,$D0,$F0,$90,$C1,$0E,$A6,$4A,$65,$E8,$F8,$DA,$F8,$CF
	.dc.b	$09,$0F,$09,$C8,$8F,$07,$B3,$5E,$C7,$13,$B1,$4E,$C7,$47,$83,$D0
	.dc.b	$D1,$D0,$A7,$43,$81,$83,$81,$82,$1E,$02,$9E,$03,$04,$30,$43,$C0
	.dc.b	$70,$29,$B2,$9B,$3C,$86,$4A,$7F,$0B,$8F,$E6,$EF,$FA,$0F,$0F,$F3
	.dc.b	$C3,$FC,$F8,$C6,$1F,$D4,$5F,$F1,$0E,$7F,$E9,$72,$69,$FE,$89,$7F
	.dc.b	$D1,$F1,$FE,$92,$D7,$FD,$32,$FF,$AD,$5A,$FF,$9A,$A3,$7B,$7F,$A8
	.dc.b	$49,$5F,$E9,$37,$9C,$76,$9E,$47,$FA,$6F,$6F,$F5,$0B,$FE,$85,$8D
	.dc.b	$3F,$EA,$93,$5B,$47,$C8,$7F,$A8,$F0,$F1,$E7,$FE,$C4,$C7,$D4,$7F
	.dc.b	$AA,$EF,$9C,$66,$BB,$4E,$48,$E0,$87,$84,$EE,$7F,$AD,$30,$F3,$BF
	.dc.b	$E8,$AC,$7D,$EF,$B0,$3F,$12,$7C,$71,$9B,$F1,$D7,$2B,$FF,$6F,$DF
	.dc.b	$2A,$F0,$3C,$5F,$F5,$43,$98,$FF,$6F,$B9,$8F,$A5,$7E,$A7,$D1,$7F
	.dc.b	$42,$3E,$91,$FE,$CD,$9E,$57,$FE,$DA,$8C,$D7,$A6,$73,$8F,$FB,$82
	.dc.b	$3E,$1E,$1F,$ED,$3A,$EC,$FC,$B8,$F8,$51,$D4,$8F,$A1,$EA,$FB,$1C
	.dc.b	$BC,$0E,$B8,$FD,$A3,$ED,$E1,$F1,$4F,$34,$79,$67,$8D,$7C,$87,$AB
	.dc.b	$F1,$1C,$CF,$09,$F8,$A7,$D1,$8F,$4C,$75,$BF,$74,$F9,$4E,$C7,$A4
	.dc.b	$86,$49,$5E,$17,$D2,$7E,$4C,$FC,$B6,$3F,$56,$3C,$91,$D0,$F8,$04
	.dc.b	$72,$F4,$AF,$44,$7C,$D3,$E8,$47,$A4,$FA,$71,$F8,$9C,$3E,$44,$78
	.dc.b	$30,$F8,$70,$8F,$6D,$76,$3D,$0F,$C4,$3F,$0F,$3F,$13,$0F,$D7,$63
	.dc.b	$F8,$FB,$E5,$5F,$9C,$AF,$9F,$87,$9A,$73,$5F,$88,$AD,$1F,$10,$FD
	.dc.b	$4E,$3F,$93,$70,$FE,$8B,$0F,$44,$7B,$4F,$19,$EE,$7E,$63,$CB,$0F
	.dc.b	$D5,$BE,$9C,$7E,$2B,$C0,$C6,$3F,$15,$E3,$7E,$17,$86,$3F,$21,$FD
	.dc.b	$E6,$1E,$27,$95,$F4,$3E,$95,$F3,$8F,$11,$E5,$29,$C1,$FC,$63,$88
	.dc.b	$E5,$18,$BF,$E2,$77,$F9,$EB,$FA,$D0,$C3,$EC,$78,$4F,$D5,$9F,$87
	.dc.b	$7F,$99,$63,$0D,$8C,$E7,$FE,$71,$9C,$B6,$32,$FD,$A1,$1A,$70,$D7
	.dc.b	$EF,$B5,$F1,$FC,$9B,$EA,$7F,$07,$93,$A3,$31,$86,$D7,$E4,$67,$4F
	.dc.b	$83,$0F,$1C,$63,$06,$0F,$27,$95,$F9,$A3,$0F,$3C,$3C,$73,$AA,$7E
	.dc.b	$CC,$F6,$71,$87,$0C,$18,$F7,$8C,$37,$0E,$57,$C6,$BB,$1E,$1D,$79
	.dc.b	$91,$CE,$3F,$57,$DC,$C4,$76,$CF,$E3,$D6,$31,$F7,$68,$BF,$47,$EA
	.dc.b	$EF,$80,$FB,$8D,$C7,$ED,$35,$FA,$AD,$7E,$AB,$5F,$C3,$B5,$FB,$AD
	.dc.b	$C8,$DD,$7E,$8E,$7F,$73,$83,$E3,$9E,$4C,$3E,$4F,$D2,$CC,$7D,$F6
	.dc.b	$8F,$A1,$FA,$C3,$E5,$35,$5F,$04,$FB,$13,$8D,$BE,$F5,$6E,$79,$FC
	.dc.b	$57,$D6,$23,$F3,$3F,$6B,$7C,$DE,$23,$A6,$FF,$86,$AC,$61,$F8,$9B
	.dc.b	$DD,$7D,$4F,$DC,$1F,$3F,$FA,$76,$C7,$E9,$5A,$FE,$81,$AF,$F1,$EF
	.dc.b	$84,$7E,$E2,$F9,$7F,$C9,$C6,$1B,$19,$9F,$05,$FE,$4F,$D6,$FD,$D4
	.dc.b	$7E,$19,$AF,$A9,$B9,$7E,$93,$0E,$93,$BE,$8E,$44,$7F,$27,$1A,$7E
	.dc.b	$1F,$EE,$7E,$87,$F8,$3B,$5F,$DE,$B7,$63,$78,$FE,$0E,$1F,$A3,$CB
	.dc.b	$F4,$E7,$EA,$E7,$3C,$DF,$C4,$BF,$77,$78,$1F,$C4,$5B,$97,$F0,$F6
	.dc.b	$BF,$C2,$9A,$FF,$0A,$C5,$F2,$5F,$B4,$24,$7E,$AA,$78,$E5,$FB,$59
	.dc.b	$CC,$FE,$02,$B8,$BB,$BC,$D6,$E3,$F2,$B5,$FC,$0E,$F9,$E1,$FC,$AB
	.dc.b	$F7,$B7,$99,$FC,$A5,$B7,$FC,$9D,$AF,$F2,$26,$BE,$9F,$B2,$FE,$16
	.dc.b	$47,$F8,$0C,$70,$C7,$F8,$EC,$3C,$57,$9D,$FC,$BE,$3F,$B4,$C6,$1C
	.dc.b	$1B,$8A,$D3,$31,$B7,$93,$5F,$E7,$4D,$7A,$C7,$F4,$5F,$39,$F1,$7F
	.dc.b	$C9,$5B,$3F,$C6,$1A,$FE,$DC,$F0,$17,$D1,$FE,$14,$9A,$4F,$03,$CE
	.dc.b	$D8,$F5,$B5,$FE,$9D,$F7,$4F,$D6,$5F,$51,$FC,$F2,$30,$D8,$CC,$FF
	.dc.b	$30,$3C,$17,$99,$91,$FC,$D2,$B8,$BB,$BF,$E4,$DA,$7D,$2F,$A6,$7C
	.dc.b	$87,$CC,$6D,$7D,$06,$BF,$55,$AF,$F6,$56,$BF,$E2,$06,$BF,$E2,$B6
	.dc.b	$A7,$32,$F9,$4F,$1F,$1F,$F6,$16,$D7,$F5,$26,$BC,$6F,$FA,$FF,$A5
	.dc.b	$CF,$8F,$F0,$AD,$9F,$C5,$35,$FE,$F4,$D7,$FC,$6A,$F8,$FC,$3E,$39
	.dc.b	$1A,$49,$FC,$8F,$D2,$3E,$45,$DF,$86,$CF,$E6,$67,$53,$93,$8F,$D6
	.dc.b	$BE,$32,$92,$B8,$46,$5F,$3F,$E4,$BF,$CD,$37,$F5,$B6,$33,$E0,$FE
	.dc.b	$A9,$FF,$2C,$6A,$9D,$47,$B9,$FF,$11,$D7,$2F,$37,$84,$AB,$84,$61
	.dc.b	$B3,$FB,$3F,$82,$7A,$EF,$A1,$FF,$0A,$F2,$FF,$85,$A7,$31,$8A,$FF
	.dc.b	$0D,$64,$F8,$9C,$66,$BD,$D7,$F8,$23,$0F,$45,$79,$1B,$1F,$C0,$DB
	.dc.b	$5F,$E3,$36,$BF,$E2,$CB,$ED,$3B,$6C,$6D,$AE,$EB,$E0,$F4,$3F,$4F
	.dc.b	$7D,$8F,$9A,$30,$FA,$32,$7E,$26,$F8,$3D,$5F,$0C,$FF,$5C,$DC,$BF
	.dc.b	$BE,$63,$CA,$7F,$C4,$B8,$72,$BE,$38,$F5,$C7,$A5,$7C,$F6,$E1,$FE
	.dc.b	$51,$BE,$13,$9B,$FB,$31,$89,$D5,$70,$7F,$D2,$31,$96,$C6,$5F,$F2
	.dc.b	$AF,$A5,$F5,$3C,$CF,$F2,$5B,$F5,$BE,$29,$F5,$BF,$E1,$7C,$BF,$C3
	.dc.b	$73,$98,$C5,$FF,$99,$6B,$7E,$BF,$E3,$1B,$7F,$CE,$DB,$3B,$AE,$2F
	.dc.b	$FA,$9A,$32,$D8,$CB,$FE,$B3,$8C,$BF,$E6,$CD,$3D,$1E,$77,$CF,$E2
	.dc.b	$FF,$32,$79,$4F,$85,$CD,$7B,$63,$19,$7F,$59,$9F,$D0,$EA,$9C,$E3
	.dc.b	$FD,$6E,$73,$8B,$EB,$FB,$9F,$1F,$CE,$7C,$F6,$32,$FF,$1F,$CD,$C3
	.dc.b	$FC,$63,$5A,$9C,$92,$3F,$BB,$4F,$3A,$E1,$A7,$F9,$F6,$F8,$57,$F8
	.dc.b	$16,$7C,$F1,$88,$FF,$47,$B5,$FF,$47,$FD,$BF,$C4,$3E,$E7,$A7,$FE
	.dc.b	$44,$C7,$FC,$A7,$18,$E2,$F8,$33,$9A,$FF,$4D,$B5,$FF,$4D,$B5,$FF
	.dc.b	$56,$FA,$BF,$C2,$9F,$F1,$A7,$77,$EF,$3C,$24,$FF,$9D,$F1,$FF,$1D
	.dc.b	$4F,$E3,$63,$86,$9F,$E9,$96,$CF,$FB,$BE,$B8,$FB,$BB,$BE,$95,$CF
	.dc.b	$0F,$F3,$35,$6A,$FF,$CB,$19,$6B,$0F,$CD,$7B,$3E,$7B,$FF,$5B,$B7
	.dc.b	$8C,$D3,$1C,$2F,$AE,$FC,$59,$FA,$3F,$6F,$F0,$19,$7F,$9E,$A3,$2D
	.dc.b	$3F,$CE,$1C,$78,$C7,$FC,$61,$F5,$1F,$94,$3F,$59,$FA,$FB,$CD,$C5
	.dc.b	$FF,$A1,$26,$D7,$F9,$DA,$B5,$39,$3F,$D1,$3F,$69,$F9,$E3,$F6,$DF
	.dc.b	$04,$BC,$4F,$FA,$B2,$30,$FF,$B6,$63,$F3,$B1,$8B,$FF,$6B,$B5,$FF
	.dc.b	$6B,$B5,$FF,$74,$B5,$FF,$68,$47,$23,$F4,$7F,$71,$FB,$03,$C7,$C3
	.dc.b	$6F,$FD,$3F,$16,$BF,$D2,$37,$F8,$DE,$5F,$F6,$54,$FE,$E3,$D7,$97
	.dc.b	$FE,$C4,$F8,$47,$C8,$63,$1F,$F2,$D5,$5C,$3F,$DC,$15,$AB,$FF,$56
	.dc.b	$CF,$F1,$9F,$B2,$FF,$76,$47,$FB,$39,$B7,$E2,$F8,$5D,$4F,$CA,$7E
	.dc.b	$29,$1F,$F8,$00,$D8,$F8,$ED,$7E,$3B,$FD,$A1,$13,$E3,$6F,$C5,$FF
	.dc.b	$A6,$23,$FB,$F5,$63,$E1,$3F,$CB,$95,$67,$89,$A1,$5C,$54,$6D,$EF
	.dc.b	$EA,$2F,$11,$E1,$38,$DF,$47,$A5,$F3,$BD,$DF,$19,$D9,$B8,$CD,$F4
	.dc.b	$8F,$09,$A3,$C2,$E6,$37,$19,$3C,$55,$F0,$6B,$B3,$E1,$8D,$46,$2F
	.dc.b	$E1,$C6,$2F,$A3,$4E,$87,$8A,$BA,$46,$9B,$1A,$BF,$1C,$77,$32,$5A
	.dc.b	$F8,$97,$E1,$AF,$80,$77,$9D,$3F,$52,$34,$FD,$6A,$EB,$5E,$8A,$EA
	.dc.b	$D3,$AD,$F9,$5B,$5E,$53,$B3,$F0,$3E,$33,$D2,$7C,$52,$B9,$38,$B9
	.dc.b	$69,$DE,$3C,$D3,$DD,$A7,$79,$E6,$FB,$9A,$FA,$CF,$19,$F0,$AB,$D6
	.dc.b	$DA,$F4,$C7,$C4,$8F,$C0,$9D,$63,$C6,$F5,$6B,$D6,$7B,$BF,$95,$6B
	.dc.b	$F6,$E3,$A9,$EB,$7A,$CF,$BD,$F5,$CD,$7C,$2D,$7F,$50,$DC,$3F,$38
	.dc.b	$F9,$E3,$F4,$35,$D5,$A7,$9C,$EA,$FC,$F6,$99,$3B,$B9,$69,$F1,$4F
	.dc.b	$9F,$5D,$EB,$F0,$F1,$EF,$9F,$86,$D3,$A9,$F1,$5C,$BF,$AE,$AF,$8B
	.dc.b	$87,$E3,$A3,$B1,$F1,$27,$C4,$FC,$06,$D7,$C0,$6E,$1F,$C5,$2F,$E2
	.dc.b	$36,$3E,$14,$F9,$AB,$E9,$1E,$A8,$E2,$AB,$8B,$63,$31,$83,$E0,$57
	.dc.b	$E4,$67,$CC,$7C,$93,$CC,$5A,$F3,$3E,$C3,$E9,$9F,$93,$C3,$F2,$2D
	.dc.b	$BF,$C2,$BB,$9F,$46,$1F,$16,$34,$7A,$4D,$5F,$8D,$EA,$7F,$14,$AE
	.dc.b	$6F,$03,$CE,$53,$67,$C4,$32,$7D,$93,$F0,$27,$D9,$3E,$D9,$D4,$F1
	.dc.b	$1C,$EB,$CE,$FE,$B8,$C4,$FF,$1E,$7F,$9B,$1F,$CE,$4A,$7F,$30,$2D
	.dc.b	$63,$D2,$7C,$6F,$FA,$DE,$78,$C6,$3D,$4D,$4E,$67,$FA,$14,$E5,$DF
	.dc.b	$C1,$B8,$7A,$36,$33,$18,$23,$FD,$06,$5F,$0D,$FD,$63,$56,$BF,$D6
	.dc.b	$A7,$35,$FD,$5D,$61,$D6,$B7,$3F,$DA,$27,$38,$6F,$0E,$98,$73,$6D
	.dc.b	$6F,$0E,$53,$DD,$F4,$6D,$D2,$7F,$DA,$B5,$FD,$B2,$31,$87,$92,$7F
	.dc.b	$34,$73,$C3,$F4,$8D,$3D,$D5,$E0,$79,$9E,$BB,$F0,$BB,$C7,$AC,$EE
	.dc.b	$FB,$47,$82,$B7,$5F,$EE,$86,$CE,$B1,$F7,$4F,$5C,$23,$E2,$3E,$21
	.dc.b	$8B,$E9,$18,$8D,$1F,$56,$FA,$9D,$B0,$F2,$1E,$57,$33,$F0,$27,$DA
	.dc.b	$7C,$42,$E3,$D9,$C6,$3F,$59,$EF,$87,$C0,$7C,$98,$FD,$3B,$E9,$5F
	.dc.b	$3E,$F5,$5C,$B0,$FA,$F7,$B8,$EB,$8F,$D7,$8F,$87,$1B,$C3,$9D,$7A
	.dc.b	$E3,$E0,$63,$E5,$3C,$D8,$76,$25,$F5,$72,$F4,$73,$3F,$7D,$E5,$1E
	.dc.b	$09,$F3,$98,$9D,$47,$12,$38,$21,$F2,$1C,$BF,$7E,$FD,$27,$A2,$3B
	.dc.b	$1F,$06,$7C,$B3,$FA,$3C,$3E,$13,$4E,$27,$C3,$3E,$39,$F3,$9D,$BD
	.dc.b	$CD,$47,$AA,$7B,$C7,$E5,$DF,$88,$79,$B1,$FA,$CE,$23,$D7,$18,$7C
	.dc.b	$47,$87,$0F,$88,$FE,$22,$BB,$3D,$4E,$37,$E4,$AE,$51,$E5,$AF,$7B
	.dc.b	$F1,$CE,$AE,$23,$F1,$D5,$D2,$7B,$5F,$C7,$23,$E4,$8D,$4F,$09,$F5
	.dc.b	$4F,$52,$47,$E6,$6B,$D2,$FE,$C6,$39,$4F,$47,$CC,$FC,$5B,$F5,$63
	.dc.b	$F9,$89,$F6,$9F,$44,$FA,$B7,$D2,$7F,$5F,$1F,$A4,$3E,$21,$EA,$3B
	.dc.b	$98,$BF,$43,$83,$47,$DB,$C3,$A3,$F8,$93,$85,$FE,$8D,$FD,$9E,$1E
	.dc.b	$9A,$E7,$1C,$9F,$73,$CE,$35,$7E,$68,$FE,$11,$8F,$9D,$E1,$1F,$C1
	.dc.b	$E3,$15,$C8,$97,$FC,$2A,$BC,$27,$A6,$3C,$24,$AE,$D8,$7D,$0A,$F4
	.dc.b	$3F,$61,$FE,$57,$E0,$39,$9E,$27,$F6,$28,$C3,$63,$38,$FF,$68,$9C
	.dc.b	$B6,$32,$FF,$B7,$A7,$81,$1E,$1E,$03,$E3,$7F,$D6,$3A,$7F,$A8,$F0
	.dc.b	$4E,$6B,$FD,$7B,$39,$8C,$47,$FB,$72,$F8,$F1,$7F,$B0,$AF,$8F,$C6
	.dc.b	$62,$91,$87,$9C,$EE,$D5,$BA,$74,$3E,$37,$15,$AC,$38,$9D,$1C,$3F
	.dc.b	$DE,$F0,$E9,$87,$3B,$E5,$3C,$30,$F6,$F0,$F4,$24,$67,$57,$8B,$88
	.dc.b	$E8,$77,$9C,$6B,$F1,$EB,$73,$D7,$E3,$5F,$0B,$F6,$5B,$1F,$5D,$AF
	.dc.b	$FB,$57,$14,$62,$3F,$DC,$71,$89,$E4,$48,$E4,$FE,$17,$5F,$BF,$87
	.dc.b	$D5,$DF,$E3,$9F,$2F,$0E,$F1,$F6,$7A,$BE,$59,$30,$F8,$9F,$1A,$F5
	.dc.b	$31,$1F,$03,$3F,$C1,$D6,$2F,$F0,$D7,$BA,$F9,$99,$3E,$07,$EB,$92
	.dc.b	$3E,$87,$92,$F0,$F1,$1D,$6B,$83,$CA,$BF,$05,$5B,$8F,$B1,$E5,$3D
	.dc.b	$EF,$D2,$36,$3F,$44,$D7,$E6,$35,$13,$CC,$7C,$63,$E3,$7F,$2D,$18
	.dc.b	$6C,$66,$7C,$37,$F9,$5F,$3F,$FC,$5D,$B7,$E9,$DB,$D5,$FA,$16,$D3
	.dc.b	$1E,$8F,$E4,$63,$3C,$26,$9C,$27,$73,$F4,$7D,$0F,$53,$FA,$D6,$BF
	.dc.b	$AC,$6D,$1B,$BF,$C3,$C7,$F7,$19,$7E,$A8,$FD,$74,$E7,$2F,$DC,$7A
	.dc.b	$5F,$E4,$5C,$7F,$22,$D7,$3E,$0F,$F1,$38,$C4,$7E,$C7,$E3,$7F,$6A
	.dc.b	$47,$F5,$D1,$FE,$4F,$C7,$F8,$44,$E6,$7F,$00,$FE,$FE,$75,$7F,$C0
	.dc.b	$E3,$55,$C2,$B9,$DF,$F0,$5F,$59,$F1,$BF,$C9,$5B,$3F,$C8,$DA,$FE
	.dc.b	$B5,$AF,$EB,$34,$7A,$5F,$C3,$09,$1F,$C1,$67,$8E,$0C,$BF,$13,$3F
	.dc.b	$E7,$AF,$8D,$FF,$4A,$B4,$FF,$4B,$F8,$A8,$CB,$63,$31,$82,$38,$3F
	.dc.b	$9D,$79,$CF,$8D,$FE,$44,$D8,$FE,$30,$D7,$F8,$99,$E1,$2F,$C5,$FF
	.dc.b	$09,$7C,$ED,$7F,$42,$D4,$D2,$78,$3E,$C6,$C7,$FA,$D3,$E0,$5F,$1A
	.dc.b	$F8,$DF,$E7,$91,$86,$C6,$67,$F9,$A9,$DA,$F3,$32,$3F,$9C,$D7,$17
	.dc.b	$77,$FC,$5B,$4F,$9F,$F1,$6F,$8C,$F9,$ED,$AF,$F6,$2B,$5F,$F6,$4B
	.dc.b	$5F,$AE,$1C,$3D,$23,$99,$1F,$F6,$8E,$EF,$8D,$F1,$47,$FB,$15,$B1
	.dc.b	$9C,$FF,$AE,$4F,$5A,$FE,$BD,$E9,$F7,$E3,$FF,$6C,$B6,$7F,$DB,$AD
	.dc.b	$7C,$AD,$7F,$1A,$F9,$FC,$2E,$39,$7F,$24,$9D,$B5,$DB,$5F,$CA,$FC
	.dc.b	$63,$E3,$7F,$88,$47,$FB,$B2,$31,$DF,$FE,$EC,$6C,$66,$76,$FC,$88
	.dc.b	$7C,$6B,$D8,$E0,$D9,$E8,$D7,$83,$5F,$E5,$F9,$3E,$33,$F9,$C7,$0F
	.dc.b	$C2,$6C,$FC,$26,$BF,$09,$AF,$27,$FC,$47,$1C,$7C,$DF,$DA,$4F,$16
	.dc.b	$BC,$5A,$FF,$40,$F8,$97,$AE,$F8,$DE,$3C,$9C,$A7,$31,$8C,$3F,$C3
	.dc.b	$F8,$BE,$37,$18,$AF,$75,$FE,$12,$C3,$DD,$5E,$0A,$46,$27,$F1,$B7
	.dc.b	$CD,$B5,$B9,$D5,$7E,$61,$AF,$F8,$D2,$B8,$39,$6B,$FC,$53,$A1,$F1
	.dc.b	$AF,$89,$F3,$C6,$1F,$4E,$4F,$8C,$7C,$6E,$AF,$86,$9F,$B4,$6E,$3F
	.dc.b	$E2,$0C,$79,$4F,$F8,$C3,$0E,$57,$C7,$1E,$B8,$F4,$AF,$D8,$E1,$F1
	.dc.b	$E3,$0F,$ED,$2F,$8C,$E9,$B1,$99,$DB,$FC,$C2,$32,$D8,$CB,$F7,$5A
	.dc.b	$FF,$90,$FD,$1F,$AF,$E6,$7F,$96,$9F,$B3,$F1,$8F,$8D,$FF,$0F,$63
	.dc.b	$FE,$18,$9C,$C6,$2F,$F8,$4D,$6F,$D6,$FB,$8D,$BF,$6B,$67,$75,$C5
	.dc.b	$FE,$8F,$19,$6C,$65,$FF,$09,$C6,$5F,$F3,$16,$7E,$8F,$27,$E8,$F0
	.dc.b	$FF,$9C,$FC,$97,$C6,$E6,$BD,$B1,$8C,$BF,$B0,$4F,$E4,$75,$4E,$71
	.dc.b	$FE,$C5,$39,$C5,$F1,$5D,$CF,$8F,$E7,$BE,$37,$19,$7F,$8F,$26,$E1
	.dc.b	$FE,$43,$AD,$4E,$49,$1F,$E3,$B9,$E7,$5C,$34,$FF,$46,$5F,$0A,$FF
	.dc.b	$30,$35,$FF,$07,$CF,$A6,$31,$1F,$E8,$5F,$8D,$F8,$87,$DC,$F4,$BF
	.dc.b	$C9,$58,$FF,$94,$23,$1C,$5F,$06,$73,$5F,$E6,$D6,$BF,$E6,$C6,$BF
	.dc.b	$E3,$7E,$CF,$C3,$FA,$9F,$C3,$A3,$FC,$FF,$E1,$FF,$8D,$7D,$EC,$D4
	.dc.b	$CB,$FC,$74,$FF,$04,$9D,$46,$0F,$F4,$AB,$8A,$E7,$7F,$E8,$C9,$FF
	.dc.b	$2D,$E3,$C6,$31,$3F,$E6,$FF,$63,$CA,$FF,$AA,$1B,$C6,$69,$8E,$17
	.dc.b	$C6,$FE,$24,$FD,$0F,$B5,$F7,$B2,$FF,$3A,$C6,$5A,$7F,$9D,$38,$F1
	.dc.b	$8F,$F9,$CF,$E2,$FF,$24,$7E,$AB,$E3,$7F,$C4,$F5,$FE,$7E,$8C,$B4
	.dc.b	$FF,$56,$46,$A7,$94,$7F,$A1,$3E,$37,$F3,$87,$ED,$3E,$05,$78,$9F
	.dc.b	$F5,$83,$F9,$98,$CC,$7F,$B6,$E3,$04,$70,$4A,$FF,$55,$4F,$07,$FD
	.dc.b	$59,$9B,$E3,$7F,$58,$78,$78,$2D,$FF,$A6,$22,$D7,$FA,$7E,$FF,$13
	.dc.b	$D1,$FE,$DF,$9F,$DA,$7A,$D2,$FF,$D9,$5F,$08,$F8,$DC,$63,$FE,$9B
	.dc.b	$AB,$87,$FB,$AE,$B5,$7F,$EA,$69,$FE,$2B,$F1,$BF,$E4,$D8,$FF,$6A
	.dc.b	$36,$FC,$5F,$09,$A9,$F9,$0F,$C5,$23,$FE,$FF,$6C,$7C,$26,$BF,$EC
	.dc.b	$AC,$D1,$FD,$56,$3F,$CB,$7F,$1B,$FE,$A4,$F8,$B7,$C4,$7F,$8A,$3D
	.dc.b	$2F,$ED,$B1,$8A,$FF,$15,$D6,$EB,$58,$D9,$C8,$AE,$0F,$FA,$C2,$F9
	.dc.b	$D7,$F4,$4B,$E1,$39,$BA,$FF,$AD,$A2,$D7,$08,$EB,$7D,$1F,$F0,$A5
	.dc.b	$70,$55,$C2,$30,$D8,$F2,$5F,$85,$A7,$F9,$F2,$FB,$3F,$10,$C9,$E5
	.dc.b	$29,$E5,$2B,$C1,$C5,$F9,$63,$11,$E4,$37,$7E,$46,$9D,$0D,$4F,$18
	.dc.b	$D3,$63,$57,$EB,$8F,$85,$1C,$2F,$E9,$BA,$3A,$C6,$A7,$84,$FC,$18
	.dc.b	$E8,$66,$77,$19,$39,$C6,$4D,$BF,$27,$E3,$3C,$87,$12,$B9,$3C,$6E
	.dc.b	$5A,$75,$9F,$A1,$7E,$E8,$EE,$71,$8C,$89,$E4,$E1,$F2,$47,$08,$FC
	.dc.b	$2B,$5F,$B5,$5E,$87,$27,$47,$9B,$F7,$66,$BF,$8C,$3C,$13,$E3,$74
	.dc.b	$D7,$51,$E2,$7A,$47,$B5,$FC,$44,$F4,$C3,$FC,$AF,$56,$D5,$D8,$FD
	.dc.b	$4C,$7C,$A6,$DF,$C8,$7D,$27,$47,$B3,$E7,$BE,$8F,$33,$F3,$53,$CD
	.dc.b	$B3,$CD,$A7,$CB,$79,$DF,$E6,$30,$FC,$3C,$F8,$8D,$14,$F8,$65,$3E
	.dc.b	$1C,$FC,$59,$F7,$BD,$30,$FB,$D3,$F1,$8E,$13,$A6,$C7,$C4,$BF,$59
	.dc.b	$2B,$94,$FE,$77,$0F,$B8,$FA,$8F,$19,$4F,$19,$D0,$F8,$87,$EE,$67
	.dc.b	$E2,$3F,$B8,$9F,$C4,$34,$F3,$1C,$EF,$CC,$DB,$F3,$4F,$92,$FF,$05
	.dc.b	$3F,$54,$FA,$E7,$D5,$8D,$11,$CD,$7C,$F7,$F6,$26,$2B,$E3,$BC,$CF
	.dc.b	$CD,$CF,$99,$FE,$00,$79,$8F,$41,$B3,$DC,$FD,$F6,$9F,$68,$E6,$71
	.dc.b	$39,$9C,$4A,$7A,$4F,$21,$FC,$34,$FF,$3E,$9F,$C7,$4B,$86,$1F,$E4
	.dc.b	$C7,$FB,$CB,$D1,$7C,$6B,$D3,$E3,$18,$6C,$66,$7F,$D1,$33,$96,$FC
	.dc.b	$17,$FA,$B1,$BC,$5C,$A3,$11,$DD,$AF,$EB,$8F,$F6,$0B,$C7,$5F,$EA
	.dc.b	$B4,$F8,$4D,$46,$1B,$9F,$19,$57,$18,$C3,$6F,$F7,$73,$CD,$FC,$29
	.dc.b	$8A,$E6,$EA,$78,$3E,$4C,$36,$FF,$B5,$A7,$6D,$33,$5E,$29,$B5,$CA
	.dc.b	$3C,$13,$FC,$05,$B8,$7B,$A7,$46,$EF,$AC,$EB,$0F,$5E,$1D,$0E,$F3
	.dc.b	$A3,$75,$CA,$33,$84,$3C,$11,$97,$C5,$3C,$5E,$86,$A7,$86,$1F,$3E
	.dc.b	$32,$F9,$E3,$37,$F4,$6B,$C3,$1A,$8C,$56,$8F,$99,$19,$7E,$81,$93
	.dc.b	$E7,$1E,$13,$CA,$67,$0F,$98,$E6,$FC,$8E,$8C,$57,$C6,$71,$1A,$3D
	.dc.b	$0E,$8F,$09,$A8,$E1,$1A,$38,$3E,$A8,$E0,$79,$A3,$2F,$27,$84,$72
	.dc.b	$38,$3A,$8E,$11,$97,$83,$A3,$0E,$8C,$90,$C1,$27,$F0,$8E,$5F,$8B
	.dc.b	$3E,$28,$C5,$FD,$47,$D3,$3C,$1F,$CC,$18,$C3,$F2,$0F,$CF,$7F,$0D
	.dc.b	$5E,$99,$E0,$76,$BF,$3B,$8B,$ED,$3E,$1B,$EA,$F1,$BE,$77,$EF,$9F
	.dc.b	$44,$61,$F3,$9F,$85,$3D,$D5,$F9,$57,$CD,$87,$E0,$E3,$BD,$78,$9D
	.dc.b	$5F,$87,$0F,$35,$F4,$3F,$01,$DE,$E1,$78,$5F,$C9,$A3,$0D,$8C,$CF
	.dc.b	$F1,$53,$FC,$07,$3A,$70,$17,$D0,$3D,$CE,$8D,$9D,$CE,$A7,$24,$8C
	.dc.b	$CE,$DF,$04,$ED,$CC,$F9,$F9,$DF,$1B,$49,$3D,$A3,$2D,$8C,$BD,$23
	.dc.b	$2D,$8F,$E8,$FE,$93,$F5,$57,$C4,$61,$B5,$F0,$27,$53,$92,$46,$67
	.dc.b	$6F,$C9,$9D,$B9,$9F,$9D,$F1,$F2,$3E,$2E,$88,$C4,$7F,$B2,$3C,$AF
	.dc.b	$A8,$47,$6D,$74,$60,$87,$CA,$FD,$BF,$A3,$F4,$9A,$E6,$7F,$05,$18
	.dc.b	$8F,$EA,$9A,$A3,$11,$F8,$26,$BF,$81,$6C,$6E,$7F,$B7,$70,$7F,$B4
	.dc.b	$6B,$FB,$5F,$7E,$08,$E0,$87,$F6,$C7,$2F,$E0,$BE,$3D,$F0,$9F,$97
	.dc.b	$6C,$7E,$31,$AE,$E3,$F0,$DC,$92,$3F,$08,$D7,$F0,$4D,$7D,$FC,$3F
	.dc.b	$E7,$B7,$67,$E6,$3E,$3D,$EC,$7E,$9D,$B1,$98,$C1,$1C,$10,$FC,$DC
	.dc.b	$70,$7F,$09,$FA,$DB,$EE,$FF,$22,$D8,$FC,$73,$5D,$C6,$88,$E9,$AF
	.dc.b	$6F,$AD,$F3,$11,$F5,$3F,$BF,$F3,$DE,$4D,$24,$FE,$DA,$32,$D8,$CB
	.dc.b	$FB,$48,$CB,$63,$F8,$57,$C0,$BC,$5A,$49,$FE,$03,$19,$6C,$65,$FD
	.dc.b	$FC,$65,$B1,$FE,$3A,$F4,$DF,$B6,$BE,$1B,$0D,$AF,$E2,$73,$A9,$C9
	.dc.b	$FC,$77,$D4,$79,$74,$95,$FC,$52,$32,$F2,$FD,$C9,$F0,$BE,$16,$C6
	.dc.b	$63,$04,$70,$43,$F9,$46,$09,$E3,$18,$FD,$D9,$F2,$5E,$D6,$C7,$A9
	.dc.b	$AE,$E3,$F9,$B7,$B5,$E6,$8F,$E4,$AD,$7F,$6E,$D7,$F9,$BB,$55,$BF
	.dc.b	$7C,$FD,$CB,$7B,$7E,$74,$60,$FD,$DF,$BB,$FB,$9A,$FE,$66,$D7,$F8
	.dc.b	$2B,$5F,$E8,$AD,$7F,$A2,$37,$BC,$D2,$77,$7F,$51,$6C,$FF,$46,$6B
	.dc.b	$FD,$47,$E0,$BF,$19,$F6,$9F,$E7,$C6,$C7,$F5,$26,$BB,$E4,$4F,$F5
	.dc.b	$DF,$82,$FC,$34,$FF,$4C,$6B,$FC,$79,$AF,$F5,$E6,$BF,$EE,$2F,$8A
	.dc.b	$7C,$BF,$FA,$25,$B1,$FE,$8E,$6B,$FD,$91,$AF,$F6,$96,$BF,$E7,$F6
	.dc.b	$BD,$63,$A1,$FE,$95,$F9,$1F,$E3,$71,$FD,$C5,$AF,$FA,$1D,$AF,$FB
	.dc.b	$D9,$AB,$EB,$DF,$37,$FE,$9F,$6C,$7F,$A9,$1A,$FF,$7A,$6B,$FD,$FD
	.dc.b	$AF,$FA,$55,$A8,$ED,$1D,$4F,$F5,$67,$C8,$3E,$4A,$92,$7F,$D6,$71
	.dc.b	$96,$C6,$5F,$F0,$5C,$65,$B1,$FE,$B7,$FB,$27,$C1,$52,$4F,$FA,$FA
	.dc.b	$32,$D8,$CB,$FE,$16,$8C,$B6,$3F,$D8,$5F,$26,$F9,$BA,$49,$FF,$64
	.dc.b	$C6,$5F,$F6,$5F,$DF,$3E,$6F,$FD,$94,$D8,$CC,$60,$CC,$C8,$E2,$D7
	.dc.b	$FC,$53,$F2,$DF,$C6,$FA,$BF,$DA,$ED,$8C,$F8,$5F,$B8,$6C,$EB,$8B
	.dc.b	$F9,$34,$7F,$B2,$69,$18,$8F,$AD,$18,$E5,$E0,$7F,$B9,$3F,$11,$7C
	.dc.b	$DB,$E1,$30,$DB,$FF,$1E,$CE,$A7,$27,$1F,$A6,$D4,$FF,$91,$65,$18
	.dc.b	$8F,$F6,$E4,$62,$78,$1F,$E4,$8B,$7E,$D7,$FE,$3A,$8C,$36,$BF,$CA
	.dc.b	$13,$AF,$73,$A4,$7F,$4D,$9D,$3F,$E5,$5F,$C0,$DF,$51,$A8,$46,$30
	.dc.b	$FF,$2B,$C6,$27,$53,$98,$FF,$2D,$B5,$E9,$1A,$24,$FF,$98,$3E,$11
	.dc.b	$F3,$8F,$3B,$CB,$18,$6C,$66,$78,$36,$FF,$CD,$95,$B9,$CD,$70,$23
	.dc.b	$E8,$FC,$11,$F2,$1E,$D6,$C6,$63,$04,$70,$4B,$FF,$3B,$CF,$07,$FC
	.dc.b	$A9,$F3,$4F,$9B,$F9,$BF,$06,$8C,$5F,$F9,$7D,$B5,$C2,$BE,$77,$E0
	.dc.b	$AF,$5E,$92,$FE,$7C,$65,$B1,$9A,$FF,$44,$CE,$5B,$1F,$7B,$CB,$78
	.dc.b	$FF,$4E,$7F,$9A,$70,$66,$31,$1F,$E9,$16,$F9,$3F,$E9,$3F,$9B,$FC
	.dc.b	$3B,$72,$FE,$8D,$82,$72,$D8,$FE,$A1,$8F,$87,$CB,$F2,$63,$FD,$7A
	.dc.b	$30,$49,$D9,$23,$FB,$2C,$65,$B1,$97,$93,$FE,$16,$FD,$F7,$F6,$ED
	.dc.b	$ED,$FC,$EF,$72,$73,$5F,$E0,$A6,$BF,$DF,$BE,$5F,$E3,$7A,$26,$93
	.dc.b	$0F,$F1,$1F,$EF,$2F,$A7,$EA,$D8,$CC,$60,$F0,$F4,$3E,$5A,$92,$7F
	.dc.b	$C5,$B1,$96,$C6,$71,$FF,$0F,$CE,$5B,$1F,$6F,$D9,$F3,$57,$F8,$BF
	.dc.b	$1C,$11,$FF,$1C,$FC,$17,$F8,$59,$BA,$FC,$48,$C6,$1F,$E2,$C6,$BF
	.dc.b	$E9,$6C,$5F,$11,$F8,$96,$DF,$F9,$42,$30,$47,$04,$9F,$F2,$8C,$7E
	.dc.b	$47,$DF,$F3,$C7,$F9,$7F,$1C,$12,$BF,$CC,$CD,$8D,$46,$0F,$F6,$FF
	.dc.b	$CC,$FF,$58,$6F,$9D,$F9,$D8,$C4,$FF,$B7,$9A,$FF,$6D,$1D,$AF,$B9
	.dc.b	$FF,$08,$DF,$FB,$45,$B3,$FE,$CD,$6B,$C6,$78,$7A,$DF,$BC,$9C,$CF
	.dc.b	$F7,$99,$CD,$73,$AD,$1E,$5F,$8E,$91,$EA,$6D,$79,$DA,$FF,$B1,$B0
	.dc.b	$FF,$80,$13,$B3,$FD,$E5,$F2,$3F,$BB,$DC,$D2,$46,$89,$5F,$EF,$58
	.dc.b	$CB,$F0,$C7,$53,$E3,$7E,$3C,$7F,$84,$E3,$1A,$FF,$09,$9C,$CF,$06
	.dc.b	$BF,$E1,$9E,$0F,$E8,$CF,$E1,$8F,$E8,$AE,$09,$87,$F4,$88,$D1,$E3
	.dc.b	$29,$E5,$8F,$E3,$1F,$1A,$F9,$2A,$4A,$FF,$78,$36,$BF,$DE,$8D,$7F
	.dc.b	$DD,$F3,$C6,$7B,$63,$EF,$F8,$6F,$61,$23,$38,$FF,$AE,$9B,$19,$BF
	.dc.b	$91,$5C,$67,$27,$0F,$8F,$F8,$C4,$70,$7B,$78,$5F,$20,$F9,$9F,$E8
	.dc.b	$AD,$9D,$37,$0F,$AD,$5C,$2B,$64,$AD,$B6,$7C,$BC,$0F,$81,$FC,$31
	.dc.b	$2B,$E0,$C6,$5B,$19,$9E,$2D,$7F,$51,$93,$E4,$BF,$9E,$47,$C2,$6B
	.dc.b	$A6,$BF,$89,$6B,$F8,$E6,$C7,$3F,$13,$FA,$A4,$60,$CE,$DF,$EC,$CC
	.dc.b	$3E,$9F,$C4,$FE,$44,$8F,$E2,$70,$7C,$BF,$F9,$A7,$E4,$3F,$0D,$1F
	.dc.b	$BF,$F5,$7F,$C6,$11,$88,$FE,$05,$3B,$E4,$FF,$11,$DD,$9F,$D2,$4D
	.dc.b	$7F,$CE,$97,$CA,$BF,$2D,$39,$9F,$F7,$A5,$57,$C6,$2F,$C0,$E8,$C5
	.dc.b	$72,$69,$E7,$9E,$55,$D6,$FC,$13,$E4,$36,$79,$27,$6E,$0E,$D7,$E0
	.dc.b	$7C,$6F,$8A,$79,$C7,$C2,$38,$38,$7F,$D2,$55,$E1,$AE,$B5,$C5,$B7
	.dc.b	$F2,$2B,$C0,$76,$71,$3E,$1A,$E5,$3A,$6C,$66,$3C,$35,$D1,$F8,$46
	.dc.b	$CC,$96,$3C,$4D,$AE,$E7,$CB,$8C,$9D,$4A,$E4,$DC,$7C,$B8,$CB,$63
	.dc.b	$2F,$C5,$8C,$B6,$32,$FC,$63,$CB,$1D,$E3,$31,$F0,$CF,$5C,$65,$B1
	.dc.b	$E3,$69,$E1,$29,$DC,$A7,$78,$F1,$B6,$3B,$D7,$CC,$7A,$47,$E1,$4F
	.dc.b	$8C,$72,$3C,$86,$67,$C8,$D8,$E2,$FE,$29,$B7,$F8,$58,$E2,$64,$F4
	.dc.b	$1C,$4A,$64,$F5,$19,$29,$93,$D4,$79,$6B,$D4,$F2,$8F,$04,$62,$38
	.dc.b	$C6,$1F,$A3,$3E,$6B,$FC,$54,$77,$3A,$94,$F4,$CF,$1B,$F9,$71,$89
	.dc.b	$E2,$F9,$4C,$1F,$28,$E5,$7E,$37,$A9,$B7,$0F,$F0,$23,$FD,$4A,$FF
	.dc.b	$17,$23,$FD,$7F,$E4,$DF,$2C,$F9,$7F,$F5,$5C,$61,$B1,$99,$E0,$D7
	.dc.b	$FB,$D4,$BE,$5D,$EA,$F2,$6C,$F7,$6B,$C1,$AF,$F8,$96,$79,$4A,$1F
	.dc.b	$31,$FC,$A6,$7F,$C1,$F1,$98,$C7,$6F,$F8,$B5,$D5,$F8,$1B,$7B,$DB
	.dc.b	$B5,$6D,$B7,$93,$C1,$87,$89,$F0,$C7,$97,$0F,$5E,$BE,$09,$FE,$6D
	.dc.b	$5E,$47,$F9,$2B,$C6,$3C,$31,$C8,$EB,$7E,$16,$9E,$11,$3E,$1C,$78
	.dc.b	$9D,$E3,$91,$A9,$E4,$F0,$BF,$81,$87,$9F,$1E,$98,$79,$70,$E4,$74
	.dc.b	$9D,$5F,$49,$D4,$7C,$48,$C5,$F4,$7E,$2B,$5F,$11,$DB,$0F,$1D,$7A
	.dc.b	$6B,$AD,$F8,$1E,$8F,$86,$B9,$E1,$F2,$AF,$FC,$CD,$8F,$BC,$F9,$26
	.dc.b	$E7,$B9,$E7,$8E,$F8,$F8,$F1,$FA,$18,$FA,$6F,$93,$EE,$79,$B9,$25
	.dc.b	$F8,$8F,$B3,$5C,$EF,$D6,$76,$C3,$E5,$C7,$E0,$AB,$8D,$79,$2F,$AB
	.dc.b	$CD,$F0,$BF,$78,$EB,$5F,$88,$AE,$D1,$F2,$4E,$B1,$1F,$C5,$DF,$C4
	.dc.b	$BE,$F8,$7B,$A7,$B4,$62,$75,$18,$3D,$98,$FD,$17,$C4,$FC,$BA,$E5
	.dc.b	$1E,$9A,$EB,$87,$B6,$39,$19,$3D,$58,$FB,$9F,$C8,$63,$E9,$7E,$A1
	.dc.b	$E9,$C3,$58,$7E,$96,$BE,$1C,$F0,$3E,$19,$EE,$3F,$04,$7C,$5A,$8F
	.dc.b	$42,$57,$A6,$FC,$D1,$E5,$25,$FB,$09,$8F,$C4,$8C,$CF,$09,$FC,$B4
	.dc.b	$7C,$4C,$7F,$2A,$7E,$60,$8E,$48,$7A,$6F,$DB,$7D,$6B,$DD,$5F,$8F
	.dc.b	$38,$D7,$E6,$2B,$E9,$3C,$63,$FD,$3D,$C0,$F9,$8F,$F4,$94,$FF,$A7
	.dc.b	$5B,$C9,$FE,$A5,$6C,$EE,$7F,$D0,$F1,$82,$38,$3F,$D1,$FF,$34,$F8
	.dc.b	$EF,$F6,$96,$AE,$2F,$F5,$74,$E6,$BF,$98,$4E,$5F,$F5,$AC,$65,$FE
	.dc.b	$7B,$3E,$8B,$E6,$FF,$AD,$FE,$69,$F2,$B8,$AF,$F5,$EC,$61,$EB,$3B
	.dc.b	$9D,$67,$FE,$7D,$F1,$2D,$EE,$FB,$9C,$AB,$27,$0E,$85,$72,$C9,$F1
	.dc.b	$14,$97,$E0,$3F,$D9,$FD,$8F,$80,$CC,$76,$CD,$7C,$A3,$84,$62,$FC
	.dc.b	$F7,$C6,$73,$C1,$F1,$77,$F9,$17,$FE,$DB,$E0,$F6,$CE,$77,$FF,$7B
	.dc.b	$78,$0F,$07,$35,$31,$E6,$FC,$9F,$07,$FD,$6F,$3F,$13,$EA,$25,$79
	.dc.b	$38,$D1,$C6,$3D,$5B,$FF,$A3,$F5,$F1,$B7,$3F,$BF,$CB,$E4,$F1,$7E
	.dc.b	$43,$2F,$DF,$5B,$7F,$DD,$D3,$F5,$3D,$49,$1E,$17,$EC,$F8,$DE,$D8
	.dc.b	$FB,$F1,$A7,$E6,$66,$8D,$ED,$F8,$0F,$9B,$7C,$DF,$E7,$9B,$7F,$EE
	.dc.b	$66,$BF,$EE,$06,$BF,$EC,$46,$BF,$E0,$7E,$34,$62,$11,$89,$EC,$7E
	.dc.b	$B3,$3F,$D5,$4F,$C8,$EA,$FC,$51,$E5,$9F,$C9,$FC,$44,$9F,$A0,$FE
	.dc.b	$0B,$C0,$C4,$7B,$A2,$F0,$FE,$0B,$9F,$FD,$65,$3F,$33,$C5,$FC,$B1
	.dc.b	$1F,$C6,$FC,$D3,$E6,$F1,$1F,$6A,$6D,$FE,$32,$B5,$3C,$3D,$AF,$E0
	.dc.b	$B0,$69,$23,$FC,$54,$7F,$BA,$A3,$4E,$0F,$90,$E1,$C6,$C8,$C4,$71
	.dc.b	$9F,$E2,$7C,$4F,$A5,$FD,$B9,$27,$C8,$7E,$2B,$E3,$3F,$68,$47,$F5
	.dc.b	$7E,$A3,$CF,$C4,$7E,$62,$6E,$7F,$AA,$AD,$4F,$E7,$FE,$5F,$F8,$29
	.dc.b	$1F,$DA,$FA,$CF,$3B,$11,$F9,$09,$B7,$FB,$4A,$D4,$FE,$F3,$D8,$70
	.dc.b	$7F,$2E,$6C,$7C,$53,$F1,$2E,$1F,$8A,$D7,$E6,$FB,$0F,$99,$7C,$D7
	.dc.b	$F2,$18,$C3,$63,$33,$E1,$BF,$E4,$3F,$3D,$FC,$3E,$3F,$89,$37,$2F
	.dc.b	$E1,$6D,$F0,$BF,$CC,$98,$F4,$AF,$E2,$B3,$97,$F3,$ED,$E5,$39,$9E
	.dc.b	$13,$FC,$7F,$DF,$F5,$3F,$A9,$6B,$FA,$86,$D1,$BC,$8F,$A7,$FE,$6E
	.dc.b	$4A,$FE,$4F,$3F,$1F,$2F,$E4,$7F,$04,$F6,$3F,$AD,$B6,$BE,$76,$28
	.dc.b	$C4,$7E,$49,$AF,$E6,$A1,$8B,$E6,$7F,$A3,$92,$7F,$9B,$D7,$2C,$3F
	.dc.b	$99,$D6,$E7,$F9,$45,$6D,$FE,$81,$5B,$D7,$FA,$D4,$E6,$F8,$BF,$D5
	.dc.b	$A7,$55,$FC,$33,$E1,$5D,$DF,$DD,$9B,$1E,$C6,$BF,$AC,$6B,$EB,$C5
	.dc.b	$F4,$5F,$D8,$09,$1F,$D4,$27,$8E,$3F,$DC,$E3,$E0,$78,$3F,$E0,$66
	.dc.b	$9F,$E0,$8C,$BF,$90,$38,$3F,$34,$E0,$91,$C5,$FE,$4D,$F4,$8F,$A7
	.dc.b	$FD,$E5,$7F,$67,$8C,$77,$7F,$7B,$6C,$7F,$6B,$3B,$61,$E5,$F9,$A7
	.dc.b	$CD,$79,$5B,$3F,$E1,$C6,$BF,$AD,$6B,$FC,$4D,$AF,$F1,$86,$BF,$EC
	.dc.b	$36,$A7,$32,$FA,$7F,$E8,$EF,$F8,$64,$C6,$7F,$E1,$36,$D7,$A6,$BC
	.dc.b	$71,$BA,$E3,$C3,$FE,$22,$6D,$7F,$B7,$1A,$F8,$70,$FF,$0F,$2F,$03
	.dc.b	$F7,$4F,$F2,$D6,$BF,$CC,$9A,$FF,$B7,$5A,$F1,$F5,$3E,$E4,$60,$8F
	.dc.b	$F9,$13,$D2,$7D,$4B,$EC,$7F,$C0,$CD,$BF,$F0,$F4,$EA,$7C,$B7,$FC
	.dc.b	$17,$E0,$3E,$43,$FE,$4E,$6D,$7F,$1F,$6B,$FD,$15,$AF,$F4,$A7,$E8
	.dc.b	$7A,$7C,$76,$34,$93,$FE,$5A,$F9,$AF,$D2,$F7,$5E,$93,$49,$3F,$CB
	.dc.b	$E7,$FC,$0D,$E3,$71,$F4,$BF,$9B,$CF,$F5,$79,$CB,$B8,$CA,$19,$3E
	.dc.b	$57,$FA,$77,$1F,$B5,$B3,$ED,$6B,$ED,$6B,$CE,$79,$7C,$DF,$EC,$63
	.dc.b	$12,$8C,$47,$82,$31,$5D,$E7,$3C,$0F,$B0,$F3,$4F,$D6,$6D,$FF,$97
	.dc.b	$B3,$7C,$CF,$F9,$56,$B5,$3F,$9B,$7D,$D3,$AF,$77,$C2,$D8,$FE,$B0
	.dc.b	$D9,$FE,$FC,$D7,$FC,$06,$D7,$9D,$77,$76,$D7,$6D,$7E,$76,$0F,$82
	.dc.b	$FE,$BF,$7F,$90,$6C,$FF,$9E,$3A,$7E,$97,$99,$E9,$85,$F1,$F2,$F0
	.dc.b	$DC,$39,$46,$EF,$8C,$60,$FF,$3D,$38,$AF,$F0,$D3,$5F,$F4,$EE,$CC
	.dc.b	$3E,$AC,$E6,$7F,$D3,$B3,$9A,$FE,$FD,$39,$93,$4C,$76,$BE,$A9,$F1
	.dc.b	$1E,$B8,$C3,$EA,$F9,$77,$CB,$E6,$B7,$8F,$ED,$27,$2F,$F8,$37,$54
	.dc.b	$63,$2F,$F1,$A4,$62,$FF,$13,$39,$9F,$F1,$DC,$E6,$B9,$4E,$67,$EA
	.dc.b	$B5,$F9,$ED,$F2,$7F,$B4,$FD,$07,$D3,$6D,$7F,$4F,$F9,$7C,$79,$4A
	.dc.b	$D7,$92,$9F,$CF,$DA,$31,$7F,$D7,$E7,$31,$8C,$7F,$C9,$78,$5D,$FF
	.dc.b	$B9,$D7,$87,$CB,$FF,$2C,$61,$FE,$9B,$8F,$09,$1F,$EF,$7E,$B5,$C3
	.dc.b	$FD,$17,$36,$FF,$C1,$B3,$6B,$FD,$05,$7D,$2F,$FC,$15,$3E,$88,$C4
	.dc.b	$7F,$82,$DA,$FF,$B6,$7E,$BF,$E0,$9F,$17,$CF,$7C,$9E,$30,$FF,$0E
	.dc.b	$BF,$0A,$33,$1F,$E1,$A6,$D7,$F9,$A9,$AF,$F8,$67,$D2,$FF,$27,$CF
	.dc.b	$F8,$E3,$D3,$FD,$27,$87,$F9,$7F,$94,$FA,$95,$AB,$FF,$21,$D6,$0D
	.dc.b	$F5,$2B,$18,$7F,$B7,$70,$E4,$D9,$D3,$FE,$5F,$CB,$8D,$7E,$07,$D6
	.dc.b	$F2,$E3,$FE,$A9,$6B,$FE,$60,$9F,$BF,$18,$E3,$FF,$18,$FD,$8F,$E1
	.dc.b	$8F,$C5,$62,$F8,$4C,$65,$FE,$D5,$9B,$8F,$FB,$B6,$B5,$39,$3F,$D9
	.dc.b	$FF,$53,$F9,$3F,$82,$3F,$95,$F7,$3E,$5F,$2F,$FB,$6A,$34,$4A,$FF
	.dc.b	$45,$EB,$71,$FF,$67,$FD,$97,$E6,$8F,$CF,$7C,$AB,$E5,$71,$7F,$EF
	.dc.b	$C8,$C3,$FE,$52,$6D,$7F,$A8,$9A,$FF,$94,$5A,$FF,$B3,$DA,$FF,$B3
	.dc.b	$9B,$B9,$A6,$34,$EB,$9F,$FB,$D3,$0F,$2F,$C6,$FC,$16,$63,$FE,$F2
	.dc.b	$72,$D3,$FE,$00,$7D,$AF,$ED,$8F,$DD,$7C,$2B,$E4,$B1,$3A,$26,$9F
	.dc.b	$C7,$1F,$F7,$34,$ED,$CE,$27,$FD,$D5,$93,$E7,$FE,$A1,$FC,$1F,$E4
	.dc.b	$3D,$33,$A8,$C6,$7F,$C1,$27,$E0,$91,$C1,$27,$FC,$FF,$19,$7F,$D0
	.dc.b	$3E,$03,$ED,$BF,$88,$9F,$C5,$7E,$3B,$FC,$1B,$3F,$DF,$72,$FF,$3D
	.dc.b	$CE,$A3,$07,$FC,$01,$AE,$4E,$34,$FF,$80,$37,$C9,$F8,$43,$87,$FD
	.dc.b	$17,$7F,$91,$F9,$89,$7C,$65,$97,$F8,$23,$2F,$8D,$87,$1D,$58,$70
	.dc.b	$8E,$B6,$9C,$E1,$CF,$0E,$4F,$CE,$BF,$E5,$FF,$27,$23,$5F,$25,$71
	.dc.b	$E7,$68,$FA,$47,$F6,$8F,$93,$EB,$3A,$8C,$D7,$23,$FC,$87,$3A,$CF
	.dc.b	$D3,$9F,$B2,$BF,$61,$DD,$2F,$E0,$71,$7F,$AD,$DF,$85,$AF,$F0,$AC
	.dc.b	$BD,$AD,$C7,$9B,$67,$F1,$BF,$11,$CE,$BE,$1F,$0F,$F6,$F6,$F6,$7F
	.dc.b	$51,$DF,$FB,$2C,$62,$FE,$85,$6F,$1F,$F5,$CD,$EF,$CE,$FE,$AA,$FF
	.dc.b	$56,$A8,$7C,$DF,$0A,$E1,$F2,$1F,$CA,$12,$3F,$22,$D8,$FF,$24,$1F
	.dc.b	$E4,$9D,$FE,$FC,$66,$3F,$DA,$1C,$3F,$CC,$DD,$9C,$3E,$63,$FC,$A1
	.dc.b	$F2,$DF,$E6,$28,$FF,$16,$E1,$8F,$7B,$70,$AC,$3A,$15,$C5,$E8,$3E
	.dc.b	$47,$0B,$D8,$FC,$8C,$62,$3A,$1B,$BF,$F1,$FE,$1F,$E6,$39,$D6,$1E
	.dc.b	$0A,$FF,$24,$CD,$7F,$CD,$11,$6F,$F3,$B8,$70,$9F,$2E,$1D,$1A,$7C
	.dc.b	$2A,$E1,$7C,$E3,$C7,$87,$A3,$0F,$3D,$FC,$2B,$F2,$3C,$E7,$B1,$BB
	.dc.b	$CD,$F5,$3D,$87,$C7,$8F,$26,$1B,$8F,$19,$F2,$1C,$1D,$2B,$B1,$E1
	.dc.b	$8E,$67,$C2,$29,$F0,$8B,$87,$A1,$F1,$57,$73,$C3,$3F,$0E,$AC,$FD
	.dc.b	$B7,$C0,$F9,$E3,$C1,$3C,$5A,$79,$23,$AB,$5F,$53,$D4,$F4,$4F,$B1
	.dc.b	$E3,$1D,$4E,$25,$3F,$13,$3D,$DC,$9E,$F7,$CE,$FC,$19,$AF,$E3,$6F
	.dc.b	$ED,$61,$F4,$A3,$85,$F8,$1A,$74,$38,$1F,$28,$FC,$0B,$EB,$7E,$65
	.dc.b	$7A,$5F,$B8,$D7,$CF,$1B,$78,$9E,$B3,$EA,$5F,$AD,$D9,$93,$F4,$33
	.dc.b	$CD,$EF,$3F,$25,$F3,$61,$F6,$DE,$27,$A8,$F9,$E7,$D9,$8F,$9E,$71
	.dc.b	$72,$7C,$77,$2E,$CF,$5B,$B7,$C2,$64,$E2,$7C,$69,$F9,$F3,$FA,$E9
	.dc.b	$E7,$3D,$4E,$67,$13,$99,$93,$9C,$71,$7D,$B8,$7D,$77,$E3,$CF,$CB
	.dc.b	$C3,$F6,$B1,$FA,$63,$99,$E3,$8F,$0B,$4F,$09,$DC,$FD,$21,$FC,$68
	.dc.b	$FE,$1E,$58,$C3,$A3,$F9,$5F,$AE,$F9,$8F,$F7,$9C,$7F,$22,$8C,$7B
	.dc.b	$BC,$DB,$19,$9D,$BF,$CB,$A1,$F3,$1F,$EB,$48,$E4,$D7,$4D,$7C,$2D
	.dc.b	$47,$FB,$5D,$E5,$9F,$F2,$F9,$FF,$14,$4F,$F2,$56,$BF,$C3,$1F,$0F
	.dc.b	$16,$6B,$8E,$5E,$EF,$03,$FA,$4E,$1C,$DB,$1E,$2A,$E0,$7E,$AE,$FA
	.dc.b	$ED,$BC,$71,$8F,$B6,$3C,$8F,$68,$F4,$CF,$93,$1E,$11,$E5,$9D,$3E
	.dc.b	$DB,$F5,$63,$E2,$8E,$2D,$74,$71,$8D,$B0,$C1,$2B,$B9,$31,$F6,$46
	.dc.b	$5F,$84,$F1,$8F,$4C,$FC,$08,$F9,$55,$E1,$C3,$D5,$7C,$AF,$75,$FC
	.dc.b	$D2,$7E,$2C,$F4,$8C,$61,$F1,$63,$81,$9A,$F2,$4F,$C6,$9F,$87,$5F
	.dc.b	$62,$3A,$5F,$F7,$EC,$7E,$49,$E0,$26,$1D,$DC,$BE,$E7,$CB,$87,$6A
	.dc.b	$5F,$A9,$F0,$63,$EF,$AE,$31,$F4,$0F,$BB,$11,$F1,$9F,$83,$C3,$C9
	.dc.b	$87,$73,$DB,$1C,$4F,$92,$7C,$BC,$3C,$EE,$6F,$BB,$F7,$1C,$BE,$6A
	.dc.b	$EA,$EE,$BB,$3E,$83,$A9,$2B,$DF,$1A,$8E,$47,$9C,$F5,$57,$6C,$3E
	.dc.b	$BC,$75,$3A,$47,$27,$D5,$5F,$26,$7E,$A3,$C6,$FC,$B3,$D8,$FB,$F8
	.dc.b	$76,$C3,$EA,$12,$3A,$BF,$19,$E1,$5D,$71,$FC,$09,$F8,$78,$F2,$1D
	.dc.b	$71,$F4,$46,$63,$15,$D5,$F9,$4F,$5C,$7E,$F4,$76,$8C,$4F,$E3,$B1
	.dc.b	$FC,$E1,$F9,$D7,$C3,$8F,$C7,$AF,$BE,$7B,$09,$8F,$C3,$AE,$13,$FE
	.dc.b	$44,$E1,$FF,$47,$FC,$A7,$F2,$F6,$DF,$FA,$A2,$31,$C2,$69,$23,$44
	.dc.b	$3F,$CA,$F3,$FE,$93,$F9,$6F,$F2,$FD,$FF,$9E,$1B,$C3,$FE,$C0,$6D
	.dc.b	$7F,$A9,$70,$EC,$FF,$9D,$3E,$63,$38,$39,$D1,$86,$CF,$FA,$A7,$A1
	.dc.b	$E2,$E6,$A4,$F3,$35,$1D,$3A,$DF,$13,$49,$3F,$E9,$4F,$80,$AF,$FA
	.dc.b	$4A,$2E,$DF,$E9,$CE,$E5,$79,$B5,$6D,$49,$87,$FA,$93,$C0,$AF,$FA
	.dc.b	$BA,$2E,$1F,$EB,$EF,$69,$E1,$7F,$AC,$6B,$FD,$2B,$F3,$1E,$F2,$3F
	.dc.b	$EB,$DB,$F5,$74,$48,$FF,$54,$BE,$3F,$43,$11,$FE,$AA,$8B,$8F,$FA
	.dc.b	$D3,$35,$7C,$BF,$84,$97,$FE,$A4,$7E,$8F,$91,$EC,$8C,$10,$FF,$57
	.dc.b	$C6,$B1,$FA,$FF,$30,$F8,$EF,$C6,$36,$BF,$D2,$BD,$08,$C4,$7E,$21
	.dc.b	$AF,$FA,$9A,$35,$3C,$89,$1C,$A7,$6F,$FA,$AF,$93,$F1,$2F,$FB,$23
	.dc.b	$AB,$EF,$1E,$4A,$FC,$47,$C4,$49,$FF,$60,$BF,$8E,$E5,$FF,$51,$46
	.dc.b	$08,$7F,$B3,$23,$5E,$07,$E2,$7C,$59,$3F,$ED,$87,$F2,$FF,$2E,$F9
	.dc.b	$8C,$CF,$FB,$6A,$6F,$1F,$FB,$92,$B5,$3F,$F0,$05,$7B,$BF,$B3,$6C
	.dc.b	$7C,$86,$BF,$AE,$6B,$F1,$DA,$FF,$B3,$5A,$B7,$46,$1B,$3F,$C0,$7B
	.dc.b	$DF,$23,$FA,$D2,$5F,$96,$7F,$51,$F1,$B2,$3F,$DE,$EF,$ED,$BD,$87
	.dc.b	$CC,$66,$7F,$DF,$33,$73,$78,$9F,$EE,$1B,$DD,$7F,$BC,$BE,$63,$F7
	.dc.b	$84,$7F,$C5,$75,$C1,$F5,$79,$DF,$ED,$57,$81,$0F,$F8,$03,$3F,$C2
	.dc.b	$23,$50,$9E,$5A,$BD,$C7,$07,$F2,$16,$D7,$E5,$9A,$FF,$1C,$6B,$FE
	.dc.b	$FA,$F7,$5F,$30,$F8,$BF,$E3,$71,$86,$C6,$67,$D7,$AF,$FC,$01,$BE
	.dc.b	$B9,$79,$3E,$03,$F8,$9D,$68,$FB,$0E,$1F,$55,$7F,$93,$B3,$4E,$6B
	.dc.b	$E3,$65,$E1,$BE,$55,$AF,$23,$F9,$5D,$EF,$0E,$4D,$A5,$F1,$7B,$9F
	.dc.b	$89,$F8,$1F,$82,$F8,$5A,$E7,$5F,$E9,$51,$8C,$7F,$9C,$7C,$C7,$F3
	.dc.b	$72,$57,$92,$7E,$0E,$3E,$BF,$84,$7B,$1F,$D4,$DB,$5F,$C2,$1A,$FF
	.dc.b	$4E,$6B,$FC,$1F,$A1,$F2,$FF,$D1,$49,$1F,$0A,$78,$EF,$F8,$2A,$DC
	.dc.b	$FC,$66,$BF,$03,$3F,$EA,$17,$FD,$4A,$F8,$E3,$F4,$6F,$78,$73,$AE
	.dc.b	$F8,$71,$BD,$BF,$B1,$F8,$77,$77,$F6,$F6,$C7,$F2,$46,$BF,$DA,$9A
	.dc.b	$FB,$7E,$63,$FB,$01,$1F,$BB,$1C,$30,$FD,$34,$7C,$2F,$07,$FD,$4F
	.dc.b	$8F,$F2,$68,$FF,$4A,$B5,$D1,$FE,$AF,$70,$48,$E2,$7F,$AC,$FE,$35
	.dc.b	$F3,$1F,$C1,$AB,$F2,$D1,$8E,$FF,$EF,$6D,$8F,$CD,$1D,$8F,$C5,$8F
	.dc.b	$94,$7C,$BF,$FA,$F1,$B1,$FE,$C1,$6B,$FE,$AC,$23,$87,$07,$38,$E4
	.dc.b	$47,$93,$FE,$CB,$C9,$F3,$1F,$D4,$1F,$E1,$86,$32,$FF,$09,$B6,$BD
	.dc.b	$37,$E2,$8D,$D7,$1E,$1F,$E2,$AD,$AF,$E1,$CD,$7C,$87,$C2,$F0,$3F
	.dc.b	$74,$FA,$1A,$FB,$5A,$FF,$B2,$9A,$9A,$4E,$AF,$F7,$1B,$C3,$D2,$7C
	.dc.b	$C3,$E6,$3F,$C5,$B1,$86,$C6,$6F,$C9,$7F,$E3,$2F,$7F,$EA,$61,$FE
	.dc.b	$F1,$6C,$FC,$E0,$E1,$FA,$2D,$7E,$B3,$F7,$7D,$3E,$5A,$9A,$49,$FF
	.dc.b	$7F,$7C,$C7,$E9,$3B,$EF,$47,$F1,$29,$F0,$4E,$5B,$19,$7F,$03,$0F
	.dc.b	$47,$F9,$FC,$7E,$1E,$31,$F0,$5D,$A7,$33,$FE,$F6,$9C,$BE,$AF,$95
	.dc.b	$FE,$B3,$C5,$EF,$D1,$18,$9D,$79,$5F,$AE,$7F,$DF,$F3,$A8,$C4,$7C
	.dc.b	$69,$7C,$C6,$2B,$9B,$63,$37,$DD,$B5,$F9,$CE,$37,$CC,$23,$F2,$F5
	.dc.b	$67,$31,$81,$A3,$E6,$3F,$CA,$B5,$FD,$2E,$B6,$FF,$4D,$9D,$7B,$BE
	.dc.b	$86,$C7,$DF,$6D,$FE,$B9,$AF,$EC,$DA,$F9,$5A,$F8,$AF,$C2,$F1,$6B
	.dc.b	$FB,$BB,$FF,$6E,$F9,$7E,$96,$DF,$F6,$1E,$9F,$A3,$E6,$67,$E1,$3F
	.dc.b	$B7,$4F,$AD,$B8,$7F,$56,$BE,$2F,$F5,$8B,$E3,$5B,$C3,$E3,$46,$27
	.dc.b	$F8,$4B,$5F,$F4,$BE,$AC,$7E,$DC,$E6,$7F,$C2,$73,$9B,$FD,$24,$E6
	.dc.b	$7F,$C1,$D3,$97,$AB,$5F,$E3,$5D,$CF,$98,$7C,$47,$D9,$8C,$3F,$0F
	.dc.b	$E5,$DF,$31,$9A,$DE,$5F,$6E,$72,$FE,$D7,$54,$63,$1F,$C5,$35,$FA
	.dc.b	$0D,$7F,$14,$D5,$1C,$AF,$F3,$6D,$AE,$D5,$BF,$2B,$FD,$69,$F3,$1F
	.dc.b	$91,$6C,$FF,$54,$F9,$8C,$79,$8A,$D7,$94,$9F,$D3,$DA,$31,$7F,$D8
	.dc.b	$E7,$31,$8C,$FF,$D9,$B8,$5E,$0F,$EF,$55,$E1,$F3,$3F,$CD,$38,$7F
	.dc.b	$7B,$8F,$09,$1F,$F0,$17,$AD,$70,$FE,$FD,$87,$39,$FE,$F9,$5C,$EB
	.dc.b	$58,$7F,$92,$30,$EB,$87,$F8,$46,$7D,$31,$88,$FF,$4C,$35,$FF,$03
	.dc.b	$7C,$C7,$C2,$3E,$37,$A0,$F9,$6C,$65,$FE,$23,$7E,$1C,$66,$3F,$D4
	.dc.b	$ED,$AF,$F7,$6B,$5F,$F5,$37,$A9,$FE,$5A,$9F,$F1,$FB,$63,$FC,$73
	.dc.b	$E2,$7E,$6F,$E3,$FE,$B4,$E7,$4F,$F2,$87,$AB,$83,$F8,$55,$6F,$0E
	.dc.b	$71,$8E,$A6,$14,$FC,$07,$AD,$E5,$C3,$FC,$A3,$3A,$BF,$F2,$8C,$F9
	.dc.b	$F2,$FF,$30,$69,$F8,$08,$C4,$7F,$A5,$5B,$C7,$FE,$FF,$F9,$8F,$C4
	.dc.b	$9F,$8D,$C3,$FD,$AF,$87,$F8,$DE,$32,$D3,$FC,$BF,$1A,$BF,$BB,$3F
	.dc.b	$95,$F8,$07,$CC,$53,$F3,$1E,$E7,$A7,$0F,$F6,$54,$65,$FF,$34,$CF
	.dc.b	$07,$FC,$D7,$3B,$BF,$F5,$64,$E5,$FC,$5F,$CC,$7E,$80,$FD,$27,$CC
	.dc.b	$3E,$5B,$18,$7F,$9D,$23,$0F,$FB,$91,$B5,$FE,$1C,$D1,$18,$8F,$F7
	.dc.b	$23,$5F,$F4,$13,$5F,$F5,$F3,$63,$8D,$7E,$BA,$31,$87,$6D,$FF,$D4
	.dc.b	$6D,$C3,$9F,$C6,$FC,$7B,$91,$F0,$9F,$5C,$63,$1F,$D4,$FC,$C7,$EF
	.dc.b	$8F,$E0,$5F,$0A,$F9,$1C,$61,$FE,$6D,$9B,$5F,$E6,$DA,$D4,$E4,$F9
	.dc.b	$FF,$31,$F5,$CF,$E1,$9F,$1B,$E5,$AF,$9F,$18,$CB,$F8,$4C,$FB,$C8
	.dc.b	$E0,$95,$FE,$19,$D9,$3A,$8C,$77,$BE,$63,$F8,$B9,$FC,$6F,$E3,$7F
	.dc.b	$C3,$13,$FE,$85,$8C,$B4,$FF,$56,$C6,$B2,$FE,$17,$3F,$8F,$F9,$89
	.dc.b	$5F,$EB,$2C,$3F,$C1,$5C,$3F,$42,$73,$7F,$E8,$FE,$14,$7C,$0F,$53
	.dc.b	$35,$FE,$97,$F9,$4F,$E5,$67,$C4,$F9,$3B,$97,$57,$0F,$F7,$FF,$93
	.dc.b	$E6,$E4,$99,$7F,$A7,$73,$57,$F7,$1A,$DC,$E6,$3F,$51,$DD,$2F,$FD
	.dc.b	$97,$C7,$FE,$B6,$7F,$4B,$19,$AF,$53,$F8,$2F,$88,$C6,$7F,$EA,$FE
	.dc.b	$DF,$E8,$67,$F4,$7D,$3F,$AD,$DF,$F3,$28,$CE,$1F,$EB,$37,$5E,$76
	.dc.b	$27,$F9,$BC,$FF,$4C,$3F,$0B,$E0,$D7,$E7,$3F,$84,$23,$F8,$1D,$BF
	.dc.b	$92,$4F,$FA,$2F,$97,$F0,$75,$A8,$FD,$84,$E9,$FF,$77,$35,$FF,$47
	.dc.b	$FC,$C7,$F8,$CF,$E5,$BF,$CB,$91,$FE,$23,$C3,$1E,$F7,$F9,$9E,$AC
	.dc.b	$E4,$57,$07,$FD,$39,$1F,$E3,$DF,$90,$C4,$7F,$8E,$AF,$17,$CE,$13
	.dc.b	$AF,$61,$5A,$9F,$29,$DA,$BF,$C7,$11,$E2,$8E,$E6,$4F,$26,$1E,$05
	.dc.b	$86,$DF,$11,$B3,$8C,$78,$67,$51,$C0,$F2,$D7,$81,$A7,$93,$0E,$0E
	.dc.b	$0F,$1B,$83,$C5,$7E,$B8,$FF,$26,$55,$7F,$CC,$51,$E5,$C3,$E0,$C7
	.dc.b	$94,$C9,$5C,$E1,$D2,$3C,$07,$B2,$F0,$78,$70,$F9,$CD,$8F,$44,$7A
	.dc.b	$E3,$94,$FA,$8B,$1E,$F7,$9B,$B3,$99,$67,$99,$F8,$39,$DB,$63,$EF
	.dc.b	$4F,$56,$9D,$4E,$71,$B3,$99,$F1,$8F,$69,$F4,$E7,$D4,$FC,$E8,$F1
	.dc.b	$3F,$69,$E8,$FE,$22,$7C,$0F,$43,$E3,$47,$81,$ED,$3E,$86,$C6,$5D
	.dc.b	$9C,$DE,$AF,$73,$F1,$EE,$EB,$EC,$CF,$39,$E2,$79,$CF,$71,$4E,$E7
	.dc.b	$53,$99,$4F,$85,$5C,$E7,$AD,$77,$C3,$F1,$55,$F9,$7A,$F7,$BE,$53
	.dc.b	$8C,$FD,$07,$A9,$93,$F4,$71,$F8,$07,$27,$E4,$1E,$6D,$8F,$82,$D3
	.dc.b	$67,$85,$CB,$4C,$E1,$F9,$DA,$FD,$6B,$63,$F3,$C7,$DD,$9F,$B6,$FE
	.dc.b	$83,$0F,$BE,$F7,$3D,$67,$C5,$3D,$31,$ED,$72,$7B,$70,$FD,$3C,$7F
	.dc.b	$0F,$8F,$F8,$00,$7E,$F4,$B5,$87,$F5,$E7,$FB,$54,$FF,$6B,$7A,$EF
	.dc.b	$98,$7B,$BC,$A3,$0D,$8C,$CF,$13,$8B,$87,$12,$F9,$87,$1F,$36,$CF
	.dc.b	$11,$3B,$9D,$12,$3F,$8D,$47,$F8,$55,$FE,$48,$D7,$33,$D3,$5F,$E6
	.dc.b	$D3,$CE,$78,$F8,$5E,$F9,$CB,$73,$F8,$55,$B9,$E3,$7E,$6D,$77,$87
	.dc.b	$F1,$DA,$F0,$46,$B1,$C5,$78,$70,$E7,$8F,$5C,$3E,$6B,$5F,$7C,$ED
	.dc.b	$FA,$2D,$7E,$1C,$6F,$0F,$76,$3D,$E7,$FA,$17,$0F,$F9,$F3,$E4,$3E
	.dc.b	$D6,$5F,$CB,$23,$12,$3E,$8F,$05,$8C,$C6,$03,$FE,$86,$F8,$FE,$73
	.dc.b	$BE,$86,$18,$BD,$C6,$1E,$D8,$78,$63,$11,$D7,$0E,$8E,$AF,$C5,$E0
	.dc.b	$BE,$8F,$A5,$23,$15,$B3,$CA,$E1,$FF,$60,$FD,$27,$08,$C6,$3F,$05
	.dc.b	$B8,$76,$C3,$55,$E9,$FC,$2F,$E2,$7A,$1F,$A6,$6E,$3F,$93,$6B,$F9
	.dc.b	$D6,$BF,$A7,$7F,$C0,$0F,$95,$AF,$CB,$6E,$3F,$D7,$9E,$6D,$87,$C7
	.dc.b	$78,$5A,$78,$4A,$78,$4A,$78,$4A,$7D,$D9,$FB,$D1,$CC,$91,$CF,$E0
	.dc.b	$BE,$02,$69,$88,$CF,$BB,$E3,$24,$F8,$DA,$F8,$DA,$F8,$DF,$57,$81
	.dc.b	$D8,$91,$F1,$44,$78,$BE,$7B,$E0,$12,$3E,$03,$5F,$80,$D7,$E0,$3F
	.dc.b	$1F,$97,$D6,$48,$ED,$1E,$EE,$6F,$98,$48,$F9,$8D,$7E,$63,$5F,$98
	.dc.b	$FD,$3E,$9F,$8E,$48,$F5,$47,$CA,$FA,$FF,$B0,$48,$FB,$0D,$7E,$C3
	.dc.b	$5F,$B0,$FD,$FF,$BB,$FA,$64,$8F,$8D,$1F,$5B,$DB,$FC,$21,$23,$F0
	.dc.b	$8D,$7F,$08,$D7,$F0,$8F,$E2,$FE,$F3,$EF,$92,$3E,$94,$7E,$0B,$EB
	.dc.b	$24,$7D,$57,$F0,$BF,$23,$F8,$6F,$87,$FC,$A1,$27,$F2,$8F,$E5,$FE
	.dc.b	$87,$F1,$64,$8F,$BD,$F2,$9F,$BA,$FD,$A5,$E4,$7F,$52,$ED,$46,$2B
	.dc.b	$FD,$78,$D7,$FD,$D0,$D7,$FC,$FC,$D7,$F8,$C3,$5F,$E2,$ED,$8E,$B3
	.dc.b	$FA,$BF,$9C,$FC,$F9,$2B,$F8,$54,$E9,$FC,$A7,$E0,$3F,$A4,$23,$FB
	.dc.b	$BF,$1F,$FA,$E7,$5F,$ED,$E7,$33,$FB,$D0,$FF,$0B,$9D,$F7,$FE,$CD
	.dc.b	$79,$39,$25,$75,$FC,$0E,$BF,$C1,$E7,$6F,$86,$7A,$4F,$CB,$F2,$24
	.dc.b	$FA,$B5,$FF,$49,$65,$FC,$FD,$7C,$B6,$A7,$F9,$84,$6B,$2F,$93,$7F
	.dc.b	$56,$BD,$1F,$45,$2F,$E0,$3C,$70,$FE,$31,$9B,$E8,$BD,$C4,$AF,$8D
	.dc.b	$75,$F5,$45,$AE,$33,$F1,$70,$95,$F3,$5F,$91,$F4,$38,$8F,$85,$16
	.dc.b	$BC,$33,$F8,$BD,$3F,$15,$A1,$A4,$BF,$C5,$B5,$FC,$5B,$5F,$E7,$19
	.dc.b	$7E,$3E,$FD,$BC,$3F,$64,$ED,$3F,$81,$F9,$9F,$B8,$47,$F0,$9F,$2B
	.dc.b	$5F,$B1,$16,$BE,$3C,$FE,$0F,$D6,$95,$F8,$B7,$FA,$EF,$3B,$9B,$33
	.dc.b	$F8,$19,$B7,$FD,$E6,$B5,$3F,$98,$5D,$5F,$AB,$6C,$7F,$8A,$79,$91
	.dc.b	$88,$FF,$1A,$46,$27,$81,$23,$84,$ED,$FF,$76,$35,$FC,$D7,$C0,$7E
	.dc.b	$74,$8F,$8E,$7F,$41,$EC,$C8,$FD,$43,$FD,$EB,$DE,$7C,$16,$67,$F3
	.dc.b	$D3,$72,$FF,$0B,$56,$A7,$F6,$1E,$BF,$EA,$C8,$FE,$E3,$E0,$DC,$18
	.dc.b	$8F,$D4,$4D,$BF,$F1,$05,$6A,$7F,$75,$F0,$FF,$DF,$78,$CD,$24,$7F
	.dc.b	$94,$7E,$A1,$F8,$13,$E8,$7F,$C3,$D1,$86,$C6,$67,$A9,$FC,$1F,$ED
	.dc.b	$7F,$81,$11,$FF,$0C,$57,$17,$44,$3F,$80,$47,$4A,$FE,$15,$F1,$3E
	.dc.b	$72,$4F,$F8,$4F,$14,$7A,$F6,$FE,$20,$49,$FE,$31,$1F,$E3,$57,$FC
	.dc.b	$E1,$EC,$FF,$9F,$FC,$B3,$49,$1F,$89,$F6,$1E,$B7,$F8,$C9,$B1,$FC
	.dc.b	$A1,$AF,$F9,$1C,$EA,$66,$38,$11,$FF,$1C,$1A,$3F,$C8,$51,$CB,$83
	.dc.b	$FC,$8D,$3B,$9E,$97,$E6,$7F,$C8,$19,$3F,$37,$FF,$0C,$47,$FA,$1E
	.dc.b	$31,$EA,$FF,$A7,$9B,$19,$9D,$BF,$E3,$98,$7C,$47,$F3,$72,$47,$69
	.dc.b	$DE,$3F,$C8,$3E,$17,$F1,$04,$9F,$A4,$D7,$A3,$5F,$E1,$AF,$E3,$F3
	.dc.b	$FE,$88,$48,$FE,$99,$1F,$E5,$EF,$DA,$5F,$15,$FD,$95,$B3,$FB,$B6
	.dc.b	$BF,$BA,$6B,$FE,$B7,$EE,$46,$23,$FC,$F8,$D7,$FC,$F6,$D8,$DC,$FF
	.dc.b	$97,$F0,$79,$5F,$E6,$36,$DF,$F9,$6D,$AF,$F9,$E6,$FC,$71,$FE,$5B
	.dc.b	$32,$59,$E4,$7B,$6B,$DD,$F1,$9E,$D2,$5F,$B5,$AF,$B5,$AF,$B6,$D1
	.dc.b	$E7,$FA,$9F,$31,$27,$FB,$DC,$E9,$F4,$7D,$4F,$B4,$8F,$F9,$EF,$14
	.dc.b	$7C,$1F,$9F,$93,$96,$BF,$ED,$3F,$9F,$7D,$03,$E0,$7F,$DD,$71,$86
	.dc.b	$C6,$67,$83,$5F,$F7,$24,$AB,$FB,$EC,$61,$AF,$FB,$C6,$35,$E9,$76
	.dc.b	$9C,$CB,$E0,$7F,$DD,$37,$F5,$23,$FD,$01,$97,$9B,$7F,$74,$E6,$78
	.dc.b	$4E,$5F,$F4,$D2,$FA,$4D,$5F,$FB,$D1,$B1,$9B,$F9,$11,$8F,$85,$FC
	.dc.b	$2C,$FE,$05,$B3,$FE,$CB,$F9,$07,$9D,$8C,$7C,$35,$79,$15,$AB,$FE
	.dc.b	$B0,$DA,$FE,$AE,$D7,$F3,$CD,$7B,$5F,$B6,$B9,$B6,$75,$3F,$ED,$A6
	.dc.b	$E4,$C7,$F9,$0E,$1F,$9D,$ED,$F2,$57,$F4,$C8,$C1,$2F,$FD,$75,$5C
	.dc.b	$1F,$F6,$D4,$ED,$CD,$72,$F1,$FC,$44,$BF,$6B,$67,$DA,$D7,$DA,$FD
	.dc.b	$0F,$A6,$F9,$18,$F9,$DB,$8F,$F5,$3B,$FD,$A4,$63,$C2,$FA,$E4,$8F
	.dc.b	$CF,$37,$0F,$BE,$D7,$FD,$B1,$5C,$DF,$C1,$FC,$DF,$63,$EC,$77,$3F
	.dc.b	$34,$FE,$2B,$87,$F0,$39,$CC,$62,$84,$15,$C7,$AF,$F2,$4D,$BF,$E7
	.dc.b	$15,$C6,$BF,$8D,$7C,$9B,$CC,$FC,$33,$6B,$FB,$4F,$DC,$BD,$DC,$D4
	.dc.b	$9F,$F3,$DD,$6E,$FE,$F4,$79,$3C,$A6,$1C,$63,$77,$C6,$B6,$39,$58
	.dc.b	$73,$DB,$F9,$1E,$3C,$B2,$E1,$18,$CB,$A6,$1F,$04,$56,$F2,$FF,$06
	.dc.b	$4E,$A7,$E3,$4E,$5F,$D9,$FC,$E7,$DE,$3E,$FF,$CD,$DB,$FE,$64,$FE
	.dc.b	$D0,$8F,$F1,$B7,$F6,$FE,$17,$E1,$8F,$D8,$5F,$F8,$6B,$E3,$31,$5F
	.dc.b	$CE,$E6,$E1,$F0,$2B,$53,$93,$FC,$2F,$E7,$2F,$9B,$63,$FA,$27,$A5
	.dc.b	$89,$FC,$AB,$FC,$20,$91,$FC,$BD,$FE,$17,$C7,$F9,$A3,$F8,$17,$A3
	.dc.b	$FD,$CA,$7F,$A7,$46,$5A,$7F,$8F,$FC,$AC,$65,$FB,$2E,$CF,$D1,$9F
	.dc.b	$C2,$FC,$FF,$F0,$BD,$7F,$57,$8C,$B4,$FF,$5F,$C6,$A7,$94,$7F,$5D
	.dc.b	$F9,$4F,$D6,$9F,$C5,$7F,$6B,$7C,$46,$27,$FD,$8F,$18,$7F,$CD,$CD
	.dc.b	$AF,$F3,$6B,$5F,$EE,$8D,$7F,$4B,$CE,$8C,$47,$E3,$DB,$19,$AF,$E3
	.dc.b	$91,$AE,$EF,$E3,$1F,$2B,$FC,$08,$F3,$7C,$5D,$C7,$FB,$B4,$5C,$7F
	.dc.b	$37,$8F,$7A,$C7,$47,$F7,$7E,$2F,$E1,$67,$F0,$CF,$4E,$E3,$FD,$FE
	.dc.b	$2E,$1F,$C8,$26,$D7,$F8,$13,$27,$D0,$7F,$63,$C3,$F8,$0C,$E6,$31
	.dc.b	$7E,$2D,$1F,$29,$FC,$72,$73,$5F,$E0,$ED,$3E,$24,$FC,$02,$61,$F3
	.dc.b	$9B,$87,$CE,$6B,$F3,$9F,$EB,$37,$DA,$7C,$67,$F1,$4C,$7C,$F3,$E1
	.dc.b	$F9,$1F,$13,$27,$77,$FD,$B6,$F9,$3F,$C1,$2F,$DF,$7B,$8C,$4B,$D5
	.dc.b	$F6,$1E,$DF,$90,$B8,$74,$C3,$E3,$CE,$2F,$8D,$5C,$FE,$47,$C8,$FC
	.dc.b	$53,$FB,$37,$ED,$6E,$2C,$65,$E6,$D9,$39,$8C,$37,$1F,$C0,$AE,$4E
	.dc.b	$77,$2B,$6F,$68,$C5,$7B,$38,$FE,$B1,$F5,$FE,$4A,$DF,$AF,$74,$E2
	.dc.b	$BE,$1D,$5D,$3F,$05,$ED,$7D,$F3,$F0,$1F,$25,$71,$E5,$16,$FE,$5C
	.dc.b	$DA,$FC,$57,$C9,$7E,$1C,$F8,$7D,$57,$0F,$7B,$FE,$01,$DB,$FC,$17
	.dc.b	$1F,$EE,$16,$E1,$FE,$E0,$6B,$FE,$67,$6B,$FB,$46,$BF,$D0,$5B,$BF
	.dc.b	$CE,$AF,$C6,$F4,$FE,$64,$FC,$D7,$C0,$DC,$3E,$27,$C9,$7E,$48,$8F
	.dc.b	$F4,$3C,$7F,$19,$1F,$96,$F9,$2F,$D1,$1F,$A3,$F9,$2B,$5F,$37,$E1
	.dc.b	$BF,$4C,$7E,$9F,$E4,$AC,$7D,$17,$F3,$E4,$7B,$75,$7E,$D2,$31,$8F
	.dc.b	$F5,$38,$C4,$EB,$95,$5A,$AF,$EF,$15,$FB,$CF,$89,$FD,$B1,$FB,$7F
	.dc.b	$32,$E1,$F8,$17,$F5,$E4,$7A,$36,$BC,$18,$7F,$5F,$7F,$64,$D7,$F5
	.dc.b	$D1,$FB,$8F,$92,$FE,$08,$7F,$88,$3D,$EF,$F0,$5D,$FE,$3E,$31,$5F
	.dc.b	$E5,$69,$CB,$ED,$6B,$EA,$6B,$FC,$05,$AF,$F9,$A5,$B8,$FF,$9D,$EB
	.dc.b	$E4,$C6,$07,$BB,$E3,$3C,$9F,$19,$67,$F7,$92,$8F,$21,$23,$9B,$6F
	.dc.b	$CC,$D7,$CC,$FA,$FE,$4B,$E0,$1F,$07,$E4,$AD,$76,$7E,$01,$1F,$3B
	.dc.b	$6B,$E1,$35,$F8,$4F,$C7,$F9,$2F,$98,$7C,$DF,$92,$B5,$EA,$7E,$61
	.dc.b	$1F,$86,$DA,$F9,$CD,$7E,$73,$F4,$FE,$4B,$EC,$1F,$67,$E4,$6D,$7C
	.dc.b	$67,$EC,$11,$F9,$EF,$DB,$F7,$BF,$00,$7E,$07,$E0,$2C,$FC,$E8,$B8
	.dc.b	$7F,$33,$D7,$FD,$0D,$DF,$FD,$4D,$FF,$39,$37,$E4,$7F,$14,$7E,$2F
	.dc.b	$E4,$6E,$1F,$62,$2D,$F3,$9B,$5F,$EA,$7F,$96,$FF,$2D,$FC,$AB,$6F
	.dc.b	$F1,$E7,$1F,$FB,$92,$31,$87,$F8,$FE,$B7,$5A,$CA,$CE,$45,$70,$7F
	.dc.b	$CC,$F7,$CE,$BF,$83,$5F,$09,$CD,$D7,$FC,$DB,$1F,$E6,$EF,$3B,$94
	.dc.b	$7E,$5A,$71,$5C,$A3,$FB,$7B,$E3,$69,$E3,$C3,$8E,$1E,$18,$C4,$EE
	.dc.b	$7A,$C3,$0F,$15,$6E,$3F,$CF,$D1,$8B,$F3,$46,$CF,$E4,$07,$A0,$A6
	.dc.b	$4C,$C6,$BE,$59,$E3,$7F,$A2,$27,$F9,$44,$63,$E0,$75,$39,$9F,$F3
	.dc.b	$74,$E5,$DF,$91,$FE,$BD,$6E,$BD,$63,$11,$C1,$AF,$FA,$5E,$51,$C7
	.dc.b	$3F,$AD,$0F,$27,$FD,$39,$18,$6E,$BC,$6B,$83,$70,$FB,$57,$C7,$0B
	.dc.b	$7C,$1F,$F5,$3C,$78,$27,$77,$E9,$DB,$B5,$7D,$C3,$77,$FE,$AD,$8F
	.dc.b	$2C,$EF,$0F,$05,$7E,$39,$F6,$57,$78,$61,$E3,$74,$78,$AB,$D2,$DC
	.dc.b	$3C,$EC,$36,$72,$8D,$F8,$EF,$8F,$7C,$3F,$FB,$06,$FF,$C7,$91,$8B
	.dc.b	$FF,$0C,$35,$FF,$2A,$AF,$87,$E1,$C3,$FE,$66,$8D,$57,$2B,$CD,$FF
	.dc.b	$AD,$67,$35,$BB,$CE,$0F,$96,$FF,$4C,$5F,$F8,$9D,$AF,$F9,$91,$B8
	.dc.b	$FF,$94,$57,$AD,$C9,$CE,$1F,$E7,$86,$D7,$FB,$15,$B7,$DE,$BC,$BE
	.dc.b	$07,$FB,$8B,$A3,$F4,$2D,$BF,$EE,$A2,$32,$D3,$3C,$3C,$9F,$F3,$39
	.dc.b	$FE,$49,$29,$FE,$53,$38,$17,$D5,$E6,$47,$BB,$5F,$F0,$10,$70,$FF
	.dc.b	$80,$0C,$1E,$6F,$57,$C4,$47,$C5,$1E,$5F,$57,$D2,$48,$F6,$B5,$EE
	.dc.b	$D7,$B3,$F0,$3D,$5F,$71,$23,$DD,$1F,$07,$D3,$F8,$A4,$8F,$92,$D7
	.dc.b	$D6,$D7,$D4,$FC,$CE,$5F,$94,$48,$F9,$51,$F3,$79,$3E,$89,$23,$EA
	.dc.b	$B5,$F8,$ED,$7E,$33,$F6,$3A,$3E,$B1,$23,$EB,$47,$D9,$F4,$FE,$E9
	.dc.b	$23,$F0,$2D,$7E,$9B,$5F,$A4,$FE,$13,$D1,$FC,$11,$23,$F0,$51,$F8
	.dc.b	$5F,$47,$F1,$24,$8F,$C6,$B5,$FB,$ED,$7E,$F3,$F9,$0F,$47,$F1,$C4
	.dc.b	$8F,$C7,$47,$E4,$7D,$19,$1F,$91,$7F,$29,$E8,$FE,$57,$D1,$FC,$C9
	.dc.b	$27,$F3,$AF,$E7,$3D,$1F,$CD,$12,$3F,$35,$E8,$FE,$DB,$E5,$9E,$57
	.dc.b	$F4,$6E,$A4,$62,$BF,$CD,$11,$89,$D4,$E6,$44,$13,$B7,$8C,$6D,$D4
	.dc.b	$65,$E0,$70,$8D,$3F,$EC,$71,$E8,$FE,$CC,$8F,$FA,$C6,$3C,$5D,$DF
	.dc.b	$EE,$38,$7A,$3F,$B9,$23,$FE,$C1,$F1,$BB,$CE,$E7,$33,$FB,$B7,$F8
	.dc.b	$4C,$EB,$C0,$FF,$6C,$3F,$E5,$29,$E5,$E4,$68,$DF,$A3,$63,$24,$3F
	.dc.b	$DB,$13,$FE,$55,$7D,$1E,$3F,$D1,$23,$EB,$F3,$EA,$CB,$FD,$E9,$9F
	.dc.b	$F8,$5D,$B5,$F6,$6B,$CF,$E8,$CB,$F7,$3C,$A9,$9F,$AB,$D1,$F6,$92
	.dc.b	$BE,$25,$D7,$CF,$16,$B8,$C3,$C4,$66,$BF,$8F,$94,$AF,$9C,$FC,$AF
	.dc.b	$47,$11,$F0,$A2,$E1,$D3,$0E,$75,$F9,$3D,$7F,$1A,$4B,$FC,$4B,$5F
	.dc.b	$C3,$F0,$FE,$5A,$31,$1F,$8C,$DB,$E7,$3F,$EA,$FC,$FF,$0B,$AF,$DA
	.dc.b	$E2,$FB,$A7,$79,$FC,$27,$A1,$F8,$02,$3F,$87,$F4,$6B,$F6,$A2,$EB
	.dc.b	$E9,$C1,$5F,$88,$F4,$65,$FE,$3D,$FE,$C7,$D0,$E7,$CC,$FE,$12,$6E
	.dc.b	$1F,$3B,$65,$EE,$BF,$5B,$39,$87,$5F,$ED,$DB,$3F,$B6,$6B,$F1,$9A
	.dc.b	$FE,$C1,$AF,$EE,$2B,$BC,$F3,$24,$7F,$B6,$B2,$7A,$3F,$A4,$24,$79
	.dc.b	$36,$E5,$87,$83,$1F,$89,$F0,$BF,$AA,$25,$6C,$FE,$0B,$ED,$3D,$EC
	.dc.b	$C7,$E9,$A6,$D7,$C3,$AD,$4F,$EE,$7E,$5B,$F6,$A4,$7A,$6A,$F7,$DC
	.dc.b	$59,$9F,$D8,$D5,$C3,$E7,$56,$A7,$F8,$27,$CB,$7F,$81,$B7,$34,$91
	.dc.b	$FC,$58,$7C,$13,$E0,$9F,$15,$FE,$24,$8C,$36,$33,$3D,$8F,$E2,$5F
	.dc.b	$2D,$FC,$38,$8F,$C5,$AE,$2E,$88,$7F,$0A,$8E,$95,$FC,$5B,$E5,$BD
	.dc.b	$04,$9F,$B0,$FA,$F7,$FE,$36,$48,$FE,$45,$1F,$E3,$A7,$FC,$DF,$EB
	.dc.b	$FF,$A1,$3C,$CF,$F2,$14,$7F,$A2,$1A,$FC,$16,$BF,$CE,$BD,$8F,$F1
	.dc.b	$BB,$FE,$4C,$F5,$FF,$C8,$31,$88,$FE,$35,$CF,$FD,$16,$B9,$F0,$FF
	.dc.b	$95,$2F,$64,$AF,$C3,$1E,$43,$A1,$0F,$F2,$BB,$E9,$9E,$75,$FD,$4B
	.dc.b	$81,$F2,$CF,$5B,$FD,$43,$18,$6C,$67,$0F,$B9,$39,$7F,$41,$F0,$BF
	.dc.b	$D0,$88,$FE,$66,$34,$FF,$3C,$F9,$6F,$C4,$11,$FD,$43,$5F,$E2,$4D
	.dc.b	$7F,$B3,$78,$BF,$CA,$A3,$F2,$1A,$FF,$50,$24,$FF,$57,$8F,$F3,$5F
	.dc.b	$CA,$BE,$23,$FB,$B3,$67,$FB,$AB,$5F,$E4,$AD,$7F,$B7,$B5,$FE,$00
	.dc.b	$D7,$FB,$BB,$7A,$CE,$70,$FB,$33,$FB,$96,$B8,$6F,$07,$F1,$99,$F1
	.dc.b	$3F,$CF,$4F,$E9,$33,$B6,$99,$DB,$DB,$87,$BB,$E2,$7D,$A4,$AF,$6B
	.dc.b	$5F,$6B,$5F,$6B,$FE,$7E,$A7,$CB,$7F,$A0,$9B,$3E,$96,$BE,$CC,$BD
	.dc.b	$F5,$BF,$96,$E4,$49,$FE,$01,$9F,$F7,$38,$ED,$F2,$B2,$73,$82,$31
	.dc.b	$1F,$C1,$FE,$59,$F2,$DF,$E8,$68,$FF,$13,$46,$3E,$0B,$FD,$EA,$D8
	.dc.b	$E1,$3B,$7F,$C5,$B1,$D7,$D3,$DD,$FF,$84,$E7,$FC,$53,$39,$7C,$11
	.dc.b	$97,$FC,$6B,$97,$F8,$2E,$FE,$9C,$63,$1F,$2E,$FE,$A9,$CC,$FF,$7F
	.dc.b	$9C,$BF,$E3,$6A,$FF,$0A,$7C,$B6,$9B,$7F,$0A,$31,$46,$FC,$2F,$F8
	.dc.b	$2A,$3F,$08,$D9,$FE,$A5,$F1,$5F,$DA,$9F,$AD,$19,$C3,$FC,$97,$39
	.dc.b	$7F,$C0,$2D,$7F,$C9,$6D,$79,$F9,$DE,$AE,$14,$E6,$79,$CE,$63,$13
	.dc.b	$A8,$F4,$F6,$F8,$73,$FF,$19,$3E,$39,$E7,$E3,$E7,$3F,$F7,$46,$7F
	.dc.b	$8D,$AC,$1B,$C7,$FA,$9C,$64,$FE,$AE,$E3,$2F,$5D,$F8,$C9,$3F,$E7
	.dc.b	$09,$D4,$FB,$A7,$2F,$09,$FC,$8F,$CB,$78,$CF,$27,$85,$63,$F4,$0F
	.dc.b	$D4,$23,$F9,$86,$CF,$DD,$C7,$FA,$DC,$E4,$91,$9F,$96,$F8,$07,$D5
	.dc.b	$EC,$7C,$B7,$F1,$5A,$FD,$74,$E6,$31,$5F,$E7,$EA,$DF,$57,$E3,$9B
	.dc.b	$7F,$CC,$AB,$75,$FE,$86,$F9,$67,$99,$F4,$5B,$5F,$D9,$FE,$5B,$FD
	.dc.b	$0A,$FF,$80,$A3,$35,$FB,$D6,$E7,$FE,$2D,$6C,$FF,$A6,$3C,$95,$F2
	.dc.b	$BE,$18,$77,$25,$77,$F7,$6C,$FF,$24,$BE,$31,$FA,$86,$CF,$F8,$1D
	.dc.b	$B3,$FE,$9E,$6B,$FD,$0B,$17,$CB,$7E,$2C,$FC,$67,$CB,$5B,$FE,$30
	.dc.b	$FE,$B0,$8F,$F1,$07,$F5,$FE,$07,$E4,$CF,$D4,$61,$FE,$15,$F9,$6C
	.dc.b	$57,$F2,$19,$B8,$FA,$7C,$E5,$63,$2F,$F3,$FE,$1C,$70,$FF,$0A,$56
	.dc.b	$9F,$E5,$5E,$97,$68,$C6,$1F,$CF,$63,$F8,$09,$27,$F9,$2B,$FC,$13
	.dc.b	$8B,$F5,$07,$EA,$FD,$17,$93,$8A,$FE,$6D,$36,$FF,$9B,$56,$A7,$27
	.dc.b	$F3,$BE,$9F,$D9,$1F,$B5,$F3,$DE,$76,$27,$44,$BF,$E8,$13,$A8,$C6
	.dc.b	$BF,$D5,$FE,$53,$F7,$87,$EF,$FE,$59,$F1,$18,$AF,$F2,$7C,$61,$FF
	.dc.b	$72,$B7,$0F,$F4,$AB,$5F,$F6,$FB,$5F,$F3,$83,$5F,$F7,$3B,$7C,$03
	.dc.b	$4D,$78,$3F,$EC,$3F,$96,$FE,$20,$79,$7E,$2E,$E1,$FD,$76,$2E,$1F
	.dc.b	$D7,$6D,$58,$CF,$FB,$A7,$2F,$F1,$93,$F8,$D6,$D7,$1F,$EC,$D1,$70
	.dc.b	$FE,$F1,$36,$BF,$D7,$1F,$2A,$F3,$3F,$93,$37,$0F,$8B,$F2,$38,$9F
	.dc.b	$94,$FF,$33,$CC,$D2,$4F,$CA,$6E,$3F,$29,$AF,$F8,$A2,$F8,$CF,$F8
	.dc.b	$4D,$B1,$FE,$13,$5F,$23,$E4,$A9,$5B,$7C,$F1,$8A,$E1,$EA,$FA,$8F
	.dc.b	$5F,$C8,$5B,$ED,$16,$7B,$4D,$CF,$E1,$FC,$8F,$C2,$3F,$A3,$7C,$B3
	.dc.b	$87,$18,$FF,$A5,$A3,$0D,$8D,$F2,$F0,$64,$64,$E1,$18,$C7,$DD,$C5
	.dc.b	$F4,$0F,$A3,$F2,$56,$BE,$24,$59,$F8,$93,$72,$FB,$3E,$D7,$D8,$3E
	.dc.b	$CF,$C9,$5C,$3E,$64,$5B,$F9,$93,$6B,$F0,$3F,$25,$F8,$03,$DF,$D7
	.dc.b	$70,$F7,$3F,$DC,$75,$FE,$E3,$1F,$EA,$9E,$94,$63,$1F,$F7,$DB,$51
	.dc.b	$FE,$D5,$70,$DA,$E8,$4A,$FF,$1A,$F2,$7E,$C7,$1F,$CA,$F5,$7E,$53
	.dc.b	$E0,$FF,$2A,$FE,$5B,$E4,$BF,$C3,$14,$69,$2F,$9F,$C9,$7E,$70,$FC
	.dc.b	$EF,$C9,$58,$FC,$87,$C2,$FE,$80,$FD,$0F,$C9,$58,$FC,$A3,$D8,$8F
	.dc.b	$99,$BA,$FF,$8A,$F1,$46,$23,$F3,$CF,$9B,$E2,$7F,$5A,$7E,$BB,$CC
	.dc.b	$B7,$FA,$17,$D6,$47,$E1,$B6,$BB,$35,$FD,$53,$F0,$BE,$4B,$F7,$27
	.dc.b	$F8,$7F,$C0,$FA,$35,$FA,$D8,$C5,$7F,$8F,$E7,$2F,$F9,$0D,$AF,$FA
	.dc.b	$71,$AF,$F9,$41,$AB,$C3,$FE,$10,$F3,$25,$7F,$9B,$A5,$EF,$78,$CF
	.dc.b	$27,$C6,$D9,$FF,$49,$4A,$3C,$84,$8E,$4D,$AF,$33,$5F,$33,$EB,$F9
	.dc.b	$2F,$80,$7C,$1F,$92,$B5,$D9,$F8,$04,$7C,$CD,$AF,$84,$D7,$E1,$3F
	.dc.b	$1F,$E4,$BE,$61,$F3,$7E,$4A,$D7,$A9,$F9,$84,$7E,$13,$6B,$E7,$35
	.dc.b	$F9,$CF,$D3,$F9,$2F,$B0,$7D,$9F,$91,$B5,$F1,$9F,$B0,$47,$E7,$3F
	.dc.b	$6F,$DE,$FC,$01,$F8,$1F,$80,$B3,$F3,$A2,$E1,$FE,$BF,$D7,$FC,$E5
	.dc.b	$D6,$AF,$BD,$5F,$89,$F9,$19,$1F,$5B,$E4,$6E,$1C,$66,$DF,$18,$FC
	.dc.b	$26,$0A,$CD,$FF,$02,$99,$F2,$DF,$C1,$3E,$59,$B6,$67,$F0,$F5,$98
	.dc.b	$C4,$2B,$0F,$3E,$3F,$05,$F2,$8E,$53,$A7,$F8,$8C,$EA,$B9,$CE,$67
	.dc.b	$B8,$AE,$4F,$98,$FE,$BB,$3E,$5A,$F7,$79,$DA,$9F,$E3,$B7,$FC,$7A
	.dc.b	$31,$3A,$AE,$F5,$D5,$E5,$87,$86,$3D,$86,$4F,$63,$9B,$F3,$57,$A6
	.dc.b	$7D,$A6,$CF,$6C,$ED,$F8,$E7,$90,$A7,$63,$33,$AF,$96,$78,$CF,$81
	.dc.b	$C3,$6B,$E8,$CE,$A7,$27,$1F,$21,$BD,$25,$70,$8C,$B6,$3E,$A5,$3C
	.dc.b	$9F,$C2,$6D,$F4,$23,$15,$AB,$E3,$39,$6E,$BF,$8B,$AD,$D7,$C3,$6D
	.dc.b	$6F,$0E,$47,$D5,$AE,$18,$7E,$4F,$6F,$03,$63,$A6,$3E,$89,$E5,$8F
	.dc.b	$5A,$F3,$D7,$28,$EA,$78,$6B,$CA,$FF,$60,$C7,$C7,$3B,$8D,$61,$F9
	.dc.b	$AF,$21,$F2,$DF,$E7,$89,$FE,$CF,$3F,$AB,$9C,$FA,$9F,$A9,$8C,$56
	.dc.b	$EB,$E8,$F0,$7F,$6E,$BE,$7E,$A6,$E3,$15,$BB,$E3,$86,$70,$61,$D9
	.dc.b	$F0,$DF,$0C,$35,$2F,$90,$7C,$36,$1B,$7C,$0E,$6E,$1F,$29,$8F,$86
	.dc.b	$E1,$32,$38,$57,$58,$E9,$3E,$5F,$79,$F8,$57,$F9,$E2,$FF,$D6,$4D
	.dc.b	$7F,$D6,$4D,$7F,$87,$47,$0F,$37,$F4,$7E,$27,$F9,$77,$E0,$FE,$03
	.dc.b	$6B,$FA,$8C,$77,$BE,$9B,$FF,$AA,$1F,$F0,$64,$F8,$3C,$EF,$D6,$E3
	.dc.b	$E4,$6C,$EA,$7C,$71,$82,$3C,$4F,$EC,$9F,$2D,$E2,$EE,$34,$F1,$4F
	.dc.b	$8F,$CA,$EA,$4A,$F8,$2D,$7E,$0B,$5F,$83,$68,$F8,$36,$F9,$6F,$01
	.dc.b	$2B,$EC,$63,$F3,$AB,$E8,$7E,$05,$FD,$E2,$7F,$DC,$0D,$7F,$DC,$0D
	.dc.b	$7E,$7B,$5F,$F7,$93,$54,$A3,$0D,$F0,$3E,$5C,$6F,$6F,$F4,$97,$CB
	.dc.b	$FC,$52,$5F,$9A,$7C,$5B,$7F,$B7,$17,$8F,$C8,$93,$E8,$6B,$E8,$6B
	.dc.b	$E8,$7E,$2F,$CB,$FD,$32,$47,$78,$F8,$7C,$3E,$62,$47,$C4,$6B,$F1
	.dc.b	$1A,$FC,$47,$E8,$FC,$07,$DF,$24,$7A,$E3,$E7,$FC,$1F,$C2,$24,$7D
	.dc.b	$06,$BF,$41,$AF,$D0,$7E,$E9,$F8,$B2,$3F,$1D,$FC,$77,$CB,$48,$FF
	.dc.b	$17,$BF,$91,$E9,$FD,$77,$C8,$3E,$4B,$F3,$AD,$AF,$EE,$3C,$28,$C4
	.dc.b	$7F,$77,$8C,$4F,$EB,$9F,$F0,$14,$7F,$00,$E0,$FC,$0F,$6F,$EC,$27
	.dc.b	$3C,$1F,$E0,$79,$FC,$2F,$CA,$FF,$13,$B7,$EB,$6D,$3F,$AC,$B6,$3F
	.dc.b	$AE,$B5,$FF,$02,$B5,$FF,$05,$35,$FF,$08,$35,$FF,$49,$B5,$EF,$1F
	.dc.b	$DA,$3B,$9E,$E7,$E9,$E7,$F8,$AB,$5F,$E3,$71,$AE,$8F,$45,$F8,$F2
	.dc.b	$F6,$66,$F9,$6F,$80,$4C,$3F,$93,$69,$FA,$4D,$B1,$1F,$1E,$B8,$3A
	.dc.b	$6B,$98,$F9,$5F,$2B,$23,$4F,$CD,$E0,$C4,$73,$8B,$8F,$92,$7F,$9D
	.dc.b	$EA,$E3,$A4,$BF,$7F,$75,$7D,$F3,$B7,$E1,$CF,$29,$FB,$9F,$2B,$27
	.dc.b	$D0,$F1,$C3,$E1,$FC,$AF,$E0,$49,$3F,$C4,$31,$7C,$8E,$23,$E1,$4D
	.dc.b	$AF,$F2,$66,$BF,$91,$AF,$C5,$78,$92,$FE,$33,$F8,$DF,$13,$11,$F2
	.dc.b	$E2,$DF,$D4,$9F,$C7,$7D,$5B,$E1,$7F,$C4,$AD,$BF,$F1,$4B,$5F,$F2
	.dc.b	$8B,$5F,$F2,$B3,$5F,$F2,$E3,$5F,$F6,$FB,$53,$47,$C4,$78,$89,$1F
	.dc.b	$DA,$4D,$CF,$E9,$FE,$5A,$47,$E0,$DF,$D5,$65,$88,$FC,$44,$5C,$7F
	.dc.b	$23,$3F,$DE,$3F,$01,$FE,$F1,$5F,$DB,$9A,$FE,$69,$B3,$FE,$79,$1F
	.dc.b	$BC,$FF,$62,$ED,$FF,$06,$4E,$9B,$DD,$D4,$FF,$05,$57,$0F,$2B,$E6
	.dc.b	$FC,$6F,$CC,$26,$3F,$31,$AF,$CC,$6B,$F3,$1F,$A7,$F1,$DF,$58,$91
	.dc.b	$F1,$77,$4F,$F0,$BF,$95,$93,$FA,$77,$F8,$DF,$C6,$E2,$3F,$59,$17
	.dc.b	$1F,$C5,$F3,$2B,$F8,$A7,$CA,$CB,$FD,$83,$FC,$93,$D1,$C4,$7E,$D6
	.dc.b	$2D,$F2,$9F,$E3,$1E,$AB,$A7,$FC,$DE,$DB,$FF,$39,$B5,$FF,$4D,$B5
	.dc.b	$FF,$52,$78,$9F,$E2,$DF,$4B,$FC,$58,$48,$FC,$24,$F1,$C3,$F8,$6F
	.dc.b	$CA,$7F,$2B,$24,$FF,$8A,$B8,$97,$88,$FE,$19,$36,$BF,$C9,$95,$A9
	.dc.b	$FF,$1C,$7B,$4F,$CF,$FF,$93,$37,$C9,$FF,$45,$B6,$7F,$D6,$6D,$7F
	.dc.b	$D7,$1E,$37,$F7,$EF,$95,$FE,$7C,$48,$EB,$7F,$C9,$7A,$11,$FE,$04
	.dc.b	$F9,$77,$CA,$7F,$8B,$1B,$1F,$D4,$1A,$FF,$4F,$6B,$FE,$CB,$C3,$FC
	.dc.b	$1B,$E3,$FF,$4D,$24,$7F,$87,$27,$8E,$BF,$CD,$27,$33,$FC,$BD,$FF
	.dc.b	$27,$4E,$B0,$F2,$DF,$4A,$FE,$65,$5C,$2F,$F8,$9D,$EE,$7D,$5C,$1E
	.dc.b	$92,$3E,$97,$FB,$27,$2F,$F6,$12,$47,$96,$3F,$CE,$5E,$A3,$E5,$1F
	.dc.b	$29,$FE,$BA,$8C,$36,$33,$5F,$EE,$49,$CB,$63,$2F,$FB,$5D,$AF,$FB
	.dc.b	$9D,$AF,$FB,$91,$AB,$3F,$F1,$BC,$7F,$77,$24,$7F,$9F,$67,$BB,$FE
	.dc.b	$7C,$F8,$F7,$ED,$CF,$1F,$FD,$A3,$18,$6C,$66,$7F,$DF,$73,$96,$C6
	.dc.b	$5F,$C2,$9F,$86,$8D,$38,$0E,$1D,$B5,$FB,$3E,$4F,$FA,$7F,$82,$F5
	.dc.b	$FF,$84,$4F,$F5,$05,$74,$E0,$FE,$EF,$19,$BF,$EE,$0E,$B0,$F2,$4A
	.dc.b	$FC,$91,$FE,$9D,$AF,$0B,$5C,$F9,$3F,$00,$8F,$B9,$AF,$AF,$0F,$93
	.dc.b	$1F,$E2,$9E,$AF,$F1,$31,$27,$FD,$59,$3A,$7E,$97,$E0,$6F,$4F,$F0
	.dc.b	$CD,$8F,$C2,$35,$F9,$1B,$23,$11,$F3,$A3,$1E,$8F,$22,$47,$27,$E8
	.dc.b	$3F,$53,$E7,$1E,$8B,$DC,$CC,$A7,$31,$86,$C7,$26,$BF,$6D,$FF,$28
	.dc.b	$5E,$FD,$3F,$F4,$44,$62,$78,$47,$FA,$36,$7D,$4F,$F9,$3A,$3F,$85
	.dc.b	$C6,$37,$F8,$33,$99,$4E,$5F,$87,$97,$AA,$BE,$3E,$3B,$AF,$8C,$D9
	.dc.b	$D5,$7F,$1E,$FA,$CF,$81,$7F,$E3,$A6,$C7,$E8,$FD,$A1,$FD,$AE,$7F
	.dc.b	$85,$CE,$B8,$BF,$15,$5F,$9A,$AD,$D7,$E9,$1B,$1F,$85,$6B,$F9,$E6
	.dc.b	$BF,$A5,$6B,$CD,$F8,$1B,$F8,$9A,$6B,$9D,$83,$E1,$3C,$D7,$FC,$06
	.dc.b	$31,$87,$E4,$2D,$EE,$EB,$0E,$38,$5B,$DE,$1F,$E7,$F6,$E5,$EE,$CB
	.dc.b	$E1,$E1,$BB,$EF,$18,$CB,$E0,$4F,$98,$8E,$09,$87,$C3,$8C,$BF,$4F
	.dc.b	$CD,$F1,$1E,$AD,$9F,$BC,$3E,$37,$38,$7F,$19,$8C,$3F,$CA,$5B,$7F
	.dc.b	$A6,$6B,$FA,$56,$BF,$8E,$6B,$F8,$C6,$D7,$82,$BF,$15,$1A,$D7,$E7
	.dc.b	$F9,$DF,$1C,$F9,$FD,$6F,$07,$F9,$FE,$5F,$8B,$9C,$C6,$32,$FE,$C1
	.dc.b	$DF,$F9,$BB,$E0,$D9,$D4,$FF,$63,$E1,$FE,$D7,$87,$E5,$7C,$3E,$7F
	.dc.b	$2F,$F5,$D9,$19,$3F,$B3,$46,$2B,$ED,$4D,$E3,$FE,$BF,$36,$BF,$BC
	.dc.b	$E5,$E9,$8F,$11,$23,$FB,$C3,$FB,$1D,$3F,$1F,$EB,$FE,$1C,$FC,$7F
	.dc.b	$90,$FC,$E1,$F2,$99,$C3,$FC,$27,$18,$7F,$C1,$ED,$D3,$FA,$15,$71
	.dc.b	$AF,$DF,$36,$BF,$8B,$E1,$F9,$EF,$67,$F2,$C7,$E7,$BC,$F7,$CF,$7F
	.dc.b	$A8,$6B,$F7,$53,$A6,$9F,$E1,$6D,$BF,$6F,$5C,$3D,$DD,$67,$FE,$C2
	.dc.b	$C3,$93,$63,$EC,$FC,$B7,$79,$CD,$FF,$82,$23,$EC,$13,$0F,$F0,$F3
	.dc.b	$F1,$F8,$9F,$17,$FA,$D3,$F0,$FF,$84,$7F,$8A,$2B,$F8,$84,$65,$A7
	.dc.b	$9C,$FA,$B1,$BC,$BF,$69,$3F,$B7,$F3,$65,$6E,$5C,$8B,$7C,$BE,$A7
	.dc.b	$F7,$97,$73,$FF,$59,$D6,$B0,$FE,$45,$3F,$C2,$AB,$FB,$AF,$C6,$5F
	.dc.b	$00,$DF,$95,$C3,$6F,$FC,$0F,$1F,$CD,$EB,$FB,$EB,$68,$D2,$4F,$02
	.dc.b	$4F,$F8,$17,$85,$FA,$83,$E0,$BF,$00,$DA,$F8,$1A,$3E,$B3,$FD,$A3
	.dc.b	$1F,$0E,$F8,$DF,$02,$47,$09,$DE,$0B,$FC,$2E,$1D,$63,$F9,$17,$C9
	.dc.b	$7F,$25,$3F,$93,$77,$DC,$3F,$A4,$45,$9F,$E5,$13,$6F,$FA,$2C,$7F
	.dc.b	$87,$BE,$67,$E5,$1F,$2F,$86,$E1,$FE,$C7,$DB,$E9,$57,$F8,$B2,$7E
	.dc.b	$41,$27,$FC,$66,$D8,$D4,$60,$FD,$3F,$23,$E0,$BF,$9E,$1F,$CF,$7E
	.dc.b	$76,$CF,$F5,$F8,$B5,$FC,$FE,$6E,$5F,$D7,$3A,$FF,$A6,$FC,$8F,$F4
	.dc.b	$53,$F3,$3E,$35,$C7,$FB,$44,$5C,$3E,$54,$7E,$62,$71,$7C,$F2,$78
	.dc.b	$5F,$D3,$CF,$DD,$7E,$DE,$F8,$2C,$61,$F6,$63,$0D,$8D,$FC,$0F,$06
	.dc.b	$46,$4E,$11,$8C,$BF,$AD,$7C,$F7,$F5,$D3,$FA,$F7,$CF,$5A,$FF,$00
	.dc.b	$45,$9F,$EC,$13,$73,$F4,$7C,$F7,$F6,$53,$FB,$37,$CA,$5C,$3F,$C1
	.dc.b	$11,$6F,$FB,$44,$DA,$FE,$1F,$9B,$E2,$3D,$E7,$EB,$7F,$5E,$FF,$67
	.dc.b	$63,$EF,$8C,$7C,$47,$EA,$67,$38,$FF,$8E,$1A,$FF,$9A,$5A,$FF,$9B
	.dc.b	$9A,$FF,$AD,$1A,$9A,$6B,$8F,$FC,$13,$E2,$FF,$80,$4F,$F0,$17,$95
	.dc.b	$71,$FE,$19,$E2,$FF,$0B,$F4,$FF,$AC,$78,$DF,$E0,$A9,$FF,$25,$CE
	.dc.b	$5B,$19,$CB,$FD,$97,$F8,$D7,$F1,$9D,$FF,$18,$DA,$FF,$5A,$BF,$E3
	.dc.b	$C7,$F8,$D6,$68,$D5,$7F,$A6,$DF,$D2,$D7,$2E,$CF,$EE,$3F,$17,$FE
	.dc.b	$7B,$C3,$8D,$69,$FE,$BB,$19,$BF,$F4,$86,$48,$C4,$7F,$A5,$A3,$13
	.dc.b	$FD,$D6,$73,$3C,$A3,$04,$3E,$0C,$6A,$3C,$DD,$FF,$E2,$DB,$3E,$6F
	.dc.b	$8F,$C3,$6F,$8C,$77,$23,$82,$57,$A2,$D3,$A8,$F9,$BF,$41,$DC,$F8
	.dc.b	$7C,$76,$FE,$CD,$FE,$45,$F3,$C7,$B4,$8E,$09,$5D,$E3,$2F,$DA,$F2
	.dc.b	$3D,$87,$D0,$F8,$6B,$5D,$9F,$90,$47,$E2,$B6,$7D,$4D,$7E,$85,$FF
	.dc.b	$B9,$E7,$FB,$FE,$7F,$D3,$9F,$B7,$E5,$7F,$8B,$87,$B1,$F3,$5C,$3F
	.dc.b	$E2,$B7,$31,$F9,$46,$DF,$E0,$1A,$FD,$F6,$BF,$E2,$26,$BF,$DA,$9B
	.dc.b	$B7,$F6,$0A,$FF,$11,$5B,$E5,$BE,$F1,$E5,$F5,$2E,$3F,$A8,$C1,$1E
	.dc.b	$52,$47,$95,$B8,$F9,$5A,$F9,$5E,$FE,$1F,$E2,$CF,$85,$E7,$5A,$EC
	.dc.b	$FC,$12,$3F,$05,$F3,$FC,$FF,$E4,$CF,$CA,$7D,$05,$9F,$D3,$C5,$C3
	.dc.b	$F9,$C9,$FA,$F9,$C7,$37,$F4,$3F,$82,$57,$FA,$25,$B7,$FB,$85,$59
	.dc.b	$99,$39,$78,$46,$2F,$B8,$27,$CF,$5F,$9E,$FF,$07,$C7,$FB,$76,$F1
	.dc.b	$AF,$FB,$6A,$AC,$E4,$57,$07,$A0,$8D,$BF,$EC,$EA,$AF,$80,$E3,$5F
	.dc.b	$EE,$E8,$CC,$71,$3F,$D5,$D1,$96,$9D,$8D,$7B,$FE,$17,$F8,$E1,$A3
	.dc.b	$C7,$5E,$0F,$7F,$FD,$93,$86,$E7,$B3,$C6,$33,$7E,$F9,$F2,$9E,$F9
	.dc.b	$E1,$5C,$F0,$F3,$BF,$EF,$A9,$F6,$1F,$EF,$C7,$0F,$AC,$F4,$7C,$F3
	.dc.b	$D2,$7B,$7E,$B9,$FA,$93,$98,$C4,$23,$0E,$DA,$F4,$E0,$FF,$16,$39
	.dc.b	$6B,$9E,$0E,$72,$9C,$CF,$D8,$8C,$4F,$D8,$7F,$CD,$71,$7D,$8F,$BB
	.dc.b	$9F,$DF,$9D,$CF,$C6,$3C,$D7,$E2,$9D,$57,$0D,$3B,$36,$BD,$F5,$B9
	.dc.b	$F1,$3E,$AD,$B8,$CF,$E5,$DA,$FD,$DC,$3C,$31,$E3,$9E,$53,$F2,$AF
	.dc.b	$B3,$F2,$5B,$5E,$8C,$38,$63,$ED,$A8,$7A,$E7,$58,$FE,$9F,$D8,$7C
	.dc.b	$F3,$E3,$3F,$3F,$87,$FA,$7A,$31,$87,$F9,$8D,$AF,$EF,$AB,$8F,$C5
	.dc.b	$70,$E1,$E9,$5A,$BE,$58,$67,$06,$1E,$08,$C3,$CA,$3F,$CE,$1F,$26
	.dc.b	$F8,$CC,$B6,$33,$18,$F8,$EC,$B2,$32,$72,$8F,$07,$73,$E7,$BF,$CB
	.dc.b	$93,$FE,$10,$6B,$FD,$F1,$AF,$F7,$B6,$BF,$C3,$7C,$87,$5B,$E0,$7D
	.dc.b	$91,$86,$C6,$67,$FC,$43,$5D,$B5,$6F,$FE,$E8,$87,$B3,$FA,$4C,$BC
	.dc.b	$CD,$8C,$DF,$9E,$30,$47,$04,$3F,$85,$65,$E6,$F9,$9F,$19,$23,$C7
	.dc.b	$3F,$12,$9E,$9F,$72,$4F,$C5,$6B,$F1,$5A,$FC,$57,$F7,$77,$F1,$3E
	.dc.b	$6F,$E0,$92,$7E,$0C,$FD,$2F,$9E,$77,$FB,$DB,$3F,$E4,$26,$BF,$E2
	.dc.b	$E6,$BF,$E2,$D6,$BF,$CF,$F3,$46,$23,$FD,$B2,$D8,$CC,$F9,$B6,$7A
	.dc.b	$1F,$58,$97,$FE,$56,$BF,$0E,$FF,$5A,$39,$F8,$BE,$72,$3E,$76,$BE
	.dc.b	$76,$BE,$77,$C1,$F1,$7F,$84,$24,$77,$8F,$89,$C5,$F0,$C9,$1F,$0D
	.dc.b	$AF,$C3,$6B,$F0,$DF,$67,$CF,$7E,$40,$91,$EB,$8F,$A1,$F3,$DF,$3C
	.dc.b	$91,$F3,$DA,$FC,$F6,$BF,$3D,$F9,$07,$E7,$08,$FC,$77,$F3,$FF,$33
	.dc.b	$23,$FB,$D3,$FA,$2E,$8F,$D8,$FC,$D3,$E4,$BF,$3C,$DA,$FF,$40,$B5
	.dc.b	$FF,$36,$B5,$FE,$E3,$C8,$E2,$FF,$71,$ED,$FE,$CD,$6F,$47,$ED,$5B
	.dc.b	$4A,$E1,$F2,$BF,$C4,$61,$F3,$CD,$3F,$AF,$36,$3F,$AA,$B5,$FE,$B4
	.dc.b	$D7,$FB,$B3,$5F,$F0,$2B,$5F,$F0,$93,$5E,$F1,$FD,$83,$B5,$F3,$2E
	.dc.b	$9F,$E4,$2D,$AF,$F0,$D6,$9F,$91,$AD,$D7,$A7,$0F,$24,$FC,$1F,$97
	.dc.b	$FE,$23,$A1,$A6,$37,$B8,$FF,$A9,$27,$4D,$9D,$39,$F9,$6E,$64,$78
	.dc.b	$70,$57,$FC,$7B,$BF,$D3,$9C,$63,$F4,$EB,$D7,$C7,$2F,$D7,$C3,$FD
	.dc.b	$3F,$7F,$E4,$91,$FD,$25,$77,$6E,$31,$57,$0F,$85,$5A,$9F,$57,$B1
	.dc.b	$F0,$08,$F6,$3F,$A6,$C7,$69,$FC,$37,$BF,$F1,$89,$3F,$ED,$4C,$9F
	.dc.b	$3D,$88,$F6,$EE,$AC,$5F,$F0,$38,$5F,$0A,$F9,$9E,$1F,$D2,$23,$F3
	.dc.b	$FC,$3A,$F4,$8B,$7F,$99,$9F,$CA,$7A,$CF,$91,$FF,$18,$36,$FF,$C4
	.dc.b	$ED,$7F,$C5,$8D,$7F,$C9,$0D,$7F,$CA,$ED,$7F,$CC,$8D,$7F,$DB,$11
	.dc.b	$DB,$D5,$F0,$92,$3E,$51,$A3,$F1,$1F,$21,$F8,$E2,$3F,$8B,$D6,$BF
	.dc.b	$22,$2E,$3F,$B5,$9F,$EF,$BF,$3A,$ED,$A4,$BF,$B1,$4F,$9E,$FE,$CD
	.dc.b	$CB,$FE,$15,$BE,$15,$F7,$7B,$11,$88,$FF,$09,$36,$FF,$9C,$1C,$FE
	.dc.b	$3B,$E6,$FC,$BF,$CC,$25,$7C,$C6,$BF,$31,$AF,$CC,$7E,$9F,$B9,$F5
	.dc.b	$89,$1F,$1B,$81,$3F,$B2,$F7,$E4,$FE,$BD,$FE,$43,$F3,$98,$8F,$C6
	.dc.b	$C5,$C7,$F1,$9C,$AA,$FD,$DF,$C6,$CB,$FD,$C3,$FC,$AB,$D0,$C4,$7E
	.dc.b	$5A,$2D,$F2,$9F,$E3,$DE,$BF,$E6,$6B,$FC,$DE,$D7,$FC,$E8,$D7,$FD
	.dc.b	$20,$D7,$FC,$89,$C4,$F4,$FF,$C6,$24,$8F,$C2,$CF,$3B,$FE,$65,$F1
	.dc.b	$BF,$C4,$89,$3F,$E3,$AA,$AF,$EB,$22,$CF,$F1,$DB,$ED,$3F,$B9,$F6
	.dc.b	$9F,$3B,$FC,$AD,$BE,$4F,$FA,$21,$B3,$FE,$8D,$6B,$FE,$20,$EE,$7C
	.dc.b	$F7,$F2,$42,$47,$5B,$FE,$53,$D1,$FE,$09,$F9,$A7,$CD,$7F,$8C,$5B
	.dc.b	$5F,$E9,$9C,$11,$88,$FF,$5B,$46,$27,$FC,$42,$FF,$3C,$8F,$E5,$DE
	.dc.b	$1F,$F3,$43,$FC,$A3,$5C,$75,$FE,$B1,$19,$9D,$9F,$C6,$6B,$A3,$83
	.dc.b	$F8,$9C,$6D,$FF,$76,$C6,$9B,$5F,$EA,$17,$D0,$D7,$CE,$D7,$FD,$B7
	.dc.b	$1E,$CE,$0F,$4F,$1F,$F4,$A2,$4F,$8E,$B9,$61,$FE,$23,$F5,$5F,$0A
	.dc.b	$F9,$6F,$F5,$DC,$61,$B1,$9B,$FF,$61,$4E,$5B,$19,$7F,$DC,$11,$96
	.dc.b	$C6,$5F,$F7,$8C,$65,$FF,$01,$E5,$E2,$7F,$AE,$4F,$F6,$23,$FD,$1D
	.dc.b	$3C,$EF,$FC,$75,$F2,$CF,$9E,$78,$DF,$ED,$28,$C3,$63,$38,$7F,$B6
	.dc.b	$67,$2D,$8C,$BF,$EF,$78,$CB,$63,$28,$91,$A7,$D3,$D1,$FD,$E0,$8F
	.dc.b	$F5,$98,$E7,$7F,$E1,$D9,$CC,$FF,$54,$7F,$D5,$53,$AE,$1E,$B7,$FE
	.dc.b	$A5,$AF,$17,$8F,$EB,$23,$E9,$6B,$A3,$07,$FB,$DF,$CD,$FF,$17,$75
	.dc.b	$7F,$83,$09,$1E,$5A,$ED,$97,$E0,$FE,$75,$E9,$FE,$09,$B5,$F8,$06
	.dc.b	$BF,$79,$AF,$DC,$7F,$C2,$6E,$7D,$0E,$44,$7E,$51,$07,$CF,$3D,$1F
	.dc.b	$F5,$9C,$64,$CB,$87,$84,$6F,$DB,$D4,$69,$B1,$FE,$55,$E2,$BE,$9F
	.dc.b	$FA,$2E,$31,$5D,$5A,$FA,$63,$FD,$0C,$DE,$0F,$81,$18,$8F,$C1,$53
	.dc.b	$2F,$F3,$15,$7C,$5C,$75,$5F,$EA,$49,$F6,$4E,$49,$1F,$EB,$CF,$98
	.dc.b	$FF,$68,$C7,$F8,$7F,$81,$18,$96,$DF,$A0,$F6,$87,$F6,$4B,$FF,$6D
	.dc.b	$56,$E7,$53,$E2,$E3,$DB,$4E,$35,$FA,$36,$DF,$E8,$5A,$FD,$69,$E3
	.dc.b	$3B,$9E,$47,$BB,$0F,$3F,$C2,$F8,$EF,$FC,$ED,$8A,$31,$87,$A3,$DD
	.dc.b	$D5,$F0,$C7,$E0,$5E,$DB,$7C,$31,$FF,$5E,$C6,$0F,$D6,$38,$C7,$D1
	.dc.b	$87,$B4,$93,$ED,$6C,$6A,$30,$7E,$96,$DE,$5F,$91,$B1,$FE,$51,$D9
	.dc.b	$F3,$CF,$8B,$D5,$FB,$E7,$31,$F8,$06,$D7,$F2,$6A,$E0,$D9,$D4,$63
	.dc.b	$0F,$E4,$ED,$7F,$92,$B5,$FC,$7B,$5F,$9B,$E6,$79,$CF,$A1,$DB,$FD
	.dc.b	$73,$BB,$18,$7F,$3F,$7F,$27,$19,$8F,$E9,$71,$8E,$FF,$E8,$11,$F0
	.dc.b	$FC,$3F,$84,$74,$87,$CC,$E2,$7E,$BC,$DC,$7F,$AB,$CD,$E4,$FF,$77
	.dc.b	$71,$FE,$6A,$B1,$7F,$DD,$6A,$DF,$FB,$07,$0E,$B3,$FD,$7A,$7C,$84
	.dc.b	$7E,$B7,$AB,$F5,$4F,$C7,$79,$0F,$9E,$7C,$A6,$6B,$FC,$21,$18,$7F
	.dc.b	$C1,$8D,$C7,$F8,$85,$71,$AF,$DB,$B6,$BF,$09,$EC,$7E,$0C,$FC,$E7
	.dc.b	$AA,$F9,$E7,$B9,$9A,$FC,$CD,$5C,$7F,$6F,$1F,$D1,$EF,$85,$67,$83
	.dc.b	$FD,$85,$87,$2A,$FF,$31,$36,$FF,$C0,$3F,$2D,$E0,$8C,$57,$FA,$5F
	.dc.b	$74,$FD,$92,$5F,$F8,$19,$FD,$A7,$4B,$E3,$3F,$38,$7E,$1B,$E5,$BF
	.dc.b	$AD,$57,$EC,$23,$2D,$3F,$C3,$11,$AC,$BF,$D8,$B5,$CE,$7F,$6F,$E7
	.dc.b	$7E,$9C,$E9,$F2,$D6,$32,$4A,$FF,$0E,$DA,$75,$1F,$18,$FE,$D7,$F1
	.dc.b	$9F,$C2,$57,$84,$8C,$2F,$94,$B1,$87,$F9,$7C,$F0,$6C,$EF,$0F,$F0
	.dc.b	$24,$60,$8E,$09,$3F,$D4,$F8,$5F,$3A,$F8,$1F,$C0,$B6,$BF,$D3,$BF
	.dc.b	$3D,$FE,$CE,$7F,$D3,$51,$9B,$F6,$5F,$2B,$E2,$7F,$25,$E2,$61,$C6
	.dc.b	$B7,$7C,$67,$F5,$3B,$BE,$4B,$F8,$A1,$FC,$57,$BA,$E7,$FC,$A2,$2E
	.dc.b	$1F,$5A,$6D,$FF,$5A,$D5,$F3,$1F,$28,$F9,$7C,$17,$1F,$F6,$2C,$FD
	.dc.b	$57,$FC,$55,$1F,$28,$8E,$09,$87,$F8,$AE,$32,$FF,$5B,$F8,$1F,$E5
	.dc.b	$47,$F2,$BF,$9E,$B5,$FC,$FA,$2C,$FE,$32,$6E,$3F,$03,$E4,$7F,$99
	.dc.b	$1F,$97,$F1,$AE,$1F,$D1,$A2,$DF,$E3,$26,$D7,$C8,$F0,$7F,$9D,$1F
	.dc.b	$B8,$F9,$B7,$C0,$E3,$1F,$F6,$84,$61,$B1,$BF,$81,$E0,$C8,$C9,$C2
	.dc.b	$31,$8F,$E9,$3E,$7B,$FA,$41,$FD,$27,$E7,$AD,$7F,$5C,$8B,$3F,$AA
	.dc.b	$9B,$97,$DE,$F9,$EF,$EA,$07,$F5,$1F,$94,$B8,$7F,$64,$8B,$7F,$B2
	.dc.b	$9B,$4A,$FE,$1D,$9B,$E1,$FE,$01,$FA,$BF,$9E,$7C,$6E,$2B,$FC,$B5
	.dc.b	$7F,$E1,$E9,$CC,$7F,$8A,$9B,$8F,$F8,$C5,$AF,$F8,$DD,$AF,$F8,$E9
	.dc.b	$AF,$F9,$95,$A9,$A6,$BC,$DF,$F0,$EF,$89,$FD,$BC,$FE,$E1,$E5,$5B
	.dc.b	$FE,$23,$E9,$FF,$58,$F1,$3F,$BF,$63,$FE,$9E,$6B,$FE,$73,$F3,$D1
	.dc.b	$88,$62,$F9,$DF,$E3,$5A,$7F,$54,$6E,$3F,$E7,$B9,$D7,$A7,$FE,$14
	.dc.b	$9C,$DF,$27,$F4,$8F,$F1,$79,$E9,$D9,$FD,$C3,$E2,$FE,$96,$1F,$E3
	.dc.b	$9B,$DE,$3F,$57,$06,$6B,$DC,$7F,$B4,$1A,$FF,$AE,$9A,$FB,$5A,$F2
	.dc.b	$6C,$73,$BF,$EF,$B1,$83,$78,$FC,$1E,$FF,$61,$E7,$F8,$FB,$5F,$ED
	.dc.b	$5A,$47,$22,$47,$FB,$9E,$32,$D8,$CE,$1E,$96,$BF,$E7,$CB,$FA,$1F
	.dc.b	$3D,$F2,$CF,$8D,$C9,$6F,$C0,$FA,$08,$F7,$6D,$7C,$56,$BF,$15,$FF
	.dc.b	$28,$69,$FC,$F6,$3E,$A3,$74,$FA,$1A,$7C,$FF,$86,$FB,$9E,$57,$D7
	.dc.b	$7E,$39,$31,$F7,$37,$0F,$A8,$D7,$EA,$3E,$0F,$2F,$F0,$C7,$F8,$C0
	.dc.b	$FF,$23,$C7,$89,$C7,$B1,$FE,$30,$73,$1F,$EF,$66,$BF,$99,$6B,$F8
	.dc.b	$66,$E1,$FD,$86,$79,$12,$FF,$D7,$99,$3E,$5B,$F3,$07,$8F,$D6,$B3
	.dc.b	$FB,$E3,$B3,$E3,$23,$FE,$AD,$BF,$ED,$D1,$8C,$7F,$D6,$91,$89,$D6
	.dc.b	$2A,$D5,$7D,$FC,$3F,$CB,$5A,$7E,$B7,$C3,$F8,$9E,$7F,$EA,$DB,$8F
	.dc.b	$28,$F8,$84,$7D,$7F,$3D,$FB,$13,$F6,$5F,$3D,$63,$F5,$91,$71,$FB
	.dc.b	$F3,$70,$FF,$27,$FC,$12,$BF,$D1,$AD,$BF,$F5,$62,$AC,$CC,$9C,$BC
	.dc.b	$23,$17,$FE,$15,$3F,$DA,$FF,$3D,$FF,$00,$7E,$7B,$FC,$29,$1F,$ED
	.dc.b	$AE,$BF,$DF,$46,$23,$61,$C2,$8D,$EF,$FE,$DD,$AE,$8F,$FB,$8E,$33
	.dc.b	$3F,$EE,$59,$CB,$DF,$1A,$F6,$F8,$0E,$B5,$FC,$3D,$EF,$0A,$3F,$EB
	.dc.b	$3B,$ED,$78,$AD,$D7,$4A,$F3,$4D,$79,$1E,$BA,$E2,$D3,$CC,$53,$FD
	.dc.b	$CB,$1C,$67,$A5,$FB,$27,$E2,$9F,$EF,$88,$FF,$7C,$BE,$66,$BE,$03
	.dc.b	$D2,$7A,$BE,$7B,$FC,$49,$EE,$7F,$C0,$06,$C6,$8D,$7A,$49,$D3,$63
	.dc.b	$42,$5C,$1F,$EE,$E8,$CC,$63,$83,$B4,$E6,$7B,$4E,$5D,$C7,$1D,$3F
	.dc.b	$05,$3F,$E7,$19,$F9,$DA,$7A,$A3,$11,$F4,$CF,$65,$71,$9D,$36,$3E
	.dc.b	$07,$B1,$F0,$AF,$AB,$6B,$78,$76,$AE,$6D,$7D,$D3,$B7,$99,$9C,$3D
	.dc.b	$57,$E8,$C3,$C1,$3C,$0D,$57,$17,$E4,$46,$AB,$8E,$1E,$DA,$D4,$7B
	.dc.b	$AF,$A6,$1E,$5E,$CF,$C9,$F1,$7D,$2C,$BF,$BB,$87,$11,$A8,$FA,$ED
	.dc.b	$7E,$F3,$5F,$EA,$07,$13,$65,$36,$53,$F1,$7C,$CF,$9D,$7A,$5E,$38
	.dc.b	$C3,$63,$33,$FC,$9D,$AF,$F2,$66,$BD,$1A,$FF,$85,$5A,$FF,$80,$9B
	.dc.b	$2F,$C3,$1F,$01,$E1,$6D,$7F,$35,$6B,$FC,$D1,$AF,$E7,$FC,$27,$BF
	.dc.b	$FA,$A6,$C7,$CD,$6B,$FA,$06,$BC,$6B,$81,$1E,$0D,$78,$35,$E0,$75
	.dc.b	$F8,$1F,$01,$1C,$1F,$D3,$7F,$2E,$FF,$73,$B7,$E0,$7E,$74,$62,$3F
	.dc.b	$A7,$B5,$FA,$AD,$7F,$C7,$CD,$7F,$C5,$6D,$C8,$E7,$5F,$F2,$07,$D1
	.dc.b	$7E,$2B,$BC,$D2,$5F,$97,$E1,$7C,$64,$70,$43,$CA,$E5,$A7,$F7,$2F
	.dc.b	$A2,$EC,$47,$07,$C1,$F9,$6F,$79,$1C,$10,$F8,$2E,$5A,$7F,$16,$E4
	.dc.b	$7D,$8B,$E2,$BF,$51,$18,$6C,$66,$78,$35,$FE,$F7,$D0,$F8,$EF,$CF
	.dc.b	$36,$3F,$C0,$4D,$77,$1A,$23,$A6,$BF,$B2,$D3,$E8,$13,$83,$E8,$1D
	.dc.b	$B0,$93,$D9,$AF,$91,$AF,$ED,$2F,$D0,$D8,$F3,$FC,$CC,$8F,$3B,$5F
	.dc.b	$6F,$9D,$EC,$23,$EC,$7D,$FE,$B4,$8F,$7B,$5F,$8B,$F2,$5F,$10,$8F
	.dc.b	$C4,$7F,$AA,$FC,$F3,$E8,$29,$27,$F8,$D4,$65,$B1,$97,$FA,$7C,$65
	.dc.b	$B1,$F6,$BE,$3F,$F2,$B1,$F7,$DA,$FF,$5B,$E0,$46,$23,$FB,$1C,$63
	.dc.b	$93,$F3,$DF,$3F,$27,$FC,$8A,$D7,$EB,$35,$FF,$1C,$35,$FF,$51,$70
	.dc.b	$3E,$83,$F5,$04,$8F,$F2,$7B,$5F,$C0,$35,$FF,$21,$3F,$DE,$BF,$7A
	.dc.b	$79,$F4,$93,$FC,$FE,$32,$D8,$CB,$FD,$D2,$32,$D8,$FE,$8D,$F1,$CF
	.dc.b	$D4,$1E,$B6,$1B,$5F,$DF,$27,$53,$93,$F8,$17,$B9,$27,$FC,$D2,$D7
	.dc.b	$F3,$4D,$7F,$CB,$8D,$7E,$2F,$67,$F0,$B2,$3F,$E7,$26,$BF,$A3,$6B
	.dc.b	$FE,$65,$7F,$C3,$BE,$E3,$E6,$69,$27,$FB,$14,$65,$B1,$97,$FC,$2F
	.dc.b	$19,$6C,$7F,$8A,$3E,$EB,$E0,$C7,$F8,$71,$AF,$F8,$8D,$AF,$F8,$A9
	.dc.b	$AF,$FB,$77,$E9,$3F,$93,$11,$FF,$46,$B5,$FD,$F3,$5F,$F4,$03,$F5
	.dc.b	$3F,$54,$7C,$15,$24,$FD,$38,$CB,$63,$2F,$F8,$DF,$44,$EA,$30,$D7
	.dc.b	$FC,$79,$8B,$E2,$7E,$59,$23,$E4,$B5,$F8,$ED,$78,$BF,$1F,$DA,$91
	.dc.b	$FE,$11,$CB,$FD,$10,$D8,$FC,$7F,$EA,$DF,$E9,$F8,$FC,$63,$5F,$43
	.dc.b	$5F,$F2,$AE,$28,$C4,$7F,$9C,$23,$1E,$E7,$42,$47,$47,$FA,$B7,$54
	.dc.b	$8F,$F5,$EB,$5F,$E5,$CD,$7F,$D6,$0D,$7F,$CC,$BF,$85,$7F,$A7,$5F
	.dc.b	$F1,$63,$5F,$F2,$FB,$5F,$F3,$33,$5F,$F4,$2F,$B0,$FB,$EF,$87,$1F
	.dc.b	$E6,$B8,$C6,$5F,$E3,$76,$C7,$F9,$A9,$AF,$FB,$23,$3F,$F1,$73,$FE
	.dc.b	$90,$F3,$11,$88,$46,$3D,$4F,$F1,$84,$E6,$31,$3F,$ED,$58,$C1,$27
	.dc.b	$7F,$01,$27,$4D,$7F,$AA,$79,$7F,$E0,$12,$3F,$EE,$87,$BF,$A1,$23
	.dc.b	$B3,$5F,$EB,$9F,$4B,$FE,$09,$23,$FE,$F0,$6B,$A3,$04,$30,$43,$FD
	.dc.b	$CC,$F0,$7F,$CA,$38,$7F,$A0,$9B,$EC,$7C,$22,$3F,$E6,$0F,$78,$7F
	.dc.b	$99,$0A,$E5,$C1,$FE,$6B,$F7,$9F,$4E,$F8,$6C,$36,$BF,$D4,$33,$A9
	.dc.b	$C9,$23,$33,$B7,$FD,$A8,$D7,$FD,$C0,$D7,$FD,$9B,$3F,$D5,$BE,$7F
	.dc.b	$C5,$E4,$9A,$49,$FF,$58,$FC,$A3,$C0,$6B,$FE,$83,$8C,$36,$33,$3C
	.dc.b	$1A,$FF,$AF,$E7,$6E,$67,$81,$1F,$E8,$9F,$40,$F9,$3F,$E9,$6D,$8C
	.dc.b	$C6,$08,$E0,$87,$F1,$78,$E0,$F8,$6F,$FB,$E3,$FE,$D6,$BF,$67,$4F
	.dc.b	$A4,$91,$E9,$6B,$C5,$AF,$F0,$AF,$BE,$78,$5F,$D8,$1B,$1F,$D7,$5A
	.dc.b	$FF,$29,$6B,$FE,$E1,$F3,$BF,$DD,$24,$7F,$C3,$27,$FB,$BB,$E8,$5F
	.dc.b	$4F,$49,$1F,$EB,$38,$CB,$7D,$5F,$DC,$47,$F9,$47,$AB,$04,$9F,$F7
	.dc.b	$5C,$3E,$89,$C7,$CD,$B3,$A6,$BF,$EF,$79,$E1,$3F,$EC,$3E,$57,$A1
	.dc.b	$F9,$EA,$FF,$7E,$46,$39,$F5,$39,$91,$27,$4F,$FB,$FB,$F6,$07,$CD
	.dc.b	$EE,$37,$18,$BF,$F6,$93,$63,$84,$EF,$C5,$FC,$B7,$B1,$79,$4D,$26
	.dc.b	$1F,$9C,$FD,$D3,$FC,$2A,$DF,$3F,$C1,$CE,$9C,$E5,$F9,$96,$BF,$89
	.dc.b	$6B,$EC,$8E,$13,$CA,$73,$29,$CD,$7E,$17,$E4,$BD,$1C,$C6,$92,$34
	.dc.b	$4A,$FC,$54,$65,$F2,$3D,$C9,$1E,$D6,$C6,$A3,$04,$7E,$99,$FC,$13
	.dc.b	$B7,$F0,$91,$FC,$41,$B1,$A3,$F0,$EE,$1E,$3C,$3F,$E6,$0E,$3F,$B0
	.dc.b	$FF,$1C,$F7,$EF,$79,$A4,$C3,$F8,$CB,$6F,$F8,$CB,$5F,$2F,$C0,$7F
	.dc.b	$2C,$F7,$FF,$08,$DA,$FE,$61,$18,$7F,$D5,$07,$F5,$3F,$8B,$FC,$57
	.dc.b	$91,$FD,$71,$BE,$29,$A7,$1F,$12,$57,$F3,$B6,$E3,$F0,$71,$FE,$7B
	.dc.b	$18,$8F,$2C,$FF,$29,$38,$11,$E0,$E5,$B1,$C5,$FE,$E1,$6F,$9B,$F2
	.dc.b	$92,$BF,$B6,$3E,$DD,$A4,$7F,$4F,$6D,$7F,$7B,$F1,$3D,$A4,$9F,$EE
	.dc.b	$8F,$C0,$F9,$99,$1F,$DD,$DB,$3F,$E0,$36,$BF,$D8,$1A,$FF,$84,$3F
	.dc.b	$00,$7C,$8F,$ED,$7C,$D4,$62,$FF,$C1,$CD,$AF,$F0,$A3,$5F,$C5,$7E
	.dc.b	$19,$FE,$6D,$9F,$C6,$47,$F0,$48,$C3,$86,$BC,$3D,$8F,$BD,$E2,$CA
	.dc.b	$FE,$E0,$D9,$F9,$ED,$7E,$AB,$5F,$F1,$F7,$A9,$F3,$E7,$FC,$8F,$B7
	.dc.b	$F0,$98,$C4,$FF,$92,$23,$04,$9D,$92,$3F,$49,$CE,$FD,$6B,$FC,$DB
	.dc.b	$3F,$DE,$1B,$3A,$6B,$FA,$86,$BF,$D4,$FD,$EF,$E4,$0F,$F0,$76,$F6
	.dc.b	$FF,$AF,$EB,$FB,$84,$60,$8E,$09,$3F,$E7,$7F,$D6,$7F,$C2,$71,$FE
	.dc.b	$AF,$6C,$66,$3F,$94,$46,$23,$FC,$CA,$D7,$FC,$C8,$D5,$E1,$A3,$0D
	.dc.b	$9F,$D9,$78,$51,$FF,$12,$36,$BF,$46,$D7,$F5,$2D,$7F,$A2,$FC,$13
	.dc.b	$F0,$AF,$81,$1F,$EE,$D6,$BF,$EE,$C6,$BC,$1A,$FF,$33,$E4,$F9,$24
	.dc.b	$BF,$F0,$C3,$63,$FC,$3A,$D7,$FC,$2C,$FF,$0B,$F8,$69,$3F,$E3,$D6
	.dc.b	$CF,$EF,$1A,$FF,$02,$6B,$FD,$7B,$AF,$FC,$87,$F0,$DF,$C5,$49,$5F
	.dc.b	$E4,$76,$CF,$F0,$66,$BF,$C3,$1F,$EB,$9F,$82,$FF,$5A,$9F,$F7,$43
	.dc.b	$67,$FD,$68,$D7,$FD,$3A,$D7,$F9,$EF,$13,$EA,$FF,$1A,$DA,$FD,$8C
	.dc.b	$FF,$11,$9C,$92,$33,$3B,$9F,$DB,$B5,$FE,$88,$D7,$E0,$43,$E2,$3F
	.dc.b	$9A,$11,$FF,$33,$36,$3F,$92,$B5,$FE,$56,$FF,$3B,$FA,$C9,$3F,$E6
	.dc.b	$D6,$CF,$F2,$C6,$BF,$CC,$9A,$FE,$0F,$C4,$FF,$05,$6F,$F8,$B6,$EF
	.dc.b	$FD,$4E,$FF,$88,$CE,$5F,$87,$F3,$DE,$96,$E3,$FB,$48,$C6,$1F,$C6
	.dc.b	$1A,$FF,$65,$6B,$FD,$B1,$AF,$F8,$81,$A9,$CF,$88,$F9,$17,$BD,$FD
	.dc.b	$61,$B8,$7F,$26,$9D,$4E,$4D,$71,$C9,$D5,$7F,$2F,$C3,$FD,$78,$D7
	.dc.b	$FD,$76,$BE,$2F,$B1,$27,$8B,$E1,$D6,$47,$46,$D7,$F8,$0F,$1F,$47
	.dc.b	$C6,$F9,$C9,$5E,$26,$CE,$F0,$46,$23,$FC,$23,$5C,$AB,$FD,$55,$F5
	.dc.b	$1F,$DE,$C8,$E0,$F6,$FC,$4F,$C5,$23,$82,$57,$78,$CB,$63,$EF,$FE
	.dc.b	$25,$FC,$C2,$1C,$08,$C3,$7E,$2B,$F1,$31,$BC,$7E,$B7,$BD,$F5,$49
	.dc.b	$87,$D0,$6E,$1E,$F6,$BE,$ED,$11,$FE,$17,$FB,$29,$5F,$EC,$96,$D7
	.dc.b	$CD,$FC,$DF,$F9,$86,$06,$92,$BF,$C3,$AD,$7F,$0C,$D9,$FF,$08,$FD
	.dc.b	$07,$F8,$94,$93,$FE,$D7,$78,$F9,$B2,$38,$B6,$7C,$4D,$7E,$BD,$7F
	.dc.b	$90,$5F,$F3,$DD,$F9,$67,$FA,$BF,$AF,$2F,$FD,$8C,$DA,$FF,$65,$FD
	.dc.b	$0F,$F5,$C2,$4F,$FB,$2D,$B1,$A8,$C1,$1F,$F6,$9F,$AD,$FD,$90,$91
	.dc.b	$FE,$D3,$7E,$1F,$85,$23,$E1,$B6,$7F,$31,$F3,$8F,$9D,$FF,$22,$B6
	.dc.b	$BF,$5C,$D9,$DC,$E8,$93,$A6,$C7,$F9,$17,$E0,$5F,$25,$FE,$56,$6D
	.dc.b	$7F,$87,$BC,$54,$62,$B8,$56,$C9,$5B,$9D,$4F,$F9,$89,$AF,$F8,$A3
	.dc.b	$E7,$BF,$D5,$7F,$3C,$CB,$FC,$27,$78,$75,$49,$D3,$FE,$78,$9D,$3F
	.dc.b	$E7,$DA,$AF,$FA,$DB,$CE,$4E,$23,$FD,$73,$36,$73,$1D,$6F,$BD,$F6
	.dc.b	$8E,$57,$C5,$5F,$18,$C3,$63,$91,$E3,$29,$C8,$A7,$88,$B7,$E2,$AF
	.dc.b	$04,$65,$B1,$9B,$F8,$73,$FD,$82,$BF,$B0,$BC,$CC,$1F,$E5,$4F,$97
	.dc.b	$FF,$43,$7A,$BF,$D9,$9B,$19,$F6,$93,$96,$C7,$1D,$BF,$86,$B7,$B3
	.dc.b	$94,$62,$3F,$D5,$97,$F9,$39,$C9,$FE,$E6,$F9,$FF,$95,$0F,$8C,$FF
	.dc.b	$5B,$C6,$1B,$9F,$FA,$EA,$79,$5A,$B8,$E1,$E2,$78,$47,$5A,$E1,$7C
	.dc.b	$A3,$FD,$EB,$7C,$AB,$8E,$1F,$84,$8F,$14,$DA,$FF,$64,$EB,$A8,$FA
	.dc.b	$F3,$C7,$0E,$B1,$DC,$EB,$5D,$D4,$ED,$F0,$E3,$E7,$AE,$31,$EB,$6E
	.dc.b	$3D,$CF,$CD,$F0,$FE,$8A,$7F,$C9,$3B,$FF,$8C,$5F,$D0,$18,$11,$B7
	.dc.b	$FA,$0C,$6F,$83,$FD,$C7,$A7,$E7,$2B,$4D,$AE,$67,$03,$99,$87,$53
	.dc.b	$CA,$BF,$C8,$4F,$F6,$09,$F1,$57,$F0,$2F,$07,$51,$E4,$9D,$3D,$7A
	.dc.b	$1F,$3C,$F9,$FD,$E1,$C6,$B9,$DF,$0C,$BD,$0D,$CB,$F9,$AE,$1C,$6B
	.dc.b	$E1,$0F,$9E,$FF,$08,$3F,$E4,$26,$BF,$EB,$46,$BF,$F0,$05,$A8,$8B
	.dc.b	$E5,$7D,$71,$D5,$B1,$98,$F0,$BD,$B0,$FE,$13,$1E,$3F,$9E,$FD,$9C
	.dc.b	$9A,$48,$F3,$35,$F3,$35,$F3,$7C,$A7,$80,$8F,$82,$3E,$4F,$CE,$7F
	.dc.b	$88,$63,$F9,$A3,$5F,$F7,$F3,$5F,$71,$F8,$87,$0F,$F9,$F5,$AE,$7C
	.dc.b	$AF,$F3,$5E,$A9,$CB,$63,$3A,$3E,$9F,$ED,$B6,$FF,$0A,$D7,$F1,$D5
	.dc.b	$C5,$FC,$4D,$BC,$BE,$ED,$8F,$5C,$60,$8E,$08,$7B,$9F,$1F,$CC,$F8
	.dc.b	$89,$1D,$A3,$F2,$BA,$BE,$79,$EF,$FF,$A2,$23,$0D,$8C,$D7,$F9,$B1
	.dc.b	$AF,$CA,$3F,$C5,$EE,$1F,$D7,$B5,$FF,$60,$B5,$DC,$7F,$A5,$BD,$A7
	.dc.b	$D9,$BD,$1F,$D4,$C6,$1B,$19,$AF,$E0,$18,$79,$9F,$93,$D1,$F2,$08
	.dc.b	$FE,$09,$AF,$E0,$9A,$FE,$0B,$5F,$B9,$1F,$6B,$E4,$FE,$C9,$27,$A9
	.dc.b	$FC,$3F,$A7,$F8,$EF,$8B,$FE,$AE,$6C,$7F,$9E,$FB,$93,$96,$C6,$63
	.dc.b	$04,$70,$7F,$82,$F8,$9F,$65,$F8,$A2,$47,$86,$B9,$BF,$C2,$D7,$B9
	.dc.b	$FC,$43,$0F,$E6,$2D,$8C,$CF,$1F,$77,$5F,$9C,$BE,$5F,$D4,$D9,$FF
	.dc.b	$4A,$35,$FF,$42,$35,$F9,$19,$3E,$CF,$F3,$E4,$8F,$37,$B7,$F8,$EF
	.dc.b	$8F,$FD,$5F,$11,$A4,$EC,$FC,$83,$6B,$FD,$72,$D7,$F1,$DC,$DF,$86
	.dc.b	$F6,$FF,$06,$48,$FC,$1C,$FE,$FE,$FF,$9F,$64,$FA,$6F,$D9,$92,$BF
	.dc.b	$5F,$F2,$F5,$FD,$9C,$ED,$FD,$BC,$FF,$0C,$9F,$DD,$FC,$4C,$9F,$DD
	.dc.b	$BF,$C4,$3E,$1B,$11,$CA,$2E,$1F,$D3,$FC,$05,$7E,$73,$E8,$9E,$B5
	.dc.b	$26,$1F,$CE,$BE,$8F,$FC,$9E,$FF,$5D,$DD,$5C,$23,$0D,$7F,$AC,$46
	.dc.b	$AB,$F9,$BD,$72,$9F,$6F,$C3,$7F,$14,$23,$CA,$7F,$A8,$FE,$4B,$FE
	.dc.b	$5A,$8F,$EB,$6D,$7F,$8B,$B7,$3F,$67,$C1,$70,$F3,$65,$70,$7F,$99
	.dc.b	$C7,$4F,$4E,$4F,$47,$F9,$BF,$CB,$E2,$3A,$C5,$BF,$91,$3F,$D4,$3E
	.dc.b	$4B,$DD,$5F,$C3,$DA,$FF,$3A,$6B,$FD,$65,$A9,$A4,$CB,$FC,$8A,$F0
	.dc.b	$D5,$F3,$DF,$CD,$70,$FF,$30,$C6,$3C,$CF,$E2,$73,$9E,$6F,$06,$1F
	.dc.b	$23,$CE,$99,$7C,$87,$FA,$9F,$76,$23,$E4,$C5,$CB,$F0,$93,$F4,$3C
	.dc.b	$E9,$5F,$41,$FE,$43,$F4,$8F,$86,$CC,$FD,$29,$B7,$E6,$AD,$4F,$F8
	.dc.b	$0B,$67,$6F,$B8,$91,$FC,$DC,$F8,$CE,$1C,$47,$07,$47,$C9,$8D,$3F
	.dc.b	$C9,$F9,$1F,$41,$F8,$82,$47,$86,$7F,$17,$E9,$C8,$FC,$5B,$FC,$F3
	.dc.b	$E4,$7F,$C1,$11,$F8,$C8,$D3,$4F,$D0,$46,$B2,$FF,$1C,$61,$FE,$0D
	.dc.b	$E0,$FF,$0B,$79,$A8,$C5,$FF,$90,$9B,$1F,$DD,$FE,$1B,$F3,$BF,$0F
	.dc.b	$2B,$F3,$AF,$F8,$22,$3F,$C6,$FE,$F3,$E8,$BE,$15,$7F,$5D,$8C,$6B
	.dc.b	$F1,$1B,$1F,$2D,$AF,$F6,$3F,$37,$F2,$1D,$BF,$CC,$F8,$7F,$51,$AF
	.dc.b	$FB,$B1,$B5,$FE,$8E,$C7,$BE,$DE,$EF,$7B,$F6,$64,$9F,$EA,$15,$C3
	.dc.b	$6F,$81,$39,$9F,$DA,$BD,$F1,$FF,$11,$FD,$17,$F8,$28,$93,$FE,$24
	.dc.b	$7F,$C3,$7C,$1F,$EE,$78,$FD,$FF,$BF,$FC,$08,$E2,$FF,$85,$3A,$FF
	.dc.b	$C2,$64,$7F,$C5,$AF,$2F,$80,$FE,$12,$48,$E0,$7F,$0A,$70,$FF,$0A
	.dc.b	$7F,$C6,$F1,$8D,$3E,$EC,$5D,$BF,$C6,$53,$FE,$2A,$7F,$C5,$A4,$AF
	.dc.b	$EC,$CF,$AB,$E7,$7F,$8C,$92,$3F,$C5,$F1,$FE,$72,$F8,$3F,$EA,$1F
	.dc.b	$33,$F9,$C6,$CF,$E8,$DA,$FF,$86,$DA,$FF,$91,$9A,$FF,$BA,$7C,$D4
	.dc.b	$F3,$6C,$73,$6B,$FE,$25,$E0,$FC,$B7,$C0,$FF,$2E,$24,$7F,$A6,$6F
	.dc.b	$C3,$A7,$F1,$0F,$A2,$7C,$D7,$F8,$95,$B5,$FB,$9F,$15,$18,$8F,$EC
	.dc.b	$AD,$7F,$CB,$B1,$A9,$E4,$48,$E5,$3B,$7F,$57,$C2,$FA,$DF,$E8,$64
	.dc.b	$8F,$EA,$53,$DF,$0F,$D9,$CE,$67,$FA,$2B,$FE,$38,$9D,$6D,$FE,$8B
	.dc.b	$C5,$F3,$DF,$E0,$9B,$C3,$54,$6B,$C4,$FF,$22,$4E,$63,$13,$F0,$63
	.dc.b	$04,$9F,$CD,$FC,$AF,$F5,$A2,$4F,$F6,$99,$FF,$0D,$65,$FC,$F3,$F7
	.dc.b	$E7,$D5,$7F,$94,$5B,$5F,$35,$AF,$F8,$41,$AF,$E0,$FF,$74,$FE,$42
	.dc.b	$FE,$25,$AF,$E2,$5A,$EF,$C5,$FE,$56,$BE,$97,$FC,$DF,$D9,$FC,$B9
	.dc.b	$B1,$FC,$D5,$AF,$F9,$B5,$AF,$FA,$61,$7C,$8B,$93,$93,$67,$93,$5E
	.dc.b	$4D,$7E,$A7,$EE,$CF,$A1,$FF,$5B,$C6,$1A,$FD,$48,$D7,$17,$E4,$5B
	.dc.b	$1F,$E3,$4F,$4D,$F2,$7C,$39,$11,$8B,$FC,$C3,$6B,$85,$76,$EA,$FE
	.dc.b	$75,$D2,$BF,$F0,$8B,$6B,$FC,$3B,$E9,$7F,$9B,$27,$27,$F8,$95,$C4
	.dc.b	$7F,$91,$9A,$FE,$89,$AF,$E4,$DA,$FE,$41,$B8,$9A,$4F,$84,$F3,$75
	.dc.b	$BE,$C3,$C9,$CF,$FD,$51,$B8,$FE,$73,$E5,$2E,$5F,$D4,$E6,$E9,$FD
	.dc.b	$5D,$B8,$73,$BF,$CB,$5E,$27,$8E,$7F,$A8,$8F,$99,$D2,$69,$2B,$F4
	.dc.b	$8D,$C7,$FC,$CF,$3B,$9C,$9F,$E0,$1C,$7F,$B4,$70,$F8,$8F,$3F,$1F
	.dc.b	$F9,$D7,$8F,$15,$FC,$57,$BF,$F1,$EF,$F1,$D8,$FC,$2C,$EE,$30,$4C
	.dc.b	$7F,$95,$57,$06,$D7,$08,$C4,$7F,$82,$FE,$C3,$E1,$9F,$1A,$9F,$49
	.dc.b	$87,$FD,$D2,$FE,$1D,$C9,$FE,$A7,$70,$7D,$6F,$B0,$F9,$A7,$3F,$0E
	.dc.b	$CF,$F9,$72,$2D,$FF,$A0,$26,$F1,$7F,$83,$A6,$D7,$F8,$38,$FE,$7E
	.dc.b	$71,$3F,$DF,$73,$E3,$23,$FE,$B9,$7F,$0F,$EA,$FD,$E3,$D9,$9D,$96
	.dc.b	$1F,$E9,$08,$F6,$12,$3F,$D8,$4F,$E4,$3B,$FF,$0E,$7E,$33,$CD,$FF
	.dc.b	$2E,$6F,$8A,$FF,$3B,$CD,$B6,$7F,$A9,$BD,$D6,$8F,$F0,$5D,$F3,$F3
	.dc.b	$3F,$77,$5F,$AF,$F3,$9F,$4B,$FE,$F3,$BD,$9E,$89,$D4,$67,$1F,$F1
	.dc.b	$33,$5F,$68,$AE,$7F,$1D,$FC,$31,$B7,$F9,$3F,$97,$F8,$6F,$53,$FC
	.dc.b	$7A,$E2,$BF,$D6,$BF,$21,$FA,$A8,$C6,$3F,$C1,$79,$91,$8B,$E3,$7F
	.dc.b	$8B,$BE,$0F,$1E,$69,$3F,$E4,$EF,$1E,$CF,$C8,$8F,$0C,$E3,$1F,$F4
	.dc.b	$1F,$AC,$BC,$57,$9B,$AB,$C8,$7E,$AB,$E9,$3F,$CF,$11,$EB,$C5,$A7
	.dc.b	$9A,$B5,$CB,$FA,$AB,$DD,$7F,$84,$5A,$FF,$84,$5A,$FE,$23,$34,$62
	.dc.b	$3F,$A1,$36,$33,$3D,$4F,$E5,$DA,$BE,$A7,$E5,$1F,$C3,$BE,$93,$F9
	.dc.b	$86,$7E,$7C,$FE,$5C,$EA,$31,$5C,$5F,$F6,$A5,$70,$CB,$E4,$1E,$E9
	.dc.b	$D3,$8B,$FA,$BE,$27,$D9,$3F,$7D,$E1,$5C,$3A,$C5,$AF,$F7,$9C,$DE
	.dc.b	$0F,$BD,$87,$FB,$A2,$FC,$7E,$F7,$F2,$E9,$7D,$17,$FB,$BA,$72,$7D
	.dc.b	$18,$CF,$C2,$FF,$5C,$6C,$7F,$A8,$DB,$8F,$E6,$0F,$D4,$D7,$37,$11
	.dc.b	$A7,$F3,$AD,$7F,$49,$1B,$72,$D7,$F3,$7F,$37,$F9,$93,$F8,$F7,$C7
	.dc.b	$DB,$F9,$B1,$67,$E4,$42,$71,$7F,$0B,$EB,$FF,$44,$7F,$27,$F9,$2B
	.dc.b	$5F,$5A,$2D,$F2,$9B,$5F,$2B,$E7,$7F,$54,$7F,$4E,$FA,$5F,$ED,$31
	.dc.b	$9C,$3F,$31,$E4,$A7,$31,$FE,$3B,$6E,$1F,$C6,$BA,$7F,$BF,$CF,$09
	.dc.b	$CD,$71,$E1,$EC,$7F,$64,$BF,$D5,$79,$D8,$BF,$C6,$FC,$07,$EF,$4F
	.dc.b	$E8,$5F,$23,$63,$24,$AF,$8B,$19,$7F,$1D,$C9,$FC,$4F,$4F,$C3,$F8
	.dc.b	$1F,$D2,$FE,$97,$F8,$43,$30,$FC,$DB,$96,$9F,$03,$E1,$9F,$15,$FD
	.dc.b	$C9,$B1,$F8,$BE,$67,$CD,$E6,$BF,$25,$18,$9F,$C6,$37,$3F,$C4,$B5
	.dc.b	$FC,$2D,$7E,$4E,$31,$1F,$8A,$8C,$12,$7F,$93,$7C,$9F,$F2,$13,$FA
	.dc.b	$17,$C4,$7F,$BF,$A3,$F6,$B1,$FD,$17,$F3,$EF,$F1,$75,$FF,$32,$F2
	.dc.b	$D1,$88,$FE,$A3,$18,$EE,$FF,$1F,$B6,$FF,$45,$28,$61,$F3,$67,$8B
	.dc.b	$67,$4B,$0F,$2E,$BF,$CD,$F2,$97,$FC,$8F,$0E,$71,$8C,$3F,$84,$C6
	.dc.b	$27,$C2,$D8,$FF,$76,$61,$F1,$9E,$F5,$E8,$8C,$4E,$E7,$E0,$46,$0E
	.dc.b	$1D,$FF,$D1,$BE,$6B,$FA,$51,$FE,$0F,$FA,$5B,$8F,$F2,$10,$FF,$83
	.dc.b	$C8,$FF,$8A,$4C,$95,$CD,$FF,$AF,$E3,$2D,$8F,$F3,$27,$03,$DC,$FC
	.dc.b	$61,$FC,$67,$E9,$7F,$A0,$F2,$7F,$97,$FA,$95,$BC,$7F,$8D,$57,$F3
	.dc.b	$FA,$34,$9E,$37,$F1,$D6,$CF,$F1,$D6,$E5,$FE,$D3,$E7,$7C,$B7,$F6
	.dc.b	$B3,$FC,$61,$F4,$B6,$FE,$13,$FE,$2F,$23,$F0,$5B,$5E,$F6,$BF,$E0
	.dc.b	$0D,$FF,$CF,$51,$FD,$43,$C5,$FE,$F0,$7F,$8F,$7E,$96,$DF,$CF,$7F
	.dc.b	$C7,$84,$7E,$73,$6F,$E6,$35,$ED,$08,$FF,$4B,$C3,$E9,$DE,$3F,$F0
	.dc.b	$F6,$D7,$F7,$39,$D7,$91,$9A,$D3,$CA,$BF,$A7,$4F,$E8,$9A,$FF,$45
	.dc.b	$6B,$FE,$1E,$6B,$FE,$3C,$6B,$DA,$BF,$56,$D3,$A7,$95,$E5,$3C,$DE
	.dc.b	$6D,$9F,$35,$7F,$9E,$6B,$C5,$3E,$52,$4F,$F9,$EF,$E6,$1E,$47,$FA
	.dc.b	$39,$B1,$FA,$FC,$9F,$1B,$9A,$93,$FD,$FE,$B7,$87,$EC,$AF,$73,$9A
	.dc.b	$FF,$4C,$F3,$BC,$BA,$62,$2C,$FF,$82,$67,$4D,$9D,$34,$FA,$57,$FE
	.dc.b	$62,$E3,$FF,$31,$FD,$2F,$D8,$8F,$F2,$4F,$9F,$FE,$C1,$8C,$47,$F9
	.dc.b	$32,$77,$97,$F7,$7A,$D3,$FC,$62,$33,$3C,$AE,$BC,$45,$F4,$7F,$DD
	.dc.b	$51,$A7,$C8,$7F,$42,$AE,$D1,$FE,$78,$F1,$78,$4F,$F9,$40,$D1,$DC
	.dc.b	$A7,$7A,$ED,$87,$9A,$3F,$C1,$D8,$77,$8C,$46,$F0,$EE,$F9,$9F,$63
	.dc.b	$4E,$D5,$E2,$AF,$C6,$C7,$A0,$FC,$7B,$87,$CC,$7F,$87,$7E,$79,$F1
	.dc.b	$FF,$E7,$58,$FE,$7F,$18,$F5,$7F,$2E,$D8,$CC,$ED,$E5,$C1,$FE,$8E
	.dc.b	$78,$B5,$CF,$67,$28,$C4,$7F,$A1,$2B,$F2,$73,$F2,$F2,$C7,$C8,$FF
	.dc.b	$A2,$35,$FB,$F4,$8C,$4F,$1C,$3C,$15,$A6,$C7,$2C,$3A,$1F,$EA,$EB
	.dc.b	$E4,$E0,$3E,$27,$EB,$B6,$BC,$37,$E0,$9E,$4D,$78,$DF,$81,$EC,$D7
	.dc.b	$AC,$6E,$BC,$31,$DA,$75,$1D,$EF,$D1,$3E,$2A,$E1,$5D,$7A,$BF,$D8
	.dc.b	$FF,$2D,$FC,$29,$B5,$FE,$FB,$11,$96,$99,$70,$47,$07,$F6,$EF,$A5
	.dc.b	$FF,$4E,$37,$D2,$FE,$2F,$E3,$27,$33,$FE,$65,$8C,$4E,$A7,$33,$FE
	.dc.b	$0F,$6B,$FE,$0E,$6B,$FE,$0B,$6A,$FA,$57,$C5,$FE,$69,$B1,$FC,$A1
	.dc.b	$AF,$F8,$51,$AF,$F9,$DD,$7C,$C7,$12,$47,$17,$F9,$8F,$63,$F1,$CF
	.dc.b	$D3,$B6,$73,$18,$23,$82,$1F,$D0,$DF,$D7,$F8,$EF,$A4,$77,$FF,$3E
	.dc.b	$8C,$36,$33,$5F,$E2,$C9,$CB,$FE,$1E,$FB,$07,$C6,$7D,$06,$C6,$63
	.dc.b	$04,$70,$43,$EA,$47,$07,$FD,$35,$C2,$F0,$7F,$CA,$F1,$FD,$1F,$CE
	.dc.b	$FC,$54,$62,$36,$D7,$FC,$91,$F4,$6E,$FF,$C1,$B6,$3F,$C6,$AD,$7F
	.dc.b	$08,$D7,$8C,$70,$23,$C1,$FC,$D7,$73,$E7,$A9,$27,$FC,$A3,$8F,$FB
	.dc.b	$F2,$3F,$27,$D6,$F8,$9A,$4A,$FF,$23,$C6,$5B,$19,$7F,$31,$19,$6C
	.dc.b	$7C,$2F,$42,$47,$F6,$B6,$BE,$CF,$4F,$E0,$91,$F8,$2D,$74,$60,$87
	.dc.b	$F8,$BF,$7F,$79,$1F,$6E,$5F,$5D,$FF,$3B,$60,$F7,$7D,$E4,$8F,$6B
	.dc.b	$5F,$5B,$5E,$2F,$D6,$F9,$49,$1F,$59,$AF,$D6,$F9,$AF,$B6,$47,$ED
	.dc.b	$BF,$C0,$3E,$03,$FC,$57,$DD,$FE,$D1,$F2,$30,$4A,$FF,$3A,$CB,$E2
	.dc.b	$B8,$12,$32,$FF,$54,$FB,$07,$D1,$52,$4F,$F8,$92,$32,$D8,$CB,$FE
	.dc.b	$92,$8C,$B6,$3F,$AF,$FC,$4B,$D1,$A4,$9F,$F1,$7C,$65,$F9,$FA,$7E
	.dc.b	$40,$8E,$08,$7F,$99,$5C,$B4,$FF,$17,$3F,$A8,$23,$82,$1F,$AA,$72
	.dc.b	$D3,$FD,$57,$E5,$3F,$0B,$7A,$5F,$E4,$48,$C3,$63,$33,$C1,$A8,$82
	.dc.b	$78,$BB,$9E,$44,$7F,$88,$7B,$2F,$8C,$FE,$2C,$D8,$CC,$60,$CF,$B5
	.dc.b	$23,$8B,$5F,$08,$F0,$7F,$D2,$F1,$98,$F2,$C6,$1E,$27,$E8,$3E,$7F
	.dc.b	$F8,$09,$1C,$10,$FF,$4B,$39,$69,$F8,$7F,$A9,$FC,$61,$1C,$10,$FC
	.dc.b	$43,$96,$9F,$84,$F8,$5F,$C4,$91,$C1,$FB,$4E,$6F,$E2,$A4,$70,$43
	.dc.b	$FD,$54,$E5,$A7,$F3,$DF,$8B,$7D,$0F,$F4,$36,$C6,$63,$04,$70,$43
	.dc.b	$E6,$E2,$9E,$31,$FA,$3F,$80,$FD,$61,$1F,$D5,$35,$D1,$82,$1F,$E6
	.dc.b	$7F,$95,$FF,$53,$F8,$86,$93,$AB,$FA,$2E,$B2,$BF,$D1,$CD,$7F,$A3
	.dc.b	$B5,$FF,$60,$78,$C8,$C4,$7E,$5F,$C5,$78,$0F,$A0,$FE,$82,$49,$FE
	.dc.b	$46,$D7,$FD,$AC,$D7,$FD,$99,$1C,$5F,$F5,$5F,$D8,$B8,$7F,$CC,$2D
	.dc.b	$8C,$C6,$0F,$F5,$8F,$B4,$FA,$2F,$F3,$CC,$A3,$1E,$CE,$A7,$31,$89
	.dc.b	$FE,$3F,$F0,$1F,$C9,$48,$E0,$87,$F8,$97,$C3,$FF,$6A,$38,$3F,$6F
	.dc.b	$E2,$23,$31,$BE,$8F,$E6,$04,$7F,$96,$BF,$E3,$FF,$CA,$7F,$DB,$BB
	.dc.b	$9A,$4E,$2F,$E7,$1F,$53,$2F,$F9,$AB,$5F,$F0,$1C,$FF,$61,$7F,$DE
	.dc.b	$CD,$7E,$B3,$55,$77,$F3,$3C,$44,$8F,$EE,$B3,$B9,$D4,$62,$74,$F8
	.dc.b	$1A,$FC,$2C,$5F,$1D,$FD,$3C,$91,$FD,$31,$F8,$1E,$44,$8E,$ED,$7F
	.dc.b	$DE,$5E,$5F,$E0,$DB,$CB,$FE,$3E,$8C,$47,$FB,$D5,$AF,$FB,$D5,$AB
	.dc.b	$D5,$FF,$18,$CF,$12,$47,$F8,$DF,$E8,$5F,$3B,$FE,$3D,$6C,$66,$30
	.dc.b	$78,$7A,$64,$7F,$26,$3B,$61,$FA,$57,$18,$7F,$90,$27,$71,$EC,$6B
	.dc.b	$E9,$7F,$C9,$9F,$38,$F2,$69,$27,$FC,$99,$19,$6C,$65,$FE,$17,$19
	.dc.b	$6C,$7F,$99,$3E,$69,$C7,$49,$3E,$68,$CB,$63,$2F,$F4,$2C,$13,$A8
	.dc.b	$C3,$5F,$E8,$77,$FE,$1F,$7F,$AD,$78,$48,$C4,$7F,$4B,$6C,$67,$2F
	.dc.b	$F6,$EF,$CF,$BE,$AA,$92,$FE,$70,$8D,$38,$7F,$5E,$FF,$58,$6B,$FD
	.dc.b	$A9,$AF,$F6,$D6,$BC,$FE,$65,$F1,$3C,$DB,$1F,$76,$51,$88,$EC,$D7
	.dc.b	$B7,$CE,$7F,$99,$5E,$CD,$7D,$0D,$93,$9F,$37,$FD,$0B,$B7,$B9,$B5
	.dc.b	$F8,$D6,$BF,$DD,$1A,$FF,$81,$97,$C7,$7D,$38,$FF,$05,$46,$27,$A4
	.dc.b	$E6,$7F,$C2,$AE,$F4,$FF,$4A,$F3,$DD,$7F,$C8,$95,$FD,$59,$AF,$F8
	.dc.b	$01,$AF,$F8,$5F,$81,$FB,$D7,$F8,$DF,$1F,$D0,$B6,$7F,$14,$D7,$FC
	.dc.b	$3B,$3C,$3D,$3F,$43,$6B,$8C,$62,$B8,$57,$76,$BF,$E1,$5B,$7C,$C7
	.dc.b	$A4,$97,$FE,$B3,$8C,$B6,$33,$3E,$D6,$BE,$4F,$B1,$EB,$89,$A4,$C7
	.dc.b	$CE,$DB,$FD,$FB,$6B,$F6,$7F,$03,$F3,$79,$3F,$D9,$35,$FB,$A8,$C1
	.dc.b	$C8,$8F,$26,$CF,$EC,$6B,$8B,$63,$2F,$9D,$FC,$07,$77,$F4,$9E,$1B
	.dc.b	$DF,$F8,$EC,$3F,$9D,$C6,$0F,$C3,$78,$4F,$CF,$DF,$19,$86,$D7,$F1
	.dc.b	$D9,$D4,$E4,$EB,$CF,$27,$FA,$3B,$73,$FE,$A6,$D7,$FA,$9B,$5F,$EA
	.dc.b	$6D,$7F,$A9,$E3,$FA,$E9,$E6,$78,$70,$FF,$17,$B7,$D1,$F3,$11,$FE
	.dc.b	$AF,$C3,$F0,$08,$E0,$E5,$F1,$9E,$D2,$38,$3B,$7C,$6F,$C1,$23,$82
	.dc.b	$63,$FE,$64,$D3,$E4,$4E,$A3,$18,$A3,$F6,$5E,$4B,$E7,$9F,$03,$86
	.dc.b	$E3,$FD,$26,$75,$39,$3F,$6D,$F6,$3F,$DA,$A7,$F9,$0B,$71,$FE,$A0
	.dc.b	$D7,$FC,$EE,$D7,$FB,$FF,$57,$E0,$89,$3F,$DF,$1B,$1F,$E0,$86,$BF
	.dc.b	$E0,$87,$F1,$71,$F4,$E3,$FD,$07,$C7,$FD,$2B,$87,$24,$9F,$F7,$D9
	.dc.b	$B9,$FF,$4C,$CE,$E3,$33,$B9,$FF,$46,$FA,$8F,$D7,$FF,$D4,$15,$FD
	.dc.b	$AE,$31,$3F,$C6,$DB,$19,$9D,$CF,$F6,$EF,$17,$F8,$E7,$5F,$F9,$56
	.dc.b	$BF,$A3,$B6,$75,$18,$9D,$91,$C1,$FE,$7D,$F8,$EF,$E4,$9E,$C1,$A4
	.dc.b	$8D,$12,$BF,$CF,$73,$B8,$C4,$7F,$8C,$BE,$0B,$F6,$64,$8F,$F1,$6B
	.dc.b	$63,$FC,$72,$D7,$84,$7C,$3F,$A0,$7B,$9F,$0D,$B1,$98,$C1,$F0,$BE
	.dc.b	$56,$47,$C2,$6D,$7F,$85,$1A,$FF,$7D,$6B,$FE,$46,$F9,$EF,$E1,$24
	.dc.b	$9F,$F2,$1B,$63,$FC,$9E,$D7,$FC,$9E,$FF,$4B,$FA,$0F,$F0,$7C,$FF
	.dc.b	$89,$7C,$64,$62,$BF,$D4,$2D,$7F,$D9,$6D,$7F,$C1,$2D,$7F,$DA,$4D
	.dc.b	$8E,$73,$DC,$FE,$B5,$F3,$0F,$80,$A4,$8F,$14,$65,$FF,$7B,$FE,$F9
	.dc.b	$FA,$7C,$3F,$DD,$CD,$F8,$43,$49,$EB,$7F,$6D,$8C,$B6,$33,$5F,$A2
	.dc.b	$F0,$7F,$52,$47,$04,$BF,$D5,$BF,$D5,$1B,$1B,$7F,$CF,$7F,$41,$FC
	.dc.b	$E8,$93,$FE,$74,$6C,$6A,$30,$47,$FD,$F3,$83,$E8,$1B,$FA,$5B,$5F
	.dc.b	$E5,$99,$D4,$E4,$FD,$5F,$CF,$49,$FD,$5B,$6F,$FC,$BA,$D7,$FC,$8C
	.dc.b	$D4,$3C,$3F,$E9,$DC,$5F,$D3,$49,$5F,$CF,$5F,$EA,$7F,$41,$FB,$CA
	.dc.b	$FF,$55,$B5,$E7,$48,$C4,$75,$6C,$67,$CA,$E6,$4B,$FE,$B5,$3F,$DA
	.dc.b	$AB,$AF,$D0,$78,$09,$3A,$7F,$B2,$FD,$03,$2A,$49,$FE,$C9,$19,$6C
	.dc.b	$66,$BF,$A3,$4E,$5B,$1F,$DD,$3E,$22,$57,$FA,$B9,$B3,$FD,$37,$07
	.dc.b	$C8,$7C,$42,$57,$A9,$B3,$B6,$BF,$21,$FE,$73,$52,$7F,$9A,$B6,$7C
	.dc.b	$BF,$23,$EE,$24,$FB,$1B,$1A,$8C,$11,$FF,$67,$F2,$FF,$83,$89,$1F
	.dc.b	$EC,$87,$FC,$1B,$FA,$17,$EE,$76,$34,$93,$A2,$1F,$E0,$F7,$2F,$F4
	.dc.b	$C9,$FE,$A6,$47,$07,$F8,$57,$EC,$FB,$11,$FF,$0E,$B5,$D1,$82,$1F
	.dc.b	$C5,$7E,$C1,$F2,$DF,$E6,$36,$C6,$63,$07,$F9,$5B,$E6,$73,$97,$FA
	.dc.b	$95,$AF,$F9,$71,$AF,$F9,$95,$AF,$FB,$4D,$BE,$21,$A4,$9F,$EB,$7E
	.dc.b	$BC,$BF,$EB,$6D,$BF,$F6,$6F,$1F,$F6,$02,$4F,$F6,$07,$FC,$7B,$F1
	.dc.b	$5F,$E2,$89,$FF,$0C,$35,$FD,$A9,$FB,$57,$0F,$EB,$9A,$E7,$D3,$E6
	.dc.b	$4A,$FF,$22,$4E,$BC,$BF,$F5,$4B,$FE,$52,$F7,$8D,$24,$FF,$B9,$E7
	.dc.b	$6D,$9D,$B6,$77,$7F,$EC,$01,$F1,$BF,$C3,$3D,$7E,$B1,$EA,$F3,$7F
	.dc.b	$DB,$31,$88,$F6,$CE,$EB,$FC,$1F,$76,$7E,$09,$D0,$FF,$30,$C6,$9A
	.dc.b	$6A,$78,$3F,$C7,$2F,$BC,$FF,$C0,$19,$CC,$A7,$2F,$13,$99,$C6,$B6
	.dc.b	$F8,$6F,$D0,$F1,$8C,$D7,$4A,$E7,$3E,$C3,$99,$B2,$9C,$EF,$D5,$7D
	.dc.b	$23,$DC,$7F,$BA,$AB,$E6,$38,$3D,$47,$C4,$1F,$2D,$FE,$5B,$F4,$FF
	.dc.b	$DE,$2D,$8D,$35,$F9,$6D,$7F,$1A,$BE,$7D,$E0,$F2,$6C,$F2,$6B,$C9
	.dc.b	$AF,$27,$78,$7D,$DA,$7B,$1F,$9A,$8C,$37,$2E,$18,$7D,$8A,$F1,$36
	.dc.b	$75,$87,$F5,$69,$E6,$EF,$0E,$F7,$CA,$3E,$CC,$E9,$B5,$C9,$F8,$F5
	.dc.b	$C1,$E8,$D7,$E5,$C6,$E7,$96,$1E,$BA,$D4,$79,$AB,$A9,$E7,$3C,$98
	.dc.b	$7E,$67,$6F,$ED,$BF,$3F,$F5,$E5,$18,$7F,$5C,$23,$51,$98,$FF,$3D
	.dc.b	$99,$37,$1A,$70,$69,$ED,$D4,$F9,$FF,$F3,$24,$FF,$B1,$5A,$FF,$A9
	.dc.b	$5A,$FF,$0E,$8E,$1E,$8F,$29,$CC,$FF,$7C,$6B,$FE,$07,$6C,$F0,$8F
	.dc.b	$77,$CF,$3E,$7F,$F8,$23,$63,$31,$83,$3F,$3F,$23,$8B,$5F,$E7,$07
	.dc.b	$FA,$1B,$AB,$F8,$77,$5F,$F3,$38,$C1,$0F,$D2,$39,$6C,$70,$3F,$8B
	.dc.b	$79,$EF,$9F,$79,$1F,$56,$30,$D8,$CD,$7F,$99,$5A,$FF,$B3,$7E,$7D
	.dc.b	$F0,$9F,$41,$B1,$FE,$F3,$6B,$B8,$D1,$1D,$35,$FF,$5C,$FC,$BF,$F5
	.dc.b	$E7,$F9,$93,$5C,$F8,$3F,$86,$8C,$47,$D9,$F9,$FF,$E3,$8F,$F0,$5E
	.dc.b	$04,$62,$3F,$DC,$8D,$78,$76,$FE,$22,$31,$1F,$E3,$A6,$BF,$E4,$56
	.dc.b	$CF,$49,$FC,$BF,$6B,$C7,$A4,$BF,$E9,$51,$97,$C9,$D2,$F9,$2F,$CB
	.dc.b	$B6,$33,$18,$23,$82,$1F,$EF,$D8,$E0,$FC,$8F,$80,$F8,$84,$7E,$4B
	.dc.b	$ED,$F2,$E4,$7C,$5C,$FF,$C7,$2D,$8C,$BF,$19,$AF,$CD,$7D,$FB,$48
	.dc.b	$F6,$B5,$F2,$7B,$5E,$C2,$3F,$01,$AE,$8C,$10,$FA,$BF,$1D,$F4,$C8
	.dc.b	$F9,$DF,$AD,$EC,$48,$F5,$35,$FE,$CF,$BB,$E0,$1E,$37,$F0,$46,$CF
	.dc.b	$E1,$5A,$F0,$6B,$FC,$0D,$7C,$97,$12,$47,$17,$FC,$01,$F3,$EF,$9F
	.dc.b	$A4,$9F,$E4,$11,$96,$C6,$5F,$EB,$51,$96,$C7,$F8,$2F,$E7,$DE,$75
	.dc.b	$24,$FF,$2B,$8C,$BF,$8E,$D3,$F2,$64,$70,$43,$FC,$17,$E2,$FF,$85
	.dc.b	$A3,$51,$8E,$54,$7E,$B4,$91,$FB,$26,$BF,$B6,$6B,$FE,$36,$97,$47
	.dc.b	$28,$E0,$7F,$AE,$7E,$7D,$F2,$1F,$D0,$FC,$84,$62,$75,$39,$8C,$47
	.dc.b	$FB,$19,$AF,$F9,$12,$78,$CE,$C9,$1F,$EF,$BF,$9F,$7C,$FF,$F1,$D6
	.dc.b	$C6,$63,$06,$7D,$B9,$1C,$5A,$FF,$8C,$BC,$35,$27,$33,$FE,$56,$9C
	.dc.b	$C6,$23,$FD,$9B,$9F,$F9,$13,$E1,$BF,$83,$92,$3F,$82,$B5,$FA,$4D
	.dc.b	$7E,$66,$C8,$FC,$7F,$CD,$49,$FC,$6B,$5F,$C5,$B5,$FF,$24,$F1,$A3
	.dc.b	$11,$FE,$6C,$E1,$7C,$EF,$E3,$C9,$1F,$8D,$7F,$8E,$F4,$C8,$FE,$36
	.dc.b	$D7,$F1,$0D,$7C,$ED,$7F,$DD,$BA,$3E,$21,$EB,$7F,$4F,$6C,$FF,$8E
	.dc.b	$E7,$53,$92,$46,$67,$6F,$D3,$6B,$F4,$5A,$FF,$A2,$32,$7B,$9F,$B6
	.dc.b	$24,$7E,$D1,$AF,$EC,$9A,$FC,$87,$FC,$31,$E2,$BE,$7F,$F8,$2C,$FF
	.dc.b	$9B,$5A,$FF,$9A,$DB,$1B,$ED,$FE,$AD,$BC,$AF,$EB,$6D,$7F,$AF,$B5
	.dc.b	$FD,$57,$47,$DF,$76,$7F,$4A,$F9,$BF,$E9,$E4,$70,$43,$F7,$4E,$5A
	.dc.b	$7F,$8C,$7E,$7F,$EF,$B7,$83,$FD,$31,$18,$8F,$F3,$63,$5F,$F5,$8F
	.dc.b	$B2,$F9,$FF,$F3,$B4,$23,$1E,$D7,$F8,$F5,$B1,$FE,$75,$6B,$FC,$B7
	.dc.b	$E7,$FF,$99,$11,$FE,$64,$D7,$46,$08,$7F,$35,$E6,$FE,$74,$47,$F9
	.dc.b	$D3,$FE,$48,$F9,$4F,$A7,$D2,$69,$37,$FE,$85,$F3,$F2,$FF,$A1,$35
	.dc.b	$FE,$4B,$3F,$DB,$1F,$E3,$8D,$7F,$76,$D7,$FC,$B6,$DA,$F0,$79,$BE
	.dc.b	$32,$4F,$FB,$2E,$74,$48,$CC,$78,$1F,$EA,$DF,$19,$23,$FA,$B3,$5F
	.dc.b	$2F,$8D,$EF,$23,$D5,$FF,$37,$7A,$BF,$E3,$4E,$2B,$E5,$9A,$4A,$FF
	.dc.b	$23,$C6,$5B,$19,$7F,$DA,$33,$CD,$FD,$B7,$CF,$BE,$7F,$FC,$90,$D8
	.dc.b	$CC,$60,$F0,$F3,$C8,$EE,$D7,$AE,$1F,$9F,$6C,$65,$FF,$80,$37,$FE
	.dc.b	$4E,$8F,$F2,$B7,$CD,$7F,$4D,$9F,$E6,$8D,$7F,$99,$35,$FE,$22,$D5
	.dc.b	$DE,$8C,$37,$C9,$EA,$49,$FF,$35,$FC,$FF,$EF,$B0,$34,$9C,$7F,$41
	.dc.b	$B3,$C2,$30,$7A,$32,$91,$FC,$7F,$1F,$F3,$2B,$E8,$8F,$F5,$32,$F9
	.dc.b	$E7,$CF,$E1,$B1,$FE,$AB,$6B,$FE,$E3,$6A,$21,$CF,$9F,$91,$1E,$4F
	.dc.b	$5F,$97,$7C,$37,$56,$CF,$56,$BD,$0F,$33,$87,$BD,$3E,$73,$FC,$D2
	.dc.b	$DF,$07,$D1,$18,$3F,$CD,$BE,$6F,$FA,$2B,$5F,$F6,$33,$6F,$FD,$FC
	.dc.b	$D7,$C0,$D7,$DA,$BE,$7F,$F9,$9B,$67,$9C,$60,$FE,$73,$C5,$FE,$98
	.dc.b	$F4,$EE,$FF,$E4,$7A,$FE,$A4,$DC,$3F,$C0,$0D,$7F,$27,$F3,$FF,$E3
	.dc.b	$49,$FF,$57,$B6,$3F,$4E,$D7,$7E,$9F,$F4,$8E,$44,$E6,$7E,$B4,$62
	.dc.b	$75,$39,$94,$E6,$BC,$BF,$2B,$E8,$23,$82,$4F,$A2,$32,$D8,$FF,$62
	.dc.b	$FC,$FF,$5A,$34,$98,$79,$DB,$3C,$23,$07,$CA,$E5,$FF,$64,$C7,$E6
	.dc.b	$FE,$07,$04,$7B,$BC,$C8,$E0,$98,$7F,$0E,$BE,$6D,$9D,$BF,$CD,$38
	.dc.b	$3F,$B8,$F8,$3F,$CF,$5B,$E0,$FE,$3F,$0F,$E5,$D1,$FD,$DB,$D8,$7C
	.dc.b	$FB,$E2,$F0,$DA,$FF,$63,$37,$1F,$F4,$4B,$5F,$E8,$3D,$1D,$89,$3F
	.dc.b	$CE,$DB,$1F,$D2,$1A,$F0,$8D,$11,$D3,$5F,$DA,$DF,$9B,$D1,$F2,$92
	.dc.b	$7C,$AF,$BB,$86,$47,$F4,$B6,$D7,$A3,$E3,$3D,$A4,$9F,$6B,$F0,$3E
	.dc.b	$7E,$47,$C0,$6C,$FC,$06,$BE,$4C,$91,$88,$FE,$19,$E5,$3E,$7D,$F0
	.dc.b	$3F,$E2,$48,$C3,$63,$38,$7F,$80,$DB,$5F,$C2,$47,$CF,$BC,$AF,$F1
	.dc.b	$5B,$6B,$53,$A9,$C9,$F8,$0E,$B9,$3F,$DB,$5B,$3F,$DE,$9A,$FF,$7A
	.dc.b	$6B,$FE,$74,$87,$07,$E3,$1A,$79,$79,$3F,$D9,$31,$92,$46,$4D,$CF
	.dc.b	$FA,$B1,$AF,$FA,$8D,$AF,$F3,$3E,$F7,$CF,$FF,$35,$9F,$85,$18,$9F
	.dc.b	$F2,$6B,$63,$33,$B9,$F9,$1D,$7F,$E8,$39,$FE,$FF,$E3,$FF,$55,$23
	.dc.b	$82,$4F,$08,$CB,$FE,$69,$F8,$EF,$E4,$7E,$C1,$A4,$AF,$F3,$83,$67
	.dc.b	$84,$60,$FF,$0E,$7C,$17,$EC,$88,$E0,$91,$FE,$2C,$8C,$B6,$3E,$0F
	.dc.b	$CF,$FF,$AA,$DB,$EF,$7C,$28,$C4,$FF,$4F,$D1,$18,$8F,$F0,$AF,$CD
	.dc.b	$7F,$75,$25,$FF,$88,$9B,$1F,$21,$AF,$08,$FF,$1A,$7C,$FF,$F0,$A2
	.dc.b	$38,$24,$7F,$91,$23,$2D,$8F,$F4,$C7,$CF,$9F,$E9,$AD,$72,$60,$86
	.dc.b	$09,$1F,$EC,$D8,$E0,$FF,$B3,$FE,$61,$F0,$1F,$D6,$1B,$5F,$EA,$76
	.dc.b	$CE,$E7,$FD,$2B,$F0,$AF,$9F,$FD,$54,$FF,$77,$8C,$73,$EA,$73,$3F
	.dc.b	$EE,$B6,$BF,$E0,$99,$DE,$1F,$B7,$6C,$73,$9D,$A8,$CB,$FA,$1E,$FF
	.dc.b	$D6,$12,$BF,$50,$D8,$FD,$43,$5E,$11,$FE,$67,$F9,$FF,$E7,$A4,$70
	.dc.b	$48,$FF,$3B,$C6,$5B,$1E,$7F,$9F,$6D,$F1,$1B,$19,$8C,$1F,$E3,$DF
	.dc.b	$9E,$91,$FE,$6A,$6D,$7E,$C1,$AF,$F9,$91,$AF,$F8,$BF,$C4,$FE,$9F
	.dc.b	$C7,$FD,$38,$95,$FD,$39,$FE,$A9,$F3,$DF,$BB,$AF,$E8,$2D,$79,$35
	.dc.b	$52,$30,$DF,$27,$99,$2F,$F9,$64,$65,$B2,$F9,$BF,$09,$23,$2F,$F6
	.dc.b	$7F,$9F,$65,$49,$3D,$23,$2D,$8C,$E3,$FE,$43,$C1,$5A,$8C,$35,$FF
	.dc.b	$55,$FC,$3F,$F7,$82,$5F,$FA,$95,$F8,$BF,$1F,$23,$D2,$D9,$ED,$92
	.dc.b	$31,$1F,$29,$AF,$F3,$DA,$FE,$78,$4A,$FE,$70,$FB,$FE,$4A,$47,$B5
	.dc.b	$B3,$E8,$6B,$FD,$DB,$04,$62,$3F,$D8,$7C,$DF,$E1,$22,$57,$FA,$FD
	.dc.b	$FF,$09,$7C,$FF,$EF,$27,$FB,$FB,$5F,$C4,$35,$FB,$ED,$CC,$D2,$57
	.dc.b	$43,$FA,$C5,$47,$FA,$A3,$6B,$EA,$7C,$FF,$32,$4F,$F8,$95,$AF,$E2
	.dc.b	$1B,$1C,$23,$FC,$CF,$F3,$EF,$96,$A4,$8F,$EF,$5E,$C2,$75,$1F,$D5
	.dc.b	$BE,$66,$DF,$FB,$29,$AF,$FA,$F1,$AF,$F9,$89,$AF,$F9,$81,$BE,$29
	.dc.b	$A4,$C3,$FD,$69,$EB,$CB,$FF,$70,$B6,$FF,$B3,$F2,$7F,$AE,$49,$3F
	.dc.b	$EE,$67,$FC,$8D,$F1,$2F,$4E,$92,$7F,$57,$19,$6C,$66,$BF,$41,$39
	.dc.b	$6C,$7F,$95,$BD,$E4,$D2,$57,$F8,$36,$32,$D8,$CD,$7F,$AB,$7C,$CF
	.dc.b	$F5,$9C,$62,$3F,$C4,$7F,$23,$FE,$BD,$F5,$BA,$C7,$FA,$FB,$C8,$FE
	.dc.b	$A4,$FF,$97,$EB,$FC,$5D,$5B,$7F,$B5,$61,$67,$FD,$75,$35,$FF,$0B
	.dc.b	$45,$7F,$D9,$51,$6B,$AD,$F8,$27,$FD,$9B,$39,$AE,$10,$8C,$3E,$33
	.dc.b	$B9,$4E,$E5,$3B,$96,$BB,$DF,$AA,$31,$5D,$AB,$AD,$F8,$E7,$89,$F1
	.dc.b	$8F,$F7,$5C,$7F,$BB,$1F,$21,$FE,$51,$F9,$7F,$F2,$EF,$A9,$FE,$4E
	.dc.b	$6C,$67,$D5,$4E,$5B,$1C,$7E,$7F,$FB,$03,$7C,$6E,$51,$88,$FF,$73
	.dc.b	$5A,$31,$1F,$EF,$48,$C4,$FF,$BD,$FE,$83,$E6,$43,$D9,$FF,$7C,$C6
	.dc.b	$1B,$AF,$1C,$3C,$33,$C1,$B1,$D6,$77,$80,$5F,$57,$C1,$3C,$F0,$B1
	.dc.b	$B9,$F2,$BF,$5D,$E0,$DA,$E8,$E6,$FE,$1E,$9E,$3B,$E7,$3E,$0A,$F1
	.dc.b	$B4,$CD,$F6,$C7,$84,$1F,$80,$EA,$78,$CE,$F1,$F6,$6F,$ED,$F9,$CF
	.dc.b	$1E,$3F,$A2,$8C,$61,$FE,$1B,$DF,$6A,$B7,$18,$FA,$0B,$2A,$E2,$FF
	.dc.b	$87,$67,$8D,$6A,$73,$7D,$E9,$79,$E2,$7C,$15,$24,$F8,$CF,$EF,$BF
	.dc.b	$41,$FE,$71,$71,$82,$B8,$C6,$1A,$F8,$CF,$33,$8C,$3F,$80,$F8,$0B
	.dc.b	$E9,$5B,$1F,$41,$F0,$5F,$F5,$6B,$5F,$F5,$A3,$5F,$BA,$70,$CF,$FD
	.dc.b	$66,$F2,$CF,$E3,$B4,$F6,$14,$F0,$4F,$5C,$3F,$8A,$F7,$3E,$7D,$E4
	.dc.b	$FF,$AF,$A3,$0D,$8C,$DF,$F4,$E6,$BE,$49,$DB,$99,$E0,$47,$FA,$AF
	.dc.b	$C7,$3C,$7F,$BB,$48,$C4,$FD,$38,$C4,$EA,$FB,$E1,$F8,$BF,$96,$FE
	.dc.b	$C5,$7F,$EC,$56,$BF,$CD,$DA,$FE,$93,$81,$18,$8F,$CB,$B5,$FC,$73
	.dc.b	$63,$7E,$07,$FB,$81,$B3,$FE,$CF,$F9,$3F,$C8,$72,$1A,$7E,$4E,$FC
	.dc.b	$78,$7E,$87,$B7,$FC,$2F,$7F,$ED,$B6,$BF,$D3,$5A,$F9,$DA,$FE,$B1
	.dc.b	$AF,$E8,$DB,$87,$F3,$DF,$23,$CD,$18,$F8,$47,$C5,$FE,$95,$B7,$F9
	.dc.b	$C6,$BF,$99,$BE,$4F,$E8,$31,$7C,$1F,$C1,$6C,$7C,$F8,$C1,$1C,$10
	.dc.b	$FA,$4F,$C9,$D1,$F1,$9F,$20,$93,$DE,$7F,$75,$93,$E2,$BD,$64,$9F
	.dc.b	$90,$D7,$E4,$35,$F9,$0F,$D7,$C9,$F5,$FF,$04,$93,$FB,$CA,$EC,$FF
	.dc.b	$01,$EA,$7C,$FB,$E4,$3F,$5F,$3F,$E8,$B8,$C3,$FE,$7C,$6B,$FE,$6F
	.dc.b	$6B,$FC,$CA,$79,$CF,$28,$FF,$44,$CE,$9E,$91,$99,$DC,$7F,$A2,$3C
	.dc.b	$07,$BD,$FC,$8A,$7F,$A3,$B6,$33,$3E,$9B,$F1,$F1,$FF,$10,$F8,$6F
	.dc.b	$F0,$8C,$FF,$15,$C1,$18,$9D,$70,$75,$BF,$F6,$ED,$FE,$17,$97,$E5
	.dc.b	$11,$F8,$7C,$6D,$FF,$41,$18,$7E,$EB,$F8,$DE,$2F,$C8,$1E,$9C,$9F
	.dc.b	$27,$FC,$5B,$E6,$DE,$77,$F1,$A6,$DF,$C7,$6B,$F4,$9A,$FF,$22,$6B
	.dc.b	$FD,$81,$AF,$F6,$D6,$A7,$3D,$9F,$C3,$F9,$3F,$88,$F6,$FE,$1A,$BF
	.dc.b	$88,$35,$FE,$6D,$5E,$2E,$0F,$EB,$5A,$3E,$5B,$F5,$84,$AF,$C6,$72
	.dc.b	$D7,$F5,$93,$B7,$F6,$13,$F9,$E9,$FE,$B1,$F5,$3F,$D5,$A7,$FA,$7B
	.dc.b	$5F,$DA,$36,$FF,$90,$4F,$0F,$0A,$57,$07,$F8,$84,$7F,$2C,$F9,$D7
	.dc.b	$CF,$52,$57,$F5,$FB,$7C,$FB,$E5,$73,$3F,$D6,$F8,$18,$72,$8F,$E6
	.dc.b	$8D,$7F,$B0,$5F,$2A,$FE,$1D,$F6,$52,$7F,$87,$3D,$27,$FC,$05,$A7
	.dc.b	$F5,$1E,$9F,$F0,$3B,$6B,$FB,$F3,$71,$F0,$FC,$07,$C6,$F2,$65,$7C
	.dc.b	$67,$DD,$1F,$C7,$7E,$CA,$4F,$F1,$D7,$F8,$5F,$C7,$62,$3E,$9C,$5C
	.dc.b	$BF,$9A,$4F,$E4,$FE,$7F,$F2,$44,$AF,$C9,$35,$FC,$93,$5F,$C9,$3F
	.dc.b	$D1,$72,$7D,$27,$DC,$24,$FF,$72,$E6,$57,$F3,$3F,$A3,$93,$FC,$CD
	.dc.b	$FE,$43,$F1,$98,$8F,$C4,$C5,$CB,$FA,$54,$FE,$23,$DB,$95,$F8,$87
	.dc.b	$FB,$5F,$C8,$FF,$AA,$63,$04,$3F,$17,$1A,$3F,$C5,$BE,$E3,$E8,$BF
	.dc.b	$83,$CF,$FA,$65,$AF,$F9,$5E,$35,$C7,$FC,$6D,$B1,$FC,$99,$AF,$F8
	.dc.b	$59,$AF,$38,$E1,$3D,$21,$C7,$FE,$01,$F1,$3F,$66,$48,$FF,$1A,$5F
	.dc.b	$4C,$FF,$13,$F0,$BF,$A8,$24,$FF,$57,$E2,$AF,$EA,$22,$CF,$F8,$1E
	.dc.b	$7F,$B1,$7C,$74,$AF,$EC,$4F,$F5,$0F,$03,$11,$FB,$88,$B5,$FD,$E6
	.dc.b	$7F,$CA,$7E,$FB,$E8,$7F,$D3,$D7,$FE,$7B,$8C,$7B,$7F,$CF,$DB,$1F
	.dc.b	$D2,$9A,$FF,$8E,$BC,$BF,$E6,$7E,$97,$F2,$52,$47,$F9,$4E,$77,$8F
	.dc.b	$F9,$A7,$5F,$F2,$94,$FF,$95,$1A,$FF,$79,$6C,$7F,$41,$AF,$F3,$24
	.dc.b	$63,$87,$FC,$EA,$DA,$FF,$36,$37,$0F,$1C,$7F,$81,$67,$FB,$51,$1F
	.dc.b	$E7,$CF,$F8,$8E,$5E,$F7,$62,$4F,$13,$FC,$1D,$F4,$3F,$DC,$C8,$FF
	.dc.b	$47,$79,$78,$BE,$62,$47,$03,$DC,$E1,$F7,$3F,$E1,$B8,$C6,$5F,$16
	.dc.b	$2E,$1F,$E9,$2F,$1D,$5F,$E2,$1F,$2B,$FC,$08,$4C,$3F,$AC,$3E,$BF
	.dc.b	$07,$E3,$92,$3F,$C4,$D1,$FE,$56,$FA,$17,$D0,$FF,$99,$5B,$3F,$E0
	.dc.b	$6E,$84,$62,$3F,$CC,$8D,$7F,$D6,$B1,$A9,$E4,$48,$E5,$3B,$7F,$92
	.dc.b	$FD,$7F,$F4,$62,$3F,$E6,$58,$ED,$AF,$F9,$EB,$E7,$FF,$C1,$93,$FA
	.dc.b	$C6,$E0,$69,$3C,$BF,$F3,$F3,$67,$D4,$F2,$F4,$5E,$2F,$FA,$63,$0F
	.dc.b	$E7,$3F,$31,$FD,$60,$FF,$3C,$D7,$F8,$43,$87,$F9,$DC,$67,$0F,$EB
	.dc.b	$6E,$B1,$DF,$AF,$E0,$24,$F8,$1A,$FF,$A2,$EF,$FC,$6A,$D8,$CB,$FE
	.dc.b	$7B,$E5,$FE,$5E,$FE,$8F,$C7,$FC,$F9,$D9,$FE,$13,$F5,$DA,$7F,$8A
	.dc.b	$35,$FF,$73,$CA,$31,$1F,$C7,$DA,$FF,$BD,$FE,$8B,$FC,$CF,$A7,$FA
	.dc.b	$A3,$F7,$BF,$F4,$BC,$BD,$AF,$F5,$D7,$0A,$73,$1F,$EB,$BB,$E3,$BF
	.dc.b	$FB,$D9,$7C,$EF,$EE,$B0,$C3,$5F,$03,$54,$70,$F8,$5F,$F2,$62,$FA
	.dc.b	$16,$B8,$6D,$7F,$96,$A7,$53,$93,$FC,$A9,$E9,$7F,$9B,$F1,$FF,$58
	.dc.b	$67,$FE,$68,$57,$ED,$EA,$AF,$F8,$1F,$85,$3C,$3D,$7B,$3B,$3F,$9E
	.dc.b	$FD,$7B,$EB,$7F,$A5,$F0,$A3,$17,$FA,$96,$D7,$FB,$2D,$AF,$DF,$F3
	.dc.b	$7A,$F1,$FF,$23,$6D,$79,$AF,$FD,$11,$ED,$E2,$FF,$8E,$D5,$9C,$D7
	.dc.b	$ED,$F3,$FF,$31,$B6,$FC,$B1,$8B,$FF,$6E,$35,$FD,$CB,$5F,$03,$5F
	.dc.b	$EB,$6D,$EB,$FD,$4F,$77,$8D,$B7,$FE,$DD,$97,$CD,$E7,$B7,$F8,$6C
	.dc.b	$FF,$90,$AF,$07,$A6,$B7,$19,$39,$57,$A0,$8F,$B5,$B8,$7A,$9A,$FA
	.dc.b	$9F,$EF,$B8,$7E,$FB,$CF,$F1,$1F,$1F,$B1,$F0,$F8,$C3,$FD,$E9,$18
	.dc.b	$7F,$83,$C7,$DC,$8C,$61,$FB,$98,$C1,$27,$64,$8F,$C5,$9F,$D4,$BC
	.dc.b	$FF,$8D,$CE,$7E,$47,$8B,$F9,$DA,$FE,$C1,$18,$EA,$F4,$CE,$6F,$9E
	.dc.b	$7F,$E9,$59,$C9,$EE,$C1,$F1,$7D,$8E,$BD,$B7,$2F,$ED,$91,$70,$FF
	.dc.b	$3C,$CD,$E5,$FE,$E5,$36,$BF,$A7,$4D,$A6,$3F,$EA,$58,$F2,$92,$7F
	.dc.b	$D4,$CF,$8F,$E3,$7D,$E7,$BB,$82,$CF,$FB,$D5,$F6,$91,$FF,$57,$3E
	.dc.b	$BF,$8E,$F8,$E7,$E7,$FC,$B7,$D1,$A3,$35,$FE,$07,$6E,$5F,$EC,$39
	.dc.b	$D4,$E4,$FF,$71,$38,$8D,$4F,$EA,$FC,$E7,$D1,$BE,$1F,$C1,$7B,$DF
	.dc.b	$F8,$0D,$6A,$7F,$C9,$2D,$C7,$FC,$45,$7C,$5B,$3F,$57,$EC,$BF,$02
	.dc.b	$7E,$03,$AE,$CF,$F8,$7E,$2E,$7F,$EA,$C9,$B8,$FF,$10,$E0,$FE,$65
	.dc.b	$3F,$AF,$F1,$BF,$8C,$E1,$FC,$6A,$B4,$FF,$30,$30,$7F,$1E,$AE,$2E
	.dc.b	$2B,$ED,$D7,$25,$5D,$E3,$0F,$4E,$1F,$04,$FE,$6B,$EC,$A5,$FF,$96
	.dc.b	$BE,$93,$6F,$F8,$CA,$3F,$8B,$72,$A7,$B5,$7F,$54,$6B,$FE,$71,$6B
	.dc.b	$FC,$F9,$B1,$C6,$3F,$CE,$ED,$AF,$F3,$A4,$E8,$E3,$97,$F5,$79,$F8
	.dc.b	$C7,$F2,$5F,$69,$E6,$E2,$FE,$15,$6D,$BA,$7F,$4E,$7E,$37,$82,$BE
	.dc.b	$31,$AC,$3F,$77,$E7,$C8,$E7,$D6,$FA,$4C,$6B,$FE,$59,$AB,$C1,$FE
	.dc.b	$EE,$9B,$7F,$57,$3F,$E1,$B3,$F4,$7B,$BE,$4F,$D2,$3E,$0B,$F0,$11
	.dc.b	$87,$E1,$C6,$6F,$FD,$2C,$DC,$BF,$D2,$CD,$7F,$C0,$6D,$7F,$D4,$0D
	.dc.b	$7F,$B6,$B7,$43,$9D,$FF,$B8,$F0,$7F,$1A,$3F,$8C,$FC,$65,$C3,$FC
	.dc.b	$F1,$17,$0F,$C5,$C7,$F9,$5E,$71,$9F,$E3,$7C,$DF,$CD,$1F,$D6,$7E
	.dc.b	$93,$FB,$7D,$7D,$28,$CB,$4F,$A7,$1A,$BE,$A7,$E6,$A7,$F9,$6F,$8F
	.dc.b	$2B,$EC,$FC,$E3,$E0,$71,$7F,$F0,$06,$31,$3A,$9F,$ED,$98,$BE,$91
	.dc.b	$F0,$1F,$DB,$6F,$78,$71,$BC,$1C,$2B,$78,$73,$3F,$17,$F3,$3F,$B7
	.dc.b	$3F,$B3,$3F,$D7,$AB,$D3,$BF,$F8,$B6,$75,$3C,$0F,$CA,$9B,$CB,$F6
	.dc.b	$91,$FD,$2B,$CB,$93,$F9,$0F,$A4,$B8,$7F,$A9,$E6,$DF,$F0,$9E,$95
	.dc.b	$62,$FF,$21,$F1,$DF,$C2,$CF,$83,$F4,$8F,$A4,$C4,$FF,$07,$C1,$39
	.dc.b	$8F,$EB,$B9,$7F,$08,$8C,$3F,$D8,$27,$75,$FE,$EE,$6D,$70,$6C,$EE
	.dc.b	$B8,$BF,$E3,$28,$CB,$63,$2F,$FA,$FB,$D4,$FE,$D6,$7F,$36,$E9,$B7
	.dc.b	$FE,$C9,$FB,$0F,$EC,$84,$7C,$FB,$FF,$92,$A3,$FE,$00,$F9,$6F,$A4
	.dc.b	$FE,$33,$8A,$30,$FF,$51,$8F,$DA,$DF,$F9,$22,$72,$FF,$2B,$7F,$BE
	.dc.b	$7C,$5F,$F5,$7F,$48,$F1,$E3,$C3,$6B,$E4,$D7,$86,$7C,$76,$69,$27
	.dc.b	$71,$96,$C6,$70,$FF,$02,$F7,$3E,$A7,$FC,$1E,$7F,$83,$BE,$6E,$DF
	.dc.b	$F1,$67,$FC,$1A,$47,$E2,$F2,$A3,$17,$FE,$10,$8C,$4E,$A7,$33,$FE
	.dc.b	$68,$DB,$FD,$0D,$E6,$7F,$55,$3F,$3F,$36,$BF,$36,$FF,$8D,$1E,$75
	.dc.b	$FE,$1C,$6B,$CD,$AF,$F5,$96,$BF,$E1,$26,$D7,$F9,$DA,$7A,$1F,$E2
	.dc.b	$3E,$0F,$F1,$79,$FE,$2E,$EA,$B5,$FC,$D9,$FF,$16,$91,$F8,$2D,$BF
	.dc.b	$80,$D7,$DD,$8A,$3F,$C9,$BE,$1B,$E7,$BF,$BA,$1F,$BE,$8B,$7F,$BA
	.dc.b	$7E,$01,$1F,$C9,$9C,$CA,$E6,$FE,$23,$5F,$F6,$27,$CA,$7F,$92,$8F
	.dc.b	$F4,$07,$D1,$3E,$73,$17,$FD,$62,$30,$FF,$BE,$5B,$5F,$E5,$4F,$2D
	.dc.b	$18,$8F,$F7,$CB,$5F,$F4,$83,$54,$70,$AF,$F4,$FC,$E6,$31,$5F,$C1
	.dc.b	$29,$F4,$4F,$8D,$FF,$4C,$46,$1B,$0E,$EB,$18,$7F,$D5,$F3,$FE,$6A
	.dc.b	$9D,$57,$56,$C7,$F1,$26,$BF,$93,$F9,$C7,$D6,$FF,$A5,$5B,$5F,$B7
	.dc.b	$C1,$F0,$19,$A9,$3F,$E0,$4A,$DD,$FF,$4E,$EA,$F5,$5E,$A3,$F9,$CF
	.dc.b	$C3,$78,$BC,$1F,$F1,$B3,$70,$FF,$18,$3F,$05,$FF,$07,$B9,$9F,$F3
	.dc.b	$5B,$79,$CD,$38,$47,$6D,$7E,$4B,$FE,$D8,$E4,$FF,$6C,$FD,$17,$CD
	.dc.b	$8F,$F7,$FE,$18,$DB,$F9,$AD,$59,$C8,$AE,$0F,$FB,$82,$34,$FF,$B9
	.dc.b	$3A,$FF,$DB,$F3,$A8,$F0,$1C,$6B,$FD,$3B,$3A,$7B,$C3,$E3,$F8,$4F
	.dc.b	$FB,$A1,$E5,$1F,$E5,$A7,$38,$7F,$BC,$9C,$E1,$E2,$9F,$F1,$91,$DC
	.dc.b	$FF,$55,$B9,$AF,$04,$60,$D1,$E9,$AE,$4F,$F9,$A3,$0E,$EE,$8F,$59
	.dc.b	$F9,$7A,$FC,$C3,$83,$D6,$7F,$AE,$BE,$7B,$FD,$21,$ED,$FE,$75,$B1
	.dc.b	$FD,$89,$AE,$FD,$D4,$EF,$DF,$FF,$36,$CE,$DA,$F5,$6A,$25,$23,$31
	.dc.b	$87,$9E,$7F,$6A,$7F,$D8,$33,$F2,$F4,$E4,$A3,$2F,$D5,$8C,$C7,$5E
	.dc.b	$0E,$13,$D5,$AF,$1F,$80,$E7,$7D,$DB,$7C,$B0,$F1,$B6,$F8,$E1,$E0
	.dc.b	$BE,$18,$62,$F6,$F7,$3C,$87,$C0,$C3,$AB,$4E,$17,$1D,$9E,$53,$BC
	.dc.b	$70,$7E,$04,$F6,$E1,$F9,$9F,$17,$FC,$55,$B3,$FE,$04,$11,$96,$99
	.dc.b	$70,$47,$07,$F7,$AF,$A2,$FE,$E0,$DF,$0F,$F9,$17,$32,$73,$3F,$50
	.dc.b	$FF,$07,$38,$7E,$BC,$6E,$78,$CE,$67,$FC,$20,$D7,$8C,$76,$FA,$DF
	.dc.b	$F6,$C4,$7F,$83,$1A,$F7,$A4,$62,$3B,$B5,$FF,$63,$74,$3E,$9F,$E3
	.dc.b	$36,$3F,$C5,$2D,$7D,$59,$23,$11,$D3,$E1,$FC,$84,$7C,$8D,$7C,$96
	.dc.b	$8C,$47,$13,$BF,$7F,$88,$8E,$0F,$F1,$4F,$DC,$FE,$95,$BB,$7D,$08
	.dc.b	$C4,$7C,$16,$BF,$39,$AF,$F8,$ED,$AF,$F9,$05,$B9,$9C,$F7,$BE,$5D
	.dc.b	$F2,$DF,$D7,$E3,$0D,$8C,$DF,$F9,$9A,$72,$FF,$90,$3C,$07,$AF,$F8
	.dc.b	$66,$C6,$63,$04,$70,$43,$EE,$47,$07,$FC,$E1,$F6,$FF,$CC,$5F,$F3
	.dc.b	$73,$5C,$F3,$FF,$75,$E7,$4E,$67,$F2,$35,$FE,$CF,$7C,$DE,$4F,$94
	.dc.b	$8E,$08,$78,$9C,$BF,$84,$C7,$F9,$5F,$1F,$F8,$22,$3F,$61,$18,$3F
	.dc.b	$97,$EC,$FD,$13,$FC,$C1,$3F,$E6,$46,$BF,$99,$6B,$FA,$66,$BF,$E7
	.dc.b	$36,$BF,$E7,$96,$C7,$5F,$37,$F6,$71,$AB,$FF,$0D,$F8,$5E,$42,$5F
	.dc.b	$FA,$1B,$04,$70,$FA,$B9,$3E,$56,$BE,$56,$BE,$56,$BE,$5F,$3B,$C2
	.dc.b	$47,$DC,$F9,$3D,$59,$1F,$05,$AF,$C1,$FA,$6F,$49,$1F,$8C,$FB,$3E
	.dc.b	$72,$47,$C9,$6B,$FE,$B6,$FB,$07,$CD,$7F,$5E,$6C,$7F,$8D,$DA,$EE
	.dc.b	$34,$47,$4D,$7F,$C9,$5F,$66,$F9,$1F,$D1,$36,$3F,$40,$D7,$71,$F8
	.dc.b	$2F,$86,$91,$F3,$DA,$FE,$25,$AF,$FA,$37,$64,$62,$3F,$B3,$6F,$FE
	.dc.b	$54,$F8,$2F,$C6,$92,$3E,$D3,$5F,$CA,$35,$E8,$FF,$80,$3E,$83,$FB
	.dc.b	$2C,$7F,$35,$6B,$FD,$A9,$AF,$F7,$56,$A6,$93,$C7,$E8,$7F,$05,$FD
	.dc.b	$75,$EF,$52,$57,$F0,$58,$CB,$F9,$1E,$6F,$D3,$91,$C1,$0F,$D7,$39
	.dc.b	$69,$F9,$9F,$47,$F5,$E4,$70,$43,$F6,$CE,$5A,$7F,$9E,$7F,$00,$73
	.dc.b	$7F,$89,$1B,$19,$8C,$11,$C1,$0F,$EB,$C2,$38,$B8,$FA,$3F,$EA,$0F
	.dc.b	$F8,$7D,$A9,$A6,$B5,$FF,$48,$F9,$DF,$C2,$89,$3F,$B7,$6B,$FC,$45
	.dc.b	$AF,$E2,$5F,$E6,$DF,$B0,$FF,$4F,$8F,$E6,$8D,$7F,$D4,$8D,$7F,$C5
	.dc.b	$3C,$28,$C4,$7F,$8D,$63,$1E,$47,$42,$47,$47,$FC,$97,$F2,$DF,$E0
	.dc.b	$A8,$F8,$6D,$7F,$C0,$4D,$7F,$C6,$ED,$7F,$C8,$4D,$4D,$27,$9F,$FE
	.dc.b	$58,$CB,$CB,$3F,$2E,$B0,$47,$FA,$97,$C6,$7E,$A9,$BC,$BF,$D6,$63
	.dc.b	$11,$FE,$C5,$6B,$F0,$9A,$FC,$16,$BF,$EC,$46,$E2,$73,$A7,$FB,$23
	.dc.b	$C2,$FE,$82,$4A,$FE,$5C,$D7,$FA,$43,$5F,$E0,$0F,$F8,$AB,$F4,$7F
	.dc.b	$ED,$71,$FE,$1B,$6B,$FE,$28,$6B,$FE,$69,$F0,$91,$88,$57,$FB,$6B
	.dc.b	$ED,$BF,$A9,$92,$3F,$A0,$3D,$FE,$AE,$47,$E6,$AF,$9C,$62,$3C,$4D
	.dc.b	$7C,$4D,$7C,$56,$FA,$9F,$51,$23,$A4,$65,$FE,$A1,$F0,$BF,$D9,$88
	.dc.b	$E0,$DF,$5F,$C3,$23,$83,$FA,$C7,$37,$F6,$E2,$38,$21,$FD,$D5,$CB
	.dc.b	$4C,$9F,$CB,$DC,$B4,$FF,$7E,$7E,$64,$F6,$7F,$D4,$ED,$8C,$C6,$0F
	.dc.b	$F1,$27,$DD,$B3,$A4,$9F,$F1,$24,$65,$FE,$0F,$E4,$3F,$09,$76,$7F
	.dc.b	$0A,$9F,$EE,$71,$87,$6D,$7F,$DA,$6D,$7F,$D4,$33,$CE,$79,$36,$38
	.dc.b	$C6,$27,$84,$7F,$4C,$E4,$FF,$2D,$B7,$D7,$FF,$0C,$91,$FE,$3B,$F0
	.dc.b	$2F,$BA,$7A,$FE,$78,$C3,$63,$33,$C1,$AF,$FA,$0E,$76,$E6,$78,$11
	.dc.b	$FF,$25,$7E,$F0,$FA,$4F,$F2,$83,$63,$31,$82,$38,$21,$FE,$91,$C5
	.dc.b	$3C,$63,$C5,$87,$F3,$67,$FD,$93,$87,$B7,$87,$D4,$48,$F5,$35,$E2
	.dc.b	$D7,$F9,$4D,$3E,$AF,$CE,$48,$F5,$BF,$46,$7F,$DA,$FF,$07,$F6,$35
	.dc.b	$FE,$8E,$D8,$CD,$7F,$B9,$77,$FC,$34,$62,$13,$FC,$DF,$EB,$B0,$D8
	.dc.b	$FF,$76,$3D,$BC,$34,$E9,$B3,$A3,$FD,$D3,$F5,$0F,$9F,$7C,$3F,$E4
	.dc.b	$E3,$0D,$8C,$CF,$F6,$09,$CB,$7C,$EF,$F2,$5B,$73,$E5,$18,$8F,$F6
	.dc.b	$24,$FF,$8A,$E7,$FB,$4E,$4F,$8B,$FF,$07,$3E,$5A,$F7,$C6,$1F,$82
	.dc.b	$7B,$9C,$47,$CC,$6C,$7C,$9D,$7F,$D7,$D3,$CE,$BF,$61,$CB,$CA,$3E
	.dc.b	$93,$6B,$83,$57,$1F,$E2,$BE,$53,$C3,$1F,$80,$6B,$FD,$9F,$89,$18
	.dc.b	$8E,$31,$8B,$FD,$5F,$CB,$3D,$6D,$72,$A3,$13,$A9,$CC,$62,$FF,$54
	.dc.b	$D7,$F3,$2D,$42,$BA,$4F,$23,$C9,$8F,$E9,$3D,$EB,$8F,$FA,$DB,$1F
	.dc.b	$D8,$C6,$08,$E0,$98,$7E,$B6,$78,$3E,$4F,$67,$CE,$49,$F7,$B6,$35
	.dc.b	$18,$23,$FE,$AB,$EA,$F9,$1E,$37,$D0,$6E,$3F,$C4,$2B,$FD,$31,$1F
	.dc.b	$37,$7F,$E5,$59,$DE,$0F,$85,$8F,$F2,$A8,$C1,$2F,$F9,$5B,$63,$F9
	.dc.b	$5B,$F1,$3E,$5B,$FD,$47,$3F,$ED,$B6,$CF,$F0,$B6,$BF,$E2,$86,$BF
	.dc.b	$40,$7E,$68,$F8,$FF,$C1,$46,$1B,$19,$AF,$E8,$0D,$9F,$F6,$66,$BF
	.dc.b	$00,$93,$FE,$B7,$6C,$6A,$30,$47,$04,$3F,$D1,$91,$C1,$F3,$79,$0F
	.dc.b	$D2,$5E,$87,$E5,$E3,$0D,$8C,$DF,$F8,$15,$B5,$FE,$E1,$F9,$1F,$AA
	.dc.b	$49,$FF,$66,$3C,$7E,$56,$47,$16,$CF,$C7,$18,$78,$FC,$7F,$01,$2B
	.dc.b	$C0,$D9,$E1,$3B,$9C,$92,$3F,$DE,$9E,$FF,$E2,$09,$1F,$ED,$F7,$D9
	.dc.b	$F4,$32,$3D,$8D,$9F,$F7,$E7,$07,$E4,$09,$3F,$EE,$76,$C6,$A3,$04
	.dc.b	$7F,$C9,$9E,$1F,$EB,$BD,$1F,$F4,$6C,$FF,$95,$23,$04,$70,$49,$E2
	.dc.b	$FF,$24,$FE,$00,$7B,$3F,$C2,$E1,$18,$75,$19,$8C,$4F,$FA,$53,$F0
	.dc.b	$6A,$FC,$F6,$BF,$97,$6B,$FC,$21,$A9,$A4,$9F,$CD,$BC,$3A,$7F,$C0
	.dc.b	$7E,$0F,$F0,$CE,$AF,$E7,$58,$FF,$A9,$27,$2D,$8C,$E1,$FD,$D5,$AF
	.dc.b	$FA,$49,$AF,$FA,$45,$AF,$F0,$3A,$FE,$07,$EB,$4A,$FF,$58,$F6,$7F
	.dc.b	$36,$8C,$57,$E3,$63,$13,$AC,$15,$AA,$FF,$78,$3F,$27,$D6,$95,$F4
	.dc.b	$5B,$3F,$EB,$EF,$1F,$FA,$93,$F9,$96,$BF,$C0,$5A,$FF,$32,$6B,$FD
	.dc.b	$B3,$D1,$FE,$07,$18,$23,$FC,$7B,$EB,$A5,$F9,$DB,$5E,$BC,$51,$88
	.dc.b	$FE,$21,$97,$F5,$68,$C4,$7F,$8F,$7E,$0F,$F9,$49,$2F,$E1,$36,$3A
	.dc.b	$35,$E7,$68,$FE,$8F,$FA,$FB,$E3,$69,$2F,$EF,$46,$5F,$F2,$D7,$ED
	.dc.b	$BF,$63,$03,$49,$BF,$F2,$46,$CF,$08,$C1,$FD,$03,$C8,$91,$F6,$5B
	.dc.b	$5F,$0D,$AF,$A1,$AF,$F9,$77,$E3,$3F,$A4,$92,$7F,$04,$D8,$F9,$8D
	.dc.b	$7E,$0B,$FD,$9B,$F6,$0F,$9F,$3F,$9B,$6C,$FF,$93,$5A,$FF,$BA,$9A
	.dc.b	$FE,$0F,$DF,$FB,$E4,$9F,$F3,$F7,$0F,$FB,$79,$B1,$99,$FC,$73,$5F
	.dc.b	$E6,$FF,$08,$F7,$FF,$9B,$36,$BF,$D7,$0D,$9F,$E4,$AD,$78,$FB,$FC
	.dc.b	$09,$3C,$1E,$7E,$84,$8E,$6D,$9E,$7E,$3F,$80,$93,$E0,$7C,$7E,$2C
	.dc.b	$8F,$1B,$67,$C6,$D7,$D6,$D7,$FC,$2F,$CD,$FE,$D8,$9F,$F0,$0B,$FE
	.dc.b	$5E,$7F,$7A,$D8,$FD,$E3,$5F,$8B,$BF,$C4,$24,$FC,$46,$C7,$C4,$6B
	.dc.b	$C2,$3A,$FD,$9F,$80,$8E,$0F,$F7,$27,$E9,$DF,$E6,$56,$FA,$BF,$D9
	.dc.b	$E3,$15,$FD,$51,$AF,$F6,$E6,$BF,$EF,$76,$BF,$D8,$1B,$E3,$1C,$FC
	.dc.b	$77,$FA,$BF,$E3,$FF,$C9,$4D,$C3,$FB,$45,$A3,$0F,$F6,$B6,$BE,$1F
	.dc.b	$9D,$F3,$12,$7D,$6D,$8F,$43,$5E,$11,$F4,$3E,$33,$E9,$91,$C1,$F1
	.dc.b	$3E,$C7,$E1,$11,$C1,$27,$E2,$46,$5B,$1F,$CE,$0F,$E7,$BB,$7F,$9F
	.dc.b	$9C,$74,$FF,$90,$9F,$E8,$3A,$3E,$FB,$F0,$17,$FE,$5E,$8C,$4F,$F8
	.dc.b	$66,$73,$3F,$E0,$C6,$CE,$A7,$E4,$B6,$7F,$C1,$73,$A3,$8D,$FF,$AB
	.dc.b	$BE,$4F,$C4,$4B,$F4,$3D,$FD,$C9,$1D,$29,$18,$BF,$F0,$BF,$C5,$7F
	.dc.b	$B1,$6B,$C2,$D8,$CF,$C0,$7F,$8C,$63,$11,$FE,$36,$6B,$EF,$6B,$FE
	.dc.b	$28,$6E,$47,$39,$7F,$AE,$BC,$AF,$F3,$5F,$6F,$F7,$E8,$C6,$3F,$DA
	.dc.b	$23,$04,$9D,$92,$3F,$1F,$87,$F9,$77,$E6,$FF,$D1,$BC,$7D,$E3,$FC
	.dc.b	$C1,$ED,$7F,$AD,$23,$11,$B0,$E1,$FF,$35,$B5,$4F,$4C,$BF,$C8,$95
	.dc.b	$CD,$FC,$FC,$66,$7B,$DD,$7B,$5F,$5B,$E1,$1F,$E4,$DA,$D3,$E0,$BE
	.dc.b	$6F,$91,$EF,$7A,$8F,$19,$93,$B9,$4E,$F1,$C1,$C1,$FE,$B0,$9E,$F5
	.dc.b	$E2,$8F,$84,$7C,$A8,$F9,$6E,$CA,$FC,$EF,$9A,$7A,$8F,$9B,$C3,$6B
	.dc.b	$F8,$0C,$EA,$72,$71,$F1,$FF,$82,$4E,$9A,$FF,$9B,$E7,$FB,$D7,$46
	.dc.b	$C9,$1E,$D8,$78,$3F,$E9,$BD,$7E,$DC,$62,$7E,$83,$C9,$E1,$7E,$09
	.dc.b	$DD,$71,$D3,$FD,$49,$7C,$6F,$78,$72,$C3,$85,$6A,$F8,$57,$FB,$6A
	.dc.b	$F8,$61,$E2,$BD,$CE,$5A,$A7,$77,$CE,$3B,$61,$1F,$76,$1D,$9E,$0D
	.dc.b	$AF,$66,$1F,$9D,$E0,$FF,$75,$FB,$1F,$A2,$6D,$FF,$94,$0F,$F5,$49
	.dc.b	$63,$4E,$0F,$F2,$91,$4F,$F1,$B9,$4F,$EC,$25,$3F,$D4,$7F,$36,$F9
	.dc.b	$BF,$D8,$F2,$23,$13,$FC,$69,$AF,$FA,$91,$AF,$FA,$F9,$AF,$F2,$76
	.dc.b	$BD,$A7,$AC,$62,$5F,$34,$F0,$FC,$2D,$BF,$E5,$2D,$7F,$D6,$6D,$7F
	.dc.b	$CA,$7F,$36,$F1,$7E,$43,$63,$FC,$EE,$D7,$FC,$EC,$D7,$FD,$AB,$3F
	.dc.b	$E7,$67,$BF,$A5,$E2,$23,$82,$1E,$27,$2F,$4F,$4F,$C2,$47,$C2,$FC
	.dc.b	$CF,$9B,$FF,$2E,$47,$F4,$46,$BF,$EC,$E6,$BF,$EE,$86,$BF,$D4,$5A
	.dc.b	$FF,$B6,$9B,$91,$A6,$BD,$17,$9C,$F0,$7F,$96,$C6,$1B,$19,$BF,$F5
	.dc.b	$53,$5F,$F7,$0F,$98,$F8,$3F,$C2,$B6,$3F,$AE,$B5,$DC,$68,$8E,$9A
	.dc.b	$FF,$82,$7E,$6F,$F9,$BB,$FE,$0E,$E8,$46,$23,$5C,$DF,$D1,$27,$2F
	.dc.b	$8B,$E6,$E4,$F8,$9A,$FA,$1A,$8F,$EF,$58,$23,$31,$FE,$44,$CF,$F9
	.dc.b	$3F,$39,$A4,$8F,$D2,$FC,$DF,$FB,$99,$FE,$FA,$D7,$B1,$FE,$07,$70
	.dc.b	$E3,$74,$62,$35,$E1,$74,$23,$D1,$FE,$05,$4F,$06,$92,$7F,$D9,$CF
	.dc.b	$93,$DB,$91,$E4,$6B,$E4,$18,$78,$A3,$31,$E5,$C1,$18,$8F,$57,$47
	.dc.b	$C0,$23,$F0,$1F,$87,$EF,$C8,$F8,$6D,$7C,$17,$F0,$3D,$EF,$92,$48
	.dc.b	$F9,$2F,$CC,$F6,$24,$7C,$C6,$BF,$E5,$2F,$9B,$79,$DF,$D7,$1B,$1F
	.dc.b	$E0,$86,$BB,$8D,$11,$D3,$5F,$89,$AB,$E6,$FF,$C5,$71,$FE,$60,$8C
	.dc.b	$79,$9F,$A8,$9C,$CF,$06,$BF,$ED,$AF,$77,$EB,$91,$FF,$34,$F1,$7E
	.dc.b	$2E,$31,$1F,$EE,$D6,$BC,$4D,$DB,$E6,$FF,$02,$48,$CB,$5F,$F7,$B3
	.dc.b	$5F,$F7,$5B,$FE,$00,$F7,$BF,$B1,$47,$E0,$5A,$FE,$59,$AF,$E9,$5A
	.dc.b	$9A,$4F,$1F,$A1,$FC,$2B,$E6,$9F,$37,$49,$5F,$26,$32,$FB,$39,$FF
	.dc.b	$36,$47,$04,$3F,$2D,$9F,$EC,$63,$51,$8C,$BF,$50,$A7,$F6,$DF,$27
	.dc.b	$FA,$62,$47,$CA,$6B,$C5,$AF,$ED,$A3,$8B,$FE,$50,$F9,$BF,$D0,$B7
	.dc.b	$9F,$FC,$53,$18,$8F,$F5,$1B,$5F,$D6,$35,$FE,$02,$D7,$F8,$83,$53
	.dc.b	$7E,$6F,$FA,$D4,$7F,$08,$6B,$FC,$5D,$A9,$CD,$7F,$15,$F3,$7F,$82
	.dc.b	$92,$7F,$12,$D7,$E4,$35,$F8,$18,$FF,$98,$DF,$E3,$DF,$25,$FD,$85
	.dc.b	$FD,$1B,$5F,$E3,$70,$8C,$47,$F2,$B8,$C7,$8F,$FD,$32,$73,$3D,$23
	.dc.b	$04,$7F,$CF,$BF,$01,$FE,$11,$4D,$27,$9D,$FE,$58,$6C,$70,$8C,$1F
	.dc.b	$E5,$CC,$BB,$C7,$C5,$AC,$11,$FD,$D7,$C3,$B9,$7F,$B0,$B6,$33,$18
	.dc.b	$23,$82,$1F,$C1,$A3,$83,$F9,$3F,$0B,$F9,$A9,$1F,$DC,$B5,$D1,$82
	.dc.b	$1F,$C9,$3E,$6B,$FB,$73,$FD,$15,$AF,$FB,$3D,$AF,$F3,$86,$ED,$FD
	.dc.b	$57,$E0,$BF,$A2,$92,$7F,$84,$BF,$E3,$A9,$7C,$C5,$24,$FF,$B5,$23
	.dc.b	$2D,$8C,$BE,$56,$BE,$5F,$9B,$E8,$47,$93,$FD,$5B,$E6,$E4,$7F,$1A
	.dc.b	$6B,$EF,$EA,$F2,$11,$F1,$3F,$D7,$B9,$64,$7F,$22,$6B,$FC,$05,$AF
	.dc.b	$E6,$9A,$FF,$79,$6B,$FC,$17,$8F,$F6,$59,$FF,$7B,$9F,$F5,$1F,$CD
	.dc.b	$FF,$A8,$2B,$FD,$9E,$D7,$FB,$30,$70,$FF,$69,$6B,$FA,$1F,$9B,$F9
	.dc.b	$2F,$F4,$A6,$BF,$A8,$6B,$BC,$FF,$DB,$BE,$43,$E5,$1D,$58,$6D,$78
	.dc.b	$A7,$53,$92,$46,$67,$6F,$FB,$69,$AF,$FB,$99,$AF,$F8,$77,$D5,$FF
	.dc.b	$05,$91,$FE,$9E,$FF,$BC,$3E,$59,$F3,$4F,$57,$0D,$AF,$8B,$3A,$9C
	.dc.b	$92,$33,$3B,$7F,$C2,$13,$B7,$33,$FE,$52,$F9,$77,$CD,$D2,$4F,$F7
	.dc.b	$18,$CB,$63,$2F,$F7,$88,$CB,$63,$CF,$BC,$8F,$1B,$5F,$1E,$1F,$CD
	.dc.b	$70,$F7,$CE,$67,$FC,$73,$18,$7C,$5F,$07,$E7,$23,$83,$E9,$4F,$F9
	.dc.b	$03,$E0,$BE,$83,$63,$FC,$A1,$B7,$DF,$8C,$47,$F8,$EF,$5F,$E9,$95
	.dc.b	$C7,$C0,$7C,$DE,$A3,$F9,$83,$63,$35,$FE,$5B,$9E,$13,$D2,$1F,$36
	.dc.b	$F8,$B7,$C3,$7E,$72,$73,$18,$7F,$3F,$3B,$9D,$36,$33,$E6,$FF,$94
	.dc.b	$5B,$A6,$A3,$11,$FE,$4D,$94,$62,$3C,$DF,$11,$FE,$0B,$CB,$D1,$18
	.dc.b	$EB,$46,$22,$CE,$DC,$9B,$8D,$61,$FA,$9E,$5C,$C7,$E7,$DB,$3F,$43
	.dc.b	$6F,$F6,$24,$F0,$6C,$74,$3F,$CF,$DC,$8F,$9B,$C4,$7F,$5D,$8C,$57
	.dc.b	$E4,$E7,$33,$C1,$BC,$6F,$8E,$7A,$98,$6D,$FE,$A1,$B7,$FA,$46,$BF
	.dc.b	$90,$6B,$FE,$97,$AE,$95,$CC,$95,$FA,$5F,$77,$C7,$DE,$CB,$FD,$6D
	.dc.b	$39,$26,$1F,$A7,$AD,$CE,$A3,$13,$A7,$C5,$EC,$7A,$49,$1E,$96,$C7
	.dc.b	$A5,$AF,$08,$FF,$2C,$F5,$FC,$6F,$03,$FD,$54,$DC,$7F,$86,$4E,$8F
	.dc.b	$E3,$59,$7C,$CE,$0F,$E4,$8D,$DF,$E1,$E3,$FA,$78,$C1,$2B,$89,$23
	.dc.b	$FC,$F7,$F3,$7F,$EA,$78,$FF,$3D,$36,$33,$7F,$22,$31,$1F,$EA,$5F
	.dc.b	$9A,$7C,$DF,$DE,$84,$62,$FF,$54,$D8,$FE,$8E,$D7,$FC,$5B,$AF,$BC
	.dc.b	$93,$FE,$7C,$6C,$7F,$96,$DA,$F0,$8D,$11,$D3,$5F,$F6,$57,$CD,$BC
	.dc.b	$EF,$C9,$FB,$68,$C5,$7F,$5A,$8C,$4E,$A7,$35,$FE,$6B,$F9,$3F,$A8
	.dc.b	$4B,$FF,$48,$BB,$F9,$B9,$1C,$E9,$18,$AF,$F5,$EC,$3E,$3B,$B1,$2B
	.dc.b	$83,$67,$6D,$7D,$6F,$E1,$7D,$F9,$3F,$EA,$26,$CF,$6F,$9B,$F5,$12
	.dc.b	$7C,$8F,$E3,$78,$64,$7F,$AA,$9B,$3F,$E8,$96,$BF,$E6,$26,$BF,$E5
	.dc.b	$BF,$47,$FD,$C3,$3F,$E4,$8F,$17,$F5,$B1,$89,$FF,$21,$46,$09,$3B
	.dc.b	$24,$7F,$BB,$D7,$CD,$BD,$9F,$E1,$31,$86,$C6,$63,$FC,$22,$64,$FE
	.dc.b	$71,$F3,$79,$AF,$F0,$AB,$5F,$F7,$E3,$51,$1C,$36,$BF,$39,$18,$83
	.dc.b	$4D,$F4,$7F,$80,$7C,$1F,$E1,$BD,$46,$93,$0F,$F7,$34,$65,$B1,$9B
	.dc.b	$FF,$76,$CE,$5F,$F5,$D7,$AD,$FC,$0C,$95,$FE,$E0,$6C,$6A,$30,$47
	.dc.b	$E6,$65,$FE,$7B,$8F,$8E,$D8,$CF,$AD,$F3,$0F,$DD,$78,$FF,$D2,$E3
	.dc.b	$E2,$9F,$19,$C3,$86,$B9,$F1,$78,$92,$FF,$B5,$4E,$A7,$FB,$6C,$FF
	.dc.b	$1D,$F9,$B9,$5F,$EF,$A6,$CF,$FB,$65,$AF,$FA,$C9,$AF,$D8,$F0,$1A
	.dc.b	$BE,$0F,$F9,$41,$2F,$FC,$AB,$3A,$8C,$57,$FB,$95,$AF,$FA,$F9,$FE
	.dc.b	$8B,$F3,$6F,$92,$A4,$BF,$F2,$FE,$09,$D4,$7F,$81,$FE,$6F,$EC,$61
	.dc.b	$F2,$03,$87,$E5,$B5,$FC,$8B,$5D,$F0,$FF,$2B,$8C,$11,$FE,$89,$E4
	.dc.b	$4A,$F6,$E1,$FD,$AD,$B1,$9B,$F1,$67,$F1,$A3,$11,$FE,$DC,$7F,$A7
	.dc.b	$7C,$64,$AE,$6D,$9E,$36,$8C,$47,$FB,$91,$AF,$FB,$77,$E6,$DB,$7F
	.dc.b	$6E,$6D,$FC,$86,$CF,$C5,$B4,$62,$3E,$47,$07,$FB,$7A,$BF,$13,$EF
	.dc.b	$FE,$50,$8E,$09,$3F,$E9,$29,$E0,$D9,$FE,$7F,$F0,$8F,$7E,$92,$FF
	.dc.b	$A0,$46,$5B,$EF,$F0,$23,$83,$7E,$87,$42,$38,$39,$F8,$FE,$02,$38
	.dc.b	$3B,$F8,$BE,$42,$38,$26,$5F,$89,$F1,$D8,$AB,$73,$98,$C5,$7E,$7B
	.dc.b	$8B,$FC,$8A,$D3,$FC,$18,$E2,$7F,$06,$D7,$F7,$B7,$DB,$8B,$E3,$92
	.dc.b	$BE,$2B,$67,$C0,$D7,$8D,$FE,$45,$EB,$F3,$7E,$02,$57,$95,$FE,$E3
	.dc.b	$F3,$7F,$E6,$F8,$FE,$E0,$D9,$F8,$0D,$7C,$AD,$7F,$9B,$E0,$8C,$42
	.dc.b	$31,$EA,$F6,$25,$7F,$B5,$BE,$3B,$FD,$6F,$F1,$DF,$E5,$46,$C6,$7D
	.dc.b	$B4,$E5,$B1,$E6,$F9,$BF,$49,$1C,$13,$1F,$4C,$65,$B1,$F5,$7E,$33
	.dc.b	$E9,$91,$C1,$F0,$BE,$6F,$E2,$91,$C1,$27,$E2,$C6,$5B,$1F,$C7,$8F
	.dc.b	$E1,$9C,$1F,$E8,$47,$1D,$9F,$E4,$D7,$F8,$86,$AF,$9B,$FC,$15,$FC
	.dc.b	$C6,$BF,$2D,$B1,$B9,$FF,$14,$C6,$08,$E0,$ED,$F2,$71,$F2,$36,$BF
	.dc.b	$39,$3F,$EA,$09,$FF,$18,$36,$33,$EE,$76,$3F,$C6,$FF,$14,$F8,$1A
	.dc.b	$4B,$FF,$1C,$46,$5B,$19,$AF,$D5,$4E,$5B,$1F,$E6,$FE,$DF,$EF,$FA
	.dc.b	$7F,$B0,$3C,$CC,$13,$0F,$E9,$33,$A2,$46,$6F,$FD,$8C,$3E,$6F,$FC
	.dc.b	$17,$C6,$CB,$FA,$17,$B5,$FE,$B4,$8C,$57,$F4,$5A,$DD,$6B,$1B,$39
	.dc.b	$3F,$27,$19,$69,$9A,$E6,$FF,$BA,$27,$8B,$FD,$E2,$AB,$FE,$7B,$9D
	.dc.b	$BF,$E2,$08,$CD,$F9,$6F,$FA,$C4,$FA,$CE,$E7,$52,$9C,$4B,$1C,$5F
	.dc.b	$24,$79,$EB,$BC,$73,$7C,$33,$B8,$F3,$1F,$2A,$BE,$5B,$E1,$3F,$DE
	.dc.b	$4F,$FB,$CF,$E6,$BF,$CD,$1F,$37,$7D,$63,$49,$5E,$A9,$FE,$F5,$D1
	.dc.b	$C7,$CA,$34,$93,$B8,$CA,$F9,$CF,$9B,$5F,$CC,$A3,$FC,$8F,$18,$EE
	.dc.b	$FF,$12,$BA,$AE,$2D,$BD,$EB,$FE,$2A,$C3,$96,$1D,$63,$18,$72,$75
	.dc.b	$87,$26,$DF,$DC,$BE,$3A,$F8,$70,$C5,$FA,$23,$C1,$1D,$4F,$F2,$EC
	.dc.b	$F8,$31,$E1,$19,$11,$F1,$9C,$36,$36,$4C,$3C,$73,$D7,$87,$FC,$D3
	.dc.b	$F3,$BF,$E4,$6C,$9C,$1F,$E3,$5A,$4E,$63,$0D,$85,$7F,$9C,$BD,$BC
	.dc.b	$3D,$23,$FC,$EB,$7D,$1D,$9D,$0C,$DF,$46,$9A,$AE,$7C,$6F,$66,$92
	.dc.b	$77,$D2,$F9,$C7,$CE,$E6,$7F,$0F,$8F,$F9,$3F,$0E,$51,$DB,$34,$62
	.dc.b	$36,$7F,$53,$BE,$6E,$FE,$75,$E4,$7E,$BA,$3E,$0C,$61,$FF,$2B,$B5
	.dc.b	$FF,$1C,$B5,$FF,$39,$DF,$CE,$EF,$F8,$F8,$9A,$7C,$AA,$E5,$B7,$C9
	.dc.b	$A4,$3D,$7F,$14,$61,$B1,$9C,$3C,$71,$82,$38,$21,$FE,$4E,$C1,$F6
	.dc.b	$FE,$42,$47,$C1,$9F,$C0,$79,$FF,$E4,$BF,$07,$E1,$36,$7F,$CD,$BE
	.dc.b	$8A,$72,$D8,$CC,$60,$8E,$0E,$3E,$1F,$69,$FF,$59,$47,$E1,$2B,$FC
	.dc.b	$B5,$39,$AF,$C9,$68,$F9,$CF,$F0,$0D,$FF,$81,$DA,$FF,$9E,$E3,$5D
	.dc.b	$5F,$65,$B1,$B8,$C4,$F0,$8F,$D2,$35,$FC,$DB,$5F,$F4,$CF,$DE,$3D
	.dc.b	$EF,$8A,$D8,$F4,$35,$FC,$6D,$F8,$1F,$C7,$F0,$7F,$A8,$9F,$B7,$9F
	.dc.b	$F5,$48,$C4,$7F,$A1,$8C,$90,$C1,$0F,$F2,$2F,$8B,$FB,$28,$34,$FD
	.dc.b	$AC,$FC,$3E,$0F,$F5,$92,$F0,$FC,$24,$BF,$F4,$AB,$5F,$F4,$5B,$5F
	.dc.b	$CD,$C7,$17,$FC,$B5,$E5,$78,$48,$FE,$BE,$34,$FE,$BB,$A5,$F3,$8F
	.dc.b	$43,$F5,$91,$86,$C6,$67,$FA,$EC,$E5,$B1,$97,$FC,$E0,$D7,$FC,$CA
	.dc.b	$D7,$FC,$A2,$D7,$F8,$CE,$6F,$4F,$E4,$12,$3F,$86,$CF,$5D,$7F,$8D
	.dc.b	$5F,$A3,$7F,$E4,$4D,$AD,$C6,$2B,$F8,$65,$7F,$21,$1F,$82,$3D,$1F
	.dc.b	$C9,$36,$FF,$CB,$CD,$7F,$30,$D7,$F8,$4E,$4F,$AF,$FB,$C4,$8F,$3F
	.dc.b	$B5,$FB,$6F,$9D,$FC,$0C,$7F,$B8,$1A,$FF,$AE,$5A,$89,$BA,$33,$18
	.dc.b	$6B,$FE,$F8,$6C,$6F,$8F,$F8,$1C,$6B,$67,$EB,$7F,$A1,$D7,$E4,$B7
	.dc.b	$93,$FD,$06,$DC,$7A,$61,$F9,$9F,$AD,$91,$FC,$DF,$54,$7E,$6B,$DF
	.dc.b	$CC,$FE,$AA,$72,$EE,$74,$F1,$FA,$CF,$D2,$11,$FC,$A7,$D0,$57,$94
	.dc.b	$5C,$FF,$98,$6C,$AF,$E4,$3F,$3E,$FA,$0A,$4C,$3F,$AC,$8F,$83,$FF
	.dc.b	$03,$E1,$FD,$87,$05,$F1,$AF,$E3,$8D,$7F,$A8,$C6,$AF,$FB,$74,$73
	.dc.b	$F4,$7F,$79,$F5,$52,$3E,$23,$E1,$7E,$17,$B5,$FC,$04,$8F,$EC,$7D
	.dc.b	$7A,$E8,$F9,$91,$FC,$12,$3F,$84,$78,$92,$7E,$83,$FC,$23,$E6,$71
	.dc.b	$1F,$3E,$2E,$9F,$1A,$7C,$FF,$41,$FA,$DA,$34,$9E,$07,$F9,$C1,$B1
	.dc.b	$FD,$AD,$AF,$FB,$7B,$C6,$FE,$34,$47,$E6,$C7,$EC,$FA,$3F,$31,$CB
	.dc.b	$FE,$12,$F2,$FF,$90,$12,$BF,$88,$7A,$75,$FC,$34,$5C,$FE,$EC,$FF
	.dc.b	$27,$F2,$E5,$7E,$3D,$FE,$BF,$F3,$2F,$89,$CC,$FE,$3E,$6D,$FD,$DA
	.dc.b	$D4,$F9,$AB,$DC,$47,$DC,$D7,$FC,$EF,$9F,$06,$C6,$5F,$DF,$FD,$EF
	.dc.b	$F3,$B2,$3D,$4F,$D2,$7C,$E7,$F3,$F2,$3F,$D9,$FE,$39,$DD,$88,$FD
	.dc.b	$34,$DC,$BF,$29,$5A,$9F,$D7,$7D,$07,$F4,$92,$3F,$CE,$7E,$22,$BF
	.dc.b	$AF,$8B,$5F,$A6,$9F,$E9,$BF,$11,$FD,$68,$95,$FE,$94,$E1,$76,$27
	.dc.b	$F9,$FF,$D0,$B3,$FE,$56,$DB,$FF,$1E,$B5,$FE,$CF,$5C,$5F,$E0,$9F
	.dc.b	$3B,$FD,$7C,$8F,$0A,$AF,$F0,$58,$B9,$7F,$3A,$BE,$8F,$EF,$BE,$87
	.dc.b	$FB,$89,$2B,$D1,$82,$3F,$C5,$F9,$7E,$F6,$7F,$8A,$78,$3F,$DB,$CF
	.dc.b	$ED,$B3,$FD,$BB,$E4,$BF,$BD,$92,$79,$BC,$BE,$B3,$FB,$B1,$23,$FB
	.dc.b	$34,$57,$F8,$E4,$5C,$7F,$89,$CD,$E9,$FD,$0E,$2A,$C5,$FF,$9D,$32
	.dc.b	$5F,$F8,$E3,$E8,$7E,$54,$FF,$3F,$3E,$9B,$87,$FA,$0B,$5C,$ED,$FE
	.dc.b	$4E,$FC,$F3,$FC,$CF,$1F,$E6,$96,$BF,$E1,$A6,$E7,$FC,$7E,$BE,$6E
	.dc.b	$7F,$E0,$EF,$73,$FC,$3A,$4B,$F9,$AD,$7C,$87,$A5,$C3,$FE,$0E,$7C
	.dc.b	$5E,$CF,$FA,$3B,$0F,$F1,$EE,$1F,$E6,$56,$DF,$7B,$FF,$24,$7C,$EB
	.dc.b	$E8,$BF,$CA,$ED,$8F,$F7,$3E,$A8,$C4,$78,$A3,$13,$C0,$91,$C2,$76
	.dc.b	$FC,$B9,$DB,$F9,$FB,$7C,$5F,$9C,$91,$E4,$BA,$FC,$98,$B8,$FF,$47
	.dc.b	$C3,$C6,$D8,$F4,$92,$7C,$4D,$7D,$13,$E0,$72,$43,$E6,$F0,$30,$EB
	.dc.b	$97,$F2,$A6,$CF,$FA,$F1,$B9,$7C,$5C,$3F,$D8,$1E,$D3,$E7,$BF,$1B
	.dc.b	$7F,$E0,$9F,$8C,$F9,$99,$75,$F8,$6F,$C7,$46,$32,$46,$36,$FF,$08
	.dc.b	$63,$CF,$8F,$18,$FF,$B4,$BC,$17,$4E,$33,$F4,$7C,$3D,$CF,$FD,$95
	.dc.b	$6D,$3E,$E7,$03,$1E,$33,$E5,$D3,$F4,$1F,$44,$F7,$FF,$33,$38,$6B
	.dc.b	$FE,$23,$8D,$7C,$37,$FA,$FD,$B1,$F9,$60,$FF,$9B,$B7,$D7,$AD,$FE
	.dc.b	$5E,$9E,$ED,$7A,$37,$27,$9B,$FE,$07,$7C,$D3,$F4,$7E,$73,$FD,$D1
	.dc.b	$AF,$B2,$B5,$3E,$FD,$3E,$C5,$69,$F9,$7C,$9F,$ED,$6C,$39,$F0,$FD
	.dc.b	$FF,$4B,$58,$7F,$9F,$EB,$4F,$33,$F0,$11,$A9,$FC,$37,$C6,$7F,$96
	.dc.b	$6F,$F0,$CD,$AF,$E7,$7D,$2F,$A2,$C5,$F0,$BF,$D8,$C7,$FC,$01,$AF
	.dc.b	$D3,$56,$E7,$F5,$53,$99,$FF,$28,$CE,$5F,$CD,$C6,$5E,$ED,$EA,$FD
	.dc.b	$C3,$67,$84,$FF,$BB,$B9,$5F,$1F,$E4,$6E,$1F,$C1,$3E,$87,$13,$23
	.dc.b	$F8,$E3,$ED,$C7,$F8,$24,$E6,$FF,$C6,$B8,$23,$C0,$4B,$FE,$08,$D9
	.dc.b	$FD,$9B,$5F,$82,$FA,$F2,$FD,$97,$D9,$F9,$4F,$93,$87,$FB,$F6,$FF
	.dc.b	$97,$46,$3E,$2B,$84,$E7,$0F,$DD,$B5,$FF,$59,$C6,$89,$5F,$D6,$9B
	.dc.b	$1F,$C0,$B3,$7C,$F7,$C5,$3E,$07,$D1,$7F,$B0,$EF,$F9,$F4,$63,$9F
	.dc.b	$E8,$4E,$6F,$F9,$33,$5F,$E4,$4D,$7F,$86,$7D,$8F,$D1,$3D,$BE,$BD
	.dc.b	$C3,$FA,$54,$59,$FE,$FB,$37,$C3,$FE,$79,$36,$BF,$DA,$F3,$6B,$FA
	.dc.b	$F3,$F1,$09,$1F,$DE,$1F,$9D,$93,$EC,$7F,$04,$7C,$EF,$9B,$B5,$FE
	.dc.b	$D5,$7E,$69,$1F,$F0,$0B,$CB,$87,$F1,$47,$E8,$7E,$11,$F4,$5F,$8C
	.dc.b	$CB,$FB,$14,$EA,$FF,$B3,$3B,$9D,$69,$FD,$5E,$31,$CB,$2B,$FD,$CD
	.dc.b	$F2,$97,$0F,$F0,$FB,$F8,$02,$47,$F8,$51,$F6,$F3,$FE,$74,$FC,$4F
	.dc.b	$C6,$D9,$FE,$ED,$16,$BF,$C5,$33,$71,$FE,$D1,$F0,$5F,$A5,$3F,$41
	.dc.b	$EB,$7F,$82,$30,$FE,$FD,$19,$69,$FE,$4D,$F2,$71,$87,$F8,$07,$EB
	.dc.b	$FF,$5C,$7F,$05,$97,$D1,$3E,$0B,$31,$FE,$CF,$C8,$FF,$B4,$4D,$C7
	.dc.b	$EF,$DA,$7F,$2F,$2D,$FF,$81,$9A,$FF,$86,$DA,$F3,$AF,$04,$71,$6C
	.dc.b	$71,$6B,$FC,$1F,$E5,$BF,$83,$1F,$C0,$7A,$FF,$9A,$D7,$F8,$7E,$32
	.dc.b	$D3,$D4,$7F,$BA,$23,$73,$FE,$23,$9F,$E2,$1F,$55,$2B,$73,$FE,$35
	.dc.b	$F7,$B0,$FF,$8A,$66,$E5,$FE,$D0,$9B,$5F,$DD,$7E,$9B,$F8,$D1,$F3
	.dc.b	$72,$FF,$3E,$C7,$F8,$CE,$31,$EF,$7C,$99,$CE,$3E,$F8,$C1,$2B,$E2
	.dc.b	$36,$35,$1F,$E7,$7F,$53,$EA,$43,$1F,$AE,$BC,$34,$EA,$7F,$DC,$FF
	.dc.b	$45,$FD,$0A,$31,$7F,$EB,$66,$DF,$16,$DF,$FA,$12,$B8,$4F,$1F,$06
	.dc.b	$5F,$C8,$AF,$F4,$4F,$C7,$E2,$BC,$93,$D6,$B1,$8F,$FB,$E2,$AC,$F8
	.dc.b	$72,$F2,$1F,$CF,$7E,$91,$E4,$62,$7E,$AF,$8C,$9C,$C6,$1B,$97,$80
	.dc.b	$0F,$A4,$FC,$0E,$5F,$EF,$F9,$DB,$70,$F0,$1F,$DA,$6B,$87,$9D,$8C
	.dc.b	$3D,$DB,$BE,$37,$E5,$1F,$D1,$3E,$52,$CE,$A2,$DF,$FA,$92,$6E,$3F
	.dc.b	$5B,$83,$EA,$F4,$7D,$33,$FA,$6F,$CC,$DC,$7B,$C5,$C3,$FD,$5B,$36
	.dc.b	$FE,$FF,$88,$FD,$71,$E6,$FF,$21,$6E,$3F,$23,$E9,$5C,$F9,$AF,$ED
	.dc.b	$51,$89,$FF,$37,$B7,$2F,$F3,$73,$5F,$F3,$53,$5F,$9B,$C8,$8C,$47
	.dc.b	$FB,$29,$B1,$9B,$FF,$02,$46,$B5,$FC,$AF,$C7,$7E,$4C,$FE,$71,$E8
	.dc.b	$BE,$1F,$19,$7E,$D3,$67,$D8,$FF,$04,$D3,$F8,$B3,$5F,$D1,$36,$37
	.dc.b	$C3,$FE,$15,$E0,$63,$FE,$DD,$C3,$F7,$F1,$8F,$88,$F0,$FC,$17,$89
	.dc.b	$FD,$37,$C6,$CA,$DD,$BD,$2C,$69,$F8,$FC,$7A,$C7,$79,$D5,$7F,$94
	.dc.b	$E7,$2D,$8C,$D7,$FA,$B2,$72,$D8,$CB,$FE,$C3,$8C,$B6,$32,$FE,$E7
	.dc.b	$5F,$DE,$1F,$E3,$ED,$FF,$9C,$DF,$E6,$23,$0F,$F8,$76,$33,$3F,$E9
	.dc.b	$06,$BF,$ED,$17,$F8,$54,$6C,$87,$F9,$B5,$CB,$5E,$31,$FC,$2F,$2F
	.dc.b	$E1,$C7,$8F,$E9,$AD,$7F,$07,$A4,$78,$C9,$1C,$DB,$5E,$56,$BE,$59
	.dc.b	$E3,$3F,$EC,$6F,$C3,$9F,$0F,$FE,$45,$6C,$7E,$27,$E9,$BF,$91,$BF
	.dc.b	$C2,$A3,$37,$FC,$1F,$85,$5A,$8F,$F6,$FC,$63,$B3,$FC,$BC,$DB,$FF
	.dc.b	$47,$36,$77,$5C,$5F,$F7,$D4,$65,$B1,$97,$FB,$73,$5F,$E6,$1F,$15
	.dc.b	$FC,$D4,$FB,$5F,$4D,$6F,$E4,$3F,$64,$8F,$D1,$6C,$FD,$B6,$BF,$1D
	.dc.b	$FF,$2E,$7C,$03,$D7,$FF,$41,$B6,$BF,$9D,$1F,$C9,$8F,$E5,$3A,$FF
	.dc.b	$AE,$67,$FD,$13,$1F,$DA,$AF,$F0,$CD,$76,$D7,$F1,$3E,$0F,$F9,$AE
	.dc.b	$31,$5F,$E0,$C8,$C1,$23,$F1,$96,$F5,$3C,$A7,$83,$E4,$EC,$F9,$9E
	.dc.b	$84,$7F,$28,$1F,$F3,$D7,$D0,$BA,$3F,$D4,$2D,$9F,$E0,$30,$F8,$BC
	.dc.b	$B2,$3F,$AE,$4E,$F3,$FD,$F4,$EE,$72,$FF,$AB,$3E,$1B,$FA,$24,$7F
	.dc.b	$63,$6C,$FF,$53,$D1,$1F,$E1,$EE,$1F,$E8,$4E,$9B,$E3,$73,$3A,$76
	.dc.b	$DC,$3F,$C2,$B8,$7F,$BE,$07,$D1,$FB,$5F,$EB,$E7,$FA,$05,$C3,$B6
	.dc.b	$B9,$CB,$FD,$CD,$BF,$F6,$69,$CF,$9B,$FE,$0A,$77,$5F,$EF,$93,$BD
	.dc.b	$76,$79,$16,$BF,$B7,$F3,$72,$C2,$D7,$6A,$FF,$09,$BC,$70,$FF,$0B
	.dc.b	$D7,$6B,$F1,$4E,$6E,$DF,$53,$C2,$7B,$AF,$CD,$3D,$70,$F0,$4E,$8F
	.dc.b	$81,$1C,$DC,$1F,$E0,$4B,$F5,$3D,$4A,$7B,$CA,$E3,$E9,$3F,$DA,$FF
	.dc.b	$4B,$7D,$C3,$49,$5D,$27,$FC,$33,$B7,$1F,$33,$FC,$6D,$39,$6B,$F2
	.dc.b	$E7,$6E,$67,$E4,$BF,$E1,$A8,$FE,$B2,$E0,$38,$7E,$69,$B9,$EF,$AF
	.dc.b	$E0,$A7,$AB,$7C,$0F,$C3,$B6,$BA,$4E,$B3,$FC,$5B,$63,$C5,$5C,$27
	.dc.b	$2D,$8F,$06,$1C,$B0,$F0,$D6,$A7,$AD,$6D,$F4,$CE,$9E,$27,$A8,$F6
	.dc.b	$C6,$EF,$DD,$87,$AE,$B8,$71,$7F,$8F,$BE,$1F,$EC,$B7,$0F,$EA,$C2
	.dc.b	$32,$D3,$2E,$08,$E0,$FE,$F1,$F8,$A3,$83,$F0,$BE,$1A,$31,$3A,$9C
	.dc.b	$C6,$23,$FB,$E3,$5F,$EF,$4D,$7F,$CA,$B3,$CE,$78,$9D,$BE,$C1,$EB
	.dc.b	$52,$57,$F8,$2E,$32,$FF,$73,$E8,$7B,$3F,$09,$B1,$98,$C1,$EA,$FC
	.dc.b	$0B,$C5,$0C,$51,$86,$E3,$B2,$4F,$F8,$6E,$74,$48,$FA,$3C,$AF,$7E
	.dc.b	$92,$BF,$9E,$BF,$2B,$C8,$7C,$0D,$24,$FF,$89,$E3,$2D,$8C,$BF,$E2
	.dc.b	$78,$CB,$63,$F7,$3C,$CF,$A7,$FF,$50,$CF,$F7,$66,$BF,$DF,$1A,$FF
	.dc.b	$76,$6B,$FE,$E1,$FC,$2D,$F2,$5D,$1B,$1F,$E0,$06,$BF,$E0,$66,$BF
	.dc.b	$E0,$06,$BC,$E3,$91,$1E,$5E,$57,$52,$38,$3F,$C9,$1F,$4D,$FE,$F5
	.dc.b	$6F,$81,$F9,$28,$C4,$7F,$92,$5A,$FF,$91,$DA,$FE,$39,$AF,$DF,$6F
	.dc.b	$01,$CF,$3B,$D4,$FC,$2B,$78,$7F,$5B,$18,$3F,$0D,$E7,$3E,$3B,$F4
	.dc.b	$AD,$C7,$FC,$C0,$D7,$FC,$BC,$D7,$FC,$B8,$D7,$F4,$6D,$7F,$28,$DC
	.dc.b	$4E,$70,$FF,$6E,$FC,$1F,$F0,$0E,$D3,$49,$7F,$2B,$93,$E4,$11,$C1
	.dc.b	$0F,$94,$E5,$A7,$F8,$FB,$E9,$9E,$DF,$E3,$9B,$19,$8C,$1F,$0B,$BA
	.dc.b	$47,$C2,$6B,$F0,$BC,$1F,$8E,$47,$E3,$BF,$2B,$C0,$91,$F2,$9A,$FF
	.dc.b	$58,$FB,$0F,$7B,$FD,$61,$AE,$72,$FE,$B6,$A3,$2F,$F7,$96,$BF,$E0
	.dc.b	$86,$BF,$EE,$86,$BC,$63,$9F,$CE,$FF,$A8,$63,$FB,$3B,$5F,$ED,$6D
	.dc.b	$7F,$C0,$CD,$7F,$C9,$7A,$3D,$6F,$B0,$48,$FF,$46,$46,$5B,$19,$7F
	.dc.b	$C2,$1B,$27,$51,$86,$BF,$E9,$9E,$2F,$F0,$CA,$9F,$CE,$78,$BF,$9A
	.dc.b	$24,$7F,$A6,$E7,$73,$A8,$C4,$E9,$FF,$41,$FC,$CB,$E8,$FF,$87,$B6
	.dc.b	$3F,$D0,$6D,$77,$1A,$23,$A6,$BF,$E2,$BB,$7E,$73,$FE,$53,$8F,$F0
	.dc.b	$AC,$63,$CF,$FF,$0A,$B6,$33,$3B,$79,$7C,$C7,$F4,$46,$F9,$1C,$23
	.dc.b	$11,$FE,$46,$A4,$62,$3F,$DF,$6D,$7F,$81,$36,$37,$3C,$CF,$D8,$78
	.dc.b	$D2,$BE,$CB,$5F,$A8,$D7,$FD,$99,$AF,$F5,$47,$27,$F8,$C3,$EA,$5F
	.dc.b	$37,$F9,$86,$C6,$63,$04,$70,$43,$FC,$77,$1C,$1F,$C9,$DF,$E6,$08
	.dc.b	$FE,$61,$FE,$11,$EE,$48,$FC,$8B,$5F,$C6,$35,$F4,$B5,$1F,$C9,$FE
	.dc.b	$39,$F0,$FF,$59,$B1,$F6,$9A,$FD,$F6,$BF,$4D,$AF,$4C,$D1,$88,$ED
	.dc.b	$3D,$4F,$DE,$F9,$CF,$A4,$A4,$9F,$F2,$DC,$65,$B1,$97,$FC,$C3,$19
	.dc.b	$6C,$7E,$3B,$EE,$5E,$E5,$24,$FE,$52,$32,$FE,$A7,$CD,$FE,$8A,$47
	.dc.b	$04,$3F,$73,$92,$35,$18,$7F,$9D,$7B,$92,$3F,$81,$B5,$FD,$F3,$5F
	.dc.b	$C5,$B5,$FF,$15,$FE,$A6,$F9,$1F,$F1,$BB,$63,$FC,$FF,$E3,$23,$11
	.dc.b	$FE,$9B,$8C,$4F,$F4,$BF,$76,$4F,$F1,$46,$BF,$D6,$A3,$FA,$F9,$1F
	.dc.b	$EB,$ED,$74,$60,$87,$7F,$C5,$3F,$9A,$D1,$A4,$8F,$EB,$9E,$3C,$7F
	.dc.b	$92,$B5,$E5,$C1,$FE,$54,$9F,$C3,$57,$38,$C1,$FD,$97,$8E,$57,$F2
	.dc.b	$E6,$BF,$CA,$5A,$FE,$D9,$AF,$F3,$86,$BF,$CA,$F9,$BF,$83,$77,$7F
	.dc.b	$33,$9F,$E9,$FF,$69,$FE,$6D,$AF,$F5,$6B,$5F,$F6,$43,$5F,$F6,$BB
	.dc.b	$5F,$F3,$77,$E0,$5F,$E9,$B7,$FD,$7C,$D7,$F5,$FE,$62,$31,$1C,$3C
	.dc.b	$7F,$F5,$CF,$DA,$BE,$13,$F8,$74,$FF,$B9,$23,$1D,$DA,$9C,$CF,$EE
	.dc.b	$A7,$2F,$FB,$11,$AF,$FB,$7D,$AF,$38,$E2,$A3,$2F,$EE,$7E,$77,$FC
	.dc.b	$35,$CE,$69,$3C,$FF,$17,$93,$2B,$C1,$87,$F7,$2C,$3D,$53,$99,$F1
	.dc.b	$4E,$49,$1F,$D1,$2F,$FA,$14,$FF,$3D,$6B,$E3,$F1,$34,$47,$FA,$6C
	.dc.b	$69,$FF,$20,$37,$1F,$44,$FC,$CF,$9E,$FF,$43,$E2,$69,$36,$D1,$30
	.dc.b	$F9,$51,$97,$FA,$DE,$CF,$86,$FE,$B6,$D8,$CC,$60,$FF,$79,$FC,$17
	.dc.b	$F5,$6F,$5B,$F9,$7C,$A2,$99,$7F,$99,$B8,$F1,$FF,$CE,$AD,$9D,$CE
	.dc.b	$A5,$F5,$98,$8E,$91,$87,$99,$B8,$D5,$7F,$68,$78,$43,$E8,$5F,$44
	.dc.b	$F8,$7F,$3C,$FA,$27,$31,$88,$FE,$93,$18,$70,$D7,$8A,$FA,$2F,$F5
	.dc.b	$BB,$76,$E5,$3E,$19,$C9,$FE,$CD,$C5,$F7,$6F,$A4,$F9,$15,$0F,$95
	.dc.b	$19,$9F,$7C,$62,$77,$1F,$3A,$75,$3C,$61,$F1,$8F,$A6,$DC,$FD,$28
	.dc.b	$C4,$F4,$37,$3A,$9F,$B5,$3A,$7C,$9C,$BF,$BD,$F6,$BC,$D1,$FD,$8F
	.dc.b	$8F,$FC,$E3,$87,$E2,$A7,$33,$FE,$97,$9C,$BF,$43,$6F,$DA,$7D,$4E
	.dc.b	$9B,$7F,$45,$B3,$BE,$65,$6D,$B5,$B6,$CE,$E7,$47,$8B,$1F,$DF,$7D
	.dc.b	$05,$E9,$34,$98,$7E,$C1,$B7,$FE,$41,$9D,$CE,$4F,$E2,$59,$FF,$BA
	.dc.b	$DB,$E3,$7C,$0C,$BF,$D2,$6D,$9D,$5F,$F1,$76,$BF,$C8,$5A,$FF,$1F
	.dc.b	$6B,$FC,$65,$B8,$FF,$17,$F8,$5F,$51,$30,$FE,$0E,$F7,$F4,$A4,$79
	.dc.b	$9B,$7E,$16,$BE,$16,$BF,$E1,$DF,$A5,$FF,$01,$4F,$F9,$79,$B1,$FE
	.dc.b	$63,$6B,$FE,$57,$6B,$FC,$0B,$E3,$FF,$C5,$F3,$FA,$26,$C7,$EB,$5A
	.dc.b	$FE,$C9,$AF,$FA,$4B,$E5,$BF,$13,$E3,$FF,$A2,$DB,$E4,$1A,$4B,$FF
	.dc.b	$4A,$B6,$BF,$D2,$0D,$7F,$CB,$AD,$7B,$47,$F8,$43,$CB,$7C,$27,$E6
	.dc.b	$3C,$CF,$F0,$03,$63,$33,$B8,$F1,$FC,$E4,$8F,$F5,$2B,$6F,$FC,$15
	.dc.b	$F0,$9C,$49,$3D,$DE,$DF,$AA,$FE,$F2,$7F,$CA,$F8,$A3,$10,$8C,$3C
	.dc.b	$09,$5F,$E2,$89,$DC,$EA,$31,$3F,$23,$8A,$47,$FA,$E9,$B1,$FE,$1F
	.dc.b	$E0,$F7,$12,$7E,$23,$F4,$3E,$6E,$47,$FB,$09,$B3,$FE,$BD,$6B,$FE
	.dc.b	$8A,$6B,$FE,$72,$DD,$C7,$F9,$C6,$F2,$7D,$E8,$C1,$1C,$12,$FF,$D8
	.dc.b	$4F,$EE,$FD,$67,$B7,$49,$3F,$E6,$17,$F8,$8E,$0F,$B6,$FE,$22,$FF
	.dc.b	$40,$6B,$FD,$61,$AF,$FA,$21,$AF,$1D,$39,$C6,$08,$FF,$0D,$F7,$1E
	.dc.b	$6D,$26,$1F,$E8,$D6,$DF,$FA,$35,$AF,$FA,$35,$FC,$FF,$E8,$1F,$DA
	.dc.b	$67,$FA,$96,$A8,$C5,$7F,$BB,$9A,$FF,$70,$8D,$4F,$22,$1C,$A7,$73
	.dc.b	$F4,$DA,$FF,$77,$FA,$0F,$CE,$BF,$D9,$25,$18,$8F,$EF,$6D,$7F,$BD
	.dc.b	$B5,$F9,$1C,$BF,$18,$93,$FD,$8B,$3D,$4F,$D3,$9C,$BF,$DB,$E3,$27
	.dc.b	$3C,$3E,$07,$CC,$FC,$42,$5F,$AD,$B5,$F0,$DA,$FC,$3C,$51,$FD,$07
	.dc.b	$D3,$FE,$C7,$5F,$EF,$86,$D7,$FB,$F5,$AF,$FB,$C5,$AF,$FA,$FD,$AF
	.dc.b	$E6,$9B,$C2,$69,$8F,$81,$7A,$BF,$D4,$5B,$7F,$E0,$DA,$46,$2F,$FC
	.dc.b	$27,$18,$AF,$ED,$3F,$1E,$FE,$00,$7D,$2E,$1B,$8F,$F8,$5E,$75,$39
	.dc.b	$24,$66,$77,$87,$F8,$82,$B7,$39,$AF,$D9,$7B,$52,$3F,$23,$8A,$31
	.dc.b	$7F,$EC,$66,$BF,$91,$8D,$4F,$09,$FC,$DF,$C0,$C8,$FC,$93,$6B,$F3
	.dc.b	$4D,$7E,$C3,$5F,$C6,$FB,$9F,$E0,$A9,$FC,$9B,$78,$FF,$A3,$CF,$E6
	.dc.b	$63,$F7,$FE,$57,$F8,$5E,$3F,$C1,$DE,$6E,$09,$7F,$E3,$49,$D4,$FF
	.dc.b	$96,$1A,$8F,$F2,$34,$F0,$7F,$DA,$5D,$EF,$03,$F9,$C9,$2F,$F9,$93
	.dc.b	$FC,$F3,$D7,$91,$FC,$D5,$B5,$CB,$A7,$FA,$09,$27,$F9,$D3,$F8,$0F
	.dc.b	$9C,$FD,$67,$09,$A4,$EA,$FE,$AD,$C9,$E3,$BF,$F7,$4B,$70,$FF,$74
	.dc.b	$35,$FF,$3A,$F8,$5F,$94,$CF,$89,$2B,$8B,$67,$F8,$77,$A5,$FC,$50
	.dc.b	$93,$FC,$2D,$FE,$B3,$E1,$C8,$FE,$A4,$D9,$F8,$ED,$7C,$3B,$23,$11
	.dc.b	$FD,$17,$B3,$F8,$98,$29,$2B,$DF,$DF,$1F,$7B,$67,$DE,$D7,$DF,$9F
	.dc.b	$F3,$CC,$FF,$67,$3F,$AC,$F6,$FF,$61,$5F,$E9,$8D,$FF,$DA,$97,$92
	.dc.b	$61,$FA,$DA,$FF,$36,$D6,$DA,$FE,$1E,$BF,$DB,$9E,$9F,$52,$57,$16
	.dc.b	$CF,$EC,$6B,$8A,$AF,$F7,$27,$93,$E5,$25,$F8,$5F,$F0,$A7,$DC,$3E
	.dc.b	$C6,$92,$7F,$D3,$1A,$27,$51,$86,$E1,$FE,$A0,$9D,$57,$18,$C4,$E9
	.dc.b	$FE,$99,$F7,$BF,$C8,$36,$34,$9E,$AF,$F8,$9F,$F0,$CF,$E9,$78,$9A
	.dc.b	$4C,$B4,$43,$CE,$E5,$FE,$BE,$77,$24,$FF,$90,$9F,$E6,$32,$FB,$AE
	.dc.b	$E4,$8F,$F1,$C3,$E2,$FC,$2D,$EF,$FF,$8E,$DB,$3F,$19,$B9,$FC,$36
	.dc.b	$BF,$A6,$6B,$FE,$34,$6B,$FE,$33,$6D,$9C,$E7,$FE,$CD,$EE,$FF,$2C
	.dc.b	$F7,$7F,$16,$8C,$61,$FD,$F2,$30,$49,$D9,$23,$FD,$93,$87,$F1,$3F
	.dc.b	$99,$FE,$29,$C2,$F8,$FF,$F6,$CF,$BB,$FE,$C8,$8C,$57,$FB,$72,$B7
	.dc.b	$5A,$BB,$39,$15,$C1,$FE,$41,$7C,$EB,$FD,$C5,$7C,$27,$37,$5F,$21
	.dc.b	$CA,$7B,$34,$ED,$7D,$55,$F5,$73,$3C,$27,$50,$BE,$D5,$B8,$FE,$68
	.dc.b	$64,$EC,$79,$0A,$79,$27,$C1,$1C,$63,$CE,$7F,$BD,$27,$E6,$BC,$CC
	.dc.b	$1F,$F0,$07,$E7,$5F,$3E,$F9,$FF,$F7,$DC,$61,$B1,$99,$E0,$D7,$CE
	.dc.b	$25,$F2,$DF,$BF,$8E,$CD,$77,$1A,$E8,$EF,$39,$9F,$83,$39,$7E,$0B
	.dc.b	$FE,$0E,$7E,$B3,$5C,$C7,$7C,$BE,$F4,$F1,$6F,$5F,$CC,$6D,$72,$AE
	.dc.b	$1A,$78,$1B,$1F,$D9,$6F,$B5,$70,$8E,$F8,$71,$3C,$78,$75,$C3,$D2
	.dc.b	$D9,$E5,$87,$47,$E3,$37,$0E,$B5,$C2,$7A,$1F,$8D,$E0,$FE,$ED,$F2
	.dc.b	$FF,$3A,$3F,$D1,$C1,$C3,$FD,$2C,$C1,$FE,$6C,$29,$F5,$A3,$9B,$98
	.dc.b	$E3,$18,$8D,$1C,$0F,$ED,$7F,$23,$F9,$C7,$F9,$8B,$5F,$E5,$CD,$77
	.dc.b	$BF,$93,$91,$39,$9F,$F6,$0B,$5F,$F6,$03,$63,$73,$C7,$E7,$FF,$C6
	.dc.b	$95,$FD,$01,$AF,$F3,$D6,$BF,$E0,$06,$BF,$EB,$3F,$9F,$FF,$2C,$3F
	.dc.b	$E8,$E6,$BF,$0D,$AF,$FA,$35,$AF,$FB,$5B,$E7,$FC,$84,$AF,$63,$63
	.dc.b	$FA,$9B,$5F,$EA,$0F,$B7,$1F,$DF,$7B,$3F,$3B,$90,$D2,$57,$F5,$CF
	.dc.b	$93,$7C,$0F,$C9,$6C,$66,$30,$47,$04,$3F,$B0,$47,$07,$FC,$C5,$F3
	.dc.b	$FF,$C8,$9F,$F3,$53,$5C,$F2,$FE,$2B,$95,$39,$9F,$F0,$47,$CF,$FF
	.dc.b	$4E,$7F,$CE,$4D,$7F,$CD,$CD,$77,$DB,$CE,$31,$1F,$E0,$96,$BF,$E0
	.dc.b	$36,$D1,$A3,$E7,$F9,$12,$7C,$2F,$E3,$7E,$7D,$E6,$D2,$4F,$F8,$1A
	.dc.b	$32,$D8,$CB,$FE,$05,$8C,$B6,$3F,$47,$EA,$FD,$92,$71,$60,$FC,$E7
	.dc.b	$9C,$F9,$2A,$4B,$FF,$0A,$C6,$5B,$19,$7F,$C2,$91,$96,$C7,$EC,$B6
	.dc.b	$FF,$76,$C7,$FB,$91,$AF,$EE,$5A,$FF,$B8,$DA,$FF,$9E,$F9,$D9,$7E
	.dc.b	$3A,$3E,$17,$17,$C7,$23,$82,$1F,$0D,$CB,$FE,$A9,$F9,$F7,$9B,$F8
	.dc.b	$56,$C7,$FA,$CD,$AE,$E3,$E2,$F5,$C8,$F8,$6D,$7E,$0F,$7F,$CA,$23
	.dc.b	$F2,$1F,$9B,$DF,$23,$E5,$B5,$FE,$45,$F3,$FF,$01,$FF,$4D,$35,$34
	.dc.b	$9E,$9F,$F8,$39,$B1,$FC,$A1,$E1,$F3,$FF,$E7,$78,$FE,$56,$D7,$F9
	.dc.b	$83,$5F,$EB,$ED,$7F,$D5,$5D,$9F,$E0,$EF,$9F,$FB,$44,$8F,$B4,$D7
	.dc.b	$ED,$35,$FB,$4F,$E4,$3C,$29,$1F,$ED,$46,$BF,$CE,$79,$7F,$C1,$2F
	.dc.b	$FA,$25,$B4,$69,$AE,$97,$CF,$BE,$7F,$F7,$CD,$BF,$F7,$63,$5F,$F7
	.dc.b	$43,$5F,$F4,$43,$5F,$F4,$13,$5F,$F3,$E3,$67,$BC,$FF,$54,$AF,$F8
	.dc.b	$03,$F3,$17,$E0,$0D,$26,$1F,$D7,$65,$F2,$FF,$CE,$A3,$D8,$D8,$CF
	.dc.b	$BF,$D6,$30,$47,$04,$9F,$D5,$7C,$17,$EB,$48,$E0,$87,$EB,$5C,$B4
	.dc.b	$FE,$DB,$F0,$EF,$9F,$FF,$2B,$B6,$33,$18,$23,$82,$1F,$DE,$63,$83
	.dc.b	$F8,$CA,$FF,$30,$F0,$FB,$E0,$D3,$E1,$6F,$FE,$5A,$F6,$FF,$83,$12
	.dc.b	$7F,$83,$35,$FE,$0C,$D7,$F8,$33,$F4,$FE,$21,$E9,$52,$4F,$D1,$8C
	.dc.b	$B6,$32,$FF,$8F,$B5,$4E,$A3,$0D,$7F,$C8,$7E,$73,$E7,$FF,$CF,$AD
	.dc.b	$8F,$F0,$BB,$5F,$F0,$E3,$5E,$31,$C0,$8F,$07,$F1,$9F,$14,$F8,$5A
	.dc.b	$49,$FC,$54,$65,$FF,$14,$67,$FA,$5F,$33,$F9,$E9,$23,$39,$23,$11
	.dc.b	$FA,$F6,$BF,$E7,$49,$E4,$FF,$35,$F9,$FF,$E7,$04,$70,$43,$F9,$C3
	.dc.b	$96,$9F,$ED,$AF,$9F,$7C,$F7,$F8,$9D,$B1,$98,$C1,$FD,$17,$E7,$E4
	.dc.b	$7F,$45,$6B,$FD,$32,$3F,$AA,$91,$FE,$A6,$D7,$46,$08,$7F,$B9,$FE
	.dc.b	$47,$FD,$67,$E4,$1A,$48,$FE,$AD,$E3,$C7,$FA,$B3,$5E,$5C,$5F,$E4
	.dc.b	$19,$FC,$24,$7F,$55,$6A,$AE,$D3,$FD,$87,$96,$4F,$F6,$16,$BF,$D8
	.dc.b	$5A,$FF,$61,$6B,$FC,$CB,$C9,$FD,$39,$23,$30,$1F,$3F,$FE,$5D,$7F
	.dc.b	$D5,$ED,$74,$D7,$7E,$3F,$F9,$BB,$CD,$7C,$D7,$FA,$46,$7F,$D6,$F1
	.dc.b	$8F,$27,$53,$99,$FF,$67,$B5,$7C,$BB,$E1,$BF,$C9,$F1,$86,$C6,$7C
	.dc.b	$0F,$F3,$EB,$67,$FD,$04,$D7,$FD,$62,$D7,$FD,$92,$D7,$E0,$35,$FE
	.dc.b	$8D,$F3,$1F,$E1,$17,$FD,$9F,$D0,$8C,$47,$C7,$8C,$7A,$1F,$CE,$E7
	.dc.b	$33,$E7,$F3,$3C,$B8,$FF,$73,$C3,$D9,$18,$24,$FA,$27,$44,$8E,$9E
	.dc.b	$2C,$9F,$E8,$57,$FD,$19,$B1,$97,$F8,$94,$65,$B1,$97,$E0,$FC,$FF
	.dc.b	$FA,$37,$03,$49,$B7,$CC,$6C,$FC,$C6,$BF,$C7,$F5,$7C,$6F,$F5,$B6
	.dc.b	$C7,$F5,$76,$BB,$8F,$F5,$9F,$C0,$FF,$5A,$F6,$3F,$DF,$6C,$F2,$7F
	.dc.b	$CD,$0D,$8C,$D7,$E5,$63,$83,$4C,$4B,$E7,$F7,$09,$FE,$CF,$18,$6C
	.dc.b	$71,$35,$3C,$5C,$0F,$9E,$7C,$FB,$E1,$7F,$D9,$71,$86,$C6,$67,$FD
	.dc.b	$A5,$39,$6F,$C7,$FF,$AE,$E0,$D2,$6D,$FE,$D9,$C5,$F3,$EF,$8F,$F5
	.dc.b	$61,$0F,$64,$66,$7D,$47,$09,$DB,$F3,$23,$53,$C6,$1F,$16,$F9,$2D
	.dc.b	$CF,$CD,$8C,$4F,$46,$C7,$16,$C7,$E9,$38,$7F,$CC,$2F,$DF,$F6,$FC
	.dc.b	$D5,$FA,$59,$CB,$F8,$78,$CB,$F2,$F9,$1F,$3F,$8E,$0F,$E0,$2D,$8C
	.dc.b	$E1,$FE,$FD,$9C,$B6,$33,$5F,$C6,$4F,$E3,$53,$A7,$07,$7A,$E8,$78
	.dc.b	$BA,$FF,$79,$F3,$F7,$A4,$D2,$63,$FB,$06,$DF,$EB,$9A,$FF,$0C,$CB
	.dc.b	$FD,$B3,$3F,$C4,$9B,$1F,$C4,$9A,$EF,$C6,$F8,$79,$FF,$A2,$9B,$3A
	.dc.b	$9F,$E2,$AD,$7F,$92,$B7,$1F,$E2,$9F,$0B,$E9,$26,$1F,$C1,$9E,$7E
	.dc.b	$9C,$8F,$23,$6F,$C8,$D7,$D4,$D7,$F8,$F7,$CF,$FF,$80,$67,$FC,$A4
	.dc.b	$D8,$FF,$10,$35,$FE,$4E,$D7,$EA,$7C,$FF,$F8,$B2,$7E,$9C,$7F,$4F
	.dc.b	$8C,$38,$6B,$F6,$1A,$FF,$59,$F9,$FF,$C4,$FA,$0F,$25,$18,$6D,$FF
	.dc.b	$B1,$63,$04,$70,$4A,$FF,$64,$46,$5B,$19,$7F,$19,$E5,$FF,$7F,$F8
	.dc.b	$5F,$CC,$B6,$7D,$6D,$9F,$F0,$34,$F8,$FE,$7E,$57,$FB,$5D,$B3,$FE
	.dc.b	$0B,$F8,$6E,$7E,$79,$A4,$AE,$BF,$2F,$FC,$52,$BF,$C0,$8D,$C4,$D2
	.dc.b	$47,$99,$B3,$CA,$3E,$37,$14,$8F,$F7,$53,$63,$FC,$3D,$C1,$E4,$24
	.dc.b	$FB,$B1,$47,$CF,$F9,$F9,$3F,$EF,$36,$D7,$FB,$CD,$AF,$FB,$CD,$AF
	.dc.b	$FB,$13,$8F,$F3,$9C,$1F,$EB,$5E,$4F,$BB,$18,$23,$82,$5F,$F9,$51
	.dc.b	$FD,$D7,$AE,$F7,$69,$27,$FC,$BE,$FF,$10,$C1,$F3,$FF,$C5,$1F,$E0
	.dc.b	$AD,$7F,$85,$35,$FF,$12,$35,$E3,$AF,$38,$C1,$1F,$E1,$9E,$EB,$CD
	.dc.b	$A4,$C3,$FD,$14,$DB,$FF,$42,$B5,$FF,$07,$CF,$19,$FF,$3A,$FC,$FF
	.dc.b	$F5,$B6,$9F,$A2,$71,$3F,$C7,$B5,$46,$23,$FC,$E8,$D7,$FD,$E6,$D5
	.dc.b	$1C,$27,$A1,$FC,$D8,$7C,$E3,$9B,$0D,$8F,$EB,$0D,$AF,$E5,$F1,$C2
	.dc.b	$7E,$31,$1F,$E0,$31,$F4,$A3,$0F,$F0,$4C,$F9,$3F,$52,$32,$D9,$F8
	.dc.b	$7F,$3F,$F0,$48,$E0,$91,$F0,$F1,$4E,$A3,$0F,$F4,$1F,$4F,$FB,$1D
	.dc.b	$7F,$BB,$5B,$3F,$EB,$B6,$BF,$E4,$26,$BF,$E4,$16,$AE,$14,$61,$BE
	.dc.b	$27,$73,$FA,$87,$C1,$BD,$5A,$4C,$3F,$DF,$D1,$97,$F9,$3F,$C2,$BE
	.dc.b	$7D,$F3,$FF,$DA,$A3,$0D,$8C,$CF,$06,$E7,$FD,$8A,$B7,$39,$AE,$04
	.dc.b	$7F,$D7,$FE,$DF,$EC,$88,$E0,$95,$F8,$FC,$D3,$A8,$C3,$F9,$5F,$82
	.dc.b	$95,$F9,$A6,$CF,$D5,$6B,$F9,$66,$BF,$E3,$FE,$2F,$E8,$2F,$F8,$ED
	.dc.b	$B3,$F9,$4F,$77,$F2,$67,$EF,$7C,$BF,$F0,$BD,$7F,$80,$1A,$FF,$81
	.dc.b	$5A,$FF,$81,$BC,$DE,$04,$AF,$CA,$CE,$A7,$FC,$81,$3F,$CA,$FC,$29
	.dc.b	$5F,$CD,$1B,$3F,$E0,$EF,$0B,$F9,$9F,$AF,$FC,$C4,$95,$FC,$E9,$FE
	.dc.b	$69,$D7,$23,$F9,$EB,$6B,$EB,$7C,$FF,$E8,$9F,$F5,$A7,$0A,$31,$08
	.dc.b	$C7,$4F,$E0,$E7,$38,$FE,$37,$2F,$EA,$FC,$BF,$D3,$72,$FE,$AE,$DC
	.dc.b	$7D,$0D,$7F,$AF,$B5,$FF,$62,$57,$39,$E3,$5D,$0F,$E2,$5E,$9C,$8F
	.dc.b	$E2,$ED,$AF,$89,$E2,$FF,$48,$9F,$EB,$7E,$27,$F5,$52,$38,$24,$F0
	.dc.b	$8C,$BF,$C7,$7B,$BF,$8C,$1F,$C6,$27,$FC,$06,$E0,$FE,$39,$8F,$EB
	.dc.b	$7C,$0F,$82,$49,$F8,$2D,$AF,$82,$D7,$E0,$BF,$AD,$F6,$FF,$65,$1F
	.dc.b	$A9,$6B,$FE,$4F,$DF,$FD,$AD,$AE,$CE,$1E,$9C,$C3,$F7,$37,$FE,$72
	.dc.b	$BE,$0D,$7E,$EB,$5F,$B8,$DB,$FF,$71,$57,$39,$F1,$79,$5E,$52,$5F
	.dc.b	$67,$FC,$2B,$F3,$CF,$9F,$A4,$9F,$F0,$BC,$65,$B1,$9A,$FC,$D4,$E5
	.dc.b	$B1,$FE,$30,$F9,$FF,$E3,$FB,$1A,$4F,$07,$FC,$4B,$F3,$FF,$48,$98
	.dc.b	$E0,$87,$F8,$A1,$CB,$4F,$F1,$E9,$D8,$91,$83,$FC,$63,$F3,$FF,$78
	.dc.b	$91,$83,$FC,$6B,$F3,$EF,$7E,$92,$3F,$C6,$F1,$96,$C6,$73,$FD,$D4
	.dc.b	$E5,$B1,$FE,$50,$ED,$FE,$27,$9F,$FB,$0F,$BB,$04,$C3,$FC,$BB,$19
	.dc.b	$6C,$F0,$24,$FF,$84,$EC,$3E,$66,$F0,$3E,$47,$FD,$33,$EE,$FF,$B0
	.dc.b	$A3,$15,$FE,$9C,$AD,$D6,$B1,$B3,$93,$FC,$CD,$19,$69,$9A,$E6,$F8
	.dc.b	$AF,$B4,$FF,$9D,$AF,$9C,$E6,$EB,$E8,$8B,$5C,$23,$A5,$77,$BE,$B1
	.dc.b	$B3,$C2,$73,$29,$97,$73,$CE,$3F,$D7,$95,$DA,$FC,$31,$F1,$CF,$F0
	.dc.b	$59,$4F,$F0,$6C,$61,$C1,$FE,$C5,$F9,$D7,$CF,$3E,$7F,$FC,$27,$18
	.dc.b	$6C,$66,$7F,$D2,$F3,$96,$FC,$BF,$F3,$16,$F4,$72,$8C,$47,$FB,$42
	.dc.b	$51,$88,$F1,$3C,$7E,$83,$E8,$CF,$F8,$0E,$7F,$07,$18,$EB,$FF,$6E
	.dc.b	$3A,$AE,$1B,$72,$C3,$6F,$B6,$B6,$DB,$C9,$F8,$EC,$3C,$75,$88,$46
	.dc.b	$1F,$AA,$F8,$DB,$5F,$EE,$DD,$BC,$55,$F8,$4A,$E1,$8F,$9E,$FA,$47
	.dc.b	$18,$DC,$1A,$6D,$C6,$1E,$76,$9E,$7C,$7D,$EF,$F9,$58,$F2,$5B,$5F
	.dc.b	$D4,$F0,$FE,$AB,$3F,$E6,$91,$8A,$DB,$7E,$2B,$31,$CE,$6D,$71,$8F
	.dc.b	$ED,$F5,$C6,$73,$0A,$DF,$13,$D6,$A4,$9E,$23,$BD,$F2,$4E,$4D,$4F
	.dc.b	$F7,$9C,$BD,$38,$73,$8E,$39,$A3,$11,$B3,$A5,$F3,$77,$F3,$FF,$E5
	.dc.b	$07,$F8,$8B,$5F,$F4,$23,$5F,$E1,$E7,$FA,$01,$C3,$F2,$AD,$F6,$9F
	.dc.b	$11,$B1,$E4,$6B,$E3,$C1,$5D,$F6,$FA,$93,$FE,$8C,$E6,$F4,$BE,$0A
	.dc.b	$FE,$3D,$AF,$FA,$09,$B1,$FC,$A2,$76,$FD,$C6,$BC,$E7,$84,$62,$78
	.dc.b	$CF,$FA,$87,$B5,$F4,$5F,$71,$B1,$F3,$1A,$FA,$6B,$E6,$E1,$E8,$1F
	.dc.b	$40,$E7,$FF,$4B,$C6,$1B,$19,$AF,$F3,$AB,$5F,$F3,$73,$5F,$F0,$33
	.dc.b	$5F,$F1,$DB,$5F,$F1,$D3,$7D,$17,$DD,$79,$36,$FE,$8B,$18,$D7,$F3
	.dc.b	$2D,$BF,$C7,$E3,$E5,$6F,$D0,$BE,$87,$D2,$D8,$FF,$45,$B5,$FF,$42
	.dc.b	$35,$FF,$0E,$35,$FF,$28,$B5,$FF,$28,$36,$4E,$6B,$D3,$E4,$F9,$B2
	.dc.b	$34,$FD,$8D,$70,$C3,$F6,$35,$F9,$BC,$3F,$D1,$9C,$BF,$85,$8C,$47
	.dc.b	$E2,$5A,$FF,$00,$6B,$FE,$DA,$6F,$0F,$FA,$CE,$79,$1F,$8F,$5F,$1B
	.dc.b	$FC,$01,$B5,$FB,$06,$BF,$9D,$AF,$26,$1F,$9C,$9F,$C2,$7A,$1F,$99
	.dc.b	$E0,$FF,$7F,$5F,$F8,$01,$AF,$FB,$CD,$AF,$E8,$5A,$FF,$18,$6B,$FE
	.dc.b	$F2,$6C,$F6,$8F,$F8,$03,$87,$F1,$6F,$6F,$CB,$87,$8D,$BC,$1C,$F1
	.dc.b	$F9,$BF,$40,$F3,$7F,$32,$DC,$7F,$D9,$6D,$7F,$D8,$8D,$7F,$CB,$8D
	.dc.b	$7F,$D2,$2D,$7F,$D2,$0D,$C0,$E7,$AD,$F4,$2F,$84,$FE,$5F,$18,$6C
	.dc.b	$66,$FF,$33,$5E,$3B,$FE,$13,$6F,$4B,$F9,$CD,$7F,$2B,$6C,$67,$C0
	.dc.b	$EB,$1F,$CD,$67,$9F,$CB,$FC,$02,$3F,$01,$FE,$21,$B3,$E0,$BE,$61
	.dc.b	$27,$F9,$CE,$1E,$1F,$0F,$FA,$3C,$3E,$73,$F3,$6F,$F2,$A3,$F9,$5B
	.dc.b	$87,$06,$FA,$3F,$C2,$B1,$98,$E0,$D7,$8C,$68,$8F,$F9,$2B,$93,$F8
	.dc.b	$BE,$BF,$91,$6F,$1F,$F6,$06,$CF,$F9,$2A,$FC,$55,$F8,$5F,$BA,$FC
	.dc.b	$41,$27,$FA,$6F,$33,$E8,$71,$1F,$CE,$F6,$56,$2F,$F4,$17,$BA,$FD
	.dc.b	$37,$DB,$FE,$A4,$8F,$F5,$8E,$9F,$F2,$5D,$79,$E7,$33,$FD,$1F,$34
	.dc.b	$7F,$00,$F2,$3F,$8D,$FC,$13,$E6,$29,$2F,$F8,$EF,$D1,$3E,$8B,$11
	.dc.b	$FD,$8F,$35,$F1,$8F,$EE,$31,$8B,$FE,$1F,$7C,$67,$FB,$6E,$CF,$99
	.dc.b	$FD,$61,$23,$8E,$1C,$DF,$47,$95,$FC,$1C,$8F,$6F,$99,$7C,$2E,$23
	.dc.b	$67,$F0,$B9,$FE,$6F,$3A,$8F,$0E,$9F,$BB,$23,$F5,$3E,$22,$BF,$4E
	.dc.b	$2F,$0F,$D2,$9F,$C9,$F2,$F3,$AF,$C7,$B5,$FC,$6E,$3F,$90,$23,$F9
	.dc.b	$43,$F3,$4E,$1E,$2F,$F2,$5F,$96,$FE,$48,$48,$FF,$0D,$9C,$0F,$93
	.dc.b	$F4,$3F,$C4,$C8,$FE,$33,$A6,$BF,$8B,$8B,$8F,$E2,$A7,$F9,$87,$8F
	.dc.b	$2B,$F9,$83,$FD,$73,$E3,$5E,$C6,$67,$F1,$13,$6F,$FC,$1B,$CA,$BD
	.dc.b	$D7,$E0,$3E,$39,$F1,$5E,$FA,$46,$27,$F6,$B1,$89,$D4,$E6,$7F,$79
	.dc.b	$39,$77,$19,$7F,$C5,$31,$97,$71,$97,$FC,$A7,$CB,$FE,$F1,$7F,$1F
	.dc.b	$F0,$1F,$CE,$4F,$EF,$33,$DF,$1F,$DE,$7D,$BF,$F5,$32,$57,$EA,$FE
	.dc.b	$8A,$BF,$A9,$8B,$3F,$D4,$F1,$57,$F4,$5F,$81,$97,$F9,$E7,$FC,$BB
	.dc.b	$DE,$FA,$2C,$CF,$EF,$E6,$E1,$FB,$E9,$DF,$8F,$FD,$74,$95,$FD,$75
	.dc.b	$FC,$57,$8D,$FC,$D7,$E8,$7F,$98,$92,$7F,$86,$CF,$F5,$7F,$62,$47
	.dc.b	$EC,$1D,$3F,$C3,$DC,$B4,$FE,$63,$E5,$2B,$D3,$EA,$FF,$6F,$24,$FF
	.dc.b	$6F,$7F,$9D,$F8,$BF,$D9,$49,$1F,$C7,$A3,$E0,$F5,$FF,$77,$24,$7F
	.dc.b	$77,$7F,$A1,$F8,$1F,$DB,$49,$1F,$C9,$A2,$BF,$C9,$62,$EB,$FC,$92
	.dc.b	$6F,$93,$F8,$09,$B5,$FD,$EE,$7F,$D1,$FE,$9B,$E7,$3F,$D5,$93,$FE
	.dc.b	$72,$8C,$79,$DF,$B5,$6C,$7F,$4A,$6B,$FD,$3F,$07,$88,$F8,$EF,$F4
	.dc.b	$DC,$61,$B1,$99,$FF,$49,$D7,$2C,$FF,$53,$F3,$8F,$A0,$FD,$33,$6B
	.dc.b	$F4,$AD,$7F,$46,$D7,$F4,$2D,$7F,$CA,$AD,$5D,$28,$C3,$67,$FD,$01
	.dc.b	$F6,$7F,$CB,$6F,$FC,$58,$DC,$FF,$D7,$6D,$C3,$9E,$1E,$3A,$F4,$F9
	.dc.b	$9E,$82,$4F,$A1,$AF,$A1,$AF,$A1,$F0,$E1,$FE,$C0,$BE,$98,$78,$23
	.dc.b	$FC,$DB,$87,$87,$CE,$56,$B0,$46,$2B,$91,$C3,$0F,$F3,$AE,$EF,$91
	.dc.b	$FE,$24,$DB,$FE,$22,$D7,$F8,$7B,$5F,$E1,$AD,$7F,$D0,$2D,$7E,$6B
	.dc.b	$74,$FC,$64,$FF,$87,$3C,$8F,$F2,$31,$2B,$FB,$1D,$77,$D7,$FD,$6B
	.dc.b	$AF,$E2,$BE,$1F,$FD,$87,$EF,$7E,$07,$5F,$7F,$C6,$63,$6F,$A6,$DC
	.dc.b	$BF,$DE,$1C,$4C,$31,$8F,$E5,$3D,$0D,$62,$F1,$BF,$D1,$D9,$F5,$6E
	.dc.b	$6C,$B9,$DF,$55,$F4,$1F,$E7,$6F,$84,$FE,$B0,$D8,$FF,$16,$B5,$FF
	.dc.b	$19,$35,$FF,$6E,$AF,$90,$FF,$42,$B7,$9F,$9C,$60,$FF,$45,$E8,$FA
	.dc.b	$07,$07,$FB,$6B,$5F,$47,$91,$75,$F4,$5E,$EB,$19,$FE,$13,$65,$E2
	.dc.b	$9E,$37,$ED,$DF,$81,$8F,$CF,$E9,$7C,$3E,$EB,$58,$AB,$53,$DB,$8B
	.dc.b	$E6,$71,$FF,$B5,$F1,$FE,$9E,$DB,$FC,$2F,$BF,$6B,$75,$F8,$28,$FF
	.dc.b	$7A,$CF,$E3,$9A,$FE,$39,$AF,$28,$FC,$BC,$63,$AB,$F8,$D3,$7A,$3F
	.dc.b	$D1,$D3,$A3,$FC,$C3,$7F,$A8,$F6,$FC,$5A,$9F,$36,$1F,$C1,$7E,$83
	.dc.b	$31,$2F,$F0,$9D,$5F,$C1,$23,$7D,$CA,$D6,$9F,$E4,$18,$C5,$FF,$18
	.dc.b	$9F,$34,$62,$7F,$8C,$C6,$09,$3B,$24,$7F,$9E,$F2,$FD,$FF,$C5,$FA
	.dc.b	$0F,$0F,$CF,$DA,$FF,$26,$E0,$8E,$E4,$8F,$E5,$2D,$BF,$E5,$2D,$7F
	.dc.b	$94,$BF,$3A,$3F,$0B,$F0,$9F,$65,$A7,$F9,$6B,$E7,$FF,$BE,$BF,$10
	.dc.b	$FC,$8F,$27,$FB,$69,$D7,$A1,$F5,$67,$78,$7F,$86,$9A,$FF,$86,$1A
	.dc.b	$FF,$85,$BC,$4F,$C3,$74,$7E,$43,$83,$FC,$FF,$F3,$D8,$99,$8F,$F9
	.dc.b	$A1,$FC,$67,$A2,$9C,$E1,$FE,$89,$9B,$7C,$AF,$81,$FE,$90,$AF,$C6
	.dc.b	$F0,$30,$E5,$3F,$EE,$99,$CC,$63,$0F,$A9,$8B,$DE,$F3,$9E,$2F,$85
	.dc.b	$B8,$7F,$A7,$1E,$24,$7F,$DE,$8F,$83,$E3,$5E,$27,$E9,$5B,$3F,$EC
	.dc.b	$0F,$95,$FD,$EE,$2D,$7F,$E0,$05,$6A,$31,$5F,$C6,$75,$4E,$6B,$78
	.dc.b	$7E,$CF,$D4,$63,$D2,$73,$87,$7C,$75,$39,$8E,$9A,$FE,$B2,$7B,$8B
	.dc.b	$E6,$FD,$2E,$CF,$8E,$4A,$FD,$8D,$71,$9F,$E5,$C7,$F5,$8C,$BE,$84
	.dc.b	$FE,$2E,$31,$3F,$73,$E7,$7E,$23,$27,$5A,$7B,$A3,$51,$8B,$E7,$7E
	.dc.b	$6A,$FE,$CF,$D3,$FC,$00,$FE,$25,$D6,$F4,$31,$87,$4E,$DF,$DB,$6F
	.dc.b	$FE,$02,$BD,$D6,$B1,$E9,$97,$FB,$17,$0F,$07,$9B,$FC,$2B,$D7,$FF
	.dc.b	$7D,$46,$0F,$F4,$17,$C3,$7F,$17,$8C,$DF,$F9,$DA,$30,$49,$FF,$3E
	.dc.b	$B6,$35,$1F,$C6,$B3,$FF,$29,$3E,$7F,$53,$F8,$CF,$B6,$F8,$9F,$E3
	.dc.b	$B1,$27,$47,$FB,$4A,$74,$E2,$FD,$7E,$57,$E4,$4F,$D8,$C5,$3E,$63
	.dc.b	$63,$FD,$E9,$17,$7F,$F1,$3F,$CD,$7F,$2C,$3F,$84,$7C,$D3,$E6,$B1
	.dc.b	$5F,$74,$46,$5F,$C1,$DF,$F5,$A8,$C3,$FE,$4D,$6B,$F8,$88,$E3,$3F
	.dc.b	$C5,$E7,$33,$C6,$30,$49,$FE,$83,$C0,$FA,$4F,$D7,$1F,$C3,$BE,$76
	.dc.b	$CF,$E2,$62,$DF,$E2,$7B,$D5,$8C,$FF,$C8,$BF,$63,$FD,$28,$FE,$93
	.dc.b	$F1,$1F,$DB,$31,$FC,$54,$65,$A7,$E7,$7A,$13,$BC,$7F,$2D,$C6,$BF
	.dc.b	$E0,$5E,$8C,$C3,$FE,$00,$FB,$56,$FF,$63,$8B,$E2,$93,$FD,$69,$B5
	.dc.b	$FA,$BC,$DF,$3C,$E3,$D5,$CA,$EA,$6E,$74,$E3,$1F,$2F,$D8,$FF,$68
	.dc.b	$3F,$53,$EF,$59,$FD,$44,$59,$FD,$86,$6A,$C6,$5F,$C6,$3D,$4F,$E3
	.dc.b	$C7,$F3,$EF,$86,$B8,$FE,$E6,$2E,$1F,$46,$6D,$7F,$9D,$3E,$C7,$FB
	.dc.b	$C9,$F8,$EF,$89,$B8,$7E,$37,$A7,$FD,$EE,$FE,$33,$03,$F9,$09,$FC
	.dc.b	$41,$1D,$19,$2C,$F2,$9F,$C9,$4E,$DF,$EE,$9D,$1F,$39,$FE,$07,$BF
	.dc.b	$F2,$8F,$84,$FE,$74,$FF,$88,$72,$7D,$07,$EC,$2B,$F3,$F1,$8E,$FF
	.dc.b	$F0,$EB,$73,$F4,$36,$7F,$4B,$E4,$FF,$38,$F8,$3F,$E6,$A7,$F3,$6B
	.dc.b	$B7,$FC,$A6,$7F,$07,$E9,$7F,$3F,$CB,$FB,$8E,$3F,$0B,$1F,$E9,$59
	.dc.b	$7E,$BA,$31,$7C,$30,$FC,$44,$FE,$B2,$B7,$DF,$E6,$8C,$5F,$B2,$30
	.dc.b	$4A,$F4,$CE,$8E,$5E,$47,$F5,$4F,$63,$FA,$F1,$FD,$FB,$DE,$B5,$FC
	.dc.b	$F3,$C1,$FF,$33,$3F,$E3,$86,$E5,$FD,$C2,$7F,$BB,$C6,$23,$FA,$5D
	.dc.b	$FF,$9F,$67,$24,$AE,$4F,$F7,$AD,$7F,$C0,$E7,$F9,$4E,$7F,$0B,$3F
	.dc.b	$D3,$3C,$BF,$8E,$FF,$6B,$8C,$DF,$F8,$CD,$AF,$81,$AE,$E3,$FC,$17
	.dc.b	$18,$8F,$F5,$AB,$4F,$F5,$BC,$68,$FF,$2D,$7D,$5F,$FA,$65,$FE,$0C
	.dc.b	$D9,$FE,$DF,$AB,$E8,$7F,$20,$FD,$BE,$B5,$6E,$3E,$F1,$BE,$1F,$C6
	.dc.b	$D7,$E5,$EF,$FB,$8C,$E7,$0F,$C7,$CE,$67,$FC,$BE,$D7,$FC,$20,$D9
	.dc.b	$E1,$1F,$9A,$8D,$61,$E2,$F8,$CF,$F3,$5E,$27,$F4,$53,$FD,$D2,$FD
	.dc.b	$13,$9C,$3F,$8B,$4F,$F7,$8C,$3C,$AD,$7C,$79,$FA,$89,$3E,$86,$C7
	.dc.b	$8D,$FF,$1D,$FC,$13,$BB,$FD,$20,$DA,$FF,$04,$FC,$CF,$F3,$1C,$BF
	.dc.b	$49,$B2,$B5,$8F,$F8,$26,$BF,$9B,$3F,$8E,$6B,$FE,$2F,$6B,$FE,$F2
	.dc.b	$6C,$9A,$4F,$6B,$FC,$9E,$EB,$E0,$5F,$19,$FE,$A5,$6E,$5F,$E1,$E3
	.dc.b	$CC,$7F,$8F,$5C,$E1,$FE,$1D,$8E,$E4,$BF,$F2,$DB,$E2,$F9,$A7,$D6
	.dc.b	$7F,$AC,$5B,$1F,$C1,$FE,$7F,$F1,$0F,$F9,$26,$33,$7F,$E2,$96,$EB
	.dc.b	$F1,$9B,$3F,$EE,$7E,$B5,$F2,$B6,$AB,$DC,$63,$DA,$F9,$D8,$7F,$8E
	.dc.b	$9B,$1F,$E3,$77,$F9,$86,$3F,$51,$CB,$77,$FF,$65,$46,$2B,$FC,$C0
	.dc.b	$F0,$6C,$7F,$B4,$6B,$FC,$9F,$AF,$FB,$F7,$C0,$7C,$AF,$F8,$42,$BF
	.dc.b	$92,$46,$2B,$FC,$27,$5B,$AD,$5D,$9C,$9F,$E6,$E8,$CB,$4C,$D7,$34
	.dc.b	$7B,$9C,$3F,$EF,$03,$FD,$21,$1C,$B8,$F3,$1F,$EE,$F3,$27,$F9,$A3
	.dc.b	$0F,$E7,$51,$AC,$3C,$93,$DA,$7D,$75,$FD,$86,$73,$3E,$18,$C3,$E7
	.dc.b	$38,$61,$A7,$C1,$87,$9A,$F9,$C7,$C0,$3F,$B6,$1E,$52,$DE,$1F,$61
	.dc.b	$FE,$3C,$F9,$E7,$C2,$BE,$7F,$8C,$61,$B1,$99,$FE,$F1,$39,$6F,$7F
	.dc.b	$FA,$85,$BB,$72,$8C,$47,$76,$BF,$E8,$E9,$FF,$7F,$4F,$D2,$D7,$EA
	.dc.b	$47,$F8,$A2,$71,$AE,$FB,$3F,$D2,$B5,$AC,$3C,$0D,$C3,$77,$F7,$30
	.dc.b	$E4,$FD,$60,$E2,$78,$46,$63,$C6,$DB,$EF,$87,$86,$3A,$9C,$B0,$F0
	.dc.b	$C7,$AD,$B1,$E1,$9E,$2D,$38,$D7,$6B,$F0,$9F,$02,$77,$3E,$66,$CE
	.dc.b	$EF,$FD,$95,$2B,$FB,$FC,$3F,$9A,$6C,$64,$FF,$17,$46,$5A,$64,$FF
	.dc.b	$3E,$B9,$69,$93,$FB,$C9,$4F,$E1,$9F,$3E,$EE,$FC,$7F,$2A,$31,$3F
	.dc.b	$E3,$D8,$C4,$EA,$73,$3F,$E2,$86,$BF,$ED,$09,$DB,$C6,$32,$F8,$27
	.dc.b	$BF,$D6,$3D,$FA,$49,$F0,$46,$5F,$F1,$BF,$CF,$BE,$1F,$F4,$BC,$08
	.dc.b	$C4,$EA,$73,$18,$8F,$F1,$EE,$DF,$E8,$E7,$D8,$D7,$3E,$BF,$C2,$8C
	.dc.b	$47,$AD,$AF,$F3,$36,$BF,$EF,$56,$E9,$F9,$49,$FB,$7D,$2F,$97,$A4
	.dc.b	$BF,$B7,$19,$7E,$76,$BF,$DA,$DB,$E0,$7D,$48,$C4,$7F,$95,$9A,$FC
	.dc.b	$F6,$BF,$D1,$DA,$B0,$FB,$D3,$CC,$FC,$7F,$89,$FC,$2F,$84,$D2,$57
	.dc.b	$E1,$3E,$7B,$FD,$9C,$FE,$0D,$AF,$F9,$A1,$AF,$E0,$5A,$FF,$59,$6B
	.dc.b	$E3,$6D,$9A,$6A,$BC,$7D,$5E,$62,$57,$E8,$AF,$FC,$14,$FF,$26,$FB
	.dc.b	$D7,$C4,$FF,$25,$6C,$66,$30,$47,$04,$3F,$D3,$51,$C1,$FF,$4E,$7C
	.dc.b	$DF,$F7,$C7,$F9,$9B,$5C,$F8,$5F,$CC,$E3,$11,$FE,$EE,$97,$CB,$78
	.dc.b	$C9,$1E,$76,$BE,$06,$BC,$5E,$9E,$7C,$8F,$0B,$5F,$3F,$2F,$BC,$8F
	.dc.b	$C3,$6B,$A3,$04,$3F,$D3,$FF,$39,$EF,$7F,$70,$D7,$FC,$20,$D7,$F1
	.dc.b	$4D,$DC,$BE,$2F,$F8,$4B,$67,$F8,$43,$5E,$F1,$F5,$7C,$D9,$1F,$65
	.dc.b	$AF,$FA,$CB,$F5,$A7,$8F,$FC,$51,$B1,$FC,$4D,$AE,$E3,$FB,$27,$DE
	.dc.b	$FE,$AE,$3F,$DA,$8D,$7F,$C0,$A7,$F8,$19,$C3,$86,$B9,$E8,$E6,$47
	.dc.b	$FB,$9C,$6B,$89,$F7,$BF,$C0,$9B,$D5,$FD,$E1,$B5,$FE,$E5,$6B,$FE
	.dc.b	$DC,$F1,$7F,$9F,$F9,$1F,$DE,$23,$F9,$EB,$5D,$18,$8D,$91,$C1,$C3
	.dc.b	$EC,$B9,$91,$C1,$0F,$F7,$5B,$96,$9F,$E0,$EF,$3F,$F9,$AB,$FE,$2A
	.dc.b	$6B,$FE,$A3,$ED,$FF,$08,$46,$08,$E0,$91,$EF,$FB,$CF,$E9,$CF,$F0
	.dc.b	$C6,$E2,$69,$AE,$C6,$4F,$BA,$F8,$A4,$BF,$8A,$D7,$FA,$FC,$6D,$C9
	.dc.b	$0C,$FD,$D7,$CE,$23,$F3,$9A,$E8,$C1,$0F,$C3,$9F,$E4,$CF,$C7,$9E
	.dc.b	$EF,$F8,$C5,$B1,$FE,$36,$6B,$C0,$38,$79,$B5,$FC,$9B,$5E,$D1,$D4
	.dc.b	$FF,$1F,$7B,$EF,$4E,$92,$7F,$C8,$B1,$97,$F1,$3F,$77,$F8,$C2,$38
	.dc.b	$21,$F2,$1C,$B4,$FC,$B7,$C7,$7F,$B2,$DB,$EC,$FE,$6E,$31,$1F,$13
	.dc.b	$74,$62,$3F,$5B,$18,$9E,$04,$8E,$67,$F9,$73,$E1,$BF,$92,$B7,$C1
	.dc.b	$FF,$2E,$46,$23,$E5,$35,$FA,$6D,$7F,$76,$D7,$FD,$02,$D4,$DF,$AB
	.dc.b	$72,$FF,$9A,$5B,$3F,$E6,$E6,$BF,$65,$AF,$B3,$F6,$67,$C2,$FE,$F5
	.dc.b	$B1,$FB,$D6,$BF,$15,$AF,$18,$E0,$47,$83,$FC,$0B,$E6,$64,$7F,$02
	.dc.b	$6B,$F9,$E6,$BF,$C6,$B9,$7F,$4F,$18,$8F,$F6,$37,$7B,$7F,$E5,$04
	.dc.b	$8F,$E5,$0F,$F1,$0F,$94,$91,$FC,$41,$AF,$F3,$17,$F9,$81,$1F,$E6
	.dc.b	$0D,$74,$60,$87,$F3,$9F,$89,$FE,$70,$47,$F9,$C3,$FC,$83,$E4,$64
	.dc.b	$7F,$20,$6B,$FD,$07,$CF,$FE,$80,$47,$FA,$03,$FC,$A2,$64,$7F,$28
	.dc.b	$6B,$FC,$11,$AF,$C7,$E9,$46,$23,$FD,$8F,$9F,$F1,$08,$C4,$7F,$BB
	.dc.b	$27,$6F,$53,$FC,$C5,$FA,$DF,$FA,$C7,$C8,$34,$91,$FC,$F3,$E3,$63
	.dc.b	$FC,$F1,$AF,$EA,$7F,$46,$FE,$7C,$FE,$A1,$AF,$FB,$6B,$9D,$18,$8F
	.dc.b	$F1,$AB,$63,$3E,$17,$F4,$F8,$C1,$1F,$E9,$BA,$4A,$FE,$9A,$D7,$F8
	.dc.b	$EF,$8E,$F8,$57,$D5,$FF,$B4,$23,$0D,$8C,$CF,$F1,$D9,$CB,$63,$2F
	.dc.b	$46,$BF,$DC,$23,$6E,$67,$FB,$37,$EF,$7F,$C2,$63,$E0,$35,$FF,$29
	.dc.b	$B7,$23,$49,$D9,$C8,$93,$E3,$F3,$FD,$04,$70,$43,$CC,$E5,$A7,$F7
	.dc.b	$79,$7D,$37,$98,$91,$97,$FD,$E5,$F4,$9F,$E9,$3C,$92,$F8,$47,$CB
	.dc.b	$7F,$84,$2B,$53,$98,$C7,$A7,$FE,$1C,$AD,$CF,$F9,$DF,$F4,$47,$1F
	.dc.b	$F8,$62,$73,$2C,$39,$C6,$1B,$87,$4C,$7C,$73,$D3,$95,$8F,$4C,$33
	.dc.b	$7B,$7F,$AC,$61,$E6,$7F,$B0,$B5,$D1,$8F,$4B,$FC,$E2,$DF,$67,$FB
	.dc.b	$03,$E5,$8C,$BF,$05,$AF,$C6,$94,$7F,$9E,$B1,$F5,$36,$FE,$5E,$2F
	.dc.b	$D2,$9F,$34,$F8,$9F,$4E,$5C,$F0,$E0,$DC,$F2,$F8,$DD,$E1,$CE,$F9
	.dc.b	$4F,$D3,$38,$17,$93,$F8,$3F,$8F,$A8,$FE,$BB,$DB,$FD,$33,$2F,$F1
	.dc.b	$E3,$63,$FB,$23,$5F,$D0,$7E,$92,$F7,$39,$47,$E1,$B8,$93,$9B,$FD
	.dc.b	$63,$67,$F6,$AD,$79,$D7,$27,$FC,$CB,$19,$7A,$35,$CC,$6F,$1F,$E4
	.dc.b	$3F,$35,$71,$34,$98,$7B,$1B,$7E,$C6,$BE,$CF,$94,$F1,$65,$EA,$6E
	.dc.b	$46,$92,$FD,$AD,$AF,$6B,$EB,$F0,$BF,$C8,$3E,$85,$F0,$FF,$C5,$F7
	.dc.b	$FE,$64,$8C,$1E,$BF,$0B,$F4,$B1,$FE,$6D,$6E,$1F,$CB,$7A,$74,$49
	.dc.b	$D3,$63,$51,$FE,$37,$C1,$FA,$4F,$EE,$A3,$0D,$9F,$F2,$1C,$E9,$F4
	.dc.b	$46,$5F,$F4,$6F,$1F,$F4,$4C,$FE,$B3,$6F,$FA,$4C,$63,$C4,$F8,$D1
	.dc.b	$DB,$97,$99,$1C,$12,$FC,$38,$27,$51,$86,$BF,$E0,$58,$D4,$F1,$8C
	.dc.b	$4F,$93,$F5,$DF,$C3,$62,$69,$39,$3D,$B9,$4C,$3E,$1B,$70,$FF,$32
	.dc.b	$DF,$1F,$8B,$E8,$4A,$FF,$31,$46,$5B,$83,$E5,$BC,$44,$8C,$B6,$FA
	.dc.b	$CE,$E7,$24,$8F,$87,$F7,$3F,$48,$91,$F6,$1F,$77,$CB,$C8,$F0,$B6
	.dc.b	$7E,$5F,$CB,$FD,$A2,$4F,$E0,$1B,$1A,$8C,$11,$FF,$31,$7E,$D1,$FC
	.dc.b	$6A,$3F,$D7,$CD,$8C,$F6,$FE,$C6,$31,$1F,$E6,$66,$BF,$DF,$DA,$BA
	.dc.b	$D1,$86,$CF,$F0,$1A,$7C,$CB,$C7,$C3,$63,$FD,$24,$6A,$32,$7F,$09
	.dc.b	$C1,$F8,$8B,$EC,$70,$DA,$FE,$76,$D9,$FF,$4E,$35,$FF,$52,$B5,$FF
	.dc.b	$76,$B5,$FE,$C6,$D7,$BC,$FF,$13,$F7,$7F,$D7,$D5,$FE,$E5,$6C,$FF
	.dc.b	$A3,$5A,$FF,$A3,$1A,$FF,$89,$5A,$FC,$76,$E4,$69,$AE,$8F,$DE,$F2
	.dc.b	$FE,$E0,$97,$FB,$E6,$DF,$EB,$9A,$FE,$AD,$FA,$1C,$5F,$C5,$23,$FD
	.dc.b	$8B,$3F,$9B,$EB,$C9,$1F,$EC,$63,$CF,$FA,$0D,$32,$E0,$8E,$09,$3F
	.dc.b	$6A,$3F,$87,$F3,$49,$FE,$2A,$D9,$FE,$0A,$D7,$F8,$13,$5F,$F2,$FF
	.dc.b	$4B,$E0,$3F,$8C,$92,$BF,$8F,$B6,$7F,$87,$B5,$FE,$18,$FF,$4B,$FB
	.dc.b	$2F,$F2,$3C,$FF,$B7,$F6,$46,$2B,$FD,$8C,$D7,$FD,$8A,$D7,$FC,$4E
	.dc.b	$3F,$43,$FD,$0B,$79,$3F,$2D,$18,$23,$82,$5F,$F8,$42,$3F,$53,$BF
	.dc.b	$F3,$AD,$BF,$9A,$92,$7F,$9D,$B6,$FF,$95,$B5,$FE,$50,$FF,$3C,$E3
	.dc.b	$93,$FD,$09,$B3,$FC,$C9,$AF,$F2,$F6,$BF,$E7,$9F,$C0,$DF,$29,$FD
	.dc.b	$C5,$B5,$F1,$DB,$3F,$EE,$B6,$BF,$10,$7C,$83,$D2,$FD,$C3,$6B,$FC
	.dc.b	$5E,$D9,$E0,$70,$9D,$92,$76,$D8,$FE,$73,$F8,$B3,$E8,$3F,$AB,$B6
	.dc.b	$BF,$9C,$B6,$77,$3F,$C2,$7C,$87,$A5,$49,$7F,$E8,$97,$9F,$04,$8E
	.dc.b	$6D,$BF,$0C,$F8,$09,$3E,$07,$B4,$C8,$FF,$4A,$36,$7F,$D2,$4D,$7F
	.dc.b	$D3,$6D,$7F,$CE,$7E,$23,$E9,$5E,$C7,$F8,$A6,$30,$D8,$CD,$FC,$A9
	.dc.b	$CB,$FE,$5C,$FB,$FF,$E3,$44,$EE,$FE,$54,$D8,$D4,$60,$8F,$D4,$F9
	.dc.b	$77,$D1,$7F,$25,$6C,$FF,$8E,$DB,$3B,$9E,$FE,$2F,$62,$47,$8D,$B9
	.dc.b	$FE,$03,$04,$62,$3F,$15,$18,$9F,$F0,$F7,$3F,$F9,$A5,$A7,$F8,$C5
	.dc.b	$C4,$FF,$87,$9A,$FF,$87,$3E,$73,$F2,$AF,$F9,$D1,$AB,$C9,$46,$1B
	.dc.b	$CF,$FE,$22,$8C,$11,$C1,$E1,$F8,$E9,$87,$99,$B7,$FE,$0E,$FB,$DF
	.dc.b	$C7,$CF,$F8,$31,$B1,$9E,$3F,$F1,$7C,$62,$3F,$C6,$6D,$7F,$D1,$6D
	.dc.b	$7D,$0D,$C4,$E7,$2F,$F4,$77,$6F,$F9,$57,$0F,$E3,$A7,$F9,$1C,$B7
	.dc.b	$FE,$44,$6B,$FE,$42,$F8,$A7,$D4,$FF,$90,$9B,$3F,$E2,$46,$CF,$F8
	.dc.b	$A9,$AF,$19,$E0,$49,$E0,$FE,$1F,$E4,$3F,$2F,$3F,$C1,$F8,$91,$88
	.dc.b	$FE,$5B,$18,$E6,$E0,$78,$FE,$B2,$61,$E3,$6D,$7F,$8F,$3E,$89,$F6
	.dc.b	$FF,$E5,$A6,$D7,$F8,$F1,$B3,$B9,$D1,$27,$4D,$8F,$F3,$4F,$5F,$F9
	.dc.b	$9D,$FF,$73,$EC,$8C,$42,$30,$FF,$A1,$63,$31,$8A,$FF,$3C,$46,$09
	.dc.b	$3F,$E1,$D3,$FD,$C2,$3E,$89,$E4,$7F,$2D,$EB,$FF,$60,$C6,$27,$F9
	.dc.b	$87,$BF,$FE,$FA,$8C,$47,$F0,$AC,$31,$7F,$CB,$AA,$CF,$F5,$AB,$E7
	.dc.b	$39,$8F,$03,$DE,$B7,$1F,$EF,$B3,$27,$17,$2F,$A2,$FC,$2F,$F9,$E8
	.dc.b	$D9,$C5,$CD,$78,$23,$D2,$68,$E1,$1C,$95,$F6,$9E,$11,$CE,$FC,$CF
	.dc.b	$13,$67,$FB,$F0,$A7,$FB,$FA,$70,$E0,$FE,$F7,$F3,$DF,$E1,$FF,$A3
	.dc.b	$FE,$A7,$08,$C3,$FD,$59,$AF,$F5,$F6,$BC,$BE,$8D,$C3,$CA,$51,$89
	.dc.b	$E4,$D7,$93,$5E,$53,$F4,$27,$F8,$AC,$5E,$BF,$F0,$8E,$7F,$7A,$7A
	.dc.b	$D6,$EF,$B4,$E6,$B7,$97,$B9,$B8,$0F,$75,$70,$9F,$0C,$7D,$8C,$3C
	.dc.b	$2D,$AE,$2F,$0C,$3C,$07,$95,$87,$1B,$E8,$F7,$AF,$3B,$CC,$DE,$1E
	.dc.b	$B9,$DE,$DE,$57,$E9,$B6,$7F,$DC,$07,$CC,$2C,$69,$C1,$FE,$E0,$29
	.dc.b	$FE,$38,$8E,$4E,$63,$EF,$F4,$3E,$8F,$FC,$9D,$3F,$EC,$26,$BF,$EC
	.dc.b	$86,$BF,$EE,$86,$BF,$2E,$39,$79,$9D,$27,$33,$FD,$3D,$B3,$A8,$FF
	.dc.b	$67,$FD,$1F,$F8,$C1,$BD,$DE,$48,$C4,$7F,$B6,$1A,$FF,$26,$FA,$3F
	.dc.b	$F0,$DB,$FD,$DD,$AF,$F2,$D6,$BB,$EE,$F7,$EA,$EF,$A4,$9F,$F7,$0C
	.dc.b	$65,$B1,$97,$FD,$D3,$19,$6C,$7D,$6F,$25,$F3,$B4,$93,$FE,$3E,$DF
	.dc.b	$FC,$8B,$1F,$43,$6F,$F0,$EC,$FF,$BC,$1A,$FF,$BD,$9A,$FF,$9C,$35
	.dc.b	$FF,$41,$76,$71,$24,$71,$6B,$FE,$53,$BF,$F6,$A3,$F8,$AF,$3F,$F8
	.dc.b	$24,$62,$3C,$91,$FC,$1E,$9F,$46,$F6,$7F,$0A,$DA,$FF,$2B,$CE,$A7
	.dc.b	$24,$8C,$CE,$DF,$F3,$24,$ED,$CC,$F8,$FA,$A4,$78,$9A,$FF,$99,$3E
	.dc.b	$85,$F4,$7F,$C6,$DB,$1F,$C5,$9A,$EE,$34,$47,$4D,$7F,$CD,$FF,$46
	.dc.b	$F6,$3F,$93,$36,$3F,$90,$B5,$DC,$78,$7E,$6A,$47,$85,$AF,$9D,$AF
	.dc.b	$FA,$0E,$D1,$88,$FF,$54,$D7,$FC,$01,$7B,$FA,$9E,$22,$3E,$DE,$5F
	.dc.b	$80,$47,$04,$3E,$33,$96,$9F,$5B,$E8,$FE,$2B,$F8,$96,$BF,$B7,$6B
	.dc.b	$FE,$43,$6E,$A5,$F4,$7F,$C1,$9B,$3D,$63,$07,$D5,$F4,$A4,$7D,$56
	.dc.b	$BF,$82,$F9,$7F,$F0,$2B,$7C,$6F,$E2,$31,$88,$FC,$F3,$5F,$F6,$7F
	.dc.b	$D1,$FE,$B1,$FE,$AA,$D7,$FD,$52,$D7,$FD,$64,$D7,$8F,$47,$F6,$38
	.dc.b	$C4,$7F,$41,$6A,$73,$CA,$FA,$3F,$DF,$4F,$FB,$82,$31,$D7,$FD,$99
	.dc.b	$B1,$FE,$EB,$9D,$BF,$E2,$6F,$89,$7B,$DF,$CE,$9B,$1F,$C0,$DA,$FE
	.dc.b	$DD,$AF,$F3,$76,$BC,$E3,$91,$1E,$5F,$3D,$D4,$8E,$08,$75,$72,$D3
	.dc.b	$FB,$FF,$9F,$FC,$D9,$FF,$29,$F7,$7F,$82,$63,$04,$70,$48,$E2,$FC
	.dc.b	$3F,$A3,$FE,$9F,$1F,$DB,$72,$46,$21,$18,$F1,$B8,$1F,$17,$E8,$E4
	.dc.b	$FC,$56,$BF,$15,$AF,$C5,$6B,$F3,$FE,$0F,$E7,$11,$F9,$CD,$7E,$73
	.dc.b	$5F,$9C,$FF,$89,$FE,$8D,$F0,$94,$93,$FE,$24,$8C,$B6,$32,$FF,$BE
	.dc.b	$A3,$2D,$8F,$F1,$97,$AC,$FA,$3A,$49,$FF,$17,$C6,$5F,$C2,$7C,$F7
	.dc.b	$E1,$C8,$E0,$87,$E1,$DC,$B4,$FF,$0E,$1F,$97,$FA,$37,$C1,$FE,$5D
	.dc.b	$B1,$A3,$EB,$38,$71,$AA,$31,$1C,$E7,$91,$1E,$47,$DC,$8D,$BE,$2F
	.dc.b	$94,$FE,$42,$FF,$93,$1A,$F0,$6B,$E3,$6B,$C7,$C2,$FF,$2D,$46,$23
	.dc.b	$FC,$70,$D4,$DF,$A3,$72,$7F,$98,$5B,$3F,$E5,$C6,$BE,$86,$BF,$BD
	.dc.b	$FA,$37,$D1,$FE,$F5,$B1,$EE,$CD,$18,$8F,$8E,$D7,$94,$F1,$23,$C6
	.dc.b	$34,$FF,$05,$FA,$3F,$E0,$84,$7F,$82,$35,$FE,$08,$D7,$F8,$23,$FC
	.dc.b	$A3,$59,$1F,$DD,$3B,$91,$88,$FE,$19,$F4,$7F,$C4,$48,$FF,$11,$7F
	.dc.b	$98,$91,$E4,$D7,$F9,$C3,$5F,$E6,$AD,$7F,$C8,$12,$F9,$9F,$E7,$64
	.dc.b	$8F,$23,$FC,$87,$E8,$E4,$7F,$21,$6B,$FD,$17,$D1,$FE,$86,$47,$D8
	.dc.b	$FF,$29,$99,$1F,$CA,$5A,$FF,$29,$6B,$FC,$A5,$AF,$F3,$36,$BF,$ED
	.dc.b	$7E,$EF,$43,$CC,$FF,$0E,$FC,$B7,$F8,$8C,$8F,$F9,$99,$FF,$14,$C3
	.dc.b	$E5,$FF,$9E,$12,$7F,$9E,$3F,$D1,$7E,$63,$F9,$CC,$7F,$20,$6F,$01
	.dc.b	$A4,$F0,$78,$12,$7F,$21,$1C,$BD,$46,$1F,$E4,$49,$FE,$9D,$9F,$F4
	.dc.b	$D2,$3F,$B4,$F1,$DF,$46,$FA,$3F,$F5,$EC,$61,$B1,$99,$E0,$D7,$F8
	.dc.b	$34,$ED,$CC,$F0,$23,$FA,$5F,$A3,$FE,$02,$FF,$3E,$6E,$86,$93,$B3
	.dc.b	$FB,$53,$67,$94,$79,$3D,$09,$1E,$46,$BE,$4C,$7F,$C5,$4D,$8C,$BF
	.dc.b	$C8,$B1,$F8,$31,$DB,$E8,$E4,$F9,$DA,$FF,$1A,$FA,$27,$A4,$BF,$F6
	.dc.b	$37,$C2,$BE,$8B,$F9,$6D,$6A,$73,$18,$AE,$17,$FE,$F8,$8F,$F5,$27
	.dc.b	$D1,$FE,$1E,$36,$7F,$B4,$6F,$93,$B3,$15,$CB,$0F,$04,$63,$0E,$3E
	.dc.b	$AB,$0E,$93,$C4,$61,$D1,$FE,$7C,$D7,$CD,$5F,$EF,$E8,$C4,$78,$FD
	.dc.b	$37,$B5,$96,$C7,$C5,$6B,$D1,$AF,$C4,$77,$8F,$A2,$3C,$71,$8C,$FF
	.dc.b	$D8,$DF,$46,$FA,$37,$C5,$7C,$4B,$DE,$37,$26,$3C,$67,$CD,$3F,$0B
	.dc.b	$0E,$57,$BA,$D7,$27,$F1,$3F,$1F,$AE,$1F,$D7,$3B,$BF,$A3,$DF,$F3
	.dc.b	$B6,$C6,$67,$6F,$E2,$7E,$8B,$FD,$0D,$5F,$AB,$6C,$7E,$91,$AE,$FD
	.dc.b	$DE,$28,$FC,$54,$69,$C7,$02,$31,$1A,$9E,$2D,$AF,$37,$D1,$E3,$2F
	.dc.b	$E5,$CD,$C8,$D2,$4F,$02,$57,$C1,$8E,$B8,$F8,$7E,$8E,$E4,$69,$2F
	.dc.b	$E0,$36,$FE,$03,$5F,$F2,$B7,$A1,$E8,$EF,$FE,$2A,$DE,$FF,$F1,$5D
	.dc.b	$FF,$97,$E3,$D5,$E0,$7F,$29,$8F,$E3,$BE,$16,$08,$E0,$98,$7F,$85
	.dc.b	$27,$83,$67,$C1,$F3,$DF,$3E,$BF,$DD,$CD,$9F,$F7,$13,$5F,$F1,$2E
	.dc.b	$28,$C4,$7D,$7E,$A7,$89,$F0,$EF,$FA,$1E,$9F,$52,$72,$D9,$D5,$76
	.dc.b	$E6,$F1,$92,$FC,$6D,$8D,$46,$08,$E0,$87,$9B,$0F,$D9,$C6,$A3,$FC
	.dc.b	$7B,$83,$E8,$FF,$12,$DE,$4F,$34,$60,$FC,$6E,$7E,$FC,$FE,$39,$30
	.dc.b	$F8,$EF,$0F,$8E,$91,$FE,$FE,$A4,$63,$1F,$F7,$DC,$3E,$7B,$B9,$2B
	.dc.b	$83,$67,$6D,$7D,$33,$C6,$7E,$BF,$D1,$FD,$A2,$38,$3C,$5F,$33,$EC
	.dc.b	$23,$83,$EF,$FC,$3F,$E0,$88,$E0,$97,$F8,$28,$CB,$63,$FC,$79,$F4
	.dc.b	$7F,$C3,$9B,$D9,$FA,$F8,$C4,$FF,$99,$1A,$FF,$B0,$DA,$FF,$A7,$DA
	.dc.b	$FF,$A7,$9B,$D6,$73,$4F,$A3,$78,$DF,$C0,$23,$0D,$8C,$C7,$F9,$B8
	.dc.b	$C9,$FE,$70,$B7,$D1,$BE,$7B,$F8,$4C,$61,$B1,$99,$E0,$D9,$FF,$3C
	.dc.b	$B5,$FF,$42,$35,$FF,$45,$B5,$FE,$38,$DC,$5E,$EB,$DE,$FE,$2C,$DB
	.dc.b	$FF,$34,$36,$FF,$D1,$AD,$7F,$DD,$8D,$7F,$D9,$0D,$7F,$D8,$ED,$AE
	.dc.b	$F5,$FC,$03,$96,$5F,$F0,$06,$DF,$F0,$06,$BF,$AB,$6B,$FC,$1F,$8B
	.dc.b	$F8,$AB,$FE,$A3,$9F,$CE,$76,$7D,$39,$CC,$FF,$AB,$27,$27,$D1,$3F
	.dc.b	$DF,$A5,$F4,$3E,$8B,$88,$FB,$11,$82,$46,$C9,$1F,$C5,$B9,$7F,$8D
	.dc.b	$92,$3F,$8D,$B6,$35,$18,$23,$FC,$7B,$E0,$7F,$91,$92,$3F,$91,$B6
	.dc.b	$35,$18,$23,$FE,$C0,$F9,$BF,$F2,$34,$7F,$A7,$DB,$19,$F4,$7F,$A9
	.dc.b	$46,$23,$FD,$95,$F4,$7F,$A0,$4D,$27,$2F,$E5,$1B,$7F,$99,$6C,$FF
	.dc.b	$36,$CF,$F9,$D9,$27,$F9,$DB,$63,$51,$82,$3F,$CF,$B9,$3F,$A1,$92
	.dc.b	$3F,$A1,$B6,$35,$18,$23,$FE,$E0,$FA,$37,$CA,$7F,$6F,$6C,$FF,$B0
	.dc.b	$5B,$3B,$9F,$D7,$FC,$8F,$FB,$BA,$BF,$68,$D9,$FF,$72,$B5,$FF,$78
	.dc.b	$35,$FF,$7B,$B5,$FF,$04,$35,$34,$D7,$7B,$E8,$DF,$47,$FD,$55,$B7
	.dc.b	$FE,$F4,$9D,$4E,$4F,$E0,$BE,$4F,$F9,$7A,$7F,$D0,$7E,$9E,$08,$F1
	.dc.b	$BF,$E3,$77,$DF,$7E,$C4,$9D,$57,$84,$8E,$0F,$F4,$7C,$78,$08,$E0
	.dc.b	$99,$7F,$A5,$E3,$2D,$8F,$F1,$53,$3C,$57,$D1,$BD,$9F,$F1,$3C,$61
	.dc.b	$B1,$9A,$E3,$5F,$E3,$3A,$D4,$7F,$29,$FA,$2F,$E3,$92,$69,$3B,$FF
	.dc.b	$95,$B6,$78,$46,$0F,$F1,$BF,$CD,$BE,$8F,$F9,$33,$63,$31,$83,$AF
	.dc.b	$8B,$D0,$95,$E3,$6E,$1E,$83,$C8,$E1,$F5,$E0,$8F,$F1,$87,$93,$FE
	.dc.b	$6A,$AE,$08,$C5,$7C,$27,$11,$C0,$FF,$0E,$7C,$EF,$E7,$27,$FC,$1A
	.dc.b	$D7,$FC,$48,$D5,$C0,$8C,$37,$97,$99,$2B,$F2,$53,$C6,$FC,$9F,$21
	.dc.b	$E8,$24,$FA,$21,$1F,$E2,$FF,$A3,$FC,$B6,$3F,$6B,$14,$62,$3F,$C4
	.dc.b	$6D,$7F,$C7,$4D,$7C,$ED,$8E,$3C,$7D,$49,$87,$C8,$1D,$9F,$E5,$B6
	.dc.b	$99,$70,$7F,$90,$FE,$2D,$F4,$74,$93,$FB,$18,$CB,$63,$38,$FA,$E7
	.dc.b	$2D,$8F,$C5,$FC,$8F,$E6,$63,$F3,$3C,$68,$C4,$7E,$92,$31,$CD,$C0
	.dc.b	$F2,$7D,$1C,$C7,$C4,$DA,$F9,$5F,$46,$FA,$3F,$F2,$E3,$6B,$F8,$43
	.dc.b	$67,$73,$A2,$4E,$9B,$1F,$8D,$EC,$FF,$35,$BF,$91,$6F,$17,$FB,$92
	.dc.b	$BF,$2B,$18,$24,$FF,$A2,$5B,$1F,$E1,$D8,$FF,$78,$8F,$A2,$FF,$29
	.dc.b	$BF,$C0,$E3,$FD,$5D,$18,$8F,$E0,$FF,$03,$8F,$1F,$F8,$0F,$67,$FB
	.dc.b	$0A,$B5,$5D,$1B,$19,$9F,$D8,$57,$1A,$EC,$78,$63,$2D,$33,$5D,$14
	.dc.b	$F1,$7F,$88,$5D,$7C,$15,$E4,$8C,$47,$03,$A1,$C0,$A7,$03,$BC,$78
	.dc.b	$AB,$D3,$3C,$AB,$E4,$57,$33,$FC,$8C,$E0,$FF,$24,$7C,$FB,$C9,$7D
	.dc.b	$1E,$1B,$5F,$C4,$67,$53,$93,$8F,$CB,$38,$A9,$2B,$84,$65,$B1,$FE
	.dc.b	$E3,$F7,$3E,$B5,$FF,$09,$9C,$69,$FC,$FA,$31,$09,$DD,$F3,$6D,$F0
	.dc.b	$9F,$F7,$75,$70,$9E,$BD,$9F,$D2,$1D,$CF,$3C,$3B,$61,$E5,$D7,$73
	.dc.b	$F5,$F0,$F0,$C7,$03,$8D,$EA,$79,$1E,$0C,$3B,$61,$D3,$1F,$24,$B0
	.dc.b	$63,$E9,$AF,$75,$F0,$AE,$47,$0A,$DD,$FB,$78,$13,$F5,$CB,$0D,$B3
	.dc.b	$3A,$D7,$F8,$24,$62,$B8,$1F,$5D,$C6,$5A,$8F,$E8,$C6,$AF,$A3,$8A
	.dc.b	$D5,$F5,$8F,$0D,$F2,$69,$FB,$FF,$90,$DF,$09,$A7,$7E,$B7,$D1,$BC
	.dc.b	$EC,$DF,$2C,$58,$71,$8C,$35,$F5,$1F,$D4,$AB,$93,$E2,$CD,$F4,$4E
	.dc.b	$EF,$F5,$34,$61,$B1,$99,$FE,$DD,$39,$7F,$B8,$7A,$6F,$BB,$FE,$DD
	.dc.b	$1E,$E8,$C7,$89,$F2,$5B,$1E,$1B,$E8,$F9,$B3,$FF,$1C,$FC,$EF,$FA
	.dc.b	$DD,$B1,$FD,$DB,$CD,$4E,$5B,$19,$8C,$11,$C1,$FD,$5F,$E2,$FE,$AF
	.dc.b	$11,$A7,$D8,$AE,$73,$E3,$C7,$F9,$66,$06,$92,$3C,$CD,$7F,$C1,$6D
	.dc.b	$7F,$C1,$50,$F3,$3F,$0A,$D8,$FB,$4D,$7C,$07,$E2,$A5,$DB,$D8,$91
	.dc.b	$D9,$AF,$F8,$5E,$36,$E4,$87,$8B,$7F,$09,$1F,$C5,$47,$07,$F0,$E3
	.dc.b	$E8,$1E,$B7,$F4,$D6,$C7,$FB,$AD,$AF,$F8,$C5,$AF,$F9,$29,$AF,$F9
	.dc.b	$51,$AF,$ED,$9A,$F7,$8F,$F7,$B4,$BE,$B5,$F7,$1F,$89,$6D,$7E,$35
	.dc.b	$AF,$A2,$FC,$7C,$7F,$B4,$8F,$57,$A9,$E9,$23,$E9,$6B,$E9,$6B,$E9
	.dc.b	$7F,$25,$ED,$BE,$1F,$F7,$AD,$9F,$D9,$35,$F1,$4F,$EE,$7E,$89,$F4
	.dc.b	$9F,$E5,$59,$FF,$21,$C6,$1F,$F4,$83,$5F,$F1,$C3,$5F,$F3,$DF,$62
	.dc.b	$31,$1F,$C7,$A3,$13,$D5,$B1,$D2,$3F,$55,$F0,$BF,$B4,$8F,$D7,$37
	.dc.b	$2F,$E3,$3C,$48,$C7,$37,$5C,$7C,$95,$F2,$FE,$93,$F2,$CD,$E5,$FE
	.dc.b	$49,$8A,$73,$7C,$7E,$93,$F8,$D7,$E1,$6F,$A8,$F3,$36,$FF,$D6,$0D
	.dc.b	$7F,$CB,$0D,$7F,$D3,$7A,$79,$7E,$03,$F9,$6E,$A6,$9F,$C9,$FE,$67
	.dc.b	$F9,$1D,$7F,$11,$FA,$5F,$F3,$2B,$FC,$59,$AF,$FA,$7B,$A1,$18,$8F
	.dc.b	$F5,$4C,$63,$D8,$FC,$9C,$E6,$7F,$A6,$CE,$5E,$93,$A3,$E7,$FD,$77
	.dc.b	$98,$91,$E3,$8F,$4F,$81,$23,$F9,$D6,$5F,$C4,$38,$BE,$9F,$9B,$FD
	.dc.b	$37,$87,$FA,$BC,$E5,$B7,$C6,$FF,$94,$56,$DF,$F4,$D5,$7C,$BB,$FB
	.dc.b	$AD,$BE,$7C,$1F,$D5,$AD,$F4,$3F,$32,$26,$1E,$47,$51,$FC,$E7,$E9
	.dc.b	$BF,$BA,$CF,$F3,$D6,$BF,$D3,$B5,$46,$32,$FD,$97,$AF,$FD,$42,$B5
	.dc.b	$39,$CB,$E5,$B6,$3F,$97,$35,$FF,$3D,$79,$5F,$80,$F7,$E4,$7C,$87
	.dc.b	$C9,$E7,$7E,$13,$D9,$93,$F3,$1F,$77,$C3,$E2,$3F,$43,$17,$3F,$9F
	.dc.b	$3F,$8C,$F3,$E5,$7D,$47,$E3,$7C,$46,$23,$F5,$31,$6B,$EB,$CF,$E8
	.dc.b	$BD,$C7,$D2,$FF,$9C,$DB,$7F,$C0,$5A,$FF,$B5,$DA,$FF,$BC,$1A,$FF
	.dc.b	$BE,$9A,$FF,$87,$1A,$9C,$F5,$FE,$E7,$EC,$7F,$4A,$49,$FF,$07,$9B
	.dc.b	$AF,$D4,$FB,$B2,$3F,$22,$FE,$0F,$E8,$71,$1F,$C1,$A2,$E5,$F9,$49
	.dc.b	$FD,$8F,$C1,$4A,$FC,$CB,$FD,$F7,$E9,$9F,$3F,$99,$FE,$1F,$36,$FF
	.dc.b	$C4,$FC,$EB,$DD,$79,$BE,$97,$F2,$6F,$F2,$76,$BF,$EF,$D6,$BF,$E3
	.dc.b	$E3,$FC,$7E,$E1,$C3,$5C,$F4,$FF,$A9,$63,$10,$75,$E5,$FF,$91,$7D
	.dc.b	$9F,$F1,$F3,$67,$FC,$6E,$D7,$FB,$2C,$F9,$31,$FD,$6F,$9F,$FC,$50
	.dc.b	$93,$FE,$21,$FA,$5F,$F1,$94,$EA,$3F,$7F,$3A,$7F,$51,$3A,$EC,$FF
	.dc.b	$2A,$F8,$5F,$D6,$6F,$F9,$AB,$63,$27,$EE,$7E,$97,$FC,$9F,$19,$B7
	.dc.b	$2A,$30,$DA,$E1,$3F,$E4,$9F,$05,$F4,$BF,$E2,$F6,$C7,$CD,$3F,$CC
	.dc.b	$EE,$1C,$57,$FA,$3E,$7F,$A8,$78,$F9,$CB,$FC,$C7,$E7,$A3,$18,$FF
	.dc.b	$57,$6E,$1E,$6C,$6E,$3E,$7C,$BD,$64,$9F,$F3,$3E,$1C,$F2,$C4,$7F
	.dc.b	$91,$E7,$F9,$DE,$5F,$19,$FE,$85,$3F,$C7,$7E,$97,$D8,$4A,$F6,$1E
	.dc.b	$17,$FC,$9D,$2F,$8E,$F9,$64,$9F,$8D,$3F,$C9,$FE,$77,$E3,$12,$39
	.dc.b	$E5,$FE,$B1,$8E,$7E,$87,$D4,$24,$FC,$F8,$AF,$A6,$2E,$5F,$D0,$26
	.dc.b	$ED,$F8,$29,$B5,$F1,$E7,$F9,$C7,$CD,$FD,$72,$47,$8B,$64,$7B,$3D
	.dc.b	$9F,$C2,$92,$7F,$01,$1F,$D0,$7E,$2B,$F0,$04,$8E,$4D,$7E,$FD,$23
	.dc.b	$11,$FD,$D2,$30,$FC,$2D,$9F,$43,$F9,$42,$47,$E3,$EB,$FD,$05,$F4
	.dc.b	$BF,$C0,$23,$FB,$93,$5F,$EE,$2D,$7F,$B7,$B5,$FF,$39,$37,$B0,$D2
	.dc.b	$77,$7D,$CF,$BB,$FD,$29,$2B,$FD,$29,$3E,$1E,$0F,$D6,$CE,$67,$F1
	.dc.b	$8F,$9A,$BC,$FF,$01,$FE,$54,$BF,$F4,$45,$7F,$A8,$1A,$FA,$A3,$44
	.dc.b	$74,$D7,$F5,$39,$BE,$E7,$F7,$24,$8F,$2C,$FF,$8B,$3E,$93,$FD,$69
	.dc.b	$DF,$FE,$DC,$F8,$D7,$D2,$FF,$21,$9C,$DA,$72,$FF,$B8,$A3,$31,$8E
	.dc.b	$B7,$D3,$7E,$1A,$BA,$5F,$3E,$5F,$F1,$4D,$F2,$96,$1C,$AB,$58,$F0
	.dc.b	$96,$3D,$39,$7F,$9A,$61,$C7,$0F,$2D,$7F,$0B,$6B,$C8,$FD,$9B,$87
	.dc.b	$D1,$CF,$FE,$93,$7D,$CD,$73,$F0,$9F,$E6,$B9,$FF,$56,$36,$7F,$D4
	.dc.b	$35,$E1,$6E,$5F,$E6,$7F,$A6,$FC,$AF,$57,$A7,$1F,$6D,$69,$EA,$7C
	.dc.b	$BA,$F0,$E1,$EF,$CD,$87,$2A,$D1,$C3,$1F,$A3,$D7,$FC,$C7,$DF,$67
	.dc.b	$FE,$F7,$E3,$FF,$3D,$3F,$EE,$5E,$C7,$D3,$6B,$1F,$E3,$97,$7A,$7F
	.dc.b	$B3,$D6,$AF,$FD,$DE,$DA,$FF,$2D,$35,$FF,$2C,$B5,$FF,$2B,$B5,$FF
	.dc.b	$6D,$37,$A8,$D3,$59,$7F,$49,$FB,$BF,$15,$FF,$54,$6D,$F2,$F8,$4B
	.dc.b	$5B,$AF,$75,$63,$43,$C3,$1E,$F7,$98,$C6,$3F,$EA,$DB,$F6,$CF,$FA
	.dc.b	$E7,$1F,$4B,$5F,$04,$68,$8F,$C2,$8D,$5F,$69,$FE,$29,$F2,$BE,$33
	.dc.b	$CF,$E0,$3E,$7B,$13,$FD,$8D,$ED,$A2,$75,$3F,$CB,$67,$31,$8C,$7F
	.dc.b	$92,$C6,$09,$3B,$3F,$92,$7A,$1F,$0C,$F8,$FE,$EA,$B1,$3F,$E8,$27
	.dc.b	$C3,$19,$8F,$E9,$AD,$8F,$EE,$2D,$7F,$D4,$BB,$3E,$57,$E8,$1D,$B9
	.dc.b	$AD,$FE,$46,$2D,$7F,$B0,$A6,$F0,$FF,$83,$A6,$D7,$FA,$42,$6D,$7F
	.dc.b	$A3,$1F,$29,$23,$CA,$FC,$8B,$FF,$7B,$C7,$E7,$FE,$27,$DF,$F4,$9F
	.dc.b	$83,$6C,$7B,$23,$DE,$47,$FA,$57,$D4,$FE,$24,$FC,$7F,$4B,$E9,$5F
	.dc.b	$41,$9A,$FD,$65,$5C,$DA,$FF,$84,$AF,$75,$A3,$FC,$16,$E2,$77,$1F
	.dc.b	$EE,$B8,$C4,$FF,$AA,$3E,$97,$C3,$18,$9F,$F5,$9C,$7D,$62,$3F,$59
	.dc.b	$F8,$5E,$27,$E8,$4F,$BF,$F4,$B6,$7F,$81,$45,$C3,$FD,$E7,$37,$0F
	.dc.b	$F1,$FF,$91,$FA,$93,$F5,$1F,$4A,$F0,$B1,$8F,$F0,$B9,$B8,$7F,$AC
	.dc.b	$6B,$53,$93,$FC,$B2,$7F,$A1,$CB,$EE,$7E,$E9,$7D,$2B,$E4,$3F,$7B
	.dc.b	$EC,$7F,$18,$AD,$4E,$6B,$FC,$52,$70,$AF,$DA,$34,$CC,$6F,$0F,$F2
	.dc.b	$54,$65,$B1,$97,$FC,$F5,$19,$6C,$65,$FF,$31,$FD,$CF,$F0,$B3,$F8
	.dc.b	$57,$C0,$3E,$0B,$17,$FC,$B6,$6D,$7F,$B4,$EB,$53,$93,$F9,$97,$DC
	.dc.b	$FF,$15,$3F,$67,$F2,$96,$32,$4A,$FC,$29,$FE,$EC,$8D,$3F,$E5,$1F
	.dc.b	$93,$FE,$3C,$7C,$BE,$7F,$E0,$9C,$5F,$E7,$88,$C7,$81,$F1,$A7,$F8
	.dc.b	$45,$7C,$06,$BF,$BF,$6B,$FD,$21,$B8,$7E,$82,$79,$11,$FF,$18,$7C
	.dc.b	$97,$F2,$F3,$F8,$6F,$CE,$D8,$FE,$A1,$16,$FD,$7E,$23,$C2,$5E,$73
	.dc.b	$FE,$A5,$3F,$E0,$6F,$47,$F9,$EF,$A6,$7F,$8C,$FC,$5E,$19,$9F,$F1
	.dc.b	$BB,$E3,$5B,$8C,$1F,$C7,$EF,$C1,$7C,$32,$F1,$D3,$E7,$BB,$9F,$D1
	.dc.b	$FC,$97,$C9,$E3,$1E,$76,$F2,$D5,$A9,$CC,$62,$BF,$AB,$D3,$E9,$FF
	.dc.b	$4A,$FD,$C9,$D5,$70,$BE,$53,$96,$E9,$FD,$D6,$39,$77,$F9,$71,$F8
	.dc.b	$BF,$6B,$30,$F2,$ED,$F3,$79,$31,$87,$BA,$AE,$3F,$3B,$8B,$F8,$64
	.dc.b	$FF,$AE,$37,$57,$D0,$F5,$66,$5D,$2F,$F5,$98,$7C,$E9,$CE,$1F,$29
	.dc.b	$FE,$9B,$DB,$89,$FA,$7E,$7F,$DA,$3F,$B9,$7D,$43,$E2,$71,$97,$C8
	.dc.b	$8C,$3F,$EA,$6E,$24,$63,$2F,$F6,$24,$62,$BF,$B7,$CE,$67,$FD,$E9
	.dc.b	$39,$AE,$53,$99,$FE,$C1,$39,$79,$C6,$5F,$85,$EA,$FE,$40,$FE,$83
	.dc.b	$F5,$1F,$EB,$5B,$F9,$BB,$FF,$88,$73,$FF,$67,$FE,$0C,$F0,$3F,$C0
	.dc.b	$DE,$6A,$31,$7F,$EC,$08,$C4,$EA,$73,$97,$FB,$13,$AD,$EE,$7E,$21
	.dc.b	$B8,$7F,$B4,$B3,$FC,$4E,$5F,$83,$9C,$DF,$E6,$BC,$6F,$D3,$9E,$8C
	.dc.b	$2E,$1C,$5F,$2E,$1C,$2B,$FC,$04,$7F,$9A,$9C,$3F,$E0,$5C,$D1,$88
	.dc.b	$D4,$F2,$25,$7B,$67,$5C,$FF,$E8,$3F,$3F,$FC,$6B,$EC,$9F,$F1,$CF
	.dc.b	$A6,$FA,$9F,$F5,$E6,$1F,$98,$AD,$3F,$AE,$9D,$7C,$0F,$F8,$C5,$B1
	.dc.b	$FE,$EA,$6E,$7E,$A0,$E1,$F6,$B5,$FF,$10,$35,$DD,$F8,$9E,$51,$9C
	.dc.b	$7F,$D3,$9E,$3F,$A8,$FF,$28,$F0,$3E,$A5,$F2,$99,$CB,$F8,$2C,$E6
	.dc.b	$3F,$C9,$0D,$8F,$F8,$02,$DC,$7D,$0D,$7E,$9B,$5F,$E3,$95,$DA,$78
	.dc.b	$B6,$38,$B5,$FB,$78,$7F,$1B,$F1,$7F,$CC,$BE,$29,$FF,$33,$C7,$FA
	.dc.b	$12,$FF,$83,$CE,$5F,$E1,$31,$9A,$F9,$88,$C3,$D9,$AF,$F2,$D6,$BB
	.dc.b	$69,$C6,$3F,$CB,$AE,$0C,$CF,$F1,$27,$F9,$97,$83,$E6,$F5,$FD,$2D
	.dc.b	$AF,$BD,$7D,$67,$99,$23,$35,$E7,$8C,$4F,$AA,$11,$88,$F8,$FC,$0F
	.dc.b	$98,$FC,$B9,$FE,$8C,$BF,$E7,$A7,$E5,$5C,$78,$DF,$E7,$68,$CD,$7C
	.dc.b	$66,$BF,$49,$AF,$F5,$16,$BF,$0E,$39,$12,$3F,$90,$7E,$2F,$FD,$66
	.dc.b	$7F,$A3,$37,$6F,$F5,$A7,$CC,$7E,$5D,$FF,$5C,$78,$09,$D5,$FF,$63
	.dc.b	$AF,$CE,$5F,$E3,$1A,$E7,$97,$F2,$B1,$88,$FB,$DF,$3E,$F5,$BF,$D5
	.dc.b	$AD,$BF,$EC,$3D,$6F,$A8,$FF,$40,$47,$F5,$09,$D5,$FF,$7E,$6E,$9F
	.dc.b	$DF,$1B,$3F,$C3,$6B,$8D,$6F,$0F,$F5,$DF,$D5,$7F,$86,$AB,$FB,$C3
	.dc.b	$67,$F2,$5D,$FF,$E9,$3C,$7F,$C0,$F3,$9E,$33,$4C,$4E,$EF,$9F,$A1
	.dc.b	$FD,$D3,$0F,$F5,$17,$1F,$FA,$8F,$E0,$3F,$C2,$71,$FC,$22,$B1,$E8
	.dc.b	$7E,$42,$AC,$E4,$FF,$5B,$46,$5A,$66,$B9,$BF,$CB,$A7,$8B,$FC,$4A
	.dc.b	$AB,$E2,$3F,$C2,$55,$FC,$4A,$3F,$D6,$DD,$1C,$A7,$4D,$ED,$FF,$7C
	.dc.b	$39,$85,$F4,$9F,$F7,$E6,$19,$8C,$5F,$82,$B6,$7A,$0F,$21,$4F,$24
	.dc.b	$F7,$8E,$13,$FB,$29,$F6,$14,$E6,$65,$D7,$D3,$BE,$9D,$F5,$38,$6D
	.dc.b	$7E,$F2,$75,$39,$38,$FC,$EF,$FA,$B2,$74,$D7,$FA,$04,$A3,$11,$FD
	.dc.b	$72,$31,$C3,$C8,$F9,$7A,$E7,$EA,$7F,$B0,$ED,$F5,$A9,$18,$9E,$58
	.dc.b	$78,$6B,$4D,$8E,$58,$74,$3F,$DD,$97,$C9,$C0,$7C,$6F,$DE,$37,$3C
	.dc.b	$DE,$37,$E1,$69,$CA,$78,$DF,$8B,$0C,$56,$EB,$B3,$6B,$78,$7A,$AB
	.dc.b	$75,$0F,$11,$E5,$C3,$C5,$C4,$87,$CC,$7D,$F8,$C3,$63,$27,$F3,$02
	.dc.b	$9F,$D2,$0A,$7F,$54,$29,$FE,$73,$29,$FD,$34,$BC,$EF,$A9,$FE,$69
	.dc.b	$1F,$E2,$96,$BF,$DF,$E3,$5D,$DF,$93,$6C,$6E,$31,$3C,$23,$FC,$3F
	.dc.b	$3A,$79,$46,$5F,$F1,$06,$7F,$D1,$9F,$F1,$C3,$5F,$03,$5E,$6D,$7F
	.dc.b	$D1,$DD,$3C,$DF,$E5,$77,$F2,$63,$18,$9D,$60,$FC,$1F,$FD,$1F,$7F
	.dc.b	$E4,$76,$BE,$F6,$BE,$A6,$BF,$E3,$76,$A8,$E9,$EB,$7C,$59,$CC,$62
	.dc.b	$7F,$B3,$F1,$BE,$7F,$EF,$36,$33,$18,$3E,$77,$D9,$7F,$45,$8F,$F2
	.dc.b	$E3,$5F,$A8,$D7,$E7,$B5,$FF,$26,$35,$F3,$36,$8D,$35,$97,$FA,$BF
	.dc.b	$A3,$F2,$5E,$0F,$F8,$FA,$FF,$BF,$B5,$FF,$0B,$B5,$FB,$FE,$23,$D6
	.dc.b	$FC,$33,$63,$FC,$E8,$D7,$F1,$8D,$78,$C7,$02,$3C,$1F,$D2,$73,$BE
	.dc.b	$AA,$92,$7F,$C2,$71,$97,$F9,$47,$D4,$3E,$27,$F9,$43,$63,$31,$82
	.dc.b	$38,$21,$FE,$22,$8E,$0F,$F8,$93,$EA,$5E,$D7,$F3,$36,$C7,$F3,$76
	.dc.b	$BB,$8F,$0F,$93,$23,$CA,$D7,$B3,$5F,$F3,$75,$7F,$B8,$1D,$9F,$E7
	.dc.b	$3F,$9A,$FF,$6B,$3F,$E7,$06,$BF,$E7,$06,$BF,$E7,$66,$A8,$7B,$DF
	.dc.b	$10,$91,$F2,$1A,$FC,$16,$BE,$47,$F8,$17,$D4,$BE,$9E,$92,$7F,$D6
	.dc.b	$51,$96,$FA,$3C,$08,$E0,$FA,$FF,$39,$F5,$88,$E0,$E7,$F4,$DE,$02
	.dc.b	$38,$3F,$B3,$7D,$8B,$CA,$FE,$CC,$D8,$CC,$60,$8E,$08,$7F,$98,$23
	.dc.b	$83,$FE,$62,$F6,$9F,$27,$FD,$C5,$B1,$FD,$D5,$AE,$E3,$F9,$D7,$BD
	.dc.b	$FD,$56,$3F,$CF,$AD,$7F,$D1,$2D,$7F,$A1,$75,$23,$11,$FE,$83,$6C
	.dc.b	$67,$C9,$EA,$49,$FF,$39,$F9,$AF,$A5,$FE,$94,$D8,$CC,$60,$8E,$08
	.dc.b	$7F,$A4,$A3,$83,$FE,$BF,$F4,$1F,$62,$F9,$6F,$F0,$A4,$61,$B1,$99
	.dc.b	$FF,$44,$CE,$5B,$19,$7A,$35,$FF,$72,$C6,$DC,$CF,$F8,$87,$BD,$F4
	.dc.b	$7F,$E2,$16,$CE,$63,$06,$7E,$99,$F2,$7C,$5B,$3F,$E9,$CA,$46,$23
	.dc.b	$FD,$B9,$18,$9E,$04,$8E,$13,$B7,$FD,$F7,$3B,$7F,$D4,$7B,$7F,$89
	.dc.b	$5B,$E7,$FE,$08,$8E,$08,$7E,$11,$CB,$FE,$B0,$F8,$5F,$E1,$EF,$FB
	.dc.b	$BD,$AF,$F8,$1C,$38,$7F,$C1,$4D,$73,$E0,$7F,$91,$E3,$04,$7F,$C9
	.dc.b	$5F,$4D,$FE,$04,$93,$49,$E8,$FE,$6B,$D5,$93,$EE,$6B,$F9,$C6,$BF
	.dc.b	$9C,$6B,$FE,$1D,$CB,$EC,$F9,$7F,$5C,$91,$F4,$F6,$FE,$C4,$FD,$26
	.dc.b	$BF,$39,$AF,$36,$A9,$EA,$FF,$99,$F2,$FE,$3E,$4F,$5B,$E1,$12,$34
	.dc.b	$D7,$E8,$3F,$E6,$FF,$93,$7B,$0F,$95,$C3,$6B,$FD,$CB,$3A,$9C,$9F
	.dc.b	$B0,$FA,$59,$3F,$B4,$6B,$FA,$D6,$BF,$6F,$A9,$D0,$9C,$CF,$F0,$4E
	.dc.b	$3F,$F1,$CD,$7F,$05,$E0,$FE,$1A,$49,$FC,$8B,$5F,$E1,$ED,$7F,$C6
	.dc.b	$7F,$37,$FD,$01,$FF,$7E,$FA,$28,$C4,$7F,$8F,$DA,$F0,$F1,$BF,$8D
	.dc.b	$FC,$AC,$8F,$CF,$B5,$E1,$F4,$7F,$E4,$5A,$34,$91,$E0,$8C,$B6,$32
	.dc.b	$F2,$35,$F4,$3C,$C8,$E0,$FD,$7F,$C6,$FF,$2E,$23,$83,$B5,$78,$C8
	.dc.b	$E0,$FD,$BF,$81,$FC,$D8,$8E,$08,$7F,$38,$72,$D3,$27,$F3,$A7,$2D
	.dc.b	$3E,$6F,$E4,$CF,$9F,$FF,$34,$36,$33,$18,$3F,$A3,$F0,$48,$FE,$16
	.dc.b	$D7,$F8,$01,$F9,$4F,$0B,$FC,$A3,$CB,$FC,$D3,$DA,$FF,$63,$36,$76
	.dc.b	$6E,$33,$3C,$08,$E0,$FE,$5D,$F4,$7F,$C7,$67,$F7,$0D,$7F,$5C,$DE
	.dc.b	$A3,$9B,$FE,$8B,$E9,$76,$25,$78,$A3,$2D,$C3,$FC,$4A,$FF,$1B,$8C
	.dc.b	$BE,$E6,$B9,$8D,$BE,$4E,$8F,$39,$1F,$E7,$97,$FB,$57,$F9,$DB,$5F
	.dc.b	$E6,$ED,$79,$1C,$23,$E4,$FC,$EB,$E6,$29,$27,$D5,$19,$6C,$65,$F4
	.dc.b	$46,$5B,$1F,$D6,$7C,$6F,$F0,$5C,$7F,$48,$6B,$FC,$95,$AF,$F1,$B6
	.dc.b	$AF,$A2,$7C,$97,$F1,$48,$7B,$DF,$E6,$B9,$CE,$9F,$EC,$B8,$C5,$6D
	.dc.b	$FF,$39,$75,$7F,$83,$7E,$8F,$FA,$54,$6A,$B8,$7B,$9F,$D3,$70,$E7
	.dc.b	$8F,$68,$C6,$3C,$DF,$F6,$C6,$3C,$E7,$89,$FE,$DB,$AE,$2F,$E6,$1A
	.dc.b	$F2,$BF,$CA,$BE,$DB,$F6,$3B,$EC,$FF,$46,$3C,$DA,$FC,$36,$D7,$F9
	.dc.b	$6B,$DE,$E6,$D9,$F4,$1C,$63,$CB,$83,$E8,$FF,$8A,$F8,$9E,$8A,$E9
	.dc.b	$56,$F9,$69,$F3,$71,$FF,$51,$3F,$2B,$0E,$37,$D8,$E2,$78,$39,$5F
	.dc.b	$A4,$BE,$03,$FC,$EB,$18,$6C,$67,$3F,$F0,$A3,$6B,$F4,$5F,$A7,$BE
	.dc.b	$AF,$FC,$FB,$C0,$8C,$5F,$EB,$5B,$5C,$2B,$71,$88,$FD,$73,$51,$FD
	.dc.b	$79,$C3,$63,$CF,$C7,$FE,$0E,$AF,$F1,$43,$67,$FC,$4E,$D7,$F3,$9E
	.dc.b	$4F,$02,$4F,$06,$C7,$EF,$7E,$D1,$F2,$BE,$C6,$D7,$F0,$76,$CE,$E7
	.dc.b	$44,$9D,$36,$3C,$7F,$39,$E8,$24,$FA,$1E,$DF,$31,$23,$B3,$67,$C8
	.dc.b	$D7,$F8,$A5,$A3,$11,$FD,$CF,$E0,$BF,$D5,$FE,$5F,$F9,$DD,$BE,$61
	.dc.b	$A7,$FA,$D3,$A1,$E2,$BE,$0F,$0D,$CF,$FA,$3C,$60,$E7,$F5,$32,$3F
	.dc.b	$C9,$AD,$CB,$FD,$16,$D7,$FD,$16,$D7,$FC,$10,$D7,$FD,$14,$DB,$FF
	.dc.b	$00,$7B,$1E,$5F,$2B,$FC,$CF,$7F,$DD,$63,$07,$B3,$8A,$47,$F9,$79
	.dc.b	$B8,$7F,$81,$FE,$8F,$91,$27,$FB,$F3,$D7,$E8,$A4,$7F,$80,$5B,$3F
	.dc.b	$E0,$F6,$BF,$E1,$66,$BF,$E2,$1F,$0B,$E6,$92,$7F,$CD,$EF,$CD,$F9
	.dc.b	$DF,$F1,$BC,$7F,$87,$1B,$3F,$DA,$33,$46,$23,$FC,$D5,$18,$E9,$FB
	.dc.b	$3F,$2F,$2F,$FC,$F6,$DA,$FF,$5C,$B5,$FF,$5C,$B5,$FF,$5C,$FD,$03
	.dc.b	$E5,$BF,$0A,$DA,$FE,$E7,$CA,$8C,$57,$F7,$C8,$C5,$70,$25,$70,$AD
	.dc.b	$D7,$F9,$1D,$AF,$EE,$BD,$77,$DB,$FF,$B0,$6B,$FC,$AA,$D9,$FF,$04
	.dc.b	$CE,$BD,$25,$69,$B3,$FE,$08,$F1,$1E,$A6,$23,$FC,$23,$18,$AF,$E7
	.dc.b	$2D,$8E,$13,$B7,$FA,$6B,$5F,$EA,$8D,$7F,$AC,$B6,$39,$47,$FA,$B3
	.dc.b	$E9,$9E,$3F,$F2,$86,$DF,$FA,$E5,$B3,$FE,$B8,$6B,$C6,$78,$12,$78
	.dc.b	$3F,$B6,$F8,$69,$1F,$EB,$86,$CF,$FB,$D9,$AF,$FB,$D9,$AF,$F2,$88
	.dc.b	$7D,$33,$D3,$F7,$46,$1B,$19,$8F,$F1,$99,$93,$FD,$1D,$F1,$5F,$0F
	.dc.b	$53,$49,$1A,$25,$FF,$A4,$23,$2F,$E2,$7E,$D7,$C8,$D3,$F8,$AB,$87
	.dc.b	$FD,$1C,$D7,$FA,$0F,$12,$31,$08,$C3,$C8,$95,$FE,$E9,$F6,$5E,$DF
	.dc.b	$F5,$C6,$C6,$63,$04,$70,$49,$FF,$77,$CF,$07,$FD,$81,$E3,$FF,$66
	.dc.b	$F7,$BF,$A0,$C7,$FB,$06,$30,$7E,$2F,$E1,$7F,$55,$DD,$79,$CD,$26
	.dc.b	$3F,$E6,$29,$DB,$67,$6D,$9D,$DF,$EC,$E7,$C2,$FE,$F2,$FF,$76,$4A
	.dc.b	$FD,$A3,$63,$83,$5F,$9F,$E9,$7D,$22,$4F,$F0,$2D,$51,$89,$FB,$4D
	.dc.b	$78,$CF,$E2,$3C,$F7,$1F,$E3,$9B,$19,$8C,$1F,$AD,$5F,$03,$48,$FF
	.dc.b	$8C,$0C,$95,$CD,$7E,$59,$AF,$F2,$6E,$F7,$BF,$FC,$BC,$95,$FC,$BD
	.dc.b	$FD,$E7,$7F,$F8,$92,$BF,$CF,$5E,$FE,$09,$5F,$CD,$78,$FF,$9C,$11
	.dc.b	$C1,$FC,$E7,$8F,$F9,$E1,$1C,$1F,$E9,$EF,$5B,$B9,$1C,$12,$FF,$D2
	.dc.b	$11,$96,$C7,$FA,$93,$67,$C2,$FE,$81,$B3,$FC,$5D,$B3,$B9,$FD,$2C
	.dc.b	$B9,$1F,$5D,$FC,$6D,$B5,$FE,$5C,$6C,$FE,$46,$74,$49,$D3,$63,$FD
	.dc.b	$7B,$6F,$A2,$78,$9E,$36,$DF,$A5,$B5,$C1,$AF,$FB,$1B,$0F,$E1,$73
	.dc.b	$E3,$EA,$F3,$11,$C1,$27,$F4,$D1,$97,$FA,$A7,$BF,$FC,$B1,$FE,$0E
	.dc.b	$D7,$F8,$33,$7A,$CE,$7C,$67,$DE,$3E,$63,$FC,$44,$DC,$7F,$DA,$B3
	.dc.b	$A9,$C9,$23,$33,$BC,$3F,$BD,$35,$FE,$F2,$D7,$F8,$DE,$1F,$D7,$27
	.dc.b	$FB,$7B,$63,$FB,$73,$5F,$F3,$09,$FC,$7D,$C3,$FE,$9C,$F8,$A7,$D3
	.dc.b	$7F,$8E,$5B,$5E,$86,$CF,$F8,$35,$AF,$F8,$31,$AF,$39,$E4,$49,$FF
	.dc.b	$01,$F9,$3F,$E4,$19,$FF,$01,$35,$F2,$DF,$F7,$CF,$47,$FC,$9F,$39
	.dc.b	$9F,$F5,$8F,$E2,$6F,$85,$FF,$27,$37,$0F,$9A,$D9,$FF,$11,$35,$FF
	.dc.b	$10,$B5,$E7,$3C,$89,$3F,$E1,$83,$C1,$1F,$D1,$5B,$87,$F8,$2B,$0F
	.dc.b	$F3,$1C,$6B,$57,$C7,$3E,$43,$FC,$C4,$D9,$FC,$0B,$70,$FF,$1B,$35
	.dc.b	$FF,$1A,$FD,$03,$EB,$BF,$CD,$4D,$AF,$C4,$B6,$7F,$C7,$CD,$78,$CF
	.dc.b	$02,$4F,$07,$FC,$F3,$C9,$FA,$89,$FE,$C6,$D7,$FB,$13,$73,$FE,$01
	.dc.b	$CD,$C0,$97,$FE,$3C,$9D,$1C,$6B,$FD,$DA,$3E,$61,$F4,$1F,$BE,$AF
	.dc.b	$F5,$8C,$62,$7F,$C5,$A6,$E7,$4D,$33,$1B,$7F,$D7,$3D,$3F,$E3,$88
	.dc.b	$C4,$7F,$C0,$13,$81,$AA,$EF,$5F,$E3,$D8,$C9,$FD,$4A,$32,$F6,$8F
	.dc.b	$0C,$F4,$87,$A1,$A7,$A1,$C8,$C9,$D2,$FC,$95,$DE,$BC,$B1,$89,$EC
	.dc.b	$7E,$C9,$C3,$F2,$0F,$F2,$69,$FB,$63,$E2,$98,$31,$F3,$DF,$D7,$BE
	.dc.b	$87,$51,$FE,$DC,$8C,$38,$6B,$C3,$C6,$4F,$08,$4B,$E5,$1F,$43,$98
	.dc.b	$C3,$63,$33,$D5,$AF,$91,$CE,$3F,$5E,$9D,$5F,$C4,$23,$0D,$CF,$F8
	.dc.b	$9C,$F6,$BF,$B2,$DA,$DC,$7D,$18,$E6,$DB,$EF,$5C,$9B,$1F,$0E,$FB
	.dc.b	$34,$E1,$84,$7B,$C6,$9F,$86,$DB,$F1,$5F,$5A,$F1,$4E,$4D,$5F,$83
	.dc.b	$81,$87,$F8,$20,$FF,$50,$94,$FF,$50,$14,$FE,$3E,$53,$FC,$AA,$53
	.dc.b	$FC,$A8,$57,$AF,$8F,$F6,$5B,$1F,$D3,$3E,$87,$FA,$83,$7B,$BE,$F7
	.dc.b	$42,$73,$3F,$EC,$56,$BF,$EC,$46,$BE,$A8,$E1,$3C,$A7,$32,$9C,$BF
	.dc.b	$D7,$34,$7D,$27,$81,$B1,$98,$C1,$E9,$FA,$2E,$78,$9A,$4C,$7D,$3F
	.dc.b	$45,$FC,$FE,$FF,$B4,$B5,$FF,$6F,$35,$FF,$6E,$B5,$F5,$35,$F4,$37
	.dc.b	$13,$4D,$5F,$F0,$0F,$97,$F9,$DC,$A6,$92,$FF,$BD,$FD,$17,$FB,$85
	.dc.b	$BE,$EF,$D0,$8C,$47,$FB,$CD,$AF,$FB,$C9,$AF,$CF,$6B,$F3,$5B,$89
	.dc.b	$CE,$0F,$8E,$FF,$35,$D7,$FB,$B9,$AF,$FB,$B5,$AF,$F6,$C6,$BF,$77
	.dc.b	$D2,$7C,$8F,$E1,$9B,$1F,$E1,$66,$BF,$C4,$F0,$46,$23,$FC,$17,$18
	.dc.b	$9E,$44,$8E,$4F,$E5,$F9,$DF,$45,$49,$3F,$C6,$B1,$FE,$91,$1F,$C5
	.dc.b	$61,$F4,$2F,$9C,$C3,$63,$F8,$D3,$5F,$E4,$6D,$7F,$95,$35,$FF,$10
	.dc.b	$B5,$FD,$BB,$5E,$F1,$E1,$FA,$2F,$F6,$F4,$7F,$27,$6B,$FC,$C1,$AF
	.dc.b	$F3,$76,$BD,$BC,$6F,$39,$1F,$3B,$5F,$3B,$5F,$3B,$FC,$D7,$BD,$F1
	.dc.b	$6F,$07,$0D,$AF,$F3,$E3,$5F,$F3,$5B,$5F,$83,$F0,$3F,$20,$8F,$C8
	.dc.b	$6B,$F2,$1A,$F0,$3F,$CA,$FF,$42,$FA,$2F,$E1,$6D,$8C,$C6,$0C,$FA
	.dc.b	$32,$38,$B5,$FA,$BF,$45,$F6,$88,$FD,$A7,$AF,$99,$23,$C2,$D7,$FA
	.dc.b	$FF,$D1,$7F,$B1,$1F,$EB,$4D,$73,$EE,$FE,$26,$31,$1F,$DA,$DA,$FF
	.dc.b	$98,$9A,$FF,$38,$6A,$6F,$9C,$FA,$2F,$CA,$36,$7F,$B5,$35,$FE,$E6
	.dc.b	$D7,$FC,$B7,$F2,$3F,$D6,$1F,$EF,$9C,$C8,$C4,$6B,$C8,$FE,$79,$18
	.dc.b	$8F,$E9,$73,$B9,$C9,$23,$F5,$13,$B5,$2F,$A1,$7C,$5F,$F4,$76,$CF
	.dc.b	$EB,$DA,$F9,$1A,$FF,$9C,$9A,$F8,$9A,$F6,$6C,$F7,$9F,$F6,$17,$B0
	.dc.b	$F8,$EA,$4B,$FF,$0B,$FA,$3F,$EA,$88,$C4,$7F,$AE,$67,$83,$A6,$B9
	.dc.b	$8F,$F0,$C7,$82,$F8,$7F,$E0,$CD,$9C,$C6,$0C,$FB,$AF,$8F,$E2,$D9
	.dc.b	$FF,$14,$D7,$FB,$92,$72,$48,$CC,$ED,$FF,$1A,$4E,$DC,$CF,$D8,$F4
	.dc.b	$24,$7D,$46,$BF,$80,$DB,$FC,$5C,$E4,$86,$63,$F9,$07,$CB,$7F,$13
	.dc.b	$27,$81,$82,$4F,$F8,$E6,$32,$D8,$FE,$51,$F4,$5F,$E0,$98,$34,$9E
	.dc.b	$97,$E3,$BD,$69,$3F,$8A,$6B,$E7,$6B,$F9,$A6,$BF,$E1,$EC,$FE,$97
	.dc.b	$99,$F4,$09,$1F,$33,$6F,$ED,$0D,$8C,$BF,$2A,$32,$D8,$F8,$3E,$0C
	.dc.b	$8F,$73,$5F,$F2,$FE,$7F,$C7,$5F,$F6,$DB,$5F,$F1,$8B,$53,$9F,$8B
	.dc.b	$7C,$33,$E7,$FF,$CC,$51,$86,$C6,$6B,$FC,$DB,$39,$7F,$51,$F4,$3F
	.dc.b	$AB,$23,$82,$1F,$AB,$72,$D3,$F6,$FA,$FE,$F8,$8E,$08,$7E,$1D,$CB
	.dc.b	$4F,$F3,$FF,$9A,$F9,$47,$8D,$FE,$3E,$6C,$E9,$AF,$F7,$96,$BF,$C5
	.dc.b	$3E,$23,$F8,$79,$1F,$E1,$6F,$3F,$89,$91,$CD,$AF,$36,$BF,$E0,$29
	.dc.b	$46,$23,$FC,$95,$0F,$92,$F3,$12,$3C,$CF,$F2,$0F,$8A,$91,$FC,$71
	.dc.b	$AF,$A6,$7D,$C4,$7D,$CF,$F2,$8E,$E9,$1F,$C9,$1A,$FE,$C5,$AF,$F3
	.dc.b	$46,$BF,$CB,$BA,$7F,$8D,$92,$33,$1F,$E5,$FF,$96,$7D,$17,$F9,$B1
	.dc.b	$B1,$F5,$1A,$FF,$89,$DA,$FF,$42,$DF,$F9,$E9,$1F,$E7,$2F,$FA,$CF
	.dc.b	$85,$DF,$FE,$57,$6F,$C2,$E6,$30,$47,$04,$AF,$F6,$1C,$F0,$7F,$8E
	.dc.b	$FD,$17,$F1,$97,$FD,$2A,$7F,$92,$5C,$3F,$BB,$6B,$9E,$FF,$07,$CA
	.dc.b	$48,$EB,$87,$F8,$79,$FD,$5B,$6F,$D3,$3B,$9F,$F6,$D4,$61,$E3,$1F
	.dc.b	$CE,$AF,$F5,$D1,$FC,$B1,$AF,$9B,$B1,$18,$8F,$37,$47,$12,$3E,$E6
	.dc.b	$BE,$0F,$A2,$7D,$17,$CD,$6C,$7C,$66,$BB,$8D,$11,$D3,$5F,$E8,$3E
	.dc.b	$27,$F8,$23,$0F,$EB,$DE,$67,$F5,$48,$C1,$FE,$0A,$FA,$27,$C8,$FF
	.dc.b	$10,$B7,$BF,$FE,$6C,$9C,$ED,$FE,$C8,$8C,$56,$DF,$F3,$9F,$6F,$F8
	.dc.b	$1F,$E6,$7E,$CC,$6A,$B8,$7B,$BF,$73,$0E,$78,$F6,$8C,$63,$CD,$FF
	.dc.b	$5E,$74,$B1,$E7,$1A,$C3,$F0,$B7,$4A,$30,$EA,$E9,$AF,$87,$BF,$FD
	.dc.b	$1A,$ED,$B1,$FC,$5B,$D8,$E7,$39,$6C,$66,$3D,$07,$FB,$A9,$C6,$7F
	.dc.b	$0B,$E8,$BF,$86,$F8,$BF,$02,$B5,$87,$FA,$8A,$BE,$63,$72,$F2,$63
	.dc.b	$BD,$98,$F2,$C3,$B9,$E3,$3C,$3C,$2F,$A2,$FF,$4F,$E7,$FE,$06,$6C
	.dc.b	$FF,$80,$9A,$FE,$5D,$A8,$FE,$9F,$F4,$4F,$A2,$FE,$B1,$BA,$31,$7F
	.dc.b	$B8,$6D,$7E,$A1,$AF,$EE,$27,$85,$72,$9C,$CF,$79,$CD,$73,$AF,$EC
	.dc.b	$D8,$7F,$8C,$FC,$8B,$C9,$FD,$92,$BF,$C4,$2D,$C3,$F3,$B3,$A2,$4E
	.dc.b	$9B,$1F,$BC,$FA,$17,$CD,$7B,$1B,$5F,$C4,$9B,$3B,$9D,$12,$74,$D8
	.dc.b	$FD,$6F,$D1,$79,$89,$3E,$67,$B7,$D1,$48,$EC,$D9,$F2,$35,$F9,$16
	.dc.b	$8C,$47,$F3,$FF,$80,$FF,$05,$7A,$DF,$E7,$36,$F4,$FF,$AC,$30,$FF
	.dc.b	$36,$C7,$F8,$3F,$8D,$E2,$BE,$07,$0D,$AF,$E8,$51,$83,$9F,$D1,$48
	.dc.b	$FF,$3B,$B7,$2F,$F2,$23,$5F,$F4,$3B,$5F,$EA,$8D,$7F,$CF,$B8,$A3
	.dc.b	$18,$7F,$80,$FC,$5F,$37,$95,$FE,$5F,$C3,$FC,$05,$18,$3D,$BC,$52
	.dc.b	$3F,$D1,$2D,$C7,$FC,$0D,$F3,$FC,$89,$3F,$E0,$B7,$AF,$D1,$48,$FF
	.dc.b	$07,$36,$7F,$C1,$CD,$7F,$C0,$8D,$7F,$B7,$78,$1F,$38,$93,$FE,$9A
	.dc.b	$7E,$77,$CF,$7F,$8D,$A3,$FC,$3E,$D9,$FE,$2B,$9A,$31,$1F,$D7,$23
	.dc.b	$1D,$3F,$6B,$E5,$A5,$FF,$A9,$9B,$5F,$E7,$36,$BF,$EB,$66,$BF,$EB
	.dc.b	$0F,$A0,$7C,$77,$E1,$9B,$5F,$C9,$39,$51,$8A,$FF,$5E,$B5,$FE,$DF
	.dc.b	$1A,$AE,$44,$AE,$55,$BA,$FE,$11,$EB,$3E,$6F,$FD,$73,$5F,$E5,$26
	.dc.b	$CF,$F3,$69,$D7,$A5,$FE,$63,$6D,$7F,$2E,$F1,$BF,$C6,$0D,$F5,$3F
	.dc.b	$CC,$D1,$8B,$FE,$7A,$D7,$FA,$0B,$5F,$E9,$F1,$C0,$93,$FE,$A8,$FA
	.dc.b	$2F,$F2,$8B,$7C,$8F,$E1,$F1,$8A,$FF,$53,$B5,$FE,$F8,$D7,$FD,$14
	.dc.b	$D5,$E1,$23,$0D,$9F,$DA,$F9,$72,$FF,$DC,$2D,$BF,$F5,$73,$5F,$F7
	.dc.b	$8B,$5F,$E6,$90,$FA,$27,$A3,$FE,$8B,$8C,$36,$33,$1F,$D8,$8C,$9F
	.dc.b	$E8,$8F,$89,$FF,$46,$11,$C1,$2B,$FD,$13,$19,$6C,$7E,$27,$E8,$BC
	.dc.b	$71,$FD,$C3,$8D,$18,$8F,$C4,$B5,$FE,$99,$1A,$7F,$8C,$C6,$5F,$F8
	.dc.b	$02,$D8,$D4,$F7,$9F,$E8,$7E,$C3,$DA,$A4,$9F,$F7,$1C,$65,$B1,$99
	.dc.b	$FF,$70,$CE,$5B,$1F,$D3,$BD,$DF,$E7,$F1,$FE,$39,$6B,$FE,$C7,$6B
	.dc.b	$F7,$9B,$F0,$7F,$AA,$EE,$FF,$5E,$37,$0F,$F0,$8B,$6B,$FC,$2A,$D7
	.dc.b	$FC,$B6,$D7,$FB,$83,$53,$4C,$63,$FB,$8F,$2B,$F7,$84,$AF,$D9,$B6
	.dc.b	$BF,$76,$D7,$F7,$7D,$1E,$AA,$FA,$FE,$87,$D2,$25,$FD,$D6,$D7,$DD
	.dc.b	$D5,$18,$8F,$BA,$FE,$47,$E4,$5C,$54,$97,$F9,$18,$CB,$FE,$2A,$1F
	.dc.b	$01,$FE,$E1,$69,$97,$04,$70,$4B,$FC,$E4,$7F,$30,$F7,$E4,$FF,$30
	.dc.b	$6C,$FF,$A3,$3D,$FF,$F7,$77,$81,$FE,$29,$8F,$F3,$EC,$60,$FE,$73
	.dc.b	$C5,$2F,$F9,$CB,$67,$F9,$CF,$84,$E4,$FE,$7E,$4A,$FE,$7E,$F6,$F5
	.dc.b	$E4,$7F,$A9,$5B,$5F,$E9,$E6,$BF,$E8,$E6,$BF,$E3,$1D,$9F,$09,$FA
	.dc.b	$56,$C2,$BF,$CB,$B1,$86,$CF,$EA,$E7,$F1,$3F,$45,$B7,$F2,$6D,$7F
	.dc.b	$26,$D7,$F2,$6D,$7F,$26,$D7,$9C,$FF,$98,$B9,$53,$96,$C7,$93,$E8
	.dc.b	$9E,$2F,$3C,$51,$89,$D4,$E6,$31,$87,$EA,$3A,$BA,$12,$BC,$0D,$8D
	.dc.b	$46,$08,$FF,$13,$F7,$FF,$9A,$73,$FF,$B5,$3B,$BF,$C4,$F1,$83,$F9
	.dc.b	$C7,$8E,$FA,$27,$CB,$FF,$89,$DB,$8F,$FA,$2A,$75,$39,$24,$66,$77
	.dc.b	$8F,$F7,$E6,$BF,$DF,$5A,$FF,$22,$C7,$ED,$4F,$F7,$36,$C7,$F2,$B6
	.dc.b	$BF,$DC,$9A,$FF,$9D,$6D,$F1,$4F,$A2,$FF,$1F,$36,$FD,$0D,$AF,$F0
	.dc.b	$8B,$5F,$F0,$83,$5E,$8D,$55,$DA,$BA,$9F,$E5,$0F,$2B,$FC,$91,$3F
	.dc.b	$D0,$5A,$FF,$82,$1A,$FF,$40,$6B,$D3,$E8,$9F,$0B,$FE,$55,$6E,$3F
	.dc.b	$3D,$B3,$FE,$28,$6B,$FE,$27,$6B,$F0,$DA,$F5,$9E,$87,$F9,$9A,$7C
	.dc.b	$39,$7E,$A7,$D5,$C1,$23,$EF,$7C,$6B,$E4,$3F,$CD,$0D,$8C,$C6,$0F
	.dc.b	$F3,$67,$D1,$3E,$8A,$92,$7F,$11,$19,$6C,$67,$2F,$F2,$1C,$E5,$B1
	.dc.b	$FE,$7C,$E8,$FD,$1C,$7F,$8C,$9A,$FF,$6A,$6A,$CB,$FA,$3F,$3F,$12
	.dc.b	$61,$FE,$3F,$AD,$CE,$A3,$FD,$B7,$5F,$EF,$91,$F3,$2F,$A1,$FF,$75
	.dc.b	$F5,$7F,$09,$8C,$4F,$F8,$B8,$DC,$EA,$BF,$D9,$13,$A9,$FF,$7F,$9A
	.dc.b	$3F,$D9,$87,$F4,$A8,$DC,$FF,$8E,$A3,$47,$F2,$18,$CB,$DE,$7A,$3C
	.dc.b	$4D,$9E,$97,$35,$E7,$AE,$91,$8A,$E7,$5E,$48,$C3,$C8,$7A,$3A,$7B
	.dc.b	$47,$81,$C9,$E0,$BE,$CF,$C3,$3F,$CB,$07,$CC,$9F,$9A,$F0,$3E,$08
	.dc.b	$F9,$FF,$EC,$3F,$45,$88,$FF,$78,$46,$1C,$35,$E2,$D7,$E5,$AF,$93
	.dc.b	$7D,$17,$39,$ED,$18,$79,$B5,$E6,$D7,$9B,$BF,$95,$FB,$55,$FC,$3A
	.dc.b	$42,$33,$D5,$F7,$1D,$D7,$1D,$79,$E1,$C2,$77,$76,$7D,$37,$C5,$DC
	.dc.b	$75,$6E,$9C,$23,$F0,$B8,$F8,$DB,$8F,$B6,$7C,$38,$F5,$9F,$97,$87
	.dc.b	$56,$99,$3E,$2C,$5B,$F9,$4D,$9E,$13,$F5,$DB,$3A,$8F,$31,$FB,$13
	.dc.b	$FC,$A8,$7C,$69,$FE,$E1,$B7,$C8,$F1,$BF,$43,$18,$9D,$9F,$63,$7F
	.dc.b	$F3,$2A,$F2,$77,$18,$AB,$3D,$DB,$7D,$63,$C1,$3C,$AB,$3E,$4B,$E1
	.dc.b	$E9,$2B,$67,$F3,$EF,$A2,$FE,$C2,$E3,$05,$71,$8C,$35,$F4,$1F,$C4
	.dc.b	$5C,$65,$FE,$0F,$EF,$5F,$4A,$DD,$BE,$87,$11,$FE,$CE,$6B,$FE,$86
	.dc.b	$6B,$FC,$75,$B1,$FE,$21,$F9,$1F,$83,$A1,$A7,$C4,$9F,$06,$3F,$4B
	.dc.b	$95,$F2,$BF,$A5,$C3,$F7,$ED,$8C,$E3,$FE,$E5,$6B,$FE,$9D,$6B,$FE
	.dc.b	$E4,$6B,$FE,$9C,$6B,$FC,$CD,$B3,$FE,$E3,$FA,$DF,$AF,$E1,$1A,$7D
	.dc.b	$A9,$EF,$87,$E2,$3D,$AF,$F3,$7E,$1F,$EF,$76,$BF,$EA,$F6,$BF,$EF
	.dc.b	$66,$BF,$EA,$E6,$BF,$D1,$5B,$87,$FB,$5B,$AB,$E1,$46,$21,$E6,$FE
	.dc.b	$31,$B5,$F8,$36,$BF,$6A,$7C,$77,$F9,$EE,$77,$D0,$7C,$D6,$D7,$F1
	.dc.b	$DB,$FE,$9B,$39,$24,$66,$76,$FF,$B2,$DA,$FF,$48,$9E,$2F,$3F,$27
	.dc.b	$C7,$7F,$AE,$6F,$0F,$FB,$39,$B3,$FD,$96,$FA,$D7,$F9,$87,$E7,$FF
	.dc.b	$DD,$33,$FC,$9D,$AF,$F3,$36,$BF,$CE,$1A,$FF,$5A,$6B,$FC,$91,$A9
	.dc.b	$A6,$B0,$7B,$EF,$8D,$FD,$CB,$6B,$F6,$0D,$7F,$5A,$73,$F7,$BF,$80
	.dc.b	$EE,$F8,$BF,$F3,$A3,$75,$FE,$80,$DA,$FF,$42,$CE,$89,$3A,$6C,$7F
	.dc.b	$51,$DD,$F6,$0F,$9B,$FE,$1F,$18,$6C,$66,$7F,$84,$E1,$DD,$FE,$07
	.dc.b	$D4,$FA,$07,$D0,$7F,$53,$8C,$36,$33,$3F,$E6,$D6,$BF,$D9,$A7,$6E
	.dc.b	$67,$81,$1F,$F3,$A7,$D8,$7F,$16,$7F,$8D,$CA,$31,$1E,$46,$F1,$7F
	.dc.b	$15,$BE,$EF,$FA,$8F,$E8,$9F,$45,$FC,$BA,$D1,$87,$F9,$39,$88,$FE
	.dc.b	$67,$87,$BF,$D7,$F8,$FF,$66,$F6,$3F,$B9,$70,$23,$18,$7C,$F6,$BF
	.dc.b	$E0,$48,$D4,$FF,$15,$F8,$27,$CA,$7F,$3E,$6D,$7F,$35,$6B,$F4,$35
	.dc.b	$FF,$2C,$3E,$6F,$51,$F4,$5F,$E3,$E6,$C7,$F8,$15,$AF,$F8,$45,$AF
	.dc.b	$F8,$59,$AF,$F8,$E1,$AF,$F8,$09,$A9,$CF,$53,$E1,$7F,$A1,$F4,$7F
	.dc.b	$82,$23,$1E,$4F,$F5,$89,$CD,$F3,$C3,$DD,$F3,$D2,$3D,$CF,$F5,$EF
	.dc.b	$95,$FF,$00,$72,$FF,$6C,$D5,$3B,$AF,$E9,$B1,$87,$FA,$D5,$70,$CF
	.dc.b	$C5,$8F,$FB,$62,$7F,$AF,$35,$FE,$AA,$DC,$3F,$BC,$FA,$4F,$8E,$F4
	.dc.b	$7C,$0C,$BE,$2F,$F7,$29,$D4,$7F,$6C,$E3,$7C,$5F,$F9,$03,$0F,$F7
	.dc.b	$94,$62,$7F,$5B,$39,$CB,$FC,$5B,$EB,$BC,$67,$C9,$6A,$E1,$D6,$F8
	.dc.b	$2A,$E1,$3F,$87,$F9,$09,$3F,$87,$78,$BE,$EF,$81,$FA,$44,$7D,$1F
	.dc.b	$0A,$F9,$3C,$46,$CF,$8F,$3F,$C1,$27,$51,$F9,$3F,$8D,$FC,$B1,$1F
	.dc.b	$A3,$ED,$D7,$EA,$C5,$DF,$FC,$4B,$C9,$FE,$59,$9F,$D6,$FD,$17,$F8
	.dc.b	$85,$FF,$2D,$35,$FF,$33,$35,$FF,$35,$B5,$FF,$46,$B5,$FF,$2B,$B5
	.dc.b	$34,$D7,$96,$E1,$FE,$0A,$7F,$97,$9C,$7B,$1F,$E3,$06,$D7,$9B,$D9
	.dc.b	$FD,$A7,$BB,$23,$F6,$8F,$E5,$FE,$33,$11,$F9,$48,$B9,$FC,$09,$FD
	.dc.b	$17,$99,$2B,$F2,$6F,$E1,$3D,$8C,$47,$E5,$22,$D7,$E2,$B9,$BF,$C0
	.dc.b	$FA,$79,$F4,$FF,$11,$5F,$FA,$2F,$4F,$F1,$CF,$B5,$FE,$83,$6C,$66
	.dc.b	$76,$FF,$A5,$23,$2D,$8C,$BF,$EB,$88,$CB,$63,$2F,$FA,$13,$A5,$F8
	.dc.b	$17,$FB,$F2,$3F,$44,$D7,$FC,$8F,$1A,$F8,$6F,$99,$5E,$2C,$FF,$CC
	.dc.b	$FF,$4F,$F4,$EF,$FC,$A4,$D7,$FC,$92,$D8,$FD,$AF,$D1,$FF,$98,$A3
	.dc.b	$1A,$7F,$0F,$6D,$7F,$B9,$5B,$E3,$7F,$A6,$6B,$F9,$67,$D5,$4A,$FE
	.dc.b	$58,$FF,$0F,$E4,$C4,$7F,$0C,$8B,$8F,$E6,$27,$F8,$A7,$C0,$FF,$92
	.dc.b	$F9,$FF,$CF,$77,$FE,$17,$6B,$FC,$85,$AB,$CF,$7C,$CB,$E4,$7F,$A5
	.dc.b	$B6,$7F,$AD,$B5,$FF,$34,$D7,$FA,$23,$1F,$E4,$5F,$25,$FD,$14,$93
	.dc.b	$C2,$EB,$FC,$92,$2C,$FF,$9D,$6F,$A3,$FB,$2F,$91,$F5,$12,$BF,$C7
	.dc.b	$18,$23,$FA,$A7,$B7,$FD,$4C,$93,$FC,$BA,$3F,$9B,$79,$3F,$0C,$91
	.dc.b	$C9,$FE,$B9,$EF,$FF,$5B,$24,$7F,$36,$8A,$FF,$3A,$8B,$87,$F0,$69
	.dc.b	$BC,$7F,$D2,$FC,$CF,$F1,$DB,$6B,$C3,$7F,$D0,$38,$BE,$A1,$1F,$5B
	.dc.b	$FD,$B7,$9B,$FB,$61,$23,$81,$FC,$E7,$E9,$3F,$8E,$BF,$DD,$5A,$FF
	.dc.b	$B8,$87,$42,$33,$1F,$A3,$6B,$A6,$C6,$FC,$6F,$F0,$2C,$60,$DF,$3B
	.dc.b	$A1,$F1,$7F,$9D,$26,$1E,$5B,$FF,$1A,$DF,$EC,$A3,$B7,$6F,$42,$3F
	.dc.b	$E3,$D8,$D6,$3E,$79,$CC,$FB,$5F,$F5,$26,$0F,$98,$FD,$61,$27,$E2
	.dc.b	$4D,$7F,$67,$17,$87,$FA,$EE,$1F,$BB,$DF,$F8,$27,$94,$F9,$5F,$F1
	.dc.b	$1B,$70,$FF,$28,$35,$FF,$11,$35,$F9,$D9,$FF,$62,$9E,$24,$8E,$2F
	.dc.b	$F8,$9F,$ED,$3F,$C4,$C4,$8F,$EE,$55,$D7,$1F,$E1,$5F,$49,$FE,$9E
	.dc.b	$F0,$14,$F8,$77,$D2,$7D,$DA,$D4,$E6,$31,$F1,$1F,$E4,$FA,$DC,$FE
	.dc.b	$EF,$B1,$F3,$9F,$85,$9C,$CB,$0E,$71,$86,$E1,$D3,$1F,$1C,$F4,$F0
	.dc.b	$58,$F4,$C3,$37,$B7,$F2,$47,$AB,$0F,$3B,$86,$BB,$31,$EE,$FF,$B1
	.dc.b	$23,$FC,$AD,$F0,$9C,$A7,$94,$E4,$91,$E0,$8C,$B7,$07,$D2,$7F,$19
	.dc.b	$E3,$D5,$7B,$EF,$FD,$B3,$9F,$BD,$B9,$7C,$D6,$CF,$F2,$3B,$E4,$FC
	.dc.b	$F3,$B9,$7A,$9D,$CF,$98,$FE,$8D,$C5,$FE,$43,$6E,$7F,$D4,$FE,$91
	.dc.b	$E2,$E3,$16,$3C,$AF,$1D,$0B,$DE,$1F,$EA,$36,$DF,$F9,$D5,$AF,$E5
	.dc.b	$1A,$F7,$C3,$FA,$B3,$67,$94,$62,$3F,$83,$CF,$07,$FA,$A7,$D3,$79
	.dc.b	$71,$FE,$00,$F9,$39,$DF,$49,$9E,$E4,$D2,$65,$F9,$E9,$B8,$7F,$6E
	.dc.b	$BE,$33,$FA,$E9,$F1,$67,$FE,$4C,$6C,$EA,$BF,$63,$18,$25,$71,$3F
	.dc.b	$DF,$58,$7F,$58,$7F,$BA,$F8,$FE,$7F,$8C,$F3,$36,$3F,$7B,$1D,$88
	.dc.b	$E0,$97,$FC,$02,$32,$FE,$73,$3F,$BF,$F1,$5F,$17,$87,$FD,$97,$7F
	.dc.b	$E7,$CF,$8C,$C4,$FC,$9E,$AF,$F6,$96,$1F,$DB,$E7,$E8,$61,$FA,$A6
	.dc.b	$CE,$A7,$F0,$78,$3E,$23,$B1,$CF,$E8,$AE,$5F,$EA,$F8,$B5,$FD,$CA
	.dc.b	$6F,$2F,$FA,$4E,$6D,$7F,$7F,$9B,$5F,$90,$C1,$1E,$22,$4F,$F6,$E7
	.dc.b	$CF,$83,$E6,$7E,$01,$ED,$E1,$B5,$FE,$9C,$7D,$64,$7F,$BB,$BC,$BD
	.dc.b	$6F,$90,$7E,$87,$B3,$FC,$1F,$F4,$58,$AF,$F6,$54,$DC,$FF,$DF,$3C
	.dc.b	$4A,$C6,$DF,$E0,$77,$94,$66,$FF,$D5,$BE,$0E,$27,$E9,$3F,$38,$91
	.dc.b	$FE,$0A,$7D,$BF,$0F,$F6,$CF,$B1,$F3,$36,$7F,$DB,$91,$71,$FF,$0F
	.dc.b	$CD,$C7,$FC,$8B,$E4,$FE,$8C,$FD,$47,$C3,$3E,$0B,$18,$FF,$92,$26
	.dc.b	$E1,$FE,$BC,$9B,$7F,$EB,$2F,$23,$FD,$3D,$7F,$EB,$76,$D2,$7F,$CA
	.dc.b	$5E,$CB,$E5,$3F,$D8,$57,$BF,$3B,$F8,$64,$E7,$D3,$FD,$F3,$67,$FC
	.dc.b	$E5,$5C,$30,$FF,$37,$35,$FF,$36,$35,$FF,$48,$35,$4F,$37,$8D,$7E
	.dc.b	$73,$BB,$CF,$8F,$F9,$A3,$1F,$E3,$7F,$44,$F1,$FF,$84,$44,$CF,$FC
	.dc.b	$D6,$EA,$31,$8F,$FB,$A3,$E2,$3F,$5C,$7E,$AF,$CF,$B7,$FD,$9B,$1F
	.dc.b	$F7,$84,$E2,$BF,$CA,$75,$67,$E3,$D3,$E2,$FE,$2B,$4E,$47,$F7,$01
	.dc.b	$F4,$4F,$67,$73,$FC,$DE,$73,$1F,$CA,$5A,$7F,$57,$2E,$1F,$D8,$9A
	.dc.b	$FF,$8D,$5A,$F7,$AE,$D1,$C5,$B1,$C5,$AF,$F8,$F7,$C3,$FE,$9F,$F0
	.dc.b	$FF,$C4,$4F,$E2,$51,$6B,$D6,$FF,$0A,$8C,$3F,$89,$8F,$F2,$B3,$8C
	.dc.b	$7F,$D3,$BF,$45,$FC,$EC,$FE,$A3,$ED,$BD,$DC,$5F,$FA,$82,$6D,$FF
	.dc.b	$AA,$2B,$53,$93,$C3,$8F,$F6,$4F,$82,$FE,$4C,$7F,$5C,$F8,$27,$C3
	.dc.b	$62,$BC,$FE,$72,$73,$18,$6E,$3E,$91,$F1,$2E,$E6,$9D,$71,$98,$71
	.dc.b	$7F,$B7,$56,$E7,$37,$FE,$B8,$FB,$1F,$EB,$47,$EE,$2A,$DF,$FA,$FA
	.dc.b	$2C,$F8,$74,$F0,$56,$37,$FF,$62,$43,$E9,$7F,$DF,$18,$7C,$26,$DF
	.dc.b	$1F,$29,$EC,$FF,$6D,$9C,$E1,$C9,$B7,$FE,$42,$C3,$A3,$FD,$3E,$B8
	.dc.b	$3F,$DC,$7C,$68,$FE,$00,$FF,$06,$F4,$63,$2F,$91,$18,$3F,$2D,$EA
	.dc.b	$FF,$98,$5B,$3F,$9B,$8C,$4F,$E6,$5A,$FE,$65,$AF,$E9,$DA,$A7,$B3
	.dc.b	$FD,$9F,$1F,$F4,$44,$7C,$3E,$EF,$E8,$1F,$48,$F8,$5F,$85,$1F,$E2
	.dc.b	$0F,$AF,$79,$3F,$E0,$46,$DF,$F9,$0F,$9D,$18,$CB,$FC,$9F,$18,$AF
	.dc.b	$F0,$E4,$3D,$EF,$F1,$CF,$0F,$FB,$DD,$B8,$7F,$9B,$BC,$E6,$9F,$9A
	.dc.b	$BF,$C7,$1C,$7E,$1B,$FA,$B7,$D7,$BE,$5A,$9E,$AE,$EB,$8F,$EC,$31
	.dc.b	$E7,$1F,$E2,$D6,$C7,$F9,$6D,$B8,$7F,$99,$9A,$FF,$AE,$5A,$F9,$71
	.dc.b	$FF,$6F,$4F,$16,$C7,$16,$BF,$C1,$79,$7D,$5A,$7E,$51,$B5,$FE,$8C
	.dc.b	$5F,$1F,$FC,$7E,$BF,$39,$1A,$7F,$2B,$19,$7F,$D9,$91,$C1,$FF,$24
	.dc.b	$36,$7F,$CE,$CD,$8F,$F4,$0B,$5E,$33,$CD,$F1,$35,$EE,$7F,$89,$3E
	.dc.b	$05,$F2,$BF,$E5,$16,$DF,$F1,$CB,$7D,$33,$D2,$D5,$FF,$9A,$E7,$31
	.dc.b	$E6,$9D,$57,$99,$B8,$79,$9A,$FB,$9A,$FC,$76,$BF,$B8,$6D,$76,$9F
	.dc.b	$F4,$B7,$5B,$E4,$DF,$25,$FE,$69,$6C,$FF,$24,$F9,$3F,$EC,$4F,$F2
	.dc.b	$68,$CE,$3F,$C3,$67,$2F,$F9,$3D,$AF,$F4,$16,$BF,$C1,$5A,$FF,$A2
	.dc.b	$9B,$66,$93,$CD,$FF,$28,$FC,$E3,$E4,$3F,$CF,$CD,$C7,$F9,$CD,$FF
	.dc.b	$12,$C3,$FD,$2B,$39,$7F,$91,$47,$F1,$C7,$FA,$A3,$5F,$E3,$0D,$7F
	.dc.b	$D3,$AD,$3F,$D3,$F3,$FE,$62,$70,$47,$FC,$CB,$9F,$FB,$61,$FE,$96
	.dc.b	$D9,$FE,$65,$EA,$3E,$93,$FD,$2D,$39,$E3,$FE,$67,$39,$BF,$AA,$DB
	.dc.b	$FF,$3D,$35,$FE,$D8,$D7,$F9,$83,$5F,$F5,$AB,$70,$FF,$18,$61,$FE
	.dc.b	$7A,$8C,$71,$7F,$B0,$6F,$FB,$0B,$6F,$FA,$44,$78,$70,$FE,$8F,$39
	.dc.b	$7F,$A4,$C7,$52,$5F,$FA,$19,$F0,$7C,$FB,$E5,$FF,$D6,$8D,$8F,$E2
	.dc.b	$1F,$48,$F9,$1C,$D4,$9F,$EE,$75,$BC,$BF,$B5,$74,$B0,$E1,$5A,$BF
	.dc.b	$F6,$07,$99,$F2,$E3,$FA,$93,$6F,$FD,$AE,$FE,$AB,$3F,$EC,$2E,$76
	.dc.b	$FE,$EB,$EE,$F2,$39,$F4,$5C,$3F,$C2,$B8,$09,$BD,$E6,$7C,$BF,$F0
	.dc.b	$AA,$C3,$BA,$95,$87,$F8,$AC,$D7,$27,$EC,$E7,$93,$4E,$42,$78,$3F
	.dc.b	$DD,$26,$B9,$1C,$9E,$47,$B9,$8A,$E6,$EE,$34,$FF,$2F,$9F,$74,$62
	.dc.b	$3D,$E7,$8C,$D1,$4C,$DE,$F9,$3F,$B6,$46,$4F,$24,$66,$BC,$F8,$7A
	.dc.b	$AB,$73,$DE,$BD,$47,$98,$A7,$A2,$70,$F9,$0F,$8B,$F4,$4F,$98,$7D
	.dc.b	$27,$18,$C3,$63,$33,$F0,$A7,$2D,$EB,$FF,$4C,$37,$87,$94,$62,$3B
	.dc.b	$B5,$FB,$D3,$FC,$AA,$7E,$4E,$B8,$FA,$4F,$C1,$6D,$F3,$A9,$18,$9E
	.dc.b	$38,$78,$2B,$4D,$8E,$58,$74,$3F,$25,$7C,$9C,$07,$C4,$FD,$F3,$73
	.dc.b	$CD,$E3,$7E,$16,$9C,$A7,$8D,$F8,$B0,$C5,$6E,$BB,$36,$B7,$87,$AA
	.dc.b	$B7,$50,$F1,$1E,$5C,$3C,$5C,$3F,$AB,$F0,$7E,$F3,$67,$FA,$A8,$8C
	.dc.b	$B4,$CB,$82,$38,$3F,$C3,$1F,$48,$F7,$3F,$15,$D0,$8C,$4E,$A7,$31
	.dc.b	$88,$FF,$0F,$35,$FF,$0F,$35,$FF,$19,$35,$4F,$59,$EF,$9B,$E5,$69
	.dc.b	$2B,$FC,$59,$19,$6F,$8F,$C0,$8E,$0D,$FC,$BF,$42,$38,$21,$FE,$35
	.dc.b	$94,$6A,$30,$FC,$2F,$AC,$FF,$2B,$C7,$F8,$F5,$AF,$F8,$F5,$AF,$F8
	.dc.b	$F5,$AF,$F9,$47,$EB,$1F,$39,$F3,$1B,$1F,$E4,$76,$BF,$E4,$76,$BF
	.dc.b	$E4,$76,$BC,$E3,$91,$1E,$5F,$2D,$D4,$8E,$0E,$DF,$1B,$E1,$23,$82
	.dc.b	$1E,$17,$2D,$3F,$90,$F1,$3F,$03,$7C,$0F,$F0,$F8,$C3,$63,$33,$C1
	.dc.b	$AF,$F2,$AF,$A3,$FB,$4D,$DF,$F8,$AC,$62,$3F,$9C,$B5,$FE,$88,$D7
	.dc.b	$FA,$63,$53,$4D,$70,$BE,$5D,$ED,$FE,$71,$B5,$FE,$77,$6B,$FE,$77
	.dc.b	$6B,$E2,$F2,$FC,$A4,$7F,$97,$5A,$31,$1E,$D6,$BC,$4F,$F4,$27,$D1
	.dc.b	$3E,$5B,$F5,$AD,$8C,$C6,$0F,$E6,$FC,$FF,$DF,$E3,$F9,$E3,$5F,$EA
	.dc.b	$6D,$7F,$AF,$B5,$FE,$D4,$D7,$FB,$93,$5F,$DE,$79,$5D,$8F,$DF,$FE
	.dc.b	$8C,$F1,$E9,$27,$FD,$39,$19,$6F,$8F,$C0,$8E,$0F,$3F,$57,$A8,$8E
	.dc.b	$0E,$7E,$F7,$80,$8E,$0F,$EB,$5F,$89,$BE,$4F,$FA,$CB,$63,$31,$82
	.dc.b	$38,$21,$FD,$E2,$38,$3F,$DE,$BF,$25,$FD,$5B,$FE,$15,$6A,$69,$AF
	.dc.b	$07,$FE,$00,$7B,$1F,$CC,$3E,$03,$FD,$8D,$5F,$E8,$46,$BF,$EE,$38
	.dc.b	$D1,$1D,$35,$FF,$74,$71,$73,$23,$CD,$AE,$8C,$10,$FF,$09,$FE,$0E
	.dc.b	$F8,$2F,$EF,$AD,$8F,$F0,$13,$5D,$C6,$88,$E9,$AF,$4F,$1F,$99,$1E
	.dc.b	$6F,$F4,$4F,$A7,$FF,$04,$47,$FB,$7D,$AF,$FB,$7D,$AF,$FB,$7D,$AF
	.dc.b	$FB,$CD,$AF,$F4,$26,$F5,$1A,$6B,$36,$7F,$C0,$5B,$D7,$F4,$C9,$7F
	.dc.b	$4D,$AF,$F9,$0A,$36,$E4,$FC,$6F,$A5,$F8,$C2,$38,$21,$F8,$97,$2D
	.dc.b	$3F,$C5,$DF,$A5,$FF,$6D,$7F,$C8,$AD,$7F,$CA,$2D,$7F,$CB,$8D,$53
	.dc.b	$EC,$E3,$FC,$61,$23,$F1,$8D,$7D,$58,$23,$11,$FE,$67,$9E,$4F,$F9
	.dc.b	$5B,$D3,$FE,$D5,$E6,$FF,$90,$DB,$19,$CB,$FC,$BE,$D9,$D4,$62,$3F
	.dc.b	$CD,$33,$C1,$B1,$97,$FC,$ED,$19,$7F,$CD,$FF,$A6,$3E,$23,$FC,$9E
	.dc.b	$D8,$FF,$2A,$35,$FF,$35,$B5,$E3,$1C,$08,$F0,$79,$FD,$D4,$8E,$6D
	.dc.b	$7F,$21,$D7,$F9,$52,$3F,$95,$6B,$A3,$04,$3F,$D0,$FC,$0F,$C4,$7F
	.dc.b	$CC,$ED,$8F,$F1,$1C,$60,$8E,$08,$7F,$A1,$A3,$83,$7E,$21,$F2,$FF
	.dc.b	$E2,$C6,$C6,$63,$07,$6E,$29,$1C,$EF,$FD,$88,$D8,$E5,$E0,$C8,$EC
	.dc.b	$D7,$D7,$7F,$D3,$5B,$19,$7F,$D5,$D8,$27,$51,$FE,$A8,$FC,$93,$FD
	.dc.b	$18,$FF,$AE,$5A,$9A,$4F,$8C,$FF,$46,$A6,$92,$3E,$17,$97,$F0,$48
	.dc.b	$E0,$F7,$7C,$07,$CE,$23,$82,$1F,$5B,$34,$6A,$30,$D7,$DB,$1A,$78
	.dc.b	$98,$8F,$F2,$EF,$EC,$9F,$E4,$08,$FB,$0D,$7F,$06,$D7,$E3,$B5,$FF
	.dc.b	$6E,$65,$F2,$7D,$4F,$B8,$48,$FC,$03,$FE,$B3,$F8,$A7,$E8,$DF,$9F
	.dc.b	$9F,$F7,$24,$63,$8F,$53,$99,$4E,$5F,$F6,$8B,$5F,$F6,$EB,$5F,$F7
	.dc.b	$43,$5E,$B1,$CE,$1F,$A5,$BE,$BF,$C8,$DA,$EA,$D7,$AB,$5F,$F7,$A7
	.dc.b	$C6,$3F,$57,$7D,$0F,$FB,$3E,$30,$D8,$CC,$FF,$BA,$27,$2D,$8C,$BF
	.dc.b	$EF,$68,$CB,$63,$2F,$81,$A8,$43,$C0,$7C,$5F,$8D,$B3,$E1,$6B,$DD
	.dc.b	$AF,$B1,$AF,$59,$E9,$3C,$C9,$1E,$9F,$A5,$F7,$11,$C1,$0F,$53,$96
	.dc.b	$9E,$33,$CC,$43,$07,$FB,$E7,$EA,$FF,$D4,$55,$EF,$F8,$47,$D5,$FC
	.dc.b	$1A,$56,$94,$E9,$F8,$D1,$98,$C7,$13,$E9,$BF,$C7,$55,$C2,$FA,$5E
	.dc.b	$DB,$E1,$7C,$E9,$D1,$D1,$C6,$1D,$2F,$9B,$E1,$9F,$2C,$F1,$7D,$06
	.dc.b	$31,$F7,$46,$63,$BC,$7C,$1F,$2D,$F1,$14,$93,$C8,$F8,$31,$A7,$19
	.dc.b	$7F,$9C,$E7,$E4,$65,$EC,$AD,$63,$F0,$BE,$AF,$FB,$57,$C5,$6B,$1F
	.dc.b	$8F,$5A,$BE,$18,$66,$7B,$BB,$D5,$87,$29,$F8,$A6,$8F,$8F,$9F,$F7
	.dc.b	$75,$8F,$E8,$FC,$9F,$CD,$36,$32,$7E,$8E,$32,$FE,$97,$C6,$7A,$9A
	.dc.b	$8F,$D9,$C6,$2B,$F5,$CD,$8E,$13,$B7,$F3,$91,$97,$8C,$65,$FD,$5C
	.dc.b	$65,$DC,$66,$DE,$BF,$FA,$16,$FF,$78,$DA,$FD,$D3,$5F,$E0,$ED,$7F
	.dc.b	$58,$D7,$F7,$2D,$9E,$BE,$3F,$78,$C6,$0F,$A7,$F1,$37,$0F,$7C,$60
	.dc.b	$8E,$09,$87,$C9,$8F,$77,$5B,$F5,$27,$D2,$61,$B7,$FC,$B9,$B7,$FC
	.dc.b	$B9,$AF,$F2,$EF,$D2,$BF,$B4,$4F,$F3,$46,$C7,$F3,$46,$BB,$F2,$3C
	.dc.b	$31,$88,$FE,$82,$D7,$F9,$FB,$70,$39,$E1,$FE,$ED,$F0,$3F,$5D,$B8
	.dc.b	$7F,$83,$63,$18,$7F,$BD,$23,$FA,$DF,$D4,$EF,$C5,$34,$91,$FE,$0D
	.dc.b	$F2,$FA,$79,$1F,$ED,$58,$C1,$34,$FE,$E1,$3A,$24,$66,$3F,$B7,$C6
	.dc.b	$1D,$C6,$5F,$EC,$FF,$34,$F9,$BF,$53,$67,$FC,$50,$D9,$FF,$19,$B5
	.dc.b	$FE,$13,$F8,$EF,$F2,$29,$FE,$05,$08,$C3,$FC,$15,$AF,$0E,$BF,$CF
	.dc.b	$3B,$FC,$06,$C9,$A4,$AF,$1F,$B7,$E6,$23,$82,$57,$9A,$32,$D8,$FF
	.dc.b	$1C,$7B,$DF,$30,$8E,$0F,$7F,$85,$F0,$88,$E0,$FF,$20,$79,$BF,$50
	.dc.b	$8E,$09,$5F,$E2,$38,$CB,$63,$FC,$A5,$F2,$DF,$C4,$1A,$7E,$25,$C4
	.dc.b	$7F,$98,$1A,$FF,$91,$1A,$FF,$87,$9A,$B9,$11,$86,$CF,$CE,$F0,$7F
	.dc.b	$DF,$0F,$D6,$6D,$7F,$39,$6B,$FD,$11,$AF,$F5,$D5,$F6,$3C,$C9,$3C
	.dc.b	$DB,$3C,$DA,$F3,$7F,$8E,$7A,$EF,$D6,$5F,$5B,$86,$DF,$FA,$9D,$B7
	.dc.b	$FE,$A0,$6B,$FE,$9E,$6B,$FE,$11,$AE,$95,$CC,$95,$FD,$57,$DB,$FF
	.dc.b	$6B,$FC,$2F,$F2,$D6,$C6,$7D,$D4,$E5,$B1,$98,$C1,$1C,$1F,$A0,$F9
	.dc.b	$AF,$F1,$8B,$7D,$2F,$CE,$C6,$32,$FE,$CC,$D7,$FB,$8A,$FB,$0D,$92
	.dc.b	$BA,$36,$78,$B5,$E2,$F9,$FE,$4A,$4F,$AD,$B3,$FE,$D6,$6B,$FE,$D5
	.dc.b	$C5,$18,$8F,$C6,$76,$FE,$4C,$95,$F9,$86,$C7,$E5,$5A,$FE,$55,$FD
	.dc.b	$DF,$B5,$F8,$88,$FF,$22,$35,$34,$C6,$BF,$D7,$63,$FC,$94,$D7,$F0
	.dc.b	$BE,$5F,$F0,$68,$C1,$1C,$12,$FE,$D7,$C5,$FD,$92,$38,$25,$7F,$BE
	.dc.b	$23,$2D,$8F,$C1,$61,$F8,$12,$38,$24,$7F,$C0,$08,$CB,$63,$F9,$0F
	.dc.b	$E5,$AF,$8D,$FE,$3A,$D8,$CC,$60,$FE,$43,$E9,$3E,$0A,$92,$7F,$95
	.dc.b	$C6,$5B,$19,$BF,$F1,$84,$E5,$B1,$FC,$EB,$DD,$95,$FC,$E9,$B3,$FE
	.dc.b	$37,$F9,$C7,$7F,$F3,$26,$D7,$F3,$A6,$CE,$E7,$FA,$2E,$32,$BF,$A2
	.dc.b	$B6,$7F,$A2,$FD,$FF,$F4,$A2,$4F,$F4,$A7,$B7,$8B,$23,$B3,$67,$C8
	.dc.b	$D7,$FC,$AB,$87,$F4,$F7,$FC,$DC,$D7,$F9,$07,$D0,$BE,$63,$FB,$93
	.dc.b	$6F,$FE,$00,$36,$7F,$CB,$FD,$E8,$C4,$7F,$A7,$FC,$2F,$EA,$7E,$A9
	.dc.b	$A4,$BF,$F5,$63,$67,$FD,$54,$D7,$FD,$F3,$F4,$5C,$67,$FC,$C9,$82
	.dc.b	$31,$3E,$26,$BC,$2B,$C3,$EC,$CB,$FF,$5D,$36,$7F,$D7,$0D,$7F,$D6
	.dc.b	$CD,$7F,$D6,$5F,$28,$FC,$0D,$F6,$3F,$E4,$B8,$C3,$63,$37,$FE,$79
	.dc.b	$9C,$BE,$8F,$9D,$7C,$7F,$F8,$75,$B1,$98,$C1,$1C,$13,$0F,$9F,$3C
	.dc.b	$1B,$F3,$1C,$C9,$3D,$DE,$DE,$8C,$8F,$13,$67,$9B,$5F,$F5,$86,$CB
	.dc.b	$FF,$13,$46,$AB,$E8,$FD,$EF,$BA,$3F,$02,$D7,$8B,$5F,$F6,$D5,$23
	.dc.b	$11,$FD,$B7,$EB,$DF,$37,$FE,$3F,$6D,$FD,$E6,$CF,$E3,$9A,$F9,$9A
	.dc.b	$F6,$6B,$D6,$7A,$1D,$7E,$42,$57,$56,$CF,$5C,$FC,$44,$9F,$13,$FE
	.dc.b	$53,$FA,$3F,$F0,$11,$23,$04,$3F,$2A,$E5,$A7,$F6,$5F,$33,$FA,$B5
	.dc.b	$7F,$BF,$7C,$C4,$62,$11,$8F,$2F,$81,$2B,$FB,$85,$6E,$75,$18,$9F
	.dc.b	$F2,$0D,$7F,$BA,$C7,$C9,$3E,$53,$FC,$25,$49,$CC,$7F,$20,$69,$FE
	.dc.b	$15,$8D,$4F,$18,$FF,$6E,$F8,$78,$AE,$71,$C0,$FF,$7C,$46,$5E,$46
	.dc.b	$4F,$24,$F3,$7C,$5E,$8F,$FB,$12,$31,$1A,$AE,$B1,$88,$FF,$80,$27
	.dc.b	$39,$E1,$3E,$99,$E3,$18,$9F,$39,$BB,$F2,$3E,$63,$07,$94,$EA,$7F
	.dc.b	$46,$3F,$CE,$26,$0F,$8C,$7F,$BF,$BE,$7D,$F4,$2F,$A2,$FF,$3B,$C6
	.dc.b	$1B,$19,$9F,$F7,$E4,$E5,$BF,$44,$E9,$E2,$D8,$CC,$60,$8F,$CF,$9F
	.dc.b	$DB,$3E,$C6,$B9,$9E,$59,$7C,$F9,$46,$33,$FC,$15,$27,$8D,$6F,$9F
	.dc.b	$2F,$1B,$E1,$5F,$89,$AE,$78,$78,$6F,$78,$74,$9E,$6D,$BE,$37,$DD
	.dc.b	$B3,$A8,$E9,$5C,$9E,$D3,$A3,$71,$98,$F4,$36,$7C,$71,$F9,$AD,$FF
	.dc.b	$3D,$E0,$FF,$AA,$9B,$1F,$E5,$93,$FA,$F1,$63,$4E,$0F,$8A,$7E,$D0
	.dc.b	$C1,$F2,$23,$9B,$A6,$BF,$CA,$BE,$1D,$EC,$7F,$AE,$BA,$91,$89,$FE
	.dc.b	$B6,$D7,$FB,$5B,$5E,$53,$C6,$31,$1F,$CB,$1A,$FF,$A5,$DB,$3F,$6F
	.dc.b	$C0,$F0,$63,$FD,$C6,$BF,$B4,$B5,$FE,$F2,$D7,$FA,$0A,$F8,$9E,$24
	.dc.b	$8E,$2F,$4F,$4A,$47,$46,$BD,$1A,$F4,$6B,$FD,$13,$E8,$9F,$4F,$F1
	.dc.b	$5B,$1F,$DF,$9A,$FF,$83,$9A,$FF,$56,$FA,$3F,$E9,$2F,$F8,$11,$AF
	.dc.b	$F8,$51,$AE,$FB,$BE,$94,$62,$3F,$AA,$B5,$FF,$60,$37,$33,$47,$6F
	.dc.b	$22,$4F,$85,$EB,$EA,$C8,$F2,$35,$F2,$35,$F5,$35,$FF,$07,$FD,$1B
	.dc.b	$E1,$BF,$E0,$0F,$3A,$31,$3F,$E0,$D6,$BF,$96,$8D,$4F,$F1,$8F,$A3
	.dc.b	$FB,$93,$FB,$E6,$BF,$D0,$5A,$9A,$4E,$1E,$44,$8F,$F1,$8F,$AD,$FE
	.dc.b	$73,$6F,$A3,$F9,$D8,$C4,$7F,$91,$1A,$FE,$AA,$9E,$5F,$94,$91,$FA
	.dc.b	$E3,$C8,$E1,$C4,$70,$74,$79,$E3,$4F,$F8,$63,$E8,$DE,$4F,$EB,$DB
	.dc.b	$1F,$E4,$86,$BF,$E5,$86,$BF,$BE,$E7,$FF,$6B,$BD,$9A,$F9,$5A,$EF
	.dc.b	$DD,$FE,$8B,$18,$8F,$E5,$6D,$7F,$B7,$35,$FE,$03,$1F,$C0,$FE,$8D
	.dc.b	$F2,$14,$93,$FE,$5E,$8C,$B7,$DB,$E0,$47,$07,$9F,$AB,$C6,$47,$07
	.dc.b	$3F,$A3,$F0,$11,$C1,$F3,$3E,$8D,$F0,$BF,$D8,$1B,$19,$8C,$11,$C1
	.dc.b	$0F,$B5,$1C,$1F,$B9,$F4,$7F,$AD,$4D,$27,$5F,$F3,$5F,$81,$FF,$54
	.dc.b	$CF,$E0,$FC,$24,$62,$3F,$97,$35,$FF,$78,$7B,$9C,$48,$F1,$8D,$3F
	.dc.b	$E2,$F6,$BF,$E2,$E7,$F1,$8F,$4E,$3F,$11,$1E,$4D,$74,$60,$FC,$9F
	.dc.b	$D1,$FF,$9F,$1B,$EA,$FF,$82,$23,$11,$F9,$D6,$BF,$D3,$1A,$FF,$72
	.dc.b	$6B,$FE,$35,$6A,$69,$FE,$23,$75,$E4,$F7,$23,$FE,$A6,$FA,$3F,$F0
	.dc.b	$AB,$7C,$CF,$EA,$51,$88,$FF,$5C,$B5,$FF,$40,$B5,$FF,$3C,$B5,$78
	.dc.b	$68,$C3,$67,$ED,$76,$CA,$FA,$4D,$7E,$CB,$5F,$ED,$59,$FF,$14,$D7
	.dc.b	$EF,$CF,$E4,$7E,$8E,$57,$E5,$5A,$FE,$39,$AF,$F7,$6E,$14,$62,$3F
	.dc.b	$C3,$BF,$39,$FD,$E1,$BE,$5F,$F8,$F2,$31,$1F,$E6,$36,$E1,$F9,$0E
	.dc.b	$6F,$CA,$12,$7F,$18,$D7,$F2,$6D,$7D,$6F,$F1,$6F,$53,$FD,$C1,$E7
	.dc.b	$DC,$FF,$CE,$75,$FC,$7E,$72,$D8,$CB,$FE,$10,$8C,$B6,$32,$FF,$8E
	.dc.b	$23,$2D,$8C,$BF,$E3,$CF,$97,$7C,$17,$F9,$69,$B1,$FC,$A9,$AF,$F3
	.dc.b	$16,$BF,$E2,$06,$BC,$E3,$91,$1E,$5F,$07,$D4,$8E,$0F,$C9,$77,$7E
	.dc.b	$40,$8E,$08,$7E,$49,$CB,$4F,$F7,$FF,$C8,$7F,$9C,$B8,$CD,$24,$7F
	.dc.b	$8B,$5A,$F0,$30,$7F,$8C,$7E,$49,$ED,$D2,$47,$F5,$1A,$4E,$A3,$C3
	.dc.b	$C5,$D4,$8E,$0E,$11,$FD,$29,$FF,$6F,$B5,$FD,$3B,$67,$AF,$8B,$E3
	.dc.b	$24,$F2,$6B,$FD,$8F,$E8,$FF,$7C,$9A,$4E,$9F,$F4,$AA,$FA,$3A,$49
	.dc.b	$FE,$CF,$19,$7C,$FE,$77,$C3,$23,$83,$E7,$FD,$1F,$CB,$23,$82,$1F
	.dc.b	$E6,$4C,$7E,$9C,$6A,$31,$AA,$31,$1F,$E8,$E9,$DB,$C8,$C4,$ED,$FA
	.dc.b	$5F,$2A,$F5,$7F,$CD,$0D,$8C,$C6,$0F,$B9,$EB,$C8,$FA,$6D,$7F,$CF
	.dc.b	$3F,$2E,$FA,$3F,$CD,$C2,$31,$CB,$FE,$0C,$6C,$7F,$84,$DA,$F2,$9E
	.dc.b	$31,$88,$FF,$54,$B5,$FF,$76,$35,$ED,$F4,$4F,$A3,$EC,$D8,$EC,$D7
	.dc.b	$CE,$D7,$FD,$6B,$F1,$2F,$A3,$7C,$27,$F8,$8A,$30,$D8,$CC,$FF,$8A
	.dc.b	$E7,$2D,$8C,$BF,$E9,$C8,$CB,$63,$2F,$81,$A8,$43,$C1,$7D,$1F,$8D
	.dc.b	$B3,$E1,$6B,$DD,$AF,$B1,$AF,$59,$E9,$3C,$C9,$1E,$9F,$A3,$F7,$11
	.dc.b	$C1,$0F,$53,$96,$9E,$33,$CC,$43,$07,$F9,$33,$E8,$9E,$07,$BA,$1F
	.dc.b	$0E,$F9,$CF,$F7,$3D,$6B,$1F,$89,$39,$AF,$8D,$39,$78,$D7,$C9,$E6
	.dc.b	$7D,$1F,$F8,$F2,$72,$70,$C7,$C1,$87,$38,$61,$CE,$30,$DC,$3A,$5F
	.dc.b	$39,$CD,$FA,$6F,$FD,$B3,$7E,$36,$C6,$5F,$5B,$5F,$4F,$9E,$F8,$9F
	.dc.b	$88,$D8,$FF,$3B,$57,$F9,$EA,$72,$48,$F5,$9E,$89,$E1,$7F,$17,$27
	.dc.b	$D1,$FF,$B3,$7E,$2B,$E4,$E9,$F3,$2F,$6D,$C7,$38,$F9,$1B,$87,$2F
	.dc.b	$29,$86,$AB,$C1,$5C,$36,$FF,$6D,$46,$8F,$D0,$57,$E6,$EB,$F4,$CD
	.dc.b	$8C,$9B,$3F,$63,$E4,$3D,$8C,$47,$E5,$E3,$15,$FA,$F6,$C7,$EB,$5A
	.dc.b	$F2,$9E,$2F,$EA,$A3,$2F,$38,$CD,$A7,$33,$FB,$0F,$6F,$FD,$0F,$7F
	.dc.b	$BD,$6C,$7E,$ED,$AF,$F0,$86,$BF,$AD,$6B,$CF,$C9,$EF,$18,$B3,$6D
	.dc.b	$F4,$7E,$46,$E1,$EF,$8C,$11,$C1,$30,$F8,$51,$F0,$3A,$DF,$44,$FA
	.dc.b	$3C,$36,$FF,$98,$36,$FF,$9C,$B5,$FE,$5F,$F4,$7F,$DD,$A7,$F9,$B3
	.dc.b	$63,$F9,$FB,$5D,$F9,$7E,$18,$C4,$7F,$36,$6B,$FD,$0D,$B8,$1C,$F0
	.dc.b	$FF,$81,$3E,$13,$EC,$B7,$0F,$F0,$0C,$63,$0F,$F7,$C4,$7F,$5A,$F9
	.dc.b	$AD,$F8,$C6,$92,$3F,$C3,$BE,$6F,$4F,$23,$FD,$B3,$18,$26,$9F,$DB
	.dc.b	$27,$44,$8C,$C7,$F6,$A8,$C3,$B8,$CB,$FC,$47,$E8,$DF,$33,$EA,$6C
	.dc.b	$FF,$88,$5B,$3F,$DF,$5A,$FF,$41,$3F,$A1,$7D,$1F,$F2,$39,$FF,$79
	.dc.b	$B6,$3F,$DF,$4D,$78,$07,$0F,$35,$F4,$7F,$A4,$8D,$36,$33,$3D,$A3
	.dc.b	$D1,$EF,$C8,$F4,$36,$BD,$0D,$7F,$C2,$16,$8C,$47,$F3,$3C,$BF,$DB
	.dc.b	$F5,$FE,$4E,$F8,$2F,$A4,$47,$E2,$78,$5F,$1C,$8E,$0F,$F2,$AF,$9B
	.dc.b	$F6,$48,$E0,$97,$FE,$24,$8C,$B6,$3F,$CA,$1F,$47,$FC,$69,$A7,$E3
	.dc.b	$9C,$47,$F8,$BD,$AF,$F8,$AD,$AF,$F4,$D6,$BF,$EB,$66,$F1,$9C,$C7
	.dc.b	$F3,$DC,$3E,$87,$7F,$C2,$8F,$AE,$66,$EB,$4C,$C3,$E8,$F8,$92,$77
	.dc.b	$19,$6C,$66,$78,$B5,$FF,$4D,$7D,$1B,$E8,$7F,$92,$FB,$68,$C5,$FF
	.dc.b	$A9,$DB,$5F,$EA,$56,$BC,$AB,$8C,$62,$3F,$C3,$15,$CE,$72,$7F,$32
	.dc.b	$F7,$5F,$00,$F8,$9C,$37,$1F,$87,$3A,$9C,$92,$33,$3B,$C3,$F3,$F5
	.dc.b	$B9,$CD,$7E,$77,$E8,$FF,$C7,$31,$F3,$1B,$5F,$41,$AF,$EA,$1A,$FE
	.dc.b	$C5,$7D,$1F,$32,$4F,$36,$CF,$36,$BC,$DF,$6F,$B1,$27,$D2,$D9,$F4
	.dc.b	$E3,$FE,$71,$9C,$92,$33,$3F,$95,$EB,$93,$F9,$A6,$CF,$E5,$DA,$FE
	.dc.b	$5D,$AF,$F0,$BF,$6B,$F2,$0F,$F0,$66,$B9,$E1,$FE,$09,$18,$8F,$F2
	.dc.b	$96,$BF,$D9,$1F,$C4,$79,$3F,$C2,$63,$04,$70,$4C,$3F,$37,$1F,$DB
	.dc.b	$36,$FC,$0F,$C7,$7E,$00,$93,$F7,$1B,$7F,$11,$AF,$F2,$39,$E3,$3F
	.dc.b	$82,$CB,$F1,$24,$70,$49,$F9,$B1,$96,$C7,$EB,$BE,$8D,$F2,$1F,$C8
	.dc.b	$DB,$19,$8C,$1F,$C9,$FD,$17,$C1,$D2,$4F,$EE,$23,$2D,$8C,$DF,$EF
	.dc.b	$A7,$2D,$8F,$E8,$9E,$EC,$AF,$E8,$8D,$9F,$E0,$BF,$45,$FC,$AE,$7F
	.dc.b	$82,$B6,$33,$DF,$FC,$EE,$31,$1F,$CC,$B1,$FE,$A0,$4A,$FE,$A0,$FF
	.dc.b	$53,$FA,$39,$1F,$D4,$DB,$3E,$0F,$13,$A1,$27,$C2,$D8,$D4,$60,$8F
	.dc.b	$FB,$5F,$E8,$DF,$11,$FD,$E9,$B3,$FF,$00,$5B,$3B,$9F,$EA,$9E,$97
	.dc.b	$FA,$12,$BF,$D6,$1E,$1E,$08,$E0,$95,$FE,$B1,$8C,$BF,$EE,$7F,$A3
	.dc.b	$E4,$D8,$FF,$05,$C6,$23,$FD,$CC,$D7,$FD,$83,$EB,$F8,$89,$5F,$EC
	.dc.b	$16,$C7,$FA,$FD,$AF,$FA,$F5,$FF,$27,$FC,$AB,$E8,$9F,$47,$86,$DF
	.dc.b	$F4,$09,$D4,$E4,$FF,$0F,$FC,$F3,$E4,$A9,$2B,$FA,$3F,$92,$9D,$46
	.dc.b	$1B,$97,$F8,$0A,$75,$5C,$63,$13,$A7,$FC,$1D,$3C,$FE,$6B,$C2,$4B
	.dc.b	$E8,$F9,$3D,$09,$1D,$DB,$5D,$1A,$F4,$BF,$F1,$6B,$99,$F4,$FD,$1F
	.dc.b	$C2,$8F,$33,$5F,$1B,$5F,$F5,$D3,$56,$0F,$A3,$7C,$B7,$F9,$19,$B7
	.dc.b	$F1,$9B,$5F,$0D,$AF,$C0,$6B,$FE,$C7,$6B,$D9,$B5,$DE,$BB,$7C,$8C
	.dc.b	$BE,$CD,$BE,$D9,$F8,$C9,$3E,$37,$FC,$B7,$F4,$7F,$E0,$62,$46,$08
	.dc.b	$7D,$E7,$2D,$3F,$B5,$79,$9F,$D5,$A9,$E1,$23,$0D,$F8,$AD,$92,$BF
	.dc.b	$B6,$CE,$89,$1F,$E4,$3A,$FF,$76,$8F,$93,$7C,$97,$F9,$33,$D2,$FF
	.dc.b	$5E,$46,$27,$FC,$9F,$E2,$1E,$15,$BC,$71,$5B,$9F,$F2,$E5,$62,$B5
	.dc.b	$1C,$0E,$65,$3A,$5F,$82,$3F,$95,$46,$4F,$F7,$45,$F3,$78,$DF,$77
	.dc.b	$26,$C5,$ED,$EB,$35,$F8,$33,$C6,$FC,$11,$D2,$FC,$F3,$D6,$39,$9E
	.dc.b	$39,$DB,$EE,$3F,$C7,$66,$8F,$E8,$C6,$2F,$FC,$7F,$F4,$2F,$96,$7D
	.dc.b	$1E,$1B,$5F,$E4,$69,$D4,$E4,$E3,$F4,$6E,$AA,$4A,$E1,$19,$6C,$7D
	.dc.b	$7F,$A4,$FA,$97,$FC,$2E,$71,$A7,$FA,$46,$31,$09,$DD,$F3,$6D,$F0
	.dc.b	$9F,$C1,$57,$09,$EB,$D1,$FE,$9B,$77,$3C,$F0,$ED,$87,$97,$5D,$CF
	.dc.b	$D5,$C3,$C3,$1C,$0E,$37,$A9,$E4,$78,$30,$ED,$87,$4C,$7C,$92,$C1
	.dc.b	$8F,$A6,$BD,$D7,$C2,$B9,$1C,$2B,$77,$ED,$E1,$FC,$F7,$C2,$7F,$AD
	.dc.b	$61,$C1,$FA,$19,$CD,$FE,$8E,$72,$FE,$D8,$FF,$79,$19,$9E,$0F,$F6
	.dc.b	$63,$06,$EB,$B4,$69,$B1,$93,$75,$FB,$1F,$0B,$2B,$C2,$7B,$58,$6D
	.dc.b	$70,$6B,$F8,$9C,$3F,$75,$D4,$F8,$7C,$E7,$EA,$C7,$A6,$3C,$C9,$87
	.dc.b	$A2,$F8,$47,$99,$FA,$1F,$13,$FE,$9C,$6C,$FF,$5A,$D5,$39,$6C,$7F
	.dc.b	$A0,$B2,$7D,$87,$B1,$B1,$F0,$1A,$FA,$4F,$E8,$B3,$C9,$C5,$BE,$91
	.dc.b	$F4,$BE,$36,$CF,$F6,$C6,$BF,$DA,$5A,$FF,$66,$6B,$F1,$9A,$FF,$54
	.dc.b	$6C,$F7,$9F,$67,$A7,$F0,$EC,$D3,$E3,$57,$83,$5F,$C3,$DB,$E9,$5D
	.dc.b	$FF,$25,$B9,$7F,$80,$9A,$FF,$7F,$6B,$FD,$F1,$AF,$DC,$6B,$FD,$A9
	.dc.b	$B3,$DE,$7F,$DA,$3B,$FF,$4C,$DB,$F1,$6F,$CB,$C3,$F0,$4D,$D7,$A5
	.dc.b	$7E,$52,$5F,$4A,$F8,$8F,$F6,$DC,$61,$B1,$9C,$FF,$C4,$4D,$7F,$A5
	.dc.b	$1F,$A9,$70,$FE,$BD,$AE,$63,$99,$F1,$78,$7F,$21,$A7,$E6,$31,$F8
	.dc.b	$11,$E5,$BE,$D8,$7C,$ED,$BF,$D0,$38,$FF,$8B,$1A,$FF,$8A,$5A,$FF
	.dc.b	$89,$9A,$FE,$99,$AF,$F8,$51,$BA,$FD,$EF,$43,$F4,$11,$AD,$5F,$57
	.dc.b	$FE,$18,$BF,$DC,$B5,$FC,$D4,$6B,$C5,$FD,$55,$79,$31,$FC,$2F,$D2
	.dc.b	$BC,$9F,$D6,$37,$0F,$F2,$2B,$5F,$F2,$13,$5F,$F1,$FB,$5F,$DF,$B5
	.dc.b	$FF,$16,$B7,$33,$7E,$EB,$F8,$64,$FE,$FD,$AF,$EB,$9B,$7E,$6B,$FE
	.dc.b	$05,$F4,$CF,$92,$FE,$04,$DB,$FF,$2D,$B5,$FF,$2C,$35,$FF,$2A,$B5
	.dc.b	$FE,$2A,$D7,$FC,$86,$DC,$4E,$7E,$3D,$F5,$CF,$7B,$F9,$4C,$61,$B1
	.dc.b	$9B,$FE,$35,$5E,$3C,$BF,$9C,$C7,$F2,$3E,$8F,$E6,$2D,$3F,$9B,$94
	.dc.b	$FE,$1F,$2F,$7F,$F9,$17,$D2,$7F,$8A,$E7,$FC,$F2,$D7,$FC,$EC,$D7
	.dc.b	$FC,$E6,$D7,$FA,$67,$C2,$3E,$2B,$F9,$EB,$63,$FA,$0B,$5F,$0F,$D3
	.dc.b	$F9,$BE,$9F,$F1,$91,$FA,$C0,$E1,$FD,$7B,$5F,$F1,$D3,$5D,$F2,$FF
	.dc.b	$86,$A3,$04,$7A,$47,$F9,$07,$77,$D7,$BE,$43,$E7,$B6,$7F,$AE,$CE
	.dc.b	$A7,$CD,$5E,$CF,$A1,$95,$EC,$7D,$FF,$07,$88,$FD,$2E,$69,$C7,$0F
	.dc.b	$F5,$58,$57,$C6,$F8,$59,$7F,$19,$F9,$3E,$F6,$23,$9C,$5B,$E5,$3F
	.dc.b	$7B,$EA,$3F,$CB,$11,$F8,$26,$F8,$5F,$C5,$9B,$7F,$E6,$17,$FD,$70
	.dc.b	$FE,$11,$CA,$69,$3D,$DF,$C9,$F8,$CF,$84,$A4,$AE,$5E,$8F,$E0,$3E
	.dc.b	$F6,$47,$E0,$1F,$C6,$FA,$F8,$8F,$8B,$17,$3F,$E9,$7D,$1F,$E4,$D9
	.dc.b	$FC,$3F,$DE,$7E,$28,$8F,$E2,$BE,$12,$BF,$3A,$2D,$7C,$D9,$FD,$0F
	.dc.b	$D1,$7F,$85,$EB,$F0,$2D,$7F,$1C,$D7,$F9,$DB,$5F,$F4,$33,$53,$49
	.dc.b	$EA,$7F,$9A,$3F,$5A,$7C,$77,$EC,$1B,$5F,$B0,$6B,$FE,$01,$39,$CF
	.dc.b	$E9,$BC,$39,$1F,$A6,$7F,$53,$E1,$62,$3F,$17,$17,$2F,$C5,$4F,$F0
	.dc.b	$6E,$6F,$E8,$FC,$BF,$9D,$6D,$BB,$11,$8F,$80,$6D,$FB,$DC,$7F,$C8
	.dc.b	$35,$AF,$47,$A1,$30,$FF,$1F,$DF,$0F,$73,$F8,$6F,$CE,$F4,$9F,$D6
	.dc.b	$87,$0F,$EC,$1A,$FE,$F5,$AF,$F6,$A6,$BF,$EB,$56,$BC,$F2,$FF,$4A
	.dc.b	$C6,$07,$C9,$F8,$89,$3F,$C7,$A3,$C3,$97,$B3,$A7,$FC,$62,$DF,$63
	.dc.b	$F3,$67,$F2,$0F,$2E,$DF,$F0,$0A,$DE,$FF,$C1,$DB,$1F,$EF,$CE,$7F
	.dc.b	$E6,$13,$FC,$82,$FF,$8F,$92,$7F,$C9,$DF,$2D,$FC,$6F,$1F,$E0,$13
	.dc.b	$99,$FE,$0A,$FF,$94,$67,$55,$FE,$5C,$5F,$55,$FE,$53,$6D,$7F,$1A
	.dc.b	$F3,$51,$88,$FE,$51,$18,$9F,$F2,$D7,$CD,$FF,$6D,$C7,$FC,$6E,$DC
	.dc.b	$3F,$CD,$2D,$B5,$F1,$7F,$9B,$FD,$8F,$F3,$A2,$3F,$E6,$99,$E1,$EA
	.dc.b	$3C,$65,$6A,$BF,$90,$BF,$E6,$DA,$ED,$0F,$AA,$7D,$5F,$FB,$41,$B1
	.dc.b	$F8,$A6,$BF,$CE,$1A,$FF,$45,$F3,$5F,$57,$FD,$24,$91,$FD,$7A,$78
	.dc.b	$EB,$FE,$56,$DF,$73,$F2,$5A,$FC,$86,$BF,$E3,$CD,$7E,$6D,$FF,$93
	.dc.b	$B4,$7C,$A7,$F5,$C2,$57,$DE,$9A,$FF,$40,$8B,$87,$86,$6E,$FF,$D9
	.dc.b	$A6,$D7,$D0,$C7,$C7,$7F,$5B,$DC,$FB,$24,$9F,$C0,$6D,$FE,$A2,$7C
	.dc.b	$D5,$FE,$33,$AF,$77,$9D,$F8,$B3,$E8,$5F,$F8,$2B,$EA,$BF,$A9,$5F
	.dc.b	$F6,$2E,$E4,$62,$3F,$B8,$46,$3C,$FF,$DD,$B6,$38,$4E,$DF,$F3,$0C
	.dc.b	$65,$E3,$19,$7F,$B7,$35,$72,$7E,$1F,$DC,$FF,$02,$92,$3F,$B7,$CF
	.dc.b	$8F,$2F,$F6,$9F,$C1,$FF,$30,$FA,$9F,$E1,$AD,$AF,$EF,$6D,$7F,$C1
	.dc.b	$0D,$7F,$CE,$AD,$47,$FB,$3F,$C9,$4F,$46,$C7,$46,$BF,$13,$F5,$97
	.dc.b	$CC,$FF,$8D,$1B,$1F,$E0,$66,$BF,$EB,$DB,$F2,$61,$FC,$B2,$72,$DA
	.dc.b	$E4,$7F,$B9,$A7,$7C,$7E,$67,$FA,$7D,$F9,$48,$79,$4A,$7C,$0C,$3F
	.dc.b	$98,$BE,$D3,$27,$C2,$F8,$3F,$F1,$F3,$FE,$F8,$9F,$F6,$77,$B5,$E1
	.dc.b	$9C,$CF,$A6,$BF,$89,$BF,$23,$EA,$3F,$A3,$DF,$CA,$F8,$D7,$D3,$FF
	.dc.b	$A8,$A7,$34,$9C,$BC,$7C,$BF,$F3,$9C,$63,$B1,$F5,$3F,$E6,$B8,$DD
	.dc.b	$2F,$9C,$61,$B8,$77,$C3,$9B,$F7,$30,$E7,$E3,$7F,$AD,$2F,$85,$7F
	.dc.b	$A0,$30,$E1,$97,$F5,$6C,$7C,$F1,$FE,$4F,$6B,$EB,$F6,$5F,$21,$B6
	.dc.b	$C7,$F9,$2A,$51,$88,$E4,$D7,$94,$FC,$7C,$EC,$7C,$57,$D1,$F5,$2F
	.dc.b	$41,$C5,$BA,$FA,$18,$78,$34,$F8,$4D,$C3,$9C,$7C,$BC,$39,$5F,$1B
	.dc.b	$F9,$57,$B5,$9B,$F2,$1F,$E9,$B5,$F8,$36,$CF,$E6,$72,$FD,$3C,$E7
	.dc.b	$C8,$F2,$CE,$4F,$F4,$79,$F9,$79,$DB,$FB,$26,$BF,$B2,$6B,$B3,$89
	.dc.b	$E4,$72,$D3,$37,$F1,$5D,$F6,$3E,$B3,$FC,$F3,$87,$F9,$D9,$B3,$FB
	.dc.b	$1F,$8E,$FF,$5D,$4E,$7E,$07,$D3,$D5,$FE,$1E,$BD,$B6,$33,$3F,$B1
	.dc.b	$AE,$36,$AD,$4F,$EA,$67,$33,$FC,$39,$AF,$1E,$4F,$4C,$62,$3E,$8B
	.dc.b	$72,$34,$93,$F1,$3B,$FF,$B6,$7B,$BE,$4B,$FE,$53,$E4,$7F,$9C,$71
	.dc.b	$FF,$4A,$FD,$53,$E4,$35,$87,$09,$CC,$7C,$89,$D4,$7F,$41,$6D,$FF
	.dc.b	$40,$6B,$FC,$F1,$AF,$FA,$9D,$AF,$F9,$2D,$BB,$9C,$F5,$7F,$85,$3E
	.dc.b	$9B,$D0,$DB,$FF,$04,$7B,$18,$F2,$55,$AF,$29,$3F,$84,$B4,$62,$FF
	.dc.b	$B0,$CE,$63,$19,$7F,$B1,$F4,$FE,$77,$C3,$F0,$BC,$5F,$CF,$E1,$FE
	.dc.b	$32,$F8,$9C,$4C,$8F,$E8,$7C,$9F,$9C,$BE,$31,$B3,$27,$F4,$87,$38
	.dc.b	$7E,$0E,$31,$87,$68,$C1,$FD,$79,$C5,$7F,$84,$70,$F0,$4E,$B3,$FF
	.dc.b	$1C,$C7,$BF,$53,$49,$87,$F9,$B3,$F3,$0F,$E5,$EF,$FB,$E9,$B7,$FE
	.dc.b	$8B,$FA,$9B,$8F,$ED,$A1,$38,$AF,$EC,$30,$AE,$2F,$FB,$BD,$AE,$9A
	.dc.b	$FF,$68,$6E,$7F,$DA,$3D,$AF,$9C,$7D,$EF,$A9,$FF,$6D,$DF,$C8,$BF
	.dc.b	$85,$3A,$8C,$57,$17,$5A,$2F,$8D,$FC,$AA,$D3,$F1,$E3,$39,$7F,$12
	.dc.b	$AE,$1F,$0F,$F8,$BF,$81,$F1,$37,$1F,$F1,$F4,$7C,$D2,$3F,$0B,$0F
	.dc.b	$B3,$F1,$1F,$87,$3F,$29,$E8,$D9,$F3,$F0,$A7,$15,$F6,$78,$57,$8C
	.dc.b	$FF,$D5,$3F,$48,$F9,$DF,$DC,$74,$A3,$19,$FC,$0F,$8C,$FE,$7E,$FC
	.dc.b	$79,$D5,$70,$AD,$E5,$F8,$C9,$CD,$71,$35,$5F,$C1,$67,$FD,$7D,$F4
	.dc.b	$76,$8E,$7C,$AF,$F6,$AD,$FD,$5A,$DB,$63,$31,$F1,$4F,$EC,$3B,$FF
	.dc.b	$36,$6C,$FF,$AF,$A7,$71,$F2,$DA,$FE,$39,$AF,$F8,$C1,$AF,$23,$FC
	.dc.b	$9F,$FB,$57,$F6,$A9,$F8,$4D,$8F,$D5,$FC,$23,$EA,$9F,$0D,$F3,$2B
	.dc.b	$5B,$FF,$B6,$1B,$87,$F8,$FB,$44,$63,$16,$3C,$F0,$E5,$E5,$FF,$07
	.dc.b	$F3,$A6,$3F,$96,$F1,$3F,$D9,$AF,$E0,$A3,$38,$7E,$57,$9B,$F5,$8F
	.dc.b	$F3,$4F,$85,$CE,$BF,$B1,$EE,$FE,$03,$87,$EC,$1B,$5F,$E5,$2F,$AA
	.dc.b	$7C,$FF,$D9,$9C,$CA,$73,$5F,$EE,$C6,$C7,$D5,$6B,$F9,$B6,$BF,$BE
	.dc.b	$6B,$DE,$BE,$D4,$62,$3B,$47,$F4,$FC,$5F,$79,$FC,$A4,$FE,$97,$A3
	.dc.b	$EA,$B1,$8F,$E9,$26,$E1,$FB,$7A,$D4,$E4,$FD,$4F,$D5,$FF,$33,$3F
	.dc.b	$A1,$F5,$7F,$9A,$A3,$24,$AF,$99,$3A,$F6,$F8,$4F,$E9,$7A,$FF,$64
	.dc.b	$DE,$EF,$D0,$7D,$5F,$E6,$38,$7F,$63,$18,$EB,$95,$F9,$FE,$67,$B9
	.dc.b	$8D,$3F,$9B,$8C,$79,$57,$F2,$66,$C7,$FA,$22,$B8,$CF,$EA,$67,$3E
	.dc.b	$27,$FA,$2B,$BF,$FA,$81,$F2,$3E,$6A,$D7,$F0,$A8,$B8,$7C,$79,$BA
	.dc.b	$7F,$51,$E4,$7D,$07,$F5,$93,$E8,$F0,$5C,$7F,$89,$C5,$C3,$E8,$4D
	.dc.b	$BF,$F1,$AF,$77,$F6,$CF,$92,$FE,$23,$E7,$1F,$E0,$DC,$6F,$82,$CC
	.dc.b	$61,$FD,$F4,$67,$3F,$E0,$59,$7F,$19,$9C,$92,$33,$3B,$9F,$E1,$0D
	.dc.b	$7F,$97,$35,$FE,$CF,$F4,$3F,$DD,$4F,$E9,$1F,$0D,$6B,$F9,$25,$FF
	.dc.b	$80,$7A,$38,$5B,$CA,$46,$1B,$3F,$C5,$7E,$AF,$FA,$2D,$FF,$5E,$6D
	.dc.b	$F1,$F3,$3C,$DB,$FF,$65,$9C,$F2,$FF,$B0,$BE,$57,$FC,$10,$7F,$5B
	.dc.b	$F5,$AE,$3F,$D5,$B0,$FF,$18,$63,$FD,$27,$3F,$EB,$15,$FE,$17,$9C
	.dc.b	$CF,$F6,$E9,$CD,$7F,$2B,$6C,$71,$9D,$BF,$E9,$96,$BE,$7A,$FF,$11
	.dc.b	$7D,$57,$AA,$8D,$FA,$B7,$CA,$FF,$8B,$A3,$0F,$F8,$A6,$33,$8F,$85
	.dc.b	$B8,$7A,$5A,$F9,$DA,$FF,$AA,$DA,$FD,$16,$D9,$CE,$7F,$EB,$AE,$AF
	.dc.b	$F2,$8F,$C2,$FF,$58,$DF,$FB,$B6,$6F,$85,$CD,$79,$DF,$69,$33,$F8
	.dc.b	$D3,$B6,$C6,$73,$FA,$13,$96,$C7,$F9,$1F,$A9,$F1,$9F,$E5,$06,$DF
	.dc.b	$F8,$4F,$17,$D5,$BE,$AF,$57,$FE,$6D,$9C,$C7,$FA,$51,$B5,$FD,$8D
	.dc.b	$B8,$FF,$82,$5A,$FF,$8B,$1A,$FF,$93,$2F,$B4,$F1,$6C,$71,$6B,$FE
	.dc.b	$34,$D5,$EB,$BE,$6B,$FC,$F6,$DB,$FF,$13,$E6,$F9,$6C,$D7,$BD,$F8
	.dc.b	$44,$9F,$C8,$4E,$DB,$19,$C3,$F3,$53,$96,$C7,$F9,$C7,$77,$DC,$FF
	.dc.b	$9E,$9B,$7F,$E3,$8F,$AC,$F5,$EB,$FE,$33,$9C,$E1,$FE,$5D,$9F,$6B
	.dc.b	$FE,$F5,$6B,$FB,$B6,$F2,$1A,$4D,$FF,$D0,$4D,$C3,$FC,$29,$B3,$E2
	.dc.b	$DF,$59,$FE,$A0,$6C,$FF,$93,$32,$B9,$7F,$92,$E6,$E7,$E7,$BF,$71
	.dc.b	$2F,$F6,$D7,$C2,$FE,$A7,$5B,$C5,$FF,$56,$B6,$BF,$CB,$3F,$59,$FE
	.dc.b	$3A,$AF,$F2,$C4,$E5,$B5,$C0,$FF,$51,$CE,$F1,$FF,$48,$56,$DF,$F4
	.dc.b	$97,$AE,$FA,$FF,$F5,$DB,$71,$FF,$34,$BF,$E0,$29,$DC,$E8,$93,$AF
	.dc.b	$4E,$E5,$FD,$78,$1F,$24,$7D,$67,$E3,$DF,$F1,$C4,$EA,$31,$1F,$E8
	.dc.b	$8D,$B1,$CD,$FE,$3F,$AB,$3F,$E6,$B9,$AF,$F8,$FE,$2B,$E0,$8B,$EE
	.dc.b	$27,$33,$E3,$E8,$EE,$A7,$4F,$E9,$2F,$FC,$A9,$7E,$AA,$DD,$69,$E3
	.dc.b	$5E,$DC,$33,$87,$6B,$F0,$B6,$7E,$0E,$1E,$B6,$9E,$B2,$9E,$B2,$9E
	.dc.b	$B3,$FB,$29,$F2,$CF,$EC,$F3,$87,$07,$FA,$C3,$EA,$DE,$1B,$EB,$7F
	.dc.b	$B6,$C6,$1B,$19,$9F,$F3,$94,$E5,$BE,$43,$7E,$2D,$8C,$C6,$08,$FA
	.dc.b	$A7,$FC,$1A,$FD,$66,$B9,$9E,$59,$FE,$02,$51,$8C,$BD,$5D,$9F,$EC
	.dc.b	$AA,$D3,$63,$DD,$3B,$BF,$C7,$DF,$2C,$3E,$C5,$72,$9A,$6A,$FF,$DA
	.dc.b	$BA,$F8,$70,$F6,$36,$FD,$13,$DE,$B8,$1E,$BA,$F0,$2C,$3C,$0F,$C1
	.dc.b	$9E,$0E,$5B,$1A,$C7,$F4,$5C,$FF,$A6,$FB,$FF,$E7,$9B,$FF,$C0,$1C
	.dc.b	$11,$85,$3B,$7F,$DB,$A3,$EC,$75,$19,$8F,$F6,$E1,$C8,$DC,$E9,$E0
	.dc.b	$7F,$B9,$BE,$1D,$F1,$3C,$9B,$19,$8C,$7D,$4E,$59,$19,$39,$47,$9F
	.dc.b	$BF,$E2,$7B,$FF,$E3,$C6,$CF,$F8,$57,$54,$E5,$B1,$9F,$77,$94,$7F
	.dc.b	$A0,$E3,$FD,$F2,$D7,$F2,$9A,$FF,$0E,$BE,$B1,$83,$FC,$75,$F2,$FE
	.dc.b	$A7,$FC,$54,$D7,$FC,$50,$D7,$93,$5E,$39,$7A,$63,$12,$69,$8C,$7F
	.dc.b	$84,$D7,$86,$7F,$D7,$6F,$F2,$2E,$2F,$E1,$B1,$88,$D5,$79,$67,$3F
	.dc.b	$5C,$FB,$DF,$90,$DA,$FF,$1E,$35,$FF,$22,$35,$FF,$21,$35,$F8,$8D
	.dc.b	$7F,$CB,$8D,$B3,$4F,$F4,$A4,$79,$F2,$E3,$EC,$FD,$B9,$CC,$FB,$5F
	.dc.b	$16,$1E,$CF,$A8,$78,$7F,$7D,$B5,$FE,$53,$6B,$FE,$5B,$6B,$FE,$59
	.dc.b	$6B,$F6,$DA,$FF,$9C,$DB,$89,$A7,$F0,$CA,$F1,$73,$7E,$46,$31,$97
	.dc.b	$3C,$3F,$29,$ED,$3D,$1F,$CA,$37,$2F,$F3,$43,$5F,$F3,$83,$5F,$F3
	.dc.b	$73,$5F,$C8,$35,$FF,$44,$37,$23,$98,$7D,$AB,$E1,$BF,$79,$18,$6C
	.dc.b	$66,$3E,$91,$D2,$FF,$59,$F5,$DF,$64,$9C,$1F,$AD,$6C,$FF,$A0,$5A
	.dc.b	$FF,$A2,$5F,$C1,$3E,$2F,$8D,$EE,$47,$BE,$49,$FB,$BA,$7E,$B7,$EB
	.dc.b	$7F,$DB,$75,$FE,$8D,$6B,$FE,$95,$6B,$FE,$93,$6B,$C7,$E0,$7F,$81
	.dc.b	$46,$23,$FD,$5C,$DB,$39,$87,$CA,$BE,$63,$E3,$B6,$3E,$29,$A8,$FE
	.dc.b	$21,$3E,$4E,$3F,$F6,$AF,$5B,$E9,$BF,$95,$7C,$12,$31,$87,$CE,$8C
	.dc.b	$4E,$A5,$F4,$DF,$C6,$FF,$66,$FF,$33,$DF,$FA,$B5,$AF,$FA,$D5,$AF
	.dc.b	$FA,$CD,$AF,$F8,$88,$F5,$FB,$3F,$CD,$BD,$27,$CB,$53,$DB,$1F,$84
	.dc.b	$FC,$09,$F5,$FF,$E0,$06,$DF,$FB,$99,$AF,$53,$FC,$5E,$E1,$C3,$5F
	.dc.b	$13,$5D,$C7,$53,$FC,$1B,$F6,$EF,$9F,$FC,$AB,$67,$E7,$B5,$FE,$5C
	.dc.b	$2B,$C7,$BF,$F8,$AB,$C9,$FF,$2D,$7C,$F7,$BF,$E0,$E5,$FE,$72,$72
	.dc.b	$DD,$55,$A8,$FE,$75,$3F,$E6,$3A,$F9,$13,$F4,$FC,$0F,$95,$F3,$B2
	.dc.b	$FF,$33,$C1,$FC,$3B,$2F,$E5,$ED,$EA,$FF,$29,$E5,$FE,$6D,$BE,$17
	.dc.b	$FC,$DE,$B4,$B9,$5F,$B5,$FF,$85,$23,$F0,$DF,$B5,$BE,$53,$F1,$6D
	.dc.b	$8F,$C5,$4E,$DF,$EA,$55,$CB,$D2,$DD,$C9,$DC,$7F,$7D,$F6,$DF,$AD
	.dc.b	$BE,$87,$0D,$AF,$E0,$73,$A9,$E5,$1D,$B2,$7C,$CF,$90,$91,$F1,$3D
	.dc.b	$9A,$EE,$2F,$0F,$C6,$9F,$CA,$7A,$B2,$BE,$A3,$F9,$7F,$5B,$11,$F5
	.dc.b	$39,$FF,$C1,$8D,$9D,$3F,$B4,$FA,$6F,$F2,$8B,$F9,$53,$FD,$06,$E1
	.dc.b	$C3,$5F,$CD,$35,$FD,$43,$5E,$37,$FE,$62,$8C,$41,$BF,$5E,$F9,$0F
	.dc.b	$E5,$AD,$9F,$E0,$ED,$7F,$33,$3E,$3C,$BF,$0F,$97,$ED,$7E,$06,$57
	.dc.b	$E3,$23,$F8,$11,$87,$E1,$46,$1F,$C0,$FC,$BF,$F0,$EE,$53,$4F,$BD
	.dc.b	$83,$E6,$31,$5C,$08,$7F,$86,$78,$BD,$91,$F8,$5C,$DF,$23,$F2,$E9
	.dc.b	$49,$CC,$7C,$F8,$C4,$ED,$B1,$99,$DB,$FC,$26,$32,$D8,$CB,$FC,$66
	.dc.b	$32,$FF,$0F,$87,$87,$F3,$89,$1F,$96,$8E,$BA,$FF,$03,$E1,$7C,$27
	.dc.b	$F2,$12,$57,$EE,$39,$6B,$F8,$D8,$B5,$FB,$A9,$FE,$50,$49,$FE,$10
	.dc.b	$FF,$93,$FF,$0F,$FE,$45,$3D,$E3,$53,$F9,$77,$4E,$FC,$4F,$F2,$A5
	.dc.b	$7F,$1B,$8F,$E6,$DE,$6A,$31,$1F,$E4,$D6,$A8,$7C,$B7,$F1,$88,$FE
	.dc.b	$D2,$D7,$FC,$84,$DC,$3F,$C8,$35,$FC,$A5,$FF,$21,$C6,$7C,$1F,$E2
	.dc.b	$B9,$FF,$59,$6F,$9F,$FE,$6B,$9D,$61,$E1,$6F,$84,$BE,$13,$FE,$86
	.dc.b	$6E,$5E,$6F,$13,$FC,$1F,$F5,$AF,$8C,$FF,$65,$B6,$FF,$18,$D7,$F7
	.dc.b	$0D,$7F,$A7,$3E,$9C,$FF,$AE,$36,$3F,$DF,$ED,$4F,$3D,$3E,$0F,$C8
	.dc.b	$7D,$02,$4F,$F0,$5C,$7E,$84,$7F,$A1,$F4,$FF,$4A,$CF,$CA,$EB,$F8
	.dc.b	$47,$DF,$9B,$1F,$42,$2B,$FE,$71,$DD,$38,$C3,$E7,$D5,$CB,$EE,$4F
	.dc.b	$E2,$7E,$B7,$F0,$64,$C3,$C9,$82,$3F,$0F,$8F,$E3,$C9,$3F,$E2,$89
	.dc.b	$D3,$FE,$0B,$FD,$6F,$F9,$9B,$FC,$87,$C6,$46,$23,$FB,$C4,$63,$C2
	.dc.b	$FD,$E3,$63,$84,$ED,$FF,$05,$46,$5E,$31,$94,$7F,$59,$70,$BE,$15
	.dc.b	$F5,$1F,$B2,$8F,$F2,$6C,$63,$A3,$FD,$9C,$D8,$F5,$61,$8F,$57,$F0
	.dc.b	$51,$DA,$7F,$C6,$9C,$7F,$C4,$DB,$1F,$CF,$DA,$FF,$85,$5A,$FF,$88
	.dc.b	$DA,$F3,$9E,$AF,$9D,$B1,$E8,$E0,$F2,$E3,$F8,$4E,$17,$0F,$8E,$30
	.dc.b	$F3,$C7,$75,$31,$F7,$D6,$9E,$D3,$ED,$E5,$F5,$65,$FA,$B9,$F5,$35
	.dc.b	$F5,$35,$F5,$35,$F5,$47,$22,$3F,$0F,$E1,$7F,$D5,$39,$7A,$23,$EA
	.dc.b	$46,$32,$F9,$8D,$8F,$85,$5F,$E8,$6F,$AB,$FE,$5F,$A7,$DE,$F8,$CF
	.dc.b	$C5,$46,$29,$18,$EE,$FC,$04,$63,$7F,$F7,$37,$CB,$7F,$82,$E3,$07
	.dc.b	$E2,$2F,$95,$F1,$C3,$B5,$FF,$96,$67,$2F,$2E,$EF,$F2,$E5,$6E,$FF
	.dc.b	$BF,$57,$8D,$FF,$2A,$35,$E5,$8F,$91,$CF,$C0,$3E,$3B,$FA,$E5,$7F
	.dc.b	$A9,$5B,$19,$9E,$AD,$78,$C7,$99,$F5,$46,$EF,$F6,$3D,$6F,$AB,$FF
	.dc.b	$45,$70,$6E,$D8,$F7,$7E,$19,$8C,$39,$69,$F0,$5B,$1F,$D0,$6F,$9B
	.dc.b	$C2,$B9,$63,$FC,$03,$57,$C4,$39,$BF,$01,$C7,$F8,$19,$CC,$63,$4F
	.dc.b	$C0,$FD,$4D,$CB,$F7,$18,$71,$9F,$DB,$F2,$2B,$F1,$55,$FA,$B8,$C4
	.dc.b	$FE,$3E,$73,$3D,$23,$04,$7F,$69,$3C,$1B,$3C,$1F,$E1,$1E,$23,$C3
	.dc.b	$F2,$65,$FE,$0F,$6D,$FF,$62,$C3,$F8,$5D,$6D,$B3,$C2,$31,$6C,$FE
	.dc.b	$15,$F1,$8C,$5F,$F0,$DA,$DD,$6B,$BB,$C9,$8B,$0E,$18,$7C,$19,$CC
	.dc.b	$62,$3F,$57,$18,$26,$1F,$B3,$79,$E0,$F2,$FC,$C7,$D1,$EE,$FF,$39
	.dc.b	$F2,$E2,$FF,$D8,$DE,$87,$E5,$DF,$E7,$8D,$C7,$F9,$DB,$5F,$E6,$ED
	.dc.b	$7F,$9A,$D7,$49,$E3,$5F,$3E,$31,$3C,$A3,$E9,$73,$7F,$84,$FE,$5B
	.dc.b	$F0,$8D,$BF,$EA,$1F,$29,$8F,$09,$56,$38,$C1,$A4,$9F,$EE,$1F,$B4
	.dc.b	$FA,$78,$7F,$8A,$60,$8C,$4F,$F8,$87,$E3,$3C,$31,$8C,$FF,$C4,$4D
	.dc.b	$AE,$19,$F8,$EB,$6C,$BF,$EA,$B8,$76,$BE,$19,$2F,$58,$FF,$79,$7D
	.dc.b	$24,$BF,$EE,$6F,$E5,$61,$F5,$5F,$E6,$B9,$EA,$7F,$BB,$DC,$47,$FB
	.dc.b	$59,$FF,$5C,$46,$A3,$FB,$3C,$6B,$0F,$D5,$BA,$8F,$F4,$C3,$5F,$F7
	.dc.b	$73,$5E,$DE,$BF,$D6,$C7,$FC,$59,$7F,$ED,$BF,$7B,$E3,$F8,$9F,$BB
	.dc.b	$F3,$3F,$C6,$D1,$89,$D1,$E3,$DB,$E8,$56,$FC,$4F,$F8,$03,$39,$C3
	.dc.b	$AE,$5F,$A6,$3E,$8D,$7E,$07,$EA,$7A,$47,$F8,$85,$FC,$09,$2B,$FC
	.dc.b	$36,$FE,$F7,$27,$EB,$5F,$E0,$89,$FF,$7A,$36,$BF,$C5,$5F,$07,$E9
	.dc.b	$8C,$6F,$FE,$FD,$6D,$7F,$BB,$1B,$5F,$E6,$5E,$15,$F1,$8E,$5C,$52
	.dc.b	$FE,$07,$E5,$6F,$89,$C5,$7F,$C0,$0C,$7F,$1D,$9F,$9B,$0F,$D0,$DE
	.dc.b	$F0,$F3,$C7,$F2,$7B,$FF,$7C,$78,$B7,$A0,$FF,$2B,$F4,$FF,$D6,$1E
	.dc.b	$CE,$63,$1E,$B7,$FB,$1F,$1F,$F5,$43,$70,$FF,$52,$35,$FF,$51,$35
	.dc.b	$FF,$17,$DF,$49,$E3,$7F,$CA,$63,$13,$CA,$3F,$7D,$EC,$FF,$1C,$AF
	.dc.b	$E0,$6D,$9F,$F5,$DD,$BE,$C7,$F8,$94,$E7,$23,$4E,$18,$FF,$B2,$E7
	.dc.b	$7E,$27,$C6,$F3,$FE,$C7,$45,$CF,$EE,$B7,$3F,$9C,$DE,$5F,$B8,$FC
	.dc.b	$BD,$FF,$9A,$47,$F4,$E6,$D7,$F9,$3A,$9F,$51,$F9,$67,$F2,$87,$F7
	.dc.b	$BF,$13,$EA,$46,$63,$81,$C3,$B3,$F7,$51,$98,$E8,$D7,$FC,$B6,$D7
	.dc.b	$85,$7F,$B8,$63,$06,$F9,$3F,$37,$EC,$7E,$5C,$FC,$C7,$5D,$AF,$CE
	.dc.b	$C5,$BF,$C7,$4D,$CF,$FC,$69,$C4,$F7,$9E,$8F,$F4,$36,$E3,$F9,$8E
	.dc.b	$2C,$4C,$8F,$E1,$95,$C6,$3F,$2F,$39,$C7,$FD,$F9,$E0,$3E,$37,$F2
	.dc.b	$8D,$C3,$8E,$3F,$EE,$67,$F8,$B7,$2A,$75,$7F,$C0,$72,$FC,$1F,$C1
	.dc.b	$79,$7D,$6F,$C3,$B7,$0F,$1F,$42,$31,$7C,$63,$13,$E0,$AF,$D6,$FD
	.dc.b	$BF,$EF,$FE,$D2,$CF,$F0,$46,$DF,$ED,$5B,$C1,$FC,$0D,$FE,$0D,$F3
	.dc.b	$5F,$C1,$FD,$CB,$7F,$C3,$1B,$7F,$BD,$6D,$FF,$0B,$7F,$0F,$DB,$2B
	.dc.b	$F9,$B5,$3E,$97,$E0,$DF,$F8,$83,$A3,$F9,$C5,$7E,$4E,$7E,$2D,$7E
	.dc.b	$39,$AE,$78,$FF,$A4,$46,$23,$FD,$4E,$D7,$B4,$68,$93,$F9,$88,$FF
	.dc.b	$56,$FC,$EF,$F0,$C8,$FE,$4F,$F2,$BF,$C4,$63,$13,$FC,$66,$3F,$BB
	.dc.b	$F6,$7F,$AC,$6B,$FD,$BC,$D9,$FD,$1F,$94,$8C,$47,$EB,$63,$1E,$9F
	.dc.b	$F3,$7F,$19,$F4,$1F,$CC,$CF,$E6,$9F,$01,$71,$F8,$F9,$FF,$25,$C7
	.dc.b	$F6,$6D,$DF,$FB,$EF,$D2,$3E,$13,$D1,$E5,$FF,$83,$23,$18,$7F,$BC
	.dc.b	$DB,$87,$99,$AF,$95,$AF,$FB,$43,$54,$62,$11,$8B,$FF,$0D,$CE,$63
	.dc.b	$15,$FE,$B0,$3E,$FE,$BF,$E6,$CC,$7F,$C2,$7F,$48,$F8,$5F,$E7,$D3
	.dc.b	$9A,$F0,$4E,$5F,$AD,$68,$C6,$5F,$39,$AF,$CD,$6B,$CA,$FF,$C5,$F1
	.dc.b	$88,$F4,$B6,$7A,$4F,$F8,$46,$DF,$03,$E4,$3F,$C6,$BF,$49,$F8,$2C
	.dc.b	$BF,$CE,$B3,$AC,$BF,$BF,$4F,$E1,$B2,$FB,$CD,$7E,$53,$5F,$92,$DC
	.dc.b	$4D,$27,$9D,$FE,$46,$75,$F8,$67,$E1,$AF,$FC,$18,$DB,$FF,$04,$5A
	.dc.b	$1C,$3F,$E4,$C7,$59,$7F,$90,$9C,$DF,$CC,$6D,$FE,$61,$AF,$E5,$DA
	.dc.b	$FE,$0D,$AF,$E0,$9B,$7E,$1B,$FF,$2E,$C6,$3E,$6D,$CD,$FE,$5E,$6E
	.dc.b	$1F,$E2,$57,$FA,$8E,$1F,$8C,$72,$FB,$1F,$F4,$03,$9E,$6F,$F3,$4C
	.dc.b	$63,$13,$49,$1A,$25,$7E,$2B,$E9,$3F,$C3,$74,$68,$F3,$BF,$CE,$6E
	.dc.b	$1F,$F3,$BB,$99,$E5,$18,$23,$82,$61,$FC,$0A,$78,$3F,$C1,$3E,$43
	.dc.b	$E7,$EB,$FE,$74,$6D,$7E,$BD,$F8,$78,$1C,$D7,$EC,$1F,$80,$4C,$7F
	.dc.b	$D1,$6F,$C6,$F3,$5D,$1F,$E3,$86,$D7,$F9,$5E,$3F,$C6,$26,$48,$E7
	.dc.b	$CF,$D3,$FE,$5D,$F2,$5F,$1D,$4C,$7C,$1D,$D3,$FC,$E1,$8A,$B7,$18
	.dc.b	$6D,$FF,$97,$67,$55,$C6,$3F,$C1,$A0,$7C,$25,$FA,$17,$8D,$FE,$0E
	.dc.b	$CD,$96,$63,$F9,$D5,$5B,$FF,$0D,$56,$95,$F2,$7F,$C2,$F3,$B7,$C1
	.dc.b	$3B,$9D,$47,$F8,$8E,$74,$FF,$B0,$E3,$2F,$9E,$3D,$1F,$45,$8A,$F0
	.dc.b	$3A,$9F,$58,$75,$87,$0C,$3D,$33,$E6,$7D,$D1,$E0,$8E,$38,$70,$74
	.dc.b	$7A,$8A,$7A,$A3,$86,$1E,$8C,$3D,$97,$DA,$BF,$B0,$1E,$42,$CE,$1F
	.dc.b	$61,$FD,$97,$E1,$1F,$20,$FA,$3E,$31,$86,$C6,$67,$83,$5F,$F2,$2C
	.dc.b	$BE,$49,$CB,$C9,$B3,$DD,$AF,$06,$BF,$E5,$59,$E5,$3F,$26,$3F,$11
	.dc.b	$1F,$C0,$9C,$07,$0F,$D1,$C7,$95,$78,$35,$FB,$F5,$D2,$B4,$74,$E8
	.dc.b	$FF,$32,$4E,$9B,$7C,$B0,$EF,$7C,$6B,$57,$8B,$DB,$E6,$C3,$C1,$87
	.dc.b	$18,$EE,$79,$E7,$6D,$3A,$CF,$07,$96,$1D,$58,$72,$C3,$C5,$7F,$AB
	.dc.b	$E0,$FD,$67,$C3,$7E,$01,$B5,$FE,$21,$6B,$FE,$7A,$6B,$FE,$83,$6B
	.dc.b	$CC,$47,$36,$9C,$CA,$7F,$88,$BD,$AF,$F5,$BB,$FD,$E9,$AE,$7B,$FF
	.dc.b	$27,$D2,$9C,$CF,$F8,$9D,$AF,$F8,$4D,$AF,$F8,$49,$B1,$C6,$79,$E2
	.dc.b	$FA,$3A,$4B,$FF,$05,$C6,$5F,$F0,$6F,$C5,$3E,$F7,$C6,$D8,$CC,$60
	.dc.b	$F3,$FD,$1F,$F0,$D8,$FF,$1F,$B5,$FF,$1F,$35,$FF,$15,$35,$FF,$14
	.dc.b	$B7,$13,$49,$CD,$F5,$FC,$F7,$A7,$49,$7F,$66,$32,$FF,$89,$BE,$8D
	.dc.b	$F1,$1F,$31,$B1,$98,$C1,$1C,$10,$FF,$2C,$C7,$07,$FD,$7B,$C6,$FB
	.dc.b	$D7,$D2,$FF,$0A,$8C,$36,$33,$3F,$E6,$D9,$CB,$FE,$3A,$F1,$BF,$66
	.dc.b	$FF,$98,$DA,$FF,$93,$DA,$FF,$93,$9B,$AF,$F8,$2B,$A3,$F0,$F1,$82
	.dc.b	$3E,$1F,$A3,$95,$E1,$C1,$18,$8F,$F4,$07,$D2,$7E,$29,$34,$9C,$BF
	.dc.b	$C8,$DB,$1F,$CA,$1A,$FF,$9F,$BE,$91,$F2,$5F,$9A,$6C,$7F,$96,$1A
	.dc.b	$EE,$3A,$7A,$92,$3A,$35,$F2,$B5,$FF,$76,$D2,$31,$1F,$D9,$FD,$F7
	.dc.b	$DC,$7E,$91,$B8,$FF,$3E,$6C,$FF,$A7,$DA,$FF,$A8,$7E,$97,$F6,$0F
	.dc.b	$F4,$46,$B9,$EC,$FE,$89,$18,$8F,$F7,$FB,$5F,$EF,$01,$C3,$FD,$ED
	.dc.b	$AE,$63,$F7,$F7,$FD,$8E,$3F,$CF,$AD,$7F,$D5,$2D,$7F,$D5,$0D,$7F
	.dc.b	$D2,$9F,$3D,$5F,$F4,$9B,$78,$4D,$35,$1E,$5F,$53,$D6,$47,$E0,$3F
	.dc.b	$C4,$BC,$E7,$CF,$52,$4F,$FA,$F9,$FE,$CD,$F4,$2F,$8A,$A4,$9F,$ED
	.dc.b	$B1,$96,$C6,$5F,$F6,$84,$65,$B1,$FD,$C7,$E7,$DF,$41,$49,$3F,$DE
	.dc.b	$23,$2F,$F3,$9F,$7F,$FD,$9C,$DC,$7F,$9F,$46,$23,$F9,$E3,$5F,$E7
	.dc.b	$8D,$7F,$A3,$35,$78,$E8,$C3,$67,$FA,$57,$D7,$7D,$E8,$34,$9C,$1A
	.dc.b	$25,$FF,$4F,$8C,$BF,$EF,$1F,$49,$F3,$2F,$8D,$FF,$0A,$C6,$1B,$19
	.dc.b	$9E,$0D,$7F,$22,$7F,$94,$5C,$3F,$94,$9E,$4F,$09,$E6,$7E,$97,$EE
	.dc.b	$5F,$0F,$49,$3F,$A7,$8C,$BE,$7E,$2F,$C8,$5F,$FB,$1B,$D0,$F3,$C6
	.dc.b	$08,$E0,$93,$C5,$FA,$BF,$33,$23,$ED,$35,$FA,$ED,$7E,$BB,$5F,$D7
	.dc.b	$FC,$DB,$E5,$3F,$C7,$0D,$8F,$F1,$EB,$5F,$8A,$1C,$3F,$21,$AF,$48
	.dc.b	$E6,$47,$9E,$3D,$88,$E0,$FC,$97,$CD,$FE,$40,$8E,$08,$7E,$49,$CB
	.dc.b	$4F,$E0,$3A,$7F,$04,$23,$82,$1F,$C1,$1C,$B4,$F7,$E5,$F0,$88,$E0
	.dc.b	$87,$C2,$C3,$FD,$FA,$E0,$F8,$51,$FC,$9F,$E1,$1F,$21,$49,$3F,$CA
	.dc.b	$63,$2F,$CC,$EB,$7D,$A7,$F9,$A5,$B1,$FE,$1C,$8C,$11,$C1,$0E,$4F
	.dc.b	$F9,$C7,$93,$F6,$1E,$1F,$EE,$89,$3F,$BF,$6B,$FB,$C6,$BF,$BC,$7F
	.dc.b	$8D,$5C,$8F,$37,$32,$31,$1F,$E3,$FF,$82,$7C,$B7,$F9,$15,$B1,$FE
	.dc.b	$45,$6B,$B8,$FE,$4D,$F2,$52,$3A,$B5,$FD,$9E,$6C,$13,$99,$FE,$8F
	.dc.b	$39,$8C,$47,$F4,$AF,$A3,$FE,$48,$FE,$F1,$AE,$7B,$BF,$D2,$D1,$88
	.dc.b	$7B,$EF,$C8,$3F,$D1,$D1,$FA,$08,$C7,$BB,$FE,$A0,$6C,$66,$76,$FA
	.dc.b	$32,$7C,$17,$F4,$32,$47,$CA,$7F,$96,$F9,$B2,$3F,$99,$35,$FE,$5E
	.dc.b	$D7,$F9,$7B,$5F,$E5,$ED,$7F,$97,$F8,$3F,$E1,$67,$99,$F8,$DF,$9B
	.dc.b	$FE,$69,$B1,$A4,$C3,$F9,$FF,$AB,$2B,$FA,$23,$5F,$E4,$BF,$16,$FA
	.dc.b	$37,$CA,$FF,$B0,$A3,$0D,$8C,$CF,$F0,$89,$CB,$63,$2F,$46,$BF,$E0
	.dc.b	$28,$DB,$99,$ED,$E3,$B9,$69,$27,$B4,$65,$ED,$B7,$8C,$8E,$08,$78
	.dc.b	$5C,$B4,$E6,$78,$08,$60,$87,$C4,$8F,$F7,$05,$9A,$63,$DF,$FF,$31
	.dc.b	$4F,$B5,$A8,$4F,$FB,$3F,$2F,$99,$3A,$9C,$B6,$75,$1F,$36,$30,$6F
	.dc.b	$CF,$7E,$6C,$FB,$1F,$F1,$D3,$6B,$FC,$74,$D7,$FB,$0C,$F0,$9F,$F7
	.dc.b	$D7,$D1,$3C,$0F,$2D,$3E,$21,$F4,$9F,$E3,$8A,$D6,$7E,$B9,$CD,$7B
	.dc.b	$67,$2F,$1A,$F8,$1C,$8F,$A5,$FE,$E5,$7C,$29,$7C,$23,$0D,$BF,$06
	.dc.b	$3E,$49,$EF,$87,$3F,$17,$FC,$97,$1C,$27,$D3,$87,$F7,$EC,$3C,$6D
	.dc.b	$8C,$BD,$32,$FE,$BE,$F8,$BD,$6F,$E0,$AF,$F6,$2A,$46,$23,$A3,$5F
	.dc.b	$88,$D7,$E1,$3C,$AF,$D6,$DC,$BE,$1F,$D2,$FF,$B6,$7E,$0F,$E5,$47
	.dc.b	$2E,$DB,$87,$26,$BE,$57,$3A,$AF,$94,$FC,$F3,$81,$78,$1F,$98,$BE
	.dc.b	$6F,$FD,$9D,$18,$6C,$67,$2F,$F4,$03,$6B,$F8,$1F,$C4,$7F,$1E,$9F
	.dc.b	$DF,$36,$33,$D1,$FE,$D9,$E0,$4E,$67,$F7,$CD,$7F,$45,$1A,$9E,$13
	.dc.b	$99,$FD,$2B,$67,$FD,$1F,$5F,$C3,$7D,$5F,$D5,$36,$BC,$11,$87,$F8
	.dc.b	$CB,$5F,$E1,$CD,$7C,$2D,$7F,$66,$D9,$FF,$10,$78,$DE,$BF,$33,$FA
	.dc.b	$FD,$FF,$1D,$6D,$FF,$26,$6B,$FC,$6D,$AF,$B9,$AF,$F0,$06,$D7,$69
	.dc.b	$FE,$51,$ED,$FF,$96,$DB,$CD,$F2,$E3,$17,$FC,$D1,$AF,$7F,$8A,$F3
	.dc.b	$12,$BC,$CD,$8F,$33,$5E,$11,$FE,$83,$FB,$33,$49,$1F,$E8,$28,$CB
	.dc.b	$FD,$DB,$C3,$7A,$18,$8F,$EE,$B1,$8A,$FE,$A5,$39,$9F,$E0,$33,$97
	.dc.b	$FD,$1B,$F1,$59,$7F,$AD,$B7,$CF,$FF,$45,$37,$C7,$34,$98,$7F,$A6
	.dc.b	$DB,$5D,$23,$04,$70,$7F,$83,$BE,$9B,$FC,$03,$1F,$E3,$36,$C6,$7A
	.dc.b	$7C,$9D,$29,$CC,$FF,$B6,$3A,$1F,$A4,$F9,$30,$FF,$73,$36,$7F,$96
	.dc.b	$B5,$E0,$D7,$FD,$C3,$F5,$1F,$AD,$7F,$D2,$6D,$73,$D5,$F1,$54,$65
	.dc.b	$63,$E5,$F1,$FC,$F9,$7F,$67,$C7,$D9,$18,$23,$82,$57,$C3,$8F,$A0
	.dc.b	$3E,$96,$92,$7F,$C7,$8F,$6F,$A5,$91,$F2,$5B,$5F,$EC,$8F,$A8,$FB
	.dc.b	$04,$9F,$F6,$83,$63,$51,$82,$3F,$E4,$2F,$59,$3F,$49,$B3,$FE,$60
	.dc.b	$8C,$11,$C1,$27,$94,$7F,$1E,$F3,$9F,$69,$49,$5F,$E9,$EE,$0F,$F2
	.dc.b	$84,$7C,$F7,$F9,$FC,$7F,$44,$6D,$EF,$C4,$D1,$27,$4D,$8D,$47,$F9
	.dc.b	$FF,$EC,$FF,$5A,$D3,$F1,$AE,$23,$FC,$C2,$D7,$FC,$BE,$D7,$FB,$4B
	.dc.b	$5F,$C0,$B7,$73,$9E,$6F,$F7,$97,$C7,$7E,$A4,$97,$FF,$00,$5B,$7F
	.dc.b	$EF,$E6,$BF,$EF,$E7,$F2,$1C,$AE,$AA,$4A,$FF,$32,$3A,$D6,$47,$F9
	.dc.b	$9D,$B7,$FE,$6B,$6B,$FE,$F0,$AF,$F3,$F4,$EC,$FF,$67,$E0,$E8,$7D
	.dc.b	$57,$F0,$12,$4E,$A3,$2D,$8C,$DF,$FB,$F2,$B9,$46,$23,$FD,$F9,$F6
	.dc.b	$FF,$E2,$5A,$FF,$6E,$FC,$0A,$31,$3F,$E3,$53,$FC,$6D,$19,$F3,$BF
	.dc.b	$8B,$36,$B8,$D7,$0A,$EB,$5A,$8C,$47,$26,$AE,$17,$BC,$F8,$0F,$EE
	.dc.b	$6D,$C3,$F2,$8D,$AF,$F1,$BB,$5F,$36,$3F,$ED,$0F,$A5,$F2,$FC,$01
	.dc.b	$A4,$BF,$7C,$65,$B1,$9A,$F8,$02,$BA,$4F,$D4,$F8,$57,$CA,$FF,$35
	.dc.b	$6C,$66,$30,$47,$04,$AF,$A9,$3C,$1F,$AD,$E7,$FE,$5C,$93,$FD,$05
	.dc.b	$AF,$E1,$1B,$1C,$23,$F9,$27,$CA,$7F,$A5,$1F,$F1,$CB,$63,$D6,$1C
	.dc.b	$3E,$E6,$BF,$7D,$FE,$58,$EF,$C2,$EA,$49,$EA,$D8,$EB,$3C,$0F,$E6
	.dc.b	$9D,$3F,$D1,$FE,$0F,$F9,$99,$2F,$FA,$73,$FC,$EF,$E7,$3F,$85,$DF
	.dc.b	$D0,$6B,$F9,$16,$BF,$E8,$76,$D7,$FB,$97,$CC,$FE,$AB,$D3,$FE,$8A
	.dc.b	$C3,$FD,$EF,$18,$3F,$86,$FD,$24,$8F,$E2,$4D,$C3,$E2,$E0,$F8,$FF
	.dc.b	$21,$2B,$FC,$C7,$19,$6C,$66,$78,$B5,$FE,$C9,$F4,$9F,$D7,$5F,$CF
	.dc.b	$37,$98,$D3,$51,$FE,$30,$AF,$F6,$67,$A5,$FD,$7F,$2F,$F2,$DC,$FF
	.dc.b	$8E,$A7,$24,$7F,$C8,$4D,$7F,$71,$A3,$E9,$1F,$1B,$FD,$E9,$B5,$F8
	.dc.b	$49,$D4,$E4,$F0,$7B,$72,$7F,$DB,$CD,$BF,$F6,$E3,$5F,$F7,$2B,$5F
	.dc.b	$F5,$E7,$CB,$BE,$D1,$F4,$9F,$E1,$D8,$C3,$63,$37,$F9,$69,$CB,$E5
	.dc.b	$FA,$47,$D3,$7F,$84,$9B,$19,$8C,$11,$C1,$30,$FD,$DE,$6A,$E3,$1F
	.dc.b	$BE,$BF,$37,$9B,$E3,$25,$F8,$F1,$47,$F8,$8B,$E6,$9F,$37,$49,$5F
	.dc.b	$2E,$32,$D8,$CE,$1E,$29,$CB,$63,$FC,$7D,$E4,$7E,$CA,$1E,$3A,$30
	.dc.b	$DF,$88,$D9,$FE,$34,$FA,$5F,$F3,$C6,$C6,$92,$F4,$4C,$BF,$01,$19
	.dc.b	$6E,$7C,$C8,$E0,$FC,$37,$D2,$FD,$6A,$F9,$6D,$7F,$1E,$D7,$FD,$BA
	.dc.b	$D7,$FC,$6B,$F3,$6F,$8A,$FF,$25,$B6,$BF,$24,$D9,$FA,$ED,$7F,$36
	.dc.b	$D7,$FD,$D2,$D7,$AC,$F4,$3F,$CD,$1E,$2B,$E4,$A9,$2F,$FB,$6C,$65
	.dc.b	$B1,$9A,$FE,$D1,$39,$6C,$7F,$94,$5F,$F7,$F0,$F9,$57,$D0,$FF,$7B
	.dc.b	$B4,$E6,$3F,$98,$FC,$4E,$2B,$73,$FD,$B7,$87,$FD,$17,$18,$8F,$F7
	.dc.b	$C9,$C4,$D5,$F1,$BE,$71,$93,$BD,$72,$8C,$D7,$39,$FF,$3D,$B5,$5F
	.dc.b	$37,$A9,$CC,$F6,$C7,$1A,$E9,$1F,$D9,$27,$83,$E0,$BA,$F0,$BF,$84
	.dc.b	$F3,$3B,$9D,$6B,$9B,$EE,$3F,$AA,$9A,$2D,$61,$F9,$E7,$F7,$4F,$8D
	.dc.b	$7D,$13,$E8,$F8,$C6,$1B,$19,$9F,$EC,$33,$96,$FC,$A3,$A7,$8B,$63
	.dc.b	$31,$82,$2F,$A4,$FA,$D4,$E9,$FF,$13,$46,$1B,$A7,$F8,$A6,$58,$75
	.dc.b	$AE,$2D,$9D,$57,$0B,$FC,$4D,$F5,$7F,$17,$7C,$30,$B7,$D7,$0F,$46
	.dc.b	$DE,$1B,$FA,$B8,$71,$7F,$0C,$F8,$AF,$35,$E1,$9E,$F8,$F8,$55,$71
	.dc.b	$C7,$C1,$1F,$07,$1F,$06,$1E,$28,$E2,$53,$8D,$F3,$BF,$04,$71,$3F
	.dc.b	$8E,$8E,$7F,$F2,$75,$7F,$BB,$3A,$3F,$53,$18,$E0,$FD,$5C,$62,$BF
	.dc.b	$A6,$C6,$17,$C4,$6D,$EB,$56,$B8,$47,$DF,$AE,$13,$98,$E2,$74,$3F
	.dc.b	$98,$F4,$FF,$BE,$F0,$F2,$36,$BC,$6D,$9D,$FD,$C7,$78,$C6,$1F,$ED
	.dc.b	$4A,$C1,$99,$F3,$4F,$CE,$FB,$4F,$D0,$B6,$FF,$BA,$66,$9C,$B6,$3F
	.dc.b	$D0,$B6,$FB,$2F,$73,$63,$D5,$8A,$31,$1F,$04,$FF,$73,$D7,$37,$18
	.dc.b	$7F,$97,$3C,$9F,$1F,$A5,$FE,$23,$9F,$F0,$33,$5F,$ED,$6D,$7F,$C1
	.dc.b	$ED,$7F,$90,$CF,$37,$8C,$F4,$3F,$7F,$87,$F9,$4F,$D8,$FA,$DE,$81
	.dc.b	$A7,$C6,$C3,$C1,$B7,$AF,$0F,$F7,$FF,$D2,$FF,$B9,$1F,$F0,$D3,$5F
	.dc.b	$EF,$ED,$7F,$C4,$6D,$7C,$AD,$79,$FC,$0F,$D2,$8C,$66,$75,$97,$F7
	.dc.b	$4C,$3F,$18,$FF,$A0,$AF,$E7,$B5,$F9,$73,$E3,$C3,$D1,$7F,$E0,$8E
	.dc.b	$6F,$BF,$E1,$FF,$04,$BF,$F1,$93,$5F,$F0,$BB,$5F,$F1,$DB,$5F,$9A
	.dc.b	$D5,$3D,$5E,$73,$F3,$BF,$05,$79,$BF,$A3,$6C,$FE,$19,$AF,$E0,$AF
	.dc.b	$C5,$87,$A3,$F0,$17,$9D,$F9,$66,$D7,$F9,$25,$AF,$F8,$B9,$AF,$F9
	.dc.b	$49,$AF,$E1,$1A,$FA,$1B,$91,$CE,$8F,$C0,$9E,$A7,$EE,$DB,$7F,$9E
	.dc.b	$6B,$F9,$AA,$F1,$E3,$F4,$3E,$9F,$FC,$21,$3F,$E5,$F6,$BF,$E4,$46
	.dc.b	$BF,$E6,$86,$BF,$9A,$6B,$F3,$9B,$89,$A6,$B1,$7E,$10,$F8,$9F,$E2
	.dc.b	$4D,$BF,$DA,$B5,$FD,$8D,$79,$30,$FC,$3F,$DC,$3E,$2B,$F8,$2B,$6B
	.dc.b	$FC,$F0,$D7,$FC,$BA,$D7,$FD,$02,$D7,$F6,$6D,$7F,$0E,$DC,$8E,$72
	.dc.b	$7B,$AF,$87,$FE,$54,$DB,$FE,$1A,$D7,$F8,$4D,$78,$F1,$FD,$B4,$3E
	.dc.b	$49,$EB,$7F,$33,$D9,$F6,$F4,$FE,$71,$5B,$9F,$E4,$BC,$4F,$A8,$FE
	.dc.b	$4A,$D9,$FF,$4C,$35,$FF,$3E,$B5,$FF,$18,$8F,$71,$F0,$EF,$A5,$FE
	.dc.b	$8D,$18,$6C,$66,$7F,$A1,$4F,$0F,$80,$F2,$FE,$04,$F8,$BF,$F0,$A3
	.dc.b	$6B,$FC,$32,$D7,$FD,$DE,$D7,$C6,$1C,$3E,$66,$BF,$E4,$86,$BB,$8F
	.dc.b	$F1,$17,$DC,$FF,$7B,$E0,$F9,$AD,$F1,$FD,$ED,$AE,$55,$FD,$87,$E1
	.dc.b	$9E,$5D,$24,$FF,$46,$8F,$E5,$B7,$FD,$B9,$DC,$66,$76,$FF,$49,$E7
	.dc.b	$4F,$08,$FF,$4B,$AF,$8E,$F0,$12,$3F,$0F,$E3,$D7,$94,$ED,$F1,$4F
	.dc.b	$E3,$A7,$EA,$FC,$B3,$6A,$4A,$FA,$BF,$16,$FA,$8C,$47,$F5,$DE,$15
	.dc.b	$F1,$8F,$B6,$D7,$55,$C5,$CD,$FC,$7E,$69,$18,$3B,$FC,$87,$CB,$F1
	.dc.b	$E4,$60,$87,$B1,$CB,$DA,$3E,$97,$95,$F4,$FE,$BE,$4F,$D1,$8C,$BF
	.dc.b	$4E,$32,$FC,$68,$FD,$2F,$CD,$7F,$4D,$8F,$C5,$B5,$FC,$FB,$5F,$F4
	.dc.b	$8B,$53,$49,$1F,$E6,$D7,$5F,$39,$FD,$C6,$7F,$C4,$41,$C3,$FE,$09
	.dc.b	$31,$BF,$92,$BF,$1D,$E9,$C9,$FC,$63,$F9,$1F,$94,$C4,$7E,$02,$2F
	.dc.b	$4F,$F8,$FF,$91,$5F,$80,$E9,$FC,$C7,$89,$30,$FB,$31,$97,$F3,$11
	.dc.b	$97,$F0,$1C,$69,$FF,$21,$7C,$03,$EE,$BF,$90,$37,$C8,$FF,$47,$B6
	.dc.b	$FF,$6A,$1C,$3F,$B8,$6B,$FC,$31,$AF,$FA,$F9,$AF,$19,$FE,$5D,$CF
	.dc.b	$F9,$2F,$BB,$FC,$D9,$27,$FC,$9D,$5D,$B5,$FF,$30,$79,$1E,$99,$FD
	.dc.b	$A7,$DC,$7E,$E4,$FE,$63,$9C,$8C,$90,$FD,$A4,$F1,$7F,$5B,$D5,$FC
	.dc.b	$07,$19,$1F,$BF,$8F,$E0,$86,$1F,$D9,$C6,$1F,$F3,$17,$95,$FD,$C3
	.dc.b	$E0,$FF,$98,$92,$3F,$98,$BF,$E3,$AE,$27,$D9,$7F,$0F,$24,$F2,$F1
	.dc.b	$FF,$DE,$4F,$F0,$FE,$7F,$E2,$84,$7F,$89,$4D,$7F,$82,$C5,$DB,$F0
	.dc.b	$D3,$FD,$1B,$F2,$97,$D0,$FF,$B1,$5B,$7F,$EC,$A6,$BF,$71,$AF,$F4
	.dc.b	$4E,$27,$E3,$2F,$AE,$FE,$86,$D9,$FE,$82,$D7,$F4,$15,$D3,$1F,$F1
	.dc.b	$DF,$ED,$5F,$D8,$67,$FA,$7E,$E8,$C4,$7F,$98,$9A,$FF,$8D,$9B,$E3
	.dc.b	$7F,$2A,$E4,$FE,$6C,$49,$FA,$33,$FD,$B6,$FF,$4D,$39,$9F,$E5,$2F
	.dc.b	$FA,$2A,$75,$E4,$7F,$39,$8D,$57,$0A,$E7,$D7,$FD,$B3,$E0,$7F,$B5
	.dc.b	$92,$3F,$B5,$BF,$6F,$E0,$3F,$A4,$12,$3E,$D4,$7F,$68,$FC,$0F,$F6
	.dc.b	$F1,$FA,$8E,$04,$62,$3F,$AF,$C6,$3B,$FF,$DF,$AD,$8E,$13,$B7,$EA
	.dc.b	$35,$1F,$DE,$5C,$36,$7F,$21,$E7,$FF,$5D,$24,$7F,$78,$9F,$16,$5F
	.dc.b	$E8,$7F,$67,$F8,$D7,$27,$EF,$AB,$FC,$B7,$E2,$7E,$11,$EF,$3A,$7F
	.dc.b	$81,$B5,$FE,$F2,$D7,$FC,$1A,$D7,$8C,$73,$7A,$CF,$E9,$7C,$AF,$EE
	.dc.b	$44,$8F,$F6,$75,$75,$BF,$E4,$53,$99,$FE,$CA,$FF,$B6,$E7,$5C,$1D
	.dc.b	$BE,$0B,$D5,$1F,$CD,$5B,$8F,$FB,$AE,$FD,$2D,$7D,$2D,$7D,$2D,$7A
	.dc.b	$1C,$23,$F7,$9F,$35,$FE,$09,$24,$7A,$27,$A6,$3F,$E3,$EF,$A9,$FF
	.dc.b	$1E,$E1,$F6,$3E,$2D,$F5,$3F,$E3,$29,$CE,$09,$CB,$F7,$A3,$31,$8E
	.dc.b	$87,$D5,$7F,$BD,$2B,$85,$F3,$F4,$3F,$D2,$37,$CA,$58,$72,$AD,$63
	.dc.b	$E3,$7F,$DA,$78,$73,$BE,$15,$E4,$BF,$F1,$AB,$5E,$27,$F4,$47,$0F
	.dc.b	$9F,$D2,$7C,$77,$E2,$E7,$FC,$40,$D8,$CC,$F1,$6B,$D9,$CD,$78,$9B
	.dc.b	$87,$FB,$A3,$C3,$FF,$5B,$F0,$FA,$6F,$C1,$85,$7D,$F7,$C1,$C9,$F1
	.dc.b	$8F,$F6,$1D,$72,$70,$70,$9E,$9D,$AF,$D7,$3F,$D7,$B8,$7E,$C3,$83
	.dc.b	$FD,$DD,$39,$DB,$FD,$E1,$1F,$AB,$FC,$11,$F1,$98,$BF,$DB,$DE,$DF
	.dc.b	$CE,$CE,$9F,$D5,$B7,$2F,$D5,$B5,$FD,$5B,$5F,$D9,$B5,$FF,$68,$B7
	.dc.b	$03,$4D,$57,$AF,$E3,$BC,$94,$8C,$5F,$93,$C4,$CC,$62,$7F,$CB,$99
	.dc.b	$AF,$2D,$BE,$07,$C5,$3D,$EE,$65,$9F,$C5,$C3,$FD,$43,$1E,$A6,$BE
	.dc.b	$A6,$E5,$EA,$6E,$5F,$D4,$F0,$F8,$F1,$82,$3F,$E1,$6C,$BE,$C7,$B9
	.dc.b	$E6,$3D,$FD,$CF,$AF,$7D,$9E,$6F,$FD,$37,$18,$7F,$9C,$37,$3F,$E6
	.dc.b	$ED,$7F,$B3,$35,$FE,$67,$5D,$6B,$D6,$DA,$E7,$18,$AE,$15,$F4,$75
	.dc.b	$FE,$07,$5F,$26,$31,$C7,$F5,$BA,$3F,$DF,$BD,$BF,$42,$31,$E5,$7F
	.dc.b	$7E,$BF,$F5,$FC,$FC,$AA,$DC,$60,$F8,$98,$3E,$8B,$EC,$9D,$7C,$77
	.dc.b	$8B,$8A,$F1,$70,$7E,$86,$B1,$A7,$DA,$9F,$EF,$15,$8E,$4F,$D1,$E3
	.dc.b	$D6,$78,$B6,$B8,$1F,$DF,$6B,$84,$F2,$3F,$C9,$B9,$7F,$81,$63,$D5
	.dc.b	$97,$92,$72,$DA,$FE,$D3,$E4,$BE,$73,$EE,$B6,$BF,$CE,$BF,$51,$FE
	.dc.b	$EA,$7F,$C2,$F3,$FA,$0A,$4E,$5F,$F2,$EB,$5F,$F6,$9B,$5F,$F7,$3B
	.dc.b	$5F,$B1,$F4,$7F,$A2,$3F,$47,$F3,$36,$FE,$0C,$59,$FD,$4C,$DD,$7E
	.dc.b	$3B,$C2,$BF,$C9,$7F,$50,$F8,$6C,$5F,$C8,$D9,$39,$AF,$8E,$D9,$FF
	.dc.b	$74,$CE,$AB,$F0,$53,$99,$FF,$7F,$F0,$FF,$1B,$F9,$5F,$DC,$FC,$87
	.dc.b	$EE,$9B,$8F,$D8,$6D,$7F,$00,$6E,$7F,$C8,$31,$75,$7F,$02,$F9,$7F
	.dc.b	$E0,$F1,$8C,$3F,$03,$18,$9D,$47,$E3,$27,$58,$7F,$0C,$AF,$DA,$FD
	.dc.b	$D9,$A7,$F1,$6F,$6D,$F2,$EF,$63,$38,$FF,$A1,$A3,$0F,$FA,$5A,$77
	.dc.b	$3F,$C3,$1B,$19,$9D,$E1,$FE,$99,$6B,$FE,$99,$6B,$FE,$99,$6B,$FE
	.dc.b	$A1,$6F,$7B,$17,$17,$F2,$4F,$8B,$F5,$B7,$0F,$DE,$56,$E3,$18,$7F
	.dc.b	$BC,$EF,$8D,$7E,$A2,$FF,$96,$F8,$12,$3F,$57,$F4,$36,$FF,$99,$CD
	.dc.b	$E0,$FD,$8C,$DA,$FF,$22,$D3,$EB,$5E,$A7,$E8,$5B,$3F,$ED,$AD,$1F
	.dc.b	$15,$9A,$FD,$34,$62,$7F,$CA,$B3,$FC,$DA,$D3,$97,$FC,$BD,$19,$8C
	.dc.b	$47,$F9,$7A,$30,$49,$FE,$27,$C1,$FD,$2B,$EE,$BF,$A5,$B6,$3F,$84
	.dc.b	$B6,$BF,$A8,$B7,$34,$FF,$30,$FC,$0D,$E7,$7F,$54,$FB,$0F,$E7,$F1
	.dc.b	$89,$FE,$27,$18,$9D,$47,$D4,$9D,$61,$FD,$7A,$BF,$27,$E3,$FF,$5E
	.dc.b	$DA,$DF,$F1,$C6,$D7,$D9,$9F,$E9,$BF,$02,$FA,$AC,$E7,$E6,$AF,$8F
	.dc.b	$C9,$FC,$F9,$E1,$8F,$E1,$2F,$83,$F8,$69,$D5,$7F,$2C,$F6,$3F,$B7
	.dc.b	$9F,$AE,$C6,$D7,$F2,$E8,$B8,$7F,$2E,$EE,$FE,$15,$58,$BF,$E6,$BE
	.dc.b	$E7,$F7,$83,$FB,$CF,$95,$71,$FE,$6F,$17,$0F,$EF,$F3,$6B,$D5,$0F
	.dc.b	$D4,$5F,$19,$FC,$AD,$B8,$7F,$59,$97,$E0,$9F,$E2,$87,$FA,$09,$C6
	.dc.b	$BF,$99,$46,$63,$F2,$E7,$0F,$4B,$F1,$D1,$98,$FC,$DB,$5F,$F5,$0B
	.dc.b	$5E,$15,$FE,$99,$8C,$1B,$D3,$FC,$75,$F1,$3F,$E1,$83,$FC,$3B,$E1
	.dc.b	$FF,$AA,$AB,$FA,$BC,$7F,$6F,$F1,$9F,$59,$FC,$A8,$93,$FC,$C1,$FD
	.dc.b	$17,$D6,$7E,$23,$6F,$F2,$1F,$D7,$3E,$77,$FA,$74,$E6,$FF,$C8,$33
	.dc.b	$97,$F9,$17,$95,$FC,$A0,$FE,$FD,$F5,$BF,$B4,$DF,$ED,$79,$49,$CF
	.dc.b	$87,$FD,$F6,$B5,$5F,$EE,$C6,$BF,$EC,$4F,$31,$18,$84,$62,$79,$13
	.dc.b	$0F,$EE,$35,$BE,$17,$C6,$7E,$F7,$6F,$F2,$FB,$6B,$B7,$D6,$FF,$69
	.dc.b	$9C,$F8,$BF,$E1,$78,$FE,$CD,$91,$A4,$F2,$7F,$C2,$CD,$BF,$F0,$0B
	.dc.b	$74,$FC,$1E,$5F,$E9,$EF,$3F,$FC,$A7,$F0,$27,$FC,$49,$F7,$CF,$A5
	.dc.b	$CC,$61,$FF,$11,$46,$71,$FF,$06,$B7,$0F,$EC,$ED,$7F,$93,$E6,$8C
	.dc.b	$47,$CB,$6B,$F2,$9B,$1B,$AF,$F2,$CE,$2F,$A6,$7A,$7F,$E6,$A6,$E3
	.dc.b	$FE,$3D,$B7,$D6,$FE,$0F,$2F,$DC,$56,$9F,$F0,$04,$65,$F2,$B5,$FB
	.dc.b	$CD,$7E,$EB,$67,$AF,$47,$F9,$0E,$30,$47,$CF,$F0,$0F,$BF,$FF,$3F
	.dc.b	$B7,$1F,$36,$4F,$9B,$CD,$7F,$94,$E3,$13,$FE,$38,$6E,$5F,$E1,$26
	.dc.b	$BF,$31,$AF,$E5,$5A,$FE,$51,$B9,$7C,$8B,$FF,$2D,$46,$B4,$FF,$65
	.dc.b	$F9,$7F,$E9,$4F,$84,$FE,$B9,$8F,$CB,$FC,$D3,$FC,$C5,$97,$F2,$19
	.dc.b	$CB,$FE,$29,$8C,$BF,$79,$AF,$E9,$DA,$FE,$99,$B9,$1A,$4F,$1F,$FC
	.dc.b	$DC,$F0,$C7,$FD,$F1,$F0,$3F,$EA,$8F,$87,$3F,$E7,$9F,$61,$F6,$19
	.dc.b	$8C,$3F,$E7,$78,$CE,$9F,$E6,$26,$E7,$FE,$3C,$6B,$FE,$3A,$FC,$AF
	.dc.b	$DF,$DB,$FE,$D2,$6D,$7F,$B3,$97,$8F,$FE,$BA,$73,$5F,$EB,$F6,$E6
	.dc.b	$69,$C2,$BF,$D0,$2E,$BD,$CF,$F5,$16,$1F,$EE,$06,$CF,$FB,$71,$7B
	.dc.b	$5F,$EC,$87,$38,$7F,$B3,$5B,$81,$A7,$0A,$FF,$47,$BC,$27,$FC,$33
	.dc.b	$A9,$FD,$67,$C4,$7F,$46,$E2,$FE,$D9,$D5,$FE,$EE,$9F,$E9,$35,$B8
	.dc.b	$C4,$71,$9D,$D7,$F4,$E8,$56,$E3,$A6,$16,$79,$9E,$03,$FC,$57,$5C
	.dc.b	$9C,$F4,$76,$51,$FD,$FE,$31,$1D,$8F,$F7,$45,$76,$69,$DB,$0E,$55
	.dc.b	$D6,$3E,$09,$E1,$3D,$57,$E5,$73,$3C,$23,$11,$F1,$F0,$B5,$CE,$30
	.dc.b	$7C,$12,$96,$30,$E0,$FF,$24,$FB,$BF,$E7,$3F,$AC,$E0,$D8,$F9,$4D
	.dc.b	$7F,$BC,$B5,$E3,$E3,$FF,$C0,$07,$6D,$76,$D7,$7B,$F6,$8E,$1A,$7D
	.dc.b	$AA,$FE,$57,$5F,$83,$8C,$74,$FF,$97,$9D,$57,$0D,$79,$D7,$CE,$3F
	.dc.b	$D1,$35,$B8,$F6,$B6,$3F,$1D,$7F,$DB,$F0,$ED,$5C,$B0,$EF,$58,$37
	.dc.b	$7C,$5F,$F4,$AE,$1E,$17,$FC,$CB,$FA,$47,$F2,$97,$F2,$1E,$F2,$31
	.dc.b	$1F,$E2,$08,$C7,$67,$D5,$6C,$70,$9D,$BF,$E7,$C8,$CB,$C6,$32,$FE
	.dc.b	$9E,$32,$EF,$F4,$8F,$F0,$A3,$E2,$6B,$E1,$6B,$BE,$FF,$14,$62,$3A
	.dc.b	$35,$FF,$40,$B6,$4E,$61,$F8,$5B,$E8,$7F,$1A,$D8,$FC,$93,$5F,$DA
	.dc.b	$35,$FF,$21,$B5,$FF,$4D,$B5,$FE,$08,$D7,$BC,$76,$FA,$E9,$1D,$9A
	.dc.b	$FF,$93,$B7,$7E,$09,$FC,$DA,$3F,$82,$46,$3B,$BF,$4C,$D8,$CC,$ED
	.dc.b	$FE,$15,$F6,$4F,$96,$FD,$4B,$63,$F5,$CD,$77,$1A,$23,$A6,$BF,$C5
	.dc.b	$7F,$19,$FE,$E8,$FE,$DD,$AE,$7C,$6F,$DB,$C6,$23,$FC,$D3,$5F,$E2
	.dc.b	$37,$C3,$E4,$79,$08,$E0,$87,$A9,$CB,$F9,$3F,$C1,$7F,$D5,$6F,$E6
	.dc.b	$1A,$E7,$E1,$7F,$2D,$18,$8F,$F3,$CF,$E6,$2F,$8C,$FE,$16,$D8,$FE
	.dc.b	$20,$D7,$F9,$93,$5E,$31,$C0,$8F,$07,$DF,$F4,$92,$3E,$23,$5F,$E7
	.dc.b	$9F,$92,$A3,$F8,$DB,$63,$F9,$03,$5D,$C6,$88,$E9,$AF,$C3,$F5,$3D
	.dc.b	$E4,$7E,$23,$FC,$A3,$F1,$EF,$ED,$F1,$FC,$B5,$AF,$F4,$F6,$BF,$E9
	.dc.b	$E6,$BF,$EF,$66,$BF,$D7,$DA,$9A,$63,$CA,$7D,$2F,$F3,$86,$CF,$F3
	.dc.b	$D6,$BF,$D7,$DA,$FF,$AB,$BF,$07,$7D,$67,$F4,$06,$C7,$F4,$56,$BF
	.dc.b	$D9,$DA,$F1,$8E,$04,$78,$3F,$BE,$CE,$47,$EF,$9A,$FF,$71,$FC,$37
	.dc.b	$DA,$4F,$3B,$FA,$8B,$F0,$2D,$F3,$5E,$52,$57,$95,$FE,$AF,$F5,$DE
	.dc.b	$48,$FF,$03,$07,$0F,$F8,$25,$AE,$71,$FE,$BF,$FA,$D3,$E5,$69,$27
	.dc.b	$FB,$34,$65,$B1,$97,$FC,$15,$19,$6C,$7F,$6A,$F1,$7F,$DC,$B1,$FD
	.dc.b	$BD,$AF,$F8,$45,$AF,$FB,$81,$AF,$AB,$F3,$77,$91,$FD,$D1,$B1,$FD
	.dc.b	$E5,$AF,$F8,$6D,$AF,$FB,$A9,$AF,$38,$E4,$47,$97,$B1,$D4,$8E,$08
	.dc.b	$75,$72,$D3,$FC,$15,$F8,$87,$F0,$F7,$FC,$58,$D7,$FD,$F6,$D7,$EA
	.dc.b	$B5,$FF,$1E,$35,$34,$9D,$5F,$E1,$4F,$D9,$7F,$71,$3F,$78,$38,$7F
	.dc.b	$02,$D7,$FC,$90,$D5,$F9,$43,$EC,$7F,$C3,$ED,$8F,$F1,$23,$5F,$F2
	.dc.b	$33,$5F,$03,$5F,$C7,$B5,$FF,$2B,$35,$EF,$1F,$E2,$CF,$B1,$7C,$75
	.dc.b	$24,$FF,$8C,$E3,$2D,$8C,$BF,$E5,$78,$CB,$63,$FC,$6F,$DE,$F9,$3A
	.dc.b	$49,$FF,$1F,$46,$5B,$19,$7F,$CC,$51,$96,$C7,$9F,$E4,$A4,$79,$5A
	.dc.b	$FF,$9A,$7E,$E7,$FB,$13,$F4,$DA,$FE,$C9,$AF,$F9,$E5,$A9,$A4,$F2
	.dc.b	$7F,$CA,$2E,$BF,$26,$F9,$44,$CF,$FC,$AA,$DA,$FF,$2D,$35,$FF,$3E
	.dc.b	$3F,$47,$F1,$77,$D7,$7F,$97,$1B,$3F,$E6,$16,$BF,$E8,$36,$BF,$8D
	.dc.b	$6B,$FC,$31,$AF,$FA,$4D,$A9,$BF,$0C,$F9,$3F,$F3,$5B,$67,$FC,$E0
	.dc.b	$D7,$FD,$24,$D7,$F3,$8F,$D3,$6E,$9C,$88,$F2,$6B,$C9,$AF,$E9,$7E
	.dc.b	$0D,$F3,$5F,$E7,$C6,$C7,$FA,$05,$AF,$FA,$8D,$AF,$18,$E0,$47,$83
	.dc.b	$FE,$86,$FC,$3B,$FD,$3B,$1F,$EB,$56,$A6,$93,$CE,$E0,$48,$FF,$49
	.dc.b	$47,$F8,$65,$FF,$1E,$FB,$3F,$E4,$62,$3F,$E4,$66,$BF,$E4,$66,$BF
	.dc.b	$E4,$67,$FD,$33,$F6,$2F,$A0,$A4,$9F,$F5,$04,$65,$FF,$39,$FC,$48
	.dc.b	$FF,$38,$15,$CB,$83,$AF,$81,$23,$9B,$5F,$F3,$D7,$E6,$EF,$B2,$FF
	.dc.b	$3D,$36,$3F,$D0,$2D,$77,$1A,$23,$A6,$BA,$30,$43,$07,$FB,$6F,$EB
	.dc.b	$7F,$C9,$CF,$F2,$56,$BF,$D9,$1A,$FF,$BA,$DB,$EE,$BF,$80,$1F,$87
	.dc.b	$FF,$A4,$5B,$5F,$E9,$86,$BF,$E9,$76,$BF,$E7,$63,$FC,$FA,$43,$FC
	.dc.b	$FA,$53,$FC,$FA,$53,$81,$FE,$A4,$FD,$07,$FC,$43,$46,$93,$CE,$FF
	.dc.b	$6A,$E0,$FB,$7A,$4B,$FF,$6E,$46,$5B,$19,$7E,$33,$5F,$F7,$DF,$E2
	.dc.b	$2F,$AF,$FF,$72,$7C,$5A,$31,$3F,$D3,$CF,$F7,$EB,$87,$FA,$94,$6E
	.dc.b	$78,$CE,$67,$AC,$60,$8F,$9B,$E9,$BD,$64,$70,$43,$EE,$39,$69,$9F
	.dc.b	$81,$7C,$6F,$FA,$FF,$F3,$CF,$E3,$7F,$1C,$FA,$6C,$78,$5F,$E2,$48
	.dc.b	$C6,$1F,$81,$C7,$FC,$4F,$8F,$28,$FC,$27,$DA,$7F,$52,$AD,$C3,$74
	.dc.b	$E7,$1F,$C6,$36,$5F,$2E,$F9,$A7,$95,$E1,$BD,$F9,$8B,$65,$DB,$2E
	.dc.b	$B3,$98,$C6,$5D,$72,$E9,$8F,$3C,$3C,$F7,$F8,$0A,$FE,$CB,$1D,$8F
	.dc.b	$D2,$38,$7F,$41,$E6,$38,$FF,$0E,$D8,$FD,$53,$5F,$D5,$35,$FD,$53
	.dc.b	$BC,$7D,$0D,$CF,$D9,$93,$ED,$1F,$7B,$F3,$6F,$78,$FB,$6F,$9B,$71
	.dc.b	$F2,$B7,$0E,$57,$C6,$B7,$5F,$21,$FF,$2F,$FB,$7F,$E7,$AE,$4F,$99
	.dc.b	$C3,$F6,$FA,$9C,$AF,$BB,$E5,$8F,$D1,$C3,$82,$BE,$15,$BC,$3C,$93
	.dc.b	$FE,$72,$F6,$FF,$0C,$FF,$8C,$6F,$8B,$63,$2F,$91,$AF,$76,$BC,$E7
	.dc.b	$F5,$8D,$3F,$5D,$3D,$1F,$C3,$62,$FC,$41,$EC,$74,$8C,$36,$33,$7F
	.dc.b	$B7,$9C,$B6,$33,$5F,$E4,$99,$CB,$63,$2F,$FA,$2A,$32,$FB,$FE,$9F
	.dc.b	$D3,$81,$A4,$BF,$5C,$65,$B1,$99,$F6,$BE,$9F,$B9,$7E,$13,$49,$5E
	.dc.b	$98,$CB,$F1,$BE,$AD,$F4,$9F,$29,$B1,$98,$C1,$1C,$13,$0F,$E3,$53
	.dc.b	$C1,$F8,$5E,$17,$FA,$3A,$7E,$43,$63,$3E,$F7,$CF,$8C,$47,$E0,$7E
	.dc.b	$17,$E1,$92,$BE,$0B,$F5,$7F,$32,$FD,$74,$7D,$36,$CF,$D1,$6B,$F3
	.dc.b	$DA,$FE,$01,$AB,$64,$61,$BD,$1D,$C9,$5D,$DB,$5D,$DF,$C8,$FD,$A3
	.dc.b	$1A,$4A,$FC,$9C,$65,$FC,$57,$E5,$2E,$DF,$C8,$36,$33,$18,$3A,$FD
	.dc.b	$94,$8E,$6D,$C7,$E8,$F5,$7D,$92,$4F,$DD,$6C,$6A,$30,$47,$FA,$96
	.dc.b	$1F,$CF,$63,$CF,$3E,$C2,$3E,$AF,$67,$DE,$47,$07,$F5,$BF,$C9,$1F
	.dc.b	$40,$FB,$9F,$DB,$C6,$1B,$19,$9E,$0D,$C7,$FD,$65,$5B,$9C,$D7,$02
	.dc.b	$3C,$3E,$3B,$99,$1C,$12,$B9,$DA,$75,$18,$7F,$85,$F7,$3E,$21,$F5
	.dc.b	$B8,$6E,$1F,$DD,$E7,$53,$93,$F8,$8F,$EB,$2F,$9B,$A4,$AF,$E2,$71
	.dc.b	$96,$C6,$71,$FF,$01,$4E,$5B,$1F,$1B,$E4,$E5,$7C,$66,$CF,$B1,$AF
	.dc.b	$F6,$11,$AF,$D2,$8C,$C7,$F6,$58,$FE,$CD,$F8,$6F,$D4,$83,$49,$5F
	.dc.b	$6E,$32,$D8,$CC,$7D,$63,$27,$FB,$A7,$E9,$9F,$A9,$7F,$88,$6B,$FC
	.dc.b	$3D,$18,$F4,$FF,$9C,$B6,$33,$3B,$AF,$F1,$27,$EB,$2F,$AA,$FE,$7A
	.dc.b	$DA,$FE,$7A,$D9,$DC,$E8,$93,$A6,$C7,$9F,$E9,$FE,$01,$27,$DC,$D8
	.dc.b	$D4,$60,$8F,$BF,$B3,$E4,$12,$3E,$33,$63,$51,$82,$3F,$E3,$CF,$D0
	.dc.b	$1F,$82,$7F,$5A,$A4,$62,$BF,$C8,$91,$89,$D4,$E6,$BF,$C9,$1D,$1F
	.dc.b	$E0,$2B,$FE,$BE,$D8,$CF,$9D,$FD,$92,$31,$1F,$85,$3F,$D9,$51,$92
	.dc.b	$4F,$27,$FB,$6A,$FC,$ED,$F7,$D8,$6C,$7F,$96,$4D,$46,$48,$E6,$37
	.dc.b	$5E,$18,$D9,$99,$E9,$F3,$92,$38,$36,$78,$52,$31,$1F,$9F,$9F,$34
	.dc.b	$62,$3F,$C4,$FF,$41,$FE,$B1,$9F,$F3,$77,$A6,$8C,$4F,$C3,$8C,$7C
	.dc.b	$2F,$F8,$31,$B5,$FE,$0A,$F5,$BF,$9C,$57,$F9,$DD,$AF,$CB,$6B,$FB
	.dc.b	$16,$BF,$E5,$36,$BF,$E3,$9F,$1B,$FC,$2F,$18,$33,$E5,$BD,$8F,$EE
	.dc.b	$44,$BF,$EE,$CF,$F8,$7B,$F1,$17,$D9,$D2,$4F,$F8,$82,$32,$FF,$88
	.dc.b	$BE,$F5,$E8,$FF,$89,$9B,$19,$8C,$19,$FA,$79,$1C,$5B,$A7,$FA,$5F
	.dc.b	$DB,$7C,$07,$F8,$C1,$B5,$FE,$30,$6C,$EE,$74,$49,$D3,$63,$F0,$9F
	.dc.b	$59,$F2,$24,$D2,$6F,$FE,$3E,$87,$C5,$52,$3F,$E4,$03,$25,$73,$5F
	.dc.b	$97,$6B,$FE,$33,$FB,$77,$D5,$FF,$90,$5B,$5F,$E3,$F6,$CF,$F8,$F5
	.dc.b	$AF,$EA,$FE,$EB,$FD,$E9,$3F,$E5,$06,$C7,$FA,$F5,$AE,$FE,$33,$FC
	.dc.b	$AF,$18,$8F,$E7,$0D,$7F,$D3,$4D,$4E,$65,$F9,$43,$EE,$BF,$6E,$D8
	.dc.b	$FD,$A3,$6B,$F6,$0D,$7F,$81,$72,$38,$3F,$CD,$2D,$AF,$F3,$4B,$67
	.dc.b	$FD,$AC,$D7,$8C,$F0,$24,$F0,$7E,$57,$E3,$12,$3E,$53,$67,$9F,$77
	.dc.b	$80,$93,$E0,$6C,$6A,$30,$47,$F8,$D7,$9E,$FC,$55,$F2,$1F,$DE,$9B
	.dc.b	$5F,$DD,$DB,$5C,$1A,$FF,$57,$F7,$7F,$D1,$3E,$59,$A4,$AF,$F4,$5C
	.dc.b	$65,$B1,$99,$FE,$BA,$F8,$7E,$5E,$4F,$85,$B3,$FE,$FF,$EA,$7D,$07
	.dc.b	$EC,$C9,$5F,$B7,$7F,$6D,$F7,$F2,$3F,$74,$DA,$E4,$D7,$F9,$E5,$23
	.dc.b	$11,$FE,$9E,$FD,$27,$F1,$9F,$00,$69,$39,$BF,$D4,$0D,$C3,$FD,$3C
	.dc.b	$DA,$FF,$4E,$7D,$AF,$ED,$27,$FD,$44,$D8,$CF,$67,$FA,$A6,$31,$1F
	.dc.b	$E9,$F6,$BF,$EB,$06,$BF,$E7,$16,$E0,$73,$1F,$E1,$BE,$EF,$F7,$E7
	.dc.b	$C1,$BC,$FF,$EB,$5E,$7B,$2C,$C7,$F1,$5B,$56,$DF,$F1,$25,$6D,$CC
	.dc.b	$EE,$75,$03,$FD,$16,$E1,$C5,$F6,$79,$DD,$7C,$C7,$43,$25,$33,$3C
	.dc.b	$EF,$C4,$F5,$85,$7F,$B6,$1E,$E5,$70,$E8,$FE,$F4,$3E,$DF,$F9,$1F
	.dc.b	$BB,$FE,$E0,$6C,$69,$AF,$F8,$05,$AF,$F9,$1D,$7D,$4F,$FB,$DE,$38
	.dc.b	$35,$E0,$D7,$83,$5F,$91,$1C,$BA,$3E,$22,$FB,$7F,$F2,$A6,$3F,$3A
	.dc.b	$31,$5A,$BF,$04,$E5,$BA,$79,$6F,$8B,$6B,$86,$11,$F9,$D8,$74,$6D
	.dc.b	$70,$7C,$B7,$D1,$FC,$87,$CE,$7C,$A6,$9F,$9C,$71,$BF,$E4,$DB,$B9
	.dc.b	$A4,$9E,$07,$D4,$97,$DB,$FF,$5A,$8C,$35,$FF,$19,$46,$B8,$7B,$CE
	.dc.b	$67,$FC,$F5,$39,$78,$CE,$88,$FA,$7E,$DF,$FA,$CB,$7C,$5F,$C5,$46
	.dc.b	$23,$F0,$ED,$7F,$C8,$4D,$7F,$D1,$0D,$7F,$44,$D4,$D3,$FD,$3C,$F8
	.dc.b	$7D,$EE,$E4,$7F,$D4,$DF,$6E,$FA,$9F,$9D,$C0,$8C,$4E,$A7,$31,$88
	.dc.b	$FC,$EF,$DB,$7F,$55,$7E,$B3,$5C,$F8,$9F,$A7,$8C,$47,$F9,$75,$AF
	.dc.b	$FA,$8D,$AF,$EF,$5A,$9B,$F6,$EF,$A4,$FD,$93,$67,$F0,$CD,$7F,$62
	.dc.b	$D7,$C1,$F5,$DE,$32,$3E,$86,$BE,$C6,$BC,$0F,$C4,$7D,$BB,$E3,$BF
	.dc.b	$22,$D8,$CC,$60,$FE,$0D,$EC,$FF,$57,$8F,$CD,$35,$FE,$0A,$D7,$FD
	.dc.b	$00,$D7,$FD,$8C,$D7,$F9,$03,$5F,$F3,$37,$8B,$D8,$F8,$7F,$6D,$23
	.dc.b	$E1,$B5,$FE,$27,$F6,$E8,$FE,$34,$D8,$FD,$73,$5D,$C6,$88,$E9,$AF
	.dc.b	$E9,$B0,$7C,$47,$C4,$24,$7C,$87,$F9,$47,$DB,$BD,$5A,$49,$FD,$F4
	.dc.b	$65,$B1,$97,$F9,$5C,$65,$B1,$FC,$CF,$DE,$7C,$6D,$24,$FF,$09,$8C
	.dc.b	$BF,$C2,$FE,$D9,$F3,$7F,$CE,$DB,$19,$8C,$11,$C1,$0F,$E7,$B1,$C1
	.dc.b	$FD,$CE,$BF,$BC,$23,$FC,$05,$FE,$91,$F6,$FE,$98,$FF,$5F,$35,$FE
	.dc.b	$B8,$1C,$3F,$D7,$DA,$A5,$F3,$5E,$92,$47,$B5,$FE,$AD,$F6,$FE,$58
	.dc.b	$FF,$66,$35,$EE,$D7,$FB,$63,$57,$DB,$FF,$9E,$9F,$E6,$4D,$7F,$AE
	.dc.b	$B5,$FF,$6C,$35,$F5,$B5,$FE,$F0,$D7,$AF,$BB,$FD,$AE,$31,$E2,$FF
	.dc.b	$89,$67,$FA,$03,$5F,$ED,$6D,$7F,$DC,$CD,$7E,$4F,$C9,$FF,$83,$1F
	.dc.b	$E9,$0D,$7F,$B9,$B5,$FF,$77,$35,$E3,$E5,$7F,$7D,$8C,$47,$F8,$3D
	.dc.b	$A9,$A3,$E7,$B8,$12,$3C,$2D,$7C,$2D,$7C,$2F,$F8,$2F,$ED,$DD,$54
	.dc.b	$93,$FD,$86,$32,$D8,$CB,$FE,$0E,$8C,$B6,$3F,$C2,$DF,$6F,$FB,$B8
	.dc.b	$FC,$68,$70,$FE,$41,$AF,$F8,$B5,$AB,$ED,$DF,$6D,$FE,$21,$6C,$7F
	.dc.b	$75,$6B,$FE,$20,$6B,$E0,$6B,$F9,$E6,$BF,$E3,$B6,$BD,$E3,$FC,$5D
	.dc.b	$EC,$7F,$BB,$E3,$FC,$04,$D7,$FC,$5A,$D7,$DA,$D7,$F5,$6D,$7F,$C9
	.dc.b	$0D,$7F,$CF,$BE,$FF,$63,$FC,$79,$DE,$FB,$3A,$49,$FF,$09,$46,$5B
	.dc.b	$19,$7F,$C7,$F1,$96,$C7,$AB,$E9,$E4,$7A,$9A,$FF,$91,$BE,$DF,$FB
	.dc.b	$63,$F5,$9A,$FF,$04,$6B,$FE,$61,$6A,$69,$3C,$8F,$F2,$B3,$AF,$B7
	.dc.b	$F9,$A4,$CF,$FC,$B4,$DA,$FF,$17,$35,$FF,$2C,$BF,$53,$ED,$DF,$5B
	.dc.b	$FE,$61,$6C,$FF,$8D,$5A,$FF,$98,$1A,$FE,$41,$AF,$F1,$E6,$BF,$E7
	.dc.b	$B6,$A6,$FD,$BB,$E5,$7F,$CE,$0D,$9F,$F2,$23,$5F,$F3,$7B,$5F,$CF
	.dc.b	$BF,$4D,$BA,$72,$23,$C9,$AF,$26,$BF,$A8,$FB,$77,$C1,$FF,$A0,$5B
	.dc.b	$1F,$E5,$56,$BF,$E8,$06,$BC,$63,$81,$1E,$0F,$FA,$2F,$ED,$FF,$D1
	.dc.b	$D1,$FE,$9D,$6A,$69,$3C,$DE,$04,$8F,$F3,$34,$7F,$87,$DF,$F2,$3F
	.dc.b	$B1,$FE,$4E,$23,$FE,$53,$6B,$FE,$50,$6B,$FE,$50,$7F,$D3,$DF,$6C
	.dc.b	$FB,$7A,$49,$FF,$3A,$46,$5F,$F3,$67,$CB,$8F,$F3,$A9,$5C,$B8,$3A
	.dc.b	$F8,$32,$39,$B5,$FF,$38,$FD,$BB,$E1,$BF,$CF,$ED,$8F,$F3,$D3,$5D
	.dc.b	$C6,$88,$E9,$AE,$8C,$10,$C1,$FE,$B8,$FB,$2F,$F2,$89,$3D,$6F,$F6
	.dc.b	$03,$FE,$56,$F7,$5F,$6E,$F8,$DF,$F4,$9B,$6F,$FD,$1C,$D7,$FD,$16
	.dc.b	$D7,$FC,$FA,$7F,$A0,$88,$7F,$A1,$CA,$7F,$A1,$0A,$70,$3F,$D3,$5F
	.dc.b	$6F,$FE,$06,$A3,$49,$E9,$7A,$3E,$DE,$57,$FA,$AF,$14,$62,$3E,$03
	.dc.b	$5F,$E8,$F1,$A7,$84,$7F,$AC,$FE,$25,$F6,$EF,$A5,$C3,$6B,$FD,$C3
	.dc.b	$3A,$9C,$92,$33,$3B,$7F,$AA,$CE,$DC,$CF,$D0,$F9,$89,$1E,$96,$BF
	.dc.b	$5D,$AF,$A1,$AF,$F8,$B3,$04,$7F,$8B,$BE,$05,$F1,$AA,$6A,$FB,$4F
	.dc.b	$E3,$7F,$1C,$FA,$CD,$4E,$7C,$3F,$F1,$BC,$E6,$F9,$67,$D9,$E9,$97
	.dc.b	$67,$3F,$6D,$FD,$6E,$B8,$42,$93,$9C,$7D,$0D,$97,$D0,$BE,$DD,$E5
	.dc.b	$F8,$AF,$7E,$6F,$AF,$2E,$B9,$74,$8C,$4E,$B1,$57,$A7,$1E,$98,$73
	.dc.b	$9F,$3D,$FE,$1F,$2F,$ED,$58,$F3,$9F,$D3,$46,$3C,$D7,$1F,$E3,$5B
	.dc.b	$3F,$AB,$6B,$F8,$C6,$BF,$55,$DE,$5E,$86,$E7,$F5,$73,$7D,$13,$E6
	.dc.b	$7E,$65,$EF,$1F,$6E,$3F,$4B,$1E,$2D,$8C,$E1,$CA,$F8,$E3,$E1,$9E
	.dc.b	$35,$FE,$6C,$F7,$FE,$C7,$07,$D9,$E2,$FC,$27,$DB,$F0,$AF,$C4,$7B
	.dc.b	$BF,$E8,$AE,$6F,$BE,$7E,$1B,$99,$5B,$AE,$93,$FE,$32,$78,$E3,$F8
	.dc.b	$57,$FC,$6B,$7C,$DD,$35,$CC,$6D,$FF,$43,$C6,$5B,$19,$7A,$CF,$7C
	.dc.b	$1F,$6F,$FE,$57,$8C,$36,$7F,$C8,$B3,$AF,$67,$93,$6B,$71,$8A,$E1
	.dc.b	$5F,$E9,$8A,$D3,$CA,$32,$F9,$BE,$DF,$C8,$FF,$A6,$1B,$89,$A4,$C3
	.dc.b	$E3,$4E,$DB,$3B,$6C,$F9,$BE,$89,$F6,$F4,$93,$F0,$63,$2F,$4F,$A6
	.dc.b	$7D,$BF,$91,$B1,$98,$C1,$1C,$13,$1F,$B5,$3C,$1F,$93,$E1,$3E,$03
	.dc.b	$E4,$B6,$BE,$93,$67,$73,$F0,$BE,$1A,$57,$B9,$B3,$F5,$3E,$DF,$F5
	.dc.b	$B3,$F6,$DB,$19,$E5,$FA,$F1,$88,$FA,$8D,$7F,$8C,$B5,$FE,$3E,$DE
	.dc.b	$03,$49,$3D,$09,$5F,$53,$ED,$7F,$DF,$4D,$CB,$EE,$C6,$30,$FC,$F3
	.dc.b	$5F,$C3,$7D,$BB,$AF,$F1,$AD,$BF,$CB,$36,$7F,$26,$D7,$FD,$33,$4F
	.dc.b	$B7,$EE,$4A,$F2,$3F,$6F,$AA,$47,$DF,$6D,$7D,$86,$BF,$EA,$4D,$FF
	.dc.b	$9F,$4E,$CF,$74,$FB,$08,$E0,$F8,$5E,$D7,$C0,$23,$83,$F7,$DF,$68
	.dc.b	$FA,$07,$DB,$FE,$D6,$30,$D8,$CC,$F0,$6E,$3F,$EB,$8A,$DC,$E6,$B8
	.dc.b	$11,$E1,$F1,$FC,$C8,$E0,$95,$CE,$D3,$A8,$C3,$FC,$27,$BF,$FD,$BD
	.dc.b	$F6,$F7,$E2,$8D,$26,$1F,$B6,$FB,$77,$CD,$FE,$F9,$B1,$98,$C1,$1C
	.dc.b	$13,$1F,$E4,$93,$C1,$F3,$7C,$A7,$C5,$24,$FC,$56,$C6,$A3,$04,$7E
	.dc.b	$40,$FB,$7F,$A0,$D8,$FA,$AD,$37,$1A,$24,$69,$B1,$FE,$EA,$FA,$87
	.dc.b	$DB,$FF,$88,$6B,$F9,$E4,$63,$D3,$FE,$64,$D8,$CC,$EE,$BF,$A1,$7D
	.dc.b	$BB,$ED,$FF,$92,$B6,$BF,$90,$B6,$77,$3A,$24,$E9,$B1,$F0,$3E,$A3
	.dc.b	$DE,$49,$F7,$B6,$35,$18,$23,$F2,$3B,$7E,$39,$23,$E3,$B6,$35,$18
	.dc.b	$23,$FE,$3A,$FB,$67,$DB,$FF,$53,$A4,$62,$BF,$C8,$AD,$AE,$15,$FD
	.dc.b	$1B,$A3,$FB,$F5,$7F,$42,$6C,$FF,$69,$6B,$FE,$39,$3F,$C7,$4E,$1C
	.dc.b	$35,$CF,$9D,$D4,$93,$FD,$CC,$7D,$B3,$ED,$FF,$B5,$46,$1B,$19,$8E
	.dc.b	$0D,$AF,$11,$FE,$4C,$70,$F9,$63,$91,$C2,$79,$9D,$3E,$76,$47,$46
	.dc.b	$CF,$48,$F2,$C6,$23,$A5,$23,$11,$FE,$27,$FA,$2F,$F5,$95,$7F,$83
	.dc.b	$FD,$34,$62,$7C,$F1,$8F,$86,$FF,$04,$36,$BF,$B8,$FA,$DF,$CE,$6B
	.dc.b	$E7,$35,$FF,$34,$35,$FF,$1A,$F8,$FF,$DF,$A3,$04,$70,$7F,$6F,$F6
	.dc.b	$65,$7F,$73,$6D,$7F,$89,$7E,$D9,$F6,$FF,$E0,$86,$D7,$F8,$11,$B3
	.dc.b	$B9,$FF,$06,$FD,$8B,$D1,$A4,$BF,$F0,$5C,$65,$BF,$0D,$C0,$8E,$0F
	.dc.b	$F0,$8F,$B7,$FE,$88,$6F,$C0,$FF,$86,$A3,$18,$FF,$8E,$DA,$FE,$09
	.dc.b	$AF,$FA,$1A,$38,$12,$7F,$21,$F5,$7F,$2D,$34,$9A,$FF,$8A,$21,$F1
	.dc.b	$54,$8F,$F8,$98,$C9,$5C,$DF,$E7,$5A,$FF,$8D,$FE,$DD,$F5,$9F,$E3
	.dc.b	$E6,$D7,$F9,$1D,$B3,$FE,$44,$6B,$F9,$EF,$B7,$FF,$7A,$4F,$F8,$D9
	.dc.b	$B1,$FE,$5D,$6B,$BF,$8C,$FF,$20,$C6,$23,$F7,$0D,$7F,$D6,$4D,$4E
	.dc.b	$65,$F6,$CF,$B7,$FD,$AB,$63,$F7,$4D,$AF,$D4,$35,$FD,$AF,$23,$83
	.dc.b	$FC,$A6,$DA,$FF,$27,$B6,$7F,$CE,$CD,$78,$CF,$02,$4F,$07,$E5,$7D
	.dc.b	$BC,$8F,$94,$D9,$E7,$DB,$E0,$24,$F8,$1B,$1A,$8C,$11,$FE,$1B,$E7
	.dc.b	$3E,$DD,$F2,$1F,$DD,$DB,$5F,$E8,$D9,$D4,$E4,$FF,$38,$F9,$2F,$76
	.dc.b	$92,$BF,$CD,$F1,$96,$C6,$70,$FE,$BA,$D7,$FD,$D5,$83,$E5,$FC,$44
	.dc.b	$AF,$14,$65,$FD,$8F,$D0,$7E,$D8,$8E,$0F,$DA,$FD,$BF,$EE,$88,$E0
	.dc.b	$97,$FE,$A9,$EB,$4E,$A3,$0F,$FA,$63,$ED,$FF,$19,$F0,$26,$93,$9B
	.dc.b	$44,$C3,$FD,$4D,$19,$7F,$D5,$1F,$6B,$FB,$46,$F6,$7F,$A9,$23,$18
	.dc.b	$FF,$AA,$1A,$FF,$A8,$DA,$86,$08,$CC,$61,$B1,$99,$FF,$53,$CF,$FB
	.dc.b	$EC,$FF,$46,$79,$BF,$D3,$BC,$D6,$1D,$7E,$0D,$3F,$C4,$EB,$67,$F6
	.dc.b	$5A,$DB,$83,$FB,$47,$83,$C6,$7A,$5D,$9F,$E3,$46,$CF,$F4,$7D,$F6
	.dc.b	$72,$D7,$DD,$7D,$A3,$11,$E9,$6C,$66,$BC,$D3,$FC,$B8,$FF,$5F,$8B
	.dc.b	$AD,$3F,$DD,$1F,$6F,$FC,$8B,$DB,$FF,$62,$B6,$33,$E0,$FF,$9B,$DB
	.dc.b	$3A,$FA,$97,$4D,$24,$F4,$8C,$B6,$3E,$C7,$DC,$7B,$70,$7D,$C7,$F8
	.dc.b	$3E,$30,$DC,$FF,$C2,$52,$C3,$9D,$71,$6C,$EA,$B8,$61,$F8,$5B,$EA
	.dc.b	$F8,$27,$9D,$70,$8F,$C3,$7C,$EF,$FB,$93,$87,$F1,$4D,$E2,$34,$D7
	.dc.b	$E9,$1F,$EE,$4A,$FC,$96,$9F,$82,$B4,$E6,$7E,$2F,$17,$E5,$E7,$3F
	.dc.b	$17,$FE,$2D,$BE,$18,$F6,$BE,$77,$BC,$39,$57,$56,$C6,$6F,$93,$D6
	.dc.b	$F9,$5F,$80,$7D,$D7,$E9,$FD,$0D,$C6,$03,$88,$D4,$78,$8F,$D6,$FD
	.dc.b	$D7,$F6,$A7,$11,$A7,$B3,$72,$FD,$1B,$70,$F8,$15,$C6,$BC,$99,$BA
	.dc.b	$7F,$97,$C7,$78,$C7,$83,$E6,$6C,$78,$C7,$DD,$7F,$6E,$78,$57,$27
	.dc.b	$AB,$72,$FA,$8D,$C7,$FC,$B9,$5C,$6B,$E3,$53,$EE,$9F,$59,$FE,$50
	.dc.b	$9F,$BB,$18,$7F,$A1,$35,$FE,$92,$D7,$FC,$D1,$8B,$F3,$37,$D3,$FA
	.dc.b	$DB,$3E,$F6,$BE,$2A,$E9,$87,$F0,$C8,$7D,$D3,$EA,$3F,$76,$D8,$F3
	.dc.b	$1A,$8C,$91,$CC,$6D,$F6,$35,$FF,$36,$B5,$FC,$B7,$E2,$0F,$9B,$FA
	.dc.b	$4D,$8F,$AC,$D7,$F1,$C7,$84,$FB,$5F,$72,$F9,$5F,$D6,$36,$3E,$66
	.dc.b	$A8,$C4,$7F,$9F,$A3,$13,$C0,$91,$C2,$76,$FC,$E6,$BF,$D0,$FD,$C7
	.dc.b	$E2,$7F,$ED,$58,$FC,$D4,$63,$E2,$7F,$42,$D8,$F3,$DF,$E9,$1F,$9F
	.dc.b	$F7,$7F,$38,$8F,$CA,$6B,$F2,$9A,$FC,$A7,$E9,$F4,$7E,$C3,$DB,$FD
	.dc.b	$62,$69,$DF,$64,$FE,$7F,$37,$C5,$FC,$82,$61,$F0,$5A,$FC,$78,$DB
	.dc.b	$92,$1F,$EA,$8C,$1F,$8A,$FF,$82,$E3,$F7,$2D,$7F,$72,$DE,$2F,$E3
	.dc.b	$55,$D2,$7E,$EF,$C1,$BE,$47,$FA,$5B,$63,$F2,$2D,$7F,$50,$D7,$FD
	.dc.b	$72,$D7,$FD,$98,$D7,$FB,$D3,$5F,$F5,$84,$7C,$AB,$7A,$FF,$C5,$5B
	.dc.b	$3E,$A8,$FB,$15,$E1,$C3,$F5,$1D,$5F,$E3,$3A,$FD,$2B,$5F,$DE,$35
	.dc.b	$FF,$66,$B5,$FE,$0B,$EA,$FF,$68,$8C,$47,$F8,$3D,$AF,$F1,$28,$FE
	.dc.b	$91,$F0,$CF,$87,$FE,$30,$D9,$FE,$38,$D7,$F4,$F5,$E3,$BF,$DF,$7D
	.dc.b	$D3,$E7,$7F,$24,$DA,$FE,$40,$D7,$F5,$FF,$76,$F8,$4F,$CC,$7E,$3C
	.dc.b	$F8,$EF,$EF,$ED,$AF,$E0,$AD,$7F,$8E,$35,$FF,$76,$F8,$BF,$D9,$9F
	.dc.b	$DD,$B7,$2F,$E2,$4D,$F1,$7F,$C6,$12,$FB,$C7,$D1,$FE,$A7,$CF,$FF
	.dc.b	$24,$CF,$F1,$46,$BF,$CA,$DA,$FF,$BF,$1A,$F4,$6B,$FE,$3D,$6B,$FE
	.dc.b	$01,$F6,$FF,$C3,$71,$AF,$09,$E1,$FF,$B5,$6B,$F8,$AF,$C4,$FF,$18
	.dc.b	$3C,$97,$FD,$77,$95,$EC,$D2,$57,$F4,$DF,$97,$7B,$78,$8F,$EB,$B5
	.dc.b	$C1,$FE,$C2,$D7,$F9,$0C,$F0,$73,$F0,$FD,$C8,$F5,$FA,$3A,$F4,$8B
	.dc.b	$C5,$FC,$FE,$95,$FC,$4F,$EE,$7F,$82,$DF,$F6,$16,$BF,$CC,$1B,$5F
	.dc.b	$D1,$FE,$ED,$F7,$7C,$7C,$69,$7C,$5F,$E3,$D3,$A8,$F5,$F6,$7C,$52
	.dc.b	$3B,$3D,$FE,$BF,$C9,$23,$FD,$CF,$EE,$DF,$75,$88,$F8,$F3,$72,$FE
	.dc.b	$B5,$5A,$9F,$93,$E8,$FD,$32,$3F,$DF,$7E,$19,$F1,$98,$8F,$A3,$36
	.dc.b	$FF,$B2,$56,$A7,$87,$D6,$7F,$97,$A0,$D2,$47,$F6,$68,$CB,$63,$2F
	.dc.b	$F7,$F8,$CB,$F9,$E1,$F8,$03,$F5,$57,$C7,$7F,$8A,$A3,$0D,$8C,$CF
	.dc.b	$90,$FC,$3F,$C1,$7E,$44,$8F,$F8,$6F,$EF,$7F,$BF,$F8,$5F,$E9,$BF
	.dc.b	$47,$F8,$55,$6A,$BF,$22,$FF,$55,$AD,$C7,$F8,$F3,$AB,$FC,$7F,$AF
	.dc.b	$F8,$C1,$B8,$73,$6C,$FF,$A9,$FE,$F7,$E9,$3F,$9C,$F8,$4F,$D2,$11
	.dc.b	$FF,$17,$46,$AF,$FC,$25,$F7,$9F,$C6,$67,$E2,$35,$FE,$06,$D7,$FD
	.dc.b	$68,$D4,$D2,$5F,$FA,$51,$E1,$3F,$E0,$3E,$0F,$F2,$AB,$6B,$FC,$A4
	.dc.b	$D7,$FC,$1F,$3F,$CA,$E3,$E0,$7C,$67,$EF,$49,$3F,$E4,$79,$7D,$E6
	.dc.b	$23,$F7,$73,$6B,$DB,$3F,$C1,$7E,$06,$57,$EF,$9F,$F0,$21,$FD,$97
	.dc.b	$4F,$98,$EA,$3F,$82,$BA,$F9,$1F,$F0,$57,$DE,$7F,$8A,$E7,$FD,$AC
	.dc.b	$D4,$D3,$1E,$97,$D6,$8F,$F2,$EE,$FF,$E1,$38,$C4,$FF,$8E,$72,$FF
	.dc.b	$47,$7E,$F9,$FE,$45,$6E,$5F,$E6,$78,$C3,$FE,$5E,$BC,$33,$0F,$F3
	.dc.b	$74,$E9,$FE,$B9,$3F,$E6,$FF,$88,$FF,$41,$C7,$F8,$D9,$AF,$F8,$D5
	.dc.b	$AF,$F9,$79,$AF,$E3,$FD,$6F,$E5,$C4,$7F,$CF,$91,$CB,$0F,$F5,$37
	.dc.b	$DD,$FF,$A0,$67,$FC,$86,$D7,$FC,$84,$D7,$FC,$D6,$D7,$F4,$DF,$9D
	.dc.b	$FF,$01,$7F,$D2,$2D,$9F,$F4,$8B,$72,$FF,$58,$57,$59,$CF,$6F,$FA
	.dc.b	$7E,$B5,$7F,$EA,$66,$F9,$CB,$E1,$87,$82,$F8,$12,$FF,$D3,$13,$A5
	.dc.b	$4F,$BC,$FF,$4A,$47,$FA,$11,$AF,$FB,$0D,$AF,$06,$F5,$7F,$82,$FE
	.dc.b	$27,$FA,$C1,$27,$FD,$4D,$3F,$E9,$B7,$BF,$DE,$7F,$80,$DF,$F3,$6B
	.dc.b	$5F,$F3,$63,$5F,$F4,$AB,$5F,$F6,$A3,$5E,$7E,$C7,$FB,$8E,$31,$81
	.dc.b	$D6,$1F,$E1,$FE,$4F,$ED,$84,$BF,$F5,$D4,$F8,$76,$6A,$9F,$EE,$3F
	.dc.b	$76,$FA,$8F,$F4,$CF,$A2,$8C,$4F,$FA,$C6,$31,$3A,$9C,$CF,$F3,$59
	.dc.b	$CB,$B8,$CB,$FE,$16,$8C,$BB,$8C,$BE,$3A,$62,$FA,$EF,$F7,$2C,$61
	.dc.b	$B1,$99,$FF,$72,$5F,$4C,$7D,$B1,$EE,$9D,$3F,$EF,$1B,$E4,$FF,$BA
	.dc.b	$DA,$FF,$BD,$6B,$94,$7F,$8B,$3E,$E7,$FD,$0D,$3F,$EE,$3F,$B8,$7D
	.dc.b	$DF,$F8,$C3,$E6,$BC,$5C,$09,$CD,$7C,$D9,$F0,$DF,$28,$FF,$5A,$B7
	.dc.b	$3F,$3D,$F1,$C3,$A6,$9F,$56,$31,$7C,$E3,$C9,$F1,$5F,$EF,$3C,$BB
	.dc.b	$CB,$2F,$1C,$61,$B8,$74,$9F,$C0,$65,$E1,$CF,$DD,$F7,$5F,$DE,$6F
	.dc.b	$CE,$DC,$B3,$CF,$E1,$C1,$F1,$AF,$BB,$FE,$EB,$87,$0A,$D5,$32,$63
	.dc.b	$CE,$3C,$2D,$C3,$97,$8B,$E5,$9E,$13,$FC,$2D,$F0,$9F,$F0,$06,$B9
	.dc.b	$DF,$FB,$B9,$B1,$9F,$8E,$FF,$21,$3F,$65,$AF,$C0,$6B,$E5,$6C,$7F
	.dc.b	$B9,$78,$BD,$2D,$9F,$99,$A7,$FB,$6F,$EA,$3F,$C1,$FB,$6B,$0F,$F2
	.dc.b	$96,$BF,$79,$B5,$E1,$C7,$E4,$6A,$C3,$8E,$3D,$0F,$24,$FC,$DE,$2F
	.dc.b	$C5,$7C,$97,$E6,$B2,$FB,$6D,$C7,$FC,$FF,$EA,$FF,$9F,$B4,$FC,$A4
	.dc.b	$E7,$1F,$DE,$4E,$57,$2F,$E3,$B9,$15,$B9,$FF,$4F,$7C,$17,$EA,$2B
	.dc.b	$53,$9B,$E5,$18,$C3,$EE,$C6,$1F,$CB,$61,$D1,$FC,$31,$FE,$D5,$FA
	.dc.b	$BE,$8F,$EF,$7E,$39,$18,$CB,$F9,$AF,$DD,$FF,$9B,$AB,$F6,$95,$A6
	.dc.b	$D6,$DE,$87,$F1,$9C,$7F,$53,$5C,$6F,$F7,$D7,$C5,$F8,$35,$FA,$B9
	.dc.b	$D4,$7F,$AA,$36,$46,$23,$5E,$F7,$F0,$A7,$51,$97,$CF,$19,$77,$5F
	.dc.b	$E5,$57,$2D,$AE,$B3,$B9,$FF,$23,$7B,$9F,$07,$C0,$FF,$6B,$3F,$0E
	.dc.b	$9F,$06,$F6,$F5,$E9,$2B,$53,$9C,$BB,$37,$1F,$F3,$36,$1C,$B0,$F0
	.dc.b	$36,$B7,$18,$AE,$15,$F0,$DA,$FF,$11,$6B,$FE,$5D,$FB,$BF,$19,$E7
	.dc.b	$F1,$7F,$D6,$B7,$FE,$B2,$8C,$7C,$1F,$82,$73,$5E,$46,$BF,$45,$AF
	.dc.b	$F8,$2B,$E3,$7E,$27,$96,$7E,$27,$DD,$7F,$B1,$B1,$FC,$01,$B2,$D7
	.dc.b	$F8,$1E,$78,$9C,$3A,$EC,$EB,$81,$87,$58,$C3,$6F,$F0,$E7,$F8,$23
	.dc.b	$0F,$03,$FE,$1C,$C3,$A5,$7D,$D9,$CC,$FE,$2D,$AF,$F3,$96,$BF,$E8
	.dc.b	$5F,$99,$FB,$67,$DF,$FB,$A7,$81,$8B,$F5,$F0,$A7,$31,$F8,$B6,$DF
	.dc.b	$DE,$6B,$F9,$B6,$BF,$D2,$7D,$2F,$E1,$FD,$1F,$E5,$6F,$77,$F6,$12
	.dc.b	$FB,$8F,$ED,$D8,$7F,$94,$A7,$F2,$57,$C2,$39,$47,$09,$DD,$7C,$6C
	.dc.b	$57,$B8,$FF,$06,$4E,$B0,$FD,$53,$70,$FF,$08,$4B,$EE,$B0,$F0,$F0
	.dc.b	$7C,$B1,$DA,$3F,$D8,$0D,$78,$B5,$FF,$0D,$37,$85,$D2,$F9,$9F,$D4
	.dc.b	$9F,$C3,$FE,$E5,$F4,$8F,$B9,$CE,$7F,$85,$BE,$0D,$BE,$15,$B8,$C1
	.dc.b	$BF,$17,$F6,$A6,$BC,$1B,$19,$26,$9F,$8C,$CB,$C3,$3A,$8F,$C7,$FD
	.dc.b	$53,$E9,$3F,$8D,$70,$FF,$86,$A3,$1E,$17,$F2,$78,$7C,$73,$EE,$7F
	.dc.b	$95,$56,$B3,$FF,$5A,$CE,$5F,$E2,$95,$C2,$BF,$89,$36,$33,$3B,$C7
	.dc.b	$FD,$DD,$39,$6C,$65,$FB,$31,$96,$C6,$5F,$F6,$F7,$CD,$FF,$18,$3E
	.dc.b	$4F,$CF,$5B,$F9,$D1,$67,$E2,$6E,$AC,$61,$FE,$E9,$FB,$9F,$E4,$27
	.dc.b	$D9,$AF,$AA,$7E,$DB,$93,$EF,$C6,$B1,$FB,$4E,$F8,$B1,$5F,$BC,$F8
	.dc.b	$B7,$EA,$CF,$1B,$FA,$94,$61,$F9,$7F,$73,$FE,$63,$C7,$F8,$6C,$E5
	.dc.b	$FE,$1F,$C2,$9D,$61,$F9,$C1,$19,$F7,$3F,$A0,$36,$BF,$A3,$36,$78
	.dc.b	$35,$FD,$2B,$5F,$F3,$B3,$5E,$B5,$E0,$7F,$57,$A7,$E8,$3E,$CF,$FA
	.dc.b	$39,$F1,$FE,$36,$E1,$F9,$D8,$B5,$F9,$31,$38,$DF,$FC,$D3,$7F,$AA
	.dc.b	$F9,$8F,$EA,$27,$D1,$F2,$6E,$3F,$A7,$8B,$87,$39,$B7,$CB,$EE,$7F
	.dc.b	$D1,$F7,$FC,$C5,$B5,$FE,$6F,$FB,$9F,$EE,$D5,$BA,$FF,$02,$5A,$73
	.dc.b	$1F,$E9,$78,$C7,$A5,$FC,$6B,$BF,$F8,$84,$62,$7F,$D0,$75,$C3,$04
	.dc.b	$E5,$FE,$7A,$7F,$06,$F1,$BB,$CE,$F4,$67,$F9,$3C,$3D,$1C,$FF,$E1
	.dc.b	$1F,$B9,$FE,$8D,$AF,$B6,$3F,$BD,$56,$F2,$FC,$A3,$67,$FD,$27,$6B
	.dc.b	$E2,$FF,$AC,$70,$EF,$87,$2F,$98,$93,$FE,$95,$A7,$D1,$E3,$2E,$B5
	.dc.b	$71,$EB,$5A,$BF,$9D,$3F,$E0,$BF,$AC,$FF,$04,$47,$F4,$C6,$DF,$F5
	.dc.b	$EF,$71,$F7,$2F,$8B,$E1,$EC,$7F,$8D,$67,$31,$FE,$19,$6E,$1F,$E1
	.dc.b	$F6,$DF,$E8,$5A,$FF,$3E,$6B,$FE,$CF,$6B,$FE,$8A,$BF,$04,$F3,$6C
	.dc.b	$7F,$8A,$FF,$0B,$7C,$FF,$F5,$86,$D7,$F8,$02,$7F,$AD,$7C,$16,$6B
	.dc.b	$FA,$EF,$DD,$3E,$53,$FC,$61,$48,$C5,$FF,$8D,$E3,$13,$A9,$CC,$7F
	.dc.b	$8E,$D7,$DD,$74,$3F,$B4,$FD,$AF,$F8,$DA,$FC,$53,$A7,$FB,$CF,$C8
	.dc.b	$7F,$98,$EF,$FB,$33,$78,$FF,$C9,$96,$FB,$7F,$F2,$04,$E6,$FF,$B2
	.dc.b	$CE,$5F,$33,$6F,$FC,$9C,$D7,$F8,$83,$5E,$35,$F0,$23,$11,$FE,$FF
	.dc.b	$6A,$73,$CF,$FC,$C6,$7F,$CA,$9C,$FD,$F1,$FC,$CF,$CC,$3D,$FC,$D7
	.dc.b	$F9,$CE,$31,$97,$F9,$84,$38,$CB,$F9,$24,$62,$BF,$CC,$4D,$38,$4E
	.dc.b	$E7,$FC,$0B,$19,$78,$C6,$50,$8D,$3F,$C8,$BE,$33,$F9,$9E,$1F,$C6
	.dc.b	$9B,$3F,$EE,$BF,$BA,$78,$9F,$E7,$19,$CD,$BD,$B5,$6A,$FF,$CF,$4D
	.dc.b	$9F,$F3,$6B,$5F,$F4,$53,$5E,$D7,$FE,$70,$8C,$47,$56,$D7,$69,$FF
	.dc.b	$42,$6B,$FE,$54,$CF,$FC,$C2,$7F,$A7,$FE,$E5,$F4,$B8,$CB,$FC,$BE
	.dc.b	$23,$2F,$D1,$6E,$5F,$41,$AF,$CC,$6B,$F2,$A3,$94,$FD,$49,$CC,$FC
	.dc.b	$D6,$CE,$8F,$E0,$1F,$84,$3E,$FB,$F8,$33,$63,$FC,$B3,$A3,$D3,$CD
	.dc.b	$7F,$A7,$63,$13,$F3,$27,$E3,$11,$C1,$30,$F9,$F1,$97,$FB,$76,$6F
	.dc.b	$8D,$7D,$1F,$FA,$AD,$B7,$FE,$84,$E1,$7D,$DB,$E1,$75,$7F,$E2,$B8
	.dc.b	$C5,$7F,$89,$9B,$8F,$F8,$89,$AF,$F8,$81,$AF,$FB,$45,$AF,$FB,$B9
	.dc.b	$B7,$E0,$BF,$F1,$7C,$6B,$89,$F9,$B7,$FA,$92,$BF,$35,$7F,$91,$8C
	.dc.b	$61,$D7,$E5,$FF,$D8,$91,$89,$ED,$1D,$09,$3E,$97,$FD,$7F,$FA,$53
	.dc.b	$E6,$3F,$D9,$8D,$9F,$E5,$1D,$7F,$CA,$AF,$FC,$F5,$39,$6D,$70,$3F
	.dc.b	$8F,$F8,$16,$FF,$D5,$5E,$A7,$DF,$AF,$F1,$93,$73,$F5,$E8,$FB,$D7
	.dc.b	$D7,$7F,$B7,$EB,$58,$7F,$61,$6E,$0E,$3F,$F1,$4F,$37,$FB,$8F,$0E
	.dc.b	$37,$E0,$8F,$F7,$37,$99,$FE,$0A,$CF,$BB,$67,$F8,$2E,$6F,$BA,$FE
	.dc.b	$27,$1F,$EA,$C9,$D6,$5F,$BF,$6D,$7F,$75,$9F,$F7,$95,$70,$AF,$F4
	.dc.b	$F5,$6A,$31,$1F,$E9,$AA,$ED,$DF,$FE,$EE,$9C,$CF,$89,$AF,$F9,$17
	.dc.b	$9F,$F8,$CF,$DE,$7F,$8C,$23,$FD,$47,$E9,$FE,$0A,$31,$1F,$EA,$69
	.dc.b	$DF,$AF,$FE,$4D,$AD,$3F,$E4,$F9,$FD,$10,$AE,$2F,$FB,$61,$AE,$9C
	.dc.b	$CF,$28,$F2,$1F,$D8,$CE,$67,$F6,$5A,$C3,$83,$FD,$73,$F0,$6F,$BD
	.dc.b	$FF,$55,$C7,$8E,$31,$F0,$3F,$DB,$5B,$19,$9D,$BF,$E6,$78,$7D,$83
	.dc.b	$83,$83,$67,$A3,$5E,$0D,$7C,$6E,$F2,$F9,$34,$FB,$DF,$F6,$7C,$61
	.dc.b	$BA,$70,$C3,$E6,$E1,$FD,$32,$FC,$71,$88,$FF,$5A,$57,$07,$4D,$7F
	.dc.b	$D7,$18,$78,$6F,$F2,$3F,$14,$EC,$FF,$07,$70,$FF,$B3,$DB,$19,$BF
	.dc.b	$B5,$39,$6F,$17,$FA,$8D,$FC,$C7,$95,$FE,$B0,$E4,$FC,$EC,$E6,$7F
	.dc.b	$DE,$57,$CD,$FC,$BD,$6D,$FC,$F3,$6F,$FC,$A4,$D9,$E1,$A7,$FA,$02
	.dc.b	$BF,$BF,$1F,$DB,$9C,$78,$FC,$FA,$3F,$B8,$B7,$8B,$FD,$8F,$9F,$F9
	.dc.b	$6E,$F8,$9F,$EB,$52,$DF,$27,$A4,$7F,$98,$67,$8C,$7F,$01,$F4,$9E
	.dc.b	$B7,$76,$C7,$03,$51,$9F,$90,$D3,$23,$59,$2B,$8C,$7F,$B1,$D5,$7A
	.dc.b	$FE,$FB,$FD,$85,$3F,$AB,$A4,$62,$3F,$A7,$B5,$FF,$57,$B5,$FF,$3D
	.dc.b	$63,$FE,$F3,$7C,$2D,$7F,$DA,$CD,$CB,$FD,$5B,$3D,$5F,$F5,$7F,$3F
	.dc.b	$F6,$7A,$F1,$35,$FE,$BE,$D7,$FD,$7E,$D7,$FD,$14,$D7,$F9,$53,$6F
	.dc.b	$FB,$37,$8F,$FE,$5F,$6C,$FB,$FD,$0F,$E6,$AF,$C8,$6B,$FE,$ED,$6D
	.dc.b	$F9,$AB,$F0,$5C,$3F,$E1,$4A,$F9,$0D,$7F,$BA,$B5,$FF,$6A,$B5,$FF
	.dc.b	$50,$35,$FE,$80,$DB,$FF,$19,$79,$9E,$78,$C4,$7F,$A0,$A7,$EE,$B5
	.dc.b	$FE,$11,$9F,$FA,$AE,$31,$87,$FB,$6A,$BC,$8F,$FB,$73,$E5,$5E,$8F
	.dc.b	$C3,$6E,$1F,$81,$6B,$FE,$0E,$6B,$FE,$EE,$6B,$FE,$B9,$6B,$FD,$65
	.dc.b	$B6,$6F,$E3,$AF,$B1,$FD,$53,$6B,$F2,$ED,$7D,$67,$3B,$FD,$87,$6F
	.dc.b	$F2,$F8,$FC,$B3,$5F,$F1,$1B,$5F,$F3,$29,$FD,$AD,$C3,$FE,$CF,$6B
	.dc.b	$9E,$EF,$C2,$46,$31,$FE,$95,$1F,$87,$FB,$97,$B7,$FC,$05,$B7,$FA
	.dc.b	$E6,$BE,$8A,$E2,$FE,$CB,$D4,$FE,$64,$FE,$B5,$AF,$F8,$E9,$AF,$F9
	.dc.b	$E7,$C0,$FC,$B4,$62,$3F,$DC,$CD,$7F,$C0,$CD,$93,$7F,$1D,$7C,$F7
	.dc.b	$F1,$76,$D7,$F0,$46,$BE,$79,$F1,$F0,$7F,$8C,$AD,$F8,$6F,$F0,$8A
	.dc.b	$F9,$71,$8E,$BF,$D5,$4E,$67,$FD,$F4,$D7,$FC,$34,$D9,$E1,$1F,$C3
	.dc.b	$63,$04,$F9,$1F,$E5,$89,$A6,$2B,$F9,$87,$33,$E6,$AF,$DC,$CC,$3F
	.dc.b	$8F,$CE,$5F,$E4,$3F,$87,$7E,$48,$9C,$9F,$C0,$5F,$CA,$FD,$1F,$F4
	.dc.b	$9D,$7E,$03,$7A,$BF,$D5,$BB,$23,$18,$7F,$8D,$5F,$E5,$75,$FE,$58
	.dc.b	$87,$C9,$FE,$7D,$FF,$76,$35,$FF,$0B,$35,$FC,$A9,$C3,$AB,$FC,$21
	.dc.b	$19,$8F,$CC,$B5,$FF,$5F,$B5,$E0,$7F,$37,$F8,$A7,$99,$FD,$55,$B1
	.dc.b	$FD,$09,$AF,$EB,$0E,$27,$EB,$BB,$5F,$AF,$3E,$0F,$FB,$33,$E9,$CB
	.dc.b	$FC,$B6,$DD,$7F,$C9,$AE,$EF,$FA,$A5,$F2,$BE,$39,$C9,$CD,$6D,$FD
	.dc.b	$37,$C4,$FF,$52,$7F,$AB,$B5,$CF,$8B,$FD,$BE,$1F,$4D,$62,$4F,$07
	.dc.b	$51,$FD,$FE,$1F,$96,$3E,$6B,$1F,$71,$D7,$6F,$B0,$DD,$BF,$83,$D6
	.dc.b	$EF,$FB,$71,$CA,$F8,$E3,$23,$33,$B7,$E6,$5F,$37,$FC,$F5,$93,$E3
	.dc.b	$FD,$47,$18,$F8,$3E,$37,$B4,$8F,$C2,$F6,$EB,$A3,$E2,$C7,$C3,$8F
	.dc.b	$89,$F7,$1F,$17,$E4,$65,$7C,$A8,$CB,$F3,$23,$2F,$D0,$8F,$8D,$F6
	.dc.b	$EF,$A6,$FF,$30,$36,$7F,$CD,$81,$C3,$FE,$6F,$6B,$F9,$A6,$BF,$CC
	.dc.b	$1A,$8E,$51,$D8,$FE,$E5,$93,$EB,$3E,$F1,$27,$E8,$D7,$83,$87,$F8
	.dc.b	$A5,$7F,$05,$9F,$C3,$F8,$5F,$85,$3F,$0D,$F0,$36,$32,$43,$F1,$EF
	.dc.b	$E2,$FE,$AA,$47,$E4,$1F,$F2,$27,$E2,$6F,$95,$CC,$FE,$5A,$6E,$3F
	.dc.b	$EB,$CA,$D4,$FE,$1B,$ED,$DF,$23,$FE,$89,$6C,$7C,$A6,$BF,$E9,$76
	.dc.b	$BF,$C0,$9A,$FF,$58,$6B,$F2,$5A,$9C,$F8,$4D,$1F,$8D,$FE,$9C,$95
	.dc.b	$E5,$BF,$D5,$7E,$13,$23,$F5,$EF,$EB,$7E,$23,$11,$FB,$48,$B9,$FD
	.dc.b	$09,$FD,$97,$BB,$2B,$F6,$EF,$ED,$7E,$8B,$11,$FB,$C8,$B5,$F6,$27
	.dc.b	$F7,$5F,$29,$F8,$02,$57,$F9,$3F,$34,$7E,$13,$EB,$3F,$80,$12,$7F
	.dc.b	$82,$C7,$F9,$C7,$F0,$8F,$E1,$58,$7F,$57,$6F,$83,$FC,$E1,$B5,$FB
	.dc.b	$2B,$E1,$E2,$C8,$FF,$32,$CE,$B8,$3E,$55,$75,$F1,$7F,$C8,$D3,$FC
	.dc.b	$C6,$4D,$27,$8F,$E3,$25,$F4,$AF,$37,$D4,$72,$24,$7F,$9B,$69,$1D
	.dc.b	$BE,$83,$D8,$49,$F2,$C5,$7F,$C1,$92,$9C,$69,$F0,$6A,$E9,$F0,$66
	.dc.b	$DB,$2F,$4D,$7C,$9E,$3F,$69,$27,$C8,$FC,$0E,$8F,$9C,$48,$F8,$F1
	.dc.b	$F4,$7D,$7F,$8A,$48,$F6,$35,$FE,$BB,$7F,$22,$31,$1F,$E0,$AC,$FF
	.dc.b	$C3,$07,$FB,$81,$FA,$7F,$11,$F7,$4E,$CF,$F4,$CF,$AE,$FF,$55,$BF
	.dc.b	$E3,$CF,$31,$18,$8F,$E1,$D1,$8F,$85,$FF,$67,$CE,$65,$39,$7F,$D9
	.dc.b	$F1,$98,$C4,$7F,$74,$6B,$E7,$F7,$7F,$20,$47,$C9,$55,$FA,$B1,$78
	.dc.b	$1C,$1E,$4C,$3F,$D8,$2D,$F9,$AF,$F2,$AB,$EB,$84,$62,$F4,$7F,$29
	.dc.b	$9D,$46,$63,$D7,$3C,$1D,$1C,$0F,$F0,$A6,$3F,$1A,$78,$36,$36,$72
	.dc.b	$8F,$F3,$3F,$C1,$FE,$52,$3F,$11,$A3,$E3,$B3,$C5,$FA,$D6,$CF,$13
	.dc.b	$F9,$F5,$71,$5F,$61,$FD,$E3,$A3,$37,$FE,$74,$8C,$4A,$B8,$46,$1F
	.dc.b	$17,$C3,$FF,$97,$3E,$C7,$BD,$74,$F4,$FF,$C2,$15,$C1,$B8,$75,$6C
	.dc.b	$FF,$B8,$7D,$F5,$F2,$C7,$C3,$3F,$E3,$26,$BC,$72,$FE,$C3,$18,$8F
	.dc.b	$EB,$4F,$FB,$A3,$D6,$CF,$A9,$FE,$2C,$9F,$63,$5D,$B5,$F8,$6F,$F8
	.dc.b	$D3,$3F,$7B,$73,$FE,$CD,$A7,$F9,$3F,$D2,$FC,$4D,$FC,$59,$EB,$87
	.dc.b	$CA,$8E,$AD,$7A,$B9,$AE,$5C,$79,$CB,$D1,$5C,$CF,$D1,$FE,$BD,$FE
	.dc.b	$B3,$9E,$9D,$5F,$C7,$BD,$EF,$0E,$BF,$83,$E5,$5E,$F0,$E5,$7B,$C3
	.dc.b	$C9,$5A,$F5,$3D,$31,$FB,$7B,$E0,$D8,$CD,$FE,$E2,$72,$D8,$CB,$E6
	.dc.b	$6D,$75,$AF,$CC,$B6,$7F,$CE,$57,$FC,$DB,$DE,$F9,$3D,$3F,$D1,$5B
	.dc.b	$7D,$F8,$2D,$FE,$7A,$6D,$F1,$AF,$E8,$B1,$AC,$3E,$0D,$EB,$9D,$39
	.dc.b	$CF,$FC,$7D,$7F,$E7,$CC,$7E,$6D,$FF,$A1,$5F,$04,$67,$0E,$6D,$AE
	.dc.b	$33,$FC,$45,$AF,$C9,$AE,$4F,$49,$FE,$95,$E6,$78,$CF,$27,$E1,$96
	.dc.b	$3F,$AE,$BC,$33,$FE,$B2,$DC,$CD,$24,$FF,$55,$6D,$FF,$3A,$9D,$CF
	.dc.b	$F5,$9F,$C2,$FE,$11,$F0,$FE,$B2,$C7,$F6,$C7,$B1,$1F,$EC,$0F,$87
	.dc.b	$B3,$E5,$1F,$2F,$A6,$CF,$F7,$48,$B8,$7F,$60,$9B,$C1,$FD,$FE,$6D
	.dc.b	$7F,$7F,$9B,$5F,$C6,$1F,$71,$23,$FB,$83,$F0,$7E,$8F,$ED,$1F,$6F
	.dc.b	$C0,$B3,$FE,$04,$7E,$31,$1F,$EF,$0F,$CB,$FD,$77,$F9,$7E,$3F,$EE
	.dc.b	$76,$CF,$FB,$21,$FF,$59,$BF,$E6,$18,$D6,$1F,$8E,$9D,$BF,$DD,$5B
	.dc.b	$B1,$E3,$09,$E3,$5C,$CE,$DD,$3C,$CE,$9F,$07,$63,$55,$D6,$3D,$04
	.dc.b	$8F,$CB,$61,$DE,$3D,$9F,$0F,$E5,$3C,$DE,$CD,$AE,$75,$E5,$9C,$5F
	.dc.b	$96,$7E,$05,$63,$1F,$89,$F3,$DE,$F3,$E0,$7C,$A5,$AE,$7A,$27,$18
	.dc.b	$73,$3E,$25,$62,$FE,$77,$DD,$FC,$93,$F7,$3F,$5E,$F9,$37,$C0,$E7
	.dc.b	$3F,$83,$39,$8F,$D4,$B6,$3F,$6C,$DC,$7F,$DE,$6D,$7F,$D3,$01,$C3
	.dc.b	$FE,$9B,$6B,$9A,$F0,$CF,$46,$C7,$F8,$8B,$37,$CC,$FE,$04,$FE,$19
	.dc.b	$FB,$7B,$E2,$F1,$87,$C4,$9B,$7F,$EF,$5A,$D4,$E4,$F9,$3F,$7F,$F8
	.dc.b	$B3,$F1,$9F,$7B,$63,$24,$AF,$99,$C7,$F8,$D9,$D4,$63,$2E,$AD,$CB
	.dc.b	$E6,$AF,$A3,$FC,$C4,$61,$F9,$71,$9C,$3E,$5B,$4F,$96,$53,$E5,$EB
	.dc.b	$FE,$C3,$9E,$24,$9E,$2F,$EB,$7E,$6F,$F4,$DF,$29,$FA,$76,$CF,$E2
	.dc.b	$1B,$7E,$96,$F1,$7E,$CF,$EB,$FF,$5C,$7E,$BF,$0B,$7F,$90,$8B,$7F
	.dc.b	$06,$6D,$7F,$70,$E9,$7D,$63,$E1,$FF,$71,$F6,$3F,$D2,$27,$31,$8C
	.dc.b	$BF,$5F,$8B,$DD,$7C,$46,$F1,$E5,$5A,$9C,$BF,$B6,$C3,$9B,$9A,$E1
	.dc.b	$87,$F5,$6C,$3A,$46,$27,$F5,$3F,$1F,$FC,$2C,$FE,$19,$F3,$56,$32
	.dc.b	$4A,$FB,$D1,$97,$F0,$1F,$65,$FC,$4F,$EA,$FF,$8A,$36,$BF,$6A,$D9
	.dc.b	$DC,$7E,$E2,$7F,$B3,$FC,$2B,$E5,$E9,$FD,$B2,$5F,$41,$FE,$EB,$7F
	.dc.b	$84,$CE,$DF,$E2,$A6,$BE,$0F,$FA,$EB,$4F,$ED,$45,$C3,$F6,$67,$F9
	.dc.b	$B2,$BA,$38,$8D,$3F,$E6,$C8,$CB,$67,$AC,$62,$07,$FA,$97,$F1,$AF
	.dc.b	$F2,$2D,$7F,$1E,$6C,$7E,$A1,$79,$3F,$E0,$67,$3C,$9F,$E3,$6E,$D7
	.dc.b	$CC,$7F,$A0,$30,$FE,$5A,$DC,$3F,$D4,$6D,$7F,$86,$CF,$09,$FE,$ED
	.dc.b	$F3,$2F,$1B,$FC,$26,$BC,$6B,$18,$7F,$C2,$FE,$53,$E1,$69,$FE,$1A
	.dc.b	$F9,$A7,$CB,$63,$1F,$E7,$31,$87,$FA,$0B,$73,$FF,$5B,$B5,$F2,$B5
	.dc.b	$F8,$F9,$A3,$11,$FE,$CD,$6C,$66,$BF,$C2,$11,$83,$E1,$6B,$F1,$DB
	.dc.b	$7D,$3E,$6F,$FA,$FC,$E6,$3F,$C6,$71,$FD,$8B,$0F,$8E,$D7,$87,$5F
	.dc.b	$F8,$76,$31,$1F,$ED,$F6,$E2,$69,$27,$FC,$49,$F0,$7F,$EF,$2B,$FE
	.dc.b	$BA,$DB,$FE,$4F,$F3,$8F,$27,$FC,$8B,$39,$CB,$F0,$13,$9B,$FE,$D6
	.dc.b	$DB,$FF,$70,$B5,$FB,$ED,$7E,$5B,$5F,$F7,$B3,$72,$34,$C7,$CE,$FE
	.dc.b	$82,$FF,$20,$DB,$FA,$BF,$3D,$FA,$DC,$3F,$9F,$CE,$56,$09,$D4,$7E
	.dc.b	$49,$AF,$2E,$AF,$F2,$44,$62,$38,$B6,$79,$4E,$89,$7F,$E0,$CF,$8E
	.dc.b	$7A,$5F,$E6,$C6,$C7,$F8,$2B,$DE,$7D,$86,$6B,$FC,$E3,$18,$9F,$F0
	.dc.b	$AB,$72,$FE,$03,$8F,$FB,$EA,$FF,$CB,$93,$99,$FD,$5B,$5F,$CB,$B5
	.dc.b	$F8,$2D,$9E,$55,$FE,$22,$E0,$7B,$5F,$E8,$26,$C7,$F8,$83,$E8,$3F
	.dc.b	$4D,$BF,$EF,$E7,$38,$FF,$6C,$9F,$D0,$93,$97,$FC,$E0,$DB,$FF,$17
	.dc.b	$B6,$7D,$6F,$EA,$3E,$0D,$E9,$7F,$A4,$9B,$5F,$E3,$03,$C2,$7E,$C9
	.dc.b	$CD,$7F,$7C,$8F,$D8,$92,$FF,$D0,$0F,$ED,$3E,$C9,$EE,$7F,$A7,$9B
	.dc.b	$1F,$E3,$C7,$F8,$DB,$FB,$E7,$35,$FE,$08,$6E,$7F,$C7,$DB,$3B,$F7
	.dc.b	$3A,$5F,$FA,$AB,$E9,$DF,$21,$4C,$7C,$85,$BF,$F2,$5C,$E9,$B3,$A6
	.dc.b	$9F,$E5,$AE,$2F,$E9,$7F,$2B,$FE,$B3,$F1,$CF,$FA,$D7,$E5,$9E,$AE
	.dc.b	$62,$5F,$1A,$DB,$CE,$B6,$E6,$79,$57,$FA,$BA,$D5,$AD,$FF,$A8,$C6
	.dc.b	$27,$53,$99,$FF,$60,$0E,$1F,$E9,$BD,$9F,$CE,$E3,$FC,$97,$CB,$FE
	.dc.b	$FC,$8C,$47,$F9,$3A,$77,$E2,$FF,$5E,$AD,$3F,$EC,$B8,$CC,$F2,$CE
	.dc.b	$BC,$45,$F4,$7F,$D1,$2D,$7C,$87,$98,$A7,$A2,$70,$E0,$FF,$30,$FD
	.dc.b	$2B,$83,$FC,$DD,$1F,$ED,$D8,$C7,$BD,$F0,$1B,$19,$9D,$BF,$DE,$23
	.dc.b	$F0,$7C,$BB,$E1,$FF,$4E,$4F,$36,$BB,$6B,$E2,$77,$8F,$C8,$AF,$EC
	.dc.b	$13,$F0,$9A,$FC,$2C,$F8,$5E,$6F,$8C,$7B,$3C,$5F,$F3,$EE,$1C,$DB
	.dc.b	$1C,$2F,$90,$E0,$FC,$1F,$B1,$F7,$9B,$1F,$EF,$BE,$04,$E5,$B1,$FE
	.dc.b	$FF,$E3,$7D,$2F,$FA,$33,$47,$0F,$F8,$4B,$1F,$F1,$3C,$E6,$31,$5C
	.dc.b	$2F,$F3,$18,$E5,$F2,$46,$5F,$F5,$45,$F2,$7F,$41,$7B,$8F,$3C,$7F
	.dc.b	$A0,$87,$C1,$F3,$EE,$FF,$11,$B6,$32,$7F,$B0,$38,$BF,$C6,$39,$FF
	.dc.b	$87,$E3,$27,$39,$E0,$73,$9D,$E7,$FE,$08,$F8,$EF,$F2,$35,$7F,$B2
	.dc.b	$9A,$FF,$A5,$1B,$7F,$E2,$B5,$F2,$1E,$57,$35,$CF,$C4,$C1,$99,$F5
	.dc.b	$7D,$2F,$F0,$2C,$3F,$DB,$AD,$7F,$A4,$B5,$FF,$18,$B5,$FF,$0E,$D7
	.dc.b	$F0,$C7,$F9,$87,$57,$C5,$8C,$47,$FA,$AA,$7A,$63,$FE,$41,$E8,$7A
	.dc.b	$DF,$C0,$B6,$FE,$04,$D8,$CD,$FF,$BC,$1A,$FF,$5C,$6B,$FE,$44,$6B
	.dc.b	$FC,$FE,$39,$1D,$7C,$27,$CD,$7D,$96,$CF,$F5,$78,$FB,$95,$DB,$0F
	.dc.b	$F2,$B7,$0B,$CE,$F3,$36,$FF,$B7,$C6,$08,$E0,$87,$F7,$28,$E0,$FF
	.dc.b	$74,$CB,$FD,$E2,$FF,$9B,$2B,$FC,$C9,$18,$8F,$F6,$8B,$63,$C7,$5F
	.dc.b	$91,$F2,$3F,$DC,$55,$FE,$09,$F8,$1F,$95,$18,$8F,$F0,$0B,$5F,$F3
	.dc.b	$4B,$5F,$F2,$AB,$5F,$F0,$FB,$6C,$DF,$29,$F1,$FF,$A1,$6D,$7A,$78
	.dc.b	$3F,$C5,$73,$E5,$C3,$F4,$DD,$BF,$C8,$EB,$FD,$26,$7F,$8B,$1C,$3F
	.dc.b	$E1,$66,$BF,$E7,$E6,$BF,$E6,$A6,$BC,$7C,$0F,$C4,$46,$32,$FE,$77
	.dc.b	$1F,$8B,$F3,$1F,$03,$FB,$76,$DF,$9C,$FF,$24,$C7,$9A,$7F,$79,$E7
	.dc.b	$7F,$24,$9F,$F4,$7B,$5F,$F3,$DB,$5F,$F2,$2B,$6C,$D2,$78,$3F,$9D
	.dc.b	$75,$E8,$3E,$23,$F8,$53,$6F,$C9,$7F,$E7,$19,$F1,$DF,$F0,$FF,$4B
	.dc.b	$F8,$0D,$7F,$A3,$BB,$3F,$51,$18,$8F,$F4,$7B,$5F,$F4,$74,$68,$8F
	.dc.b	$FA,$69,$E3,$EF,$75,$24,$7F,$A8,$23,$F8,$F7,$23,$E4,$2F,$C8,$4A
	.dc.b	$F3,$CE,$5F,$4F,$AF,$F8,$92,$71,$FE,$F5,$FC,$67,$B1,$FC,$C7,$A7
	.dc.b	$FD,$84,$DE,$7F,$F5,$3B,$6F,$FC,$E2,$FF,$1A,$7F,$DA,$3F,$4B,$F9
	.dc.b	$47,$FC,$12,$D7,$F1,$AD,$7F,$DC,$2D,$7F,$DB,$21,$C3,$D3,$AB,$FB
	.dc.b	$F4,$62,$35,$1F,$C5,$7E,$09,$E7,$7F,$41,$6C,$FD,$86,$AC,$7F,$AB
	.dc.b	$BD,$4F,$F1,$A7,$D2,$BE,$57,$FA,$73,$E9,$9D,$F7,$A3,$1A,$7F,$5F
	.dc.b	$78,$5F,$2B,$E3,$94,$8C,$1F,$8A,$F8,$0F,$E8,$6D,$F1,$FF,$AC,$46
	.dc.b	$23,$CD,$87,$FB,$5A,$1F,$2D,$AA,$95,$C4,$E9,$1F,$DA,$D7,$D2,$BE
	.dc.b	$17,$1F,$4B,$D7,$6F,$83,$48,$C6,$5F,$4A,$F8,$4F,$EC,$CE,$77,$CB
	.dc.b	$19,$19,$9D,$BE,$80,$75,$8F,$6F,$CA,$C8,$E9,$87,$81,$FF,$5D,$E7
	.dc.b	$FC,$61,$E5,$DC,$8C,$67,$F6,$23,$1E,$2F,$C8,$F9,$0F,$92,$CB,$F9
	.dc.b	$CE,$5F,$A1,$F4,$BF,$3B,$DE,$91,$82,$1E,$0C,$51,$A8,$C4,$FC,$DF
	.dc.b	$A4,$FF,$61,$C7,$F9,$99,$AF,$F2,$A6,$BF,$9C,$0E,$1F,$CF,$B5,$F8
	.dc.b	$ED,$77,$F0,$7F,$E5,$D8,$D6,$5F,$0F,$E9,$7F,$0E,$4A,$FA,$D3,$E2
	.dc.b	$D3,$E3,$64,$F0,$BF,$1A,$4A,$FC,$97,$C6,$57,$E0,$45,$AE,$33,$F9
	.dc.b	$4F,$A5,$95,$F9,$97,$FC,$65,$F3,$2F,$8C,$CC,$FC,$99,$B7,$FA,$8A
	.dc.b	$D4,$FE,$2F,$E9,$5F,$41,$FE,$79,$6C,$7F,$A4,$1A,$FF,$54,$6B,$F2
	.dc.b	$DA,$FF,$08,$6B,$F9,$06,$A7,$3B,$7E,$2F,$E9,$7F,$5E,$49,$F2,$57
	.dc.b	$EC,$BE,$96,$47,$ED,$DF,$DA,$FA,$D8,$8F,$C3,$45,$CB,$E7,$CF,$EE
	.dc.b	$BE,$06,$57,$EF,$DF,$DE,$FB,$78,$8F,$C7,$45,$AF,$AF,$3F,$C0,$BE
	.dc.b	$97,$FC,$61,$C4,$69,$2F,$FB,$BE,$9F,$A5,$FA,$5F,$E1,$24,$8F,$E1
	.dc.b	$D3,$FE,$63,$F9,$EF,$E2,$B1,$FA,$C6,$E3,$FA,$B6,$F8,$3F,$C3,$EF
	.dc.b	$87,$89,$2F,$FB,$0D,$69,$FD,$BC,$F5,$F1,$3F,$C8,$D3,$FC,$E3,$23
	.dc.b	$49,$E3,$F8,$C9,$7D,$2B,$CD,$F4,$5C,$89,$1E,$EA,$47,$6F,$A5,$F6
	.dc.b	$12,$7C,$B1,$5F,$74,$5D,$3D,$D3,$74,$FE,$2B,$97,$9A,$B1,$68,$F4
	.dc.b	$D7,$C9,$E3,$F6,$92,$7C,$8F,$C0,$E8,$F9,$C4,$8F,$8F,$1F,$47,$D4
	.dc.b	$F8,$A4,$8F,$63,$5F,$E0,$D7,$F2,$23,$11,$FD,$92,$30,$FC,$3F,$88
	.dc.b	$FB,$84,$7E,$AC,$FF,$A9,$7E,$97,$FD,$37,$1F,$DB,$3C,$C4,$62,$3F
	.dc.b	$D6,$CD,$7F,$D8,$8D,$8D,$FC,$2F,$FB,$16,$30,$47,$EF,$FB,$7F,$8D
	.dc.b	$25,$7F,$8E,$6B,$8B,$F8,$88,$CB,$F9,$17,$CD,$3E,$7F,$8D,$FF,$00
	.dc.b	$EF,$E7,$C1,$18,$BF,$3B,$5F,$3B,$70,$F6,$B6,$78,$CF,$F5,$2A,$E9
	.dc.b	$39,$24,$7F,$56,$F8,$1F,$C6,$BF,$81,$E2,$7D,$2F,$F8,$52,$73,$B7
	.dc.b	$E7,$23,$1C,$3E,$4B,$E4,$FF,$86,$BE,$97,$FB,$BF,$46,$27,$FB,$1C
	.dc.b	$61,$61,$E0,$51,$98,$F1,$FC,$3F,$F9,$63,$E4,$3B,$D7,$2F,$4F,$FC
	.dc.b	$1F,$5C,$6F,$86,$3E,$19,$CB,$FD,$CB,$0E,$57,$E3,$3B,$CF,$F0,$47
	.dc.b	$86,$3F,$E8,$27,$2D,$BF,$44,$7F,$5B,$7F,$DB,$7E,$AE,$7D,$2F,$F1
	.dc.b	$6C,$FC,$16,$BF,$01,$AF,$B9,$FF,$1A,$E5,$E9,$6E,$5E,$7F,$5D,$E9
	.dc.b	$FF,$8D,$AB,$58,$7A,$71,$F0,$E3,$F2,$B0,$F0,$4F,$CF,$D1,$8F,$2E
	.dc.b	$2E,$19,$F6,$C1,$C8,$FA,$5F,$F5,$4D,$7D,$7E,$2F,$E3,$3E,$EF,$87
	.dc.b	$4F,$F0,$DD,$EF,$D3,$C6,$3F,$87,$AB,$39,$C3,$F2,$13,$C3,$1E,$AF
	.dc.b	$F8,$C2,$74,$D8,$CC,$FE,$DD,$B1,$C6,$FF,$CA,$F3,$96,$C6,$63,$BB
	.dc.b	$F9,$9F,$77,$FC,$BB,$E0,$7F,$3D,$6C,$F8,$F8,$AD,$6E,$3F,$9E,$CE
	.dc.b	$34,$FE,$72,$EB,$2F,$8D,$79,$E7,$4E,$73,$F9,$D8,$FF,$87,$EF,$78
	.dc.b	$FF,$B3,$1B,$19,$9D,$BE,$FC,$3F,$CA,$6F,$96,$B8,$4E,$A3,$07,$F2
	.dc.b	$6C,$1E,$67,$8C,$F2,$7D,$2D,$BF,$EA,$EF,$4A,$FD,$E3,$73,$FE,$AA
	.dc.b	$DB,$34,$95,$FD,$45,$B5,$E1,$F9,$5F,$84,$7C,$3F,$A5,$B5,$FD,$99
	.dc.b	$E8,$47,$FA,$D3,$DF,$B3,$E5,$1F,$2F,$AE,$CF,$F6,$E8,$B7,$FD,$DE
	.dc.b	$6F,$07,$F7,$79,$B5,$FD,$6B,$25,$62,$FF,$8D,$3E,$D2,$4F,$F6,$B7
	.dc.b	$E0,$7C,$DF,$DB,$3E,$E7,$7D,$9F,$EF,$EF,$C5,$23,$FD,$CD,$F9,$5F
	.dc.b	$45,$F9,$EC,$BF,$CE,$8D,$9F,$ED,$AF,$FA,$D1,$FF,$74,$B9,$C3,$FB
	.dc.b	$73,$76,$34,$E1,$1F,$E8,$27,$1D,$5F,$95,$73,$7E,$0F,$82,$C4,$E5
	.dc.b	$F2,$93,$0F,$F1,$3D,$23,$D3,$F0,$BC,$CE,$9E,$CD,$AF,$3C,$5C,$7C
	.dc.b	$53,$71,$F7,$FD,$2F,$8C,$F2,$7C,$A5,$C3,$DF,$16,$FD,$53,$6B,$E3
	.dc.b	$FC,$E7,$AC,$FD,$9F,$D1,$BE,$4D,$F0,$39,$CB,$FD,$57,$7F,$ED,$8A
	.dc.b	$D4,$FE,$75,$B3,$FE,$8F,$C5,$18,$7F,$D2,$4D,$7F,$DD,$ED,$7F,$D6
	.dc.b	$ED,$78,$DF,$8A,$7A,$4E,$78,$3E,$BF,$D2,$FD,$33,$F8,$37,$D2,$BE
	.dc.b	$2B,$17,$F8,$09,$B8,$FF,$B1,$7B,$D7,$BA,$D1,$F8,$6F,$9F,$FB,$C7
	.dc.b	$DF,$F9,$EB,$19,$25,$FE,$16,$32,$FE,$1F,$1F,$C3,$1F,$31,$7D,$2E
	.dc.b	$2B,$E5,$9F,$02,$32,$FB,$9A,$7C,$82,$9F,$20,$A7,$18,$FA,$4E,$0F
	.dc.b	$A2,$5D,$3F,$DC,$71,$FA,$8F,$9B,$FC,$D7,$C9,$FE,$6D,$B7,$FA,$66
	.dc.b	$DF,$E8,$5B,$C3,$FA,$EF,$9D,$FD,$11,$FA,$3C,$2D,$FE,$BA,$2D,$FE
	.dc.b	$A6,$6D,$2B,$FD,$EB,$D4,$FA,$4F,$F6,$34,$7E,$DB,$E9,$7F,$A2,$CE
	.dc.b	$7E,$23,$F6,$11,$FD,$23,$17,$BA,$F8,$8D,$E3,$CA,$59,$77,$47,$09
	.dc.b	$DE,$1C,$EF,$96,$5F,$52,$99,$78,$63,$11,$C5,$D4,$66,$3F,$83,$7C
	.dc.b	$7F,$F0,$53,$F8,$37,$CD,$DA,$FE,$23,$16,$7F,$86,$4D,$CB,$F8,$CF
	.dc.b	$D2,$FF,$0D,$FA,$5F,$E1,$CD,$C3,$F8,$CB,$6F,$F8,$A3,$6B,$FD,$DF
	.dc.b	$F2,$DF,$DA,$7E,$18,$FF,$6E,$B7,$D0,$7F,$B9,$DF,$77,$6A,$B7,$1F
	.dc.b	$04,$DF,$C1,$FF,$59,$69,$FB,$D2,$E5,$F3,$0F,$F2,$F5,$F5,$70,$7D
	.dc.b	$F7,$0F,$F9,$B9,$AF,$09,$ED,$18,$8F,$F5,$17,$D2,$FF,$4C,$AF,$3B
	.dc.b	$63,$EE,$C7,$FA,$AA,$1E,$57,$F8,$25,$D6,$7F,$81,$CB,$FC,$67,$5F
	.dc.b	$DC,$7D,$9F,$C9,$3F,$ED,$A6,$E1,$FE,$A2,$6B,$FE,$A1,$F9,$87,$8D
	.dc.b	$FE,$15,$5E,$35,$8C,$3F,$E1,$9F,$29,$F0,$74,$FF,$0E,$7C,$CB,$E5
	.dc.b	$71,$9F,$D2,$8C,$3F,$EE,$C6,$E5,$FE,$B4,$6B,$FE,$B3,$6B,$FE,$CA
	.dc.b	$6B,$FE,$C9,$6E,$86,$92,$7C,$67,$F8,$BF,$E6,$BF,$AD,$BC,$CC,$74
	.dc.b	$FF,$85,$E3,$FB,$1D,$FF,$B0,$DA,$FF,$B6,$9A,$FF,$B6,$5B,$66,$92
	.dc.b	$7F,$C4,$5F,$37,$FE,$EC,$BF,$C9,$B6,$FE,$E9,$FE,$E3,$F9,$C7,$91
	.dc.b	$FE,$45,$9D,$71,$7D,$D9,$CD,$FF,$51,$C9,$18,$C3,$FD,$BA,$D7,$FD
	.dc.b	$B8,$D7,$FD,$E6,$D7,$C1,$87,$F8,$DE,$31,$3D,$23,$FC,$75,$F3,$BF
	.dc.b	$9F,$C3,$F1,$CD,$C3,$FD,$FB,$F3,$DF,$AC,$C7,$D7,$82,$B5,$1F,$EF
	.dc.b	$F9,$D3,$FE,$EF,$6B,$CB,$AB,$FC,$91,$18,$8E,$11,$82,$5F,$C1,$9D
	.dc.b	$5B,$E3,$9E,$97,$F9,$B5,$B5,$FC,$0B,$CE,$7D,$2E,$6B,$FC,$E5,$18
	.dc.b	$9F,$98,$DC,$BF,$7B,$87,$FA,$6A,$FF,$CB,$93,$99,$FE,$09,$39,$AE
	.dc.b	$35,$A2,$3F,$5F,$81,$ED,$FF,$A0,$5B,$1F,$C5,$BE,$83,$F4,$7B,$FE
	.dc.b	$EE,$73,$8F,$CC,$9F,$CF,$93,$97,$FC,$DC,$DB,$FC,$23,$67,$D2,$FE
	.dc.b	$9B,$E2,$1E,$8F,$FA,$45,$B5,$FC,$98,$F0,$9F,$D2,$B4,$FF,$74,$CF
	.dc.b	$FA,$5E,$3F,$5E,$4C,$3F,$CF,$CF,$EC,$BE,$95,$EE,$FF,$A7,$5B,$1F
	.dc.b	$CC,$9F,$E2,$EF,$57,$39,$7E,$21,$BA,$7F,$1B,$6C,$EF,$DC,$E9,$7F
	.dc.b	$EA,$9F,$A5,$7C,$85,$31,$F2,$16,$FF,$A0,$CE,$9B,$3A,$69,$FD,$8F
	.dc.b	$8F,$FA,$4F,$CA,$FF,$AC,$FC,$73,$FE,$B5,$F9,$67,$AB,$98,$97,$C6
	.dc.b	$B6,$F3,$AD,$B9,$9E,$55,$FE,$A7,$B5,$6B,$7F,$F5,$E3,$5F,$EA,$11
	.dc.b	$A9,$FF,$57,$47,$F5,$3E,$3F,$EF,$FD,$9F,$CE,$23,$FA,$B7,$37,$F8
	.dc.b	$0E,$31,$1B,$3F,$69,$E2,$FF,$58,$9D,$3F,$E5,$BA,$E6,$E7,$3B,$5F
	.dc.b	$D7,$A6,$BF,$E1,$C8,$AF,$F8,$42,$3F,$CC,$8F,$72,$B8,$74,$7F,$9A
	.dc.b	$7E,$95,$C1,$FE,$73,$8C,$35,$FE,$DB,$1A,$F7,$7A,$B6,$39,$70,$7F
	.dc.b	$88,$9B,$CB,$A8,$C4,$72,$6B,$FD,$EE,$7F,$13,$3E,$C7,$FA,$DC,$5F
	.dc.b	$17,$FC,$01,$AF,$CA,$9D,$CF,$CA,$3F,$C5,$EE,$23,$53,$C3,$4E,$CD
	.dc.b	$AD,$E1,$E7,$AF,$B1,$BB,$F5,$67,$C6,$FC,$A8,$C3,$63,$37,$FE,$FF
	.dc.b	$6B,$FE,$12,$97,$D9,$3E,$0B,$FC,$31,$AF,$F9,$16,$73,$18,$E1,$FF
	.dc.b	$23,$C6,$2F,$F4,$58,$74,$BE,$18,$73,$9F,$D3,$35,$FF,$47,$36,$3B
	.dc.b	$4F,$1B,$FE,$15,$BF,$F9,$3A,$FF,$C6,$6D,$7F,$A3,$B6,$BF,$C3,$4F
	.dc.b	$FB,$26,$37,$C5,$E7,$5D,$BF,$EB,$98,$C3,$63,$AF,$95,$8F,$90,$7C
	.dc.b	$9D,$24,$FF,$91,$21,$F1,$B8,$8F,$EB,$31,$C1,$D4,$F3,$9E,$FF,$4D
	.dc.b	$FE,$9E,$9F,$EE,$4D,$7F,$A3,$B5,$FE,$8A,$D7,$F9,$66,$8F,$9A,$F4
	.dc.b	$B6,$3F,$99,$47,$6F,$80,$EF,$ED,$7E,$B3,$46,$08,$C3,$6F,$C7,$18
	.dc.b	$23,$82,$1F,$D7,$3E,$0F,$FD,$1A,$FF,$99,$9A,$FF,$7A,$6F,$07,$F9
	.dc.b	$36,$BB,$4F,$F7,$EE,$37,$C4,$79,$5B,$1F,$DA,$A3,$04,$70,$43,$FB
	.dc.b	$6C,$70,$7F,$B7,$E2,$FA,$2F,$BA,$D8,$FF,$3D,$B5,$F1,$57,$E0,$AB
	.dc.b	$F8,$FC,$7F,$74,$6B,$FD,$6D,$AF,$E2,$9B,$26,$93,$AF,$D8,$7D,$18
	.dc.b	$7D,$07,$90,$95,$FE,$0D,$9F,$C9,$F5,$3E,$63,$EC,$36,$7F,$C7,$0D
	.dc.b	$7F,$C1,$AD,$7F,$C1,$6D,$7F,$B9,$35,$FD,$0B,$6F,$FC,$5D,$3F,$85
	.dc.b	$F4,$FF,$AA,$D7,$FA,$81,$AF,$F8,$D1,$BA,$FF,$8E,$A7,$CF,$F2,$3E
	.dc.b	$62,$57,$99,$AF,$99,$AF,$99,$F0,$7B,$FF,$B2,$B3,$4E,$B3,$FE,$81
	.dc.b	$F9,$6F,$CD,$7B,$26,$92,$3F,$D0,$8D,$7F,$D0,$8D,$7F,$D0,$9F,$0B
	.dc.b	$FC,$02,$7F,$DC,$91,$FE,$8D,$8C,$4F,$F0,$7E,$A7,$E5,$8F,$81,$FE
	.dc.b	$0F,$DA,$D2,$9B,$9D,$63,$FE,$97,$EC,$F3,$91,$FF,$4C,$3F,$B3,$F7
	.dc.b	$FD,$51,$FE,$CE,$6B,$FE,$62,$6F,$07,$FA,$D7,$57,$CF,$FF,$1B,$A7
	.dc.b	$D3,$BE,$53,$FB,$33,$63,$FB,$5B,$5F,$F6,$79,$C2,$36,$47,$6D,$7F
	.dc.b	$CB,$3F,$11,$FC,$AD,$FF,$3A,$B7,$0F,$F6,$43,$7C,$1F,$F6,$F5,$F7
	.dc.b	$9F,$F6,$67,$C3,$7A,$A7,$FC,$88,$DB,$FE,$66,$DE,$7F,$F7,$77,$87
	.dc.b	$FC,$D2,$30,$4E,$4C,$47,$F3,$BE,$C7,$D3,$FF,$0E,$C7,$FA,$9B,$FD
	.dc.b	$27,$1F,$DA,$36,$77,$A7,$18,$E1,$3A,$F2,$B0,$66,$7E,$77,$D3,$FF
	.dc.b	$98,$24,$7F,$AC,$07,$D3,$FD,$47,$3B,$32,$FF,$80,$13,$99,$E1,$5C
	.dc.b	$DB,$ED,$FD,$B9,$D5,$F1,$9E,$F1,$8B,$F0,$C2,$F9,$56,$E3,$C7,$C5
	.dc.b	$FD,$73,$E5,$7F,$D2,$F1,$E4,$BF,$F3,$A4,$62,$3D,$58,$75,$7F,$B2
	.dc.b	$FC,$1F,$3C,$3C,$5C,$48,$7D,$47,$F6,$86,$D7,$53,$8E,$39,$35,$32
	.dc.b	$35,$5F,$EB,$AB,$EB,$7D,$67,$2D,$7F,$D8,$6D,$8D,$F1,$78,$62,$4F
	.dc.b	$07,$C9,$1F,$A5,$FA,$77,$CA,$7F,$92,$5B,$5F,$E4,$E6,$BF,$A4,$6B
	.dc.b	$F3,$83,$87,$E8,$B5,$FC,$1B,$5D,$C7,$F6,$FE,$0F,$F3,$4E,$1F,$A6
	.dc.b	$F1,$7F,$C1,$F1,$8D,$39,$D7,$CB,$E5,$F9,$9D,$73,$1F,$87,$19,$7E
	.dc.b	$2C,$65,$F8,$B1,$FA,$3F,$1B,$F6,$51,$F6,$BE,$77,$EB,$19,$F0,$24
	.dc.b	$7C,$E8,$CB,$FE,$99,$9E,$2B,$E9,$FF,$D4,$AF,$F9,$D1,$AF,$F0,$A6
	.dc.b	$BF,$1D,$AF,$E5,$9A,$FE,$A1,$AF,$5F,$57,$FC,$F9,$1A,$E2,$FC,$77
	.dc.b	$CC,$FE,$34,$95,$F8,$D3,$81,$FE,$AF,$FD,$E3,$F0,$9B,$7F,$B3,$F9
	.dc.b	$3F,$BE,$71,$7E,$33,$CD,$FE,$0D,$5A,$BF,$EE,$CD,$C7,$F3,$93,$F9
	.dc.b	$B3,$77,$F9,$DF,$84,$FC,$F7,$D4,$CC,$3F,$2B,$19,$7F,$2F,$19,$7F
	.dc.b	$2F,$1F,$A5,$F9,$5F,$F1,$85,$1A,$49,$FE,$67,$87,$EB,$7C,$BF,$D6
	.dc.b	$12,$3F,$47,$3F,$AF,$E0,$91,$FA,$67,$4F,$E9,$9C,$B4,$FD,$5C,$FE
	.dc.b	$DF,$EA,$7C,$64,$8F,$0E,$08,$F3,$F7,$7E,$F4,$93,$FB,$48,$FE,$01
	.dc.b	$F5,$3E,$C2,$47,$27,$DF,$F0,$5F,$C1,$49,$1F,$BC,$8A,$FE,$F2,$2F
	.dc.b	$07,$F0,$29,$BC,$3F,$6E,$6D,$7F,$14,$9F,$E2,$1F,$07,$F3,$09,$1E
	.dc.b	$A7,$E7,$FB,$3F,$C5,$49,$1F,$C3,$23,$F8,$C7,$D4,$FD,$42,$47,$C2
	.dc.b	$6B,$FC,$07,$83,$F9,$04,$62,$3F,$98,$C6,$1F,$B3,$FB,$13,$E9,$FF
	.dc.b	$D1,$6D,$8F,$E3,$ED,$7F,$8D,$D7,$F5,$9F,$A4,$72,$FF,$AD,$DB,$3F
	.dc.b	$DD,$BC,$B4,$62,$3F,$D6,$CD,$7F,$DF,$AD,$7F,$D6,$6D,$7F,$CA,$8D
	.dc.b	$8E,$73,$FE,$83,$B7,$C8,$FA,$09,$7F,$CE,$EB,$6F,$AA,$32,$B8,$3F
	.dc.b	$A2,$D7,$28,$E1,$87,$9E,$7F,$D9,$0B,$9F,$FA,$CF,$C3,$FF,$B4,$B1
	.dc.b	$F6,$C6,$3C,$CF,$4B,$70,$F4,$F5,$5A,$DC,$7F,$72,$BE,$58,$75,$8C
	.dc.b	$61,$C9,$A7,$22,$D7,$25,$5C,$B0,$ED,$7C,$E7,$37,$F3,$30,$F2,$43
	.dc.b	$E9,$FF,$D8,$BB,$F6,$8C,$78,$5F,$81,$CB,$FD,$DB,$5A,$9E,$63,$D7
	.dc.b	$7C,$AB,$E3,$BA,$D6,$EF,$8E,$1C,$F3,$FF,$00,$CE,$5F,$F0,$9E,$1C
	.dc.b	$AF,$8D,$7E,$3F,$1D,$B5,$FF,$0C,$DF,$8D,$FE,$BC,$D7,$AF,$97,$FE
	.dc.b	$ED,$6F,$C1,$FF,$92,$1F,$81,$19,$79,$B5,$CC,$7C,$39,$FF,$3B,$3F
	.dc.b	$4B,$07,$AF,$FE,$E8,$EC,$D6,$1F,$EF,$DD,$3E,$3B,$74,$F0,$E5,$F2
	.dc.b	$30,$E3,$E4,$FF,$89,$6F,$79,$7D,$CF,$DF,$4E,$70,$FF,$1B,$F2,$27
	.dc.b	$39,$EF,$2D,$1F,$EB,$98,$CB,$F5,$A7,$85,$E9,$B5,$C4,$F0,$46,$6B
	.dc.b	$8E,$1F,$9D,$6E,$1E,$27,$FC,$93,$5C,$0F,$CA,$F5,$39,$DE,$67,$F9
	.dc.b	$07,$A3,$F5,$AD,$AF,$5F,$C3,$3E,$6F,$1B,$31,$E5,$78,$6D,$FF,$12
	.dc.b	$8D,$EC,$C3,$94,$E4,$FE,$00,$E2,$FF,$C4,$53,$BC,$3F,$C9,$D1,$E2
	.dc.b	$6B,$FE,$E1,$9E,$0F,$F9,$5A,$78,$57,$F1,$B9,$CC,$F1,$8C,$57,$2C
	.dc.b	$3E,$16,$9E,$9C,$CD,$FA,$BC,$36,$FF,$98,$C7,$19,$34,$91,$A2,$61
	.dc.b	$FC,$F6,$77,$1F,$CE,$BE,$AF,$C0,$78,$7D,$FB,$1F,$CE,$9E,$C4,$7F
	.dc.b	$A0,$BE,$1F,$8C,$F4,$1E,$9F,$3E,$CF,$F4,$28,$B8,$7F,$29,$8F,$F4
	.dc.b	$5E,$7F,$81,$AC,$F0,$7F,$2C,$BB,$5A,$C9,$58,$C3,$F0,$11,$89,$FE
	.dc.b	$A7,$3F,$0C,$8F,$F5,$6F,$8E,$F9,$87,$CD,$F8,$4B,$1F,$D5,$9F,$94
	.dc.b	$47,$FA,$EB,$F3,$7E,$AB,$EA,$1F,$57,$EA,$AC,$FF,$5E,$8B,$8F,$B4
	.dc.b	$F2,$CE,$33,$FE,$8E,$F1,$AF,$EC,$BF,$09,$6B,$86,$BF,$A3,$8F,$B8
	.dc.b	$49,$FE,$DE,$FD,$FF,$AA,$FC,$37,$CC,$7E,$1D,$B3,$FD,$C1,$B7,$D3
	.dc.b	$64,$63,$2F,$F5,$77,$33,$D4,$79,$7F,$91,$F8,$4F,$D3,$C6,$32,$FE
	.dc.b	$F7,$18,$9D,$47,$F9,$4D,$B9,$77,$F9,$6F,$CD,$1F,$B6,$F8,$D7,$C8
	.dc.b	$3C,$AC,$E1,$FE,$A9,$9C,$C7,$EB,$1B,$1F,$EF,$86,$DF,$FA,$55,$AF
	.dc.b	$FB,$11,$AF,$FB,$59,$A9,$A6,$33,$7A,$EF,$A4,$FD,$87,$95,$FC,$1A
	.dc.b	$31,$87,$F8,$6A,$31,$3A,$8F,$F4,$43,$70,$F9,$DD,$FF,$B9,$FA,$9F
	.dc.b	$DD,$36,$BF,$C4,$6D,$9D,$C7,$F8,$9A,$7F,$8C,$78,$4F,$8C,$A7,$C9
	.dc.b	$AF,$19,$F8,$68,$D5,$7C,$98,$C9,$F0,$CA,$7C,$32,$9F,$0C,$BA,$7F
	.dc.b	$BF,$31,$FA,$11,$82,$3F,$C5,$7F,$51,$FE,$73,$1F,$96,$6F,$27,$FB
	.dc.b	$99,$FE,$92,$D8,$FE,$65,$1A,$BF,$DF,$B6,$3F,$9C,$3C,$23,$31,$C3
	.dc.b	$E0,$38,$FC,$85,$9F,$F7,$4D,$23,$15,$E3,$6D,$F8,$DF,$E8,$BE,$4B
	.dc.b	$E9,$BE,$FE,$1C,$BE,$7B,$31,$8C,$3D,$91,$FD,$31,$FE,$A7,$EC,$EB
	.dc.b	$8F,$1B,$A3,$18,$7F,$AD,$63,$0A,$BF,$81,$63,$E1,$BE,$37,$C2,$7E
	.dc.b	$57,$CB,$FC,$6F,$86,$F8,$ED,$C3,$DE,$D9,$DC,$7C,$29,$FE,$BD,$F1
	.dc.b	$4F,$A6,$A6,$3E,$5A,$D7,$8F,$80,$FF,$30,$AE,$0E,$4C,$57,$17,$18
	.dc.b	$7F,$6D,$9D,$BF,$C5,$BE,$41,$F1,$9F,$DB,$DB,$5C,$07,$CD,$BC,$9D
	.dc.b	$57,$F7,$89,$CC,$7F,$6A,$69,$F2,$6E,$38,$70,$7A,$AB,$A4,$62,$31
	.dc.b	$3C,$E7,$31,$C2,$79,$3D,$23,$2F,$FA,$73,$E9,$5E,$0F,$F8,$25,$B5
	.dc.b	$F0,$BC,$97,$4F,$F7,$48,$C4,$EA,$FF,$9C,$FC,$A3,$E3,$7F,$BE,$36
	.dc.b	$BF,$DB,$ED,$9F,$F5,$73,$5F,$81,$EC,$BF,$33,$FF,$6A,$57,$87,$E9
	.dc.b	$9C,$DF,$E2,$0A,$D7,$C4,$FF,$87,$E7,$35,$E7,$D1,$EA,$FF,$88,$DB
	.dc.b	$1E,$BC,$FF,$99,$3B,$9F,$F1,$44,$63,$3F,$2F,$91,$EA,$25,$7A,$9B
	.dc.b	$1A,$8C,$1F,$EF,$CE,$3F,$E9,$11,$FD,$71,$BC,$FF,$92,$8F,$F1,$C7
	.dc.b	$CC,$7A,$1D,$79,$7F,$E3,$98,$C4,$7F,$B4,$1A,$FF,$82,$1A,$F2,$DF
	.dc.b	$FC,$55,$18,$E3,$34,$91,$FE,$2F,$F8,$2F,$F2,$7E,$5F,$75,$B7,$FD
	.dc.b	$CA,$9F,$57,$FE,$4B,$9C,$E5,$E9,$C7,$AC,$F8,$09,$87,$E4,$E3,$2D
	.dc.b	$BE,$6D,$9D,$E1,$FE,$EF,$9C,$BF,$EF,$1F,$C0,$3F,$CA,$58,$77,$6D
	.dc.b	$FF,$8A,$3C,$DF,$D9,$CE,$7C,$CF,$F3,$04,$7E,$DA,$FF,$DE,$ED,$78
	.dc.b	$77,$7F,$93,$E3,$11,$FE,$69,$6E,$66,$92,$7F,$CA,$DF,$11,$FC,$C7
	.dc.b	$1F,$D2,$36,$FF,$C3,$34,$FA,$9F,$F3,$94,$E7,$2F,$91,$39,$BF,$8C
	.dc.b	$DB,$FD,$EE,$3E,$1C,$3C,$C4,$AD,$B6,$77,$5C,$5F,$E1,$15,$FE,$DC
	.dc.b	$F1,$FF,$D0,$5F,$02,$7F,$D0,$98,$7F,$AC,$6B,$FC,$5F,$39,$8F,$DF
	.dc.b	$19,$8E,$6D,$7F,$CA,$ED,$7F,$D2,$0D,$AF,$E5,$BF,$05,$FE,$7B,$8C
	.dc.b	$11,$FF,$3F,$53,$E9,$FC,$A7,$93,$9A,$D7,$89,$E4,$47,$F1,$87,$ED
	.dc.b	$70,$F7,$FD,$23,$D9,$FF,$50,$36,$7F,$8C,$2F,$5B,$33,$27,$2F,$08
	.dc.b	$C5,$7F,$94,$BE,$93,$F8,$3B,$7B,$3F,$D5,$90,$F8,$0B,$13,$0F,$E4
	.dc.b	$D3,$C3,$8B,$F4,$CE,$E3,$33,$FE,$B2,$FA,$27,$D3,$53,$FD,$6B,$F0
	.dc.b	$CF,$84,$C5,$FF,$52,$E5,$5F,$08,$C3,$74,$FF,$38,$4F,$FA,$F6,$B8
	.dc.b	$C6,$2F,$A5,$FF,$AC,$E7,$FD,$73,$5B,$8C,$1F,$EF,$FA,$FE,$F5,$C5
	.dc.b	$FD,$EF,$E8,$5E,$87,$F5,$2F,$4F,$F8,$D4,$62,$BF,$AA,$D6,$EB,$57
	.dc.b	$67,$22,$B8,$3F,$E0,$7B,$E7,$5F,$D6,$EF,$84,$E6,$EB,$E4,$3C,$D1
	.dc.b	$E7,$4D,$3F,$B2,$2F,$A4,$7C,$2F,$F9,$CA,$3F,$B4,$46,$3D,$FD,$B6
	.dc.b	$33,$3B,$7F,$B8,$C7,$FA,$CB,$8B,$7F,$0C,$69,$27,$AC,$65,$F8,$19
	.dc.b	$7F,$7D,$AF,$C5,$47,$B9,$AE,$8F,$F1,$53,$DD,$C1,$EB,$9E,$9A,$7C
	.dc.b	$A9,$E2,$DE,$4F,$F0,$53,$99,$DD,$FF,$8D,$76,$70,$7F,$BB,$B1,$FF
	.dc.b	$7B,$7C,$47,$C8,$9C,$B6,$75,$3F,$E1,$29,$FC,$9F,$8D,$FE,$17,$D3
	.dc.b	$FD,$57,$C7,$FD,$DA,$BF,$2B,$39,$E1,$F0,$61,$CA,$F7,$39,$AF,$E6
	.dc.b	$17,$C5,$FC,$DB,$67,$FD,$6F,$5C,$6F,$FD,$D3,$C0,$E2,$E5,$DB,$FE
	.dc.b	$AE,$6C,$67,$1F,$F0,$D3,$FE,$1A,$6C,$68,$8E,$BC,$AE,$7E,$6B,$E3
	.dc.b	$69,$27,$FC,$49,$0F,$03,$11,$FE,$DA,$8E,$0E,$A7,$9C,$F9,$BE,$8D
	.dc.b	$F4,$9F,$96,$CD,$18,$BF,$ED,$4D,$7F,$A2,$B5,$FE,$6B,$ED,$7C,$0D
	.dc.b	$7F,$CE,$91,$FC,$EA,$3B,$5F,$7F,$6B,$B5,$9A,$4A,$EE,$D7,$BB,$5F
	.dc.b	$E9,$BF,$02,$F7,$7E,$7B,$63,$FC,$8C,$D7,$FD,$F7,$5D,$DF,$F2,$57
	.dc.b	$2F,$FA,$65,$FE,$C3,$EF,$FC,$18,$C4,$7F,$5C,$8C,$11,$D9,$0F,$EC
	.dc.b	$FA,$7F,$BA,$DF,$F2,$B3,$5F,$DA,$69,$FE,$93,$81,$9E,$4F,$EE,$75
	.dc.b	$F8,$8B,$FE,$27,$7F,$E0,$8E,$8F,$21,$1F,$EE,$ED,$7F,$B8,$B5,$FE
	.dc.b	$E3,$7F,$EE,$C7,$C7,$EF,$7E,$52,$31,$1F,$E0,$69,$FC,$BF,$5B,$E2
	.dc.b	$BF,$04,$D9,$FF,$15,$35,$FF,$02,$B5,$FE,$FE,$D7,$DC,$D7,$DE,$DA
	.dc.b	$FE,$61,$3F,$8F,$F4,$1F,$4B,$FA,$86,$DF,$F9,$FD,$AF,$CA,$39,$CF
	.dc.b	$A3,$E2,$BC,$E4,$8F,$3B,$5F,$3B,$5F,$3B,$E1,$F8,$5F,$01,$23,$FD
	.dc.b	$1D,$8A,$7F,$77,$EC,$3E,$3B,$F4,$CD,$AF,$F4,$5C,$60,$8E,$08,$7F
	.dc.b	$A3,$9E,$3F,$35,$D0,$91,$FE,$83,$8F,$E2,$5D,$6F,$A5,$FF,$44,$CF
	.dc.b	$FA,$2F,$0F,$E1,$BD,$BE,$58,$FF,$62,$D6,$FC,$BF,$3F,$6F,$98,$93
	.dc.b	$FE,$A0,$7F,$7D,$F2,$3E,$98,$F9,$9C,$48,$C7,$3F,$F9,$B5,$BD,$1F
	.dc.b	$99,$DD,$F3,$5F,$C9,$F0,$7D,$2F,$F8,$75,$FE,$BE,$D7,$FD,$88,$D7
	.dc.b	$FD,$92,$D7,$FD,$9C,$D7,$FD,$F0,$D7,$AF,$A1,$FE,$03,$8D,$65,$FC
	.dc.b	$07,$6F,$E5,$FD,$DF,$CA,$EF,$D6,$DD,$3F,$D3,$D1,$C8,$FF,$65,$7C
	.dc.b	$3F,$B1,$FD,$9B,$6C,$D3,$F9,$4C,$7F,$BB,$F8,$FF,$9E,$C6,$5B,$E1
	.dc.b	$FF,$49,$ED,$7D,$2F,$F1,$1C,$7F,$D5,$D1,$8F,$2B,$53,$FD,$27,$1F
	.dc.b	$F3,$F3,$BE,$1F,$EC,$8D,$8C,$CF,$03,$8E,$9F,$47,$E9,$1F,$4B,$F4
	.dc.b	$FE,$97,$EA,$D8,$FD,$40,$72,$E2,$78,$BB,$CF,$E3,$C6,$23,$8C,$F0
	.dc.b	$9F,$96,$DC,$BB,$4A,$B8,$56,$B8,$FF,$B1,$FC,$67,$F9,$3A,$BF,$DF
	.dc.b	$4D,$7A,$C7,$F6,$8F,$A0,$FA,$F5,$FE,$BD,$5F,$0B,$93,$58,$CB,$F4
	.dc.b	$3F,$CB,$67,$71,$83,$F9,$7B,$87,$F3,$EF,$68,$E0,$F8,$49,$1C,$78
	.dc.b	$D1,$F8,$52,$C6,$5A,$F4,$7D,$5C,$5F,$A5,$FA,$57,$D2,$FF,$96,$DB
	.dc.b	$5F,$E4,$46,$BF,$28,$FB,$0E,$1C,$35,$F9,$ED,$77,$1D,$4F,$EF,$5C
	.dc.b	$1F,$E5,$CC,$FF,$DC,$CD,$CB,$FA,$93,$6B,$89,$EB,$E4,$FA,$5D,$72
	.dc.b	$30,$43,$E4,$B9,$7E,$5C,$7E,$87,$C7,$FD,$85,$FD,$3F,$8B,$FB,$86
	.dc.b	$7A,$65,$FC,$98,$CB,$FE,$8D,$AE,$2F,$E6,$7E,$95,$EA,$FF,$A0,$DB
	.dc.b	$1F,$E6,$C6,$BF,$24,$38,$7E,$63,$5F,$CB,$B5,$ED,$1D,$4F,$F1,$17
	.dc.b	$03,$E9,$7F,$28,$49,$FC,$61,$DA,$7F,$80,$7C,$BF,$E0,$B7,$FC,$C7
	.dc.b	$C0,$FE,$6C,$CE,$12,$FF,$11,$19,$7F,$D5,$95,$C7,$5F,$CD,$7D,$07
	.dc.b	$E8,$FE,$86,$57,$E3,$E3,$F4,$C6,$1F,$CD,$C6,$1F,$CA,$7D,$2F,$A8
	.dc.b	$8F,$C5,$E7,$FC,$DC,$7B,$BC,$AF,$D8,$12,$78,$9F,$9D,$E0,$FD,$A1
	.dc.b	$1F,$D0,$45,$7F,$57,$17,$5F,$DC,$4F,$EE,$FE,$43,$E5,$92,$BC,$5B
	.dc.b	$23,$E7,$F5,$FF,$01,$24,$FE,$B6,$3F,$82,$7D,$27,$D4,$24,$72,$7E
	.dc.b	$BF,$D2,$FF,$09,$24,$7E,$D6,$2B,$FC,$02,$2E,$3F,$C3,$66,$EF,$FC
	.dc.b	$26,$6D,$7F,$08,$9F,$E2,$9F,$23,$F8,$42,$47,$A9,$FC,$3F,$B5,$FC
	.dc.b	$64,$91,$FC,$16,$3F,$8E,$7D,$2F,$E3,$09,$1F,$09,$AF,$F0,$1E,$0F
	.dc.b	$E1,$F1,$88,$FE,$5B,$18,$7E,$D7,$CF,$BE,$93,$FD,$1E,$D8,$FE,$30
	.dc.b	$D7,$F8,$FD,$7F,$55,$FA,$57,$27,$FA,$B5,$B3,$FA,$9F,$29,$18,$8F
	.dc.b	$F5,$73,$5F,$E7,$F1,$A9,$E4,$48,$E5,$3B,$7F,$CC,$16,$F9,$0F,$39
	.dc.b	$23,$F9,$B5,$F3,$7D,$31,$97,$FC,$2B,$3C,$5D,$D7,$97,$83,$FB,$6A
	.dc.b	$E7,$FE,$A7,$F1,$1F,$EC,$BC,$7D,$71,$8F,$2F,$FD,$A1,$5B,$C3,$FB
	.dc.b	$35,$EE,$3F,$CF,$0D,$AE,$12,$BE,$6F,$0B,$E7,$3A,$9F,$93,$D7,$70
	.dc.b	$ED,$6C,$3A,$3C,$4E,$91,$A8,$F1,$4F,$82,$5F,$4B,$FE,$C1,$DF,$84
	.dc.b	$63,$C2,$FC,$05,$FF,$B9,$2B,$53,$DF,$3F,$F0,$17,$A8,$F9,$67,$D0
	.dc.b	$F5,$BD,$E1,$C7,$1E,$78,$7F,$7C,$9C,$BC,$AF,$8E,$3E,$29,$EE,$FF
	.dc.b	$59,$6B,$C6,$BF,$24,$DC,$BF,$9B,$3E,$CF,$2B,$FD,$C8,$F1,$6B,$E1
	.dc.b	$6B,$FE,$1B,$AF,$F3,$97,$C2,$72,$6C,$7B,$CD,$4F,$FB,$BF,$26,$5F
	.dc.b	$0B,$D6,$FF,$73,$76,$7A,$6B,$53,$DF,$1C,$1F,$57,$6F,$F3,$27,$92
	.dc.b	$BC,$E5,$DF,$89,$F4,$BF,$F0,$06,$7E,$E4,$63,$0E,$15,$BA,$FF,$1A
	.dc.b	$56,$9E,$11,$99,$DE,$9F,$99,$9C,$B6,$73,$97,$F8,$F7,$0E,$D8,$73
	.dc.b	$C3,$35,$F9,$7C,$38,$DE,$DF,$0C,$77,$AF,$C9,$75,$39,$FF,$DD,$91
	.dc.b	$FC,$2B,$CB,$F4,$F4,$7F,$96,$BE,$1D,$F2,$D8,$AF,$DE,$D5,$C7,$FD
	.dc.b	$BB,$87,$27,$F7,$31,$9D,$58,$72,$BE,$C7,$F0,$27,$15,$FE,$23,$9D
	.dc.b	$E1,$FE,$7A,$8F,$81,$3A,$7F,$89,$B5,$CC,$6D,$FD,$C4,$65,$B1,$99
	.dc.b	$EB,$5F,$CB,$34,$F5,$67,$F0,$DB,$93,$E9,$6C,$61,$ED,$3F,$25,$B3
	.dc.b	$BA,$FE,$6F,$18,$23,$82,$63,$FD,$06,$5F,$1F,$E0,$3C,$3E,$E5,$AF
	.dc.b	$E7,$AF,$62,$3F,$D1,$1F,$0F,$C6,$FA,$0F,$4F,$9B,$67,$FA,$2C,$5C
	.dc.b	$3F,$96,$CD,$E0,$FC,$BE,$7F,$83,$AC,$5F,$F3,$AA,$B7,$F7,$9F,$79
	.dc.b	$23,$FA,$AB,$F0,$BE,$3F,$E4,$9F,$2B,$E0,$EC,$FF,$56,$7E,$39,$1F
	.dc.b	$EB,$AF,$CA,$F9,$FF,$A2,$7D,$2F,$A5,$B3,$FD,$7A,$2E,$1F,$D1,$A6
	.dc.b	$E7,$EA,$D9,$1C,$6F,$FB,$2F,$C1,$5B,$FE,$9A,$FD,$82,$47,$F6,$E7
	.dc.b	$ED,$FD,$2F,$E0,$BE,$57,$F0,$6D,$9F,$ED,$ED,$AF,$EA,$ED,$C7,$FD
	.dc.b	$83,$EA,$3C,$BF,$C5,$FC,$1F,$E8,$E3,$18,$7F,$78,$8C,$4E,$A3,$FC
	.dc.b	$8A,$DC,$3F,$B1,$FD,$07,$E5,$0F,$D8,$7C,$73,$E4,$5E,$56,$70,$FF
	.dc.b	$51,$4E,$63,$F4,$8D,$8F,$F3,$43,$6F,$FD,$24,$D7,$FD,$7A,$D7,$FD
	.dc.b	$98,$D4,$D3,$1D,$6F,$5D,$F4,$9F,$AA,$F3,$3F,$81,$46,$30,$FF,$0B
	.dc.b	$C6,$27,$51,$FE,$83,$6E,$1F,$E0,$2E,$FF,$D9,$FD,$2F,$ED,$1B,$5F
	.dc.b	$E2,$06,$CE,$E3,$FC,$47,$3F,$BC,$F0,$9F,$19,$4F,$93,$5E,$31,$C7
	.dc.b	$FE,$A8,$8C,$E3,$F2,$27,$47,$C4,$29,$F1,$0A,$7C,$42,$EB,$FE,$F3
	.dc.b	$8E,$44,$C7,$FC,$E9,$F4,$BF,$CD,$30,$FD,$8B,$63,$FD,$F5,$DE,$8F
	.dc.b	$E8,$F3,$AC,$3F,$8B,$C6,$D4,$ED,$CF,$8D,$A7,$7F,$01,$C3,$E4,$6E
	.dc.b	$3F,$ED,$FC,$11,$8A,$F8,$73,$E2,$9C,$BB,$AE,$4F,$FB,$BE,$DF,$4B
	.dc.b	$F8,$18,$E9,$F3,$BF,$D1,$67,$3C,$7E,$88,$FE,$91,$8B,$81,$EC,$EF
	.dc.b	$1E,$53,$E2,$06,$E3,$57,$CA,$B8,$E3,$E3,$BE,$3B,$7F,$72,$C3,$84
	.dc.b	$E6,$FE,$4F,$CC,$7C,$AF,$94,$B3,$F0,$DB,$5E,$1E,$0F,$A5,$18,$C3
	.dc.b	$EB,$BF,$D2,$3E,$25,$F3,$F4,$C7,$CB,$DC,$7D,$37,$FC,$BA,$B7,$18
	.dc.b	$3E,$FB,$8C,$7C,$0A,$B9,$47,$8F,$E4,$1F,$19,$FD,$61,$B7,$CE,$7F
	.dc.b	$C3,$1F,$4A,$F2,$75,$7F,$DD,$E9,$39,$7F,$DD,$46,$27,$FB,$53,$4C
	.dc.b	$C6,$EB,$E9,$9F,$E1,$C7,$0E,$23,$83,$A3,$C7,$3D,$27,$71,$F9,$EF
	.dc.b	$A5,$78,$5F,$DC,$9B,$7E,$BD,$DC,$FF,$E3,$C8,$C4,$EA,$FE,$B7,$CA
	.dc.b	$3E,$37,$FB,$AB,$6B,$FD,$C4,$D9,$FF,$58,$35,$F5,$FB,$2F,$A5,$FF
	.dc.b	$6B,$57,$87,$E9,$5C,$DF,$E1,$FA,$D7,$C4,$FF,$82,$27,$35,$E7,$D1
	.dc.b	$EA,$FF,$83,$1B,$1E,$BC,$FF,$9B,$BB,$9F,$F0,$8C,$63,$3F,$2F,$91
	.dc.b	$EA,$25,$7A,$9B,$1A,$8C,$1F,$EF,$FE,$3F,$E9,$B1,$F9,$96,$F3,$FF
	.dc.b	$7E,$F9,$9F,$3C,$E7,$CB,$FF,$10,$C6,$23,$FD,$A2,$D7,$F8,$9B,$5F
	.dc.b	$E1,$ED,$E3,$34,$9C,$1F,$E2,$67,$5F,$07,$FE,$45,$C3,$FC,$1E,$DC
	.dc.b	$BA,$FD,$2F,$F8,$CA,$31,$8F,$68,$E4,$49,$FC,$64,$65,$B8,$F8,$A7
	.dc.b	$53,$FE,$ED,$C1,$18,$8F,$F7,$6F,$D2,$FF,$94,$31,$FF,$10,$B7,$0F
	.dc.b	$E6,$3E,$6F,$EC,$E7,$3E,$67,$F9,$1A,$3F,$6D,$7F,$EF,$66,$BC,$3B
	.dc.b	$FF,$C8,$B1,$88,$FE,$6A,$DC,$FF,$D9,$93,$A2,$61,$EA,$F8,$8F,$E5
	.dc.b	$F7,$FC,$45,$B8,$FF,$8D,$B0,$7D,$2F,$F9,$62,$73,$7E,$88,$F0,$91
	.dc.b	$F8,$B3,$C1,$B1,$9B,$FD,$B6,$3F,$ED,$F8,$C4,$7E,$FA,$3F,$DB,$5E
	.dc.b	$3F,$F9,$9B,$E0,$4F,$F9,$A2,$FF,$D5,$B5,$FC,$9E,$73,$1F,$B9,$33
	.dc.b	$1C,$DA,$FF,$5E,$6B,$FD,$69,$B5,$FE,$B1,$F8,$2F,$F3,$7C,$60,$8F
	.dc.b	$F9,$CA,$9F,$4B,$E5,$3C,$9C,$D6,$BC,$4F,$22,$3F,$88,$3F,$61,$87
	.dc.b	$BF,$E9,$5E,$C7,$FA,$11,$B3,$FC,$4D,$7A,$D9,$99,$39,$78,$46,$2B
	.dc.b	$FA,$57,$D2,$FF,$02,$6F,$67,$FA,$42,$1F,$01,$62,$61,$FC,$86,$78
	.dc.b	$71,$FF,$9D,$9D,$C6,$67,$FD,$5D,$F4,$6F,$A5,$A7,$FA,$C7,$E1,$BF
	.dc.b	$D1,$B5,$FD,$17,$95,$5B,$8C,$37,$3F,$EE,$13,$AB,$E5,$F0,$9D,$2F
	.dc.b	$FD,$70,$FF,$AA,$F0,$EB,$7F,$EB,$96,$BF,$EB,$68,$FF,$5A,$3F,$D2
	.dc.b	$38,$BF,$BD,$FD,$0B,$D0,$FE,$8D,$EA,$FF,$BA,$23,$15,$FD,$26,$B7
	.dc.b	$5A,$C6,$CE,$45,$70,$7F,$C0,$F7,$CE,$BF,$83,$DF,$09,$CD,$D7,$FC
	.dc.b	$1B,$1F,$C3,$1F,$31,$43,$5A,$7F,$12,$FA,$57,$C2,$FF,$9C,$E3,$0D
	.dc.b	$7F,$B1,$46,$BD,$FE,$CD,$8E,$5F,$0D,$FE,$52,$6F,$1E,$A3,$11,$C9
	.dc.b	$AF,$F7,$29,$47,$1D,$3E,$14,$7E,$32,$7E,$03,$5F,$8F,$9F,$1A,$E0
	.dc.b	$DB,$DF,$27,$F7,$EC,$39,$B6,$75,$7C,$87,$23,$F5,$A7,$CE,$7C,$B8
	.dc.b	$C3,$63,$33,$FC,$EA,$72,$FF,$3C,$E4,$7D,$87,$F0,$E8,$FC,$C3,$5F
	.dc.b	$CA,$B7,$37,$0F,$F9,$1A,$F9,$D6,$BB,$FF,$C2,$F5,$A8,$61,$DA,$73
	.dc.b	$7F,$E4,$E8,$C4,$FF,$8A,$EF,$9D,$F1,$97,$E2,$DF,$8F,$8C,$36,$F9
	.dc.b	$36,$FB,$BF,$E0,$A7,$F6,$4E,$6B,$9E,$9E,$29,$DB,$63,$BF,$93,$7E
	.dc.b	$5D,$F5,$D4,$97,$FB,$D8,$7D,$3E,$23,$F7,$71,$C1,$D4,$F3,$9F,$04
	.dc.b	$3F,$07,$FE,$CE,$FE,$8D,$AF,$F7,$06,$BF,$E3,$D6,$BF,$27,$D5,$F7
	.dc.b	$3F,$C2,$9B,$AF,$C9,$6E,$BF,$E4,$AA,$EB,$3F,$54,$7E,$A3,$FE,$94
	.dc.b	$BF,$EF,$F1,$8F,$84,$FD,$B3,$63,$33,$B7,$FC,$A1,$19,$6C,$65,$FC
	.dc.b	$1B,$5F,$F3,$AB,$5F,$C2,$FD,$BB,$E2,$7E,$93,$63,$F9,$53,$5F,$F4
	.dc.b	$B5,$77,$3F,$D2,$1F,$0B,$FA,$F8,$FF,$31,$35,$FF,$34,$B5,$FF,$4E
	.dc.b	$B5,$FF,$51,$F4,$7F,$11,$8C,$11,$FE,$2D,$C2,$FC,$C7,$FA,$9D,$7F
	.dc.b	$3D,$6B,$FC,$E9,$B7,$E4,$AF,$C7,$2F,$9D,$79,$5F,$31,$B7,$F4,$1A
	.dc.b	$FC,$C6,$BF,$AC,$6B,$FE,$9E,$6B,$FE,$EC,$6C,$F7,$9F,$4F,$E9,$1F
	.dc.b	$EC,$1A,$FE,$A8,$D7,$FA,$8B,$67,$F0,$95,$E2,$7F,$D7,$9F,$82,$BE
	.dc.b	$CF,$F9,$B3,$6F,$F9,$D3,$5F,$F2,$3B,$5F,$F4,$A3,$5F,$F4,$DB,$5F
	.dc.b	$F6,$7B,$53,$9C,$FE,$87,$81,$FB,$5D,$3F,$98,$D7,$D1,$8F,$D5,$D7
	.dc.b	$87,$0F,$E2,$1D,$6F,$B2,$FF,$4F,$DF,$F2,$16,$BF,$EA,$78,$D7,$C1
	.dc.b	$FF,$29,$9C,$CF,$FA,$C2,$72,$F1,$9D,$11,$FE,$55,$F9,$EF,$E8,$B9
	.dc.b	$FF,$8E,$36,$3F,$90,$B6,$B9,$C7,$FB,$BC,$7E,$73,$FB,$C7,$FD,$E2
	.dc.b	$D7,$3E,$D7,$F6,$58,$C1,$1C,$12,$3F,$B6,$47,$F2,$37,$AF,$7F,$F1
	.dc.b	$08,$C1,$FC,$0B,$0F,$E6,$5F,$9B,$FC,$CA,$FF,$09,$FE,$70,$F9,$BF
	.dc.b	$E4,$8D,$8F,$F0,$E4,$EF,$6F,$9F,$F9,$C3,$E5,$BF,$A9,$36,$BF,$A8
	.dc.b	$35,$FA,$0D,$7F,$DB,$BC,$9F,$CB,$BA,$BF,$99,$78,$46,$9F,$CD,$7E
	.dc.b	$AB,$EF,$6B,$FE,$E4,$FC,$51,$F5,$7F,$E0,$96,$E3,$FE,$0D,$6B,$FE
	.dc.b	$9A,$6B,$FE,$E7,$6B,$CE,$39,$11,$FF,$0B,$7E,$2B,$FC,$ED,$FB,$0D
	.dc.b	$D7,$FA,$F4,$23,$17,$CE,$07,$DF,$FB,$17,$C1,$7F,$42,$FB,$27,$D2
	.dc.b	$D3,$FA,$2C,$EA,$B8,$E3,$DA,$31,$1F,$89,$AE,$2E,$B0,$EE,$A3,$2F
	.dc.b	$87,$0E,$BF,$67,$E1,$FA,$F9,$3C,$4F,$3E,$09,$DC,$64,$E9,$3C,$1F
	.dc.b	$64,$7A,$3C,$9F,$F1,$D5,$7F,$3D,$93,$49,$1A,$24,$7F,$90,$E3,$87
	.dc.b	$8B,$FE,$FC,$8F,$03,$5F,$D4,$37,$3F,$89,$5F,$DB,$23,$F1,$93,$F0
	.dc.b	$9A,$FD,$26,$BF,$CC,$B9,$7F,$CA,$11,$82,$38,$3F,$CA,$9A,$BE,$97
	.dc.b	$FB,$CF,$D9,$FF,$A8,$AF,$E2,$35,$FD,$E4,$ED,$F9,$0D,$7F,$80,$46
	.dc.b	$F1,$FC,$4B,$77,$FB,$35,$B9,$FC,$6A,$FA,$7F,$19,$2F,$FC,$D5,$19
	.dc.b	$6C,$65,$FF,$7C,$C6,$5B,$1F,$E6,$FB,$FF,$4D,$47,$AB,$DF,$F6,$9F
	.dc.b	$E1,$6A,$E1,$7F,$9E,$FC,$81,$F6,$3F,$E1,$AF,$C5,$5F,$67,$4F,$F0
	.dc.b	$EC,$EB,$1F,$C4,$35,$FE,$41,$3C,$1F,$C6,$35,$FE,$49,$1B,$BF,$D6
	.dc.b	$FD,$AF,$F5,$79,$FC,$E8,$70,$FE,$81,$AF,$EC,$1A,$EF,$E8,$FF,$D1
	.dc.b	$D1,$82,$3F,$E9,$1A,$FD,$3C,$F6,$6B,$FA,$C6,$BF,$BC,$6A,$69,$32
	.dc.b	$E8,$78,$3E,$A7,$FC,$83,$66,$9F,$22,$3F,$C8,$BF,$19,$FE,$3C,$BF
	.dc.b	$7B,$5F,$DE,$B5,$FE,$1A,$D7,$FD,$EB,$ED,$7F,$AA,$E3,$04,$71,$D3
	.dc.b	$27,$FD,$65,$3F,$C3,$1E,$F1,$97,$E7,$9F,$F0,$01,$C3,$A3,$FD,$19
	.dc.b	$FA,$0F,$E7,$5F,$ED,$0D,$7F,$D0,$CD,$7F,$D3,$2D,$F0,$9F,$BE,$FC
	.dc.b	$39,$FF,$74,$35,$FF,$72,$B7,$6F,$EF,$B5,$FA,$5E,$5F,$F4,$57,$67
	.dc.b	$E6,$AF,$FD,$D6,$DA,$FF,$45,$36,$7F,$D2,$17,$CD,$E3,$C9,$FE,$81
	.dc.b	$9D,$5F,$FA,$61,$B8,$F9,$B6,$FF,$44,$79,$6F,$C1,$5F,$6D,$FE,$B1
	.dc.b	$6E,$1F,$EA,$F6,$BF,$E0,$F6,$BF,$D8,$3E,$57,$FC,$27,$AF,$FA,$7D
	.dc.b	$BA,$7F,$A9,$5B,$3F,$EC,$0A,$E7,$5F,$E1,$EF,$DD,$1F,$73,$FE,$C1
	.dc.b	$6D,$7F,$AF,$DA,$FF,$87,$DA,$FF,$70,$F7,$9F,$41,$FE,$A1,$6C,$7F
	.dc.b	$43,$DB,$FD,$2D,$3F,$02,$B9,$E1,$FE,$B5,$BD,$DA,$31,$5C,$CE,$18
	.dc.b	$59,$E9,$5C,$2B,$71,$FE,$C1,$BE,$B9,$FF,$68,$FA,$17,$DF,$FF,$09
	.dc.b	$AF,$EB,$B1,$8E,$EF,$CA,$C2,$30,$FE,$61,$AF,$F0,$8D,$1F,$8B,$7F
	.dc.b	$76,$8F,$F6,$AB,$63,$3E,$77,$FB,$16,$B8,$E7,$FD,$FB,$F5,$5F,$E7
	.dc.b	$D8,$7F,$12,$6B,$FD,$C1,$AF,$F8,$01,$A9,$A4,$EA,$F4,$C6,$5F,$F4
	.dc.b	$7F,$E2,$EE,$3F,$F6,$A3,$67,$EA,$35,$FF,$25,$57,$F4,$97,$FC,$1F
	.dc.b	$FA,$CF,$F8,$DD,$FD,$CB,$5F,$E5,$2D,$7F,$C0,$ED,$78,$FA,$BE,$48
	.dc.b	$C4,$1A,$6F,$47,$DC,$7F,$85,$33,$FF,$7C,$37,$87,$FD,$FC,$DB,$EA
	.dc.b	$7E,$35,$C5,$72,$EE,$FF,$79,$46,$E7,$C1,$8F,$FB,$DF,$EE,$DF,$B8
	.dc.b	$7F,$85,$67,$FC,$5B,$8F,$F8,$82,$73,$E7,$FF,$88,$A3,$15,$BA,$FF
	.dc.b	$1B,$D6,$9E,$27,$F8,$D9,$C3,$A8,$E1,$3C,$AB,$A4,$EE,$72,$D8,$FF
	.dc.b	$18,$FC,$AF,$F7,$3C,$FE,$2F,$DA,$63,$CC,$F8,$6D,$B6,$5E,$0B,$FE
	.dc.b	$ED,$F7,$8F,$BD,$CD,$27,$31,$86,$D7,$F9,$39,$FF,$01,$73,$3E,$9F
	.dc.b	$FC,$E1,$1A,$BE,$59,$7F,$91,$B1,$ED,$87,$18,$C6,$3D,$5F,$D9,$61
	.dc.b	$C9,$CC,$FE,$C2,$B8,$DF,$F4,$66,$BF,$DD,$9A,$F7,$F2,$5F,$35,$C5
	.dc.b	$B1,$C5,$AF,$16,$BC,$63,$CC,$FF,$9B,$27,$FC,$D3,$18,$CD,$F7,$AF
	.dc.b	$A3,$71,$70,$C1,$87,$07,$E1,$9C,$F1,$F1,$ED,$F4,$9B,$A7,$F1,$3C
	.dc.b	$7A,$61,$CF,$0E,$57,$FE,$31,$D1,$F9,$0F,$F2,$CA,$F3,$F0,$BE,$D3
	.dc.b	$FC,$FF,$1E,$AA,$DF,$8B,$E6,$D7,$C7,$7B,$D7,$19,$75,$CB,$A5,$7E
	.dc.b	$69,$B8,$FE,$85,$B1,$F9,$B6,$BF,$8E,$6E,$3E,$09,$FD,$0C,$E7,$1F
	.dc.b	$14,$73,$6C,$7F,$A6,$B1,$F5,$7B,$9E,$7E,$AF,$44,$62,$FF,$D0,$DE
	.dc.b	$BF,$C6,$BF,$E8,$91,$C7,$5F,$F5,$BB,$FD,$35,$B1,$A1,$5C,$DF,$F7
	.dc.b	$35,$73,$BE,$F5,$C1,$FF,$58,$C7,$79,$FF,$22,$B5,$FF,$51,$B5,$FF
	.dc.b	$59,$B5,$FF,$6C,$70,$FF,$AE,$9D,$B5,$DB,$5D,$E1,$FE,$24,$C1,$F1
	.dc.b	$AF,$47,$E2,$4F,$C7,$8C,$3F,$EA,$9F,$C6,$5F,$84,$6B,$0F,$7D,$69
	.dc.b	$F8,$42,$77,$5F,$D7,$9B,$1F,$E3,$46,$BF,$E6,$46,$BF,$EC,$26,$BF
	.dc.b	$ED,$26,$BD,$6B,$D9,$1A,$C1,$F2,$7F,$E8,$FB,$FF,$01,$36,$FE,$B7
	.dc.b	$3B,$EA,$3F,$C4,$35,$AE,$8F,$64,$E7,$CF,$F9,$4D,$9F,$EE,$95,$C2
	.dc.b	$FF,$CF,$73,$97,$FB,$B7,$E3,$2F,$7F,$E3,$66,$8C,$61,$FE,$D7,$F4
	.dc.b	$9F,$72,$FA,$9D,$D6,$B1,$FC,$66,$1C,$71,$C9,$F4,$70,$E3,$78,$C7
	.dc.b	$FC,$41,$87,$1E,$5F,$CE,$B6,$3F,$DF,$75,$C1,$DD,$F3,$CB,$C5,$1F
	.dc.b	$E6,$89,$D6,$5F,$E9,$86,$BF,$EF,$66,$BF,$AD,$F9,$1F,$CF,$4D,$E5
	.dc.b	$F3,$7E,$30,$FB,$8F,$87,$5A,$B6,$0A,$D6,$1E,$16,$CF,$85,$AF,$85
	.dc.b	$AF,$EF,$7F,$19,$7D,$BF,$D0,$6D,$7F,$90,$FD,$B7,$D0,$E6,$A4,$FF
	.dc.b	$91,$EB,$77,$FE,$72,$AC,$B6,$79,$47,$D1,$E9,$79,$3F,$7B,$1E,$B5
	.dc.b	$F7,$67,$2F,$CB,$6D,$FC,$B6,$BF,$2D,$AF,$F8,$13,$F2,$C7,$D2,$FE
	.dc.b	$EE,$91,$87,$FD,$3A,$62,$35,$19,$9F,$F6,$7C,$E4,$DC,$66,$3F,$C1
	.dc.b	$CD,$7C,$13,$FC,$27,$C9,$7E,$AD,$FA,$99,$12,$33,$D9,$B9,$D4,$E5
	.dc.b	$AF,$F6,$86,$CF,$0C,$3F,$D6,$AD,$9F,$F7,$2B,$5E,$E7,$1F,$85,$FE
	.dc.b	$21,$1E,$47,$F8,$94,$3E,$F7,$FC,$21,$5E,$6E,$9F,$EA,$97,$B8,$C1
	.dc.b	$D2,$77,$5F,$EC,$66,$BF,$EF,$66,$BB,$F3,$BF,$95,$46,$23,$E1,$B5
	.dc.b	$F1,$46,$8F,$E6,$56,$FC,$25,$F6,$F8,$6E,$1F,$ED,$69,$D4,$E4,$91
	.dc.b	$99,$DD,$FC,$43,$FD,$30,$E1,$F8,$F5,$CA,$7F,$D1,$BF,$87,$FF,$30
	.dc.b	$8F,$87,$A7,$F8,$E2,$7E,$1D,$7D,$CA,$D4,$7D,$9F,$2F,$E6,$C6,$27
	.dc.b	$ED,$46,$38,$7E,$9D,$F3,$9F,$EC,$5A,$BF,$00,$7C,$AF,$D8,$6E,$5E
	.dc.b	$0A,$65,$FD,$27,$3F,$AF,$5A,$9F,$E9,$73,$FD,$35,$F8,$0D,$7F,$0A
	.dc.b	$D7,$D5,$E4,$FF,$5F,$8C,$47,$68,$C1,$2F,$FD,$63,$F8,$07,$FA,$5D
	.dc.b	$F8,$79,$7A,$7E,$07,$F1,$B3,$9B,$F6,$E2,$9C,$FB,$DE,$D6,$E1,$CA
	.dc.b	$74,$FF,$3F,$FC,$65,$ED,$7F,$75,$F0,$11,$89,$FE,$60,$DC,$7F,$98
	.dc.b	$35,$FE,$60,$D7,$F9,$DB,$5F,$E5,$6D,$9E,$B5,$FB,$BB,$75,$FD,$33
	.dc.b	$DB,$F6,$57,$1F,$A9,$A2,$3F,$C1,$0D,$8F,$8F,$18,$23,$82,$65,$F9
	.dc.b	$51,$3C,$5C,$74,$FD,$DE,$1F,$BF,$E2,$FE,$EA,$31,$7F,$C3,$72,$FE
	.dc.b	$CA,$FE,$1E,$33,$7F,$E7,$29,$CB,$FE,$10,$8C,$F0,$7E,$46,$FA,$7C
	.dc.b	$C7,$E4,$7E,$46,$E3,$F5,$5B,$87,$DC,$6E,$5F,$92,$7F,$31,$E7,$4A
	.dc.b	$F1,$79,$16,$FC,$0F,$F8,$B7,$3F,$F1,$A1,$1C,$12,$7F,$DF,$F1,$97
	.dc.b	$FC,$65,$F1,$5F,$E1,$5C,$7F,$CA,$6D,$8F,$F2,$C6,$5F,$B1,$F3,$53
	.dc.b	$9C,$7F,$B4,$F1,$7F,$8E,$AF,$FB,$C3,$63,$FB,$C3,$5F,$EF,$0D,$7F
	.dc.b	$CC,$9F,$34,$FC,$1B,$FC,$2F,$F3,$8F,$A1,$FF,$0C,$CE,$70,$FE,$0F
	.dc.b	$92,$B5,$1F,$BB,$F2,$1F,$09,$FC,$D1,$B1,$FB,$FE,$8F,$D5,$6F,$FE
	.dc.b	$87,$D3,$FB,$4D,$FF,$3A,$9F,$D7,$47,$F1,$D6,$BF,$D0,$9A,$FF,$52
	.dc.b	$6B,$F9,$CE,$CF,$F2,$F4,$60,$8F,$F9,$8F,$27,$D0,$FF,$0F,$C7,$FD
	.dc.b	$08,$DB,$FF,$09,$FD,$E7,$EE,$1F,$F1,$8C,$67,$B1,$39,$9F,$F1,$13
	.dc.b	$5F,$F1,$6B,$5E,$7D,$DF,$BF,$8C,$4F,$28,$C1,$2F,$FC,$E9,$F8,$57
	.dc.b	$F7,$9D,$7F,$D2,$AD,$8F,$EC,$7F,$A5,$7F,$AC,$B1,$FF,$22,$CE,$5F
	.dc.b	$EE,$B1,$97,$FA,$33,$5E,$3F,$05,$FE,$83,$8C,$47,$08,$C1,$2B,$F7
	.dc.b	$50,$9D,$E8,$FA,$97,$D5,$7F,$97,$9B,$7F,$D3,$32,$7E,$9D,$F8,$6C
	.dc.b	$7F,$CB,$75,$A7,$FD,$51,$3A,$E6,$FE,$CF,$5B,$9F,$EB,$6D,$7F,$BE
	.dc.b	$35,$4F,$1B,$FE,$D5,$39,$8C,$4F,$F6,$76,$BF,$E0,$CC,$5F,$9F,$BE
	.dc.b	$AB,$FC,$EE,$DC,$3F,$AB,$7E,$A8,$F0,$73,$5F,$E1,$08,$C4,$FC,$66
	.dc.b	$E1,$FC,$85,$AF,$F7,$76,$BF,$E1,$C6,$BF,$E2,$C6,$A6,$9A,$C3,$C5
	.dc.b	$F7,$FF,$E8,$CC,$0F,$FA,$CF,$2F,$F4,$77,$C1,$E6,$31,$EF,$7E,$FF
	.dc.b	$1F,$8E,$DA,$DE,$5F,$E1,$18,$C1,$2F,$FC,$32,$D8,$FF,$64,$7C,$47
	.dc.b	$F9,$E7,$23,$FE,$7B,$C3,$F8,$55,$FF,$A8,$23,$1D,$FE,$19,$CD,$FE
	.dc.b	$32,$30,$79,$6D,$F8,$BF,$F3,$CC,$BF,$DE,$AD,$AF,$F7,$D2,$F8,$9F
	.dc.b	$F5,$6B,$9C,$3F,$DE,$2D,$BE,$7E,$46,$AB,$6F,$FA,$D3,$F1,$7F,$F1
	.dc.b	$5E,$2C,$7F,$CB,$B6,$9C,$C6,$1F,$E0,$93,$BC,$7F,$DF,$F1,$A6,$FA
	.dc.b	$DD,$4E,$DD,$76,$FF,$94,$FD,$BF,$ED,$F1,$3F,$B9,$ED,$7D,$C7,$DC
	.dc.b	$AD,$FB,$9F,$EC,$AA,$E0,$6E,$75,$97,$FB,$1A,$31,$C9,$FE,$0D,$9C
	.dc.b	$CF,$FA,$DD,$AF,$FB,$09,$AF,$F9,$2F,$E3,$7F,$CF,$BF,$72,$FA,$FF
	.dc.b	$F3,$9E,$58,$7F,$BE,$4D,$AF,$F3,$AF,$A3,$F0,$EF,$84,$73,$0E,$1F
	.dc.b	$F4,$45,$F1,$7F,$CF,$B5,$55,$73,$BE,$F1,$FC,$48,$4D,$69,$97,$F8
	.dc.b	$BF,$DD,$3E,$F1,$F7,$98,$6D,$7F,$85,$67,$53,$93,$8F,$77,$F0,$C9
	.dc.b	$D3,$5F,$F3,$84,$FF,$07,$E5,$D9,$23,$DF,$8F,$FA,$62,$7F,$92,$C7
	.dc.b	$BD,$AE,$8F,$F5,$5B,$CD,$C0,$71,$A7,$C9,$78,$B7,$C3,$FF,$52,$B9
	.dc.b	$C3,$D0,$FD,$2E,$2F,$C5,$F1,$7F,$3A,$6E,$1F,$EF,$DE,$0F,$F6,$6B
	.dc.b	$67,$F9,$FF,$1F,$E5,$3E,$13,$FD,$17,$C5,$FA,$4E,$4F,$F7,$4D,$FE
	.dc.b	$6A,$72,$FF,$BE,$B0,$E6,$FE,$4E,$33,$7C,$E3,$81,$FE,$90,$C7,$C6
	.dc.b	$FF,$3A,$C3,$F6,$3F,$79,$F8,$7E,$0F,$D7,$5A,$31,$3A,$E0,$EB,$0F
	.dc.b	$6E,$C4,$C3,$FC,$2F,$3C,$15,$F3,$AF,$2F,$DE,$3E,$7E,$92,$BF,$C3
	.dc.b	$F0,$FB,$CC,$47,$9C,$E7,$1C,$5D,$CF,$49,$F8,$B0,$FA,$47,$DB,$7E
	.dc.b	$6D,$B1,$FD,$B9,$AF,$F9,$A5,$AF,$F5,$7F,$BC,$F7,$3F,$E3,$26,$BE
	.dc.b	$3C,$91,$8D,$BF,$C5,$D5,$D1,$FE,$30,$3E,$F1,$F2,$5F,$DF,$A3,$0D
	.dc.b	$8C,$E1,$FE,$D5,$6B,$FD,$AD,$AF,$FB,$41,$AF,$FB,$C5,$AF,$F7,$66
	.dc.b	$E0,$FB,$C7,$CB,$FE,$19,$B5,$FE,$4E,$6B,$FE,$06,$AF,$0E,$1F,$E4
	.dc.b	$9F,$87,$FD,$5C,$FF,$B8,$DA,$88,$E1,$BD,$7F,$C2,$E3,$11,$FE,$D6
	.dc.b	$8C,$11,$FF,$4D,$62,$FB,$B7,$E0,$1F,$9C,$6C,$FF,$99,$1A,$FE,$14
	.dc.b	$E8,$7F,$88,$FC,$C7,$DA,$FC,$F6,$C7,$FC,$01,$6B,$FE,$17,$6B,$FE
	.dc.b	$FC,$6B,$ED,$C9,$18,$8F,$F1,$1B,$63,$33,$FB,$CF,$51,$F0,$7F,$B4
	.dc.b	$6D,$FF,$9F,$1A,$FF,$8D,$2F,$C7,$7F,$E7,$7F,$96,$7D,$EF,$F3,$36
	.dc.b	$D7,$F9,$19,$AF,$FB,$05,$AF,$D3,$3E,$A3,$87,$0D,$73,$1D,$4F,$E2
	.dc.b	$79,$7F,$8D,$63,$E7,$78,$7F,$44,$FD,$6D,$78,$2F,$FB,$A7,$73,$EF
	.dc.b	$7F,$D3,$57,$FE,$D3,$8C,$7B,$1F,$C8,$E7,$33,$E5,$13,$A7,$84,$69
	.dc.b	$F5,$46,$5E,$0D,$73,$1B,$7F,$D1,$A7,$F2,$2F,$77,$FC,$A5,$E0,$E2
	.dc.b	$3F,$91,$4F,$8B,$0F,$F5,$27,$DE,$FF,$00,$9F,$90,$D7,$F3,$8D,$4D
	.dc.b	$27,$67,$F6,$76,$C7,$F9,$E9,$FE,$44,$F0,$F0,$BF,$8A,$C6,$0F,$E0
	.dc.b	$D8,$7F,$B1,$35,$FE,$73,$F0,$6F,$85,$A7,$F3,$CF,$85,$FA,$98,$FF
	.dc.b	$23,$D9,$F7,$AF,$A0,$FE,$A0,$D9,$FA,$4D,$7E,$4B,$5F,$F3,$A7,$88
	.dc.b	$ED,$FE,$92,$D8,$FF,$68,$B5,$FE,$41,$F1,$3F,$81,$D7,$F1,$DF,$7A
	.dc.b	$FA,$3F,$F0,$43,$6B,$FD,$36,$D7,$F0,$01,$C3,$F8,$36,$BD,$23,$99
	.dc.b	$1F,$F5,$2F,$DE,$FF,$41,$7F,$80,$37,$93,$FD,$BC,$DB,$E5,$3F,$B9
	.dc.b	$7F,$49,$F7,$9F,$EE,$88,$C1,$FC,$67,$EF,$5F,$0F,$FD,$29,$B1,$A1
	.dc.b	$3C,$5F,$E2,$0D,$C3,$FD,$DD,$3C,$27,$F8,$A3,$63,$FD,$E5,$3B,$BF
	.dc.b	$1F,$D5,$77,$F8,$C9,$87,$7A,$4F,$F5,$C6,$C7,$8F,$05,$70,$8F,$1D
	.dc.b	$6F,$CA,$FF,$1D,$5F,$F4,$49,$F4,$C6,$08,$FF,$B4,$23,$44,$8F,$F2
	.dc.b	$17,$88,$FC,$03,$FB,$6B,$6B,$E2,$35,$F8,$73,$AE,$4E,$24,$8F,$F6
	.dc.b	$EC,$FE,$45,$FE,$2C,$D7,$FA,$53,$5F,$E7,$31,$C5,$FC,$4E,$AF,$BE
	.dc.b	$FE,$FF,$F7,$CF,$B3,$A7,$F8,$0A,$75,$5F,$65,$AF,$F4,$79,$E0,$FD
	.dc.b	$D6,$BF,$D2,$E3,$7C,$1F,$E9,$18,$FC,$2F,$D0,$FF,$99,$64,$D2,$47
	.dc.b	$FB,$EA,$32,$D8,$CB,$EA,$7F,$CD,$D4,$FB,$9F,$59,$27,$D1,$3F,$E1
	.dc.b	$CF,$BD,$7C,$1F,$F8,$7B,$EF,$BF,$DC,$D5,$F9,$56,$BF,$D8,$E7,$6F
	.dc.b	$E6,$5A,$FF,$67,$8D,$E5,$ED,$6F,$2F,$F4,$9A,$DC,$FC,$1F,$BE,$FE
	.dc.b	$AF,$7F,$CE,$5A,$FF,$4F,$6B,$FE,$03,$6A,$69,$3C,$EF,$F4,$73,$AA
	.dc.b	$FD,$44,$FF,$45,$6B,$E3,$D7,$FD,$29,$18,$8F,$F0,$83,$53,$49,$1D
	.dc.b	$BE,$F3,$FC,$8B,$46,$9F,$36,$3F,$C9,$3B,$FF,$8F,$AF,$FA,$B3,$5F
	.dc.b	$ED,$8D,$7F,$C4,$0D,$7F,$25,$ED,$7F,$AA,$63,$04,$71,$F2,$F2,$7F
	.dc.b	$91,$9F,$99,$70,$FF,$26,$9F,$E1,$CF,$8A,$32,$D3,$FB,$77,$DF,$7E
	.dc.b	$71,$FE,$CE,$D7,$FC,$69,$E1,$A3,$11,$FD,$C1,$BE,$A3,$EF,$BE,$24
	.dc.b	$FF,$24,$8C,$67,$FE,$8C,$9C,$CF,$F5,$1A,$FD,$2F,$07,$FB,$EB,$B7
	.dc.b	$F3,$53,$C5,$AF,$F9,$9D,$BC,$3F,$E8,$EB,$E6,$FF,$A4,$23,$35,$CF
	.dc.b	$97,$FD,$37,$18,$BF,$F4,$5D,$F0,$AF,$6F,$CD,$BE,$77,$FD,$60,$DA
	.dc.b	$FF,$09,$35,$FF,$04,$B5,$FF,$13,$68,$F9,$BF,$F0,$25,$FF,$96,$E3
	.dc.b	$1E,$97,$FA,$B2,$73,$5F,$E1,$8B,$E9,$A7,$F8,$8B,$EF,$5F,$7D,$FE
	.dc.b	$C1,$6D,$7F,$88,$DA,$FF,$86,$9A,$F8,$BD,$F7,$C9,$7F,$AE,$9B,$1F
	.dc.b	$4D,$AF,$F8,$06,$79,$DF,$FA,$E2,$72,$DA,$E4,$7F,$B2,$38,$D5,$C3
	.dc.b	$5F,$75,$F0,$6D,$70,$C3,$FD,$91,$5B,$5F,$62,$FB,$EF,$E1,$71,$FE
	.dc.b	$4F,$8C,$77,$7E,$71,$B1,$FE,$00,$6B,$FE,$41,$E0,$7D,$F7,$F7,$78
	.dc.b	$FF,$63,$B6,$33,$E7,$7F,$B8,$2B,$8E,$DF,$CC,$FE,$F7,$F9,$BE,$1F
	.dc.b	$E4,$B6,$BF,$E6,$16,$BF,$E9,$56,$A6,$93,$AF,$F6,$71,$97,$FD,$EB
	.dc.b	$F7,$DF,$85,$8F,$D2,$35,$FF,$61,$36,$BC,$93,$FE,$D2,$FB,$EF,$F1
	.dc.b	$D5,$7E,$F9,$AF,$F8,$89,$AF,$F9,$8D,$AF,$F9,$E5,$AF,$FA,$B5,$A9
	.dc.b	$A6,$B2,$7D,$F7,$F8,$36,$7F,$DE,$2D,$8C,$F7,$79,$7D,$6F,$F4,$36
	.dc.b	$3F,$F0,$02,$B4,$49,$D5,$FF,$89,$BE,$B9,$F7,$DF,$E1,$28,$FE,$C9
	.dc.b	$18,$F4,$3F,$D2,$1C,$DF,$E9,$18,$C4,$FF,$B1,$63,$15,$C5,$B5,$C2
	.dc.b	$7F,$D9,$55,$C1,$EB,$5B,$72,$D5,$1B,$F9,$6F,$ED,$FA,$7C,$2F,$A7
	.dc.b	$FF,$51,$46,$33,$46,$3C,$BE,$79,$5C,$7F,$DB,$5F,$78,$FB,$2E,$B0
	.dc.b	$C7,$FB,$DA,$9D,$67,$FE,$E1,$9C,$C6,$27,$86,$1F,$EE,$5E,$57,$DF
	.dc.b	$7F,$9C,$EF,$78,$F6,$6E,$3C,$EF,$FD,$75,$8F,$39,$FF,$05,$E1,$CB
	.dc.b	$D1,$F4,$65,$E3,$C7,$CE,$FF,$AB,$6F,$C4,$D8,$CB,$D1,$AF,$C0,$F2
	.dc.b	$DF,$37,$FD,$52,$BF,$CC,$6D,$8C,$CF,$F9,$EA,$72,$D8,$CE,$3F,$09
	.dc.b	$B9,$7C,$1F,$BE,$7D,$63,$93,$DD,$87,$87,$2A,$FC,$DC,$38,$B9,$3E
	.dc.b	$59,$FC,$4A,$F9,$D7,$0C,$BC,$79,$7F,$8C,$36,$7D,$27,$F2,$CA,$EF
	.dc.b	$C8,$FB,$EF,$F3,$FC,$78,$AB,$7E,$1F,$97,$3E,$F7,$BD,$31,$97,$5C
	.dc.b	$BA,$57,$E8,$9B,$8F,$E8,$5B,$1F,$9C,$6B,$F9,$86,$E3,$E0,$9F,$D3
	.dc.b	$4E,$F1,$F1,$47,$36,$C7,$F5,$5C,$BF,$D6,$13,$FD,$DD,$B0,$E9,$FF
	.dc.b	$2C,$71,$7C,$68,$C5,$71,$F7,$7D,$D5,$EF,$D7,$E2,$46,$27,$47,$FA
	.dc.b	$86,$F9,$56,$F0,$EE,$F6,$E1,$FF,$33,$62,$25,$78,$5F,$F8,$00,$D7
	.dc.b	$47,$FB,$66,$38,$BC,$1A,$E6,$FF,$C4,$F3,$C6,$35,$1F,$EC,$CC,$5F
	.dc.b	$1A,$F4,$BD,$54,$9C,$C6,$21,$F7,$CF,$B5,$DD,$6B,$0F,$F6,$04,$E5
	.dc.b	$FF,$1B,$B6,$BF,$DB,$6D,$7D,$67,$4A,$F4,$C6,$63,$DE,$D7,$E3,$35
	.dc.b	$E1,$3E,$BF,$95,$FF,$46,$BF,$E3,$36,$C7,$DA,$3F,$B9,$FD,$F7,$F8
	.dc.b	$86,$75,$D1,$F0,$3A,$93,$9A,$F2,$C6,$3D,$1F,$9E,$DB,$FE,$ED,$5B
	.dc.b	$AF,$F1,$2F,$D4,$BD,$FF,$8A,$DA,$FF,$74,$F9,$EF,$A7,$7D,$F6,$AE
	.dc.b	$57,$F9,$2E,$F8,$61,$E9,$C3,$85,$EA,$3D,$9C,$0A,$DF,$2F,$E5,$5B
	.dc.b	$87,$F9,$73,$2E,$EF,$5C,$BE,$FE,$5D,$23,$E9,$E2,$8C,$3F,$59,$AF
	.dc.b	$E4,$9A,$FF,$90,$7E,$4B,$F3,$F3,$79,$BA,$FD,$EB,$EF,$BF,$0F,$5A
	.dc.b	$BE,$D3,$9B,$F0,$E0,$8C,$4F,$85,$AF,$85,$AF,$F9,$47,$EF,$5F,$7D
	.dc.b	$F9,$46,$DF,$F9,$B7,$E8,$B1,$32,$3F,$CF,$13,$AA,$FF,$38,$56,$5B
	.dc.b	$3C,$4F,$B3,$D1,$FD,$91,$C4,$7D,$58,$C3,$F1,$DA,$FC,$76,$BF,$E8
	.dc.b	$E6,$BF,$CF,$7E,$F9,$F4,$DF,$BC,$A4,$61,$FA,$A6,$23,$51,$99,$FD
	.dc.b	$5C,$E4,$DC,$66,$3F,$C1,$ED,$7F,$C1,$CD,$7F,$AD,$FD,$E7,$EA,$1F
	.dc.b	$E1,$FE,$42,$31,$1F,$C5,$E3,$1D,$5F,$C4,$9B,$7F,$8A,$6C,$FE,$F9
	.dc.b	$AF,$4A,$E6,$F8,$2B,$94,$3E,$1B,$F8,$9D,$78,$5F,$E2,$5F,$7D,$FE
	.dc.b	$0F,$C3,$F9,$47,$3F,$F5,$1A,$D4,$60,$FE,$4F,$3B,$C7,$F3,$AD,$7F
	.dc.b	$88,$B5,$DF,$97,$FC,$AA,$31,$1F,$CD,$5A,$9A,$6A,$DF,$7C,$FA,$FF
	.dc.b	$E6,$B1,$86,$C6,$70,$FD,$7C,$E5,$B1,$9B,$FE,$4B,$39,$6C,$65,$FF
	.dc.b	$29,$B5,$FE,$8B,$F7,$DF,$CB,$E7,$E2,$69,$FE,$40,$BF,$89,$5F,$D3
	.dc.b	$EB,$51,$F6,$E7,$53,$FC,$9C,$4E,$9F,$C2,$F9,$3F,$D0,$E3,$1A,$7E
	.dc.b	$16,$3E,$BF,$DD,$BE,$5A,$9B,$F1,$AE,$5E,$2C,$7A,$C7,$F5,$8C,$7F
	.dc.b	$A9,$4F,$FB,$F9,$B3,$FC,$51,$AF,$F4,$86,$BF,$02,$79,$12,$7F,$8D
	.dc.b	$C3,$EF,$3F,$DB,$11,$F0,$B2,$62,$F7,$FF,$AF,$56,$AF,$FA,$45,$6A
	.dc.b	$73,$F0,$BE,$E6,$E1,$CF,$0E,$4F,$F3,$EF,$BE,$7B,$7F,$DD,$BB,$91
	.dc.b	$89,$FE,$3E,$DC,$BF,$97,$35,$FE,$5C,$D7,$FC,$0D,$3C,$AB,$A4,$E6
	.dc.b	$7F,$9C,$37,$2F,$6F,$C4,$FD,$43,$DB,$F7,$D7,$1F,$A5,$AA,$3F,$C1
	.dc.b	$2D,$8F,$92,$DC,$3F,$A2,$B5,$FE,$F2,$D7,$FC,$2A,$D7,$FC,$38,$D7
	.dc.b	$E9,$4F,$F8,$5B,$A3,$F0,$D5,$F2,$7C,$5F,$F0,$74,$62,$BE,$8E,$BF
	.dc.b	$D9,$EB,$F1,$13,$9E,$5F,$E0,$13,$9A,$FC,$C4,$E6,$BF,$8B,$4E,$6B
	.dc.b	$FC,$AD,$5F,$C3,$7E,$67,$F2,$FF,$71,$E0,$6D,$F6,$6C,$F4,$6D,$78
	.dc.b	$3C,$DF,$CF,$9E,$8F,$1E,$DF,$99,$F7,$11,$FC,$9E,$5F,$E3,$3C,$3F
	.dc.b	$8F,$4E,$5F,$EE,$31,$93,$FC,$71,$E5,$3E,$2B,$FC,$37,$8F,$F9,$45
	.dc.b	$B8,$7F,$92,$65,$97,$ED,$DE,$1E,$07,$F7,$EE,$4F,$F1,$FD,$7F,$6E
	.dc.b	$6E,$5F,$DF,$5A,$FF,$7D,$6B,$D7,$E7,$1F,$71,$FE,$40,$F9,$E7,$C7
	.dc.b	$7F,$90,$A7,$38,$7F,$0E,$C9,$5A,$8F,$E0,$1F,$02,$F8,$4F,$F2,$63
	.dc.b	$63,$F8,$27,$47,$EC,$78,$3E,$BC,$E7,$4F,$E2,$93,$FB,$57,$FC,$92
	.dc.b	$D7,$FC,$CA,$D7,$FC,$D8,$D4,$D2,$76,$FF,$9A,$1E,$19,$3E,$93,$F8
	.dc.b	$8E,$1F,$E7,$F6,$E3,$FB,$2F,$BE,$FD,$FE,$3F,$E3,$29,$CB,$FE,$29
	.dc.b	$8C,$BF,$E2,$76,$BC,$7B,$FF,$82,$46,$23,$84,$60,$95,$FE,$24,$9D
	.dc.b	$60,$FB,$EF,$F0,$06,$3F,$E9,$36,$D7,$FA,$1B,$EF,$BF,$D6,$F8,$FF
	.dc.b	$91,$A7,$2F,$FA,$3E,$32,$FF,$9C,$9A,$F1,$F8,$2F,$F4,$5C,$62,$38
	.dc.b	$46,$09,$5F,$CB,$A7,$58,$3E,$A9,$F5,$9F,$E8,$66,$D7,$F5,$2C,$9F
	.dc.b	$7D,$F8,$5C,$BF,$CB,$75,$A7,$FC,$F1,$3A,$E5,$FE,$D7,$5B,$9F,$F5
	.dc.b	$1C,$E6,$FF,$B5,$CE,$67,$FC,$0A,$D7,$FD,$7C,$D9,$E1,$1F,$DB,$A3
	.dc.b	$58,$BE,$F9,$F4,$FF,$E9,$E6,$E1,$FD,$67,$EF,$9E,$0E,$6B,$FD,$4D
	.dc.b	$18,$9F,$EA,$2D,$C7,$FC,$DA,$D7,$FD,$42,$D7,$FD,$7A,$D7,$FD,$88
	.dc.b	$D4,$D3,$58,$78,$BE,$F3,$FD,$67,$81,$FF,$5D,$E7,$FE,$85,$F8,$3C
	.dc.b	$C6,$3D,$EF,$DE,$E3,$FE,$E1,$6D,$6F,$2F,$F0,$AC,$60,$97,$FE,$1C
	.dc.b	$6C,$7F,$B2,$FE,$23,$FC,$F9,$91,$FF,$3F,$61,$FC,$22,$FF,$D8,$F1
	.dc.b	$8E,$EF,$0C,$E6,$FF,$BB,$C6,$0F,$2D,$BE,$FB,$F9,$D4,$FF,$1F,$6D
	.dc.b	$7F,$B5,$D7,$C4,$FF,$B5,$1C,$D7,$1F,$27,$06,$66,$DF,$F1,$7F,$BE
	.dc.b	$FF,$18,$6B,$FE,$FD,$6C,$7F,$2D,$7F,$7F,$19,$F5,$BF,$DC,$31,$AC
	.dc.b	$BF,$DD,$0D,$C4,$F0,$BE,$55,$FE,$E8,$FB,$AF,$D9,$D1,$FD,$AF,$A2
	.dc.b	$FB,$AF,$B5,$7F,$CA,$E7,$64,$F7,$3F,$D7,$33,$AC,$BF,$D8,$73,$CA
	.dc.b	$31,$1F,$EB,$76,$BF,$E0,$A6,$BF,$EB,$F6,$A2,$7C,$6D,$FB,$B7,$D8
	.dc.b	$7F,$9E,$F1,$C3,$FE,$C8,$9B,$5F,$E8,$89,$AF,$31,$18,$5E,$AA,$71
	.dc.b	$1A,$17,$E0,$7A,$18,$BF,$03,$FE,$3E,$38,$1F,$E4,$0B,$C3,$83,$C5
	.dc.b	$F7,$8F,$BE,$7D,$F7,$F9,$1A,$30,$D8,$CC,$FF,$B9,$27,$2D,$ED,$71
	.dc.b	$F1,$6C,$66,$30,$47,$D0,$7C,$D9,$FE,$4A,$FC,$66,$BA,$0F,$F1,$28
	.dc.b	$D4,$F6,$CB,$E2,$56,$BC,$2F,$9D,$18,$BE,$6D,$AE,$19,$FF,$81,$F7
	.dc.b	$7C,$27,$F9,$9B,$8F,$F2,$4D,$8C,$E5,$F6,$4E,$11,$BF,$C2,$5F,$27
	.dc.b	$FC,$CB,$87,$FC,$25,$18,$AF,$F0,$13,$5F,$CA,$D6,$D6,$09,$D6,$DF
	.dc.b	$83,$9C,$C6,$2F,$AE,$1C,$6F,$CF,$5F,$E9,$CF,$8E,$FB,$7F,$1B,$D5
	.dc.b	$B1,$9E,$24,$F1,$FB,$FF,$93,$9F,$7B,$DD,$FD,$CC,$BF,$45,$18,$9D
	.dc.b	$61,$CA,$BF,$C3,$D7,$CA,$F7,$F2,$0F,$8C,$F1,$36,$BC,$CD,$7B,$1F
	.dc.b	$0B,$EF,$DF,$41,$C2,$A4,$F0,$7E,$04,$EA,$3F,$C5,$DF,$7C,$F8,$5F
	.dc.b	$B4,$D8,$FB,$07,$CE,$70,$FD,$BD,$D1,$88,$FD,$B6,$DF,$B3,$F1,$BE
	.dc.b	$56,$86,$9F,$46,$B9,$EF,$F5,$F0,$7D,$F3,$E5,$7F,$18,$DC,$BC,$73
	.dc.b	$A9,$C9,$23,$33,$B7,$CC,$D7,$F8,$54,$F1,$7D,$B9,$3E,$93,$FD,$75
	.dc.b	$1F,$83,$6B,$F3,$A3,$5E,$B7,$E2,$6F,$B6,$3E,$9F,$BB,$7D,$EF,$E8
	.dc.b	$DB,$87,$A1,$AF,$C9,$6B,$E4,$6B,$FE,$73,$6B,$F9,$E6,$E4,$73,$88
	.dc.b	$7C,$37,$E6,$5B,$5F,$9D,$6B,$F8,$8A,$E5,$7F,$A3,$F5,$5F,$0B,$FB
	.dc.b	$B6,$D7,$D0,$6B,$F8,$16,$BF,$31,$AF,$FA,$2D,$AF,$ED,$9B,$5F,$E3
	.dc.b	$D9,$FF,$76,$DB,$EF,$1F,$55,$FB,$16,$E1,$FB,$56,$BF,$E3,$EA,$FD
	.dc.b	$D6,$7F,$E2,$E8,$F9,$DE,$0F,$D1,$23,$82,$1F,$45,$CB,$FC,$57,$07
	.dc.b	$B8,$FC,$53,$F8,$13,$67,$F8,$33,$5F,$95,$7E,$17,$FC,$F4,$7F,$23
	.dc.b	$F7,$BF,$A6,$69,$FE,$83,$8C,$10,$FC,$DC,$68,$8F,$C4,$FC,$57,$F9
	.dc.b	$4B,$F9,$16,$F0,$FF,$0F,$6F,$47,$FA,$E6,$BF,$97,$4F,$FB,$5B,$EF
	.dc.b	$7F,$9E,$DF,$F3,$16,$BF,$DA,$DA,$FF,$77,$6A,$69,$3C,$6F,$EC,$CF
	.dc.b	$0C,$7F,$87,$F5,$7F,$2B,$EE,$FD,$F5,$FF,$AE,$5A,$FF,$16,$3A,$5F
	.dc.b	$F3,$9F,$3D,$F7,$AF,$A0,$C3,$6B,$FC,$15,$9F,$F3,$BA,$DD,$7F,$39
	.dc.b	$F8,$77,$C4,$FF,$3B,$F8,$B7,$DF,$FF,$53,$6E,$3F,$C2,$1A,$FF,$19
	.dc.b	$6B,$FC,$13,$47,$CA,$7F,$41,$FA,$47,$CD,$53,$DD,$1F,$D4,$69,$F7
	.dc.b	$BF,$CA,$DF,$E4,$ED,$7F,$B0,$35,$FF,$0D,$35,$FF,$13,$35,$E7,$7F
	.dc.b	$E1,$38,$C1,$BE,$AF,$DA,$F9,$BF,$D5,$7B,$7F,$A5,$5F,$F1,$56,$F0
	.dc.b	$FF,$61,$8E,$67,$CA,$E8,$FF,$15,$7D,$EB,$E2,$BF,$AF,$7C,$B3,$E5
	.dc.b	$29,$ED,$9D,$5F,$13,$FC,$78,$E1,$FE,$49,$3C,$5F,$ED,$4D,$7F,$01
	.dc.b	$1B,$C3,$C8,$FF,$BA,$A3,$FB,$87,$CE,$3E,$CF,$FB,$83,$63,$C0,$27
	.dc.b	$83,$D3,$E4,$37,$52,$37,$3A,$BF,$85,$F7,$6F,$BC,$FB,$0D,$AE,$AD
	.dc.b	$7F,$2D,$F7,$BF,$2B,$6F,$F2,$84,$E6,$F9,$DF,$23,$F0,$55,$BB,$FF
	.dc.b	$30,$56,$EF,$FC,$C5,$E0,$3F,$13,$FC,$BD,$7C,$B8,$C7,$5F,$38,$F9
	.dc.b	$DF,$82,$BF,$23,$F9,$5A,$7F,$65,$EC,$7D,$67,$E7,$CF,$F1,$15,$6F
	.dc.b	$92,$DF,$0D,$3C,$37,$C3,$1F,$03,$5F,$F0,$4D,$F1,$7F,$C4,$5F,$87
	.dc.b	$7E,$9B,$2F,$F0,$E3,$6F,$FC,$36,$D9,$FF,$3F,$7E,$00,$FB,$5F,$F1
	.dc.b	$44,$E7,$AC,$D3,$85,$2B,$83,$F7,$9F,$C0,$3F,$E1,$06,$BF,$E4,$96
	.dc.b	$BF,$E8,$36,$BE,$2E,$5F,$F3,$D4,$62,$3B,$46,$0F,$E9,$F2,$FC,$0F
	.dc.b	$CE,$49,$F1,$4F,$EF,$7F,$01,$FF,$7A,$C7,$C4,$C7,$DA,$67,$E5,$A4
	.dc.b	$64,$87,$F1,$F9,$E2,$FF,$8D,$B4,$FF,$6E,$78,$DF,$E5,$9F,$7D,$F2
	.dc.b	$BF,$A5,$DF,$FD,$51,$18,$EF,$FF,$2F,$5F,$0C,$3F,$64,$D7,$FC,$83
	.dc.b	$3B,$7F,$6C,$D7,$C1,$F2,$BF,$CA,$DF,$E6,$5F,$80,$3F,$00,$FF,$50
	.dc.b	$36,$3F,$C7,$E1,$C3,$FE,$42,$6B,$FE,$7B,$6B,$FE,$B5,$6B,$FE,$C1
	.dc.b	$6B,$CE,$3E,$8F,$DB,$7E,$18,$91,$FB,$E9,$EF,$87,$F3,$FF,$CE,$9F
	.dc.b	$69,$FD,$37,$A2,$5F,$F1,$3F,$3D,$5A,$8C,$35,$FF,$0F,$79,$4A,$E3
	.dc.b	$1F,$E6,$8A,$E3,$7F,$E8,$3F,$01,$F8,$9F,$E6,$BE,$56,$5F,$ED,$E7
	.dc.b	$2D,$9C,$BF,$E4,$5E,$A5,$73,$9F,$F4,$77,$E0,$3F,$13,$C3,$34,$9C
	.dc.b	$3F,$EA,$9F,$23,$F4,$3C,$1F,$E6,$36,$F3,$7F,$0A,$6E,$7F,$EA,$0B
	.dc.b	$F8,$F3,$9E,$CF,$D8,$FD,$94,$AF,$E1,$D7,$C3,$23,$CB,$C0,$FD,$AD
	.dc.b	$6E,$BF,$C8,$93,$BF,$A5,$7D,$1F,$FA,$F5,$B3,$FE,$72,$6B,$FE,$83
	.dc.b	$6B,$FE,$82,$E0,$FE,$C9,$AF,$F0,$AE,$5F,$E0,$A7,$FA,$C6,$B8,$C7
	.dc.b	$FA,$07,$F0,$17,$CE,$FF,$B2,$DB,$87,$FA,$01,$AF,$FA,$45,$AF,$7F
	.dc.b	$C1,$FF,$88,$91,$FE,$5D,$1C,$2B,$F8,$DC,$E6,$7E,$C3,$FE,$C3,$9D
	.dc.b	$70,$7F,$1E,$8D,$57,$0A,$E7,$7F,$EC,$9F,$C0,$7F,$BD,$70,$1A,$4C
	.dc.b	$FF,$87,$FC,$17,$F2,$42,$5F,$F4,$19,$F8,$B9,$7F,$46,$FC,$01,$AF
	.dc.b	$FB,$AD,$B5,$FE,$A0,$6B,$FE,$B1,$6B,$FE,$9C,$6B,$CE,$39,$11,$FF
	.dc.b	$4F,$EC,$FA,$AF,$E6,$E4,$8F,$EA,$95,$DB,$5F,$F5,$A7,$DF,$3E,$B3
	.dc.b	$FD,$FC,$DA,$FF,$59,$B5,$FF,$62,$35,$FF,$57,$B5,$FE,$C6,$D7,$FD
	.dc.b	$F2,$DD,$8E,$79,$7F,$A4,$D7,$F4,$C2,$57,$F6,$99,$F0,$5F,$F5,$69
	.dc.b	$CC,$FE,$CD,$FF,$7F,$4E,$B4,$4B,$ED,$1F,$7E,$F0,$F3,$2F,$A1,$7D
	.dc.b	$87,$A3,$7C,$C6,$27,$D3,$18,$9E,$B8,$74,$BC,$57,$7C,$3B,$3E,$2C
	.dc.b	$39,$6D,$F0,$31,$DB,$67,$58,$F4,$BE,$31,$E5,$8F,$23,$A3,$C6,$59
	.dc.b	$E4,$EC,$F1,$61,$E4,$5F,$7D,$FE,$5E,$CB,$94,$63,$E0,$3E,$33,$C4
	.dc.b	$C4,$77,$F2,$5F,$7F,$FD,$37,$5F,$89,$5A,$38,$61,$D5,$B8,$74,$9F
	.dc.b	$95,$7C,$BC,$DF,$25,$EE,$FE,$99,$CC,$AE,$5F,$AE,$D7,$EE,$BF,$E5
	.dc.b	$DF,$A5,$FE,$B6,$D8,$E6,$7C,$0F,$35,$3A,$6C,$6A,$78,$61,$EC,$6E
	.dc.b	$DF,$DD,$BE,$0F,$FC,$D1,$C5,$F2,$27,$37,$E2,$C3,$9E,$1C,$5B,$87
	.dc.b	$1F,$5D,$87,$19,$CD,$B8,$7F,$C7,$F1,$A8,$FB,$51,$89,$FC,$07,$03
	.dc.b	$CF,$FE,$53,$1F,$81,$F6,$3F,$CF,$CE,$A3,$37,$CA,$72,$FE,$16,$B9
	.dc.b	$78,$1D,$F4,$FC,$AD,$69,$FD,$04,$EB,$0E,$D9,$6F,$4F,$D0,$E5,$DA
	.dc.b	$F8,$CE,$72,$75,$7F,$91,$32,$F6,$75,$7F,$93,$9B,$97,$F9,$A1,$7C
	.dc.b	$F6,$2F,$8F,$6F,$F9,$8A,$BF,$CA,$F5,$9C,$3A,$47,$F9,$B6,$38,$BA
	.dc.b	$31,$7F,$EA,$5B,$EC,$D7,$E8,$4F,$07,$F2,$3B,$FF,$9C,$27,$4D,$AF
	.dc.b	$C5,$B6,$77,$5C,$5F,$E1,$71,$97,$FB,$27,$E2,$5E,$7C,$4F,$B6,$5F
	.dc.b	$14,$F9,$BD,$4E,$63,$18,$F8,$1B,$87,$99,$AF,$81,$AF,$FA,$49,$AF
	.dc.b	$E3,$5B,$7E,$0B,$F6,$C6,$B3,$FF,$3C,$FD,$6F,$9C,$F3,$7C,$DF,$F2
	.dc.b	$4C,$7F,$D0,$F1,$8E,$2F,$8B,$39,$CB,$E4,$B5,$F8,$6D,$7F,$D1,$7E
	.dc.b	$67,$C3,$3E,$17,$AE,$F8,$7C,$63,$FE,$94,$9B,$4E,$2F,$D2,$6D,$FE
	.dc.b	$6A,$BC,$C7,$FA,$62,$AD,$FF,$AC,$B0,$E8,$F3,$3F,$D6,$4E,$30,$FB
	.dc.b	$ED,$7E,$CB,$5E,$57,$F8,$7D,$1F,$90,$FE,$EA,$FF,$DC,$D1,$87,$EE
	.dc.b	$7D,$FF,$F0,$6B,$F4,$47,$F5,$29,$D3,$FE,$C3,$9D,$4F,$E2,$1A,$FF
	.dc.b	$B2,$9A,$FF,$05,$6E,$6F,$BE,$7D,$0F,$E8,$F1,$7E,$1D,$F8,$D9,$CE
	.dc.b	$3E,$69,$CB,$FE,$D3,$8C,$E3,$E6,$3F,$6F,$3A,$71,$87,$7C,$3B,$61
	.dc.b	$D5,$98,$7F,$94,$3E,$9A,$DF,$F9,$31,$FD,$4F,$A6,$69,$27,$F9,$AE
	.dc.b	$9F,$33,$A7,$D5,$D5,$EB,$6C,$FA,$5B,$87,$9E,$B8,$4F,$B6,$BF,$84
	.dc.b	$7B,$6F,$CA,$7F,$C0,$33,$FE,$F1,$FB,$9F,$F3,$FB,$FE,$8B,$8C,$E1
	.dc.b	$EE,$9C,$BF,$E8,$78,$CF,$2F,$E4,$27,$2F,$F3,$5B,$ED,$5F,$BB,$9C
	.dc.b	$DF,$DF,$E0,$79,$1F,$54,$FE,$39,$E3,$3E,$F9,$F2,$19,$BF,$F0,$BE
	.dc.b	$7F,$D0,$6B,$53,$FC,$5D,$B3,$FD,$71,$B9,$7F,$68,$6B,$FE,$1C,$6B
	.dc.b	$FE,$22,$6B,$FD,$8A,$FC,$33,$CD,$B1,$FD,$CB,$37,$D6,$3E,$87,$DD
	.dc.b	$F8,$17,$CE,$8C,$61,$F3,$63,$13,$A8,$FF,$5C,$4E,$AF,$E7,$57,$C2
	.dc.b	$C3,$FC,$69,$5F,$95,$F2,$71,$E4,$7E,$6D,$B1,$93,$E2,$9E,$B9,$E3
	.dc.b	$7F,$83,$DF,$F6,$1B,$7F,$69,$E3,$FD,$BF,$3F,$E9,$4F,$A3,$F7,$8F
	.dc.b	$96,$C6,$5D,$31,$FF,$02,$CE,$63,$E7,$36,$3E,$73,$5F,$9C,$D7,$E7
	.dc.b	$35,$E7,$3F,$66,$31,$C2,$75,$87,$F0,$CF,$A2,$FD,$D1,$E6,$FA,$6B
	.dc.b	$8F,$EE,$62,$E3,$FB,$7C,$39,$56,$33,$FF,$04,$7C,$7B,$F2,$27,$BF
	.dc.b	$FD,$3F,$F3,$4F,$8B,$9F,$F9,$92,$B5,$39,$E2,$FE,$E1,$18,$9F,$89
	.dc.b	$E0,$AF,$8E,$8A,$D3,$FC,$1A,$33,$A7,$B7,$C0,$FE,$29,$F2,$9F,$41
	.dc.b	$B8,$FC,$F9,$D3,$78,$7F,$DA,$EF,$D3,$C1,$F3,$2F,$BF,$E5,$3B,$C2
	.dc.b	$D7,$F1,$93,$73,$8E,$0F,$D5,$60,$FB,$7F,$EE,$5E,$27,$E0,$9B,$8F
	.dc.b	$85,$B8,$7C,$DA,$E1,$3F,$CF,$BD,$1F,$E3,$4F,$9F,$E4,$BF,$C2,$5F
	.dc.b	$37,$FE,$AD,$8C,$1F,$D1,$1C,$5F,$39,$DF,$BF,$F9,$26,$D7,$F4,$AD
	.dc.b	$BF,$A6,$57,$F1,$D6,$E9,$EA,$FA,$9F,$F1,$04,$62,$BF,$1F,$1F,$8D
	.dc.b	$23,$F9,$86,$D7,$E6,$1A,$FE,$61,$FF,$0E,$F8,$DF,$DE,$6F,$F5,$1F
	.dc.b	$8A,$FE,$DD,$FE,$41,$E9,$7F,$8B,$E3,$18,$7F,$AB,$39,$DF,$24,$F8
	.dc.b	$CF,$F1,$3B,$67,$FC,$4E,$DC,$FF,$C5,$AD,$7F,$D6,$3E,$43,$E6,$BF
	.dc.b	$C4,$75,$FB,$38,$C7,$AF,$FE,$27,$FA,$07,$C7,$7F,$3E,$BD,$E1,$FC
	.dc.b	$E7,$CD,$FE,$64,$7F,$92,$5F,$F7,$07,$C4,$63,$2F,$C6,$72,$FF,$B7
	.dc.b	$5F,$F4,$2C,$E9,$B1,$98,$FF,$44,$B5,$FF,$4A,$B5,$FF,$64,$B5,$FF
	.dc.b	$67,$B5,$E8,$7F,$A7,$3C,$67,$D2,$BD,$9F,$F2,$D3,$6F,$F5,$FF,$4B
	.dc.b	$FE,$3E,$E5,$FE,$35,$39,$C3,$FB,$9C,$FF,$8E,$89,$EE,$FF,$95,$9B
	.dc.b	$7F,$E5,$66,$CF,$F9,$71,$FF,$1F,$7C,$3B,$F1,$1F,$F3,$53,$6B,$F7
	.dc.b	$FF,$80,$3E,$5B,$35,$FE,$57,$8C,$4F,$F9,$89,$B8,$7F,$9A,$1A,$FF
	.dc.b	$98,$9A,$FE,$1D,$AF,$FA,$7D,$BD,$06,$9A,$D9,$F6,$BF,$C0,$6F,$FC
	.dc.b	$5E,$DB,$FF,$0A,$FD,$6B,$ED,$BF,$D0,$93,$9E,$0F,$F2,$3C,$E6,$FF
	.dc.b	$CF,$2D,$BF,$F3,$FB,$5F,$F3,$CB,$5F,$CF,$35,$FF,$5B,$37,$23,$4D
	.dc.b	$63,$F3,$7F,$14,$FF,$4B,$68,$7F,$D0,$1F,$7E,$FA,$FC,$C6,$1F,$F4
	.dc.b	$7C,$67,$2F,$F4,$7B,$72,$FF,$4A,$B5,$FF,$47,$B5,$FD,$A3,$5F,$F6
	.dc.b	$63,$72,$39,$C7,$FD,$C7,$F3,$7F,$EA,$BF,$34,$FF,$A5,$7C,$67,$C4
	.dc.b	$66,$30,$FF,$A9,$23,$39,$7F,$A9,$1B,$97,$FA,$A9,$AF,$FA,$91,$AF
	.dc.b	$F0,$96,$BF,$EE,$16,$E4,$73,$8B,$E2,$DF,$47,$FE,$C1,$6E,$1F,$D5
	.dc.b	$BE,$37,$FA,$AD,$6E,$BF,$D5,$D1,$FE,$68,$27,$AB,$FE,$BC,$7F,$CD
	.dc.b	$B3,$FD,$6F,$4F,$F6,$67,$AB,$F0,$B0,$FE,$C3,$F7,$BF,$D0,$B5,$FF
	.dc.b	$28,$4E,$5B,$5C,$0F,$F5,$E4,$EF,$97,$FC,$25,$5B,$7F,$C2,$9E,$37
	.dc.b	$E5,$70,$FF,$42,$37,$06,$EF,$BD,$7D,$EF,$FB,$7E,$B5,$A1,$A7,$1A
	.dc.b	$E1,$3B,$F5,$7F,$D9,$DF,$74,$F7,$29,$FE,$D1,$FB,$BF,$DC,$EB,$FE
	.dc.b	$AC,$9D,$6B,$FE,$00,$6E,$BF,$E8,$26,$CF,$FA,$CA,$B8,$D7,$F4,$39
	.dc.b	$CB,$FD,$16,$3F,$7A,$4F,$03,$FD,$D6,$FF,$02,$B1,$39,$6F,$73,$EC
	.dc.b	$7F,$CB,$98,$61,$DD,$4A,$C3,$FE,$5F,$9A,$E4,$FE,$13,$3C,$9A,$72
	.dc.b	$3F,$18,$60,$FF,$33,$47,$27,$C8,$7F,$46,$36,$7F,$48,$9C,$38,$3C
	.dc.b	$BF,$0C,$FB,$C7,$DE,$7F,$4E,$8C,$36,$33,$3E,$33,$FC,$DE,$E1,$F3
	.dc.b	$4F,$F0,$7D,$79,$77,$FF,$B8,$E7,$84,$E5,$B1,$97,$8B,$BC,$FE,$64
	.dc.b	$7D,$09,$FE,$50,$FC,$86,$BF,$0D,$E9,$1C,$23,$FD,$07,$3C,$F3,$F7
	.dc.b	$F8,$7F,$52,$B5,$7C,$9B,$5B,$BF,$CE,$EB,$FE,$FF,$7E,$07,$1A,$31
	.dc.b	$1F,$56,$31,$F2,$3F,$DD,$9B,$1C,$3F,$12,$F2,$68,$E0,$FF,$0A,$46
	.dc.b	$1B,$AF,$E2,$27,$F3,$18,$59,$CD,$FF,$97,$AF,$83,$FB,$26,$E1,$E0
	.dc.b	$8F,$F6,$86,$FF,$ED,$B7,$FE,$00,$B5,$CF,$BD,$D5,$74,$7E,$0B,$CA
	.dc.b	$C3,$F8,$49,$B5,$C7,$3F,$F4,$C3,$72,$FD,$2D,$71,$C3,$FD,$2C,$BE
	.dc.b	$6D,$F0,$BD,$1B,$3A,$6B,$FE,$FD,$5E,$0F,$F2,$D8,$D4,$F1,$23,$FE
	.dc.b	$A3,$9D,$F8,$7E,$8F,$C6,$DE,$87,$F8,$A3,$14,$62,$FF,$DF,$6D,$7F
	.dc.b	$C4,$0D,$7F,$73,$0F,$C5,$DF,$19,$F1,$1B,$3E,$E6,$BF,$EF,$FA,$E9
	.dc.b	$B7,$DF,$8F,$E4,$FF,$76,$FA,$3F,$C3,$36,$3F,$1A,$D7,$F1,$8D,$7F
	.dc.b	$85,$F1,$A3,$11,$FE,$5D,$6B,$DA,$7A,$9F,$07,$07,$BD,$E9,$24,$FF
	.dc.b	$AA,$2B,$C1,$87,$FB,$4F,$0F,$C3,$4F,$C2,$6B,$F7,$9A,$FE,$21,$B9
	.dc.b	$FF,$44,$E3,$FA,$11,$82,$3F,$4E,$3F,$C5,$53,$F8,$66,$AA,$FF,$72
	.dc.b	$46,$1F,$24,$FE,$5E,$9F,$87,$BD,$7F,$D8,$37,$0F,$DB,$B5,$FD,$B3
	.dc.b	$5F,$96,$D7,$FD,$0C,$D7,$FA,$EB,$67,$BC,$FA,$7E,$77,$F4,$B8,$9A
	.dc.b	$79,$6B,$C1,$8F,$FB,$F7,$B5,$F2,$3F,$93,$6E,$5F,$CA,$F0,$FF,$24
	.dc.b	$CE,$49,$19,$9D,$BF,$E7,$56,$BF,$96,$6B,$FB,$2F,$BB,$FC,$8F,$45
	.dc.b	$E4,$FD,$DC,$62,$7F,$7F,$5D,$31,$FC,$B6,$AF,$D1,$DF,$80,$7F,$1B
	.dc.b	$6D,$FF,$22,$6B,$FC,$85,$AF,$E5,$9A,$FF,$AC,$5A,$FF,$82,$1B,$3D
	.dc.b	$E7,$F6,$D8,$7F,$8E,$AB,$F8,$4B,$5F,$C2,$65,$ED,$F4,$BF,$05,$87
	.dc.b	$8B,$1F,$D7,$74,$BD,$1F,$EB,$98,$7F,$57,$8C,$78,$DF,$EA,$A6,$C6
	.dc.b	$67,$6F,$F6,$F8,$CB,$63,$2F,$F7,$C8,$CB,$63,$2F,$F2,$2F,$C5,$FF
	.dc.b	$DF,$4F,$F2,$26,$BE,$A3,$F9,$7B,$8B,$F4,$CF,$89,$FA,$7F,$03,$FB
	.dc.b	$8A,$FF,$08,$C6,$3A,$3F,$45,$39,$9F,$E5,$1F,$8E,$71,$D7,$FC,$2F
	.dc.b	$E8,$3E,$63,$FD,$55,$5F,$DE,$63,$1E,$D7,$FB,$31,$B1,$99,$DB,$FE
	.dc.b	$E2,$E0,$FE,$87,$F4,$9F,$CD,$23,$F9,$EB,$5F,$67,$27,$FB,$BA,$7F
	.dc.b	$A6,$43,$EE,$9F,$2F,$FE,$15,$6C,$7F,$B7,$1A,$FF,$83,$5A,$FF,$88
	.dc.b	$5A,$FF,$8B,$1A,$FF,$8D,$1A,$F7,$8F,$E8,$1F,$09,$FD,$56,$3F,$AE
	.dc.b	$B7,$AC,$E6,$38,$1F,$D6,$3E,$37,$C7,$C5,$FC,$E6,$91,$8D,$FF,$AF
	.dc.b	$B6,$BF,$B4,$BF,$D6,$AB,$83,$9A,$E3,$87,$F6,$C8,$C1,$BB,$FE,$D3
	.dc.b	$F3,$BF,$D1,$F0,$FC,$35,$A1,$F3,$DF,$DD,$1B,$3F,$DB,$9E,$18,$F9
	.dc.b	$DB,$7D,$2B,$81,$C5,$FA,$BF,$2A,$FA,$6F,$F2,$43,$70,$FF,$1F,$07
	.dc.b	$0F,$F9,$05,$AF,$F9,$55,$AF,$F9,$81,$AF,$F9,$A1,$AF,$38,$EB,$EE
	.dc.b	$BE,$E3,$F1,$59,$7F,$12,$F8,$E7,$D3,$7F,$3F,$6E,$1F,$A2,$CB,$FC
	.dc.b	$E7,$87,$1F,$43,$FA,$27,$D9,$3D,$FA,$4B,$E0,$7F,$A6,$21,$F7,$59
	.dc.b	$8F,$CC,$C7,$17,$7C,$E8,$C4,$6A,$7F,$C4,$BF,$3E,$F9,$6B,$F1,$92
	.dc.b	$BF,$C4,$33,$96,$CE,$5F,$BB,$1F,$D7,$74,$FF,$14,$BB,$8F,$E0,$C3
	.dc.b	$ED,$FF,$3A,$4E,$8F,$F3,$F3,$63,$EB,$B5,$FF,$3B,$3F,$A6,$E7,$FC
	.dc.b	$FF,$DE,$7F,$8F,$3A,$8D,$3F,$6F,$3F,$9B,$D9,$F3,$8F,$8E,$FC,$86
	.dc.b	$53,$2F,$8B,$39,$6C,$E5,$FC,$CF,$5F,$39,$D4,$7C,$0F,$67,$E4,$91
	.dc.b	$F9,$FF,$BB,$EB,$A7,$F6,$1B,$FF,$90,$67,$81,$F3,$2B,$FC,$05,$5B
	.dc.b	$E0,$FF,$57,$FD,$CB,$E5,$FF,$D4,$4D,$BF,$CD,$B5,$FF,$4D,$35,$FF
	.dc.b	$57,$35,$E7,$1C,$88,$FF,$77,$E1,$FF,$7E,$BF,$E5,$E6,$BF,$B8,$6E
	.dc.b	$5F,$BB,$AF,$03,$FD,$BB,$9B,$FB,$86,$1F,$EC,$CE,$0F,$C7,$7A,$7F
	.dc.b	$9A,$3F,$C3,$DF,$B2,$3E,$D7,$19,$FE,$53,$0E,$2F,$E6,$EB,$73,$F9
	.dc.b	$4E,$D9,$3F,$92,$7F,$C5,$1F,$BE,$7E,$BB,$2F,$C7,$C6,$A7,$F2,$6E
	.dc.b	$BB,$7F,$3B,$C7,$FE,$69,$E3,$FE,$B3,$5F,$E6,$46,$BF,$98,$8D,$7B
	.dc.b	$1F,$A4,$24,$7E,$8A,$2B,$FA,$0F,$A2,$FD,$49,$8E,$0F,$BF,$58,$9F
	.dc.b	$D4,$CE,$72,$FD,$2F,$AD,$E3,$24,$FF,$8A,$71,$FF,$06,$F6,$17,$C7
	.dc.b	$7E,$D8,$93,$FB,$4A,$FD,$CF,$C7,$FB,$49,$1C,$DF,$81,$F4,$DF,$BE
	.dc.b	$24,$7E,$F2,$2B,$FB,$A8,$BA,$7E,$62,$6E,$5F,$BC,$9B,$5F,$EA,$EC
	.dc.b	$55,$FC,$2B,$E7,$7E,$71,$27,$D8,$FF,$A6,$7E,$27,$FD,$1B,$8F,$F5
	.dc.b	$36,$E3,$FE,$B0,$6D,$75,$AF,$F5,$CC,$7F,$7A,$8F,$F4,$DB,$5E,$1E
	.dc.b	$97,$52,$3F,$E5,$38,$D3,$FE,$C4,$6E,$A6,$9F,$DA,$A3,$FD,$7E,$D7
	.dc.b	$FA,$CB,$70,$F4,$57,$F9,$E7,$C8,$FF,$5F,$D7,$FA,$71,$AF,$E8,$DA
	.dc.b	$BA,$7F,$A1,$FC,$37,$FB,$B6,$73,$18,$9F,$F7,$AC,$60,$E3,$8F,$9F
	.dc.b	$D6,$F8,$64,$9F,$E7,$B3,$FD,$C5,$F7,$46,$5F,$4B,$E7,$BF,$47,$A6
	.dc.b	$99,$AF,$A6,$31,$1F,$DE,$71,$FE,$B3,$18,$9D,$4F,$03,$DD,$19,$69
	.dc.b	$93,$E0,$61,$FE,$39,$71,$5D,$1B,$19,$5F,$1F,$FE,$47,$CB,$EB,$7C
	.dc.b	$EF,$92,$31,$B2,$31,$A7,$82,$F8,$BE,$AF,$B0,$FF,$24,$7C,$4F,$BF
	.dc.b	$0F,$F2,$74,$62,$B6,$7C,$37,$11,$FE,$57,$F7,$5F,$37,$FE,$08,$8E
	.dc.b	$3B,$7F,$BA,$1B,$87,$5C,$7C,$13,$97,$E7,$5F,$2F,$4F,$CF,$97,$7A
	.dc.b	$FF,$77,$BC,$2F,$FC,$2F,$97,$96,$3E,$98,$70,$FB,$7E,$AD,$F3,$5C
	.dc.b	$1B,$1C,$0E,$0E,$1E,$72,$8C,$47,$39,$F5,$3F,$0E,$39,$E7,$F0,$FD
	.dc.b	$B7,$80,$E4,$DD,$F6,$C2,$E3,$D5,$B8,$F2,$DB,$E1,$E3,$CB,$0C,$D7
	.dc.b	$83,$91,$FB,$F3,$E5,$7B,$67,$F5,$B9,$7F,$71,$EB,$63,$7F,$F4,$66
	.dc.b	$1C,$2F,$1A,$7F,$0B,$AB,$87,$EC,$71,$E9,$9F,$E4,$9B,$1F,$9F,$6C
	.dc.b	$7E,$8B,$0E,$8E,$F0,$F1,$B9,$9D,$D7,$F6,$16,$BD,$E7,$6F,$F1,$DF
	.dc.b	$5B,$D7,$D7,$FC,$85,$B7,$C5,$B3,$FC,$5D,$BA,$7F,$17,$6F,$47,$F9
	.dc.b	$02,$BF,$C7,$B8,$DB,$FE,$2F,$5C,$B2,$F9,$6F,$F9,$CE,$3B,$57,$E1
	.dc.b	$67,$2D,$AE,$53,$ED,$AE,$0F,$C9,$6D,$70,$8F,$CA,$B5,$FE,$6B,$97
	.dc.b	$98,$F3,$FC,$A5,$C3,$F9,$6B,$D3,$2F,$E6,$4D,$C8,$D2,$4F,$F3,$B6
	.dc.b	$DF,$E9,$2B,$8C,$FF,$3F,$F8,$4F,$8A,$7C,$6F,$C6,$2C,$7F,$3C,$7B
	.dc.b	$11,$FE,$8C,$F8,$7C,$BF,$9A,$7C,$EF,$BB,$B3,$FD,$12,$2E,$1F,$D1
	.dc.b	$26,$EF,$FD,$6E,$6D,$7F,$5B,$9B,$5F,$D4,$5F,$71,$23,$FA,$B3,$F0
	.dc.b	$7F,$0A,$FB,$A7,$DE,$F9,$DB,$3F,$D5,$1F,$8C,$47,$FA,$F3,$F2,$7E
	.dc.b	$F7,$F0,$BF,$5F,$F8,$66,$CF,$F5,$C6,$E1,$FD,$71,$B9,$7F,$73,$78
	.dc.b	$4F,$F6,$2F,$56,$D7,$F6,$27,$EB,$12,$3F,$B7,$3F,$67,$BF,$F2,$A7
	.dc.b	$E4,$7D,$07,$95,$8A,$FE,$DB,$36,$FF,$DD,$F5,$A9,$C9,$FD,$CB,$E5
	.dc.b	$3F,$3E,$7E,$6F,$EE,$FF,$8C,$46,$49,$5F,$DD,$A7,$5D,$FC,$27,$FC
	.dc.b	$B9,$F2,$5F,$AA,$F6,$0F,$EB,$BE,$55,$F2,$EF,$73,$53,$98,$C5,$FF
	.dc.b	$2B,$AD,$D7,$E9,$1B,$19,$9D,$E5,$FC,$E2,$72,$D8,$CB,$FD,$BA,$32
	.dc.b	$D8,$CB,$FE,$23,$FB,$5F,$DF,$9F,$BB,$F7,$7F,$D9,$77,$FE,$16,$8C
	.dc.b	$B4,$FF,$48,$F9,$D8,$A7,$27,$F2,$AF,$03,$FD,$47,$F2,$9F,$C3,$3F
	.dc.b	$1F,$FF,$74,$61,$FE,$A2,$8C,$72,$7F,$09,$F6,$91,$89,$D1,$FE,$D2
	.dc.b	$F0,$17,$CB,$2F,$F4,$46,$7F,$3C,$F9,$5F,$15,$6F,$FD,$B3,$C1,$F4
	.dc.b	$23,$E5,$70,$1A,$49,$F9,$8D,$BF,$94,$D7,$FC,$2F,$8F,$F6,$57,$F9
	.dc.b	$3F,$99,$FD,$92,$BC,$CF,$F8,$5F,$7F,$E9,$75,$FD,$AF,$CB,$DD,$F0
	.dc.b	$6D,$F0,$79,$FC,$14,$8E,$7E,$35,$AF,$9B,$1F,$E2,$BB,$FE,$ED,$59
	.dc.b	$C3,$BE,$3F,$E4,$FA,$F1,$CF,$F3,$EF,$88,$7D,$8B,$EC,$7F,$9C,$7D
	.dc.b	$8F,$F6,$5C,$FA,$8A,$DC,$EB,$D2,$FE,$8F,$F2,$2F,$C4,$3F,$C4,$99
	.dc.b	$FF,$80,$63,$1E,$CF,$F3,$A6,$C6,$67,$78,$FF,$83,$A7,$2D,$8C,$BF
	.dc.b	$E4,$48,$CB,$63,$2F,$C1,$FB,$1F,$F7,$75,$7E,$15,$F8,$11,$EA,$E0
	.dc.b	$46,$A3,$0E,$B8,$3E,$AB,$96,$F0,$FF,$62,$8C,$5F,$78,$FE,$CB,$F6
	.dc.b	$3F,$26,$3F,$28,$DB,$F0,$2F,$13,$7B,$4A,$DB,$CA,$31,$5F,$EE,$1F
	.dc.b	$AE,$FE,$10,$7F,$07,$7F,$18,$66,$BF,$36,$E0,$FD,$E1,$4F,$F3,$77
	.dc.b	$47,$F9,$BF,$9B,$F8,$44,$64,$46,$67,$F8,$74,$66,$30,$FF,$0B,$38
	.dc.b	$77,$39,$3F,$91,$C7,$CD,$FB,$07,$95,$F8,$AF,$1B,$FC,$13,$18,$C7
	.dc.b	$FD,$01,$C3,$FE,$2C,$E4,$7E,$8A,$F8,$6F,$C3,$36,$BF,$14,$DC,$BF
	.dc.b	$0C,$D7,$FC,$4D,$F3,$3F,$C0,$70,$FE,$1C,$DA,$FF,$30,$FE,$17,$FC
	.dc.b	$92,$73,$E2,$FF,$87,$3B,$9E,$D5,$3F,$C4,$3F,$00,$F9,$47,$CA,$67
	.dc.b	$2F,$49,$FE,$6C,$8C,$BF,$E5,$F9,$E1,$3F,$E1,$86,$C6,$67,$7A,$7F
	.dc.b	$9B,$27,$2D,$8C,$BF,$EA,$28,$CB,$63,$2F,$EF,$FC,$4F,$F7,$AD,$7F
	.dc.b	$23,$6C,$7E,$E3,$95,$EB,$3D,$7F,$F1,$DD,$6B,$6F,$E6,$93,$9C,$3F
	.dc.b	$C6,$8D,$C3,$FC,$5A,$D7,$FC,$A8,$D7,$FD,$75,$87,$6C,$3F,$C6,$B3
	.dc.b	$99,$FF,$68,$36,$75,$5F,$E8,$5F,$03,$FD,$3B,$F0,$3F,$E4,$EF,$83
	.dc.b	$3F,$E5,$0F,$9E,$7C,$AE,$63,$0F,$F8,$E6,$33,$7C,$DB,$97,$F8,$D5
	.dc.b	$AF,$F3,$66,$BF,$C9,$DA,$B0,$46,$1B,$5F,$E4,$7F,$C7,$BF,$19,$5B
	.dc.b	$6E,$4F,$B3,$FC,$76,$3F,$CC,$E7,$31,$FD,$B2,$BF,$DE,$4D,$AE,$1E
	.dc.b	$A7,$F9,$72,$30,$47,$04,$BF,$91,$F1,$3F,$C0,$B8,$3F,$A5,$B6,$3F
	.dc.b	$A1,$5B,$EE,$FF,$CF,$33,$9C,$7D,$D3,$9A,$FA,$4D,$8F,$F2,$E3,$5F
	.dc.b	$EB,$AD,$7A,$D7,$F9,$A6,$31,$1D,$23,$07,$F9,$B7,$E1,$FF,$96,$57
	.dc.b	$F5,$56,$CA,$9F,$AA,$FF,$AE,$EB,$FB,$E5,$6A,$73,$1F,$E8,$09,$D5
	.dc.b	$7F,$A0,$1A,$FF,$A9,$DA,$FE,$51,$AF,$76,$CF,$49,$FF,$49,$46,$B6
	.dc.b	$FF,$80,$3B,$7F,$A8,$30,$FC,$5E,$7F,$DE,$2D,$F1,$79,$AF,$F5,$2C
	.dc.b	$60,$FC,$E1,$72,$FF,$43,$35,$FF,$03,$B5,$FE,$F1,$E8,$3E,$0B,$FD
	.dc.b	$5E,$D9,$FE,$1F,$2E,$97,$E7,$2F,$B2,$DE,$12,$FF,$C4,$98,$71,$C3
	.dc.b	$FB,$CE,$1C,$2F,$51,$FD,$8F,$C4,$7C,$5F,$FA,$F9,$B8,$7F,$8B,$5F
	.dc.b	$EE,$BA,$FF,$86,$1C,$D7,$1A,$E0,$71,$F4,$ED,$FA,$38,$9F,$56,$FA
	.dc.b	$1F,$F6,$23,$67,$5F,$5A,$F9,$8C,$D4,$9E,$55,$BC,$BF,$8E,$5E,$E7
	.dc.b	$35,$FE,$CF,$8C,$4F,$FB,$11,$AF,$FB,$AD,$AF,$F1,$86,$A2,$0E,$4C
	.dc.b	$7D,$3B,$EB,$73,$3F,$DC,$A3,$0E,$27,$73,$AF,$49,$8E,$2B,$42,$F8
	.dc.b	$BF,$DB,$9A,$F8,$EF,$83,$9B,$F0,$BE,$52,$9B,$29,$93,$31,$AF,$AE
	.dc.b	$7C,$77,$FA,$42,$79,$C6,$3D,$FD,$4E,$67,$DF,$39,$77,$F1,$FF,$E7
	.dc.b	$C6,$F2,$75,$8C,$47,$06,$BF,$32,$7F,$DD,$B3,$EC,$5F,$21,$F4,$33
	.dc.b	$F8,$71,$8A,$D5,$F7,$9C,$B7,$3E,$97,$C5,$B5,$C3,$0F,$A1,$C4,$DD
	.dc.b	$EE,$7C,$88,$C3,$63,$35,$F3,$A7,$2D,$E2,$FA,$F7,$FE,$9B,$6F,$07
	.dc.b	$F6,$26,$CF,$F1,$DB,$E6,$FF,$73,$8C,$DF,$E9,$31,$FC,$7E,$1C,$DB
	.dc.b	$87,$E5,$71,$B8,$7F,$55,$DD,$F5,$78,$EF,$FF,$64,$B6,$33,$3F,$EA
	.dc.b	$55,$F5,$D9,$99,$39,$78,$47,$3E,$BF,$E3,$B3,$FE,$AD,$6B,$FE,$69
	.dc.b	$6E,$7F,$9C,$8F,$E2,$9F,$03,$FE,$D9,$78,$78,$3C,$C8,$FF,$1A,$A4
	.dc.b	$EE,$3F,$CE,$9F,$5D,$FE,$BE,$7F,$C0,$8D,$7F,$5F,$9A,$31,$1F,$EB
	.dc.b	$06,$BF,$EA,$1F,$73,$FC,$12,$FF,$B1,$5A,$FF,$A0,$5B,$8F,$FA,$CA
	.dc.b	$7B,$3F,$ED,$5F,$AE,$7C,$77,$E1,$5B,$7F,$83,$6B,$FC,$0B,$89,$18
	.dc.b	$8F,$E1,$31,$89,$E4,$48,$E4,$FE,$E3,$1F,$F3,$8F,$7F,$DB,$CB,$F9
	.dc.b	$9D,$7A,$63,$F0,$37,$CE,$3F,$DC,$0F,$DE,$7F,$BB,$EF,$CE,$31,$1F
	.dc.b	$DE,$23,$04,$7F,$BF,$B5,$F9,$B7,$FE,$5F,$7F,$DD,$8D,$7F,$D5,$0D
	.dc.b	$CB,$FD,$B3,$5E,$17,$FD,$F9,$F5,$CF,$8E,$FD,$63,$6F,$F5,$4D,$7E
	.dc.b	$43,$5F,$8E,$D7,$FC,$96,$D7,$FC,$0C,$DA,$39,$BF,$F1,$C7,$CF,$FE
	.dc.b	$8D,$34,$F9,$31,$E1,$C3,$FD,$E3,$D7,$FE,$63,$C3,$FC,$A4,$D7,$F1
	.dc.b	$ED,$7F,$DF,$AD,$7F,$DD,$CD,$7F,$C9,$0D,$BF,$E7,$9E,$9F,$DD,$8C
	.dc.b	$7A,$1F,$8E,$E8,$FF,$28,$7C,$67,$EE,$63,$18,$72,$BF,$BF,$F5,$CF
	.dc.b	$9F,$FE,$2C,$D9,$FE,$28,$D7,$F2,$4D,$7F,$22,$D7,$FC,$EA,$D7,$FC
	.dc.b	$60,$DD,$8E,$73,$FF,$30,$F8,$BE,$52,$57,$F7,$B9,$F0,$6B,$FC,$07
	.dc.b	$A1,$F5,$BF,$D3,$2B,$FD,$44,$D7,$FD,$3C,$D7,$FD,$38,$D7,$FB,$84
	.dc.b	$72,$F1,$3F,$AD,$B6,$39,$C6,$27,$84,$7F,$9D,$FE,$B9,$F2,$3F,$C8
	.dc.b	$DB,$1F,$E2,$8C,$D1,$88,$FF,$11,$D7,$97,$1F,$F6,$3F,$3F,$EB,$BE
	.dc.b	$07,$F6,$35,$FE,$A9,$6C,$FB,$FE,$BB,$87,$27,$F8,$1F,$CD,$7C,$FB
	.dc.b	$D6,$FE,$E7,$18,$6C,$67,$1F,$EF,$33,$97,$FA,$77,$D2,$7F,$89,$5F
	.dc.b	$47,$D0,$7F,$26,$6F,$87,$FC,$0A,$75,$0C,$FF,$99,$2F,$AE,$7D,$57
	.dc.b	$F8,$09,$B1,$FE,$06,$6B,$FE,$0A,$6B,$FE,$1C,$6B,$FE,$22,$6B,$FE
	.dc.b	$25,$6B,$DE,$3F,$8D,$7C,$D7,$F0,$3D,$7F,$C5,$5C,$68,$C5,$7F,$AB
	.dc.b	$5A,$F1,$3E,$B7,$D7,$79,$31,$FE,$99,$48,$C5,$F4,$6D,$7F,$AD,$DF
	.dc.b	$EA,$71,$9C,$3F,$B2,$46,$AB,$94,$63,$C8,$FE,$D1,$F5,$7F,$23,$1E
	.dc.b	$EB,$E1,$32,$6B,$19,$97,$E2,$E7,$4D,$9C,$BF,$4E,$7F,$C9,$3F,$5C
	.dc.b	$F8,$AF,$F1,$9B,$67,$FC,$6E,$D7,$FC,$76,$D7,$FC,$9A,$D7,$9C,$72
	.dc.b	$23,$FE,$42,$F6,$DF,$5D,$F8,$78,$FF,$80,$1C,$2F,$9B,$FC,$41,$FE
	.dc.b	$0A,$AD,$F1,$FF,$4C,$FA,$B7,$C6,$52,$61,$FE,$79,$87,$D3,$62,$3D
	.dc.b	$C7,$FA,$E6,$38,$BB,$9E,$93,$FE,$15,$FA,$A7,$C7,$5F,$8E,$97,$FB
	.dc.b	$59,$CB,$67,$2F,$6E,$64,$FF,$51,$E1,$FC,$6B,$C2,$7C,$63,$EB,$BF
	.dc.b	$2A,$4E,$7F,$F3,$1B,$63,$FC,$CE,$D7,$FC,$D6,$FE,$7A,$FF,$31,$F0
	.dc.b	$FF,$E2,$DE,$A3,$4F,$F1,$9D,$77,$BF,$EE,$BE,$2B,$E4,$3F,$1B,$12
	.dc.b	$7F,$D4,$53,$96,$CE,$5F,$77,$4F,$E7,$3C,$BF,$E3,$73,$F2,$7E,$2E
	.dc.b	$57,$C7,$7F,$C0,$7F,$45,$CE,$3E,$36,$DF,$E2,$BE,$DF,$D4,$B7,$83
	.dc.b	$E3,$DF,$0C,$3D,$3F,$5C,$FA,$6F,$F4,$3B,$71,$FF,$45,$B5,$FF,$47
	.dc.b	$B5,$FF,$50,$B5,$FF,$53,$B5,$FF,$55,$35,$39,$E1,$FE,$09,$F1,$5F
	.dc.b	$34,$93,$FE,$DE,$37,$5F,$95,$F4,$A4,$7E,$4D,$FF,$11,$FD,$77,$FB
	.dc.b	$1E,$30,$43,$F2,$BB,$BA,$BF,$BE,$6F,$FE,$57,$85,$71,$AF,$93,$DD
	.dc.b	$F9,$72,$3F,$ED,$9F,$A6,$FD,$65,$FC,$C9,$CC,$F0,$35,$DB,$F9,$FE
	.dc.b	$5F,$F1,$E7,$1F,$F4,$F9,$FF,$65,$EF,$F9,$A9,$C9,$AF,$67,$F4,$C4
	.dc.b	$9F,$D2,$45,$7E,$EF,$CC,$7E,$AC,$C6,$5F,$7A,$B1,$3F,$AB,$9C,$F2
	.dc.b	$FE,$9B,$D6,$F1,$92,$7F,$D9,$38,$FE,$B2,$3C,$FF,$5D,$FB,$72,$4F
	.dc.b	$ED,$63,$F7,$5F,$5D,$EC,$24,$72,$D1,$1F,$03,$EA,$7F,$80,$12,$7F
	.dc.b	$7D,$15,$FC,$BC,$5D,$3F,$2F,$37,$2F,$DE,$CD,$A6,$5F,$C0,$AB,$F8
	.dc.b	$67,$D7,$7C,$E2,$4F,$8D,$FF,$36,$7C,$3F,$F9,$FB,$0F,$EA,$6D,$9E
	.dc.b	$6D,$BF,$E1,$B5,$FE,$AB,$8F,$EE,$D3,$FE,$90,$6B,$F6,$9A,$F4,$F4
	.dc.b	$BB,$91,$F9,$B1,$A7,$BC,$79,$3C,$1E,$C4,$8F,$F0,$B4,$F8,$33,$FC
	.dc.b	$B7,$91,$FE,$B0,$9F,$F4,$5B,$50,$E7,$FE,$7D,$E6,$FF,$B6,$E7,$53
	.dc.b	$96,$CE,$A3,$FD,$B9,$18,$37,$9F,$FB,$93,$D5,$F8,$44,$9E,$F8,$79
	.dc.b	$30,$C4,$7F,$35,$C7,$FC,$4B,$AF,$CE,$BF,$F4,$BE,$3F,$DD,$2B,$C5
	.dc.b	$E9,$FA,$1F,$F0,$BE,$7F,$D5,$63,$17,$EE,$6C,$FB,$A7,$4F,$C0,$9D
	.dc.b	$34,$D5,$7F,$8C,$9C,$1E,$02,$CF,$27,$C3,$1B,$5F,$1D,$FE,$B0,$D1
	.dc.b	$93,$E4,$7D,$13,$98,$C5,$7D,$28,$FF,$5B,$E1,$D5,$FF,$24,$FD,$47
	.dc.b	$F9,$13,$E6,$75,$AF,$C1,$8C,$5F,$82,$31,$AF,$FB,$62,$FF,$D8,$3E
	.dc.b	$EB,$E6,$BF,$C0,$D1,$D2,$75,$5C,$31,$EB,$87,$E1,$31,$EB,$7C,$B0
	.dc.b	$E3,$7B,$AF,$A4,$6F,$0F,$F7,$2B,$93,$37,$E6,$6F,$CF,$BE,$53,$8C
	.dc.b	$77,$8C,$38,$94,$62,$38,$CF,$05,$3A,$8F,$4C,$62,$79,$E3,$F0,$BD
	.dc.b	$A7,$80,$E3,$DE,$3C,$70,$F9,$47,$8F,$1C,$E3,$CB,$3F,$F4,$6E,$3C
	.dc.b	$B0,$CD,$7C,$8D,$DF,$5C,$F9,$3E,$D9,$74,$E4,$FC,$1F,$AB,$8D,$FE
	.dc.b	$FD,$EF,$2F,$A3,$5A,$7A,$46,$6F,$F3,$ED,$8F,$D1,$63,$E0,$C7,$BC
	.dc.b	$62,$75,$5F,$A1,$6B,$FB,$F6,$D7,$19,$FC,$EC,$6A,$FF,$33,$EA,$FA
	.dc.b	$BA,$7F,$CD,$7C,$DF,$E3,$AB,$FD,$D9,$D6,$74,$FF,$26,$6E,$3F,$C9
	.dc.b	$9B,$7D,$27,$8C,$F8,$E7,$35,$D6,$3F,$B8,$E1,$5F,$64,$FF,$A5,$9F
	.dc.b	$05,$7E,$2B,$1F,$A3,$39,$9E,$93,$9A,$E5,$1C,$0F,$CC,$D7,$09,$FE
	.dc.b	$5F,$97,$88,$F1,$FC,$A5,$8F,$E5,$AF,$5C,$BF,$7A,$DC,$8D,$24,$FF
	.dc.b	$3A,$6E,$1F,$D0,$DE,$1F,$07,$F0,$0F,$83,$F5,$D6,$BF,$9D,$BD,$48
	.dc.b	$FF,$44,$7C,$1E,$67,$C8,$3E,$4F,$CF,$59,$FE,$87,$17,$0F,$E5,$D3
	.dc.b	$77,$FB,$73,$6B,$FA,$EC,$DA,$FE,$BA,$FB,$49,$1F,$D5,$1F,$81,$F3
	.dc.b	$3F,$60,$FB,$3F,$5D,$67,$FA,$9B,$F1,$48,$FF,$5C,$7E,$47,$D1,$FE
	.dc.b	$03,$EB,$BF,$02,$D9,$FE,$B6,$DC,$3F,$A3,$37,$2F,$C9,$BC,$27,$FB
	.dc.b	$0F,$A5,$6B,$FA,$63,$F5,$49,$1F,$DB,$1F,$B1,$F0,$1F,$90,$3F,$2F
	.dc.b	$C6,$F4,$5E,$56,$6F,$FB,$6D,$5C,$3F,$C8,$B5,$A9,$C9,$FD,$CB,$E4
	.dc.b	$FF,$36,$74,$FA,$7F,$E3,$71,$92,$57,$F6,$19,$D7,$7F,$09,$FD,$2F
	.dc.b	$3F,$FB,$87,$E4,$BF,$53,$E5,$1F,$D6,$FC,$A3,$E6,$1E,$E6,$A7,$31
	.dc.b	$8C,$3F,$97,$D6,$EB,$F4,$CD,$8C,$CE,$F3,$FE,$7B,$39,$6C,$65,$FE
	.dc.b	$E5,$19,$6C,$65,$FF,$13,$FD,$77,$EF,$0F,$9B,$EF,$7F,$B2,$EF,$FC
	.dc.b	$2F,$19,$69,$FC,$13,$89,$3B,$AF,$F0,$35,$7D,$6F,$61,$F2,$7F,$C3
	.dc.b	$3E,$BB,$FD,$CD,$8F,$FA,$9A,$31,$C9,$B6,$C6,$4F,$9B,$3C,$30,$FF
	.dc.b	$63,$6F,$DF,$A3,$F9,$CF,$0A,$31,$7F,$E2,$8D,$FE,$87,$C5,$7D,$4F
	.dc.b	$4B,$E7,$BF,$33,$84,$D2,$57,$CE,$6E,$5F,$31,$AF,$F9,$FB,$CC,$FF
	.dc.b	$0A,$67,$E2,$7F,$B4,$57,$E1,$1C,$BF,$E3,$AC,$FF,$B1,$F9,$5F,$CB
	.dc.b	$A7,$37,$FE,$60,$78,$BF,$EE,$37,$33,$C2,$BA,$7C,$0F,$13,$97,$8D
	.dc.b	$63,$24,$BE,$51,$97,$FD,$F7,$F1,$0F,$AE,$7D,$5F,$EB,$3E,$AB,$FB
	.dc.b	$45,$6A,$73,$87,$DF,$9C,$EB,$FD,$A3,$E4,$1F,$5D,$FE,$29,$AF,$F0
	.dc.b	$3C,$63,$D9,$FE,$7C,$D8,$CC,$EF,$1F,$F0,$AC,$E5,$B1,$97,$FC,$95
	.dc.b	$19,$6C,$65,$F9,$9F,$5D,$FE,$F3,$AF,$D9,$BF,$E3,$C6,$ED,$FD,$72
	.dc.b	$31,$2C,$7F,$13,$1C,$63,$15,$D3,$6E,$17,$C5,$FE,$CB,$F5,$DF,$12
	.dc.b	$B6,$7F,$A6,$1C,$61,$F9,$45,$E1,$F0,$DE,$63,$FC,$22,$76,$FE,$15
	.dc.b	$DB,$98,$FE,$17,$F5,$D2,$3F,$B9,$F4,$7F,$A1,$39,$71,$97,$8A,$31
	.dc.b	$3F,$C2,$A3,$F1,$67,$F0,$22,$9F,$A5,$28,$8E,$2F,$F0,$C3,$2E,$1E
	.dc.b	$53,$FD,$E7,$C0,$71,$7F,$28,$9F,$87,$CA,$FA,$E7,$93,$F8,$DF,$17
	.dc.b	$FC,$1B,$18,$CF,$FC,$1B,$CA,$FA,$E7,$C3,$7E,$21,$B3,$F8,$C6,$E5
	.dc.b	$F8,$86,$BF,$E2,$EF,$98,$FE,$0D,$C9,$FF,$00,$5B,$5F,$EA,$AF,$80
	.dc.b	$7D,$77,$F2,$9A,$D7,$8B,$FE,$20,$E9,$7B,$74,$FF,$12,$DB,$E5,$1F
	.dc.b	$29,$9C,$BF,$C4,$B3,$98,$FF,$0E,$36,$3F,$CD,$2D,$D3,$FC,$E6,$D7
	.dc.b	$FD,$4C,$D7,$CD,$5D,$A7,$8B,$63,$8B,$5F,$F5,$77,$1F,$F1,$AF,$8A
	.dc.b	$FF,$1C,$F8,$87,$FB,$3F,$AC,$F5,$F3,$18,$7E,$D4,$67,$1F,$F1,$B3
	.dc.b	$70,$FF,$18,$B5,$FF,$2D,$B5,$FF,$5C,$35,$FF,$6B,$37,$A4,$E7,$1F
	.dc.b	$F4,$F7,$C2,$7F,$93,$BE,$10,$FF,$94,$3E,$79,$F2,$99,$8C,$3F,$E3
	.dc.b	$B8,$CE,$5F,$E9,$4B,$46,$32,$FF,$4F,$C6,$2B,$81,$2B,$85,$6E,$BE
	.dc.b	$4B,$5F,$E6,$DF,$5D,$F8,$85,$68,$C6,$6F,$AE,$FC,$56,$1C,$39,$7F
	.dc.b	$C2,$B1,$FD,$B7,$0F,$F7,$C3,$6B,$87,$AB,$FA,$98,$C1,$1C,$12,$FD
	.dc.b	$DF,$15,$FD,$13,$A3,$FA,$FB,$63,$FA,$65,$BE,$BB,$FC,$F3,$39,$C7
	.dc.b	$D3,$39,$AF,$90,$D8,$F8,$ED,$7F,$16,$D7,$AD,$7F,$9A,$E3,$11,$D2
	.dc.b	$30,$7F,$9B,$FE,$1B,$F9,$6E,$1F,$DD,$1B,$7F,$EF,$DF,$AE,$FF,$5E
	.dc.b	$3F,$E2,$88,$CD,$B0,$56,$A3,$FD,$02,$D7,$FD,$56,$D7,$A5,$7F,$13
	.dc.b	$8C,$47,$36,$CF,$49,$D1,$E7,$E0,$FF,$4F,$DF,$E2,$70,$FF,$05,$CB
	.dc.b	$E3,$33,$5F,$EA,$48,$C1,$F9,$12,$E3,$F9,$06,$BF,$B4,$6B,$FE,$01
	.dc.b	$F4,$5F,$05,$FE,$AE,$6C,$FF,$0A,$FA,$E7,$CB,$66,$A4,$FF,$93,$AB
	.dc.b	$77,$FD,$43,$B5,$87,$0A,$D5,$FF,$AE,$BC,$57,$C6,$53,$1E,$A5,$AF
	.dc.b	$F2,$CC,$E9,$B3,$A6,$9F,$E3,$FF,$AB,$7D,$27,$F9,$C9,$B1,$FE,$43
	.dc.b	$FA,$C7,$D7,$66,$A4,$F1,$7F,$CF,$33,$A8,$C1,$C6,$7F,$A6,$46,$75
	.dc.b	$FF,$60,$B5,$FF,$75,$B5,$10,$BC,$78,$FA,$77,$D6,$E6,$7F,$C5,$B1
	.dc.b	$87,$13,$B9,$D7,$A4,$C7,$15,$A1,$7C,$5F,$EC,$ED,$7C,$26,$AB,$9C
	.dc.b	$F2,$8F,$35,$F8,$DE,$85,$0D,$79,$9E,$9F,$AE,$7C,$77,$FA,$2E,$39
	.dc.b	$46,$3D,$FE,$4D,$8C,$CE,$DE,$5F,$1F,$FE,$21,$78,$B5,$CF,$57,$28
	.dc.b	$C4,$7C,$DA,$FF,$74,$CF,$AD,$7C,$87,$D1,$D3,$E2,$46,$2B,$57,$DE
	.dc.b	$72,$DD,$3A,$5F,$16,$D7,$0C,$3E,$8F,$63,$F6,$27,$D6,$7C,$78,$C3
	.dc.b	$63,$35,$F3,$67,$2D,$FB,$17,$B1,$F8,$7C,$3E,$F4,$FF,$11,$9C,$BF
	.dc.b	$E9,$D6,$BF,$E3,$EA,$E6,$D8,$CC,$F2,$94,$63,$87,$F2,$B1,$8B,$F0
	.dc.b	$46,$30,$F3,$E1,$FE,$28,$E0,$FC,$66,$3F,$E9,$C6,$D7,$FA,$25,$B3
	.dc.b	$FD,$51,$FE,$CE,$FF,$B4,$9C,$F9,$1F,$EC,$FD,$3B,$CE,$DB,$1F,$ED
	.dc.b	$3F,$02,$F7,$3E,$9A,$92,$FF,$DB,$B1,$FE,$20,$6C,$65,$FF,$1F,$C7
	.dc.b	$17,$7E,$97,$6F,$AF,$FF,$71,$47,$F7,$76,$BF,$DD,$9A,$FF,$86,$9A
	.dc.b	$FF,$A9,$FD,$D7,$91,$E8,$6C,$7F,$BC,$1A,$FF,$94,$27,$AE,$DF,$D4
	.dc.b	$7C,$8F,$D6,$69,$FA,$C6,$F5,$7F,$67,$BF,$F0,$53,$5F,$F0,$4B,$5F
	.dc.b	$F1,$6B,$5F,$F5,$D3,$5F,$F0,$2B,$65,$F8,$7B,$E7,$7E,$7B,$6B,$F8
	.dc.b	$BE,$1F,$ED,$88,$C4,$7F,$9A,$2F,$C5,$3F,$ED,$CE,$57,$D8,$FA,$9B
	.dc.b	$5F,$DE,$A3,$04,$70,$43,$FB,$EB,$F0,$B1,$79,$9F,$82,$6C,$F9,$39
	.dc.b	$7F,$DA,$D1,$88,$FF,$3D,$DF,$81,$FF,$3E,$EF,$FE,$21,$7F,$C2,$BE
	.dc.b	$6F,$CE,$8C,$47,$F8,$C9,$AF,$F9,$59,$AF,$FB,$8D,$AF,$F8,$B9,$B5
	.dc.b	$F8,$09,$FB,$3F,$87,$FE,$36,$BF,$D1,$AD,$CF,$FE,$00,$B7,$1E,$95
	.dc.b	$F9,$EE,$EF,$F2,$8D,$FF,$91,$DA,$FF,$91,$9A,$FF,$9A,$9A,$FF,$BD
	.dc.b	$DA,$FF,$90,$9B,$87,$F4,$2F,$53,$F1,$B1,$8F,$4B,$F2,$38,$7F,$BF
	.dc.b	$DB,$7F,$E4,$A6,$DF,$18,$FA,$07,$F9,$3B,$D4,$78,$3F,$98,$6C,$7F
	.dc.b	$97,$DA,$FF,$97,$9A,$FF,$A0,$9A,$FF,$64,$94,$62,$3B,$4F,$53,$F4
	.dc.b	$BF,$73,$F7,$75,$FE,$47,$68,$C6,$3C,$DB,$7C,$8F,$4C,$BE,$C1,$F4
	.dc.b	$1F,$B0,$6C,$FF,$9D,$DA,$FF,$9D,$9A,$FF,$A6,$9A,$FA,$5A,$FF,$9C
	.dc.b	$9B,$3D,$E7,$B7,$D2,$7F,$1C,$93,$4C,$47,$F2,$0F,$79,$F3,$94,$98
	.dc.b	$7A,$A3,$FD,$AE,$FB,$3F,$22,$FC,$99,$38,$BF,$85,$BF,$96,$FA,$17
	.dc.b	$D5,$FF,$2C,$7F,$C9,$15,$F1,$1A,$FF,$B9,$1D,$BF,$80,$3F,$98,$7B
	.dc.b	$BF,$DF,$FC,$3F,$F5,$1C,$7F,$80,$DA,$FF,$81,$DA,$FF,$86,$DA,$FF
	.dc.b	$88,$5A,$FF,$89,$1A,$F6,$3F,$D5,$1F,$29,$F6,$B7,$34,$FB,$B1,$B3
	.dc.b	$FD,$F7,$F5,$DF,$AD,$7F,$D4,$6F,$F4,$37,$3E,$27,$F4,$C8,$D5,$1A
	.dc.b	$6F,$93,$F1,$AD,$E4,$FC,$B3,$C3,$EB,$07,$F5,$62,$E1,$F8,$E6,$AA
	.dc.b	$FC,$BC,$7F,$6F,$8F,$F5,$BB,$A8,$CB,$CC,$E9,$E5,$FF,$64,$9F,$F0
	.dc.b	$1F,$D1,$5F,$C2,$DF,$85,$D3,$54,$9A,$66,$BE,$D5,$70,$7E,$E4,$ED
	.dc.b	$FD,$2F,$CA,$FF,$76,$7D,$19,$FE,$D6,$31,$97,$EC,$1B,$1F,$E5,$BB
	.dc.b	$E9,$5C,$0F,$F6,$9F,$D4,$FC,$96,$FC,$4F,$F7,$C8,$37,$79,$3F,$B6
	.dc.b	$9C,$B6,$72,$FF,$87,$7E,$B7,$FC,$70,$DE,$EF,$F0,$4F,$8A,$9F,$E2
	.dc.b	$3A,$7F,$87,$DD,$B7,$D5,$DC,$6B,$4F,$F1,$0B,$70,$FF,$5F,$3C,$23
	.dc.b	$FC,$F1,$F5,$EF,$9F,$FF,$30,$36,$BF,$CC,$8D,$7F,$CD,$0D,$7F,$CF
	.dc.b	$8D,$79,$C7,$22,$3F,$E6,$FE,$57,$CE,$7F,$8C,$32,$FE,$3E,$DC,$3F
	.dc.b	$94,$36,$BF,$D1,$B8,$78,$6B,$7E,$5F,$F8,$AA,$BF,$54,$4F,$53,$11
	.dc.b	$2F,$F9,$2C,$E5,$FE,$7F,$AB,$F7,$AF,$BF,$0C,$FC,$BE,$AF,$53,$E8
	.dc.b	$CE,$AB,$84,$66,$BA,$BA,$3F,$39,$F6,$0F,$A2,$FF,$45,$36,$7F,$D1
	.dc.b	$CD,$7F,$D2,$4D,$7F,$D4,$CD,$7F,$D5,$8D,$7F,$D5,$ED,$4E,$70,$F9
	.dc.b	$BF,$13,$F1,$49,$3F,$C7,$E7,$C1,$AF,$D0,$EB,$F9,$5F,$82,$CA,$F8
	.dc.b	$F1,$F3,$4C,$3F,$A1,$C3,$EC,$CE,$63,$E6,$FD,$CF,$D3,$F1,$A4,$60
	.dc.b	$87,$1D,$FF,$C2,$B8,$FE,$02,$7E,$D5,$7D,$BF,$83,$FF,$23,$D7,$FB
	.dc.b	$6A,$FF,$10,$D7,$FD,$95,$8F,$E2,$A7,$7F,$39,$F8,$32,$4F,$E0,$22
	.dc.b	$BE,$1F,$9E,$FC,$41,$8C,$BF,$23,$58,$9F,$C3,$CE,$79,$BF,$27,$F4
	.dc.b	$9E,$42,$4F,$91,$F3,$FD,$37,$E4,$C9,$1F,$90,$8F,$CB,$7D,$27,$B0
	.dc.b	$91,$EC,$7D,$FE,$EF,$E6,$C9,$1F,$98,$8A,$FC,$98,$B9,$7E,$A2,$6E
	.dc.b	$1F,$A8,$9B,$5F,$9B,$D5,$5F,$A7,$F8,$9F,$9A,$49,$F9,$AF,$D0,$FA
	.dc.b	$0F,$D6,$12,$3F,$53,$1F,$E9,$8F,$CE,$3F,$D6,$F1,$F4,$9A,$FF,$C0
	.dc.b	$1E,$04,$62,$3F,$C4,$4D,$7F,$5C,$D8,$DE,$5F,$E8,$68,$C1,$3E,$67
	.dc.b	$C0,$4A,$FF,$0F,$C7,$8B,$4F,$F6,$07,$D6,$BE,$67,$FD,$A8,$DA,$FF
	.dc.b	$69,$35,$10,$76,$23,$51,$98,$C4,$F4,$24,$74,$7D,$1E,$67,$C5,$24
	.dc.b	$7F,$0E,$AF,$05,$FF,$8F,$30,$FE,$C9,$97,$C4,$BD,$DF,$FB,$76,$F7
	.dc.b	$3F,$ED,$87,$FA,$D4,$3E,$05,$F2,$19,$F3,$BD,$51,$89,$F5,$D7,$08
	.dc.b	$C6,$9E,$DF,$8D,$DE,$1C,$6F,$E1,$D6,$AF,$84,$E6,$7C,$0F,$8E,$B8
	.dc.b	$C7,$6C,$2E,$1F,$25,$B7,$CA,$7C,$75,$C1,$FE,$83,$EC,$FF,$8B,$B1
	.dc.b	$F4,$46,$3D,$4F,$1E,$5F,$57,$97,$FC,$B7,$18,$97,$C4,$7F,$9B,$3E
	.dc.b	$63,$A5,$6A,$7C,$E7,$DE,$71,$87,$66,$E1,$F8,$2B,$E5,$5D,$4F,$11
	.dc.b	$F8,$53,$96,$3F,$E3,$CA,$ED,$19,$6B,$E7,$F4,$DF,$1F,$C6,$3C,$51
	.dc.b	$87,$12,$8C,$47,$36,$BC,$E7,$C9,$3B,$7E,$15,$BD,$87,$D7,$B8,$B7
	.dc.b	$8F,$7D,$3E,$03,$74,$E9,$9F,$C6,$C7,$96,$1C,$63,$E4,$9F,$DD,$AB
	.dc.b	$83,$FA,$6F,$C7,$5D,$1B,$C7,$F1,$B1,$8C,$BE,$B7,$22,$3C,$91,$87
	.dc.b	$57,$F5,$A7,$57,$CF,$3F,$DA,$C6,$2F,$81,$9A,$F1,$55,$9D,$47,$3B
	.dc.b	$EB,$87,$92,$F5,$18,$C3,$F8,$9E,$3E,$FE,$7F,$DA,$36,$BC,$3F,$1C
	.dc.b	$EB,$C6,$5F,$C6,$2A,$F5,$FF,$1E,$6F,$57,$DB,$C3,$FC,$1B,$87,$17
	.dc.b	$F8,$BC,$F2,$D5,$39,$C3,$F1,$D1,$8C,$7F,$94,$36,$FF,$14,$D9,$DD
	.dc.b	$71,$7F,$8B,$35,$F4,$D7,$28,$C4,$7F,$30,$E4,$F3,$9E,$8E,$0B,$87
	.dc.b	$F2,$C7,$F9,$C6,$7F,$CE,$9B,$9F,$E3,$A7,$FC,$74,$DB,$FE,$85,$3A
	.dc.b	$24,$E9,$E3,$EE,$7C,$63,$E3,$FC,$CD,$9F,$E7,$6F,$62,$3D,$9F,$0F
	.dc.b	$BB,$F3,$8F,$9F,$EC,$D9,$FE,$87,$17,$0F,$F0,$B6,$2A,$C7,$07,$F3
	.dc.b	$1A,$B7,$FD,$52,$6D,$FF,$54,$7D,$E4,$8F,$7B,$F0,$BD,$8F,$BE,$7E
	.dc.b	$03,$E2,$AC,$FF,$54,$7E,$39,$1F,$8E,$FC,$AF,$B3,$FC,$3F,$AF,$F8
	.dc.b	$86,$CF,$F5,$C6,$E1,$EA,$6E,$5F,$D2,$5E,$13,$FD,$8B,$E0,$ED,$7C
	.dc.b	$07,$EB,$92,$3E,$BB,$F6,$BE,$8B,$F2,$FF,$67,$F9,$86,$CF,$F6,$C6
	.dc.b	$D7,$CA,$6E,$1F,$D6,$7E,$7F,$F3,$FF,$21,$FA,$06,$DF,$F7,$46,$DF
	.dc.b	$D2,$6D,$7F,$61,$FA,$FF,$D3,$9F,$AC,$F9,$37,$CA,$35,$CE,$3F,$DD
	.dc.b	$AF,$F8,$A5,$6A,$7F,$C1,$8D,$9F,$E5,$CD,$C7,$F9,$C3,$5F,$EF,$CD
	.dc.b	$7F,$C0,$2D,$4D,$31,$D7,$FE,$85,$F3,$3F,$7B,$F6,$7F,$B9,$F7,$51
	.dc.b	$8C,$3F,$D1,$71,$87,$FC,$2B,$3B,$9D,$61,$FE,$69,$FC,$C1,$E5,$7F
	.dc.b	$0A,$6D,$FB,$CF,$F0,$AF,$6A,$1F,$29,$BA,$95,$BB,$FF,$4C,$57,$27
	.dc.b	$B5,$69,$F0,$7C,$57,$73,$E8,$70,$7F,$BF,$AF,$96,$3D,$E7,$3E,$87
	.dc.b	$CE,$AD,$5F,$CC,$6B,$F3,$1A,$FC,$C6,$BF,$31,$BB,$FD,$E9,$E8,$7F
	.dc.b	$0E,$CB,$FC,$37,$8F,$89,$7C,$EF,$C5,$F1,$BE,$33,$67,$4D,$C7,$E6
	.dc.b	$3F,$CD,$3C,$DF,$F1,$EC,$FF,$B0,$3C,$DC,$56,$5B,$19,$3F,$D8,$B1
	.dc.b	$97,$FD,$ED,$E1,$7F,$3E,$F8,$83,$FD,$0A,$9F,$81,$BA,$B5,$3F,$D1
	.dc.b	$B0,$7C,$63,$EC,$78,$5E,$F2,$FB,$15,$A9,$CE,$1C,$6F,$85,$7D,$DC
	.dc.b	$7B,$CE,$6B,$73,$F6,$3E,$37,$DF,$87,$FA,$D5,$BD,$9F,$DE,$F9,$51
	.dc.b	$FE,$14,$AE,$15,$FD,$6A,$37,$1F,$90,$6C,$7F,$B8,$DE,$0F,$53,$FB
	.dc.b	$0F,$D6,$3E,$06,$FC,$06,$1B,$7F,$8C,$9D,$36,$74,$D3,$F2,$3D,$1F
	.dc.b	$D8,$F2,$F4,$36,$3F,$6F,$83,$E2,$7F,$B8,$CE,$63,$F7,$26,$67,$F8
	.dc.b	$1B,$4F,$E0,$65,$3F,$81,$94,$EB,$3F,$C3,$1C,$1D,$1C,$1F,$D9,$3E
	.dc.b	$A9,$F4,$5F,$E0,$25,$F2,$76,$30,$FF,$81,$BE,$4B,$F8,$25,$FF,$B6
	.dc.b	$1A,$FF,$A9,$DA,$FF,$00,$6A,$DD,$0F,$AD,$E8,$75,$F9,$FB,$9F,$EE
	.dc.b	$BE,$57,$F5,$2D,$EB,$FF,$0B,$FC,$B7,$E6,$A3,$07,$F0,$07,$19,$7F
	.dc.b	$BB,$1A,$FF,$AE,$DA,$FF,$14,$6B,$E5,$6E,$46,$93,$AB,$FC,$17,$F2
	.dc.b	$FF,$5E,$4D,$3F,$1E,$E3,$D9,$FF,$16,$C7,$E7,$2F,$FD,$98,$D7,$F9
	.dc.b	$2B,$5F,$8C,$D9,$EB,$1E,$02,$61,$FE,$73,$8C,$B7,$DD,$79,$1F,$E3
	.dc.b	$C6,$C7,$F2,$4F,$99,$FC,$57,$37,$EA,$E7,$38,$23,$F5,$D3,$F8,$8C
	.dc.b	$0D,$27,$2F,$F8,$B1,$B8,$78,$DB,$7F,$E8,$FA,$7C,$D7,$DA,$8E,$D5
	.dc.b	$FD,$12,$31,$87,$EF,$7E,$73,$FC,$D6,$FE,$FE,$33,$3F,$D2,$A7,$FC
	.dc.b	$DF,$5F,$EF,$56,$B9,$CB,$FC,$7F,$18,$8F,$C8,$37,$03,$49,$3F,$E4
	.dc.b	$6F,$27,$EC,$9F,$6B,$CD,$7C,$47,$F9,$9A,$30,$FF,$93,$23,$38,$FD
	.dc.b	$36,$E3,$F0,$5A,$FF,$BD,$71,$FF,$7A,$D7,$12,$57,$17,$FC,$AF,$E1
	.dc.b	$7E,$D2,$3F,$AD,$B6,$BE,$49,$FA,$BF,$9D,$7B,$5F,$E7,$A9,$D6,$DF
	.dc.b	$C9,$E7,$37,$F8,$76,$DF,$D5,$6B,$E5,$6B,$FD,$E1,$AF,$7B,$FF,$35
	.dc.b	$C6,$31,$37,$E0,$BF,$1B,$7F,$E0,$06,$DF,$F0,$18,$7B,$BF,$E9,$09
	.dc.b	$CD,$FF,$AF,$E3,$FC,$EF,$3A,$3F,$3C,$5A,$FC,$63,$5F,$99,$F4,$AF
	.dc.b	$8A,$FF,$4D,$B6,$7F,$C0,$6B,$E7,$B3,$32,$72,$F0,$8C,$57,$CE,$C1
	.dc.b	$F4,$9F,$E1,$8B,$FF,$06,$B6,$74,$BC,$4F,$F5,$5B,$9D,$BF,$AC,$36
	.dc.b	$BF,$C4,$6D,$AF,$E4,$8F,$18,$EB,$C5,$FE,$AD,$FA,$17,$B9,$4F,$F5
	.dc.b	$87,$C2,$FC,$CE,$54,$EA,$FF,$C5,$51,$8A,$E1,$1F,$EB,$AE,$AB,$97
	.dc.b	$E2,$AB,$FD,$73,$1F,$EA,$D2,$4F,$FA,$DB,$0F,$E9,$34,$27,$15,$E9
	.dc.b	$FE,$B5,$19,$3F,$B6,$B8,$71,$1B,$BD,$46,$9A,$A3,$4F,$3E,$6F,$F7
	.dc.b	$D5,$73,$8F,$09,$FC,$32,$B9,$CE,$9E,$75,$D2,$7C,$25,$3B,$15,$D7
	.dc.b	$D1,$3E,$33,$FC,$F3,$3E,$48,$C7,$B1,$A9,$CC,$FA,$A7,$2E,$FE,$7F
	.dc.b	$F9,$53,$78,$BA,$C6,$23,$83,$5F,$91,$28,$E3,$9F,$C9,$87,$CF,$FC
	.dc.b	$D8,$C3,$73,$E3,$5C,$1B,$87,$CA,$C1,$87,$27,$0D,$7E,$51,$F7,$5F
	.dc.b	$14,$7D,$2D,$FF,$DF,$BE,$3F,$CB,$6D,$7E,$16,$31,$B9,$D7,$13,$E2
	.dc.b	$BF,$09,$83,$8B,$F1,$58,$FE,$62,$73,$18,$AE,$18,$FE,$6B,$1E,$8F
	.dc.b	$E7,$23,$2F,$E6,$5B,$87,$58,$61,$D5,$E5,$8F,$67,$FC,$25,$C1,$F9
	.dc.b	$5A,$FF,$26,$37,$1F,$F4,$BB,$67,$7F,$39,$C7,$A3,$FD,$7B,$B5,$AD
	.dc.b	$46,$2B,$FD,$23,$0A,$E1,$F1,$BF,$CD,$A7,$FD,$2A,$D7,$F8,$73,$71
	.dc.b	$FF,$2F,$3F,$EF,$F8,$DF,$7F,$22,$3C,$BE,$27,$C9,$F4,$4F,$95,$FD
	.dc.b	$06,$68,$C5,$7F,$81,$DA,$FF,$75,$6B,$FD,$57,$E0,$3F,$D6,$AF,$FA
	.dc.b	$A9,$AF,$F2,$36,$E5,$FE,$73,$9E,$6F,$F9,$D7,$A5,$F2,$9E,$26,$DF
	.dc.b	$F8,$5D,$AF,$F8,$15,$AF,$F6,$56,$BF,$D2,$63,$91,$1F,$EB,$5F,$0F
	.dc.b	$F4,$FC,$53,$4F,$AD,$5D,$AF,$FD,$11,$C8,$F9,$EF,$8A,$DC,$3F,$BC
	.dc.b	$46,$08,$E0,$87,$F7,$B7,$E6,$E8,$F8,$CF,$C2,$36,$7F,$DA,$2D,$7F
	.dc.b	$A5,$57,$77,$FA,$67,$83,$FE,$8F,$7F,$C7,$4D,$7F,$C5,$0D,$7F,$C1
	.dc.b	$0D,$7F,$DD,$6D,$7F,$CD,$8D,$9F,$F7,$6F,$B5,$F8,$28,$C7,$97,$F8
	.dc.b	$DE,$8F,$F1,$57,$27,$E6,$E3,$18,$72,$BF,$CF,$78,$1F,$D9,$E3,$FC
	.dc.b	$9C,$D7,$FC,$78,$D7,$FC,$38,$D7,$FC,$79,$E3,$7E,$4A,$31,$1F,$E8
	.dc.b	$16,$E5,$FE,$44,$9F,$CB,$F9,$9F,$8A,$CB,$FA,$74,$FF,$94,$39,$BF
	.dc.b	$61,$39,$C3,$95,$FE,$CF,$E8,$BF,$94,$47,$F9,$8D,$AF,$F9,$45,$AF
	.dc.b	$F8,$C5,$AF,$66,$BF,$E9,$56,$D9,$A6,$3C,$FF,$B5,$C5,$FE,$D3,$6E
	.dc.b	$5F,$E7,$AE,$2F,$27,$87,$FC,$1F,$D2,$7C,$8F,$ED,$1B,$9F,$F9,$E5
	.dc.b	$AF,$FA,$5E,$34,$47,$4D,$7F,$D3,$9F,$1B,$D4,$8F,$FA,$86,$3F,$8B
	.dc.b	$7A,$DF,$EB,$A8,$FF,$55,$7B,$38,$33,$F5,$9F,$C6,$FD,$AF,$C4,$13
	.dc.b	$97,$F8,$4B,$F8,$BF,$94,$FF,$1B,$DF,$F8,$07,$83,$FD,$E5,$0F,$67
	.dc.b	$F9,$23,$67,$5C,$1F,$6A,$38,$C7,$F9,$60,$FF,$2B,$FC,$57,$F7,$3E
	.dc.b	$A3,$49,$1F,$DE,$63,$2D,$8C,$BF,$E0,$88,$CB,$FE,$8F,$F4,$FE,$9B
	.dc.b	$FE,$CF,$6F,$37,$F8,$D7,$93,$FA,$14,$F8,$8F,$39,$FE,$D5,$F8,$4F
	.dc.b	$D1,$C9,$FE,$65,$1F,$E4,$36,$E9,$FD,$3E,$75,$87,$F4,$59,$E2,$6B
	.dc.b	$3F,$CD,$B6,$FB,$4F,$F4,$BF,$8A,$14,$91,$E4,$BC,$F9,$F8,$8F,$F2
	.dc.b	$A4,$F0,$74,$D7,$31,$EE,$1F,$2F,$F0,$3E,$8D,$F3,$D4,$D4,$4A,$F8
	.dc.b	$62,$76,$E1,$AF,$DB,$AE,$AF,$9F,$E1,$BF,$B0,$37,$CE,$FE,$D1,$18
	.dc.b	$8F,$C1,$57,$2D,$BE,$57,$C5,$7C,$5D,$0B,$E5,$3F,$B8,$36,$FE,$23
	.dc.b	$B2,$46,$8D,$F9,$FF,$DF,$FC,$87,$D1,$FF,$87,$F1,$FF,$03,$46,$3C
	.dc.b	$1D,$4F,$F0,$89,$CB,$BE,$4F,$F0,$BB,$72,$FF,$06,$BB,$8C,$D7,$2F
	.dc.b	$63,$F1,$BF,$46,$F9,$6F,$F2,$83,$70,$FF,$2A,$35,$FF,$30,$B5,$FF
	.dc.b	$32,$B5,$FF,$36,$35,$FF,$39,$35,$39,$D9,$F1,$FF,$E2,$89,$FF,$4D
	.dc.b	$37,$3F,$CC,$37,$2E,$B8,$7F,$87,$63,$F4,$E4,$E9,$C7,$AF,$96,$61
	.dc.b	$F9,$59,$DC,$FC,$FF,$93,$FB,$D8,$F6,$E5,$E8,$64,$97,$C5,$CB,$FE
	.dc.b	$98,$9E,$39,$7F,$A2,$FE,$85,$F4,$7F,$E7,$96,$D7,$F9,$F5,$AF,$FA
	.dc.b	$29,$AF,$FA,$39,$AF,$38,$E4,$47,$FD,$23,$AF,$C8,$FA,$3F,$84,$48
	.dc.b	$FD,$1D,$76,$CB,$E4,$F5,$FC,$7F,$A3,$95,$F0,$E3,$E5,$18,$7E,$8C
	.dc.b	$61,$F9,$1F,$47,$F3,$FC,$69,$18,$21,$F6,$1C,$BF,$6A,$3E,$B7,$C8
	.dc.b	$FF,$8F,$B8,$4D,$24,$FB,$21,$F4,$7F,$74,$91,$F6,$66,$BF,$86,$FA
	.dc.b	$3F,$C1,$18,$CF,$F0,$F5,$89,$FC,$04,$E7,$83,$FB,$BE,$0F,$A3,$F2
	.dc.b	$12,$BC,$4F,$9F,$D5,$FC,$71,$23,$F1,$51,$F9,$1F,$8D,$F6,$12,$3D
	.dc.b	$4F,$BF,$DE,$FC,$B1,$23,$F2,$51,$5F,$CF,$C5,$CF,$F3,$F3,$71,$F3
	.dc.b	$CD,$A7,$1F,$EA,$2B,$F4,$5F,$21,$F3,$49,$3F,$2D,$FA,$1F,$37,$FA
	.dc.b	$72,$47,$E8,$E3,$FD,$0D,$F3,$FF,$E9,$38,$FF,$6F,$DA,$31,$1F,$EE
	.dc.b	$78,$C6,$3F,$E7,$89,$CC,$FF,$88,$9A,$F9,$9B,$3C,$23,$F3,$31,$81
	.dc.b	$F2,$3E,$22,$47,$F4,$B8,$F2,$67,$FC,$8B,$E8,$DF,$21,$FE,$BD,$6D
	.dc.b	$7F,$BD,$3C,$44,$62,$3F,$DF,$31,$89,$E0,$48,$E1,$3B,$47,$FB,$01
	.dc.b	$C3,$ED,$F3,$3E,$39,$1F,$EE,$75,$D5,$F4,$46,$55,$FF,$71,$C3,$FB
	.dc.b	$03,$EB,$C3,$FA,$06,$1F,$D5,$29,$F0,$4F,$9C,$F1,$79,$DE,$49,$CC
	.dc.b	$61,$F3,$CE,$DF,$F6,$5D,$3E,$3B,$84,$E7,$4F,$8B,$86,$30,$E5,$87
	.dc.b	$43,$8C,$65,$E7,$76,$B8,$C7,$82,$B8,$DF,$5C,$3D,$4D,$BE,$55,$FC
	.dc.b	$FB,$DA,$FF,$12,$CF,$9E,$31,$EA,$73,$7E,$8F,$47,$F9,$4A,$31,$F1
	.dc.b	$3F,$E5,$DF,$A3,$E7,$39,$AF,$41,$C8,$AE,$6F,$E2,$57,$19,$EC,$6C
	.dc.b	$F9,$85,$3E,$74,$F5,$72,$D7,$3E,$9B,$E4,$38,$C7,$A2,$30,$E2,$51
	.dc.b	$88,$E6,$D7,$9C,$F9,$27,$6F,$C0,$F6,$1F,$44,$E8,$D6,$1D,$B2,$F9
	.dc.b	$58,$70,$C7,$E0,$CE,$78,$3F,$37,$9B,$2E,$78,$F2,$CF,$E8,$70,$BE
	.dc.b	$8F,$F6,$73,$86,$E5,$F8,$59,$D4,$F8,$E3,$13,$AD,$3F,$1C,$DC,$BF
	.dc.b	$57,$7D,$7A,$38,$56,$B0,$B5,$AB,$E1,$83,$0E,$15,$E3,$C3,$73,$9A
	.dc.b	$F8,$93,$93,$F1,$5D,$3F,$C6,$75,$F3,$73,$FE,$C5,$B3,$E4,$F9,$17
	.dc.b	$5E,$31,$FE,$31,$57,$0E,$57,$FC,$4E,$35,$9B,$0E,$33,$93,$F8,$1B
	.dc.b	$8B,$FE,$22,$D7,$F7,$53,$C7,$0F,$E5,$2D,$AF,$E3,$55,$BA,$F3,$56
	.dc.b	$A3,$11,$F1,$5B,$87,$F3,$7E,$5F,$47,$57,$FA,$EB,$0F,$E5,$5C,$98
	.dc.b	$9C,$BF,$CE,$2B,$F9,$F3,$6B,$F0,$B8,$7F,$42,$6B,$FD,$05,$AF,$F3
	.dc.b	$F8,$E0,$4B,$FE,$7B,$F2,$DE,$23,$C7,$EE,$D8,$FE,$78,$F7,$23,$FD
	.dc.b	$11,$F1,$7B,$7E,$A3,$D7,$EB,$D9,$FE,$89,$17,$0F,$E5,$B1,$FD,$46
	.dc.b	$71,$C1,$FC,$BE,$AC,$E4,$E2,$7F,$4D,$9F,$80,$47,$FA,$AB,$F0,$FD
	.dc.b	$8F,$96,$7C,$CF,$8B,$B3,$FD,$55,$F9,$04,$7F,$AE,$BF,$2F,$E8,$7E
	.dc.b	$9F,$AD,$F5,$1B,$3F,$D7,$5B,$97,$A8,$FE,$D0,$E3,$2F,$E9,$0F,$19
	.dc.b	$FE,$C9,$F0,$96,$B8,$70,$7F,$77,$8F,$B4,$49,$FE,$DE,$FD,$DF,$A3
	.dc.b	$FC,$2F,$CB,$7E,$19,$B3,$FD,$BD,$B7,$D3,$89,$18,$CB,$FA,$D7,$D0
	.dc.b	$7E,$3B,$E4,$3F,$1E,$DC,$3F,$BC,$37,$0E,$4D,$AF,$EC,$5F,$47,$F9
	.dc.b	$63,$F6,$1F,$28,$F9,$56,$F9,$C7,$FD,$BF,$39,$8F,$F0,$7B,$63,$F9
	.dc.b	$8B,$71,$FE,$76,$D7,$FB,$43,$5F,$EF,$AD,$4D,$31,$D6,$F6,$DE,$6F
	.dc.b	$EB,$3E,$6F,$F8,$0C,$63,$0F,$F0,$B4,$62,$75,$1F,$E8,$56,$E1,$F5
	.dc.b	$BE,$8D,$E5,$7F,$0D,$6D,$7F,$BA,$57,$CA,$E6,$64,$E5,$E1,$18,$AE
	.dc.b	$9F,$15,$CC,$F9,$7B,$7F,$BD,$AF,$A4,$63,$D1,$F9,$33,$9A,$F8,$ED
	.dc.b	$7E,$3B,$5F,$8E,$D7,$E3,$B7,$6F,$B1,$3D,$0F,$E1,$99,$7F,$86,$70
	.dc.b	$F6,$7C,$EF,$BF,$C8,$F8,$0D,$8C,$9F,$06,$32,$FF,$AD,$F1,$FE,$65
	.dc.b	$E6,$FF,$6F,$8C,$1F,$EB,$BF,$2F,$33,$A6,$C6,$4F,$F7,$8F,$85,$FC
	.dc.b	$EF,$E2,$0F,$F3,$F9,$7D,$1B,$A7,$53,$FD,$0F,$07,$C6,$3E,$8F,$85
	.dc.b	$EF,$2F,$F2,$3F,$5F,$F9,$A6,$B5,$18,$3F,$C9,$79,$76,$CB,$A4,$EB
	.dc.b	$1F,$E0,$13,$9A,$E1,$1F,$E5,$29,$D5,$FF,$6F,$AD,$4F,$F0,$0F,$8D
	.dc.b	$FF,$2B,$DF,$0F,$09,$18,$9D,$57,$F5,$B8,$C5,$FF,$79,$AD,$D6,$4D
	.dc.b	$55,$BE,$9F,$46,$F8,$1F,$EC,$8F,$FA,$EA,$33,$18,$F8,$0C,$B6,$32
	.dc.b	$72,$8F,$ED,$9D,$1F,$D9,$72,$F1,$36,$7F,$6B,$8B,$E2,$37,$E8,$7F
	.dc.b	$06,$8C,$4F,$F0,$06,$9F,$C0,$0A,$7F,$00,$29,$D2,$7F,$84,$B8,$3F
	.dc.b	$80,$97,$A7,$FD,$03,$1F,$D9,$7E,$8D,$F4,$7F,$E0,$35,$F2,$76,$30
	.dc.b	$FF,$81,$FE,$4B,$F8,$26,$1F,$EA,$A6,$BF,$EA,$96,$BF,$E8,$EE,$14
	.dc.b	$62,$3F,$D7,$AB,$E8,$FA,$1D,$7E,$7E,$E9,$F7,$FE,$59,$DD,$FE,$17
	.dc.b	$6C,$7E,$0F,$E5,$FF,$3E,$EE,$7F,$C3,$91,$8C,$7F,$D7,$ED,$7F,$D7
	.dc.b	$CD,$7C,$CD,$7F,$DA,$2D,$C8,$D2,$76,$7F,$83,$7E,$63,$EE,$57,$FB
	.dc.b	$61,$B4,$7F,$27,$87,$C3,$72,$7E,$66,$7F,$DA,$0D,$74,$75,$24,$7F
	.dc.b	$9C,$A7,$A4,$60,$FF,$6D,$CF,$17,$FD,$D3,$EE,$3C,$8F,$F1,$F3,$6F
	.dc.b	$F5,$5F,$37,$F8,$CE,$7F,$D3,$4E,$6F,$FB,$7C,$FE,$27,$13,$49,$C9
	.dc.b	$FE,$2D,$6E,$1E,$56,$D7,$FB,$BB,$E7,$3E,$BB,$FD,$E9,$B8,$FE,$D7
	.dc.b	$E7,$7F,$CE,$2F,$ED,$E3,$37,$F8,$99,$FF,$3A,$D7,$FB,$D5,$AE,$72
	.dc.b	$FF,$1F,$46,$21,$92,$30,$D9,$D9,$2B,$E6,$79,$3F,$5A,$F8,$B6,$BF
	.dc.b	$CE,$DF,$11,$FE,$67,$8C,$79,$C8,$FF,$27,$36,$FE,$B3,$6F,$FD,$63
	.dc.b	$7C,$A7,$35,$C0,$8E,$0F,$BD,$E0,$7E,$D1,$FE,$40,$DA,$FF,$11,$FC
	.dc.b	$F3,$DB,$FF,$3D,$4E,$76,$FE,$41,$39,$BF,$C4,$36,$FC,$6D,$7F,$DE
	.dc.b	$18,$A3,$11,$FB,$96,$BF,$31,$B1,$BC,$3F,$CD,$D1,$8F,$82,$FC,$65
	.dc.b	$FF,$01,$6D,$FD,$48,$7B,$BF,$E9,$19,$CE,$1F,$EC,$18,$FF,$3C,$4E
	.dc.b	$8F,$D0,$16,$BE,$5B,$5F,$4F,$D1,$BE,$2B,$FD,$38,$D9,$FE,$90,$BE
	.dc.b	$7F,$33,$27,$2F,$08,$C5,$7F,$94,$3E,$8F,$FC,$2F,$5F,$D4,$5B,$19
	.dc.b	$F1,$7F,$D5,$51,$8D,$BF,$A8,$36,$BF,$91,$36,$BF,$2A,$AB,$8F,$17
	.dc.b	$FA,$C3,$E8,$9E,$ED,$3F,$D6,$7F,$0B,$F2,$B9,$53,$AC,$BF,$B5,$C6
	.dc.b	$2B,$84,$7F,$AD,$BA,$AE,$7F,$E7,$1A,$FF,$5E,$47,$FA,$BC,$93,$FE
	.dc.b	$B4,$7F,$DC,$13,$FF,$00,$07,$11,$74,$FF,$59,$70,$7F,$86,$4C,$CE
	.dc.b	$1E,$13,$C6,$78,$47,$F8,$22,$B1,$5B,$8E,$E7,$23,$AF,$37,$FB,$CE
	.dc.b	$B8,$C7,$43,$47,$32,$99,$3D,$57,$EB,$70,$77,$7D,$DF,$46,$F8,$CF
	.dc.b	$F3,$B4,$F8,$E3,$1E,$C6,$A7,$33,$ED,$9C,$BB,$F9,$E7,$1F,$46,$C6
	.dc.b	$63,$07,$BF,$57,$D2,$3E,$73,$E8,$46,$1B,$19,$AE,$0D,$7E,$74,$ED
	.dc.b	$CC,$F0,$23,$EE,$FA,$9F,$85,$89,$A4,$8F,$27,$EC,$EF,$91,$F1,$B6
	.dc.b	$33,$18,$23,$82,$1E,$48,$E0,$F9,$BF,$16,$79,$5E,$E6,$C7,$C3,$6B
	.dc.b	$B8,$D1,$1D,$35,$FC,$47,$E2,$2F,$B8,$FB,$CD,$8F,$CD,$35,$DC,$68
	.dc.b	$8E,$9A,$FE,$4F,$EA,$DF,$0F,$F8,$96,$C7,$E9,$5A,$EE,$3F,$1D,$F8
	.dc.b	$43,$E9,$29,$27,$F5,$91,$97,$97,$C6,$F5,$23,$82,$1E,$37,$2D,$39
	.dc.b	$FD,$0F,$80,$8E,$0F,$DD,$7E,$60,$FA,$2F,$D1,$36,$33,$18,$23,$82
	.dc.b	$1F,$B5,$8E,$0F,$ED,$ED,$EE,$FC,$02,$47,$B9,$FD,$87,$CA,$FF,$8E
	.dc.b	$A3,$F8,$63,$5F,$E0,$0D,$7F,$84,$35,$FE,$1A,$D7,$F8,$93,$53,$4C
	.dc.b	$4B,$E7,$FE,$99,$27,$E8,$BF,$C0,$FF,$3B,$7D,$35,$24,$FF,$21,$8C
	.dc.b	$B6,$32,$FF,$17,$8C,$B6,$3F,$87,$70,$3E,$5A,$92,$7F,$96,$46,$5F
	.dc.b	$E5,$DC,$ED,$FF,$8B,$36,$33,$18,$23,$82,$1F,$CA,$63,$83,$FC,$AF
	.dc.b	$F3,$67,$D0,$7F,$21,$6C,$7F,$3D,$6B,$B8,$FE,$4D,$F9,$BF,$F9,$D6
	.dc.b	$3F,$A1,$35,$FE,$6C,$D7,$F9,$F3,$5F,$E8,$6D,$7F,$A4,$35,$34,$C7
	.dc.b	$E6,$FF,$DD,$63,$FA,$83,$5F,$E8,$8D,$7F,$A6,$35,$FE,$A2,$FF,$39
	.dc.b	$78,$F8,$5F,$CF,$E3,$11,$CA,$3F,$A1,$7E,$8A,$F9,$8A,$49,$FE,$C1
	.dc.b	$19,$6F,$A9,$C0,$8E,$0F,$EC,$DF,$3E,$FB,$EF,$E9,$CD,$8C,$C6,$08
	.dc.b	$E0,$87,$F6,$18,$E0,$FF,$63,$F6,$7F,$94,$BF,$DA,$DA,$FF,$70,$6B
	.dc.b	$FD,$D5,$AB,$EB,$DF,$09,$FD,$85,$B1,$FD,$F5,$AF,$F7,$16,$BF,$DE
	.dc.b	$5A,$FF,$7E,$6B,$FE,$03,$6B,$DE,$3B,$7D,$54,$8E,$CD,$7B,$35,$EC
	.dc.b	$D7,$FC,$17,$6F,$96,$F9,$64,$8F,$9E,$D7,$E4,$35,$FF,$08,$47,$17
	.dc.b	$DD,$F3,$5F,$04,$8E,$0F,$F0,$F7,$D6,$FF,$95,$1B,$A7,$F8,$1E,$31
	.dc.b	$1F,$E1,$56,$BF,$E1,$D6,$BF,$E2,$26,$BF,$E2,$76,$A6,$8F,$A0,$EC
	.dc.b	$48,$F1,$B5,$F0,$35,$F0,$3F,$E1,$BF,$CA,$5F,$45,$49,$3F,$E3,$98
	.dc.b	$CB,$63,$2F,$F8,$BE,$32,$D8,$FF,$13,$7D,$6F,$F9,$56,$3F,$C8,$4D
	.dc.b	$7F,$C6,$CD,$7F,$C7,$CD,$7F,$C8,$6B,$D2,$E6,$48,$E6,$D7,$9B,$5E
	.dc.b	$6F,$8B,$86,$47,$89,$AF,$89,$AF,$89,$AF,$F9,$53,$F4,$EF,$F5,$B3
	.dc.b	$FE,$5B,$6B,$FE,$4F,$6B,$FE,$57,$6B,$C7,$E1,$7F,$C9,$11,$88,$FF
	.dc.b	$31,$35,$37,$C7,$FE,$F9,$1F,$E5,$B6,$BF,$E6,$36,$BF,$E6,$86,$BF
	.dc.b	$E6,$D5,$F2,$BF,$E5,$76,$C7,$F9,$CD,$AF,$58,$D1,$1D,$35,$F1,$7C
	.dc.b	$AF,$32,$3C,$DA,$E8,$C1,$0F,$F3,$A7,$CF,$7F,$39,$27,$C8,$FF,$9A
	.dc.b	$DF,$E7,$DE,$E3,$F3,$7F,$C0,$57,$F9,$EE,$31,$AF,$FA,$2D,$B1,$99
	.dc.b	$DB,$FE,$81,$FA,$57,$CF,$7F,$9D,$DB,$1F,$E9,$36,$BB,$8D,$11,$D3
	.dc.b	$5E,$9F,$2F,$CC,$8F,$37,$FC,$DF,$FB,$FF,$F3,$08,$FF,$4D,$FB,$88
	.dc.b	$C4,$7F,$A8,$63,$1C,$5C,$0F,$37,$93,$27,$CC,$D7,$FD,$45,$9F,$C5
	.dc.b	$FB,$2F,$85,$A1,$A4,$9F,$88,$D7,$8C,$6D,$C9,$0C,$C6,$DF,$F5,$3C
	.dc.b	$6D,$FF,$56,$7D,$17,$F5,$B6,$F8,$9F,$EA,$38,$C4,$7F,$AD,$DA,$FB
	.dc.b	$FD,$6F,$F1,$EB,$FE,$97,$6B,$FE,$90,$6B,$F2,$1A,$FF,$B1,$21,$F4
	.dc.b	$FD,$C9,$1F,$EC,$D8,$CB,$63,$2F,$A1,$AF,$B3,$E6,$7E,$49,$1F,$C5
	.dc.b	$3E,$BF,$A7,$91,$D5,$AF,$C2,$6B,$EF,$6B,$E7,$7F,$DB,$7F,$06,$F8
	.dc.b	$8F,$C0,$7D,$57,$F2,$6F,$87,$FE,$99,$C4,$8C,$78,$FC,$71,$FF,$73
	.dc.b	$61,$FE,$EA,$D1,$97,$4C,$B9,$F8,$3F,$EE,$8C,$F9,$E5,$CB,$1F,$F7
	.dc.b	$76,$1C,$1E,$77,$F2,$FC,$BF,$F7,$7B,$F5,$DB,$19,$3F,$DE,$D1,$96
	.dc.b	$99,$3F,$DF,$A2,$53,$CA,$3C,$2D,$74,$D8,$DC,$7F,$BF,$33,$FD,$27
	.dc.b	$ED,$0F,$8E,$F6,$46,$1B,$3D,$0E,$07,$BB,$EA,$DF,$53,$88,$F2,$46
	.dc.b	$16,$9F,$DE,$DC,$CF,$2F,$AD,$7D,$B7,$FB,$3A,$B5,$9B,$0E,$51,$86
	.dc.b	$E5,$F9,$56,$E2,$C3,$9F,$C4,$FA,$23,$B9,$FA,$A9,$F2,$3F,$A7,$6B
	.dc.b	$CF,$2F,$CE,$79,$EE,$BA,$49,$FD,$9C,$65,$B1,$F0,$73,$4E,$73,$FF
	.dc.b	$0D,$64,$FA,$EF,$F4,$FF,$0E,$EB,$5A,$FE,$BD,$B9,$7E,$29,$B7,$C7
	.dc.b	$E0,$7F,$39,$3F,$32,$B8,$67,$F4,$7C,$67,$A3,$AD,$3F,$3F,$C7,$FD
	.dc.b	$EF,$4D,$4E,$70,$FC,$0F,$8D,$72,$0C,$79,$DF,$E4,$A3,$31,$87,$F2
	.dc.b	$6D,$8E,$31,$AB,$F3,$C6,$23,$A4,$61,$DC,$F1,$6C,$65,$E5,$3E,$17
	.dc.b	$C9,$F0,$1F,$0B,$AB,$E1,$F5,$FE,$EF,$D0,$F5,$56,$AF,$F6,$2A,$75
	.dc.b	$1F,$1E,$F9,$3F,$C1,$AB,$8E,$3F,$ED,$18,$E3,$5C,$2F,$E8,$1F,$0A
	.dc.b	$FF,$66,$D9,$F7,$34,$B5,$CD,$B3,$C1,$F3,$18,$9E,$2E,$0F,$EB,$7F
	.dc.b	$9A,$3F,$05,$EF,$82,$31,$5E,$08,$C4,$EA,$73,$5E,$2F,$B5,$F0,$62
	.dc.b	$69,$30,$F0,$C6,$5B,$19,$9E,$51,$DF,$CD,$FC,$DC,$7D,$DC,$51,$88
	.dc.b	$FE,$E6,$D7,$FB,$93,$5F,$C5,$FA,$DE,$42,$57,$89,$F7,$FE,$2E,$F9
	.dc.b	$0A,$49,$F8,$11,$96,$C6,$6B,$C9,$39,$6C,$65,$FF,$03,$C6,$5B,$19
	.dc.b	$7F,$2B,$F1,$BF,$20,$95,$F1,$9F,$A7,$F6,$8F,$6A,$92,$7E,$A4,$65
	.dc.b	$FB,$BF,$61,$F5,$2F,$F4,$8D,$7F,$52,$D7,$F6,$6D,$4B,$E0,$BF,$14
	.dc.b	$D9,$EB,$18,$23,$82,$4F,$E5,$A7,$83,$F6,$7F,$4D,$7B,$5F,$8E,$6D
	.dc.b	$7E,$35,$B3,$B9,$F0,$FC,$2C,$AF,$6B,$67,$F8,$07,$D6,$3E,$59,$E8
	.dc.b	$7E,$A2,$30,$D8,$CD,$FE,$AA,$72,$D8,$CD,$74,$6B,$FC,$42,$76,$E6
	.dc.b	$BA,$7D,$84,$BE,$4D,$9E,$4D,$79,$5A,$31,$1F,$C9,$3C,$17,$E2,$DF
	.dc.b	$ED,$DB,$FF,$33,$36,$7F,$C8,$ED,$7F,$91,$CF,$0F,$03,$F8,$07,$D7
	.dc.b	$BE,$32,$93,$0F,$E0,$31,$96,$C6,$6F,$F8,$6C,$E5,$B1,$F2,$49,$3F
	.dc.b	$1D,$B1,$F1,$DA,$FB,$9A,$FF,$40,$D3,$EB,$79,$FF,$4C,$95,$F4,$5B
	.dc.b	$3F,$45,$AF,$C8,$7F,$91,$F8,$2F,$CA,$9E,$B6,$1B,$7F,$C9,$A7,$53
	.dc.b	$93,$F9,$27,$E5,$4E,$7A,$4A,$FE,$4B,$19,$6C,$67,$1F,$E6,$33,$96
	.dc.b	$C7,$E3,$7E,$D6,$57,$E2,$DB,$3F,$8B,$6B,$F8,$26,$BF,$17,$89,$F5
	.dc.b	$3F,$97,$25,$7E,$55,$B3,$F9,$56,$BF,$8C,$7F,$A3,$F0,$BE,$D7,$F3
	.dc.b	$33,$F8,$D6,$BF,$E1,$16,$C6,$F9,$7E,$2F,$CC,$4C,$3E,$1B,$6F,$E0
	.dc.b	$B5,$FE,$E1,$A2,$31,$1F,$7C,$7D,$CF,$EC,$DB,$5F,$B1,$69,$FB,$A2
	.dc.b	$9F,$AF,$29,$CE,$39,$12,$39,$7D,$D7,$52,$38,$24,$F5,$8C,$B6,$3F
	.dc.b	$B5,$FD,$3B,$F0,$67,$BD,$FD,$C2,$30,$D8,$CC,$F0,$6D,$7F,$74,$FB
	.dc.b	$87,$D5,$7F,$71,$6D,$7F,$70,$6C,$EE,$74,$49,$D3,$63,$FB,$87,$07
	.dc.b	$F7,$BF,$27,$F0,$B5,$FD,$E9,$B3,$A8,$FF,$75,$79,$1F,$E0,$1E,$13
	.dc.b	$49,$1F,$E0,$2F,$AE,$7C,$D7,$F8,$1D,$B1,$98,$C1,$9F,$87,$91,$C5
	.dc.b	$B8,$7F,$85,$29,$E1,$F3,$25,$77,$6C,$ED,$AF,$F8,$E6,$78,$CF,$F8
	.dc.b	$78,$7E,$72,$F8,$FF,$F1,$1C,$61,$B1,$98,$E0,$DA,$FE,$7D,$C7,$FE
	.dc.b	$23,$F0,$0D,$25,$7F,$89,$23,$2D,$8C,$CF,$28,$F2,$7D,$EF,$C6,$8F
	.dc.b	$F2,$84,$23,$11,$FE,$B2,$6B,$FE,$B0,$6B,$FE,$B2,$F9,$8F,$F1,$B4
	.dc.b	$9A,$47,$FC,$6E,$64,$AE,$63,$E3,$3F,$0B,$E5,$A4,$FC,$26,$CF,$F9
	.dc.b	$23,$EB,$D1,$FE,$45,$6C,$FF,$91,$1B,$3B,$9D,$12,$74,$D8,$F8,$39
	.dc.b	$3C,$FF,$A6,$4A,$FA,$6F,$F9,$4F,$E0,$BF,$88,$D7,$FA,$F3,$BB,$04
	.dc.b	$AF,$F2,$AC,$E8,$91,$99,$FF,$3F,$72,$3F,$51,$7C,$8F,$F2,$06,$E1
	.dc.b	$FE,$66,$9D,$4E,$4F,$F3,$27,$6F,$DE,$BF,$E0,$3C,$18,$25,$FF,$99
	.dc.b	$A7,$53,$FE,$8D,$6B,$FE,$92,$6B,$FE,$9C,$6F,$8C,$FC,$E2,$92,$FF
	.dc.b	$AC,$BC,$29,$4F,$A4,$F2,$93,$0F,$13,$E8,$DA,$47,$95,$B5,$FC,$69
	.dc.b	$AF,$9D,$AF,$FA,$C3,$AD,$FB,$1F,$FB,$CE,$BF,$CF,$8D,$9F,$F3,$73
	.dc.b	$5F,$F5,$7C,$F0,$F8,$BF,$F4,$2C,$E6,$7A,$35,$FF,$51,$36,$78,$57
	.dc.b	$FA,$27,$EA,$FF,$C7,$B5,$D1,$AF,$95,$AF,$FA,$AD,$B2,$69,$3D,$3F
	.dc.b	$F4,$BB,$C2,$7F,$D9,$35,$FE,$9E,$7F,$D9,$CD,$F7,$3F,$D4,$91,$83
	.dc.b	$FD,$19,$3F,$B9,$9F,$37,$4F,$42,$38,$24,$F0,$8C,$BD,$FE,$13,$C0
	.dc.b	$9A,$47,$FD,$73,$F5,$7F,$BF,$F3,$3F,$D6,$6D,$8C,$F6,$7F,$96,$DB
	.dc.b	$3A,$F8,$47,$2D,$24,$F4,$8C,$BE,$CF,$73,$FA,$36,$9F,$E9,$D8,$FA
	.dc.b	$51,$82,$38,$21,$C4,$F2,$F8,$5F,$2F,$2F,$E7,$B1,$82,$1F,$EB,$5F
	.dc.b	$AB,$7C,$6F,$F5,$56,$C7,$FB,$25,$AE,$E3,$44,$74,$D7,$FD,$81,$F5
	.dc.b	$7F,$C3,$5F,$F6,$93,$5C,$F3,$7F,$63,$8C,$47,$E6,$1A,$FE,$85,$AF
	.dc.b	$FB,$DD,$BC,$87,$30,$FA,$B7,$CB,$FE,$2D,$B1,$FE,$EA,$6B,$FE,$F4
	.dc.b	$6A,$21,$CE,$39,$11,$E4,$FE,$5F,$EA,$FF,$D3,$31,$FE,$F6,$6B,$FF
	.dc.b	$00,$1A,$F4,$0E,$1E,$FF,$53,$FE,$33,$7F,$DF,$ED,$78,$07,$0F,$36
	.dc.b	$BE,$88,$7D,$47,$94,$91,$E5,$6B,$E5,$6B,$E5,$8E,$2F,$8B,$C9,$FD
	.dc.b	$A5,$7F,$7B,$8C,$10,$F2,$FD,$5F,$F2,$37,$CE,$1C,$3A,$EE,$FD,$F4
	.dc.b	$62,$3E,$23,$5F,$A2,$D7,$EA,$35,$E9,$1E,$4F,$AB,$91,$F2,$1A,$FC
	.dc.b	$8F,$90,$FF,$22,$BE,$46,$B9,$F5,$7F,$88,$46,$23,$EB,$35,$FC,$1B
	.dc.b	$5F,$C3,$35,$37,$07,$C6,$FC,$82,$4F,$DC,$7F,$8F,$7D,$5F,$F8,$EA
	.dc.b	$3E,$3B,$5F,$BE,$D7,$F1,$0D,$7F,$2A,$D7,$F3,$0D,$4D,$31,$B7,$FB
	.dc.b	$36,$3E,$AB,$5F,$C6,$B5,$FC,$B3,$5F,$D1,$F3,$7E,$11,$FC,$9B,$5F
	.dc.b	$CE,$35,$DE,$FF,$CE,$E3,$11,$FA,$D6,$BF,$E5,$08,$D1,$FD,$07,$EA
	.dc.b	$DF,$13,$49,$3F,$8C,$8C,$BF,$8E,$FA,$B7,$87,$FD,$25,$B1,$98,$C1
	.dc.b	$1C,$10,$FD,$6C,$70,$7F,$5F,$F5,$7F,$E8,$47,$F3,$6D,$73,$EC,$FF
	.dc.b	$57,$8C,$47,$EF,$5F,$F2,$23,$C3,$C2,$D1,$1F,$D2,$FD,$5F,$FB,$89
	.dc.b	$BF,$07,$FD,$86,$31,$1F,$C0,$1A,$FF,$09,$5F,$1B,$B2,$47,$47,$FB
	.dc.b	$57,$CB,$BD,$AA,$49,$FD,$AC,$65,$B1,$97,$F8,$7C,$65,$B1,$FD,$D7
	.dc.b	$E1,$FF,$A1,$C7,$F1,$B6,$BF,$CA,$1A,$FF,$2B,$6A,$FA,$BF,$F2,$DB
	.dc.b	$FC,$19,$AF,$F1,$E6,$BF,$C9,$DA,$FF,$34,$6B,$FC,$DD,$AF,$5F,$67
	.dc.b	$FC,$15,$18,$3F,$D5,$71,$DF,$DE,$C9,$1C,$10,$F1,$39,$7E,$1F,$C5
	.dc.b	$FC,$C2,$3F,$29,$AE,$8C,$10,$F6,$7C,$B7,$90,$8F,$C1,$7F,$C4,$1F
	.dc.b	$57,$F7,$23,$F9,$4B,$5F,$E8,$2D,$7F,$A5,$35,$FE,$AE,$D7,$FA,$D3
	.dc.b	$53,$49,$F2,$FE,$02,$47,$81,$AF,$81,$AF,$81,$FF,$1B,$7D,$5B,$E8
	.dc.b	$29,$27,$FA,$0C,$65,$B1,$97,$FA,$DC,$65,$B1,$FE,$41,$FA,$B7,$D8
	.dc.b	$D2,$4F,$F4,$C8,$CB,$7D,$1E,$04,$70,$43,$FB,$4C,$23,$51,$87,$C1
	.dc.b	$C1,$23,$B3,$5E,$2D,$7F,$BD,$4F,$E9,$5B,$1F,$E5,$9F,$AC,$7B,$14
	.dc.b	$93,$FD,$92,$32,$D8,$CB,$FD,$FA,$32,$D8,$FF,$32,$79,$3F,$DF,$A3
	.dc.b	$FC,$2C,$D7,$FD,$DD,$F1,$9A,$37,$F5,$92,$3F,$B9,$CA,$31,$1F,$E0
	.dc.b	$C8,$C3,$A8,$CB,$FE,$17,$8C,$BB,$F9,$AF,$19,$1E,$AD,$74,$60,$87
	.dc.b	$F8,$6B,$EB,$3F,$97,$3F,$E2,$36,$BF,$E3,$06,$BF,$E3,$36,$FC,$1B
	.dc.b	$EA,$3E,$D4,$FF,$B1,$DA,$FF,$B3,$9A,$FF,$8E,$A3,$86,$DF,$E8,$DF
	.dc.b	$AC,$FF,$68,$CF,$F8,$59,$AF,$F8,$B9,$AF,$F8,$DD,$AF,$F9,$11,$AF
	.dc.b	$F9,$1D,$A9,$A4,$FA,$CF,$01,$23,$C0,$FF,$89,$BD,$57,$D6,$7E,$2E
	.dc.b	$7F,$D8,$0D,$51,$AE,$1E,$13,$99,$F0,$79,$7E,$D2,$3E,$D7,$FC,$67
	.dc.b	$97,$C9,$FA,$CB,$A1,$A4,$AF,$98,$D7,$E6,$35,$F9,$91,$C5,$D9,$0D
	.dc.b	$9F,$E4,$2F,$A7,$FE,$B2,$FF,$9B,$1A,$FF,$9B,$DA,$9C,$E4,$F8,$B7
	.dc.b	$A9,$F1,$DB,$5F,$E6,$49,$D4,$E4,$FF,$65,$53,$EB,$29,$2B,$C8,$D7
	.dc.b	$FC,$DB,$1B,$72,$43,$F1,$9E,$AF,$C8,$23,$F9,$17,$BF,$D6,$48,$F8
	.dc.b	$0D,$7A,$B5,$FF,$41,$4A,$31,$1F,$15,$FF,$48,$FC,$13,$E1,$BF,$D2
	.dc.b	$99,$3E,$B3,$F9,$47,$C3,$3C,$5C,$C6,$1E,$58,$FF,$A6,$F0,$E4,$FC
	.dc.b	$FC,$7A,$E3,$C3,$C1,$FF,$4F,$E3,$CE,$BF,$1F,$8F,$37,$81,$CE,$BE
	.dc.b	$D7,$93,$FE,$A6,$7D,$6D,$8C,$9F,$EA,$B8,$CB,$4C,$9E,$43,$F0,$51
	.dc.b	$C5,$A6,$CA,$6C,$FC,$33,$E8,$CB,$E2,$FD,$43,$E2,$B1,$5F,$A8,$8C
	.dc.b	$3F,$EB,$DC,$7D,$CF,$23,$15,$F9,$5F,$AB,$7C,$8E,$27,$F2,$D1,$85
	.dc.b	$5C,$B8,$13,$98,$F9,$5F,$54,$FA,$CF,$F7,$4C,$6E,$75,$8F,$7C,$7B
	.dc.b	$63,$FA,$19,$CA,$C3,$97,$C3,$FB,$23,$55,$FA,$DA,$F2,$47,$E3,$1A
	.dc.b	$F5,$3F,$3B,$E8,$BA,$FF,$68,$D8,$CC,$60,$8F,$BB,$1F,$DC,$4E,$F2
	.dc.b	$F6,$FD,$67,$FA,$83,$87,$D3,$B2,$B5,$7C,$31,$CC,$63,$1F,$66,$1C
	.dc.b	$5E,$C7,$F0,$38,$D3,$FE,$7C,$F3,$FF,$33,$97,$E5,$79,$DC,$3F,$DE
	.dc.b	$F4,$F9,$3E,$D2,$F7,$A7,$E1,$3C,$4B,$97,$FB,$F6,$B5,$7F,$8D,$8E
	.dc.b	$CD,$8C,$C7,$E3,$DB,$1F,$EF,$96,$C0,$E9,$7F,$AE,$9D,$36,$34,$76
	.dc.b	$3A,$7C,$07,$ED,$3A,$BF,$6B,$9F,$EF,$7D,$0F,$3D,$7F,$A7,$2B,$58
	.dc.b	$7E,$FA,$73,$8F,$FB,$3A,$3F,$69,$5F,$B8,$9C,$D7,$08,$C4,$FC,$68
	.dc.b	$FD,$54,$7C,$36,$BD,$E3,$FD,$F3,$5D,$63,$D2,$6E,$33,$3C,$27,$53
	.dc.b	$F1,$63,$AD,$BE,$AD,$F5,$98,$6E,$1F,$CE,$5B,$7D,$AB,$85,$75,$FA
	.dc.b	$C9,$7E,$46,$D7,$90,$FF,$60,$B8,$7E,$B3,$5F,$E5,$3E,$5F,$E3,$DF
	.dc.b	$A7,$68,$C4,$6B,$E1,$7D,$31,$88,$FC,$47,$AD,$EF,$26,$1F,$19,$F7
	.dc.b	$FD,$63,$E3,$E9,$27,$FA,$B4,$65,$B1,$9A,$F8,$13,$96,$C6,$5F,$3C
	.dc.b	$65,$B1,$97,$F1,$BB,$3E,$37,$EA,$12,$FE,$DB,$F5,$3E,$97,$FD,$93
	.dc.b	$1F,$DA,$5B,$5F,$45,$AF,$C6,$6B,$F9,$AF,$AC,$FA,$EF,$E1,$1A,$E7
	.dc.b	$7F,$C5,$C6,$23,$F4,$CF,$FB,$F2,$BE,$C4,$23,$0F,$DB,$6B,$D7,$E1
	.dc.b	$B9,$92,$78,$CE,$A7,$C9,$3F,$90,$FA,$CF,$F3,$8D,$7F,$81,$5B,$3F
	.dc.b	$8D,$6B,$F8,$26,$BE,$DF,$89,$F4,$92,$7B,$BF,$A8,$F9,$E7,$CA,$FF
	.dc.b	$85,$67,$F2,$51,$8F,$3F,$53,$9B,$FD,$4C,$E6,$B7,$39,$9F,$E0,$B3
	.dc.b	$97,$71,$97,$F8,$74,$65,$DC,$3E,$B3,$B9,$2B,$AB,$67,$9B,$5E,$2F
	.dc.b	$F0,$2F,$09,$F5,$9F,$E0,$BA,$FD,$64,$63,$BF,$53,$9B,$FE,$07,$39
	.dc.b	$AF,$B7,$F5,$8F,$AC,$FE,$16,$D8,$CC,$60,$8E,$09,$5F,$BB,$9E,$0F
	.dc.b	$C2,$9F,$96,$49,$F9,$2D,$8D,$46,$08,$FC,$AF,$37,$E9,$92,$3E,$8B
	.dc.b	$63,$51,$82,$3F,$C8,$3C,$07,$D6,$3E,$2B,$F9,$4C,$61,$B1,$9A,$FE
	.dc.b	$29,$39,$7F,$3F,$F5,$4E,$7F,$E5,$ED,$8C,$C6,$08,$E0,$98,$7F,$1F
	.dc.b	$9E,$0F,$E1,$BE,$B3,$F1,$E4,$9F,$C6,$B6,$35,$18,$23,$F8,$EF,$A8
	.dc.b	$FC,$B9,$23,$F2,$AD,$8D,$46,$08,$FE,$0B,$EB,$3F,$2C,$FF,$83,$5B
	.dc.b	$B9,$A6,$A1,$83,$E1,$3E,$19,$2F,$E3,$B6,$7E,$03,$5F,$EA,$73,$C6
	.dc.b	$7F,$63,$0F,$AC,$FD,$B3,$63,$31,$82,$38,$24,$7E,$D8,$E0,$5F,$A4
	.dc.b	$E6,$48,$EE,$D8,$D4,$60,$8F,$F6,$8F,$A3,$7D,$63,$DC,$FE,$E3,$18
	.dc.b	$6C,$66,$BF,$AE,$4E,$5F,$E4,$FF,$58,$F9,$6F,$EE,$ED,$8C,$C6,$08
	.dc.b	$E0,$98,$7F,$67,$9E,$0F,$FA,$A7,$7F,$F0,$0F,$94,$69,$2B,$FD,$93
	.dc.b	$E4,$7F,$81,$B8,$4D,$24,$7F,$3C,$FA,$B7,$D6,$7F,$83,$1B,$19,$8C
	.dc.b	$19,$F8,$79,$1C,$5B,$97,$F7,$DA,$78,$7C,$C9,$5D,$DB,$3B,$6B,$FE
	.dc.b	$18,$9E,$33,$FE,$07,$1F,$56,$F8,$CF,$F1,$1C,$61,$B1,$98,$E0,$DA
	.dc.b	$FE,$7F,$C7,$FE,$28,$F1,$0D,$25,$7F,$5E,$8C,$B6,$33,$3C,$A3,$C9
	.dc.b	$F4,$FF,$1A,$3F,$C5,$F0,$8C,$47,$FA,$BD,$AF,$FA,$B5,$AF,$FA,$BF
	.dc.b	$E6,$7F,$C7,$72,$69,$1F,$EE,$26,$4A,$E6,$3E,$33,$F0,$BE,$5E,$4F
	.dc.b	$C2,$6C,$FF,$8C,$7E,$B1,$1F,$E4,$A6,$CF,$F7,$E6,$CE,$E7,$44,$9D
	.dc.b	$36,$3E,$0E,$5F,$E1,$B9,$FA,$3E,$77,$D4,$23,$FE,$0A,$F8,$0F,$E1
	.dc.b	$84,$F2,$FF,$CB,$6D,$8D,$46,$08,$FF,$5E,$FA,$C7,$D6,$7F,$63,$6C
	.dc.b	$66,$30,$7F,$99,$BA,$FE,$DE,$5F,$B6,$DF,$04,$AF,$F0,$EC,$EA,$7F
	.dc.b	$81,$4E,$5E,$13,$FD,$1B,$EB,$35,$1F,$EF,$76,$AF,$15,$18,$6D,$78
	.dc.b	$FE,$92,$5F,$81,$B7,$FC,$2B,$5F,$41,$27,$CA,$D8,$D4,$60,$8F,$F9
	.dc.b	$CF,$EB,$3F,$DD,$91,$FE,$6F,$ED,$46,$27,$5F,$13,$FE,$80,$8C,$47
	.dc.b	$F9,$B9,$AF,$FA,$9E,$78,$CE,$C9,$5D,$7E,$B3,$FC,$67,$4A,$46,$1B
	.dc.b	$F3,$1F,$E8,$E8,$C1,$1C,$12,$BF,$C9,$F4,$F6,$3F,$D3,$ED,$8C,$C6
	.dc.b	$0E,$9D,$12,$3C,$8D,$9F,$F4,$B5,$FE,$D6,$72,$48,$CC,$F6,$5F,$09
	.dc.b	$E2,$69,$E4,$70,$7F,$99,$3E,$B3,$F7,$1E,$B5,$EB,$FF,$27,$D7,$FA
	.dc.b	$7E,$72,$D8,$CB,$FE,$B6,$87,$C2,$7E,$26,$38,$35,$D3,$5F,$F6,$AC
	.dc.b	$70,$E3,$FB,$DF,$82,$7E,$F2,$7F,$D9,$BB,$A3,$11,$F9,$5D,$3E,$E4
	.dc.b	$62,$BF,$DB,$7A,$FF,$A8,$EB,$8C,$E6,$B7,$D9,$FE,$E4,$BF,$F7,$0C
	.dc.b	$E6,$BF,$01,$7C,$5E,$C7,$E0,$FE,$B5,$F1,$9F,$E8,$AC,$91,$89,$FC
	.dc.b	$B3,$5F,$F6,$63,$5F,$F6,$0D,$F5,$C3,$C3,$EF,$F8,$F3,$F5,$B6,$BC
	.dc.b	$18,$DC,$7E,$14,$DC,$39,$5F,$0B,$F8,$F8,$F8,$71,$4A,$5F,$69,$FE
	.dc.b	$8A,$E2,$EC,$D8,$FF,$7D,$B5,$F8,$39,$3E,$BB,$FD,$7D,$1A,$E2,$FE
	.dc.b	$8C,$DB,$E7,$18,$8E,$33,$C1,$F9,$63,$EB,$BF,$B6,$BE,$5C,$BC,$EF
	.dc.b	$9E,$32,$F0,$6C,$6A,$39,$7C,$CF,$C7,$7A,$79,$16,$B7,$A7,$C1,$9C
	.dc.b	$9D,$DC,$42,$B9,$7D,$67,$F9,$16,$BF,$D8,$ED,$7F,$E0,$03,$5E,$81
	.dc.b	$C3,$FC,$3F,$E3,$1F,$19,$F6,$5B,$1F,$9C,$6B,$F1,$27,$B6,$5E,$D7
	.dc.b	$FC,$0D,$1E,$2F,$2B,$F8,$5C,$62,$3C,$71,$82,$3B,$21,$E6,$E0,$79
	.dc.b	$5F,$E4,$38,$FB,$ED,$7F,$3F,$1A,$F5,$3E,$1D,$7C,$4F,$98,$F0,$92
	.dc.b	$BE,$2B,$5F,$8A,$D7,$E2,$E0,$8F,$F7,$8D,$FE,$4B,$E5,$FF,$22,$9A
	.dc.b	$77,$AF,$AC,$7E,$66,$9F,$1D,$F0,$89,$1F,$59,$AF,$D6,$6B,$E2,$7C
	.dc.b	$34,$F9,$8F,$09,$27,$B9,$F6,$E7,$F1,$9B,$FF,$B3,$63,$EF,$9F,$AB
	.dc.b	$70,$E1,$AF,$E1,$1A,$FE,$45,$AF,$1F,$82,$FE,$8B,$18,$8F,$C9,$47
	.dc.b	$C6,$FB,$67,$89,$FB,$66,$CF,$ED,$9A,$FE,$B2,$BC,$7C,$1F,$A5,$FA
	.dc.b	$B7,$C5,$7F,$52,$6D,$7A,$83,$87,$DA,$D7,$F3,$AD,$7F,$54,$D7,$F8
	.dc.b	$63,$5E,$71,$FE,$51,$E6,$7A,$3F,$C6,$F9,$3F,$89,$B6,$33,$5F,$42
	.dc.b	$FC,$58,$7E,$F7,$EA,$BF,$D2,$13,$E7,$6B,$F4,$9A,$FE,$C9,$AF,$EF
	.dc.b	$DA,$FF,$21,$6A,$69,$8F,$CF,$BF,$1B,$1F,$82,$6B,$FC,$61,$B3,$E5
	.dc.b	$F7,$7F,$88,$7D,$5F,$FB,$D6,$BF,$12,$D7,$F8,$FB,$67,$D9,$E4,$F0
	.dc.b	$E6,$FF,$02,$CF,$E8,$6B,$EA,$35,$FC,$7B,$5F,$E1,$FF,$A5,$7E,$65
	.dc.b	$FC,$BB,$5F,$E5,$AD,$BF,$E4,$D2,$FA,$CF,$E4,$DF,$15,$FD,$4E,$FF
	.dc.b	$92,$35,$FE,$8C,$D4,$D2,$79,$1F,$E1,$C6,$C7,$F9,$CF,$F1,$27,$CA
	.dc.b	$FF,$59,$F7,$11,$89,$FD,$DC,$62,$7C,$77,$FD,$4B,$E0,$DF,$41,$49
	.dc.b	$7F,$B4,$FA,$DF,$EE,$DD,$EA,$DC,$FF,$58,$6B,$FC,$69,$BB,$7F,$47
	.dc.b	$AE,$3F,$56,$FB,$4F,$EC,$2D,$BF,$EC,$2D,$7F,$AB,$CF,$FB,$41,$CC
	.dc.b	$F0,$23,$E5,$F8,$5F,$EA,$52,$F1,$E9,$27,$F8,$6F,$CE,$BE,$53,$11
	.dc.b	$DA,$6E,$5F,$D7,$6B,$53,$FD,$D3,$F0,$1F,$ED,$93,$FD,$9B,$54,$63
	.dc.b	$B7,$FB,$F4,$63,$D0,$E7,$18,$DB,$FC,$0F,$87,$F8,$09,$B4,$73,$0F
	.dc.b	$AF,$FC,$43,$F1,$3D,$5F,$8E,$47,$F9,$3C,$69,$FE,$93,$F4,$3F,$58
	.dc.b	$9E,$C7,$F7,$59,$FE,$BF,$A7,$A6,$73,$5F,$E1,$07,$FC,$64,$F8,$7D
	.dc.b	$0F,$EF,$9F,$5D,$E5,$8F,$E3,$0D,$7F,$99,$B5,$FE,$E6,$D7,$FC,$00
	.dc.b	$D7,$FC,$44,$D4,$D3,$54,$FC,$57,$FC,$55,$AF,$F8,$55,$B5,$FE,$A5
	.dc.b	$6B,$CB,$4F,$F1,$5F,$CA,$3E,$FA,$92,$7F,$A6,$E3,$FE,$66,$7F,$C5
	.dc.b	$F3,$B6,$CE,$DF,$F1,$3C,$ED,$FE,$EB,$5F,$7D,$FB,$5C,$1F,$9A,$3E
	.dc.b	$3F,$C6,$D8,$FC,$DC,$57,$9C,$7F,$89,$FD,$4F,$F0,$BC,$7F,$91,$5A
	.dc.b	$9A,$4F,$94,$FF,$4D,$B6,$3F,$A5,$3C,$BE,$27,$FC,$79,$F8,$3F,$F9
	.dc.b	$65,$B3,$FD,$E9,$AF,$F9,$0A,$7F,$C3,$FB,$7E,$55,$FD,$99,$23,$FB
	.dc.b	$9A,$FB,$0C,$47,$C2,$8B,$C1,$FE,$43,$75,$1F,$E5,$AF,$B0,$FF,$20
	.dc.b	$BF,$E6,$06,$F3,$9C,$C7,$BF,$E8,$FF,$CC,$71,$A2,$7B,$BF,$1E,$3F
	.dc.b	$CE,$9F,$28,$FA,$CF,$A9,$7F,$E4,$D8,$C7,$C3,$FF,$6D,$6C,$7F,$82
	.dc.b	$1A,$FF,$82,$BE,$C3,$FE,$00,$BF,$E1,$06,$BF,$E6,$66,$E9,$FE,$64
	.dc.b	$9E,$2F,$F9,$9B,$E1,$1F,$A3,$FF,$B3,$6F,$FD,$4E,$D7,$FC,$F2,$D9
	.dc.b	$FF,$34,$5F,$4E,$0F,$F7,$24,$E5,$BF,$17,$FE,$8F,$8C,$5F,$FA,$26
	.dc.b	$B8,$3F,$EE,$5F,$B0,$73,$7F,$A8,$DB,$1F,$EF,$76,$BF,$E6,$6C,$91
	.dc.b	$88,$FF,$3A,$72,$3F,$1E,$DE,$9F,$E8,$28,$C6,$1F,$8B,$9C,$E9,$FE
	.dc.b	$FB,$C2,$E0,$D7,$EA,$7D,$8F,$D1,$E0,$34,$93,$FE,$3A,$B7,$E3,$2F
	.dc.b	$98,$FF,$58,$46,$1B,$19,$9F,$F2,$9E,$1D,$9F,$F2,$AD,$6D,$B3,$B7
	.dc.b	$FD,$93,$19,$7F,$D9,$73,$C5,$CC,$EF,$5F,$F5,$0D,$6D,$B1,$98,$FF
	.dc.b	$5D,$57,$19,$FC,$15,$FF,$9B,$6F,$FC,$AB,$B7,$AF,$E1,$FF,$DA,$D3
	.dc.b	$99,$FA,$4D,$7F,$D4,$9C,$FF,$A3,$F8,$1F,$D0,$92,$3F,$DE,$33,$C7
	.dc.b	$0F,$C9,$6B,$FE,$87,$9F,$D0,$35,$F2,$B5,$FA,$8D,$CB,$F5,$1F,$11
	.dc.b	$F5,$49,$3F,$A5,$9E,$1A,$7E,$FF,$CC,$FA,$7D,$9F,$D8,$AB,$E5,$B7
	.dc.b	$1F,$1E,$1F,$EC,$1F,$1B,$FC,$51,$1F,$A5,$FB,$2F,$F0,$1F,$C8,$FF
	.dc.b	$6A,$E1,$46,$32,$FF,$62,$5F,$0C,$EF,$B9,$FE,$EF,$AD,$68,$C3,$8F
	.dc.b	$A9,$FE,$CB,$CB,$C9,$8F,$1A,$D5,$F0,$BD,$C7,$FB,$C6,$BC,$53,$D2
	.dc.b	$15,$C6,$B7,$0A,$DD,$F7,$CB,$FD,$F5,$5A,$9F,$0E,$3F,$07,$D9,$D6
	.dc.b	$1E,$7F,$AE,$7C,$9E,$30,$46,$27,$C9,$8F,$C2,$BE,$31,$E4,$78,$4E
	.dc.b	$E3,$C9,$3E,$28,$C6,$3F,$0A,$B8,$E7,$FE,$48,$BE,$A7,$8F,$1E,$66
	.dc.b	$A7,$7F,$61,$FE,$82,$BF,$96,$DC,$7E,$AB,$6D,$F6,$2F,$B2,$7B,$7C
	.dc.b	$BE,$1F,$BD,$7C,$CB,$E0,$D8,$F1,$DF,$1A,$E1,$3F,$EB,$CD,$17,$FE
	.dc.b	$94,$8F,$F7,$EC,$62,$FF,$D5,$0D,$8E,$1E,$BF,$F1,$16,$F4,$6A,$31
	.dc.b	$1C,$9A,$FE,$41,$E1,$A7,$E2,$9B,$A2,$7D,$FF,$65,$FE,$6C,$E5,$F6
	.dc.b	$E2,$BD,$D6,$E3,$18,$EA,$73,$AB,$0E,$57,$F6,$8E,$13,$C9,$78,$DF
	.dc.b	$E5,$9C,$7E,$CF,$1F,$E2,$3C,$74,$7C,$E9,$D5,$73,$AD,$4E,$7D,$9D
	.dc.b	$CE,$79,$58,$63,$7F,$DE,$63,$C9,$FD,$B4,$EB,$0E,$47,$89,$C5,$FA
	.dc.b	$4F,$3E,$1C,$F7,$FC,$A6,$36,$7F,$67,$EA,$DC,$FF,$80,$B9,$8F,$7F
	.dc.b	$C7,$79,$BA,$FF,$84,$4E,$FA,$CD,$38,$6C,$AE,$1A,$FE,$D2,$B6,$FC
	.dc.b	$58,$CE,$1E,$28,$C1,$F7,$70,$EE,$E2,$11,$87,$DD,$7D,$E7,$81,$DF
	.dc.b	$0F,$72,$F8,$E7,$E1,$1E,$08,$C3,$E2,$CD,$F6,$6F,$B1,$D5,$F0,$BE
	.dc.b	$B9,$7F,$1F,$AD,$47,$F4,$C6,$C7,$F7,$B6,$BF,$97,$38,$57,$E0,$1C
	.dc.b	$5F,$82,$31,$1F,$C6,$A7,$8B,$63,$2F,$2A,$E8,$FF,$A7,$3D,$1F,$EF
	.dc.b	$71,$FE,$3D,$F2,$7D,$9E,$87,$E7,$36,$FF,$52,$56,$AF,$E5,$DF,$F2
	.dc.b	$8A,$FD,$0C,$79,$1A,$E7,$E0,$FC,$39,$A7,$33,$FA,$0D,$3F,$82,$DF
	.dc.b	$F8,$01,$B1,$D8,$FE,$8D,$F6,$4F,$7F,$E6,$F7,$CB,$E0,$FD,$39,$D4
	.dc.b	$60,$95,$FE,$08,$BE,$BA,$7F,$8F,$EB,$73,$9B,$E5,$87,$A2,$73,$3F
	.dc.b	$41,$AF,$F4,$86,$AF,$45,$E2,$BE,$D3,$F2,$F1,$87,$F1,$1F,$58,$F8
	.dc.b	$9C,$D6,$F5,$FC,$3E,$EA,$D4,$7F,$84,$9B,$8F,$F9,$05,$AF,$E1,$1A
	.dc.b	$FF,$94,$BE,$83,$F2,$DE,$B9,$FD,$4C,$3E,$CD,$F2,$9A,$99,$87,$F9
	.dc.b	$81,$DC,$EA,$30,$7E,$72,$B9,$56,$8F,$D2,$7D,$9B,$E5,$F1,$5E,$C8
	.dc.b	$C3,$FE,$2C,$6D,$7F,$96,$DA,$FE,$79,$AF,$E5,$FC,$0F,$E0,$3E,$AF
	.dc.b	$FA,$3D,$FE,$07,$F2,$CF,$9C,$FF,$28,$61,$FE,$97,$9D,$F5,$FF,$1E
	.dc.b	$E8,$FC,$DC,$70,$AD,$E3,$F8,$FC,$7A,$61,$FB,$87,$FD,$F5,$F5,$9E
	.dc.b	$0F,$84,$3C,$7E,$CF,$0C,$AE,$13,$A7,$F8,$0E,$9E,$5A,$D6,$1F,$9D
	.dc.b	$E8,$7C,$E7,$F5,$5B,$FF,$5E,$36,$7F,$CB,$BF,$66,$F2,$7F,$17,$39
	.dc.b	$DF,$FD,$35,$39,$BF,$F5,$23,$6F,$F7,$FE,$82,$31,$1F,$EA,$D6,$BF
	.dc.b	$E6,$26,$BF,$E8,$16,$C7,$1C,$3F,$85,$B6,$BF,$85,$D3,$66,$BF,$CB
	.dc.b	$63,$FD,$DB,$E4,$FC,$FF,$07,$E5,$32,$FE,$3B,$96,$9F,$47,$F1,$3F
	.dc.b	$F5,$1B,$7E,$0B,$EC,$7A,$FF,$82,$9C,$E1,$F6,$5B,$B7,$F6,$6C,$D1
	.dc.b	$8C,$3F,$C0,$5C,$5F,$87,$C3,$96,$3F,$B6,$FA,$4F,$E6,$78,$FF,$A9
	.dc.b	$DB,$7F,$D3,$7E,$C9,$F6,$7B,$F2,$FF,$A9,$DA,$73,$5F,$12,$31,$5F
	.dc.b	$CC,$9B,$19,$9D,$EB,$FC,$D2,$72,$D8,$CB,$FD,$0E,$32,$D8,$CB,$FC
	.dc.b	$EF,$83,$F3,$E7,$E7,$7E,$3A,$DF,$FB,$E6,$2C,$FF,$2D,$D7,$F5,$15
	.dc.b	$8C,$3E,$45,$7E,$D1,$BD,$BF,$A7,$FB,$0F,$D1,$33,$1F,$CF,$B9,$69
	.dc.b	$FA,$9F,$03,$F8,$BD,$7D,$36,$C7,$FA,$C3,$EC,$FF,$63,$39,$F1,$FF
	.dc.b	$BE,$F3,$A7,$35,$BF,$83,$FE,$33,$F6,$6F,$8E,$DD,$6B,$83,$FD,$B0
	.dc.b	$DC,$39,$5F,$0A,$DB,$FB,$BF,$22,$63,$FB,$6C,$3F,$70,$7F,$03,$DF
	.dc.b	$F9,$94,$7E,$F2,$77,$3F,$24,$DE,$9F,$22,$3F,$8A,$FD,$A3,$E6,$29
	.dc.b	$FC,$5A,$9F,$5B,$8C,$3F,$87,$56,$DB,$5B,$9D,$57,$F6,$3F,$9C,$7E
	.dc.b	$26,$FB,$6F,$F1,$1C,$61,$FE,$9F,$F6,$CF,$8A,$CD,$6F,$2F,$F0,$9F
	.dc.b	$8E,$AD,$47,$EE,$A3,$15,$FE,$08,$6C,$66,$77,$8F,$F8,$2A,$72,$D8
	.dc.b	$CB,$FE,$19,$8C,$B6,$32,$FF,$87,$7B,$7F,$43,$BF,$F8,$75,$B7,$FE
	.dc.b	$1E,$E9,$FE,$AF,$18,$AF,$F7,$BF,$87,$FE,$1C,$5C,$BF,$C3,$DF,$E3
	.dc.b	$F1,$98,$C4,$7F,$8A,$DF,$9C,$FF,$83,$4E,$3C,$1F,$D9,$3E,$D5,$F6
	.dc.b	$DF,$D9,$AB,$79,$FF,$8C,$7F,$33,$7C,$4F,$F6,$A6,$EB,$FD,$8F,$0F
	.dc.b	$E1,$D9,$7F,$A3,$E7,$2F,$78,$FE,$2E,$FF,$8C,$9A,$FF,$8F,$DA,$FF
	.dc.b	$A5,$DA,$9A,$4F,$13,$E1,$3C,$3C,$1F,$F4,$5F,$C8,$7F,$64,$C0,$FF
	.dc.b	$96,$7E,$DD,$F3,$19,$8C,$3F,$E4,$69,$4E,$9F,$E7,$A6,$23,$51,$9C
	.dc.b	$7F,$C9,$73,$93,$71,$98,$FF,$2D,$46,$5D,$C6,$5F,$F5,$A7,$94,$E3
	.dc.b	$FD,$D3,$7B,$BF,$CE,$1F,$35,$F5,$30,$FE,$E9,$5F,$BF,$51,$9B,$F1
	.dc.b	$B5,$F0,$B5,$FE,$DC,$D7,$FD,$62,$D8,$E7,$C5,$F0,$E3,$06,$F0,$7E
	.dc.b	$21,$F7,$9B,$BF,$CD,$FA,$0F,$F0,$AC,$FE,$0B,$0F,$97,$1F,$20,$8F
	.dc.b	$C6,$6E,$3F,$11,$AF,$C8,$D1,$1F,$E7,$AC,$3C,$39,$76,$3E,$E7,$DA
	.dc.b	$DC,$3F,$D1,$E7,$DA,$7E,$B9,$1F,$AA,$DC,$3E,$9B,$5F,$03,$DF,$E3
	.dc.b	$0F,$F8,$C1,$B3,$FE,$9C,$E6,$7D,$C3,$EE,$35,$7F,$CE,$A3,$15,$FE
	.dc.b	$06,$AD,$D7,$F8,$2A,$73,$87,$F8,$59,$AF,$ED,$DA,$FF,$BE,$67,$8D
	.dc.b	$73,$9C,$CF,$F0,$26,$E4,$C7,$FA,$35,$7F,$8E,$DB,$5F,$E8,$DE,$87
	.dc.b	$DC,$7F,$34,$AD,$78,$FF,$DF,$23,$15,$FD,$D5,$AF,$F7,$36,$BF,$ED
	.dc.b	$16,$A3,$FD,$09,$F1,$9F,$E7,$0F,$00,$DD,$B0,$CC,$3F,$CD,$93,$A6
	.dc.b	$CE,$6B,$FC,$AD,$3F,$C3,$31,$FF,$0F,$36,$BC,$1F,$02,$FB,$87,$CB
	.dc.b	$FF,$9A,$2B,$58,$7F,$8D,$DB,$B7,$FA,$67,$BF,$FD,$B9,$7C,$31,$FF
	.dc.b	$31,$57,$0C,$5F,$6F,$FD,$93,$1F,$F5,$B4,$63,$CB,$FF,$72,$7D,$C7
	.dc.b	$F8,$27,$0F,$1D,$EF,$54,$62,$BF,$DD,$53,$98,$F4,$D7,$1C,$18,$75
	.dc.b	$9F,$F5,$2C,$FF,$A1,$27,$2D,$93,$4C,$7E,$05,$E0,$AF,$F5,$6B,$6B
	.dc.b	$D3,$3F,$C2,$A7,$27,$FB,$F6,$75,$8F,$A2,$BC,$71,$89,$F3,$F6,$6D
	.dc.b	$FF,$56,$CF,$E0,$79,$3F,$CC,$AF,$FB,$ED,$AE,$4A,$EB,$D7,$7D,$87
	.dc.b	$FA,$6E,$7F,$CD,$B1,$8F,$81,$D4,$E6,$7F,$DF,$73,$97,$7F,$04,$CF
	.dc.b	$A3,$63,$31,$83,$FC,$F3,$AF,$DC,$E8,$FF,$27,$67,$F6,$1B,$5F,$4B
	.dc.b	$7F,$F1,$83,$FE,$D6,$FC,$E7,$FD,$28,$FF,$B6,$23,$2F,$E1,$1B,$87
	.dc.b	$C9,$C3,$B4,$F1,$F7,$FA,$C2,$FF,$2F,$3E,$46,$BF,$AB,$3F,$3A,$E1
	.dc.b	$C4,$71,$E7,$F9,$EB,$85,$F1,$BD,$9B,$5F,$82,$8F,$0D,$F6,$C7,$CD
	.dc.b	$5A,$AF,$CE,$B6,$33,$5C,$30,$F8,$BD,$1F,$ED,$1F,$1A,$F1,$7F,$6E
	.dc.b	$C7,$FD,$94,$D7,$FD,$9C,$FF,$BC,$DC,$C7,$02,$3C,$3E,$23,$AF,$CC
	.dc.b	$BE,$76,$92,$7F,$DA,$D0,$F8,$3C,$47,$FB,$6E,$38,$3A,$9E,$73,$E8
	.dc.b	$FC,$43,$FD,$ED,$3F,$EB,$AC,$91,$88,$FA,$07,$FA,$F5,$C3,$F4,$E3
	.dc.b	$7E,$FF,$E7,$97,$9D,$FE,$C2,$9F,$F7,$53,$5F,$F7,$73,$77,$FF,$5B
	.dc.b	$5F,$77,$FD,$71,$D4,$F9,$1E,$8D,$BF,$33,$5F,$2E,$1F,$CF,$E7,$64
	.dc.b	$8D,$B5,$DF,$A0,$F3,$FE,$CB,$63,$FE,$00,$35,$FD,$57,$57,$F8,$5B
	.dc.b	$0F,$06,$3F,$EC,$DE,$77,$D7,$7C,$16,$DF,$CB,$6B,$E6,$6B,$F2,$5A
	.dc.b	$FF,$88,$9A,$FF,$BA,$DB,$9F,$F2,$29,$F9,$CB,$E5,$FF,$22,$DA,$FE
	.dc.b	$A1,$7F,$EE,$18,$C4,$7A,$8E,$95,$F9,$8E,$67,$D1,$7D,$96,$CF,$E0
	.dc.b	$DA,$FC,$E6,$BF,$81,$6B,$FE,$3A,$6A,$AF,$EA,$33,$D4,$FC,$37,$E0
	.dc.b	$EF,$99,$FD,$4B,$6B,$D3,$AF,$F9,$12,$31,$1E,$9B,$E6,$E8,$ED,$F8
	.dc.b	$23,$E5,$7F,$24,$D8,$FC,$E3,$5F,$C3,$B5,$FC,$D3,$5F,$F2,$9B,$5E
	.dc.b	$B1,$D0,$FC,$F7,$DC,$BC,$8F,$E0,$2D,$9F,$25,$A3,$11,$E9,$BF,$16
	.dc.b	$DF,$D2,$E9,$F4,$7F,$D0,$2B,$F6,$4D,$7F,$3A,$D7,$F6,$0D,$7F,$CD
	.dc.b	$2D,$7E,$83,$63,$AF,$8F,$FB,$58,$C7,$E1,$3F,$94,$BF,$13,$75,$FF
	.dc.b	$39,$B7,$0E,$77,$FC,$7B,$F0,$A7,$CE,$7E,$FD,$B3,$FC,$25,$AF,$ED
	.dc.b	$1A,$FF,$07,$6B,$FE,$80,$6B,$F8,$76,$E4,$68,$FB,$CF,$11,$27,$E7
	.dc.b	$C7,$83,$2F,$EE,$DE,$03,$F1,$A7,$D7,$FF,$36,$F4,$DF,$07,$4F,$EB
	.dc.b	$B5,$B9,$FE,$77,$F8,$67,$E6,$B4,$34,$9D,$1F,$C7,$FE,$9D,$F2,$BF
	.dc.b	$D0,$9B,$B7,$E0,$9A,$FE,$65,$FC,$DE,$BF,$E0,$B8,$7D,$B3,$E1,$BF
	.dc.b	$C0,$AD,$8F,$F0,$4B,$5F,$E6,$67,$08,$D9,$1D,$B5,$FE,$4B,$F1,$FF
	.dc.b	$6F,$33,$4F,$BD,$1C,$CF,$CC,$FD,$BF,$F4,$B6,$F9,$5F,$D7,$A3,$11
	.dc.b	$FB,$35,$F4,$39,$99,$3C,$0E,$71,$FD,$9B,$E5,$BF,$B5,$75,$7F,$B0
	.dc.b	$1F,$B0,$DE,$8D,$7A,$32,$FF,$57,$5A,$7F,$86,$37,$33,$D2,$3F,$CE
	.dc.b	$1F,$81,$BE,$77,$FC,$33,$0F,$B7,$F4,$F0,$FD,$16,$E7,$FE,$0D,$BE
	.dc.b	$15,$FC,$34,$38,$7F,$95,$1C,$F5,$EB,$5C,$A3,$C3,$86,$27,$FC,$0B
	.dc.b	$F7,$1F,$E0,$1C,$7F,$80,$37,$1F,$F0,$E6,$1C,$7C,$AF,$E7,$9D,$7F
	.dc.b	$1F,$E6,$66,$7F,$C7,$E7,$2D,$F2,$FF,$D6,$AF,$C6,$9D,$BF,$E7,$48
	.dc.b	$CB,$FC,$A7,$EB,$7F,$C6,$78,$FF,$BC,$1B,$1F,$CD,$DB,$C0,$7F,$10
	.dc.b	$60,$9E,$E7,$F8,$81,$FC,$4F,$DB,$3D,$EF,$F2,$EB,$6F,$FC,$C2,$D7
	.dc.b	$F3,$2D,$7F,$C1,$AD,$7F,$C7,$CD,$7F,$D5,$4D,$4E,$6F,$F0,$3F,$6F
	.dc.b	$F8,$D2,$4F,$D9,$9F,$07,$07,$F9,$47,$F1,$1F,$E5,$8D,$FA,$7F,$F2
	.dc.b	$8C,$1B,$E4,$4B,$FA,$13,$96,$CE,$5F,$F5,$0F,$E7,$5F,$92,$7F,$2F
	.dc.b	$E0,$FE,$7C,$FF,$2E,$C6,$63,$7D,$39,$76,$7D,$D8,$D6,$3F,$C1,$7D
	.dc.b	$D7,$D5,$FF,$A2,$1B,$5F,$E8,$C6,$BF,$C1,$1A,$FF,$8E,$1A,$FF,$9A
	.dc.b	$5A,$FF,$B5,$5A,$9C,$EC,$FA,$8F,$8F,$C7,$FE,$DB,$8C,$6D,$FE,$60
	.dc.b	$9C,$DF,$3C,$3F,$7D,$EA,$7E,$FF,$DE,$93,$FB,$A8,$CB,$FB,$C8,$CB
	.dc.b	$FD,$F7,$5F,$E4,$B9,$77,$C3,$E8,$4F,$F8,$05,$7E,$1B,$E1,$3C,$5C
	.dc.b	$12,$7F,$B8,$CE,$9B,$39,$79,$7D,$FF,$F9,$5E,$4D,$24,$7F,$98,$E3
	.dc.b	$FB,$BD,$F8,$FD,$6F,$80,$7C,$1A,$F8,$5F,$87,$CB,$F0,$BA,$69,$0F
	.dc.b	$53,$97,$D9,$E6,$F8,$C8,$F8,$5F,$37,$9F,$F3,$C9,$1E,$E8,$FA,$5F
	.dc.b	$2F,$EB,$24,$7A,$5F,$77,$DA,$FD,$72,$47,$C6,$8A,$FC,$D8,$BC,$BF
	.dc.b	$E1,$5D,$15,$8D,$3E,$A5,$5B,$FF,$1F,$56,$9F,$A3,$F2,$9F,$34,$8F
	.dc.b	$CB,$7E,$87,$CA,$FE,$24,$91,$C0,$FE,$3D,$F9,$DF,$FA,$61,$FF,$48
	.dc.b	$F8,$68,$C4,$7F,$B0,$1A,$84,$6F,$93,$FD,$87,$3A,$9C,$B6,$75,$1F
	.dc.b	$EE,$28,$D7,$43,$EA,$BF,$D8,$9A,$7E,$85,$B5,$FE,$5D,$6D,$F2,$8F
	.dc.b	$F3,$73,$E7,$E7,$FF,$80,$1F,$19,$E5,$8C,$4F,$9F,$E0,$BF,$DD,$31
	.dc.b	$88,$46,$32,$FF,$04,$B6,$F9,$CF,$A9,$B5,$C6,$78,$77,$62,$FF,$CD
	.dc.b	$31,$C6,$FC,$18,$78,$E7,$2D,$AE,$18,$4A,$D9,$DC,$B8,$7D,$39,$F0
	.dc.b	$FD,$9B,$F0,$AA,$4A,$DC,$7F,$8E,$5B,$19,$7D,$F8,$7C,$59,$E5,$1C
	.dc.b	$3D,$CD,$67,$F1,$A7,$33,$C0,$FB,$EE,$23,$53,$F1,$A7,$FA,$D7,$A0
	.dc.b	$F9,$5F,$5C,$61,$BE,$42,$75,$3D,$7D,$BF,$F3,$AF,$3F,$3A,$D5,$F0
	.dc.b	$BD,$E2,$CF,$F2,$58,$F5,$39,$7B,$EC,$79,$5F,$69,$FF,$76,$E3,$D1
	.dc.b	$FF,$1E,$B5,$CC,$6D,$EB,$0F,$B7,$7C,$D7,$F9,$12,$7F,$BC,$46,$1E
	.dc.b	$6D,$7B,$B5,$E6,$EF,$2F,$54,$7E,$AE,$BB,$57,$E0,$B4,$7B,$9F,$E8
	.dc.b	$2E,$6F,$93,$8F,$1D,$79,$DE,$0E,$15,$BC,$7E,$C1,$E6,$8F,$C5,$E8
	.dc.b	$EA,$63,$86,$D7,$DF,$E4,$F9,$5D,$58,$DB,$F5,$38,$71,$C3,$81,$2F
	.dc.b	$84,$2B,$86,$7F,$B3,$8E,$13,$AC,$72,$FF,$0A,$8C,$E1,$DA,$F3,$8F
	.dc.b	$DB,$BB,$F9,$3F,$D2,$CB,$F4,$AD,$C7,$D0,$FF,$15,$8C,$E7,$C0,$8F
	.dc.b	$09,$A7,$08,$B8,$FF,$06,$AE,$53,$9A,$E6,$47,$B9,$C3,$3F,$E3,$1E
	.dc.b	$DF,$CA,$EF,$FF,$68,$75,$FF,$A0,$AF,$FC,$99,$F2,$5F,$0A,$73,$5B
	.dc.b	$9F,$E3,$13,$F1,$B0,$FC,$EE,$A8,$C4,$6B,$5F,$F4,$E3,$6B,$FD,$2F
	.dc.b	$5B,$9C,$CF,$F3,$6A,$E2,$D8,$CC,$EE,$BF,$D3,$FD,$BF,$94,$7F,$DD
	.dc.b	$B9,$FC,$2C,$BF,$DD,$4F,$CF,$6E,$1E,$91,$5F,$B5,$DB,$C2,$DB,$EB
	.dc.b	$7F,$5E,$BF,$73,$39,$7F,$76,$72,$87,$A9,$D6,$7F,$80,$36,$1F,$33
	.dc.b	$62,$5F,$A1,$D4,$62,$D8,$FA,$A7,$0F,$B2,$6D,$7F,$8C,$6F,$AB,$E8
	.dc.b	$8F,$5C,$62,$FC,$F8,$7F,$13,$9C,$FC,$E7,$94,$F3,$7A,$97,$0F,$96
	.dc.b	$F2,$23,$F1,$1E,$BF,$17,$ED,$3D,$DE,$75,$9F,$A7,$17,$1F,$5C,$79
	.dc.b	$E7,$19,$FD,$67,$8C,$7D,$AF,$97,$B5,$C0,$F9,$CF,$B0,$91,$F5,$9F
	.dc.b	$7F,$A3,$F3,$FE,$4F,$E8,$36,$7F,$0A,$DA,$E9,$AF,$F3,$58,$C6,$3F
	.dc.b	$A2,$F3,$3E,$D7,$33,$CB,$B1,$86,$E3,$FE,$0D,$AF,$D4,$C6,$31,$EB
	.dc.b	$7C,$6B,$F2,$BF,$05,$F8,$33,$F2,$5F,$24,$F9,$4F,$EC,$D3,$A3,$A1
	.dc.b	$FE,$19,$8D,$79,$1F,$A6,$6C,$7F,$C0,$06,$E1,$F9,$46,$BF,$D4,$1A
	.dc.b	$F3,$AE,$EE,$DA,$ED,$AF,$F8,$57,$D6,$7D,$77,$C4,$F4,$7F,$55,$D6
	.dc.b	$9C,$E3,$E8,$BF,$F1,$15,$6A,$73,$1F,$12,$31,$7C,$EB,$F7,$3F,$73
	.dc.b	$FA,$7F,$88,$FD,$43,$63,$27,$EF,$A3,$2D,$3C,$1C,$7F,$E3,$5F,$4F
	.dc.b	$EC,$9F,$3B,$47,$9F,$8A,$F3,$6D,$F4,$63,$E6,$4E,$A7,$E6,$37,$4F
	.dc.b	$8D,$3A,$24,$E9,$B1,$FA,$EF,$4F,$FC,$19,$3F,$E9,$0F,$3B,$F8,$24
	.dc.b	$3E,$46,$C4,$9D,$1F,$E9,$68,$C5,$7F,$BD,$FE,$CB,$F2,$7E,$57,$D0
	.dc.b	$F8,$04,$62,$FF,$E0,$03,$6B,$84,$7C,$4E,$7F,$F0,$73,$F5,$7C,$57
	.dc.b	$D8,$F9,$AB,$F4,$7F,$61,$FD,$CF,$B3,$FB,$6D,$6B,$5F,$E3,$93,$9C
	.dc.b	$3F,$C3,$7F,$60,$E6,$FE,$59,$7B,$9F,$C1,$62,$AD,$47,$87,$2F,$06
	.dc.b	$5F,$DE,$AB,$58,$72,$9C,$D7,$68,$C6,$5F,$B4,$F2,$7F,$6B,$E7,$5A
	.dc.b	$F7,$37,$0F,$D8,$B7,$1F,$EB,$AF,$F3,$5F,$AF,$FF,$6E,$57,$E2,$1B
	.dc.b	$5F,$C1,$17,$87,$BF,$0E,$56,$DE,$51,$8A,$FE,$17,$E6,$BE,$57,$FA
	.dc.b	$13,$6F,$F1,$F3,$F6,$1F,$CB,$C6,$63,$F8,$F7,$57,$FA,$5F,$93,$F7
	.dc.b	$D1,$B3,$F7,$85,$3F,$78,$53,$F8,$31,$44,$73,$AF,$E0,$D1,$98,$C3
	.dc.b	$FD,$13,$F1,$2F,$F7,$4C,$FE,$A9,$B1,$FC,$99,$FF,$71,$C6,$7D,$FF
	.dc.b	$EB,$DF,$22,$F9,$AF,$ED,$2D,$9F,$EE,$AD,$BF,$EA,$2D,$7F,$CC,$7E
	.dc.b	$BB,$F1,$6F,$C6,$47,$F7,$49,$FE,$D7,$5F,$C2,$EB,$5F,$21,$E0,$F5
	.dc.b	$3F,$B7,$1F,$E6,$D9,$FE,$02,$DB,$FE,$7B,$F2,$AF,$8C,$DF,$AD,$FD
	.dc.b	$D6,$31,$7F,$E0,$A6,$D7,$F6,$66,$BF,$E0,$06,$BF,$EC,$BF,$1D,$18
	.dc.b	$8F,$F0,$5B,$63,$37,$FE,$0C,$8C,$7C,$B7,$DF,$E9,$FF,$2C,$D2,$31
	.dc.b	$87,$5F,$7F,$FC,$A8,$FF,$1C,$8C,$E3,$FE,$18,$9C,$BF,$DF,$1A,$FF
	.dc.b	$85,$9A,$F3,$EA,$FF,$0E,$C6,$23,$94,$60,$95,$FE,$2C,$FA,$8F,$E6
	.dc.b	$F8,$F9,$9B,$1F,$07,$E6,$5E,$C7,$F8,$7E,$73,$87,$F8,$B2,$73,$5F
	.dc.b	$E3,$66,$C7,$F8,$49,$AF,$F8,$B1,$AF,$5A,$FF,$19,$C6,$23,$A4,$60
	.dc.b	$FF,$1B,$FC,$CF,$DF,$C3,$E5,$B6,$FE,$B7,$CD,$7F,$A2,$1F,$E8,$B1
	.dc.b	$9B,$FF,$20,$CE,$5F,$F1,$2B,$5F,$F1,$FB,$5E,$73,$FE,$46,$8C,$47
	.dc.b	$28,$C1,$0F,$F2,$AF,$D4,$7F,$9A,$5F,$C4,$36,$3F,$CA,$3E,$A3,$EA
	.dc.b	$FF,$C9,$13,$9E,$04,$E6,$FF,$CB,$ED,$BF,$F1,$DB,$5F,$F2,$AB,$5F
	.dc.b	$CB,$35,$EF,$7F,$E5,$E8,$C5,$76,$8F,$F3,$0F,$91,$FB,$EA,$FF,$12
	.dc.b	$B7,$3F,$CB,$FC,$EB,$E4,$FF,$CB,$D3,$9C,$BF,$49,$39,$7F,$CE,$AD
	.dc.b	$BF,$F2,$8B,$5F,$F3,$5B,$5E,$35,$FE,$72,$8C,$47,$F9,$D9,$B8,$9B
	.dc.b	$F0,$6F,$80,$FF,$38,$37,$0F,$F0,$19,$FC,$32,$91,$9B,$FF,$20,$CF
	.dc.b	$FB,$45,$FF,$3E,$B5,$FD,$F3,$5F,$F4,$3B,$65,$F6,$0F,$92,$FF,$40
	.dc.b	$36,$7F,$C2,$6B,$E4,$73,$32,$72,$F0,$8C,$5F,$F9,$53,$57,$CF,$FF
	.dc.b	$A2,$DB,$7F,$E1,$E5,$F8,$9E,$66,$4E,$5E,$11,$8A,$FF,$35,$70,$BE
	.dc.b	$8D,$F1,$DF,$E3,$A6,$E1,$FE,$C0,$FB,$9C,$4C,$8F,$F1,$E5,$71,$8F
	.dc.b	$F2,$0F,$C5,$2B,$58,$76,$8C,$5F,$FA,$F9,$B1,$FE,$A8,$BF,$F6,$35
	.dc.b	$7F,$AF,$27,$33,$FE,$CA,$FA,$57,$DD,$FF,$93,$9B,$87,$FB,$53,$F0
	.dc.b	$2C,$4C,$8F,$2C,$EA,$BC,$55,$97,$FD,$14,$09,$C7,$FE,$BA,$7F,$D7
	.dc.b	$6D,$74,$7E,$20,$C1,$4F,$F6,$0F,$D4,$7E,$07,$E0,$78,$36,$3A,$35
	.dc.b	$DF,$1F,$F9,$FD,$CF,$CE,$FF,$AF,$9B,$C7,$CA,$31,$1D,$9A,$FA,$77
	.dc.b	$67,$FC,$CA,$3E,$0F,$7F,$A2,$DA,$EE,$B5,$5F,$57,$C7,$D4,$E6,$EC
	.dc.b	$FF,$10,$AD,$DF,$28,$C4,$EB,$0F,$F7,$F6,$17,$83,$FC,$B9,$5F,$7F
	.dc.b	$E0,$3E,$55,$27,$33,$F8,$16,$BF,$3A,$3F,$29,$F2,$9D,$79,$4D,$3B
	.dc.b	$5D,$AF,$C9,$CE,$9B,$3A,$7F,$D0,$7C,$0F,$5F,$DB,$D7,$FE,$AF,$6C
	.dc.b	$66,$77,$EB,$E2,$64,$72,$77,$1C,$BC,$E7,$CC,$52,$57,$FA,$4A,$1F
	.dc.b	$41,$88,$FF,$67,$C7,$07,$53,$CE,$7C,$BF,$4E,$FA,$8F,$C9,$66,$8C
	.dc.b	$5F,$E5,$E3,$13,$A9,$CC,$FF,$BF,$BD,$87,$B5,$F0,$1B,$1F,$EA,$66
	.dc.b	$BF,$EE,$29,$E3,$CB,$FE,$14,$EB,$FF,$1E,$4F,$76,$BF,$A4,$18,$FF
	.dc.b	$2C,$F5,$FC,$F3,$A9,$FD,$4C,$E5,$E5,$3A,$23,$FD,$1B,$C2,$FA,$BE
	.dc.b	$37,$F2,$08,$FF,$5D,$B5,$FF,$7B,$DF,$8B,$0F,$F1,$4F,$33,$E4,$FE
	.dc.b	$3B,$6B,$FB,$3C,$60,$8E,$08,$7F,$6A,$7E,$7E,$4F,$85,$FC,$3B,$67
	.dc.b	$FD,$9A,$D5,$7E,$2A,$FC,$6F,$43,$E2,$BE,$CB,$6B,$EC,$35,$F9,$57
	.dc.b	$FE,$C3,$9D,$92,$36,$D7,$FC,$2B,$F3,$5F,$9A,$F3,$CD,$3C,$33,$F9
	.dc.b	$EF,$41,$F3,$BF,$8A,$6D,$FE,$25,$AF,$89,$AF,$D8,$6B,$FE,$5F,$6B
	.dc.b	$F8,$06,$E0,$6F,$A5,$FE,$A1,$9F,$F7,$9B,$5F,$8F,$E1,$FF,$92,$7E
	.dc.b	$03,$CD,$87,$ED,$3D,$6F,$ED,$55,$F9,$D6,$BF,$E5,$DF,$1F,$F4,$71
	.dc.b	$88,$FF,$2F,$46,$08,$F1,$7F,$53,$EB,$FE,$0E,$3C,$27,$E6,$1C,$70
	.dc.b	$FF,$9C,$9B,$8F,$F9,$C6,$BF,$83,$FB,$0F,$6F,$F6,$6D,$C3,$FD,$11
	.dc.b	$18,$23,$82,$1F,$E8,$B7,$8F,$9F,$D0,$91,$FD,$4F,$1F,$F4,$1D,$75
	.dc.b	$8F,$3F,$82,$FA,$87,$C9,$FF,$1A,$F7,$5F,$21,$4F,$E7,$35,$B9,$FE
	.dc.b	$3F,$EF,$FE,$3F,$0F,$D9,$B5,$FF,$4D,$35,$FF,$58,$73,$7F,$10,$F9
	.dc.b	$97,$C7,$FF,$28,$6C,$F9,$F8,$7F,$D4,$51,$88,$F9,$BD,$08,$FC,$AC
	.dc.b	$6C,$FE,$E1,$F5,$0F,$95,$FE,$D6,$D8,$FE,$E0,$D7,$FB,$A3,$5F,$E2
	.dc.b	$0D,$7F,$BA,$B5,$EB,$1D,$0F,$E5,$3E,$3F,$DC,$8F,$E2,$ED,$BF,$E4
	.dc.b	$BD,$FF,$D8,$25,$7E,$1E,$3F,$F5,$AF,$D4,$7F,$3A,$9F,$55,$A3,$11
	.dc.b	$F0,$1B,$3C,$3E,$A3,$FA,$7C,$63,$4F,$F6,$43,$6B,$F9,$4B,$6F,$A5
	.dc.b	$7F,$56,$FA,$2F,$EB,$95,$F3,$1F,$C2,$61,$FE,$C8,$6E,$1F,$ED,$37
	.dc.b	$71,$9F,$47,$97,$4C,$BF,$4D,$6A,$FF,$C3,$BF,$50,$FA,$8F,$85,$F4
	.dc.b	$FE,$5B,$36,$75,$1F,$6C,$38,$7F,$C5,$B1,$C4,$E1,$3D,$3D,$0F,$EE
	.dc.b	$3F,$51,$FD,$BE,$BF,$0E,$DD,$3F,$C1,$D1,$E0,$9F,$F0,$77,$92,$EC
	.dc.b	$F8,$9F,$15,$2B,$FC,$A7,$39,$6C,$E5,$F9,$13,$CA,$3F,$C0,$3F,$47
	.dc.b	$FE,$1C,$D7,$FD,$F4,$DA,$FE,$B6,$DE,$23,$F8,$73,$04,$F6,$F9,$F3
	.dc.b	$7E,$27,$EA,$1E,$F7,$F9,$19,$B8,$7F,$92,$DA,$FF,$93,$DA,$FF,$6B
	.dc.b	$6B,$FE,$50,$6B,$FE,$57,$6A,$73,$83,$8F,$FD,$69,$3F,$7B,$E1,$FF
	.dc.b	$14,$79,$6F,$FC,$71,$F5,$1F,$CB,$38,$7F,$C7,$9E,$8A,$3F,$D2,$D7
	.dc.b	$FE,$7E,$6B,$FE,$88,$78,$6D,$FE,$51,$6C,$7F,$B5,$9D,$D7,$46,$CE
	.dc.b	$FC,$8F,$F2,$47,$CF,$79,$EF,$F2,$7E,$0F,$E7,$CC,$E9,$23,$24,$3F
	.dc.b	$D3,$33,$C5,$FF,$3D,$FB,$8F,$A2,$FF,$3A,$B6,$3F,$CF,$4D,$7F,$CF
	.dc.b	$CD,$7F,$C3,$4D,$7F,$CF,$ED,$7F,$D0,$CD,$7F,$C9,$51,$FE,$34,$F9
	.dc.b	$DF,$83,$C0,$69,$F1,$27,$67,$EC,$3D,$3F,$DE,$7B,$B2,$30,$43,$F7
	.dc.b	$CE,$5F,$E0,$11,$FE,$58,$9F,$97,$1F,$E3,$05,$F2,$59,$35,$F1,$52
	.dc.b	$7F,$8B,$CE,$9B,$39,$7F,$9B,$C3,$EA,$3D,$84,$8F,$F2,$74,$65,$BE
	.dc.b	$A7,$8C,$8F,$69,$F3,$7D,$14,$8F,$0B,$A7,$BB,$96,$9E,$69,$F7,$79
	.dc.b	$DF,$34,$91,$E2,$7E,$87,$A1,$F0,$C9,$1E,$E8,$F8,$DF,$41,$F5,$49
	.dc.b	$1E,$A7,$EC,$7D,$07,$CB,$24,$7C,$68,$AF,$C4,$8B,$D1,$FE,$A1,$E0
	.dc.b	$FB,$D5,$8E,$7F,$F6,$54,$7E,$06,$B1,$7F,$56,$BE,$C7,$CD,$7E,$1C
	.dc.b	$91,$F3,$1F,$C5,$7C,$B7,$DD,$24,$7D,$88,$FF,$57,$7D,$47,$FA,$1E
	.dc.b	$3F,$DD,$0D,$7F,$D5,$0D,$7F,$D8,$8D,$5D,$88,$C3,$79,$7F,$D6,$B1
	.dc.b	$83,$86,$4F,$A1,$FF,$5B,$EB,$F0,$5B,$AF,$C0,$6E,$3D,$AF,$CD,$5F
	.dc.b	$C3,$1F,$F0,$0C,$3E,$05,$F1,$9E,$6E,$8F,$F7,$6C,$62,$79,$46,$27
	.dc.b	$D7,$DB,$67,$27,$FB,$E2,$39,$34,$C9,$DD,$E2,$E7,$5F,$85,$5C,$DB
	.dc.b	$19,$C3,$CF,$87,$8E,$7E,$46,$1B,$9C,$E1,$E0,$BE,$C7,$82,$F8,$4F
	.dc.b	$F8,$8D,$E5,$F5,$0F,$A4,$F0,$36,$BC,$8D,$79,$E2,$87,$BB,$C2,$A5
	.dc.b	$70,$79,$C6,$23,$8C,$F3,$BF,$0D,$6A,$7E,$31,$FD,$73,$D2,$7C,$FE
	.dc.b	$3C,$AF,$F7,$7B,$63,$33,$EE,$F7,$1F,$00,$E9,$FA,$93,$CE,$F8,$11
	.dc.b	$E1,$3B,$70,$7F,$B2,$70,$EE,$FF,$B2,$CE,$45,$72,$EF,$EA,$1F,$33
	.dc.b	$B8,$FE,$EB,$18,$78,$35,$E1,$1B,$94,$EF,$4F,$C8,$57,$2A,$FC,$56
	.dc.b	$5F,$E5,$9F,$5D,$CF,$F1,$70,$E1,$C6,$C2,$65,$D3,$1D,$1C,$BD,$DF
	.dc.b	$F4,$46,$18,$AF,$BB,$B3,$AF,$AC,$61,$BA,$FE,$46,$5D,$98,$CF,$C1
	.dc.b	$CB,$F4,$6F,$18,$7E,$FA,$FA,$63,$F9,$F6,$C7,$83,$0D,$DF,$1A,$D5
	.dc.b	$E2,$F7,$8F,$86,$F1,$3C,$3E,$A3,$E8,$DF,$EA,$DB,$8F,$E6,$1F,$89
	.dc.b	$19,$D3,$81,$1E,$13,$4E,$11,$70,$FD,$85,$72,$9C,$D7,$32,$3D,$CE
	.dc.b	$19,$FF,$13,$F7,$7E,$5F,$5F,$F1,$76,$D7,$C8,$F9,$2F,$87,$39,$8F
	.dc.b	$F4,$A6,$C9,$F8,$F8,$7D,$D3,$F9,$4B,$87,$96,$9F,$E5,$66,$D7,$E6
	.dc.b	$6B,$94,$E6,$78,$B6,$33,$3B,$AE,$94,$EE,$FE,$45,$3F,$E8,$EC,$BE
	.dc.b	$16,$3F,$D4,$5F,$9E,$DA,$F6,$9C,$63,$E9,$BE,$9A,$FE,$5F,$AF,$8D
	.dc.b	$B7,$FA,$F6,$CF,$BC,$E5,$EA,$74,$8F,$E6,$ED,$B7,$D4,$58,$97,$E7
	.dc.b	$75,$18,$BF,$F6,$1D,$7F,$9E,$6B,$47,$B6,$75,$8F,$B5,$F3,$4F,$69
	.dc.b	$F3,$36,$BC,$D8,$7F,$4A,$AD,$FC,$DF,$94,$F3,$7A,$D7,$0F,$94,$F4
	.dc.b	$23,$D1,$EF,$EB,$7B,$4F,$77,$9D,$67,$E9,$45,$C7,$D7,$37,$3F,$A8
	.dc.b	$F0,$AF,$B1,$F5,$16,$BE,$0B,$EC,$24,$7B,$1F,$7F,$A1,$F3,$7E,$4F
	.dc.b	$E7,$36,$7F,$04,$DA,$F9,$6D,$C3,$FB,$A7,$9B,$F5,$FC,$87,$97,$63
	.dc.b	$0D,$C3,$FD,$B3,$3A,$6C,$E9,$FC,$87,$C1,$FD,$F3,$F1,$BF,$24,$F9
	.dc.b	$4F,$EB,$53,$A3,$FD,$F9,$19,$F1,$FF,$40,$D8,$FE,$78,$DD,$3F,$1C
	.dc.b	$D7,$F5,$CD,$7F,$96,$35,$34,$C7,$4B,$D5,$7D,$47,$C1,$F4,$7F,$47
	.dc.b	$39,$8C,$61,$FB,$EA,$DB,$F0,$AB,$71,$8C,$1B,$FF,$87,$E7,$F4,$1F
	.dc.b	$4B,$FA,$1F,$88,$B8,$7E,$D5,$B7,$F9,$E6,$DF,$ED,$DF,$A9,$E9,$CA
	.dc.b	$F8,$FB,$3D,$1C,$61,$E0,$9F,$83,$39,$7E,$4B,$63,$E4,$B5,$F9,$2D
	.dc.b	$78,$CF,$D2,$8C,$47,$CA,$6E,$DF,$6E,$7F,$82,$7A,$DF,$E0,$2B,$FF
	.dc.b	$43,$F9,$DF,$C0,$61,$F2,$36,$24,$E8,$FF,$46,$C6,$31,$F5,$1F,$E5
	.dc.b	$0F,$81,$7D,$27,$E5,$7C,$BF,$AB,$18,$BF,$C9,$C6,$27,$51,$FE,$FC
	.dc.b	$6D,$74,$F1,$5F,$4D,$E7,$8E,$FE,$23,$EA,$3F,$B7,$C6,$E7,$5C,$7F
	.dc.b	$E7,$99,$CB,$FD,$C2,$5F,$50,$E5,$FE,$57,$8F,$6D,$FF,$96,$D6,$A7
	.dc.b	$38,$F4,$D3,$FA,$36,$1C,$5F,$F0,$26,$3F,$DF,$31,$E5,$18,$37,$7C
	.dc.b	$B1,$FD,$C7,$95,$FB,$9F,$3A,$D7,$ED,$9B,$9F,$BB,$BF,$FC,$07,$3F
	.dc.b	$BA,$AE,$0F,$59,$FE,$95,$F5,$0F,$0B,$F9,$D3,$67,$5C,$08,$78,$7A
	.dc.b	$A9,$5A,$9E,$17,$F3,$9D,$C6,$67,$FA,$37,$9A,$F9,$3A,$7F,$08,$BF
	.dc.b	$BC,$7D,$68,$D5,$7F,$08,$8C,$9F,$C0,$8A,$7F,$02,$29,$FC,$08,$BD
	.dc.b	$5F,$E3,$1E,$5F,$E1,$F1,$82,$3F,$D7,$FE,$A3,$FD,$CB,$5F,$01,$B3
	.dc.b	$F8,$87,$FD,$C1,$19,$F8,$0F,$EC,$1F,$22,$F9,$AF,$ED,$0D,$9F,$ED
	.dc.b	$0D,$BF,$EA,$6D,$7F,$C1,$1E,$CB,$EA,$3F,$1F,$5F,$EC,$6A,$FE,$C5
	.dc.b	$5F,$A6,$AD,$7B,$5E,$0F,$53,$FB,$81,$FE,$A6,$BF,$AE,$DA,$FC,$FF
	.dc.b	$CA,$BE,$33,$7E,$B7,$F7,$78,$C5,$FF,$7D,$6D,$7F,$68,$6B,$FE,$05
	.dc.b	$6B,$FE,$DA,$6B,$FE,$13,$6F,$31,$A6,$3E,$5B,$ED,$CF,$EC,$9B,$DD
	.dc.b	$FE,$E7,$F7,$BF,$2F,$39,$E6,$FF,$06,$46,$23,$FC,$2C,$D7,$FD,$D6
	.dc.b	$D7,$FC,$48,$DC,$3F,$D2,$3D,$9F,$E1,$B8,$C1,$1F,$F0,$F7,$D4,$7F
	.dc.b	$3B,$C3,$FD,$F0,$DC,$FF,$DE,$FF,$30,$F6,$3F,$C3,$D3,$9C,$BF,$C6
	.dc.b	$53,$9A,$FF,$13,$36,$3F,$C2,$2D,$7F,$C5,$CD,$7A,$D7,$F8,$BE,$31
	.dc.b	$1D,$23,$07,$F8,$CF,$E6,$7E,$E6,$1F,$E8,$2C,$D1,$8C,$3F,$D1,$D0
	.dc.b	$F9,$AF,$F4,$9C,$7E,$EA,$75,$87,$F9,$26,$72,$FF,$89,$9A,$FF,$90
	.dc.b	$DA,$F4,$6B,$C6,$7F,$C8,$D1,$8A,$F0,$46,$0F,$F2,$57,$D4,$7F,$9A
	.dc.b	$9F,$2B,$6B,$FC,$B5,$EA,$3E,$A3,$FC,$95,$39,$B4,$E6,$FF,$CA,$4D
	.dc.b	$BF,$F1,$EB,$5F,$F2,$D3,$5F,$86,$D7,$BD,$FF,$97,$63,$15,$DA,$3F
	.dc.b	$CC,$1E,$3F,$EF,$AB,$FA,$3B,$73,$FA,$1F,$3A,$F9,$3F,$F3,$04,$E7
	.dc.b	$2F,$A3,$39,$7F,$CD,$0D,$BF,$F2,$9B,$5F,$F3,$83,$5E,$35,$FE,$71
	.dc.b	$8C,$47,$F9,$FD,$B8,$9B,$F0,$6F,$80,$FF,$39,$37,$0F,$E7,$27,$F0
	.dc.b	$CB,$FF,$44,$46,$5F,$EC,$71,$FD,$89,$FF,$41,$35,$FC,$5B,$5F,$F4
	.dc.b	$83,$6D,$F5,$0F,$92,$FF,$41,$36,$7F,$A4,$AF,$91,$CC,$C9,$CB,$C2
	.dc.b	$31,$7F,$DD,$B5,$7C,$FF,$FA,$35,$B7,$FD,$51,$7D,$46,$66,$4E,$5E
	.dc.b	$11,$8A,$FF,$07,$F0,$BE,$8D,$F1,$FF,$E4,$26,$E1,$FE,$BE,$F8,$97
	.dc.b	$D4,$66,$A4,$FF,$B0,$AB,$78,$74,$BD,$CE,$6B,$FC,$9B,$18,$9F,$F5
	.dc.b	$46,$1F,$EC,$9C,$3F,$D7,$D3,$99,$FF,$65,$4E,$6F,$FD,$8D,$F4,$AF
	.dc.b	$A8,$FF,$2B,$B6,$3C,$DF,$51,$89,$91,$FE,$D6,$9D,$5F,$96,$B2,$FF
	.dc.b	$8C,$0F,$F6,$C7,$1F,$FA,$ED,$FF,$5E,$35,$C9,$43,$9F,$A8,$FC,$07
	.dc.b	$C0,$DE,$3F,$F4,$05,$73,$9C,$B6,$32,$FF,$B2,$61,$F3,$9F,$E6,$A8
	.dc.b	$E0,$D7,$4D,$7F,$DD,$D1,$C3,$8F,$E2,$FC,$77,$FB,$C5,$68,$AF,$E1
	.dc.b	$3A,$2B,$53,$9D,$93,$9A,$FF,$7E,$FD,$4F,$FB,$61,$F9,$57,$C1,$E6
	.dc.b	$DA,$E2,$D9,$E1,$3F,$8E,$FC,$FD,$ED,$7C,$BB,$46,$27,$E8,$46,$27
	.dc.b	$53,$99,$FC,$3E,$0F,$8F,$F0,$36,$3E,$F4,$78,$E7,$86,$7F,$0A,$7E
	.dc.b	$0C,$F0,$23,$FE,$41,$C3,$C0,$FF,$8F,$B7,$7C,$6F,$BF,$B3,$FD,$74
	.dc.b	$D8,$CC,$F1,$F8,$A7,$C2,$66,$A4,$F3,$7E,$34,$EA,$3C,$1F,$A1,$FE
	.dc.b	$6D,$BC,$FF,$2E,$31,$1F,$E8,$4D,$1F,$98,$BD,$9D,$5C,$AE,$27,$F9
	.dc.b	$6E,$B7,$3F,$57,$F2,$2F,$95,$3F,$AC,$6B,$F8,$16,$BF,$D3,$9B,$AF
	.dc.b	$FB,$2B,$D3,$F8,$5E,$E9,$A7,$5E,$4F,$C1,$7E,$C0,$ED,$FB,$CD,$C7
	.dc.b	$EE,$B5,$FE,$02,$D7,$F7,$ED,$7F,$28,$D7,$FB,$0B,$78,$CE,$71,$FF
	.dc.b	$6F,$D7,$98,$95,$F7,$67,$C1,$A7,$E0,$ED,$F8,$27,$F6,$FA,$FC,$8B
	.dc.b	$5F,$E2,$AD,$7F,$89,$B5,$FD,$43,$5F,$EE,$AD,$8E,$BE,$5F,$E7,$E3
	.dc.b	$58,$87,$C9,$7E,$35,$B7,$FE,$8E,$6B,$FC,$DA,$7C,$98,$7A,$7E,$CD
	.dc.b	$ED,$7E,$AD,$B5,$FA,$A6,$BF,$CB,$9A,$FF,$2C,$6B,$FC,$05,$AF,$F8
	.dc.b	$39,$B9,$1C,$E1,$E9,$FC,$5F,$D0,$4A,$FD,$54,$F8,$30,$F9,$DF,$91
	.dc.b	$AB,$F8,$03,$6B,$F7,$ED,$7F,$A1,$B5,$FE,$82,$D7,$F8,$B3,$5F,$F1
	.dc.b	$1B,$70,$39,$C1,$EA,$FF,$A4,$EB,$FD,$72,$D7,$FD,$C2,$DC,$3C,$F8
	.dc.b	$7F,$1B,$FC,$23,$ED,$13,$6F,$E3,$0D,$BF,$E2,$ED,$7F,$AD,$BF,$7A
	.dc.b	$FF,$73,$7F,$E5,$56,$F8,$5F,$E6,$AE,$7E,$D8,$7F,$2B,$FB,$47,$D0
	.dc.b	$3E,$4B,$FA,$84,$61,$B1,$9C,$FF,$AB,$4E,$5B,$19,$7F,$A0,$1F,$D0
	.dc.b	$63,$4E,$03,$87,$6D,$7F,$A2,$7D,$7F,$C5,$C6,$FC,$37,$F3,$F8,$C4
	.dc.b	$F2,$BF,$06,$9F,$AD,$FC,$E1,$F1,$5F,$D1,$5B,$5F,$CB,$DA,$A7,$F9
	.dc.b	$B6,$4F,$C4,$3F,$6F,$9F,$F5,$9F,$87,$7E,$19,$FC,$A7,$0F,$54,$63
	.dc.b	$C6,$FE,$EC,$D8,$FE,$F4,$D7,$FC,$1F,$F3,$5F,$8F,$CB,$FD,$11,$F4
	.dc.b	$4F,$A8,$FE,$7C,$DA,$FE,$89,$5B,$CB,$E5,$FD,$8F,$FB,$26,$7F,$C0
	.dc.b	$8D,$7F,$C1,$0D,$7E,$73,$5E,$3F,$17,$FE,$09,$8C,$47,$F7,$56,$A7
	.dc.b	$3D,$5F,$2B,$E8,$BF,$A9,$71,$1A,$7F,$73,$AE,$17,$FD,$5F,$EC,$DF
	.dc.b	$39,$49,$3F,$D7,$FE,$AD,$F7,$78,$8F,$A5,$A7,$FA,$A2,$F8,$C7,$F8
	.dc.b	$27,$83,$FD,$5D,$18,$8F,$F2,$4D,$F2,$7F,$B5,$61,$FC,$06,$FF,$D6
	.dc.b	$3F,$5D,$F1,$70,$FF,$5B,$36,$3F,$C2,$13,$BA,$FF,$4E,$B7,$55,$6E
	.dc.b	$7F,$94,$37,$C2,$F6,$F8,$9F,$07,$EC,$3F,$17,$5F,$9D,$6B,$FA,$26
	.dc.b	$BF,$E2,$36,$A6,$93,$83,$FC,$7C,$F0,$CF,$F1,$47,$F8,$2F,$95,$F8
	.dc.b	$3B,$EB,$3F,$C4,$91,$87,$FC,$4D,$3C,$36,$FF,$69,$FD,$13,$EC,$BF
	.dc.b	$C1,$8D,$CF,$E7,$4E,$A7,$FC,$59,$F9,$0F,$C3,$AD,$57,$D6,$3F,$A5
	.dc.b	$38,$BF,$F7,$13,$6F,$FC,$79,$5F,$37,$67,$E3,$9F,$89,$C3,$B9,$FD
	.dc.b	$4F,$EB,$F0,$F5,$E2,$46,$32,$FE,$D1,$57,$1F,$F2,$43,$6F,$97,$47
	.dc.b	$F8,$9E,$F1,$8F,$82,$3E,$B7,$D4,$FC,$C2,$57,$F9,$B7,$85,$18,$8F
	.dc.b	$F3,$94,$61,$D4,$65,$FE,$3B,$9F,$F0,$EF,$73,$F4,$A4,$8F,$69,$F7
	.dc.b	$67,$FC,$AD,$F5,$5F,$8C,$8F,$83,$5F,$8E,$FA,$AF,$F1,$F7,$39,$DB
	.dc.b	$27,$6E,$5B,$19,$7F,$D5,$7F,$69,$FE,$35,$7F,$C9,$8D,$73,$ED,$7F
	.dc.b	$99,$FF,$23,$78,$3F,$E6,$9A,$D7,$1F,$F8,$55,$BC,$7F,$E7,$56,$E0
	.dc.b	$E1,$61,$CA,$3D,$FF,$48,$FA,$9F,$F4,$B3,$63,$FD,$34,$D7,$FD,$3C
	.dc.b	$D7,$FA,$23,$5F,$E9,$2D,$7F,$D1,$4D,$4E,$76,$FE,$7E,$FF,$16,$24
	.dc.b	$79,$27,$F3,$FE,$64,$8F,$A3,$D0,$8F,$F4,$57,$E4,$6F,$A2,$D5,$7E
	.dc.b	$4E,$B4,$F0,$AD,$CF,$C9,$F0,$A4,$FD,$07,$FD,$29,$FB,$5F,$F9,$DA
	.dc.b	$3E,$AC,$69,$A7,$FA,$06,$35,$C1,$FE,$45,$FC,$C7,$F8,$E4,$FF,$B5
	.dc.b	$9A,$FF,$B5,$1A,$FF,$AA,$5A,$FD,$5F,$C3,$5F,$37,$FE,$A5,$6C,$7F
	.dc.b	$A3,$DA,$FF,$BC,$A7,$BD,$7F,$95,$67,$2F,$F9,$5F,$E4,$DF,$80,$7F
	.dc.b	$AB,$DB,$7C,$F0,$EC,$66,$64,$64,$FF,$4F,$7C,$13,$E6,$3F,$D6,$4D
	.dc.b	$8F,$F5,$8B,$5F,$F7,$AB,$5F,$D7,$FC,$23,$F0,$8F,$F5,$23,$63,$F5
	.dc.b	$6D,$7F,$8B,$73,$FF,$B0,$2F,$FD,$8D,$F9,$8F,$F6,$08,$34,$9E,$87
	.dc.b	$FB,$BF,$A3,$FD,$1D,$F8,$57,$62,$61,$FE,$B7,$AF,$E8,$93,$A8,$F0
	.dc.b	$CE,$9F,$0C,$EB,$14,$E7,$7F,$F6,$25,$F1,$BD,$E1,$FE,$DB,$FC,$C7
	.dc.b	$FD,$C5,$1F,$E2,$7C,$91,$88,$FF,$16,$C6,$3E,$4F,$FD,$B9,$39,$9F
	.dc.b	$F7,$85,$FF,$A9,$FE,$BB,$E3,$92,$3F,$DA,$13,$DB,$1F,$F4,$A7,$03
	.dc.b	$E8,$7F,$DC,$CD,$AF,$F7,$2B,$5F,$23,$5F,$1B,$5F,$F7,$E3,$5F,$E1
	.dc.b	$0D,$CB,$F1,$73,$F3,$FE,$E7,$FD,$79,$5F,$C9,$1A,$FF,$43,$6D,$F9
	.dc.b	$A7,$FE,$00,$5F,$F9,$62,$87,$FB,$BC,$EF,$EF,$FF,$94,$BC,$2F,$2B
	.dc.b	$63,$CA,$D7,$67,$28,$DB,$4D,$94,$D9,$DF,$1F,$D2,$CF,$F9,$83,$E6
	.dc.b	$BD,$7A,$B3,$7E,$15,$9E,$2F,$AB,$18,$C7,$D3,$7F,$0B,$1E,$71,$8A
	.dc.b	$F0,$A9,$EB,$5C,$63,$13,$D2,$FB,$47,$9A,$BA,$7E,$77,$FC,$01,$BD
	.dc.b	$1A,$C5,$39,$C3,$E5,$C6,$2F,$53,$97,$FA,$13,$63,$8D,$6A,$76,$C7
	.dc.b	$D4,$70,$74,$E0,$FE,$FF,$9B,$ED,$71,$E8,$7F,$A5,$1B,$19,$9F,$78
	.dc.b	$F5,$5F,$59,$FE,$75,$D3,$F1,$38,$71,$AD,$9D,$B0,$E6,$79,$2F,$8F
	.dc.b	$C0,$7C,$89,$D1,$FE,$97,$29,$FE,$99,$AE,$8D,$C7,$C8,$E7,$D7,$FE
	.dc.b	$5A,$FF,$A8,$A7,$FC,$6D,$9F,$38,$FF,$0E,$CE,$9F,$CC,$CE,$9F,$03
	.dc.b	$72,$F5,$CF,$F3,$A7,$F4,$F8,$BD,$FF,$EB,$1C,$7F,$22,$B5,$C1,$E1
	.dc.b	$6E,$3D,$9B,$7C,$7C,$55,$F1,$CF,$EB,$F0,$3F,$3C,$EB,$8F,$D0,$8C
	.dc.b	$5F,$08,$C4,$FF,$B2,$AF,$8B,$C3,$4F,$CE,$37,$3F,$C6,$B6,$BF,$C1
	.dc.b	$8F,$17,$9D,$F5,$AE,$2D,$AE,$2D,$9E,$3B,$5C,$7C,$B8,$FA,$BF,$25
	.dc.b	$7D,$16,$F9,$FF,$69,$F2,$9B,$9C,$D7,$E5,$A7,$2F,$1A,$E1,$D0,$BE
	.dc.b	$11,$86,$CF,$F0,$5C,$18,$74,$CB,$F8,$33,$67,$F9,$CB,$63,$F5,$ED
	.dc.b	$7F,$C5,$CD,$9E,$77,$F3,$EB,$8B,$6B,$F8,$5F,$C1,$7D,$3E,$CF,$E9
	.dc.b	$B9,$7F,$AB,$AF,$FC,$DF,$87,$92,$B5,$1C,$0C,$E3,$BC,$3F,$A8,$35
	.dc.b	$F3,$CF,$F2,$D7,$CE,$D7,$CE,$DA,$ED,$5F,$A9,$6C,$F4,$9E,$19,$B6
	.dc.b	$7C,$5F,$92,$93,$98,$C5,$BE,$B9,$EA,$EE,$B5,$97,$BA,$72,$F8,$5B
	.dc.b	$87,$F6,$B6,$BF,$DA,$9A,$FF,$68,$C7,$FC,$3F,$3F,$E0,$CE,$EF,$0F
	.dc.b	$C0,$7F,$9A,$23,$04,$AF,$F4,$C5,$70,$7F,$C2,$33,$B7,$35,$E7,$9D
	.dc.b	$6F,$FE,$02,$6D,$FF,$30,$BE,$13,$9A,$F0,$FA,$DF,$36,$7A,$61,$FE
	.dc.b	$02,$6B,$C6,$3E,$6E,$0F,$AF,$F0,$1F,$2B,$F3,$1F,$F7,$AE,$5F,$E2
	.dc.b	$88,$C7,$C2,$FD,$99,$CE,$1F,$E1,$16,$BF,$E1,$06,$BF,$D1,$BC,$FF
	.dc.b	$F4,$05,$FE,$3B,$F3,$17,$E0,$7F,$62,$73,$5F,$E3,$49,$CD,$7F,$94
	.dc.b	$FD,$8F,$BB,$18,$9F,$F1,$CD,$70,$DF,$F8,$77,$1A,$B5,$1F,$E5,$7B
	.dc.b	$EC,$7E,$1B,$D2,$7C,$FE,$32,$FD,$0C,$61,$FC,$A3,$72,$FF,$17,$35
	.dc.b	$FF,$16,$B5,$FC,$27,$C8,$7D,$93,$F2,$7F,$BF,$FF,$24,$C7,$ED,$46
	.dc.b	$6F,$FC,$A6,$EB,$BF,$17,$FE,$1B,$F1,$FF,$31,$DA,$7F,$3B,$ED,$BE
	.dc.b	$AF,$31,$31,$FF,$12,$3A,$8C,$5F,$FB,$9F,$E2,$9E,$8F,$E8,$DB,$7F
	.dc.b	$EE,$EF,$9D,$7D,$13,$F0,$7D,$5F,$EB,$E7,$31,$FB,$B6,$D7,$F6,$16
	.dc.b	$E1,$FE,$2E,$3F,$C5,$EE,$1C,$35,$CD,$F7,$9E,$4D,$8E,$42,$78,$3F
	.dc.b	$C0,$34,$7B,$7F,$18,$FE,$03,$EC,$3C,$EC,$63,$FE,$47,$9B,$7F,$E8
	.dc.b	$0A,$D4,$E4,$FE,$43,$F0,$CF,$C0,$3F,$84,$7D,$EB,$E9,$7F,$86,$CE
	.dc.b	$6B,$FD,$EB,$39,$78,$D7,$0A,$DB,$63,$FD,$F7,$C1,$FC,$69,$BE,$4F
	.dc.b	$CD,$DD,$F2,$1F,$56,$30,$FC,$F8,$CC,$FC,$68,$C1,$1C,$13,$4F,$93
	.dc.b	$C1,$F7,$27,$F9,$87,$C7,$BE,$6B,$F9,$27,$E3,$4F,$AF,$FE,$53,$39
	.dc.b	$C7,$FC,$CF,$D9,$FE,$92,$AD,$47,$F4,$69,$D6,$4A,$D5,$F2,$C3,$AB
	.dc.b	$F9,$AF,$32,$61,$FD,$3B,$E4,$FF,$8D,$E1,$F6,$E7,$2F,$F9,$ED,$FE
	.dc.b	$77,$1A,$F3,$BF,$CC,$D1,$FD,$07,$E6,$9F,$57,$F4,$6F,$FD,$77,$F9
	.dc.b	$97,$F3,$AD,$3F,$C2,$15,$AE,$5F,$E8,$B3,$9C,$FD,$BF,$8A,$FD,$BC
	.dc.b	$3F,$1F,$82,$31,$1F,$9B,$8C,$73,$7F,$5B,$6D,$7F,$90,$5B,$3C,$AB
	.dc.b	$8B,$DA,$B8,$46,$21,$E4,$BC,$DF,$D9,$C7,$CB,$C3,$FD,$6B,$D3,$FC
	.dc.b	$05,$FE,$EB,$3A,$C7,$FD,$8B,$39,$AF,$F6,$DC,$E6,$B8,$F8,$5F,$D9
	.dc.b	$23,$13,$FD,$A2,$B8,$5F,$FB,$3B,$EF,$DF,$2D,$FD,$9D,$B8,$FF,$81
	.dc.b	$3F,$41,$FE,$45,$5B,$AD,$12,$75,$E5,$DC,$7F,$88,$7A,$BF,$C3,$CF
	.dc.b	$CF,$7D,$DD,$AF,$E2,$1D,$C8,$FC,$F1,$23,$F8,$1B,$6B,$F3,$73,$B9
	.dc.b	$C9,$23,$FD,$C5,$C5,$FE,$15,$CB,$FD,$D3,$F8,$03,$E3,$7F,$03,$ED
	.dc.b	$7F,$80,$E3,$18,$7E,$78,$FF,$62,$FC,$13,$E7,$FF,$CB,$F1,$FE,$7D
	.dc.b	$8C,$3F,$E1,$B6,$B9,$8D,$E5,$FE,$F3,$FA,$2F,$E2,$B3,$FE,$2C,$6C
	.dc.b	$7D,$D1,$D5,$FE,$47,$F0,$FF,$C3,$33,$AD,$3F,$AD,$FE,$5D,$FC,$5A
	.dc.b	$F8,$74,$A3,$13,$F9,$EF,$CB,$9E,$AF,$F8,$8A,$73,$9F,$FB,$83,$35
	.dc.b	$6A,$3F,$D3,$11,$8A,$FF,$1B,$36,$33,$3B,$C7,$F9,$9C,$E5,$B1,$97
	.dc.b	$FA,$04,$65,$B1,$97,$FD,$E3,$E0,$FD,$FF,$A8,$FF,$1A,$72,$9F,$F1
	.dc.b	$CF,$EA,$DE,$8A,$FF,$1E,$CE,$71,$FF,$2A,$CF,$C3,$9F,$E9,$DA,$A3
	.dc.b	$11,$FE,$D4,$6B,$FA,$E6,$C6,$FC,$FF,$F2,$94,$60,$8F,$F9,$42,$FF
	.dc.b	$A9,$6D,$FE,$99,$CB,$F2,$DE,$8F,$F9,$3B,$F0,$DF,$E1,$59,$79,$A3
	.dc.b	$1F,$05,$FE,$5D,$9F,$F2,$3C,$FE,$F9,$B9,$9A,$49,$E0,$4D,$3F,$D9
	.dc.b	$98,$3E,$F7,$F8,$F6,$5E,$36,$CF,$FB,$CB,$EB,$5F,$82,$FF,$97,$E7
	.dc.b	$39,$FF,$0E,$9C,$D7,$F9,$BD,$B1,$FD,$BB,$34,$62,$3E,$6B,$5E,$D7
	.dc.b	$FE,$79,$8C,$47,$58,$C1,$FE,$78,$BF,$E1,$B7,$F8,$06,$CF,$F2,$0F
	.dc.b	$77,$FA,$DB,$FA,$3C,$BD,$DE,$37,$F9,$EA,$73,$3F,$79,$A8,$E5,$D1
	.dc.b	$FE,$8B,$8C,$C6,$30,$FF,$46,$46,$33,$34,$DE,$0F,$8A,$7C,$67,$FA
	.dc.b	$1D,$B7,$FE,$A4,$FB,$17,$CE,$66,$BF,$D1,$F1,$89,$FF,$4D,$37,$2F
	.dc.b	$98,$D7,$F2,$ED,$7D,$59,$23,$11,$FD,$11,$B1,$9B,$FF,$54,$46,$B1
	.dc.b	$F4,$7B,$7E,$73,$FD,$43,$F9,$77,$F1,$2E,$2F,$CC,$68,$9C,$E7,$FE
	.dc.b	$B1,$AF,$E2,$F9,$FE,$A9,$AF,$91,$AF,$F5,$6F,$D7,$7F,$90,$BF,$C4
	.dc.b	$9B,$87,$FB,$2F,$F6,$4F,$B6,$FF,$9C,$A3,$37,$EE,$6D,$73,$F0,$7F
	.dc.b	$D7,$54,$BE,$37,$B9,$FF,$5C,$4F,$F7,$53,$F6,$45,$8F,$93,$F8,$B7
	.dc.b	$F3,$F7,$F8,$6C,$66,$FF,$D8,$AD,$D3,$FC,$2D,$E8,$FF,$B1,$AF,$85
	.dc.b	$6A,$3F,$D9,$3F,$75,$FE,$14,$E1,$37,$F6,$9F,$F4,$76,$5F,$62,$34
	.dc.b	$4F,$33,$FD,$A3,$3A,$C3,$FB,$EC,$F1,$E5,$FF,$70,$4F,$FA,$7A,$7F
	.dc.b	$8C,$B5,$FC,$1B,$5E,$3F,$0D,$E4,$3F,$DB,$BF,$79,$FE,$C0,$C7,$C5
	.dc.b	$19,$7F,$DF,$38,$23,$7F,$1F,$9B,$FE,$A0,$7F,$9D,$79,$7F,$CD,$AF
	.dc.b	$F9,$B9,$AE,$4A,$7E,$94,$DF,$B0,$FB,$EF,$F4,$3C,$61,$AF,$F9,$E2
	.dc.b	$35,$E8,$F4,$6C,$72,$FB,$EF,$F2,$43,$79,$35,$18,$8E,$4D,$7E,$87
	.dc.b	$D0,$7F,$8C,$F5,$B1,$F0,$23,$0F,$F8,$B3,$6F,$D0,$39,$21,$FD,$AB
	.dc.b	$64,$F0,$E2,$57,$0A,$FE,$DF,$F7,$DF,$E8,$E7,$E3,$CA,$31,$1F,$CA
	.dc.b	$5A,$FF,$54,$6B,$F8,$7F,$5B,$B6,$1F,$E2,$E8,$FA,$F1,$E1,$9E,$79
	.dc.b	$FF,$B6,$AB,$55,$E7,$AE,$26,$FE,$23,$DF,$D1,$FE,$B8,$F5,$EF,$07
	.dc.b	$FB,$72,$FF,$C3,$82,$74,$E3,$3F,$F4,$E7,$87,$C2,$E4,$F0,$AD,$BF
	.dc.b	$E1,$0F,$77,$FD,$A4,$F1,$FB,$EF,$F4,$53,$7E,$0B,$E5,$46,$23,$FC
	.dc.b	$3D,$B3,$EF,$9E,$5E,$AE,$57,$13,$CF,$5B,$9F,$A9,$F7,$DF,$2A,$7F
	.dc.b	$CC,$ED,$D4,$D3,$59,$FF,$B1,$FE,$5B,$E2,$FC,$41,$A7,$4E,$1F,$C0
	.dc.b	$79,$0F,$9A,$FD,$0E,$3F,$CC,$DB,$19,$CB,$F9,$EC,$60,$8E,$08,$7F
	.dc.b	$B6,$31,$FF,$6C,$47,$88,$91,$FD,$2D,$FC,$87,$DE,$BE,$2F,$C2,$D9
	.dc.b	$FC,$7B,$5F,$EE,$CD,$7F,$C4,$2D,$7F,$A5,$B5,$FF,$46,$B7,$03,$9B
	.dc.b	$FC,$27,$DF,$7E,$29,$34,$F2,$DF,$E9,$7D,$27,$DB,$7C,$66,$E3,$FE
	.dc.b	$0B,$8C,$11,$C1,$0F,$F0,$74,$70,$7E,$35,$FF,$90,$5F,$C8,$35,$FF
	.dc.b	$02,$37,$85,$DD,$FE,$1F,$BF,$DC,$7D,$F6,$EF,$FC,$86,$D7,$FB,$6B
	.dc.b	$5F,$F5,$7B,$74,$34,$9E,$D7,$E0,$9E,$18,$BE,$F7,$F9,$65,$FE,$95
	.dc.b	$AF,$F8,$65,$B9,$79,$A3,$BD,$FF,$0D,$FB,$EF,$A4,$4D,$7F,$1E,$D9
	.dc.b	$FE,$1A,$D7,$FC,$68,$FD,$7B,$FD,$65,$FF,$94,$DB,$A7,$F9,$B1,$B5
	.dc.b	$FE,$CC,$AF,$C4,$57,$FB,$8B,$E8,$5F,$7D,$FC,$FB,$E0,$11,$8B,$FF
	.dc.b	$41,$B5,$FF,$3F,$B5,$FF,$3E,$B5,$FF,$46,$35,$FF,$4D,$B6,$3A,$CF
	.dc.b	$EB,$3E,$C7,$E2,$E5,$FE,$20,$6D,$FF,$B8,$9B,$5C,$23,$FD,$51,$F2
	.dc.b	$9F,$CD,$7E,$14,$D3,$F9,$FC,$BE,$FB,$F6,$D5,$FE,$AD,$F8,$77,$DF
	.dc.b	$7F,$18,$8F,$F2,$41,$FE,$BE,$70,$E2,$37,$E2,$FF,$62,$6C,$7F,$40
	.dc.b	$F4,$3F,$25,$A7,$FA,$6B,$E9,$FF,$D8,$D3,$FC,$9D,$AF,$FA,$DA,$76
	.dc.b	$FF,$95,$FE,$F9,$F6,$1F,$DB,$1B,$1F,$DC,$1A,$FF,$74,$6B,$E8,$0E
	.dc.b	$1F,$5B,$5E,$D1,$D4,$FE,$F7,$CF,$F2,$3E,$F7,$FA,$3E,$E6,$9F,$DA
	.dc.b	$2B,$AD,$7F,$4C,$FB,$E7,$CA,$D2,$4F,$F4,$9F,$82,$7C,$86,$23,$FC
	.dc.b	$F5,$5C,$1F,$E9,$4D,$7F,$DB,$33,$C1,$FE,$A5,$5F,$C0,$FA,$3F,$D5
	.dc.b	$DF,$79,$F0,$9B,$8F,$F7,$59,$D7,$07,$F9,$A1,$B7,$FD,$76,$F8,$4F
	.dc.b	$FB,$39,$B0,$EF,$F5,$F8,$1E,$FF,$BE,$FC,$5C,$FF,$88,$1A,$FF,$89
	.dc.b	$1A,$FD,$16,$BF,$8A,$6B,$FE,$93,$6A,$69,$AC,$BF,$17,$1F,$95,$FA
	.dc.b	$A7,$C7,$7F,$86,$39,$91,$89,$FE,$C7,$D1,$F9,$9B,$FF,$69,$7D,$CB
	.dc.b	$E2,$BF,$C0,$2D,$AF,$EF,$0D,$53,$FC,$4F,$47,$CE,$7C,$8E,$75,$EE
	.dc.b	$BA,$E1,$D0,$FF,$12,$57,$CE,$B7,$CF,$7E,$27,$1F,$F0,$37,$D7,$D7
	.dc.b	$FD,$74,$DC,$3C,$13,$78,$BF,$D3,$ED,$AF,$F0,$A5,$5C,$CE,$70,$EF
	.dc.b	$D1,$F5,$FE,$FB,$FC,$A7,$C4,$69,$30,$FB,$4D,$7E,$D3,$5F,$E0,$38
	.dc.b	$7F,$0A,$F6,$BF,$46,$48,$FF,$13,$9F,$66,$7F,$C8,$9F,$44,$E6,$FF
	.dc.b	$17,$8F,$BE,$A6,$99,$5F,$E2,$97,$4D,$8C,$BF,$53,$6F,$F6,$0F,$DF
	.dc.b	$7F,$89,$23,$FC,$68,$D7,$47,$E6,$5C,$70,$7D,$BF,$BE,$78,$3F,$E5
	.dc.b	$99,$CD,$F4,$F6,$30,$7F,$97,$6B,$78,$7B,$BE,$FB,$FD,$C6,$FF,$A0
	.dc.b	$1A,$FF,$A1,$1A,$FF,$11,$6B,$FC,$A9,$AF,$FB,$CD,$A9,$A6,$B7,$FE
	.dc.b	$72,$FF,$12,$24,$78,$FE,$2B,$F3,$7E,$54,$8F,$CD,$BF,$E7,$5F,$AA
	.dc.b	$7D,$5E,$67,$E2,$E1,$F9,$AA,$C6,$DF,$9F,$BD,$D7,$CC,$F0,$FE,$31
	.dc.b	$1F,$F3,$17,$DF,$7F,$99,$DF,$9F,$D0,$E0,$FF,$7F,$C7,$12,$1A,$9E
	.dc.b	$18,$7E,$B7,$EF,$BF,$8E,$4F,$F3,$7B,$46,$23,$FA,$8C,$63,$C7,$FF
	.dc.b	$27,$CE,$64,$7F,$17,$FB,$07,$CC,$7F,$A5,$5B,$1F,$E7,$96,$BE,$89
	.dc.b	$F0,$E1,$FB,$F9,$CB,$FC,$07,$E8,$1F,$25,$FE,$9F,$6D,$F3,$C3,$B1
	.dc.b	$99,$91,$93,$FD,$11,$EF,$BE,$8F,$FC,$CE,$D8,$FF,$52,$B5,$F8,$4D
	.dc.b	$7F,$63,$F0,$2F,$BE,$FF,$4C,$B6,$3F,$D2,$0D,$7E,$9C,$F3,$CB,$F8
	.dc.b	$D7,$DF,$7F,$5F,$DC,$D2,$79,$FF,$D8,$30,$7D,$77,$62,$63,$FE,$A1
	.dc.b	$AF,$E8,$BC,$FF,$E8,$59,$CC,$FE,$32,$B8,$BB,$BC,$E3,$FE,$3D,$BE
	.dc.b	$13,$E8,$BF,$F6,$56,$1F,$D4,$31,$7D,$F7,$FB,$3A,$BF,$D9,$2D,$7F
	.dc.b	$0C,$D7,$F0,$AD,$7F,$07,$7F,$E9,$9F,$B0,$F8,$E4,$8F,$F5,$D4,$F3
	.dc.b	$C7,$F9,$F6,$EF,$BE,$FF,$4D,$B6,$BF,$DB,$2D,$7F,$2C,$D7,$F2,$AD
	.dc.b	$7F,$27,$8F,$E2,$27,$99,$23,$FC,$28,$7F,$B3,$7E,$83,$FD,$5C,$FF
	.dc.b	$B2,$5A,$FE,$D9,$B7,$E6,$9F,$F7,$85,$FF,$94,$AB,$F4,$3E,$0F,$8B
	.dc.b	$DE,$FF,$1B,$B6,$32,$79,$27,$83,$FE,$6F,$3F,$BF,$38,$8E,$11,$9C
	.dc.b	$7F,$43,$5D,$23,$93,$5E,$2F,$F9,$1B,$E6,$3D,$7A,$B3,$7D,$F6,$78
	.dc.b	$BE,$AC,$63,$0F,$DC,$DF,$C2,$C7,$BC,$62,$BB,$B4,$F9,$17,$CD,$A6
	.dc.b	$67,$8A,$AE,$9F,$7D,$FC,$01,$BC,$FC,$69,$39,$C3,$EC,$BF,$E4,$9A
	.dc.b	$DD,$F5,$8D,$3F,$E2,$56,$D7,$28,$E9,$18,$25,$E2,$3D,$07,$F7,$8C
	.dc.b	$DF,$7B,$ED,$8C,$37,$CF,$FF,$47,$BF,$07,$D6,$7D,$F7,$F9,$D7,$3F
	.dc.b	$C3,$D6,$AF,$E0,$69,$FE,$78,$8C,$4E,$B1,$EF,$5A,$18,$74,$F7,$FC
	.dc.b	$B7,$CE,$7F,$CD,$4D,$7A,$57,$27,$2D,$7C,$1E,$CF,$F2,$E6,$E9,$FA
	.dc.b	$07,$FC,$2F,$19,$7F,$9C,$D7,$F8,$F2,$74,$D8,$F3,$1C,$23,$E1,$7B
	.dc.b	$FF,$D7,$38,$FE,$35,$FF,$3E,$CF,$E1,$B7,$4F,$0E,$1E,$FD,$18,$71
	.dc.b	$8E,$B9,$77,$C3,$F2,$5F,$7D,$96,$F9,$3F,$ED,$57,$3A,$DC,$BF,$7F
	.dc.b	$5B,$6D,$67,$0F,$F3,$54,$E5,$B5,$9E,$24,$E5,$B5,$98,$C1,$FE,$D0
	.dc.b	$D7,$F3,$CE,$FE,$F5,$F7,$DD,$F9,$FF,$61,$F1,$FD,$67,$37,$FA,$99
	.dc.b	$CB,$FA,$78,$CE,$7F,$83,$9C,$BF,$85,$C9,$3A,$D7,$F7,$BD,$0C,$7B
	.dc.b	$65,$FE,$F9,$CB,$FB,$BD,$F4,$6C,$67,$1F,$E4,$B3,$96,$C6,$5F,$F4
	.dc.b	$84,$65,$B1,$9C,$7C,$6E,$F2,$FE,$1D,$E3,$B0,$FE,$54,$DB,$FF,$0A
	.dc.b	$E1,$FE,$B0,$9C,$9F,$E0,$8E,$CF,$F3,$F4,$66,$7F,$C2,$4D,$7D,$75
	.dc.b	$FC,$9D,$F4,$B5,$F3,$B7,$1F,$8E,$D9,$E5,$3C,$2F,$FA,$0E,$0E,$07
	.dc.b	$C5,$79,$F2,$FE,$CD,$18,$AF,$31,$FD,$33,$ED,$9E,$96,$E3,$FA,$FC
	.dc.b	$EF,$1F,$F1,$9D,$6A,$7F,$A7,$E8,$8C,$4F,$F8,$BD,$AF,$F8,$B9,$AF
	.dc.b	$4E,$EE,$CB,$3F,$44,$62,$BF,$D6,$3E,$9E,$A7,$38,$7F,$85,$2B,$18
	.dc.b	$FF,$8A,$A3,$D7,$87,$FB,$37,$E0,$3F,$CE,$71,$B2,$47,$F8,$2E,$BB
	.dc.b	$4E,$A3,$13,$CB,$1F,$0B,$EE,$9D,$4F,$F9,$21,$AF,$F7,$46,$BE,$BB
	.dc.b	$7D,$F7,$98,$F6,$7D,$F7,$FB,$EA,$FF,$C5,$B1,$8F,$84,$FB,$D3,$9B
	.dc.b	$FF,$2C,$35,$FF,$2B,$B5,$FF,$01,$F9,$CF,$57,$E4,$7D,$EB,$EF,$BE
	.dc.b	$54,$E7,$2F,$64,$E5,$FF,$2C,$57,$0A,$DB,$63,$24,$BF,$F3,$9F,$A2
	.dc.b	$F9,$CD,$CF,$C7,$8C,$3F,$84,$6C,$FF,$9C,$5A,$FF,$9C,$1A,$FE,$3B
	.dc.b	$E9,$7E,$D9,$F8,$1F,$BE,$FE,$55,$CF,$F9,$88,$CD,$FF,$95,$1D,$78
	.dc.b	$18,$BF,$F0,$E7,$8F,$F8,$7E,$F3,$F8,$BF,$65,$F7,$D9,$89,$8F,$F8
	.dc.b	$95,$D4,$62,$FF,$D3,$7F,$18,$F4,$7F,$24,$DB,$FC,$7F,$CD,$3E,$81
	.dc.b	$F6,$5A,$BF,$CE,$CE,$63,$F7,$8D,$AF,$F1,$BB,$70,$FF,$47,$9F,$E9
	.dc.b	$07,$0E,$1A,$E6,$FB,$CF,$26,$C7,$21,$3C,$1F,$F2,$27,$B1,$F1,$4F
	.dc.b	$D7,$7A,$EF,$3F,$18,$7F,$91,$E6,$D7,$FA,$CA,$B5,$39,$3F,$4D,$F0
	.dc.b	$8F,$AA,$FD,$B7,$DD,$BE,$8F,$F7,$53,$9A,$FD,$AF,$8A,$AD,$47,$2B
	.dc.b	$E3,$7C,$1B,$1C,$38,$3F,$89,$73,$FF,$9F,$DB,$1F,$0F,$E3,$BE,$AF
	.dc.b	$95,$F4,$38,$51,$F4,$1B,$3F,$1E,$30,$47,$04,$DB,$E5,$70,$FD,$D9
	.dc.b	$FE,$65,$F1,$AF,$A1,$FE,$29,$F7,$CF,$BE,$FE,$31,$39,$C7,$DF,$39
	.dc.b	$7F,$86,$46,$71,$FF,$73,$CE,$5F,$F6,$FC,$67,$5F,$F4,$E4,$E5,$FF
	.dc.b	$4F,$75,$FF,$2F,$BF,$F3,$E7,$9B,$FC,$9F,$E4,$7F,$91,$63,$FE,$7D
	.dc.b	$8F,$D1,$12,$7F,$8E,$4E,$AF,$9C,$FF,$32,$F9,$87,$D5,$FD,$4C,$3F
	.dc.b	$DE,$3F,$77,$FC,$F2,$FF,$B0,$56,$B9,$7F,$9B,$4E,$72,$FE,$C1,$F7
	.dc.b	$DF,$7F,$0F,$C9,$60,$8C,$47,$E7,$A3,$1C,$DF,$D7,$DB,$5F,$E9,$A6
	.dc.b	$CF,$2A,$E2,$F6,$AE,$11,$88,$7A,$0F,$37,$F6,$D0,$BF,$F6,$37,$4F
	.dc.b	$F0,$4C,$7F,$9F,$56,$A7,$3E,$17,$F5,$18,$C4,$FF,$B0,$2B,$85,$F3
	.dc.b	$8F,$F5,$9C,$EA,$3F,$AD,$4E,$AF,$FD,$A9,$F7,$CF,$94,$FE,$B4,$DC
	.dc.b	$3F,$A7,$7D,$F7,$F2,$6A,$DD,$68,$93,$AF,$2E,$E3,$FC,$53,$D4,$FE
	.dc.b	$26,$7E,$7F,$EE,$AD,$7F,$14,$EE,$47,$E8,$49,$1F,$C1,$DB,$5F,$9E
	.dc.b	$9D,$CE,$49,$1F,$C2,$F7,$71,$7F,$86,$F3,$FF,$78,$7D,$F3,$E3,$7F
	.dc.b	$07,$ED,$FF,$77,$8C,$63,$FD,$D3,$E0,$DF,$6B,$FE,$2B,$8F,$F6,$BC
	.dc.b	$61,$FF,$10,$35,$CC,$6F,$3F,$17,$D0,$BC,$AF,$F0,$33,$6B,$FC,$03
	.dc.b	$D7,$FE,$49,$F1,$33,$5F,$E0,$BF,$BE,$FE,$33,$87,$D9,$3F,$0E,$E3
	.dc.b	$6F,$CE,$FD,$F3,$D5,$E1,$E7,$7E,$96,$31,$9F,$FB,$BB,$34,$63,$1F
	.dc.b	$E6,$11,$8B,$FF,$1D,$36,$B8,$56,$EB,$F9,$FC,$E5,$E3,$19,$42,$34
	.dc.b	$FF,$46,$F3,$3E,$EF,$D3,$FF,$89,$79,$CF,$F8,$A7,$EF,$BD,$98,$F2
	.dc.b	$AD,$BF,$E1,$D8,$F8,$77,$FD,$43,$9D,$18,$8F,$0C,$63,$4F,$F2,$94
	.dc.b	$E6,$7F,$5E,$D9,$D5,$7F,$91,$23,$06,$EF,$F2,$7E,$8F,$9C,$FF,$1F
	.dc.b	$7D,$F7,$F0,$9C,$3F,$C8,$59,$FF,$5A,$DF,$FC,$DD,$92,$BF,$C9,$35
	.dc.b	$FB,$E6,$C6,$7E,$0B,$FC,$CD,$18,$23,$82,$67,$FE,$66,$FB,$DF,$E3
	.dc.b	$D8,$7F,$9E,$F3,$46,$27,$E0,$FD,$63,$EF,$BF,$CA,$53,$9C,$FF,$87
	.dc.b	$CE,$6B,$FC,$C6,$D8,$FD,$43,$5F,$9C,$D7,$AD,$7F,$9F,$63,$11,$D2
	.dc.b	$30,$7F,$A0,$EF,$CB,$E3,$78,$8F,$F3,$37,$BB,$FD,$72,$BF,$39,$3A
	.dc.b	$C3,$FC,$B5,$39,$9F,$BC,$D7,$EE,$B5,$FE,$52,$DD,$7F,$C9,$9D,$1F
	.dc.b	$E9,$18,$C1,$1F,$27,$C4,$BE,$3B,$FC,$EC,$DC,$3F,$CD,$5F,$60,$F9
	.dc.b	$CC,$D7,$F9,$F6,$31,$3F,$E8,$A6,$E5,$FC,$3D,$AF,$E5,$5A,$FE,$51
	.dc.b	$AF,$F4,$06,$E4,$69,$AC,$D8,$BD,$2F,$39,$FE,$8F,$FB,$EF,$E1,$DC
	.dc.b	$5F,$83,$9D,$65,$FE,$83,$9F,$E2,$59,$7E,$9D,$AF,$E9,$9A,$FF,$52
	.dc.b	$6E,$4F,$BE,$FE,$3D,$3F,$C4,$63,$1E,$0F,$FA,$8B,$17,$DF,$7D,$B7
	.dc.b	$F4,$37,$C1,$B3,$FE,$A5,$8C,$63,$FE,$CF,$AE,$0F,$EE,$23,$38,$7F
	.dc.b	$AB,$6F,$FB,$A1,$F9,$F2,$CF,$EF,$BE,$FB,$F9,$E3,$FC,$1E,$33,$3C
	.dc.b	$7D,$2F,$F5,$94,$63,$25,$F0,$AD,$5F,$FA,$A7,$EE,$3F,$C1,$CF,$F0
	.dc.b	$A6,$C7,$F7,$6F,$BE,$FF,$4A,$3F,$0E,$33,$66,$98,$9D,$E1,$FC,$12
	.dc.b	$31,$CB,$FE,$E5,$9E,$77,$FC,$6D,$AF,$F1,$A6,$BF,$E0,$69,$7C,$37
	.dc.b	$FB,$3D,$B8,$FF,$B0,$7E,$EF,$FD,$89,$8F,$7A,$D1,$27,$5F,$21,$70
	.dc.b	$FE,$98,$7F,$9E,$B8,$BF,$CE,$4F,$F9,$CD,$AE,$4A,$7E,$50,$DF,$B2
	.dc.b	$FB,$EF,$F4,$64,$61,$AF,$F9,$F6,$35,$E8,$F4,$6C,$72,$FB,$EF,$F1
	.dc.b	$03,$73,$D4,$62,$39,$35,$FE,$7F,$F9,$5F,$FA,$A1,$FF,$7E,$B5,$FF
	.dc.b	$01,$35,$FE,$90,$D7,$8F,$77,$BB,$89,$39,$9F,$F1,$AB,$63,$33,$F1
	.dc.b	$3C,$97,$E8,$3F,$E1,$8B,$FE,$DC,$D7,$FB,$6B,$5F,$ED,$2D,$7F,$AB
	.dc.b	$FD,$FF,$F4,$77,$E8,$DA,$31,$1F,$E1,$B6,$BF,$D8,$5A,$F2,$F1,$BD
	.dc.b	$51,$88,$FF,$26,$B6,$7A,$4F,$BF,$DE,$C1,$1C,$12,$BC,$71,$96,$C7
	.dc.b	$E0,$BF,$2B,$7C,$AD,$24,$FE,$22,$32,$D8,$CB,$F8,$E8,$CB,$63,$F1
	.dc.b	$BF,$98,$BE,$F6,$92,$7F,$27,$19,$7F,$2B,$F9,$13,$EB,$BF,$2A,$D8
	.dc.b	$CC,$60,$EB,$E1,$48,$E6,$D7,$F4,$35,$FE,$40,$72,$43,$31,$FA,$4F
	.dc.b	$1D,$F8,$A3,$AF,$0D,$AF,$D5,$CE,$A7,$27,$EA,$BE,$51,$F4,$D4,$95
	.dc.b	$FB,$08,$CB,$63,$2F,$ED,$A3,$2D,$8F,$83,$F1,$F2,$3E,$2B,$5F,$DE
	.dc.b	$FE,$A6,$F9,$8F,$DD,$B6,$3F,$7E,$D7,$71,$A2,$3A,$6B,$FE,$6F,$FC
	.dc.b	$01,$F6,$9F,$AB,$6C,$7E,$71,$AE,$E3,$F8,$5F,$D1,$7D,$B8,$FE,$22
	.dc.b	$D7,$FC,$96,$7F,$93,$1C,$38,$6B,$9E,$2E,$64,$7F,$8C,$C6,$97,$E5
	.dc.b	$BF,$E8,$17,$F8,$EB,$5F,$E4,$2D,$7C,$27,$0F,$85,$F3,$ED,$23,$AB
	.dc.b	$5F,$4B,$5E,$35,$F8,$37,$27,$F2,$EF,$95,$7D,$4F,$F9,$E6,$30,$D7
	.dc.b	$FD,$F6,$D7,$E1,$C7,$0F,$81,$FE,$67,$F9,$5B,$D0,$A4,$AF,$E7,$31
	.dc.b	$96,$C6,$5F,$E7,$F1,$96,$C7,$4F,$B4,$91,$C9,$AF,$FA,$22,$5F,$3B
	.dc.b	$FC,$8D,$B1,$FE,$8F,$6B,$B8,$F3,$FD,$44,$8F,$2B,$5F,$F4,$9F,$73
	.dc.b	$E5,$7F,$97,$36,$3F,$D3,$4D,$77,$1F,$D5,$3F,$1C,$FD,$C4,$7F,$5A
	.dc.b	$6B,$F6,$9A,$FE,$31,$AF,$FA,$61,$A9,$A4,$EC,$FE,$C3,$F7,$2F,$9D
	.dc.b	$A4,$AF,$EC,$F1,$96,$F0,$F0,$23,$82,$1F,$DC,$21,$1A,$8C,$3E,$0F
	.dc.b	$2E,$47,$66,$BC,$5A,$FE,$62,$51,$88,$F8,$1F,$0F,$F1,$C8,$FF,$7E
	.dc.b	$C5,$18,$8F,$94,$D7,$E5,$3F,$0B,$F1,$89,$1F,$01,$AF,$AF,$EE,$BE
	.dc.b	$39,$1F,$8A,$D7,$46,$08,$7C,$3D,$3E,$79,$1F,$9A,$D7,$46,$08,$7F
	.dc.b	$86,$BF,$2F,$7C,$A7,$F8,$61,$B1,$FE,$1C,$6B,$B8,$D1,$1D,$35,$FF
	.dc.b	$12,$F8,$CF,$C4,$BF,$C4,$6D,$8F,$F1,$3B,$5D,$C6,$BE,$F6,$46,$9A
	.dc.b	$F4,$6B,$FC,$42,$51,$88,$FE,$47,$0F,$77,$C2,$48,$F0,$B5,$F3,$B5
	.dc.b	$E8,$FC,$3F,$97,$91,$F0,$DA,$FC,$96,$BF,$C8,$B1,$46,$23,$C7,$E2
	.dc.b	$FC,$22,$3F,$09,$F8,$BE,$DC,$8F,$8A,$D7,$E5,$B5,$F8,$0D,$7F,$A0
	.dc.b	$FE,$D3,$FD,$21,$FE,$66,$D7,$3E,$27,$F9,$52,$31,$00,$E1,$B1,$B2
	.dc.b	$3F,$E5,$EF,$A0,$7A,$3F,$E5,$E6,$C6,$63,$07,$F9,$1F,$F2,$CF,$E2
	.dc.b	$51,$FD,$5B,$D3,$46,$23,$C5,$18,$E2,$E0,$7F,$92,$7F,$07,$7C,$B5
	.dc.b	$25,$7F,$9C,$BD,$84,$EA,$30,$D7,$FA,$AC,$6A,$78,$C6,$27,$4F,$F6
	.dc.b	$A9,$FF,$67,$7C,$67,$F9,$67,$E2,$CD,$24,$FF,$94,$B2,$FF,$0E,$B7
	.dc.b	$D4,$F8,$04,$7E,$47,$95,$F1,$C8,$E0,$FF,$2D,$FD,$CF,$F6,$17,$E3
	.dc.b	$35,$F8,$AD,$5A,$A3,$0D,$F3,$3F,$CE,$D1,$82,$38,$24,$F8,$1A,$FF
	.dc.b	$7A,$A7,$DC,$FF,$91,$63,$0D,$7F,$C1,$F1,$AE,$9F,$F5,$0B,$63,$FC
	.dc.b	$EB,$E4,$7E,$89,$BD,$1F,$E8,$98,$C4,$7F,$82,$32,$FE,$CB,$3F,$EA
	.dc.b	$8F,$B5,$79,$74,$95,$FE,$AE,$8C,$B6,$32,$FF,$AD,$23,$2D,$8C,$FC
	.dc.b	$07,$F9,$3E,$DF,$8E,$FF,$62,$F8,$5F,$F3,$8C,$63,$CB,$FF,$5F,$7B
	.dc.b	$BF,$E2,$CC,$B9,$E1,$FE,$15,$5F,$80,$FF,$8C,$38,$7F,$C4,$31,$8D
	.dc.b	$3F,$59,$18,$9D,$47,$36,$BF,$B1,$9E,$0F,$E0,$5B,$3C,$63,$99,$C4
	.dc.b	$A7,$19,$F3,$BF,$84,$FC,$AD,$F8,$4E,$33,$46,$27,$F8,$06,$9F,$E8
	.dc.b	$CC,$7C,$2D,$C3,$85,$6E,$93,$F1,$A3,$0E,$E3,$9E,$7F,$C2,$F8,$5F
	.dc.b	$96,$FF,$B3,$74,$FC,$3B,$6B,$4D,$7F,$C7,$AD,$9E,$3F,$0B,$F1,$9B
	.dc.b	$3F,$01,$AF,$F9,$1D,$B5,$DA,$7E,$07,$F0,$0F,$FB,$D7,$0F,$E1,$AD
	.dc.b	$BE,$F3,$F2,$16,$8F,$53,$FC,$9F,$1F,$02,$FF,$05,$18,$CF,$E4,$36
	.dc.b	$3F,$95,$7D,$0B,$EF,$7F,$DD,$95,$AD,$3E,$2E,$1F,$7F,$0E,$0D,$8C
	.dc.b	$DF,$1C,$BC,$99,$78,$E7,$F4,$17,$E4,$BF,$C7,$B5,$F2,$C7,$E9,$7D
	.dc.b	$07,$C1,$D2,$4F,$E9,$63,$2D,$8F,$2B,$F0,$A3,$39,$7F,$B5,$32,$7D
	.dc.b	$EB,$E1,$3F,$D7,$13,$F8,$9A,$DE,$DF,$A7,$6E,$1F,$9D,$6E,$1C,$AF
	.dc.b	$8D,$6F,$1F,$8B,$93,$F0,$2F,$E7,$FC,$3F,$23,$95,$C4,$F9,$7E,$3A
	.dc.b	$FF,$A1,$2F,$6F,$D7,$3F,$92,$CE,$EF,$AC,$E9,$FE,$32,$D9,$FC,$C3
	.dc.b	$5E,$9E,$3F,$FB,$56,$76,$D8,$DB,$5D,$C7,$E5,$67,$94,$79,$FE,$C7
	.dc.b	$CF,$8F,$EC,$BD,$8F,$4C,$E6,$FD,$DD,$1F,$E5,$27,$E9,$E6,$C3,$8C
	.dc.b	$E6,$BD,$C7,$13,$E0,$61,$D6,$F8,$D7,$96,$75,$1F,$EB,$6B,$E8,$FF
	.dc.b	$0A,$8C,$9F,$EC,$A8,$CB,$FD,$4D,$B1,$C6,$7A,$46,$27,$87,$A1,$FE
	.dc.b	$FB,$EC,$FE,$5D,$DC,$FA,$DF,$7D,$AB,$58,$FF,$80,$2B,$4F,$C0,$8F
	.dc.b	$EB,$D3,$B8,$FF,$66,$E1,$FE,$17,$8C,$38,$6B,$ED,$9E,$24,$9E,$31
	.dc.b	$F8,$98,$FE,$D9,$93,$D1,$7E,$01,$86,$DF,$82,$75,$39,$3F,$BA,$71
	.dc.b	$3E,$CB,$95,$79,$A3,$17,$A9,$CE,$3F,$DB,$5A,$FF,$88,$FE,$F3,$FC
	.dc.b	$E3,$7D,$DB,$1F,$D9,$5A,$EF,$D9,$F4,$C6,$23,$FC,$52,$D7,$FC,$0A
	.dc.b	$DC,$8D,$24,$F2,$25,$7C,$6D,$1F,$5B,$F4,$89,$19,$7E,$4F,$DE,$7F
	.dc.b	$62,$8F,$8C,$DC,$7F,$BE,$35,$FF,$20,$35,$FF,$28,$7A,$2F,$CE,$DF
	.dc.b	$53,$FE,$1F,$8C,$36,$33,$7F,$E3,$F6,$D7,$F8,$F3,$CA,$7C,$4F,$16
	.dc.b	$D7,$F9,$15,$B3,$B9,$D1,$27,$4D,$8F,$4F,$83,$EE,$24,$FB,$9F,$F1
	.dc.b	$97,$E0,$EF,$A3,$A4,$9F,$C7,$C6,$5F,$CC,$FC,$5B,$A9,$CD,$FE,$3D
	.dc.b	$8C,$36,$33,$3C,$1B,$8F,$F9,$DC,$FF,$3C,$57,$09,$D5,$71,$23,$E0
	.dc.b	$F9,$97,$A3,$DD,$B1,$98,$C1,$1C,$12,$FC,$02,$D5,$CA,$33,$2F,$D1
	.dc.b	$7F,$D4,$2F,$EE,$0C,$7A,$5F,$E5,$B6,$99,$8D,$CF,$EF,$BE,$CD,$F1
	.dc.b	$BF,$E5,$F6,$D7,$EE,$9B,$3B,$9D,$12,$74,$D8,$F4,$E3,$E8,$FA,$BF
	.dc.b	$84,$4A,$F8,$CD,$9D,$E4,$8C,$47,$D1,$AE,$55,$F3,$7E,$47,$E6,$11
	.dc.b	$C1,$27,$9C,$65,$B1,$FC,$73,$C9,$7E,$40,$F2,$FF,$D0,$11,$86,$C6
	.dc.b	$67,$83,$6B,$E4,$7B,$BF,$1C,$93,$F1,$DB,$1A,$8C,$11,$F7,$F3,$FE
	.dc.b	$14,$91,$F8,$96,$C6,$A3,$04,$7E,$3C,$FE,$34,$91,$F9,$16,$C6,$A3
	.dc.b	$04,$7F,$1B,$87,$C2,$F8,$EF,$79,$27,$DE,$FE,$0F,$D2,$91,$F8,$36
	.dc.b	$D7,$E2,$5A,$FE,$63,$83,$FC,$79,$3B,$3F,$49,$EB,$FE,$8C,$8E,$09
	.dc.b	$3F,$82,$8C,$B6,$3F,$59,$E3,$7E,$AC,$8E,$09,$1F,$8A,$8C,$B6,$3F
	.dc.b	$AD,$F3,$BE,$DB,$F3,$73,$FB,$D8,$C7,$27,$FB,$05,$B5,$C2,$B7,$5F
	.dc.b	$C0,$FF,$20,$78,$3F,$EC,$56,$D7,$F6,$06,$CF,$F6,$A6,$BF,$BC,$6B
	.dc.b	$CE,$79,$12,$7F,$DD,$DF,$3F,$FC,$6E,$7F,$DD,$6D,$7F,$DD,$0D,$7F
	.dc.b	$9E,$B5,$FF,$7E,$F7,$7F,$B5,$89,$3F,$DD,$3C,$74,$7F,$41,$F9,$1F
	.dc.b	$6D,$3C,$44,$61,$BA,$EC,$FF,$72,$71,$FF,$80,$AB,$F9,$33,$51,$FE
	.dc.b	$64,$F2,$F6,$70,$EE,$99,$FF,$81,$AB,$F8,$9D,$7E,$52,$72,$FE,$DD
	.dc.b	$AE,$67,$F9,$EF,$E3,$5E,$FE,$C3,$49,$E1,$FF,$BE,$7D,$7F,$F0,$6D
	.dc.b	$FF,$87,$1A,$FF,$2E,$6B,$FC,$FD,$AF,$FA,$E3,$97,$91,$30,$FF,$0E
	.dc.b	$CE,$97,$E5,$2F,$80,$FE,$60,$D8,$FE,$68,$DB,$FE,$72,$D7,$F9,$3F
	.dc.b	$DE,$A7,$FC,$53,$9A,$31,$3F,$D7,$1A,$F0,$9D,$C6,$23,$F9,$43,$5F
	.dc.b	$EB,$2D,$7F,$A6,$3F,$E3,$4F,$99,$FF,$47,$4F,$F8,$C5,$B3,$FD,$4B
	.dc.b	$91,$18,$8F,$EC,$91,$8F,$6B,$A7,$E3,$5F,$E2,$E9,$FE,$D0,$D7,$F0
	.dc.b	$6D,$4D,$27,$A7,$C8,$98,$7C,$8F,$C6,$FE,$59,$1C,$1F,$E4,$E9,$47
	.dc.b	$88,$91,$96,$D7,$F9,$72,$77,$39,$24,$7F,$24,$C7,$FC,$EC,$FF,$AF
	.dc.b	$1B,$C7,$F0,$A3,$FA,$3C,$69,$FD,$47,$57,$CD,$E7,$FD,$EC,$7F,$8C
	.dc.b	$23,$04,$BF,$F3,$2C,$EA,$78,$C6,$27,$FA,$9F,$D8,$7F,$AB,$A7,$F5
	.dc.b	$AD,$7F,$5B,$82,$31,$1F,$ED,$26,$BF,$BB,$DB,$C8,$4B,$F0,$BE,$7F
	.dc.b	$94,$91,$E4,$6C,$FF,$A0,$B1,$46,$23,$FC,$2B,$18,$9E,$07,$F9,$E7
	.dc.b	$EA,$DF,$65,$FE,$80,$6C,$66,$30,$47,$04,$AF,$C8,$4F,$07,$F2,$5F
	.dc.b	$59,$FE,$6D,$27,$AF,$FE,$8C,$6D,$7F,$A2,$5B,$3C,$27,$FC,$BD,$F8
	.dc.b	$73,$E6,$69,$2F,$DD,$19,$7F,$D2,$5F,$59,$FE,$7B,$25,$7F,$A6,$9B
	.dc.b	$1A,$8C,$11,$FF,$50,$DF,$F2,$2E,$AA,$9A,$47,$FD,$5F,$E1,$BE,$AF
	.dc.b	$FC,$7F,$18,$6B,$FE,$44,$8D,$7A,$5F,$EA,$E6,$C7,$2F,$82,$FF,$30
	.dc.b	$37,$C9,$D4,$62,$39,$35,$FE,$6F,$D8,$FA,$B7,$C5,$7F,$47,$6C,$FF
	.dc.b	$AD,$5A,$FF,$05,$6B,$FC,$11,$AF,$46,$A9,$ED,$3D,$4F,$D1,$FD,$5B
	.dc.b	$ED,$29,$2B,$F2,$51,$97,$FD,$8D,$F5,$8F,$AC,$FE,$B8,$D8,$CC,$60
	.dc.b	$8E,$08,$7F,$1D,$8E,$0F,$F1,$FF,$9A,$AF,$F1,$E6,$BF,$19,$AF,$AD
	.dc.b	$B8,$9A,$4A,$FE,$E0,$F0,$97,$D2,$7F,$94,$5F,$F7,$3B,$5F,$F7,$8B
	.dc.b	$5F,$F4,$1B,$5F,$F4,$A3,$5F,$ED,$ED,$7A,$FB,$3F,$92,$8C,$7D,$5B
	.dc.b	$DB,$FC,$AB,$6B,$FD,$EE,$D7,$FE,$00,$B5,$FF,$49,$FD,$67,$F5,$D7
	.dc.b	$FD,$FA,$D7,$FD,$2E,$1C,$3C,$3C,$BE,$FF,$2D,$23,$BB,$5E,$ED,$7F
	.dc.b	$D4,$34,$8C,$47,$F2,$0F,$99,$FF,$08,$93,$A7,$FC,$14,$FF,$5D,$87
	.dc.b	$96,$FA,$CF,$D9,$B6,$3C,$E6,$A3,$24,$73,$1B,$7D,$AD,$7F,$D6,$AD
	.dc.b	$7F,$D8,$D9,$3D,$3F,$E0,$2D,$8F,$8F,$18,$3F,$82,$FD,$63,$EB,$29
	.dc.b	$27,$CB,$19,$6C,$65,$F9,$F1,$96,$C7,$F8,$F7,$CC,$7D,$15,$24,$FF
	.dc.b	$7D,$7F,$8A,$FB,$1F,$76,$3F,$DB,$CD,$7F,$C7,$2D,$7F,$C8,$7C,$3C
	.dc.b	$08,$FC,$A8,$D1,$BF,$AC,$7C,$4F,$F2,$26,$CF,$D0,$6B,$F8,$86,$BF
	.dc.b	$EE,$09,$FB,$44,$E0,$F3,$92,$3C,$0D,$7F,$23,$1F,$E1,$DF,$53,$FD
	.dc.b	$FF,$F5,$9F,$E2,$AF,$89,$34,$95,$F8,$2F,$AC,$FF,$70,$B7,$DD,$FE
	.dc.b	$73,$18,$8F,$CE,$B5,$FF,$7A,$B5,$FF,$2F,$87,$0F,$F9,$89,$AE,$63
	.dc.b	$AF,$C3,$48,$F1,$35,$FE,$DB,$2F,$8A,$FF,$03,$B6,$3F,$BA,$35,$DC
	.dc.b	$7A,$3E,$16,$47,$B5,$AF,$F7,$8E,$C7,$D2,$7F,$84,$9B,$1F,$DF,$DA
	.dc.b	$EE,$3F,$AD,$FD,$5F,$EC,$E3,$F5,$2D,$7F,$80,$B5,$F8,$61,$C3,$F1
	.dc.b	$9A,$FF,$A1,$9A,$EF,$B3,$B9,$FD,$AB,$EB,$1F,$13,$49,$3F,$BA,$8C
	.dc.b	$B7,$87,$81,$1C,$10,$FE,$2F,$1F,$E9,$E7,$07,$A1,$C3,$AF,$2B,$C4
	.dc.b	$47,$C0,$D7,$46,$08,$7C,$4F,$85,$F9,$24,$7E,$4B,$5D,$18,$21,$E6
	.dc.b	$FA,$CF,$84,$47,$E0,$3F,$17,$D5,$91,$F0,$DA,$FB,$9A,$FB,$5A,$FE
	.dc.b	$62,$FE,$FB,$F0,$74,$FA,$04,$7E,$73,$5F,$92,$D7,$EF,$FD,$53,$D4
	.dc.b	$FF,$10,$B6,$3F,$97,$B5,$FE,$8E,$D7,$8C,$70,$23,$C1,$FF,$15,$F8
	.dc.b	$DF,$E5,$28,$FE,$72,$D7,$FA,$8B,$5F,$C8,$B5,$FD,$AA,$FA,$CE,$64
	.dc.b	$8E,$6D,$79,$B5,$E6,$F8,$BE,$B2,$47,$89,$AF,$89,$AF,$89,$AF,$F0
	.dc.b	$6B,$7C,$4F,$C6,$24,$7C,$66,$BF,$19,$AF,$CE,$7E,$1F,$89,$23,$E1
	.dc.b	$B5,$F8,$7F,$0B,$F2,$08,$FC,$86,$BA,$30,$43,$FB,$DF,$D6,$7F,$4A
	.dc.b	$7F,$B3,$B5,$CF,$87,$FE,$5E,$8C,$47,$F0,$06,$BF,$C9,$1A,$83,$86
	.dc.b	$C7,$F9,$A3,$EB,$3F,$C5,$51,$FD,$D9,$AF,$F8,$39,$AF,$F0,$E6,$BF
	.dc.b	$E5,$AF,$AA,$FC,$DB,$77,$FE,$FF,$E8,$A7,$33,$FE,$53,$6F,$C0,$BE
	.dc.b	$B1,$F4,$5F,$EB,$66,$D7,$FA,$8D,$AF,$FA,$89,$AF,$F4,$29,$E5,$3C
	.dc.b	$49,$1C,$5F,$F6,$0F,$D6,$3E,$52,$92,$7F,$D6,$11,$97,$FC,$5B,$8F
	.dc.b	$A7,$D2,$F5,$E4,$69,$27,$E4,$78,$FF,$2C,$8E,$0E,$7F,$58,$F2,$BE
	.dc.b	$2B,$63,$31,$82,$38,$21,$FC,$FB,$04,$F1,$8C,$35,$FE,$CB,$1A,$9E
	.dc.b	$71,$A7,$FC,$67,$4F,$AC,$FF,$24,$CF,$E1,$E3,$1D,$3A,$9C,$CA,$72
	.dc.b	$FC,$6F,$1B,$F5,$AD,$E7,$FC,$0C,$62,$3D,$0D,$7F,$AB,$FD,$4B,$C9
	.dc.b	$FF,$5C,$B6,$3F,$C8,$ED,$7F,$CB,$ED,$78,$C7,$02,$3C,$1A,$BD,$FF
	.dc.b	$9D,$A3,$E8,$BF,$B5,$7C,$27,$FA,$0E,$31,$E4,$7F,$83,$23,$FD,$09
	.dc.b	$8F,$86,$FE,$72,$F9,$7F,$F0,$8F,$07,$A9,$B9,$7F,$85,$27,$4F,$EC
	.dc.b	$67,$4D,$35,$3C,$1F,$DB,$4F,$07,$ED,$36,$78,$C6,$27,$CD,$ED,$FF
	.dc.b	$B2,$BE,$95,$F5,$98,$AF,$F6,$04,$61,$FF,$72,$4F,$81,$A7,$82,$B8
	.dc.b	$68,$AE,$13,$FE,$E9,$AE,$D7,$D6,$18,$F9,$F3,$FF,$56,$70,$BE,$B3
	.dc.b	$FD,$AF,$A7,$F8,$DA,$31,$F0,$9E,$F6,$E1,$F0,$CF,$C6,$57,$07,$F2
	.dc.b	$4D,$73,$3C,$63,$13,$CA,$7F,$44,$DB,$FB,$9F,$53,$FE,$24,$6C,$FC
	.dc.b	$59,$D6,$7F,$5F,$D2,$FC,$E4,$FB,$CF,$A6,$E2,$BF,$06,$D9,$FE,$59
	.dc.b	$F5,$4F,$AC,$FF,$7A,$D6,$B0,$F9,$B8,$7C,$6C,$38,$3D,$23,$E8,$DF
	.dc.b	$1C,$FE,$66,$AB,$F1,$DF,$E9,$9A,$F9,$EF,$F4,$B1,$88,$FC,$4F,$9E
	.dc.b	$F9,$4F,$D4,$B6,$3F,$52,$D7,$F6,$2D,$7F,$30,$EF,$2F,$D0,$B7,$4F
	.dc.b	$7F,$D6,$7F,$AF,$B8,$3D,$B9,$AB,$58,$78,$B1,$CC,$63,$56,$1C,$6B
	.dc.b	$E7,$1F,$8A,$9E,$3A,$3E,$B3,$FA,$1E,$5F,$96,$E4,$70,$BE,$A3,$F9
	.dc.b	$46,$FF,$90,$BD,$E5,$F6,$6F,$75,$AD,$FF,$8E,$56,$B1,$FC,$A3,$6F
	.dc.b	$3E,$27,$F3,$39,$CB,$63,$31,$F9,$7C,$FC,$D3,$99,$FE,$49,$5C,$DB
	.dc.b	$19,$7D,$1F,$59,$E3,$7F,$6B,$EA,$D7,$38,$7B,$FA,$13,$9C,$D7,$C7
	.dc.b	$4F,$DB,$4E,$63,$15,$C2,$FF,$23,$7C,$1F,$D7,$61,$D8,$FE,$8D,$19
	.dc.b	$6D,$79,$E3,$13,$FE,$8C,$6C,$FF,$AC,$1A,$FF,$83,$5A,$FD,$5F,$3B
	.dc.b	$FA,$E7,$5F,$F3,$0E,$D7,$D6,$7A,$E7,$38,$7F,$85,$74,$FE,$69,$39
	.dc.b	$8F,$F4,$DE,$1D,$6B,$FA,$E3,$63,$2F,$76,$E2,$9D,$CE,$4E,$71,$99
	.dc.b	$DC,$F4,$B7,$9E,$F9,$EC,$36,$BF,$C0,$6D,$C3,$FB,$93,$5F,$E3,$DF
	.dc.b	$59,$CA,$7D,$3C,$48,$C4,$FF,$AD,$A3,$15,$EC,$6D,$7F,$84,$3E,$B3
	.dc.b	$FC,$F8,$DF,$5B,$CD,$18,$BF,$F0,$13,$5F,$EF,$ED,$7E,$63,$56,$48
	.dc.b	$C3,$67,$A9,$2B,$C0,$4A,$F9,$3F,$59,$F3,$35,$34,$93,$FE,$28,$F9
	.dc.b	$EF,$EC,$CD,$F1,$3E,$74,$62,$FF,$C3,$4D,$7F,$C7,$5F,$56,$FA,$CF
	.dc.b	$BF,$E9,$23,$18,$7E,$12,$31,$3A,$9C,$D7,$FB,$8A,$7F,$C1,$9F,$0F
	.dc.b	$CB,$CA,$34,$98,$73,$8C,$B6,$33,$5C,$E7,$2F,$A3,$C0,$F5,$92,$BE
	.dc.b	$03,$F9,$0F,$AC,$7D,$2D,$24,$FE,$62,$32,$FE,$7B,$E2,$3F,$C2,$DC
	.dc.b	$9F,$9E,$6C,$67,$AF,$FC,$60,$D9,$D4,$63,$1F,$F8,$03,$5C,$1B,$19
	.dc.b	$7F,$D0,$42,$34,$E1,$7C,$CB,$CF,$F1,$37,$0F,$33,$6B,$BB,$5E,$4D
	.dc.b	$7F,$D2,$15,$D2,$B9,$92,$BF,$7A,$3E,$B3,$FD,$6A,$E1,$B5,$E8,$3F
	.dc.b	$CD,$0E,$1F,$F1,$8B,$5F,$6F,$D5,$3E,$2B,$F8,$0B,$6B,$F8,$3B,$67
	.dc.b	$F8,$6B,$5E,$CD,$79,$CF,$22,$4F,$C2,$FA,$CF,$8A,$47,$04,$9F,$8B
	.dc.b	$19,$6C,$7C,$AF,$8E,$F9,$A4,$70,$48,$F9,$B1,$96,$C7,$F2,$4F,$35
	.dc.b	$F5,$8F,$2B,$F9,$2C,$61,$B1,$99,$E0,$DA,$F8,$3E,$EF,$C6,$24,$FC
	.dc.b	$66,$C6,$A3,$04,$7F,$03,$CB,$F8,$62,$47,$E1,$9B,$1A,$8C,$11,$FC
	.dc.b	$4C,$FE,$38,$91,$F8,$E6,$C6,$A3,$04,$7F,$09,$87,$C2,$F8,$AF,$79
	.dc.b	$27,$DE,$FE,$0B,$D2,$91,$F8,$26,$D7,$E0,$9A,$FE,$09,$AF,$E6,$B8
	.dc.b	$BF,$8B,$37,$B7,$F4,$A4,$AF,$D2,$B6,$3F,$4A,$D7,$F4,$BE,$2F,$EB
	.dc.b	$09,$3F,$AC,$6C,$7E,$B1,$AF,$08,$FE,$C7,$CA,$FA,$CF,$CB,$4F,$FA
	.dc.b	$75,$B1,$9E,$2F,$EC,$ED,$AF,$EC,$FF,$58,$F0,$3F,$B2,$B6,$33,$18
	.dc.b	$23,$82,$61,$FD,$D2,$78,$3F,$CB,$FE,$7F,$F8,$DB,$7B,$3F,$9C,$C6
	.dc.b	$23,$FD,$EC,$D7,$FD,$AB,$DB,$FD,$DB,$C5,$FF,$25,$61,$FD,$F1,$B3
	.dc.b	$FE,$F2,$9F,$E8,$7F,$1B,$F0,$3C,$13,$49,$AF,$F8,$03,$93,$FB,$F5
	.dc.b	$FE,$9F,$CA,$C1,$32,$FF,$04,$CF,$F0,$99,$1F,$E6,$9F,$73,$91,$27
	.dc.b	$6D,$8F,$D9,$35,$FE,$55,$F5,$9E,$D7,$F9,$73,$7C,$13,$4D,$43,$D4
	.dc.b	$FF,$05,$3F,$ED,$1E,$5F,$F0,$CC,$62,$3C,$AD,$78,$46,$89,$87,$F8
	.dc.b	$92,$78,$2F,$AC,$7C,$07,$F8,$AA,$30,$D8,$CC,$7F,$34,$6D,$FF,$55
	.dc.b	$FA,$9F,$F1,$5E,$46,$92,$BF,$C6,$71,$96,$C6,$67,$94,$7F,$8F,$38
	.dc.b	$9F,$2D,$FE,$83,$9F,$EA,$B1,$8F,$9A,$FF,$1A,$4E,$67,$FC,$8C,$D7
	.dc.b	$EB,$FD,$5F,$F8,$B1,$FE,$8C,$D7,$FA,$FB,$5D,$FA,$5F,$E4,$48,$C4
	.dc.b	$1A,$48,$F8,$9F,$59,$31,$F8,$8D,$AF,$F2,$AD,$23,$A1,$27,$FC,$AB
	.dc.b	$19,$6C,$66,$78,$B5,$FE,$EB,$C5,$F1,$1F,$E7,$78,$7F,$9E,$1F,$F5
	.dc.b	$E3,$63,$55,$FE,$62,$E6,$FD,$9D,$7F,$88,$7A,$FF,$CC,$31,$88,$FF
	.dc.b	$00,$B5,$FC,$FB,$53,$49,$1F,$D3,$3E,$B3,$FD,$5B,$3E,$66,$BF,$A8
	.dc.b	$B4,$62,$3F,$DA,$4D,$7A,$ED,$E4,$26,$1E,$97,$CF,$F5,$92,$3D,$8D
	.dc.b	$9F,$F2,$03,$5F,$F1,$F3,$5F,$F4,$57,$D5,$BE,$B3,$FC,$F8,$DA,$FC
	.dc.b	$83,$67,$FC,$19,$C2,$8C,$47,$E4,$9A,$F4,$AE,$64,$AF,$CD,$7D,$67
	.dc.b	$F9,$A8,$9E,$BF,$FA,$2D,$B1,$A8,$C1,$1F,$F2,$D7,$D6,$3E,$67,$FC
	.dc.b	$BE,$D8,$CC,$60,$FF,$4B,$7D,$67,$F9,$E0,$95,$82,$4F,$EB,$63,$2D
	.dc.b	$8F,$F5,$1E,$3F,$C2,$8F,$F5,$49,$FC,$44,$BD,$58,$70,$7F,$8C,$7C
	.dc.b	$27,$D6,$7F,$91,$A3,$FD,$6B,$18,$F2,$7F,$D6,$6D,$8C,$CE,$DE,$5F
	.dc.b	$04,$F2,$36,$D8,$E6,$D7,$71,$FC,$1F,$EB,$9F,$37,$FE,$5A,$86,$9F
	.dc.b	$EC,$99,$CD,$7F,$3A,$9C,$BF,$EA,$08,$CB,$E1,$9F,$F6,$87,$D6,$BE
	.dc.b	$33,$A4,$73,$C2,$D7,$16,$F2,$B0,$E7,$8F,$68,$63,$CE,$FC,$58,$7F
	.dc.b	$B7,$7E,$AD,$F5,$7F,$D7,$5B,$3F,$C8,$DA,$FE,$51,$AF,$F2,$4F,$C6
	.dc.b	$DE,$1F,$89,$B1,$D8,$FF,$5E,$B8,$7D,$C7,$FB,$02,$7A,$E5,$E7,$9D
	.dc.b	$36,$BA,$1E,$E9,$DE,$1F,$D9,$3E,$CF,$FD,$0B,$F4,$1F,$11,$B8,$F8
	.dc.b	$5A,$F8,$34,$7D,$6F,$28,$4E,$B2,$FA,$AD,$C3,$B7,$23,$1F,$9F,$87
	.dc.b	$1C,$1F,$4A,$F5,$FE,$AC,$61,$B1,$98,$F8,$87,$0F,$6F,$33,$23,$33
	.dc.b	$B7,$E3,$9B,$38,$FD,$5F,$F6,$87,$FD,$24,$7F,$81,$9C,$38,$6B,$FE
	.dc.b	$97,$6B,$F2,$7C,$B7,$D6,$7C,$96,$C7,$C9,$6B,$E2,$9E,$BB,$7E,$6B
	.dc.b	$C6,$7C,$E7,$F8,$74,$9B,$FF,$86,$9B,$7F,$E8,$56,$BF,$E6,$67,$FC
	.dc.b	$5D,$F1,$EF,$80,$FC,$DB,$67,$F1,$5C,$68,$C4,$7E,$2B,$0F,$0E,$1F
	.dc.b	$43,$27,$AB,$FC,$29,$B5,$F1,$83,$87,$E3,$46,$C8,$EE,$34,$FC,$98
	.dc.b	$E5,$C9,$F8,$8F,$A9,$FC,$82,$69,$E3,$8F,$DA,$F9,$CF,$A3,$FF,$25
	.dc.b	$36,$FF,$C2,$0D,$7F,$A0,$B5,$FF,$08,$B5,$FF,$07,$C7,$22,$3F,$82
	.dc.b	$B7,$E3,$8F,$9D,$FD,$23,$67,$F4,$8D,$7F,$49,$7E,$1E,$1F,$E2,$B0
	.dc.b	$FA,$B7,$C6,$7F,$34,$6C,$7C,$B6,$BF,$92,$6B,$FE,$E5,$6B,$FE,$FE
	.dc.b	$6B,$FE,$5C,$6B,$DE,$3F,$C6,$FB,$3E,$2F,$F8,$FF,$2F,$F2,$C6,$C6
	.dc.b	$6B,$F5,$77,$E2,$C3,$F9,$6D,$F1,$DF,$FC,$83,$87,$D5,$BF,$D3,$CE
	.dc.b	$5B,$19,$7F,$DF,$91,$96,$C6,$5F,$F3,$98,$8D,$38,$6B,$FE,$78,$E3
	.dc.b	$7E,$2E,$F7,$7F,$9B,$B6,$7F,$3A,$D7,$F9,$2D,$78,$71,$FA,$7F,$58
	.dc.b	$F8,$8F,$EC,$AD,$AF,$CE,$35,$FE,$06,$D7,$CE,$1C,$3E,$A6,$BF,$E8
	.dc.b	$E6,$BB,$8E,$FF,$87,$7F,$1C,$93,$4C,$46,$7E,$22,$57,$F1,$EA,$47
	.dc.b	$F5,$8F,$6B,$C5,$F1,$DF,$D6,$3D,$6F,$E1,$B7,$FE,$6E,$6B,$F2,$5A
	.dc.b	$B6,$FF,$15,$77,$FF,$AC,$FE,$CD,$F2,$FF,$D8,$9B,$5F,$C2,$1A,$FF
	.dc.b	$50,$C7,$F9,$BF,$D4,$7F,$36,$FA,$CF,$E5,$D3,$F1,$9A,$FD,$E6,$BF
	.dc.b	$EB,$26,$A6,$93,$C5,$FF,$0C,$BC,$38,$5F,$89,$7F,$71,$AF,$EC,$31
	.dc.b	$8F,$4B,$FB,$DB,$63,$FA,$2D,$7F,$41,$8F,$EA,$7F,$8A,$7F,$20,$6F
	.dc.b	$8D,$FD,$7E,$31,$D9,$E7,$F8,$AC,$4C,$C3,$FB,$15,$69,$F5,$4F,$F6
	.dc.b	$7F,$0D,$F1,$14,$93,$E2,$F5,$AB,$FD,$E3,$C5,$FE,$35,$38,$CF,$F8
	.dc.b	$3E,$8B,$F9,$9F,$5F,$30,$F5,$BF,$D9,$3E,$C1,$F1,$B9,$9F,$EB,$B6
	.dc.b	$AC,$61,$FC,$4A,$F7,$5F,$DE,$FF,$25,$7D,$D7,$F7,$F6,$C6,$63,$8F
	.dc.b	$0F,$F8,$17,$91,$F3,$9F,$E0,$56,$D7,$F8,$15,$AF,$8B,$E0,$B5,$32
	.dc.b	$34,$FE,$1E,$3F,$C1,$DF,$4F,$C3,$3F,$F0,$3B,$70,$FF,$0A,$36,$FF
	.dc.b	$11,$D1,$C3,$09,$1F,$E1,$59,$D3,$F8,$B9,$FF,$04,$FC,$7B,$E2,$3F
	.dc.b	$CC,$6D,$9F,$EB,$4D,$7F,$C0,$2D,$7F,$7E,$D7,$F8,$BB,$51,$FE,$F0
	.dc.b	$8E,$A7,$FA,$DF,$D5,$FE,$F7,$1F,$E3,$9E,$CF,$F0,$EF,$AB,$FB,$99
	.dc.b	$CC,$FF,$85,$2F,$C5,$A7,$F8,$A3,$F2,$3F,$F3,$9C,$FF,$8C,$5A,$FE
	.dc.b	$25,$B1,$FD,$3A,$7F,$CA,$FA,$FF,$8E,$E7,$2D,$F8,$2F,$F1,$E4,$62
	.dc.b	$FD,$35,$C1,$FF,$1C,$F0,$7F,$85,$2F,$F0,$13,$F7,$CF,$97,$ED,$5A
	.dc.b	$FC,$04,$57,$D5,$1F,$43,$D5,$F9,$E4,$8F,$9E,$D7,$E7,$B5,$F9,$EF
	.dc.b	$F8,$37,$C5,$7E,$2B,$FE,$4D,$6F,$C0,$7F,$9D,$A3,$FC,$91,$87,$F8
	.dc.b	$47,$5F,$91,$E0,$7B,$89,$8F,$F9,$1A,$5F,$21,$88,$F9,$53,$6B,$E5
	.dc.b	$4F,$F9,$87,$DC,$FF,$13,$57,$F9,$85,$AF,$EE,$5B,$3F,$DF,$BE,$CF
	.dc.b	$FC,$01,$B7,$F9,$AE,$72,$DF,$77,$A1,$FE,$5C,$9D,$DF,$F3,$3F,$B0
	.dc.b	$74,$7C,$16,$C7,$F9,$53,$74,$62,$3F,$DB,$AD,$7F,$E0,$0F,$D9,$3E
	.dc.b	$A3,$FA,$33,$63,$F2,$4D,$7F,$CF,$53,$CF,$4F,$F4,$F7,$E9,$B7,$4F
	.dc.b	$EE,$0D,$AF,$E9,$0D,$EA,$61,$CE,$F8,$E7,$FE,$AA,$AE,$19,$9E,$15
	.dc.b	$BC,$7F,$BE,$65,$FA,$2C,$BF,$C0,$77,$EE,$EA,$F8,$44,$8F,$F3,$AE
	.dc.b	$7E,$0F,$43,$DA,$48,$FF,$2E,$DF,$EB,$34,$FF,$5C,$7D,$97,$F8,$CA
	.dc.b	$7A,$B5,$FF,$06,$35,$FE,$C8,$DF,$25,$FA,$FF,$F8,$EA,$BF,$D3,$6D
	.dc.b	$D3,$FD,$50,$DB,$E3,$1F,$EA,$98,$C5,$72,$8C,$1F,$EC,$EB,$CC,$C8
	.dc.b	$CD,$7F,$9D,$A7,$A6,$8F,$4F,$FC,$39,$3F,$19,$AF,$F8,$95,$AF,$F7
	.dc.b	$D6,$E1,$EA,$FB,$BF,$21,$27,$FD,$07,$3F,$E2,$7C,$BF,$CA,$FE,$17
	.dc.b	$F8,$85,$FF,$70,$7D,$5F,$F9,$77,$E3,$DD,$1F,$E3,$38,$C3,$FE,$3A
	.dc.b	$AE,$1A,$DF,$67,$FD,$D7,$5A,$CB,$FC,$20,$D5,$5C,$1C,$FC,$27,$F9
	.dc.b	$13,$1E,$F1,$83,$F9,$7B,$8D,$97,$CE,$F9,$57,$4F,$A0,$E5,$9F,$F9
	.dc.b	$27,$EB,$1F,$67,$8A,$FF,$11,$46,$1E,$B3,$EB,$AD,$47,$33,$27,$0A
	.dc.b	$E5,$8A,$F9,$4B,$2F,$4B,$C4,$C6,$1D,$5E,$75,$C6,$F9,$DF,$2B,$E3
	.dc.b	$8B,$1F,$1B,$70,$E3,$5B,$C7,$DD,$86,$AF,$DD,$F0,$4F,$B3,$FF,$14
	.dc.b	$65,$F2,$7E,$C3,$FC,$B6,$DC,$B7,$3A,$C7,$FA,$ED,$F1,$F6,$BD,$F8
	.dc.b	$70,$9F,$F3,$A3,$6F,$FD,$1D,$7C,$5B,$5C,$18,$7C,$AB,$E8,$FC,$C3
	.dc.b	$F0,$7E,$FF,$FA,$4F,$1F,$82,$DC,$7F,$1C,$D9,$7D,$93,$EC,$FC,$11
	.dc.b	$CB,$0E,$9B,$FF,$9A,$1B,$7C,$6F,$86,$5F,$32,$3F,$D4,$57,$FE,$7E
	.dc.b	$BE,$0F,$FA,$00,$E8,$57,$2E,$FD,$87,$C6,$FF,$B9,$67,$FC,$6A,$D8
	.dc.b	$CC,$F1,$6B,$C6,$3C,$F1,$AD,$3F,$43,$8B,$EC,$DF,$24,$E8,$DE,$13
	.dc.b	$1F,$C6,$9F,$E1,$4C,$BC,$2E,$30,$E5,$7C,$63,$67,$EC,$B2,$F0,$6F
	.dc.b	$F8,$0F,$B3,$FE,$95,$97,$E5,$B8,$9F,$63,$C7,$6F,$DA,$F0,$AF,$79
	.dc.b	$78,$B9,$D8,$70,$C7,$F0,$15,$AC,$3F,$25,$8F,$3C,$BC,$1E,$45,$CB
	.dc.b	$FD,$2D,$8F,$17,$DB,$9F,$79,$D6,$7D,$B0,$E1,$39,$CF,$D1,$1C,$EF
	.dc.b	$D3,$E5,$70,$E9,$FF,$37,$57,$E4,$69,$39,$F7,$B8,$46,$27,$F5,$F1
	.dc.b	$89,$D4,$71,$6D,$FC,$CA,$E0,$FC,$3B,$EE,$E6,$77,$AF,$F0,$6A,$DB
	.dc.b	$FC,$76,$B6,$FF,$1F,$EB,$FF,$5A,$E1,$FC,$21,$B8,$77,$F4,$7E,$E7
	.dc.b	$7F,$F3,$8F,$85,$F4,$C6,$2F,$F9,$C4,$61,$FF,$2C,$CE,$DF,$85,$3B
	.dc.b	$AF,$E6,$F5,$B5,$5B,$7F,$1F,$E1,$FF,$66,$F8,$1E,$0D,$C3,$A9,$FE
	.dc.b	$F9,$F3,$FF,$D4,$AE,$29,$3B,$7A,$C6,$FE,$07,$FB,$B7,$17,$F8,$E1
	.dc.b	$B7,$CA,$BF,$DE,$73,$9A,$FD,$B5,$FF,$78,$E7,$F1,$F9,$1F,$EA,$69
	.dc.b	$F0,$53,$EC,$9F,$37,$AB,$95,$E5,$C3,$93,$C2,$B6,$E7,$0E,$98,$FF
	.dc.b	$AE,$DB,$AF,$F8,$1B,$0F,$95,$97,$7C,$BA,$DF,$B8,$FF,$79,$63,$D1
	.dc.b	$F5,$B5,$FF,$11,$B5,$FF,$47,$CF,$3C,$7F,$C4,$7E,$3F,$EC,$31,$FD
	.dc.b	$E7,$9A,$8C,$4F,$0F,$A2,$FE,$94,$FF,$A4,$E3,$38,$30,$56,$A3,$E6
	.dc.b	$B5,$FF,$1A,$B5,$E9,$E3,$FD,$6F,$C4,$1F,$8B,$7C,$06,$E3,$DF,$E4
	.dc.b	$F1,$32,$3E,$9C,$EA,$FE,$26,$7E,$9B,$D4,$63,$34,$7E,$56,$75,$7F
	.dc.b	$73,$2F,$A2,$FF,$AD,$23,$37,$F3,$F3,$F6,$EF,$F8,$CB,$FB,$33,$EC
	.dc.b	$8C,$1F,$E4,$5E,$1F,$F2,$0F,$8F,$F8,$A3,$F4,$9F,$1F,$FA,$2E,$FC
	.dc.b	$67,$F8,$39,$B8,$7D,$DA,$D4,$E4,$FF,$5D,$77,$FE,$9B,$DE,$FE,$07
	.dc.b	$C7,$FC,$CE,$9F,$31,$FE,$3F,$BD,$93,$0F,$F6,$06,$1E,$29,$D4,$62
	.dc.b	$7F,$31,$F6,$4E,$BF,$D9,$B6,$3F,$60,$3E,$C9,$F2,$9A,$AF,$E0,$13
	.dc.b	$98,$FF,$33,$34,$FE,$C4,$5B,$FF,$70,$9F,$EE,$27,$0E,$1A,$E6,$BB
	.dc.b	$C7,$26,$C7,$23,$FA,$E3,$87,$AE,$EF,$82,$7E,$1D,$FC,$39,$B7,$FC
	.dc.b	$2F,$5F,$93,$F6,$59,$A9,$3F,$BB,$AD,$E1,$F5,$75,$5E,$A3,$FD,$79
	.dc.b	$9F,$F1,$AF,$2B,$CB,$C1,$F5,$6D,$F2,$19,$A9,$97,$FB,$15,$FA,$91
	.dc.b	$8C,$3F,$37,$F3,$7F,$15,$A7,$F2,$DF,$45,$F6,$5F,$E6,$CC,$3F,$92
	.dc.b	$75,$AF,$86,$7F,$DD,$EF,$7A,$7F,$A6,$5B,$7F,$E1,$56,$E3,$F2,$DA
	.dc.b	$FD,$66,$A3,$FC,$2F,$8F,$85,$E0,$D7,$83,$5E,$1C,$2F,$B9,$FC,$8E
	.dc.b	$3F,$E0,$B8,$C7,$B5,$F9,$5E,$FF,$E8,$75,$AC,$B8,$B7,$0F,$E8,$2D
	.dc.b	$E4,$FD,$0D,$70,$BF,$BD,$E4,$CC,$3E,$F7,$13,$EC,$F1,$97,$F2,$BC
	.dc.b	$FF,$0D,$78,$D3,$EC,$55,$C3,$D3,$D2,$EC,$FD,$29,$F4,$F7,$FC,$74
	.dc.b	$FE,$AA,$31,$CF,$23,$FA,$C6,$6F,$B4,$E5,$1F,$D6,$F8,$BE,$DD,$6E
	.dc.b	$BF,$9F,$46,$27,$8D,$70,$9C,$E1,$E0,$7D,$5F,$85,$BE,$43,$FB,$5B
	.dc.b	$6F,$E0,$F8,$F8,$99,$1F,$D5,$E7,$55,$F9,$DA,$CB,$F1,$70,$7D,$37
	.dc.b	$F7,$66,$DF,$CC,$F3,$71,$32,$3F,$B1,$4E,$AB,$F5,$15,$97,$F9,$BF
	.dc.b	$D3,$BE,$8F,$F9,$A3,$6F,$FB,$F7,$13,$ED,$5F,$67,$AB,$FF,$07,$4E
	.dc.b	$63,$FD,$E8,$DA,$FF,$38,$37,$0F,$E0,$CD,$7F,$88,$B5,$FF,$0D,$DF
	.dc.b	$69,$E2,$D8,$E2,$D7,$F8,$CF,$94,$FC,$29,$F2,$7F,$DE,$1B,$7F,$B4
	.dc.b	$CD,$F6,$B9,$AF,$EF,$BE,$23,$EF,$7F,$C2,$ED,$BF,$F4,$46,$3F,$CB
	.dc.b	$A7,$27,$F9,$C7,$2F,$F0,$CF,$9B,$FE,$2F,$7F,$B9,$3F,$D7,$7E,$47
	.dc.b	$D1,$B6,$EF,$FA,$ED,$7F,$80,$BE,$27,$84,$1E,$5D,$2E,$EC,$C6,$1F
	.dc.b	$05,$7F,$03,$6B,$FC,$0E,$78,$67,$FE,$9F,$8F,$F4,$DC,$E9,$B1,$99
	.dc.b	$DC,$7F,$11,$9D,$46,$23,$E6,$FC,$AB,$EE,$7F,$C4,$6D,$BF,$E7,$1F
	.dc.b	$6F,$F3,$1F,$F1,$AC,$66,$BF,$B5,$4E,$5F,$E8,$DE,$32,$31,$1F,$D4
	.dc.b	$A3,$1C,$5E,$C6,$DF,$FA,$E5,$B3,$CA,$B8,$BF,$C8,$E3,$2F,$38,$CA
	.dc.b	$8F,$F5,$87,$E1,$5F,$E0,$2A,$3F,$DA,$3E,$E1,$F2,$F9,$8C,$3F,$DA
	.dc.b	$A3,$33,$FD,$A9,$B7,$FD,$86,$11,$88,$FE,$D4,$D7,$FB,$73,$5F,$F6
	.dc.b	$13,$63,$8C,$FF,$86,$7F,$21,$FE,$19,$C3,$FC,$D5,$B3,$FE,$4A,$FA
	.dc.b	$CF,$F6,$5B,$FE,$5F,$8C,$E1,$FE,$43,$9C,$BF,$E6,$E6,$BF,$A8,$6B
	.dc.b	$CF,$E2,$7F,$CE,$F1,$88,$E5,$18,$25,$7F,$9D,$BF,$06,$FE,$2F,$87
	.dc.b	$F4,$96,$C7,$BB,$EE,$7F,$D1,$71,$8B,$F7,$C7,$CB,$24,$FC,$B6,$C7
	.dc.b	$CB,$CD,$18,$8F,$96,$FF,$8F,$78,$9F,$2D,$FE,$98,$6D,$FF,$3B,$CF
	.dc.b	$FB,$8B,$FD,$82,$33,$7F,$CF,$27,$FB,$E3,$FD,$B5,$AF,$EA,$DA,$A1
	.dc.b	$18,$6C,$7F,$7B,$8C,$11,$FF,$29,$7C,$33,$F1,$9F,$F5,$6B,$6F,$FA
	.dc.b	$87,$DD,$7F,$9E,$BC,$F4,$E6,$FF,$A8,$CF,$FB,$D1,$FF,$01,$35,$FE
	.dc.b	$00,$D7,$CC,$D5,$EB,$3F,$20,$FE,$2C,$DC,$3F,$B1,$AF,$A2,$CC,$C9
	.dc.b	$CB,$C2,$31,$5F,$DD,$F8,$7F,$C6,$35,$FE,$AF,$6C,$7F,$9D,$FE,$EB
	.dc.b	$E7,$C7,$F8,$34,$C9,$7D,$BF,$F6,$7B,$8A,$FF,$52,$CF,$07,$FC,$2F
	.dc.b	$19,$AF,$F4,$F7,$DA,$FF,$8E,$A7,$FD,$86,$D8,$FF,$43,$7D,$D7,$FA
	.dc.b	$CE,$95,$FE,$B0,$AD,$46,$2B,$9F,$BB,$FE,$BE,$9C,$D7,$FA,$DA,$F9
	.dc.b	$3C,$67,$FD,$87,$1F,$EB,$BA,$FF,$58,$37,$C2,$34,$C7,$DC,$7F,$AD
	.dc.b	$2F,$FD,$B2,$DA,$FF,$4E,$7D,$CF,$67,$FD,$75,$19,$CB,$CD,$82,$31
	.dc.b	$7F,$EC,$A6,$D7,$FB,$06,$F9,$56,$FE,$1B,$FA,$A4,$FF,$91,$F8,$BF
	.dc.b	$A2,$34,$AE,$1F,$C8,$9F,$E7,$B1,$E5,$FF,$6C,$F6,$FF,$A6,$36,$34
	.dc.b	$D7,$FD,$04,$D7,$FD,$1A,$BE,$11,$A7,$26,$CF,$26,$BC,$9A,$FB,$FE
	.dc.b	$A3,$FC,$0F,$AD,$8F,$97,$18,$7F,$DC,$5F,$70,$F9,$5C,$6B,$F2,$6A
	.dc.b	$EE,$69,$AA,$E1,$7F,$E6,$DC,$BE,$8F,$E3,$2F,$B2,$FE,$D9,$8A,$31
	.dc.b	$5F,$E0,$66,$BF,$EA,$38,$D4,$F0,$C7,$FA,$6C,$FE,$0B,$C5,$F1,$9E
	.dc.b	$6B,$E3,$87,$E0,$2B,$75,$FE,$E1,$6C,$66,$B8,$5F,$C1,$EC,$FF,$68
	.dc.b	$B7,$83,$9E,$DF,$E8,$06,$E2,$CD,$F2,$1F,$EA,$48,$F8,$15,$BB,$E4
	.dc.b	$48,$C9,$C6,$78,$3F,$E7,$9F,$9A,$7C,$2F,$8D,$B1,$C3,$24,$62,$3F
	.dc.b	$07,$87,$FA,$6E,$1E,$3F,$FA,$F2,$78,$46,$F5,$FF,$4B,$36,$BF,$D3
	.dc.b	$6D,$9E,$B5,$C5,$EA,$7B,$FF,$02,$7A,$5F,$69,$B1,$FE,$3E,$6B,$F6
	.dc.b	$5A,$FF,$B6,$BC,$CC,$E3,$FE,$A7,$6E,$3F,$EA,$C6,$CF,$7B,$FA,$9F
	.dc.b	$95,$3A,$BF,$C6,$9A,$23,$15,$FA,$F8,$C4,$EA,$73,$3F,$EE,$D9,$CB
	.dc.b	$B8,$CB,$FE,$EE,$6B,$FE,$BE,$6A,$8F,$F1,$17,$9B,$F8,$7C,$3F,$C4
	.dc.b	$33,$F5,$5A,$FF,$AC,$EB,$C3,$97,$F1,$2F,$17,$FA,$D4,$FF,$99,$1A
	.dc.b	$FE,$7D,$AF,$F9,$D1,$AF,$FB,$5F,$89,$18,$8F,$C9,$36,$33,$E3,$7E
	.dc.b	$BA,$35,$97,$A7,$D1,$FC,$F3,$FE,$53,$BF,$C6,$35,$FB,$F7,$C9,$FC
	.dc.b	$83,$FE,$A8,$7E,$73,$5F,$98,$D7,$FD,$D8,$7C,$77,$0F,$FB,$A1,$AE
	.dc.b	$7D,$EF,$B7,$18,$9F,$D6,$47,$DE,$7F,$25,$8F,$FB,$B5,$BB,$FF,$BC
	.dc.b	$9B,$7C,$CF,$D6,$DB,$E9,$FF,$D0,$51,$FE,$92,$6B,$FC,$35,$AF,$FA
	.dc.b	$89,$AF,$CF,$6B,$FB,$F6,$C7,$5F,$53,$F8,$DC,$6A,$9F,$8F,$3E,$73
	.dc.b	$F8,$63,$70,$FD,$D3,$5F,$9F,$5E,$4C,$BF,$CA,$DF,$65,$F9,$EF,$50
	.dc.b	$D2,$57,$E7,$E3,$2D,$8C,$BF,$9F,$8C,$BF,$A1,$F6,$3F,$D6,$15,$FD
	.dc.b	$E7,$2F,$F5,$46,$BF,$C8,$67,$2A,$BF,$41,$7C,$AF,$FC,$01,$D2,$FB
	.dc.b	$57,$D1,$FF,$A9,$E3,$0D,$8C,$DF,$F6,$39,$CB,$63,$2F,$E2,$0F,$E3
	.dc.b	$AE,$1F,$C5,$C7,$07,$4D,$7F,$DB,$3D,$DF,$99,$BF,$EF,$ED,$C7,$F5
	.dc.b	$0D,$9E,$B8,$7F,$3F,$E5,$7C,$3D,$24,$FF,$36,$C5,$F9,$57,$9B,$2F
	.dc.b	$83,$F9,$F7,$F8,$0A,$4D,$25,$7F,$16,$FC,$9B,$C1,$C0,$FA,$67,$C8
	.dc.b	$7F,$4F,$6D,$7F,$3D,$AF,$32,$FB,$7F,$2B,$FE,$D3,$6B,$FE,$0D,$6B
	.dc.b	$F2,$5A,$FE,$C9,$AF,$DB,$6B,$D7,$4F,$F0,$DC,$6B,$07,$C7,$FF,$7D
	.dc.b	$AF,$EA,$0D,$7F,$A1,$C6,$BC,$CF,$F0,$44,$73,$3F,$AB,$7D,$C3,$EF
	.dc.b	$7F,$B6,$36,$33,$18,$EF,$CB,$23,$27,$28,$FF,$08,$FE,$2E,$F9,$BF
	.dc.b	$EE,$C3,$E4,$69,$A6,$57,$F6,$77,$4D,$8C,$BF,$D7,$0F,$F6,$BF,$B9
	.dc.b	$FE,$3C,$F3,$FF,$CF,$51,$FF,$00,$5A,$FF,$8E,$DA,$FE,$5D,$AF,$F2
	.dc.b	$06,$BF,$A9,$6B,$D8,$FF,$22,$7B,$CF,$B7,$7D,$97,$F8,$BB,$B3,$FB
	.dc.b	$97,$8D,$FE,$27,$F4,$3F,$C6,$97,$BA,$FF,$19,$FD,$B7,$FA,$E3,$1F
	.dc.b	$E3,$ED,$7F,$9C,$B7,$B3,$FC,$C3,$F8,$17,$F8,$63,$14,$E7,$C1,$FE
	.dc.b	$98,$DB,$FF,$0E,$36,$55,$FE,$20,$E6,$7D,$77,$F8,$AB,$0F,$E0,$D1
	.dc.b	$8F,$B9,$FF,$15,$49,$BC,$13,$2F,$F0,$D4,$E5,$B3,$97,$EA,$4B,$E3
	.dc.b	$7E,$A1,$23,$EA,$35,$FA,$8D,$7E,$A3,$FE,$06,$EA,$F3,$47,$F7,$96
	.dc.b	$E1,$F9,$EA,$FC,$C9,$DE,$7F,$11,$F1,$FF,$8E,$FB,$39,$18,$21,$FA
	.dc.b	$2E,$74,$6A,$31,$3F,$E4,$BF,$9C,$FF,$16,$60,$69,$F7,$BF,$18,$FF
	.dc.b	$1B,$D7,$F8,$F6,$33,$CB,$C4,$FF,$38,$CE,$A7,$E2,$FD,$5F,$E5,$C8
	.dc.b	$FE,$5D,$AF,$E5,$DA,$FE,$5D,$FF,$14,$FD,$2F,$F8,$BB,$03,$4F,$F1
	.dc.b	$8C,$B4,$57,$E4,$3E,$4E,$4F,$E4,$1F,$F4,$2F,$E3,$EF,$96,$CC,$FE
	.dc.b	$4A,$6F,$47,$FA,$CA,$B5,$3F,$E5,$BF,$AC,$F6,$4F,$F5,$36,$F3,$7E
	.dc.b	$31,$B3,$FE,$31,$5F,$80,$7F,$9D,$DC,$EA,$69,$23,$5F,$3D,$FE,$48
	.dc.b	$83,$49,$5F,$E4,$98,$FE,$0E,$FF,$1E,$6E,$1E,$5F,$07,$9C,$FA,$5F
	.dc.b	$F5,$2F,$D6,$3E,$7A,$9E,$58,$93,$FE,$14,$9D,$36,$72,$F7,$E5,$F3
	.dc.b	$11,$F2,$3E,$9E,$6F,$80,$48,$F9,$51,$FE,$B6,$F8,$43,$FE,$AE,$6C
	.dc.b	$7F,$B4,$FB,$BF,$D2,$F3,$93,$E2,$79,$DF,$2C,$93,$FE,$2A,$AE,$2F
	.dc.b	$C3,$8C,$B4,$D4,$E3,$3F,$B5,$37,$83,$FC,$B1,$7C,$A7,$EF,$74,$7D
	.dc.b	$32,$61,$F4,$5F,$AD,$F5,$DF,$74,$91,$F8,$78,$F7,$7D,$27,$F9,$3B
	.dc.b	$0F,$F2,$7E,$7F,$EB,$28,$C5,$78,$1A,$FF,$AB,$DB,$1B,$8F,$F0,$F4
	.dc.b	$F2,$C5,$5A,$6C,$E8,$E8,$53,$A1,$FE,$27,$F8,$E7,$D3,$FF,$24,$FB
	.dc.b	$17,$DF,$FF,$99,$27,$35,$FE,$8B,$9C,$BC,$6B,$AA,$9D,$E5,$FE,$E7
	.dc.b	$F9,$AE,$55,$FE,$83,$F8,$E4,$62,$FF,$DA,$91,$87,$A1,$83,$55,$CD
	.dc.b	$CB,$5F,$05,$75,$AD,$D7,$F8,$D6,$FC,$6F,$5A,$D9,$E0,$9D,$9C,$1A
	.dc.b	$76,$8E,$13,$CA,$75,$3E,$2F,$B2,$7D,$17,$99,$B1,$98,$C7,$D7,$E5
	.dc.b	$91,$93,$94,$61,$F2,$E4,$AE,$71,$86,$E1,$F3,$95,$7C,$B9,$FE,$BF
	.dc.b	$B3,$EB,$FE,$5C,$61,$BE,$57,$E2,$1F,$9F,$EE,$3E,$D7,$FC,$BB,$97
	.dc.b	$E0,$2F,$81,$F2,$0A,$E6,$F9,$47,$CC,$AE,$5F,$01,$E3,$9C,$CF,$F9
	.dc.b	$9D,$AF,$FA,$C1,$AF,$F9,$E5,$AF,$F7,$8F,$A5,$FB,$0E,$DA,$ED,$AE
	.dc.b	$F7,$FF,$3A,$39,$7D,$F1,$F8,$DC,$3F,$37,$87,$F9,$77,$E4,$9C,$DA
	.dc.b	$B9,$87,$BF,$16,$3C,$78,$FF,$D1,$D8,$F1,$C3,$79,$7E,$83,$B5,$DD
	.dc.b	$D3,$2F,$B7,$18,$AF,$CC,$AF,$C1,$75,$7C,$6E,$E3,$FE,$E9,$AE,$18
	.dc.b	$72,$9C,$AE,$4F,$E7,$58,$61,$FE,$79,$36,$B8,$E1,$D7,$1F,$0D,$E6
	.dc.b	$B8,$7E,$3E,$DE,$98,$E3,$B7,$FB,$09,$D3,$67,$4D,$26,$7F,$9A,$9D
	.dc.b	$B6,$B9,$B9,$AB,$8F,$F0,$8A,$FE,$37,$7C,$A3,$F8,$5F,$E3,$CF,$C7
	.dc.b	$FE,$0F,$8E,$EB,$7C,$AF,$C1,$AD,$61,$FC,$2A,$B5,$39,$BF,$A9,$3F
	.dc.b	$CC,$AB,$79,$FD,$73,$E6,$38,$8D,$4F,$09,$FF,$67,$70,$2B,$51,$86
	.dc.b	$BC,$A7,$83,$6B,$F9,$96,$1F,$E0,$1F,$9F,$FC,$37,$85,$FE,$0B,$6C
	.dc.b	$FF,$82,$FD,$FB,$7F,$E0,$B9,$B7,$FE,$09,$9F,$44,$EE,$BF,$C0,$91
	.dc.b	$88,$FF,$09,$36,$7F,$C2,$1E,$FF,$2F,$24,$F6,$D3,$D9,$E1,$66,$26
	.dc.b	$3F,$E2,$17,$5A,$FF,$46,$8C,$13,$2F,$F0,$94,$EA,$B8,$C3,$4F,$D2
	.dc.b	$61,$DA,$FF,$BF,$61,$CE,$BF,$DD,$93,$96,$C6,$5F,$F7,$F1,$FE,$2D
	.dc.b	$FA,$2F,$51,$EB,$E7,$B7,$FE,$37,$7E,$06,$7F,$E1,$F8,$FF,$14,$46
	.dc.b	$0C,$FD,$87,$C1,$F8,$E7,$DD,$FC,$58,$C3,$69,$8F,$AD,$FC,$1D,$70
	.dc.b	$7F,$C8,$D5,$C1,$A4,$9F,$F1,$B7,$E1,$9C,$DB,$1F,$E3,$B8,$F3,$91
	.dc.b	$FE,$37,$C0,$FC,$3B,$F9,$E4,$78,$B7,$FE,$1D,$39,$F1,$3F,$0B,$5E
	.dc.b	$3C,$3F,$DC,$2F,$06,$C6,$8D,$E7,$70,$E8,$7F,$A9,$BB,$BF,$CD,$6E
	.dc.b	$29,$87,$F8,$23,$D5,$FC,$47,$E3,$BF,$BE,$CF,$C7,$87,$0B,$FF,$78
	.dc.b	$B6,$33,$5C,$3B,$7F,$37,$39,$C7,$FD,$6B,$EC,$3E,$8B,$F3,$ED,$8F
	.dc.b	$17,$CD,$BE,$7D,$C5,$AB,$FD,$3C,$E6,$3F,$44,$DA,$FE,$8A,$DD,$3F
	.dc.b	$16,$1C,$3F,$8E,$6B,$F8,$06,$BB,$BF,$0C,$F4,$6C,$74,$DB,$F8,$B7
	.dc.b	$93,$FB,$8F,$00,$FE,$BF,$DC,$DF,$07,$DA,$9C,$E3,$F3,$EF,$8B,$63
	.dc.b	$38,$7E,$44,$FE,$85,$7C,$A7,$FC,$75,$87,$F0,$6F,$B3,$73,$DF,$85
	.dc.b	$C3,$6B,$B4,$E9,$B3,$A6,$9E,$3F,$81,$7D,$BF,$F1,$26,$C7,$D4,$D9
	.dc.b	$DB,$9A,$FA,$71,$89,$F8,$D3,$F5,$23,$1A,$7D,$2E,$0F,$B9,$39,$25
	.dc.b	$70,$24,$FC,$DF,$B6,$E3,$C3,$FA,$27,$F7,$D1,$F1,$78,$51,$AA,$FF
	.dc.b	$80,$2E,$AB,$85,$FF,$2C,$8C,$4E,$A3,$1D,$BF,$67,$E9,$64,$FE,$8F
	.dc.b	$C1,$FE,$F5,$E7,$63,$3F,$07,$2A,$F1,$9F,$F2,$0F,$37,$FC,$17,$8F
	.dc.b	$DB,$9F,$E7,$BF,$12,$FB,$7F,$F3,$C5,$FF,$26,$FB,$37,$E0,$9A,$9C
	.dc.b	$E1,$FD,$4A,$72,$FE,$D7,$D4,$7D,$C7,$F9,$F2,$3F,$C4,$71,$8F,$03
	.dc.b	$FA,$23,$63,$33,$BD,$BF,$47,$39,$6C,$65,$FE,$63,$19,$6C,$65,$FE
	.dc.b	$CB,$F2,$3F,$29,$FE,$0F,$D6,$FB,$7F,$E9,$D3,$9E,$74,$E7,$1F,$99
	.dc.b	$39,$8C,$5F,$E7,$F9,$13,$9C,$3F,$B1,$FD,$C3,$E4,$E9,$8F,$B1,$B9
	.dc.b	$7E,$AE,$74,$D9,$D3,$4F,$CF,$7D,$27,$E7,$4F,$CC,$FC,$25,$8F,$CF
	.dc.b	$3F,$99,$23,$FC,$05,$B9,$FE,$6F,$B9,$18,$8F,$E1,$91,$89,$FE,$7B
	.dc.b	$F3,$CF,$07,$FB,$EF,$E3,$5F,$E8,$C9,$FE,$05,$39,$7F,$82,$46,$6F
	.dc.b	$F8,$17,$BA,$F9,$E7,$D5,$7F,$80,$63,$0D,$8C,$DF,$F9,$9A,$72,$FF
	.dc.b	$4E,$FC,$9F,$FD,$6A,$FF,$11,$6D,$F6,$FC,$A1,$F2,$BF,$E1,$29,$CF
	.dc.b	$77,$E5,$FE,$B3,$FC,$15,$DA,$7F,$C3,$39,$3E,$3B,$FC,$DF,$9F,$EA
	.dc.b	$6B,$4F,$F4,$D9,$D3,$E9,$9D,$E7,$FC,$DA,$73,$3E,$99,$CC,$FF,$81
	.dc.b	$27,$2E,$E3,$2F,$F9,$E1,$AA,$36,$FF,$83,$3A,$DF,$4C,$F7,$BF,$C5
	.dc.b	$ED,$C3,$F9,$A7,$D4,$3E,$DF,$35,$FE,$34,$8C,$4F,$F7,$86,$E1,$FD
	.dc.b	$0D,$AF,$76,$BF,$E0,$CC,$11,$88,$FE,$DE,$D8,$CD,$FF,$8E,$E3,$18
	.dc.b	$7F,$8E,$AF,$FA,$43,$6F,$F5,$1F,$8C,$FF,$10,$7F,$92,$46,$76,$FF
	.dc.b	$19,$CE,$67,$FC,$5B,$DE,$8C,$47,$F9,$81,$AF,$FA,$45,$B1,$BF,$2B
	.dc.b	$FC,$A0,$DC,$3F,$D6,$6D,$AF,$F1,$ED,$72,$CD,$F7,$4F,$47,$FC,$BA
	.dc.b	$DC,$3F,$B1,$61,$FD,$3B,$0F,$E7,$93,$97,$F9,$F4,$65,$FB,$4D,$7E
	.dc.b	$63,$5F,$F0,$CB,$6F,$FD,$55,$E8,$FF,$99,$63,$04,$7F,$CD,$15,$FC
	.dc.b	$4B,$3F,$F1,$7B,$70,$FF,$41,$FD,$73,$E7,$BD,$3E,$27,$F9,$AE,$31
	.dc.b	$3F,$09,$B9,$FF,$97,$9A,$FD,$06,$BD,$2B,$E5,$C6,$23,$FD,$80,$DC
	.dc.b	$8E,$73,$7C,$47,$F4,$6A,$FF,$1B,$B6,$3F,$DB,$FF,$15,$FE,$8A,$8C
	.dc.b	$63,$FE,$B0,$AF,$F0,$C3,$5F,$D2,$35,$FC,$B3,$5F,$F2,$1B,$6B,$AB
	.dc.b	$FE,$58,$30,$66,$7F,$83,$BF,$9F,$6C,$FF,$B6,$FE,$C1,$F3,$7E,$5F
	.dc.b	$33,$FD,$29,$18,$BF,$C5,$B6,$BF,$D0,$ED,$7F,$32,$D7,$F5,$9F,$88
	.dc.b	$7F,$92,$5F,$F0,$23,$78,$3F,$DE,$DD,$2F,$C5,$5E,$27,$FA,$BA,$B5
	.dc.b	$8F,$2F,$7B,$07,$F1,$CA,$DC,$7F,$8D,$E7,$78,$FF,$99,$B1,$FD,$74
	.dc.b	$FE,$F5,$B1,$FE,$F9,$F9,$67,$C3,$FF,$AF,$27,$3C,$1F,$EA,$7F,$8B
	.dc.b	$C1,$FE,$B5,$AD,$E1,$FE,$45,$AE,$7C,$8F,$B3,$79,$BF,$EB,$B6,$D7
	.dc.b	$F8,$5F,$E5,$DE,$D6,$6A,$4F,$F8,$6E,$B7,$AF,$F6,$FB,$DC,$E6,$BF
	.dc.b	$D8,$D1,$89,$FF,$5B,$B5,$FE,$12,$D7,$FD,$CE,$D5,$87,$9F,$2F,$31
	.dc.b	$FE,$D1,$FB,$77,$D1,$E3,$1F,$24,$DB,$FF,$01,$56,$A7,$27,$FA,$56
	.dc.b	$3F,$D8,$BD,$1F,$E7,$C7,$FC,$FA,$D7,$25,$3F,$4E,$6F,$ED,$9F,$35
	.dc.b	$FE,$8D,$8C,$35,$FF,$43,$C6,$BD,$2E,$8D,$8E,$5C,$DF,$ED,$F6,$F0
	.dc.b	$EA,$31,$1C,$9A,$FF,$B3,$35,$FB,$5F,$55,$FE,$DE,$AF,$F7,$3F,$0A
	.dc.b	$31,$A6,$E3,$17,$FE,$53,$FB,$6F,$F2,$83,$F6,$EB,$6F,$D0,$9D,$BF
	.dc.b	$E2,$B6,$DF,$FB,$D2,$B8,$CF,$E0,$7E,$A7,$E4,$62,$E2,$46,$1B,$3F
	.dc.b	$CF,$9A,$FF,$BE,$B0,$7D,$37,$8A,$FF,$2C,$D8,$CC,$FE,$56,$F1,$32
	.dc.b	$3F,$28,$F1,$BE,$B8,$FF,$B8,$36,$BA,$7F,$69,$6D,$FF,$B5,$9B,$93
	.dc.b	$89,$F2,$DF,$13,$A3,$FD,$95,$A5,$BE,$31,$8C,$3F,$CF,$D7,$C1,$FF
	.dc.b	$5A,$76,$BE,$7F,$93,$63,$31,$F4,$21,$F1,$FF,$EA,$49,$C9,$0F,$F4
	.dc.b	$B4,$71,$7F,$C7,$71,$B7,$FC,$C1,$E0,$3E,$67,$E9,$B6,$3F,$91,$C6
	.dc.b	$0F,$CF,$63,$FE,$DB,$7D,$DF,$17,$B9,$FF,$52,$37,$81,$C1,$F9,$1C
	.dc.b	$30,$FF,$84,$FE,$DB,$FC,$97,$7F,$AB,$C5,$18,$8F,$F6,$F3,$5F,$F1
	.dc.b	$53,$5E,$5E,$D7,$92,$31,$09,$E7,$39,$3D,$38,$FF,$92,$67,$FD,$C4
	.dc.b	$D7,$D3,$EB,$7D,$9B,$F1,$65,$EA,$9F,$EC,$BE,$2F,$E6,$3C,$9F,$E7
	.dc.b	$71,$FD,$C6,$30,$47,$04,$3F,$BA,$3F,$2B,$27,$87,$F8,$76,$CF,$FB
	.dc.b	$D5,$AF,$FA,$56,$BB,$BF,$E1,$4F,$B6,$FA,$7C,$E6,$92,$3E,$64,$65
	.dc.b	$B1,$94,$7F,$87,$1F,$27,$C6,$78,$C9,$1F,$E2,$68,$FD,$77,$DA,$FF
	.dc.b	$42,$8F,$F1,$8B,$5E,$67,$F0,$F7,$0F,$F9,$01,$AE,$7C,$5F,$E0,$31
	.dc.b	$88,$FE,$72,$DB,$39,$C7,$FC,$B3,$F6,$DF,$AD,$F5,$8D,$3C,$B8,$7E
	.dc.b	$C7,$ED,$9F,$23,$F9,$86,$E5,$F8,$F6,$BE,$83,$F1,$0E,$1F,$F3,$C3
	.dc.b	$5F,$F3,$53,$5D,$C7,$53,$F4,$BE,$C7,$E1,$E3,$FC,$C7,$CD,$FC,$0E
	.dc.b	$31,$C5,$F8,$BC,$D8,$78,$A3,$FC,$0F,$F6,$CF,$93,$FE,$99,$D8,$8C
	.dc.b	$65,$FE,$AD,$6B,$FE,$99,$6B,$FE,$97,$6B,$F7,$A3,$94,$F4,$9C,$CF
	.dc.b	$FA,$5D,$BC,$1F,$CC,$B9,$3F,$96,$78,$1F,$92,$C7,$F4,$53,$9B,$FF
	.dc.b	$56,$5F,$8A,$7D,$9C,$CF,$B3,$FE,$44,$DA,$FF,$67,$C7,$F1,$9F,$B6
	.dc.b	$E1,$C1,$E5,$F8,$4F,$ED,$1B,$7F,$B3,$EF,$E0,$C6,$0F,$ED,$9C,$0F
	.dc.b	$B5,$F2,$5F,$FA,$77,$ED,$BF,$DC,$77,$F9,$CD,$3E,$74,$62,$11,$F4
	.dc.b	$6B,$8C,$7E,$8B,$ED,$BC,$CF,$F7,$D6,$BF,$85,$6B,$F9,$E6,$BC,$74
	.dc.b	$FF,$00,$46,$20,$D3,$78,$FF,$BC,$FC,$CF,$E8,$DF,$13,$FD,$9B,$3F
	.dc.b	$F5,$B3,$6F,$A1,$EB,$8F,$F5,$D7,$DB,$7F,$B2,$63,$07,$F1,$F7,$0F
	.dc.b	$BA,$37,$F0,$9F,$D4,$E7,$DF,$81,$DE,$9F,$93,$6D,$70,$BE,$B5,$BE
	.dc.b	$5F,$ED,$DF,$1F,$FE,$F6,$9C,$FC,$4B,$EB,$BF,$B0,$36,$3C,$BC,$EA
	.dc.b	$E1,$1E,$53,$FC,$6E,$E3,$3F,$D2,$57,$19,$D6,$28,$C4,$FE,$4A,$B8
	.dc.b	$65,$FC,$43,$D0,$7C,$77,$F8,$C5,$B8,$FF,$8D,$5A,$FF,$8E,$5A,$FF
	.dc.b	$01,$6B,$CE,$39,$11,$FF,$20,$7C,$B3,$ED,$BF,$D9,$91,$EE,$EC,$FE
	.dc.b	$DD,$97,$E9,$9B,$8F,$F2,$EB,$E1,$F0,$FF,$E1,$7F,$B6,$7D,$DD,$26
	.dc.b	$3F,$23,$ED,$FF,$C1,$B3,$A8,$F9,$6D,$7E,$57,$0A,$31,$9F,$ED,$AB
	.dc.b	$FC,$1F,$F5,$BF,$E2,$2B,$F9,$6F,$FB,$2D,$FB,$AD,$73,$BF,$F8,$66
	.dc.b	$38,$57,$F1,$77,$4D,$F8,$8D,$4F,$F9,$0A,$BF,$C6,$7F,$03,$F4,$48
	.dc.b	$FD,$16,$BF,$45,$AF,$D1,$7F,$C1,$BD,$3E,$0D,$FF,$47,$5F,$A6,$3B
	.dc.b	$1F,$E7,$AA,$FC,$8F,$D9,$7E,$43,$E4,$24,$60,$87,$E2,$B9,$91,$A8
	.dc.b	$C4,$FF,$8D,$FE,$DF,$EB,$C7,$F8,$25,$BD,$07,$3C,$2F,$A2,$FF,$18
	.dc.b	$D7,$E9,$A7,$64,$F6,$BF,$CA,$93,$A9,$FD,$01,$FE,$58,$FA,$5F,$CB
	.dc.b	$11,$FC,$B3,$5F,$CB,$35,$FC,$B3,$FE,$32,$FA,$5F,$F1,$5C,$7F,$87
	.dc.b	$DB,$89,$CC,$77,$3F,$13,$F1,$BF,$91,$23,$FB,$6F,$B6,$7D,$BE,$23
	.dc.b	$F1,$BD,$1F,$EE,$87,$89,$B8,$FF,$2F,$7D,$67,$B7,$9F,$FA,$E3,$79
	.dc.b	$3F,$C1,$B9,$A3,$17,$FE,$35,$5F,$6F,$CB,$E0,$64,$FF,$10,$9D,$D7
	.dc.b	$1A,$EF,$F6,$FF,$E5,$29,$34,$95,$FE,$4B,$8F,$E0,$F8,$78,$9B,$1E
	.dc.b	$1F,$13,$9B,$DD,$FF,$45,$7D,$9B,$EA,$69,$EE,$89,$3F,$C7,$27,$4D
	.dc.b	$9C,$BF,$C6,$29,$CF,$E7,$24,$79,$5F,$57,$47,$CA,$24,$7C,$18,$FF
	.dc.b	$4E,$7C,$2F,$28,$FF,$68,$74,$FF,$A3,$E7,$FD,$53,$39,$9F,$F0,$D4
	.dc.b	$E5,$F8,$5F,$41,$F5,$88,$FF,$A4,$2A,$BE,$78,$BC,$3F,$DF,$AF,$A3
	.dc.b	$F7,$23,$3A,$FC,$B7,$CB,$7F,$82,$E9,$FA,$84,$C3,$E9,$3F,$5F,$ED
	.dc.b	$FF,$0E,$48,$FB,$D1,$EE,$F9,$7F,$F2,$4E,$1F,$E4,$AC,$FF,$D5,$91
	.dc.b	$8A,$FF,$6D,$66,$8C,$47,$FC,$00,$8C,$3F,$E5,$CA,$E5,$18,$8F,$F1
	.dc.b	$A5,$74,$74,$74,$AE,$71,$98,$DC,$F1,$F9,$17,$D3,$7E,$73,$EA,$9F
	.dc.b	$65,$FE,$60,$9C,$DF,$FA,$46,$72,$FF,$9B,$A3,$38,$75,$6F,$C7,$BE
	.dc.b	$6B,$96,$3F,$EF,$78,$C5,$F1,$8C,$4E,$A1,$7F,$EE,$97,$0F,$FB,$AE
	.dc.b	$2D,$72,$95,$F8,$27,$31,$A6,$3D,$CE,$B3,$C6,$7C,$F1,$A3,$9D,$F9
	.dc.b	$1E,$17,$E1,$6D,$7F,$9B,$6B,$E0,$FD,$BB,$E5,$BC,$ED,$AE,$2D,$7A
	.dc.b	$9E,$E5,$F6,$FC,$26,$4F,$07,$9C,$62,$38,$E0,$AE,$97,$E2,$BD,$D7
	.dc.b	$BC,$79,$7F,$8D,$FB,$2C,$36,$3C,$66,$B8,$13,$F1,$FD,$E7,$DB,$FF
	.dc.b	$9D,$31,$FC,$1D,$EE,$B9,$DF,$08,$E7,$87,$5A,$FC,$4E,$1C,$BE,$0F
	.dc.b	$EF,$57,$13,$FD,$2C,$53,$FD,$31,$5E,$56,$DF,$57,$3F,$47,$F8,$17
	.dc.b	$FD,$6B,$28,$C4,$7F,$B0,$A3,$1C,$3D,$A3,$E1,$CE,$9B,$1F,$12,$3F
	.dc.b	$2B,$8F,$A3,$E0,$1E,$27,$FB,$9E,$3E,$0E,$1D,$30,$F7,$DF,$0E,$3F
	.dc.b	$1B,$61,$A3,$1E,$7C,$FD,$F3,$FC,$6F,$7B,$C0,$ED,$97,$FB,$6B,$93
	.dc.b	$F9,$C4,$63,$2F,$D2,$D3,$E7,$31,$5A,$C7,$9D,$E2,$BA,$47,$6A,$C4
	.dc.b	$C8,$D9,$92,$DF,$E3,$BE,$DF,$CE,$DE,$0E,$F1,$FC,$5E,$7F,$4A,$EB
	.dc.b	$83,$F8,$1B,$73,$F5,$36,$FF,$7A,$F1,$7A,$57,$2B,$EA,$76,$E1,$B1
	.dc.b	$92,$4F,$96,$32,$FE,$93,$ED,$9F,$6F,$F1,$FC,$87,$63,$E6,$3E,$3D
	.dc.b	$6B,$1F,$E0,$95,$A9,$CD,$A7,$F9,$5E,$1F,$C5,$2F,$83,$FC,$C1,$B1
	.dc.b	$FC,$F5,$B1,$F6,$5A,$F1,$9E,$8F,$16,$CF,$16,$C7,$F7,$DF,$9A,$DF
	.dc.b	$8B,$FE,$12,$6D,$7F,$80,$7E,$0A,$D7,$F8,$46,$6E,$5F,$E1,$59,$FF
	.dc.b	$7C,$4E,$EB,$FC,$15,$18,$8F,$F0,$C3,$67,$FC,$39,$F0,$3C,$BC,$93
	.dc.b	$DB,$99,$A7,$F4,$6C,$7F,$C3,$93,$B2,$3F,$EA,$68,$D6,$3F,$E0,$99
	.dc.b	$CB,$67,$5E,$1E,$34,$FF,$79,$DF,$0A,$F9,$93,$E7,$C3,$C2,$D9,$D4
	.dc.b	$FF,$88,$FE,$4F,$D0,$7A,$7A,$2E,$3F,$E3,$17,$DD,$9F,$F8,$8E,$3F
	.dc.b	$C5,$91,$83,$3F,$69,$EF,$F8,$F7,$DB,$7C,$38,$C3,$6B,$FD,$87,$3A
	.dc.b	$7F,$C7,$D3,$A6,$93,$1F,$F1,$57,$D1,$71,$6C,$7F,$8F,$63,$CA,$47
	.dc.b	$F9,$2F,$03,$ED,$FF,$9A,$C7,$FB,$69,$B5,$FE,$CC,$7C,$4D,$F1,$3F
	.dc.b	$07,$1A,$E1,$FC,$8D,$F3,$9C,$E7,$A7,$F1,$3D,$6F,$62,$9F,$81,$FB
	.dc.b	$7F,$E0,$78,$72,$9D,$67,$FE,$EC,$6E,$7F,$ED,$E6,$CF,$E8,$39,$17
	.dc.b	$CA,$FF,$CD,$75,$A7,$FC,$DB,$EC,$BE,$8B,$F3,$6D,$86,$3F,$EE,$FF
	.dc.b	$9D,$7D,$03,$8B,$78,$7E,$92,$73,$1F,$9F,$6D,$FF,$50,$6D,$F4,$3A
	.dc.b	$B8,$70,$D7,$37,$DE,$79,$36,$39,$09,$E0,$FF,$8A,$7C,$AF,$DA,$F7
	.dc.b	$9F,$D6,$7B,$BC,$77,$FA,$B3,$9C,$7D,$F1,$8A,$E2,$70,$CA,$CE,$8F
	.dc.b	$F4,$AF,$D9,$39,$FF,$81,$AF,$87,$B1,$86,$CF,$1C,$DC,$CB,$E1,$39
	.dc.b	$8C,$5F,$18,$C7,$C1,$BE,$AB,$F8,$8B,$63,$A6,$FF,$5B,$B7,$35,$F5
	.dc.b	$78,$51,$F5,$5B,$5F,$32,$30,$47,$04,$CF,$E7,$DF,$58,$FD,$3F,$DB
	.dc.b	$F2,$97,$07,$E8,$23,$F8,$04,$62,$7F,$D7,$71,$B9,$D7,$6F,$F2,$A8
	.dc.b	$C4,$EA,$31,$86,$5F,$B7,$F6,$F2,$7E,$57,$4F,$F7,$5F,$43,$19,$77
	.dc.b	$3F,$CD,$55,$8C,$BF,$49,$E6,$FF,$86,$70,$F9,$DC,$CA,$FE,$7B,$F1
	.dc.b	$6F,$B6,$FF,$3E,$E1,$FA,$BF,$B0,$7D,$BE,$A7,$38,$FF,$82,$67,$2F
	.dc.b	$DD,$F4,$9F,$6F,$FE,$80,$8F,$F1,$8C,$63,$C0,$FE,$88,$D8,$CC,$EF
	.dc.b	$6F,$CE,$CE,$5B,$19,$7F,$55,$19,$6C,$65,$FF,$0E,$7C,$9F,$D2,$CF
	.dc.b	$F8,$9F,$5B,$EB,$FF,$A7,$4E,$71,$FC,$DC,$63,$93,$F3,$6D,$8F,$CA
	.dc.b	$57,$17,$F2,$DF,$6C,$F9,$5F,$EC,$6D,$8C,$FD,$BE,$26,$47,$07,$71
	.dc.b	$8C,$FE,$C7,$7B,$B9,$F4,$5F,$9E,$3F,$35,$F0,$D7,$1F,$CF,$F7,$A3
	.dc.b	$F3,$84,$8F,$E0,$ED,$C3,$F3,$33,$B9,$C9,$23,$F8,$5E,$0F,$9F,$F1
	.dc.b	$4F,$EC,$FE,$DF,$FD,$20,$FE,$D6,$33,$5E,$89,$CB,$F9,$1F,$79,$F3
	.dc.b	$EF,$AB,$FF,$01,$46,$1B,$19,$C3,$FC,$EB,$39,$7F,$A9,$FD,$B7,$FA
	.dc.b	$F5,$FE,$08,$DB,$ED,$F6,$EF,$96,$FF,$09,$CE,$7A,$3F,$31,$F5,$BF
	.dc.b	$E0,$BE,$D3,$FE,$1A,$C9,$F1,$FF,$E3,$1C,$FF,$57,$5A,$7F,$8B,$4E
	.dc.b	$9F,$4C,$EF,$3F,$E6,$33,$99,$F4,$CE,$67,$F9,$F4,$E5,$DC,$65,$FE
	.dc.b	$9B,$19,$77,$19,$7F,$92,$FD,$3B,$DD,$FF,$17,$B6,$FF,$93,$7D,$4B
	.dc.b	$ED,$F3,$5F,$E3,$48,$C4,$FF,$7C,$6D,$FD,$E6,$BF,$D4,$F9,$11,$88
	.dc.b	$FE,$C3,$18,$BF,$F1,$C4,$E6,$7F,$9C,$B6,$75,$5F,$E0,$AC,$3F,$C7
	.dc.b	$6F,$F3,$B6,$D7,$F1,$5F,$29,$F6,$FF,$C4,$E3,$F4,$33,$AD,$FF,$BD
	.dc.b	$4E,$67,$FB,$45,$70,$F2,$FF,$C6,$D3,$99,$FF,$4D,$35,$FF,$7E,$B6
	.dc.b	$78,$61,$FE,$C2,$6C,$FF,$6A,$6B,$F0,$FE,$DF,$FA,$9B,$FB,$46,$E1
	.dc.b	$FE,$1C,$9F,$EA,$13,$9F,$4B,$FC,$C7,$18,$8F,$98,$D7,$E5,$B5,$FE
	.dc.b	$C2,$DC,$CD,$27,$A3,$FE,$67,$75,$87,$F1,$1A,$F2,$78,$7F,$E6,$7F
	.dc.b	$AF,$7D,$07,$F9,$D2,$B5,$9F,$F8,$8E,$73,$A7,$C1,$6E,$1F,$05,$AF
	.dc.b	$CE,$6B,$FE,$C5,$6B,$E6,$6E,$26,$9A,$CD,$F1,$5F,$D1,$AB,$FB,$FB
	.dc.b	$63,$FD,$41,$F1,$1F,$E8,$A8,$C6,$1F,$EB,$7A,$FC,$A3,$5F,$C9,$B5
	.dc.b	$FF,$04,$B6,$B9,$BF,$E5,$A3,$04,$AF,$F4,$85,$7F,$05,$8F,$CE,$B6
	.dc.b	$BE,$8F,$D8,$BE,$7B,$C9,$E6,$FF,$A4,$E3,$17,$F8,$96,$D7,$E2,$5A
	.dc.b	$FE,$59,$AF,$EA,$BE,$DB,$FC,$98,$FF,$3E,$6E,$BF,$EB,$6E,$A7,$DB
	.dc.b	$FF,$31,$C3,$FC,$45,$5A,$79,$7B,$DF,$EB,$08,$C6,$1F,$F0,$02,$B8
	.dc.b	$3D,$67,$FD,$6B,$87,$EA,$EF,$F7,$6D,$9F,$C7,$FC,$BB,$E1,$B7,$F1
	.dc.b	$52,$FF,$C6,$57,$BE,$35,$EE,$73,$5F,$EB,$7F,$B4,$FE,$AF,$1F,$E1
	.dc.b	$46,$E3,$FB,$FF,$98,$7B,$9B,$F3,$A5,$ED,$FF,$60,$4E,$A3,$07,$F8
	.dc.b	$79,$C5,$7E,$DD,$AF,$F0,$46,$BF,$E7,$8C,$91,$88,$FD,$6E,$1E,$AA
	.dc.b	$F3,$65,$FE,$CF,$FB,$77,$D2,$FA,$E2,$4E,$8F,$F1,$7C,$E9,$C5,$FF
	.dc.b	$8B,$CF,$F3,$E7,$37,$F9,$D1,$FF,$3A,$B5,$C9,$4F,$D2,$9B,$FB,$77
	.dc.b	$CC,$FF,$A3,$A3,$0D,$7F,$CF,$F1,$AF,$47,$A3,$63,$97,$1F,$F9,$49
	.dc.b	$BC,$5A,$8C,$47,$26,$BF,$1F,$57,$D3,$7F,$A5,$6B,$FD,$55,$18,$9F
	.dc.b	$B9,$39,$CF,$E7,$6E,$FA,$9C,$57,$CE,$AB,$C2,$C3,$94,$FF,$41,$6E
	.dc.b	$3F,$EB,$7C,$39,$4F,$F9,$73,$DA,$75,$FC,$9C,$11,$89,$FE,$74,$D7
	.dc.b	$F9,$D3,$5F,$E8,$FF,$72,$FA,$4F,$1B,$63,$FC,$D2,$D7,$FA,$A4,$F3
	.dc.b	$C7,$FD,$95,$39,$6D,$72,$3D,$B3,$BD,$7F,$B3,$F0,$3E,$6B,$E1,$F8
	.dc.b	$3F,$EC,$86,$C6,$72,$FC,$E6,$F8,$99,$1F,$EA,$19,$D3,$FA,$99,$E7
	.dc.b	$DC,$FA,$1A,$49,$FF,$0B,$FC,$D7,$FB,$55,$FF,$55,$4E,$DB,$3B,$7F
	.dc.b	$C6,$D3,$B7,$FC,$A3,$F6,$FF,$E3,$77,$F3,$59,$23,$11,$FE,$46,$6B
	.dc.b	$FE,$09,$6B,$FE,$08,$F8,$BF,$F1,$7B,$FE,$BB,$6B,$FE,$5E,$6E,$ED
	.dc.b	$7F,$C1,$F5,$DA,$31,$5F,$D2,$3E,$D7,$F5,$58,$7F,$4C,$6F,$51,$A4
	.dc.b	$9F,$4C,$65,$B1,$97,$FC,$BF,$F7,$BF,$C8,$5F,$F6,$83,$5F,$F3,$9B
	.dc.b	$6F,$C7,$7E,$19,$FF,$6B,$72,$3E,$E3,$CA,$D8,$F9,$2D,$7F,$CD,$AD
	.dc.b	$7F,$C0,$B1,$C0,$8F,$07,$DF,$3F,$ED,$88,$FF,$72,$35,$FF,$43,$B6
	.dc.b	$BF,$C2,$93,$F9,$0F,$97,$FC,$4D,$7F,$8F,$5A,$FF,$8F,$1A,$FF,$BC
	.dc.b	$9B,$46,$93,$8F,$E8,$BA,$FC,$43,$FD,$21,$5F,$EF,$76,$BF,$E9,$A6
	.dc.b	$E1,$F2,$27,$F3,$6F,$CA,$FB,$87,$C9,$7D,$C6,$DF,$E4,$1A,$FF,$A5
	.dc.b	$1A,$FF,$95,$DA,$FF,$95,$9A,$B1,$46,$1B,$3F,$8A,$F5,$7F,$1B,$5F
	.dc.b	$E5,$0E,$1F,$DB,$C6,$30,$F3,$57,$EE,$BE,$91,$F4,$DF,$92,$6E,$1F
	.dc.b	$A6,$6B,$FE,$AA,$6B,$FE,$6D,$6B,$FE,$6C,$6B,$E7,$6E,$26,$FE,$51
	.dc.b	$F8,$9C,$3F,$22,$DC,$BF,$CE,$AD,$BE,$58,$7F,$32,$E8,$FF,$58,$C7
	.dc.b	$F5,$16,$BF,$EA,$EF,$82,$FE,$A5,$18,$8F,$F4,$DC,$60,$8F,$17,$FA
	.dc.b	$C6,$4F,$B7,$79,$9F,$C7,$9B,$43,$FD,$B6,$E1,$FF,$62,$56,$5F,$F6
	.dc.b	$57,$CC,$7F,$26,$F8,$73,$4F,$57,$E8,$9B,$9F,$F7,$2F,$C2,$7E,$58
	.dc.b	$9C,$DF,$C1,$1F,$CC,$FD,$3B,$EA,$BF,$98,$BF,$C8,$6B,$FD,$DA,$DE
	.dc.b	$2F,$DD,$BB,$8F,$44,$7E,$13,$C2,$FF,$01,$C7,$F7,$96,$BF,$DF,$5A
	.dc.b	$F9,$5A,$FB,$5A,$FC,$36,$BF,$DB,$BD,$0F,$F0,$14,$6B,$83,$FE,$00
	.dc.b	$3F,$D2,$3E,$33,$FB,$2F,$2F,$FA,$C9,$B7,$CE,$FF,$A6,$FD,$AF,$F0
	.dc.b	$1A,$FF,$5B,$B6,$7F,$44,$DC,$0F,$F6,$A8,$F7,$CF,$0F,$9A,$E6,$48
	.dc.b	$C9,$CB,$C8,$FE,$B7,$F1,$4F,$AC,$BF,$13,$2B,$79,$7F,$BC,$E7,$66
	.dc.b	$61,$B7,$F2,$AB,$FF,$78,$C7,$07,$A3,$E0,$3F,$C2,$7E,$BF,$C6,$23
	.dc.b	$F4,$DA,$FD,$36,$BF,$4D,$FF,$04,$7D,$83,$E6,$7F,$B9,$FE,$1C,$FB
	.dc.b	$1A,$7F,$85,$A7,$57,$EA,$FB,$3F,$C2,$3E,$8E,$0F,$F7,$1C,$63,$3F
	.dc.b	$E8,$6D,$8F,$F2,$87,$D1,$3D,$0F,$F0,$0C,$E6,$FF,$8A,$37,$73,$9F
	.dc.b	$2D,$7C,$6B,$E5,$FD,$7F,$9D,$BC,$DF,$E0,$C8,$FC,$9E,$7F,$D5,$1D
	.dc.b	$4F,$9C,$FE,$44,$E2,$35,$3C,$4E,$5A,$FF,$8E,$2B,$FC,$83,$F4,$DF
	.dc.b	$94,$23,$FA,$16,$BF,$A1,$6B,$FA,$17,$FB,$AF,$D2,$BE,$BB,$FC,$42
	.dc.b	$D9,$FC,$2B,$5F,$F3,$47,$5A,$B7,$E3,$F0,$FB,$59,$5C,$23,$2F,$28
	.dc.b	$CB,$EF,$E0,$61,$FC,$CB,$4F,$F0,$4E,$EE,$3F,$F5,$E5,$FE,$72,$31
	.dc.b	$F1,$1F,$81,$3F,$CA,$FF,$69,$F1,$F9,$7F,$D1,$57,$C2,$FE,$65,$F0
	.dc.b	$79,$46,$70,$EF,$7C,$67,$F9,$17,$E1,$CE,$AF,$F4,$03,$63,$FD,$08
	.dc.b	$D7,$FD,$10,$D7,$F9,$03,$5F,$E5,$0D,$7A,$C7,$43,$FC,$4B,$F4,$7F
	.dc.b	$E1,$F8,$F8,$8D,$BF,$F0,$CB,$72,$E7,$5F,$E5,$6F,$AD,$FC,$85,$FC
	.dc.b	$B6,$DF,$F8,$81,$B5,$FC,$F7,$64,$7F,$99,$63,$0E,$AB,$FB,$93,$67
	.dc.b	$7D,$3D,$FD,$DF,$D8,$7A,$32,$FE,$5B,$FB,$3F,$93,$C4,$7D,$0D,$FF
	.dc.b	$C0,$79,$FD,$C6,$E7,$F5,$AB,$FC,$DF,$F5,$AF,$1B,$FC,$8E,$DC,$3F
	.dc.b	$CB,$6D,$7F,$A6,$68,$8C,$47,$F5,$58,$C6,$FD,$09,$1F,$C7,$A7,$96
	.dc.b	$5F,$E4,$45,$3F,$EA,$0F,$C6,$1F,$29,$4E,$37,$F9,$7A,$DB,$F9,$39
	.dc.b	$DB,$67,$77,$D7,$EC,$3D,$64,$9F,$5B,$EE,$EC,$F8,$24,$8F,$34,$7F
	.dc.b	$AC,$BE,$15,$F5,$7F,$EA,$B6,$CF,$FB,$5F,$A3,$FD,$CF,$39,$3E,$4F
	.dc.b	$CF,$7C,$C2,$4F,$FB,$CE,$2B,$EB,$8B,$8F,$FA,$EE,$6F,$0F,$E1,$F5
	.dc.b	$FF,$3F,$B6,$BC,$93,$F6,$3E,$CB,$EB,$93,$0F,$AE,$FD,$AF,$8B,$FB
	.dc.b	$A4,$8F,$A5,$1E,$DF,$89,$FF,$78,$F0,$FF,$B5,$1B,$2D,$1A,$FF,$B1
	.dc.b	$FE,$13,$E8,$C6,$27,$6E,$AA,$D6,$EF,$A5,$7F,$87,$6F,$8C,$FF,$BF
	.dc.b	$A7,$32,$9C,$D7,$23,$9C,$76,$FB,$4F,$F7,$FB,$CE,$BF,$C5,$18,$27
	.dc.b	$33,$CE,$73,$18,$BF,$4B,$FE,$50,$6F,$C7,$75,$25,$75,$3E,$14,$EE
	.dc.b	$B9,$CE,$A7,$B1,$A7,$CD,$E1,$E6,$FA,$CF,$3C,$3C,$D5,$A8,$C7,$A0
	.dc.b	$F9,$AF,$96,$BC,$FA,$62,$24,$F0,$8C,$BC,$A3,$2F,$92,$FF,$E0,$0C
	.dc.b	$EA,$31,$5D,$DF,$70,$F2,$DF,$2E,$FB,$1C,$C6,$1B,$19,$9F,$83,$F0
	.dc.b	$1F,$E7,$1F,$73,$F1,$31,$CE,$75,$AF,$EE,$23,$17,$C8,$F9,$D7,$C6
	.dc.b	$30,$7F,$96,$4D,$97,$0F,$13,$6F,$C4,$FF,$97,$E1,$F3,$AE,$2E,$53
	.dc.b	$ED,$8C,$3C,$DA,$F1,$6B,$FE,$6F,$9E,$73,$F0,$A7,$8E,$3F,$C4,$B0
	.dc.b	$7B,$3F,$E1,$9E,$9D,$5C,$C7,$F0,$A7,$85,$F7,$CE,$E3,$38,$F6,$7E
	.dc.b	$C7,$ED,$0F,$97,$FF,$7D,$E5,$FB,$0F,$C3,$BE,$A6,$5F,$8B,$AD,$3F
	.dc.b	$8D,$8C,$DF,$2B,$E3,$CD,$FA,$69,$CE,$1F,$BA,$C3,$8D,$FE,$1E,$73
	.dc.b	$5F,$AC,$9C,$D7,$87,$7E,$77,$FC,$16,$73,$5D,$67,$3C,$6E,$27,$BF
	.dc.b	$86,$E9,$FA,$0F,$98,$C4,$C8,$FD,$1B,$B8,$C6,$5F,$98,$8B,$9F,$F1
	.dc.b	$2B,$56,$32,$FE,$07,$1D,$AF,$86,$1D,$CE,$A7,$EC,$FE,$DB,$E9,$61
	.dc.b	$FC,$71,$B7,$FC,$31,$FD,$1C,$EF,$CC,$F6,$C6,$AF,$F9,$33,$73,$3C
	.dc.b	$23,$F6,$4E,$49,$3C,$B9,$79,$BF,$33,$C7,$FE,$D9,$D4,$C7,$FB,$43
	.dc.b	$70,$FE,$C5,$D7,$7C,$3F,$ED,$53,$6B,$FA,$44,$FF,$BD,$AA,$C7,$F7
	.dc.b	$A8,$C4,$7F,$BF,$FC,$3F,$F7,$AC,$FF,$7E,$F4,$F9,$4E,$5B,$AF,$B3
	.dc.b	$AF,$39,$FF,$80,$66,$E0,$69,$C3,$6F,$D9,$E3,$FE,$01,$8C,$B4,$FF
	.dc.b	$26,$46,$B1,$FE,$FD,$3C,$70,$F3,$E3,$E1,$BF,$EE,$17,$BB,$FF,$72
	.dc.b	$4E,$5F,$F0,$AE,$1F,$E9,$9A,$FF,$08,$7B,$7E,$6F,$4F,$CE,$D8,$FF
	.dc.b	$0E,$C7,$C4,$CF,$FC,$38,$DA,$37,$CE,$F8,$7C,$AF,$B3,$F8,$31,$86
	.dc.b	$E1,$FE,$67,$9D,$3F,$E2,$19,$D3,$49,$97,$F8,$77,$CE,$E2,$D8,$FF
	.dc.b	$18,$C7,$8C,$8F,$F8,$D3,$CC,$FA,$27,$D2,$FB,$DB,$1F,$E2,$F8,$B5
	.dc.b	$FE,$29,$9B,$C9,$FE,$31,$EB,$FB,$27,$DA,$F8,$DB,$87,$F8,$DE,$2D
	.dc.b	$FF,$8C,$66,$D7,$F9,$77,$A3,$F0,$7E,$9F,$E7,$E3,$F2,$DF,$27,$FE
	.dc.b	$D8,$C3,$35,$F8,$C8,$C6,$3F,$CF,$5B,$88,$81,$CB,$86,$BB,$6B,$BB
	.dc.b	$FE,$42,$ED,$AE,$DA,$B8,$BF,$CA,$3E,$CF,$E7,$3B,$3F,$D5,$98,$7F
	.dc.b	$A7,$FE,$D7,$F9,$B6,$1F,$E8,$2A,$D1,$27,$5D,$B7,$1F,$F4,$4F,$B7
	.dc.b	$CF,$E8,$BA,$36,$BF,$DF,$F9,$23,$15,$C2,$73,$5F,$AE,$F7,$1E,$8D
	.dc.b	$3E,$57,$CB,$FE,$32,$FF,$D7,$53,$AB,$F9,$5A,$7D,$2A,$D5,$7C,$96
	.dc.b	$BF,$25,$AF,$CF,$6E,$E6,$93,$A7,$F8,$4C,$65,$FD,$47,$DA,$BE,$9F
	.dc.b	$F8,$2A,$F8,$FB,$18,$6E,$5E,$63,$FD,$49,$3B,$71,$8F,$D5,$8E,$91
	.dc.b	$8E,$34,$6E,$7F,$DB,$5E,$B3,$D1,$FE,$2A,$DB,$FF,$6C,$7C,$1F,$C7
	.dc.b	$DB,$F9,$56,$3F,$E7,$CB,$D9,$2B,$7A,$5C,$7F,$DC,$DF,$09,$FE,$D3
	.dc.b	$9F,$13,$FC,$5F,$0E,$6E,$67,$7E,$37,$F2,$AA,$7E,$0E,$F8,$4F,$E6
	.dc.b	$31,$86,$C6,$63,$FC,$40,$64,$AE,$6B,$F4,$A7,$F4,$57,$0F,$EA,$23
	.dc.b	$83,$A6,$BF,$EB,$5F,$B1,$78,$7F,$CE,$97,$87,$63,$0D,$AF,$A1,$3A
	.dc.b	$6C,$E9,$FA,$9F,$63,$FE,$A6,$6F,$0F,$F4,$68,$7D,$55,$89,$3A,$38
	.dc.b	$C6,$31,$6D,$F5,$BE,$0B,$FC,$C9,$5F,$81,$6C,$EA,$3F,$35,$F1,$9F
	.dc.b	$D5,$27,$35,$FA,$18,$FC,$BF,$49,$A4,$AF,$E2,$51,$96,$C6,$6F,$E7
	.dc.b	$F4,$FE,$8E,$3F,$A3,$7D,$73,$E0,$3F,$B1,$BF,$B1,$AF,$21,$FC,$35
	.dc.b	$D5,$FE,$65,$D6,$3F,$92,$F9,$07,$D7,$7F,$51,$6D,$7F,$98,$7A,$51
	.dc.b	$8C,$3F,$DE,$2D,$7F,$DE,$0F,$F2,$99,$FD,$33,$63,$C7,$EF,$3E,$49
	.dc.b	$F3,$BF,$DE,$AB,$59,$7E,$93,$E5,$1C,$3F,$DF,$9B,$97,$ED,$31,$7C
	.dc.b	$AF,$8E,$F8,$5F,$F8,$22,$73,$AF,$F8,$39,$B7,$FE,$7B,$6E,$9F,$9A
	.dc.b	$6B,$FD,$01,$AF,$F4,$36,$A6,$98,$D5,$F2,$EF,$A5,$FF,$0A,$B6,$FF
	.dc.b	$84,$7C,$CB,$ED,$B3,$5F,$E1,$A8,$C4,$FC,$F6,$E3,$FA,$EC,$11,$88
	.dc.b	$FE,$AB,$18,$BF,$F0,$44,$E6,$7F,$AD,$4E,$6B,$A5,$68,$FF,$07,$7C
	.dc.b	$DF,$F1,$CA,$FC,$3D,$23,$1A,$FF,$16,$FA,$DF,$D4,$E1,$FC,$86,$72
	.dc.b	$F6,$8C,$DF,$9D,$AF,$99,$AE,$FA,$7F,$C3,$D1,$88,$FE,$C4,$DC,$0D
	.dc.b	$31,$F3,$AF,$57,$FC,$82,$DC,$3F,$9A,$47,$E3,$1F,$E6,$D1,$9B,$F6
	.dc.b	$CE,$5F,$98,$D7,$E5,$B5,$F9,$2D,$7F,$BB,$36,$FC,$5C,$BF,$E3,$78
	.dc.b	$C1,$9F,$AC,$FE,$2D,$5F,$31,$BA,$7F,$40,$F7,$7F,$82,$E5,$F5,$A7
	.dc.b	$2F,$D8,$8F,$E1,$B5,$F7,$9A,$E7,$E3,$3F,$C8,$D1,$88,$FF,$05,$37
	.dc.b	$23,$49,$3F,$E4,$DF,$8B,$FF,$37,$D7,$E0,$1B,$8F,$F5,$4F,$A1,$7C
	.dc.b	$AF,$F9,$BA,$73,$97,$E3,$27,$2F,$F0,$56,$DF,$E5,$DA,$FE,$59,$AF
	.dc.b	$1A,$FF,$2F,$46,$23,$FC,$44,$DC,$8D,$F8,$6F,$E0,$B7,$F9,$C6,$D7
	.dc.b	$F8,$A6,$3F,$C8,$06,$5F,$F4,$2B,$FE,$A1,$7F,$4E,$D7,$F4,$AD,$79
	.dc.b	$7B,$5F,$E7,$1F,$AA,$FE,$4F,$1F,$A4,$6E,$7F,$EA,$1E,$97,$D5,$78
	.dc.b	$B3,$56,$B1,$FD,$8B,$70,$FF,$1F,$36,$F8,$C7,$FA,$6A,$35,$5F,$D4
	.dc.b	$3D,$4F,$F4,$EE,$3F,$AE,$AF,$B7,$F2,$5F,$C7,$F1,$FD,$EC,$E6,$F8
	.dc.b	$47,$F8,$56,$38,$70,$AB,$85,$FF,$93,$E7,$3F,$4C,$FA,$6F,$F5,$6B
	.dc.b	$70,$FF,$10,$E1,$FE,$E9,$F2,$B3,$52,$73,$5B,$C3,$FA,$7D,$EE,$73
	.dc.b	$5F,$EB,$8F,$8C,$FE,$E9,$5F,$E0,$8A,$D3,$63,$2F,$FA,$A9,$AF,$F9
	.dc.b	$CE,$1F,$4E,$F9,$4F,$F6,$1B,$70,$F3,$E9,$FE,$F2,$7D,$31,$9B,$F2
	.dc.b	$B7,$4F,$F6,$D6,$1D,$30,$FF,$66,$E1,$C6,$F7,$1F,$ED,$27,$FD,$07
	.dc.b	$C7,$FE,$20,$7F,$C4,$2D,$72,$50,$E7,$EA,$5F,$1F,$FE,$5B,$9E,$51
	.dc.b	$8F,$3F,$53,$99,$FF,$16,$CE,$5D,$FB,$DF,$E1,$46,$E7,$D6,$31,$1C
	.dc.b	$1A,$FC,$9D,$3F,$CF,$CF,$C4,$F8,$0F,$F4,$ED,$7C,$29,$CC,$FD,$AD
	.dc.b	$1F,$53,$AD,$F9,$61,$6B,$F8,$6D,$F0,$C3,$94,$E7,$7F,$C8,$61,$78
	.dc.b	$1D,$9F,$1E,$51,$8C,$7E,$64,$62,$75,$39,$9F,$C4,$71,$FF,$14,$7F
	.dc.b	$23,$D9,$DE,$31,$1F,$92,$9E,$19,$7D,$E9,$CB,$6B,$81,$E5,$BF,$67
	.dc.b	$37,$FA,$3F,$E4,$6F,$19,$A7,$B7,$83,$E0,$70,$7F,$6E,$9D,$69,$FE
	.dc.b	$85,$6B,$FD,$B1,$B9,$7E,$42,$B8,$CF,$FB,$07,$AB,$F6,$4F,$FA,$35
	.dc.b	$AF,$F7,$46,$CF,$F1,$1F,$84,$FF,$80,$13,$BE,$8F,$4D,$6B,$1F,$F0
	.dc.b	$0B,$67,$FC,$31,$5C,$1F,$EF,$FF,$53,$FE,$40,$7F,$C7,$CD,$7F,$C2
	.dc.b	$0D,$7F,$C1,$ED,$7F,$C1,$BE,$AB,$E6,$7E,$4B,$63,$FD,$4C,$D7,$BF
	.dc.b	$9D,$F3,$BA,$FF,$C8,$31,$FE,$4D,$6B,$FE,$1F,$6B,$FE,$1E,$6B,$FD
	.dc.b	$07,$C6,$FD,$87,$0F,$EB,$9B,$3A,$9F,$F7,$63,$67,$FA,$5F,$CB,$7D
	.dc.b	$EF,$10,$D3,$C9,$7F,$82,$E8,$7D,$4F,$66,$E3,$FE,$63,$6B,$FE,$35
	.dc.b	$6B,$FD,$DE,$38,$11,$E0,$FA,$30,$FF,$41,$47,$FB,$2D,$AF,$F8,$C9
	.dc.b	$BC,$5F,$E0,$79,$FC,$AF,$91,$F8,$CA,$FF,$0B,$F3,$7C,$28,$C4,$7F
	.dc.b	$86,$23,$04,$7F,$CF,$51,$A7,$FC,$37,$1F,$2B,$EA,$5F,$61,$FA,$26
	.dc.b	$CF,$FB,$7D,$AF,$F8,$AA,$7F,$4F,$E5,$FF,$AE,$A3,$FD,$0C,$D7,$FC
	.dc.b	$B0,$D7,$FC,$AE,$D7,$FC,$AA,$D5,$92,$30,$DF,$73,$EF,$C6,$3C,$F7
	.dc.b	$47,$EE,$1B,$7F,$E5,$58,$F2,$D7,$EF,$3D,$17,$CB,$FE,$21,$B5,$FE
	.dc.b	$6F,$8C,$11,$C1,$0F,$F3,$94,$70,$7F,$CE,$9E,$A7,$E0,$B8,$3F,$DF
	.dc.b	$6D,$C3,$FC,$E2,$D9,$F6,$D7,$1A,$FF,$05,$7A,$DF,$D1,$FB,$3F,$D3
	.dc.b	$75,$FD,$2D,$AF,$FA,$5E,$34,$47,$4D,$7F,$D6,$DF,$54,$F8,$9F,$E3
	.dc.b	$6D,$81,$FE,$7E,$70,$FF,$AD,$27,$FD,$F7,$7F,$EB,$1F,$87,$FE,$47
	.dc.b	$ED,$1A,$7F,$41,$C7,$FD,$75,$7F,$CA,$3D,$BF,$C6,$71,$7F,$B1,$B8
	.dc.b	$FF,$6B,$18,$33,$F3,$8F,$AA,$FE,$5E,$FF,$15,$A3,$4D,$5F,$A3,$3F
	.dc.b	$BA,$BC,$C7,$D3,$7F,$6F,$6C,$7F,$73,$6B,$FD,$DD,$AF,$C2,$3E,$1B
	.dc.b	$87,$0D,$73,$1D,$4F,$E1,$FE,$F7,$F3,$8C,$3F,$DA,$B5,$FD,$22,$3F
	.dc.b	$9D,$D7,$96,$BF,$A5,$FD,$57,$EE,$B9,$7F,$D6,$6D,$9F,$CF,$B7,$03
	.dc.b	$FD,$92,$3E,$04,$F0,$F8,$3E,$64,$8C,$9C,$BC,$FF,$EB,$3F,$4A,$FA
	.dc.b	$9B,$F5,$52,$B7,$92,$B5,$18,$6B,$FE,$C2,$53,$C6,$31,$5F,$E0,$17
	.dc.b	$E9,$7B,$1F,$44,$8F,$D1,$6B,$F4,$5A,$FD,$13,$89,$FE,$00,$FA,$77
	.dc.b	$CF,$7F,$72,$FA,$A7,$C7,$53,$FC,$19,$3A,$AF,$5F,$C8,$7E,$19,$FF
	.dc.b	$6C,$B7,$93,$F8,$1E,$A8,$C7,$07,$F9,$2F,$E4,$9E,$9F,$F8,$02,$73
	.dc.b	$87,$F1,$26,$DF,$1E,$8F,$F0,$E5,$F0,$AF,$97,$F4,$9E,$66,$F3,$7F
	.dc.b	$82,$A3,$F2,$39,$F8,$9D,$4F,$FB,$B9,$B1,$99,$E0,$71,$D7,$E7,$74
	.dc.b	$3E,$63,$F3,$C4,$C3,$F3,$CD,$7F,$3C,$D7,$E7,$47,$17,$FC,$B1,$9B
	.dc.b	$E5,$9F,$4F,$FE,$20,$6C,$FE,$19,$AB,$2F,$D8,$DF,$38,$C7,$07,$F0
	.dc.b	$CC,$3A,$78,$FC,$FE,$8E,$57,$38,$EC,$61,$E2,$7F,$40,$C5,$3A,$8F
	.dc.b	$F4,$8F,$1B,$E8,$7F,$02,$48,$FF,$4A,$FD,$57,$CA,$CB,$FD,$0D,$5B
	.dc.b	$9E,$35,$B7,$FD,$2B,$5B,$7F,$91,$FD,$57,$E8,$DF,$F3,$D3,$5F,$F3
	.dc.b	$F3,$5F,$CC,$35,$FE,$14,$D7,$9F,$67,$FA,$06,$31,$07,$5A,$3E,$77
	.dc.b	$FC,$3F,$BF,$E5,$9B,$9F,$CB,$6D,$F4,$3F,$73,$F4,$1F,$91,$7E,$4D
	.dc.b	$A3,$17,$FA,$06,$DF,$D5,$7F,$CC,$2D,$E9,$FF,$2F,$C6,$AF,$FB,$8B
	.dc.b	$6F,$A7,$C1,$7E,$C3,$CF,$95,$F5,$1F,$D9,$FC,$CE,$23,$FD,$83,$C1
	.dc.b	$FE,$01,$CF,$FD,$5A,$DC,$FE,$AD,$7F,$9B,$BE,$91,$E3,$7F,$8E,$9B
	.dc.b	$87,$E5,$9A,$FF,$07,$CD,$18,$8F,$E7,$51,$8D,$BA,$12,$3F,$8F,$CF
	.dc.b	$2C,$BF,$81,$BF,$46,$3F,$DB,$BF,$54,$FA,$4F,$F4,$C3,$63,$8D,$F3
	.dc.b	$7F,$DB,$B1,$96,$CE,$6F,$AF,$D5,$7A,$C9,$3E,$B7,$DD,$D5,$F0,$49
	.dc.b	$1E,$68,$FF,$46,$7C,$77,$FA,$97,$E1,$8D,$24,$FC,$5F,$84,$F9,$44
	.dc.b	$7F,$DC,$31,$5F,$C2,$E9,$FE,$71,$C7,$E0,$D7,$69,$F8,$AE,$E3,$31
	.dc.b	$C8,$F9,$3F,$37,$F5,$48,$FD,$57,$EC,$7C,$6F,$DB,$24,$70,$3F,$DE
	.dc.b	$7F,$15,$FE,$E8,$DF,$FD,$F5,$9F,$FA,$EF,$E1,$7E,$97,$C1,$23,$13
	.dc.b	$AB,$E0,$27,$57,$86,$C7,$F8,$72,$F9,$DF,$28,$E0,$FF,$BD,$AF,$C3
	.dc.b	$7C,$63,$57,$CA,$BF,$DE,$F5,$E1,$7A,$3F,$E2,$88,$7D,$57,$FB,$EA
	.dc.b	$72,$79,$5C,$3C,$8C,$7C,$87,$9E,$3F,$C9,$18,$FC,$07,$6D,$8D,$1B
	.dc.b	$F0,$ED,$76,$BF,$0C,$E2,$3B,$47,$92,$71,$5C,$E7,$85,$F6,$C3,$D4
	.dc.b	$F8,$9F,$67,$A2,$F9,$5F,$98,$BD,$0A,$62,$25,$F8,$E7,$4D,$9C,$BE
	.dc.b	$48,$E8,$E4,$86,$67,$EE,$AF,$2D,$F2,$AF,$A5,$F9,$B3,$98,$C3,$F3
	.dc.b	$4C,$47,$D2,$F6,$9F,$01,$FE,$57,$9E,$37,$D2,$AD,$F2,$D7,$F8,$15
	.dc.b	$72,$3E,$C6,$1C,$BD,$CF,$2E,$1C,$67,$FC,$30,$6C,$FF,$0D,$39,$6D
	.dc.b	$F9,$9C,$FD,$0B,$93,$8C,$7B,$E3,$0F,$4A,$46,$23,$8C,$F0,$9C,$BE
	.dc.b	$C9,$E3,$8F,$F1,$BC,$7F,$CD,$3E,$3F,$FA,$8A,$3D,$B7,$C3,$0F,$90
	.dc.b	$D8,$CD,$F2,$E0,$FF,$42,$E1,$CB,$C3,$ED,$97,$E4,$7E,$A7,$FD,$CF
	.dc.b	$1F,$93,$FA,$AF,$AC,$E7,$E3,$3C,$18,$7E,$9E,$73,$A7,$E5,$AF,$8B
	.dc.b	$F9,$88,$CE,$7F,$C0,$A7,$2F,$F0,$48,$CD,$7E,$C6,$72,$F5,$E0,$EF
	.dc.b	$7F,$C1,$67,$2F,$F0,$7A,$E2,$FF,$09,$E3,$7C,$07,$0E,$84,$63,$4F
	.dc.b	$CE,$7D,$4E,$63,$17,$F9,$BB,$C1,$99,$B5,$FA,$38,$B3,$FA,$19,$BA
	.dc.b	$7F,$1E,$BE,$6D,$AE,$57,$C4,$FD,$F7,$D5,$7F,$95,$1B,$C7,$E7,$8F
	.dc.b	$D2,$C6,$09,$8F,$F0,$28,$E0,$DE,$4E,$15,$2B,$78,$7F,$30,$BE,$77
	.dc.b	$C6,$BE,$5F,$8F,$FD,$BB,$B3,$FD,$D7,$E0,$7F,$68,$6E,$3F,$E3,$FF
	.dc.b	$0B,$FD,$D6,$FF,$6E,$8C,$E1,$FE,$FA,$9C,$BF,$DE,$5A,$FF,$78,$6B
	.dc.b	$CF,$BB,$FD,$7A,$F5,$F4,$B9,$E3,$FE,$00,$7A,$F0,$B4,$FD,$3E,$5F
	.dc.b	$DF,$A7,$78,$FF,$92,$A3,$07,$FA,$9A,$B8,$74,$EA,$A4,$EB,$1F,$F2
	.dc.b	$9D,$6A,$31,$A1,$FE,$85,$8F,$82,$BF,$D3,$55,$B6,$C6,$5F,$F6,$74
	.dc.b	$65,$F8,$1E,$E7,$98,$F3,$FA,$76,$FF,$C2,$EF,$C2,$9F,$94,$DD,$3F
	.dc.b	$C4,$2D,$D3,$FC,$C9,$5F,$E2,$7F,$3B,$E2,$F2,$BE,$9B,$E1,$C6,$1B
	.dc.b	$3F,$0E,$74,$FF,$89,$A7,$4D,$26,$3F,$E0,$AF,$3B,$8B,$63,$FC,$6D
	.dc.b	$1E,$32,$3F,$E3,$9F,$33,$E9,$9F,$53,$EA,$AC,$7F,$8D,$22,$D7,$F8
	.dc.b	$5A,$6F,$07,$F8,$DB,$B3,$ED,$9F,$73,$E3,$AE,$1F,$E3,$C8,$B7,$FE
	.dc.b	$1E,$9B,$5F,$E6,$1E,$9F,$C4,$FA,$7F,$9E,$8F,$B5,$4F,$93,$FF,$4E
	.dc.b	$61,$AB,$FC,$CC,$63,$1F,$E7,$ED,$AF,$F6,$C3,$51,$1C,$36,$FF,$90
	.dc.b	$BA,$3C,$0E,$1D,$5F,$DF,$9E,$30,$E2,$FF,$43,$7B,$5F,$A2,$F0,$7F
	.dc.b	$D8,$58,$FE,$6F,$EA,$BF,$7F,$97,$FA,$26,$B4,$49,$D7,$6D,$C7,$FD
	.dc.b	$0F,$83,$DB,$E9,$F5,$5D,$5B,$7E,$CC,$7F,$D2,$91,$8B,$E2,$D8,$FF
	.dc.b	$76,$FB,$8F,$43,$F6,$ED,$BE,$47,$C5,$F9,$8F,$C7,$E1,$FE,$98,$9D
	.dc.b	$3F,$EC,$2D,$D3,$A8,$F9,$11,$8E,$9F,$8D,$39,$9E,$91,$82,$5F,$D8
	.dc.b	$9D,$4F,$DF,$6F,$03,$EA,$9F,$3D,$FC,$39,$7C,$7D,$8C,$37,$0F,$5E
	.dc.b	$1F,$56,$B7,$18,$BE,$97,$C2,$31,$EB,$BD,$0F,$E3,$6D,$8F,$E2,$9F
	.dc.b	$07,$F1,$38,$3F,$2B,$87,$FA,$02,$F6,$4A,$DE,$B7,$3F,$F6,$0F,$C2
	.dc.b	$BE,$07,$F9,$3B,$FC,$57,$1E,$6E,$6F,$CC,$E5,$FC,$C8,$FA,$A7,$C2
	.dc.b	$7F,$11,$8C,$36,$33,$1F,$E2,$23,$25,$73,$87,$E7,$CF,$E5,$0E,$1F
	.dc.b	$D1,$47,$07,$4D,$7F,$D7,$7F,$54,$F0,$FF,$A0,$AF,$0E,$C6,$1B,$5F
	.dc.b	$46,$74,$D9,$D3,$F5,$7E,$AB,$FD,$56,$DE,$1F,$E9,$B0,$FA,$6B,$12
	.dc.b	$74,$71,$8C,$63,$FE,$F0,$F8,$3F,$F3,$44,$FE,$05,$B1,$F9,$BE,$77
	.dc.b	$C6,$FF,$5A,$9C,$F0,$7F,$12,$E9,$FC,$FC,$FE,$70,$98,$7E,$65,$B8
	.dc.b	$7F,$10,$C3,$9C,$E4,$91,$FE,$E5,$FA,$A7,$C0,$FF,$68,$7F,$5B,$8F
	.dc.b	$E3,$1C,$D7,$E5,$DC,$BF,$C6,$B9,$DF,$22,$F9,$DF,$EA,$0D,$9F,$F3
	.dc.b	$0C,$EA,$72,$6B,$1F,$EE,$F5,$F8,$C5,$87,$F3,$0F,$75,$F5,$4F,$9D
	.dc.b	$DD,$7F,$7E,$F9,$57,$0D,$3F,$C1,$76,$F9,$6F,$15,$74,$D7,$D1,$7C
	.dc.b	$1F,$DA,$D6,$F4,$FE,$D4,$D8,$FF,$3C,$B6,$BF,$2C,$D7,$F9,$B3,$5F
	.dc.b	$F3,$1D,$F8,$1C,$B5,$CD,$7F,$75,$F9,$87,$D5,$7F,$86,$1B,$87,$F0
	.dc.b	$5F,$85,$7D,$56,$6B,$FC,$3B,$18,$9F,$9A,$DC,$3F,$55,$C3,$FE,$9C
	.dc.b	$BF,$F0,$14,$E6,$7F,$A6,$4E,$6B,$8D,$68,$8F,$D8,$F9,$BF,$E3,$EF
	.dc.b	$F1,$76,$D7,$CB,$3F,$A8,$7D,$57,$E8,$E7,$5E,$37,$F8,$CA,$31,$3E
	.dc.b	$46,$BE,$26,$BF,$8B,$6B,$D3,$AB,$FC,$35,$18,$C4,$D3,$1F,$3A,$F5
	.dc.b	$7F,$C8,$2D,$C7,$F9,$74,$7E,$25,$FE,$61,$19,$C3,$D5,$7C,$79,$3F
	.dc.b	$C5,$17,$B9,$F9,$6D,$7E,$4B,$5F,$D1,$35,$FE,$E4,$D9,$E9,$7F,$A2
	.dc.b	$8C,$7D,$57,$F1,$0A,$F0,$B7,$9F,$F9,$DF,$BB,$FC,$02,$FF,$5D,$39
	.dc.b	$7F,$61,$1F,$C2,$2B,$ED,$B5,$CF,$C5,$FF,$90,$63,$11,$FE,$05,$6E
	.dc.b	$26,$92,$7F,$C9,$1F,$15,$FE,$72,$C7,$FA,$6B,$70,$FE,$01,$F4,$4F
	.dc.b	$95,$FF,$37,$CE,$72,$FF,$09,$4E,$6B,$F8,$0B,$63,$F2,$AD,$7F,$26
	.dc.b	$D7,$AD,$7F,$96,$63,$11,$D2,$30,$7F,$97,$7E,$1F,$F8,$46,$1F,$D8
	.dc.b	$9B,$7F,$C5,$4F,$F1,$D1,$FC,$5D,$CD,$7F,$89,$E3,$FD,$36,$FE,$91
	.dc.b	$AF,$1F,$6F,$FC,$D5,$18,$8F,$F1,$77,$D4,$FF,$24,$7F,$12,$DB,$FE
	.dc.b	$31,$F5,$5D,$B6,$4E,$6F,$F9,$43,$6F,$9F,$87,$FC,$62,$F8,$4E,$6F
	.dc.b	$F9,$1F,$37,$FA,$73,$1F,$D8,$9F,$E5,$4E,$07,$C9,$7F,$1E,$8F,$F0
	.dc.b	$19,$AC,$FF,$9B,$37,$3F,$F2,$33,$6B,$8F,$6F,$5C,$3F,$D4,$FF,$4E
	.dc.b	$FA,$7A,$7F,$AA,$BE,$3B,$FB,$1E,$3F,$EA,$09,$D6,$5F,$EA,$76,$C6
	.dc.b	$6B,$87,$97,$CA,$BF,$D5,$97,$CA,$BF,$D3,$ED,$7F,$CC,$B9,$7F,$BB
	.dc.b	$AB,$FD,$38,$DA,$FF,$03,$4F,$E6,$7E,$A5,$F2,$9F,$EB,$E6,$CF,$FA
	.dc.b	$E7,$67,$C3,$E6,$A4,$FF,$AF,$6B,$77,$FE,$BA,$BD,$CE,$6B,$FD,$AB
	.dc.b	$3F,$E7,$3E,$5F,$0B,$4A,$1A,$D3,$CF,$F5,$4F,$90,$FF,$2B,$C6,$1A
	.dc.b	$FF,$88,$63,$5E,$77,$66,$C7,$2F,$7F,$FB,$2B,$79,$35,$18,$8E,$4D
	.dc.b	$7E,$5F,$E6,$6F,$82,$F4,$F0,$23,$13,$F3,$23,$13,$A9,$CC,$FF,$B6
	.dc.b	$1A,$FF,$B5,$9A,$F6,$8E,$07,$C6,$F8,$77,$E5,$AE,$DC,$36,$BE,$DC
	.dc.b	$EA,$72,$78,$7E,$FE,$4F,$97,$04,$62,$3C,$8D,$7C,$8D,$7F,$C2,$DA
	.dc.b	$FE,$09,$FF,$09,$B5,$FE,$CE,$D7,$FB,$2B,$5E,$3C,$9F,$80,$EB,$FF
	.dc.b	$47,$B6,$75,$3F,$EB,$16,$BF,$E9,$DF,$BD,$FC,$2E,$C6,$92,$3F,$17
	.dc.b	$19,$6C,$65,$FC,$FC,$65,$FD,$0F,$C2,$FD,$07,$FD,$96,$D4,$D3,$54
	.dc.b	$FC,$03,$FD,$75,$3F,$AA,$8C,$7C,$3F,$E7,$1B,$19,$9D,$BE,$09,$FC
	.dc.b	$FD,$62,$3F,$CF,$0D,$7F,$DB,$0D,$7C,$B1,$C0,$8F,$EA,$A3,$55,$E5
	.dc.b	$FB,$6F,$15,$9A,$4A,$FF,$19,$FE,$AC,$FA,$6F,$D9,$FA,$C8,$C4,$EA
	.dc.b	$73,$18,$8F,$F1,$CB,$5F,$F1,$CB,$5F,$F7,$AC,$F3,$9E,$27,$E0,$79
	.dc.b	$7F,$23,$3F,$E5,$C6,$BF,$E3,$06,$BF,$DD,$38,$3C,$BF,$B3,$3E,$4A
	.dc.b	$92,$7F,$C9,$B1,$96,$C6,$5F,$F2,$7C,$65,$B1,$F9,$3F,$3B,$F2,$3C
	.dc.b	$7F,$E2,$1E,$3C,$12,$BF,$CB,$3F,$98,$3E,$57,$E3,$B6,$33,$18,$23
	.dc.b	$82,$1F,$E6,$08,$E0,$FF,$93,$BD,$2F,$F5,$73,$FE,$7E,$6B,$9F,$85
	.dc.b	$FD,$44,$62,$3F,$CC,$FF,$54,$F9,$AF,$B2,$D8,$FF,$37,$B5,$FF,$37
	.dc.b	$B5,$E3,$1C,$08,$F0,$7F,$6D,$ED,$3E,$6A,$92,$7F,$CC,$EF,$E2,$3E
	.dc.b	$6D,$F5,$34,$93,$FE,$80,$8C,$B6,$32,$FF,$A0,$A3,$2D,$8F,$C9,$FC
	.dc.b	$03,$CC,$A4,$9F,$F4,$5C,$65,$FF,$46,$FC,$1D,$7F,$D1,$AD,$7F,$D1
	.dc.b	$AD,$7F,$2A,$D5,$D1,$FE,$F2,$C7,$F4,$7F,$19,$F0,$67,$FD,$32,$D7
	.dc.b	$FD,$32,$D7,$FD,$32,$FD,$C6,$F0,$F2,$23,$FB,$7E,$BF,$E0,$24,$70
	.dc.b	$43,$F8,$AB,$96,$9F,$BF,$FB,$AF,$E1,$24,$70,$7F,$B1,$BE,$B5,$F2
	.dc.b	$3F,$CB,$DB,$19,$8C,$11,$C1,$0F,$F5,$0C,$70,$7F,$D4,$7F,$22,$FB
	.dc.b	$1F,$E7,$4D,$8F,$F6,$AB,$5D,$C6,$88,$E9,$A8,$83,$EF,$9F,$8C,$38
	.dc.b	$FF,$C0,$F1,$86,$C6,$67,$FC,$37,$39,$6C,$65,$EA,$75,$70,$F8,$23
	.dc.b	$83,$A6,$BE,$4F,$01,$F8,$93,$E4,$BF,$C0,$2D,$9F,$F0,$03,$5E,$F3
	.dc.b	$C2,$7F,$B0,$7C,$AF,$F9,$0E,$7F,$DF,$6D,$7F,$D9,$ED,$7F,$D9,$AD
	.dc.b	$7E,$37,$7A,$31,$08,$C7,$B5,$D8,$91,$FE,$3E,$87,$D3,$BE,$DF,$FB
	.dc.b	$93,$67,$4D,$7F,$DC,$2D,$7F,$DB,$ED,$7C,$6D,$7C,$4D,$9E,$F3,$FD
	.dc.b	$7B,$AA,$57,$F5,$76,$BF,$DC,$9A,$FF,$62,$6B,$F8,$5F,$85,$7E,$6E
	.dc.b	$F3,$FF,$C9,$31,$86,$C6,$67,$FC,$BB,$39,$7E,$17,$9C,$FA,$4F,$F0
	.dc.b	$93,$63,$31,$82,$38,$21,$F8,$AC,$BF,$C0,$B1,$F8,$2F,$AB,$93,$F8
	.dc.b	$26,$BF,$82,$E1,$46,$23,$F1,$91,$87,$81,$FE,$64,$F9,$DF,$EA,$CF
	.dc.b	$E6,$1A,$FE,$99,$AF,$EA,$5A,$9A,$4F,$2F,$FC,$DA,$F0,$E3,$7D,$43
	.dc.b	$F1,$7F,$F2,$DB,$6F,$FC,$B4,$D7,$F3,$6D,$7F,$6F,$9B,$E9,$7F,$C5
	.dc.b	$91,$86,$BF,$98,$6B,$F9,$76,$C7,$1F,$5B,$FC,$89,$18,$23,$FE,$47
	.dc.b	$FA,$77,$CC,$52,$5F,$E7,$63,$2D,$8C,$BF,$5E,$32,$D8,$FF,$0F,$FD
	.dc.b	$73,$E5,$E9,$27,$FD,$4B,$E4,$27,$51,$FE,$7D,$FB,$0F,$94,$4E,$0F
	.dc.b	$F4,$1B,$F5,$29,$F8,$EF,$FA,$71,$B5,$FE,$2E,$8C,$11,$C1,$0F,$0B
	.dc.b	$AF,$9B,$E0,$48,$CB,$F1,$7E,$8E,$47,$C3,$6B,$FC,$4B,$F5,$BF,$EA
	.dc.b	$CF,$FB,$11,$AB,$8D,$18,$6F,$9B,$F4,$7E,$72,$57,$CF,$6B,$F3,$5A
	.dc.b	$FE,$13,$81,$18,$8F,$E3,$31,$87,$91,$0E,$47,$FA,$A6,$FF,$C3,$CF
	.dc.b	$9F,$D6,$FF,$57,$C6,$23,$D0,$DE,$45,$7E,$56,$7B,$F3,$7E,$24,$8F
	.dc.b	$FB,$43,$C1,$FF,$55,$7C,$1F,$F5,$1F,$89,$7D,$3F,$F5,$29,$CD,$7F
	.dc.b	$B4,$E3,$1F,$0A,$BE,$11,$FE,$5E,$F5,$DC,$3F,$ED,$D8,$DC,$EA,$FF
	.dc.b	$AF,$61,$D3,$1E,$F1,$8C,$7A,$3D,$CF,$F0,$9E,$3D,$1E,$47,$F3,$A8
	.dc.b	$DC,$F2,$8C,$5E,$AF,$B6,$1E,$5C,$7C,$95,$DE,$B5,$87,$2A,$F3,$CF
	.dc.b	$87,$91,$97,$86,$3C,$87,$62,$E3,$D9,$E8,$71,$8E,$19,$79,$67,$53
	.dc.b	$C6,$31,$87,$86,$3D,$F9,$79,$9D,$63,$E4,$AD,$E5,$F2,$F2,$F2,$DE
	.dc.b	$F1,$EA,$4C,$7A,$4E,$B1,$F7,$DC,$7D,$33,$AC,$BC,$F9,$77,$24,$FB
	.dc.b	$32,$FA,$F8,$73,$8E,$38,$73,$9F,$2C,$7C,$1C,$BC,$98,$FB,$63,$D6
	.dc.b	$73,$C3,$EA,$3E,$33,$FC,$2B,$97,$E4,$1B,$3F,$84,$6E,$1F,$99,$F6
	.dc.b	$1E,$1D,$24,$FF,$87,$8F,$EF,$F1,$A7,$17,$FE,$87,$9F,$C3,$4E,$72
	.dc.b	$F6,$FD,$BB,$EA,$5F,$1D,$AB,$98,$FE,$03,$0E,$1A,$7F,$AF,$B0,$E1
	.dc.b	$7A,$C7,$F0,$15,$AB,$FD,$0B,$76,$5F,$09,$E9,$8F,$1C,$7F,$65,$7B
	.dc.b	$AD,$61,$D2,$B5,$7D,$1F,$C2,$4E,$AF,$8C,$6A,$7E,$1B,$AC,$BD,$B1
	.dc.b	$E0,$25,$F5,$7F,$45,$7C,$DE,$C6,$AF,$AB,$AA,$E7,$5C,$49,$3A,$7E
	.dc.b	$6C,$6A,$30,$49,$EB,$3A,$CB,$D1,$9F,$D7,$9F,$01,$FE,$96,$C3,$C8
	.dc.b	$FB,$0C,$96,$B9,$3C,$38,$9F,$8D,$7F,$9B,$F0,$FF,$22,$36,$7F,$C3
	.dc.b	$8D,$7F,$BE,$B5,$FE,$91,$C1,$F2,$1F,$F3,$27,$27,$F2,$0A,$FE,$71
	.dc.b	$39,$9D,$C7,$6D,$71,$1F,$E5,$7B,$FC,$CC,$E4,$91,$E2,$9D,$E1,$E8
	.dc.b	$6D,$70,$9D,$57,$07,$CF,$3A,$24,$7E,$8F,$A1,$FB,$2F,$1A,$FF,$3E
	.dc.b	$35,$FF,$49,$46,$AB,$CE,$DA,$F8,$FD,$5E,$22,$38,$3F,$BB,$FD,$07
	.dc.b	$BB,$8F,$FC,$02,$DE,$C3,$4C,$7B,$32,$BF,$BF,$B7,$2F,$F4,$BD,$FF
	.dc.b	$82,$A3,$11,$FE,$A6,$8C,$4F,$03,$ED,$F9,$BF,$18,$8E,$0F,$F0,$5F
	.dc.b	$C2,$F8,$C8,$E0,$FF,$15,$76,$FD,$0E,$D3,$49,$1A,$26,$1F,$E2,$D8
	.dc.b	$CB,$FE,$65,$F1,$9F,$65,$F6,$9B,$5F,$E9,$36,$CE,$E7,$44,$9D,$36
	.dc.b	$3F,$D5,$7F,$30,$EC,$FA,$AD,$AF,$BC,$D9,$DC,$FE,$41,$FF,$54,$7B
	.dc.b	$B7,$D4,$34,$97,$FC,$A6,$76,$D9,$DB,$67,$73,$FE,$BE,$F9,$87,$BF
	.dc.b	$F8,$86,$D7,$E4,$5B,$3B,$9D,$12,$74,$D8,$FF,$47,$77,$B7,$FD,$1B
	.dc.b	$6B,$FD,$78,$D9,$DC,$E8,$93,$A6,$C7,$FB,$53,$E3,$BF,$CC,$13,$FA
	.dc.b	$56,$C6,$7C,$BF,$D0,$C6,$23,$FD,$D3,$9F,$F9,$8E,$BE,$27,$97,$F2
	.dc.b	$C8,$E0,$93,$FD,$4E,$78,$36,$7F,$7B,$E7,$BF,$08,$A4,$BF,$F6,$C4
	.dc.b	$65,$B1,$9A,$FF,$59,$4E,$5B,$1F,$C2,$39,$5F,$61,$49,$7F,$EE,$38
	.dc.b	$CB,$63,$35,$FE,$BD,$9C,$B6,$3E,$E7,$4C,$AF,$B8,$D9,$10,$7B,$6F
	.dc.b	$85,$7D,$07,$F2,$28,$C3,$63,$37,$FD,$E6,$72,$D8,$CD,$75,$3A,$B8
	.dc.b	$7C,$13,$C1,$D3,$5F,$F5,$BF,$D5,$FE,$69,$FF,$6D,$35,$CF,$2F,$F3
	.dc.b	$08,$C4,$7B,$3C,$44,$62,$11,$89,$E0,$4C,$3F,$C2,$50,$FA,$B7,$83
	.dc.b	$FC,$F1,$B3,$A6,$DF,$FB,$B5,$AF,$FB,$39,$AF,$8D,$AF,$89,$B5,$DE
	.dc.b	$BF,$7D,$CD,$2F,$F7,$4D,$BF,$CA,$7A,$9F,$C0,$C9,$3F,$BF,$6C,$6A
	.dc.b	$30,$47,$F9,$AF,$EA,$FE,$2F,$04,$D2,$61,$FD,$0D,$FE,$83,$5F,$EC
	.dc.b	$C6,$D7,$2E,$0D,$1F,$D2,$7E,$21,$DD,$49,$7F,$ED,$48,$CB,$FE,$D6
	.dc.b	$F8,$97,$CD,$7F,$51,$6C,$66,$30,$7F,$55,$FD,$45,$E9,$D2,$4F,$FB
	.dc.b	$82,$32,$D8,$CE,$5F,$D5,$27,$2D,$8F,$EE,$FE,$A7,$F8,$0E,$3F,$2D
	.dc.b	$C3,$FE,$76,$BF,$C1,$CE,$67,$F4,$D3,$97,$FA,$FF,$EA,$5F,$EA,$DA
	.dc.b	$FF,$78,$B6,$33,$EA,$FF,$6D,$8C,$47,$F6,$36,$BF,$B2,$E1,$46,$21
	.dc.b	$18,$9E,$47,$F8,$3B,$EA,$1D,$94,$98,$7E,$6A,$32,$FF,$9B,$69,$F9
	.dc.b	$83,$E6,$FF,$C0,$0D,$9D,$37,$1F,$EF,$6D,$7F,$BD,$35,$F4,$35,$F3
	.dc.b	$B6,$BB,$D7,$F8,$93,$E5,$1F,$67,$49,$87,$EC,$23,$2F,$9F,$F6,$67
	.dc.b	$C4,$FF,$85,$5B,$19,$8C,$11,$C1,$32,$FF,$09,$CF,$07,$EE,$FA,$6F
	.dc.b	$96,$FF,$1B,$B6,$BF,$80,$B6,$77,$3F,$E2,$5F,$D9,$DF,$03,$49,$7F
	.dc.b	$36,$32,$D8,$CD,$7F,$89,$27,$2D,$8F,$F2,$6F,$CE,$3F,$1A,$F3,$4F
	.dc.b	$F3,$38,$C6,$7A,$9C,$CF,$F3,$89,$CE,$1F,$E0,$1F,$C0,$3F,$C2,$B3
	.dc.b	$FC,$BD,$AF,$F3,$D6,$BF,$D0,$1A,$9A,$4F,$53,$FC,$24,$F0,$AF,$37
	.dc.b	$07,$8C,$97,$FE,$B9,$8C,$BE,$3F,$96,$E2,$47,$07,$E4,$FF,$44,$7C
	.dc.b	$37,$F9,$45,$B1,$98,$C1,$1C,$13,$2F,$F2,$7C,$F0,$7F,$73,$F4,$5F
	.dc.b	$DB,$5F,$DB,$B5,$FE,$88,$D7,$FA,$1B,$7A,$CD,$27,$9F,$FE,$7C,$78
	.dc.b	$71,$BD,$3F,$C3,$5F,$F6,$98,$C7,$27,$F8,$48,$99,$7F,$85,$DB,$3F
	.dc.b	$E1,$B6,$BF,$DC,$7B,$1F,$9C,$7F,$01,$69,$FE,$92,$2D,$7F,$B9,$DB
	.dc.b	$3F,$E2,$5E,$4F,$F1,$09,$27,$FC,$4C,$D8,$D4,$60,$8F,$F9,$24,$7C
	.dc.b	$25,$69,$97,$07,$F9,$3F,$E9,$1F,$4A,$E3,$C3,$6B,$FD,$F9,$3A,$9C
	.dc.b	$9C,$7E,$53,$E7,$CE,$9A,$FF,$81,$E5,$18,$8F,$F0,$B4,$63,$7F,$85
	.dc.b	$F9,$1F,$F8,$A2,$7F,$C3,$71,$FE,$E6,$D1,$39,$9F,$BB,$9F,$F8,$82
	.dc.b	$73,$EB,$7F,$88,$6F,$FD,$DB,$C1,$F7,$79,$F1,$7B,$C7,$95,$75,$6C
	.dc.b	$66,$F7,$8B,$0E,$4F,$92,$7F,$D6,$FC,$5F,$E6,$58,$FF,$2D,$7B,$7F
	.dc.b	$E7,$9A,$FC,$63,$67,$53,$FE,$29,$CD,$F7,$EF,$8F,$EF,$18,$6C,$66
	.dc.b	$7C,$07,$8E,$B9,$E9,$F1,$67,$53,$F1,$05,$F3,$78,$B5,$D6,$3E,$67
	.dc.b	$AF,$86,$FD,$11,$F4,$1F,$0A,$30,$D8,$CC,$F9,$3A,$BF,$11,$D3,$9A
	.dc.b	$93,$D2,$B6,$F9,$6B,$6F,$B3,$07,$D9,$FF,$AE,$23,$0D,$7C,$ED,$D1
	.dc.b	$C1,$FE,$5F,$E5,$7C,$9F,$D8,$C3,$9F,$83,$70,$E1,$7F,$EE,$E6,$E3
	.dc.b	$FE,$68,$C3,$8B,$F3,$BA,$1F,$4E,$FA,$8F,$F5,$A4,$61,$B1,$99,$FE
	.dc.b	$50,$D7,$FD,$D3,$F8,$97,$D2,$7E,$6B,$5F,$AD,$C4,$8C,$6D,$EC,$9F
	.dc.b	$D9,$57,$38,$C6,$1E,$4F,$8B,$F0,$92,$BC,$2D,$7F,$CE,$D1,$B7,$24
	.dc.b	$3F,$DF,$BF,$64,$FA,$9F,$BC,$D8,$FC,$1B,$5F,$C6,$4F,$37,$ED,$70
	.dc.b	$7E,$71,$FF,$15,$B5,$FF,$3C,$35,$FF,$3B,$B7,$8B,$F5,$9D,$5F,$ED
	.dc.b	$F6,$CF,$F7,$88,$C4,$F1,$D5,$F5,$4F,$BD,$FC,$A3,$6F,$F3,$2D,$7C
	.dc.b	$B7,$FB,$AC,$1E,$77,$F8,$1E,$7F,$C2,$DD,$3E,$18,$C4,$7F,$92,$1A
	.dc.b	$FF,$91,$9A,$FF,$A4,$DA,$FF,$A4,$9B,$3D,$A7,$FD,$F5,$8F,$FB,$A7
	.dc.b	$B7,$F5,$58,$7E,$DB,$83,$EF,$4E,$71,$FD,$56,$3E,$29,$FC,$16,$2F
	.dc.b	$A9,$7C,$7F,$C7,$6D,$FF,$A7,$9A,$FF,$98,$1A,$FF,$97,$9A,$FF,$AA
	.dc.b	$DA,$FF,$AA,$9B,$3D,$E7,$D3,$F5,$5D,$2F,$CE,$DE,$9F,$F2,$F3,$70
	.dc.b	$E5,$5F,$3F,$D7,$FF,$16,$C7,$F9,$C3,$D0,$FB,$D1,$88,$FF,$3A,$B5
	.dc.b	$FF,$39,$B5,$FF,$60,$35,$FF,$5F,$B7,$13,$9D,$3F,$8B,$FC,$D7,$F1
	.dc.b	$7F,$00,$D3,$C9,$87,$F1,$BF,$73,$FC,$DF,$8F,$FA,$67,$D8,$FA,$F1
	.dc.b	$88,$FF,$6F,$60,$8C,$47,$FB,$B6,$31,$3C,$09,$1C,$CF,$F3,$B7,$D3
	.dc.b	$3F,$00,$FE,$54,$D8,$FE,$5E,$D7,$F8,$1D,$F8,$78,$3F,$0B,$F2,$1F
	.dc.b	$1D,$B9,$7F,$33,$8C,$70,$FF,$A3,$FE,$03,$83,$FE,$E7,$F8,$4F,$F1
	.dc.b	$9E,$3F,$EE,$26,$BF,$EA,$46,$BF,$EA,$26,$BF,$EE,$FF,$9B,$7A,$9F
	.dc.b	$A9,$6C,$7F,$3F,$6B,$FC,$6B,$07,$C4,$FF,$68,$87,$D1,$BE,$AF,$FC
	.dc.b	$28,$D8,$10,$8E,$5C,$1A,$70,$E1,$AF,$08,$EA,$47,$89,$DB,$F5,$57
	.dc.b	$E1,$5E,$56,$C7,$F6,$C0,$E1,$FE,$D4,$7A,$A7,$F8,$57,$D6,$FF,$48
	.dc.b	$E9,$34,$9E,$EE,$BC,$E9,$7F,$D9,$67,$2F,$F6,$7F,$AD,$FE,$59,$C9
	.dc.b	$FB,$E6,$F1,$7C,$4C,$51,$8A,$46,$3E,$B7,$F8,$E4,$E6,$37,$E5,$60
	.dc.b	$CC,$FF,$63,$FA,$77,$D6,$52,$61,$FE,$0C,$F9,$47,$D6,$BB,$33,$3F
	.dc.b	$E0,$6B,$E2,$FF,$77,$6B,$FE,$0F,$AE,$33,$FD,$C7,$E7,$3F,$BD,$CF
	.dc.b	$F7,$D6,$BF,$E1,$06,$ED,$FE,$36,$BE,$6F,$F2,$0F,$9B,$FC,$6B,$7A
	.dc.b	$BF,$9D,$46,$38,$3E,$32,$FA,$DC,$CC,$C7,$FC,$29,$3B,$7E,$4B,$B7
	.dc.b	$FC,$31,$F4,$FF,$B1,$C3,$F9,$7B,$70,$FA,$4D,$9F,$A2,$FF,$B6,$9E
	.dc.b	$3E,$0F,$F4,$D8,$D7,$11,$A6,$E7,$FC,$69,$F4,$6F,$AD,$FF,$31,$C7
	.dc.b	$E5,$A3,$1E,$CF,$E6,$21,$18,$75,$3C,$1F,$D2,$C6,$5B,$19,$7F,$59
	.dc.b	$19,$6C,$65,$FD,$87,$CF,$7F,$04,$CF,$FC,$21,$08,$C6,$7A,$F8,$AF
	.dc.b	$F3,$1E,$1D,$EF,$FA,$AF,$7F,$F0,$0C,$3F,$C8,$4D,$7F,$C7,$CD,$8D
	.dc.b	$FD,$6F,$F7,$38,$C6,$1F,$9F,$6D,$7E,$79,$BA,$FF,$0A,$AF,$F1,$E7
	.dc.b	$E4,$1F,$12,$BF,$C9,$0D,$7F,$CB,$6D,$C3,$F1,$9F,$5D,$C3,$29,$5C
	.dc.b	$1F,$D2,$47,$F9,$1F,$EB,$9F,$5B,$FE,$79,$6D,$7F,$0A,$F6,$91,$88
	.dc.b	$FE,$2B,$18,$9E,$04,$8E,$13,$B7,$FD,$98,$D7,$F9,$0F,$53,$F1,$BF
	.dc.b	$9F,$08,$C7,$07,$FA,$49,$B1,$FE,$67,$AF,$16,$7F,$E0,$6F,$A8,$7E
	.dc.b	$2F,$FE,$1F,$6D,$7F,$9D,$5A,$FF,$9C,$5F,$F2,$FB,$63,$81,$1E,$1F
	.dc.b	$0B,$FE,$29,$FA,$A7,$D1,$D2,$4F,$FA,$06,$1F,$47,$88,$FF,$3F,$47
	.dc.b	$07,$53,$CE,$7F,$CE,$DF,$61,$FE,$98,$9F,$F7,$83,$5F,$E8,$BC,$08
	.dc.b	$C4,$7F,$AC,$DB,$EB,$3E,$B3,$FA,$3E,$BF,$E2,$C6,$F4,$FF,$4E,$9D
	.dc.b	$5F,$83,$E0,$3C,$3F,$5F,$23,$FD,$25,$8A,$3F,$D4,$18,$72,$9F,$04
	.dc.b	$6D,$FF,$54,$CE,$DF,$F5,$4C,$EF,$A9,$B6,$67,$FA,$DC,$62,$FF,$CC
	.dc.b	$13,$99,$FE,$C5,$39,$7F,$AC,$E3,$ED,$FA,$9F,$19,$23,$E8,$5F,$D5
	.dc.b	$E5,$E6,$48,$FE,$C3,$7E,$28,$FF,$6A,$7D,$73,$EB,$7F,$D4,$CD,$AF
	.dc.b	$F5,$93,$5F,$B3,$3A,$99,$1A,$7C,$0D,$21,$FE,$CC,$9E,$5B,$FF,$96
	.dc.b	$B3,$FC,$EE,$1F,$13,$A7,$E0,$92,$FC,$F3,$FA,$57,$FD,$E1,$EA,$3E
	.dc.b	$F3,$FD,$78,$D8,$FF,$64,$35,$FC,$7D,$7F,$84,$7C,$47,$D2,$FF,$AC
	.dc.b	$7E,$6B,$F2,$55,$FC,$D6,$31,$D1,$FE,$11,$AD,$FC,$9F,$F8,$52,$FF
	.dc.b	$31,$C7,$FE,$4E,$E2,$FF,$0B,$47,$2A,$DE,$28,$C6,$3D,$23,$FC,$3B
	.dc.b	$3A,$9E,$7F,$07,$F8,$37,$78,$F7,$CB,$C5,$87,$1C,$3C,$D3,$95,$97
	.dc.b	$8D,$C1,$A3,$C5,$5C,$A7,$38,$F8,$63,$F4,$58,$F8,$5D,$57,$03,$CF
	.dc.b	$87,$86,$B8,$47,$82,$B8,$3A,$BE,$B5,$B3,$E0,$61,$E5,$9E,$87,$97
	.dc.b	$1F,$7B,$8C,$3A,$CE,$E7,$24,$C7,$CA,$FC,$8C,$BE,$8C,$60,$97,$E3
	.dc.b	$3D,$55,$CE,$33,$1E,$17,$71,$98,$E1,$5D,$DB,$3F,$1A,$BB,$C6,$08
	.dc.b	$73,$3E,$23,$A7,$E9,$1F,$E5,$8B,$FC,$2B,$70,$FF,$80,$0F,$D0,$35
	.dc.b	$F5,$FF,$D6,$5F,$C9,$35,$FC,$4B,$5F,$F0,$63,$64,$FF,$63,$C3,$F2
	.dc.b	$8D,$C3,$F9,$07,$BB,$FE,$A2,$EA,$FC,$1C,$E7,$0E,$BE,$26,$0F,$C7
	.dc.b	$68,$BE,$13,$F5,$EB,$83,$E2,$AE,$0F,$D0,$9D,$E3,$E2,$9C,$B6,$75
	.dc.b	$F0,$6C,$73,$97,$C0,$BB,$5F,$9E,$C3,$9C,$E6,$3C,$53,$AC,$BC,$D8
	.dc.b	$73,$9C,$93,$1F,$01,$FB,$AC,$39,$C7,$C0,$35,$87,$59,$D5,$73,$78
	.dc.b	$C6,$49,$3E,$59,$D4,$62,$3E,$21,$D0,$D6,$3E,$C7,$35,$D1,$C9,$B3
	.dc.b	$33,$DB,$0F,$6D,$7C,$4E,$DF,$F4,$CF,$AF,$EF,$E7,$F8,$11,$89,$FF
	.dc.b	$55,$C7,$C3,$E7,$7D,$87,$FA,$BA,$7F,$93,$DF,$17,$FD,$1D,$19,$CB
	.dc.b	$FA,$AB,$5F,$EA,$6D,$76,$7F,$91,$8F,$EE,$BC,$CF,$07,$A4,$7F,$38
	.dc.b	$6E,$1E,$2E,$0C,$E7,$FE,$B2,$9B,$7C,$71,$FE,$77,$1C,$31,$FF,$30
	.dc.b	$C6,$09,$8B,$3F,$83,$7C,$63,$13,$EC,$C7,$AE,$1C,$DF,$44,$FF,$AA
	.dc.b	$B0,$EA,$FC,$08,$CE,$5E,$7B,$FE,$E3,$DD,$FE,$F9,$7F,$BE,$F9,$DE
	.dc.b	$5F,$80,$F7,$EF,$FE,$11,$8C,$4F,$F7,$0C,$95,$D1,$B1,$98,$C7,$D5
	.dc.b	$FB,$CE,$78,$3E,$47,$15,$F7,$0F,$AF,$38,$D3,$DE,$7F,$44,$AE,$13
	.dc.b	$B8,$F7,$4F,$4A,$93,$9C,$FF,$0B,$F0,$D8,$BE,$71,$89,$F5,$4F,$A4
	.dc.b	$8F,$F7,$6F,$6B,$EE,$F8,$1F,$EC,$C8,$FF,$21,$7C,$46,$26,$47,$F9
	.dc.b	$15,$DC,$63,$0F,$F2,$97,$C0,$7E,$1B,$CB,$37,$E0,$30,$CC,$3F,$CB
	.dc.b	$4E,$5A,$7F,$B6,$7D,$FF,$F6,$C3,$7D,$DF,$C7,$FD,$5F,$F2,$68,$C1
	.dc.b	$F1,$7A,$91,$9C,$BF,$CF,$6D,$47,$F9,$49,$C3,$7C,$0F,$C5,$CF,$26
	.dc.b	$CE,$A5,$39,$AF,$F4,$54,$E6,$1E,$A3,$CC,$FD,$2B,$6B,$FD,$D1,$F0
	.dc.b	$B8,$99,$1F,$EA,$07,$F5,$7E,$FA,$73,$8F,$DA,$F0,$5D,$FF,$B2,$6E
	.dc.b	$1F,$81,$5F,$1B,$99,$93,$97,$84,$62,$BF,$1E,$3D,$DF,$9E,$E1,$BE
	.dc.b	$17,$F1,$4F,$99,$FF,$27,$61,$F9,$3A,$E1,$C3,$FC,$0A,$73,$1F,$EA
	.dc.b	$A6,$BF,$40,$F9,$2E,$1F,$F4,$1B,$5C,$FA,$1F,$4A,$31,$C0,$69,$BE
	.dc.b	$BF,$F6,$E7,$27,$F1,$4F,$70,$DF,$7F,$0C,$CF,$F4,$D3,$A6,$CE,$72
	.dc.b	$FD,$3F,$6B,$AF,$F8,$FB,$6B,$F5,$CB,$E6,$B3,$32,$72,$F0,$8C,$57
	.dc.b	$EF,$34,$7E,$6A,$F9,$5F,$E5,$2D,$C3,$F3,$5C,$EF,$0B,$35,$F8,$46
	.dc.b	$4F,$F0,$3F,$73,$F1,$DF,$61,$8C,$BB,$61,$C6,$B9,$DF,$F3,$AB,$DC
	.dc.b	$60,$FE,$11,$F6,$0F,$81,$FE,$7A,$D8,$CF,$C6,$62,$64,$70,$77,$18
	.dc.b	$C7,$F8,$7F,$D8,$7F,$32,$E9,$FE,$B0,$DB,$FD,$8F,$32,$3F,$94,$D7
	.dc.b	$0F,$3B,$FA,$6C,$6E,$13,$B9,$CF,$97,$A7,$FA,$AC,$BF,$3D,$FF,$6D
	.dc.b	$E1,$F9,$56,$C7,$E6,$FC,$0D,$C6,$32,$FF,$07,$4F,$E8,$67,$FA,$5B
	.dc.b	$6B,$F3,$0F,$E6,$7B,$B9,$B8,$6B,$99,$E3,$18,$8F,$E1,$ED,$7A,$1F
	.dc.b	$EC,$8F,$C0,$5F,$0B,$FD,$B1,$B7,$FC,$FF,$D1,$7D,$97,$C6,$CB,$FC
	.dc.b	$05,$5A,$C3,$F9,$EE,$AF,$77,$FA,$FB,$71,$FF,$63,$B6,$7F,$DA,$ED
	.dc.b	$7F,$69,$8F,$6E,$5F,$EC,$39,$7E,$26,$7F,$A9,$CB,$E4,$7F,$96,$E1
	.dc.b	$E5,$AD,$67,$FE,$37,$A7,$D8,$BD,$7F,$ED,$6D,$BF,$67,$CA,$3E,$CB
	.dc.b	$35,$FD,$C2,$31,$3F,$EE,$66,$E7,$F4,$B0,$46,$23,$D2,$D7,$F9,$EC
	.dc.b	$6A,$FF,$C0,$AD,$AE,$91,$8A,$E1,$5F,$EB,$2F,$96,$FE,$AF,$5F,$45
	.dc.b	$B1,$FD,$AB,$E0,$9F,$03,$FE,$06,$9C,$E9,$FE,$35,$9C,$DF,$FB,$D9
	.dc.b	$B7,$F8,$AE,$8F,$F6,$84,$E7,$0F,$F0,$AD,$6A,$BF,$93,$56,$9E,$11
	.dc.b	$97,$F5,$2D,$7F,$D9,$3F,$61,$FD,$02,$FF,$10,$D8,$FF,$00,$FD,$8B
	.dc.b	$C7,$FF,$0E,$CE,$72,$FF,$26,$4E,$6F,$F3,$F7,$FB,$08,$C3,$E9,$6B
	.dc.b	$E6,$6B,$FB,$C6,$BC,$6F,$FC,$63,$18,$C4,$EB,$67,$E0,$9F,$E5,$7C
	.dc.b	$7F,$14,$DC,$7F,$83,$FC,$E3,$EC,$BF,$C6,$B3,$9C,$FF,$82,$CE,$6B
	.dc.b	$D3,$92,$31,$3F,$45,$AF,$CE,$6B,$DA,$FF,$C8,$F1,$88,$EB,$18,$3F
	.dc.b	$17,$F6,$2F,$AC,$FF,$23,$B7,$0F,$F1,$97,$94,$FA,$0C,$D7,$F9,$A2
	.dc.b	$31,$3E,$66,$E1,$F8,$56,$BF,$82,$6B,$FC,$A1,$AF,$F2,$46,$E5,$FE
	.dc.b	$14,$BF,$F2,$FC,$6B,$57,$EB,$3E,$5B,$FD,$DB,$76,$FF,$57,$7D,$23
	.dc.b	$EC,$9F,$5F,$FE,$62,$AD,$69,$FE,$28,$AD,$4F,$F9,$C1,$B1,$F8,$86
	.dc.b	$E1,$FC,$A1,$AF,$F8,$05,$AF,$3B,$EF,$3C,$5B,$1C,$5A,$FF,$16,$FA
	.dc.b	$87,$CA,$7F,$9D,$9B,$5F,$E6,$5B,$F5,$7D,$76,$6B,$FD,$33,$18,$9E
	.dc.b	$D3,$E9,$C0,$D3,$FC,$83,$F7,$5E,$2D,$3F,$D6,$6D,$AF,$E1,$BF,$64
	.dc.b	$F9,$6F,$F5,$0C,$E7,$4F,$F2,$2B,$78,$0E,$78,$17,$C7,$0F,$F2,$C4
	.dc.b	$E7,$EB,$BF,$85,$C7,$F1,$A6,$E1,$FD,$8B,$27,$B3,$FE,$96,$9C,$E2
	.dc.b	$69,$C3,$93,$FD,$59,$3B,$F8,$6F,$F5,$DF,$CA,$BE,$ED,$F5,$F8,$6E
	.dc.b	$9F,$E6,$BF,$80,$FC,$AE,$9F,$EF,$3B,$E1,$7B,$25,$6F,$A6,$ED,$FE
	.dc.b	$1F,$C3,$F2,$FC,$3F,$E3,$99,$FF,$1E,$12,$BF,$87,$E1,$FE,$68,$F0
	.dc.b	$3C,$BF,$0F,$FC,$43,$3F,$F2,$87,$56,$26,$5F,$FA,$8E,$F8,$3E,$79
	.dc.b	$CE,$7F,$E5,$C3,$FC,$FF,$F2,$7F,$C7,$1F,$F2,$93,$5C,$95,$D7,$D9
	.dc.b	$7E,$F7,$AA,$F7,$1A,$4A,$FE,$4F,$2F,$A2,$FF,$6D,$46,$1A,$FF,$82
	.dc.b	$A3,$5A,$76,$9C,$CF,$FA,$53,$7F,$F4,$4F,$C8,$FF,$99,$E9,$A6,$77
	.dc.b	$FF,$65,$5F,$FA,$7D,$AF,$FB,$23,$67,$D9,$7F,$95,$63,$58,$73,$BE
	.dc.b	$4D,$8F,$F4,$DD,$F2,$C3,$EF,$5F,$19,$E9,$F6,$5F,$B0,$7E,$BB,$5F
	.dc.b	$F1,$2E,$28,$C4,$7F,$9F,$E3,$1E,$07,$D1,$CF,$FD,$4B,$F5,$5E,$16
	.dc.b	$D7,$AE,$3C,$95,$C2,$30,$43,$FC,$D5,$92,$77,$87,$8F,$2F,$8F,$5F
	.dc.b	$E4,$8D,$DF,$05,$EF,$EB,$FB,$4D,$8C,$CF,$F8,$BF,$B1,$E6,$E6,$A4
	.dc.b	$FF,$BB,$6B,$6F,$F8,$92,$B6,$FF,$BC,$7B,$BF,$6A,$DE,$EF,$04,$62
	.dc.b	$3F,$D0,$5D,$8F,$91,$CD,$49,$E0,$7F,$B6,$E7,$4F,$F9,$0F,$C5,$70
	.dc.b	$FE,$6F,$04,$62,$7F,$BC,$C6,$27,$53,$99,$FF,$09,$7E,$0A,$FC,$03
	.dc.b	$E2,$36,$3F,$DC,$BC,$1F,$2B,$81,$39,$9F,$F2,$1D,$F3,$AF,$F2,$07
	.dc.b	$C3,$7F,$AC,$27,$C2,$D7,$FC,$70,$D7,$FC,$9E,$D7,$FC,$6E,$D5,$97
	.dc.b	$FA,$97,$E2,$FD,$D3,$98,$C4,$FF,$7E,$F8,$1F,$17,$5F,$FA,$31,$B1
	.dc.b	$FE,$88,$6C,$F0,$8F,$F2,$DF,$D8,$BE,$CB,$E5,$36,$3E,$4B,$5F,$F2
	.dc.b	$83,$5F,$F3,$3B,$5F,$F2,$7B,$5F,$4B,$64,$E7,$5F,$F4,$DF,$8D,$F9
	.dc.b	$37,$F9,$95,$7D,$0C,$7F,$D4,$B1,$88,$F3,$E1,$FB,$7C,$3F,$DF,$BD
	.dc.b	$9F,$81,$F9,$0F,$D7,$D7,$F9,$81,$AF,$CA,$6C,$F0,$8F,$C2,$C6,$09
	.dc.b	$3F,$C9,$6D,$F8,$87,$D9,$DF,$F0,$4D,$CB,$FD,$72,$DC,$3B,$5F,$97
	.dc.b	$EC,$BF,$DD,$11,$F9,$26,$BF,$E7,$96,$BF,$E9,$06,$BF,$E7,$86,$BF
	.dc.b	$82,$6E,$46,$9A,$C5,$F4,$72,$BE,$51,$FE,$C8,$70,$B2,$FC,$E4,$61
	.dc.b	$FF,$4C,$57,$99,$FF,$4D,$76,$3E,$83,$F5,$CD,$BF,$D6,$B5,$FF,$4A
	.dc.b	$35,$FF,$55,$B5,$FF,$49,$B5,$FF,$47,$47,$43,$E8,$7E,$08,$DB,$F8
	.dc.b	$B3,$67,$D7,$92,$31,$1F,$EB,$4B,$F1,$61,$FE,$B1,$F9,$57,$D0,$7F
	.dc.b	$05,$6D,$7F,$04,$6B,$FE,$AB,$6B,$FE,$C2,$6B,$FE,$AA,$6B,$FB,$46
	.dc.b	$EA,$73,$8B,$F1,$07,$E0,$FF,$CB,$DB,$7E,$A6,$BF,$EC,$CA,$F1,$E3
	.dc.b	$FE,$C9,$F9,$A7,$DA,$FF,$37,$F8,$9F,$F1,$75,$FC,$26,$BF,$ED,$49
	.dc.b	$DB,$FE,$E7,$5F,$23,$FC,$95,$B1,$FC,$91,$AF,$FB,$2D,$AF,$FA,$C3
	.dc.b	$E9,$5F,$2D,$FD,$1D,$B1,$F4,$9A,$FE,$2F,$A0,$B6,$4F,$F3,$FC,$FF
	.dc.b	$A9,$CB,$ED,$5F,$3B,$FE,$11,$6C,$7F,$85,$5A,$FF,$AC,$9A,$F6,$3B
	.dc.b	$B8,$70,$D7,$31,$D4,$FE,$E1,$F3,$3E,$2A,$FE,$CB,$E7,$A3,$19,$7F
	.dc.b	$84,$5B,$7C,$8F,$C8,$FE,$59,$FC,$B3,$33,$4C,$7D,$B7,$F5,$69,$F8
	.dc.b	$87,$C3,$8D,$79,$1F,$DC,$5D,$F9,$7F,$DC,$FE,$35,$F6,$74,$97,$FA
	.dc.b	$08,$FE,$1D,$89,$DC,$66,$76,$F8,$27,$5E,$47,$F8,$37,$8B,$FC,$BB
	.dc.b	$3F,$C1,$36,$FB,$BE,$66,$4F,$F0,$97,$DA,$3E,$9B,$15,$BB,$E2,$FB
	.dc.b	$2F,$8B,$CE,$B8,$3B,$BF,$F0,$77,$DD,$BE,$92,$9D,$27,$CD,$F3,$5F
	.dc.b	$85,$E4,$FC,$4E,$A8,$FC,$6F,$2F,$F7,$16,$CE,$9E,$15,$F8,$F6,$D7
	.dc.b	$1D,$3F,$C3,$D1,$C2,$75,$E9,$7F,$88,$7E,$B3,$F5,$F8,$78,$1F,$F6
	.dc.b	$B4,$FF,$0E,$6B,$F2,$DD,$BF,$9D,$6E,$07,$78,$FF,$81,$DB,$5D,$7C
	.dc.b	$3F,$F2,$75,$BE,$93,$F7,$2F,$E6,$CC,$7C,$47,$F9,$CE,$33,$1F,$9F
	.dc.b	$6B,$FA,$E6,$BC,$0F,$F3,$E3,$83,$7C,$3F,$E8,$B8,$FF,$33,$7C,$1F
	.dc.b	$F9,$3F,$DC,$3E,$4A,$FE,$17,$9F,$F8,$7F,$ED,$3E,$DD,$FE,$2D,$B1
	.dc.b	$9E,$5F,$F2,$54,$7F,$0A,$8C,$4F,$F3,$68,$DE,$9F,$71,$B1,$FC,$15
	.dc.b	$E1,$1F,$E7,$BF,$53,$F2,$DF,$4D,$2B,$FC,$F9,$1E,$73,$0F,$E4,$78
	.dc.b	$D3,$98,$FE,$BF,$F4,$8F,$B4,$F8,$ED,$E0,$FF,$50,$36,$7E,$80,$70
	.dc.b	$FD,$36,$BF,$C3,$DA,$FF,$1E,$6B,$C6,$7E,$7F,$C0,$BE,$7B,$FC,$F8
	.dc.b	$D9,$FD,$83,$5F,$E0,$B5,$E0,$D7,$FC,$9D,$F5,$4F,$B3,$FF,$42,$72
	.dc.b	$FF,$88,$67,$F8,$BB,$77,$FD,$3D,$6F,$0F,$EC,$AD,$7F,$6D,$3B,$5C
	.dc.b	$BF,$D9,$3D,$AF,$F4,$15,$6F,$0F,$F2,$F7,$D6,$3E,$93,$FD,$29,$F8
	.dc.b	$77,$FA,$2F,$3F,$F5,$A3,$72,$FE,$7D,$5B,$7F,$BA,$B5,$FE,$0D,$1B
	.dc.b	$C8,$F1,$EB,$56,$F1,$FF,$4B,$FA,$8F,$AE,$FF,$6A,$37,$1F,$F3,$F4
	.dc.b	$60,$FF,$6C,$62,$FB,$37,$CF,$FF,$AA,$DB,$5F,$C7,$DA,$FF,$34,$BE
	.dc.b	$8E,$7E,$B3,$81,$1E,$15,$5F,$E4,$15,$FE,$C3,$C7,$FA,$4B,$75,$FF
	.dc.b	$6C,$3D,$A3,$D7,$97,$F2,$1A,$FF,$49,$B5,$FE,$01,$8F,$BF,$C3,$F6
	.dc.b	$9F,$E5,$1F,$B0,$FF,$5E,$4F,$F8,$81,$AF,$F3,$D6,$DF,$69,$FF,$65
	.dc.b	$7D,$9F,$E9,$2B,$F3,$BA,$A3,$11,$FC,$3E,$31,$D7,$FE,$D3,$9C,$CF
	.dc.b	$F6,$DE,$06,$3F,$EE,$3D,$7F,$9A,$B6,$7F,$BF,$37,$0E,$71,$E7,$F8
	.dc.b	$EF,$F6,$74,$EE,$F9,$C7,$9B,$1C,$B2,$B8,$1F,$D5,$71,$F3,$E7,$FD
	.dc.b	$DB,$D2,$FF,$2C,$CF,$F8,$07,$2F,$F0,$4F,$C0,$FF,$BA,$E7,$33,$FE
	.dc.b	$12,$9C,$BF,$BE,$CB,$D7,$3E,$42,$47,$F6,$37,$FC,$C5,$A7,$F8,$37
	.dc.b	$C6,$7D,$9F,$FB,$2F,$E4,$FF,$39,$18,$F4,$FF,$C3,$F1,$89,$D7,$4F
	.dc.b	$F8,$8B,$8D,$F4,$7F,$90,$AD,$5F,$0A,$ED,$7C,$33,$FF,$14,$E3,$D5
	.dc.b	$5F,$EB,$71,$E7,$CF,$FE,$21,$8E,$73,$AC,$39,$E3,$D6,$39,$E3,$D7
	.dc.b	$0E,$31,$93,$FA,$C4,$78,$5B,$39,$C7,$C5,$87,$87,$2F,$56,$3D,$AF
	.dc.b	$C0,$76,$BE,$18,$77,$AD,$CE,$4D,$9D,$8B,$87,$6B,$F9,$13,$C2,$75
	.dc.b	$87,$B8,$93,$AC,$BE,$13,$E3,$8C,$CF,$77,$D1,$3A,$8C,$4E,$C8,$FF
	.dc.b	$BB,$AB,$C1,$7E,$38,$D9,$32,$F9,$2F,$A3,$1F,$91,$3B,$8F,$3D,$F9
	.dc.b	$5F,$87,$94,$78,$3F,$56,$F9,$E5,$F6,$A7,$FC,$05,$7F,$81,$6E,$0B
	.dc.b	$FE,$AF,$1F,$24,$DF,$D9,$3C,$6C,$DA,$31,$3F,$D7,$A3,$13,$A9,$CC
	.dc.b	$F2,$8F,$C0,$47,$0C,$3F,$21,$EC,$FF,$B0,$76,$CD,$4B,$FB,$37,$BD
	.dc.b	$3F,$41,$7B,$9C,$D6,$F0,$FD,$4D,$EE,$5A,$7F,$A1,$BE,$11,$96,$F2
	.dc.b	$FD,$95,$EE,$FF,$C5,$CF,$1B,$E1,$5B,$9F,$D4,$CE,$F3,$F7,$C7,$E5
	.dc.b	$AB,$6F,$90,$98,$F9,$32,$FA,$99,$FA,$63,$13,$AC,$BC,$75,$AC,$FC
	.dc.b	$F3,$93,$C3,$1E,$3C,$3B,$DF,$22,$5F,$2A,$F2,$61,$DE,$31,$3A,$9E
	.dc.b	$66,$B0,$EF,$19,$8F,$87,$87,$7B,$E5,$1C,$27,$58,$FC,$27,$ED,$DF
	.dc.b	$2C,$3B,$9F,$CF,$3E,$13,$D7,$CB,$EC,$E3,$FE,$83,$18,$DB,$F9,$67
	.dc.b	$1F,$FA,$45,$FE,$A9,$19,$CF,$FC,$1B,$39,$7F,$A7,$35,$FE,$70,$D7
	.dc.b	$9F,$AD,$FE,$C4,$7F,$A6,$FD,$DF,$3C,$3F,$A8,$BD,$79,$BF,$D8,$FF
	.dc.b	$69,$9E,$C4,$D2,$67,$FC,$EE,$6E,$1F,$D9,$E6,$E9,$FD,$2E,$6D,$7B
	.dc.b	$1F,$15,$7A,$9B,$87,$82,$7F,$B5,$E6,$9D,$E5,$FD,$9B,$E7,$BC,$87
	.dc.b	$C3,$E4,$FF,$0D,$78,$D8,$C3,$FD,$CB,$9F,$F8,$12,$3B,$CE,$A7,$FC
	.dc.b	$0B,$18,$F0,$0F,$6D,$9E,$FF,$C1,$3E,$17,$D6,$5B,$FF,$03,$45,$DF
	.dc.b	$FC,$0D,$37,$2F,$B9,$68,$E3,$7E,$43,$FA,$C6,$2F,$03,$38,$7F,$89
	.dc.b	$E3,$13,$FE,$1A,$9F,$31,$1F,$ED,$DE,$2F,$FB,$25,$FF,$67,$7C,$DB
	.dc.b	$E7,$7E,$E4,$E7,$0F,$F6,$54,$E6,$BF,$C9,$3E,$9F,$DF,$8C,$4F,$FB
	.dc.b	$6E,$B8,$69,$FE,$F8,$9C,$BF,$E6,$8C,$3F,$0D,$6F,$A4,$FC,$57,$97
	.dc.b	$AC,$BF,$C0,$93,$97,$FC,$CF,$5C,$2B,$6D,$8F,$F7,$37,$AF,$F4,$9B
	.dc.b	$C3,$F9,$FF,$3D,$EE,$BE,$1F,$F2,$D7,$BD,$FF,$CF,$B3,$9C,$3F,$D1
	.dc.b	$F7,$C2,$FF,$1E,$D8,$CC,$EE,$BF,$D2,$4D,$7F,$D2,$4D,$7F,$D2,$4D
	.dc.b	$7F,$D3,$4D,$EA,$61,$FE,$F1,$AF,$F8,$01,$F0,$6F,$8A,$F3,$F9,$DF
	.dc.b	$AD,$8C,$4F,$C0,$8C,$4E,$A3,$8B,$FB,$5F,$07,$F0,$79,$38,$BF,$84
	.dc.b	$46,$1B,$DD,$BE,$69,$87,$E3,$ED,$DB,$FE,$8F,$9E,$95,$A9,$E9,$5F
	.dc.b	$E3,$BD,$1D,$BF,$38,$FE,$21,$2F,$77,$E4,$65,$FC,$D3,$7F,$83,$5A
	.dc.b	$9F,$8D,$5B,$E2,$FD,$FB,$71,$FE,$DB,$3A,$24,$E9,$B1,$FE,$4B,$F6
	.dc.b	$FF,$02,$FA,$F5,$FF,$7E,$7C,$0F,$F1,$99,$CE,$1E,$A3,$DE,$A3,$5C
	.dc.b	$7B,$6D,$F3,$9D,$3D,$7D,$D7,$D9,$FF,$25,$5F,$1D,$63,$0D,$9E,$0D
	.dc.b	$C7,$C5,$97,$FC,$00,$8F,$C0,$7C,$97,$F2,$FD,$1E,$9E,$30,$F6,$46
	.dc.b	$1F,$C4,$CE,$70,$FF,$79,$78,$57,$8B,$E8,$74,$AB,$17,$FE,$73,$F7
	.dc.b	$FF,$88,$3C,$8F,$91,$F6,$6F,$8C,$FE,$80,$D8,$CF,$D9,$E2,$64,$70
	.dc.b	$77,$18,$CF,$DB,$F6,$67,$FA,$53,$6B,$EB,$3F,$E2,$76,$C6,$88,$EB
	.dc.b	$D3,$B1,$F6,$7A,$5A,$7F,$87,$75,$F7,$FA,$1F,$92,$FB,$3F,$F7,$46
	.dc.b	$3F,$AE,$8C,$77,$FE,$3A,$73,$8F,$94,$8E,$1C,$07,$0E,$A3,$91,$2B
	.dc.b	$FC,$87,$3A,$9F,$D5,$FD,$93,$EF,$FF,$B5,$36,$BF,$2C,$BE,$D3,$33
	.dc.b	$FD,$5F,$EC,$DF,$27,$FD,$51,$B7,$E4,$C3,$FC,$6F,$18,$7F,$DB,$2D
	.dc.b	$7D,$9E,$92,$3E,$FC,$7E,$42,$9F,$0F,$DF,$A3,$FB,$F4,$63,$0F,$D5
	.dc.b	$3F,$CD,$6B,$87,$99,$FD,$AF,$EC,$FF,$C1,$9C,$3F,$E1,$5A,$D3,$FE
	.dc.b	$32,$CD,$3F,$E0,$DB,$FF,$26,$B6,$33,$EB,$FF,$70,$8C,$11,$C1,$31
	.dc.b	$FF,$64,$7C,$BB,$EC,$FF,$C2,$8D,$8F,$E3,$57,$FC,$AF,$D0,$CD,$7F
	.dc.b	$80,$7C,$DF,$EB,$39,$7C,$46,$BF,$EC,$56,$BF,$E6,$26,$D7,$3B,$FA
	.dc.b	$0D,$9F,$C2,$B5,$E9,$7F,$E1,$2F,$05,$F1,$DF,$E2,$96,$DF,$9B,$EC
	.dc.b	$9E,$E6,$6B,$FC,$2D,$18,$9E,$ED,$C3,$BB,$5F,$BA,$D7,$FD,$BC,$D7
	.dc.b	$FC,$F6,$DC,$4D,$35,$E3,$3E,$EF,$FD,$39,$7F,$CA,$5B,$7F,$C9,$ED
	.dc.b	$F6,$5F,$B9,$8F,$E3,$33,$AE,$2F,$C0,$4E,$67,$F2,$2D,$7F,$DE,$4D
	.dc.b	$7F,$D1,$ED,$AE,$B1,$FE,$32,$8C,$11,$FF,$1E,$FE,$11,$FE,$10,$C3
	.dc.b	$F8,$66,$28,$C4,$FF,$A6,$BE,$C9,$F5,$9F,$E5,$99,$CE,$29,$CE,$1F
	.dc.b	$71,$B7,$F7,$1A,$FE,$A9,$AF,$F1,$B9,$E5,$7F,$E4,$89,$CC,$F8,$63
	.dc.b	$07,$F9,$53,$C3,$EB,$9F,$F9,$ED,$B8,$73,$FB,$2B,$7F,$DC,$A6,$E2
	.dc.b	$9E,$AD,$C3,$FC,$AF,$18,$23,$82,$5F,$E6,$23,$CB,$9B,$E9,$7F,$7D
	.dc.b	$8F,$F5,$26,$E5,$FB,$9F,$A5,$FE,$B7,$87,$F6,$B9,$CB,$F0,$63,$35
	.dc.b	$FA,$26,$BF,$C4,$1A,$EF,$CE,$FF,$3A,$FE,$0D,$FC,$7E,$BF,$BB,$36
	.dc.b	$7F,$9E,$7D,$8B,$EC,$3F,$D2,$13,$9C,$BF,$D4,$4D,$BF,$C0,$B6,$DE
	.dc.b	$53,$0E,$55,$DB,$C5,$FF,$4D,$61,$FD,$EA,$FF,$B8,$B4,$FE,$31,$F5
	.dc.b	$5F,$E5,$58,$FF,$11,$C6,$6B,$F1,$ED,$8C,$FA,$DF,$E8,$DA,$E1,$BF
	.dc.b	$F8,$C6,$3F,$CA,$3D,$53,$0F,$F1,$77,$C6,$BE,$59,$F3,$B9,$CB,$FA
	.dc.b	$E7,$BB,$FB,$FD,$3F,$C5,$0D,$C3,$FB,$56,$1C,$6F,$85,$7F,$B8,$23
	.dc.b	$13,$FD,$15,$B7,$FD,$69,$AF,$F5,$F6,$BF,$E3,$BF,$7B,$FC,$CE,$7F
	.dc.b	$B2,$63,$FC,$AF,$ED,$62,$BF,$CE,$33,$6B,$FC,$55,$5A,$9C,$9F,$E4
	.dc.b	$B8,$FE,$D9,$F1,$9F,$D5,$DA,$50,$D6,$9F,$D7,$7E,$C3,$F7,$3F,$59
	.dc.b	$C5,$B1,$9E,$B4,$E5,$B1,$C7,$EC,$3F,$CC,$4D,$D3,$94,$62,$3F,$BE
	.dc.b	$DA,$31,$1F,$E5,$DF,$83,$FF,$02,$57,$D2,$6E,$59,$9F,$B7,$F3,$8F
	.dc.b	$9A,$CC,$7E,$4B,$87,$FD,$70,$FF,$A6,$2F,$3D,$3F,$E8,$27,$7B,$B0
	.dc.b	$E5,$5F,$86,$FA,$FF,$F2,$5B,$F6,$DA,$FD,$B6,$BF,$6D,$AF,$F9,$D7
	.dc.b	$EC,$3B,$E2,$69,$E0,$9E,$7A,$7E,$3A,$33,$5F,$EA,$E7,$5D,$DE,$CE
	.dc.b	$9F,$9F,$F6,$17,$8C,$D3,$E0,$F8,$6E,$CF,$F4,$E6,$3F,$E7,$3A,$DB
	.dc.b	$6B,$83,$FD,$E2,$76,$FF,$99,$2B,$FB,$4B,$FE,$7B,$6B,$9F,$2B,$E5
	.dc.b	$F5,$3E,$97,$CD,$5A,$D5,$CB,$FD,$1A,$30,$DB,$E3,$5E,$8F,$0D,$C7
	.dc.b	$F9,$7B,$46,$32,$FF,$00,$B5,$FE,$41,$3F,$9D,$DB,$F9,$67,$D8,$7C
	.dc.b	$7F,$14,$D3,$B4,$FC,$6F,$84,$FF,$6B,$DF,$F8,$31,$AF,$F8,$31,$AF
	.dc.b	$F8,$E9,$AF,$F5,$16,$AC,$7F,$CF,$DF,$11,$E9,$9C,$C6,$27,$FC,$91
	.dc.b	$8B,$C9,$F5,$6F,$FD,$4F,$1F,$BD,$18,$DF,$B6,$5F,$73,$EC,$1F,$61
	.dc.b	$F1,$5B,$97,$F8,$91,$AF,$F8,$91,$AF,$F9,$49,$AF,$F6,$96,$BE,$96
	.dc.b	$E6,$73,$9F,$FB,$53,$C4,$FC,$93,$FC,$A6,$BF,$D7,$CD,$7F,$C2,$D3
	.dc.b	$E2,$D3,$FC,$A5,$E2,$FE,$92,$7E,$33,$70,$FF,$18,$F4,$7D,$F8,$C1
	.dc.b	$1C,$12,$7E,$4F,$D8,$7D,$6B,$FC,$6C,$7F,$BA,$23,$1A,$FF,$85,$1B
	.dc.b	$1E,$0A,$FD,$2F,$D8,$7F,$B2,$23,$FC,$96,$D7,$FC,$96,$D7,$FC,$F2
	.dc.b	$D7,$FC,$20,$D7,$F0,$0D,$C4,$D3,$59,$FF,$B2,$BD,$3F,$DD,$1F,$EC
	.dc.b	$CA,$FF,$73,$35,$FF,$1F,$4F,$8F,$2F,$F3,$D7,$AA,$F8,$3F,$D2,$36
	.dc.b	$BF,$CC,$6D,$7F,$CC,$6D,$7F,$CD,$D1,$C0,$8F,$07,$F5,$1E,$C7,$F9
	.dc.b	$8A,$3F,$DF,$0D,$7F,$DC,$AD,$AF,$1C,$FF,$0E,$F8,$D7,$C8,$FE,$DD
	.dc.b	$B7,$FE,$76,$6B,$FE,$76,$6B,$FE,$A4,$6B,$FE,$3B,$6B,$FA,$B6,$E8
	.dc.b	$73,$A3,$EB,$DF,$61,$FC,$95,$B7,$FE,$DF,$CB,$F0,$B5,$BA,$F4,$63
	.dc.b	$FC,$9B,$E6,$BF,$DD,$D7,$FE,$AE,$FA,$9F,$E5,$DF,$02,$75,$F1,$BB
	.dc.b	$FB,$0F,$E2,$96,$69,$2F,$F7,$BF,$40,$F9,$1F,$E7,$0D,$8F,$27,$47
	.dc.b	$F9,$A2,$31,$0D,$BC,$F1,$FC,$9A,$3F,$DF,$1F,$60,$F5,$BF,$BC,$36
	.dc.b	$3F,$BE,$35,$FF,$6A,$35,$FF,$74,$B5,$FF,$77,$35,$EB,$1D,$0F,$F0
	.dc.b	$37,$A3,$FE,$FC,$C3,$FB,$3F,$C7,$FE,$BA,$BF,$E0,$03,$71,$C6,$1E
	.dc.b	$17,$E7,$FD,$7F,$F1,$9A,$71,$A3,$0D,$F1,$3F,$AE,$7D,$87,$F4,$9A
	.dc.b	$F8,$A7,$0A,$D9,$27,$7E,$57,$F6,$3F,$88,$7B,$94,$97,$CB,$14,$FF
	.dc.b	$02,$6C,$EA,$7E,$86,$BF,$E9,$AA,$E7,$1C,$5F,$E1,$3B,$7D,$BF,$3B
	.dc.b	$FC,$13,$C5,$FE,$48,$AF,$3C,$E5,$BF,$60,$F9,$CD,$DC,$9D,$DF,$17
	.dc.b	$C2,$7F,$9A,$2B,$83,$A9,$FF,$00,$7D,$83,$E7,$29,$CE,$78,$EA,$F9
	.dc.b	$9F,$C4,$37,$1F,$F0,$44,$7E,$3B,$8F,$F4,$6E,$89,$1A,$37,$E9,$FF
	.dc.b	$83,$FE,$97,$F7,$35,$FA,$A5,$F4,$19,$35,$E0,$4C,$BE,$84,$E9,$B3
	.dc.b	$97,$DE,$3E,$77,$F5,$CF,$F9,$55,$A8,$85,$72,$E0,$E5,$E0,$7F,$96
	.dc.b	$E3,$31,$C1,$AF,$18,$D1,$F9,$EF,$71,$F0,$9F,$E3,$62,$4F,$F9,$A6
	.dc.b	$BC,$79,$7F,$0A,$FB,$0F,$AB,$3F,$41,$B1,$9E,$4F,$F1,$B4,$7F,$05
	.dc.b	$9F,$C9,$3A,$E7,$FF,$73,$B6,$3F,$4A,$DE,$13,$C6,$3F,$D2,$7E,$AF
	.dc.b	$E4,$3E,$B2,$47,$E4,$34,$FF,$05,$36,$72,$FE,$33,$A3,$F8,$95,$7F
	.dc.b	$94,$A7,$FC,$CB,$F4,$2F,$B0,$F8,$AD,$E0,$FF,$43,$37,$0F,$9C,$1C
	.dc.b	$3F,$45,$AF,$D6,$6B,$F6,$5A,$F1,$9F,$A7,$87,$F9,$BF,$E6,$3F,$CE
	.dc.b	$7F,$06,$69,$E6,$9F,$A3,$B7,$E0,$F6,$7D,$1B,$EC,$3F,$CF,$7C,$FF
	.dc.b	$E1,$1E,$3F,$F0,$1E,$7F,$D5,$3D,$AF,$F3,$45,$EF,$5F,$E0,$8D,$7E
	.dc.b	$35,$F1,$CB,$F2,$17,$C6,$77,$8F,$F9,$C7,$E9,$5F,$41,$FE,$8A,$FB
	.dc.b	$0F,$F3,$CE,$1C,$1B,$7F,$C8,$2B,$6F,$F1,$36,$BF,$56,$37,$81,$E3
	.dc.b	$AA,$B7,$AF,$F8,$7B,$D4,$7D,$3F,$FA,$E5,$B8,$FF,$99,$E3,$07,$FA
	.dc.b	$F7,$17,$D8,$3E,$BF,$FD,$42,$DA,$FE,$32,$D7,$F8,$9D,$F4,$73,$F1
	.dc.b	$BC,$08,$F0,$AA,$FE,$3E,$BF,$D5,$2D,$9E,$F3,$FE,$94,$C3,$F8,$D4
	.dc.b	$FE,$32,$FD,$7E,$27,$A4,$91,$FE,$66,$FA,$AF,$F5,$43,$FC,$ED,$AF
	.dc.b	$F2,$06,$E7,$F1,$6F,$AC,$FE,$87,$EC,$1D,$9F,$D7,$1B,$1F,$9C,$D5
	.dc.b	$18,$8F,$E1,$91,$89,$FF,$46,$61,$FE,$CA,$9F,$07,$C7,$FF,$AC,$E3
	.dc.b	$18,$FE,$43,$0E,$4F,$F8,$42,$33,$C5,$F9,$96,$DF,$F6,$46,$E3,$DE
	.dc.b	$3D,$D8,$75,$23,$FD,$3B,$0E,$B7,$F9,$6F,$4F,$FC,$91,$3F,$C9,$AF
	.dc.b	$FC,$0D,$F0,$1F,$DE,$A7,$33,$B8,$F5,$C7,$A4,$91,$FD,$69,$FF,$29
	.dc.b	$65,$FD,$FF,$C6,$7D,$7F,$FA,$C7,$E5,$3F,$33,$18,$F4,$BF,$9B,$46
	.dc.b	$27,$5D,$3F,$CE,$78,$DF,$61,$F8,$DA,$D5,$F0,$AF,$E7,$B8,$73,$CD
	.dc.b	$5A,$6C,$F3,$BE,$4F,$0C,$7B,$CF,$F4,$7B,$E5,$7C,$4F,$D8,$E3,$E2
	.dc.b	$E5,$FE,$97,$19,$C3,$96,$5E,$3C,$39,$63,$D3,$0E,$11,$91,$8F,$6B
	.dc.b	$E3,$1E,$63,$8D,$F3,$AE,$31,$93,$8D,$FA,$1B,$3D,$6F,$A4,$7F,$B2
	.dc.b	$F2,$F7,$BD,$0E,$51,$A7,$89,$E1,$BE,$F8,$FB,$63,$83,$63,$38,$FA
	.dc.b	$23,$53,$C9,$F8,$F7,$E0,$8C,$4E,$C8,$F9,$72,$F9,$71,$88,$D9,$E0
	.dc.b	$7E,$C4,$6A,$30,$43,$77,$E4,$7D,$C7,$F7,$DC,$3F,$08,$DC,$BF,$E0
	.dc.b	$03,$F4,$0D,$7D,$87,$F6,$26,$F8,$BF,$D6,$1F,$AF,$19,$7F,$1E,$D7
	.dc.b	$F1,$0F,$F6,$AC,$FF,$3A,$DC,$BF,$35,$EC,$FF,$AC,$78,$BE,$B4,$E6
	.dc.b	$F8,$E1,$83,$F0,$15,$B8,$FA,$F3,$B7,$C7,$3B,$7E,$C4,$EF,$0F,$0C
	.dc.b	$E5,$B3,$AF,$81,$FF,$6F,$C6,$09,$8F,$B2,$BC,$33,$A8,$C5,$78,$A7
	.dc.b	$A4,$7B,$27,$53,$EA,$34,$4A,$FD,$4D,$F9,$6B,$9C,$78,$27,$83,$AA
	.dc.b	$E7,$1E,$E3,$94,$F0,$3C,$18,$F8,$A3,$8C,$65,$68,$C3,$C1,$1A,$21
	.dc.b	$D7,$0F,$05,$F3,$8C,$5F,$3B,$E5,$87,$93,$46,$7F,$67,$B5,$CE,$F8
	.dc.b	$76,$7F,$D3,$DB,$9F,$F8,$AF,$93,$FD,$17,$39,$3F,$94,$7A,$BF,$EA
	.dc.b	$59,$FE,$68,$D5,$CF,$FE,$B9,$8C,$36,$7F,$AF,$7D,$87,$6E,$7F,$E9
	.dc.b	$0D,$8F,$E7,$DF,$61,$70,$FE,$91,$37,$4F,$4B,$74,$3B,$C7,$FD,$91
	.dc.b	$79,$89,$87,$F6,$B7,$59,$A7,$38,$FB,$DF,$3C,$66,$BF,$B4,$36,$7F
	.dc.b	$B3,$B5,$FF,$3B,$E3,$F3,$3E,$C3,$C8,$7A,$78,$7F,$C3,$3E,$2E,$30
	.dc.b	$FF,$5F,$C6,$1F,$F0,$27,$7A,$7E,$54,$E7,$0F,$F4,$03,$5E,$FA,$BE
	.dc.b	$03,$E1,$1F,$0F,$EA,$2E,$3F,$E0,$98,$B5,$FE,$09,$9B,$C1,$FD,$DD
	.dc.b	$E1,$5E,$4C,$9E,$06,$2F,$FC,$4B,$18,$9F,$3C,$F9,$A3,$13,$E6,$6B
	.dc.b	$F1,$3C,$5F,$F6,$25,$7F,$B2,$3D,$47,$D2,$FD,$B9,$CD,$7F,$9E,$67
	.dc.b	$35,$FE,$49,$F4,$FE,$E4,$62,$7F,$D8,$35,$C3,$3F,$F3,$0C,$E5,$FC
	.dc.b	$CF,$0F,$F9,$A3,$1F,$C2,$60,$FA,$8F,$C5,$79,$9B,$CF,$FC,$C3,$39
	.dc.b	$7F,$CC,$F5,$C2,$B6,$D8,$FF,$49,$71,$7E,$3F,$D3,$FB,$11,$FE,$FA
	.dc.b	$F4,$9F,$20,$F8,$9D,$5F,$E6,$E7,$31,$FA,$76,$D7,$FA,$41,$BC,$1F
	.dc.b	$E8,$FB,$E5,$39,$6C,$66,$B9,$46,$23,$FD,$28,$D7,$FC,$FB,$87,$FB
	.dc.b	$AE,$BF,$DF,$3F,$08,$F8,$BF,$57,$9D,$FA,$98,$C4,$FF,$AB,$E3,$13
	.dc.b	$A8,$E2,$FE,$CF,$C4,$FC,$0F,$17,$D0,$6C,$FF,$A6,$57,$66,$F9,$E4
	.dc.b	$FF,$B7,$EB,$84,$73,$9C,$E7,$FE,$BB,$CD,$D9,$F3,$CF,$E0,$91,$FF
	.dc.b	$00,$76,$C6,$3C,$CF,$F7,$4F,$BB,$F0,$67,$FD,$D8,$D7,$84,$7F,$67
	.dc.b	$8C,$12,$7F,$8D,$4E,$8F,$F6,$1C,$73,$E6,$FF,$7D,$70,$7F,$13,$F6
	.dc.b	$9F,$03,$63,$0D,$C7,$C9,$82,$D7,$C2,$73,$1D,$1B,$19,$96,$DF,$9C
	.dc.b	$F7,$1F,$61,$FC,$8D,$7C,$75,$8C,$37,$0E,$B3,$A6,$CE,$9F,$C1,$FC
	.dc.b	$97,$F2,$DD,$5E,$9F,$E0,$E3,$0D,$9D,$9E,$CF,$1D,$F0,$2F,$0F,$5A
	.dc.b	$FF,$BC,$FA,$98,$E3,$6F,$F6,$BD,$EF,$1E,$77,$EE,$7F,$DC,$1F,$60
	.dc.b	$F8,$BF,$E8,$0D,$8F,$8E,$BE,$C3,$33,$27,$2F,$08,$C5,$75,$3E,$97
	.dc.b	$D8,$1F,$E9,$8D,$AF,$A6,$BD,$4C,$CC,$9C,$BC,$23,$0F,$BF,$D2,$FD
	.dc.b	$51,$F9,$6F,$B0,$FF,$74,$D7,$EC,$23,$1D,$FF,$8E,$9C,$D7,$8C,$8E
	.dc.b	$1C,$07,$0E,$A3,$91,$2B,$FC,$85,$3A,$9F,$C8,$7D,$83,$EC,$3F,$B5
	.dc.b	$36,$BF,$22,$BE,$93,$33,$FD,$67,$EC,$1F,$27,$FD,$51,$B7,$FE,$DB
	.dc.b	$6E,$1F,$D5,$9A,$FA,$FD,$24,$7D,$F8,$FA,$5D,$DD,$3E,$1F,$B7,$47
	.dc.b	$F7,$DC,$51,$7C,$C9,$97,$F6,$8F,$B0,$FF,$09,$E3,$CA,$FF,$9C,$3F
	.dc.b	$EF,$A8,$CB,$E1,$6B,$FE,$B5,$6B,$FE,$30,$6E,$66,$93,$D7,$FE,$E6
	.dc.b	$F0,$CD,$F2,$EF,$B0,$FF,$0A,$37,$2F,$EF,$BE,$67,$F4,$DC,$BF,$80
	.dc.b	$4E,$5F,$C9,$65,$FC,$C7,$D0,$FE,$FF,$5F,$D7,$67,$E4,$35,$FF,$66
	.dc.b	$35,$FF,$23,$B6,$79,$5F,$E1,$A7,$35,$D0,$FF,$1A,$63,$FB,$CF,$8E
	.dc.b	$FF,$15,$78,$47,$FC,$2D,$F6,$1F,$E4,$3C,$3F,$89,$CE,$5F,$F0,$BC
	.dc.b	$67,$0F,$F0,$9B,$5F,$BC,$D7,$FD,$C4,$D7,$FC,$BA,$DC,$4D,$27,$B9
	.dc.b	$E6,$FB,$07,$C2,$FF,$8F,$DB,$87,$F8,$A3,$EC,$3F,$6D,$8F,$E2,$A7
	.dc.b	$2F,$CC,$8C,$BF,$91,$6B,$FE,$F5,$6B,$FE,$70,$6E,$26,$92,$3F,$C6
	.dc.b	$6F,$0D,$1F,$61,$FE,$11,$C3,$F8,$54,$63,$D8,$FF,$20,$7D,$83,$EB
	.dc.b	$3F,$CB,$17,$BC,$7F,$CF,$F3,$9C,$3E,$C6,$48,$C4,$FF,$90,$1A,$FE
	.dc.b	$AD,$AF,$F1,$98,$E1,$7F,$E4,$89,$CC,$F6,$8C,$1F,$82,$F0,$BA,$F9
	.dc.b	$1F,$E2,$66,$E1,$CF,$EC,$2D,$FF,$57,$9B,$8A,$7A,$B7,$0F,$F2,$BC
	.dc.b	$60,$8E,$09,$7E,$E8,$F2,$E8,$FA,$5F,$DF,$CF,$F6,$26,$E3,$FA,$6F
	.dc.b	$A6,$FE,$B0,$FF,$08,$8C,$E7,$FE,$96,$C9,$5A,$AF,$E2,$0D,$7E,$33
	.dc.b	$5F,$F5,$7B,$70,$7D,$7F,$F2,$1C,$7F,$B9,$B6,$BF,$A4,$79,$EF,$AF
	.dc.b	$7D,$87,$FA,$4E,$B5,$9F,$F9,$41,$B8,$7E,$11,$B8,$7F,$A9,$2F,$8C
	.dc.b	$F6,$F1,$BF,$D3,$97,$FD,$FC,$FF,$2C,$96,$3E,$F7,$D5,$7F,$0B,$CF
	.dc.b	$FC,$07,$EC,$70,$24,$FF,$A3,$27,$55,$C7,$1F,$F5,$9F,$B9,$FB,$98
	.dc.b	$FF,$5E,$EA,$8C,$5F,$F6,$3E,$97,$CB,$BE,$77,$FC,$BD,$D9,$2B,$FC
	.dc.b	$01,$7C,$32,$FD,$BD,$EE,$73,$5F,$EB,$9C,$39,$E1,$FE,$C5,$6C,$FF
	.dc.b	$17,$6D,$FF,$5E,$6B,$FD,$8F,$1F,$37,$BF,$FE,$6F,$3A,$64,$F6,$B1
	.dc.b	$87,$92,$6D,$FF,$7F,$C3,$DD,$58,$D3,$FD,$AD,$F1,$DF,$C9,$EB,$FB
	.dc.b	$83,$5F,$F0,$23,$53,$BF,$B0,$FD,$E7,$D6,$DE,$B3,$49,$5F,$CC,$E5
	.dc.b	$F6,$1F,$E3,$A8,$C3,$5F,$EF,$B1,$AC,$3B,$4E,$67,$FC,$2D,$B3,$F3
	.dc.b	$0F,$F3,$1C,$7F,$3F,$8C,$7C,$27,$AD,$B1,$99,$DB,$FE,$0A,$8C,$B6
	.dc.b	$32,$FC,$F6,$BE,$06,$BF,$CD,$3F,$38,$FF,$1D,$BF,$65,$AF,$D9,$6B
	.dc.b	$FE,$1B,$D9,$18,$8E,$9C,$7E,$7A,$34,$91,$E9,$6B,$E7,$6B,$FE,$F0
	.dc.b	$D3,$FB,$93,$FD,$87,$9F,$E9,$F9,$9F,$75,$B1,$99,$DB,$FE,$2C,$8C
	.dc.b	$B6,$32,$F5,$9E,$51,$88,$FF,$24,$2F,$D9,$1F,$6F,$F8,$B8,$C3,$63
	.dc.b	$33,$FD,$D6,$72,$D8,$CB,$FE,$3E,$8C,$B6,$32,$FF,$91,$63,$2F,$F1
	.dc.b	$EF,$BC,$FC,$D7,$83,$FC,$89,$BE,$09,$A7,$C3,$FC,$57,$9D,$7F,$58
	.dc.b	$6B,$FE,$95,$6E,$67,$37,$FE,$7A,$F3,$7B,$FC,$17,$F8,$26,$FF,$AC
	.dc.b	$B5,$FF,$2B,$CE,$DC,$CF,$02,$38,$3E,$C7,$E1,$9F,$EB,$0C,$0D,$27
	.dc.b	$67,$BB,$EC,$9F,$5F,$49,$7F,$DA,$A3,$2D,$8C,$BE,$48,$CB,$63,$F0
	.dc.b	$DC,$CF,$98,$A4,$9F,$C6,$46,$5F,$EE,$FF,$82,$FF,$07,$D0,$D2,$47
	.dc.b	$D4,$6B,$C0,$C1,$F9,$5F,$69,$F6,$D4,$91,$F9,$B8,$CB,$FE,$06,$FB
	.dc.b	$37,$D2,$7E,$01,$B1,$98,$C1,$1C,$10,$FA,$B1,$C1,$FD,$17,$AB,$FE
	.dc.b	$CA,$7F,$54,$D7,$3F,$1D,$FA,$B8,$C4,$7F,$95,$BE,$73,$F2,$3E,$D9
	.dc.b	$A4,$8F,$F0,$EC,$65,$B1,$97,$F0,$F1,$97,$F6,$9F,$45,$FE,$BD,$7F
	.dc.b	$78,$D7,$3F,$15,$FB,$D8,$C4,$7F,$A9,$BE,$AD,$F2,$BF,$CE,$DB,$1F
	.dc.b	$CF,$DA,$FF,$99,$DA,$F1,$8E,$04,$78,$3C,$FE,$DE,$47,$36,$BC,$FE
	.dc.b	$1B,$C0,$47,$C0,$F8,$FC,$D9,$1E,$36,$BF,$E7,$9F,$94,$7D,$67,$F5
	.dc.b	$36,$C7,$F5,$76,$BB,$8D,$7B,$F2,$34,$D7,$A3,$5F,$F6,$5C,$A3,$11
	.dc.b	$FD,$3A,$7A,$BF,$DC,$5A,$E5,$FE,$9C,$61,$76,$BE,$C1,$CB,$FC,$C5
	.dc.b	$B3,$FC,$D5,$AF,$F3,$16,$BF,$EE,$19,$E5,$3C,$49,$1C,$5F,$E7,$9F
	.dc.b	$0F,$FC,$3E,$38,$35,$FF,$39,$B5,$F2,$B6,$4D,$27,$81,$D0,$FF,$02
	.dc.b	$7C,$D3,$F1,$A7,$26,$1B,$87,$FA,$BE,$75,$39,$24,$66,$76,$FF,$BE
	.dc.b	$A7,$6E,$67,$E5,$F4,$C8,$FE,$F9,$A7,$ED,$5F,$A5,$1F,$D6,$FE,$5D
	.dc.b	$F6,$F4,$93,$FD,$86,$32,$D8,$CB,$FD,$7E,$32,$D8,$FE,$CD,$F4,$BF
	.dc.b	$C3,$A3,$FB,$23,$5F,$A8,$D7,$FD,$3E,$D7,$F0,$2D,$D4,$D2,$76,$FE
	.dc.b	$6F,$ED,$65,$7E,$6D,$AF,$E8,$5A,$FE,$59,$AF,$F8,$8B,$F0,$1F,$88
	.dc.b	$DC,$BF,$C4,$11,$88,$78,$48,$C6,$2E,$A7,$CE,$7F,$81,$5B,$7F,$E0
	.dc.b	$96,$BF,$E0,$56,$BF,$95,$6B,$FE,$C0,$6B,$FA,$26,$CF,$79,$FD,$3F
	.dc.b	$CF,$4A,$FB,$AD,$7F,$C6,$9B,$A3,$11,$FB,$16,$A1,$C2,$FC,$61,$F8
	.dc.b	$2F,$E6,$1B,$3F,$95,$6B,$F9,$16,$BC,$CE,$6E,$1E,$D3,$D6,$7A,$12
	.dc.b	$3F,$CA,$19,$FC,$0D,$3E,$76,$5F,$48,$93,$FE,$33,$FA,$3F,$D0,$BF
	.dc.b	$BD,$6B,$FE,$E8,$6B,$FC,$29,$BA,$1A,$4E,$4F,$F1,$E3,$AF,$A5,$7E
	.dc.b	$29,$FE,$3F,$6D,$FF,$90,$DA,$FF,$8F,$DA,$FF,$0F,$6B,$FE,$FA,$6B
	.dc.b	$FC,$71,$B6,$73,$E4,$BE,$9D,$F2,$FF,$E3,$26,$DF,$FB,$F1,$AF,$F8
	.dc.b	$91,$AF,$87,$EB,$7D,$04,$7D,$0F,$3F,$A6,$91,$CD,$AF,$36,$BF,$6E
	.dc.b	$FC,$91,$88,$FF,$64,$FA,$3F,$E4,$6F,$89,$34,$91,$FE,$8B,$FB,$07
	.dc.b	$CD,$FF,$A2,$DB,$19,$8C,$1F,$E8,$FF,$88,$7D,$BD,$24,$FF,$A5,$63
	.dc.b	$2F,$F9,$D7,$EB,$19,$7F,$9D,$5B,$19,$8C,$11,$C1,$0F,$F3,$C4,$70
	.dc.b	$6C,$6D,$A6,$CF,$F6,$57,$EB,$EF,$A3,$FF,$53,$36,$3F,$D5,$4D,$7F
	.dc.b	$D4,$CD,$7F,$AA,$57,$FC,$00,$7D,$5F,$29,$E9,$23,$FE,$CB,$F0,$DF
	.dc.b	$60,$F8,$8F,$F6,$5E,$3F,$AF,$9C,$C6,$2B,$F7,$91,$87,$FD,$9F,$3F
	.dc.b	$BF,$E1,$7A,$6E,$3F,$F6,$BC,$F3,$C7,$B5,$F0,$39,$B8,$F8,$EF,$66
	.dc.b	$3D,$5F,$81,$8F,$4C,$39,$DF,$0C,$BC,$8F,$F3,$AC,$39,$63,$DE,$7D
	.dc.b	$17,$C6,$B7,$8F,$5A,$D4,$78,$CE,$55,$C3,$2F,$54,$70,$C3,$AB,$E1
	.dc.b	$38,$5F,$35,$87,$77,$33,$C1,$EA,$62,$FB,$63,$EC,$8C,$E5,$F0,$AB
	.dc.b	$86,$1D,$70,$E9,$39,$C3,$C8,$4B,$F9,$93,$AC,$BD,$B5,$B9,$F0,$1A
	.dc.b	$CB,$DA,$E2,$34,$49,$FB,$93,$A8,$C4,$FC,$16,$D7,$78,$C5,$78,$08
	.dc.b	$FC,$4A,$E5,$1F,$42,$3F,$9E,$1F,$10,$B8,$7E,$35,$B8,$FE,$21,$CF
	.dc.b	$D7,$BE,$67,$F3,$4D,$8E,$67,$F7,$C7,$0F,$08,$D9,$F8,$28,$E1,$97
	.dc.b	$E3,$BE,$C1,$F4,$FF,$C4,$2F,$EE,$D6,$B2,$FD,$23,$63,$F4,$6D,$D1
	.dc.b	$7C,$B9,$FB,$65,$2F,$9B,$F9,$49,$D4,$60,$99,$7E,$8B,$0E,$93,$A8
	.dc.b	$C5,$73,$6D,$7E,$02,$75,$8F,$9A,$B4,$D9,$D6,$5E,$DB,$E2,$49,$D3
	.dc.b	$F0,$E3,$51,$82,$4F,$C0,$AF,$54,$EA,$30,$7A,$A7,$79,$7C,$49,$D4
	.dc.b	$FD,$5A,$F2,$BC,$2F,$C1,$96,$23,$47,$33,$F5,$B8,$FB,$9E,$18,$F9
	.dc.b	$1D,$D7,$27,$87,$63,$89,$F5,$DF,$76,$7F,$93,$B5,$FF,$0C,$4E,$E3
	.dc.b	$11,$FE,$88,$9E,$38,$FF,$32,$F9,$8C,$F8,$6E,$64,$61,$B3,$B3,$FA
	.dc.b	$67,$D7,$F4,$CB,$C8,$DC,$0D,$24,$F0,$26,$3F,$D3,$23,$2D,$8C,$BF
	.dc.b	$D0,$A3,$C5,$E4,$C9,$FE,$92,$D9,$FE,$AB,$E7,$78,$49,$3F,$D8,$5F
	.dc.b	$83,$C5,$FE,$09,$F8,$CB,$E0,$9A,$4A,$FF,$05,$6F,$A2,$38,$26,$1F
	.dc.b	$E0,$D8,$CB,$63,$FB,$87,$85,$F0,$C8,$E0,$FF,$0B,$7B,$5E,$22,$38
	.dc.b	$3F,$C5,$5E,$87,$CD,$F1,$4D,$24,$68,$97,$FE,$2F,$8C,$BF,$C5,$38
	.dc.b	$3F,$C7,$9F,$31,$F5,$FC,$83,$49,$7F,$E4,$96,$D7,$19,$DC,$E4,$FE
	.dc.b	$2B,$CA,$FC,$15,$F6,$BF,$8A,$6C,$E9,$B7,$FC,$9A,$B8,$57,$E5,$3C
	.dc.b	$FF,$F1,$E7,$A9,$78,$BF,$D5,$F8,$FF,$AF,$DB,$87,$FA,$ED,$AF,$F9
	.dc.b	$AE,$B8,$CE,$EB,$91,$1F,$3F,$C9,$BE,$6F,$F3,$ED,$8C,$C6,$08,$E0
	.dc.b	$97,$E9,$9E,$0F,$CD,$ED,$7C,$07,$E8,$DB,$5F,$CE,$B8,$3F,$C4,$53
	.dc.b	$92,$46,$67,$75,$CE,$72,$FC,$7D,$9F,$19,$F3,$49,$7F,$35,$FD,$EC
	.dc.b	$BD,$97,$C3,$61,$B5,$FD,$32,$75,$39,$24,$66,$77,$3F,$D4,$9A,$FF
	.dc.b	$5E,$6B,$FC,$3B,$E0,$7F,$D4,$73,$EF,$6C,$7E,$45,$AF,$FA,$87,$34
	.dc.b	$62,$39,$7A,$FF,$C2,$E3,$10,$9D,$CE,$4F,$E2,$1F,$04,$F8,$2A,$4C
	.dc.b	$3E,$84,$65,$B1,$9B,$FD,$24,$E5,$B1,$F8,$4F,$22,$57,$E2,$1B,$3F
	.dc.b	$DF,$B8,$DF,$56,$FA,$DF,$CD,$36,$FF,$BC,$CE,$A7,$24,$8C,$CE,$EB
	.dc.b	$A3,$5F,$CE,$35,$FE,$77,$F5,$DF,$ED,$79,$FC,$63,$63,$F7,$CD,$7E
	.dc.b	$73,$5E,$3E,$27,$F3,$78,$C4,$7C,$D6,$EA,$6F,$C2,$FF,$A5,$2F,$F3
	.dc.b	$4D,$AF,$E1,$8D,$7E,$E3,$5F,$EA,$2D,$7E,$D3,$68,$D3,$5D,$6E,$77
	.dc.b	$9B,$FB,$62,$61,$FB,$A7,$F3,$9E,$E4,$8F,$CC,$37,$1F,$CC,$35,$FC
	.dc.b	$C3,$5F,$F7,$97,$27,$C2,$7F,$51,$A7,$E9,$C9,$5F,$A3,$FA,$CF,$D5
	.dc.b	$11,$C1,$FA,$7F,$73,$F5,$C4,$70,$7E,$AF,$AF,$F6,$44,$70,$7E,$BF
	.dc.b	$E4,$BF,$6C,$47,$04,$C3,$F6,$51,$96,$C7,$F0,$DF,$51,$F0,$7F,$DC
	.dc.b	$DB,$19,$8C,$1F,$D8,$BD,$F7,$C7,$D2,$4F,$F9,$43,$CD,$4E,$A3,$0D
	.dc.b	$C3,$FC,$BF,$3A,$AE,$31,$89,$D3,$FB,$FF,$C0,$FF,$DE,$F7,$FC,$8D
	.dc.b	$B1,$FD,$55,$AE,$FC,$CF,$EE,$FF,$50,$F8,$BA,$4B,$F3,$46,$5B,$19
	.dc.b	$AF,$2C,$E5,$B1,$FE,$17,$F8,$BF,$EB,$15,$FC,$D5,$B3,$FD,$9D,$AF
	.dc.b	$F1,$76,$BF,$EC,$5F,$90,$7C,$CF,$F8,$31,$B5,$F2,$5B,$3F,$19,$A8
	.dc.b	$E5,$3C,$49,$3C,$5F,$F1,$67,$4F,$F8,$DA,$7F,$BD,$35,$FE,$54,$D7
	.dc.b	$FC,$7E,$D7,$FD,$73,$F4,$DF,$E3,$87,$EB,$B5,$F1,$1A,$F6,$7F,$C4
	.dc.b	$B1,$98,$FF,$71,$35,$D5,$FE,$16,$31,$C0,$75,$DA,$F9,$5F,$F4,$1D
	.dc.b	$FF,$3F,$6B,$FE,$54,$6B,$FC,$F5,$B4,$FA,$67,$DF,$FF,$8E,$1B,$8F
	.dc.b	$E5,$5B,$7F,$92,$6B,$EC,$AE,$55,$C4,$95,$C5,$FF,$2C,$3F,$C3,$8F
	.dc.b	$EE,$C5,$9F,$F1,$43,$5F,$EB,$0D,$7E,$D4,$BE,$67,$FA,$54,$78,$5A
	.dc.b	$E9,$AF,$FB,$B1,$BF,$34,$DF,$F9,$33,$7E,$2B,$FC,$EE,$DC,$3F,$D8
	.dc.b	$6D,$BF,$F6,$7B,$5F,$B6,$D7,$F1,$CD,$4D,$D1,$F4,$0F,$B3,$FF,$3D
	.dc.b	$36,$FF,$C3,$0D,$BF,$F2,$53,$5F,$EF,$4D,$7F,$D1,$ED,$7F,$BC,$36
	.dc.b	$BB,$D7,$FA,$3B,$E7,$FF,$D0,$B7,$FE,$4A,$6D,$FF,$91,$5A,$FE,$5B
	.dc.b	$89,$18,$8F,$D5,$46,$3E,$03,$A1,$2B,$A3,$FE,$A0,$F8,$87,$E8,$1F
	.dc.b	$9B,$6F,$2E,$63,$07,$E7,$EF,$BF,$C5,$F9,$C9,$87,$9D,$B8,$F9,$DA
	.dc.b	$F9,$F2,$47,$F8,$D2,$3F,$80,$7C,$37,$F8,$C5,$A5,$70,$9F,$F1,$B7
	.dc.b	$D1,$BE,$93,$EC,$CF,$EE,$0F,$E1,$51,$9E,$1D,$CE,$A7,$53,$97,$80
	.dc.b	$5F,$49,$F8,$28,$CC,$63,$8B,$C3,$39,$9E,$6D,$7E,$8F,$D3,$BE,$97
	.dc.b	$F8,$5D,$35,$CC,$61,$B5,$FC,$86,$74,$B6,$FB,$B5,$B8,$C5,$F5,$78
	.dc.b	$5F,$27,$F3,$2D,$BE,$77,$CA,$76,$75,$DB,$F9,$BB,$F8,$FE,$DF,$F3
	.dc.b	$14,$FE,$31,$B3,$A9,$FF,$4B,$7E,$20,$F9,$FF,$25,$A3,$13,$DA,$31
	.dc.b	$3A,$AE,$58,$79,$2F,$77,$E8,$3E,$03,$88,$D5,$71,$C3,$E1,$73,$FE
	.dc.b	$E3,$EE,$2F,$09,$A6,$3E,$9E,$E5,$F0,$2B,$83,$6B,$83,$67,$AF,$6B
	.dc.b	$ED,$69,$27,$E4,$D3,$EA,$3F,$DC,$D1,$D6,$B8,$36,$B8,$3E,$DA,$E1
	.dc.b	$F1,$DE,$AF,$A8,$E3,$3F,$ED,$36,$BF,$E8,$36,$BF,$D9,$72,$46,$23
	.dc.b	$FD,$05,$F5,$9F,$98,$7D,$CD,$7E,$64,$A3,$1B,$FE,$3A,$39,$DF,$CC
	.dc.b	$F1,$7F,$D4,$F1,$FD,$5F,$93,$FD,$48,$DC,$3F,$D7,$13,$A2,$4E,$9B
	.dc.b	$1F,$EB,$DF,$95,$79,$1F,$61,$B1,$F4,$DA,$F9,$A7,$F4,$2F,$A3,$99
	.dc.b	$F5,$1E,$26,$C7,$FB,$25,$AF,$F8,$0A,$34,$47,$4D,$7F,$E0,$06,$2F
	.dc.b	$B0,$FC,$5B,$63,$F0,$CD,$7C,$53,$F9,$0C,$5F,$29,$F1,$1B,$3F,$E2
	.dc.b	$78,$C1,$1C,$10,$FF,$15,$3F,$33,$F0,$1F,$F7,$84,$7E,$55,$AF,$D8
	.dc.b	$6E,$DF,$5E,$BB,$CF,$DC,$FA,$9F,$F6,$DB,$FE,$EB,$6B,$FC,$8F,$14
	.dc.b	$62,$3F,$D9,$CD,$7F,$18,$D7,$FD,$98,$D8,$E3,$F1,$3F,$82,$8C,$7D
	.dc.b	$8B,$EA,$7F,$66,$DB,$FD,$73,$5F,$35,$7E,$E3,$D8,$7D,$57,$E3,$1B
	.dc.b	$3F,$E7,$78,$C1,$1C,$10,$F6,$47,$07,$DD,$F6,$5F,$82,$D3,$F3,$8D
	.dc.b	$E4,$FD,$43,$67,$FD,$D1,$5C,$6B,$F1,$FF,$52,$F9,$1F,$EA,$4D,$AF
	.dc.b	$E8,$A1,$C3,$FD,$21,$AF,$F5,$16,$BF,$E4,$A6,$BF,$E5,$D6,$BC,$E3
	.dc.b	$FB,$07,$E0,$0F,$A4,$FE,$7D,$28,$C3,$FC,$E0,$C4,$6A,$BC,$77,$EA
	.dc.b	$F8,$FC,$12,$B8,$4B,$E7,$3F,$A5,$75,$3F,$08,$FE,$0D,$5F,$E6,$08
	.dc.b	$C7,$97,$F0,$1B,$1F,$D9,$9A,$FF,$6F,$EB,$7D,$83,$2F,$E8,$6D,$9F
	.dc.b	$E8,$4D,$7F,$9F,$0F,$94,$7D,$67,$F4,$A2,$C7,$F7,$4B,$CC,$7F,$54
	.dc.b	$33,$1E,$16,$BF,$E2,$36,$BF,$E2,$26,$BF,$DA,$DB,$1D,$23,$FC,$03
	.dc.b	$D2,$F8,$5F,$ED,$91,$FC,$A1,$BB,$FF,$27,$6F,$2F,$5C,$3F,$A9,$FC
	.dc.b	$77,$1B,$FE,$9C,$D7,$F9,$8B,$6F,$F9,$83,$FC,$81,$B7,$C8,$93,$CB
	.dc.b	$B3,$FA,$FF,$5F,$F4,$C8,$FE,$76,$DA,$FE,$74,$DC,$7F,$A3,$2F,$AE
	.dc.b	$E3,$E3,$CB,$FE,$C5,$3B,$5A,$7E,$EE,$BE,$97,$C3,$C9,$FE,$F5,$AF
	.dc.b	$FB,$A9,$FF,$04,$FC,$9B,$EB,$BF,$DC,$71,$AA,$E0,$FF,$7B,$E0,$FB
	.dc.b	$B1,$FE,$19,$AE,$1F,$11,$FD,$EB,$EF,$FF,$C8,$77,$FD,$D5,$AF,$F5
	.dc.b	$86,$EF,$FE,$1A,$CB,$C5,$3F,$D7,$3E,$3F,$FD,$EE,$FF,$7D,$6B,$9F
	.dc.b	$8B,$FF,$03,$C3,$E7,$2C,$49,$E0,$EA,$3F,$C1,$BF,$50,$F4,$E9,$2B
	.dc.b	$FC,$15,$1F,$CC,$B1,$FE,$D4,$EE,$33,$3B,$7F,$B8,$C6,$F9,$3F,$CE
	.dc.b	$BF,$5D,$EA,$9F,$F3,$D3,$5C,$EF,$FE,$32,$8C,$47,$F9,$0D,$BC,$06
	.dc.b	$92,$7F,$8B,$F0,$3E,$FB,$E3,$E7,$FE,$4C,$8C,$4F,$D3,$9C,$E3,$E0
	.dc.b	$C7,$FC,$4D,$E0,$FF,$4A,$C3,$FC,$0C,$D9,$FF,$09,$36,$7F,$12,$BE
	.dc.b	$53,$FC,$6A,$E7,$9B,$F9,$47,$93,$83,$33,$FE,$3A,$F9,$17,$D3,$52
	.dc.b	$61,$F9,$0F,$95,$AF,$F8,$6A,$76,$FF,$8E,$DA,$FE,$3F,$99,$3F,$E5
	.dc.b	$4E,$0F,$24,$FF,$0B,$6B,$FC,$65,$AF,$D2,$C3,$E8,$11,$E0,$D7,$FD
	.dc.b	$1D,$C1,$F3,$7E,$6B,$E5,$12,$3F,$CC,$15,$DB,$7F,$F1,$77,$D1,$BF
	.dc.b	$19,$FF,$2C,$B6,$BF,$CA,$8D,$7F,$C6,$EF,$EB,$1B,$1C,$08,$F0,$EB
	.dc.b	$FF,$31,$FD,$3B,$E6,$69,$27,$FC,$C1,$0F,$99,$C4,$7F,$91,$23,$83
	.dc.b	$A9,$E7,$3D,$3E,$B3,$FC,$83,$46,$92,$BF,$D5,$3E,$FB,$E9,$5F,$5B
	.dc.b	$FE,$97,$8C,$36,$33,$3E,$4A,$FF,$3D,$FD,$57,$F5,$AE,$AF,$F2,$F3
	.dc.b	$67,$F8,$AB,$70,$E7,$1F,$E7,$B8,$D7,$49,$A6,$F0,$EF,$3F,$C4,$FE
	.dc.b	$BD,$F2,$BF,$E9,$E6,$C7,$F4,$F6,$BF,$D7,$5A,$FF,$BB,$DF,$EA,$3C
	.dc.b	$5F,$EA,$C6,$E5,$FE,$F0,$6C,$FF,$AE,$5E,$73,$FE,$6F,$C3,$C7,$23
	.dc.b	$DF,$A7,$F9,$BA,$33,$18,$3F,$AA,$BF,$C3,$5F,$F3,$DB,$67,$A7,$83
	.dc.b	$F2,$AB,$E5,$B2,$BE,$33,$96,$90,$E6,$E5,$EB,$F2,$1E,$32,$3E,$27
	.dc.b	$CD,$F0,$FF,$60,$91,$F4,$E3,$FC,$33,$E1,$BE,$B9,$F2,$9F,$EF,$EE
	.dc.b	$FF,$81,$6A,$D4,$E6,$3E,$24,$61,$FF,$80,$13,$F1,$F8,$9E,$DF,$E4
	.dc.b	$E7,$81,$CF,$0E,$EF,$03,$9B,$8F,$92,$5F,$8B,$0E,$D1,$8A,$61,$DB
	.dc.b	$97,$CF,$39,$9E,$D9,$78,$2F,$8E,$1C,$F1,$EC,$F9,$F2,$F3,$5F,$8C
	.dc.b	$E2,$5A,$E2,$F9,$AF,$E1,$65,$E1,$8E,$D8,$74,$74,$71,$71,$1B,$8C
	.dc.b	$4F,$02,$3F,$0E,$35,$8F,$AE,$FC,$33,$E4,$8C,$5F,$4C,$3D,$31,$8C
	.dc.b	$39,$B6,$BA,$4E,$B1,$F8,$55,$A6,$CE,$B2,$F9,$71,$8B,$E4,$47,$E4
	.dc.b	$46,$A7,$DF,$39,$6C,$E9,$EC,$60,$99,$7B,$A3,$C3,$1F,$90,$3E,$89
	.dc.b	$72,$FC,$73,$72,$FC,$4B,$9F,$A0,$7A,$FF,$E8,$2A,$46,$27,$A1,$F9
	.dc.b	$B7,$0F,$19,$E1,$1B,$C7,$E8,$37,$0F,$CE,$7D,$73,$E9,$5C,$DF,$8D
	.dc.b	$AD,$63,$FA,$46,$BD,$F1,$FC,$3D,$F1,$AE,$11,$27,$F1,$D5,$CB,$45
	.dc.b	$EE,$73,$5B,$CB,$CA,$F4,$C7,$C7,$3A,$24,$E9,$FC,$65,$73,$8C,$13
	.dc.b	$2F,$D7,$63,$E1,$9D,$46,$30,$F0,$36,$F9,$E1,$E0,$C3,$94,$7C,$1A
	.dc.b	$DB,$FB,$6B,$E6,$E4,$F2,$DF,$3A,$FD,$E6,$3F,$03,$0E,$93,$92,$57
	.dc.b	$1A,$DD,$F4,$8F,$9C,$74,$35,$3F,$62,$B9,$C6,$08,$70,$AE,$71,$A8
	.dc.b	$C4,$FF,$5E,$FA,$9F,$9B,$CB,$FD,$15,$BA,$7F,$62,$F1,$ED,$FF,$62
	.dc.b	$9B,$D7,$FD,$7E,$7E,$AC,$FF,$2C,$8C,$57,$F4,$89,$CC,$F3,$8F,$E9
	.dc.b	$7F,$15,$CB,$A4,$DF,$99,$C3,$27,$FB,$73,$96,$92,$BF,$A9,$C6,$5B
	.dc.b	$1F,$DD,$67,$C3,$1F,$DE,$9B,$87,$F7,$76,$D9,$BF,$69,$E5,$3C,$DF
	.dc.b	$13,$70,$FF,$00,$3C,$08,$FF,$66,$79,$FC,$B7,$B4,$F9,$3C,$5F,$E1
	.dc.b	$5D,$31,$5F,$E0,$99,$B9,$B6,$FC,$FD,$EE,$B4,$7F,$77,$8C,$F8,$1F
	.dc.b	$E5,$88,$C1,$FD,$55,$C5,$7F,$76,$FA,$3F,$0C,$62,$BF,$C0,$51,$F0
	.dc.b	$09,$5F,$DF,$9F,$8B,$E1,$3E,$23,$EF,$B6,$7F,$C7,$7F,$5D,$89,$91
	.dc.b	$FE,$3E,$77,$18,$C7,$FC,$97,$EB,$7E,$23,$CA,$37,$EB,$B0,$CC,$3F
	.dc.b	$CA,$6E,$5A,$7F,$B0,$3E,$7B,$FD,$EE,$DF,$4B,$F2,$BE,$BB,$A5,$F5
	.dc.b	$DF,$98,$F4,$98,$70,$D3,$FD,$89,$5A,$C7,$FD,$FC,$D8,$FE,$6D,$87
	.dc.b	$3C,$3F,$12,$DA,$DC,$62,$B8,$57,$F8,$32,$B4,$F2,$8C,$BF,$E2,$FE
	.dc.b	$AF,$F0,$0C,$7E,$AB,$B3,$D1,$87,$DA,$A5,$BE,$A1,$F5,$DB,$C2,$5F
	.dc.b	$FA,$45,$F8,$33,$A8,$C1,$FE,$95,$F8,$AE,$C7,$7F,$8D,$B1,$92,$63
	.dc.b	$FE,$9D,$8C,$BF,$93,$8F,$F2,$36,$DF,$2B,$8F,$E6,$6E,$7E,$8E,$4F
	.dc.b	$AE,$CC,$61,$F8,$C4,$BF,$9E,$DC,$3E,$33,$5F,$95,$5C,$E7,$FA,$EE
	.dc.b	$EF,$88,$FF,$4D,$47,$C7,$8C,$7A,$DF,$C5,$25,$F5,$D6,$25,$6E,$78
	.dc.b	$E1,$ED,$17,$D6,$F7,$19,$87,$C5,$FB,$EF,$C0,$D9,$D2,$FA,$EF,$E4
	.dc.b	$AE,$7B,$7F,$AC,$DF,$2D,$72,$6A,$AE,$1E,$BF,$83,$FE,$E2,$D3,$EE
	.dc.b	$F9,$1F,$CC,$23,$39,$7A,$79,$BF,$0D,$3A,$AF,$C5,$71,$3E,$B9,$EE
	.dc.b	$E7,$0E,$1C,$FF,$DC,$76,$FE,$DB,$97,$2C,$78,$E1,$C1,$FC,$A4,$66
	.dc.b	$FE,$27,$D7,$3E,$93,$FA,$1B,$6B,$EA,$3F,$EE,$26,$C6,$88,$EB,$E0
	.dc.b	$6D,$7C,$FF,$8E,$FD,$B6,$5F,$21,$B5,$FA,$97,$FD,$52,$FE,$05,$CC
	.dc.b	$FD,$A6,$F1,$1A,$70,$8E,$DD,$FF,$D6,$BE,$3B,$FA,$05,$FE,$2D,$B3
	.dc.b	$F1,$BD,$3D,$79,$DF,$AE,$7F,$81,$F0,$7F,$8C,$AF,$FC,$63,$BF,$ED
	.dc.b	$E7,$24,$8C,$CE,$EF,$F8,$6B,$5F,$E1,$0D,$7F,$D6,$94,$F8,$F7,$99
	.dc.b	$FC,$79,$B7,$FC,$1E,$9C,$DF,$3D,$FC,$24,$E9,$FC,$37,$E0,$AF,$73
	.dc.b	$F8,$DB,$6F,$D2,$D9,$FE,$42,$D7,$FC,$B1,$7F,$EF,$D9,$F0,$36,$3F
	.dc.b	$54,$78,$BE,$07,$FC,$0E,$F9,$3E,$0A,$F3,$FF,$A8,$2F,$D0,$D9,$E3
	.dc.b	$77,$93,$FC,$17,$36,$BF,$D6,$F3,$BA,$F1,$F8,$FF,$DF,$1F,$82,$D8
	.dc.b	$D4,$70,$8C,$12,$7F,$CC,$F1,$C1,$FE,$1B,$F2,$AF,$4F,$FC,$3A,$D8
	.dc.b	$FE,$1D,$EF,$3E,$2B,$35,$FE,$22,$8C,$4F,$F1,$66,$DF,$F9,$FF,$A7
	.dc.b	$FD,$8F,$7F,$E1,$09,$CC,$FF,$89,$DA,$FF,$45,$6B,$FC,$51,$B3,$CA
	.dc.b	$BF,$91,$FC,$B7,$F3,$FA,$FE,$2C,$DD,$7F,$87,$FD,$7B,$D4,$FF,$1B
	.dc.b	$CE,$6F,$FA,$94,$E5,$FE,$5A,$DB,$F4,$35,$FF,$1E,$35,$E3,$5F,$E2
	.dc.b	$A8,$C4,$7F,$2A,$6E,$06,$FC,$CB,$EC,$7F,$C9,$2D,$C3,$F9,$D7,$CA
	.dc.b	$FF,$62,$C3,$F9,$84,$E5,$FE,$65,$19,$7F,$C9,$CD,$7F,$B6,$35,$FE
	.dc.b	$76,$DC,$0D,$27,$99,$FE,$42,$75,$F6,$FF,$CD,$F0,$FE,$EC,$DB,$FE
	.dc.b	$8B,$D7,$FE,$29,$9C,$F9,$3F,$E6,$28,$FF,$17,$CF,$F9,$85,$AF,$F7
	.dc.b	$F6,$BC,$BD,$BF,$F2,$8C,$63,$2F,$EF,$F1,$82,$5F,$F6,$0F,$81,$7C
	.dc.b	$BF,$F9,$C1,$B1,$FD,$87,$E7,$BF,$9D,$3F,$E0,$D8,$CE,$3F,$D6,$67
	.dc.b	$FD,$5C,$FF,$9D,$5A,$FF,$84,$DA,$FF,$62,$6E,$06,$93,$CC,$FF,$34
	.dc.b	$BA,$F9,$FF,$E5,$E7,$F3,$0C,$3F,$8F,$72,$FF,$60,$8C,$78,$9F,$E7
	.dc.b	$09,$FE,$3D,$8F,$F7,$06,$E3,$FE,$D0,$9F,$F3,$AF,$DB,$FF,$0D,$C3
	.dc.b	$FB,$4B,$69,$6F,$A2,$7D,$67,$F3,$DB,$DF,$83,$FE,$96,$C3,$85,$FF
	.dc.b	$7E,$6D,$F0,$F9,$0D,$57,$FA,$72,$FF,$C1,$38,$FF,$83,$5B,$87,$F8
	.dc.b	$0D,$FF,$25,$C6,$7D,$4E,$04,$78,$7B,$16,$BF,$BB,$7D,$3F,$06,$C7
	.dc.b	$F9,$63,$EA,$3F,$A7,$E1,$FD,$CA,$B4,$D9,$FF,$2E,$C6,$2F,$FA,$2D
	.dc.b	$70,$7F,$A3,$C6,$6F,$FC,$93,$3F,$D6,$9F,$F2,$A3,$5E,$3E,$6F,$F9
	.dc.b	$72,$31,$1F,$E5,$0F,$A9,$FE,$40,$FF,$83,$5B,$8F,$F8,$1A,$5F,$21
	.dc.b	$FE,$70,$9C,$E3,$FD,$61,$B7,$CF,$C5,$F7,$56,$DF,$E8,$B1,$F6,$BE
	.dc.b	$2F,$F9,$9B,$4A,$E1,$3F,$CE,$3E,$AB,$FC,$21,$C3,$7D,$6F,$C1,$57
	.dc.b	$E1,$09,$3A,$70,$7E,$21,$C3,$87,$F5,$BA,$75,$EB,$FF,$59,$4F,$01
	.dc.b	$3A,$70,$D7,$FD,$01,$FA,$C3,$EB,$FF,$A6,$F7,$A3,$13,$FE,$06,$6B
	.dc.b	$FE,$8E,$6B,$CA,$78,$C6,$23,$FC,$06,$D7,$FD,$0A,$DF,$93,$7E,$DA
	.dc.b	$F8,$4F,$EB,$B1,$86,$C6,$6B,$D9,$39,$7C,$9D,$1E,$02,$38,$21,$E7
	.dc.b	$72,$D3,$FB,$5E,$6F,$AD,$FB,$6D,$8C,$C6,$08,$E0,$87,$CF,$8E,$0F
	.dc.b	$D8,$F8,$E7,$E1,$0F,$57,$FB,$CC,$61,$B1,$99,$FB,$13,$96,$C6,$5E
	.dc.b	$8D,$7F,$11,$1B,$73,$3D,$BE,$C7,$51,$FC,$42,$51,$88,$46,$27,$81
	.dc.b	$FE,$0E,$F0,$9F,$5A,$F9,$2C,$36,$FF,$D3,$AD,$7F,$C1,$0D,$7C,$BF
	.dc.b	$12,$FA,$2F,$82,$D8,$F7,$B5,$F9,$0D,$78,$C7,$02,$3C,$1F,$37,$90
	.dc.b	$F5,$29,$27,$FD,$63,$19,$7E,$6F,$DD,$3E,$BB,$E9,$B6,$33,$18,$23
	.dc.b	$82,$1F,$76,$38,$3F,$E2,$7E,$8F,$F2,$FB,$FE,$C0,$6B,$9F,$80,$F9
	.dc.b	$91,$88,$FB,$3E,$57,$F8,$7D,$FC,$0B,$5F,$C5,$35,$DF,$AF,$F8,$A8
	.dc.b	$C4,$7D,$D6,$BF,$71,$BC,$3F,$D3,$A7,$EE,$7A,$6F,$9C,$A4,$BF,$F6
	.dc.b	$BC,$65,$FC,$9F,$DD,$3E,$17,$F3,$2D,$8C,$C6,$08,$E0,$87,$E8,$E3
	.dc.b	$83,$FE,$4D,$FA,$EF,$F2,$A3,$FE,$E6,$6B,$9F,$77,$F2,$51,$88,$FC
	.dc.b	$FF,$D0,$8F,$D5,$95,$FD,$4B,$5F,$D9,$35,$E2,$F0,$23,$C1,$FD,$07
	.dc.b	$BC,$F6,$69,$27,$FC,$CC,$FE,$EB,$E2,$7F,$CE,$11,$FB,$86,$BF,$C0
	.dc.b	$5A,$FE,$D9,$AF,$EC,$DA,$FE,$C9,$BB,$1A,$63,$F0,$A7,$D7,$7F,$0A
	.dc.b	$6D,$7F,$08,$6B,$FC,$45,$AF,$F0,$6F,$C2,$EB,$FC,$0D,$AF,$F0,$26
	.dc.b	$D9,$BF,$81,$BE,$9F,$F8,$CB,$6B,$F8,$BB,$5F,$E4,$0D,$7F,$8A,$FE
	.dc.b	$BA,$F9,$6F,$E4,$2D,$8F,$E3,$ED,$7F,$94,$35,$FE,$C3,$F2,$DF,$EF
	.dc.b	$77,$FD,$42,$D7,$FA,$F3,$5D,$FA,$1F,$DD,$23,$11,$FD,$B5,$AF,$F8
	.dc.b	$C9,$A9,$CF,$4B,$F3,$1F,$E2,$53,$FE,$DC,$6A,$8D,$78,$1F,$DA,$27
	.dc.b	$3E,$27,$FB,$E7,$E2,$BF,$95,$7C,$91,$A4,$C3,$FB,$B4,$ED,$B3,$B6
	.dc.b	$CE,$DF,$F5,$A7,$D8,$76,$26,$1F,$CD,$DB,$1D,$1A,$F0,$8F,$F6,$4F
	.dc.b	$CB,$3F,$59,$FE,$AF,$3F,$E1,$18,$C7,$9D,$A9,$CC,$FF,$96,$27,$2F
	.dc.b	$F8,$35,$AF,$F8,$59,$AF,$38,$E2,$A3,$2F,$CB,$FB,$1F,$90,$47,$FC
	.dc.b	$31,$AF,$D1,$8F,$EA,$5F,$20,$F9,$FA,$4A,$F9,$31,$96,$C6,$5F,$F6
	.dc.b	$D4,$65,$B1,$FD,$77,$E9,$DF,$2B,$49,$3F,$52,$32,$D8,$CB,$FE,$E5
	.dc.b	$8C,$B6,$3F,$21,$F7,$32,$3F,$2E,$D7,$F2,$ED,$7F,$3A,$D7,$FD,$0B
	.dc.b	$F1,$AF,$B4,$FF,$14,$37,$DC,$FF,$80,$2D,$9F,$F1,$F3,$5F,$F7,$C7
	.dc.b	$D8,$7F,$27,$7F,$D1,$FC,$A8,$C4,$23,$1E,$37,$F7,$F9,$CC,$62,$7F
	.dc.b	$2B,$18,$24,$FE,$7F,$D2,$93,$F6,$3A,$11,$88,$FF,$2C,$F0,$FE,$AE
	.dc.b	$31,$1F,$E5,$F8,$C3,$C6,$3F,$71,$ED,$48,$FD,$FB,$5F,$DF,$B5,$FE
	.dc.b	$0A,$D7,$F6,$DF,$0B,$FB,$32,$3E,$57,$FC,$BB,$BF,$F4,$38,$FF,$59
	.dc.b	$DF,$E0,$BC,$8C,$92,$3F,$27,$3B,$7F,$80,$36,$78,$46,$A7,$FC,$D3
	.dc.b	$F8,$B3,$E8,$29,$27,$F3,$B1,$96,$C6,$5F,$8F,$19,$6C,$7F,$9D,$3F
	.dc.b	$14,$7D,$05,$24,$FE,$A2,$32,$FF,$0C,$FA,$8F,$ED,$EF,$FA,$37,$85
	.dc.b	$18,$8F,$C2,$B5,$FC,$23,$72,$7E,$13,$F9,$E9,$FB,$0D,$7E,$4B,$5F
	.dc.b	$E4,$8D,$8E,$3A,$7F,$A3,$23,$04,$7F,$D1,$FE,$5B,$E6,$A9,$2F,$F8
	.dc.b	$1C,$65,$FE,$4F,$F5,$6F,$B0,$FF,$2B,$36,$33,$18,$3F,$4F,$FB,$23
	.dc.b	$8E,$92,$7F,$53,$19,$7F,$88,$FC,$7B,$A7,$FD,$46,$D8,$CC,$60,$8E
	.dc.b	$08,$7E,$82,$38,$36,$36,$D3,$67,$E8,$FD,$CF,$F6,$5E,$C6,$92,$3F
	.dc.b	$CE,$9F,$2E,$E7,$A4,$8F,$E9,$71,$97,$FD,$84,$3B,$DF,$61,$EF,$F8
	.dc.b	$FF,$E9,$D3,$98,$C7,$95,$FE,$D3,$AD,$ED,$FE,$DF,$8F,$F4,$57,$AE
	.dc.b	$FB,$1F,$F6,$C6,$3C,$AF,$78,$FF,$B8,$2F,$96,$3D,$20,$77,$C7,$A3
	.dc.b	$FE,$92,$C7,$A3,$83,$47,$33,$8D,$F6,$C3,$C1,$7D,$7D,$55,$F2,$BE
	.dc.b	$07,$2A,$DE,$1D,$70,$E5,$8F,$B3,$2F,$5B,$E8,$9D,$65,$E9,$C3,$94
	.dc.b	$ED,$CC,$C7,$BD,$F4,$7D,$97,$D1,$F4,$DF,$39,$D4,$74,$36,$73,$C3
	.dc.b	$CE,$F4,$9F,$76,$3E,$D9,$C9,$31,$F4,$57,$C7,$CB,$E1,$E3,$E1,$8E
	.dc.b	$C4,$C7,$CD,$3A,$8F,$05,$72,$BE,$91,$E4,$3C,$59,$69,$F5,$BF,$6A
	.dc.b	$B9,$46,$8F,$BE,$79,$0A,$7D,$C2,$DF,$E2,$1B,$9F,$EA,$1F,$C9,$35
	.dc.b	$FC,$0B,$5F,$CE,$36,$0F,$ED,$EF,$F2,$8D,$C3,$FD,$C5,$F6,$2F,$9C
	.dc.b	$FD,$35,$6B,$87,$F0,$97,$BC,$7B,$C7,$B6,$F8,$6F,$A8,$93,$F8,$0A
	.dc.b	$E1,$9A,$F7,$39,$AD,$E1,$DE,$3E,$B4,$EA,$30,$4C,$BE,$0D,$78,$72
	.dc.b	$F3,$C6,$2F,$A1,$2B,$FC,$91,$87,$86,$B7,$3E,$A3,$55,$C6,$B7,$87
	.dc.b	$86,$30,$49,$E1,$3A,$AE,$71,$F4,$30,$F5,$D6,$F0,$F0,$65,$8A,$E3
	.dc.b	$39,$C7,$B4,$60,$87,$33,$E1,$E1,$DB,$3F,$B7,$8F,$39,$D6,$5E,$3C
	.dc.b	$7B,$75,$FE,$03,$E1,$7F,$90,$57,$F5,$FD,$51,$88,$FF,$11,$C6,$2B
	.dc.b	$F9,$3C,$E6,$7F,$CF,$9C,$5F,$DE,$67,$FC,$A6,$D8,$FF,$0D,$B5,$FE
	.dc.b	$71,$DF,$C7,$0F,$C8,$FA,$17,$13,$49,$87,$F3,$66,$DF,$E9,$2B,$8C
	.dc.b	$E4,$91,$9A,$E7,$18,$8F,$75,$7A,$79,$7F,$A4,$75,$5C,$BF,$D8,$F7
	.dc.b	$EA,$8C,$1E,$7E,$A7,$CB,$52,$4F,$F6,$E7,$A7,$E0,$5F,$ED,$38,$FF
	.dc.b	$3A,$B7,$2F,$F1,$EB,$5F,$F1,$E3,$5F,$F1,$6D,$BD,$1F,$33,$6B,$FB
	.dc.b	$EC,$60,$8E,$09,$3F,$E4,$A9,$E0,$FC,$7C,$3F,$BB,$4F,$F8,$13,$87
	.dc.b	$D8,$47,$FD,$17,$E0,$6A,$BF,$CA,$6D,$7F,$C3,$73,$F2,$63,$11,$FE
	.dc.b	$44,$FC,$73,$F8,$5D,$7F,$A5,$1B,$1F,$E5,$C6,$BF,$E5,$B6,$BC,$77
	.dc.b	$FA,$91,$88,$FF,$64,$37,$63,$7F,$26,$7C,$67,$DB,$6E,$1F,$EA,$46
	.dc.b	$DF,$F9,$B1,$AF,$F9,$AD,$AF,$F9,$71,$AF,$FB,$5D,$B6,$6F,$E6,$1F
	.dc.b	$EA,$2A,$FF,$58,$B6,$BF,$CF,$2D,$7F,$CF,$0D,$7F,$9E,$F2,$3F,$00
	.dc.b	$FD,$54,$7F,$98,$5B,$19,$AF,$F0,$9B,$67,$FA,$19,$B8,$CC,$F0,$23
	.dc.b	$E4,$F9,$AF,$CF,$59,$A4,$8D,$12,$BD,$31,$97,$F5,$DD,$9F,$B9,$7F
	.dc.b	$D4,$AD,$7F,$D4,$8D,$E7,$FF,$7C,$63,$FB,$D8,$C1,$1C,$1E,$4E,$29
	.dc.b	$87,$CA,$6D,$FB,$A5,$F1,$3F,$10,$95,$F1,$1B,$3B,$6B,$FD,$A2,$78
	.dc.b	$CF,$C8,$C1,$ED,$BC,$DF,$E2,$4D,$9D,$36,$FF,$87,$35,$FF,$42,$35
	.dc.b	$FF,$63,$B5,$FF,$63,$36,$BB,$D7,$F1,$FB,$7D,$A5,$26,$1E,$58,$CB
	.dc.b	$63,$38,$7F,$1F,$9C,$B6,$3E,$EC,$49,$FC,$9B,$67,$FC,$13,$EC,$BE
	.dc.b	$C5,$F0,$5F,$CE,$63,$0D,$8C,$DF,$E3,$A7,$2D,$8C,$D7,$46,$BF,$E1
	.dc.b	$89,$DB,$9A,$FE,$7D,$F5,$EF,$95,$A4,$C3,$EA,$46,$5B,$19,$AF,$E7
	.dc.b	$D3,$96,$C7,$F4,$9F,$80,$FF,$36,$57,$DD,$6C,$FF,$45,$6B,$FE,$BB
	.dc.b	$6B,$FE,$FE,$6B,$FE,$FD,$6E,$A6,$9A,$E6,$72,$7F,$20,$A4,$63,$AF
	.dc.b	$F5,$E4,$C3,$FD,$17,$EE,$BE,$8B,$F7,$ED,$8C,$C6,$08,$E0,$97,$FC
	.dc.b	$02,$78,$3F,$E4,$4F,$25,$F0,$6F,$B3,$FE,$E9,$18,$6C,$66,$FF,$84
	.dc.b	$CE,$5B,$F8,$57,$AE,$0D,$26,$5D,$3E,$21,$F4,$14,$97,$D2,$32,$F6
	.dc.b	$F8,$A7,$D3,$F9,$1B,$19,$8C,$1E,$8F,$C2,$BF,$C1,$DA,$FB,$5A,$FF
	.dc.b	$99,$33,$46,$23,$FD,$4B,$18,$F3,$79,$12,$FD,$F5,$BA,$56,$E7,$F8
	.dc.b	$27,$A8,$FB,$BF,$F0,$A3,$63,$31,$83,$E6,$F7,$3E,$2A,$92,$7E,$6C
	.dc.b	$65,$B1,$9C,$3E,$C4,$E5,$B1,$FE,$23,$FA,$DF,$C6,$74,$9A,$4E,$1F
	.dc.b	$BB,$F8,$3B,$ED,$A9,$32,$FB,$B1,$96,$C6,$71,$FC,$44,$E5,$B1,$FE
	.dc.b	$33,$F9,$6F,$EA,$D0,$C5,$18,$6F,$C0,$6C,$FC,$7F,$CC,$3E,$8A,$93
	.dc.b	$2F,$C7,$C6,$5B,$19,$C3,$F3,$53,$96,$C7,$F9,$17,$B7,$FA,$C5,$7F
	.dc.b	$39,$6C,$FF,$90,$5A,$FF,$AA,$30,$FD,$9F,$97,$FA,$2F,$C1,$BF,$BD
	.dc.b	$F2,$1A,$4F,$3F,$44,$C7,$F4,$B1,$97,$FA,$47,$CE,$BE,$03,$FC,$B0
	.dc.b	$D8,$CC,$60,$FD,$8F,$E0,$CF,$A3,$A4,$9F,$D8,$C6,$5B,$19,$D3,$F7
	.dc.b	$93,$96,$C7,$F9,$9E,$7F,$CB,$71,$FE,$61,$6B,$F0,$5A,$8F,$B4,$E1
	.dc.b	$FD,$0F,$D1,$3E,$6B,$F8,$3B,$67,$F8,$2B,$67,$F8,$73,$5F,$D9,$7D
	.dc.b	$1B,$EE,$3F,$87,$B6,$BF,$86,$B6,$7F,$8B,$35,$E3,$3C,$09,$3C,$1E
	.dc.b	$7F,$3F,$23,$9B,$67,$9B,$5F,$1B,$5F,$81,$C4,$FA,$6F,$EC,$F1,$FE
	.dc.b	$EC,$8C,$7A,$7F,$E8,$19,$CC,$F0,$6B,$FC,$F1,$B3,$C2,$FF,$C6,$2D
	.dc.b	$9F,$F7,$AB,$5F,$F7,$CF,$C3,$BF,$6D,$FF,$2D,$37,$DD,$FF,$54,$36
	.dc.b	$FF,$C7,$6D,$AE,$78,$7F,$A0,$DF,$17,$CF,$F9,$C9,$5E,$29,$D6,$3E
	.dc.b	$F9,$CD,$72,$7F,$C9,$AF,$E0,$FE,$1B,$FC,$92,$FE,$14,$38,$74,$57
	.dc.b	$7F,$50,$F9,$E7,$D5,$61,$B5,$FE,$54,$9D,$4E,$4E,$3E,$B3,$EA,$A9
	.dc.b	$2B,$84,$65,$F9,$7F,$1E,$FA,$2F,$35,$36,$FB,$B3,$9A,$FC,$F4,$E5
	.dc.b	$FE,$E7,$19,$7C,$13,$F7,$7E,$73,$9B,$CA,$EC,$F0,$6E,$CB,$E5,$87
	.dc.b	$58,$FC,$66,$1A,$DF,$F7,$D8,$7D,$3E,$9F,$9B,$5F,$91,$6C,$EA,$7F
	.dc.b	$0F,$F5,$8F,$BA,$EE,$D8,$F5,$E9,$F9,$A8,$C4,$7A,$8F,$CA,$5F,$0A
	.dc.b	$91,$C0,$FF,$09,$CF,$0C,$BD,$18,$78,$5F,$57,$D5,$BE,$5F,$DE,$DB
	.dc.b	$E5,$9A,$31,$1E,$13,$FC,$75,$F1,$BC,$2A,$47,$09,$DD,$FD,$38,$D9
	.dc.b	$E5,$F1,$BF,$B8,$F9,$FF,$41,$B1,$9F,$1B,$FB,$63,$EB,$5D,$FF,$EB
	.dc.b	$E8,$D5,$71,$9D,$C7,$C5,$AE,$6F,$C7,$8E,$3E,$37,$7F,$AB,$F6,$4F
	.dc.b	$FA,$0D,$AF,$E8,$CF,$EC,$0E,$1F,$E9,$8D,$7F,$96,$53,$EC,$7F,$49
	.dc.b	$1E,$D6,$BE,$36,$E5,$F9,$59,$57,$82,$31,$E3,$BD,$6F,$DA,$D2,$31
	.dc.b	$87,$F5,$29,$CC,$62,$3D,$13,$B9,$C9,$23,$81,$FE,$A1,$F4,$5F,$75
	.dc.b	$F4,$5B,$1F,$55,$AF,$CB,$BE,$D8,$7A,$78,$9F,$59,$E6,$6D,$7F,$75
	.dc.b	$8C,$11,$C1,$0F,$EF,$0F,$BB,$2F,$C3,$F4,$7C,$C6,$CF,$F8,$45,$A7
	.dc.b	$F9,$BA,$7F,$11,$E8,$BE,$B3,$E3,$B6,$BF,$D7,$CD,$7E,$2B,$5F,$F0
	.dc.b	$33,$5F,$B6,$D7,$FC,$40,$DA,$FF,$10,$CF,$D4,$FB,$57,$D4,$FE,$65
	.dc.b	$B7,$F9,$D6,$BF,$93,$BF,$16,$7F,$81,$FA,$C7,$D3,$7D,$E6,$D7,$FB
	.dc.b	$55,$AF,$DC,$6B,$FE,$1D,$6B,$F9,$26,$BF,$E3,$76,$E4,$6F,$0B,$C9
	.dc.b	$FD,$83,$6B,$F6,$8D,$7F,$55,$3E,$2C,$3F,$31,$EE,$7F,$81,$E7,$FC
	.dc.b	$CF,$E8,$7E,$5A,$31,$1F,$92,$6B,$FE,$33,$6B,$FA,$86,$BF,$E4,$D6
	.dc.b	$E2,$6F,$DB,$7E,$32,$7F,$78,$DB,$FD,$53,$78,$7A,$61,$FC,$EB,$A5
	.dc.b	$F0,$0F,$91,$FE,$BF,$18,$6C,$67,$1F,$F4,$B4,$60,$8E,$08,$7E,$BB
	.dc.b	$F1,$37,$D6,$7F,$19,$6C,$7F,$1D,$6B,$FC,$46,$BB,$3F,$C4,$FC,$9F
	.dc.b	$F7,$8B,$FC,$75,$AE,$7E,$3B,$D5,$F0,$BC,$36,$FE,$11,$F8,$EB,$E7
	.dc.b	$3F,$82,$B6,$FF,$8C,$1F,$E7,$77,$0F,$F0,$46,$BF,$E6,$5F,$A1,$FE
	.dc.b	$4F,$5F,$C4,$DB,$C1,$FC,$71,$B3,$FE,$A6,$F8,$B7,$D7,$FF,$64,$FA
	.dc.b	$BF,$E5,$F3,$FE,$2B,$6B,$FE,$30,$6B,$FE,$E4,$6F,$80,$69,$3C,$7F
	.dc.b	$EC,$EF,$0E,$87,$DB,$BD,$9F,$E8,$CD,$C3,$FA,$53,$5F,$EC,$98,$75
	.dc.b	$C3,$FA,$67,$C9,$FF,$47,$9F,$E8,$8D,$DB,$F9,$93,$76,$FE,$76,$BE
	.dc.b	$6B,$8F,$97,$2F,$FA,$C4,$ED,$FF,$52,$3B,$7F,$AE,$7D,$17,$FB,$C5
	.dc.b	$BF,$25,$FD,$76,$31,$1F,$D5,$5F,$9B,$B1,$CE,$5F,$D1,$5B,$97,$F4
	.dc.b	$E6,$D7,$FA,$CD,$FA,$BA,$7C,$BF,$1B,$EA,$12,$7F,$C0,$5F,$2E,$FB
	.dc.b	$07,$A7,$99,$F8,$FC,$7F,$C5,$B3,$FF,$0E,$37,$3F,$F4,$46,$1C,$6F
	.dc.b	$FB,$B7,$E0,$7F,$EF,$AC,$3F,$BD,$35,$FE,$E0,$DC,$39,$E5,$FD,$F7
	.dc.b	$E5,$9F,$67,$49,$7F,$E0,$38,$FD,$3D,$7F,$65,$77,$19,$9D,$BF,$DF
	.dc.b	$63,$79,$7F,$72,$FA,$67,$DC,$FF,$82,$DB,$5F,$E0,$F6,$BF,$E0,$55
	.dc.b	$F2,$1B,$99,$3B,$79,$47,$F9,$13,$D3,$7D,$93,$DF,$FF,$34,$C6,$1B
	.dc.b	$19,$BF,$D9,$CE,$5B,$19,$7F,$C1,$D1,$96,$C6,$5F,$F3,$A4,$65,$FF
	.dc.b	$3B,$62,$FC,$5F,$E2,$C7,$F9,$0E,$31,$A7,$F9,$69,$B1,$FE,$43,$AE
	.dc.b	$DA,$FF,$84,$7E,$CD,$F5,$9F,$E3,$16,$D7,$F8,$D5,$AF,$F8,$AD,$FE
	.dc.b	$85,$7F,$E2,$07,$35,$CE,$B9,$1C,$FD,$3F,$F1,$E7,$6F,$F1,$29,$FF
	.dc.b	$10,$36,$FF,$C5,$2D,$C3,$F8,$37,$D9,$F0,$F1,$A4,$7F,$91,$E7,$5D
	.dc.b	$5F,$C2,$6B,$F8,$95,$FF,$08,$25,$7F,$AD,$3A,$7F,$CA,$0F,$F1,$58
	.dc.b	$CB,$FE,$B6,$8C,$BF,$C5,$E3,$FC,$5F,$F7,$FF,$E5,$7C,$BF,$C6,$CD
	.dc.b	$8F,$F1,$FB,$67,$F8,$E4,$7F,$96,$61,$F7,$54,$93,$FE,$6A,$E4,$7C
	.dc.b	$DE,$23,$FC,$AD,$5C,$1F,$F3,$23,$5F,$E3,$73,$C1,$FF,$34,$7D,$93
	.dc.b	$ED,$BF,$CD,$0D,$8C,$C7,$FA,$63,$ED,$9F,$51,$A8,$93,$A3,$FC,$FB
	.dc.b	$3A,$7A,$FC,$A7,$F9,$42,$CD,$24,$7F,$B8,$3E,$BB,$F5,$2F,$FA,$19
	.dc.b	$AF,$F3,$26,$ED,$FE,$5F,$AE,$73,$FE,$81,$FA,$F7,$E1,$7F,$E8,$16
	.dc.b	$C6,$63,$FD,$63,$7C,$B8,$3F,$CD,$F5,$AA,$FF,$3F,$37,$0F,$F3,$63
	.dc.b	$67,$FE,$00,$D7,$1A,$78,$EA,$7D,$AF,$F9,$B9,$B3,$FD,$AA,$75,$39
	.dc.b	$35,$F0,$7E,$72,$4F,$AE,$3F,$D8,$FF,$20,$ED,$FF,$60,$B6,$7F,$BA
	.dc.b	$F4,$7F,$B2,$E7,$26,$A3,$89,$0F,$F5,$8C,$F1,$7F,$AE,$BF,$E9,$5D
	.dc.b	$3F,$DA,$4D,$D3,$FD,$C7,$3E,$2A,$E4,$F0,$2D,$70,$8F,$F5,$87,$C1
	.dc.b	$7E,$8A,$7E,$2B,$73,$FF,$23,$5F,$FA,$4A,$30,$7F,$91,$6F,$FD,$34
	.dc.b	$D8,$FF,$51,$B6,$B9,$C7,$FB,$B1,$78,$8F,$9C,$FE,$F5,$F2,$DF,$DA
	.dc.b	$67,$31,$8E,$D4,$4A,$E1,$E9,$E2,$FF,$B8,$7B,$FF,$E4,$7F,$0B,$76
	.dc.b	$9E,$97,$C2,$31,$86,$A3,$13,$AC,$7C,$14,$C3,$A3,$DE,$F8,$4E,$A3
	.dc.b	$CF,$8E,$5F,$34,$72,$AE,$31,$E0,$33,$3B,$78,$99,$2D,$FB,$3E,$0B
	.dc.b	$F9,$8B,$E5,$9E,$18,$FA,$31,$F1,$C7,$87,$0E,$F5,$BC,$3A,$5E,$DF
	.dc.b	$09,$E3,$9D,$DF,$28,$DE,$3E,$DA,$E1,$5A,$9E,$37,$D4,$97,$F5,$A7
	.dc.b	$53,$EA,$8E,$93,$A8,$F9,$98,$FB,$27,$24,$8E,$AF,$38,$D4,$60,$87
	.dc.b	$53,$9E,$5F,$80,$AE,$D1,$93,$57,$E4,$C3,$F0,$51,$FE,$FD,$6B,$F0
	.dc.b	$4C,$63,$FE,$4B,$7F,$0D,$A7,$EC,$5B,$BF,$F9,$B9,$FC,$DC,$65,$FC
	.dc.b	$B3,$5C,$C7,$D8,$8F,$E0,$58,$7E,$63,$D8,$7D,$A3,$7D,$5C,$C3,$EC
	.dc.b	$E1,$C3,$6F,$9D,$87,$0B,$D6,$3F,$9E,$AD,$5F,$8B,$0F,$C2,$5E,$E3
	.dc.b	$9C,$70,$C3,$D7,$1F,$83,$9D,$5F,$3A,$D4,$F3,$35,$8F,$9D,$E6,$49
	.dc.b	$E0,$EB,$1E,$2F,$C2,$35,$87,$AA,$B9,$57,$02,$47,$91,$F8,$D1,$A8
	.dc.b	$C1,$27,$C9,$3A,$AE,$F9,$FF,$B7,$70,$F4,$E8,$CB,$DB,$8F,$1C,$7C
	.dc.b	$98,$F8,$1F,$E0,$13,$A7,$AC,$EF,$0E,$91,$ED,$3D,$67,$F2,$8D,$3E
	.dc.b	$27,$07,$F2,$46,$E7,$FC,$93,$E1,$6D,$6E,$3F,$8E,$D7,$F2,$A9,$FF
	.dc.b	$09,$F1,$7F,$BD,$5B,$B9,$A6,$37,$E7,$3F,$EF,$B6,$CF,$FA,$2E,$FA
	.dc.b	$9F,$CD,$23,$24,$73,$87,$F4,$0F,$4F,$0B,$1F,$E7,$B5,$87,$F9,$FC
	.dc.b	$DA,$FE,$7E,$F9,$88,$FF,$69,$BF,$15,$FF,$34,$3F,$B8,$F4,$3F,$02
	.dc.b	$F4,$B6,$BF,$C0,$9E,$3E,$27,$2F,$C4,$9F,$F7,$04,$FF,$75,$6D,$75
	.dc.b	$CF,$FB,$E5,$6F,$F0,$9E,$A7,$D0,$E9,$7C,$27,$FA,$8B,$1F,$F0,$94
	.dc.b	$E9,$A7,$C4,$9D,$FA,$99,$C7,$FD,$9B,$7C,$5B,$1E,$AE,$7E,$D3,$9D
	.dc.b	$FF,$C2,$71,$E9,$26,$1E,$97,$EF,$F8,$0F,$B9,$FB,$6D,$9F,$17,$DA
	.dc.b	$FF,$72,$CB,$EA,$47,$F6,$4D,$95,$B9,$C9,$FE,$1E,$71,$8F,$F8,$C7
	.dc.b	$0F,$85,$87,$38,$C6,$1F,$E5,$0F,$73,$F1,$7E,$49,$F9,$1F,$6A,$F7
	.dc.b	$73,$12,$B8,$1F,$1E,$74,$E3,$3F,$F2,$EF,$2F,$43,$E9,$7D,$87,$FA
	.dc.b	$C2,$7F,$CB,$F1,$8F,$33,$FC,$E7,$C6,$AD,$5F,$E2,$63,$13,$A9,$CC
	.dc.b	$FF,$80,$27,$37,$FE,$2E,$9C,$CF,$F8,$2A,$72,$F3,$8C,$BF,$E4,$4E
	.dc.b	$DE,$FA,$FF,$A5,$B0,$F3,$FD,$2F,$EB,$A7,$38,$FF,$A6,$27,$06,$9F
	.dc.b	$F4,$B3,$BF,$B4,$91,$FE,$99,$F3,$AE,$3F,$5B,$91,$82,$F3,$97,$FA
	.dc.b	$9A,$EE,$1F,$E2,$23,$FD,$1B,$DF,$FC,$29,$A7,$E5,$F4,$FC,$EB,$FB
	.dc.b	$C3,$39,$7E,$13,$5F,$99,$CF,$F2,$49,$3F,$1D,$B3,$F1,$5A,$FC,$DA
	.dc.b	$EA,$FE,$BF,$ED,$1F,$37,$FC,$49,$B7,$FC,$33,$64,$7F,$A8,$23,$0E
	.dc.b	$AB,$E1,$37,$4F,$E9,$95,$C2,$FA,$7A,$3D,$5F,$E3,$FA,$FE,$73,$1F
	.dc.b	$D1,$37,$1F,$F3,$EA,$FA,$3D,$F4,$CA,$F2,$4E,$F1,$FF,$52,$D6,$6F
	.dc.b	$EA,$7B,$EF,$99,$FE,$5A,$DC,$3E,$76,$18,$9F,$CD,$B2,$3F,$95,$FC
	.dc.b	$0B,$ED,$1F,$17,$9C,$7E,$97,$0F,$FA,$D2,$F9,$9C,$2B,$6E,$27,$F0
	.dc.b	$DF,$68,$F8,$1F,$E7,$ED,$9F,$9C,$BE,$83,$33,$27,$2F,$08,$C5,$7E
	.dc.b	$47,$E2,$1E,$FF,$F4,$A6,$DF,$D9,$5E,$96,$66,$4E,$5E,$11,$8A,$FC
	.dc.b	$E7,$C4,$FC,$D3,$EA,$75,$7E,$EF,$C7,$FE,$AF,$18,$DB,$FD,$4D,$8F
	.dc.b	$ED,$E7,$F7,$84,$70,$4B,$FD,$C4,$65,$FE,$81,$F8,$7F,$F1,$07,$F5
	.dc.b	$8D,$EA,$FF,$0B,$7A,$8F,$8C,$7C,$5F,$F6,$8A,$D6,$3F,$CA,$FE,$37
	.dc.b	$F8,$1F,$59,$A4,$F1,$3F,$C4,$19,$FF,$A3,$B2,$7D,$A7,$AA,$3F,$98
	.dc.b	$C6,$35,$FE,$FE,$BE,$3D,$F6,$98,$BE,$1B,$FE,$83,$C6,$7B,$9F,$DF
	.dc.b	$1B,$B7,$F7,$6C,$5F,$24,$FB,$4D,$5F,$68,$C5,$7F,$6F,$D7,$FD,$35
	.dc.b	$18,$7F,$B9,$5F,$2B,$F5,$C7,$F9,$E6,$74,$FB,$5A,$E6,$36,$F4,$6C
	.dc.b	$F3,$AF,$F7,$47,$CB,$FF,$74,$CB,$F7,$6D,$8F,$EB,$1E,$FB,$EC,$FF
	.dc.b	$C3,$33,$9C,$7F,$DE,$B3,$9A,$FF,$79,$36,$3F,$BF,$35,$FF,$5F,$F1
	.dc.b	$A3,$11,$DA,$FF,$C5,$71,$88,$EB,$18,$3F,$C5,$FE,$EB,$EF,$BF,$C5
	.dc.b	$6D,$C3,$F8,$D7,$DA,$7D,$67,$FB,$84,$66,$FF,$9D,$4F,$E0,$5F,$1B
	.dc.b	$5F,$EE,$CD,$58,$A3,$0D,$E4,$FF,$20,$C6,$08,$FF,$91,$3E,$75,$F2
	.dc.b	$1F,$E4,$06,$E3,$FC,$BF,$E7,$7F,$AA,$E3,$DA,$72,$FF,$4F,$8F,$CE
	.dc.b	$93,$A3,$FC,$9E,$DA,$FC,$46,$48,$C5,$7F,$8B,$5F,$D2,$F7,$7F,$2D
	.dc.b	$AF,$E7,$CD,$9F,$EB,$9D,$EF,$53,$F7,$FE,$6F,$F9,$5A,$31,$7D,$9B
	.dc.b	$5F,$E3,$96,$BF,$71,$AF,$F8,$7D,$AF,$CA,$6E,$26,$98,$D7,$F9,$34
	.dc.b	$FF,$6F,$6F,$27,$F4,$EF,$9F,$7A,$9F,$E6,$A9,$CE,$3F,$E2,$D9,$CB
	.dc.b	$F1,$1B,$7F,$E4,$F6,$BF,$92,$6B,$C6,$BF,$CF,$71,$88,$FC,$0B,$71
	.dc.b	$37,$E8,$BF,$97,$DF,$F6,$96,$D7,$F8,$1E,$3F,$CD,$78,$FF,$66,$9C
	.dc.b	$BF,$DA,$23,$FC,$DF,$3F,$A4,$6B,$FE,$47,$6B,$F9,$26,$DB,$EF,$FF
	.dc.b	$91,$CF,$E5,$9B,$D2,$FA,$47,$D7,$7F,$A4,$27,$77,$C7,$DD,$C1,$FC
	.dc.b	$82,$B7,$8F,$F8,$0E,$BA,$EE,$F8,$BF,$C0,$E9,$FE,$CB,$8C,$71,$71
	.dc.b	$F9,$0F,$F3,$C5,$6B,$0F,$F1,$D3,$76,$FC,$03,$76,$F3,$57,$0B,$FF
	.dc.b	$40,$7D,$43,$E5,$29,$FE,$83,$FA,$8F,$CC,$69,$E1,$9D,$67,$FE,$81
	.dc.b	$6E,$7F,$E2,$96,$CF,$9E,$B8,$D7,$F9,$36,$72,$FF,$93,$E3,$F4,$95
	.dc.b	$FE,$55,$6B,$9E,$6F,$F2,$FC,$62,$3D,$7F,$11,$F0,$0F,$37,$D6,$DB
	.dc.b	$FF,$4A,$6D,$FC,$9F,$7F,$95,$5A,$9C,$BF,$EC,$EB,$13,$E4,$3F,$A2
	.dc.b	$38,$0E,$1D,$15,$DF,$D6,$3E,$B5,$F5,$B8,$6D,$7F,$4F,$9D,$4E,$4E
	.dc.b	$3E,$BF,$FB,$2E,$74,$D7,$FD,$89,$3F,$E1,$9E,$0D,$9F,$D6,$73,$FC
	.dc.b	$37,$D2,$79,$E3,$F0,$DA,$A3,$1A,$FE,$11,$B8,$7F,$03,$DF,$FC,$50
	.dc.b	$FC,$88,$CE,$1D,$23,$07,$FA,$FE,$76,$FF,$99,$D7,$D6,$FC,$89,$46
	.dc.b	$27,$CC,$D7,$FD,$2C,$D7,$FD,$29,$C5,$FE,$66,$7F,$19,$F0,$7E,$28
	.dc.b	$C4,$77,$BC,$4C,$8D,$9E,$58,$F4,$FC,$0B,$EB,$69,$27,$FD,$F6,$7F
	.dc.b	$90,$FE,$3F,$0F,$FA,$06,$78,$3F,$ED,$D6,$BF,$EC,$83,$81,$FE,$79
	.dc.b	$EA,$FF,$61,$37,$E3,$3E,$6F,$7A,$73,$3F,$E2,$0E,$EF,$F6,$33,$B2
	.dc.b	$1F,$E9,$F8,$DF,$17,$F8,$D2,$77,$3F,$EA,$3F,$AD,$FF,$38,$3F,$D4
	.dc.b	$9A,$FF,$BF,$5A,$FF,$80,$1A,$FF,$B8,$FC,$2F,$CB,$3F,$CA,$7A,$3C
	.dc.b	$F1,$88,$ED,$9F,$AB,$C3,$79,$9F,$A4,$D3,$F9,$9B,$63,$35,$FD,$06
	.dc.b	$30,$47,$04,$3F,$58,$7F,$7E,$F1,$9F,$5B,$F2,$1B,$1F,$EA,$56,$BF
	.dc.b	$EC,$3A,$EE,$FF,$6A,$FA,$DF,$27,$39,$A4,$8F,$ED,$4D,$7F,$B5,$35
	.dc.b	$FF,$12,$E3,$F7,$B7,$FF,$56,$B7,$A7,$F8,$2B,$FE,$14,$E4,$FB,$B8
	.dc.b	$77,$BF,$F0,$67,$CD,$FC,$5F,$2C,$D2,$47,$F8,$22,$32,$D8,$CB,$FE
	.dc.b	$09,$7E,$77,$9A,$FA,$7F,$C5,$B6,$7F,$D9,$0D,$7F,$C3,$13,$F9,$0F
	.dc.b	$AD,$FF,$4B,$C7,$F8,$59,$AF,$D0,$C9,$18,$8F,$F2,$2B,$5F,$7B,$5F
	.dc.b	$F1,$9B,$63,$8F,$C1,$FE,$0E,$31,$C4,$F9,$1F,$D0,$B6,$FF,$DB,$CD
	.dc.b	$7F,$DF,$93,$E4,$DF,$FD,$B7,$EB,$3E,$B7,$F1,$CD,$AF,$F2,$EC,$60
	.dc.b	$8E,$08,$7F,$98,$1F,$CB,$7B,$1F,$56,$15,$FE,$6E,$E3,$FD,$AC,$E6
	.dc.b	$3C,$35,$FC,$93,$B3,$FD,$13,$F2,$3F,$D0,$9B,$87,$FA,$1D,$AF,$FA
	.dc.b	$1B,$DB,$4E,$DB,$1B,$6B,$B8,$D1,$FA,$BF,$AC,$7D,$6F,$F0,$86,$CF
	.dc.b	$F5,$4C,$7F,$DF,$F1,$88,$FF,$54,$DF,$49,$FF,$5C,$F8,$FF,$EE,$A9
	.dc.b	$FF,$51,$35,$FF,$57,$B6,$3F,$D7,$FF,$05,$C7,$AB,$FD,$79,$E7,$7E
	.dc.b	$F7,$E0,$CD,$26,$1F,$E6,$0F,$9F,$FE,$2B,$5F,$A4,$6C,$7F,$9E,$1B
	.dc.b	$B7,$FB,$2F,$EB,$7F,$A5,$FD,$37,$F2,$2A,$FE,$D9,$DC,$8C,$47,$F7
	.dc.b	$28,$C7,$91,$FD,$2E,$73,$3F,$E5,$96,$CE,$A3,$FB,$14,$60,$DF,$5B
	.dc.b	$DC,$FF,$63,$CF,$F9,$CB,$E0,$FF,$9A,$C6,$23,$F0,$BA,$78,$6F,$D7
	.dc.b	$6F,$8C,$FE,$63,$C2,$AF,$E0,$71,$86,$FB,$9F,$D0,$21,$F2,$F6,$26
	.dc.b	$3F,$6E,$B6,$D3,$FD,$3F,$5C,$1C,$60,$F9,$FF,$F7,$9C,$61,$AF,$A0
	.dc.b	$FF,$53,$B8,$C7,$CD,$18,$DB,$FA,$9C,$E7,$0F,$E3,$6D,$9E,$3F,$17
	.dc.b	$AA,$FA,$FE,$2C,$9F,$37,$07,$D7,$8F,$F0,$4F,$C9,$BE,$B5,$E9,$6E
	.dc.b	$FC,$97,$B7,$8F,$0F,$F1,$7D,$7C,$B8,$71,$BF,$EC,$FF,$2D,$FE,$F7
	.dc.b	$BF,$82,$D7,$DD,$BA,$31,$7F,$E6,$3C,$3A,$6F,$FC,$DB,$EB,$7F,$D8
	.dc.b	$38,$7C,$A6,$BC,$9B,$5F,$1B,$04,$7E,$B6,$30,$EA,$BA,$7C,$3F,$F7
	.dc.b	$C8,$C5,$FF,$B5,$9E,$13,$FD,$1B,$E8,$DF,$5B,$FE,$02,$6D,$7D,$C6
	.dc.b	$BE,$C5,$F1,$3B,$99,$3B,$79,$47,$F8,$E3,$D2,$7D,$6F,$F8,$CA,$BF
	.dc.b	$CA,$D1,$8F,$96,$FF,$2C,$B6,$3F,$B7,$87,$0F,$F7,$26,$BC,$A7,$A3
	.dc.b	$FE,$64,$8C,$B6,$32,$FE,$8F,$EB,$7E,$1E,$46,$9F,$1A,$BB,$EB,$F3
	.dc.b	$FE,$B5,$F3,$1F,$E2,$66,$E1,$FA,$06,$BF,$80,$7F,$A3,$61,$FD,$C9
	.dc.b	$CD,$73,$AE,$47,$3F,$4B,$FC,$63,$DD,$FC,$6B,$0F,$D5,$B5,$FC,$5B
	.dc.b	$6B,$F1,$5F,$5B,$C3,$C5,$95,$C1,$FE,$33,$1F,$C6,$B2,$FE,$28,$49
	.dc.b	$FE,$22,$D7,$FC,$3B,$D3,$FC,$76,$31,$1F,$C8,$39,$BF,$6A,$FF,$73
	.dc.b	$FA,$2B,$8F,$E4,$DB,$5F,$E1,$06,$CF,$F2,$18,$FF,$26,$C7,$FA,$DE
	.dc.b	$3F,$83,$B5,$FD,$3B,$67,$FD,$91,$C4,$FA,$0E,$3F,$19,$2F,$8B,$FE
	.dc.b	$80,$9D,$47,$F1,$2F,$A4,$7D,$6F,$F9,$71,$B1,$98,$FF,$43,$FD,$6B
	.dc.b	$E8,$B5,$12,$74,$7F,$0A,$9D,$3D,$7E,$4B,$FC,$99,$66,$92,$3F,$D3
	.dc.b	$DF,$59,$FA,$67,$F8,$9B,$5F,$E1,$ED,$D3,$F7,$95,$CE,$7F,$95,$FD
	.dc.b	$4B,$EB,$7F,$CE,$AD,$8C,$C7,$FA,$82,$F9,$70,$7F,$90,$2B,$55,$FC
	.dc.b	$49,$B8,$7F,$09,$6C,$FF,$59,$AE,$34,$F0,$54,$FA,$DF,$F3,$53,$67
	.dc.b	$F6,$B3,$A9,$C9,$AF,$80,$F3,$92,$7D,$71,$FE,$B6,$F8,$FF,$DE,$47
	.dc.b	$FB,$0D,$AF,$E4,$9A,$FF,$70,$6B,$FD,$BD,$FF,$39,$E5,$FE,$C8,$8E
	.dc.b	$87,$5E,$DA,$FF,$BB,$0A,$7F,$3E,$9E,$F3,$C6,$34,$43,$53,$FE,$A4
	.dc.b	$F8,$1F,$D0,$C7,$FB,$B7,$1F,$F2,$46,$9F,$E8,$D9,$CC,$FF,$9C,$1A
	.dc.b	$FC,$B7,$FC,$95,$7F,$CE,$5B,$3B,$7B,$3F,$ED,$E9,$78,$6F,$99,$FF
	.dc.b	$01,$FC,$A7,$F6,$F9,$CC,$63,$BB,$FE,$00,$56,$FD,$35,$2F,$FB,$B7
	.dc.b	$AA,$F8,$1F,$E2,$F1,$CE,$F8,$56,$F0,$E7,$8F,$58,$E7,$3A,$AE,$77
	.dc.b	$CA,$F7,$8F,$82,$7A,$61,$C7,$0F,$3C,$E6,$BC,$B8,$F8,$FC,$3F,$3D
	.dc.b	$78,$4E,$F7,$C7,$2F,$05,$F6,$9D,$5F,$17,$A1,$C3,$0A,$F8,$0E,$85
	.dc.b	$C7,$D3,$8F,$81,$F3,$CF,$0C,$3B,$D7,$33,$BE,$5E,$F7,$15,$C1,$EB
	.dc.b	$3C,$30,$F2,$46,$32,$F5,$13,$0F,$3C,$EA,$F9,$13,$2F,$73,$E6,$BF
	.dc.b	$25,$F1,$7A,$5F,$2C,$7C,$91,$ED,$BE,$24,$7F,$DF,$91,$AC,$7D,$B1
	.dc.b	$B3,$D7,$3F,$52,$75,$7F,$E4,$F9,$FA,$07,$FB,$8A,$BF,$0E,$D8,$D3
	.dc.b	$F8,$6D,$3F,$6E,$FE,$69,$B3,$FE,$AF,$DF,$FB,$EC,$E5,$B1,$98,$F9
	.dc.b	$73,$FC,$1B,$2F,$CC,$FA,$EF,$AD,$FF,$23,$C7,$E6,$67,$59,$FE,$89
	.dc.b	$B7,$E3,$C7,$E7,$E1,$C7,$7F,$BD,$5A,$AF,$11,$E0,$AD,$E3,$DB,$0E
	.dc.b	$91,$C0,$93,$F9,$2C,$3A,$3C,$49,$8F,$96,$79,$E3,$C2,$31,$3A,$BF
	.dc.b	$D1,$DF,$29,$C9,$23,$C3,$3D,$F0,$F2,$3E,$D3,$91,$67,$E3,$C6,$A3
	.dc.b	$DB,$1C,$DD,$4F,$82,$75,$97,$D1,$AD,$57,$0A,$D6,$1E,$97,$C7,$3C
	.dc.b	$F6,$63,$E9,$9C,$92,$FB,$89,$DE,$3E,$0C,$FE,$BF,$0F,$FB,$CF,$5F
	.dc.b	$F4,$1F,$0F,$F2,$76,$E9,$FC,$B7,$E1,$2E,$5F,$CB,$66,$DF,$FB,$EE
	.dc.b	$7F,$99,$B6,$BF,$AB,$E0,$9C,$B6,$33,$8F,$28,$FB,$78,$FF,$37,$F5
	.dc.b	$71,$32,$BF,$D3,$F5,$A7,$FA,$04,$E7,$2F,$F4,$0C,$E5,$FE,$85,$5D
	.dc.b	$67,$57,$D6,$FC,$B1,$E4,$23,$FD,$BB,$9D,$F5,$BE,$E6,$C7,$F8,$1B
	.dc.b	$C7,$C4,$E5,$F8,$33,$FE,$E4,$9E,$51,$83,$3F,$5B,$CC,$F9,$7E,$3B
	.dc.b	$E0,$BF,$D4,$B5,$FE,$12,$9D,$34,$FF,$56,$C6,$B4,$FF,$07,$4F,$28
	.dc.b	$C6,$7F,$EB,$CB,$E9,$E9,$FA,$79,$F4,$DC,$7D,$D1,$E8,$23,$FE,$0E
	.dc.b	$F0,$1F,$55,$F7,$9B,$1F,$EB,$EF,$AD,$FE,$E7,$A7,$FA,$C2,$72,$DA
	.dc.b	$E0,$7B,$A7,$79,$7F,$8C,$23,$F1,$3E,$53,$D9,$A7,$C4,$FA,$D7,$BB
	.dc.b	$8C,$FF,$CB,$33,$72,$FF,$64,$D6,$A7,$27,$F9,$7B,$9B,$A9,$F3,$BE
	.dc.b	$A7,$FD,$53,$19,$3F,$DA,$EE,$2B,$FC,$E6,$D7,$FB,$2F,$02,$31,$1F
	.dc.b	$DF,$A3,$1E,$67,$F8,$49,$B5,$FE,$AA,$6C,$F4,$9D,$57,$28,$FF,$20
	.dc.b	$76,$F8,$76,$FF,$67,$37,$0E,$D4,$F9,$EF,$D7,$CE,$6F,$8E,$18,$3D
	.dc.b	$D3,$6F,$FD,$37,$F4,$BC,$0E,$3E,$75,$8C,$92,$FF,$D3,$F1,$95,$92
	.dc.b	$7F,$C3,$A3,$CF,$FF,$13,$3F,$E2,$08,$D4,$EE,$72,$FF,$BF,$B3,$FC
	.dc.b	$F5,$6C,$E7,$3A,$AF,$CC,$4E,$AB,$E9,$F3,$FC,$82,$5F,$CD,$9D,$4F
	.dc.b	$D6,$6B,$F2,$27,$6F,$F8,$78,$FF,$37,$7D,$6B,$E6,$7F,$8C,$B6,$FC
	.dc.b	$93,$AF,$4B,$35,$27,$3E,$12,$F8,$5F,$C6,$9F,$F2,$0C,$E6,$F8,$4E
	.dc.b	$4F,$F3,$F6,$9F,$A6,$CF,$E3,$D7,$FA,$8F,$9F,$F9,$5C,$BE,$82,$C4
	.dc.b	$C7,$C3,$3C,$5E,$EE,$DC,$C7,$F3,$4F,$79,$F2,$B4,$FD,$15,$5A,$FD
	.dc.b	$0B,$87,$E0,$FC,$03,$EB,$1F,$5B,$9D,$7E,$D5,$DD,$7E,$FD,$EE,$B1
	.dc.b	$87,$C2,$FA,$D7,$C0,$7F,$46,$6C,$7C,$A5,$F4,$39,$99,$39,$78,$46
	.dc.b	$32,$F9,$5C,$0F,$88,$7B,$FF,$D4,$9B,$87,$D6,$5E,$AE,$66,$4E,$5E
	.dc.b	$11,$8B,$EB,$F1,$3F,$35,$FF,$5A,$37,$BF,$F8,$17,$8F,$FD,$82,$31
	.dc.b	$D6,$8F,$DE,$93,$1F,$CB,$36,$FF,$86,$B5,$FC,$C3,$FD,$A3,$EB,$3F
	.dc.b	$8D,$47,$B9,$BC,$5F,$AC,$F4,$9F,$5A,$F8,$BF,$EE,$35,$AC,$7F,$23
	.dc.b	$F1,$BF,$C3,$78,$4D,$27,$89,$FE,$A8,$C5,$F5,$9E,$9B,$FD,$0C,$63
	.dc.b	$5F,$EF,$EB,$E3,$DF,$5B,$8B,$E1,$BF,$D4,$F1,$9E,$EF,$F8,$11,$B9
	.dc.b	$74,$F9,$27,$D6,$E6,$BA,$C6,$27,$F6,$D7,$FE,$17,$8C,$3F,$B9,$CD
	.dc.b	$18,$8F,$F4,$CC,$62,$FD,$93,$99,$FF,$4F,$B5,$E9,$1A,$3F,$C2,$FF
	.dc.b	$2F,$FD,$DE,$3F,$1A,$DE,$4F,$DA,$FC,$03,$EB,$7F,$C3,$F3,$9C,$BF
	.dc.b	$7F,$39,$7F,$DE,$AD,$BF,$E2,$1C,$7F,$EC,$1B,$FF,$12,$CE,$67,$FD
	.dc.b	$B8,$D7,$9C,$68,$95,$FF,$00,$3D,$D7,$D6,$FF,$8C,$1B,$1F,$C5,$3E
	.dc.b	$B7,$E9,$E7,$FC,$32,$73,$87,$EA,$27,$EB,$BF,$EE,$B6,$BF,$C6,$1A
	.dc.b	$B2,$46,$1B,$C9,$FE,$3E,$8C,$11,$FF,$20,$FC,$EB,$E5,$3F,$C8,$AD
	.dc.b	$CB,$F9,$4F,$CF,$7F,$59,$C1,$39,$C7,$F7,$B3,$F9,$92,$74,$7F,$93
	.dc.b	$1B,$7F,$86,$B4,$62,$BE,$7B,$FA,$1E,$EF,$E4,$91,$FB,$CF,$37,$FC
	.dc.b	$AF,$E0,$7F,$6B,$C3,$F9,$DD,$6B,$7F,$E6,$73,$9B,$FB,$2D,$7F,$DC
	.dc.b	$F8,$A3,$11,$FD,$01,$AF,$C9,$6C,$6F,$D3,$FF,$32,$C6,$0C,$ED,$FC
	.dc.b	$92,$BF,$8E,$B6,$FF,$A2,$FD,$07,$F6,$5C,$BF,$A6,$CE,$5F,$E9,$F1
	.dc.b	$FD,$BA,$BC,$AD,$73,$F0,$9F,$E7,$48,$C4,$7D,$D6,$E4,$69,$27,$FC
	.dc.b	$F7,$F4,$5F,$CA,$AF,$FB,$0B,$6F,$F9,$9C,$3E,$13,$FD,$07,$39,$CB
	.dc.b	$FA,$EC,$E4,$FB,$05,$AF,$D2,$B5,$F9,$AD,$7F,$B5,$FD,$67,$F1,$D7
	.dc.b	$F2,$8D,$EC,$7D,$33,$EB,$7F,$D2,$73,$BD,$4D,$38,$4E,$F0,$FE,$F7
	.dc.b	$1D,$77,$7C,$67,$E0,$73,$FF,$35,$36,$99,$FF,$AA,$F0,$7C,$8F,$F9
	.dc.b	$EA,$B5,$3F,$85,$6E,$1D,$38,$BF,$95,$DF,$0C,$3F,$C1,$F3,$96,$FD
	.dc.b	$4F,$F8,$B5,$BE,$EF,$FA,$1B,$EA,$BF,$2D,$1F,$E5,$19,$D6,$5F,$E0
	.dc.b	$DB,$46,$34,$FF,$41,$36,$BF,$CB,$17,$CA,$FF,$35,$39,$7F,$39,$1F
	.dc.b	$A8,$7F,$CC,$2D,$78,$F3,$FF,$97,$E3,$11,$EF,$F8,$9F,$84,$79,$FE
	.dc.b	$B6,$E1,$EE,$DB,$F9,$45,$6E,$3E,$0D,$AB,$73,$FE,$27,$3E,$78,$F8
	.dc.b	$FF,$E8,$AE,$1A,$FD,$46,$BB,$3F,$A5,$FD,$63,$EA,$9F,$5B,$CA,$30
	.dc.b	$D8,$CC,$F0,$6B,$FE,$49,$9F,$F0,$BE,$DC,$7D,$7F,$F3,$2C,$F6,$6B
	.dc.b	$B6,$BF,$E8,$7D,$1F,$49,$EC,$AF,$C6,$B7,$5D,$36,$BE,$9F,$B3,$78
	.dc.b	$BF,$2B,$37,$0E,$75,$F3,$63,$5B,$FF,$8C,$23,$E6,$FD,$6F,$C8,$B4
	.dc.b	$A3,$0D,$9F,$F4,$C3,$5F,$F3,$44,$FE,$33,$E1,$7C,$5C,$9F,$E6,$99
	.dc.b	$FC,$BC,$78,$2F,$2C,$8C,$9E,$23,$F3,$FF,$02,$FA,$BF,$63,$63,$31
	.dc.b	$F0,$5F,$F2,$2F,$C8,$6A,$24,$68,$FF,$3B,$9A,$3F,$CE,$BD,$FF,$EC
	.dc.b	$B6,$FC,$67,$CB,$EF,$4E,$67,$FC,$E8,$FF,$0A,$F8,$DD,$39,$E3,$FF
	.dc.b	$3A,$B7,$8F,$FC,$FA,$DA,$FE,$05,$5D,$7E,$B7,$FC,$DF,$1F,$D5,$5A
	.dc.b	$FF,$80,$9A,$FF,$4A,$6B,$FE,$88,$F0,$FF,$2C,$FF,$2E,$E8,$F4,$46
	.dc.b	$23,$FC,$C3,$3D,$31,$FE,$8B,$E7,$7E,$93,$3F,$E7,$4D,$F1,$CD,$25
	.dc.b	$FF,$69,$8C,$B6,$32,$FF,$86,$3C,$87,$D6,$FC,$96,$C7,$FA,$89,$AF
	.dc.b	$F9,$D2,$B9,$BF,$E7,$6E,$67,$D6,$79,$5B,$1F,$DE,$5A,$FF,$6E,$8D
	.dc.b	$11,$D3,$5F,$F1,$5E,$3F,$87,$7F,$D0,$6D,$CB,$FD,$56,$DE,$0F,$EE
	.dc.b	$95,$F8,$1F,$31,$F3,$7F,$1D,$B8,$7F,$82,$DA,$FF,$82,$A3,$44,$74
	.dc.b	$D7,$FC,$7B,$E7,$BE,$A3,$F1,$CD,$8F,$F6,$3B,$5F,$F0,$EC,$FE,$4F
	.dc.b	$EB,$7F,$D3,$31,$FE,$1B,$6B,$FE,$48,$6B,$FE,$10,$6B,$FE,$CE,$6B
	.dc.b	$FE,$37,$6E,$46,$98,$E0,$7C,$97,$E8,$DB,$5F,$ED,$C6,$BF,$EA,$E9
	.dc.b	$F1,$EB,$FE,$62,$FA,$DF,$C7,$7B,$06,$92,$BF,$CC,$2D,$7F,$CC,$31
	.dc.b	$B7,$27,$F8,$E7,$D9,$FC,$0D,$7F,$B8,$9B,$C1,$FE,$EA,$6C,$FF,$94
	.dc.b	$EB,$93,$FB,$8F,$6D,$F2,$5F,$D1,$3A,$FF,$D0,$AD,$8C,$CF,$FA,$32
	.dc.b	$72,$D8,$CB,$FE,$8D,$6B,$FE,$8C,$6B,$FE,$8A,$9F,$D7,$7D,$63,$EB
	.dc.b	$7F,$86,$36,$7F,$C0,$B9,$FF,$BF,$23,$11,$E7,$AF,$E7,$98,$7F,$42
	.dc.b	$F8,$DB,$E3,$FF,$BB,$AB,$FD,$62,$D7,$F8,$AF,$C2,$79,$B1,$6D,$FE
	.dc.b	$BE,$F4,$3F,$80,$7C,$29,$A4,$BF,$F3,$37,$CF,$7F,$17,$E6,$34,$FE
	.dc.b	$35,$A7,$FB,$33,$EB,$7F,$A7,$7D,$37,$F2,$5B,$FE,$E3,$DC,$8C,$47
	.dc.b	$F8,$8A,$31,$E3,$7F,$4D,$9C,$CF,$F9,$95,$B3,$A8,$FF,$5E,$B5,$FE
	.dc.b	$9B,$1C,$31,$7B,$DF,$EC,$A9,$FF,$3A,$FC,$2F,$F3,$98,$C4,$7F,$87
	.dc.b	$EB,$D9,$F1,$EF,$27,$F9,$F3,$67,$E9,$F0,$7F,$8E,$31,$4E,$65,$C3
	.dc.b	$F3,$1F,$F5,$04,$3E,$63,$9E,$12,$79,$9A,$7B,$C7,$F4,$DF,$80,$FF
	.dc.b	$7A,$4E,$B0,$FF,$32,$7C,$66,$4D,$6F,$67,$55,$C1,$F1,$E7,$E7,$8C
	.dc.b	$67,$E5,$C3,$FC,$DD,$7F,$5B,$C4,$98,$7F,$92,$B5,$FE,$1D,$7E,$BF
	.dc.b	$92,$7D,$6B,$D0,$DD,$FC,$9C,$2D,$EE,$B5,$3E,$9F,$96,$FF,$7B,$B7
	.dc.b	$E2,$FF,$B5,$46,$23,$91,$EF,$FA,$DF,$F6,$03,$7E,$27,$FB,$84,$62
	.dc.b	$3D,$A7,$FB,$35,$FD,$2E,$1E,$D7,$4D,$9E,$27,$2F,$0B,$FB,$D7,$D1
	.dc.b	$BE,$B6,$93,$0F,$A1,$0F,$8B,$C4,$78,$38,$93,$C6,$37,$5D,$2B,$FC
	.dc.b	$61,$E9,$3E,$B7,$FC,$55,$5F,$E5,$21,$19,$F7,$FF,$CA,$6D,$8D,$1A
	.dc.b	$9E,$2F,$F9,$6A,$34,$F3,$6B,$98,$DB,$FE,$8E,$8C,$BF,$E9,$0E,$77
	.dc.b	$D6,$FC,$67,$0D,$8F,$E0,$AD,$9E,$FB,$7E,$27,$EB,$5F,$31,$FE,$23
	.dc.b	$6D,$FE,$5D,$AF,$C7,$7F,$A0,$E1,$FA,$97,$35,$CE,$B9,$1C,$FE,$67
	.dc.b	$FC,$5D,$DB,$FC,$4F,$0F,$D1,$35,$FA,$4D,$AF,$D8,$78,$5F,$C3,$7E
	.dc.b	$B7,$8F,$8F,$2F,$8B,$FC,$52,$3F,$8A,$E7,$FC,$3C,$93,$FC,$33,$B7
	.dc.b	$F7,$4F,$F0,$C8,$CB,$4D,$3F,$E2,$7F,$A2,$DE,$3F,$AD,$6C,$7F,$07
	.dc.b	$6C,$FF,$1B,$8F,$F2,$54,$3E,$32,$92,$7F,$D7,$7C,$6F,$A0,$C4,$7E
	.dc.b	$46,$B8,$3F,$C1,$5A,$FF,$AE,$67,$83,$FC,$27,$E9,$1F,$5B,$FE,$59
	.dc.b	$6C,$66,$3F,$D0,$9F,$5A,$FA,$1D,$44,$9D,$1F,$EC,$E9,$D3,$D7,$E5
	.dc.b	$3F,$C9,$17,$FE,$DA,$8F,$CC,$F6,$23,$11,$FA,$18,$C4,$FF,$A2,$FE
	.dc.b	$B3,$F5,$D3,$FE,$DA,$6B,$FE,$D3,$6E,$FF,$C6,$6F,$9B,$FC,$6F,$EA
	.dc.b	$5F,$5B,$FE,$75,$6D,$FF,$29,$6B,$CF,$5D,$4C,$8D,$3F,$EA,$5C,$3B
	.dc.b	$F0,$FF,$3F,$B7,$D6,$FF,$99,$5B,$7E,$E6,$BE,$D6,$BF,$ED,$0D,$11
	.dc.b	$FE,$44,$CB,$D1,$F0,$3E,$73,$D1,$7F,$EB,$3E,$CF,$F5,$77,$C8,$7F
	.dc.b	$06,$C7,$FA,$FB,$5F,$F5,$5B,$5F,$ED,$AF,$F0,$DB,$FD,$1B,$72,$FF
	.dc.b	$64,$4F,$53,$FD,$D4,$7F,$B0,$48,$7F,$BB,$6A,$BF,$CB,$BA,$55,$CE
	.dc.b	$11,$87,$A1,$FC,$97,$E0,$BF,$52,$FC,$16,$BF,$DA,$5A,$B4,$FF,$25
	.dc.b	$E1,$FE,$91,$9F,$F2,$4E,$3F,$D5,$DB,$1C,$5F,$F5,$91,$FA,$CA,$FE
	.dc.b	$67,$F3,$7F,$DF,$FE,$57,$FB,$84,$0C,$FA,$BF,$EF,$DE,$EC,$46,$04
	.dc.b	$F0,$F5,$DF,$01,$FC,$6A,$39,$CE,$DB,$E4,$31,$F0,$61,$D2,$31,$8F
	.dc.b	$7B,$FF,$63,$E1,$CE,$B5,$7C,$27,$CD,$7C,$1F,$1C,$EA,$F9,$63,$E4
	.dc.b	$F0,$FD,$39,$78,$B1,$E5,$87,$97,$0E,$0F,$23,$C3,$5C,$23,$35,$C7
	.dc.b	$0F,$35,$71,$36,$63,$1F,$24,$EE,$3A,$1B,$C7,$C8,$EF,$1F,$1D,$70
	.dc.b	$C3,$B9,$F4,$EB,$75,$CA,$74,$FC,$2A,$E2,$7C,$59,$DD,$F8,$49,$7D
	.dc.b	$2B,$CE,$73,$AE,$38,$78,$CF,$93,$97,$D1,$9E,$D1,$97,$83,$AC,$BE
	.dc.b	$34,$68,$C9,$30,$FC,$35,$73,$8F,$C5,$1F,$DF,$72,$FB,$4D,$C3,$FC
	.dc.b	$9E,$FD,$43,$59,$FE,$E5,$FC,$0B,$5F,$F0,$6B,$67,$FD,$5F,$C3,$FE
	.dc.b	$04,$9C,$BF,$56,$7F,$83,$61,$F9,$9F,$65,$F5,$BF,$E5,$38,$FC,$BC
	.dc.b	$EB,$5F,$CB,$37,$0F,$0E,$1F,$8E,$C3,$8E,$FD,$62,$4F,$D7,$AE,$1B
	.dc.b	$FF,$A0,$B2,$ED,$8F,$4B,$DE,$5E,$07,$A4,$EB,$0E,$E7,$38,$DD,$7E
	.dc.b	$17,$2F,$0C,$E4,$97,$CD,$FC,$E6,$1E,$1C,$B8,$C7,$49,$D1,$F0,$F1
	.dc.b	$F0,$47,$C8,$39,$E1,$E7,$79,$E1,$E0,$AE,$33,$96,$CB,$0F,$03,$C8
	.dc.b	$F0,$4E,$CF,$55,$F8,$33,$FB,$33,$B3,$C5,$7E,$5C,$7D,$31,$8B,$E4
	.dc.b	$F7,$C7,$C9,$18,$E1,$FF,$78,$EB,$FE,$8F,$E1,$FE,$4C,$DC,$3F,$97
	.dc.b	$7C,$2D,$C7,$F9,$76,$FF,$C5,$E7,$FD,$18,$A7,$F9,$AE,$1F,$EF,$C6
	.dc.b	$C6,$67,$FB,$DB,$6B,$FB,$D3,$F5,$B0,$FE,$6D,$AF,$7C,$12,$F5,$B0
	.dc.b	$EF,$0F,$E8,$55,$83,$33,$63,$CB,$17,$0F,$E9,$93,$70,$FE,$88,$F9
	.dc.b	$89,$5E,$67,$D7,$D0,$FA,$DF,$23,$67,$FC,$09,$E3,$E2,$72,$F0,$24
	.dc.b	$7F,$7D,$79,$FD,$6F,$23,$E6,$74,$3E,$0B,$FD,$49,$5F,$E0,$F9,$D3
	.dc.b	$4F,$F5,$FC,$EF,$D7,$CE,$9F,$DB,$6F,$8B,$63,$CF,$CF,$DA,$73,$AF
	.dc.b	$F8,$7E,$3D,$A4,$C3,$FC,$24,$FD,$AF,$05,$F5,$5F,$4D,$B3,$E2,$FA
	.dc.b	$DF,$ED,$F9,$7F,$A6,$E7,$2D,$AE,$07,$BA,$77,$87,$F8,$C6,$3E,$F7
	.dc.b	$90,$F6,$E9,$F1,$3E,$B7,$FD,$D5,$97,$F9,$56,$32,$D3,$E1,$C6,$B1
	.dc.b	$FF,$2E,$4F,$4E,$59,$5F,$23,$EA,$BF,$D6,$4F,$F9,$7E,$3E,$8C,$62
	.dc.b	$BF,$D5,$9C,$08,$C4,$7F,$7E,$8C,$78,$FF,$E1,$26,$DF,$FA,$B9,$B3
	.dc.b	$CA,$B8,$BF,$E0,$98,$CB,$CE,$32,$FF,$86,$7A,$FC,$5A,$FF,$BB,$5B
	.dc.b	$7C,$69,$F4,$1F,$A5,$9C,$DF,$1C,$70,$7B,$26,$DF,$FA,$83,$E9,$B8
	.dc.b	$1C,$7C,$DB,$19,$25,$FF,$A9,$23,$2F,$FB,$9A,$3F,$D1,$DD,$DF,$C0
	.dc.b	$DA,$7C,$DE,$5C,$79,$FF,$C1,$6B,$FC,$81,$A7,$E7,$6B,$E3,$36,$3E
	.dc.b	$35,$7C,$B8,$C4,$7C,$58,$C1,$27,$8B,$FC,$3B,$EB,$5F,$33,$FC,$1D
	.dc.b	$B3,$FE,$37,$DC,$FF,$8E,$7C,$04,$FF,$A8,$E3,$46,$5E,$38,$7F,$BD
	.dc.b	$5B,$5C,$A3,$FC,$9D,$18,$9E,$1E,$8F,$87,$0F,$E3,$1A,$FE,$8B,$0E
	.dc.b	$18,$FF,$A8,$F9,$F3,$F4,$3A,$66,$3E,$2C,$FF,$CB,$58,$71,$8F,$42
	.dc.b	$AD,$CB,$0E,$BE,$F3,$E6,$BF,$94,$37,$0F,$D0,$61,$89,$FC,$FB,$23
	.dc.b	$C1,$F0,$6F,$AD,$7C,$6E,$72,$FB,$37,$72,$F8,$57,$BA,$C6,$1F,$37
	.dc.b	$EB,$5E,$FF,$F3,$96,$C7,$B7,$74,$3E,$83,$55,$2B,$53,$C3,$3F,$74
	.dc.b	$75,$AD,$46,$0F,$F4,$07,$C5,$3D,$EF,$E9,$0D,$AF,$02,$F4,$F3,$32
	.dc.b	$72,$F0,$8C,$57,$5C,$9F,$17,$F4,$E3,$FD,$1C,$DF,$03,$F8,$37,$8F
	.dc.b	$FD,$5E,$31,$D6,$8F,$CE,$13,$2F,$D0,$B6,$FF,$83,$35,$FE,$08,$FF
	.dc.b	$6A,$FA,$CF,$E3,$9C,$5F,$8A,$6D,$7F,$98,$7E,$B5,$F1,$9F,$D9,$E7
	.dc.b	$38,$7F,$A8,$FE,$3F,$F8,$6E,$A6,$93,$C4,$FF,$55,$FA,$EF,$AC,$F4
	.dc.b	$67,$F9,$38,$C6,$BF,$E0,$15,$F2,$2F,$AD,$C5,$F0,$CF,$E4,$78,$EF
	.dc.b	$73,$FB,$D3,$72,$E9,$F2,$AF,$8F,$CD,$75,$8C,$4F,$FA,$B2,$FF,$CF
	.dc.b	$D1,$87,$19,$A3,$11,$FE,$B2,$8C,$5F,$B2,$73,$3C,$A3,$04,$AE,$D1
	.dc.b	$FE,$C9,$F9,$67,$AD,$FE,$11,$6C,$FE,$67,$DF,$7D,$6E,$6B,$FC,$2D
	.dc.b	$18,$9E,$7C,$7F,$EC,$0A,$FF,$11,$CE,$70,$F6,$4E,$6B,$73,$99,$FF
	.dc.b	$7F,$35,$FE,$30,$D7,$FD,$BC,$D9,$FF,$18,$7D,$6F,$F8,$9F,$DE,$3F
	.dc.b	$E2,$9F,$AD,$FA,$9D,$5F,$C4,$A7,$38,$FE,$CE,$7E,$C6,$3F,$ED,$AC
	.dc.b	$11,$88,$FE,$50,$D7,$FD,$DC,$D8,$DF,$2F,$F9,$02,$30,$47,$FC,$87
	.dc.b	$F3,$8F,$93,$FF,$1E,$37,$0F,$E0,$9F,$3B,$FD,$6B,$04,$E7,$1F,$D9
	.dc.b	$4F,$E6,$89,$D3,$FE,$4E,$6D,$FD,$06,$CF,$CE,$7F,$43,$DF,$FC,$A6
	.dc.b	$7F,$89,$36,$7F,$7D,$E0,$7F,$6C,$9C,$F9,$BF,$E5,$18,$C4,$FD,$86
	.dc.b	$BF,$1D,$AF,$F4,$D6,$AC,$91,$86,$FA,$BF,$E6,$68,$C1,$9D,$7F,$90
	.dc.b	$70,$7F,$48,$6E,$3F,$D6,$7E,$7F,$FB,$64,$E7,$C9,$FF,$33,$47,$F6
	.dc.b	$FC,$BE,$D3,$5E,$1E,$B7,$F9,$D2,$31,$1E,$56,$E6,$69,$27,$FC,$F7
	.dc.b	$F4,$5F,$CA,$71,$FE,$74,$DB,$FE,$4D,$1F,$E7,$53,$3E,$3F,$F9,$E5
	.dc.b	$C4,$7E,$81,$AF,$E3,$DA,$FF,$75,$6B,$F2,$7E,$B3,$F8,$DE,$27,$F8
	.dc.b	$EF,$73,$EB,$1F,$5B,$FE,$87,$9D,$93,$E7,$7F,$8C,$4E,$B1,$FE,$A5
	.dc.b	$3F,$C9,$B8,$5F,$19,$F7,$B8,$7F,$BF,$46,$38,$F8,$FC,$8F,$F9,$DA
	.dc.b	$B5,$87,$F8,$ED,$BB,$7D,$E6,$EB,$E5,$AE,$17,$FE,$7E,$FA,$97,$BD
	.dc.b	$4F,$F4,$0F,$D5,$7E,$4F,$4F,$0C,$EB,$3F,$F3,$FB,$73,$FE,$C1,$A2
	.dc.b	$31,$5E,$8B,$E5,$7F,$E4,$F9,$CB,$FE,$51,$8F,$D0,$5F,$F9,$6D,$AE
	.dc.b	$7A,$3F,$CB,$91,$88,$F6,$78,$7F,$04,$F3,$FD,$6D,$BF,$F4,$9E,$FF
	.dc.b	$C9,$AB,$CA,$DC,$BC,$95,$B9,$FE,$F0,$09,$F1,$FF,$CC,$9C,$07,$0E
	.dc.b	$8A,$EF,$EB,$1F,$54,$FA,$DC,$36,$BF,$9E,$CE,$A7,$27,$1F,$63,$FC
	.dc.b	$55,$3A,$6B,$FD,$46,$7F,$C2,$D8,$EC,$FF,$3F,$7E,$DA,$F9,$CF,$47
	.dc.b	$12,$31,$7A,$9C,$C6,$23,$FD,$1A,$D7,$FD,$18,$D7,$FD,$0C,$D7,$F8
	.dc.b	$4B,$79,$BF,$C1,$9D,$5F,$1D,$B3,$FD,$69,$AF,$F6,$6F,$07,$F4,$AF
	.dc.b	$A3,$BF,$C6,$47,$04,$3C,$AE,$5A,$7F,$66,$D5,$F9,$13,$E2,$3E,$C4
	.dc.b	$61,$B1,$99,$E0,$D7,$FB,$E4,$ED,$CC,$F0,$23,$EB,$ED,$FF,$23,$37
	.dc.b	$83,$E0,$46,$23,$FC,$06,$D7,$FC,$04,$D7,$F1,$D9,$7F,$07,$9E,$24
	.dc.b	$8F,$91,$F6,$3F,$E5,$86,$F8,$9F,$36,$31,$1F,$E1,$06,$BF,$C6,$79
	.dc.b	$FF,$31,$E0,$FF,$A3,$9A,$7F,$A1,$E3,$F9,$03,$F5,$F8,$5F,$5F,$49
	.dc.b	$3F,$66,$32,$D8,$CB,$FE,$20,$8C,$B6,$3C,$3E,$2B,$E5,$69,$27,$FC
	.dc.b	$27,$19,$7F,$5D,$F1,$5F,$A8,$6F,$4F,$EB,$A3,$11,$FE,$24,$6B,$FE
	.dc.b	$33,$6B,$FB,$66,$BF,$BB,$6B,$FE,$AE,$8F,$D2,$FC,$FF,$FB,$F6,$3F
	.dc.b	$DC,$CD,$7F,$DD,$ED,$7F,$DD,$CD,$7B,$7C,$FF,$94,$8F,$A9,$AF,$9D
	.dc.b	$AF,$9D,$FE,$19,$CC,$FA,$2F,$B3,$3F,$C4,$63,$1C,$1A,$9C,$CF,$F9
	.dc.b	$E5,$BC,$4F,$D4,$BF,$89,$B6,$7F,$65,$18,$23,$82,$1F,$B5,$7F,$8D
	.dc.b	$F0,$BE,$B7,$FD,$29,$3F,$E4,$86,$BF,$E4,$66,$BF,$E3,$56,$BF,$9C
	.dc.b	$FC,$3B,$FD,$C2,$FF,$01,$6B,$FE,$66,$6B,$FE,$65,$6B,$C7,$E3,$FF
	.dc.b	$85,$46,$21,$82,$30,$D9,$FC,$FF,$AF,$FC,$32,$BF,$CB,$6D,$7F,$CB
	.dc.b	$3D,$BA,$3F,$8B,$7E,$CC,$F9,$9A,$49,$FE,$31,$19,$6C,$65,$FF,$3F
	.dc.b	$C6,$5B,$1F,$C8,$7C,$57,$D3,$D2,$4F,$F2,$38,$CB,$FC,$97,$E0,$DF
	.dc.b	$61,$FC,$A5,$B1,$98,$C1,$9F,$AE,$91,$C5,$AF,$FA,$49,$FA,$6F,$C8
	.dc.b	$6E,$A6,$9A,$87,$D1,$7F,$98,$DF,$EC,$0D,$7F,$D2,$6D,$7F,$D4,$4D
	.dc.b	$7F,$B2,$B5,$FE,$D4,$D7,$AF,$B7,$FD,$A2,$35,$E4,$3F,$3B,$FF,$4E
	.dc.b	$D7,$FA,$82,$31,$E7,$7F,$5A,$9C,$CF,$FA,$C6,$72,$FF,$A7,$3E,$39
	.dc.b	$F6,$9F,$C8,$1B,$19,$8C,$11,$C1,$0F,$F5,$0C,$70,$7F,$D4,$7E,$0B
	.dc.b	$EB,$3F,$95,$B6,$3F,$D5,$6D,$77,$1A,$23,$A6,$BF,$EC,$6F,$0B,$FD
	.dc.b	$CA,$FF,$65,$6B,$9F,$8D,$FE,$D5,$18,$8F,$F6,$53,$5F,$E9,$2D,$7F
	.dc.b	$4A,$DD,$FF,$CD,$13,$F2,$7D,$79,$5F,$37,$44,$62,$3F,$D8,$3F,$5B
	.dc.b	$FC,$E9,$FF,$58,$35,$FF,$12,$76,$A3,$11,$FE,$8B,$6C,$67,$C6,$FE
	.dc.b	$A3,$18,$23,$FD,$53,$85,$F4,$74,$97,$FE,$D6,$8C,$B6,$32,$FF,$B5
	.dc.b	$A3,$2D,$8F,$C6,$6B,$23,$F3,$2D,$7F,$32,$D7,$F3,$2D,$7F,$C8,$1C
	.dc.b	$3F,$EC,$3C,$BF,$C5,$3C,$7F,$1A,$7F,$28,$D7,$F2,$4D,$7F,$DA,$DF
	.dc.b	$58,$F6,$FF,$B8,$B6,$3F,$DE,$2D,$7F,$DD,$ED,$78,$C7,$02,$3C,$1F
	.dc.b	$D1,$7D,$5C,$8F,$B4,$D7,$F4,$9B,$A3,$11,$FE,$58,$C7,$F5,$D1,$88
	.dc.b	$FF,$2F,$7C,$67,$EC,$7E,$97,$FC,$BC,$D8,$FF,$2F,$9A,$B4,$E9,$B1
	.dc.b	$A1,$3C,$5E,$2D,$7F,$CD,$2D,$7A,$4F,$CE,$E9,$91,$F3,$9A,$F8,$B8
	.dc.b	$9F,$46,$F8,$AF,$F0,$EB,$67,$C5,$3A,$9C,$92,$33,$3B,$7A,$35,$FF
	.dc.b	$3E,$CF,$17,$9F,$D2,$FF,$70,$83,$49,$E5,$7F,$8B,$9B,$3C,$23,$07
	.dc.b	$F9,$CB,$F0,$27,$CD,$52,$4F,$F1,$F8,$CB,$E0,$FB,$EE,$84,$70,$7F
	.dc.b	$25,$FA,$37,$C7,$FF,$9F,$9B,$19,$8C,$11,$C1,$0F,$E5,$51,$C1,$FE
	.dc.b	$59,$F8,$07,$F9,$8D,$FC,$43,$5F,$D2,$B5,$71,$A3,$1A,$3E,$A7,$F4
	.dc.b	$95,$FC,$21,$AA,$35,$D1,$FE,$56,$9C,$DB,$F0,$1F,$F5,$6B,$6F,$FD
	.dc.b	$37,$18,$3F,$C2,$FF,$4F,$FE,$5B,$8F,$B2,$D7,$F8,$BB,$5F,$03,$70
	.dc.b	$34,$9E,$D7,$F9,$A1,$B3,$F8,$C6,$BF,$93,$7A,$7A,$AF,$C0,$BF,$C6
	.dc.b	$AD,$E2,$FF,$1E,$C6,$0F,$F1,$CE,$5F,$EF,$CF,$A3,$FF,$5A,$FC,$41
	.dc.b	$A4,$C7,$FB,$8F,$C2,$7E,$B1,$FF,$68,$7C,$17,$F6,$E8,$C7,$C4,$7F
	.dc.b	$A6,$7C,$54,$62,$7F,$D1,$BE,$B7,$F8,$E2,$B5,$39,$BF,$EE,$AD,$7F
	.dc.b	$C0,$2E,$7C,$A5,$71,$BF,$07,$81,$FE,$E4,$6D,$F9,$0E,$71,$97,$FC
	.dc.b	$6F,$3C,$5C,$9B,$3B,$C6,$DA,$73,$AF,$0F,$C7,$FF,$BC,$31,$F4,$D6
	.dc.b	$F2,$F7,$B7,$2F,$46,$3D,$F1,$E7,$7C,$1F,$05,$F0,$7B,$4E,$E3,$13
	.dc.b	$B2,$65,$F2,$B1,$F1,$3F,$02,$33,$7E,$C7,$FD,$99,$97,$A2,$3C,$99
	.dc.b	$47,$C3,$97,$9B,$0E,$71,$C7,$0E,$84,$7E,$4C,$69,$F5,$C6,$5A,$68
	.dc.b	$ED,$87,$91,$CE,$1E,$37,$C3,$87,$E3,$08,$FF,$8F,$30,$FC,$73,$FE
	.dc.b	$3C,$E1,$70,$FF,$90,$9B,$1F,$E4,$26,$BB,$8D,$1F,$56,$37,$87,$C4
	.dc.b	$FA,$D7,$D6,$B9,$35,$73,$0F,$AB,$87,$0D,$3F,$DC,$B8,$70,$BD,$63
	.dc.b	$F9,$FA,$D5,$FF,$93,$9B,$A2,$F8,$4F,$4C,$78,$E3,$FE,$56,$BD,$D6
	.dc.b	$B0,$E9,$5A,$BE,$8F,$E6,$F0,$E9,$97,$A6,$7A,$12,$B8,$BF,$99,$BE
	.dc.b	$98,$F7,$C3,$A4,$FF,$9B,$B1,$F0,$E1,$CF,$0E,$24,$BE,$15,$CE,$75
	.dc.b	$18,$9D,$4F,$98,$D6,$5D,$B6,$FE,$3B,$8F,$AF,$3F,$4C,$E7,$2F,$93
	.dc.b	$7E,$3A,$F7,$67,$E8,$8E,$98,$F6,$74,$76,$3E,$06,$5F,$CB,$BE,$0F
	.dc.b	$F9,$87,$37,$F2,$C6,$E2,$69,$FC,$D7,$91,$F7,$8F,$47,$0D,$D7,$FD
	.dc.b	$04,$DC,$BF,$D0,$0D,$7F,$9F,$F7,$76,$24,$FF,$37,$6C,$7F,$37,$6B
	.dc.b	$C2,$34,$47,$4D,$7F,$C5,$1D,$5F,$CD,$63,$FD,$28,$D8,$FF,$49,$35
	.dc.b	$DF,$3F,$C2,$EB,$7C,$DD,$25,$FF,$85,$A3,$2F,$FA,$7F,$B9,$F0,$5F
	.dc.b	$15,$B1,$98,$C1,$F2,$7C,$0F,$85,$8F,$F9,$61,$AF,$F7,$46,$EC,$69
	.dc.b	$36,$E4,$4B,$F2,$57,$A4,$8E,$0F,$F0,$B7,$5F,$D9,$F0,$8D,$24,$7F
	.dc.b	$8A,$3D,$9F,$C1,$F8,$A6,$92,$34,$4B,$FF,$1F,$4E,$E3,$11,$FE,$34
	.dc.b	$F1,$9F,$87,$FE,$39,$B3,$FE,$3C,$8C,$11,$C1,$27,$FC,$7F,$1F,$91
	.dc.b	$F4,$7F,$E0,$04,$FF,$B3,$DB,$3F,$EC,$D6,$BF,$E6,$96,$BF,$E6,$8F
	.dc.b	$4B,$FD,$33,$3F,$E6,$1F,$09,$18,$9F,$F3,$0B,$5F,$F3,$0B,$5E,$55
	.dc.b	$FE,$17,$9E,$2D,$9D,$4F,$F9,$B1,$BB,$BC,$87,$D5,$FE,$AD,$B8,$FF
	.dc.b	$BA,$5B,$7F,$EE,$76,$BF,$E8,$36,$BF,$EA,$06,$BF,$E3,$C6,$D7,$7A
	.dc.b	$FD,$EF,$6F,$F4,$4B,$FF,$37,$B6,$FF,$CE,$CD,$7F,$CD,$6D,$7F,$D9
	.dc.b	$2D,$7F,$C8,$4D,$ED,$34,$D5,$FF,$B1,$BC,$3F,$B0,$4B,$FB,$0F,$F8
	.dc.b	$F7,$A5,$E6,$52,$4F,$FA,$57,$1F,$F1,$74,$61,$FF,$4C,$4F,$09,$D3
	.dc.b	$63,$33,$FC,$73,$E1,$FF,$03,$1F,$E5,$96,$E4,$69,$3D,$EE,$04,$C3
	.dc.b	$FD,$2B,$1C,$30,$FC,$F5,$FF,$A7,$5A,$FF,$A4,$9A,$FF,$BC,$1A,$FF
	.dc.b	$9A,$5B,$7F,$E4,$9E,$4F,$E6,$31,$83,$79,$FE,$1F,$4F,$C3,$13,$1F
	.dc.b	$C3,$3F,$E6,$7F,$17,$FC,$E9,$F5,$97,$13,$49,$5F,$EB,$F9,$DB,$67
	.dc.b	$6D,$9D,$E5,$FE,$B9,$FA,$8F,$E3,$6D,$EC,$FE,$8F,$18,$8F,$F5,$8B
	.dc.b	$51,$FE,$87,$70,$D8,$E2,$4A,$FF,$60,$FB,$6F,$9E,$FE,$A6,$D8,$CC
	.dc.b	$60,$8E,$09,$5F,$EC,$F9,$E0,$FE,$A7,$E6,$FF,$5C,$49,$FD,$73,$F9
	.dc.b	$1F,$80,$91,$F8,$A6,$CF,$E3,$9A,$FF,$BB,$37,$46,$23,$DF,$AF,$E6
	.dc.b	$BD,$8F,$A9,$C8,$69,$30,$FA,$5F,$02,$F5,$3E,$D3,$63,$31,$83,$3F
	.dc.b	$61,$23,$8B,$71,$FF,$7F,$E3,$FC,$DF,$B3,$F0,$9C,$DF,$C4,$A3,$17
	.dc.b	$F8,$18,$C7,$C2,$7E,$02,$F8,$B5,$FB,$22,$91,$98,$F7,$B5,$D3,$63
	.dc.b	$7C,$9F,$95,$8C,$1B,$E5,$FF,$4A,$FA,$FF,$E1,$0F,$53,$FD,$DB,$97
	.dc.b	$83,$1F,$C4,$D6,$A3,$11,$FE,$D1,$9D,$BF,$03,$E1,$DF,$6F,$FA,$36
	.dc.b	$D7,$E7,$DB,$3B,$9D,$12,$74,$D8,$F1,$5B,$E8,$3F,$37,$5E,$28,$C4
	.dc.b	$FF,$8A,$27,$2F,$06,$BF,$2B,$47,$C5,$BE,$8F,$F7,$0D,$C3,$F6,$8D
	.dc.b	$AF,$B6,$D7,$93,$5F,$C7,$B5,$FA,$CD,$AE,$F5,$FE,$37,$FA,$9F,$F0
	.dc.b	$55,$FC,$AE,$F4,$63,$0F,$A1,$18,$F3,$78,$1F,$CD,$6E,$5F,$F2,$E6
	.dc.b	$D7,$F3,$26,$BF,$A5,$6B,$FC,$0A,$BB,$FB,$1E,$32,$57,$8F,$0F,$E9
	.dc.b	$94,$F8,$5F,$EC,$D3,$FD,$05,$B1,$F8,$26,$BF,$87,$6B,$F9,$86,$BF
	.dc.b	$C4,$9A,$F5,$F1,$FF,$D0,$4F,$17,$FC,$C3,$F0,$EF,$62,$92,$FF,$01
	.dc.b	$19,$7F,$95,$7D,$EB,$C1,$FE,$64,$D8,$CC,$60,$8E,$09,$87,$EE,$27
	.dc.b	$83,$F9,$08,$7A,$FF,$E7,$46,$9F,$91,$2C,$6E,$3F,$A1,$7C,$93,$E4
	.dc.b	$A9,$2B,$FA,$04,$65,$FE,$85,$F2,$9F,$D5,$38,$CD,$27,$6F,$F4,$D6
	.dc.b	$CF,$08,$C1,$F2,$B1,$FE,$7F,$F2,$AF,$DB,$3F,$C7,$B1,$FB,$B8,$C7
	.dc.b	$A7,$B8,$46,$2F,$F7,$AD,$8E,$15,$C9,$FE,$27,$19,$6C,$65,$F0,$B5
	.dc.b	$72,$BE,$61,$F3,$DF,$E9,$96,$E1,$FA,$36,$D7,$E8,$9A,$FF,$22,$AE
	.dc.b	$55,$C4,$95,$C5,$FE,$E5,$1F,$E8,$49,$FE,$66,$D7,$F8,$13,$5F,$E8
	.dc.b	$ED,$7C,$F5,$FC,$AE,$3D,$1E,$17,$9C,$97,$E7,$6D,$79,$DA,$FF,$89
	.dc.b	$FE,$07,$FC,$50,$FF,$72,$6A,$69,$FE,$2D,$F9,$BF,$C4,$67,$78,$7E
	.dc.b	$AD,$7F,$30,$9C,$B6,$32,$FF,$37,$87,$CC,$BE,$6F,$A3,$67,$93,$5E
	.dc.b	$0D,$7F,$83,$7C,$43,$C9,$FE,$A3,$5F,$EF,$7D,$53,$9A,$FF,$06,$CE
	.dc.b	$63,$11,$E1,$D7,$F8,$86,$DF,$ED,$98,$DC,$76,$C3,$AB,$63,$27,$E2
	.dc.b	$2B,$9D,$F8,$6F,$94,$F5,$3E,$BF,$C4,$7F,$95,$71,$FC,$73,$64,$D3
	.dc.b	$5F,$3D,$FE,$34,$AE,$F4,$8C,$47,$58,$C7,$07,$9A,$3B,$9C,$27,$78
	.dc.b	$7C,$56,$E5,$FC,$BD,$B3,$C0,$F4,$78,$BF,$B0,$F9,$6E,$8D,$8F,$82
	.dc.b	$D7,$DD,$C2,$F9,$CF,$F3,$EC,$6A,$15,$C9,$B1,$98,$F4,$47,$1F,$63
	.dc.b	$BF,$53,$E5,$69,$2F,$E8,$C7,$F1,$2E,$3C,$47,$A2,$38,$3E,$66,$BF
	.dc.b	$4D,$E0,$FA,$B4,$FE,$1C,$B3,$46,$1A,$FF,$A3,$5B,$1F,$DF,$7E,$87
	.dc.b	$FD,$C0,$FC,$46,$BF,$29,$BB,$74,$8E,$D5,$FD,$32,$31,$4F,$05,$F0
	.dc.b	$BD,$DB,$7F,$D3,$23,$04,$70,$43,$FC,$27,$1C,$1F,$F0,$AF,$8F,$F5
	.dc.b	$1F,$BA,$DB,$F0,$B7,$0F,$F1,$5D,$71,$9F,$D6,$7C,$EB,$EC,$7F,$34
	.dc.b	$DC,$3F,$C0,$CD,$7F,$C0,$AD,$7F,$92,$35,$FE,$FC,$D7,$F9,$8B,$79
	.dc.b	$4E,$73,$7D,$1B,$E1,$3F,$26,$DB,$FC,$9B,$5F,$D2,$57,$8F,$1F,$A1
	.dc.b	$A3,$E7,$DF,$27,$FE,$C7,$8C,$36,$33,$7F,$E4,$A9,$CB,$63,$2F,$F9
	.dc.b	$46,$32,$D8,$CB,$FE,$64,$8C,$BF,$E6,$6D,$9F,$6E,$F9,$AF,$DC,$46
	.dc.b	$1B,$19,$9F,$E1,$F5,$DB,$1F,$C3,$FD,$03,$E3,$3F,$28,$DA,$FF,$28
	.dc.b	$35,$FF,$2A,$B5,$FF,$3B,$EE,$8C,$47,$F9,$A9,$AF,$E8,$DB,$1B,$9F
	.dc.b	$D1,$79,$7F,$C6,$2B,$F8,$1B,$5F,$E1,$2D,$BF,$3C,$FF,$0A,$F6,$1F
	.dc.b	$3F,$FA,$B6,$DF,$F9,$9D,$AF,$F9,$B1,$AF,$A1,$AF,$FA,$01,$AF,$F9
	.dc.b	$FE,$3A,$1F,$B7,$FA,$77,$C0,$FF,$1A,$6C,$FF,$1A,$6B,$FC,$82,$BC
	.dc.b	$38,$7E,$DF,$A9,$E6,$FF,$18,$AF,$F5,$4B,$5F,$F3,$C3,$5F,$F5,$EB
	.dc.b	$5F,$F5,$23,$63,$9F,$81,$FE,$96,$9C,$B6,$33,$1F,$EF,$1F,$A8,$FF
	.dc.b	$64,$BF,$C8,$1A,$FF,$1B,$6D,$F9,$EF,$CC,$FA,$7E,$07,$DA,$DC,$FF
	.dc.b	$A3,$60,$9C,$F2,$7F,$B5,$1E,$6F,$FA,$B3,$EA,$DF,$57,$FC,$85,$B9
	.dc.b	$FF,$A7,$5A,$FF,$A8,$9A,$FE,$3B,$E6,$7F,$98,$E5,$FC,$79,$B7,$FE
	.dc.b	$B8,$6C,$FF,$B2,$72,$FC,$C5,$7F,$7B,$F9,$E7,$D1,$FF,$69,$6D,$7F
	.dc.b	$A5,$5A,$FF,$A5,$1A,$FF,$8C,$DA,$FF,$A3,$5A,$FF,$90,$1B,$DE,$73
	.dc.b	$CC,$FB,$CF,$8B,$5F,$D1,$5B,$19,$C7,$FB,$85,$F1,$C3,$FB,$3F,$C5
	.dc.b	$BE,$36,$93,$1F,$ED,$30,$FA,$4C,$47,$F7,$08,$E0,$EA,$79,$CF,$F7
	.dc.b	$2F,$1B,$F8,$24,$FF,$76,$6E,$BF,$D2,$9B,$BF,$FB,$91,$7D,$27,$1F
	.dc.b	$02,$5F,$F7,$A9,$DB,$FD,$85,$DB,$FE,$10,$E0,$7C,$5F,$B9,$B1,$98
	.dc.b	$FC,$5F,$91,$F6,$B7,$F9,$51,$B2,$4F,$F7,$89,$FB,$FC,$2F,$B3,$FB
	.dc.b	$3A,$FF,$62,$6E,$BF,$EF,$C6,$DF,$D7,$C3,$FC,$27,$F1,$FF,$CF,$EF
	.dc.b	$FB,$4B,$6D,$5F,$DE,$E3,$0F,$F7,$DF,$83,$7C,$B7,$2F,$2E,$4F,$F8
	.dc.b	$72,$B7,$9F,$FB,$36,$BF,$C4,$3F,$40,$FB,$3A,$4B,$FF,$11,$C3,$E4
	.dc.b	$31,$1F,$E2,$A8,$E0,$EA,$79,$CF,$E2,$38,$FE,$5C,$FF,$BB,$1A,$FF
	.dc.b	$A2,$DA,$FF,$B9,$5A,$FF,$A6,$1B,$C3,$FD,$2F,$7F,$F2,$74,$60,$DE
	.dc.b	$4F,$A6,$FC,$6D,$FF,$8F,$9B,$C1,$FE,$12,$6F,$07,$7C,$3F,$C8,$30
	.dc.b	$F8,$8A,$4B,$FF,$21,$C3,$E4,$B1,$1D,$66,$E1,$D6,$7F,$C9,$7E,$B7
	.dc.b	$F0,$88,$FF,$10,$B7,$1F,$B3,$92,$31,$87,$ED,$FE,$9B,$8F,$85,$23
	.dc.b	$FC,$A7,$3A,$DB,$F7,$55,$F3,$7B,$FE,$19,$2B,$F0,$A7,$CD,$70,$FD
	.dc.b	$16,$B9,$36,$79,$3E,$6F,$FD,$DE,$FF,$98,$9A,$FF,$9A,$1B,$9F,$FA
	.dc.b	$4F,$83,$F8,$3C,$3E,$CF,$FC,$D6,$DA,$E5,$18,$FA,$6C,$B2,$32,$77
	.dc.b	$8F,$F3,$9F,$CF,$7F,$AA,$A3,$FC,$E6,$D7,$FC,$F2,$DC,$3F,$C9,$AF
	.dc.b	$F8,$61,$BE,$D7,$22,$4F,$2F,$5B,$AF,$1F,$F9,$7E,$FF,$01,$3F,$08
	.dc.b	$FF,$33,$38,$7E,$2F,$0A,$31,$1F,$EE,$3F,$91,$F8,$7C,$7F,$B4,$E1
	.dc.b	$46,$27,$FD,$E2,$DB,$E5,$87,$FA,$43,$E8,$77,$5F,$E9,$06,$BF,$E9
	.dc.b	$66,$E9,$FE,$C1,$BE,$51,$8C,$7F,$16,$DA,$FF,$3F,$37,$1F,$1F,$A9
	.dc.b	$DF,$A3,$FC,$FD,$3F,$A8,$23,$8F,$AD,$F2,$12,$FC,$91,$FE,$A6,$F9
	.dc.b	$1F,$F4,$41,$31,$FF,$44,$38,$34,$49,$FF,$55,$BF,$E8,$38,$FD,$0A
	.dc.b	$35,$1F,$EB,$47,$0F,$08,$CE,$7F,$E8,$A9,$CC,$ED,$F2,$4F,$F5,$6F
	.dc.b	$86,$78,$1F,$ED,$36,$D7,$DD,$D3,$FD,$D3,$39,$35,$F2,$7C,$49,$3F
	.dc.b	$ED,$4A,$E2,$FF,$B7,$C7,$81,$FE,$DC,$F0,$BF,$99,$FC,$5F,$FA,$46
	.dc.b	$31,$28,$C7,$C3,$FF,$9E,$DF,$F6,$4F,$A6,$F5,$FF,$CB,$51,$AC,$39
	.dc.b	$F8,$1F,$BA,$C7,$9C,$E7,$1E,$93,$FD,$0B,$0E,$5E,$D7,$F0,$0A,$D1
	.dc.b	$27,$96,$5E,$17,$86,$5E,$07,$39,$78,$31,$FE,$9B,$97,$82,$73,$1E
	.dc.b	$01,$97,$81,$E0,$7C,$4C,$3B,$3C,$9A,$E6,$FB,$4F,$18,$E8,$64,$B7
	.dc.b	$E1,$79,$4F,$0C,$3D,$11,$83,$77,$DA,$B8,$5F,$5C,$7E,$15,$6D,$B1
	.dc.b	$9C,$7D,$B1,$A9,$F1,$3F,$5B,$0E,$F1,$89,$D9,$1F,$87,$1A,$BF,$44
	.dc.b	$6C,$F0,$4F,$D0,$CB,$EB,$3C,$63,$46,$F0,$F2,$3E,$D3,$F9,$AD,$7E
	.dc.b	$31,$B8,$FF,$B0,$DF,$AA,$6B,$81,$EE,$7F,$9C,$6B,$F5,$CD,$8C,$CF
	.dc.b	$F1,$F9,$CB,$63,$39,$FE,$6D,$B8,$FE,$3F,$D4,$7C,$FF,$FB,$FE,$B5
	.dc.b	$8F,$E0,$B1,$E7,$18,$C7,$F1,$19,$B0,$E3,$C7,$BC,$FA,$D7,$E4,$1A
	.dc.b	$FE,$B9,$CD,$71,$CB,$C1,$9F,$E9,$67,$79,$78,$23,$19,$FE,$3B,$2E
	.dc.b	$F8,$FE,$32,$F6,$49,$D6,$1D,$27,$C5,$1E,$52,$3E,$AA,$E7,$9F,$B4
	.dc.b	$98,$78,$27,$57,$CD,$F5,$5F,$2A,$DC,$7E,$AA,$F9,$DF,$19,$CC,$F0
	.dc.b	$39,$E7,$C6,$74,$48,$E5,$87,$7C,$BE,$14,$7A,$8F,$81,$8F,$BA,$F9
	.dc.b	$46,$70,$F1,$E7,$F3,$67,$3D,$7F,$CD,$7E,$7D,$1F,$CE,$9B,$7F,$CD
	.dc.b	$FE,$9A,$F3,$FF,$35,$9B,$5F,$CF,$A7,$75,$FE,$F4,$DD,$18,$8D,$4F
	.dc.b	$F7,$1C,$3F,$A2,$4E,$65,$1F,$D1,$F8,$3A,$C6,$2D,$18,$9F,$E9,$9D
	.dc.b	$76,$73,$87,$F5,$2A,$C1,$99,$B1,$92,$5F,$F5,$38,$CB,$63,$CD,$49
	.dc.b	$F3,$5F,$82,$73,$18,$EB,$7B,$9E,$56,$CF,$F7,$FF,$A4,$C4,$E5,$E0
	.dc.b	$48,$F8,$92,$8E,$9F,$82,$73,$3E,$1F,$67,$F8,$43,$D4,$C5,$FF,$82
	.dc.b	$E6,$E8,$E0,$FF,$05,$56,$1B,$5F,$E8,$C6,$CF,$F8,$44,$EB,$E0,$2B
	.dc.b	$D5,$1E,$D2,$57,$99,$F8,$7E,$13,$E4,$3E,$A3,$67,$FC,$77,$EF,$62
	.dc.b	$64,$7F,$8F,$9D,$C6,$2B,$FC,$A1,$F7,$7F,$77,$CA,$37,$E5,$30,$CC
	.dc.b	$3F,$CB,$0E,$5A,$7F,$96,$F2,$E4,$7C,$8F,$49,$D1,$89,$F8,$47,$F8
	.dc.b	$EE,$32,$FF,$82,$5B,$8F,$F9,$A3,$83,$FD,$49,$5F,$97,$6D,$6E,$31
	.dc.b	$5C,$2B,$FC,$EB,$5A,$79,$46,$5F,$F4,$2F,$73,$EE,$3F,$38,$DC,$3C
	.dc.b	$B2,$BF,$F4,$C7,$D2,$EA,$E5,$7F,$A2,$1F,$C2,$F9,$49,$CD,$FE,$0E
	.dc.b	$BB,$BD,$8E,$FD,$F6,$3F,$D2,$31,$4E,$07,$ED,$31,$FE,$9F,$1A,$3F
	.dc.b	$D4,$FE,$5F,$EC,$FD,$43,$F3,$FC,$3C,$36,$7E,$64,$7C,$62,$38,$25
	.dc.b	$FC,$68,$CB,$FE,$74,$FA,$5F,$E1,$73,$F1,$3B,$91,$89,$F3,$C6,$38
	.dc.b	$3F,$83,$4E,$7A,$7F,$8D,$CF,$0F,$07,$26,$AA,$E1,$FE,$E9,$E8,$FE
	.dc.b	$1D,$E5,$FE,$6F,$2E,$AB,$A7,$33,$2B,$81,$FE,$F1,$8C,$57,$93,$D9
	.dc.b	$FF,$76,$4F,$A5,$B1,$F3,$B9,$7F,$DA,$9C,$3F,$C7,$67,$3A,$FF,$32
	.dc.b	$9C,$DF,$E1,$1B,$7F,$5B,$C4,$FE,$53,$F2,$5F,$82,$C6,$5E,$F9,$BF
	.dc.b	$DA,$4F,$E3,$2B,$6E,$67,$F9,$67,$D2,$BD,$FA,$63,$E0,$2D,$7D,$19
	.dc.b	$D3,$67,$4D,$3E,$B7,$C0,$BE,$8B,$F9,$C3,$63,$3F,$4B,$89,$91,$C1
	.dc.b	$DC,$63,$5F,$BB,$F0,$BF,$C9,$6B,$F0,$4D,$8F,$95,$AB,$E7,$BF,$A3
	.dc.b	$4E,$78,$7E,$54,$E6,$FF,$D0,$CD,$BF,$DF,$35,$FD,$CB,$5F,$F6,$03
	.dc.b	$5F,$DA,$37,$84,$D3,$5D,$0F,$B3,$FA,$35,$F9,$66,$E1,$FD,$BB,$89
	.dc.b	$EE,$BE,$03,$FA,$F5,$6B,$36,$1F,$D9,$BE,$23,$F8,$33,$FE,$D8,$6B
	.dc.b	$A3,$5E,$1F,$F5,$88,$CC,$7F,$A2,$FE,$25,$ED,$FF,$6E,$CD,$18,$C3
	.dc.b	$F0,$CB,$E6,$74,$E6,$FF,$21,$F1,$7E,$8A,$FC,$63,$63,$F2,$9D,$BF
	.dc.b	$9A,$9F,$F4,$D4,$E7,$7F,$81,$3F,$A2,$AF,$46,$3F,$AD,$F2,$3F,$C0
	.dc.b	$8D,$C3,$FD,$F0,$D9,$F7,$1F,$E9,$6A,$E4,$FF,$A9,$DA,$E6,$B9,$C6
	.dc.b	$38,$5F,$19,$F8,$5C,$3E,$6C,$63,$D1,$FF,$07,$7B,$7F,$97,$C7,$F8
	.dc.b	$6D,$ED,$FF,$0B,$C6,$5F,$A1,$5F,$EB,$FE,$AF,$F0,$3C,$E6,$7A,$B5
	.dc.b	$E3,$1A,$26,$1F,$E2,$59,$D5,$7F,$84,$7E,$6F,$FC,$41,$EB,$1F,$F1
	.dc.b	$17,$D1,$7E,$12,$FF,$92,$4E,$5F,$E4,$D1,$9C,$7E,$3D,$A3,$11,$E8
	.dc.b	$6B,$F0,$9A,$FF,$8D,$1B,$1C,$63,$FC,$47,$18,$37,$3E,$5F,$67,$FC
	.dc.b	$72,$D3,$F8,$C7,$D0,$BD,$6C,$E5,$FE,$38,$C3,$F9,$14,$E6,$BF,$DF
	.dc.b	$4D,$8F,$EE,$4D,$7F,$B6,$B5,$FD,$2F,$6A,$31,$1F,$DA,$5B,$19,$BF
	.dc.b	$EF,$51,$AC,$9F,$2B,$FD,$A7,$0F,$E9,$CD,$C3,$CB,$E8,$FF,$94,$63
	.dc.b	$1C,$BF,$CD,$70,$FE,$CB,$5E,$6A,$EA,$4B,$F5,$B6,$35,$18,$3F,$CC
	.dc.b	$90,$F1,$BF,$CB,$CD,$8F,$EC,$1F,$1A,$F7,$33,$5F,$E6,$38,$C4,$FF
	.dc.b	$9B,$5B,$7F,$80,$6B,$F9,$66,$BF,$A0,$6B,$FE,$72,$6E,$66,$98,$F9
	.dc.b	$77,$DA,$FF,$9C,$DB,$7F,$DD,$4F,$F3,$41,$FD,$DD,$CD,$FF,$76,$8C
	.dc.b	$BF,$94,$6B,$FA,$A6,$BF,$B4,$6B,$FE,$86,$FB,$1F,$E4,$8D,$F3,$3F
	.dc.b	$8D,$7D,$0F,$F7,$5A,$FF,$03,$56,$B2,$34,$E3,$DA,$AE,$18,$7F,$82
	.dc.b	$A7,$87,$03,$E9,$7F,$D2,$0D,$C7,$FC,$6B,$F3,$AF,$A2,$7B,$DA,$B9
	.dc.b	$5F,$E1,$7B,$E1,$8F,$F8,$B3,$0E,$17,$A8,$FF,$01,$7D,$06,$AB,$FC
	.dc.b	$00,$DB,$FC,$EF,$D0,$FE,$F6,$B7,$8F,$FA,$12,$B5,$1F,$E4,$76,$F4
	.dc.b	$AB,$86,$1D,$2F,$FD,$17,$1C,$EF,$FC,$AE,$D7,$BF,$57,$F9,$7E,$31
	.dc.b	$1F,$E5,$3F,$8B,$FE,$45,$81,$B2,$F5,$7F,$D2,$F1,$A2,$78,$7E,$08
	.dc.b	$CB,$FE,$BE,$BF,$EB,$1F,$0F,$FC,$F1,$FE,$7A,$D7,$25,$0E,$7E,$8B
	.dc.b	$FD,$25,$C1,$7D,$93,$49,$5D,$65,$F1,$FF,$EA,$78,$C3,$5F,$E9,$B1
	.dc.b	$AD,$FB,$4E,$67,$FB,$73,$FE,$8C,$7C,$CD,$7D,$4D,$7F,$AB,$B5,$E3
	.dc.b	$EE,$7A,$E3,$11,$FD,$F9,$AF,$9A,$3E,$07,$77,$EC,$23,$E2,$35,$F9
	.dc.b	$0D,$7F,$B2,$B5,$FF,$02,$7D,$7B,$E3,$7E,$53,$63,$E7,$35,$FA,$4D
	.dc.b	$7F,$B7,$B5,$E7,$1C,$88,$FF,$BF,$72,$79,$9F,$67,$B3,$E9,$B6,$33
	.dc.b	$3B,$8C,$11,$C1,$0F,$EF,$FE,$13,$E6,$BF,$CE,$51,$F8,$78,$C7,$9B
	.dc.b	$F8,$76,$C6,$67,$6F,$26,$BF,$E2,$78,$DB,$99,$E5,$3A,$8F,$F1,$8C
	.dc.b	$BF,$48,$7C,$07,$76,$CF,$8D,$AF,$A1,$AF,$7F,$C6,$BF,$DD,$CF,$A5
	.dc.b	$AF,$B9,$AE,$FE,$23,$F4,$5F,$71,$FE,$C7,$F9,$AB,$EC,$9A,$4A,$FD
	.dc.b	$54,$ED,$B3,$B6,$CE,$DF,$F6,$F7,$E9,$AF,$A0,$FD,$87,$7A,$31,$3F
	.dc.b	$B5,$8C,$4E,$A7,$33,$F5,$BC,$27,$E0,$BF,$4D,$B1,$F5,$DA,$EE,$34
	.dc.b	$47,$4D,$7F,$C3,$37,$FE,$66,$7F,$83,$72,$7F,$B9,$7D,$FF,$C7,$4E
	.dc.b	$67,$71,$F2,$3F,$0E,$FF,$2D,$47,$C4,$6B,$FE,$4A,$D5,$18,$8F,$F3
	.dc.b	$54,$63,$DC,$E4,$48,$E4,$D7,$FC,$A7,$6F,$DA,$DE,$77,$CA,$6C,$FC
	.dc.b	$96,$BC,$1A,$FF,$9E,$67,$FC,$AB,$E0,$FD,$2F,$6C,$D2,$4F,$D3,$8C
	.dc.b	$B6,$32,$F9,$23,$2F,$F9,$0B,$E9,$3F,$D1,$CF,$CA,$6B,$9F,$82,$FD
	.dc.b	$C4,$62,$3F,$D1,$9F,$32,$FB,$AF,$9D,$18,$6B,$FE,$96,$8D,$6F,$FC
	.dc.b	$E5,$B1,$FC,$F5,$AF,$F9,$E9,$AF,$38,$E1,$3D,$21,$93,$F6,$97,$D5
	.dc.b	$7E,$59,$B3,$F9,$56,$BF,$51,$AF,$CD,$F9,$7F,$E1,$4F,$E6,$DA,$FD
	.dc.b	$A6,$BB,$EF,$FD,$1C,$62,$3F,$C9,$ED,$7C,$AD,$B3,$7F,$5B,$FF,$3C
	.dc.b	$4F,$EA,$5A,$FE,$21,$AF,$E0,$5A,$FF,$97,$3F,$69,$FC,$2B,$78,$BF
	.dc.b	$6B,$18,$8F,$8A,$DD,$5F,$B3,$7F,$41,$94,$63,$DB,$FA,$93,$98,$C4
	.dc.b	$FE,$F7,$F6,$7F,$EF,$4D,$F4,$7F,$81,$C6,$23,$F3,$ED,$7F,$D6,$7F
	.dc.b	$8D,$7F,$75,$4D,$23,$FD,$E5,$AF,$F7,$D6,$BF,$AA,$FD,$7B,$FD,$D6
	.dc.b	$FF,$0F,$6B,$9F,$8E,$FE,$2B,$18,$8F,$F5,$8F,$C9,$3E,$67,$FA,$83
	.dc.b	$63,$F8,$3B,$5F,$F5,$1B,$5E,$31,$C0,$8F,$07,$FA,$D7,$8C,$FA,$AA
	.dc.b	$49,$FE,$27,$19,$6C,$65,$FF,$58,$C6,$5B,$1F,$E1,$BF,$C1,$5F,$7F
	.dc.b	$49,$3F,$E2,$18,$CB,$63,$2F,$F8,$9E,$32,$D8,$F8,$BD,$92,$3E,$53
	.dc.b	$5F,$F6,$1F,$98,$F9,$5F,$EE,$CD,$8F,$E5,$AD,$77,$1A,$23,$A6,$BF
	.dc.b	$EC,$EF,$9B,$7C,$07,$F8,$05,$B1,$FC,$E1,$AE,$E3,$44,$74,$D7,$FC
	.dc.b	$89,$F4,$9F,$E6,$E7,$FC,$88,$D7,$3F,$05,$FE,$45,$8C,$40,$81,$CB
	.dc.b	$86,$BB,$6B,$B8,$FF,$06,$7D,$8B,$F0,$2A,$49,$FF,$09,$46,$5F,$EC
	.dc.b	$DF,$95,$FF,$A2,$5B,$F0,$BF,$DB,$23,$11,$FD,$21,$AF,$F4,$16,$BF
	.dc.b	$ED,$C6,$BF,$C8,$9B,$E0,$1B,$F5,$5F,$E4,$A8,$34,$9E,$F7,$F7,$C6
	.dc.b	$D7,$F7,$B6,$BF,$3B,$EA,$FE,$51,$1F,$9E,$D7,$46,$08,$60,$7C,$0F
	.dc.b	$E0,$2D,$FA,$07,$C0,$F8,$4F,$EC,$71,$8C,$FF,$D0,$3E,$27,$D3,$C7
	.dc.b	$EB,$E1,$C7,$27,$C9,$FD,$AE,$2F,$EC,$D1,$8D,$3E,$F4,$62,$75,$1F
	.dc.b	$81,$CB,$E1,$C6,$23,$FA,$5D,$71,$7F,$27,$3A,$79,$9C,$63,$55,$D6
	.dc.b	$32,$7F,$54,$FC,$25,$EF,$7A,$A3,$0D,$B4,$EA,$7F,$29,$AF,$C2,$8F
	.dc.b	$F4,$FE,$9E,$77,$53,$FE,$9E,$8C,$5F,$39,$CD,$71,$AF,$01,$F9,$DC
	.dc.b	$BF,$9F,$BC,$4F,$CE,$38,$8F,$F0,$8F,$27,$CC,$F1,$3C,$78,$FE,$9F
	.dc.b	$E3,$9F,$2F,$FE,$83,$BD,$E1,$C6,$BF,$50,$DC,$7A,$E1,$C9,$61,$CB
	.dc.b	$E1,$3C,$B9,$77,$9F,$21,$E7,$AE,$0D,$3D,$15,$C5,$CC,$F1,$C7,$C7
	.dc.b	$58,$C3,$C5,$7B,$BE,$B7,$BC,$7C,$F7,$B3,$CD,$5B,$C3,$9D,$F0,$C3
	.dc.b	$D8,$F8,$E3,$FC,$73,$3B,$BE,$73,$BB,$E8,$FA,$32,$F9,$13,$AB,$EA
	.dc.b	$47,$E7,$46,$A7,$E4,$9C,$C9,$3E,$08,$D4,$78,$E7,$A1,$1F,$9F,$1A
	.dc.b	$C7,$E0,$47,$4A,$F0,$3A,$CB,$E9,$E5,$EE,$AD,$CF,$53,$47,$B2,$B8
	.dc.b	$B9,$CB,$E1,$3D,$71,$F1,$39,$C7,$C5,$87,$52,$65,$F4,$23,$2F,$5A
	.dc.b	$F5,$5F,$4B,$F1,$13,$1F,$5C,$65,$F3,$93,$0F,$C0,$E1,$D2,$BE,$4D
	.dc.b	$F4,$8C,$63,$EA,$BE,$71,$89,$F4,$C6,$32,$F8,$B9,$79,$70,$F7,$1F
	.dc.b	$10,$EA,$60,$F8,$11,$C1,$C9,$83,$F0,$A7,$F5,$23,$F3,$C5,$9F,$D2
	.dc.b	$37,$1F,$DC,$B9,$F8,$6F,$E0,$6F,$F5,$76,$BF,$D5,$DA,$FF,$5D,$6C
	.dc.b	$9F,$E2,$19,$7F,$00,$6E,$3F,$BF,$FA,$C7,$D2,$FF,$76,$AD,$6F,$FB
	.dc.b	$0B,$DE,$3D,$E3,$F5,$37,$C3,$87,$51,$27,$F6,$35,$C3,$75,$EE,$73
	.dc.b	$5B,$C3,$BC,$7F,$00,$9D,$46,$09,$97,$C1,$AF,$0E,$5E,$78,$C5,$F4
	.dc.b	$25,$7E,$E2,$F9,$CE,$A3,$9C,$FA,$CF,$05,$73,$71,$1A,$7E,$65,$F8
	.dc.b	$32,$F4,$5F,$32,$4F,$F2,$0C,$3B,$BC,$A3,$2F,$23,$BC,$F2,$74,$4C
	.dc.b	$7E,$3E,$3E,$38,$E1,$8F,$8F,$3F,$B9,$97,$8E,$7D,$F8,$F6,$9C,$C7
	.dc.b	$63,$27,$63,$9C,$7E,$06,$FB,$4F,$0B,$ED,$9F,$D9,$CF,$D7,$87,$03
	.dc.b	$95,$6F,$0E,$59,$FA,$E7,$46,$F3,$F5,$C6,$6B,$84,$FB,$EF,$AC,$62
	.dc.b	$38,$BA,$AE,$31,$EA,$33,$87,$97,$2F,$4B,$F9,$68,$CE,$7F,$17,$3F
	.dc.b	$5B,$F3,$EF,$9C,$62,$7E,$2C,$63,$3F,$95,$7C,$F1,$F1,$65,$E6,$8F
	.dc.b	$D1,$F0,$BC,$1F,$BE,$DD,$3F,$14,$E7,$6B,$3A,$24,$E2,$32,$F0,$FC
	.dc.b	$23,$FD,$35,$BF,$F8,$3A,$D1,$8D,$7F,$C2,$ED,$9F,$F2,$EC,$EB,$C5
	.dc.b	$FF,$37,$C6,$27,$55,$FE,$23,$6B,$FE,$EC,$6D,$71,$9F,$F0,$EC,$BC
	.dc.b	$17,$C1,$FF,$9E,$BA,$3F,$C7,$9D,$6F,$AF,$FC,$B5,$6B,$0F,$F3,$85
	.dc.b	$6A,$73,$7F,$E8,$0C,$3F,$CE,$77,$B7,$F9,$21,$DE,$34,$FF,$9E,$63
	.dc.b	$4F,$F2,$96,$BC,$A3,$6F,$1B,$E6,$D3,$F9,$86,$3F,$2B,$E5,$BA,$B6
	.dc.b	$7C,$2D,$9D,$E6,$AD,$B6,$B6,$D9,$DC,$E8,$EB,$F9,$53,$C6,$A4,$BE
	.dc.b	$F1,$96,$C6,$6B,$CF,$39,$6C,$7A,$3F,$55,$7D,$2D,$25,$FA,$A3,$2F
	.dc.b	$C2,$D1,$F2,$3F,$19,$B1,$98,$C1,$1C,$13,$0F,$AB,$3C,$1F,$F5,$F7
	.dc.b	$17,$D6,$F8,$47,$93,$FC,$F6,$7F,$D5,$F1,$89,$D4,$E6,$7F,$D6,$93
	.dc.b	$9B,$FF,$62,$56,$88,$FF,$B4,$BE,$43,$B6,$06,$92,$34,$4B,$FF,$6E
	.dc.b	$C6,$5F,$C0,$F9,$AF,$86,$FC,$0B,$6B,$F0,$2D,$9D,$CF,$E3,$39,$9F
	.dc.b	$09,$49,$7F,$8E,$8C,$B6,$33,$5F,$9C,$9C,$B6,$3E,$47,$D3,$FF,$B2
	.dc.b	$EB,$F4,$3A,$23,$15,$FE,$07,$8C,$79,$BC,$09,$5C,$2B,$75,$F4,$1A
	.dc.b	$FC,$E6,$AA,$7A,$4F,$8A,$FB,$4D,$BF,$1B,$6B,$C4,$D7,$EE,$B5,$FB
	.dc.b	$6D,$7B,$36,$BB,$D7,$D0,$FC,$46,$5F,$D0,$6D,$FF,$02,$FC,$61,$F2
	.dc.b	$1F,$BC,$6D,$7E,$F1,$B3,$B9,$D1,$27,$4D,$8F,$E0,$3E,$E3,$EB,$7F
	.dc.b	$80,$B6,$BF,$80,$B6,$77,$3F,$C4,$BC,$C7,$E8,$9F,$AC,$9F,$F1,$D4
	.dc.b	$63,$AB,$53,$99,$12,$75,$87,$F0,$EF,$D8,$3F,$D8,$ED,$F3,$7F,$89
	.dc.b	$C6,$2B,$F8,$BB,$5F,$E7,$BF,$B2,$3F,$13,$FE,$32,$DB,$FE,$32,$D9
	.dc.b	$FE,$3E,$D7,$FA,$2F,$DA,$FC,$86,$ED,$FC,$8A,$31,$1F,$EE,$76,$AF
	.dc.b	$11,$18,$6C,$F1,$25,$FF,$2A,$F8,$3F,$C7,$C9,$A4,$E7,$F9,$9F,$89
	.dc.b	$BE,$E2,$92,$FF,$A1,$EE,$9D,$46,$1B,$9F,$D2,$6B,$F4,$04,$F1,$AE
	.dc.b	$91,$FA,$BF,$8E,$7C,$1F,$F1,$96,$E1,$FB,$26,$CE,$E7,$FA,$3F,$C7
	.dc.b	$BD,$7A,$4B,$FE,$93,$19,$6C,$66,$BF,$A8,$CE,$5B,$1F,$D4,$BE,$47
	.dc.b	$F5,$1B,$1A,$4E,$BF,$EC,$7F,$24,$F9,$97,$D6,$E1,$B9,$FF,$6A,$9D
	.dc.b	$4E,$49,$19,$9D,$E5,$FE,$93,$AD,$CE,$6B,$F9,$E6,$3F,$C6,$2B,$F9
	.dc.b	$DB,$5F,$E7,$2D,$7F,$86,$37,$A1,$F2,$CF,$BB,$FE,$F7,$18,$6C,$67
	.dc.b	$0F,$F0,$0C,$E5,$B1,$9B,$FF,$54,$4E,$5B,$19,$7F,$01,$19,$7F,$C4
	.dc.b	$7F,$9D,$FE,$11,$FD,$9E,$08,$C4,$7E,$75,$B1,$9E,$2F,$9F,$FB,$4E
	.dc.b	$B4,$34,$91,$A2,$65,$F4,$63,$2F,$B3,$E7,$3F,$D6,$4D,$F6,$FE,$1C
	.dc.b	$63,$2F,$6B,$5F,$5F,$CE,$FC,$B2,$79,$3F,$29,$F9,$DF,$9B,$3E,$EF
	.dc.b	$E6,$B7,$1F,$96,$DB,$F9,$2D,$7E,$3B,$5F,$E2,$FB,$A3,$10,$8C,$57
	.dc.b	$63,$FC,$01,$ED,$BE,$4E,$93,$0F,$EB,$71,$96,$C6,$6F,$FA,$E4,$E5
	.dc.b	$B1,$F9,$EF,$D5,$FF,$AB,$F0,$1A,$4E,$FD,$13,$0F,$D0,$46,$5F,$F2
	.dc.b	$0F,$E5,$EF,$C5,$3F,$C8,$AD,$8C,$C6,$0F,$F2,$7F,$C3,$3E,$AF,$FC
	.dc.b	$61,$9F,$F0,$B8,$C7,$B9,$A9,$CC,$FF,$3E,$9C,$DF,$F9,$69,$B1,$FE
	.dc.b	$FF,$6B,$CE,$78,$A8,$CB,$FE,$62,$F8,$8F,$F2,$CB,$FD,$29,$A9,$A7
	.dc.b	$F9,$7B,$E9,$9F,$42,$FA,$5C,$36,$BF,$CD,$B3,$A9,$C9,$C7,$DB,$71
	.dc.b	$52,$57,$08,$CB,$FE,$76,$5F,$23,$E6,$6C,$7F,$9C,$1A,$EE,$34,$47
	.dc.b	$4D,$7F,$D0,$3D,$5F,$D1,$DF,$F3,$CB,$5C,$F9,$7F,$12,$31,$1F,$CF
	.dc.b	$C7,$D3,$7F,$35,$7F,$D0,$2D,$7F,$D1,$AD,$78,$35,$FE,$90,$D7,$FB
	.dc.b	$CB,$5E,$BE,$3F,$D4,$8D,$7B,$1F,$3F,$C6,$FF,$47,$B6,$BF,$C7,$7A
	.dc.b	$7F,$37,$6C,$EA,$31,$1F,$E3,$D6,$BF,$E6,$86,$BF,$E6,$66,$BF,$B3
	.dc.b	$F9,$FF,$F1,$2B,$FC,$A1,$AF,$F2,$66,$BF,$C9,$1A,$FE,$F9,$AF,$3F
	.dc.b	$3F,$FD,$4F,$DA,$9C,$B6,$33,$DF,$FE,$C3,$AF,$E6,$ED,$7F,$9A,$B5
	.dc.b	$FE,$33,$EF,$F4,$F0,$9F,$33,$49,$3F,$C9,$5F,$F5,$BF,$7F,$FB,$3B
	.dc.b	$E9,$AF,$C4,$9A,$4A,$FE,$BD,$19,$6C,$65,$FE,$B9,$19,$7F,$D8,$3F
	.dc.b	$4D,$FE,$5F,$7F,$D7,$ED,$73,$E7,$7F,$B2,$3C,$A4,$E6,$7F,$A6,$F9
	.dc.b	$9F,$E8,$97,$FA,$AF,$C1,$7C,$88,$C4,$7F,$54,$8C,$11,$D9,$0F,$F3
	.dc.b	$E5,$3E,$97,$F1,$2D,$8F,$F4,$1B,$5D,$C7,$C6,$FA,$6F,$F4,$9C,$7F
	.dc.b	$7F,$6B,$FD,$ED,$AF,$F8,$8F,$44,$62,$3F,$CA,$B1,$8F,$33,$A1,$23
	.dc.b	$A3,$F0,$FD,$27,$C9,$D2,$4F,$FB,$A2,$32,$FF,$88,$BC,$2F,$95,$EA
	.dc.b	$7F,$14,$8C,$10,$FF,$11,$C6,$88,$E4,$FF,$64,$FA,$DF,$EE,$97,$FC
	.dc.b	$69,$E4,$FE,$BA,$31,$1F,$E3,$B8,$FF,$7F,$FA,$EF,$A6,$F9,$13,$FE
	.dc.b	$43,$8C,$6D,$A9,$CC,$FF,$9D,$5A,$FF,$92,$1B,$3C,$23,$FC,$A7,$18
	.dc.b	$37,$BB,$E9,$5F,$4D,$F8,$96,$D7,$F1,$53,$FC,$D4,$E1,$FE,$31,$3C
	.dc.b	$67,$F2,$1F,$4D,$FB,$19,$E1,$28,$C4,$70,$6B,$FE,$5F,$8D,$77,$F2
	.dc.b	$24,$72,$9D,$BF,$E2,$26,$BF,$E7,$DF,$A6,$7D,$4F,$E8,$5B,$1E,$26
	.dc.b	$BE,$86,$BE,$1F,$A7,$FC,$0B,$EC,$6B,$9E,$3F,$D6,$46,$23,$FD,$49
	.dc.b	$F0,$FF,$CE,$9B,$E5,$7D,$A8,$C4,$7F,$9E,$9B,$9B,$E9,$FF,$C2,$13
	.dc.b	$F1,$9A,$FC,$C6,$BF,$15,$AF,$FA,$17,$E9,$FF,$B5,$91,$FE,$DE,$D7
	.dc.b	$EC,$07,$0F,$DB,$7F,$BA,$CB,$E9,$FF,$D1,$73,$F6,$1A,$FD,$E6,$BF
	.dc.b	$5D,$AF,$FB,$A3,$E6,$7F,$A4,$FC,$71,$A4,$8F,$F7,$9B,$5F,$F7,$9B
	.dc.b	$5F,$F5,$07,$99,$FC,$BB,$E4,$7F,$A9,$F8,$C6,$92,$7F,$29,$3B,$6C
	.dc.b	$ED,$B3,$B7,$F3,$5F,$36,$FA,$7F,$F0,$AB,$63,$E3,$B5,$DC,$68,$8E
	.dc.b	$9A,$FC,$5E,$BF,$E6,$19,$FE,$F2,$3E,$34,$60,$FE,$DB,$E5,$3E,$9E
	.dc.b	$92,$7D,$71,$96,$C6,$5F,$F0,$36,$69,$D4,$61,$AF,$CA,$F9,$67,$B9
	.dc.b	$FD,$F1,$B1,$F2,$1A,$F4,$6B,$C6,$38,$11,$E0,$FF,$8F,$3E,$7D,$F4
	.dc.b	$F4,$93,$F9,$08,$CB,$63,$2F,$F0,$B8,$CB,$63,$FB,$CF,$D3,$BE,$0E
	.dc.b	$92,$7F,$D5,$B1,$97,$F8,$6F,$D3,$BE,$9F,$FB,$03,$63,$31,$82,$38
	.dc.b	$21,$FC,$6A,$38,$3F,$C7,$3E,$9B,$FC,$82,$FF,$15,$6B,$FE,$B2,$6A
	.dc.b	$E7,$46,$1B,$EE,$FF,$73,$8C,$11,$F9,$5F,$4F,$2B,$E4,$35,$F8,$CD
	.dc.b	$7F,$C8,$DA,$23,$10,$24,$65,$C7,$C0,$7E,$03,$07,$CE,$7B,$BE,$0F
	.dc.b	$FA,$FC,$63,$4E,$57,$FE,$75,$F0,$AE,$3C,$E3,$FC,$DB,$C3,$F0,$7E
	.dc.b	$93,$6F,$C5,$6C,$64,$F2,$46,$5A,$64,$FC,$8E,$1F,$22,$78,$C7,$CB
	.dc.b	$D9,$18,$9D,$4F,$23,$E8,$7D,$3F,$FA,$F5,$FF,$47,$63,$FE,$E6,$8C
	.dc.b	$47,$AC,$DF,$BD,$EE,$D3,$E0,$CF,$F9,$C5,$B1,$9A,$E4,$DA,$E2,$76
	.dc.b	$3C,$26,$46,$7E,$8F,$37,$51,$8C,$91,$89,$D4,$F9,$7E,$19,$F4,$FF
	.dc.b	$E7,$FC,$BF,$A9,$D6,$9B,$87,$2B,$F6,$1D,$70,$F2,$61,$CB,$E1,$3C
	.dc.b	$98,$71,$CF,$ED,$E1,$C7,$1E,$78,$70,$CB,$C1,$5A,$8F,$35,$6E,$31
	.dc.b	$97,$9E,$B6,$FE,$42,$74,$DA,$E2,$75,$C7,$C7,$3A,$9F,$41,$EA,$C7
	.dc.b	$CC,$FF,$B2,$EF,$AB,$AC,$7C,$EE,$EB,$BE,$5E,$F8,$EE,$67,$1B,$5F
	.dc.b	$E3,$C9,$D4,$E4,$91,$D0,$D5,$F8,$30,$F1,$CF,$87,$2F,$83,$5A,$CB
	.dc.b	$D7,$8F,$69,$C9,$FE,$EB,$C3,$96,$5E,$FC,$79,$9E,$EC,$BC,$78,$F8
	.dc.b	$6F,$8E,$5F,$24,$F9,$D9,$76,$7C,$F1,$AA,$E2,$F0,$8F,$31,$C6,$BC
	.dc.b	$CF,$94,$C4,$71,$C7,$CB,$96,$67,$8D,$F4,$76,$43,$33,$C1,$F1,$9A
	.dc.b	$8E,$0F,$88,$98,$FB,$E7,$86,$1E,$F9,$E1,$18,$C3,$D5,$5B,$C7,$E0
	.dc.b	$5E,$EF,$C3,$18,$AD,$E5,$F8,$2C,$3C,$95,$FC,$9A,$3F,$4C,$DC,$BF
	.dc.b	$AA,$46,$23,$F8,$EF,$C2,$FF,$06,$6E,$9F,$D6,$63,$11,$FC,$B1,$AF
	.dc.b	$EF,$9F,$E2,$19,$7E,$D9,$B8,$FF,$33,$D1,$F4,$EF,$A1,$FE,$E3,$7B
	.dc.b	$C7,$F8,$16,$1C,$31,$ED,$1F,$BA,$BE,$1C,$5A,$89,$3F,$A7,$AE,$18
	.dc.b	$F6,$C7,$AE,$5E,$AA,$D5,$F0,$CB,$F4,$59,$74,$9C,$92,$3C,$8F,$F1
	.dc.b	$38,$D4,$60,$95,$D5,$E7,$7E,$0A,$F4,$9C,$E7,$47,$C3,$AE,$6E,$23
	.dc.b	$75,$D1,$61,$EB,$CB,$D1,$5B,$9F,$A1,$8F,$C5,$C7,$C1,$18,$3C,$07
	.dc.b	$49,$F0,$46,$B7,$67,$F3,$EB,$86,$5E,$1C,$BC,$11,$C8,$CE,$1E,$78
	.dc.b	$CC,$F9,$E7,$27,$73,$D3,$9F,$B9,$C1,$A3,$73,$CA,$BD,$98,$79,$32
	.dc.b	$F7,$D7,$2C,$BD,$19,$79,$EB,$53,$EE,$C3,$93,$DA,$33,$97,$AF,$2F
	.dc.b	$14,$F8,$67,$39,$7A,$E7,$31,$89,$E5,$87,$E3,$27,$2F,$84,$99,$7D
	.dc.b	$1C,$FC,$F5,$D6,$78,$67,$E8,$AC,$E1,$D7,$3F,$3D,$EE,$7D,$F1,$89
	.dc.b	$F1,$4E,$4F,$C9,$57,$37,$FB,$D7,$0B,$C3,$F9,$CD,$9F,$C9,$39,$DA
	.dc.b	$CE,$89,$38,$8C,$BC,$3E,$9B,$FD,$31,$A7,$F8,$36,$D1,$8C,$3F,$C2
	.dc.b	$ED,$9F,$F6,$C3,$63,$FC,$65,$3C,$3C,$7F,$F3,$7D,$69,$B3,$AA,$E0
	.dc.b	$FF,$92,$5A,$FF,$9F,$7E,$07,$FC,$EF,$D3,$FE,$39,$E9,$7D,$3F,$E8
	.dc.b	$3C,$25,$6B,$2F,$0D,$69,$FF,$08,$CE,$9F,$F6,$1B,$70,$FF,$7E,$35
	.dc.b	$FF,$20,$9E,$18,$DD,$7F,$A1,$23,$FC,$9D,$3B,$7F,$C8,$ED,$73,$1B
	.dc.b	$CF,$E4,$FD,$3D,$CC,$D2,$61,$E0,$BF,$85,$18,$7C,$75,$CA,$74,$D8
	.dc.b	$CF,$D3,$BC,$7F,$03,$6B,$6D,$9D,$CE,$89,$3A,$6C,$78,$3E,$9D,$F4
	.dc.b	$5E,$B6,$D7,$89,$B3,$B9,$F8,$CF,$FA,$D7,$E9,$EE,$A6,$92,$FE,$9B
	.dc.b	$67,$E9,$B5,$FF,$5D,$71,$7D,$1F,$25,$E4,$FF,$39,$9C,$DF,$FA,$EE
	.dc.b	$B5,$39,$9F,$F4,$25,$70,$9C,$92,$BF,$D9,$FF,$1F,$D7,$13,$49,$1A
	.dc.b	$25,$FF,$B3,$E3,$2F,$D6,$F3,$9F,$0B,$F8,$16,$D7,$D8,$6C,$EE,$7E
	.dc.b	$3F,$23,$E0,$E9,$2F,$F0,$95,$FE,$98,$8C,$3F,$87,$9E,$13,$A6,$C6
	.dc.b	$67,$E4,$7D,$3F,$FB,$2E,$BF,$1D,$A2,$31,$5F,$EE,$C6,$BF,$EA,$28
	.dc.b	$D7,$9B,$C8,$95,$CA,$B7,$5F,$39,$AA,$9E,$9B,$E2,$7E,$D3,$6F,$C6
	.dc.b	$DA,$FF,$7E,$B5,$F1,$35,$FB,$6D,$7B,$36,$BB,$D7,$D1,$FA,$79,$7F
	.dc.b	$45,$B7,$FA,$FF,$A7,$7C,$7F,$E6,$1B,$5F,$41,$B3,$B9,$D1,$27,$4D
	.dc.b	$8F,$DB,$FB,$4F,$A7,$FE,$04,$DA,$FD,$CB,$67,$73,$FC,$37,$CB,$7D
	.dc.b	$3F,$EB,$27,$FD,$A7,$18,$E9,$D4,$E6,$7F,$DC,$13,$9C,$3F,$82,$FD
	.dc.b	$3B,$C9,$FE,$24,$D8,$CC,$60,$FE,$2D,$F4,$CF,$A7,$A4,$9F,$E1,$F1
	.dc.b	$97,$F8,$8F,$D3,$FC,$4C,$7F,$DC,$FE,$12,$31,$1F,$D0,$9A,$A3,$5B
	.dc.b	$FF,$25,$9C,$C6,$2B,$AC,$E4,$FE,$55,$F0,$3F,$8C,$93,$49,$CF,$F2
	.dc.b	$7E,$9D,$F4,$74,$98,$7F,$2D,$DD,$3A,$8C,$37,$3F,$E8,$33,$AA,$E3
	.dc.b	$18,$9D,$3F,$A5,$F8,$D7,$D3,$FF,$15,$6E,$1F,$AE,$6C,$FF,$4C,$E9
	.dc.b	$46,$23,$E4,$FC,$83,$D8,$FE,$90,$DB,$FE,$7E,$D9,$FE,$86,$D7,$8C
	.dc.b	$F0,$24,$F0,$7F,$A9,$7C,$8F,$E9,$F7,$34,$9D,$9F,$D7,$7E,$49,$F3
	.dc.b	$4F,$A7,$C3,$71,$FE,$AD,$3A,$9C,$92,$33,$3B,$CB,$FB,$5D,$6E,$73
	.dc.b	$5F,$CE,$F1,$FE,$2F,$5F,$C3,$DA,$FF,$38,$6B,$FC,$2D,$BD,$0F,$96
	.dc.b	$7D,$3F,$F7,$58,$C3,$63,$38,$7F,$6D,$9C,$B6,$33,$7F,$E0,$09,$CB
	.dc.b	$63,$28,$91,$A7,$F4,$3F,$4F,$F8,$47,$FC,$42,$D7,$F4,$0D,$C0,$E7
	.dc.b	$27,$D3,$D2,$72,$FD,$36,$E5,$F4,$1B,$87,$CC,$7B,$62,$F9,$CF,$F3
	.dc.b	$ED,$7B,$DB,$5E,$D6,$BE,$B3,$F8,$33,$87,$DB,$F3,$BF,$38,$9E,$47
	.dc.b	$CD,$7E,$8F,$D3,$BE,$9F,$E8,$37,$0F,$9C,$DB,$F9,$8D,$7C,$18,$A3
	.dc.b	$11,$F3,$1A,$FF,$19,$6C,$6E,$BF,$C0,$5E,$D3,$E9,$E9,$30,$FE,$BB
	.dc.b	$19,$6C,$67,$0F,$D9,$4E,$5B,$1F,$A1,$FA,$7F,$EB,$3B,$9A,$4E,$FD
	.dc.b	$13,$0F,$D1,$C6,$5F,$EE,$FF,$4E,$FA,$7F,$F1,$1B,$63,$31,$83,$FC
	.dc.b	$99,$F0,$6F,$A7,$FF,$1B,$67,$F5,$63,$1E,$DE,$A7,$33,$FE,$ED,$6B
	.dc.b	$FE,$C8,$6C,$F0,$BF,$F2,$63,$67,$FC,$BA,$D7,$FC,$1B,$F1,$1F,$E0
	.dc.b	$E7,$F3,$CD,$4D,$3F,$C2,$7F,$4E,$FA,$27,$D3,$61,$B5,$FE,$5B,$9D
	.dc.b	$4E,$4E,$3E,$D3,$8A,$92,$B8,$46,$5F,$F3,$0F,$D3,$3D,$BF,$CC,$53
	.dc.b	$4F,$F4,$7C,$E6,$BF,$6D,$39,$7F,$A8,$46,$5F,$04,$FE,$86,$DF,$25
	.dc.b	$D2,$39,$E1,$6B,$8B,$7B,$D8,$73,$C7,$B5,$7F,$92,$30,$E0,$BE,$9F
	.dc.b	$F9,$E3,$FE,$3A,$6B,$FE,$7F,$6B,$FE,$90,$6B,$FC,$3F,$EA,$DF,$A4
	.dc.b	$3E,$07,$C9,$18,$6C,$66,$7E,$0E,$9F,$BD,$BE,$F8,$FC,$3C,$38,$56
	.dc.b	$83,$87,$E2,$98,$BE,$8E,$0F,$67,$95,$F1,$7C,$DF,$C0,$E9,$78,$3E
	.dc.b	$2D,$75,$9C,$B6,$3F,$A6,$B6,$78,$11,$D7,$93,$D7,$95,$F4,$74,$93
	.dc.b	$E9,$F0,$33,$E9,$E2,$3D,$39,$2F,$8C,$78,$DA,$FB,$AB,$8B,$CF,$C5
	.dc.b	$7C,$EF,$F9,$C2,$7F,$96,$36,$33,$3F,$CB,$63,$FC,$EB,$6F,$AD,$FF
	.dc.b	$20,$4F,$C1,$6B,$E4,$6F,$0F,$E5,$70,$EB,$3F,$4F,$99,$F5,$5F,$E6
	.dc.b	$78,$FF,$5F,$B5,$FF,$61,$B5,$FF,$18,$B5,$FF,$78,$B5,$FF,$18,$36
	.dc.b	$3A,$F9,$FF,$E8,$C6,$CF,$FA,$13,$0F,$E3,$0F,$A7,$C0,$FB,$D1,$88
	.dc.b	$FC,$2C,$F0,$7E,$7A,$FA,$9F,$86,$D8,$FF,$47,$B5,$FF,$11,$35,$FF
	.dc.b	$00,$47,$02,$3C,$1F,$9B,$F6,$0F,$9F,$FC,$83,$67,$F1,$CD,$7C,$73
	.dc.b	$FB,$7F,$4D,$F2,$FE,$F6,$CF,$FB,$A9,$AF,$FB,$B9,$AF,$F9,$71,$AF
	.dc.b	$F8,$07,$0F,$F1,$4C,$F3,$24,$7F,$AB,$2D,$F7,$3E,$62,$47,$DD,$BE
	.dc.b	$FB,$FF,$15,$E0,$FF,$21,$F8,$DF,$8A,$F5,$FF,$DF,$57,$FE,$49,$8C
	.dc.b	$11,$C1,$0F,$F2,$64,$70,$7F,$C9,$DF,$69,$FB,$8B,$EC,$DF,$03,$FD
	.dc.b	$C6,$D9,$FF,$5B,$57,$1A,$FE,$51,$AB,$EA,$DF,$57,$F8,$76,$DF,$F8
	.dc.b	$24,$F8,$8E,$1C,$4F,$19,$E0,$48,$E0,$7F,$A0,$E7,$83,$FE,$8A,$6B
	.dc.b	$E6,$FB,$CF,$E1,$D8,$9A,$7F,$33,$9F,$06,$1E,$8F,$AA,$7D,$5F,$E7
	.dc.b	$5B,$87,$83,$24,$62,$3C,$ED,$7F,$D3,$6D,$7F,$00,$D7,$FD,$34,$D8
	.dc.b	$E7,$3F,$C0,$BD,$F7,$E0,$3F,$C9,$5B,$7F,$C8,$DA,$FE,$AA,$FC,$58
	.dc.b	$7F,$57,$F8,$17,$C5,$7F,$A7,$AB,$F9,$B4,$63,$E0,$FF,$AA,$36,$3F
	.dc.b	$84,$7D,$67,$E3,$38,$BF,$D8,$7F,$AA,$7F,$A8,$A7,$FD,$40,$D7,$FD
	.dc.b	$3A,$D7,$FD,$9A,$D7,$FC,$BD,$CA,$FA,$1F,$E4,$F9,$FE,$5D,$B7,$A6
	.dc.b	$D7,$FA,$F7,$EB,$7F,$C6,$8F,$F1,$BF,$AB,$FF,$52,$BF,$45,$AF,$E1
	.dc.b	$9A,$FF,$B2,$9A,$FE,$BD,$AF,$FB,$25,$BC,$46,$9A,$C7,$F6,$3F,$01
	.dc.b	$FD,$57,$AC,$D3,$FB,$D4,$F8,$B3,$FD,$97,$C9,$3D,$5F,$EB,$8D,$CB
	.dc.b	$FA,$D3,$5F,$E9,$AB,$E7,$F7,$32,$76,$F2,$8F,$EC,$DF,$28,$E2,$A4
	.dc.b	$AF,$EC,$DE,$4F,$F8,$2F,$87,$E8,$D6,$E7,$FB,$53,$5F,$EB,$AD,$CB
	.dc.b	$F8,$55,$71,$9F,$8D,$DF,$EE,$23,$FD,$FC,$7D,$0B,$EB,$33,$1F,$CF
	.dc.b	$36,$FE,$05,$97,$D1,$BF,$F0,$DD,$EE,$BF,$BE,$7D,$F7,$FA,$C1,$BF
	.dc.b	$17,$FE,$03,$8C,$47,$3A,$FF,$03,$7C,$D7,$F2,$48,$FF,$02,$35,$FE
	.dc.b	$EE,$DD,$3F,$92,$3F,$CE,$DB,$C7,$C8,$93,$CB,$C9,$FF,$09,$7C,$DB
	.dc.b	$C4,$A4,$9F,$F0,$94,$7F,$8D,$AF,$F9,$83,$B8,$CC,$ED,$FF,$04,$C6
	.dc.b	$F5,$FF,$5F,$FD,$63,$E1,$7F,$C0,$8D,$AF,$F6,$63,$5F,$F6,$53,$5F
	.dc.b	$F4,$FB,$5E,$71,$C8,$8F,$F2,$49,$FC,$BF,$97,$FE,$51,$6C,$7F,$8D
	.dc.b	$DA,$FF,$86,$6B,$C1,$8F,$F4,$BE,$4F,$91,$5F,$8A,$CF,$F2,$27,$F9
	.dc.b	$67,$EB,$1F,$57,$8C,$7F,$1B,$35,$FE,$FB,$3A,$7E,$BE,$5F,$68,$8F
	.dc.b	$F8,$5F,$EA,$FF,$CD,$2E,$88,$7D,$C8,$FF,$34,$46,$A3,$FC,$95,$F5
	.dc.b	$0F,$AC,$FF,$36,$7D,$07,$79,$FF,$47,$DF,$F6,$F8,$C7,$C4,$7F,$76
	.dc.b	$BE,$15,$F9,$B6,$BF,$E1,$49,$DD,$71,$CF,$FD,$BD,$7C,$70,$EF,$18
	.dc.b	$9D,$5F,$FA,$81,$AF,$88,$FF,$80,$18,$76,$DB,$FB,$2E,$4F,$A9,$E5
	.dc.b	$4F,$09,$CD,$F8,$3D,$3F,$C1,$F1,$8E,$3D,$F8,$6C,$79,$61,$E1,$11
	.dc.b	$9C,$7A,$DF,$19,$D3,$5F,$34,$ED,$E4,$BE,$B1,$ED,$78,$1B,$3A,$6B
	.dc.b	$C9,$FE,$F9,$97,$F8,$9A,$B9,$57,$12,$57,$13,$E0,$8D,$7E,$27,$83
	.dc.b	$EB,$8C,$65,$FC,$6F,$0F,$B4,$DD,$D3,$FE,$77,$FA,$CF,$F5,$2D,$6A
	.dc.b	$F9,$61,$D2,$BE,$A1,$D2,$3E,$47,$BA,$BE,$38,$74,$AD,$1C,$2F,$96
	.dc.b	$1D,$63,$EF,$5F,$1F,$83,$F5,$D6,$AF,$93,$6B,$F0,$79,$7A,$5F,$45
	.dc.b	$70,$8C,$47,$83,$0E,$C7,$F9,$43,$0E,$D1,$EC,$9D,$DF,$58,$E1,$87
	.dc.b	$C4,$AD,$C7,$A6,$B6,$4C,$BE,$0E,$5F,$02,$7C,$B1,$D4,$8F,$BA,$34
	.dc.b	$F6,$36,$75,$8E,$79,$7A,$DE,$39,$7A,$B1,$F2,$5F,$2C,$BD,$04,$C7
	.dc.b	$E2,$65,$E8,$BE,$19,$76,$6C,$67,$0F,$9B,$5E,$BA,$F1,$1E,$7C,$BC
	.dc.b	$4E,$2F,$8E,$3E,$1A,$E1,$18,$AD,$BE,$0A,$F7,$61,$DD,$F1,$61,$E0
	.dc.b	$CB,$DA,$FB,$2F,$8B,$99,$F0,$39,$76,$72,$25,$75,$7E,$61,$32,$FC
	.dc.b	$44,$65,$F8,$D7,$CA,$31,$8F,$AA,$FC,$18,$FE,$1F,$2F,$75,$F8,$31
	.dc.b	$F5,$C6,$1F,$29,$E2,$70,$FF,$9D,$71,$FD,$43,$70,$FF,$05,$3E,$D3
	.dc.b	$5F,$59,$FC,$09,$FC,$FB,$5F,$E4,$2D,$7F,$7C,$D9,$3F,$C3,$30,$FE
	.dc.b	$6C,$DC,$7F,$35,$E5,$BE,$21,$D7,$FB,$CA,$D6,$3D,$B7,$C1,$FB,$CB
	.dc.b	$E1,$58,$9F,$DD,$57,$07,$C7,$5B,$BE,$B1,$A8,$E1,$8F,$91,$D5,$E3
	.dc.b	$1F,$7C,$5A,$6D,$FE,$61,$C3,$9E,$1C,$67,$A1,$1F,$DE,$65,$E6,$8C
	.dc.b	$61,$D1,$A7,$48,$D4,$F3,$24,$ED,$E7,$5E,$08,$C1,$2B,$E3,$1F,$C6
	.dc.b	$2B,$9C,$66,$3E,$19,$E0,$BF,$56,$3E,$D8,$CE,$1E,$07,$C9,$87,$77
	.dc.b	$AC,$65,$DD,$FD,$A9,$D4,$60,$DC,$79,$6F,$BC,$7C,$03,$46,$DC,$D7
	.dc.b	$5C,$FF,$13,$97,$D1,$9C,$DF,$AE,$73,$7E,$BC,$BD,$EF,$80,$87,$57
	.dc.b	$2F,$82,$38,$E1,$EF,$BE,$B3,$C2,$73,$7E,$9B,$E3,$18,$9D,$67,$F9
	.dc.b	$6B,$E3,$1D,$63,$F2,$53,$B9,$CB,$C6,$75,$8F,$E4,$1F,$89,$19,$38
	.dc.b	$5F,$82,$F9,$92,$B8,$63,$F5,$F8,$3F,$CF,$39,$FD,$76,$E9,$FE,$00
	.dc.b	$F1,$3F,$C2,$8D,$9D,$47,$C8,$9D,$E7,$FE,$08,$8D,$B6,$7A,$7D,$57
	.dc.b	$FA,$76,$FF,$C1,$FC,$EF,$BE,$CF,$95,$F8,$7E,$7F,$F0,$FC,$62,$B6
	.dc.b	$DC,$7F,$C5,$53,$A9,$FF,$0A,$F6,$BE,$B7,$D4,$D8,$FF,$04,$62,$FA
	.dc.b	$DF,$DD,$F6,$FE,$F3,$75,$EF,$2F,$F0,$1D,$EF,$0F,$F0,$5D,$EE,$B5
	.dc.b	$C0,$E6,$FF,$02,$5E,$F0,$FF,$24,$67,$F9,$C8,$C6,$1C,$63,$19,$F8
	.dc.b	$B2,$FF,$2A,$47,$C1,$A7,$D6,$E7,$0F,$F2,$14,$62,$BC,$67,$F8,$7B
	.dc.b	$47,$06,$6B,$FD,$43,$7C,$76,$FF,$24,$B7,$0D,$67,$F0,$6E,$E3,$FE
	.dc.b	$F6,$9B,$87,$F9,$47,$C3,$78,$FC,$DB,$5E,$29,$FF,$26,$FD,$6B,$C9
	.dc.b	$D5,$CA,$FF,$67,$BE,$BD,$55,$B9,$C9,$FE,$93,$F6,$9F,$2F,$F0,$61
	.dc.b	$18,$7C,$14,$FA,$D7,$D4,$6E,$B5,$9B,$7F,$BD,$5A,$C3,$FC,$75,$3F
	.dc.b	$26,$30,$E3,$0E,$98,$7F,$40,$8F,$F2,$04,$62,$B8,$C7,$F9,$B2,$B8
	.dc.b	$BA,$6B,$D6,$B9,$36,$78,$F0,$BC,$CF,$D3,$E3,$FE,$CA,$F0,$7E,$C4
	.dc.b	$7F,$41,$8C,$57,$FB,$42,$72,$6B,$AE,$C7,$FB,$8B,$AF,$EF,$1F,$7F
	.dc.b	$D3,$B1,$92,$65,$FE,$E6,$8C,$BF,$1E,$3F,$CD,$39,$3C,$F7,$E0,$5F
	.dc.b	$8B,$6C,$FD,$2F,$94,$7D,$6E,$6B,$EF,$46,$27,$53,$F9,$BC,$3F,$5F
	.dc.b	$EC,$6A,$24,$E8,$FC,$64,$E9,$C3,$E5,$3F,$29,$DD,$F9,$33,$F2,$9E
	.dc.b	$6D,$9F,$CA,$66,$EA,$FD,$B4,$61,$B7,$CE,$75,$5C,$63,$1E,$C3,$EA
	.dc.b	$3F,$54,$DC,$3A,$AF,$8B,$CC,$C9,$CB,$C2,$31,$5D,$FB,$1F,$49,$FB
	.dc.b	$56,$DF,$CB,$FA,$DC,$4F,$CA,$7F,$13,$5F,$EB,$AE,$04,$63,$AD,$18
	.dc.b	$BF,$CA,$C6,$27,$53,$9A,$FC,$F3,$5F,$CE,$B5,$FE,$18,$D7,$FD,$A5
	.dc.b	$F7,$FF,$C0,$1F,$90,$DE,$9F,$E5,$5F,$04,$FA,$DF,$E2,$13,$9C,$BF
	.dc.b	$D6,$7E,$93,$F4,$C7,$A9,$EC,$8C,$36,$33,$3F,$AE,$9C,$B7,$EB,$BF
	.dc.b	$6F,$97,$F8,$D2,$51,$89,$FE,$49,$F5,$DF,$ED,$67,$F8,$B4,$EB,$C2
	.dc.b	$FE,$49,$F1,$1F,$22,$7F,$0E,$DD,$BF,$1D,$C2,$FA,$EF,$C0,$47,$E1
	.dc.b	$E7,$59,$FF,$49,$9C,$CF,$F0,$66,$BF,$CD,$5A,$FF,$BC,$9B,$5D,$79
	.dc.b	$BE,$D4,$60,$8F,$E0,$3A,$5F,$01,$FC,$73,$0F,$F2,$9C,$63,$C8,$FE
	.dc.b	$59,$F1,$0E,$CF,$E8,$17,$BD,$7F,$A3,$4E,$5F,$C2,$B7,$4F,$E4,$0D
	.dc.b	$7F,$8F,$B5,$E3,$5F,$8F,$8C,$43,$44,$61,$B3,$E9,$B7,$BB,$E4,$3F
	.dc.b	$AA,$7C,$73,$EB,$71,$97,$EE,$23,$0F,$F2,$66,$E5,$FC,$E3,$44,$62
	.dc.b	$3F,$DE,$91,$8A,$F9,$13,$99,$FE,$36,$D5,$3B,$AE,$D5,$F3,$FE,$45
	.dc.b	$F4,$3F,$DA,$5B,$3F,$D4,$3E,$4B,$F8,$03,$FC,$3A,$33,$87,$E2,$27
	.dc.b	$2F,$F4,$B6,$BF,$D2,$9A,$FF,$79,$6B,$F2,$DB,$A9,$A4,$F0,$7C,$DF
	.dc.b	$31,$E5,$3F,$BF,$4B,$C7,$C6,$3F,$C8,$A3,$0F,$F5,$06,$E1,$F2,$DA
	.dc.b	$FC,$A6,$BF,$D0,$9A,$FC,$86,$E9,$F8,$1A,$FC,$B4,$63,$E5,$5F,$7B
	.dc.b	$FE,$11,$6E,$1F,$DE,$3E,$B1,$F5,$99,$AF,$F0,$24,$62,$7F,$8C,$B7
	.dc.b	$2F,$EE,$8D,$7F,$B9,$B5,$FF,$11,$35,$FC,$A3,$70,$34,$D6,$4C,$7F
	.dc.b	$84,$57,$E1,$1B,$F0,$0F,$9B,$FF,$69,$E3,$FD,$32,$73,$18,$AF,$F0
	.dc.b	$5D,$6F,$E0,$BF,$9B,$B6,$FF,$C0,$ED,$9F,$CD,$7D,$0F,$F2,$57,$F2
	.dc.b	$6D,$87,$D5,$3E,$AB,$FC,$75,$3B,$CC,$D3,$84,$EF,$0F,$EC,$31,$BE
	.dc.b	$D7,$D2,$7F,$90,$DB,$3F,$D9,$9F,$E0,$D9,$7E,$91,$CB,$FD,$95,$A7
	.dc.b	$F1,$02,$C7,$F8,$19,$E3,$87,$F8,$BE,$BF,$CB,$16,$FD,$61,$F6,$58
	.dc.b	$6E,$1F,$E4,$9F,$AC,$7D,$67,$F1,$3C,$BF,$DD,$F8,$71,$C3,$81,$2F
	.dc.b	$85,$6D,$67,$F1,$31,$EA,$FF,$81,$F1,$EB,$E0,$7F,$6A,$6D,$7F,$8E
	.dc.b	$9B,$3F,$E4,$46,$BF,$E4,$0F,$B8,$FE,$EA,$FF,$44,$6E,$6A,$F0,$FC
	.dc.b	$73,$EB,$3F,$CB,$97,$BF,$53,$FC,$C1,$87,$0B,$FF,$3C,$36,$57,$C7
	.dc.b	$E1,$37,$7F,$E2,$79,$F6,$FC,$57,$F9,$41,$F7,$87,$0E,$8A,$EF,$EB
	.dc.b	$3F,$8D,$FD,$55,$F1,$0D,$25,$7F,$97,$25,$F0,$1F,$E6,$98,$C3,$5F
	.dc.b	$F3,$4C,$6B,$6E,$D3,$99,$F2,$F2,$FF,$7E,$7D,$0D,$7F,$CD,$2D,$7F
	.dc.b	$CE,$6D,$78,$FA,$9E,$C8,$C4,$7D,$E6,$BF,$9C,$8F,$83,$F6,$5F,$D3
	.dc.b	$A3,$E3,$35,$FF,$3B,$B5,$FF,$40,$B5,$F9,$7F,$9B,$7F,$A7,$9F,$9C
	.dc.b	$D7,$FC,$FE,$D7,$FD,$12,$D7,$8F,$B9,$F4,$A3,$11,$F9,$46,$A7,$3E
	.dc.b	$13,$F6,$D7,$DB,$7D,$D8,$C3,$63,$35,$FE,$3C,$6B,$FE,$97,$6B,$FE
	.dc.b	$96,$6B,$FE,$C5,$6B,$FD,$DD,$BC,$EF,$AB,$FF,$57,$CF,$E8,$DA,$FE
	.dc.b	$8D,$AF,$E4,$23,$87,$A5,$F9,$36,$C7,$18,$C4,$F0,$8F,$DA,$CE,$9F
	.dc.b	$D7,$7C,$8B,$EB,$B9,$36,$3F,$68,$D7,$71,$F1,$6B,$FC,$E1,$1F,$E5
	.dc.b	$E6,$BF,$EB,$96,$BF,$EB,$86,$BE,$29,$FF,$4D,$BD,$BE,$1B,$C2,$47
	.dc.b	$04,$3C,$EE,$5F,$EA,$7F,$35,$F5,$FC,$A3,$49,$1F,$C3,$7E,$EB,$FB
	.dc.b	$A3,$7C,$EF,$91,$18,$8F,$E2,$0D,$7E,$73,$5F,$92,$D5,$9A,$30,$D9
	.dc.b	$FC,$57,$A2,$F8,$7A,$4B,$FF,$42,$C6,$5F,$37,$E2,$9E,$42,$38,$21
	.dc.b	$E6,$72,$D3,$9F,$27,$80,$8E,$0F,$F3,$D7,$DD,$BE,$BB,$F8,$D7,$C0
	.dc.b	$23,$13,$A9,$CC,$62,$3F,$DE,$F3,$B9,$C9,$23,$81,$FE,$A0,$FA,$C7
	.dc.b	$D7,$7E,$AD,$B1,$98,$C1,$DB,$E9,$E4,$79,$9A,$F8,$30,$46,$23,$F9
	.dc.b	$8E,$1E,$98,$C4,$7F,$52,$E9,$61,$FB,$06,$F4,$F8,$C9,$3F,$E9,$DF
	.dc.b	$89,$FF,$34,$B7,$E0,$3E,$2C,$62,$3E,$0B,$5F,$0B,$5F,$F5,$6B,$56
	.dc.b	$28,$C3,$67,$E8,$FD,$57,$F8,$CA,$BF,$D5,$CD,$7E,$7B,$5F,$7B,$5F
	.dc.b	$F5,$DF,$C9,$7F,$B4,$9F,$F5,$B3,$5F,$AE,$D7,$E3,$B5,$F3,$FB,$5E
	.dc.b	$E2,$3F,$0D,$FC,$2F,$EB,$4F,$2E,$92,$7F,$D8,$91,$97,$FB,$27,$D0
	.dc.b	$FF,$1E,$4D,$26,$9F,$D9,$9B,$1C,$23,$07,$F6,$DF,$65,$F4,$FF,$EA
	.dc.b	$B9,$FF,$1C,$B5,$FF,$1F,$B5,$FE,$E7,$1C,$3D,$3F,$E7,$5F,$31,$FE
	.dc.b	$9D,$9F,$F7,$9B,$5F,$EF,$27,$FA,$A5,$C3,$FD,$F3,$85,$18,$8F,$C6
	.dc.b	$35,$E1,$F0,$7D,$88,$FF,$C0,$0F,$A6,$FE,$B8,$FF,$AD,$5B,$46,$93
	.dc.b	$E1,$BF,$A8,$36,$79,$47,$F8,$33,$E9,$7D,$D1,$FE,$10,$6B,$FE,$30
	.dc.b	$6A,$69,$31,$E4,$48,$F8,$27,$FB,$27,$F5,$A7,$8D,$FD,$85,$B1,$AC
	.dc.b	$51,$88,$F9,$31,$89,$FE,$D1,$E2,$3E,$D6,$92,$B9,$46,$5B,$19,$79
	.dc.b	$C6,$5B,$1F,$DC,$FE,$81,$F2,$94,$93,$E5,$8C,$B6,$32,$F9,$A3,$2D
	.dc.b	$8F,$C9,$7D,$44,$8F,$CC,$B5,$F8,$BC,$88,$C4,$7E,$22,$30,$F0,$20
	.dc.b	$F5,$3F,$CA,$8F,$FB,$D7,$E9,$3F,$C4,$5E,$CF,$F3,$FD,$51,$8F,$23
	.dc.b	$FC,$97,$87,$4B,$E5,$EE,$FF,$1B,$E4,$FF,$01,$E3,$C5,$B7,$9F,$47
	.dc.b	$FC,$A7,$39,$F5,$AC,$E7,$0F,$F2,$CB,$5F,$F2,$C3,$67,$84,$68,$67
	.dc.b	$FD,$B3,$E7,$BF,$CB,$F9,$7E,$97,$89,$18,$AF,$E6,$11,$87,$FD,$07
	.dc.b	$3B,$9D,$47,$49,$D3,$F1,$8C,$46,$AB,$C1,$1C,$8E,$B1,$C9,$FE,$1B
	.dc.b	$8D,$D3,$FD,$5B,$1B,$3F,$D2,$6E,$1E,$C7,$03,$45,$32,$EF,$0F,$BD
	.dc.b	$3D,$49,$3F,$16,$DF,$9C,$3E,$AB,$E7,$CF,$C8,$8C,$3F,$2C,$ED,$9F
	.dc.b	$E0,$7A,$1D,$9C,$30,$FD,$C3,$63,$35,$DD,$FE,$2F,$5C,$67,$44,$9D
	.dc.b	$1C,$ED,$EE,$D2,$47,$A6,$1F,$1B,$88,$FC,$19,$E9,$9C,$5F,$E5,$EB
	.dc.b	$73,$EB,$D7,$F0,$3E,$3D,$DF,$FA,$FC,$F2,$3F,$A0,$B8,$BF,$F0,$F7
	.dc.b	$C0,$3E,$AB,$FD,$25,$1F,$E2,$1A,$DD,$F3,$23,$E4,$AE,$1E,$EF,$CE
	.dc.b	$8D,$9D,$F1,$F1,$B7,$1F,$07,$27,$FB,$13,$2E,$EF,$EB,$A3,$33,$F8
	.dc.b	$66,$BD,$32,$F3,$D7,$08,$D6,$5E,$49,$F4,$61,$E0,$77,$8F,$96,$31
	.dc.b	$8F,$7A,$E0,$F8,$E7,$58,$F8,$2B,$78,$76,$3E,$06,$3E,$17,$C9,$7C
	.dc.b	$DF,$76,$1F,$B7,$CB,$E2,$C7,$63,$96,$5E,$6C,$7D,$F3,$A9,$E2,$49
	.dc.b	$D5,$73,$C3,$C1,$8F,$92,$BC,$27,$4C,$7D,$4F,$73,$26,$B0,$ED,$87
	.dc.b	$48,$ED,$3E,$98,$F8,$F1,$C6,$BB,$1F,$0A,$B8,$D6,$C9,$3E,$13,$FD
	.dc.b	$EF,$7E,$6C,$7E,$51,$31,$F5,$65,$F3,$9E,$D1,$F0,$A7,$84,$60,$DD
	.dc.b	$53,$CE,$76,$AE,$4E,$0F,$74,$F2,$71,$8F,$A6,$31,$97,$C1,$26,$3F
	.dc.b	$22,$B7,$8F,$9D,$FB,$F8,$76,$AD,$C7,$BF,$2F,$86,$FE,$14,$FE,$A5
	.dc.b	$5F,$B0,$6E,$5F,$E3,$28,$C4,$7F,$4C,$F8,$BF,$E1,$6F,$F5,$96,$B9
	.dc.b	$DF,$FC,$73,$18,$8F,$E5,$2F,$F1,$2C,$3F,$3E,$DC,$92,$C5,$F5,$4F
	.dc.b	$2B,$FB,$0E,$1C,$31,$FE,$11,$8F,$1C,$3A,$C7,$E6,$6F,$87,$36,$A2
	.dc.b	$4F,$EF,$6B,$86,$3D,$B1,$EB,$97,$A6,$B5,$7C,$32,$FE,$1F,$97,$49
	.dc.b	$C9,$23,$C8,$FF,$0D,$8D,$46,$09,$5D,$5E,$77,$E0,$AF,$49,$CE,$74
	.dc.b	$7C,$4C,$3C,$D9,$78,$E3,$33,$C2,$F9,$CE,$63,$13,$AA,$E3,$1D,$4C
	.dc.b	$91,$CD,$79,$A3,$27,$F9,$C3,$2F,$8F,$8F,$B2,$FC,$33,$9C,$BC,$18
	.dc.b	$72,$9C,$EE,$C3,$EB,$5E,$CF,$8D,$9F,$C1,$71,$97,$85,$B7,$F0,$72
	.dc.b	$F0,$47,$93,$0E,$99,$78,$27,$33,$C0,$D6,$3E,$8C,$7B,$5F,$09,$F8
	.dc.b	$46,$F2,$F4,$3C,$30,$E9,$5C,$2F,$D3,$7C,$72,$F8,$D8,$70,$C3,$C0
	.dc.b	$49,$D6,$5E,$9C,$39,$63,$F5,$6B,$C1,$97,$A2,$F7,$1E,$13,$E2,$E7
	.dc.b	$B8,$CC,$70,$BE,$6F,$AA,$38,$E3,$E3,$79,$F2,$3E,$D1,$F1,$3F,$8F
	.dc.b	$8C,$36,$1F,$87,$58,$99,$7F,$85,$67,$8B,$67,$38,$FF,$7E,$B7,$D6
	.dc.b	$7D,$66,$E7,$DE,$3F,$C7,$B7,$FE,$1F,$74,$48,$D7,$0E,$DF,$C2,$7D
	.dc.b	$4B,$E0,$7D,$1C,$EF,$85,$CC,$61,$B8,$FF,$8C,$AB,$83,$67,$FC,$D2
	.dc.b	$FF,$AD,$A3,$39,$FF,$8E,$9B,$19,$9D,$CF,$F1,$41,$3A,$70,$D7,$FC
	.dc.b	$BB,$E2,$BE,$9B,$8C,$61,$BA,$A7,$58,$7F,$93,$63,$13,$A9,$CD,$7F
	.dc.b	$96,$E7,$35,$B9,$CB,$FE,$CD,$6C,$FF,$2A,$9F,$F5,$6F,$27,$DF,$F2
	.dc.b	$7F,$C8,$B7,$FE,$4F,$7F,$CC,$5C,$FF,$E8,$97,$37,$F6,$DB,$19,$AE
	.dc.b	$53,$CD,$FA,$BE,$4F,$FA,$53,$C6,$FF,$4A,$E3,$FE,$92,$FA,$BC,$E5
	.dc.b	$FC,$F2,$73,$E3,$A7,$38,$FF,$B2,$36,$FF,$63,$DF,$FB,$2A,$3E,$7E
	.dc.b	$3F,$E6,$E7,$FA,$2C,$E9,$E1,$19,$77,$3C,$67,$51,$F7,$E7,$96,$2F
	.dc.b	$D7,$9F,$6F,$86,$E1,$FE,$D5,$6D,$FF,$B5,$1A,$FF,$BE,$DA,$FC,$F6
	.dc.b	$BF,$E9,$E6,$D7,$7A,$F0,$EE,$FA,$DA,$4C,$3F,$DA,$0D,$C3,$FD,$9C
	.dc.b	$D7,$EF,$3E,$AF,$21,$F6,$34,$95,$FE,$E4,$7E,$27,$87,$FD,$5E,$FF
	.dc.b	$34,$1C,$3F,$9D,$6B,$FD,$C9,$AE,$FA,$39,$92,$BE,$B4,$F0,$CD,$F5
	.dc.b	$CF,$5F,$E7,$B6,$BF,$E0,$03,$70,$FF,$7F,$B5,$FD,$75,$72,$AE,$24
	.dc.b	$AE,$2F,$DA,$F3,$7C,$04,$C7,$04,$9F,$7D,$FF,$B4,$27,$51,$8B,$47
	.dc.b	$A7,$D4,$7D,$17,$E1,$9B,$5B,$6D,$FC,$AB,$46,$23,$F8,$67,$85,$F9
	.dc.b	$3E,$4F,$E4,$F1,$F9,$D8,$C1,$2F,$F2,$CD,$8D,$47,$A3,$C8,$7E,$37
	.dc.b	$FC,$42,$30,$D9,$FB,$73,$AB,$FD,$33,$6B,$97,$D2,$FC,$DA,$FF,$16
	.dc.b	$CA,$31,$08,$C6,$DC,$A7,$30,$F9,$07,$C2,$76,$6E,$1E,$76,$E1,$E4
	.dc.b	$6B,$FC,$A6,$9F,$4D,$F9,$1A,$F7,$36,$7D,$6D,$78,$35,$F4,$E1,$F3
	.dc.b	$DB,$BF,$9C,$95,$F0,$1B,$1F,$1A,$78,$3E,$F6,$EE,$F8,$07,$D8,$7F
	.dc.b	$06,$6E,$1F,$87,$6D,$FE,$05,$AF,$B7,$E9,$9F,$5B,$F4,$9B,$5F,$6D
	.dc.b	$B3,$F6,$1A,$F1,$9E,$04,$9E,$0F,$8F,$DC,$91,$E3,$6C,$F3,$FA,$1F
	.dc.b	$01,$27,$C0,$D8,$D4,$60,$8F,$E6,$FE,$31,$F0,$9F,$C9,$9B,$3F,$A7
	.dc.b	$6C,$EE,$7F,$9A,$7C,$77,$F2,$AF,$B1,$BE,$11,$A4,$C3,$F6,$73,$B6
	.dc.b	$CE,$DB,$3B,$BF,$27,$D1,$F8,$C9,$3E,$37,$BF,$CC,$C8,$E6,$D9,$FE
	.dc.b	$3F,$87,$F3,$E8,$C4,$7F,$7B,$8C,$4F,$03,$F8,$27,$8D,$FC,$6C,$9D
	.dc.b	$FF,$D2,$1F,$E4,$B2,$FB,$CF,$B3,$7F,$E9,$88,$C7,$07,$C4,$9C,$CA
	.dc.b	$73,$87,$C2,$9C,$C6,$27,$CF,$18,$38,$F2,$3E,$51,$C3,$FD,$69,$B8
	.dc.b	$7F,$1B,$6D,$FF,$15,$6B,$EF,$BA,$F9,$DA,$F9,$5B,$87,$B6,$FE,$EE
	.dc.b	$5F,$56,$BE,$5B,$5F,$94,$D7,$E1,$35,$F8,$0D,$93,$49,$B7,$F7,$4F
	.dc.b	$90,$7C,$7D,$26,$7F,$CD,$E3,$2D,$8C,$E1,$FC,$CE,$72,$D8,$FF,$04
	.dc.b	$79,$AF,$CE,$BF,$CC,$57,$FD,$12,$31,$EC,$EA,$73,$7F,$D9,$A7,$35
	.dc.b	$B9,$CC,$FF,$90,$E7,$2E,$E3,$2F,$F8,$7D,$AF,$FB,$73,$EB,$DF,$2F
	.dc.b	$FE,$17,$6D,$FF,$87,$DB,$3F,$D6,$1A,$FE,$5F,$E8,$1D,$5F,$B0,$6D
	.dc.b	$7E,$B1,$B3,$FA,$06,$BC,$67,$81,$27,$83,$FE,$29,$8F,$F7,$8F,$A9
	.dc.b	$FE,$31,$69,$5C,$27,$FC,$6D,$E1,$BE,$0D,$F3,$B8,$6D,$7F,$72,$9D
	.dc.b	$4E,$4E,$3E,$D3,$6A,$4A,$E1,$19,$7F,$BD,$FC,$47,$F8,$65,$FF,$23
	.dc.b	$35,$CF,$A1,$E9,$8C,$47,$F8,$81,$AF,$F9,$ED,$A8,$38,$6C,$7C,$3F
	.dc.b	$9D,$7C,$75,$24,$FF,$95,$E3,$2F,$F9,$67,$E7,$5F,$31,$F2,$DB,$19
	.dc.b	$8C,$11,$C1,$0F,$F0,$CC,$70,$7F,$C2,$1F,$3A,$F9,$4F,$F3,$37,$80
	.dc.b	$8C,$4F,$F8,$75,$AF,$09,$DC,$62,$3F,$C5,$ED,$7F,$D1,$0D,$53,$D6
	.dc.b	$7F,$CE,$9D,$AF,$9D,$7C,$86,$1B,$7F,$CF,$9A,$FF,$32,$6B,$FC,$C5
	.dc.b	$AF,$F2,$96,$BF,$C9,$DB,$3D,$E7,$C3,$DF,$FE,$C3,$AF,$E9,$6D,$7F
	.dc.b	$8A,$FC,$1E,$8F,$8B,$6F,$1E,$92,$7F,$C8,$51,$97,$9F,$67,$62,$38
	.dc.b	$21,$FC,$89,$CB,$4F,$E6,$5E,$DF,$D4,$F1,$4D,$24,$7F,$62,$F9,$D7
	.dc.b	$CE,$FC,$C6,$C6,$63,$04,$70,$43,$FA,$EC,$70,$7F,$BD,$7C,$6F,$E1
	.dc.b	$3C,$E3,$49,$1E,$1F,$9D,$F0,$11,$C1,$0E,$AE,$5A,$7F,$80,$F9,$7B
	.dc.b	$11,$C1,$FE,$59,$F9,$DF,$F2,$3B,$7D,$8F,$F5,$E7,$B0,$9C,$CF,$F9
	.dc.b	$69,$AF,$F9,$1D,$AF,$F9,$61,$AF,$F9,$5D,$B1,$CE,$7F,$47,$F3,$AF
	.dc.b	$2E,$92,$FF,$D0,$11,$97,$CD,$E5,$F9,$30,$FE,$A3,$18,$21,$E5,$8D
	.dc.b	$11,$C9,$FE,$98,$9F,$D4,$74,$9A,$4E,$9F,$83,$F3,$BF,$E4,$FA,$FF
	.dc.b	$3F,$35,$FF,$3D,$B5,$FF,$34,$35,$FF,$3C,$B5,$FF,$3C,$37,$33,$4C
	.dc.b	$7C,$EF,$F8,$22,$7F,$D1,$8D,$7F,$D1,$2D,$7F,$CE,$CD,$7F,$D0,$FE
	.dc.b	$93,$EB,$FE,$CB,$63,$FD,$2C,$D7,$FD,$26,$D7,$D3,$EB,$7C,$02,$3F
	.dc.b	$01,$FC,$1F,$CF,$7F,$B1,$E3,$FD,$40,$D7,$FD,$3A,$D7,$FD,$14,$D7
	.dc.b	$F9,$C0,$F9,$2F,$E3,$0D,$CF,$FA,$D4,$60,$8E,$09,$1F,$EC,$63,$81
	.dc.b	$FE,$C8,$F8,$C7,$CF,$7F,$70,$8C,$35,$FF,$75,$C6,$BC,$FE,$4D,$8F
	.dc.b	$F6,$47,$CF,$7F,$36,$F8,$63,$49,$1A,$21,$FE,$CD,$8D,$B8,$3F,$DC
	.dc.b	$BF,$0F,$FD,$39,$FF,$69,$35,$CF,$95,$FD,$12,$31,$1F,$D7,$79,$3F
	.dc.b	$DB,$F3,$92,$47,$FB,$8E,$76,$FF,$63,$F9,$AF,$73,$F3,$83,$87,$E8
	.dc.b	$B5,$E1,$97,$F8,$26,$31,$1C,$23,$04,$7F,$DE,$5F,$21,$FE,$44,$6F
	.dc.b	$8F,$FD,$72,$31,$1F,$EE,$F6,$BF,$80,$C7,$F9,$6F,$CF,$7F,$62,$F0
	.dc.b	$CD,$24,$FF,$BF,$E3,$2D,$8C,$BF,$EF,$B8,$CB,$FE,$FD,$F8,$E7,$CF
	.dc.b	$7F,$6E,$6C,$7C,$33,$FB,$2B,$87,$84,$6C,$8E,$ED,$3B,$8F,$53,$5F
	.dc.b	$CD,$F1,$BE,$7B,$F3,$24,$8F,$CB,$46,$5B,$19,$7C,$91,$96,$C6,$7D
	.dc.b	$47,$B3,$F8,$3F,$93,$FC,$57,$B5,$FD,$1A,$FF,$BF,$C6,$3C,$7F,$1E
	.dc.b	$1F,$E5,$FB,$FC,$57,$1F,$F7,$CC,$38,$68,$8C,$5E,$BD,$0F,$DA,$D6
	.dc.b	$A7,$FC,$5B,$97,$F9,$3A,$B4,$D3,$FC,$1F,$1C,$9B,$5C,$A7,$C3,$EB
	.dc.b	$59,$FF,$0A,$F0,$3E,$5B,$F8,$1E,$7F,$85,$8C,$4F,$F8,$F6,$31,$3A
	.dc.b	$8F,$D2,$4E,$9F,$8A,$62,$34,$F6,$6E,$7F,$DF,$1B,$3C,$6B,$F4,$DB
	.dc.b	$FB,$AF,$99,$FE,$4D,$71,$1A,$C3,$AC,$FF,$19,$6C,$EA,$3F,$07,$18
	.dc.b	$23,$EB,$32,$7C,$9C,$1F,$2C,$F9,$EF,$8D,$18,$7E,$44,$F7,$38,$74
	.dc.b	$3B,$36,$D8,$CC,$61,$CB,$1C,$9C,$8E,$F7,$FE,$83,$7C,$8D,$7F,$D1
	.dc.b	$1A,$FF,$13,$8C,$70,$7E,$29,$7C,$F7,$2F,$7A,$5F,$27,$AC,$7A,$B5
	.dc.b	$FC,$07,$8F,$F8,$9A,$FE,$85,$C2,$8C,$4E,$A7,$C2,$3E,$6D,$F3,$DF
	.dc.b	$E8,$9C,$FF,$45,$3C,$AF,$8E,$1C,$E3,$89,$93,$9D,$F2,$F7,$7C,$93
	.dc.b	$BB,$E7,$8F,$85,$B1,$9D,$7F,$D2,$17,$C7,$0E,$F8,$C7,$BE,$1C,$6B
	.dc.b	$79,$79,$DF,$E0,$F8,$73,$AF,$C7,$35,$F3,$63,$FB,$DC,$7B,$47,$49
	.dc.b	$D6,$1D,$63,$8E,$3F,$B3,$AD,$DF,$A2,$3C,$8D,$9F,$83,$8F,$8D,$E2
	.dc.b	$75,$9D,$E3,$E8,$CB,$E2,$C6,$27,$77,$E5,$BE,$51,$E8,$BF,$6D,$6F
	.dc.b	$0F,$1C,$62,$7B,$12,$F9,$61,$DA,$31,$3B,$9C,$C7,$84,$EC,$78,$A3
	.dc.b	$CC,$E7,$1F,$88,$EA,$FC,$2E,$C8,$67,$1F,$4C,$F0,$26,$5F,$56,$B8
	.dc.b	$D7,$80,$93,$F4,$63,$2F,$0C,$BE,$05,$F4,$BF,$AE,$48,$FA,$D1,$97
	.dc.b	$D1,$1D,$E3,$17,$F0,$8F,$45,$F6,$9E,$51,$98,$C5,$72,$3A,$E5,$F2
	.dc.b	$F2,$FC,$01,$32,$F8,$F7,$F4,$8F,$E4,$38,$7E,$9D,$B7,$FD,$62,$31
	.dc.b	$1F,$D7,$BE,$7B,$F8,$73,$FC,$A1,$AE,$77,$FE,$BF,$18,$8F,$E5,$4F
	.dc.b	$F1,$5C,$3F,$5A,$DC,$3F,$57,$F3,$CF,$27,$F9,$DD,$6B,$0F,$E0,$77
	.dc.b	$BC,$39,$C7,$E7,$EF,$87,$36,$A2,$4F,$EC,$AB,$86,$AB,$DC,$E6,$B7
	.dc.b	$87,$78,$FE,$0B,$3A,$8C,$13,$2F,$74,$EB,$2F,$3C,$62,$75,$3E,$63
	.dc.b	$59,$76,$9D,$12,$3F,$8E,$5F,$78,$C4,$7F,$99,$72,$F8,$0D,$C7,$E0
	.dc.b	$63,$E0,$8E,$D8,$F8,$1D,$1E,$03,$DC,$47,$CB,$97,$C3,$9E,$53,$C2
	.dc.b	$79,$6E,$C3,$E4,$E5,$E5,$AE,$F9,$79,$AB,$59,$79,$33,$F7,$56,$89
	.dc.b	$9F,$78,$CC,$77,$BF,$85,$39,$8F,$34,$75,$C7,$C3,$9F,$9B,$1E,$F5
	.dc.b	$A3,$AE,$5E,$57,$A1,$E1,$2E,$1D,$73,$F6,$E1,$F3,$2F,$9C,$7F,$2F
	.dc.b	$BE,$51,$8C,$7C,$59,$FB,$71,$F0,$1D,$8E,$B5,$C3,$0E,$84,$9E,$18
	.dc.b	$78,$1F,$8B,$8F,$8F,$0E,$B8,$7A,$E3,$0F,$A7,$1F,$13,$EE,$CB,$E1
	.dc.b	$63,$FA,$3E,$57,$C3,$3E,$5B,$F0,$91,$86,$C3,$E7,$AC,$4D,$FF,$C2
	.dc.b	$53,$C5,$B3,$9C,$7F,$B8,$DB,$E7,$BF,$04,$DC,$BB,$C7,$F8,$EE,$FF
	.dc.b	$C3,$8E,$89,$1A,$E1,$DB,$F9,$0F,$97,$7C,$07,$A3,$A5,$F0,$B9,$8C
	.dc.b	$37,$1F,$F1,$56,$61,$7C,$A3,$31,$8A,$FF,$22,$C7,$FA,$8E,$75,$3F
	.dc.b	$E3,$E3,$51,$C2,$32,$EE,$78,$CF,$FB,$62,$72,$FF,$AD,$7C,$47,$CF
	.dc.b	$6A,$3F,$D1,$AD,$D7,$FD,$12,$D9,$DE,$1E,$A9,$CB,$63,$35,$FE,$52
	.dc.b	$9C,$B6,$33,$3F,$EA,$BF,$23,$FC,$84,$FF,$92,$DB,$3F,$61,$FF,$28
	.dc.b	$73,$FF,$A0,$1C,$B7,$9F,$F0,$F1,$8E,$7F,$F6,$83,$C1,$B1,$F5,$FC
	.dc.b	$9F,$F3,$7F,$8F,$FE,$70,$CD,$E3,$BE,$7B,$5A,$7D,$5A,$D4,$E6,$3F
	.dc.b	$C8,$1B,$B0,$FA,$5A,$FD,$B8,$C4,$FF,$BF,$5B,$1F,$EB,$96,$BF,$6E
	.dc.b	$78,$BF,$82,$6C,$6A,$3E,$C4,$76,$C5,$F3,$CF,$93,$C3,$70,$FF,$60
	.dc.b	$36,$FF,$DD,$55,$C2,$B6,$4A,$DB,$67,$FD,$A5,$F3,$DD,$F8,$4D,$25
	.dc.b	$7F,$B3,$63,$2D,$8C,$CF,$28,$FF,$71,$7C,$F7,$A7,$C9,$34,$91,$FE
	.dc.b	$DD,$97,$87,$FD,$79,$FD,$31,$FA,$77,$0E,$1A,$E7,$A7,$E5,$C6,$23
	.dc.b	$9C,$60,$93,$FE,$E5,$F9,$E7,$CE,$FC,$D6,$C6,$63,$04,$70,$4A,$F9
	.dc.b	$B3,$C1,$F9,$FE,$6F,$93,$6F,$F6,$B6,$1F,$66,$30,$47,$04,$BF,$F7
	.dc.b	$C4,$71,$F4,$DF,$39,$F8,$16,$CF,$FB,$F5,$B8,$7D,$E6,$BF,$C1,$FC
	.dc.b	$3F,$C5,$92,$7F,$28,$D8,$FC,$A3,$5E,$11,$F4,$1F,$E2,$DE,$43,$E7
	.dc.b	$BF,$8B,$CE,$67,$FD,$41,$18,$78,$4E,$F9,$BF,$4A,$DA,$E0,$BE,$77
	.dc.b	$E4,$57,$DE,$8C,$6F,$C2,$73,$3E,$26,$BF,$EA,$FF,$92,$7C,$27,$66
	.dc.b	$E1,$D9,$B3,$E4,$6B,$DF,$E7,$BF,$1F,$3E,$46,$C7,$A9,$AE,$F8,$FC
	.dc.b	$9B,$CA,$F5,$E3,$F4,$2B,$CD,$39,$7E,$1C,$65,$FF,$63,$37,$3F,$85
	.dc.b	$3F,$BD,$F9,$D7,$CF,$52,$61,$F3,$F9,$13,$A8,$F9,$7F,$3C,$F9,$DF
	.dc.b	$9E,$D8,$CC,$60,$8E,$09,$9F,$D8,$9E,$0F,$F0,$8A,$7B,$9E,$52,$57
	.dc.b	$8E,$32,$F8,$BE,$73,$C2,$47,$04,$AE,$D1,$96,$C7,$B7,$E3,$5F,$09
	.dc.b	$FC,$7D,$B1,$98,$C1,$FC,$C3,$E5,$3F,$91,$7C,$E5,$F0,$8D,$25,$7E
	.dc.b	$5E,$76,$D9,$DB,$67,$78,$7F,$28,$C1,$F2,$BE,$32,$57,$9D,$F0,$7C
	.dc.b	$F4,$8E,$8D,$AF,$D5,$63,$FC,$DE,$31,$1F,$C7,$23,$13,$C0,$FB,$DE
	.dc.b	$37,$F0,$92,$78,$1F,$D0,$5F,$E3,$F2,$F9,$EF,$AF,$7C,$DA,$FF,$83
	.dc.b	$63,$5C,$5F,$1E,$73,$29,$CE,$1F,$16,$73,$18,$9F,$4C,$70,$E3,$7C
	.dc.b	$AB,$87,$FA,$A3,$70,$FC,$B3,$6F,$FA,$8B,$5F,$34,$F8,$C9,$3E,$17
	.dc.b	$EE,$65,$F5,$27,$E4,$B5,$F9,$0D,$7D,$CD,$7D,$8D,$C7,$FB,$B6,$FD
	.dc.b	$09,$7F,$AB,$F9,$17,$C8,$7F,$6D,$6C,$66,$30,$47,$04,$C3,$FB,$6C
	.dc.b	$F0,$7F,$94,$79,$8F,$9E,$FF,$2E,$D7,$F2,$78,$C7,$B5,$FE,$08,$6C
	.dc.b	$66,$77,$5C,$9A,$FF,$62,$9D,$B9,$AE,$53,$A8,$FE,$CF,$F3,$CF,$98
	.dc.b	$FF,$06,$B7,$0F,$ED,$CD,$9F,$E7,$4D,$7F,$29,$F3,$DF,$BC,$7F,$32
	.dc.b	$D7,$3D,$7F,$B1,$8C,$47,$E3,$DA,$FE,$35,$BB,$1A,$6A,$3F,$BE,$FA
	.dc.b	$DF,$E0,$17,$FC,$C8,$D4,$D3,$FC,$0D,$E2,$3E,$11,$F3,$D8,$6D,$7F
	.dc.b	$55,$9D,$4E,$4E,$3E,$DB,$82,$92,$B8,$46,$5F,$EC,$5F,$44,$FB,$BF
	.dc.b	$F7,$F4,$7F,$91,$77,$FF,$3D,$4E,$7B,$3E,$B4,$62,$B7,$5D,$E3,$FC
	.dc.b	$37,$F1,$5F,$E1,$E8,$E4,$2F,$93,$86,$BC,$4D,$57,$4C,$57,$D3,$C9
	.dc.b	$F0,$D7,$FA,$37,$E7,$DF,$35,$F3,$5B,$1F,$E2,$F6,$BF,$DE,$DA,$FE
	.dc.b	$3F,$F4,$9F,$33,$E1,$C1,$18,$7F,$CD,$6D,$C3,$FC,$6F,$5D,$27,$38
	.dc.b	$79,$1E,$2D,$AE,$17,$E7,$C3,$D3,$F4,$BE,$F8,$F8,$0D,$BF,$7C,$23
	.dc.b	$18,$FF,$AA,$FC,$07,$D0,$BE,$5F,$9E,$1C,$30,$E6,$D3,$9E,$1C,$A1
	.dc.b	$7C,$AB,$C7,$F1,$6F,$BA,$A4,$AF,$B3,$D2,$FA,$27,$C9,$E6,$7D,$37
	.dc.b	$C5,$FB,$F9,$23,$11,$F7,$6F,$95,$7D,$AC,$BA,$79,$5F,$E6,$56,$D7
	.dc.b	$F8,$0F,$1F,$EA,$AD,$9F,$F3,$56,$2F,$AA,$7D,$E7,$CC,$6D,$78,$9A
	.dc.b	$FE,$37,$0E,$AD,$F2,$5F,$01,$E1,$6C,$7F,$3E,$8C,$11,$C1,$0F,$E8
	.dc.b	$2F,$92,$9F,$5F,$E3,$24,$FE,$8A,$FF,$A6,$D7,$C1,$F3,$DF,$11,$F0
	.dc.b	$5B,$5F,$E0,$78,$C1,$1C,$10,$FE,$F4,$FC,$4F,$AD,$FC,$A7,$37,$FB
	.dc.b	$C7,$54,$62,$BF,$18,$DC,$58,$78,$27,$E3,$3F,$A6,$7F,$20,$D7,$F1
	.dc.b	$2D,$8E,$3C,$FE,$92,$4F,$F8,$7D,$AF,$F9,$BE,$FC,$78,$FE,$8B,$0F
	.dc.b	$07,$37,$E5,$9B,$97,$83,$0F,$C3,$7D,$13,$CB,$FC,$33,$72,$FF,$24
	.dc.b	$35,$FF,$61,$B5,$FF,$56,$E6,$8C,$47,$EA,$DA,$FE,$91,$B1,$B9,$FC
	.dc.b	$7F,$D7,$BE,$F3,$F7,$2D,$BF,$CF,$B5,$FC,$75,$F8,$F3,$FF,$79,$3F
	.dc.b	$B7,$8F,$3F,$57,$A8,$8E,$08,$7A,$9C,$BF,$26,$9F,$3B,$E2,$24,$7E
	.dc.b	$36,$BB,$3F,$CA,$FD,$C7,$D1,$FF,$86,$E3,$FD,$2A,$D7,$FD,$36,$D7
	.dc.b	$FD,$2C,$D7,$FD,$4A,$D5,$1D,$3D,$4F,$F6,$E3,$63,$AC,$7E,$EF,$EE
	.dc.b	$BF,$83,$C7,$E8,$5A,$FE,$CD,$B9,$FF,$20,$AE,$53,$E9,$FA,$5F,$F3
	.dc.b	$3B,$FD,$2A,$11,$88,$D7,$B7,$FD,$12,$5E,$4B,$E9,$71,$9F,$F7,$19
	.dc.b	$F8,$77,$E8,$8C,$1F,$DD,$79,$DF,$37,$FC,$63,$3F,$E5,$2D,$E6,$FD
	.dc.b	$63,$6F,$E1,$FD,$2F,$FB,$B9,$FE,$BF,$F2,$6F,$A5,$FF,$46,$C7,$FB
	.dc.b	$2E,$31,$EA,$FF,$B3,$61,$18,$75,$3C,$1F,$F6,$B4,$65,$B1,$94,$48
	.dc.b	$D3,$83,$AC,$6D,$DE,$0F,$A9,$FD,$7C,$7F,$6E,$6C,$67,$B7,$FB,$7D
	.dc.b	$F8,$73,$FE,$97,$EB,$FF,$85,$70,$FE,$7E,$D7,$FA,$DB,$63,$FA,$DF
	.dc.b	$63,$E3,$5F,$4D,$FD,$8A,$B5,$97,$F1,$76,$E1,$FD,$85,$B7,$C6,$BF
	.dc.b	$B4,$FD,$B3,$E9,$29,$2F,$F9,$77,$C8,$BE,$9F,$11,$FD,$BA,$B8,$3F
	.dc.b	$D5,$9A,$FF,$45,$9E,$0F,$F3,$7F,$9A,$7C,$27,$F5,$76,$C6,$63,$FC
	.dc.b	$26,$BE,$97,$EC,$6F,$F0,$2B,$73,$FD,$45,$B1,$9A,$E3,$5C,$36,$FE
	.dc.b	$71,$F6,$FA,$6C,$FF,$82,$23,$18,$7F,$7A,$C2,$F4,$3E,$4F,$ED,$B7
	.dc.b	$93,$FC,$1F,$1F,$7B,$87,$FA,$FF,$87,$89,$98,$7F,$78,$AD,$79,$7F
	.dc.b	$C3,$A3,$FB,$DF,$DC,$3E,$63,$FC,$32,$D8,$CC,$7F,$7B,$F9,$A7,$D4
	.dc.b	$EA,$24,$E8,$FE,$C5,$3A,$7F,$C8,$A3,$E9,$9F,$53,$FE,$05,$F9,$FF
	.dc.b	$A7,$3F,$E6,$6D,$3F,$C4,$51,$8F,$88,$FE,$B1,$7C,$2B,$9B,$5F,$F0
	.dc.b	$04,$EE,$B8,$F4,$FF,$9D,$1F,$D3,$56,$F9,$FF,$2F,$5B,$BF,$F1,$3B
	.dc.b	$78,$4E,$63,$D1,$97,$F8,$6F,$A3,$F8,$EE,$7F,$DE,$AD,$18,$CB,$FC
	.dc.b	$74,$DB,$FF,$19,$AF,$84,$F2,$46,$3D,$BB,$7C,$9B,$5C,$63,$AD,$6B
	.dc.b	$89,$F4,$9C,$B9,$BF,$09,$E2,$5E,$6F,$E1,$9A,$F9,$6F,$7A,$7F,$A5
	.dc.b	$35,$FF,$27,$46,$23,$E8,$1F,$E5,$3A,$E3,$3F,$E0,$D6,$C7,$F9,$32
	.dc.b	$78,$57,$D4,$CB,$D3,$F4,$9F,$AE,$8F,$88,$DC,$3E,$8C,$FC,$0B,$FC
	.dc.b	$13,$C3,$BE,$47,$94,$FE,$3F,$C7,$F1,$DC,$1C,$0D,$6C,$C3,$B4,$7F
	.dc.b	$98,$1A,$EC,$F6,$C7,$EE,$7C,$3F,$9F,$DF,$FC,$D2,$79,$B5,$DB,$FC
	.dc.b	$B5,$B1,$C0,$8F,$03,$E0,$FC,$54,$7C,$B7,$FC,$9F,$4F,$87,$F5,$39
	.dc.b	$9F,$1C,$E7,$5F,$F4,$03,$63,$D5,$3F,$1F,$3F,$BF,$E5,$FC,$26,$CF
	.dc.b	$D8,$3F,$59,$C7,$FD,$E1,$FB,$7F,$02,$FA,$9F,$F5,$0E,$7F,$AA,$AD
	.dc.b	$37,$0E,$91,$8C,$FF,$A7,$DF,$1F,$86,$F1,$4E,$71,$FA,$AD,$7C,$37
	.dc.b	$F6,$6B,$84,$74,$88,$F9,$F2,$F4,$E3,$FE,$62,$AD,$CE,$6F,$A5,$6B
	.dc.b	$1F,$DC,$C7,$86,$B7,$7E,$9B,$E1,$87,$68,$E9,$97,$C0,$C3,$8C,$E5
	.dc.b	$B5,$F1,$A7,$4F,$B2,$74,$D3,$59,$7C,$28,$CC,$7C,$03,$26,$32,$F4
	.dc.b	$3C,$70,$F1,$D7,$18,$F3,$92,$32,$F1,$CB,$E0,$9E,$FC,$3C,$B1,$DE
	.dc.b	$7E,$01,$31,$FA,$F3,$AC,$3C,$35,$B9,$F7,$D7,$68,$F0,$46,$08,$FF
	.dc.b	$AE,$EB,$95,$F8,$B1,$F8,$79,$7D,$0B,$DD,$F5,$3D,$CF,$8B,$2F,$76
	.dc.b	$3E,$59,$C9,$D6,$33,$3C,$B2,$F8,$F7,$D6,$B6,$75,$3B,$39,$71,$1E
	.dc.b	$E3,$8E,$5E,$8C,$7E,$51,$F9,$37,$13,$F0,$0F,$1D,$73,$AE,$59,$7B
	.dc.b	$E7,$ED,$9F,$58,$FE,$05,$1F,$B0,$6C,$7E,$BD,$B7,$FE,$3F,$97,$D4
	.dc.b	$FF,$0B,$9F,$EA,$2D,$7F,$8E,$B5,$FE,$C8,$D8,$E3,$BF,$EF,$E3,$F8
	.dc.b	$B6,$9F,$A7,$F3,$BF,$CC,$3E,$3E,$6A,$61,$FA,$DB,$DE,$BF,$BA,$BD
	.dc.b	$CE,$6B,$78,$7F,$6C,$BD,$CF,$FB,$EE,$B7,$39,$AD,$DF,$A9,$FD,$F4
	.dc.b	$EA,$31,$5D,$49,$5E,$7F,$45,$8F,$29,$F8,$38,$5D,$7F,$CE,$75,$A9
	.dc.b	$E3,$87,$4C,$3C,$51,$FC,$73,$1F,$03,$C8,$D9,$63,$C1,$8F,$78,$F9
	.dc.b	$B8,$F7,$AE,$31,$83,$89,$6B,$8B,$8C,$FE,$9B,$DB,$2F,$3E,$1D,$67
	.dc.b	$35,$E6,$38,$97,$1F,$51,$2F,$E7,$65,$EF,$9D,$47,$A6,$75,$87,$68
	.dc.b	$F1,$67,$F5,$9F,$34,$FD,$DB,$E4,$F5,$8C,$9C,$CE,$F8,$F9,$EB,$96
	.dc.b	$3E,$8A,$E5,$39,$CF,$E3,$61,$D3,$0E,$47,$43,$AE,$3E,$5C,$BD,$31
	.dc.b	$9C,$7A,$3E,$03,$D3,$7D,$30,$ED,$1F,$2F,$3F,$A7,$5C,$23,$96,$7F
	.dc.b	$94,$C7,$B6,$7E,$DA,$D1,$DA,$33,$5E,$FE,$47,$C8,$BE,$87,$D5,$18
	.dc.b	$6C,$3E,$A6,$C4,$D3,$FC,$1B,$3C,$5B,$39,$BF,$F0,$5F,$D4,$FE,$4B
	.dc.b	$03,$4F,$F1,$AD,$7F,$85,$9C,$70,$78,$9D,$71,$58,$F9,$D7,$FE,$7D
	.dc.b	$FA,$9F,$3F,$3B,$E4,$33,$DC,$9C,$C6,$1B,$97,$F8,$9A,$74,$FF,$84
	.dc.b	$A7,$4D,$35,$3C,$2F,$FC,$4B,$E3,$BE,$A7,$F0,$1E,$17,$FA,$3F,$E7
	.dc.b	$FE,$4C,$E7,$0F,$F1,$B4,$E5,$FF,$19,$C6,$6F,$FC,$97,$39,$7A,$DF
	.dc.b	$4C,$98,$7D,$4F,$3B,$FD,$CB,$8F,$F8,$F1,$B7,$FE,$38,$7F,$DB,$6D
	.dc.b	$E9,$E0,$49,$E1,$14,$FF,$1F,$F9,$0F,$A9,$FC,$0B,$67,$E3,$7A,$3F
	.dc.b	$C3,$FC,$BF,$F4,$1D,$7C,$E8,$C4,$FF,$A4,$A3,$0F,$FB,$CA,$76,$F2
	.dc.b	$30,$6C,$C5,$FF,$A5,$27,$E5,$F9,$FF,$D1,$27,$FD,$5F,$DB,$DA,$9C
	.dc.b	$EF,$A9,$7C,$FE,$78,$98,$F1,$B6,$7F,$C9,$E7,$38,$EE,$B5,$3F,$EF
	.dc.b	$4C,$39,$E5,$E0,$3F,$D8,$5F,$53,$D7,$88,$F8,$F4,$72,$62,$B6,$4D
	.dc.b	$3C,$2F,$FA,$16,$B8,$4E,$E7,$27,$FA,$2E,$3D,$5E,$A3,$CC,$A7,$73
	.dc.b	$FD,$A1,$7F,$2A,$9B,$2B,$84,$E7,$1E,$5E,$97,$13,$E5,$5B,$C8,$FE
	.dc.b	$BF,$B7,$F4,$BA,$DD,$F1,$CB,$BD,$F9,$A7,$2F,$A0,$E5,$D9,$D6,$BC
	.dc.b	$99,$BD,$57,$C3,$FD,$56,$CF,$97,$EA,$5F,$51,$9A,$93,$CA,$B7,$8F
	.dc.b	$86,$F7,$39,$AF,$C0,$7B,$4F,$8F,$A7,$D5,$F5,$7D,$B4,$C3,$E3,$D6
	.dc.b	$E3,$CE,$D8,$FA,$79,$A3,$1B,$FC,$5B,$E7,$5C,$B0,$FC,$A6,$0F,$B9
	.dc.b	$7A,$F8,$6C,$7D,$3F,$75,$F4,$D9,$AF,$C5,$46,$27,$FD,$55,$87,$E7
	.dc.b	$A3,$0F,$FA,$B7,$89,$18,$8F,$61,$BB,$FF,$02,$38,$9D,$C6,$23,$E1
	.dc.b	$46,$2B,$BB,$63,$B5,$75,$7E,$46,$AF,$B8,$7B,$1F,$AF,$6E,$1F,$93
	.dc.b	$F5,$1F,$56,$ED,$D5,$FE,$E6,$5C,$FF,$C5,$B2,$F4,$C6,$1F,$30,$71
	.dc.b	$87,$C4,$6C,$7F,$B3,$D7,$D4,$FE,$12,$3F,$57,$18,$E8,$F0,$7C,$FB
	.dc.b	$EA,$FC,$57,$BE,$6F,$D0,$7C,$53,$E0,$BF,$78,$DC,$FD,$9F,$57,$FA
	.dc.b	$77,$F5,$91,$9D,$3B,$CE,$5F,$23,$5F,$6B,$5F,$C9,$68,$8C,$47,$BD
	.dc.b	$B1,$9E,$FF,$86,$DB,$F8,$53,$C3,$8D,$F1,$EF,$81,$FE,$44,$DA,$FE
	.dc.b	$05,$F5,$5F,$D2,$9F,$DA,$C6,$72,$FD,$E4,$E5,$FC,$73,$5F,$C5,$35
	.dc.b	$FC,$4B,$5F,$E6,$6D,$E8,$34,$9E,$87,$DC,$C7,$CF,$D7,$E6,$3F,$95
	.dc.b	$4B,$E0,$F1,$97,$A1,$FC,$1E,$1D,$1B,$3B,$9F,$CC,$B7,$1F,$4C,$E8
	.dc.b	$93,$A7,$C5,$F2,$9F,$E0,$78,$FE,$2E,$D9,$FE,$3D,$F2,$AF,$8F,$FE
	.dc.b	$99,$39,$E4,$FE,$4D,$39,$BF,$D4,$B6,$FF,$6C,$D7,$F6,$4D,$7F,$62
	.dc.b	$D7,$FA,$DB,$72,$34,$D6,$8C,$1F,$0B,$E4,$3F,$A9,$4B,$E6,$71,$8F
	.dc.b	$F4,$18,$C3,$E6,$8F,$97,$18,$C7,$CF,$18,$24,$EC,$91,$FC,$38,$FE
	.dc.b	$9F,$F3,$1C,$73,$FE,$D8,$D9,$FE,$1B,$EA,$7F,$26,$CD,$39,$BF,$EF
	.dc.b	$33,$FE,$AE,$7F,$8B,$35,$E3,$DB,$F8,$08,$C4,$70,$8C,$12,$FF,$8C
	.dc.b	$CE,$AF,$BE,$DF,$E0,$3C,$3E,$A6,$3F,$D7,$7E,$65,$F2,$99,$AF,$EF
	.dc.b	$51,$8C,$7F,$91,$B7,$3F,$E6,$0D,$7F,$96,$35,$FE,$4B,$F4,$3F,$C8
	.dc.b	$9F,$E9,$2D,$C3,$FD,$6D,$F3,$8D,$1F,$4F,$FE,$17,$AD,$63,$CB,$C3
	.dc.b	$C1,$FE,$1F,$BE,$15,$89,$FE,$DB,$83,$E4,$3F,$C4,$0D,$8F,$F0,$07
	.dc.b	$CD,$BE,$9F,$35,$27,$8B,$FE,$19,$9D,$46,$0F,$F0,$66,$0F,$17,$FA
	.dc.b	$8E,$1E,$A6,$CF,$F8,$F7,$E9,$FF,$D1,$2F,$08,$CE,$CB,$E5,$E6,$7A
	.dc.b	$E7,$38,$30,$E3,$7F,$E2,$2B,$DE,$1F,$E1,$DA,$D7,$C4,$7F,$44,$6C
	.dc.b	$7F,$66,$6D,$7E,$55,$AF,$F6,$4D,$DF,$65,$FE,$95,$8F,$F1,$7C,$63
	.dc.b	$E2,$BE,$77,$D3,$FF,$09,$BF,$F7,$64,$E7,$0F,$F1,$73,$6F,$8E,$1C
	.dc.b	$EB,$EE,$DF,$1B,$FF,$07,$5E,$E7,$FA,$DF,$C8,$7F,$7C,$C3,$FB,$BB
	.dc.b	$5F,$F0,$A3,$53,$BD,$7F,$7F,$F4,$F7,$B8,$D2,$57,$F8,$22,$7F,$D4
	.dc.b	$57,$C7,$E2,$7F,$CC,$33,$96,$BF,$BB,$9D,$BF,$BD,$C5,$F4,$FF,$DC
	.dc.b	$E3,$E3,$B6,$FE,$2B,$73,$FF,$08,$F9,$1F,$E4,$BA,$F8,$F3,$95,$7D
	.dc.b	$3B,$FF,$C2,$F8,$72,$6E,$1C,$BC,$BB,$7F,$E3,$7F,$A7,$FF,$24,$BF
	.dc.b	$25,$AF,$F8,$99,$AF,$F1,$56,$BF,$E9,$1F,$A7,$F2,$3F,$E6,$EC,$11
	.dc.b	$87,$FD,$38,$DB,$FF,$15,$D7,$69,$CD,$F9,$1E,$2D,$AE,$17,$E7,$F6
	.dc.b	$FD,$3E,$1F,$F9,$96,$FF,$D2,$07,$F8,$D0,$B8,$FB,$BC,$77,$D3,$BE
	.dc.b	$8F,$8D,$6E,$BF,$C6,$2D,$C7,$FC,$FE,$DB,$E3,$5E,$0F,$93,$7D,$3D
	.dc.b	$25,$FE,$F3,$E9,$FF,$CA,$AF,$F9,$DA,$76,$D9,$DB,$FB,$39,$DB,$FE
	.dc.b	$81,$7F,$A7,$3F,$E4,$86,$BF,$E5,$56,$BF,$E5,$16,$BF,$E4,$AF,$27
	.dc.b	$FC,$EC,$FF,$A1,$9A,$FF,$9C,$1B,$8F,$FA,$4E,$7A,$BF,$CE,$7E,$0F
	.dc.b	$F5,$59,$3C,$8F,$F5,$36,$1F,$CE,$27,$31,$88,$FE,$73,$18,$23,$B3
	.dc.b	$FC,$DF,$B7,$F5,$6F,$A7,$FB,$39,$1A,$7D,$7B,$E9,$8F,$F8,$0B,$E6
	.dc.b	$3C,$7E,$71,$A4,$C7,$FC,$0A,$D7,$FC,$0A,$D7,$FC,$F3,$8F,$F7,$DE
	.dc.b	$2F,$C4,$F9,$FF,$91,$E2,$FF,$38,$B6,$BF,$C0,$F7,$F8,$9C,$3F,$3F
	.dc.b	$1F,$E0,$AB,$F1,$91,$FF,$0A,$B5,$FF,$0A,$B5,$F0,$D7,$A5,$FF,$11
	.dc.b	$D7,$88,$ED,$3F,$97,$FA,$67,$91,$F4,$9B,$5F,$E8,$A6,$BF,$E9,$76
	.dc.b	$BF,$E9,$36,$BF,$E9,$26,$BF,$EF,$C6,$F0,$1C,$E7,$FE,$5E,$FA,$7F
	.dc.b	$D6,$FA,$C6,$9E,$5C,$3C,$DC,$DE,$52,$63,$E5,$6B,$E5,$6B,$E5,$7F
	.dc.b	$9B,$5F,$F9,$B3,$CB,$EE,$49,$FF,$76,$63,$FB,$77,$FD,$41,$F4,$BF
	.dc.b	$E0,$C7,$FD,$2C,$D7,$FD,$2C,$D7,$FD,$3E,$D7,$FD,$06,$D7,$FD,$40
	.dc.b	$DF,$64,$D3,$5A,$3E,$9F,$F6,$BC,$7F,$EF,$96,$F1,$FF,$AF,$9B,$87
	.dc.b	$4F,$53,$F9,$6E,$2F,$6F,$FC,$BD,$3F,$EB,$96,$BF,$EA,$A6,$E0,$C3
	.dc.b	$FD,$33,$E0,$BE,$9F,$9E,$9F,$D6,$F1,$F3,$E3,$FD,$7D,$AF,$C4,$3E
	.dc.b	$6F,$CF,$7F,$10,$7F,$DD,$ED,$C7,$B4,$38,$13,$9C,$BF,$D5,$3F,$4E
	.dc.b	$F8,$2F,$E8,$FF,$2E,$FA,$7F,$F4,$56,$3F,$EE,$A8,$C7,$BB,$FD,$C5
	.dc.b	$B1,$99,$DB,$FE,$5D,$8C,$B6,$32,$FF,$BB,$A3,$2D,$8C,$A3,$E3,$7D
	.dc.b	$3F,$EA,$D6,$9E,$CC,$93,$98,$C3,$63,$95,$F8,$A7,$E1,$E8,$F8,$DF
	.dc.b	$F4,$0E,$1D,$E7,$31,$88,$EF,$7F,$EF,$88,$C4,$FC,$CE,$57,$C5,$3E
	.dc.b	$9F,$FA,$BD,$6A,$F8,$7A,$78,$3F,$AB,$56,$F1,$F0,$4F,$BF,$27,$C8
	.dc.b	$3D,$FF,$EC,$2D,$AD,$9F,$6D,$C3,$E6,$19,$BE,$3D,$F4,$FC,$AE,$57
	.dc.b	$2C,$3B,$C7,$F6,$9A,$DC,$F6,$B7,$CE,$7F,$9B,$E3,$0D,$7C,$11,$AF
	.dc.b	$83,$FF,$07,$C3,$E9,$BF,$07,$C3,$F1,$EF,$83,$6B,$FB,$C4,$63,$5F
	.dc.b	$C0,$5F,$1E,$5F,$F0,$0F,$CA,$E6,$3F,$34,$DD,$4E,$73,$F1,$4F,$B7
	.dc.b	$E3,$FF,$04,$FE,$85,$B4,$73,$2F,$3F,$FD,$05,$5C,$7C,$AF,$F0,$8D
	.dc.b	$70,$37,$3A,$9F,$A1,$6F,$99,$7C,$B7,$F8,$61,$B3,$F3,$DA,$FC,$8F
	.dc.b	$98,$7C,$86,$EA,$4E,$DF,$F0,$E4,$EA,$3F,$02,$3E,$91,$F4,$FF,$E2
	.dc.b	$1F,$9F,$FB,$33,$FE,$60,$D7,$F5,$5B,$27,$3F,$09,$FA,$BC,$38,$DF
	.dc.b	$46,$BF,$72,$77,$7C,$9B,$1F,$B5,$9D,$F2,$FE,$8E,$73,$87,$7E,$9F
	.dc.b	$F3,$AE,$1C,$5B,$87,$1C,$BF,$21,$F2,$38,$BF,$CC,$37,$A7,$FC,$4F
	.dc.b	$A7,$F6,$B9,$FC,$E7,$C1,$66,$A6,$1F,$E8,$0C,$BD,$F8,$70,$6D,$75
	.dc.b	$E3,$7D,$1F,$8B,$93,$F0,$5C,$F7,$9B,$F8,$CE,$BD,$EF,$79,$F5,$3F
	.dc.b	$5A,$E1,$F9,$C7,$E3,$6B,$91,$DF,$0F,$7C,$7D,$7C,$7D,$DF,$47,$FB
	.dc.b	$8C,$3D,$8D,$CF,$F4,$3C,$6A,$E2,$DB,$F1,$6E,$BE,$11,$FC,$01,$AF
	.dc.b	$73,$D0,$53,$CF,$1C,$7B,$7C,$67,$6B,$FA,$5D,$AF,$0E,$92,$FC,$31
	.dc.b	$FB,$67,$9B,$B7,$31,$B7,$27,$C0,$F0,$63,$EA,$C7,$F9,$76,$9F,$0B
	.dc.b	$E9,$F3,$3E,$69,$CE,$BF,$9A,$6C,$7B,$E7,$E2,$EC,$F2,$BD,$FB,$FD
	.dc.b	$E6,$C6,$6B,$EB,$FB,$EF,$A7,$FF,$4D,$E7,$FA,$3A,$D3,$6F,$89,$F8
	.dc.b	$1A,$E1,$F0,$7E,$08,$CE,$1F,$82,$3F,$69,$7C,$B4,$FE,$B5,$7C,$27
	.dc.b	$78,$75,$78,$65,$E7,$7C,$B9,$79,$2F,$F0,$78,$73,$9F,$D8,$35,$EF
	.dc.b	$8F,$96,$B8,$46,$A3,$BD,$72,$75,$97,$A6,$7C,$37,$CF,$0F,$16,$3F
	.dc.b	$E5,$9C,$BD,$15,$AC,$39,$BD,$4F,$3E,$1C,$EB,$D8,$4C,$7E,$0E,$5E
	.dc.b	$4A,$F3,$9C,$C9,$3D,$E3,$58,$F8,$AB,$D1,$5A,$79,$DF,$77,$37,$DE
	.dc.b	$58,$FA,$23,$11,$E0,$75,$19,$9E,$24,$9F,$64,$F8,$EB,$B4,$62,$75
	.dc.b	$5C,$63,$04,$3B,$47,$99,$F8,$E6,$0D,$C7,$17,$04,$BF,$80,$FA,$B0
	.dc.b	$F8,$79,$7D,$52,$65,$F6,$A7,$AD,$70,$7B,$5F,$58,$C1,$2B,$F1,$51
	.dc.b	$97,$9C,$FC,$D3,$C1,$8F,$C0,$7C,$91,$C3,$0F,$1E,$1E,$08,$C5,$7E
	.dc.b	$2E,$BF,$82,$61,$F9,$B6,$CF,$EC,$5B,$5F,$C2,$5F,$E9,$5F,$43,$FC
	.dc.b	$25,$E2,$D7,$F9,$1B,$5D,$EF,$C5,$FE,$C5,$1F,$C4,$74,$FE,$B8,$DD
	.dc.b	$3F,$63,$E7,$3D,$1F,$F3,$1D,$6B,$0F,$E0,$F7,$BC,$3A,$9F,$C1,$AF
	.dc.b	$87,$8B,$AA,$93,$FB,$8A,$E1,$BF,$F9,$B3,$1E,$B9,$7A,$2F,$79,$77
	.dc.b	$C3,$A5,$69,$EF,$19,$25,$7A,$27,$53,$EE,$AE,$73,$D6,$39,$E3,$87
	.dc.b	$9C,$66,$FF,$75,$97,$AA,$7D,$17,$B7,$A1,$C8,$C9,$27,$C6,$FE,$B2
	.dc.b	$79,$46,$23,$8B,$AC,$7E,$16,$1D,$75,$67,$F3,$6F,$79,$7B,$6B,$53
	.dc.b	$9A,$DE,$5F,$CB,$F2,$EF,$7E,$B2,$61,$EE,$3C,$99,$FC,$6B,$E3,$1E
	.dc.b	$99,$DE,$5E,$48,$C1,$C4,$F1,$9C,$B2,$F1,$E3,$F2,$E7,$86,$7E,$FC
	.dc.b	$3A,$BE,$FC,$BC,$B8,$73,$9C,$E3,$E0,$9C,$CF,$62,$4E,$9F,$01,$E4
	.dc.b	$33,$9F,$E3,$27,$8C,$EE,$73,$8F,$C5,$CF,$E0,$5F,$C9,$24,$67,$3F
	.dc.b	$5D,$7C,$0C,$FD,$79,$7C,$A8,$E6,$7E,$07,$3F,$BF,$CC,$F8,$E7,$C7
	.dc.b	$7E,$5E,$30,$D8,$7D,$3D,$89,$97,$F8,$5A,$78,$B6,$73,$87,$F8,$4B
	.dc.b	$E9,$FE,$4E,$06,$9F,$E3,$6A,$FF,$0F,$B8,$DB,$C4,$EB,$8A,$C7,$E9
	.dc.b	$BB,$5F,$4F,$E7,$E8,$7C,$76,$63,$0D,$C7,$FC,$24,$DC,$FF,$C2,$0D
	.dc.b	$DF,$FC,$6A,$DE,$0F,$F0,$93,$F0,$FC,$77,$D3,$FE,$EF,$C2,$FF,$46
	.dc.b	$7C,$DE,$63,$1A,$A3,$19,$FF,$92,$EB,$6F,$F8,$F2,$F9,$B9,$AE,$0F
	.dc.b	$F9,$03,$CE,$FF,$1A,$37,$9F,$EA,$47,$FB,$9E,$30,$4D,$7F,$C7,$F1
	.dc.b	$C1,$A7,$09,$FB,$9E,$4B,$E9,$E9,$F2,$FD,$06,$1F,$EC,$39,$D3,$FE
	.dc.b	$4C,$C5,$3A,$C3,$FD,$11,$7C,$6B,$83,$63,$31,$C6,$78,$79,$9F,$E8
	.dc.b	$9A,$F9,$5E,$7F,$F0,$8A,$FF,$54,$76,$E9,$FF,$4D,$FD,$33,$E9,$F1
	.dc.b	$93,$83,$FD,$65,$78,$6D,$F2,$3F,$DB,$97,$C5,$C5,$7F,$AE,$B4,$7D
	.dc.b	$3F,$16,$CF,$FB,$8F,$86,$D6,$EB,$44,$9D,$36,$3F,$D8,$BE,$9F,$FB
	.dc.b	$36,$7F,$DE,$0D,$8F,$F7,$4F,$9B,$FE,$C8,$9C,$BE,$CE,$EF,$F6,$A7
	.dc.b	$A5,$63,$55,$FE,$FF,$1E,$3F,$F3,$DE,$2F,$A5,$7C,$30,$FF,$7D,$5F
	.dc.b	$09,$CB,$63,$FD,$DF,$5C,$5F,$F6,$3C,$66,$FE,$67,$A8,$F8,$7F,$98
	.dc.b	$DA,$4B,$0F,$0F,$D3,$BD,$CD,$E1,$2F,$75,$F8,$7E,$26,$3C,$F1,$E5
	.dc.b	$5F,$ED,$8C,$79,$3C,$0E,$F5,$C3,$0F,$BD,$EC,$BE,$9E,$9F,$27,$D5
	.dc.b	$F6,$63,$F8,$59,$D6,$1C,$E3,$15,$C2,$31,$BA,$F8,$E2,$C3,$8E,$1E
	.dc.b	$63,$F4,$BF,$4E,$F5,$FF,$27,$8A,$30,$F4,$F6,$DF,$3B,$A9,$CD,$F3
	.dc.b	$CD,$5A,$BF,$F1,$34,$62,$FF,$5D,$5F,$40,$46,$67,$75,$C6,$7F,$D6
	.dc.b	$D3,$96,$C6,$5E,$51,$96,$C6,$5F,$9F,$F4,$EF,$67,$F5,$8D,$BF,$95
	.dc.b	$EA,$3E,$9D,$D9,$AB,$FD,$AC,$FF,$8C,$70,$F4,$9F,$88,$71,$7F,$E4
	.dc.b	$36,$C7,$FB,$35,$AF,$BD,$7D,$37,$E1,$23,$E9,$46,$39,$FB,$FC,$DB
	.dc.b	$E9,$FC,$37,$BE,$9F,$D2,$FC,$43,$E0,$BF,$5C,$DC,$FD,$7F,$4F,$FA
	.dc.b	$C7,$F0,$71,$9D,$3B,$CE,$5F,$33,$5F,$33,$5F,$73,$5F,$F6,$B3,$74
	.dc.b	$34,$9D,$DF,$1A,$7F,$8E,$7C,$8B,$E0,$7F,$8E,$37,$1F,$CB,$FD,$3F
	.dc.b	$F4,$6A,$FC,$A7,$5A,$B5,$1F,$8D,$9D,$3F,$80,$6B,$FB,$AE,$74,$62
	.dc.b	$3F,$1A,$D8,$CF,$9F,$E5,$6E,$1F,$96,$9D,$1C,$70,$F1,$75,$FA,$0F
	.dc.b	$E3,$90,$F8,$3C,$57,$A9,$E9,$19,$8E,$91,$82,$38,$26,$3F,$9E,$9E
	.dc.b	$0F,$E9,$4F,$E8,$9F,$27,$FE,$02,$9F,$E0,$18,$23,$13,$FB,$1F,$94
	.dc.b	$7D,$37,$F4,$99,$CF,$27,$F2,$C9,$CE,$1F,$B4,$6D,$FE,$B9,$AF,$EB
	.dc.b	$1A,$FC,$D6,$BF,$B8,$6E,$86,$9A,$C7,$CD,$F0,$BE,$33,$FA,$24,$3E
	.dc.b	$9F,$18,$7F,$18,$8C,$3E,$58,$FA,$D1,$8C,$3C,$D1,$82,$4E,$C9,$1F
	.dc.b	$26,$DF,$33,$C7,$2F,$E7,$0D,$AF,$C4,$7A,$7F,$C9,$A3,$17,$FC,$52
	.dc.b	$3F,$D2,$0F,$F0,$E6,$BC,$3D,$DF,$C5,$46,$23,$84,$60,$95,$FC,$8A
	.dc.b	$75,$87,$93,$6F,$EF,$D8,$7D,$2C,$7F,$B0,$FC,$C3,$E4,$B3,$5F,$DA
	.dc.b	$E3,$18,$FF,$2C,$6E,$7F,$C9,$9A,$FF,$24,$6B,$FD,$43,$E7,$FF,$90
	.dc.b	$3F,$AE,$6E,$1F,$E2,$1F,$A4,$FF,$11,$FD,$3F,$F8,$46,$73,$87,$1F
	.dc.b	$0F,$07,$F8,$62,$B7,$1F,$94,$9D,$E1,$FE,$1B,$C1,$F3,$1F,$E1,$B6
	.dc.b	$D7,$F5,$6F,$9A,$7D,$3E,$6A,$4F,$E7,$AB,$77,$FD,$36,$F7,$39,$AF
	.dc.b	$17,$8B,$FD,$1E,$FD,$4D,$9F,$EE,$FF,$4F,$FE,$83,$7F,$B4,$66,$9D
	.dc.b	$62,$8C,$79,$9D,$0F,$F1,$AE,$0C,$39,$57,$F6,$9A,$DE,$1F,$DC,$EB
	.dc.b	$5F,$15,$FC,$85,$B1,$FB,$BA,$FF,$07,$5F,$27,$FB,$14,$E9,$FE,$01
	.dc.b	$AB,$E9,$FF,$D2,$31,$FD,$DE,$31,$F1,$9F,$47,$E9,$FF,$83,$DF,$F9
	.dc.b	$66,$73,$8F,$F8,$15,$B7,$C7,$0E,$75,$F7,$EF,$8D,$FF,$88,$AF,$73
	.dc.b	$FC,$E7,$E2,$3D,$18,$7F,$77,$6B,$FE,$28,$6A,$77,$B3,$B5,$F4,$F8
	.dc.b	$6D,$7C,$39,$D4,$E4,$E3,$F1,$1F,$E3,$39,$D3,$5F,$E4,$93,$FE,$A4
	.dc.b	$C3,$67,$9B,$E9,$BF,$C7,$D3,$7D,$9F,$3C,$62,$3E,$93,$5F,$E5,$CD
	.dc.b	$7F,$C2,$8D,$7F,$C8,$AD,$4D,$FC,$0D,$EE,$FC,$46,$CF,$C8,$6B,$F8
	.dc.b	$16,$BF,$CF,$7A,$1F,$57,$F3,$1B,$1F,$41,$AF,$E2,$5A,$F1,$8E,$04
	.dc.b	$78,$3F,$63,$C5,$7D,$3F,$FA,$96,$7F,$C6,$6D,$7F,$C5,$AD,$7F,$C5
	.dc.b	$6D,$7F,$B6,$35,$FF,$1B,$B6,$3A,$FC,$47,$78,$D7,$4B,$E9,$BF,$C0
	.dc.b	$17,$FA,$18,$C7,$9F,$F8,$C6,$C6,$67,$6F,$F6,$18,$CB,$63,$2F,$F9
	.dc.b	$26,$32,$D8,$CB,$FE,$7C,$8F,$F5,$EF,$D5,$77,$F1,$CD,$24,$FF,$2B
	.dc.b	$DB,$E4,$78,$7F,$5A,$3F,$C1,$8D,$7F,$CA,$ED,$BF,$F1,$24,$7F,$09
	.dc.b	$F4,$7F,$C4,$0D,$EA,$F7,$46,$27,$64,$70,$70,$D9,$7F,$ED,$4B,$FE
	.dc.b	$7D,$97,$90,$8F,$CB,$FB,$07,$D8,$7C,$B6,$C6,$63,$04,$70,$43,$E7
	.dc.b	$47,$07,$FC,$05,$CB,$F8,$9F,$37,$F4,$58,$FF,$7D,$8C,$1F,$67,$F0
	.dc.b	$07,$C7,$D2,$5F,$DF,$8C,$B6,$32,$FD,$D8,$CB,$63,$F3,$1F,$45,$FD
	.dc.b	$12,$3F,$D1,$2D,$7F,$D0,$6D,$7F,$D0,$2D,$7F,$CE,$9F,$63,$57,$B2
	.dc.b	$8C,$37,$A7,$F8,$EB,$67,$FC,$F5,$3A,$24,$E9,$E9,$93,$EA,$1F,$53
	.dc.b	$FA,$F6,$DF,$FA,$81,$AF,$FA,$69,$AF,$FA,$0F,$E0,$5F,$77,$E2,$6C
	.dc.b	$7F,$A2,$DA,$FF,$A2,$DA,$F1,$8E,$04,$78,$3F,$C0,$7D,$87,$D2,$D2
	.dc.b	$4F,$F9,$7D,$F8,$3E,$FF,$F9,$22,$3F,$D3,$4D,$7F,$D3,$4D,$7F,$D3
	.dc.b	$4D,$7F,$D4,$4D,$7E,$7B,$75,$34,$C7,$D4,$7F,$4A,$83,$49,$3F,$53
	.dc.b	$D9,$FF,$11,$3F,$EA,$D6,$BF,$EA,$D6,$BF,$EA,$D6,$BF,$EB,$5F,$19
	.dc.b	$EA,$FD,$E6,$C7,$FA,$DD,$AF,$FA,$DD,$AF,$FA,$DF,$EA,$3F,$D1,$6F
	.dc.b	$FA,$F5,$AF,$FA,$F5,$AE,$F3,$FE,$CB,$F4,$2F,$A6,$A4,$9F,$ED,$71
	.dc.b	$96,$C6,$5F,$F0,$34,$65,$B1,$9F,$29,$E0,$FF,$A7,$FE,$65,$F4,$FF
	.dc.b	$DD,$7E,$77,$0D,$C0,$31,$E9,$87,$28,$FF,$78,$FC,$6F,$B3,$1F,$F7
	.dc.b	$4E,$7F,$EF,$56,$CF,$26,$BF,$EF,$AE,$8F,$F7,$FC,$EE,$3F,$D6,$55
	.dc.b	$CD,$B1,$97,$BB,$65,$F2,$0F,$85,$F4,$46,$1B,$19,$AF,$F0,$86,$9F
	.dc.b	$EB,$F8,$FE,$8F,$EA,$23,$15,$C0,$95,$C0,$5F,$83,$CC,$F8,$DC,$CF
	.dc.b	$D8,$BF,$91,$6B,$FE,$F7,$6D,$FF,$BF,$DA,$FF,$0C,$6C,$F1,$F1,$BD
	.dc.b	$CD,$9F,$90,$71,$8C,$CF,$03,$F9,$86,$7E,$99,$F2,$FC,$EF,$E5,$23
	.dc.b	$19,$23,$1A,$73,$AF,$64,$FF,$84,$BE,$87,$E6,$79,$A6,$92,$7C,$67
	.dc.b	$C2,$8D,$38,$9E,$4D,$8C,$BF,$7D,$BA,$7E,$32,$3E,$5F,$CD,$3E,$9F
	.dc.b	$EC,$B6,$3C,$AD,$7F,$C8,$D8,$FF,$BD,$E3,$FC,$8D,$1A,$9E,$24,$78
	.dc.b	$FB,$3F,$9B,$7F,$35,$C1,$F8,$B6,$9E,$C2,$9F,$9D,$3F,$43,$CD,$FE
	.dc.b	$0C,$C7,$FD,$D8,$D8,$F5,$35,$DC,$F5,$8C,$47,$4D,$BE,$9F,$D1,$3E
	.dc.b	$13,$F2,$ED,$9F,$C7,$1B,$D3,$F6,$DF,$87,$FE,$73,$AC,$D3,$1E,$3D
	.dc.b	$AF,$B7,$3A,$6C,$E9,$FB,$FF,$53,$FD,$A3,$2F,$C2,$B7,$0F,$DF,$F1
	.dc.b	$23,$15,$FA,$07,$EC,$3F,$48,$C7,$99,$FB,$08,$DC,$F1,$8C,$4F,$E9
	.dc.b	$BE,$C5,$EA,$7E,$79,$B1,$F8,$CE,$A4,$62,$3E,$94,$62,$78,$12,$38
	.dc.b	$4E,$F8,$BF,$19,$5B,$9C,$D7,$F0,$4F,$D4,$3D,$3A,$7E,$91,$BA,$7F
	.dc.b	$1E,$BE,$17,$ED,$CD,$C1,$F8,$E8,$C3,$70,$FB,$7C,$9F,$6B,$97,$ED
	.dc.b	$B2,$BF,$5B,$5B,$C7,$E6,$BF,$9B,$BF,$DA,$FC,$2F,$ED,$2B,$F8,$6B
	.dc.b	$FE,$39,$7C,$4D,$8D,$F8,$DF,$C7,$A3,$86,$47,$58,$7D,$A6,$DF,$4A
	.dc.b	$FD,$17,$37,$EF,$74,$FC,$7B,$67,$EE,$1F,$CF,$FA,$BF,$C4,$2F,$DF
	.dc.b	$FA,$97,$D5,$FD,$5D,$3F,$88,$4E,$F0,$E7,$7C,$27,$38,$75,$7F,$8B
	.dc.b	$61,$CB,$87,$C7,$5B,$D3,$F5,$6D,$C7,$F6,$2D,$AE,$11,$DB,$18,$FA
	.dc.b	$31,$ED,$7F,$D6,$2B,$73,$9B,$E8,$F9,$CC,$5F,$F3,$D9,$DE,$3E,$8A
	.dc.b	$DE,$1D,$2F,$C7,$7B,$CB,$E1,$63,$E8,$8F,$1E,$1C,$5D,$65,$EE,$73
	.dc.b	$3B,$BF,$99,$7C,$9E,$A7,$90,$97,$F2,$23,$51,$E9,$8E,$CE,$A7,$E0
	.dc.b	$4E,$B0,$F4,$4E,$E7,$C1,$5C,$F0,$F2,$4F,$13,$C3,$19,$7C,$E6,$27
	.dc.b	$89,$DA,$78,$CF,$88,$91,$99,$E2,$78,$4E,$43,$2F,$9F,$1D,$23,$26
	.dc.b	$E3,$2E,$2F,$D6,$7D,$38,$CC,$75,$24,$67,$0F,$85,$97,$DA,$26,$5F
	.dc.b	$52,$BC,$B5,$D2,$7F,$1D,$5C,$2B,$31,$8C,$BD,$F8,$F9,$F0,$F3,$5E
	.dc.b	$DF,$94,$62,$76,$FA,$6B,$93,$F9,$83,$FA,$5B,$FA,$E6,$DF,$F8,$FA
	.dc.b	$31,$1F,$D3,$BE,$93,$FD,$3E,$DE,$5F,$F2,$1C,$62,$3F,$AB,$B5,$FF
	.dc.b	$1C,$3F,$EA,$AD,$3F,$66,$DC,$BF,$4D,$F5,$6F,$2F,$FC,$79,$5A,$C3
	.dc.b	$F8,$2D,$EF,$0E,$B1,$FA,$7B,$E1,$CD,$A8,$93,$FB,$3A,$E1,$BA,$F7
	.dc.b	$39,$AD,$E1,$DE,$3F,$87,$4E,$A3,$04,$CB,$E0,$4E,$B2,$F3,$C6,$27
	.dc.b	$53,$E6,$35,$97,$69,$D1,$23,$F2,$57,$DE,$31,$1F,$23,$0E,$0E,$B0
	.dc.b	$EE,$F5,$C3,$BB,$B3,$B9,$ED,$23,$E4,$C7,$E1,$4F,$29,$E1,$3C,$B7
	.dc.b	$67,$F2,$32,$F2,$57,$7C,$BC,$B5,$AC,$BE,$96,$3C,$E3,$15,$E0,$AF
	.dc.b	$76,$5E,$0C,$FD,$73,$E0,$9C,$E3,$DA,$30,$77,$25,$71,$7A,$19,$36
	.dc.b	$E7,$3F,$56,$7F,$87,$9C,$D7,$5C,$7C,$B5,$AC,$FD,$39,$79,$27,$39
	.dc.b	$79,$33,$F6,$65,$CE,$73,$97,$93,$2E,$93,$92,$63,$D2,$7C,$E7,$B0
	.dc.b	$F1,$DF,$48,$F0,$4F,$CF,$8C,$4F,$0B,$EF,$9F,$B7,$1F,$C4,$71,$3E
	.dc.b	$E7,$FD,$E3,$A7,$F9,$AB,$F3,$37,$07,$E1,$AB,$5E,$8D,$9C,$D7,$F9
	.dc.b	$C2,$F8,$E7,$FE,$0F,$9C,$BF,$E1,$18,$CF,$07,$F9,$EA,$72,$FF,$82
	.dc.b	$3E,$2B,$B7,$31,$BF,$5B,$86,$63,$FE,$7E,$9D,$36,$73,$8F,$F8,$EF
	.dc.b	$47,$E7,$AF,$9E,$F5,$76,$BD,$6F,$D0,$D6,$A7,$31,$8A,$E1,$8F,$F8
	.dc.b	$ED,$B8,$7F,$8E,$9B,$1F,$E3,$66,$C7,$F8,$C5,$AF,$CA,$6D,$F7,$9E
	.dc.b	$97,$FE,$D1,$7F,$D1,$5C,$0F,$AE,$FD,$3B,$79,$7E,$94,$7F,$A6,$E7
	.dc.b	$FD,$3E,$E8,$91,$AE,$9D,$BF,$67,$AB,$D3,$C3,$FE,$AA,$6F,$2F,$F9
	.dc.b	$99,$7A,$FB,$CE,$65,$FE,$AC,$9D,$ED,$FE,$C8,$77,$19,$9F,$E0,$5B
	.dc.b	$BE,$BB,$F1,$1D,$EF,$8B,$CC,$61,$B7,$FE,$97,$AE,$0D,$9F,$89,$8F
	.dc.b	$FA,$B6,$72,$FF,$A6,$CF,$B5,$1A,$7F,$DF,$B1,$A9,$FF,$3E,$B6,$39
	.dc.b	$4E,$F8,$1E,$33,$E7,$3F,$3C,$DC,$3F,$D2,$26,$22,$3F,$E9,$47,$6E
	.dc.b	$2F,$FC,$DD,$E5,$BE,$BB,$B3,$67,$FD,$98,$D9,$FF,$5A,$60,$8C,$4F
	.dc.b	$E0,$9B,$3F,$EC,$5A,$E6,$49,$FF,$47,$7D,$AF,$FA,$A9,$FF,$70,$F9
	.dc.b	$DF,$CC,$1F,$F7,$F7,$3A,$75,$7F,$E8,$3A,$D3,$FA,$78,$CD,$7F,$A4
	.dc.b	$1B,$1F,$E8,$B6,$BF,$C1,$1B,$5D,$7C,$BF,$A4,$DA,$F2,$C6,$27,$8E
	.dc.b	$5F,$EF,$9F,$13,$F8,$67,$9E,$6E,$38,$66,$1F,$EF,$C9,$D3,$67,$27
	.dc.b	$FA,$77,$D3,$7D,$9F,$DD,$6C,$FF,$BE,$DB,$3F,$41,$B1,$FB,$06,$C7
	.dc.b	$FB,$21,$AB,$04,$61,$B5,$F9,$19,$FE,$19,$F0,$CB,$43,$D8,$C3,$CF
	.dc.b	$89,$F1,$9F,$A8,$BC,$D6,$B1,$FC,$AC,$E7,$0F,$F6,$BE,$1F,$9E,$BF
	.dc.b	$D1,$36,$32,$F8,$9B,$CA,$9D,$CE,$4E,$71,$99,$DC,$F4,$E0,$7D,$23
	.dc.b	$F1,$4C,$36,$BC,$73,$A9,$C9,$C7,$DE,$7C,$D5,$25,$70,$8C,$BC,$BE
	.dc.b	$03,$F9,$DE,$7F,$EF,$4C,$11,$88,$FF,$71,$B5,$FF,$6F,$35,$FE,$AF
	.dc.b	$E1,$7D,$0A,$F2,$B6,$3A,$35,$FF,$76,$B5,$FF,$74,$7D,$27,$F3,$E7
	.dc.b	$FD,$E2,$D7,$3D,$FF,$16,$31,$1F,$D9,$9B,$81,$A4,$9F,$AF,$C0,$FC
	.dc.b	$8D,$F7,$D8,$6E,$5F,$7A,$75,$39,$3E,$A7,$E2,$AE,$CA,$4A,$FA,$71
	.dc.b	$96,$C6,$73,$F9,$93,$96,$C7,$93,$E0,$5F,$4B,$49,$7E,$38,$CB,$F8
	.dc.b	$0F,$BC,$7C,$B7,$E1,$DB,$19,$8C,$11,$C1,$30,$FB,$D3,$C1,$FD,$39
	.dc.b	$FA,$8C,$DF,$1A,$FB,$4F,$92,$DB,$F8,$CD,$AE,$27,$ED,$DC,$3D,$38
	.dc.b	$5F,$28,$F8,$2F,$CF,$36,$FF,$36,$DA,$FC,$93,$5E,$4D,$7F,$7B,$5D
	.dc.b	$2B,$99,$2B,$ED,$7C,$97,$E7,$AB,$CB,$92,$31,$1F,$E0,$06,$BF,$05
	.dc.b	$AF,$A3,$27,$BF,$FC,$03,$2F,$88,$D8,$CE,$1F,$C1,$A7,$2D,$8C,$D7
	.dc.b	$C6,$9C,$B6,$32,$FD,$28,$CB,$FE,$14,$EC,$FE,$52,$DF,$07,$F3,$71
	.dc.b	$88,$FB,$0D,$7E,$7F,$A5,$F7,$5E,$8D,$7F,$92,$D7,$F8,$7F,$83,$D9
	.dc.b	$39,$9E,$31,$82,$63,$ED,$9D,$69,$FC,$77,$A7,$C6,$49,$FD,$56,$3F
	.dc.b	$EB,$6C,$3D,$9F,$77,$D7,$23,$49,$97,$CD,$AF,$E4,$95,$F1,$DA,$FC
	.dc.b	$56,$BE,$A6,$BF,$8E,$6D,$7F,$31,$7A,$12,$FF,$7D,$F4,$4F,$89,$FE
	.dc.b	$0A,$D8,$CC,$60,$8E,$09,$7F,$8C,$9E,$0F,$E8,$BC,$87,$CB,$3D,$4F
	.dc.b	$E8,$F1,$86,$C6,$6F,$FA,$6C,$E5,$B1,$9A,$E8,$D7,$F6,$F3,$B7,$35
	.dc.b	$FD,$3B,$F3,$B7,$D7,$52,$61,$FD,$5A,$32,$FF,$5C,$F2,$5F,$44,$F7
	.dc.b	$7F,$AE,$C6,$1B,$19,$9E,$0D,$C7,$F8,$0D,$6E,$73,$5C,$08,$FF,$66
	.dc.b	$9F,$EE,$BE,$CD,$4D,$23,$FD,$C3,$C6,$7C,$3B,$E7,$FF,$B7,$C6,$1B
	.dc.b	$19,$9E,$0D,$7F,$8A,$CF,$F9,$8F,$1E,$3E,$C7,$F8,$16,$7B,$35,$DB
	.dc.b	$5F,$E4,$BF,$3E,$F9,$AF,$53,$63,$FC,$02,$D7,$FC,$0E,$D7,$F9,$2B
	.dc.b	$5E,$71,$C8,$8F,$F8,$2B,$E7,$FF,$A7,$37,$C7,$F9,$11,$88,$FF,$09
	.dc.b	$B5,$FE,$63,$F3,$5F,$D1,$1F,$F0,$9B,$5F,$F0,$CB,$5D,$F5,$7D,$38
	.dc.b	$C4,$7F,$42,$6B,$FE,$2A,$6A,$73,$DA,$F9,$FF,$F1,$5C,$FF,$69,$6B
	.dc.b	$FC,$C1,$AF,$FA,$B9,$AF,$F4,$F6,$BF,$E3,$F6,$C7,$5F,$53,$FC,$5A
	.dc.b	$D9,$FE,$5D,$F3,$EF,$93,$FF,$18,$F7,$A3,$13,$A9,$CC,$62,$3F,$99
	.dc.b	$35,$FE,$60,$D7,$F9,$2B,$5F,$EB,$2D,$9F,$F1,$E7,$CF,$F8,$3C,$03
	.dc.b	$49,$5F,$E0,$BE,$EF,$AC,$FF,$6D,$6B,$FE,$5D,$6E,$3F,$C2,$37,$F5
	.dc.b	$FC,$17,$F9,$23,$13,$49,$D3,$A2,$61,$FD,$46,$32,$F8,$72,$F9,$38
	.dc.b	$FF,$94,$E3,$04,$3F,$AA,$7C,$FF,$F2,$37,$FA,$D3,$5C,$F8,$5F,$3A
	.dc.b	$31,$1F,$D5,$DA,$FF,$4C,$6B,$FE,$00,$6E,$27,$39,$FF,$73,$E4,$FB
	.dc.b	$BE,$4F,$E9,$6F,$FD,$00,$D9,$D4,$7F,$6D,$F9,$FF,$F4,$C3,$7C,$3F
	.dc.b	$C0,$C6,$23,$FB,$73,$5F,$ED,$8D,$7F,$B0,$B5,$FF,$0A,$B7,$13,$9A
	.dc.b	$FF,$11,$F2,$FE,$4B,$CD,$FD,$15,$FF,$88,$A3,$07,$FA,$0B,$CD,$7C
	.dc.b	$FE,$A7,$FC,$D8,$D7,$FC,$48,$D7,$FC,$36,$D5,$1C,$A7,$A4,$E6,$31
	.dc.b	$3F,$E8,$28,$E1,$6F,$9F,$FF,$17,$4F,$F9,$6D,$AF,$F8,$89,$AF,$F9
	.dc.b	$02,$38,$79,$FE,$9F,$71,$F1,$F4,$95,$FE,$90,$8C,$B6,$32,$FF,$A2
	.dc.b	$63,$2D,$8F,$DC,$FA,$6F,$88,$A4,$9F,$F2,$F3,$F2,$7E,$7F,$FD,$33
	.dc.b	$1F,$E9,$E6,$BF,$E9,$56,$BF,$E4,$D6,$BF,$E4,$46,$BF,$29,$BA,$1A
	.dc.b	$63,$E7,$FF,$AA,$41,$A4,$9F,$B9,$E8,$BE,$96,$92,$3F,$D3,$EF,$E0
	.dc.b	$7C,$8F,$F2,$3C,$7F,$AB,$7D,$6C,$11,$FF,$59,$67,$F8,$7F,$78,$D2
	.dc.b	$47,$FB,$33,$E5,$1F,$41,$FD,$A5,$B1,$98,$C1,$1C,$10,$FF,$6C,$C7
	.dc.b	$06,$C7,$FA,$03,$CC,$FF,$70,$7C,$23,$E8,$3F,$D0,$7E,$8F,$F8,$66
	.dc.b	$FF,$DD,$11,$8F,$5B,$FD,$0F,$8F,$3C,$3F,$DD,$1F,$41,$EA,$BF,$F5
	.dc.b	$F7,$42,$31,$1F,$EF,$23,$FD,$F8,$E3,$C7,$FC,$93,$63,$FD,$24,$D7
	.dc.b	$7A,$FF,$A5,$A7,$6D,$8D,$B5,$53,$E1,$DF,$41,$E9,$9C,$C6,$1F,$F0
	.dc.b	$9E,$7F,$EA,$18,$C4,$7F,$85,$23,$1E,$6F,$1B,$E9,$76,$F7,$B7,$C7
	.dc.b	$9C,$E1,$D0,$EF,$7E,$1E,$77,$D0,$7F,$2D,$D3,$FD,$4C,$DA,$FF,$4E
	.dc.b	$B5,$FE,$06,$D9,$E3,$E4,$7B,$DB,$3F,$EE,$D6,$BF,$3D,$B5,$DA,$7D
	.dc.b	$5F,$15,$F9,$6A,$F7,$1F,$21,$C6,$5F,$81,$BE,$6F,$B7,$CD,$7C,$97
	.dc.b	$CF,$D3,$FD,$C8,$D8,$CE,$1F,$82,$6B,$F8,$19,$DB,$99,$E0,$47,$F1
	.dc.b	$3F,$1E,$F9,$DF,$A0,$D8,$F9,$94,$8C,$47,$AA,$31,$EC,$F0,$A9,$3C
	.dc.b	$2B,$78,$FD,$77,$FC,$9B,$1F,$5C,$73,$7F,$81,$F6,$FF,$6E,$B6,$3F
	.dc.b	$43,$C1,$F8,$F7,$6D,$8D,$40,$C9,$67,$9C,$EC,$91,$EC,$F9,$37,$D0
	.dc.b	$7E,$3D,$B1,$E5,$BF,$F0,$D6,$3F,$8F,$EA,$7C,$8F,$E5,$5B,$5F,$19
	.dc.b	$AF,$C5,$7F,$A1,$CF,$19,$E0,$49,$E1,$E3,$FE,$7F,$E8,$3F,$B6,$57
	.dc.b	$E7,$5B,$D3,$ED,$E1,$46,$2F,$F6,$2F,$C1,$7E,$BB,$98,$DF,$9F,$D4
	.dc.b	$ED,$3F,$A5,$FA,$0F,$F0,$FB,$F9,$B6,$BF,$9B,$6F,$07,$5E,$64,$62
	.dc.b	$7E,$84,$63,$D6,$FD,$34,$E6,$7F,$B6,$B6,$75,$1D,$E7,$F7,$DF,$31
	.dc.b	$E9,$8F,$C6,$37,$4F,$D2,$5F,$91,$FE,$01,$C3,$F8,$AA,$FC,$93,$63
	.dc.b	$3C,$DF,$6F,$9B,$EE,$32,$BF,$00,$E5,$FC,$EF,$C0,$FE,$CB,$83,$E7
	.dc.b	$3F,$E3,$98,$FC,$F3,$6F,$F8,$83,$B8,$FE,$42,$DD,$FF,$D7,$9E,$C7
	.dc.b	$F1,$98,$E1,$7D,$3C,$6F,$E4,$97,$FA,$0E,$AF,$F0,$8B,$FC,$10,$C7
	.dc.b	$07,$E4,$63,$39,$7F,$45,$F9,$D7,$D0,$7C,$3C,$7F,$A2,$D6,$9B,$7C
	.dc.b	$4F,$E4,$95,$C3,$8B,$C3,$D3,$FC,$32,$75,$97,$86,$30,$FF,$16,$AE
	.dc.b	$0E,$B2,$FE,$AC,$F9,$8E,$18,$F8,$54,$F0,$73,$7C,$71,$F2,$C7,$5C
	.dc.b	$BD,$4F,$EE,$71,$F0,$E3,$D9,$B5,$C2,$3E,$15,$70,$AF,$14,$6F,$1F
	.dc.b	$14,$62,$BF,$82,$63,$E3,$AD,$63,$E3,$AF,$05,$FA,$6B,$57,$DC,$93
	.dc.b	$AA,$F9,$78,$78,$71,$F2,$CE,$4F,$79,$ED,$CB,$E7,$E3,$E4,$9D,$CF
	.dc.b	$94,$F8,$F5,$CE,$74,$4B,$EF,$7C,$9F,$45,$71,$C7,$CE,$F8,$72,$F6
	.dc.b	$BE,$27,$E3,$D7,$18,$C6,$5E,$CA,$E3,$1E,$32,$65,$EC,$8C,$BE,$82
	.dc.b	$63,$F1,$F2,$F4,$65,$E4,$9E,$11,$8B,$EE,$79,$70,$FA,$71,$9C,$BC
	.dc.b	$E4,$8C,$E5,$E7,$8C,$57,$13,$E3,$C6,$67,$EA,$C7,$C3,$AE,$91,$8B
	.dc.b	$F1,$D6,$EB,$91,$B7,$E2,$3F,$D4,$31,$FD,$9B,$71,$FF,$22,$46,$23
	.dc.b	$FC,$67,$F3,$FF,$E9,$BB,$34,$9C,$DF,$E4,$B7,$FD,$53,$9F,$E6,$DB
	.dc.b	$97,$E5,$3E,$81,$E6,$BA,$3F,$A8,$56,$B2,$FD,$23,$5F,$EC,$EE,$77
	.dc.b	$5F,$2A,$E3,$94,$BF,$DE,$5E,$EB,$51,$89,$D6,$5E,$57,$F8,$3C,$EA
	.dc.b	$31,$3B,$23,$FC,$4E,$35,$97,$C4,$AE,$6D,$AF,$E1,$53,$A7,$D7,$3A
	.dc.b	$69,$AC,$BE,$13,$DA,$B9,$5F,$88,$F0,$4E,$9C,$46,$89,$3F,$C4,$AF
	.dc.b	$C0,$FC,$8B,$F0,$34,$F0,$63,$8C,$FE,$DC,$F9,$73,$F9,$B9,$77,$C7
	.dc.b	$AE,$1C,$72,$F0,$BD,$E7,$4F,$B2,$32,$EE,$7F,$01,$3A,$8C,$1B,$8F
	.dc.b	$4D,$78,$31,$F2,$67,$F8,$1C,$BC,$B7,$E8,$3E,$31,$C4,$95,$C5,$C9
	.dc.b	$D8,$EE,$E7,$1F,$2C,$F4,$7D,$71,$9C,$3D,$F3,$97,$CB,$9F,$CC,$8C
	.dc.b	$5F,$84,$EB,$19,$C3,$D0,$4C,$FF,$1B,$9F,$C1,$BE,$37,$E5,$79,$E3
	.dc.b	$E1,$8E,$91,$A3,$38,$75,$BF,$8A,$74,$CF,$F1,$9C,$6F,$A0,$FF,$76
	.dc.b	$C7,$F9,$6B,$E8,$3F,$DF,$EF,$F9,$72,$33,$8F,$F9,$6E,$72,$FF,$97
	.dc.b	$63,$3B,$7F,$A3,$E7,$2F,$FA,$46,$33,$B7,$F8,$D6,$72,$FF,$82,$BE
	.dc.b	$83,$F0,$78,$B9,$51,$85,$F4,$16,$26,$7F,$E1,$47,$51,$8C,$FF,$D1
	.dc.b	$3C,$6F,$A0,$7C,$C7,$6E,$E7,$B1,$F9,$2A,$D4,$E6,$31,$5C,$32,$FF
	.dc.b	$4D,$37,$0F,$F4,$D3,$63,$FC,$80,$D8,$FF,$1F,$B5,$5F,$69,$E7,$7F
	.dc.b	$BA,$78,$B6,$7F,$CB,$5F,$23,$FC,$0F,$53,$4F,$F4,$FD,$7F,$A4,$1C
	.dc.b	$73,$7C,$A7,$5D,$16,$3E,$6F,$4F,$A7,$93,$FD,$88,$DE,$2F,$F3,$42
	.dc.b	$F6,$37,$A4,$CB,$FD,$3B,$3B,$D3,$FD,$B6,$EE,$33,$3F,$B7,$DD,$F4
	.dc.b	$1F,$73,$C1,$7C,$66,$63,$0D,$BF,$F4,$96,$3F,$E7,$5B,$E3,$18,$8F
	.dc.b	$F4,$C4,$61,$FF,$7C,$1F,$6A,$BE,$43,$A6,$B9,$9E,$4D,$38,$CF,$38
	.dc.b	$FF,$42,$70,$3C,$77,$D0,$7E,$51,$B3,$FE,$99,$31,$11,$FF,$4E,$3B
	.dc.b	$71,$7F,$EC,$1F,$2D,$F4,$1E,$6C,$3F,$D8,$CD,$8C,$D7,$FB,$3A,$30
	.dc.b	$47,$04,$AF,$F7,$57,$D0,$7F,$AB,$A3,$FD,$A3,$E7,$7F,$6E,$7F,$DF
	.dc.b	$DC,$E9,$D5,$7F,$B4,$EB,$4F,$E9,$A3,$35,$FE,$99,$6C,$7F,$A6,$1A
	.dc.b	$FE,$69,$B5,$D7,$CC,$FA,$2D,$AF,$F8,$01,$3A,$38,$E4,$F3,$DE,$0F
	.dc.b	$F0,$06,$D7,$FB,$A9,$FC,$1F,$3B,$9F,$FD,$79,$1A,$CF,$FD,$D8,$D9
	.dc.b	$D4,$30,$ED,$87,$58,$EC,$FF,$0E,$F4,$9F,$2D,$F4,$DB,$97,$E3,$ED
	.dc.b	$18,$BF,$ED,$E7,$57,$13,$F8,$86,$C7,$FB,$45,$AE,$EB,$B1,$F9,$4E
	.dc.b	$37,$95,$FD,$EB,$D7,$FE,$F7,$8F,$E4,$7E,$33,$F5,$53,$9C,$BF,$DB
	.dc.b	$9C,$C9,$CC,$7F,$B0,$F2,$FD,$2D,$7E,$91,$B1,$97,$F8,$FC,$65,$B1
	.dc.b	$9C,$7C,$6D,$9D,$46,$27,$FD,$A1,$C2,$FA,$07,$D0,$77,$8C,$36,$33
	.dc.b	$87,$96,$72,$DF,$75,$F4,$1C,$5B,$19,$8C,$1D,$7D,$FF,$E8,$71,$CD
	.dc.b	$B8,$F9,$1A,$FF,$BC,$5A,$FF,$BC,$3C,$2F,$E0,$2F,$F3,$26,$E1,$FE
	.dc.b	$4C,$8F,$54,$63,$E8,$3F,$9F,$47,$FB,$ED,$AF,$FB,$E9,$AF,$F3,$C6
	.dc.b	$C9,$A4,$F0,$7E,$33,$C3,$85,$F4,$0F,$A0,$FB,$11,$86,$C6,$73,$FB
	.dc.b	$D3,$97,$E6,$7D,$03,$BB,$E9,$36,$33,$18,$23,$82,$67,$F7,$27,$82
	.dc.b	$3F,$91,$FC,$03,$E8,$3C,$AD,$AF,$1B,$67,$81,$C7,$04,$66,$33,$AB
	.dc.b	$E8,$1F,$2F,$F8,$96,$E1,$F8,$76,$D7,$E4,$1A,$F2,$6B,$E7,$AE,$95
	.dc.b	$CC,$95,$F2,$3E,$39,$F4,$1F,$39,$B1,$98,$C1,$F9,$BF,$91,$7C,$0D
	.dc.b	$24,$FE,$6E,$32,$D8,$CE,$3F,$A6,$9C,$B6,$3E,$EF,$C9,$7E,$7A,$3E
	.dc.b	$39,$FC,$5D,$C3,$FD,$49,$AF,$CB,$6A,$9F,$5F,$BD,$FB,$F6,$DF,$F0
	.dc.b	$36,$CF,$66,$BF,$59,$AF,$DF,$6B,$D7,$05,$75,$6D,$7E,$4B,$B7,$F9
	.dc.b	$25,$7F,$68,$6B,$F8,$16,$BF,$CC,$DB,$7F,$E1,$AC,$FD,$1E,$97,$DA
	.dc.b	$AF,$63,$67,$57,$EB,$8C,$12,$B8,$9F,$A8,$BF,$F5,$BD,$F8,$FA,$30
	.dc.b	$4B,$F8,$9F,$41,$D4,$97,$F2,$5F,$05,$FF,$1F,$BF,$8C,$D7,$E3,$35
	.dc.b	$FA,$4D,$C7,$F9,$DB,$F4,$CC,$10,$FD,$EF,$D0,$3E,$1E,$93,$1F,$DE
	.dc.b	$C6,$5B,$19,$BF,$C3,$F2,$2B,$51,$86,$BF,$B2,$F1,$DF,$2F,$FD,$A7
	.dc.b	$0F,$E8,$D1,$8F,$2F,$FA,$33,$63,$33,$BA,$FC,$AC,$E5,$B1,$97,$B3
	.dc.b	$5F,$E1,$11,$B7,$F3,$BF,$3E,$FA,$0F,$EA,$0D,$BF,$EA,$8D,$9F,$EB
	.dc.b	$4D,$7F,$85,$F9,$2F,$A0,$FF,$42,$57,$F5,$F8,$C7,$BD,$FD,$7D,$B1
	.dc.b	$99,$DD,$7E,$BA,$72,$D8,$CB,$D9,$AF,$F2,$18,$DB,$FD,$A7,$CE,$FE
	.dc.b	$D4,$FF,$25,$6A,$69,$FD,$BF,$C5,$7C,$33,$E8,$30,$DA,$FE,$F1,$3A
	.dc.b	$9C,$9C,$7D,$86,$14,$95,$C2,$32,$FB,$7D,$37,$B5,$F2,$34,$F5,$46
	.dc.b	$27,$D9,$18,$9D,$47,$F3,$FF,$A0,$7C,$DE,$31,$FF,$24,$55,$C3,$95
	.dc.b	$F9,$30,$E4,$F4,$39,$5F,$45,$F3,$5F,$CF,$9F,$F0,$93,$5F,$F0,$C3
	.dc.b	$5F,$E4,$6D,$7F,$8A,$7E,$95,$FE,$06,$7B,$75,$A3,$11,$E3,$3F,$92
	.dc.b	$38,$CB,$FC,$6F,$5D,$D6,$1D,$3C,$FF,$83,$87,$36,$DF,$19,$FE,$B7
	.dc.b	$7C,$5F,$EC,$9F,$51,$F5,$3D,$DF,$92,$DC,$BE,$13,$5F,$27,$63,$E8
	.dc.b	$7F,$C7,$11,$AD,$3F,$D1,$AD,$BE,$71,$88,$E3,$3C,$1F,$75,$BE,$2D
	.dc.b	$F3,$7E,$26,$CF,$89,$AF,$C5,$D9,$F4,$2F,$94,$E1,$72,$B8,$3F,$12
	.dc.b	$75,$1F,$77,$4F,$E1,$7E,$97,$F9,$2D,$B1,$FE,$14,$F0,$3F,$C2,$4D
	.dc.b	$9F,$F2,$86,$6F,$A8,$7D,$E7,$C8,$6D,$7C,$86,$BF,$22,$F9,$B7,$C6
	.dc.b	$F0,$47,$F9,$A2,$3C,$11,$82,$38,$21,$FC,$A2,$38,$3F,$DC,$2F,$FD
	.dc.b	$23,$F5,$9F,$6B,$C4,$34,$FD,$2E,$7F,$CA,$B0,$BE,$77,$E9,$B0,$FF
	.dc.b	$19,$B5,$FE,$6A,$D7,$FB,$D7,$4F,$C7,$8C,$47,$F7,$B8,$C1,$1F,$F3
	.dc.b	$D6,$2F,$AC,$FF,$4D,$C7,$E0,$1B,$CB,$F9,$A8,$4E,$FC,$7F,$2D,$FA
	.dc.b	$3D,$1F,$D8,$D7,$F8,$87,$AB,$EC,$46,$23,$FA,$5B,$5F,$F7,$0B,$5F
	.dc.b	$F1,$D3,$5F,$F5,$63,$74,$39,$C1,$F4,$5F,$E2,$8A,$FD,$1B,$5F,$CC
	.dc.b	$B7,$2F,$F4,$FD,$78,$DF,$F5,$0F,$D1,$6E,$BF,$DE,$6D,$7F,$C9,$8D
	.dc.b	$7F,$D7,$CD,$C0,$D2,$7A,$7F,$97,$78,$63,$FE,$FC,$FA,$BF,$DC,$F6
	.dc.b	$9A,$7E,$02,$7C,$DD,$DF,$E5,$BF,$5D,$F1,$9F,$A1,$6E,$7F,$E6,$28
	.dc.b	$C1,$1C,$10,$FE,$F3,$1C,$1F,$C9,$7E,$16,$FB,$BF,$E1,$6D,$8F,$E1
	.dc.b	$6D,$7F,$85,$CF,$1D,$BF,$05,$EC,$BE,$93,$FA,$AB,$6B,$FD,$80,$D7
	.dc.b	$FD,$65,$1A,$23,$A6,$BF,$EB,$5F,$81,$FE,$8B,$2F,$82,$46,$1B,$3F
	.dc.b	$D0,$67,$B3,$FC,$CD,$79,$DF,$E4,$A8,$D3,$5D,$37,$0F,$5F,$B2,$FA
	.dc.b	$4F,$D8,$FB,$3F,$74,$9C,$BF,$C4,$5F,$B7,$F3,$1F,$C7,$AF,$F4,$9C
	.dc.b	$C8,$C7,$1F,$F3,$86,$E6,$D3,$D9,$C1,$FD,$5F,$E6,$5F,$49,$FC,$03
	.dc.b	$E2,$DF,$49,$FE,$9C,$F6,$1E,$0F,$FB,$CA,$30,$FF,$A8,$2F,$8E,$5F
	.dc.b	$D8,$C3,$87,$FA,$1D,$F4,$7F,$C1,$53,$CD,$A7,$FA,$C6,$78,$6D,$FB
	.dc.b	$FE,$D5,$F1,$AF,$F0,$85,$F1,$68,$9D,$E1,$E6,$8E,$99,$7F,$1C,$FA
	.dc.b	$4F,$F5,$AC,$78,$38,$1F,$66,$FA,$0E,$11,$87,$96,$2B,$E3,$7E,$08
	.dc.b	$C6,$FC,$30,$E5,$3D,$5A,$EE,$B8,$EF,$F3,$23,$D1,$F2,$2F,$A5,$EE
	.dc.b	$D8,$F8,$B8,$A3,$11,$E1,$7F,$AB,$B6,$78,$91,$E3,$1A,$AF,$07,$13
	.dc.b	$E9,$7F,$CE,$95,$DB,$E5,$3F,$2D,$5E,$5C,$51,$8E,$1F,$F6,$85,$F0
	.dc.b	$BF,$A3,$87,$3F,$85,$F8,$F7,$B9,$FA,$8D,$9D,$47,$7F,$89,$D9,$1F
	.dc.b	$C1,$C6,$9F,$9D,$E9,$BE,$6B,$EA,$FC,$9F,$F5,$89,$FB,$4D,$77,$1F
	.dc.b	$62,$78,$46,$BD,$1E,$26,$67,$EE,$FC,$C7,$D6,$DF,$E0,$37,$6F,$F7
	.dc.b	$E3,$6F,$F1,$CF,$F9,$79,$FA,$AE,$77,$34,$9F,$09,$F9,$39,$FD,$2F
	.dc.b	$B5,$F8,$CF,$9A,$7D,$2D,$35,$DF,$F5,$F2,$FD,$75,$66,$FF,$07,$7C
	.dc.b	$1F,$D4,$9D,$0B,$5C,$5F,$D2,$46,$DF,$CC,$37,$3F,$ED,$F3,$BC,$FF
	.dc.b	$5F,$F3,$3F,$A0,$EF,$34,$C7,$89,$71,$FC,$AC,$E9,$B3,$A7,$F3,$3F
	.dc.b	$4A,$F6,$7F,$56,$D8,$CC,$61,$CB,$1C,$9C,$8F,$DE,$72,$BE,$97,$FC
	.dc.b	$EB,$A7,$E5,$FE,$7D,$5F,$B8,$6B,$F8,$C9,$DF,$06,$AF,$97,$C3,$EE
	.dc.b	$A1,$B7,$BD,$70,$8F,$E1,$D7,$D1,$FE,$04,$DC,$3F,$5B,$3B,$D9,$C1
	.dc.b	$F8,$7B,$FA,$3C,$7F,$81,$3F,$05,$E1,$58,$FB,$B1,$5F,$AF,$1F,$C4
	.dc.b	$BE,$33,$F5,$96,$7F,$C7,$B8,$FE,$DD,$BA,$FC,$29,$D7,$5F,$13,$97
	.dc.b	$BB,$31,$FE,$1B,$7F,$B7,$7F,$9C,$F4,$FF,$47,$7F,$96,$98,$E0,$FC
	.dc.b	$6C,$67,$2F,$E3,$5F,$08,$FA,$5F,$8D,$7B,$38,$F4,$7E,$E7,$0E,$27
	.dc.b	$F4,$0A,$E1,$C7,$FA,$0C,$7A,$6D,$FD,$35,$B8,$7F,$45,$6D,$70,$8E
	.dc.b	$67,$68,$DB,$E1,$CB,$CB,$97,$93,$2F,$14,$EF,$2E,$35,$AC,$BC,$51
	.dc.b	$8C,$BC,$4F,$9F,$0E,$58,$F6,$C7,$9E,$1F,$AA,$8F,$2C,$EA,$B9,$63
	.dc.b	$E9,$79,$63,$DA,$3C,$44,$BF,$13,$C0,$F5,$63,$E2,$7F,$AF,$E5,$F0
	.dc.b	$5E,$11,$99,$F1,$5F,$69,$E5,$97,$9A,$B4,$7E,$9A,$F9,$5F,$0A,$E6
	.dc.b	$79,$CE,$F8,$FA,$DC,$9E,$43,$FD,$68,$E6,$39,$12,$BB,$46,$5F,$0C
	.dc.b	$F5,$9F,$7B,$D2,$FC,$CE,$67,$8E,$3F,$1E,$BA,$4E,$4F,$0C,$66,$7A
	.dc.b	$C7,$97,$0F,$6C,$62,$FA,$CE,$70,$F2,$3C,$CC,$5F,$37,$13,$E4,$BE
	.dc.b	$71,$8C,$BF,$04,$7E,$02,$B8,$47,$C5,$39,$9F,$96,$3F,$84,$57,$E4
	.dc.b	$9B,$3F,$AF,$6E,$5F,$C3,$5F,$F1,$27,$D2,$FF,$27,$78,$B5,$FF,$19
	.dc.b	$35,$DF,$2F,$17,$FC,$67,$1F,$CB,$F4,$FE,$7E,$DD,$3F,$05,$E6,$3D
	.dc.b	$0F,$F5,$75,$6B,$0F,$E9,$77,$BC,$3A,$9F,$92,$BE,$1E,$76,$AA,$4F
	.dc.b	$F0,$0A,$E1,$B7,$F9,$57,$1E,$B9,$7A,$2F,$79,$77,$C3,$A5,$69,$EF
	.dc.b	$19,$25,$7A,$27,$53,$EE,$AE,$73,$D6,$39,$E3,$87,$9C,$66,$FF,$8C
	.dc.b	$E5,$EA,$9F,$45,$ED,$E8,$72,$32,$49,$F1,$AC,$7D,$D8,$71,$D5,$9F
	.dc.b	$C9,$AE,$77,$E0,$CF,$D3,$5A,$C7,$85,$FC,$8C,$BC,$97,$DB,$0E,$39
	.dc.b	$FA,$4F,$85,$97,$0C,$3D,$D9,$F8,$AB,$CD,$5B,$7E,$09,$E7,$C3,$AC
	.dc.b	$66,$3A,$9A,$9E,$58,$78,$72,$F3,$4F,$43,$96,$7E,$AC,$FD,$D3,$AA
	.dc.b	$F8,$B7,$CB,$0E,$D3,$97,$C9,$19,$C3,$CA,$78,$8F,$1B,$97,$13,$C5
	.dc.b	$F3,$63,$EA,$78,$C6,$5E,$31,$97,$8E,$1E,$63,$E7,$63,$E9,$CB,$E5
	.dc.b	$4F,$DF,$CB,$E0,$56,$E7,$FA,$C5,$F3,$C3,$C5,$C6,$FA,$57,$C3,$7E
	.dc.b	$66,$1F,$4B,$FE,$B1,$DF,$FC,$F7,$5A,$7F,$CF,$B1,$9D,$3F,$D1,$93
	.dc.b	$97,$FD,$1D,$19,$D3,$FC,$E9,$39,$7F,$CE,$D1,$9D,$FF,$C0,$1C,$8F
	.dc.b	$A5,$F1,$37,$1F,$EF,$FF,$4B,$89,$91,$FE,$02,$77,$18,$CB,$FC,$FB
	.dc.b	$F4,$BF,$EE,$2A,$FF,$49,$F7,$3C,$8F,$DB,$73,$2B,$57,$FE,$24,$9C
	.dc.b	$C6,$35,$FF,$1B,$46,$3D,$9F,$8D,$87,$3C,$7F,$D0,$37,$BD,$7F,$C3
	.dc.b	$53,$9C,$7C,$77,$C3,$0F,$9D,$F3,$1F,$81,$E1,$34,$FF,$44,$65,$FE
	.dc.b	$AA,$71,$CF,$C0,$E3,$D3,$67,$FD,$5F,$D5,$E1,$CB,$FD,$98,$DC,$0B
	.dc.b	$DE,$FB,$CE,$49,$9E,$23,$2F,$F1,$6C,$DF,$4B,$F8,$5F,$09,$F2,$79
	.dc.b	$8C,$37,$3F,$F3,$FB,$73,$FF,$3C,$B7,$8B,$FD,$6D,$7C,$AB,$FD,$07
	.dc.b	$87,$67,$89,$FE,$76,$F8,$3F,$CA,$79,$06,$C6,$18,$EC,$FF,$3E,$98
	.dc.b	$C7,$FD,$C1,$E6,$3E,$97,$F9,$B7,$B1,$F8,$3F,$9D,$FD,$04,$E6,$FF
	.dc.b	$D6,$B1,$8A,$DE,$48,$C4,$E8,$FF,$6F,$FA,$3F,$CA,$1B,$E2,$7E,$3B
	.dc.b	$E5,$3F,$D2,$71,$83,$FD,$8E,$E3,$3F,$F6,$3C,$EF,$CE,$FD,$95,$F3
	.dc.b	$9F,$F6,$BD,$6F,$83,$FD,$BD,$39,$7F,$DC,$B1,$9E,$BF,$F3,$54,$E5
	.dc.b	$FF,$36,$7A,$7F,$E9,$BC,$BF,$DD,$1E,$3F,$EF,$E3,$FD,$23,$18,$26
	.dc.b	$2C,$7F,$DD,$93,$C6,$31,$7D,$1E,$57,$FC,$27,$D6,$7D,$2F,$CE,$F6
	.dc.b	$9F,$63,$98,$C3,$73,$FF,$50,$E6,$CB,$FD,$47,$C8,$C3,$8D,$6A,$7F
	.dc.b	$D0,$D8,$F6,$8C,$66,$BE,$18,$FE,$23,$E9,$A5,$7E,$3B,$CF,$3F,$EB
	.dc.b	$0F,$80,$B8,$3C,$07,$53,$0F,$D1,$61,$AB,$DD,$6A,$73,$87,$33,$FD
	.dc.b	$35,$7C,$5F,$01,$9C,$3B,$46,$5F,$F4,$BF,$3B,$1E,$D5,$CA,$FF,$2F
	.dc.b	$F6,$7F,$CB,$F2,$F3,$E8,$8C,$57,$FB,$17,$E9,$9F,$4F,$CB,$3F,$89
	.dc.b	$5A,$7D,$DF,$01,$F9,$BB,$FF,$1C,$6C,$8C,$4F,$FB,$1E,$31,$D7,$F1
	.dc.b	$A3,$FD,$AB,$18,$AF,$F2,$07,$94,$F8,$77,$CD,$78,$63,$0D,$F8,$8F
	.dc.b	$E7,$98,$37,$FA,$35,$AC,$3E,$32,$DB,$F9,$54,$E9,$B3,$FE,$E3,$F2
	.dc.b	$BF,$33,$7C,$E3,$E1,$71,$F8,$67,$33,$FE,$EC,$C3,$FC,$90,$FA,$CE
	.dc.b	$45,$0E,$AB,$DB,$3C,$A7,$71,$87,$84,$ED,$FF,$77,$70,$BE,$29,$F7
	.dc.b	$FF,$4E,$30,$FD,$5D,$DF,$4C,$FA,$7D,$5F,$0C,$BF,$11,$5A,$9F,$33
	.dc.b	$63,$EC,$D7,$16,$CE,$A3,$18,$FD,$B6,$BF,$69,$AF,$FB,$C5,$AF,$EA
	.dc.b	$FC,$77,$C1,$7A,$5B,$7E,$4F,$88,$FC,$3F,$0F,$E5,$27,$37,$F4,$CF
	.dc.b	$D6,$4F,$E3,$1F,$C4,$B5,$FF,$80,$0D,$77,$CB,$EE,$8C,$66,$69,$23
	.dc.b	$EC,$6C,$F9,$27,$AF,$F3,$DB,$9F,$CA,$F9,$27,$D2,$E6,$BE,$94,$62
	.dc.b	$7F,$46,$DC,$FF,$44,$D7,$F4,$2D,$47,$F0,$97,$0D,$AF,$AF,$18,$CC
	.dc.b	$EB,$17,$CB,$7F,$63,$BF,$91,$18,$F4,$3F,$0D,$F2,$FF,$AC,$CB,$F8
	.dc.b	$54,$E5,$FC,$2C,$66,$BF,$70,$D7,$F6,$ED,$7C,$F8,$A3,$11,$FC,$85
	.dc.b	$B1,$9E,$BF,$C7,$C6,$0D,$DB,$C9,$FC,$0E,$1F,$CE,$1B,$87,$F2,$BF
	.dc.b	$0B,$F9,$B3,$EC,$8C,$E3,$FC,$7F,$C9,$FC,$DD,$6A,$BF,$86,$B5,$F3
	.dc.b	$B5,$E9,$1C,$09,$7F,$C6,$A7,$58,$3E,$65,$F6,$5F,$B1,$6D,$7E,$B2
	.dc.b	$BF,$D5,$5C,$DF,$CB,$27,$38,$FE,$67,$2F,$E6,$9D,$BE,$62,$61,$DD
	.dc.b	$FC,$8F,$D7,$FF,$2D,$CC,$FF,$2F,$F9,$EF,$E7,$99,$FF,$01,$8D,$13
	.dc.b	$BF,$F9,$B4,$EB,$0F,$BB,$3C,$F4,$F9,$27,$F4,$02,$C7,$DA,$FA,$67
	.dc.b	$C8,$7F,$0D,$9C,$EB,$FD,$8D,$B9,$FF,$2F,$E0,$FE,$21,$7C,$30,$FC
	.dc.b	$5D,$70,$C9,$F8,$F7,$F2,$C8,$FE,$CB,$18,$C7,$FA,$67,$13,$E7,$BF
	.dc.b	$BC,$61,$FD,$76,$B4,$DC,$39,$3F,$A6,$8C,$DF,$F0,$5B,$E5,$E7,$FC
	.dc.b	$2B,$DC,$FE,$37,$0E,$AF,$EB,$8F,$D7,$D7,$0C,$31,$18,$9D,$BB,$9D
	.dc.b	$3A,$C1,$E3,$BE,$CF,$E8,$B7,$0F,$EE,$D9,$BE,$9B,$F8,$4E,$1F,$C6
	.dc.b	$F0,$E3,$87,$02,$5F,$0A,$DD,$FC,$8C,$3A,$61,$FC,$8C,$F5,$7A,$7E
	.dc.b	$87,$D2,$D7,$25,$3F,$C6,$46,$E1,$F1,$EF,$A6,$E1,$18,$6C,$66,$78
	.dc.b	$35,$F8,$F3,$FE,$7A,$BE,$3D,$BF,$DC,$E7,$B3,$5D,$B5,$FE,$3F,$6F
	.dc.b	$A6,$7C,$F7,$A6,$30,$D8,$CC,$FB,$E7,$2D,$8C,$BF,$6A,$32,$D8,$CB
	.dc.b	$F8,$18,$CB,$F8,$0F,$C6,$3F,$42,$FC,$96,$BF,$35,$AE,$F9,$FC,$9F
	.dc.b	$45,$23,$C4,$D7,$C0,$D7,$C1,$68,$C4,$7F,$95,$B2,$7D,$D7,$D9,$E9
	.dc.b	$FF,$0E,$36,$33,$3F,$7E,$72,$D8,$CB,$FE,$08,$6B,$F9,$F8,$DB,$99
	.dc.b	$FC,$37,$23,$F0,$5F,$F5,$24,$FE,$2E,$31,$E5,$EA,$73,$3F,$A6,$9C
	.dc.b	$BB,$8C,$BF,$A7,$6B,$FA,$E6,$BF,$CF,$DA,$BF,$0F,$FC,$B3,$4F,$CC
	.dc.b	$38,$3F,$35,$E0,$BF,$0A,$FF,$69,$4F,$F9,$41,$AF,$F9,$3D,$AF,$F9
	.dc.b	$39,$AF,$F0,$1F,$93,$7C,$DF,$E9,$5B,$1F,$A9,$6B,$FA,$D6,$BF,$C1
	.dc.b	$9A,$F3,$8E,$44,$7E,$57,$94,$F9,$DF,$9E,$D8,$CC,$60,$FA,$7F,$67
	.dc.b	$F9,$0E,$CF,$E2,$0D,$9F,$E2,$ED,$7F,$B4,$37,$53,$49,$C9,$FC,$7E
	.dc.b	$B9,$B6,$B9,$BF,$C2,$BC,$1F,$F5,$3F,$99,$E5,$9F,$F5,$A9,$FE,$A3
	.dc.b	$71,$DB,$FE,$AD,$8F,$0D,$71,$7F,$D4,$7F,$8D,$FF,$8A,$1F,$F3,$3E
	.dc.b	$08,$C4,$7F,$A0,$63,$1E,$7F,$8E,$38,$B6,$33,$1C,$A7,$8C,$7A,$DB
	.dc.b	$1F,$ED,$06,$BC,$8F,$F5,$DF,$37,$F8,$93,$D1,$F3,$36,$3F,$D7,$F5
	.dc.b	$D9,$F7,$8E,$17,$C7,$FF,$9E,$A3,$31,$8F,$2B,$FD,$8F,$1B,$9D,$69
	.dc.b	$F1,$7E,$65,$F3,$4F,$91,$F8,$D7,$BC,$38,$E3,$CE,$3A,$E3,$CB,$87
	.dc.b	$FC,$87,$8F,$2C,$38,$DE,$F2,$FA,$58,$73,$38,$E3,$FE,$DA,$BE,$11
	.dc.b	$DB,$1E,$8F,$F9,$3A,$F9,$63,$FA,$39,$CD,$70,$AE,$64,$C3,$BD,$F0
	.dc.b	$84,$EE,$73,$5D,$A7,$C9,$18,$C3,$C9,$97,$C3,$9F,$24,$47,$B6,$3F
	.dc.b	$02,$75,$18,$3D,$17,$D7,$1F,$16,$3E,$1B,$F5,$61,$CF,$1F,$06,$1F
	.dc.b	$9A,$AD,$3C,$A7,$55,$C2,$35,$8F,$82,$F8,$92,$74,$FD,$D8,$D4,$60
	.dc.b	$93,$D4,$E7,$97,$D8,$CB,$D0,$4C,$3E,$1C,$65,$F1,$CF,$1C,$3B,$4E
	.dc.b	$63,$E1,$1F,$32,$FE,$34,$66,$7C,$D3,$C6,$72,$F4,$38,$19,$3E,$CE
	.dc.b	$1E,$A7,$06,$8D,$E3,$F2,$1C,$65,$F8,$17,$33,$C5,$CE,$5E,$D7,$E0
	.dc.b	$9E,$19,$E0,$7D,$89,$E0,$E2,$78,$61,$F0,$2B,$84,$63,$0F,$5C,$7C
	.dc.b	$2A,$E1,$97,$CA,$8D,$BF,$2B,$1F,$44,$EA,$7F,$15,$18,$BE,$91,$F0
	.dc.b	$E3,$15,$CA,$3D,$27,$79,$DC,$EA,$31,$3E,$16,$9D,$B1,$F6,$D7,$68
	.dc.b	$F8,$F9,$7D,$7C,$7F,$0F,$18,$3C,$19,$7D,$AC,$7B,$3F,$72,$34,$76
	.dc.b	$75,$87,$6C,$78,$DF,$C6,$9E,$19,$7C,$17,$E6,$5F,$19,$FC,$BE,$1D
	.dc.b	$9F,$1D,$7D,$5C,$7C,$D5,$F8,$48,$C3,$D8,$F9,$18,$F8,$5E,$87,$E3
	.dc.b	$27,$8B,$F8,$4C,$7E,$36,$3E,$4C,$BE,$5D,$F1,$8C,$4E,$B1,$F1,$D6
	.dc.b	$B0,$E7,$8F,$AE,$7D,$C6,$AB,$F0,$18,$73,$8C,$63,$F6,$B1,$EA,$FD
	.dc.b	$5C,$39,$E3,$D6,$3D,$87,$CE,$9D,$9F,$3A,$77,$87,$66,$CF,$33,$8E
	.dc.b	$3F,$19,$F8,$67,$3C,$3C,$77,$CE,$37,$87,$F3,$0C,$7D,$11,$F3,$E3
	.dc.b	$18,$76,$AE,$13,$D3,$0E,$D1,$E6,$C3,$ED,$E1,$F4,$EB,$FC,$2D,$87
	.dc.b	$F5,$96,$E3,$FE,$1D,$8C,$47,$F8,$8F,$E2,$FF,$DE,$4F,$F8,$7D,$AF
	.dc.b	$F8,$7D,$AE,$F9,$3F,$C4,$D1,$C3,$1F,$E8,$AD,$C7,$FA,$CF,$CD,$3D
	.dc.b	$16,$FF,$D8,$2B,$58,$FF,$53,$6B,$FD,$C5,$CE,$EB,$E5,$5C,$72,$97
	.dc.b	$FD,$D2,$F7,$5A,$8C,$4E,$B2,$F2,$BF,$D9,$67,$51,$89,$D9,$1F,$EF
	.dc.b	$51,$AC,$BE,$04,$E5,$B3,$AC,$BD,$B3,$93,$D9,$7E,$1A,$F8,$73,$AA
	.dc.b	$F0,$12,$39,$9A,$C7,$85,$F2,$3A,$65,$F0,$EF,$95,$70,$9D,$5F,$37
	.dc.b	$FB,$AE,$3E,$97,$33,$B7,$E7,$63,$F0,$EF,$9C,$68,$86,$8F,$ED,$55
	.dc.b	$CD,$C5,$F7,$6D,$FA,$5F,$84,$60,$E9,$87,$A7,$1F,$6C,$64,$D4,$65
	.dc.b	$E3,$B3,$3F,$C2,$B7,$3F,$45,$F2,$8C,$57,$C6,$CB,$CD,$18,$CB,$CC
	.dc.b	$DC,$BC,$D1,$8C,$7A,$9F,$8A,$CF,$E1,$E1,$E5,$BE,$0F,$AB,$0E,$B5
	.dc.b	$C3,$2F,$AF,$9F,$A7,$1E,$91,$8A,$DB,$E7,$AE,$D9,$D8,$D6,$5E,$0C
	.dc.b	$BB,$E5,$E1,$BD,$BF,$11,$F0,$3F,$52,$78,$46,$31,$FC,$16,$5E,$3C
	.dc.b	$FF,$01,$87,$43,$D7,$19,$C3,$EF,$C6,$6F,$BD,$FA,$A7,$E6,$61,$DB
	.dc.b	$1F,$44,$78,$72,$B5,$E0,$CF,$DF,$3E,$7A,$FA,$ED,$39,$D7,$0C,$FF
	.dc.b	$35,$9F,$8E,$7E,$05,$7C,$4A,$D4,$7D,$1C,$7C,$4F,$3C,$7F,$65,$3A
	.dc.b	$8F,$C5,$E3,$E2,$9F,$51,$A8,$E1,$8F,$82,$74,$FC,$6C,$BE,$0B,$71
	.dc.b	$F1,$57,$03,$C5,$87,$7C,$FE,$5E,$1C,$A7,$E5,$E1,$C9,$B7,$CB,$1F
	.dc.b	$1E,$3E,$1C,$BD,$71,$8A,$FD,$CC,$62,$7C,$51,$F3,$CF,$11,$EE,$8D
	.dc.b	$3D,$8D,$4F,$37,$59,$79,$F0,$F2,$47,$CE,$3E,$99,$D3,$2F,$5D,$F4
	.dc.b	$8D,$5F,$C9,$3A,$4E,$AF,$A6,$3F,$7A,$73,$87,$6A,$FC,$64,$FC,$1C
	.dc.b	$7D,$75,$F8,$7C,$7B,$56,$B1,$F4,$D6,$8F,$8D,$8F,$F2,$41,$F5,$EF
	.dc.b	$C1,$7F,$8C,$F1,$39,$DF,$80,$FF,$1C,$BD,$D6,$A7,$31,$FE,$C2,$BE
	.dc.b	$4F,$FB,$61,$B8,$FF,$B5,$DB,$1F,$EE,$B6,$C0,$FE,$BA,$E1,$B5,$FD
	.dc.b	$12,$78,$36,$38,$47,$90,$AF,$FB,$FB,$EF,$7C,$DF,$23,$FC,$D3,$7F
	.dc.b	$F7,$EC,$62,$BC,$1B,$FF,$BC,$E7,$31,$CA,$D1,$87,$FA,$A9,$8E,$B5
	.dc.b	$F8,$E3,$0D,$9E,$15,$ED,$9C,$BD,$63,$31,$E7,$8F,$07,$80,$F8,$67
	.dc.b	$A1,$E0,$AF,$8D,$39,$8C,$42,$31,$5D,$E3,$75,$CA,$31,$1F,$3F,$E1
	.dc.b	$1F,$37,$CA,$11,$89,$DC,$62,$75,$39,$8E,$6D,$7C,$0D,$7E,$03,$5F
	.dc.b	$04,$7A,$2D,$F3,$CF,$C1,$30,$DC,$3C,$F3,$A9,$C9,$E7,$FC,$41,$F4
	.dc.b	$54,$95,$E8,$8C,$BF,$03,$57,$94,$EE,$F9,$31,$86,$C6,$67,$83,$74
	.dc.b	$F9,$95,$B9,$CD,$70,$23,$F3,$BA,$1E,$6F,$F8,$AE,$FF,$05,$18,$EC
	.dc.b	$FB,$2D,$AE,$15,$BA,$FD,$24,$E5,$FB,$8D,$79,$C6,$D4,$65,$FD,$67
	.dc.b	$A2,$F5,$9E,$47,$E2,$23,$0D,$8C,$E1,$F8,$69,$CB,$63,$35,$F8,$D9
	.dc.b	$CB,$63,$2F,$E7,$E3,$2F,$E4,$FF,$03,$FE,$06,$DE,$BF,$C8,$46,$23
	.dc.b	$F4,$CD,$53,$E9,$FC,$57,$CC,$4C,$3C,$8D,$9F,$13,$5F,$16,$08,$FC
	.dc.b	$F4,$7F,$00,$3F,$3A,$3A,$EB,$4C,$BF,$A2,$F5,$DE,$C7,$92,$7F,$57
	.dc.b	$18,$C7,$53,$99,$FD,$94,$E5,$DF,$B1,$F8,$46,$EF,$D6,$31,$1C,$1A
	.dc.b	$FD,$AF,$59,$EB,$7E,$D3,$14,$62,$7E,$C3,$5F,$AC,$D7,$EA,$35,$E9
	.dc.b	$3C,$E3,$11,$F4,$1A,$BD,$7A,$49,$F9,$2F,$0A,$F1,$FC,$7F,$32,$57
	.dc.b	$91,$AF,$91,$AF,$91,$FE,$1F,$C7,$FA,$8F,$26,$E7,$FE,$59,$AF,$D2
	.dc.b	$C6,$08,$E0,$87,$E9,$A3,$83,$FB,$8E,$87,$B1,$FC,$73,$A7,$F8,$1B
	.dc.b	$63,$33,$FC,$02,$30,$47,$04,$3F,$95,$F9,$4F,$57,$F1,$6D,$8F,$E5
	.dc.b	$2D,$77,$1E,$1F,$55,$EA,$D2,$4F,$F1,$78,$CB,$CB,$EA,$BA,$91,$C1
	.dc.b	$0F,$1B,$96,$99,$CF,$FD,$5B,$E1,$3E,$49,$F5,$7F,$D5,$7A,$5E,$87
	.dc.b	$F4,$E9,$CE,$1F,$D9,$A3,$FA,$D6,$AC,$BF,$AE,$65,$FD,$D7,$D7,$7B
	.dc.b	$1F,$D7,$FB,$E4,$63,$1F,$ED,$90,$8C,$47,$F6,$C8,$C4,$F1,$8F,$EF
	.dc.b	$B3,$A9,$E7,$39,$9F,$F7,$1C,$E5,$DC,$66,$79,$C7,$A4,$FE,$4B,$E8
	.dc.b	$BD,$8F,$81,$9F,$F2,$E8,$C4,$EE,$B8,$BF,$C1,$67,$C7,$F6,$1D,$71
	.dc.b	$34,$93,$EE,$8C,$B6,$32,$FB,$E3,$2D,$8F,$1B,$FC,$EB,$E2,$9F,$05
	.dc.b	$8E,$A4,$62,$7C,$D7,$CB,$87,$EC,$C7,$F8,$AF,$83,$FA,$F7,$B3,$F0
	.dc.b	$61,$18,$7D,$86,$FC,$A4,$FC,$7F,$65,$ED,$3D,$BF,$2D,$F2,$BD,$BC
	.dc.b	$8C,$5F,$7C,$3B,$63,$E4,$C7,$9E,$1C,$A7,$3B,$FE,$33,$1E,$77,$EA
	.dc.b	$C7,$99,$E6,$C3,$8B,$D6,$77,$5C,$71,$C1,$E9,$CB,$D1,$58,$3C,$D8
	.dc.b	$F8,$1F,$81,$3B,$CB,$C9,$87,$49,$C9,$C1,$C6,$3E,$28,$C5,$75,$CB
	.dc.b	$D1,$87,$5A,$D4,$62,$FA,$5F,$B2,$31,$7E,$18,$E5,$97,$AD,$EA,$7C
	.dc.b	$1C,$BD,$6E,$2B,$8B,$67,$FC,$D5,$87,$81,$E5,$3A,$23,$AA,$E7,$18
	.dc.b	$AF,$7B,$6B,$94,$62,$B9,$91,$46,$9F,$64,$65,$A6,$AB,$BC,$67,$1F
	.dc.b	$A5,$3A,$AE,$51,$82,$63,$F4,$2F,$8D,$78,$C8,$71,$72,$F8,$09,$3F
	.dc.b	$4E,$32,$F4,$C7,$E7,$CF,$97,$2F,$C0,$C6,$23,$A4,$61,$F9,$C7,$87
	.dc.b	$0F,$3E,$1F,$72,$FA,$4F,$D7,$AE,$19,$7C,$D8,$E6,$75,$9D,$C7,$E9
	.dc.b	$71,$F2,$DF,$32,$63,$F8,$5B,$F0,$9E,$C8,$F8,$98,$76,$CB,$E6,$46
	.dc.b	$27,$B3,$C9,$F2,$18,$AF,$83,$97,$A9,$C4,$68,$EF,$97,$9A,$FF,$11
	.dc.b	$87,$79,$FA,$33,$F3,$63,$11,$E5,$AF,$46,$5F,$06,$B9,$4F,$CF,$3B
	.dc.b	$C7,$E3,$F2,$F8,$35,$8C,$3B,$D5,$3E,$84,$E9,$F2,$E5,$EF,$6D,$7B
	.dc.b	$27,$57,$E1,$AD,$36,$75,$8F,$A3,$1F,$2E,$5F,$0F,$1E,$67,$C3,$BE
	.dc.b	$0F,$E0,$A3,$51,$83,$ED,$57,$3C,$3E,$84,$62,$BD,$47,$E0,$8E,$AF
	.dc.b	$C1,$9E,$18,$79,$27,$78,$7C,$D6,$DF,$AB,$2F,$BF,$5B,$9E,$71,$FB
	.dc.b	$5C,$7C,$AE,$A3,$D4,$DC,$7C,$11,$D2,$BE,$BC,$E4,$E5,$97,$CA,$71
	.dc.b	$3F,$64,$EE,$7D,$4C,$BD,$2F,$C5,$9E,$0F,$E1,$B1,$F4,$36,$FD,$18
	.dc.b	$7C,$9C,$3B,$E1,$CA,$3C,$B5,$FC,$88,$FF,$6E,$65,$FD,$65,$B8,$7F
	.dc.b	$52,$6E,$1F,$E3,$0E,$FF,$F7,$CC,$7F,$8D,$9A,$F2,$6B,$FE,$1C,$6B
	.dc.b	$FD,$A5,$E3,$7F,$CF,$1B,$8F,$F2,$EF,$6D,$E7,$38,$FF,$B0,$56,$B0
	.dc.b	$E5,$E0,$E0,$FE,$D9,$7C,$2B,$13,$FD,$52,$B8,$3F,$DD,$9E,$5B,$AB
	.dc.b	$84,$EE,$F8,$E5,$E5,$7A,$E3,$E3,$9D,$12,$74,$FF,$6B,$AE,$71,$82
	.dc.b	$65,$F0,$EF,$9E,$5D,$A3,$17,$CD,$F8,$58,$77,$BE,$33,$F0,$0D,$1F
	.dc.b	$D1,$B0,$EF,$5C,$A7,$73,$EF,$C3,$BB,$C0,$C1,$CB,$1F,$85,$87,$7B
	.dc.b	$E4,$49,$F0,$5F,$F8,$3F,$0E,$F1,$89,$D3,$D2,$75,$5C,$E3,$44,$35
	.dc.b	$3E,$EC,$7E,$14,$67,$56,$7F,$5B,$1E,$F3,$C0,$D4,$67,$0E,$D1,$92
	.dc.b	$57,$2C,$FE,$0E,$5F,$12,$31,$3F,$07,$0E,$B1,$89,$D6,$1D,$71,$F2
	.dc.b	$DF,$48,$C3,$F2,$EB,$86,$5F,$1F,$3F,$97,$7C,$27,$35,$F0,$A7,$2F
	.dc.b	$1B,$FC,$AC,$EA,$73,$87,$8B,$2F,$8B,$18,$C7,$E7,$D7,$4C,$3E,$24
	.dc.b	$E5,$F8,$B5,$C3,$2F,$6E,$3D,$89,$1F,$13,$0E,$71,$8A,$FC,$D6,$7F
	.dc.b	$33,$1E,$C4,$9F,$83,$97,$96,$BE,$55,$6E,$79,$4F,$E6,$32,$F2,$57
	.dc.b	$C9,$C7,$C9,$7B,$7E,$35,$F4,$C7,$C7,$1A,$CF,$2D,$BF,$90,$DC,$7C
	.dc.b	$47,$7C,$FD,$F1,$C2,$7D,$75,$6B,$58,$F8,$67,$99,$A2,$C6,$A7,$86
	.dc.b	$1F,$77,$0E,$B5,$A8,$FC,$D3,$4F,$BD,$1A,$AF,$2C,$E9,$B3,$AB,$F1
	.dc.b	$46,$32,$F8,$F9,$FC,$FC,$BD,$D8,$7C,$4C,$3E,$7E,$5E,$47,$93,$5F
	.dc.b	$C1,$E5,$E1,$9E,$18,$FE,$D2,$B5,$87,$E3,$EB,$57,$F8,$83,$E1,$57
	.dc.b	$99,$E7,$97,$9B,$0E,$67,$AC,$B1,$E2,$3E,$11,$C3,$0E,$F9,$FD,$D8
	.dc.b	$ED,$7F,$35,$DE,$3E,$3C,$3B,$46,$B0,$EC,$D3,$B5,$79,$6B,$86,$1D
	.dc.b	$B2,$FC,$E5,$7F,$10,$87,$EA,$9F,$ED,$57,$FD,$B1,$C7,$FE,$C0,$FC
	.dc.b	$7B,$F8,$B4,$6B,$A3,$F8,$C4,$E6,$3F,$DB,$15,$C6,$FF,$DD,$8D,$8F
	.dc.b	$F7,$A3,$60,$7F,$59,$70,$DA,$FE,$95,$3C,$1B,$1C,$23,$C4,$7F,$BF
	.dc.b	$27,$8A,$F7,$1E,$87,$9B,$A9,$EE,$FA,$A7,$35,$EE,$9C,$BF,$EF,$18
	.dc.b	$CE,$1F,$D6,$CF,$ED,$CE,$23,$D3,$18,$BF,$2B,$63,$CC,$F9,$E7,$83
	.dc.b	$67,$73,$E7,$6C,$EA,$7E,$2F,$B0,$F8,$7F,$ED,$9D,$FF,$DB,$7C,$37
	.dc.b	$84,$F9,$6F,$E7,$B7,$BC,$FE,$6D,$EE,$B5,$1F,$EF,$39,$D6,$3E,$A3
	.dc.b	$BE,$3E,$F7,$DB,$1A,$74,$D7,$31,$B7,$8C,$F3,$69,$67,$82,$C5,$ED
	.dc.b	$BD,$CE,$51,$86,$C6,$6F,$A4,$E5,$ED,$4B,$7B,$9E,$38,$C3,$63,$33
	.dc.b	$C1,$B3,$F0,$79,$5E,$67,$F6,$CA,$F5,$46,$3A,$FD,$CD,$8C,$CE,$EB
	.dc.b	$93,$5F,$C2,$CE,$DC,$D7,$29,$D4,$7E,$8B,$99,$BB,$BF,$E6,$46,$1B
	.dc.b	$19,$C7,$E7,$4E,$5B,$19,$AF,$B9,$39,$6C,$65,$FC,$CC,$65,$FC,$DF
	.dc.b	$9C,$F5,$9F,$27,$F7,$63,$0D,$8C,$E1,$F7,$E7,$2D,$8C,$D7,$E3,$67
	.dc.b	$2D,$8C,$BE,$06,$BF,$AC,$F6,$1E,$EF,$DB,$6D,$FE,$01,$B3,$F8,$F6
	.dc.b	$BC,$FD,$AF,$31,$27,$C8,$D8,$F1,$35,$E1,$1F,$8C,$7F,$8A,$F8,$15
	.dc.b	$34,$8F,$E7,$FD,$87,$B2,$CF,$F2,$B1,$86,$C6,$67,$83,$5F,$D9,$4E
	.dc.b	$D7,$A8,$F4,$7F,$87,$CE,$63,$13,$FC,$AA,$31,$3A,$AE,$D5,$AA,$78
	.dc.b	$9F,$E9,$97,$F8,$CB,$5F,$E6,$6D,$77,$C5,$FE,$9D,$76,$D7,$6D,$77
	.dc.b	$E4,$FE,$D2,$79,$12,$74,$D8,$D4,$60,$8E,$09,$3F,$C2,$E7,$83,$67
	.dc.b	$6D,$9D,$BF,$EB,$23,$04,$30,$43,$D6,$F0,$69,$B2,$9B,$3D,$EE,$DA
	.dc.b	$6C,$A6,$CF,$86,$64,$A6,$67,$D6,$D9,$F5,$B6,$7B,$3C,$0F,$19,$0E
	.dc.b	$E5,$34,$60,$86,$09,$3E,$07,$83,$5D,$B5,$DC,$F9,$1D,$B5,$DB,$5D
	.dc.b	$CF,$A2,$79,$12,$74,$D8,$D4,$60,$8E,$09,$3F,$26,$78,$36,$76,$D9
	.dc.b	$DB,$F2,$4C,$10,$C1,$0F,$5B,$C1,$A6,$CA,$6C,$F7,$BB,$69,$B2,$9B
	.dc.b	$3E,$19,$92,$99,$9F,$5B,$67,$D6,$D9,$EC,$F0,$3C,$64,$3B,$94,$D1
	.dc.b	$82,$18,$24,$F8,$1E,$0D,$76,$D7,$73,$E4,$76,$D7,$6D,$77,$3E,$89
	.dc.b	$E4,$49,$D3,$63,$51,$82,$38,$24,$FC,$99,$E0,$D9,$DB,$67,$6F,$C9
	.dc.b	$30,$43,$04,$3D,$6F,$06,$9B,$29,$B3,$DE,$ED,$A6,$CA,$6C,$F8,$66
	.dc.b	$4A,$66,$7D,$6D,$9F,$8B,$CA,$E6,$FE,$54,$D9,$FE,$44,$D7,$FC,$71
	.dc.b	$5F,$EF,$66,$C6,$5F,$E7,$AD,$7F,$9E,$B5,$FE,$86,$D8,$E5,$1C,$27
	.dc.b	$C2,$47,$C2,$D7,$BB,$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$3B,$D5
	.dc.b	$A7,$22,$9C,$4A,$71,$3D,$AF,$16,$9C,$4A,$71,$29,$C6,$3E,$1B,$D0
	.dc.b	$86,$8A,$E9,$C1,$1C,$12,$7E,$71,$27,$C9,$19,$6C,$66,$7C,$D3,$96
	.dc.b	$C6,$5F,$44,$65,$B1,$97,$D4,$D7,$CE,$D7,$CC,$D8,$E5,$1C,$27,$C6
	.dc.b	$47,$C2,$D7,$BB,$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$3B,$D5,$A7
	.dc.b	$22,$9C,$4A,$71,$3D,$AF,$16,$9C,$4A,$71,$29,$C6,$3E,$1B,$D0,$86
	.dc.b	$8A,$E9,$C1,$1C,$12,$7E,$71,$27,$C9,$19,$6C,$66,$7C,$D3,$96,$C6
	.dc.b	$5F,$44,$65,$B1,$97,$D4,$D7,$CE,$D7,$CC,$D8,$E5,$1C,$27,$C6,$47
	.dc.b	$C2,$D7,$BB,$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F,$3B,$D5,$A7,$22
	.dc.b	$9C,$4A,$71,$3D,$AF,$16,$9C,$4A,$71,$29,$C6,$3E,$1B,$D0,$86,$8A
	.dc.b	$E9,$67,$FE,$72,$7F,$CE,$99,$7F,$9C,$B3,$FF,$3A,$68,$9F,$F3,$EC
	.dc.b	$61,$A7,$08,$FF,$43,$91,$FF,$43,$B7,$4F,$F3,$E5,$F7,$9F,$0D,$F7
	.dc.b	$79,$1D,$0F,$41,$30,$F4,$13,$2F,$C1,$13,$2F,$91,$19,$6C,$66,$7E
	.dc.b	$54,$E5,$B1,$97,$E6,$46,$5B,$19,$7E,$73,$5F,$96,$D7,$CC,$D8,$E5
	.dc.b	$1C,$27,$C6,$47,$C2,$D7,$BB,$5E,$06,$88,$68,$A6,$8C,$10,$C1,$0F
	.dc.b	$3B,$D5,$A7,$22,$9C,$4A,$71,$3D,$AF,$16,$9C,$4A,$71,$29,$C6,$3E
	.dc.b	$1B,$D0,$86,$8A,$E9,$C1,$1C,$12,$7E,$71,$27,$C9,$19,$6C,$66,$7C
	.dc.b	$D3,$96,$C6,$5F,$44,$65,$B1,$97,$D4,$D7,$CE,$D7,$FB,$56,$7F,$E3
	.dc.b	$E7,$FC,$4D,$A7,$F8,$C5,$FF,$15,$9F,$E3,$62,$1F,$E3,$22,$9A,$30
	.dc.b	$43,$04,$AF,$F1,$9B,$C1,$AE,$DA,$EE,$7C,$8E,$DA,$ED,$AE,$E7,$D1
	.dc.b	$3C,$89,$3A,$6C,$6A,$30,$47,$04,$9F,$F2,$64,$F0,$6C,$ED,$B3,$B7
	.dc.b	$E4,$98,$21,$82,$1E,$B7,$83,$4D,$94,$D9,$EF,$76,$D3,$65,$36,$7C
	.dc.b	$33,$25,$33,$3E,$B6,$CF,$AD,$B3,$D9,$E0,$78,$C8,$77,$29,$A3,$04
	.dc.b	$30,$49,$F0,$3C,$1A,$ED,$AE,$E7,$C8,$ED,$AF,$5C,$7F,$9A,$13,$1F
	.dc.b	$E6,$91,$FC,$DB,$3F,$E8,$59,$FF,$3A,$94,$7F,$3F,$9F,$E8,$C7,$F3
	.dc.b	$A2,$9C,$4A,$6B,$5A,$D2,$86,$B4,$E7,$0E,$27,$03,$83,$AC,$E6,$31
	.dc.b	$5C,$2B,$75,$A6,$C6,$67,$6F,$AA,$B9,$56,$E7,$33,$FD,$8E,$72,$FB
	.dc.b	$38,$5C,$5A,$8F,$ED,$11,$8B,$F7,$B6,$38,$4E,$EB,$D9,$C7,$71,$34
	.dc.b	$98,$7A,$5B,$7E,$D6,$BF,$4B,$91,$CA,$E0,$F2,$46,$1B,$19,$BF,$AB
	.dc.b	$39,$6C,$66,$BE,$DC,$E5,$B1,$97,$C0,$D7,$FC,$11,$C8,$E6,$72,$FC
	.dc.b	$58,$C3,$63,$38,$7E,$1A,$72,$D8,$CD,$7E,$32,$72,$D8,$CB,$F9,$28
	.dc.b	$CB,$F9,$2E,$77,$47,$D5,$84,$62,$FF,$26,$D8,$FC,$BB,$5F,$CE,$F4
	.dc.b	$BA,$9D,$7F,$7E,$30,$D8,$CD,$FE,$82,$72,$F6,$EC,$F1,$60,$69,$23
	.dc.b	$44,$C3,$C7,$19,$6C,$7F,$8D,$75,$FC,$FD,$BB,$5D,$DF,$A2,$E8,$78
	.dc.b	$DF,$B6,$CF,$C1,$8F,$13,$FC,$7D,$87,$1B,$E1,$5F,$E4,$19,$78,$0F
	.dc.b	$07,$F5,$B1,$FC,$00,$CB,$8D,$38,$4E,$E7,$94,$62,$74,$FE,$E6,$74
	.dc.b	$FE,$CA,$79,$9C,$1D,$4F,$08,$CC,$6E,$78,$E5,$EE,$87,$23,$C2,$78
	.dc.b	$7E,$C9,$E7,$87,$F9,$76,$F8,$5E,$EB,$52,$F1,$1E,$2F,$85,$B1,$E5
	.dc.b	$35,$19,$23,$98,$DE,$1E,$D6,$C7,$C9,$6C,$7C,$F7,$78,$79,$3C,$D7
	.dc.b	$8F,$E4,$9C,$ED,$FE,$75,$CD,$5A,$C7,$D9,$3F,$CD,$3C,$5F,$D8,$F4
	.dc.b	$E3,$3F,$F3,$E4,$61,$FE,$61,$3F,$E7,$FF,$29,$E5,$3C,$AF,$91,$4B
	.dc.b	$E1,$B7,$CF,$8C,$77,$32,$ED,$18,$7C,$18,$F4,$C3,$9E,$1C,$B2,$F3
	.dc.b	$61,$C7,$0F,$19,$E6,$71,$8F,$DD,$C3,$96,$7D,$B0,$E5,$7B,$CB,$18
	.dc.b	$70,$C7,$D9,$9F,$AF,$0C,$D4,$BE,$E7,$9C,$F0,$E3,$D9,$CE,$3D,$9D
	.dc.b	$65,$F0,$63,$C3,$3B,$8C,$65,$E1,$AF,$21,$C9,$C4,$EF,$2F,$83,$87
	.dc.b	$8E,$F8,$BF,$23,$2F,$46,$1C,$63,$55,$C2,$3C,$F3,$AA,$E3,$1B,$23
	.dc.b	$F2,$63,$53,$C6,$36,$47,$E3,$E7,$F5,$B1,$F5,$5F,$C0,$9E,$06,$B3
	.dc.b	$F1,$C7,$BA,$F9,$2C,$BD,$CE,$6B,$8B,$4D,$63,$E4,$70,$77,$35,$3C
	.dc.b	$9C,$13,$3F,$A3,$3C,$63,$18,$FA,$27,$39,$7A,$F3,$F3,$D7,$5C,$3A
	.dc.b	$57,$DD,$25,$FD,$6C,$3A,$46,$2F,$DF,$5C,$2B,$B1,$23,$EC,$61,$DF
	.dc.b	$0F,$A3,$8F,$C0,$BE,$15,$E2,$C3,$CB,$9D,$C7,$96,$1F,$73,$1E,$59
	.dc.b	$FC,$C8,$F4,$67,$E4,$AD,$CF,$D0,$CB,$AD,$78,$0F,$93,$5B,$CB,$B1
	.dc.b	$E6,$C6,$CE,$EB,$8D,$F7,$CB,$CB,$3C,$CF,$9B,$7C,$5C,$47,$C5,$7D
	.dc.b	$17,$D2,$BC,$98,$78,$67,$78,$FC,$DC,$BD,$B9,$FE,$12,$7D,$B1,$87
	.dc.b	$F0,$78,$F8,$B0,$F2,$5F,$C0,$BF,$C9,$57,$4C,$BE,$AE,$3D,$EF,$71
	.dc.b	$E7,$6C,$FB,$E7,$55,$E1,$C3,$B3,$70,$ED,$5F,$96,$C7,$C8,$FC,$23
	.dc.b	$DE,$78,$4E,$8E,$AB,$E2,$61,$FB,$18,$C5,$74,$CF,$E7,$E1,$DF,$2F
	.dc.b	$33,$71,$E7,$9F,$CB,$8F,$C5,$67,$E8,$9F,$8B,$87,$CB,$AD,$47,$E0
	.dc.b	$31,$EF,$87,$C4,$6C,$74,$3D,$19,$7D,$68,$CB,$6B,$9C,$FD,$46,$E7
	.dc.b	$F1,$A3,$C3,$7D,$67,$38,$78,$E7,$AE,$5E,$8C,$7A,$B6,$B7,$7F,$02
	.dc.b	$B9,$9E,$5C,$7E,$B9,$E3,$7E,$71,$FE,$8E,$C7,$FA,$D3,$71,$FE,$94
	.dc.b	$DA,$FF,$59,$4F,$F9,$8F,$C9,$FE,$91,$E0,$1A,$7F,$56,$F3,$1E,$67
	.dc.b	$FA,$5F,$4F,$ED,$55,$BC,$7F,$A1,$B7,$3F,$E8,$4D,$9F,$ED,$57,$CA
	.dc.b	$BA,$9C,$30,$B7,$FD,$6E,$75,$39,$24,$74,$9F,$26,$1E,$39,$FE,$F3
	.dc.b	$87,$76,$CF,$4C,$3C,$35,$B9,$FE,$F1,$5B,$6D,$6D,$FE,$6D,$87,$8B
	.dc.b	$0F,$04,$78,$B0,$F8,$35,$E2,$8E,$07,$23,$47,$2C,$3C,$18,$FC,$F8
	.dc.b	$C9,$BB,$F8,$13,$AC,$7C,$95,$B9,$F0,$9A,$AE,$F1,$9C,$3C,$71,$FD
	.dc.b	$DB,$1F,$44,$62,$B8,$B4,$F8,$73,$CA,$FB,$9F,$27,$75,$F2,$9F,$96
	.dc.b	$7A,$30,$F0,$3E,$D3,$96,$5F,$26,$B8,$E5,$EB,$8C,$CF,$96,$30,$F3
	.dc.b	$C3,$C1,$87,$78,$F8,$98,$F9,$DF,$0E,$5F,$05,$E1,$87,$7A,$E1,$97
	.dc.b	$92,$78,$9C,$2F,$DE,$F0,$3E,$AE,$5E,$67,$CB,$87,$C6,$7D,$86,$DC
	.dc.b	$46,$E3,$13,$E5,$8C,$4F,$19,$F0,$9E,$28,$CC,$FA,$E3,$35,$F2,$A7
	.dc.b	$31,$E7,$21,$C6,$FD,$B8,$FD,$6C,$7E,$06,$5F,$2B,$2F,$06,$3C,$EF
	.dc.b	$75,$EC,$CB,$E0,$D6,$B0,$E5,$8F,$D3,$34,$78,$72,$F1,$BE,$68,$DC
	.dc.b	$FD,$13,$95,$F8,$A7,$89,$E7,$CB,$C9,$87,$BF,$2F,$24,$62,$7D,$B1
	.dc.b	$87,$8B,$5F,$9A,$78,$F2,$F5,$B8,$CB,$CB,$3F,$7C,$99,$7E,$2A,$FB
	.dc.b	$57,$A7,$1F,$DB,$3D,$0F,$31,$E0,$9E,$0D,$8D,$E5,$EE,$8F,$C3,$E3
	.dc.b	$E2,$8C,$1F,$0F,$2F,$75,$FC,$67,$E7,$98,$BF,$8D,$7E,$1C,$7D,$D7
	.dc.b	$CA,$BF,$15,$87,$57,$85,$7D,$53,$86,$1E,$66,$E1,$E6,$9D,$DF,$E5
	.dc.b	$AF,$95,$7C,$EB,$E1,$3D,$E3,$E6,$63,$F1,$B1,$F9,$38,$F3,$9E,$0F
	.dc.b	$A2,$B8,$47,$C8,$3C,$27,$95,$CC,$FB,$31,$F1,$9E,$CB,$ED,$87,$C5
	.dc.b	$BF,$D6,$DF,$FB,$57,$57,$9A,$F3,$FF,$9E,$71,$FF,$AE,$BC,$FB,$C8
	.dc.b	$7F,$A0,$5F,$0A,$B7,$C7,$1F,$F6,$93,$5F,$F7,$53,$67,$FD,$EC,$D8
	.dc.b	$FF,$7E,$35,$47,$57,$9C,$FF,$BF,$5D,$B5,$DC,$BD,$0F,$4F,$57,$FB
	.dc.b	$9A,$7F,$D9,$D9,$7F,$BC,$7A,$7C,$F5,$27,$CE,$F3,$C7,$E1,$B7,$0E
	.dc.b	$93,$A2,$1E,$C8,$D1,$E3,$7F,$DA,$9E,$89,$9E,$17,$B1,$78,$0F,$09
	.dc.b	$E8,$FF,$6B,$8C,$3F,$D6,$2F,$95,$2F,$93,$D8,$C9,$FE,$ED,$8A,$D2
	.dc.b	$9B,$33,$CA,$8C,$C6,$4F,$F0,$D3,$88,$D4,$74,$E9,$7A,$4F,$4F,$AC
	.dc.b	$61,$B1,$99,$E0,$D7,$C3,$EA,$75,$23,$E6,$B4,$74,$F2,$DE,$9D,$25
	.dc.b	$78,$E3,$2D,$8C,$BF,$13,$14,$EA,$30,$D7,$E6,$F1,$FC,$DE,$FF,$8E
	.dc.b	$D8,$F9,$6D,$7E,$7B,$5E,$3A,$A7,$8B,$63,$8B,$5E,$2D,$7E,$EF,$9E
	.dc.b	$F6,$1E,$B7,$D8,$8C,$36,$33,$3F,$7A,$72,$D8,$CB,$F8,$48,$CB,$63
	.dc.b	$2F,$81,$AF,$E3,$BD,$77,$B1,$F8,$46,$C7,$E2,$9A,$FE,$39,$AF,$38
	.dc.b	$F3,$11,$F2,$35,$F1,$35,$E0,$7E,$5B,$D8,$7B,$3F,$9F,$8F,$CD,$C6
	.dc.b	$32,$D4,$E6,$7F,$43,$39,$77,$EC,$B2,$E8,$D8,$CC,$60,$FD,$0A,$F0
	.dc.b	$69,$1F,$D2,$BF,$3F,$D6,$7B,$5F,$B2,$9F,$DF,$52,$73,$DD,$BA,$D5
	.dc.b	$78,$E7,$2F,$D5,$6B,$F8,$26,$BD,$1A,$F1,$8E,$6B,$D9,$7A,$38,$6D
	.dc.b	$FF,$00,$6C,$FF,$04,$6B,$DF,$DA,$F2,$E3,$E9,$6E,$26,$92,$FE,$13
	.dc.b	$6B,$D4,$FB,$B9,$3F,$8B,$79,$B7,$80,$D2,$5F,$E9,$1B,$7F,$A4,$6B
	.dc.b	$FC,$83,$9F,$F7,$3E,$47,$CA,$6D,$7F,$01,$8C,$74,$1A,$48,$D1,$2B
	.dc.b	$F9,$2F,$B6,$F7,$3E,$5B,$67,$F9,$3B,$67,$73,$F6,$BC,$27,$B8,$F7
	.dc.b	$70,$DC,$3F,$98,$CE,$A7,$27,$D8,$F6,$1E,$E5,$25,$7F,$36,$8C,$B6
	.dc.b	$33,$8F,$F3,$C9,$CB,$63,$C9,$ED,$BD,$CA,$4B,$F1,$C6,$5F,$E7,$5E
	.dc.b	$FB,$DF,$FE,$A9,$1F,$3E,$30,$FF,$59,$31,$1A,$C3,$FA,$E7,$0A,$B7
	.dc.b	$1F,$D5,$23,$13,$B8,$46,$2B,$A4,$72,$AE,$31,$D8,$CC,$70,$8E,$BD
	.dc.b	$6F,$53,$FB,$67,$33,$D0,$7C,$07,$F6,$DA,$D6,$1C,$73,$C6,$1C,$2F
	.dc.b	$79,$7F,$75,$BF,$F1,$17,$BB,$FA,$68,$FE,$DE,$1C,$61,$F0,$DA,$FF
	.dc.b	$85,$9B,$1B,$EE,$F8,$0D,$3D,$65,$9F,$EF,$2D,$3B,$CF,$ED,$AB,$D3
	.dc.b	$9F,$B7,$DD,$F4,$DF,$7C,$2C,$E7,$0F,$56,$5F,$E0,$6F,$2F,$FA,$96
	.dc.b	$F8,$C3,$FC,$25,$18,$7F,$8F,$4F,$F8,$57,$DA,$7B,$7F,$B6,$C3,$85
	.dc.b	$2F,$86,$9F,$32,$31,$E5,$FD,$C7,$99,$D9,$C4,$7D,$2C,$3A,$F7,$7A
	.dc.b	$AB,$8E,$7D,$B0,$EB,$9F,$A7,$1E,$2F,$97,$3F,$25,$F6,$38,$96,$B8
	.dc.b	$BE,$1B,$F8,$18,$F8,$33,$ED,$8F,$4A,$D1,$C5,$C4,$EE,$31,$5C,$08
	.dc.b	$FB,$A3,$59,$FC,$3C,$3C,$33,$E4,$8C,$5F,$4C,$7D,$31,$8C,$39,$B6
	.dc.b	$BA,$4E,$B2,$F8,$55,$A6,$CE,$B3,$F9,$58,$F8,$09,$8F,$78,$CB,$EF
	.dc.b	$9F,$25,$72,$C7,$EA,$63,$DE,$7E,$1C,$78,$C9,$3F,$46,$F9,$E3,$E3
	.dc.b	$AD,$CE,$73,$C4,$76,$32,$33,$F8,$F3,$E1,$24,$6A,$32,$F0,$39,$1C
	.dc.b	$DC,$B8,$CF,$DF,$7F,$7B,$2F,$54,$E7,$3F,$81,$5F,$84,$BE,$2F,$09
	.dc.b	$FB,$A7,$CB,$C7,$D3,$19,$CF,$E2,$E7,$F0,$2B,$71,$F1,$1F,$66,$3E
	.dc.b	$0A,$E1,$9F,$B3,$1E,$F9,$78,$89,$9F,$9F,$2E,$99,$73,$CB,$DD,$97
	.dc.b	$3B,$DC,$78,$31,$F9,$CD,$3E,$36,$5E,$DA,$E1,$8F,$AA,$FF,$05,$9D
	.dc.b	$8F,$B1,$9F,$91,$C6,$1F,$04,$E8,$7C,$97,$CD,$3A,$CB,$E0,$57,$1A
	.dc.b	$DE,$3E,$1C,$FD,$79,$FD,$C9,$F5,$C6,$1F,$3E,$5E,$47,$8B,$EE,$C3
	.dc.b	$E7,$E3,$DE,$FE,$56,$7E,$CB,$DE,$7E,$46,$9F,$0F,$3F,$26,$1F,$2F
	.dc.b	$3F,$26,$3C,$EB,$F1,$D8,$73,$AD,$65,$E3,$6E,$3F,$1A,$75,$97,$AF
	.dc.b	$1E,$78,$78,$2F,$E1,$D6,$A3,$DB,$7E,$59,$E7,$1E,$59,$D3,$F3,$4F
	.dc.b	$A1,$1A,$39,$E7,$F1,$E3,$51,$F1,$1F,$06,$1F,$12,$75,$87,$E4,$27
	.dc.b	$F2,$71,$8C,$7E,$65,$7E,$22,$FB,$3C,$72,$F8,$97,$F5,$9B,$87,$BA
	.dc.b	$B7,$3F,$23,$0E,$99,$7C,$56,$C7,$B5,$B3,$EA,$BE,$78,$FC,$DC,$7E
	.dc.b	$53,$E4,$BF,$3C,$EB,$0F,$91,$5A,$7B,$46,$67,$E3,$CE,$5F,$F6,$CD
	.dc.b	$7F,$51,$7F,$DC,$38,$FF,$5D,$6C,$65,$FE,$C9,$DA,$F6,$BF,$D9,$CD
	.dc.b	$8F,$F7,$1B,$5E,$6D,$78,$C7,$F5,$78,$E1,$97,$F4,$16,$E1,$FD,$AF
	.dc.b	$DB,$7B,$7F,$EF,$8A,$D5,$F1,$26,$1F,$D2,$EB,$6E,$27,$6F,$F5,$DC
	.dc.b	$3A,$BF,$CA,$E7,$6E,$6B,$86,$1E,$68,$E9,$87,$86,$31,$87,$42,$3E
	.dc.b	$78,$D4,$FF,$81,$B0,$EE,$FF,$80,$5B,$97,$F9,$B7,$0F,$ED,$B8,$F8
	.dc.b	$B1,$E7,$7C,$67,$C6,$F9,$F0,$EA,$F3,$8C,$B4,$F3,$98,$BE,$CF,$B4
	.dc.b	$C9,$6B,$BB,$C5,$FF,$07,$5F,$8F,$1F,$79,$27,$C8,$72,$9E,$35,$D9
	.dc.b	$F4,$EE,$BE,$55,$C2,$77,$97,$D2,$AD,$65,$EC,$C3,$96,$3D,$F0,$F2
	.dc.b	$3F,$35,$F6,$5F,$CA,$39,$63,$DA,$36,$77,$9F,$46,$1D,$A3,$8E,$5F
	.dc.b	$32,$31,$1F,$0E,$7C,$18,$78,$A3,$E8,$1C,$0D,$18,$29,$E0,$3C,$33
	.dc.b	$D3,$2F,$7D,$F7,$C3,$CD,$7D,$6F,$83,$DE,$F8,$61,$E2,$C7,$DD,$87
	.dc.b	$E1,$B2,$F8,$D1,$8C,$3C,$77,$C1,$F8,$04,$3C,$78,$F8,$27,$86,$5F
	.dc.b	$5A,$31,$87,$B6,$3F,$03,$87,$68,$C6,$3F,$47,$0F,$7E,$5E,$93,$D1
	.dc.b	$87,$28,$CE,$3F,$53,$0E,$51,$8C,$3D,$4F,$43,$CF,$97,$A5,$F8,$75
	.dc.b	$E6,$C7,$C1,$97,$CA,$9F,$7E,$3D,$E3,$07,$A2,$3A,$1F,$03,$1F,$0C
	.dc.b	$6A,$B9,$9F,$37,$1F,$A9,$3A,$8C,$57,$E7,$5B,$5F,$1E,$75,$3F,$37
	.dc.b	$2F,$83,$18,$8F,$4D,$FB,$9F,$80,$7A,$23,$53,$F2,$71,$F8,$D3,$E8
	.dc.b	$8E,$6D,$8C,$BC,$AF,$C4,$E6,$FB,$47,$B6,$3E,$5E,$3F,$17,$1F,$49
	.dc.b	$F0,$23,$D1,$7F,$35,$FC,$D5,$71,$7E,$41,$D4,$F8,$47,$0B,$F8,$EE
	.dc.b	$63,$E8,$3E,$5A,$FD,$7D,$F8,$63,$81,$C8,$F8,$A7,$C3,$39,$47,$95
	.dc.b	$CC,$FB,$9E,$87,$B3,$1F,$93,$87,$A7,$2F,$C4,$65,$F2,$F2,$F2,$DF
	.dc.b	$03,$D6,$7F,$AB,$34,$7A,$EF,$5F,$F9,$BF,$17,$FB,$87,$D8,$BC,$E7
	.dc.b	$F9,$DD,$F0,$AB,$7C,$70,$FF,$69,$B5,$FF,$69,$B6,$7F,$DC,$2D,$80
	.dc.b	$8E,$8F,$29,$FF,$73,$BA,$8C,$BD,$23,$2F,$FB,$EF,$C6,$F3,$FA,$B2
	.dc.b	$3F,$DA,$DC,$8E,$C5,$7F,$54,$8C,$3C,$9B,$3D,$CF,$2B,$87,$CC,$62
	.dc.b	$79,$BB,$3C,$45,$8E,$18,$7B,$A3,$27,$22,$B9,$87,$AA,$C3,$85,$66
	.dc.b	$30,$E0,$4F,$09,$C4,$62,$76,$EF,$77,$AB,$A8,$FF,$79,$36,$33,$1C
	.dc.b	$9A,$F1,$8D,$11,$D3,$E0,$F1,$1F,$02,$FA,$2C,$36,$BD,$13,$A9,$C9
	.dc.b	$E3,$F8,$3F,$24,$9A,$47,$DB,$9B,$CB,$7C,$26,$1B,$3F,$16,$75,$39
	.dc.b	$24,$66,$76,$FC,$E9,$DB,$99,$F9,$7C,$AF,$35,$EC,$E1,$B5,$F4,$E7
	.dc.b	$53,$92,$46,$67,$6F,$DC,$9D,$B9,$9F,$B3,$E8,$3D,$87,$AF,$86,$D7
	.dc.b	$E0,$67,$53,$92,$46,$67,$6F,$E2,$A7,$6E,$67,$F0,$5E,$C7,$E6,$A3
	.dc.b	$F1,$8D,$7F,$22,$D7,$F3,$4D,$53,$E7,$F8,$5F,$31,$23,$C4,$D7,$C5
	.dc.b	$68,$C4,$7E,$96,$79,$3F,$9A,$F6,$1E,$CF,$D1,$8F,$D0,$46,$34,$D4
	.dc.b	$E6,$7F,$4F,$39,$77,$EC,$F9,$9B,$7D,$63,$11,$C1,$AF,$EB,$7C,$4F
	.dc.b	$C9,$91,$FD,$43,$FB,$1F,$86,$5B,$D1,$FE,$0B,$97,$F0,$69,$CC,$63
	.dc.b	$B3,$F8,$FC,$BD,$8F,$E6,$11,$83,$F9,$1D,$70,$71,$8A,$30,$FE,$4E
	.dc.b	$37,$87,$85,$DD,$F5,$F0,$3C,$15,$C6,$B4,$F1,$3F,$84,$FC,$3B,$C9
	.dc.b	$FE,$29,$1F,$C8,$23,$18,$7E,$AD,$B1,$FA,$76,$BF,$C7,$71,$7C,$47
	.dc.b	$F2,$9C,$BE,$13,$6F,$A6,$1E,$4A,$DD,$DA,$D3,$EF,$6F,$43,$1F,$0E
	.dc.b	$3D,$2B,$DB,$CF,$FB,$5F,$89,$B2,$7E,$0F,$6B,$E0,$FF,$A0,$EB,$F3
	.dc.b	$2F,$83,$6F,$83,$6F,$8D,$70,$C3,$E4,$5F,$06,$FC,$57,$FB,$06,$FA
	.dc.b	$36,$3E,$76,$AF,$89,$95,$AF,$2F,$D3,$7B,$CF,$FA,$6E,$28,$C5,$7E
	.dc.b	$12,$F8,$E3,$E0,$C3,$DB,$0F,$47,$EC,$CB,$5F,$04,$61,$B7,$FC,$CA
	.dc.b	$B8,$3F,$C6,$7E,$2D,$F1,$9F,$31,$B1,$F9,$2A,$E3,$87,$C2,$9C,$BF
	.dc.b	$8B,$8F,$76,$FD,$F0,$F9,$73,$FD,$6A,$73,$7E,$1A,$D1,$1C,$12,$7F
	.dc.b	$93,$70,$FF,$84,$63,$EC,$36,$3C,$17,$DB,$6F,$93,$39,$7F,$BB,$E0
	.dc.b	$F8,$D7,$C7,$7C,$36,$E3,$FD,$51,$B8,$F9,$DA,$FF,$79,$AE,$55,$C4
	.dc.b	$95,$C5,$FA,$3C,$0F,$43,$F8,$CC,$FE,$3D,$B7,$F1,$F0,$EF,$8F,$E0
	.dc.b	$27,$2F,$A3,$CB,$7C,$7F,$D9,$6E,$1F,$DA,$9B,$87,$CD,$6B,$E7,$6B
	.dc.b	$FE,$2E,$CF,$FC,$1B,$5C,$C9,$5F,$D6,$7E,$2F,$F8,$9D,$7E,$79,$B8
	.dc.b	$78,$EB,$F5,$F1,$8A,$E9,$F2,$1F,$4E,$FE,$33,$5C,$FA,$7F,$91,$8C
	.dc.b	$47,$E5,$5B,$89,$A4,$9F,$97,$E4,$BE,$47,$F7,$ED,$C3,$F2,$95,$F7
	.dc.b	$DF,$F2,$5E,$FF,$E0,$4A,$FE,$03,$F2,$4F,$93,$FD,$13,$6B,$FC,$26
	.dc.b	$DC,$FF,$2C,$D7,$F0,$0D,$7E,$53,$5F,$D7,$37,$33,$9C,$1F,$28,$F9
	.dc.b	$4F,$E2,$ED,$C3,$F5,$95,$E1,$E0,$FE,$03,$39,$C7,$F9,$B7,$5F,$EF
	.dc.b	$5F,$DE,$B6,$FF,$C8,$4D,$8F,$F2,$3B,$5F,$F2,$83,$5F,$F3,$7B,$53
	.dc.b	$49,$F2,$AF,$87,$FE,$93,$08,$C5,$7F,$54,$8F,$06,$1F,$E6,$47,$55
	.dc.b	$C3,$E5,$BF,$3B,$5E,$06,$E1,$E0,$F1,$5F,$11,$BE,$AF,$E6,$FF,$11
	.dc.b	$F8,$42,$73,$7F,$19,$7F,$13,$C2,$F9,$7F,$E4,$B8,$A3,$1A,$FA,$37
	.dc.b	$7C,$87,$F4,$6B,$DD,$FB,$7B,$FF,$A2,$FC,$9B,$E5,$FF,$CF,$BE,$7F
	.dc.b	$F6,$3A,$FF,$4A,$F1,$FF,$30,$E4,$FE,$17,$7B,$E4,$FF,$4C,$DE,$EF
	.dc.b	$A4,$7F,$27,$9D,$42,$75,$7C,$EB,$F6,$95,$A9,$F0,$74,$23,$1E,$66
	.dc.b	$B2,$F2,$BF,$E8,$BC,$39,$63,$FC,$FE,$B6,$FF,$41,$F9,$9F,$E6,$38
	.dc.b	$FA,$F8,$5F,$34,$F9,$AE,$77,$BF,$1E,$D6,$A7,$3A,$B1,$E7,$A7,$EB
	.dc.b	$B1,$E5,$B7,$C2,$8F,$E1,$D3,$AC,$BE,$A5,$69,$F8,$53,$FC,$4F,$E6
	.dc.b	$FF,$D3,$37,$C6,$7F,$AC,$46,$1D,$98,$3C,$6E,$74,$F8,$ED,$AE,$9E
	.dc.b	$46,$AB,$6F,$BB,$99,$F3,$AF,$9E,$F6,$FC,$F3,$E3,$75,$39,$D3,$C9
	.dc.b	$BA,$B5,$1E,$7C,$39,$E1,$CB,$0E,$2D,$9E,$35,$C3,$2F,$5E,$1C,$EF
	.dc.b	$75,$A7,$E7,$FC,$FB,$E5,$3E,$9F,$37,$F5,$FE,$0F,$C2,$CE,$6F,$E4
	.dc.b	$72,$7E,$DF,$BB,$8E,$1C,$1B,$3A,$C3,$97,$0F,$E4,$3C,$9F,$BB,$D1
	.dc.b	$FE,$AA,$AF,$7B,$FE,$F3,$6C,$68,$8E,$8B,$B7,$D6,$3F,$84,$7D,$0F
	.dc.b	$E2,$3E,$87,$FA,$FE,$DF,$A0,$8C,$7B,$DF,$90,$AD,$DF,$CE,$BD,$E5
	.dc.b	$F9,$DB,$DC,$FE,$16,$B6,$7C,$AA,$DB,$FB,$69,$DB,$F3,$BE,$41,$F4
	.dc.b	$5F,$9A,$6D,$FD,$C7,$FD,$AC,$D8,$D1,$1D,$7C,$45,$AF,$B3,$F4,$7F
	.dc.b	$E2,$86,$FB,$1F,$A6,$87,$C9,$58,$97,$F8,$A9,$E1,$B7,$EF,$9D,$C6
	.dc.b	$67,$F2,$DF,$4B,$E6,$C3,$F0,$FF,$4B,$FC,$C1,$FD,$1F,$62,$75,$8F
	.dc.b	$F0,$8A,$D3,$FC,$66,$75,$A7,$EE,$A7,$3C,$1F,$D0,$AF,$8B,$FE,$05
	.dc.b	$8C,$E1,$FD,$12,$72,$FF,$92,$E3,$37,$D3,$2F,$D2,$C7,$0E,$A7,$C0
	.dc.b	$FF,$3C,$9F,$D4,$B6,$FF,$42,$BE,$9B,$F8,$6B,$9D,$3F,$90,$B7,$3F
	.dc.b	$F0,$C3,$6B,$FC,$B8,$F1,$BE,$BE,$77,$F1,$6F,$A8,$7C,$F7,$F1,$8F
	.dc.b	$A8,$FF,$39,$65,$FC,$B2,$73,$7F,$BC,$9C,$E5,$FE,$25,$9C,$E5,$FB
	.dc.b	$69,$CB,$FE,$6B,$8C,$DF,$F3,$69,$CC,$BE,$4F,$F8,$57,$8B,$F9,$76
	.dc.b	$C7,$F1,$AF,$98,$7C,$D7,$E0,$B2,$4C,$AD,$E3,$FD,$2D,$B3,$A9,$ED
	.dc.b	$E2,$31,$FE,$51,$8F,$4E,$6F,$D6,$61,$C2,$FA,$5E,$DF,$DB,$46,$6B
	.dc.b	$86,$BE,$7A,$D6,$5E,$16,$E5,$E1,$BE,$19,$FA,$71,$EB,$9F,$12,$3E
	.dc.b	$F8,$D3,$E8,$8C,$E1,$D5,$F4,$CE,$AB,$D1,$5B,$3C,$F8,$78,$96,$5E
	.dc.b	$87,$37,$D3,$1F,$35,$6E,$B5,$8F,$7B,$D9,$1F,$EF,$D1,$AA,$E3,$7C
	.dc.b	$EB,$DA,$7A,$F3,$F9,$B8,$78,$E3,$CD,$87,$53,$9E,$5E,$B9,$D6,$5E
	.dc.b	$C3,$E4,$DF,$2C,$3B,$61,$CB,$1F,$C1,$E5,$E3,$8F,$0E,$5D,$67,$CB
	.dc.b	$1E,$3C,$F8,$E3,$D7,$2F,$81,$9F,$87,$2C,$C6,$33,$F0,$61,$F0,$71
	.dc.b	$E7,$97,$A2,$72,$F7,$9F,$0C,$F9,$30,$E9,$1C,$4C,$9F,$51,$CC,$F1
	.dc.b	$8F,$61,$9A,$EB,$5E,$02,$46,$72,$F3,$E3,$F4,$73,$F8,$53,$C8,$99
	.dc.b	$FC,$AA,$DC,$78,$E7,$73,$F0,$EB,$F5,$13,$F2,$0B,$1F,$B4,$6E,$7F
	.dc.b	$A5,$73,$D1,$FE,$64,$78,$9F,$BD,$70,$EB,$97,$2E,$5B,$5F,$D0,$9F
	.dc.b	$E0,$53,$CF,$4F,$DE,$37,$3F,$BF,$F3,$4E,$BF,$DF,$56,$AF,$89,$31
	.dc.b	$FE,$15,$5B,$71,$3B,$7F,$7B,$87,$57,$F4,$F3,$B7,$35,$C3,$0F,$DE
	.dc.b	$56,$F4,$5E,$F0,$E8,$4A,$DD,$FF,$10,$9D,$46,$30,$EA,$4A,$FD,$6C
	.dc.b	$EA,$3F,$8F,$61,$E5,$7E,$41,$BA,$F8,$78,$FC,$27,$A5,$F8,$67,$8E
	.dc.b	$1E,$1A,$D4,$60,$E4,$59,$F0,$C6,$A3,$BC,$73,$6D,$78,$1F,$88,$71
	.dc.b	$7C,$26,$5E,$A7,$03,$47,$CE,$9E,$D9,$7E,$06,$36,$DA,$F1,$C7,$63
	.dc.b	$91,$CF,$2F,$A3,$5C,$B1,$F8,$D7,$C7,$0F,$1D,$F7,$39,$46,$5F,$1D
	.dc.b	$F2,$73,$5C,$8F,$BD,$97,$D3,$8F,$0D,$71,$7C,$51,$98,$C1,$27,$8E
	.dc.b	$1E,$5C,$BA,$9F,$4E,$E9,$F4,$B0,$F1,$3F,$0A,$BC,$53,$B8,$C3,$F4
	.dc.b	$27,$79,$7D,$5C,$3C,$13,$B8,$FF,$02,$FD,$4F,$E0,$39,$FF,$C1,$FE
	.dc.b	$8D,$E7,$36,$66,$5F,$E1,$28,$D1,$87,$FC,$95,$F1,$BC,$FC,$03,$4C
	.dc.b	$7D,$0D,$9F,$F2,$B4,$57,$FC,$A5,$1F,$92,$CD,$F5,$5F,$9F,$CD,$DB
	.dc.b	$FC,$4B,$5F,$F1,$B6,$1B,$AF,$D2,$5F,$0E,$1F,$65,$EF,$8F,$FC,$69
	.dc.b	$7B,$98,$FF,$8C,$0D,$98,$3F,$D1,$BF,$57,$FC,$6F,$1F,$F6,$5B,$6B
	.dc.b	$FC,$1F,$92,$31,$48,$C3,$F3,$63,$31,$BE,$6C,$19,$9F,$A5,$E3,$7F
	.dc.b	$AA,$DF,$F4,$4F,$9B,$83,$3F,$35,$63,$24,$3F,$D4,$72,$8F,$D4,$FC
	.dc.b	$17,$D3,$FA,$C9,$5F,$E2,$B8,$CB,$FE,$87,$8C,$BF,$E1,$C8,$CB,$FE
	.dc.b	$85,$8C,$E9,$FE,$2B,$9F,$C4,$79,$EE,$2A,$4B,$FF,$5F,$C7,$E0,$9B
	.dc.b	$19,$7F,$D7,$31,$C5,$DF,$1F,$E4,$FD,$0F,$F4,$0C,$7F,$AB,$FE,$73
	.dc.b	$F8,$AF,$A2,$F3,$D1,$BE,$E9,$5B,$21,$FE,$C6,$BE,$6F,$FB,$23,$D3
	.dc.b	$76,$7E,$85,$B1,$FE,$D6,$6B,$FE,$86,$FA,$D7,$CF,$FE,$55,$B1,$BA
	.dc.b	$E4,$66,$64,$64,$87,$37,$F6,$1E,$BB,$EB,$A9,$27,$FD,$CF,$0E,$1C
	.dc.b	$47,$FB,$82,$38,$3A,$9E,$73,$FB,$BF,$80,$7C,$47,$C8,$F6,$BE,$75
	.dc.b	$B9,$5D,$BF,$E6,$8E,$35,$6A,$73,$5F,$9B,$C3,$9B,$FE,$FA,$6B,$CE
	.dc.b	$F9,$57,$16,$F3,$78,$F1,$FC,$EF,$D5,$CB,$FF,$72,$BF,$CF,$BE,$AF
	.dc.b	$FC,$E9,$F0,$0F,$33,$FC,$CD,$18,$7F,$5F,$DA,$AE,$57,$FE,$72,$AD
	.dc.b	$3F,$E8,$16,$F8,$0A,$E1,$3E,$03,$FD,$19,$3C,$37,$FF,$45,$FA,$37
	.dc.b	$CF,$FF,$50,$D7,$FA,$33,$3F,$22,$AD,$B9,$CF,$FD,$23,$E3,$3E,$1B
	.dc.b	$E2,$DF,$0D,$1F,$5F,$FB,$5C,$3F,$C4,$71,$8F,$0F,$FC,$42,$D8,$CC
	.dc.b	$ED,$FF,$18,$E0,$F8,$7F,$CA,$E1,$E4,$8C,$75,$FC,$26,$D7,$94,$2F
	.dc.b	$49,$F6,$1B,$D3,$FD,$FB,$5B,$AF,$57,$B3,$93,$A7,$D8,$FF,$90,$B2
	.dc.b	$F0,$3F,$E3,$8A,$EA,$DA,$34,$91,$FE,$3E,$8C,$B6,$32,$FF,$91,$B8
	.dc.b	$1F,$14,$F9,$DF,$AC,$D9,$F6,$4E,$A7,$EF,$F4,$3D,$EA,$7C,$C7,$15
	.dc.b	$F4,$1A,$FF,$97,$38,$BF,$4F,$3B,$24,$6D,$AF,$F9,$8A,$DF,$02,$FA
	.dc.b	$FF,$C5,$36,$7C,$AD,$5A,$FE,$0F,$EC,$BE,$E3,$79,$BF,$03,$1F,$7E
	.dc.b	$7F,$DA,$D9,$23,$11,$FD,$41,$AF,$F4,$F6,$C6,$E7,$99,$27,$E6,$7C
	.dc.b	$8F,$83,$A3,$F8,$03,$63,$F3,$8D,$A7,$C9,$BE,$8B,$F5,$6D,$AF,$CC
	.dc.b	$B5,$F9,$4D,$7D,$2D,$7F,$B3,$B5,$FE,$CC,$D9,$EF,$3F,$A2,$8D,$6A
	.dc.b	$F9,$5F,$E2,$77,$F2,$B1,$FE,$31,$18,$D3,$F6,$B5,$BF,$0B,$F8,$0F
	.dc.b	$CB,$3E,$CE,$9F,$AE,$71,$8F,$EC,$9A,$FE,$0D,$AF,$D2,$6B,$FE,$00
	.dc.b	$6B,$FD,$FD,$BA,$1C,$DB,$D6,$FE,$AF,$5F,$98,$6B,$EB,$3F,$86,$38
	.dc.b	$C7,$F8,$47,$CE,$FF,$8C,$A7,$F7,$58,$23,$11,$F9,$96,$BF,$88,$6B
	.dc.b	$FE,$16,$6B,$FE,$15,$6C,$73,$F7,$7F,$88,$C6,$0D,$E0,$F9,$97,$DA
	.dc.b	$7F,$27,$6E,$1F,$B2,$6B,$EB,$C3,$CD,$E9,$FF,$32,$C0,$F6,$24,$FF
	.dc.b	$22,$8C,$BF,$C0,$7E,$0B,$F9,$8B,$F1,$71,$46,$33,$FB,$ED,$E1,$53
	.dc.b	$E7,$5F,$6B,$FD,$A6,$3F,$83,$D6,$DF,$E7,$4D,$C3,$AF,$93,$FD,$2A
	.dc.b	$F8,$FA,$3F,$D1,$72,$FD,$FE,$1F,$C4,$1A,$F9,$5A,$7F,$3F,$FB,$67
	.dc.b	$CA,$F1,$ED,$95,$C5,$FE,$AB,$3A,$8F,$E6,$D8,$3E,$DD,$F3,$5E,$36
	.dc.b	$CE,$9A,$FE,$24,$FD,$B7,$DC,$EE,$64,$6D,$FE,$CA,$E8,$FE,$88,$53
	.dc.b	$FD,$EB,$DA,$9F,$E7,$8D,$8F,$F7,$AB,$5F,$E8,$5F,$2D,$FE,$E3,$D3
	.dc.b	$FB,$28,$71,$C5,$FD,$A5,$B3,$F5,$BE,$9F,$F5,$F5,$C2,$72,$F1,$F8
	.dc.b	$6F,$EF,$6D,$AF,$3D,$70,$BF,$BF,$1F,$D8,$FE,$57,$F9,$6B,$83,$F0
	.dc.b	$7E,$8A,$33,$19,$8C,$5E,$E3,$73,$A7,$AC,$FF,$A6,$BE,$9F,$51,$FD
	.dc.b	$C7,$E1,$11,$89,$D4,$F3,$6B,$C3,$DC,$FE,$CC,$FF,$8C,$4F,$ED,$0E
	.dc.b	$0F,$16,$0F,$A6,$7C,$6F,$F8,$42,$BF,$C4,$B1,$8C,$F6,$D8,$FE,$E8
	.dc.b	$D7,$73,$C5,$FF,$0D,$46,$5B,$19,$7F,$AA,$C6,$5F,$D5,$FD,$CB,$EC
	.dc.b	$B5,$08,$C5,$FC,$46,$C7,$CB,$6B,$F5,$BC,$4F,$E3,$1E,$3F,$76,$D7
	.dc.b	$D1,$6C,$FD,$96,$BC,$7C,$0F,$F2,$BC,$ED,$B1,$B6,$BB,$AF,$67,$84
	.dc.b	$FA,$7F,$F3,$CD,$FF,$8C,$1B,$5F,$E0,$16,$BF,$DC,$1A,$FF,$AF,$9A
	.dc.b	$FF,$AF,$5B,$3D,$7E,$13,$BC,$6B,$AF,$FD,$D3,$E3,$FC,$D6,$E1,$F8
	.dc.b	$F6,$E1,$F9,$76,$BF,$9C,$6B,$FC,$71,$AF,$F1,$D9,$FF,$7A,$CF,$36
	.dc.b	$C7,$3F,$6F,$F7,$53,$DD,$B1,$9E,$BF,$24,$62,$3F,$C1,$F5,$FE,$45
	.dc.b	$AF,$F2,$AB,$63,$FC,$4C,$D7,$7E,$D7,$B7,$E6,$79,$47,$9D,$AF,$F3
	.dc.b	$36,$E2,$69,$31,$E4,$4C,$3C,$3F,$3D,$F6,$B1,$34,$91,$F2,$FE,$97
	.dc.b	$FD,$76,$DF,$7B,$E8,$46,$32,$F9,$CD,$7E,$73,$5F,$9C,$D7,$FA,$53
	.dc.b	$72,$39,$DF,$FB,$F7,$83,$F8,$EF,$3B,$F9,$BE,$3F,$E0,$B8,$C1,$FA
	.dc.b	$DF,$6F,$FA,$A7,$A9,$71,$FF,$7B,$57,$F1,$19,$CB,$63,$39,$7F,$A0
	.dc.b	$1A,$FF,$20,$9D,$B9,$AE,$87,$E8,$3D,$3F,$3D,$FF,$8D,$B3,$C1,$1F
	.dc.b	$07,$DD,$7D,$6C,$0D,$27,$07,$A5,$B3,$C2,$30,$7E,$BB,$D9,$7C,$F5
	.dc.b	$24,$FF,$9B,$DF,$85,$F7,$7F,$E0,$28,$F8,$0D,$D3,$FD,$3A,$D7,$DC
	.dc.b	$D7,$FD,$5A,$D7,$FD,$36,$DC,$4D,$35,$C0,$FA,$57,$D3,$FF,$04,$6E
	.dc.b	$1F,$EB,$26,$E1,$FE,$83,$6B,$F3,$FE,$F1,$F0,$5F,$6D,$B5,$F6,$5B
	.dc.b	$3F,$EB,$66,$BC,$67,$81,$27,$83,$F8,$5E,$9F,$F0,$CC,$7E,$0D,$B3
	.dc.b	$FE,$C0,$6B,$F8,$16,$BF,$ED,$0F,$19,$F3,$FF,$8F,$6D,$7E,$35,$B3
	.dc.b	$FE,$CA,$6B,$F8,$BF,$84,$FC,$CB,$FE,$CE,$6B,$9E,$7F,$CD,$46,$23
	.dc.b	$FD,$BF,$F4,$EF,$A7,$FC,$F3,$70,$FC,$E3,$67,$FD,$B2,$D7,$FB,$B7
	.dc.b	$CA,$FF,$18,$6F,$77,$F4,$E8,$C4,$7F,$7E,$6B,$FE,$3E,$6A,$69,$23
	.dc.b	$91,$FE,$14,$AF,$F2,$2F,$0B,$E2,$FF,$C6,$B1,$FE,$48,$F3,$3F,$D1
	.dc.b	$B7,$FC,$AB,$CC,$FF,$7C,$56,$B1,$53,$C6,$E7,$F2,$3F,$C5,$31,$FA
	.dc.b	$B9,$23,$17,$FD,$E6,$31,$E1,$AB,$75,$A7,$FD,$7E,$DA,$EB,$3B,$79
	.dc.b	$46,$55,$74,$AF,$0F,$C9,$FF,$A3,$DF,$66,$1F,$02,$31,$5E,$A8,$C3
	.dc.b	$D7,$1F,$63,$D2,$77,$18,$B4,$62,$BA,$79,$5C,$5E,$78,$78,$30,$E5
	.dc.b	$39,$5D,$7F,$E4,$5F,$98,$F6,$57,$F1,$BF,$55,$18,$AF,$F3,$DD,$6E
	.dc.b	$BF,$42,$D8,$FF,$12,$36,$39,$4F,$19,$EF,$5D,$27,$8B,$63,$8B,$73
	.dc.b	$F7,$7C,$03,$E7,$FD,$18,$F7,$9C,$C7,$AF,$57,$CC,$3E,$97,$C1,$18
	.dc.b	$6C,$67,$3F,$C2,$B7,$3F,$C2,$36,$3F,$DF,$ED,$8F,$CB,$B5,$FB,$ED
	.dc.b	$AF,$CD,$7B,$BE,$3E,$33,$6F,$0C,$7E,$A4,$F0,$6C,$FE,$4D,$B3,$F5
	.dc.b	$EF,$F5,$9F,$4B,$F2,$DB,$3F,$C4,$39,$93,$96,$C6,$63,$04,$70,$49
	.dc.b	$FC,$91,$F9,$AF,$9F,$FF,$3B,$47,$E2,$63,$B9,$AC,$53,$A8,$C7,$B9
	.dc.b	$FC,$C3,$6F,$CB,$37,$93,$F7,$EB,$E3,$7E,$C3,$9A,$FD,$4B,$70,$34
	.dc.b	$E1,$8F,$C8,$75,$E8,$FF,$88,$F1,$FD,$23,$71,$FE,$1A,$BE,$6B,$F0
	.dc.b	$CE,$71,$FD,$BB,$71,$FF,$11,$78,$7A,$AD,$BF,$90,$FA,$5F,$E9,$79
	.dc.b	$FF,$7F,$6E,$1F,$A4,$6B,$FA,$C6,$F2,$9A,$4F,$17,$F4,$91,$96,$C7
	.dc.b	$C1,$F9,$FF,$E5,$57,$FB,$9A,$D3,$F8,$88,$CD,$FD,$3F,$73,$FA,$5E
	.dc.b	$DF,$C1,$5B,$B7,$F2,$E5,$F4,$BF,$A7,$73,$5F,$C6,$1B,$B7,$FA,$7B
	.dc.b	$C8,$D5,$6D,$FD,$77,$D2,$BE,$23,$F6,$1F,$4B,$FC,$2B,$5F,$CD,$CE
	.dc.b	$5F,$CE,$C6,$72,$FC,$44,$E5,$FC,$54,$67,$4F,$E3,$93,$97,$F8,$F4
	.dc.b	$67,$4F,$F6,$6F,$D2,$7F,$03,$F0,$BF,$28,$DC,$3F,$33,$F1,$8F,$99
	.dc.b	$FB,$D0,$9D,$63,$FC,$B2,$31,$E0,$B0,$E6,$D8,$E7,$7C,$AF,$8E,$3E
	.dc.b	$2A,$D6,$1C,$E7,$36,$C7,$C5,$CD,$FB,$99,$CC,$F8,$32,$EE,$7F,$35
	.dc.b	$AD,$EB,$89,$FD,$99,$B2,$D7,$69,$DC,$63,$1F,$1E,$5E,$5C,$39,$65
	.dc.b	$FD,$2A,$F6,$F9,$EB,$76,$3B,$46,$6B,$C9,$9F,$C1,$AD,$D6,$B0,$EC
	.dc.b	$FC,$6C,$FD,$D3,$AC,$BD,$E7,$9E,$F9,$61,$E4,$C3,$8E,$3E,$59,$CE
	.dc.b	$3E,$BA,$D3,$F2,$E3,$33,$E5,$8D,$1E,$1C,$F6,$F9,$4F,$66,$1E,$1C
	.dc.b	$BC,$B8,$6A,$31,$57,$2F,$41,$F1,$33,$F8,$F9,$F9,$89,$8F,$49,$D6
	.dc.b	$7E,$49,$CC,$F8,$6B,$8B,$BC,$7B,$1F,$33,$2F,$1C,$7A,$B3,$F6,$1F
	.dc.b	$0E,$32,$E2,$FC,$37,$D2,$BD,$C4,$CF,$D6,$F3,$9F,$1C,$62,$B9,$4F
	.dc.b	$8C,$D1,$F3,$6F,$BD,$70,$CB,$DB,$18,$9E,$07,$E1,$CA,$7E,$08,$C1
	.dc.b	$F9,$72,$DF,$F3,$1E,$4C,$C7,$D4,$E7,$FF,$08,$B8,$9D,$36,$32,$F2
	.dc.b	$77,$8F,$EB,$DB,$8F,$F0,$46,$CF,$EC,$7E,$97,$FA,$4B,$FC,$1A,$32
	.dc.b	$FF,$03,$6D,$F8,$2F,$F4,$D5,$C2,$BF,$5F,$5A,$26,$1F,$B7,$9D,$57
	.dc.b	$8E,$3B,$12,$BF,$89,$CE,$B0,$F3,$4E,$6B,$85,$6B,$AB,$E6,$46,$09
	.dc.b	$5C,$27,$47,$17,$11,$A8,$F8,$13,$CA,$72,$48,$FA,$15,$E9,$9D,$46
	.dc.b	$0F,$7C,$75,$79,$94,$F9,$0E,$74,$61,$ED,$9E,$58,$78,$DD,$9E,$83
	.dc.b	$A9,$1F,$44,$F2,$BF,$1C,$F0,$9E,$59,$7D,$5C,$7F,$8B,$5E,$EB,$C9
	.dc.b	$39,$7C,$33,$AC,$3C,$91,$CE,$3E,$13,$4F,$69,$E3,$73,$3C,$EB,$B4
	.dc.b	$E4,$EA,$47,$37,$E6,$CB,$F0,$B3,$EF,$8C,$65,$F1,$F1,$F1,$E1,$E7
	.dc.b	$AF,$11,$E3,$C3,$9E,$3E,$2C,$79,$E3,$F9,$4C,$BE,$2E,$1C,$49,$3F
	.dc.b	$A0,$BE,$38,$79,$8E,$B3,$EB,$C7,$C3,$5D,$31,$FC,$25,$F0,$E5,$73
	.dc.b	$3D,$DF,$C2,$C6,$3C,$BB,$26,$CC,$CF,$FC,$B7,$39,$6C,$E5,$FF,$24
	.dc.b	$77,$3E,$1B,$AB,$63,$FC,$C6,$D7,$7F,$3B,$A9,$91,$A7,$8C,$7E,$6B
	.dc.b	$85,$F4,$BF,$93,$CD,$DB,$FC,$53,$8F,$FD,$50,$DD,$FF,$D3,$75,$F9
	.dc.b	$9B,$E3,$C7,$FC,$77,$A3,$F8,$F4,$63,$4F,$F4,$FD,$7F,$05,$9C,$B6
	.dc.b	$39,$D7,$16,$9C,$1C,$3F,$EB,$08,$E0,$FF,$B2,$3E,$97,$FC,$20,$FF
	.dc.b	$9B,$5B,$97,$FA,$FD,$B8,$FF,$BA,$1F,$54,$F1,$E8,$FA,$F1,$A8,$C1
	.dc.b	$FE,$8E,$77,$A7,$FB,$BF,$CB,$FB,$BE,$3F,$FA,$FA,$BF,$D2,$31,$8F
	.dc.b	$7F,$2C,$8C,$9F,$E9,$68,$FC,$9F,$C8,$7D,$AF,$A5,$93,$FE,$34,$8C
	.dc.b	$BF,$E0,$DD,$FF,$A1,$C7,$C2,$8F,$F1,$A3,$7A,$CD,$24,$71,$3F,$C8
	.dc.b	$2E,$1F,$F5,$BF,$1F,$E5,$BC,$B3,$4C,$72,$D9,$FF,$57,$CE,$9B,$3A
	.dc.b	$7F,$D6,$7E,$6F,$FA,$1D,$F0,$1F,$EF,$C7,$0F,$FC,$01,$31,$D9,$F7
	.dc.b	$37,$4E,$B3,$FE,$CD,$18,$9D,$4E,$6B,$FC,$61,$39,$7B,$B5,$FF,$75
	.dc.b	$FD,$23,$E1,$FF,$67,$E8,$FE,$D6,$7F,$D9,$2D,$7F,$D7,$3D,$2A,$E1
	.dc.b	$1F,$E8,$96,$BD,$2B,$F2,$73,$B6,$CE,$B2,$FD,$0B,$6B,$FE,$00,$B5
	.dc.b	$57,$17,$E3,$B5,$F8,$D8,$3D,$27,$D2,$7F,$0C,$6C,$FF,$B8,$9A,$FF
	.dc.b	$B3,$5F,$C1,$B9,$8E,$04,$78,$70,$7F,$16,$FA,$47,$C0,$FD,$36,$CF
	.dc.b	$CC,$6B,$E4,$6E,$7F,$59,$B1,$F5,$1A,$B1,$46,$1B,$5F,$99,$FA,$2F
	.dc.b	$F1,$27,$B0,$C3,$EE,$F4,$BE,$8F,$F5,$E7,$F3,$AA,$DC,$FF,$A7,$1F
	.dc.b	$F1,$2E,$5F,$E5,$86,$BF,$E9,$96,$F3,$9A,$48,$E3,$19,$6C,$74,$DD
	.dc.b	$EE,$7C,$29,$FE,$3A,$D7,$FD,$05,$1A,$CF,$84,$E6,$7E,$FF,$BD,$F8
	.dc.b	$F7,$F9,$2B,$5E,$2D,$77,$E3,$F8,$7C,$A7,$D1,$D2,$4F,$F9,$13,$14
	.dc.b	$EA,$31,$F4,$7F,$0A,$4D,$24,$7B,$7E,$F5,$F0,$34,$91,$D6,$32,$D8
	.dc.b	$CB,$FE,$58,$A4,$EA,$30,$D7,$F7,$7A,$3D,$9F,$AF,$C9,$F9,$06,$C6
	.dc.b	$67,$FC,$BF,$39,$7D,$DE,$DF,$FB,$A5,$BE,$AF,$D5,$8C,$47,$8D,$AF
	.dc.b	$F5,$26,$BF,$59,$AF,$EC,$9B,$89,$CD,$3E,$85,$E2,$FA,$1B,$7E,$66
	.dc.b	$BF,$E7,$A9,$E1,$3F,$89,$FA,$17,$CC,$52,$57,$D9,$8C,$B6,$32,$FC
	.dc.b	$D8,$CB,$63,$E4,$FC,$6E,$23,$F9,$3B,$5F,$E4,$AD,$7F,$64,$DF,$8E
	.dc.b	$7D,$9F,$F6,$89,$FF,$4A,$46,$3C,$DF,$D5,$B6,$3F,$D2,$2D,$7F,$D2
	.dc.b	$ED,$78,$CF,$37,$FD,$C9,$19,$6C,$65,$FE,$09,$F2,$0F,$A1,$FC,$1B
	.dc.b	$63,$F0,$2D,$7E,$63,$5F,$3F,$DF,$7F,$34,$6F,$8F,$E8,$8C,$47,$89
	.dc.b	$AB,$04,$61,$B3,$C4,$93,$F9,$2F,$95,$FF,$1A,$3F,$75,$AF,$F5,$66
	.dc.b	$BF,$D5,$1A,$F7,$F5,$7F,$B0,$3F,$39,$AF,$C6,$6B,$BF,$07,$E8,$C6
	.dc.b	$2C,$D2,$47,$EA,$AB,$FB,$4D,$79,$FC,$BC,$11,$FB,$3E,$C3,$EF,$FE
	.dc.b	$EB,$63,$31,$83,$F0,$3F,$43,$FD,$2A,$3E,$FB,$5F,$C3,$B5,$FB,$ED
	.dc.b	$7E,$73,$72,$34,$9E,$57,$F0,$2F,$9F,$7D,$05,$25,$FF,$02,$8C,$B6
	.dc.b	$32,$FE,$BE,$32,$D8,$FE,$83,$CA,$F9,$5A,$49,$FE,$8F,$19,$6C,$65
	.dc.b	$FF,$77,$C6,$5B,$1F,$D3,$3F,$00,$F9,$F1,$FE,$F4,$6B,$FE,$FA,$6B
	.dc.b	$FE,$FF,$6A,$1F,$41,$F9,$B7,$06,$39,$7F,$AC,$C6,$63,$83,$5D,$1F
	.dc.b	$D8,$1C,$12,$3F,$B0,$F9,$0F,$A0,$79,$FF,$D9,$A3,$0D,$8C,$DF,$88
	.dc.b	$F3,$4E,$9C,$07,$FD,$25,$F4,$1C,$FD,$93,$49,$3F,$DC,$BD,$EF,$EE
	.dc.b	$69,$A4,$7C,$18,$3E,$83,$FD,$57,$D1,$E0,$8C,$7C,$1F,$EA,$1B,$19
	.dc.b	$9D,$D7,$26,$BF,$E0,$C6,$BF,$46,$78,$BB,$AF,$D8,$FC,$FB,$E8,$29
	.dc.b	$30,$FF,$06,$C6,$5F,$87,$E0,$BB,$1F,$05,$FB,$C8,$C3,$63,$33,$C1
	.dc.b	$B8,$FC,$FA,$DC,$E6,$B8,$11,$FE,$C7,$F4,$1F,$E5,$B6,$F8,$1F,$C2
	.dc.b	$BC,$74,$E7,$0F,$EA,$4D,$7F,$CD,$6D,$7F,$CD,$2D,$7F,$C6,$CD,$8E
	.dc.b	$73,$FC,$5B,$B7,$FD,$B1,$DF,$4F,$F6,$D6,$1F,$C3,$1C,$11,$C1,$30
	.dc.b	$FE,$8F,$1C,$1F,$E9,$3F,$41,$FB,$59,$FE,$9A,$D8,$CF,$57,$86,$31
	.dc.b	$1F,$E8,$4A,$FF,$0F,$D7,$F8,$19,$B1,$FD,$C5,$AE,$FD,$2F,$57,$C9
	.dc.b	$72,$8F,$EA,$AD,$7F,$A2,$37,$13,$49,$8F,$22,$61,$E1,$F9,$7F,$A5
	.dc.b	$89,$A4,$8F,$ED,$3F,$40,$F5,$FE,$63,$63,$31,$82,$38,$26,$5F,$DC
	.dc.b	$27,$83,$FD,$93,$D5,$FC,$17,$95,$FE,$3A,$AF,$ED,$31,$83,$FA,$7F
	.dc.b	$A3,$FD,$87,$CF,$BB,$7F,$87,$2B,$FC,$94,$DC,$3F,$C7,$8D,$7F,$DF
	.dc.b	$6D,$7F,$CB,$ED,$7F,$D3,$6D,$AE,$D3,$FE,$3A,$F3,$99,$7E,$5D,$B1
	.dc.b	$98,$C1,$E6,$FA,$0F,$A9,$46,$93,$7D,$12,$BF,$CC,$51,$97,$FC,$93
	.dc.b	$EA,$3E,$83,$F4,$CD,$8C,$C6,$0F,$8B,$F4,$1F,$ED,$28,$FF,$36,$37
	.dc.b	$4F,$F2,$F3,$5F,$E6,$F8,$A3,$11,$FE,$82,$6B,$FE,$BE,$6C,$6F,$CF
	.dc.b	$EE,$7E,$E7,$E7,$DF,$41,$49,$87,$F9,$8A,$32,$FF,$9F,$FE,$81,$F3
	.dc.b	$5F,$71,$B1,$98,$C1,$1C,$13,$2F,$F3,$C4,$F0,$7F,$CF,$5D,$3F,$EF
	.dc.b	$87,$FD,$36,$D7,$FD,$AE,$DC,$BF,$D1,$F1,$F8,$CF,$25,$F3,$14,$98
	.dc.b	$7F,$A5,$63,$2F,$FA,$5F,$E0,$3F,$2D,$5F,$E8,$96,$BF,$EA,$CE,$7F
	.dc.b	$CB,$C6,$23,$FD,$5B,$F3,$EF,$A0,$FC,$DB,$6F,$FD,$44,$D9,$FF,$4A
	.dc.b	$35,$FE,$B9,$EE,$FF,$11,$6F,$77,$F8,$C2,$31,$1F,$DC,$9A,$FF,$77
	.dc.b	$6B,$FE,$85,$8E,$04,$AF,$F6,$8C,$FF,$1D,$F2,$BF,$DB,$55,$FE,$E1
	.dc.b	$F8,$6F,$F0,$ED,$7F,$80,$7C,$DF,$E8,$F3,$D7,$0D,$DF,$0F,$15,$87
	.dc.b	$0A,$FF,$7E,$78,$5D,$31,$FF,$03,$D3,$E2,$7F,$88,$47,$F8,$34,$CB
	.dc.b	$8A,$F1,$9F,$EF,$37,$11,$AA,$E9,$3F,$E1,$5A,$E9,$18,$8E,$0D,$8C
	.dc.b	$CF,$79,$E1,$F1,$6F,$2B,$E0,$E9,$EE,$B4,$E6,$FE,$17,$95,$C3,$0D
	.dc.b	$BC,$B0,$B0,$BE,$31,$86,$D7,$29,$D5,$F8,$F0,$F0,$3D,$E3,$FC,$21
	.dc.b	$87,$FB,$DB,$E8,$1F,$23,$ED,$E7,$46,$31,$F6,$5A,$73,$5B,$AD,$4F
	.dc.b	$CC,$9C,$D7,$06,$C6,$67,$73,$FC,$21,$B1,$E3,$AF,$67,$6B,$E6,$3E
	.dc.b	$0E,$1C,$B3,$F9,$53,$9C,$7E,$06,$8F,$91,$7D,$07,$D5,$6C,$FA,$A7
	.dc.b	$53,$92,$46,$67,$79,$FE,$1A,$B7,$39,$AF,$C0,$71,$3E,$83,$E9,$36
	.dc.b	$7C,$69,$CC,$47,$2F,$07,$17,$F5,$0F,$D7,$7D,$07,$E3,$DF,$F7,$93
	.dc.b	$79,$CD,$24,$FF,$17,$8C,$B6,$33,$3F,$EB,$09,$CB,$DF,$17,$CC,$FF
	.dc.b	$99,$6F,$F1,$53,$E0,$7E,$6C,$66,$38,$7C,$27,$F2,$7D,$BF,$38,$DE
	.dc.b	$3F,$D4,$2F,$8A,$FC,$B3,$9A,$FC,$D3,$71,$FF,$5D,$F8,$7A,$AD,$BF
	.dc.b	$9D,$F4,$7F,$C3,$78,$FE,$A5,$B8,$FE,$D5,$7C,$A6,$FC,$49,$AF,$E1
	.dc.b	$27,$7A,$7E,$A5,$DC,$66,$7F,$1B,$F4,$1F,$D0,$63,$FA,$DB,$6B,$FD
	.dc.b	$2D,$CC,$8C,$47,$E8,$DB,$19,$CF,$F5,$51,$82,$38,$3F,$71,$F3,$4E
	.dc.b	$1F,$0C,$61,$FD,$E7,$B7,$FD,$1B,$0F,$AC,$DD,$BF,$85,$2F,$A0,$DF
	.dc.b	$91,$34,$FD,$1C,$EF,$83,$F8,$D3,$B8,$CC,$FF,$09,$FA,$07,$C7,$FF
	.dc.b	$0B,$FA,$0F,$DF,$57,$EB,$27,$2F,$EB,$A3,$39,$FE,$26,$72,$FE,$2E
	.dc.b	$33,$9F,$F1,$F9,$CB,$FC,$86,$33,$97,$FA,$FB,$E5,$7F,$7D,$E1,$FE
	.dc.b	$41,$B8,$7E,$8F,$E7,$9F,$41,$F6,$A1,$3A,$C7,$FA,$4C,$63,$C2,$FE
	.dc.b	$60,$FE,$AC,$C6,$1F,$C9,$70,$E7,$CD,$FC,$B2,$72,$6F,$0E,$99,$79
	.dc.b	$70,$E7,$8F,$F5,$6C,$38,$EB,$E7,$9C,$D6,$F2,$C1,$DF,$3F,$46,$3D
	.dc.b	$32,$FF,$70,$35,$F0,$E5,$D6,$77,$87,$49,$DB,$5D,$E1,$FE,$E9,$CB
	.dc.b	$C7,$1E,$99,$D6,$1D,$4F,$04,$EB,$0F,$8F,$19,$9D,$9D,$E7,$57,$EB
	.dc.b	$BE,$98,$FB,$1F,$A9,$8F,$9A,$31,$8F,$4C,$39,$4E,$6B,$89,$27,$DD
	.dc.b	$9F,$D1,$CB,$C5,$87,$2C,$7D,$98,$71,$BE,$93,$9C,$19,$79,$9F,$8B
	.dc.b	$19,$7E,$29,$AA,$E5,$7E,$BB,$E7,$3E,$AC,$7C,$59,$7C,$2C,$3B,$63
	.dc.b	$D6,$B9,$93,$2F,$97,$97,$48,$F0,$9E,$DC,$BD,$38,$F4,$C3,$C9,$7C
	.dc.b	$23,$15,$B3,$8E,$5F,$13,$3D,$E7,$E3,$9F,$34,$63,$0F,$2E,$5E,$19
	.dc.b	$F2,$57,$D0,$C3,$AE,$1E,$E2,$65,$E4,$C7,$95,$88,$7A,$8A,$7E,$90
	.dc.b	$B9,$7E,$D1,$BD,$1F,$E0,$37,$67,$ED,$1C,$3A,$E7,$CB,$96,$CF,$F3
	.dc.b	$87,$F8,$04,$F3,$C3,$F2,$AD,$C3,$F3,$DF,$40,$EC,$F3,$D6,$AF,$89
	.dc.b	$31,$FE,$11,$5B,$71,$3B,$7F,$7D,$87,$57,$F1,$73,$B7,$35,$C3,$0F
	.dc.b	$E2,$D5,$BD,$D7,$BC,$3A,$12,$B7,$7F,$C3,$E7,$51,$8C,$3A,$92,$BF
	.dc.b	$86,$4E,$A3,$F8,$2E,$1E,$57,$E4,$1B,$AF,$87,$8F,$C2,$7A,$5F,$86
	.dc.b	$78,$E1,$E1,$AD,$46,$0E,$45,$9F,$0C,$6A,$3B,$C7,$36,$D7,$81,$F8
	.dc.b	$87,$17,$C2,$65,$EA,$70,$34,$7C,$E9,$ED,$97,$E0,$63,$6D,$AF,$1C
	.dc.b	$76,$39,$1C,$F2,$FA,$35,$CB,$1F,$8D,$7C,$70,$F1,$DF,$73,$94,$65
	.dc.b	$F1,$DF,$27,$35,$C8,$FB,$D9,$7D,$38,$F0,$D7,$17,$C5,$19,$8C,$12
	.dc.b	$78,$E1,$E5,$CB,$A9,$F4,$EE,$9F,$4B,$0F,$13,$F0,$AB,$C5,$3B,$8C
	.dc.b	$3F,$42,$77,$97,$D5,$C3,$C1,$3B,$8F,$F0,$2F,$BB,$F8,$0E,$D7,$33
	.dc.b	$CB,$C3,$72,$FF,$03,$D6,$0C,$CC,$8C,$9F,$E4,$2F,$86,$E9,$E0,$1A
	.dc.b	$63,$E7,$6C,$EC,$87,$F9,$3E,$51,$F9,$4C,$DF,$41,$FA,$1C,$DD,$FF
	.dc.b	$C4,$B5,$FF,$54,$37,$8F,$FC,$67,$5F,$A7,$BE,$3C,$5F,$C6,$B9,$3F
	.dc.b	$8D,$C6,$34,$FF,$49,$57,$EF,$67,$2D,$8E,$75,$C5,$A7,$07,$0F,$FA
	.dc.b	$86,$38,$3F,$EB,$3F,$A0,$FF,$07,$BF,$E6,$B6,$E5,$FE,$BF,$6E,$3F
	.dc.b	$EC,$B7,$D5,$3C,$7A,$7E,$B4,$6A,$30,$7F,$A7,$1D,$E9,$FE,$E2,$F2
	.dc.b	$BE,$E7,$8F,$FE,$B8,$AF,$F4,$7C,$63,$E0,$32,$C8,$C9,$FE,$94,$8F
	.dc.b	$D8,$FC,$8F,$DA,$FA,$09,$3F,$E3,$18,$CB,$FE,$C0,$3F,$C5,$EF,$C2
	.dc.b	$69,$9E,$84,$65,$B1,$98,$C1,$1F,$F5,$77,$1F,$E5,$7C,$CF,$C1,$C7
	.dc.b	$FA,$E9,$AF,$FA,$55,$FC,$73,$98,$E0,$47,$87,$37,$E7,$FC,$DF,$F4
	.dc.b	$1C,$7F,$AF,$DA,$FF,$AE,$DB,$C9,$FD,$83,$64,$62,$7F,$E0,$03,$5F
	.dc.b	$F7,$83,$63,$7D,$9F,$84,$6C,$FF,$B7,$FE,$81,$F0,$DF,$B1,$F4,$5D
	.dc.b	$B4,$FD,$04,$EA,$BF,$DF,$1D,$48,$C4,$7F,$BA,$DB,$9F,$E7,$1B,$1F
	.dc.b	$E9,$46,$BF,$EF,$56,$CF,$3C,$3F,$32,$D9,$F3,$47,$A7,$D3,$7D,$07
	.dc.b	$F0,$76,$CF,$FB,$B1,$AF,$FB,$1D,$FB,$AD,$8E,$04,$78,$70,$FF,$11
	.dc.b	$FA,$07,$C0,$FC,$76,$CF,$FB,$CD,$AF,$FB,$BD,$B9,$FD,$56,$C7,$F8
	.dc.b	$5F,$64,$62,$11,$8A,$EC,$7E,$4F,$E7,$FF,$C3,$BE,$C7,$FB,$CE,$B7
	.dc.b	$D4,$FA,$0F,$D5,$9F,$D5,$2B,$79,$7F,$A5,$5F,$F0,$CD,$7F,$A4,$1B
	.dc.b	$D0,$69,$23,$FC,$3A,$D7,$FC,$4B,$1B,$7F,$C4,$FC,$0F,$73,$E0,$C7
	.dc.b	$F1,$86,$BF,$ED,$06,$A8,$E1,$9F,$F8,$B1,$7B,$DF,$8E,$9E,$AD,$7F
	.dc.b	$C7,$71,$AE,$4E,$13,$99,$FE,$4F,$E5,$3E,$83,$A3,$63,$CE,$D7,$9B
	.dc.b	$5F,$E5,$FF,$41,$F1,$B0,$34,$91,$E7,$FA,$07,$C0,$D2,$47,$C4,$8C
	.dc.b	$B6,$32,$FA,$A3,$2D,$8F,$F2,$6F,$1F,$E3,$BD,$AB,$A9,$A4,$AF,$A3
	.dc.b	$EB,$FF,$B8,$1B,$EA,$FC,$B8,$C4,$7C,$86,$BF,$D3,$1A,$FF,$9A,$73
	.dc.b	$FF,$7D,$4F,$12,$47,$F9,$93,$07,$D0,$3C,$5F,$3B,$67,$4D,$7F,$CE
	.dc.b	$33,$C2,$7E,$F7,$D0,$3E,$6A,$92,$BF,$17,$19,$6C,$65,$FC,$0C,$65
	.dc.b	$B1,$F2,$3E,$37,$11,$FC,$8D,$AF,$EB,$DA,$FE,$B9,$BF,$1C,$FA,$0F
	.dc.b	$EC,$33,$FE,$8A,$8C,$79,$BF,$E8,$06,$C7,$FA,$19,$AF,$FA,$39,$AF
	.dc.b	$19,$E6,$FF,$A5,$63,$2D,$8C,$BF,$C0,$3E,$41,$F3,$3F,$82,$6C,$7D
	.dc.b	$16,$BF,$31,$AF,$9F,$E8,$3F,$97,$B7,$C6,$F4,$46,$23,$C4,$D5,$82
	.dc.b	$30,$D9,$E2,$49,$FC,$17,$CA,$FF,$8B,$DF,$BA,$D7,$FA,$93,$5F,$E3
	.dc.b	$CD,$7B,$FA,$BF,$C6,$5F,$90,$D7,$E8,$35,$F9,$0D,$78,$F8,$3F,$4E
	.dc.b	$31,$7F,$EA,$B8,$C1,$FA,$AA,$FE,$C7,$5E,$8F,$33,$04,$7E,$D7,$B0
	.dc.b	$FA,$0F,$BC,$D8,$CC,$60,$FC,$17,$D0,$7F,$48,$8F,$C3,$35,$FB,$CD
	.dc.b	$7E,$6B,$71,$34,$9E,$57,$43,$F8,$0F,$D0,$3E,$82,$92,$FF,$65,$19
	.dc.b	$6C,$65,$FD,$7C,$65,$B1,$FD,$03,$95,$F2,$D4,$93,$FE,$DC,$8C,$B6
	.dc.b	$32,$FF,$B9,$23,$2D,$8F,$E9,$7F,$41,$F3,$A3,$FD,$D2,$D7,$FD,$E2
	.dc.b	$D7,$FD,$EA,$D7,$FD,$F1,$F4,$1F,$9A,$7F,$DF,$2D,$73,$CB,$FD,$62
	.dc.b	$31,$06,$92,$34,$7F,$5E,$F1,$DF,$40,$F3,$FF,$DF,$31,$86,$C6,$6F
	.dc.b	$FE,$00,$CE,$5B,$EB,$3E,$83,$8B,$63,$31,$83,$FB,$6A,$F7,$A9,$1F
	.dc.b	$EE,$29,$85,$FA,$F7,$E0,$4F,$8B,$FE,$FD,$9F,$F7,$F9,$CC,$63,$AB
	.dc.b	$F8,$67,$E0,$4F,$AD,$C6,$2B,$DD,$77,$C3,$96,$3D,$63,$07,$85,$C5
	.dc.b	$7C,$78,$EB,$7D,$2F,$9C,$E4,$F9,$9F,$49,$FE,$B4,$AF,$5C,$77,$8C
	.dc.b	$38,$C1,$18,$8F,$F0,$DB,$5F,$E4,$DB,$3E,$CF,$D7,$D5,$FE,$04,$7D
	.dc.b	$B7,$CB,$E2,$73,$5B,$C7,$D7,$87,$F1,$C8,$C3,$E7,$C3,$94,$EA,$F9
	.dc.b	$3C,$23,$33,$ED,$FA,$7F,$C4,$7D,$5F,$CC,$6D,$76,$F2,$DF,$4B,$FD
	.dc.b	$6E,$3C,$33,$AB,$E5,$8B,$0E,$4D,$AE,$57,$C6,$B7,$5D,$F6,$7C,$35
	.dc.b	$3E,$1F,$85,$FE,$72,$C3,$BC,$E9,$B3,$B7,$D9,$3B,$BF,$65,$6D,$FB
	.dc.b	$59,$BC,$4F,$F1,$45,$FF,$3E,$F6,$3F,$99,$4E,$5B,$3A,$AF,$BF,$AB
	.dc.b	$EA,$7F,$C4,$78,$7C,$26,$CF,$6A,$E1,$B7,$F9,$66,$F8,$7A,$DF,$98
	.dc.b	$CB,$FD,$F9,$EC,$F8,$35,$34,$99,$FF,$84,$E3,$2D,$8C,$E5,$FE,$10
	.dc.b	$9C,$BE,$6C,$DF,$65,$F6,$DB,$7F,$6E,$B7,$9F,$96,$73,$7F,$1B,$CD
	.dc.b	$AF,$FA,$BD,$AF,$F9,$ED,$AF,$F7,$BC,$3E,$4C,$63,$23,$49,$1F,$D7
	.dc.b	$F4,$7D,$63,$BF,$F2,$0D,$D3,$F2,$17,$E4,$CF,$E5,$4E,$72,$F3,$6D
	.dc.b	$E0,$25,$FF,$82,$5B,$1F,$E0,$96,$BE,$B7,$C5,$3F,$CA,$27,$F3,$6D
	.dc.b	$9E,$B1,$8F,$03,$EC,$7B,$3F,$C4,$63,$FC,$8C,$D9,$FF,$20,$B5,$FF
	.dc.b	$6B,$35,$FF,$51,$35,$FF,$50,$B7,$5F,$F5,$7F,$67,$E4,$A3,$5A,$BE
	.dc.b	$6F,$FD,$7F,$87,$EC,$1B,$87,$E7,$2B,$C5,$A7,$E3,$E7,$2F,$FA,$A7
	.dc.b	$1F,$F5,$DF,$D2,$F5,$26,$1E,$A6,$E1,$EA,$6B,$EA,$7E,$C5,$BC,$BF
	.dc.b	$E0,$CD,$AF,$E0,$D8,$AB,$35,$FC,$DB,$D9,$7E,$07,$FE,$19,$C7,$FD
	.dc.b	$3A,$D7,$FA,$2C,$6B,$D1,$FF,$12,$37,$1F,$F5,$8D,$6E,$BF,$CF,$4D
	.dc.b	$7F,$D5,$CD,$9E,$11,$FE,$74,$F7,$5F,$63,$FB,$16,$DF,$F2,$AA,$DE
	.dc.b	$DF,$1E,$73,$7F,$D0,$65,$F4,$8F,$9C,$F6,$37,$0E,$5E,$1B,$EF,$73
	.dc.b	$5F,$07,$B7,$FB,$15,$7F,$AD,$31,$F6,$C6,$1F,$F5,$A9,$FA,$5F,$B3
	.dc.b	$FE,$31,$A7,$F0,$96,$D7,$FA,$67,$E9,$5F,$05,$FC,$DE,$73,$7F,$EE
	.dc.b	$CF,$8B,$FE,$86,$FF,$BC,$FC,$AF,$E3,$B1,$88,$FF,$7A,$35,$E1,$1A
	.dc.b	$26,$1F,$EA,$39,$D2,$FA,$EF,$C6,$BE,$C6,$D7,$28,$FE,$AA,$6F,$E3
	.dc.b	$7F,$A6,$7D,$9B,$F0,$4A,$7F,$23,$F9,$0F,$E2,$75,$FE,$07,$EE,$56
	.dc.b	$E3,$FA,$7B,$70,$FE,$B0,$DD,$FF,$3D,$7C,$A5,$7C,$B0,$FE,$B9,$F8
	.dc.b	$37,$FA,$43,$1F,$EB,$2D,$AF,$E8,$BF,$83,$7D,$B7,$FB,$24,$67,$1F
	.dc.b	$E7,$4D,$AF,$33,$65,$E0,$2F,$95,$F0,$F6,$3F,$A0,$7C,$0B,$EE,$A9
	.dc.b	$FE,$1A,$C1,$DF,$E9,$C5,$5B,$CF,$FB,$93,$6B,$FA,$DB,$71,$FF,$47
	.dc.b	$5F,$18,$C5,$7C,$BF,$C0,$F9,$BF,$DE,$1B,$7C,$B1,$FF,$00,$46,$30
	.dc.b	$FE,$B1,$CC,$FC,$07,$FC,$08,$DC,$3F,$C0,$9F,$35,$71,$FF,$05,$4D
	.dc.b	$BE,$37,$C0,$FE,$C1,$59,$BF,$BD,$56,$FE,$BF,$E1,$0F,$A4,$FF,$09
	.dc.b	$36,$BF,$C2,$5F,$84,$3E,$D7,$35,$27,$FC,$27,$5B,$BF,$EF,$F7,$B9
	.dc.b	$CD,$7C,$FC,$1E,$9F,$F7,$3F,$9A,$FF,$49,$57,$F9,$5F,$E0,$5F,$43
	.dc.b	$FE,$58,$BD,$CF,$F8,$8E,$B7,$97,$F9,$86,$B5,$87,$3F,$88,$5F,$39
	.dc.b	$CB,$C6,$33,$5C,$4F,$C4,$DF,$4C,$7C,$31,$89,$DC,$62,$3F,$80,$E1
	.dc.b	$DE,$FB,$47,$E4,$BE,$75,$F2,$1C,$32,$FF,$39,$73,$A7,$37,$C2,$31
	.dc.b	$7B,$39,$5D,$8E,$0D,$AE,$33,$E9,$9E,$35,$CE,$73,$8F,$AA,$5C,$2F
	.dc.b	$AE,$ED,$DB,$FC,$6B,$D0,$ED,$39,$C7,$D7,$39,$7C,$D1,$9C,$BE,$44
	.dc.b	$E5,$F1,$46,$71,$F9,$33,$97,$E4,$70,$3E,$BB,$D9,$F4,$DF,$DF,$E7
	.dc.b	$31,$8F,$5B,$E1,$63,$FC,$BA,$31,$97,$CE,$8C,$3F,$26,$F9,$70,$7E
	.dc.b	$1E,$B7,$5F,$74,$D9,$F1,$E7,$6F,$F0,$8F,$AE,$FF,$1C,$E3,$F1,$DB
	.dc.b	$7F,$61,$7D,$17,$D3,$73,$5C,$7C,$CC,$19,$9B,$97,$DD,$F8,$6F,$F4
	.dc.b	$FE,$3F,$4D,$B8,$FE,$11,$7D,$6F,$DF,$73,$8F,$F1,$C6,$F1,$9A,$70
	.dc.b	$C3,$B7,$C5,$FC,$8F,$A0,$FD,$D7,$3F,$C7,$6E,$1F,$9F,$FA,$97,$D5
	.dc.b	$FA,$A1,$3A,$D7,$F3,$CD,$D7,$FD,$92,$D9,$5F,$2D,$FC,$1D,$BF,$A0
	.dc.b	$9F,$A1,$5B,$7A,$C6,$70,$FD,$6D,$F1,$D3,$CD,$39,$C7,$BB,$6F,$FD
	.dc.b	$B5,$87,$86,$F8,$47,$69,$D3,$DB,$3E,$67,$ED,$27,$78,$FB,$2B,$77
	.dc.b	$E5,$3A,$67,$F0,$DC,$46,$8F,$15,$72,$C7,$CB,$87,$6C,$7C,$4F,$7C
	.dc.b	$39,$5F,$18,$D6,$3F,$02,$73,$18,$BE,$77,$F5,$67,$2F,$42,$57,$1C
	.dc.b	$FE,$7D,$F1,$9F,$8B,$87,$9A,$F8,$DF,$BE,$B5,$8F,$A6,$F7,$1C,$5F
	.dc.b	$90,$78,$70,$EA,$F8,$31,$F2,$DF,$87,$3E,$19,$FC,$6C,$3E,$06,$1C
	.dc.b	$2B,$C1,$87,$13,$F0,$F9,$79,$31,$E6,$48,$CD,$71,$8C,$47,$92,$BC
	.dc.b	$0F,$33,$11,$C7,$2F,$64,$60,$F6,$9E,$17,$38,$F9,$E3,$18,$78,$DD
	.dc.b	$DF,$A0,$EE,$5B,$F1,$D7,$E1,$B0,$F3,$57,$60,$7D,$A2,$99,$30,$7E
	.dc.b	$AB,$81,$EB,$EA,$30,$D7,$F3,$F1,$A7,$15,$FC,$4A,$75,$1F,$B8,$9D
	.dc.b	$61,$FB,$86,$E1,$D2,$7F,$80,$7D,$5F,$F8,$79,$FE,$09,$3A,$7B,$DE
	.dc.b	$0F,$D4,$4E,$E3,$F7,$73,$B7,$C1,$3B,$7A,$F9,$1F,$03,$0F,$14,$E5
	.dc.b	$B3,$A3,$18,$79,$63,$27,$F0,$4C,$3C,$0E,$23,$44,$9F,$E2,$33,$A8
	.dc.b	$C4,$79,$0F,$3D,$7A,$AF,$C9,$3A,$C7,$E5,$56,$8F,$21,$70,$F0,$3F
	.dc.b	$10,$CB,$87,$B3,$AD,$18,$F8,$E7,$84,$EE,$BB,$BE,$DC,$3C,$73,$AC
	.dc.b	$3C,$99,$7D,$3C,$79,$D6,$AB,$BD,$6B,$2F,$64,$7C,$A3,$91,$2B,$8B
	.dc.b	$93,$89,$9C,$BC,$CE,$09,$97,$9A,$FF,$37,$5C,$A7,$E2,$47,$85,$F8
	.dc.b	$6D,$CB,$E1,$5F,$4B,$E5,$5E,$43,$C3,$7C,$EB,$E2,$65,$F8,$BA,$DC
	.dc.b	$E7,$0E,$87,$DF,$71,$7C,$E7,$AB,$D3,$0F,$05,$F3,$26,$3F,$1A,$7D
	.dc.b	$27,$18,$F9,$D3,$F5,$F8,$9C,$AF,$83,$FA,$B1,$86,$C6,$6F,$FC,$21
	.dc.b	$F4,$F8,$99,$1F,$E1,$37,$71,$CB,$EA,$7F,$B2,$4F,$F8,$DD,$BB,$7F
	.dc.b	$8D,$5B,$B7,$F9,$19,$FE,$09,$19,$E2,$E4,$4A,$FF,$24,$4E,$AB,$FB
	.dc.b	$AF,$BF,$E6,$F8,$B9,$5F,$E0,$18,$FC,$57,$01,$CC,$7F,$88,$E7,$31
	.dc.b	$23,$FC,$59,$1A,$C7,$FC,$07,$3F,$13,$CD,$FF,$11,$D7,$F9,$1B,$B9
	.dc.b	$F0,$7F,$9C,$F0,$3F,$DC,$AE,$F8,$E5,$6E,$35,$5C,$23,$FC,$C5,$7D
	.dc.b	$1F,$F2,$E7,$A6,$F5,$7E,$9B,$63,$FC,$D0,$D7,$79,$6A,$64,$69,$E3
	.dc.b	$1F,$73,$C9,$7D,$3D,$25,$7F,$89,$FE,$9F,$F9,$CF,$A1,$FE,$86,$8F
	.dc.b	$F1,$2C,$63,$BF,$FC,$F7,$7C,$A7,$FC,$F4,$D7,$FC,$43,$3B,$7F,$C3
	.dc.b	$2D,$E9,$4E,$F8,$7F,$CE,$CD,$E4,$FF,$17,$D7,$1E,$27,$9D,$FE,$D9
	.dc.b	$C7,$FD,$5F,$E9,$66,$31,$0D,$3F,$D3,$B9,$FF,$A0,$3E,$0B,$57,$2B
	.dc.b	$89,$FE,$CA,$8E,$DC,$7F,$E9,$DB,$FF,$68,$E7,$83,$D1,$C6,$F8,$EC
	.dc.b	$61,$FD,$82,$B6,$DE,$0F,$F6,$6D,$F0,$AF,$57,$A9,$FE,$C0,$AF,$F7
	.dc.b	$17,$A1,$83,$E2,$7A,$19,$8F,$56,$1F,$EA,$88,$C6,$3E,$C6,$C6,$6B
	.dc.b	$E7,$53,$B3,$F1,$8F,$23,$FD,$69,$1D,$88,$F9,$CF,$EF,$31,$C2,$FF
	.dc.b	$D8,$BC,$2F,$5D,$D9,$F4,$DB,$7C,$67,$53,$F7,$7A,$BF,$05,$7C,$AF
	.dc.b	$77,$E3,$BD,$BC,$63,$33,$F6,$2B,$8E,$5F,$E0,$CB,$7B,$3F,$C4,$70
	.dc.b	$E4,$DB,$F3,$B6,$BF,$D9,$FE,$DB,$E9,$3F,$15,$39,$B6,$1D,$3C,$3C
	.dc.b	$9F,$95,$BE,$13,$F1,$70,$7D,$D7,$E1,$5F,$F0,$F9,$8E,$3F,$CF,$36
	.dc.b	$7E,$67,$86,$FA,$1F,$99,$3F,$E2,$68,$C7,$85,$FE,$1A,$6D,$7F,$88
	.dc.b	$5A,$FF,$89,$5A,$FF,$8B,$1A,$F3,$AF,$BE,$E5,$AE,$67,$FC,$4F,$9F
	.dc.b	$FA,$83,$F0,$3F,$D6,$79,$A6,$9F,$B7,$F2,$9F,$52,$F8,$4C,$7D,$47
	.dc.b	$F8,$C3,$2F,$F1,$CB,$68,$D3,$1F,$12,$F8,$EE,$8D,$AF,$C8,$52,$31
	.dc.b	$1F,$97,$87,$A4,$FC,$13,$F8,$47,$E1,$3F,$C7,$FA,$7F,$7A,$DE,$3F
	.dc.b	$F2,$8F,$8C,$8C,$63,$E6,$F9,$5F,$49,$FE,$48,$CB,$CE,$DD,$4D,$24
	.dc.b	$FF,$94,$E7,$6D,$9D,$B6,$7E,$6F,$C8,$7F,$05,$8F,$D2,$35,$FD,$03
	.dc.b	$72,$FE,$2B,$F5,$2F,$A9,$FE,$48,$DA,$FF,$31,$66,$8C,$47,$F9,$9E
	.dc.b	$31,$3F,$E6,$26,$C7,$09,$DB,$FE,$72,$8C,$BC,$63,$2F,$C8,$6B,$FA
	.dc.b	$4B,$7D,$A7,$F3,$16,$C7,$B2,$3C,$FF,$35,$23,$C0,$F8,$E8,$D2,$46
	.dc.b	$88,$78,$DF,$CA,$FC,$AF,$EB,$63,$F6,$AD,$E8,$39,$C5,$F2,$CF,$8B
	.dc.b	$FE,$98,$DB,$F9,$2D,$7E,$43,$5F,$88,$D7,$F1,$2D,$53,$DA,$7E,$74
	.dc.b	$E6,$31,$3F,$E9,$7B,$F2,$D7,$F5,$7B,$F0,$6B,$FC,$A9,$BC,$5D,$BB
	.dc.b	$BC,$E7,$42,$61,$E8,$6B,$E8,$6B,$E8,$7F,$77,$F3,$0F,$BD,$FE,$CC
	.dc.b	$D9,$FE,$68,$D7,$FA,$16,$6F,$9A,$79,$FF,$DC,$DB,$3F,$90,$6B,$F8
	.dc.b	$F6,$BF,$8B,$6B,$FB,$06,$BF,$21,$B3,$DE,$7F,$2D,$1A,$D5,$6F,$90
	.dc.b	$F4,$93,$0F,$EA,$98,$3E,$99,$F3,$7F,$E0,$96,$CF,$E9,$9A,$FE,$95
	.dc.b	$AF,$E8,$5A,$FF,$4B,$F8,$2F,$E4,$0F,$F5,$96,$BF,$D5,$5B,$C0,$DF
	.dc.b	$FD,$BB,$F3,$BE,$3F,$7B,$FC,$29,$5A,$9E,$9E,$26,$0F,$F1,$05,$6F
	.dc.b	$0F,$EC,$D8,$7F,$8A,$CF,$ED,$A5,$AF,$EB,$CD,$8F,$C8,$FD,$37,$F0
	.dc.b	$08,$DC,$FF,$88,$DD,$6F,$FD,$F6,$BE,$ED,$6E,$BF,$C2,$9F,$6B,$FE
	.dc.b	$D9,$BF,$55,$7F,$89,$63,$13,$AF,$86,$FF,$04,$FD,$27,$E5,$AB,$51
	.dc.b	$FE,$39,$B5,$F0,$9C,$9F,$E0,$91,$5C,$5B,$D1,$FD,$1E,$3F,$59,$18
	.dc.b	$3F,$DF,$2F,$EB,$FE,$81,$F6,$3F,$D0,$27,$FC,$29,$18,$F1,$BF,$C9
	.dc.b	$4D,$8F,$07,$C2,$78,$E3,$11,$FE,$1F,$6E,$5F,$01,$BA,$26,$57,$AD
	.dc.b	$7C,$F3,$E4,$7F,$C9,$51,$B2,$46,$70,$C4,$62,$76,$E2,$74,$E8,$EB
	.dc.b	$EA,$3E,$97,$11,$86,$CF,$FA,$C5,$B3,$B9,$FE,$E1,$F0,$DF,$DC,$53
	.dc.b	$48,$F9,$33,$7D,$23,$E3,$FF,$03,$9F,$C3,$8C,$6B,$FE,$41,$E6,$7D
	.dc.b	$2F,$FA,$6A,$3E,$8C,$EB,$1F,$98,$DB,$E5,$87,$4A,$FD,$2D,$F1,$BD
	.dc.b	$CE,$5F,$F3,$A7,$D2,$BE,$4B,$FD,$19,$1F,$E6,$98,$C5,$FC,$7C,$11
	.dc.b	$89,$F9,$F1,$8A,$E3,$0C,$BE,$CF,$CF,$F7,$6E,$1F,$E5,$DB,$E0,$FB
	.dc.b	$F8,$15,$B6,$C6,$63,$D1,$3C,$2F,$FC,$D5,$7D,$1F,$F2,$D4,$7F,$A0
	.dc.b	$EB,$F7,$BF,$0D,$EF,$F0,$7F,$B0,$FD,$47,$C1,$ED,$FF,$4A,$4D,$BE
	.dc.b	$18,$72,$8E,$1E,$5A,$F5,$1C,$78,$BF,$CE,$39,$7F,$9B,$DB,$7E,$4F
	.dc.b	$85,$7D,$7E,$FE,$16,$5E,$DF,$89,$3A,$8C,$1F,$CB,$BC,$9F,$F1,$0B
	.dc.b	$FD,$07,$D3,$F9,$BB,$FF,$47,$6C,$EB,$1F,$D0,$EF,$FC,$EB,$D9,$F7
	.dc.b	$79,$7F,$E3,$FC,$7A,$61,$89,$F8,$99,$7F,$8F,$27,$FC,$3A,$DC,$3F
	.dc.b	$C3,$8D,$7F,$C3,$0D,$7F,$C7,$AD,$7F,$BA,$B7,$23,$4D,$63,$FD,$3B
	.dc.b	$CA,$F2,$9E,$2F,$02,$E1,$FD,$3B,$D0,$FE,$9A,$FF,$82,$27,$E0,$C6
	.dc.b	$31,$FF,$04,$C6,$09,$3B,$24,$7C,$4F,$41,$DF,$F8,$16,$CF,$7A,$C4
	.dc.b	$FE,$1E,$FE,$45,$FF,$85,$B4,$F0,$92,$FF,$C3,$6D,$8D,$46,$0F,$F0
	.dc.b	$96,$1F,$C6,$5F,$F5,$A3,$63,$FC,$03,$5D,$B9,$BF,$C2,$9E,$0F,$E5
	.dc.b	$FE,$9F,$F7,$15,$FE,$47,$6B,$FE,$6C,$6B,$FE,$24,$6E,$E6,$93,$DE
	.dc.b	$FC,$4B,$C3,$77,$AA,$F3,$BF,$48,$DD,$7F,$BF,$D7,$F0,$9C,$FF,$DA
	.dc.b	$33,$9C,$BF,$C3,$32,$F9,$BE,$A4,$BF,$F3,$3C,$65,$B1,$99,$FF,$34
	.dc.b	$4E,$5B,$1F,$C2,$B1,$F2,$F8,$FE,$43,$E2,$45,$BF,$F4,$BF,$D3,$BC
	.dc.b	$DF,$E2,$3E,$D2,$31,$3F,$E8,$76,$E3,$FD,$1A,$75,$5C,$27,$33,$FD
	.dc.b	$36,$73,$5C,$6B,$47,$EC,$BD,$D7,$C6,$FF,$0D,$6C,$F9,$27,$13,$FC
	.dc.b	$BE,$9F,$4F,$FE,$B9,$AF,$F5,$7B,$70,$FF,$57,$78,$CF,$A7,$E1,$39
	.dc.b	$E1,$FF,$03,$E1,$FD,$CF,$0F,$F5,$17,$5F,$FB,$83,$1F,$63,$5F,$53
	.dc.b	$5F,$F7,$C7,$D3,$7F,$16,$BF,$F0,$8F,$0A,$31,$7F,$DC,$F0,$7D,$3B
	.dc.b	$E4,$7F,$94,$56,$B1,$FF,$39,$FC,$57,$F3,$67,$FD,$44,$D7,$FD,$3C
	.dc.b	$D7,$FE,$00,$F9,$7F,$C6,$A3,$11,$CA,$30,$4C,$7F,$BB,$0F,$A7,$FC
	.dc.b	$86,$9F,$E2,$88,$C7,$4F,$F3,$DC,$3C,$53,$9A,$DD,$7F,$83,$75,$7C
	.dc.b	$67,$F9,$76,$3F,$C2,$15,$F0,$A3,$19,$7F,$9D,$3E,$3B,$F9,$14,$E7
	.dc.b	$DE,$FE,$91,$5A,$8C,$5F,$F9,$92,$B6,$F7,$9D,$3E,$1F,$8F,$FF,$52
	.dc.b	$3E,$03,$1E,$D7,$F5,$5F,$A7,$FB,$D1,$83,$59,$33,$5F,$28,$65,$F6
	.dc.b	$A3,$0F,$F9,$33,$C1,$61,$D2,$B9,$56,$B0,$FF,$7A,$FC,$0B,$E5,$3F
	.dc.b	$B3,$36,$3F,$C9,$5D,$9E,$CB,$FF,$58,$CE,$5F,$F4,$CB,$72,$F0,$36
	.dc.b	$BF,$C6,$D5,$C7,$2F,$F4,$EE,$1F,$DC,$7E,$53,$FD,$3F,$5E,$ED,$11
	.dc.b	$8C,$BA,$D6,$F1,$FE,$F1,$CC,$FA,$7A,$7F,$7D,$FA,$3C,$65,$CE,$6E
	.dc.b	$1F,$1A,$3F,$05,$18,$D7,$F4,$2D,$8F,$F2,$96,$8B,$C5,$FD,$1A,$BA
	.dc.b	$3E,$5D,$F4,$1F,$E0,$B6,$E3,$FE,$A6,$FA,$67,$D3,$E6,$A4,$FB,$EB
	.dc.b	$78,$FF,$B0,$AF,$73,$9A,$FF,$1B,$4F,$EC,$7D,$1F,$F0,$D7,$CE,$3E
	.dc.b	$9F,$FD,$13,$F0,$6F,$8D,$FF,$32,$6D,$FD,$C3,$2F,$F5,$BD,$ED,$FF
	.dc.b	$37,$56,$F1,$FF,$38,$56,$DF,$F6,$9D,$6D,$FF,$2D,$DF,$4C,$7C,$38
	.dc.b	$F8,$23,$E3,$D6,$F0,$E9,$C0,$8C,$4F,$1C,$BC,$77,$C7,$1F,$F7,$DF
	.dc.b	$CE,$3E,$67,$18,$A3,$17,$E8,$C3,$EA,$C7,$F1,$FC,$BB,$BA,$C3,$CE
	.dc.b	$74,$C3,$9C,$66,$77,$38,$9E,$FC,$4F,$A7,$F2,$71,$FF,$1F,$F3,$B3
	.dc.b	$18,$E0,$FE,$AD,$8F,$BA,$F8,$69,$E5,$CB,$F0,$D8,$70,$8C,$63,$CD
	.dc.b	$D4,$FF,$2B,$FA,$7F,$7F,$D2,$7F,$80,$F1,$FA,$93,$9A,$E9,$BA,$B5
	.dc.b	$BF,$DE,$C3,$8F,$AB,$F2,$F0,$E0,$7C,$2B,$E1,$19,$D5,$5C,$31,$F1
	.dc.b	$C6,$1E,$F8,$F6,$C1,$83,$E9,$FF,$C8,$99,$FD,$E6,$E7,$FA,$85,$F4
	.dc.b	$7F,$69,$CD,$F2,$6D,$F3,$F3,$75,$5B,$7F,$03,$F0,$FF,$E9,$DC,$7F
	.dc.b	$16,$DC,$7F,$6A,$BE,$9F,$7F,$15,$35,$F8,$F3,$BD,$3D,$2E,$E3,$33
	.dc.b	$F1,$FE,$97,$F7,$3C,$DF,$51,$B7,$F8,$AF,$94,$7C,$FF,$AA,$13,$AD
	.dc.b	$3E,$BB,$74,$FF,$6A,$36,$57,$CB,$5F,$07,$6F,$E7,$E7,$F2,$E7,$47
	.dc.b	$18,$74,$6E,$1F,$EE,$5B,$E5,$A7,$9E,$73,$7F,$EE,$9C,$7C,$0F,$3C
	.dc.b	$FF,$29,$7C,$23,$BC,$E8,$95,$CA,$75,$8F,$86,$FB,$12,$FE,$26,$7F
	.dc.b	$0F,$0E,$33,$9A,$FE,$19,$7C,$B2,$F8,$78,$F9,$B1,$F2,$1E,$27,$C1
	.dc.b	$8F,$37,$58,$F4,$9E,$F9,$FC,$8C,$BB,$61,$D1,$F6,$E1,$C5,$CD,$7C
	.dc.b	$3C,$3A,$DF,$18,$F0,$4E,$B1,$F3,$46,$E3,$8B,$E6,$CB,$D7,$87,$58
	.dc.b	$E2,$48,$CD,$F5,$7E,$44,$67,$1F,$B1,$5C,$67,$24,$8C,$D7,$17,$D9
	.dc.b	$5C,$4F,$C4,$D7,$18,$CD,$71,$9C,$D7,$17,$D4,$71,$C3,$DA,$FA,$EB
	.dc.b	$89,$1E,$39,$FC,$0C,$FE,$86,$1E,$C8,$C6,$7F,$50,$F9,$6E,$33,$F9
	.dc.b	$59,$F9,$70,$F0,$DF,$97,$2F,$0E,$3F,$6E,$47,$B8,$A7,$A0,$A7,$E1
	.dc.b	$8A,$FF,$0C,$F4,$F3,$1F,$9F,$E0,$46,$23,$F6,$B1,$87,$73,$C5,$B1
	.dc.b	$9B,$FD,$CB,$74,$FD,$CB,$67,$F7,$3F,$4F,$FE,$8D,$7F,$83,$46,$5E
	.dc.b	$F8,$60,$FD,$7C,$EE,$3F,$5F,$3B,$C3,$C5,$87,$82,$F7,$E4,$7C,$28
	.dc.b	$C5,$F0,$9E,$8F,$F1,$3C,$3A,$4E,$89,$8F,$A2,$B9,$D7,$09,$CB,$F1
	.dc.b	$E3,$2F,$52,$46,$B2,$5F,$6B,$E9,$5A,$6D,$7E,$FA,$75,$3D,$4D,$16
	.dc.b	$35,$87,$B6,$B8,$92,$74,$FF,$1D,$AE,$71,$F3,$0E,$65,$AE,$6F,$B6
	.dc.b	$79,$39,$79,$57,$78,$EA,$43,$47,$DC,$74,$E0,$98,$FD,$69,$E7,$1D
	.dc.b	$CE,$51,$97,$91,$93,$8E,$5F,$89,$C7,$E5,$DF,$6A,$E1,$5A,$CB,$E9
	.dc.b	$E3,$F5,$A3,$07,$DF,$AE,$18,$7E,$24,$95,$C3,$1F,$83,$87,$43,$C9
	.dc.b	$8F,$83,$0F,$24,$66,$FC,$99,$7C,$E8,$C6,$1E,$DB,$E1,$7F,$42,$72
	.dc.b	$F8,$CE,$77,$E2,$70,$7F,$83,$F9,$5F,$05,$F4,$79,$7F,$C3,$0D,$8C
	.dc.b	$CF,$F8,$3B,$E9,$F1,$32,$3F,$C2,$2E,$E3,$97,$D2,$FF,$67,$9F,$F0
	.dc.b	$E3,$77,$FF,$27,$B7,$6F,$F3,$CB,$FC,$2A,$33,$C3,$C8,$95,$FE,$2C
	.dc.b	$9D,$57,$F7,$8F,$7B,$CD,$F1,$32,$BF,$C0,$91,$F6,$4C,$3E,$F8,$C3
	.dc.b	$A3,$FC,$52,$EB,$0F,$F0,$2C,$FC,$3F,$2F,$FC,$4B,$5F,$E4,$4E,$C7
	.dc.b	$C1,$7E,$5F,$C0,$FF,$77,$3B,$E1,$95,$B8,$D5,$70,$8F,$F2,$2D,$F4
	.dc.b	$7F,$C8,$1E,$9B,$D7,$FA,$4D,$8F,$F2,$5B,$5D,$E5,$A9,$91,$A7,$8C
	.dc.b	$7D,$BF,$21,$F4,$F4,$95,$FE,$29,$FA,$7F,$E7,$7E,$7B,$BB,$FC,$47
	.dc.b	$18,$7F,$D0,$F5,$C2,$BF,$C2,$73,$97,$FC,$34,$DE,$8F,$F4,$54,$EF
	.dc.b	$87,$FC,$5F,$3B,$E1,$FF,$59,$7A,$37,$CF,$FF,$6D,$63,$FE,$AD,$8E
	.dc.b	$5E,$53,$E0,$35,$12,$74,$7F,$BB,$B3,$FF,$60,$4F,$5E,$3F,$F3,$9E
	.dc.b	$DF,$EC,$6E,$7C,$1E,$5E,$47,$C7,$63,$1F,$F6,$84,$D7,$FD,$CB,$1E
	.dc.b	$7F,$51,$E8,$D2,$4F,$87,$D0,$CB,$63,$E1,$46,$23,$FD,$B4,$7F,$AA
	.dc.b	$23,$E4,$2E,$CF,$C2,$3A,$D7,$FD,$55,$3C,$89,$1F,$DC,$B3,$FA,$CF
	.dc.b	$F7,$4A,$E3,$18,$E1,$7A,$AE,$DF,$A0,$DB,$FF,$7C,$35,$F0,$75,$FE
	.dc.b	$19,$CC,$F0,$23,$F3,$E3,$4F,$B2,$9E,$C7,$F0,$68,$C3,$70,$F2,$37
	.dc.b	$A7,$FD,$EF,$ED,$BE,$9F,$F0,$93,$9B,$61,$D3,$C3,$C9,$F8,$DB,$E1
	.dc.b	$3F,$0F,$0F,$F0,$A8,$FA,$7F,$C0,$B8,$6C,$7E,$75,$BE,$1B,$E9,$FE
	.dc.b	$4D,$7F,$87,$A3,$1E,$1F,$F8,$55,$B7,$FE,$1A,$6B,$FE,$20,$6B,$FE
	.dc.b	$24,$6B,$CE,$BE,$EB,$96,$B9,$9F,$F1,$16,$5F,$E9,$8F,$9D,$FD,$37
	.dc.b	$9A,$69,$FA,$FF,$2D,$F4,$EF,$A7,$C7,$D2,$FF,$8A,$F2,$FF,$16,$B6
	.dc.b	$8D,$31,$F1,$2F,$8F,$E8,$DA,$F9,$87,$F8,$D9,$C3,$F9,$AA,$7A,$8F
	.dc.b	$A7,$FE,$01,$F4,$FF,$C5,$7A,$3F,$2B,$DC,$8C,$72,$7F,$92,$A3,$1E
	.dc.b	$77,$A3,$E4,$E6,$7F,$E4,$6C,$FC,$F1,$FE,$45,$6C,$67,$64,$E5,$B1
	.dc.b	$98,$C1,$1F,$F2,$8F,$1B,$E4,$3F,$80,$C7,$BA,$31,$E2,$7F,$15,$9C
	.dc.b	$CF,$EC,$FE,$95,$F4,$FF,$C7,$5B,$1F,$E5,$CD,$11,$88,$FF,$31,$C6
	.dc.b	$27,$FC,$BC,$D8,$E1,$3B,$7F,$CD,$71,$97,$8C,$65,$F9,$0D,$7F,$4B
	.dc.b	$4F,$A7,$F4,$12,$3F,$92,$4B,$E7,$3D,$07,$89,$C4,$78,$A3,$04,$70
	.dc.b	$43,$C6,$FE,$5B,$E5,$BF,$57,$99,$A7,$EB,$B1,$7C,$B3,$E3,$3F,$A2
	.dc.b	$37,$0F,$90,$D7,$E3,$B5,$F8,$6D,$7F,$10,$D5,$3D,$A7,$E6,$CE,$63
	.dc.b	$13,$F8,$0A,$ED,$C9,$FC,$95,$BC,$7F,$E9,$CC,$7D,$58,$F9,$BC,$C9
	.dc.b	$87,$99,$EA,$47,$04,$3D,$2E,$5F,$F5,$1C,$BE,$61,$F4,$FF,$D8,$5B
	.dc.b	$3F,$91,$6B,$FD,$07,$E6,$9E,$87,$F6,$B6,$C7,$E3,$DA,$FE,$39,$AF
	.dc.b	$E2,$9A,$FE,$BD,$AF,$C8,$6E,$A6,$9A,$C3,$D1,$F2,$1E,$72,$57,$F6
	.dc.b	$FF,$A7,$7C,$DF,$F7,$F6,$C7,$E8,$DA,$FE,$89,$AF,$E7,$9A,$FD,$1F
	.dc.b	$82,$FE,$3C,$FE,$CD,$AF,$F4,$D6,$F0,$FF,$60,$F9,$F5,$7F,$0C,$AF
	.dc.b	$E8,$F5,$AE,$1E,$9E,$2F,$F8,$3E,$31,$2B,$E3,$DD,$FE,$13,$C7,$FB
	.dc.b	$5C,$FF,$5B,$69,$FD,$AC,$B8,$7F,$19,$FA,$7F,$DF,$46,$63,$91,$27
	.dc.b	$FC,$29,$3A,$8F,$EF,$1F,$4F,$FE,$CD,$AF,$EF,$21,$C4,$6B,$E1,$FF
	.dc.b	$C0,$BF,$4F,$F9,$6C,$17,$B9,$FF,$18,$36,$3F,$9B,$B5,$FF,$07,$89
	.dc.b	$E4,$DE,$7F,$E8,$D1,$FA,$B8,$C1,$FE,$F3,$7F,$5D,$F4,$2F,$A7,$FE
	.dc.b	$7F,$3F,$D1,$23,$1E,$3F,$F9,$05,$B1,$FD,$20,$7C,$2F,$B1,$46,$9F
	.dc.b	$F1,$1B,$71,$FF,$0D,$B7,$34,$CA,$E4,$BE,$7D,$F2,$5F,$12,$36,$48
	.dc.b	$CE,$18,$8C,$4E,$DC,$4E,$9D,$1D,$7D,$57,$D3,$E2,$30,$D9,$FF,$53
	.dc.b	$B6,$77,$3F,$DC,$7E,$1B,$FB,$92,$69,$1F,$27,$23,$E9,$BF,$D1,$7A
	.dc.b	$78,$A3,$1E,$A7,$E5,$2B,$E1,$C6,$27,$55,$C2,$79,$B5,$F8,$73,$B7
	.dc.b	$35,$CD,$FD,$14,$ED,$F9,$3E,$CB,$EA,$B5,$1F,$E3,$C6,$C6,$70,$F8
	.dc.b	$8D,$9F,$9D,$B3,$E9,$DF,$27,$DE,$30,$D8,$CD,$FD,$09,$CB,$63,$35
	.dc.b	$D1,$AF,$FA,$59,$AF,$F9,$8D,$AF,$F4,$4F,$25,$F4,$FE,$FF,$13,$FB
	.dc.b	$13,$63,$37,$FD,$E5,$B5,$FD,$E1,$AF,$F7,$46,$BF,$E1,$96,$BD,$E7
	.dc.b	$E6,$6D,$FE,$86,$FA,$7B,$F0,$86,$93,$0F,$BD,$19,$6C,$66,$FF,$0B
	.dc.b	$39,$7F,$0F,$F7,$7F,$E2,$0A,$FC,$33,$63,$3E,$9F,$8A,$31,$1F,$D0
	.dc.b	$E3,$E1,$79,$18,$8F,$E8,$91,$83,$87,$D5,$CB,$EA,$D9,$F0,$B5,$FC
	.dc.b	$6D,$A3,$11,$FE,$3D,$C3,$FC,$8B,$5F,$E2,$56,$C7,$F8,$91,$AF,$F8
	.dc.b	$81,$AF,$E7,$BE,$7D,$F7,$FF,$31,$B5,$FA,$56,$CF,$EA,$9A,$FE,$81
	.dc.b	$AF,$39,$E4,$49,$FF,$08,$F9,$3F,$85,$F4,$7F,$AD,$C6,$09,$5D,$75
	.dc.b	$F1,$12,$3C,$4D,$8D,$46,$08,$FF,$87,$3D,$CE,$C4,$8F,$F1,$0B,$FB
	.dc.b	$3F,$65,$F7,$9F,$56,$91,$8D,$F5,$39,$8C,$5F,$F9,$9E,$30,$49,$D9
	.dc.b	$FA,$1F,$AB,$FD,$F5,$FF,$8A,$1B,$CA,$69,$AC,$3C,$7C,$1E,$02,$63
	.dc.b	$DD,$B7,$FC,$B6,$77,$39,$24,$7F,$98,$A9,$F3,$FE,$32,$4F,$F9,$A6
	.dc.b	$32,$FF,$A0,$7E,$A5,$F4,$FF,$15,$B1,$98,$C1,$1C,$12,$FC,$93,$C1
	.dc.b	$F3,$7D,$5B,$CC,$FA,$0D,$AF,$F4,$83,$67,$73,$F5,$7C,$57,$D5,$D2
	.dc.b	$5F,$FA,$5E,$32,$FA,$7E,$13,$D0,$47,$07,$FA,$77,$EA,$DF,$27,$F8
	.dc.b	$16,$C6,$63,$07,$F3,$21,$F4,$34,$91,$FD,$56,$32,$D8,$CE,$5F,$D7
	.dc.b	$27,$2D,$8F,$E4,$9F,$1B,$F1,$63,$FD,$22,$D7,$FD,$EB,$B6,$8F,$E5
	.dc.b	$5F,$1F,$FB,$68,$FF,$38,$B7,$90,$D2,$7A,$1C,$09,$97,$FA,$9E,$38
	.dc.b	$7D,$4F,$FB,$66,$3F,$D5,$4D,$BF,$F5,$4B,$5F,$F5,$2B,$5F,$F6,$53
	.dc.b	$5F,$F4,$1B,$6C,$D3,$51,$F8,$1B,$FE,$BB,$C1,$EB,$8C,$5F,$F7,$F8
	.dc.b	$C1,$27,$64,$8F,$F6,$07,$BA,$F9,$7F,$E9,$CD,$8C,$C6,$0F,$EA,$5F
	.dc.b	$30,$FA,$9A,$49,$FF,$65,$46,$5B,$19,$C7,$FD,$97,$39,$6C,$7F,$5D
	.dc.b	$FA,$97,$87,$49,$7F,$ED,$58,$CB,$63,$35,$FE,$D6,$9C,$B6,$3F,$21
	.dc.b	$E2,$4A,$FC,$73,$67,$F1,$8D,$7F,$C5,$DC,$68,$C4,$7F,$8D,$23,$13
	.dc.b	$FE,$3C,$F2,$CF,$F8,$FB,$81,$F3,$CF,$A9,$CE,$1F,$E2,$9F,$86,$E5
	.dc.b	$97,$F9,$A3,$0E,$17,$B7,$FC,$8F,$EA,$2B,$A7,$CD,$3E,$57,$9C,$E7
	.dc.b	$1F,$F6,$4C,$E6,$F8,$DF,$09,$DB,$CA,$EC,$70,$6E,$6B,$E5,$3E,$73
	.dc.b	$A9,$A7,$38,$78,$65,$C2,$FA,$9F,$7E,$5F,$E5,$A9,$CC,$63,$0F,$8D
	.dc.b	$5B,$AF,$F3,$29,$B7,$11,$FE,$9A,$AE,$33,$A8,$E3,$5C,$DF,$F7,$4D
	.dc.b	$70,$9E,$53,$99,$FE,$2B,$F5,$37,$80,$D2,$57,$59,$FF,$3B,$CE,$A3
	.dc.b	$11,$C9,$B1,$98,$FF,$2E,$36,$3F,$13,$3F,$33,$E9,$3F,$CF,$B1,$F5
	.dc.b	$2B,$AC,$E7,$DC,$FA,$3F,$53,$FE,$0C,$D3,$E6,$37,$7F,$C6,$2F,$A4
	.dc.b	$DF,$54,$CB,$A6,$EA,$E1,$9A,$31,$7B,$8F,$B9,$1F,$E5,$CF,$03,$F0
	.dc.b	$3E,$C7,$F2,$CC,$FE,$E2,$FA,$9C,$CC,$CB,$BC,$EF,$4F,$D0,$3B,$8C
	.dc.b	$CF,$E3,$B1,$7D,$3B,$EA,$7F,$35,$1F,$8D,$32,$E3,$2F,$CE,$7C,$53
	.dc.b	$E7,$3D,$75,$AE,$8F,$94,$D3,$96,$1D,$63,$E7,$57,$1E,$1F,$14,$E6
	.dc.b	$B7,$8F,$7D,$7F,$D5,$F3,$9C,$BC,$39,$78,$23,$13,$AA,$E1,$7F,$EB
	.dc.b	$C9,$CA,$C7,$C1,$A7,$9A,$73,$3E,$7C,$7F,$70,$DC,$3F,$D8,$B7,$C2
	.dc.b	$3C,$B9,$F1,$C7,$BC,$6C,$ED,$97,$96,$33,$8F,$63,$C7,$7C,$F1,$EF
	.dc.b	$19,$9D,$9E,$A9,$D6,$3E,$DC,$7C,$04,$C7,$F8,$14,$E9,$F8,$15,$CA
	.dc.b	$75,$87,$C8,$CF,$E7,$CE,$6B,$D6,$72,$CB,$D1,$7C,$31,$F0,$4E,$6F
	.dc.b	$8C,$E6,$3B,$BF,$72,$B8,$C7,$4C,$3D,$B8,$7A,$E7,$31,$8A,$E5,$8F
	.dc.b	$A6,$FE,$51,$32,$F8,$F7,$C3,$2F,$1E,$1C,$6B,$C9,$9F,$8D,$E1,$97
	.dc.b	$7C,$FC,$F1,$E3,$BE,$27,$DC,$C3,$A4,$7A,$8D,$1F,$6A,$78,$E1,$DC
	.dc.b	$95,$C6,$32,$F4,$9E,$53,$F3,$F0,$F2,$63,$F7,$73,$F8,$D8,$70,$BD
	.dc.b	$E1,$C7,$3F,$85,$7B,$91,$0F,$31,$4F,$D7,$15,$FD,$53,$78,$3F,$CF
	.dc.b	$AE,$CF,$DF,$B8,$75,$E7,$E5,$CB,$67,$F8,$E3,$FA,$79,$E7,$9F,$E6
	.dc.b	$1B,$87,$E9,$BE,$A7,$FC,$C1,$86,$EE,$3F,$C0,$AB,$6F,$EC,$E7,$6E
	.dc.b	$67,$77,$D5,$FE,$0F,$3A,$8C,$4E,$C8,$FF,$14,$8D,$5F,$F0,$8B,$EF
	.dc.b	$5E,$C3,$71,$E7,$F1,$3E,$65,$F4,$24,$71,$C3,$C3,$3A,$8C,$57,$23
	.dc.b	$F8,$75,$FB,$32,$63,$E2,$9C,$E1,$EC,$BE,$38,$F9,$F1,$E9,$5A,$3A
	.dc.b	$1E,$78,$D3,$C4,$D3,$E6,$7E,$54,$72,$78,$4F,$62,$3A,$39,$1A,$39
	.dc.b	$BF,$0D,$CE,$5F,$80,$7D,$D8,$FB,$2B,$B4,$E6,$FC,$55,$CE,$B4,$F2
	.dc.b	$CB,$E8,$D6,$F0,$ED,$3E,$E8,$C3,$E2,$3C,$C6,$67,$89,$F3,$72,$F9
	.dc.b	$91,$E9,$24,$F2,$BF,$0C,$71,$CB,$E4,$46,$31,$F3,$D7,$82,$BD,$17
	.dc.b	$D2,$30,$4C,$3C,$19,$7C,$07,$C7,$8F,$4A,$E1,$18,$BF,$8F,$18,$9F
	.dc.b	$84,$7E,$2A,$3F,$C0,$7C,$EF,$83,$FC,$0F,$2A,$31,$3A,$9C,$C3,$EA
	.dc.b	$6C,$49,$D1,$D2,$39,$FD,$27,$F7,$A8,$FF,$05,$B7,$5F,$F2,$23,$77
	.dc.b	$FF,$32,$3F,$C2,$E3,$3C,$3C,$89,$5F,$E1,$A9,$D5,$7F,$74,$F8,$7F
	.dc.b	$2F,$D1,$CA,$FE,$D9,$1F,$86,$3F,$8C,$72,$7F,$15,$6C,$71,$3F,$C7
	.dc.b	$3C,$48,$CC,$7F,$79,$6B,$FE,$16,$6C,$6E,$38,$C7,$0D,$9E,$4F,$F8
	.dc.b	$CE,$7F,$C9,$11,$8E,$2D,$4E,$67,$FC,$B4,$DC,$3F,$CC,$8D,$8F,$F4
	.dc.b	$43,$5F,$F1,$C3,$6B,$A4,$76,$F2,$DF,$1F,$F7,$1B,$1F,$E4,$26,$BF
	.dc.b	$E3,$E7,$D4,$D8,$E0,$47,$86,$3F,$B0,$FA,$8F,$E8,$5E,$47,$FA,$63
	.dc.b	$0F,$F4,$DF,$D4,$FE,$2B,$C7,$3B,$C3,$FD,$45,$A7,$F8,$AB,$0F,$E7
	.dc.b	$D1,$FE,$40,$6E,$C6,$92,$3F,$A1,$35,$FE,$95,$1B,$7F,$D9,$7F,$51
	.dc.b	$FD,$27,$1F,$F4,$7B,$7C,$73,$49,$3C,$08,$7F,$B4,$1E,$5B,$BE,$7A
	.dc.b	$92,$7F,$DC,$6D,$7F,$B1,$5F,$E0,$23,$11,$F6,$9F,$27,$93,$FE,$C7
	.dc.b	$8F,$EC,$53,$FE,$AB,$F2,$B2,$48,$FE,$C5,$CC,$F8,$CF,$8E,$D8,$CC
	.dc.b	$60,$8E,$08,$7F,$6F,$8E,$0F,$9B,$93,$EA,$BE,$26,$B9,$E0,$F9,$51
	.dc.b	$88,$FF,$5B,$35,$FF,$6D,$35,$FF,$45,$37,$03,$7D,$37,$D3,$7D,$56
	.dc.b	$D7,$C2,$6B,$F2,$9A,$FF,$B1,$5A,$FF,$BA,$5A,$FF,$A6,$5B,$47,$3D
	.dc.b	$4F,$A5,$F8,$35,$FE,$19,$8C,$67,$FE,$0F,$6C,$7F,$85,$5A,$FE,$3B
	.dc.b	$E8,$3F,$2A,$47,$FC,$28,$D7,$46,$08,$7F,$89,$7C,$17,$D2,$BD,$9F
	.dc.b	$F0,$F4,$61,$B1,$99,$FF,$13,$4E,$5B,$F4,$AF,$84,$E2,$D8,$CC,$60
	.dc.b	$EB,$E4,$BE,$96,$92,$7A,$C6,$5E,$FF,$48,$F8,$5F,$2B,$63,$31,$83
	.dc.b	$C3,$F0,$72,$3C,$AD,$7E,$06,$28,$C4,$7F,$92,$72,$4E,$63,$89,$FA
	.dc.b	$4F,$A2,$6D,$FB,$B6,$C6,$63,$07,$CA,$FB,$17,$C8,$52,$4F,$CA,$8C
	.dc.b	$B6,$32,$FD,$58,$CB,$63,$C1,$F2,$32,$3C,$8D,$78,$7C,$2F,$84,$8F
	.dc.b	$76,$BA,$30,$43,$F8,$3F,$C6,$3E,$6B,$F8,$C3,$63,$F8,$2B,$5D,$C7
	.dc.b	$F9,$97,$E4,$9F,$60,$FB,$0C,$36,$BF,$CD,$D3,$A9,$C9,$23,$33,$B7
	.dc.b	$FC,$F5,$3B,$73,$3E,$4F,$42,$47,$91,$AF,$0A,$7D,$4F,$88,$91,$D9
	.dc.b	$AE,$DA,$FF,$A2,$EF,$D7,$1F,$CE,$7E,$57,$FB,$4C,$FF,$26,$6B,$FC
	.dc.b	$B9,$AF,$FA,$4F,$BD,$18,$8F,$F4,$BD,$3C,$47,$CC,$7C,$96,$CF,$C6
	.dc.b	$6B,$F0,$9A,$F9,$5A,$F4,$6A,$9E,$D3,$D4,$FE,$A5,$F4,$2F,$95,$A4
	.dc.b	$AF,$E8,$31,$97,$C5,$F0,$9E,$52,$38,$3E,$DC,$7F,$78,$7E,$43,$5F
	.dc.b	$7B,$5F,$23,$6C,$D2,$7A,$9F,$8B,$78,$66,$FA,$0F,$D7,$DF,$F5,$C6
	.dc.b	$BF,$3D,$AF,$FB,$12,$38,$73,$FF,$71,$9C,$CA,$72,$FF,$60,$8C,$C6
	.dc.b	$23,$FD,$89,$F3,$FF,$9F,$7F,$D9,$8D,$7F,$DA,$0D,$7F,$DA,$6D,$4D
	.dc.b	$27,$1F,$F8,$39,$D7,$E1,$5F,$D5,$E4,$D2,$79,$7F,$E1,$4F,$5F,$FD
	.dc.b	$87,$3F,$ED,$D6,$BF,$EE,$16,$BF,$EE,$46,$BF,$E4,$9F,$7F,$91,$1F
	.dc.b	$F6,$E4,$6B,$E4,$9F,$10,$F6,$BF,$DC,$11,$86,$C6,$6B,$FD,$D1,$39
	.dc.b	$6F,$BE,$F9,$EE,$2D,$8C,$C6,$0F,$F1,$7A,$F7,$E9,$1F,$F1,$9B,$FC
	.dc.b	$23,$E7,$BF,$C2,$1E,$ED,$F4,$4D,$25,$7F,$08,$9D,$B6,$76,$D9,$DA
	.dc.b	$20,$F9,$EF,$DA,$C7,$FC,$01,$6C,$68,$E0,$E1,$C3,$5E,$9A,$3E,$79
	.dc.b	$F1,$5E,$29,$FF,$2B,$C6,$2F,$A4,$7D,$18,$C3,$DE,$B9,$57,$18,$EB
	.dc.b	$3A,$79,$C6,$5E,$4D,$9E,$35,$FE,$3E,$F1,$BF,$AA,$B7,$C7,$E1,$CE
	.dc.b	$9C,$E1,$FE,$4A,$6B,$FC,$75,$AF,$EB,$DA,$FE,$B5,$B1,$CE,$7C,$9D
	.dc.b	$9F,$C8,$7E,$72,$F8,$26,$93,$1F,$E4,$4D,$C7,$F9,$13,$5F,$E6,$DF
	.dc.b	$3D,$FE,$05,$9F,$E9,$AD,$8F,$E9,$2D,$77,$E7,$FB,$BC,$4C,$47,$F0
	.dc.b	$76,$BF,$C1,$5B,$9F,$F3,$A9,$F4,$D3,$E7,$29,$30,$F0,$B7,$0F,$0B
	.dc.b	$5F,$3B,$F3,$B2,$7A,$54,$95,$FE,$72,$8C,$BF,$DA,$3E,$4D,$F3,$DF
	.dc.b	$29,$B1,$98,$C1,$1C,$13,$0F,$ED,$73,$C1,$FE,$FF,$E3,$FE,$03,$CD
	.dc.b	$FE,$89,$5F,$DF,$A3,$07,$7D,$E4,$78,$DB,$5D,$1A,$FF,$69,$BF,$B9
	.dc.b	$3B,$3F,$C1,$3E,$B7,$72,$38,$3F,$C9,$BF,$3D,$F4,$BD,$43,$49,$C1
	.dc.b	$F9,$D8,$CB,$63,$33,$CA,$3F,$37,$EA,$FF,$02,$C7,$FB,$A3,$5F,$E8
	.dc.b	$CD,$7F,$A2,$36,$FC,$3C,$1D,$49,$7E,$66,$D7,$F9,$AF,$0F,$C2,$4E
	.dc.b	$49,$1F,$EB,$2B,$7C,$F7,$90,$93,$FE,$B0,$8C,$BF,$E6,$EF,$9E,$7C
	.dc.b	$E7,$BD,$B1,$98,$C1,$1C,$12,$FF,$CF,$D3,$C1,$FF,$40,$7C,$F3,$C7
	.dc.b	$F9,$4D,$AF,$F3,$EB,$67,$73,$F3,$FC,$57,$AD,$49,$7F,$E8,$48,$CB
	.dc.b	$E9,$F7,$7D,$04,$70,$7F,$A2,$BE,$79,$F3,$1F,$6D,$B1,$98,$C1,$FC
	.dc.b	$6F,$1D,$FC,$B5,$83,$49,$3F,$CA,$A7,$6D,$9D,$B6,$77,$3F,$EB,$DD
	.dc.b	$7F,$90,$FC,$1F,$C7,$8F,$F5,$F4,$60,$FE,$4B,$F0,$BF,$B4,$8F,$F6
	.dc.b	$C7,$6E,$09,$87,$FB,$65,$B1,$A8,$C1,$FE,$EE,$F9,$EF,$F5,$83,$74
	.dc.b	$FE,$63,$18,$9F,$F7,$CB,$5F,$F2,$63,$5F,$F0,$F3,$5F,$F0,$E3,$7C
	.dc.b	$03,$98,$FC,$05,$FC,$BE,$0F,$54,$7F,$5B,$6B,$FD,$89,$AF,$F6,$56
	.dc.b	$BC,$B0,$F9,$D1,$83,$76,$F8,$A7,$CF,$7F,$4C,$6D,$7F,$5E,$9D,$4E
	.dc.b	$4F,$EA,$3F,$1A,$F9,$EA,$4A,$E1,$19,$6C,$67,$1F,$EE,$96,$AD,$46
	.dc.b	$1A,$FF,$9D,$7E,$79,$DF,$FD,$81,$B8,$79,$5B,$3D,$1A,$F1,$9E,$04
	.dc.b	$9E,$0F,$E5,$7C,$49,$1F,$91,$6C,$FF,$7E,$E4,$FF,$41,$4E,$49,$19
	.dc.b	$9D,$F9,$5F,$E1,$8C,$9F,$24,$F9,$EF,$F0,$DF,$A4,$F8,$3F,$F0,$CE
	.dc.b	$7D,$32,$FF,$1D,$67,$FE,$63,$C3,$83,$FE,$1F,$BE,$13,$FE,$63,$F8
	.dc.b	$F7,$C6,$74,$C7,$FC,$D5,$18,$BE,$11,$8A,$D1,$C6,$AE,$6B,$E1,$3F
	.dc.b	$E3,$DA,$E7,$5A,$9F,$F1,$85,$F4,$AE,$4D,$BF,$B5,$F3,$DF,$E3,$88
	.dc.b	$FF,$56,$70,$A3,$15,$FE,$7F,$8C,$4F,$48,$C4,$E9,$FF,$40,$4E,$9F
	.dc.b	$F1,$95,$76,$AE,$11,$AA,$E1,$39,$8F,$F1,$86,$8F,$83,$E9,$1F,$C6
	.dc.b	$9B,$19,$BE,$73,$96,$C6,$63,$9C,$F0,$24,$E9,$F9,$7F,$31,$FE,$74
	.dc.b	$8F,$8D,$3C,$CF,$98,$7F,$A6,$E3,$4E,$2D,$F3,$DF,$C5,$F6,$FB,$F1
	.dc.b	$8E,$9F,$AF,$2F,$97,$B1,$31,$FA,$53,$C7,$83,$FD,$58,$EE,$33,$3F
	.dc.b	$80,$F5,$BF,$96,$E5,$E1,$6E,$5E,$05,$F3,$DB,$F0,$A6,$1F,$7E,$77
	.dc.b	$97,$FA,$E5,$DC,$66,$7F,$1F,$83,$E5,$DF,$3D,$F9,$88,$F8,$66,$5C
	.dc.b	$65,$F9,$EF,$9D,$7C,$F7,$AE,$BF,$D0,$17,$CB,$9F,$E6,$B6,$39,$B6
	.dc.b	$F9,$57,$19,$ED,$5C,$30,$EF,$A7,$FA,$2B,$0E,$0F,$FA,$37,$1F,$0E
	.dc.b	$DE,$A9,$CE,$1E,$2B,$E2,$DC,$3F,$49,$7C,$34,$ED,$12,$7C,$B5,$C2
	.dc.b	$B5,$18,$9D,$67,$FB,$E8,$F3,$9F,$88,$9D,$C6,$70,$F1,$CE,$A3,$17
	.dc.b	$D7,$3F,$81,$5A,$24,$67,$2F,$63,$EA,$BE,$59,$FC,$38,$D6,$7E,$48
	.dc.b	$C6,$76,$99,$F9,$27,$31,$EB,$26,$7F,$4B,$1E,$F7,$C6,$7B,$9C,$B3
	.dc.b	$F3,$E5,$E3,$CF,$CF,$5F,$32,$72,$F1,$8F,$A7,$97,$86,$FE,$91,$E0
	.dc.b	$CB,$C7,$97,$87,$3E,$19,$75,$3D,$B5,$B3,$8C,$FC,$AC,$7C,$19,$FA
	.dc.b	$27,$47,$A3,$1F,$03,$EA,$26,$5E,$4C,$FC,$F1,$C2,$31,$5C,$F0,$ED
	.dc.b	$3E,$1B,$E7,$5F,$02,$31,$7C,$AB,$64,$CF,$F2,$D1,$97,$DE,$49,$F9
	.dc.b	$98,$77,$8F,$AB,$1F,$EE,$3B,$FD,$43,$72,$FD,$A3,$72,$FF,$1A,$4F
	.dc.b	$F0,$FF,$42,$F0,$9A,$4A,$FF,$71,$CE,$DF,$D9,$CE,$DB,$3F,$A6,$F9
	.dc.b	$EF,$F2,$E4,$7E,$EA,$75,$A7,$E7,$5B,$97,$ED,$5B,$3F,$90,$AE,$2F
	.dc.b	$42,$57,$1B,$ED,$E1,$A7,$B4,$47,$78,$7E,$FE,$F9,$DF,$19,$C9,$FC
	.dc.b	$43,$0F,$0E,$CA,$DD,$F9,$70,$E9,$87,$13,$F8,$D4,$EB,$0E,$53,$AA
	.dc.b	$E1,$1F,$BD,$9D,$3F,$C9,$F0,$F7,$B6,$B8,$3F,$07,$1F,$63,$E6,$34
	.dc.b	$7A,$31,$F3,$46,$B1,$F2,$63,$E0,$6D,$73,$3A,$9C,$9C,$D7,$28,$CE
	.dc.b	$5F,$76,$73,$97,$A2,$31,$7D,$49,$5D,$4F,$34,$69,$F8,$67,$C1,$C3
	.dc.b	$CD,$3C,$23,$C4,$F2,$C3,$C5,$1C,$B2,$F8,$B7,$CE,$73,$8F,$B2,$B8
	.dc.b	$9D,$9F,$A7,$87,$68,$C6,$1D,$EF,$96,$3E,$BC,$7A,$93,$2F,$C5,$DF
	.dc.b	$C2,$8C,$3E,$93,$11,$F5,$B2,$F1,$E1,$D6,$B7,$97,$8A,$F6,$7D,$7C
	.dc.b	$3C,$18,$62,$75,$3F,$E0,$5E,$57,$C1,$7A,$79,$91,$89,$FF,$19,$35
	.dc.b	$E1,$F3,$DB,$8C,$5F,$F8,$31,$B9,$FE,$96,$F9,$E1,$BB,$ED,$F3,$5F
	.dc.b	$C1,$30,$FF,$09,$B6,$F8,$37,$16,$08,$FC,$AC,$7F,$87,$5D,$46,$78
	.dc.b	$3A,$9D,$A7,$EA,$1F,$E0,$7F,$86,$F4,$7C,$F4,$8F,$EF,$F1,$AE,$13
	.dc.b	$FA,$2E,$6F,$E3,$8D,$9F,$EF,$8D,$7F,$C1,$2D,$7B,$1F,$E4,$09,$E6
	.dc.b	$D8,$D4,$62,$3F,$CA,$3E,$47,$F8,$C5,$FF,$28,$E3,$FE,$A9,$E2,$EC
	.dc.b	$D8,$CC,$ED,$E3,$3A,$8C,$61,$FB,$AA,$E1,$5A,$3A,$F9,$6F,$8F,$FC
	.dc.b	$03,$63,$FC,$E0,$D7,$FC,$94,$F9,$F7,$FF,$24,$B9,$6C,$F2,$39,$E7
	.dc.b	$F2,$FE,$7B,$F8,$5F,$8D,$FE,$99,$F1,$5F,$15,$F8,$F6,$E5,$FE,$30
	.dc.b	$AD,$E9,$FE,$29,$7F,$DF,$51,$FE,$3F,$6E,$E6,$92,$3F,$A6,$B5,$FE
	.dc.b	$92,$FE,$21,$FF,$43,$7C,$F5,$7F,$D0,$6D,$DB,$FD,$90,$DF,$18,$D2
	.dc.b	$4F,$FB,$49,$AF,$FB,$43,$81,$F3,$DC,$DB,$1F,$ED,$F6,$BB,$8D,$11
	.dc.b	$D3,$5F,$F6,$DF,$90,$F2,$BC,$CD,$8F,$F7,$53,$5D,$C7,$C8,$E7,$7C
	.dc.b	$4D,$24,$FC,$18,$CB,$63,$2F,$C7,$8C,$B6,$3E,$37,$2F,$D8,$8F,$F7
	.dc.b	$7B,$5F,$C8,$35,$FE,$04,$D7,$F7,$6D,$D0,$D2,$70,$FD,$3F,$4D,$F3
	.dc.b	$D4,$97,$F6,$ED,$3A,$8C,$35,$FE,$0F,$1A,$9E,$31,$89,$D3,$FD,$83
	.dc.b	$A9,$F3,$DF,$02,$30,$D7,$FC,$01,$1A,$CB,$F1,$4D,$8F,$F0,$EF,$C9
	.dc.b	$FE,$58,$8E,$08,$7F,$87,$1C,$B4,$FC,$B7,$82,$F9,$E7,$AD,$F9,$78
	.dc.b	$C3,$63,$33,$C1,$AF,$F2,$69,$7C,$E3,$E0,$F9,$36,$7B,$B5,$E0,$D7
	.dc.b	$FC,$07,$2F,$29,$F3,$DE,$36,$CF,$A1,$AF,$85,$AF,$2F,$9E,$7C,$2F
	.dc.b	$AD,$B1,$F0,$1A,$FA,$5A,$F8,$B0,$7C,$17,$A8,$91,$EA,$6B,$EA,$6B
	.dc.b	$C5,$FD,$F7,$CF,$36,$A4,$9F,$C6,$46,$5F,$A7,$F3,$CF,$8F,$FA,$4D
	.dc.b	$8C,$C6,$08,$E0,$87,$D3,$8E,$0F,$E5,$25,$F2,$1E,$42,$47,$89,$EF
	.dc.b	$F0,$92,$3C,$6D,$7A,$35,$FA,$F4,$8C,$47,$F7,$3F,$8C,$7C,$C7,$F1
	.dc.b	$F6,$C7,$EA,$DA,$FE,$A1,$AF,$F9,$13,$E4,$5F,$3A,$F9,$EF,$E5,$51
	.dc.b	$86,$C6,$67,$F9,$84,$E5,$B1,$97,$FB,$DC,$65,$B1,$97,$C0,$D7,$D5
	.dc.b	$E8,$FA,$48,$FA,$5F,$0F,$CF,$48,$EE,$D7,$AB,$5E,$AD,$7F,$C3,$1F
	.dc.b	$28,$F1,$FF,$A2,$36,$3F,$88,$35,$FE,$1A,$D7,$E3,$78,$AF,$9E,$F8
	.dc.b	$AD,$8F,$86,$D7,$E0,$B5,$E3,$1C,$08,$F0,$7F,$AB,$7C,$F3,$E5,$29
	.dc.b	$27,$F9,$24,$65,$EF,$F0,$7E,$22,$38,$3E,$AC,$7F,$7F,$7E,$F3,$5F
	.dc.b	$F4,$F6,$A8,$C4,$23,$1E,$A7,$E0,$A7,$31,$89,$EB,$39,$3F,$B7,$7C
	.dc.b	$F7,$EB,$63,$F9,$AB,$5F,$F0,$C3,$5F,$F2,$53,$5F,$F2,$1B,$7C,$83
	.dc.b	$49,$D3,$FE,$08,$F9,$EF,$CC,$D7,$FA,$71,$AF,$FA,$E5,$A9,$A4,$E5
	.dc.b	$E4,$48,$FF,$04,$7C,$F7,$F5,$B4,$D2,$79,$7F,$E1,$8F,$5F,$FD,$97
	.dc.b	$3F,$E5,$86,$BF,$EA,$E6,$BF,$EC,$B6,$BF,$E5,$2F,$80,$E4,$47,$FC
	.dc.b	$31,$1A,$F9,$17,$CF,$3D,$AF,$F0,$E4,$61,$B1,$9A,$FF,$11,$4E,$5B
	.dc.b	$EF,$BE,$7B,$8B,$63,$31,$83,$FC,$66,$BD,$DA,$47,$FC,$6E,$FF,$75
	.dc.b	$FC,$2D,$F3,$6F,$7B,$FD,$D1,$9D,$BF,$F3,$ED,$6A,$72,$7F,$BB,$7E
	.dc.b	$15,$F8,$67,$F6,$28,$CD,$70,$C3,$96,$1C,$F9,$98,$F3,$AF,$89,$7F
	.dc.b	$E9,$38,$C5,$78,$71,$EB,$E9,$F2,$AD,$63,$70,$FC,$F7,$D0,$7F,$8A
	.dc.b	$A7,$F8,$DB,$63,$F8,$CB,$5F,$E4,$CD,$43,$E8,$5F,$64,$F8,$CE,$D1
	.dc.b	$87,$C1,$DA,$BE,$DA,$7C,$7A,$D6,$3F,$09,$B8,$FB,$8E,$D8,$F5,$9F
	.dc.b	$8F,$87,$3F,$67,$BC,$E6,$B8,$7D,$3F,$E2,$79,$FE,$03,$70,$F9,$EF
	.dc.b	$E3,$7B,$F3,$9F,$FA,$9A,$72,$DB,$E6,$D3,$EF,$57,$19,$DD,$F6,$CF
	.dc.b	$FA,$9C,$76,$6D,$FA,$FD,$A7,$DD,$3E,$FB,$87,$7C,$C3,$83,$F5,$EB
	.dc.b	$73,$9B,$E5,$F8,$2B,$E3,$FF,$21,$1F,$E3,$F8,$C6,$1F,$E8,$5B,$46
	.dc.b	$27,$FD,$9F,$18,$AE,$31,$F7,$34,$7D,$5B,$E7,$7E,$4B,$71,$FC,$BE
	.dc.b	$1C,$5F,$97,$19,$BF,$B1,$D4,$F3,$3F,$C4,$B5,$FD,$3B,$DB,$EF,$5F
	.dc.b	$EE,$29,$CB,$FD,$42,$33,$5C,$5B,$3A,$9F,$F6,$13,$5F,$EB,$B1,$D3
	.dc.b	$2F,$4F,$BF,$E8,$3F,$41,$7C,$F0,$F3,$C7,$E1,$21,$E8,$71,$6C,$FF
	.dc.b	$AB,$6B,$FD,$EF,$39,$24,$66,$77,$87,$FB,$EA,$B7,$39,$AF,$27,$D6
	.dc.b	$BE,$43,$F2,$6D,$9F,$C9,$E1,$CB,$9B,$F1,$B3,$9C,$7E,$5F,$AD,$FD
	.dc.b	$96,$FF,$C3,$9E,$87,$C6,$8C,$4F,$F8,$76,$30,$49,$D9,$23,$CD,$4F
	.dc.b	$1B,$C6,$7C,$6B,$E7,$87,$8A,$3F,$53,$F8,$6F,$FA,$63,$0F,$88,$DC
	.dc.b	$3E,$03,$5F,$E0,$A3,$54,$66,$3F,$1C,$D7,$4D,$8D,$FC,$37,$DB,$8C
	.dc.b	$7D,$87,$FB,$BA,$FF,$72,$DB,$FD,$FC,$F8,$F1,$FD,$AC,$E5,$FE,$55
	.dc.b	$EF,$FF,$19,$A6,$DF,$E7,$0F,$03,$F1,$F3,$98,$C6,$3F,$EB,$38,$C1
	.dc.b	$27,$67,$E3,$F0,$77,$FF,$0E,$6D,$7F,$3C,$BE,$78,$7F,$14,$9C,$BF
	.dc.b	$A6,$FC,$39,$F6,$1F,$A0,$6D,$FF,$B0,$DB,$3F,$EC,$06,$BF,$3A,$D1
	.dc.b	$88,$FE,$6B,$18,$AE,$84,$AF,$CE,$7D,$3B,$EE,$BF,$92,$B6,$3D,$18
	.dc.b	$78,$35,$FE,$3F,$39,$C7,$F9,$A7,$C4,$3E,$9F,$F8,$05,$7F,$61,$8C
	.dc.b	$77,$7E,$C3,$F0,$F7,$7F,$F5,$19,$CE,$7F,$D8,$17,$D4,$3D,$5F,$E0
	.dc.b	$B1,$86,$C6,$67,$FD,$96,$DC,$FF,$61,$F7,$3F,$CB,$B3,$FE,$6A,$DA
	.dc.b	$F8,$7F,$12,$F9,$4F,$E8,$73,$9B,$FC,$1F,$0F,$FB,$0A,$FF,$DE,$4D
	.dc.b	$8F,$F7,$7B,$5F,$A0,$D7,$E7,$B5,$E7,$F1,$BF,$C6,$A3,$1C,$9F,$3E
	.dc.b	$3F,$8B,$79,$3F,$DE,$2F,$F9,$83,$6F,$8C,$7F,$56,$8C,$76,$3E,$BF
	.dc.b	$EA,$4F,$F3,$76,$D7,$EB,$DF,$DC,$C6,$79,$FF,$B0,$C6,$AB,$94,$63
	.dc.b	$C1,$CB,$FD,$9B,$ED,$9F,$0D,$4C,$76,$DC,$7F,$B4,$CE,$9B,$3A,$69
	.dc.b	$FA,$1D,$9F,$5B,$FC,$C9,$B1,$FE,$20,$D1,$F3,$7F,$C8,$63,$E5,$D6
	.dc.b	$C8,$EE,$35,$3F,$62,$7F,$BE,$FD,$BF,$FA,$56,$BF,$BE,$B6,$7F,$A9
	.dc.b	$DF,$81,$FF,$02,$C6,$79,$7F,$BD,$FD,$2B,$E9,$BF,$C0,$ED,$C3,$FC
	.dc.b	$0E,$FF,$95,$1B,$1A,$23,$AF,$81,$B8,$7F,$84,$BE,$E1,$F4,$1F,$E1
	.dc.b	$26,$D7,$F8,$49,$7C,$B6,$66,$4E,$5E,$11,$8A,$FE,$DB,$F3,$3F,$DB
	.dc.b	$DF,$E0,$4D,$7F,$80,$B5,$FE,$3E,$DE,$93,$49,$E1,$7F,$70,$78,$60
	.dc.b	$FB,$9F,$E2,$FB,$7F,$97,$63,$1E,$47,$F6,$CA,$FE,$03,$1F,$E4,$43
	.dc.b	$27,$E1,$BA,$FF,$18,$7F,$72,$FA,$EF,$F3,$06,$7F,$E3,$6E,$7F,$D9
	.dc.b	$75,$9A,$7E,$D7,$55,$70,$CF,$FC,$63,$3C,$38,$DF,$5B,$F6,$0F,$EF
	.dc.b	$FF,$43,$8D,$3E,$CC,$65,$A7,$F3,$58,$D6,$7F,$E4,$8A,$B8,$7F,$8B
	.dc.b	$17,$C9,$7F,$82,$BA,$9F,$57,$FE,$7F,$F8,$9F,$D3,$E1,$FE,$64,$BD
	.dc.b	$E5,$FD,$06,$F6,$FF,$9A,$2B,$74,$AD,$BF,$E2,$8A,$DB,$FE,$93,$ED
	.dc.b	$FE,$D7,$3A,$CB,$FD,$09,$5A,$BF,$0E,$FF,$AA,$CB,$C9,$87,$F4,$2A
	.dc.b	$D3,$CB,$3F,$54,$EB,$2F,$EC,$7A,$7E,$7D,$F0,$1F,$DE,$73,$7D,$86
	.dc.b	$F2,$56,$F3,$F8,$3C,$3E,$BA,$D4,$E7,$3F,$F2,$66,$1C,$B6,$F8,$51
	.dc.b	$FD,$5A,$75,$97,$7B,$ED,$F5,$9F,$A4,$BD,$CF,$F7,$88,$C3,$E6,$CD
	.dc.b	$1F,$AC,$8C,$3A,$CF,$E2,$B7,$1F,$EC,$AD,$9E,$9D,$3E,$0E,$9F,$89
	.dc.b	$0F,$A6,$FF,$37,$79,$9E,$28,$C7,$BD,$FD,$71,$F8,$FF,$14,$FA,$AE
	.dc.b	$91,$AA,$E1,$7C,$A3,$E6,$99,$38,$CF,$08,$D6,$9F,$86,$6E,$9F,$E2
	.dc.b	$C6,$D3,$0F,$06,$5E,$89,$CC,$F5,$BE,$59,$79,$DF,$CB,$DF,$0D,$BC
	.dc.b	$B3,$9C,$7B,$36,$FF,$D7,$F8,$78,$2F,$84,$7A,$A3,$B9,$27,$84,$EB
	.dc.b	$0F,$0E,$7D,$89,$7F,$03,$2F,$45,$F7,$8F,$46,$3E,$2A,$FC,$0E,$1C
	.dc.b	$E3,$13,$E1,$AE,$18,$F9,$73,$F9,$51,$82,$61,$CA,$32,$F8,$B2,$F2
	.dc.b	$DF,$8F,$3C,$63,$DA,$72,$7C,$8B,$E3,$5E,$1B,$DB,$F4,$E3,$33,$C2
	.dc.b	$B9,$1F,$F0,$06,$32,$E2,$7E,$6E,$1F,$22,$73,$18,$BE,$B9,$7C,$18
	.dc.b	$C1,$E1,$26,$1E,$5C,$FE,$5D,$79,$27,$E5,$9E,$58,$CB,$C3,$1F,$15
	.dc.b	$F5,$8D,$93,$3F,$99,$5C,$30,$F6,$4F,$89,$F0,$67,$DE,$BF,$05,$8F
	.dc.b	$CF,$BE,$18,$FB,$F0,$F1,$E3,$DE,$F7,$1E,$33,$CD,$87,$3C,$FF,$5D
	.dc.b	$1F,$E8,$C3,$F5,$05,$9F,$D6,$36,$36,$FE,$9F,$47,$9F,$C6,$3F,$37
	.dc.b	$18,$71,$5F,$DD,$A7,$51,$FC,$06,$74,$F7,$78,$69,$FB,$56,$E5,$DE
	.dc.b	$BF,$2F,$F5,$5F,$E5,$BB,$FE,$03,$5A,$7C,$33,$83,$7E,$87,$FA,$A3
	.dc.b	$0E,$91,$99,$FD,$75,$71,$D1,$87,$59,$CD,$6E,$F8,$C7,$F1,$49,$D4
	.dc.b	$60,$98,$F9,$E7,$58,$F8,$63,$13,$A9,$F3,$9A,$3B,$46,$88,$FF,$13
	.dc.b	$AE,$F1,$88,$F9,$57,$C1,$D5,$F7,$7A,$DF,$77,$67,$73,$DA,$47,$C9
	.dc.b	$87,$C2,$9E,$53,$C2,$79,$65,$F6,$E3,$94,$63,$2F,$A5,$1B,$C4,$F9
	.dc.b	$0F,$31,$93,$CA,$EB,$0F,$04,$6A,$FB,$5F,$93,$2F,$83,$3C,$08,$73
	.dc.b	$AE,$2F,$AE,$FE,$35,$F2,$7B,$57,$29,$F1,$12,$74,$F9,$CE,$47,$67
	.dc.b	$35,$DA,$FE,$06,$5F,$46,$31,$87,$83,$1F,$87,$7C,$31,$F0,$3D,$B1
	.dc.b	$F0,$4F,$E4,$67,$2F,$02,$61,$F9,$A9,$F8,$91,$87,$E8,$DF,$79,$F1
	.dc.b	$57,$C6,$E1,$7D,$2F,$EF,$B3,$FF,$0D,$B6,$FF,$DB,$6D,$C3,$FD,$F2
	.dc.b	$FF,$03,$9D,$F0,$7E,$2E,$F9,$B6,$78,$47,$6A,$D7,$CD,$7F,$9D,$1F
	.dc.b	$F0,$6F,$3F,$F8,$CF,$CA,$FC,$A3,$63,$FC,$23,$4A,$E1,$1F,$E1,$16
	.dc.b	$E3,$FE,$22,$E8,$61,$C9,$B5,$CA,$B7,$9F,$F8,$D7,$E4,$FD,$1D,$86
	.dc.b	$9F,$9C,$E1,$7D,$47,$E5,$B0,$77,$39,$FF,$C0,$93,$98,$FF,$22,$DF
	.dc.b	$2A,$FF,$08,$CE,$6B,$71,$FE,$55,$9D,$3F,$DE,$E7,$4F,$F9,$07,$1B
	.dc.b	$BF,$F9,$5D,$BA,$7F,$96,$B1,$F1,$53,$7F,$8D,$F5,$1A,$B9,$3F,$EA
	.dc.b	$A7,$AC,$EA,$3F,$D9,$7F,$53,$DB,$C2,$34,$DF,$D4,$7C,$59,$D9,$0F
	.dc.b	$F4,$54,$6B,$5F,$4E,$CF,$84,$F3,$3F,$E9,$56,$D7,$C9,$6F,$33,$EA
	.dc.b	$3E,$BC,$FC,$EA,$FE,$0B,$39,$E5,$FE,$8A,$D9,$FE,$98,$D7,$FA,$B3
	.dc.b	$5F,$B0,$D7,$9D,$76,$72,$D7,$33,$FD,$4B,$85,$E6,$3D,$EF,$A6,$DA
	.dc.b	$FF,$65,$46,$31,$CB,$23,$27,$FB,$3E,$3E,$CF,$93,$23,$FD,$BA,$FD
	.dc.b	$CD,$71,$1F,$EA,$38,$BC,$7F,$EE,$69,$FC,$1F,$A0,$FB,$1A,$4B,$FE
	.dc.b	$57,$5F,$84,$E3,$7E,$14,$F9,$6F,$C0,$B6,$B7,$9F,$F3,$2A,$DD,$68
	.dc.b	$FC,$9C,$FF,$9F,$BE,$A3,$0F,$03,$F2,$17,$CD,$FB,$06,$27,$8F,$90
	.dc.b	$AF,$C7,$78,$52,$B9,$70,$FF,$B4,$63,$F3,$FF,$62,$F5,$B5,$5F,$ED
	.dc.b	$79,$BA,$FE,$82,$D7,$BA,$E7,$ED,$3D,$6F,$D4,$B6,$33,$18,$F5,$F2
	.dc.b	$C8,$C9,$CA,$3F,$63,$C9,$F9,$B8,$F3,$B5,$F7,$E0,$8C,$63,$F4,$77
	.dc.b	$7C,$37,$D0,$3F,$C3,$BC,$3F,$51,$B1,$AC,$3F,$C4,$D3,$96,$C6,$5F
	.dc.b	$F1,$94,$65,$B1,$97,$F8,$B4,$65,$FE,$31,$87,$FA,$97,$EB,$BF,$85
	.dc.b	$F9,$E6,$9F,$C1,$3E,$E9,$EC,$7F,$12,$E1,$7D,$63,$E3,$FF,$C6,$B1
	.dc.b	$86,$C6,$72,$FF,$1F,$4E,$5B,$F0,$6F,$89,$E8,$D8,$FC,$54,$A3,$11
	.dc.b	$F8,$AD,$5F,$88,$7F,$08,$D9,$F1,$2F,$B9,$FE,$4A,$DB,$FC,$5B,$5E
	.dc.b	$5F,$88,$BE,$07,$F9,$0B,$63,$E7,$B5,$F9,$AD,$7E,$5B,$5F,$D3,$35
	.dc.b	$FE,$60,$DC,$4D,$35,$87,$F0,$9F,$BC,$FE,$77,$E2,$1A,$7F,$34,$FB
	.dc.b	$07,$D8,$53,$F0,$EE,$31,$FC,$2B,$5F,$C1,$B5,$FC,$0B,$5F,$DE,$35
	.dc.b	$FE,$8C,$DC,$4E,$70,$FE,$3D,$EE,$F9,$89,$5F,$B6,$F9,$0F,$DF,$75
	.dc.b	$FE,$C5,$FE,$01,$1F,$8E,$6B,$FE,$3C,$A4,$62,$11,$8F,$13,$F9,$74
	.dc.b	$E6,$31,$3F,$E8,$76,$BF,$E9,$16,$BF,$9C,$F9,$3F,$D8,$3F,$C2,$9A
	.dc.b	$FE,$D1,$B7,$FC,$8F,$E4,$FF,$A8,$E6,$5E,$9F,$ED,$1B,$7F,$B2,$6B
	.dc.b	$FB,$06,$BF,$AE,$6B,$FC,$A5,$AF,$F7,$36,$CF,$79,$FE,$99,$7F,$DB
	.dc.b	$AB,$D1,$E5,$FF,$7E,$8C,$47,$F0,$CA,$7D,$7F,$F8,$09,$B1,$F1,$5A
	.dc.b	$FC,$56,$BF,$C0,$1A,$FD,$36,$BF,$49,$B9,$7F,$82,$67,$FB,$04,$63
	.dc.b	$E6,$1F,$79,$FE,$12,$6D,$FF,$38,$6B,$FB,$7F,$AC,$7C,$7F,$F8,$39
	.dc.b	$B1,$FC,$71,$AF,$F1,$A6,$BF,$C6,$1A,$FF,$53,$6B,$FE,$1D,$6D,$9A
	.dc.b	$6B,$15,$BE,$4B,$D0,$4B,$FE,$B1,$EA,$BE,$6F,$FA,$4C,$61,$AF,$F8
	.dc.b	$29,$AF,$F8,$29,$AF,$F8,$51,$AF,$E7,$3E,$17,$F9,$63,$FD,$71,$AF
	.dc.b	$F4,$66,$F1,$7F,$8E,$2B,$F9,$97,$CC,$FF,$8F,$A7,$3E,$9F,$EA,$5B
	.dc.b	$7D,$7C,$8F,$F2,$2D,$F0,$8F,$ED,$3F,$4E,$FA,$DF,$F2,$3B,$63,$31
	.dc.b	$8F,$B6,$CB,$23,$27,$28,$FF,$28,$FD,$0B,$EE,$DD,$B8,$6D,$7F,$80
	.dc.b	$67,$53,$FE,$06,$FA,$3F,$F7,$26,$FF,$E5,$49,$CD,$6E,$F8,$BF,$E0
	.dc.b	$68,$CB,$FE,$08,$15,$CB,$6F,$F3,$83,$FE,$EF,$8F,$F3,$91,$AF,$8D
	.dc.b	$AE,$9A,$FF,$B3,$23,$83,$FE,$1D,$CF,$FC,$0C,$FF,$4D,$6E,$3F,$E2
	.dc.b	$76,$DF,$F9,$FF,$D5,$7C,$27,$F9,$A2,$73,$7F,$E6,$F6,$F4,$7D,$1F
	.dc.b	$67,$FC,$4F,$7C,$2B,$FC,$4C,$3E,$AF,$FC,$29,$97,$99,$B1,$EA,$6B
	.dc.b	$EE,$6B,$FE,$91,$5F,$40,$E2,$E4,$D9,$E4,$D7,$93,$5F,$8F,$F1,$9F
	.dc.b	$19,$FF,$78,$B5,$34,$F9,$BC,$CF,$AB,$FF,$25,$6F,$86,$D7,$D4,$9D
	.dc.b	$78,$DF,$E9,$48,$FF,$54,$46,$2B,$84,$E6,$7F,$DF,$95,$C9,$DB,$5C
	.dc.b	$C6,$D0,$FC,$4B,$F8,$AD,$7E,$01,$B1,$F8,$46,$BF,$88,$6B,$CB,$EA
	.dc.b	$1F,$53,$FE,$C8,$AF,$C5,$C6,$3E,$1B,$C0,$D8,$CC,$EE,$BF,$27,$39
	.dc.b	$6C,$65,$FF,$64,$46,$5B,$19,$7F,$B3,$6F,$FE,$30,$AF,$EC,$0D,$8F
	.dc.b	$EC,$0D,$7F,$C7,$6D,$78,$FB,$3F,$17,$C2,$FF,$6A,$B6,$75,$3F,$DC
	.dc.b	$1B,$DA,$FA,$87,$D4,$7C,$D8,$C3,$63,$38,$FF,$AB,$9B,$7F,$EA,$E6
	.dc.b	$BF,$EA,$E6,$BF,$EB,$86,$AA,$EB,$3F,$EB,$8F,$C4,$FF,$D8,$CD,$F8
	.dc.b	$BF,$14,$62,$BF,$D7,$6D,$7F,$99,$7C,$BF,$9F,$C5,$34,$97,$F6,$BE
	.dc.b	$F3,$FC,$B0,$DF,$7F,$ED,$C6,$2F,$EC,$B5,$FA,$ED,$7F,$0C,$D5,$C2
	.dc.b	$8C,$36,$7F,$2D,$DA,$FA,$AE,$0D,$8C,$C6,$0E,$9A,$BF,$00,$FF,$5E
	.dc.b	$65,$FE,$DC,$6E,$5F,$ED,$C6,$BF,$ED,$C6,$BF,$EE,$66,$BF,$3D,$B3
	.dc.b	$D7,$E3,$3E,$F4,$6B,$91,$97,$CD,$BF,$F0,$37,$07,$E2,$E3,$11,$FE
	.dc.b	$07,$8F,$D1,$F9,$0F,$AE,$A4,$CF,$F5,$31,$96,$C6,$6B,$F4,$F3,$96
	.dc.b	$C7,$C8,$F5,$7F,$C1,$35,$FE,$31,$F4,$30,$47,$FD,$4B,$F8,$5B,$C9
	.dc.b	$F9,$ED,$8C,$C6,$08,$E0,$98,$7F,$AA,$A7,$83,$FE,$4D,$F5,$FF,$D6
	.dc.b	$F3,$FE,$51,$6C,$67,$E2,$3F,$5D,$18,$8F,$E2,$63,$E0,$FF,$D2,$BC
	.dc.b	$9F,$C5,$DB,$7F,$EA,$4F,$80,$FC,$54,$62,$BF,$AD,$B5,$E6,$D7,$F8
	.dc.b	$FB,$5E,$D1,$F8,$CF,$84,$FE,$15,$1F,$69,$BC,$3F,$EA,$FF,$07,$F2
	.dc.b	$71,$8F,$B1,$7A,$9F,$96,$6E,$5F,$ED,$06,$E1,$FE,$CE,$6B,$FE,$CC
	.dc.b	$6B,$FC,$CB,$04,$62,$3F,$16,$D8,$CD,$7E,$93,$EB,$5F,$51,$49,$87
	.dc.b	$FB,$82,$32,$FF,$B8,$BE,$B3,$F8,$43,$7A,$FF,$5B,$18,$CB,$FD,$C6
	.dc.b	$D7,$E4,$FD,$6B,$EA,$7E,$93,$6F,$FD,$AA,$D9,$FA,$2D,$7F,$DD,$3F
	.dc.b	$15,$FE,$49,$9F,$F7,$9B,$63,$FD,$E2,$D7,$7E,$DF,$F5,$55,$F8,$87
	.dc.b	$C3,$6E,$B9,$8C,$11,$C1,$2F,$FC,$33,$1F,$27,$D5,$7D,$67,$F0,$C6
	.dc.b	$CF,$F8,$83,$81,$18,$C3,$FC,$4B,$18,$AF,$E8,$9F,$27,$FA,$C9,$FF
	.dc.b	$55,$35,$FC,$CF,$82,$8C,$47,$FA,$E5,$AF,$FA,$E1,$B1,$BE,$DE,$A4
	.dc.b	$C7,$FD,$71,$F3,$9F,$94,$93,$49,$CF,$FD,$5D,$B3,$C2,$30,$7C,$BF
	.dc.b	$A0,$FE,$23,$89,$A4,$F0,$F4,$4A,$FF,$20,$C6,$5F,$8D,$AB,$EB,$3E
	.dc.b	$51,$23,$2F,$F6,$9F,$9F,$7B,$74,$93,$FE,$D3,$8C,$B6,$33,$8F,$FB
	.dc.b	$56,$72,$D8,$FE,$E9,$F3,$9F,$E6,$AA,$FF,$6E,$36,$7F,$DB,$8D,$7E
	.dc.b	$9B,$5F,$F7,$53,$5F,$F7,$4B,$75,$34,$D4,$2D,$F6,$3F,$95,$6D,$FF
	.dc.b	$99,$9B,$3F,$E6,$A6,$BF,$E6,$E6,$BF,$C7,$9A,$FF,$8C,$9A,$F7,$9F
	.dc.b	$EC,$3F,$5B,$FA,$B8,$FE,$0C,$D7,$FA,$7B,$78,$CD,$FA,$DF,$F2,$04
	.dc.b	$FF,$54,$6C,$9A,$4F,$67,$FB,$7B,$71,$FD,$23,$C3,$EA,$9F,$5B,$FD
	.dc.b	$D1,$B3,$FA,$96,$E1,$FA,$86,$BF,$CF,$30,$46,$23,$F9,$03,$5F,$ED
	.dc.b	$2D,$8D,$D7,$CF,$FA,$D9,$7F,$3D,$B7,$F3,$DA,$FC,$F6,$BF,$3D,$47
	.dc.b	$F9,$83,$D9,$B3,$FD,$E3,$8D,$F5,$B8,$F8,$07,$C8,$F1,$C3,$81,$FC
	.dc.b	$DE,$B6,$FF,$A0,$23,$D7,$F3,$FD,$23,$E0,$6A,$8C,$63,$A9,$FF,$38
	.dc.b	$9A,$8C,$C7,$C2,$6E,$BF,$21,$B5,$CE,$B8,$D7,$33,$FD,$0D,$4F,$A2
	.dc.b	$FD,$26,$5E,$A9,$FF,$0A,$C6,$31,$FE,$A0,$BE,$B3,$E0,$46,$3E,$36
	.dc.b	$D7,$06,$E1,$D5,$B7,$C7,$1F,$85,$5B,$6A,$FA,$9F,$F2,$6F,$93,$86
	.dc.b	$C7,$FA,$9B,$1F,$F4,$9F,$C3,$3E,$83,$CD,$3F,$E9,$6A,$DE,$BE,$B8
	.dc.b	$E9,$EF,$FF,$98,$DE,$46,$DC,$47,$C1,$BE,$9A,$7B,$2B,$8E,$5E,$3B
	.dc.b	$DE,$5F,$E1,$2C,$7B,$BE,$4C,$BC,$77,$C0,$EE,$5C,$BC,$95,$C2,$19
	.dc.b	$78,$33,$ED,$94,$9E,$35,$BA,$D4,$63,$2F,$1B,$EB,$AE,$18,$7A,$F1
	.dc.b	$F1,$E1,$DE,$BC,$59,$7A,$70,$E3,$8F,$4C,$3C,$B1,$D2,$B6,$75,$3E
	.dc.b	$4C,$EE,$B9,$63,$E0,$6C,$EB,$1F,$0D,$F3,$C7,$C4,$7A,$E3,$37,$C6
	.dc.b	$BB,$63,$E9,$CF,$E9,$5F,$0C,$39,$1E,$8C,$38,$DF,$C3,$C3,$96,$5E
	.dc.b	$58,$E7,$97,$93,$2F,$6D,$7E,$4A,$B5,$5E,$E2,$63,$E5,$CB,$C3,$97
	.dc.b	$82,$FA,$E5,$E0,$C7,$9E,$1E,$13,$E0,$5F,$0A,$F6,$47,$CA,$C7,$BE
	.dc.b	$3E,$B9,$E7,$5B,$AE,$57,$D0,$F8,$59,$F2,$BE,$13,$DA,$31,$97,$96
	.dc.b	$30,$72,$CF,$E0,$11,$FC,$7E,$7E,$98,$C4,$79,$63,$0F,$DF,$3F,$CD
	.dc.b	$E7,$E6,$CB,$7F,$B4,$6D,$7E,$B5,$CE,$DF,$E7,$57,$89,$FB,$D7,$0E
	.dc.b	$BD,$2C,$B9,$6D,$7F,$0F,$7F,$81,$4F,$3C,$BF,$52,$DC,$7F,$81,$FD
	.dc.b	$67,$FA,$57,$0D,$DC,$7F,$53,$5B,$7F,$5F,$3B,$73,$3B,$BE,$AF,$E1
	.dc.b	$67,$51,$89,$D9,$1F,$E1,$11,$AB,$FE,$09,$7D,$EB,$D8,$6E,$3C,$FE
	.dc.b	$C7,$FA,$6A,$FA,$12,$38,$E1,$E1,$9D,$46,$2B,$91,$FC,$7E,$FD,$9B
	.dc.b	$B1,$F1,$4E,$70,$F6,$5F,$1C,$7C,$F8,$F4,$AD,$1D,$0F,$3C,$69,$E2
	.dc.b	$69,$F3,$3F,$2A,$39,$3C,$27,$B1,$1D,$1C,$8D,$1C,$DF,$86,$E7,$2F
	.dc.b	$C0,$3E,$EC,$7D,$95,$DA,$73,$7E,$2A,$E7,$5A,$79,$65,$F4,$6B,$78
	.dc.b	$76,$9F,$74,$61,$F1,$1E,$63,$33,$C4,$F9,$B9,$7C,$C8,$F4,$92,$79
	.dc.b	$5F,$86,$38,$E5,$F2,$23,$18,$F9,$EB,$C1,$5E,$8B,$E9,$18,$26,$1E
	.dc.b	$0C,$BE,$03,$E3,$C7,$A5,$70,$8C,$5F,$C7,$8C,$4F,$C2,$3F,$15,$1F
	.dc.b	$DC,$38,$5F,$59,$FB,$7C,$BF,$C0,$11,$8E,$1F,$C2,$36,$B8,$57,$F0
	.dc.b	$2A,$DB,$C7,$5F,$F6,$A3,$65,$5C,$5F,$ED,$EF,$81,$EB,$F5,$9F,$E0
	.dc.b	$B7,$FC,$2A,$70,$70,$FF,$85,$9B,$C3,$FE,$45,$6C,$7F,$B8,$DA,$F1
	.dc.b	$D0,$F4,$A4,$E9,$B3,$A9,$FF,$1E,$9F,$E0,$9F,$AC,$C7,$4F,$F9,$02
	.dc.b	$BF,$D4,$5F,$57,$F1,$3A,$4F,$E4,$EB,$79,$7F,$80,$DF,$E3,$78,$7F
	.dc.b	$86,$DB,$C4,$69,$23,$F8,$FB,$5F,$E4,$F1,$B7,$FD,$A1,$9B,$E4,$7F
	.dc.b	$93,$E5,$EC,$6C,$66,$77,$18,$23,$82,$1F,$ED,$8F,$AC,$F5,$BF,$ED
	.dc.b	$D6,$B9,$CB,$BC,$62,$3F,$CE,$CD,$7E,$7D,$A3,$11,$F8,$26,$C6,$67
	.dc.b	$F9,$FF,$0B,$EB,$1E,$5E,$1B,$87,$F4,$89,$D4,$E4,$F9,$7F,$69,$27
	.dc.b	$FA,$3B,$5F,$92,$D7,$FA,$BB,$5F,$F5,$EE,$DF,$D4,$3C,$8F,$F5,$DB
	.dc.b	$7C,$8F,$C1,$CF,$FB,$12,$30,$47,$FD,$96,$D7,$FD,$A5,$F3,$DC,$FC
	.dc.b	$93,$49,$1F,$ED,$B6,$BC,$0C,$1F,$D9,$B5,$7D,$4B,$C8,$FB,$ED,$9D
	.dc.b	$35,$FE,$EB,$3C,$27,$F2,$3E,$13,$EB,$29,$2B,$FB,$7C,$65,$B1,$97
	.dc.b	$FB,$BC,$65,$B1,$F1,$FE,$B1,$E7,$D2,$4F,$F7,$ED,$93,$A8,$C3,$5F
	.dc.b	$F0,$5C,$6A,$78,$C6,$27,$4F,$ED,$65,$E9,$FF,$A9,$63,$B3,$5E,$6D
	.dc.b	$78,$35,$FE,$07,$1C,$BC,$8F,$C0,$4E,$65,$39,$7A,$CF,$D0,$F8,$39
	.dc.b	$1F,$41,$AF,$F8,$73,$F0,$FF,$DF,$A6,$91,$FE,$06,$D7,$FC,$3E,$D7
	.dc.b	$EC,$FD,$63,$E4,$FF,$7A,$D8,$FC,$23,$5D,$C7,$F8,$A3,$C5,$7B,$8F
	.dc.b	$83,$C3,$6B,$FC,$63,$3A,$9C,$9F,$C2,$BE,$AF,$FD,$4D,$3F,$EF,$16
	.dc.b	$BF,$BC,$6B,$FC,$F1,$BC,$6F,$A9,$7D,$57,$F1,$46,$D7,$E6,$DA,$FE
	.dc.b	$59,$AF,$FB,$FF,$F0,$37,$D1,$7F,$1C,$6C,$7E,$8D,$AF,$E7,$9A,$F1
	.dc.b	$8E,$04,$78,$3F,$C9,$BE,$A1,$F5,$34,$93,$FA,$F8,$CB,$F1,$BE,$ED
	.dc.b	$F5,$3F,$35,$B1,$98,$C1,$1C,$10,$FF,$33,$EC,$9E,$31,$FA,$0F,$8F
	.dc.b	$FD,$C3,$F9,$D6,$BF,$96,$6B,$F9,$46,$EA,$F9,$07,$D0,$7F,$42,$6D
	.dc.b	$7F,$09,$6B,$FC,$11,$AF,$F5,$F9,$E5,$3C,$49,$1C,$5F,$E9,$BF,$25
	.dc.b	$F7,$89,$E9,$E0,$FF,$3B,$FC,$C3,$EE,$9F,$0D,$86,$E1,$FE,$80,$9D
	.dc.b	$4E,$49,$19,$9D,$BF,$E8,$E9,$DB,$99,$FE,$6D,$8F,$CA,$8F,$E0,$8D
	.dc.b	$7F,$7A,$D7,$F7,$6D,$F2,$DF,$2E,$F8,$7F,$EA,$2D,$AF,$E9,$CD,$7F
	.dc.b	$D3,$F3,$C2,$76,$48,$DB,$5F,$E6,$3F,$52,$F8,$AF,$EE,$AD,$8F,$E7
	.dc.b	$8D,$77,$1F,$2B,$E1,$1F,$4B,$49,$3F,$26,$32,$D8,$CB,$F1,$E3,$2D
	.dc.b	$8F,$ED,$BF,$3B,$FC,$CA,$3F,$B5,$35,$F6,$B5,$FF,$11,$62,$8C,$47
	.dc.b	$F8,$AA,$1F,$3C,$F3,$FF,$BC,$B6,$7F,$BB,$35,$F9,$2D,$7A,$F5,$BE
	.dc.b	$43,$FC,$02,$D8,$FE,$FC,$D7,$E8,$B5,$E3,$1C,$08,$F0,$7F,$BD,$7D
	.dc.b	$23,$E7,$29,$27,$FA,$A4,$65,$B1,$97,$FA,$9C,$65,$B1,$FE,$3A,$FA
	.dc.b	$6F,$F6,$34,$7F,$B8,$9A,$FF,$92,$5A,$FF,$94,$1A,$9A,$4F,$85,$E8
	.dc.b	$7F,$91,$BE,$1F,$F9,$B4,$FF,$B9,$9A,$FF,$BC,$1A,$FF,$95,$DA,$FF
	.dc.b	$97,$BE,$43,$FC,$80,$FF,$BB,$9A,$FF,$BE,$1A,$FF,$97,$DA,$F1,$F7
	.dc.b	$BF,$CA,$F1,$88,$34,$C7,$D3,$63,$E1,$3F,$DE,$ED,$9F,$F7,$EB,$5F
	.dc.b	$F3,$60,$70,$FF,$9B,$DF,$E3,$39,$F2,$EE,$34,$93,$FE,$6A,$F8,$7F
	.dc.b	$ED,$A4,$70,$76,$F8,$97,$D3,$7F,$89,$FD,$DF,$0C,$A7,$3E,$DF,$F8
	.dc.b	$F2,$31,$5B,$79,$35,$FF,$42,$4F,$0A,$E6,$D8,$DD,$73,$8C,$47,$C0
	.dc.b	$FA,$67,$D5,$77,$84,$62,$FC,$B1,$89,$D4,$E6,$3F,$91,$7D,$3F,$F9
	.dc.b	$92,$BF,$95,$36,$33,$F0,$7E,$BF,$11,$39,$9F,$F4,$E4,$62,$75,$39
	.dc.b	$9F,$F5,$2C,$E6,$BA,$D7,$C5,$F6,$9C,$3F,$E1,$39,$FF,$1E,$36,$FF
	.dc.b	$C7,$2D,$7F,$C6,$CD,$7F,$C5,$ED,$7F,$C5,$AD,$9E,$BE,$B7,$78,$D7
	.dc.b	$6B,$E9,$9F,$4F,$F5,$E3,$0D,$8C,$E3,$FE,$85,$6E,$3F,$CE,$1A,$FF
	.dc.b	$1C,$6B,$FC,$6D,$AA,$BA,$CF,$F1,$AF,$23,$C1,$DF,$FD,$72,$30,$4A
	.dc.b	$FE,$53,$D5,$E5,$F0,$4D,$24,$FF,$61,$FA,$77,$D2,$7D,$76,$C6,$63
	.dc.b	$04,$70,$4B,$FE,$C9,$3C,$1F,$F3,$FF,$D3,$D7,$FB,$E3,$5F,$F5,$9B
	.dc.b	$5F,$F7,$9B,$73,$FE,$C7,$F4,$FF,$37,$CB,$34,$98,$7F,$64,$6D,$FF
	.dc.b	$64,$9D,$CE,$4F,$F4,$07,$99,$F1,$E7,$FC,$FA,$D7,$FB,$BE,$DF,$8D
	.dc.b	$8C,$47,$E4,$BC,$47,$D3,$FE,$6D,$B7,$FE,$09,$6C,$FF,$81,$5A,$FE
	.dc.b	$79,$AF,$39,$E4,$49,$FF,$56,$7A,$1F,$DD,$5B,$E6,$7F,$03,$C1,$39
	.dc.b	$BF,$F5,$5B,$5F,$F1,$47,$A2,$F1,$F8,$37,$0F,$F1,$84,$60,$8E,$09
	.dc.b	$3F,$E3,$38,$FD,$77,$A4,$F8,$0A,$4A,$FF,$55,$46,$5F,$F3,$CF,$D3
	.dc.b	$FD,$BF,$58,$D2,$72,$FF,$1D,$8C,$B6,$33,$3C,$A3,$E0,$7B,$DF,$C0
	.dc.b	$F2,$6B,$FE,$89,$EF,$C9,$2F,$FD,$A9,$F4,$FF,$DE,$5B,$E7,$FC,$98
	.dc.b	$C5,$FF,$B4,$9A,$FF,$93,$DA,$8E,$EE,$1B,$1C,$CF,$A9,$F4,$CF,$A7
	.dc.b	$A4,$BF,$F7,$34,$65,$FF,$74,$7D,$3F,$EE,$DB,$D5,$F7,$63,$18,$7F
	.dc.b	$B9,$9A,$F8,$C6,$EF,$A6,$7D,$3F,$CB,$6E,$1F,$EA,$A6,$D7,$F9,$F9
	.dc.b	$AF,$F9,$B7,$E1,$5F,$4F,$F8,$D6,$D7,$FB,$F5,$B3,$FE,$F8,$6B,$FD
	.dc.b	$A3,$E9,$FE,$22,$69,$36,$FE,$98,$DA,$FE,$96,$D9,$FE,$E3,$C3,$FE
	.dc.b	$F6,$FA,$7F,$CF,$7C,$71,$A4,$BF,$F7,$5F,$67,$F4,$1F,$8D,$FD,$06
	.dc.b	$1F,$EE,$86,$BF,$EE,$66,$F9,$9F,$EF,$49,$E0,$4A,$E6,$7F,$72,$87
	.dc.b	$D3,$FE,$3D,$BC,$FF,$D3,$E3,$04,$70,$49,$E5,$1F,$97,$C1,$F3,$9F
	.dc.b	$C2,$9F,$C6,$35,$CF,$87,$FD,$C6,$30,$47,$04,$AF,$F0,$3E,$AD,$7F
	.dc.b	$D0,$D1,$86,$DF,$CC,$6B,$FE,$AB,$6B,$F9,$DC,$BF,$CC,$DE,$D7,$F6
	.dc.b	$BF,$7B,$FC,$57,$7F,$0D,$B5,$FE,$26,$AD,$CE,$6B,$81,$1C,$1F,$2F
	.dc.b	$E5,$5F,$43,$FD,$DD,$B1,$98,$C1,$1C,$12,$FC,$13,$C1,$FF,$13,$E0
	.dc.b	$FA,$7F,$F0,$C3,$6B,$FC,$2E,$D9,$DC,$E8,$93,$A6,$C7,$E4,$BE,$9F
	.dc.b	$F4,$AF,$F0,$1D,$11,$88,$FE,$22,$D8,$CF,$4F,$F3,$5F,$A7,$FF,$1A
	.dc.b	$57,$16,$BF,$C6,$1B,$26,$93,$DB,$E4,$4C,$7E,$FF,$D2,$BE,$9F,$FA
	.dc.b	$13,$63,$31,$82,$38,$26,$1F,$AE,$9E,$0F,$CA,$F8,$7F,$92,$49,$F9
	.dc.b	$2D,$8D,$46,$08,$E0,$FF,$18,$BF,$1F,$91,$F4,$E8,$FE,$55,$F0,$2F
	.dc.b	$8C,$FE,$35,$C2,$C7,$58,$F2,$3F,$9E,$E3,$CB,$0E,$37,$F4,$F2,$7C
	.dc.b	$BF,$09,$FE,$49,$BA,$73,$1F,$CE,$4C,$B8,$C3,$B9,$FD,$29,$C4,$6A
	.dc.b	$BA,$4F,$76,$BB,$6D,$77,$6F,$CD,$7E,$57,$4F,$75,$7F,$56,$8C,$63
	.dc.b	$F8,$B5,$F4,$9F,$02,$B7,$82,$77,$7C,$79,$B5,$87,$16,$C7,$5C,$2D
	.dc.b	$F1,$9C,$CB,$E8,$FF,$C7,$FE,$47,$CA,$8F,$EB,$91,$85,$97,$A3,$E1
	.dc.b	$5F,$4F,$E2,$9E,$5E,$E7,$F9,$45,$DE,$9F,$D4,$F0,$E5,$87,$C0,$9D
	.dc.b	$E3,$DF,$1E,$79,$79,$2F,$78,$71,$C7,$A6,$3C,$67,$38,$FF,$A8,$F2
	.dc.b	$F1,$46,$23,$A9,$E2,$3D,$75,$CA,$FF,$BC,$63,$E0,$DF,$E0,$57,$0C
	.dc.b	$7B,$5E,$E3,$8E,$7E,$7C,$BB,$CE,$67,$86,$3E,$7A,$D6,$7E,$BA,$D6
	.dc.b	$7E,$53,$D3,$9F,$87,$0F,$31,$C8,$EE,$62,$37,$8F,$83,$0F,$3C,$73
	.dc.b	$BF,$35,$F3,$CB,$D3,$5A,$9F,$09,$AB,$E5,$87,$58,$CC,$FA,$33,$E0
	.dc.b	$C9,$D3,$96,$92,$FB,$BF,$3F,$0F,$16,$7F,$0E,$34,$D9,$FC,$56,$3E
	.dc.b	$27,$89,$23,$EA,$C6,$5E,$04,$9F,$C0,$67,$F3,$70,$F3,$5F,$19,$CE
	.dc.b	$38,$BE,$11,$F3,$71,$F2,$BF,$3A,$B8,$39,$C7,$C0,$E6,$BC,$15,$C2
	.dc.b	$3E,$B1,$33,$FA,$B9,$78,$E3,$DE,$71,$3C,$66,$0E,$19,$FD,$38,$E1
	.dc.b	$8F,$8A,$B9,$E5,$E9,$BF,$C1,$9F,$E3,$D3,$F3,$A5,$8F,$DB,$36,$FF
	.dc.b	$64,$E7,$5F,$F2,$1B,$C4,$FE,$00,$E1,$D7,$95,$FE,$FA,$AF,$E2,$2D
	.dc.b	$9E,$38,$7E,$C5,$B9,$FE,$F5,$B6,$CD,$F4,$FF,$E8,$D8,$FC,$BC,$E9
	.dc.b	$F0,$E1,$83,$F8,$14,$EE,$3F,$41,$3B,$C7,$C7,$18,$AD,$FA,$BF,$E5
	.dc.b	$FC,$BC,$91,$CE,$74,$7E,$4F,$0E,$93,$A8,$CB,$4F,$E3,$91,$AC,$BD
	.dc.b	$95,$CE,$75,$1F,$C2,$27,$55,$FC,$9B,$1F,$31,$2F,$CB,$1C,$6B,$9B
	.dc.b	$E0,$79,$61,$E9,$73,$87,$69,$E9,$5C,$9E,$06,$0B,$97,$CE,$BE,$51
	.dc.b	$9A,$E4,$FC,$E3,$E1,$46,$9C,$46,$B2,$F9,$AF,$C2,$9D,$46,$30,$F2
	.dc.b	$92,$BE,$F6,$5F,$09,$E7,$19,$69,$AA,$E7,$97,$C5,$8F,$4D,$71,$BE
	.dc.b	$D1,$E9,$BE,$D8,$FB,$63,$E5,$19,$23,$97,$E2,$9E,$03,$C2,$E7,$1F
	.dc.b	$76,$1E,$17,$DD,$3D,$0E,$17,$E2,$AE,$17,$EB,$8F,$2C,$F4,$C3,$D1
	.dc.b	$87,$3A,$F3,$C6,$32,$E9,$7C,$E3,$17,$E6,$8C,$65,$F6,$38,$1F,$4F
	.dc.b	$FB,$8D,$3F,$BF,$B7,$2F,$F2,$73,$70,$FF,$6E,$BF,$C0,$E7,$7C,$3C
	.dc.b	$E3,$57,$D2,$31,$5D,$FE,$9B,$FC,$1A,$FF,$88,$4F,$F2,$D3,$87,$FC
	.dc.b	$43,$28,$C7,$07,$F8,$A9,$B3,$C3,$6F,$F6,$84,$E5,$B1,$93,$B1,$FE
	.dc.b	$F4,$9E,$2F,$F8,$3B,$E9,$FF,$DB,$1F,$0B,$FE,$3E,$C3,$FD,$47,$F4
	.dc.b	$DF,$17,$A4,$FE,$52,$B7,$97,$F8,$35,$FE,$39,$87,$33,$F8,$D3,$87
	.dc.b	$FC,$3C,$DE,$03,$49,$1C,$0F,$F5,$4E,$4F,$90,$EB,$97,$C3,$6C,$66
	.dc.b	$77,$18,$23,$82,$1F,$ED,$AF,$A7,$F5,$BF,$EB,$66,$B9,$CB,$BC,$62
	.dc.b	$3F,$CB,$2D,$7F,$D2,$0D,$56,$8C,$36,$7E,$2E,$CF,$A7,$79,$98,$6E
	.dc.b	$1F,$D1,$67,$53,$93,$E4,$7D,$3C,$9F,$E8,$CD,$7F,$A6,$6A,$8C,$47
	.dc.b	$F5,$56,$BF,$EC,$BF,$23,$FC,$DA,$8C,$6F,$F6,$BC,$A4,$EA,$72,$7F
	.dc.b	$B1,$E7,$51,$98,$4F,$09,$C9,$CF,$C2,$7D,$3D,$25,$7F,$B6,$5A,$FF
	.dc.b	$B6,$63,$6E,$48,$7F,$71,$D9,$F4,$EF,$1F,$F0,$AD,$9F,$F7,$8B,$5E
	.dc.b	$0D,$7F,$BD,$78,$4F,$A7,$FC,$B3,$63,$F3,$6D,$7F,$BC,$35,$E3,$1C
	.dc.b	$08,$F0,$7E,$57,$D3,$FF,$C0,$18,$FF,$00,$68,$8C,$47,$FB,$DD,$AF
	.dc.b	$D3,$3E,$5B,$87,$EB,$47,$1F,$3F,$A9,$23,$AC,$ED,$FA,$7E,$9F,$FA
	.dc.b	$C9,$F0,$B5,$FF,$0D,$52,$31,$1E,$18,$C7,$8F,$F8,$39,$CC,$F8,$9A
	.dc.b	$F6,$8D,$1F,$4B,$E0,$A4,$7D,$26,$BF,$E2,$1F,$A7,$FE,$08,$9A,$47
	.dc.b	$F8,$43,$5F,$E1,$AD,$7F,$05,$F4,$EF,$91,$FE,$0A,$D8,$FF,$17,$70
	.dc.b	$23,$11,$C2,$7F,$8B,$F8,$EF,$6D,$F0,$58,$6D,$7F,$8D,$E7,$53,$93
	.dc.b	$F8,$9F,$D3,$FF,$AA,$E7,$FD,$BA,$D7,$F5,$98,$A3,$11,$FC,$19,$B0
	.dc.b	$FA,$67,$D3,$FF,$1E,$6D,$7A,$9A,$FE,$85,$AF,$FB,$B3,$E9,$DF,$3F
	.dc.b	$FC,$99,$B1,$F0,$9A,$FE,$A5,$AF,$18,$E0,$47,$83,$FC,$C3,$E9,$9F
	.dc.b	$4F,$49,$3F,$3A,$32,$FC,$7F,$A6,$7D,$3F,$CE,$6C,$66,$30,$47,$04
	.dc.b	$3F,$CD,$DB,$27,$8C,$7E,$63,$E3,$BF,$74,$FE,$91,$AF,$F1,$26,$F3
	.dc.b	$9C,$EC,$F8,$F7,$D0,$FF,$4C,$6D,$FE,$21,$AF,$F0,$F6,$BF,$B5,$9E
	.dc.b	$53,$C4,$91,$C5,$FE,$AF,$F2,$3F,$78,$9E,$A6,$0F,$EC,$FF,$38,$FA
	.dc.b	$77,$C3,$61,$B8,$7F,$A1,$E7,$53,$92,$46,$67,$6F,$FA,$56,$76,$E6
	.dc.b	$7F,$9E,$63,$F2,$A3,$F7,$CD,$7F,$86,$B5,$FE,$76,$DF,$31,$F2,$CF
	.dc.b	$A7,$FE,$B6,$DA,$FE,$B0,$D7,$FD,$4D,$3C,$27,$64,$8D,$B5,$FE,$79
	.dc.b	$F4,$EF,$8B,$FE,$FE,$D8,$FD,$DB,$5D,$C7,$CA,$F8,$37,$D3,$D2,$4F
	.dc.b	$C9,$8C,$B6,$32,$FC,$78,$CB,$63,$FB,$BF,$CE,$FF,$37,$8F,$EE,$8D
	.dc.b	$7D,$AD,$7F,$9C,$62,$8C,$47,$F6,$D8,$7C,$F3,$CE,$FF,$01,$B6,$7F
	.dc.b	$C0,$0D,$7E,$4B,$5E,$BD,$8F,$91,$FF,$05,$B6,$3F,$C1,$0D,$7E,$8B
	.dc.b	$5E,$31,$C0,$8F,$07,$FC,$01,$F4,$AF,$9C,$A4,$9F,$EB,$31,$96,$C6
	.dc.b	$5F,$E9,$71,$96,$C7,$F9,$07,$E9,$FF,$D9,$91,$FE,$E9,$6B,$FD,$7D
	.dc.b	$AF,$F8,$4D,$A9,$A4,$F8,$4E,$87,$F9,$37,$E1,$FF,$9B,$CF,$FB,$B5
	.dc.b	$AF,$FB,$DD,$AF,$F6,$D6,$BF,$E2,$1F,$91,$FF,$25,$3F,$EF,$56,$BF
	.dc.b	$EF,$F6,$BF,$DD,$5A,$F1,$F7,$FF,$CB,$D1,$88,$34,$C7,$D3,$E3,$E1
	.dc.b	$3F,$CC,$ED,$9F,$F0,$01,$FE,$33,$70,$E1,$AF,$F8,$15,$FE,$33,$9F
	.dc.b	$2E,$E3,$49,$3F,$E7,$1F,$87,$FE,$DE,$47,$07,$F8,$AF,$ED,$DF,$26
	.dc.b	$F7,$7F,$C8,$99,$7F,$A0,$E7,$31,$83,$FA,$EF,$E2,$9F,$DC,$5C,$5F
	.dc.b	$F9,$8A,$76,$AB,$95,$F5,$8C,$57,$F6,$6A,$DD,$7F,$A4,$6F,$A5,$68
	.dc.b	$E9,$7C,$67,$5E,$77,$F9,$57,$BD,$7F,$DA,$B1,$FF,$00,$B7,$0F,$F2
	.dc.b	$43,$5F,$E7,$5D,$FF,$E9,$C9,$F4,$D7,$F1,$EF,$84,$ED,$1E,$EC,$2D
	.dc.b	$FF,$9F,$AF,$83,$E5,$BE,$15,$CA,$73,$8F,$9A,$31,$87,$2F,$23,$E2
	.dc.b	$36,$77,$3F,$E7,$3D,$9F,$8A,$FF,$9E,$E3,$E6,$DF,$06,$D7,$0C,$BF
	.dc.b	$CE,$2D,$BE,$3F,$05,$D7,$1F,$07,$D5,$BC,$8A,$7B,$3E,$39,$F5,$7F
	.dc.b	$C3,$73,$EF,$5B,$6D,$70,$7E,$FD,$70,$ED,$C6,$9F,$5F,$5F,$C5,$7D
	.dc.b	$87,$FA,$25,$B7,$FE,$62,$11,$96,$9F,$E6,$6F,$AE,$FF,$19,$BF,$01
	.dc.b	$B5,$D6,$36,$F7,$8C,$EF,$F8,$DB,$7E,$2C,$ED,$F1,$37,$0F,$F3,$9B
	.dc.b	$70,$FF,$31,$B5,$FF,$31,$35,$FF,$38,$B5,$FF,$4F,$B6,$BB,$D7,$BF
	.dc.b	$C4,$7C,$BF,$DF,$6E,$1F,$6A,$FC,$18,$F9,$E7,$38,$FE,$2F,$F1,$74
	.dc.b	$7E,$75,$B7,$FE,$5C,$6C,$FF,$A8,$DA,$FF,$85,$5A,$FF,$94,$9A,$FF
	.dc.b	$B5,$9B,$CE,$73,$9B,$EC,$5F,$41,$F9,$76,$E1,$FA,$6A,$F0,$E5,$F3
	.dc.b	$E7,$38,$FA,$7F,$15,$F4,$12,$FD,$0D,$8F,$43,$5F,$43,$FE,$CF,$8F
	.dc.b	$F3,$1F,$CA,$FE,$C1,$A7,$86,$31,$D8,$9F,$D6,$7E,$31,$FE,$DB,$AF
	.dc.b	$F5,$7B,$6F,$FD,$3E,$D7,$FD,$3C,$D7,$FD,$5A,$D7,$FD,$B6,$DD,$0D
	.dc.b	$31,$C7,$FC,$37,$3F,$C8,$36,$F9,$DF,$EF,$E3,$0F,$FB,$9B,$E6,$1F
	.dc.b	$79,$FA,$56,$D7,$FB,$19,$B3,$FE,$B6,$6B,$FE,$B5,$6B,$FE,$C4,$6B
	.dc.b	$D6,$7A,$1F,$B1,$F6,$3F,$96,$D7,$FB,$29,$B3,$B8,$FE,$1B,$18,$E8
	.dc.b	$FF,$07,$79,$7F,$F0,$05,$B3,$FD,$5A,$91,$87,$FA,$CB,$5F,$F6,$D3
	.dc.b	$51,$C3,$A7,$FC,$1D,$3C,$E7,$53,$9A,$F0,$7E,$33,$FE,$8A,$AF,$E1
	.dc.b	$ED,$9F,$E8,$95,$E1,$D3,$FA,$62,$9D,$46,$3F,$1A,$7D,$6F,$F1,$F6
	.dc.b	$E1,$FC,$69,$FF,$38,$B6,$3F,$9B,$7E,$36,$F5,$7F,$8D,$36,$7F,$DF
	.dc.b	$6D,$CB,$FD,$D6,$D7,$FD,$D5,$F7,$DF,$CA,$5F,$E4,$8D,$C7,$CF,$F0
	.dc.b	$2F,$94,$FE,$89,$39,$C3,$FD,$93,$F1,$AF,$95,$FE,$BC,$D8,$FF,$6E
	.dc.b	$B7,$1F,$F0,$4F,$72,$31,$1F,$E9,$76,$BD,$2B,$99,$2B,$EE,$74,$3E
	.dc.b	$F3,$F0,$AD,$F3,$BF,$B0,$E3,$D6,$FF,$AD,$56,$BE,$47,$F2,$FB,$7F
	.dc.b	$31,$6E,$3F,$80,$C5,$0F,$A5,$FE,$C4,$EB,$34,$EB,$A7,$26,$AA,$DF
	.dc.b	$F6,$2F,$A7,$7D,$F7,$F6,$B6,$C6,$7C,$7C,$4C,$8E,$0E,$E3,$18,$FF
	.dc.b	$79,$E0,$FA,$0B,$7F,$95,$E5,$7F,$65,$D3,$FC,$27,$B7,$F8,$66,$73
	.dc.b	$97,$F8,$23,$0E,$31,$8C,$3F,$2B,$7C,$2B,$ED,$E1,$FE,$01,$FB,$EF
	.dc.b	$D0,$D7,$F6,$76,$E3,$D7,$1F,$1E,$09,$CE,$1F,$2F,$DE,$FF,$6C,$BF
	.dc.b	$E0,$46,$E1,$FD,$C5,$FE,$F4,$DF,$5F,$FC,$21,$1A,$C7,$FB,$33,$6B
	.dc.b	$A7,$7E,$AB,$FC,$2D,$F3,$6F,$BC,$A6,$3D,$CB,$7F,$E1,$49,$D3,$67
	.dc.b	$4D,$3F,$C5,$BE,$13,$E9,$7F,$C7,$0D,$8F,$E1,$9C,$3F,$E9,$19,$C9
	.dc.b	$23,$33,$BD,$7F,$3E,$D7,$FD,$E8,$D7,$F2,$1F,$8C,$3E,$67,$FC,$60
	.dc.b	$DA,$F1,$DE,$27,$FC,$69,$C7,$FD,$1F,$83,$FC,$14,$DA,$FC,$D3,$FD
	.dc.b	$81,$B8,$70,$23,$FE,$36,$8D,$47,$E7,$1E,$4D,$F0,$7B,$9F,$AC,$E4
	.dc.b	$FE,$D3,$7E,$08,$CD,$F3,$75,$E7,$E2,$FF,$CA,$1A,$38,$3E,$59,$2F
	.dc.b	$E3,$B6,$7E,$3B,$5F,$8E,$FF,$8D,$7F,$15,$7C,$DF,$F9,$55,$B5,$F6
	.dc.b	$27,$13,$9F,$C4,$24,$FF,$95,$2B,$FC,$AD,$7F,$E4,$15,$3A,$8C,$3D
	.dc.b	$7D,$AC,$69,$FE,$37,$F9,$5F,$C9,$57,$F0,$26,$F5,$1F,$CD,$CE,$F9
	.dc.b	$FF,$CD,$51,$A2,$74,$62,$3F,$CD,$DF,$69,$FD,$E4,$98,$7F,$94,$DF
	.dc.b	$EF,$DC,$EF,$4F,$DF,$3F,$E7,$B6,$E3,$FE,$4B,$AF,$B3,$5B,$DF,$AF
	.dc.b	$DD,$CE,$0F,$F4,$4E,$1C,$1F,$E4,$16,$9D,$63,$FE,$65,$8C,$4F,$0B
	.dc.b	$CE,$3D,$FC,$6F,$31,$27,$F9,$AD,$7F,$AB,$2F,$FB,$59,$FC,$A1,$7E
	.dc.b	$0D,$F0,$0F,$83,$56,$BF,$D1,$3E,$D7,$F9,$A2,$F8,$F0,$7F,$99,$7D
	.dc.b	$3F,$F3,$B4,$E6,$7F,$9C,$35,$FE,$6A,$FF,$8B,$EB,$9F,$AB,$FE,$9B
	.dc.b	$8F,$F3,$9E,$7F,$E7,$17,$55,$FE,$83,$6C,$66,$78,$5F,$16,$5F,$43
	.dc.b	$CF,$18,$BE,$5D,$FE,$62,$4F,$F5,$8C,$3F,$88,$CF,$FB,$AB,$1F,$F4
	.dc.b	$07,$3F,$94,$F3,$5D,$BF,$F6,$BD,$3E,$93,$FD,$0D,$ED,$7F,$85,$BE
	.dc.b	$17,$FD,$4D,$3F,$E1,$89,$CA,$8C,$FA,$AF,$AC,$FE,$29,$1C,$27,$A6
	.dc.b	$9F,$E7,$A7,$87,$83,$F6,$30,$E7,$5B,$6C,$7F,$89,$B0,$E9,$E1,$FF
	.dc.b	$B9,$EB,$58,$73,$6E,$5E,$7B,$E9,$39,$79,$65,$E4,$73,$3B,$CB,$D9
	.dc.b	$5B,$6C,$67,$1F,$0E,$7B,$9F,$5C,$DC,$FC,$F1,$C0,$F6,$9F,$E4,$4C
	.dc.b	$7C,$0E,$72,$F3,$BA,$39,$9C,$30,$EE,$E6,$BA,$BA,$9E,$27,$D0,$AE
	.dc.b	$B1,$9C,$BD,$AF,$1C,$BD,$98,$74,$BE,$35,$A6,$D7,$9F,$3F,$9D,$39
	.dc.b	$24,$7A,$5F,$97,$1A,$8C,$10,$F1,$1C,$F3,$F8,$D8,$7C,$D8,$C9,$AC
	.dc.b	$3C,$97,$DA,$3E,$19,$FB,$C2,$9E,$C3,$19,$7E,$49,$BC,$0F,$98,$FF
	.dc.b	$3A,$47,$E6,$E3,$0F,$F9,$DE,$7F,$D4,$D3,$A6,$C6,$67,$79,$FE,$4D
	.dc.b	$B7,$F9,$76,$CF,$FA,$03,$D9,$7C,$57,$F1,$2B,$E1,$8F,$E1,$6F,$77
	.dc.b	$E3,$8F,$C6,$5F,$08,$D4,$47,$F0,$13,$DB,$25,$6E,$73,$5B,$BE,$F1
	.dc.b	$F8,$99,$D4,$60,$98,$FA,$6B,$C3,$8F,$9E,$31,$7D,$09,$5F,$BC,$BE
	.dc.b	$73,$A8,$E7,$3F,$C0,$F1,$F3,$56,$A3,$18,$79,$6F,$81,$FB,$B9,$D4
	.dc.b	$FC,$33,$BC,$69,$F9,$98,$79,$5E,$66,$8B,$87,$96,$35,$1C,$E3,$CA
	.dc.b	$D9,$F0,$4E,$B2,$FB,$F5,$A6,$CE,$AB,$F8,$27,$5B,$E2,$1F,$6F,$FD
	.dc.b	$3B,$9F,$FC,$C1,$18,$AF,$F5,$DC,$62,$75,$C1,$FE,$BE,$BE,$5E,$5E
	.dc.b	$2F,$FC,$D9,$82,$3F,$2F,$3A,$9F,$F2,$0C,$E5,$B3,$F9,$CE,$77,$14
	.dc.b	$9F,$EB,$6F,$58,$CB,$46,$15,$93,$C0,$FE,$C7,$18,$7F,$AB,$45,$CB
	.dc.b	$FA,$D4,$F3,$3C,$51,$D2,$7F,$A8,$46,$0F,$37,$5B,$EA,$BC,$8D,$9F
	.dc.b	$8F,$3F,$E0,$49,$CF,$B5,$27,$FB,$EB,$C8,$8F,$F8,$03,$C0,$E6,$47
	.dc.b	$FC,$1F,$D9,$5F,$F0,$64,$5C,$BF,$D2,$CA,$7E,$96,$DF,$E1,$F6,$C6
	.dc.b	$6B,$99,$E6,$3F,$C6,$5E,$57,$CD,$F1,$3F,$D9,$91,$FE,$C6,$6B,$FD
	.dc.b	$BD,$AF,$F8,$31,$AF,$FB,$05,$7C,$67,$32,$47,$7D,$FF,$C7,$DE,$66
	.dc.b	$21,$9F,$F9,$16,$B0,$FF,$91,$E7,$F0,$1E,$4B,$EB,$E9,$2F,$FC,$A9
	.dc.b	$F5,$95,$FF,$2A,$45,$BF,$B5,$8A,$D7,$D3,$59,$87,$F9,$75,$F9,$3F
	.dc.b	$80,$BD,$DA,$7E,$69,$C4,$FF,$A6,$9A,$FF,$BD,$1A,$FF,$8E,$38,$11
	.dc.b	$88,$FD,$43,$5F,$F6,$93,$63,$73,$FA,$3E,$97,$DD,$7E,$85,$B7,$FE
	.dc.b	$91,$8F,$BF,$F8,$EB,$F1,$5D,$44,$9D,$1F,$7F,$05,$6E,$3E,$A7,$A9
	.dc.b	$F7,$B7,$F8,$ED,$E8,$FF,$56,$F2,$FE,$C6,$1F,$01,$62,$5F,$27,$51
	.dc.b	$BF,$62,$4F,$67,$E6,$F4,$C8,$F9,$4F,$CD,$23,$82,$1F,$35,$CB,$FE
	.dc.b	$41,$F0,$1F,$2F,$FC,$15,$B1,$F8,$F6,$BF,$8E,$C8,$B9,$BF,$C8,$9E
	.dc.b	$AF,$1B,$93,$C5,$E9,$1F,$C4,$7C,$0F,$ED,$F3,$FE,$F1,$F1,$B0,$7F
	.dc.b	$15,$FC,$75,$F5,$78,$9F,$F7,$C4,$D7,$C4,$78,$67,$6F,$E9,$7D,$B7
	.dc.b	$E1,$DF,$C7,$DB,$19,$8F,$BF,$B4,$8F,$B9,$C8,$8C,$47,$6F,$31,$F1
	.dc.b	$AF,$C7,$73,$CF,$FC,$77,$A3,$F0,$78,$63,$8F,$FC,$A9,$7B,$C7,$F7
	.dc.b	$1E,$B7,$E6,$9B,$DD,$FC,$CE,$31,$1E,$CE,$5F,$F6,$6F,$CC,$E6,$A4
	.dc.b	$F0,$39,$C7,$F3,$BF,$84,$74,$D2,$4F,$F0,$78,$7C,$0E,$23,$B4,$70
	.dc.b	$75,$3C,$E7,$FA,$37,$C2,$FE,$D3,$7F,$E0,$2D,$D7,$F0,$39,$A3,$18
	.dc.b	$37,$FD,$2F,$1F,$EB,$0F,$EB,$3C,$7F,$4A,$7F,$C7,$13,$A7,$FC,$74
	.dc.b	$71,$8D,$3F,$C0,$5A,$FF,$67,$6B,$B9,$EA,$7F,$61,$F5,$7E,$84,$7C
	.dc.b	$B6,$E5,$FC,$69,$B9,$7F,$A2,$7C,$3F,$F1,$44,$FF,$1E,$F8,$8F,$E0
	.dc.b	$B7,$FE,$A5,$6B,$FE,$C2,$6B,$F0,$9A,$B9,$9F,$13,$F1,$1B,$7F,$DF
	.dc.b	$5B,$3F,$D2,$5D,$FE,$19,$9F,$8B,$34,$93,$BF,$81,$FF,$6C,$B7,$DF
	.dc.b	$FF,$2D,$DA,$73,$3F,$E6,$58,$C4,$EA,$73,$3F,$EE,$49,$CB,$F4,$9A
	.dc.b	$F1,$8D,$BF,$EA,$F6,$BF,$C0,$7E,$33,$F0,$3C,$66,$98,$F6,$9F,$3B
	.dc.b	$FE,$11,$6D,$7E,$02,$BF,$CB,$33,$FE,$19,$9C,$CF,$FB,$55,$AF,$FB
	.dc.b	$C9,$AF,$E4,$9A,$FF,$13,$6C,$F4,$8E,$DE,$6B,$DC,$FF,$11,$36,$3F
	.dc.b	$B6,$35,$FD,$97,$D2,$7E,$13,$8F,$F3,$EF,$F9,$B6,$3F,$90,$9F,$C8
	.dc.b	$5C,$3A,$F7,$7F,$C6,$51,$88,$D4,$60,$8F,$C1,$8D,$4B,$F1,$CF,$E0
	.dc.b	$7C,$BF,$E2,$98,$C7,$7F,$F9,$15,$B7,$FE,$11,$F8,$07,$DA,$3D,$7F
	.dc.b	$F2,$54,$61,$FF,$4F,$B6,$7F,$5E,$D7,$FD,$54,$D7,$DA,$D7,$F9,$FB
	.dc.b	$53,$4C,$6C,$F9,$3F,$E7,$B9,$7F,$82,$1B,$8F,$F0,$86,$DF,$F0,$7F
	.dc.b	$C7,$9F,$55,$FE,$30,$6C,$7F,$04,$F0,$11,$88,$FE,$12,$D7,$FC,$0C
	.dc.b	$D7,$FC,$D8,$D7,$B4,$FF,$9A,$A3,$13,$D2,$3F,$CD,$19,$3D,$7F,$F3
	.dc.b	$33,$6F,$FC,$60,$D7,$F9,$B7,$90,$F2,$3F,$CD,$8D,$8F,$CE,$35,$FF
	.dc.b	$6B,$6B,$FD,$2A,$7F,$CE,$D3,$99,$FC,$5C,$E5,$E3,$3A,$23,$FA,$7F
	.dc.b	$97,$FD,$F3,$FE,$43,$6B,$FD,$25,$BB,$7F,$8E,$36,$FF,$39,$4F,$F4
	.dc.b	$D6,$BE,$06,$BF,$CF,$1A,$FF,$7D,$6D,$FF,$1D,$F2,$FF,$D1,$5D,$9F
	.dc.b	$CA,$AB,$FC,$04,$D1,$F3,$3F,$E5,$A8,$FF,$2C,$46,$EB,$94,$62,$BF
	.dc.b	$92,$56,$E3,$FB,$17,$CD,$FF,$1A,$C7,$FC,$A2,$DC,$7F,$A8,$B6,$4B
	.dc.b	$F1,$FF,$AD,$5C,$3C,$0F,$F5,$04,$EF,$2D,$C7,$FA,$BE,$38,$70,$7F
	.dc.b	$A0,$3E,$79,$F6,$14,$99,$7F,$90,$7E,$83,$F8,$36,$5F,$E0,$1A,$DC
	.dc.b	$FF,$91,$5A,$FF,$85,$B9,$51,$89,$FF,$04,$57,$2D,$7F,$D8,$4F,$F1
	.dc.b	$1B,$FF,$30,$35,$CF,$7F,$FB,$0A,$31,$1F,$EC,$6F,$A7,$FF,$4B,$FC
	.dc.b	$49,$A7,$FA,$6A,$5F,$1B,$8A,$FF,$2D,$56,$DB,$5B,$9F,$F6,$94,$3E
	.dc.b	$8B,$FD,$27,$97,$F7,$68,$C7,$C3,$7F,$A4,$1F,$EF,$33,$A6,$C6,$5F
	.dc.b	$E4,$E7,$3F,$92,$FA,$EF,$06,$BA,$94,$62,$3F,$97,$C6,$38,$7C,$51
	.dc.b	$FE,$52,$F8,$EA,$9A,$47,$FC,$1F,$A7,$DC,$DB,$F8,$63,$63,$3C,$3F
	.dc.b	$A2,$E1,$4E,$67,$E9,$CE,$63,$11,$F6,$23,$07,$EA,$BA,$3F,$C4,$5F
	.dc.b	$23,$6C,$D2,$61,$E2,$F9,$EE,$84,$70,$4C,$3D,$31,$96,$C7,$F8,$A3
	.dc.b	$37,$D1,$3E,$57,$F7,$71,$86,$C6,$67,$83,$6B,$EA,$D6,$E7,$35,$C0
	.dc.b	$8F,$F8,$17,$AF,$FC,$1C,$DF,$3B,$E0,$46,$2F,$FC,$20,$D7,$FC,$1C
	.dc.b	$D7,$FD,$13,$A7,$F0,$69,$E2,$4A,$FF,$09,$7E,$3D,$F8,$26,$F1,$7C
	.dc.b	$D8,$C5,$FF,$87,$5A,$FF,$1B,$F3,$3F,$8D,$F6,$AF,$FD,$F8,$7F,$83
	.dc.b	$CB,$3F,$CD,$1A,$FF,$88,$FF,$20,$7C,$FF,$DB,$6C,$FF,$89,$5B,$3F
	.dc.b	$E2,$D6,$BC,$67,$81,$27,$83,$E4,$F1,$1F,$29,$49,$3F,$D2,$1F,$E6
	.dc.b	$3F,$04,$F8,$AA,$49,$FD,$94,$65,$B1,$9C,$3F,$CD,$73,$96,$C7,$E7
	.dc.b	$7E,$8B,$FB,$CD,$7F,$6D,$6C,$FF,$9D,$1A,$FF,$BC,$5A,$FF,$88,$3E
	.dc.b	$1D,$F2,$DF,$D1,$9B,$5F,$C1,$5B,$3F,$E7,$C6,$BF,$C0,$9A,$F3,$9E
	.dc.b	$44,$9F,$27,$CA,$79,$C8,$E0,$FF,$2B,$7E,$43,$CA,$BF,$CE,$0D,$E4
	.dc.b	$34,$97,$FB,$56,$D7,$28,$FC,$3F,$D1,$FF,$B7,$23,$FC,$A8,$DB,$FE
	.dc.b	$F2,$D7,$FB,$BB,$5F,$CD,$FE,$44,$FA,$6F,$E0,$4D,$AF,$F3,$63,$67
	.dc.b	$FC,$E8,$D7,$FC,$E4,$D7,$9C,$F2,$24,$FF,$99,$FD,$A7,$D2,$7E,$79
	.dc.b	$B1,$98,$C1,$FC,$4F,$F2,$3F,$E2,$F7,$FE,$95,$6E,$86,$93,$C2,$E0
	.dc.b	$4B,$FF,$42,$47,$0E,$B7,$C8,$7F,$1F,$6C,$FF,$A2,$5B,$87,$FA,$26
	.dc.b	$7F,$92,$7C,$1F,$F5,$5A,$FF,$47,$36,$7F,$D2,$CD,$7F,$D2,$8D,$7F
	.dc.b	$83,$BF,$EF,$58,$E3,$F1,$BD,$08,$FF,$A6,$7E,$11,$F2,$7F,$CC,$DB
	.dc.b	$19,$8C,$1F,$99,$F8,$49,$1F,$99,$6E,$1F,$99,$6B,$F9,$96,$A3,$FD
	.dc.b	$C1,$D1,$FC,$09,$C7,$91,$FB,$9E,$DF,$E9,$93,$99,$DC,$7F,$16,$F8
	.dc.b	$D7,$D2,$D2,$63,$FE,$A5,$8C,$B6,$33,$5F,$E4,$C9,$CB,$63,$F9,$0F
	.dc.b	$D2,$BE,$3E,$92,$FF,$D6,$11,$96,$C6,$6B,$FC,$AF,$39,$6C,$7F,$5D
	.dc.b	$F0,$FF,$12,$4E,$4C,$12,$FF,$D9,$51,$96,$C7,$C2,$F5,$E4,$7F,$9F
	.dc.b	$74,$FF,$38,$36,$3F,$9C,$F7,$FF,$47,$C5,$82,$30,$DE,$5D,$92,$FF
	.dc.b	$D9,$93,$A2,$47,$F4,$4F,$A5,$7D,$0D,$25,$7F,$B4,$E3,$2D,$8C,$E1
	.dc.b	$FE,$7A,$9C,$B6,$3F,$91,$72,$4A,$FE,$44,$D9,$FE,$44,$D7,$F9,$13
	.dc.b	$5F,$C3,$F3,$BE,$87,$FC,$24,$DA,$FF,$51,$36,$7F,$C5,$7C,$9F,$B2
	.dc.b	$AF,$EB,$FA,$FF,$36,$9F,$8E,$DD,$7F,$D3,$9F,$01,$A2,$63,$FE,$F1
	.dc.b	$9D,$12,$3F,$2F,$F4,$92,$7E,$B3,$6F,$C3,$C0,$8C,$47,$FB,$93,$2F
	.dc.b	$EB,$11,$88,$FD,$37,$CB,$3E,$5F,$F8,$2C,$7F,$BA,$A3,$1A,$FF,$A4
	.dc.b	$9B,$8F,$F9,$22,$51,$87,$FC,$96,$D7,$95,$74,$7B,$B5,$E2,$DC,$3F
	.dc.b	$8C,$EF,$F3,$89,$7F,$39,$FF,$04,$FC,$F7,$F8,$96,$FF,$D6,$AD,$7F
	.dc.b	$80,$6C,$8C,$47,$E5,$9B,$19,$F4,$79,$92,$FD,$91,$96,$FD,$07,$E7
	.dc.b	$09,$C9,$FE,$1A,$6C,$6A,$30,$47,$FD,$CD,$F9,$23,$E3,$FF,$CB,$4D
	.dc.b	$8C,$C6,$0F,$F1,$2F,$D1,$FF,$59,$CF,$F5,$AD,$C4,$D3,$5A,$BC,$CF
	.dc.b	$F7,$06,$1F,$EF,$56,$DF,$E5,$1A,$FE,$0D,$AF,$F2,$46,$A9,$E9,$F1
	.dc.b	$1F,$E7,$19,$CC,$62,$BF,$04,$BF,$25,$FE,$29,$7F,$8F,$6B,$FC,$C1
	.dc.b	$AF,$56,$E3,$EC,$F6,$FE,$09,$2F,$E9,$E5,$FE,$76,$7B,$FE,$37,$FC
	.dc.b	$3D,$BD,$BF,$E8,$58,$C6,$3F,$B2,$6B,$FE,$FA,$F4,$FF,$C9,$6F,$E1
	.dc.b	$9A,$FF,$1A,$6E,$3F,$E4,$2F,$5B,$FC,$AF,$18,$23,$82,$61,$F8,$88
	.dc.b	$CB,$63,$FD,$31,$F9,$37,$C8,$26,$F8,$3A,$FC,$1C,$AE,$4D,$C7,$F5
	.dc.b	$DF,$41,$FE,$54,$EE,$FF,$56,$77,$FF,$5D,$94,$E5,$FF,$57,$47,$F9
	.dc.b	$EB,$E6,$7F,$8A,$7C,$12,$38,$56,$E3,$15,$FD,$9E,$31,$3A,$D5,$87
	.dc.b	$82,$9E,$83,$1E,$B7,$BA,$D5,$79,$2F,$8B,$E4,$3B,$38,$C7,$C3,$87
	.dc.b	$3F,$0F,$FD,$09,$5B,$39,$5F,$4C,$BC,$B1,$CA,$3C,$95,$C7,$4D,$63
	.dc.b	$D7,$2F,$25,$F3,$C3,$8D,$EE,$31,$5B,$7A,$18,$C3,$AE,$7E,$E8,$CE
	.dc.b	$7F,$21,$F2,$9B,$3D,$D7,$D6,$35,$87,$A2,$77,$18,$C3,$C1,$5B,$C7
	.dc.b	$CF,$39,$C7,$D6,$78,$B3,$F8,$B1,$9C,$3C,$0D,$3E,$5E,$5E,$78,$EF
	.dc.b	$7D,$08,$FD,$A8,$D6,$5E,$52,$65,$E4,$8C,$BF,$16,$7D,$E7,$E4,$4A
	.dc.b	$7C,$E3,$13,$F9,$36,$FC,$47,$F4,$67,$F3,$6D,$73,$E0,$FF,$4B,$7F
	.dc.b	$37,$19,$7F,$2E,$F1,$C7,$E5,$B7,$1C,$D7,$E0,$BD,$3F,$E5,$D7,$F8
	.dc.b	$7A,$D3,$F9,$D6,$BD,$A7,$87,$5F,$EA,$F0,$EB,$1C,$49,$3F,$72,$FA
	.dc.b	$BF,$8B,$BE,$78,$9A,$78,$36,$57,$0C,$3A,$DF,$1C,$3A,$C7,$E8,$A7
	.dc.b	$59,$7B,$2B,$44,$9D,$65,$E7,$9D,$47,$42,$3F,$EF,$AA,$E7,$5F,$06
	.dc.b	$FB,$CF,$03,$57,$CE,$74,$77,$8D,$3F,$C0,$2B,$9B,$F2,$4D,$1C,$2F
	.dc.b	$CA,$F6,$31,$3C,$B0,$F8,$77,$DE,$79,$57,$A1,$D9,$DE,$3C,$9D,$4F
	.dc.b	$88,$FF,$27,$4F,$F5,$06,$BF,$E3,$06,$BF,$E1,$F6,$BF,$E8,$3E,$07
	.dc.b	$E4,$6F,$92,$FC,$84,$61,$B1,$99,$FF,$0F,$4E,$5F,$E7,$DC,$7F,$97
	.dc.b	$E8,$71,$7F,$18,$9F,$D2,$46,$27,$53,$99,$FE,$A5,$18,$23,$D0,$87
	.dc.b	$28,$E0,$FF,$57,$8D,$BF,$D7,$4F,$D4,$64,$E9,$FD,$54,$7F,$79,$6C
	.dc.b	$66,$3D,$1D,$6F,$4E,$92,$3F,$B7,$BD,$3B,$BF,$8F,$47,$F7,$BE,$2C
	.dc.b	$11,$FF,$07,$76,$78,$BB,$CD,$24,$68,$87,$F8,$4D,$CB,$F2,$3C,$27
	.dc.b	$93,$F3,$5B,$1F,$E3,$28,$C1,$F0,$3C,$2F,$F7,$7C,$7F,$8B,$BD,$6C
	.dc.b	$11,$FF,$1C,$F2,$73,$23,$82,$1E,$17,$2D,$3F,$C8,$7E,$33,$E9,$7F
	.dc.b	$00,$D8,$CC,$60,$8E,$08,$7E,$12,$38,$3F,$E8,$6F,$95,$79,$BF,$69
	.dc.b	$B1,$FE,$2C,$6B,$B8,$FE,$4D,$DC,$F9,$D7,$DE,$E1,$B5,$FE,$75,$9D
	.dc.b	$4E,$49,$19,$9D,$BF,$E7,$86,$BF,$E8,$16,$BF,$EE,$85,$F4,$DF,$90
	.dc.b	$6C,$7F,$90,$9A,$FF,$A8,$9A,$FF,$A4,$5A,$F3,$8E,$44,$7F,$CE,$3F
	.dc.b	$4A,$F6,$BF,$5C,$D8,$CC,$60,$8E,$08,$7E,$CA,$38,$3F,$3F,$9B,$E7
	.dc.b	$11,$F9,$CF,$EA,$3C,$57,$D1,$D2,$4F,$F5,$58,$CB,$63,$2F,$F8,$D9
	.dc.b	$AF,$F8,$D7,$BB,$FC,$6B,$E9,$7F,$07,$A3,$49,$3F,$E9,$08,$CB,$63
	.dc.b	$2F,$F0,$C8,$CB,$FC,$3B,$E1,$3F,$D7,$0F,$FA,$61,$AE,$7D,$3F,$E2
	.dc.b	$D1,$88,$F2,$35,$FF,$53,$B5,$FF,$7D,$B6,$CE,$7B,$BF,$21,$97,$E0
	.dc.b	$49,$5F,$81,$7F,$BD,$78,$EF,$85,$7C,$AE,$1B,$5F,$EC,$99,$D4,$E4
	.dc.b	$91,$99,$DB,$FE,$0F,$9D,$B9,$9F,$E6,$5E,$7F,$D4,$8F,$E6,$2D,$7E
	.dc.b	$93,$5F,$F6,$2B,$56,$48,$C3,$76,$EA,$49,$FF,$62,$FB,$8F,$A3,$FE
	.dc.b	$84,$D8,$CC,$60,$8E,$08,$7F,$46,$8E,$0F,$E9,$FE,$8B,$F5,$84,$7F
	.dc.b	$58,$FE,$57,$A6,$47,$E4,$DA,$FF,$A1,$9A,$FF,$8C,$9A,$8F,$F3,$CF
	.dc.b	$4F,$F8,$69,$FF,$45,$69,$FC,$4B,$CC,$FB,$33,$99,$E1,$39,$7E,$C7
	.dc.b	$C0,$FF,$14,$78,$B5,$DB,$5F,$F4,$3B,$60,$BE,$07,$F0,$ED,$9E,$B1
	.dc.b	$83,$F1,$7E,$C3,$E5,$A9,$27,$F1,$51,$97,$F1,$9F,$06,$F7,$3F,$26
	.dc.b	$D8,$CC,$60,$8E,$08,$7E,$3E,$38,$3F,$DF,$3D,$47,$CF,$FF,$80,$5B
	.dc.b	$1F,$E6,$33,$F1,$6E,$1E,$11,$FA,$4F,$86,$7D,$75,$24,$FE,$8E,$32
	.dc.b	$D8,$CB,$FA,$18,$CB,$63,$FC,$25,$F4,$1F,$5E,$3F,$36,$DE,$63,$4D
	.dc.b	$6E,$F8,$BF,$F0,$3B,$6B,$F7,$31,$82,$38,$21,$FB,$97,$FC,$1F,$EB
	.dc.b	$3E,$7F,$FC,$51,$3F,$E8,$58,$C7,$97,$FE,$2E,$9C,$CF,$FB,$66,$72
	.dc.b	$FF,$A0,$7E,$11,$F0,$DF,$E1,$A6,$C6,$63,$04,$70,$43,$E9,$47,$07
	.dc.b	$FD,$2D,$E0,$BE,$79,$F3,$DF,$E2,$86,$CF,$D9,$9D,$4E,$4F,$17,$C9
	.dc.b	$49,$F3,$35,$EE,$D7,$FD,$49,$68,$C4,$7E,$C7,$7F,$E4,$9E,$0F,$F9
	.dc.b	$27,$C3,$FE,$21,$3F,$89,$9C,$B6,$32,$FF,$21,$F8,$3F,$E6,$DC,$26
	.dc.b	$93,$C7,$FE,$62,$D9,$FE,$5C,$D7,$FC,$AB,$EE,$7F,$93,$1F,$E4,$2D
	.dc.b	$C0,$D3,$56,$F9,$2F,$B3,$5F,$C1,$1A,$A3,$5C,$7F,$D0,$A7,$30,$F9
	.dc.b	$37,$CC,$FF,$46,$6D,$FF,$43,$6B,$FC,$FD,$AF,$F1,$66,$BF,$C5,$1A
	.dc.b	$F8,$5B,$3D,$E7,$FD,$99,$F2,$AF,$9D,$FF,$2D,$DF,$F2,$38,$C7,$BD
	.dc.b	$A9,$CC,$A7,$2F,$EF,$5A,$FF,$B8,$5A,$FF,$2D,$38,$47,$28,$CC,$7F
	.dc.b	$31,$E4,$7C,$CF,$FB,$1E,$3F,$20,$D7,$F2,$ED,$7A,$47,$0F,$87,$FF
	.dc.b	$3F,$CE,$67,$A3,$54,$ED,$EB,$3F,$DD,$3A,$7F,$8A,$47,$F4,$06,$BF
	.dc.b	$CF,$9A,$FD,$B6,$CF,$A3,$87,$C2,$49,$F4,$B5,$F4,$B5,$F4,$DF,$E5
	.dc.b	$DF,$F7,$F7,$CD,$7E,$0B,$E6,$EE,$A6,$92,$7F,$96,$4F,$EA,$F3,$E3
	.dc.b	$CF,$F5,$67,$2D,$7E,$90,$9E,$0F,$FA,$93,$E2,$7F,$D4,$A9,$A4,$7F
	.dc.b	$D9,$5E,$6F,$E8,$9F,$27,$2F,$D9,$F8,$94,$E7,$2D,$6C,$FC,$9D,$E4
	.dc.b	$E7,$0E,$17,$71,$FA,$B8,$F2,$9F,$EE,$98,$F2,$8C,$4F,$80,$FB,$78
	.dc.b	$74,$AD,$9D,$0B,$1A,$FC,$53,$FD,$8F,$97,$8D,$B4,$69,$AF,$9F,$7C
	.dc.b	$F7,$96,$91,$8C,$3A,$C7,$36,$C6,$70,$F3,$4F,$0A,$E5,$39,$CB,$E2
	.dc.b	$E3,$E1,$C3,$8D,$77,$CB,$FD,$A5,$1D,$9B,$8F,$C1,$E8,$FE,$11,$EE
	.dc.b	$6F,$D8,$97,$B7,$E6,$4E,$A3,$07,$C4,$EE,$7D,$83,$F0,$2F,$A5,$18
	.dc.b	$78,$F7,$7F,$7C,$AE,$15,$B6,$CF,$08,$C6,$5E,$5A,$E0,$FD,$7B,$E9
	.dc.b	$7E,$13,$FC,$98,$7F,$3D,$2C,$7F,$32,$F0,$FF,$DF,$B3,$FC,$C1,$BF
	.dc.b	$44,$F5,$BC,$ED,$BE,$93,$89,$F5,$F8,$AF,$03,$9B,$6B,$FA,$54,$60
	.dc.b	$8E,$09,$8F,$EA,$31,$FF,$3C,$4F,$93,$C0,$78,$9F,$21,$B7,$F3,$70
	.dc.b	$ED,$87,$C1,$9C,$BF,$0F,$9F,$ED,$BF,$EA,$36,$BF,$D8,$DA,$FF,$9C
	.dc.b	$9B,$8F,$F1,$AE,$1F,$8B,$18,$23,$F2,$3E,$8F,$F1,$95,$F4,$5B,$CF
	.dc.b	$FB,$9B,$E5,$97,$E0,$E7,$2F,$F7,$3F,$9C,$7D,$6F,$D2,$6C,$7F,$73
	.dc.b	$6E,$1F,$15,$AF,$FA,$F5,$AF,$39,$E4,$49,$FE,$F9,$F7,$2F,$6F,$F2
	.dc.b	$2D,$8F,$CB,$5F,$6C,$BD,$13,$9C,$3F,$19,$F4,$0D,$BF,$0E,$DC,$3F
	.dc.b	$C1,$0D,$9F,$B4,$D7,$FD,$A4,$D7,$FC,$24,$D7,$FD,$42,$DD,$0D,$F8
	.dc.b	$F7,$C5,$FE,$A1,$B7,$FA,$D9,$F0,$65,$F3,$E7,$38,$7E,$8F,$D8,$7C
	.dc.b	$FF,$E7,$9B,$87,$F8,$7D,$B3,$FE,$6D,$9D,$12,$74,$D8,$FC,$B7,$D3
	.dc.b	$FF,$26,$C3,$EF,$36,$BF,$C5,$D7,$D2,$73,$E2,$7F,$06,$FA,$87,$CE
	.dc.b	$FF,$3D,$C7,$FD,$43,$18,$C3,$FD,$A8,$D8,$FF,$68,$35,$DC,$F1,$7F
	.dc.b	$D4,$B1,$96,$C6,$5F,$F1,$24,$65,$FF,$67,$FD,$1B,$E3,$FF,$89,$B6
	.dc.b	$FF,$9D,$F8,$4B,$F1,$65,$F9,$BA,$D6,$1F,$C9,$7F,$28,$FA,$38,$7F
	.dc.b	$AD,$9B,$1F,$EB,$0F,$AE,$FE,$4B,$18,$E3,$7D,$5F,$E6,$2B,$FD,$C2
	.dc.b	$D8,$CF,$4F,$F1,$48,$C7,$D9,$FF,$20,$CF,$F5,$8D,$CB,$FD,$8B,$F9
	.dc.b	$4F,$FB,$F9,$FE,$61,$19,$C7,$FC,$27,$F5,$3F,$8F,$EF,$34,$9D,$5F
	.dc.b	$D6,$1B,$7F,$E7,$36,$D7,$F1,$5E,$E7,$D3,$7F,$3A,$6D,$7F,$56,$AF
	.dc.b	$74,$62,$BF,$9A,$7C,$63,$EF,$FF,$A0,$B6,$FF,$A3,$3F,$CB,$67,$73
	.dc.b	$A2,$4E,$BC,$6B,$3F,$CF,$FE,$A7,$F8,$FE,$FF,$C3,$DB,$5F,$E7,$F7
	.dc.b	$FC,$BE,$FF,$53,$73,$2E,$24,$61,$BE,$37,$F5,$68,$E3,$3D,$E3,$FB
	.dc.b	$B6,$8F,$A7,$A7,$F8,$37,$EE,$9F,$66,$F9,$DC,$E7,$F1,$6A,$E3,$FE
	.dc.b	$0D,$BD,$D6,$34,$FE,$51,$F6,$4F,$B3,$FE,$CE,$DC,$79,$1F,$DB,$5C
	.dc.b	$61,$FD,$87,$1F,$A1,$7F,$CC,$5B,$5F,$EA,$17,$F9,$8C,$FC,$17,$2D
	.dc.b	$E3,$FE,$EB,$18,$DF,$DD,$EC,$62,$7F,$BC,$FC,$2F,$FB,$AB,$0F,$EF
	.dc.b	$CD,$C3,$FB,$92,$FA,$DD,$FC,$6C,$AD,$BC,$A3,$15,$FE,$1A,$E1,$7D
	.dc.b	$27,$E4,$C9,$7F,$93,$6C,$FF,$B3,$27,$73,$92,$47,$F9,$49,$FF,$16
	.dc.b	$FA,$5F,$E4,$16,$9F,$97,$AC,$7B,$5F,$DA,$B7,$FE,$B1,$F0,$CF,$13
	.dc.b	$FC,$2E,$DB,$FF,$0E,$3F,$81,$6C,$68,$8E,$B8,$6E,$5F,$E1,$5A,$FE
	.dc.b	$9D,$3F,$ED,$36,$C7,$ED,$BD,$3E,$FE,$1C,$AE,$EF,$E4,$23,$13,$F1
	.dc.b	$E3,$F7,$04,$9F,$DC,$36,$3E,$33,$5F,$F4,$DE,$BF,$C3,$FE,$85,$F5
	.dc.b	$1F,$E3,$46,$DF,$EE,$EF,$13,$9F,$8B,$93,$FE,$54,$FC,$23,$F8,$77
	.dc.b	$3B,$0F,$F7,$EC,$FF,$90,$5B,$97,$F8,$D5,$BB,$FF,$9E,$2F,$8C,$62
	.dc.b	$BF,$DE,$DF,$4D,$F9,$13,$F2,$BE,$2B,$F0,$4F,$2E,$3F,$E3,$F9,$D7
	.dc.b	$4F,$F9,$31,$DC,$EB,$2F,$F0,$94,$7E,$6F,$EA,$7F,$B7,$13,$C7,$FF
	.dc.b	$20,$BF,$C9,$79,$DF,$31,$FE,$73,$6E,$9F,$E5,$B9,$FE,$7B,$A7,$E5
	.dc.b	$AB,$FC,$B7,$0F,$A6,$A7,$18,$B5,$C6,$2E,$BF,$E7,$8C,$3A,$BF,$E5
	.dc.b	$9B,$E5,$5D,$FD,$3F,$29,$1F,$F5,$6C,$FF,$19,$9F,$F4,$BB,$FE,$E1
	.dc.b	$E1,$43,$E8,$FC,$A7,$9A,$ED,$FF,$9D,$79,$3F,$CC,$78,$71,$6A,$3F
	.dc.b	$93,$38,$6F,$BF,$FE,$68,$87,$D6,$FF,$9F,$1B,$5D,$67,$13,$23,$C0
	.dc.b	$EE,$31,$5F,$E7,$9A,$E2,$D9,$D5,$70,$3D,$1E,$D7,$84,$8F,$C4,$CF
	.dc.b	$F3,$71,$E5,$EF,$F2,$1E,$56,$CF,$FA,$E2,$DF,$49,$FA,$2E,$2F,$81
	.dc.b	$F1,$6E,$EF,$85,$5A,$D3,$FC,$84,$FF,$B6,$63,$E3,$7A,$EF,$AD,$F4
	.dc.b	$DF,$1A,$F9,$58,$F5,$78,$18,$C3,$9C,$EE,$3B,$63,$E2,$C7,$AE,$1D
	.dc.b	$1F,$1E,$1C,$EF,$83,$63,$38,$74,$78,$ED,$FE,$EB,$AD,$CE,$71,$F1
	.dc.b	$63,$DD,$A7,$9B,$2F,$65,$F1,$C3,$AD,$6E,$BC,$E7,$B0,$B8,$7C,$1B
	.dc.b	$E3,$9E,$F2,$F3,$57,$6A,$F1,$DF,$86,$F7,$3C,$EB,$6F,$4C,$3C,$2F
	.dc.b	$97,$3F,$88,$F3,$8C,$E1,$D2,$35,$97,$A5,$D1,$0E,$83,$0F,$0C,$F1
	.dc.b	$25,$74,$3E,$5E,$5E,$BC,$FE,$55,$F2,$39,$DF,$2A,$E1,$1D,$F2,$F8
	.dc.b	$33,$98,$C5,$7E,$31,$FC,$79,$F3,$0B,$3F,$89,$6C,$6D,$FC,$1E,$BF
	.dc.b	$E5,$B7,$F8,$2C,$FF,$B4,$FD,$9F,$E0,$F1,$F9,$E9,$D3,$D1,$AF,$28
	.dc.b	$FA,$D1,$CF,$2F,$C6,$B7,$1F,$CF,$FB,$7F,$EE,$5B,$DD,$C7,$EC,$56
	.dc.b	$DF,$99,$3B,$73,$3B,$BE,$AF,$E0,$67,$51,$89,$D9,$1F,$C6,$46,$AF
	.dc.b	$D1,$39,$6C,$EB,$D1,$FD,$CB,$CC,$91,$A3,$D7,$1A,$71,$3F,$B8,$C3
	.dc.b	$D4,$FC,$9C,$3C,$D5,$B8,$D4,$7A,$EB,$8E,$6C,$7C,$D5,$A3,$CC,$5C
	.dc.b	$7C,$CF,$13,$15,$CA,$78,$C7,$84,$E2,$59,$E2,$E3,$2F,$C0,$3D,$AF
	.dc.b	$C5,$8F,$9F,$87,$FB,$FE,$9F,$C7,$3A,$FF,$DE,$CF,$F8,$0A,$FF,$A2
	.dc.b	$39,$DB,$FA,$15,$7F,$70,$9F,$E6,$D1,$F8,$C8,$C4,$7F,$49,$8C,$4F
	.dc.b	$FA,$C6,$1D,$4F,$89,$FC,$EE,$0B,$FE,$9D,$39,$AF,$01,$C2,$B7,$3A
	.dc.b	$7F,$4D,$5C,$67,$47,$0A,$E4,$FF,$59,$8D,$46,$21,$18,$9E,$75,$D6
	.dc.b	$79,$47,$81,$EA,$73,$8F,$17,$1E,$8F,$D6,$DF,$B5,$B5,$FB,$5A,$D9
	.dc.b	$FD,$57,$A1,$F7,$1E,$EC,$FF,$BB,$B6,$33,$3F,$E0,$FF,$B4,$C1,$F3
	.dc.b	$E0,$D2,$47,$5F,$A9,$E4,$47,$D0,$7D,$2F,$CA,$9E,$36,$63,$FC,$21
	.dc.b	$37,$6F,$8F,$5A,$9E,$3A,$A3,$11,$F2,$B6,$FB,$31,$8A,$E3,$87,$B7
	.dc.b	$C1,$7C,$6D,$3F,$10,$E3,$0F,$EF,$0D,$7F,$90,$EC,$8C,$47,$F8,$75
	.dc.b	$7C,$DF,$52,$47,$9B,$F0,$9F,$F5,$A3,$F0,$2B,$94,$EA,$3F,$C9,$0F
	.dc.b	$E1,$E7,$71,$FE,$55,$F3,$BE,$A7,$90,$69,$F8,$DC,$BF,$D3,$93,$FE
	.dc.b	$5D,$8C,$B4,$F8,$F1,$A7,$FC,$C1,$AF,$42,$3F,$33,$E7,$3F,$34,$FC
	.dc.b	$B3,$F4,$EF,$EA,$77,$34,$93,$F9,$06,$CF,$F9,$12,$77,$39,$3F,$6D
	.dc.b	$D8,$F0,$FF,$20,$DA,$FF,$4D,$C7,$EA,$6D,$F3,$3A,$89,$3A,$3D,$11
	.dc.b	$F9,$9F,$AB,$FE,$79,$C5,$FB,$3E,$04,$63,$2F,$E1,$2D,$BF,$F7,$82
	.dc.b	$F7,$79,$76,$49,$FB,$13,$BB,$E7,$5F,$C1,$2D,$F6,$EF,$8E,$C3,$70
	.dc.b	$FC,$FC,$EA,$7F,$85,$F6,$BE,$7D,$EF,$E1,$B7,$FE,$6C,$9D,$4F,$F6
	.dc.b	$86,$C6,$67,$6F,$F7,$68,$CB,$63,$2F,$F9,$22,$32,$D8,$CB,$FC,$B3
	.dc.b	$C2,$7D,$2F,$F0,$36,$C7,$E6,$5A,$FE,$17,$4F,$F2,$BF,$2E,$EA,$4E
	.dc.b	$DE,$51,$FC,$37,$EB,$3E,$24,$F1,$6F,$0F,$D7,$6E,$BF,$CF,$97,$D5
	.dc.b	$F1,$DA,$5F,$EA,$E7,$6B,$24,$70,$8F,$D8,$7B,$CF,$8C,$FE,$3C,$D8
	.dc.b	$CC,$7E,$3F,$EE,$A4,$7E,$8D,$AF,$E8,$FE,$0D,$F9,$5F,$FA,$D2,$34
	.dc.b	$43,$F5,$1D,$9F,$A6,$AF,$E9,$35,$C3,$D6,$FE,$5D,$E9,$BE,$BA,$93
	.dc.b	$0F,$E0,$D0,$F6,$F1,$1F,$3E,$38,$3A,$9E,$73,$FC,$E7,$E2,$1F,$35
	.dc.b	$49,$5F,$C4,$63,$FC,$F9,$C9,$FA,$57,$71,$99,$DB,$F7,$63,$7C,$9F
	.dc.b	$DC,$3E,$2B,$FC,$6F,$7F,$AD,$6C,$67,$D4,$FD,$C6,$33,$0F,$E0,$5E
	.dc.b	$1F,$AD,$B1,$FC,$1C,$8E,$08,$7F,$94,$38,$3F,$87,$C7,$F7,$DF,$5F
	.dc.b	$94,$FE,$45,$B9,$7E,$FD,$B8,$FF,$A8,$BE,$31,$F6,$FF,$D7,$7E,$3B
	.dc.b	$F8,$76,$1F,$CF,$1A,$FF,$B5,$DA,$FF,$83,$9A,$FF,$9C,$FE,$B9,$F1
	.dc.b	$9F,$E0,$4F,$61,$18,$9F,$EB,$B1,$89,$D7,$D7,$3E,$03,$FB,$77,$E2
	.dc.b	$7F,$95,$BF,$D5,$37,$6F,$E3,$0D,$D7,$FD,$77,$F2,$9C,$6B,$AB,$5F
	.dc.b	$F3,$AE,$48,$C4,$7F,$B1,$A3,$18,$7F,$9C,$1B,$1F,$EF,$86,$BC,$A7
	.dc.b	$8B,$E0,$9E,$11,$88,$FD,$77,$CA,$BE,$BF,$FC,$10,$D8,$FE,$BC,$D7
	.dc.b	$F8,$37,$C0,$3F,$13,$FF,$07,$36,$3F,$DC,$CD,$7F,$B7,$B5,$FF,$6F
	.dc.b	$F8,$9F,$E5,$99,$FF,$1F,$4E,$67,$FC,$86,$D7,$F8,$13,$67,$84,$7F
	.dc.b	$BC,$FE,$13,$FC,$98,$FF,$19,$6E,$DF,$D1,$DB,$AF,$F9,$3B,$EB,$7F
	.dc.b	$A2,$57,$F7,$96,$BF,$E0,$46,$BE,$66,$BC,$7D,$0F,$F2,$A4,$62,$3F
	.dc.b	$8B,$37,$03,$4C,$7C,$E3,$C4,$FF,$1A,$36,$BF,$C1,$4D,$7F,$99,$7C
	.dc.b	$EF,$F8,$D9,$BF,$05,$FE,$63,$8C,$47,$CA,$6B,$FE,$6A,$6B,$FC,$A5
	.dc.b	$B8,$1A,$49,$FF,$37,$43,$EB,$3F,$C9,$0D,$AF,$F0,$F3,$5F,$E8,$DD
	.dc.b	$DF,$83,$DB,$FB,$A3,$F8,$58,$FB,$AD,$7F,$CF,$AD,$7F,$9E,$36,$7F
	.dc.b	$CF,$3C,$5F,$E7,$F8,$C1,$1F,$F4,$17,$A4,$FB,$CF,$F2,$E3,$6F,$FC
	.dc.b	$64,$D7,$FA,$D7,$D0,$38,$BF,$CC,$4D,$8F,$F7,$09,$FD,$35,$C3,$FE
	.dc.b	$44,$6B,$F9,$66,$BF,$E9,$66,$BC,$63,$FD,$2D,$18,$C4,$DF,$A1,$7D
	.dc.b	$47,$F9,$CD,$B7,$FE,$49,$6B,$FD,$C7,$F2,$C7,$CD,$FF,$9E,$5B,$1E
	.dc.b	$8C,$11,$88,$FF,$2D,$B5,$FD,$6B,$5F,$F5,$6F,$67,$F2,$A7,$FB,$63
	.dc.b	$64,$E4,$7D,$27,$F4,$A9,$FF,$02,$46,$C9,$E3,$7F,$27,$9D,$70,$FF
	.dc.b	$A8,$B1,$FC,$33,$78,$BF,$D5,$11,$8E,$96,$33,$EA,$FF,$B8,$3F,$DB
	.dc.b	$A7,$4F,$FA,$F1,$BC,$9F,$12,$FF,$D7,$32,$AE,$38,$7F,$AC,$BE,$A1
	.dc.b	$F3,$9F,$EB,$76,$D7,$FA,$39,$AF,$F7,$CF,$A8,$7D,$9E,$EA,$4E,$DF
	.dc.b	$F6,$04,$EA,$3F,$D2,$ED,$7F,$AE,$35,$F6,$9F,$ED,$47,$0F,$FB,$49
	.dc.b	$AF,$FB,$47,$C7,$FF,$61,$91,$EB,$EF,$D7,$FC,$23,$F2,$1F,$D5,$1F
	.dc.b	$F5,$3B,$5F,$54,$7C,$4F,$AB,$F8,$99,$FE,$7A,$51,$87,$F4,$2D,$7C
	.dc.b	$E6,$B4,$FF,$6C,$BA,$E7,$7D,$67,$3A,$FC,$DB,$63,$33,$B8,$F0,$3F
	.dc.b	$7B,$E5,$2A,$69,$1F,$F0,$26,$5F,$E8,$9E,$3F,$6F,$C4,$7F,$BC,$73
	.dc.b	$FF,$76,$EC,$FA,$CC,$57,$FB,$DE,$AE,$1F,$EF,$B6,$F2,$31,$E9,$3F
	.dc.b	$7E,$F3,$18,$C7,$AF,$D1,$FF,$A9,$AB,$EB,$35,$FB,$4D,$B6,$48,$C7
	.dc.b	$27,$FA,$EF,$3F,$C7,$7C,$67,$8A,$B5,$18,$26,$1F,$E4,$CA,$E0,$FE
	.dc.b	$4E,$FD,$31,$8C,$55,$FA,$8C,$7D,$9F,$3F,$EA,$3E,$0F,$5F,$EE,$3E
	.dc.b	$4B,$15,$F8,$69,$B9,$7F,$B0,$6B,$53,$93,$FD,$B5,$F5,$9F,$E4,$E6
	.dc.b	$FB,$9D,$3B,$5F,$65,$F3,$BC,$05,$EF,$1F,$F4,$43,$75,$FF,$5B,$37
	.dc.b	$1F,$DF,$E5,$FC,$33,$0E,$53,$9B,$FF,$3F,$53,$DF,$FA,$18,$FC,$06
	.dc.b	$C6,$4F,$E4,$45,$9F,$F7,$DF,$86,$F8,$EF,$33,$67,$FD,$EC,$6F,$3F
	.dc.b	$F6,$54,$66,$BF,$A3,$FC,$D7,$83,$C6,$34,$97,$FC,$C6,$32,$D8,$CC
	.dc.b	$FF,$49,$7D,$3E,$47,$FB,$0A,$7F,$AA,$F8,$1F,$26,$F8,$DF,$FA,$BE
	.dc.b	$B4,$FF,$6C,$E2,$F8,$3C,$86,$93,$0F,$E8,$51,$96,$C6,$67,$F7,$7B
	.dc.b	$2B,$51,$8F,$7F,$F2,$6F,$FA,$FD,$BA,$FF,$7A,$BF,$BD,$18,$C3,$FC
	.dc.b	$19,$F1,$5F,$53,$CE,$34,$91,$FD,$56,$32,$D8,$CE,$1E,$49,$CB,$E6
	.dc.b	$F3,$FF,$DD,$D5,$FE,$22,$F8,$9F,$C7,$56,$B2,$FF,$80,$15,$AB,$FF
	.dc.b	$6B,$7D,$63,$6F,$C3,$B7,$1F,$EC,$CD,$9F,$86,$D7,$FC,$A2,$D7,$FA
	.dc.b	$9E,$C8,$C4,$7F,$B4,$5B,$19,$AF,$CA,$F9,$CF,$AD,$FD,$23,$70,$FE
	.dc.b	$BB,$97,$EE,$31,$E5,$7A,$BF,$D5,$7A,$CF,$AD,$FC,$FB,$70,$FF,$34
	.dc.b	$C6,$08,$E0,$97,$FE,$6C,$8F,$D3,$FB,$3F,$C6,$E7,$D6,$D9,$FF,$34
	.dc.b	$CF,$EE,$63,$1D,$9F,$E8,$DF,$8A,$FE,$3C,$DB,$FF,$45,$7B,$7F,$EE
	.dc.b	$D6,$CE,$A3,$18,$7F,$A3,$63,$04,$9F,$F0,$B7,$D3,$3E,$1B,$F8,$43
	.dc.b	$63,$FD,$3F,$38,$9F,$E3,$3F,$59,$F5,$F7,$F9,$19,$23,$17,$FC,$2B
	.dc.b	$E0,$5F,$5B,$C3,$87,$F8,$A7,$C2,$7E,$33,$5F,$F6,$07,$4F,$F0,$A8
	.dc.b	$C4,$7F,$B0,$63,$F8,$6F,$D6,$FF,$11,$E6,$3F,$C4,$F4,$7C,$3F,$FB
	.dc.b	$EF,$4F,$B6,$7E,$1A,$7F,$8C,$77,$3E,$A3,$F0,$ED,$D3,$FA,$34,$60
	.dc.b	$8E,$09,$5F,$E5,$29,$E0,$FF,$8F,$7B,$9F,$4F,$FB,$23,$FA,$DD,$FC
	.dc.b	$2D,$5A,$2B,$77,$E6,$9C,$FC,$63,$E1,$7F,$9C,$36,$75,$68,$FE,$45
	.dc.b	$C9,$F9,$78,$C5,$70,$25,$70,$AD,$BF,$A8,$7A,$5A,$F9,$7D,$6F,$F1
	.dc.b	$C6,$F8,$BF,$D2,$63,$FC,$71,$7E,$27,$4F,$EC,$9B,$5E,$E6,$E2,$78
	.dc.b	$C7,$79,$E1,$5F,$1B,$EA,$A4,$7C,$3F,$91,$7D,$6F,$FB,$FA,$3E,$37
	.dc.b	$07,$D9,$AD,$E1,$CC,$FE,$51,$BA,$3F,$C1,$B8,$72,$F8,$DD,$EB,$F1
	.dc.b	$FE,$53,$FC,$EF,$97,$CC,$6E,$3C,$F2,$ED,$39,$F8,$4F,$ED,$59,$7D
	.dc.b	$8A,$F0,$6F,$DF,$8F,$3E,$B6,$99,$87,$D8,$BE,$19,$FD,$77,$F9,$B6
	.dc.b	$1F,$84,$73,$3F,$C7,$3E,$B7,$FC,$C6,$FE,$01,$B3,$E0,$5F,$5B,$FD
	.dc.b	$F5,$CD,$FE,$0D,$B5,$CF,$E1,$35,$5B,$7F,$2B,$F1,$FF,$A0,$24,$FE
	.dc.b	$77,$89,$18,$C7,$FC,$9F,$BF,$EA,$63,$11,$FE,$68,$8C,$57,$FA,$D8
	.dc.b	$7B,$9F,$DA,$33,$FE,$59,$18,$F4,$BC,$2F,$E8,$4C,$1F,$9E,$F9,$F7
	.dc.b	$8F,$FE,$19,$6C,$67,$7C,$4C,$8E,$0E,$E3,$19,$7D,$6A,$FE,$A9,$5F
	.dc.b	$D2,$1B,$1F,$88,$F4,$BB,$78,$52,$BC,$4F,$EE,$A3,$13,$F8,$5D,$BF
	.dc.b	$8A,$BF,$90,$9F,$E1,$24,$AF,$94,$D8,$F9,$33,$C3,$5F,$E2,$53,$FE
	.dc.b	$28,$FA,$07,$D4,$7F,$8D,$DB,$87,$E8,$2F,$CF,$CC,$9C,$E3,$F9,$25
	.dc.b	$F1,$7C,$4F,$F3,$5F,$D6,$FF,$27,$8C,$E1,$FC,$29,$B5,$C6,$B7,$EF
	.dc.b	$71,$AF,$E1,$FF,$4B,$F9,$4A,$FE,$32,$DD,$5F,$5B,$E2,$E4,$FE,$25
	.dc.b	$3B,$D0,$D3,$84,$EF,$0F,$EF,$31,$BF,$AA,$FE,$EE,$4F,$1F,$FC,$96
	.dc.b	$FF,$36,$E8,$7C,$E7,$FA,$19,$B7,$FC,$92,$7F,$97,$67,$F9,$8A,$FF
	.dc.b	$30,$43,$EB,$69,$C6,$2D,$71,$8B,$AF,$F0,$1C,$3A,$BF,$BE,$BE,$55
	.dc.b	$DF,$E2,$7C,$A4,$7F,$DF,$D1,$FE,$98,$9F,$F6,$5E,$A8,$7D,$6F,$90
	.dc.b	$F2,$DD,$BF,$E7,$BC,$BF,$E6,$5C,$38,$B5,$FF,$0F,$B5,$FE,$96,$DC
	.dc.b	$3F,$13,$F5,$BF,$E7,$F8,$37,$E0,$F0,$CB,$EE,$E5,$A4,$BF,$E3,$B5
	.dc.b	$CA,$3C,$B5,$C6,$73,$EE,$7F,$A0,$28,$D2,$57,$FB,$5A,$7C,$FD,$DE
	.dc.b	$53,$CD,$36,$7F,$B1,$4B,$E9,$3F,$47,$C5,$FE,$D3,$ED,$FF,$7A,$C6
	.dc.b	$27,$FD,$A9,$1F,$E9,$6F,$9E,$F3,$46,$01,$CF,$2F,$F2,$83,$93,$17
	.dc.b	$D5,$DD,$F5,$F8,$E6,$1D,$75,$FF,$22,$46,$2B,$78,$74,$8C,$47,$43
	.dc.b	$B5,$F3,$BE,$59,$77,$9D,$57,$0A,$D4,$74,$C7,$C9,$EA,$33,$F6,$DE
	.dc.b	$F2,$F0,$65,$F1,$1B,$87,$13,$E1,$5F,$1C,$FD,$35,$BC,$7C,$59,$78
	.dc.b	$2F,$C7,$3E,$67,$D3,$87,$47,$27,$5C,$7E,$16,$3D,$73,$E7,$5B,$C3
	.dc.b	$A1,$EC,$9D,$5F,$4C,$76,$47,$CF,$9F,$BB,$2F,$34,$6C,$99,$78,$27
	.dc.b	$51,$E3,$9D,$CF,$C5,$AF,$93,$5A,$AF,$49,$30,$F2,$46,$5F,$3E,$3F
	.dc.b	$86,$9F,$C4,$9F,$2C,$A7,$DD,$2C,$7E,$4D,$BA,$FF,$9A,$5F,$D0,$47
	.dc.b	$FB,$53,$D9,$FD,$14,$7E,$76,$74,$F0,$8C,$CE,$E3,$F1,$33,$CF,$1F
	.dc.b	$C3,$37,$2F,$F7,$AF,$BD,$FE,$E9,$AF,$B7,$39,$7F,$36,$D7,$C1,$87
	.dc.b	$E3,$27,$84,$76,$8E,$24,$7F,$13,$1A,$BF,$0C,$E5,$B3,$AF,$86,$F8
	.dc.b	$F1,$82,$61,$FB,$8C,$3C,$9A,$B1,$E9,$39,$C7,$A3,$71,$E8,$60,$F5
	.dc.b	$C5,$AE,$58,$F8,$2B,$86,$3E,$0B,$D9,$C8,$F7,$11,$D5,$72,$8E,$C7
	.dc.b	$C6,$35,$3F,$3A,$BB,$47,$4A,$EC,$ED,$FD,$1E,$1E,$C8,$D5,$79,$08
	.dc.b	$AF,$C0,$FF,$17,$E2,$FF,$00,$E9,$F9,$0E,$CF,$F3,$E3,$FE,$04,$9F
	.dc.b	$E8,$AE,$78,$3F,$A1,$96,$3F,$A1,$BF,$8E,$AF,$F0,$2C,$7E,$61,$AF
	.dc.b	$F4,$C8,$FC,$9F,$53,$E3,$7F,$3B,$87,$FB,$86,$31,$5D,$5A,$FF,$9C
	.dc.b	$A7,$F4,$D3,$B7,$8D,$7F,$30,$AD,$D7,$F9,$AE,$72,$A3,$C1,$C5,$D6
	.dc.b	$73,$1C,$2F,$A5,$66,$A3,$FE,$B1,$AF,$DB,$3F,$E5,$16,$BE,$16,$BF
	.dc.b	$EB,$0E,$87,$CC,$FC,$7C,$BF,$BA,$36,$33,$3F,$E0,$BF,$AD,$C1,$F2
	.dc.b	$E0,$D2,$47,$F8,$03,$EB,$79,$11,$FF,$4B,$FD,$6B,$C7,$C4,$7F,$82
	.dc.b	$E6,$E9,$F3,$EB,$53,$C2,$BE,$84,$62,$3F,$D2,$8D,$9F,$F1,$7F,$C7
	.dc.b	$FA,$FC,$23,$F8,$16,$D7,$F7,$1D,$3F,$C6,$73,$93,$5F,$5B,$C4,$93
	.dc.b	$FE,$AE,$FA,$DF,$F5,$53,$FE,$42,$8C,$BF,$EA,$0B,$47,$E1,$67,$73
	.dc.b	$FE,$53,$F3,$BE,$77,$93,$FE,$82,$8F,$F2,$84,$7E,$2F,$4F,$F4,$CC
	.dc.b	$60,$87,$13,$FD,$7E,$E9,$FF,$2F,$72,$74,$23,$FE,$C4,$FA,$44,$7E
	.dc.b	$5D,$B1,$FE,$F7,$6B,$FE,$9A,$6B,$FE,$99,$6B,$F9,$96,$BF,$97,$6F
	.dc.b	$29,$A6,$B5,$FF,$03,$F8,$7F,$8D,$E9,$34,$F2,$7D,$6D,$C3,$FD,$3D
	.dc.b	$15,$FF,$68,$5A,$7F,$2D,$F5,$BF,$CD,$B8,$BF,$DF,$CD,$CF,$FC,$3B
	.dc.b	$92,$31,$82,$31,$EF,$F2,$EF,$93,$FE,$24,$C3,$9C,$76,$9F,$E1,$58
	.dc.b	$3E,$B5,$F2,$18,$6E,$1F,$EB,$76,$BF,$EF,$1E,$D7,$D6,$BE,$03,$F8
	.dc.b	$A4,$61,$FE,$C2,$D9,$FF,$22,$35,$FE,$DA,$D7,$FB,$8B,$5F,$F1,$DB
	.dc.b	$53,$4C,$69,$FE,$49,$FA,$DF,$DF,$78,$E6,$98,$E5,$B7,$F0,$F4,$F9
	.dc.b	$75,$B8,$C1,$F2,$DC,$3F,$F0,$01,$FD,$94,$7D,$2F,$AC,$F8,$7C,$1F
	.dc.b	$90,$C5,$18,$9D,$6D,$FC,$46,$1F,$5B,$62,$63,$E0,$BE,$31,$B9,$FE
	.dc.b	$31,$F0,$0F,$8D,$A4,$BF,$D1,$7D,$3F,$E8,$48,$E0,$F5,$FC,$23,$EB
	.dc.b	$7F,$D5,$71,$93,$F2,$33,$C1,$6B,$FA,$AA,$E1,$EB,$7F,$29,$F4,$DF
	.dc.b	$5B,$49,$7E,$98,$FF,$69,$B6,$32,$F4,$E0,$FC,$CC,$F2,$8E,$1E,$D7
	.dc.b	$F3,$5F,$89,$7B,$74,$93,$F1,$63,$FC,$DF,$9F,$CF,$77,$19,$9D,$BD
	.dc.b	$CF,$D2,$47,$0C,$3F,$53,$F1,$5F,$E3,$2B,$FB,$0D,$8C,$FA,$9F,$C2
	.dc.b	$F1,$98,$7F,$0B,$7F,$B4,$F8,$3E,$E8,$FE,$19,$18,$23,$B2,$1F,$DC
	.dc.b	$FD,$6E,$18,$FE,$1D,$B8,$7E,$F1,$B3,$FE,$07,$F8,$C7,$D6,$FF,$59
	.dc.b	$F8,$EF,$E1,$B3,$FC,$F5,$AF,$F7,$F6,$BF,$ED,$26,$BF,$EA,$4F,$59
	.dc.b	$F4,$AF,$8C,$FF,$05,$46,$1B,$19,$9F,$D1,$CB,$E4,$5F,$01,$FD,$AF
	.dc.b	$EB,$7F,$37,$5F,$99,$6B,$F4,$7B,$D1,$8D,$FF,$9A,$E0,$F9,$4E,$77
	.dc.b	$D9,$AF,$F9,$C2,$35,$87,$F9,$A9,$B1,$FE,$CD,$6B,$B9,$E2,$FF,$9E
	.dc.b	$A3,$2D,$8C,$BF,$0D,$AF,$F2,$9F,$95,$7D,$5F,$F8,$1D,$B1,$FB,$56
	.dc.b	$BE,$DF,$7D,$F5,$BF,$E0,$D6,$C7,$F7,$16,$BF,$E3,$36,$BF,$E8,$ED
	.dc.b	$3F,$CC,$93,$FE,$39,$9C,$CF,$F3,$E6,$BF,$CF,$5B,$3C,$23,$FB,$F7
	.dc.b	$C2,$7F,$91,$DF,$A0,$DD,$3F,$8A,$37,$97,$FA,$4F,$D6,$BC,$FF,$F1
	.dc.b	$23,$6F,$FC,$10,$D7,$FC,$94,$D7,$D0,$D7,$9C,$7F,$94,$63,$11,$FD
	.dc.b	$4D,$B8,$1B,$F3,$8F,$0F,$FC,$68,$DA,$FE,$44,$D7,$F1,$5F,$3B,$FE
	.dc.b	$36,$6F,$C1,$7F,$97,$E3,$11,$F2,$DA,$FF,$66,$6B,$FD,$95,$B8,$1A
	.dc.b	$49,$FF,$35,$43,$EA,$FF,$C9,$0D,$AF,$E6,$CD,$7F,$3D,$EC,$FE,$13
	.dc.b	$6F,$E8,$4F,$E1,$A3,$EF,$35,$FE,$F2,$D7,$FB,$C3,$67,$FC,$CD,$C5
	.dc.b	$FE,$7B,$8C,$11,$FF,$3F,$7A,$2F,$AD,$FF,$2E,$36,$FF,$A5,$35,$FD
	.dc.b	$97,$D0,$38,$BF,$CC,$4D,$8F,$F2,$33,$5F,$F4,$43,$5F,$CB,$35,$FF
	.dc.b	$08,$B5,$FF,$08,$37,$13,$4C,$7D,$0B,$EA,$7F,$CE,$4D,$AF,$EC,$0D
	.dc.b	$7F,$84,$7D,$6B,$E6,$FF,$CF,$0D,$8F,$F2,$E3,$5F,$F4,$F3,$5F,$D6
	.dc.b	$35,$FF,$13,$F5,$FF,$2C,$C7,$FD,$65,$E2,$FF,$A0,$E3,$F9,$78,$FA
	.dc.b	$6F,$E9,$75,$FC,$66,$38,$12,$7F,$93,$CE,$B8,$7F,$D5,$18,$FE,$0D
	.dc.b	$BC,$5F,$EA,$98,$C7,$4B,$19,$F5,$9F,$DC,$9F,$EA,$DC,$9E,$FB,$E0
	.dc.b	$48,$FF,$09,$89,$E1,$87,$FA,$DF,$EA,$1F,$35,$FE,$B7,$6D,$7F,$86
	.dc.b	$B0,$46,$23,$BF,$D4,$BE,$B7,$85,$C9,$E1,$5B,$C3,$FD,$89,$7B,$9E
	.dc.b	$8D,$7D,$87,$FB,$05,$C3,$8C,$11,$88,$D6,$1F,$ED,$2F,$90,$FE,$A7
	.dc.b	$1F,$EC,$DF,$1F,$FD,$88,$7F,$B4,$6B,$FC,$99,$19,$21,$FE,$CD,$8D
	.dc.b	$7B,$FF,$EE,$38,$7D,$6F,$C7,$CB,$CB,$1E,$68,$C3,$83,$7A,$FF,$88
	.dc.b	$1C,$46,$9E,$2F,$E6,$7E,$B7,$84,$76,$E8,$46,$27,$53,$CA,$72,$F0
	.dc.b	$1F,$29,$5A,$65,$C1,$FE,$DF,$F8,$8F,$F4,$DE,$3F,$EE,$F6,$F1,$FC
	.dc.b	$FF,$AD,$B7,$FE,$ED,$DF,$FD,$23,$3F,$EF,$7A,$CF,$0F,$FB,$FD,$BC
	.dc.b	$0C,$33,$7C,$30,$FC,$2F,$D1,$FF,$A9,$30,$FB,$2D,$B3,$4D,$71,$BE
	.dc.b	$33,$BB,$73,$FC,$3C,$E2,$64,$7E,$39,$F1,$C6,$30,$FD,$1E,$2F,$9E
	.dc.b	$F2,$9E,$EE,$BF,$DB,$7C,$96,$31,$FC,$F5,$6D,$B5,$BE,$BF,$F7,$2C
	.dc.b	$60,$FF,$37,$7D,$67,$F9,$39,$BE,$E6,$FE,$B7,$E5,$F8,$0A,$D6,$9F
	.dc.b	$E1,$B6,$F2,$7F,$9F,$DB,$87,$F8,$B7,$16,$1C,$A7,$37,$FE,$B5,$A7
	.dc.b	$BD,$FE,$D0,$C7,$FD,$B2,$D8,$C9,$FE,$F9,$2C,$FE,$6A,$5E,$1B,$E3
	.dc.b	$7C,$ED,$AF,$F4,$51,$C3,$3F,$F4,$A4,$66,$FF,$A1,$FC,$D7,$87,$C6
	.dc.b	$34,$97,$FD,$11,$B3,$FD,$11,$AF,$F8,$67,$C8,$FF,$60,$4F,$F5,$0F
	.dc.b	$03,$E4,$DF,$48,$C1,$FE,$9D,$E9,$FA,$D7,$EB,$6B,$FD,$9B,$8B,$E2
	.dc.b	$46,$23,$FB,$3C,$60,$93,$FB,$9D,$55,$BA,$FF,$71,$7B,$FF,$96,$AF
	.dc.b	$F5,$63,$63,$FB,$BD,$FD,$E8,$C6,$BF,$EB,$9F,$35,$F1,$7F,$55,$B8
	.dc.b	$77,$6E,$1F,$4D,$AF,$F8,$46,$78,$12,$78,$3F,$7F,$CF,$FF,$76,$C7
	.dc.b	$F8,$7B,$E2,$7F,$1F,$7D,$30,$FF,$65,$56,$B0,$FF,$35,$7D,$6F,$FB
	.dc.b	$83,$0F,$80,$D8,$FC,$23,$5F,$F2,$63,$5F,$EA,$1A,$A3,$10,$8C,$70
	.dc.b	$7E,$4E,$73,$18,$AF,$F2,$6F,$D6,$FE,$93,$CF,$3F,$A6,$AF,$16,$5F
	.dc.b	$E4,$E8,$FD,$4F,$AC,$FA,$DF,$CF,$37,$0F,$F3,$24,$60,$8E,$09,$9F
	.dc.b	$F9,$9E,$3F,$4D,$EC,$FF,$15,$C3,$FD,$DC,$DA,$FF,$33,$4F,$EE,$23
	.dc.b	$14,$F0,$3F,$D1,$DF,$14,$BF,$F4,$47,$B7,$FE,$EB,$6C,$66,$BF,$D1
	.dc.b	$4D,$7F,$D1,$0D,$7F,$D1,$0D,$7B,$4F,$37,$8C,$FE,$57,$E9,$5F,$0D
	.dc.b	$FC,$31,$B1,$FE,$9F,$BC,$4F,$F1,$DF,$AD,$F9,$73,$FE,$B7,$6D,$FA
	.dc.b	$FE,$05,$F5,$BB,$E4,$FE,$2F,$F0,$9F,$92,$AF,$BE,$D7,$EF,$37,$94
	.dc.b	$DF,$AD,$FE,$2B,$07,$F8,$BE,$EF,$87,$FF,$7F,$E9,$FE,$19,$E3,$F8
	.dc.b	$D7,$BF,$23,$FA,$6F,$D3,$FE,$33,$C0,$34,$9A,$FF,$4B,$6E,$1F,$DC
	.dc.b	$67,$73,$93,$F9,$27,$7B,$EA,$FF,$6C,$76,$8B,$5E,$33,$F3,$BF,$19
	.dc.b	$FE,$9B,$71,$C1,$FF,$00,$3E,$17,$F9,$DC,$FF,$21,$D7,$D2,$EE,$7F
	.dc.b	$DA,$CD,$8C,$CF,$0F,$23,$8B,$FD,$13,$EB,$7F,$8F,$DF,$FB,$75,$B5
	.dc.b	$F1,$1F,$F1,$B6,$1E,$2D,$D1,$A8,$FD,$9B,$6B,$F6,$4D,$9D,$FB,$5D
	.dc.b	$FC,$8F,$9B,$F5,$53,$0F,$6F,$C8,$BE,$B7,$FD,$F5,$1F,$E3,$FE,$35
	.dc.b	$6F,$3F,$F2,$6C,$62,$78,$D7,$09,$CD,$7F,$60,$F9,$4F,$F3,$85,$7A
	.dc.b	$5B,$3A,$8C,$7C,$1F,$F6,$6A,$FA,$D1,$EF,$6C,$FF,$C0,$17,$F9,$6C
	.dc.b	$7F,$12,$73,$B7,$E2,$9B,$19,$E3,$E9,$EB,$75,$7F,$B9,$FD,$6F,$F9
	.dc.b	$7A,$7E,$63,$70,$61,$F9,$48,$7D,$6F,$0F,$84,$97,$C1,$E7,$18,$BF
	.dc.b	$8B,$F1,$DF,$95,$25,$7E,$57,$74,$62,$7F,$C9,$7D,$5F,$9E,$8C,$47
	.dc.b	$F9,$7E,$77,$5F,$E4,$D3,$FC,$49,$E9,$FF,$8D,$9A,$7D,$7A,$C7,$B7
	.dc.b	$FD,$9E,$FD,$73,$FC,$E3,$E7,$DE,$47,$F8,$51,$B1,$F8,$D7,$EE,$4F
	.dc.b	$03,$F9,$E3,$A7,$1D,$08,$C6,$FF,$BA,$6D,$77,$E2,$ED,$87,$7F,$4E
	.dc.b	$4F,$E3,$7D,$4B,$97,$4C,$3F,$AC,$4F,$9E,$BF,$81,$D6,$B1,$FE,$59
	.dc.b	$BF,$F0,$49,$FE,$04,$4A,$F9,$CD,$C3,$E6,$EF,$FC,$66,$B8,$56,$8F
	.dc.b	$F0,$07,$D0,$3E,$A3,$FC,$6E,$D8,$FD,$16,$18,$9C,$FC,$54,$9F,$F5
	.dc.b	$27,$D6,$FF,$20,$B4,$E7,$2F,$8F,$18,$AD,$F8,$B9,$D3,$F9,$0F,$D2
	.dc.b	$FE,$62,$FF,$B5,$37,$47,$D6,$F8,$E7,$E8,$CE,$F8,$CD,$38,$4E,$EF
	.dc.b	$AF,$57,$F9,$4B,$EA,$BF,$B9,$93,$C7,$FF,$24,$3F,$CB,$FA,$1F,$39
	.dc.b	$FE,$7E,$6E,$DF,$C4,$67,$FA,$4E,$BF,$9A,$AE,$9F,$5B,$2B,$FD,$8D
	.dc.b	$7C,$36,$FC,$3E,$0A,$D4,$61,$ED,$78,$CF,$FA,$1D,$7F,$A2,$6B,$FD
	.dc.b	$87,$B2,$31,$1F,$E1,$26,$BF,$E1,$16,$C6,$FE,$B7,$E0,$1F,$07,$E2
	.dc.b	$AE,$7F,$CA,$F9,$7F,$CC,$78,$76,$6B,$FE,$1E,$6B,$FE,$1D,$6D,$FE
	.dc.b	$1F,$EB,$7F,$CF,$B0,$6F,$C1,$E1,$97,$DD,$CB,$49,$7F,$AC,$AE,$51
	.dc.b	$E5,$AE,$33,$9F,$7B,$FC,$FD,$46,$92,$BF,$CB,$51,$FC,$07,$4F,$ED
	.dc.b	$AD,$8F,$64,$FA,$3C,$0F,$F5,$FD,$BE,$8F,$F4,$5C,$68,$FF,$66,$F7
	.dc.b	$7F,$BD,$27,$31,$83,$FC,$49,$F3,$DE,$87,$07,$3C,$97,$CF,$E3,$BF
	.dc.b	$84,$4F,$5B,$E7,$18,$8E,$13,$BC,$3C,$5E,$A3,$1E,$D7,$BC,$38,$D7
	.dc.b	$97,$1E,$F8,$F8,$A7,$2F,$9B,$7F,$F1,$A5,$F1,$AD,$B7,$2F,$05,$79
	.dc.b	$67,$38,$78,$B1,$F0,$BD,$EB,$75,$E8,$3C,$15,$C1,$F0,$6B,$AC,$BC
	.dc.b	$71,$BC,$7C,$0F,$CB,$CF,$DD,$9F,$9B,$2E,$D3,$D3,$2E,$59,$75,$BD
	.dc.b	$E5,$89,$95,$B3,$A6,$7E,$87,$D9,$1A,$CB,$C2,$7D,$4B,$E7,$3C,$63
	.dc.b	$38,$F8,$9D,$5F,$39,$F5,$4E,$48,$FC,$03,$93,$87,$66,$67,$AB,$E4
	.dc.b	$3F,$20,$7C,$02,$9F,$34,$B9,$FE,$49,$BA,$7F,$99,$1D,$9F,$9B,$70
	.dc.b	$EB,$DA,$CB,$96,$CF,$FB,$61,$FA,$D3,$CE,$BF,$1C,$DB,$FC,$9F,$BF
	.dc.b	$FE,$E7,$C3,$77,$1F,$C8,$56,$DF,$F4,$44,$ED,$CC,$EE,$FA,$BF,$90
	.dc.b	$C9,$5B,$9C,$D7,$02,$3F,$5E,$35,$87,$ED,$70,$F0,$57,$B4,$DC,$7A
	.dc.b	$3D,$3F,$9B,$7D,$09,$1C,$71,$F1,$61,$D2,$FF,$6D,$87,$4C,$BE,$34
	.dc.b	$74,$3D,$B8,$F4,$C3,$94,$64,$F1,$63,$E7,$9D,$47,$4C,$7E,$15,$70
	.dc.b	$AD,$4E,$0D,$1D,$F2,$F8,$B7,$CA,$3A,$D7,$12,$3E,$43,$91,$A3,$9D
	.dc.b	$76,$7F,$8D,$71,$FF,$80,$77,$FC,$8F,$17,$F2,$86,$F0,$7F,$43,$6D
	.dc.b	$7F,$43,$8B,$5F,$DD,$63,$F2,$ED,$7F,$A6,$47,$E4,$23,$11,$FC,$B7
	.dc.b	$A9,$F1,$5F,$9E,$C3,$FC,$3A,$DB,$D3,$69,$5C,$38,$BC,$6F,$28,$B3
	.dc.b	$B8,$C4,$7F,$A6,$9E,$0D,$8D,$11,$E3,$31,$FF,$56,$47,$F2,$C7,$FC
	.dc.b	$AE,$D7,$F6,$F3,$D1,$F8,$3D,$0F,$98,$F8,$AD,$8F,$F0,$54,$63,$EB
	.dc.b	$64,$7E,$E3,$3F,$EF,$D1,$F1,$E8,$D3,$1F,$5B,$D0,$95,$ED,$FA,$D7
	.dc.b	$8F,$88,$FF,$08,$CD,$DB,$FC,$93,$5A,$9E,$15,$F2,$A3,$11,$FE,$61
	.dc.b	$CD,$18,$AF,$F1,$BF,$C7,$FE,$03,$C4,$3F,$75,$B7,$EB,$6B,$F8,$46
	.dc.b	$BF,$E1,$C5,$F5,$BC,$C9,$1F,$5F,$EB,$7F,$D6,$2F,$FB,$06,$76,$D9
	.dc.b	$DB,$E7,$9D,$BF,$E5,$0F,$3F,$F2,$BE,$49,$A7,$E2,$34,$FF,$4E,$CF
	.dc.b	$FB,$36,$74,$D9,$DB,$F8,$99,$DB,$FE,$5C,$E4,$EA,$47,$F1,$DF,$48
	.dc.b	$8F,$CA,$B6,$3F,$DE,$CD,$7F,$D4,$2D,$7F,$16,$D7,$F4,$CD,$7F,$42
	.dc.b	$DD,$8D,$35,$C5,$FE,$95,$F0,$FF,$6B,$D2,$69,$E4,$FA,$DB,$87,$FA
	.dc.b	$86,$2B,$F8,$1B,$4F,$EF,$BE,$B7,$F9,$D7,$17,$AF,$14,$63,$3F,$F1
	.dc.b	$44,$63,$B3,$87,$BD,$C5,$99,$7F,$BE,$EF,$82,$AE,$75,$FC,$3A,$DF
	.dc.b	$5A,$F9,$0C,$36,$FF,$C6,$D9,$7C,$DA,$DD,$7F,$16,$ED,$7D,$6B,$DF
	.dc.b	$C3,$70,$FE,$CF,$3A,$9F,$EC,$AD,$8C,$CE,$DF,$EE,$71,$96,$C6,$5F
	.dc.b	$EF,$31,$96,$C6,$5F,$8B,$E2,$3E,$B7,$F9,$4B,$63,$D7,$8E,$A7,$3C
	.dc.b	$BA,$64,$E8,$F9,$15,$C5,$FF,$6A,$BF,$B7,$AF,$A7,$F5,$BF,$1E,$7F
	.dc.b	$DB,$2D,$D3,$ED,$36,$77,$F5,$BF,$CF,$23,$15,$F5,$5B,$C0,$68,$9F
	.dc.b	$E8,$7F,$00,$F8,$DA,$4B,$FD,$2F,$D3,$7E,$90,$8E,$0F,$67,$C2,$3E
	.dc.b	$B7,$FD,$67,$19,$3F,$29,$3C,$1F,$F7,$B4,$EF,$D4,$FE,$A7,$E9,$BE
	.dc.b	$B6,$92,$BE,$46,$A9,$FF,$6B,$C6,$1E,$0D,$9E,$0F,$FC,$00,$8D,$AD
	.dc.b	$FF,$45,$F1,$2F,$6F,$FB,$1B,$67,$BB,$5E,$1C,$3F,$A6,$8F,$F4,$05
	.dc.b	$72,$AE,$24,$AE,$3E,$07,$F6,$EF,$8A,$FF,$1D,$D7,$EC,$5B,$87,$EC
	.dc.b	$1B,$5F,$C1,$BD,$DF,$E2,$27,$F9,$03,$C3,$FE,$0E,$DF,$0B,$D7,$3F
	.dc.b	$C4,$23,$04,$7F,$8B,$B5,$FE,$FB,$EB,$7C,$2C,$7F,$81,$37,$1F,$E0
	.dc.b	$2D,$97,$A6,$F8,$B7,$D6,$FF,$82,$3E,$3B,$F8,$A5,$7F,$61,$6B,$FE
	.dc.b	$0A,$6B,$FE,$E2,$6B,$FE,$A4,$FA,$47,$D6,$FF,$85,$30,$46,$27,$F4
	.dc.b	$D1,$89,$D7,$C9,$3E,$07,$FC,$33,$F5,$BF,$99,$BF,$C2,$B5,$FA,$E7
	.dc.b	$F1,$57,$1B,$7F,$3B,$C1,$F2,$BC,$EB,$BB,$5F,$F3,$CC,$6A,$FF,$CE
	.dc.b	$2D,$8F,$F3,$BB,$5D,$CF,$17,$FD,$07,$19,$6C,$65,$FF,$62,$B5,$FE
	.dc.b	$55,$F2,$CF,$AB,$FF,$1A,$B6,$3F,$42,$D7,$DB,$D7,$FE,$70,$FA,$DF
	.dc.b	$F1,$DF,$C1,$1F,$F1,$C3,$67,$FC,$2C,$D7,$FC,$7E,$D7,$FD,$33,$3C
	.dc.b	$DD,$CF,$22,$3F,$E1,$FF,$84,$7C,$07,$F9,$29,$B1,$FB,$66,$BE,$6F
	.dc.b	$AD,$7A,$3F,$E5,$06,$C7,$F8,$A5,$AF,$F9,$49,$AF,$91,$AF,$F5,$66
	.dc.b	$BF,$D4,$5B,$81,$A6,$3E,$75,$E2,$FF,$98,$1B,$5F,$C3,$1A,$FC,$CF
	.dc.b	$9E,$FF,$20,$37,$E0,$BF,$CC,$51,$88,$F8,$ED,$7F,$B4,$B5,$FE,$C8
	.dc.b	$DC,$0D,$24,$FF,$9B,$61,$F5,$7F,$E7,$46,$D7,$F2,$06,$BF,$80,$F0
	.dc.b	$3F,$05,$CB,$FD,$ED,$FC,$24,$7D,$A6,$BF,$DF,$1A,$FF,$77,$6C,$FF
	.dc.b	$9D,$38,$BF,$CF,$D1,$82,$3F,$E8,$1F,$49,$F5,$BF,$E8,$86,$DF,$F3
	.dc.b	$46,$BF,$94,$FA,$17,$0F,$FA,$39,$B1,$FE,$67,$6B,$FE,$90,$6B,$F9
	.dc.b	$26,$BF,$E1,$46,$BF,$E0,$F6,$E2,$69,$8F,$A2,$7D,$4F,$FA,$79,$B5
	.dc.b	$FD,$29,$AF,$EA,$BE,$B5,$F3,$7F,$EA,$46,$C7,$F9,$F5,$AF,$FA,$99
	.dc.b	$AF,$EA,$1A,$FF,$8A,$BA,$FF,$95,$63,$FC,$AB,$C6,$FF,$57,$C7,$F2
	.dc.b	$D1,$F4,$FF,$D3,$78,$7F,$C6,$91,$C0,$93,$FC,$96,$75,$E3,$FF,$AC
	.dc.b	$72,$FC,$1D,$FF,$95,$78,$BD,$3C,$7F,$EA,$C9,$CD,$33,$9F,$3D,$FD
	.dc.b	$D1,$FF,$43,$4E,$9F,$F5,$AB,$6F,$9C,$7F,$85,$C4,$F1,$C3,$C1,$F5
	.dc.b	$0F,$9C,$FF,$5D,$B6,$BF,$D7,$0D,$7F,$D2,$9F,$52,$FA,$DD,$D4,$9D
	.dc.b	$BF,$EC,$49,$D4,$7F,$B0,$5A,$FA,$CF,$F6,$93,$87,$FD,$76,$D7,$33
	.dc.b	$FD,$A7,$E4,$3F,$A6,$C7,$FB,$27,$C9,$FF,$67,$9D,$FE,$02,$C6,$48
	.dc.b	$7A,$1F,$F7,$4A,$FA,$CF,$87,$9F,$8E,$3C,$91,$87,$06,$F4,$FF,$08
	.dc.b	$38,$8D,$3C,$5F,$CB,$FD,$6F,$08,$ED,$D0,$8C,$4E,$A7,$94,$E5,$E0
	.dc.b	$3E,$4E,$B4,$CB,$83,$CB,$C0,$D9,$F5,$BF,$E9,$98,$C3,$63,$3C,$7C
	.dc.b	$1B,$7F,$06,$B7,$39,$AE,$04,$7E,$1F,$CA,$BF,$2D,$FE,$93,$18,$6D
	.dc.b	$FC,$B9,$D7,$91,$FE,$AB,$6D,$78,$7E,$93,$C8,$47,$04,$AF,$26,$49
	.dc.b	$D4,$61,$FF,$5B,$75,$FE,$CB,$E4,$EF,$5F,$FB,$93,$0F,$EF,$AD,$BF
	.dc.b	$F0,$1B,$5F,$EF,$AD,$7F,$5D,$5C,$E7,$8D,$74,$3A,$FD,$74,$8E,$AD
	.dc.b	$AE,$AD,$7A,$B5,$EB,$F6,$3F,$DF,$67,$FC,$22,$D8,$FF,$0B,$B5,$FF
	.dc.b	$08,$B5,$FF,$5E,$C7,$F9,$A4,$FC,$AF,$7F,$FB,$7F,$DE,$FF,$6C,$36
	.dc.b	$32,$6E,$7F,$C4,$7F,$7B,$FB,$98,$FF,$10,$B6,$33,$D9,$F3,$23,$11
	.dc.b	$FE,$22,$6B,$F1,$9A,$FF,$93,$9B,$99,$CD,$FF,$3D,$F9,$BF,$2F,$8A
	.dc.b	$69,$31,$FF,$19,$FD,$E7,$F9,$ED,$BF,$01,$F7,$A3,$19,$7F,$8F,$1A
	.dc.b	$FF,$8D,$9A,$FD,$A6,$BF,$E6,$26,$E4,$73,$8B,$9D,$F4,$9F,$29,$B8
	.dc.b	$FC,$E6,$E1,$FC,$1E,$B8,$57,$E4,$BF,$2E,$7E,$25,$49,$87,$F9,$3A
	.dc.b	$32,$D8,$CD,$FF,$95,$A7,$2D,$8F,$B7,$E8,$3E,$3E,$92,$FF,$03,$19
	.dc.b	$7F,$CB,$7F,$97,$BF,$06,$FD,$2B,$63,$31,$82,$38,$26,$1F,$E6,$89
	.dc.b	$E0,$FB,$FE,$D1,$E8,$FE,$39,$B5,$F9,$26,$CE,$E7,$F6,$BF,$58,$F3
	.dc.b	$E9,$2F,$FC,$E3,$19,$6C,$66,$BF,$CF,$33,$96,$C7,$E7,$BD,$B7,$D9
	.dc.b	$D2,$5F,$F9,$99,$FE,$0B,$F9,$87,$F8,$6A,$3F,$CF,$ED,$BF,$F4,$2B
	.dc.b	$5F,$F3,$FB,$5F,$E0,$6D,$7F,$D5,$6D,$E5,$34,$C7,$BF,$FE,$F1,$AF
	.dc.b	$F4,$6B,$6B,$FD,$1B,$F2,$1F,$C6,$23,$11,$FE,$8E,$F8,$37,$D9,$FF
	.dc.b	$1C,$6D,$FF,$A5,$9B,$3F,$E9,$C6,$BF,$E9,$6F,$B4,$FA,$2D,$DF,$F9
	.dc.b	$2C,$62,$3F,$8E,$B5,$FF,$62,$FC,$4F,$FB,$DE,$FF,$D4,$6D,$8F,$F5
	.dc.b	$4B,$5F,$F5,$1B,$5F,$EF,$5F,$2A,$FB,$E5,$7F,$61,$8C,$36,$33,$3F
	.dc.b	$DB,$27,$2D,$8C,$BF,$DC,$E3,$2D,$8C,$BE,$07,$FA,$F7,$D7,$FA,$FB
	.dc.b	$4D,$26,$5F,$C9,$3E,$45,$F3,$D4,$99,$7F,$2A,$8C,$B6,$33,$8F,$EF
	.dc.b	$E7,$2D,$8F,$E5,$FF,$69,$FE,$2C,$AF,$E6,$6D,$9F,$E0,$CD,$7F,$D6
	.dc.b	$AD,$7F,$C7,$9E,$02,$31,$08,$C7,$AD,$D8,$95,$FE,$14,$F9,$67,$DA
	.dc.b	$7F,$BC,$A3,$0D,$BF,$F0,$E4,$EB,$E3,$FE,$4C,$23,$15,$C2,$73,$18
	.dc.b	$9E,$35,$CD,$FF,$1C,$46,$5F,$F1,$D6,$DF,$43,$EC,$FE,$99,$30,$FF
	.dc.b	$0A,$EE,$8F,$EA,$A7,$F7,$B8,$FE,$B2,$DA,$FE,$58,$D7,$FD,$B6,$D7
	.dc.b	$E3,$B5,$F8,$CD,$CF,$B7,$A3,$D8,$FE,$C7,$F6,$8E,$FA,$4C,$3F,$B4
	.dc.b	$C6,$5B,$19,$C3,$F9,$F4,$E5,$B1,$F2,$9F,$67,$DA,$5C,$4D,$25,$FF
	.dc.b	$90,$E3,$2D,$8C,$CF,$F9,$3A,$72,$FF,$8C,$F2,$7C,$8F,$E7,$89,$7F
	.dc.b	$31,$FE,$FF,$F3,$1F,$36,$3F,$C0,$6D,$AF,$EA,$CD,$7F,$CE,$1A,$23
	.dc.b	$11,$F9,$26,$BF,$91,$6C,$6F,$0E,$E7,$EC,$3C,$A9,$7F,$79,$B7,$E9
	.dc.b	$6B,$F8,$6C,$7F,$B4,$B6,$3E,$47,$E0,$92,$BE,$33,$67,$E2,$35,$FF
	.dc.b	$3B,$5A,$31,$1F,$EA,$5D,$FF,$80,$F9,$3F,$C1,$09,$7F,$92,$7F,$CC
	.dc.b	$9F,$6B,$FE,$04,$93,$49,$E7,$68,$95,$FE,$6E,$8C,$BF,$E7,$0F,$B5
	.dc.b	$FF,$2F,$E1,$FA,$26,$BE,$36,$BF,$E0,$D6,$D9,$A4,$F6,$FF,$CF,$0E
	.dc.b	$BE,$BB,$F8,$E1,$3B,$7F,$CF,$8F,$F3,$9E,$07,$D8,$D2,$4F,$F9,$09
	.dc.b	$FE,$49,$C6,$F9,$FF,$F4,$23,$74,$FF,$44,$B7,$4F,$F4,$13,$5F,$DE
	.dc.b	$B5,$F9,$CD,$7F,$C5,$0D,$D3,$F2,$F5,$FE,$91,$FC,$03,$FD,$1B,$5F
	.dc.b	$C1,$9A,$FD,$76,$BF,$E3,$16,$E5,$EC,$FA,$BF,$51,$31,$FF,$2F,$E4
	.dc.b	$8F,$F4,$F7,$E0,$3F,$A6,$9F,$F5,$1B,$70,$FF,$4E,$35,$FE,$2E,$D7
	.dc.b	$F9,$8E,$FF,$E6,$57,$F4,$AD,$73,$ED,$FF,$9B,$23,$11,$FC,$D5,$B8
	.dc.b	$9A,$49,$E0,$4B,$FF,$3B,$C6,$5F,$3F,$2F,$A8,$8E,$0D,$FC,$97,$84
	.dc.b	$8E,$0F,$F6,$77,$D3,$78,$78,$59,$7F,$B2,$7A,$BF,$DD,$D3,$98,$C0
	.dc.b	$F3,$DF,$40,$FB,$4F,$EE,$BF,$11,$FC,$16,$B9,$DF,$18,$C4,$23,$18
	.dc.b	$76,$D7,$FC,$8B,$3D,$B1,$EB,$87,$4A,$D6,$5E,$6C,$79,$E3,$CB,$0E
	.dc.b	$F8,$F2,$9F,$16,$3C,$B3,$E1,$8F,$2B,$F5,$9E,$3B,$E1,$19,$BE,$39
	.dc.b	$FB,$6F,$8D,$F6,$CA,$BC,$5B,$97,$8F,$2F,$04,$F6,$8F,$33,$B3,$19
	.dc.b	$78,$1F,$99,$87,$77,$37,$D2,$75,$1C,$71,$F2,$E7,$F1,$63,$C8,$70
	.dc.b	$24,$6B,$2E,$57,$27,$39,$FF,$B6,$31,$F0,$5F,$CC,$C7,$BE,$5E,$83
	.dc.b	$D9,$3A,$AF,$8C,$4B,$F0,$DF,$77,$A9,$88,$D6,$7F,$12,$7B,$63,$E3
	.dc.b	$AF,$C8,$9E,$A2,$9F,$1C,$B8,$7E,$4D,$BE,$EB,$E1,$BF,$36,$D8,$E2
	.dc.b	$7E,$3D,$C3,$C2,$36,$7D,$88,$E1,$7F,$8C,$6E,$3F,$EF,$CF,$5B,$FC
	.dc.b	$83,$7B,$B8,$FD,$DA,$DB,$FA,$59,$DB,$99,$DD,$F5,$7F,$17,$92,$B7
	.dc.b	$39,$AE,$04,$7F,$37,$1A,$C3,$F6,$B8,$78,$2B,$DD,$87,$69,$F4,$79
	.dc.b	$7F,$33,$0E,$84,$8E,$33,$E2,$C3,$A5,$FE,$E3,$0E,$99,$7F,$BB,$EB
	.dc.b	$A4,$FB,$71,$E9,$87,$28,$F3,$37,$1F,$3C,$EA,$7A,$57,$27,$CE,$7E
	.dc.b	$FF,$1F,$0B,$8A,$F3,$3C,$0D,$57,$37,$57,$DC,$8F,$94,$ED,$87,$A5
	.dc.b	$F2,$99,$3F,$8D,$F1,$FF,$6E,$7F,$9A,$6F,$F8,$86,$9F,$CC,$1C,$74
	.dc.b	$B9,$75,$3F,$F0,$05,$B1,$98,$DC,$63,$E3,$FF,$8E,$C7,$E7,$25,$18
	.dc.b	$8F,$F3,$1B,$5F,$F0,$A3,$54,$70,$E0,$F1,$CE,$63,$13,$FA,$06,$BD
	.dc.b	$E3,$44,$9E,$27,$BC,$7E,$05,$FA,$69,$FD,$9C,$63,$7D,$4E,$67,$FA
	.dc.b	$3C,$E5,$FE,$DB,$F6,$BF,$27,$3F,$F0,$6B,$7A,$8D,$3E,$5F,$6B,$E5
	.dc.b	$69,$2F,$FC,$1A,$FD,$0E,$F7,$87,$49,$3E,$F8,$CB,$63,$2F,$CF,$8C
	.dc.b	$B6,$3C,$DF,$68,$F8,$9A,$49,$FE,$13,$9A,$75,$1F,$37,$F0,$27,$DC
	.dc.b	$FE,$11,$B1,$98,$C1,$F8,$8F,$B5,$FF,$4D,$47,$D5,$6B,$F8,$36,$BF
	.dc.b	$EA,$46,$BF,$E5,$56,$BF,$82,$6F,$19,$A6,$32,$FF,$4C,$4F,$E0,$DA
	.dc.b	$FE,$45,$AF,$FA,$C9,$AF,$F9,$8D,$AF,$E4,$1B,$3F,$4B,$CD,$FD,$1C
	.dc.b	$63,$EA,$3F,$D3,$92,$69,$39,$FF,$4F,$E7,$FE,$27,$0F,$F0,$7B,$53
	.dc.b	$49,$CD,$C0,$91,$FB,$58,$E5,$7D,$FE,$CF,$F6,$B8,$1A,$4A,$FD,$24
	.dc.b	$65,$B1,$97,$F6,$91,$97,$FA,$8E,$EF,$B3,$7D,$2F,$E7,$5B,$3F,$86
	.dc.b	$6B,$C1,$AF,$FA,$A1,$AF,$FB,$E1,$AF,$FB,$DD,$B3,$DE,$7F,$69,$F6
	.dc.b	$2F,$9A,$A4,$BF,$EB,$BD,$49,$D4,$7C,$5F,$B6,$F9,$E4,$70,$43,$FB
	.dc.b	$2E,$A8,$D4,$61,$FD,$D7,$7F,$EA,$23,$FD,$82,$D7,$FC,$8F,$97,$F8
	.dc.b	$EF,$87,$64,$8F,$D4,$CE,$D4,$ED,$CC,$FF,$07,$F7,$9F,$63,$49,$3F
	.dc.b	$B2,$8C,$B6,$32,$FD,$68,$CB,$63,$F0,$7F,$4F,$23,$F0,$2D,$7F,$C1
	.dc.b	$BE,$D3,$EC,$3F,$D3,$D1,$FE,$08,$8C,$7A,$7F,$D0,$DB,$19,$9D,$BC
	.dc.b	$9A,$FF,$86,$A3,$6E,$67,$94,$EA,$3E,$4F,$A2,$F8,$AF,$E5,$0D,$8F
	.dc.b	$E2,$2D,$7F,$26,$D7,$FD,$E6,$D7,$9C,$72,$23,$FC,$6B,$E0,$7F,$CE
	.dc.b	$2F,$FB,$E9,$AF,$DA,$6B,$F5,$9B,$A9,$A4,$F2,$7F,$9F,$3C,$3A,$7F
	.dc.b	$80,$5F,$EC,$89,$7F,$B0,$7F,$80,$FC,$DC,$8F,$DD,$35,$FC,$9B,$5F
	.dc.b	$F2,$0F,$1A,$31,$1F,$E9,$6F,$3D,$F0,$8F,$AE,$FE,$F1,$18,$6C,$66
	.dc.b	$7F,$C8,$73,$96,$FD,$03,$E0,$F8,$B6,$33,$18,$3A,$FC,$43,$CB,$A4
	.dc.b	$9E,$B1,$97,$BF,$C5,$57,$FC,$B7,$82,$31,$1F,$E6,$08,$C5,$7A,$A7
	.dc.b	$33,$EF,$F9,$D7,$C5,$FB,$9B,$1E,$C6,$BD,$9A,$F5,$6B,$CE,$39,$11
	.dc.b	$FE,$BF,$EB,$3E,$AF,$FB,$D3,$63,$31,$83,$E7,$F9,$3F,$D1,$B2,$34
	.dc.b	$86,$89,$3F,$49,$CB,$FD,$BB,$EA,$FF,$90,$93,$C5,$FF,$05,$3F,$C9
	.dc.b	$A9,$F9,$8F,$F7,$2A,$F9,$6D,$7E,$53,$5F,$C3,$35,$FF,$6C,$36,$39
	.dc.b	$FA,$1F,$88,$8C,$1B,$EA,$7C,$AB,$EA,$FF,$C3,$8D,$C3,$FC,$0A,$D7
	.dc.b	$FA,$3B,$5F,$B9,$F3,$0F,$8B,$FC,$AB,$63,$F2,$6D,$7F,$0A,$D7,$8C
	.dc.b	$70,$23,$C1,$FF,$17,$76,$3E,$AE,$92,$7F,$C3,$31,$97,$F4,$3E,$E3
	.dc.b	$ED,$3F,$4E,$D8,$CC,$60,$8E,$08,$7E,$6A,$38,$3F,$D9,$BE,$71,$F0
	.dc.b	$BF,$E4,$16,$C7,$F8,$A9,$AE,$E3,$F6,$FE,$9F,$F1,$98,$70,$FF,$79
	.dc.b	$F0,$B2,$48,$FD,$CC,$EE,$75,$18,$9D,$3F,$DE,$3E,$A9,$F5,$5F,$E5
	.dc.b	$36,$C7,$F8,$F1,$AE,$E3,$F8,$4F,$D0,$BE,$3A,$92,$7F,$84,$C6,$5F
	.dc.b	$E1,$7F,$47,$5F,$DE,$B5,$FD,$E3,$5F,$E3,$0D,$7D,$ED,$C4,$D2,$57
	.dc.b	$FA,$3B,$E5,$7F,$A4,$50,$38,$6F,$9B,$B2,$3F,$E9,$B8,$D1,$23,$FD
	.dc.b	$37,$F4,$8F,$A3,$A4,$AF,$F4,$54,$65,$B1,$97,$FD,$35,$19,$6C,$7F
	.dc.b	$AA,$7D,$C7,$E6,$4F,$A2,$C3,$6B,$FD,$EB,$3A,$9C,$9E,$4F,$72,$4F
	.dc.b	$89,$AF,$81,$AF,$C3,$3F,$D4,$B8,$7F,$88,$9D,$1F,$EB,$BF,$A7,$F8
	.dc.b	$FF,$53,$74,$34,$93,$CC,$F9,$D3,$FD,$47,$EA,$79,$7A,$26,$92,$7F
	.dc.b	$D8,$7F,$1D,$FE,$C4,$4D,$23,$FE,$B3,$F2,$9F,$07,$F6,$F1,$B3,$FE
	.dc.b	$BE,$F8,$DF,$E7,$F3,$93,$F0,$1D,$EF,$CC,$DF,$35,$C3,$0E,$33,$FE
	.dc.b	$CC,$C7,$17,$FE,$CA,$6E,$3D,$5B,$3C,$EE,$E3,$E0,$AF,$8B,$7E,$0C
	.dc.b	$5F,$2D,$E6,$9F,$A1,$87,$8E,$31,$1B,$6B,$FE,$D8,$9E,$35,$FE,$D5
	.dc.b	$FA,$C7,$DF,$78,$9B,$1E,$FA,$E1,$79,$A9,$39,$C1,$7C,$2F,$CE,$70
	.dc.b	$9D,$46,$74,$F9,$BE,$57,$E0,$3E,$E7,$E1,$B6,$BC,$0F,$E1,$31,$CD
	.dc.b	$2A,$92,$B7,$1D,$5F,$F3,$1B,$B7,$9F,$DF,$FE,$B5,$BD,$5F,$4B,$89
	.dc.b	$E5,$3E,$8F,$C7,$7B,$D7,$D7,$A2,$31,$97,$A9,$B9,$7F,$03,$C7,$9E
	.dc.b	$1C,$B2,$F6,$CB,$E6,$7E,$9E,$3F,$90,$6C,$64,$FF,$6F,$97,$2F,$F5
	.dc.b	$0F,$D8,$7F,$B9,$63,$E2,$B6,$3C,$B1,$BC,$FE,$6C,$E5,$F9,$FE,$4B
	.dc.b	$E6,$FC,$2D,$AF,$E8,$91,$82,$38,$25,$FF,$4D,$8F,$4F,$C6,$FF,$20
	.dc.b	$9F,$AE,$D9,$F8,$F5,$D5,$F9,$31,$9B,$FC,$0F,$1B,$E9,$BE,$1B,$70
	.dc.b	$FE,$CF,$18,$23,$82,$5F,$FB,$12,$78,$3F,$EC,$6F,$B6,$FF,$14,$4F
	.dc.b	$E2,$9B,$1F,$6A,$77,$7F,$92,$9C,$BF,$94,$F4,$9F,$73,$F5,$9B,$87
	.dc.b	$F8,$3A,$30,$47,$04,$C3,$FC,$23,$1F,$7B,$ED,$1F,$33,$F9,$E6,$D7
	.dc.b	$E7,$AF,$AD,$FE,$3A,$72,$FE,$43,$EB,$3F,$C3,$15,$FC,$8E,$3F,$D7
	.dc.b	$31,$87,$FD,$CA,$D7,$F9,$97,$12,$31,$1F,$EC,$66,$BC,$BD,$6F,$C9
	.dc.b	$C6,$27,$84,$7E,$5B,$D3,$7D,$47,$ED,$9B,$87,$ED,$AF,$C5,$AF,$EF
	.dc.b	$27,$38,$FE,$A3,$DF,$FF,$7B,$DF,$F9,$73,$DD,$FD,$0C,$62,$7F,$DF
	.dc.b	$AD,$7D,$2D,$7F,$DB,$AD,$7F,$DB,$2D,$CC,$DF,$B5,$FE,$51,$8F,$F0
	.dc.b	$E6,$E1,$CB,$0F,$E1,$F1,$8C,$3F,$BD,$7C,$8B,$E3,$3F,$93,$C2,$31
	.dc.b	$7F,$46,$31,$3A,$9C,$C7,$D5,$6B,$FD,$65,$AF,$F6,$56,$BE,$08,$EF
	.dc.b	$F8,$C7,$2C,$7F,$9E,$36,$F8,$4F,$F4,$8A,$4E,$5F,$4F,$66,$70,$FE
	.dc.b	$27,$1E,$CF,$4F,$13,$FD,$3B,$F3,$47,$CA,$7F,$44,$E9,$46,$33,$FC
	.dc.b	$74,$62,$75,$39,$C3,$FB,$A4,$BF,$35,$FF,$30,$5F,$F4,$36,$CF,$F4
	.dc.b	$23,$E5,$7C,$3F,$F5,$E8,$CE,$DF,$D9,$BF,$0E,$FE,$3D,$5F,$D6,$9A
	.dc.b	$FF,$6E,$6A,$C1,$18,$6F,$85,$EB,$8C,$11,$FE,$5F,$F7,$3F,$D8,$2B
	.dc.b	$F9,$DB,$6F,$C5,$3F,$D4,$A3,$19,$7F,$BC,$FE,$1B,$FA,$AF,$C9,$7F
	.dc.b	$2B,$8F,$F7,$B7,$8D,$89,$91,$FD,$62,$75,$8F,$FB,$E6,$B2,$FF,$4A
	.dc.b	$FA,$FF,$F7,$55,$7F,$5E,$6C,$67,$6F,$EC,$71,$8C,$FF,$B3,$B6,$BF
	.dc.b	$CA,$1D,$28,$C5,$A3,$17,$CB,$D4,$FF,$06,$ED,$FC,$CF,$1F,$A1,$E3
	.dc.b	$FF,$84,$BE,$5B,$F1,$7F,$3F,$A8,$9A,$7F,$7D,$BE,$0F,$F9,$7E,$B6
	.dc.b	$E6,$7F,$BC,$F4,$7F,$8A,$A3,$FB,$CB,$74,$FF,$01,$F1,$FD,$8A,$DD
	.dc.b	$7F,$80,$3E,$A7,$F9,$8B,$EC,$6D,$7A,$D7,$DB,$EF,$E3,$E5,$7F,$81
	.dc.b	$A7,$7C,$07,$71,$99,$FF,$06,$FC,$BB,$E5,$E9,$8F,$4E,$D7,$F8,$42
	.dc.b	$74,$D9,$D3,$4F,$F1,$07,$96,$F8,$AF,$D5,$36,$3F,$52,$DD,$3F,$4E
	.dc.b	$D7,$FC,$6E,$D7,$FC,$96,$D7,$F3,$CD,$DB,$FA,$9D,$7F,$8E,$B8,$3F
	.dc.b	$C6,$F9,$FD,$9F,$57,$E3,$D6,$B1,$FF,$1C,$56,$9F,$F1,$5F,$87,$FE
	.dc.b	$81,$6F,$C0,$FF,$8C,$21,$F6,$B6,$24,$E8,$E3,$18,$D3,$FC,$47,$5F
	.dc.b	$D6,$39,$FF,$C7,$6D,$AF,$DC,$79,$7D,$76,$91,$E2,$7F,$27,$18,$BF
	.dc.b	$C2,$C3,$EE,$7E,$59,$2B,$E5,$B6,$7E,$5B,$5F,$96,$FC,$5D,$7F,$63
	.dc.b	$A7,$F9,$73,$BF,$FC,$DB,$87,$5C,$3E,$2D,$7E,$DB,$95,$F7,$EF,$BD
	.dc.b	$FF,$29,$B6,$BF,$CA,$6B,$E7,$F3,$32,$72,$F0,$8C,$69,$FE,$48,$FA
	.dc.b	$77,$DA,$7F,$97,$9B,$7F,$E5,$E7,$F3,$26,$27,$84,$EC,$E1,$CB,$6B
	.dc.b	$FC,$CB,$F5,$7F,$DD,$C9,$E9,$FF,$93,$1F,$EF,$BF,$1C,$FC,$DB,$FC
	.dc.b	$3F,$5F,$ED,$78,$C7,$B5,$FE,$70,$AF,$EF,$33,$9A,$DE,$5F,$C2,$E7
	.dc.b	$F9,$BC,$FF,$10,$6C,$2F,$9C,$F9,$3E,$5B,$A1,$1F,$F4,$54,$6A,$7F
	.dc.b	$CD,$75,$CB,$80,$FF,$A7,$E3,$FC,$32,$D7,$C1,$F0,$FE,$D3,$DD,$F2
	.dc.b	$5E,$93,$D4,$DC,$3E,$2F,$C2,$FC,$32,$3F,$E2,$9C,$3B,$C7,$C7,$FC
	.dc.b	$0B,$E1,$1F,$0D,$B3,$F3,$A2,$E1,$E8,$9B,$D9,$FE,$9A,$C9,$58,$BF
	.dc.b	$8D,$56,$FE,$56,$DF,$54,$93,$E3,$BF,$F2,$CC,$FC,$0E,$7F,$A8,$7D
	.dc.b	$5A,$B5,$FE,$B3,$97,$D2,$FF,$A2,$FE,$05,$1F,$ED,$EE,$EF,$F7,$94
	.dc.b	$E6,$30,$7D,$7F,$B1,$FF,$3A,$7C,$67,$F8,$7E,$38,$CE,$E3,$11,$DB
	.dc.b	$D4,$61,$DA,$73,$5B,$C3,$F1,$98,$74,$7C,$37,$C3,$0F,$09,$CF,$0E
	.dc.b	$53,$A6,$C7,$97,$CB,$FF,$78,$E3,$DA,$F9,$36,$B7,$5C,$CF,$6F,$23
	.dc.b	$1F,$14,$7B,$2F,$8E,$3D,$E3,$C7,$97,$A5,$F0,$67,$D2,$31,$1D,$72
	.dc.b	$F0,$3D,$B0,$EA,$E6,$FA,$4E,$8F,$99,$8F,$8F,$3F,$5C,$E8,$95,$E5
	.dc.b	$CF,$E0,$D6,$F2,$F1,$67,$E8,$AD,$1E,$87,$13,$B8,$C5,$70,$23,$E7
	.dc.b	$C3,$C2,$F2,$C7,$CD,$8F,$8A,$F9,$46,$27,$53,$9C,$BD,$F9,$FC,$43
	.dc.b	$F1,$93,$F9,$03,$CA,$53,$EA,$15,$DB,$F9,$4F,$B1,$7C,$97,$F9,$AE
	.dc.b	$51,$89,$FF,$4F,$46,$27,$53,$99,$E8,$EF,$0F,$C9,$B7,$0E,$D5,$F3
	.dc.b	$3D,$F7,$D8,$F9,$EF,$73,$F7,$EF,$86,$1F,$A2,$6E,$7F,$E8,$CB,$E1
	.dc.b	$8F,$4A,$FC,$4D,$70,$7F,$66,$F0,$C7,$BC,$75,$C2,$3C,$1F,$25,$F4
	.dc.b	$8F,$04,$6B,$1F,$2C,$7E,$72,$B9,$BC,$49,$1A,$3F,$1F,$1A,$71,$87
	.dc.b	$9D,$B3,$FC,$DE,$75,$39,$24,$78,$8F,$05,$F9,$5F,$6D,$F9,$67,$78
	.dc.b	$FC,$59,$C1,$C7,$0E,$F1,$98,$F3,$CF,$67,$99,$D8,$E2,$62,$78,$E1
	.dc.b	$E6,$8F,$EA,$5E,$DF,$E3,$F9,$BF,$98,$37,$87,$FA,$A3,$67,$FA,$C7
	.dc.b	$C7,$DE,$8F,$E9,$AF,$FA,$E5,$B1,$FC,$C6,$31,$CC,$6F,$CE,$71,$83
	.dc.b	$4C,$7D,$85,$AF,$EE,$71,$5F,$EB,$B1,$72,$FE,$8D,$37,$2F,$ED,$53
	.dc.b	$CA,$BF,$A1,$3E,$22,$3F,$EB,$2F,$CD,$DE,$47,$B7,$CE,$46,$27,$FA
	.dc.b	$0E,$2C,$3F,$A1,$D6,$AF,$85,$BD,$BA,$7F,$B6,$E7,$9B,$63,$58,$78
	.dc.b	$63,$E2,$F9,$DF,$0C,$91,$FD,$F3,$C3,$67,$88,$FE,$FB,$37,$3F,$24
	.dc.b	$7F,$81,$EB,$6F,$1A,$FB,$1C,$9F,$E1,$76,$CE,$A7,$FC,$55,$CD,$F6
	.dc.b	$FC,$7F,$D1,$C7,$F1,$1E,$24,$62,$3F,$D4,$AD,$7F,$D4,$8D,$8D,$FC
	.dc.b	$07,$32,$4F,$F8,$CF,$E0,$6B,$FE,$39,$8B,$9F,$AA,$BB,$D7,$E3,$7C
	.dc.b	$CF,$F4,$65,$FF,$91,$BD,$3C,$19,$FA,$1B,$19,$20,$C3,$A5,$27,$F3
	.dc.b	$5E,$83,$F3,$87,$B9,$86,$E1,$FE,$7E,$6B,$FE,$AE,$FC,$D1,$F4,$FF
	.dc.b	$87,$6C,$7F,$BF,$1A,$FF,$BE,$9A,$FF,$3B,$E7,$46,$23,$F1,$6D,$7F
	.dc.b	$DE,$4D,$8D,$CF,$E2,$E3,$5B,$3C,$07,$91,$FB,$36,$E1,$FE,$94,$8F
	.dc.b	$CE,$75,$3E,$03,$51,$27,$47,$E7,$BB,$3F,$3B,$5B,$8C,$1F,$EE,$1F
	.dc.b	$AF,$7D,$E7,$F0,$16,$C6,$B2,$46,$23,$E9,$5F,$FB,$B2,$3E,$2F,$1F
	.dc.b	$DF,$74,$DB,$E8,$75,$D7,$EA,$7C,$54,$C3,$EA,$3F,$44,$8F,$D1,$7E
	.dc.b	$31,$1C,$10,$FA,$8E,$5F,$ED,$98,$FF,$90,$3C,$0F,$E3,$9E,$47,$F1
	.dc.b	$49,$F6,$35,$F5,$89,$FF,$23,$FA,$1C,$2A,$4F,$07,$9C,$7F,$27,$F8
	.dc.b	$1F,$F5,$14,$FF,$BE,$BD,$8C,$19,$FA,$2B,$19,$21,$E0,$C1,$49,$FE
	.dc.b	$65,$F8,$1F,$F2,$DA,$FA,$4D,$78,$37,$6F,$E0,$FC,$DF,$C8,$35,$FC
	.dc.b	$29,$2B,$EF,$C6,$5F,$F6,$FF,$E7,$3F,$A9,$73,$C6,$AE,$11,$F1,$2B
	.dc.b	$83,$F1,$BA,$97,$C6,$3D,$5F,$5C,$FC,$3F,$FA,$53,$63,$F1,$4D,$77
	.dc.b	$F6,$3A,$99,$1A,$78,$C7,$F5,$3F,$8B,$7C,$65,$25,$7E,$56,$1F,$29
	.dc.b	$88,$F9,$91,$C1,$D4,$F3,$9F,$EB,$CB,$DD,$A4,$9F,$87,$B7,$EB,$8F
	.dc.b	$E8,$07,$F4,$97,$FC,$71,$AF,$F4,$E8,$C9,$23,$FC,$78,$23,$4E,$1A
	.dc.b	$F2,$8E,$8F,$F0,$BF,$8E,$7C,$57,$F5,$96,$C7,$DA,$6B,$F9,$81,$F1
	.dc.b	$3E,$5E,$2F,$E9,$3E,$DB,$EC,$DF,$43,$E4,$BF,$F4,$AC,$61,$F3,$CE
	.dc.b	$E7,$4D,$8E,$1E,$07,$82,$FF,$AA,$C7,$F0,$4C,$BF,$C1,$4D,$3F,$D6
	.dc.b	$12,$F8,$AC,$FE,$34,$FA,$FA,$4A,$DF,$AB,$CA,$CF,$9E,$3F,$88,$B6
	.dc.b	$33,$1E,$48,$C1,$1C,$10,$FE,$07,$F2,$AF,$C2,$7F,$C3,$CD,$8F,$E3
	.dc.b	$CD,$7F,$69,$EE,$BE,$2B,$FC,$48,$D8,$FF,$72,$35,$FE,$1E,$D7,$FB
	.dc.b	$73,$5F,$F6,$7F,$84,$8C,$47,$F0,$F6,$C6,$67,$FC,$61,$18,$F9,$77
	.dc.b	$6F,$F8,$D5,$B7,$FC,$D5,$AF,$F0,$BF,$AF,$FE,$63,$E5,$F8,$1F,$E7
	.dc.b	$11,$FE,$01,$6B,$9F,$2F,$FC,$7D,$18,$8F,$E4,$4D,$C8,$D2,$4F,$F9
	.dc.b	$17,$E6,$9F,$23,$FE,$49,$6D,$FF,$49,$6B,$FC,$7F,$BD,$E5,$7F,$93
	.dc.b	$DB,$1F,$A8,$C7,$F7,$73,$FE,$51,$9C,$CF,$F3,$C6,$BF,$E1,$B6,$BF
	.dc.b	$39,$AF,$F3,$B6,$CF,$48,$EC,$BE,$BB,$FC,$C0,$D8,$FE,$C0,$D7,$F9
	.dc.b	$B7,$7F,$E1,$7A,$7F,$C3,$8F,$E1,$E3,$FC,$58,$D7,$EF,$B5,$FE,$98
	.dc.b	$D9,$FE,$09,$C7,$FE,$6D,$8C,$11,$FF,$38,$75,$7F,$1C,$AF,$EE,$4D
	.dc.b	$7F,$A4,$36,$FF,$C5,$DF,$3D,$FB,$46,$F0,$7F,$9E,$63,$F6,$F5,$FE
	.dc.b	$41,$6B,$F9,$46,$BF,$D8,$5B,$89,$A4,$9F,$F4,$27,$CF,$FF,$09,$AF
	.dc.b	$EB,$4D,$BF,$F2,$03,$67,$F4,$3F,$5A,$F8,$FF,$F4,$73,$6F,$F0,$0D
	.dc.b	$7F,$BC,$B5,$FF,$2C,$35,$FD,$5F,$C1,$7F,$4B,$C3,$FC,$9C,$DB,$CF
	.dc.b	$93,$FE,$9D,$FA,$17,$E1,$DD,$6F,$73,$FE,$9F,$BE,$18,$7F,$80,$5B
	.dc.b	$7C,$32,$FF,$52,$2F,$89,$A4,$AF,$F1,$34,$7F,$82,$F0,$FF,$30,$3B
	.dc.b	$8C,$CE,$DF,$F0,$04,$6F,$2F,$F4,$8F,$D2,$BE,$B3,$FD,$30,$DA,$FE
	.dc.b	$F2,$D7,$FB,$A7,$D3,$BE,$FF,$75,$27,$6F,$FA,$86,$75,$1F,$E0,$26
	.dc.b	$BF,$D6,$9A,$FF,$96,$DA,$FF,$A9,$1A,$FF,$5B,$FA,$9F,$ED,$0F,$F8
	.dc.b	$31,$AF,$F8,$1D,$B9,$FF,$85,$26,$E3,$FE,$AE,$9B,$5C,$FC,$EF,$F5
	.dc.b	$B4,$63,$9D,$7A,$A7,$D5,$7F,$5A,$C7,$FA,$D3,$63,$FB,$7B,$51,$0E
	.dc.b	$3E,$5F,$FA,$25,$DF,$D1,$37,$E6,$D8,$CC,$60,$FE,$A4,$BE,$46,$91
	.dc.b	$FE,$AC,$FF,$7D,$C7,$FD,$C9,$AD,$F9,$1F,$D7,$E9,$EC,$6C,$EA,$FE
	.dc.b	$94,$60,$95,$C4,$FF,$05,$FC,$93,$F3,$A7,$DC,$7E,$EA,$30,$D8,$CC
	.dc.b	$F0,$6D,$FD,$FF,$9A,$F0,$92,$7D,$0D,$8D,$46,$08,$FF,$9B,$38,$1F
	.dc.b	$3F,$FC,$2B,$BF,$FD,$68,$D8,$CD,$7F,$3B,$6D,$71,$9D,$12,$74,$FF
	.dc.b	$14,$DD,$F5,$9F,$EF,$19,$FF,$15,$C6,$3E,$57,$53,$9B,$FB,$C7,$DB
	.dc.b	$70,$FE,$06,$78,$57,$29,$CC,$FE,$1A,$72,$F2,$6B,$F8,$9E,$C7,$D8
	.dc.b	$BE,$97,$F9,$4C,$61,$B1,$9C,$3F,$C8,$33,$97,$FA,$E0,$F9,$DF,$90
	.dc.b	$D3,$2E,$0E,$BF,$83,$3E,$C2,$92,$7A,$C6,$5B,$19,$C7,$E1,$DA,$B5
	.dc.b	$18,$6B,$FD,$0F,$E7,$FE,$D7,$8E,$69,$30,$F5,$FE,$0D,$F8,$AA,$F8
	.dc.b	$8D,$7F,$5F,$82,$31,$08,$C7,$2F,$C8,$9C,$C6,$2B,$AC,$E4,$FC,$67
	.dc.b	$4F,$FB,$8A,$3F,$C0,$8D,$BF,$F5,$EB,$5F,$EE,$3F,$1D,$F5,$38,$7F
	.dc.b	$92,$57,$D2,$6C,$FC,$46,$BF,$75,$AF,$A9,$AF,$A5,$B8,$7E,$53,$C2
	.dc.b	$EC,$7E,$87,$CD,$7D,$0D,$26,$1F,$E1,$B8,$CB,$F8,$4F,$C1,$9F,$85
	.dc.b	$FE,$29,$B1,$98,$C1,$1C,$13,$2F,$BB,$3C,$1F,$F6,$FF,$D7,$BE,$9B
	.dc.b	$F6,$8D,$AF,$F1,$63,$67,$73,$F9,$AF,$99,$7C,$8D,$25,$FE,$6A,$32
	.dc.b	$D8,$CD,$7E,$3A,$72,$D8,$FE,$09,$EE,$7F,$B2,$EB,$FC,$CF,$F1,$D8
	.dc.b	$23,$FE,$93,$F9,$1F,$F5,$EB,$7E,$2F,$FC,$27,$18,$C3,$FD,$56,$D7
	.dc.b	$F7,$ED,$7F,$9F,$35,$FE,$82,$D7,$FD,$DB,$1D,$3E,$5A,$57,$85,$B3
	.dc.b	$DB,$E7,$3C,$24,$9F,$33,$E4,$F9,$D9,$1E,$96,$CF,$9B,$E6,$3D,$24
	.dc.b	$9F,$73,$EC,$F9,$89,$1F,$05,$B3,$EE,$6B,$EE,$6B,$FD,$73,$A5,$F5
	.dc.b	$9F,$C2,$63,$FD,$A6,$D7,$FB,$04,$6B,$C1,$FE,$A3,$39,$9F,$F4,$9F
	.dc.b	$C9,$7F,$22,$5E,$4A,$30,$DF,$1F,$F9,$F4,$60,$8E,$09,$97,$F9,$E7
	.dc.b	$EB,$5A,$7F,$45,$6C,$66,$30,$47,$04,$C3,$F9,$DE,$0A,$E3,$1F,$CF
	.dc.b	$BE,$59,$F5,$EF,$9E,$FF,$2D,$C6,$1B,$19,$C3,$FD,$E7,$39,$6C,$66
	.dc.b	$BA,$35,$FF,$04,$4E,$DC,$D7,$CA,$F1,$65,$FF,$BE,$F5,$FE,$50,$D8
	.dc.b	$FE,$C9,$F2,$5F,$D0,$EF,$D6,$D7,$FC,$29,$9F,$FB,$FB,$CB,$D9,$2F
	.dc.b	$FD,$45,$5B,$AF,$F6,$4B,$5F,$AE,$DA,$FF,$80,$3F,$5D,$FD,$CF,$E5
	.dc.b	$4D,$25,$FF,$AB,$63,$2D,$8C,$DF,$D0,$9C,$BF,$B0,$F8,$AF,$D9,$92
	.dc.b	$BF,$70,$D8,$D4,$60,$8F,$F7,$4E,$4F,$E4,$BF,$5D,$FE,$22,$F1,$4D
	.dc.b	$25,$7F,$B0,$7A,$BF,$E0,$0B,$7E,$47,$FC,$1D,$18,$C3,$F0,$AD,$7E
	.dc.b	$5B,$5F,$F7,$7B,$5F,$CE,$B7,$43,$9C,$7F,$57,$F1,$9F,$A6,$26,$1F
	.dc.b	$75,$B7,$FA,$E6,$BF,$AE,$E1,$FF,$70,$CF,$E3,$BE,$95,$F3,$BF,$94
	.dc.b	$6C,$66,$30,$47,$04,$AF,$CF,$4F,$07,$E5,$6A,$FA,$EF,$94,$4A,$FF
	.dc.b	$10,$CB,$2F,$EA,$73,$FE,$35,$FA,$37,$D7,$52,$5F,$FB,$BE,$32,$D8
	.dc.b	$CD,$FE,$CA,$72,$D8,$FF,$20,$7D,$33,$EB,$A9,$2F,$FD,$F5,$19,$6C
	.dc.b	$66,$BF,$7B,$39,$6C,$7F,$8B,$33,$7C,$E5,$25,$FF,$8B,$23,$2F,$C0
	.dc.b	$E1,$7D,$5F,$F5,$DC,$3F,$5A,$D7,$FD,$07,$1A,$F3,$3F,$CB,$13,$99
	.dc.b	$FF,$0B,$5F,$F8,$2D,$FF,$0D,$B5,$CF,$A9,$E5,$26,$1E,$36,$C7,$F8
	.dc.b	$C2,$78,$7A,$67,$FD,$C1,$5E,$9F,$8C,$F7,$11,$F5,$60,$F9,$4F,$29
	.dc.b	$23,$2F,$FA,$2B,$F0,$2F,$E4,$78,$FF,$3E,$6B,$FD,$15,$AF,$E0,$DB
	.dc.b	$A7,$C6,$F8,$1F,$82,$4B,$F0,$B6,$BF,$98,$E0,$8C,$47,$CC,$6B,$F3
	.dc.b	$23,$53,$C8,$91,$FC,$06,$FF,$C8,$7C,$86,$93,$D9,$FA,$7B,$4B,$F8
	.dc.b	$6D,$BF,$CE,$FD,$37,$F9,$A7,$E5,$3F,$D7,$BF,$11,$FB,$A9,$FF,$5F
	.dc.b	$CE,$5F,$F5,$B7,$6F,$F9,$53,$EB,$BF,$9A,$C6,$0D,$F9,$CC,$39,$D7
	.dc.b	$FB,$1E,$F9,$61,$D6,$D1,$89,$E9,$E3,$B3,$E3,$5D,$30,$AF,$0C,$7B
	.dc.b	$3E,$13,$11,$C3,$95,$9F,$9B,$3F,$06,$7D,$EF,$A1,$C5,$C5,$70,$CB
	.dc.b	$C5,$39,$C7,$BD,$F2,$6E,$7E,$D6,$CF,$F9,$D7,$1F,$04,$E7,$2F,$4E
	.dc.b	$5E,$1A,$F3,$9F,$27,$2F,$06,$7A,$C7,$9C,$EB,$2F,$06,$1E,$DC,$BA
	.dc.b	$CF,$AA,$72,$F9,$E7,$59,$FA,$A7,$24,$CF,$C9,$5E,$1B,$E6,$F7,$8C
	.dc.b	$CE,$E3,$D9,$5C,$E7,$24,$8F,$13,$F6,$B1,$F5,$61,$27,$C8,$77,$AE
	.dc.b	$4F,$03,$E9,$9F,$ED,$0C,$FF,$20,$DC,$BE,$FB,$6B,$F9,$AB,$F8,$AF
	.dc.b	$AE,$7C,$46,$2D,$18,$9F,$CF,$B5,$EC,$D7,$F2,$CE,$F3,$FC,$3B,$73
	.dc.b	$CD,$FE,$37,$D7,$7D,$63,$0F,$BB,$7B,$7F,$52,$D7,$CD,$5C,$67,$84
	.dc.b	$7E,$46,$75,$8F,$A2,$B9,$9B,$9A,$6D,$B8,$FE,$DF,$76,$3E,$4A,$CE
	.dc.b	$5E,$87,$81,$F8,$F8,$D3,$8A,$F7,$DF,$08,$FF,$67,$4E,$A7,$31,$BA
	.dc.b	$E9,$1E,$19,$D4,$60,$93,$F0,$AB,$D4,$F1,$30,$5B,$F4,$E5,$F1,$30
	.dc.b	$EE,$4B,$E7,$3A,$AF,$31,$F3,$0E,$58,$78,$9E,$B5,$C5,$B5,$FE,$EC
	.dc.b	$9D,$73,$7F,$5E,$F8,$AF,$C5,$B6,$BF,$A4,$C6,$39,$8F,$E4,$39,$BF
	.dc.b	$98,$61,$79,$BF,$8E,$5F,$F4,$98,$C1,$C7,$EB,$64,$7F,$54,$6B,$FD
	.dc.b	$11,$AF,$F6,$E6,$BF,$D0,$DA,$FF,$A3,$30,$46,$2D,$F4,$4F,$B1,$F3
	.dc.b	$36,$FF,$C3,$73,$A9,$C9,$EB,$FA,$F7,$AD,$49,$5C,$23,$2F,$F8,$93
	.dc.b	$81,$F9,$DF,$F2,$68,$FC,$D8,$8C,$F4,$FE,$D9,$B1,$C6,$78,$3F,$A2
	.dc.b	$8F,$EF,$99,$EF,$C0,$F6,$CF,$F8,$56,$30,$47,$04,$3F,$C0,$CF,$DA
	.dc.b	$F1,$3F,$3B,$1F,$E8,$46,$BF,$E1,$86,$BF,$EB,$66,$D7,$F8,$37,$C2
	.dc.b	$F1,$12,$7F,$C4,$CF,$5F,$6E,$47,$F8,$D1,$AF,$F8,$C9,$AF,$85,$AF
	.dc.b	$F8,$A3,$C7,$7C,$07,$E4,$1B,$1F,$E4,$98,$C1,$1C,$10,$FF,$25,$3F
	.dc.b	$93,$FA,$E7,$B9,$49,$3F,$80,$8C,$BF,$EB,$8F,$CF,$3F,$D3,$6D,$F5
	.dc.b	$7F,$95,$FB,$29,$CC,$FF,$B1,$5A,$FF,$AE,$9A,$FF,$21,$8E,$13,$CA
	.dc.b	$73,$29,$CB,$F9,$2F,$AB,$FF,$6D,$B7,$DD,$FD,$64,$62,$3F,$D3,$CD
	.dc.b	$7F,$DE,$2D,$7F,$DF,$0D,$7F,$CE,$CD,$E7,$39,$ED,$7D,$47,$C2,$AF
	.dc.b	$F7,$1B,$5F,$F6,$4B,$5F,$F2,$EB,$5F,$F7,$DB,$63,$9E,$9F,$C2,$23
	.dc.b	$06,$ED,$A3,$E0,$FE,$B1,$30,$FA,$8F,$F5,$A5,$ED,$7F,$A0,$5F,$EB
	.dc.b	$AD,$7F,$D2,$6D,$7F,$5E,$D7,$FA,$5B,$5F,$EA,$8D,$7A,$FA,$3D,$E3
	.dc.b	$59,$7F,$8C,$BC,$FF,$E4,$3E,$3F,$F9,$9E,$FF,$D0,$4D,$7F,$7B,$3B
	.dc.b	$73,$3C,$08,$E0,$FF,$45,$7C,$1F,$FA,$E1,$BE,$B7,$F2,$E8,$C4,$74
	.dc.b	$6B,$FE,$FA,$6B,$FE,$8E,$6A,$A4,$61,$B3,$F8,$AC,$A5,$7E,$3D,$AF
	.dc.b	$F1,$6E,$17,$D6,$FE,$C2,$30,$D7,$FB,$AC,$6B,$97,$F4,$B3,$99,$FD
	.dc.b	$03,$5F,$CF,$36,$78,$47,$FB,$29,$AF,$CA,$FA,$77,$D6,$FF,$50,$6C
	.dc.b	$7F,$AC,$5A,$EE,$34,$47,$4D,$7E,$AF,$C2,$3E,$2F,$FA,$DB,$63,$FD
	.dc.b	$78,$D7,$71,$A2,$3A,$6B,$FB,$2E,$6F,$DA,$91,$FD,$D3,$F8,$BF,$82
	.dc.b	$91,$F9,$86,$BF,$98,$6B,$F9,$86,$BF,$98,$8F,$CA,$11,$FD,$0B,$FA
	.dc.b	$5E,$67,$C1,$52,$4F,$F9,$0A,$32,$DF,$6B,$81,$1C,$1B,$F8,$0E,$84
	.dc.b	$70,$73,$F8,$BF,$01,$1C,$10,$FE,$BE,$7F,$0C,$74,$E2,$91,$FE,$0C
	.dc.b	$F4,$1F,$53,$49,$5F,$EE,$F8,$CB,$FD,$B7,$89,$F3,$DF,$DA,$5B,$19
	.dc.b	$8C,$11,$C1,$0F,$EE,$7E,$2F,$F4,$E8,$FF,$0E,$FD,$43,$E1,$A9,$2B
	.dc.b	$FE,$00,$46,$5F,$EF,$DF,$75,$FC,$65,$BD,$1F,$DE,$E3,$11,$D9,$AF
	.dc.b	$F7,$B6,$BF,$E6,$5C,$11,$88,$FF,$03,$36,$33,$3F,$E3,$2E,$3F,$E2
	.dc.b	$15,$FE,$66,$EA,$FC,$E7,$66,$49,$1F,$E1,$3F,$91,$FE,$55,$81,$A4
	.dc.b	$8F,$F0,$93,$5E,$06,$0F,$F2,$0F,$84,$FA,$7A,$48,$FB,$F8,$27,$51
	.dc.b	$FE,$23,$FB,$A7,$C8,$7F,$87,$DB,$19,$8C,$11,$C1,$0F,$AF,$1C,$1F
	.dc.b	$E5,$FF,$2B,$FE,$26,$7F,$D1,$D6,$8C,$47,$F2,$36,$C6,$7D,$2F,$F1
	.dc.b	$87,$C6,$7F,$0A,$AF,$F1,$73,$5F,$A8,$D7,$FC,$70,$DD,$0D,$27,$7F
	.dc.b	$43,$FC,$CF,$F4,$EF,$88,$A4,$BF,$8F,$19,$7F,$C9,$1F,$32,$F9,$3F
	.dc.b	$F2,$23,$63,$31,$83,$FD,$7B,$F3,$4F,$A7,$E5,$3F,$E4,$E8,$C6,$1A
	.dc.b	$9C,$CF,$FA,$B2,$72,$FE,$D9,$AF,$18,$DB,$FE,$B9,$8C,$BC,$63,$2F
	.dc.b	$F4,$F8,$7D,$8F,$FA,$19,$B1,$F8,$46,$BF,$D9,$9A,$FE,$65,$AF,$DA
	.dc.b	$6B,$D6,$3A,$1F,$E8,$DF,$A0,$7C,$D5,$24,$FF,$9E,$A3,$2D,$8C,$BF
	.dc.b	$E7,$2E,$44,$EA,$30,$D7,$FD,$A3,$F0,$EF,$93,$FB,$F1,$FC,$72,$31
	.dc.b	$C1,$FE,$EE,$6C,$66,$76,$F4,$FA,$5F,$41,$1F,$33,$5D,$18,$21,$FC
	.dc.b	$AB,$E9,$1F,$48,$FA,$2F,$F7,$EC,$61,$B1,$99,$E0,$D7,$FD,$23,$2F
	.dc.b	$A4,$6F,$C9,$B3,$DD,$AF,$06,$BF,$CF,$3E,$33,$FC,$68,$9A,$47,$E7
	.dc.b	$7D,$87,$FA,$C3,$A5,$1E,$96,$F0,$27,$53,$91,$B6,$7C,$E5,$F2,$D6
	.dc.b	$D7,$F0,$FB,$E3,$18,$C3,$A7,$C4,$EA,$31,$3A,$BF,$1E,$3D,$5E,$87
	.dc.b	$92,$35,$87,$8B,$E5,$DF,$69,$FC,$3E,$D1,$8B,$FC,$23,$6B,$FD,$7F
	.dc.b	$3A,$AF,$E2,$D9,$7B,$F9,$7C,$98,$77,$6D,$F7,$AB,$5B,$AF,$2C,$FB
	.dc.b	$EB,$38,$FE,$8A,$73,$87,$77,$77,$C6,$FE,$67,$77,$E2,$7E,$6F,$BB
	.dc.b	$6B,$AF,$7F,$FB,$63,$EB,$33,$B2,$E9,$2B,$7F,$14,$AC,$63,$DB,$ED
	.dc.b	$DF,$35,$D9,$B7,$E4,$FA,$97,$CA,$66,$A4,$FB,$B5,$5F,$0C,$3F,$09
	.dc.b	$B3,$1E,$37,$BA,$D6,$BF,$84,$D9,$E1,$FF,$7B,$AF,$E5,$BE,$87,$F4
	.dc.b	$79,$CB,$67,$59,$7F,$9E,$BE,$B7,$FD,$FF,$7E,$36,$C7,$E5,$6F,$8C
	.dc.b	$E7,$E4,$7E,$CC,$3E,$9F,$F5,$D3,$FE,$22,$8C,$76,$78,$67,$37,$FE
	.dc.b	$12,$6B,$FE,$FF,$6B,$FE,$05,$6A,$9E,$75,$D6,$73,$BB,$E6,$7A,$EB
	.dc.b	$FA,$68,$C6,$3E,$6A,$CD,$FF,$68,$BB,$8F,$5F,$A7,$EF,$91,$A4,$C3
	.dc.b	$E5,$36,$FD,$CD,$AF,$F5,$1F,$DB,$BE,$03,$F1,$ED,$AF,$C7,$D7,$83
	.dc.b	$34,$E6,$FE,$77,$E7,$BF,$C2,$2F,$E5,$36,$3F,$C7,$2D,$7F,$86,$9F
	.dc.b	$15,$C3,$FE,$2D,$6B,$93,$F0,$4E,$32,$3A,$B7,$B1,$FC,$4A,$FF,$28
	.dc.b	$DB,$F2,$9F,$AA,$71,$B7,$E4,$7E,$A7,$F6,$93,$FE,$4E,$6B,$E6,$C5
	.dc.b	$18,$8F,$F2,$13,$5F,$BC,$D8,$DF,$77,$E6,$63,$06,$F1,$7D,$8F,$FA
	.dc.b	$A3,$2F,$D5,$37,$0F,$CC,$57,$93,$1F,$C2,$4E,$5F,$D4,$7B,$9F,$BB
	.dc.b	$BF,$F3,$77,$67,$E9,$E3,$18,$7F,$9A,$1A,$FA,$DA,$FF,$96,$DA,$FE
	.dc.b	$5D,$B9,$9C,$E4,$FB,$2F,$F0,$4E,$1F,$C0,$5B,$7F,$AF,$AF,$0E,$3F
	.dc.b	$9B,$9C,$BF,$BF,$FC,$29,$F0,$FF,$BD,$6E,$1F,$B1,$6E,$1F,$E8,$06
	.dc.b	$BF,$4D,$AF,$F9,$CD,$AF,$EB,$DB,$91,$CE,$4F,$96,$FF,$66,$5F,$F1
	.dc.b	$96,$DF,$F0,$7A,$F0,$E3,$FB,$39,$CB,$FC,$5F,$F0,$A7,$C3,$FF,$34
	.dc.b	$9F,$E1,$F3,$98,$C7,$57,$F0,$EA,$F0,$63,$FC,$35,$B8,$FF,$A7,$1A
	.dc.b	$FE,$3D,$AF,$F9,$9B,$F0,$F7,$D6,$FF,$42,$6D,$7F,$31,$7F,$95,$BB
	.dc.b	$8F,$EA,$B2,$FA,$C7,$D0,$FF,$9B,$9B,$5F,$EC,$09,$D4,$E4,$91,$99
	.dc.b	$DC,$FF,$B1,$9A,$FE,$E8,$E3,$1F,$EC,$AF,$11,$F9,$F3,$E9,$FF,$B2
	.dc.b	$B6,$7F,$B2,$D7,$03,$FA,$3C,$65,$FE,$C3,$F9,$E7,$F0,$9D,$FF,$92
	.dc.b	$36,$BF,$D1,$2F,$F2,$97,$FA,$93,$99,$71,$23,$0D,$8F,$ED,$71,$C5
	.dc.b	$EE,$7F,$6E,$FA,$F7,$D7,$D3,$1F,$1B,$70,$FE,$C1,$3A,$6C,$E9,$A7
	.dc.b	$F7,$CC,$BF,$D2,$2D,$F6,$FF,$C0,$FF,$6D,$F8,$A8,$C1,$31,$F8,$73
	.dc.b	$AE,$5F,$92,$C9,$FE,$E1,$F2,$FF,$C2,$9E,$1B,$7F,$49,$8C,$5F,$F8
	.dc.b	$2B,$D7,$7C,$E7,$F8,$29,$FE,$D3,$5F,$D3,$DC,$DF,$F7,$B7,$2F,$FA
	.dc.b	$D9,$CD,$73,$F3,$79,$5F,$37,$FC,$2B,$F3,$4F,$B3,$BE,$AE,$1B,$8F
	.dc.b	$F8,$1E,$74,$D9,$D3,$4F,$C2,$E6,$F9,$EF,$F6,$2B,$63,$F3,$0D,$CB
	.dc.b	$F2,$41,$C3,$F9,$56,$BF,$D4,$1A,$FF,$97,$9A,$F1,$9F,$F2,$0F,$B0
	.dc.b	$F8,$0F,$F1,$C9,$FA,$CA,$F0,$6D,$FE,$56,$8F,$F1,$AF,$5B,$E5,$29
	.dc.b	$8F,$76,$E7,$FE,$27,$9D,$36,$74,$D3,$FC,$87,$8B,$E9,$3F,$53,$E4
	.dc.b	$7F,$92,$9B,$3A,$6E,$BF,$E3,$0A,$E1,$3F,$91,$AF,$F2,$BF,$CD,$7D
	.dc.b	$5D,$BF,$DE,$4D,$9F,$92,$DB,$FD,$E3,$5F,$EE,$ED,$7F,$D0,$AD,$4D
	.dc.b	$31,$DF,$FE,$87,$FA,$EF,$F3,$09,$E2,$E3,$F9,$37,$FE,$79,$8F,$F2
	.dc.b	$EF,$D1,$FF,$48,$8F,$F4,$7F,$A9,$FC,$C1,$FF,$49,$F8,$29,$D6,$4C
	.dc.b	$7F,$C3,$1F,$29,$CB,$CE,$FF,$39,$E3,$C6,$BF,$81,$DF,$0B,$DE,$3F
	.dc.b	$E5,$9F,$A5,$FF,$0D,$73,$7F,$A4,$FE,$BF,$AE,$89,$CD,$7F,$A8,$A7
	.dc.b	$31,$8C,$7C,$51,$8C,$BF,$D0,$6F,$9B,$CE,$7D,$3F,$E0,$2B,$F9,$44
	.dc.b	$63,$9B,$FC,$BD,$39,$9F,$F0,$83,$5F,$83,$F5,$5F,$C1,$DF,$F4,$2B
	.dc.b	$74,$FF,$35,$4F,$57,$FC,$95,$19,$CF,$8F,$E0,$7F,$E8,$A8,$37,$C3
	.dc.b	$C3,$27,$FC,$F9,$3A,$6C,$E7,$0F,$F3,$66,$0E,$3F,$49,$2B,$FC,$D5
	.dc.b	$19,$7F,$6B,$BB,$0F,$D6,$47,$6E,$CF,$F3,$9D,$6F,$0F,$6C,$7F,$AB
	.dc.b	$3B,$1F,$39,$FE,$B4,$6D,$FF,$4F,$6D,$7F,$53,$E3,$46,$23,$FC,$9F
	.dc.b	$97,$87,$E6,$3E,$11,$D7,$0B,$87,$C0,$8B,$5F,$E9,$AB,$EB,$BF,$CC
	.dc.b	$9C,$B6,$BF,$D7,$33,$A7,$EB,$CF,$A7,$D4,$F4,$12,$3D,$0F,$FB,$92
	.dc.b	$FC,$BA,$FF,$8F,$E7,$EF,$FB,$1C,$5F,$A6,$FF,$7C,$FA,$47,$D7,$FF
	.dc.b	$A2,$6F,$FD,$F9,$E1,$7F,$9B,$27,$3F,$01,$FE,$FA,$F8,$CF,$F0,$0D
	.dc.b	$7F,$A4,$FE,$C3,$FB,$74,$62,$15,$C1,$E6,$0C,$E1,$DE,$78,$4E,$DE
	.dc.b	$BD,$2B,$EB,$8F,$8E,$F8,$37,$1F,$23,$71,$EF,$7C,$7D,$36,$9A,$9F
	.dc.b	$16,$7D,$EF,$86,$7E,$2C,$BA,$D7,$B7,$0E,$2F,$A8,$F4,$E5,$CD,$A6
	.dc.b	$70,$E8,$F1,$C7,$51,$88,$DE,$1D,$EB,$C9,$97,$9E,$F7,$8F,$8A,$F7
	.dc.b	$97,$98,$F9,$55,$B8,$F0,$4E,$F1,$EC,$78,$A7,$58,$F7,$AD,$E3,$D8
	.dc.b	$F0,$E7,$F3,$A3,$E2,$E1,$CC,$93,$AC,$FE,$1D,$F3,$26,$5E,$5B,$F8
	.dc.b	$07,$62,$46,$9C,$B5,$F9,$27,$F9,$07,$2F,$C4,$37,$0F,$CA,$37,$0F
	.dc.b	$F2,$74,$BE,$C1,$F1,$98,$6D,$7F,$7A,$9D,$4E,$49,$1F,$5E,$77,$87
	.dc.b	$E4,$9B,$8F,$CE,$FA,$17,$D8,$7F,$B3,$AB,$58,$7D,$CB,$DE,$1F,$8C
	.dc.b	$6E,$CC,$38,$CE,$EA,$3E,$17,$F0,$31,$A8,$C1,$2F,$BB,$F9,$C8,$D4
	.dc.b	$60,$98,$FE,$DF,$0F,$0C,$EA,$31,$7E,$0A,$DD,$B0,$F3,$63,$ED,$BD
	.dc.b	$CF,$85,$B5,$FB,$4C,$3A,$4F,$BF,$0E,$95,$A3,$E4,$5F,$39,$D1,$CC
	.dc.b	$B7,$CE,$35,$1F,$C2,$31,$F9,$6D,$BE,$AF,$9F,$0F,$0B,$93,$A1,$6F
	.dc.b	$A3,$C0,$C6,$1E,$57,$96,$1E,$5E,$7F,$E9,$5C,$BF,$C5,$67,$FB,$1B
	.dc.b	$79,$BF,$B0,$B7,$8B,$F8,$D7,$47,$F4,$4F,$80,$C5,$FF,$82,$6B,$FD
	.dc.b	$E6,$FF,$4B,$8F,$F5,$24,$EE,$30,$7F,$5A,$E6,$FF,$1F,$3F,$D9,$BB
	.dc.b	$F9,$E5,$FD,$75,$EB,$4F,$C2,$5F,$71,$AC,$38,$E1,$FD,$EA,$F0,$70
	.dc.b	$AD,$C7,$52,$1B,$3F,$9A,$69,$FC,$7B,$0E,$D8,$71,$3D,$35,$CA,$BF
	.dc.b	$B2,$35,$F9,$13,$C1,$F6,$7D,$37,$94,$8F,$6D,$7F,$8C,$F7,$FF,$94
	.dc.b	$E3,$07,$E4,$DC,$3F,$E1,$C6,$BF,$E1,$AF,$3B,$FD,$9D,$CD,$FE,$22
	.dc.b	$7E,$6F,$D8,$3E,$97,$E3,$B6,$BA,$E4,$8C,$43,$5F,$81,$1F,$B1,$6D
	.dc.b	$72,$24,$F2,$9D,$65,$F8,$26,$C7,$F9,$25,$B7,$DF,$8F,$F0,$B9,$7F
	.dc.b	$8E,$2A,$E7,$FE,$E7,$7E,$01,$EC,$71,$1B,$8F,$F0,$FF,$91,$FD,$B7
	.dc.b	$3F,$AD,$F9,$FB,$D5,$FC,$2B,$6B,$FC,$A5,$CF,$F8,$AB,$E1,$AF,$E6
	.dc.b	$2B,$55,$F0,$63,$17,$BB,$FC,$8F,$96,$FB,$4F,$C8,$F9,$0F,$97,$A6
	.dc.b	$A2,$63,$F3,$A7,$4D,$9C,$BD,$CF,$F3,$77,$E3,$4F,$86,$FC,$EB,$63
	.dc.b	$EB,$B5,$E1,$E9,$BE,$FD,$F5,$9F,$B0,$8C,$3F,$EB,$89,$FF,$44,$CE
	.dc.b	$9F,$F4,$9B,$5F,$F4,$9B,$5F,$F4,$9B,$5F,$F4,$9B,$5E,$71,$D6,$7B
	.dc.b	$C7,$FB,$1B,$95,$EC,$7F,$56,$E3,$FD,$57,$BF,$FA,$76,$F6,$FE,$EA
	.dc.b	$75,$C7,$FC,$29,$B8,$7F,$A4,$6F,$84,$FF,$0E,$6C,$7F,$A5,$67,$7E
	.dc.b	$03,$EB,$1E,$AF,$F0,$86,$D7,$EB,$27,$53,$9F,$B0,$D3,$27,$55,$C1
	.dc.b	$FA,$D5,$D6,$BD,$1F,$01,$27,$FD,$96,$FF,$20,$5E,$FF,$CD,$D7,$FC
	.dc.b	$A4,$FC,$98,$CF,$07,$E7,$E3,$04,$70,$48,$FB,$51,$F4,$70,$FF,$25
	.dc.b	$F2,$FF,$24,$F3,$1E,$C5,$31,$12,$FE,$04,$E9,$B3,$97,$DB,$EF,$3E
	.dc.b	$9F,$F9,$7A,$E1,$A6,$22,$4F,$CB,$9D,$36,$72,$FE,$77,$93,$97,$CE
	.dc.b	$FF,$3A,$FB,$FF,$E3,$7B,$FF,$92,$B5,$46,$39,$7E,$F5,$7F,$3E,$C4
	.dc.b	$D2,$63,$FC,$6B,$F3,$DF,$AD,$8F,$E3,$A7,$33,$E1,$AE,$2E,$F9,$7F
	.dc.b	$A5,$FB,$0F,$83,$A4,$9F,$C3,$43,$E0,$71,$1F,$82,$8E,$0E,$A7,$9C
	.dc.b	$FF,$56,$FB,$0F,$F1,$7C,$FE,$21,$B9,$7B,$DB,$A7,$FB,$C3,$F2,$F7
	.dc.b	$DC,$71,$F5,$E6,$1F,$97,$BD,$AE,$15,$F0,$8F,$E7,$7F,$13,$FE,$9F
	.dc.b	$CB,$FA,$C3,$72,$FE,$EA,$D9,$FF,$48,$7E,$01,$FE,$BA,$7F,$A7,$87
	.dc.b	$0F,$F8,$04,$DF,$AF,$FC,$D6,$33,$19,$6B,$FE,$07,$6B,$C0,$FE,$24
	.dc.b	$E1,$ED,$1F,$C5,$71,$FF,$4E,$70,$FF,$7E,$F7,$BF,$17,$5F,$B0,$6E
	.dc.b	$5F,$87,$7F,$2D,$1F,$D9,$BC,$EF,$F7,$7F,$81,$FD,$8B,$E1,$0D,$24
	.dc.b	$FA,$CF,$F1,$F4,$3E,$2F,$FC,$1B,$F2,$7F,$B7,$DF,$FD,$69,$A2,$31
	.dc.b	$7C,$2B,$FC,$27,$F8,$9B,$EC,$69,$27,$C9,$F1,$7F,$EF,$96,$F8,$7F
	.dc.b	$DD,$A3,$11,$F3,$75,$46,$23,$FC,$61,$18,$9E,$04,$8F,$54,$65,$B3
	.dc.b	$C5,$F4,$FC,$BB,$E1,$BF,$C5,$2D,$8F,$E4,$4D,$7F,$A9,$53,$E3,$5F
	.dc.b	$0F,$FD,$D9,$B3,$E0,$6B,$E9,$6B,$E8,$6B,$FE,$B4,$6B,$FE,$6D,$6C
	.dc.b	$F7,$9F,$F0,$CC,$6B,$67,$C1,$7F,$92,$2F,$F9,$8B,$79,$3F,$84,$37
	.dc.b	$87,$FD,$83,$F6,$8F,$8F,$FF,$06,$B6,$FE,$3B,$5F,$A0,$D7,$E7,$B5
	.dc.b	$E7,$1F,$E2,$C8,$C4,$7F,$A1,$5B,$99,$CE,$2F,$C0,$7F,$84,$57,$F4
	.dc.b	$F6,$BF,$D2,$9B,$97,$FB,$AA,$FE,$B7,$B5,$FE,$24,$DF,$FC,$27,$7F
	.dc.b	$82,$6B,$FE,$DE,$6B,$FE,$97,$6C,$F2,$8F,$C2,$C6,$09,$3F,$AF,$C3
	.dc.b	$FC,$27,$3F,$D7,$1B,$7F,$CC,$DB,$9F,$FB,$B3,$EC,$DF,$6F,$FE,$38
	.dc.b	$6D,$FE,$49,$AF,$E6,$DA,$FE,$69,$AF,$38,$FF,$2B,$C6,$23,$FD,$60
	.dc.b	$DC,$8E,$6D,$F1,$7F,$C8,$9B,$E1,$7F,$A1,$23,$F9,$2D,$3E,$4F,$E6
	.dc.b	$E4,$C7,$FD,$79,$1F,$49,$FD,$6B,$5D,$4F,$1F,$17,$FC,$D7,$39,$9E
	.dc.b	$31,$82,$3F,$B2,$8D,$62,$FA,$37,$DF,$FF,$A6,$1B,$5F,$E1,$16,$BF
	.dc.b	$B0,$FA,$5F,$49,$FD,$21,$FE,$02,$D7,$CA,$D7,$FD,$AC,$DA,$FF,$6C
	.dc.b	$7A,$3F,$EA,$6C,$BF,$92,$D7,$F8,$61,$BC,$1F,$E9,$BF,$37,$FA,$44
	.dc.b	$E6,$B8,$11,$FF,$42,$C6,$A7,$FB,$B7,$D5,$BE,$DF,$FD,$66,$DA,$FF
	.dc.b	$16,$35,$FE,$17,$F5,$9F,$C2,$DC,$CF,$02,$3F,$CE,$63,$4F,$F8,$1F
	.dc.b	$DB,$7C,$67,$F9,$14,$91,$FE,$A4,$E4,$7E,$71,$FE,$78,$CB,$FD,$6D
	.dc.b	$7C,$2B,$FD,$3A,$D7,$A5,$FF,$B5,$2F,$8F,$45,$C3,$F5,$71,$8A,$FE
	.dc.b	$8C,$D7,$FB,$13,$5F,$F5,$ED,$FF,$B3,$B0,$F0,$FC,$7F,$F9,$6C,$ED
	.dc.b	$F0,$5F,$EA,$C8,$CD,$79,$27,$4F,$AE,$75,$CF,$FE,$EF,$FA,$E7,$D7
	.dc.b	$FF,$AC,$EB,$FA,$C4,$63,$0F,$F3,$E3,$63,$33,$B7,$FB,$C4,$3E,$A9
	.dc.b	$C3,$C1,$B3,$D1,$AF,$06,$BF,$D9,$FE,$53,$FD,$68,$9A,$47,$FC,$0D
	.dc.b	$EE,$7F,$A0,$3D,$26,$7F,$49,$B3,$FE,$9F,$DF,$FD,$0B,$F0,$58,$D7
	.dc.b	$EC,$D5,$BF,$F4,$13,$6F,$FC,$FC,$DB,$E3,$3D,$4F,$C2,$7C,$47,$F9
	.dc.b	$19,$FB,$2D,$AF,$B2,$D7,$F8,$20,$9F,$F5,$D7,$3F,$9A,$3E,$EF,$87
	.dc.b	$E1,$CF,$FD,$36,$F8,$6F,$A6,$5F,$90,$8D,$71,$FF,$6F,$CB,$F2,$31
	.dc.b	$BC,$72,$FA,$FA,$FF,$DD,$DE,$D5,$C3,$FC,$A5,$7F,$A3,$F9,$CC,$4C
	.dc.b	$8F,$F4,$C4,$EB,$4F,$D6,$D6,$5F,$F4,$AF,$D6,$3E,$BB,$B3,$6F,$FD
	.dc.b	$FF,$F5,$CF,$AF,$CD,$49,$FF,$52,$56,$EF,$FC,$F1,$7B,$9C,$D7,$D3
	.dc.b	$F0,$5E,$17,$FB,$BE,$7F,$C2,$6D,$BF,$F0,$E3,$5F,$E3,$7D,$DF,$35
	.dc.b	$FF,$16,$F8,$7F,$ED,$FA,$FE,$49,$F1,$BD,$6F,$96,$DF,$E8,$5A,$D5
	.dc.b	$FF,$85,$BE,$BF,$F3,$98,$7F,$88,$9B,$1F,$E2,$B6,$BF,$E2,$A6,$BF
	.dc.b	$CF,$9A,$B2,$46,$1B,$D7,$E7,$9D,$4F,$0E,$2F,$EA,$73,$F6,$FC,$9F
	.dc.b	$0E,$5F,$DE,$EA,$E3,$FE,$4A,$9B,$87,$7F,$AF,$EC,$4C,$3B,$64,$8C
	.dc.b	$5F,$F8,$F9,$B5,$FE,$47,$7C,$5F,$5F,$FE,$DE,$9F,$F6,$E3,$67,$FC
	.dc.b	$77,$3E,$0C,$D3,$9C,$FF,$C2,$1F,$5F,$FD,$97,$0F,$F2,$5B,$70,$FF
	.dc.b	$2A,$35,$FF,$29,$B5,$FE,$CE,$D7,$E5,$37,$13,$4D,$60,$FA,$FF,$D8
	.dc.b	$5F,$FB,$CD,$B7,$FE,$52,$AF,$16,$1F,$EC,$59,$CB,$FE,$27,$FA,$FF
	.dc.b	$D7,$57,$F9,$A5,$AF,$F9,$A1,$AF,$F7,$C6,$BF,$7D,$B8,$9A,$4E,$9F
	.dc.b	$C0,$65,$FE,$52,$F2,$FF,$71,$EA,$7F,$B9,$74,$FF,$48,$5F,$8B,$2F
	.dc.b	$F2,$9D,$7E,$97,$D5,$75,$FD,$C6,$DF,$F9,$A6,$30,$47,$04,$D3,$FC
	.dc.b	$D9,$1E,$2F,$5D,$F6,$5F,$C2,$5B,$5F,$EA,$1B,$EB,$6C,$7F,$DD,$95
	.dc.b	$A8,$FF,$3E,$FD,$83,$E2,$7F,$1E,$DC,$3F,$D1,$6D,$9F,$F4,$A3,$5F
	.dc.b	$F4,$9B,$5F,$F1,$3B,$5F,$D5,$37,$80,$E7,$7F,$F5,$57,$D4,$FF,$21
	.dc.b	$F8,$13,$FC,$2E,$7C,$19,$FF,$46,$AD,$E3,$FE,$BD,$FB,$07,$D5,$71
	.dc.b	$EE,$FE,$25,$18,$9F,$F1,$F5,$F1,$CB,$F8,$3B,$FE,$6D,$9F,$F5,$D4
	.dc.b	$FF,$AA,$1B,$A7,$FA,$C5,$AF,$FA,$C1,$AF,$FB,$0B,$EC,$1F,$61,$FC
	.dc.b	$E5,$B5,$E0,$35,$E1,$66,$7F,$AC,$53,$EC,$1F,$3D,$FC,$2E,$30,$D8
	.dc.b	$CD,$FF,$B5,$67,$2D,$8C,$C7,$FB,$65,$AF,$FB,$D5,$AF,$F5,$C6,$AF
	.dc.b	$31,$F1,$BF,$EB,$CA,$F4,$4E,$63,$0F,$E2,$63,$95,$FF,$61,$9C,$E9
	.dc.b	$F5,$7E,$C3,$F6,$B5,$FE,$E6,$6D,$B8,$51,$FE,$78,$9E,$F1,$89,$FE
	.dc.b	$C5,$3C,$23,$13,$A8,$E1,$3E,$33,$A3,$BF,$AF,$FF,$7C,$4F,$46,$CE
	.dc.b	$E9,$0F,$B0,$FE,$E0,$EB,$0F,$C7,$36,$FF,$DF,$ED,$BF,$53,$C6,$3B
	.dc.b	$7C,$5F,$C3,$F8,$09,$1F,$2B,$57,$C7,$FE,$4F,$0F,$44,$7D,$0A,$DC
	.dc.b	$63,$5F,$F4,$B5,$F1,$BE,$19,$7F,$89,$B0,$E1,$3D,$6B,$77,$ED,$F9
	.dc.b	$7F,$F7,$2C,$E7,$1F,$F0,$64,$62,$7E,$17,$AF,$FE,$E3,$AF,$A2,$FF
	.dc.b	$6C,$7F,$D4,$9B,$23,$53,$C3,$CE,$FF,$0B,$46,$33,$FD,$A3,$C6,$7E
	.dc.b	$F3,$9B,$E6,$E6,$DF,$34,$F8,$4F,$F1,$03,$FE,$19,$C3,$F7,$CE,$6F
	.dc.b	$F0,$AE,$5F,$2B,$9C,$3E,$E3,$97,$F1,$CE,$67,$9F,$A9,$FB,$0C,$DF
	.dc.b	$17,$FA,$F6,$E1,$FA,$D6,$E1,$FE,$42,$3F,$D8,$CE,$1C,$35,$FF,$23
	.dc.b	$B5,$DC,$F5,$3F,$C0,$DB,$BD,$77,$D8,$7F,$91,$0F,$24,$DC,$3F,$47
	.dc.b	$CE,$F8,$1F,$F2,$23,$63,$3E,$E6,$26,$47,$07,$71,$8C,$BF,$D5,$1F
	.dc.b	$53,$FB,$7F,$23,$FC,$A6,$DB,$FF,$51,$B6,$77,$1F,$8D,$9F,$F3,$77
	.dc.b	$D0,$7D,$7C,$BF,$C4,$0D,$9F,$8D,$48,$C3,$F2,$5A,$FF,$9B,$1A,$FF
	.dc.b	$BC,$5A,$F1,$E2,$F9,$51,$8D,$BE,$47,$C7,$FF,$9A,$8F,$1C,$FD,$6E
	.dc.b	$2F,$F5,$EC,$7F,$9A,$7E,$8F,$FA,$FC,$7E,$9F,$E6,$BF,$A5,$E5,$FC
	.dc.b	$3A,$72,$FF,$C0,$1C,$13,$AC,$BF,$83,$56,$9F,$E4,$D1,$9B,$FE,$A5
	.dc.b	$39,$AE,$3E,$4F,$F9,$E7,$E9,$7F,$C5,$58,$F5,$F9,$EE,$5D,$29,$CE
	.dc.b	$7F,$CC,$67,$31,$8C,$11,$8C,$FF,$D1,$CF,$8F,$CC,$7D,$37,$E0,$F0
	.dc.b	$FC,$B4,$63,$97,$FC,$8F,$39,$9F,$F4,$DC,$E7,$3F,$AF,$9B,$EA,$FF
	.dc.b	$86,$C7,$F0,$06,$D7,$F9,$86,$FB,$6F,$FD,$C6,$73,$7C,$7E,$63,$FD
	.dc.b	$31,$06,$F9,$D8,$66,$3F,$CF,$A7,$4D,$9C,$E3,$FC,$EF,$2F,$6F,$17
	.dc.b	$A8,$95,$F8,$E8,$CB,$EA,$BF,$D5,$BD,$7A,$BF,$D2,$5A,$2F,$CB,$1F
	.dc.b	$EB,$7E,$97,$D8,$7E,$D1,$B7,$F9,$AE,$24,$63,$3F,$F6,$CB,$5F,$F6
	.dc.b	$2E,$DF,$27,$EC,$3E,$09,$D7,$0B,$87,$BE,$2D,$7F,$40,$BE,$9C,$1F
	.dc.b	$2E,$72,$DA,$FE,$91,$3A,$7E,$A4,$FA,$7D,$2F,$41,$23,$D1,$3F,$E2
	.dc.b	$BD,$BF,$CA,$11,$F6,$FD,$9E,$0F,$D1,$7F,$C2,$3F,$46,$FB,$07,$C0
	.dc.b	$B4,$FF,$7D,$78,$3F,$EE,$2A,$D4,$E7,$E2,$FF,$C2,$F5,$FE,$A5,$F9
	.dc.b	$EF,$F0,$2C,$62,$15,$C1,$E6,$75,$71,$1B,$6D,$F8,$79,$97,$D2,$3C
	.dc.b	$58,$F9,$B0,$E6,$76,$71,$87,$36,$CF,$3A,$E5,$5E,$19,$CB,$71,$F6
	.dc.b	$57,$66,$C7,$B7,$E6,$DA,$62,$3C,$13,$73,$F4,$C7,$33,$DF,$9F,$AA
	.dc.b	$F8,$C7,$A2,$76,$F2,$31,$8F,$92,$B8,$61,$D6,$3C,$F3,$BC,$FD,$D8
	.dc.b	$F9,$AF,$73,$CA,$38,$11,$F1,$E7,$F0,$9E,$59,$FB,$DE,$06,$B3,$F3
	.dc.b	$E1,$DA,$32,$D3,$F0,$38,$78,$1C,$57,$36,$CF,$C9,$CB,$C5,$18,$8F
	.dc.b	$09,$93,$AC,$F8,$DF,$EE,$19,$FE,$25,$B9,$FE,$05,$B9,$FF,$98,$E5
	.dc.b	$F6,$0F,$8E,$C3,$6B,$FA,$AC,$EA,$72,$48,$FA,$93,$BC,$7F,$28,$DC
	.dc.b	$7F,$29,$F6,$0F,$AD,$FD,$35,$F0,$C3,$F0,$57,$B8,$FC,$F3,$75,$5F
	.dc.b	$19,$DD,$47,$91,$F8,$88,$D3,$8C,$7C,$73,$B7,$E8,$CE,$9C,$CE,$F1
	.dc.b	$FD,$C6,$1E,$1A,$D4,$62,$F9,$36,$BC,$18,$FB,$30,$E6,$49,$D3,$E0
	.dc.b	$8D,$46,$09,$3F,$02,$BF,$29,$7E,$08,$C1,$E5,$2D,$F3,$7C,$66,$2B
	.dc.b	$9E,$3F,$23,$0E,$F3,$92,$1F,$C3,$B1,$F7,$46,$A3,$18,$F4,$69,$EE
	.dc.b	$3F,$7B,$CF,$FD,$87,$8F,$ED,$F3,$7F,$5D,$6F,$07,$F6,$36,$D7,$FA
	.dc.b	$9D,$7C,$1E,$3A,$7F,$A3,$BF,$EB,$56,$CF,$F3,$48,$C4,$F0,$8F,$EB
	.dc.b	$1D,$FD,$31,$FE,$E4,$DE,$7F,$F2,$04,$FF,$A2,$7E,$C1,$F5,$79,$A9
	.dc.b	$3F,$DA,$DF,$14,$EA,$30,$43,$FD,$41,$5D,$1F,$F4,$8C,$6D,$B5,$E9
	.dc.b	$8C,$47,$E7,$B2,$46,$23,$FC,$50,$D7,$FD,$79,$4F,$98,$F5,$12,$3F
	.dc.b	$D9,$7C,$3F,$E1,$8F,$33,$E5,$B6,$3F,$BE,$78,$29,$FF,$6F,$4E,$67
	.dc.b	$FC,$3B,$39,$7F,$BB,$7D,$7B,$EC,$3E,$93,$63,$FD,$60,$D7,$FD,$5E
	.dc.b	$FE,$B3,$8B,$FD,$54,$E5,$B3,$C8,$E7,$C9,$2B,$9B,$5F,$F0,$B5,$FF
	.dc.b	$A1,$6B,$F9,$EE,$1F,$E3,$49,$F4,$CF,$FB,$DD,$FF,$15,$18,$3F,$DF
	.dc.b	$5C,$1F,$E4,$3E,$6F,$C5,$7D,$83,$D5,$A7,$E1,$7C,$AF,$ED,$99,$7F
	.dc.b	$4A,$BE,$18,$7F,$B3,$6F,$83,$6F,$86,$1F,$8D,$F3,$1F,$61,$F9,$7F
	.dc.b	$21,$F2,$F4,$D4,$4C,$3F,$DB,$13,$A6,$CE,$5F,$F6,$CF,$CD,$BE,$1B
	.dc.b	$F4,$4D,$8F,$F7,$03,$5F,$F7,$B7,$A4,$FB,$07,$D7,$FE,$B2,$30,$FF
	.dc.b	$A2,$A7,$FD,$15,$3A,$7F,$D1,$B3,$A6,$C6,$5F,$F4,$73,$5F,$F4,$73
	.dc.b	$5F,$F4,$73,$5E,$B3,$FB,$7F,$5F,$FA,$67,$0F,$FB,$B7,$DE,$FD,$23
	.dc.b	$7B,$BF,$7D,$3A,$F0,$BF,$D1,$0D,$C1,$DF,$FE,$FE,$AF,$E0,$98,$72
	.dc.b	$BD,$F9,$1F,$BF,$FB,$0F,$F4,$7E,$1F,$E2,$FC,$BE,$E4,$62,$11,$8F
	.dc.b	$57,$87,$D5,$71,$64,$F1,$C3,$FD,$EF,$7C,$6F,$9D,$FA,$3D,$E9,$7B
	.dc.b	$3C,$FA,$FF,$BF,$5C,$1F,$EF,$4D,$BF,$8F,$B4,$FA,$9E,$F7,$C9,$70
	.dc.b	$7F,$B0,$1C,$12,$3F,$40,$D8,$FF,$47,$F9,$BF,$E7,$37,$C5,$D9,$FE
	.dc.b	$46,$F6,$3F,$91,$4F,$FA,$82,$72,$ED,$F8,$7A,$FF,$95,$6B,$FC,$9B
	.dc.b	$D1,$C2,$E5,$70,$8F,$E5,$9E,$EB,$E3,$EF,$1C,$AF,$0C,$E5,$B3,$97
	.dc.b	$B8,$C1,$2A,$EB,$5F,$7B,$E7,$3F,$9D,$7D,$87,$F1,$5C,$38,$E0,$8C
	.dc.b	$71,$7E,$36,$B8,$57,$D3,$E7,$E9,$18,$8F,$0C,$3E,$C3,$EE,$6D,$FE
	.dc.b	$D7,$CD,$7B,$C3,$A6,$1C,$EF,$6F,$A7,$D8,$7C,$17,$F4,$F6,$C7,$D2
	.dc.b	$6B,$E8,$5F,$01,$B9,$93,$B7,$94,$7F,$59,$FA,$FF,$F1,$74,$FC,$06
	.dc.b	$E1,$FC,$5D,$B9,$FD,$9F,$AF,$7D,$87,$1F,$5A,$61,$F8,$4B,$DB,$F4
	.dc.b	$6B,$6F,$C7,$F8,$7F,$F3,$5B,$F1,$9B,$8F,$F9,$BB,$81,$18,$CB,$FC
	.dc.b	$6F,$F6,$1F,$EB,$CB,$FC,$FB,$5F,$F1,$D9,$FE,$A1,$70,$E1,$AF,$2F
	.dc.b	$5F,$F9,$E4,$62,$38,$46,$08,$FF,$BB,$BD,$EF,$C4,$3F,$A0,$6B,$F8
	.dc.b	$66,$E3,$FC,$F5,$FC,$A4,$7F,$B8,$FE,$11,$E0,$7F,$64,$6C,$7F,$B8
	.dc.b	$DA,$FA,$BD,$1F,$F7,$94,$EF,$E2,$7F,$C1,$DF,$25,$FC,$9F,$1F,$E9
	.dc.b	$4D,$D3,$FD,$61,$8A,$BF,$C2,$9F,$60,$FB,$0A,$4C,$FF,$BD,$FC,$57
	.dc.b	$ED,$72,$FE,$C2,$FE,$DE,$3F,$CC,$BA,$23,$11,$FE,$BD,$6A,$8D,$78
	.dc.b	$1F,$E0,$09,$CC,$62,$7F,$E0,$04,$60,$E3,$DC,$F9,$67,$C3,$7F,$8A
	.dc.b	$DB,$5F,$C3,$9A,$FE,$F2,$DF,$18,$F8,$7F,$F0,$9B,$67,$F0,$67,$95
	.dc.b	$C3,$F3,$5A,$FB,$9A,$FF,$B6,$5A,$F1,$8F,$F0,$DC,$62,$7C,$31,$FE
	.dc.b	$02,$F8,$3F,$F2,$25,$7F,$08,$6E,$9F,$DF,$DB,$C3,$FE,$E5,$FA,$F7
	.dc.b	$D8,$7F,$8A,$9B,$7E,$8C,$51,$88,$FC,$23,$5F,$AA,$D7,$A4,$FF,$8C
	.dc.b	$23,$11,$F5,$5B,$3D,$27,$FC,$33,$F6,$1F,$C1,$AB,$F9,$DB,$5F,$E4
	.dc.b	$2D,$C7,$FC,$4B,$AF,$F8,$5A,$7E,$93,$5F,$C3,$B5,$1F,$85,$70,$DF
	.dc.b	$6F,$FC,$8B,$18,$C3,$F0,$31,$82,$4F,$DA,$DF,$F9,$D7,$B1,$FE,$6F
	.dc.b	$7F,$C3,$91,$FC,$D9,$B8,$7F,$8E,$BE,$BD,$F6,$1F,$E5,$66,$D7,$CD
	.dc.b	$6B,$FA,$F6,$BF,$A0,$6B,$EA,$C5,$18,$8F,$D0,$36,$33,$3F,$E6,$08
	.dc.b	$D5,$BD,$5F,$E4,$6D,$F0,$FF,$D1,$11,$FC,$9A,$DF,$25,$F3,$F8,$7C
	.dc.b	$0F,$D3,$CB,$F6,$6D,$73,$E3,$7F,$9B,$A3,$11,$FB,$36,$D7,$49,$D1
	.dc.b	$27,$E6,$7D,$2B,$EC,$3F,$D3,$4D,$8F,$EF,$AD,$7F,$61,$F4,$FE,$73
	.dc.b	$FA,$3B,$FC,$19,$AF,$C7,$6B,$FC,$15,$BA,$7F,$07,$F4,$3F,$D5,$18
	.dc.b	$FF,$26,$E0,$FF,$32,$37,$4F,$C4,$F8,$FF,$D9,$BC,$AD,$DC,$CB,$FC
	.dc.b	$23,$7B,$7F,$C7,$3E,$77,$F4,$E7,$FB,$CF,$D5,$BE,$C3,$FD,$6C,$D8
	.dc.b	$FF,$0C,$B5,$FE,$19,$F6,$1F,$C3,$1C,$CF,$02,$3F,$EA,$28,$D3,$FD
	.dc.b	$FB,$D9,$FF,$54,$3F,$E0,$36,$B9,$F8,$CF,$F6,$8F,$33,$EC,$3F,$55
	.dc.b	$C1,$FD,$7E,$F6,$DA,$FF,$2B,$9C,$27,$7C,$1F,$AC,$BE,$5D,$7F,$E6
	.dc.b	$F6,$E3,$FC,$81,$AF,$F6,$46,$BF,$EC,$3C,$7F,$D9,$58,$F8,$7E,$43
	.dc.b	$FC,$CA,$79,$BE,$0F,$FD,$59,$8F,$F9,$9A,$33,$7E,$D7,$5F,$19,$FE
	.dc.b	$4E,$FA,$F7,$D8,$7F,$98,$2B,$F9,$8C,$63,$1F,$F4,$23,$63,$33,$B7
	.dc.b	$FB,$CC,$3E,$B1,$C5,$C1,$B3,$D1,$AF,$06,$BF,$D0,$3E,$57,$FD,$70
	.dc.b	$9A,$47,$E2,$6A,$F8,$36,$BF,$E6,$DD,$D3,$98,$FF,$1F,$7E,$0A,$FB
	.dc.b	$1D,$E1,$FE,$4E,$BD,$E1,$8D,$B3,$18,$C3,$A6,$3C,$EF,$FC,$8B,$8F
	.dc.b	$39,$CE,$1C,$5F,$D0,$C3,$E0,$BF,$C4,$F5,$86,$BF,$E9,$88,$D4,$FF
	.dc.b	$60,$EC,$46,$27,$F0,$5E,$37,$F9,$42,$BD,$CD,$8F,$1D,$F8,$56,$5F
	.dc.b	$EC,$EE,$AC,$36,$F8,$C6,$F1,$EF,$7A,$C3,$A6,$9F,$EE,$CE,$97,$D4
	.dc.b	$F5,$EC,$46,$32,$FF,$45,$D7,$F9,$FD,$FD,$75,$6E,$75,$A7,$ED,$A3
	.dc.b	$CD,$DF,$FE,$AE,$EE,$D4,$78,$FE,$C3,$E6,$57,$F9,$C9,$B8,$FF,$BF
	.dc.b	$BF,$08,$FF,$6C,$E5,$FE,$79,$9C,$BF,$C4,$1B,$5F,$EC,$46,$E6,$AE
	.dc.b	$33,$D7,$93,$E4,$73,$BF,$3A,$7D,$07,$D6,$8C,$36,$33,$97,$F8,$5A
	.dc.b	$72,$FF,$80,$7C,$47,$81,$F1,$DB,$1F,$EC,$19,$DE,$3F,$CD,$2B,$C3
	.dc.b	$7F,$02,$3F,$B8,$9F,$AA,$70,$FF,$74,$9F,$36,$3E,$32,$5F,$8E,$B7
	.dc.b	$5C,$67,$2F,$F9,$57,$EC,$BF,$6D,$5F,$ED,$3D,$11,$89,$FF,$58,$E1
	.dc.b	$DA,$B5,$39,$DB,$F5,$3C,$3F,$EF,$0E,$4F,$CD,$B6,$BF,$C0,$1C,$BF
	.dc.b	$99,$C7,$D7,$39,$7F,$CF,$ED,$73,$1B,$7F,$D9,$CD,$7F,$D8,$4D,$C9
	.dc.b	$F6,$5E,$3B,$FE,$FC,$DA,$DE,$1F,$97,$8C,$60,$FA,$47,$DB,$FA,$1B
	.dc.b	$87,$F8,$5D,$B8,$79,$9A,$FA,$9A,$FF,$B8,$DA,$FF,$B5,$DB,$5D,$EB
	.dc.b	$F7,$F8,$BE,$CB,$FC,$5B,$87,$D1,$6E,$1F,$ED,$8C,$3C,$3A,$7F,$16
	.dc.b	$BE,$13,$FA,$8F,$C0,$FF,$89,$3F,$31,$AF,$FB,$D5,$AF,$FB,$A5,$B9
	.dc.b	$1A,$4E,$EF,$B2,$F0,$A7,$D9,$7F,$9F,$33,$FC,$33,$70,$FF,$76,$D7
	.dc.b	$F2,$3C,$7C,$D3,$97,$F0,$BF,$48,$F9,$FF,$C2,$B7,$0F,$F2,$23,$70
	.dc.b	$FC,$1B,$5F,$C3,$B5,$FF,$29,$E6,$8C,$42,$31,$5D,$8F,$9F,$ED,$FF
	.dc.b	$10,$C3,$E5,$37,$0E,$78,$78,$67,$3D,$9F,$C7,$7F,$0A,$7D,$9F,$D8
	.dc.b	$6D,$FF,$9F,$5B,$9F,$FA,$FD,$AF,$FA,$C1,$AF,$F9,$45,$AF,$FA,$BD
	.dc.b	$BB,$1B,$F6,$6F,$8F,$FE,$5C,$DB,$FD,$B4,$F8,$32,$FC,$24,$E7,$0F
	.dc.b	$F5,$67,$D9,$BE,$B3,$F9,$B3,$70,$FD,$F7,$E1,$6F,$C1,$73,$5F,$CC
	.dc.b	$BE,$2D,$F4,$1F,$BA,$6D,$FF,$9E,$5B,$8F,$EE,$1A,$FE,$F3,$ED,$5F
	.dc.b	$6B,$FC,$F1,$B5,$FC,$37,$35,$E6,$BF,$AB,$5B,$E6,$9F,$4B,$FD,$92
	.dc.b	$30,$D8,$CE,$1F,$EC,$49,$CB,$63,$31,$FE,$C8,$6B,$FC,$A8,$D9,$99
	.dc.b	$ED,$F6,$9C,$38,$3F,$C0,$72,$8C,$5F,$F3,$2B,$B3,$FE,$A4,$9B,$D2
	.dc.b	$FB,$4F,$F4,$35,$FF,$AA,$30,$FE,$4B,$18,$B4,$63,$3F,$EC,$33,$98
	.dc.b	$DF,$C1,$E0,$CC,$DB,$E8,$7E,$8B,$EC,$DF,$69,$FD,$A9,$B1,$AC,$7F
	.dc.b	$49,$0F,$95,$D5,$4A,$D6,$6A,$E3,$7E,$69,$ED,$7B,$8C,$0F,$1D,$F3
	.dc.b	$DF,$8D,$8C,$C6,$39,$7F,$C2,$5F,$4E,$FA,$EF,$F3,$D6,$3E,$5A,$D3
	.dc.b	$70,$E5,$BF,$F8,$D2,$73,$8F,$83,$0E,$5F,$09,$FE,$17,$BD,$E1,$FE
	.dc.b	$8E,$F9,$CF,$F5,$8B,$D5,$B7,$E4,$9D,$BE,$98,$CE,$3F,$D6,$BC,$7F
	.dc.b	$F7,$DD,$FC,$46,$D7,$C0,$DF,$FC,$9D,$F0,$3F,$E1,$89,$CD,$F1,$F8
	.dc.b	$FC,$19,$9B,$87,$CE,$FA,$37,$C5,$FF,$88,$1B,$19,$F6,$B1,$32,$38
	.dc.b	$3B,$8C,$67,$FE,$50,$FA,$7F,$F2,$4F,$0F,$F9,$19,$BF,$00,$69,$30
	.dc.b	$FF,$26,$46,$5B,$19,$AF,$D4,$7C,$B7,$ED,$3B,$4F,$F9,$16,$F8,$4F
	.dc.b	$E0,$8F,$DC,$47,$F8,$EB,$B9,$F3,$B4,$C7,$C9,$5C,$BF,$1F,$3A,$6C
	.dc.b	$E9,$A7,$FA,$C7,$A3,$FC,$A9,$F0,$9F,$89,$8F,$C3,$7C,$E6,$26,$46
	.dc.b	$CC,$97,$2F,$C9,$F8,$EF,$B2,$F8,$ED,$9F,$F3,$73,$67,$F8,$37,$07
	.dc.b	$F2,$5A,$D9,$2B,$6D,$9F,$E1,$9D,$9F,$73,$6F,$CF,$F1,$FC,$8B,$FC
	.dc.b	$F6,$3D,$79,$FF,$D1,$F7,$B9,$FD,$8F,$D5,$FF,$53,$C7,$FA,$D3,$7A
	.dc.b	$8B,$E4,$3F,$CC,$EE,$49,$E8,$E2,$32,$FF,$9B,$3E,$B5,$F5,$54,$C7
	.dc.b	$6D,$CB,$F8,$0C,$E9,$B3,$A6,$9F,$C1,$BE,$C3,$FC,$12,$4E,$FF,$F1
	.dc.b	$FB,$FE,$07,$FA,$F7,$C8,$FF,$A1,$5B,$5C,$6B,$13,$FE,$80,$AF,$D9
	.dc.b	$EA,$7F,$6B,$EA,$B0,$E3,$BF,$EC,$67,$64,$F1,$B9,$E5,$E5,$7A,$79
	.dc.b	$5E,$42,$47,$F4,$1A,$FF,$03,$57,$F5,$28,$C6,$1D,$FE,$7F,$D0,$72
	.dc.b	$8F,$F3,$26,$28,$F4,$79,$9F,$00,$93,$D7,$3F,$54,$7E,$D3,$B9,$F6
	.dc.b	$9F,$EA,$D6,$D7,$CB,$9C,$4C,$8F,$E4,$93,$AC,$FE,$05,$65,$FE,$5B
	.dc.b	$19,$E7,$FE,$4C,$DA,$FF,$12,$E9,$FD,$2D,$57,$F3,$79,$D3,$F0,$8F
	.dc.b	$B2,$48,$F5,$4F,$F9,$4A,$3F,$D2,$3D,$9F,$BC,$F5,$FD,$47,$D9,$8B
	.dc.b	$5F,$D3,$D7,$D5,$FE,$BF,$2F,$0F,$C8,$FF,$7C,$9F,$1C,$E5,$FF,$7E
	.dc.b	$78,$7E,$6E,$17,$D2,$BE,$87,$FD,$EB,$87,$1B,$EB,$E1,$32,$EB,$8F
	.dc.b	$FA,$9B,$2E,$97,$B6,$C7,$5C,$7A,$67,$C3,$1F,$1E,$3C,$DF,$19,$8C
	.dc.b	$74,$F8,$1E,$19,$F9,$B3,$F2,$E3,$E1,$BE,$59,$F8,$5B,$19,$BE,$39
	.dc.b	$78,$B1,$EE,$75,$2E,$7E,$87,$C4,$79,$32,$F0,$67,$E4,$9D,$4F,$1C
	.dc.b	$BC,$18,$76,$CF,$CB,$3C,$A7,$2F,$6C,$FC,$AB,$1E,$AE,$6F,$B4,$EA
	.dc.b	$3B,$67,$F2,$F0,$EB,$18,$3B,$47,$BD,$F4,$67,$EF,$C7,$C2,$49,$D3
	.dc.b	$F8,$5C,$7C,$2E,$6B,$99,$23,$33,$C0,$91,$9A,$ED,$18,$8F,$98,$7F
	.dc.b	$90,$72,$FC,$83,$71,$FC,$03,$72,$FF,$26,$4B,$EC,$DF,$29,$86,$D7
	.dc.b	$F0,$B9,$D4,$E4,$91,$F6,$A7,$79,$FE,$4D,$B8,$FE,$7F,$DF,$FF,$52
	.dc.b	$57,$DB,$9C,$BF,$A6,$6B,$FA,$C7,$3A,$7F,$A7,$A3,$86,$11,$F0,$3F
	.dc.b	$96,$AE,$B1,$82,$61,$E3,$9D,$7C,$0F,$F0,$69,$CE,$3F,$AB,$C7,$C3
	.dc.b	$97,$BA,$7F,$25,$8F,$4C,$79,$47,$03,$D1,$8F,$47,$C9,$1A,$26,$1F
	.dc.b	$07,$1F,$04,$62,$78,$9C,$F2,$F7,$3D,$4F,$85,$7D,$2F,$94,$53,$8D
	.dc.b	$73,$AE,$2F,$58,$E2,$75,$C3,$D5,$5E,$FB,$E5,$3A,$AF,$CA,$56,$A3
	.dc.b	$F9,$57,$D1,$7C,$7E,$5F,$E6,$2D,$E1,$FE,$5F,$C7,$FE,$22,$8C,$57
	.dc.b	$FB,$57,$BF,$FD,$5A,$FF,$70,$F7,$7F,$D3,$5D,$29,$CC,$FF,$A5,$E3
	.dc.b	$13,$A9,$CC,$FF,$34,$FB,$3F,$F1,$F3,$FE,$6B,$6B,$FE,$72,$6E,$4C
	.dc.b	$BF,$DD,$34,$FA,$EF,$2E,$6B,$DD,$7A,$9B,$87,$F9,$FD,$B9,$79,$EF
	.dc.b	$8C,$6A,$F9,$7A,$38,$3E,$0D,$6E,$FF,$C1,$F7,$EB,$23,$FD,$A7,$1F
	.dc.b	$23,$FD,$D3,$BB,$FD,$10,$FF,$72,$F3,$3E,$5F,$1F,$F8,$15,$EB,$5F
	.dc.b	$9E,$D7,$DD,$5F,$3F,$93,$E4,$11,$EC,$FF,$B2,$74,$7D,$9F,$1D,$FE
	.dc.b	$CB,$6B,$FD,$3C,$DC,$BF,$C2,$91,$F2,$23,$55,$CA,$F0,$79,$B3,$FF
	.dc.b	$68,$63,$E4,$C3,$8E,$1F,$E0,$C3,$47,$3F,$91,$8F,$F8,$5F,$0F,$2C
	.dc.b	$79,$F1,$34,$FC,$27,$37,$C2,$25,$7F,$8F,$3E,$F2,$BF,$E9,$9D,$7F
	.dc.b	$B5,$65,$F1,$B8,$3F,$21,$7F,$2A,$FE,$BF,$96,$FB,$3A,$4C,$7F,$CA
	.dc.b	$FE,$1D,$7F,$D4,$93,$B7,$FC,$B7,$3C,$A7,$F4,$3E,$6B,$EC,$FE,$ED
	.dc.b	$7F,$AE,$DB,$9F,$F9,$91,$BC,$7F,$AA,$FC,$07,$F4,$73,$FE,$C3,$7F
	.dc.b	$43,$59,$E9,$FB,$51,$88,$FF,$6A,$37,$88,$D2,$4F,$03,$F0,$5E,$A7
	.dc.b	$F4,$2B,$FF,$68,$37,$0F,$DD,$B7,$0F,$F7,$07,$D9,$3E,$13,$F5,$33
	.dc.b	$9C,$8D,$31,$3B,$AF,$F7,$7F,$DA,$7F,$BC,$DB,$F0,$3F,$A6,$8C,$47
	.dc.b	$FB,$3D,$FF,$7E,$BA,$21,$98,$E1,$F0,$7F,$C3,$A1,$EF,$7C,$A7,$FC
	.dc.b	$82,$D7,$3D,$DF,$C5,$3C,$77,$E1,$0F,$A2,$C3,$6B,$FC,$D1,$3A,$9F
	.dc.b	$F3,$54,$7F,$8C,$67,$4F,$F4,$93,$11,$A9,$E1,$1C,$A7,$6F,$F4,$F8
	.dc.b	$CB,$C5,$AF,$F5,$3E,$AF,$F2,$5F,$A1,$FC,$2B,$C9,$FF,$39,$4F,$F9
	.dc.b	$73,$8F,$F3,$51,$88,$FF,$7B,$2F,$98,$E1,$52,$78,$72,$7F,$99,$30
	.dc.b	$EB,$18,$0F,$F9,$6A,$BF,$8C,$FC,$03,$E0,$69,$2B,$85,$25,$F6,$99
	.dc.b	$96,$7F,$AF,$9E,$51,$C1,$AE,$B1,$F2,$47,$28,$F2,$73,$3E,$15,$F4
	.dc.b	$3F,$CE,$1B,$3A,$6B,$CC,$FC,$BF,$B7,$FC,$FB,$A3,$FD,$87,$1D,$1A
	.dc.b	$EE,$D1,$8C,$BF,$DA,$FF,$8B,$BE,$0B,$96,$13,$0E,$4F,$EF,$A7,$51
	.dc.b	$F1,$3D,$77,$C1,$7F,$42,$6C,$66,$31,$F8,$2E,$59,$19,$39,$47,$F4
	.dc.b	$AF,$B5,$FF,$71,$47,$D0,$6B,$EF,$E2,$46,$33,$FE,$3C,$BE,$0B,$97
	.dc.b	$C1,$CB,$E4,$F5,$8F,$D5,$7D,$AB,$DF,$A4,$AF,$F4,$F7,$84,$FA,$AF
	.dc.b	$D8,$36,$3F,$4A,$D7,$FD,$A0,$D7,$F4,$CD,$7F,$C6,$0D,$7F,$D6,$ED
	.dc.b	$4D,$35,$9F,$FA,$9B,$A7,$FB,$6F,$BD,$F8,$AD,$7F,$9E,$37,$A7,$FD
	.dc.b	$59,$F0,$FF,$C3,$27,$FA,$07,$C1,$7F,$32,$CB,$F5,$CD,$7F,$48,$D7
	.dc.b	$F4,$6D,$F5,$54,$FB,$5F,$EE,$6D,$AF,$D5,$4E,$A7,$FB,$BF,$C9,$BE
	.dc.b	$1F,$FB,$CF,$57,$ED,$F2,$FE,$C8,$DC,$FF,$D8,$4D,$C7,$FD,$29,$F2
	.dc.b	$BC,$5E,$AD,$7F,$88,$64,$8C,$47,$F9,$6E,$31,$87,$F1,$86,$C7,$FC
	.dc.b	$00,$9D,$BF,$ED,$16,$BE,$09,$E1,$18,$97,$CB,$FF,$B5,$6B,$DE,$D7
	.dc.b	$F6,$11,$AF,$86,$FF,$10,$FC,$17,$C5,$9F,$CB,$35,$FE,$3C,$D7,$F8
	.dc.b	$A3,$5F,$E2,$6D,$E2,$34,$9B,$7F,$92,$9E,$16,$FB,$27,$C5,$FF,$8A
	.dc.b	$1B,$87,$D6,$6B,$FA,$9F,$BB,$BC,$1F,$E0,$57,$31,$FC,$CD,$AF,$F2
	.dc.b	$96,$BF,$CA,$1B,$91,$A4,$BF,$F2,$FB,$C3,$07,$C7,$7F,$87,$2F,$F1
	.dc.b	$2D,$7E,$5B,$70,$FF,$0C,$7C,$E3,$E8,$3F,$C9,$AD,$AF,$EA,$47,$F3
	.dc.b	$A7,$0F,$F9,$AD,$AF,$F4,$D6,$BF,$D0,$DA,$F1,$8F,$F3,$AC,$63,$23
	.dc.b	$AC,$1F,$62,$FB,$2F,$F2,$AB,$70,$FD,$1B,$5F,$E2,$DE,$47,$C0,$7F
	.dc.b	$A4,$B5,$FF,$40,$35,$FE,$C4,$D7,$FA,$C3,$5F,$EA,$ED,$C0,$D2,$67
	.dc.b	$FA,$1F,$B2,$FE,$63,$5F,$96,$6E,$9F,$E5,$56,$F3,$A5,$F4,$AF,$CD
	.dc.b	$9F,$29,$FE,$9C,$8C,$3F,$E0,$46,$CF,$D2,$6B,$FE,$05,$6B,$FE,$CF
	.dc.b	$6B,$DD,$AF,$82,$7B,$BC,$DA,$FF,$00,$FA,$8F,$F7,$9B,$FB,$06,$E3
	.dc.b	$FE,$6F,$6E,$1D,$69,$F7,$7E,$96,$DF,$58,$F5,$46,$23,$D0,$D7,$FD
	.dc.b	$CD,$A7,$F7,$2F,$1F,$FD,$53,$E8,$7F,$4F,$C3,$FD,$82,$DA,$F9,$5E
	.dc.b	$C3,$E0,$F7,$72,$7F,$BE,$D6,$DF,$F1,$94,$FF,$A7,$BE,$B7,$F8,$7C
	.dc.b	$7F,$2C,$6B,$FC,$39,$BA,$FF,$9E,$F6,$FC,$FB,$FE,$8F,$8C,$E5,$F6
	.dc.b	$F9,$51,$8B,$46,$3B,$BF,$D6,$F8,$72,$BF,$C1,$7C,$9B,$EC,$BF,$65
	.dc.b	$18,$6E,$09,$D7,$C6,$7F,$B3,$3E,$6B,$F3,$9A,$7F,$AD,$27,$35,$FE
	.dc.b	$2F,$6B,$FE,$17,$6C,$7F,$AE,$B8,$55,$CB,$1F,$F7,$6B,$FA,$18,$CF
	.dc.b	$3F,$FA,$EE,$30,$7E,$96,$DF,$64,$F9,$CF,$F7,$9C,$61,$B1,$9C,$BF
	.dc.b	$C5,$F4,$FA,$DC,$D4,$9F,$45,$6D,$FF,$14,$56,$DF,$F2,$50,$FB,$2F
	.dc.b	$F3,$66,$9F,$E6,$86,$C7,$F1,$66,$BF,$E6,$96,$BF,$B1,$39,$7D,$97
	.dc.b	$EC,$9E,$0D,$78,$35,$E1,$08,$C4,$7F,$9D,$7E,$6B,$FC,$F2,$FE,$F5
	.dc.b	$A9,$A7,$FB,$46,$FF,$CD,$5B,$5F,$80,$FF,$76,$D7,$E8,$A7,$2F,$FA
	.dc.b	$07,$7F,$AF,$3A,$8F,$AE,$D7,$F8,$AC,$F0,$79,$46,$67,$8D,$7F,$3A
	.dc.b	$D9,$F9,$57,$E3,$63,$0D,$BF,$F4,$84,$E8,$EC,$59,$F5,$78,$9D,$71
	.dc.b	$FE,$91,$18,$25,$7A,$63,$2D,$9E,$04,$9F,$E9,$BE,$0F,$F9,$F3,$D2
	.dc.b	$F6,$36,$33,$BF,$FA,$1D,$B3,$A8,$C5,$FF,$A9,$EB,$83,$63,$2F,$FA
	.dc.b	$85,$AF,$EE,$7E,$C5,$F0,$FF,$2F,$A9,$18,$C7,$FB,$0C,$62,$75,$39
	.dc.b	$AF,$F1,$3B,$5F,$F1,$33,$5E,$D1,$C0,$FB,$0B,$EE,$FF,$CD,$57,$FE
	.dc.b	$C2,$6C,$7F,$70,$6B,$FE,$C3,$6B,$FC,$F7,$EC,$BF,$59,$3F,$D1,$5B
	.dc.b	$1F,$D1,$5A,$FF,$45,$6B,$FA,$39,$78,$1E,$46,$D7,$8E,$30,$47,$04
	.dc.b	$9F,$34,$7E,$4F,$B1,$7F,$93,$9F,$37,$1E,$63,$11,$FD,$89,$B9,$BE
	.dc.b	$C7,$FC,$C5,$8F,$D0,$6D,$7F,$95,$1A,$FF,$94,$DA,$FF,$33,$6B,$FC
	.dc.b	$C9,$B3,$D7,$CF,$FB,$D1,$AD,$FF,$BE,$F0,$FE,$43,$97,$F2,$F8,$7F
	.dc.b	$81,$A3,$07,$E1,$7F,$05,$F3,$93,$4C,$13,$3F,$C3,$C6,$5B,$1F,$9C
	.dc.b	$F3,$1F,$65,$49,$3F,$A2,$8C,$BF,$90,$FB,$0F,$AF,$8F,$F4,$C6,$EC
	.dc.b	$69,$3B,$FF,$34,$DA,$E5,$1F,$AC,$FB,$1F,$F4,$44,$7E,$C1,$B7,$FA
	.dc.b	$96,$BF,$C9,$F6,$46,$23,$FB,$07,$E8,$1F,$C1,$DF,$F4,$23,$5C,$F5
	.dc.b	$FE,$AE,$31,$1F,$D9,$1B,$91,$A4,$9F,$E0,$3E,$A3,$E8,$29,$32,$FE
	.dc.b	$0F,$19,$7F,$6B,$F6,$2F,$A6,$FD,$D3,$63,$31,$82,$38,$26,$5F,$EA
	.dc.b	$09,$E0,$FF,$A8,$BD,$97,$D9,$7F,$05,$6D,$7F,$A7,$E3,$07,$F0,$9F
	.dc.b	$75,$F5,$74,$93,$FC,$2A,$32,$FF,$0C,$FB,$17,$D6,$7F,$11,$6C,$66
	.dc.b	$30,$67,$E0,$A4,$71,$6E,$5F,$E7,$6F,$91,$7D,$17,$F0,$E6,$D7,$F9
	.dc.b	$EA,$51,$87,$84,$EC,$93,$B9,$D4,$FF,$69,$38,$CF,$5F,$A3,$95,$B6
	.dc.b	$CF,$F3,$3F,$B0,$61,$FC,$F1,$B5,$FD,$01,$B3,$B9,$D1,$27,$4D,$8F
	.dc.b	$E7,$DF,$60,$F9,$DF,$E8,$ED,$AF,$E9,$6D,$9D,$CE,$89,$3A,$6C,$7B
	.dc.b	$7E,$8D,$F5,$2F,$B0,$FE,$DD,$18,$6C,$66,$FF,$D4,$B3,$96,$C6,$6B
	.dc.b	$A3,$5F,$F0,$94,$ED,$CD,$7F,$77,$F9,$9F,$C0,$C7,$F8,$05,$AF,$F9
	.dc.b	$99,$AF,$FA,$2F,$C5,$46,$21,$DC,$F8,$AF,$F1,$B6,$5F,$DB,$9B,$3F
	.dc.b	$D9,$DA,$FE,$FD,$AF,$FA,$5E,$79,$7A,$FF,$DD,$27,$33,$D9,$B3,$AA
	.dc.b	$FE,$F7,$F3,$FF,$DB,$AB,$C6,$D9,$34,$9E,$97,$F8,$01,$B5,$CA,$3F
	.dc.b	$35,$F7,$F2,$3F,$28,$DB,$FD,$03,$5F,$CB,$72,$A3,$11,$FE,$5E,$FA
	.dc.b	$CF,$F0,$C1,$2B,$FC,$68,$FF,$84,$39,$7F,$DD,$31,$FE,$14,$6C,$FF
	.dc.b	$83,$5A,$FF,$25,$6B,$F1,$DA,$FD,$26,$E4,$69,$AD,$BE,$CF,$C7,$FE
	.dc.b	$04,$97,$FE,$F0,$8C,$B7,$74,$EB,$0F,$C3,$35,$F8,$3C,$7F,$C0,$49
	.dc.b	$5F,$BB,$6C,$7F,$07,$6B,$C2,$3C,$1D,$6F,$AF,$FF,$25,$36,$33,$18
	.dc.b	$3F,$C6,$FF,$55,$FE,$08,$AF,$F1,$93,$5F,$E9,$4D,$7F,$18,$D7,$F2
	.dc.b	$ED,$D0,$D2,$7A,$7F,$E4,$4F,$AB,$7C,$7D,$26,$3F,$E4,$A8,$CB,$63
	.dc.b	$38,$7F,$91,$67,$2D,$8F,$F2,$07,$D6,$FD,$8F,$28,$D2,$7C,$0F,$F9
	.dc.b	$53,$EC,$3F,$B1,$5F,$EB,$9B,$A1,$A6,$AF,$9F,$C6,$F6,$26,$7F,$E4
	.dc.b	$CC,$7F,$E0,$0B,$63,$38,$7F,$2F,$AE,$33,$93,$87,$87,$E8,$23,$83
	.dc.b	$FC,$D7,$87,$F2,$CA,$FE,$F4,$D7,$F7,$8D,$7D,$5E,$0F,$FA,$1F,$EB
	.dc.b	$DF,$4D,$49,$7F,$93,$8C,$BD,$A3,$E2,$11,$C1,$2B,$FC,$E9,$7C,$1B
	.dc.b	$3B,$6C,$ED,$F3,$C6,$5F,$F5,$C7,$3B,$F1,$4F,$F0,$04,$7F,$A9,$5A
	.dc.b	$E9,$AF,$F2,$56,$BF,$5C,$FB,$44,$AF,$8E,$FF,$9D,$3E,$9F,$F4,$17
	.dc.b	$FE,$BB,$EA,$FF,$64,$46,$27,$FD,$79,$1F,$F0,$07,$E6,$5F,$30,$C3
	.dc.b	$85,$05,$F1,$F8,$6F,$E0,$51,$98,$C6,$1D,$BA,$7E,$8C,$F6,$1C,$2C
	.dc.b	$BB,$E3,$DA,$7A,$D7,$0C,$7A,$C6,$33,$D6,$5D,$A3,$9E,$1C,$9E,$C6
	.dc.b	$CB,$1B,$71,$5E,$8C,$7C,$59,$FB,$E7,$78,$F8,$73,$F5,$65,$E0,$8C
	.dc.b	$4F,$82,$B8,$61,$E6,$C7,$B1,$2B,$78,$F6,$CF,$96,$33,$1E,$EF,$CE
	.dc.b	$9D,$46,$09,$87,$5B,$E7,$9F,$BE,$31,$7C,$DB,$5F,$3E,$F9,$C6,$09
	.dc.b	$9F,$B2,$75,$7C,$E3,$13,$A9,$FB,$98,$78,$33,$F7,$D6,$89,$3A,$C7
	.dc.b	$D5,$87,$38,$C1,$EB,$3F,$C6,$19,$FE,$45,$B8,$7E,$19,$B5,$FE,$3D
	.dc.b	$9F,$F3,$17,$A7,$7D,$13,$49,$5F,$E3,$B3,$F2,$B1,$A7,$F0,$2D,$CF
	.dc.b	$EB,$37,$2F,$CE,$68,$F4,$BF,$CE,$D7,$F8,$EA,$D3,$F9,$D6,$BD,$B0
	.dc.b	$FB,$13,$C2,$3B,$C7,$12,$3F,$4E,$35,$87,$8A,$72,$D9,$D7,$A7,$FA
	.dc.b	$F7,$99,$23,$47,$E6,$6B,$9E,$3F,$BC,$C3,$CB,$97,$C8,$BE,$13,$EE
	.dc.b	$AD,$C7,$AF,$1F,$0C,$F1,$7D,$78,$F8,$71,$EF,$39,$AE,$67,$EE,$AF
	.dc.b	$9B,$F0,$CC,$96,$7C,$D5,$CE,$B9,$47,$49,$F1,$1F,$00,$F1,$98,$75
	.dc.b	$97,$D5,$AE,$53,$9A,$E6,$47,$C9,$1A,$E8,$75,$BE,$B3,$FD,$11,$18
	.dc.b	$6C,$66,$BF,$B8,$47,$FA,$37,$93,$F9,$B7,$65,$E5,$FE,$EF,$5F,$E1
	.dc.b	$B6,$BF,$CE,$63,$9F,$C1,$C8,$FE,$F4,$D7,$FB,$C3,$5F,$F6,$A6,$08
	.dc.b	$C4,$7F,$95,$E3,$0F,$22,$1C,$8F,$17,$C0,$79,$48,$E0,$FF,$05,$67
	.dc.b	$EE,$23,$83,$FB,$C7,$57,$FA,$AF,$6F,$F0,$3B,$7B,$7F,$D3,$F3,$A3
	.dc.b	$FB,$F1,$92,$18,$21,$FD,$FD,$FF,$5B,$3F,$D0,$1B,$3F,$E1,$48,$F9
	.dc.b	$DF,$60,$F8,$CA,$48,$FF,$24,$46,$5F,$DF,$4F,$F9,$8F,$C7,$F4,$79
	.dc.b	$06,$92,$7F,$C5,$5E,$9F,$DE,$23,$82,$1F,$E2,$B7,$2D,$3F,$86,$71
	.dc.b	$FF,$8E,$BE,$43,$F1,$3E,$31,$A4,$9D,$C6,$5B,$19,$7F,$C9,$6F,$E3
	.dc.b	$7E,$BB,$F2,$11,$FE,$D3,$E5,$46,$23,$FC,$0E,$D7,$FD,$4C,$D7,$FB
	.dc.b	$A7,$9D,$F1,$5F,$E4,$4D,$7F,$DB,$CD,$7F,$86,$35,$E3,$B7,$D6,$8C
	.dc.b	$47,$FA,$72,$30,$7E,$8F,$EA,$1F,$5D,$49,$3F,$E7,$D8,$CB,$63,$2F
	.dc.b	$B6,$32,$D8,$FD,$9F,$CF,$BE,$B6,$92,$7E,$8C,$65,$B1,$97,$FD,$F5
	.dc.b	$AA,$75,$18,$6B,$FC,$AF,$C2,$7D,$6B,$E6,$3F,$A4,$4F,$F9,$1A,$30
	.dc.b	$FF,$4B,$9D,$CE,$9B,$1C,$2B,$FD,$A8,$DF,$57,$F3,$F1,$88,$FE,$A8
	.dc.b	$D7,$F9,$E1,$E7,$8D,$91,$E4,$7E,$43,$C4,$7C,$3F,$F1,$86,$C6,$63
	.dc.b	$04,$70,$43,$E2,$C7,$07,$E3,$FA,$AF,$AC,$FE,$46,$D8,$FC,$FB,$5D
	.dc.b	$C6,$88,$E9,$AF,$DD,$FA,$CF,$C3,$11,$FC,$23,$F3,$BD,$B7,$D6,$3E
	.dc.b	$93,$0D,$AF,$EF,$53,$A9,$C9,$23,$33,$B7,$E5,$CE,$DC,$CF,$F3,$EF
	.dc.b	$AC,$7D,$5D,$24,$FE,$F2,$32,$D8,$CB,$F8,$C8,$CB,$63,$FA,$4F,$D5
	.dc.b	$BE,$AE,$92,$7F,$83,$C6,$5B,$19,$7F,$2F,$19,$6C,$7E,$23,$EA,$64
	.dc.b	$7E,$35,$AF,$E2,$BD,$9F,$C6,$91,$FC,$93,$5D,$18,$21,$FE,$B1,$F7
	.dc.b	$9E,$F7,$F3,$F6,$C7,$F2,$C6,$BB,$8D,$79,$52,$34,$D7,$FD,$79,$F0
	.dc.b	$FF,$9D,$7F,$C5,$6D,$7F,$CF,$ED,$F1,$CD,$F8,$99,$CF,$FD,$45,$B5
	.dc.b	$FC,$FB,$81,$F0,$BF,$D5,$5B,$19,$8C,$11,$C1,$0F,$F6,$84,$70,$7F
	.dc.b	$83,$FA,$AC,$BF,$C0,$6D,$8F,$E7,$ED,$77,$1F,$D9,$7E,$7D,$EC,$D2
	.dc.b	$4F,$F5,$48,$CB,$63,$2F,$FB,$86,$32,$D8,$FF,$09,$FD,$47,$F8,$2A
	.dc.b	$3F,$C8,$1E,$0A,$31,$1F,$EB,$18,$C7,$A9,$C0,$FE,$F1,$F5,$3F,$62
	.dc.b	$7F,$C9,$1C,$08,$C4,$7F,$97,$9A,$FF,$AC,$9B,$1B,$E1,$E6,$49,$FE
	.dc.b	$DF,$3C,$2D,$F2,$4F,$95,$FF,$15,$B6,$BF,$B3,$B5,$FE,$6A,$D7,$CB
	.dc.b	$FA,$0F,$F5,$1C,$4D,$27,$95,$FE,$13,$6C,$FF,$80,$1A,$FF,$40,$F3
	.dc.b	$5F,$0B,$FE,$3C,$6C,$7F,$78,$6B,$B8,$FF,$0F,$7D,$4B,$EB,$69,$27
	.dc.b	$FC,$1B,$19,$6C,$65,$FE,$73,$8A,$75,$18,$6B,$FE,$79,$F9,$B7,$D3
	.dc.b	$BB,$7F,$A7,$C6,$1B,$19,$9F,$F4,$D4,$E5,$FE,$9D,$F5,$1F,$D3,$08
	.dc.b	$E0,$87,$F4,$E7,$2D,$39,$CA,$3B,$92,$32,$FF,$58,$F9,$89,$1F,$D7
	.dc.b	$5A,$FF,$AB,$7C,$EF,$F0,$E3,$7D,$5F,$F1,$EC,$62,$3F,$AE,$B5,$FF
	.dc.b	$4D,$35,$1F,$DC,$63,$89,$1F,$F1,$7E,$1F,$87,$7F,$40,$D5,$C6,$8C
	.dc.b	$37,$8F,$FC,$FB,$18,$23,$83,$FC,$F9,$F4,$3F,$84,$AF,$F5,$77,$0A
	.dc.b	$31,$1E,$68,$C7,$0F,$02,$47,$FB,$3E,$32,$D9,$E2,$A3,$2F,$F8,$03
	.dc.b	$E1,$5F,$95,$BE,$1F,$FC,$23,$18,$6C,$66,$7F,$DB,$93,$97,$B6,$DE
	.dc.b	$A2,$38,$21,$FE,$4B,$C1,$1A,$8C,$3F,$E7,$4F,$A5,$FF,$11,$7D,$1F
	.dc.b	$F8,$89,$B3,$FE,$EB,$6B,$FE,$18,$6B,$FE,$D5,$5F,$4B,$F4,$E3,$83
	.dc.b	$5E,$0D,$78,$35,$FF,$11,$FC,$57,$FA,$19,$34,$8F,$F9,$97,$EE,$5F
	.dc.b	$36,$F8,$FF,$C5,$E3,$6B,$E8,$56,$A7,$27,$DB,$FC,$5B,$F8,$C3,$89
	.dc.b	$DD,$F1,$BE,$5C,$0C,$39,$52,$73,$18,$77,$87,$49,$D6,$1F,$89,$AD
	.dc.b	$3F,$8B,$FB,$B7,$99,$E7,$AF,$C8,$C6,$27,$F1,$8D,$8F,$F4,$23,$5F
	.dc.b	$CB,$FE,$21,$FE,$39,$9F,$56,$48,$C4,$F8,$8F,$C9,$57,$3C,$BD,$B3
	.dc.b	$A6,$FB,$1F,$07,$0F,$D0,$4E,$5F,$46,$39,$C3,$F3,$93,$97,$F6,$76
	.dc.b	$F9,$8D,$74,$FC,$77,$86,$6D,$5F,$31,$AB,$E1,$9F,$D3,$C3,$83,$63
	.dc.b	$38,$FA,$F0,$E5,$7C,$71,$EF,$F8,$33,$EF,$E9,$E7,$F0,$DF,$A1,$7F
	.dc.b	$10,$CF,$B5,$6D,$B5,$C1,$FC,$25,$70,$CB,$07,$CB,$D3,$F2,$1D,$7F
	.dc.b	$66,$3F,$BA,$57,$EE,$CF,$F6,$39,$5F,$F4,$AB,$5F,$EF,$BC,$BE,$37
	.dc.b	$A3,$76,$FC,$D5,$F6,$AF,$89,$39,$7F,$C0,$5F,$86,$3E,$CF,$C5,$87
	.dc.b	$F2,$48,$C4,$FF,$B6,$1B,$1F,$EA,$66,$BF,$E0,$E6,$BD,$27,$9D,$A7
	.dc.b	$33,$D6,$73,$F0,$9F,$E7,$9B,$FB,$0D,$9F,$8D,$5E,$0C,$7D,$73,$97
	.dc.b	$DB,$C2,$FA,$0E,$ED,$C3,$FB,$C4,$60,$8E,$09,$87,$F8,$06,$3C,$FF
	.dc.b	$84,$3D,$0F,$C8,$B6,$BF,$19,$7D,$6F,$ED,$4E,$5F,$B9,$E7,$3E,$CF
	.dc.b	$E0,$B6,$FF,$C3,$71,$82,$38,$24,$FF,$87,$A3,$E4,$FA,$5F,$E2,$69
	.dc.b	$FC,$E3,$67,$BC,$FE,$9A,$31,$9F,$E0,$3E,$A1,$F4,$1F,$49,$B8,$7F
	.dc.b	$B2,$30,$FE,$C3,$18,$7F,$DC,$2D,$7F,$C9,$2D,$7F,$C9,$0D,$78,$CF
	.dc.b	$53,$EF,$FD,$E7,$EB,$AF,$F2,$2D,$BE,$33,$FB,$C8,$C5,$7F,$72,$F6
	.dc.b	$5F,$35,$F8,$86,$DF,$F9,$AA,$30,$47,$04,$9F,$F3,$9C,$7E,$43,$F0
	.dc.b	$47,$CA,$FF,$0A,$6D,$7F,$05,$BE,$B7,$FB,$39,$CB,$FC,$7F,$A5,$EF
	.dc.b	$FF,$0E,$BF,$91,$18,$EF,$FD,$3B,$63,$33,$BA,$FF,$1E,$B5,$FF,$35
	.dc.b	$B5,$FF,$0C,$CF,$17,$75,$FA,$8F,$99,$7C,$B7,$F2,$26,$E1,$FC,$33
	.dc.b	$0E,$DB,$FE,$FA,$73,$87,$F0,$DF,$C7,$7F,$7C,$FF,$87,$9B,$97,$FA
	.dc.b	$A3,$8F,$F9,$6C,$6B,$B0,$F9,$6B,$86,$9F,$BA,$8C,$4E,$A3,$F7,$BF
	.dc.b	$80,$FF,$21,$AF,$E3,$8B,$D5,$FE,$72,$E7,$3F,$E2,$6E,$75,$FE,$D1
	.dc.b	$F8,$27,$F8,$C7,$F0,$6F,$E2,$2D,$8F,$CB,$37,$0F,$F1,$F9,$FE,$40
	.dc.b	$70,$E1,$AE,$7D,$65,$74,$6D,$74,$6C,$F4,$3F,$C8,$7C,$3F,$BD,$C7
	.dc.b	$F9,$4B,$6B,$A5,$FF,$54,$8C,$57,$F9,$2B,$EA,$FF,$B2,$57,$FA,$B5
	.dc.b	$B1,$9E,$8F,$EB,$71,$8D,$8D,$24,$6B,$CB,$FE,$C3,$F5,$6F,$91,$BE
	.dc.b	$0E,$1B,$97,$F5,$F9,$D3,$67,$4D,$3F,$C2,$C3,$F0,$47,$C5,$FF,$40
	.dc.b	$8C,$36,$33,$1F,$C1,$CC,$95,$CE,$9F,$26,$32,$D8,$CB,$F6,$23,$2F
	.dc.b	$FB,$53,$E4,$DF,$23,$FD,$F9,$FF,$71,$CF,$F7,$97,$2D,$F8,$8F,$F0
	.dc.b	$1E,$8F,$A1,$A7,$F8,$13,$F4,$37,$C8,$E2,$FF,$C0,$55,$B6,$D6,$E7
	.dc.b	$59,$7F,$73,$F9,$DF,$E2,$7C,$65,$F7,$BF,$E1,$27,$26,$30,$FF,$07
	.dc.b	$38,$8D,$7A,$BB,$7F,$C3,$DD,$5F,$E8,$DC,$7F,$24,$4B,$FC,$93,$71
	.dc.b	$FC,$94,$EE,$72,$48,$FC,$B6,$2F,$9A,$FD,$54,$7F,$82,$DB,$D5,$FE
	.dc.b	$59,$BE,$53,$9E,$DF,$F1,$5F,$D3,$FF,$AE,$71,$FF,$13,$3F,$E7,$C6
	.dc.b	$FC,$3E,$BE,$07,$6D,$8C,$9C,$23,$9F,$D9,$C8,$FF,$24,$FE,$88,$FA
	.dc.b	$BC,$65,$FE,$1F,$C5,$78,$D3,$FC,$CB,$7B,$C3,$C7,$3F,$22,$5F,$0D
	.dc.b	$FC,$11,$B3,$FC,$11,$B8,$7C,$86,$BF,$C0,$DA,$FF,$07,$6A,$D7,$FC
	.dc.b	$E5,$5D,$4F,$F0,$BF,$D2,$7F,$8B,$EB,$FC,$52,$DB,$E5,$3F,$E5,$48
	.dc.b	$C5,$FF,$B2,$BE,$A3,$C7,$7F,$E3,$06,$DF,$FB,$5B,$85,$1F,$DD,$A3
	.dc.b	$0E,$AB,$96,$1F,$E6,$08,$C5,$70,$F5,$31,$39,$F9,$29,$87,$F9,$7E
	.dc.b	$DF,$61,$8C,$7F,$C9,$D0,$AC,$65,$E5,$AB,$7E,$2F,$07,$FC,$6F,$7F
	.dc.b	$E1,$BC,$3F,$97,$E7,$E8,$9C,$9A,$F0,$FC,$C7,$9D,$B7,$E3,$8F,$F3
	.dc.b	$BF,$C6,$FE,$D2,$30,$7F,$A3,$9C,$5F,$9A,$77,$D1,$CA,$3F,$CE,$D3
	.dc.b	$AC,$FE,$67,$CA,$F9,$89,$87,$F8,$6F,$2F,$9B,$1E,$AF,$A3,$F9,$47
	.dc.b	$CB,$6C,$FF,$A7,$BF,$1C,$FE,$1F,$8F,$F4,$1F,$71,$18,$8F,$F7,$DB
	.dc.b	$5F,$F7,$D3,$5F,$C0,$7C,$CF,$D7,$3F,$D2,$B1,$D3,$2F,$C1,$46,$6F
	.dc.b	$EB,$3A,$F1,$31,$7F,$5A,$2D,$7F,$B2,$2F,$95,$FE,$1A,$7F,$09,$F3
	.dc.b	$5F,$58,$91,$F0,$78,$7F,$CD,$18,$7D,$87,$FD,$D7,$1F,$85,$F8,$EE
	.dc.b	$2D,$3E,$17,$67,$FB,$7F,$E1,$1D,$BE,$E8,$C3,$F0,$3E,$16,$E2,$DF
	.dc.b	$E3,$D6,$1B,$82,$7F,$DE,$37,$CE,$31,$87,$69,$E5,$AF,$FA,$76,$31
	.dc.b	$7C,$A7,$89,$AC,$3B,$57,$F0,$78,$CC,$FD,$8C,$73,$5E,$3F,$8C,$7E
	.dc.b	$05,$4C,$72,$DC,$FC,$7A,$7F,$BA,$E7,$19,$74,$BF,$F7,$96,$9E,$47
	.dc.b	$9D,$F0,$9F,$F7,$1D,$70,$AD,$61,$70,$E9,$7B,$CD,$8F,$2C,$7B,$5E
	.dc.b	$DF,$99,$F9,$B3,$EC,$BE,$67,$D5,$7F,$1E,$CF,$E2,$CE,$72,$E9,$39
	.dc.b	$CF,$E8,$66,$AD,$63,$F1,$AB,$4F,$6C,$7A,$F7,$7D,$9A,$DE,$5F,$4A
	.dc.b	$F7,$87,$3B,$DE,$5F,$07,$17,$D3,$39,$DF,$13,$98,$C3,$F1,$FD,$E7
	.dc.b	$D8,$7D,$9D,$3F,$05,$87,$2C,$38,$BF,$80,$31,$7C,$E3,$33,$F8,$9B
	.dc.b	$E7,$C7,$F9,$4B,$D9,$C7,$1F,$0B,$FE,$9F,$6B,$FE,$3F,$6B,$FE,$4C
	.dc.b	$6B,$F9,$95,$C9,$FD,$4A,$39,$35,$E4,$D7,$93,$5E,$4F,$1B,$F7,$CF
	.dc.b	$4E,$0F,$8D,$39,$E1,$FF,$41,$FC,$1F,$F8,$77,$0C,$D4,$7D,$95,$B5
	.dc.b	$92,$B8,$47,$E6,$7E,$2B,$FC,$CF,$97,$E6,$7E,$52,$E5,$F8,$6E,$64
	.dc.b	$62,$FF,$79,$1F,$EE,$B7,$F3,$0D,$7F,$49,$1B,$E3,$FC,$DB,$67,$57
	.dc.b	$FA,$7C,$31,$32,$F9,$C7,$1F,$BF,$75,$52,$4F,$E9,$E1,$D9,$88,$FD
	.dc.b	$FC,$70,$75,$3C,$E7,$58,$7F,$0A,$AD,$57,$0A,$D3,$C0,$86,$4F,$05
	.dc.b	$7B,$7E,$4B,$F2,$71,$FC,$39,$B7,$FC,$59,$BA,$7E,$C1,$FF,$45,$46
	.dc.b	$79,$39,$12,$BF,$81,$4E,$AB,$84,$E6,$7B,$3E,$17,$EA,$78,$5F,$E0
	.dc.b	$8F,$3F,$FD,$43,$E0,$FF,$83,$5B,$D3,$FE,$0A,$6D,$7F,$84,$FE,$42
	.dc.b	$DF,$F8,$41,$FF,$80,$11,$93,$FA,$43,$88,$D4,$7F,$4B,$FA,$BE,$3E
	.dc.b	$11,$A7,$6E,$46,$AF,$5F,$37,$FE,$13,$AA,$B4,$69,$F1,$EB,$31,$FE
	.dc.b	$27,$AB,$39,$AF,$E9,$53,$C1,$FE,$A1,$7D,$DF,$F5,$C1,$8B,$F6,$E1
	.dc.b	$E4,$6C,$EE,$7F,$CD,$38,$BE,$B7,$DA,$48,$FF,$15,$7E,$29,$F0,$0F
	.dc.b	$89,$26,$92,$3F,$B8,$70,$7C,$3F,$C3,$BF,$C5,$D1,$E8,$D3,$F9,$9C
	.dc.b	$62,$3F,$C6,$73,$F3,$0E,$1E,$74,$8F,$F1,$D4,$65,$A4,$3F,$C7,$7F
	.dc.b	$07,$F5,$0F,$29,$1F,$F0,$13,$F0,$BE,$C3,$FD,$91,$1E,$33,$FB,$C3
	.dc.b	$87,$EA,$61,$FD,$22,$31,$BF,$CF,$DD,$0F,$3F,$F0,$0E,$AB,$87,$9F
	.dc.b	$83,$33,$F9,$0F,$4B,$FD,$05,$8F,$87,$91,$18,$8F,$08,$8C,$F8,$9C
	.dc.b	$7E,$57,$93,$27,$95,$71,$D3,$FC,$EB,$5F,$9E,$D9,$F1,$74,$97,$F0
	.dc.b	$47,$E4,$5F,$ED,$07,$0D,$7F,$16,$D7,$FC,$34,$7F,$86,$DC,$38,$6B
	.dc.b	$9F,$82,$FD,$6B,$D9,$B1,$A8,$C1,$F9,$8E,$0F,$9B,$E9,$7C,$C2,$47
	.dc.b	$D7,$87,$E8,$BF,$EF,$68,$C7,$12,$B7,$1F,$2A,$B9,$BC,$6B,$83,$FE
	.dc.b	$C5,$F0,$3F,$81,$FB,$AF,$52,$98,$89,$3F,$5E,$74,$D9,$CB,$FE,$CF
	.dc.b	$EB,$E6,$47,$E7,$7B,$BF,$EB,$E6,$FB,$7F,$12,$31,$1F,$B8,$6B,$F3
	.dc.b	$83,$87,$E8,$B5,$FF,$26,$35,$DC,$75,$3F,$8D,$79,$DF,$03,$8F,$F8
	.dc.b	$47,$37,$F8,$5B,$EA,$7F,$91,$C9,$BF,$0D,$30,$FC,$B4,$E5,$B3,$97
	.dc.b	$F2,$F4,$F8,$57,$D3,$7F,$2E,$6C,$FE,$7D,$AF,$D5,$F8,$6F,$CC,$77
	.dc.b	$7F,$B0,$E7,$39,$F9,$9B,$19,$BF,$E5,$D5,$C2,$FF,$73,$EE,$7F,$3C
	.dc.b	$FC,$03,$DD,$7F,$EC,$D6,$CE,$8F,$A7,$E8,$7D,$12,$3F,$D0,$B8,$3F
	.dc.b	$C0,$2D,$8C,$BF,$55,$AF,$FB,$0B,$A9,$F5,$5F,$D3,$17,$E3,$34,$C4
	.dc.b	$49,$FE,$01,$3A,$6C,$E5,$F9,$7F,$5C,$F8,$FF,$EA,$ED,$8F,$E1,$0D
	.dc.b	$77,$EA,$EA,$64,$69,$E3,$1F,$D8,$3C,$7F,$C5,$CF,$F1,$16,$BF,$69
	.dc.b	$BC,$5F,$D3,$63,$F2,$3C,$9F,$DA,$9A,$7C,$77,$F3,$44,$70,$43,$F3
	.dc.b	$8E,$5F,$CE,$72,$7F,$98,$38,$FF,$B9,$FB,$6F,$A1,$A6,$23,$FB,$BF
	.dc.b	$CA,$7F,$A6,$2F,$FD,$B8,$D7,$9B,$5F,$F7,$AB,$5F,$F5,$85,$BE,$57
	.dc.b	$FC,$87,$97,$E8,$9B,$CB,$A6,$D7,$F9,$2B,$EB,$7F,$65,$5F,$A8,$FC
	.dc.b	$41,$F1,$BF,$E0,$A6,$D7,$F3,$B6,$BF,$CE,$BD,$E7,$CC,$72,$CB,$FD
	.dc.b	$85,$18,$9F,$F0,$8B,$5F,$EF,$2D,$7F,$81,$35,$FE,$62,$D7,$9C,$FF
	.dc.b	$2E,$8C,$47,$F9,$8A,$3F,$98,$E6,$F9,$97,$E1,$1F,$E2,$26,$D7,$F5
	.dc.b	$36,$BF,$BC,$F8,$1F,$F2,$7B,$FD,$5D,$AF,$DA,$6B,$FD,$12,$BF,$CE
	.dc.b	$1E,$3F,$F8,$A6,$73,$3F,$49,$AF,$F5,$D6,$CF,$08,$FF,$0C,$B5,$FC
	.dc.b	$3F,$CE,$FE,$4F,$6F,$E3,$2D,$D3,$EE,$B6,$7F,$B4,$7C,$FF,$F7,$36
	.dc.b	$FA,$BF,$E3,$B8,$FE,$FD,$1E,$76,$BF,$86,$6B,$C0,$F1,$B8,$C0,$D3
	.dc.b	$1F,$88,$3E,$CB,$FC,$9A,$DA,$FE,$FC,$D7,$F9,$3F,$C8,$FE,$67,$3F
	.dc.b	$E6,$6F,$E5,$E3,$FC,$00,$DC,$0D,$27,$C7,$7F,$95,$5B,$3F,$21,$E1
	.dc.b	$3F,$C1,$6B,$F9,$FB,$74,$FD,$1B,$6B,$FC,$1D,$E8,$7E,$02,$BF,$A2
	.dc.b	$BF,$82,$8F,$B8,$D7,$F5,$6D,$78,$F0,$FF,$99,$E3,$14,$69,$23,$FC
	.dc.b	$D9,$F4,$7F,$EF,$BA,$FF,$11,$35,$FE,$96,$DB,$FE,$AF,$F0,$BF,$E1
	.dc.b	$A9,$FF,$14,$35,$FE,$16,$D7,$F2,$4D,$7F,$80,$B5,$FF,$16,$37,$03
	.dc.b	$49,$EB,$7F,$A0,$7E,$4F,$FC,$FD,$5F,$E3,$76,$BF,$D8,$1B,$7F,$DB
	.dc.b	$3F,$15,$7D,$07,$FA,$39,$B5,$F9,$C6,$BF,$C8,$9A,$FE,$A5,$AF,$F1
	.dc.b	$8E,$9F,$E4,$8F,$F9,$1D,$AF,$F7,$06,$D7,$FA,$0D,$7B,$AF,$94,$FF
	.dc.b	$4F,$C6,$B8,$3A,$78,$58,$3F,$D4,$B1,$C2,$7F,$D4,$BF,$3F,$F9,$5A
	.dc.b	$FE,$52,$DC,$7F,$CA,$8D,$E2,$FF,$4F,$F9,$DC,$39,$65,$7F,$97,$6B
	.dc.b	$5F,$0B,$FD,$33,$04,$FE,$17,$E9,$DF,$01,$FE,$A9,$6C,$7F,$8F,$DA
	.dc.b	$FF,$8E,$FC,$BF,$F6,$8F,$07,$F9,$CE,$72,$DA,$E4,$7F,$AC,$E7,$77
	.dc.b	$FE,$FD,$F8,$BF,$F5,$1C,$7F,$B6,$B0,$FF,$2D,$EA,$8C,$46,$CF,$F2
	.dc.b	$A7,$D4,$7A,$5F,$CE,$36,$4D,$CD,$F6,$DF,$EC,$38,$D1,$37,$FF,$63
	.dc.b	$4E,$AF,$FC,$E5,$F5,$4F,$8E,$FF,$62,$47,$FA,$EA,$31,$B7,$FB,$59
	.dc.b	$B1,$99,$DB,$FE,$47,$87,$C7,$B8,$78,$36,$7A,$35,$E0,$D7,$FD,$93
	.dc.b	$D7,$FE,$78,$7F,$4A,$7F,$9E,$9C,$1F,$13,$D7,$FF,$69,$63,$FE,$A5
	.dc.b	$D1,$18,$E1,$FB,$FB,$FF,$B9,$3E,$63,$18,$FC,$AA,$B3,$F2,$9B,$7C
	.dc.b	$AF,$FD,$01,$B3,$0C,$D7,$0E,$1F,$F7,$75,$7F,$A2,$5B,$8F,$F9,$A7
	.dc.b	$1F,$B9,$F1,$DF,$E1,$B8,$F1,$CE,$9F,$C0,$F2,$FF,$84,$6F,$F2,$1D
	.dc.b	$3D,$6F,$CB,$39,$AF,$F4,$63,$75,$FC,$95,$6E,$F8,$9F,$EB,$D8,$7C
	.dc.b	$EF,$C0,$F0,$3F,$E0,$03,$F0,$36,$FE,$0F,$5B,$8D,$57,$08,$C1,$F0
	.dc.b	$BB,$B3,$3C,$FE,$A5,$F2,$54,$F9,$BD,$CF,$AA,$FD,$7E,$1F,$EC,$FA
	.dc.b	$DB,$6B,$83,$FE,$E3,$AE,$1B,$60,$FF,$3E,$C3,$E5,$BE,$54,$7F,$B5
	.dc.b	$63,$07,$F7,$82,$BF,$E3,$E6,$BF,$CD,$0F,$85,$CD,$E4,$7A,$CE,$1F
	.dc.b	$8D,$E2,$FF,$04,$AF,$F0,$1B,$63,$FC,$92,$D7,$FC,$20,$D7,$FC,$7E
	.dc.b	$D7,$FD,$D4,$DD,$CD,$31,$DC,$F3,$BE,$B3,$6F,$FA,$AF,$81,$89,$FB
	.dc.b	$5C,$EF,$96,$F8,$78,$FF,$5F,$6C,$66,$FF,$C3,$6D,$CB,$FC,$C2,$D7
	.dc.b	$FB,$6C,$F0,$24,$FF,$88,$BE,$A9,$F2,$3F,$89,$6C,$7F,$2A,$E5,$FE
	.dc.b	$EF,$59,$AF,$C7,$79,$7F,$F0,$02,$BF,$C0,$9E,$57,$9A,$31,$8F,$F9
	.dc.b	$C9,$AF,$F8,$E5,$AF,$F9,$A1,$AF,$F8,$52,$39,$1F,$03,$CE,$FF,$0A
	.dc.b	$D7,$8C,$FF,$15,$57,$8A,$7F,$41,$3A,$AF,$F6,$3F,$D5,$3E,$9B,$E4
	.dc.b	$B7,$0F,$F2,$0B,$67,$FD,$0E,$D7,$FC,$9C,$D7,$FC,$FA,$D5,$B2,$30
	.dc.b	$DA,$FA,$FE,$97,$EB,$6B,$FD,$B2,$DD,$7F,$DC,$B7,$E1,$7F,$C9,$FD
	.dc.b	$5F,$B9,$F4,$DF,$55,$F8,$06,$E7,$FE,$58,$6E,$3F,$E6,$59,$D1,$27
	.dc.b	$4D,$8F,$F5,$1F,$AA,$FA,$AF,$E0,$CD,$AF,$F3,$85,$62,$7F,$7B,$EB
	.dc.b	$FF,$08,$BF,$F4,$2B,$5F,$F4,$2B,$5F,$F4,$2F,$81,$FB,$8E,$DF,$F4
	.dc.b	$9B,$67,$53,$FE,$87,$8C,$1F,$9C,$F6,$DF,$01,$FC,$65,$B9,$7F,$07
	.dc.b	$BF,$16,$DF,$EA,$BA,$FE,$3D,$EE,$FE,$F6,$15,$FE,$B2,$8C,$3F,$EA
	.dc.b	$2F,$77,$F2,$73,$AC,$BE,$CD,$69,$B9,$7E,$BA,$3C,$FC,$86,$98,$FA
	.dc.b	$AF,$E3,$B4,$BE,$94,$9F,$ED,$D8,$FE,$5E,$BF,$0F,$1C,$2B,$6E,$A3
	.dc.b	$94,$FF,$86,$3D,$8F,$ED,$9F,$05,$FC,$89,$B1,$FE,$1C,$99,$18,$70
	.dc.b	$1C,$3A,$F4,$7F,$DF,$D3,$A8,$D4,$61,$E3,$3C,$1F,$14,$65,$FF,$58
	.dc.b	$6B,$FC,$37,$1F,$7B,$6F,$78,$7F,$51,$8C,$57,$E9,$BE,$AB,$FB,$3D
	.dc.b	$FC,$66,$DF,$FB,$A1,$FD,$13,$7C,$DF,$EB,$51,$AA,$E5,$18,$EA,$CB
	.dc.b	$FD,$7F,$EA,$9E,$FD,$F0,$30,$DC,$7E,$97,$02,$B7,$18,$7F,$8F,$E3
	.dc.b	$F4,$8F,$DE,$BA,$71,$3C,$6F,$FC,$1B,$4F,$AA,$7C,$7F,$F5,$88,$C3
	.dc.b	$63,$31,$F4,$4C,$95,$CD,$FD,$A8,$CB,$63,$2F,$E0,$63,$2F,$FB,$8F
	.dc.b	$E3,$9F,$15,$FE,$01,$5F,$43,$63,$0F,$F8,$17,$57,$D5,$53,$FC,$4B
	.dc.b	$F2,$1F,$E7,$BC,$7E,$3C,$E9,$B3,$B7,$F3,$9E,$52,$B8,$65,$F1,$AF
	.dc.b	$83,$F2,$3E,$AB,$F9,$05,$FF,$9D,$37,$43,$E2,$7F,$C3,$0E,$AB,$F9
	.dc.b	$6B,$9A,$FB,$2E,$6B,$8D,$70,$F1,$78,$BF,$E4,$0E,$AF,$F2,$A7,$C1
	.dc.b	$FE,$A7,$AF,$CB,$97,$E8,$E3,$04,$CB,$F4,$95,$C1,$B5,$C2,$BF,$47
	.dc.b	$5F,$DC,$BE,$5B,$F7,$1D,$5F,$C0,$1B,$7A,$3F,$1E,$E2,$BF,$A3,$FC
	.dc.b	$DB,$E0,$FF,$C6,$CB,$EA,$AC,$61,$B7,$F9,$89,$D3,$67,$4E,$7E,$5B
	.dc.b	$99,$FE,$5E,$FA,$AF,$F5,$AC,$64,$98,$7E,$6A,$D5,$BF,$89,$E3,$5F
	.dc.b	$C5,$F6,$FE,$39,$F4,$BF,$C4,$75,$34,$93,$F2,$23,$2D,$8C,$E3,$FC
	.dc.b	$32,$72,$FF,$16,$E2,$7C,$FF,$F8,$EA,$7F,$94,$B6,$FF,$91,$47,$03
	.dc.b	$3E,$DF,$F9,$77,$E8,$F1,$3F,$8B,$6E,$1F,$DD,$1F,$F0,$15,$7F,$34
	.dc.b	$73,$81,$A4,$F5,$BA,$65,$D7,$E8,$66,$3F,$C2,$29,$F5,$58,$CB,$F8
	.dc.b	$1C,$2B,$19,$77,$AB,$7E,$5F,$0F,$FC,$73,$7F,$C3,$B0,$FE,$75,$9F
	.dc.b	$8E,$72,$6B,$C5,$F7,$1E,$F6,$DF,$8E,$3F,$D3,$7F,$11,$FB,$0C,$BD
	.dc.b	$D1,$8E,$8D,$E3,$FC,$6F,$BF,$FD,$39,$18,$AF,$EC,$51,$8C,$3D,$DF
	.dc.b	$41,$E6,$23,$FD,$1F,$3F,$91,$1E,$AF,$9C,$FA,$47,$D3,$6C,$FF,$04
	.dc.b	$FA,$AF,$E2,$2F,$F4,$FB,$46,$32,$FF,$54,$35,$FF,$72,$B5,$FB,$9F
	.dc.b	$41,$F7,$CF,$F5,$CC,$71,$CB,$EE,$46,$6F,$EB,$3A,$F1,$71,$7F,$5A
	.dc.b	$2D,$7C,$0B,$E5,$7D,$8F,$BD,$F0,$1F,$58,$8F,$C1,$E1,$FF,$3B,$61
	.dc.b	$F9,$97,$FD,$C3,$1F,$85,$F8,$EF,$CA,$57,$E5,$5F,$3F,$57,$F8,$3F
	.dc.b	$E1,$5D,$BF,$ED,$6C,$BF,$6B,$92,$72,$B6,$F8,$FD,$5C,$6F,$86,$16
	.dc.b	$B5,$87,$2C,$38,$E1,$F3,$AB,$4F,$2C,$D7,$DF,$0E,$71,$8B,$E3,$87
	.dc.b	$39,$CD,$6C,$EF,$8F,$49,$D4,$79,$FE,$31,$F2,$5E,$06,$E3,$D7,$0F
	.dc.b	$F7,$1F,$3E,$6A,$4E,$6F,$94,$71,$9C,$C7,$87,$36,$1D,$23,$0D,$CF
	.dc.b	$E0,$BF,$EF,$9C,$3A,$57,$0C,$39,$D6,$9E,$D3,$AA,$E5,$8F,$C6,$B7
	.dc.b	$D5,$3E,$A7,$E5,$FD,$57,$F1,$AC,$BE,$7C,$E7,$1E,$F3,$9C,$39,$4E
	.dc.b	$73,$F5,$4E,$5F,$BD,$19,$BF,$B3,$39,$C7,$C3,$8F,$83,$BB,$ED,$DE
	.dc.b	$F1,$F3,$5F,$3C,$5F,$45,$FE,$00,$F8,$AF,$99,$18,$E8,$47,$CF,$F7
	.dc.b	$DF,$3B,$F6,$AB,$53,$C2,$F9,$47,$E0,$4C,$9F,$83,$AE,$3C,$BF,$93
	.dc.b	$8C,$EF,$F9,$A7,$67,$F9,$52,$B9,$B7,$1F,$1B,$FE,$53,$8C,$BF,$E5
	.dc.b	$D5,$CB,$FD,$42,$34,$D7,$A3,$5E,$0D,$7F,$3F,$1C,$B6,$F6,$4F,$3B
	.dc.b	$F3,$4E,$6B,$D9,$F0,$9F,$E2,$47,$77,$1F,$85,$3B,$AF,$DB,$E4,$BE
	.dc.b	$15,$F9,$4F,$89,$FF,$3B,$57,$E5,$3E,$4E,$F0,$7E,$37,$95,$18,$BF
	.dc.b	$D5,$5D,$E4,$FC,$FB,$6B,$73,$F9,$7A,$C4,$C8,$C1,$0F,$11,$FA,$2F
	.dc.b	$AA,$75,$75,$6C,$66,$31,$D7,$96,$46,$4E,$51,$87,$F6,$93,$C9,$B3
	.dc.b	$C9,$E0,$60,$B9,$FF,$16,$95,$7B,$FE,$4B,$F0,$75,$FB,$F6,$D7,$F1
	.dc.b	$C6,$E9,$FB,$57,$FD,$3F,$19,$E4,$E4,$4A,$FE,$03,$3A,$AE,$13,$99
	.dc.b	$EC,$F8,$5F,$A5,$E1,$7F,$81,$7E,$5F,$FD,$5D,$E0,$7F,$82,$5B,$D9
	.dc.b	$FE,$0F,$6D,$7F,$83,$FE,$46,$DF,$F8,$35,$FF,$57,$46,$4F,$E9,$4E
	.dc.b	$23,$51,$FD,$37,$EA,$B8,$F8,$46,$9D,$B9,$1A,$BD,$7C,$DF,$F8,$56
	.dc.b	$AB,$FD,$FE,$74,$E4,$87,$F8,$62,$3A,$46,$A3,$84,$FA,$2B,$BB,$67
	.dc.b	$53,$FE,$D1,$9C,$BF,$D9,$7E,$97,$D2,$47,$FC,$51,$F5,$5E,$C3,$E0
	.dc.b	$62,$69,$23,$FB,$77,$07,$BF,$EA,$BF,$C5,$91,$FE,$8E,$6B,$FE,$30
	.dc.b	$8F,$8E,$6F,$CE,$91,$FE,$37,$8C,$B4,$87,$F8,$E3,$E1,$3E,$79,$E4
	.dc.b	$23,$FD,$FD,$F8,$5F,$53,$FE,$CC,$8F,$F4,$FB,$5F,$F1,$63,$78,$FF
	.dc.b	$C5,$6B,$CF,$FB,$0E,$79,$3F,$A6,$36,$BF,$BF,$37,$63,$C6,$3F,$DA
	.dc.b	$FE,$93,$C4,$FC,$6B,$63,$31,$8F,$95,$CB,$23,$27,$28,$FC,$AE,$8F
	.dc.b	$94,$A4,$9F,$F0,$2E,$8C,$FF,$C0,$F0,$FA,$AF,$F6,$6B,$A8,$CC,$62
	.dc.b	$36,$7F,$83,$9C,$38,$6B,$C3,$E0,$BF,$4A,$F5,$6C,$6A,$39,$46,$0F
	.dc.b	$A7,$E8,$FC,$C2,$47,$FB,$4B,$EA,$BF,$DE,$2F,$D6,$8C,$BF,$59,$FD
	.dc.b	$54,$6A,$3F,$DF,$7E,$EF,$F1,$27,$FD,$6F,$E9,$FE,$EA,$0D,$F0,$64
	.dc.b	$E8,$87,$27,$87,$4C,$8F,$F6,$0B,$FC,$19,$FF,$60,$7B,$57,$DC,$FF
	.dc.b	$58,$C7,$F0,$66,$BF,$28,$D9,$98,$F8,$F1,$A8,$F9,$ED,$7E,$BB,$5F
	.dc.b	$F1,$FB,$7B,$FF,$C0,$BF,$0D,$FC,$5B,$CC,$7D,$35,$31,$12,$FE,$F0
	.dc.b	$AF,$F0,$6C,$EE,$33,$1D,$1B,$F0,$CF,$A6,$FE,$46,$D8,$CC,$7F,$1B
	.dc.b	$F8,$6F,$CD,$43,$97,$F2,$0F,$7A,$E0,$E7,$97,$FC,$B3,$82,$73,$5C
	.dc.b	$23,$18,$75,$E9,$FD,$8F,$B7,$FC,$D3,$EA,$BF,$01,$7B,$6E,$3F,$B4
	.dc.b	$9F,$A9,$E7,$FC,$E2,$67,$F5,$5A,$FD,$5E,$0F,$EE,$6D,$8C,$BF,$EB
	.dc.b	$CE,$87,$D5,$7F,$44,$5F,$55,$4C,$44,$9F,$97,$3A,$6C,$E5,$F8,$47
	.dc.b	$D3,$FA,$A7,$C8,$7F,$50,$6C,$7D,$46,$BC,$38,$11,$FE,$6A,$8D,$4F
	.dc.b	$12,$3C,$7D,$4F,$EB,$7E,$33,$4A,$49,$FC,$29,$9E,$5F,$CE,$1F,$96
	.dc.b	$23,$F9,$D6,$BF,$25,$AF,$E6,$9F,$EC,$9E,$D7,$DC,$E0,$FB,$EB,$E8
	.dc.b	$7F,$B6,$B6,$BF,$22,$ED,$FC,$B7,$C9,$FF,$A4,$9F,$F6,$C3,$5F,$E2
	.dc.b	$BA,$A3,$11,$FE,$F2,$6B,$F8,$EF,$97,$FF,$1D,$F2,$FE,$35,$B8,$E7
	.dc.b	$E3,$FF,$C0,$3E,$DB,$E5,$BF,$90,$DF,$E4,$FE,$A9,$F1,$FF,$E0,$46
	.dc.b	$E5,$FA,$86,$BF,$72,$DF,$31,$CB,$2F,$F5,$BC,$62,$7F,$C1,$6D,$7F
	.dc.b	$83,$35,$FE,$62,$D7,$F9,$CB,$5E,$73,$FC,$C2,$31,$06,$EA,$F9,$B7
	.dc.b	$D4,$7F,$87,$1B,$5F,$C0,$1A,$FE,$37,$E0,$7F,$95,$37,$C7,$FF,$0F
	.dc.b	$46,$23,$FA,$5D,$FF,$99,$E7,$24,$8F,$E4,$11,$96,$CF,$17,$F8,$0B
	.dc.b	$5F,$E2,$9F,$3D,$FC,$21,$FC,$F3,$76,$FE,$08,$DD,$3F,$84,$FC,$FF
	.dc.b	$F7,$1C,$0F,$F7,$BB,$F2,$B5,$DF,$C7,$7F,$8D,$E3,$11,$FC,$8D,$B3
	.dc.b	$CA,$74,$49,$F5,$FD,$4B,$EA,$BF,$C8,$AD,$8F,$E5,$2D,$7F,$6D,$F2
	.dc.b	$5F,$97,$CF,$F8,$E3,$F9,$48,$FE,$64,$DC,$0D,$27,$C7,$7F,$93,$5B
	.dc.b	$3F,$31,$E1,$3F,$C1,$2B,$F9,$43,$73,$FE,$5A,$DA,$FE,$81,$EC,$FD
	.dc.b	$EA,$FE,$5C,$FE,$02,$3E,$DB,$5F,$E6,$AD,$78,$F0,$FF,$97,$E3,$14
	.dc.b	$69,$23,$FC,$C9,$F4,$7F,$EE,$FA,$FE,$9A,$D7,$F8,$D3,$6F,$FA,$07
	.dc.b	$C3,$7F,$86,$E7,$FA,$9B,$5F,$DF,$B5,$FC,$8B,$5F,$E9,$6D,$7F,$AF
	.dc.b	$37,$03,$49,$EB,$FF,$9E,$7E,$53,$FC,$ED,$5F,$D8,$9A,$FF,$2F,$6D
	.dc.b	$FF,$55,$FA,$A7,$D0,$7F,$A1,$9B,$5F,$A0,$6B,$FC,$59,$AF,$EA,$1A
	.dc.b	$FF,$62,$E8,$FE,$44,$FF,$3D,$6D,$7F,$63,$6E,$DF,$C2,$BC,$E7,$CA
	.dc.b	$FF,$A5,$A7,$37,$FD,$F5,$BA,$FF,$03,$F0,$FF,$8E,$DF,$0A,$FE,$FF
	.dc.b	$F3,$3F,$95,$C7,$FB,$9B,$63,$FC,$0C,$DE,$2F,$F4,$BF,$8D,$FE,$A4
	.dc.b	$8D,$62,$69,$BF,$86,$FE,$9D,$7F,$E1,$0B,$7D,$3B,$E0,$3F,$D4,$4D
	.dc.b	$9D,$35,$FE,$CB,$2F,$2F,$FD,$97,$C1,$FE,$0C,$8E,$77,$C8,$95,$DE
	.dc.b	$B7,$A7,$FB,$D7,$E3,$3F,$D4,$51,$FE,$D4,$9F,$F5,$03,$5D,$1F,$E3
	.dc.b	$7B,$7D,$47,$AA,$3C,$6D,$C7,$C4,$DC,$3F,$DB,$5F,$55,$FE,$BD,$8C
	.dc.b	$57,$E9,$5B,$7D,$37,$FF,$5B,$D7,$F9,$2F,$EA,$9F,$1F,$FE,$29,$C3
	.dc.b	$FC,$91,$18,$CB,$FD,$A2,$D8,$CC,$ED,$FF,$3B,$C3,$E4,$7E,$A4,$72
	.dc.b	$6B,$A6,$BF,$E8,$48,$E1,$C3,$FE,$7A,$8F,$D3,$9F,$E7,$CE,$BC,$11
	.dc.b	$FF,$42,$E7,$F1,$BE,$97,$FC,$7F,$97,$C4,$6E,$CA,$FE,$0D,$FA,$37
	.dc.b	$F9,$5A,$3F,$09,$BA,$B7,$5F,$1D,$B8,$75,$AF,$C5,$5F,$1A,$FA,$75
	.dc.b	$AE,$47,$CD,$FF,$49,$9F,$06,$1F,$EE,$E8,$C4,$7F,$17,$9D,$E1,$F7
	.dc.b	$DB,$1E,$CC,$5E,$D7,$89,$B5,$F9,$BA,$DE,$7F,$82,$AD,$12,$75,$85
	.dc.b	$C3,$FC,$71,$5F,$E0,$A9,$FF,$37,$B6,$3F,$3F,$93,$E6,$FE,$1F,$47
	.dc.b	$F9,$AF,$CF,$B7,$B3,$FC,$CC,$5C,$3D,$F7,$C5,$EB,$3C,$BE,$AB,$F1
	.dc.b	$B1,$FE,$7E,$6E,$3F,$E1,$7F,$AB,$7A,$9B,$E6,$97,$B7,$EA,$F7,$AB
	.dc.b	$73,$93,$F7,$DF,$8C,$3E,$2F,$E7,$F2,$23,$13,$FA,$D6,$E5,$F8,$A6
	.dc.b	$BF,$CF,$3B,$7E,$0F,$39,$FC,$1C,$F0,$E0,$FE,$8F,$1F,$1A,$DF,$37
	.dc.b	$F7,$B8,$BF,$DC,$0D,$8C,$9F,$E1,$42,$E3,$FE,$4A,$6B,$FE,$1C,$6B
	.dc.b	$FE,$96,$6B,$FE,$A9,$6C,$FE,$99,$FE,$AF,$1E,$39,$F4,$BF,$9A,$FD
	.dc.b	$1D,$F4,$7F,$11,$B5,$FC,$3D,$B3,$FA,$C6,$BF,$EF,$8D,$11,$88,$FC
	.dc.b	$EB,$5F,$EE,$4D,$8D,$D7,$C3,$FA,$D7,$CA,$FE,$31,$B8,$7F,$35,$BF
	.dc.b	$0E,$9F,$EB,$19,$CE,$3F,$E8,$FC,$5F,$3B,$F3,$5B,$87,$F8,$F9,$B3
	.dc.b	$FE,$74,$6B,$FE,$0C,$9E,$04,$9E,$0F,$97,$E5,$BC,$87,$E8,$2B,$9E
	.dc.b	$1E,$38,$FE,$07,$FA,$43,$E8,$3F,$1A,$DC,$3F,$9C,$37,$1F,$E2,$AD
	.dc.b	$7E,$53,$5F,$E0,$AD,$7F,$C3,$4D,$C0,$E7,$77,$B1,$FA,$FA,$FF,$2B
	.dc.b	$F5,$79,$31,$FD,$BC,$E6,$9F,$5C,$FA,$BF,$CA,$37,$3F,$F3,$2B,$71
	.dc.b	$FF,$4D,$B5,$FF,$36,$B5,$FF,$73,$B5,$FF,$7D,$36,$BB,$D7,$E8,$6D
	.dc.b	$F5,$1E,$43,$F8,$5D,$77,$DB,$FC,$D7,$1F,$BC,$F8,$96,$5F,$C9,$BA
	.dc.b	$FF,$DB,$AD,$8C,$BF,$E9,$66,$E7,$FE,$96,$6B,$FE,$96,$6B,$FE,$A0
	.dc.b	$6B,$FE,$95,$6E,$EF,$C6,$9F,$43,$FD,$01,$B7,$FD,$27,$2F,$2E,$1C
	.dc.b	$63,$55,$FC,$9B,$E1,$BF,$83,$46,$CF,$EA,$1B,$FF,$B1,$7A,$BF,$8C
	.dc.b	$FE,$92,$F8,$DE,$11,$FC,$73,$E1,$7F,$C7,$77,$FE,$B2,$6E,$5F,$EB
	.dc.b	$26,$BF,$EB,$26,$BF,$EC,$DF,$8B,$FE,$6F,$F5,$FF,$DC,$6B,$C5,$2C
	.dc.b	$BF,$96,$D6,$BD,$0F,$EB,$FE,$DF,$F8,$A7,$E0,$BF,$AA,$B7,$0F,$ED
	.dc.b	$18,$48,$C3,$80,$E1,$D7,$A3,$FE,$95,$9D,$46,$A3,$0F,$19,$E0,$FF
	.dc.b	$75,$6B,$E7,$C3,$FA,$FA,$E1,$FA,$39,$27,$31,$F5,$2B,$CA,$EA,$33
	.dc.b	$87,$D2,$FA,$F7,$DD,$7F,$5C,$6C,$7C,$6C,$10,$FB,$7D,$54,$AD,$4F
	.dc.b	$0C,$BC,$91,$D6,$B5,$1F,$E4,$8F,$D2,$BF,$C9,$13,$D7,$EC,$BF,$8D
	.dc.b	$E7,$FE,$49,$9C,$D7,$23,$EB,$4E,$AB,$FC,$13,$3A,$AF,$83,$39,$AF
	.dc.b	$F5,$04,$E6,$BF,$1F,$39,$E2,$7E,$76,$FA,$3F,$F0,$8C,$61,$B1,$9A
	.dc.b	$FA,$F3,$96,$C6,$70,$FC,$04,$E5,$B1,$97,$F0,$F1,$97,$F1,$3F,$7A
	.dc.b	$FC,$4B,$FC,$65,$7F,$E7,$3E,$B7,$CC,$FF,$8C,$2B,$57,$FB,$18,$FF
	.dc.b	$08,$FD,$5B,$ED,$A9,$FE,$34,$F6,$7F,$97,$70,$7E,$12,$75,$97,$DB
	.dc.b	$6E,$1F,$37,$91,$18,$AF,$F3,$D5,$F2,$BF,$DB,$CE,$5F,$DC,$FD,$AB
	.dc.b	$E6,$BF,$C4,$6F,$CF,$C7,$F5,$8E,$67,$F1,$4E,$71,$39,$9F,$13,$93
	.dc.b	$18,$FE,$71,$E9,$D1,$FE,$9B,$C3,$F6,$7D,$3C,$6F,$F5,$B1,$82,$63
	.dc.b	$FA,$EA,$E0,$DA,$E1,$5F,$AD,$6B,$FA,$DF,$37,$FC,$7B,$D5,$FD,$7A
	.dc.b	$B5,$86,$27,$FC,$83,$CC,$F9,$1B,$E7,$61,$B7,$F9,$D6,$E3,$F3,$EB
	.dc.b	$84,$FE,$86,$BE,$DF,$CC,$48,$FF,$5A,$7D,$07,$FB,$29,$FB,$11,$9B
	.dc.b	$FB,$1C,$28,$FF,$39,$D6,$E7,$31,$FC,$6B,$5C,$47,$F1,$2F,$8B,$F9
	.dc.b	$24,$C3,$E2,$36,$3F,$89,$4F,$07,$E3,$47,$F8,$E7,$E9,$BF,$CA,$BC
	.dc.b	$5F,$C4,$1B,$7A,$3F,$84,$38,$AF,$E5,$DD,$8F,$A4,$FF,$34,$B6,$BF
	.dc.b	$DA,$5F,$6A,$F8,$DC,$D4,$9F,$E2,$55,$BB,$FF,$68,$DE,$E7,$35,$FE
	.dc.b	$74,$FA,$E7,$D8,$D3,$1F,$2D,$6B,$F8,$E4,$E9,$B3,$A6,$9F,$C8,$7E
	.dc.b	$B3,$FC,$F1,$8F,$F3,$DE,$B4,$62,$3F,$A1,$C6,$3E,$1F,$FC,$BD,$39
	.dc.b	$9F,$EB,$EF,$D5,$DF,$F9,$BB,$6F,$35,$FE,$8C,$8C,$57,$F5,$56,$FA
	.dc.b	$3F,$8F,$CF,$5C,$3F,$EB,$2E,$84,$63,$1F,$F7,$E4,$61,$F5,$4E,$E7
	.dc.b	$B7,$C7,$F9,$89,$3F,$D5,$34,$F6,$C7,$AB,$AF,$C8,$70,$9F,$88,$FF
	.dc.b	$B0,$BF,$26,$7D,$87,$FA,$31,$B3,$FE,$F7,$6E,$BF,$ED,$46,$BF,$23
	.dc.b	$5F,$F1,$A1,$FE,$39,$D5,$3C,$37,$F8,$FE,$E7,$C5,$26,$1F,$DE,$AB
	.dc.b	$78,$E3,$1F,$E3,$98,$F4,$BF,$34,$E7,$0F,$A3,$EC,$7D,$62,$63,$F0
	.dc.b	$6F,$FC,$6A,$DF,$C2,$3E,$A1,$F5,$66,$C7,$87,$37,$D9,$7F,$88,$A3
	.dc.b	$FD,$09,$18,$7F,$9E,$E7,$FE,$E0,$8C,$3F,$E2,$81,$A7,$C0,$AE,$51
	.dc.b	$B6,$BA,$9E,$35,$DA,$73,$5C,$63,$BD,$6E,$7F,$C6,$BF,$0D,$FD,$7E
	.dc.b	$3F,$C4,$1D,$EF,$AD,$CF,$5F,$7C,$14,$BE,$1C,$7F,$D1,$AF,$73,$E8
	.dc.b	$FB,$6F,$D9,$3B,$C3,$FD,$FB,$18,$BF,$06,$68,$FF,$5E,$C6,$1D,$4F
	.dc.b	$0E,$DE,$B5,$B6,$CF,$09,$D6,$5E,$6B,$E5,$5A,$37,$F0,$F6,$75,$5C
	.dc.b	$31,$F3,$1E,$9A,$DC,$EA,$30,$DC,$5C,$CF,$9E,$F2,$B6,$74,$BC,$AC
	.dc.b	$CC,$9C,$BC,$23,$17,$E2,$BE,$B3,$8C,$FE,$3E,$1C,$F0,$E2,$79,$6D
	.dc.b	$F5,$DF,$EA,$FF,$47,$0D,$8F,$6F,$C0,$3E,$9F,$E6,$C7,$DC,$9D,$63
	.dc.b	$F8,$38,$E7,$F0,$BF,$EF,$B7,$E2,$98,$C3,$E2,$61,$CF,$6F,$C0,$CE
	.dc.b	$4D,$E1,$E1,$7F,$CC,$4D,$7F,$CC,$ED,$7C,$57,$DD,$DF,$07,$F3,$87
	.dc.b	$FC,$CB,$28,$C4,$6B,$C2,$FF,$34,$BF,$0A,$32,$F6,$79,$65,$ED,$AE
	.dc.b	$55,$F1,$A7,$3C,$3F,$9E,$C9,$F4,$7F,$DA,$F1,$D5,$C7,$DB,$7C,$15
	.dc.b	$70,$F2,$7F,$D0,$79,$7E,$7B,$ED,$1E,$EE,$33,$FD,$5F,$22,$73,$8F
	.dc.b	$E3,$5A,$FE,$0D,$7D,$2F,$18,$FE,$0B,$1A,$C3,$B6,$19,$BE,$8D,$9E
	.dc.b	$35,$D6,$30,$47,$C1,$7C,$7E,$5B,$E9,$61,$FB,$66,$E3,$E8,$6D,$7F
	.dc.b	$16,$7F,$D3,$D3,$BE,$2E,$44,$9F,$D4,$4E,$AB,$94,$E6,$7F,$41,$3D
	.dc.b	$32,$3B,$8E,$D3,$BB,$ED,$1D,$0F,$81,$F2,$BF,$85,$7F,$71,$D7,$FE
	.dc.b	$5A,$F0,$FE,$43,$63,$F7,$53,$B7,$F6,$0D,$E7,$FF,$5D,$E4,$8C,$4F
	.dc.b	$D8,$6B,$FC,$D9,$B1,$BE,$1F,$F6,$1D,$72,$9C,$92,$33,$5F,$E0,$2F
	.dc.b	$9E,$FF,$64,$F8,$9F,$E0,$26,$DE,$67,$FC,$21,$F5,$B8,$7F,$C0,$4F
	.dc.b	$E8,$32,$FF,$08,$B7,$C5,$34,$C7,$C1,$71,$9F,$F0,$BB,$64,$F5,$D3
	.dc.b	$E0,$7D,$66,$62,$61,$FE,$21,$75,$AA,$73,$3F,$DB,$A6,$E8,$CF,$E4
	.dc.b	$57,$03,$F6,$F8,$76,$BF,$F7,$0B,$5F,$F7,$43,$5F,$F0,$27,$C5,$79
	.dc.b	$C8,$FF,$8C,$E7,$D4,$7C,$12,$3F,$DE,$5F,$89,$F8,$BF,$B7,$C1,$FE
	.dc.b	$73,$3F,$E0,$76,$BF,$E0,$68,$F8,$86,$FC,$19,$1F,$E4,$18,$CB,$48
	.dc.b	$7F,$83,$BE,$F7,$E6,$1F,$4C,$8F,$F8,$2D,$FA,$DE,$0F,$D1,$24,$7F
	.dc.b	$93,$7E,$86,$BF,$E1,$88,$BA,$FF,$91,$A7,$EB,$7D,$AB,$E6,$29,$2F
	.dc.b	$FC,$45,$F6,$8F,$B3,$C4,$7F,$BA,$78,$17,$C6,$3F,$C5,$0D,$75,$5C
	.dc.b	$5F,$F3,$CF,$95,$F8,$9F,$67,$FC,$0D,$18,$3F,$C0,$B9,$FF,$86,$7E
	.dc.b	$93,$FC,$F9,$18,$E4,$FF,$04,$A8,$CB,$83,$FC,$1A,$E1,$C3,$5E,$2D
	.dc.b	$77,$EC,$FF,$40,$7A,$36,$35,$18,$3F,$CB,$5F,$67,$FA,$47,$C8,$3A
	.dc.b	$91,$98,$FF,$71,$37,$0F,$CD,$47,$F8,$D3,$DD,$FF,$4A,$3C,$AB,$8F
	.dc.b	$67,$EB,$6F,$86,$5F,$E2,$89,$CD,$7F,$AE,$B8,$FF,$65,$F2,$1F,$A9
	.dc.b	$8E,$4B,$AB,$26,$BA,$A4,$EA,$32,$F0,$8F,$DE,$7C,$17,$CD,$8F,$23
	.dc.b	$5F,$F7,$E3,$75,$FA,$DF,$63,$FC,$2E,$7E,$A3,$F2,$48,$FF,$6C,$DD
	.dc.b	$18,$8F,$B0,$70,$34,$E5,$FF,$50,$F9,$DD,$F5,$F8,$4B,$E7,$3F,$89
	.dc.b	$36,$7E,$53,$B7,$EA,$37,$6F,$F0,$EC,$F0,$38,$E1,$FA,$2F,$B4,$97
	.dc.b	$F9,$A7,$F8,$D7,$E9,$6F,$A8,$CC,$FE,$87,$A5,$58,$D7,$EE,$F5,$30
	.dc.b	$E1,$7F,$86,$F8,$C7,$E9,$4F,$87,$FE,$4D,$9F,$DE,$AF,$F3,$BB,$6B
	.dc.b	$73,$F3,$74,$7C,$A7,$E5,$49,$7F,$35,$AF,$CC,$8D,$B9,$21,$FC,$93
	.dc.b	$E4,$9C,$FF,$8A,$F1,$24,$FF,$A8,$67,$2D,$9C,$AD,$BF,$37,$C5,$CA
	.dc.b	$FF,$A0,$5F,$0A,$D7,$27,$E6,$7E,$D5,$F0,$1F,$D2,$5B,$3F,$A1,$6B
	.dc.b	$C0,$FD,$62,$F8,$2E,$13,$27,$83,$CE,$3F,$33,$F7,$F2,$7F,$32,$FE
	.dc.b	$57,$F0,$07,$DE,$D3,$F8,$03,$89,$FF,$20,$B5,$FC,$AB,$5F,$E1,$CD
	.dc.b	$7F,$89,$B5,$FF,$4E,$B5,$39,$D1,$F6,$FE,$B9,$FE,$01,$F0,$1F,$DA
	.dc.b	$1B,$E7,$FF,$6D,$9D,$47,$C3,$F7,$5F,$37,$FC,$5D,$B1,$FE,$58,$6B
	.dc.b	$FA,$C6,$BF,$C9,$70,$7D,$4F,$B7,$8F,$EC,$B7,$87,$F8,$73,$6B,$FC
	.dc.b	$A7,$FA,$67,$F9,$65,$FE,$81,$D5,$E4,$BF,$CF,$B6,$FF,$CA,$6D,$9F
	.dc.b	$F2,$DF,$E9,$3C,$6B,$F9,$0E,$48,$E9,$3F,$CF,$23,$1F,$03,$EB,$9C
	.dc.b	$CF,$F9,$AD,$AF,$48,$D1,$1F,$81,$1A,$CD,$F4,$7F,$EE,$3C,$3F,$2E
	.dc.b	$DC,$BF,$96,$36,$FF,$9D,$7D,$C3,$E4,$3F,$C3,$CD,$8F,$E2,$F8,$23
	.dc.b	$11,$F3,$5A,$FF,$56,$8D,$4F,$F8,$B6,$73,$3F,$CE,$DA,$F8,$23,$47
	.dc.b	$F8,$CF,$E9,$1F,$6B,$FE,$2F,$6C,$FF,$47,$6B,$FD,$CF,$F1,$B7,$D8
	.dc.b	$7F,$8C,$5B,$1F,$E2,$46,$BF,$E2,$46,$BF,$E5,$66,$BF,$E8,$C6,$AF
	.dc.b	$19,$18,$6C,$FF,$8C,$A3,$79,$7F,$63,$F9,$2F,$F2,$32,$69,$FE,$49
	.dc.b	$A7,$DC,$53,$FC,$A4,$E3,$0F,$9E,$D7,$F3,$4D,$7E,$73,$5F,$EC,$CD
	.dc.b	$7F,$BA,$36,$BD,$53,$FE,$5F,$7F,$2C,$7F,$79,$ED,$F3,$9F,$E5,$9F
	.dc.b	$AC,$FD,$3D,$FF,$7E,$B4,$62,$3F,$C9,$AD,$7F,$CF,$ED,$7F,$D6,$AD
	.dc.b	$7E,$53,$63,$9F,$3F,$F9,$6A,$30,$6E,$DF,$3A,$FA,$9F,$F3,$83,$70
	.dc.b	$FF,$04,$35,$FE,$51,$2F,$B6,$7D,$BF,$F9,$DD,$B3,$F9,$A6,$BF,$C1
	.dc.b	$5A,$FE,$65,$AF,$F8,$45,$AF,$F8,$7D,$B3,$DE,$7F,$D1,$51,$8A,$FF
	.dc.b	$1E,$D7,$91,$BC,$DF,$D2,$1B,$CD,$FE,$DA,$FB,$0B,$5F,$8F,$78,$BF
	.dc.b	$EA,$A6,$BB,$F7,$7F,$D0,$31,$88,$FF,$19,$75,$FF,$6B,$CF,$FD,$D7
	.dc.b	$ED,$7F,$A5,$E3,$FB,$70,$F6,$7F,$29,$5F,$E3,$18,$E0,$49,$FF,$53
	.dc.b	$3B,$E4,$FF,$50,$7C,$FF,$F7,$CA,$F9,$ED,$D7,$FB,$1B,$74,$FF,$5F
	.dc.b	$7C,$D7,$0F,$02,$5F,$F8,$FE,$B4,$FF,$B0,$67,$FD,$51,$F6,$9F,$E1
	.dc.b	$9E,$63,$4C,$7C,$7F,$08,$CF,$A1,$C0,$97,$F9,$79,$D1,$E7,$7F,$DA
	.dc.b	$6D,$7F,$DE,$58,$FF,$AA,$A3,$11,$FB,$FB,$79,$7E,$42,$47,$E8,$FF
	.dc.b	$4D,$7D,$2E,$23,$FD,$09,$5C,$1F,$D4,$35,$FF,$44,$4F,$07,$FC,$D2
	.dc.b	$3E,$DD,$F6,$BF,$E0,$D8,$FF,$22,$C6,$38,$3F,$C8,$F1,$B9,$D3,$FE
	.dc.b	$81,$9D,$3C,$3E,$59,$C1,$C2,$51,$89,$D4,$E6,$31,$1D,$BE,$4F,$FD
	.dc.b	$08,$9A,$47,$FD,$9B,$97,$FB,$F7,$8B,$DF,$F1,$FF,$E7,$EC,$FF,$C8
	.dc.b	$7F,$6F,$FE,$60,$9C,$92,$FE,$3D,$6E,$11,$FC,$BE,$78,$3A,$8F,$B9
	.dc.b	$F3,$9F,$E8,$A8,$FB,$6D,$78,$F2,$38,$93,$99,$FE,$75,$18,$AF,$F7
	.dc.b	$3E,$5F,$D2,$1F,$AF,$E3,$78,$EB,$77,$70,$FC,$9C,$DB,$E1,$5F,$80
	.dc.b	$8D,$61,$D6,$31,$87,$FA,$77,$3F,$F1,$D3,$7C,$4E,$9A,$3E,$8F,$C1
	.dc.b	$DB,$FE,$4D,$9B,$7C,$30,$FE,$3A,$D3,$FC,$DF,$7C,$70,$E8,$7F,$AA
	.dc.b	$FE,$DF,$F1,$6D,$E4,$F6,$76,$3E,$DF,$FC,$47,$39,$26,$3F,$CA,$AF
	.dc.b	$8B,$FE,$31,$9D,$B9,$BF,$89,$E0,$7F,$BA,$6F,$FD,$37,$C2,$8C,$57
	.dc.b	$FB,$56,$31,$F2,$1F,$4D,$B5,$FE,$0C,$E5,$7C,$CF,$C2,$6C,$7F,$51
	.dc.b	$C3,$9E,$DF,$E7,$09,$CE,$3F,$D0,$29,$F3,$1F,$7F,$85,$18,$3F,$C3
	.dc.b	$25,$9F,$F2,$23,$5F,$F0,$C3,$5F,$F6,$83,$5F,$F6,$F3,$63,$AC,$F7
	.dc.b	$AF,$CC,$CF,$FB,$6D,$B5,$E5,$9F,$63,$F8,$BF,$27,$FD,$75,$5F,$21
	.dc.b	$B1,$FD,$B3,$E1,$FE,$34,$62,$3F,$B6,$C6,$09,$3C,$5F,$81,$F6,$CF
	.dc.b	$B7,$FC,$5B,$67,$FB,$D5,$F6,$CF,$FD,$33,$39,$C3,$FB,$4E,$CF,$AB
	.dc.b	$F9,$4D,$C3,$FC,$27,$18,$23,$82,$4F,$F8,$5A,$3C,$7F,$0B,$E2,$3F
	.dc.b	$3B,$5C,$E7,$FB,$BD,$FF,$B0,$EF,$FC,$1D,$F6,$EF,$A3,$FC,$33,$63
	.dc.b	$F1,$AD,$AF,$C6,$35,$FE,$12,$7F,$8D,$DC,$3F,$C1,$1A,$E6,$7A,$9F
	.dc.b	$87,$F6,$3F,$59,$5F,$E6,$4E,$AF,$E0,$17,$DF,$4F,$F3,$4C,$7E,$DB
	.dc.b	$D5,$7D,$BF,$E3,$DB,$7F,$E6,$66,$E5,$FE,$91,$6B,$FE,$65,$6B,$FC
	.dc.b	$83,$91,$18,$84,$62,$BB,$1E,$7F,$83,$FE,$0F,$9F,$FB,$33,$0F,$F3
	.dc.b	$24,$E2,$7F,$7F,$F2,$5F,$A2,$9F,$F5,$23,$70,$F8,$E3,$B5,$19,$8F
	.dc.b	$F4,$A3,$5F,$E6,$B1,$A9,$FE,$58,$D7,$A9,$89,$E1,$3F,$E6,$9F,$B0
	.dc.b	$FF,$1D,$4F,$F9,$2D,$B1,$FE,$BD,$9F,$26,$BF,$B8,$8E,$1E,$B7,$F2
	.dc.b	$4F,$B7,$FD,$FC,$FF,$BC,$1B,$87,$93,$ED,$FF,$DF,$6F,$F4,$A9,$D7
	.dc.b	$6F,$F2,$0F,$82,$FF,$21,$CF,$FB,$15,$B8,$FC,$2D,$3F,$D9,$BE,$CF
	.dc.b	$B2,$73,$3F,$E9,$BF,$69,$F2,$9F,$CD,$BE,$DF,$FB,$5E,$3D,$CF,$F6
	.dc.b	$4C,$67,$CE,$FE,$95,$ED,$FF,$72,$F8,$6F,$EA,$4D,$AF,$A3,$8F,$F9
	.dc.b	$7B,$D1,$56,$9B,$3A,$AE,$15,$F6,$1A,$FE,$81,$AF,$F6,$A6,$BF,$8A
	.dc.b	$C3,$FA,$BE,$FF,$37,$A3,$D7,$E7,$79,$B1,$EF,$7F,$E6,$AB,$DE,$3F
	.dc.b	$95,$FB,$77,$C7,$FF,$56,$6C,$7C,$1A,$FC,$BC,$3E,$DF,$55,$2B,$59
	.dc.b	$2B,$8D,$FF,$9E,$1E,$13,$A7,$39,$FE,$0F,$C2,$7C,$73,$E8,$7F,$B3
	.dc.b	$FD,$9F,$F1,$9A,$D4,$E7,$0E,$F3,$97,$FC,$0B,$19,$C7,$FD,$0F,$39
	.dc.b	$7F,$D7,$D1,$9A,$F8,$D3,$9E,$87,$DB,$BE,$A3,$FC,$19,$18,$6C,$67
	.dc.b	$1F,$BB,$39,$6C,$67,$1F,$D8,$CE,$5B,$19,$7F,$8D,$C6,$5F,$E3,$BF
	.dc.b	$2A,$FB,$7F,$E9,$B9,$7C,$5F,$05,$F6,$FF,$E1,$9A,$D5,$FF,$86,$67
	.dc.b	$39,$7D,$2F,$B0,$FF,$2C,$B7,$E0,$3F,$C3,$DE,$EF,$F1,$F8,$FE,$29
	.dc.b	$3A,$BF,$F5,$6B,$72,$FC,$53,$67,$FC,$63,$5C,$6B,$F8,$CC,$E5,$FE
	.dc.b	$37,$F6,$FF,$EE,$3A,$FC,$1B,$F3,$DF,$E4,$4E,$63,$FD,$76,$E5,$E3
	.dc.b	$F0,$DF,$E2,$68,$C7,$31,$E1,$8F,$E7,$5E,$1D,$9F,$E6,$2A,$FD,$61
	.dc.b	$30,$FD,$67,$5F,$7C,$7F,$54,$D7,$F5,$4D,$7F,$6D,$3C,$5F,$E5,$38
	.dc.b	$FF,$04,$F4,$7F,$C7,$3D,$9F,$D7,$73,$FE,$35,$97,$87,$2F,$D0,$CE
	.dc.b	$5F,$E8,$1C,$CF,$8E,$FC,$DF,$A5,$F9,$C6,$CF,$E6,$9B,$9F,$E6,$1B
	.dc.b	$8F,$D2,$F9,$1F,$F2,$4F,$01,$FF,$2C,$7D,$13,$ED,$73,$13,$1F,$C1
	.dc.b	$3A,$8C,$62,$C3,$D3,$C3,$C6,$FE,$3D,$7F,$15,$23,$82,$57,$F1,$59
	.dc.b	$E0,$D9,$FF,$1A,$7D,$2F,$F9,$32,$FF,$89,$36,$FE,$7C,$FF,$98,$A3
	.dc.b	$15,$FD,$8B,$B1,$F4,$BF,$E6,$66,$DF,$F1,$1F,$01,$F6,$FF,$ED,$D8
	.dc.b	$FE,$1F,$3A,$BF,$64,$7F,$6F,$8C,$65,$FC,$8D,$B1,$FE,$8A,$AE,$55
	.dc.b	$C0,$F1,$61,$FE,$76,$FB,$77,$D9,$53,$1F,$6F,$70,$FE,$3F,$3A,$6C
	.dc.b	$E9,$A7,$F2,$4F,$AF,$FF,$3D,$61,$F8,$C6,$BF,$E1,$CE,$94,$62,$3F
	.dc.b	$C1,$0D,$7F,$C0,$EF,$D0,$D3,$F9,$F3,$6B,$7F,$11,$FE,$8E,$8C,$57
	.dc.b	$0E,$0F,$F4,$F7,$2B,$4D,$46,$1B,$87,$CF,$9D,$3E,$99,$D3,$EB,$F1
	.dc.b	$BF,$CE,$58,$7F,$2F,$C3,$D4,$D7,$D3,$1A,$EC,$F2,$1E,$56,$D7,$DD
	.dc.b	$FB,$77,$D9,$7F,$A2,$5B,$1F,$E9,$06,$DF,$FA,$3D,$AF,$83,$DC,$F8
	.dc.b	$07,$FA,$EE,$37,$B7,$C6,$8C,$D7,$D3,$75,$78,$BF,$C1,$6F,$F4,$A7
	.dc.b	$17,$FD,$57,$0E,$78,$7D,$49,$F9,$DE,$CF,$C2,$24,$7F,$80,$2F,$ED
	.dc.b	$C7,$C6,$F9,$3F,$A8,$70,$9F,$C2,$3E,$7F,$B7,$FF,$12,$47,$F9,$7B
	.dc.b	$A1,$8A,$73,$9F,$FB,$8A,$31,$5F,$E9,$28,$57,$0A,$DC,$62,$B6,$FF
	.dc.b	$BB,$67,$55,$C5,$B1,$99,$DC,$F8,$DE,$3F,$0D,$FD,$96,$7F,$A0,$FC
	.dc.b	$BF,$F8,$EE,$31,$7C,$9D,$66,$EF,$56,$FA,$FF,$DF,$7F,$6F,$FB,$2B
	.dc.b	$E2,$D8,$CF,$C4,$6F,$2F,$6C,$E5,$BF,$31,$9C,$D3,$96,$D6,$63,$04
	.dc.b	$9C,$46,$59,$AA,$D8,$F9,$27,$73,$98,$C3,$71,$E1,$8F,$F8,$67,$EC
	.dc.b	$FC,$9C,$A6,$F9,$58,$65,$F1,$CB,$EC,$D6,$E3,$17,$FE,$23,$BD,$D7
	.dc.b	$93,$0E,$F7,$B8,$C1,$CB,$0E,$F8,$72,$9D,$63,$F6,$69,$F5,$9F,$EB
	.dc.b	$3F,$47,$C5,$18,$F8,$7F,$F7,$BB,$D3,$DF,$7D,$1F,$D1,$8D,$57,$0B
	.dc.b	$E5,$1E,$23,$27,$19,$E1,$1A,$C3,$C4,$F0,$9F,$F4,$BB,$5F,$F5,$FB
	.dc.b	$5F,$A8,$D7,$D3,$C1,$FC,$D9,$FF,$37,$C2,$31,$1D,$1A,$F4,$6B,$D1
	.dc.b	$E5,$3E,$46,$EB,$BF,$0B,$EB,$64,$FB,$7F,$EE,$58,$7C,$1B,$DB,$CE
	.dc.b	$B0,$70,$9D,$C7,$D6,$E1,$7D,$BB,$DE,$FF,$6E,$E7,$FA,$A8,$C5,$7E
	.dc.b	$09,$AF,$15,$F6,$FC,$34,$FC,$8C,$6A,$B9,$5E,$6F,$F8,$2C,$E6,$B7
	.dc.b	$79,$C3,$C7,$76,$B5,$7C,$23,$15,$BF,$96,$FC,$2C,$FA,$DB,$8F,$F0
	.dc.b	$26,$E3,$FA,$47,$FD,$A7,$19,$E0,$E4,$4A,$FC,$FC,$EA,$B8,$4E,$67
	.dc.b	$B3,$E1,$71,$5C,$DE,$31,$97,$F5,$3F,$2D,$F8,$37,$F5,$DD,$7F,$C4
	.dc.b	$FC,$2F,$88,$D8,$DF,$07,$C3,$AE,$11,$FC,$9D,$BC,$DF,$CD,$5B,$1F
	.dc.b	$C0,$1A,$FF,$32,$6E,$E6,$92,$BF,$87,$4F,$FB,$8C,$FE,$FB,$F6,$F8
	.dc.b	$F0,$FF,$BE,$B7,$0F,$F0,$1B,$67,$FC,$07,$F3,$F7,$2F,$F0,$4B,$FE
	.dc.b	$D0,$8C,$9F,$D4,$FC,$64,$66,$77,$39,$F8,$3E,$55,$FE,$D6,$8F,$F0
	.dc.b	$C4,$77,$D3,$D5,$F6,$5A,$89,$3A,$3F,$C6,$DA,$A7,$31,$FD,$B2,$FA
	.dc.b	$36,$BA,$56,$E3,$8D,$7A,$6F,$BC,$7F,$BA,$9A,$FC,$83,$E7,$47,$AE
	.dc.b	$DF,$19,$EA,$24,$FF,$8C,$67,$DA,$6C,$8F,$C5,$79,$7D,$67,$BF,$BF
	.dc.b	$FD,$A5,$3F,$E5,$16,$BE,$B8,$F8,$E6,$FB,$E4,$7F,$8F,$E3,$2D,$21
	.dc.b	$F0,$FE,$DF,$E7,$9E,$92,$3F,$5D,$F6,$78,$5F,$50,$91,$FE,$4B,$FB
	.dc.b	$7A,$FC,$D8,$BC,$7F,$E4,$49,$FB,$1F,$66,$F9,$AA,$4B,$FF,$36,$7D
	.dc.b	$B3,$ED,$F1,$1F,$E6,$3E,$15,$F1,$8F,$F3,$93,$5D,$57,$17,$FE,$00
	.dc.b	$FB,$1F,$DF,$9B,$E5,$7E,$56,$31,$1F,$E6,$F1,$F4,$DF,$E8,$07,$FB
	.dc.b	$86,$9F,$A3,$8C,$47,$DF,$37,$EC,$7E,$75,$A7,$E7,$E3,$6F,$E1,$5A
	.dc.b	$FE,$B9,$AF,$F7,$D6,$BC,$8F,$F5,$FF,$DB,$FE,$99,$F1,$F6,$7F,$80
	.dc.b	$FA,$93,$99,$FF,$23,$B7,$05,$7F,$80,$BD,$D7,$17,$ED,$6B,$53,$CD
	.dc.b	$B6,$C3,$AF,$67,$EB,$F0,$E3,$3E,$23,$5F,$67,$FA,$77,$FC,$AE,$FE
	.dc.b	$81,$D5,$23,$11,$AE,$9F,$E0,$D1,$C7,$3F,$3C,$F4,$EA,$C9,$AA,$FE
	.dc.b	$19,$F0,$3F,$5A,$7F,$DB,$BB,$60,$FE,$24,$FD,$AF,$B7,$B0,$7E,$B3
	.dc.b	$63,$EB,$46,$08,$E0,$87,$DF,$8E,$0F,$FB,$CB,$CD,$F1,$3F,$E7,$65
	.dc.b	$F3,$FF,$C8,$5B,$1F,$09,$DE,$5E,$ED,$3F,$C3,$F7,$BB,$97,$B8,$FD
	.dc.b	$3F,$DB,$C8,$FC,$FB,$FC,$87,$ED,$DF,$57,$99,$FD,$37,$4A,$B1,$97
	.dc.b	$CC,$EA,$61,$C2,$FE,$D7,$C6,$3E,$DD,$F0,$FF,$CB,$73,$FC,$1E,$1D
	.dc.b	$B9,$BF,$9F,$5E,$EB,$E6,$FC,$9F,$CC,$26,$3F,$A1,$6B,$F2,$5A,$FE
	.dc.b	$8D,$FE,$59,$F2,$5F,$85,$8F,$C3,$F3,$BC,$4F,$C2,$9F,$CF,$E7,$53
	.dc.b	$FC,$0D,$BB,$7C,$BA,$E1,$3C,$AB,$84,$7F,$53,$FB,$77,$C0,$52,$57
	.dc.b	$E5,$A1,$F0,$38,$8F,$F5,$8C,$70,$75,$3C,$E7,$F2,$7F,$5B,$27,$F3
	.dc.b	$FD,$48,$FC,$9F,$DB,$3E,$DE,$9F,$C1,$DC,$57,$F6,$56,$BF,$C3,$5A
	.dc.b	$FF,$4D,$6B,$FD,$D9,$AF,$FA,$15,$A9,$CF,$53,$ED,$FD,$B3,$FB,$4F
	.dc.b	$7B,$FB,$0E,$FF,$93,$6F,$87,$FC,$36,$B7,$4A,$FD,$AF,$BC,$F9,$DF
	.dc.b	$E4,$0D,$9F,$EF,$CD,$7F,$93,$35,$FE,$CD,$8B,$E7,$7D,$F7,$FA,$36
	.dc.b	$E3,$FC,$75,$B4,$B7,$DB,$FF,$96,$2F,$F3,$9D,$7E,$66,$F0,$FE,$64
	.dc.b	$FF,$0A,$53,$ED,$F8,$65,$FE,$AB,$8C,$7C,$97,$A2,$7F,$CC,$79,$7F
	.dc.b	$88,$A7,$2F,$AA,$32,$F1,$6B,$E5,$6C,$70,$8F,$F0,$23,$5F,$F1,$6E
	.dc.b	$8F,$A5,$FF,$17,$65,$FC,$0D,$B8,$FF,$3D,$6D,$7F,$33,$FB,$77,$C7
	.dc.b	$FF,$8B,$5B,$1F,$EF,$D6,$BF,$39,$AF,$81,$AF,$F4,$96,$BD,$63,$FC
	.dc.b	$65,$18,$C0,$DF,$A7,$7D,$BF,$F8,$F5,$B7,$FC,$F9,$AF,$F7,$7F,$B7
	.dc.b	$7D,$1F,$F8,$DD,$B1,$FE,$29,$6B,$FE,$29,$F1,$11,$88,$FF,$29,$C6
	.dc.b	$27,$FC,$65,$39,$9F,$F6,$EB,$54,$ED,$ED,$3F,$E4,$DC,$5F,$25,$FE
	.dc.b	$55,$6C,$FF,$56,$6B,$F9,$FD,$7F,$B7,$5B,$ED,$FF,$CB,$8D,$9F,$F0
	.dc.b	$15,$78,$27,$FC,$B5,$39,$9F,$CF,$B5,$FC,$0B,$5E,$B1,$C0,$8F,$49
	.dc.b	$FC,$E5,$7F,$7F,$6C,$7F,$75,$6D,$3E,$B7,$F5,$93,$FD,$D6,$31,$CF
	.dc.b	$FE,$59,$9C,$CF,$CF,$6B,$F3,$9A,$8F,$96,$E1,$B1,$CC,$8F,$C2,$F9
	.dc.b	$C7,$D5,$7F,$A0,$5B,$1F,$E0,$26,$BF,$34,$FF,$07,$7D,$BB,$EA,$7F
	.dc.b	$D1,$0D,$8F,$9A,$D7,$F8,$4B,$5F,$D5,$B5,$FF,$12,$35,$EB,$1F,$E8
	.dc.b	$D8,$C6,$46,$D7,$F9,$0B,$97,$FB,$A3,$74,$FF,$17,$37,$0F,$C6,$FD
	.dc.b	$85,$D4,$FF,$92,$A3,$F2,$ED,$7E,$54,$6B,$DC,$FF,$45,$4E,$67,$FC
	.dc.b	$5D,$D1,$FE,$30,$CF,$F8,$F7,$9B,$FE,$A9,$8F,$F1,$99,$FE,$38,$E2
	.dc.b	$C4,$4A,$FF,$16,$CE,$BD,$AF,$CC,$EB,$FE,$0D,$F9,$FF,$EF,$D5,$F3
	.dc.b	$F1,$46,$27,$5D,$FF,$EA,$0F,$9A,$FF,$5C,$4E,$78,$3F,$C9,$CD,$C3
	.dc.b	$FC,$88,$DD,$FF,$D9,$55,$FE,$B3,$FB,$4F,$F0,$ED,$FC,$06,$DF,$ED
	.dc.b	$1B,$4F,$90,$E5,$18,$27,$AF,$FE,$BD,$8E,$C7,$AA,$BF,$81,$B5,$FF
	.dc.b	$7D,$65,$FE,$B5,$8C,$47,$7F,$2B,$C8,$47,$F4,$9F,$6E,$FA,$5C,$47
	.dc.b	$FA,$37,$1F,$EF,$0F,$EA,$5A,$FF,$A4,$67,$83,$FE,$61,$1F,$6E,$FB
	.dc.b	$4F,$EC,$D1,$FE,$85,$8C,$70,$7F,$B8,$5B,$19,$9D,$BF,$EC,$98,$7C
	.dc.b	$B7,$D6,$8E,$4D,$74,$D7,$FD,$EF,$1C,$38,$7F,$D2,$50,$F9,$4A,$47
	.dc.b	$FD,$2C,$FF,$B6,$72,$FD,$87,$0F,$BF,$B3,$FC,$E3,$9F,$F9,$1F,$F4
	.dc.b	$E7,$C8,$62,$BE,$E5,$5C,$7E,$E3,$79,$18,$F4,$9F,$F5,$C6,$1C,$AF
	.dc.b	$B1,$FC,$57,$EE,$31,$3F,$D0,$5A,$FF,$40,$6B,$FA,$A6,$BF,$CF,$BE
	.dc.b	$E5,$CF,$E1,$6D,$75,$CF,$13,$23,$F9,$24,$EA,$BF,$D9,$97,$E2,$D3
	.dc.b	$FC,$CD,$39,$7F,$CE,$C7,$F4,$8A,$FF,$4D,$B6,$BF,$DA,$9F,$A7,$FF
	.dc.b	$9B,$E7,$3D,$BF,$17,$AB,$FC,$50,$DB,$E0,$7F,$9C,$EB,$85,$FE,$D6
	.dc.b	$33,$7F,$CC,$61,$EF,$3E,$C7,$E8,$46,$1B,$0F,$7A,$C4,$AF,$CB,$4F
	.dc.b	$16,$CE,$6B,$F9,$97,$CC,$FF,$49,$EA,$FF,$58,$C7,$86,$31,$5F,$ED
	.dc.b	$F6,$C7,$94,$38,$7F,$C3,$BC,$BF,$E3,$D9,$FF,$60,$B6,$3B,$F9,$1F
	.dc.b	$B1,$9C,$F9,$1F,$8B,$B7,$EA,$1F,$D4,$67,$C0,$DB,$F0,$4E,$9F,$39
	.dc.b	$FD,$5D,$C4,$F0,$8C,$9E,$69,$E4,$FB,$5A,$F1,$6B,$FE,$E1,$9F,$AB
	.dc.b	$AF,$FA,$DD,$B5,$B9,$FB,$D1,$87,$FD,$DF,$FA,$8B,$ED,$FC,$34,$8C
	.dc.b	$3F,$26,$77,$3A,$9C,$C7,$CD,$9E,$13,$92,$3C,$4C,$9F,$E1,$2F,$B9
	.dc.b	$FF,$47,$3F,$EF,$06,$CF,$F6,$F8,$DB,$FD,$0E,$33,$87,$FB,$17,$17
	.dc.b	$8B,$F2,$3A,$FE,$9B,$63,$38,$7A,$67,$2D,$8C,$E1,$FE,$3B,$6B,$FE
	.dc.b	$3A,$6B,$FD,$91,$AF,$F9,$77,$DE,$FD,$47,$A6,$7F,$86,$4F,$4C,$7F
	.dc.b	$DD,$13,$9B,$FF,$0B,$41,$FC,$46,$8F,$1F,$E6,$12,$3E,$44,$65,$B1
	.dc.b	$9C,$7E,$84,$E5,$B1,$F8,$AF,$33,$B7,$9D,$FC,$85,$B7,$E2,$3C,$33
	.dc.b	$97,$FC,$46,$66,$92,$FB,$D7,$CA,$F5,$6D,$7E,$16,$75,$39,$24,$66
	.dc.b	$77,$5F,$B3,$AD,$CE,$6B,$ED,$7C,$6F,$FB,$72,$3F,$C8,$8D,$AF,$F1
	.dc.b	$95,$70,$7F,$C6,$9B,$A7,$5A,$7F,$04,$E3,$FF,$3D,$E1,$F9,$9B,$FE
	.dc.b	$03,$E9,$FE,$B9,$B1,$99,$DC,$7F,$09,$9D,$46,$23,$F7,$71,$83,$96
	.dc.b	$1F,$EA,$5A,$FE,$DC,$D9,$FF,$24,$1D,$8A,$E6,$FF,$8C,$AF,$A4,$FE
	.dc.b	$4E,$DA,$FA,$5D,$6F,$49,$F3,$3A,$BF,$EB,$5F,$2E,$F9,$1F,$D1,$37
	.dc.b	$0F,$D0,$37,$3F,$F4,$C3,$5F,$F4,$BF,$D9,$7F,$7A,$9F,$F3,$93,$63
	.dc.b	$FC,$C1,$97,$F8,$23,$87,$FB,$44,$E7,$5F,$EE,$3F,$8D,$79,$30,$FE
	.dc.b	$0D,$08,$C6,$1F,$CC,$63,$1A,$7F,$1E,$6D,$70,$AD,$D7,$F4,$B9,$CB
	.dc.b	$C6,$32,$FE,$A9,$AF,$F5,$0F,$39,$F7,$9F,$DF,$F3,$FF,$0F,$BF,$E0
	.dc.b	$1B,$E6,$7E,$C6,$B8,$47,$F6,$1F,$B8,$FF,$54,$63,$FA,$1F,$21,$18
	.dc.b	$AF,$C8,$FA,$0F,$9D,$E1,$EE,$4C,$38,$5E,$F2,$ED,$D3,$FE,$5A,$BD
	.dc.b	$63,$FE,$5D,$FB,$3F,$E3,$78,$7F,$A6,$73,$46,$27,$F6,$39,$BE,$3F
	.dc.b	$FB,$B4,$E6,$B7,$F1,$98,$3F,$7D,$5B,$C3,$D5,$5C,$FA,$DF,$71,$FC
	.dc.b	$4F,$2F,$ED,$6D,$D7,$FB,$5E,$CF,$A4,$FE,$BD,$39,$C7,$FA,$C3,$6F
	.dc.b	$FB,$5B,$5F,$8F,$3C,$3C,$5F,$F0,$67,$CE,$73,$BF,$98,$DA,$F0,$CE
	.dc.b	$AB,$FC,$24,$FE,$CB,$1F,$AD,$B7,$F2,$DE,$FF,$6F,$D1,$E9,$99,$7D
	.dc.b	$7B,$E0,$FF,$B0,$F8,$FF,$D1,$35,$AC,$BF,$D1,$9F,$6C,$FB,$1F,$F1
	.dc.b	$03,$6F,$F9,$05,$FF,$A4,$3E,$0F,$35,$27,$F0,$95,$BB,$FC,$1D,$65
	.dc.b	$F9,$7F,$51,$FD,$EB,$8B,$E7,$B7,$84,$D2,$61,$FE,$17,$8C,$B6,$33
	.dc.b	$5F,$E1,$9E,$A7,$DC,$7F,$83,$78,$7F,$35,$18,$9F,$AF,$5F,$A1,$AD
	.dc.b	$5F,$E3,$61,$E0,$FF,$8F,$DB,$1F,$03,$9F,$FC,$94,$EE,$74,$47,$5E
	.dc.b	$D5,$C3,$FD,$6D,$C4,$FB,$7F,$B2,$7E,$4B,$D9,$B7,$F5,$62,$D7,$EA
	.dc.b	$A6,$EB,$FB,$3A,$7E,$0F,$E4,$AF,$F0,$63,$5F,$F2,$74,$6B,$2F,$F0
	.dc.b	$BC,$E6,$7F,$C6,$4D,$7F,$B7,$36,$78,$1F,$E1,$37,$5C,$EE,$97,$CC
	.dc.b	$FE,$60,$FF,$44,$57,$0B,$FF,$6A,$4E,$A3,$F8,$37,$D2,$FF,$AA,$DB
	.dc.b	$F0,$9F,$B2,$E8,$FF,$36,$4E,$6F,$FD,$B8,$DD,$3F,$3A,$DB,$FC,$C7
	.dc.b	$0F,$E3,$67,$38,$7F,$9D,$7E,$A7,$F5,$11,$FA,$46,$D7,$E7,$E1,$F5
	.dc.b	$FB,$EE,$93,$FC,$6E,$B7,$97,$F7,$CB,$DC,$E6,$BF,$D4,$9F,$0B,$FD
	.dc.b	$46,$BF,$CD,$1E,$7A,$31,$1F,$E7,$F8,$C7,$9D,$FE,$31,$9C,$C8,$FF
	.dc.b	$41,$F6,$BE,$B7,$F9,$A3,$71,$FF,$07,$61,$C9,$FE,$BF,$19,$AF,$EF
	.dc.b	$BF,$90,$3E,$B3,$FD,$70,$DA,$DE,$18,$99,$1D,$27,$55,$FD,$6A,$FA
	.dc.b	$69,$FD,$07,$15,$6A,$31,$97,$F2,$9B,$F8,$2D,$7E,$03,$5F,$75,$7C
	.dc.b	$5F,$23,$FC,$F3,$DE,$C7,$FD,$11,$5E,$1F,$80,$F9,$B5,$F0,$9F,$F6
	.dc.b	$97,$E6,$1F,$EC,$6C,$0D,$27,$C3,$FF,$A1,$3A,$9C,$7F,$57,$5F,$C1
	.dc.b	$71,$72,$AF,$C3,$33,$83,$F0,$0E,$5A,$4D,$15,$CA,$3E,$9E,$3F,$EC
	.dc.b	$D7,$FA,$A6,$1D,$DF,$7F,$D5,$FF,$61,$7F,$CD,$8D,$BF,$C3,$B5,$FC
	.dc.b	$33,$5E,$7D,$7F,$9B,$3F,$DE,$51,$C7,$0F,$CE,$47,$FB,$23,$ED,$3F
	.dc.b	$9D,$D3,$57,$E8,$0F,$B0,$FE,$F5,$F0,$DF,$CB,$2B,$F2,$17,$C2,$F7
	.dc.b	$5A,$F1,$BF,$DA,$D3,$9C,$39,$69,$F9,$7C,$79,$62,$AD,$61,$C6,$3C
	.dc.b	$57,$C1,$57,$87,$1E,$F8,$73,$C7,$BF,$9B,$FE,$0F,$D3,$F2,$18,$F1
	.dc.b	$C3,$BE,$6F,$98,$ED,$93,$EC,$BF,$84,$56,$A7,$31,$8C,$BD,$71,$8E
	.dc.b	$FF,$36,$BE,$98,$C6,$BE,$38,$F8,$B1,$87,$D3,$5C,$A7,$58,$F8,$6F
	.dc.b	$9E,$5E,$59,$CE,$3E,$1C,$39,$C6,$2B,$E3,$77,$BF,$12,$FD,$0E,$9F
	.dc.b	$17,$F5,$0F,$99,$7E,$3A,$D7,$4F,$C0,$9C,$E3,$E4,$3F,$8D,$E1,$CB
	.dc.b	$5F,$D7,$CE,$B1,$FC,$06,$1C,$B1,$F4,$CE,$6B,$F8,$94,$E7,$0E,$9A
	.dc.b	$AB,$51,$DB,$2F,$16,$BF,$2B,$EB,$FF,$CE,$9E,$7F,$C9,$A4,$61,$F3
	.dc.b	$7C,$C3,$E9,$7F,$D9,$7B,$7E,$2A,$B4,$DE,$B6,$1C,$E7,$3B,$7F,$27
	.dc.b	$C3,$97,$C4,$F8,$6F,$79,$7C,$33,$C2,$E1,$FE,$5C,$7F,$2F,$BF,$23
	.dc.b	$83,$78,$7F,$A8,$63,$31,$C1,$FF,$1F,$6F,$9F,$1B,$FA,$4C,$FF,$BF
	.dc.b	$E7,$2D,$8C,$BD,$63,$39,$FB,$E7,$8B,$6B,$F2,$19,$3E,$D5,$F4,$5E
	.dc.b	$8A,$D6,$3E,$DC,$79,$EB,$E5,$CB,$9B,$F1,$70,$E3,$5F,$89,$FD,$22
	.dc.b	$BF,$73,$8B,$E0,$63,$F8,$F8,$ED,$7F,$82,$AD,$CF,$E2,$1B,$1F,$8C
	.dc.b	$9D,$9F,$C3,$EB,$9C,$71,$CB,$3A,$2F,$8D,$EC,$E3,$C7,$F8,$D7,$71
	.dc.b	$A7,$F1,$3E,$1E,$BA,$1F,$77,$4F,$0C,$E8,$FC,$34,$E5,$FD,$8C,$6B
	.dc.b	$7F,$F4,$24,$68,$C9,$0F,$99,$3B,$39,$3F,$A2,$AF,$0B,$C4,$FD,$2F
	.dc.b	$5B,$EC,$FD,$6D,$8F,$E0,$AD,$7E,$C7,$8D,$FE,$B7,$EA,$FC,$63,$63
	.dc.b	$F9,$3D,$72,$3D,$11,$98,$FD,$5B,$BD,$3F,$9E,$CE,$E7,$F4,$D3,$CF
	.dc.b	$C6,$FE,$C3,$F4,$97,$C5,$34,$E5,$E7,$FF,$97,$AF,$FC,$1A,$D7,$FC
	.dc.b	$36,$D7,$FC,$34,$D7,$FB,$DB,$55,$7F,$70,$CB,$C0,$48,$FF,$0C,$FE
	.dc.b	$65,$FE,$CB,$7F,$B5,$C7,$F6,$83,$4E,$0E,$78,$6F,$C3,$FF,$09,$55
	.dc.b	$AE,$15,$FE,$37,$78,$CE,$9F,$8D,$3D,$E7,$FC,$64,$D7,$FC,$26,$D7
	.dc.b	$FB,$43,$7C,$A7,$D1,$BE,$CB,$EB,$46,$1B,$19,$AF,$F1,$AF,$E1,$AF
	.dc.b	$B1,$FA,$CD,$8F,$E1,$C7,$F7,$57,$0F,$F9,$21,$AF,$F9,$1D,$AF,$F8
	.dc.b	$A3,$F2,$37,$E3,$3F,$81,$6C,$7F,$91,$1A,$FC,$9C,$7D,$3E,$3E,$EA
	.dc.b	$4E,$DF,$4C,$48,$FF,$29,$6B,$FE,$A6,$C9,$39,$9F,$A9,$C8,$CB,$BD
	.dc.b	$FF,$91,$5F,$0C,$FA,$70,$F7,$BE,$F6,$D7,$17,$3C,$1F,$EB,$EF,$2B
	.dc.b	$F3,$1D,$DF,$DF,$6F,$C7,$C3,$FE,$B3,$9F,$D1,$7E,$1D,$FE,$90,$9C
	.dc.b	$9F,$90,$AE,$0F,$E2,$6F,$A3,$A6,$BF,$8D,$13,$CD,$FF,$47,$FC,$87
	.dc.b	$EA,$7D,$83,$4F,$D9,$7C,$9B,$D2,$C5,$7F,$A7,$26,$BD,$67,$4F,$F9
	.dc.b	$E3,$ED,$DF,$82,$7E,$E5,$B1,$98,$EB,$E9,$48,$F9,$34,$8F,$97,$08
	.dc.b	$C3,$A8,$CC,$62,$3F,$1E,$D7,$FA,$F3,$5F,$EC,$6D,$7F,$73,$F0,$3D
	.dc.b	$5F,$90,$D7,$FD,$B3,$1F,$0B,$E5,$3F,$87,$4E,$79,$1C,$3F,$E9,$BD
	.dc.b	$D3,$98,$C3,$70,$FF,$50,$60,$C3,$E8,$7D,$2C,$9F,$F6,$3B,$F4,$3E
	.dc.b	$3B,$11,$FE,$C4,$8B,$97,$F0,$BB,$57,$ED,$BE,$DF,$F5,$B3,$FF,$00
	.dc.b	$1B,$97,$FA,$BF,$A9,$18,$C7,$C7,$F4,$1F,$78,$F9,$64,$7E,$9B,$5F
	.dc.b	$9A,$D7,$E6,$BF,$9C,$F2,$1F,$51,$FC,$CD,$B3,$F9,$36,$BF,$93,$FB
	.dc.b	$4F,$D5,$E3,$F2,$67,$39,$7C,$CA,$FE,$5B,$18,$9D,$5F,$F4,$5B,$E3
	.dc.b	$5F,$6A,$3E,$A9,$1F,$D3,$88,$7C,$8E,$63,$F5,$1C,$1D,$5E,$47,$09
	.dc.b	$DB,$FA,$BF,$C3,$9F,$75,$FD,$3D,$B1,$98,$F2,$7D,$8C,$8E,$4D,$7F
	.dc.b	$65,$F8,$D3,$DA,$C6,$DF,$D1,$71,$5E,$35,$FF,$1C,$DE,$F0,$FD,$E7
	.dc.b	$C0,$BE,$5B,$FB,$03,$63,$31,$F1,$7E,$07,$2C,$8C,$9E,$D8,$FE,$D1
	.dc.b	$F7,$2F,$90,$A4,$9F,$85,$DF,$5F,$E1,$D3,$B7,$F8,$73,$5F,$B9,$1F
	.dc.b	$DF,$75,$7D,$A7,$F8,$C9,$FF,$4F,$37,$4F,$E6,$2D,$A1,$F5,$F8,$BE
	.dc.b	$F1,$FC,$0B,$7F,$D6,$C7,$F2,$E6,$BF,$D3,$9A,$FF,$97,$5A,$FF,$98
	.dc.b	$9A,$F6,$6B,$DC,$FE,$AF,$E3,$BE,$C7,$FD,$95,$1F,$CC,$23,$1F,$0D
	.dc.b	$FE,$13,$9C,$CF,$F5,$1E,$AF,$C9,$D7,$F6,$9F,$BF,$FE,$83,$E8,$1A
	.dc.b	$4A,$FE,$8F,$FA,$23,$E0,$BF,$C4,$0D,$8F,$CC,$B5,$FE,$BC,$BD,$5D
	.dc.b	$FE,$01,$27,$6F,$2F,$62,$47,$67,$F9,$5D,$1A,$48,$D1,$0F,$EF,$31
	.dc.b	$B7,$E2,$FE,$31,$FE,$27,$E3,$FD,$B6,$88,$C7,$1F,$DA,$8C,$7C,$A7
	.dc.b	$F8,$FF,$E6,$9F,$04,$FB,$DC,$37,$2F,$F0,$3C,$EA,$7F,$B5,$B6,$33
	.dc.b	$3B,$7F,$D3,$51,$96,$C6,$5F,$F5,$0C,$65,$B1,$97,$F1,$F1,$FC,$EA
	.dc.b	$3C,$D8,$7F,$A6,$B2,$FE,$25,$39,$AF,$E1,$7F,$A9,$3E,$53,$FC,$AC
	.dc.b	$D8,$FF,$00,$B5,$FF,$10,$B5,$FF,$58,$35,$FF,$5A,$B5,$FD,$0B,$53
	.dc.b	$4D,$79,$CF,$C4,$3F,$CD,$D8,$7F,$27,$8C,$7C,$0F,$F9,$8E,$73,$97
	.dc.b	$D3,$FA,$CF,$F0,$05,$FD,$87,$FC,$05,$8F,$FB,$1B,$74,$62,$38,$78
	.dc.b	$7F,$E1,$A8,$C4,$27,$73,$92,$47,$EA,$A7,$79,$7F,$99,$B7,$FF,$3B
	.dc.b	$3F,$67,$0F,$43,$71,$FF,$6B,$FC,$E7,$F7,$B9,$FF,$34,$3F,$DD,$F0
	.dc.b	$34,$9E,$9F,$F8,$19,$B1,$FE,$32,$6B,$FE,$DF,$CD,$ED,$7F,$17,$BF
	.dc.b	$CC,$B7,$0F,$CC,$36,$BF,$DE,$7F,$01,$FE,$71,$6C,$FF,$8F,$BE,$3B
	.dc.b	$FC,$C0,$D9,$FF,$36,$46,$27,$FC,$C2,$D7,$FD,$E0,$D7,$FD,$EA,$D7
	.dc.b	$F9,$0B,$5F,$F4,$6F,$1F,$F0,$97,$FC,$78,$D7,$FC,$80,$DC,$3E,$3F
	.dc.b	$A9,$FE,$A9,$CF,$FC,$35,$5F,$EF,$F6,$BF,$3A,$3E,$41,$FE,$2E,$EF
	.dc.b	$FE,$57,$5F,$E9,$76,$E5,$FC,$D2,$BF,$C1,$3E,$1F,$FA,$D6,$5E,$E5
	.dc.b	$89,$3F,$DC,$27,$8B,$FC,$F1,$DB,$FE,$55,$F9,$8F,$F1,$13,$7D,$8F
	.dc.b	$F5,$D4,$62,$3F,$CE,$3F,$45,$FE,$48,$7F,$CB,$D1,$9C,$BF,$40,$DC
	.dc.b	$BF,$C6,$4D,$9F,$CB,$57,$1A,$FE,$2F,$F6,$6F,$92,$FF,$66,$36,$BF
	.dc.b	$9F,$35,$FF,$40,$FD,$BF,$1C,$7F,$8E,$4E,$6F,$9D,$F2,$3F,$DA,$D5
	.dc.b	$BB,$FD,$33,$E9,$7F,$5C,$D7,$F9,$33,$5D,$FB,$BF,$E8,$8F,$B7,$FF
	.dc.b	$1E,$C7,$F5,$16,$BF,$E9,$46,$DF,$F3,$2F,$D4,$FF,$B5,$73,$A7,$39
	.dc.b	$FF,$33,$6D,$FE,$E5,$B2,$AF,$EE,$52,$FA,$77,$A6,$C7,$FD,$3B,$18
	.dc.b	$6C,$66,$7F,$D7,$F3,$96,$FC,$03,$83,$8B,$63,$31,$83,$FC,$FE,$BE
	.dc.b	$52,$91,$FF,$41,$BF,$DD,$BE,$E1,$F3,$9F,$ED,$DD,$11,$8D,$B5,$39
	.dc.b	$8C,$5F,$F0,$36,$BF,$C8,$5A,$FF,$59,$6B,$FC,$7D,$BF,$04,$FC,$0B
	.dc.b	$FC,$97,$7F,$ED,$E8,$C7,$C0,$7F,$BB,$DB,$19,$9D,$DF,$F1,$3F,$B9
	.dc.b	$EB,$91,$A4,$CB,$D8,$DB,$FF,$0A,$B6,$BF,$8E,$7D,$CF,$FB,$DE,$7F
	.dc.b	$C3,$2D,$8C,$FA,$BF,$16,$31,$1F,$CF,$DA,$FF,$77,$6B,$FC,$F5,$B8
	.dc.b	$1C,$C3,$27,$C2,$FF,$9D,$E3,$F8,$31,$97,$0F,$F0,$98,$DE,$1F,$EE
	.dc.b	$66,$C7,$C9,$9D,$BF,$E9,$76,$BC,$E7,$84,$62,$5F,$82,$D2,$4F,$57
	.dc.b	$9F,$86,$FB,$9F,$F7,$EC,$FF,$34,$6E,$1F,$CD,$1A,$FF,$34,$6B,$E9
	.dc.b	$FA,$2F,$01,$27,$D4,$D8,$FC,$13,$5F,$87,$82,$3C,$7E,$5F,$F9,$7A
	.dc.b	$BF,$A2,$B6,$BF,$A2,$B5,$FE,$8A,$D7,$F9,$97,$E2,$2F,$B9,$F8,$AD
	.dc.b	$AF,$A3,$3F,$E6,$A8,$C3,$86,$BF,$59,$AF,$49,$E6,$49,$E6,$3E,$DF
	.dc.b	$73,$FE,$E0,$6C,$66,$7E,$CB,$4F,$1C,$A9,$F1,$DE,$82,$4E,$A3,$2D
	.dc.b	$8C,$CF,$A5,$AF,$F8,$5F,$0F,$E7,$53,$FE,$0F,$F0,$7F,$81,$46,$27
	.dc.b	$FC,$00,$D7,$FC,$8F,$B7,$E3,$3C,$AF,$C4,$DF,$F9,$3A,$30,$47,$04
	.dc.b	$9F,$F2,$AC,$7B,$B3,$79,$34,$95,$F2,$23,$2F,$C2,$EA,$FF,$5C,$37
	.dc.b	$E1,$7F,$3B,$18,$C3,$EC,$B5,$FD,$5B,$5F,$EE,$4D,$7F,$D4,$ED,$7F
	.dc.b	$94,$47,$EA,$B1,$7D,$B3,$ED,$70,$DC,$3F,$CD,$ED,$BF,$F3,$5B,$5F
	.dc.b	$F0,$BF,$B3,$FE,$1C,$24,$FF,$88,$5B,$1F,$E1,$E6,$BC,$23,$FC,$4D
	.dc.b	$E8,$FF,$89,$08,$E0,$FE,$51,$F0,$AF,$B4,$FE,$50,$D8,$CC,$60,$FE
	.dc.b	$07,$D7,$FE,$E9,$8F,$C6,$B6,$FF,$D7,$3F,$17,$A3,$F9,$97,$8B,$EB
	.dc.b	$8F,$F6,$6F,$4F,$F5,$DD,$72,$4C,$3F,$2F,$F9,$67,$ED,$DB,$D5,$FC
	.dc.b	$4A,$31,$7F,$AE,$6B,$FC,$81,$AF,$F8,$9D,$AF,$FB,$71,$A9,$CF,$90
	.dc.b	$FB,$5F,$81,$3F,$CA,$A3,$1A,$FF,$24,$6E,$1F,$BE,$6C,$FF,$A8,$BC
	.dc.b	$7F,$C2,$71,$FF,$6F,$E4,$FC,$14,$60,$8E,$09,$7F,$EB,$08,$D7,$D0
	.dc.b	$70,$24,$65,$B7,$DD,$AF,$FA,$BA,$51,$FC,$E3,$ED,$3F,$DB,$95,$FC
	.dc.b	$09,$B5,$FC,$59,$AF,$F4,$86,$BF,$E4,$F6,$A0,$E1,$BF,$0D,$DC,$FE
	.dc.b	$7F,$F6,$4F,$5A,$92,$FF,$CE,$71,$97,$E0,$7D,$A7,$C6,$23,$82,$5F
	.dc.b	$CD,$8C,$B6,$3E,$47,$C3,$7C,$E2,$38,$24,$7D,$58,$CB,$63,$E2,$4F
	.dc.b	$C7,$23,$82,$47,$C7,$8C,$B6,$3D,$7E,$CF,$C0,$23,$83,$EE,$7C,$7F
	.dc.b	$E0,$88,$E0,$93,$F8,$98,$CB,$63,$F0,$9F,$65,$F8,$A2,$38,$24,$7E
	.dc.b	$46,$32,$D8,$FC,$07,$17,$E0,$C8,$E0,$91,$F8,$38,$CB,$63,$FA,$8F
	.dc.b	$D9,$3B,$BF,$A6,$B6,$33,$18,$3F,$3F,$F1,$72,$3F,$38,$DA,$FD,$33
	.dc.b	$5F,$CC,$35,$FC,$77,$27,$F9,$7F,$E2,$3F,$95,$73,$1A,$4B,$FE,$DF
	.dc.b	$3B,$6C,$ED,$B3,$BA,$FE,$EB,$F6,$5F,$C4,$DF,$CE,$35,$34,$D7,$7F
	.dc.b	$F8,$27,$BF,$FC,$BD,$D9,$FC,$7E,$BF,$DF,$0D,$7F,$4D,$3B,$C3,$FC
	.dc.b	$A9,$C6,$F8,$9F,$F2,$A3,$63,$31,$82,$38,$25,$7F,$86,$A7,$83,$FE
	.dc.b	$1C,$FC,$75,$F8,$17,$F9,$7D,$B5,$FE,$63,$6C,$EE,$7D,$3D,$72,$BF
	.dc.b	$C2,$F8,$FF,$89,$63,$15,$FE,$1A,$FB,$07,$CD,$7F,$90,$9B,$7F,$E0
	.dc.b	$C6,$CE,$E7,$FD,$05,$F6,$5F,$E6,$C2,$7B,$18,$25,$F8,$23,$2D,$8C
	.dc.b	$D7,$F9,$6B,$15,$6A,$30,$FF,$A1,$38,$5F,$86,$7F,$86,$27,$FC,$C3
	.dc.b	$18,$F4,$7F,$CB,$6D,$C7,$FC,$6A,$D9,$ED,$3F,$DE,$F0,$FF,$12,$F8
	.dc.b	$7F,$E9,$B8,$C1,$FD,$F7,$47,$95,$F0,$C9,$8F,$C3,$6E,$1F,$25,$AF
	.dc.b	$F9,$D2,$78,$CE,$CF,$E5,$9C,$C7,$F1,$3F,$54,$F9,$6C,$5F,$FA,$CF
	.dc.b	$C0,$FE,$64,$B0,$E6,$FF,$B1,$6F,$8D,$70,$F3,$BF,$DF,$F3,$BF,$8B
	.dc.b	$E7,$5E,$36,$C7,$4B,$46,$23,$FA,$04,$62,$7F,$CC,$4D,$8F,$F4,$6B
	.dc.b	$63,$A1,$A9,$E5,$3E,$18,$FF,$13,$E1,$FA,$6F,$AB,$FF,$6E,$C6,$27
	.dc.b	$BD,$7A,$6B,$5F,$05,$67,$35,$BE,$06,$1C,$A7,$E3,$1A,$2C,$6A,$B9
	.dc.b	$70,$BE,$CF,$FD,$23,$87,$CC,$8C,$7D,$3F,$C1,$6C,$7F,$A0,$DB,$87
	.dc.b	$FA,$BD,$B1,$C6,$B9,$CF,$F6,$79,$CB,$63,$2F,$F1,$B7,$FD,$43,$AB
	.dc.b	$3F,$E1,$3F,$61,$F7,$70,$F8,$98,$78,$2B,$58,$33,$5E,$E3,$E0,$8F
	.dc.b	$A2,$78,$0F,$7B,$E7,$4E,$63,$0F,$C8,$FB,$27,$D9,$FF,$9E,$72,$FC
	.dc.b	$4E,$1C,$B0,$E2,$DB,$EE,$D9,$F2,$DF,$2F,$77,$EC,$57,$7B,$E3,$8F
	.dc.b	$82,$3F,$DA,$F8,$73,$6C,$65,$FC,$FC,$65,$BE,$87,$F4,$37,$FD,$5F
	.dc.b	$08,$C4,$74,$6B,$C3,$5F,$F5,$1C,$ED,$F5,$CF,$3C,$BD,$ED,$DF,$FC
	.dc.b	$41,$9B,$EC,$DF,$67,$FE,$82,$9E,$99,$7C,$2B,$E0,$FF,$89,$F2,$EC
	.dc.b	$FC,$DB,$E2,$FD,$CF,$B3,$AF,$E5,$B8,$3F,$D1,$F8,$7E,$0A,$3B,$1F
	.dc.b	$D8,$67,$79,$FE,$0B,$44,$72,$E2,$FC,$D3,$6B,$F2,$4D,$79,$69,$FC
	.dc.b	$2A,$76,$D8,$DB,$5D,$9F,$A1,$9E,$4F,$E3,$7C,$3F,$2F,$1F,$EC,$BE
	.dc.b	$CE,$F4,$1F,$4D,$6C,$FC,$4D,$6A,$24,$60,$87,$F2,$03,$E5,$47,$8D
	.dc.b	$E4,$64,$B8,$FE,$2A,$35,$BF,$F9,$8A,$7F,$DF,$0D,$78,$36,$7F,$23
	.dc.b	$E3,$FF,$2C,$EE,$FE,$5F,$93,$B9,$F3,$7E,$E6,$E3,$FA,$2B,$E1,$87
	.dc.b	$E5,$9F,$F7,$7C,$7E,$C5,$B8,$9A,$48,$FC,$9B,$5F,$CD,$BF,$EA,$CF
	.dc.b	$37,$FC,$61,$1F,$DD,$9A,$FF,$65,$6B,$FD,$13,$14,$62,$3F,$89,$7E
	.dc.b	$77,$FE,$E6,$7F,$B0,$B5,$FE,$CE,$D7,$F9,$13,$5F,$F0,$DF,$D9,$7F
	.dc.b	$B0,$09,$8F,$FB,$01,$B1,$FE,$0A,$6B,$FD,$E5,$AF,$95,$AF,$48,$F8
	.dc.b	$FF,$71,$FB,$68,$FE,$BA,$D7,$FD,$20,$D4,$E7,$07,$D8,$BE,$6F,$FD
	.dc.b	$A6,$DA,$FF,$0E,$B5,$FF,$06,$35,$FE,$D7,$F0,$0F,$C0,$9E,$D7,$F6
	.dc.b	$E9,$CC,$61,$C0,$9E,$13,$88,$C4,$ED,$E5,$2F,$B3,$78,$BB,$8C,$36
	.dc.b	$33,$3D,$A7,$2D,$8C,$BD,$1A,$FA,$9A,$FF,$BB,$9A,$FF,$47,$C1,$F6
	.dc.b	$7F,$18,$91,$F1,$9F,$A7,$F3,$72,$3D,$9A,$A3,$11,$F0,$BE,$CD,$F4
	.dc.b	$5F,$A0,$6C,$7C,$36,$BB,$8D,$11,$D3,$5F,$9B,$F1,$6F,$43,$F5,$6D
	.dc.b	$8F,$9C,$D7,$71,$A2,$3A,$6B,$D3,$D0,$E6,$47,$9B,$FB,$8F,$C3,$BF
	.dc.b	$51,$1F,$ED,$F6,$A6,$93,$DD,$E0,$48,$FC,$14,$70,$B7,$CF,$FA,$89
	.dc.b	$3E,$A6,$BF,$01,$AF,$E4,$63,$8B,$F0,$3E,$CB,$E1,$91,$C1,$0F,$92
	.dc.b	$E5,$A7,$C8,$F9,$4F,$96,$47,$07,$A3,$D9,$F8,$44,$70,$43,$F4,$38
	.dc.b	$23,$51,$87,$F9,$0F,$D6,$7E,$3A,$3F,$48,$D7,$F5,$2D,$7F,$5C,$D7
	.dc.b	$FC,$BA,$D4,$1C,$37,$87,$B9,$E0,$F9,$A9,$1E,$06,$BE,$06,$BE,$06
	.dc.b	$BE,$0F,$B0,$F4,$91,$F4,$BF,$CF,$3E,$C3,$FC,$57,$1F,$BF,$6B,$FC
	.dc.b	$11,$AF,$F0,$96,$BF,$E7,$86,$BE,$C6,$A6,$98,$B7,$AF,$F3,$09,$3F
	.dc.b	$31,$AF,$CC,$6B,$F3,$1F,$A7,$F2,$32,$3E,$9B,$5F,$A6,$D7,$E9,$B5
	.dc.b	$FA,$7F,$17,$F7,$08,$FD,$C7,$E2,$78,$F2,$3E,$23,$5F,$88,$D7,$E2
	.dc.b	$35,$FB,$BD,$6F,$A4,$FC,$03,$75,$FE,$4D,$18,$3F,$03,$F5,$7F,$E3
	.dc.b	$FA,$FE,$60,$D7,$F9,$AB,$5F,$E7,$4D,$7B,$BF,$D9,$9F,$E5,$CD,$7F
	.dc.b	$9B,$B5,$FE,$8A,$D7,$FD,$62,$FF,$7B,$7C,$3E,$3F,$F8,$02,$31,$1E
	.dc.b	$28,$C7,$C5,$48,$F3,$B5,$F3,$B5,$F3,$B5,$FD,$0F,$C4,$BD,$27,$D4
	.dc.b	$7F,$4F,$8C,$36,$33,$3F,$D5,$A7,$2F,$F5,$5D,$BF,$C5,$0F,$FB,$29
	.dc.b	$AF,$EB,$9A,$9A,$4F,$23,$FC,$3A,$D8,$FE,$C1,$F5,$4F,$89,$FF,$10
	.dc.b	$B6,$33,$18,$38,$FC,$74,$8F,$23,$5F,$ED,$B4,$8C,$47,$F7,$38,$C3
	.dc.b	$C0,$FE,$E7,$E5,$BE,$67,$FA,$7C,$7F,$88,$DA,$FF,$0A,$8D,$78,$7F
	.dc.b	$DD,$DB,$1F,$DE,$7F,$4C,$7D,$F7,$F8,$E9,$B1,$98,$C1,$1C,$10,$FF
	.dc.b	$02,$C7,$07,$FC,$0F,$F5,$4F,$17,$FC,$8A,$D8,$FF,$04,$35,$DC,$7F
	.dc.b	$92,$BE,$0D,$F5,$F4,$93,$FE,$0F,$8C,$B6,$32,$FF,$85,$A3,$2D,$8F
	.dc.b	$F0,$B7,$C2,$7F,$89,$F8,$6F,$A5,$F8,$28,$C5,$7F,$87,$9A,$FF,$88
	.dc.b	$3F,$15,$FF,$30,$B7,$B7,$FC,$BF,$18,$8F,$F1,$43,$5F,$E7,$21,$C3
	.dc.b	$FC,$F1,$AE,$63,$99,$F1,$BC,$89,$1F,$19,$AF,$8F,$8F,$C2,$47,$FC
	.dc.b	$65,$5F,$86,$7F,$C6,$FD,$9F,$EE,$86,$FC,$07,$F9,$DA,$31,$1F,$E3
	.dc.b	$F6,$BE,$86,$BF,$D4,$9A,$FF,$17,$8E,$47,$E2,$FC,$E7,$D5,$52,$4F
	.dc.b	$E5,$A3,$2D,$8C,$BF,$E4,$FE,$4F,$F1,$74,$62,$3F,$CF,$9F,$57,$C1
	.dc.b	$D3,$5C,$C7,$E6,$E3,$11,$F9,$F6,$A9,$46,$1B,$3C,$4F,$F2,$A7,$E3
	.dc.b	$0F,$BB,$A4,$BF,$F2,$A4,$65,$FD,$8F,$D6,$7F,$B6,$9F,$F4,$93,$5F
	.dc.b	$D8,$35,$F1,$36,$CD,$27,$C2,$7E,$D5,$E1,$2F,$A0,$FE,$C7,$F0,$3F
	.dc.b	$E6,$F6,$C7,$F9,$D1,$AF,$F9,$E5,$AF,$E4,$97,$D6,$3E,$7B,$93,$67
	.dc.b	$93,$5E,$4D,$7F,$D0,$3F,$11,$FE,$BE,$4D,$23,$F6,$FE,$B3,$FC,$4D
	.dc.b	$D3,$FC,$59,$B1,$9F,$4F,$F9,$9E,$A9,$CC,$FF,$AE,$DA,$FF,$19,$6B
	.dc.b	$F6,$DB,$1C,$67,$F9,$EF,$9A,$FA,$C7,$A9,$86,$E3,$FE,$8D,$6E,$3F
	.dc.b	$E9,$66,$BF,$EC,$BF,$AC,$EA,$4C,$3C,$ED,$AF,$E6,$2D,$9F,$CD,$3E
	.dc.b	$4F,$AB,$7D,$67,$B9,$B5,$FC,$E1,$B5,$FA,$16,$BF,$ED,$B6,$BF,$CE
	.dc.b	$5A,$FE,$6D,$B9,$9C,$C7,$FA,$D7,$D9,$FE,$BB,$A1,$A4,$BF,$F5,$93
	.dc.b	$67,$FD,$60,$D7,$FD,$73,$F5,$3C,$89,$3F,$EB,$97,$FB,$5F,$82,$FA
	.dc.b	$CF,$F5,$EC,$FF,$29,$6D,$7F,$B2,$E7,$5F,$09,$C2,$73,$3F,$EC,$6F
	.dc.b	$A0,$F3,$92,$FF,$D8,$CD,$8F,$F6,$23,$5E,$11,$FC,$DF,$CD,$FF,$33
	.dc.b	$B6,$3C,$B1,$87,$C4,$D7,$FD,$BD,$F5,$6F,$AC,$FF,$01,$B6,$BF,$DB
	.dc.b	$8D,$9F,$F7,$0B,$5E,$33,$C0,$93,$C1,$FF,$07,$AF,$AC,$E1,$3E,$66
	.dc.b	$A8,$D4,$F0,$9C,$CA,$BF,$D8,$95,$E4,$F8,$4F,$39,$1C,$12,$7C,$F1
	.dc.b	$97,$F9,$76,$0F,$33,$F6,$2D,$AF,$F0,$5B,$67,$73,$FE,$26,$F5,$5B
	.dc.b	$D2,$5F,$F8,$CD,$B5,$FE,$31,$6B,$FD,$88,$4F,$29,$C4,$7C,$1C,$9E
	.dc.b	$57,$C1,$24,$E8,$F9,$4F,$F8,$E3,$AD,$F3,$D4,$93,$D2,$FF,$8A,$C6
	.dc.b	$1F,$E3,$2D,$7C,$93,$CA,$7F,$C9,$1F,$58,$FA,$CF,$F2,$3E,$48,$C4
	.dc.b	$EA,$73,$18,$BF,$F0,$47,$C3,$FF,$82,$C9,$5F,$E1,$06,$C6,$A3,$04
	.dc.b	$7F,$C2,$1E,$97,$F8,$4C,$91,$FE,$18,$7F,$8E,$FA,$EF,$72,$92,$7E
	.dc.b	$5F,$57,$F9,$66,$3F,$CB,$BD,$8F,$9B,$A4,$AF,$F2,$13,$FC,$A3,$C6
	.dc.b	$F5,$65,$FE,$18,$C3,$FA,$86,$B9,$25,$FC,$DF,$AC,$75,$7F,$9A,$9B
	.dc.b	$19,$8C,$11,$C1,$2F,$F0,$33,$C1,$FC,$37,$90,$FA,$CF,$75,$7E,$1A
	.dc.b	$31,$AF,$F9,$E9,$B1,$99,$DD,$7E,$3F,$8B,$FA,$F7,$27,$DA,$F2,$7E
	.dc.b	$F4,$E5,$AF,$F8,$DA,$77,$7F,$E8,$99,$CC,$FF,$A4,$9A,$FE,$69,$B3
	.dc.b	$C2,$3A,$7C,$B7,$72,$57,$36,$C6,$A3,$04,$7F,$47,$F5,$8F,$9D,$FF
	.dc.b	$4A,$B6,$7F,$38,$D9,$DC,$E8,$93,$A6,$C7,$F3,$1F,$AC,$7C,$3F,$F3
	.dc.b	$86,$D7,$F9,$7D,$B3,$B9,$F9,$1F,$29,$2B,$E2,$36,$7E,$53,$5F,$95
	.dc.b	$9A,$31,$1F,$27,$E1,$BE,$91,$2B,$E7,$36,$3E,$A3,$5E,$6F,$C2,$99
	.dc.b	$3F,$09,$B3,$F2,$1A,$FF,$0F,$CB,$C5,$3B,$3D,$DE,$C7,$C1,$23,$83
	.dc.b	$F0,$1E,$DF,$E1,$C8,$E0,$95,$F8,$98,$CB,$63,$F1,$1F,$59,$F8,$F2
	.dc.b	$38,$24,$7E,$46,$32,$D8,$FB,$BC,$7F,$80,$23,$82,$47,$E1,$A3,$2D
	.dc.b	$8F,$E4,$DF,$58,$EE,$FE,$A2,$D8,$CC,$60,$FD,$27,$CC,$C8,$FD,$03
	.dc.b	$6B,$F4,$CD,$7E,$EB,$5F,$F7,$AF,$37,$FB,$D6,$7F,$DE,$0D,$8F,$D9
	.dc.b	$35,$FD,$73,$5E,$3F,$11,$FC,$BA,$31,$1F,$D1,$9B,$CA,$73,$0F,$AC
	.dc.b	$FE,$2E,$FF,$AB,$9A,$8D,$75,$FF,$BF,$A3,$3D,$FF,$C6,$3B,$FF,$CA
	.dc.b	$3D,$BF,$C9,$73,$FF,$75,$B7,$0F,$F4,$A3,$5F,$E4,$9C,$BF,$EF,$F9
	.dc.b	$FE,$5E,$D8,$FE,$BD,$DC,$8C,$47,$0F,$88,$FF,$2A,$C6,$23,$CD,$3B
	.dc.b	$9C,$92,$BF,$9D,$FD,$63,$EB,$3F,$CB,$AD,$8C,$C6,$0F,$5F,$74,$8F
	.dc.b	$5B,$70,$FE,$DF,$F5,$0F,$AC,$FF,$08,$46,$1A,$FF,$77,$8D,$57,$F6
	.dc.b	$B6,$C7,$FA,$DF,$EB,$3F,$CE,$24,$F6,$7F,$CF,$6D,$AF,$EF,$BC,$08
	.dc.b	$C5,$71,$AD,$92,$B7,$3A,$9F,$F7,$4F,$2B,$EB,$3F,$C4,$31,$FE,$0B
	.dc.b	$8C,$7A,$3E,$88,$FC,$04,$62,$FF,$C1,$AD,$8F,$F6,$74,$FF,$3A,$C4
	.dc.b	$D2,$78,$BC,$E7,$F1,$59,$BC,$AC,$42,$31,$8F,$C5,$9C,$C6,$27,$E7
	.dc.b	$C6,$09,$3B,$24,$7E,$75,$FC,$C6,$0F,$A2,$7D,$67,$D6,$F3,$BF,$DF
	.dc.b	$58,$F5,$EB,$FB,$37,$BA,$D7,$87,$FA,$1C,$BC,$DF,$59,$F7,$B3,$FB
	.dc.b	$8D,$8D,$62,$8C,$47,$C0,$6B,$CA,$7F,$CD,$0D,$3E,$84,$F1,$8C,$B6
	.dc.b	$33,$3E,$47,$FD,$E5,$F4,$CF,$97,$C4,$73,$D1,$8F,$C2,$AD,$63,$FB
	.dc.b	$B9,$F8,$38,$73,$CD,$5A,$C3,$FD,$A7,$87,$B2,$B5,$39,$BE,$B4,$E1
	.dc.b	$7D,$67,$FA,$5B,$2F,$C7,$B6,$7F,$16,$D7,$83,$67,$8F,$B7,$F2,$1B
	.dc.b	$5F,$E8,$D6,$D7,$EB,$1B,$1D,$A7,$D5,$F5,$3F,$8B,$8F,$76,$7E,$09
	.dc.b	$CE,$DF,$1B,$E9,$FF,$C7,$5E,$F6,$B2,$FE,$3D,$18,$7E,$08,$FA,$B7
	.dc.b	$D6,$7F,$9E,$F6,$F8,$32,$AE,$11,$86,$E5,$F8,$96,$CF,$3A,$E5,$39
	.dc.b	$F1,$18,$F8,$EB,$A4,$61,$FF,$3E,$DF,$95,$FF,$41,$35,$E5,$3F,$E7
	.dc.b	$78,$FF,$10,$6D,$7D,$3F,$AD,$5C,$9A,$F2,$6B,$E7,$77,$97,$AD,$B8
	.dc.b	$E6,$BE,$46,$6F,$AC,$7C,$5F,$E9,$F0,$E1,$5E,$BC,$BC,$79,$7C,$C3
	.dc.b	$E4,$65,$D1,$F9,$D8,$72,$7E,$A5,$3E,$B3,$11,$F5,$39,$9C,$7F,$E9
	.dc.b	$0D,$7F,$7E,$D9,$FC,$A5,$70,$CF,$F2,$F8,$73,$D3,$F7,$9C,$7F,$C9
	.dc.b	$2B,$AB,$63,$37,$FC,$8A,$72,$D8,$CB,$D9,$B5,$D6,$3F,$31,$3C,$1F
	.dc.b	$C6,$F8,$5F,$17,$2F,$DA,$7D,$65,$9F,$DD,$B5,$FC,$64,$DD,$CE,$F5
	.dc.b	$FD,$BD,$5E,$73,$B9,$97,$FC,$3A,$F8,$B6,$3D,$2F,$63,$04,$3F,$A1
	.dc.b	$47,$06,$C7,$08,$FE,$E9,$1F,$E0,$5E,$BF,$E5,$3D,$8F,$95,$F3,$36
	.dc.b	$7F,$05,$E4,$FF,$2B,$AF,$EC,$B7,$CA,$7F,$3B,$3F,$EC,$F7,$F5,$8D
	.dc.b	$8D,$47,$E8,$23,$04,$7F,$40,$FF,$B5,$F3,$79,$DF,$E3,$99,$FE,$D2
	.dc.b	$D7,$FD,$D5,$1A,$F3,$F8,$4E,$67,$FB,$3F,$D6,$7F,$BA,$9F,$EC,$2D
	.dc.b	$7F,$B4,$35,$DF,$BF,$E1,$FA,$CF,$F6,$39,$39,$70,$49,$FE,$F7,$19
	.dc.b	$6C,$65,$F2,$C6,$5B,$1F,$17,$EA,$7F,$6F,$1F,$D5,$5A,$FF,$8D,$DA
	.dc.b	$9C,$E4,$FA,$A7,$CD,$7C,$76,$D7,$F8,$39,$AF,$C6,$6B,$FE,$DB,$FA
	.dc.b	$CF,$F4,$A3,$FE,$E6,$6B,$FE,$38,$6B,$BF,$73,$FB,$53,$FE,$E3,$FA
	.dc.b	$C7,$8B,$88,$FF,$1C,$C6,$1F,$F7,$8B,$5F,$F2,$1B,$5E,$51,$C5,$7A
	.dc.b	$89,$D4,$74,$9D,$4E,$5F,$81,$F4,$B2,$7E,$03,$5F,$26,$8F,$9B,$FA
	.dc.b	$04,$8F,$04,$65,$F1,$7D,$63,$E8,$FF,$34,$D8,$CC,$60,$8E,$08,$7C
	.dc.b	$38,$E0,$FC,$5E,$2F,$F1,$6B,$EF,$6B,$9F,$47,$F5,$11,$88,$F9,$ED
	.dc.b	$7F,$C3,$4D,$7F,$CF,$8D,$4D,$3F,$C6,$CF,$7F,$43,$A9,$1F,$A1,$F5
	.dc.b	$5F,$A7,$7F,$C4,$CD,$7F,$D1,$0D,$4D,$27,$47,$EF,$DB,$1E,$DF,$AC
	.dc.b	$F5,$91,$C1,$0F,$88,$E5,$A7,$C5,$FA,$CF,$86,$47,$04,$3E,$63,$96
	.dc.b	$9F,$37,$E5,$3E,$59,$1C,$1F,$03,$D8,$F4,$11,$C1,$0F,$CE,$5A,$35
	.dc.b	$18,$7F,$8F,$7D,$53,$86,$92,$7F,$39,$19,$6C,$65,$FD,$44,$65,$B1
	.dc.b	$D3,$E6,$A4,$72,$6B,$C1,$AF,$ED,$25,$18,$8F,$F3,$04,$3E,$B3,$CE
	.dc.b	$48,$F2,$BF,$CD,$FE,$B1,$E9,$D2,$4F,$ED,$63,$2D,$8C,$BF,$C0,$63
	.dc.b	$2D,$8F,$8B,$EC,$48,$F8,$6D,$7E,$0B,$5F,$53,$5F,$4F,$C8,$7C,$E2
	.dc.b	$3F,$31,$AF,$CA,$6B,$F0,$DF,$AB,$F1,$92,$3E,$9B,$5F,$47,$93,$F0
	.dc.b	$48,$FB,$DA,$E8,$C1,$0F,$F6,$EF,$D2,$7D,$97,$FD,$1D,$D8,$8C,$47
	.dc.b	$FB,$71,$AF,$F1,$E6,$AF,$AC,$7A,$DF,$DA,$5B,$1F,$C9,$5A,$FF,$2F
	.dc.b	$6B,$E0,$7F,$B3,$BF,$CC,$EB,$A7,$91,$FC,$A2,$73,$3F,$EF,$46,$BF
	.dc.b	$CF,$1F,$EE,$AF,$FA,$9D,$A8,$F3,$79,$3E,$53,$CD,$F1,$52,$3C,$CD
	.dc.b	$7C,$CD,$7B,$35,$E9,$F1,$3F,$E9,$1F,$AC,$FE,$84,$D8,$FE,$98,$D7
	.dc.b	$FA,$8B,$5F,$EA,$7B,$7F,$B4,$93,$49,$E5,$7F,$84,$DB,$1F,$D3,$DA
	.dc.b	$FF,$59,$FA,$C7,$C4,$7F,$86,$5B,$1F,$D5,$5A,$EE,$3A,$7C,$74,$8E
	.dc.b	$8D,$7F,$B2,$52,$31,$1E,$86,$BF,$ED,$EF,$99,$FE,$8E,$FF,$B5,$FC
	.dc.b	$E4,$62,$35,$E3,$7F,$89,$63,$11,$F5,$7E,$AD,$F5,$9F,$E3,$16,$C7
	.dc.b	$F6,$D6,$BF,$DD,$DA,$FF,$7C,$6B,$FE,$E7,$6B,$D6,$3A,$1F,$E3,$DF
	.dc.b	$AC,$FE,$9B,$1F,$E0,$16,$BF,$E0,$66,$BF,$EF,$16,$BF,$8B,$F9,$C7
	.dc.b	$D6,$7F,$91,$DB,$1F,$E0,$36,$BF,$E0,$D6,$BF,$E1,$26,$BF,$EF,$D6
	.dc.b	$BD,$63,$A1,$FE,$0D,$F8,$4F,$F1,$3F,$05,$F5,$0D,$25,$7F,$84,$FE
	.dc.b	$B3,$FC,$C0,$FF,$87,$5A,$FF,$88,$9A,$FE,$94,$38,$7F,$50,$D7,$3D
	.dc.b	$9F,$E6,$58,$C1,$9C,$BF,$CA,$53,$F1,$FC,$CF,$9C,$47,$C9,$C3,$C4
	.dc.b	$8E,$0F,$F1,$57,$6F,$FB,$71,$FF,$1B,$B5,$F3,$35,$FD,$E3,$5F,$E3
	.dc.b	$1E,$F7,$F9,$E6,$30,$47,$FC,$E9,$E8,$BE,$AE,$92,$BF,$C8,$5C,$BF
	.dc.b	$E2,$F8,$C4,$7F,$9D,$A7,$83,$A6,$B9,$8F,$F4,$07,$D6,$62,$3F,$44
	.dc.b	$D7,$F8,$95,$23,$10,$8C,$4F,$22,$47,$92,$77,$3A,$8F,$E4,$DF,$56
	.dc.b	$FA,$CF,$F2,$6B,$63,$31,$83,$FD,$29,$F5,$9F,$ED,$A8,$FD,$B3,$5F
	.dc.b	$2B,$5F,$23,$68,$D2,$7C,$37,$43,$FD,$51,$F4,$0F,$21,$F0,$3F,$E6
	.dc.b	$58,$C3,$63,$38,$7F,$9C,$67,$2D,$FA,$C7,$CF,$71,$6C,$66,$30,$7F
	.dc.b	$AD,$D7,$C5,$52,$3F,$EB,$B7,$EE,$FD,$6B,$DB,$FD,$4C,$F9,$BA,$BE
	.dc.b	$AC,$E5,$FF,$66,$1F,$63,$EB,$5F,$67,$C3,$5E,$3C,$0B,$E1,$87,$33
	.dc.b	$EE,$38,$9D,$B7,$0E,$97,$CD,$F0,$E3,$6F,$F0,$9F,$5A,$F7,$3C,$F1
	.dc.b	$FC,$B6,$31,$5F,$98,$6C,$7F,$B3,$1A,$FF,$33,$C3,$FC,$85,$3E,$C8
	.dc.b	$F2,$FB,$3D,$F3,$5F,$2C,$FC,$F3,$9C,$38,$DF,$B6,$FE,$04,$F1,$BE
	.dc.b	$18,$7C,$A8,$F5,$7D,$87,$E0,$BC,$EB,$91,$F0,$7D,$6B,$F0,$1C,$C4
	.dc.b	$C3,$CD,$7C,$5F,$46,$4B,$E1,$1A,$C3,$A3,$AF,$D5,$1E,$27,$B5,$B1
	.dc.b	$F0,$65,$F9,$8B,$E2,$35,$72,$B9,$3F,$4E,$B7,$39,$DF,$E1,$7E,$A9
	.dc.b	$FC,$CA,$FF,$84,$B6,$3F,$DF,$AD,$7F,$B2,$35,$FF,$0A,$69,$F9,$8F
	.dc.b	$99,$F8,$FC,$E7,$EA,$57,$3C,$7F,$2F,$5B,$C3,$E1,$FC,$5B,$E9,$7F
	.dc.b	$0B,$97,$F0,$E8,$C4,$FF,$1C,$6C,$7F,$B8,$B8,$11,$88,$FE,$F0,$D7
	.dc.b	$95,$74,$7F,$8B,$46,$5B,$19,$BF,$D8,$F9,$9E,$A3,$F1,$95,$E0,$C7
	.dc.b	$D3,$A2,$7F,$21,$CF,$F4,$73,$FE,$62,$DC,$BC,$CD,$7F,$C1,$8D,$7F
	.dc.b	$95,$B5,$FE,$54,$DC,$FF,$51,$B7,$A5,$B5,$FE,$FF,$CD,$F8,$33,$C0
	.dc.b	$FC,$BB,$67,$F1,$F8,$78,$F4,$FA,$33,$9C,$BF,$45,$F6,$4F,$BB,$EC
	.dc.b	$DC,$3F,$A5,$B6,$7E,$83,$5F,$F1,$33,$5F,$E8,$AD,$7F,$A2,$37,$33
	.dc.b	$9C,$FD,$3E,$47,$A0,$FD,$D4,$77,$CB,$CF,$1F,$BD,$F2,$1F,$69,$F0
	.dc.b	$DB,$97,$F6,$06,$E5,$F8,$56,$BF,$E3,$B6,$BF,$D6,$9A,$FF,$59,$6E
	.dc.b	$5F,$DB,$6B,$F0,$9E,$3B,$EA,$FF,$81,$B7,$0F,$DD,$DF,$83,$2F,$CE
	.dc.b	$CE,$71,$FE,$1D,$F8,$CF,$FA,$F2,$FF,$BA,$B6,$3F,$34,$D7,$FC,$A0
	.dc.b	$D7,$FB,$7B,$5F,$ED,$CD,$C8,$D3,$5A,$FF,$6B,$FA,$8F,$39,$FC,$96
	.dc.b	$3C,$18,$F9,$A3,$F9,$B7,$C1,$3F,$55,$7C,$07,$E9,$E3,$0D,$8C,$E7
	.dc.b	$FE,$9D,$6E,$7F,$E3,$11,$5C,$27,$55,$C4,$8F,$F8,$5B,$F5,$6F,$1B
	.dc.b	$FE,$0C,$6D,$72,$BE,$B8,$FF,$56,$9C,$E8,$FD,$5D,$F2,$BF,$D0,$5B
	.dc.b	$8F,$F4,$5C,$FF,$D2,$7D,$39,$AF,$EA,$9F,$47,$FC,$AA,$BD,$1A,$23
	.dc.b	$0F,$AD,$AF,$F9,$35,$AF,$F8,$8B,$F0,$9F,$E7,$2F,$F6,$7E,$17,$E3
	.dc.b	$5F,$E9,$AB,$FE,$7F,$5A,$DB,$F9,$0C,$E6,$FF,$B7,$7C,$37,$F8,$1B
	.dc.b	$93,$F8,$B3,$70,$FF,$58,$FA,$1F,$E2,$56,$CE,$A3,$15,$FE,$B4,$6B
	.dc.b	$FE,$63,$AE,$4E,$67,$F8,$F7,$C5,$FF,$62,$C7,$F9,$9B,$70,$FF,$61
	.dc.b	$E1,$D2,$73,$E6,$7F,$62,$FB,$27,$D7,$53,$FB,$87,$EB,$0F,$98,$C6
	.dc.b	$3F,$D8,$2B,$6D,$AD,$CE,$B4,$FE,$E5,$F5,$8F,$C2,$BF,$B7,$FE,$17
	.dc.b	$FC,$57,$4F,$F0,$AC,$E7,$1F,$ED,$73,$97,$FC,$A7,$19,$BF,$A5,$39
	.dc.b	$7F,$CA,$91,$9A,$FE,$A3,$39,$5F,$6F,$FD,$C2,$BF,$35,$18,$F4,$FF
	.dc.b	$A4,$37,$0F,$C3,$B6,$77,$5C,$5F,$F4,$94,$65,$B1,$97,$FD,$F5,$19
	.dc.b	$7F,$DF,$9D,$2F,$6D,$E2,$FF,$86,$F9,$FF,$86,$63,$A9,$CB,$F8,$CF
	.dc.b	$17,$FC,$2B,$E8,$3E,$66,$9F,$E2,$3F,$D6,$7E,$E6,$6E,$CF,$EE,$93
	.dc.b	$AC,$3F,$C2,$B7,$C5,$FF,$0A,$DF,$0A,$DC,$7F,$86,$3F,$5A,$7D,$8F
	.dc.b	$F8,$93,$E8,$9F,$67,$FE,$31,$9C,$D7,$F8,$82,$72,$F1,$AE,$15,$B6
	.dc.b	$C7,$F8,$EF,$BF,$FC,$CF,$E1,$FE,$74,$91,$FA,$26,$EB,$F9,$F6,$BF
	.dc.b	$9F,$EE,$FE,$D5,$9B,$BD,$F4,$5F,$E3,$A6,$E1,$DE,$BF,$B2,$3F,$E1
	.dc.b	$A8,$CD,$FF,$91,$7F,$17,$FC,$27,$3F,$E2,$39,$BF,$C9,$D1,$8B,$FF
	.dc.b	$10,$F5,$7D,$4B,$E1,$D5,$F9,$AB,$DB,$6B,$E7,$F8,$BF,$60,$FF,$2C
	.dc.b	$7D,$F7,$F2,$0D,$BF,$C5,$11,$99,$D9,$FE,$48,$9D,$73,$7F,$1E,$C7
	.dc.b	$F8,$4E,$DF,$D3,$DF,$85,$E7,$FF,$05,$25,$FF,$0A,$6C,$7F,$07,$6B
	.dc.b	$FE,$E9,$C7,$E4,$77,$FC,$73,$FC,$F5,$C7,$FE,$9A,$AC,$5F,$E2,$3D
	.dc.b	$4F,$C8,$9F,$E7,$1F,$C6,$1F,$39,$89,$D1,$32,$FF,$38,$4E,$A3,$1C
	.dc.b	$7F,$E4,$4F,$8B,$FF,$46,$57,$E5,$DB,$E6,$BF,$1D,$7C,$A7,$F9,$D2
	.dc.b	$77,$B9,$A7,$09,$DE,$3F,$E7,$28,$E7,$E9,$7F,$17,$8F,$F7,$3B,$71
	.dc.b	$FF,$26,$35,$1F,$EE,$07,$0F,$FB,$8B,$F0,$FF,$F4,$3B,$FE,$E1,$6E
	.dc.b	$5F,$DA,$BC,$9F,$F4,$A4,$60,$CF,$E6,$AF,$97,$CF,$B9,$FE,$99,$8C
	.dc.b	$65,$FE,$86,$AD,$D6,$A3,$FD,$13,$3A,$C7,$FD,$35,$F9,$C3,$CD,$FF
	.dc.b	$23,$36,$BF,$AE,$E6,$8C,$57,$F9,$A5,$AF,$E0,$75,$7D,$EF,$CD,$3F
	.dc.b	$D5,$93,$C3,$8B,$EB,$47,$FA,$BF,$EF,$3F,$D9,$78,$F6,$6E,$1F,$E7
	.dc.b	$56,$BF,$28,$62,$8C,$C7,$FA,$2B,$1F,$E2,$07,$F1,$89,$D7,$C6,$FD
	.dc.b	$FF,$52,$61,$F6,$DF,$14,$62,$FF,$B9,$5F,$29,$CD,$6C,$FF,$58,$F9
	.dc.b	$9F,$E3,$E2,$79,$BF,$E6,$D7,$FC,$87,$CD,$F8,$BF,$1F,$F1,$C7,$FB
	.dc.b	$36,$3F,$99,$E3,$F9,$58,$FF,$68,$7E,$3D,$FE,$02,$8F,$76,$88,$C4
	.dc.b	$7E,$2A,$31,$F1,$7F,$E7,$B9,$CC,$FC,$46,$CE,$B2,$FB,$31,$83,$78
	.dc.b	$FF,$AB,$6F,$FA,$09,$E2,$2C,$FF,$B3,$A5,$F8,$FA,$3F,$36,$79,$A3
	.dc.b	$2E,$30,$F3,$46,$1F,$F7,$BE,$1E,$BA,$E5,$18,$9D,$57,$39,$CB,$4E
	.dc.b	$93,$ED,$8C,$BD,$9E,$7F,$0B,$FB,$98,$FE,$AD,$E2,$3E,$83,$3C,$1B
	.dc.b	$C3,$F6,$77,$BD,$3F,$D9,$D5,$AC,$3B,$56,$AF,$FD,$AB,$87,$38,$C4
	.dc.b	$E8,$61,$CE,$75,$18,$AF,$B1,$F8,$FF,$F0,$A8,$C1,$F9,$A7,$15,$E9
	.dc.b	$D1,$0F,$A6,$F2,$46,$3C,$0B,$3C,$30,$EB,$18,$AC,$B3,$2F,$3D,$F1
	.dc.b	$6D,$F5,$6C,$FC,$1C,$5F,$68,$FA,$8F,$F6,$B4,$7A,$0C,$B8,$CF,$E8
	.dc.b	$7C,$03,$EF,$7C,$95,$FC,$82,$F9,$7C,$5F,$C8,$8E,$2D,$8C,$C7,$C4
	.dc.b	$AE,$58,$F8,$E7,$38,$7D,$DA,$E2,$7F,$9C,$23,$8B,$FE,$89,$6B,$FC
	.dc.b	$1D,$AF,$8F,$57,$DD,$FF,$9E,$23,$DF,$18,$7A,$35,$E8,$D7,$A3,$BA
	.dc.b	$F2,$57,$D7,$AF,$64,$E7,$7F,$E8,$38,$BE,$F7,$FD,$DD,$8E,$AE,$3F
	.dc.b	$37,$0E,$4F,$C3,$9D,$BF,$88,$E0,$7E,$33,$F5,$23,$0D,$CF,$F1,$4D
	.dc.b	$CB,$F8,$43,$FE,$AD,$8C,$EB,$FC,$AD,$B5,$F8,$69,$E0,$FF,$2E,$78
	.dc.b	$57,$2A,$E0,$D9,$E1,$18,$78,$16,$DD,$7F,$9B,$8F,$D4,$F2,$B9,$1F
	.dc.b	$4B,$C6,$30,$F1,$AE,$15,$F0,$5B,$9F,$EB,$E7,$85,$F2,$AF,$CF,$D6
	.dc.b	$B2,$FE,$50,$F2,$E4,$FE,$5A,$D9,$CF,$0F,$F4,$D9,$CB,$63,$27,$F1
	.dc.b	$2C,$FD,$71,$98,$FE,$A7,$5C,$DB,$19,$CF,$DD,$87,$F9,$F3,$0F,$E1
	.dc.b	$6D,$AF,$E2,$ED,$C3,$E3,$FE,$01,$71,$3F,$97,$8C,$D7,$F7,$49,$CB
	.dc.b	$67,$51,$88,$FF,$76,$4F,$06,$C6,$5F,$87,$3F,$A3,$C3,$FC,$45,$E8
	.dc.b	$79,$5B,$3F,$E2,$18,$F3,$7D,$CB,$F1,$9A,$79,$5C,$4F,$F8,$21,$AF
	.dc.b	$89,$AF,$F8,$1F,$D4,$ED,$99,$A7,$C3,$FB,$A7,$E1,$F8,$AF,$F1,$94
	.dc.b	$D5,$E3,$FF,$4C,$D9,$7C,$27,$64,$3F,$BE,$D7,$5B,$91,$FE,$1E,$9E
	.dc.b	$0F,$C8,$6B,$FD,$D1,$AF,$F8,$3F,$F1,$7F,$49,$1F,$F5,$AF,$07,$F7
	.dc.b	$7F,$61,$F5,$FF,$75,$B3,$FE,$2C,$6B,$F5,$9A,$FF,$80,$9A,$FF,$91
	.dc.b	$F9,$FF,$0D,$E9,$7E,$B2,$3B,$9F,$E2,$47,$0F,$C4,$BE,$0F,$F9,$0E
	.dc.b	$78,$F2,$48,$C1,$0E,$07,$F9,$4E,$3F,$29,$E8,$7F,$0F,$DF,$E7,$7E
	.dc.b	$06,$F1,$FF,$30,$DB,$FF,$23,$B5,$FC,$8B,$5F,$F0,$EB,$5F,$F0,$E7
	.dc.b	$E2,$8F,$9A,$FD,$13,$63,$DB,$C6,$8C,$47,$9A,$1F,$8F,$BD,$6E,$37
	.dc.b	$2B,$8D,$F0,$CB,$F0,$D7,$B9,$F9,$3D,$4F,$C1,$DF,$6B,$FA,$16,$CE
	.dc.b	$9A,$FF,$A4,$BC,$7D,$4C,$8D,$3F,$BB,$F6,$BF,$D1,$F1,$DB,$E1,$1F
	.dc.b	$5D,$FA,$D6,$C6,$63,$F8,$23,$AF,$6E,$C1,$FE,$B1,$19,$3F,$AD,$46
	.dc.b	$5B,$1A,$8C,$47,$F9,$A4,$FF,$49,$B8,$70,$D7,$FC,$DC,$D7,$FC,$DF
	.dc.b	$AB,$D4,$FA,$A4,$8F,$AB,$F8,$7B,$D6,$C4,$7D,$5A,$E0,$FD,$2A,$FE
	.dc.b	$3D,$5B,$8F,$BB,$F0,$CF,$B5,$FE,$20,$3F,$03,$A7,$F2,$58,$D4,$F0
	.dc.b	$9C,$BF,$69,$E0,$E4,$FA,$1F,$27,$F3,$C8,$FE,$A0,$7D,$DB,$EF,$3E
	.dc.b	$8B,$63,$FC,$0C,$D7,$E5,$1C,$23,$E9,$C6,$63,$E7,$35,$FF,$4C,$35
	.dc.b	$FF,$61,$B5,$E4,$7F,$AB,$BE,$41,$F7,$FF,$CC,$7D,$4B,$3F,$96,$6B
	.dc.b	$FA,$19,$BD,$3F,$C8,$5B,$E1,$7F,$B4,$3A,$7F,$C7,$78,$73,$3F,$A7
	.dc.b	$4E,$AD,$3F,$A1,$F8,$EF,$D0,$12,$3F,$63,$F2,$4F,$C6,$B1,$1F,$C1
	.dc.b	$AB,$83,$FA,$5A,$FE,$B5,$5B,$8F,$E2,$B9,$3F,$14,$FF,$0D,$BF,$E8
	.dc.b	$2F,$6B,$FD,$75,$EE,$7F,$63,$6D,$70,$AD,$EB,$F2,$BA,$FE,$79,$27
	.dc.b	$EA,$B5,$FA,$BC,$08,$C4,$7D,$47,$FB,$57,$C3,$79,$63,$D5,$C9,$FD
	.dc.b	$03,$3F,$F4,$AE,$DF,$93,$F3,$FF,$BA,$5E,$E7,$FA,$73,$72,$F9,$D7
	.dc.b	$C5,$B5,$C3,$83,$FB,$57,$E2,$BF,$EE,$DC,$BF,$87,$B5,$F9,$4D,$C7
	.dc.b	$FA,$D7,$89,$FC,$A5,$F1,$C6,$70,$FB,$8D,$AF,$B2,$DF,$1B,$F9,$57
	.dc.b	$91,$FC,$37,$9A,$57,$F1,$17,$F8,$6F,$E6,$1E,$78,$07,$FA,$EB,$86
	.dc.b	$B9,$D3,$F8,$8C,$60,$8E,$09,$1F,$A6,$BF,$E0,$9F,$45,$FE,$0B,$F8
	.dc.b	$07,$D5,$53,$FC,$41,$E1,$A7,$F0,$5C,$DF,$E6,$0C,$3F,$E0,$07,$3F
	.dc.b	$E2,$7E,$3F,$FC,$F0,$D8,$FD,$9B,$5F,$F0,$FF,$D0,$FE,$05,$FE,$92
	.dc.b	$D7,$FA,$43,$78,$BF,$DA,$7F,$38,$FD,$53,$FD,$91,$5F,$D4,$3C,$B7
	.dc.b	$D8,$FF,$8D,$5B,$19,$8F,$F1,$D5,$7F,$AD,$FF,$09,$BF,$02,$7F,$D1
	.dc.b	$D1,$9A,$FE,$12,$D7,$F9,$A3,$5F,$A6,$1C,$3F,$59,$AE,$67,$B3,$D1
	.dc.b	$AF,$F3,$FE,$D7,$E3,$CF,$9C,$FF,$1E,$36,$7F,$B0,$B5,$FE,$D9,$F8
	.dc.b	$43,$E3,$FF,$CA,$8D,$8F,$B8,$D7,$FC,$9A,$D7,$D3,$CE,$8C,$47,$E3
	.dc.b	$A3,$13,$FE,$5B,$9C,$CA,$72,$F6,$9F,$F2,$A5,$7D,$08,$FE,$20,$DD
	.dc.b	$FF,$6E,$DE,$DE,$DF,$80,$FE,$2E,$BF,$C1,$2D,$7F,$CC,$6D,$7C,$ED
	.dc.b	$78,$F3,$7F,$9B,$E3,$11,$E5,$6E,$26,$9A,$D3,$F9,$5D,$7F,$9A,$5F
	.dc.b	$06,$3F,$C3,$1B,$87,$C3,$FA,$CF,$E1,$58,$FC,$17,$F8,$A3,$F2,$9A
	.dc.b	$EF,$C1,$FF,$41,$46,$23,$84,$60,$8F,$EC,$63,$55,$FC,$FB,$EF,$7C
	.dc.b	$84,$9F,$F1,$6F,$E3,$1F,$6F,$1F,$F3,$53,$F5,$34,$34,$9B,$7F,$A4
	.dc.b	$DB,$5F,$BF,$6B,$F9,$1F,$9D,$FD,$5B,$FD,$1D,$B7,$FC,$B9,$BA,$7E
	.dc.b	$17,$EC,$6D,$7E,$0D,$E2,$FE,$31,$AE,$FE,$4F,$FD,$49,$18,$8E,$11
	.dc.b	$82,$3F,$C4,$E3,$5A,$2D,$F1,$1F,$E9,$F6,$D7,$F9,$11,$AF,$F9,$3F
	.dc.b	$F5,$B7,$C2,$FF,$AD,$DB,$1F,$C7,$DA,$FF,$AC,$1A,$FE,$91,$AF,$E8
	.dc.b	$FF,$5C,$FD,$86,$9F,$E9,$96,$E9,$9E,$7F,$F6,$34,$BF,$49,$FF,$15
	.dc.b	$DF,$F9,$86,$F6,$FE,$A5,$BC,$27,$35,$C7,$DA,$FF,$5E,$7D,$17,$F9
	.dc.b	$86,$FF,$BD,$36,$7F,$B0,$B7,$2F,$DC,$7D,$27,$F7,$2A,$FF,$03,$CE
	.dc.b	$5B,$F0,$9D,$0F,$F6,$D4,$EF,$27,$DD,$3D,$BF,$F6,$F3,$6F,$FC,$FA
	.dc.b	$D7,$FC,$F9,$F8,$07,$CA,$CF,$FD,$79,$39,$6D,$72,$3F,$DD,$13,$BC
	.dc.b	$FF,$87,$3F,$3A,$FF,$86,$37,$03,$4C,$61,$FE,$D6,$C3,$BF,$C8,$F9
	.dc.b	$8F,$F7,$A7,$DE,$3E,$57,$18,$FA,$26,$BF,$E9,$39,$D3,$FE,$4D,$FA
	.dc.b	$B7,$DF,$7F,$B5,$A3,$FC,$93,$18,$C3,$FC,$97,$08,$C3,$A9,$E0,$FF
	.dc.b	$9E,$61,$F7,$CE,$1E,$31,$86,$C6,$67,$BB,$5E,$FF,$35,$FE,$8A,$4D
	.dc.b	$23,$FE,$87,$DB,$EC,$F9,$4E,$0F,$AE,$D9,$F8,$1E,$8D,$D3,$EB,$CD
	.dc.b	$AF,$E8,$1C,$1F,$DA,$9E,$27,$FA,$F3,$6F,$F4,$43,$8C,$3F,$D7,$9F
	.dc.b	$7D,$FE,$9F,$8F,$F0,$53,$63,$C8,$D7,$FA,$7B,$5E,$FF,$25,$E1,$7F
	.dc.b	$A6,$5A,$31,$9A,$BA,$65,$F8,$E9,$CD,$6F,$0F,$CA,$C7,$0A,$DC,$76
	.dc.b	$3F,$D9,$34,$FA,$BF,$07,$80,$8C,$4F,$FB,$6B,$EF,$BF,$DD,$0F,$0A
	.dc.b	$DC,$EB,$3F,$F3,$63,$4F,$9D,$9A,$F8,$CE,$AB,$D7,$F7,$DF,$E2,$CC
	.dc.b	$3F,$DC,$ED,$9F,$F7,$DE,$0F,$BE,$7C,$37,$0F,$66,$61,$C1,$F9,$15
	.dc.b	$B9,$CD,$FF,$BB,$FE,$FB,$F0,$57,$FE,$3D,$6C,$7E,$1D,$AF,$F8,$09
	.dc.b	$AF,$FA,$4F,$93,$F5,$53,$FE,$F0,$6C,$75,$8F,$93,$18,$BF,$83,$E3
	.dc.b	$BE,$07,$F2,$17,$FD,$0D,$AF,$F4,$96,$BF,$D2,$5B,$BA,$79,$61,$FD
	.dc.b	$0E,$73,$18,$AF,$D0,$35,$FF,$19,$F0,$7E,$9B,$E5,$FC,$A7,$E2,$A7
	.dc.b	$86,$3F,$E8,$8C,$93,$F8,$EF,$91,$F9,$F9,$7F,$98,$DB,$8F,$EB,$5A
	.dc.b	$FF,$8A,$1A,$FF,$89,$DA,$FF,$89,$5B,$99,$A6,$B2,$7D,$F3,$DD,$FC
	.dc.b	$9B,$70,$FF,$3F,$E7,$F8,$CB,$F1,$BF,$EA,$09,$D6,$3F,$99,$AF,$F0
	.dc.b	$B7,$DF,$7C,$AF,$48,$D2,$47,$F8,$3D,$B3,$FC,$26,$77,$39,$3F,$DD
	.dc.b	$39,$79,$BC,$5F,$29,$FB,$49,$DD,$7F,$B3,$70,$4F,$EE,$3E,$F9,$E4
	.dc.b	$7D,$C6,$E3,$FE,$8F,$6D,$FF,$19,$6B,$FE,$54,$6B,$FE,$53,$6B,$FE
	.dc.b	$51,$6E,$C7,$39,$3E,$E9,$F6,$7F,$BE,$6E,$1F,$30,$E9,$8E,$67,$F8
	.dc.b	$7F,$DF,$3E,$3B,$F2,$6D,$AF,$F5,$33,$73,$FE,$60,$D7,$FC,$D6,$D7
	.dc.b	$FC,$D4,$D7,$FC,$D0,$DC,$CE,$72,$F4,$7B,$3E,$73,$C7,$85,$C3,$D1
	.dc.b	$87,$FA,$3F,$EF,$9F,$3F,$FA,$AF,$80,$46,$2B,$FD,$32,$DA,$FF,$4C
	.dc.b	$B5,$FF,$4C,$CF,$0A,$E5,$39,$9F,$F4,$D3,$5F,$04,$7E,$DB,$EF,$BE
	.dc.b	$17,$2F,$F2,$06,$E1,$C2,$7F,$9C,$C6,$2B,$FD,$11,$F0,$BF,$E3,$4A
	.dc.b	$FF,$5F,$FA,$FF,$D0,$72,$7B,$5F,$9E,$9F,$F4,$95,$6E,$FF,$3C,$3E
	.dc.b	$1B,$F9,$54,$FF,$B3,$3C,$2F,$E3,$33,$96,$CE,$A3,$FB,$37,$DD,$7F
	.dc.b	$34,$D7,$FB,$27,$0B,$EF,$9F,$47,$FD,$42,$B5,$97,$FB,$C2,$B5,$3F
	.dc.b	$D5,$FD,$DF,$EF,$9D,$3F,$C6,$9B,$5F,$EC,$4C,$BF,$DE,$F1,$87,$0D
	.dc.b	$7F,$D9,$2D,$77,$E9,$7E,$66,$79,$47,$F9,$86,$30,$F2,$6B,$FC,$83
	.dc.b	$E4,$FF,$B1,$3F,$1B,$14,$63,$4F,$F0,$16,$1E,$37,$E0,$C6,$70,$F7
	.dc.b	$7D,$EF,$F9,$05,$FC,$7D,$A3,$18,$7D,$2E,$A7,$DF,$3D,$CF,$ED,$75
	.dc.b	$AC,$BE,$6B,$71,$EB,$EE,$7F,$7C,$BE,$11,$DA,$75,$8F,$FB,$37,$E5
	.dc.b	$5F,$55,$FD,$DB,$EE,$FF,$8F,$E5,$FD,$BA,$73,$87,$9A,$72,$FF,$B6
	.dc.b	$63,$37,$F0,$E7,$2F,$5B,$E9,$E4,$FF,$6A,$FB,$EF,$EF,$93,$FE,$FC
	.dc.b	$C5,$18,$8F,$E6,$6D,$7F,$C1,$0D,$8D,$FA,$7F,$D7,$1B,$97,$DA,$6D
	.dc.b	$7E,$15,$AF,$EC,$BD,$CF,$E5,$33,$FB,$EE,$07,$3F,$F2,$6B,$FF,$1D
	.dc.b	$76,$FF,$94,$7C,$8F,$CB,$78,$FF,$E1,$AF,$7D,$F6,$B4,$FF,$14,$FC
	.dc.b	$DF,$DD,$D5,$D1,$FE,$65,$9D,$6B,$F7,$AF,$8B,$E1,$BE,$15,$B8,$F2
	.dc.b	$7C,$E3,$E1,$BF,$C5,$3F,$7C,$FB,$EF,$F1,$BC,$E6,$BF,$13,$39,$78
	.dc.b	$D7,$0A,$DB,$63,$F8,$EF,$77,$F9,$9F,$D9,$FD,$67,$81,$FD,$B2,$31
	.dc.b	$3F,$A4,$6E,$DF,$A3,$AD,$92,$75,$5F,$DB,$FC,$07,$C6,$7F,$90,$1B
	.dc.b	$3E,$09,$FE,$21,$7F,$1A,$72,$FE,$7F,$E8,$7F,$C7,$D5,$FC,$0F,$B7
	.dc.b	$EB,$CE,$7A,$BF,$2D,$CB,$FE,$58,$BD,$CF,$D0,$AD,$F3,$FF,$91,$A7
	.dc.b	$3C,$49,$CE,$3F,$C3,$F9,$3F,$90,$5F,$F1,$66,$C3,$07,$4B,$EB,$BF
	.dc.b	$97,$63,$FE,$A1,$F2,$FF,$08,$4B,$FC,$64,$EB,$0E,$78,$7F,$0D,$F4
	.dc.b	$BE,$79,$2B,$E4,$B6,$3F,$88,$35,$FE,$33,$C1,$FD,$B1,$FE,$41,$8F
	.dc.b	$F1,$0F,$03,$E7,$1F,$C9,$E7,$5A,$7F,$10,$8F,$CC,$78,$13,$1F,$D0
	.dc.b	$7D,$3B,$EF,$B1,$97,$F1,$BD,$3F,$73,$9F,$FA,$F7,$0F,$F4,$2D,$EF
	.dc.b	$0F,$CD,$47,$F9,$F3,$E3,$FF,$D3,$58,$7F,$23,$6E,$7F,$E7,$6F,$BE
	.dc.b	$7D,$56,$FE,$26,$63,$FC,$82,$F6,$FC,$FE,$35,$F0,$8D,$57,$FA,$5B
	.dc.b	$D4,$FE,$37,$97,$FA,$CD,$AF,$D7,$3E,$BB,$87,$A3,$5F,$B7,$F5,$DF
	.dc.b	$E9,$86,$FB,$DD,$7C,$EF,$F5,$1C,$E7,$23,$7E,$F5,$F7,$DA,$BF,$F5
	.dc.b	$4C,$63,$3F,$E1,$95,$BA,$D4,$7F,$0F,$9D,$63,$FC,$F7,$EF,$9E,$77
	.dc.b	$F9,$D1,$B5,$FD,$53,$24,$62,$BF,$DA,$4D,$7F,$CE,$7B,$BE,$FB,$EA
	.dc.b	$1F,$EB,$C9,$E1,$C5,$FD,$CA,$75,$87,$F2,$1F,$BE,$FF,$67,$E1,$FE
	.dc.b	$DD,$6B,$F2,$F1,$46,$23,$EB,$35,$FE,$3F,$5F,$C6,$8F,$E4,$53,$C3
	.dc.b	$E4,$3F,$0D,$F1,$D3,$0F,$C0,$3F,$ED,$0C,$3C,$38,$F9,$63,$58,$FE
	.dc.b	$2D,$D5,$E2,$7F,$D8,$3E,$6F,$F9,$50,$9E,$77,$FA,$71,$FF,$2A,$E7
	.dc.b	$FC,$32,$7E,$07,$8F,$FE,$DE,$9F,$E7,$D9,$FF,$83,$EB,$4F,$E4,$7E
	.dc.b	$FB,$FD,$97,$9F,$C1,$6B,$9F,$A8,$FF,$54,$C6,$23,$F8,$2B,$76,$34
	.dc.b	$93,$FE,$B2,$ED,$77,$7F,$88,$2F,$FA,$3B,$6F,$8D,$7F,$4F,$3F,$DE
	.dc.b	$B6,$FB,$E6,$3F,$C5,$4F,$33,$8C,$3F,$D8,$77,$E2,$8C,$4E,$8E,$13
	.dc.b	$EF,$8C,$B6,$32,$FB,$23,$2D,$8C,$CF,$57,$F9,$37,$C5,$33,$FE,$59
	.dc.b	$F0,$1E,$38,$C5,$F1,$0E,$34,$F8,$73,$BA,$F1,$56,$9E,$77,$D9,$CD
	.dc.b	$F1,$D3,$FC,$B7,$39,$AF,$E6,$FF,$7D,$FC,$25,$E4,$63,$C0,$EC,$EB
	.dc.b	$E8,$AE,$8A,$E1,$8F,$A2,$31,$7A,$9C,$E5,$FC,$5D,$5E,$66,$63,$E3
	.dc.b	$BE,$2D,$AD,$B6,$7C,$78,$BE,$BB,$FD,$49,$F1,$BE,$B8,$C7,$C3,$FF
	.dc.b	$BA,$9F,$5F,$C1,$BE,$A3,$C7,$1F,$D5,$EF,$8B,$F0,$0E,$37,$CE,$33
	.dc.b	$B7,$D4,$BE,$58,$F8,$A7,$3B,$7F,$2B,$79,$E1,$FE,$6F,$6B,$9C,$BC
	.dc.b	$8E,$5A,$F1,$D9,$F7,$DF,$E6,$E8,$F7,$46,$1F,$F3,$8D,$7C,$A9,$D3
	.dc.b	$63,$33,$BC,$BF,$17,$1C,$AB,$D9,$39,$C3,$FB,$C6,$4F,$BE,$FF,$80
	.dc.b	$17,$AB,$8F,$C5,$C7,$B6,$3F,$0E,$B7,$3F,$88,$E0,$7D,$F7,$BA,$BF
	.dc.b	$08,$DC,$BF,$AC,$36,$BF,$C0,$EF,$FA,$56,$77,$B7,$F5,$56,$CF,$E1
	.dc.b	$67,$85,$F6,$8D,$5F,$4B,$B1,$D1,$B3,$E4,$6D,$77,$AF,$ED,$5D,$BF
	.dc.b	$94,$BF,$D2,$72,$FF,$10,$FB,$EE,$6D,$8F,$CB,$4E,$DF,$80,$DC,$BF
	.dc.b	$5D,$5C,$39,$4F,$8A,$75,$87,$83,$83,$FB,$97,$27,$F4,$99,$E8,$D8
	.dc.b	$CE,$3B,$3F,$85,$CE,$9C,$1F,$C6,$9C,$3F,$E3,$26,$E5,$FC,$AA,$FF
	.dc.b	$D1,$57,$FE,$3B,$6D,$7F,$2C,$6E,$AB,$7D,$F6,$1B,$5F,$DB,$DA,$FF
	.dc.b	$B3,$67,$84,$FE,$65,$B1,$B8,$C4,$F0,$8F,$81,$18,$3F,$41,$87,$F2
	.dc.b	$5F,$3F,$CA,$D9,$FF,$11,$47,$9B,$EF,$9F,$41,$4F,$03,$89,$FF,$02
	.dc.b	$35,$F3,$35,$FF,$01,$FA,$9D,$B3,$FF,$2C,$47,$F8,$BE,$3E,$1F,$DF
	.dc.b	$3E,$C3,$51,$27,$47,$FA,$96,$74,$E0,$FF,$1F,$46,$DF,$F6,$B4,$69
	.dc.b	$B3,$F1,$E3,$11,$FE,$0B,$6B,$FE,$3F,$6B,$FE,$0A,$FB,$EF,$41,$1F
	.dc.b	$F5,$B7,$B0,$F8,$DF,$B6,$D8,$FF,$13,$35,$FB,$0D,$7F,$C3,$4D,$7F
	.dc.b	$C9,$DE,$7F,$F1,$8F,$1F,$FB,$47,$13,$D2,$FC,$35,$7F,$B1,$5A,$FF
	.dc.b	$BC,$EB,$87,$27,$2E,$AF,$C6,$D6,$B0,$E6,$D9,$D5,$FF,$91,$2B,$F2
	.dc.b	$77,$CA,$3F,$CA,$71,$F9,$6F,$BE,$78,$F4,$FC,$83,$89,$FF,$22,$35
	.dc.b	$FC,$B3,$5F,$F1,$C7,$DD,$BE,$73,$F4,$2D,$8F,$F7,$03,$56,$7F,$0B
	.dc.b	$EF,$9E,$B7,$0B,$95,$C1,$FD,$5C,$EA,$3F,$D4,$D5,$FE,$11,$E9,$7D
	.dc.b	$F3,$EC,$FF,$3E,$DA,$DE,$FF,$EA,$DA,$DD,$7E,$DB,$DC,$FF,$43,$D6
	.dc.b	$8F,$F7,$5D,$F1,$7C,$F7,$C6,$31,$87,$85,$B1,$FE,$8B,$AE,$75,$C0
	.dc.b	$FF,$68,$7D,$F7,$EC,$BE,$14,$D3,$F8,$4C,$63,$DB,$B0,$7F,$AD,$46
	.dc.b	$4F,$EB,$91,$96,$C6,$A3,$11,$FE,$95,$3F,$D9,$8E,$1C,$35,$FF,$4D
	.dc.b	$35,$FF,$68,$EE,$F5,$3E,$D9,$23,$E3,$73,$3E,$F9,$EB,$66,$7E,$04
	.dc.b	$AF,$94,$7E,$0A,$FF,$92,$DF,$08,$E7,$F0,$8F,$BB,$FE,$2E,$3E,$FA
	.dc.b	$9F,$CB,$23,$53,$C2,$72,$FD,$9E,$3E,$55,$B9,$FA,$1F,$2D,$2B,$5B
	.dc.b	$A3,$E8,$FD,$EB,$EF,$A9,$F4,$1C,$57,$F8,$2D,$AF,$CD,$0E,$1F,$A0
	.dc.b	$D7,$FD,$8C,$D7,$FD,$FC,$D7,$8C,$7F,$3F,$F9,$17,$DE,$FF,$38,$F4
	.dc.b	$ED,$7B,$5A,$FC,$E9,$BD,$1F,$C3,$1B,$CF,$F9,$FA,$B4,$DB,$FC,$81
	.dc.b	$87,$F5,$6B,$CC,$7E,$AB,$E3,$BF,$52,$4A,$FB,$FF,$26,$FB,$6C,$47
	.dc.b	$E7,$6B,$83,$F4,$7A,$D7,$FD,$8A,$F8,$4F,$EB,$74,$7D,$E7,$F8,$81
	.dc.b	$FF,$45,$7B,$3F,$EC,$0F,$6B,$FB,$5B,$6B,$85,$6F,$7F,$97,$C7,$F5
	.dc.b	$49,$3F,$55,$AF,$D5,$E2,$46,$23,$EA,$BF,$DB,$7E,$1B,$CB,$1E,$BE
	.dc.b	$5F,$E9,$99,$FF,$0D,$6E,$09,$75,$FF,$BA,$B8,$7F,$BF,$E1,$CB,$0E
	.dc.b	$1E,$6F,$E6,$8C,$CF,$F7,$2F,$BE,$FF,$78,$E1,$F9,$76,$BF,$2D,$B5
	.dc.b	$FD,$1B,$C4,$FE,$60,$FF,$1F,$8C,$E5,$FC,$6D,$B5,$F7,$5B,$97,$F3
	.dc.b	$3F,$1F,$F8,$87,$3C,$AF,$D0,$BF,$C4,$3E,$FB,$DB,$1E,$D7,$FB,$2A
	.dc.b	$69,$34,$FE,$26,$D8,$FF,$37,$35,$FF,$0D,$78,$2F,$A4,$FF,$08,$FC
	.dc.b	$03,$E8,$69,$FE,$26,$9D,$57,$F3,$1E,$6F,$C2,$FC,$7F,$F9,$E5,$B1
	.dc.b	$FA,$F6,$BF,$C7,$5A,$FD,$53,$EA,$FD,$17,$E0,$1F,$E2,$ED,$7F,$8A
	.dc.b	$B7,$87,$FD,$B5,$F3,$6F,$BE,$7C,$3F,$F8,$EB,$CD,$7C,$8D,$26,$3F
	.dc.b	$D1,$EF,$FD,$67,$F7,$DF,$E4,$26,$C7,$F0,$56,$BF,$CC,$DA,$F9,$DA
	.dc.b	$FE,$35,$A9,$A4,$F8,$14,$FF,$8F,$BE,$F5,$F3,$9F,$E3,$E6,$DF,$F2
	.dc.b	$36,$BF,$CD,$3E,$F9,$F2,$1F,$E5,$66,$C7,$F1,$BE,$A4,$62,$3F,$CA
	.dc.b	$8D,$7E,$74,$6A,$7F,$CA,$D3,$99,$FD,$34,$E5,$E9,$19,$7F,$52,$DF
	.dc.b	$21,$7F,$3A,$7D,$1A,$7A,$CF,$F3,$5F,$DF,$7E,$1A,$3F,$A6,$B5,$F1
	.dc.b	$B5,$F0,$B5,$FD,$BB,$70,$34,$9D,$1F,$E6,$F7,$86,$1F,$C9,$E3,$FC
	.dc.b	$D2,$F8,$78,$BF,$C2,$0D,$EE,$F7,$FD,$6F,$F0,$16,$F8,$3F,$E7,$98
	.dc.b	$FE,$1B,$97,$C8,$6B,$A3,$E8,$38,$7F,$86,$B7,$23,$4D,$5F,$F3,$BF
	.dc.b	$BE,$F2,$92,$BF,$BA,$FD,$F7,$D9,$CB,$FC,$9E,$FD,$1C,$0D,$27,$07
	.dc.b	$FA,$3D,$B5,$F8,$36,$BF,$E0,$BC,$1F,$43,$FA,$68,$FE,$E0,$D7,$FA
	.dc.b	$23,$70,$FF,$2F,$FD,$8D,$D7,$F0,$AE,$EB,$F1,$6D,$73,$D1,$FE,$A1
	.dc.b	$8C,$47,$F3,$06,$E6,$69,$27,$FD,$43,$87,$FB,$E3,$0F,$F2,$33,$70
	.dc.b	$FF,$21,$37,$0F,$E7,$7F,$7C,$F8,$2F,$F5,$CB,$63,$F4,$8D,$7F,$D6
	.dc.b	$CD,$7F,$46,$D7,$F4,$3F,$7B,$FB,$0C,$3F,$CF,$ED,$C3,$3C,$FF,$EC
	.dc.b	$89,$7D,$F7,$F8,$9A,$FF,$C5,$37,$B7,$FA,$7B,$75,$39,$AE,$3E,$D7
	.dc.b	$FA,$FB,$E8,$7F,$CC,$D7,$FD,$D5,$B3,$FE,$67,$6E,$5F,$B6,$FA,$4F
	.dc.b	$ED,$95,$FE,$05,$9C,$B7,$E0,$BA,$1F,$ED,$C9,$DE,$4F,$BA,$FF,$15
	.dc.b	$57,$F9,$11,$AF,$F9,$09,$B8,$FF,$A8,$3E,$FB,$E5,$3F,$EA,$28,$CE
	.dc.b	$5F,$C2,$1B,$5F,$DB,$9B,$6B,$E3,$5F,$C3,$9F,$A1,$5F,$DD,$1B,$3B
	.dc.b	$E0,$FF,$75,$46,$2F,$FD,$B5,$87,$87,$E3,$FF,$DE,$F1,$8C,$7F,$8C
	.dc.b	$FD,$E3,$E5,$73,$53,$3F,$73,$FE,$FE,$9D,$47,$FA,$24,$7D,$63,$EF
	.dc.b	$BF,$D3,$F1,$FE,$47,$8C,$65,$FE,$49,$8D,$CE,$9F,$F3,$C4,$E9,$E1
	.dc.b	$F7,$CE,$1E,$12,$8C,$4E,$A7,$31,$88,$ED,$F3,$3F,$E8,$A4,$D2,$3F
	.dc.b	$E8,$7E,$0F,$B9,$CC,$DB,$EE,$36,$7F,$CF,$BF,$9C,$7F,$66,$9C,$92
	.dc.b	$3F,$01,$3B,$D3,$FD,$BF,$5B,$9C,$D7,$FA,$2A,$1F,$41,$D6,$7F,$A6
	.dc.b	$C6,$27,$FA,$A3,$63,$CC,$D7,$FA,$9F,$E3,$7F,$EE,$B9,$FF,$4A,$B6
	.dc.b	$3C,$F5,$E3,$9F,$EA,$B3,$9B,$FF,$79,$B6,$33,$F2,$5B,$BF,$F1,$FE
	.dc.b	$1E,$1B,$DC,$E7,$BF,$FD,$B7,$F5,$56,$4D,$FA,$EC,$33,$1F,$F4,$B4
	.dc.b	$E9,$B3,$9C,$7F,$CD,$9F,$8D,$7D,$D9,$FF,$4E,$36,$33,$F2,$DE,$18
	.dc.b	$C5,$FF,$B8,$DB,$5F,$ED,$F6,$D7,$FA,$05,$57,$1E,$1F,$55,$3E,$8F
	.dc.b	$E9,$5F,$FB,$9E,$30,$7F,$A7,$F3,$FC,$24,$EA,$7F,$D1,$93,$A7,$FA
	.dc.b	$37,$13,$F1,$BF,$8C,$D9,$FF,$5B,$1B,$CF,$FD,$F5,$19,$AF,$F4,$67
	.dc.b	$E3,$4F,$A0,$ED,$92,$31,$87,$8D,$B5,$FD,$55,$AF,$F9,$0E,$78,$57
	.dc.b	$29,$CC,$FF,$9E,$E7,$35,$CE,$BB,$FC,$CB,$EF,$3E,$F3,$67,$FD,$9D
	.dc.b	$87,$83,$2F,$D2,$60,$AD,$63,$FE,$A5,$E7,$FF,$17,$E1,$F2,$5B,$1F
	.dc.b	$DA,$DA,$F9,$DA,$F9,$5A,$FF,$AE,$5B,$5F,$E9,$2F,$2F,$E2,$C6,$3E
	.dc.b	$F7,$FA,$D5,$FF,$B8,$DB,$7E,$3A,$FC,$C4,$63,$1F,$4F,$E3,$9F,$BF
	.dc.b	$BF,$BC,$DB,$FF,$00,$35,$F9,$8D,$7E,$53,$5F,$F6,$73,$71,$34,$C7
	.dc.b	$E4,$2F,$A2,$FD,$53,$6F,$FD,$F5,$3E,$1C,$3E,$6C,$E7,$0F,$F6,$87
	.dc.b	$AC,$FC,$4B,$F0,$AD,$C3,$F2,$CD,$9F,$F0,$C3,$5F,$C2,$35,$FC,$13
	.dc.b	$5F,$F7,$33,$72,$FF,$2A,$D7,$E4,$BE,$F1,$F6,$FF,$C0,$5B,$87,$F9
	.dc.b	$BE,$71,$3F,$C1,$7E,$AD,$F5,$FF,$9A,$6D,$7E,$AD,$B9,$7F,$8B,$5A
	.dc.b	$FE,$65,$AF,$E5,$DA,$FF,$BE,$DB,$89,$BF,$94,$3E,$CF,$F8,$B3,$6F
	.dc.b	$FD,$B5,$B7,$E6,$2B,$C2,$FE,$6A,$75,$87,$FB,$F7,$F2,$67,$C6,$7E
	.dc.b	$C9,$B8,$7F,$03,$6C,$FF,$91,$5A,$FE,$C5,$AF,$EB,$DA,$B2,$46,$1B
	.dc.b	$5F,$C0,$3D,$CF,$C0,$63,$FA,$E6,$E1,$CE,$7F,$99,$46,$30,$7C,$3B
	.dc.b	$F2,$7F,$E1,$D7,$FB,$FF,$CA,$17,$FC,$F6,$B5,$E0,$7F,$3B,$9C,$D7
	.dc.b	$C6,$38,$7C,$17,$F1,$28,$FF,$61,$C7,$F2,$36,$E3,$FE,$65,$6B,$FC
	.dc.b	$47,$F0,$17,$B1,$FD,$1D,$B5,$D6,$7E,$14,$7F,$08,$D7,$F8,$5B,$A8
	.dc.b	$47,$F7,$1F,$71,$0F,$AF,$FE,$65,$18,$6C,$66,$FF,$DC,$8F,$09,$D4
	.dc.b	$62,$3F,$CF,$66,$CD,$C6,$9A,$64,$FF,$77,$39,$7F,$97,$7C,$BF,$E2
	.dc.b	$2F,$F9,$1B,$6B,$C1,$3D,$67,$3C,$BF,$DA,$7F,$00,$7D,$3D,$31,$F8
	.dc.b	$2D,$C3,$E3,$72,$AB,$71,$86,$E3,$F1,$A3,$F9,$E4,$2B,$94,$E4,$FC
	.dc.b	$87,$DF,$7F,$21,$AF,$E8,$4F,$E8,$9B,$EE,$FF,$7D,$8C,$4F,$08,$C7
	.dc.b	$5E,$5B,$5F,$E4,$1F,$C8,$1F,$57,$FE,$09,$F2,$1F,$90,$3E,$E3,$FC
	.dc.b	$19,$92,$F7,$87,$E0,$EF,$75,$A9,$CD,$7A,$67,$35,$DF,$0E,$D8,$75
	.dc.b	$8E,$4D,$AF,$81,$B3,$E9,$7F,$47,$1F,$D5,$E7,$5C,$5F,$D6,$A7,$3E
	.dc.b	$5B,$E7,$FE,$C6,$3F,$1B,$E5,$FF,$C4,$93,$9E,$8F,$8D,$18,$AD,$E5
	.dc.b	$F3,$A3,$13,$AB,$FB,$7F,$91,$F8,$1F,$EE,$4D,$AF,$E0,$8B,$E4,$FF
	.dc.b	$C5,$CE,$75,$34,$91,$AE,$1F,$F1,$C7,$57,$F9,$9F,$EA,$FF,$4B,$F0
	.dc.b	$06,$93,$2F,$D1,$4E,$DB,$3B,$6C,$EF,$2F,$F7,$17,$3B,$EA,$BF,$0C
	.dc.b	$78,$2F,$84,$FD,$09,$D5,$FE,$6B,$D0,$FB,$7C,$7F,$E0,$85,$F3,$7F
	.dc.b	$E4,$B7,$33,$FE,$15,$72,$F1,$E6,$D4,$ED,$B1,$FC,$2F,$91,$F5,$9F
	.dc.b	$D5,$32,$F1,$7C,$5F,$D8,$8C,$65,$FE,$18,$CD,$39,$F3,$BF,$40,$DC
	.dc.b	$7F,$63,$3A,$7F,$8C,$70,$7F,$A4,$FE,$4F,$E6,$92,$3E,$4B,$72,$FE
	.dc.b	$17,$BA,$31,$1F,$C2,$A3,$13,$FE,$4D,$DD,$F1,$BF,$0C,$F9,$97,$67
	.dc.b	$C7,$F8,$C7,$ED,$7F,$14,$FD,$B3,$63,$27,$A2,$32,$D3,$E9,$78,$9F
	.dc.b	$3F,$C3,$FA,$0D,$8C,$9F,$5E,$32,$D3,$ED,$7D,$77,$F9,$53,$23,$49
	.dc.b	$1E,$BF,$A3,$FB,$87,$DD,$8B,$1F,$83,$F8,$EF,$C0,$18,$DB,$F6,$F6
	.dc.b	$8C,$CE,$CF,$C5,$5E,$2B,$7E,$B7,$8C,$93,$F2,$DF,$37,$C0,$FE,$34
	.dc.b	$FC,$73,$63,$F2,$9F,$1F,$EB,$23,$F4,$DF,$77,$A7,$F9,$53,$F2,$CD
	.dc.b	$8F,$CE,$45,$BF,$4C,$DE,$7F,$CF,$4D,$AF,$CE,$4D,$AF,$F0,$16,$7F
	.dc.b	$E6,$99,$FC,$17,$8F,$F3,$48,$FE,$16,$7F,$4C,$7E,$9E,$2C,$7F,$B2
	.dc.b	$FC,$9F,$F4,$A7,$5B,$E2,$7F,$CC,$37,$FE,$CC,$9C,$C6,$23,$FB,$5C
	.dc.b	$EE,$72,$4A,$FF,$68,$BF,$E9,$2E,$97,$9D,$FD,$14,$FE,$47,$87,$F4
	.dc.b	$BC,$3F,$A6,$C7,$F2,$6F,$D6,$BD,$91,$FE,$F0,$CF,$BC,$63,$E1,$3F
	.dc.b	$CC,$52,$CD,$3F,$DF,$B3,$FE,$FF,$5A,$9C,$E9,$FE,$F3,$8C,$57,$3C
	.dc.b	$3A,$CE,$5E,$F8,$53,$B4,$ED,$B3,$9B,$EB,$87,$82,$F8,$B6,$B6,$D9
	.dc.b	$DC,$7B,$4C,$95,$F5,$BD,$6F,$C1,$0F,$27,$FD,$2B,$8F,$B5,$B8,$7D
	.dc.b	$47,$FC,$CB,$19,$D7,$CD,$1A,$C3,$E4,$36,$FA,$79,$DA,$A9,$97,$9E
	.dc.b	$B4,$74,$C7,$B4,$FC,$5C,$1F,$5E,$FA,$CF,$F5,$E4,$7B,$4C,$B8,$CF
	.dc.b	$EC,$FB,$0F,$A5,$FF,$71,$47,$DC,$AD,$EF,$F2,$5A,$73,$C3,$B4,$7C
	.dc.b	$DB,$E5,$EC,$F8,$E7,$35,$BC,$7C,$6F,$F8,$F2,$32,$D8,$CB,$FE,$46
	.dc.b	$3B,$9B,$FA,$0F,$F0,$93,$FE,$41,$94,$62,$35,$E6,$7F,$91,$5F,$87
	.dc.b	$19,$7A,$4F,$09,$F6,$36,$3D,$93,$DB,$2F,$8B,$EF,$FD,$F9,$DC,$E7
	.dc.b	$6E,$6D,$8F,$F2,$ED,$F3,$BE,$3B,$FA,$DF,$D1,$7E,$3B,$FD,$2E,$BF
	.dc.b	$24,$D7,$F2,$0D,$8D,$FC,$7F,$1C,$7F,$7F,$39,$6F,$8B,$9B,$FD,$94
	.dc.b	$E5,$B5,$9E,$54,$E5,$B5,$98,$C0,$9E,$1F,$96,$BF,$04,$A4,$C3,$F5
	.dc.b	$51,$FE,$15,$6C,$65,$FD,$54,$71,$77,$E8,$E7,$3F,$E0,$93,$99,$DD
	.dc.b	$78,$2B,$59,$FF,$1E,$AD,$5F,$0A,$D5,$FE,$F7,$D4,$76,$FF,$8C,$63
	.dc.b	$FC,$03,$E9,$FE,$2F,$C1,$FF,$04,$37,$AF,$A7,$D4,$5A,$FE,$35,$9A
	.dc.b	$3F,$15,$C2,$8C,$4F,$FA,$46,$31,$3A,$9C,$CF,$FA,$62,$72,$EE,$32
	.dc.b	$FE,$76,$32,$EE,$32,$FF,$93,$3B,$7C,$99,$FF,$92,$1B,$EA,$FF,$BD
	.dc.b	$27,$FC,$A7,$DD,$83,$F4,$15,$FE,$4E,$6C,$9A,$63,$97,$99,$2B,$FC
	.dc.b	$BB,$EA,$D7,$FC,$E3,$5F,$A1,$6C,$F8,$A7,$5B,$35,$FF,$31,$CE,$1F
	.dc.b	$F3,$2C,$EE,$BF,$D7,$18,$FF,$98,$A3,$CA,$48,$CE,$1F,$E3,$C7,$FC
	.dc.b	$DF,$E8,$FC,$03,$E0,$F9,$B2,$36,$FB,$88,$F2,$F3,$BE,$39,$1F,$F3
	.dc.b	$CF,$D6,$57,$DB,$17,$4F,$54,$7F,$9E,$A7,$E7,$91,$E2,$7F,$A0,$BE
	.dc.b	$37,$E9,$9F,$38,$8F,$A9,$FA,$3F,$21,$F6,$3E,$06,$4F,$FA,$2A,$3E
	.dc.b	$D9,$87,$E7,$C6,$1F,$F6,$4F,$D6,$FE,$97,$0F,$2E,$C8,$FC,$7C,$61
	.dc.b	$BB,$FF,$BB,$A7,$5D,$1F,$8A,$9E,$53,$97,$76,$AE,$9C,$DB,$BF,$C7
	.dc.b	$FC,$8C,$AF,$F5,$0B,$FA,$77,$FD,$C3,$E8,$5F,$50,$FF,$31,$8C,$9F
	.dc.b	$CC,$E3,$2D,$8D,$46,$23,$FD,$F8,$D7,$FE,$00,$B5,$FE,$88,$D7,$FC
	.dc.b	$21,$EC,$FC,$37,$E3,$F7,$FB,$CF,$DE,$75,$FF,$A3,$34,$B5,$BA,$FF
	.dc.b	$66,$C7,$F1,$96,$D6,$A3,$8F,$DA,$32,$BC,$32,$FF,$D3,$13,$96,$CE
	.dc.b	$5F,$F7,$C6,$29,$57,$4A,$FA,$9F,$25,$2F,$EF,$BF,$37,$EB,$DF,$37
	.dc.b	$4F,$98,$E2,$7F,$AF,$35,$FE,$D0,$03,$97,$2D,$7F,$B7,$35,$E0,$D7
	.dc.b	$8C,$7F,$0C,$F1,$DE,$AF,$F1,$46,$CF,$7E,$5F,$CA,$C6,$23,$FD,$F2
	.dc.b	$FF,$A0,$DD,$72,$FE,$56,$35,$3C,$2B,$A1,$D7,$D7,$FB,$3F,$02,$F5
	.dc.b	$FF,$35,$F1,$73,$2F,$F6,$44,$E5,$B3,$97,$EE,$CF,$28,$FC,$C7,$D6
	.dc.b	$BE,$CB,$F9,$6F,$E4,$AE,$8A,$6A,$3F,$3D,$F8,$E3,$F0,$8C,$65,$EA
	.dc.b	$3E,$76,$1C,$DC,$09,$E0,$E1,$AB,$6F,$E1,$EA,$79,$39,$E0,$FF,$11
	.dc.b	$46,$27,$75,$C9,$FE,$01,$E9,$7E,$F7,$D6,$98,$7F,$08,$8F,$E0,$06
	.dc.b	$1F,$CF,$C6,$1F,$57,$80,$FB,$3E,$D9,$7F,$09,$6C,$FF,$05,$6C,$EF
	.dc.b	$EC,$3F,$AA,$C6,$09,$AE,$23,$FA,$CF,$BB,$F8,$1B,$F6,$1F,$C3,$DC
	.dc.b	$67,$FC,$71,$B8,$FF,$90,$7F,$30,$FF,$75,$D7,$CC,$6B,$F9,$E0,$E1
	.dc.b	$FD,$0B,$5E,$5F,$09,$FC,$4E,$31,$1F,$A5,$6B,$CA,$34,$7F,$6A,$F7
	.dc.b	$BF,$82,$61,$FC,$C1,$B7,$FC,$B5,$B5,$FC,$A5,$FE,$5A,$FF,$2B,$F9
	.dc.b	$17,$C5,$7F,$74,$6C,$7F,$62,$6B,$FB,$8D,$91,$88,$FF,$31,$7B,$EF
	.dc.b	$CA,$9E,$7F,$F8,$13,$A7,$FB,$35,$7C,$CC,$3F,$4D,$E8,$7F,$86,$F0
	.dc.b	$E1,$7F,$E0,$AF,$95,$FF,$1E,$5F,$E0,$1A,$F1,$6E,$DF,$D3,$97,$2F
	.dc.b	$13,$FC,$1A,$E2,$7F,$BD,$B5,$F5,$35,$FF,$3D,$62,$8C,$47,$F5,$56
	.dc.b	$BF,$D4,$5B,$1B,$9F,$F0,$DF,$CB,$7E,$12,$BF,$AB,$36,$4D,$F8,$0F
	.dc.b	$E6,$53,$F0,$DA,$F8,$9A,$FF,$65,$6B,$FD,$85,$B2,$69,$3C,$6F,$F1
	.dc.b	$6B,$AF,$98,$7C,$A7,$F8,$D9,$B7,$FA,$C6,$BF,$57,$E6,$5E,$AF,$F8
	.dc.b	$F1,$B1,$FE,$24,$6B,$F8,$36,$BF,$29,$AF,$F7,$D6,$BF,$DE,$5B,$89
	.dc.b	$A6,$3E,$6B,$F2,$B0,$69,$8F,$7B,$F2,$7B,$FF,$83,$9F,$C9,$5F,$F8
	.dc.b	$2D,$AF,$F8,$1D,$B4,$69,$39,$BF,$CA,$6D,$9F,$F2,$0C,$BC,$6F,$F2
	.dc.b	$EB,$63,$F8,$9B,$5F,$CE,$79,$FF,$7C,$FE,$14,$FE,$31,$AF,$F8,$7D
	.dc.b	$AF,$F8,$6D,$B0,$69,$3C,$8F,$F3,$3B,$AF,$9D,$7D,$07,$F9,$C1,$B7
	.dc.b	$FC,$99,$AF,$EB,$FE,$09,$F3,$5F,$E7,$66,$C7,$F9,$79,$AF,$EF,$5A
	.dc.b	$FE,$91,$AF,$F8,$DD,$AF,$F8,$CD,$B8,$9A,$63,$E4,$9E,$1F,$FA,$19
	.dc.b	$B5,$FC,$F1,$AF,$F0,$5F,$CA,$9F,$81,$FF,$A3,$1B,$1F,$E7,$46,$BF
	.dc.b	$C4,$DA,$FE,$E9,$AF,$F9,$37,$B7,$FC,$93,$89,$A4,$F1,$FF,$D3,$2E
	.dc.b	$3D,$CA,$62,$3F,$D3,$BF,$29,$FD,$C6,$FF,$A9,$B5,$FE,$38,$DE,$5F
	.dc.b	$F3,$0A,$F9,$4E,$3E,$84,$AE,$2F,$48,$FF,$20,$FD,$1B,$CE,$A4,$AF
	.dc.b	$EF,$5C,$AF,$92,$FF,$64,$47,$F8,$E2,$B8,$3A,$6B,$9A,$FF,$26,$CF
	.dc.b	$0C,$BF,$D1,$37,$CF,$A7,$FD,$93,$3D,$2F,$FD,$80,$D7,$FC,$6D,$3B
	.dc.b	$7F,$D9,$9F,$51,$FE,$24,$F3,$32,$D8,$FC,$D3,$67,$D1,$F1,$9F,$E5
	.dc.b	$E8,$C1,$3D,$7F,$2C,$7F,$9D,$FE,$9F,$FC,$97,$9F,$F6,$9C,$11,$87
	.dc.b	$FB,$63,$5F,$F2,$5B,$5F,$F2,$61,$C3,$C6,$70,$F3,$9F,$F3,$33,$63
	.dc.b	$33,$D1,$AF,$87,$AB,$FD,$0C,$FE,$74,$FF,$44,$38,$3E,$1F,$D3,$FF
	.dc.b	$AD,$30,$F9,$B9,$A3,$1A,$FC,$AD,$9F,$5B,$8C,$7E,$3D,$5D,$D8,$71
	.dc.b	$C3,$F0,$2D,$AF,$F3,$BE,$1C,$A7,$37,$C6,$1F,$51,$FE,$38,$AF,$03
	.dc.b	$5F,$E1,$D1,$AA,$FF,$0B,$61,$F8,$DA,$D2,$E3,$7C,$8F,$95,$B5,$F9
	.dc.b	$4B,$E5,$87,$E4,$EB,$44,$9D,$61,$67,$FC,$05,$A7,$F1,$A9,$FF,$1E
	.dc.b	$B6,$33,$F5,$1F,$17,$AF,$FC,$A5,$E1,$5A,$D1,$FE,$CE,$2D,$FF,$B2
	.dc.b	$D5,$6E,$79,$7D,$47,$FA,$96,$BF,$C9,$6D,$C3,$FC,$28,$FF,$A8,$23
	.dc.b	$3C,$5C,$08,$F0,$F8,$3B,$5F,$E1,$B8,$FF,$08,$9F,$E3,$FC,$8E,$23
	.dc.b	$FC,$50,$D7,$FC,$1E,$D7,$F9,$67,$CD,$FB,$F9,$0F,$C1,$9E,$5A,$7F
	.dc.b	$BB,$27,$31,$FC,$FF,$E1,$3E,$E6,$0F,$19,$18,$6C,$F9,$5B,$8F,$F8
	.dc.b	$ED,$AF,$F8,$89,$AF,$FB,$D5,$AF,$FB,$C5,$B4,$F9,$97,$C6,$FD,$76
	.dc.b	$DF,$94,$FF,$3C,$B8,$BF,$EC,$FF,$51,$E8,$E6,$34,$95,$F2,$E3,$2D
	.dc.b	$8C,$CF,$F9,$4E,$72,$FF,$95,$7E,$A1,$F2,$9F,$89,$6D,$FF,$A2,$AB
	.dc.b	$7B,$7D,$6B,$66,$BD,$C7,$F9,$5F,$CF,$FB,$78,$7D,$D6,$C7,$F9,$8D
	.dc.b	$AF,$F9,$29,$AF,$FB,$76,$BF,$C3,$53,$F5,$A7,$33,$E0,$6C,$EA,$7F
	.dc.b	$C5,$DD,$FF,$A1,$F4,$0F,$E8,$EB,$C3,$8F,$A6,$FF,$C5,$53,$FA,$7F
	.dc.b	$4B,$FD,$C3,$87,$E4,$DB,$8F,$F9,$E5,$AF,$F9,$81,$AF,$9D,$AF,$C4
	.dc.b	$6D,$1A,$63,$EA,$1F,$41,$FB,$76,$DF,$F9,$37,$C4,$C4,$FE,$EF,$D6
	.dc.b	$7C,$BF,$E6,$1B,$5F,$A9,$6E,$5F,$E8,$F6,$BF,$E6,$59,$E0,$49,$E0
	.dc.b	$FE,$8B,$D8,$FF,$2F,$C7,$FB,$21,$B3,$FE,$6F,$9F,$E1,$91,$8D,$7F
	.dc.b	$D1,$5F,$13,$FA,$AF,$6C,$D2,$61,$FE,$8A,$6E,$1F,$E8,$A6,$BF,$E8
	.dc.b	$AF,$7B,$EB,$CF,$FB,$61,$B1,$FE,$99,$9F,$E3,$91,$8B,$FC,$97,$A2
	.dc.b	$FA,$8F,$DF,$E1,$F8,$9F,$81,$5F,$F4,$7A,$D7,$BD,$FC,$96,$73,$5F
	.dc.b	$EE,$67,$FC,$7E,$FF,$AE,$7D,$9F,$E0,$B1,$8C,$FF,$D7,$4D,$7F,$D4
	.dc.b	$3F,$50,$F9,$5F,$E6,$0D,$BF,$F7,$8B,$FB,$9D,$3F,$20,$E6,$BF,$DC
	.dc.b	$FE,$D1,$7D,$27,$F4,$F8,$C3,$63,$37,$FE,$E5,$9C,$B6,$32,$7F,$7A
	.dc.b	$11,$A7,$0D,$7F,$BF,$B5,$FD,$87,$B9,$F7,$66,$FC,$0F,$F6,$FA,$E5
	.dc.b	$8F,$FB,$EE,$3F,$A7,$7D,$43,$DA,$A6,$3D,$AB,$87,$B6,$FF,$61,$5B
	.dc.b	$8C,$1F,$BF,$71,$97,$FB,$7D,$FF,$21,$46,$A3,$F8,$7C,$6A,$7F,$75
	.dc.b	$F3,$BF,$C4,$1B,$D1,$FD,$86,$31,$7F,$C2,$DF,$D2,$CF,$0F,$13,$55
	.dc.b	$B6,$D7,$7C,$1F,$50,$FA,$6F,$F0,$87,$90,$FA,$87,$D1,$FF,$84,$EF
	.dc.b	$75,$A9,$CE,$1C,$F0,$E4,$FA,$F0,$E5,$87,$19,$CE,$0C,$FF,$97,$56
	.dc.b	$AF,$86,$1D,$2B,$51,$FC,$9F,$EA,$3F,$43,$1F,$D3,$67,$57,$FC,$7E
	.dc.b	$73,$E5,$7F,$AD,$79,$7E,$37,$D0,$FF,$8A,$7E,$1F,$F1,$37,$F3,$23
	.dc.b	$13,$AC,$BF,$CD,$75,$A7,$E6,$4F,$F8,$3F,$E8,$3B,$C7,$F3,$96,$E7
	.dc.b	$FD,$19,$FF,$65,$47,$DB,$F2,$51,$A6,$EF,$CC,$E9,$F0,$56,$BF,$05
	.dc.b	$D3,$FE,$54,$F9,$CF,$D3,$7C,$11,$A4,$9F,$D3,$46,$5B,$19,$CB,$F4
	.dc.b	$73,$97,$F5,$18,$B9,$DF,$43,$F8,$B3,$F4,$55,$BE,$0F,$C8,$C7,$F8
	.dc.b	$BF,$CE,$FB,$31,$FD,$35,$7C,$5E,$7A,$34,$DD,$3E,$14,$E9,$B3,$A7
	.dc.b	$FD,$33,$F3,$DF,$D1,$34,$FB,$E7,$F7,$0C,$DF,$1F,$FE,$61,$9D,$65
	.dc.b	$FA,$48,$C7,$9D,$FA,$36,$E1,$F4,$6F,$8B,$F3,$35,$7D,$27,$CC,$24
	.dc.b	$7C,$A6,$E5,$F2,$1A,$FF,$0A,$9E,$33,$FE,$3B,$D1,$F2,$7F,$14,$F9
	.dc.b	$37,$67,$C5,$F4,$5F,$B7,$F9,$2F,$DC,$36,$32,$79,$E3,$2D,$3E,$4F
	.dc.b	$89,$F4,$7C,$4F,$A4,$D8,$C9,$F3,$A3,$2D,$3E,$8F,$D4,$7F,$90,$B2
	.dc.b	$34,$91,$EB,$FA,$2F,$BC,$7D,$F8,$B1,$F6,$BE,$4F,$F0,$46,$36,$FA
	.dc.b	$FC,$FF,$E1,$28,$CC,$EC,$FC,$0D,$E2,$B7,$87,$F8,$20,$9E,$B7,$94
	.dc.b	$F3,$7C,$0F,$E3,$CF,$C8,$36,$FF,$17,$F5,$1E,$B2,$3F,$4D,$F7,$7A
	.dc.b	$7F,$97,$3F,$30,$D8,$FC,$9C,$5C,$3C,$73,$79,$FF,$39,$36,$BF,$C6
	.dc.b	$D8,$AB,$17,$FE,$44,$8F,$F2,$F4,$FE,$0F,$C7,$F9,$C4,$7F,$0D,$3F
	.dc.b	$A9,$3F,$55,$16,$3F,$D4,$DE,$4F,$FA,$B3,$E2,$BF,$C9,$B5,$FE,$EE
	.dc.b	$8C,$12,$7F,$D6,$73,$A9,$FF,$64,$4E,$5F,$DE,$F4,$3C,$EF,$E8,$C7
	.dc.b	$FB,$06,$6D,$FF,$4F,$F2,$1F,$51,$EA,$C1,$F0,$9F,$E4,$C5,$3A,$CF
	.dc.b	$FC,$C9,$39,$8E,$33,$8D,$3F,$BF,$46,$F0,$FF,$00,$56,$F0,$EC,$DC
	.dc.b	$38,$61,$DA,$7F,$C0,$D5,$C3,$0E,$71,$C1,$B5,$C3,$0E,$D7,$E8,$9D
	.dc.b	$B9,$AE,$0F,$5A,$C1,$96,$DF,$77,$DF,$E4,$7F,$9F,$B1,$F7,$37,$0F
	.dc.b	$6B,$FE,$52,$8C,$E5,$E3,$8D,$57,$28,$C7,$9D,$96,$61,$DF,$35,$F0
	.dc.b	$6D,$6A,$FD,$87,$F8,$A7,$E8,$1F,$51,$FE,$AD,$8C,$88,$CE,$9E,$BF
	.dc.b	$61,$F3,$5F,$EC,$FD,$7E,$E5,$EF,$0E,$51,$8C,$38,$CF,$46,$C7,$E0
	.dc.b	$EF,$9F,$B3,$F6,$2B,$96,$3F,$25,$A7,$F8,$D4,$A7,$F8,$DE,$FA,$B7
	.dc.b	$0F,$53,$9F,$94,$FF,$07,$BC,$65,$18,$8D,$79,$DD,$5F,$81,$19,$7F
	.dc.b	$C9,$C6,$23,$C9,$3C,$9B,$3D,$B2,$F6,$7B,$FF,$4A,$F3,$51,$F8,$55
	.dc.b	$BE,$2F,$F2,$D6,$8B,$E1,$5F,$5B,$77,$D4,$7F,$4E,$8F,$BC,$D7,$4D
	.dc.b	$CB,$F5,$4B,$E4,$3D,$31,$8F,$1A,$E5,$FA,$CA,$E0,$D8,$CD,$FE,$DA
	.dc.b	$79,$3E,$08,$CE,$1E,$46,$CF,$29,$E1,$86,$BE,$A1,$F5,$14,$93,$F9
	.dc.b	$78,$FF,$09,$36,$32,$FC,$18,$E2,$EF,$D1,$CE,$7F,$C1,$E7,$33,$BA
	.dc.b	$F0,$56,$B3,$FE,$13,$5A,$BE,$15,$AB,$FD,$B7,$89,$FE,$19,$F4,$FE
	.dc.b	$FF,$83,$FE,$66,$6F,$57,$F9,$8D,$B5,$FC,$6B,$17,$CC,$E3,$81,$18
	.dc.b	$9F,$CC,$46,$27,$F1,$CD,$8C,$CE,$DF,$F4,$94,$65,$B1,$97,$F2,$71
	.dc.b	$96,$C6,$5F,$F2,$57,$57,$9B,$2F,$F2,$3B,$7D,$5F,$F7,$1C,$FF,$92
	.dc.b	$D7,$6E,$4F,$D1,$57,$F9,$39,$B1,$A9,$FD,$94,$E5,$B1,$F9,$8D,$1A
	.dc.b	$7F,$96,$FB,$7C,$07,$87,$D4,$B7,$E1,$8A,$FF,$80,$2F,$FC,$92,$E4
	.dc.b	$86,$63,$93,$E9,$9F,$29,$1F,$1B,$E8,$F4,$3E,$01,$23,$E0,$79,$BF
	.dc.b	$0C,$F6,$91,$F5,$BF,$03,$CE,$F9,$04,$8F,$91,$F3,$95,$F4,$45,$CF
	.dc.b	$E0,$C7,$F9,$E6,$7E,$81,$1E,$27,$D1,$F8,$CF,$A8,$7C,$E2,$3F,$31
	.dc.b	$FA,$3F,$1F,$F6,$7E,$06,$4F,$D9,$8F,$B8,$61,$F9,$51,$87,$FD,$11
	.dc.b	$F5,$1F,$A6,$C3,$CB,$C0,$8F,$C8,$46,$1B,$C1,$FE,$ED,$9D,$73,$FE
	.dc.b	$2E,$79,$4E,$5D,$DA,$BA,$73,$6E,$FF,$21,$F2,$32,$BF,$20,$FE,$A9
	.dc.b	$FF,$71,$7A,$17,$D4,$3F,$CC,$E3,$27,$F3,$58,$CB,$63,$51,$88,$FE
	.dc.b	$8C,$D7,$FD,$F8,$D7,$F9,$DB,$5F,$F0,$9F,$B1,$F0,$DF,$8F,$DD,$EF
	.dc.b	$3F,$79,$19,$8F,$F4,$5F,$5A,$31,$B7,$C9,$AE,$15,$FE,$92,$D7,$F8
	.dc.b	$27,$D4,$33,$BC,$32,$BF,$C5,$13,$96,$CE,$5E,$F0,$9E,$73,$F4,$FE
	.dc.b	$4A,$57,$DE,$7E,$67,$D4,$3E,$A2,$9F,$2D,$C4,$FF,$5F,$6B,$FD,$B5
	.dc.b	$AF,$F6,$00,$E1,$FE,$C6,$D4,$6A,$3B,$1F,$C2,$BC,$A7,$AB,$FC,$4D
	.dc.b	$B3,$FE,$DB,$6B,$FE,$80,$5E,$BE,$E6,$4E,$DE,$51,$FC,$BF,$E0,$1E
	.dc.b	$BF,$E5,$7E,$2E,$5F,$FB,$C2,$34,$61,$FF,$46,$EE,$9C,$C7,$AB,$87
	.dc.b	$F7,$2B,$E9,$9F,$3F,$FC,$AF,$EA,$1C,$F4,$FC,$D4,$EA,$B9,$76,$7F
	.dc.b	$82,$3E,$9D,$F5,$1A,$C6,$56,$8F,$93,$3A,$70,$1C,$3F,$ED,$A3,$C9
	.dc.b	$1C,$9C,$46,$B2,$FD,$D7,$A7,$FB,$7F,$5E,$57,$F0,$28,$FD,$D1,$87
	.dc.b	$E1,$C6,$1F,$27,$80,$FA,$8F,$64,$7F,$AD,$1B,$7F,$C2,$9B,$9F,$F0
	.dc.b	$75,$F5,$1F,$D4,$5C,$D7,$F0,$56,$DF,$5C,$F5,$5F,$D5,$FD,$CF,$C0
	.dc.b	$DF,$89,$AF,$FB,$11,$B8,$FD,$6F,$A8,$FF,$73,$CF,$C8,$6B,$FE,$47
	.dc.b	$6B,$F9,$B6,$BF,$E3,$C6,$BF,$9A,$6A,$69,$3E,$13,$FB,$2F,$BB,$FE
	.dc.b	$18,$9F,$F6,$BB,$6B,$F9,$6B,$74,$FE,$54,$FF,$28,$9F,$E4,$DF,$22
	.dc.b	$F8,$9F,$ED,$ED,$8F,$ED,$AD,$7F,$65,$AA,$31,$1F,$E5,$4F,$71,$F5
	.dc.b	$0F,$3F,$FC,$01,$D6,$F4,$29,$FE,$18,$AD,$CF,$F8,$17,$E5,$7F,$C7
	.dc.b	$B8,$7F,$BC,$1B,$B7,$F4,$66,$ED,$FD,$0D,$72,$F1,$3F,$C0,$EE,$2B
	.dc.b	$FC,$0A,$D7,$D2,$D7,$FC,$D7,$AA,$31,$1D,$67,$99,$1F,$F0,$7F,$CB
	.dc.b	$7E,$02,$FF,$AA,$36,$3F,$A8,$36,$FF,$CD,$5E,$37,$F8,$87,$DF,$FE
	.dc.b	$59,$5F,$29,$AF,$89,$AF,$F5,$F6,$BF,$D6,$DB,$3D,$23,$FC,$43,$18
	.dc.b	$33,$F3,$0F,$94,$FF,$18,$36,$7F,$56,$D5,$A7,$F7,$2F,$9A,$7A,$BF
	.dc.b	$E3,$76,$CF,$F8,$B9,$AF,$E0,$9A,$FC,$A6,$BF,$DD,$DA,$FF,$73,$6E
	.dc.b	$26,$98,$F9,$AF,$CA,$41,$A6,$3D,$EF,$C9,$61,$FE,$08,$7F,$23,$7F
	.dc.b	$E0,$56,$BF,$E0,$16,$D1,$A4,$E7,$FF,$26,$B6,$7F,$C9,$72,$F1,$FF
	.dc.b	$CB,$0D,$8F,$E2,$6D,$7E,$4F,$07,$F8,$57,$D2,$FC,$01,$FC,$1A,$3F
	.dc.b	$1A,$D7,$5E,$4F,$F9,$72,$31,$1F,$E1,$76,$D1,$A4,$9F,$F3,$1F,$CE
	.dc.b	$BE,$7F,$FC,$D8,$DB,$FE,$4E,$D7,$CA,$7F,$88,$FE,$05,$F3,$5F,$E7
	.dc.b	$36,$C7,$F9,$A9,$AF,$EF,$9A,$FE,$99,$AF,$F8,$CD,$AF,$58,$FF,$3B
	.dc.b	$C6,$31,$37,$E5,$1E,$1F,$FA,$0D,$B7,$FC,$F9,$AF,$A2,$9F,$4E,$FA
	.dc.b	$8F,$F4,$53,$67,$FD,$02,$D7,$F8,$AB,$5F,$DE,$B5,$FF,$25,$77,$7F
	.dc.b	$91,$A8,$D2,$78,$FF,$E9,$77,$1E,$E5,$31,$1F,$E9,$BF,$94,$FE,$E3
	.dc.b	$7F,$D5,$9A,$FA,$5B,$CD,$FE,$5E,$5F,$29,$C7,$D0,$95,$C5,$E9,$1F
	.dc.b	$EB,$9F,$A4,$79,$D4,$95,$FD,$DF,$99,$F2,$5F,$9D,$8C,$57,$06,$D7
	.dc.b	$07,$FC,$8F,$3B,$6F,$57,$FB,$0A,$31,$1F,$EB,$C6,$BF,$E2,$CB,$EF
	.dc.b	$3F,$EC,$8F,$A6,$FF,$13,$E9,$FE,$8E,$94,$63,$AB,$83,$6F,$F3,$98
	.dc.b	$3E,$2F,$FD,$AD,$39,$BE,$3E,$C6,$0F,$F6,$D5,$6E,$FF,$C2,$7F,$51
	.dc.b	$FE,$4B,$CB,$FC,$89,$1F,$E4,$08,$C3,$86,$BF,$E4,$96,$BF,$E4,$60
	.dc.b	$FF,$98,$F8,$39,$C6,$3C,$64,$62,$75,$3D,$67,$2F,$FA,$17,$AB,$FD
	.dc.b	$0C,$FE,$80,$C1,$0F,$F7,$3F,$07,$FB,$53,$6F,$AE,$7F,$AD,$B1,$F9
	.dc.b	$5F,$52,$F9,$8C,$4F,$D1,$AE,$17,$FE,$80,$6D,$FD,$16,$CF,$D7,$BE
	.dc.b	$4F,$63,$F1,$0F,$EA,$DF,$93,$B2,$31,$7F,$F0,$01,$AF,$FB,$F9,$AF
	.dc.b	$DC,$F1,$FC,$AF,$F1,$7E,$0F,$EA,$17,$DB,$AF,$38,$7E,$12,$ED,$6A
	.dc.b	$FF,$D3,$2D,$CB,$E0,$DE,$AB,$AF,$53,$ED,$1F,$27,$86,$E7,$FA,$3F
	.dc.b	$C5,$BF,$9F,$3F,$E4,$88,$CE,$7F,$C2,$DB,$97,$FB,$41,$B3,$F9,$DA
	.dc.b	$E3,$5B,$BF,$6B,$E1,$F9,$9B,$66,$FA,$58,$66,$1F,$E6,$89,$D3,$67
	.dc.b	$35,$FC,$D3,$17,$E8,$0F,$C1,$3F,$CF,$F3,$FE,$81,$8C,$5F,$E7,$AB
	.dc.b	$75,$AA,$FE,$F3,$5A,$7F,$9C,$74,$FC,$8E,$5F,$F4,$FE,$1F,$E7,$DC
	.dc.b	$38,$BF,$E6,$C8,$CD,$7F,$46,$F9,$97,$DA,$3E,$53,$B4,$61,$B1,$9B
	.dc.b	$FD,$8C,$E5,$B1,$9A,$FF,$5C,$4E,$5B,$19,$7F,$65,$19,$7F,$A7,$69
	.dc.b	$E9,$3F,$1B,$1C,$71,$5F,$F5,$19,$FC,$8A,$FC,$A7,$FD,$62,$FF,$04
	.dc.b	$31,$1F,$1D,$A6,$63,$77,$FE,$CD,$8C,$B6,$32,$FF,$07,$8C,$B6,$32
	.dc.b	$FF,$B4,$38,$9F,$52,$F9,$7F,$CC,$B6,$FF,$D5,$C7,$6C,$7D,$76,$BE
	.dc.b	$37,$EC,$C1,$F3,$FE,$96,$E1,$E9,$8C,$11,$C1,$2F,$D5,$1F,$87,$BF
	.dc.b	$2F,$CB,$79,$0E,$D1,$6B,$CB,$DB,$E7,$BF,$88,$D7,$E1,$35,$FF,$77
	.dc.b	$C7,$F1,$7C,$7E,$74,$E6,$7E,$BB,$67,$47,$D5,$75,$83,$EA,$7F,$C3
	.dc.b	$93,$FE,$C0,$6D,$7F,$86,$A3,$8B,$FE,$05,$8C,$E7,$F2,$A7,$F0,$5E
	.dc.b	$3F,$C7,$23,$F7,$DB,$5F,$75,$AF,$DB,$7F,$59,$E6,$7E,$3B,$EE,$FF
	.dc.b	$8E,$B6,$BE,$CE,$DF,$EF,$07,$FD,$D1,$19,$BF,$F7,$27,$A0,$DF,$F6
	.dc.b	$D0,$8C,$57,$F4,$E6,$C7,$F5,$A6,$BF,$D5,$4E,$13,$CA,$33,$1F,$D6
	.dc.b	$5A,$8D,$CF,$F5,$AF,$CF,$7F,$8E,$C6,$5A,$E8,$FF,$7D,$38,$E3,$7C
	.dc.b	$1F,$F2,$06,$D7,$5D,$7F,$2F,$7E,$BE,$1D,$5E,$FE,$B3,$DF,$81,$A4
	.dc.b	$C3,$FB,$6F,$C0,$BE,$93,$F9,$6B,$63,$B1,$FD,$D7,$C5,$FC,$23,$8C
	.dc.b	$BD,$D3,$B7,$FC,$BB,$DE,$FD,$5D,$F8,$4F,$F2,$58,$C3,$63,$37,$FE
	.dc.b	$08,$9C,$B6,$33,$5F,$46,$72,$D8,$CB,$FE,$09,$8C,$BF,$E0,$BF,$AF
	.dc.b	$7E,$B0,$FC,$23,$FB,$2C,$61,$FE,$CF,$5D,$38,$7F,$80,$CE,$71,$F7
	.dc.b	$75,$3E,$BF,$E8,$E5,$FE,$63,$6D,$FF,$08,$E7,$FC,$07,$8B,$FD,$9A
	.dc.b	$73,$87,$F1,$36,$DF,$F0,$D6,$D7,$09,$EF,$1A,$C3,$FC,$EB,$E4,$FF
	.dc.b	$70,$F6,$BF,$BA,$57,$E3,$6F,$F1,$DF,$59,$9A,$93,$E8,$AD,$E2,$BD
	.dc.b	$CE,$6B,$F9,$DF,$E9,$BF,$B8,$8F,$CC,$73,$A3,$17,$F9,$BF,$B0,$FE
	.dc.b	$AB,$B7,$F2,$6A,$D7,$BF,$C8,$98,$7E,$16,$D7,$C2,$3A,$E1,$D3,$57
	.dc.b	$D7,$BE,$5F,$FC,$1A,$DA,$FC,$76,$5D,$9F,$A7,$19,$C7,$FD,$29,$B7
	.dc.b	$F8,$9B,$4F,$E7,$38,$FF,$A7,$3D,$1F,$E0,$FE,$BF,$E6,$2F,$71,$F2
	.dc.b	$C9,$8F,$F0,$89,$D5,$F2,$C3,$A4,$FF,$4B,$FA,$7F,$F5,$3B,$F8,$46
	.dc.b	$E1,$FE,$BC,$E1,$FF,$2A,$FC,$CF,$F8,$9A,$73,$87,$91,$B7,$CF,$0F
	.dc.b	$2D,$6D,$FF,$1C,$D3,$F1,$5F,$D6,$63,$FE,$65,$8C,$77,$FF,$83,$DB
	.dc.b	$87,$F8,$4E,$B7,$5F,$C1,$5A,$A7,$6F,$F1,$B6,$BD,$63,$7D,$3F,$E7
	.dc.b	$2E,$EF,$D2,$FB,$47,$CB,$3F,$BF,$BF,$EC,$F5,$BC,$3F,$4F,$F5,$FF
	.dc.b	$CA,$A7,$E9,$B6,$3F,$31,$BB,$E8,$BF,$C9,$73,$9C,$FF,$B8,$37,$88
	.dc.b	$E7,$87,$FD,$03,$5C,$3C,$8F,$D2,$78,$93,$2F,$B5,$F0,$57,$2F,$C2
	.dc.b	$CD,$D3,$F6,$DA,$2B,$17,$FA,$0F,$B1,$FF,$51,$3F,$E2,$A6,$E1,$FE
	.dc.b	$28,$6B,$FE,$26,$6B,$C7,$E0,$3F,$C6,$B1,$88,$74,$A3,$0D,$9F,$E4
	.dc.b	$9F,$90,$FF,$78,$AF,$E5,$2D,$CB,$F5,$D7,$E2,$7E,$7C,$67,$0F,$F0
	.dc.b	$F7,$33,$C0,$FE,$0C,$7E,$83,$E7,$2C,$FF,$03,$8B,$8F,$E5,$F5,$56
	.dc.b	$31,$FC,$B7,$D0,$7E,$64,$FD,$4F,$83,$71,$FD,$04,$5C,$39,$4D,$AF
	.dc.b	$F6,$87,$E6,$7F,$E9,$49,$79,$68,$C3,$7B,$FF,$CD,$0D,$C7,$FD,$7F
	.dc.b	$E0,$3E,$93,$FC,$FE,$D8,$FF,$77,$5F,$ED,$30,$FE,$DD,$39,$7F,$3D
	.dc.b	$3F,$EE,$5F,$C3,$3F,$DF,$AF,$FB,$D5,$B1,$A3,$8D,$FF,$A9,$23,$39
	.dc.b	$7F,$BA,$5B,$5C,$FE,$27,$C7,$5B,$7F,$D5,$7E,$3F,$F1,$3C,$3F,$B3
	.dc.b	$35,$FF,$7B,$35,$F8,$6D,$EE,$7C,$F3,$E5,$3F,$A9,$B7,$1F,$ED,$DB
	.dc.b	$2F,$F9,$06,$9F,$B0,$AD,$63,$EB,$A7,$E2,$4F,$A9,$F9,$AD,$C7,$FD
	.dc.b	$9A,$DA,$FF,$0E,$57,$0A,$FF,$3A,$76,$3E,$37,$FC,$F6,$DC,$3C,$78
	.dc.b	$71,$CB,$FC,$7D,$39,$C7,$FC,$88,$D8,$CD,$FF,$B4,$2B,$85,$FF,$95
	.dc.b	$AB,$4D,$9E,$E6,$F0,$B1,$F0,$BE,$45,$F3,$5F,$88,$6C,$7F,$BA,$1B
	.dc.b	$97,$B1,$AF,$F8,$CC,$FF,$3E,$79,$6F,$B9,$FE,$6C,$7F,$B9,$27,$8E
	.dc.b	$9F,$CD,$E3,$FD,$CF,$FA,$AF,$F2,$1C,$3F,$DE,$AD,$C3,$E4,$B5,$FF
	.dc.b	$13,$35,$FF,$21,$6C,$8C,$47,$E9,$1B,$19,$F2,$BF,$47,$1A,$AF,$E2
	.dc.b	$DF,$59,$F4,$8F,$1C,$D8,$FE,$DD,$F1,$5F,$E3,$D7,$F0,$FF,$66,$F9
	.dc.b	$8F,$F4,$DF,$C5,$3C,$7F,$85,$7B,$D3,$F1,$17,$BC,$3F,$CB,$17,$B9
	.dc.b	$FF,$2D,$56,$F6,$FF,$52,$EA,$C3,$85,$2F,$83,$FE,$F3,$F2,$7F,$D5
	.dc.b	$77,$C2,$B7,$5D,$6F,$96,$7E,$6C,$FE,$3D,$69,$F3,$E1,$CB,$3C,$D6
	.dc.b	$F2,$FE,$1B,$7B,$6C,$65,$F1,$4F,$C4,$8C,$66,$FB,$BE,$59,$3F,$2B
	.dc.b	$78,$9A,$9C,$C6,$30,$F8,$BA,$7C,$A9,$CD,$6E,$73,$97,$F2,$0B,$E0
	.dc.b	$FC,$5B,$E1,$3A,$C3,$C1,$7C,$6B,$47,$17,$D5,$F5,$6F,$B7,$FE,$97
	.dc.b	$3A,$11,$9F,$37,$C3,$F3,$4F,$A7,$FD,$1F,$17,$D4,$AD,$5F,$16,$C6
	.dc.b	$70,$EA,$FE,$06,$F9,$75,$78,$A7,$3A,$FE,$AE,$F9,$E1,$FE,$A5,$6B
	.dc.b	$99,$F3,$B9,$6B,$C7,$B1,$F2,$7F,$EA,$D8,$FF,$54,$57,$F1,$18,$C4
	.dc.b	$7B,$A7,$6E,$E7,$4D,$8C,$E5,$EF,$6E,$1D,$EF,$E2,$FD,$13,$ED,$3F
	.dc.b	$82,$5E,$E7,$EC,$E1,$DD,$B8,$F6,$9F,$89,$7C,$6B,$3E,$A7,$2C,$7F
	.dc.b	$7F,$C6,$F9,$CC,$57,$E0,$DA,$FD,$8A,$B8,$FF,$88,$E3,$F0,$95,$C5
	.dc.b	$B5,$AA,$91,$82,$1F,$6D,$EF,$C7,$FB,$5F,$9D,$BD,$26,$9E,$0F,$17
	.dc.b	$FC,$99,$CF,$FB,$08,$CE,$1F,$95,$AE,$0F,$C2,$AD,$BF,$59,$B7,$DE
	.dc.b	$38,$57,$57,$FD,$05,$9F,$F5,$E7,$F8,$AC,$76,$7F,$25,$5D,$30,$7D
	.dc.b	$CF,$F0,$AE,$D4,$7F,$BA,$CC,$ED,$B9,$D4,$7F,$A6,$3F,$37,$FF,$67
	.dc.b	$E1,$FE,$88,$6B,$FE,$F9,$E9,$46,$23,$FA,$24,$63,$E2,$7F,$1E,$D8
	.dc.b	$FC,$A3,$5E,$53,$C5,$F0,$4F,$08,$C4,$7F,$4E,$F8,$FF,$8F,$EE,$BC
	.dc.b	$74,$61,$B3,$FD,$FB,$D4,$7D,$17,$CC,$6C,$7E,$97,$89,$18,$8F,$EF
	.dc.b	$2D,$7F,$DE,$CD,$7E,$47,$E0,$9C,$1F,$F7,$C6,$A8,$C4,$7F,$88,$E3
	.dc.b	$13,$F7,$B5,$F8,$FE,$4F,$D5,$BD,$F0,$FD,$4C,$3E,$AC,$63,$0E,$F8
	.dc.b	$71,$DB,$FB,$75,$78,$2B,$4F,$09,$E8,$6F,$09,$1F,$42,$B8,$BF,$42
	.dc.b	$3E,$71,$FC,$13,$D0,$7E,$0E,$F7,$BF,$2B,$18,$6C,$66,$BF,$C5,$7F
	.dc.b	$96,$57,$FC,$50,$D7,$FC,$0D,$B2,$31,$1F,$DE,$9A,$FF,$97,$1B,$0F
	.dc.b	$59,$E4,$FE,$91,$B5,$FE,$51,$8F,$D4,$AF,$C6,$9F,$39,$B9,$93,$B3
	.dc.b	$FD,$D5,$3A,$6C,$F2,$8C,$47,$F9,$6C,$E9,$1F,$47,$CA,$91,$F2,$DF
	.dc.b	$CB,$11,$FE,$61,$CF,$F9,$5F,$C0,$BF,$28,$48,$FF,$7D,$7C,$BB,$D3
	.dc.b	$C4,$7F,$9A,$6B,$83,$FE,$E3,$AF,$6D,$6E,$07,$FB,$D9,$FD,$1F,$C0
	.dc.b	$9A,$70,$ED,$B3,$C3,$A3,$FD,$E5,$38,$8F,$73,$FC,$19,$D7,$47,$6E
	.dc.b	$47,$E4,$CF,$AE,$FC,$1B,$6B,$62,$77,$3E,$8F,$A0,$93,$FD,$43,$04
	.dc.b	$7F,$52,$8F,$43,$63,$31,$B8,$C1,$1C,$1F,$3B,$37,$CE,$7C,$22,$47
	.dc.b	$90,$7D,$B7,$E5,$74,$FD,$76,$EB,$E3,$5E,$A8,$C5,$F2,$8F,$C1,$5F
	.dc.b	$1D,$7E,$8D,$3E,$39,$F6,$FF,$C5,$1B,$7C,$DA,$F2,$E3,$D4,$C8,$D3
	.dc.b	$F9,$AC,$3F,$C4,$F1,$F3,$33,$7E,$1B,$FD,$96,$3F,$D5,$D1,$8F,$47
	.dc.b	$FA,$4B,$63,$E0,$57,$B5,$BA,$FC,$F8,$C4,$7F,$81,$5A,$FF,$9E,$CF
	.dc.b	$8F,$1B,$23,$FE,$08,$8D,$BF,$E8,$16,$AC,$FF,$C5,$3F,$41,$FC,$DB
	.dc.b	$CB,$34,$FD,$BF,$DB,$FF,$9A,$EB,$E8,$4E,$9B,$3B,$7E,$D4,$ED,$FD
	.dc.b	$36,$1F,$EB,$37,$EC,$35,$CF,$BF,$DB,$E4,$FF,$4A,$C8,$DB,$97,$F0
	.dc.b	$1F,$71,$F8,$27,$F6,$4D,$D3,$F0,$ED,$E2,$FF,$50,$7D,$B7,$F9,$7E
	.dc.b	$BF,$1C,$D7,$F8,$43,$6B,$F9,$E4,$23,$13,$FE,$A9,$F6,$7F,$B7,$CF
	.dc.b	$EF,$27,$2D,$F7,$7F,$8E,$C6,$23,$FC,$B1,$E8,$7F,$00,$FA,$F9,$1F
	.dc.b	$09,$DF,$CF,$E2,$3E,$3C,$5C,$BC,$76,$AF,$ED,$7F,$6F,$FE,$14,$BF
	.dc.b	$E2,$4D,$AF,$D3,$B7,$3F,$83,$C3,$C3,$D5,$97,$FA,$9A,$D3,$EB,$9F
	.dc.b	$F0,$EF,$CE,$BF,$2D,$FF,$67,$CF,$F0,$88,$C7,$C0,$6A,$73,$3F,$CC
	.dc.b	$7D,$AF,$D0,$67,$F2,$9F,$D1,$47,$E5,$DA,$FF,$B4,$DA,$FF,$26,$6A
	.dc.b	$69,$3A,$BF,$89,$3A,$FD,$69,$F8,$0F,$F4,$C6,$D7,$F0,$B6,$BF,$D1
	.dc.b	$7F,$2C,$74,$7F,$52,$7F,$C6,$91,$FE,$71,$6B,$FA,$D6,$BF,$EE,$96
	.dc.b	$BF,$CF,$17,$CE,$F3,$24,$7F,$1B,$FC,$3D,$F6,$7F,$D8,$3F,$5D,$7C
	.dc.b	$5D,$25,$6F,$F0,$2F,$EC,$B6,$7F,$98,$36,$3F,$D0,$6D,$7F,$D1,$0D
	.dc.b	$7F,$82,$35,$FF,$80,$0D,$7F,$A9,$B5,$39,$F9,$EF,$F2,$7F,$C4,$FF
	.dc.b	$97,$9B,$3F,$D2,$1A,$FF,$6F,$FC,$B5,$F2,$FF,$DF,$1B,$1F,$E9,$86
	.dc.b	$BF,$E9,$D6,$BF,$C6,$5A,$FF,$66,$0E,$1F,$ED,$2D,$73,$1F,$D4,$E3
	.dc.b	$5E,$33,$EE,$DF,$37,$FE,$50,$6D,$FF,$47,$6B,$FE,$07,$F8,$67,$CA
	.dc.b	$FF,$85,$1B,$1F,$EB,$06,$BF,$EB,$56,$BF,$CC,$5A,$FA,$5A,$FF,$80
	.dc.b	$1A,$FF,$98,$23,$FB,$54,$6A,$FF,$99,$72,$FF,$99,$9F,$CB,$F4,$7F
	.dc.b	$60,$6F,$7F,$C5,$F9,$FF,$C4,$E3,$FD,$9D,$AF,$FB,$25,$AF,$F4,$56
	.dc.b	$BC,$78,$BF,$BD,$46,$23,$FC,$2A,$D4,$D3,$55,$FD,$1F,$E2,$BC,$C4
	.dc.b	$9F,$17,$C4,$7F,$8F,$2B,$FC,$E1,$1D,$23,$B9,$1F,$F6,$DC,$69,$EF
	.dc.b	$19,$7D,$4F,$FB,$82,$9C,$3F,$E9,$AC,$BF,$C0,$6D,$CB,$FB,$EB,$6F
	.dc.b	$D7,$EA,$F8,$CF,$79,$1F,$0B,$5F,$90,$D7,$E4,$3F,$EF,$0F,$80,$79
	.dc.b	$3F,$D6,$F0,$FF,$40,$46,$3E,$4F,$FD,$83,$39,$C7,$EB,$7A,$FF,$0C
	.dc.b	$F9,$E4,$C7,$E0,$BF,$E2,$BF,$69,$EE,$7F,$99,$1B,$3F,$E2,$36,$BF
	.dc.b	$E5,$D7,$FD,$4E,$D8,$E0,$47,$87,$1F,$FA,$C3,$E7,$BF,$B1,$47,$F7
	.dc.b	$F6,$EE,$73,$A3,$F5,$E7,$D9,$71,$F7,$26,$1F,$E2,$FB,$DD,$6A,$3F
	.dc.b	$C9,$76,$FB,$6F,$EF,$9C,$3F,$19,$BB,$FF,$98,$A7,$5E,$AF,$FB,$A3
	.dc.b	$F2,$BF,$E8,$F3,$9A,$DE,$0B,$DB,$DA,$33,$0A,$E3,$3B,$7F,$A5,$C6
	.dc.b	$7C,$AF,$B5,$18,$33,$5F,$E1,$9A,$63,$E7,$F3,$BF,$DF,$73,$98,$FB
	.dc.b	$FF,$71,$E0,$8C,$65,$E7,$AD,$9F,$C6,$6B,$79,$FF,$A7,$BE,$A1,$EF
	.dc.b	$F6,$CB,$FC,$70,$23,$35,$88,$C4,$ED,$FF,$1F,$4E,$9E,$2F,$F4,$CD
	.dc.b	$F8,$36,$33,$E4,$FD,$D6,$CF,$F9,$E5,$7C,$BD,$23,$FE,$7D,$7F,$DD
	.dc.b	$DA,$FF,$9D,$B4,$BC,$5F,$E2,$4A,$FB,$53,$97,$FC,$A5,$C2,$8C,$47
	.dc.b	$FB,$29,$AF,$09,$E4,$FD,$D8,$CB,$63,$2F,$FA,$A3,$E2,$2B,$FE,$CC
	.dc.b	$BF,$F4,$94,$60,$E7,$7E,$1F,$16,$E1,$FD,$0A,$BD,$71,$82,$38,$25
	.dc.b	$FA,$A7,$83,$EC,$FB,$87,$CB,$76,$6D,$7D,$AC,$11,$8A,$E1,$5B,$25
	.dc.b	$6E,$75,$3F,$EE,$E6,$BF,$D8,$7E,$1B,$F6,$33,$FE,$FD,$E8,$F8,$71
	.dc.b	$89,$FC,$4B,$5F,$F4,$13,$5E,$51,$C0,$93,$F8,$D9,$7D,$17,$42,$46
	.dc.b	$5F,$AB,$C6,$FB,$87,$D7,$61,$B5,$F9,$A9,$D4,$E4,$F3,$92,$3C,$0D
	.dc.b	$C7,$D2,$D7,$D2,$D7,$FD,$CF,$F2,$CF,$B8,$7C,$77,$E2,$23,$0D,$8C
	.dc.b	$DF,$EB,$27,$2F,$EA,$FD,$17,$CB,$7E,$31,$B1,$98,$C1,$1C,$13,$0F
	.dc.b	$DB,$CF,$07,$FD,$61,$0F,$B6,$FA,$10,$8C,$6F,$F9,$E6,$9F,$C0,$0B
	.dc.b	$1F,$C0,$BD,$EF,$F5,$7C,$7F,$01,$6C,$67,$BB,$F4,$71,$88,$FF,$60
	.dc.b	$35,$FF,$14,$35,$07,$0D,$8F,$25,$EE,$BD,$51,$82,$3F,$C4,$CF,$F2
	.dc.b	$C9,$7A,$FF,$6A,$E2,$BD,$91,$82,$46,$C9,$1F,$23,$A5,$F6,$9F,$15
	.dc.b	$B3,$F1,$5B,$3B,$9F,$E0,$7E,$67,$E4,$23,$F9,$4B,$5F,$F7,$13,$5F
	.dc.b	$F2,$53,$5F,$86,$D7,$FC,$17,$C3,$D0,$97,$F5,$72,$79,$9F,$18,$91
	.dc.b	$97,$F8,$AF,$DB,$7F,$78,$8F,$E7,$0D,$BF,$E7,$AD,$7F,$DE,$CD,$7F
	.dc.b	$CB,$CD,$7E,$BB,$53,$49,$F2,$FE,$02,$4F,$81,$F1,$E5,$23,$C6,$DA
	.dc.b	$FD,$27,$C2,$7F,$2C,$E3,$FF,$65,$57,$F4,$F6,$CE,$A3,$FC,$41,$F6
	.dc.b	$FF,$CC,$BB,$4D,$24,$7F,$57,$F8,$4F,$F4,$B3,$7D,$CF,$E7,$31,$8B
	.dc.b	$FE,$BE,$D7,$FC,$FC,$1C,$3F,$E8,$16,$BF,$9A,$6B,$B8,$FE,$93,$DA
	.dc.b	$FB,$67,$CF,$61,$B8,$7F,$26,$6D,$FF,$26,$6B,$F8,$63,$F0,$DC,$7F
	.dc.b	$A8,$E3,$FE,$AA,$DA,$FC,$7C,$E8,$93,$A6,$C6,$BE,$BB,$04,$70,$48
	.dc.b	$FC,$B8,$8D,$38,$7F,$B0,$FD,$9B,$D4,$A4,$AF,$F0,$34,$65,$B1,$9A
	.dc.b	$FF,$06,$CE,$5B,$1F,$CA,$7E,$CF,$F8,$D4,$7F,$AB,$F8,$51,$88,$FE
	.dc.b	$62,$D5,$1A,$EE,$F8,$BE,$57,$E1,$31,$FD,$9B,$6B,$F6,$6D,$7F,$66
	.dc.b	$D7,$F6,$6D,$7F,$CF,$AD,$D3,$FC,$A9,$C5,$D8,$EF,$F2,$D2,$FB,$B6
	.dc.b	$FB,$B5,$EE,$D7,$FD,$99,$A3,$EB,$3E,$69,$2B,$E5,$B6,$7E,$4B,$5F
	.dc.b	$CB,$CF,$19,$FE,$C1,$F6,$7F,$F0,$01,$BF,$0D,$FD,$A2,$31,$5F,$9E
	.dc.b	$F1,$91,$88,$F8,$7F,$15,$F2,$89,$7F,$29,$B1,$F2,$9A,$F0,$8F,$9F
	.dc.b	$F0,$5F,$48,$8E,$09,$1F,$4A,$32,$D8,$FF,$27,$7D,$A7,$FB,$81,$B3
	.dc.b	$FE,$30,$8C,$4F,$F9,$59,$AF,$EE,$1A,$FF,$BD,$DA,$FF,$50,$6A,$73
	.dc.b	$C2,$FB,$37,$C9,$7F,$7B,$6E,$1F,$DE,$5B,$7F,$C1,$AB,$85,$7E,$43
	.dc.b	$EC,$E5,$FE,$41,$B5,$F9,$06,$BF,$90,$6B,$F9,$0E,$4F,$CC,$92,$7F
	.dc.b	$32,$D8,$FC,$CB,$5F,$CC,$BF,$E5,$4F,$B2,$FE,$6C,$4F,$7F,$07,$F9
	.dc.b	$4B,$EC,$FF,$C9,$35,$FD,$C8,$FF,$4D,$38,$7F,$D4,$4D,$73,$EA,$F5
	.dc.b	$F8,$B9,$8F,$FA,$0F,$04,$62,$BF,$D1,$31,$89,$D4,$E6,$7F,$96,$CE
	.dc.b	$5D,$FD,$7F,$A4,$95,$D5,$FF,$34,$F3,$3E,$42,$92,$7F,$CB,$D1,$97
	.dc.b	$FD,$2D,$EC,$FF,$88,$2C,$D2,$7A,$3F,$E7,$4F,$B2,$7B,$D4,$97,$FE
	.dc.b	$69,$8C,$B6,$33,$7F,$E8,$C9,$CB,$63,$2F,$CD,$F2,$53,$A8,$C3,$5F
	.dc.b	$F6,$83,$5F,$F4,$67,$95,$FE,$8E,$C7,$FB,$0C,$7F,$54,$6B,$F6,$5A
	.dc.b	$FF,$B4,$FE,$89,$F2,$BF,$E9,$06,$E5,$FE,$83,$6C,$FF,$A9,$1A,$F5
	.dc.b	$FB,$5F,$01,$27,$E5,$B6,$3F,$D8,$B6,$8C,$47,$19,$FF,$48,$7D,$9B
	.dc.b	$83,$FD,$42,$D8,$CC,$60,$F7,$7C,$7C,$8F,$B2,$DA,$EE,$D7,$E1,$35
	.dc.b	$F8,$4F,$F9,$75,$FF,$73,$C7,$FB,$EF,$EB,$5F,$1B,$9F,$83,$FD,$ED
	.dc.b	$7F,$AE,$6F,$87,$87,$C7,$85,$87,$0C,$7F,$CA,$F8,$70,$C3,$F1,$78
	.dc.b	$70,$9E,$D8,$F5,$ED,$F1,$C7,$F9,$7B,$2E,$98,$F3,$C3,$FD,$BB,$5F
	.dc.b	$67,$E3,$7D,$17,$F7,$5B,$39,$8F,$F3,$39,$92,$B9,$3F,$DC,$1B,$A7
	.dc.b	$8C,$7F,$B5,$5A,$FE,$89,$B1,$B9,$F6,$D7,$F9,$D3,$47,$D7,$3E,$1F
	.dc.b	$D9,$8F,$3B,$DA,$AD,$CE,$A1,$F4,$2F,$9E,$FE,$EB,$1F,$00,$FF,$40
	.dc.b	$C6,$7C,$3E,$7F,$62,$FB,$2F,$CA,$F1,$FC,$FA,$D5,$F1,$3A,$E1,$D2
	.dc.b	$74,$DD,$7F,$D1,$B8,$73,$E8,$F2,$61,$CA,$F7,$97,$F8,$FB,$1F,$1E
	.dc.b	$3E,$67,$FC,$83,$5E,$06,$C6,$5F,$E5,$10,$FB,$27,$C2,$73,$9F,$83
	.dc.b	$18,$7B,$4E,$E7,$4D,$8C,$CE,$F0,$F8,$0D,$CF,$BB,$65,$9B,$EC,$9F
	.dc.b	$21,$FC,$EF,$0E,$39,$7D,$6B,$F1,$B7,$2F,$04,$FE,$8B,$0E,$57,$C7
	.dc.b	$2F,$6F,$9B,$71,$FB,$DC,$6F,$8C,$C7,$1F,$E1,$9D,$4F,$E0,$2A,$E6
	.dc.b	$7F,$A7,$72,$7C,$C6,$CF,$63,$F3,$67,$E1,$1C,$B4,$C9,$D4,$B3,$D5
	.dc.b	$B1,$C1,$FC,$8F,$CB,$FF,$5D,$E9,$FD,$8F,$D0,$7A,$3A,$8F,$9A,$75
	.dc.b	$8F,$DC,$6D,$FE,$76,$B6,$FE,$83,$3F,$E0,$B7,$C6,$3F,$7D,$7C,$5F
	.dc.b	$BB,$7F,$83,$7C,$66,$08,$7E,$11,$D6,$46,$9F,$EA,$1F,$B0,$FF,$44
	.dc.b	$CF,$E5,$F9,$91,$88,$FF,$09,$35,$FF,$34,$35,$FC,$BF,$CD,$76,$23
	.dc.b	$F9,$26,$BF,$96,$6B,$FA,$27,$F3,$99,$3C,$E7,$CC,$E1,$B5,$FA,$B9
	.dc.b	$D4,$E4,$F6,$FA,$B2,$7F,$42,$D7,$D3,$F6,$1E,$F2,$3E,$FC,$51,$88
	.dc.b	$D9,$97,$0F,$C0,$6A,$3F,$D0,$B1,$C9,$E0,$7E,$DF,$EB,$5F,$35,$F5
	.dc.b	$DB,$19,$8C,$1F,$0B,$EB,$DF,$13,$49,$3F,$06,$32,$F5,$FB,$4E,$84
	.dc.b	$70,$78,$7E,$4F,$C0,$47,$04,$3F,$8B,$60,$8D,$46,$1F,$C3,$78,$92
	.dc.b	$3E,$43,$5F,$E3,$DF,$67,$FE,$22,$7F,$8B,$35,$CF,$8D,$F9,$78,$C4
	.dc.b	$7F,$B8,$BE,$39,$E7,$7E,$6D,$B1,$FC,$7D,$AF,$F2,$B6,$BC,$63,$81
	.dc.b	$1E,$0F,$3F,$B0,$91,$CD,$AF,$36,$BE,$36,$BF,$E5,$99,$FF,$69,$3E
	.dc.b	$3F,$86,$F3,$11,$FE,$69,$F9,$7B,$DA,$FD,$A3,$63,$31,$82,$38,$21
	.dc.b	$FD,$1E,$38,$3F,$D2,$7E,$BB,$EF,$B7,$87,$F8,$04,$62,$3F,$CD,$C1
	.dc.b	$C3,$FE,$71,$6B,$F9,$76,$BB,$8E,$67,$4D,$64,$7F,$4F,$9F,$EE,$CD
	.dc.b	$8C,$BE,$56,$BE,$28,$DB,$E9,$FA,$DF,$39,$1E,$8F,$F1,$9F,$C0,$FF
	.dc.b	$90,$C7,$F6,$76,$BE,$C6,$BF,$E8,$B6,$BF,$B1,$6A,$69,$3C,$0F,$E4
	.dc.b	$DF,$5F,$FE,$36,$9F,$ED,$4D,$7F,$BB,$35,$F9,$6D,$7F,$D3,$4D,$7F
	.dc.b	$80,$35,$34,$9E,$E7,$80,$91,$E0,$6B,$E0,$6B,$E0,$7F,$9E,$7D,$6F
	.dc.b	$F8,$BE,$3F,$C0,$0D,$7F,$C1,$8D,$7F,$02,$D7,$FD,$61,$E8,$7F,$92
	.dc.b	$DF,$C2,$B5,$CF,$95,$FD,$26,$31,$1F,$C6,$DA,$FF,$8B,$63,$44,$7C
	.dc.b	$7E,$CF,$98,$8E,$08,$79,$9C,$B4,$F5,$FC,$CF,$B8,$8E,$08,$7B,$9C
	.dc.b	$B4,$FF,$11,$79,$5F,$E1,$46,$F9,$BF,$D8,$E3,$11,$FE,$31,$6B,$F0
	.dc.b	$B8,$7E,$01,$1F,$F1,$66,$3F,$A4,$24,$66,$3F,$C7,$51,$FA,$8F,$AB
	.dc.b	$FF,$63,$47,$F8,$CD,$AF,$F8,$FD,$AF,$ED,$1A,$FF,$B8,$D7,$85,$FD
	.dc.b	$E5,$B1,$CE,$30,$7F,$7D,$FA,$A7,$CD,$D2,$4F,$F9,$0E,$32,$DF,$96
	.dc.b	$E0,$47,$07,$F9,$27,$F5,$FF,$E6,$2F,$FB,$D5,$AF,$F5,$B6,$A6,$93
	.dc.b	$BF,$FC,$20,$D8,$FF,$2A,$7D,$5F,$FB,$D9,$BF,$0D,$FE,$15,$8C,$47
	.dc.b	$F9,$8D,$AF,$F1,$8F,$88,$7E,$15,$FE,$1B,$6C,$7F,$97,$DA,$FF,$9A
	.dc.b	$DA,$F1,$8E,$04,$78,$3F,$E2,$5F,$21,$F1,$B4,$93,$FE,$6B,$8C,$BF
	.dc.b	$E6,$DF,$D1,$1F,$39,$FE,$2D,$6C,$66,$30,$47,$04,$3F,$D0,$11,$C1
	.dc.b	$FF,$41,$7D,$63,$E1,$3F,$C7,$0D,$8F,$F3,$E3,$5D,$C7,$F8,$F7,$E0
	.dc.b	$7F,$C5,$D1,$FE,$0F,$6B,$FE,$F2,$F9,$BD,$11,$FF,$43,$46,$89,$1E
	.dc.b	$0F,$4A,$4F,$81,$AF,$FA,$5F,$F4,$47,$D2,$FF,$93,$5B,$1F,$E9,$16
	.dc.b	$BB,$8D,$11,$D3,$5F,$83,$E5,$FC,$02,$3F,$01,$FF,$2D,$7E,$63,$F7
	.dc.b	$23,$FC,$58,$1C,$3F,$E2,$F6,$B9,$E0,$E2,$47,$FD,$4D,$1A,$37,$A7
	.dc.b	$F7,$A8,$FA,$7E,$67,$D2,$23,$C3,$D7,$F1,$91,$C1,$FE,$B1,$FA,$EF
	.dc.b	$F1,$CB,$FE,$BD,$6B,$FD,$C9,$AF,$A9,$AF,$F9,$1D,$A9,$A4,$E2,$FF
	.dc.b	$31,$FD,$6B,$E4,$29,$2B,$FC,$C5,$19,$6C,$65,$FF,$01,$FB,$49,$D4
	.dc.b	$61,$AF,$CE,$87,$D6,$FF,$26,$8F,$F7,$B3,$5F,$F7,$93,$63,$7E,$2F
	.dc.b	$F9,$E2,$30,$47,$07,$FA,$57,$EB,$1D,$54,$97,$FE,$DB,$8C,$BF,$E7
	.dc.b	$DF,$AC,$7C,$9F,$FA,$11,$B1,$98,$C1,$1C,$10,$F8,$31,$C1,$FF,$79
	.dc.b	$4B,$D3,$7C,$2F,$85,$B3,$A6,$BF,$E2,$C9,$E1,$3E,$7F,$15,$F3,$F4
	.dc.b	$95,$DA,$32,$FF,$AD,$7E,$2B,$FD,$6C,$9A,$47,$CF,$F5,$0F,$87,$F8
	.dc.b	$4D,$8C,$C6,$08,$E0,$87,$DB,$CD,$3C,$63,$FC,$8A,$3C,$7C,$3F,$E4
	.dc.b	$86,$BF,$B0,$6B,$FE,$9D,$73,$C2,$AF,$F2,$F5,$FD,$7F,$6B,$31,$8B
	.dc.b	$FA,$31,$82,$4E,$C9,$1F,$96,$FA,$CF,$F8,$01,$1F,$EB,$D6,$C6,$7E
	.dc.b	$2B,$C3,$18,$8F,$CC,$35,$FF,$59,$60,$8C,$42,$31,$3C,$8F,$6F,$D3
	.dc.b	$7E,$B6,$FE,$2D,$23,$0F,$C8,$6B,$FE,$6E,$6B,$FC,$59,$AF,$FB,$0D
	.dc.b	$AF,$3E,$7E,$E7,$83,$C0,$95,$E0,$AF,$E5,$ED,$8F,$A7,$C2,$FA,$C7
	.dc.b	$D3,$61,$B5,$F7,$27,$53,$93,$DA,$48,$EC,$DC,$7D,$8D,$7D,$8D,$7F
	.dc.b	$DB,$9F,$14,$FA,$CF,$F6,$25,$7E,$2A,$31,$F0,$BF,$88,$6C,$66,$77
	.dc.b	$5F,$90,$F9,$1F,$F7,$E4,$FC,$E6,$C6,$7C,$1F,$C7,$C6,$23,$FD,$36
	.dc.b	$D7,$FA,$1B,$5F,$F7,$83,$53,$99,$7D,$5F,$D4,$6E,$FF,$9E,$69,$FA
	.dc.b	$42,$D7,$E9,$BE,$B1,$ED,$FE,$89,$B3,$F8,$66,$CE,$E7,$44,$9D,$36
	.dc.b	$3F,$D6,$B1,$E4,$8F,$15,$F2,$9F,$EB,$F3,$93,$C9,$19,$8C,$F5,$7E
	.dc.b	$D2,$7F,$C8,$67,$E6,$0B,$1E,$B8,$D1,$23,$4D,$8F,$1F,$4F,$F8,$7A
	.dc.b	$7E,$13,$63,$3E,$87,$C7,$8C,$47,$FB,$3B,$E9,$FF,$22,$FF,$0C,$6B
	.dc.b	$9E,$1F,$E1,$31,$88,$FE,$F6,$D4,$1C,$36,$38,$9F,$1B,$CB,$98,$7C
	.dc.b	$46,$CF,$F1,$CF,$AC,$FE,$9B,$3F,$BA,$6C,$67,$C3,$FE,$37,$18,$8F
	.dc.b	$F7,$3B,$5F,$F0,$7B,$5F,$2B,$53,$47,$B9,$D0,$95,$E1,$7B,$E7,$23
	.dc.b	$C8,$DA,$FD,$57,$BD,$FC,$BB,$94,$D2,$57,$F1,$4E,$C7,$D4,$7F,$34
	.dc.b	$6C,$66,$30,$7F,$38,$FA,$C7,$B7,$49,$3F,$C7,$A3,$2D,$8C,$E3,$FD
	.dc.b	$1A,$72,$D8,$FE,$8B,$DC,$FA,$C7,$AB,$86,$E1,$FC,$9D,$B7,$FC,$9D
	.dc.b	$AF,$E0,$8F,$C1,$71,$FE,$0B,$8F,$FA,$8B,$6B,$F1,$93,$A2,$4E,$9B
	.dc.b	$1A,$FA,$CC,$11,$C1,$23,$F3,$02,$34,$E1,$FE,$BF,$F5,$4F,$4A,$92
	.dc.b	$BF,$A2,$C6,$5B,$19,$AF,$ED,$B3,$96,$C7,$F2,$5F,$AC,$FE,$31,$C4
	.dc.b	$69,$3B,$FE,$0F,$97,$F7,$2F,$F5,$CD,$7F,$52,$D7,$F5,$3A,$7F,$84
	.dc.b	$B8,$B8,$93,$3F,$DA,$56,$EB,$F5,$CD,$AF,$F0,$CF,$D6,$78,$09,$5C
	.dc.b	$1B,$5E,$26,$BE,$27,$E3,$FD,$34,$9F,$8A,$D9,$F8,$6D,$7F,$CF,$63
	.dc.b	$54,$66,$3F,$81,$F8,$AF,$AC,$FF,$72,$DF,$F5,$D6,$CF,$29,$D7,$C2
	.dc.b	$7F,$66,$9C,$CF,$C6,$F8,$7F,$96,$4B,$F8,$6D,$8F,$9A,$D7,$84,$7C
	.dc.b	$8F,$81,$FA,$64,$70,$48,$FA,$B1,$96,$C7,$F8,$2B,$EB,$3F,$D9,$ED
	.dc.b	$9F,$F1,$5C,$62,$7F,$C6,$6D,$7E,$2B,$5F,$F5,$0B,$5F,$E4,$ED,$4E
	.dc.b	$78,$9F,$58,$F8,$AF,$EE,$8D,$C3,$FB,$C3,$6F,$E9,$D7,$0A,$FC,$8F
	.dc.b	$D6,$4B,$FC,$4B,$6B,$F2,$6D,$7F,$26,$D7,$F2,$7C,$5F,$DD,$78,$BF
	.dc.b	$CC,$BC,$9F,$9C,$25,$FE,$51,$B5,$F9,$E6,$BF,$86,$FA,$BF,$E6,$04
	.dc.b	$F1,$3F,$CA,$6F,$F3,$5F,$01,$F5,$9F,$E4,$FA,$FF,$4D,$35,$FF,$21
	.dc.b	$B5,$FF,$21,$35,$4B,$E2,$7A,$93,$1F,$F2,$1C,$65,$B1,$99,$FF,$35
	.dc.b	$CE,$5B,$1E,$6F,$A9,$95,$C9,$B3,$FE,$B4,$E7,$7C,$87,$F9,$B1,$B5
	.dc.b	$FE,$76,$CB,$F4,$B3,$93,$FC,$E7,$EC,$FF,$93,$09,$E8,$60,$FF,$3A
	.dc.b	$FD,$67,$FA,$A2,$F9,$37,$1F,$F6,$1B,$5F,$F6,$03,$5F,$F3,$4B,$5F
	.dc.b	$F3,$43,$70,$34,$9E,$F7,$80,$95,$E0,$7C,$FE,$4C,$8F,$F3,$6E,$3F
	.dc.b	$D7,$9B,$1F,$E8,$CF,$99,$7C,$55,$24,$FC,$38,$CB,$CB,$EB,$3E,$19
	.dc.b	$1C,$13,$3F,$F4,$FE,$09,$D4,$61,$FF,$4D,$FD,$63,$82,$92,$FE,$94
	.dc.b	$65,$F4,$FC,$6F,$D2,$23,$82,$5F,$82,$32,$D8,$CF,$85,$FE,$BB,$DD
	.dc.b	$F4,$AF,$AC,$FF,$5F,$7C,$1F,$EE,$70,$FF,$58,$63,$FE,$C1,$C3,$85
	.dc.b	$EF,$1F,$F1,$4D,$EE,$FF,$D7,$97,$B9,$FF,$51,$D6,$E7,$FD,$A1,$DD
	.dc.b	$FE,$BD,$AD,$47,$FA,$96,$75,$7C,$F6,$7D,$67,$FB,$47,$1F,$F6,$33
	.dc.b	$6F,$FD,$8C,$D7,$FD,$8C,$D7,$FD,$9C,$D9,$E7,$3F,$E4,$03,$99,$67
	.dc.b	$9B,$5E,$6F,$BB,$1F,$F5,$EF,$D4,$7F,$BE,$63,$13,$A3,$DA,$7F,$AF
	.dc.b	$E3,$4E,$07,$D4,$3E,$61,$E1,$F2,$9C,$C6,$1E,$7F,$52,$FA,$CF,$C8
	.dc.b	$C7,$3B,$E5,$7C,$5B,$19,$BE,$5A,$FD,$BC,$39,$73,$FD,$ED,$3F,$CC
	.dc.b	$37,$E0,$C7,$96,$1E,$18,$FE,$CF,$8F,$46,$C6,$5E,$31,$96,$FD,$67
	.dc.b	$FA,$89,$FF,$7F,$C2,$31,$1D,$1A,$F0,$F8,$2F,$F7,$0C,$ED,$F6,$4F
	.dc.b	$3C,$FF,$1D,$39,$AF,$8B,$F5,$8F,$8E,$F5,$D6,$AF,$86,$7D,$6F,$E4
	.dc.b	$46,$16,$AC,$39,$47,$CB,$F3,$79,$DF,$E0,$3E,$26,$CF,$DD,$6B,$F7
	.dc.b	$26,$E6,$7F,$D5,$71,$F1,$9A,$FF,$A4,$23,$4F,$E3,$CE,$45,$72,$B8
	.dc.b	$BE,$9C,$F2,$8D,$99,$3B,$1C,$7E,$B3,$EE,$F4,$7E,$E3,$E8,$3D,$3D
	.dc.b	$7F,$C0,$1F,$3C,$EB,$0F,$B6,$DB,$FD,$A4,$F0,$E9,$3C,$23,$FC,$4F
	.dc.b	$5C,$1F,$B3,$5F,$E2,$E7,$07,$E0,$9C,$3F,$05,$AF,$E9,$1B,$91,$A7
	.dc.b	$0E,$47,$D5,$7F,$A0,$2B,$B4,$63,$DC,$D4,$E6,$7F,$C1,$AD,$7F,$37
	.dc.b	$F5,$9D,$88,$FE,$41,$AF,$E6,$1A,$F0,$3F,$B2,$79,$AF,$9B,$FF,$52
	.dc.b	$36,$33,$18,$3D,$7E,$B4,$8F,$CE,$B5,$F2,$FD,$67,$A0,$8F,$B3,$04
	.dc.b	$62,$36,$65,$C3,$FB,$AC,$BE,$2C,$62,$3F,$DB,$B1,$87,$81,$F4,$BE
	.dc.b	$A9,$F3,$1F,$49,$B1,$98,$C1,$FE,$C1,$FA,$CF,$F6,$7C,$7F,$81,$5A
	.dc.b	$FF,$9B,$38,$D1,$88,$FE,$1B,$18,$F8,$7E,$DF,$59,$27,$C4,$D7,$F1
	.dc.b	$18,$3E,$47,$C8,$48,$F4,$35,$DB,$5F,$F3,$EC,$71,$7E,$57,$CC,$7E
	.dc.b	$10,$8E,$0F,$C9,$FD,$63,$C7,$FC,$9B,$63,$31,$83,$F2,$FF,$1A,$F3
	.dc.b	$E9,$27,$F3,$B1,$96,$C6,$5F,$D2,$46,$5B,$1D,$3E,$B2,$47,$26,$BC
	.dc.b	$1A,$FF,$A6,$A5,$18,$8F,$E7,$71,$FE,$C9,$7C,$7F,$0D,$E8,$23,$FB
	.dc.b	$2F,$AC,$7C,$CF,$EC,$9B,$19,$8C,$11,$C1,$0F,$DF,$47,$07,$F8,$07
	.dc.b	$D6,$7E,$05,$FE,$9A,$D4,$1C,$37,$83,$F8,$1C,$60,$8E,$0C,$EB,$27
	.dc.b	$F8,$5C,$A3,$11,$E2,$6B,$FE,$C0,$8D,$3C,$23,$BF,$D6,$48,$E6,$D7
	.dc.b	$F8,$D7,$D6,$7F,$20,$7F,$8C,$B5,$CF,$83,$FC,$6A,$31,$1F,$EC,$E6
	.dc.b	$BF,$DB,$9A,$FC,$26,$A6,$FD,$67,$F8,$CE,$3F,$93,$B5,$FE,$58,$D7
	.dc.b	$FD,$B6,$D7,$FB,$DB,$5F,$9E,$D4,$D2,$7B,$9F,$CC,$9F,$F4,$1C,$78
	.dc.b	$23,$04,$70,$43,$C6,$FF,$3A,$FA,$CF,$F1,$BC,$7F,$3E,$6B,$FD,$0D
	.dc.b	$AF,$FB,$C9,$AF,$F8,$4B,$D0,$FF,$25,$3F,$EF,$66,$B9,$F2,$BF,$A3
	.dc.b	$C6,$23,$F1,$6D,$7F,$C5,$91,$A2,$3E,$3F,$6B,$CC,$47,$04,$3C,$CE
	.dc.b	$5A,$7A,$FE,$67,$DC,$47,$04,$3D,$CE,$5A,$7F,$60,$F2,$FF,$C1,$AF
	.dc.b	$F6,$56,$BF,$E3,$23,$FC,$11,$E8,$FF,$64,$8C,$C7,$C5,$E1,$F8,$24
	.dc.b	$7F,$B5,$E3,$FA,$A2,$46,$63,$FB,$94,$7E,$BB,$EB,$3E,$EC,$7F,$75
	.dc.b	$6B,$E1,$6B,$FE,$44,$6B,$FB,$85,$E2,$7F,$78,$6C,$7F,$7C,$6B,$D6
	.dc.b	$3F,$BE,$FD,$63,$E6,$A9,$27,$FC,$07,$19,$6F,$CB,$70,$23,$83,$FC
	.dc.b	$13,$F5,$9F,$CD,$5F,$F0,$73,$5F,$94,$D7,$FC,$BA,$D7,$F8,$8B,$53
	.dc.b	$49,$DD,$FE,$14,$FA,$CF,$F7,$A4,$FF,$86,$1A,$FF,$87,$5A,$FD,$96
	.dc.b	$BF,$E6,$CF,$89,$7D,$57,$F8,$79,$B1,$FE,$22,$6B,$FE,$27,$6B,$F8
	.dc.b	$36,$BC,$E3,$91,$1F,$F1,$5F,$90,$F9,$0F,$F1,$5B,$63,$31,$83,$FC
	.dc.b	$5F,$F5,$8F,$9C,$A4,$9F,$F1,$AC,$65,$B1,$97,$FC,$79,$19,$6C,$7F
	.dc.b	$8E,$BE,$A9,$F0,$74,$93,$FE,$40,$8C,$BF,$E4,$2F,$80,$FF,$1A,$3F
	.dc.b	$E9,$26,$BF,$D1,$9A,$FF,$BD,$3D,$5F,$F2,$3C,$60,$8E,$0F,$17,$A7
	.dc.b	$27,$C0,$D7,$FC,$A9,$F5,$8F,$A5,$FF,$28,$B6,$3F,$CA,$AD,$77,$1A
	.dc.b	$23,$A6,$BF,$E5,$FC,$1E,$57,$C3,$24,$7C,$17,$FC,$BF,$F5,$9F,$72
	.dc.b	$3F,$B1,$B5,$34,$9C,$1C,$09,$1F,$E6,$A8,$E1,$5F,$DF,$E3,$C5,$E5
	.dc.b	$FD,$32,$3B,$F5,$FC,$24,$70,$7F,$9D,$3E,$B3,$FC,$82,$FF,$9E,$5A
	.dc.b	$FF,$15,$6B,$FE,$C8,$6B,$FD,$F1,$A9,$A4,$E3,$FF,$40,$FA,$8F,$AC
	.dc.b	$7C,$7E,$1B,$7F,$E8,$99,$D4,$E4,$91,$99,$DB,$FE,$51,$6B,$FE,$50
	.dc.b	$6B,$FE,$10,$97,$D6,$7F,$2C,$84,$63,$C4,$F2,$CE,$63,$13,$E3,$8C
	.dc.b	$12,$76,$7F,$A6,$BE,$B1,$D5,$49,$5F,$EA,$08,$CB,$ED,$FA,$C7,$C8
	.dc.b	$FC,$46,$C6,$63,$04,$70,$43,$FC,$E1,$1C,$1F,$F5,$84,$FF,$88,$FE
	.dc.b	$13,$C1,$E9,$1A,$49,$DF,$88,$F9,$8F,$33,$63,$31,$83,$FD,$6E,$BE
	.dc.b	$2A,$91,$FF,$5D,$B7,$F0,$47,$DA,$7F,$5D,$F8,$1F,$F6,$B4,$62,$BF
	.dc.b	$B0,$57,$F9,$12,$31,$1F,$E9,$8F,$87,$C5,$FF,$B2,$EA,$CF,$FA,$7E
	.dc.b	$6B,$FF,$00,$05,$72,$57,$E0,$BF,$C1,$7E,$BD,$E4,$BF,$C7,$36,$BF
	.dc.b	$D3,$4D,$7F,$D3,$2D,$7F,$15,$F8,$13,$F1,$8F,$F7,$86,$BF,$EF,$1E
	.dc.b	$54,$E7,$0F,$F7,$2D,$FE,$3B,$0E,$D7,$BA,$D4,$FD,$5A,$EB,$87,$46
	.dc.b	$9C,$B0,$61,$CA,$33,$5C,$47,$D9,$BE,$A9,$F1,$DC,$E7,$31,$8A,$F2
	.dc.b	$D6,$FD,$ED,$3C,$6E,$D6,$EF,$B4,$E6,$F8,$CA,$F7,$7C,$BE,$BD,$F8
	.dc.b	$6D,$31,$E9,$5C,$3C,$33,$A6,$CE,$9A,$79,$87,$BA,$7F,$C5,$19,$FC
	.dc.b	$07,$26,$33,$D4,$EE,$7F,$C5,$D1,$89,$D7,$27,$CB,$9F,$92,$D9,$D4
	.dc.b	$7C,$08,$C7,$D7,$3D,$AE,$93,$FE,$9C,$3E,$8C,$66,$FC,$4D,$8F,$0B
	.dc.b	$5E,$16,$8C,$47,$FA,$9A,$31,$5D,$A3,$13,$A7,$BC,$65,$FC,$0F,$2B
	.dc.b	$E2,$BE,$DB,$70,$FB,$F7,$E3,$E5,$F8,$98,$FE,$9A,$FF,$03,$C7,$FE
	.dc.b	$C4,$BF,$94,$DC,$7E,$4B,$5F,$4B,$5F,$43,$5F,$C9,$B6,$BF,$97,$FC
	.dc.b	$2F,$9D,$B5,$F5,$F8,$7F,$D8,$AF,$E5,$1B,$7F,$90,$BF,$16,$3E,$89
	.dc.b	$CB,$EA,$FC,$1D,$F5,$5D,$5B,$87,$E0,$DB,$87,$E0,$9A,$FD,$26,$BF
	.dc.b	$45,$AF,$EB,$1B,$89,$BD,$4F,$47,$F5,$2D,$BF,$D6,$CF,$83,$1F,$9F
	.dc.b	$39,$C3,$F5,$7F,$6D,$FC,$35,$FD,$CB,$71,$34,$9D,$9F,$25,$B8,$FE
	.dc.b	$7D,$E1,$9B,$F1,$3F,$DD,$B6,$7F,$7F,$3F,$9D,$8C,$61,$FB,$CF,$C2
	.dc.b	$7E,$7B,$F9,$46,$BF,$93,$6B,$FC,$39,$B8,$9A,$48,$FB,$CE,$3E,$33
	.dc.b	$F8,$6A,$6C,$61,$FE,$23,$F8,$53,$EF,$FF,$0C,$DC,$BF,$78,$DA,$FD
	.dc.b	$DB,$5F,$D5,$B5,$FD,$53,$5F,$E4,$AD,$C0,$DF,$33,$EA,$6F,$FA,$46
	.dc.b	$DF,$2C,$3F,$92,$C6,$30,$FE,$5F,$F8,$63,$DF,$FE,$52,$DB,$FE,$5A
	.dc.b	$BE,$1B,$33,$AF,$A0,$FD,$14,$9A,$4A,$FE,$35,$F8,$97,$EE,$70,$FE
	.dc.b	$62,$BF,$10,$FE,$6E,$E7,$4F,$DD,$B9,$AF,$EA,$3F,$58,$BF,$C7,$37
	.dc.b	$97,$F8,$43,$67,$F9,$5B,$67,$FD,$A8,$D7,$F8,$F3,$5E,$8D,$4D,$E0
	.dc.b	$FE,$89,$C3,$FD,$33,$C5,$FC,$05,$F9,$2F,$F9,$74,$61,$FE,$9B,$F5
	.dc.b	$EF,$94,$FE,$A6,$BE,$02,$C6,$1B,$7F,$D6,$A7,$4D,$9D,$3F,$D7,$BF
	.dc.b	$01,$FE,$0E,$FF,$2D,$6D,$FF,$68,$7F,$C0,$F1,$B8,$FE,$C7,$1A,$C8
	.dc.b	$D3,$71,$FD,$91,$E6,$BE,$E1,$F3,$3F,$DC,$7C,$57,$DA,$BE,$C7,$FB
	.dc.b	$A5,$EE,$B5,$39,$CB,$FC,$13,$39,$7F,$C1,$77,$CA,$F8,$E1,$DA,$73
	.dc.b	$7C,$6F,$FB,$BF,$D8,$7E,$9F,$93,$F9,$FB,$6B,$F0,$98,$F3,$7F,$C0
	.dc.b	$71,$9A,$FF,$05,$FD,$17,$C1,$3D,$F7,$FD,$21,$C5,$FF,$82,$1B,$7F
	.dc.b	$E0,$C6,$D7,$F8,$23,$E7,$1F,$2D,$FE,$0E,$7F,$55,$B9,$CD,$FF,$85
	.dc.b	$1C,$98,$AE,$5E,$37,$1B,$E4,$FF,$87,$79,$BF,$CB,$5F,$6B,$F9,$CF
	.dc.b	$5C,$D2,$63,$F9,$E8,$CB,$63,$38,$7E,$82,$72,$FF,$82,$31,$FF,$0F
	.dc.b	$57,$97,$9F,$C1,$5A,$DB,$FB,$5D,$6A,$FE,$07,$D4,$7B,$CF,$81,$E5
	.dc.b	$5C,$7D,$D1,$67,$E3,$4D,$BF,$F2,$67,$E3,$9F,$C1,$6F,$FC,$6F,$F4
	.dc.b	$1F,$CE,$27,$3E,$07,$F8,$E2,$31,$3F,$E9,$FA,$E1,$87,$F8,$E2,$72
	.dc.b	$FF,$A8,$23,$37,$FE,$02,$9C,$ED,$FE,$83,$FA,$5F,$94,$4C,$3E,$43
	.dc.b	$70,$F8,$CD,$7F,$82,$EA,$8F,$EF,$D6,$F8,$7F,$31,$F2,$22,$DF,$B3
	.dc.b	$A7,$D5,$F2,$FE,$B6,$C6,$4F,$83,$19,$69,$F1,$3E,$97,$E1,$7B,$BF
	.dc.b	$0D,$B1,$93,$E4,$C6,$5A,$7C,$CF,$8D,$F3,$91,$FF,$25,$64,$8F,$5E
	.dc.b	$3F,$3C,$FA,$0D,$9F,$AB,$1F,$4F,$EB,$31,$AF,$D1,$6E,$FF,$3D,$B5
	.dc.b	$FE,$3F,$C1,$1F,$6F,$E5,$BE,$49,$27,$C6,$FC,$CF,$A1,$FC,$01,$B8
	.dc.b	$FC,$2B,$F8,$3F,$C2,$3E,$89,$23,$D6,$FD,$4F,$6F,$F1,$06,$E3,$F1
	.dc.b	$AE,$1A,$7E,$42,$31,$97,$E2,$E6,$E5,$F8,$99,$B5,$EB,$F4,$3F,$02
	.dc.b	$47,$E5,$BF,$84,$F7,$7F,$30,$7E,$65,$B1,$F9,$EF,$95,$FC,$49,$1F
	.dc.b	$A6,$DC,$7F,$CF,$7A,$A3,$11,$FE,$6C,$8C,$4F,$DA,$FA,$BF,$D3,$1F
	.dc.b	$A7,$AB,$1F,$EA,$DF,$84,$FE,$A5,$7F,$EC,$0F,$37,$FC,$D9,$EE,$7F
	.dc.b	$AC,$1B,$3A,$9F,$F6,$5C,$60,$95,$FE,$CC,$7F,$43,$F3,$DF,$B5,$3F
	.dc.b	$D6,$57,$CE,$7F,$77,$19,$69,$E3,$9C,$57,$FA,$F6,$9F,$53,$FE,$5A
	.dc.b	$F8,$2F,$E5,$72,$9C,$FB,$7F,$E9,$F9,$CC,$70,$C3,$1E,$3F,$F8,$0E
	.dc.b	$37,$1F,$E0,$59,$DE,$1D,$6B,$4D,$C3,$B3,$FE,$B1,$AE,$13,$AC,$3A
	.dc.b	$D6,$9B,$87,$68,$61,$DA,$F3,$87,$59,$79,$2C,$FF,$DF,$31,$C3,$2F
	.dc.b	$F5,$74,$E7,$CA,$F0,$CB,$EA,$DF,$59,$F8,$08,$E0,$7F,$88,$6B,$83
	.dc.b	$83,$BE,$1D,$B3,$F9,$71,$8C,$3A,$E1,$D3,$83,$DB,$39,$39,$61,$D1
	.dc.b	$FF,$61,$19,$2B,$97,$D3,$87,$B1,$CF,$0B,$C7,$FF,$6D,$C7,$C2,$8C
	.dc.b	$3F,$ED,$EA,$46,$23,$8C,$F0,$9C,$BF,$12,$39,$36,$33,$1F,$13,$07
	.dc.b	$D6,$3E,$47,$FD,$B5,$A7,$C2,$BD,$B7,$1F,$1B,$71,$F8,$B7,$C7,$1F
	.dc.b	$A7,$C2,$FC,$95,$F6,$9D,$38,$7F,$21,$18,$AD,$B5,$FA,$F5,$C7,$E3
	.dc.b	$B8,$47,$E7,$67,$59,$7E,$71,$B7,$D3,$0D,$E1,$C6,$FA,$C6,$2B,$71
	.dc.b	$CE,$31,$87,$6F,$45,$D3,$FE,$08,$AF,$ED,$FE,$87,$FB,$F7,$BB,$FB
	.dc.b	$33,$7A,$3D,$7F,$15,$69,$87,$2A,$C3,$63,$99,$FE,$73,$B5,$F2,$8C
	.dc.b	$35,$FC,$5E,$9F,$2F,$0F,$C2,$B6,$3F,$1E,$7C,$58,$CE,$86,$92,$37
	.dc.b	$3C,$5F,$E5,$3D,$5E,$6C,$FF,$DD,$31,$FE,$0E,$8C,$61,$23,$FB,$1B
	.dc.b	$F9,$6C,$BE,$3B,$6B,$D7,$3A,$21,$FC,$DA,$34,$71,$E3,$FF,$50,$47
	.dc.b	$F9,$1F,$CE,$F8,$F1,$88,$69,$FE,$4F,$EC,$C9,$C4,$91,$F0,$B1,$FE
	.dc.b	$13,$1D,$7E,$D3,$99,$27,$FC,$A9,$F1,$35,$EB,$97,$FA,$96,$FF,$88
	.dc.b	$D6,$70,$FF,$2F,$5D,$AD,$62,$AC,$61,$FE,$66,$AF,$49,$ED,$23,$E5
	.dc.b	$94,$7C,$1F,$9F,$F7,$92,$7F,$CD,$1F,$73,$F0,$8F,$8C,$47,$93,$F2
	.dc.b	$7E,$D3,$E3,$FD,$D4,$9F,$F3,$84,$7C,$A3,$0F,$B8,$FF,$3D,$C6,$5D
	.dc.b	$90,$D6,$9F,$CC,$A7,$E9,$FC,$04,$9F,$53,$F6,$48,$FB,$7E,$AB,$EC
	.dc.b	$FC,$04,$8F,$F4,$14,$7D,$C3,$0F,$6D,$3F,$D7,$CE,$8F,$F4,$47,$D0
	.dc.b	$FE,$0B,$E8,$A4,$60,$87,$17,$2F,$38,$FC,$4F,$DE,$48,$FF,$49,$3F
	.dc.b	$B5,$5F,$5E,$FA,$6C,$36,$3F,$98,$B5,$FD,$6B,$5F,$E7,$AD,$7F,$DD
	.dc.b	$87,$48,$FE,$72,$D3,$9B,$88,$E1,$1F,$E7,$1F,$9D,$FC,$F9,$1F,$E0
	.dc.b	$26,$4B,$D4,$F7,$53,$B9,$F9,$4F,$F1,$19,$D3,$FA,$AF,$A3,$93,$FE
	.dc.b	$B3,$7F,$5B,$F1,$38,$8F,$A5,$17,$A7,$FD,$79,$3F,$B2,$F8,$E9,$5F
	.dc.b	$EB,$B7,$E5,$7D,$77,$F9,$5A,$3F,$DA,$AD,$7F,$89,$B5,$FE,$C6,$D7
	.dc.b	$E3,$B5,$07,$0D,$F6,$3E,$5C,$60,$E1,$93,$F1,$17,$D9,$FC,$15,$F3
	.dc.b	$F4,$C4,$4C,$3E,$5C,$E9,$B3,$97,$FD,$CF,$EC,$7F,$38,$F8,$97,$B7
	.dc.b	$4C,$44,$9F,$F7,$3C,$65,$FF,$74,$C7,$F4,$0F,$8D,$FA,$F1,$F7,$3B
	.dc.b	$51,$88,$FD,$7B,$78,$3F,$94,$C3,$EC,$1D,$1F,$99,$F5,$7F,$B3,$56
	.dc.b	$A7,$35,$F8,$83,$85,$6F,$CE,$91,$99,$DB,$FC,$0C,$D9,$C6,$FA,$F8
	.dc.b	$DF,$C8,$FC,$DF,$EC,$53,$FC,$11,$B8,$FF,$3E,$CF,$FD,$DB,$1F,$D7
	.dc.b	$8F,$F1,$C4,$FB,$DD,$D7,$37,$46,$EF,$FB,$2F,$C6,$BD,$EB,$B4,$C3
	.dc.b	$E4,$4E,$5B,$39,$7F,$8B,$C2,$79,$CF,$F7,$0F,$8B,$75,$52,$5F,$C7
	.dc.b	$9F,$F1,$56,$9F,$41,$FF,$16,$47,$F8,$E9,$AF,$EF,$9A,$FE,$61,$AF
	.dc.b	$A7,$D5,$FE,$03,$18,$23,$FD,$9F,$DB,$7B,$FF,$E0,$45,$EF,$D3,$11
	.dc.b	$FE,$08,$F9,$2F,$E4,$B8,$7F,$93,$38,$51,$88,$FE,$23,$18,$F2,$3F
	.dc.b	$BA,$4E,$67,$FA,$E7,$CA,$3E,$4F,$FC,$26,$D8,$FC,$A3,$5F,$E8,$2B
	.dc.b	$F0,$7F,$F0,$CF,$CB,$7F,$B8,$67,$F3,$4D,$7F,$A3,$B7,$7F,$C2,$3F
	.dc.b	$E5,$68,$DF,$BD,$FE,$07,$8C,$47,$AD,$AF,$A9,$AF,$F7,$36,$BF,$E2
	.dc.b	$C6,$E6,$69,$8F,$97,$FB,$F3,$FA,$F6,$BF,$D7,$1B,$7F,$99,$F7,$FF
	.dc.b	$81,$4F,$F6,$06,$BF,$31,$AF,$F8,$15,$AF,$F8,$F9,$B7,$FE,$6A,$F0
	.dc.b	$3F,$C4,$91,$82,$3F,$E2,$7F,$98,$7A,$BF,$E4,$46,$DF,$F0,$56,$BF
	.dc.b	$DC,$BE,$69,$F3,$FF,$E4,$A6,$C7,$F8,$7D,$AF,$E1,$5A,$FE,$11,$AF
	.dc.b	$F8,$75,$AF,$F9,$59,$B8,$9A,$63,$E7,$3F,$61,$3F,$E0,$46,$E7,$FC
	.dc.b	$05,$BD,$FF,$E2,$9F,$94,$FC,$BC,$FF,$05,$7F,$25,$5F,$E6,$46,$E0
	.dc.b	$69,$39,$7F,$C9,$8D,$9F,$F1,$E3,$C2,$5F,$17,$FE,$6C,$6C,$FF,$2D
	.dc.b	$6B,$FE,$1A,$F2,$BE,$FE,$3F,$E3,$D7,$F0,$71,$FA,$86,$B9,$ED,$FF
	.dc.b	$2E,$46,$23,$FC,$F6,$DA,$FF,$77,$4E,$89,$3F,$E5,$1F,$9E,$7C,$8F
	.dc.b	$FA,$01,$B1,$FD,$05,$AF,$F8,$BB,$E1,$FF,$C9,$4F,$F4,$56,$BF,$BF
	.dc.b	$6B,$FB,$E6,$BF,$E5,$76,$BF,$E9,$06,$DF,$FB,$BB,$DF,$FF,$3B,$46
	.dc.b	$0C,$FC,$AF,$D1,$AF,$EA,$8D,$7F,$C8,$2D,$DF,$F9,$F7,$E2,$1F,$94
	.dc.b	$6F,$2F,$FA,$06,$3F,$2F,$5F,$C4,$9A,$FF,$9A,$DA,$FF,$A9,$3E,$2F
	.dc.b	$F0,$ED,$F1,$3F,$D4,$B1,$8C,$FF,$DF,$5E,$36,$26,$57,$F6,$9A,$D1
	.dc.b	$FC,$FD,$FE,$D9,$F2,$CF,$5B,$FD,$5E,$D8,$CC,$63,$D7,$CB,$23,$27
	.dc.b	$28,$FF,$15,$7D,$17,$F5,$A8,$FF,$34,$72,$23,$1B,$7F,$A7,$A3,$1E
	.dc.b	$67,$F8,$C7,$E4,$5E,$CF,$28,$98,$7F,$8E,$B0,$E3,$D1,$FE,$20,$C1
	.dc.b	$87,$0A,$FF,$1F,$B5,$FF,$60,$B5,$FF,$27,$C7,$0C,$3F,$D8,$73,$99
	.dc.b	$FF,$29,$FD,$1F,$F8,$8B,$2F,$F1,$6F,$77,$C3,$94,$FF,$8A,$1C,$EB
	.dc.b	$C3,$D8,$FF,$2D,$D6,$EB,$95,$7F,$05,$AE,$BE,$17,$FA,$D2,$30,$7F
	.dc.b	$8E,$A7,$A3,$83,$BD,$F4,$AE,$38,$78,$27,$85,$F3,$7F,$CD,$67,$99
	.dc.b	$C3,$C0,$FF,$6F,$DF,$91,$DA,$F8,$D7,$0F,$FB,$92,$7D,$F1,$87,$8B
	.dc.b	$5E,$2D,$7F,$DD,$7D,$3F,$E7,$A4,$D2,$3F,$E7,$E7,$D4,$FF,$96,$BE
	.dc.b	$47,$FD,$C3,$A7,$CF,$8F,$F2,$EC,$E9,$4E,$B3,$EF,$1F,$1C,$CF,$06
	.dc.b	$DF,$09,$99,$E0,$B8,$3F,$13,$5F,$E4,$88,$FC,$66,$FF,$A3,$BF,$F7
	.dc.b	$FB,$63,$F1,$73,$B7,$97,$0B,$E5,$7F,$DF,$3C,$1F,$9D,$8C,$61,$F9
	.dc.b	$4B,$CB,$F9,$98,$CD,$FE,$99,$B5,$C6,$F7,$3E,$7C,$3C,$53,$9B,$FF
	.dc.b	$32,$7D,$1F,$F4,$36,$F8,$5C,$3A,$92,$F4,$F0,$EF,$1F,$F6,$BB,$70
	.dc.b	$FF,$2A,$37,$0F,$F7,$E3,$C5,$EA,$7F,$9C,$FE,$8D,$F2,$1D,$DB,$19
	.dc.b	$F2,$31,$32,$38,$3B,$8C,$61,$FE,$E8,$9F,$F1,$E9,$F9,$A5,$92,$33
	.dc.b	$1F,$E3,$B6,$BF,$E3,$A6,$BF,$CF,$77,$F8,$5C,$5F,$22,$7F,$D4,$F5
	.dc.b	$CF,$D2,$CD,$7C,$7F,$13,$FC,$D5,$5F,$D4,$FB,$FE,$F5,$A7,$38,$FF
	.dc.b	$92,$DA,$FF,$92,$9A,$F6,$6B,$EA,$6C,$73,$9E,$BC,$AF,$A3,$FB,$0D
	.dc.b	$C3,$FB,$04,$E2,$7E,$E7,$33,$E3,$7C,$AD,$AF,$91,$87,$E7,$67,$24
	.dc.b	$8C,$CE,$F2,$FF,$32,$35,$FE,$F9,$3F,$03,$93,$FD,$9B,$5F,$EB,$16
	.dc.b	$CF,$F8,$16,$7F,$1D,$18,$E3,$FF,$0E,$7C,$8F,$C8,$F3,$0D,$26,$1E
	.dc.b	$68,$CB,$63,$38,$7F,$9D,$E7,$2F,$F8,$D7,$E7,$7F,$39,$D6,$7F,$3D
	.dc.b	$3A,$C3,$FD,$59,$39,$C3,$FC,$89,$C3,$F6,$BC,$EF,$DD,$63,$FE,$80
	.dc.b	$6B,$F8,$46,$BF,$90,$6E,$C6,$92,$7F,$0B,$9B,$E8,$FF,$60,$DC,$3C
	.dc.b	$27,$FB,$65,$C6,$1F,$E6,$B3,$F1,$BE,$8F,$CC,$7F,$D2,$6D,$7F,$2E
	.dc.b	$D7,$F4,$6D,$C4,$D2,$47,$4F,$91,$FE,$05,$06,$C6,$1F,$E0,$DE,$9F
	.dc.b	$FA,$AB,$0F,$C4,$B6,$7F,$D1,$DE,$CF,$E8,$A3,$11,$FE,$8F,$8C,$12
	.dc.b	$78,$BF,$A8,$F5,$FE,$94,$FF,$A9,$38,$7B,$CF,$F1,$59,$CF,$07,$FB
	.dc.b	$8F,$D9,$7D,$1F,$F1,$96,$E3,$FF,$00,$5F,$F0,$2B,$63,$5F,$27,$FB
	.dc.b	$69,$34,$93,$FA,$4F,$A3,$FD,$66,$6C,$3F,$DE,$51,$FE,$BD,$94,$E5
	.dc.b	$FE,$0F,$1C,$27,$7E,$E7,$F3,$4F,$53,$7C,$9F,$D0,$DB,$7F,$C6,$5B
	.dc.b	$87,$F7,$46,$BD,$83,$87,$C2,$D7,$27,$E5,$1C,$71,$7F,$3B,$E1,$FE
	.dc.b	$7D,$E2,$FD,$DB,$F4,$E5,$FC,$9F,$0E,$38,$EA,$FF,$A2,$7D,$1B,$DA
	.dc.b	$BF,$25,$86,$E3,$C8,$FE,$4F,$3B,$71,$87,$F3,$48,$57,$19,$CE,$69
	.dc.b	$CC,$63,$E1,$BF,$83,$3F,$CC,$9B,$7F,$CF,$9F,$EF,$2F,$D1,$30,$7F
	.dc.b	$5B,$77,$46,$9C,$33,$F8,$11,$99,$FE,$D2,$3E,$09,$F4,$7F,$DB,$3C
	.dc.b	$47,$D1,$BE,$87,$75,$AC,$FD,$F5,$A9,$CD,$F1,$BE,$15,$F1,$30,$ED
	.dc.b	$39,$AF,$EE,$5F,$17,$FA,$BC,$FC,$AD,$BF,$F1,$06,$1C,$B0,$F8,$19
	.dc.b	$AB,$51,$FD,$8B,$E5,$3E,$09,$EF,$BF,$EB,$6E,$33,$FF,$00,$36,$3E
	.dc.b	$9B,$67,$AF,$D1,$BE,$73,$FC,$0C,$FE,$CA,$BF,$B2,$B9,$7E,$DB,$99
	.dc.b	$FE,$E6,$E5,$E7,$DD,$CA,$B9,$BF,$A5,$F5,$9F,$15,$FA,$56,$E1,$FA
	.dc.b	$56,$DF,$E9,$5A,$FE,$97,$A7,$FC,$A7,$5C,$49,$5C,$5F,$EE,$B8,$FF
	.dc.b	$85,$6B,$DB,$D3,$E1,$9F,$6B,$EE,$F8,$A9,$7E,$CF,$26,$DF,$C4,$9B
	.dc.b	$C5,$F8,$AC,$7F,$C4,$53,$F0,$27,$FC,$67,$E5,$BE,$4B,$F8,$7E,$3F
	.dc.b	$A0,$F9,$BF,$E7,$B8,$FF,$A7,$67,$38,$7D,$C9,$CD,$7F,$A6,$67,$35
	.dc.b	$F9,$89,$CD,$7F,$95,$67,$35,$FE,$18,$9C,$EB,$FE,$76,$FA,$0F,$98
	.dc.b	$49,$F9,$8D,$BF,$98,$D7,$E6,$6E,$8F,$F0,$ED,$BE,$8F,$FC,$90,$DB
	.dc.b	$FD,$FE,$1D,$DF,$74,$EB,$A3,$7F,$11,$67,$E1,$37,$2F,$80,$DD,$3D
	.dc.b	$CF,$6F,$A3,$EF,$EE,$5B,$F9,$4D,$BF,$90,$DB,$F8,$CF,$9B,$E8,$FD
	.dc.b	$24,$AF,$F1,$0E,$88,$F6,$EB,$EC,$38,$4F,$D7,$70,$FD,$9F,$97,$B8
	.dc.b	$FD,$76,$E3,$F5,$5B,$3F,$4D,$F8,$FF,$1B,$F2,$89,$1E,$27,$E6,$FC
	.dc.b	$AF,$CC,$37,$1F,$87,$7E,$7F,$C5,$7D,$22,$47,$A9,$FA,$BE,$CF,$D4
	.dc.b	$37,$1F,$8F,$70,$D3,$F2,$51,$8C,$7F,$21,$37,$1F,$C7,$4D,$AF,$C6
	.dc.b	$7A,$1F,$82,$23,$F2,$9F,$C2,$FB,$7F,$84,$3F,$0A,$D8,$FD,$07,$D1
	.dc.b	$7E,$28,$8F,$D2,$6E,$3F,$ED,$4F,$0D,$C0,$9C,$CF,$F6,$39,$CC,$7E
	.dc.b	$53,$E4,$7F,$26,$7E,$51,$B7,$FE,$FE,$F3,$3F,$D6,$5D,$6F,$83,$FE
	.dc.b	$C5,$8F,$FC,$00,$9C,$C6,$23,$FE,$00,$B5,$FF,$5F,$C6,$89,$5F,$8A
	.dc.b	$F8,$AF,$D1,$9F,$C0,$2B,$AE,$FF,$B7,$8C,$CF,$F0,$17,$55,$8B,$FE
	.dc.b	$B5,$83,$E8,$FF,$CB,$7F,$03,$FC,$C6,$7F,$DA,$51,$8F,$1B,$FC,$23
	.dc.b	$5B,$C7,$FB,$D5,$EE,$39,$7B,$58,$AD,$DF,$1C,$39,$CE,$5E,$B8,$5B
	.dc.b	$E2,$FF,$BB,$2F,$8D,$66,$F9,$57,$86,$B5,$7D,$1A,$BC,$86,$7F,$EF
	.dc.b	$48,$FF,$7F,$E3,$FE,$1D,$9C,$F9,$3C,$BE,$35,$F2,$9F,$81,$C7,$9D
	.dc.b	$6B,$0E,$27,$F8,$C9,$C6,$BF,$1E,$3C,$19,$77,$8F,$95,$87,$4E,$1F
	.dc.b	$81,$5A,$CB,$AD,$F9,$63,$FC,$60,$64,$FF,$67,$61,$E1,$72,$D7,$D5
	.dc.b	$C4,$F1,$BF,$DA,$91,$F0,$63,$0F,$26,$BC,$9A,$F2,$8F,$7C,$70,$CB
	.dc.b	$D9,$48,$C6,$BF,$E5,$2C,$DF,$26,$FA,$3F,$F2,$46,$1C,$2F,$79,$7C
	.dc.b	$76,$BE,$5C,$BE,$85,$F1,$8F,$B7,$C0,$FA,$37,$CC,$74,$D3,$F2,$31
	.dc.b	$8A,$DB,$5F,$AB,$5C,$7E,$37,$84,$23,$A5,$E1,$B1,$CA,$31,$4A,$B1
	.dc.b	$27,$47,$82,$3A,$F4,$FF,$08,$F3,$FF,$CD,$1D,$3F,$D9,$DB,$CB,$FD
	.dc.b	$95,$B7,$E3,$C5,$F1,$58,$C3,$94,$FF,$77,$AC,$76,$7F,$85,$A3,$FC
	.dc.b	$BB,$A7,$C5,$BE,$6D,$8F,$C4,$9D,$63,$F0,$71,$98,$FC,$6B,$5F,$F3
	.dc.b	$F3,$5F,$9B,$1C,$30,$F2,$C7,$49,$D2,$C3,$F9,$3F,$57,$98,$91,$FE
	.dc.b	$0F,$CB,$D2,$7E,$52,$7F,$96,$B7,$1F,$F4,$14,$FF,$31,$6B,$FC,$BD
	.dc.b	$AF,$F8,$36,$38,$11,$F7,$CF,$EF,$FD,$DF,$95,$CD,$FE,$AE,$9F,$F2
	.dc.b	$6C,$63,$B2,$47,$09,$FF,$68,$36,$3C,$D1,$88,$FE,$88,$BE,$57,$A9
	.dc.b	$23,$FC,$B3,$F1,$15,$FF,$32,$CE,$DF,$F2,$EC,$F9,$27,$5C,$1F,$E8
	.dc.b	$2A,$D5,$70,$AD,$3D,$0F,$35,$79,$08,$FA,$7E,$57,$DE,$47,$FC,$D7
	.dc.b	$F4,$7F,$08,$F5,$91,$F8,$0F,$BB,$E8,$BE,$3F,$D1,$C9,$FF,$39,$C7
	.dc.b	$CA,$30,$FB,$23,$0E,$88,$7C,$08,$FA,$3E,$FC,$8F,$9C,$FC,$C2,$3F
	.dc.b	$47,$E8,$FE,$BF,$C0,$C8,$FF,$41,$C7,$DA,$30,$FC,$C8,$C3,$FE,$88
	.dc.b	$F9,$DF,$BF,$F4,$72,$30,$43,$E9,$B9,$7E,$AC,$7E,$1B,$E8,$E4,$7F
	.dc.b	$A4,$9F,$DA,$AF,$A2,$7D,$1E,$1B,$1F,$CB,$9A,$FE,$C1,$AF,$F3,$A6
	.dc.b	$BF,$EE,$B6,$BF,$EF,$A6,$BD,$E3,$F9,$CB,$4F,$D8,$7B,$8F,$9D,$FC
	.dc.b	$E9,$23,$F8,$1C,$65,$B1,$97,$F2,$6F,$F0,$68,$D4,$7E,$5B,$E7,$FF
	.dc.b	$4E,$47,$FD,$67,$F1,$35,$D1,$FE,$B6,$8F,$D6,$C7,$EB,$BE,$7A,$4F
	.dc.b	$FA,$E9,$F9,$3F,$46,$F6,$29,$EC,$71,$3F,$D5,$9A,$FF,$18,$6B,$FD
	.dc.b	$89,$A8,$8E,$1B,$1D,$4F,$E3,$FF,$46,$FA,$0F,$E4,$AB,$E8,$29,$88
	.dc.b	$95,$FC,$B2,$74,$D9,$CB,$FE,$E0,$F8,$97,$B3,$FC,$DD,$7B,$74,$C4
	.dc.b	$49,$F4,$77,$FF,$BB,$27,$51,$8F,$0F,$D9,$E1,$FF,$BA,$EF,$FA,$17
	.dc.b	$C6,$FD,$6C,$15,$A8,$C3,$78,$FF,$2F,$F4,$6E,$8F,$CC,$FA,$BF,$D8
	.dc.b	$F0,$FE,$5B,$39,$CF,$C0,$7C,$B7,$0A,$3F,$99,$CF,$1C,$3E,$84,$FF
	.dc.b	$8A,$BC,$EE,$F8,$48,$E6,$64,$EB,$A7,$FB,$B3,$CD,$FE,$B8,$7F,$59
	.dc.b	$8E,$D7,$F0,$23,$18,$FB,$A7,$FA,$64,$7F,$8E,$1E,$6E,$A7,$83,$D0
	.dc.b	$EB,$8F,$F6,$9F,$8E,$7B,$D7,$79,$87,$92,$72,$D9,$CB,$E4,$3F,$AE
	.dc.b	$47,$DE,$74,$7B,$3E,$2D,$D5,$FD,$DD,$B1,$98,$FE,$13,$5F,$E2,$EC
	.dc.b	$FF,$AE,$BF,$E2,$E9,$F9,$6D,$7F,$38,$D7,$D3,$EA,$7F,$02,$8C,$11
	.dc.b	$C1,$FD,$AF,$DB,$7B,$FF,$E0,$95,$EF,$D3,$11,$FE,$0D,$F9,$27,$8B
	.dc.b	$FD,$CD,$B8,$FF,$6A,$6B,$FE,$A3,$6B,$FE,$A1,$F9,$47,$C9,$7F,$85
	.dc.b	$DB,$1F,$80,$6B,$F2,$F9,$10,$FA,$3F,$F0,$F7,$CB,$7F,$B8,$6B,$F0
	.dc.b	$ED,$78,$37,$7F,$C2,$5A,$3F,$CA,$53,$C3,$DD,$FF,$06,$46,$23,$FD
	.dc.b	$74,$D7,$FD,$70,$D7,$F5,$ED,$7F,$C3,$4D,$9E,$D3,$D0,$FF,$0C,$7C
	.dc.b	$BF,$DE,$AF,$63,$6F,$CA,$DE,$8F,$DD,$FB,$DF,$C2,$AB,$E1,$35,$FF
	.dc.b	$65,$35,$DF,$83,$FE,$25,$8C,$47,$F8,$B5,$B8,$FF,$9B,$A7,$44,$9F
	.dc.b	$F1,$0F,$CC,$BD,$3F,$F2,$53,$63,$F6,$6D,$7E,$A7,$CD,$3E,$7B,$FC
	.dc.b	$A0,$D8,$FF,$15,$B5,$FF,$72,$35,$FF,$71,$35,$FE,$32,$D7,$FC,$88
	.dc.b	$DC,$0D,$31,$F3,$9F,$B2,$9F,$C3,$37,$5F,$BE,$DD,$3F,$92,$FC,$AF
	.dc.b	$E5,$27,$F0,$6F,$E5,$2B,$3C,$BF,$E4,$D8,$C4,$7F,$95,$5B,$89,$A4
	.dc.b	$9E,$07,$F9,$56,$1F,$17,$FE,$73,$6D,$7F,$1C,$6B,$F9,$DF,$80,$FC
	.dc.b	$03,$FE,$63,$7F,$0C,$FF,$28,$8F,$F4,$87,$6F,$F9,$86,$73,$3F,$D0
	.dc.b	$1A,$F1,$8D,$11,$FF,$2E,$C6,$A5,$F3,$CF,$91,$FF,$44,$36,$BF,$98
	.dc.b	$B5,$FD,$97,$C3,$FF,$93,$B4,$FD,$5B,$FE,$50,$8F,$33,$5F,$EA,$6D
	.dc.b	$7F,$CF,$CD,$B3,$49,$EF,$7F,$9F,$5D,$7C,$A7,$CF,$AF,$E8,$6D,$7F
	.dc.b	$82,$36,$FF,$7B,$F4,$7F,$95,$6F,$2F,$FA,$1E,$3F,$31,$5F,$2D,$AF
	.dc.b	$F6,$86,$BF,$E9,$3F,$8B,$FC,$43,$7C,$4F,$F5,$64,$63,$33,$FC,$F6
	.dc.b	$7F,$8B,$39,$24,$F0,$F1,$FF,$D6,$9F,$2C,$F5,$E9,$2F,$FA,$FC,$7F
	.dc.b	$74,$7F,$D4,$8E,$DC,$C6,$DF,$E4,$91,$BE,$17,$D1,$7F,$5F,$9F,$EA
	.dc.b	$51,$8F,$2F,$FD,$7B,$39,$DB,$FB,$34,$BE,$45,$EC,$EA,$E6,$1F,$EC
	.dc.b	$6C,$38,$75,$7F,$8A,$70,$E1,$7F,$EC,$8C,$38,$DF,$FA,$F5,$AF,$FB
	.dc.b	$19,$AF,$FB,$19,$AF,$F8,$13,$E8,$FF,$C5,$99,$7F,$8D,$7B,$7E,$04
	.dc.b	$A7,$FC,$64,$E7,$5E,$1E,$C7,$F8,$32,$B7,$5C,$AB,$F8,$3D,$75,$F0
	.dc.b	$7F,$C2,$91,$83,$FC,$85,$3D,$1C,$1D,$EF,$A5,$71,$C3,$C1,$3C,$2F
	.dc.b	$9B,$FE,$1C,$3C,$CE,$1E,$07,$FB,$AA,$FC,$8E,$D7,$C6,$B8,$3F,$DD
	.dc.b	$F3,$EF,$8C,$3C,$5A,$F1,$6B,$FE,$F6,$E8,$FF,$3F,$BF,$AC,$3F,$D0
	.dc.b	$2E,$08,$FB,$87,$C8,$7F,$89,$F4,$F9,$F1,$FE,$65,$9F,$7B,$FE,$67
	.dc.b	$9D,$67,$DE,$39,$46,$1F,$85,$3C,$1E,$BC,$1F,$23,$81,$F4,$8F,$84
	.dc.b	$FF,$79,$C6,$1B,$19,$9F,$99,$4A,$D4,$61,$7E,$08,$F8,$BF,$F2,$07
	.dc.b	$0F,$E7,$23,$15,$FE,$56,$AE,$37,$FA,$49,$CB,$6F,$8D,$7E,$76,$73
	.dc.b	$7F,$A6,$C3,$C9,$7B,$AF,$F2,$AF,$D2,$BE,$4B,$8C,$BB,$13,$9A,$DF
	.dc.b	$CE,$EA,$73,$76,$B5,$7F,$E1,$F6,$E3,$FE,$18,$57,$BA,$E5,$F4,$AF
	.dc.b	$A5,$A7,$8B,$BD,$F5,$8F,$8A,$CE,$BF,$E6,$CB,$E0,$DB,$E1,$5B,$8C
	.dc.b	$1F,$E7,$23,$FC,$86,$7D,$4D,$7F,$D1,$7E,$C7,$F9,$BF,$6F,$EA,$D3
	.dc.b	$A7,$F4,$93,$A8,$E1,$C9,$FE,$D1,$7F,$CF,$AD,$9F,$F1,$F4,$71,$DF
	.dc.b	$FA,$6F,$C0,$FC,$CE,$47,$E5,$DF,$E9,$AA,$FC,$13,$6F,$F6,$53,$A9
	.dc.b	$FC,$6C,$E6,$7F,$D2,$33,$99,$E2,$68,$87,$9B,$E9,$BE,$B6,$27,$C5
	.dc.b	$7D,$F1,$FE,$C3,$1F,$6E,$DE,$4F,$C3,$C7,$FB,$A3,$5F,$EF,$8D,$CF
	.dc.b	$F3,$8D,$9E,$39,$7F,$76,$8C,$11,$FA,$1E,$7F,$EB,$2F,$FC,$C8,$DC
	.dc.b	$7F,$CB,$B7,$CA,$73,$D5,$F2,$7C,$FF,$F0,$9E,$1F,$A9,$6C,$FC,$86
	.dc.b	$BF,$19,$AF,$F3,$F6,$BF,$B3,$6D,$7F,$A0,$BC,$FF,$DD,$C6,$B3,$7E
	.dc.b	$20,$F9,$AF,$D0,$37,$1F,$DA,$57,$8B,$3F,$F3,$AC,$E7,$2F,$F4,$07
	.dc.b	$A9,$F8,$57,$FC,$44,$D7,$FD,$BE,$D7,$F6,$0D,$C4,$D2,$71,$7D,$17
	.dc.b	$5E,$AB,$DF,$FD,$A3,$73,$FF,$2F,$D6,$27,$F2,$5E,$B3,$E6,$BF,$8C
	.dc.b	$B6,$BF,$8A,$37,$2F,$C9,$B5,$FC,$8B,$5F,$ED,$4D,$7F,$8F,$37,$83
	.dc.b	$FC,$D9,$5F,$87,$F9,$97,$D0,$FF,$09,$6E,$1F,$C7,$6F,$C1,$BF,$FA
	.dc.b	$7A,$73,$8F,$FB,$4F,$B9,$F1,$4F,$4B,$F2,$91,$86,$C6,$71,$FE,$63
	.dc.b	$39,$6C,$66,$BF,$B0,$4E,$5B,$19,$7F,$DF,$F1,$97,$FD,$81,$F5,$3F
	.dc.b	$C8,$7E,$00,$FF,$17,$AE,$99,$7F,$AD,$A7,$38,$7F,$39,$FD,$83,$F0
	.dc.b	$93,$FE,$18,$6C,$30,$FF,$19,$FE,$2F,$F8,$B8,$FF,$0B,$4E,$A7,$F9
	.dc.b	$9F,$E5,$CF,$A5,$FD,$A7,$92,$8C,$5F,$F5,$08,$C4,$EA,$73,$5F,$C3
	.dc.b	$BF,$2E,$7B,$3F,$CE,$FE,$B3,$FD,$A9,$8F,$F4,$F9,$CB,$E5,$D3,$FD
	.dc.b	$A3,$3A,$B4,$27,$D5,$EF,$7F,$6F,$7F,$B0,$B6,$7F,$C1,$CD,$7E,$00
	.dc.b	$70,$F2,$3F,$83,$1A,$72,$EC,$C9,$BF,$07,$F8,$A7,$C5,$BE,$83,$FA
	.dc.b	$B3,$6F,$D4,$6C,$B5,$FD,$7E,$34,$FF,$BD,$FE,$B1,$F3,$BF,$D7,$DB
	.dc.b	$5C,$3C,$14,$3E,$7B,$55,$2B,$53,$C2,$FA,$56,$5F,$93,$3F,$81,$B7
	.dc.b	$CA,$B9,$BF,$B7,$3F,$CE,$67,$C0,$E5,$BE,$56,$A7,$6D,$8C,$9E,$EE
	.dc.b	$57,$E0,$DF,$DC,$A3,$F0,$CD,$C7,$FC,$85,$3A,$F5,$3F,$1B,$39,$9F
	.dc.b	$AF,$39,$AE,$35,$A2,$3F,$E2,$0F,$49,$F8,$9F,$C1,$E4,$FF,$3B,$F3
	.dc.b	$3E,$9B,$FC,$1F,$5A,$E2,$FE,$CD,$39,$C3,$E7,$7D,$27,$BF,$4F,$8E
	.dc.b	$7F,$85,$5C,$63,$F9,$18,$C6,$DF,$E1,$59,$CD,$7C,$D6,$CF,$1D,$F5
	.dc.b	$5B,$78,$47,$1E,$6B,$7E,$26,$C7,$8B,$89,$87,$F5,$9A,$ED,$3C,$5F
	.dc.b	$C6,$F3,$FF,$99,$7C,$2F,$ED,$4D,$8F,$D2,$FB,$E9,$CB,$63,$31,$82
	.dc.b	$38,$3F,$C3,$5D,$0F,$A4,$FC,$F9,$FE,$3A,$BE,$D8,$79,$E7,$55,$FE
	.dc.b	$4B,$F9,$D7,$D2,$7F,$8E,$FF,$13,$FE,$1D,$E5,$A7,$35,$F9,$F9,$CC
	.dc.b	$63,$0F,$85,$18,$BE,$5E,$97,$F9,$17,$89,$F4,$54,$FF,$11,$7E,$43
	.dc.b	$E2,$C3,$E5,$4E,$B2,$FD,$63,$63,$35,$C3,$6E,$55,$F3,$38,$3F,$BD
	.dc.b	$EF,$FE,$91,$F5,$7E,$49,$27,$E6,$B7,$0F,$E2,$35,$C0,$FE,$FD,$C0
	.dc.b	$F8,$EF,$E0,$E7,$C9,$8F,$E6,$53,$FC,$16,$3D,$9F,$61,$30,$FF,$3A
	.dc.b	$78,$5F,$C9,$AB,$D7,$39,$6C,$EB,$C6,$C7,$57,$E5,$FA,$3F,$C6,$9C
	.dc.b	$5F,$1C,$FF,$23,$7E,$46,$FA,$FC,$61,$F1,$26,$E3,$FE,$E4,$AD,$4E
	.dc.b	$4F,$93,$F7,$DF,$E3,$5E,$43,$49,$1F,$E6,$18,$FE,$75,$97,$8D,$B1
	.dc.b	$FC,$4D,$E2,$FD,$4F,$9A,$FB,$87,$DD,$FC,$16,$C7,$D8,$6E,$3F,$71
	.dc.b	$B7,$D5,$FC,$27,$CE,$79,$49,$5E,$37,$D1,$F0,$9F,$89,$37,$1F,$84
	.dc.b	$7F,$19,$EB,$7B,$49,$1E,$B7,$E0,$7B,$1F,$91,$37,$1F,$8C,$70,$D3
	.dc.b	$F1,$F1,$8C,$7F,$27,$37,$1F,$89,$36,$BE,$27,$D1,$FC,$D2,$3F,$2D
	.dc.b	$FA,$1F,$1D,$FA,$13,$F4,$4D,$8F,$CE,$FC,$87,$D5,$23,$F4,$DB,$8F
	.dc.b	$FA,$43,$64,$62,$3F,$B3,$F1,$FF,$9E,$E3,$EF,$7C,$8F,$EB,$CF,$D8
	.dc.b	$36,$FF,$D9,$1F,$5A,$FB,$7F,$F6,$BB,$63,$FD,$A4,$DC,$3F,$D2,$2D
	.dc.b	$7F,$D1,$ED,$7F,$C2,$DE,$2A,$31,$1F,$ED,$D6,$C6,$6B,$F2,$47,$F0
	.dc.b	$33,$FD,$45,$1D,$F6,$FC,$84,$66,$FF,$80,$BA,$31,$7F,$D0,$2F,$C1
	.dc.b	$8F,$F3,$FA,$F2,$B6,$7C,$B8,$23,$11,$E5,$6D,$72,$AF,$F1,$81,$C8
	.dc.b	$B3,$CA,$77,$5C,$E7,$33,$CD,$FF,$1C,$7C,$47,$FB,$A6,$3F,$DD,$1A
	.dc.b	$7F,$90,$14,$67,$0F,$04,$63,$E3,$3F,$5D,$EE,$FF,$59,$AE,$53,$FD
	.dc.b	$33,$AD,$E5,$F6,$BF,$F1,$3F,$97,$FE,$38,$8C,$53,$DE,$7D,$17,$F6
	.dc.b	$98,$EC,$7F,$93,$6B,$83,$83,$B5,$72,$69,$C2,$B9,$FA,$5E,$5C,$3C
	.dc.b	$38,$F9,$2B,$96,$1D,$1F,$F1,$D9,$C8,$AE,$5E,$38,$69,$F2,$7C,$FF
	.dc.b	$FA,$65,$C4,$A3,$11,$C9,$AF,$C2,$6B,$CB,$E0,$FD,$53,$C5,$B3,$D7
	.dc.b	$4F,$AD,$9B,$E1,$7F,$D7,$98,$6A,$E3,$EC,$BE,$19,$FF,$9D,$6F,$87
	.dc.b	$85,$FE,$58,$C7,$EE,$79,$7E,$7C,$3E,$9F,$4F,$FB,$BF,$D7,$D6,$3F
	.dc.b	$CA,$DB,$2E,$3F,$EF,$19,$E1,$B1,$F8,$E6,$BF,$E1,$D3,$86,$1D,$DC
	.dc.b	$46,$B9,$7F,$CD,$6D,$8F,$DB,$CE,$E7,$2D,$8C,$D7,$F9,$CA,$7F,$86
	.dc.b	$F2,$FE,$4B,$A7,$FA,$26,$5F,$E7,$66,$DF,$F9,$A3,$1F,$F0,$87,$C6
	.dc.b	$67,$0F,$E9,$17,$6B,$F4,$6D,$D1,$8F,$5A,$F2,$DE,$F0,$E7,$87,$E1
	.dc.b	$5A,$FE,$31,$AF,$AA,$3F,$14,$FC,$46,$C6,$A3,$A9,$CF,$34,$F0,$C3
	.dc.b	$C7,$E5,$F9,$71,$34,$C7,$C5,$4C,$3F,$B4,$3C,$64,$D2,$46,$88,$7F
	.dc.b	$3B,$78,$71,$77,$24,$7F,$A2,$FE,$9D,$F6,$BF,$1D,$B1,$FA,$7C,$51
	.dc.b	$88,$FD,$7C,$62,$7D,$93,$99,$FE,$8E,$FF,$82,$DE,$DE,$57,$53,$B7
	.dc.b	$CF,$58,$C9,$0F,$F2,$F4,$A3,$51,$89,$BC,$3F,$E1,$BA,$D4,$FA,$EB
	.dc.b	$C6,$47,$D5,$1A,$7D,$2B,$E2,$7E,$01,$23,$E2,$7D,$9F,$C3,$3D,$A4
	.dc.b	$79,$BF,$03,$EC,$FE,$47,$C3,$C9,$F9,$91,$F2,$CC,$3E,$C8,$C3,$A2
	.dc.b	$1F,$02,$3E,$97,$B5,$23,$D2,$FC,$E2,$3E,$CF,$B3,$FB,$04,$7F,$4D
	.dc.b	$E8,$FF,$A7,$DF,$BD,$19,$7E,$6B,$FA,$18,$D4,$7F,$5F,$FA,$C7,$9B
	.dc.b	$FA,$56,$C7,$EC,$5A,$FF,$65,$FC,$8D,$F3,$5B,$A9,$3B,$7F,$5D,$3A
	.dc.b	$8F,$F6,$D7,$3B,$E5,$FF,$6C,$D8,$CC,$7E,$ED,$7D,$87,$F9,$31,$FF
	.dc.b	$74,$98,$F5,$BF,$66,$D3,$F6,$B1,$B7,$FA,$03,$5F,$E8,$ED,$44,$70
	.dc.b	$D8,$EB,$18,$E7,$FD,$1F,$B7,$FA,$72,$57,$F1,$5F,$7B,$FD,$D4,$FE
	.dc.b	$B6,$32,$FE,$71,$FE,$21,$1A,$8F,$F0,$77,$05,$E7,$FF,$3B,$7D,$63
	.dc.b	$EA,$FF,$8D,$36,$BF,$D5,$5F,$09,$FC,$66,$F8,$78,$7F,$EF,$BA,$D5
	.dc.b	$7F,$B1,$DB,$87,$FB,$5F,$C3,$FF,$57,$61,$F4,$3E,$52,$5F,$D4,$7E
	.dc.b	$87,$D7,$F0,$C7,$E2,$CA,$3A,$BF,$DC,$1A,$E9,$AF,$CC,$6B,$C7,$3F
	.dc.b	$AF,$18,$23,$FC,$2B,$E5,$7F,$89,$7C,$AC,$AF,$E2,$D1,$FC,$54,$C3
	.dc.b	$FC,$1A,$30,$FE,$57,$D4,$FC,$C1,$1F,$CC,$7A,$F5,$FC,$F4,$5D,$7E
	.dc.b	$27,$57,$F9,$5E,$7F,$07,$1F,$EF,$DF,$8B,$7E,$25,$FA,$68,$F9,$5D
	.dc.b	$DF,$EE,$1A,$FE,$A8,$D5,$1C,$67,$86,$DF,$2B,$E9,$BF,$C0,$35,$FE
	.dc.b	$13,$6B,$FE,$1A,$6B,$BF,$47,$F3,$B1,$88,$FC,$BB,$53,$4D,$74,$39
	.dc.b	$7F,$BD,$63,$F8,$CF,$8B,$FC,$59,$F8,$CC,$66,$1F,$01,$FC,$54,$6A
	.dc.b	$30,$2F,$A4,$FF,$6D,$FA,$EF,$C3,$4E,$5B,$26,$9F,$D8,$A3,$9C,$67
	.dc.b	$8F,$91,$23,$33,$B7,$E5,$D7,$ED,$AB,$FB,$3F,$D3,$7F,$08,$24,$FF
	.dc.b	$04,$FD,$4B,$C6,$9F,$C8,$93,$0F,$E0,$ED,$8F,$EF,$6D,$7F,$C8,$CF
	.dc.b	$E6,$BA,$5C,$0F,$A3,$FF,$06,$7B,$BF,$72,$FE,$23,$6B,$73,$FE,$13
	.dc.b	$F9,$3F,$F2,$67,$A3,$FE,$08,$6D,$FF,$94,$DA,$FF,$96,$9A,$FF,$8B
	.dc.b	$3E,$6B,$FC,$30,$FF,$8E,$5B,$67,$3E,$D3,$F1,$A7,$C2,$FF,$88,$7E
	.dc.b	$BD,$E1,$D2,$63,$F7,$2D,$F8,$AF,$08,$C3,$5F,$F3,$6C,$6A,$FF,$C4
	.dc.b	$2D,$8F,$F3,$63,$5D,$CF,$17,$F8,$4C,$65,$B1,$97,$FC,$BC,$D7,$F1
	.dc.b	$DF,$3F,$F3,$DF,$C3,$B5,$FC,$19,$AD,$BF,$C7,$5F,$91,$BD,$3A,$7F
	.dc.b	$8D,$1C,$4F,$F8,$F5,$AF,$F8,$F5,$AF,$F9,$69,$AF,$FB,$41,$AF,$F9
	.dc.b	$A5,$BB,$1C,$F3,$30,$7C,$AF,$F9,$45,$B7,$EA,$8F,$F2,$A7,$D4,$FF
	.dc.b	$6A,$EC,$75,$7E,$76,$3F,$B6,$D7,$FB,$C1,$AE,$7D,$1F,$F2,$54,$62
	.dc.b	$3F,$68,$DC,$4D,$24,$FF,$94,$61,$EA,$7F,$99,$5B,$5F,$B4,$6B,$FE
	.dc.b	$E1,$FC,$65,$F2,$9F,$E6,$76,$C7,$F9,$9D,$AF,$F9,$9D,$AF,$FA,$11
	.dc.b	$AF,$FB,$E9,$AF,$FA,$4D,$BB,$1A,$6A,$BF,$88,$79,$3E,$72,$57,$F0
	.dc.b	$AA,$77,$FF,$9F,$DB,$1F,$E5,$96,$BF,$E5,$86,$BF,$D7,$2B,$FD,$49
	.dc.b	$3F,$E7,$59,$CC,$FF,$0F,$6B,$FC,$85,$B3,$C2,$3F,$CC,$DF,$2D,$FB
	.dc.b	$68,$34,$FD,$BF,$E4,$4F,$95,$FF,$48,$36,$BF,$D2,$0D,$7F,$D2,$0D
	.dc.b	$7F,$D5,$AD,$7F,$C2,$F6,$8C,$47,$FA,$F5,$B1,$99,$FF,$4A,$C6,$AF
	.dc.b	$CB,$F0,$1F,$EA,$5C,$3F,$DB,$FC,$3F,$92,$C7,$F9,$F4,$63,$87,$F9
	.dc.b	$87,$D7,$FE,$6A,$7F,$63,$82,$31,$1F,$2D,$AF,$0E,$3F,$F5,$04,$62
	.dc.b	$3F,$A5,$74,$FF,$3D,$C7,$C1,$68,$C6,$1F,$CF,$DB,$59,$F8,$D7,$AB
	.dc.b	$FE,$BD,$9C,$DF,$1F,$84,$C1,$FE,$C3,$85,$70,$BF,$EA,$3F,$31,$F8
	.dc.b	$4C,$BC,$2D,$CB,$FA,$73,$67,$FB,$02,$FB,$1F,$F6,$63,$9C,$8D,$24
	.dc.b	$7F,$B3,$7E,$B1,$F5,$B4,$97,$FE,$DB,$FA,$DF,$F6,$AB,$FE,$DD,$9D
	.dc.b	$B6,$76,$FF,$20,$E5,$57,$08,$C3,$5F,$F7,$9B,$5F,$C3,$B5,$47,$2F
	.dc.b	$89,$FF,$34,$FD,$73,$E5,$BF,$C3,$35,$FC,$2B,$0F,$F2,$0C,$BE,$5B
	.dc.b	$FC,$87,$E0,$79,$A9,$39,$7F,$CD,$71,$98,$E9,$C1,$8B,$E7,$87,$6A
	.dc.b	$D6,$1D,$27,$BB,$70,$FF,$3A,$E1,$C9,$F2,$18,$BF,$3B,$FE,$56,$3C
	.dc.b	$6E,$1E,$F8,$7F,$BE,$27,$94,$6D,$FE,$CF,$B6,$7C,$6F,$B7,$3C,$CF
	.dc.b	$6B,$87,$83,$5D,$7C,$8F,$FA,$49,$34,$8F,$FA,$59,$F8,$4F,$F9,$B3
	.dc.b	$E5,$3F,$61,$A7,$CB,$8F,$F4,$9C,$E9,$FF,$4A,$CE,$B4,$EF,$1F,$08
	.dc.b	$CF,$1E,$DE,$26,$44,$F1,$7F,$03,$C1,$FE,$B8,$9F,$77,$07,$E3,$3E
	.dc.b	$1F,$F8,$15,$7E,$29,$AF,$E2,$9A,$FE,$2B,$8B,$FC,$49,$5F,$99,$F2
	.dc.b	$3F,$85,$70,$7F,$A3,$F1,$ED,$7F,$99,$BD,$B6,$B7,$7F,$C7,$27,$37
	.dc.b	$C7,$0C,$DF,$7B,$BC,$AF,$39,$F2,$18,$6E,$5F,$E4,$E5,$F2,$59,$99
	.dc.b	$39,$78,$46,$30,$FE,$3B,$EE,$3E,$67,$BB,$6F,$FC,$51,$F5,$DF,$ED
	.dc.b	$97,$FC,$B7,$19,$AE,$35,$C0,$F4,$F6,$AF,$84,$EA,$FF,$CC,$63,$BB
	.dc.b	$E6,$E9,$FE,$04,$6C,$64,$FC,$A4,$65,$BC,$AF,$84,$F8,$AD,$8E,$B5
	.dc.b	$89,$F9,$7C,$8F,$AE,$FF,$08,$D7,$E0,$9B,$8F,$EA,$E7,$53,$F8,$06
	.dc.b	$BB,$31,$3C,$27,$FD,$8F,$3A,$79,$46,$5F,$F3,$D6,$0E,$EF,$39,$E4
	.dc.b	$9F,$0F,$27,$F4,$49,$D6,$1F,$DF,$FE,$BB,$E2,$E2,$69,$23,$FC,$00
	.dc.b	$DC,$7F,$C0,$0D,$7F,$C0,$1E,$BF,$E9,$27,$FD,$26,$D8,$FE,$BB,$5D
	.dc.b	$67,$3D,$3F,$23,$EB,$BF,$A8,$C7,$EA,$1B,$5F,$1D,$AF,$C2,$6B,$FD
	.dc.b	$E1,$AF,$F3,$26,$D1,$A6,$B7,$7D,$73,$E0,$BF,$42,$DC,$3F,$D6,$B7
	.dc.b	$C7,$C6,$CD,$7E,$9B,$D3,$FC,$16,$5F,$C9,$71,$FF,$2B,$71,$7E,$3A
	.dc.b	$73,$3F,$E2,$36,$BF,$EA,$F6,$BF,$E7,$48,$E0,$4C,$7F,$27,$F3,$2F
	.dc.b	$33,$F6,$AD,$8F,$F3,$CC,$E2,$7F,$23,$F3,$CF,$57,$F1,$8D,$AF,$E2
	.dc.b	$6D,$C3,$F2,$4D,$7F,$1A,$D7,$FC,$46,$D7,$FA,$EB,$75,$39,$CD,$F0
	.dc.b	$4F,$94,$FE,$16,$DC,$3E,$5D,$78,$79,$BF,$C2,$B3,$9C,$7F,$D2,$3D
	.dc.b	$6F,$51,$F1,$5F,$C9,$23,$0D,$8C,$E3,$FE,$A6,$6D,$7F,$AB,$1A,$FF
	.dc.b	$A9,$9A,$8F,$F1,$23,$86,$CF,$F0,$6F,$8C,$7C,$57,$F2,$46,$DF,$FA
	.dc.b	$F6,$FB,$E5,$FE,$2E,$9C,$E1,$FE,$0B,$F8,$1F,$C5,$37,$97,$F8,$BF
	.dc.b	$D7,$7E,$72,$FF,$C8,$9C,$3F,$EC,$F9,$FE,$67,$E3,$3E,$B9,$F2,$DF
	.dc.b	$D1,$23,$0D,$8C,$EB,$FE,$DF,$6E,$5F,$C3,$7E,$9D,$F3,$FF,$CF,$3E
	.dc.b	$B9,$F4,$BF,$C5,$A7,$38,$7C,$3D,$7F,$C8,$F5,$A8,$C1,$F1,$FD,$CF
	.dc.b	$EC,$78,$FF,$80,$DA,$FC,$B3,$FC,$EA,$7E,$FC,$C3,$A7,$91,$92,$F8
	.dc.b	$5F,$D8,$23,$51,$CA,$3F,$9B,$FC,$C3,$EB,$BF,$AA,$B6,$FC,$B5,$B6
	.dc.b	$F6,$27,$53,$FE,$6C,$FA,$E7,$D5,$FF,$5E,$6D,$F0,$5F,$53,$99,$93
	.dc.b	$97,$84,$62,$BF,$CE,$FF,$5C,$E9,$FE,$D2,$FF,$2C,$8C,$C6,$3C,$DC
	.dc.b	$B7,$0F,$3D,$71,$6C,$FE,$0B,$EB,$BF,$B6,$C7,$F8,$9F,$9D,$18,$C3
	.dc.b	$FC,$69,$18,$F4,$FF,$0D,$39,$9F,$A9,$39,$AE,$53,$99,$FD,$14,$E5
	.dc.b	$E7,$19,$5E,$9B,$EB,$BE,$3C,$7F,$99,$F9,$BF,$D0,$38,$7D,$2A,$D6
	.dc.b	$DF,$8D,$F8,$2F,$F0,$5F,$D4,$7B,$70,$FF,$4E,$B7,$A3,$FD,$2A,$FC
	.dc.b	$76,$F0,$70,$26,$3F,$72,$74,$FE,$E7,$91,$1C,$E3,$0F,$EF,$5E,$1C
	.dc.b	$DE,$2C,$3C,$6D,$8C,$9D,$63,$2F,$E3,$39,$FF,$D2,$3E,$2F,$E5,$7D
	.dc.b	$FF,$E3,$B3,$FE,$59,$9C,$B6,$33,$5F,$A5,$6B,$FA,$26,$BF,$AB,$6B
	.dc.b	$FE,$DC,$6E,$3F,$99,$FA,$7F,$CF,$9F,$E3,$6B,$EF,$5F,$EC,$6E,$94
	.dc.b	$FF,$84,$3E,$41,$F0,$FF,$E3,$AF,$AE,$FE,$23,$C1,$FC,$3A,$73,$A7
	.dc.b	$E9,$A7,$2F,$FB,$82,$33,$8F,$FB,$4A,$72,$FF,$A7,$23,$35,$FE,$CC
	.dc.b	$9C,$E7,$FE,$0F,$FA,$EF,$F2,$4F,$51,$FF,$11,$FD,$07,$9B,$2F,$DA
	.dc.b	$CE,$5B,$3C,$0D,$E9,$63,$E9,$6F,$FD,$F3,$7F,$F6,$1F,$A9,$F4,$C9
	.dc.b	$3F,$45,$BB,$7D,$16,$BF,$C5,$73,$FE,$09,$F1,$FF,$C2,$8F,$99,$87
	.dc.b	$F3,$3B,$FC,$67,$0A,$7D,$7F,$1B,$27,$FC,$07,$E0,$7F,$28,$C7,$FD
	.dc.b	$EF,$B2,$B5,$18,$7F,$C8,$73,$BE,$1F,$F7,$9D,$6E,$FE,$1F,$91,$F2
	.dc.b	$0F,$F2,$4F,$D2,$BE,$A3,$13,$A2,$65,$FE,$D7,$9D,$46,$30,$EB,$F4
	.dc.b	$9F,$E5,$EC,$4D,$24,$FF,$97,$A3,$F9,$F5,$FA,$BC,$8F,$B3,$5F,$69
	.dc.b	$95,$F6,$3E,$BA,$D7,$C0,$6E,$1E,$C6,$CF,$ED,$70,$47,$E1,$3E,$BB
	.dc.b	$C8,$49,$F1,$3E,$7F,$84,$FC,$49,$B8,$FC,$3B,$F8,$CF,$5B,$D8,$48
	.dc.b	$F5,$3E,$FF,$63,$F2,$26,$E3,$F1,$EE,$1A,$7E,$4A,$31,$97,$DA,$9B
	.dc.b	$97,$D6,$9B,$5E,$BF,$AE,$F9,$84,$7E,$53,$F3,$FE,$37,$F4,$27,$E8
	.dc.b	$9B,$1F,$A0,$F8,$FF,$A8,$47,$E9,$37,$1F,$F2,$56,$A8,$C4,$7E,$66
	.dc.b	$31,$3F,$67,$E4,$3F,$5E,$7E,$C2,$AC,$7F,$96,$3E,$B9,$F4,$DF,$E8
	.dc.b	$D6,$C7,$FB,$51,$B8,$7F,$A4,$DA,$FF,$A3,$9A,$FF,$89,$FD,$74,$62
	.dc.b	$11,$8A,$EC,$7E,$44,$FE,$06,$7F,$A9,$27,$BE,$BF,$A2,$8C,$DF,$F0
	.dc.b	$47,$46,$2D,$3F,$E2,$6A,$62,$BF,$E7,$D5,$E5,$9C,$C6,$23,$CB,$18
	.dc.b	$7F,$DE,$38,$7A,$AB,$94,$62,$75,$5F,$06,$72,$D3,$A5,$71,$6C,$F5
	.dc.b	$F8,$7F,$F7,$3D,$BB,$7F,$D6,$5E,$E7,$69,$CE,$1C,$69,$1C,$2B,$78
	.dc.b	$7F,$95,$63,$9C,$FF,$5E,$F5,$DE,$4E,$7E,$2F,$FC,$61,$E5,$FF,$8F
	.dc.b	$A3,$14,$F7,$DF,$3D,$FD,$DA,$3B,$DF,$0C,$39,$C6,$0F,$F3,$26,$1C
	.dc.b	$DC,$1E,$1C,$3A,$E1,$D3,$CF,$F6,$61,$CE,$7C,$35,$B3,$FC,$88,$74
	.dc.b	$C3,$BB,$8C,$3B,$36,$32,$FB,$3E,$BB,$FD,$40,$F2,$3F,$BF,$B8,$79
	.dc.b	$35,$E1,$F0,$7B,$76,$F9,$29,$3D,$34,$F9,$B1,$89,$F6,$FC,$2F,$FB
	.dc.b	$0E,$F8,$5E,$F2,$EB,$1F,$12,$F8,$3F,$27,$45,$F1,$8F,$99,$C1,$FD
	.dc.b	$DB,$0F,$F3,$36,$5F,$7F,$9D,$EC,$E3,$3F,$E5,$0D,$C3,$F5,$BE,$67
	.dc.b	$A1,$FB,$B8,$F3,$7F,$27,$19,$7F,$3A,$DE,$2F,$F3,$A4,$EE,$73,$97
	.dc.b	$F9,$DE,$73,$5B,$9F,$DF,$35,$F0,$CE,$DF,$E7,$98,$7F,$9F,$F9,$FF
	.dc.b	$AD,$B6,$F3,$3F,$EE,$8B,$7C,$66,$63,$F4,$99,$FE,$73,$0E,$31,$FE
	.dc.b	$04,$C3,$59,$7F,$87,$2E,$E1,$CA,$BF,$05,$39,$6C,$65,$FC,$7C,$65
	.dc.b	$B1,$97,$DF,$3F,$85,$9D,$46,$0F,$EC,$9E,$57,$9F,$99,$97,$FA,$2E
	.dc.b	$7F,$C3,$33,$98,$7C,$55,$3F,$21,$5C,$1B,$1A,$AF,$EE,$11,$82,$38
	.dc.b	$21,$FD,$0B,$8F,$B1,$1F,$F4,$EF,$D6,$3E,$BB,$CC,$D8,$FD,$36,$08
	.dc.b	$C4,$7E,$CA,$31,$3E,$69,$CC,$FF,$89,$5F,$F0,$83,$DB,$C9,$EA,$76
	.dc.b	$FA,$EB,$19,$21,$FC,$DB,$87,$FC,$E9,$49,$DC,$E4,$FE,$49,$1D,$9C
	.dc.b	$61,$E7,$BE,$D1,$EE,$AF,$31,$1F,$6D,$75,$6F,$C4,$FC,$22,$3F,$1B
	.dc.b	$E7,$FE,$29,$EF,$23,$C9,$F8,$5F,$3D,$F2,$BE,$1E,$4F,$CE,$8F,$9A
	.dc.b	$61,$F3,$DF,$FB,$D1,$D1,$0C,$C7,$07,$EB,$FB,$3F,$58,$FA,$04,$7D
	.dc.b	$6F,$D3,$FA,$3F,$B6,$48,$FF,$6E,$F1,$BD,$07,$83,$99,$F1,$D5,$E9
	.dc.b	$FE,$B7,$7B,$AF,$D5,$FD,$53,$CC,$A4,$C3,$AF,$D7,$3E,$5F,$11,$F5
	.dc.b	$F0,$5F,$18,$F1,$B5,$D5,$71,$7F,$DE,$3D,$4F,$AE,$FD,$53,$63,$31
	.dc.b	$FC,$0D,$7D,$77,$F9,$59,$FF,$7D,$98,$F5,$7F,$9D,$34,$FE,$79,$1B
	.dc.b	$7F,$A2,$35,$FE,$9E,$D4,$47,$0D,$8E,$B1,$8E,$8F,$D4,$7B,$7F,$AC
	.dc.b	$25,$7D,$2F,$7B,$FD,$FC,$FE,$CA,$32,$FE,$11,$FE,$33,$1A,$8F,$95
	.dc.b	$C7,$78,$7F,$C1,$F6,$FA,$B7,$D5,$7F,$0F,$6D,$FE,$02,$F8,$76,$FF
	.dc.b	$AD,$7E,$13,$57,$2B,$51,$F3,$BE,$52,$47,$D2,$7E,$77,$D7,$74,$8F
	.dc.b	$D9,$07,$0F,$F7,$66,$BF,$20,$D5,$7D,$28,$CC,$7C,$C6,$BA,$3F,$B8
	.dc.b	$38,$37,$97,$F1,$3F,$95,$FE,$2F,$F2,$B2,$BF,$8D,$C7,$F1,$A3,$0F
	.dc.b	$EB,$A3,$0F,$E4,$3D,$3F,$CD,$91,$FC,$9F,$AD,$5F,$CC,$73,$A7,$1C
	.dc.b	$1F,$A4,$8F,$F3,$1C,$FE,$12,$3F,$C8,$3F,$14,$FA,$EF,$D6,$CF,$EC
	.dc.b	$3B,$BF,$DD,$AD,$F3,$3F,$AD,$56,$FB,$FF,$35,$F5,$DF,$DF,$EB,$FC
	.dc.b	$3E,$D7,$E5,$35,$FC,$9B,$53,$49,$F0,$5F,$A0,$78,$70,$7E,$1F,$97
	.dc.b	$F0,$DF,$17,$2F,$D5,$1F,$9E,$9D,$46,$0F,$F2,$97,$47,$F9,$79,$FE
	.dc.b	$5E,$AB,$94,$62,$BF,$B0,$7D,$77,$E1,$F7,$FC,$BB,$63,$B3,$67,$FC
	.dc.b	$BC,$FF,$71,$9D,$F3,$72,$24,$F8,$F1,$56,$EF,$FD,$DF,$F5,$1F,$C1
	.dc.b	$89,$5F,$C0,$BE,$BB,$B2,$7F,$1E,$4C,$3F,$82,$B6,$3F,$C8,$0D,$7F
	.dc.b	$C9,$0F,$E6,$3C,$37,$13,$E8,$7F,$BF,$7B,$BF,$80,$BF,$EA,$6D,$D7
	.dc.b	$F9,$5D,$6E,$7E,$67,$CA,$7F,$92,$BD,$3F,$F2,$73,$63,$FC,$AC,$D7
	.dc.b	$FC,$C2,$D7,$FC,$67,$F3,$7F,$E2,$1D,$BF,$D5,$8D,$8F,$F5,$5B,$7D
	.dc.b	$C7,$D7,$3E,$17,$FC,$29,$F5,$CF,$16,$93,$0F,$C0,$5B,$EB,$B8,$46
	.dc.b	$1A,$FF,$9D,$23,$57,$FE,$67,$6C,$7F,$9B,$DA,$EE,$78,$BF,$C1,$63
	.dc.b	$2D,$8C,$BF,$CC,$23,$2F,$E5,$7E,$7F,$E5,$3F,$E0,$06,$EF,$FC,$91
	.dc.b	$BA,$7F,$B3,$FE,$B9,$EB,$7F,$91,$1B,$7F,$E4,$26,$BF,$E4,$06,$BF
	.dc.b	$E6,$06,$BC,$E3,$FC,$81,$18,$8F,$F6,$CB,$79,$0E,$7A,$58,$3E,$5B
	.dc.b	$FC,$88,$DB,$FC,$E3,$5F,$F0,$DF,$D4,$FF,$65,$CF,$F9,$FB,$FD,$9E
	.dc.b	$3F,$96,$5A,$31,$1F,$B9,$6B,$FE,$C9,$6C,$6F,$D4,$FF,$26,$C6,$08
	.dc.b	$FF,$94,$21,$EB,$7F,$96,$5B,$5F,$B1,$6B,$FD,$F7,$EB,$9F,$29,$FE
	.dc.b	$70,$6C,$7F,$9B,$9A,$FF,$9B,$1A,$FF,$A3,$1A,$FF,$6E,$D5,$18,$84
	.dc.b	$62,$7F,$CE,$13,$98,$C4,$FF,$96,$29,$E5,$FA,$49,$1F,$C2,$65,$DF
	.dc.b	$FE,$75,$6C,$7F,$98,$1A,$FF,$97,$DA,$FD,$06,$BF,$C9,$DA,$FF,$BD
	.dc.b	$DB,$5E,$89,$FF,$3D,$C6,$3E,$5B,$F6,$D2,$69,$FB,$98,$7D,$73,$E5
	.dc.b	$BF,$D3,$6D,$C3,$FD,$32,$D7,$FD,$2E,$D7,$FD,$6E,$D7,$E6,$35,$F9
	.dc.b	$6D,$9E,$F3,$FE,$9B,$8D,$5F,$97,$E0,$BF,$D3,$58,$7F,$B9,$F8,$3F
	.dc.b	$C2,$2D,$E5,$FE,$91,$F5,$DF,$99,$9F,$F0,$AB,$F9,$BC,$FF,$2E,$D7
	.dc.b	$7C,$9F,$E9,$E8,$C4,$31,$73,$FF,$3D,$8C,$36,$33,$F0,$BF,$EC,$38
	.dc.b	$FF,$2D,$FA,$FF,$EA,$F7,$44,$8D,$79,$89,$FF,$5A,$7C,$C7,$DE,$CF
	.dc.b	$FC,$56,$DD,$3F,$B1,$37,$2F,$22,$FA,$EE,$3C,$72,$7F,$AA,$4E,$F4
	.dc.b	$3B,$8F,$F3,$8F,$D5,$BE,$B7,$FD,$AA,$D8,$CC,$7F,$B6,$3E,$B7,$FD
	.dc.b	$8F,$18,$21,$FE,$6B,$9E,$0F,$F9,$B6,$36,$DF,$8A,$FF,$6A,$46,$23
	.dc.b	$FD,$F2,$D7,$F5,$CD,$7F,$CB,$BF,$5C,$F9,$66,$DF,$E6,$0E,$FE,$F2
	.dc.b	$F9,$6F,$F1,$E4,$61,$B9,$FF,$8F,$E7,$95,$71,$8C,$5F,$4C,$3F,$8C
	.dc.b	$46,$EF,$8C,$78,$9A,$AF,$9B,$C0,$F3,$D7,$37,$FC,$9C,$6D,$C3,$E0
	.dc.b	$BC,$BC,$1A,$F5,$F1,$7E,$BB,$AA,$FE,$CD,$AE,$9B,$19,$9D,$BC,$BE
	.dc.b	$47,$FD,$1A,$9A,$47,$FD,$22,$FC,$37,$FC,$C9,$F2,$AE,$3C,$47,$F9
	.dc.b	$9E,$6A,$9D,$3F,$04,$D1,$4D,$4F,$07,$C4,$66,$38,$EB,$F8,$FE,$E6
	.dc.b	$EF,$B9,$FF,$1E,$5F,$E1,$1A,$FF,$30,$8D,$61,$F2,$AB,$8D,$69,$7D
	.dc.b	$8B,$E3,$BC,$FB,$FF,$B2,$A3,$17,$FE,$D2,$C3,$9E,$3F,$EC,$59,$CB
	.dc.b	$6F,$8D,$7F,$43,$9C,$DF,$FB,$52,$F8,$BD,$6B,$77,$FC,$27,$CF,$FF
	.dc.b	$3C,$61,$FE,$47,$6C,$7F,$B4,$FF,$61,$7C,$87,$0E,$85,$6B,$1F,$F5
	.dc.b	$6B,$70,$FF,$32,$37,$0F,$EA,$78,$72,$9E,$DF,$03,$CF,$DE,$7C,$ED
	.dc.b	$3C,$BB,$BF,$62,$FF,$18,$5F,$F9,$6E,$B6,$DA,$E0,$FF,$BE,$27,$78
	.dc.b	$FE,$D2,$3F,$CE,$D5,$C2,$3F,$DD,$DE,$8F,$FB,$F2,$7E,$B3,$63,$F4
	.dc.b	$B8,$7F,$8E,$FC,$1F,$B1,$39,$9F,$F0,$5F,$37,$FB,$BE,$FF,$CD,$CD
	.dc.b	$8E,$B5,$F1,$E3,$1B,$7E,$53,$0F,$E5,$9E,$CF,$E0,$FC,$2F,$F3,$26
	.dc.b	$3F,$87,$6E,$3E,$26,$BF,$EC,$5A,$E3,$3B,$AE,$44,$7F,$73,$C3,$FA
	.dc.b	$3D,$BE,$C7,$47,$C1,$CB,$C5,$97,$67,$FD,$CB,$19,$BF,$F0,$67,$E7
	.dc.b	$DF,$C7,$5F,$CD,$37,$3F,$8A,$D7,$C8,$D7,$8F,$6F,$E7,$E3,$11,$FE
	.dc.b	$14,$6E,$87,$39,$BD,$87,$C4,$FE,$3D,$B8,$FF,$AB,$6F,$86,$39,$AF
	.dc.b	$C3,$FE,$77,$FC,$BE,$BF,$60,$DC,$BE,$DB,$5F,$94,$D7,$FC,$54,$D7
	.dc.b	$FC,$52,$DB,$34,$D6,$3F,$E4,$BE,$5F,$D2,$76,$FE,$7E,$B7,$E0,$E2
	.dc.b	$7F,$37,$E7,$FE,$EB,$1F,$B4,$D7,$FC,$74,$D7,$FC,$72,$DA,$FF,$1A
	.dc.b	$F7,$FF,$09,$8C,$11,$FE,$23,$8B,$E6,$BE,$56,$9F,$EC,$16,$E1,$FA
	.dc.b	$33,$AD,$7E,$EE,$33,$C5,$F9,$5F,$C7,$BF,$BC,$5F,$F1,$96,$DF,$E9
	.dc.b	$1A,$FE,$4D,$AF,$F9,$4D,$AF,$F9,$49,$B9,$1A,$6B,$17,$DC,$BE,$BB
	.dc.b	$F8,$73,$70,$FC,$05,$78,$B1,$FE,$7B,$7C,$2F,$F8,$DE,$3F,$E8,$7F
	.dc.b	$57,$F9,$9B,$70,$FE,$B3,$E8,$7F,$9C,$1B,$3A,$8C,$4F,$F6,$06,$BF
	.dc.b	$D9,$9A,$88,$E1,$FF,$37,$FE,$64,$F9,$4F,$E3,$2D,$BE,$19,$7F,$54
	.dc.b	$AF,$21,$F4,$F5,$56,$E3,$EB,$78,$6F,$9E,$7C,$E7,$F3,$88,$C3,$FC
	.dc.b	$9B,$E0,$9F,$2B,$9A,$DE,$7F,$A7,$F8,$47,$DA,$7F,$51,$6C,$7F,$55
	.dc.b	$6E,$5F,$C8,$1A,$FF,$0F,$CD,$F3,$DF,$D0,$7E,$C5,$F4,$FE,$E8,$C3
	.dc.b	$FD,$33,$C3,$7E,$06,$F9,$6F,$EC,$B1,$86,$C6,$73,$FF,$07,$CE,$5B
	.dc.b	$19,$CB,$FC,$2F,$39,$6C,$65,$FF,$02,$B5,$FA,$DD,$0F,$8E,$FF,$04
	.dc.b	$5F,$D2,$32,$E2,$7F,$49,$7D,$79,$7F,$1B,$39,$7F,$C1,$71,$EC,$F1
	.dc.b	$FF,$B1,$7D,$97,$FB,$57,$0E,$98,$7F,$84,$72,$FF,$34,$46,$32,$56
	.dc.b	$EB,$BD,$69,$FD,$A4,$EB,$A7,$E5,$4E,$72,$FC,$FD,$F1,$7C,$71,$FD
	.dc.b	$79,$FD,$1F,$D9,$7F,$AD,$B1,$F1,$E7,$F4,$BE,$2F,$FB,$C4,$E7,$3F
	.dc.b	$0C,$63,$DD,$EE,$DB,$E3,$8B,$E8,$BC,$67,$EB,$B7,$FE,$F9,$97,$DE
	.dc.b	$8C,$D7,$07,$5E,$2E,$32,$F7,$C7,$F8,$27,$E8,$9F,$2B,$9D,$BF,$77
	.dc.b	$18,$C3,$E5,$FA,$8F,$B1,$7D,$87,$F8,$13,$EC,$7F,$BB,$61,$FE,$3E
	.dc.b	$AD,$4E,$7D,$6F,$F0,$EC,$63,$0F,$E9,$17,$C1,$FD,$D4,$67,$4F,$D9
	.dc.b	$4E,$5F,$F5,$66,$4F,$85,$F4,$1F,$53,$DC,$B8,$7D,$D9,$D3,$67,$4D
	.dc.b	$3E,$5F,$D3,$7E,$80,$8F,$E7,$9B,$7F,$9C,$6B,$F9,$C7,$F9,$ED,$3E
	.dc.b	$5F,$E5,$1F,$86,$8B,$5F,$80,$FC,$59,$F1,$9F,$3F,$5F,$DA,$46,$27
	.dc.b	$D1,$18,$9D,$47,$16,$E3,$FC,$4B,$E1,$3F,$7F,$3F,$DC,$7C,$37,$C8
	.dc.b	$FF,$13,$9C,$D7,$F1,$B9,$CD,$7E,$93,$CB,$FF,$4A,$78,$7F,$C9,$30
	.dc.b	$E0,$D9,$F4,$79,$BF,$C1,$B4,$FE,$45,$8F,$FA,$01,$AF,$F3,$1E,$A4
	.dc.b	$62,$3F,$D0,$0D,$8C,$CF,$32,$57,$F8,$93,$81,$F5,$1F,$93,$3E,$2C
	.dc.b	$D9,$F4,$7E,$11,$FC,$FD,$F8,$1F,$5D,$FC,$C6,$31,$97,$BE,$30,$FF
	.dc.b	$40,$6C,$F0,$3F,$0D,$F3,$BF,$A1,$FA,$CF,$D1,$36,$32,$7E,$2E,$32
	.dc.b	$D3,$FC,$41,$DD,$FB,$9B,$FF,$21,$35,$FE,$A9,$E9,$7E,$DF,$C4,$FF
	.dc.b	$47,$4E,$65,$2F,$CC,$7F,$15,$7F,$E2,$B6,$CF,$F5,$3A,$FE,$0D,$DF
	.dc.b	$FE,$57,$9C,$B7,$AB,$F7,$5F,$3F,$FE,$76,$9C,$E1,$FC,$C3,$85,$39
	.dc.b	$AD,$CF,$76,$E3,$FD,$87,$F3,$1F,$F3,$63,$EA,$6B,$E8,$6E,$67,$3E
	.dc.b	$5B,$1F,$E6,$D8,$78,$4F,$F4,$4C,$FE,$EE,$75,$1F,$E8,$CF,$C5,$DF
	.dc.b	$23,$FE,$B3,$6C,$7F,$AD,$DB,$8F,$FA,$09,$AF,$F8,$04,$53,$CB,$FF
	.dc.b	$0B,$47,$F3,$A6,$ED,$FE,$7F,$AE,$6F,$F3,$6F,$67,$F9,$99,$30,$FF
	.dc.b	$2E,$4E,$9F,$EB,$35,$C9,$B3,$BB,$FF,$04,$56,$DF,$1D,$6D,$FB,$FF
	.dc.b	$6B,$FD,$E2,$FF,$DA,$4D,$8F,$F4,$EB,$5F,$F4,$63,$5F,$0F,$C4,$7F
	.dc.b	$48,$3F,$D6,$71,$C7,$1F,$E9,$F1,$FE,$B6,$F6,$BF,$C7,$A4,$F9,$4F
	.dc.b	$F6,$F3,$70,$FF,$71,$37,$0F,$F5,$6B,$FE,$41,$C5,$F7,$7F,$D5,$8E
	.dc.b	$97,$6F,$FD,$01,$F4,$AE,$9F,$F7,$53,$63,$FD,$DC,$DC,$7F,$D7,$2D
	.dc.b	$7F,$D5,$0D,$7F,$4E,$D7,$F4,$AD,$E3,$39,$E0,$FB,$CF,$F6,$A3,$FD
	.dc.b	$AA,$77,$D7,$FB,$8B,$9B,$FF,$00,$67,$C9,$F4,$E3,$9C,$7F,$CE,$5E
	.dc.b	$2B,$E3,$5F,$3A,$BF,$F2,$77,$C7,$7F,$8C,$2F,$CB,$DB,$95,$19,$C3
	.dc.b	$95,$E6,$38,$6D,$EA,$BC,$7B,$5F,$E5,$9B,$DE,$3D,$A3,$E1,$CE,$B1
	.dc.b	$FE,$93,$97,$92,$B5,$1E,$26,$E1,$D5,$BA,$AF,$96,$3D,$AF,$CF,$8F
	.dc.b	$6B,$E7,$0F,$B1,$E1,$E2,$61,$B1,$C0,$7C,$CB,$E9,$3F,$DE,$DB,$7C
	.dc.b	$59,$E5,$9B,$0E,$4D,$8E,$46,$A3,$D7,$7C,$FC,$DF,$81,$39,$37,$8F
	.dc.b	$89,$FE,$00,$64,$AE,$5F,$E0,$8D,$7F,$D0,$CF,$F4,$ED,$38,$C6,$3C
	.dc.b	$7F,$EE,$AE,$27,$4D,$8C,$BD,$9D,$E5,$EF,$9D,$B6,$33,$AF,$FA,$57
	.dc.b	$1F,$F6,$17,$D3,$7F,$17,$AD,$63,$F3,$6F,$78,$F8,$E3,$E1,$FC,$42
	.dc.b	$F8,$63,$FD,$33,$F2,$67,$BD,$8C,$3F,$A5,$C6,$1F,$C1,$B5,$FB,$18
	.dc.b	$3F,$62,$7E,$0F,$C6,$30,$F3,$E5,$5F,$1D,$7F,$32,$DC,$BF,$31,$87
	.dc.b	$2C,$7F,$27,$8F,$27,$C3,$97,$3B,$E1,$1F,$E2,$2A,$E0,$FE,$2F,$AF
	.dc.b	$F9,$C7,$C4,$FF,$3B,$EC,$FF,$34,$B7,$9F,$F9,$8B,$6B,$DE,$DA,$E3
	.dc.b	$1F,$BE,$8C,$76,$1E,$1B,$7B,$1B,$96,$6B,$F0,$73,$96,$C6,$5F,$C7
	.dc.b	$C6,$5B,$19,$7F,$A9,$46,$5B,$19,$7E,$C4,$FE,$41,$FE,$8B,$9F,$98
	.dc.b	$8F,$FA,$62,$9F,$AF,$3E,$8F,$D2,$D9,$F3,$B5,$F2,$B5,$F1,$B5,$FF
	.dc.b	$09,$B5,$FF,$08,$B6,$7B,$CF,$95,$B3,$FA,$6C,$BF,$B6,$7D,$D7,$CC
	.dc.b	$24,$79,$FF,$31,$7D,$B7,$CE,$6C,$7C,$C6,$BF,$29,$AF,$C8,$6B,$FD
	.dc.b	$BB,$DF,$FB,$24,$7E,$27,$E6,$3F,$CD,$1F,$E7,$71,$97,$FB,$F3,$F7
	.dc.b	$A3,$51,$83,$64,$3F,$DF,$CE,$B8,$7F,$9D,$64,$8E,$33,$F7,$1A,$FF
	.dc.b	$8F,$1A,$FF,$8E,$9B,$97,$F8,$47,$1F,$C8,$92,$7E,$E6,$6F,$8E,$FC
	.dc.b	$9B,$63,$F2,$2D,$7F,$1E,$D7,$F1,$AD,$7F,$C9,$BE,$93,$CC,$FC,$FF
	.dc.b	$C6,$C9,$FF,$2A,$C7,$E8,$27,$51,$FD,$8A,$74,$FF,$93,$67,$44,$78
	.dc.b	$9F,$93,$FC,$C5,$F3,$9F,$AA,$6C,$7E,$A1,$AF,$E9,$9A,$FE,$91,$AF
	.dc.b	$F9,$93,$EC,$5E,$C7,$ED,$BE,$56,$4F,$E8,$67,$2D,$9C,$BF,$E3,$B9
	.dc.b	$FD,$DC,$EA,$1C,$3F,$DC,$FD,$F7,$E2,$2F,$BE,$FE,$03,$C2,$F9,$3A
	.dc.b	$7F,$0B,$AD,$CD,$AE,$18,$7E,$CE,$F8,$36,$F5,$5F,$C3,$BC,$39,$3F
	.dc.b	$E0,$E7,$F8,$B2,$FB,$8F,$8B,$97,$F8,$1E,$31,$BF,$F0,$36,$9F,$C1
	.dc.b	$67,$E3,$A3,$27,$F5,$87,$2E,$0F,$EC,$6E,$0E,$33,$FE,$1B,$F1,$7F
	.dc.b	$94,$79,$0F,$5F,$FC,$09,$7D,$6B,$FC,$35,$39,$F3,$7F,$97,$47,$19
	.dc.b	$D7,$89,$BF,$43,$FA,$A4,$60,$FF,$01,$E1,$D1,$FA,$73,$FD,$86,$31
	.dc.b	$5F,$E2,$73,$A4,$E8,$F7,$7D,$B3,$E4,$FF,$9C,$FE,$15,$27,$B0,$9D
	.dc.b	$38,$6B,$FE,$7F,$8E,$0F,$53,$84,$FF,$44,$FB,$A9,$1F,$E4,$47,$EB
	.dc.b	$FE,$17,$FA,$68,$E0,$D7,$FC,$1E,$D7,$FD,$20,$D7,$EB,$35,$34,$9D
	.dc.b	$7F,$7D,$E1,$E0,$7F,$A5,$FE,$EB,$FA,$B9,$2B,$DF,$2C,$71,$1F,$E9
	.dc.b	$B9,$B9,$FE,$2E,$B5,$3F,$4F,$E4,$1F,$72,$F5,$7F,$B0,$F4,$FF,$9A
	.dc.b	$AB,$FA,$73,$5F,$A5,$5C,$27,$FB,$73,$63,$7E,$C7,$23,$37,$D6,$3E
	.dc.b	$AF,$C9,$BE,$EB,$F6,$BF,$39,$27,$31,$FA,$49,$D4,$60,$87,$D7,$FD
	.dc.b	$6F,$FD,$86,$E3,$5F,$D4,$4F,$08,$FB,$8D,$8F,$D6,$4F,$09,$FC,$07
	.dc.b	$DD,$BC,$8F,$F0,$37,$CE,$7D,$DB,$FF,$3C,$71,$A3,$1D,$3F,$EC,$BB
	.dc.b	$E1,$87,$E4,$1A,$FE,$5E,$B8,$72,$F4,$3F,$7F,$3A,$C3,$F7,$CF,$E8
	.dc.b	$2F,$F7,$54,$F9,$AF,$91,$87,$36,$DF,$0E,$CE,$67,$F8,$76,$33,$7C
	.dc.b	$9B,$5F,$C1,$70,$EC,$FF,$B8,$BE,$3F,$E1,$61,$EE,$6D,$F8,$9B,$3F
	.dc.b	$C3,$7F,$63,$FE,$EB,$5F,$C5,$5F,$EE,$71,$9F,$17,$F8,$B4,$62,$0D
	.dc.b	$24,$68,$91,$FE,$2C,$5F,$2D,$FE,$37,$F8,$27,$C0,$53,$F9,$AC,$7E
	.dc.b	$5B,$B3,$FC,$25,$D6,$69,$30,$FF,$3C,$F9,$4F,$D8,$3F,$89,$C7,$F0
	.dc.b	$B8,$C7,$8D,$FC,$F6,$73,$AF,$F7,$EF,$A1,$75,$FF,$44,$F8,$99,$8F
	.dc.b	$F3,$F7,$F9,$CF,$E6,$2F,$B4,$A7,$F3,$77,$13,$FE,$3B,$6B,$FE,$8C
	.dc.b	$6B,$FE,$96,$6B,$FC,$EC,$38,$7F,$9F,$35,$CC,$7F,$91,$FA,$1F,$4A
	.dc.b	$F9,$5F,$F3,$03,$6B,$F9,$94,$EA,$7F,$CB,$FF,$4B,$FE,$A6,$D3,$FC
	.dc.b	$3A,$FF,$4F,$F6,$CD,$25,$FF,$90,$5A,$FF,$99,$1A,$FF,$AC,$2D,$F7
	.dc.b	$4F,$7F,$FC,$E4,$D9,$F1,$B5,$FE,$7D,$6F,$D1,$1F,$5B,$FE,$74,$6C
	.dc.b	$FF,$9C,$9A,$FF,$9C,$1A,$FF,$9B,$9A,$FF,$65,$9E,$93,$FE,$59,$9C
	.dc.b	$CA,$72,$F5,$9F,$99,$F8,$7B,$E8,$7F,$D1,$4D,$9F,$96,$D7,$D3,$F5
	.dc.b	$AE,$FF,$F4,$5B,$63,$FD,$12,$D7,$FD,$0E,$D7,$FD,$0A,$D7,$D0,$D7
	.dc.b	$CC,$DC,$0D,$35,$A0,$7D,$4F,$FA,$85,$B5,$F8,$26,$BF,$43,$37,$D0
	.dc.b	$7C,$F3,$E3,$12,$3E,$83,$5F,$A0,$D7,$E8,$3F,$93,$FC,$3D,$F4,$BF
	.dc.b	$EB,$16,$CF,$E4,$9A,$FE,$03,$EF,$1F,$37,$FE,$B2,$6C,$7F,$AC,$1A
	.dc.b	$FF,$AB,$9A,$FF,$AB,$1A,$FD,$F6,$BF,$75,$BA,$1A,$6B,$27,$8C,$EA
	.dc.b	$FF,$64,$36,$FF,$52,$D7,$F2,$FF,$85,$BE,$6B,$FD,$92,$D8,$FF,$63
	.dc.b	$B5,$FF,$62,$B5,$FF,$61,$B5,$FE,$83,$EE,$BE,$83,$FD,$B0,$D8,$FD
	.dc.b	$CB,$5F,$D3,$FE,$01,$FC,$D3,$0E,$35,$C0,$93,$FE,$C7,$9D,$3F,$85
	.dc.b	$FB,$BF,$E7,$B9,$9A,$7F,$40,$D5,$F8,$B7,$F1,$8E,$DF,$EC,$73,$AF
	.dc.b	$A1,$FF,$76,$3B,$9D,$67,$FA,$FF,$BB,$7D,$07,$FB,$B1,$B9,$FF,$B6
	.dc.b	$5A,$FF,$00,$FB,$BF,$E2,$CF,$F0,$28,$CE,$5F,$EB,$06,$DF,$5B,$FF
	.dc.b	$7B,$DF,$0A,$FF,$5A,$BC,$AF,$FD,$DE,$D7,$F8,$13,$5F,$E0,$0D,$5F
	.dc.b	$59,$FC,$42,$7F,$CD,$9F,$6B,$FC,$E2,$1F,$6B,$FC,$D6,$30,$DF,$1F
	.dc.b	$F9,$C4,$F1,$B5,$71,$F0,$BF,$D2,$31,$8A,$DD,$F1,$8F,$F3,$D5,$F6
	.dc.b	$7C,$33,$D2,$B8,$DF,$3C,$3F,$A3,$56,$5A,$F2,$9E,$EE,$AF,$C8,$DF
	.dc.b	$BC,$FC,$03,$C2,$11,$88,$D6,$FD,$1F,$8B,$19,$7B,$FD,$0F,$FA,$55
	.dc.b	$34,$8F,$FA,$65,$F5,$BF,$D7,$7E,$93,$F5,$EF,$C3,$3D,$71,$A7,$FD
	.dc.b	$39,$1A,$CB,$FB,$2C,$E6,$53,$9C,$FC,$71,$CC,$E3,$BF,$83,$D0,$7E
	.dc.b	$C8,$FA,$8F,$F5,$75,$7F,$AA,$67,$31,$88,$46,$2B,$FC,$17,$5C,$5F
	.dc.b	$9F,$CC,$E2,$7C,$A7,$FA,$66,$7F,$1E,$DC,$39,$DF,$17,$F4,$11,$9B
	.dc.b	$E4,$DB,$E7,$7F,$EA,$9C,$3C,$57,$C2,$3F,$D5,$96,$FB,$C7,$DE,$F3
	.dc.b	$6C,$E9,$FE,$66,$D8,$D1,$1D,$78,$96,$FF,$B9,$FD,$EB,$EE,$3C,$CD
	.dc.b	$AF,$0A,$FA,$BC,$CC,$9C,$BC,$23,$15,$FE,$AF,$FC,$2B,$F9,$75,$7E
	.dc.b	$35,$B1,$F9,$A6,$BF,$9E,$6B,$FD,$EF,$E5,$3F,$D3,$13,$F0,$9B,$1D
	.dc.b	$7C,$6F,$99,$18,$EB,$FF,$66,$7C,$A7,$2A,$47,$FB,$7E,$BF,$00,$DB
	.dc.b	$FD,$5B,$67,$FC,$4D,$5C,$63,$11,$FE,$37,$6B,$FE,$E7,$8E,$DB,$FF
	.dc.b	$B5,$F9,$3E,$F7,$C9,$FE,$06,$FC,$78,$E2,$7F,$0D,$83,$F6,$5F,$F6
	.dc.b	$F6,$5F,$B1,$6C,$7E,$F9,$AF,$F0,$26,$BF,$E4,$96,$A3,$A6,$1F,$9D
	.dc.b	$6C,$F5,$8F,$D0,$62,$FB,$FF,$F1,$8C,$FE,$39,$B5,$FE,$E7,$BF,$1E
	.dc.b	$5F,$BA,$BE,$1D,$FF,$9F,$87,$EC,$CF,$A2,$F9,$AD,$CB,$E6,$B7,$1F
	.dc.b	$F2,$93,$5F,$98,$D7,$D5,$5D,$2B,$99,$2B,$FC,$5B,$E1,$7E,$AF,$D3
	.dc.b	$FE,$6F,$1C,$0E,$38,$77,$BF,$44,$EB,$0F,$C1,$D7,$F9,$C7,$F4,$17
	.dc.b	$C5,$FC,$76,$C7,$F2,$36,$CF,$F3,$06,$BF,$CD,$1A,$F3,$9E,$44,$9F
	.dc.b	$E1,$3F,$7F,$FB,$E9,$FF,$30,$B7,$0F,$E4,$B0,$BE,$F1,$F3,$E7,$55
	.dc.b	$FC,$67,$DA,$FD,$ED,$FF,$44,$ED,$FF,$38,$F5,$FF,$1B,$6D,$7F,$A9
	.dc.b	$2B,$75,$FE,$8F,$9C,$BC,$63,$2F,$5A,$E5,$18,$8F,$F3,$47,$B9,$FE
	.dc.b	$21,$BF,$E2,$6D,$8F,$F5,$74,$F9,$31,$FF,$46,$FA,$5F,$9E,$F8,$F7
	.dc.b	$D2,$7E,$4D,$B8,$FF,$4F,$6E,$1F,$D7,$1A,$FF,$60,$6B,$FE,$A7,$6B
	.dc.b	$F9,$F6,$BF,$EF,$D9,$FD,$07,$D4,$BE,$AB,$F5,$ED,$BF,$C3,$57,$1A
	.dc.b	$CD,$7F,$35,$FC,$E1,$DF,$FC,$DF,$E1,$BF,$15,$87,$FA,$36,$72,$FF
	.dc.b	$34,$8C,$E1,$FD,$2B,$E1,$BF,$88,$70,$FF,$B3,$3B,$FE,$A4,$60,$FE
	.dc.b	$0F,$F7,$6F,$B1,$FE,$88,$DC,$7F,$9F,$FC,$63,$EA,$73,$5F,$C2,$3E
	.dc.b	$13,$F9,$0E,$7F,$E0,$46,$BF,$EC,$B6,$BF,$C0,$DA,$FF,$A5,$FE,$77
	.dc.b	$F8,$24,$60,$8F,$F6,$3F,$45,$F7,$FF,$E8,$CC,$BE,$3C,$63,$E1,$BC
	.dc.b	$55,$FD,$7E,$31,$3F,$33,$0F,$F5,$26,$79,$C3,$F8,$4C,$F9,$63,$1E
	.dc.b	$3F,$AD,$B1,$C7,$24,$62,$FF,$6E,$7A,$FE,$43,$A7,$9B,$F0,$DB,$7C
	.dc.b	$5B,$3C,$0F,$F5,$B3,$8E,$BF,$D7,$FC,$B3,$EF,$FF,$93,$36,$FF,$C4
	.dc.b	$4D,$BF,$F1,$7B,$5F,$F1,$9B,$5F,$F7,$D3,$5F,$E6,$4D,$4E,$75,$F9
	.dc.b	$BD,$9F,$14,$FE,$87,$1D,$F0,$FF,$13,$CE,$AF,$FC,$09,$F8,$83,$97
	.dc.b	$F7,$CD,$8F,$F2,$1F,$3A,$31,$97,$F0,$56,$BF,$C0,$5A,$FF,$88,$DA
	.dc.b	$FF,$88,$9B,$1C,$EB,$E2,$F6,$7D,$BF,$07,$CE,$DC,$3F,$01,$9F,$EE
	.dc.b	$B3,$F4,$F1,$FF,$03,$9F,$8B,$5B,$CF,$EA,$7E,$55,$FC,$25,$FF,$2E
	.dc.b	$B5,$FE,$B4,$1C,$3F,$D7,$5A,$E7,$C0,$FE,$BD,$18,$23,$F9,$BB,$70
	.dc.b	$FE,$4C,$FE,$E3,$3F,$C6,$71,$F3,$C6,$3A,$7F,$39,$F7,$CF,$B6,$C6
	.dc.b	$9F,$A2,$8C,$3F,$9F,$C3,$F6,$CE,$67,$79,$FE,$63,$89,$7C,$23,$F3
	.dc.b	$79,$3F,$2A,$FF,$1C,$C7,$F9,$FA,$31,$EB,$7F,$85,$A7,$33,$FE,$11
	.dc.b	$6B,$FD,$99,$AF,$F8,$79,$B3,$CB,$1F,$F0,$EC,$6A,$5F,$3C,$F9,$9F
	.dc.b	$F0,$33,$6B,$FD,$04,$D9,$FF,$47,$B5,$FF,$49,$B5,$FA,$ED,$7F,$C2
	.dc.b	$4D,$7B,$CF,$F0,$DC,$5D,$7F,$38,$FF,$11,$D7,$7C,$7F,$43,$3A,$C3
	.dc.b	$E8,$77,$F1,$23,$FE,$A2,$D3,$E6,$C6,$32,$FF,$53,$C6,$27,$55,$FC
	.dc.b	$B6,$72,$D7,$F1,$F0,$FD,$17,$FB,$B4,$61,$AF,$F8,$B6,$35,$EE,$7F
	.dc.b	$A2,$5A,$7F,$AD,$0B,$1F,$EB,$86,$BC,$E3,$84,$F4,$8F,$F4,$8F,$0B
	.dc.b	$C8,$FE,$F4,$DB,$FD,$EB,$67,$FC,$00,$D7,$F4,$FE,$9B,$B5,$F2,$BF
	.dc.b	$E8,$F6,$DF,$FA,$36,$F9,$6D,$FE,$3B,$9C,$E1,$FC,$BB,$07,$DF,$BE
	.dc.b	$33,$FD,$30,$DC,$7F,$D2,$77,$C3,$0F,$F2,$24,$E7,$0E,$4D,$8C,$DF
	.dc.b	$8F,$2E,$6F,$FA,$9F,$BD,$F9,$9B,$F0,$2F,$F0,$D3,$73,$FE,$3A,$DC
	.dc.b	$BF,$C4,$0D,$7E,$0F,$DF,$FF,$A0,$67,$FC,$C5,$AA,$31,$3F,$EB,$EA
	.dc.b	$E5,$9F,$F9,$6E,$73,$8F,$FA,$87,$EC,$7F,$C5,$5E,$31,$A4,$CB,$FD
	.dc.b	$87,$EF,$3E,$3B,$FD,$78,$D8,$FF,$5D,$57,$17,$FD,$73,$19,$CF,$E7
	.dc.b	$37,$2F,$F3,$63,$67,$FD,$AD,$5C,$6B,$77,$D8,$93,$DB,$0B,$3F,$D9
	.dc.b	$BE,$45,$F2,$1F,$DE,$63,$F9,$58,$8C,$F9,$DF,$EE,$16,$DE,$A7,$55
	.dc.b	$C5,$FC,$6F,$CA,$7E,$A7,$4F,$F4,$13,$6B,$FD,$21,$7C,$AB,$F1,$53
	.dc.b	$97,$FD,$29,$F8,$EF,$F2,$67,$B3,$5F,$2B,$5F,$E8,$2D,$78,$F3,$FF
	.dc.b	$A6,$E3,$10,$69,$BE,$27,$A5,$FC,$F7,$1F,$C1,$B6,$B9,$C7,$FB,$B2
	.dc.b	$31,$0F,$9F,$FF,$13,$93,$D2,$FF,$55,$36,$BE,$34,$EE,$72,$48,$F9
	.dc.b	$56,$FA,$A7,$C9,$7F,$C0,$16,$CF,$F8,$9F,$0F,$CA,$D6,$EB,$44,$9D
	.dc.b	$5D,$C3,$FC,$47,$E4,$3E,$1F,$FD,$E9,$5F,$C5,$67,$FD,$85,$39,$F8
	.dc.b	$84,$E6,$7F,$AE,$57,$07,$FD,$23,$5B,$9D,$36,$33,$3B,$9F,$EE,$B3
	.dc.b	$96,$C6,$67,$C2,$F5,$F8,$EF,$F2,$F4,$32,$F8,$9F,$03,$E4,$84,$EB
	.dc.b	$0F,$03,$C2,$B7,$87,$FB,$76,$39,$D3,$EF,$FF,$D7,$3E,$3E,$7C,$8F
	.dc.b	$F0,$4C,$62,$7A,$C7,$E5,$BE,$8D,$B7,$87,$0E,$77,$B6,$E8,$C7,$A6
	.dc.b	$3F,$EE,$DA,$D6,$3C,$F0,$E4,$F8,$2F,$85,$78,$0E,$0E,$1F,$D9,$D7
	.dc.b	$95,$B1,$97,$95,$FE,$3A,$7F,$BB,$69,$7C,$CF,$F0,$A5,$7E,$DE,$72
	.dc.b	$D8,$CB,$D9,$DE,$5E,$E6,$E3,$F1,$9B,$3F,$C1,$7E,$FF,$F4,$4F,$C0
	.dc.b	$3E,$74,$69,$B8,$67,$0F,$0E,$1F,$7B,$A1,$87,$2B,$F8,$56,$FB,$F7
	.dc.b	$DA,$FF,$8F,$B0,$FB,$DF,$7D,$FD,$86,$FE,$7F,$12,$31,$1C,$AF,$8B
	.dc.b	$F5,$30,$E9,$87,$F7,$A8,$FD,$0C,$E9,$FC,$EB,$78,$BF,$32,$D8,$DC
	.dc.b	$F1,$7F,$C8,$11,$96,$C6,$6B,$BC,$7E,$AF,$0F,$37,$C1,$FF,$96,$3B
	.dc.b	$7F,$39,$18,$9F,$F6,$7B,$5F,$F4,$A7,$E7,$7F,$E7,$B7,$39,$FF,$B5
	.dc.b	$DB,$7C,$F3,$FE,$B7,$39,$9E,$3E,$E6,$50,$BE,$8F,$F3,$B6,$BF,$D0
	.dc.b	$9A,$FF,$9B,$9A,$F2,$6B,$CE,$3F,$8F,$B4,$FE,$43,$1C,$30,$7D,$FF
	.dc.b	$FB,$FA,$7D,$ED,$7F,$C1,$8D,$CB,$F8,$04,$3F,$3C,$7D,$AF,$AE,$30
	.dc.b	$F8,$23,$31,$FC,$C1,$AF,$F5,$76,$BF,$D7,$9A,$FF,$A2,$9A,$FC,$66
	.dc.b	$BD,$8F,$E6,$FF,$80,$3E,$B7,$E6,$62,$8C,$4F,$DA,$6B,$FD,$BF,$E9
	.dc.b	$1F,$05,$F4,$F9,$91,$89,$FE,$D9,$18,$9F,$E9,$2D,$8C,$CE,$DF,$EE
	.dc.b	$F6,$FB,$F7,$81,$E4,$6C,$FE,$31,$AF,$FB,$9B,$CE,$7E,$B2,$FA,$AE
	.dc.b	$17,$2B,$83,$F8,$9A,$DC,$E6,$7F,$C8,$57,$D3,$83,$F3,$6D,$E5,$FC
	.dc.b	$7D,$F0,$AF,$C8,$D7,$F3,$B8,$C4,$7F,$84,$1A,$FF,$AF,$1A,$F2,$C7
	.dc.b	$CC,$FF,$68,$F8,$47,$CA,$FD,$26,$C7,$EA,$9A,$FB,$FE,$2D,$E7,$FA
	.dc.b	$DB,$1F,$EE,$46,$BE,$36,$BE,$36,$BF,$E4,$8F,$69,$F6,$5F,$82,$6C
	.dc.b	$7F,$98,$23,$F0,$BF,$80,$3E,$F7,$51,$27,$47,$F0,$09,$D3,$67,$89
	.dc.b	$1F,$93,$B7,$CC,$FC,$3D,$F8,$BF,$D5,$6C,$FF,$BF,$9A,$FC,$D6,$BF
	.dc.b	$35,$AF,$FB,$AB,$AF,$F8,$43,$FE,$42,$6B,$9F,$03,$F7,$B1,$88,$FE
	.dc.b	$16,$D7,$F6,$B1,$A2,$3F,$E2,$28,$7C,$5F,$22,$46,$5A,$FF,$93,$63
	.dc.b	$6E,$48,$7F,$94,$7E,$FF,$E3,$AE,$4F,$E0,$F1,$86,$E7,$FC,$1F,$F2
	.dc.b	$97,$9B,$4F,$E2,$EE,$2F,$FC,$6E,$D7,$FC,$BC,$D7,$FC,$CA,$D7,$F9
	.dc.b	$58,$70,$FF,$2E,$6B,$98,$FE,$4B,$7F,$ED,$C8,$FF,$22,$B5,$F6,$7C
	.dc.b	$46,$88,$E9,$AE,$8F,$3F,$77,$F2,$3F,$C0,$1F,$27,$4F,$81,$12,$7F
	.dc.b	$12,$7F,$42,$9D,$B8,$0E,$1F,$99,$C3,$EE,$CB,$F8,$86,$9F,$CD,$3E
	.dc.b	$FF,$F3,$F5,$A3,$FA,$6B,$88,$D7,$27,$F3,$7D,$7F,$C6,$71,$FC,$F1
	.dc.b	$FF,$1A,$D7,$FA,$41,$AF,$CC,$6B,$FD,$61,$AF,$DA,$F4,$3F,$A7,$C6
	.dc.b	$08,$E3,$93,$FA,$36,$1F,$D0,$AF,$D1,$C3,$E3,$99,$87,$C1,$AD,$67
	.dc.b	$F7,$1B,$1E,$19,$FE,$A1,$F7,$9F,$E1,$78,$FF,$53,$B5,$FC,$1B,$5F
	.dc.b	$ED,$ED,$4D,$27,$AF,$FD,$A1,$C7,$DE,$F6,$24,$FD,$5F,$BD,$FF,$65
	.dc.b	$37,$E0,$7F,$B7,$C6,$23,$FD,$68,$D7,$FD,$76,$D7,$F2,$ED,$7F,$C0
	.dc.b	$6D,$4D,$27,$A1,$FD,$BF,$EF,$5D,$B4,$FE,$E5,$3F,$E2,$9D,$FF,$BB
	.dc.b	$B7,$6F,$D2,$D7,$0C,$3F,$1C,$D7,$F1,$07,$83,$0E,$7D,$7D,$F8,$A4
	.dc.b	$FC,$18,$CB,$CA,$32,$F2,$C3,$FC,$41,$8F,$F8,$23,$EF,$7F,$B8,$B7
	.dc.b	$CC,$FF,$04,$46,$33,$FD,$CF,$EC,$9F,$EB,$0C,$3F,$44,$D7,$FD,$BC
	.dc.b	$D7,$FD,$CA,$D7,$8F,$BB,$FE,$20,$8C,$47,$F8,$B9,$A9,$A6,$BC,$BF
	.dc.b	$F0,$57,$CA,$FF,$82,$B1,$34,$FF,$06,$FC,$D3,$EF,$BF,$B9,$7D,$F7
	.dc.b	$E3,$7C,$C3,$49,$D1,$FE,$D0,$C5,$F3,$CF,$A7,$FF,$0B,$FC,$FF,$F9
	.dc.b	$03,$8B,$FC,$A2,$DE,$1F,$F2,$2D,$F0,$AF,$F1,$77,$DF,$3E,$C2,$93
	.dc.b	$8B,$F8,$44,$3F,$01,$D9,$FE,$42,$71,$1F,$EE,$26,$BF,$CA,$0F,$F2
	.dc.b	$9B,$87,$0D,$7F,$96,$B5,$DC,$75,$3D,$DF,$4B,$CF,$3F,$F3,$5B,$72
	.dc.b	$F5,$B6,$9B,$BE,$F7,$FC,$0B,$7F,$E1,$37,$FC,$0F,$3F,$E1,$16,$BF
	.dc.b	$E1,$A6,$BC,$9B,$3C,$FD,$3F,$F7,$FB,$67,$DC,$6E,$33,$3C,$B5,$7D
	.dc.b	$EF,$D5,$AF,$F4,$0B,$5F,$E5,$91,$AD,$BF,$CC,$1F,$29,$F8,$39,$FE
	.dc.b	$70,$D7,$BB,$5F,$9C,$D7,$FA,$BB,$5F,$F4,$1B,$53,$49,$C9,$F8,$BF
	.dc.b	$AA,$FE,$AD,$A7,$FA,$61,$BB,$7E,$21,$B7,$FD,$CF,$E3,$BF,$8F,$BF
	.dc.b	$D4,$29,$18,$8F,$B0,$D7,$FD,$48,$D7,$FC,$8E,$D7,$A7,$8B,$F3,$1A
	.dc.b	$7C,$E9,$EB,$19,$6C,$7D,$2C,$FE,$7F,$DE,$FF,$9E,$35,$34,$FF,$46
	.dc.b	$75,$3E,$B6,$9F,$E9,$C7,$17,$E1,$6B,$F7,$9A,$FE,$75,$AF,$F8,$19
	.dc.b	$AF,$FA,$C9,$AF,$F0,$08,$FD,$C7,$DE,$BE,$53,$FD,$1A,$D9,$FF,$62
	.dc.b	$B5,$FE,$E1,$93,$F5,$8F,$E8,$4B,$7F,$F3,$7C,$7F,$3F,$8D,$79,$3F
	.dc.b	$16,$30,$47,$04,$AF,$99,$1F,$D1,$F5,$7D,$F7,$F8,$D2,$7F,$2A,$DB
	.dc.b	$FF,$6D,$B7,$0F,$F1,$2F,$BF,$FC,$02,$7F,$D7,$6F,$EE,$EB,$FD,$31
	.dc.b	$D5,$FE,$C8,$8C,$1F,$BF,$C1,$F7,$DF,$CD,$E9,$18,$F0,$FF,$D6,$33
	.dc.b	$98,$DF,$E0,$D8,$99,$5F,$EE,$CA,$D0,$9F,$F5,$CF,$DF,$FF,$1D,$9F
	.dc.b	$F7,$8B,$5F,$D8,$B7,$2F,$F7,$BB,$ED,$8F,$1B,$9C,$FF,$94,$B6,$33
	.dc.b	$E0,$76,$F7,$3F,$D8,$FF,$72,$FA,$2A,$4A,$FF,$1A,$F8,$2F,$BA,$7D
	.dc.b	$3E,$67,$F8,$1D,$F1,$7F,$C7,$7C,$08,$C4,$7F,$AF,$6B,$8B,$D1,$AF
	.dc.b	$FA,$BD,$AF,$FA,$BD,$AF,$FA,$E1,$A8,$7D,$F7,$F4,$BD,$B1,$F7,$9F
	.dc.b	$1E,$3F,$46,$DE,$AF,$F3,$2B,$9F,$0F,$FC,$ED,$68,$C5,$F2,$DF,$FC
	.dc.b	$D7,$87,$26,$E1,$C9,$CE,$1C,$AF,$8E,$1F,$CB,$AF,$C8,$D9,$EE,$FF
	.dc.b	$4F,$8C,$B7,$EF,$5F,$7D,$FC,$2E,$51,$89,$E8,$D7,$84,$FF,$11,$7E
	.dc.b	$27,$CB,$D4,$D2,$3B,$7D,$CF,$FA,$23,$E6,$FF,$6C,$FC,$3C,$BF,$A3
	.dc.b	$CE,$A3,$E8,$4E,$B2,$FC,$A4,$78,$0E,$07,$3E,$5E,$0E,$63,$73,$C6
	.dc.b	$FE,$E7,$DE,$78,$2B,$F0,$0D,$73,$7F,$D6,$38,$FF,$25,$39,$9F,$D8
	.dc.b	$F0,$FE,$52,$BF,$22,$D8,$FC,$84,$FB,$38,$3D,$BC,$5F,$95,$E3,$C6
	.dc.b	$1F,$E0,$EC,$38,$56,$9B,$1E,$2C,$3F,$3F,$5C,$1F,$E0,$3F,$7C,$FB
	.dc.b	$1C,$62,$8C,$5F,$E9,$17,$C2,$E9,$CD,$FF,$6C,$6E,$5E,$66,$D7,$F8
	.dc.b	$89,$E2,$F3,$C3,$BF,$DF,$3E,$A2,$98,$F9,$3B,$5F,$B4,$9D,$36,$74
	.dc.b	$D3,$FD,$61,$F7,$1F,$D9,$1F,$EB,$8D,$CB,$FC,$12,$D7,$FC,$50,$D7
	.dc.b	$FC,$6F,$F4,$8F,$A3,$F8,$6D,$AE,$BE,$76,$27,$E5,$6A,$FA,$AE,$58
	.dc.b	$7F,$BA,$5B,$19,$AF,$C0,$36,$FF,$B7,$4E,$89,$3A,$79,$F0,$7F,$B4
	.dc.b	$B8,$FE,$AF,$C6,$9F,$0C,$FB,$70,$FC,$2C,$E7,$0F,$F6,$97,$DE,$FF
	.dc.b	$B6,$9F,$EF,$CD,$CB,$FC,$56,$D7,$FC,$94,$D7,$FD,$E6,$D7,$F8,$C3
	.dc.b	$53,$4D,$62,$FB,$EF,$F3,$35,$7F,$2C,$6D,$7F,$A5,$AB,$C5,$87,$E7
	.dc.b	$27,$2F,$E7,$FE,$F9,$EC,$7C,$86,$E1,$FE,$15,$6E,$1F,$E5,$76,$BF
	.dc.b	$CB,$45,$23,$31,$FC,$C6,$31,$5D,$49,$5F,$E1,$DF,$57,$FC,$30,$FF
	.dc.b	$44,$6D,$7F,$81,$AF,$C3,$7F,$E3,$9F,$0F,$F5,$92,$FB,$E7,$C2,$FE
	.dc.b	$FE,$30,$D8,$CE,$3F,$E3,$16,$CF,$F9,$79,$AF,$F9,$F5,$AF,$B5,$AF
	.dc.b	$F4,$B6,$EA,$FB,$EF,$E0,$33,$FD,$F9,$AF,$8A,$BF,$AF,$C6,$BB,$9F
	.dc.b	$7D,$FB,$6A,$FF,$46,$46,$3B,$7F,$8C,$36,$FF,$8E,$56,$EB,$E8,$4E
	.dc.b	$5E,$31,$97,$FB,$04,$65,$DC,$65,$FF,$09,$FB,$AF,$2F,$F8,$9B,$6F
	.dc.b	$FB,$DD,$79,$76,$FF,$4A,$D7,$D0,$F8,$B7,$CD,$FC,$F6,$DF,$F9,$75
	.dc.b	$B8,$FF,$A2,$9A,$FF,$AA,$5A,$FE,$25,$AF,$F7,$76,$BF,$EF,$89,$FB
	.dc.b	$9F,$6A,$F8,$BF,$C0,$B6,$FF,$1C,$73,$9C,$CF,$F2,$EF,$84,$77,$FF
	.dc.b	$30,$F8,$4F,$C5,$DF,$FB,$22,$72,$FF,$85,$23,$38,$FF,$AC,$3E,$97
	.dc.b	$F8,$7F,$01,$A4,$EF,$FE,$9D,$C8,$F9,$C7,$CA,$7F,$3E,$6E,$5F,$E2
	.dc.b	$0F,$8A,$7D,$7E,$6B,$F3,$7F,$33,$FC,$87,$4F,$F5,$8B,$5F,$F6,$83
	.dc.b	$5F,$D8,$35,$FF,$0F,$35,$34,$9E,$2F,$F6,$3F,$3D,$F7,$6F,$9F,$FE
	.dc.b	$AD,$18,$7C,$73,$89,$D6,$7F,$3B,$0E,$F9,$E7,$C6,$F4,$B6,$75,$3E
	.dc.b	$28,$C3,$FC,$1A,$3D,$31,$8E,$AF,$EF,$99,$BE,$17,$C1,$E5,$FC,$46
	.dc.b	$DF,$49,$D3,$E3,$D7,$D7,$5B,$8C,$67,$FD,$6B,$2F,$01,$FE,$43,$FB
	.dc.b	$C7,$DF,$7F,$07,$6D,$7F,$AD,$5B,$3F,$EE,$26,$BF,$F0,$05,$AF,$F2
	.dc.b	$16,$BD,$67,$A1,$F3,$E9,$DB,$F2,$CF,$E3,$35,$DB,$4F,$F1,$5E,$5F
	.dc.b	$52,$7F,$C1,$1C,$EF,$BE,$FC,$86,$7F,$E1,$D6,$E7,$FC,$21,$AF,$F0
	.dc.b	$46,$BF,$C0,$9A,$A7,$A7,$47,$F0,$E9,$CC,$62,$BE,$17,$6F,$E0,$7C
	.dc.b	$2F,$3B,$63,$EE,$E9,$FB,$ED,$3D,$3A,$FF,$07,$9F,$89,$5B,$CF,$E8
	.dc.b	$8F,$BE,$FE,$1C,$FF,$49,$3F,$CE,$CE,$1C,$35,$FE,$9C,$D7,$7E,$0F
	.dc.b	$F3,$F8,$C4,$75,$8C,$1F,$47,$8F,$F3,$47,$F4,$89,$F0,$E3,$E8,$8C
	.dc.b	$74,$FE,$8B,$EF,$1F,$5D,$8C,$FF,$1D,$18,$7F,$3B,$87,$EE,$9C,$CE
	.dc.b	$F2,$FC,$7F,$1A,$F8,$47,$FB,$83,$EF,$BF,$C7,$6F,$F6,$9C,$D1,$88
	.dc.b	$FE,$D0,$D7,$FB,$6B,$63,$7E,$AF,$F8,$7A,$30,$47,$FB,$17,$DF,$3E
	.dc.b	$D2,$93,$3F,$A1,$87,$FA,$96,$75,$18,$84,$61,$DB,$5F,$B0,$D7,$FC
	.dc.b	$04,$D5,$97,$C1,$EA,$F9,$87,$F7,$89,$E9,$97,$E8,$33,$FE,$2F,$87
	.dc.b	$CE,$EF,$FE,$2D,$1F,$CA,$31,$FC,$36,$3F,$2E,$72,$4A,$F5,$C6,$5B
	.dc.b	$5C,$6B,$FD,$30,$D7,$FD,$89,$0F,$BE,$FF,$41,$47,$F8,$6E,$31,$ED
	.dc.b	$7F,$A2,$5A,$7C,$82,$C7,$E6,$5A,$F1,$9E,$6F,$66,$A8,$DB,$FD,$DB
	.dc.b	$85,$E4,$FF,$80,$1B,$3F,$E9,$B6,$CF,$F8,$19,$AF,$F8,$C7,$C4,$FE
	.dc.b	$1B,$F0,$9F,$E8,$CE,$EF,$F7,$B5,$FF,$BD,$2F,$8E,$DF,$E9,$C9,$CE
	.dc.b	$1F,$C9,$30,$7D,$F3,$EF,$BF,$D2,$AD,$C7,$FD,$EF,$7C,$30,$FF,$52
	.dc.b	$CE,$70,$E4,$D8,$CD,$F8,$F2,$E6,$FF,$1C,$FB,$D7,$DF,$7F,$87,$9B
	.dc.b	$97,$FA,$E5,$B8,$FF,$89,$1A,$FB,$FE,$63,$FB,$36,$3F,$EB,$66,$D7
	.dc.b	$F8,$82,$B8,$CE,$7C,$8F,$E5,$7E,$33,$EF,$BF,$C5,$AD,$BF,$F6,$43
	.dc.b	$71,$FF,$19,$B5,$F9,$DE,$E3,$E4,$FF,$D7,$2D,$AF,$EC,$9E,$37,$FA
	.dc.b	$B7,$0C,$D4,$9F,$ED,$1C,$5F,$E9,$98,$F1,$CE,$70,$FA,$58,$73,$6C
	.dc.b	$EA,$F8,$CE,$73,$FF,$78,$FD,$93,$EF,$BF,$BA,$E1,$FC,$AE,$31,$E7
	.dc.b	$7F,$B8,$5B,$87,$F0,$06,$CF,$F3,$1C,$DF,$2F,$FA,$FC,$7F,$DC,$2D
	.dc.b	$B5,$2F,$8D,$6B,$B7,$FC,$E7,$F7,$DF,$C8,$EB,$F8,$6B,$5F,$E7,$CD
	.dc.b	$7F,$D1,$0D,$7A,$B5,$34,$9D,$3F,$E2,$EF,$4B,$F9,$CE,$3E,$4C,$D1
	.dc.b	$8B,$E7,$5F,$EE,$E8,$C4,$3E,$7F,$FC,$4C,$4E,$4F,$F4,$43,$70,$FE
	.dc.b	$44,$D9,$FE,$4E,$FF,$8B,$F0,$7D,$F3,$E8,$FF,$DF,$6D,$BF,$F1,$0D
	.dc.b	$7E,$52,$D7,$C2,$F6,$4A,$DE,$16,$BF,$C2,$FE,$43,$E1,$9F,$11,$FC
	.dc.b	$26,$61,$FD,$5A,$33,$7F,$C7,$23,$EB,$D7,$1A,$E1,$3F,$EB,$78,$C3
	.dc.b	$B9,$E5,$3A,$78,$D7,$47,$BD,$70,$6C,$66,$BA,$3B,$F8,$DF,$F3,$0C
	.dc.b	$FC,$0D,$3F,$D5,$F1,$8B,$E4,$EB,$3F,$97,$F0,$3E,$D8,$E3,$6F,$BD
	.dc.b	$FF,$5E,$F9,$1F,$92,$9C,$C6,$25,$1E,$0E,$57,$DF,$7F,$9A,$E3,$78
	.dc.b	$7F,$54,$AD,$F2,$FF,$B6,$70,$E5,$5A,$C7,$B4,$FE,$83,$1E,$7C,$1E
	.dc.b	$58,$E1,$8F,$3B,$E1,$8F,$F6,$CC,$7C,$8F,$F8,$75,$AF,$0C,$7F,$B7
	.dc.b	$47,$F7,$5D,$2F,$99,$FE,$10,$AE,$4D,$79,$35,$F4,$3B,$BF,$86,$DC
	.dc.b	$73,$5E,$AA,$FD,$C7,$DF,$7E,$85,$C1,$F3,$6B,$83,$FB,$EC,$3C,$8F
	.dc.b	$C6,$31,$1A,$D1,$5C,$E7,$56,$FB,$E7,$D5,$7F,$08,$9E,$5C,$0F,$BE
	.dc.b	$FE,$BF,$1E,$58,$C7,$97,$BD,$BF,$0D,$87,$2A,$F5,$63,$D7,$1F,$EE
	.dc.b	$D1,$F9,$E9,$D3,$F9,$66,$F2,$FE,$7D,$B1,$B9,$E2,$FF,$8F,$A3,$2D
	.dc.b	$8C,$D7,$78,$FD,$36,$3E,$9F,$83,$FF,$1C,$61,$FC,$47,$3F,$F6,$A5
	.dc.b	$7F,$3E,$6B,$FE,$60,$FB,$EF,$F6,$43,$9A,$E2,$FF,$04,$9D,$F7,$7E
	.dc.b	$F7,$0E,$D3,$FE,$AA,$BE,$2F,$3F,$77,$FC,$6C,$A3,$2F,$F9,$AD,$AF
	.dc.b	$D1,$6B,$FC,$E9,$AF,$18,$E7,$18,$37,$2F,$BE,$FF,$7F,$D7,$F5,$A6
	.dc.b	$BF,$E0,$96,$DF,$F1,$4F,$BE,$7D,$9F,$BA,$91,$87,$FC,$9C,$D7,$FC
	.dc.b	$A6,$D7,$FD,$08,$D7,$F1,$4D,$7F,$A9,$B5,$EB,$1F,$12,$35,$6F,$BE
	.dc.b	$7D,$0F,$D0,$8C,$36,$33,$7F,$ED,$DE,$57,$DF,$3E,$0B,$EB,$C6,$1F
	.dc.b	$F3,$33,$67,$FB,$64,$ED,$B3,$FE,$6B,$C1,$F4,$7F,$CE,$A7,$FC,$06
	.dc.b	$D7,$FC,$64,$DD,$3F,$9F,$79,$CF,$BD,$7D,$F7,$A6,$B5,$CB,$FD,$85
	.dc.b	$B8,$7F,$4E,$6D,$F1,$AE,$1E,$2C,$AF,$0B,$F9,$29,$D4,$7F,$86,$B0
	.dc.b	$EE,$FF,$A1,$A3,$2F,$FA,$DD,$AF,$EF,$5A,$FF,$73,$F8,$37,$CA,$FC
	.dc.b	$F6,$C7,$F8,$89,$AF,$FB,$D2,$5F,$16,$F3,$FC,$8D,$9F,$13,$5F,$13
	.dc.b	$5F,$13,$5F,$F0,$27,$B6,$FB,$8F,$C0,$36,$3F,$CB,$71,$F8,$3F,$BE
	.dc.b	$7D,$EE,$A2,$4E,$8F,$F1,$CC,$E9,$B3,$C4,$8F,$F9,$0B,$EF,$9F,$75
	.dc.b	$F3,$5B,$1F,$2D,$AF,$CB,$6B,$F2,$DA,$FF,$29,$EB,$FE,$16,$FF,$B9
	.dc.b	$DA,$E7,$C0,$FE,$07,$18,$8F,$F1,$BB,$5F,$D9,$C6,$88,$FF,$AF,$E1
	.dc.b	$F1,$5C,$89,$19,$6B,$FE,$C8,$8D,$B9,$21,$FE,$CC,$FB,$EF,$8C,$FF
	.dc.b	$C0,$06,$F3,$FF,$86,$1B,$CF,$FD,$17,$EF,$5E,$77,$F0,$76,$DF,$FB
	.dc.b	$4D,$AF,$FB,$61,$AF,$F4,$D0,$E1,$E9,$1F,$C9,$A3,$11,$FE,$5A,$6B
	.dc.b	$C2,$3F,$95,$D3,$E2,$29,$27,$FD,$C9,$19,$6C,$65,$F8,$0D,$7C,$95
	.dc.b	$FC,$0D,$61,$FE,$38,$FB,$E7,$C9,$FF,$25,$6C,$FF,$9B,$A7,$FC,$7B
	.dc.b	$5C,$5B,$DD,$BB,$95,$BC,$FE,$54,$FF,$2F,$FB,$EF,$CF,$4F,$82,$7F
	.dc.b	$B9,$46,$2F,$FC,$92,$D8,$FE,$ED,$A7,$F8,$C2,$7F,$C9,$AF,$F8,$CA
	.dc.b	$BE,$73,$5D,$FA,$3F,$D4,$A3,$11,$FE,$86,$6B,$F7,$23,$44,$7E,$C1
	.dc.b	$BE,$2F,$46,$F2,$3D,$86,$4A,$7E,$1A,$34,$FB,$4F,$F4,$AE,$3F,$D2
	.dc.b	$F0,$F4,$CF,$F5,$1F,$B5,$FF,$0B,$57,$E1,$DA,$FF,$85,$5A,$FF,$A7
	.dc.b	$9A,$9A,$4F,$67,$FB,$8B,$8F,$BE,$EC,$49,$F3,$FD,$EF,$FB,$1D,$BF
	.dc.b	$05,$FD,$E2,$31,$1F,$09,$AF,$E6,$DA,$FF,$8A,$5A,$FF,$AC,$9A,$9A
	.dc.b	$4F,$43,$FB,$87,$DF,$3B,$69,$FD,$CE,$7F,$C4,$FA,$7F,$A2,$9B,$C3
	.dc.b	$FE,$37,$AE,$18,$7D,$06,$BF,$3C,$F0,$61,$CF,$A7,$BF,$14,$9F,$83
	.dc.b	$19,$79,$46,$5E,$58,$7F,$88,$30,$FF,$23,$7D,$F7,$F7,$29,$EC,$DC
	.dc.b	$BF,$D3,$ED,$9F,$F2,$67,$DF,$7F,$AC,$6B,$F0,$AD,$7F,$1A,$D7,$F8
	.dc.b	$33,$5E,$3E,$FF,$F8,$AA,$31,$1F,$ED,$E6,$A6,$9A,$F2,$BF,$C1,$7F
	.dc.b	$2B,$FE,$0B,$F3,$CD,$3F,$C1,$DF,$34,$FB,$BF,$ED,$DF,$7D,$F8,$EF
	.dc.b	$24,$D2,$74,$7F,$81,$FE,$79,$F6,$5F,$E1,$7F,$9F,$FF,$20,$F1,$7F
	.dc.b	$91,$5B,$C3,$FE,$E1,$AD,$F0,$7E,$9B,$EF,$9F,$5F,$FE,$31,$6C,$66
	.dc.b	$31,$F7,$12,$3F,$66,$25,$1F,$E4,$E6,$C6,$9A,$FE,$E9,$AF,$F3,$36
	.dc.b	$BF,$E8,$16,$A3,$B4,$7F,$89,$7E,$97,$F9,$D6,$5F,$E6,$16,$E3,$FE
	.dc.b	$FD,$6D,$36,$7D,$F7,$F8,$2A,$FF,$C4,$EF,$F8,$32,$7F,$C2,$8D,$7C
	.dc.b	$13,$C3,$D3,$F7,$B6,$3F,$C3,$D3,$B7,$B3,$63,$51,$FE,$21,$8D,$72
	.dc.b	$BE,$FB,$E8,$CF,$F1,$38,$C7,$07,$F9,$3E,$73,$8F,$FB,$E3,$E5,$BF
	.dc.b	$09,$7F,$C7,$1A,$FF,$24,$6B,$FD,$75,$AF,$FA,$AD,$AF,$CB,$6A,$69
	.dc.b	$39,$3F,$CA,$5F,$59,$FD,$66,$7C,$CD,$AF,$F4,$83,$76,$FE,$A5,$F1
	.dc.b	$FF,$C8,$6B,$F9,$9D,$23,$11,$F6,$1A,$FF,$AF,$E3,$5E,$2F,$E3,$1B
	.dc.b	$1F,$D0,$1A,$F1,$8D,$4F,$28,$F0,$C7,$F9,$A7,$27,$DF,$7C,$B2,$4F
	.dc.b	$F4,$4E,$A7,$D6,$7F,$A3,$DB,$1F,$9C,$6B,$FD,$41,$AF,$F8,$25,$AF
	.dc.b	$FB,$5D,$AF,$E5,$DA,$FE,$FA,$3F,$D6,$31,$AD,$DF,$7C,$F9,$5F,$F4
	.dc.b	$73,$6F,$FA,$DC,$EA,$7F,$D8,$3A,$3E,$FB,$FA,$1A,$DB,$FC,$05,$1F
	.dc.b	$D0,$23,$5E,$4F,$CB,$8C,$11,$C1,$30,$F9,$D1,$FD,$23,$07,$DF,$7F
	.dc.b	$8E,$27,$F2,$CD,$BF,$F6,$83,$70,$FF,$7C,$7C,$0F,$EF,$A7,$C9,$5F
	.dc.b	$B9,$8F,$DD,$BF,$E9,$CE,$AF,$F6,$9C,$60,$FD,$FE,$0F,$BE,$FE,$71
	.dc.b	$48,$C7,$87,$FE,$B3,$9C,$C6,$FE,$FB,$13,$2B,$E4,$D6,$84,$FF,$AE
	.dc.b	$FE,$FB,$F8,$EC,$FE,$C5,$B9,$7F,$BA,$DB,$9F,$F1,$1C,$51,$EF,$AF
	.dc.b	$23,$A6,$F8,$1D,$4E,$DE,$D7,$FB,$27,$EE,$5F,$47,$49,$5F,$E1,$8F
	.dc.b	$01,$F7,$CF,$A7,$CC,$FF,$05,$BE,$2F,$E1,$1A,$FF,$A8,$E7,$83,$CD
	.dc.b	$AF,$FA,$C5,$AF,$FA,$C5,$AF,$F1,$E6,$BF,$CB,$FE,$FB,$FA,$5E,$DF
	.dc.b	$CC,$7E,$F3,$E3,$C7,$E9,$1B,$D7,$FC,$D6,$91,$AF,$0F,$FC,$B3,$18
	.dc.b	$DB,$57,$D2,$30,$77,$BE,$57,$D2,$3C,$15,$C2,$F9,$CF,$5C,$3F,$A8
	.dc.b	$E1,$CE,$FF,$CF,$8D,$78,$46,$BE,$F9,$F7,$D4,$93,$FE,$4C,$A4,$EA
	.dc.b	$3F,$AE,$8F,$97,$AD,$32,$E0,$8F,$B0,$FE,$B1,$F3,$9F,$A2,$C7,$F1
	.dc.b	$F1,$FD,$6E,$7D,$AF,$C9,$9D,$67,$FE,$5E,$9C,$CF,$2E,$5A,$F1,$3B
	.dc.b	$0A,$E2,$FE,$9F,$F3,$97,$DA,$FE,$CA,$3E,$B7,$3A,$73,$7F,$E6,$C8
	.dc.b	$C5,$6E,$BF,$41,$5A,$7F,$D5,$12,$FC,$4B,$FD,$CD,$7F,$C5,$9B,$3F
	.dc.b	$D6,$EB,$86,$FF,$C0,$AF,$87,$C5,$FF,$04,$8C,$5F,$06,$C6,$70,$F0
	.dc.b	$E3,$71,$FE,$05,$7B,$9F,$F0,$27,$D3,$7F,$7D,$F7,$F4,$D8,$FF,$41
	.dc.b	$BF,$E0,$1E,$4C,$F4,$2A,$92,$B7,$1F,$D1,$1B,$1D,$3F,$05,$7D,$4D
	.dc.b	$31,$E4,$DA,$EF,$3A,$6C,$E9,$A7,$90,$FF,$43,$1F,$E9,$CC,$7F,$9D
	.dc.b	$79,$7F,$47,$44,$62,$74,$70,$8F,$F6,$D4,$7E,$57,$E1,$3E,$0F,$21
	.dc.b	$F8,$75,$BE,$1F,$E6,$75,$B8,$F1,$7D,$1F,$78,$FE,$94,$D8,$34,$8F
	.dc.b	$89,$B8,$78,$A3,$67,$AF,$0F,$E3,$EF,$D2,$6C,$7F,$AC,$6B,$BC,$E7
	.dc.b	$BF,$EC,$F2,$3E,$FF,$C8,$D9,$F8,$CD,$BF,$82,$D7,$F8,$95,$FF,$47
	.dc.b	$AE,$24,$AE,$2F,$BF,$F1,$27,$97,$F8,$86,$CF,$E2,$30,$ED,$A7,$C8
	.dc.b	$9C,$E3,$E9,$97,$C1,$79,$89,$7E,$66,$CF,$99,$AF,$9A,$78,$CF,$E0
	.dc.b	$7F,$1E,$7C,$AF,$E6,$9B,$1E,$3B,$C4,$FE,$7B,$EF,$3F,$CD,$F3,$F8
	.dc.b	$96,$E1,$F8,$46,$BF,$11,$AF,$EA,$1A,$FE,$39,$BA,$9A,$63,$E9,$5F
	.dc.b	$23,$FA,$E6,$DF,$EB,$A7,$C3,$AF,$E3,$A7,$38,$7D,$F9,$79,$3F,$8F
	.dc.b	$6D,$FA,$63,$04,$70,$49,$F6,$47,$E5,$FF,$0E,$7D,$BF,$F0,$06,$D7
	.dc.b	$8E,$71,$3F,$C1,$3F,$17,$FD,$45,$FE,$19,$AF,$F0,$A6,$BF,$B0,$6E
	.dc.b	$86,$93,$A3,$F5,$0E,$BF,$14,$FE,$19,$8F,$F1,$06,$DF,$E7,$E7,$F6
	.dc.b	$53,$FC,$62,$73,$AF,$F8,$9B,$E8,$3F,$C5,$FF,$3D,$FC,$6B,$F2,$2F
	.dc.b	$8B,$8A,$B5,$EB,$AB,$5B,$7B,$7E,$DB,$C6,$4C,$FF,$86,$68,$8F,$DF
	.dc.b	$7E,$2A,$F8,$4F,$E4,$ED,$AF,$47,$CB,$62,$7F,$98,$43,$F0,$97,$DC
	.dc.b	$7F,$15,$8C,$36,$33,$1F,$E0,$3D,$BF,$D7,$0E,$8C,$3F,$E5,$78,$DB
	.dc.b	$B8,$D3,$63,$2F,$F8,$36,$32,$FE,$FB,$F0,$07,$37,$F4,$06,$CF,$FB
	.dc.b	$1F,$83,$FC,$A2,$F4,$7F,$90,$C6,$AB,$F8,$97,$DE,$FE,$C9,$BD,$9F
	.dc.b	$D2,$A1,$F6,$96,$26,$1F,$D3,$27,$86,$1F,$CB,$1D,$C6,$67,$FA,$9F
	.dc.b	$E2,$1F,$6A,$BF,$98,$BF,$EE,$E9,$FE,$AA,$E5,$FE,$B6,$E7,$0F,$E4
	.dc.b	$CE,$5F,$E8,$6E,$67,$9F,$17,$6F,$4F,$FB,$35,$7E,$DB,$E1,$7F,$B4
	.dc.b	$7A,$8F,$C0,$5F,$77,$AF,$09,$5A,$9C,$E9,$F2,$4F,$F1,$E4,$E9,$C6
	.dc.b	$1F,$E0,$F9,$DB,$F9,$58,$DD,$73,$B9,$1E,$EF,$26,$D7,$B1,$F8,$BD
	.dc.b	$DF,$1B,$C8,$B5,$F1,$1B,$5F,$09,$BB,$7C,$A7,$E6,$C4,$9F,$97,$A5
	.dc.b	$8F,$92,$5A,$FA,$31,$65,$2F,$C5,$EF,$DA,$FF,$71,$6C,$FE,$61,$B3
	.dc.b	$F9,$D6,$BF,$E6,$16,$BD,$1A,$F5,$13,$D5,$B3,$FE,$01,$D9,$F8,$97
	.dc.b	$F8,$69,$B1,$F4,$2F,$C3,$8F,$F7,$A9,$CD,$FF,$6B,$FC,$7B,$FA,$3B
	.dc.b	$7E,$2B,$FC,$47,$18,$AF,$F0,$0B,$FE,$CF,$C3,$FC,$4A,$E5,$FF,$15
	.dc.b	$B9,$AE,$9E,$57,$2B,$FF,$19,$79,$9F,$8A,$D3,$AF,$CB,$63,$8B,$D7
	.dc.b	$87,$8E,$31,$3A,$CB,$FD,$B7,$5A,$CF,$C7,$5F,$E3,$DD,$7F,$CE,$BE
	.dc.b	$47,$D1,$F8,$0F,$E6,$71,$8C,$3E,$94,$60,$93,$B2,$47,$F8,$AF,$6F
	.dc.b	$E3,$99,$FC,$5F,$07,$B5,$FF,$32,$8F,$F1,$6C,$EB,$2F,$F2,$B7,$E7
	.dc.b	$5F,$D2,$A7,$E6,$7A,$FF,$02,$31,$3F,$2A,$31,$E6,$FC,$E6,$E5,$F3
	.dc.b	$1F,$A3,$F5,$5F,$4B,$D7,$B3,$F4,$1B,$8F,$D9,$6E,$1F,$59,$FB,$3F
	.dc.b	$83,$FF,$8D,$73,$34,$97,$EF,$FC,$53,$EF,$9C,$1F,$BA,$FE,$0B,$C8
	.dc.b	$91,$83,$F0,$2E,$31,$FC,$53,$CA,$32,$7E,$1B,$E9,$7C,$44,$7E,$73
	.dc.b	$E5,$F8,$4F,$C7,$9F,$90,$6C,$7E,$3B,$E1,$BD,$44,$7E,$B3,$ED,$F8
	.dc.b	$0F,$CB,$9F,$98,$6C,$7E,$5A,$2E,$1F,$9F,$9B,$AF,$E7,$26,$D7,$E6
	.dc.b	$A6,$D7,$DE,$F7,$FE,$51,$27,$F0,$CF,$CD,$F8,$2F,$D3,$9F,$A8,$6C
	.dc.b	$7E,$9B,$ED,$FE,$91,$1F,$C7,$36,$BF,$CE,$7B,$A3,$11,$FE,$54,$8C
	.dc.b	$4F,$E5,$7D,$0F,$DA,$9F,$B6,$AB,$1F,$CB,$BB,$7F,$D6,$DD,$FF,$E8
	.dc.b	$36,$C7,$FA,$DF,$E1,$7F,$7F,$8F,$FA,$F1,$AF,$FA,$E2,$B8,$46,$23
	.dc.b	$FD,$89,$18,$3F,$19,$E0,$FF,$06,$3F,$7B,$3F,$C0,$E7,$F4,$B1,$9A
	.dc.b	$FE,$0E,$EA,$B1,$3F,$D0,$45,$BE,$E3,$FD,$A9,$7F,$ED,$FC,$BF,$97
	.dc.b	$CE,$7C,$44,$E6,$F8,$DF,$F8,$21,$FF,$78,$5F,$08,$E7,$18,$37,$5C
	.dc.b	$9C,$5F,$5B,$E8,$F8,$E3,$5F,$00,$C3,$A5,$E1,$B7,$C3,$0E,$F5,$BB
	.dc.b	$F2,$9B,$3C,$33,$E9,$5F,$78,$FA,$5F,$F7,$A4,$79,$CC,$B8,$AE,$9E
	.dc.b	$E3,$EF,$7F,$D5,$35,$BB,$E5,$9F,$56,$C6,$70,$E5,$9F,$F5,$AC,$39
	.dc.b	$7C,$27,$96,$BB,$E1,$C6,$FC,$71,$FE,$35,$C3,$9B,$63,$2F,$F6,$43
	.dc.b	$89,$FE,$38,$87,$DE,$BE,$F3,$FC,$77,$3F,$02,$30,$ED,$AF,$76,$BD
	.dc.b	$67,$9C,$FB,$67,$8E,$9F,$3E,$BA,$4F,$E1,$72,$7C,$1B,$C9,$F3,$5E
	.dc.b	$F1,$E7,$87,$C9,$C7,$9E,$BD,$DB,$0C,$79,$E1,$CA,$BF,$2B,$F9,$DB
	.dc.b	$ED,$B1,$7F,$9C,$8C,$3F,$8C,$6B,$F9,$05,$E5,$70,$E4,$46,$B4,$FC
	.dc.b	$54,$62,$F5,$39,$C3,$C3,$78,$99,$87,$77,$71,$E4,$F8,$47,$53,$E1
	.dc.b	$FE,$C7,$81,$FD,$F9,$BC,$FF,$DF,$1B,$8F,$AF,$F0,$AB,$87,$29,$B8
	.dc.b	$73,$AF,$F0,$2C,$70,$D7,$D7,$77,$53,$F8,$28,$CD,$7F,$9F,$9A,$FF
	.dc.b	$A0,$DA,$FE,$35,$AF,$D0,$8E,$53,$D2,$39,$35,$FE,$5B,$C9,$E6,$23
	.dc.b	$FE,$54,$C5,$F5,$2F,$9E,$F6,$B6,$7F,$D2,$33,$A9,$FC,$CB,$63,$33
	.dc.b	$B7,$FD,$35,$19,$6C,$65,$FD,$34,$65,$B1,$97,$FC,$9B,$1E,$4F,$63
	.dc.b	$C6,$48,$F4,$4F,$FB,$69,$BE,$6F,$92,$31,$1E,$08,$C1,$9F,$B0,$F6
	.dc.b	$92,$39,$FC,$CD,$7F,$9A,$61,$F1,$27,$1C,$3F,$C9,$E3,$E3,$D6,$38
	.dc.b	$7C,$F8,$AF,$18,$2F,$6F,$63,$E3,$DF,$9C,$8F,$CA,$FA,$FF,$A2,$47
	.dc.b	$DF,$F1,$3F,$50,$F7,$11,$FA,$2F,$C1,$F5,$BE,$C9,$23,$E3,$FD,$55
	.dc.b	$7D,$B9,$7F,$A7,$1F,$97,$3F,$80,$99,$1C,$08,$70,$FB,$1F,$C2,$9F
	.dc.b	$3C,$8F,$E0,$9F,$A5,$D1,$F8,$AE,$49,$3F,$62,$3F,$1A,$61,$F1,$C6
	.dc.b	$1F,$EF,$5E,$AF,$E5,$1F,$BF,$D3,$F9,$43,$F5,$91,$98,$FE,$D5,$CE
	.dc.b	$8C,$71,$7D,$7D,$3F,$86,$D7,$23,$AB,$F9,$EF,$B9,$91,$F8,$C7,$F7
	.dc.b	$EB,$ED,$5E,$76,$1B,$1F,$CE,$9A,$FF,$BD,$9A,$FF,$BE,$DA,$FF,$45
	.dc.b	$6B,$FE,$AA,$6B,$DE,$3F,$A2,$B4,$FE,$09,$BB,$DF,$FD,$AF,$C6,$49
	.dc.b	$FF,$42,$C7,$F0,$F9,$D4,$7E,$9A,$74,$FC,$68,$CB,$F8,$2F,$0F,$F7
	.dc.b	$FE,$44,$8F,$D6,$C7,$F0,$23,$0F,$E2,$23,$0F,$F0,$9F,$31,$F0,$7F
	.dc.b	$C9,$1B,$19,$8F,$97,$F6,$6F,$9D,$A7,$B9,$C4,$FF,$64,$6B,$FD,$AC
	.dc.b	$FF,$67,$B8,$70,$D7,$FB,$83,$5D,$C7,$53,$F8,$26,$FF,$C7,$09,$1F
	.dc.b	$C2,$BC,$BA,$FE,$92,$2F,$0F,$E2,$67,$F9,$1F,$77,$F2,$4F,$8C,$97
	.dc.b	$FC,$4A,$32,$FF,$14,$8C,$BF,$AE,$8F,$C5,$FD,$9B,$B7,$FA,$4F,$C5
	.dc.b	$BE,$6A,$9A,$8F,$C8,$FD,$93,$EC,$F1,$8F,$FB,$C3,$4F,$F1,$C2,$C7
	.dc.b	$A3,$9C,$BF,$C2,$57,$C6,$30,$7F,$BE,$1C,$3F,$E5,$13,$96,$1D,$9C
	.dc.b	$46,$AF,$FC,$8D,$D6,$FC,$55,$C7,$FD,$75,$FF,$08,$4E,$9B,$8F,$F9
	.dc.b	$61,$DE,$1F,$E7,$56,$BF,$B8,$76,$FF,$98,$9B,$7D,$EF,$FB,$7F,$C8
	.dc.b	$FF,$91,$6B,$FA,$C3,$5F,$E6,$ED,$C3,$F9,$B3,$FE,$30,$8D,$F0,$72
	.dc.b	$23,$CB,$9B,$F3,$1D,$32,$3F,$38,$FF,$02,$FC,$25,$F0,$B4,$FD,$AB
	.dc.b	$89,$FF,$1D,$B5,$FC,$C0,$70,$FE,$6D,$AF,$F9,$1D,$A8,$E1,$1D,$8F
	.dc.b	$ED,$FE,$BF,$B7,$6F,$EC,$AD,$9F,$E9,$CD,$F2,$1F,$2B,$F8,$49,$FE
	.dc.b	$B6,$BB,$3F,$BC,$B6,$3F,$BE,$B5,$FE,$F2,$D7,$FC,$B9,$EA,$BF,$1C
	.dc.b	$7B,$DF,$E2,$2D,$7F,$DD,$B5,$FE,$0B,$6B,$FE,$9D,$AE,$13,$FE,$28
	.dc.b	$B7,$95,$49,$5F,$E1,$38,$F5,$1F,$75,$FF,$03,$B5,$FF,$02,$B5,$FE
	.dc.b	$DC,$DC,$BE,$0F,$07,$B8,$93,$EE,$7F,$C3,$3F,$32,$F8,$EF,$F1,$DB
	.dc.b	$67,$FC,$48,$D7,$FC,$07,$F4,$1E,$CE,$5F,$F0,$13,$EE,$8F,$F0,$DB
	.dc.b	$5F,$F0,$CB,$5F,$F0,$13,$70,$34,$9F,$11,$FE,$2C,$75,$6F,$A5,$F2
	.dc.b	$92,$FF,$C7,$12,$F8,$6F,$F2,$9B,$63,$FC,$6E,$D7,$FC,$66,$D7,$CF
	.dc.b	$1D,$C8,$F0,$6B,$FE,$3A,$C3,$FC,$0C,$FF,$89,$1B,$8F,$F8,$8D,$BC
	.dc.b	$9F,$E3,$8E,$AF,$E8,$35,$FE,$2A,$6B,$FE,$40,$6B,$BF,$1B,$FC,$97
	.dc.b	$18,$8F,$F1,$43,$6B,$FD,$CB,$3A,$24,$FF,$95,$23,$FB,$EB,$E5,$F3
	.dc.b	$7F,$CE,$B1,$8C,$3F,$C7,$16,$F9,$1F,$F3,$CB,$6F,$FC,$BA,$D7,$FC
	.dc.b	$B2,$D7,$D1,$1E,$62,$3C,$1A,$FF,$98,$3E,$3F,$F8,$ED,$7F,$91,$DB
	.dc.b	$97,$F9,$71,$B3,$FE,$45,$FC,$6D,$F3,$1F,$E8,$C6,$C7,$F9,$C1,$AF
	.dc.b	$F9,$B1,$AF,$F9,$A9,$AF,$F9,$A1,$AF,$58,$FF,$3B,$C6,$32,$36,$DE
	.dc.b	$CF,$FA,$71,$B7,$FE,$85,$6B,$FE,$64,$97,$E2,$9E,$73,$A9,$23,$D0
	.dc.b	$FF,$A1,$BA,$7F,$99,$47,$F9,$B1,$B9,$FF,$9A,$DB,$7F,$E8,$07,$F8
	.dc.b	$94,$67,$C0,$FF,$57,$C6,$AB,$FC,$D8,$DC,$BF,$DA,$9D,$BF,$EB,$4F
	.dc.b	$91,$FF,$7B,$E1,$FE,$77,$6E,$3F,$E8,$C6,$E1,$FE,$75,$7F,$C4,$73
	.dc.b	$BF,$86,$E4,$49,$FF,$53,$4E,$AB,$F9,$B7,$0B,$E8,$DF,$31,$FE,$BF
	.dc.b	$6D,$FF,$70,$9D,$4F,$FB,$0F,$E2,$FF,$AE,$D6,$AB,$FC,$B2,$DC,$BF
	.dc.b	$B7,$B7,$0E,$F5,$FD,$90,$FF,$7F,$57,$FB,$0D,$AE,$BE,$17,$FD,$99
	.dc.b	$18,$8F,$E9,$9F,$49,$FE,$5A,$C3,$FA,$77,$57,$ED,$A1,$F1,$FF,$D4
	.dc.b	$6B,$F8,$AF,$55,$9C,$C7,$08,$C7,$B0,$C3,$13,$DA,$30,$F2,$3C,$37
	.dc.b	$D7,$0F,$03,$6F,$9D,$F2,$A6,$1E,$77,$FB,$5B,$5F,$F2,$23,$5E,$26
	.dc.b	$BE,$41,$C1,$FD,$CA,$39,$C6,$1D,$46,$65,$39,$9E,$1F,$13,$FE,$72
	.dc.b	$4D,$23,$FE,$76,$7D,$0B,$89,$BE,$23,$89,$F0,$CF,$EF,$D1,$A7,$07
	.dc.b	$F8,$06,$79,$C6,$74,$FE,$4F,$18,$AF,$9F,$C4,$DD,$F2,$1F,$E4,$4C
	.dc.b	$7F,$0E,$D5,$3B,$8C,$36,$BF,$C2,$EF,$E6,$FD,$4F,$26,$DF,$92,$6C
	.dc.b	$F4,$3F,$39,$C7,$FE,$59,$6C,$6B,$0E,$F3,$97,$FD,$6B,$8F,$9A,$B6
	.dc.b	$7F,$95,$39,$9F,$49,$FE,$0C,$8C,$36,$BF,$DE,$6F,$E5,$A3,$3E,$4F
	.dc.b	$7A,$E2,$D9,$E1,$97,$FB,$DD,$E1,$7C,$A7,$87,$D2,$3E,$56,$98,$F3
	.dc.b	$EE,$5E,$59,$D3,$67,$4D,$3F,$93,$78,$9F,$3F,$3F,$C0,$B6,$0D,$35
	.dc.b	$C2,$FA,$4F,$53,$67,$99,$F1,$5C,$67,$FA,$F8,$FF,$23,$5F,$E3,$BB
	.dc.b	$7B,$12,$3F,$CA,$31,$96,$9F,$E4,$12,$C6,$E7,$FD,$5D,$3B,$6D,$3E
	.dc.b	$6F,$CA,$7C,$FB,$E5,$3E,$31,$5F,$A6,$9F,$A9,$C6,$FA,$4F,$88,$DB
	.dc.b	$FE,$DF,$18,$23,$82,$5F,$F7,$28,$F9,$9F,$1F,$FD,$E6,$7F,$10,$D9
	.dc.b	$FF,$07,$4F,$E0,$E3,$1C,$5F,$E0,$0F,$9F,$F0,$92,$FF,$BF,$B6,$FF
	.dc.b	$BF,$B5,$FF,$0B,$3F,$6F,$E9,$1F,$2F,$F8,$F6,$D7,$E6,$72,$56,$6B
	.dc.b	$F2,$9E,$73,$EA,$BF,$08,$DB,$FA,$AD,$C7,$E8,$B5,$FE,$65,$B2,$31
	.dc.b	$1F,$80,$6B,$F4,$1B,$1B,$AF,$C9,$7C,$BF,$FA,$DA,$FF,$5C,$DB,$FB
	.dc.b	$15,$B9,$CF,$B7,$E8,$F1,$FC,$E4,$8F,$3B,$6B,$CE,$D7,$CF,$8F,$F6
	.dc.b	$29,$FC,$8F,$C1,$3E,$47,$F6,$CD,$8F,$F3,$6D,$E2,$7F,$75,$EB,$FE
	.dc.b	$9A,$7F,$D2,$7D,$1F,$AB,$8C,$63,$FE,$94,$8C,$12,$76,$48,$FC,$DF
	.dc.b	$B3,$FC,$0A,$3F,$4E,$D8,$FF,$43,$CF,$F0,$98,$C7,$98,$FA,$47,$C8
	.dc.b	$7F,$0D,$FA,$5F,$66,$1F,$E9,$B9,$CE,$3F,$C6,$67,$3C,$FF,$CE,$F8
	.dc.b	$7F,$D6,$3F,$17,$E9,$25,$7F,$AB,$9F,$E0,$DF,$4A,$FA,$0F,$E3,$4D
	.dc.b	$9F,$F5,$57,$A5,$89,$FE,$59,$0F,$A5,$7C,$FF,$F3,$D8,$C3,$63,$31
	.dc.b	$FE,$D2,$32,$57,$3A,$FF,$7F,$3F,$AB,$B8,$7F,$C0,$51,$C1,$D3,$5F
	.dc.b	$CA,$FC,$2B,$9B,$F9,$93,$67,$FD,$B5,$AF,$F5,$D7,$93,$FE,$F4,$9F
	.dc.b	$E7,$1F,$17,$FA,$89,$FE,$84,$DB,$FC,$AB,$FE,$64,$8C,$F6,$F0,$23
	.dc.b	$C3,$DB,$B5,$F9,$CF,$A5,$FA,$73,$FC,$99,$7A,$DF,$D1,$DC,$CF,$F1
	.dc.b	$87,$33,$FD,$4D,$CF,$11,$CC,$F5,$EB,$E5,$7F,$D7,$E1,$DB,$FE,$6A
	.dc.b	$C3,$E0,$9F,$DF,$BE,$95,$F2,$3B,$9D,$57,$F2,$9F,$55,$5A,$BE,$32
	.dc.b	$3A,$FB,$DF,$DB,$6B,$78,$73,$8C,$5F,$1A,$E8,$DE,$1F,$CA,$69,$FC
	.dc.b	$EF,$1E,$B8,$F2,$7E,$07,$93,$F1,$7B,$BE,$37,$C0,$59,$F8,$8D,$CB
	.dc.b	$E7,$37,$1E,$AF,$CD,$89,$3F,$2F,$0B,$1F,$54,$B5,$E4,$8B,$3E,$FF
	.dc.b	$80,$F8,$1F,$49,$FE,$19,$6C,$7E,$5E,$D1,$87,$F3,$4D,$7F,$CB,$4D
	.dc.b	$7A,$09,$E8,$D9,$E9,$3B,$8F,$EB,$7B,$BE,$17,$FC,$1C,$DA,$FA,$B5
	.dc.b	$E1,$C7,$F8,$04,$E6,$FF,$96,$FC,$F7,$F4,$66,$FD,$2F,$F8,$5E,$31
	.dc.b	$5F,$C1,$D4,$67,$CB,$D3,$71,$FF,$13,$4E,$9F,$F4,$67,$1F,$A7,$0E
	.dc.b	$FF,$41,$FE,$31,$F3,$7F,$1F,$B7,$78,$C5,$6F,$3F,$04,$62,$75,$87
	.dc.b	$B3,$E0,$3F,$97,$E3,$F2,$9A,$FD,$56,$BF,$49,$BB,$7F,$9C,$3C,$9F
	.dc.b	$B1,$18,$23,$FE,$25,$D3,$F8,$F5,$7C,$6F,$0B,$FC,$91,$7C,$70,$F7
	.dc.b	$C7,$AF,$E6,$3F,$A4,$57,$CC,$F6,$31,$E6,$FC,$B6,$DF,$C6,$6C,$EE
	.dc.b	$3E,$4F,$D2,$FD,$1F,$63,$E9,$36,$32,$7C,$E8,$CB,$4F,$A3,$F1,$5E
	.dc.b	$92,$3F,$E2,$ED,$11,$ED,$FA,$5F,$BC,$7D,$F6,$CF,$DA,$F2,$7F,$04
	.dc.b	$63,$1F,$C3,$B8,$8D,$1F,$80,$AC,$4F,$E3,$3E,$2B,$E5,$12,$7C,$4F
	.dc.b	$CD,$F8,$4F,$C7,$1F,$8F,$6C,$7E,$2B,$E1,$FE,$91,$1F,$53,$F5,$7E
	.dc.b	$1F,$F2,$C7,$E5,$DB,$1F,$92,$8B,$7F,$9E,$9B,$CF,$F9,$B9,$B5,$FE
	.dc.b	$56,$D1,$58,$BF,$F2,$3C,$7F,$9F,$5F,$F0,$FB,$5E,$1F,$01,$F8,$73
	.dc.b	$F1,$1F,$0B,$FA,$93,$F5,$4D,$BF,$D2,$FD,$2F,$E3,$48,$FD,$56,$D7
	.dc.b	$F9,$7A,$D1,$88,$F0,$BF,$95,$F4,$7F,$6C,$7E,$DD,$B5,$FE,$FB,$F8
	.dc.b	$4F,$DE,$57,$FB,$C7,$C0,$FF,$55,$C6,$23,$FD,$EF,$18,$24,$FF,$AB
	.dc.b	$A7,$53,$FE,$B4,$6F,$98,$C7,$FC,$6D,$B7,$F0,$5F,$0B,$F7,$95,$FC
	.dc.b	$23,$0F,$99,$1F,$C2,$C8,$F9,$71,$B8,$FF,$9D,$7C,$17,$D2,$FF,$B5
	.dc.b	$6B,$FD,$9D,$AF,$F4,$D9,$CF,$8B,$FE,$D3,$F8,$14,$E6,$FF,$A9,$5F
	.dc.b	$0C,$3F,$D9,$71,$CD,$FF,$72,$CE,$EF,$93,$6B,$86,$1E,$2A,$C9,$CA
	.dc.b	$AC,$78,$A7,$77,$D7,$0F,$1D,$F8,$0D,$C6,$70,$FE,$CD,$0F,$A5,$7C
	.dc.b	$7F,$FB,$CE,$3C,$A7,$F6,$B8,$CE,$5E,$BF,$71,$F2,$5F,$E9,$CC,$BB
	.dc.b	$DF,$3C,$39,$46,$30,$E3,$3D,$0D,$61,$CE,$F9,$63,$E4,$AD,$DF,$1C
	.dc.b	$3C,$98,$7F,$8C,$67,$9B,$63,$2F,$16,$BE,$9F,$8D,$7D,$27,$F6,$38
	.dc.b	$F8,$11,$87,$FB,$35,$23,$11,$DD,$AF,$95,$E1,$7E,$D6,$E1,$9B,$F8
	.dc.b	$DF,$04,$F2,$BE,$7D,$EE,$7F,$09,$87,$89,$B8,$F8,$5E,$35,$C3,$0F
	.dc.b	$C8,$7D,$2B,$E8,$71,$83,$89,$39,$8F,$C4,$B5,$FB,$71,$AF,$2B,$8B
	.dc.b	$C2,$AD,$6F,$3F,$C4,$CE,$6F,$9D,$EA,$E4,$76,$78,$C7,$93,$A7,$F6
	.dc.b	$9F,$0F,$F8,$EE,$9F,$EF,$0D,$E6,$FE,$3C,$DC,$3D,$59,$BE,$87,$18
	.dc.b	$72,$9F,$EF,$75,$8E,$B3,$C7,$4F,$83,$77,$0E,$57,$FE,$21,$9C,$B6
	.dc.b	$32,$FF,$9E,$E3,$2D,$8C,$BF,$A6,$8C,$B6,$32,$FD,$49,$FC,$66,$1F
	.dc.b	$C4,$F8,$3C,$C4,$9F,$F2,$AC,$3E,$75,$F4,$BE,$E6,$CF,$F8,$DE,$75
	.dc.b	$3D,$27,$33,$FE,$91,$9C,$BB,$8C,$BF,$EE,$26,$BF,$E6,$38,$DB,$C6
	.dc.b	$32,$A3,$C9,$EC,$F8,$C9,$1E,$89,$FF,$6A,$37,$CC,$F2,$11,$F1,$3E
	.dc.b	$7F,$A5,$F6,$12,$39,$7D,$2D,$7F,$DD,$74,$9C,$70,$FF,$B0,$E1,$58
	.dc.b	$E1,$F5,$46,$EF,$C9,$7F,$26,$A4,$7C,$37,$E4,$11,$F8,$DF,$3D,$F3
	.dc.b	$C8,$FB,$7E,$27,$E9,$1F,$3C,$8F,$CD,$7E,$97,$AD,$F5,$C9,$1F,$17
	.dc.b	$E6,$AB,$EC,$8B,$9F,$AE,$7E,$F1,$27,$81,$F3,$7E,$97,$F0,$47,$DF
	.dc.b	$23,$F7,$5F,$C1,$74,$7E,$1F,$8E,$4F,$D5,$8F,$C5,$18,$7E,$6C,$61
	.dc.b	$FF,$52,$F2,$BD,$4F,$C9,$47,$DD,$E9,$FC,$91,$FA,$B8,$CC,$7F,$63
	.dc.b	$6E,$5F,$56,$77,$3D,$27,$3C,$3F,$5B,$E9,$7F,$38,$49,$FF,$5A,$8F
	.dc.b	$A5,$FF,$0E,$B8,$6B,$FE,$97,$6B,$FE,$EF,$6B,$FD,$61,$AF,$FA,$FD
	.dc.b	$AF,$5F,$37,$F9,$F3,$4F,$E8,$11,$C3,$2F,$F6,$6F,$BD,$FB,$1F,$8B
	.dc.b	$95,$FE,$86,$8F,$E1,$9C,$27,$31,$F9,$89,$CC,$7E,$E7,$C4,$FD,$D7
	.dc.b	$C9,$CA,$FD,$44,$65,$FD,$54,$65,$FD,$0C,$7F,$1E,$F3,$3F,$DD,$F1
	.dc.b	$FC,$19,$AB,$83,$FD,$4F,$F0,$9F,$C9,$E1,$F4,$AF,$A5,$C3,$63,$FA
	.dc.b	$F3,$5F,$F6,$13,$5F,$F0,$09,$CA,$3E,$3C,$66,$3F,$C0,$AD,$7F,$DD
	.dc.b	$8D,$78,$1F,$C4,$7B,$3F,$8D,$11,$FE,$11,$E5,$57,$F7,$F1,$70,$FD
	.dc.b	$F4,$FF,$21,$E3,$FE,$45,$F1,$72,$FF,$88,$46,$5F,$E2,$51,$97,$F8
	.dc.b	$4C,$7C,$3F,$A4,$76,$FF,$48,$F8,$C7,$CD,$D3,$51,$F2,$7E,$91,$F4
	.dc.b	$B8,$C7,$F2,$B8,$72,$7F,$2F,$3A,$6C,$ED,$FA,$98,$FF,$8A,$AB,$84
	.dc.b	$60,$0A,$E4,$FF,$93,$7E,$95,$C5,$FD,$65,$FF,$14,$EC,$C3,$FC,$9D
	.dc.b	$18,$6C,$EC,$E1,$E7,$4B,$FD,$ED,$74,$7F,$A7,$FC,$8F,$F8,$F1,$BC
	.dc.b	$1F,$DB,$23,$11,$E7,$F2,$51,$FE,$21,$8D,$90,$D3,$C3,$2F,$34,$7F
	.dc.b	$49,$8E,$0F,$3C,$7F,$50,$FE,$6F,$A3,$F3,$C7,$E6,$7E,$95,$F0,$B4
	.dc.b	$FD,$BB,$89,$FF,$1B,$35,$FC,$B8,$70,$FE,$69,$AF,$F9,$71,$AF,$D2
	.dc.b	$6B,$C6,$3F,$BB,$FA,$BE,$BD,$BF,$B3,$36,$7F,$B2,$37,$37,$CB,$FD
	.dc.b	$ED,$3E,$0C,$63,$B3,$FC,$03,$26,$92,$FD,$3E,$C3,$E9,$5E,$F7,$F8
	.dc.b	$7B,$6F,$F7,$1D,$7F,$81,$9A,$FE,$2E,$B8,$4F,$F8,$93,$07,$97,$49
	.dc.b	$5F,$E0,$E8,$F4,$1F,$79,$FF,$3F,$E5,$EA,$E2,$F3,$92,$3F,$C2,$13
	.dc.b	$B7,$FC,$10,$D7,$FC,$0C,$D9,$E5,$1F,$E1,$7F,$99,$7C,$77,$F8,$E1
	.dc.b	$B1,$FE,$20,$6B,$F8,$FF,$A1,$F6,$EF,$F2,$9F,$7C,$79,$9A,$FF,$86
	.dc.b	$5A,$FF,$85,$DB,$66,$93,$E2,$3F,$C6,$6E,$AD,$F4,$BE,$52,$5F,$F8
	.dc.b	$D2,$5F,$0D,$FE,$50,$6C,$7F,$8C,$DA,$FF,$8C,$5A,$F9,$E3,$B9,$1E
	.dc.b	$0D,$7F,$C7,$18,$3C,$8F,$F2,$EB,$63,$FC,$8A,$D7,$F6,$9E,$4F,$F4
	.dc.b	$2C,$FF,$1E,$FF,$44,$8F,$B6,$D7,$FC,$7C,$D7,$FC,$78,$D9,$FF,$43
	.dc.b	$78,$FF,$E5,$68,$C1,$1F,$F2,$CC,$3C,$EF,$29,$2B,$FC,$B3,$4F,$91
	.dc.b	$FF,$3A,$B6,$3F,$CB,$0D,$7F,$CA,$CD,$7C,$F1,$E1,$23,$C1,$AF,$F9
	.dc.b	$77,$E3,$BF,$8D,$3F,$C5,$5B,$8F,$EC,$1B,$D5,$FE,$5D,$FA,$57,$CB
	.dc.b	$FF,$A2,$1B,$7F,$E6,$B6,$BF,$E6,$96,$BF,$A4,$6B,$CE,$3F,$CF,$71
	.dc.b	$88,$FF,$33,$37,$33,$69,$ED,$7F,$A6,$1B,$5F,$E8,$06,$BF,$CB,$E5
	.dc.b	$F4,$BE,$73,$B9,$23,$D0,$FF,$A3,$3A,$BF,$96,$E7,$FC,$41,$B7,$FE
	.dc.b	$7D,$6D,$7F,$9E,$DF,$E2,$4F,$FA,$41,$CE,$26,$93,$B7,$FD,$5F,$F2
	.dc.b	$5F,$EE,$EC,$3F,$8E,$B7,$0F,$F4,$53,$70,$FF,$44,$3F,$CF,$9F,$E9
	.dc.b	$4E,$70,$34,$9D,$BF,$EB,$7F,$A4,$7C,$CD,$25,$FF,$95,$BE,$2F,$FB
	.dc.b	$03,$FE,$DD,$9D,$BF,$E5,$A6,$BF,$DB,$F8,$D1,$8C,$FF,$AA,$C2,$B9
	.dc.b	$B7,$C8,$FE,$97,$7F,$EE,$99,$46,$27,$53,$FE,$C9,$9C,$BC,$3E,$95
	.dc.b	$EB,$FF,$8D,$63,$F8,$97,$47,$F8,$01,$4B,$E3,$F1,$D9,$FE,$3C,$8C
	.dc.b	$4F,$F8,$1A,$31,$7C,$DB,$9E,$70,$E9,$7C,$2F,$B9,$FE,$0E,$71,$5C
	.dc.b	$23,$27,$8A,$FC,$CF,$F9,$02,$7D,$37,$EB,$9C,$CF,$66,$BD,$7E,$43
	.dc.b	$ED,$BF,$DC,$A9,$18,$8E,$8D,$7C,$ED,$7F,$C9,$FF,$11,$FE,$74,$4D
	.dc.b	$23,$FE,$78,$7E,$03,$F2,$FE,$33,$FC,$1F,$8F,$F1,$F9,$CC,$FF,$8B
	.dc.b	$A7,$39,$7F,$81,$A7,$32,$9C,$BE,$23,$E2,$99,$E4,$EF,$B7,$FB,$4F
	.dc.b	$C0,$FF,$0E,$CF,$FC,$01,$F8,$3F,$E9,$35,$FE,$12,$AD,$36,$33,$3F
	.dc.b	$E6,$C9,$7C,$E3,$E1,$7F,$C4,$FC,$1F,$94,$8C,$61,$FE,$B4,$C3,$8C
	.dc.b	$E6,$B7,$3F,$E4,$AB,$E7,$18,$C1,$18,$9E,$58,$E6,$31,$3F,$E3,$DF
	.dc.b	$A5,$FC,$BD,$7F,$95,$30,$FF,$0C,$C6,$27,$5B,$F5,$E9,$4E,$6B,$7C
	.dc.b	$59,$8C,$61,$FE,$5E,$AD,$DE,$6B,$97,$D2,$BE,$62,$98,$F8,$EB,$87
	.dc.b	$8C,$E5,$3B,$71,$9B,$14,$E6,$3F,$CB,$0A,$B9,$C7,$F9,$6E,$9F,$0D
	.dc.b	$F6,$A3,$CD,$18,$3F,$C9,$77,$FE,$6F,$9D,$4F,$79,$D3,$FE,$A6,$9F
	.dc.b	$CB,$FB,$3F,$17,$93,$FD,$13,$5F,$CF,$6F,$13,$F2,$7E,$FF,$8E,$1F
	.dc.b	$E3,$06,$BF,$1D,$AF,$B5,$B8,$1A,$4A,$F2,$C6,$5B,$CA,$F9,$8F,$B0
	.dc.b	$DC,$7E,$4D,$6D,$B1,$9C,$3F,$C1,$BE,$EF,$4E,$3F,$F5,$3D,$7F,$7A
	.dc.b	$8C,$11,$C1,$27,$FB,$EC,$79,$3E,$3D,$F3,$9F,$89,$6D,$7F,$82,$27
	.dc.b	$13,$F8,$DF,$B2,$7C,$1F,$B1,$B5,$F5,$1B,$87,$D1,$6B,$FE,$5A,$6B
	.dc.b	$F9,$16,$BF,$87,$6E,$66,$F6,$FF,$9C,$2B,$FC,$81,$EB,$F8,$F0,$FD
	.dc.b	$04,$E7,$0F,$F5,$27,$A4,$FA,$7F,$A2,$DC,$3F,$CC,$11,$82,$38,$24
	.dc.b	$FE,$26,$78,$3F,$8B,$F9,$5F,$E9,$F3,$F5,$9B,$1E,$0A,$FD,$94,$63
	.dc.b	$B3,$FD,$1F,$F8,$AF,$F1,$7A,$FC,$A3,$5F,$F3,$EB,$5F,$DA,$35,$FD
	.dc.b	$53,$75,$34,$9C,$5F,$8E,$F6,$3F,$DA,$59,$7F,$A2,$CF,$E0,$93,$E1
	.dc.b	$C3,$FD,$43,$3A,$C3,$FD,$94,$FE,$68,$9C,$BC,$C9,$7F,$E8,$E6,$CF
	.dc.b	$EB,$67,$F4,$3E,$57,$F6,$7A,$FF,$51,$78,$DF,$C4,$E7,$F6,$78,$7F
	.dc.b	$AC,$2B,$57,$FE,$D9,$FA,$47,$8D,$FC,$65,$7D,$95,$8C,$3F,$C7,$7D
	.dc.b	$EF,$45,$7F,$A9,$1A,$FF,$B0,$B3,$FD,$7C,$62,$3F,$89,$7D,$2F,$F0
	.dc.b	$F9,$FD,$6B,$63,$FD,$D4,$BE,$6F,$F9,$53,$9C,$FF,$DC,$F5,$FD,$13
	.dc.b	$9F,$F8,$7B,$4F,$F6,$E9,$6F,$FC,$14,$1C,$3F,$E0,$D6,$A3,$5E,$57
	.dc.b	$E6,$5D,$35,$D4,$70,$7F,$DF,$1D,$1F,$CF,$FE,$13,$F5,$F8,$79,$AF
	.dc.b	$F4,$4F,$FB,$7A,$33,$7F,$CF,$FE,$27,$FD,$7E,$FF,$0E,$5F,$53,$FD
	.dc.b	$25,$CE,$2B,$FF,$73,$C6,$AB,$87,$C0,$6E,$F8,$36,$3A,$7D,$37,$33
	.dc.b	$A7,$B7,$63,$24,$C7,$B3,$E5,$8D,$46,$0F,$8D,$F0,$FF,$EE,$89,$FE
	.dc.b	$9B,$D4,$FA,$67,$CB,$7F,$6D,$B5,$ED,$E5,$5B,$9D,$4E,$75,$FA,$D5
	.dc.b	$FE,$0F,$BE,$78,$7F,$5A,$8E,$7F,$0D,$FD,$E7,$1E,$53,$E2,$AD,$D7
	.dc.b	$9F,$B9,$BB,$E9,$38,$56,$B2,$F9,$71,$F3,$FD,$39,$8F,$D3,$F7,$AE
	.dc.b	$5C,$F1,$FF,$06,$CF,$D9,$AF,$AD,$5A,$C3,$EA,$FE,$0F,$F6,$7E,$BB
	.dc.b	$ED,$36,$FE,$F3,$71,$E8,$DA,$F8,$3F,$07,$F9,$3C,$3F,$B5,$B6,$FF
	.dc.b	$37,$68,$C3,$F9,$E6,$BF,$E6,$96,$BF,$9D,$6B,$C4,$FF,$54,$C7,$48
	.dc.b	$CE,$BF,$23,$E4,$FF,$24,$7F,$89,$E3,$BE,$1F,$92,$8F,$CB,$78,$F2
	.dc.b	$FF,$35,$E8,$DC,$7E,$44,$DB,$FC,$CC,$DA,$FD,$17,$D5,$7E,$7C,$FF
	.dc.b	$18,$7E,$82,$C7,$13,$FA,$59,$B8,$FF,$A5,$EB,$53,$93,$FC,$3F,$97
	.dc.b	$FA,$A3,$E6,$7E,$4F,$C1,$9A,$49,$F8,$11,$96,$C6,$6F,$E6,$4E,$5F
	.dc.b	$F0,$D7,$17,$F2,$4A,$FD,$C7,$89,$F1,$6B,$5A,$FE,$DA,$3F,$75,$F0
	.dc.b	$72,$BF,$7D,$E3,$5C,$7F,$19,$36,$FF,$7B,$36,$BF,$83,$7C,$47,$F0
	.dc.b	$33,$F8,$27,$A5,$63,$F8,$44,$5C,$3F,$2D,$36,$FF,$83,$F9,$5F,$E2
	.dc.b	$0D,$0D,$26,$1F,$E3,$A8,$FE,$A8,$FF,$15,$EB,$FE,$25,$3F,$C4,$DF
	.dc.b	$E2,$98,$49,$FE,$31,$16,$3F,$4C,$5B,$FE,$2F,$16,$7F,$CB,$DF,$87
	.dc.b	$BE,$AB,$FC,$66,$D8,$FF,$25,$B6,$FF,$C9,$0D,$7F,$D8,$9C,$E8,$FE
	.dc.b	$8D,$3F,$E3,$66,$CF,$F0,$17,$96,$1F,$C3,$2B,$87,$95,$FE,$75,$9F
	.dc.b	$EA,$15,$D1,$AF,$9B,$DB,$F2,$9E,$6F,$77,$81,$C5,$B8,$F8,$62,$CF
	.dc.b	$86,$6E,$3E,$1B,$FF,$38,$36,$B9,$5F,$13,$D5,$F2,$3F,$10,$8F,$C4
	.dc.b	$7E,$3F,$B7,$E9,$3D,$4D,$8F,$85,$F5,$5F,$30,$8F,$CC,$6E,$5F,$31
	.dc.b	$AF,$C1,$7E,$9F,$C8,$7C,$53,$E3,$36,$7F,$D1,$7E,$EB,$E8,$DF,$17
	.dc.b	$FE,$80,$8C,$36,$33,$3F,$EB,$F6,$E1,$FE,$BD,$6B,$FE,$FF,$AE,$33
	.dc.b	$BA,$E4,$7E,$1B,$EA,$7E,$C1,$FE,$B5,$BE,$7A,$FC,$D8,$CD,$7E,$0D
	.dc.b	$D5,$63,$0F,$EB,$58,$7A,$7E,$77,$FC,$17,$1F,$EC,$5C,$51,$88,$FF
	.dc.b	$62,$C6,$30,$FF,$00,$C6,$E7,$55,$CE,$73,$3C,$27,$35,$C6,$3F,$DC
	.dc.b	$75,$B6,$C6,$7E,$0D,$F1,$9F,$A6,$FA,$AF,$EC,$71,$8A,$E9,$8F,$F8
	.dc.b	$5E,$31,$CB,$FA,$3A,$DD,$FF,$86,$ED,$F4,$EF,$96,$FF,$7D,$43,$0E
	.dc.b	$8E,$A3,$31,$DF,$E4,$5E,$E7,$D8,$8D,$DF,$3E,$BF,$16,$3D,$E3,$18
	.dc.b	$F5,$8C,$63,$D1,$FF,$17,$63,$D3,$83,$CF,$5F,$E3,$1B,$E3,$8F,$6C
	.dc.b	$3C,$F3,$FE,$3A,$9E,$AD,$8C,$BD,$B1,$FE,$CA,$FC,$0F,$9F,$7D,$36
	.dc.b	$2D,$18,$9F,$ED,$71,$89,$D4,$E6,$7C,$0E,$F0,$F8,$2D,$CF,$E0,$36
	.dc.b	$59,$BE,$13,$FD,$D9,$1F,$18,$F7,$4E,$DF,$36,$19,$69,$F2,$4E,$73
	.dc.b	$C9,$C7,$13,$F3,$47,$5F,$DF,$CF,$F0,$7D,$8F,$99,$C6,$BF,$CC,$1B
	.dc.b	$8B,$1E,$58,$61,$B1,$F9,$A6,$BF,$98,$C7,$BB,$96,$B9,$9F,$F3,$63
	.dc.b	$FE,$6B,$9E,$3C,$BF,$BD,$AD,$B6,$B7,$4E,$5F,$E7,$9F,$1F,$F9,$9E
	.dc.b	$AF,$E5,$0D,$BF,$E9,$2D,$AF,$56,$1F,$D2,$DF,$E8,$51,$9B,$E7,$18
	.dc.b	$3F,$8C,$CE,$DF,$E0,$6D,$7C,$DE,$0F,$E2,$B2,$F9,$CD,$9D,$4F,$B5
	.dc.b	$AF,$8A,$34,$47,$FA,$DF,$C7,$F8,$F6,$34,$C7,$C3,$49,$FE,$C2,$FE
	.dc.b	$6B,$1F,$EB,$CD,$A3,$49,$3C,$08,$7F,$68,$F8,$3F,$01,$1F,$EC,$DF
	.dc.b	$57,$E3,$3A,$11,$FE,$E2,$F7,$FA,$BF,$41,$23,$FB,$77,$15,$7F,$B7
	.dc.b	$45,$E0,$FE,$BD,$37,$83,$FB,$FC,$DA,$FB,$D9,$7F,$10,$BD,$C7,$F8
	.dc.b	$02,$FD,$C4,$7F,$C0,$7E,$EF,$C9,$23,$FD,$FF,$E3,$FE,$61,$F1,$48
	.dc.b	$FF,$82,$9F,$91,$CF,$F4,$49,$1F,$EB,$FB,$FF,$5A,$7D,$4E,$23,$FC
	.dc.b	$11,$36,$73,$1B,$CF,$FD,$8D,$BB,$E1,$58,$7E,$3D,$B5,$B3,$F3,$8E
	.dc.b	$1F,$F4,$A2,$F9,$AE,$07,$DF,$71,$1F,$E7,$EA,$ED,$39,$35,$F2,$DD
	.dc.b	$C9,$3D,$E7,$C9,$F2,$D8,$8E,$F3,$72,$FF,$69,$4F,$9F,$E8,$A5,$71
	.dc.b	$C5,$1E,$BF,$A9,$CC,$FA,$22,$DF,$FB,$6E,$7E,$07,$D4,$CA,$E4,$FE
	.dc.b	$D5,$7D,$3B,$97,$0D,$8F,$E6,$ED,$7F,$D3,$AD,$7F,$DF,$2D,$7F,$AE
	.dc.b	$35,$FF,$7C,$35,$EF,$1F,$D0,$9A,$7C,$BC,$5F,$41,$F5,$C9,$1F,$E1
	.dc.b	$CF,$39,$D1,$88,$FA,$F3,$73,$FF,$6A,$F6,$AF,$75,$F2,$7E,$27,$F0
	.dc.b	$44,$7E,$67,$C4,$D7,$F0,$31,$6F,$FC,$A9,$92,$BF,$15,$F5,$32,$FE
	.dc.b	$A3,$F2,$FE,$A1,$EC,$D3,$DC,$E2,$7F,$B0,$B5,$FF,$63,$B5,$FF,$02
	.dc.b	$07,$0F,$F8,$1D,$A8,$D4,$76,$3F,$2D,$F4,$5F,$9F,$F9,$69,$3F,$89
	.dc.b	$8F,$D1,$18,$7F,$3F,$18,$7F,$1D,$E9,$FE,$9C,$8F,$E4,$7E,$06,$BF
	.dc.b	$A6,$8B,$A7,$C6,$9F,$C6,$7D,$3F,$F8,$6E,$BF,$31,$E9,$7E,$C4,$CF
	.dc.b	$A5,$F9,$1F,$A7,$7C,$8E,$2B,$F6,$82,$F9,$2C,$5E,$0F,$F8,$A6,$F8
	.dc.b	$CF,$06,$BB,$8F,$F1,$75,$72,$AE,$33,$F2,$63,$13,$CE,$FF,$83,$61
	.dc.b	$2B,$F6,$0F,$F0,$9F,$47,$11,$FC,$1A,$2E,$1F,$86,$9F,$EC,$1C,$CF
	.dc.b	$A7,$FF,$0B,$4F,$E5,$17,$8B,$A6,$E1,$FE,$78,$77,$97,$F7,$56,$C7
	.dc.b	$F8,$0D,$DB,$FE,$66,$6D,$75,$F1,$7F,$B6,$78,$DF,$E3,$48,$FF,$56
	.dc.b	$36,$FF,$92,$B6,$FF,$15,$F4,$FC,$A7,$FC,$A8,$F6,$8F,$C9,$B5,$CE
	.dc.b	$3F,$B9,$8C,$47,$28,$C1,$1F,$F1,$FC,$6B,$C5,$FF,$2B,$FD,$47,$F8
	.dc.b	$07,$D8,$34,$C7,$CC,$7F,$81,$7E,$69,$F1,$BF,$DB,$DB,$8F,$F3,$16
	.dc.b	$BF,$A8,$C1,$18,$8F,$F6,$9F,$C2,$3D,$DF,$F0,$82,$F7,$E9,$88,$FF
	.dc.b	$0A,$7C,$DB,$C4,$A4,$BF,$E7,$50,$CF,$67,$F8,$0D,$C4,$7F,$45,$6B
	.dc.b	$E5,$6B,$FE,$E6,$6B,$FD,$61,$AF,$F8,$BD,$B9,$1B,$F3,$8F,$88,$FF
	.dc.b	$15,$36,$BF,$A7,$35,$FD,$11,$FE,$39,$F7,$BF,$C2,$AF,$F5,$36,$BF
	.dc.b	$25,$AF,$FB,$E9,$AF,$F6,$C6,$BC,$FD,$7F,$F1,$14,$63,$0F,$F3,$64
	.dc.b	$60,$FF,$13,$7C,$EB,$E5,$7F,$C8,$0D,$BF,$EC,$8D,$7C,$F4,$F9,$E7
	.dc.b	$BD,$FE,$47,$6C,$FF,$71,$6B,$F8,$16,$BF,$EA,$D9,$E5,$3F,$E3,$79
	.dc.b	$CC,$FF,$81,$9A,$FF,$97,$5B,$3C,$23,$FB,$F7,$CF,$FF,$B4,$1F,$EF
	.dc.b	$4D,$7C,$ED,$EB,$FE,$1B,$3F,$BE,$E5,$FF,$2F,$74,$FE,$4E,$FF,$C2
	.dc.b	$AD,$7F,$CD,$8D,$9E,$11,$FE,$0D,$8C,$12,$7F,$CA,$0E,$E5,$E2,$FF
	.dc.b	$9A,$DB,$5F,$E0,$F6,$BF,$33,$BD,$CF,$FE,$71,$6C,$7F,$86,$DA,$FE
	.dc.b	$AD,$AF,$CB,$6B,$FE,$2E,$6B,$FE,$85,$6C,$FF,$A6,$67,$FC,$CD,$18
	.dc.b	$FA,$17,$AB,$FE,$81,$6D,$FF,$89,$5A,$FE,$0B,$E0,$DF,$39,$FE,$87
	.dc.b	$6C,$7F,$8C,$5A,$FF,$01,$6B,$F8,$16,$BF,$E4,$46,$BF,$E9,$B6,$E3
	.dc.b	$FE,$EF,$9F,$F3,$E4,$63,$C0,$7B,$3F,$E9,$76,$DF,$F8,$ED,$AF,$E6
	.dc.b	$3E,$9D,$F3,$7F,$E9,$D6,$C7,$F9,$1D,$AF,$F1,$76,$BF,$97,$6B,$FE
	.dc.b	$5A,$F8,$7F,$E1,$3C,$5F,$8C,$6D,$E7,$B7,$FD,$55,$1F,$D6,$AB,$FD
	.dc.b	$6C,$E9,$B1,$A3,$7E,$4F,$FA,$C6,$3F,$D5,$D7,$F9,$86,$DF,$F9,$A9
	.dc.b	$A7,$FB,$07,$E5,$3F,$C2,$B3,$BF,$85,$E4,$49,$FF,$2F,$4E,$AB,$FD
	.dc.b	$9D,$F4,$1F,$CF,$A7,$FC,$E1,$C6,$8C,$47,$F1,$68,$C7,$8D,$FD,$AB
	.dc.b	$E0,$1F,$39,$FD,$B2,$F7,$9F,$F8,$F9,$B8,$F6,$BE,$15,$31,$FF,$15
	.dc.b	$9F,$EE,$BC,$39,$B8,$C5,$18,$8F,$F1,$8C,$62,$7F,$CA,$DF,$4B,$FE
	.dc.b	$C4,$CB,$FC,$61,$DA,$F9,$5F,$F1,$96,$0A,$FE,$57,$39,$8F,$F1,$B4
	.dc.b	$62,$76,$F3,$8F,$E3,$D1,$8F,$89,$FF,$0F,$46,$0E,$3E,$1F,$6B,$E9
	.dc.b	$1D,$E3,$15,$C7,$0F,$2B,$67,$E0,$E1,$DD,$FF,$26,$B5,$D1,$8F,$05
	.dc.b	$C1,$F2,$23,$D9,$18,$79,$B5,$E1,$1F,$E7,$B5,$D1,$48,$FF,$9F,$DB
	.dc.b	$97,$AA,$1F,$1D,$FE,$34,$8C,$47,$23,$E1,$1F,$E5,$F8,$D6,$3F,$72
	.dc.b	$BD,$F3,$F4,$A7,$2F,$F9,$AA,$32,$F1,$E1,$DE,$FF,$80,$E8,$FF,$21
	.dc.b	$CF,$DA,$8C,$4F,$E1,$78,$7F,$76,$DA,$DD,$7E,$73,$A7,$FC,$97,$BF
	.dc.b	$E4,$DB,$5F,$E7,$5C,$39,$D7,$E7,$79,$3F,$85,$CF,$E8,$9B,$19,$AE
	.dc.b	$2F,$E5,$F0,$B7,$FA,$79,$CF,$5B,$E9,$5C,$98,$6E,$5F,$EC,$F7,$FB
	.dc.b	$4B,$63,$44,$75,$C5,$72,$FE,$E3,$F4,$AF,$8E,$EE,$DA,$FF,$6D,$2F
	.dc.b	$95,$CC,$C9,$CB,$C2,$31,$5F,$DF,$25,$EE,$7C,$A8,$67,$FE,$7C,$8C
	.dc.b	$36,$7F,$DD,$C7,$03,$FA,$07,$CE,$FC,$2E,$33,$F1,$23,$57,$FD,$26
	.dc.b	$3E,$3F,$85,$88,$FF,$02,$35,$FF,$55,$5A,$31,$1E,$46,$C6,$70,$F2
	.dc.b	$C6,$08,$FA,$39,$5F,$2D,$F5,$9B,$8F,$F5,$D9,$C4,$FD,$AE,$27,$D2
	.dc.b	$FC,$06,$D7,$F7,$38,$C1,$1C,$13,$2F,$EE,$D1,$F2,$3C,$8F,$F6,$8C
	.dc.b	$FF,$B1,$79,$FF,$C2,$71,$8A,$FF,$01,$57,$E2,$A3,$13,$FE,$18,$F2
	.dc.b	$BF,$CB,$97,$FE,$20,$F0,$FE,$9C,$63,$0F,$F3,$73,$5F,$F1,$73,$5F
	.dc.b	$96,$D7,$EE,$B7,$63,$7B,$3F,$DF,$77,$E7,$3F,$C6,$9F,$11,$F9,$E9
	.dc.b	$F0,$63,$FE,$2C,$AD,$E1,$FE,$46,$F3,$1F,$3F,$F8,$86,$E3,$FD,$03
	.dc.b	$24,$62,$BF,$D0,$CD,$7F,$C8,$CD,$7F,$06,$D7,$F2,$AD,$8E,$75,$F9
	.dc.b	$6F,$3D,$F2,$1F,$B4,$6E,$1F,$E5,$C9,$C4,$FE,$E3,$D1,$FE,$3F,$3E
	.dc.b	$56,$E3,$FE,$99,$6B,$FE,$59,$6B,$F9,$66,$BF,$A8,$6E,$06,$98,$F4
	.dc.b	$FF,$DC,$F5,$FE,$98,$3C,$5C,$BF,$C2,$A3,$37,$FE,$70,$7F,$3E,$4E
	.dc.b	$5F,$D9,$B6,$BF,$D3,$91,$82,$3F,$3F,$CF,$78,$FF,$C5,$9B,$1F,$EA
	.dc.b	$89,$C4,$FF,$1A,$F5,$FF,$B5,$4F,$E0,$70,$FE,$07,$0F,$9C,$D7,$8F
	.dc.b	$B7,$F9,$0F,$B3,$E8,$AF,$F3,$F3,$5F,$DD,$B5,$FE,$1A,$DD,$9F,$4B
	.dc.b	$FC,$67,$1F,$04,$63,$C0,$FE,$5B,$2F,$9B,$B1,$FC,$D2,$BF,$C3,$1D
	.dc.b	$3F,$C6,$5A,$7F,$83,$8B,$8F,$F8,$10,$38,$7F,$C0,$ED,$46,$BC,$9F
	.dc.b	$CE,$BA,$6B,$A8,$E0,$FF,$28,$F8,$BF,$D2,$C7,$F1,$A6,$C7,$3A,$FE
	.dc.b	$8F,$18,$C9,$F3,$8F,$A4,$FE,$96,$BE,$96,$C6,$1B,$8F,$C6,$9D,$36
	.dc.b	$74,$E7,$E1,$79,$9D,$3D,$8B,$19,$26,$7F,$EB,$EB,$FF,$3E,$C6,$3B
	.dc.b	$FF,$B9,$FC,$13,$E3,$BF,$B1,$F5,$BE,$97,$FD,$E9,$7F,$87,$C5,$7B
	.dc.b	$8F,$C5,$56,$FE,$07,$FB,$5D,$6B,$0E,$78,$72,$8F,$F7,$29,$9B,$E3
	.dc.b	$19,$C7,$EC,$CF,$F8,$5A,$31,$3F,$D2,$32,$7D,$2F,$4C,$7E,$5B,$F3
	.dc.b	$3D,$49,$5F,$27,$E9,$6E,$3F,$06,$6E,$5F,$EC,$8C,$55,$8B,$ED,$EC
	.dc.b	$FD,$6F,$82,$FA,$ED,$8F,$A8,$DC,$7E,$63,$70,$F4,$FD,$1F,$E3,$B0
	.dc.b	$FE,$E0,$DB,$FF,$32,$B6,$FF,$30,$D7,$FC,$BA,$D7,$F2,$ED,$4D,$31
	.dc.b	$CB,$F8,$EF,$83,$FC,$69,$FE,$1C,$8E,$D8,$7B,$E3,$F2,$3E,$3C,$C3
	.dc.b	$F1,$DE,$95,$C7,$F0,$D3,$6F,$E3,$4D,$AF,$A5,$F4,$BF,$9A,$3F,$C4
	.dc.b	$FF,$4A,$C7,$13,$F9,$99,$B8,$FF,$A3,$EB,$53,$93,$F8,$96,$5F,$E6
	.dc.b	$2F,$85,$F8,$FE,$F7,$FB,$BA,$7E,$3B,$6F,$E3,$B5,$F8,$ED,$7E,$3B
	.dc.b	$5F,$8E,$DA,$ED,$3F,$D5,$F8,$7F,$90,$3F,$AF,$F1,$3E,$45,$77,$D7
	.dc.b	$F0,$D1,$FB,$8F,$A5,$95,$FB,$5F,$1A,$E5,$FA,$D9,$B7,$F8,$D9,$B5
	.dc.b	$F9,$9F,$87,$FE,$02,$7F,$02,$F9,$3B,$1F,$BF,$8B,$87,$ED,$E6,$DF
	.dc.b	$E5,$FC,$CF,$F1,$BE,$86,$93,$0F,$67,$6F,$F0,$E3,$F8,$7C,$58,$FE
	.dc.b	$19,$87,$F1,$23,$07,$F0,$53,$0E,$8F,$E2,$53,$88,$5F,$ED,$B9,$5F
	.dc.b	$3C,$F8,$5F,$F2,$5B,$70,$FF,$02,$36,$FF,$9B,$57,$0A,$DF,$93,$FE
	.dc.b	$6E,$AF,$E8,$57,$FD,$19,$E1,$87,$77,$38,$78,$A7,$FC,$97,$1F,$D3
	.dc.b	$63,$AB,$5F,$F6,$3E,$1E,$6F,$27,$CA,$79,$BD,$DF,$01,$E1,$6D,$F8
	.dc.b	$62,$D7,$73,$FC,$D3,$7D,$F0,$F2,$4E,$9B,$5E,$9B,$C3,$E6,$FA,$5F
	.dc.b	$8A,$47,$E2,$BF,$23,$DC,$F8,$47,$C3,$6C,$7C,$3F,$A5,$F9,$A4,$7E
	.dc.b	$6B,$72,$F8,$2D,$7E,$0E,$48,$FA,$BF,$33,$F3,$CF,$A0,$DA,$FF,$02
	.dc.b	$7B,$CF,$A5,$7C,$B7,$FA,$26,$30,$D8,$CC,$FF,$BF,$5B,$8F,$FB,$F1
	.dc.b	$AF,$FA,$FD,$AF,$FA,$DD,$AA,$BA,$CF,$E2,$3E,$43,$F0,$87,$FA,$D2
	.dc.b	$BA,$ED,$F8,$38,$CC,$FE,$1D,$D6,$18,$C3,$EC,$61,$EB,$F9,$7F,$E9
	.dc.b	$B7,$E3,$6C,$F8,$DA,$F8,$F1,$46,$2B,$95,$7F,$81,$6B,$83,$5E,$13
	.dc.b	$BA,$E3,$39,$AE,$35,$FE,$F8,$F6,$FF,$33,$F2,$3F,$D8,$3E,$17,$FC
	.dc.b	$BB,$3C,$AF,$FC,$2F,$18,$E4,$FD,$7D,$6E,$FF,$C3,$76,$FA,$57,$C7
	.dc.b	$FF,$BF,$A3,$B1,$FE,$20,$8C,$F4,$F4,$AF,$F1,$1F,$C8,$3D,$BF,$B3
	.dc.b	$9F,$8B,$0E,$98,$F3,$8C,$63,$CA,$7A,$9A,$C7,$A2,$C3,$A7,$07,$9E
	.dc.b	$3C,$58,$73,$CB,$D3,$87,$39,$FF,$1D,$35,$CC,$FA,$DC,$B5,$E3,$F2
	.dc.b	$CF,$95,$FF,$21,$47,$BE,$30,$E2,$91,$88,$EE,$D7,$9C,$FA,$9F,$99
	.dc.b	$1C,$B3,$F8,$2D,$C3,$E4,$FC,$1F,$FB,$D6,$B8,$61,$DB,$0B,$7B,$9F
	.dc.b	$24,$7C,$6C,$D7,$C6,$7E,$BF,$0B,$E9,$1D,$BF,$7B,$4F,$C1,$75,$3E
	.dc.b	$97,$1B,$7F,$22,$6E,$3F,$AB,$C3,$8B,$F5,$31,$E9,$8F,$F9,$92,$3F
	.dc.b	$CC,$71,$87,$F2,$D3,$C1,$FF,$19,$9C,$E7,$83,$4D,$B8,$9E,$51,$C1
	.dc.b	$BC,$BF,$C4,$BE,$3B,$F2,$FD,$5F,$CE,$DB,$1F,$CE,$9B,$5E,$9F,$06
	.dc.b	$F4,$7F,$4E,$9B,$A9,$A7,$0C,$7E,$45,$5C,$FF,$AA,$61,$F8,$49,$CB
	.dc.b	$63,$2F,$FA,$26,$32,$D8,$CB,$FA,$E8,$CB,$63,$2F,$CA,$6B,$FD,$1B
	.dc.b	$E3,$FC,$E4,$7F,$AE,$7C,$27,$A8,$FC,$BB,$FD,$31,$B9,$7F,$6C,$6E
	.dc.b	$5F,$EB,$E9,$FE,$D7,$18,$23,$83,$FB,$7F,$C0,$FC,$62,$3F,$DA,$FE
	.dc.b	$7F,$E4,$9D,$08,$FF,$76,$7B,$FD,$2F,$CE,$24,$7F,$73,$E2,$AF,$F6
	.dc.b	$3C,$7F,$C0,$33,$8E,$0F,$EE,$D5,$78,$3F,$BA,$CD,$BF,$ED,$33,$F6
	.dc.b	$8F,$69,$1F,$F0,$3B,$F0,$3D,$EF,$BE,$48,$FF,$02,$7C,$77,$E0,$8F
	.dc.b	$8A,$47,$FC,$1E,$FC,$8E,$8F,$C3,$92,$3F,$DA,$3F,$08,$FA,$5C,$47
	.dc.b	$AF,$83,$F8,$26,$9F,$E0,$FB,$FC,$8D,$EE,$B8,$DE,$08,$F4,$FA,$5D
	.dc.b	$9F,$6A,$4D,$24,$7F,$16,$D9,$2F,$96,$EE,$49,$F3,$FC,$B5,$7E,$84
	.dc.b	$ED,$F5,$CF,$9A,$7C,$FF,$4B,$27,$DE,$FA,$BE,$6F,$11,$CA,$2E,$9F
	.dc.b	$E3,$BC,$15,$F0,$3E,$96,$5F,$C8,$7F,$6A,$BE,$95,$CB,$86,$C7,$E7
	.dc.b	$1A,$FF,$34,$6B,$FE,$F7,$6B,$FD,$69,$AF,$FB,$D9,$AF,$78,$FD,$3B
	.dc.b	$4F,$95,$6F,$A5,$FA,$E4,$8F,$F0,$DF,$9C,$E8,$C4,$7C,$A9,$B9,$FF
	.dc.b	$B4,$3B,$17,$BA,$FC,$3F,$C3,$FE,$08,$8F,$E2,$BE,$26,$BF,$52,$2D
	.dc.b	$FC,$B9,$FC,$4F,$D2,$CA,$FC,$8B,$F2,$BE,$95,$EC,$D3,$DA,$E2,$7F
	.dc.b	$82,$B5,$FE,$BA,$D7,$FC,$02,$1C,$3F,$E0,$46,$A3,$51,$D8,$FC,$B7
	.dc.b	$D2,$FE,$7B,$E5,$24,$FE,$9A,$3F,$42,$61,$FC,$8C,$61,$FD,$5F,$A7
	.dc.b	$FA,$62,$3F,$AE,$F8,$2A,$FE,$62,$2E,$7F,$8F,$9F,$C5,$FD,$27,$F8
	.dc.b	$62,$BF,$6B,$E8,$FE,$C0,$CF,$A3,$F9,$0F,$A5,$7C,$8E,$2B,$D9,$8F
	.dc.b	$F5,$6C,$39,$47,$F9,$DC,$57,$07,$7E,$3F,$F8,$A2,$31,$1B,$9E,$35
	.dc.b	$CF,$0F,$06,$1F,$C0,$99,$5F,$C2,$9F,$E0,$BE,$96,$23,$F6,$B1,$6B
	.dc.b	$F5,$93,$FD,$57,$E9,$7F,$86,$E5,$FD,$C1,$78,$D9,$35,$E2,$CC,$7F
	.dc.b	$CE,$73,$A6,$CE,$5F,$E6,$5E,$3F,$F8,$97,$1F,$F5,$A3,$70,$FC,$07
	.dc.b	$88,$8C,$57,$E0,$FE,$97,$A3,$F1,$DA,$FE,$31,$AF,$F9,$39,$AF,$1C
	.dc.b	$3F,$67,$18,$83,$49,$1F,$D7,$3D,$77,$D2,$7F,$7C,$6D,$7D,$41,$87
	.dc.b	$F9,$4E,$75,$F3,$FF,$E0,$2F,$9A,$7C,$6F,$F6,$66,$DF,$F2,$B6,$BF
	.dc.b	$EC,$E6,$BF,$EC,$2F,$84,$7B,$BF,$E0,$C5,$EF,$D3,$11,$FE,$10,$F9
	.dc.b	$B7,$85,$49,$7E,$7D,$12,$CF,$81,$FD,$E9,$C4,$FF,$3F,$6B,$FE,$E2
	.dc.b	$6B,$FE,$D6,$6B,$F6,$1A,$FC,$46,$CF,$82,$7F,$C0,$FF,$38,$F8,$8F
	.dc.b	$F1,$3B,$6F,$CC,$D7,$FB,$C7,$BD,$FE,$0B,$6F,$AF,$FE,$10,$8C,$47
	.dc.b	$FB,$A1,$AF,$E2,$9A,$FD,$66,$E1,$FE,$5A,$9E,$44,$9F,$EB,$DF,$3A
	.dc.b	$F9,$5F,$F1,$DB,$63,$E5,$35,$FF,$06,$7C,$F7,$F8,$F5,$F9,$CD,$7F
	.dc.b	$8F,$DA,$31,$1F,$EF,$E6,$BF,$A0,$6B,$F8,$C6,$C7,$1F,$7B,$FC,$5F
	.dc.b	$18,$33,$F3,$FF,$EC,$0A,$FB,$CD,$7F,$C4,$0D,$EB,$75,$FE,$BB,$A7
	.dc.b	$F2,$31,$8A,$FC,$DC,$7E,$49,$D7,$3F,$F8,$FA,$31,$1A,$8C,$11,$FF
	.dc.b	$03,$46,$B2,$FE,$0F,$3F,$E4,$65,$E2,$7F,$99,$9B,$5F,$95,$6B,$FE
	.dc.b	$37,$F3,$5D,$1F,$E6,$C6,$C7,$F8,$4D,$AF,$CE,$6B,$FE,$CE,$9F,$F7
	.dc.b	$B4,$FF,$95,$A7,$33,$FC,$39,$AF,$EE,$9B,$3C,$23,$FC,$41,$F4,$3F
	.dc.b	$E0,$E7,$F5,$4D,$7F,$C9,$8D,$CB,$E5,$4B,$E0,$DF,$39,$FE,$84,$6D
	.dc.b	$FF,$8A,$DA,$FE,$19,$AF,$A5,$AF,$F2,$46,$BF,$C4,$1B,$3D,$E7,$FC
	.dc.b	$E7,$18,$D5,$EC,$FF,$A5,$1B,$7F,$C0,$DA,$FF,$99,$7E,$95,$F3,$7F
	.dc.b	$E9,$A6,$C7,$F9,$05,$AF,$E7,$1A,$FC,$F6,$BF,$CE,$BE,$1F,$F8,$4E
	.dc.b	$7F,$11,$B7,$9E,$BF,$F5,$3C,$7F,$5A,$8C,$57,$F1,$78,$DB,$FC,$B9
	.dc.b	$BA,$7F,$BD,$FC,$8F,$F5,$7C,$FF,$A9,$30,$FA,$2D,$BF,$E9,$6D,$3F
	.dc.b	$9B,$FC,$A7,$F8,$4A,$77,$F0,$DC,$89,$3F,$C9,$E7,$55,$FE,$D0,$FA
	.dc.b	$0F,$E8,$32,$BF,$F1,$94,$61,$BE,$2F,$FA,$F7,$E0,$1F,$39,$FE,$BF
	.dc.b	$AD,$61,$CE,$B0,$6E,$66,$1F,$EC,$1B,$E2,$D8,$FF,$58,$F2,$FF,$9C
	.dc.b	$5F,$F2,$6F,$D2,$BE,$23,$C9,$5F,$C3,$FE,$57,$FC,$54,$A7,$F9,$57
	.dc.b	$72,$73,$18,$6D,$72,$9D,$57,$F8,$CE,$BB,$6B,$FE,$35,$9C,$CF,$5F
	.dc.b	$0B,$BC,$62,$BA,$4F,$1B,$E9,$87,$6C,$3F,$D9,$75,$D2,$BF,$C8,$A7
	.dc.b	$9E,$BD,$6F,$F9,$22,$34,$DF,$01,$C1,$C9,$B1,$CA,$7F,$DE,$13,$93
	.dc.b	$FC,$E7,$1F,$B5,$3F,$CE,$BD,$38,$23,$83,$E5,$1F,$2B,$E3,$BF,$93
	.dc.b	$67,$F0,$23,$FC,$AD,$3F,$0F,$3F,$F2,$ED,$69,$F5,$C6,$67,$97,$0D
	.dc.b	$7E,$AC,$55,$5C,$56,$EF,$B0,$7C,$DF,$F3,$79,$FF,$7F,$46,$30,$F9
	.dc.b	$35,$BA,$D5,$7F,$9A,$E3,$13,$AE,$4F,$E8,$6F,$F8,$6E,$11,$89,$FC
	.dc.b	$F4,$F1,$BE,$D1,$83,$CB,$3B,$E0,$FD,$1D,$7F,$4B,$6B,$D6,$B8,$C6
	.dc.b	$2B,$FD,$E9,$F4,$CF,$90,$E1,$7F,$E4,$28,$C3,$E5,$FC,$5F,$FD,$7E
	.dc.b	$FF,$86,$E3,$35,$CA,$B8,$BF,$E3,$48,$CD,$FF,$A9,$EF,$8C,$EE,$FE
	.dc.b	$07,$E2,$EF,$82,$E3,$A2,$31,$3F,$E3,$FA,$7E,$0B,$FD,$6A,$38,$DF
	.dc.b	$0A,$DD,$7C,$5B,$E5,$3E,$48,$E3,$E3,$6E,$3C,$F9,$3F,$38,$FE,$0F
	.dc.b	$8F,$E8,$A3,$1D,$5F,$6B,$0F,$E2,$15,$A7,$F8,$D3,$5F,$DF,$E2,$FB
	.dc.b	$8F,$F3,$BE,$1F,$75,$B3,$F6,$70,$EB,$39,$F2,$BC,$78,$3E,$93,$C1
	.dc.b	$5E,$56,$C6,$6B,$CC,$DB,$F1,$CE,$89,$3A,$7D,$5E,$17,$F6,$5F,$9F
	.dc.b	$FA,$ED,$9F,$F3,$66,$1D,$DF,$F2,$C7,$42,$BE,$AF,$EA,$3F,$DB,$27
	.dc.b	$F3,$ED,$C7,$FD,$9B,$AA,$31,$1F,$11,$AF,$E8,$DA,$FE,$B9,$B1,$C7
	.dc.b	$CA,$FA,$0D,$AF,$E2,$BE,$33,$F5,$2F,$C6,$C7,$FB,$5E,$31,$D1,$E6
	.dc.b	$BF,$C9,$C6,$27,$FD,$10,$3F,$39,$FF,$0E,$BF,$C8,$0C,$79,$DF,$7D
	.dc.b	$A6,$63,$79,$FF,$BA,$A3,$2D,$8C,$BF,$D0,$23,$2D,$8C,$BF,$EA,$EF
	.dc.b	$B7,$79,$3F,$49,$B5,$E6,$AE,$FC,$1F,$E1,$C9,$CD,$FE,$77,$67,$D6
	.dc.b	$7C,$66,$E1,$F5,$23,$04,$70,$49,$FA,$F3,$C1,$FF,$33,$FA,$AF,$9A
	.dc.b	$FC,$2B,$6B,$F8,$B5,$6F,$1F,$F1,$7C,$E6,$FF,$8A,$57,$F2,$6C,$FE
	.dc.b	$57,$85,$E8,$9F,$88,$D4,$7F,$99,$1C,$36,$38,$92,$7E,$B4,$EB,$97
	.dc.b	$FD,$8B,$F2,$FF,$9A,$F3,$4F,$F1,$2A,$F1,$1F,$EB,$18,$CB,$FC,$2F
	.dc.b	$C0,$B8,$FF,$60,$6B,$FD,$E0,$FE,$1F,$5F,$C0,$5B,$3F,$D9,$9A,$71
	.dc.b	$8D,$4F,$28,$FF,$00,$B5,$FE,$97,$EA,$3E,$43,$F8,$FB,$6F,$C9,$78
	.dc.b	$9F,$E3,$1E,$AB,$E4,$E9,$FC,$6F,$17,$CC,$63,$0F,$E4,$D4,$F9,$CF
	.dc.b	$43,$63,$D0,$DC,$FF,$BF,$B5,$F8,$1F,$9F,$FF,$2D,$7F,$A7,$07,$1C
	.dc.b	$5F,$D8,$7F,$14,$7C,$4F,$F2,$D9,$CE,$1F,$B3,$FC,$C1,$F4,$DF,$CA
	.dc.b	$7B,$51,$89,$FD,$DB,$72,$FA,$13,$AA,$E1,$39,$9F,$F1,$5C,$E6,$B8
	.dc.b	$D6,$8F,$E0,$FF,$69,$FE,$C9,$8F,$ED,$8D,$AF,$ED,$15,$C5,$FE,$8D
	.dc.b	$19,$CF,$FD,$13,$F3,$5F,$EF,$5B,$FF,$7C,$36,$BF,$3B,$E8,$BE,$6B
	.dc.b	$F9,$4C,$E6,$F8,$FB,$B8,$3F,$25,$5B,$C3,$9C,$70,$E8,$91,$FC,$52
	.dc.b	$5F,$11,$8C,$BF,$BF,$75,$32,$61,$9C,$BF,$D9,$18,$FF,$82,$30,$C3
	.dc.b	$FA,$CF,$55,$F5,$EF,$96,$FE,$EB,$18,$7F,$B2,$7C,$7B,$E5,$73,$5B
	.dc.b	$BF,$DA,$E3,$FE,$D5,$C3,$84,$60,$ED,$3B,$D7,$F0,$F8,$73,$9D,$47
	.dc.b	$F8,$B2,$75,$7F,$3A,$7E,$49,$E0,$C6,$BC,$35,$60,$FB,$2F,$EA,$ED
	.dc.b	$9E,$BF,$59,$EF,$78,$46,$72,$F9,$2D,$C7,$FD,$8A,$D9,$FE,$23,$5C
	.dc.b	$6B,$77,$DB,$C7,$95,$DA,$7F,$CF,$0F,$68,$CC,$76,$7F,$C4,$35,$FC
	.dc.b	$73,$DC,$CE,$3F,$EF,$DF,$6D,$F8,$CF,$F2,$7C,$3F,$D0,$B1,$8F,$2B
	.dc.b	$FC,$1C,$DC,$BF,$CE,$6D,$9F,$E2,$8D,$78,$D7,$37,$F9,$5C,$65,$B1
	.dc.b	$95,$F6,$8F,$C8,$FF,$D9,$F7,$F5,$A3,$1F,$13,$FE,$3A,$C3,$C1,$92
	.dc.b	$BF,$9B,$D6,$9F,$9F,$F6,$9F,$50,$F8,$38,$BE,$D3,$13,$A2,$47,$F9
	.dc.b	$67,$6F,$E8,$33,$A8,$FA,$FE,$9F,$DE,$3E,$5F,$C2,$7E,$82,$FE,$EC
	.dc.b	$65,$A7,$E0,$A3,$53,$CA,$3F,$C3,$7F,$2F,$FE,$F2,$7F,$C7,$8D,$8F
	.dc.b	$F6,$93,$5F,$F5,$BE,$68,$C4,$7E,$89,$AF,$FB,$45,$B1,$BF,$8D,$FF
	.dc.b	$1B,$C6,$B5,$FF,$60,$7D,$6F,$CB,$F8,$83,$FE,$50,$9F,$15,$FE,$16
	.dc.b	$73,$8F,$E6,$BE,$6F,$F3,$E7,$C2,$F9,$6B,$19,$26,$7F,$2F,$44,$EA
	.dc.b	$3F,$43,$B7,$EA,$0F,$97,$F1,$D6,$FF,$4F,$16,$79,$4D,$CB,$FD,$3D
	.dc.b	$F9,$0B,$EA,$BF,$B1,$37,$0F,$EC,$4D,$BF,$F6,$4B,$5F,$ED,$5E,$76
	.dc.b	$6F,$FA,$45,$FF,$9C,$EB,$17,$FE,$E6,$AB,$7F,$EB,$0F,$C6,$7F,$85
	.dc.b	$3F,$E7,$B6,$E1,$CA,$BF,$D5,$11,$87,$FD,$5C,$D9,$E5,$E1,$FF,$A0
	.dc.b	$6F,$AF,$C0,$AB,$4B,$F0,$57,$C9,$FF,$94,$1B,$8F,$FA,$F1,$B8,$FF
	.dc.b	$97,$9A,$FF,$BF,$BC,$3F,$E2,$FF,$0D,$FC,$58,$FF,$4D,$CF,$0D,$3F
	.dc.b	$88,$C7,$FA,$7B,$F1,$9F,$F6,$7E,$1F,$EC,$86,$E1,$FE,$6A,$6B,$FE
	.dc.b	$8C,$6B,$FE,$49,$F2,$7F,$C9,$07,$3F,$26,$D7,$F2,$9F,$81,$FE,$96
	.dc.b	$D9,$D3,$49,$5F,$D3,$A7,$6F,$FA,$A6,$FC,$78,$FF,$94,$6F,$FD,$6B
	.dc.b	$EF,$FF,$A6,$AB,$FD,$3E,$D7,$FD,$BC,$D7,$FD,$D2,$DD,$3F,$9D,$F8
	.dc.b	$5F,$CE,$8E,$7F,$13,$73,$FF,$71,$7B,$4F,$A5,$FF,$40,$B6,$3F,$DD
	.dc.b	$AD,$C7,$FD,$1E,$D7,$FD,$62,$D7,$FD,$E0,$D7,$FD,$F6,$DE,$2F,$F1
	.dc.b	$4D,$7F,$A9,$3C,$CF,$F0,$15,$FF,$95,$32,$FE,$A3,$F3,$3F,$ED,$4C
	.dc.b	$3C,$1B,$7F,$4A,$8F,$F6,$BF,$D7,$7E,$E3,$3F,$F5,$0B,$5F,$F6,$1B
	.dc.b	$5F,$F2,$FE,$68,$C4,$23,$1E,$17,$FA,$82,$73,$18,$CB,$FC,$A9,$5B
	.dc.b	$9E,$31,$FC,$03,$CF,$FF,$30,$77,$FF,$6E,$8E,$F3,$E5,$CB,$FC,$0F
	.dc.b	$39,$BF,$E7,$8D,$AE,$47,$F7,$8C,$3C,$13,$FE,$7C,$9E,$0B,$E7,$1E
	.dc.b	$17,$FB,$E3,$F1,$9F,$F2,$D4,$BE,$37,$F8,$43,$FE,$7E,$AD,$E3,$D6
	.dc.b	$B7,$5A,$AE,$97,$CF,$E0,$FF,$A8,$47,$3F,$0B,$FD,$0E,$EB,$0E,$53
	.dc.b	$AB,$EF,$9B,$1E,$B8,$7D,$06,$E1,$E3,$C7,$B6,$3F,$E9,$2C,$3C,$5F
	.dc.b	$61,$FC,$0B,$B3,$18,$F8,$23,$18,$78,$C7,$B8,$FA,$2F,$E2,$F8,$72
	.dc.b	$F0,$58,$F2,$BF,$F4,$FE,$1C,$AB,$53,$9C,$39,$78,$9E,$BC,$BC,$18
	.dc.b	$F2,$C7,$C0,$31,$F1,$BF,$E9,$76,$BC,$30,$FF,$4B,$46,$23,$C3,$F4
	.dc.b	$9F,$E7,$87,$F1,$70,$8C,$47,$36,$BF,$05,$AF,$2F,$82,$F6,$3E,$38
	.dc.b	$E9,$9F,$CA,$6E,$3F,$EF,$5F,$8A,$7D,$A7,$DA,$AD,$61,$CB,$1F,$13
	.dc.b	$70,$E3,$8F,$CE,$C3,$8B,$F2,$F8,$BE,$F7,$D5,$F3,$CF,$EF,$36,$7F
	.dc.b	$59,$C6,$8C,$47,$C9,$F9,$6F,$EF,$FC,$7F,$75,$B1,$C2,$79,$61,$DA
	.dc.b	$3C,$D3,$A7,$91,$F9,$B9,$DD,$F7,$8D,$9F,$E6,$2B,$7C,$B7,$E7,$FA
	.dc.b	$7F,$CF,$2D,$8F,$F7,$19,$AC,$BC,$7C,$2F,$CE,$3F,$DE,$98,$73,$BE
	.dc.b	$0F,$F8,$D5,$B7,$C6,$3F,$CE,$35,$C7,$DB,$E7,$7C,$A7,$FD,$BD,$39
	.dc.b	$7F,$86,$B5,$CC,$6D,$FD,$8C,$65,$B1,$97,$FD,$20,$D7,$F4,$2D,$F0
	.dc.b	$7F,$CF,$78,$7C,$E6,$CF,$F9,$E2,$3E,$8E,$CF,$C5,$DE,$3E,$1B,$5F
	.dc.b	$C8,$A7,$53,$FE,$F3,$6C,$66,$76,$FF,$BE,$E3,$2D,$8C,$BF,$C3,$23
	.dc.b	$2D,$8C,$BF,$D0,$FC,$47,$C7,$FA,$5B,$1F,$E9,$08,$F4,$FE,$2C,$F3
	.dc.b	$69,$E5,$71,$3E,$96,$BE,$56,$BF,$E0,$BF,$37,$F6,$7C,$5F,$C8,$B5
	.dc.b	$46,$38,$7F,$AB,$36,$BF,$A5,$7E,$0E,$FB,$3F,$9D,$39,$BE,$3D,$58
	.dc.b	$3E,$7D,$6E,$FF,$C1,$11,$96,$9F,$D0,$E3,$4B,$1F,$E6,$BB,$A7,$33
	.dc.b	$F2,$DA,$FF,$60,$6B,$E6,$C7,$FA,$DA,$C3,$FD,$7F,$F7,$9F,$34,$91
	.dc.b	$FD,$4F,$E3,$07,$E1,$8A,$FE,$0D,$AF,$DF,$6B,$F7,$5A,$FF,$91,$B4
	.dc.b	$FC,$7F,$A7,$FD,$3A,$3F,$C9,$D1,$FB,$DF,$D5,$5F,$79,$A8,$93,$FE
	.dc.b	$50,$75,$12,$3F,$DB,$E2,$78,$B8,$6B,$F2,$67,$FA,$DF,$E6,$2F,$9D
	.dc.b	$FE,$18,$D9,$F8,$AD,$7F,$C0,$2D,$7F,$B1,$87,$0F,$E5,$FE,$FF,$F2
	.dc.b	$C4,$7C,$AF,$F1,$27,$FD,$CE,$E5,$FF,$73,$BF,$C7,$4D,$3F,$EE,$AF
	.dc.b	$99,$FD,$01,$1F,$7F,$DB,$BE,$5F,$11,$B3,$FD,$E7,$3F,$C9,$E7,$50
	.dc.b	$3E,$11,$F8,$BF,$C7,$3F,$9E,$B8,$E4,$46,$1F,$F0,$19,$8F,$27,$A7
	.dc.b	$5C,$AF,$C5,$DA,$3F,$1B,$08,$C3,$A8,$CC,$62,$3F,$C5,$AD,$7E,$6B
	.dc.b	$5F,$F0,$8B,$5F,$CC,$47,$DC,$AF,$DB,$37,$5F,$E1,$1B,$FF,$47,$FC
	.dc.b	$59,$F6,$DF,$C2,$62,$61,$F0,$F0,$E3,$1B,$9F,$E1,$DF,$1D,$2B,$58
	.dc.b	$7F,$13,$8F,$EA,$5F,$71,$FC,$4A,$7E,$5F,$3F,$C1,$9E,$11,$8C,$D1
	.dc.b	$C3,$BF,$FA,$FF,$C6,$BF,$23,$7A,$38,$6E,$1E,$01,$3B,$9F,$EC,$5F
	.dc.b	$73,$F5,$74,$FF,$1A,$BF,$6A,$BF,$2A,$D7,$3C,$5F,$52,$31,$1F,$83
	.dc.b	$6A,$69,$23,$F8,$3E,$9F,$77,$A3,$F9,$91,$2B,$F6,$35,$FC,$7F,$5F
	.dc.b	$D4,$70,$FD,$4A,$D5,$7E,$40,$ED,$E0,$7F,$3B,$F7,$A4,$7B,$F1,$FC
	.dc.b	$54,$7F,$7F,$FB,$8F,$B3,$39,$C7,$FD,$2C,$D9,$E2,$F5,$9D,$BF,$47
	.dc.b	$D9,$FE,$90,$47,$F8,$3F,$E2,$CF,$7F,$FC,$18,$D8,$CC,$60,$FF,$08
	.dc.b	$7E,$2C,$F9,$BC,$4E,$88,$7B,$63,$4F,$9F,$E9,$BF,$AC,$11,$FE,$2F
	.dc.b	$F1,$EF,$83,$C4,$7E,$1E,$6F,$0F,$F2,$5A,$D4,$FC,$1F,$4B,$FB,$11
	.dc.b	$1E,$7F,$0B,$5F,$C8,$45,$AE,$53,$FD,$FB,$E6,$1F,$0D,$49,$7F,$CA
	.dc.b	$BE,$C5,$F5,$9F,$E2,$F6,$C7,$F4,$86,$BF,$EA,$46,$BF,$C9,$1A,$FF
	.dc.b	$A2,$1A,$FF,$11,$6A,$69,$8F,$39,$F6,$5F,$E3,$F6,$CF,$F3,$E6,$BF
	.dc.b	$8B,$FC,$FB,$FD,$46,$FF,$42,$FC,$DD,$F7,$5F,$C4,$9B,$1F,$D8,$1A
	.dc.b	$FF,$AF,$1A,$FF,$3C,$FC,$DF,$FD,$5C,$FF,$4C,$6B,$F9,$96,$ED,$FD
	.dc.b	$67,$F3,$7F,$F0,$FB,$FD,$D3,$E4,$3F,$8E,$37,$B3,$FC,$5B,$18,$F0
	.dc.b	$BF,$D5,$7E,$93,$D5,$FE,$C0,$7F,$28,$71,$97,$F7,$66,$BF,$ED,$26
	.dc.b	$BF,$D4,$9A,$F4,$8E,$44,$7F,$DB,$3E,$03,$EE,$9F,$39,$FE,$43,$6C
	.dc.b	$FE,$C5,$AF,$ED,$FE,$99,$F2,$1F,$E7,$16,$C7,$F8,$29,$AF,$FB,$99
	.dc.b	$AF,$F6,$66,$BF,$EC,$A6,$BF,$D4,$DA,$9A,$6A,$FD,$1F,$03,$E7,$24
	.dc.b	$FF,$B9,$FD,$47,$D2,$7F,$96,$9B,$1F,$E5,$76,$BF,$E5,$36,$BF,$E5
	.dc.b	$16,$BF,$D9,$7D,$DF,$EC,$33,$FE,$5D,$9C,$CA,$72,$F5,$9F,$E4,$B6
	.dc.b	$F8,$8F,$F3,$73,$67,$F8,$E3,$5F,$E4,$5F,$94,$FF,$0F,$6F,$8D,$FD
	.dc.b	$A2,$31,$1F,$E0,$E6,$BF,$EF,$26,$BF,$DE,$DA,$9A,$48,$FF,$1F,$FC
	.dc.b	$7B,$F6,$77,$CE,$7F,$AC,$23,$0D,$8C,$DF,$CD,$F9,$1F,$98,$7C,$32
	.dc.b	$3F,$35,$AF,$CD,$6B,$F3,$5F,$F4,$37,$86,$F9,$4F,$E2,$44,$9F,$F1
	.dc.b	$CE,$5F,$D2,$39,$FD,$DE,$8F,$C3,$24,$FF,$96,$27,$FC,$5E,$FC,$18
	.dc.b	$CB,$FE,$34,$3E,$13,$87,$FD,$1D,$FA,$13,$E0,$7F,$D3,$ED,$8F,$ED
	.dc.b	$DC,$48,$C4,$7F,$88,$FF,$4D,$FE,$FD,$DD,$FE,$22,$7F,$C0,$11,$FE
	.dc.b	$2B,$3F,$CE,$8E,$1F,$F0,$63,$5C,$FB,$5F,$EA,$7F,$17,$FD,$23,$5F
	.dc.b	$E7,$A6,$E9,$F9,$16,$D7,$F8,$DF,$B9,$F4,$FF,$EB,$69,$CE,$7F,$E8
	.dc.b	$FF,$8A,$C1,$FE,$69,$9F,$F5,$F5,$FF,$9D,$31,$FF,$4D,$37,$1F,$F2
	.dc.b	$17,$3F,$F9,$0F,$E1,$FF,$D8,$3F,$8B,$3E,$17,$8C,$4B,$FE,$E8,$71
	.dc.b	$3F,$45,$3B,$7E,$4F,$D6,$7F,$74,$7F,$C9,$BA,$23,$19,$7F,$A6,$1B
	.dc.b	$97,$EB,$FE,$EF,$FC,$71,$7F,$AD,$9C,$B7,$DB,$FF,$6A,$C6,$30,$57
	.dc.b	$06,$FC,$2F,$FB,$4E,$3F,$CF,$F7,$FE,$CD,$6B,$DE,$3F,$CA,$BF,$49
	.dc.b	$FD,$47,$2F,$F3,$27,$DE,$7F,$8D,$AB,$FC,$CD,$3F,$D3,$A3,$10,$8C
	.dc.b	$7C,$2F,$E0,$5E,$50,$9D,$C6,$1B,$E9,$7E,$1D,$B3,$E0,$AE,$5E,$9F
	.dc.b	$FA,$F6,$39,$46,$30,$FF,$5B,$E1,$DA,$7F,$D2,$2D,$74,$7E,$85,$7D
	.dc.b	$EB,$E9,$B8,$B6,$74,$D7,$9B,$5F,$D5,$7C,$E7,$F9,$E9,$34,$8F,$F9
	.dc.b	$F9,$FF,$46,$7B,$3F,$1A,$1E,$D3,$8F,$39,$F4,$95,$6E,$B5,$18,$9D
	.dc.b	$4E,$6B,$8C,$FF,$AC,$2E,$BF,$E9,$BB,$E3,$5E,$18,$F3,$9C,$CF,$DB
	.dc.b	$46,$AF,$A9,$FE,$BA,$71,$1F,$CF,$78,$3F,$45,$87,$F0,$BE,$1F,$C7
	.dc.b	$F3,$7F,$B5,$EB,$F3,$0D,$78,$36,$37,$3C,$6B,$9D,$7F,$B2,$27,$2D
	.dc.b	$8C,$BF,$C2,$E1,$F7,$AE,$BE,$8D,$BF,$23,$6B,$A3,$5E,$4D,$7F,$AE
	.dc.b	$35,$FE,$B4,$DA,$EF,$5E,$B9,$7D,$A7,$F9,$0A,$FF,$96,$46,$3D,$BD
	.dc.b	$4E,$63,$F9,$83,$5F,$F7,$2B,$5F,$E4,$CD,$7F,$DA,$4D,$7A,$4F,$68
	.dc.b	$FF,$49,$37,$E1,$BB,$91,$FE,$7D,$C6,$FB,$B7,$A3,$F4,$E3,$0D,$8C
	.dc.b	$CF,$06,$E1,$FD,$1F,$EA,$BF,$C6,$53,$EE,$6C,$67,$D7,$F8,$91,$88
	.dc.b	$FF,$80,$2D,$7F,$A1,$35,$FF,$77,$35,$39,$C3,$D1,$E1,$79,$49,$7E
	.dc.b	$57,$E9,$7C,$47,$FB,$12,$3E,$8B,$6F,$FA,$F3,$5F,$EA,$41,$C3,$FD
	.dc.b	$55,$AF,$FC,$01,$6B,$BF,$8A,$EE,$7E,$67,$EE,$3F,$C8,$B5,$FA,$16
	.dc.b	$CF,$E6,$5A,$FE,$5D,$AF,$7F,$B8,$F4,$92,$7D,$2D,$8F,$4B,$5F,$4D
	.dc.b	$A3,$F5,$92,$FB,$97,$C3,$61,$B1,$FD,$FC,$D4,$64,$FC,$D7,$DF,$7E
	.dc.b	$7F,$03,$49,$C9,$A2,$57,$E7,$A3,$2F,$C1,$F2,$7E,$21,$1C,$1F,$A4
	.dc.b	$3F,$97,$97,$BB,$F5,$4E,$2F,$CD,$18,$24,$6C,$91,$F6,$9F,$EE,$5F
	.dc.b	$39,$F0,$BA,$CD,$24,$7E,$D3,$BF,$FC,$A0,$DF,$67,$F7,$51,$8B,$FE
	.dc.b	$77,$18,$24,$EC,$91,$FB,$AD,$BF,$7A,$48,$FD,$A3,$FC,$03,$E1,$E4
	.dc.b	$7E,$DD,$B3,$FE,$31,$FB,$97,$DD,$7F,$0B,$6D,$7F,$0A,$6C,$EE,$7F
	.dc.b	$88,$7D,$77,$E6,$6B,$F8,$7B,$67,$FC,$78,$D7,$F3,$ED,$7F,$C6,$5F
	.dc.b	$51,$F1,$A7,$F8,$BB,$63,$FC,$88,$D7,$F4,$ED,$78,$E9,$FC,$86,$31
	.dc.b	$00,$E1,$B1,$FD,$07,$DD,$7D,$CB,$D1,$C3,$70,$FF,$2A,$4E,$A7,$27
	.dc.b	$E7,$BE,$E7,$FD,$41,$3F,$9B,$6E,$1F,$83,$6B,$FE,$4F,$84,$62,$3C
	.dc.b	$D1,$8F,$83,$E8,$47,$A3,$DF,$E1,$E4,$77,$6C,$F7,$6B,$DE,$91,$88
	.dc.b	$FF,$30,$7C,$47,$F0,$A2,$76,$FF,$49,$6D,$FF,$48,$6C,$FF,$9D,$5F
	.dc.b	$E2,$7E,$53,$EE,$1F,$5D,$FD,$AE,$30,$D8,$CD,$FF,$9F,$67,$2F,$F9
	.dc.b	$FF,$ED,$7F,$C0,$4D,$C3,$FC,$22,$D9,$FE,$ED,$3A,$9F,$E4,$ED,$7F
	.dc.b	$CF,$4D,$7F,$00,$D4,$D3,$F0,$9E,$77,$48,$34,$95,$E3,$6C,$78,$0D
	.dc.b	$C6,$4F,$ED,$BF,$72,$F9,$0F,$EE,$4D,$8C,$C6,$08,$E0,$93,$FE,$A1
	.dc.b	$9E,$0F,$F5,$6F,$B9,$7D,$4F,$F8,$5F,$C9,$46,$2F,$FB,$13,$6B,$85
	.dc.b	$7B,$7E,$26,$57,$CD,$6C,$FA,$DA,$FC,$A3,$FD,$9B,$1B,$3E,$B7,$AB
	.dc.b	$FE,$F6,$6F,$C5,$FD,$48,$C4,$FD,$46,$BF,$51,$AF,$D4,$6B,$F5,$1B
	.dc.b	$AF,$F9,$7E,$7F,$C3,$5F,$84,$3E,$E6,$93,$0F,$F0,$DC,$65,$B1,$9C
	.dc.b	$3F,$D9,$53,$96,$C7,$F8,$37,$E5,$9F,$73,$49,$7F,$E1,$48,$CB,$F0
	.dc.b	$3E,$47,$DC,$47,$04,$BF,$81,$19,$6C,$7C,$8F,$B9,$F8,$C4,$70,$48
	.dc.b	$F9,$11,$96,$C7,$F8,$E3,$EE,$5D,$1F,$E3,$D6,$C6,$63,$06,$7E,$BE
	.dc.b	$47,$16,$D7,$F3,$27,$F9,$B9,$23,$F9,$BB,$63,$51,$82,$3F,$EF,$5A
	.dc.b	$7E,$17,$E3,$24,$F9,$DF,$F3,$57,$DD,$3E,$4E,$92,$7F,$CE,$91,$97
	.dc.b	$DD,$E2,$7B,$48,$E0,$FF,$3D,$7D,$D7,$F0,$4C,$7F,$CE,$4D,$7F,$CE
	.dc.b	$0D,$7F,$A0,$7B,$E8,$C4,$23,$1D,$BF,$E7,$F9,$CC,$62,$BB,$CE,$48
	.dc.b	$E0,$95,$FE,$38,$1F,$45,$FE,$47,$27,$B7,$FE,$73,$7F,$C9,$97,$FE
	.dc.b	$02,$F8,$3F,$F4,$76,$5F,$EA,$AC,$3F,$D2,$CD,$C3,$FD,$1C,$D7,$FB
	.dc.b	$3F,$DE,$FF,$6E,$24,$FF,$6E,$6C,$7F,$6E,$6B,$C2,$3F,$D4,$5E,$7B
	.dc.b	$E2,$FF,$D3,$CD,$8C,$C6,$0F,$47,$DD,$C8,$FF,$44,$E6,$8C,$5F,$C1
	.dc.b	$8C,$4E,$A7,$33,$FE,$5C,$9C,$BF,$EA,$9F,$BB,$7D,$17,$FA,$BD,$B7
	.dc.b	$FE,$B7,$6C,$EE,$74,$49,$D3,$63,$C5,$F7,$7F,$3C,$93,$E8,$6C,$6A
	.dc.b	$30,$47,$03,$E0,$BF,$DB,$7B,$BE,$DD,$F3,$5F,$ED,$EF,$69,$D5,$FE
	.dc.b	$0D,$C3,$F0,$39,$7F,$B8,$B1,$E3,$87,$0C,$BF,$C2,$58,$70,$BF,$F0
	.dc.b	$A7,$DD,$FE,$97,$9B,$39,$73,$8C,$4F,$FB,$E2,$3F,$C3,$5E,$A7,$F8
	.dc.b	$6B,$D0,$F1,$46,$FA,$FF,$DE,$F5,$B3,$AB,$8C,$7C,$10,$C3,$AC,$78
	.dc.b	$F0,$E9,$97,$8A,$B8,$E5,$5E,$18,$76,$7F,$D2,$F8,$79,$9B,$19,$7C
	.dc.b	$93,$FE,$D0,$97,$DD,$FF,$A7,$67,$F5,$11,$8F,$84,$D4,$E6,$7A,$B5
	.dc.b	$F8,$8F,$1C,$BD,$75,$C6,$7E,$01,$ED,$9D,$67,$F2,$3E,$ED,$F5,$9F
	.dc.b	$5F,$0E,$35,$AB,$FA,$6D,$79,$63,$F0,$F5,$61,$CA,$7E,$26,$4F,$7F
	.dc.b	$F0,$5C,$3F,$80,$FB,$BE,$DA,$FF,$12,$E2,$46,$30,$FC,$0D,$F0,$E1
	.dc.b	$FC,$03,$FE,$F6,$8F,$AC,$DC,$7B,$47,$CB,$8C,$11,$C1,$0F,$F6,$5F
	.dc.b	$03,$EE,$FF,$C1,$31,$FE,$31,$6B,$A6,$BF,$9B,$8E,$1E,$87,$F1,$D8
	.dc.b	$46,$1E,$51,$98,$C4,$73,$8F,$47,$E1,$32,$3F,$04,$D7,$F0,$2D,$7F
	.dc.b	$16,$D7,$F9,$CF,$DE,$7F,$AA,$1F,$E5,$DD,$28,$C4,$7F,$BA,$23,$1E
	.dc.b	$0F,$9D,$B1,$E9,$F9,$CF,$98,$9A,$47,$E8,$35,$E0,$60,$E9,$F7,$8F
	.dc.b	$8E,$A4,$8F,$2C,$65,$FD,$07,$C3,$7D,$92,$38,$3D,$7F,$76,$F8,$CF
	.dc.b	$43,$63,$31,$82,$38,$21,$F0,$63,$83,$63,$6D,$36,$7F,$17,$1F,$9C
	.dc.b	$3E,$2F,$F2,$0D,$8F,$C0,$35,$FE,$BC,$D7,$F9,$06,$7F,$EA,$7F,$85
	.dc.b	$FA,$07,$FA,$A6,$3E,$BB,$5F,$AC,$D7,$C5,$2F,$C3,$3C,$24,$8F,$0B
	.dc.b	$F5,$3E,$76,$47,$DB,$6B,$FD,$C7,$64,$62,$3D,$D1,$87,$81,$F9,$BF
	.dc.b	$BC,$7D,$37,$EB,$DB,$19,$8C,$19,$F8,$49,$1C,$5A,$FF,$80,$7E,$99
	.dc.b	$F2,$FF,$B9,$6C,$7E,$A5,$AE,$E3,$44,$74,$D7,$FC,$19,$F7,$6F,$BC
	.dc.b	$FE,$08,$D8,$FD,$A3,$5D,$C6,$88,$E9,$AF,$4F,$6B,$99,$1E,$6F,$F1
	.dc.b	$1F,$C8,$3F,$8B,$C7,$F0,$16,$BF,$E1,$F6,$BF,$8B,$6B,$FD,$8D,$AF
	.dc.b	$FB,$C5,$A9,$A6,$3E,$F5,$F7,$BF,$C8,$5B,$3F,$C4,$5A,$FF,$8B,$DA
	.dc.b	$FE,$6D,$AF,$F7,$36,$BF,$F0,$05,$A9,$BF,$7A,$FA,$AF,$E6,$0D,$9F
	.dc.b	$E4,$0D,$7F,$C7,$CD,$7F,$54,$D7,$FC,$04,$D4,$1C,$36,$3F,$9F,$7C
	.dc.b	$FF,$F4,$B8,$FF,$25,$35,$FD,$B3,$5F,$F0,$73,$5E,$BE,$87,$E0,$5F
	.dc.b	$DE,$35,$FF,$0A,$35,$DE,$7F,$D3,$63,$11,$FE,$75,$8C,$11,$FE,$81
	.dc.b	$3F,$DC,$5F,$0F,$9B,$E8,$23,$A6,$BE,$A6,$BF,$C3,$FE,$EF,$E8,$93
	.dc.b	$3F,$EB,$CD,$8F,$E9,$CD,$7F,$CD,$8F,$D8,$FB,$B7,$D9,$7F,$65,$6C
	.dc.b	$FF,$55,$6B,$FE,$73,$6B,$FC,$7F,$EE,$9F,$77,$FD,$B5,$B1,$FD,$75
	.dc.b	$AF,$F9,$ED,$AF,$18,$E0,$47,$83,$FD,$DF,$E2,$3F,$CB,$71,$FD,$9D
	.dc.b	$AF,$FA,$15,$AF,$F3,$16,$BF,$E4,$96,$BF,$88,$6A,$69,$3E,$07,$C0
	.dc.b	$48,$F0,$35,$F0,$35,$F0,$3F,$E0,$DF,$AA,$7D,$3D,$24,$FF,$7F,$8C
	.dc.b	$BF,$E0,$1F,$DA,$3F,$A0,$3F,$E5,$F6,$BF,$9E,$6A,$69,$3C,$3F,$F0
	.dc.b	$E3,$63,$FC,$19,$F7,$8F,$77,$FC,$40,$D8,$CC,$60,$FF,$12,$7D,$DF
	.dc.b	$FA,$BA,$3F,$C2,$8D,$7F,$D6,$4D,$7F,$AF,$35,$FF,$3B,$7D,$DF,$F8
	.dc.b	$21,$FE,$C4,$D7,$3E,$6F,$F8,$C6,$31,$1F,$BA,$F0,$DF,$11,$FE,$06
	.dc.b	$24,$7F,$83,$5F,$2E,$12,$3C,$AD,$7C,$AD,$7C,$AD,$7F,$D1,$5F,$3C
	.dc.b	$F8,$FF,$F2,$13,$63,$FC,$64,$D7,$FD,$A0,$D7,$FB,$E7,$E7,$7E,$84
	.dc.b	$D2,$65,$FE,$4D,$6C,$7F,$8F,$5A,$FF,$8A,$FE,$E9,$F2,$BF,$E2,$B6
	.dc.b	$C7,$F8,$C9,$AE,$E3,$FC,$B1,$FB,$4F,$EC,$23,$FD,$48,$D7,$F9,$1B
	.dc.b	$53,$49,$D3,$C8,$91,$FE,$51,$FB,$9E,$04,$8F,$F3,$33,$CF,$E8,$FE
	.dc.b	$AC,$FF,$95,$DA,$FF,$BD,$5A,$FF,$88,$5A,$FF,$AE,$5A,$FF,$38,$6A
	.dc.b	$69,$AF,$45,$8B,$C2,$F8,$C4,$AF,$8C,$FF,$9E,$FF,$39,$FE,$17,$1F
	.dc.b	$EC,$76,$BF,$D1,$9A,$9A,$4E,$CE,$44,$8F,$89,$E6,$FC,$22,$38,$3F
	.dc.b	$CF,$12,$F5,$7C,$44,$8C,$B5,$FF,$22,$1E,$17,$0F,$F9,$1E,$39,$3A
	.dc.b	$F8,$37,$DC,$FF,$5F,$6F,$91,$FE,$87,$6C,$EE,$7F,$B0,$F4,$3E,$67
	.dc.b	$FC,$DE,$49,$FF,$3B,$3F,$E7,$2F,$AB,$91,$FE,$79,$6B,$FE,$3F,$6B
	.dc.b	$FE,$3E,$6B,$FD,$C3,$77,$DC,$3E,$B3,$FD,$08,$D9,$FF,$45,$B5,$FF
	.dc.b	$45,$35,$FF,$2E,$B5,$E8,$D5,$3D,$A7,$A9,$FE,$91,$FB,$4F,$F2,$84
	.dc.b	$FF,$A5,$5A,$FF,$9A,$1A,$F9,$DA,$F4,$6C,$9A,$4F,$5F,$FD,$8D,$F6
	.dc.b	$AF,$15,$B7,$F8,$43,$C7,$7A,$FF,$E1,$2C,$96,$FB,$7C,$56,$A7,$FC
	.dc.b	$2F,$39,$79,$57,$18,$ED,$78,$9D,$E1,$E0,$57,$CD,$EF,$87,$6C,$39
	.dc.b	$36,$38,$DF,$37,$CB,$7C,$EF,$96,$1E,$39,$EB,$87,$27,$84,$79,$1C
	.dc.b	$9D,$0F,$CD,$B9,$7F,$D4,$0D,$7D,$95,$E9,$73,$E4,$3E,$8B,$FD,$C1
	.dc.b	$5F,$EA,$56,$C6,$67,$AB,$5F,$F7,$37,$C8,$7F,$B3,$88,$FF,$BE,$DB
	.dc.b	$8F,$C5,$8F,$F7,$5F,$C5,$7C,$17,$FC,$8F,$37,$E1,$D3,$99,$4E,$63
	.dc.b	$0D,$8F,$82,$7D,$F7,$15,$DE,$38,$9A,$3C,$27,$C8,$E0,$FC,$77,$0F
	.dc.b	$82,$3F,$30,$D7,$F3,$0D,$7F,$30,$D7,$F3,$0D,$8E,$77,$FE,$F8,$E1
	.dc.b	$FD,$93,$67,$5F,$6D,$FE,$E1,$8F,$F7,$EB,$6F,$FD,$F8,$D7,$FD,$9E
	.dc.b	$D7,$F7,$CD,$7F,$66,$DB,$34,$D4,$3E,$DB,$FC,$43,$5F,$C8,$23,$1E
	.dc.b	$7F,$AD,$B1,$99,$DC,$7F,$26,$6B,$FE,$EB,$6B,$FA,$F6,$BF,$E8,$26
	.dc.b	$BE,$1F,$5B,$B4,$1A,$49,$FE,$5D,$F6,$DF,$EC,$C6,$FB,$DF,$3F,$95
	.dc.b	$39,$C3,$FD,$F9,$18,$9F,$57,$D4,$7F,$86,$2F,$DC,$DA,$FE,$7E,$D7
	.dc.b	$F8,$30,$70,$FF,$09,$6B,$FE,$9C,$6B,$BF,$3F,$B9,$E5,$C1,$F1,$14
	.dc.b	$97,$EC,$7E,$77,$B4,$F9,$6A,$49,$FA,$91,$96,$C6,$70,$FE,$AF,$39
	.dc.b	$6C,$7E,$47,$ED,$9F,$7D,$49,$7F,$D7,$B8,$93,$A8,$EF,$E3,$79,$08
	.dc.b	$E0,$98,$78,$23,$2D,$8F,$ED,$23,$ED,$DF,$6F,$FA,$58,$C3,$63,$31
	.dc.b	$C1,$B3,$FD,$CB,$ED,$FF,$2E,$4F,$8C,$FC,$A3,$6B,$F3,$0D,$9E,$13
	.dc.b	$F0,$7C,$59,$5E,$FC,$91,$8A,$F3,$F7,$7E,$7E,$8D,$23,$FA,$53,$25
	.dc.b	$73,$1C,$A3,$E0,$F5,$3D,$5F,$80,$D8,$CC,$60,$FD,$77,$73,$E4,$E9
	.dc.b	$27,$F9,$5B,$6F,$FC,$2B,$3B,$9C,$92,$3F,$69,$BF,$EB,$09,$1F,$B9
	.dc.b	$7F,$5F,$F4,$52,$3F,$78,$D9,$FF,$10,$FD,$BB,$ED,$FF,$81,$36,$BF
	.dc.b	$83,$36,$77,$3F,$C1,$FE,$DD,$CF,$49,$7F,$C3,$63,$2F,$F0,$FF,$B7
	.dc.b	$F8,$97,$FE,$30,$6B,$FB,$26,$BF,$DE,$5A,$83,$86,$EB,$FC,$66,$30
	.dc.b	$70,$F9,$EF,$F2,$D7,$C9,$FF,$37,$6D,$FF,$90,$1B,$7F,$C0,$9A,$FF
	.dc.b	$82,$61,$F6,$EF,$7F,$F3,$8D,$8F,$D0,$B6,$7F,$85,$C7,$09,$D9,$1D
	.dc.b	$B6,$37,$2F,$B7,$EA,$4A,$EA,$D9,$DB,$5F,$34,$F1,$9F,$E7,$9F,$21
	.dc.b	$FC,$16,$7F,$8E,$B5,$FF,$12,$35,$F9,$8D,$4D,$27,$77,$F4,$57,$87
	.dc.b	$96,$FB,$67,$DB,$FF,$63,$8C,$36,$33,$87,$F9,$BA,$72,$FF,$9C,$3E
	.dc.b	$CF,$EE,$47,$F6,$B6,$C6,$71,$FE,$EE,$7F,$8E,$E3,$2F,$F3,$86,$B9
	.dc.b	$36,$7E,$1D,$CB,$4C,$87,$F1,$9E,$97,$86,$3F,$C9,$33,$E1,$8C,$11
	.dc.b	$C1,$23,$C4,$70,$3F,$D1,$7F,$6C,$FB,$7F,$ED,$8D,$9F,$EE,$0D,$9D
	.dc.b	$CE,$89,$3A,$6C,$7F,$56,$F2,$5F,$6E,$FA,$0F,$F0,$94,$61,$B1,$9B
	.dc.b	$FE,$BE,$DA,$F3,$FD,$BF,$BC,$93,$F3,$DB,$1A,$8C,$11,$FA,$1E,$AF
	.dc.b	$FB,$9A,$3E,$D3,$63,$3F,$11,$F6,$E3,$11,$F6,$5A,$FD,$56,$BF,$4D
	.dc.b	$BA,$FF,$94,$E7,$FC,$29,$F6,$FF,$EA,$9B,$FF,$0C,$36,$FF,$D7,$2D
	.dc.b	$7F,$BD,$B5,$FF,$41,$35,$FE,$14,$D4,$D3,$5E,$8B,$E4,$5F,$6F,$FE
	.dc.b	$11,$6D,$FF,$B0,$DB,$7F,$E1,$26,$BE,$FF,$8D,$F8,$44,$9F,$8E,$D8
	.dc.b	$F9,$6D,$78,$47,$CC,$FB,$7F,$94,$47,$04,$8F,$A7,$19,$6C,$7F,$8F
	.dc.b	$7E,$DF,$F7,$6D,$E8,$FF,$1F,$46,$27,$FD,$C2,$D7,$FC,$46,$BE,$AB
	.dc.b	$64,$AE,$8F,$F4,$12,$3F,$CF,$DB,$3F,$EE,$CF,$0B,$94,$E4,$91,$99
	.dc.b	$F1,$FD,$BC,$9F,$43,$67,$FC,$6B,$E9,$BE,$DD,$F1,$9F,$E6,$F6,$DF
	.dc.b	$F8,$75,$B5,$C1,$AF,$B3,$C6,$F8,$24,$9F,$8C,$FF,$9E,$7E,$DF,$F8
	.dc.b	$2E,$06,$93,$BB,$44,$AF,$F1,$64,$65,$B1,$9B,$F4,$CE,$5F,$F2,$78
	.dc.b	$FB,$4F,$F2,$59,$3D,$BF,$F3,$93,$FE,$50,$C3,$FC,$19,$F0,$5F,$E8
	.dc.b	$DC,$BF,$D9,$F9,$7F,$8E,$DB,$87,$C6,$6B,$FE,$5C,$DD,$F6,$FF,$DD
	.dc.b	$C9,$5F,$DD,$5B,$3F,$5D,$AF,$17,$FD,$39,$E7,$BE,$DE,$92,$BF,$C9
	.dc.b	$91,$97,$FD,$19,$9B,$ED,$FD,$64,$8C,$B7,$0F,$87,$3B,$9C,$92,$3F
	.dc.b	$03,$F6,$EF,$B7,$FF,$57,$36,$7F,$CB,$8D,$9D,$CE,$89,$3A,$6C,$78
	.dc.b	$3E,$6B,$E7,$92,$7D,$2D,$8D,$46,$08,$E0,$76,$FF,$B2,$78,$1F,$66
	.dc.b	$FB,$7F,$F6,$67,$B1,$FA,$BB,$FF,$0D,$68,$CB,$FD,$A1,$8F,$1C,$38
	.dc.b	$65,$FE,$E4,$EA,$FF,$10,$60,$FB,$3F,$D0,$F4,$66,$3F,$C4,$91,$87
	.dc.b	$FD,$CF,$3F,$E2,$7F,$4F,$FC,$4F,$E8,$78,$A3,$9D,$6F,$5F,$25,$F5
	.dc.b	$EC,$63,$D6,$30,$F7,$C7,$9E,$59,$C3,$96,$3D,$30,$E3,$87,$E7,$A7
	.dc.b	$C8,$FF,$AB,$9A,$EF,$0F,$F6,$94,$3E,$DB,$FC,$97,$3C,$5A,$F1,$6B
	.dc.b	$EA,$6B,$F0,$9E,$39,$7B,$27,$A7,$AF,$E7,$9C,$E3,$F1,$7E,$D9,$F6
	.dc.b	$FF,$3B,$0E,$15,$AC,$7A,$61,$C4,$F8,$57,$C2,$3E,$17,$BB,$F5,$B8
	.dc.b	$7E,$D7,$DB,$F3,$CF,$FC,$B1,$CD,$F8,$48,$C5,$FD,$8B,$E1,$C1,$F5
	.dc.b	$1F,$F7,$A4,$7D,$F6,$ED,$FE,$87,$8F,$8F,$18,$23,$82,$1F,$ED,$CD
	.dc.b	$DF,$6F,$FE,$07,$8F,$F1,$93,$5D,$35,$FC,$7C,$70,$F4,$7F,$8B,$C7
	.dc.b	$E1,$27,$4F,$F8,$E1,$AE,$63,$73,$C6,$3E,$EF,$DB,$FA,$88,$E0,$87
	.dc.b	$E1,$DC,$B4,$FF,$6C,$7D,$BF,$FA,$B5,$BE,$17,$93,$54,$E6,$7F,$A3
	.dc.b	$B5,$FF,$7E,$FC,$CF,$CC,$4D,$23,$F4,$1A,$FC,$06,$BF,$D4,$BE,$DF
	.dc.b	$FD,$FA,$FF,$BA,$DA,$E7,$E3,$7C,$91,$88,$FD,$07,$C2,$7D,$A2,$3F
	.dc.b	$A0,$7D,$5F,$6E,$F8,$AA,$49,$FF,$7C,$46,$5B,$19,$7F,$B2,$C6,$5B
	.dc.b	$19,$7F,$84,$E4,$9D,$46,$1A,$FE,$59,$AF,$FA,$BF,$ED,$FF,$DC,$AF
	.dc.b	$E0,$9A,$FF,$5A,$6B,$FC,$58,$38,$7F,$43,$F0,$83,$EC,$15,$FE,$C7
	.dc.b	$59,$8C,$9F,$A2,$97,$DB,$D2,$57,$8D,$FB,$1F,$37,$23,$EC,$35,$FE
	.dc.b	$DF,$AA,$31,$1F,$84,$6B,$FC,$BB,$ED,$DF,$45,$FB,$26,$C7,$E8,$9A
	.dc.b	$FF,$7A,$6B,$F0,$57,$C1,$F1,$24,$71,$7F,$7B,$F6,$EF,$96,$A4,$9F
	.dc.b	$D7,$C6,$5B,$19,$7F,$C1,$31,$96,$C7,$F0,$8F,$B7,$7D,$A5,$24,$FE
	.dc.b	$EA,$32,$D8,$CB,$FE,$14,$8C,$B6,$3A,$7B,$52,$39,$35,$FF,$0D,$FD
	.dc.b	$BB,$C0,$FE,$32,$D8,$FE,$12,$D7,$71,$A2,$3A,$6B,$FE,$26,$FB,$77
	.dc.b	$DB,$7F,$24,$6C,$7F,$14,$6B,$B8,$D1,$1D,$35,$FF,$18,$FD,$BB,$EA
	.dc.b	$3F,$97,$B6,$3F,$8F,$B5,$DC,$68,$8E,$9A,$FF,$8F,$3E,$7B,$FA,$23
	.dc.b	$7C,$CF,$E7,$71,$88,$FD,$33,$5F,$EF,$BE,$6F,$D0,$7F,$54,$D7,$3A
	.dc.b	$FF,$44,$8C,$40,$FF,$31,$C6,$48,$FF,$38,$8D,$A9,$FE,$CE,$F1,$6B
	.dc.b	$9F,$2F,$CE,$47,$FA,$0C,$68,$91,$FD,$5B,$ED,$FE,$8C,$FF,$01,$6B
	.dc.b	$FE,$18,$6B,$E7,$6A,$69,$33,$E8,$7F,$62,$FB,$77,$D8,$52,$57,$F5
	.dc.b	$48,$CB,$FD,$5B,$ED,$DF,$5D,$FD,$A9,$B1,$98,$C1,$1C,$10,$FF,$3A
	.dc.b	$C7,$07,$FC,$EF,$F1,$1F,$E5,$17,$FB,$1B,$5C,$FB,$5F,$DE,$23,$11
	.dc.b	$FC,$8D,$AF,$F8,$E9,$AF,$D9,$6B,$FE,$CF,$8E,$5F,$03,$23,$BB,$5F
	.dc.b	$1B,$5E,$CD,$7F,$C8,$FF,$52,$FA,$6F,$F0,$5B,$63,$FB,$DB,$5F,$F4
	.dc.b	$9B,$5F,$E7,$5F,$6F,$FC,$E5,$FC,$73,$53,$49,$F3,$FF,$E1,$86,$C7
	.dc.b	$F8,$21,$E1,$F6,$FF,$E5,$F8,$FF,$05,$B5,$FF,$51,$B5,$FE,$94,$D7
	.dc.b	$FC,$C3,$F6,$FF,$EA,$77,$FC,$26,$D7,$FD,$56,$D7,$FA,$93,$5F,$F3
	.dc.b	$4F,$DB,$FF,$80,$DB,$E8,$7F,$8B,$A3,$11,$FA,$B6,$BF,$E0,$6F,$89
	.dc.b	$FF,$02,$11,$FF,$02,$3F,$E3,$6A,$F3,$47,$EC,$DA,$9A,$4F,$81,$E0
	.dc.b	$48,$F5,$BF,$E4,$0F,$9E,$7C,$85,$24,$FF,$8C,$A3,$2F,$F8,$D3,$ED
	.dc.b	$FC,$E4,$CB,$FC,$92,$D8,$D4,$60,$8F,$F7,$0F,$B4,$7C,$AF,$F8,$AD
	.dc.b	$B1,$98,$C1,$FE,$54,$FB,$7F,$D7,$C7,$FB,$69,$AF,$F8,$05,$AF,$FA
	.dc.b	$61,$AF,$F1,$76,$A6,$93,$A7,$FC,$C3,$F6,$FA,$23,$83,$FC,$CB,$F6
	.dc.b	$ED,$E9,$2F,$FC,$AF,$19,$6C,$65,$FF,$78,$C6,$5B,$1F,$07,$C0,$91
	.dc.b	$F1,$5A,$FF,$BE,$3E,$DF,$F8,$23,$FE,$1F,$6B,$FE,$B6,$6B,$FC,$D5
	.dc.b	$A9,$A4,$EC,$FF,$3F,$BA,$C7,$FC,$17,$3F,$EC,$06,$BF,$CF,$DA,$9B
	.dc.b	$0F,$63,$99,$27,$FC,$F3,$19,$6C,$65,$F6,$35,$FF,$4D,$7D,$BF,$F6
	.dc.b	$16,$F8,$DF,$E8,$2F,$85,$4E,$67,$FA,$94,$62,$7F,$CD,$5F,$6F,$27
	.dc.b	$FC,$D4,$D7,$FC,$67,$F5,$7F,$E6,$F2,$3F,$E6,$F6,$BA,$30,$43,$FC
	.dc.b	$9D,$ED,$3E,$DD,$F5,$9F,$E8,$36,$CF,$F9,$31,$AF,$06,$BF,$E6,$46
	.dc.b	$BF,$EB,$66,$A9,$ED,$3D,$4F,$F4,$87,$DA,$FF,$92,$E7,$FC,$B2,$D7
	.dc.b	$CA,$D7,$FC,$E2,$D7,$FD,$86,$D7,$B3,$64,$D3,$50,$FB,$67,$84,$D3
	.dc.b	$FC,$05,$E2,$7F,$94,$2B,$FC,$09,$4F,$B7,$FF,$03,$42,$74,$F3,$8C
	.dc.b	$C6,$27,$AD,$74,$F6,$38,$C6,$2F,$94,$62,$F8,$B6,$38,$1E,$2C,$3C
	.dc.b	$CF,$03,$17,$A7,$C4,$5B,$F4,$36,$78,$E1,$E5,$79,$C7,$FA,$3E,$7C
	.dc.b	$8D,$8C,$BE,$86,$BF,$13,$C5,$7D,$1F,$FB,$72,$91,$89,$FF,$4D,$1F
	.dc.b	$3D,$C3,$DD,$AE,$BE,$43,$FD,$98,$47,$FD,$F2,$DF,$B8,$F9,$F0,$F8
	.dc.b	$97,$C7,$7F,$8E,$67,$32,$AC,$6B,$FE,$F9,$AB,$3F,$EA,$D9,$AF,$F9
	.dc.b	$06,$F9,$AB,$E7,$3A,$8F,$05,$57,$B1,$F4,$CE,$4F,$D3,$7F,$C9,$91
	.dc.b	$6B,$AE,$9D,$27,$B6,$1E,$56,$9E,$52,$9D,$0A,$74,$2C,$FA,$30,$E8
	.dc.b	$7C,$67,$13,$FE,$5D,$FB,$A7,$D4,$3E,$BB,$DF,$C1,$F0,$27,$31,$8D
	.dc.b	$38,$53,$EA,$B1,$9A,$F7,$5F,$C6,$0D,$DE,$A3,$33,$FA,$D9,$E7,$59
	.dc.b	$BF,$14,$71,$9E,$9F,$80,$7F,$B7,$63,$16,$8C,$4F,$33,$7E,$3F,$78
	.dc.b	$CE,$5E,$B6,$DF,$C1,$6D,$70,$8E,$B1,$8F,$97,$79,$4F,$83,$FC,$1D
	.dc.b	$FF,$00,$9C,$C6,$23,$FD,$11,$18,$BF,$CF,$B6,$3F,$DF,$59,$7E,$4F
	.dc.b	$BB,$E2,$71,$FF,$A9,$EF,$E0,$5E,$F3,$FA,$53,$9C,$3D,$D2,$FD,$67
	.dc.b	$E0,$C7,$8D,$AF,$89,$AA,$38,$7C,$5F,$9E,$31,$3A,$9C,$E1,$E8,$8C
	.dc.b	$1C,$B2,$FD,$77,$AB,$F6,$F9,$0F,$E0,$AF,$8E,$1F,$4E,$B7,$87,$B3
	.dc.b	$AD,$F1,$1E,$1B,$FE,$54,$D8,$CC,$FC,$86,$DF,$F2,$E3,$D4,$E1,$FA
	.dc.b	$4D,$7E,$8B,$5D,$CF,$BE,$FF,$9B,$7C,$47,$E4,$FA,$4F,$A6,$FB,$EF
	.dc.b	$FA,$19,$CE,$3F,$53,$F2,$CF,$EC,$6F,$DC,$6E,$5E,$5C,$D1,$88,$FC
	.dc.b	$33,$5F,$C2,$B5,$FA,$8D,$8E,$3E,$0F,$C5,$8D,$67,$FC,$47,$2F,$D5
	.dc.b	$75,$FF,$51,$C7,$C9,$78,$9F,$A5,$F0,$FF,$EB,$69,$FC,$8B,$72,$F3
	.dc.b	$35,$FC,$D3,$5F,$CC,$B5,$FC,$43,$72,$34,$D6,$2F,$98,$7D,$9F,$EF
	.dc.b	$DB,$87,$E7,$2B,$C5,$B7,$F0,$D9,$CE,$3F,$9C,$FC,$21,$F4,$9F,$85
	.dc.b	$6E,$1F,$AA,$6C,$FC,$F6,$BF,$B3,$6B,$FB,$26,$BF,$A1,$6E,$67,$39
	.dc.b	$BF,$0D,$FF,$3B,$DF,$F0,$E6,$DF,$97,$0F,$E3,$51,$8C,$BF,$A2,$E2
	.dc.b	$FB,$77,$D6,$FE,$75,$B8,$FF,$02,$6E,$3F,$87,$6B,$FC,$2D,$AF,$F0
	.dc.b	$A6,$BF,$B9,$6D,$77,$AF,$CE,$7D,$D7,$F8,$62,$FF,$92,$B6,$FF,$83
	.dc.b	$DF,$83,$0F,$E7,$93,$97,$F8,$DF,$DC,$BD,$4F,$E6,$DF,$87,$7E,$72
	.dc.b	$73,$18,$E4,$FD,$F7,$EA,$EF,$A9,$FD,$D3,$74,$FE,$3E,$DD,$3F,$48
	.dc.b	$D7,$FC,$4F,$C2,$FC,$5D,$E3,$7F,$45,$6D,$FF,$3E,$F9,$45,$E6,$BF
	.dc.b	$A7,$AF,$AD,$7C,$77,$F1,$B8,$C3,$63,$37,$FE,$10,$9C,$B6,$33,$1F
	.dc.b	$E1,$46,$BF,$C3,$DA,$FF,$8F,$9A,$F9,$3F,$0C,$7C,$57,$91,$B5,$FD
	.dc.b	$66,$7B,$9F,$D5,$44,$69,$FE,$4D,$93,$E4,$BE,$11,$F0,$FC,$7B,$7F
	.dc.b	$D8,$27,$4D,$9D,$34,$D7,$C9,$7C,$B3,$FB,$87,$33,$E9,$B1,$3A,$26
	.dc.b	$3F,$CD,$E7,$51,$8E,$5F,$F1,$0F,$B0,$FC,$07,$37,$FD,$BE,$31,$87
	.dc.b	$F8,$09,$7D,$33,$E6,$B1,$7C,$30,$E6,$DC,$7E,$ED,$F2,$7F,$C0,$0F
	.dc.b	$2A,$E3,$84,$9F,$85,$E5,$DC,$BE,$4B,$E1,$FC,$D3,$F3,$53,$F1,$FE
	.dc.b	$CF,$1D,$9F,$8D,$6C,$64,$EF,$3C,$2F,$E9,$CF,$E4,$FF,$14,$FC,$A7
	.dc.b	$D0,$5B,$FA,$4D,$AF,$B2,$DD,$BF,$C5,$1C,$68,$FC,$BE,$DF,$C2,$AB
	.dc.b	$FD,$5B,$F1,$9F,$97,$8C,$47,$FA,$29,$AF,$C4,$8D,$13,$0F,$CD,$4E
	.dc.b	$A7,$FB,$8F,$87,$FA,$4D,$BF,$7D,$5C,$EF,$13,$FA,$DF,$6A,$4F,$E2
	.dc.b	$FD,$AB,$87,$E4,$A6,$E9,$EA,$9B,$5F,$5B,$8F,$FC,$46,$FE,$4F,$C8
	.dc.b	$FB,$BE,$C7,$EE,$5B,$19,$3E,$DC,$F0,$C7,$F3,$F3,$FE,$48,$F6,$BF
	.dc.b	$D6,$1C,$3F,$E1,$36,$DF,$FA,$19,$B5,$FE,$01,$6B,$FE,$B0,$6B,$FC
	.dc.b	$19,$AF,$F9,$73,$DD,$57,$56,$D7,$FA,$4B,$83,$F3,$9E,$EF,$E5,$8F
	.dc.b	$24,$EA,$FF,$3D,$97,$FA,$6A,$FF,$96,$F8,$7F,$C5,$EF,$F5,$AD,$8C
	.dc.b	$9F,$B7,$8C,$B4,$FD,$D7,$A3,$FC,$7F,$DF,$FE,$40,$D8,$C9,$FB,$F8
	.dc.b	$CB,$4F,$F2,$2F,$8E,$EE,$FE,$62,$48,$FE,$62,$FF,$87,$FA,$3E,$87
	.dc.b	$1F,$E5,$38,$BF,$9C,$CF,$F2,$F7,$F9,$87,$A5,$2F,$F8,$74,$5D,$BF
	.dc.b	$8A,$CD,$BE,$55,$FC,$4E,$3F,$99,$FD,$AF,$F6,$EA,$34,$9C,$9F,$E4
	.dc.b	$FE,$27,$E8,$EF,$8B,$FE,$10,$DC,$7F,$94,$5F,$F7,$EC,$BF,$88,$CE
	.dc.b	$71,$FF,$30,$FE,$BA,$FB,$8F,$F2,$53,$70,$FF,$37,$B6,$7F,$C6,$8D
	.dc.b	$7F,$A8,$EE,$FA,$CF,$EA,$07,$F9,$66,$78,$63,$FB,$A8,$CD,$FF,$31
	.dc.b	$75,$86,$30,$FE,$73,$16,$BF,$C0,$77,$CA,$FF,$D5,$D8,$77,$7F,$CF
	.dc.b	$BF,$03,$FE,$6F,$27,$93,$FE,$63,$7F,$CD,$D3,$FC,$97,$1F,$F5,$63
	.dc.b	$7D,$54,$FF,$A6,$3D,$0F,$ED,$5E,$3F,$F3,$1F,$D9,$BF,$D5,$38,$7F
	.dc.b	$99,$F8,$91,$88,$FE,$1C,$D7,$F9,$8B,$54,$70,$F0,$7E,$C4,$E6,$31
	.dc.b	$A7,$EA,$63,$86,$1F,$DE,$FB,$3F,$BD,$1E,$3B,$B5,$FE,$BB,$F4,$1F
	.dc.b	$AB,$3E,$D3,$FD,$25,$18,$6C,$66,$7F,$D0,$19,$FF,$7E,$8C,$3F,$E8
	.dc.b	$3A,$E1,$3A,$6C,$66,$76,$FF,$82,$8E,$55,$FA,$5F,$7F,$FC,$20,$7F
	.dc.b	$B6,$AB,$8F,$1F,$F3,$A8,$CE,$1F,$DC,$5D,$61,$E2,$9D,$4F,$B6,$1F
	.dc.b	$0F,$EA,$24,$79,$63,$2D,$8C,$E3,$E7,$6B,$F9,$6C,$3F,$C4,$87,$82
	.dc.b	$2C,$FF,$83,$7E,$C7,$FC,$71,$F0,$9F,$E2,$6F,$86,$FE,$FD,$87,$F8
	.dc.b	$A2,$72,$A3,$3C,$8F,$C0,$DF,$68,$C7,$97,$2F,$C7,$C7,$8E,$1D,$31
	.dc.b	$46,$27,$5E,$93,$3E,$35,$FE,$CF,$CB,$AE,$38,$8E,$35,$B7,$C9,$7C
	.dc.b	$AF,$FC,$6D,$87,$A2,$7F,$D8,$0D,$74,$7E,$11,$7D,$A7,$F9,$0A,$30
	.dc.b	$D7,$8B,$5F,$63,$5F,$8C,$F1,$CB,$DD,$3D,$36,$F4,$36,$BC,$E7,$FB
	.dc.b	$AF,$E0,$FF,$C7,$0E,$AF,$96,$3D,$AF,$07,$C4,$DB,$FD,$4D,$EC,$7E
	.dc.b	$4F,$2F,$AF,$CF,$F4,$FE,$E3,$F2,$DC,$9F,$4F,$34,$E5,$FF,$31,$C6
	.dc.b	$63,$10,$8E,$18,$BF,$4A,$7B,$5D,$38,$BF,$CD,$73,$98,$FF,$32,$36
	.dc.b	$FF,$55,$5C,$1B,$3A,$8C,$47,$EC,$5A,$FE,$B6,$79,$39,$9F,$F3,$86
	.dc.b	$1F,$8F,$F9,$4F,$14,$FB,$39,$11,$89,$D4,$FF,$07,$C1,$E0,$66,$F7
	.dc.b	$3F,$0B,$2F,$F4,$2D,$F0,$8F,$85,$18,$9D,$C6,$0F,$F4,$3D,$F4,$7B
	.dc.b	$5F,$B4,$8F,$C2,$6B,$EF,$6B,$EF,$8E,$4F,$F0,$D3,$D9,$EA,$79,$C8
	.dc.b	$FF,$97,$3F,$67,$7E,$09,$F0,$BC,$14,$62,$7F,$C2,$8D,$7F,$9D,$35
	.dc.b	$FF,$23,$35,$FF,$34,$B5,$FE,$72,$D8,$EB,$3F,$E4,$88,$D6,$7F,$C0
	.dc.b	$7C,$6F,$98,$4B,$FE,$AB,$C7,$F5,$B9,$3F,$A7,$BF,$EA,$96,$C7,$F3
	.dc.b	$46,$BF,$05,$AF,$FA,$FB,$57,$D9,$FD,$92,$47,$D9,$12,$F1,$9F,$5D
	.dc.b	$A9,$FB,$19,$7F,$17,$BC,$72,$FF,$6E,$AB,$C5,$F7,$E6,$DF,$4C,$39
	.dc.b	$BD,$F1,$ED,$7E,$B9,$CC,$FF,$81,$A3,$07,$F9,$BB,$C9,$FE,$AD,$1F
	.dc.b	$E9,$46,$BF,$EA,$16,$E3,$FB,$78,$7D,$97,$E1,$23,$0F,$E1,$A3,$F0
	.dc.b	$CD,$7F,$AA,$35,$FF,$6F,$B7,$3F,$E6,$DF,$6F,$FA,$12,$4F,$C9,$8F
	.dc.b	$E7,$5F,$8D,$E2,$3E,$3E,$CA,$C6,$1F,$C1,$E1,$5F,$AB,$25,$73,$8C
	.dc.b	$BF,$E1,$6F,$67,$F0,$5E,$5F,$F9,$AA,$3F,$00,$D7,$F0,$0D,$7F,$B7
	.dc.b	$6D,$FE,$16,$9F,$E0,$35,$FE,$C4,$9E,$DF,$77,$FB,$B3,$F7,$9F,$07
	.dc.b	$23,$24,$3D,$6F,$F0,$5F,$8D,$FC,$5C,$79,$FE,$97,$F8,$21,$9E,$39
	.dc.b	$19,$21,$F8,$A9,$E2,$FF,$B6,$7E,$89,$87,$EC,$DB,$1F,$ED,$26,$BF
	.dc.b	$C2,$7D,$F7,$E7,$4F,$8B,$FE,$37,$18,$7F,$C2,$F2,$8C,$3F,$E9,$B9
	.dc.b	$E0,$D9,$DC,$EA,$3F,$D3,$D3,$A6,$CE,$A3,$11,$FE,$C7,$AF,$5F,$0F
	.dc.b	$DD,$E6,$FE,$4C,$7A,$F3,$95,$F7,$63,$2F,$B6,$3F,$96,$7D,$8C,$8F
	.dc.b	$C1,$3F,$CB,$FB,$71,$1F,$8C,$8B,$7F,$8B,$9F,$E7,$73,$FD,$3F,$96
	.dc.b	$F8,$1F,$91,$8F,$DF,$19,$2F,$E8,$AF,$B1,$E2,$D8,$F9,$2D,$7F,$C8
	.dc.b	$2D,$7F,$97,$35,$FF,$1E,$35,$07,$0D,$8F,$AA,$D3,$FC,$9B,$D3,$FE
	.dc.b	$23,$8F,$E9,$DF,$1B,$27,$F4,$91,$FD,$44,$C3,$FA,$E8,$C3,$FC,$6F
	.dc.b	$C2,$73,$3E,$83,$FA,$C7,$A3,$2B,$53,$96,$CE,$5F,$DC,$CF,$F4,$59
	.dc.b	$D4,$7E,$5F,$C7,$FE,$C5,$E6,$C8,$FC,$B4,$7E,$64,$8E,$0E,$FF,$35
	.dc.b	$87,$91,$93,$FB,$26,$3E,$88,$FE,$DD,$E9,$49,$FE,$08,$FE,$C7,$83
	.dc.b	$11,$FC,$3A,$2F,$17,$F8,$4F,$AF,$F0,$13,$FC,$4B,$E2,$7F,$BC,$11
	.dc.b	$FE,$1D,$F4,$B5,$D1,$FC,$46,$3F,$BF,$47,$E3,$39,$DF,$0D,$49,$5F
	.dc.b	$8C,$FA,$D7,$D5,$FE,$49,$B1,$FB,$26,$BF,$E8,$A6,$BF,$DD,$DA,$FF
	.dc.b	$A1,$5A,$FE,$29,$A9,$A6,$B6,$7E,$20,$F3,$FF,$BE,$36,$BF,$08,$D7
	.dc.b	$E7,$BF,$E1,$4D,$7F,$C7,$DE,$37,$FA,$F3,$DE,$FE,$DD,$E4,$1A,$4B
	.dc.b	$FE,$1B,$FA,$9F,$F9,$81,$FE,$85,$F1,$0F,$86,$FF,$15,$36,$3F,$63
	.dc.b	$3B,$7F,$91,$78,$7F,$CA,$C8,$FF,$65,$FD,$AB,$F8,$86,$7E,$47,$F8
	.dc.b	$7C,$7E,$B3,$34,$62,$3F,$BC,$B6,$33,$E3,$FF,$81,$E3,$04,$70,$7F
	.dc.b	$91,$3F,$1D,$65,$FE,$41,$6D,$FF,$5D,$6B,$FE,$39,$B7,$E2,$8F,$C2
	.dc.b	$7F,$C2,$CD,$9F,$F0,$A3,$5F,$F0,$A3,$5F,$EA,$AD,$7D,$8D,$7F,$C2
	.dc.b	$6D,$9E,$F3,$FE,$21,$8D,$78,$8F,$C5,$7F,$DE,$77,$FD,$CD,$AF,$F8
	.dc.b	$61,$BA,$7D,$EF,$BA,$FF,$4A,$37,$DC,$FF,$16,$47,$FA,$4E,$BF,$C4
	.dc.b	$6D,$C8,$D2,$4F,$02,$4F,$F1,$AF,$C7,$7F,$D7,$6F,$F8,$15,$AF,$F8
	.dc.b	$A9,$B7,$F9,$0F,$9E,$7C,$97,$F8,$FD,$B5,$FE,$3D,$6B,$FE,$3D,$6B
	.dc.b	$FD,$F9,$AF,$E1,$5A,$FF,$8F,$1B,$89,$A6,$B3,$FF,$4E,$F9,$5F,$E8
	.dc.b	$17,$FA,$06,$1F,$E9,$D6,$FB,$5F,$91,$FD,$7D,$F5,$DF,$E5,$56,$E5
	.dc.b	$FE,$53,$6B,$FE,$53,$6B,$FE,$14,$6B,$CE,$3F,$CB,$91,$88,$FF,$29
	.dc.b	$B7,$23,$9B,$7E,$23,$FC,$7B,$1F,$F5,$93,$71,$FF,$49,$B7,$0F,$F2
	.dc.b	$EE,$4F,$0F,$FC,$D8,$D8,$FF,$35,$35,$FF,$35,$35,$FF,$15,$B5,$FD
	.dc.b	$93,$5E,$B1,$FE,$75,$8C,$65,$FD,$B6,$3F,$D6,$36,$F7,$FF,$D5,$CD
	.dc.b	$BF,$F2,$53,$5F,$D7,$7C,$C3,$D4,$FF,$41,$36,$3F,$D0,$0D,$7F,$D0
	.dc.b	$0D,$7F,$C8,$2D,$7F,$D6,$B8,$7F,$99,$73,$FF,$6C,$36,$F2,$7F,$B0
	.dc.b	$57,$8F,$62,$57,$F9,$7E,$78,$3F,$EE,$47,$6F,$F9,$8C,$FF,$6A,$D7
	.dc.b	$FB,$95,$B3,$FE,$C6,$69,$FE,$EC,$3F,$DF,$27,$FA,$38,$CE,$BF,$E0
	.dc.b	$AF,$83,$FF,$6A,$46,$0D,$D7,$F9,$FF,$1F,$D4,$BF,$E1,$0F,$A7,$7C
	.dc.b	$5F,$F8,$4A,$95,$FD,$69,$D4,$66,$31,$09,$E2,$FF,$86,$A3,$AB,$99
	.dc.b	$E1,$C9,$FE,$FF,$8C,$1B,$F2,$7C,$17,$CA,$3C,$31,$8A,$E3,$87,$8A
	.dc.b	$7B,$5F,$3F,$8D,$E1,$8F,$91,$B8,$7A,$DC,$CE,$F0,$E7,$7E,$23,$B9
	.dc.b	$D4,$A6,$6F,$D7,$8F,$2C,$75,$1F,$0A,$3B,$CE,$2F,$DD,$1B,$8F,$F7
	.dc.b	$71,$F3,$23,$E1,$3F,$EF,$19,$F8,$ED,$8C,$BF,$3F,$E6,$1F,$5B,$FE
	.dc.b	$76,$A4,$62,$7A,$C6,$27,$53,$99,$FF,$3E,$FC,$57,$F9,$29,$34,$8F
	.dc.b	$F9,$39,$FF,$29,$E7,$F4,$E1,$F1,$CE,$0C,$EB,$EC,$BB,$5F,$89,$AD
	.dc.b	$2B,$E4,$F8,$0C,$57,$E3,$DE,$B7,$C9,$FC,$95,$D7,$27,$03,$83,$9B
	.dc.b	$EF,$7D,$A3,$15,$C6,$BC,$13,$EB,$AD,$DF,$A6,$7A,$1F,$E7,$28,$CB
	.dc.b	$D8,$C9,$E2,$9E,$CE,$23,$7C,$1F,$EA,$9F,$83,$FF,$76,$4F,$F9,$F2
	.dc.b	$31,$87,$F9,$E3,$E0,$15,$AA,$56,$9F,$88,$D7,$F6,$8D,$7A,$4F,$2A
	.dc.b	$E9,$5C,$F8,$DF,$63,$FE,$0F,$9F,$C8,$C6,$3C,$3D,$4E,$70,$FA,$53
	.dc.b	$9A,$DC,$E6,$7F,$DC,$AD,$7F,$C5,$8D,$53,$C5,$E7,$F3,$2F,$98,$F2
	.dc.b	$36,$FC,$6D,$9D,$CF,$A7,$F5,$B7,$E1,$54,$97,$EA,$8C,$B6,$B8,$11
	.dc.b	$C1,$1F,$6C,$7F,$AB,$A3,$11,$FF,$00,$23,$0F,$D0,$D9,$D3,$F5,$77
	.dc.b	$FD,$0B,$63,$37,$FE,$B2,$6D,$7C,$AE,$9F,$F0,$44,$FD,$06,$C6,$7C
	.dc.b	$2F,$AD,$18,$8F,$E6,$1A,$FF,$A2,$27,$24,$AE,$4F,$E1,$FA,$9F,$8A
	.dc.b	$FC,$2B,$FD,$54,$63,$5D,$4E,$67,$F9,$E4,$E6,$FF,$07,$F5,$AF,$B0
	.dc.b	$FC,$43,$63,$31,$82,$38,$25,$7E,$22,$78,$3E,$2B,$FB,$8F,$81,$AF
	.dc.b	$FB,$7A,$FD,$3C,$1E,$63,$F4,$50,$F1,$DE,$FE,$1B,$8F,$F2,$A9,$D4
	.dc.b	$E4,$91,$99,$DC,$FF,$30,$6B,$FC,$3D,$AF,$FA,$17,$EB,$5F,$45,$FA
	.dc.b	$96,$D7,$EA,$1B,$3F,$A6,$6B,$CF,$E0,$FC,$C4,$9F,$13,$63,$C4,$D7
	.dc.b	$84,$7F,$44,$F6,$9F,$5A,$FC,$57,$F8,$24,$61,$B1,$99,$E0,$DA,$FE
	.dc.b	$9B,$F5,$DF,$E3,$79,$F8,$CD,$8C,$F9,$5F,$3A,$31,$1F,$CC,$5A,$FF
	.dc.b	$80,$5A,$83,$86,$C7,$F0,$FF,$5D,$F0,$B4,$97,$FC,$42,$32,$FF,$12
	.dc.b	$F7,$3E,$F3,$77,$FE,$31,$18,$C3,$F8,$9B,$5F,$37,$C1,$FA,$49,$5E
	.dc.b	$E7,$F9,$07,$D2,$3E,$B6,$92,$7F,$90,$C6,$5F,$C5,$7A,$4F,$A5,$FC
	.dc.b	$8B,$63,$31,$82,$38,$26,$1F,$DF,$27,$83,$7C,$0E,$64,$9E,$EF,$F0
	.dc.b	$BF,$93,$FE,$D5,$1F,$E0,$4F,$09,$18,$AF,$E8,$2D,$7F,$A0,$B5,$FE
	.dc.b	$98,$D7,$FA,$4B,$63,$9C,$F7,$3F,$8D,$7C,$3B,$D5,$A4,$8F,$14,$65
	.dc.b	$B1,$9B,$FE,$A7,$39,$6C,$7F,$59,$F2,$9F,$48,$F9,$CC,$37,$0F,$F1
	.dc.b	$0C,$EA,$72,$48,$CC,$EE,$FF,$C0,$95,$B9,$CD,$7E,$93,$E3,$1F,$57
	.dc.b	$49,$3F,$AB,$8C,$BE,$0F,$D1,$5F,$1B,$E7,$6C,$66,$30,$47,$04,$C7
	.dc.b	$C7,$3C,$1F,$EE,$BF,$5E,$F8,$AF,$E8,$CD,$AF,$B8,$D9,$DC,$E8,$93
	.dc.b	$A6,$C7,$F9,$1B,$CD,$74,$7F,$24,$6D,$7F,$21,$6C,$EE,$74,$49,$D3
	.dc.b	$63,$A7,$C6,$F3,$24,$F3,$7D,$3F,$21,$23,$D2,$D9,$F7,$B5,$FF,$09
	.dc.b	$E2,$8C,$47,$F0,$7F,$39,$DB,$F9,$56,$DF,$F0,$72,$1F,$C1,$4B,$3F
	.dc.b	$E1,$29,$E0,$7E,$75,$78,$6E,$DF,$E8,$ED,$AF,$E8,$8D,$9F,$E5,$93
	.dc.b	$A2,$4E,$9B,$1F,$E5,$19,$F9,$FF,$29,$FC,$C3,$43,$49,$7F,$E2,$AF
	.dc.b	$CF,$FF,$DF,$6D,$F8,$3F,$F1,$84,$63,$0F,$F1,$5B,$5F,$EF,$AD,$7E
	.dc.b	$33,$5F,$F1,$4B,$73,$39,$EA,$6D,$49,$7F,$90,$26,$1F,$90,$6E,$1F
	.dc.b	$96,$7F,$92,$FA,$12,$7F,$91,$B6,$7F,$C9,$FF,$12,$F8,$AF,$F1,$03
	.dc.b	$6B,$F8,$B3,$67,$73,$FC,$C2,$24,$FF,$2E,$6C,$FF,$93,$3C,$24,$62
	.dc.b	$3F,$C3,$F1,$89,$E0,$7F,$7D,$FA,$9F,$E5,$73,$FE,$85,$6B,$FE,$D5
	.dc.b	$6B,$FD,$55,$BC,$0F,$D3,$7F,$B2,$4F,$FB,$6E,$31,$E5,$FF,$6C,$9C
	.dc.b	$CF,$F5,$D9,$CD,$FF,$46,$E7,$FE,$96,$47,$07,$EF,$2F,$F7,$E4,$70
	.dc.b	$4A,$FE,$0B,$19,$6C,$65,$FD,$4C,$65,$B1,$FE,$76,$FC,$C5,$F3,$3F
	.dc.b	$E7,$B6,$C6,$63,$07,$F1,$0E,$99,$1F,$C4,$1B,$5F,$E1,$FF,$98,$7D
	.dc.b	$6F,$FA,$3F,$CB,$46,$2F,$FC,$4C,$DA,$E1,$5B,$8C,$47,$F8,$0C,$FF
	.dc.b	$1B,$38,$7F,$C0,$D5,$D2,$78,$1D,$FE,$49,$F4,$54,$98,$7F,$8D,$63
	.dc.b	$2F,$8B,$E9,$3C,$04,$70,$4C,$3D,$91,$96,$C7,$F8,$B9,$F7,$7A,$7F
	.dc.b	$E7,$C7,$D1,$8A,$31,$1E,$86,$BF,$E5,$16,$C6,$E3,$99,$27,$FC,$99
	.dc.b	$F3,$8F,$AF,$F9,$8D,$8C,$C6,$08,$E0,$93,$FE,$53,$9E,$0F,$F1,$3F
	.dc.b	$C4,$3F,$36,$4F,$47,$FC,$86,$FE,$82,$1F,$53,$FC,$8F,$93,$FC,$6D
	.dc.b	$F1,$2F,$A8,$FA,$D3,$9A,$FF,$7F,$C7,$F3,$5F,$4B,$E1,$67,$C2,$37
	.dc.b	$E3,$7F,$BF,$AB,$67,$F9,$06,$F9,$61,$D6,$31,$3B,$8C,$78,$9F,$E1
	.dc.b	$AE,$0F,$F8,$03,$1B,$AF,$3D,$6F,$0E,$F9,$FF,$9E,$F1,$F3,$57,$E3
	.dc.b	$4D,$C6,$6B,$C9,$19,$9F,$F4,$2F,$07,$1F,$73,$FD,$11,$3C,$4F,$8A
	.dc.b	$E1,$E0,$D7,$E2,$BC,$31,$F3,$57,$49,$F8,$27,$2A,$F8,$CF,$FB,$7B
	.dc.b	$EA,$7F,$36,$E0,$D1,$C2,$79,$57,$4B,$ED,$97,$F6,$5B,$E3,$7F,$3B
	.dc.b	$B1,$FB,$4F,$F8,$5A,$3F,$D0,$8D,$7F,$AB,$35,$FE,$38,$D7,$F6,$3F
	.dc.b	$1E,$F8,$AF,$F5,$13,$63,$FD,$44,$D7,$FD,$44,$D7,$F4,$AD,$79,$C7
	.dc.b	$22,$3F,$EA,$BF,$C7,$BF,$D6,$4F,$FA,$DD,$BA,$9A,$48,$FF,$5A,$35
	.dc.b	$E4,$7F,$AD,$FE,$BF,$F5,$4F,$FA,$DD,$AF,$FA,$DD,$AF,$EE,$9A,$FF
	.dc.b	$10,$6B,$FE,$C9,$6C,$9A,$62,$7F,$A9,$CF,$AF,$C5,$F2,$91,$F1,$7B
	.dc.b	$5E,$52,$38,$21,$E5,$72,$D3,$FC,$83,$87,$FB,$5F,$CB,$FE,$F2,$DF
	.dc.b	$2F,$E7,$4A,$72,$FF,$41,$F3,$BF,$DB,$DC,$EF,$17,$FC,$87,$39,$21
	.dc.b	$FE,$52,$FC,$3F,$A1,$1F,$F2,$8B,$5D,$18,$21,$82,$1F,$E3,$C7,$83
	.dc.b	$FC,$FE,$7F,$CA,$7D,$FE,$4B,$F9,$F1,$89,$FF,$5E,$4F,$F9,$57,$F1
	.dc.b	$17,$CA,$52,$4F,$F9,$76,$32,$FF,$BE,$FF,$0D,$7D,$8F,$E3,$1B,$19
	.dc.b	$8C,$1E,$1F,$92,$91,$FE,$CB,$6B,$FE,$C8,$6B,$E9,$C9,$18,$8F,$F6
	.dc.b	$E6,$BF,$E1,$4F,$17,$EB,$71,$1A,$49,$FF,$6C,$7D,$97,$C1,$23,$83
	.dc.b	$FC,$FF,$F0,$7F,$74,$8E,$08,$7F,$A2,$1C,$B4,$FF,$26,$1F,$80,$F5
	.dc.b	$7F,$5B,$C1,$FE,$42,$8C,$10,$D9,$0F,$F2,$79,$FE,$62,$AF,$F5,$9B
	.dc.b	$FC,$8F,$AD,$18,$8F,$F3,$03,$5F,$F1,$33,$5F,$EF,$DE,$63,$BF,$F7
	.dc.b	$ED,$8F,$E5,$C7,$F8,$2A,$32,$47,$31,$B7,$F9,$A3,$5F,$E7,$CD,$7F
	.dc.b	$D1,$1F,$81,$FF,$24,$7F,$20,$D7,$3E,$67,$E9,$23,$10,$C5,$18,$6C
	.dc.b	$EC,$93,$FE,$B4,$F3,$FF,$64,$47,$04,$3F,$D7,$AE,$5A,$7B,$3D,$DF
	.dc.b	$86,$47,$07,$C0,$FA,$6F,$8E,$47,$04,$3E,$1B,$96,$9F,$EC,$8F,$7F
	.dc.b	$F8,$31,$1C,$10,$FF,$68,$B9,$69,$FE,$EE,$CD,$F2,$DF,$5B,$E0,$BE
	.dc.b	$D3,$63,$33,$B8,$C1,$1C,$10,$FF,$53,$60,$FA,$9F,$F5,$7B,$63,$FD
	.dc.b	$E1,$18,$3C,$1F,$9C,$7D,$6C,$0D,$26,$9A,$25,$7F,$79,$9D,$C6,$23
	.dc.b	$FC,$09,$CC,$F8,$7F,$F1,$A4,$61,$AF,$F9,$56,$35,$E7,$FF,$31,$9C
	.dc.b	$CF,$F0,$06,$BF,$CC,$5B,$3C,$23,$FC,$14,$D7,$FD,$FF,$D5,$FC,$E4
	.dc.b	$8F,$FB,$F9,$AE,$8C,$10,$FF,$0A,$78,$AF,$8E,$FE,$D2,$D8,$FD,$5B
	.dc.b	$5D,$C7,$F6,$EF,$C1,$BF,$91,$C7,$F8,$75,$AF,$F8,$9D,$AF,$F9,$D5
	.dc.b	$A8,$38,$6F,$77,$52,$3F,$E2,$9E,$56,$0F,$A9,$FE,$80,$49,$D4,$65
	.dc.b	$FE,$07,$EC,$7E,$ED,$F0,$35,$34,$D5,$3F,$2C,$FC,$9C,$FF,$90,$63
	.dc.b	$1C,$9F,$D8,$A7,$33,$FD,$4D,$AF,$F0,$88,$FE,$08,$47,$D5,$D3,$A7
	.dc.b	$C7,$C1,$FE,$18,$C3,$F0,$31,$E4,$8C,$11,$F5,$B5,$FF,$26,$47,$E2
	.dc.b	$3B,$3F,$87,$FA,$7F,$C8,$9B,$3F,$CB,$ED,$39,$6C,$7F,$28,$F9,$B7
	.dc.b	$C7,$FF,$8A,$9B,$19,$8C,$11,$C1,$0F,$F2,$F4,$70,$7E,$47,$D3,$BC
	.dc.b	$2F,$F0,$4B,$63,$EE,$35,$DC,$78,$70,$91,$E5,$6B,$D9,$AF,$F9,$D6
	.dc.b	$BD,$6E,$4F,$F3,$87,$CA,$BF,$03,$79,$9F,$D1,$A3,$0D,$8C,$CF,$06
	.dc.b	$BF,$81,$F9,$A7,$CD,$FF,$88,$9B,$1F,$95,$6B,$B8,$D1,$1D,$35,$FF
	.dc.b	$46,$5F,$97,$E3,$3C,$44,$8F,$13,$FE,$30,$FA,$5F,$F1,$AF,$29,$A4
	.dc.b	$F4,$F4,$4A,$FD,$44,$65,$FE,$13,$F2,$8F,$B2,$FF,$21,$36,$33,$18
	.dc.b	$3F,$C7,$FF,$2A,$F8,$1A,$49,$FD,$A4,$65,$B1,$97,$F5,$11,$96,$C7
	.dc.b	$F9,$2B,$E1,$5F,$11,$49,$3F,$BF,$8C,$BF,$C0,$7D,$D7,$D7,$FF,$94
	.dc.b	$9B,$19,$8C,$1F,$E5,$6E,$CF,$C1,$C7,$F9,$1D,$AF,$F1,$1D,$BF,$DE
	.dc.b	$1C,$DB,$24,$7F,$0F,$9D,$BF,$DE,$DB,$3C,$23,$F9,$98,$F8,$8F,$F3
	.dc.b	$63,$4C,$B8,$23,$82,$1F,$D2,$8E,$07,$FB,$1F,$0F,$F2,$BF,$17,$F2
	.dc.b	$7E,$E7,$D0,$FF,$29,$A5,$7E,$42,$73,$1F,$CB,$63,$13,$B7,$9C,$7F
	.dc.b	$72,$8C,$7B,$4B,$ED,$1B,$E6,$ED,$87,$4B,$EF,$18,$9E,$35,$C2,$7C
	.dc.b	$17,$CD,$EC,$78,$EB,$83,$4E,$07,$AA,$38,$5F,$7C,$3C,$ED,$AE,$35
	.dc.b	$E2,$38,$CE,$B0,$F1,$C7,$80,$D1,$63,$E1,$BE,$6C,$38,$E1,$E7,$8F
	.dc.b	$F7,$F4,$FC,$66,$C6,$5E,$A6,$BE,$89,$F4,$7F,$E5,$0A,$46,$27,$ED
	.dc.b	$35,$E1,$3F,$ED,$65,$F0,$D4,$8F,$FB,$69,$BF,$49,$F4,$61,$F1,$3F
	.dc.b	$AF,$8C,$47,$20,$EE,$35,$9F,$E1,$E7,$FB,$85,$7F,$71,$AD,$72,$DA
	.dc.b	$D5,$73,$9C,$BD,$23,$35,$F6,$23,$C7,$55,$F1,$6D,$C0,$F2,$CF,$58
	.dc.b	$FC,$C6,$1E,$47,$C9,$19,$7F,$C0,$91,$97,$F3,$D1,$97,$B9,$E3,$38
	.dc.b	$94,$E3,$5D,$FC,$D7,$A0,$FB,$AF,$E2,$DC,$1F,$C5,$E7,$31,$8A,$FF
	.dc.b	$58,$D3,$E0,$71,$87,$FB,$8E,$39,$4F,$F8,$67,$0E,$51,$89,$FD,$54
	.dc.b	$F2,$BE,$38,$78,$B0,$B7,$BC,$3C,$5F,$3B,$FE,$23,$C3,$A1,$FE,$C2
	.dc.b	$71,$8F,$43,$FD,$E8,$FC,$29,$E3,$1D,$A3,$71,$89,$E7,$3C,$33,$D3
	.dc.b	$EC,$E5,$7D,$2F,$F8,$B2,$BF,$C5,$CD,$9F,$F1,$03,$5F,$F7,$83,$5F
	.dc.b	$E5,$4B,$E8,$7A,$36,$7C,$35,$DB,$CE,$CD,$78,$7E,$95,$F0,$5F,$D1
	.dc.b	$71,$FE,$69,$18,$C3,$FC,$80,$D8,$FF,$19,$35,$FE,$45,$82,$31,$1C
	.dc.b	$AB,$A3,$FE,$B3,$8C,$B6,$33,$3C,$63,$C8,$79,$62,$D7,$AB,$A7,$E0
	.dc.b	$4F,$FA,$F5,$B5,$FE,$04,$C7,$9C,$60,$8E,$09,$5F,$E5,$B8,$78,$DF
	.dc.b	$4D,$B1,$DE,$71,$3F,$5F,$E9,$3E,$66,$1F,$E5,$36,$BF,$09,$AF,$F5
	.dc.b	$D6,$BF,$ED,$36,$E0,$69,$33,$F8,$6B,$E9,$7C,$07,$86,$2E,$3F,$73
	.dc.b	$EF,$BF,$C9,$6F,$F9,$D5,$B8,$7F,$99,$DA,$FD,$66,$BF,$DC,$5A,$FF
	.dc.b	$B9,$DB,$81,$A6,$A1,$F1,$BF,$F0,$06,$BF,$3B,$18,$F7,$FF,$39,$5E
	.dc.b	$2E,$2F,$F0,$85,$EE,$BF,$55,$83,$E9,$5F,$0B,$F8,$26,$E5,$FE,$8F
	.dc.b	$6D,$7F,$A0,$5A,$FE,$41,$AF,$F8,$2D,$AF,$FC,$01,$6D,$77,$AF,$C9
	.dc.b	$CF,$AB,$C3,$FE,$1A,$DB,$FD,$3D,$F8,$31,$F4,$D7,$ED,$FE,$95,$F4
	.dc.b	$BF,$98,$6D,$FF,$A9,$5B,$87,$FA,$5D,$AF,$E9,$9A,$FF,$88,$5A,$B2
	.dc.b	$46,$1B,$5E,$8E,$8F,$F4,$2D,$FD,$06,$DF,$CC,$BF,$0E,$3F,$E3,$F9
	.dc.b	$CB,$FC,$BF,$E3,$BF,$19,$7F,$C5,$3B,$11,$8C,$3F,$8F,$46,$38,$7F
	.dc.b	$94,$CE,$6F,$FC,$9D,$9F,$FB,$0B,$7F,$DC,$F2,$FE,$13,$2F,$F6,$0C
	.dc.b	$60,$F4,$78,$FC,$B9,$BE,$7E,$BF,$85,$AF,$F6,$96,$9E,$29,$FA,$7F
	.dc.b	$76,$FA,$5F,$AE,$DA,$FE,$CB,$BA,$30,$FF,$68,$6B,$FD,$BD,$AF,$F8
	.dc.b	$DD,$AF,$FB,$85,$AF,$39,$FC,$2F,$9E,$F9,$9F,$C2,$B6,$FF,$3F,$7E
	.dc.b	$0D,$FF,$CD,$93,$9C,$3E,$4F,$07,$F9,$1A,$FE,$72,$F8,$6F,$97,$E3
	.dc.b	$7C,$C6,$CE,$9B,$5C,$9F,$E8,$DF,$98,$FF,$A1,$E7,$9F,$E0,$18,$F4
	.dc.b	$FE,$AB,$63,$27,$F4,$09,$E1,$7E,$49,$FE,$D7,$EF,$3F,$37,$7C,$8E
	.dc.b	$3F,$38,$7C,$B6,$A7,$39,$7F,$5A,$B5,$6A,$31,$3A,$C7,$FB,$35,$6A
	.dc.b	$F9,$61,$D5,$FE,$F7,$E5,$3E,$AB,$E5,$62,$C3,$FC,$0F,$18,$58,$73
	.dc.b	$9F,$ED,$75,$C7,$97,$D1,$F1,$92,$7E,$07,$DC,$DC,$3C,$7A,$AB,$1C
	.dc.b	$BF,$9A,$E0,$5E,$30,$FC,$9F,$23,$E2,$7F,$42,$7E,$8B,$E9,$6C,$FD
	.dc.b	$C8,$B9,$FE,$02,$6E,$3F,$84,$FA,$5F,$C5,$92,$FF,$33,$D0,$8C,$5F
	.dc.b	$F9,$02,$31,$3A,$9C,$CF,$F9,$27,$D0,$FF,$02,$5F,$FA,$ED,$B5,$FE
	.dc.b	$93,$BE,$35,$FE,$17,$3F,$CA,$FE,$0F,$E2,$BC,$6F,$DD,$F9,$D6,$FF
	.dc.b	$2A,$DA,$FC,$CB,$71,$FE,$10,$FE,$0F,$E5,$E5,$7D,$2E,$FF,$92,$FE
	.dc.b	$0A,$33,$7F,$82,$EA,$47,$E0,$2B,$73,$98,$C7,$CA,$7F,$81,$B9,$FF
	.dc.b	$60,$DA,$34,$98,$7F,$99,$A7,$6D,$9D,$B6,$7F,$9F,$F4,$7F,$0F,$C3
	.dc.b	$F1,$CD,$BE,$31,$F1,$E3,$5E,$17,$F2,$4E,$0F,$E4,$27,$EC,$BE,$26
	.dc.b	$E3,$FB,$C8,$B7,$FD,$73,$9F,$F9,$7C,$FE,$2A,$7F,$94,$CC,$8F,$F1
	.dc.b	$9F,$CD,$3E,$A7,$13,$FC,$20,$BC,$3F,$CB,$A7,$55,$FC,$2A,$7F,$C9
	.dc.b	$10,$FA,$7F,$F1,$4B,$67,$FD,$13,$E5,$23,$15,$FE,$91,$8C,$57,$F8
	.dc.b	$C3,$E9,$DD,$3F,$E5,$36,$E1,$E6,$9C,$4C,$8F,$E9,$1C,$BF,$8E,$CB
	.dc.b	$FA,$1D,$F0,$C7,$FA,$4C,$E6,$BF,$DF,$D3,$9A,$FF,$40,$4E,$6B,$FB
	.dc.b	$04,$E6,$BF,$A1,$CE,$65,$F4,$FF,$C1,$E7,$FA,$0B,$63,$FD,$4E,$D7
	.dc.b	$FD,$5C,$D7,$FD,$EC,$FF,$35,$B4,$E3,$93,$FC,$03,$3C,$E7,$35,$FC
	.dc.b	$4E,$38,$57,$F1,$5C,$FE,$67,$C4,$7D,$02,$47,$D0,$D5,$1F,$37,$7F
	.dc.b	$98,$7D,$86,$CF,$F9,$D3,$4F,$F4,$53,$F1,$A7,$4E,$5D,$E3,$FE,$28
	.dc.b	$9D,$BF,$E0,$7B,$EF,$5B,$3F,$1B,$F3,$3F,$84,$23,$E5,$7F,$0F,$F3
	.dc.b	$5F,$8A,$3F,$16,$D8,$FE,$7D,$F3,$3F,$EB,$EB,$FE,$68,$7F,$32,$70
	.dc.b	$FE,$A9,$AE,$71,$FF,$3F,$F3,$27,$2D,$8C,$C6,$0E,$3E,$83,$E6,$DF
	.dc.b	$0F,$FE,$C0,$8C,$3F,$D6,$27,$E0,$C6,$31,$FF,$56,$AF,$86,$FF,$42
	.dc.b	$B7,$C0,$FF,$48,$B6,$7F,$DD,$AD,$CF,$FD,$E8,$D7,$DE,$D7,$FA,$A3
	.dc.b	$5F,$F3,$0D,$76,$C7,$F8,$ED,$78,$9B,$3F,$EC,$7A,$ED,$C7,$F8,$C7
	.dc.b	$DD,$DD,$ED,$AD,$CE,$5B,$7E,$0C,$3F,$9E,$E1,$FE,$70,$9C,$BE,$2F
	.dc.b	$91,$F6,$C9,$A4,$AF,$7C,$65,$B1,$99,$F8,$33,$97,$E1,$FC,$57,$FB
	.dc.b	$C6,$BE,$6B,$63,$D9,$3B,$C3,$D9,$39,$7F,$53,$2F,$A3,$7A,$CE,$0F
	.dc.b	$EF,$DF,$0A,$FA,$3F,$C6,$D6,$A7,$31,$FA,$FF,$A4,$78,$FF,$B4,$BE
	.dc.b	$18,$FF,$81,$F0,$E0,$A3,$2F,$87,$1F,$03,$9C,$3B,$C6,$27,$9F,$5B
	.dc.b	$1E,$FB,$78,$B0,$ED,$7B,$C3,$8E,$7F,$E5,$2C,$BC,$35,$FA,$63,$C1
	.dc.b	$19,$AF,$0C,$66,$5F,$01,$FE,$AA,$8E,$07,$C7,$70,$ED,$AF,$C6,$6B
	.dc.b	$F0,$5E,$59,$FC,$09,$EB,$C7,$E9,$AE,$8F,$AF,$47,$D2,$7F,$88,$E7
	.dc.b	$55,$C1,$F8,$C7,$5A,$CE,$1D,$B3,$FE,$45,$7C,$6B,$75,$FB,$1E,$DF
	.dc.b	$E0,$FC,$1F,$9F,$E2,$FE,$47,$5F,$B6,$6B,$FA,$5E,$17,$81,$C3,$87
	.dc.b	$F5,$F1,$89,$FD,$BB,$5F,$DB,$B5,$FD,$BB,$5E,$33,$FE,$86,$C7,$F1
	.dc.b	$2D,$9D,$4F,$FB,$06,$35,$F4,$6F,$88,$F1,$4F,$FB,$7A,$30,$FF,$0E
	.dc.b	$6B,$E9,$F0,$F7,$82,$73,$9F,$EF,$DB,$97,$F0,$A6,$DF,$FB,$46,$F8
	.dc.b	$E5,$FD,$CA,$3D,$F1,$E7,$23,$82,$1F,$09,$CB,$FC,$6A,$7F,$CF,$2F
	.dc.b	$9F,$D5,$F2,$C6,$23,$FC,$EB,$FB,$4F,$E9,$B1,$FE,$BF,$DE,$8F,$D4
	.dc.b	$CF,$F4,$08,$C7,$0F,$FA,$C1,$B1,$F9,$86,$BF,$A1,$6B,$C6,$79,$BF
	.dc.b	$E3,$06,$BF,$E7,$C6,$F2,$7F,$85,$FB,$7E,$8B,$67,$FC,$4F,$18,$F0
	.dc.b	$9E,$FD,$3F,$D8,$75,$FE,$1F,$6C,$6A,$BF,$BA,$46,$AB,$E8,$D1,$A6
	.dc.b	$2B,$FD,$6D,$5F,$E9,$3C,$BF,$DB,$F3,$AE,$1F,$2B,$BF,$6E,$C4,$AE
	.dc.b	$1C,$3F,$E0,$7A,$D4,$7F,$90,$2B,$CA,$47,$FC,$10,$FF,$B7,$17,$D1
	.dc.b	$FF,$A4,$5E,$67,$FA,$95,$C3,$FE,$7D,$6E,$1F,$CC,$1F,$C5,$46,$78
	.dc.b	$7A,$46,$23,$F8,$B3,$5F,$EF,$AD,$59,$FC,$5F,$84,$FD,$1F,$9D,$FB
	.dc.b	$6F,$0E,$F3,$FF,$9F,$2F,$CB,$7B,$6E,$8C,$37,$72,$73,$1F,$28,$8E
	.dc.b	$0F,$F3,$2F,$A9,$F7,$CF,$B6,$47,$FC,$5C,$FF,$28,$58,$FF,$0A,$F5
	.dc.b	$FC,$A4,$8F,$F3,$79,$4F,$F2,$AB,$75,$FF,$37,$47,$B3,$9E,$3F,$E5
	.dc.b	$E7,$F2,$7F,$39,$FC,$7E,$7F,$CB,$31,$A9,$FF,$2F,$BA,$AF,$E1,$31
	.dc.b	$FE,$1B,$EE,$BE,$37,$F4,$08,$FB,$66,$4F,$F6,$DA,$F8,$4F,$E2,$3E
	.dc.b	$9F,$F8,$8D,$F7,$36,$7F,$D4,$13,$B6,$CE,$A3,$11,$FD,$4E,$78,$36
	.dc.b	$32,$FF,$8F,$23,$2F,$EF,$FD,$57,$CE,$FF,$22,$EC,$46,$27,$FD,$FF
	.dc.b	$3E,$6F,$A0,$7A,$5B,$AD,$57,$FB,$06,$B0,$6E,$7E,$5F,$5F,$E2,$BC
	.dc.b	$E9,$5F,$8A,$E1,$FD,$A9,$FC,$9B,$8B,$FA,$2C,$EA,$5C,$5F,$E7,$6E
	.dc.b	$75,$FF,$2C,$9F,$F2,$57,$AF,$F4,$08,$FD,$0D,$5F,$25,$F5,$9B,$1F
	.dc.b	$8B,$6B,$FD,$D1,$AF,$F7,$D6,$BF,$E6,$46,$A0,$E1,$B1,$FB,$D8,$DE
	.dc.b	$5F,$D8,$3E,$3F,$F5,$64,$AF,$F5,$6E,$95,$FF,$55,$C5,$CF,$FD,$53
	.dc.b	$3F,$57,$DE,$FE,$75,$E0,$7F,$B7,$1B,$6E,$24,$63,$6F,$8F,$A7,$1F
	.dc.b	$12,$5F,$F0,$EB,$DC,$7E,$82,$7F,$27,$E9,$7E,$87,$1F,$77,$2F,$F9
	.dc.b	$F7,$A7,$27,$8F,$E0,$3F,$05,$F2,$77,$3E,$9A,$A7,$31,$DF,$A1,$5B
	.dc.b	$8F,$BB,$8F,$AE,$B7,$18,$3F,$8A,$F2,$BE,$43,$F8,$9F,$C1,$C9,$FC
	.dc.b	$54,$7E,$E0,$C3,$FE,$D4,$8C,$3F,$ED,$AF,$03,$F8,$DF,$D0,$48,$C1
	.dc.b	$0F,$F6,$DB,$97,$FD,$BB,$1F,$BE,$FA,$0F,$E6,$39,$7F,$A9,$9B,$3F
	.dc.b	$83,$6A,$FC,$31,$F0,$7F,$C1,$9B,$1F,$C6,$DA,$FF,$91,$1A,$FF,$93
	.dc.b	$5A,$FF,$AD,$1A,$FE,$45,$AF,$78,$FE,$99,$18,$FA,$1F,$E2,$53,$FE
	.dc.b	$F6,$FB,$7F,$E7,$67,$F1,$88,$CF,$57,$EB,$76,$7C,$B7,$F6,$E6,$DF
	.dc.b	$F9,$63,$AB,$F6,$33,$93,$FC,$0D,$F0,$6F,$B9,$FE,$14,$DA,$FC,$73
	.dc.b	$5F,$B9,$D8,$87,$D9,$F8,$FD,$59,$5F,$CA,$34,$FF,$41,$BF,$D4,$3E
	.dc.b	$E3,$FA,$93,$F8,$5C,$D1,$88,$FF,$6C,$C6,$3C,$DF,$F0,$3C,$E6,$7F
	.dc.b	$DF,$8D,$7F,$B0,$B6,$78,$47,$BE,$30,$6D,$7D,$1B,$E2,$7F,$A0,$B6
	.dc.b	$3F,$9F,$B5,$F3,$7D,$23,$E9,$3F,$C3,$ED,$8F,$98,$D7,$FD,$19,$82
	.dc.b	$31,$1E,$C6,$BF,$C4,$63,$53,$FE,$24,$9C,$CF,$F7,$C6,$CE,$A3,$FB
	.dc.b	$55,$7F,$AD,$9F,$3F,$C0,$FA,$8F,$F1,$9F,$DA,$3E,$42,$9F,$E3,$07
	.dc.b	$13,$F8,$36,$BE,$86,$BF,$49,$AF,$9D,$AF,$F8,$41,$B6,$6F,$CA,$BE
	.dc.b	$73,$FC,$8A,$DA,$FF,$08,$B5,$FC,$13,$FC,$15,$CF,$81,$FE,$46,$8C
	.dc.b	$47,$CC,$6B,$F8,$66,$BF,$2D,$AF,$F8,$91,$B6,$69,$8F,$A3,$FF,$4E
	.dc.b	$F4,$FF,$55,$6F,$77,$D5,$6D,$FF,$8B,$9F,$BD,$3B,$DF,$FC,$B7,$1F
	.dc.b	$58,$8E,$09,$3F,$B1,$8C,$BF,$B3,$FA,$3F,$F7,$8D,$7F,$65,$6D,$FE
	.dc.b	$15,$B3,$FE,$3E,$F8,$A7,$D2,$7F,$9C,$1B,$1F,$BD,$6B,$F9,$26,$BF
	.dc.b	$AC,$6B,$F9,$16,$BD,$63,$FC,$EB,$18,$CC,$DF,$85,$7D,$1F,$FA,$01
	.dc.b	$B7,$FE,$4C,$6B,$FB,$D5,$86,$CF,$F4,$0B,$88,$FE,$28,$D7,$F5,$2D
	.dc.b	$7F,$BB,$1F,$E4,$8F,$86,$F6,$E6,$CF,$FB,$D7,$A6,$9C,$C6,$1B,$5C
	.dc.b	$7B,$7F,$CC,$B1,$8A,$E1,$AE,$0C,$CF,$F9,$AA,$9E,$9D,$23,$FE,$7E
	.dc.b	$AF,$F0,$57,$A7,$88,$FE,$C1,$1F,$E0,$D9,$C6,$5E,$79,$FF,$30,$9F
	.dc.b	$E2,$1C,$FF,$B5,$C7,$F8,$4F,$E8,$9F,$35,$F0,$27,$F4,$D1,$8A,$FF
	.dc.b	$0C,$46,$27,$5E,$77,$F8,$BA,$55,$B5,$5D,$9C,$D7,$5E,$7F,$9B,$5B
	.dc.b	$AE,$95,$DA,$FA,$4E,$5E,$DE,$2E,$EF,$EA,$B6,$BF,$C5,$8E,$F1,$B8
	.dc.b	$F9,$9F,$73,$E1,$9A,$7B,$2B,$C7,$1B,$8F,$39,$E3,$9D,$BE,$C9,$DB
	.dc.b	$C4,$CE,$3E,$FC,$7C,$51,$F8,$A3,$91,$4F,$8B,$8F,$C9,$75,$3F,$1D
	.dc.b	$BE,$A3,$E8,$FF,$D3,$53,$FE,$54,$6C,$66,$7A,$4E,$5F,$F6,$37,$C6
	.dc.b	$FF,$B1,$D3,$48,$E8,$F7,$9F,$E5,$5F,$8E,$FC,$46,$3F,$6E,$3F,$CB
	.dc.b	$13,$F0,$94,$EB,$4F,$64,$E6,$31,$3F,$06,$31,$C9,$C6,$3C,$67,$F8
	.dc.b	$22,$BE,$35,$7E,$5F,$6F,$04,$E9,$F1,$46,$67,$BC,$FA,$0F,$1D,$75
	.dc.b	$BE,$F1,$93,$BE,$1C,$A7,$51,$C6,$FC,$FC,$0F,$59,$ED,$FE,$A7,$87
	.dc.b	$F5,$53,$98,$C6,$1F,$C0,$A9,$F0,$38,$C7,$FD,$9F,$1C,$A7,$F7,$0D
	.dc.b	$8F,$DB,$B7,$0E,$AF,$29,$DD,$F3,$C1,$F3,$AF,$85,$C3,$6B,$FD,$DB
	.dc.b	$F7,$5F,$C5,$5F,$E0,$D1,$9C,$BF,$82,$B7,$2F,$5B,$67,$FC,$09,$5C
	.dc.b	$6B,$77,$EB,$EE,$FF,$35,$7C,$9F,$F0,$B9,$FF,$13,$F8,$68,$C4,$7F
	.dc.b	$83,$63,$18,$7F,$90,$5B,$1F,$A3,$CF,$F2,$75,$ED,$E1,$CE,$9F,$EB
	.dc.b	$7B,$B8,$7F,$2E,$9B,$83,$17,$D2,$3E,$2B,$C3,$39,$8C,$5F,$F9,$45
	.dc.b	$B1,$FE,$26,$6B,$FE,$40,$6B,$FE,$25,$6B,$CE,$7A,$D7,$81,$B8,$3E
	.dc.b	$8F,$E3,$37,$0F,$F5,$F6,$5F,$ED,$1C,$7C,$2F,$B6,$BE,$67,$5B,$E5
	.dc.b	$BA,$B6,$FF,$CD,$0D,$CF,$FC,$7A,$D7,$FC,$AE,$D7,$FC,$78,$D7,$E3
	.dc.b	$37,$3F,$F1,$25,$7C,$6D,$BF,$61,$7F,$D2,$FA,$3F,$A0,$BE,$83,$9C
	.dc.b	$FE,$46,$33,$87,$FA,$B7,$27,$DA,$7C,$66,$ED,$F4,$9B,$87,$F1,$F6
	.dc.b	$D7,$F2,$36,$BF,$DB,$9A,$FC,$8A,$E8,$7D,$2C,$BF,$C7,$9F,$1D,$F2
	.dc.b	$0F,$D7,$57,$43,$FD,$FD,$A2,$3F,$31,$F4,$8F,$8C,$FB,$2D,$CB,$FD
	.dc.b	$28,$DC,$BF,$CD,$2D,$7F,$D0,$6D,$7F,$CD,$0D,$7F,$24,$DC,$4E,$76
	.dc.b	$7C,$AF,$FB,$46,$FF,$BC,$36,$FF,$D2,$55,$E1,$CF,$F8,$6C,$E5,$FE
	.dc.b	$F1,$83,$E9,$1F,$03,$F9,$26,$E3,$FE,$AD,$6E,$3F,$E8,$26,$BF,$E9
	.dc.b	$C6,$BF,$E8,$16,$BF,$AB,$6D,$77,$AF,$8F,$EA,$7C,$A3,$F8,$14,$F6
	.dc.b	$C3,$CB,$97,$FA,$6F,$1F,$C4,$63,$FB,$5F,$A4,$FF,$41,$47,$FB,$05
	.dc.b	$B1,$FE,$96,$6B,$FE,$B0,$6B,$FE,$95,$6B,$CF,$D2,$FD,$8C,$63,$33
	.dc.b	$AD,$7F,$92,$78,$9F,$C9,$FA,$CF,$F1,$99,$F0,$E7,$FC,$EA,$73,$97
	.dc.b	$A3,$2F,$E5,$BA,$FF,$39,$6C,$7B,$32,$EF,$F1,$F9,$EE,$F8,$15,$FE
	.dc.b	$CB,$FB,$24,$FF,$06,$6D,$FF,$B8,$5B,$3F,$EB,$66,$BF,$E4,$ED,$7F
	.dc.b	$9E,$79,$1E,$93,$D1,$F1,$B6,$FC,$DF,$4B,$C2,$FF,$CA,$AD,$41,$C3
	.dc.b	$71,$FE,$4A,$D9,$FF,$09,$34,$FF,$0A,$4F,$03,$A7,$BE,$F9,$37,$DD
	.dc.b	$FF,$5D,$8C,$3F,$1E,$7D,$BC,$3F,$E4,$F9,$CE,$1F,$E5,$7E,$1F,$F5
	.dc.b	$A3,$7A,$FF,$AE,$7D,$5B,$EC,$FE,$C4,$49,$D1,$FE,$5F,$9D,$38,$D3
	.dc.b	$FC,$BF,$7F,$A7,$8F,$F3,$9B,$61,$C0,$AF,$F6,$35,$78,$FC,$4F,$C3
	.dc.b	$11,$ED,$ED,$DB,$69,$FD,$7F,$E7,$1F,$65,$F8,$56,$DF,$E2,$7E,$D5
	.dc.b	$F0,$79,$A9,$3F,$8D,$AD,$E1,$D2,$3F,$D1,$B7,$C1,$D5,$7C,$DE,$17
	.dc.b	$CB,$D3,$9D,$F1,$1C,$5F,$7E,$3F,$41,$F2,$F3,$0F,$CC,$74,$FC,$8B
	.dc.b	$F8,$F3,$96,$CE,$B9,$31,$B7,$E9,$7E,$47,$F5,$47,$D6,$F7,$AD,$7D
	.dc.b	$B8,$B7,$FE,$90,$E3,$FF,$5A,$CE,$4F,$D6,$7D,$37,$F8,$6E,$7F,$CC
	.dc.b	$0D,$7F,$CC,$8D,$7F,$D5,$ED,$7F,$36,$D4,$D2,$7A,$9F,$9C,$FC,$13
	.dc.b	$FD,$65,$5F,$BB,$6D,$7C,$73,$8C,$FE,$F6,$33,$97,$F0,$3F,$98,$FE
	.dc.b	$10,$7E,$53,$E0,$6E,$1F,$99,$8B,$87,$BE,$6D,$7E,$7F,$F0,$37,$C3
	.dc.b	$FE,$95,$B8,$7F,$24,$F6,$5F,$6E,$E8,$D5,$CA,$FE,$53,$CC,$C3,$8E
	.dc.b	$7F,$7F,$1E,$38,$6E,$3D,$9E,$DF,$E3,$1F,$F3,$4F,$4F,$EF,$A3,$11
	.dc.b	$FE,$DA,$6B,$FE,$7D,$8D,$13,$1F,$E0,$73,$A9,$FC,$6F,$DC,$7F,$23
	.dc.b	$C8,$FF,$25,$9F,$CD,$EB,$FE,$43,$9C,$DF,$AB,$AB,$F6,$51,$FC,$4F
	.dc.b	$BB,$FC,$95,$D5,$FE,$4D,$BF,$E0,$D5,$A2,$4E,$AF,$F4,$4F,$F3,$BE
	.dc.b	$B9,$1F,$CC,$FE,$D1,$F5,$58,$D7,$F8,$1D,$5C,$7F,$B4,$D6,$AF,$F8
	.dc.b	$94,$FF,$96,$E3,$E1,$F9,$46,$93,$7F,$E9,$FE,$5F,$F4,$93,$83,$6F
	.dc.b	$F8,$E7,$AF,$FD,$3C,$C6,$5F,$C5,$1C,$46,$8F,$E9,$5E,$6F,$F4,$8C
	.dc.b	$FF,$A9,$4F,$F4,$FF,$AA,$75,$7F,$36,$6D,$FF,$36,$6E,$1F,$EC,$96
	.dc.b	$BF,$EC,$DE,$BC,$E1,$FE,$75,$6D,$7F,$44,$95,$61,$FE,$8F,$37,$5F
	.dc.b	$85,$F5,$3F,$E5,$ED,$8D,$26,$1F,$ED,$9F,$1F,$FD,$25,$F1,$3F,$DB
	.dc.b	$CF,$99,$36,$7F,$A0,$45,$C7,$F8,$C6,$1D,$A7,$24,$8F,$E7,$73,$CA
	.dc.b	$31,$9F,$F8,$8B,$E5,$9F,$15,$FE,$72,$6C,$FF,$BA,$F0,$46,$30,$FF
	.dc.b	$78,$46,$2B,$D9,$E7,$7E,$04,$FF,$4E,$56,$F0,$FC,$3C,$7F,$5D,$FC
	.dc.b	$3D,$F3,$9F,$E8,$CE,$B4,$63,$3F,$F7,$D4,$62,$75,$39,$CF,$FE,$00
	.dc.b	$4E,$6B,$73,$99,$FA,$93,$97,$71,$97,$FB,$DE,$DF,$D6,$7D,$6F,$EC
	.dc.b	$E7,$97,$0B,$87,$F8,$C3,$67,$CA,$3E,$7F,$FD,$47,$18,$6C,$66,$7F
	.dc.b	$09,$8F,$F8,$22,$B5,$18,$84,$61,$DB,$5F,$C6,$35,$E5,$3D,$5F,$83
	.dc.b	$F5,$0F,$9E,$FF,$04,$36,$FD,$15,$BD,$FF,$C1,$13,$9B,$FE,$BD,$B2
	.dc.b	$31,$3A,$C3,$FD,$DD,$8F,$96,$31,$8F,$C3,$F9,$8F,$F1,$1B,$EB,$94
	.dc.b	$63,$1F,$5B,$5F,$5B,$5F,$88,$D7,$E1,$B6,$39,$C7,$CE,$8C,$7E,$06
	.dc.b	$FA,$5F,$2B,$6F,$C9,$58,$9F,$F2,$77,$D5,$7F,$48,$F1,$7F,$C7,$3F
	.dc.b	$15,$FC,$0A,$BF,$C7,$73,$95,$19,$E8,$7D,$5F,$B3,$1E,$3D,$DF,$BF
	.dc.b	$C3,$85,$F3,$C1,$18,$9D,$79,$8C,$B8,$D7,$FB,$B7,$1E,$8F,$82,$B7
	.dc.b	$3A,$8E,$75,$C5,$FC,$45,$F9,$DF,$C6,$B5,$E2,$6B,$E1,$7F,$1A,$DD
	.dc.b	$FF,$CA,$73,$FE,$B7,$6C,$EA,$79,$BC,$32,$F6,$56,$E7,$CC,$7C,$89
	.dc.b	$D6,$3F,$E6,$4D,$3F,$D6,$7C,$BF,$C8,$F0,$E3,$5A,$BE,$18,$FB,$71
	.dc.b	$E9,$5F,$3F,$E1,$7F,$59,$5F,$1A,$9F,$2D,$FE,$40,$8F,$D9,$35,$FF
	.dc.b	$1B,$46,$BD,$AF,$F6,$F7,$07,$EB,$FE,$9B,$FD,$C4,$DA,$FF,$39,$D2
	.dc.b	$7F,$CE,$B1,$FB,$26,$C6,$63,$71,$82,$38,$3F,$CF,$BD,$0F,$C0,$FF
	.dc.b	$CD,$D1,$FE,$0E,$6B,$FA,$E6,$F3,$B0,$F3,$7B,$DE,$8A,$FF,$27,$D6
	.dc.b	$A7,$FD,$1F,$7D,$7D,$BD,$36,$BF,$8A,$62,$C3,$96,$5E,$D7,$D2,$49
	.dc.b	$FD,$B4,$ED,$B1,$97,$F8,$8C,$65,$B1,$FE,$9A,$EF,$FF,$3F,$7C,$0F
	.dc.b	$9D,$B3,$FE,$89,$8F,$2F,$E5,$8F,$B1,$A7,$C6,$71,$3F,$CF,$1A,$FF
	.dc.b	$3A,$6B,$FC,$E1,$AF,$C2,$6B,$FE,$74,$6E,$47,$3C,$DF,$E0,$FF,$AD
	.dc.b	$F9,$8D,$AF,$F0,$E4,$63,$C6,$FF,$0E,$C7,$FB,$45,$AF,$F5,$C6,$BF
	.dc.b	$DF,$67,$FB,$0F,$A3,$FE,$BC,$6C,$7F,$77,$F7,$5F,$0B,$F1,$5B,$1F
	.dc.b	$EA,$D8,$F1,$F9,$8F,$8C,$D4,$49,$D1,$FE,$31,$8C,$3F,$E8,$18,$BC
	.dc.b	$DF,$E3,$09,$E6,$78,$EB,$A5,$7F,$77,$6B,$FD,$D9,$AF,$F5,$4F,$C5
	.dc.b	$73,$5F,$93,$8F,$F6,$EC,$63,$9B,$FC,$38,$D8,$FF,$70,$FD,$6B,$EC
	.dc.b	$FF,$31,$08,$C3,$FE,$E0,$6B,$FD,$7D,$AF,$F6,$46,$BF,$DC,$B3,$7D
	.dc.b	$2F,$C9,$24,$7D,$5F,$46,$BF,$EA,$38,$B9,$7F,$86,$67,$EE,$12,$78
	.dc.b	$1F,$E5,$BF,$1F,$E8,$9F,$1C,$8F,$C7,$7F,$3D,$6F,$A6,$FB,$44,$9F
	.dc.b	$F3,$21,$4F,$F5,$A3,$6F,$FC,$C7,$1F,$B6,$F9,$4E,$4F,$FA,$41,$AF
	.dc.b	$F7,$06,$C7,$D1,$E1,$7D,$C7,$F2,$2A,$D5,$F3,$22,$3F,$C7,$F1,$BA
	.dc.b	$FB,$50,$FA,$F7,$C2,$FD,$66,$C7,$19,$DC,$62,$3F,$D7,$4B,$EE,$78
	.dc.b	$1F,$C9,$9C,$47,$F8,$64,$ED,$19,$23,$98,$DB,$FE,$1F,$6B,$FE,$24
	.dc.b	$6B,$FC,$7A,$3F,$CA,$2F,$FA,$D9,$AF,$F8,$59,$B3,$F9,$7F,$84,$79
	.dc.b	$2F,$6F,$F9,$2D,$6B,$97,$A7,$B3,$83,$F9,$75,$F0,$AF,$DB,$7C,$24
	.dc.b	$CB,$E2,$5F,$E4,$E3,$F7,$5E,$9E,$67,$FD,$E5,$17,$3F,$D4,$78,$6E
	.dc.b	$3F,$F0,$BD,$7F,$B8,$FF,$02,$7B,$FF,$A6,$6C,$76,$3E,$D3,$87,$FD
	.dc.b	$D5,$E3,$BE,$5B,$F5,$AD,$8F,$A0,$D7,$FC,$94,$D7,$FC,$9E,$D7,$FD
	.dc.b	$28,$D7,$F1,$0D,$7F,$B1,$47,$F6,$28,$D6,$1E,$BF,$B7,$FE,$34,$4A
	.dc.b	$F4,$47,$BF,$8B,$10,$CF,$E1,$E6,$BC,$C6,$2B,$F9,$27,$C9,$4B,$E7
	.dc.b	$68,$FE,$6F,$CC,$F9,$5D,$57,$5E,$1F,$E0,$D5,$8C,$7F,$51,$7B,$AF
	.dc.b	$D7,$7B,$DF,$B1,$23,$F3,$C7,$E7,$2F,$A1,$FE,$32,$D8,$D3,$5F,$01
	.dc.b	$FA,$F7,$17,$FA,$E5,$F5,$3C,$AA,$57,$27,$AC,$7F,$43,$E9,$93,$F0
	.dc.b	$1F,$E8,$DE,$D6,$23,$BF,$0F,$E2,$E7,$18,$FC,$CA,$FE,$99,$E7,$4B
	.dc.b	$F9,$2F,$F4,$FF,$87,$C4,$72,$E1,$FE,$DB,$7F,$52,$B5,$3F,$91,$F3
	.dc.b	$FF,$D1,$1F,$11,$FE,$0C,$F8,$03,$4F,$E7,$BE,$53,$D0,$A7,$F8,$A1
	.dc.b	$C5,$FF,$03,$6B,$FE,$9D,$6B,$FD,$3E,$38,$11,$E0,$FF,$30,$F8,$BF
	.dc.b	$ED,$04,$8F,$D3,$78,$CF,$AC,$7D,$7F,$E1,$BC,$3F,$EE,$5D,$1F,$C0
	.dc.b	$6F,$FD,$AC,$D7,$FD,$AA,$D7,$FD,$67,$F0,$BF,$B6,$E0,$FE,$4E,$DE
	.dc.b	$BF,$F7,$D3,$65,$97,$87,$DB,$FF,$18,$FD,$4B,$C5,$A4,$AF,$3E,$BF
	.dc.b	$CC,$F1,$FF,$32,$3F,$CE,$E3,$FC,$14,$D7,$3F,$03,$FE,$0B,$8C,$47
	.dc.b	$81,$B8,$FC,$19,$D1,$27,$FD,$E9,$F8,$0F,$F9,$E7,$4F,$E9,$4D,$C3
	.dc.b	$FC,$E8,$D9,$F5,$FD,$73,$D1,$FF,$12,$B6,$3F,$DF,$CD,$7F,$DF,$AD
	.dc.b	$7F,$C4,$4D,$7F,$D0,$0D,$7A,$C7,$F8,$96,$31,$89,$D7,$6B,$F0,$2F
	.dc.b	$C3,$E5,$FC,$A9,$B9,$7F,$BD,$1B,$8F,$D2,$F9,$B7,$BB,$FE,$3C,$6C
	.dc.b	$7F,$BA,$F2,$46,$23,$FD,$E3,$18,$9F,$F1,$C4,$E6,$7F,$C7,$8D,$7F
	.dc.b	$D3,$4D,$7C,$11,$C0,$FF,$20,$7D,$6F,$F9,$1B,$1F,$EF,$0D,$C7,$FD
	.dc.b	$44,$DA,$FE,$49,$F7,$BF,$D4,$DB,$E8,$FF,$69,$B4,$FF,$5E,$9F,$EF
	.dc.b	$86,$8A,$7F,$77,$74,$D9,$E5,$18,$8E,$07,$09,$EB,$3F,$E7,$9D,$3F
	.dc.b	$38,$0A,$7E,$7F,$36,$0F,$57,$FC,$D2,$DC,$3E,$73,$5F,$9A,$D7,$FC
	.dc.b	$CC,$D7,$E1,$C7,$C8,$23,$C9,$FF,$35,$7D,$5F,$D8,$8F,$DF,$B5,$FE
	.dc.b	$63,$48,$C6,$3F,$E2,$7F,$85,$7D,$EF,$F9,$F9,$B7,$F8,$36,$BF,$82
	.dc.b	$6B,$FE,$7B,$6B,$FE,$D9,$6B,$FC,$01,$B7,$FE,$D2,$9F,$F4,$04,$6B
	.dc.b	$2F,$F5,$D7,$B1,$FE,$A3,$DB,$FC,$76,$7F,$8D,$63,$38,$7B,$3D,$5F
	.dc.b	$F4,$9E,$7F,$E0,$68,$FF,$46,$35,$FF,$75,$35,$FE,$22,$D9,$7C,$27
	.dc.b	$E3,$5B,$F0,$3F,$EB,$38,$FC,$86,$8E,$DD,$44,$9D,$1F,$EA,$48,$E5
	.dc.b	$83,$E3,$FF,$D7,$8D,$3F,$BF,$94,$FE,$AA,$7F,$84,$77,$FF,$10,$B9
	.dc.b	$6C,$72,$39,$FA,$DF,$E8,$6C,$DF,$43,$FE,$07,$B7,$D6,$7F,$6F,$97
	.dc.b	$D1,$7E,$8A,$30,$DF,$4B,$F4,$B3,$CB,$2F,$F3,$15,$71,$C3,$A3,$C3
	.dc.b	$0C,$DF,$0C,$39,$47,$F9,$A7,$0E,$4F,$8A,$7A,$DF,$97,$C5,$E5,$3F
	.dc.b	$E4,$D9,$E4,$FF,$9C,$63,$C1,$8D,$9E,$2F,$82,$7B,$61,$E0,$7F,$34
	.dc.b	$DC,$3C,$38,$F2,$9F,$11,$C0,$E4,$70,$2C,$70,$BE,$F1,$FC,$18,$FD
	.dc.b	$36,$3F,$35,$FE,$13,$3E,$E6,$C6,$5F,$97,$EB,$3E,$B7,$FC,$F9,$5F
	.dc.b	$E3,$16,$C6,$67,$B4,$E5,$FF,$41,$0F,$8D,$AD,$32,$E0,$8F,$CB,$3F
	.dc.b	$8F,$FC,$87,$FB,$47,$5F,$F1,$84,$7F,$A7,$67,$E6,$3A,$E6,$AF,$F2
	.dc.b	$68,$AA,$B9,$67,$F0,$E7,$31,$89,$FE,$57,$18,$AE,$31,$E7,$BF,$0D
	.dc.b	$6E,$3B,$1E,$73,$FD,$60,$E5,$F3,$19,$3C,$53,$D2,$BB,$46,$4E,$D7
	.dc.b	$FC,$F2,$BC,$9C,$AF,$7D,$F3,$DF,$EB,$AB,$FE,$15,$18,$C7,$EF,$56
	.dc.b	$EB,$55,$FC,$EE,$B4,$FF,$48,$9D,$57,$09,$CC,$FF,$4E,$9C,$BB,$8C
	.dc.b	$BF,$EC,$DE,$77,$D6,$2F,$BC,$61,$B1,$97,$FB,$93,$67,$FB,$8B,$5F
	.dc.b	$EB,$B3,$C6,$76,$72,$3A,$FC,$DF,$F5,$17,$FB,$E3,$67,$FB,$DB,$5E
	.dc.b	$6D,$7D,$0B,$EB,$39,$92,$79,$BD,$BD,$19,$1D,$9B,$5D,$9A,$F6,$6B
	.dc.b	$F1,$7E,$B1,$F5,$7F,$11,$B5,$FE,$13,$6C,$FF,$84,$9A,$FC,$2E,$7F
	.dc.b	$F1,$94,$FF,$86,$5B,$1F,$E1,$86,$BB,$F2,$BE,$8C,$62,$3E,$B3,$5F
	.dc.b	$F2,$BB,$73,$FE,$57,$3C,$3C,$39,$7D,$DB,$7D,$FD,$DE,$A4,$9F,$23
	.dc.b	$63,$51,$82,$3F,$E0,$FF,$AB,$F6,$A3,$FD,$88,$E1,$B9,$FE,$63,$B1
	.dc.b	$39,$9F,$EB,$BE,$EF,$E7,$1B,$7F,$C7,$5A,$7F,$1F,$F0,$93,$B6,$CE
	.dc.b	$DB,$3B,$9D,$1F,$9A,$FA,$A7,$C1,$D2,$5F,$F9,$2E,$32,$F3,$8E,$C4
	.dc.b	$70,$4A,$F2,$46,$5B,$1F,$E5,$5F,$AB,$FF,$7E,$37,$C2,$FD,$5C,$62
	.dc.b	$7F,$CA,$ED,$7F,$C8,$3F,$56,$FA,$4F,$88,$DB,$F8,$8D,$9F,$E8,$4D
	.dc.b	$78,$CF,$02,$4F,$07,$F7,$7F,$46,$F8,$6A,$49,$FF,$36,$46,$5B,$E8
	.dc.b	$F0,$23,$83,$C7,$F0,$9E,$62,$38,$39,$FC,$A7,$80,$8E,$0F,$BD,$E1
	.dc.b	$BE,$1B,$F0,$4D,$8C,$C6,$08,$E0,$99,$FF,$41,$8F,$C5,$7D,$27,$FA
	.dc.b	$22,$7F,$14,$D9,$FE,$D0,$D7,$FC,$DE,$D7,$FC,$75,$F0,$EF,$99,$FE
	.dc.b	$5F,$E2,$23,$17,$FC,$D1,$B5,$FC,$C5,$AF,$F2,$F6,$BD,$2B,$9C,$62
	.dc.b	$3F,$CE,$8D,$AE,$DE,$C6,$89,$7D,$DB,$5D,$DA,$F7,$7F,$55,$EF,$3E
	.dc.b	$02,$92,$BF,$59,$19,$6C,$66,$BF,$C1,$93,$96,$C7,$ED,$7E,$AD,$F1
	.dc.b	$F4,$97,$FB,$78,$CB,$FE,$B5,$F8,$5F,$77,$86,$69,$35,$FE,$9A,$D9
	.dc.b	$E1,$18,$3E,$3F,$D5,$48,$FA,$0D,$CB,$E8,$35,$FA,$0D,$7F,$D5,$BF
	.dc.b	$52,$FA,$BF,$E1,$CD,$AF,$E1,$CD,$9F,$F1,$63,$5F,$F5,$9B,$5E,$73
	.dc.b	$C8,$93,$CB,$CE,$EA,$47,$07,$E1,$FE,$53,$F0,$24,$70,$4A,$FC,$54
	.dc.b	$65,$B1,$FE,$3F,$F3,$DF,$54,$E1,$FF,$01,$C6,$1B,$19,$9E,$0D,$AF
	.dc.b	$F6,$6D,$6E,$73,$5C,$08,$FF,$31,$F4,$5F,$4F,$FC,$D5,$B1,$98,$C1
	.dc.b	$1C,$12,$FF,$CA,$D3,$C1,$BC,$5C,$C9,$3D,$DF,$C1,$B0,$FC,$51,$63
	.dc.b	$FD,$FB,$C1,$FE,$0D,$7F,$C3,$0D,$5C,$3F,$8F,$9E,$11,$FE,$12,$F9
	.dc.b	$B7,$BB,$FE,$1C,$6C,$FF,$9A,$E9,$18,$AF,$F7,$9C,$62,$B8,$12,$B8
	.dc.b	$56,$EB,$B4,$FF,$2E,$F1,$65,$7F,$2C,$6C,$F4,$AF,$D3,$B7,$1F,$21
	.dc.b	$2B,$F9,$9F,$77,$F3,$82,$38,$24,$F9,$BC,$04,$EA,$30,$FF,$6D,$FA
	.dc.b	$97,$C7,$D2,$5F,$F6,$F8,$CB,$7E,$93,$81,$1C,$1F,$D1,$BC,$5F,$E9
	.dc.b	$64,$70,$7C,$4F,$47,$E7,$11,$C1,$31,$F8,$51,$96,$C6,$5F,$B1,$19
	.dc.b	$6C,$7D,$7F,$D9,$BF,$86,$93,$B3,$FC,$AE,$FF,$10,$FA,$07,$C4,$7F
	.dc.b	$96,$9B,$5F,$7D,$B7,$F7,$9A,$FF,$86,$C7,$CA,$3F,$17,$4F,$F9,$CA
	.dc.b	$30,$D8,$CD,$7F,$AD,$67,$2D,$8C,$D7,$53,$AB,$87,$C1,$3C,$1D,$35
	.dc.b	$CF,$D5,$7F,$BE,$6B,$B3,$63,$6D,$77,$F0,$BF,$E8,$8F,$4D,$F3,$AF
	.dc.b	$AF,$C3,$70,$FF,$17,$36,$FC,$D5,$C2,$B6,$4A,$DB,$67,$7F,$10,$F9
	.dc.b	$2F,$03,$6B,$FC,$74,$D9,$DC,$E8,$93,$A6,$C7,$8B,$0F,$EA,$8F,$D5
	.dc.b	$C3,$D7,$E4,$FA,$89,$7E,$A6,$CF,$16,$BC,$7E,$2B,$FC,$B1,$3F,$E4
	.dc.b	$D6,$C7,$C0,$6B,$B3,$FC,$AD,$F5,$1F,$E5,$2F,$07,$FD,$B1,$CF,$F8
	.dc.b	$78,$C4,$FF,$B9,$63,$FD,$59,$F5,$2B,$FF,$56,$4E,$4F,$F3,$45,$70
	.dc.b	$F4,$97,$CF,$DA,$FF,$6F,$56,$C5,$6E,$7A,$DF,$4B,$E3,$5D,$2F,$8C
	.dc.b	$62,$38,$C7,$47,$F2,$57,$E8,$6C,$65,$F1,$63,$FE,$73,$9F,$F3,$4E
	.dc.b	$97,$C1,$34,$95,$F9,$D3,$CF,$1A,$7D,$53,$C9,$B1,$9C,$3D,$3E,$1B
	.dc.b	$EA,$7F,$15,$5F,$38,$F8,$F5,$BD,$3E,$83,$63,$38,$EE,$F9,$70,$FE
	.dc.b	$EF,$1F,$8B,$2F,$A0,$FF,$47,$5F,$A9,$B1,$98,$FF,$46,$46,$3E,$AD
	.dc.b	$F6,$3F,$E8,$F6,$CF,$FA,$55,$AF,$EA,$1A,$FE,$5D,$AF,$FA,$9E,$91
	.dc.b	$88,$F9,$6D,$8C,$CF,$8B,$EA,$A5,$7F,$A8,$ED,$18,$8F,$DB,$C6,$1D
	.dc.b	$46,$5F,$D4,$46,$5F,$37,$9F,$EB,$23,$EB,$6B,$A3,$04,$3F,$8A,$FD
	.dc.b	$53,$E8,$7F,$D7,$1D,$A8,$C4,$FF,$23,$6B,$C2,$77,$FA,$73,$DC,$A4
	.dc.b	$9F,$E5,$51,$96,$C6,$5F,$E4,$B1,$96,$C7,$FB,$27,$83,$F9,$BF,$9B
	.dc.b	$7C,$2F,$F0,$0D,$7F,$41,$8C,$1B,$F1,$1E,$B5,$24,$FF,$4F,$7A,$79
	.dc.b	$CF,$AB,$A4,$9F,$F0,$1B,$5F,$EF,$CD,$7F,$C2,$4D,$7F,$DD,$18,$FF
	.dc.b	$84,$23,$11,$FE,$1A,$9E,$CF,$F5,$7F,$63,$D4,$47,$07,$FB,$9F,$EA
	.dc.b	$9F,$57,$F9,$C6,$C6,$63,$06,$7E,$46,$47,$16,$BF,$D8,$FC,$DF,$90
	.dc.b	$47,$FB,$43,$5D,$18,$21,$FE,$1D,$FA,$AF,$8E,$47,$FC,$4E,$FD,$3F
	.dc.b	$90,$91,$FD,$CD,$AF,$F8,$94,$F9,$E4,$3F,$C5,$E5,$34,$60,$87,$F8
	.dc.b	$E3,$E8,$FF,$84,$B7,$BB,$F0,$91,$88,$FE,$FD,$B2,$31,$1F,$51,$AF
	.dc.b	$F9,$31,$B1,$B9,$E6,$7F,$0A,$F2,$7F,$DB,$55,$FC,$84,$FE,$DF,$F0
	.dc.b	$79,$23,$98,$FC,$C3,$FE,$03,$6B,$FE,$11,$6A,$3F,$BC,$F2,$F7,$23
	.dc.b	$CD,$AF,$F9,$FB,$C1,$7C,$A7,$8F,$DC,$46,$27,$FD,$18,$D7,$84,$EE
	.dc.b	$31,$1F,$E2,$39,$E3,$39,$24,$7F,$B8,$FC,$9F,$81,$F0,$5F,$EE,$76
	.dc.b	$FC,$1F,$F6,$19,$D1,$0F,$F7,$44,$68,$8F,$C9,$F0,$5F,$13,$FC,$D9
	.dc.b	$B1,$98,$C1,$FC,$EB,$ED,$DC,$74,$93,$F4,$E3,$2D,$8C,$BF,$D5,$E3
	.dc.b	$2D,$8F,$4E,$92,$3F,$E0,$03,$5F,$F8,$00,$D7,$FE,$00,$35,$FC,$57
	.dc.b	$CB,$FE,$40,$8F,$FA,$3F,$8D,$18,$8F,$CC,$35,$FC,$AC,$69,$FC,$67
	.dc.b	$D4,$39,$FF,$AC,$B6,$33,$18,$33,$F2,$32,$38,$B5,$FE,$ED,$28,$C4
	.dc.b	$7F,$95,$63,$0F,$03,$FB,$E6,$ED,$DE,$87,$DE,$24,$EA,$32,$D8,$CB
	.dc.b	$FE,$61,$9E,$51,$88,$FF,$38,$47,$5F,$7B,$F1,$24,$8E,$8D,$79,$57
	.dc.b	$FA,$81,$C9,$0F,$F0,$97,$D3,$BE,$5B,$FB,$FB,$63,$F6,$CD,$77,$1F
	.dc.b	$8B,$F8,$29,$1F,$89,$6B,$F8,$76,$BF,$E7,$BE,$14,$62,$3F,$9E,$FD
	.dc.b	$2B,$E9,$FF,$AA,$36,$3F,$DA,$0D,$7F,$D9,$ED,$7F,$D0,$DE,$D7,$EA
	.dc.b	$5F,$F1,$63,$5C,$F3,$FF,$86,$A3,$11,$E0,$F6,$BB,$91,$EE,$F8,$BB
	.dc.b	$A4,$78,$9A,$F6,$F8,$9E,$84,$7F,$8E,$52,$31,$1B,$32,$E1,$FE,$5B
	.dc.b	$F4,$AF,$0F,$FB,$A3,$63,$FD,$E0,$D7,$FD,$DE,$D7,$FD,$4D,$F6,$BF
	.dc.b	$8A,$7F,$C9,$8D,$73,$C9,$FE,$3A,$8C,$47,$FA,$E5,$AF,$F2,$96,$A6
	.dc.b	$9A,$F0,$DF,$1C,$F9,$0F,$F0,$53,$6B,$FC,$B9,$3A,$9C,$9E,$1E,$D9
	.dc.b	$3F,$D0,$6F,$F9,$7B,$63,$2F,$A9,$AF,$B9,$FE,$91,$F2,$CF,$A5,$FF
	.dc.b	$6C,$CF,$F9,$AE,$31,$F0,$DA,$9C,$CF,$FB,$2E,$72,$F9,$BE,$95,$E6
	.dc.b	$FF,$88,$DB,$19,$8C,$11,$C1,$0F,$F6,$B6,$69,$E3,$18,$5F,$35,$D4
	.dc.b	$91,$E1,$7C,$1D,$52,$3C,$8D,$79,$35,$FE,$E7,$3F,$B3,$6C,$7F,$8E
	.dc.b	$FE,$91,$F4,$B4,$93,$F5,$23,$2F,$D6,$FA,$1F,$F1,$DB,$7D,$4F,$F2
	.dc.b	$24,$62,$3D,$AD,$7F,$C0,$6D,$7F,$C6,$AD,$59,$23,$0D,$9E,$BD,$52
	.dc.b	$BC,$4D,$7B,$F0,$78,$88,$F9,$DF,$F2,$C7,$47,$F9,$5A,$3E,$8B,$5F
	.dc.b	$F0,$BB,$5F,$F2,$4B,$5F,$2B,$6C,$D2,$7A,$DF,$DD,$A7,$FB,$59,$FC
	.dc.b	$CC,$A7,$F3,$22,$9F,$E1,$1F,$0F,$FC,$25,$E5,$71,$24,$7F,$44,$36
	.dc.b	$23,$6F,$F8,$8F,$37,$D1,$B8,$FF,$C3,$9E,$6F,$F5,$9A,$F9,$9F,$3F
	.dc.b	$FE,$1F,$7F,$C3,$3E,$55,$C1,$39,$AE,$11,$D2,$F9,$D6,$F0,$ED,$87
	.dc.b	$23,$FC,$55,$5C,$9D,$98,$C3,$A5,$71,$BC,$CF,$DF,$BE,$78,$78,$2F
	.dc.b	$73,$C4,$DE,$1E,$37,$C0,$71,$BC,$CF,$08,$F7,$9C,$2F,$AE,$1E,$18
	.dc.b	$E3,$7E,$76,$D7,$18,$ED,$87,$C2,$BF,$7D,$7C,$93,$C0,$7F,$BC,$E7
	.dc.b	$C8,$E5,$AF,$CD,$F3,$1F,$47,$F5,$DB,$1C,$5A,$F1,$6B,$FE,$FF,$F8
	.dc.b	$4F,$F7,$5A,$69,$1F,$F7,$7B,$F3,$17,$C4,$3E,$2B,$11,$C4,$F9,$26
	.dc.b	$9C,$CF,$19,$E1,$3A,$38,$65,$F8,$B7,$81,$99,$E1,$5C,$A3,$AD,$72
	.dc.b	$AE,$11,$CB,$4F,$95,$39,$7D,$D1,$C2,$B9,$57,$7A,$D4,$FD,$1B,$EE
	.dc.b	$FE,$47,$67,$A6,$F8,$2F,$4C,$FF,$88,$23,$18,$FF,$8B,$9B,$1F,$E2
	.dc.b	$76,$BB,$9E,$2F,$ED,$63,$2D,$8C,$BF,$B3,$8C,$BF,$C6,$BE,$8F,$E8
	.dc.b	$57,$F5,$C6,$C7,$F4,$06,$BF,$EA,$66,$BF,$ED,$86,$BC,$E3,$BF,$2A
	.dc.b	$72,$D8,$CF,$D1,$FF,$48,$7F,$B5,$B6,$BF,$A7,$B5,$FF,$5A,$B5,$FF
	.dc.b	$73,$AF,$6B,$99,$27,$9B,$DB,$B6,$47,$66,$D7,$66,$BD,$9A,$FF,$BD
	.dc.b	$3E,$8D,$F4,$7E,$C6,$D7,$F8,$09,$B3,$FD,$99,$AF,$FB,$33,$9B,$FC
	.dc.b	$41,$3F,$E0,$86,$C7,$F6,$C6,$BB,$F2,$BE,$4C,$62,$3F,$D8,$19,$FF
	.dc.b	$20,$9C,$92,$BA,$78,$5C,$08,$E0,$F0,$7A,$DD,$08,$E0,$97,$E3,$8C
	.dc.b	$B6,$3F,$C2,$1F,$47,$EC,$8F,$F6,$97,$5A,$31,$1F,$CE,$DB,$19,$D3
	.dc.b	$F2,$32,$F8,$AF,$F4,$3D,$FF,$B9,$5B,$1F,$C5,$CD,$79,$9C,$23,$31
	.dc.b	$FC,$6D,$AF,$F4,$56,$BF,$EE,$56,$BC,$A7,$FC,$6F,$F4,$7F,$EE,$06
	.dc.b	$F8,$5F,$96,$8C,$4F,$F8,$79,$AF,$78,$F1,$12,$7D,$0D,$8F,$45,$70
	.dc.b	$7C,$91,$97,$FA,$D6,$CF,$A3,$FF,$6D,$DF,$F9,$29,$B3,$FE,$2E,$6B
	.dc.b	$C1,$AF,$F6,$3F,$A3,$7C,$C7,$CC,$6D,$7C,$99,$46,$1F,$98,$D7,$FA
	.dc.b	$23,$5E,$93,$CC,$93,$FE,$5E,$F9,$B7,$D1,$FE,$F1,$B1,$98,$C1,$9F
	.dc.b	$46,$47,$16,$DF,$8B,$DD,$F4,$12,$7D,$18,$A3,$B7,$C1,$49,$F1,$36
	.dc.b	$BF,$A7,$7C,$27,$E3,$3C,$13,$49,$5F,$0E,$32,$D8,$CC,$F2,$8F,$91
	.dc.b	$F4,$7F,$E7,$F6,$E5,$F9,$88,$C4,$FE,$55,$AF,$FB,$5B,$E0,$1F,$37
	.dc.b	$FC,$BB,$C4,$FF,$72,$36,$33,$87,$FB,$8E,$30,$47,$04,$AF,$F6,$DF
	.dc.b	$B1,$C4,$91,$FE,$EB,$6C,$6A,$30,$47,$F5,$5F,$40,$F7,$7F,$62,$D9
	.dc.b	$FC,$33,$67,$73,$A2,$4E,$9B,$1F,$B5,$F5,$FF,$DB,$B3,$F8,$F6,$C6
	.dc.b	$7E,$03,$F7,$F1,$88,$FD,$0F,$CF,$FC,$27,$FD,$16,$D7,$39,$7F,$48
	.dc.b	$8C,$47,$F2,$06,$BF,$E7,$06,$F4,$1A,$6B,$56,$AF,$5B,$E9,$13,$3F
	.dc.b	$E1,$D3,$A8,$C6,$1F,$E2,$78,$C4,$EA,$B9,$36,$BF,$8D,$FD,$1B,$E8
	.dc.b	$E9,$27,$F6,$51,$96,$C6,$70,$FE,$3D,$39,$6C,$74,$F3,$65,$72,$6C
	.dc.b	$F9,$70,$FF,$78,$4F,$97,$E0,$3F,$1C,$47,$04,$9E,$11,$97,$FC,$89
	.dc.b	$E7,$3E,$8D,$C7,$FE,$06,$8C,$36,$33,$7F,$CD,$67,$2D,$8C,$D7,$46
	.dc.b	$BF,$E7,$99,$DB,$9A,$FE,$87,$E8,$3E,$8E,$93,$0F,$E3,$51,$96,$C6
	.dc.b	$6B,$FA,$34,$E5,$B1,$D3,$92,$57,$26,$CF,$E4,$23,$F1,$C4,$3F,$DF
	.dc.b	$7C,$3F,$E1,$62,$39,$8E,$0F,$FC,$00,$F8,$F7,$B5,$FE,$1E,$6C,$66
	.dc.b	$30,$47,$04,$9F,$F7,$5C,$F0,$7F,$DD,$BE,$0F,$F4,$97,$FD,$9A,$DE
	.dc.b	$53,$4C,$4F,$EB,$67,$FC,$0A,$D7,$FD,$AA,$D9,$36,$7F,$97,$57,$F3
	.dc.b	$DE,$FF,$E7,$84,$70,$4B,$F2,$C6,$5F,$EE,$DF,$44,$FA,$3F,$EF,$AD
	.dc.b	$AF,$EA,$4D,$9D,$CE,$BE,$72,$56,$9B,$3F,$0B,$C5,$FE,$9C,$49,$FE
	.dc.b	$A4,$FC,$FF,$42,$47,$D3,$6C,$FC,$C6,$BF,$DC,$33,$FF,$15,$3F,$12
	.dc.b	$32,$FC,$86,$B9,$9F,$F0,$96,$8F,$94,$FE,$1E,$4E,$DF,$F2,$EB,$FC
	.dc.b	$4B,$E5,$5F,$0D,$FE,$60,$6D,$F2,$6E,$1F,$EB,$3C,$11,$88,$FF,$6F
	.dc.b	$7C,$CB,$E8,$F3,$7F,$E2,$38,$C5,$7F,$9E,$9B,$19,$9D,$D7,$FA,$FE
	.dc.b	$72,$D8,$CB,$D9,$A8,$47,$07,$C7,$F4,$6F,$A3,$F0,$36,$FC,$8D,$9F
	.dc.b	$03,$5F,$F1,$A7,$CB,$3E,$8F,$FD,$23,$E9,$23,$17,$FE,$28,$6D,$7F
	.dc.b	$8F,$5A,$F2,$AE,$31,$88,$FF,$1F,$35,$FF,$26,$36,$9F,$0E,$F8,$FF
	.dc.b	$0B,$70,$FF,$1B,$B6,$FF,$C9,$8D,$7F,$C9,$8D,$7F,$C9,$8D,$7F,$CB
	.dc.b	$4D,$AE,$F5,$E6,$F2,$A5,$FB,$5B,$7E,$D6,$BF,$19,$AF,$F9,$67,$E2
	.dc.b	$7E,$A4,$FF,$93,$1B,$1F,$E6,$16,$BF,$E6,$16,$BF,$6C,$7D,$13,$D8
	.dc.b	$FF,$76,$C7,$FB,$CF,$A7,$53,$9F,$0B,$F0,$5E,$92,$7F,$D1,$FF,$46
	.dc.b	$C7,$FD,$DB,$39,$C3,$F1,$18,$71,$7A,$18,$C7,$BC,$EE,$3C,$71,$8C
	.dc.b	$3A,$DF,$47,$C7,$19,$9E,$58,$FF,$96,$F0,$ED,$1E,$D3,$9C,$66,$7D
	.dc.b	$B1,$99,$7B,$1F,$E8,$F8,$FF,$46,$46,$36,$E9,$39,$9E,$0D,$7E,$2B
	.dc.b	$C7,$2F,$55,$71,$9F,$79,$F3,$A7,$58,$FF,$AD,$BE,$8F,$F2,$13,$F3
	.dc.b	$23,$9E,$16,$3E,$4D,$70,$6C,$67,$A1,$7C,$A7,$E6,$CB,$E6,$BF,$CF
	.dc.b	$D1,$FE,$AB,$6C,$66,$3F,$D5,$91,$8F,$A3,$7D,$1F,$FA,$BD,$B3,$FE
	.dc.b	$98,$6B,$FE,$AD,$6B,$FE,$AD,$6B,$FE,$AD,$6B,$FE,$BB,$6D,$1B,$4F
	.dc.b	$A3,$F1,$92,$BF,$D4,$F1,$96,$C6,$5F,$F5,$F4,$65,$B1,$E9,$F4,$24
	.dc.b	$7A,$5A,$FA,$5A,$FA,$5A,$FF,$1F,$EB,$7D,$1B,$E8,$FF,$D5,$11,$86
	.dc.b	$C6,$67,$F9,$93,$5F,$E5,$DF,$46,$F7,$78,$B6,$3F,$9C,$35,$DC,$68
	.dc.b	$8E,$9A,$FF,$45,$F0,$7F,$C1,$AF,$F4,$9F,$33,$FD,$79,$C3,$FD,$21
	.dc.b	$B3,$A9,$FE,$99,$E2,$7F,$97,$DF,$EB,$2D,$73,$EB,$F2,$8C,$47,$F6
	.dc.b	$5F,$A3,$EB,$E4,$9A,$48,$FF,$10,$35,$FF,$0C,$B5,$E4,$68,$86,$8F
	.dc.b	$EC,$5E,$D7,$A0,$8E,$0F,$F6,$C7,$D1,$BE,$8F,$F3,$AD,$8C,$C6,$0C
	.dc.b	$FC,$8C,$8E,$2D,$7E,$57,$9B,$F1,$88,$FF,$73,$6B,$A3,$04,$3F,$C6
	.dc.b	$BF,$47,$F1,$08,$FF,$90,$5F,$A3,$F1,$F2,$3F,$BF,$35,$FF,$1E,$9F
	.dc.b	$30,$87,$F9,$24,$A6,$8C,$10,$FF,$03,$7D,$1F,$F0,$17,$F9,$E7,$0A
	.dc.b	$31,$1F,$DE,$63,$1E,$47,$E1,$A7,$32,$9C,$BF,$E1,$C8,$CC,$62,$3B
	.dc.b	$CF,$F0,$CF,$2B,$FD,$D7,$1F,$CB,$CF,$F0,$33,$87,$0D,$7F,$9A,$35
	.dc.b	$FE,$A4,$FE,$85,$CF,$27,$52,$3D,$4F,$F0,$23,$87,$AB,$5F,$F4,$57
	.dc.b	$80,$F9,$3E,$1E,$DF,$F3,$E6,$C6,$67,$FD,$2C,$D7,$FD,$3E,$D7,$9C
	.dc.b	$70,$23,$FE,$BF,$F2,$3C,$9F,$03,$FE,$CA,$6F,$BF,$FD,$8E,$74,$43
	.dc.b	$FD,$69,$1A,$23,$ED,$F0,$5F,$31,$FC,$E1,$B1,$98,$C1,$FC,$F3,$E8
	.dc.b	$5C,$54,$93,$F1,$E3,$2D,$8C,$BF,$5E,$32,$D8,$F4,$E7,$23,$FD,$B8
	.dc.b	$D7,$FD,$98,$D7,$FC,$15,$97,$FB,$AA,$31,$1F,$8F,$E1,$7C,$B7,$E3
	.dc.b	$C9,$1F,$8D,$6B,$F9,$A6,$BF,$9A,$8E,$2F,$E0,$7E,$8D,$CD,$FD,$71
	.dc.b	$B1,$98,$C1,$9F,$91,$91,$C5,$AF,$E5,$E5,$18,$8F,$F0,$44,$61,$E0
	.dc.b	$7F,$BF,$7C,$EF,$BA,$47,$04,$3F,$DE,$0E,$5A,$7E,$97,$27,$03,$DD
	.dc.b	$FC,$39,$27,$51,$96,$C6,$5F,$F7,$E3,$5F,$F0,$F7,$D1,$BE,$5B,$FB
	.dc.b	$FB,$63,$F5,$0D,$7F,$6C,$D7,$FC,$4D,$C0,$F8,$2F,$C6,$12,$3F,$14
	.dc.b	$D7,$F1,$0D,$78,$BF,$D4,$3E,$89,$F4,$74,$93,$FE,$A6,$8C,$BF,$BB
	.dc.b	$F6,$7F,$4C,$FF,$09,$6B,$FE,$39,$6B,$FE,$57,$6B,$FC,$5E,$1E,$CF
	.dc.b	$84,$91,$DD,$F2,$76,$C8,$F1,$35,$F0,$7C,$47,$F8,$9A,$CD,$24,$79
	.dc.b	$DA,$F0,$30,$7F,$B0,$BE,$8D,$E1,$7F,$72,$6C,$66,$30,$7F,$8D,$3E
	.dc.b	$8F,$FC,$E1,$1F,$C7,$5A,$FF,$2B,$6B,$FE,$5B,$6B,$FE,$79,$6B,$FE
	.dc.b	$B5,$6A,$69,$AF,$09,$F1,$AF,$8F,$FF,$03,$B6,$BF,$DB,$6D,$7F,$DA
	.dc.b	$4D,$7C,$5D,$9E,$92,$3E,$96,$BE,$96,$BC,$0F,$E8,$DF,$2C,$FA,$3F
	.dc.b	$F7,$1C,$61,$AF,$F9,$E6,$35,$F0,$BF,$D0,$5B,$1F,$EF,$0F,$A3,$79
	.dc.b	$9F,$E2,$06,$C6,$63,$04,$70,$43,$FD,$DD,$1C,$1B,$F3,$5C,$C8,$F7
	.dc.b	$7B,$74,$C8,$F1,$35,$E6,$D7,$FB,$9C,$FE,$BD,$FE,$E4,$D7,$FB,$87
	.dc.b	$D1,$3E,$8F,$FC,$72,$D8,$FE,$D9,$BA,$31,$1F,$E9,$D8,$C4,$FF,$90
	.dc.b	$3E,$8B,$FC,$6F,$3F,$DF,$DA,$FF,$7D,$6C,$9A,$4F,$53,$91,$27,$C3
	.dc.b	$1F,$B0,$9E,$1E,$57,$72,$4E,$B8,$3C,$44,$70,$7A,$B9,$FF,$CA,$4F
	.dc.b	$F8,$59,$AF,$F8,$55,$AF,$F8,$4D,$B5,$FE,$F6,$F5,$BF,$CB,$B1,$82
	.dc.b	$3F,$DC,$A7,$FB,$39,$FE,$C1,$F0,$8E,$0F,$EA,$E5,$32,$EB,$CA,$E4
	.dc.b	$48,$FE,$70,$6C,$46,$DF,$F0,$A6,$4F,$A3,$71,$F8,$3C,$DF,$EA,$B5
	.dc.b	$F3,$FE,$7F,$FD,$E1,$18,$B4,$63,$CA,$FF,$7A,$47,$3B,$E5,$18,$C3
	.dc.b	$AE,$1F,$E1,$D8,$DD,$F1,$8F,$7D,$F4,$78,$98,$AE,$38,$78,$5B,$7C
	.dc.b	$A7,$67,$9C,$B5,$E3,$7C,$07,$3C,$33,$87,$86,$38,$1A,$3F,$1D,$7D
	.dc.b	$9E,$06,$6B,$CC,$D9,$E5,$1D,$F0,$F3,$5F,$CE,$BF,$83,$7F,$EF,$DA
	.dc.b	$F2,$36,$32,$FC,$AF,$31,$F4,$7F,$59,$B1,$FE,$57,$AF,$F0,$AC,$E4
	.dc.b	$FF,$73,$C3,$E1,$A9,$1F,$F7,$5B,$7E,$93,$EE,$C3,$E2,$1F,$15,$F9
	.dc.b	$09,$CC,$F3,$0F,$09,$DC,$63,$6F,$F3,$55,$F1,$7F,$33,$1F,$9A,$9D
	.dc.b	$D7,$4A,$D4,$EA,$73,$5C,$A3,$E9,$1E,$5B,$FB,$AF,$E9,$2A,$D7,$8B
	.dc.b	$5E,$93,$E3,$3A,$57,$67,$C7,$55,$E8,$6C,$C9,$4C,$E1,$D0,$F8,$13
	.dc.b	$E8,$DD,$EA,$3C,$6F,$87,$C1,$F1,$27,$31,$8A,$FD,$BD,$7F,$80,$1F
	.dc.b	$DE,$D6,$E3,$94,$60,$DC,$F1,$C3,$F7,$33,$A7,$F6,$F3,$D3,$C7,$D6
	.dc.b	$1E,$08,$E5,$3F,$C0,$3E,$73,$FC,$DF,$18,$94,$62,$79,$9F,$EA,$BF
	.dc.b	$A4,$FF,$05,$D6,$E7,$59,$7B,$5B,$1F,$C4,$9B,$5C,$7C,$9F,$1C,$63
	.dc.b	$0F,$C6,$FC,$B7,$E2,$2B,$FC,$E3,$68,$C4,$7F,$1C,$8C,$1F,$E5,$12
	.dc.b	$BF,$C9,$76,$78,$5F,$01,$B3,$ED,$37,$97,$D3,$8C,$D7,$BB,$07,$CD
	.dc.b	$F4,$8F,$F4,$1B,$5F,$E5,$AD,$7F,$94,$B6,$38,$E3,$E8,$9C,$B6,$33
	.dc.b	$87,$F9,$A6,$B8,$71,$7D,$9C,$7F,$0C,$D9,$EF,$7F,$6A,$31,$87,$F3
	.dc.b	$8C,$3F,$6B,$5F,$E7,$36,$C7,$F4,$76,$BF,$E9,$C6,$BF,$D1,$9A,$F3
	.dc.b	$8F,$34,$60,$FE,$8A,$FA,$78,$9E,$F7,$E3,$1B,$1F,$89,$9F,$0E,$3F
	.dc.b	$9F,$9C,$DF,$CD,$FC,$15,$EA,$FC,$36,$E1,$FE,$8B,$6C,$FF,$5F,$6B
	.dc.b	$FE,$B4,$6B,$FD,$79,$AF,$F5,$A6,$E0,$6F,$2F,$F5,$BA,$FC,$F3,$6B
	.dc.b	$C9,$7F,$A6,$8C,$63,$FC,$6B,$37,$EA,$3F,$B5,$BF,$E3,$27,$F2,$57
	.dc.b	$0F,$F1,$76,$BF,$C7,$5A,$EF,$93,$F0,$93,$D1,$B3,$A8,$C6,$3F,$EA
	.dc.b	$DE,$07,$C4,$FE,$6C,$FE,$01,$3A,$C3,$F3,$13,$AA,$FD,$EC,$BF,$47
	.dc.b	$FF,$03,$C7,$A2,$BF,$9E,$46,$1F,$5C,$ED,$F2,$C6,$5F,$E6,$4D,$7F
	.dc.b	$9C,$35,$E6,$70,$3F,$3D,$0F,$AC,$FC,$C1,$E8,$9F,$06,$7F,$9E,$8F
	.dc.b	$E3,$5E,$1B,$E8,$DE,$17,$EA,$63,$0D,$8C,$E5,$FE,$4A,$6E,$5F,$E4
	.dc.b	$96,$BF,$D3,$AB,$8C,$EE,$B9,$1D,$7E,$93,$FD,$B1,$1F,$C9,$25,$18
	.dc.b	$AF,$E4,$31,$C9,$B1,$9C,$7F,$8C,$7C,$7B,$C8,$FE,$69,$48,$C6,$1D
	.dc.b	$72,$F8,$8F,$F3,$29,$D5,$7F,$3B,$F8,$D7,$CE,$FF,$9A,$6B,$EC,$46
	.dc.b	$3C,$CF,$E1,$8D,$8C,$CE,$EF,$FC,$8D,$EF,$3E,$6F,$FA,$0B,$6B,$C5
	.dc.b	$F1,$98,$9F,$E9,$50,$F8,$2F,$F6,$FB,$F8,$73,$1E,$EF,$F1,$B6,$99
	.dc.b	$8D,$E1,$FA,$48,$CB,$63,$2F,$F8,$02,$32,$D8,$CB,$FE,$58,$AF,$F5
	.dc.b	$D4,$FF,$53,$6C,$7F,$52,$8E,$BB,$FF,$6B,$9C,$BF,$DB,$71,$FE,$23
	.dc.b	$3F,$E6,$26,$D7,$F9,$7B,$F0,$8F,$EF,$B5,$BA,$F8,$A4,$3F,$B1,$46
	.dc.b	$9E,$47,$F6,$AB,$7C,$3D,$3F,$B6,$FD,$5B,$E2,$F1,$5E,$99,$B8,$FF
	.dc.b	$4D,$AD,$4E,$4F,$EE,$7B,$BF,$3D,$CD,$7F,$70,$8C,$61,$FD,$FF,$CB
	.dc.b	$60,$FB,$2F,$C4,$61,$C3,$5F,$F1,$8E,$1C,$32,$EB,$39,$7F,$D1,$38
	.dc.b	$72,$BE,$38,$FF,$81,$B5,$7D,$7D,$33,$7F,$11,$F8,$55,$23,$D3,$D5
	.dc.b	$F9,$CD,$7F,$B5,$6F,$E5,$9F,$B5,$5B,$9F,$65,$6F,$5F,$B5,$3E,$1F
	.dc.b	$76,$5F,$D2,$BA,$7F,$75,$C9,$38,$BF,$F1,$64,$7F,$0C,$AF,$F5,$0B
	.dc.b	$A7,$FC,$3E,$3F,$28,$7D,$CF,$E6,$DB,$1F,$9F,$6E,$1F,$9B,$6B,$F9
	.dc.b	$96,$BF,$CB,$1A,$FF,$99,$9B,$1D,$E7,$FC,$05,$F9,$27,$F9,$AB,$EC
	.dc.b	$3F,$C7,$0D,$9F,$2D,$78,$31,$FF,$34,$CE,$70,$F8,$1E,$D7,$D5,$3F
	.dc.b	$C7,$7F,$39,$FB,$0A,$F8,$31,$9C,$3F,$32,$EB,$A7,$17,$F5,$3E,$C7
	.dc.b	$F3,$47,$F0,$FF,$D5,$BF,$B7,$E1,$FA,$B8,$CE,$1F,$B1,$75,$E6,$62
	.dc.b	$FF,$C8,$79,$FF,$89,$BE,$93,$DB,$E9,$9A,$4C,$3F,$A8,$4E,$DB,$3B
	.dc.b	$6C,$EE,$BF,$B6,$F5,$3E,$59,$97,$F2,$56,$DF,$F9,$6B,$1E,$97,$FA
	.dc.b	$79,$CB,$F9,$2F,$AA,$FD,$C1,$FB,$6F,$B0,$B7,$FC,$36,$2C,$FE,$D6
	.dc.b	$6E,$BF,$8C,$F1,$3F,$48,$7D,$0F,$42,$E1,$FC,$1A,$2D,$FF,$92,$78
	.dc.b	$FF,$C0,$D7,$C6,$BF,$CD,$91,$FE,$75,$9F,$F5,$0F,$A6,$8C,$47,$FB
	.dc.b	$E1,$AF,$FB,$D1,$BE,$5B,$EA,$3F,$8F,$DA,$93,$9F,$8C,$FC,$2F,$BB
	.dc.b	$FD,$B2,$F7,$5F,$C8,$7E,$47,$F7,$C6,$6A,$E1,$FC,$82,$2C,$FF,$60
	.dc.b	$AE,$39,$3B,$55,$F9,$2F,$A0,$FF,$2E,$68,$69,$31,$FF,$08,$5F,$1F
	.dc.b	$8C,$FE,$1C,$71,$8F,$E5,$EF,$E7,$3E,$9B,$99,$23,$8D,$A3,$FC,$D5
	.dc.b	$8B,$EF,$7F,$8C,$1C,$6B,$F9,$CB,$86,$9F,$4A,$71,$A7,$F0,$B9,$BC
	.dc.b	$1F,$EB,$5C,$3B,$5F,$0A,$FB,$9F,$49,$EE,$24,$79,$1F,$1F,$C9,$FF
	.dc.b	$29,$3F,$6A,$D8,$FE,$7B,$AB,$EF,$3F,$BC,$B6,$3F,$BE,$B6,$FF,$42
	.dc.b	$D7,$FD,$CF,$D7,$F3,$6B,$89,$2B,$8B,$FE,$8D,$F2,$3F,$0D,$1F,$DA
	.dc.b	$78,$BF,$C1,$98,$75,$D3,$FB,$44,$63,$EB,$FC,$64,$AF,$1B,$72,$F1
	.dc.b	$B5,$F1,$DA,$3C,$1F,$29,$FD,$94,$E9,$5F,$DE,$5C,$34,$F8,$71,$8A
	.dc.b	$FE,$BD,$FB,$27,$F8,$46,$7F,$C3,$ED,$8F,$E0,$6D,$7E,$1B,$5F,$D1
	.dc.b	$68,$8C,$42,$31,$E9,$FF,$8A,$27,$31,$8A,$FC,$AF,$1F,$FB,$0B,$4F
	.dc.b	$E9,$7F,$0B,$FE,$33,$BF,$14,$62,$BF,$C7,$1F,$53,$FE,$9A,$F7,$BF
	.dc.b	$65,$EF,$BE,$A7,$F6,$9A,$AB,$4A,$74,$FE,$E3,$EA,$BF,$C8,$AE,$6B
	.dc.b	$8D,$70,$AD,$CF,$3F,$25,$87,$82,$F7,$18,$9E,$47,$5F,$71,$96,$E3
	.dc.b	$F8,$46,$3D,$6F,$8E,$1F,$E8,$59,$EC,$D9,$F4,$C7,$FA,$81,$AF,$9B
	.dc.b	$E4,$FF,$CB,$CF,$FA,$36,$11,$88,$E0,$D7,$87,$BF,$D6,$38,$E7,$EE
	.dc.b	$AE,$33,$E7,$AF,$95,$5A,$C3,$F9,$07,$87,$FC,$A7,$C3,$F9,$F8,$F1
	.dc.b	$AD,$5F,$B9,$AF,$2C,$3E,$15,$F1,$8F,$89,$E0,$FF,$1D,$F2,$3E,$CF
	.dc.b	$0F,$F8,$31,$BB,$7F,$83,$1B,$5F,$8D,$E9,$79,$D8,$E7,$FE,$44,$FC
	.dc.b	$B6,$CF,$EF,$A3,$13,$C2,$30,$43,$F8,$0C,$68,$FA,$98,$FF,$9E,$FE
	.dc.b	$53,$C2,$D9,$FF,$33,$47,$87,$F6,$97,$97,$AE,$F4,$D2,$57,$F9,$42
	.dc.b	$B6,$DA,$F4,$46,$23,$FB,$9B,$5F,$E2,$6D,$7F,$62,$D7,$F6,$0D,$C8
	.dc.b	$D3,$58,$7F,$A5,$7E,$37,$CC,$4A,$F2,$79,$DE,$B3,$C6,$47,$FD,$47
	.dc.b	$5F,$E0,$46,$C6,$5F,$3B,$5F,$ED,$FD,$AF,$B0,$F9,$2D,$8F,$F3,$FC
	.dc.b	$7C,$CF,$19,$F7,$F4,$FA,$2E,$27,$FA,$9C,$60,$FA,$B4,$F9,$AE,$A4
	.dc.b	$9F,$F0,$0F,$A5,$5F,$F2,$D4,$5E,$5F,$EA,$13,$79,$7F,$CB,$33,$6B
	.dc.b	$FC,$05,$3E,$53,$C6,$47,$FB,$2B,$E2,$F6,$3D,$A4,$8F,$F0,$8F,$B7
	.dc.b	$ED,$3D,$64,$7F,$B6,$BE,$AF,$17,$E2,$92,$3F,$DC,$9E,$47,$FB,$75
	.dc.b	$FF,$52,$C6,$5F,$F3,$A3,$FA,$4E,$AF,$F0,$B7,$85,$6B,$85,$7F,$82
	.dc.b	$E3,$07,$CF,$F8,$79,$1F,$DF,$9F,$9E,$47,$F9,$AF,$C7,$7D,$82,$3F
	.dc.b	$E2,$CF,$85,$AF,$FA,$16,$2E,$FF,$DE,$E7,$F4,$5B,$CA,$FF,$66,$3F
	.dc.b	$AA,$F8,$CF,$B9,$1F,$E9,$88,$D3,$4F,$F1,$75,$7F,$2A,$5D,$0A,$E3
	.dc.b	$3F,$EB,$0F,$D0,$7F,$19,$5F,$EB,$96,$C6,$79,$BF,$84,$FB,$2F,$0B
	.dc.b	$F5,$D3,$FE,$80,$7F,$55,$19,$ED,$FE,$68,$DC,$3F,$CD,$73,$A2,$4E
	.dc.b	$9B,$1F,$E3,$0F,$85,$FF,$5D,$BE,$AD,$3F,$C9,$1A,$7F,$B2,$BE,$0B
	.dc.b	$F8,$5D,$6A,$BF,$DA,$1F,$0B,$FE,$3E,$78,$1F,$02,$35,$9F,$C3,$BE
	.dc.b	$53,$A6,$BC,$BD,$8F,$D7,$FC,$9C,$8F,$F7,$6B,$FA,$FF,$96,$C4,$7F
	.dc.b	$AF,$62,$DF,$F8,$CE,$7E,$3E,$7F,$D2,$E3,$FD,$4A,$D3,$FC,$C5,$B7
	.dc.b	$F5,$ED,$7F,$9A,$E0,$FC,$09,$C7,$86,$C7,$F5,$F6,$BF,$DE,$0F,$F3
	.dc.b	$63,$87,$0D,$7F,$BE,$B5,$DC,$7F,$69,$7C,$0D,$8D,$69,$FE,$69,$F8
	.dc.b	$7F,$E2,$24,$AF,$F4,$2F,$CC,$57,$FD,$BD,$17,$3F,$F2,$A4,$FD,$6F
	.dc.b	$87,$95,$F3,$F8,$3F,$DC,$6F,$D6,$F1,$DF,$1F,$FC,$96,$18,$7D,$7A
	.dc.b	$DC,$63,$7F,$C1,$C6,$27,$58,$7F,$5B,$BE,$51,$FE,$EE,$F8,$3F,$E5
	.dc.b	$24,$C3,$C1,$C9,$F2,$67,$F4,$D9,$23,$19,$7F,$8D,$9B,$5F,$E3,$56
	.dc.b	$FC,$53,$E4,$BF,$D1,$53,$D0,$FE,$D9,$3C,$2F,$A4,$7F,$72,$AE,$1E
	.dc.b	$7F,$E6,$BE,$22,$4F,$C5,$7F,$7D,$E3,$62,$3E,$01,$F9,$B8,$C6,$7F
	.dc.b	$9B,$EE,$FF,$6D,$4F,$E6,$FC,$87,$D3,$7F,$48,$24,$7F,$AF,$3E,$6A
	.dc.b	$BD,$39,$BF,$D4,$D3,$8C,$3F,$D1,$95,$FA,$EF,$87,$97,$F6,$5F,$CB
	.dc.b	$FC,$63,$EB,$A9,$F8,$B7,$13,$FE,$35,$6B,$F9,$70,$E1,$FC,$D3,$5F
	.dc.b	$F2,$9B,$5F,$F6,$43,$5E,$31,$FD,$6B,$E5,$3F,$B2,$92,$3F,$80,$F9
	.dc.b	$4F,$A1,$7C,$57,$F8,$7F,$E9,$DF,$03,$FE,$1F,$6D,$F9,$79,$91,$88
	.dc.b	$FF,$7E,$B5,$FD,$87,$D1,$3E,$9B,$F7,$2D,$8F,$E2,$21,$C3,$F6,$B2
	.dc.b	$F2,$7C,$67,$F5,$1F,$BA,$FE,$09,$3F,$EE,$46,$C9,$BA,$3E,$37,$FA
	.dc.b	$C3,$6B,$E6,$57,$BE,$7F,$CC,$8D,$8C,$CE,$DF,$E1,$11,$96,$C6,$5F
	.dc.b	$F4,$1C,$65,$B1,$97,$FD,$33,$E6,$3E,$F3,$FC,$94,$D8,$FD,$E3,$5F
	.dc.b	$CC,$FC,$3B,$E3,$BF,$AA,$46,$1A,$FF,$A4,$A3,$5E,$17,$F9,$F5,$B1
	.dc.b	$F8,$96,$BF,$C8,$5A,$F1,$9E,$68,$FF,$48,$B8,$6C,$FF,$95,$BC,$CF
	.dc.b	$F6,$F4,$7F,$43,$6B,$FA,$56,$F4,$FA,$3F,$19,$7B,$5F,$CE,$1B,$5F
	.dc.b	$E9,$36,$BF,$9F,$6B,$FC,$DD,$AF,$FA,$C1,$AF,$9D,$A9,$A6,$BB,$7F
	.dc.b	$88,$7C,$2F,$F9,$B3,$C6,$FF,$4C,$D7,$F5,$B6,$BF,$BD,$F9,$DF,$F1
	.dc.b	$F3,$FD,$7D,$AF,$F4,$9A,$46,$23,$FD,$87,$18,$F7,$7F,$CF,$73,$99
	.dc.b	$FE,$E6,$D7,$FC,$1A,$D9,$E1,$1F,$BF,$6B,$DD,$F2,$E5,$FD,$CD,$BA
	.dc.b	$7F,$7E,$6C,$FF,$86,$3A,$7F,$D8,$CF,$F1,$76,$BE,$66,$BF,$E0,$66
	.dc.b	$BC,$77,$FF,$49,$C6,$27,$FD,$DD,$18,$24,$F9,$FE,$51,$F7,$BF,$E9
	.dc.b	$86,$C7,$F8,$1D,$AF,$F2,$9F,$89,$FF,$08,$37,$D7,$FF,$50,$C6,$23
	.dc.b	$E5,$B5,$FF,$0D,$B5,$FF,$1A,$B7,$0F,$F7,$84,$F2,$24,$FC,$BC,$BF
	.dc.b	$C6,$27,$F8,$90,$AF,$F8,$F9,$B2,$FA,$07,$C8,$7F,$AC,$9B,$5F,$55
	.dc.b	$AF,$E0,$DA,$FE,$09,$AF,$F8,$CB,$D7,$78,$DF,$EB,$C6,$C7,$F8,$AD
	.dc.b	$AF,$F1,$FF,$A4,$EB,$C1,$FE,$31,$9C,$B6,$B9,$1F,$E7,$39,$DE,$5F
	.dc.b	$E4,$0B,$7B,$FF,$EC,$A6,$9F,$E3,$92,$9F,$D6,$CF,$F1,$2D,$FF,$8F
	.dc.b	$DC,$B6,$39,$1C,$FD,$9F,$F1,$8F,$D2,$3D,$8F,$F0,$8D,$FF,$99,$FD
	.dc.b	$CF,$E1,$4A,$5E,$6E,$3D,$05,$6A,$73,$5F,$C4,$67,$2F,$F1,$38,$ED
	.dc.b	$96,$6B,$86,$1E,$0B,$DD,$F7,$C3,$B5,$75,$6C,$73,$AE,$13,$CA,$FF
	.dc.b	$90,$74,$72,$C7,$C7,$7F,$BE,$AE,$4E,$9E,$0E,$4E,$95,$E5,$7D,$18
	.dc.b	$78,$1F,$21,$93,$E2,$61,$E2,$78,$9F,$24,$E2,$6E,$FC,$78,$F9,$67
	.dc.b	$84,$F2,$BF,$F0,$33,$4F,$F0,$45,$7C,$86,$CF,$D6,$7F,$CC,$50,$F7
	.dc.b	$1F,$49,$CA,$30,$D8,$CC,$F4,$6B,$E0,$E6,$FF,$32,$26,$91,$FF,$34
	.dc.b	$3F,$25,$FE,$B1,$BD,$7F,$0F,$15,$74,$FF,$AC,$F6,$FB,$B1,$FD,$56
	.dc.b	$75,$39,$DB,$FC,$4D,$18,$BE,$55,$E3,$3E,$B1,$FD,$76,$B8,$3F,$E8
	.dc.b	$39,$E0,$E8,$EC,$58,$DB,$E6,$31,$3D,$A7,$8D,$F9,$A7,$C0,$FA,$2B
	.dc.b	$BF,$33,$D6,$7C,$CF,$F7,$08,$C3,$63,$38,$FF,$43,$6E,$1F,$D4,$5A
	.dc.b	$FF,$50,$6B,$FE,$C0,$6B,$FE,$DA,$6F,$8E,$FA,$47,$C6,$7F,$80,$23
	.dc.b	$0D,$8C,$DF,$F8,$99,$B7,$FE,$22,$6B,$FE,$20,$6B,$FE,$09,$AE,$73
	.dc.b	$C6,$BB,$7A,$9F,$E4,$C8,$FF,$18,$B6,$BF,$C5,$AD,$7F,$C5,$6D,$7F
	.dc.b	$9C,$7D,$27,$C7,$93,$48,$FF,$8F,$3E,$F5,$F0,$9E,$D6,$C6,$63,$04
	.dc.b	$70,$4A,$FF,$1F,$CF,$07,$FC,$01,$E0,$F8,$27,$FB,$05,$FD,$98,$C4
	.dc.b	$EA,$3E,$67,$4B,$E0,$29,$2F,$FC,$A3,$19,$6C,$66,$BF,$C9,$F3,$96
	.dc.b	$C7,$E1,$F8,$15,$48,$7F,$82,$5C,$BF,$E5,$EF,$C3,$3F,$AD,$B7,$C2
	.dc.b	$FC,$04,$62,$FF,$CB,$AD,$7F,$CB,$6D,$7E,$E3,$56,$E8,$C3,$67,$F3
	.dc.b	$9E,$3F,$FA,$22,$FF,$C3,$8D,$BF,$F1,$53,$5F,$F1,$4B,$5F,$E9,$3C
	.dc.b	$4F,$4B,$FC,$A4,$DA,$FE,$3E,$D9,$FF,$53,$35,$FF,$7F,$35,$E7,$3C
	.dc.b	$89,$3F,$E6,$5F,$A6,$FD,$3F,$31,$A4,$E6,$F0,$FB,$B2,$BC,$2D,$C3
	.dc.b	$C2,$D7,$C2,$D7,$F8,$87,$C1,$7F,$9B,$67,$FC,$E8,$D8,$FF,$3A,$35
	.dc.b	$FF,$3E,$35,$FE,$D7,$E3,$3C,$CF,$F3,$A3,$6B,$FA,$0B,$67,$FD,$90
	.dc.b	$D7,$FB,$98,$70,$F4,$9E,$64,$9F,$F5,$2F,$D2,$BE,$37,$F8,$33,$63
	.dc.b	$31,$83,$3E,$3C,$8E,$2D,$BF,$1F,$C5,$FA,$09,$3E,$87,$AF,$91,$23
	.dc.b	$C2,$D9,$FF,$71,$47,$F5,$73,$FA,$F9,$5C,$F8,$7F,$E9,$78,$C4,$7C
	.dc.b	$06,$BF,$E1,$46,$BF,$4E,$38,$1F,$EA,$28,$7C,$6B,$E5,$30,$D8,$F1
	.dc.b	$B6,$BB,$35,$FF,$41,$FC,$1F,$F5,$A9,$FF,$3E,$36,$3F,$D1,$4D,$77
	.dc.b	$C7,$FD,$0E,$31,$1F,$EB,$3E,$A4,$62,$11,$89,$E0,$7F,$4B,$F8,$87
	.dc.b	$C4,$D2,$61,$FE,$8D,$8C,$B6,$33,$7F,$E9,$D9,$CB,$63,$FD,$8B,$D0
	.dc.b	$97,$D1,$61,$B7,$FB,$26,$DF,$EC,$9A,$FE,$F1,$AF,$FB,$3E,$BA,$57
	.dc.b	$32,$57,$E6,$25,$EA,$3E,$27,$CB,$1F,$E4,$E8,$C6,$3F,$ED,$A6,$D7
	.dc.b	$16,$CF,$E8,$9F,$CE,$71,$6F,$8F,$FC,$CF,$5B,$6D,$6D,$AF,$76,$BF
	.dc.b	$EE,$7A,$E7,$3C,$6B,$A1,$F6,$E6,$47,$E0,$9A,$7E,$08,$A7,$E2,$0A
	.dc.b	$7F,$9A,$BD,$9F,$E3,$4F,$FB,$D5,$A9,$A4,$EC,$FF,$7C,$36,$7F,$C8
	.dc.b	$AF,$25,$F4,$5F,$EF,$47,$F5,$AD,$AF,$D4,$35,$FD,$43,$5F,$CA,$7C
	.dc.b	$97,$E7,$DB,$C9,$FE,$1B,$8C,$47,$E5,$9B,$C4,$69,$27,$81,$F6,$7E
	.dc.b	$8B,$FA,$15,$7E,$01,$BA,$1A,$4F,$86,$E0,$4C,$BF,$01,$1C,$AB,$FC
	.dc.b	$A9,$F2,$DF,$DA,$1F,$F2,$3B,$5F,$F2,$13,$5F,$F2,$9B,$5F,$F2,$83
	.dc.b	$5E,$73,$FE,$61,$8C,$7A,$07,$86,$AF,$29,$F1,$98,$6D,$FE,$76,$75
	.dc.b	$39,$3F,$75,$F1,$D2,$7F,$70,$DC,$7F,$6C,$D7,$FD,$5F,$BF,$ED,$63
	.dc.b	$11,$F5,$BC,$4F,$A8,$4A,$FD,$65,$7D,$38,$C4,$FF,$46,$8C,$4E,$B8
	.dc.b	$3F,$C9,$71,$89,$FE,$97,$E4,$3C,$1F,$E9,$97,$FD,$23,$C3,$FE,$8E
	.dc.b	$47,$97,$CC,$FE,$14,$8E,$0F,$DF,$53,$E5,$1F,$45,$FE,$55,$6C,$E9
	.dc.b	$B8,$F8,$9A,$FF,$B4,$2B,$95,$71,$25,$71,$7A,$76,$C8,$E8,$DA,$F8
	.dc.b	$5A,$3E,$47,$F3,$84,$AF,$88,$F9,$7E,$5E,$47,$95,$B5,$CB,$E8,$BF
	.dc.b	$48,$49,$F9,$8D,$8D,$46,$08,$E0,$87,$DE,$B4,$F1,$8F,$81,$FB,$27
	.dc.b	$F9,$49,$BE,$9F,$F9,$F6,$31,$1F,$E7,$56,$BF,$B5,$F8,$8F,$DC,$13
	.dc.b	$0F,$C0,$3F,$A1,$E0,$7D,$2B,$E4,$F0,$DA,$FF,$07,$B6,$FF,$C1,$F5
	.dc.b	$C2,$B6,$4A,$DB,$67,$F9,$A6,$AF,$A4,$FE,$74,$DA,$FE,$74,$D9,$DC
	.dc.b	$F8,$BC,$59,$5E,$26,$CF,$8B,$1F,$E8,$E8,$FF,$11,$38,$6C,$70,$7F
	.dc.b	$A6,$5B,$E8,$FC,$64,$BF,$1B,$67,$D2,$D7,$8B,$F2,$7F,$5F,$7C,$3D
	.dc.b	$25,$7F,$8D,$A3,$2D,$8C,$D7,$CA,$9C,$B6,$3F,$AF,$7D,$27,$F8,$5A
	.dc.b	$BF,$AF,$36,$7F,$AF,$35,$FE,$CC,$D7,$E1,$0F,$A3,$78,$9F,$EF,$08
	.dc.b	$FF,$7D,$7D,$1E,$A7,$3E,$7F,$FB,$C6,$3F,$D7,$DF,$48,$F5,$7F,$96
	.dc.b	$78,$E9,$E1,$8F,$38,$C6,$3F,$47,$1E,$6F,$63,$13,$BA,$EB,$7D,$1E
	.dc.b	$78,$F8,$6B,$6F,$F8,$15,$F5,$57,$A5,$FF,$2F,$DF,$66,$C6,$5F,$67
	.dc.b	$AD,$FE,$63,$7F,$C1,$D3,$FE,$6D,$E5,$E8,$D8,$CC,$ED,$EA,$F1,$CB
	.dc.b	$DB,$1F,$2E,$31,$8F,$6B,$F4,$77,$7F,$61,$F0,$3E,$8E,$1C,$31,$EC
	.dc.b	$DC,$38,$E3,$F1,$DB,$7C,$63,$E1,$79,$5F,$C5,$7C,$BF,$84,$D8,$FD
	.dc.b	$3C,$63,$97,$FC,$B5,$1F,$E8,$4F,$9C,$7D,$FF,$FA,$21,$B1,$98,$C1
	.dc.b	$1C,$10,$FD,$EC,$70,$7F,$C5,$9F,$46,$FA,$0F,$1B,$63,$FC,$64,$D7
	.dc.b	$71,$A2,$3A,$6B,$FC,$2B,$DE,$7C,$AF,$FA,$79,$B1,$FE,$9E,$6B,$B8
	.dc.b	$D1,$1D,$35,$F3,$F8,$BE,$62,$3E,$67,$B7,$C2,$C8,$EC,$D7,$C8,$D7
	.dc.b	$FD,$6B,$48,$C4,$7F,$89,$BE,$8F,$F2,$DE,$67,$F8,$85,$BE,$41,$A7
	.dc.b	$E6,$F8,$5E,$33,$E9,$B0,$DC,$3F,$B3,$B5,$FE,$E6,$D7,$FA,$37,$A1
	.dc.b	$D8,$8F,$F7,$B6,$BF,$D7,$9A,$F0,$34,$43,$45,$3F,$B7,$79,$BF,$DD
	.dc.b	$DF,$EE,$8F,$A4,$C3,$FD,$D5,$F7,$FB,$12,$3F,$C1,$8D,$7F,$BD,$7A
	.dc.b	$9C,$48,$FF,$7F,$7A,$78,$12,$3F,$C0,$4D,$7F,$C0,$4D,$7C,$AD,$7F
	.dc.b	$CA,$FE,$C7,$D0,$23,$FE,$1D,$7D,$1E,$04,$8F,$F0,$73,$5F,$F0,$1F
	.dc.b	$BF,$F6,$08,$FF,$89,$5A,$E8,$C1,$0F,$F1,$67,$C8,$FD,$F7,$FC,$62
	.dc.b	$D7,$3B,$FE,$1A,$31,$1F,$E3,$C6,$BF,$EC,$36,$BF,$EC,$26,$F0,$9C
	.dc.b	$FA,$DF,$EF,$FE,$7F,$F4,$0F,$85,$FB,$1B,$FF,$46,$35,$FF,$46,$35
	.dc.b	$FE,$CD,$E5,$7F,$97,$DF,$EC,$5E,$47,$F2,$F6,$9F,$CC,$63,$44,$8D
	.dc.b	$36,$3F,$D6,$5E,$DB,$E5,$3F,$8B,$B6,$33,$18,$23,$82,$1F,$EB,$98
	.dc.b	$E0,$FF,$8D,$BE,$07,$F5,$64,$7F,$CB,$0D,$74,$60,$87,$FB,$37,$E8
	.dc.b	$9B,$FB,$5B,$1F,$D3,$5A,$EE,$3D,$BF,$35,$23,$E1,$35,$FF,$6C,$62
	.dc.b	$8C,$47,$F7,$03,$FC,$18,$E1,$FF,$55,$FC,$9F,$F0,$42,$3F,$E7,$16
	.dc.b	$BF,$E5,$46,$BF,$E4,$37,$FA,$0F,$C2,$BC,$0A,$49,$FF,$71,$46,$5B
	.dc.b	$19,$7F,$DC,$91,$96,$C7,$D6,$61,$F6,$8A,$7A,$FE,$09,$E0,$FF,$99
	.dc.b	$E3,$FD,$F1,$18,$F1,$BF,$BB,$36,$33,$3B,$79,$35,$FF,$0A,$C6,$DC
	.dc.b	$CF,$29,$D4,$7E,$8F,$B7,$F5,$04,$7F,$5A,$D7,$F6,$2D,$7F,$58,$FF
	.dc.b	$32,$F8,$B9,$1F,$E9,$F6,$BF,$E8,$26,$BF,$E7,$16,$BF,$E2,$AF,$8D
	.dc.b	$7D,$17,$E9,$5B,$1F,$E1,$56,$BF,$81,$0E,$1F,$C2,$7E,$3D,$FD,$D9
	.dc.b	$D3,$5C,$F9,$7F,$B1,$8C,$47,$F8,$E9,$AF,$F9,$75,$A9,$A6,$BC,$57
	.dc.b	$CE,$BE,$8F,$FA,$83,$6B,$FD,$66,$D7,$FD,$7E,$D7,$F3,$5F,$42,$FA
	.dc.b	$2F,$E0,$8D,$8F,$F1,$A3,$5F,$80,$D7,$FD,$6D,$2F,$8A,$F0,$92,$3C
	.dc.b	$2F,$93,$E8,$64,$79,$1A,$F4,$F1,$3C,$24,$7C,$2D,$74,$60,$87,$EC
	.dc.b	$FD,$87,$C0,$FF,$BF,$5B,$E0,$7F,$93,$5B,$3F,$65,$AF,$B3,$E0,$BE
	.dc.b	$09,$1F,$82,$D7,$46,$08,$7E,$0F,$E8,$3F,$C3,$2F,$F0,$56,$BF,$E7
	.dc.b	$66,$BF,$E9,$96,$AD,$3F,$87,$3F,$43,$B3,$E9,$91,$C1,$0F,$A6,$E5
	.dc.b	$FF,$80,$1F,$30,$FA,$0F,$F0,$BB,$63,$F3,$1E,$4A,$31,$1C,$27,$FC
	.dc.b	$31,$F2,$3F,$E6,$28,$FF,$79,$35,$FF,$58,$72,$23,$10,$8C,$7B,$DC
	.dc.b	$89,$1F,$A7,$8C,$B6,$78,$E0,$F6,$BF,$CF,$93,$FE,$BD,$8C,$6D,$FE
	.dc.b	$30,$9C,$CA,$72,$F5,$FA,$C7,$C6,$FF,$8C,$5B,$19,$8C,$11,$C1,$0F
	.dc.b	$E7,$98,$A7,$8C,$7F,$3F,$A7,$BD,$FD,$62,$3F,$D6,$31,$8F,$2F,$FC
	.dc.b	$8B,$39,$9F,$53,$5F,$8D,$F5,$9F,$E4,$D7,$E2,$35,$F0,$B5,$FF,$5E
	.dc.b	$E0,$8C,$47,$D2,$6B,$D3,$DA,$FF,$28,$46,$27,$84,$7F,$96,$3E,$85
	.dc.b	$F2,$F4,$97,$F8,$48,$CB,$F5,$FE,$7D,$F5,$1F,$E5,$B6,$C6,$63,$07
	.dc.b	$F9,$7F,$E8,$B8,$C7,$CA,$6B,$EA,$6B,$F8,$A6,$BF,$89,$6E,$26,$93
	.dc.b	$0F,$F3,$8C,$FF,$7A,$3E,$89,$4F,$C4,$14,$FC,$31,$75,$34,$9E,$7F
	.dc.b	$43,$FC,$49,$8B,$CE,$7D,$17,$0E,$87,$D0,$FF,$BC,$29,$E5,$7F,$BC
	.dc.b	$A7,$38,$F6,$9C,$BC,$6B,$86,$1D,$23,$F6,$51,$8C,$35,$1C,$70,$E9
	.dc.b	$3D,$DB,$1D,$6B,$96,$1E,$08,$61,$E9,$C3,$9C,$62,$F8,$B4,$CC,$ED
	.dc.b	$F1,$9D,$67,$33,$C2,$7D,$73,$A3,$BD,$F8,$B0,$F5,$46,$67,$78,$79
	.dc.b	$5B,$3A,$8C,$57,$C3,$AF,$F1,$65,$78,$DB,$19,$7A,$9A,$F4,$DF,$45
	.dc.b	$FD,$CE,$3E,$0C,$61,$FE,$ED,$48,$C4,$71,$9F,$F3,$13,$FB,$C3,$FC
	.dc.b	$C9,$C9,$82,$38,$3E,$00,$E9,$78,$3F,$3A,$3F,$1B,$36,$73,$1C,$0E
	.dc.b	$D9,$7D,$77,$95,$71,$73,$9F,$F5,$B8,$C5,$F3,$6C,$F8,$AF,$AD,$F3
	.dc.b	$8E,$27,$E7,$8E,$A7,$86,$B9,$34,$E5,$5D,$27,$C9,$5D,$2F,$F9,$97
	.dc.b	$A0,$F9,$6F,$5E,$1F,$C8,$23,$18,$7F,$4C,$6C,$7F,$27,$6B,$B9,$E2
	.dc.b	$F4,$6B,$FC,$EE,$36,$E6,$7F,$A4,$7D,$13,$E8,$B7,$CC,$8C,$4E,$A7
	.dc.b	$31,$8A,$FE,$76,$D7,$F9,$5B,$5F,$EB,$6D,$7F,$AD,$36,$BF,$C7,$BE
	.dc.b	$67,$83,$CF,$FF,$17,$61,$FD,$6D,$B7,$FD,$21,$AF,$F4,$48,$78,$FE
	.dc.b	$EE,$7F,$C3,$B6,$32,$7F,$74,$2C,$FF,$52,$FA,$27,$6F,$C0,$6C,$FF
	.dc.b	$6B,$6C,$EE,$74,$49,$D3,$63,$FA,$D7,$CD,$76,$7F,$AC,$B5,$FE,$E0
	.dc.b	$D7,$FB,$73,$73,$FE,$3F,$F4,$5F,$4B,$A4,$D2,$61,$FD,$FE,$32,$D8
	.dc.b	$CD,$FF,$6E,$9C,$BF,$DB,$73,$FD,$5D,$7F,$85,$9B,$19,$8F,$95,$18
	.dc.b	$8F,$EC,$BF,$45,$FD,$72,$7F,$C1,$CD,$8F,$EF,$6D,$77,$DB,$F8,$A8
	.dc.b	$C4,$7F,$87,$1A,$FF,$86,$DB,$89,$CD,$FF,$A9,$BC,$DF,$B9,$EB,$1A
	.dc.b	$4C,$7F,$D4,$FE,$87,$E7,$B1,$72,$FF,$99,$E7,$24,$8F,$F5,$7C,$EE
	.dc.b	$75,$18,$9D,$3F,$BD,$E8,$7D,$17,$EA,$23,$F8,$A3,$63,$3D,$1F,$AE
	.dc.b	$6D,$7F,$AE,$BE,$8B,$B1,$1C,$12,$FF,$D8,$31,$96,$C7,$F9,$A7,$DB
	.dc.b	$FE,$D6,$D8,$ED,$18,$7F,$CE,$8D,$7F,$DD,$1E,$FB,$E4,$3E,$23,$6B
	.dc.b	$F9,$D3,$67,$FD,$5C,$D7,$8C,$F0,$24,$F0,$7F,$82,$FC,$F3,$E2,$69
	.dc.b	$27,$FC,$A5,$19,$6F,$C3,$F0,$23,$83,$C7,$F1,$1E,$62,$38,$39,$F9
	.dc.b	$3E,$02,$38,$3F,$AA,$1F,$D4,$4B,$E3,$7E,$0D,$C6,$5F,$EC,$B6,$BF
	.dc.b	$CE,$4F,$E9,$B1,$B2,$47,$23,$87,$A4,$F7,$7F,$95,$C7,$F9,$D5,$B1
	.dc.b	$98,$DC,$7F,$13,$FA,$2F,$E8,$B7,$FE,$5A,$6B,$FE,$71,$6B,$FE,$6F
	.dc.b	$6F,$41,$A4,$E2,$E8,$7F,$1F,$FA,$27,$CD,$52,$63,$FE,$92,$8C,$B6
	.dc.b	$33,$87,$F9,$BE,$72,$D8,$FE,$8F,$CE,$97,$CF,$E1,$B7,$FE,$F5,$9D
	.dc.b	$4E,$49,$19,$9D,$DF,$CF,$AD,$CE,$6B,$FA,$AF,$AE,$F8,$8F,$1D,$FF
	.dc.b	$6D,$8C,$63,$A9,$CC,$FF,$93,$A7,$37,$F6,$E3,$FD,$3D,$C8,$E3,$EA
	.dc.b	$D9,$D5,$7E,$66,$72,$48,$CC,$EE,$BF,$0D,$5B,$9C,$D7,$DD,$99,$1F
	.dc.b	$EC,$0D,$7F,$65,$1F,$74,$C9,$FE,$9D,$72,$FD,$F3,$27,$E4,$FE,$2B
	.dc.b	$F8,$9B,$7B,$3F,$8E,$46,$23,$FC,$20,$D7,$FC,$DE,$D4,$D2,$47,$E6
	.dc.b	$3E,$89,$F2,$74,$98,$7E,$96,$32,$FF,$B6,$BE,$4F,$F3,$98,$7F,$AD
	.dc.b	$5A,$FF,$AD,$1A,$FF,$B2,$DA,$FF,$B2,$5B,$C6,$69,$39,$3F,$03,$F4
	.dc.b	$5F,$CF,$6F,$EA,$B6,$FE,$8B,$5F,$C2,$35,$FC,$13,$5F,$F6,$F3,$76
	.dc.b	$34,$D4,$7F,$B3,$BE,$47,$FA,$47,$A6,$69,$2F,$FC,$C1,$19,$6C,$66
	.dc.b	$7F,$21,$39,$7E,$D7,$94,$F8,$AF,$F1,$8F,$02,$31,$87,$E5,$E3,$13
	.dc.b	$A9,$CD,$7E,$D7,$DF,$FD,$F1,$2F,$F7,$8D,$8D,$46,$08,$FD,$7F,$13
	.dc.b	$FC,$49,$B7,$F5,$9D,$FE,$94,$60,$95,$F7,$DB,$5C,$63,$F9,$DF,$85
	.dc.b	$FD,$00,$8E,$0F,$C3,$7C,$CF,$62,$38,$3F,$69,$E5,$3E,$4D,$F4,$5F
	.dc.b	$E4,$E6,$CE,$9B,$87,$F3,$5A,$E1,$5B,$25,$6D,$B3,$BA,$FE,$1E,$FF
	.dc.b	$BF,$1A,$FF,$B5,$2B,$C5,$DB,$E1,$3E,$2F,$C7,$4C,$3F,$40,$DA,$E9
	.dc.b	$F1,$FE,$72,$4F,$9D,$F9,$BF,$45,$23,$F5,$0D,$9F,$D4,$35,$FD,$43
	.dc.b	$5F,$D4,$35,$FD,$43,$70,$7C,$B7,$F9,$14,$9E,$B7,$F9,$D9,$FF,$25
	.dc.b	$70,$7E,$B7,$E3,$3E,$F9,$31,$FD,$D3,$FE,$82,$E0,$7D,$03,$E8,$B0
	.dc.b	$DA,$FF,$01,$B7,$4F,$F0,$5B,$5F,$F5,$05,$72,$AE,$24,$AE,$2F,$F3
	.dc.b	$CD,$9F,$45,$49,$3F,$CC,$A3,$2F,$83,$C6,$F1,$91,$C1,$30,$F1,$C6
	.dc.b	$5B,$1C,$FE,$57,$C0,$47,$04,$8F,$E7,$74,$9D,$46,$1F,$8D,$F4,$4F
	.dc.b	$A2,$A4,$BF,$F1,$3C,$65,$B1,$9A,$FF,$18,$CE,$5B,$1F,$D7,$BE,$89
	.dc.b	$E7,$52,$5F,$F5,$48,$CB,$FE,$EF,$FA,$1F,$F3,$07,$9F,$FE,$E0,$F9
	.dc.b	$0F,$F6,$84,$63,$17,$86,$9F,$CF,$7D,$13,$D4,$FF,$72,$CE,$68,$2F
	.dc.b	$94,$66,$31,$ED,$7F,$BC,$5E,$B8,$78,$23,$13,$CF,$A9,$8F,$7C,$FC
	.dc.b	$D8,$76,$C7,$AE,$3C,$73,$F6,$65,$D6,$BF,$C6,$55,$EB,$6C,$65,$F7
	.dc.b	$FA,$EE,$6F,$88,$D8,$FF,$00,$57,$F9,$7E,$72,$48,$F4,$9E,$39,$E1
	.dc.b	$A7,$BE,$B8,$4E,$B4,$FE,$AD,$E0,$FF,$BF,$6B,$57,$C3,$3F,$1E,$3E
	.dc.b	$EC,$3E,$54,$E5,$61,$C9,$F8,$7E,$53,$97,$FC,$95,$1F,$A1,$F2,$FF
	.dc.b	$C3,$0A,$75,$18,$6C,$7F,$9D,$BE,$85,$F4,$5F,$E7,$A6,$C6,$63,$04
	.dc.b	$70,$43,$FD,$0B,$1C,$1F,$F0,$EF,$D1,$3E,$7F,$C0,$D8,$FF,$13,$B5
	.dc.b	$DC,$68,$8E,$9A,$FF,$A4,$BD,$FF,$F0,$7B,$FE,$8E,$6B,$9F,$4B,$FD
	.dc.b	$27,$18,$8F,$E1,$AD,$7F,$2F,$B2,$31,$1F,$83,$6C,$66,$7D,$3E,$34
	.dc.b	$AF,$4B,$5F,$27,$C2,$78,$08,$F9,$5A,$E8,$C1,$0F,$F1,$17,$D0,$FD
	.dc.b	$8F,$17,$FC,$38,$DF,$34,$D3,$EE,$F1,$7F,$3A,$FA,$2B,$E4,$1A,$4C
	.dc.b	$3F,$C6,$1E,$8F,$32,$38,$21,$FE,$33,$72,$D3,$27,$F8,$D5,$CB,$4F
	.dc.b	$ED,$DE,$6F,$90,$FE,$E8,$E0,$87,$F8,$FF,$D9,$F6,$11,$FF,$22,$3C
	.dc.b	$7D,$49,$1F,$DE,$5A,$FF,$73,$F0,$BB,$11,$FE,$FC,$D7,$46,$08,$7F
	.dc.b	$93,$FD,$9F,$98,$47,$FC,$A8,$FA,$7C,$09,$1F,$E0,$96,$BF,$E5,$7F
	.dc.b	$7B,$EA,$11,$FF,$2E,$35,$D1,$82,$1F,$EC,$5E,$07,$D1,$7D,$C8,$FF
	.dc.b	$47,$46,$38,$3F,$05,$39,$9E,$0D,$7F,$97,$89,$E0,$F2,$8D,$3F,$CD
	.dc.b	$23,$2F,$29,$FE,$0B,$EB,$FF,$21,$E8,$BC,$FF,$B3,$AF,$F4,$1B,$5F
	.dc.b	$F4,$1B,$5F,$E7,$FE,$5B,$C6,$FE,$1F,$1F,$D8,$5B,$19,$8F,$E2,$6D
	.dc.b	$7F,$D5,$8D,$7F,$DD,$0D,$7D,$AD,$7B,$C7,$F1,$9F,$71,$F2,$F4,$93
	.dc.b	$FE,$B2,$8C,$B6,$32,$FF,$AC,$23,$2D,$8F,$D7,$7C,$04,$8F,$F4,$B3
	.dc.b	$5F,$F4,$B3,$5F,$F4,$B3,$5F,$A7,$F4,$4D,$FE,$1B,$63,$F9,$9B,$5F
	.dc.b	$F6,$63,$5F,$83,$F3,$7F,$20,$8F,$C8,$6B,$F2,$1A,$F0,$3F,$D5,$9F
	.dc.b	$29,$FC,$1C,$8E,$08,$7F,$AB,$DC,$B4,$FF,$6C,$FC,$2B,$BF,$FA,$33
	.dc.b	$63,$31,$82,$38,$21,$FE,$DC,$8E,$0F,$CF,$3E,$C1,$0F,$AC,$7F,$56
	.dc.b	$F8,$3F,$E6,$5E,$45,$F0,$FF,$AE,$4F,$FB,$C6,$72,$D8,$CB,$FF,$00
	.dc.b	$5A,$FE,$92,$36,$E6,$7A,$1F,$AC,$ED,$91,$FA,$76,$BF,$AE,$6B,$FA
	.dc.b	$EE,$6F,$DB,$C6,$23,$FD,$A9,$F1,$5F,$CD,$48,$FF,$B7,$5A,$FF,$B7
	.dc.b	$5A,$FF,$B7,$5F,$D3,$FC,$6B,$E8,$A9,$27,$FB,$CC,$65,$FE,$F7,$F4
	.dc.b	$5F,$DF,$9F,$E0,$27,$F8,$C1,$C3,$86,$BF,$C1,$9A,$EF,$CC,$FD,$BC
	.dc.b	$62,$0D,$37,$E2,$BE,$85,$F2,$3F,$D4,$1B,$5F,$EB,$26,$BF,$EA,$D6
	.dc.b	$BE,$6F,$9F,$7D,$17,$F0,$66,$C7,$F8,$59,$AF,$B0,$38,$7F,$8F,$DF
	.dc.b	$8F,$E2,$BC,$44,$8F,$13,$E5,$FA,$29,$1E,$56,$BD,$69,$E2,$78,$C9
	.dc.b	$1E,$36,$BB,$6B,$EB,$7F,$C5,$BE,$C3,$E0,$BF,$8B,$37,$C0,$CC,$60
	.dc.b	$FE,$35,$5E,$CF,$83,$F8,$44,$7E,$13,$5F,$33,$5F,$94,$FF,$2E,$FA
	.dc.b	$2F,$F0,$BC,$1A,$4F,$4B,$E7,$77,$49,$F9,$CD,$7E,$13,$5F,$AC,$D7
	.dc.b	$FA,$47,$CC,$BE,$8B,$FC,$28,$D8,$FE,$AE,$D7,$FA,$E3,$5F,$AF,$F0
	.dc.b	$1F,$E8,$16,$FB,$7F,$E1,$88,$C4,$7F,$BA,$5A,$FF,$97,$9B,$C5,$FE
	.dc.b	$9C,$9E,$04,$9F,$EC,$FE,$D7,$F9,$ED,$FE,$BC,$D7,$FB,$23,$7B,$7F
	.dc.b	$CC,$5C,$1F,$E2,$8F,$A2,$7C,$65,$25,$FE,$47,$14,$EA,$30,$D7,$F6
	.dc.b	$31,$A9,$E3,$18,$9D,$3F,$C2,$7D,$EF,$E9,$ED,$F2,$BF,$C7,$D1,$88
	.dc.b	$FE,$EC,$D7,$FC,$FF,$F4,$4F,$6F,$FC,$7E,$D8,$F3,$35,$F1,$B5,$FD
	.dc.b	$3B,$5F,$F5,$AE,$28,$C4,$76,$9E,$A7,$F9,$47,$E8,$9F,$31,$49,$3F
	.dc.b	$E1,$68,$CB,$F1,$FE,$7D,$F4,$5F,$E5,$46,$C6,$63,$07,$F9,$63,$E8
	.dc.b	$B8,$C7,$CA,$6B,$FB,$E6,$BE,$86,$BF,$EA,$26,$E4,$69,$30,$FF,$35
	.dc.b	$4F,$F6,$D3,$FC,$44,$53,$FC,$50,$5E,$0F,$F6,$CF,$A1,$C0,$91,$FE
	.dc.b	$34,$34,$71,$18,$BC,$F7,$D1,$7F,$B8,$B9,$DF,$43,$FE,$36,$A7,$97
	.dc.b	$FE,$38,$9C,$E1,$FE,$E8,$9C,$BC,$6B,$86,$1D,$21,$5C,$1F,$0C,$66
	.dc.b	$31,$5C,$A7,$75,$E1,$BF,$03,$D3,$0E,$D3,$E1,$C3,$94,$EE,$32,$7B
	.dc.b	$AB,$8B,$BA,$CD,$FC,$16,$CE,$A7,$81,$AB,$EA,$F1,$3C,$A7,$C1,$8E
	.dc.b	$15,$E9,$C3,$D7,$5C,$2B,$DF,$7F,$ED,$CA,$F0,$47,$F4,$76,$BF,$D9
	.dc.b	$1A,$FD,$1F,$4D,$F4,$5C,$5B,$1C,$5A,$F1,$6B,$FD,$4F,$93,$FC,$BA
	.dc.b	$9A,$47,$FC,$C2,$9F,$F3,$EE,$7F,$DD,$BA,$3F,$CC,$5E,$DF,$D9,$9D
	.dc.b	$56,$E3,$13,$AA,$F0,$57,$23,$FC,$FF,$19,$7F,$C0,$35,$C6,$7B,$78
	.dc.b	$76,$72,$7C,$E3,$C9,$5F,$E7,$3D,$7C,$51,$EB,$34,$23,$55,$E2,$C3
	.dc.b	$CB,$7D,$6B,$FD,$23,$85,$78,$45,$7A,$4F,$07,$D4,$70,$29,$C0,$F7
	.dc.b	$70,$3D,$17,$9B,$FE,$89,$DF,$FD,$15,$39,$8C,$61,$FE,$9D,$F6,$3E
	.dc.b	$83,$F4,$6B,$77,$CA,$30,$7F,$A8,$DC,$61,$FB,$88,$C5,$F4,$8F,$F5
	.dc.b	$65,$F4,$74,$7F,$AB,$EB,$8D,$F9,$3A,$3F,$A5,$E1,$DA,$BF,$D6,$91
	.dc.b	$87,$8F,$3B,$E8,$5F,$80,$77,$D1,$7B,$C7,$A7,$7E,$63,$19,$7F,$3A
	.dc.b	$C3,$8B,$DE,$F8,$D7,$A6,$DF,$94,$BE,$63,$FD,$59,$18,$6C,$66,$FF
	.dc.b	$C6,$13,$97,$FC,$65,$E1,$3F,$36,$70,$7C,$58,$C3,$DE,$F9,$6F,$F1
	.dc.b	$A7,$37,$E5,$1E,$57,$FA,$F7,$4F,$F6,$43,$63,$27,$FB,$88,$B5,$FE
	.dc.b	$E7,$6B,$FE,$73,$6B,$FE,$DD,$6B,$FE,$E3,$6C,$FE,$DF,$BB,$E7,$72
	.dc.b	$FF,$5C,$9F,$25,$E2,$7E,$95,$7F,$5A,$9F,$F7,$9B,$6F,$FD,$F0,$D7
	.dc.b	$FD,$08,$D7,$FD,$DC,$D7,$FD,$E8,$D3,$FD,$ED,$E0,$7C,$17,$1D,$2F
	.dc.b	$1B,$F0,$4D,$BF,$B9,$3E,$1D,$FF,$43,$39,$C3,$F2,$73,$EF,$AF,$39
	.dc.b	$D4,$86,$09,$1E,$A7,$2F,$F8,$E3,$EA,$FC,$27,$C2,$8A,$7E,$6E,$DF
	.dc.b	$81,$BD,$1F,$C5,$B6,$7F,$12,$DB,$FF,$4A,$D7,$0A,$D9,$2B,$6D,$9F
	.dc.b	$F4,$DE,$1F,$CB,$7E,$57,$EB,$9F,$9B,$C3,$A6,$1F,$C0,$B7,$4F,$ED
	.dc.b	$FB,$9F,$47,$F9,$2C,$7F,$93,$B5,$FE,$42,$D7,$F8,$F3,$63,$8F,$33
	.dc.b	$4F,$F4,$94,$61,$B3,$B3,$96,$7F,$D0,$FE,$63,$F1,$A7,$8E,$3F,$47
	.dc.b	$9F,$9E,$3F,$86,$76,$3E,$C1,$F2,$5D,$F4,$FF,$25,$C6,$32,$FF,$29
	.dc.b	$36,$33,$3B,$7F,$D4,$71,$96,$C6,$5F,$F1,$FC,$65,$B1,$FB,$1F,$8F
	.dc.b	$7B,$9F,$C4,$5B,$7F,$C6,$70,$F3,$56,$6B,$F9,$25,$FF,$02,$8B,$D7
	.dc.b	$E5,$CF,$D5,$7E,$DF,$5B,$35,$F0,$3E,$8D,$F1,$BE,$5C,$56,$9C,$E7
	.dc.b	$FE,$71,$6C,$FF,$9B,$9A,$FF,$BC,$F4,$7C,$3F,$A4,$F9,$BB,$7F,$B3
	.dc.b	$AF,$F9,$24,$E9,$FE,$69,$DC,$FB,$07,$D3,$F8,$E3,$0D,$8C,$CF,$F2
	.dc.b	$26,$E3,$FE,$F0,$AD,$CE,$6B,$81,$1F,$F6,$D7,$23,$07,$CE,$7C,$F3
	.dc.b	$F9,$56,$3D,$38,$7F,$AE,$CE,$AF,$D3,$EB,$FD,$53,$FA,$37,$D2,$BD
	.dc.b	$BC,$4E,$89,$AF,$F0,$49,$D4,$63,$0E,$BF,$80,$7D,$F3,$9F,$85,$67
	.dc.b	$E3,$C5,$BE,$53,$6B,$F2,$DB,$FF,$3A,$F5,$FF,$A1,$78,$DF,$E9,$5C
	.dc.b	$7F,$B0,$60,$FD,$2B,$FD,$DB,$1F,$42,$75,$3F,$D1,$5B,$87,$46,$CF
	.dc.b	$F8,$A6,$79,$74,$F6,$C3,$F2,$DE,$2C,$AE,$91,$6B,$F9,$1F,$07,$F5
	.dc.b	$69,$FC,$0F,$5F,$C8,$F1,$7F,$3C,$C8,$C9,$FD,$A7,$81,$39,$BF,$E5
	.dc.b	$BE,$33,$E6,$7F,$BB,$36,$F8,$7A,$58,$99,$1F,$DC,$E7,$55,$F4,$EB
	.dc.b	$2F,$F8,$2A,$7F,$A0,$7A,$3F,$A3,$69,$F0,$63,$FC,$37,$C7,$FE,$16
	.dc.b	$7E,$24,$69,$D3,$5C,$C6,$DE,$33,$D6,$31,$1F,$D4,$BE,$8F,$D6,$7F
	.dc.b	$87,$EB,$A7,$17,$F8,$BB,$14,$FC,$0F,$B2,$95,$F1,$FC,$8B,$87,$C6
	.dc.b	$9B,$87,$F8,$DB,$6F,$E6,$B7,$F0,$E7,$E4,$FD,$24,$8F,$D9,$F8,$6F
	.dc.b	$95,$C6,$3F,$42,$AE,$3F,$90,$AD,$5F,$CD,$9F,$47,$D3,$FE,$7A,$FE
	.dc.b	$1B,$5F,$8E,$D7,$FC,$8A,$D7,$F9,$F3,$64,$D2,$76,$7F,$17,$F9,$5F
	.dc.b	$E8,$9E,$47,$F9,$15,$B5,$FE,$3F,$38,$CE,$5F,$E5,$77,$F8,$AF,$92
	.dc.b	$FC,$49,$FE,$29,$FA,$7F,$D2,$C6,$67,$F1,$13,$A7,$F8,$BC,$EB,$4F
	.dc.b	$A9,$3F,$92,$F9,$69,$5F,$99,$F5,$AE,$1F,$98,$9B,$7F,$4E,$6D,$7F
	.dc.b	$A5,$7C,$67,$13,$E7,$BF,$C6,$8D,$9F,$E3,$2D,$C7,$FB,$AB,$5F,$DB
	.dc.b	$79,$8F,$C9,$BF,$C2,$99,$7F,$42,$6D,$F7,$BE,$53,$9E,$9F,$D5,$FC
	.dc.b	$4C,$AF,$EA,$17,$C3,$1F,$F1,$3C,$E5,$FD,$9B,$FD,$AA,$35,$9B,$A9
	.dc.b	$3F,$C9,$BD,$0F,$59,$30,$F5,$BF,$E5,$19,$7D,$77,$EF,$0E,$13,$FB
	.dc.b	$97,$F9,$7F,$A7,$F1,$09,$1F,$11,$E5,$EE,$FF,$04,$37,$1F,$C0,$5C
	.dc.b	$34,$FE,$1D,$18,$CF,$F8,$6C,$DD,$7F,$51,$36,$BE,$45,$7D,$42,$3F
	.dc.b	$51,$F5,$7C,$9F,$F1,$43,$FA,$2B,$63,$FD,$09,$F0,$1F,$C0,$35,$34
	.dc.b	$9E,$0F,$F9,$CD,$B3,$FD,$61,$B5,$FE,$34,$D3,$FC,$E7,$97,$F0,$5F
	.dc.b	$2B,$FD,$31,$5F,$EA,$2B,$FC,$65,$69,$FF,$71,$79,$8F,$95,$FE,$B9
	.dc.b	$8F,$F6,$96,$BF,$CF,$E3,$5E,$8F,$F7,$39,$CC,$F2,$6A,$9D,$D7,$FA
	.dc.b	$5A,$73,$18,$9F,$47,$CE,$FF,$38,$3F,$D3,$13,$E2,$CF,$F8,$F4,$66
	.dc.b	$FF,$9C,$3A,$C3,$17,$FE,$9B,$BF,$F0,$53,$E9,$F5,$7C,$C4,$BF,$13
	.dc.b	$63,$51,$83,$FC,$07,$93,$E7,$BF,$A5,$9E,$4A,$B3,$FD,$AB,$E8,$9E
	.dc.b	$8F,$FB,$CF,$07,$B8,$F5,$7F,$B3,$56,$B9,$55,$A9,$BD,$AF,$A3,$7C
	.dc.b	$FF,$F9,$3F,$1E,$57,$C3,$1E,$4F,$13,$19,$78,$67,$70,$9D,$C7,$2C
	.dc.b	$3B,$5F,$5C,$7C,$B8,$71,$6E,$5F,$D7,$71,$EA,$D7,$A4,$FB,$DF,$EC
	.dc.b	$58,$F9,$1B,$EE,$7F,$B4,$9E,$51,$FE,$6A,$F4,$F9,$36,$33,$3B,$78
	.dc.b	$CF,$C6,$CF,$D7,$5B,$6D,$75,$7D,$9E,$A3,$E8,$FF,$C2,$91,$AB,$E5
	.dc.b	$9F,$D2,$BE,$0D,$8C,$E3,$FE,$5E,$C3,$93,$8E,$BF,$E3,$1F,$43,$FE
	.dc.b	$1F,$E8,$FE,$E0,$DE,$2F,$D5,$B6,$BF,$41,$F4,$6F,$23,$1D,$68,$C4
	.dc.b	$FF,$BE,$1A,$FF,$95,$DA,$FE,$BD,$AF,$ED,$9A,$F6,$9F,$A9,$3D,$5B
	.dc.b	$3A,$C7,$F9,$87,$D7,$79,$5B,$7F,$E2,$18,$F2,$CB,$CA,$D7,$52,$69
	.dc.b	$2B,$FC,$EB,$5B,$6D,$78,$08,$FB,$5A,$FB,$5A,$FB,$5F,$C6,$5F,$F4
	.dc.b	$5E,$6F,$09,$27,$F9,$2F,$C1,$7A,$7B,$CF,$06,$C7,$F2,$38,$C1,$1C
	.dc.b	$10,$FF,$6C,$C7,$07,$F9,$47,$1F,$C0,$23,$FC,$D3,$97,$C0,$77,$23
	.dc.b	$FC,$C5,$F4,$E9,$FC,$0A,$3F,$C9,$9C,$FF,$28,$F4,$76,$D8,$C9,$07
	.dc.b	$1F,$FB,$37,$4F,$E8,$33,$B9,$FF,$2E,$D5,$AD,$65,$FC,$DE,$7F,$A4
	.dc.b	$D7,$C3,$23,$FD,$15,$F2,$FC,$D7,$DC,$24,$7F,$50,$FA,$9F,$98,$7C
	.dc.b	$B2,$3F,$D3,$5F,$6E,$5F,$84,$24,$7F,$B4,$7E,$C1,$F3,$58,$8F,$F3
	.dc.b	$9C,$DC,$BF,$D9,$F5,$A9,$E1,$5F,$A7,$EE,$4E,$67,$E1,$F4,$7F,$24
	.dc.b	$8C,$5F,$C0,$BF,$C1,$1F,$81,$25,$7F,$63,$7E,$9D,$7E,$68,$91,$FD
	.dc.b	$BB,$E9,$6B,$FE,$88,$8B,$5F,$32,$7F,$4D,$E7,$3E,$C2,$92,$FF,$DA
	.dc.b	$BF,$4C,$F9,$AC,$47,$FA,$EE,$6E,$1F,$EB,$2A,$D4,$FF,$40,$86,$3F
	.dc.b	$D0,$BD,$6B,$E0,$7F,$09,$6D,$6E,$74,$7F,$25,$AE,$31,$FD,$2B,$E9
	.dc.b	$9F,$33,$FC,$46,$11,$87,$F9,$C3,$5F,$E7,$8D,$7F,$D3,$4D,$7F,$D7
	.dc.b	$8D,$7B,$47,$F4,$17,$AB,$63,$5C,$FF,$E5,$6F,$89,$F9,$24,$AF,$F7
	.dc.b	$DF,$D3,$7F,$07,$7F,$DC,$11,$96,$9A,$EE,$FA,$3F,$5D,$23,$FD,$80
	.dc.b	$FF,$01,$FA,$6C,$47,$FB,$A6,$2E,$7F,$A4,$9F,$B3,$F5,$52,$BF,$D9
	.dc.b	$0F,$C5,$F7,$9E,$A5,$3D,$8E,$27,$FA,$F3,$5F,$EC,$4D,$7F,$D8,$0D
	.dc.b	$7F,$DC,$4D,$7F,$BE,$B5,$39,$D3,$FC,$F7,$C9,$F8,$E2,$4F,$F9,$0B
	.dc.b	$E0,$6B,$FA,$21,$3C,$1F,$F7,$14,$FF,$23,$9F,$CB,$FD,$34,$9F,$F7
	.dc.b	$33,$FC,$9F,$E0,$F1,$1C,$EF,$FD,$17,$38,$D3,$F8,$5D,$7E,$83,$E4
	.dc.b	$E5,$FF,$BB,$DF,$EB,$3F,$87,$BE,$7A,$92,$7F,$2C,$7F,$8F,$E3,$4E
	.dc.b	$07,$CB,$FF,$BE,$1D,$F9,$C9,$DC,$7B,$E3,$15,$D6,$3E,$11,$CB,$E3
	.dc.b	$7F,$6B,$F4,$D2,$3C,$3A,$FF,$A9,$A3,$FA,$37,$4E,$67,$E0,$77,$A7
	.dc.b	$1B,$7F,$2A,$AF,$DF,$FC,$C4,$BE,$76,$8F,$EA,$1F,$3B,$99,$E7,$16
	.dc.b	$FF,$99,$CF,$F0,$8F,$B4,$95,$C9,$FF,$00,$FD,$9F,$FA,$86,$3E,$0B
	.dc.b	$5F,$F2,$19,$FE,$72,$70,$FF,$9B,$DA,$E7,$C8,$FE,$FF,$18,$8F,$F3
	.dc.b	$73,$77,$34,$D6,$BF,$EC,$AF,$93,$FE,$3A,$4A,$FE,$75,$E1,$FF,$33
	.dc.b	$FA,$9F,$8D,$1F,$E4,$1F,$29,$18,$8F,$38,$8C,$F0,$FF,$91,$5B,$1F
	.dc.b	$E4,$7F,$A9,$FE,$C8,$FD,$A6,$EF,$FC,$6A,$11,$8F,$17,$FC,$C7,$F0
	.dc.b	$0F,$95,$FE,$67,$F5,$12,$FE,$AB,$FD,$26,$5F,$2B,$4F,$13,$89,$F5
	.dc.b	$35,$F4,$E5,$FE,$7D,$9D,$92,$36,$D7,$DF,$BB,$CA,$7C,$27,$F8,$AD
	.dc.b	$B3,$F9,$26,$BF,$98,$FA,$97,$C9,$7F,$35,$6C,$7F,$9A,$DA,$FF,$9B
	.dc.b	$DA,$FC,$E6,$BF,$86,$6B,$FE,$8A,$6A,$69,$AE,$5F,$EB,$5F,$17,$FE
	.dc.b	$52,$F4,$3F,$DD,$75,$FB,$26,$BE,$7F,$96,$F3,$1F,$CF,$EC,$D2,$47
	.dc.b	$A5,$AF,$A5,$AF,$E7,$33,$FE,$33,$ED,$FF,$94,$53,$4F,$F2,$A6,$3F
	.dc.b	$E8,$BE,$4F,$F1,$13,$FE,$8D,$AE,$4D,$7F,$D5,$35,$FE,$CD,$F8,$6F
	.dc.b	$F2,$4C,$E6,$7F,$93,$B6,$75,$1F,$E9,$B6,$BF,$72,$35,$1F,$91,$8F
	.dc.b	$E5,$ED,$3F,$98,$FB,$7E,$68,$87,$39,$FE,$C9,$3F,$ED,$F8,$F1,$90
	.dc.b	$D1,$4F,$01,$FE,$60,$E5,$FF,$40,$E1,$FE,$41,$69,$EE,$2C,$7F,$42
	.dc.b	$FA,$2F,$E2,$8F,$F2,$06,$BF,$96,$6B,$F4,$5A,$FC,$86,$BC,$FC,$0F
	.dc.b	$F3,$8C,$62,$CD,$37,$83,$F3,$37,$3F,$FA,$71,$B8,$7F,$3C,$6B,$FD
	.dc.b	$67,$27,$E3,$5F,$80,$8F,$E8,$F1,$8D,$BF,$D9,$8D,$8F,$F6,$83,$5F
	.dc.b	$E1,$ED,$7F,$19,$DB,$FE,$2C,$7F,$AE,$36,$3F,$CD,$4D,$D7,$FD,$DB
	.dc.b	$F5,$3F,$D4,$6B,$7E,$77,$FA,$8E,$B5,$1F,$DB,$9B,$D4,$AE,$18,$7F
	.dc.b	$9D,$69,$E2,$7F,$AA,$DA,$7F,$52,$29,$FD,$BC,$FE,$BB,$1C,$63,$81
	.dc.b	$23,$87,$91,$FE,$39,$FA,$8F,$ED,$D4,$97,$C0,$3E,$73,$FA,$3D,$3E
	.dc.b	$A7,$FA,$4C,$E6,$B8,$4E,$5E,$51,$95,$1D,$9E,$B5,$D3,$0F,$F4,$9C
	.dc.b	$60,$E7,$7D,$A7,$C2,$D8,$EF,$3C,$2F,$B5,$F5,$F9,$EB,$7F,$D7,$9B
	.dc.b	$7E,$D7,$77,$E0,$6E,$3E,$8C,$7B,$D5,$3C,$65,$3D,$75,$E1,$8F,$54
	.dc.b	$E7,$0F,$85,$5E,$B3,$E0,$E1,$E9,$78,$CE,$EF,$E0,$C7,$F5,$EA,$F8
	.dc.b	$CD,$8C,$BD,$8F,$EC,$67,$CC,$C9,$F5,$38,$8C,$36,$BF,$12,$D9,$DC
	.dc.b	$FF,$9B,$7E,$3B,$FC,$DC,$9A,$47,$43,$3F,$B8,$BA,$FF,$D0,$B1,$88
	.dc.b	$E4,$6F,$4F,$EF,$33,$A7,$D3,$5F,$4D,$CF,$15,$AF,$EF,$F3,$5F,$CA
	.dc.b	$57,$17,$27,$53,$FC,$36,$E5,$FF,$0E,$C7,$17,$CD,$35,$F2,$DF,$AA
	.dc.b	$B8,$47,$E8,$6B,$84,$62,$37,$7E,$67,$C0,$76,$29,$D8,$F2,$74,$B8
	.dc.b	$BF,$59,$8F,$F4,$36,$CF,$F7,$E6,$BF,$C3,$8F,$F0,$B1,$FD,$41,$CB
	.dc.b	$4C,$9F,$C2,$4B,$3D,$DC,$F1,$3E,$A1,$F2,$D8,$6E,$1F,$E2,$F6,$D7
	.dc.b	$F8,$59,$AF,$F9,$1A,$B9,$57,$12,$57,$17,$BF,$B0,$F7,$29,$27,$FC
	.dc.b	$79,$19,$7F,$97,$FA,$AF,$89,$F5,$F2,$FF,$9E,$1B,$19,$9D,$C7,$6F
	.dc.b	$66,$47,$46,$E7,$FE,$2E,$C1,$18,$8F,$54,$62,$78,$1F,$DD,$47,$CA
	.dc.b	$FC,$A6,$99,$70,$7C,$CF,$DA,$FF,$C1,$71,$FE,$57,$6D,$7F,$8F,$9A
	.dc.b	$FC,$96,$BF,$CB,$75,$46,$23,$FC,$B8,$D8,$CF,$6F,$73,$EE,$F0,$3D
	.dc.b	$4A,$4C,$3F,$C1,$D1,$97,$FC,$D5,$D6,$F8,$1F,$C1,$B6,$33,$18,$23
	.dc.b	$82,$65,$FE,$58,$9E,$0F,$F9,$52,$93,$E4,$21,$E2,$3F,$2B,$0F,$AA
	.dc.b	$FE,$F3,$1F,$C5,$5B,$1F,$C6,$9A,$FF,$9F,$9A,$FF,$A5,$9A,$FF,$55
	.dc.b	$6B,$D7,$C3,$EF,$1A,$E2,$7D,$4B,$DD,$FC,$6B,$70,$FE,$FE,$DC,$3F
	.dc.b	$BF,$B5,$F1,$4F,$42,$4F,$A5,$B1,$E9,$6B,$C2,$3F,$9A,$F5,$3E,$A7
	.dc.b	$F3,$53,$FC,$89,$AF,$F1,$86,$BF,$C7,$9A,$FF,$73,$EA,$79,$5F,$31
	.dc.b	$B7,$FC,$FD,$B3,$FD,$11,$AF,$FA,$A1,$AF,$FB,$01,$AF,$59,$E8,$7F
	.dc.b	$08,$FA,$8F,$C9,$13,$8B,$06,$62,$5F,$FA,$DE,$79,$F1,$9E,$B7,$FC
	.dc.b	$86,$FD,$04,$AF,$43,$DF,$E3,$64,$73,$6C,$FF,$5E,$F2,$DF,$0D,$F8
	.dc.b	$C6,$D7,$F5,$F6,$CE,$E7,$44,$9D,$36,$3F,$D9,$7E,$63,$EA,$3F,$3B
	.dc.b	$1F,$EE,$48,$C7,$3F,$F2,$E6,$DF,$F7,$06,$CF,$FA,$DB,$DF,$FF,$57
	.dc.b	$CF,$F9,$FD,$B1,$9F,$53,$F9,$AC,$62,$3F,$D2,$CD,$7F,$D1,$AD,$5D
	.dc.b	$08,$C3,$67,$9F,$D0,$4B,$F0,$36,$FC,$8D,$7B,$B5,$FF,$80,$3F,$4F
	.dc.b	$F5,$89,$C5,$FB,$86,$D7,$F8,$2D,$B3,$FE,$10,$7F,$13,$97,$F2,$8F
	.dc.b	$7F,$E3,$77,$9A,$4B,$F8,$9F,$07,$F0,$88,$E0,$98,$7C,$B8,$CB,$63
	.dc.b	$E6,$79,$1F,$28,$8E,$09,$1F,$4E,$32,$D8,$FF,$12,$1F,$E6,$A2,$F9
	.dc.b	$7F,$C5,$9C,$47,$F8,$B1,$AF,$F9,$8C,$FE,$F3,$1B,$24,$72,$38,$68
	.dc.b	$F9,$EF,$EE,$B0,$8C,$3A,$8C,$C6,$2B,$F1,$5E,$B7,$E1,$89,$5F,$8B
	.dc.b	$6C,$6A,$30,$47,$F3,$7F,$0E,$F8,$7F,$CC,$36,$7F,$30,$D9,$DC,$E8
	.dc.b	$93,$A6,$C7,$F9,$3B,$DC,$FE,$48,$FC,$36,$BF,$E8,$76,$A6,$93,$C0
	.dc.b	$FE,$86,$DA,$F9,$DF,$4B,$F9,$F4,$D2,$72,$7E,$C7,$E9,$7F,$1E,$4E
	.dc.b	$5C,$12,$BF,$69,$19,$6C,$7E,$C3,$E9,$64,$7E,$D5,$B9,$7E,$E5,$AF
	.dc.b	$ED,$1A,$FF,$77,$FA,$4F,$EA,$3E,$09,$A4,$F2,$3F,$7B,$F4,$B3,$0F
	.dc.b	$E0,$4D,$BF,$C5,$0B,$E5,$C5,$FE,$36,$C3,$F0,$D7,$FE,$7C,$9C,$B6
	.dc.b	$33,$5C,$A7,$2D,$8C,$BF,$E3,$76,$BE,$9A,$ED,$CB,$2F,$A3,$67,$F8
	.dc.b	$D7,$D2,$7F,$16,$24,$FF,$1B,$7C,$DC,$92,$3C,$8D,$9F,$E4,$5E,$3F
	.dc.b	$F1,$E2,$4F,$F2,$36,$C6,$A3,$04,$70,$43,$F9,$2C,$70,$7F,$CA,$5F
	.dc.b	$94,$2F,$FC,$BA,$DA,$F9,$0D,$9F,$8C,$D7,$F9,$79,$FC,$C8,$91,$FC
	.dc.b	$DD,$FE,$7D,$F6,$AF,$9B,$A4,$9F,$E8,$11,$96,$C6,$67,$FA,$5C,$E5
	.dc.b	$B1,$FE,$74,$F7,$7F,$7D,$1F,$C2,$FC,$1F,$F7,$6F,$5E,$49,$87,$F7
	.dc.b	$9F,$B4,$F4,$57,$F8,$05,$AF,$FA,$2D,$BE,$59,$A4,$D3,$FD,$0B,$19
	.dc.b	$6C,$72,$F0,$9F,$38,$F8,$8F,$EB,$CD,$C7,$FA,$F3,$71,$FE,$CE,$D7
	.dc.b	$FB,$33,$50,$AE,$B5,$D0,$95,$ED,$E1,$F1,$91,$C1,$2B,$FC,$31,$82
	.dc.b	$75,$18,$7C,$FF,$4D,$2B,$CE,$D9,$FF,$72,$FD,$2F,$F8,$A3,$8F,$FD
	.dc.b	$79,$F4,$BF,$CE,$E3,$13,$FE,$DF,$8F,$F3,$A7,$D3,$3E,$87,$FC,$BB
	.dc.b	$39,$3D,$B5,$C3,$E0,$57,$CF,$BF,$FD,$C5,$5B,$15,$B9,$EB,$7D,$2F
	.dc.b	$8D,$74,$BE,$31,$88,$E3,$1D,$1F,$E1,$37,$E8,$6C,$65,$F1,$63,$FF
	.dc.b	$00,$65,$E9,$7F,$B0,$E7,$8B,$5F,$84,$D7,$DA,$D7,$D2,$FF,$B2,$EF
	.dc.b	$C9,$3D,$38,$FE,$5D,$70,$6F,$6B,$E9,$BF,$55,$3A,$39,$CD,$AE,$51
	.dc.b	$F1,$DB,$7F,$12,$F9,$74,$FF,$BE,$B5,$FC,$6F,$23,$E8,$DF,$4B,$FC
	.dc.b	$B3,$2C,$C6,$1F,$F1,$B3,$5F,$F1,$FB,$5F,$F3,$17,$D8,$BC,$6F,$F3
	.dc.b	$53,$63,$FC,$D4,$D7,$FC,$D4,$D7,$8C,$70,$23,$C1,$F2,$3F,$A3,$9F
	.dc.b	$F2,$CB,$E4,$F5,$33,$87,$AD,$B3,$A9,$F3,$C6,$08,$F1,$3C,$BE,$7F
	.dc.b	$A0,$8E,$08,$7A,$1C,$B4,$F9,$1D,$CE,$FF,$F7,$94,$61,$AF,$F5,$88
	.dc.b	$D7,$C0,$FB,$1B,$1F,$C9,$3C,$FF,$5F,$84,$69,$23,$44,$3F,$93,$B9
	.dc.b	$7B,$5F,$E2,$1B,$C1,$DE,$31,$1F,$CA,$5A,$FF,$BA,$FC,$3F,$9E,$FF
	.dc.b	$6D,$6F,$11,$A6,$B1,$F8,$1E,$0F,$52,$57,$C1,$6B,$E4,$6B,$E1,$6B
	.dc.b	$FC,$82,$39,$BC,$48,$7F,$77,$FA,$4F,$96,$FF,$92,$ED,$18,$8F,$E5
	.dc.b	$31,$8D,$3F,$12,$D8,$FA,$9F,$4B,$F0,$C8,$E0,$FF,$04,$57,$F3,$C6
	.dc.b	$E1,$CA,$31,$1E,$38,$C7,$D2,$C8,$E2,$D7,$D0,$D7,$D1,$28,$C4,$7C
	.dc.b	$1C,$F0,$FE,$26,$3E,$C1,$1F,$2F,$D0,$F8,$C8,$E0,$FC,$87,$D2,$FD
	.dc.b	$F2,$38,$21,$F8,$67,$2D,$3F,$D6,$9F,$4A,$F9,$1F,$D6,$B6,$33,$18
	.dc.b	$23,$82,$1F,$EC,$48,$E0,$FF,$0D,$FA,$57,$D0,$FE,$39,$B1,$FB,$E6
	.dc.b	$BB,$8F,$C9,$F8,$CF,$A9,$A4,$9F,$E0,$B1,$96,$C6,$5F,$E3,$91,$96
	.dc.b	$C7,$F0,$8F,$A6,$FE,$0D,$1F,$ED,$76,$BF,$C9,$9A,$FF,$73,$6F,$59
	.dc.b	$A4,$EA,$E8,$7E,$BF,$59,$5F,$C0,$DA,$FE,$D9,$AF,$EC,$DA,$FF,$A9
	.dc.b	$FE,$2D,$F4,$BF,$B4,$6C,$7F,$24,$6B,$FC,$E5,$AF,$F9,$9B,$EC,$7F
	.dc.b	$D5,$4F,$F2,$A6,$BF,$CF,$DA,$EF,$D1,$FE,$03,$18,$8F,$F5,$C3,$5F
	.dc.b	$F0,$AB,$53,$9E,$9F,$E4,$5E,$EF,$F0,$F2,$4F,$F2,$36,$BF,$C5,$9A
	.dc.b	$FF,$13,$7F,$8B,$7B,$B2,$3F,$95,$35,$FE,$3A,$D7,$F8,$D3,$5F,$EB
	.dc.b	$BF,$4C,$F9,$7F,$E8,$AD,$8F,$EB,$DB,$A3,$11,$FE,$8C,$8C,$4F,$F2
	.dc.b	$2F,$91,$7C,$0D,$25,$7F,$54,$8C,$B6,$32,$FF,$6B,$8C,$B6,$3F,$99
	.dc.b	$72,$48,$FE,$8C,$D7,$F9,$CB,$5F,$E6,$CD,$7F,$83,$78,$5F,$CE,$C8
	.dc.b	$FF,$4F,$6B,$FD,$09,$AF,$F3,$F7,$FB,$3F,$D2,$BE,$42,$92,$7E,$2C
	.dc.b	$65,$FE,$E3,$F9,$2B,$E7,$7F,$A2,$B6,$33,$18,$23,$82,$1F,$E0,$D8
	.dc.b	$E0,$FF,$3B,$F4,$3F,$96,$11,$FE,$62,$FF,$52,$FA,$47,$AB,$49,$3F
	.dc.b	$E0,$48,$CB,$FE,$BF,$FA,$37,$D2,$7F,$40,$6C,$66,$30,$67,$E9,$24
	.dc.b	$71,$6B,$FE,$23,$AF,$F2,$0B,$DF,$D7,$F2,$11,$C1,$0E,$0E,$5F,$F1
	.dc.b	$67,$D1,$7F,$87,$DF,$F3,$E8,$70,$FF,$A0,$1A,$E7,$D2,$F8,$5E,$0C
	.dc.b	$8F,$80,$D7,$CC,$D7,$FD,$CB,$7F,$C7,$DB,$1F,$DF,$2B,$FD,$05,$1F
	.dc.b	$E2,$C6,$BE,$6F,$77,$47,$D1,$C6,$47,$CF,$6B,$F1,$5A,$F8,$1A,$FF
	.dc.b	$7B,$E7,$FF,$36,$37,$DA,$FF,$10,$46,$21,$D9,$FE,$9F,$9F,$EE,$FF
	.dc.b	$64,$F8,$7A,$4A,$FF,$7E,$46,$5B,$19,$7F,$DF,$91,$96,$C7,$81,$87
	.dc.b	$F8,$E2,$FF,$C3,$47,$F8,$7B,$A7,$FC,$15,$0F,$43,$F8,$0F,$A3,$FE
	.dc.b	$0A,$9C,$C6,$27,$FC,$D0,$D7,$EE,$9C,$63,$64,$9F,$F2,$07,$C8,$FE
	.dc.b	$8A,$7F,$DB,$4D,$7F,$C4,$4D,$7E,$2B,$69,$F4,$1F,$B5,$9E,$F1,$8E
	.dc.b	$BF,$F0,$FC,$E6,$7F,$C4,$CD,$7F,$C4,$AD,$7D,$CD,$9E,$51,$EC,$8D
	.dc.b	$7C,$BB,$E8,$BF,$47,$3E,$E8,$C7,$37,$F9,$D9,$B1,$FE,$8B,$6B,$E6
	.dc.b	$FD,$B1,$F2,$9F,$E3,$46,$C7,$CA,$6B,$B8,$FF,$1C,$FD,$1F,$F2,$A8
	.dc.b	$FF,$20,$35,$FF,$1F,$B5,$FB,$0D,$C8,$D2,$6D,$D0,$FF,$39,$4F,$F4
	.dc.b	$23,$F6,$BC,$7F,$EF,$2F,$13,$24,$8F,$F3,$F9,$B3,$FC,$B0,$53,$F3
	.dc.b	$65,$8E,$4F,$FA,$6B,$5F,$EA,$99,$FF,$BF,$BE,$89,$F1,$7F,$C1,$25
	.dc.b	$F3,$BF,$C1,$A7,$36,$3F,$03,$19,$AE,$18,$B0,$E9,$18,$6D,$71,$C3
	.dc.b	$AB,$83,$99,$5F,$2A,$C3,$C2,$FE,$42,$FB,$39,$BF,$46,$1E,$38,$EC
	.dc.b	$D7,$86,$1D,$6B,$8D,$F4,$C3,$C7,$5C,$CE,$26,$EA,$39,$3C,$E5,$8E
	.dc.b	$33,$CE,$BA,$E1,$F0,$6B,$FD,$B9,$5E,$C7,$F6,$AD,$74,$63,$C5,$7D
	.dc.b	$1F,$D7,$6C,$72,$6B,$B8,$FF,$53,$AF,$85,$A4,$7F,$D5,$6D,$C7,$E5
	.dc.b	$C3,$A9,$F1,$5F,$4E,$3E,$F5,$59,$CC,$7C,$13,$25,$73,$5D,$23,$C4
	.dc.b	$66,$BA,$38,$8F,$0C,$F0,$CB,$FE,$00,$5F,$7B,$DB,$6F,$C0,$71,$9F
	.dc.b	$E3,$B5,$C6,$3F,$8F,$CE,$EF,$C4,$DA,$FF,$5E,$78,$BF,$99,$C3,$FD
	.dc.b	$CE,$DB,$FF,$72,$35,$FD,$8B,$5E,$27,$C2,$32,$53,$31,$D1,$C1,$FC
	.dc.b	$E7,$E8,$DF,$47,$8E,$44,$62,$BF,$9C,$B6,$BF,$95,$B5,$FF,$67,$35
	.dc.b	$FF,$66,$B5,$ED,$5D,$63,$14,$F4,$9F,$47,$E1,$6E,$1F,$D2,$1B,$87
	.dc.b	$F3,$A6,$BF,$E0,$6E,$27,$9C,$EE,$F6,$C6,$1B,$19,$BF,$E6,$0D,$AE
	.dc.b	$9F,$47,$E1,$24,$F8,$5B,$1A,$8C,$11,$FE,$B7,$CF,$F9,$6F,$9A,$F9
	.dc.b	$AD,$9F,$EB,$F3,$FD,$E9,$4E,$9A,$7F,$66,$FA,$37,$77,$D3,$6C,$66
	.dc.b	$30,$47,$04,$AF,$EB,$F3,$C1,$F2,$DF,$F9,$6E,$7C,$6F,$F8,$33,$C0
	.dc.b	$F3,$B9,$9D,$C7,$E0,$FA,$FF,$A9,$D7,$F7,$A6,$CF,$F6,$C6,$BF,$E3
	.dc.b	$8E,$0F,$E6,$BD,$DC,$49,$5C,$5B,$3E,$48,$F0,$90,$F2,$E1,$FC,$01
	.dc.b	$F9,$30,$FA,$37,$CC,$7E,$6A,$30,$D8,$CC,$F0,$6C,$FD,$26,$BF,$E1
	.dc.b	$26,$BF,$E5,$06,$BF,$D2,$DB,$C2,$FA,$37,$CE,$FE,$39,$B7,$FD,$ED
	.dc.b	$B7,$FE,$01,$6B,$DE,$7C,$C4,$9F,$33,$63,$CC,$D7,$84,$7E,$2B,$B1
	.dc.b	$F4,$7F,$99,$9F,$F3,$3C,$63,$97,$F7,$AD,$AE,$15,$BA,$FF,$36,$F5
	.dc.b	$7F,$A2,$27,$E6,$B6,$3F,$2E,$D7,$FC,$6E,$D7,$8F,$A5,$F4,$63,$11
	.dc.b	$FD,$C5,$A9,$CF,$03,$E8,$FF,$29,$3F,$87,$6B,$FD,$E6,$35,$CB,$FC
	.dc.b	$4E,$73,$3F,$E4,$F7,$99,$2F,$47,$8E,$66,$1E,$C6,$D7,$F9,$0A,$9F
	.dc.b	$47,$D4,$95,$FE,$56,$8C,$BF,$90,$F2,$5F,$01,$F8,$F6,$C6,$63,$04
	.dc.b	$70,$4B,$FE,$03,$3C,$1F,$F7,$57,$D1,$FE,$A1,$FF,$0C,$79,$88,$C4
	.dc.b	$23,$1C,$FF,$CE,$1B,$7F,$BA,$FA,$37,$AB,$FC,$41,$B1,$98,$C1,$1C
	.dc.b	$12,$FF,$8C,$CF,$06,$FC,$EF,$32,$4F,$76,$C6,$A3,$04,$7F,$8F,$FD
	.dc.b	$1F,$E0,$89,$C1,$FB,$76,$CF,$EF,$1B,$3C,$27,$DD,$DE,$F9,$3A,$4B
	.dc.b	$FE,$57,$9F,$F2,$A8,$F8,$1E,$74,$9F,$88,$DC,$3D,$AD,$7F,$CB,$39
	.dc.b	$A3,$11,$FE,$49,$EE,$4F,$F9,$42,$BE,$77,$91,$F3,$48,$E0,$95,$F2
	.dc.b	$A3,$2F,$F4,$08,$79,$9F,$C5,$DA,$7F,$1C,$2C,$6E,$34,$48,$D3,$63
	.dc.b	$FD,$53,$0D,$9F,$07,$FD,$F2,$30,$D8,$CC,$78,$DB,$3F,$80,$F5,$BF
	.dc.b	$10,$49,$FC,$63,$63,$51,$82,$3F,$EB,$AD,$5F,$04,$F8,$2F,$CB,$36
	.dc.b	$BE,$BB,$6B,$83,$5F,$F7,$FD,$72,$AE,$24,$AE,$2F,$F3,$FF,$A3,$FE
	.dc.b	$57,$5F,$E7,$F0,$E1,$FF,$41,$35,$CF,$83,$C4,$95,$FD,$32,$74,$71
	.dc.b	$5F,$47,$FA,$3A,$ED,$18,$E6,$F9,$D3,$99,$FF,$1C,$35,$FD,$C7,$D1
	.dc.b	$FF,$24,$C8,$D2,$73,$7E,$ED,$B8,$7E,$71,$B5,$FB,$4F,$67,$F7,$84
	.dc.b	$9F,$E0,$0D,$8D,$46,$08,$FF,$B8,$3E,$8B,$FA,$DF,$82,$69,$3C,$9F
	.dc.b	$E1,$1F,$47,$2F,$F8,$63,$6F,$97,$1F,$F9,$16,$FF,$21,$5F,$EF,$C6
	.dc.b	$CE,$A3,$13,$B2,$38,$38,$73,$73,$23,$83,$F8,$E7,$D1,$FF,$1B,$23
	.dc.b	$83,$C1,$CB,$E3,$23,$83,$F9,$27,$8F,$FC,$8C,$8E,$09,$87,$F2,$08
	.dc.b	$CB,$63,$2F,$F0,$68,$CB,$63,$FC,$77,$A3,$E5,$D8,$7F,$9A,$5B,$3A
	.dc.b	$6D,$7E,$5C,$57,$1A,$FE,$71,$12,$BF,$9E,$34,$FD,$77,$CC,$BE,$67
	.dc.b	$FA,$13,$67,$F9,$1B,$67,$73,$A2,$4E,$9B,$1F,$E5,$8F,$96,$FE,$06
	.dc.b	$FF,$BF,$1A,$FF,$23,$6A,$EE,$46,$3C,$A7,$D1,$7B,$AB,$FC,$1A,$D7
	.dc.b	$FC,$08,$D7,$FC,$08,$D8,$E3,$97,$FA,$51,$B8,$FF,$84,$A7,$47,$1E
	.dc.b	$77,$CC,$BE,$8F,$FB,$23,$6B,$FA,$2B,$71,$FE,$2C,$D7,$F8,$7B,$5F
	.dc.b	$EE,$2D,$7F,$B8,$36,$BB,$D7,$96,$DC,$34,$98,$79,$DB,$87,$F8,$A2
	.dc.b	$77,$39,$24,$79,$3E,$53,$D2,$48,$F4,$BF,$E5,$DF,$A0,$75,$FF,$B8
	.dc.b	$EB,$FD,$D5,$F4,$1F,$C8,$67,$3C,$BF,$EE,$48,$FF,$15,$7D,$1B,$E8
	.dc.b	$FF,$DC,$BF,$02,$9E,$18,$F3,$8C,$62,$18,$F4,$7B,$99,$71,$1C,$39
	.dc.b	$D8,$F7,$D3,$C5,$7B,$C3,$B6,$3E,$0C,$FF,$CF,$79,$75,$AF,$D2,$35
	.dc.b	$D1,$FE,$A8,$7F,$D4,$7C,$5C,$3D,$3F,$F2,$64,$E5,$AF,$26,$BF,$15
	.dc.b	$DE,$9E,$E8,$F6,$C6,$30,$D5,$7A,$34,$7D,$1F,$EB,$EF,$D6,$70,$9B
	.dc.b	$8F,$92,$3E,$29,$93,$E3,$D7,$2E,$8F,$EA,$D9,$7C,$EE,$47,$D0,$BE
	.dc.b	$8F,$FD,$3B,$18,$6C,$67,$0F,$F0,$FB,$5F,$F3,$9F,$D1,$FF,$90,$1F
	.dc.b	$F3,$BB,$5C,$F8,$DF,$E7,$C8,$C4,$7F,$A0,$1A,$FF,$9D,$DA,$FF,$9D
	.dc.b	$DB,$A1,$CE,$1E,$8F,$57,$CA,$4A,$F2,$B5,$FF,$02,$61,$FE,$68,$72
	.dc.b	$43,$D3,$E8,$79,$48,$FA,$9A,$E8,$C1,$0F,$E3,$1E,$CF,$FB,$13,$BF
	.dc.b	$F8,$9B,$7B,$CD,$3E,$6F,$81,$FE,$32,$AF,$E4,$FE,$7E,$08,$E0,$87
	.dc.b	$F3,$07,$2F,$F2,$7E,$1E,$C4,$7F,$DD,$77,$F8,$77,$F9,$F7,$17,$F4
	.dc.b	$57,$FD,$B2,$D7,$FA,$21,$8F,$0B,$FD,$BA,$FF,$2D,$9F,$E3,$AF,$4F
	.dc.b	$07,$D8,$47,$04,$3F,$9D,$39,$69,$93,$F9,$E3,$97,$FD,$ED,$F4,$5F
	.dc.b	$19,$FE,$B6,$D7,$D2,$D7,$79,$FC,$DF,$A3,$91,$F0,$1A,$F0,$AF,$7D
	.dc.b	$7A,$A3,$07,$68,$CF,$D1,$C8,$F0,$B5,$F0,$B5,$F3,$B5,$FE,$A8,$FD
	.dc.b	$F2,$3F,$49,$F0,$FC,$F4,$8F,$6B,$5F,$EC,$1F,$47,$F8,$72,$3F,$69
	.dc.b	$AE,$8C,$10,$FF,$30,$7D,$13,$E8,$FF,$72,$D8,$FF,$4A,$B5,$DC,$68
	.dc.b	$8E,$9A,$FF,$81,$47,$D1,$BE,$7F,$F0,$AD,$8D,$1A,$8C,$9F,$8A,$F1
	.dc.b	$DF,$33,$49,$3C,$23,$2D,$8C,$BE,$18,$CB,$63,$F8,$AF,$D1,$FE,$EE
	.dc.b	$3F,$D0,$0D,$7F,$CF,$2D,$7F,$DB,$6D,$7F,$DB,$4D,$EB,$34,$9D,$7F
	.dc.b	$B9,$D6,$57,$EB,$1A,$FF,$85,$DA,$FF,$85,$DA,$FF,$A0,$FE,$35,$F4
	.dc.b	$7F,$AC,$6C,$7C,$86,$BF,$3D,$AF,$F8,$D7,$E8,$FF,$D5,$6F,$CE,$6B
	.dc.b	$F5,$5A,$EF,$DD,$FD,$CC,$62,$3F,$D2,$8D,$7F,$B6,$35,$39,$ED,$FF
	.dc.b	$15,$FB,$BF,$C6,$49,$3F,$C3,$1A,$FF,$8D,$DA,$FF,$8D,$DF,$E4,$1E
	.dc.b	$EC,$8F,$E2,$6D,$7F,$C7,$CD,$7F,$C7,$CD,$7F,$C0,$1A,$BE,$8D,$F2
	.dc.b	$FF,$D5,$1B,$3F,$EF,$06,$BF,$EC,$16,$BF,$E5,$DF,$91,$FF,$76,$3F
	.dc.b	$96,$6B,$F9,$D6,$BB,$F8,$2F,$E4,$31,$88,$FF,$5E,$35,$FF,$0A,$35
	.dc.b	$39,$CB,$FC,$A7,$C9,$FD,$04,$93,$FC,$D1,$AF,$F9,$7D,$AF,$F9,$7D
	.dc.b	$FE,$91,$E1,$48,$FE,$76,$D7,$FC,$CC,$D7,$FC,$CC,$D7,$E4,$FD,$1B
	.dc.b	$E4,$3F,$BD,$B6,$3F,$79,$A2,$31,$1F,$EE,$96,$BF,$E8,$5F,$A2,$7D
	.dc.b	$1F,$F4,$46,$C7,$F0,$06,$BF,$C1,$DA,$F1,$8E,$04,$78,$3F,$CD,$3C
	.dc.b	$B9,$1F,$D0,$1A,$FF,$11,$FA,$1F,$F2,$7B,$FC,$39,$AE,$7D,$5F,$EA
	.dc.b	$F1,$88,$FF,$79,$79,$0F,$A3,$7D,$1F,$F6,$06,$CF,$FA,$A1,$AF,$F6
	.dc.b	$66,$BC,$97,$D1,$71,$24,$71,$7C,$3E,$BC,$8F,$33,$5F,$53,$5F,$F5
	.dc.b	$86,$08,$C4,$7F,$97,$C7,$D1,$FF,$88,$9B,$E9,$FF,$74,$8C,$19,$AF
	.dc.b	$E4,$B3,$D3,$C2,$F8,$64,$7E,$43,$5F,$98,$D7,$FD,$85,$3F,$E8,$A7
	.dc.b	$F9,$DB,$5F,$E8,$4D,$77,$EF,$7D,$0C,$64,$7D,$56,$BF,$65,$AF,$C0
	.dc.b	$6B,$FE,$7D,$E6,$71,$FF,$A2,$E3,$FD,$A9,$18,$F6,$BF,$C5,$93,$99
	.dc.b	$F8,$D3,$97,$FD,$C3,$F4,$6F,$87,$FF,$0E,$B6,$33,$18,$23,$82,$1F
	.dc.b	$E2,$68,$E0,$F8,$AB,$FC,$47,$5E,$68,$F1,$90,$FF,$78,$F5,$FF,$85
	.dc.b	$5F,$F4,$C7,$12,$31,$1F,$EE,$E6,$C6,$7D,$0F,$F1,$9C,$60,$8E,$0F
	.dc.b	$F2,$67,$C8,$FE,$B2,$BE,$43,$5F,$F7,$13,$6C,$E6,$DF,$41,$FB,$CA
	.dc.b	$F2,$B5,$FF,$80,$2D,$8D,$F5,$FF,$91,$A3,$04,$70,$7F,$80,$3E,$5D
	.dc.b	$F4,$7F,$9B,$BF,$F0,$44,$63,$9B,$53,$99,$FC,$F0,$9D,$3F,$E0,$FC
	.dc.b	$1F,$46,$F9,$4F,$F2,$BB,$67,$4D,$7F,$CB,$AD,$7F,$80,$FD,$1F,$F3
	.dc.b	$7A,$34,$9B,$FF,$98,$9B,$3E,$16,$BF,$92,$3F,$A5,$9F,$97,$28,$DB
	.dc.b	$FD,$AD,$E2,$FF,$A0,$63,$31,$87,$FC,$54,$D7,$A4,$6C,$FF,$37,$ED
	.dc.b	$FD,$7B,$3F,$F7,$07,$D1,$3E,$2F,$F8,$24,$BE,$77,$F8,$34,$E7,$0F
	.dc.b	$F7,$34,$E5,$E3,$5C,$30,$F9,$D0,$AE,$0F,$E1,$23,$31,$8A,$E5,$3B
	.dc.b	$AF,$0D,$F8,$1E,$98,$76,$9F,$0E,$1C,$A7,$71,$93,$F2,$35,$C5,$DD
	.dc.b	$66,$FE,$04,$77,$3D,$73,$B7,$89,$BB,$EC,$F2,$38,$1A,$2C,$FA,$B0
	.dc.b	$F6,$4F,$09,$F8,$17,$FE,$58,$AF,$0C,$7E,$51,$AF,$B8,$FF,$31,$2F
	.dc.b	$15,$F4,$7C,$5B,$3A,$6B,$CD,$AF,$F9,$AF,$E1,$3F,$D5,$A9,$A4,$7F
	.dc.b	$D6,$29,$7B,$BF,$61,$75,$BE,$23,$53,$99,$E8,$70,$DB,$F5,$F5,$B8
	.dc.b	$C1,$FB,$2B,$E3,$3A,$BB,$39,$36,$78,$EB,$95,$F7,$BF,$BD,$3E,$4A
	.dc.b	$AF,$8B,$5F,$45,$F8,$5E,$11,$D0,$E0,$53,$D3,$7C,$2B,$C7,$1F,$D9
	.dc.b	$6E,$BC,$27,$99,$B7,$DB,$BB,$C6,$EB,$B7,$C1,$8C,$65,$F0,$E3,$12
	.dc.b	$F8,$DC,$65,$FD,$56,$38,$B6,$B3,$87,$F0,$19,$CB,$6B,$38,$76,$BF
	.dc.b	$B5,$87,$87,$A9,$E0,$69,$B3,$A3,$FA,$4F,$D1,$3E,$BF,$55,$27,$59
	.dc.b	$7F,$40,$8F,$1F,$42,$F8,$56,$B2,$FE,$87,$5E,$48,$FF,$49,$9F,$C5
	.dc.b	$0B,$7F,$DC,$1A,$FF,$6E,$6B,$FD,$DF,$77,$D4,$7B,$9B,$3D,$CE,$7A
	.dc.b	$7C,$F8,$CD,$7F,$3F,$9F,$F0,$07,$DB,$F4,$8F,$E5,$4D,$C4,$D2,$7A
	.dc.b	$FF,$99,$9F,$F0,$0C,$62,$BF,$96,$BC,$78,$5E,$0F,$D2,$6C,$FC,$39
	.dc.b	$F2,$BF,$E1,$18,$CD,$79,$4F,$EA,$9E,$8F,$F2,$79,$FE,$B6,$D8,$FE
	.dc.b	$B3,$9F,$FA,$03,$97,$C9,$1F,$7A,$74,$FF,$8E,$E7,$55,$CA,$B7,$3D
	.dc.b	$9B,$3A,$8F,$EC,$BE,$1B,$F5,$E7,$B5,$F9,$08,$C3,$F9,$2A,$E0,$7E
	.dc.b	$F6,$33,$87,$E1,$A3,$FD,$2B,$DD,$DB,$2F,$E9,$53,$EC,$8C,$11,$C1
	.dc.b	$31,$F7,$CF,$07,$CB,$8F,$E3,$9F,$91,$87,$9F,$97,$C9,$5B,$C7,$C9
	.dc.b	$7B,$7F,$65,$C1,$FC,$67,$1F,$EC,$AD,$C7,$FC,$82,$D7,$FC,$7E,$D7
	.dc.b	$FC,$90,$D7,$FB,$93,$4F,$EE,$91,$F3,$9A,$FF,$72,$D3,$FC,$68,$FE
	.dc.b	$95,$B3,$F8,$19,$DB,$63,$35,$FC,$27,$E2,$9E,$87,$56,$E1,$FD,$F9
	.dc.b	$B7,$FE,$5A,$6B,$FE,$58,$6B,$FE,$61,$6B,$FE,$0B,$6E,$06,$F5,$FF
	.dc.b	$B8,$AB,$F7,$8D,$AF,$44,$F8,$30,$FE,$3D,$39,$7F,$8F,$7D,$17,$F9
	.dc.b	$96,$FF,$C6,$AD,$BF,$CD,$B5,$FC,$C3,$5F,$E3,$2D,$7F,$C7,$0D,$93
	.dc.b	$4D,$74,$FF,$85,$FC,$3F,$E2,$BC,$3F,$2B,$0E,$7E,$56,$27,$F8,$CF
	.dc.b	$DC,$BD,$0A,$63,$E6,$2D,$FE,$6F,$24,$7E,$A1,$B1,$EC,$8C,$1F,$81
	.dc.b	$D7,$F2,$51,$FE,$8F,$6E,$AD,$18,$BE,$5B,$F9,$CD,$6A,$DF,$35,$FD
	.dc.b	$15,$FF,$2C,$E1,$FD,$C6,$30,$FF,$97,$23,$7E,$17,$EF,$9A,$70,$2C
	.dc.b	$6E,$78,$BF,$EC,$18,$CB,$63,$2F,$FA,$9F,$F3,$97,$BF,$FD,$39,$B5
	.dc.b	$FD,$10,$EB,$C5,$FD,$7E,$33,$5F,$9D,$C5,$F0,$1E,$E3,$FA,$BF,$23
	.dc.b	$E8,$F1,$8F,$E3,$A6,$DF,$E3,$A6,$E3,$D7,$C0,$F1,$1F,$D8,$3B,$5F
	.dc.b	$0D,$8C,$7F,$A9,$69,$FC,$BA,$F8,$47,$F5,$5A,$DB,$AB,$FE,$3D,$FA
	.dc.b	$3F,$D6,$93,$F6,$29,$FA,$73,$E3,$37,$D7,$31,$FE,$CB,$87,$08,$FE
	.dc.b	$4D,$5B,$9C,$E1,$F8,$2E,$CF,$F1,$6F,$C4,$7E,$00,$E5,$BF,$C7,$7E
	.dc.b	$77,$C8,$4B,$EB,$F6,$1F,$D2,$AB,$FD,$05,$BA,$B5,$1F,$C2,$E3,$15
	.dc.b	$C3,$C8,$CE,$FF,$B6,$F9,$CF,$C7,$9F,$4B,$DA,$B5,$FB,$E8,$B8,$7E
	.dc.b	$FA,$6D,$7F,$8E,$EB,$F2,$BE,$17,$E3,$FA,$BF,$6D,$18,$3F,$20,$E0
	.dc.b	$91,$B2,$47,$FA,$5B,$2F,$8D,$87,$A7,$8F,$FC,$3F,$3C,$E3,$0F,$F1
	.dc.b	$2F,$99,$FD,$39,$F8,$A6,$99,$24,$7F,$16,$72,$FC,$5F,$92,$FC,$D1
	.dc.b	$FD,$77,$E9,$FF,$77,$19,$9F,$C4,$CE,$9F,$BB,$3A,$CF,$F3,$D3,$FD
	.dc.b	$97,$E3,$9F,$39,$FD,$BB,$A1,$18,$C3,$F7,$31,$89,$D4,$E7,$1F,$EE
	.dc.b	$2D,$7F,$CE,$42,$78,$57,$29,$D4,$FF,$9D,$E7,$2F,$FB,$7B,$97,$FC
	.dc.b	$F3,$F6,$DF,$C3,$0F,$36,$17,$0F,$D2,$FC,$A7,$EE,$CF,$D5,$7D,$5D
	.dc.b	$8C,$93,$0F,$EA,$D8,$A7,$51,$FB,$DF,$33,$F8,$C9,$FE,$52,$1F,$5F
	.dc.b	$8B,$FE,$93,$16,$7F,$AF,$4D,$DB,$FC,$CD,$3F,$EF,$1A,$FE,$D1,$F0
	.dc.b	$3F,$DA,$E3,$17,$ED,$6B,$F1,$BC,$9F,$E5,$6F,$F3,$C3,$C5,$36,$BF
	.dc.b	$AC,$FB,$5F,$CB,$8C,$4F,$F1,$57,$11,$A3,$FC,$BF,$5D,$1C,$4F,$F7
	.dc.b	$5F,$AD,$F2,$12,$3F,$BF,$3F,$DD,$7E,$2F,$F8,$E9,$F0,$DB,$1F,$E0
	.dc.b	$1F,$B2,$F6,$11,$FF,$03,$3F,$DF,$7E,$23,$F9,$29,$F2,$DB,$1F,$E6
	.dc.b	$DF,$25,$38,$C3,$FC,$11,$57,$5F,$F0,$44,$DB,$FF,$5E,$72,$7F,$BC
	.dc.b	$EB,$17,$F8,$DF,$CC,$BF,$C2,$6F,$E4,$38,$91,$88,$FC,$73,$5E,$1E
	.dc.b	$9F,$F4,$FF,$05,$F1,$1F,$B8,$6E,$5F,$C9,$2F,$FB,$95,$FF,$B2,$67
	.dc.b	$2F,$F8,$EB,$F6,$B7,$CD,$7E,$A5,$B7,$FA,$86,$CF,$F9,$59,$AF,$9D
	.dc.b	$AF,$E8,$DA,$FE,$A5,$B9,$1C,$E8,$F0,$3F,$DD,$D7,$FD,$23,$E1,$3F
	.dc.b	$CC,$15,$E0,$E0,$FE,$B9,$18,$EA,$F3,$12,$BC,$CD,$CB,$CC,$D7,$CC
	.dc.b	$FF,$91,$70,$FF,$13,$4F,$F6,$DF,$87,$E8,$FF,$58,$7F,$B7,$11,$FE
	.dc.b	$9B,$8A,$B1,$7F,$E6,$2F,$C8,$3F,$46,$FF,$03,$6B,$FC,$25,$BA,$1A
	.dc.b	$4E,$9F,$E3,$0D,$C3,$FC,$99,$EE,$FF,$76,$3F,$D0,$51,$FB,$2D,$7F
	.dc.b	$CA,$91,$F7,$F9,$7F,$DF,$DF,$02,$F1,$3F,$C1,$1E,$9E,$16,$9F,$81
	.dc.b	$AC,$3F,$D6,$A6,$E4,$FD,$55,$E7,$6B,$93,$FC,$A9,$F7,$5A,$9C,$D7
	.dc.b	$4C,$FC,$18,$FF,$96,$EB,$AA,$FA,$4F,$F3,$4F,$C1,$FF,$8A,$A3,$1F
	.dc.b	$00,$8C,$4E,$A7,$FC,$5E,$D7,$F9,$99,$CF,$B3,$FD,$FC,$E9,$AE,$A3
	.dc.b	$F4,$7E,$1F,$49,$CC,$FC,$2C,$9F,$81,$3E,$EB,$F9,$FE,$9E,$AD,$BF
	.dc.b	$CE,$5F,$0B,$8E,$3F,$F3,$C3,$71,$63,$CB,$0C,$3D,$39,$35,$87,$E7
	.dc.b	$78,$AF,$67,$CC,$C3,$E4,$65,$E3,$D3,$FE,$00,$46,$0F,$F4,$27,$D3
	.dc.b	$7F,$B2,$F2,$FF,$44,$36,$32,$7F,$88,$3A,$3F,$CD,$4D,$F8,$4F,$9D
	.dc.b	$D7,$F8,$F6,$CE,$A7,$FD,$9E,$D7,$F8,$83,$5F,$CD,$47,$02,$3F,$E0
	.dc.b	$7C,$7C,$5A,$7E,$E5,$B9,$FE,$0A,$7F,$86,$79,$B8,$99,$1F,$E4,$47
	.dc.b	$D5,$8F,$F0,$39,$BD,$C7,$D2,$DA,$FE,$25,$18,$23,$82,$1F,$C5,$5F
	.dc.b	$C7,$61,$FC,$35,$F2,$92,$3F,$77,$7E,$03,$B9,$1F,$E4,$0D,$7F,$90
	.dc.b	$35,$FE,$40,$FE,$6F,$4F,$C5,$F5,$FD,$56,$CF,$F9,$D7,$6F,$E8,$EF
	.dc.b	$EC,$BB,$3F,$12,$FE,$87,$67,$B5,$F7,$7A,$7F,$83,$3F,$A7,$6C,$FF
	.dc.b	$A4,$9A,$FE,$91,$AF,$FB,$3F,$E5,$3E,$AF,$90,$69,$F6,$3C,$CF,$F3
	.dc.b	$BC,$FF,$5B,$8C,$B4,$FF,$73,$7B,$B8,$7F,$D9,$F1,$72,$FF,$03,$CF
	.dc.b	$33,$E7,$D7,$4A,$FF,$03,$46,$0F,$A7,$6F,$B8,$F2,$92,$7F,$DE,$9E
	.dc.b	$87,$94,$E6,$47,$FD,$F4,$F2,$FB,$7F,$69,$23,$FC,$0F,$F0,$8F,$57
	.dc.b	$11,$FE,$E6,$9B,$9F,$D0,$F4,$BF,$D8,$15,$FA,$B6,$D7,$28,$C5,$F3
	.dc.b	$8F,$B1,$C5,$FD,$EB,$57,$D8,$7C,$B3,$E0,$92,$7F,$D8,$FC,$E9,$DC
	.dc.b	$63,$67,$9D,$F5,$89,$1D,$85,$71,$F8,$CC,$C7,$FC,$00,$9B,$AF,$F8
	.dc.b	$E2,$7F,$67,$4F,$22,$92,$FF,$D6,$1E,$8D,$57,$F9,$2A,$EF,$59,$8C
	.dc.b	$76,$AF,$F1,$7F,$65,$30,$FF,$11,$3F,$BA,$F6,$3F,$D9,$51,$FE,$B2
	.dc.b	$6B,$FE,$6C,$F8,$4F,$E2,$71,$88,$FF,$0F,$35,$FF,$35,$35,$FF,$38
	.dc.b	$36,$FF,$CF,$73,$C8,$FD,$EE,$4F,$1F,$E0,$92,$FF,$D0,$FF,$0F,$5F
	.dc.b	$7F,$17,$F8,$CA,$B8,$47,$CF,$7E,$7D,$7F,$14,$F6,$D7,$FA,$AF,$BB
	.dc.b	$97,$FE,$41,$7F,$55,$F3,$98,$8E,$A7,$D8,$E3,$4E,$78,$3F,$CD,$B5
	.dc.b	$F5,$BC,$89,$7E,$1C,$D1,$FC,$F6,$1F,$35,$FA,$BB,$FE,$0F,$18,$EE
	.dc.b	$FE,$D4,$D3,$FB,$64,$FE,$C5,$FE,$CE,$D7,$FD,$36,$D7,$FC,$06,$D7
	.dc.b	$91,$FD,$B8,$A7,$FA,$8A,$D9,$3E,$73,$F8,$69,$27,$E5,$74,$3E,$1B
	.dc.b	$11,$F1,$2F,$8D,$63,$2F,$F4,$7D,$7E,$63,$C1,$97,$F0,$DF,$E8,$BE
	.dc.b	$2F,$C1,$E3,$FB,$11,$A9,$E9,$8F,$F1,$D7,$FE,$00,$D6,$DF,$A5,$E7
	.dc.b	$7F,$22,$23,$FA,$DE,$67,$C3,$FF,$4D,$8F,$B3,$18,$F2,$BF,$AA,$CE
	.dc.b	$74,$FE,$6F,$39,$AF,$DB,$43,$DF,$F8,$91,$D6,$B8,$69,$FE,$C5,$6C
	.dc.b	$7F,$B2,$DB,$5C,$76,$FD,$CF,$8D,$2B,$F0,$8F,$F6,$EC,$5F,$2B,$99
	.dc.b	$F8,$33,$6F,$F6,$F3,$FD,$03,$EE,$E5,$7F,$A5,$DF,$E8,$1F,$57,$88
	.dc.b	$FA,$9C,$89,$C5,$FF,$B0,$AB,$F8,$5F,$91,$2F,$F2,$CF,$E5,$79,$9D
	.dc.b	$34,$FC,$9B,$89,$FC,$9C,$60,$8E,$08,$7E,$3E,$38,$3F,$EE,$9C,$7F
	.dc.b	$DF,$6F,$FA,$BF,$E1,$3F,$AD,$9F,$CD,$3D,$EF,$F1,$16,$4F,$C0,$FF
	.dc.b	$96,$D7,$F9,$1A,$31,$E4,$7F,$93,$DB,$1F,$9D,$6B,$FE,$4C,$F2,$DE
	.dc.b	$57,$F7,$28,$FE,$97,$E7,$7F,$42,$3F,$A1,$FB,$BF,$D1,$FE,$79,$F0
	.dc.b	$B4,$95,$F8,$4C,$DF,$01,$E6,$6C,$78,$5A,$F9,$DA,$FF,$9D,$8F,$E6
	.dc.b	$AE,$1F,$F0,$43,$5C,$C7,$B6,$31,$9F,$C3,$8F,$ED,$FE,$E7,$F8,$08
	.dc.b	$95,$FB,$6F,$92,$FF,$01,$1F,$DE,$08,$FF,$58,$F1,$11,$88,$FF,$80
	.dc.b	$0D,$7F,$82,$46,$9F,$EB,$FE,$17,$FB,$49,$FE,$39,$5F,$D5,$0F,$CD
	.dc.b	$7A,$92,$3F,$B2,$F6,$7F,$A0,$5B,$1F,$E3,$88,$C4,$7F,$A0,$29,$18
	.dc.b	$8F,$E2,$8D,$7F,$DD,$AD,$7F,$B4,$B4,$7D,$DF,$E3,$B3,$FD,$DB,$7C
	.dc.b	$0F,$F5,$2C,$EB,$A7,$F1,$BF,$3F,$FD,$1B,$0F,$E7,$B8,$23,$11,$F5
	.dc.b	$E3,$1E,$6F,$E3,$E7,$33,$FE,$3F,$6B,$FD,$C1,$B3,$C2,$3F,$29,$18
	.dc.b	$37,$7E,$8F,$5B,$FC,$CF,$91,$A7,$E8,$BF,$07,$FF,$07,$5F,$F1,$46
	.dc.b	$BF,$A3,$6B,$E7,$6B,$FE,$53,$6B,$FE,$04,$6E,$06,$93,$7F,$F3,$6F
	.dc.b	$CC,$BB,$FF,$D0,$6D,$BF,$ED,$8D,$7F,$A1,$7C,$0F,$A5,$FE,$E0,$D7
	.dc.b	$EF,$35,$FE,$A0,$D7,$FD,$B3,$C0,$8C,$47,$F8,$99,$B1,$9C,$BF,$D0
	.dc.b	$31,$83,$3C,$1F,$D1,$EB,$FA,$A3,$73,$FF,$11,$34,$EF,$F3,$9F,$D6
	.dc.b	$E7,$FC,$0C,$D7,$F2,$8D,$7F,$B3,$35,$E3,$E6,$FF,$A4,$BE,$0F,$DD
	.dc.b	$7F,$E2,$F6,$DF,$B1,$B4,$EE,$7C,$FF,$E8,$AF,$86,$7F,$EB,$0B,$DE
	.dc.b	$87,$8D,$F0,$E6,$FF,$5A,$E3,$FC,$6C,$FE,$E4,$5C,$3F,$C8,$0D,$CB
	.dc.b	$E6,$AE,$DE,$3D,$D2,$3F,$C4,$C6,$CF,$F2,$A9,$B3,$FC,$B0,$3E,$7D
	.dc.b	$E1,$7F,$93,$A1,$F2,$FF,$D6,$91,$BA,$FE,$41,$19,$8C,$1A,$71,$5D
	.dc.b	$31,$C7,$93,$FC,$9E,$31,$5D,$A3,$17,$D1,$B1,$C0,$F0,$CF,$0B,$EB
	.dc.b	$87,$9F,$9B,$BE,$1F,$CC,$DB,$7E,$87,$78,$DC,$7C,$57,$EF,$C3,$77
	.dc.b	$E3,$3B,$E1,$E3,$7C,$33,$BA,$E3,$7E,$59,$F8,$27,$12,$DF,$81,$C6
	.dc.b	$3F,$01,$FE,$90,$7D,$1B,$FA,$0E,$23,$4F,$63,$FC,$7C,$BE,$2D,$F3
	.dc.b	$FF,$E4,$19,$FB,$51,$87,$6D,$7A,$35,$FF,$24,$73,$7F,$95,$1F,$DB
	.dc.b	$1F,$E5,$67,$04,$7F,$D4,$99,$7F,$B6,$76,$F9,$51,$EA,$9A,$FE,$1B
	.dc.b	$7F,$F7,$3E,$DF,$85,$8F,$C4,$CE,$94,$EA,$FC,$71,$D8,$FE,$D9,$19
	.dc.b	$7B,$19,$3B,$5F,$37,$D3,$7D,$E7,$84,$7F,$75,$BF,$04,$62,$3C,$67
	.dc.b	$33,$C7,$5C,$E7,$6D,$8D,$4F,$8F,$8B,$F7,$5E,$BF,$F9,$FA,$7F,$82
	.dc.b	$35,$FF,$61,$F1,$FF,$A5,$2F,$FD,$9A,$D8,$FE,$CC,$D7,$FC,$F0,$D7
	.dc.b	$9C,$70,$9E,$9C,$6F,$9F,$7C,$4E,$1B,$5F,$E8,$39,$D4,$E4,$91,$99
	.dc.b	$DD,$FF,$8A,$DA,$FF,$86,$EB,$8C,$F3,$FA,$A3,$E2,$6B,$93,$07,$C1
	.dc.b	$CB,$7F,$05,$63,$FD,$D2,$7F,$BB,$8B,$7F,$E0,$76,$BE,$1F,$D9,$7A
	.dc.b	$BF,$E3,$F6,$B9,$8F,$87,$18,$8F,$E2,$B8,$23,$11,$FC,$D5,$B1,$99
	.dc.b	$E2,$79,$FB,$7C,$35,$FD,$61,$AF,$FB,$F1,$B8,$1C,$D3,$EF,$5F,$03
	.dc.b	$F5,$9B,$87,$D3,$6E,$1F,$E5,$96,$BF,$3D,$AF,$A9,$AF,$F4,$D6,$D7
	.dc.b	$7A,$FC,$77,$4B,$E8,$5C,$98,$6E,$3F,$E6,$96,$E3,$FE,$6F,$6B,$FE
	.dc.b	$23,$5F,$1B,$C4,$93,$C5,$B3,$C5,$AF,$17,$C1,$D9,$27,$C0,$D9,$FF
	.dc.b	$3C,$3F,$E6,$E8,$FC,$B3,$63,$3E,$5F,$E7,$A3,$11,$F9,$36,$BF,$79
	.dc.b	$AF,$F7,$16,$E6,$73,$B0,$FC,$19,$6F,$F2,$CE,$0F,$C2,$FC,$FF,$FA
	.dc.b	$9A,$3F,$54,$DB,$FF,$48,$35,$FD,$33,$5F,$C8,$B5,$FF,$02,$36,$CD
	.dc.b	$35,$5F,$EF,$6F,$90,$FD,$47,$A8,$69,$30,$F8,$FE,$97,$C5,$23,$82
	.dc.b	$63,$F3,$A3,$2D,$8F,$F5,$57,$D0,$3E,$67,$F1,$2D,$8C,$C6,$0F,$E1
	.dc.b	$F9,$7F,$A3,$23,$E1,$FA,$38,$23,$F1,$BE,$47,$64,$70,$7C,$AC,$79
	.dc.b	$91,$C1,$FD,$67,$B5,$F1,$DF,$CA,$73,$FF,$13,$36,$33,$3B,$8C,$11
	.dc.b	$C1,$27,$FB,$6A,$F9,$AF,$E5,$91,$FC,$69,$B1,$9E,$FF,$E7,$0D,$3F
	.dc.b	$BA,$CF,$E2,$77,$7C,$4F,$D6,$24,$EB,$85,$3A,$8C,$36,$BF,$DC,$D5
	.dc.b	$C1,$B3,$FE,$AC,$F8,$6F,$E5,$1E,$91,$A4,$8D,$12,$7F,$D5,$D3,$B8
	.dc.b	$C4,$7F,$A7,$7E,$79,$F3,$7F,$C1,$9B,$3F,$EE,$B6,$CF,$FB,$C5,$AF
	.dc.b	$19,$E0,$49,$E0,$FF,$5D,$A7,$C1,$FE,$BE,$7E,$3C,$63,$A3,$53,$9B
	.dc.b	$FF,$0F,$CE,$6B,$7E,$8F,$F2,$E8,$34,$9E,$1F,$76,$CF,$08,$C1,$FA
	.dc.b	$8F,$9C,$91,$F4,$1B,$7F,$ED,$3D,$91,$88,$FF,$7C,$6F,$F6,$23,$11
	.dc.b	$FE,$FB,$FC,$AF,$F4,$8F,$E0,$9A,$FF,$8E,$1A,$EF,$A7,$F3,$11,$88
	.dc.b	$FF,$33,$B5,$34,$D6,$1F,$95,$9F,$C8,$F1,$E0,$8F,$F0,$48,$87,$E3
	.dc.b	$0B,$5E,$D6,$BE,$CC,$7F,$8F,$9F,$D8,$B4,$7C,$6F,$EE,$E7,$E9,$B5
	.dc.b	$FF,$3E,$9F,$EF,$87,$0F,$FB,$B9,$AE,$7A,$7F,$05,$18,$9E,$B1,$83
	.dc.b	$E1,$F7,$4C,$3F,$CA,$D9,$FF,$38,$C3,$E1,$CE,$5F,$9C,$D7,$E7,$3F
	.dc.b	$8F,$F8,$E7,$C1,$52,$5F,$E4,$23,$2F,$EC,$F1,$FF,$78,$37,$D5,$FD
	.dc.b	$69,$1C,$13,$0F,$D5,$C6,$5F,$F3,$AF,$07,$F6,$CF,$9D,$FF,$0E,$78
	.dc.b	$A6,$92,$FF,$81,$46,$5B,$19,$AE,$73,$97,$FD,$03,$E1,$BF,$66,$FE
	.dc.b	$8C,$4F,$7F,$FC,$7E,$FF,$46,$AF,$EB,$64,$DB,$C6,$7F,$36,$B9,$8F
	.dc.b	$F5,$86,$9F,$D6,$0A,$7F,$58,$29,$FE,$28,$A7,$83,$E0,$24,$FF,$92
	.dc.b	$E3,$2F,$BB,$E0,$BE,$19,$1C,$1B,$E5,$F4,$91,$C1,$F1,$BE,$7B,$E5
	.dc.b	$91,$C1,$30,$F8,$F1,$96,$C6,$5F,$DC,$E4,$9D,$46,$1F,$E9,$5F,$71
	.dc.b	$FD,$02,$0D,$27,$85,$FC,$6F,$E7,$DF,$29,$49,$8F,$F1,$C8,$CB,$FD
	.dc.b	$13,$EF,$DB,$FF,$7F,$6F,$07,$FA,$05,$B3,$B9,$D1,$27,$4D,$8F,$E5
	.dc.b	$3F,$36,$FA,$0F,$E5,$CD,$8C,$C6,$0E,$3C,$D2,$3B,$37,$5F,$F6,$6E
	.dc.b	$3E,$68,$C4,$7E,$FA,$31,$3C,$0F,$0F,$D0,$78,$08,$E0,$93,$FC,$EB
	.dc.b	$04,$EA,$30,$FF,$A8,$7C,$A5,$BE,$5F,$0D,$C3,$FB,$6C,$EA,$72,$48
	.dc.b	$CC,$EE,$FF,$DC,$F5,$B9,$CD,$7F,$53,$F9,$CF,$F2,$1F,$59,$A4,$D7
	.dc.b	$FB,$EF,$89,$FE,$0A,$C7,$FD,$36,$0F,$F4,$BF,$99,$8C,$7F,$91,$89
	.dc.b	$E0,$E1,$B3,$C6,$1F,$41,$FE,$6A,$71,$19,$8C,$3E,$03,$FD,$56,$E2
	.dc.b	$3B,$B6,$3F,$99,$FD,$07,$F9,$0F,$DD,$FF,$7F,$36,$3F,$C2,$AD,$7F
	.dc.b	$9C,$87,$0F,$2F,$8D,$6D,$C9,$B1,$C9,$AF,$29,$46,$23,$DD,$F3,$2F
	.dc.b	$99,$F8,$B1,$86,$BF,$EC,$68,$D3,$F0,$8C,$46,$B2,$4F,$FB,$6A,$75
	.dc.b	$18,$F8,$0D,$4E,$67,$E4,$4E,$5E,$9C,$1F,$7F,$D4,$FF,$1B,$C2,$30
	.dc.b	$FF,$85,$38,$53,$A9,$CC,$FE,$9B,$F3,$B6,$DC,$B9,$11,$89,$FE,$00
	.dc.b	$D7,$84,$EE,$31,$1F,$AD,$6B,$F8,$D9,$E7,$3B,$3C,$5C,$CF,$9A,$A4
	.dc.b	$AF,$E0,$6D,$7F,$60,$D7,$FB,$BE,$09,$E5,$1F,$C5,$7D,$B7,$99,$EB
	.dc.b	$6C,$66,$30,$47,$04,$3F,$84,$C7,$07,$F8,$07,$D0,$7A,$08,$FF,$18
	.dc.b	$7B,$7A,$52,$3C,$4D,$79,$B5,$FE,$25,$3F,$E4,$07,$67,$F9,$C3,$2F
	.dc.b	$DD,$7A,$9F,$71,$B1,$9D,$FF,$D7,$CD,$9F,$C0,$AF,$CA,$5F,$19,$86
	.dc.b	$C7,$E1,$9A,$FF,$9F,$9A,$FF,$4C,$F6,$7F,$17,$E7,$FF,$BB,$1B,$E4
	.dc.b	$FF,$9D,$67,$FB,$FC,$60,$8E,$08,$7F,$80,$1C,$B4,$C9,$FE,$30,$F3
	.dc.b	$BF,$C0,$5E,$47,$E6,$9B,$1F,$DC,$63,$1E,$69,$F1,$F8,$B2,$7F,$C2
	.dc.b	$2D,$7F,$C1,$B3,$F1,$2B,$FC,$1D,$1C,$88,$7F,$A7,$9F,$F0,$0F,$D0
	.dc.b	$F8,$7C,$F3,$49,$3B,$8C,$B6,$32,$FF,$87,$23,$2F,$F8,$83,$E5,$FE
	.dc.b	$49,$1F,$F1,$3B,$EF,$FD,$9B,$71,$F0,$DA,$FF,$35,$B4,$62,$11,$8B
	.dc.b	$FA,$BF,$45,$27,$FC,$66,$D7,$FC,$62,$D7,$FC,$12,$D7,$FA,$16,$DF
	.dc.b	$C2,$7C,$EF,$C0,$F5,$7F,$81,$27,$FC,$99,$18,$23,$82,$1F,$EB,$47
	.dc.b	$F8,$77,$A9,$FE,$C8,$8F,$F2,$B7,$BD,$82,$3F,$C6,$3E,$45,$F4,$3F
	.dc.b	$9A,$6C,$66,$30,$47,$04,$3F,$DA,$F1,$C1,$FF,$3E,$F1,$7F,$1F,$F6
	.dc.b	$BF,$11,$1F,$EB,$66,$BA,$31,$1B,$23,$83,$FC,$BB,$E9,$7E,$88,$8E
	.dc.b	$08,$7F,$97,$9C,$B4,$FE,$6B,$F3,$7F,$A9,$7F,$DD,$AD,$7F,$AC,$B5
	.dc.b	$FF,$49,$B5,$FE,$ED,$ED,$3E,$93,$F7,$ED,$8F,$E8,$2D,$7F,$DF,$2D
	.dc.b	$7F,$B2,$B5,$FF,$4F,$B5,$EB,$1D,$0F,$DF,$F8,$12,$3F,$CF,$AD,$7F
	.dc.b	$CF,$6D,$7F,$CA,$6D,$7F,$C1,$9B,$FF,$5E,$E9,$FE,$A1,$E4,$FA,$E3
	.dc.b	$13,$FE,$E8,$8C,$12,$76,$48,$FF,$76,$FC,$07,$F3,$E7,$FD,$D6,$D7
	.dc.b	$FA,$1B,$5F,$0B,$57,$C3,$BE,$85,$F4,$3F,$DA,$23,$0D,$8C,$CF,$F8
	.dc.b	$30,$E2,$E1,$FF,$07,$CF,$07,$4D,$73,$1B,$7F,$D9,$B1,$97,$FD,$E3
	.dc.b	$8B,$E4,$7F,$5E,$48,$FF,$5D,$35,$FF,$59,$C6,$DC,$90,$FF,$56,$FC
	.dc.b	$FF,$F3,$62,$3F,$EB,$46,$BA,$30,$43,$F2,$7C,$1F,$D2,$B8,$0D,$24
	.dc.b	$7F,$83,$BF,$31,$FE,$C8,$FF,$8C,$1A,$FF,$BA,$9A,$FF,$93,$5A,$9A
	.dc.b	$4F,$2B,$FA,$CB,$C3,$B7,$FD,$67,$F2,$FF,$CC,$C9,$5F,$CC,$5F,$E6
	.dc.b	$DE,$3C,$8F,$E6,$8D,$7F,$99,$78,$FF,$CE,$C8,$FF,$39,$7F,$01,$F3
	.dc.b	$FF,$E9,$A8,$FB,$2D,$7E,$EB,$5F,$B2,$D7,$F9,$A7,$D0,$3C,$7F,$EB
	.dc.b	$0D,$8F,$E9,$8D,$7F,$DB,$ED,$7F,$B0,$35,$E7,$1C,$88,$FF,$BB,$33
	.dc.b	$FE,$E2,$47,$07,$F5,$3F,$37,$FA,$C1,$1C,$10,$FE,$A8,$E5,$A7,$E7
	.dc.b	$7C,$2F,$F2,$47,$99,$FD,$00,$91,$9E,$2F,$F1,$DB,$FD,$43,$17,$C1
	.dc.b	$FC,$02,$47,$C0,$6B,$F0,$1A,$FC,$07,$F6,$1F,$3E,$FA,$0A,$49,$FD
	.dc.b	$5C,$65,$BE,$BF,$02,$38,$21,$FB,$48,$FF,$34,$3F,$B0,$6B,$FE,$9F
	.dc.b	$E8,$46,$23,$F8,$2F,$CF,$BC,$FF,$E0,$8D,$8F,$DE,$B5,$FE,$02,$D7
	.dc.b	$FC,$35,$E4,$FF,$B6,$5F,$F0,$63,$5F,$F4,$AF,$37,$FB,$BB,$E0,$FF
	.dc.b	$C3,$D3,$99,$E3,$18,$23,$C9,$FF,$19,$FC,$CF,$E1,$E1,$C2,$8C,$37
	.dc.b	$93,$67,$F8,$AF,$F0,$7F,$B3,$89,$A4,$E3,$D1,$30,$FF,$12,$46,$5F
	.dc.b	$C3,$7C,$FF,$F9,$5D,$FF,$57,$E0,$8C,$47,$FB,$2C,$F2,$B8,$7F,$C1
	.dc.b	$AD,$7F,$C3,$3F,$59,$FD,$77,$03,$49,$E2,$7F,$90,$FC,$BF,$49,$32
	.dc.b	$C1,$2F,$FC,$75,$19,$6C,$7F,$96,$9F,$F2,$E9,$FE,$1E,$29,$F2,$8B
	.dc.b	$8F,$FB,$8F,$AF,$81,$23,$F4,$C6,$8E,$23,$E7,$DD,$DF,$D3,$A3,$F1
	.dc.b	$5E,$1F,$C3,$9F,$F3,$BC,$60,$FF,$78,$38,$E4,$FE,$79,$1C,$EB,$94
	.dc.b	$62,$7A,$F3,$7F,$BD,$E3,$1D,$FF,$3D,$B7,$9C,$3C,$18,$72,$6D,$E6
	.dc.b	$7C,$37,$C6,$7C,$CD,$8F,$2D,$71,$6D,$71,$7E,$04,$ED,$A6,$70,$F3
	.dc.b	$BE,$F3,$D9,$39,$9E,$13,$F0,$6F,$CB,$5B,$8E,$2F,$FB,$F1,$AE,$8C
	.dc.b	$5F,$CF,$75,$F3,$CF,$9F,$FF,$71,$D2,$31,$3F,$DE,$E3,$13,$A9,$CC
	.dc.b	$FF,$BA,$B9,$7F,$D5,$6F,$EB,$4F,$F5,$6B,$82,$3E,$D3,$FD,$D3,$F0
	.dc.b	$1F,$15,$1C,$74,$FB,$AE,$8E,$06,$72,$FD,$43,$98,$DB,$93,$F5,$9B
	.dc.b	$AB,$B4,$64,$EF,$19,$7A,$DF,$8E,$31,$5D,$23,$15,$C9,$B5,$C0,$F6
	.dc.b	$57,$0B,$FF,$11,$4E,$AF,$FC,$25,$BB,$E1,$3F,$D6,$98,$FF,$4B,$6E
	.dc.b	$3F,$C1,$9A,$FF,$05,$9E,$1E,$7F,$9E,$31,$3A,$7F,$D2,$6D,$7F,$8F
	.dc.b	$B6,$78,$D7,$FA,$E7,$E7,$DD,$DC,$EF,$FD,$51,$18,$9D,$4E,$67,$FD
	.dc.b	$53,$39,$7F,$9A,$B5,$FF,$6A,$35,$E7,$1C,$79,$D3,$99,$FF,$59,$FC
	.dc.b	$FF,$D5,$BF,$F6,$63,$63,$FA,$0B,$5F,$F7,$03,$5F,$F5,$8C,$3E,$23
	.dc.b	$CD,$8F,$DF,$6C,$64,$FE,$F0,$59,$FE,$9B,$F3,$FC,$DF,$F6,$2B,$71
	.dc.b	$34,$93,$EB,$6B,$FE,$E3,$79,$5B,$E7,$FB,$5F,$D9,$6B,$FE,$C8,$6B
	.dc.b	$FD,$51,$B6,$F9,$F7,$CF,$FC,$96,$E3,$FE,$F1,$6D,$FF,$66,$6B,$FD
	.dc.b	$2E,$B9,$57,$12,$57,$17,$EE,$74,$BE,$0F,$FB,$D4,$FF,$9D,$1B,$7F
	.dc.b	$E7,$86,$BF,$E7,$46,$BF,$E0,$55,$F3,$FD,$1B,$3C,$E3,$04,$70,$49
	.dc.b	$ED,$1C,$7A,$E4,$F8,$DB,$3F,$E0,$45,$F1,$FF,$8D,$6C,$FF,$A4,$72
	.dc.b	$46,$2B,$85,$6C,$95,$B9,$D4,$FC,$96,$BF,$EF,$ED,$7E,$E9,$27,$F1
	.dc.b	$EF,$E1,$FE,$77,$FD,$03,$1E,$16,$CF,$F8,$69,$AF,$D5,$6B,$FC,$83
	.dc.b	$04,$62,$11,$8F,$47,$F5,$53,$98,$C5,$7F,$94,$FE,$65,$F3,$FF,$A2
	.dc.b	$6C,$66,$30,$7C,$7F,$85,$91,$F0,$5B,$8F,$CD,$6B,$F0,$DA,$FF,$90
	.dc.b	$FE,$7F,$FC,$B7,$3F,$EB,$A6,$C7,$FA,$F9,$AF,$FA,$E9,$AF,$BF,$17
	.dc.b	$A7,$EC,$25,$7C,$97,$DF,$F1,$D2,$3E,$63,$6B,$E6,$61,$F1,$09,$3F
	.dc.b	$45,$FE,$31,$1F,$CB,$FE,$4E,$F6,$1A,$4A,$FF,$6C,$4E,$DB,$3B,$6C
	.dc.b	$EE,$7F,$A8,$C3,$E7,$3F,$98,$37,$BB,$F9,$43,$4F,$F7,$21,$67,$FC
	.dc.b	$87,$B3,$E1,$3E,$C1,$27,$EE,$36,$76,$D7,$FA,$24,$F1,$9F,$F4,$87
	.dc.b	$95,$FC,$8B,$D6,$34,$91,$A2,$57,$FB,$9A,$32,$FF,$BD,$BE,$7D,$F3
	.dc.b	$FF,$C1,$9B,$5F,$EF,$86,$CE,$E7,$44,$9D,$36,$3F,$BB,$DB,$E0,$3F
	.dc.b	$6B,$5F,$DE,$A5,$39,$E8,$FE,$A0,$D9,$D3,$67,$9F,$C1,$7F,$2A,$83
	.dc.b	$49,$E1,$72,$6D,$72,$6D,$7D,$2F,$9F,$FD,$41,$27,$F5,$8D,$8D,$46
	.dc.b	$08,$FA,$7E,$7F,$F3,$AF,$F8,$69,$A9,$A4,$F6,$FF,$2A,$D9,$FC,$8B
	.dc.b	$C9,$FC,$64,$7E,$2B,$8B,$04,$7F,$7D,$10,$FE,$04,$5A,$FC,$6B,$5F
	.dc.b	$C3,$B5,$FF,$17,$E4,$F8,$7F,$D9,$CF,$CE,$6C,$FD,$86,$BF,$E9,$79
	.dc.b	$E1,$D1,$F5,$A7,$33,$D1,$AF,$DC,$6C,$F0,$AF,$7F,$7F,$C2,$23,$82
	.dc.b	$57,$CD,$8C,$B6,$3F,$07,$F1,$2F,$88,$FC,$3B,$63,$31,$83,$F4,$9E
	.dc.b	$CC,$8F,$F7,$B3,$6B,$FD,$6B,$82,$31,$1F,$E3,$18,$C4,$F0,$3F,$31
	.dc.b	$E1,$7F,$B8,$7E,$7F,$FC,$23,$E7,$1A,$49,$DC,$65,$B1,$9B,$FF,$62
	.dc.b	$CE,$5F,$F7,$0F,$C7,$FF,$3F,$27,$8B,$FE,$2B,$7F,$A3,$61,$FE,$F1
	.dc.b	$D7,$C2,$4C,$7B,$BF,$D2,$2E,$47,$F5,$56,$9F,$D5,$4A,$7F,$55,$29
	.dc.b	$FD,$56,$9E,$07,$84,$93,$FD,$BE,$32,$FB,$BE,$7F,$E1,$91,$C1,$BE
	.dc.b	$2F,$51,$1C,$1F,$1B,$E7,$FE,$59,$1C,$13,$0F,$8F,$19,$6C,$65,$FF
	.dc.b	$34,$68,$9D,$46,$1F,$E8,$BF,$3F,$FE,$67,$83,$49,$E0,$FF,$12,$F9
	.dc.b	$F7,$C0,$D2,$63,$FC,$52,$32,$FF,$1B,$F9,$8F,$F0,$AF,$0F,$F9,$A5
	.dc.b	$B1,$9E,$0F,$EF,$D1,$8C,$BF,$D3,$6D,$7F,$D4,$61,$C3,$FE,$A6,$6B
	.dc.b	$98,$FF,$04,$71,$BE,$5D,$F0,$7F,$CA,$9B,$87,$F2,$A6,$DF,$F3,$46
	.dc.b	$BE,$0C,$3F,$C5,$BC,$DE,$9C,$7F,$43,$3E,$66,$BB,$6B,$E5,$6B,$E5
	.dc.b	$8E,$44,$BF,$17,$CF,$F9,$48,$E0,$93,$E5,$8C,$B6,$3F,$A4,$79,$3F
	.dc.b	$17,$E7,$FF,$D2,$CD,$8C,$E2,$9C,$B6,$33,$18,$23,$83,$FA,$7F,$CC
	.dc.b	$FF,$90,$AF,$FA,$AB,$75,$34,$D4,$FF,$A4,$FC,$BF,$F6,$63,$F6,$8C
	.dc.b	$1F,$ED,$E3,$1E,$27,$F8,$36,$FF,$D7,$91,$A6,$FC,$FF,$F9,$A9,$BE
	.dc.b	$97,$47,$94,$66,$7B,$36,$3F,$13,$F3,$FF,$E4,$4F,$7F,$FC,$3A,$D8
	.dc.b	$FE,$E0,$D7,$FD,$D6,$D7,$8F,$CF,$FF,$B5,$DD,$B5,$DB,$5D,$EF,$FE
	.dc.b	$1E,$F9,$57,$CF,$FC,$19,$F8,$11,$87,$E1,$18,$8D,$3F,$EF,$78,$D3
	.dc.b	$F1,$1F,$F6,$BB,$C0,$D7,$C0,$70,$CD,$3A,$9E,$13,$D3,$83,$EB,$FA
	.dc.b	$9F,$E3,$E6,$CF,$DE,$8C,$70,$9F,$F2,$1F,$1B,$E7,$DB,$E1,$B7,$FB
	.dc.b	$F6,$BF,$BA,$6B,$FA,$B6,$BF,$B9,$6B,$FB,$D6,$CF,$79,$F3,$74,$3E
	.dc.b	$7E,$92,$FF,$74,$D7,$F7,$4D,$7F,$BB,$47,$17,$F8,$A7,$B7,$FE,$70
	.dc.b	$6F,$99,$EE,$8C,$47,$F1,$26,$BF,$C0,$DA,$FF,$11,$6B,$FC,$55,$B8
	.dc.b	$7F,$B2,$E7,$D3,$F2,$52,$BF,$88,$35,$ED,$E8,$F8,$48,$F9,$9A,$E8
	.dc.b	$C1,$0F,$E4,$A7,$D4,$D7,$FD,$7E,$FF,$92,$F3,$FD,$AF,$A5,$FE,$7C
	.dc.b	$6C,$66,$76,$FF,$94,$7E,$7D,$F1,$7F,$E8,$28,$46,$1F,$F4,$2B,$5E
	.dc.b	$11,$FE,$88,$F3,$5E,$4F,$F9,$C2,$7F,$BF,$FB,$19,$8C,$47,$F7,$E8
	.dc.b	$C1,$1D,$90,$D1,$4F,$EB,$6E,$DF,$F0,$2F,$91,$FE,$98,$F7,$9E,$72
	.dc.b	$30,$D9,$FE,$ED,$1E,$3F,$16,$47,$F8,$51,$AF,$F8,$40,$E2,$43,$FB
	.dc.b	$F1,$DF,$CF,$7C,$ED,$24,$7F,$BC,$63,$2D,$8C,$BF,$63,$0F,$F0,$5C
	.dc.b	$62,$3F,$BB,$FC,$FF,$C7,$23,$FE,$28,$7D,$DF,$3F,$D2,$39,$35,$FE
	.dc.b	$CF,$48,$C4,$7F,$BB,$5A,$B5,$FA,$FF,$3F,$F5,$49,$1F,$E3,$66,$BF
	.dc.b	$E3,$46,$BF,$E0,$C7,$F0,$1D,$BF,$E0,$48,$FF,$21,$B5,$FF,$25,$F9
	.dc.b	$FA,$23,$A6,$BA,$3F,$CB,$7E,$FF,$F1,$5F,$5F,$FD,$89,$18,$21,$FE
	.dc.b	$A7,$F9,$07,$CF,$FE,$5D,$B1,$FE,$DA,$6B,$B8,$D1,$1D,$35,$FF,$3F
	.dc.b	$70,$FF,$25,$F6,$BF,$11,$18,$3F,$D0,$91,$92,$38,$21,$FE,$5A,$F4
	.dc.b	$FF,$40,$47,$FC,$C2,$D7,$46,$08,$7F,$B0,$FE,$6F,$F4,$EF,$EF,$9A
	.dc.b	$FF,$89,$1A,$83,$87,$FD,$09,$F3,$FF,$BB,$F7,$0D,$24,$7F,$C0,$08
	.dc.b	$CB,$63,$2F,$FB,$56,$32,$FF,$9D,$BB,$FF,$78,$47,$FC,$FA,$D7,$46
	.dc.b	$08,$7F,$B8,$BC,$9F,$4F,$07,$F5,$8E,$AF,$EA,$91,$82,$38,$24,$FF
	.dc.b	$B8,$A3,$8F,$C0,$7F,$3D,$8F,$F7,$6B,$5F,$F5,$B3,$5F,$F5,$FB,$57
	.dc.b	$C3,$BE,$79,$F3,$FF,$EF,$39,$CC,$61,$FF,$7B,$4E,$E7,$4D,$8C,$CE
	.dc.b	$DF,$E6,$B1,$96,$C6,$5F,$F2,$EC,$65,$FF,$57,$7C,$8F,$EB,$08,$FE
	.dc.b	$B1,$AF,$FA,$D9,$AF,$03,$FD,$5B,$F3,$FF,$CB,$C8,$E0,$87,$FA,$B9
	.dc.b	$CB,$4F,$81,$BB,$83,$FA,$13,$63,$31,$83,$FA,$37,$CF,$FF,$62,$83
	.dc.b	$49,$E4,$E8,$91,$F3,$63,$2F,$F2,$7F,$97,$FE,$58,$47,$07,$F2,$BF
	.dc.b	$27,$F9,$81,$1C,$1F,$CB,$FC,$6F,$E6,$84,$70,$7D,$8F,$9F,$7B,$DF
	.dc.b	$61,$B1,$98,$C1,$FD,$2F,$E7,$FF,$DC,$51,$FD,$B8,$78,$68,$CC,$7F
	.dc.b	$3B,$6B,$FC,$C9,$AF,$ED,$63,$87,$8F,$D4,$91,$D6,$32,$FF,$B9,$B4
	.dc.b	$FE,$D0,$47,$07,$F4,$DF,$37,$FA,$91,$1C,$10,$FE,$9D,$E4,$23,$51
	.dc.b	$87,$F9,$CF,$9B,$23,$F9,$CB,$5F,$67,$C2,$7C,$32,$3E,$7C,$91,$88
	.dc.b	$D9,$97,$0F,$E7,$BE,$79,$F3,$FF,$A8,$6C,$7E,$B1,$AF,$EA,$1A,$F1
	.dc.b	$F5,$F8,$11,$E0,$D7,$83,$5E,$07,$EE,$3E,$7F,$FC,$2A,$DF,$43,$F7
	.dc.b	$31,$88,$FD,$BB,$5F,$ED,$7E,$57,$FB,$29,$FA,$CD,$7F,$BD,$35,$DF
	.dc.b	$C2,$7F,$84,$E3,$11,$F8,$16,$BF,$8A,$6F,$07,$FB,$B2,$7F,$C5,$5F
	.dc.b	$33,$F8,$8A,$FE,$EC,$DF,$38,$D3,$54,$F9,$FF,$F3,$5B,$78,$3F,$C4
	.dc.b	$91,$82,$38,$25,$FE,$46,$3F,$CE,$76,$F9,$F7,$8F,$FE,$3D,$6D,$7E
	.dc.b	$8E,$75,$39,$3F,$C6,$5F,$3F,$FD,$53,$13,$49,$E3,$FF,$8D,$FC,$EF
	.dc.b	$39,$33,$C1,$31,$FD,$24,$65,$B1,$FE,$4F,$7F,$CA,$47,$F6,$13,$FC
	.dc.b	$38,$60,$FE,$C9,$8F,$F6,$5E,$CE,$04,$8F,$51,$B3,$8B,$97,$F8,$07
	.dc.b	$CF,$BB,$FF,$A3,$C7,$F9,$7B,$C4,$F0,$4F,$F9,$AB,$91,$18,$F0,$3C
	.dc.b	$91,$8A,$FE,$ED,$5C,$FC,$5F,$EF,$13,$99,$4E,$5E,$C7,$0A,$E7,$7D
	.dc.b	$63,$8B,$E2,$C3,$CB,$7C,$EF,$93,$6F,$8D,$F3,$38,$9E,$4A,$E3,$7C
	.dc.b	$F0,$CC,$F9,$2B,$8B,$4C,$E1,$DA,$B8,$47,$A0,$D1,$63,$E3,$46,$B0
	.dc.b	$F3,$61,$F3,$2B,$FD,$37,$5F,$05,$B1,$97,$A9,$AF,$9E,$7C,$FF,$F7
	.dc.b	$7A,$46,$27,$EC,$B5,$E1,$3F,$E9,$D7,$F5,$67,$FA,$7F,$97,$04,$70
	.dc.b	$3E,$83,$EB,$3F,$EE,$2F,$82,$FF,$18,$46,$23,$9B,$AE,$04,$EB,$4F
	.dc.b	$E9,$57,$6F,$F3,$15,$A7,$B4,$E9,$A6,$AB,$83,$F9,$FA,$FA,$AF,$8C
	.dc.b	$FA,$07,$F9,$0D,$CD,$75,$D3,$CD,$1C,$A7,$86,$18,$7B,$9F,$E4,$BA
	.dc.b	$F2,$46,$4F,$65,$F1,$9F,$7E,$1F,$ED,$2A,$DB,$FB,$0F,$84,$7C,$77
	.dc.b	$B7,$83,$DD,$C4,$9C,$E3,$FE,$59,$9C,$C7,$F4,$C8,$C1,$CE,$B9,$34
	.dc.b	$CE,$3F,$E9,$B9,$E6,$FF,$8F,$23,$C5,$E0,$70,$8D,$55,$C3,$F8,$F6
	.dc.b	$37,$A7,$F7,$38,$74,$E8,$FD,$DF,$C1,$66,$A4,$F8,$2F,$8B,$B8,$4E
	.dc.b	$B1,$F1,$1F,$EA,$F9,$7D,$0F,$DF,$8F,$E3,$17,$F8,$68,$C4,$7F,$18
	.dc.b	$9D,$9F,$CB,$CA,$FF,$C0,$09,$FE,$87,$EE,$7A,$F8,$4F,$B6,$B7,$BF
	.dc.b	$D0,$9C,$C7,$A4,$7C,$5F,$83,$0F,$33,$70,$FC,$CB,$67,$67,$8D,$C1
	.dc.b	$1C,$1F,$D1,$69,$EE,$79,$4F,$24,$D9,$F9,$54,$F7,$3C,$0D,$8F,$E9
	.dc.b	$CD,$C3,$E0,$35,$F9,$2D,$7F,$9F,$B5,$FF,$6F,$34,$FF,$70,$4F,$B7
	.dc.b	$95,$F0,$DF,$83,$6D,$F9,$27,$13,$F8,$9C,$DF,$08,$BF,$91,$18,$6C
	.dc.b	$66,$FF,$D0,$ED,$C7,$FD,$06,$D7,$FD,$02,$D7,$F4,$4D,$7F,$D2,$AD
	.dc.b	$F6,$1F,$57,$FE,$2A,$AF,$D0,$C6,$3C,$AF,$35,$77,$C7,$F8,$64,$E6
	.dc.b	$BE,$7F,$CB,$7C,$E2,$3F,$39,$B8,$FC,$E6,$BF,$39,$F9,$1E,$77,$C7
	.dc.b	$3F,$55,$B2,$31,$5F,$70,$FB,$64,$3E,$D9,$6B,$ED,$B5,$FB,$6F,$D5
	.dc.b	$F2,$7E,$A1,$D9,$A7,$ED,$7C,$2F,$F4,$B3,$FE,$69,$6C,$FF,$80,$CF
	.dc.b	$F3,$BB,$87,$FC,$0D,$D0,$8C,$47,$F9,$D5,$AF,$0F,$53,$F4,$93,$B6
	.dc.b	$CE,$B1,$FF,$01,$5F,$F0,$3E,$0F,$E1,$38,$7F,$00,$C3,$84,$E7,$A7
	.dc.b	$F8,$94,$78,$74,$B8,$9B,$EB,$61,$FC,$87,$CD,$FF,$23,$C7,$FD,$96
	.dc.b	$D7,$FD,$92,$D7,$F9,$53,$56,$7F,$97,$FA,$CF,$CF,$1E,$0D,$AE,$9F
	.dc.b	$8A,$FC,$0D,$F2,$BF,$C0,$F8,$91,$89,$FF,$51,$46,$27,$53,$9C,$7E
	.dc.b	$84,$E6,$B7,$39,$9F,$F1,$B4,$E5,$DC,$65,$FF,$53,$7E,$0C,$FB,$2F
	.dc.b	$8E,$DB,$F3,$61,$89,$FD,$C7,$87,$27,$F9,$A6,$2F,$99,$C6,$3F,$CC
	.dc.b	$B5,$FD,$C5,$E3,$9F,$FA,$9D,$EF,$0F,$B5,$3F,$C0,$FC,$49,$1D,$3D
	.dc.b	$AB,$8F,$3E,$7F,$E9,$F8,$FF,$5C,$6D,$7F,$94,$31,$FF,$67,$BF,$85
	.dc.b	$C3,$FA,$94,$3F,$12,$FF,$26,$4F,$EE,$9B,$87,$F4,$81,$FA,$A7,$F3
	.dc.b	$07,$FA,$5D,$71,$F3,$B9,$D7,$0C,$BF,$7E,$DC,$3F,$CC,$98,$71,$9C
	.dc.b	$D7,$F1,$FF,$86,$95,$E1,$8B,$8F,$B7,$CB,$FE,$48,$7F,$67,$C9,$F3
	.dc.b	$78,$BF,$D0,$4D,$AF,$9D,$5A,$9C,$9F,$A5,$8F,$E5,$C7,$3E,$3B,$19
	.dc.b	$24,$72,$8C,$BF,$DA,$A1,$F3,$8C,$BF,$AC,$46,$1B,$19,$8F,$F7,$C1
	.dc.b	$92,$B9,$9F,$E0,$D1,$96,$C6,$5F,$F4,$64,$65,$FE,$EF,$F5,$CF,$B1
	.dc.b	$FF,$05,$36,$7F,$C2,$67,$3D,$7F,$C3,$F1,$9A,$FC,$92,$FB,$1E,$27
	.dc.b	$CD,$F5,$ED,$FF,$83,$67,$4D,$9D,$34,$D7,$D8,$F8,$8F,$C6,$FB,$5F
	.dc.b	$7E,$3F,$D4,$34,$9D,$47,$03,$FA,$4C,$EF,$0F,$E0,$F5,$B9,$F5,$E7
	.dc.b	$FC,$2D,$FA,$3C,$FF,$38,$8F,$CC,$6D,$7C,$A6,$BF,$23,$E8,$BE,$39
	.dc.b	$FE,$33,$9E,$67,$BD,$F9,$7F,$25,$23,$FA,$EF,$C7,$BE,$17,$17,$F2
	.dc.b	$AA,$EF,$FD,$7A,$B5,$7F,$26,$7E,$AF,$D0,$48,$FE,$0B,$F0,$F7,$0F
	.dc.b	$A7,$37,$0F,$E0,$73,$6B,$F7,$5E,$AF,$F8,$8F,$9B,$FC,$53,$87,$F8
	.dc.b	$C5,$AF,$F8,$E9,$A8,$79,$4F,$D9,$FF,$D2,$B7,$FE,$4D,$6C,$FE,$A2
	.dc.b	$FF,$A2,$C6,$31,$FC,$4F,$C1,$FE,$38,$CF,$BD,$70,$FC,$34,$5C,$3F
	.dc.b	$A8,$5F,$19,$FC,$0C,$FE,$E7,$D7,$F5,$12,$3F,$DF,$59,$FF,$8D,$9F
	.dc.b	$F7,$E7,$23,$EE,$BF,$3A,$71,$8F,$CA,$3F,$C0,$FD,$1F,$88,$48,$E6
	.dc.b	$FF,$92,$65,$F7,$3F,$A6,$38,$4F,$E7,$5C,$34,$FE,$73,$38,$CF,$F4
	.dc.b	$F3,$75,$FE,$6F,$36,$BE,$76,$2F,$A4,$FF,$51,$61,$FD,$39,$AE,$9A
	.dc.b	$A3,$87,$81,$E7,$FB,$9F,$DD,$9F,$CD,$AB,$FD,$59,$8F,$EE,$A3,$F9
	.dc.b	$17,$5F,$E0,$C9,$8F,$CA,$6E,$3F,$85,$6B,$F8,$57,$E2,$79,$7F,$C2
	.dc.b	$0E,$73,$FB,$F7,$FA,$2F,$CE,$B8,$7F,$B9,$36,$7F,$B8,$B6,$FF,$B7
	.dc.b	$B5,$FE,$DA,$D7,$E2,$B5,$F8,$0D,$D0,$E7,$CB,$70,$7E,$26,$BF,$81
	.dc.b	$F2,$79,$31,$FE,$3E,$CC,$7F,$8B,$39,$69,$FE,$54,$C1,$D1,$EA,$24
	.dc.b	$7A,$9B,$7E,$A6,$BE,$A7,$F1,$5E,$37,$F2,$E3,$E2,$B6,$7F,$D8,$3C
	.dc.b	$2F,$9E,$7A,$9F,$EE,$BF,$79,$F1,$DF,$C8,$BD,$7C,$3F,$ED,$DA,$DE
	.dc.b	$1F,$EE,$78,$D3,$73,$FF,$50,$FD,$27,$07,$F9,$2F,$C4,$59,$FF,$54
	.dc.b	$CE,$78,$BE,$4C,$EE,$9F,$44,$FA,$A6,$DF,$E5,$59,$FE,$0D,$39,$8C
	.dc.b	$47,$09,$DB,$B9,$D3,$7E,$8F,$FD,$D4,$DF,$88,$E3,$29,$CC,$FF,$98
	.dc.b	$E3,$13,$FD,$93,$E6,$5F,$23,$FC,$5E,$9C,$3F,$21,$FE,$87,$F4,$B9
	.dc.b	$8C,$6B,$F8,$68,$C7,$07,$F1,$FE,$06,$3D,$30,$F9,$5B,$7D,$5F,$2B
	.dc.b	$71,$FD,$3B,$E7,$2E,$5F,$90,$73,$5B,$9F,$A5,$3F,$94,$C7,$CD,$E0
	.dc.b	$7F,$6C,$6E,$5F,$D7,$9B,$5F,$9E,$F8,$BB,$DD,$F9,$E7,$E6,$B6,$3F
	.dc.b	$57,$18,$EB,$FF,$7C,$C6,$08,$7E,$BA,$34,$7D,$6C,$7F,$4E,$F8,$1B
	.dc.b	$1F,$DE,$63,$C3,$3F,$2E,$3F,$C0,$31,$BE,$EE,$F1,$83,$F5,$18,$70
	.dc.b	$3C,$84,$7F,$80,$B5,$FE,$02,$D7,$F8,$0B,$E5,$8F,$F6,$B4,$7F,$82
	.dc.b	$FD,$AF,$01,$E7,$C7,$F4,$F1,$FE,$68,$EB,$F6,$46,$23,$FA,$F4,$60
	.dc.b	$8F,$F4,$86,$BA,$3F,$A4,$F2,$7B,$48,$FF,$B5,$BF,$37,$7C,$3F,$D4
	.dc.b	$6C,$7F,$A3,$1A,$FF,$A2,$9A,$FF,$A2,$1A,$FF,$BF,$BE,$6B,$E8,$6C
	.dc.b	$69,$ED,$E3,$B3,$FD,$CF,$85,$C3,$C6,$30,$D7,$FC,$B3,$3C,$1F,$F4
	.dc.b	$45,$F5,$76,$7F,$AB,$1C,$3C,$0F,$F4,$8C,$7C,$A2,$3F,$DB,$9F,$C4
	.dc.b	$43,$B3,$D2,$49,$FF,$4C,$FC,$57,$D5,$3E,$99,$1F,$EE,$EF,$2E,$9F
	.dc.b	$82,$48,$F2,$F9,$CF,$C8,$7E,$5E,$7F,$02,$7D,$15,$B7,$DA,$FE,$A2
	.dc.b	$B8,$6D,$6F,$8E,$9F,$E1,$B9,$CD,$F6,$9C,$BF,$EC,$EE,$1F,$F6,$D3
	.dc.b	$FE,$B2,$F8,$7F,$C5,$91,$FF,$07,$BF,$0B,$ED,$FE,$C9,$23,$FD,$73
	.dc.b	$F0,$D5,$FA,$31,$74,$FF,$5B,$CF,$E0,$33,$95,$F5,$37,$FF,$6F,$C7
	.dc.b	$F0,$0F,$57,$FD,$0D,$39,$9F,$39,$A7,$13,$BF,$77,$F1,$7F,$53,$27
	.dc.b	$83,$FC,$31,$7C,$EB,$DE,$C3,$63,$F9,$CB,$5F,$E7,$CD,$7F,$97,$1C
	.dc.b	$A3,$F9,$FB,$4E,$2E,$23,$84,$7F,$35,$6B,$F9,$1F,$B7,$FD,$09,$1F
	.dc.b	$CA,$EF,$FC,$CA,$B1,$1F,$9A,$9B,$AF,$E0,$EB,$53,$F1,$7E,$DF,$F5
	.dc.b	$64,$7F,$DD,$7F,$1B,$5F,$D1,$C5,$AF,$F7,$4C,$FE,$CB,$E4,$E5,$7C
	.dc.b	$F7,$E2,$FA,$D2,$3E,$2B,$ED,$23,$82,$1F,$21,$CB,$F2,$37,$0F,$93
	.dc.b	$FE,$02,$48,$FF,$80,$3F,$1D,$5F,$DC,$C5,$C7,$FD,$FF,$3F,$C2,$3E
	.dc.b	$C2,$57,$E1,$5F,$E4,$9F,$05,$88,$FE,$03,$16,$BF,$55,$87,$F9,$26
	.dc.b	$BF,$88,$FD,$24,$BF,$C7,$3F,$D4,$3E,$ED,$F0,$74,$93,$F5,$A3,$2F
	.dc.b	$F0,$C3,$F8,$27,$AB,$A7,$FD,$FF,$1A,$6C,$FE,$4F,$C1,$57,$F2,$2F
	.dc.b	$A3,$95,$F9,$F7,$F9,$F7,$C7,$62,$3F,$8D,$C5,$D3,$D7,$C8,$AF,$E5
	.dc.b	$7F,$59,$2F,$F5,$0F,$F4,$7E,$AC,$47,$F2,$48,$B7,$CA,$7F,$85,$7E
	.dc.b	$16,$F8,$7A,$4B,$FC,$AF,$DE,$7F,$88,$DB,$EC,$FF,$74,$8C,$47,$E9
	.dc.b	$B3,$46,$23,$FC,$2D,$18,$9E,$04,$8F,$F1,$64,$65,$B3,$C5,$FF,$5A
	.dc.b	$71,$FE,$2B,$EE,$BF,$A4,$92,$3F,$94,$FD,$2B,$E4,$7F,$C1,$7A,$BE
	.dc.b	$99,$F2,$BF,$E3,$68,$C3,$63,$37,$FE,$3D,$9C,$BF,$E3,$FF,$29,$87
	.dc.b	$C5,$8E,$FB,$7F,$31,$8C,$42,$BF,$9C,$57,$0F,$8D,$FE,$BF,$F5,$32
	.dc.b	$BF,$8E,$3C,$7B,$E4,$77,$C1,$1F,$E4,$CB,$F0,$B6,$33,$3B,$8C,$11
	.dc.b	$C1,$FE,$57,$DB,$F6,$3F,$01,$FD,$C0,$91,$FC,$C7,$87,$FC,$AE,$FA
	.dc.b	$7C,$EF,$9F,$18,$8F,$F7,$37,$B7,$FE,$0E,$23,$FE,$D2,$6B,$FE,$ED
	.dc.b	$6B,$FE,$55,$BF,$E0,$3C,$9F,$E0,$02,$47,$84,$FF,$0E,$1F,$E1,$A2
	.dc.b	$1F,$ED,$E2,$9F,$EF,$92,$9F,$EF,$93,$FC,$8B,$F7,$5F,$4B,$7F,$E4
	.dc.b	$CD,$C7,$FD,$9E,$D9,$FA,$DF,$4F,$FE,$2E,$7E,$01,$8F,$67,$F0,$D1
	.dc.b	$98,$FF,$2C,$B5,$FF,$0A,$35,$E0,$7E,$25,$C5,$9A,$6F,$37,$CC,$FF
	.dc.b	$23,$AF,$E8,$2D,$7C,$71,$AF,$1B,$FC,$C7,$2F,$1E,$9F,$E6,$47,$15
	.dc.b	$FA,$A6,$BF,$B3,$6B,$FB,$26,$BF,$EF,$26,$BF,$EE,$D6,$D7,$9A,$7F
	.dc.b	$CD,$19,$BC,$AF,$F3,$43,$6F,$FB,$D3,$5F,$97,$B3,$E8,$5F,$19,$FE
	.dc.b	$6F,$6C,$FE,$55,$AF,$F4,$76,$BF,$89,$6B,$FC,$09,$AF,$F4,$C6,$CF
	.dc.b	$79,$FF,$3F,$46,$29,$E6,$7A,$09,$7F,$E1,$19,$FF,$34,$6D,$FD,$19
	.dc.b	$FF,$35,$47,$87,$DC,$E6,$47,$F5,$9E,$E3,$AB,$FD,$14,$D8,$FF,$0E
	.dc.b	$35,$FC,$6F,$4B,$EF,$1F,$2D,$C2,$E5,$70,$7F,$C7,$93,$A8,$FF,$13
	.dc.b	$0F,$85,$FF,$4C,$34,$CB,$8F,$84,$CB,$23,$27,$28,$FF,$12,$7D,$2B
	.dc.b	$E3,$3F,$8C,$4B,$E9,$1E,$AF,$FB,$36,$BF,$AB,$C8,$D4,$67,$CE,$FF
	.dc.b	$69,$57,$09,$D5,$75,$E1,$E1,$39,$94,$E5,$E6,$70,$AE,$77,$D6,$38
	.dc.b	$BE,$3C,$3C,$F7,$CE,$F9,$76,$5C,$3F,$DC,$8D,$C3,$D0,$EE,$FD,$ED
	.dc.b	$C7,$AC,$F8,$4A,$79,$AB,$C6,$FB,$2B,$C3,$8F,$8E,$B7,$87,$C6,$AF
	.dc.b	$59,$C4,$CD,$70,$79,$4F,$F9,$31,$A5,$9F,$88,$FF,$94,$1A,$FD,$3F
	.dc.b	$8E,$7D,$2F,$18,$F9,$71,$87,$93,$5E,$4D,$7C,$1F,$17,$FE,$5A,$4D
	.dc.b	$23,$FE,$5E,$7E,$4E,$3F,$E6,$2F,$0B,$FD,$29,$A7,$DB,$AF,$C2,$56
	.dc.b	$9F,$C3,$09,$DD,$78,$21,$5C,$67,$31,$FE,$6E,$AE,$4F,$E3,$E3,$2F
	.dc.b	$48,$CB,$D7,$A2,$BE,$A3,$AD,$F9,$E3,$33,$C2,$B9,$D7,$A0,$D9,$E8
	.dc.b	$72,$F3,$BF,$35,$72,$8D,$9E,$57,$E0,$75,$39,$9E,$B7,$FA,$E6,$BF
	.dc.b	$B1,$46,$2B,$F8,$4C,$E6,$7F,$C3,$B3,$97,$FC,$D0,$D7,$8C,$6D,$FF
	.dc.b	$35,$C6,$5E,$31,$9E,$7F,$DC,$FC,$4D,$D0,$D2,$61,$FE,$1D,$8C,$B6
	.dc.b	$33,$7F,$EA,$0A,$E7,$3F,$E2,$3F,$BA,$78,$BD,$DB,$19,$8C,$1E,$7C
	.dc.b	$FF,$DF,$3E,$05,$9F,$EB,$47,$F8,$54,$B8,$7F,$A1,$DA,$FE,$2B,$06
	.dc.b	$09,$F0,$10,$F2,$16,$7D,$11,$B8,$C9,$23,$B7,$87,$E3,$24,$78,$DF
	.dc.b	$95,$F7,$1F,$D2,$23,$FC,$84,$D9,$F8,$8D,$7C,$CD,$7F,$DE,$0D,$56
	.dc.b	$8C,$37,$C0,$EE,$7C,$7C,$9F,$4F,$49,$87,$F8,$EE,$32,$F9,$3D,$BF
	.dc.b	$39,$1C,$13,$1F,$3C,$65,$B1,$FE,$5F,$FA,$6F,$D9,$B7,$8F,$F1,$DD
	.dc.b	$29,$CD,$7F,$B8,$A3,$13,$F1,$BC,$A9,$7F,$19,$B5,$F1,$9A,$FC,$66
	.dc.b	$BF,$E7,$CC,$5F,$4D,$F0,$49,$5E,$F7,$F3,$9D,$BF,$E7,$A8,$FF,$3A
	.dc.b	$36,$BF,$2C,$D7,$F1,$0D,$7F,$AC,$6D,$FE,$B0,$F4,$79,$92,$B9,$BF
	.dc.b	$45,$87,$CF,$2C,$F6,$F9,$BF,$19,$23,$C6,$D8,$D4,$60,$8F,$9B,$CE
	.dc.b	$F5,$92,$3D,$6F,$BB,$C7,$91,$EE,$6C,$FB,$BE,$1F,$E1,$92,$7E,$1B
	.dc.b	$F1,$BE,$1A,$47,$C6,$6C,$FF,$4F,$ED,$7A,$5F,$C2,$E3,$F8,$C3,$63
	.dc.b	$35,$FD,$35,$B3,$C6,$74,$49,$D3,$FC,$57,$CE,$7D,$2B,$E8,$B0,$DA
	.dc.b	$FF,$5F,$B6,$FF,$D9,$0D,$7F,$D1,$15,$FE,$66,$9F,$F4,$D3,$63,$FA
	.dc.b	$68,$E2,$FF,$2C,$4E,$BD,$1F,$E2,$15,$BA,$E7,$18,$24,$F4,$E0,$D9
	.dc.b	$1C,$12,$7B,$DA,$FF,$19,$1E,$16,$C6,$6B,$E6,$7D,$1B,$B6,$93,$0F
	.dc.b	$EF,$11,$96,$C6,$70,$FF,$02,$4E,$5B,$1F,$37,$E9,$3F,$C7,$15,$FE
	.dc.b	$E8,$6C,$FF,$BC,$5A,$FF,$BC,$1A,$FF,$AF,$3D,$9F,$8A,$49,$F9,$AD
	.dc.b	$8F,$7E,$5F,$EE,$B9,$C9,$23,$3A,$FC,$E2,$47,$D6,$6C,$6A,$30,$47
	.dc.b	$F2,$5F,$03,$F8,$F2,$47,$E3,$DB,$1A,$8C,$11,$FC,$0A,$FA,$27,$BF
	.dc.b	$FD,$96,$30,$D8,$CC,$7F,$8B,$CC,$9F,$8C,$F8,$7F,$C3,$91,$C1,$2B
	.dc.b	$E1,$C6,$5B,$1F,$E3,$9C,$7F,$D3,$BA,$7E,$87,$D7,$FB,$15,$FE,$22
	.dc.b	$9C,$D6,$E7,$33,$FE,$6B,$9C,$BF,$9C,$6B,$C6,$36,$FE,$45,$AB,$ED
	.dc.b	$7F,$C0,$4F,$F9,$D5,$AE,$7C,$BF,$DF,$46,$08,$E0,$97,$F1,$72,$E6
	.dc.b	$FC,$D8,$F9,$44,$7E,$07,$C6,$3E,$47,$F8,$03,$63,$31,$82,$38,$26
	.dc.b	$1E,$F9,$E0,$FE,$43,$DF,$FD,$C9,$27,$F8,$0B,$63,$51,$82,$3F,$E1
	.dc.b	$7F,$A2,$7C,$FF,$F8,$8D,$B3,$FE,$6F,$F0,$51,$8A,$E1,$5F,$C2,$3D
	.dc.b	$B9,$7F,$C4,$1B,$3D,$7D,$4E,$84,$9E,$96,$8C,$4E,$E7,$31,$88,$FF
	.dc.b	$0E,$4B,$E6,$FC,$C4,$AE,$6F,$F1,$AF,$12,$47,$F2,$26,$D7,$BF,$DA
	.dc.b	$F7,$12,$7C,$6F,$F2,$2F,$16,$47,$F2,$A6,$CF,$EF,$9A,$FF,$30,$6B
	.dc.b	$FE,$BF,$EE,$FE,$41,$18,$8F,$E4,$0D,$5F,$8E,$BE,$9B,$FA,$63,$6F
	.dc.b	$FD,$2C,$D9,$FF,$50,$35,$FD,$37,$D1,$7F,$9E,$DF,$D1,$35,$CF,$B1
	.dc.b	$FD,$36,$31,$1F,$E8,$5F,$9F,$71,$FF,$99,$69,$18,$C3,$FD,$5D,$18
	.dc.b	$9D,$4E,$6B,$FD,$38,$D7,$FD,$04,$D7,$B4,$70,$3F,$A6,$F4,$CB,$FE
	.dc.b	$AA,$D9,$FE,$AB,$F6,$AF,$11,$AF,$F9,$E6,$30,$D8,$CD,$FF,$B0,$E7
	.dc.b	$2D,$8C,$D7,$46,$BF,$EA,$E9,$DB,$9A,$FE,$E5,$F3,$CF,$A2,$A4,$C3
	.dc.b	$F8,$C4,$65,$B1,$9A,$FE,$25,$39,$6C,$79,$F2,$95,$E5,$6C,$F9,$5A
	.dc.b	$F9,$71,$F9,$4B,$1F,$89,$3E,$6F,$A2,$97,$E4,$6D,$7D,$4F,$3B,$FD
	.dc.b	$00,$FF,$AA,$F6,$FB,$B1,$8A,$FA,$ED,$9F,$C2,$35,$FF,$44,$78,$2F
	.dc.b	$13,$15,$FD,$D5,$B1,$98,$4F,$FA,$E5,$D1,$83,$FC,$23,$F4,$5F,$EA
	.dc.b	$2C,$3F,$DE,$4D,$9F,$F7,$E3,$5D,$FB,$9F,$EE,$37,$3E,$9F,$FB,$1D
	.dc.b	$BE,$7F,$28,$C4,$76,$6B,$E1,$FA,$27,$CE,$FF,$BB,$B2,$91,$89,$F3
	.dc.b	$1B,$9D,$34,$C9,$FE,$F5,$E6,$FE,$74,$E3,$3F,$D4,$F8,$9C,$23,$4D
	.dc.b	$8C,$C7,$2F,$21,$CD,$FE,$CE,$8F,$C6,$FB,$FF,$E3,$67,$FA,$F4,$65
	.dc.b	$FC,$6C,$65,$FC,$27,$B3,$E2,$7F,$52,$D7,$F3,$CD,$7E,$AF,$37,$FA
	.dc.b	$57,$1E,$4D,$8F,$D5,$4E,$DE,$B3,$A8,$C6,$1E,$8F,$A2,$F1,$36,$BF
	.dc.b	$4B,$18,$C4,$D2,$46,$88,$7A,$3A,$FB,$11,$F1,$B5,$D1,$82,$1F,$E3
	.dc.b	$0E,$6F,$89,$DF,$FE,$4F,$6E,$06,$9F,$27,$5F,$E3,$FF,$33,$77,$34
	.dc.b	$98,$7F,$22,$8C,$B6,$32,$FF,$22,$7E,$1F,$D0,$BE,$76,$92,$7F,$C4
	.dc.b	$11,$97,$ED,$7D,$A7,$F6,$16,$F9,$1F,$75,$46,$5F,$C2,$35,$FF,$3F
	.dc.b	$F0,$FB,$88,$FF,$A0,$1A,$FF,$95,$1A,$F0,$34,$43,$45,$3F,$76,$7F
	.dc.b	$BC,$BE,$83,$F8,$13,$FE,$03,$6B,$FE,$8C,$F3,$11,$88,$E3,$18,$EE
	.dc.b	$FA,$7F,$43,$27,$FD,$2A,$D7,$95,$3E,$87,$F3,$D7,$D1,$B1,$99,$F1
	.dc.b	$46,$34,$91,$E5,$6B,$D1,$AF,$92,$7E,$93,$93,$FD,$4F,$EE,$FE,$20
	.dc.b	$8E,$0F,$81,$F0,$FE,$22,$38,$3F,$D5,$FE,$F7,$E4,$08,$E0,$87,$FA
	.dc.b	$19,$CB,$4F,$F2,$2F,$AF,$FA,$CF,$4F,$FC,$EB,$18,$21,$FE,$41,$8D
	.dc.b	$11,$C9,$FE,$BE,$FA,$07,$B1,$F8,$B6,$C7,$F0,$96,$BF,$C3,$DA,$FF
	.dc.b	$B1,$BD,$F7,$A3,$F9,$36,$C7,$F1,$16,$BF,$C5,$DA,$F1,$8E,$04,$78
	.dc.b	$3F,$C2,$3C,$3E,$F1,$FE,$7B,$6B,$FE,$C7,$6F,$59,$A4,$C3,$91,$27
	.dc.b	$FD,$BF,$F1,$3F,$C0,$08,$E0,$87,$FA,$D9,$CB,$4F,$89,$E7,$FC,$72
	.dc.b	$38,$3F,$98,$7D,$87,$F2,$54,$D2,$73,$FE,$E5,$B1,$C2,$30,$7F,$14
	.dc.b	$EB,$91,$FE,$F6,$6B,$FE,$CA,$6B,$FE,$FC,$6B,$FE,$4D,$C9,$EC,$FF
	.dc.b	$1E,$24,$7F,$C0,$06,$BF,$ED,$46,$BF,$BC,$F0,$3F,$DE,$91,$FC,$EB
	.dc.b	$E8,$1F,$41,$49,$5F,$EB,$78,$CB,$FD,$47,$B1,$F0,$9F,$C5,$DB,$19
	.dc.b	$8C,$11,$C1,$0F,$EB,$D1,$C1,$FF,$74,$FC,$FF,$F3,$82,$3F,$D8,$F1
	.dc.b	$64,$9C,$CF,$03,$51,$96,$C7,$F4,$2E,$59,$1A,$6B,$FE,$F8,$6B,$F1
	.dc.b	$67,$F6,$2E,$4F,$F6,$C7,$CE,$BD,$0F,$EC,$2D,$8C,$C6,$0F,$E7,$5F
	.dc.b	$7A,$F8,$CA,$49,$FE,$F5,$19,$6C,$65,$FF,$01,$46,$5B,$1F,$C9,$3E
	.dc.b	$71,$F0,$B4,$93,$FE,$B7,$8C,$B7,$D8,$E0,$47,$06,$FE,$07,$A1,$1C
	.dc.b	$1C,$F9,$FC,$04,$70,$77,$E6,$F2,$11,$C1,$0F,$E8,$D7,$FE,$0D,$F1
	.dc.b	$4D,$3F,$C2,$1C,$3F,$7B,$CB,$FE,$F2,$4A,$FB,$AF,$C1,$E7,$91,$EF
	.dc.b	$6B,$D9,$AF,$F5,$2A,$FD,$3B,$B3,$F0,$E7,$F8,$18,$86,$0F,$F7,$07
	.dc.b	$82,$FA,$AF,$EA,$6D,$8C,$C6,$08,$E0,$87,$F6,$58,$E0,$FF,$6C,$1F
	.dc.b	$1A,$F9,$C7,$C7,$7F,$8F,$63,$0D,$8C,$CF,$F9,$16,$72,$F9,$BD,$FF
	.dc.b	$29,$1C,$10,$F1,$39,$69,$FE,$4F,$E7,$7C,$63,$C8,$FF,$1E,$B6,$74
	.dc.b	$D7,$D5,$3C,$27,$FB,$BF,$E2,$6F,$97,$A4,$AF,$F2,$E7,$12,$75,$18
	.dc.b	$6B,$FE,$0B,$6B,$FE,$0A,$6B,$FE,$8E,$8E,$53,$FE,$40,$F9,$C7,$AD
	.dc.b	$49,$5F,$C2,$63,$2F,$8B,$EC,$3F,$A7,$D9,$A4,$F2,$3F,$C2,$2D,$9E
	.dc.b	$11,$83,$C1,$EB,$C8,$EC,$D7,$FC,$49,$3F,$ED,$A7,$C1,$D3,$D0,$8F
	.dc.b	$C4,$F9,$C7,$CE,$FF,$88,$1B,$19,$8C,$11,$C1,$0F,$F1,$94,$70,$7F
	.dc.b	$1D,$D5,$FE,$B4,$F3,$7F,$CC,$B1,$FD,$D8,$FF,$5B,$C6,$5F,$E5,$C6
	.dc.b	$43,$97,$99,$97,$0F,$F4,$B3,$80,$F9,$BF,$C1,$72,$FF,$B1,$39,$DE
	.dc.b	$6F,$F9,$32,$7F,$A0,$46,$2B,$FD,$93,$18,$9D,$71,$7E,$9A,$7F,$4B
	.dc.b	$18,$BE,$8F,$3F,$27,$FC,$AF,$87,$3A,$F1,$46,$27,$84,$72,$3B,$DF
	.dc.b	$5A,$E1,$87,$86,$7C,$77,$C5,$DE,$1E,$06,$9C,$CE,$E6,$4A,$66,$7E
	.dc.b	$1B,$6F,$C1,$87,$A6,$38,$4F,$85,$F9,$53,$A7,$B9,$FC,$26,$39,$B6
	.dc.b	$3E,$33,$6B,$FD,$8D,$2F,$61,$F3,$78,$6D,$7F,$B3,$E7,$53,$93,$FD
	.dc.b	$25,$0F,$9B,$A4,$7F,$D2,$CD,$CF,$E8,$43,$C1,$7C,$1F,$C3,$81,$CC
	.dc.b	$F9,$06,$DD,$39,$9E,$73,$CA,$36,$67,$4F,$E5,$51,$99,$F0,$CE,$4F
	.dc.b	$F4,$CC,$66,$7F,$DD,$97,$E3,$8C,$4F,$F3,$38,$C5,$F8,$9B,$1E,$06
	.dc.b	$DF,$66,$DF,$EA,$BE,$17,$FC,$67,$7F,$E1,$76,$DF,$EC,$9A,$FF,$8C
	.dc.b	$5A,$F1,$F0,$FD,$CE,$5A,$E6,$7A,$46,$21,$D1,$FA,$CF,$83,$C3,$6B
	.dc.b	$F9,$7E,$09,$CB,$63,$31,$82,$38,$3F,$9B,$FC,$DF,$F8,$7D,$BE,$67
	.dc.b	$68,$C6,$1F,$EC,$16,$BF,$E4,$D9,$FE,$94,$7F,$8E,$4A,$FF,$14,$6B
	.dc.b	$BE,$FE,$91,$0F,$09,$63,$B3,$5F,$F6,$6C,$FE,$3A,$36,$77,$F8,$8F
	.dc.b	$8D,$CD,$F9,$26,$C1,$A7,$C8,$F9,$B7,$CD,$D2,$4F,$F5,$88,$CB,$63
	.dc.b	$38,$7F,$B8,$67,$2D,$8F,$0D,$FF,$90,$6B,$FC,$B4,$D9,$FE,$6A,$D7
	.dc.b	$FC,$E8,$D7,$FA,$F5,$3E,$6F,$CC,$4A,$F2,$36,$7C,$4D,$7F,$DE,$33
	.dc.b	$C6,$7F,$C1,$1F,$0F,$FE,$0E,$6F,$1F,$E2,$FA,$13,$9B,$FF,$0B,$46
	.dc.b	$27,$E3,$7C,$54,$BF,$88,$DA,$F8,$4D,$7C,$2D,$7F,$A7,$E4,$F9,$BF
	.dc.b	$8A,$4A,$F8,$AF,$E6,$BB,$1F,$1D,$49,$3F,$E0,$A8,$CB,$63,$37,$FE
	.dc.b	$2D,$E5,$FF,$4A,$C6,$23,$E6,$CF,$D4,$21,$F5,$0F,$07,$C9,$47,$B3
	.dc.b	$63,$A5,$A3,$11,$FC,$70,$FD,$9C,$66,$78,$9E,$7F,$0B,$D8,$47,$07
	.dc.b	$AF,$C4,$F8,$04,$70,$7B,$FE,$0B,$E2,$11,$C1,$F0,$FC,$EF,$90,$47
	.dc.b	$07,$F9,$3F,$0F,$E6,$3D,$8F,$80,$FE,$15,$08,$C4,$EE,$31,$3A,$9C
	.dc.b	$C7,$36,$BF,$E5,$B6,$BF,$0D,$AF,$82,$3F,$8C,$79,$8F,$9B,$7C,$CE
	.dc.b	$1B,$87,$F9,$4D,$B7,$FE,$4E,$6B,$FE,$B5,$F3,$FF,$7B,$7F,$E5,$8A
	.dc.b	$FF,$60,$C6,$08,$E0,$93,$FE,$24,$9E,$0D,$E0,$E6,$49,$FF,$65,$B6
	.dc.b	$35,$18,$23,$F3,$7E,$6F,$F8,$64,$7C,$96,$C6,$7B,$3E,$7C,$62,$3F
	.dc.b	$D0,$4D,$7F,$14,$D7,$FB,$3B,$53,$9B,$7C,$DF,$F9,$BE,$FF,$CF,$4D
	.dc.b	$AF,$F3,$AB,$5F,$CB,$4F,$0F,$27,$E4,$FB,$32,$FF,$DD,$2D,$AF,$8C
	.dc.b	$D7,$FD,$C8,$D7,$FC,$93,$5F,$51,$BA,$FD,$42,$57,$FB,$D1,$B1,$F4
	.dc.b	$5A,$FE,$27,$DF,$FC,$71,$27,$F2,$5C,$08,$C4,$FE,$45,$AF,$19,$FC
	.dc.b	$F0,$F9,$A7,$BF,$FD,$9E,$30,$D8,$CC,$70,$6C,$FE,$17,$D2,$FC,$69
	.dc.b	$27,$F5,$5D,$8C,$D3,$9A,$E1,$5A,$9C,$B6,$7F,$43,$EA,$7D,$92,$70
	.dc.b	$E0,$93,$FE,$B3,$8C,$B6,$3F,$6F,$F3,$7F,$E0,$8C,$3F,$C5,$6D,$4D
	.dc.b	$27,$97,$C0,$95,$FB,$D8,$E4,$FC,$78,$F8,$84,$7E,$23,$FB,$8F,$88
	.dc.b	$7B,$94,$93,$FE,$C9,$8C,$B6,$33,$7F,$EC,$69,$CB,$63,$F7,$3E,$F4
	.dc.b	$AF,$90,$D9,$FD,$B3,$5F,$E5,$DA,$FD,$18,$C4,$7F,$B0,$FE,$69,$F3
	.dc.b	$7F,$E2,$26,$DF,$FB,$41,$B3,$FD,$15,$AF,$4D,$DF,$01,$FC,$30,$95
	.dc.b	$F6,$DE,$9E,$B4,$8E,$8D,$AE,$8D,$7F,$A2,$78,$AC,$3F,$CE,$55,$C0
	.dc.b	$F2,$7C,$DF,$9C,$8E,$0F,$C5,$78,$DF,$C7,$88,$E0,$F6,$7B,$1E,$F2
	.dc.b	$38,$3F,$25,$E3,$7F,$26,$23,$82,$63,$FC,$8E,$32,$D8,$CB,$F1,$63
	.dc.b	$2D,$8F,$E8,$DF,$34,$F9,$BF,$E9,$6D,$8C,$C6,$0F,$E8,$3F,$37,$FE
	.dc.b	$7F,$AF,$ED,$07,$EA,$DC,$3F,$C5,$9A,$FF,$A2,$37,$46,$23,$FD,$1D
	.dc.b	$F2,$AE,$2F,$F3,$27,$B4,$8C,$63,$FD,$86,$31,$3A,$9C,$D7,$FA,$29
	.dc.b	$AF,$FA,$21,$AF,$FA,$62,$78,$D7,$39,$CC,$FF,$B9,$B0,$75,$7F,$54
	.dc.b	$25,$7F,$03,$7F,$CF,$1F,$31,$FE,$1C,$D6,$F8,$C6,$92,$BF,$BC,$4E
	.dc.b	$DB,$3B,$6C,$EF,$0F,$97,$83,$E6,$1F,$09,$FD,$C1,$B7,$FE,$03,$9D
	.dc.b	$4E,$49,$19,$9D,$D7,$46,$BF,$D2,$9A,$FF,$AF,$31,$FF,$1B,$EB,$EC
	.dc.b	$25,$7A,$1B,$3F,$80,$9D,$CE,$49,$1D,$FE,$6F,$D0,$48,$F2,$3F,$67
	.dc.b	$CF,$FF,$62,$47,$D7,$6C,$FD,$7D,$91,$88,$FC,$3B,$67,$FD,$05,$E2
	.dc.b	$5F,$07,$FA,$74,$7F,$B5,$8F,$F1,$E1,$A3,$15,$98,$DB,$FE,$B6,$F9
	.dc.b	$B7,$B9,$FE,$E1,$8F,$F1,$84,$63,$8F,$FC,$58,$D8,$CC,$ED,$FD,$1C
	.dc.b	$3D,$4F,$F1,$A4,$72,$6B,$A6,$BF,$E5,$18,$E1,$E4,$7F,$62,$F9,$27
	.dc.b	$CD,$FA,$AB,$D9,$18,$7D,$67,$FB,$D9,$C3,$FE,$D2,$CD,$18,$7F,$49
	.dc.b	$E2,$73,$9F,$D8,$B7,$3F,$D8,$37,$7F,$E8,$F3,$DB,$97,$F2,$9E,$FF
	.dc.b	$F9,$35,$B1,$F8,$F8,$C7,$2F,$FB,$5E,$3F,$43,$ED,$72,$7F,$10,$D7
	.dc.b	$FC,$E1,$CB,$FE,$98,$C7,$FC,$B0,$D8,$E3,$18,$9E,$11,$F9,$76,$A9
	.dc.b	$E0,$FE,$7F,$E6,$FC,$64,$7F,$48,$D7,$46,$08,$79,$BA,$FC,$04,$7D
	.dc.b	$6D,$74,$60,$87,$F9,$57,$9B,$FC,$E5,$E3,$FF,$83,$DB,$A7,$E7,$27
	.dc.b	$FC,$1D,$1F,$E7,$7D,$BF,$90,$7C,$9D,$DC,$D2,$57,$F1,$96,$BF,$C8
	.dc.b	$5A,$FF,$91,$BE,$6D,$F3,$1F,$5D,$B1,$FC,$25,$AF,$F1,$06,$BF,$CF
	.dc.b	$3E,$6D,$E4,$FF,$A2,$E1,$18,$7F,$D1,$CD,$7F,$D2,$CD,$7F,$BA,$70
	.dc.b	$FC,$02,$3F,$E2,$D6,$BF,$E7,$B6,$BC,$0D,$10,$D1,$4F,$EE,$BF,$33
	.dc.b	$FE,$4D,$7F,$D4,$DE,$52,$31,$1F,$EC,$B6,$BC,$3B,$BE,$9F,$CD,$C8
	.dc.b	$FF,$20,$35,$E1,$F3,$7F,$EA,$BB,$E8,$DB,$34,$C6,$92,$BC,$73,$F3
	.dc.b	$10,$AF,$34,$EE,$3C,$CD,$7D,$AF,$E2,$3D,$E9,$1F,$E5,$56,$BE,$DF
	.dc.b	$87,$F2,$11,$F0,$BF,$90,$F7,$E4,$7F,$97,$5A,$FF,$AA,$DA,$FF,$8A
	.dc.b	$DA,$FF,$79,$F4,$BF,$CE,$6F,$F9,$1F,$D7,$FE,$05,$18,$8F,$F2,$34
	.dc.b	$60,$8E,$C8,$7F,$B6,$FE,$6F,$FC,$E2,$FF,$B5,$DA,$E7,$D7,$FC,$84
	.dc.b	$62,$3F,$C1,$FF,$00,$F4,$7F,$28,$D8,$FF,$70,$35,$FF,$74,$35,$E3
	.dc.b	$1C,$08,$F0,$7F,$8E,$78,$7D,$E3,$FC,$30,$D7,$FD,$10,$DE,$BF,$E5
	.dc.b	$B8,$70,$24,$FF,$75,$9D,$1C,$7B,$3F,$D2,$1F,$13,$FC,$08,$93,$FE
	.dc.b	$8F,$6B,$FE,$DD,$6B,$FE,$5D,$7E,$1F,$9D,$23,$E4,$35,$FF,$19,$8F
	.dc.b	$9B,$FE,$4A,$DE,$8F,$DD,$46,$08,$E0,$91,$C8,$E0,$7F,$A8,$FB,$3F
	.dc.b	$8C,$91,$C1,$0F,$F7,$A3,$96,$9F,$EA,$BF,$67,$F9,$09,$1C,$10,$FF
	.dc.b	$7E,$39,$69,$FE,$1F,$F9,$A7,$CD,$FF,$51,$6C,$66,$30,$7F,$17,$EC
	.dc.b	$7C,$2D,$24,$FB,$63,$2D,$8C,$BF,$2A,$32,$D8,$FE,$73,$F3,$72,$3F
	.dc.b	$D8,$CD,$7F,$CC,$7B,$7F,$B3,$63,$11,$FE,$95,$6B,$FE,$F0,$E5,$FE
	.dc.b	$86,$47,$FD,$A0,$D7,$9F,$17,$EB,$5F,$F4,$F4,$65,$FE,$16,$7F,$B6
	.dc.b	$FE,$6F,$FC,$32,$FF,$91,$5A,$FF,$93,$1A,$F0,$94,$62,$39,$FC,$DB
	.dc.b	$E6,$FF,$A0,$36,$3F,$08,$D7,$F1,$ED,$7F,$D0,$6D,$79,$C7,$22,$3F
	.dc.b	$CD,$BE,$6D,$F0,$BF,$CB,$5B,$19,$8C,$19,$F6,$24,$71,$6B,$FE,$92
	.dc.b	$F0,$55,$FE,$C2,$8E,$DF,$03,$D4,$8F,$87,$9F,$C0,$47,$07,$97,$9B
	.dc.b	$C8,$47,$04,$3C,$15,$FE,$12,$BF,$0C,$EE,$7E,$CF,$95,$FE,$00,$23
	.dc.b	$FB,$6E,$16,$FF,$E4,$98,$FF,$67,$E0,$E6,$F8,$84,$8F,$90,$D7,$C2
	.dc.b	$D7,$8B,$FE,$0E,$61,$DE,$7F,$BD,$57,$82,$3F,$AE,$78,$0F,$9B,$A4
	.dc.b	$8F,$ED,$71,$96,$C6,$5F,$A1,$19,$6C,$7F,$09,$F8,$E7,$CD,$BE,$3F
	.dc.b	$0D,$AF,$E2,$F3,$A9,$C9,$E4,$F7,$A4,$F8,$9A,$F8,$1A,$FF,$BD,$B0
	.dc.b	$FF,$53,$BB,$3F,$D4,$DF,$1A,$F1,$FF,$CB,$0D,$8C,$C6,$0F,$EF,$DF
	.dc.b	$36,$F8,$CA,$49,$FF,$06,$C6,$5B,$19,$7F,$25,$19,$6C,$7F,$90,$BE
	.dc.b	$6F,$FD,$9B,$1F,$81,$6B,$FC,$21,$AF,$EF,$36,$46,$23,$FC,$13,$C8
	.dc.b	$F9,$BF,$EA,$8D,$F1,$FF,$C2,$D1,$82,$38,$24,$FF,$89,$A3,$FA,$A5
	.dc.b	$7F,$94,$E3,$C5,$EA,$F9,$C8,$EF,$A3,$A9,$1C,$1F,$E3,$0F,$9A,$7C
	.dc.b	$DF,$F8,$9D,$B1,$98,$C1,$1C,$10,$FD,$D4,$70,$7F,$95,$9F,$DE,$4F
	.dc.b	$F1,$4E,$7F,$E2,$AF,$2F,$FC,$DB,$19,$11,$97,$F5,$C6,$5C,$3D,$23
	.dc.b	$89,$FE,$5D,$F9,$BF,$C0,$72,$FF,$8D,$B9,$5E,$5F,$F2,$E9,$FE,$87
	.dc.b	$18,$BF,$F1,$DC,$62,$75,$C2,$BE,$51,$86,$D7,$0F,$1B,$F9,$CE,$1C
	.dc.b	$E1,$87,$37,$AE,$1C,$DF,$D5,$61,$CD,$A7,$33,$A9,$CA,$FA,$C6,$EF
	.dc.b	$B5,$FB,$AB,$89,$B3,$8B,$E1,$7C,$77,$E1,$9E,$2D,$AF,$89,$7D,$AB
	.dc.b	$C4,$74,$BE,$D7,$C3,$0F,$F0,$64,$78,$67,$FA,$C9,$DA,$33,$87,$BE
	.dc.b	$33,$3F,$E9,$2F,$9B,$E3,$EB,$9A,$49,$FF,$4B,$7C,$DF,$FA,$5D,$34
	.dc.b	$8E,$87,$CE,$7D,$15,$E0,$BE,$0B,$FC,$EF,$39,$9E,$67,$0D,$BD,$35
	.dc.b	$B8,$C1,$EC,$BE,$33,$AB,$B3,$93,$67,$FA,$12,$B9,$5F,$7B,$FB,$73
	.dc.b	$E4,$AA,$F8,$B5,$F4,$5F,$85,$E1,$1D,$0E,$05,$3D,$37,$C2,$BC,$71
	.dc.b	$FE,$0A,$BA,$F0,$9E,$66,$DF,$6F,$03,$E1,$3F,$DA,$B7,$FE,$9C,$DF
	.dc.b	$E1,$4E,$7D,$0F,$89,$18,$AF,$D5,$D6,$CF,$DD,$BC,$AF,$F5,$11,$8A
	.dc.b	$E2,$D3,$84,$7E,$FA,$79,$BC,$0C,$4F,$07,$F8,$1F,$3B,$BB,$18,$FF
	.dc.b	$2C,$75,$D7,$FB,$EF,$9D,$CD,$6F,$2E,$F7,$B6,$C6,$70,$F0,$5F,$19
	.dc.b	$CE,$1E,$19,$7C,$0F,$B5,$FC,$FB,$5F,$F1,$CB,$5F,$F6,$2B,$5E,$3C
	.dc.b	$3C,$E3,$DF,$AF,$E3,$E7,$AE,$99,$AF,$81,$E2,$3F,$2A,$7C,$0E,$EF
	.dc.b	$FD,$A3,$39,$8C,$47,$F8,$4A,$77,$9F,$FA,$2A,$B4,$D8,$CC,$FE,$E2
	.dc.b	$72,$D8,$CB,$FE,$4D,$6B,$FC,$E3,$91,$F1,$DF,$49,$B7,$F7,$2F,$9E
	.dc.b	$7F,$8A,$9C,$DF,$DD,$E0,$FD,$A7,$C5,$FA,$1B,$87,$F8,$AC,$46,$5A
	.dc.b	$65,$C1,$1C,$1F,$E3,$3E,$57,$C8,$7E,$19,$B1,$F8,$23,$9D,$F8,$E3
	.dc.b	$2F,$97,$EF,$72,$FF,$79,$6B,$FD,$61,$AF,$F5,$66,$E0,$69,$30,$F7
	.dc.b	$3A,$C7,$FC,$6D,$87,$E5,$1B,$7E,$1A,$F9,$9E,$57,$E7,$A9,$F3,$BF
	.dc.b	$0C,$8F,$C7,$6D,$7F,$A6,$23,$66,$48,$FF,$A7,$32,$78,$7F,$14,$ED
	.dc.b	$56,$BF,$4F,$97,$2B,$FF,$51,$B5,$F8,$93,$EC,$24,$FF,$98,$E7,$53
	.dc.b	$F1,$E7,$2F,$FA,$B3,$2F,$F0,$6D,$7C,$1F,$0F,$F7,$77,$DF,$6F,$C4
	.dc.b	$47,$EF,$3F,$13,$FF,$35,$57,$FB,$93,$A1,$18,$C7,$FD,$10,$D7,$FD
	.dc.b	$57,$1A,$F3,$7F,$4A,$DA,$E5,$5B,$AF,$F1,$9B,$54,$ED,$FF,$4A,$7C
	.dc.b	$A3,$E1,$7F,$87,$B6,$FF,$8D,$D6,$F2,$FE,$94,$A7,$57,$FC,$3B,$03
	.dc.b	$ED,$3D,$DE,$C5,$C3,$F0,$9F,$5F,$FA,$5D,$8D,$24,$7F,$A4,$63,$FD
	.dc.b	$D6,$FE,$2F,$DC,$FE,$47,$F9,$97,$E3,$34,$FD,$94,$7F,$2B,$5F,$64
	.dc.b	$F5,$FF,$84,$C6,$1B,$19,$BF,$F3,$EB,$73,$FE,$FF,$5B,$9C,$D7,$02
	.dc.b	$3F,$B3,$EF,$FD,$91,$E1,$BE,$55,$FD,$3B,$9B,$FA,$43,$FC,$1F,$E5
	.dc.b	$3F,$86,$1F,$84,$FA,$BB,$87,$E9,$E2,$E5,$FB,$79,$B5,$CF,$93,$F8
	.dc.b	$01,$FA,$9F,$84,$B8,$7F,$42,$C7,$F1,$15,$B8,$FE,$09,$3A,$C3,$FA
	.dc.b	$97,$77,$F1,$2F,$96,$FE,$67,$DA,$7E,$57,$CB,$BE,$B7,$31,$31,$FE
	.dc.b	$45,$5B,$7F,$AF,$E4,$BE,$11,$AA,$FE,$55,$ED,$48,$C1,$FC,$15,$FE
	.dc.b	$35,$F0,$72,$3F,$97,$7D,$E7,$F8,$75,$F6,$46,$74,$FE,$38,$FF,$4A
	.dc.b	$9D,$46,$27,$F8,$87,$D9,$7F,$39,$3F,$63,$F5,$D6,$32,$49,$FE,$49
	.dc.b	$19,$7F,$BB,$74,$BE,$33,$F8,$CC,$7E,$70,$FF,$18,$B8,$7F,$C4,$8D
	.dc.b	$73,$E2,$FF,$58,$9E,$4D,$9D,$46,$2F,$FC,$9B,$5C,$1F,$E7,$DD,$BF
	.dc.b	$CF,$4F,$F0,$34,$EE,$7F,$9E,$47,$F5,$5F,$B2,$95,$FC,$73,$EB,$ED
	.dc.b	$7F,$53,$B5,$63,$87,$FC,$43,$E2,$AF,$18,$7F,$A6,$39,$FF,$C6,$75
	.dc.b	$F9,$9C,$3F,$BB,$FE,$C0,$DF,$FA,$84,$4B,$FC,$83,$A8,$C6,$7F,$2B
	.dc.b	$EE,$BF,$81,$D7,$BA,$91,$89,$FF,$2A,$B5,$F9,$D1,$A7,$E5,$46,$5F
	.dc.b	$F2,$93,$5F,$F3,$1B,$63,$84,$FF,$09,$FA,$2F,$EF,$47,$CA,$9F,$06
	.dc.b	$3F,$DE,$63,$FB,$55,$CA,$FE,$D5,$F3,$F6,$3F,$B5,$16,$BF,$C9,$32
	.dc.b	$9C,$3F,$E0,$9F,$B4,$FF,$04,$1F,$D0,$3E,$C6,$C7,$F6,$38,$B7,$FD
	.dc.b	$CE,$6D,$FF,$9D,$7D,$B7,$E6,$FF,$D6,$B0,$FC,$2C,$63,$CF,$FF,$11
	.dc.b	$B6,$33,$3B,$C3,$FD,$2F,$E6,$3C,$97,$BB,$FC,$A0,$FF,$28,$61,$C6
	.dc.b	$BF,$C8,$B3,$AC,$3F,$BC,$57,$F8,$84,$E0,$E0,$93,$F8,$D8,$CB,$63
	.dc.b	$FB,$6F,$87,$FE,$04,$23,$FD,$A5,$FE,$81,$EF,$FF,$8B,$4F,$EC,$AD
	.dc.b	$8F,$EE,$BB,$FF,$83,$08,$FF,$71,$7F,$A4,$7C,$07,$F8,$D4,$FE,$DA
	.dc.b	$D8,$FF,$08,$45,$BF,$F1,$04,$DD,$BF,$C7,$13,$6B,$FC,$41,$36,$BF
	.dc.b	$C4,$1F,$47,$FE,$1C,$24,$FF,$80,$5F,$EB,$1E,$87,$F9,$14,$FE,$FA
	.dc.b	$D8,$FF,$23,$7E,$AC,$ED,$FF,$2C,$F9,$08,$C4,$FF,$96,$5B,$87,$F9
	.dc.b	$5D,$AF,$F0,$F9,$E1,$5C,$A7,$33,$FE,$66,$6A,$9D,$D7,$F0,$CF,$55
	.dc.b	$F8,$83,$CD,$FF,$46,$B6,$7C,$F8,$E2,$7F,$CE,$BF,$10,$E0,$F4,$B6
	.dc.b	$BF,$CE,$0D,$CB,$FC,$DC,$D7,$D0,$D7,$FC,$F6,$D7,$CC,$DC,$8E,$72
	.dc.b	$F3,$F9,$1E,$63,$C7,$85,$C3,$FD,$3F,$9A,$71,$8F,$F5,$CC,$3C,$9C
	.dc.b	$3F,$EB,$EA,$D5,$FF,$9E,$65,$D3,$EC,$24,$7B,$1B,$7E,$C6,$BE,$C7
	.dc.b	$FC,$31,$F0,$1F,$E8,$53,$E3,$B6,$7F,$D0,$DF,$98,$FF,$86,$2B,$FD
	.dc.b	$24,$D7,$EE,$35,$FF,$4F,$B5,$FF,$14,$F9,$5F,$12,$7F,$D2,$1F,$9E
	.dc.b	$3E,$CB,$FD,$F5,$E2,$BE,$8F,$FA,$A5,$EF,$0F,$EF,$B7,$BC,$3F,$C6
	.dc.b	$77,$B9,$FF,$1A,$FD,$43,$9D,$E2,$FF,$52,$94,$E6,$30,$D8,$FE,$AE
	.dc.b	$D7,$FB,$21,$C3,$AB,$FD,$E0,$E9,$AE,$A3,$F9,$47,$C3,$74,$9C,$CF
	.dc.b	$F6,$D9,$7C,$7F,$F6,$FE,$46,$7F,$EC,$4D,$3F,$D0,$F7,$EC,$DB,$1E
	.dc.b	$FF,$F7,$5D,$BD,$D3,$EA,$C7,$96,$1F,$E5,$4D,$5F,$81,$3E,$33,$E0
	.dc.b	$46,$1B,$19,$BF,$EB,$0D,$7F,$CF,$D1,$D1,$75,$7C,$A8,$C3,$63,$F8
	.dc.b	$1F,$3E,$FA,$FF,$81,$3F,$3D,$B1,$FC,$0E,$31,$DD,$FA,$48,$C1,$0F
	.dc.b	$E0,$D1,$A3,$EC,$57,$EF,$BC,$BF,$0F,$5F,$F7,$7A,$FD,$3C,$78,$6D
	.dc.b	$F5,$DA,$89,$3A,$3F,$8C,$C6,$CF,$19,$1F,$E3,$0D,$7F,$8C,$35,$FE
	.dc.b	$31,$3D,$9F,$E2,$55,$FE,$12,$7F,$73,$E8,$79,$8F,$89,$FA,$07,$EE
	.dc.b	$E3,$FD,$DC,$D7,$FD,$F4,$D7,$F9,$23,$5F,$F4,$7E,$68,$C4,$7F,$23
	.dc.b	$6C,$67,$BF,$E8,$46,$0D,$E7,$FA,$2E,$4F,$B3,$C1,$F8,$AE,$6F,$F7
	.dc.b	$F4,$7D,$CF,$45,$F3,$34,$FA,$2E,$33,$FF,$63,$46,$0F,$AB,$6F,$AA
	.dc.b	$EA,$49,$FF,$69,$FD,$9D,$7F,$C7,$11,$71,$FF,$1B,$CD,$E6,$FF,$56
	.dc.b	$E7,$FB,$8A,$C5,$FF,$BD,$EB,$CC,$79,$08,$FF,$B6,$5F,$1F,$C8,$7B
	.dc.b	$89,$1F,$EE,$4B,$F7,$1E,$C2,$3F,$EE,$17,$D7,$DD,$F1,$89,$1F,$EC
	.dc.b	$6F,$0F,$F7,$2F,$FC,$01,$8C,$BF,$E5,$37,$F5,$1E,$77,$F3,$2E,$BB
	.dc.b	$5C,$29,$A7,$FB,$3A,$30,$D9,$FD,$87,$9C,$FC,$59,$F1,$B8,$6E,$1F
	.dc.b	$E2,$86,$BF,$C5,$A7,$84,$FF,$1F,$6C,$7F,$BF,$7D,$FF,$BC,$47,$FD
	.dc.b	$59,$F2,$F5,$F2,$EB,$F4,$B9,$3F,$D4,$D1,$FE,$E0,$CF,$FC,$DB,$7F
	.dc.b	$C0,$BC,$16,$1F,$B4,$F3,$A5,$FF,$5A,$7F,$83,$FC,$8B,$E7,$33,$3D
	.dc.b	$8F,$E1,$BD,$DF,$EE,$F7,$81,$DF,$4F,$C1,$D7,$18,$C1,$FE,$03,$C1
	.dc.b	$F6,$FF,$C4,$5B,$1A,$EE,$F2,$CF,$F1,$BE,$AF,$F0,$24,$F2,$6B,$FC
	.dc.b	$F9,$AF,$F9,$DB,$D0,$FE,$6A,$FF,$9F,$5A,$ED,$AE,$F1,$FE,$5F,$18
	.dc.b	$8F,$E6,$8D,$7D,$FE,$7F,$F1,$A2,$3E,$7F,$6D,$18,$8F,$8D,$DB,$FE
	.dc.b	$1E,$8C,$71,$79,$33,$5F,$0B,$F8,$B6,$E3,$7D,$07,$EF,$49,$3F,$EE
	.dc.b	$BF,$82,$AF,$69,$DB,$C9,$AF,$8B,$93,$FD,$D4,$FF,$85,$BC,$9F,$E6
	.dc.b	$64,$7E,$A8,$FA,$47,$C7,$7F,$4B,$8C,$3F,$DC,$DA,$7F,$75,$8D,$B6
	.dc.b	$35,$18,$8F,$F6,$5B,$5F,$EE,$AD,$7F,$DB,$4D,$7F,$AC,$DB,$EB,$BF
	.dc.b	$8F,$12,$3E,$3E,$1F,$8D,$7E,$14,$65,$A6,$B9,$FF,$A8,$F8,$B2,$3F
	.dc.b	$C6,$EF,$F6,$1F,$C5,$DF,$09,$99,$FB,$3D,$5F,$CB,$F9,$BF,$8E,$B6
	.dc.b	$3F,$CB,$95,$C1,$FA,$1F,$55,$FC,$CC,$8F,$E6,$E1,$F2,$9D,$C9,$19
	.dc.b	$7F,$BC,$73,$3F,$51,$7B,$1A,$AF,$D8,$74,$30,$E7,$1C,$4F,$D6,$DF
	.dc.b	$1A,$FE,$DF,$E5,$49,$FF,$2A,$BF,$BC,$F4,$31,$1E,$E8,$B5,$FA,$8E
	.dc.b	$97,$81,$FD,$9E,$BF,$1D,$F5,$5F,$D3,$C8,$FD,$2C,$2B,$A3,$F2,$51
	.dc.b	$FC,$3E,$3F,$7D,$ED,$49,$FD,$F7,$85,$FE,$49,$7F,$23,$D2,$F9,$0F
	.dc.b	$C8,$36,$3F,$1D,$1F,$8C,$23,$82,$1F,$8E,$72,$FF,$90,$B2,$FD,$67
	.dc.b	$DE,$FF,$67,$24,$7F,$45,$F4,$3F,$B5,$78,$BF,$DA,$09,$1F,$E9,$7E
	.dc.b	$DF,$E2,$AF,$FB,$6B,$CC,$FE,$17,$C1,$FD,$0F,$A7,$FA,$39,$FD,$2B
	.dc.b	$F0,$2F,$D7,$D7,$F2,$7F,$39,$F3,$CF,$8D,$FE,$9E,$D9,$FE,$4B,$3A
	.dc.b	$9E,$5C,$72,$7F,$CB,$62,$57,$F5,$60,$48,$CC,$FF,$9A,$5B,$1C,$9A
	.dc.b	$FF,$BD,$27,$93,$C5,$AF,$76,$EF,$FC,$57,$8F,$FC,$20,$49,$FD,$9F
	.dc.b	$B3,$FD,$34,$FF,$20,$F8,$7F,$E4,$78,$FA,$4D,$7C,$E7,$F9,$59,$C3
	.dc.b	$FE,$0A,$6B,$98,$E6,$47,$FD,$95,$A3,$8B,$FC,$BA,$D8,$FE,$56,$D7
	.dc.b	$F9,$9F,$3B,$E2,$3F,$AE,$9F,$E5,$67,$11,$FE,$D5,$6B,$F8,$96,$BE
	.dc.b	$8C,$BF,$4B,$3C,$49,$1C,$5F,$F1,$8F,$A5,$FE,$35,$24,$7F,$18,$FA
	.dc.b	$6F,$EE,$47,$C6,$23,$FE,$E5,$CD,$18,$8F,$92,$D7,$C5,$1A,$7E,$1F
	.dc.b	$2F,$C2,$23,$FD,$33,$17,$CE,$7F,$7C,$3E,$01,$23,$E0,$35,$F8,$0D
	.dc.b	$7E,$04,$72,$7F,$27,$DF,$FE,$0A,$F7,$BF,$BD,$12,$3F,$BD,$7D,$E3
	.dc.b	$E1,$BF,$D1,$CD,$8E,$2D,$7F,$D6,$F4,$8C,$47,$E1,$A3,$13,$FE,$95
	.dc.b	$9C,$CF,$F9,$B9,$AA,$76,$F6,$9F,$EB,$1F,$85,$FF,$97,$E3,$FA,$FB
	.dc.b	$5F,$F0,$E0,$71,$D3,$FE,$08,$7F,$A0,$46,$7C,$9E,$64,$7F,$DA,$D1
	.dc.b	$FE,$BE,$7F,$0D,$EC,$BE,$77,$FD,$62,$D8,$FE,$FE,$D7,$F8,$37,$D1
	.dc.b	$7F,$A9,$5F,$F0,$4C,$67,$0F,$E6,$8D,$E5,$FF,$37,$B6,$7F,$82,$57
	.dc.b	$1A,$FE,$75,$1F,$D5,$F9,$7F,$82,$E5,$FE,$C2,$8C,$1E,$EC,$3F,$CF
	.dc.b	$2E,$23,$46,$FC,$5F,$F0,$17,$73,$E2,$FF,$CC,$74,$FA,$2F,$F2,$DC
	.dc.b	$BE,$33,$F7,$B1,$86,$FB,$BF,$C0,$27,$96,$9F,$C6,$2B,$8E,$1D,$1E
	.dc.b	$18,$66,$F8,$61,$CA,$3F,$8F,$E1,$C9,$F1,$4F,$5B,$F2,$FB,$FC,$A4
	.dc.b	$D3,$BE,$3F,$CA,$1B,$1C,$8E,$B8,$77,$7F,$9C,$B7,$0F,$06,$3C,$0E
	.dc.b	$C4,$C7,$C5,$87,$A6,$FC,$F8,$FC,$87,$F9,$09,$FD,$11,$C3,$C0,$FE
	.dc.b	$7F,$7E,$D7,$6B,$E3,$5E,$FF,$F4,$49,$FE,$9B,$18,$78,$B5,$E2,$D7
	.dc.b	$FA,$5F,$5F,$F8,$ED,$34,$8F,$F8,$FD,$F8,$D8,$FF,$9B,$F7,$B3,$EA
	.dc.b	$9A,$FF,$52,$F7,$7F,$CE,$3B,$7F,$A7,$A3,$FA,$AC,$E9,$4E,$AF,$C7
	.dc.b	$1F,$59,$EC,$D3,$B1,$4E,$D7,$CD,$F4,$99,$3A,$B9,$AF,$05,$77,$9E
	.dc.b	$77,$E5,$8C,$47,$F5,$49,$DD,$71,$8E,$93,$C7,$85,$C7,$FE,$BE,$AF
	.dc.b	$EE,$B1,$8C,$7F,$D1,$4D,$8C,$CE,$EF,$FC,$28,$D7,$F8,$3D,$70,$8C
	.dc.b	$47,$EF,$A3,$FD,$29,$E2,$7F,$B3,$3D,$CB,$1F,$E9,$0B,$FE,$F8,$DA
	.dc.b	$FC,$6D,$6E,$73,$5C,$08,$E0,$FC,$57,$D0,$BE,$67,$FD,$40,$D8,$CC
	.dc.b	$60,$CF,$9B,$23,$8B,$70,$FC,$ED,$3F,$6D,$FE,$A3,$5F,$E1,$A8,$C7
	.dc.b	$83,$FE,$AF,$6C,$66,$77,$5C,$9A,$FF,$82,$27,$6E,$6B,$94,$EA,$3F
	.dc.b	$C4,$7D,$CF,$A2,$C4,$7F,$45,$6C,$66,$BF,$D8,$0D,$C3,$FC,$4D,$F7
	.dc.b	$C7,$E8,$B5,$C9,$82,$18,$25,$FD,$28,$E0,$FF,$94,$E1,$F3,$9F,$86
	.dc.b	$69,$FE,$1C,$70,$7E,$03,$F0,$CE,$33,$FE,$28,$DD,$18,$8F,$F2,$03
	.dc.b	$5F,$F2,$33,$63,$75,$CC,$87,$F9,$0E,$38,$71,$7F,$B4,$6F,$F3,$3E
	.dc.b	$8F,$93,$0F,$F3,$53,$5F,$F1,$13,$5F,$F2,$87,$C2,$FF,$86,$30,$FF
	.dc.b	$27,$36,$3F,$26,$D7,$D4,$D7,$8F,$93,$F9,$E8,$C4,$7F,$97,$DB,$91
	.dc.b	$CE,$3F,$57,$C7,$F9,$44,$C3,$FC,$8A,$FE,$A3,$E8,$DF,$4D,$49,$3F
	.dc.b	$E6,$38,$CB,$63,$38,$FE,$AA,$72,$D8,$F8,$B5,$2B,$FD,$0C,$D9,$F7
	.dc.b	$FD,$5F,$AC,$93,$F1,$1B,$1A,$8C,$11,$F0,$73,$7A,$49,$1E,$77,$D9
	.dc.b	$E4,$C8,$F5,$36,$7D,$07,$C1,$24,$7B,$DF,$E1,$7F,$31,$FE,$E0,$8F
	.dc.b	$F4,$23,$63,$F8,$4B,$5F,$C8,$35,$FD,$95,$FF,$06,$24,$FF,$0A,$6C
	.dc.b	$7E,$FD,$AF,$F0,$47,$F5,$8F,$F6,$2F,$0D,$F4,$4F,$A3,$CC,$FF,$A1
	.dc.b	$A7,$31,$88,$FF,$4F,$C6,$2F,$FD,$4A,$D8,$FC,$BF,$D1,$FE,$68,$93
	.dc.b	$FE,$BB,$6C,$6A,$30,$47,$F6,$DF,$37,$FB,$52,$47,$FA,$41,$B1,$A8
	.dc.b	$C1,$1F,$67,$D1,$BE,$3F,$D1,$0C,$13,$9A,$F4,$B6,$78,$D7,$0A,$DB
	.dc.b	$67,$53,$F1,$5A,$FC,$46,$BF,$E6,$3E,$77,$C6,$7F,$1E,$6E,$1F,$E7
	.dc.b	$A6,$CF,$F9,$C9,$AF,$F9,$C7,$E8,$FF,$86,$76,$1A,$4A,$FE,$52,$D7
	.dc.b	$FD,$0A,$D9,$FE,$13,$EE,$FF,$0C,$24,$FF,$BB,$5B,$1A,$8C,$11,$FE
	.dc.b	$91,$B7,$F5,$42,$47,$F5,$86,$C6,$A3,$04,$7F,$C5,$FE,$33,$E8,$FF
	.dc.b	$35,$1F,$E1,$26,$C6,$7E,$2B,$EB,$B6,$BF,$BE,$F7,$3E,$8F,$F9,$29
	.dc.b	$23,$2D,$C3,$F9,$3B,$5F,$E4,$EF,$F4,$EF,$91,$7C,$CD,$25,$7F,$AB
	.dc.b	$23,$2D,$8C,$D7,$FA,$9A,$72,$D8,$FE,$EF,$F1,$D2,$BF,$BD,$B6,$7F
	.dc.b	$B8,$35,$FE,$EA,$D7,$FC,$71,$F4,$67,$E0,$B6,$7F,$CA,$F1,$FE,$38
	.dc.b	$8C,$3F,$31,$AF,$FB,$96,$1F,$3D,$FD,$74,$8F,$FB,$E1,$B1,$FD,$8D
	.dc.b	$AF,$17,$FA,$8F,$C3,$49,$FA,$2D,$9F,$EA,$6D,$7F,$A9,$B5,$F5,$0F
	.dc.b	$D4,$7F,$46,$DE,$7F,$D9,$B6,$BF,$CD,$53,$FA,$8F,$51,$F4,$4F,$83
	.dc.b	$FD,$CC,$61,$B1,$9A,$F8,$B3,$97,$F9,$5F,$AD,$FC,$B8,$8E,$09,$87
	.dc.b	$F2,$E8,$CB,$63,$F9,$9F,$B5,$FC,$D8,$8E,$0F,$C9,$F7,$7F,$76,$23
	.dc.b	$83,$F9,$DF,$A7,$FC,$F8,$8E,$0F,$CD,$ED,$FD,$F8,$8E,$09,$7F,$E0
	.dc.b	$08,$CB,$63,$2F,$F8,$0A,$32,$D8,$FE,$97,$EB,$FF,$4E,$23,$83,$F5
	.dc.b	$7F,$43,$FE,$0E,$23,$83,$FC,$8D,$F9,$B3,$C7,$FF,$23,$36,$33,$18
	.dc.b	$23,$82,$5F,$F9,$4A,$78,$3F,$EB,$0E,$F7,$CF,$E2,$3F,$0F,$18,$BF
	.dc.b	$F2,$1C,$E6,$7F,$C9,$6D,$5E,$B2,$DE,$BF,$98,$98,$FA,$9B,$5E,$36
	.dc.b	$BF,$E5,$B9,$E3,$3F,$AE,$C5,$F1,$0D,$FF,$CA,$AD,$9D,$36,$FD,$8D
	.dc.b	$7D,$6D,$7F,$CC,$6D,$55,$DA,$BA,$9F,$E7,$3F,$D5,$7C,$73,$FE,$7A
	.dc.b	$6B,$FE,$71,$6B,$FC,$1F,$24,$62,$3E,$2B,$63,$39,$75,$26,$3E,$AE
	.dc.b	$8F,$69,$1C,$1F,$ED,$DF,$91,$7C,$27,$EE,$1B,$19,$8C,$19,$D2,$47
	.dc.b	$16,$E5,$FC,$6B,$E8,$DC,$3F,$ED,$28,$C3,$5F,$F7,$AC,$6B,$C2,$FF
	.dc.b	$77,$36,$39,$72,$7F,$B0,$9B,$EF,$6A,$31,$1C,$9A,$FF,$14,$97,$C5
	.dc.b	$7F,$BE,$61,$18,$7B,$C6,$63,$11,$E9,$F8,$BC,$11,$F2,$BF,$CD,$9F
	.dc.b	$BB,$CB,$F8,$A7,$F1,$8D,$7F,$44,$D7,$F4,$2D,$7E,$E7,$D4,$FF,$B0
	.dc.b	$89,$AF,$FB,$19,$B1,$FE,$BB,$6B,$FE,$C0,$7F,$96,$F2,$FE,$BB,$E3
	.dc.b	$FC,$8D,$9F,$1C,$62,$CD,$24,$68,$87,$FB,$33,$77,$87,$FE,$D8,$6C
	.dc.b	$7F,$B3,$5A,$EE,$34,$47,$4D,$7F,$82,$7D,$03,$E8,$7C,$ED,$8F,$DA
	.dc.b	$35,$DC,$78,$7E,$52,$47,$85,$AF,$9D,$AF,$FB,$66,$D1,$88,$FE,$E1
	.dc.b	$EE,$3C,$B7,$C9,$FD,$B8,$C3,$63,$33,$FE,$14,$8F,$C0,$7B,$EE,$AF
	.dc.b	$F2,$84,$FF,$78,$8C,$79,$3A,$9C,$CF,$F6,$61,$3A,$7F,$C3,$FF,$43
	.dc.b	$DC,$8E,$08,$7F,$B1,$1C,$B4,$C9,$FE,$C6,$72,$D3,$C9,$2F,$96,$F2
	.dc.b	$12,$32,$FB,$F9,$24,$7F,$8C,$9A,$FA,$6F,$E0,$7C,$9F,$B4,$91,$D9
	.dc.b	$F8,$1D,$12,$3C,$2D,$79,$61,$FA,$E7,$FD,$DC,$D7,$FC,$49,$5F,$36
	.dc.b	$78,$47,$F8,$3B,$C9,$7D,$0B,$E2,$BF,$65,$18,$6C,$66,$7F,$C0,$ED
	.dc.b	$7F,$C5,$BF,$92,$3E,$7F,$F8,$23,$63,$F7,$8D,$77,$1F,$85,$F8,$DF
	.dc.b	$C5,$C7,$F8,$DB,$8D,$18,$8F,$BD,$18,$E4,$E0,$48,$E1,$3B,$7F,$C5
	.dc.b	$33,$B7,$33,$F9,$3F,$96,$FD,$6C,$69,$AF,$E6,$9A,$A5,$18,$6F,$4F
	.dc.b	$32,$4F,$A6,$78,$F8,$4F,$A2,$FF,$1B,$B7,$8F,$F8,$F3,$6B,$FC,$AA
	.dc.b	$D7,$FC,$AF,$F6,$0F,$87,$FE,$5C,$D8,$FE,$4C,$D7,$71,$A2,$3A,$6B
	.dc.b	$FE,$71,$F4,$3F,$8D,$FB,$06,$92,$3F,$CF,$4D,$78,$18,$3E,$7F,$C0
	.dc.b	$7D,$92,$38,$21,$F6,$37,$FF,$5C,$B8,$3E,$AB,$87,$FC,$EB,$F3,$7F
	.dc.b	$85,$7F,$78,$D7,$FC,$C6,$D7,$FC,$FC,$D5,$8B,$A7,$E3,$92,$3E,$3B
	.dc.b	$5F,$9A,$D7,$F8,$54,$71,$7F,$17,$BF,$E5,$48,$E0,$87,$B5,$CB,$4F
	.dc.b	$F1,$17,$CE,$3D,$DF,$E9,$2D,$8C,$C6,$08,$E0,$87,$F8,$FE,$38,$37
	.dc.b	$E7,$B9,$91,$EE,$FF,$69,$F8,$57,$CF,$3D,$1C,$36,$BF,$D4,$73,$A9
	.dc.b	$C9,$23,$33,$B7,$F9,$74,$ED,$CC,$FE,$AF,$C8,$91,$FA,$16,$BF,$B8
	.dc.b	$E6,$FE,$CA,$E4,$86,$63,$F8,$11,$1F,$DD,$B5,$FC,$53,$5F,$C4,$B5
	.dc.b	$FC,$2F,$A5,$FE,$13,$DF,$FC,$B7,$1F,$E0,$B6,$BF,$8E,$8F,$D5,$FD
	.dc.b	$1C,$8F,$D5,$B5,$FD,$8B,$5F,$C2,$B5,$FE,$57,$E1,$7F,$5A,$7F,$98
	.dc.b	$F8,$DF,$D5,$48,$FF,$46,$E9,$FE,$B0,$47,$07,$F5,$DF,$9F,$FD,$13
	.dc.b	$FE,$C0,$F1,$11,$88,$FE,$72,$D7,$FA,$FB,$63,$79,$FF,$6A,$8F,$9F
	.dc.b	$3F,$D3,$5A,$F5,$6B,$FD,$13,$E5,$DF,$07,$FD,$C9,$B1,$FD,$C5,$AF
	.dc.b	$F5,$76,$BE,$76,$BC,$E3,$91,$1F,$E6,$BE,$5F,$F3,$F2,$38,$3E,$17
	.dc.b	$C7,$FA,$48,$E0,$87,$AD,$CB,$4F,$E6,$1F,$03,$FC,$F4,$8E,$0F,$57
	.dc.b	$97,$ED,$23,$82,$1E,$D7,$2D,$3F,$A6,$FC,$97,$F5,$72,$38,$3E,$D7
	.dc.b	$85,$F4,$48,$E0,$87,$D3,$72,$D3,$FA,$47,$CA,$FF,$55,$23,$83,$FC
	.dc.b	$45,$F5,$BF,$E5,$D7,$FC,$3E,$DD,$CD,$26,$7F,$E2,$C6,$CF,$28,$FF
	.dc.b	$2D,$FC,$FF,$FB,$56,$3F,$D7,$0D,$7F,$D9,$CD,$7F,$D9,$AD,$7F,$D5
	.dc.b	$1F,$33,$FE,$81,$F7,$8D,$24,$7F,$9E,$E3,$2D,$8C,$BF,$E3,$93,$FD
	.dc.b	$A6,$FF,$A2,$D7,$CF,$BE,$7B,$FD,$1C,$D8,$FF,$43,$35,$E4,$D7,$FC
	.dc.b	$88,$BE,$93,$89,$23,$8B,$FE,$50,$FA,$07,$C6,$D2,$4F,$F9,$46,$32
	.dc.b	$D8,$CB,$FE,$39,$8C,$B6,$3F,$CC,$0B,$E8,$29,$1F,$F0,$F1,$92,$B9
	.dc.b	$7F,$C3,$E6,$4A,$FF,$AB,$F9,$9F,$3E,$F0,$94,$F8,$AF,$F1,$A5,$3E
	.dc.b	$2F,$33,$9F,$5E,$CE,$63,$8D,$70,$C3,$A4,$6C,$EB,$3E,$0C,$37,$7C
	.dc.b	$B0,$EB,$87,$81,$B1,$DB,$0E,$73,$D6,$B8,$E1,$D2,$F6,$D3,$23,$0F
	.dc.b	$03,$E3,$31,$7D,$30,$F1,$47,$09,$F6,$3B,$78,$E1,$E1,$8C,$5F,$81
	.dc.b	$EC,$D9,$DD,$7A,$70,$FF,$6A,$47,$8A,$BF,$D2,$0D,$74,$7E,$6D,$7C
	.dc.b	$63,$D9,$E2,$D9,$D3,$5E,$6D,$7F,$49,$E7,$FE,$8C,$8F,$FB,$BD,$B9
	.dc.b	$7F,$9C,$61,$D2,$F0,$7E,$4C,$7E,$12,$AC,$E6,$3E,$51,$92,$B9,$AE
	.dc.b	$91,$FE,$72,$33,$5D,$1C,$47,$86,$78,$67,$E8,$9D,$9D,$8D,$BF,$EB
	.dc.b	$78,$DC,$7F,$AE,$6F,$CB,$7D,$5B,$1C,$5B,$7F,$E9,$1F,$55,$F0,$DF
	.dc.b	$E9,$7B,$FF,$44,$36,$FF,$D0,$ED,$7F,$D0,$6D,$7F,$DA,$ED,$7F,$CF
	.dc.b	$CD,$9E,$B8,$FC,$79,$E2,$D7,$FD,$56,$FF,$9A,$6B,$FD,$AC,$D8,$FF
	.dc.b	$5E,$35,$FE,$F8,$D7,$8F,$A7,$CD,$FF,$6C,$C6,$5B,$19,$EC,$FF,$5A
	.dc.b	$3B,$87,$4B,$E6,$B0,$DC,$3F,$98,$46,$0C,$F9,$32,$3F,$9A,$36,$BF
	.dc.b	$DD,$55,$FC,$EB,$BF,$C8,$DA,$FE,$6C,$D9,$DF,$62,$B6,$DA,$DB,$67
	.dc.b	$73,$A3,$E0,$7C,$1B,$E7,$F5,$3F,$C5,$A3,$13,$A9,$CC,$FF,$1E,$9C
	.dc.b	$DF,$F4,$9D,$1F,$3F,$FA,$18,$C3,$6B,$FD,$DC,$D7,$FC,$5C,$D7,$FD
	.dc.b	$CA,$D5,$3D,$27,$E9,$4E,$63,$11,$FE,$D7,$1F,$2F,$F6,$F8,$3F,$64
	.dc.b	$D8,$C9,$B7,$EE,$7C,$FF,$32,$46,$08,$7B,$5C,$B4,$ED,$F0,$9D,$31
	.dc.b	$FF,$72,$B5,$FF,$70,$B5,$FE,$8B,$68,$C4,$7B,$BE,$7D,$E3,$FE,$3D
	.dc.b	$B7,$F3,$5B,$3F,$D2,$8E,$13,$B2,$4E,$DB,$1F,$EF,$88,$FF,$6D,$1F
	.dc.b	$1F,$DD,$F8,$84,$7E,$D7,$CF,$BC,$CF,$D0,$36,$33,$18,$23,$82,$57
	.dc.b	$97,$55,$71,$8F,$8B,$7F,$20,$95,$FE,$11,$CF,$E2,$47,$C1,$F9,$E9
	.dc.b	$3F,$05,$B5,$D7,$34,$62,$3F,$CF,$11,$89,$E0,$7B,$39,$BD,$64,$70
	.dc.b	$7C,$0F,$94,$F7,$91,$C1,$F1,$2B,$E1,$91,$C1,$FA,$1F,$9F,$FF,$67
	.dc.b	$37,$DC,$FE,$11,$18,$9F,$88,$D7,$F7,$F8,$FF,$08,$25,$7F,$8D,$DB
	.dc.b	$1F,$C0,$1A,$F0,$8F,$F6,$4F,$CF,$BE,$6F,$F8,$D7,$8A,$8C,$4E,$A7
	.dc.b	$31,$8A,$FC,$BF,$CF,$FE,$68,$95,$F4,$F9,$11,$89,$DC,$E6,$31,$1F
	.dc.b	$E1,$CE,$0F,$EC,$1F,$2F,$FA,$F2,$57,$ED,$5B,$3F,$93,$6B,$C6,$11
	.dc.b	$EA,$F9,$F7,$C7,$52,$4F,$FB,$A2,$32,$D8,$CD,$7F,$B9,$A7,$2D,$8F
	.dc.b	$E3,$FD,$2F,$86,$A4,$BF,$F3,$17,$67,$F0,$A8,$FE,$4B,$F3,$FF,$C4
	.dc.b	$E0,$D2,$4E,$88,$72,$72,$FC,$BF,$9F,$FE,$0E,$47,$04,$C7,$F8,$4C
	.dc.b	$65,$B1,$FE,$79,$E0,$FE,$9A,$47,$04,$8F,$E8,$B1,$96,$C7,$F9,$F7
	.dc.b	$C9,$7C,$FF,$E4,$67,$FC,$AF,$18,$E6,$FE,$C4,$DA,$E1,$5B,$A7,$07
	.dc.b	$F1,$9F,$6F,$F9,$09,$2B,$F1,$0D,$9F,$E4,$6D,$7F,$91,$BF,$D4,$3E
	.dc.b	$2B,$FC,$FF,$3F,$41,$B3,$FC,$77,$54,$62,$3F,$AA,$35,$FE,$A4,$D5
	.dc.b	$1C,$3D,$3E,$C4,$AF,$F5,$17,$C4,$7F,$76,$23,$82,$57,$F6,$E8,$CB
	.dc.b	$63,$E5,$7C,$FF,$F9,$F9,$B1,$F2,$23,$0F,$F8,$B5,$AF,$E7,$7E,$67
	.dc.b	$FA,$91,$27,$FA,$C3,$63,$F9,$B3,$5E,$11,$FB,$3F,$83,$FE,$9E,$47
	.dc.b	$04,$8F,$EA,$31,$96,$C7,$FA,$F5,$7C,$FF,$E9,$1F,$E9,$4D,$7F,$D2
	.dc.b	$AD,$7F,$D3,$6D,$F5,$1F,$3E,$F8,$2F,$F1,$44,$61,$B1,$9B,$FF,$66
	.dc.b	$4E,$5F,$E4,$DE,$B7,$F2,$42,$38,$26,$1F,$C9,$A3,$2D,$8F,$E5,$DE
	.dc.b	$C7,$F2,$C2,$38,$3F,$8C,$76,$FF,$74,$23,$83,$F9,$B7,$A7,$FC,$D0
	.dc.b	$8E,$0F,$E4,$1A,$FF,$7C,$23,$82,$5F,$F7,$E8,$CB,$63,$2F,$F8,$02
	.dc.b	$32,$D8,$FE,$8D,$EB,$7F,$44,$23,$83,$F9,$87,$CE,$7F,$83,$08,$E0
	.dc.b	$FF,$20,$FC,$FB,$C8,$FF,$22,$B6,$33,$18,$23,$82,$5F,$F9,$3A,$78
	.dc.b	$3F,$AF,$F9,$AA,$FF,$90,$5A,$FF,$8F,$DA,$FE,$B9,$BE,$B2,$5E,$BF
	.dc.b	$90,$97,$E8,$6D,$F8,$5A,$FF,$B9,$2B,$94,$7F,$95,$3E,$0F,$FA,$1D
	.dc.b	$7F,$34,$6D,$7F,$05,$6B,$FE,$57,$6B,$FE,$55,$6B,$FC,$11,$BD,$26
	.dc.b	$9A,$B7,$CF,$FA,$AF,$FC,$DE,$DB,$FF,$3A,$B5,$F4,$35,$FF,$80,$33
	.dc.b	$CB,$0F,$F3,$CC,$E6,$7F,$D0,$ED,$9D,$57,$B7,$9F,$E0,$11,$C1,$FB
	.dc.b	$9F,$8D,$7C,$FF,$FA,$45,$B1,$98,$C1,$9C,$A4,$71,$6E,$3F,$E4,$91
	.dc.b	$F3,$EE,$0F,$F6,$CC,$66,$31,$E1,$7F,$09,$6C,$66,$76,$FF,$94,$E1
	.dc.b	$C7,$FE,$C4,$8E,$4D,$74,$D4,$48,$E3,$EF,$77,$F9,$AF,$E3,$13,$E6
	.dc.b	$6C,$9A,$63,$E7,$E4,$F9,$DA,$FE,$77,$2F,$E4,$5C,$4E,$4F,$C1,$4F
	.dc.b	$F8,$01,$AE,$CF,$F5,$7F,$CF,$FF,$AF,$09,$B7,$F2,$BE,$24,$62,$77
	.dc.b	$39,$8C,$47,$EB,$7E,$3B,$C3,$66,$92,$3C,$4D,$78,$18,$3F,$D8,$38
	.dc.b	$BC,$2F,$13,$63,$31,$82,$38,$21,$FE,$CE,$8E,$0F,$F6,$FF,$9F,$7C
	.dc.b	$FF,$C6,$6C,$7F,$89,$1A,$EE,$3C,$3F,$37,$23,$CA,$D7,$B3,$5F,$DF
	.dc.b	$52,$31,$1F,$EA,$4F,$51,$F2,$5F,$D5,$FC,$BF,$F5,$33,$63,$33,$FE
	.dc.b	$55,$8F,$EB,$9E,$F3,$AF,$FC,$8F,$3F,$ED,$36,$BF,$E9,$36,$BF,$90
	.dc.b	$3F,$D9,$AE,$1F,$17,$C8,$F7,$23,$FE,$63,$6B,$FE,$BE,$6B,$FE,$BE
	.dc.b	$6B,$C8,$E0,$43,$C5,$F3,$FE,$89,$34,$91,$FE,$71,$E2,$F8,$04,$70
	.dc.b	$7B,$7E,$6F,$91,$1C,$1F,$07,$9F,$B1,$1C,$10,$F8,$8E,$5A,$7F,$78
	.dc.b	$F9,$F7,$C5,$7F,$82,$7C,$A4,$62,$75,$39,$8C,$47,$F0,$DF,$9F,$7C
	.dc.b	$FF,$EE,$9B,$1F,$E1,$36,$BB,$8F,$C0,$7C,$6F,$E1,$63,$FD,$88,$D7
	.dc.b	$FD,$C1,$C4,$8C,$47,$FA,$45,$AF,$FB,$B9,$AF,$F9,$61,$B1,$C7,$8F
	.dc.b	$B9,$F8,$FF,$9A,$FD,$1D,$7F,$A6,$DA,$FF,$BE,$1A,$FF,$98,$9B,$26
	.dc.b	$93,$A3,$A1,$FE,$31,$F0,$DF,$3E,$E1,$C3,$70,$FF,$3B,$35,$FF,$7F
	.dc.b	$4F,$09,$FE,$47,$F3,$EF,$88,$A4,$AF,$F1,$F4,$65,$B1,$97,$FA,$04
	.dc.b	$65,$B1,$FC,$93,$D7,$79,$F4,$93,$FE,$75,$6B,$FE,$75,$8D,$B9,$21
	.dc.b	$F5,$B6,$FF,$57,$3F,$6B,$DE,$FB,$A4,$70,$43,$83,$97,$FA,$B7,$CE
	.dc.b	$7E,$19,$FD,$C0,$70,$FE,$ED,$AF,$F4,$C6,$AC,$5C,$FF,$28,$91,$F2
	.dc.b	$1A,$FC,$66,$BE,$08,$E2,$FE,$53,$6F,$CB,$91,$C1,$0F,$84,$E5,$A7
	.dc.b	$FB,$3B,$E7,$DE,$EF,$F5,$46,$C6,$63,$04,$70,$43,$FD,$4D,$1C,$1B
	.dc.b	$F3,$FC,$C8,$F7,$7F,$D0,$5F,$0E,$F9,$F7,$A3,$86,$D7,$F8,$16,$75
	.dc.b	$39,$24,$66,$76,$FD,$49,$DB,$99,$FD,$67,$8F,$23,$FC,$21,$CF,$FA
	.dc.b	$78,$C4,$7E,$F3,$9F,$FB,$23,$A6,$B9,$8D,$BF,$BF,$7F,$83,$91,$FE
	.dc.b	$0E,$D7,$F2,$4D,$78,$1F,$EA,$3D,$FF,$CA,$ED,$F4,$FF,$C1,$F1,$88
	.dc.b	$FC,$8C,$7E,$C3,$E7,$E4,$7E,$B9,$AF,$EB,$1A,$FE,$1D,$AF,$F8,$7B
	.dc.b	$C4,$FE,$AD,$E2,$FF,$41,$24,$7F,$57,$7F,$A2,$74,$C8,$FE,$B4,$D7
	.dc.b	$F4,$32,$F9,$FF,$D6,$C6,$1A,$FF,$5D,$8D,$73,$7F,$66,$9C,$CF,$8D
	.dc.b	$F9,$AE,$DA,$E7,$3E,$07,$F6,$EF,$97,$7C,$15,$24,$FF,$92,$F0,$4E
	.dc.b	$A3,$0D,$7D,$2D,$7C,$CD,$7E,$33,$5F,$E7,$1E,$57,$F4,$42,$3F,$D1
	.dc.b	$1F,$67,$C8,$C8,$F1,$62,$8C,$47,$C5,$6B,$EF,$6B,$FC,$F7,$E0,$3F
	.dc.b	$A1,$11,$FC,$93,$F0,$7C,$B9,$1E,$C6,$BE,$C6,$BE,$C6,$BF,$D4,$FE
	.dc.b	$4B,$FA,$E9,$1F,$EB,$AF,$D5,$F0,$A4,$7A,$9A,$FD,$D6,$BF,$5D,$AF
	.dc.b	$F5,$5F,$96,$FE,$B4,$47,$F5,$4F,$F8,$9B,$E7,$FF,$CB,$B1,$F8,$56
	.dc.b	$BF,$95,$6A,$D5,$18,$6E,$9C,$C9,$3F,$81,$9E,$3E,$AB,$E7,$FF,$DA
	.dc.b	$D3,$FE,$8F,$8C,$7B,$FF,$E7,$29,$CC,$FF,$3C,$9C,$BF,$E9,$5F,$5D
	.dc.b	$F3,$3F,$E8,$26,$C7,$F9,$C6,$30,$47,$04,$39,$3F,$E8,$A5,$F3,$CF
	.dc.b	$9F,$FF,$46,$B6,$3F,$D4,$86,$A3,$26,$BE,$7E,$46,$A5,$18,$8F,$DA
	.dc.b	$FC,$FB,$E3,$3F,$CA,$AD,$8F,$D3,$35,$DC,$68,$8E,$9A,$FF,$88,$07
	.dc.b	$CF,$FF,$99,$5A,$7F,$2C,$29,$B7,$44,$74,$D7,$FD,$85,$D0,$F9,$F7
	.dc.b	$8B,$FE,$B4,$F8,$8F,$F1,$DC,$BE,$2B,$FC,$D1,$18,$6F,$B1,$FE,$6B
	.dc.b	$9E,$55,$C6,$31,$7D,$30,$61,$C5,$D1,$FE,$C2,$BE,$CE,$0F,$E1,$73
	.dc.b	$C6,$7C,$6D,$AF,$06,$1E,$2A,$E4,$D7,$8E,$1C,$F0,$F2,$C7,$8A,$B8
	.dc.b	$B9,$AE,$13,$AC,$3B,$DE,$6A,$1C,$4F,$46,$1E,$36,$C7,$E2,$AB,$BC
	.dc.b	$7F,$86,$1A,$FC,$48,$D7,$C5,$BD,$AA,$49,$FA,$B1,$97,$FD,$C7,$E9
	.dc.b	$7E,$6C,$8E,$09,$F4,$1F,$E7,$25,$D9,$FE,$4D,$8C,$47,$F9,$A4,$3B
	.dc.b	$8D,$6B,$FE,$2E,$9D,$4E,$7C,$5F,$F7,$BB,$CA,$ED,$7F,$95,$A6,$D7
	.dc.b	$F8,$DA,$7E,$D3,$E2,$8A,$F8,$75,$E8,$FC,$B3,$FD,$69,$3C,$91,$C4
	.dc.b	$F3,$D7,$96,$35,$19,$BF,$23,$A3,$FD,$7D,$87,$B1,$F2,$19,$3A,$B9
	.dc.b	$F2,$DF,$0C,$FA,$DF,$F3,$A6,$DF,$E7,$59,$CC,$63,$4F,$2C,$70,$FB
	.dc.b	$5C,$62,$8D,$CF,$AA,$F9,$61,$D6,$30,$72,$71,$5F,$B2,$8C,$5F,$4B
	.dc.b	$E7,$39,$3F,$D2,$51,$9F,$38,$F8,$BC,$1F,$F5,$CF,$1E,$62,$61,$E0
	.dc.b	$BE,$2D,$AD,$B9,$C3,$FD,$39,$5E,$CA,$76,$3B,$FF,$D2,$D1,$FE,$EB
	.dc.b	$8C,$5F,$F2,$A6,$C7,$F3,$B9,$DB,$F7,$23,$FC,$A3,$3E,$38,$EB,$E7
	.dc.b	$F7,$9E,$3A,$FF,$A8,$DD,$CF,$C1,$3C,$7E,$77,$46,$DF,$86,$31,$26
	.dc.b	$92,$34,$49,$FE,$A1,$C7,$E1,$3C,$53,$67,$E2,$0F,$7F,$D7,$A7,$FA
	.dc.b	$ED,$B1,$93,$FC,$BE,$5B,$FF,$6C,$B5,$FC,$E9,$F8,$77,$0F,$E3,$1A
	.dc.b	$E6,$3B,$F2,$3E,$4F,$F0,$8D,$AF,$2C,$E2,$7F,$11,$0F,$96,$F2,$91
	.dc.b	$F2,$B6,$FC,$AD,$7C,$B8,$7F,$B5,$E3,$C3,$F1,$BE,$03,$AB,$6B,$F1
	.dc.b	$9A,$BE,$EB,$FD,$3B,$1F,$E8,$B6,$E1,$FE,$87,$6B,$F0,$A7,$87,$B1
	.dc.b	$F8,$66,$D7,$18,$C5,$70,$AF,$D6,$35,$78,$CF,$63,$F7,$77,$F9,$F8
	.dc.b	$C7,$47,$ED,$B1,$EB,$7F,$AB,$9C,$D7,$DE,$F7,$BE,$D9,$1F,$B6,$D9
	.dc.b	$FB,$6D,$7E,$3C,$7F,$BD,$4F,$F4,$FF,$03,$CF,$FA,$E7,$D4,$AB,$3F
	.dc.b	$B1,$F9,$F7,$9D,$FA,$6E,$A4,$62,$7F,$CD,$CD,$C7,$FD,$06,$D7,$FD
	.dc.b	$00,$D7,$FD,$20,$D7,$FC,$78,$D9,$EB,$5F,$8C,$3F,$16,$7F,$09,$8E
	.dc.b	$D9,$FF,$21,$CA,$E1,$D3,$E7,$74,$6F,$D2,$B1,$FC,$2F,$E8,$1F,$41
	.dc.b	$EB,$6C,$7F,$A1,$DB,$A7,$FA,$71,$AF,$E4,$30,$FF,$3D,$4F,$F1,$36
	.dc.b	$C7,$F2,$1D,$3E,$2F,$C2,$FE,$8B,$1F,$B9,$F3,$FE,$FD,$0D,$26,$DF
	.dc.b	$8D,$6E,$5F,$ED,$F6,$E3,$FE,$EC,$F1,$BF,$9D,$F1,$7E,$2D,$EB,$78
	.dc.b	$E4,$FF,$0D,$E1,$E5,$E0,$FD,$99,$FB,$0F,$9F,$B1,$92,$63,$FC,$E7
	.dc.b	$A9,$09,$DC,$E7,$81,$BF,$EF,$8F,$DD,$FC,$E5,$BF,$89,$16,$BA,$42
	.dc.b	$71,$BF,$69,$FD,$2F,$D9,$BE,$67,$DA,$DB,$FE,$65,$E8,$3E,$7F,$35
	.dc.b	$27,$F5,$55,$BB,$FE,$65,$7B,$9C,$D7,$F1,$4F,$1E,$46,$0F,$B4,$FF
	.dc.b	$18,$F2,$A4,$7F,$2C,$F7,$6E,$1F,$06,$6E,$7F,$DC,$6B,$FD,$A2,$7E
	.dc.b	$0E,$3F,$90,$F8,$D2,$3F,$9A,$FC,$7D,$BF,$95,$36,$F9,$4D,$AF,$EE
	.dc.b	$9D,$1F,$CA,$6F,$F9,$99,$FE,$62,$70,$FF,$15,$6B,$C3,$C5,$FC,$51
	.dc.b	$27,$F1,$4D,$8E,$7C,$8E,$3F,$EA,$55,$F9,$BF,$1F,$F2,$F7,$FC,$E9
	.dc.b	$FE,$77,$DD,$27,$FA,$7F,$05,$C7,$F0,$93,$74,$F8,$93,$6B,$FC,$27
	.dc.b	$8F,$F6,$69,$FD,$C7,$93,$F9,$AF,$9F,$FC,$7D,$7E,$AE,$3F,$A4,$91
	.dc.b	$FE,$4D,$1A,$C7,$F2,$53,$FE,$1C,$F3,$9F,$3F,$FC,$BD,$BE,$47,$F7
	.dc.b	$C6,$E1,$FE,$49,$6D,$FC,$C6,$BF,$E3,$4A,$E4,$4A,$FE,$BB,$E6,$FF
	.dc.b	$5C,$3C,$11,$FD,$23,$2F,$E2,$51,$FD,$82,$E6,$1F,$CF,$7E,$2F,$FD
	.dc.b	$60,$FF,$07,$8C,$E3,$FA,$B7,$F7,$73,$A8,$C4,$FF,$0D,$7F,$B5,$9F
	.dc.b	$DA,$3E,$66,$C6,$49,$1F,$AE,$8C,$BF,$E3,$3F,$51,$F3,$AF,$07,$FC
	.dc.b	$3F,$18,$6C,$66,$FF,$CD,$D3,$97,$FC,$E1,$E3,$BE,$21,$F5,$FF,$E3
	.dc.b	$B6,$CF,$F8,$E6,$B8,$63,$FC,$8E,$72,$FF,$0D,$FA,$87,$C6,$7F,$92
	.dc.b	$5B,$7F,$E3,$FA,$DF,$07,$F9,$76,$73,$7C,$6F,$81,$E1,$C6,$E1,$FE
	.dc.b	$22,$F9,$DF,$F0,$F9,$1F,$F0,$FB,$FE,$24,$BF,$B1,$1F,$E6,$66,$DF
	.dc.b	$F8,$7E,$7C,$9C,$7F,$E1,$6F,$8B,$FF,$14,$92,$FF,$C5,$2F,$3F,$9B
	.dc.b	$F5,$1B,$8F,$F2,$B6,$08,$CC,$61,$D6,$1F,$0D,$CB,$63,$FC,$77,$58
	.dc.b	$CF,$F8,$77,$C4,$FF,$8E,$89,$3F,$E3,$A7,$D9,$F6,$1F,$2C,$FF,$16
	.dc.b	$B6,$3F,$CD,$9C,$AF,$C1,$1A,$FE,$95,$B3,$FA,$56,$E5,$FA,$56,$BF
	.dc.b	$E9,$86,$BF,$E9,$66,$BF,$EB,$C6,$D7,$7A,$FA,$1B,$7D,$F3,$F7,$13
	.dc.b	$DB,$2F,$BD,$1F,$D7,$3E,$77,$FC,$EB,$87,$F9,$4B,$D1,$46,$23,$FD
	.dc.b	$63,$18,$E9,$FF,$3C,$B7,$3F,$F6,$33,$67,$F9,$4B,$5E,$95,$CD,$FF
	.dc.b	$1A,$B7,$C2,$70,$BE,$5B,$F2,$C7,$97,$1B,$8F,$DE,$8B,$87,$F8,$DF
	.dc.b	$0F,$F4,$2C,$FE,$36,$7F,$BB,$FC,$67,$F5,$5A,$F6,$35,$F5,$B5,$FF
	.dc.b	$23,$B7,$23,$49,$E0,$FC,$07,$86,$7E,$8F,$9C,$FD,$61,$E0,$9B,$8F
	.dc.b	$F9,$3F,$CE,$FF,$0D,$E3,$F2,$9A,$FF,$95,$9B,$8F,$F6,$FF,$13,$E6
	.dc.b	$FD,$7F,$EE,$AB,$EA,$35,$FA,$6D,$7F,$CB,$ED,$5F,$41,$F9,$0E,$5F
	.dc.b	$F7,$D4,$63,$95,$18,$9D,$4F,$FB,$FD,$AF,$F6,$25,$E0,$FF,$B9,$23
	.dc.b	$4D,$74,$D7,$FD,$37,$1C,$3D,$CF,$E8,$5F,$5C,$F8,$BF,$1E,$08,$C5
	.dc.b	$EA,$73,$18,$8F,$C5,$FE,$83,$FC,$69,$FC,$7B,$5C,$F8,$1E,$ED,$93
	.dc.b	$99,$FF,$4C,$F4,$3D,$3F,$F4,$77,$17,$F1,$36,$CF,$FB,$DE,$74,$FF
	.dc.b	$A1,$FA,$1F,$29,$AC,$7F,$4A,$F7,$8C,$CF,$E8,$9A,$FE,$89,$AF,$E8
	.dc.b	$A3,$81,$1E,$0F,$D6,$E7,$FD,$1F,$D7,$F9,$1B,$3F,$C7,$A3,$C5,$F4
	.dc.b	$4E,$9D,$64,$9A,$4A,$FF,$50,$56,$DB,$5F,$1E,$31,$1F,$B8,$6B,$FC
	.dc.b	$0D,$AF,$F0,$26,$BF,$EA,$E6,$E4,$69,$AC,$3F,$7B,$F6,$5E,$62,$57
	.dc.b	$F2,$6F,$01,$E1,$FD,$2E,$4F,$E5,$0F,$E1,$A3,$33,$FC,$9E,$30,$47
	.dc.b	$04,$3F,$95,$3F,$2F,$47,$D9,$7C,$62,$4F,$FB,$EB,$87,$E7,$1F,$E6
	.dc.b	$57,$FD,$EA,$FD,$33,$0F,$FA,$F1,$E9,$F1,$BF,$40,$90,$3F,$CF,$9F
	.dc.b	$18,$FA,$0C,$C7,$F7,$39,$BE,$2F,$F9,$CF,$65,$EE,$BF,$AD,$D6,$B0
	.dc.b	$FE,$D0,$FE,$0E,$75,$3D,$B0,$FC,$44,$62,$3F,$DE,$ED,$7C,$97,$DA
	.dc.b	$7E,$27,$C6,$FE,$20,$93,$E9,$C7,$C9,$6F,$9A,$F0,$1D,$89,$3F,$E6
	.dc.b	$61,$87,$AA,$3E,$87,$C8,$FE,$50,$93,$FE,$0F,$ED,$AF,$F8,$56,$2E
	.dc.b	$1F,$E9,$5F,$3F,$FC,$41,$5F,$B4,$6D,$71,$21,$FE,$2F,$FA,$2F,$F5
	.dc.b	$2B,$FE,$23,$6B,$FE,$7A,$F4,$11,$88,$FA,$C2,$33,$F0,$5F,$C9,$9B
	.dc.b	$1E,$DC,$FF,$AA,$FC,$4F,$EB,$89,$1F,$1F,$87,$E6,$EC,$F9,$EC,$CF
	.dc.b	$F8,$C6,$AE,$BF,$32,$7F,$6F,$F4,$32,$BF,$C3,$2F,$F0,$DF,$5C,$E6
	.dc.b	$3F,$C7,$4D,$AF,$F5,$D7,$62,$B7,$5F,$E2,$2F,$17,$F8,$19,$1F,$31
	.dc.b	$F9,$5F,$57,$E2,$DF,$E7,$DF,$89,$1F,$11,$AE,$F7,$FE,$92,$FF,$94
	.dc.b	$DB,$1A,$8C,$1F,$D3,$9C,$1C,$B6,$FE,$FB,$F3,$1F,$C5,$89,$3E,$4C
	.dc.b	$9F,$10,$F5,$F3,$3F,$E5,$9A,$B9,$7C,$0A,$D4,$FF,$90,$3E,$5B,$F9
	.dc.b	$09,$1F,$7E,$3F,$AC,$F1,$71,$1F,$E6,$39,$B7,$F9,$E9,$FE,$79,$D6
	.dc.b	$FA,$FA,$4B,$F9,$27,$F9,$CF,$E0,$3F,$A3,$B4,$F8,$0F,$C4,$23,$82
	.dc.b	$1F,$A7,$D9,$1A,$8F,$F6,$2F,$80,$F8,$CF,$E7,$64,$8F,$5F,$3F,$F8
	.dc.b	$51,$FF,$2F,$C6,$5F,$F3,$F3,$FC,$17,$D5,$FE,$87,$DB,$27,$FC,$CA
	.dc.b	$FF,$09,$F9,$DC,$47,$FA,$16,$2E,$DF,$C0,$27,$F9,$1F,$AB,$2B,$E5
	.dc.b	$F6,$23,$FB,$17,$D7,$FA,$A7,$ED,$35,$5F,$E5,$A3,$0D,$D7,$E4,$FD
	.dc.b	$DB,$E2,$F9,$C4,$BE,$6E,$A3,$F9,$AF,$D0,$C9,$F1,$BF,$C8,$FE,$1F
	.dc.b	$11,$E0,$CF,$F8,$F8,$9C,$E3,$FC,$72,$BF,$9F,$FD,$94,$BF,$6B,$FC
	.dc.b	$B7,$E6,$31,$1C,$CF,$E3,$B8,$27,$37,$FC,$9A,$BF,$C0,$9F,$41,$FA
	.dc.b	$CA,$FF,$33,$B6,$DD,$C8,$C5,$FF,$A1,$7E,$53,$FC,$22,$DA,$FE,$58
	.dc.b	$D7,$F8,$B9,$F3,$3C,$F4,$FF,$8F,$27,$53,$FC,$71,$AF,$F3,$B9,$DB
	.dc.b	$D1,$AF,$48,$DB,$EE,$FB,$2F,$EC,$64,$7F,$84,$F9,$4F,$9E,$7C,$A7
	.dc.b	$F3,$5F,$A3,$FF,$0E,$D7,$F6,$1A,$46,$21,$18,$F0,$3F,$CA,$6D,$8F
	.dc.b	$F6,$FD,$BE,$53,$FA,$03,$63,$C2,$D7,$EA,$4F,$8F,$E4,$BF,$BC,$FD
	.dc.b	$44,$9F,$CB,$BC,$BE,$1A,$47,$67,$C3,$66,$92,$34,$43,$E8,$DF,$C1
	.dc.b	$8F,$F0,$27,$D5,$7F,$82,$09,$3F,$B6,$F9,$CF,$EE,$C7,$F7,$42,$3F
	.dc.b	$D2,$FC,$24,$62,$3E,$EF,$91,$FE,$02,$8C,$47,$F7,$A7,$FC,$B9,$C8
	.dc.b	$F9,$0F,$C7,$92,$7F,$9E,$F1,$BF,$0B,$FF,$3B,$C6,$1A,$FD,$F6,$BF
	.dc.b	$87,$6B,$FE,$01,$6B,$FB,$A6,$C7,$5F,$87,$FC,$84,$60,$DE,$2E,$97
	.dc.b	$D4,$FF,$8B,$09,$87,$F1,$AF,$9C,$FF,$0D,$1F,$E1,$82,$3F,$15,$AF
	.dc.b	$C4,$D5,$18,$8F,$F0,$EB,$FE,$35,$F2,$FF,$C7,$24,$8F,$E4,$FD,$7F
	.dc.b	$E2,$73,$FC,$4A,$47,$E9,$35,$EC,$D7,$FC,$52,$FF,$60,$FA,$57,$9D
	.dc.b	$FD,$01,$B3,$FD,$11,$AF,$F0,$70,$BD,$0E,$07,$32,$47,$FA,$D2,$34
	.dc.b	$D8,$CB,$FE,$2D,$8C,$B6,$3C,$9E,$AF,$8C,$91,$E3,$F5,$FC,$C7,$A4
	.dc.b	$8F,$77,$FD,$1D,$E6,$7E,$4A,$3F,$B3,$B5,$FE,$41,$BA,$31,$C5,$FE
	.dc.b	$35,$FA,$57,$C1,$7F,$A5,$67,$3C,$5F,$E4,$E6,$E1,$FE,$44,$6F,$52
	.dc.b	$F8,$CF,$F7,$63,$FB,$01,$7D,$0F,$F5,$03,$83,$C6,$7F,$B3,$F1,$FF
	.dc.b	$34,$39,$6C,$70,$38,$DF,$F7,$6E,$C7,$BD,$FD,$52,$DF,$4A,$F9,$CF
	.dc.b	$F2,$1D,$3D,$FF,$F2,$2C,$E6,$BF,$CF,$93,$97,$FC,$FD,$19,$7F,$92
	.dc.b	$5F,$58,$C4,$F8,$32,$DE,$1D,$B0,$EA,$DC,$32,$76,$C3,$94,$79,$1F
	.dc.b	$46,$5D,$2B,$FC,$B2,$DA,$FF,$48,$3E,$09,$DD,$71,$BE,$B5,$FC,$E6
	.dc.b	$FC,$2F,$B1,$AF,$AA,$37,$87,$A6,$3E,$3D,$79,$DA,$71,$37,$7E,$87
	.dc.b	$D5,$5F,$E8,$57,$A1,$FE,$87,$AE,$8F,$F5,$C6,$BF,$4B,$E0,$7F,$CE
	.dc.b	$8E,$DA,$ED,$AE,$FD,$FF,$F2,$AA,$F9,$BA,$47,$FC,$B2,$DC,$FE,$C4
	.dc.b	$3E,$17,$F8,$C3,$FE,$31,$CF,$FD,$39,$5F,$31,$FE,$E3,$3A,$AE,$35
	.dc.b	$DC,$E0,$7F,$A3,$23,$53,$CA,$32,$7F,$B1,$63,$2F,$F7,$CA,$EC,$F1
	.dc.b	$E1,$AF,$FA,$5A,$2B,$E7,$8A,$F8,$6F,$D7,$5D,$23,$81,$E0,$38,$14
	.dc.b	$E1,$7E,$CA,$E7,$7D,$AB,$CB,$C7,$FC,$07,$D2,$FF,$70,$E1,$FB,$B6
	.dc.b	$C6,$6F,$FD,$4D,$C3,$FE,$3B,$6C,$EA,$31,$5F,$D7,$9A,$FF,$AC,$AB
	.dc.b	$93,$FD,$BB,$1F,$DB,$DF,$FB,$37,$8F,$F6,$DC,$1E,$26,$C6,$67,$75
	.dc.b	$C6,$B5,$18,$8F,$F5,$FD,$70,$75,$5D,$7E,$CF,$FC,$04,$DF,$1B,$B4
	.dc.b	$62,$7F,$C0,$2D,$7F,$C6,$7F,$4C,$F9,$5F,$8B,$08,$C5,$FE,$85,$B1
	.dc.b	$FE,$E7,$6B,$CB,$E5,$38,$C1,$A4,$9E,$4D,$9E,$4D,$7C,$7E,$BF,$22
	.dc.b	$4F,$85,$F5,$FC,$9C,$8F,$80,$D9,$F6,$B5,$F8,$F8,$A3,$11,$FE,$6C
	.dc.b	$FA,$6F,$C0,$D1,$A4,$7E,$61,$F2,$88,$60,$91,$F3,$1C,$B5,$FF,$78
	.dc.b	$0F,$A6,$EE,$47,$2F,$E0,$FB,$BF,$DE,$11,$F8,$76,$CF,$F8,$E1,$AF
	.dc.b	$E0,$9A,$FF,$95,$DA,$FF,$95,$9B,$B1,$A6,$BA,$9F,$4D,$FE,$D7,$BF
	.dc.b	$F3,$BB,$6F,$FC,$D8,$D7,$FC,$E8,$D7,$C1,$EF,$FA,$49,$3E,$96,$C7
	.dc.b	$A5,$AF,$A5,$FA,$3F,$4D,$27,$E3,$B6,$7E,$07,$9F,$F1,$49,$3F,$15
	.dc.b	$B1,$A8,$C1,$1F,$91,$E4,$FC,$D2,$47,$CD,$7E,$87,$91,$23,$E8,$36
	.dc.b	$7E,$87,$9D,$F5,$49,$3F,$55,$FB,$1F,$0F,$23,$EC,$36,$7F,$D3,$E3
	.dc.b	$F2,$DF,$F4,$7C,$FF,$4B,$6C,$69,$AF,$FA,$99,$A8,$8E,$1B,$EC,$7E
	.dc.b	$E2,$31,$1D,$E3,$F8,$FE,$CF,$A3,$FF,$28,$DF,$FA,$D1,$B5,$FE,$A9
	.dc.b	$6B,$FE,$B1,$6B,$FE,$D2,$FA,$6F,$D3,$92,$7F,$3A,$D8,$FF,$40,$C2
	.dc.b	$31,$1F,$C9,$63,$07,$E9,$2F,$F6,$24,$70,$48,$E7,$19,$6C,$7F,$73
	.dc.b	$F9,$6F,$F7,$D3,$7E,$17,$F8,$CC,$62,$7E,$8B,$5F,$F6,$8B,$5F,$98
	.dc.b	$1C,$3F,$3D,$AE,$63,$FA,$4F,$BE,$FA,$6F,$E5,$53,$F5,$63,$1D,$FA
	.dc.b	$9C,$CF,$FB,$A1,$BE,$17,$FA,$37,$DC,$FE,$22,$4C,$BF,$83,$B6,$BE
	.dc.b	$5B,$5F,$94,$FF,$18,$F7,$24,$FF,$0E,$6C,$FD,$26,$BF,$45,$AF,$C0
	.dc.b	$F0,$7F,$D6,$53,$FC,$F7,$8F,$F9,$C9,$1C,$12,$7F,$A0,$46,$5F,$C4
	.dc.b	$7C,$1F,$E3,$B8,$4D,$25,$7F,$22,$D3,$F9,$79,$1C,$12,$BF,$13,$19
	.dc.b	$6C,$7E,$AB,$57,$D2,$FF,$AD,$63,$0D,$9F,$B6,$27,$73,$F6,$27,$53
	.dc.b	$F8,$09,$CC,$F2,$34,$43,$FB,$05,$FF,$5C,$23,$82,$4F,$F6,$38,$CB
	.dc.b	$63,$F8,$0F,$92,$FA,$4F,$E5,$13,$FE,$37,$6C,$67,$E3,$BF,$C2,$AD
	.dc.b	$AF,$81,$F2,$1F,$95,$6F,$27,$E4,$A3,$18,$7C,$0C,$91,$88,$F7,$35
	.dc.b	$FC,$B3,$5F,$8C,$D8,$E3,$3F,$D9,$3D,$E9,$7F,$D6,$9B,$7F,$BE,$6B
	.dc.b	$FB,$D6,$BF,$AC,$FB,$07,$C4,$FF,$83,$1B,$5F,$E6,$B6,$CF,$F2,$66
	.dc.b	$BF,$B4,$F4,$9F,$4B,$F9,$68,$FB,$B1,$8E,$3F,$F1,$DB,$6F,$F9,$7B
	.dc.b	$67,$F0,$7E,$33,$E3,$DF,$45,$FE,$31,$6D,$FF,$8D,$5B,$5C,$1A,$FF
	.dc.b	$8C,$1A,$FF,$33,$AE,$95,$CC,$95,$CF,$E6,$BB,$11,$C1,$FE,$04,$F3
	.dc.b	$3F,$C2,$24,$70,$78,$F7,$F3,$11,$C1,$FE,$0C,$F9,$9F,$F0,$C9,$1C
	.dc.b	$13,$1F,$E6,$71,$96,$C6,$5F,$E6,$91,$96,$C7,$C3,$F6,$3E,$31,$1C
	.dc.b	$1F,$E1,$CF,$A3,$FF,$14,$91,$C1,$FA,$0F,$B7,$FE,$93,$5F,$E9,$56
	.dc.b	$F3,$1A,$4F,$27,$F4,$2D,$AE,$51,$FA,$4F,$8E,$FF,$1D,$47,$EB,$1B
	.dc.b	$7F,$E9,$C6,$BF,$E9,$96,$BF,$DA,$7B,$DF,$63,$FE,$58,$8C,$35,$FF
	.dc.b	$30,$36,$37,$EB,$7F,$A2,$1B,$7F,$E6,$19,$D1,$27,$FA,$2F,$CF,$F3
	.dc.b	$EB,$34,$91,$A2,$57,$F1,$D8,$CB,$D7,$D6,$F0,$59,$A4,$8D,$13,$1F
	.dc.b	$F3,$DC,$EE,$31,$1F,$C7,$FE,$57,$FA,$9E,$26,$93,$C3,$FF,$40,$FE
	.dc.b	$60,$F9,$EA,$4C,$3F,$DC,$51,$97,$FD,$C9,$F5,$2F,$95,$FF,$44,$B6
	.dc.b	$33,$18,$3F,$D1,$BF,$4B,$FE,$73,$E4,$FF,$76,$37,$4F,$F1,$3B,$5F
	.dc.b	$E7,$6D,$7F,$BB,$AF,$0D,$F0,$1C,$9B,$3C,$9A,$F2,$6B,$CB,$E4,$BF
	.dc.b	$DF,$8A,$91,$86,$C7,$9A,$37,$F9,$2F,$FB,$D8,$91,$83,$FC,$7E,$7F
	.dc.b	$51,$E3,$7C,$AE,$38,$BF,$06,$D8,$CC,$FE,$06,$3F,$D5,$DC,$4F,$9F
	.dc.b	$A4,$9F,$F5,$84,$65,$B1,$97,$FD,$77,$19,$6C,$78,$7E,$89,$E0,$D2
	.dc.b	$4F,$F6,$D8,$CB,$63,$2F,$F4,$D8,$CB,$63,$CF,$F3,$D2,$3C,$AD,$7C
	.dc.b	$6D,$7F,$C2,$16,$8C,$47,$F9,$77,$81,$D6,$F0,$3E,$5C,$61,$B1,$99
	.dc.b	$FE,$05,$1A,$EE,$73,$52,$4F,$F1,$36,$BF,$E0,$6B,$FE,$25,$18,$8F
	.dc.b	$F2,$1C,$61,$FF,$5C,$F3,$7D,$CF,$4F,$FC,$13,$1F,$E5,$B8,$C1,$FE
	.dc.b	$E8,$F1,$7F,$A7,$7C,$5F,$E0,$9B,$3F,$D3,$63,$1E,$39,$E9,$F2,$9F
	.dc.b	$0E,$7F,$D0,$F8,$23,$11,$FE,$1C,$8C,$6D,$C0,$91,$C2,$76,$FF,$97
	.dc.b	$27,$6E,$67,$6F,$39,$DB,$FE,$8D,$9D,$B9,$9F,$3F,$D0,$48,$FE,$CC
	.dc.b	$D7,$97,$D1,$7A,$88,$F8,$DF,$7F,$CA,$C8,$FE,$DC,$D7,$FB,$77,$6F
	.dc.b	$C4,$23,$FD,$CD,$AE,$8C,$10,$FF,$01,$FD,$13,$E0,$3F,$63,$E7,$23
	.dc.b	$13,$FE,$04,$6B,$C2,$7F,$80,$FD,$8B,$D0,$A4,$9F,$DD,$C6,$5F,$9D
	.dc.b	$D7,$F6,$08,$E0,$87,$CD,$72,$D3,$FC,$77,$F2,$BF,$C0,$FD,$53,$49
	.dc.b	$1A,$21,$FE,$3D,$8D,$B8,$3F,$CB,$FC,$7E,$C6,$E3,$FC,$72,$31,$1F
	.dc.b	$E9,$E6,$BF,$EC,$BF,$B1,$7C,$D7,$F2,$26,$C7,$F1,$A6,$BF,$ED,$A6
	.dc.b	$BF,$E6,$E6,$BF,$EA,$A6,$BD,$63,$A1,$FC,$87,$E9,$1E,$CD,$24,$FF
	.dc.b	$9F,$63,$2D,$8C,$BF,$E5,$D8,$CB,$63,$F5,$3E,$14,$8F,$D1,$35,$FC
	.dc.b	$F3,$5F,$D2,$73,$FF,$30,$72,$7F,$38,$C7,$FB,$73,$7C,$EF,$94,$47
	.dc.b	$E2,$7B,$3F,$1C,$8E,$08,$7C,$77,$2D,$3F,$6D,$E1,$7F,$00,$23,$82
	.dc.b	$1F,$B5,$72,$D3,$F7,$DF,$13,$FC,$20,$8E,$08,$7E,$F5,$CB,$4F,$F5
	.dc.b	$57,$D1,$FB,$9F,$F4,$43,$5F,$F5,$23,$5F,$F6,$E3,$50,$FA,$37,$CA
	.dc.b	$7F,$6A,$6C,$7F,$62,$6B,$FE,$B4,$38,$46,$C8,$ED,$AF,$F1,$5F,$7F
	.dc.b	$F9,$29,$1F,$E3,$CD,$74,$60,$87,$F1,$DF,$43,$F9,$69,$1F,$E4,$CD
	.dc.b	$74,$60,$87,$FB,$2B,$E2,$3F,$A9,$FC,$57,$F7,$C6,$C7,$AD,$AE,$FC
	.dc.b	$A4,$FF,$81,$BA,$64,$FF,$81,$3C,$FF,$EE,$8F,$FB,$59,$AF,$CC,$0E
	.dc.b	$1F,$9E,$D7,$31,$D1,$FF,$4E,$FC,$BB,$E2,$7F,$A8,$36,$3E,$87,$90
	.dc.b	$8C,$47,$09,$DF,$C3,$C8,$DC,$A3,$11,$DB,$E1,$FA,$91,$DB,$E0,$F6
	.dc.b	$64,$72,$6B,$E6,$F5,$FC,$A4,$7B,$B5,$D1,$82,$1F,$D5,$FE,$33,$FB
	.dc.b	$39,$1F,$EC,$0F,$C0,$F7,$E4,$7A,$9A,$FC,$66,$BF,$3F,$3F,$89,$18
	.dc.b	$8F,$CE,$F0,$3E,$7B,$FC,$9D,$1F,$7D,$AF,$E1,$5A,$FD,$98,$E1,$87
	.dc.b	$F8,$07,$EC,$1E,$65,$25,$7D,$48,$CB,$63,$2F,$FB,$62,$32,$D8,$FF
	.dc.b	$03,$FC,$0C,$8F,$F0,$03,$5E,$1D,$7E,$12,$3D,$DA,$E8,$C1,$0F,$F0
	.dc.b	$27,$27,$F8,$50,$8F,$F8,$35,$F6,$6B,$23,$D4,$D7,$D0,$D7,$E7,$60
	.dc.b	$8C,$47,$E7,$FE,$79,$F3,$DF,$E7,$26,$C7,$F9,$A9,$AF,$F0,$96,$BF
	.dc.b	$A7,$0B,$D0,$E4,$48,$E4,$D7,$93,$5E,$51,$C5,$F0,$7C,$9F,$8C,$8E
	.dc.b	$0F,$D5,$FD,$CE,$DF,$17,$12,$31,$1F,$8D,$6B,$FC,$A2,$35,$7F,$E4
	.dc.b	$5F,$9E,$7C,$FD,$25,$7E,$E2,$32,$D8,$CB,$E0,$8C,$B6,$3F,$CA,$CF
	.dc.b	$FA,$78,$FF,$1E,$14,$7A,$DF,$CE,$BD,$4D,$92,$3F,$CA,$26,$E3,$4E
	.dc.b	$23,$FD,$4F,$D4,$F9,$C7,$37,$F4,$4F,$9E,$7C,$EF,$F1,$BA,$FF,$2F
	.dc.b	$61,$98,$C7,$B5,$FC,$7E,$B7,$3F,$D3,$2B,$A5,$EE,$31,$87,$65,$5C
	.dc.b	$1F,$E3,$97,$CF,$0E,$B1,$C4,$FE,$41,$5C,$6B,$86,$1D,$A7,$9D,$70
	.dc.b	$69,$E3,$BE,$8F,$8B,$0E,$95,$C6,$F9,$D6,$AB,$C6,$6B,$0F,$06,$1F
	.dc.b	$0D,$B1,$D4,$E7,$3C,$6F,$A5,$6F,$0F,$F6,$5C,$79,$EB,$F4,$0D,$74
	.dc.b	$7F,$6D,$5F,$16,$F6,$78,$B6,$74,$D7,$9B,$5F,$EE,$DF,$3D,$FE,$E4
	.dc.b	$4D,$23,$FE,$E8,$7F,$C0,$7F,$3F,$F4,$E1,$EF,$BD,$FC,$E7,$D2,$55
	.dc.b	$BA,$D4,$62,$75,$39,$AE,$33,$FB,$FB,$AF,$FB,$1E,$2B,$FE,$14,$8F
	.dc.b	$C0,$3E,$28,$AF,$A8,$FB,$A7,$12,$9C,$4F,$11,$FE,$C2,$7A,$BC,$67
	.dc.b	$C8,$E8,$FE,$F1,$3E,$2B,$F5,$CF,$C3,$9E,$17,$FE,$CF,$BF,$5F,$03
	.dc.b	$83,$FD,$31,$1F,$D4,$0F,$F5,$93,$8A,$FD,$6C,$63,$DC,$F8,$FB,$7C
	.dc.b	$8B,$E5,$EC,$E2,$B8,$B6,$FF,$7F,$87,$3B,$E0,$FF,$A7,$EB,$C1,$3A
	.dc.b	$AE,$17,$FE,$D1,$6F,$A9,$C6,$BF,$D8,$CF,$1A,$F2,$9F,$C3,$1D,$36
	.dc.b	$78,$9C,$B1,$B5,$E0,$E7,$FE,$0F,$C7,$F8,$2B,$FE,$E9,$A7,$F0,$DA
	.dc.b	$FF,$0C,$B6,$3F,$97,$B5,$FF,$3D,$F9,$DE,$63,$E0,$5F,$4A,$F6,$E7
	.dc.b	$FE,$EB,$9C,$DF,$FA,$C2,$7A,$FB,$7E,$1C,$3F,$9B,$4E,$5E,$F1,$95
	.dc.b	$87,$F1,$E8,$C3,$6B,$99,$27,$FD,$DD,$F5,$FF,$28,$F0,$D5,$9F,$85
	.dc.b	$6F,$7F,$9B,$63,$E0,$C6,$08,$E0,$97,$F1,$23,$CD,$93,$6F,$E6,$75
	.dc.b	$F0,$9B,$5F,$63,$2F,$E6,$F7,$F7,$B0,$F9,$3D,$CF,$A6,$F6,$B7,$1F
	.dc.b	$F2,$5B,$70,$FE,$C2,$D7,$FC,$58,$D7,$FB,$03,$5F,$92,$DC,$3F,$BC
	.dc.b	$D7,$A3,$17,$AF,$F9,$03,$C7,$17,$0F,$C4,$5B,$E7,$FC,$A4,$8F,$2B
	.dc.b	$71,$F2,$B5,$F2,$BD,$FE,$87,$F3,$A7,$C0,$6C,$F6,$F3,$BE,$09,$1F
	.dc.b	$82,$DA,$F8,$2D,$7E,$0B,$EB,$F8,$4F,$D4,$9F,$31,$B3,$EA,$3E,$69
	.dc.b	$0F,$9A,$59,$F9,$AD,$7E,$6B,$F1,$FE,$43,$F6,$A7,$D8,$69,$F7,$E3
	.dc.b	$F7,$27,$EE,$BA,$AC,$7E,$C7,$E0,$5E,$DF,$E6,$9B,$1F,$E8,$86,$D7
	.dc.b	$F8,$79,$AF,$FA,$5F,$8D,$F8,$27,$FB,$F2,$BF,$5E,$D9,$FE,$0B,$3A
	.dc.b	$FA,$BF,$E1,$B3,$9E,$DF,$E0,$1F,$3F,$FE,$8F,$C3,$FD,$8D,$82,$30
	.dc.b	$FF,$B2,$5A,$FF,$90,$5A,$FF,$AA,$5A,$FF,$B5,$9A,$F3,$F5,$BF,$74
	.dc.b	$D9,$FF,$6A,$7C,$73,$F4,$AF,$C6,$47,$F2,$48,$C7,$17,$F2,$BA,$E0
	.dc.b	$7F,$20,$8C,$CF,$C0,$FB,$0F,$69,$EE,$FB,$1B,$19,$25,$7C,$4D,$13
	.dc.b	$A8,$C7,$D8,$FC,$63,$E3,$FC,$DD,$BF,$9B,$16,$79,$4D,$DB,$FB,$5E
	.dc.b	$5E,$BF,$A8,$FE,$84,$DC,$3D,$9A,$BE,$2F,$2A,$A9,$27,$73,$71,$F4
	.dc.b	$7C,$97,$DA,$31,$8F,$F5,$3D,$51,$F7,$BE,$6E,$61,$F8,$FF,$16,$E1
	.dc.b	$F0,$66,$E1,$FE,$02,$F0,$5C,$EB,$CE,$0C,$3F,$CB,$39,$7F,$9A,$23
	.dc.b	$F1,$FE,$8F,$E3,$3E,$BB,$F1,$AC,$8C,$9F,$94,$8C,$57,$F5,$81,$F3
	.dc.b	$8E,$DF,$C2,$B6,$BF,$26,$D9,$FF,$48,$1C,$27,$64,$9D,$B6,$37,$2E
	.dc.b	$FF,$C5,$36,$BF,$B8,$CF,$36,$C6,$63,$F0,$DE,$7F,$EA,$0F,$D4,$FC
	.dc.b	$DD,$AF,$D7,$C5,$9F,$C6,$4D,$DF,$F0,$7F,$87,$3E,$9F,$F0,$6D,$C3
	.dc.b	$FC,$03,$2F,$D9,$1F,$55,$AB,$95,$FA,$3B,$E1,$8F,$F8,$03,$0E,$17
	.dc.b	$A8,$F9,$99,$FF,$3A,$BF,$93,$E4,$FC,$C2,$38,$24,$FC,$68,$CB,$F2
	.dc.b	$3C,$BF,$8E,$7C,$88,$B3,$FC,$3B,$CF,$FE,$16,$7F,$80,$39,$7E,$B6
	.dc.b	$5F,$93,$8C,$CE,$CF,$E4,$1C,$38,$C7,$F6,$5E,$BF,$F1,$43,$93,$67
	.dc.b	$F8,$DC,$58,$FE,$D1,$87,$E4,$E7,$27,$F8,$6F,$F0,$0F,$E4,$13,$F9
	.dc.b	$73,$FC,$0E,$E1,$FF,$61,$35,$CF,$CA,$7F,$83,$7B,$DF,$A4,$7F,$B9
	.dc.b	$EF,$F8,$93,$6B,$FB,$C5,$FF,$29,$9F,$D2,$57,$0F,$81,$FE,$5B,$F0
	.dc.b	$12,$78,$C5,$BF,$E2,$13,$76,$FE,$0F,$8F,$57,$F7,$51,$FE,$49,$F6
	.dc.b	$DF,$8A,$3C,$CF,$F1,$14,$61,$B1,$9C,$BF,$4B,$39,$7F,$4F,$E6,$FF
	.dc.b	$46,$F5,$BF,$A0,$9D,$6B,$86,$7F,$C8,$E3,$FC,$73,$B7,$F0,$AB,$FD
	.dc.b	$7F,$84,$8C,$57,$FB,$C1,$AF,$CF,$8D,$78,$9F,$E2,$2F,$09,$F0,$1F
	.dc.b	$D4,$0E,$B1,$70,$FE,$AF,$16,$FE,$6C,$DD,$FF,$8E,$4D,$AF,$B3,$36
	.dc.b	$BE,$EF,$7F,$F9,$0B,$23,$49,$5F,$0B,$17,$C1,$7F,$62,$3D,$F3,$67
	.dc.b	$FC,$E7,$E9,$3F,$68,$FE,$45,$1F,$C4,$23,$1E,$3F,$F9,$4D,$B1,$99
	.dc.b	$DE,$5F,$91,$3F,$D6,$6E,$1F,$CA,$4F,$07,$4D,$73,$1B,$7F,$98,$70
	.dc.b	$FF,$7E,$F6,$7F,$BC,$9F,$E8,$DA,$E5,$BF,$F4,$18,$FF,$48,$78,$0F
	.dc.b	$86,$FF,$03,$E1,$FC,$E6,$31,$EA,$FF,$9A,$DB,$9F,$F2,$6A,$DD,$7F
	.dc.b	$34,$6C,$EA,$3D,$8D,$7F,$5C,$D7,$CF,$BB,$E0,$BF,$C1,$E7,$92,$BF
	.dc.b	$C3,$79,$7F,$57,$8C,$E1,$FE,$18,$75,$8E,$30,$FF,$4E,$E3,$EB,$C3
	.dc.b	$D0,$7B,$6F,$D0,$E0,$8E,$09,$3E,$88,$E0,$FD,$5D,$3F,$C5,$3E,$FF
	.dc.b	$F8,$90,$F3,$4D,$AF,$C1,$FC,$83,$E2,$BF,$00,$D8,$FC,$23,$70,$FB
	.dc.b	$6D,$7E,$FA,$FA,$5E,$24,$9E,$2F,$F8,$93,$E8,$F5,$C3,$FE,$32,$6D
	.dc.b	$FF,$60,$6B,$F8,$D6,$BF,$C7,$DF,$36,$5C,$71,$FF,$6F,$4F,$16,$BC
	.dc.b	$5A,$FF,$7D,$A6,$2F,$A0,$FF,$21,$C6,$1B,$19,$9F,$83,$18,$F8,$F9
	.dc.b	$1F,$11,$AF,$CF,$EB,$7D,$67,$F3,$5E,$3F,$F7,$54,$62,$7F,$D8,$ED
	.dc.b	$7F,$03,$F4,$4F,$93,$E1,$C1,$F9,$B8,$C4,$FF,$7E,$6B,$F9,$66,$BF
	.dc.b	$96,$6B,$C6,$7F,$C1,$B1,$88,$FF,$09,$B7,$8C,$E7,$8F,$FD,$A3,$F3
	.dc.b	$5E,$2A,$46,$1F,$4C,$66,$B8,$FD,$6E,$EB,$55,$FE,$74,$C9,$79,$6C
	.dc.b	$7F,$8B,$F3,$67,$EF,$C3,$A6,$47,$D2,$4A,$F8,$4D,$7E,$13,$5F,$84
	.dc.b	$FF,$9F,$B1,$FF,$75,$FC,$37,$84,$93,$FC,$89,$FE,$56,$DE,$DF,$C1
	.dc.b	$C6,$23,$F8,$CB,$5F,$E3,$AD,$7F,$18,$D5,$C5,$FE,$DD,$9E,$64,$8F
	.dc.b	$EA,$F9,$BE,$CF,$E2,$12,$3F,$BF,$79,$FF,$46,$7F,$AA,$3F,$E9,$86
	.dc.b	$C7,$F5,$8F,$05,$39,$6C,$76,$FB,$2F,$A0,$47,$FC,$15,$E2,$57,$EB
	.dc.b	$45,$E0,$FF,$33,$57,$F5,$C9,$FC,$BC,$E7,$83,$F9,$BE,$28,$F3,$56
	.dc.b	$EB,$FC,$13,$87,$8C,$8F,$9D,$FA,$94,$FA,$7F,$C2,$12,$7F,$09,$87
	.dc.b	$F9,$CF,$C8,$7E,$4C,$F9,$7F,$CA,$46,$1F,$F5,$DB,$6B,$FC,$0B,$3B
	.dc.b	$6C,$F8,$3D,$FF,$C9,$12,$3F,$D6,$5F,$4D,$FB,$3D,$BF,$6F,$39,$9F
	.dc.b	$CD,$3F,$A3,$F4,$55,$BC,$BF,$C4,$AD,$A4,$EB,$AF,$84,$E6,$7E,$BF
	.dc.b	$C4,$7E,$B5,$A7,$FB,$53,$89,$1F,$ED,$68,$C9,$FE,$84,$8C,$BF,$8A
	.dc.b	$F8,$2F,$D6,$11,$F9,$23,$81,$F3,$5F,$B0,$8F,$DA,$4E,$DA,$70,$9D
	.dc.b	$F6,$7E,$DF,$E2,$64,$FF,$18,$7F,$29,$7F,$99,$8C,$10,$FE,$01,$1A
	.dc.b	$E7,$F7,$7C,$6B,$CC,$A4,$AF,$F4,$AC,$7F,$86,$1B,$E7,$FC,$18,$C4
	.dc.b	$7F,$A9,$1A,$FF,$84,$5A,$FF,$98,$DA,$FF,$AE,$DA,$9A,$6B,$81,$F6
	.dc.b	$7F,$72,$7F,$DE,$8D,$7F,$DE,$4D,$E4,$FE,$2F,$05,$F5,$0F,$86,$FC
	.dc.b	$35,$6F,$97,$FC,$D2,$DB,$FC,$2B,$6F,$8D,$7F,$25,$F8,$29,$5F,$D1
	.dc.b	$9F,$E5,$1F,$03,$88,$FE,$59,$16,$BF,$A6,$CF,$CE,$F5,$A5,$7C,$C7
	.dc.b	$E8,$FB,$32,$3F,$D9,$A7,$C3,$7E,$71,$1C,$EA,$8C,$47,$D1,$6B,$FE
	.dc.b	$91,$F1,$7F,$1D,$E4,$FF,$3D,$24,$7F,$5E,$F1,$EB,$FD,$0E,$2E,$5F
	.dc.b	$D9,$27,$FA,$27,$54,$AF,$EC,$AF,$E3,$B3,$C4,$7F,$4B,$8B,$5F,$EF
	.dc.b	$BE,$6F,$F6,$5C,$FF,$6C,$F8,$EF,$E9,$C4,$7F,$47,$C7,$FC,$5F,$DC
	.dc.b	$FD,$D7,$8C,$69,$27,$70,$F9,$FC,$47,$DF,$E5,$FF,$51,$CF,$18,$DD
	.dc.b	$74,$AF,$EB,$DE,$84,$9F,$EF,$EF,$F6,$1C,$B1,$1F,$D9,$62,$E7,$FE
	.dc.b	$06,$9F,$EC,$DE,$7C,$AF,$F0,$3B,$FD,$A7,$D3,$C4,$7F,$6D,$8B,$5F
	.dc.b	$E0,$E9,$F6,$FD,$04,$AF,$5B,$FC,$3F,$E8,$3F,$92,$47,$AC,$07,$2F
	.dc.b	$F9,$F5,$AF,$FB,$3D,$AF,$06,$BC,$7C,$7F,$E3,$11,$82,$3F,$D0,$36
	.dc.b	$FF,$01,$12,$7F,$C4,$7E,$97,$F0,$CC,$3F,$AF,$67,$FE,$4B,$9F,$F5
	.dc.b	$AB,$5F,$F4,$87,$87,$FC,$72,$7F,$BF,$7A,$9E,$CD,$3F,$6B,$D0,$8C
	.dc.b	$65,$FE,$8E,$6E,$1F,$E1,$C7,$F6,$93,$BF,$A0,$FF,$0D,$41,$A7,$F7
	.dc.b	$3F,$C5,$31,$5F,$C9,$DA,$F9,$1A,$FF,$B1,$DA,$FC,$EC,$91,$88,$FE
	.dc.b	$6C,$D8,$CC,$FF,$7F,$8C,$1B,$BF,$67,$D2,$F9,$89,$7E,$3E,$4F,$31
	.dc.b	$C4,$8F,$FB,$7A,$BF,$30,$F9,$E3,$2D,$34,$FC,$CF,$67,$E1,$92,$3E
	.dc.b	$07,$CB,$FC,$33,$CC,$47,$B3,$5F,$88,$D7,$E2,$3F,$63,$DB,$F9,$E4
	.dc.b	$8F,$99,$EB,$7C,$F3,$E1,$11,$F5,$35,$FA,$0D,$7E,$83,$F8,$4E,$5F
	.dc.b	$B6,$48,$FE,$BD,$F3,$AF,$94,$FF,$17,$36,$3F,$DB,$87,$F6,$97,$0F
	.dc.b	$E4,$1A,$FE,$A0,$D4,$7F,$8D,$63,$31,$F7,$DA,$E8,$EB,$BA,$33,$B7
	.dc.b	$E5,$FA,$BF,$22,$4A,$F3,$FC,$8F,$98,$EC,$47,$9D,$23,$11,$E7,$6B
	.dc.b	$E7,$BF,$04,$7E,$A3,$6F,$CF,$92,$7E,$2F,$8F,$F1,$0F,$61,$1E,$EF
	.dc.b	$F4,$AE,$EF,$F1,$1F,$87,$FE,$1E,$24,$FF,$8B,$3C,$3A,$FF,$0F,$9D
	.dc.b	$BF,$E3,$09,$FF,$12,$CF,$F8,$B7,$F0,$9F,$F0,$B7,$4F,$F6,$EB,$46
	.dc.b	$3A,$7F,$8B,$C6,$3C,$FE,$1E,$9F,$16,$65,$FD,$02,$F8,$3F,$E3,$47
	.dc.b	$FC,$65,$F3,$DF,$E6,$8C,$7F,$90,$FC,$EB,$E3,$FE,$E4,$BE,$37,$EF
	.dc.b	$4E,$70,$FE,$4F,$39,$78,$D7,$0F,$5B,$FC,$09,$0A,$E0,$AB,$C0,$E6
	.dc.b	$BB,$DF,$66,$C7,$43,$BD,$F4,$BE,$77,$CB,$E1,$2E,$1F,$93,$6D,$FF
	.dc.b	$3C,$77,$7E,$46,$E3,$D6,$7C,$95,$E1,$7D,$AD,$7D,$15,$E1,$C7,$DD
	.dc.b	$5C,$70,$F8,$55,$EB,$38,$99,$AE,$0F,$AA,$7F,$C2,$33,$F0,$A3,$F5
	.dc.b	$47,$68,$CC,$FC,$A8,$CC,$BE,$49,$EE,$72,$6C,$FE,$E2,$75,$39,$3F
	.dc.b	$CD,$50,$F9,$EA,$47,$FC,$DC,$DD,$3E,$BC,$7F,$AE,$B8,$7E,$33,$FD
	.dc.b	$9A,$6F,$6F,$F9,$0A,$B7,$3D,$4F,$84,$7F,$6C,$8D,$69,$FC,$32,$B8
	.dc.b	$DF,$4A,$E1,$B7,$82,$FF,$C9,$AF,$4A,$FB,$57,$F1,$A7,$2F,$3A,$EE
	.dc.b	$79,$CA,$78,$F0,$E9,$87,$7B,$F0,$CE,$9B,$3F,$9D,$EB,$FE,$D1,$F3
	.dc.b	$DF,$E5,$ED,$FF,$B5,$F8,$A8,$C6,$3F,$DA,$FE,$7A,$E5,$FD,$AE,$6E
	.dc.b	$1C,$AB,$FD,$26,$D9,$FF,$08,$3F,$DD,$6B,$77,$D6,$B7,$F3,$D6,$F6
	.dc.b	$DC,$7F,$BE,$AB,$EE,$E6,$BF,$C3,$78,$23,$16,$8C,$63,$C2,$F8,$BE
	.dc.b	$AE,$DF,$E0,$5C,$7F,$E7,$EB,$FE,$51,$B7,$FB,$BB,$2F,$E7,$2D,$8F
	.dc.b	$F0,$9B,$5F,$F3,$26,$4F,$8C,$F2,$9D,$67,$86,$9E,$A8,$C7,$AD,$DE
	.dc.b	$FF,$C0,$D4,$8C,$47,$FA,$EE,$31,$87,$8A,$73,$97,$F1,$96,$CE,$AB
	.dc.b	$CB,$18,$39,$EB,$F1,$7C,$5F,$92,$78,$B0,$FA,$2F,$9F,$DB,$E6,$48
	.dc.b	$F8,$2D,$BF,$82,$D7,$E3,$BD,$B4,$FE,$59,$3F,$73,$1E,$8D,$3F,$DA
	.dc.b	$1E,$BB,$E1,$BD,$6D,$8F,$EC,$ED,$C3,$FC,$82,$D7,$FA,$E3,$5F,$F1
	.dc.b	$73,$5F,$F7,$13,$74,$FF,$61,$57,$97,$E5,$3F,$12,$7E,$06,$7B,$6F
	.dc.b	$F8,$E8,$C7,$BF,$E4,$26,$3E,$46,$E3,$E4,$6B,$E4,$7E,$17,$CF,$7E
	.dc.b	$58,$E7,$3F,$9E,$7C,$1E,$C7,$C0,$24,$7C,$06,$D7,$C0,$6B,$F0,$1F
	.dc.b	$9D,$F3,$1F,$A2,$39,$CF,$EA,$9F,$63,$F3,$08,$7C,$C2,$CF,$CC,$6B
	.dc.b	$F3,$1F,$B5,$F2,$DF,$AE,$39,$BF,$B6,$7F,$63,$12,$3F,$73,$D7,$63
	.dc.b	$FD,$1D,$B3,$E5,$9E,$7F,$E6,$1B,$7F,$E2,$46,$D7,$F9,$F9,$AF,$FB
	.dc.b	$3F,$E4,$1F,$2D,$FC,$39,$B5,$FC,$39,$7C,$AE,$67,$F8,$AD,$3E,$79
	.dc.b	$F0,$1F,$AC,$8C,$36,$33,$87,$FB,$99,$B3,$FE,$44,$6B,$FD,$61,$AF
	.dc.b	$F9,$0D,$AF,$FA,$B1,$BE,$5B,$E6,$DC,$7F,$CA,$23,$0F,$9A,$7B,$F2
	.dc.b	$FF,$88,$6B,$78,$7B,$7E,$7B,$E0,$1F,$07,$D5,$B1,$92,$65,$FC,$BB
	.dc.b	$6F,$F2,$F4,$63,$0F,$F5,$2C,$3E,$7B,$E5,$1F,$2F,$E7,$AD,$7C,$B8
	.dc.b	$B7,$CE,$6D,$F9,$7E,$33,$D5,$99,$BF,$1D,$86,$63,$FE,$F2,$9D,$36
	.dc.b	$73,$87,$9F,$E6,$7E,$E1,$8C,$3C,$EF,$DF,$F6,$25,$FF,$B6,$24,$FF
	.dc.b	$AB,$79,$7F,$98,$3F,$83,$9D,$DF,$C3,$7F,$9F,$4E,$A3,$13,$F8,$9F
	.dc.b	$5F,$F1,$87,$E3,$7E,$7A,$C6,$49,$3F,$36,$32,$FF,$BA,$C7,$CB,$BB
	.dc.b	$3F,$1C,$DA,$FF,$46,$CE,$A7,$24,$8C,$CE,$E7,$FD,$26,$D7,$FC,$4E
	.dc.b	$D7,$FD,$24,$BB,$7F,$3E,$D9,$F8,$C6,$AB,$33,$FA,$9F,$4E,$4F,$EA
	.dc.b	$7E,$5A,$E1,$F8,$99,$BA,$FE,$22,$6D,$7F,$9C,$BE,$75,$F1,$DF,$9B
	.dc.b	$6C,$78,$30,$5B,$E7,$9F,$1D,$BC,$25,$EF,$B5,$8F,$2C,$78,$CF,$DF
	.dc.b	$C3,$8B,$B9,$D6,$3F,$DD,$F6,$78,$DF,$40,$95,$F3,$1B,$87,$CC,$6B
	.dc.b	$F3,$1F,$AF,$E4,$FC,$A3,$E5,$B6,$7F,$87,$7A,$1F,$C4,$4F,$C5,$F0
	.dc.b	$D8,$C9,$2F,$F9,$0F,$27,$DB,$CB,$F6,$37,$FC,$67,$D9,$97,$F1,$FC
	.dc.b	$6F,$E4,$D8,$60,$FD,$D4,$DB,$FC,$FC,$27,$75,$FC,$86,$BF,$C3,$5F
	.dc.b	$31,$FC,$92,$BF,$CD,$B4,$8C,$47,$FB,$0D,$A8,$E8,$E1,$70,$3E,$79
	.dc.b	$F0,$3F,$C6,$5B,$8F,$A6,$FC,$19,$7F,$AE,$67,$38,$7F,$13,$F9,$1F
	.dc.b	$E7,$07,$1C,$2E,$1F,$CE,$62,$CF,$FB,$1E,$B8,$DF,$F1,$89,$FF,$27
	.dc.b	$FB,$8F,$9E,$FE,$7F,$87,$FB,$E2,$31,$E5,$FF,$8A,$1B,$19,$9D,$E3
	.dc.b	$FE,$90,$F5,$5E,$37,$F8,$9F,$2F,$E4,$BE,$B7,$6B,$DC,$63,$0F,$77
	.dc.b	$85,$FE,$30,$DC,$D2,$47,$FA,$73,$2F,$F2,$86,$2F,$9A,$FE,$B0,$7D
	.dc.b	$6A,$B5,$FD,$66,$2D,$7F,$68,$BE,$B3,$92,$47,$F3,$D9,$E5,$18,$C3
	.dc.b	$F0,$BC,$5F,$7F,$BB,$CC,$49,$F2,$46,$5F,$36,$BF,$0B,$E5,$BF,$B6
	.dc.b	$1F,$8A,$AB,$3F,$BB,$F4,$5F,$3D,$FC,$A7,$3F,$C4,$08,$CF,$B5,$FE
	.dc.b	$58,$6D,$7C,$36,$CF,$FA,$6E,$B9,$3F,$EA,$76,$BC,$DB,$1C,$23,$FC
	.dc.b	$05,$EC,$FF,$80,$4F,$F4,$8C,$74,$D3,$FA,$EC,$7F,$A5,$3B,$DF,$15
	.dc.b	$FE,$0A,$BF,$CF,$46,$3D,$5F,$F3,$6B,$72,$FE,$5F,$5B,$AF,$E7,$6D
	.dc.b	$9D,$47,$F3,$28,$C1,$BD,$BF,$17,$F0,$3F,$E1,$53,$C5,$1F,$E1,$CC
	.dc.b	$BF,$B8,$C6,$70,$FF,$0C,$3A,$C7,$18,$7F,$20,$3C,$37,$E6,$3D,$75
	.dc.b	$E6,$70,$47,$04,$8F,$34,$7F,$A0,$39,$1E,$FF,$F8,$A4,$F2,$C5,$AF
	.dc.b	$BD,$F3,$AF,$95,$FB,$4D,$8F,$B4,$DC,$3F,$DA,$8D,$7F,$07,$BA,$1F
	.dc.b	$21,$C8,$95,$C9,$FF,$19,$7C,$F6,$AB,$FB,$2B,$6F,$F5,$4D,$7F,$1E
	.dc.b	$1C,$3F,$92,$7D,$19,$72,$C7,$DF,$3C,$9A,$F2,$9D,$BE,$AF,$96,$FE
	.dc.b	$EB,$7E,$96,$E4,$69,$8F,$94,$95,$F1,$1A,$FF,$0D,$E8,$7C,$5F,$E9
	.dc.b	$B8,$BF,$9C,$F1,$FF,$39,$8C,$6F,$FE,$D8,$AD,$DF,$F6,$9E,$47,$CF
	.dc.b	$3E,$3B,$73,$FE,$0E,$8C,$3F,$E0,$66,$BF,$E0,$66,$BF,$E0,$66,$BF
	.dc.b	$A2,$6B,$CE,$3F,$C2,$31,$89,$F0,$C7,$F8,$63,$1F,$EA,$1F,$1B,$E2
	.dc.b	$6D,$FF,$BF,$B0,$4F,$9A,$FD,$3F,$1D,$B9,$CD,$6E,$B0,$6F,$15,$53
	.dc.b	$C6,$49,$F7,$35,$F7,$35,$F7,$3E,$7A,$FF,$66,$47,$97,$E1,$BC,$47
	.dc.b	$CA,$E1,$FF,$68,$76,$D8,$3F,$82,$6C,$FF,$19,$6B,$F9,$29,$E2,$E6
	.dc.b	$78,$11,$C1,$FC,$BA,$9F,$37,$F0,$C9,$1F,$DA,$7C,$CF,$BB,$3F,$DD
	.dc.b	$DF,$F2,$3B,$63,$F9,$CF,$84,$9C,$B6,$3B,$7C,$CF,$CF,$23,$FD,$D7
	.dc.b	$C3,$AF,$F0,$DE,$0F,$F0,$54,$E3,$93,$FA,$4D,$FF,$57,$AF,$B9,$39
	.dc.b	$E3,$FC,$B3,$E6,$9D,$4F,$F5,$7C,$3C,$64,$7C,$EF,$D3,$F9,$EF,$C1
	.dc.b	$12,$3F,$C8,$DE,$43,$E7,$9F,$35,$F9,$E8,$C3,$FE,$68,$6C,$FF,$7F
	.dc.b	$9D,$B6,$7F,$01,$EF,$7E,$3C,$91,$FE,$4D,$F9,$EF,$D6,$BF,$07,$81
	.dc.b	$3A,$8C,$3F,$AE,$F4,$55,$C3,$7F,$F1,$8C,$63,$AF,$53,$9E,$0F,$F1
	.dc.b	$54,$E7,$0F,$F3,$11,$F5,$FE,$23,$F7,$2D,$3F,$D0,$7C,$7F,$ED,$46
	.dc.b	$C7,$FA,$19,$D1,$FC,$2F,$E0,$BF,$54,$47,$FC,$91,$F3,$7F,$AC,$7D
	.dc.b	$91,$96,$9A,$EC,$FD,$97,$C4,$C8,$FD,$C3,$F9,$0A,$FC,$AC,$60,$87
	.dc.b	$C7,$8D,$74,$7A,$BE,$35,$E6,$52,$57,$C0,$87,$9F,$B3,$D8,$E2,$3F
	.dc.b	$D4,$6D,$7F,$C5,$0D,$7F,$9F,$B5,$FF,$07,$B5,$FF,$2C,$35,$39,$D5
	.dc.b	$F3,$BF,$6A,$7F,$D1,$CD,$7F,$B7,$46,$B8,$3F,$96,$F1,$3E,$79,$F0
	.dc.b	$BF,$83,$C3,$84,$FE,$09,$BC,$7F,$80,$6E,$3F,$EA,$0B,$E2,$FF,$25
	.dc.b	$F8,$1F,$E4,$44,$7F,$94,$7C,$15,$7F,$29,$17,$0F,$E8,$FD,$1F,$EA
	.dc.b	$89,$F9,$9E,$BF,$D1,$23,$F3,$FD,$9F,$A8,$7C,$E2,$3F,$39,$AF,$CE
	.dc.b	$6B,$F3,$9F,$E5,$1E,$4F,$F3,$72,$47,$F3,$DF,$1E,$BF,$A9,$8B,$87
	.dc.b	$9A,$7F,$9F,$74,$CA,$FE,$88,$FE,$37,$3C,$47,$EC,$62,$D7,$F9,$63
	.dc.b	$9F,$DD,$3F,$B5,$F8,$EF,$E9,$04,$7F,$D8,$DE,$27,$FA,$49,$BE,$E7
	.dc.b	$ED,$E3,$11,$F7,$1B,$7F,$B8,$5F,$33,$BA,$95,$C8,$EB,$1F,$D5,$FD
	.dc.b	$19,$1F,$D7,$5F,$EB,$58,$E2,$3F,$85,$45,$C3,$EB,$CF,$F5,$FF,$3E
	.dc.b	$57,$F6,$57,$FB,$17,$A7,$88,$FE,$25,$16,$BE,$FC,$FA,$FE,$6A,$57
	.dc.b	$C0,$7F,$84,$FC,$F7,$F2,$08,$FF,$23,$07,$0F,$F8,$99,$AF,$F9,$D9
	.dc.b	$AF,$FA,$FD,$AF,$1F,$1B,$F8,$84,$60,$8F,$F6,$BE,$9F,$EF,$44,$9F
	.dc.b	$F0,$07,$A5,$FC,$27,$0F,$EA,$F9,$FF,$90,$27,$FD,$1E,$D7,$FD,$A6
	.dc.b	$D7,$FD,$13,$E6,$FF,$95,$30,$FF,$06,$7A,$BE,$DD,$3F,$C4,$ED,$CB
	.dc.b	$FD,$AD,$5F,$E5,$AE,$ED,$FC,$F4,$C7,$FD,$B5,$87,$08,$FE,$D7,$F3
	.dc.b	$A6,$9F,$DB,$1C,$4F,$FA,$91,$AF,$FB,$A1,$AF,$FA,$7B,$B1,$18,$8F
	.dc.b	$F7,$44,$62,$7A,$12,$3F,$D5,$B9,$3E,$7B,$DA,$48,$FF,$80,$17,$F9
	.dc.b	$6C,$F9,$BF,$9A,$8F,$03,$5C,$F2,$FA,$48,$FA,$5A,$FA,$5F,$F1,$C5
	.dc.b	$FF,$A8,$A3,$E2,$D7,$CA,$8C,$47,$F8,$0D,$B9,$7F,$B2,$AD,$F2,$FF
	.dc.b	$28,$F9,$04,$AF,$90,$D7,$E4,$35,$F4,$BE,$2F,$73,$EE,$92,$3E,$0F
	.dc.b	$AB,$F5,$8F,$A8,$47,$EA,$35,$FA,$8D,$7E,$2B,$EE,$E5,$FC,$49,$23
	.dc.b	$DE,$36,$7C,$F3,$E4,$FF,$C5,$ED,$9F,$F7,$2B,$5F,$F4,$7C,$F0,$9F
	.dc.b	$F1,$9C,$E6,$7F,$08,$D7,$F0,$6D,$7C,$0D,$9E,$51,$F8,$4E,$AF,$CF
	.dc.b	$91,$F3,$C7,$9B,$E4,$3C,$A7,$80,$91,$E6,$6B,$E6,$A4,$62,$3C,$AF
	.dc.b	$E0,$B8,$3F,$5A,$48,$EF,$3F,$D5,$9F,$F6,$85,$7C,$6F,$1F,$DE,$48
	.dc.b	$F8,$AF,$FA,$C9,$FF,$0F,$F2,$7F,$A4,$38,$3F,$C4,$AD,$E2,$CF,$85
	.dc.b	$FE,$26,$F0,$BF,$C4,$EC,$BF,$F1,$A3,$97,$EB,$F9,$0F,$9E,$FF,$0E
	.dc.b	$63,$FE,$30,$6E,$1F,$E2,$E6,$D2,$B8,$7A,$5F,$E3,$99,$CC,$73,$F4
	.dc.b	$30,$66,$7F,$C6,$7F,$3D,$FE,$6B,$C5,$83,$E7,$5F,$1D,$88,$53,$E2
	.dc.b	$F1,$5A,$9E,$33,$97,$9C,$65,$E3,$EB,$30,$CE,$1D,$6F,$C5,$5E,$18
	.dc.b	$C6,$1D,$9B,$1C,$0E,$F7,$CD,$E0,$62,$FD,$7F,$0B,$E1,$C3,$C9,$28
	.dc.b	$C6,$1E,$D7,$57,$CD,$B7,$DE,$BE,$0D,$78,$DF,$7B,$5F,$6C,$6E,$FD
	.dc.b	$78,$F6,$9C,$56,$E7,$38,$FC,$1C,$7D,$55,$FE,$FE,$9F,$86,$D8,$CB
	.dc.b	$BC,$3F,$24,$FD,$4F,$91,$7B,$78,$B4,$62,$7F,$31,$18,$9D,$4E,$67
	.dc.b	$9F,$CF,$7F,$9C,$53,$48,$FF,$9D,$5F,$88,$BB,$5C,$58,$8E,$27,$D4
	.dc.b	$3F,$1B,$1A,$70,$7E,$42,$79,$C6,$75,$FC,$44,$E4,$DC,$67,$4F,$D9
	.dc.b	$57,$DE,$7F,$DA,$97,$6B,$F3,$07,$8B,$6E,$77,$E1,$7A,$C7,$9B,$0C
	.dc.b	$D7,$A7,$0F,$1B,$67,$C2,$D9,$DC,$7E,$43,$C4,$78,$4F,$A1,$F7,$6F
	.dc.b	$FD,$E9,$B8,$7F,$51,$FA,$2B,$83,$0E,$35,$86,$D7,$F3,$A6,$CF,$F5
	.dc.b	$7A,$D6,$1D,$EB,$4B,$E7,$F2,$6B,$E8,$2C,$EA,$2D,$FF,$61,$9B,$97
	.dc.b	$F8,$43,$E3,$7E,$B4,$FF,$85,$B0,$46,$23,$FB,$53,$5F,$54,$6B,$0F
	.dc.b	$F0,$0F,$8D,$E6,$3C,$F5,$63,$CF,$C3,$C7,$4F,$92,$DB,$FE,$23,$3F
	.dc.b	$E0,$A8,$C1,$1C,$13,$0F,$BF,$5F,$4B,$E5,$3E,$29,$F1,$AA,$D7,$C6
	.dc.b	$F9,$FE,$C4,$7F,$C2,$ED,$BE,$78,$23,$11,$F0,$DF,$37,$CC,$7D,$23
	.dc.b	$E9,$B6,$BE,$9F,$CB,$7A,$C8,$FF,$89,$1B,$5D,$9A,$FC,$F7,$E1,$72
	.dc.b	$FD,$E3,$EF,$B6,$7E,$FC,$FC,$72,$3F,$E3,$06,$9E,$A2,$9F,$6C,$FA
	.dc.b	$1F,$18,$F9,$4F,$C8,$36,$7A,$9F,$91,$71,$3F,$CE,$63,$FC,$A7,$0F
	.dc.b	$9F,$E6,$48,$FE,$CD,$B2,$75,$18,$6D,$7D,$86,$BF,$E6,$46,$BD,$AB
	.dc.b	$CF,$E5,$F9,$8F,$3B,$6B,$CD,$F3,$DE,$C2,$3D,$DB,$5D,$DA,$F7,$7E
	.dc.b	$1F,$CE,$FC,$23,$E1,$B6,$7E,$17,$CB,$7C,$63,$E3,$FC,$F5,$8F,$8D
	.dc.b	$E7,$7C,$C2,$3E,$F7,$E7,$F9,$7F,$38,$F9,$FF,$3B,$67,$FC,$03,$FA
	.dc.b	$4B,$DE,$FD,$67,$94,$8C,$4F,$F8,$46,$31,$3A,$9C,$EB,$FC,$82,$73
	.dc.b	$5B,$9C,$CF,$F7,$B9,$CB,$B8,$CB,$F7,$3E,$4B,$F0,$67,$F1,$69,$DF
	.dc.b	$6F,$F3,$9E,$3F,$B9,$3F,$88,$F2,$26,$3F,$87,$F1,$AE,$3F,$87,$3F
	.dc.b	$8F,$72,$AB,$38,$F3,$8F,$EA,$F7,$F8,$F9,$FC,$9F,$C1,$C8,$FC,$97
	.dc.b	$95,$70,$E9,$9F,$E3,$6B,$18,$F3,$AB,$7F,$E1,$CD,$3E,$57,$C9,$FA
	.dc.b	$F6,$37,$E3,$30,$CA,$FA,$6E,$5A,$7E,$8F,$E1,$BF,$4E,$6F,$4F,$1E
	.dc.b	$C8,$FD,$5F,$C5,$CA,$FD,$57,$C9,$5B,$DC,$DD,$7E,$14,$DA,$FD,$6F
	.dc.b	$9B,$FB,$43,$F6,$BF,$1B,$63,$F6,$91,$70,$F0,$CD,$BF,$ED,$DB,$3F
	.dc.b	$63,$FF,$10,$61,$F8,$96,$DF,$E1,$DA,$FE,$15,$AF,$C9,$6B,$FE,$4F
	.dc.b	$6C,$F5,$C3,$F2,$F1,$AE,$2F,$E0,$1E,$2F,$F0,$C3,$F1,$51,$E0,$C7
	.dc.b	$F0,$11,$FD,$7D,$7A,$74,$C7,$A7,$73,$FE,$C5,$3A,$6C,$E9,$A6,$B8
	.dc.b	$7A,$1E,$0E,$5B,$19,$26,$BF,$95,$D7,$E3,$4E,$A3,$19,$BA,$FD,$A4
	.dc.b	$BF,$94,$D9,$F9,$4D,$7E,$53,$FD,$77,$27,$27,$BC,$F8,$71,$6B,$DB
	.dc.b	$D5,$F0,$CF,$D8,$7B,$ED,$B1,$3A,$26,$3E,$99,$D4,$63,$1F,$DB,$0F
	.dc.b	$93,$F9,$C7,$EE,$FF,$59,$7C,$16,$2B,$E6,$CD,$C3,$E2,$56,$A7,$27
	.dc.b	$D2,$F9,$3F,$61,$1F,$E7,$F8,$23,$FC,$09,$EE,$BF,$1D,$FB,$18,$7F
	.dc.b	$AB,$9B,$5E,$6A,$FF,$11,$D6,$F9,$3F,$09,$F2,$93,$1F,$F6,$2E,$1C
	.dc.b	$1F,$F4,$BF,$3F,$FA,$2E,$75,$97,$F0,$4C,$91,$D2,$F8,$4F,$F8,$BE
	.dc.b	$73,$7F,$8C,$F9,$4F,$AC,$47,$D6,$FD,$9F,$90,$FC,$B1,$F9,$76,$C7
	.dc.b	$FB,$63,$E5,$BF,$AB,$DF,$FB,$BD,$BA,$1A,$6B,$A1,$F3,$CF,$85,$FE
	.dc.b	$5C,$DC,$3A,$79,$D8,$99,$1F,$EC,$69,$D6,$3F,$ED,$8C,$3F,$B5,$E9
	.dc.b	$FC,$02,$72,$DB,$E3,$5B,$C2,$CF,$F9,$CF,$DF,$7E,$D9,$F8,$7C,$BF
	.dc.b	$DB,$D1,$8E,$5F,$F1,$F3,$63,$33,$BC,$3F,$C6,$36,$F9,$4F,$DE,$1F
	.dc.b	$E5,$EB,$E1,$3F,$56,$3F,$80,$7C,$2F,$E9,$49,$87,$E2,$9B,$3D,$1A
	.dc.b	$FF,$B4,$F9,$3F,$D9,$51,$FA,$DF,$85,$FE,$14,$74,$AF,$E0,$EF,$F5
	.dc.b	$9E,$CF,$F3,$24,$7F,$69,$6D,$FF,$67,$6B,$FD,$95,$AF,$F4,$96,$BF
	.dc.b	$EE,$F6,$F0,$FF,$77,$F7,$FF,$BB,$C6,$B1,$FF,$6F,$7C,$A7,$F1,$D3
	.dc.b	$C5,$37,$0F,$E3,$71,$70,$FE,$25,$7E,$4D,$7F,$73,$3E,$9F,$80,$EA
	.dc.b	$4B,$F5,$36,$FD,$4D,$7D,$4F,$F1,$1F,$94,$FE,$5C,$73,$9F,$E5,$6F
	.dc.b	$F3,$1F,$8F,$FE,$2C,$48,$FD,$DB,$FC,$67,$E5,$24,$7E,$F9,$B7,$FE
	.dc.b	$AE,$1F,$36,$F9,$2F,$F6,$24,$66,$31,$EF,$FF,$B6,$1B,$19,$9D,$BF
	.dc.b	$EB,$88,$57,$FA,$B2,$39,$35,$D3,$51,$23,$8F,$97,$FC,$52,$BF,$DC
	.dc.b	$FC,$0D,$BC,$D7,$92,$1F,$9A,$F9,$5D,$C1,$A4,$8F,$E3,$FF,$29,$E0
	.dc.b	$7F,$88,$36,$7F,$98,$B7,$4F,$F0,$97,$ED,$BF,$C5,$2B,$CF,$5F,$DD
	.dc.b	$A3,$11,$FE,$39,$6B,$FE,$37,$6B,$CB,$C3,$FF,$80,$3C,$A9,$CC,$A3
	.dc.b	$13,$A9,$CC,$62,$7F,$9B,$DF,$95,$FE,$48,$DC,$BF,$C7,$11,$FD,$96
	.dc.b	$BF,$A5,$3F,$DC,$A7,$58,$7F,$B1,$8F,$E9,$EF,$3B,$F2,$75,$DB,$E3
	.dc.b	$7E,$D8,$F2,$11,$FF,$2E,$46,$9F,$81,$19,$78,$1F,$E8,$CE,$A7,$C7
	.dc.b	$F9,$DB,$1F,$EE,$76,$BF,$E4,$65,$F2,$BB,$3F,$06,$E2,$3A,$46,$08
	.dc.b	$E0,$87,$FB,$3A,$38,$37,$C9,$F1,$11,$FB,$B9,$7E,$52,$5F,$B4,$BD
	.dc.b	$5F,$A1,$18,$7F,$55,$48,$C3,$FC,$CA,$78,$36,$7F,$03,$8F,$F8,$16
	.dc.b	$3B,$9F,$81,$70,$FF,$87,$5B,$AF,$FB,$F3,$DA,$75,$FD,$79,$CE,$5F
	.dc.b	$E6,$76,$DF,$3F,$3B,$F0,$CF,$F9,$E6,$B8,$44,$8F,$0F,$07,$D4,$8D
	.dc.b	$E1,$DA,$7C,$A7,$13,$DE,$E5,$FE,$14,$D7,$F5,$86,$CF,$A1,$E7,$7D
	.dc.b	$F2,$3F,$7F,$E7,$BF,$04,$7D,$C2,$3F,$87,$7E,$FF,$CB,$7E,$1C,$91
	.dc.b	$FE,$0C,$F0,$EB,$F8,$78,$BA,$B5,$FE,$1B,$E5,$7F,$86,$EF,$F5,$0D
	.dc.b	$BE,$44,$3F,$21,$F2,$DF,$95,$3F,$1E,$47,$F3,$4F,$E4,$BE,$5B,$F3
	.dc.b	$44,$8F,$97,$E9,$3C,$6C,$47,$E6,$E6,$E5,$FE,$7E,$AD,$4F,$E8,$7C
	.dc.b	$BF,$D2,$11,$FA,$BB,$7F,$8D,$5D,$10,$FD,$34,$7E,$E6,$35,$1F,$A9
	.dc.b	$F9,$0F,$D7,$11,$FB,$FF,$0E,$F8,$3F,$E0,$0D,$8F,$F6,$B3,$5F,$F6
	.dc.b	$A3,$5F,$F6,$93,$5F,$F5,$B3,$5F,$E3,$4D,$E4,$34,$D7,$77,$E9,$FE
	.dc.b	$47,$F8,$11,$2B,$CB,$F1,$35,$FE,$05,$17,$5F,$F5,$2D,$AB,$F8,$56
	.dc.b	$F2,$FF,$85,$3F,$C3,$7C,$BC,$47,$F0,$B8,$B7,$FE,$E3,$ED,$57,$F1
	.dc.b	$3F,$90,$97,$FC,$4D,$FE,$49,$FA,$43,$E3,$69,$FC,$E9,$C4,$FF,$BF
	.dc.b	$5A,$FF,$BE,$DA,$FF,$BE,$5A,$FF,$B7,$5A,$FF,$41,$6E,$C7,$3B,$7F
	.dc.b	$0D,$F9,$0F,$E5,$04,$AF,$E5,$1F,$25,$5F,$E5,$11,$74,$F7,$4F,$F2
	.dc.b	$FF,$90,$95,$FC,$BD,$FE,$65,$F2,$58,$8F,$E6,$11,$6B,$E3,$4F,$F3
	.dc.b	$7F,$92,$95,$FC,$DD,$F8,$B8,$48,$FE,$17,$92,$31,$1F,$EB,$78,$56
	.dc.b	$2F,$FD,$BF,$1B,$9E,$35,$F4,$2B,$FA,$37,$A3,$27,$FA,$33,$FD,$27
	.dc.b	$5C,$47,$F4,$78,$B8,$7D,$D9,$FE,$C9,$2F,$92,$A4,$7F,$A8,$14,$FF
	.dc.b	$07,$77,$FF,$A5,$9E,$07,$F3,$6D,$BF,$C2,$F3,$FD,$66,$3A,$FB,$31
	.dc.b	$D0,$7F,$13,$C3,$23,$F0,$CF,$E2,$48,$E0,$87,$E2,$5C,$BF,$CB,$2B
	.dc.b	$FC,$4F,$F2,$5E,$C2,$47,$83,$9B,$DF,$CF,$FE,$28,$8F,$D8,$9F,$E3
	.dc.b	$FA,$FF,$17,$C6,$4D,$7C,$97,$C6,$24,$FA,$BE,$4B,$E4,$FC,$94,$8F
	.dc.b	$6B,$F3,$3B,$24,$75,$3F,$7A,$47,$C4,$47,$4D,$7D,$87,$FB,$DC,$F9
	.dc.b	$3F,$0B,$F6,$0F,$B3,$F2,$52,$3B,$3E,$D2,$38,$21,$D1,$CB,$DB,$E4
	.dc.b	$BF,$06,$47,$EB,$F9,$5F,$86,$3E,$41,$1F,$63,$5F,$3B,$5F,$F4,$E0
	.dc.b	$7E,$87,$C9,$7E,$40,$91,$F8,$4F,$8B,$FC,$99,$F5,$48,$FC,$96,$BF
	.dc.b	$11,$AF,$67,$EE,$7C,$87,$E7,$09,$1F,$90,$F9,$0F,$CF,$9F,$81,$23
	.dc.b	$F5,$5A,$FD,$06,$BE,$A7,$F1,$1E,$97,$EA,$09,$1F,$9C,$F9,$0F,$D5
	.dc.b	$9F,$8D,$23,$F8,$16,$BF,$71,$AF,$C6,$7F,$29,$F2,$1F,$B4,$24,$7E
	.dc.b	$A3,$DE,$FD,$B9,$F9,$92,$3F,$8D,$7F,$CD,$3F,$2C,$F8,$DF,$F2,$C3
	.dc.b	$67,$FC,$36,$D7,$F7,$67,$F9,$B5,$FF,$12,$F8,$C7,$5D,$FF,$E0,$79
	.dc.b	$D5,$74,$3A,$FA,$BF,$E2,$FF,$84,$93,$FE,$28,$F6,$11,$FE,$33,$F6
	.dc.b	$33,$3F,$E3,$98,$B8,$75,$CB,$FC,$C3,$5F,$E3,$4F,$96,$FC,$44,$FF
	.dc.b	$A8,$BE,$57,$F9,$16,$09,$72,$7F,$A5,$E7,$31,$89,$E1,$E4,$7F,$A6
	.dc.b	$E3,$9C,$27,$7C,$DF,$EA,$98,$C5,$6E,$FA,$36,$39,$9F,$EB,$4B,$E7
	.dc.b	$7C,$B0,$FF,$5B,$F0,$F8,$B0,$FF,$55,$36,$F3,$87,$8F,$1F,$04,$F1
	.dc.b	$9F,$09,$D6,$BC,$2D,$3C,$27,$C1,$9F,$46,$1C,$4E,$D1,$C8,$F8,$D3
	.dc.b	$C9,$E1,$87,$2C,$7E,$1C,$7F,$95,$1A,$F9,$63,$FC,$F3,$18,$8F,$81
	.dc.b	$CF,$FE,$7B,$7F,$DC,$D0,$8C,$47,$36,$BF,$45,$AF,$F9,$8B,$E5,$BF
	.dc.b	$D0,$AF,$FB,$79,$A9,$A4,$7D,$83,$C8,$FC,$A6,$3F,$84,$8E,$46,$A3
	.dc.b	$3A,$7F,$BE,$EB,$E4,$CF,$1F,$1E,$BD,$CE,$0E,$23,$EB,$C7,$FB,$8E
	.dc.b	$BF,$45,$AF,$78,$E0,$F8,$63,$37,$DE,$79,$9E,$83,$B6,$1B,$C3,$B4
	.dc.b	$6A,$3A,$61,$E6,$BF,$4C,$BC,$27,$33,$D1,$FD,$A6,$DF,$B5,$9C,$C6
	.dc.b	$32,$5F,$1B,$7A,$59,$8C,$55,$BE,$37,$FD,$EE,$FB,$36,$F5,$18,$BF
	.dc.b	$E1,$35,$E4,$F4,$B2,$6B,$D1,$B8,$73,$8B,$5C,$2F,$FC,$9B,$39,$3C
	.dc.b	$1E,$4F,$F0,$F9,$F4,$9E,$97,$0F,$B3,$0F,$87,$18,$8D,$4F,$F7,$BA
	.dc.b	$7A,$3E,$83,$D3,$56,$7D,$DC,$1C,$70,$F9,$AD,$BF,$F1,$B4,$F3,$8C
	.dc.b	$11,$C1,$30,$FF,$09,$4F,$F8,$EA,$3E,$57,$9D,$F2,$2B,$E4,$BF,$2B
	.dc.b	$D1,$EC,$49,$F4,$36,$F9,$B5,$F8,$78,$7E,$0A,$3C,$DE,$8F,$D4,$3A
	.dc.b	$57,$D9,$7E,$BF,$A1,$EB,$24,$7C,$56,$DF,$A1,$AF,$6C,$BF,$CF,$93
	.dc.b	$F2,$F8,$BF,$02,$74,$7F,$0C,$FE,$12,$7E,$41,$23,$E9,$34,$F8,$A5
	.dc.b	$3B,$1F,$47,$E5,$7F,$CA,$51,$FE,$4B,$6C,$7E,$1C,$FC,$B3,$8E,$3F
	.dc.b	$D0,$3F,$E4,$B9,$FF,$1D,$36,$7F,$C8,$8D,$7E,$4B,$5F,$C5,$35,$FD
	.dc.b	$73,$73,$34,$9E,$77,$88,$F1,$FA,$16,$FD,$1E,$87,$AC,$8F,$91,$B8
	.dc.b	$79,$1A,$F9,$1F,$85,$E8,$7B,$8F,$7B,$67,$E2,$7A,$1F,$08,$F8,$7E
	.dc.b	$85,$8F,$91,$E8,$7C,$B2,$3F,$11,$F9,$DE,$7F,$CA,$3E,$5F,$A1,$67
	.dc.b	$F8,$3E,$FF,$DC,$FE,$59,$F1,$5F,$AD,$F1,$D1,$8A,$FD,$AE,$DF,$C3
	.dc.b	$A7,$35,$BA,$D4,$EA,$73,$5C,$21,$18,$75,$3C,$1F,$E0,$31,$97,$EC
	.dc.b	$FA,$5F,$82,$3F,$8D,$47,$1C,$3F,$07,$1F,$86,$F1,$E6,$1F,$8B,$F1
	.dc.b	$6E,$1F,$C6,$39,$3F,$CA,$93,$F8,$FA,$FC,$6D,$6B,$0F,$C6,$FA,$5F
	.dc.b	$90,$3F,$23,$E4,$5B,$FC,$B4,$5C,$7A,$4D,$AF,$95,$F2,$DE,$6D,$4D
	.dc.b	$F9,$5C,$33,$1F,$A2,$7C,$B9,$DB,$8C,$3E,$3C,$2B,$84,$7A,$FD,$4F
	.dc.b	$D2,$98,$C3,$F4,$AF,$EA,$3D,$39,$87,$EB,$7D,$4B,$7F,$0E,$6D,$FE
	.dc.b	$B2,$6D,$7E,$9B,$CC,$FD,$89,$FB,$2F,$52,$C7,$ED,$E2,$E1,$F3,$66
	.dc.b	$DF,$F8,$F7,$E5,$95,$F8,$B6,$DF,$E0,$F9,$11,$8C,$3F,$10,$D7,$F0
	.dc.b	$6D,$7F,$20,$D7,$F1,$6D,$8E,$75,$FC,$37,$C1,$FE,$12,$7E,$22,$7B
	.dc.b	$61,$FC,$2A,$3F,$B1,$3F,$E6,$9C,$3F,$DB,$8D,$C3,$FA,$FA,$F4,$F7
	.dc.b	$EA,$4A,$DB,$CA,$31,$5C,$38,$7A,$9E,$1E,$3B,$7D,$A2,$CF,$F9,$D3
	.dc.b	$4F,$87,$58,$C7,$E5,$75,$7C,$92,$61,$F2,$74,$46,$30,$F9,$2D,$7E
	.dc.b	$4B,$FD,$7E,$BF,$15,$5E,$EE,$2F,$89,$3F,$09,$F8,$7D,$32,$7F,$25
	.dc.b	$93,$D9,$F8,$51,$F1,$A7,$58,$FC,$67,$F6,$B3,$A8,$C4,$FC,$CF,$85
	.dc.b	$F9,$E7,$EE,$BD,$C7,$C2,$62,$74,$4A,$F8,$93,$A8,$C6,$9F,$E2,$11
	.dc.b	$F2,$DE,$D2,$47,$FB,$F6,$32,$FF,$11,$F8,$17,$CB,$7D,$D9,$FF,$4F
	.dc.b	$46,$39,$3F,$07,$7F,$E2,$78,$C5,$FE,$13,$E2,$7F,$0E,$67,$0E,$51
	.dc.b	$FE,$15,$C1,$3A,$8E,$07,$EE,$EB,$9E,$5F,$8C,$9F,$C8,$FC,$AF,$D5
	.dc.b	$25,$7A,$5F,$B1,$F2,$DF,$96,$37,$1F,$96,$7C,$DF,$2D,$FD,$6F,$03
	.dc.b	$49,$E5,$FF,$3B,$F9,$67,$C2,$FF,$2A,$6E,$1E,$5A,$FE,$D5,$A7,$FA
	.dc.b	$D6,$73,$A7,$E5,$5B,$19,$BF,$0E,$1C,$D9,$87,$E2,$2B,$6F,$F9,$53
	.dc.b	$1F,$2D,$6B,$0F,$F6,$9F,$C2,$BE,$59,$C9,$FE,$3D,$8C,$36,$33,$87
	.dc.b	$F4,$B9,$CB,$F7,$FE,$5B,$F7,$67,$F9,$7E,$33,$7F,$BA,$8F,$DF,$FC
	.dc.b	$57,$E8,$C9,$5F,$87,$6C,$FF,$5E,$F4,$DB,$27,$33,$FE,$34,$8C,$57
	.dc.b	$E3,$FE,$5B,$F8,$59,$D9,$FE,$16,$FF,$58,$EA,$7B,$BF,$DB,$5B,$3F
	.dc.b	$C8,$DB,$7F,$DA,$5A,$FF,$64,$6B,$FD,$C5,$AF,$F6,$C6,$F6,$7F,$7A
	.dc.b	$AF,$E0,$1F,$09,$FC,$78,$F1,$36,$FF,$8F,$C5,$BF,$E2,$57,$E2,$C7
	.dc.b	$F5,$D3,$E9,$F7,$FD,$04,$BF,$43,$6F,$D0,$D7,$D0,$FF,$11,$F8,$5F
	.dc.b	$E5,$E7,$39,$FE,$5E,$FF,$32,$F9,$6F,$E2,$C4,$8F,$DD,$3F,$C6,$7E
	.dc.b	$5A,$47,$EF,$5B,$7F,$E0,$DF,$96,$7C,$B7,$FB,$1E,$30,$D7,$FC,$33
	.dc.b	$1A,$F7,$7F,$C2,$90,$8C,$3C,$EB,$FD,$62,$DF,$2F,$72,$9C,$CF,$46
	.dc.b	$BF,$C0,$EF,$AF,$1F,$E2,$9B,$B7,$9A,$7F,$8B,$46,$3E,$5A,$47,$F1
	.dc.b	$96,$BF,$93,$97,$CB,$7F,$A9,$9F,$E4,$4D,$D3,$F8,$93,$67,$FC,$5F
	.dc.b	$F2,$DF,$C5,$1F,$E7,$6D,$7F,$C6,$CD,$7F,$C6,$2D,$7F,$C7,$CD,$79
	.dc.b	$F8,$5F,$23,$95,$39,$6C,$66,$38,$67,$FA,$AC,$3C,$8D,$BF,$ED,$F1
	.dc.b	$E7,$BE,$7D,$7A,$CB,$FD,$88,$7F,$3F,$9C,$1B,$C3,$F8,$3D,$FB,$2B
	.dc.b	$C0,$47,$FC,$1D,$1A,$7D,$F1,$96,$9A,$7E,$BF,$53,$E3,$7C,$ED,$9F
	.dc.b	$F1,$A6,$48,$C4,$7F,$B9,$17,$CB,$70,$39,$92,$3F,$CC,$53,$C6,$31
	.dc.b	$1F,$C3,$E7,$8B,$A6,$B9,$8F,$2F,$93,$E4,$24,$7F,$91,$E5,$F2,$CF
	.dc.b	$57,$F0,$51,$87,$F3,$ED,$9F,$E5,$53,$B6,$CF,$DD,$C7,$FC,$41,$1F
	.dc.b	$E0,$26,$BF,$E1,$76,$EB,$FE,$F9,$F6,$9D,$7F,$6E,$73,$BF,$FA,$1D
	.dc.b	$B7,$FE,$82,$6A,$AE,$2F,$7F,$4F,$91,$1F,$F3,$21,$BB,$FC,$F3,$4F
	.dc.b	$DD,$CF,$27,$FD,$20,$D7,$FD,$32,$D4,$36,$FB,$9E,$77,$E0,$09,$1F
	.dc.b	$80,$D5,$F2,$DF,$84,$3F,$0E,$48,$FB,$0F,$E2,$BE,$5B,$F1,$24,$8F
	.dc.b	$F0,$17,$9B,$FD,$2D,$E9,$19,$7F,$14,$FE,$E7,$C8,$FF,$0B,$F8,$B6
	.dc.b	$B8,$57,$EA,$67,$2F,$49,$FC,$BF,$CB,$48,$FC,$53,$F9,$C2,$3F,$8D
	.dc.b	$F9,$6F,$CE,$91,$FF,$1F,$FA,$5F,$D5,$9F,$81,$19,$7F,$3C,$FE,$DE
	.dc.b	$35,$1F,$A1,$F3,$7F,$4E,$47,$FC,$93,$AF,$F8,$D9,$F9,$51,$97,$81
	.dc.b	$AF,$47,$F5,$FF,$1F,$23,$E8,$3F,$C1,$7E,$1D,$F0,$74,$FE,$24,$E2
	.dc.b	$7F,$D1,$8D,$7F,$DA,$4D,$7F,$D9,$AD,$7F,$DB,$4D,$7F,$DA,$AD,$EF
	.dc.b	$39,$E3,$FE,$01,$F2,$1F,$C1,$09,$5F,$E6,$69,$7C,$4E,$23,$F8,$24
	.dc.b	$DE,$B7,$5F,$82,$BF,$86,$F9,$D2,$FF,$1C,$FF,$10,$F3,$71,$1F,$C3
	.dc.b	$62,$DF,$28,$FC,$C7,$4A,$BF,$8B,$FC,$7C,$9F,$CC,$3F,$CB,$7E,$59
	.dc.b	$F1,$94,$FE,$74,$E2,$7F,$D6,$8D,$7F,$DF,$4D,$7F,$DE,$AD,$7F,$53
	.dc.b	$BA,$31,$1F,$EF,$E6,$C6,$67,$F9,$6F,$C8,$7F,$2C,$25,$7E,$B7,$E4
	.dc.b	$6B,$FC,$AE,$2E,$7E,$FE,$1F,$E6,$15,$FC,$D3,$E4,$25,$FE,$D5,$FE
	.dc.b	$6F,$F2,$38,$8F,$E6,$91,$6F,$96,$CA,$FE,$7B,$F2,$32,$FF,$7C,$FC
	.dc.b	$7C,$A4,$7F,$A8,$B5,$46,$23,$FD,$71,$1C,$DD,$46,$70,$FA,$DA,$FC
	.dc.b	$58,$FA,$58,$7F,$4C,$F4,$E5,$FF,$0D,$7F,$A7,$ED,$88,$FE,$99,$17
	.dc.b	$0F,$4C,$FF,$6A,$97,$CB,$52,$3F,$C5,$CA,$7F,$6F,$DF,$F8,$FC,$62
	.dc.b	$BE,$1C,$F3,$F6,$E3,$FD,$78,$3F,$89,$E1,$91,$F8,$67,$F1,$24,$70
	.dc.b	$43,$F1,$2E,$5F,$F3,$60,$F9,$2F,$51,$1F,$07,$2F,$B7,$A3,$FC,$53
	.dc.b	$1F,$B2,$3F,$B4,$94,$E6,$EB,$E4,$BE,$21,$23,$D5,$F2,$5F,$1F,$E4
	.dc.b	$A4,$7B,$5F,$95,$D9,$23,$AB,$E0,$23,$82,$1E,$A7,$2F,$B3,$E1,$BE
	.dc.b	$A1,$1F,$95,$F2,$5F,$5C,$F5,$91,$F5,$B5,$F8,$AD,$7F,$AA,$9F,$D5
	.dc.b	$43,$F1,$3E,$47,$F0,$44,$8F,$B1,$E6,$7E,$14,$F9,$24,$7E,$4B,$5F
	.dc.b	$A6,$D7,$BF,$17,$F6,$48,$FA,$3F,$23,$F8,$F2,$4F,$E1,$7E,$2F,$F2
	.dc.b	$47,$D6,$23,$F5,$9A,$FE,$01,$AF,$60,$FD,$EF,$90,$FC,$E1,$23,$F2
	.dc.b	$5F,$21,$F9,$F3,$F0,$64,$7F,$06,$D7,$F1,$AD,$7B,$3F,$8A,$F5,$3F
	.dc.b	$50,$48,$FC,$F7,$C8,$7E,$AC,$FC,$79,$1F,$C7,$B5,$FC,$CB,$5F,$53
	.dc.b	$F9,$6F,$90,$FD,$A1,$23,$F5,$5F,$05,$FB,$73,$F3,$64,$7F,$36,$FF
	.dc.b	$9A,$FE,$59,$F1,$DF,$E6,$36,$CF,$F8,$5B,$AF,$F8,$5C,$62,$3F,$CD
	.dc.b	$5E,$0A,$3F,$C5,$AD,$AE,$44,$9E,$5E,$4F,$F3,$8A,$DF,$B2,$F8,$6F
	.dc.b	$F1,$C1,$2B,$FC,$81,$EC,$D7,$FC,$81,$17,$2F,$E3,$B5,$FE,$2F,$7F
	.dc.b	$AA,$4B,$E5,$BF,$C7,$91,$FE,$2E,$F9,$6F,$E5,$33,$FE,$30,$97,$27
	.dc.b	$E7,$A7,$33,$C6,$31,$E3,$FE,$8A,$B9,$CA,$B7,$18,$6D,$70,$9E,$CD
	.dc.b	$7C,$9C,$DF,$EF,$C7,$95,$F5,$8C,$70,$F2,$9F,$F2,$2B,$53,$E0,$8D
	.dc.b	$CF,$1C,$7D,$CF,$9E,$FC,$18,$FB,$B0,$C9,$F0,$6F,$C5,$8F,$3C,$33
	.dc.b	$3F,$04,$B5,$C5,$F3,$9C,$63,$35,$EB,$8F,$E0,$E7,$BA,$3F,$DB,$B3
	.dc.b	$F2,$23,$2D,$7E,$67,$47,$FA,$6D,$E2,$7D,$97,$0F,$06,$BC,$3C,$AF
	.dc.b	$F7,$4B,$FE,$6A,$F9,$6A,$9A,$47,$6F,$B9,$79,$1F,$E1,$BC,$BF,$03
	.dc.b	$1C,$CF,$79,$FE,$76,$8D,$69,$E6,$AE,$51,$89,$F8,$31,$8E,$2E,$D1
	.dc.b	$E3,$3E,$CD,$7C,$6A,$FF,$15,$6D,$E0,$9D,$3E,$28,$CC,$F7,$9F,$19
	.dc.b	$E8,$AE,$B7,$DE,$32,$77,$C3,$94,$EA,$38,$DF,$9D,$FE,$7B,$5C,$6B
	.dc.b	$C4,$DB,$F0,$47,$29,$EB,$39,$7D,$E7,$4B,$ED,$3B,$39,$C6,$4E,$25
	.dc.b	$73,$87,$52,$DF,$C5,$AF,$D0,$BE,$23,$15,$CA,$BB,$47,$C5,$AE,$4D
	.dc.b	$36,$F4,$9F,$EB,$78,$7C,$C7,$D3,$1C,$1F,$5D,$74,$6D,$78,$23,$94
	.dc.b	$7E,$0E,$FC,$AF,$29,$EE,$78,$5E,$F1,$A7,$E7,$9F,$82,$32,$53,$33
	.dc.b	$E1,$6C,$F8,$4E,$8F,$84,$E0,$7E,$5E,$BD,$4F,$66,$B9,$8F,$2B,$EC
	.dc.b	$3C,$33,$F5,$8F,$03,$B3,$CB,$3F,$01,$FC,$81,$E0,$3F,$2F,$3D,$1F
	.dc.b	$03,$5F,$D5,$4F,$47,$F2,$91,$EB,$7E,$41,$D9,$F6,$BD,$A7,$C1,$7F
	.dc.b	$5E,$3A,$9B,$3D,$47,$C3,$AF,$55,$75,$8D,$9F,$1E,$B9,$DF,$D3,$8F
	.dc.b	$71,$99,$F0,$DF,$78,$F4,$C7,$36,$9F,$21,$FC,$A3,$4F,$19,$F1,$AF
	.dc.b	$BB,$F1,$F0,$F0,$57,$58,$F8,$67,$13,$EF,$9C,$F9,$1F,$A4,$3C,$6F
	.dc.b	$EB,$B3,$FE,$A0,$8C,$65,$FC,$A2,$B7,$5A,$A4,$62,$75,$5C,$27,$F9
	.dc.b	$BC,$E5,$B1,$97,$FC,$83,$19,$6F,$2F,$FB,$2A,$BF,$BD,$C7,$68,$C3
	.dc.b	$FE,$04,$AE,$DC,$1F,$E0,$8A,$FF,$64,$3F,$E0,$AF,$95,$B1,$FE,$03
	.dc.b	$6D,$7A,$3E,$51,$EC,$E3,$2F,$F0,$75,$5C,$8D,$35,$5C,$30,$F2,$D7
	.dc.b	$06,$FC,$8B,$9F,$A3,$63,$7F,$AB,$3C,$0C,$D4,$9E,$2F,$8E,$75,$18
	.dc.b	$3B,$EC,$F5,$BD,$0D,$8F,$8F,$F3,$0F,$D6,$9F,$0F,$AB,$95,$C9,$F8
	.dc.b	$75,$B9,$CF,$0F,$F8,$C3,$81,$BF,$C3,$33,$C5,$FE,$36,$7D,$B5,$31
	.dc.b	$F2,$FB,$97,$0F,$F1,$CF,$07,$8A,$7E,$2D,$7A,$2B,$58,$7F,$90,$3E
	.dc.b	$47,$E8,$1F,$4F,$F6,$2F,$F6,$9D,$FC,$EC,$D3,$A8,$C3,$BE,$AC,$67
	.dc.b	$F8,$5F,$7F,$F2,$DC,$BF,$9E,$6F,$11,$A4,$C3,$F3,$2D,$AF,$CB,$BC
	.dc.b	$B6,$7C,$4F,$DC,$3E,$8C,$5A,$F3,$7C,$57,$E2,$4F,$B5,$E1,$D8,$C9
	.dc.b	$30,$FF,$2E,$F3,$7F,$91,$A3,$1C,$1F,$EE,$CC,$BC,$D1,$F8,$CC,$7F
	.dc.b	$2D,$FB,$4B,$E0,$7F,$1B,$12,$74,$7E,$52,$74,$E3,$1F,$F3,$5C,$74
	.dc.b	$7F,$6A,$D7,$F6,$6D,$47,$EF,$9C,$36,$7E,$6C,$63,$63,$4C,$70,$BE
	.dc.b	$0B,$F5,$87,$AA,$AE,$1F,$A8,$F8,$9F,$D3,$9F,$B3,$F8,$47,$C2,$E2
	.dc.b	$74,$4C,$FF,$D0,$B3,$A8,$C6,$DF,$E8,$AF,$8D,$FD,$D1,$FB,$DF,$01
	.dc.b	$F0,$58,$AF,$C6,$4D,$C3,$FD,$1F,$5A,$9C,$9F,$BE,$FD,$27,$B4,$FF
	.dc.b	$1C,$6B,$FC,$69,$AF,$C8,$E0,$46,$21,$E8,$7F,$0C,$F0,$BF,$89,$B6
	.dc.b	$FB,$6B,$8C,$7F,$84,$5F,$FA,$8B,$EC,$7F,$8F,$D7,$56,$C7,$49,$DE
	.dc.b	$AA,$D7,$8D,$FC,$76,$F7,$A1,$E3,$97,$87,$1E,$D5,$FC,$8F,$EB,$3D
	.dc.b	$16,$69,$33,$FD,$6F,$23,$E7,$3F,$94,$1C,$62,$E7,$FC,$9B,$F0,$B7
	.dc.b	$B9,$FB,$46,$C7,$F4,$D6,$E7,$FD,$21,$AF,$F4,$4F,$9D,$FE,$68,$7F
	.dc.b	$37,$8D,$EF,$FE,$C8,$8C,$D7,$F0,$77,$57,$8B,$FE,$39,$16,$BE,$34
	.dc.b	$7F,$5A,$BE,$77,$FE,$C9,$AF,$E8,$9F,$B6,$9F,$E1,$6D,$8F,$EC,$8D
	.dc.b	$AF,$EB,$ED,$7F,$AD,$FC,$E7,$F4,$B3,$9D,$5A,$FE,$6F,$FA,$93,$DC
	.dc.b	$FE,$D9,$E5,$A3,$13,$FE,$98,$8C,$4E,$A7,$38,$7F,$A9,$27,$35,$B9
	.dc.b	$CC,$FE,$E8,$4E,$9E,$11,$A7,$F7,$FF,$56,$FD,$5B,$FA,$F4,$35,$F8
	.dc.b	$F3,$9F,$47,$FA,$4D,$F8,$A7,$37,$C1,$FE,$E7,$F8,$07,$F6,$AC,$BF
	.dc.b	$D6,$30,$8C,$63,$FE,$BC,$8C,$74,$FF,$83,$5B,$5C,$2B,$75,$EB,$9C
	.dc.b	$BC,$63,$2F,$F1,$88,$CB,$F0,$73,$FF,$7C,$7C,$3F,$F6,$0F,$68,$FF
	.dc.b	$81,$2B,$99,$F8,$49,$DE,$1F,$DB,$2B,$6D,$AE,$18,$7A,$AF,$85,$6B
	.dc.b	$0F,$91,$7F,$EB,$DF,$BE,$7C,$0F,$F8,$A7,$D7,$46,$27,$FD,$AB,$18
	.dc.b	$9D,$4E,$6B,$FD,$78,$D7,$FD,$1E,$D7,$B4,$70,$3F,$17,$F7,$5F,$CE
	.dc.b	$AF,$FC,$0A,$D9,$F2,$9F,$ED,$6A,$ED,$8F,$C1,$9D,$3F,$A6,$F7,$3E
	.dc.b	$C7,$C9,$FF,$90,$5B,$87,$FB,$B1,$B8,$7F,$BD,$9A,$FE,$45,$FC,$07
	.dc.b	$AB,$C7,$53,$49,$5F,$EF,$6A,$7D,$03,$E7,$FF,$DD,$71,$FE,$FA,$8C
	.dc.b	$6D,$B6,$C6,$67,$6F,$E8,$4E,$7F,$41,$FC,$C9,$E4,$D7,$52,$8C,$47
	.dc.b	$EA,$63,$1E,$27,$BF,$EF,$BF,$DD,$92,$D1,$18,$6F,$AF,$EB,$6C,$F1
	.dc.b	$F9,$8C,$51,$A4,$8F,$E0,$1F,$5A,$F8,$FF,$D0,$6D,$FD,$D3,$A1,$39
	.dc.b	$AF,$F2,$0B,$5F,$BA,$75,$F9,$6E,$2F,$F7,$CC,$11,$F3,$E3,$47,$D1
	.dc.b	$8C,$B6,$35,$18,$8F,$98,$70,$2B,$97,$E8,$35,$FD,$A7,$E0,$AF,$93
	.dc.b	$F1,$F3,$23,$13,$FE,$80,$8C,$4F,$A2,$3E,$6F,$C6,$F0,$AD,$C7,$E6
	.dc.b	$AB,$6F,$18,$CC,$7F,$3E,$8E,$27,$FA,$30,$A7,$AB,$55,$7C,$36,$C7
	.dc.b	$E4,$DA,$EE,$34,$47,$4F,$E6,$7C,$17,$EA,$AF,$A5,$FA,$31,$86,$C6
	.dc.b	$6B,$C9,$F4,$DE,$8E,$0F,$E7,$0F,$BA,$3F,$34,$D7,$3A,$7D,$98,$C4
	.dc.b	$7F,$8E,$1A,$9A,$48,$FD,$77,$9E,$F3,$9F,$3D,$F8,$68,$C3,$63,$37
	.dc.b	$FE,$A0,$F2,$FF,$4C,$DE,$AE,$71,$88,$FE,$C7,$18,$3B,$CF,$8E,$3F
	.dc.b	$CF,$98,$F5,$3C,$1E,$67,$FA,$96,$32,$43,$FD,$29,$1A,$F8,$9E,$11
	.dc.b	$9E,$2F,$EF,$79,$FA,$EB,$53,$FE,$11,$AF,$7D,$F5,$8F,$F0,$74,$60
	.dc.b	$F2,$FC,$87,$BC,$91,$FE,$BE,$F4,$BF,$DC,$CD,$F4,$FE,$4C,$62,$3F
	.dc.b	$DB,$ED,$7F,$DB,$AD,$7F,$C7,$1D,$1F,$37,$D4,$FD,$56,$9E,$86,$F1
	.dc.b	$7F,$9C,$3B,$DF,$B4,$7F,$56,$C3,$FD,$9F,$5A,$7F,$DE,$2D,$DF,$FC
	.dc.b	$79,$8F,$EF,$AF,$8F,$97,$70,$E5,$87,$F4,$09,$CB,$FE,$EE,$6B,$FD
	.dc.b	$12,$FF,$CB,$3A,$7E,$0F,$D7,$F5,$4F,$FC,$00,$6C,$EA,$3F,$81,$B5
	.dc.b	$FF,$80,$3C,$4F,$91,$FC,$51,$23,$FD,$D3,$E7,$7F,$36,$DF,$EC,$D6
	.dc.b	$E7,$FD,$81,$5B,$7F,$CE,$BE,$3F,$E5,$7E,$07,$F9,$44,$7F,$BC,$3C
	.dc.b	$1C,$1F,$99,$F7,$3F,$A6,$C6,$48,$7F,$B5,$23,$5E,$57,$E7,$7B,$E4
	.dc.b	$FF,$B3,$DF,$D1,$7E,$A3,$FE,$58,$8F,$F7,$FB,$5F,$E9,$9B,$23,$11
	.dc.b	$FE,$2F,$8C,$7B,$1F,$AE,$9C,$CF,$FA,$15,$AA,$76,$FC,$A8,$CC,$62
	.dc.b	$3B,$E5,$F8,$FD,$BF,$D8,$DC,$3F,$0C,$F3,$FC,$DD,$AF,$F1,$17,$47
	.dc.b	$FA,$4E,$F8,$36,$B7,$3F,$EE,$DE,$EF,$DF,$11,$E2,$79,$7E,$11,$F2
	.dc.b	$39,$8F,$C3,$72,$7E,$FF,$3F,$B3,$7F,$C2,$6F,$75,$CB,$46,$6F,$A0
	.dc.b	$FB,$44,$9D,$7E,$CF,$FC,$85,$F8,$8F,$F1,$CA,$34,$9E,$2F,$F1,$A6
	.dc.b	$CF,$E4,$9A,$FE,$4B,$D1,$FE,$34,$47,$C9,$F1,$AF,$A1,$C4,$7B,$26
	.dc.b	$EB,$FC,$32,$B5,$3E,$BC,$DE,$97,$E4,$C9,$1E,$FF,$8D,$FF,$67,$BB
	.dc.b	$8C,$BF,$AD,$9E,$2E,$FE,$17,$F9,$77,$D4,$C8,$F9,$1C,$48,$FE,$69
	.dc.b	$87,$8A,$7F,$6C,$DC,$BF,$90,$B7,$97,$FD,$A4,$D9,$FF,$40,$FC,$9B
	.dc.b	$E4,$F9,$EF,$27,$D0,$FF,$81,$BD,$55,$6E,$7E,$07,$8F,$FB,$12,$3F
	.dc.b	$77,$9B,$FC,$BA,$E8,$87,$BE,$3F,$09,$EE,$FF,$4B,$FA,$99,$3E,$E7
	.dc.b	$FA,$87,$CA,$3E,$3F,$33,$F8,$E9,$BB,$FF,$3E,$AD,$4F,$D3,$D3,$F8
	.dc.b	$21,$1F,$A5,$D5,$F2,$0F,$EB,$FB,$FE,$46,$3F,$8B,$B5,$FD,$DB,$5F
	.dc.b	$DD,$35,$F9,$71,$C8,$8F,$F9,$5F,$E0,$5F,$31,$FE,$A7,$87,$CB,$FD
	.dc.b	$E5,$2F,$92,$FF,$5A,$4E,$63,$12,$38,$F4,$FF,$AE,$63,$AE,$99,$AE
	.dc.b	$18,$77,$F0,$39,$63,$FE,$BF,$6E,$3C,$DD,$DD,$C7,$A5,$F9,$F0,$DD
	.dc.b	$F8,$0F,$16,$3E,$47,$07,$AA,$78,$57,$2B,$F2,$3E,$7C,$7D,$AD,$AC
	.dc.b	$9D,$2B,$8B,$6B,$C8,$F8,$2B,$D5,$5C,$63,$F8,$8B,$DD,$A7,$12,$9F
	.dc.b	$3A,$38,$BC,$CE,$A7,$A6,$6D,$FC,$C8,$F6,$BE,$FC,$3E,$19,$F2,$23
	.dc.b	$38,$7C,$FC,$3C,$98,$FC,$83,$44,$72,$49,$F3,$47,$3A,$EB,$3E,$D7
	.dc.b	$0F,$B6,$3A,$3C,$67,$D0,$7B,$AF,$D1,$3A,$8F,$71,$9B,$F3,$3F,$16
	.dc.b	$FB,$BA,$7B,$BE,$4B,$ED,$7F,$7E,$FD,$53,$93,$AB,$DD,$F1,$BC,$9F
	.dc.b	$A0,$7D,$E2,$9F,$23,$0C,$4F,$A2,$BB,$E3,$F5,$AB,$9C,$7C,$13,$2F
	.dc.b	$D6,$C7,$D6,$FC,$2C,$7D,$6F,$67,$C5,$7D,$63,$CF,$5E,$CA,$E3,$39
	.dc.b	$9F,$2E,$3E,$B8,$C1,$F8,$3B,$EA,$FE,$0E,$FF,$21,$1F,$12,$B8,$C7
	.dc.b	$5C,$7C,$C7,$BB,$1F,$D0,$CF,$CF,$9E,$24,$70,$43,$D8,$E5,$F6,$3F
	.dc.b	$85,$31,$7E,$57,$11,$C6,$30,$FD,$F8,$CB,$C8,$C1,$0C,$12,$38,$BF
	.dc.b	$48,$F9,$24,$30,$79,$8F,$04,$FB,$9C,$12,$3E,$43,$E5,$C7,$F7,$EF
	.dc.b	$E6,$8F,$F7,$A4,$7E,$C5,$B1,$FC,$A5,$B7,$FC,$B7,$E6,$3F,$DD,$D1
	.dc.b	$FE,$D0,$A4,$62,$39,$B5,$FE,$50,$D7,$F9,$61,$FC,$B7,$E6,$3F,$DC
	.dc.b	$09,$A4,$7F,$DC,$8F,$E6,$D7,$96,$E9,$C4,$71,$3F,$92,$9A,$73,$3C
	.dc.b	$67,$84,$E8,$E1,$A7,$F4,$57,$81,$99,$FE,$2E,$E3,$5D,$3D,$DE,$15
	.dc.b	$D1,$CC,$F7,$BF,$0D,$79,$2B,$C3,$7A,$8C,$4E,$A3,$9B,$75,$F8,$F3
	.dc.b	$E4,$9E,$06,$9F,$31,$C8,$AE,$4D,$C7,$2A,$F0,$BD,$27,$E0,$B9,$AF
	.dc.b	$04,$7D,$69,$DB,$E2,$BF,$63,$E0,$C3,$C4,$7A,$CB,$7D,$67,$78,$78
	.dc.b	$AB,$A6,$1F,$12,$F7,$1D,$4F,$E9,$27,$CF,$BE,$4E,$30,$F9,$D1,$89
	.dc.b	$E3,$3F,$8C,$C3,$BC,$F0,$C3,$E1,$3B,$3D,$D3,$D8,$FB,$4E,$2B,$D4
	.dc.b	$D3,$84,$7A,$0E,$15,$D9,$F4,$57,$67,$D5,$5E,$47,$F3,$C7,$A7,$0F
	.dc.b	$C5,$BE,$03,$A4,$78,$1B,$7E,$93,$CF,$5F,$26,$3F,$C2,$98,$7D,$47
	.dc.b	$E0,$1D,$4F,$29,$6B,$A3,$C2,$FC,$38,$7E,$72,$BE,$5C,$FA,$B0,$EE
	.dc.b	$78,$5F,$81,$87,$27,$EA,$1E,$E8,$F8,$CD,$3D,$25,$3C,$07,$92,$3E
	.dc.b	$33,$E8,$6B,$F1,$B0,$F9,$0F,$C5,$38,$BE,$1A,$F3,$C7,$C0,$8D,$DF
	.dc.b	$D3,$8E,$97,$EA,$7D,$87,$A9,$F6,$3E,$4B,$E9,$87,$C4,$8F,$91,$3A
	.dc.b	$7F,$60,$DC,$3F,$47,$E1,$38,$1E,$27,$F0,$CB,$FF,$2B,$B5,$FF,$40
	.dc.b	$B5,$FF,$67,$47,$0C,$3F,$CC,$33,$9A,$E8,$D9,$D5,$7F,$99,$23,$5D
	.dc.b	$BF,$C8,$6F,$FD,$2A,$7F,$1C,$DF,$FD,$23,$36,$3F,$CB,$58,$2B,$85
	.dc.b	$FF,$B2,$3C,$DD,$3F,$EE,$18,$46,$25,$CA,$F8,$9C,$56,$B0,$FF,$52
	.dc.b	$DE,$0E,$15,$B8,$C9,$FE,$B7,$F8,$FE,$9E,$11,$ED,$DA,$F5,$F3,$12
	.dc.b	$B8,$1F,$E9,$D8,$C6,$9F,$EB,$0F,$07,$CD,$C1,$FE,$AB,$AF,$F7,$3F
	.dc.b	$83,$FE,$A8,$F9,$8C,$D4,$9F,$F6,$23,$F0,$27,$51,$83,$FD,$7F,$A7
	.dc.b	$B8,$E1,$87,$FB,$B9,$F4,$DC,$9F,$F6,$77,$9D,$6B,$FD,$99,$37,$8F
	.dc.b	$FD,$87,$36,$BE,$9F,$03,$E6,$3E,$59,$F3,$FC,$7F,$F6,$DC,$FF,$B4
	.dc.b	$23,$2D,$3F,$DC,$FC,$F8,$CB,$FD,$A3,$F2,$9F,$87,$24,$FF,$AB,$1B
	.dc.b	$1F,$EA,$C6,$BF,$EA,$C7,$F2,$1C,$6F,$98,$FA,$A7,$CC,$8B,$5F,$EE
	.dc.b	$2F,$98,$FC,$09,$F6,$3C,$1B,$19,$26,$0D,$BF,$B7,$F4,$B9,$FF,$D6
	.dc.b	$15,$A9,$CE,$6C,$BF,$0A,$7E,$47,$E6,$3F,$DA,$55,$FE,$FF,$8C,$B4
	.dc.b	$FF,$5B,$57,$69,$DE,$3F,$EF,$DA,$E4,$7E,$B6,$3B,$57,$F6,$1C,$51
	.dc.b	$8A,$FE,$DF,$18,$AE,$04,$AE,$15,$BA,$FF,$0D,$D6,$E7,$A7,$03,$E6
	.dc.b	$3F,$4E,$78,$A3,$CE,$7E,$09,$FD,$1F,$CC,$48,$CD,$AD,$EB,$BE,$5B
	.dc.b	$53,$F3,$6B,$F7,$BA,$B0,$CE,$BF,$43,$0E,$15,$D3,$15,$EE,$31,$9B
	.dc.b	$E6,$3F,$7A,$7F,$01,$EE,$7A,$18,$AD,$CD,$D3,$C4,$79,$EB,$73,$A3
	.dc.b	$C1,$ED,$F8,$A7,$E6,$B5,$F9,$2D,$7F,$8B,$B5,$7B,$9F,$C5,$3C,$1F
	.dc.b	$E2,$ED,$AE,$BC,$18,$C3,$F8,$64,$7B,$FD,$FF,$E5,$3C,$1F,$92,$6C
	.dc.b	$7F,$BF,$6F,$96,$3E,$AD,$D5,$A8,$FA,$4D,$8C,$F8,$9D,$F3,$FE,$4D
	.dc.b	$39,$F8,$0F,$45,$1A,$4D,$3F,$9C,$FC,$C7,$F2,$A3,$F9,$74,$7C,$5D
	.dc.b	$3F,$4F,$C2,$9F,$E6,$5F,$06,$F7,$7F,$A3,$37,$7F,$CD,$37,$6F,$CC
	.dc.b	$B5,$FC,$AD,$BE,$5F,$F9,$C9,$D2,$ED,$FE,$0A,$2D,$7C,$3B,$E9,$8F
	.dc.b	$92,$72,$DA,$FC,$64,$E9,$FE,$3B,$3F,$D2,$7E,$1B,$C1,$1F,$AF,$6D
	.dc.b	$FE,$B9,$AF,$EA,$9A,$FF,$26,$F9,$8F,$E9,$E7,$3A,$B5,$FC,$FB,$E5
	.dc.b	$DF,$31,$FD,$B7,$C9,$46,$27,$FC,$BD,$18,$9D,$4E,$70,$FF,$36,$CE
	.dc.b	$6B,$73,$99,$FF,$4B,$CE,$5D,$C6,$50,$F8,$C7,$CC,$7F,$61,$8F,$D2
	.dc.b	$C6,$3C,$FF,$EA,$15,$E1,$DB,$FD,$27,$39,$C7,$F8,$0C,$3E,$57,$FB
	.dc.b	$76,$1F,$E8,$28,$C7,$57,$F8,$2D,$B5,$FE,$7B,$6C,$EE,$B8,$BF,$EB
	.dc.b	$08,$CB,$63,$2F,$F2,$A6,$AC,$DE,$B3,$E2,$BF,$B4,$37,$0F,$3D,$F9
	.dc.b	$B0,$FF,$56,$DF,$0B,$D9,$2B,$78,$DC,$7F,$84,$FA,$6F,$98,$7B,$DF
	.dc.b	$E2,$78,$C3,$63,$37,$FE,$99,$9C,$B6,$33,$5F,$E9,$F6,$BF,$E9,$B6
	.dc.b	$BF,$E6,$A6,$BE,$FF,$98,$FB,$73,$FE,$0A,$6C,$79,$A7,$AE,$7E,$E9
	.dc.b	$CB,$FE,$CC,$F9,$6F,$B1,$F3,$1F,$E3,$F6,$E1,$FE,$AF,$6E,$1F,$EB
	.dc.b	$C6,$BF,$EE,$47,$F0,$1E,$9F,$1D,$CD,$25,$7F,$AE,$69,$F3,$0F,$86
	.dc.b	$FF,$72,$C7,$FA,$F6,$31,$AE,$DB,$19,$9D,$BF,$EC,$E8,$7C,$4F,$F3
	.dc.b	$08,$E4,$D7,$4D,$7F,$DF,$71,$C3,$C6,$F7,$7C,$AF,$FB,$9A,$5A,$23
	.dc.b	$0D,$F5,$7D,$4D,$9E,$3F,$1D,$8A,$34,$91,$FB,$CF,$97,$7C,$C7,$E7
	.dc.b	$76,$FD,$9F,$42,$73,$5F,$E4,$46,$BF,$D6,$0E,$BF,$13,$C5,$FD,$DE
	.dc.b	$08,$F9,$D1,$A3,$E8,$46,$5B,$1A,$8C,$47,$FB,$B5,$AF,$FB,$E9,$AF
	.dc.b	$F0,$83,$A1,$FB,$DF,$98,$7C,$77,$8F,$99,$18,$9F,$E0,$AD,$7D,$5A
	.dc.b	$3E,$63,$84,$E6,$BF,$D0,$8D,$D3,$FC,$4E,$DB,$FE,$33,$7C,$5F,$F0
	.dc.b	$89,$F5,$EF,$E0,$C6,$63,$D0,$70,$8C,$91,$FC,$A4,$6D,$DB,$5F,$80
	.dc.b	$D7,$F9,$47,$CC,$3E,$4F,$E9,$78,$08,$C4,$FF,$89,$23,$13,$F5,$E1
	.dc.b	$F2,$FE,$B6,$E9,$F6,$23,$E3,$47,$A1,$AF,$D1,$6B,$FC,$E9,$AF,$23
	.dc.b	$F5,$CE,$0D,$F9,$FF,$EA,$BF,$98,$FC,$43,$67,$FC,$57,$18,$F2,$BF
	.dc.b	$57,$1F,$E3,$16,$BF,$D8,$7A,$7F,$75,$49,$CC,$FF,$63,$9C,$F9,$BE
	.dc.b	$E8,$FF,$1B,$E1,$E0,$8C,$47,$F7,$7F,$87,$C4,$C8,$D9,$FD,$DA,$30
	.dc.b	$FF,$87,$E7,$8F,$17,$F8,$42,$BD,$7E,$6F,$FA,$CE,$B8,$1F,$E0,$EB
	.dc.b	$F3,$C6,$63,$FC,$1F,$1E,$6F,$98,$F2,$12,$3F,$CA,$1E,$8F,$FB,$99
	.dc.b	$BE,$AF,$C0,$8C,$47,$F9,$C1,$AF,$F9,$B1,$AF,$FB,$6B,$D4,$FD,$6E
	.dc.b	$5F,$E3,$BE,$9F,$A3,$18,$E8,$FF,$1E,$7C,$C7,$F5,$7B,$FF,$21,$56
	.dc.b	$88,$FF,$9B,$BC,$04,$EE,$3F,$CB,$71,$8F,$2F,$F8,$23,$63,$84,$EF
	.dc.b	$1F,$F4,$EF,$AF,$EA,$6E,$9F,$72,$3D,$35,$FE,$7B,$6F,$21,$BC,$4F
	.dc.b	$98,$FB,$84,$AF,$F3,$C7,$9B,$FC,$DB,$1F,$F2,$2C,$E6,$7E,$D3,$F9
	.dc.b	$29,$D6,$DF,$E8,$4E,$FF,$CA,$FB,$7F,$CA,$AF,$FD,$04,$D7,$FC,$C1
	.dc.b	$EE,$7F,$4E,$73,$3C,$08,$FF,$A2,$63,$4F,$E2,$7B,$3F,$24,$47,$FD
	.dc.b	$1D,$F3,$1F,$E5,$76,$FB,$1F,$9E,$8C,$47,$FA,$B9,$AF,$FA,$A9,$AF
	.dc.b	$F8,$B3,$74,$62,$3F,$D4,$8D,$8C,$CF,$53,$F5,$78,$FD,$3C,$BF,$D4
	.dc.b	$BC,$1F,$08,$FD,$BD,$BE,$63,$18,$FC,$49,$AF,$FA,$0E,$74,$FF,$A9
	.dc.b	$BB,$3F,$60,$47,$FD,$59,$F0,$6F,$8F,$C4,$7E,$B6,$6F,$17,$FB,$8A
	.dc.b	$B5,$3F,$4B,$E4,$BE,$A9,$1F,$F5,$B7,$CC,$7F,$35,$DC,$FF,$10,$27
	.dc.b	$87,$FC,$21,$B3,$FE,$DB,$6B,$FE,$DA,$7F,$8B,$F3,$7F,$81,$63,$FD
	.dc.b	$7F,$E7,$FF,$09,$3F,$8C,$FC,$5B,$E5,$31,$3A,$21,$FE,$A5,$8D,$3F
	.dc.b	$8A,$F4,$3F,$1A,$47,$FD,$9D,$F2,$1F,$EC,$87,$44,$3F,$D5,$DC,$A9
	.dc.b	$FE,$53,$3B,$9F,$E2,$9F,$31,$FC,$74,$8F,$4C,$36,$47,$07,$FB,$A3
	.dc.b	$D0,$7C,$C3,$E4,$B5,$3F,$EB,$DC,$3D,$2D,$BF,$F0,$05,$F0,$AF,$D3
	.dc.b	$F8,$92,$BF,$4E,$FF,$3E,$E3,$FF,$2A,$C6,$08,$7B,$23,$5E,$E7,$F3
	.dc.b	$5F,$97,$93,$FE,$EE,$7F,$A3,$FC,$9B,$E6,$33,$3F,$CC,$E6,$F4,$3B
	.dc.b	$BF,$45,$7B,$AF,$DC,$EB,$FB,$C2,$3F,$43,$A7,$E5,$9F,$D3,$37,$34
	.dc.b	$91,$FD,$E8,$74,$A7,$71,$98,$C4,$F0,$8F,$F0,$0C,$61,$FF,$29,$7C
	.dc.b	$03,$E6,$19,$FF,$A9,$3E,$5B,$EF,$57,$FA,$9A,$9F,$25,$FE,$21,$9C
	.dc.b	$CF,$FA,$3A,$31,$CF,$FE,$25,$AE,$97,$CE,$31,$87,$6B,$FD,$84,$6E
	.dc.b	$FF,$C5,$9D,$FD,$30,$EF,$5F,$EB,$68,$C5,$F1,$6E,$1D,$67,$CD,$7E
	.dc.b	$26,$99,$3F,$D4,$75,$E1,$C7,$8C,$1F,$04,$F1,$8F,$61,$C7,$0F,$5B
	.dc.b	$BC,$7D,$78,$79,$5F,$54,$53,$C2,$7B,$8F,$95,$87,$A6,$38,$C6,$4E
	.dc.b	$E5,$AF,$82,$D9,$E0,$F2,$AF,$25,$F9,$EB,$78,$7B,$0F,$6E,$3E,$CB
	.dc.b	$F2,$1D,$9C,$E3,$E2,$75,$1E,$1B,$FA,$73,$93,$CF,$5A,$23,$83,$53
	.dc.b	$F0,$2B,$99,$D9,$F4,$1D,$4F,$1D,$FB,$A3,$51,$F5,$B0,$F6,$B0,$FA
	.dc.b	$07,$2B,$DD,$71,$BF,$A3,$3F,$1E,$BA,$E3,$F2,$23,$E0,$4E,$B0,$E3
	.dc.b	$3F,$83,$8E,$38,$7B,$2B,$B4,$7B,$1E,$27,$B5,$C6,$1F,$84,$BF,$7C
	.dc.b	$75,$33,$7D,$6B,$F0,$77,$D5,$F7,$1E,$C2,$D7,$27,$B1,$F2,$1F,$7E
	.dc.b	$3F,$52,$32,$70,$BF,$A4,$F6,$7D,$66,$1D,$1C,$30,$F9,$0F,$58,$D4
	.dc.b	$7E,$29,$E3,$1F,$08,$87,$E5,$AF,$D5,$3F,$A4,$7F,$21,$3C,$27,$4F
	.dc.b	$4C,$3E,$14,$60,$F3,$1D,$2B,$94,$F1,$7C,$51,$99,$E3,$18,$8E,$47
	.dc.b	$CA,$AF,$8A,$FE,$32,$FE,$33,$F0,$0D,$5F,$58,$FD,$45,$F4,$3A,$BF
	.dc.b	$AB,$33,$8F,$F0,$18,$FE,$31,$5F,$EA,$AC,$3F,$81,$B6,$32,$FE,$F0
	.dc.b	$D7,$CC,$7F,$BA,$9F,$E7,$12,$8C,$46,$BD,$DF,$E6,$71,$88,$FE,$7D
	.dc.b	$F3,$1F,$ED,$A4,$D2,$3F,$ED,$E7,$F7,$AB,$E0,$3F,$A9,$3F,$C3,$41
	.dc.b	$C8,$C3,$AD,$7F,$9D,$C6,$A7,$9C,$E7,$4F,$1C,$F0,$E3,$EE,$F3,$BE
	.dc.b	$35,$B8,$FF,$67,$DF,$2A,$DB,$63,$C8,$71,$BF,$F6,$AD,$71,$8F,$F7
	.dc.b	$74,$EE,$FC,$B1,$B3,$FD,$E5,$3B,$7D,$07,$94,$EB,$3C,$6B,$D3,$5E
	.dc.b	$6B,$F1,$D6,$E3,$AD,$6E,$3D,$26,$6B,$89,$E6,$AE,$35,$CE,$7B,$9B
	.dc.b	$38,$B9,$3A,$B9,$3D,$A5,$3D,$2E,$E7,$BC,$7B,$A7,$BC,$78,$63,$BC
	.dc.b	$ED,$EF,$19,$8E,$E6,$63,$BB,$5E,$F1,$B3,$BB,$F8,$6B,$F9,$8F,$CF
	.dc.b	$31,$5E,$27,$C8,$78,$A7,$93,$E2,$33,$5E,$28,$CC,$78,$9A,$F8,$8D
	.dc.b	$BE,$27,$F3,$53,$DD,$F0,$9D,$CD,$9D,$CC,$CF,$77,$27,$72,$9D,$CD
	.dc.b	$CF,$7A,$F7,$DF,$E0,$67,$81,$A3,$B1,$E4,$3C,$46,$4D,$99,$AF,$13
	.dc.b	$93,$C0,$66,$BC,$17,$ED,$8C,$1D,$AF,$C6,$FE,$14,$FA,$46,$8E,$EE
	.dc.b	$AB,$EE,$46,$A3,$15,$F3,$A3,$15,$DE,$73,$1F,$10,$C9,$EC,$3B,$1E
	.dc.b	$87,$F7,$9B,$FF,$93,$FC,$5F,$E1,$B3,$FC,$30,$FF,$29,$F1,$23,$31
	.dc.b	$FE,$C3,$8C,$5F,$F9,$55,$B1,$FE,$54,$E4,$7C,$3F,$08,$FF,$2E,$C6
	.dc.b	$2B,$FC,$B8,$D8,$E1,$3B,$C1,$39,$9F,$24,$62,$75,$39,$9F,$F3,$AC
	.dc.b	$71,$EA,$7A,$B4,$98,$7F,$9A,$E3,$2D,$8C,$D7,$F9,$B2,$72,$D8,$F5
	.dc.b	$76,$FF,$A7,$3E,$62,$F7,$9A,$4C,$3F,$CF,$93,$B6,$CE,$DB,$3B,$BF
	.dc.b	$F3,$DE,$9F,$EA,$79,$F9,$9D,$FF,$0E,$31,$3A,$8C,$7C,$1C,$AF,$A4
	.dc.b	$D9,$E6,$D7,$E4,$D2,$31,$1F,$EF,$DF,$0F,$FD,$85,$5F,$E8,$F6,$C7
	.dc.b	$FA,$39,$AF,$D2,$6B,$C7,$8B,$EC,$46,$23,$FD,$36,$DE,$03,$98,$FE
	.dc.b	$B3,$F2,$3F,$85,$6D,$FF,$A8,$DA,$58,$D2,$69,$23,$84,$65,$FC,$A7
	.dc.b	$CA,$FC,$92,$38,$3F,$05,$D9,$F8,$72,$38,$24,$FE,$0E,$32,$D8,$FF
	.dc.b	$5A,$7C,$AF,$FB,$25,$BE,$17,$E6,$63,$13,$FE,$B6,$6B,$F9,$A6,$BF
	.dc.b	$DE,$B2,$46,$23,$FD,$86,$D8,$CC,$FE,$4B,$F6,$47,$D0,$D2,$61,$F9
	.dc.b	$08,$CB,$63,$38,$7E,$F4,$71,$AB,$71,$98,$C4,$FA,$F3,$97,$F0,$1B
	.dc.b	$5F,$ED,$2F,$49,$F2,$7F,$B9,$6D,$7F,$B6,$5B,$3B,$9D,$12,$74,$D8
	.dc.b	$FF,$6E,$FA,$BF,$6D,$BB,$FF,$05,$8C,$47,$F0,$46,$BF,$2D,$AF,$FB
	.dc.b	$B1,$BA,$9A,$6B,$D9,$74,$9F,$E3,$06,$0A,$E7,$1F,$F7,$C4,$65,$FF
	.dc.b	$7B,$7B,$1E,$EC,$3F,$8B,$B5,$FB,$CD,$7F,$DF,$ED,$9F,$27,$CA,$78
	.dc.b	$30,$34,$97,$E0,$8C,$B6,$33,$7F,$76,$B9,$CF,$9B,$F0,$A7,$CE,$FB
	.dc.b	$1B,$19,$8C,$1F,$CC,$7D,$FF,$36,$86,$93,$6F,$8B,$FB,$6B,$D3,$A4
	.dc.b	$C3,$E2,$C6,$5F,$C8,$66,$C1,$F0,$4F,$6F,$FA,$03,$6B,$73,$A9,$C9
	.dc.b	$23,$33,$BA,$FC,$E5,$6E,$73,$5B,$9C,$D6,$DF,$E9,$6D,$7F,$64,$D5
	.dc.b	$F4,$FE,$7F,$13,$FA,$D3,$6F,$D6,$D9,$FD,$63,$5F,$EA,$FF,$0A,$F9
	.dc.b	$6F,$EB,$8D,$AF,$6B,$67,$D2,$D7,$F6,$DF,$AC,$F8,$D3,$49,$97,$F2
	.dc.b	$96,$D7,$F2,$96,$CF,$EE,$FE,$55,$F2,$FF,$D6,$9B,$5F,$D5,$9B,$3B
	.dc.b	$9F,$E6,$9F,$AC,$BE,$7E,$92,$FF,$9B,$46,$5B,$19,$AF,$BF,$39,$6C
	.dc.b	$7F,$80,$FD,$57,$CA,$FE,$CE,$FF,$0B,$18,$EA,$D4,$E6,$FF,$8C,$4E
	.dc.b	$6B,$73,$99,$FF,$0C,$4E,$5D,$C6,$5F,$F5,$3C,$65,$FE,$45,$FA,$EF
	.dc.b	$CF,$7F,$C4,$0D,$7F,$D5,$AD,$4D,$26,$FF,$D6,$5B,$7F,$9E,$F4,$5F
	.dc.b	$32,$F7,$7F,$C4,$51,$86,$C6,$67,$83,$70,$FD,$3F,$E1,$8F,$A3,$FE
	.dc.b	$CC,$DA,$FE,$CC,$D9,$DC,$EB,$EB,$A5,$69,$B3,$FB,$2F,$9A,$FD,$C6
	.dc.b	$5F,$DB,$DB,$19,$EC,$FF,$68,$36,$75,$F1,$1F,$C9,$63,$93,$5F,$E9
	.dc.b	$94,$8C,$47,$F9,$2E,$31,$E1,$7F,$7D,$FC,$29,$F1,$BF,$DE,$1B,$59
	.dc.b	$8C,$19,$F9,$C9,$1C,$5A,$FF,$0C,$87,$C3,$7F,$80,$9B,$1F,$E0,$B6
	.dc.b	$BB,$8F,$F0,$3F,$E8,$CF,$36,$92,$7F,$C2,$51,$96,$C6,$5F,$E3,$71
	.dc.b	$96,$C7,$F8,$4B,$E6,$5F,$35,$49,$3F,$E1,$D8,$CB,$63,$2F,$F2,$68
	.dc.b	$CB,$63,$FC,$3B,$FA,$53,$E8,$69,$27,$FC,$51,$19,$6C,$65,$FE,$63
	.dc.b	$19,$6C,$7F,$8A,$3E,$65,$F2,$34,$93,$FE,$33,$8C,$BF,$E3,$5E,$3F
	.dc.b	$F0,$8B,$7D,$4F,$F1,$8C,$62,$3F,$9F,$B5,$FF,$08,$35,$FF,$47,$35
	.dc.b	$1F,$D2,$E3,$99,$FE,$43,$F2,$9E,$3F,$E3,$67,$F9,$4F,$2E,$63,$11
	.dc.b	$FC,$AA,$3F,$C9,$5E,$43,$9D,$F5,$98,$6D,$7F,$3A,$6B,$FC,$E9,$AF
	.dc.b	$F4,$6F,$82,$F1,$79,$9F,$E8,$C8,$FE,$A3,$18,$23,$82,$1F,$D4,$1C
	.dc.b	$B4,$C9,$D4,$FF,$05,$FE,$DA,$F8,$EF,$F3,$17,$5A,$31,$3A,$9C,$C6
	.dc.b	$23,$FB,$0F,$A3,$FE,$B7,$7F,$B4,$B5,$CF,$C3,$7C,$B8,$C4,$7F,$AF
	.dc.b	$ED,$F8,$93,$E6,$3C,$2D,$9F,$F0,$FB,$5F,$F0,$73,$5E,$5F,$32,$F9
	.dc.b	$AF,$4B,$63,$FC,$4E,$D7,$FC,$2C,$D7,$8C,$70,$23,$C1,$FC,$17,$B5
	.dc.b	$E8,$8F,$F1,$5E,$58,$23,$FE,$34,$FD,$7B,$F0,$6F,$B5,$AF,$E3,$B2
	.dc.b	$46,$23,$FC,$80,$DE,$5F,$F2,$37,$0F,$E4,$7C,$FF,$E4,$75,$FE,$54
	.dc.b	$8C,$11,$FF,$0B,$C6,$9F,$F2,$A4,$65,$FF,$33,$7C,$37,$E6,$FD,$FF
	.dc.b	$F5,$94,$7F,$99,$63,$04,$70,$43,$89,$FE,$24,$F9,$9F,$F1,$43,$7D
	.dc.b	$8F,$D3,$46,$23,$FC,$46,$D7,$FD,$F9,$F0,$2F,$D0,$DF,$2D,$FE,$AF
	.dc.b	$8C,$36,$33,$3F,$ED,$A6,$BF,$E8,$56,$BD,$1A,$FE,$EA,$79,$BC,$67
	.dc.b	$CB,$F3,$72,$3D,$4D,$7C,$ED,$7F,$DF,$F6,$8C,$47,$FA,$9F,$E0,$DF
	.dc.b	$21,$FC,$15,$B1,$FE,$FD,$8C,$11,$C1,$0F,$F7,$F3,$FC,$37,$E6,$FF
	.dc.b	$8E,$C7,$F9,$29,$AF,$F1,$43,$FC,$B0,$E1,$FF,$5C,$35,$FE,$33,$F2
	.dc.b	$BF,$C1,$5F,$F2,$FB,$5F,$F5,$83,$5E,$8D,$78,$F6,$FF,$B6,$A3,$10
	.dc.b	$69,$BE,$2F,$F7,$5F,$6F,$F2,$5F,$8C,$FE,$05,$5F,$EF,$58,$C1,$1F
	.dc.b	$F7,$E3,$5F,$F7,$E3,$FC,$B7,$E3,$BF,$DD,$91,$FE,$70,$6B,$FE,$6C
	.dc.b	$6B,$F0,$E9,$18,$8F,$F6,$D3,$5F,$F3,$F3,$63,$7F,$13,$DC,$F0,$54
	.dc.b	$AF,$03,$5F,$F6,$97,$C8,$BE,$43,$F2,$CD,$8F,$F4,$5F,$12,$31,$1C
	.dc.b	$27,$64,$77,$1A,$7F,$46,$D7,$F1,$9E,$E7,$E2,$C8,$FF,$A4,$B8,$3F
	.dc.b	$C7,$4D,$8C,$BF,$E9,$08,$CB,$63,$FA,$C7,$C9,$3E,$3A,$92,$7F,$D3
	.dc.b	$B1,$96,$C6,$5F,$F4,$DC,$65,$B1,$F5,$FE,$62,$47,$D7,$6B,$F9,$0F
	.dc.b	$97,$7C,$1F,$F6,$CE,$57,$C9,$FE,$43,$E4,$FF,$39,$18,$BF,$CC,$C6
	.dc.b	$27,$F2,$B7,$CA,$7F,$5F,$18,$BE,$8F,$38,$A7,$22,$9F,$99,$9E,$57
	.dc.b	$DE,$31,$5A,$E5,$61,$E1,$9E,$47,$80,$FD,$6D,$F2,$C3,$C5,$19,$3F
	.dc.b	$4D,$3C,$5A,$75,$9E,$0F,$2B,$F0,$4E,$EF,$C3,$3B,$98,$F5,$8C,$B6
	.dc.b	$78,$57,$48,$F5,$99,$3A,$9C,$B0,$F8,$14,$9E,$0F,$48,$CD,$74,$BE
	.dc.b	$18,$7C,$DC,$3A,$1E,$3C,$38,$9C,$AB,$D4,$71,$8D,$57,$89,$E3,$7E
	.dc.b	$2A,$E3,$39,$8F,$0C,$F8,$0F,$8C,$70,$BF,$14,$71,$9D,$4F,$CE,$34
	.dc.b	$70,$8D,$38,$34,$E0,$93,$D0,$91,$C5,$FB,$B1,$E4,$8E,$58,$7C,$9A
	.dc.b	$F0,$BF,$26,$B8,$CF,$D2,$AE,$AF,$E2,$9E,$46,$1D,$47,$09,$E5,$1E
	.dc.b	$12,$1E,$88,$D5,$FC,$88,$F1,$E1,$F3,$EF,$A5,$F2,$9E,$07,$27,$D0
	.dc.b	$F0,$3D,$35,$F6,$5F,$A5,$7F,$00,$8F,$A5,$F0,$E1,$F0,$30,$FB,$F3
	.dc.b	$CA,$38,$BB,$3C,$15,$D5,$F9,$2F,$A7,$0F,$87,$7D,$27,$F0,$E7,$94
	.dc.b	$B1,$EC,$BE,$F1,$89,$F2,$BE,$E2,$61,$EE,$AE,$11,$E4,$3E,$79,$F8
	.dc.b	$C3,$FB,$1D,$FF,$0D,$6E,$1F,$DB,$DF,$DE,$9F,$DA,$3E,$59,$F2,$F9
	.dc.b	$8F,$ED,$F1,$87,$FB,$7B,$5E,$11,$B7,$FB,$D7,$CB,$D7,$FD,$1E,$D4
	.dc.b	$D2,$2F,$98,$FE,$30,$FF,$A2,$FE,$13,$FD,$2F,$18,$8E,$26,$F4,$F5
	.dc.b	$78,$8A,$DC,$7A,$6E,$D7,$F8,$36,$B4,$F1,$3E,$14,$69,$FF,$52,$C6
	.dc.b	$AF,$C1,$3F,$CB,$A7,$C8,$7F,$23,$3D,$A5,$9E,$BA,$70,$8E,$C7,$4C
	.dc.b	$3B,$3C,$2F,$DF,$5D,$F0,$F4,$4F,$03,$BC,$F3,$37,$5C,$A3,$E1,$C6
	.dc.b	$1E,$06,$27,$E2,$CE,$F0,$F0,$DF,$7A,$F8,$27,$77,$0F,$DA,$31,$3C
	.dc.b	$5A,$76,$9E,$71,$9B,$F1,$1F,$07,$0F,$21,$70,$F1,$DF,$C9,$3C,$C7
	.dc.b	$D5,$AE,$51,$AA,$E1,$39,$38,$5F,$D3,$C3,$E7,$D7,$C8,$C3,$BB,$4E
	.dc.b	$EF,$C9,$7A,$1F,$3A,$FC,$EE,$EF,$B4,$F8,$1F,$B0,$78,$0F,$BD,$87
	.dc.b	$E0,$DF,$C9,$C7,$81,$F9,$75,$D1,$B5,$E0,$7E,$7C,$ED,$F9,$CD,$7F
	.dc.b	$16,$D7,$AC,$7A,$8F,$41,$CD,$FA,$D3,$DD,$E7,$1B,$7B,$B5,$C9,$EE
	.dc.b	$7E,$03,$EE,$9F,$8C,$FC,$F9,$AF,$C2,$7E,$23,$FA,$28,$B7,$F4,$A3
	.dc.b	$B9,$D4,$A7,$5A,$F4,$BD,$5A,$F5,$7C,$6F,$56,$C7,$58,$F4,$9C,$CE
	.dc.b	$A6,$CF,$9E,$75,$29,$D4,$A7,$EB,$23,$C2,$D3,$9C,$77,$7E,$9C,$7C
	.dc.b	$97,$F7,$0F,$D1,$69,$DC,$F2,$9E,$32,$C7,$37,$ED,$9F,$14,$CD,$FE
	.dc.b	$2F,$0F,$D5,$C7,$EB,$4F,$D7,$E1,$F4,$9F,$31,$F8,$0B,$F1,$3F,$AA
	.dc.b	$3D,$C6,$4F,$21,$E6,$32,$53,$99,$6F,$F8,$0B,$6F,$F6,$6F,$23,$B1
	.dc.b	$C0,$B1,$C1,$B1,$C1,$EE,$60,$F6,$57,$E5,$DE,$C6,$2B,$D7,$7F,$0A
	.dc.b	$7E,$05,$F5,$8C,$4F,$D8,$8C,$57,$C6,$8F,$F5,$2F,$91,$FC,$FA,$3F
	.dc.b	$D5,$5D,$68,$C4,$6B,$1F,$F5,$94,$62,$3F,$D5,$6D,$EB,$FF,$58,$B6
	.dc.b	$B9,$57,$03,$FB,$34,$BE,$6F,$FA,$94,$2F,$FD,$0A,$D9,$D3,$FD,$4B
	.dc.b	$7F,$EA,$71,$DA,$38,$46,$2B,$9D,$72,$14,$F9,$B7,$CB,$F3,$C3,$B4
	.dc.b	$61,$E6,$7F,$BE,$FC,$0D,$5F,$29,$D3,$D6,$FA,$5E,$9B,$7C,$A7,$C3
	.dc.b	$D4,$C3,$58,$72,$CB,$D6,$BE,$6B,$27,$3D,$1E,$46,$32,$E9,$9F,$AE
	.dc.b	$B1,$9F,$8E,$F7,$5A,$C3,$A3,$C2,$1F,$03,$F5,$49,$1F,$EE,$D6,$D7
	.dc.b	$FB,$B5,$AF,$E1,$1F,$BD,$4F,$8F,$F5,$9E,$78,$B5,$E1,$7D,$E7,$C5
	.dc.b	$F9,$AF,$F6,$84,$64,$97,$F2,$EF,$CB,$1E,$E9,$DC,$E6,$BE,$4F,$D4
	.dc.b	$CA,$F8,$7E,$6B,$E6,$31,$7C,$EA,$EF,$F8,$8A,$D5,$FC,$59,$FB,$1E
	.dc.b	$97,$C8,$BF,$F1,$F6,$E8,$C4,$7E,$54,$FC,$5B,$87,$F2,$CD,$5E,$93
	.dc.b	$E7,$9E,$57,$E2,$A3,$0F,$83,$8F,$13,$AC,$BE,$EA,$F0,$FF,$18,$D8
	.dc.b	$FC,$65,$F2,$77,$3A,$23,$A9,$E9,$87,$E5,$DE,$4F,$E6,$7D,$8F,$46
	.dc.b	$66,$93,$1F,$BF,$F6,$BF,$98,$3F,$3D,$1F,$1F,$7F,$CC,$C7,$E8,$7D
	.dc.b	$B7,$D7,$FF,$03,$6E,$7F,$C0,$1B,$97,$CE,$6B,$F3,$72,$FD,$B7,$17
	.dc.b	$E9,$8E,$97,$6F,$F3,$91,$6B,$F6,$57,$D3,$1F,$D5,$CE,$5B,$5F,$A8
	.dc.b	$9D,$3F,$C0,$27,$F7,$1F,$00,$FA,$2F,$E3,$4D,$9F,$E2,$ED,$C3,$F0
	.dc.b	$AD,$7D,$1C,$EF,$9B,$FD,$E9,$D2,$ED,$FF,$03,$F8,$17,$DC,$BD,$6F
	.dc.b	$E5,$51,$86,$C6,$67,$FD,$19,$39,$6C,$67,$1F,$F4,$A4,$E5,$B1,$97
	.dc.b	$F3,$11,$97,$CF,$C5,$FC,$48,$FE,$3B,$1C,$76,$FE,$27,$1F,$C8,$3D
	.dc.b	$37,$CD,$BE,$6F,$FA,$04,$61,$B1,$9C,$BF,$D4,$53,$96,$C6,$71,$FF
	.dc.b	$55,$4E,$5B,$19,$7F,$5D,$19,$7F,$61,$F4,$8F,$AA,$F9,$D7,$FC,$B6
	.dc.b	$31,$BF,$9F,$0E,$D9,$FF,$A6,$27,$35,$BB,$E3,$87,$F4,$59,$CB,$FD
	.dc.b	$1F,$1F,$25,$F1,$AD,$BF,$D4,$7E,$71,$F0,$1F,$D8,$61,$18,$C7,$FD
	.dc.b	$85,$18,$9D,$4E,$70,$FF,$65,$4E,$6B,$73,$99,$FE,$1D,$39,$77,$19
	.dc.b	$7F,$B3,$60,$F4,$DF,$05,$F5,$DB,$87,$F5,$1C,$39,$1F,$EB,$AA,$E1
	.dc.b	$5F,$D7,$3D,$2F,$4F,$77,$F7,$B6,$DF,$FB,$6D,$B8,$FF,$B8,$5A,$FF
	.dc.b	$23,$7E,$0F,$CA,$71,$CC,$D2,$57,$FB,$7F,$E5,$9F,$18,$F4,$B0,$DC
	.dc.b	$7F,$DD,$53,$A9,$C9,$C7,$E2,$FF,$96,$CE,$9A,$FF,$BD,$65,$18,$8F
	.dc.b	$E7,$11,$8F,$13,$D9,$F2,$4F,$95,$FF,$79,$6A,$8C,$5E,$A7,$31,$88
	.dc.b	$F8,$4B,$DC,$C9,$23,$AB,$FE,$FD,$E2,$7D,$3B,$E4,$7E,$EE,$FF,$E1
	.dc.b	$88,$C5,$FD,$76,$BF,$E1,$93,$C9,$F2,$9C,$1F,$F1,$36,$28,$FE,$A9
	.dc.b	$12,$30,$E0,$DC,$56,$96,$36,$ED,$AF,$38,$E0,$F8,$8E,$9D,$2F,$A7
	.dc.b	$7D,$37,$96,$30,$D8,$CC,$FF,$88,$C7,$CA,$7B,$B4,$FD,$B4,$EA,$7F
	.dc.b	$68,$D8,$CD,$7C,$19,$E3,$A5,$BF,$8F,$1F,$09,$AF,$F7,$76,$BF,$E4
	.dc.b	$D6,$BC,$8F,$88,$E0,$91,$F1,$FC,$AF,$F2,$3F,$CD,$7D,$16,$D7,$F8
	.dc.b	$CB,$DA,$4F,$97,$E0,$9F,$4F,$4F,$92,$E2,$BE,$93,$5F,$AC,$D7,$ED
	.dc.b	$B5,$FF,$08,$35,$FF,$33,$35,$39,$BF,$F2,$F7,$17,$9F,$CA,$FC,$B5
	.dc.b	$7F,$90,$E3,$D5,$E5,$BE,$4E,$9E,$F7,$13,$FD,$82,$30,$7C,$3A,$7C
	.dc.b	$4F,$52,$4F,$F6,$DF,$86,$AF,$F8,$EE,$2F,$0F,$F9,$E3,$D0,$56,$3A
	.dc.b	$3F,$CE,$35,$6F,$FC,$EF,$5A,$7A,$1F,$E0,$9A,$FA,$91,$98,$FF,$05
	.dc.b	$47,$97,$C0,$F8,$04,$8F,$F3,$3F,$A7,$FC,$51,$B1,$F7,$63,$11,$FE
	.dc.b	$40,$6B,$FE,$13,$6B,$FE,$93,$F5,$1F,$05,$F8,$36,$C7,$F8,$E2,$3F
	.dc.b	$6D,$DE,$FC,$0D,$F2,$FB,$99,$5B,$3F,$C7,$73,$A6,$CF,$28,$C4,$7F
	.dc.b	$A6,$EB,$A4,$FE,$4F,$D9,$F4,$EF,$FE,$A4,$7C,$F5,$FE,$56,$CF,$F3
	.dc.b	$51,$8E,$4F,$F2,$E4,$7D,$BF,$07,$F0,$C4,$AF,$F3,$5F,$A3,$5F,$F3
	.dc.b	$A4,$5C,$3F,$CD,$53,$F8,$DF,$84,$95,$FE,$8F,$7F,$4D,$F1,$FF,$E1
	.dc.b	$D8,$C1,$0F,$F3,$F4,$6B,$D6,$FC,$B7,$99,$27,$FD,$30,$FE,$BF,$E3
	.dc.b	$FF,$83,$73,$FF,$B0,$DF,$DD,$EA,$69,$3C,$0F,$DC,$36,$FF,$D9,$ED
	.dc.b	$7F,$CF,$BD,$3F,$61,$FF,$48,$72,$7E,$9C,$FD,$F7,$C1,$7F,$9E,$23
	.dc.b	$FD,$55,$19,$69,$FE,$AB,$F7,$7F,$5D,$F3,$72,$3F,$D6,$0F,$F0,$7F
	.dc.b	$86,$7C,$6E,$67,$FD,$3D,$37,$8F,$FD,$B7,$5A,$9F,$F5,$C7,$C7,$FE
	.dc.b	$E4,$8F,$FA,$E7,$83,$CA,$7F,$15,$CF,$FD,$0D,$1F,$EE,$C6,$BF,$E9
	.dc.b	$A6,$BF,$EF,$C8,$E0,$47,$83,$FC,$FB,$D6,$FE,$17,$87,$FB,$CF,$E2
	.dc.b	$FE,$34,$7E,$BD,$BD,$FF,$EA,$37,$3E,$2A,$B8,$6F,$FE,$A7,$BE,$78
	.dc.b	$AF,$8B,$A3,$FE,$00,$47,$59,$E1,$7C,$E7,$84,$62,$F8,$47,$F8,$87
	.dc.b	$0F,$14,$F7,$6C,$74,$AE,$58,$79,$AF,$8E,$1D,$EB,$C3,$7C,$1F,$F1
	.dc.b	$65,$F0,$C3,$C1,$1E,$23,$67,$29,$E1,$5E,$EC,$3B,$4E,$CF,$31,$B3
	.dc.b	$C3,$7D,$5F,$36,$19,$78,$57,$95,$F1,$C7,$03,$55,$F2,$A3,$75,$CD
	.dc.b	$D9,$CA,$FC,$A7,$9A,$35,$7E,$7C,$3E,$45,$74,$7D,$F7,$EA,$3B,$11
	.dc.b	$C9,$23,$E1,$9D,$CF,$31,$E4,$C0,$BE,$33,$D0,$49,$F9,$AF,$C8,$31
	.dc.b	$1D,$1C,$47,$78,$F9,$67,$73,$B1,$23,$8B,$A7,$8B,$F3,$8E,$15,$EB
	.dc.b	$70,$FB,$E7,$E1,$C6,$9F,$05,$FE,$6A,$BA,$BF,$D4,$5E,$B3,$F4,$1C
	.dc.b	$E3,$F0,$23,$E0,$57,$B8,$E7,$1F,$52,$3E,$3D,$F2,$BE,$95,$E2,$BC
	.dc.b	$63,$FA,$37,$AD,$FC,$C7,$38,$FD,$98,$FA,$78,$7A,$DF,$81,$3C,$1F
	.dc.b	$19,$8B,$F2,$3F,$3D,$F0,$18,$7F,$17,$3F,$1D,$FA,$2F,$57,$E0,$47
	.dc.b	$07,$91,$D4,$87,$D3,$74,$F0,$3E,$8B,$F4,$23,$B6,$1F,$95,$9E,$2F
	.dc.b	$C4,$7A,$9C,$48,$E5,$EC,$70,$C3,$F3,$F8,$FC,$F8,$FC,$3D,$7C,$F8
	.dc.b	$FA,$D8,$7D,$CA,$F9,$D8,$7C,$1C,$3A,$12,$33,$87,$A0,$F9,$F3,$E9
	.dc.b	$3F,$0F,$1A,$8F,$19,$27,$F1,$53,$F8,$28,$FB,$07,$CE,$70,$FE,$44
	.dc.b	$EC,$64,$EE,$43,$04,$8F,$E1,$04,$9F,$DF,$9F,$0A,$3B,$C7,$CD,$78
	.dc.b	$5F,$C5,$C3,$EE,$4F,$9A,$7E,$7B,$F0,$71,$FC,$86,$1E,$6A,$FA,$F8
	.dc.b	$76,$9E,$B8,$7F,$0F,$7F,$81,$E1,$FE,$96,$C3,$FA,$1B,$6B,$F9,$F3
	.dc.b	$67,$FD,$2D,$2F,$8D,$7A,$58,$6D,$7F,$A7,$67,$53,$93,$FD,$2F,$0F
	.dc.b	$8D,$A4,$7F,$D3,$6D,$CB,$FC,$05,$1F,$ED,$3E,$8F,$E9,$AF,$F1,$89
	.dc.b	$BE,$B2,$73,$3D,$4F,$E7,$C7,$12,$99,$AE,$B3,$C6,$33,$5F,$29,$FE
	.dc.b	$0D,$19,$DB,$C7,$1D,$5E,$37,$B8,$FE,$19,$5E,$48,$F1,$B6,$3A,$9E
	.dc.b	$4A,$F0,$5F,$49,$D3,$67,$55,$C9,$A7,$28,$F0,$B7,$6F,$F6,$7D,$F4
	.dc.b	$7F,$D9,$D7,$D1,$B3,$CA,$BB,$1C,$8B,$1C,$AF,$C1,$87,$63,$C8,$58
	.dc.b	$E7,$3C,$5A,$71,$AF,$7C,$ED,$FF,$71,$DF,$C7,$7E,$A4,$74,$9E,$0D
	.dc.b	$8F,$A3,$7B,$7E,$C4,$F5,$BF,$3C,$68,$EF,$1C,$8F,$7D,$78,$63,$EA
	.dc.b	$9F,$CB,$8F,$EC,$38,$7D,$D9,$D9,$C6,$3C,$E7,$C7,$7D,$AF,$7A,$F3
	.dc.b	$3C,$AF,$CD,$87,$B6,$FA,$4F,$2C,$3C,$4F,$E6,$4F,$11,$F0,$B0,$EA
	.dc.b	$F3,$3E,$E1,$DD,$F2,$BE,$F8,$F0,$BC,$4E,$D1,$E9,$7D,$F1,$E0,$7E
	.dc.b	$49,$CC,$F6,$E1,$F6,$5E,$27,$59,$EE,$F8,$E7,$BD,$7B,$23,$AC,$F7
	.dc.b	$7D,$95,$CD,$F9,$ED,$7E,$05,$78,$EF,$E3,$61,$F2,$27,$E1,$E1,$D5
	.dc.b	$DE,$1D,$5F,$49,$F1,$70,$FC,$0B,$63,$E4,$BF,$BF,$8C,$B6,$3D,$6F
	.dc.b	$C7,$3D,$27,$BC,$F8,$6F,$E3,$4E,$73,$FA,$B8,$F8,$A7,$C3,$8E,$71
	.dc.b	$E7,$8F,$E0,$91,$E7,$BF,$6C,$7D,$E3,$D2,$78,$CE,$E7,$D3,$3E,$3C
	.dc.b	$78,$5C,$9E,$D2,$9C,$4D,$9C,$4F,$A2,$F1,$72,$71,$29,$C4,$D9,$C4
	.dc.b	$F2,$CF,$BD,$DB,$5D,$9C,$CA,$73,$38,$DF,$F1,$7B,$FC,$04,$60,$FC
	.dc.b	$D9,$C4,$C3,$F5,$A7,$B4,$62,$BE,$8D,$73,$7D,$91,$E0,$9F,$A9,$D2
	.dc.b	$E1,$75,$FF,$3B,$AF,$F1,$F4,$E6,$31,$1F,$E6,$08,$C6,$1F,$EB,$59
	.dc.b	$CC,$A7,$2F,$FA,$DE,$33,$18,$8E,$B5,$E0,$AF,$F2,$64,$75,$F8,$DF
	.dc.b	$EA,$91,$FE,$6E,$E0,$FE,$9D,$69,$CD,$7F,$AD,$6B,$C9,$7C,$E3,$07
	.dc.b	$8E,$77,$7C,$BA,$5E,$77,$F5,$98,$C3,$70,$F7,$73,$BE,$37,$18,$FB
	.dc.b	$70,$E3,$5E,$53,$E2,$38,$8F,$0E,$68,$C6,$3F,$EE,$4B,$E7,$39,$C6
	.dc.b	$D3,$E3,$69,$AD,$1E,$0E,$33,$F0,$67,$E3,$AC,$3E,$0A,$DC,$EB,$1F
	.dc.b	$15,$73,$F8,$BF,$A7,$46,$92,$BF,$DE,$0D,$BF,$F7,$83,$5F,$C2,$E2
	.dc.b	$F8,$DF,$49,$D6,$BC,$11,$96,$D7,$B8,$F8,$3F,$1A,$F4,$71,$3A,$26
	.dc.b	$1F,$2A,$75,$18,$DF,$ED,$7C,$67,$C8,$3E,$5F,$C6,$BE,$2F,$15,$E2
	.dc.b	$9B,$87,$D4,$AD,$4E,$4F,$8D,$E5,$FC,$89,$FC,$9F,$02,$31,$1F,$A4
	.dc.b	$6B,$FA,$36,$AF,$35,$E7,$BC,$7F,$C4,$C6,$1F,$C0,$D7,$7D,$3F,$C5
	.dc.b	$53,$9C,$7E,$C5,$BC,$4F,$C3,$B6,$3F,$03,$3B,$C7,$F1,$F3,$96,$D7
	.dc.b	$03,$C3,$1E,$3A,$EB,$1F,$92,$F4,$3D,$39,$9A,$4E,$6F,$DC,$F1,$BE
	.dc.b	$37,$F2,$E7,$18,$B8,$7E,$3F,$D2,$7C,$6F,$F0,$06,$C7,$EF,$B4,$FE
	.dc.b	$0D,$18,$7E,$7B,$5F,$C6,$F1,$FE,$90,$FD,$3C,$6F,$7F,$DD,$46,$6B
	.dc.b	$F1,$EE,$AF,$17,$FB,$2C,$BE,$FB,$67,$55,$C0,$FD,$37,$AC,$F8,$DF
	.dc.b	$E2,$ED,$8F,$46,$C8,$C5,$FE,$19,$AF,$E9,$3E,$2B,$F7,$67,$EF,$A3
	.dc.b	$79,$FF,$11,$8F,$E1,$7F,$18,$F8,$D7,$A7,$FC,$AA,$30,$D8,$CE,$5F
	.dc.b	$26,$72,$D8,$CE,$3F,$CA,$A7,$2D,$8C,$BF,$EF,$98,$CB,$FB,$AE,$0F
	.dc.b	$E2,$27,$92,$79,$E3,$FC,$9E,$3F,$90,$FC,$4B,$E3,$1F,$1B,$FD,$02
	.dc.b	$30,$D8,$CE,$5F,$7E,$72,$D8,$CE,$3F,$D0,$27,$2D,$8C,$A2,$46,$9C
	.dc.b	$0F,$83,$7C,$6B,$6F,$E6,$F1,$87,$D3,$7E,$0C,$FF,$D3,$13,$9C,$7F
	.dc.b	$95,$B7,$0F,$E9,$2D,$9F,$F4,$F6,$3E,$67,$8E,$3D,$EF,$F9,$9C,$3E
	.dc.b	$2D,$F1,$BF,$D9,$23,$0D,$8C,$CF,$E8,$E7,$2D,$8C,$E3,$FD,$92,$72
	.dc.b	$D8,$CB,$F0,$A3,$2F,$C4,$C1,$E9,$3E,$37,$EC,$B7,$0F,$3E,$1D,$B0
	.dc.b	$F6,$CE,$70,$FE,$D3,$E9,$FA,$FB,$7F,$BE,$B7,$0F,$DD,$36,$7F,$BC
	.dc.b	$B5,$FA,$AF,$C3,$F7,$B8,$E2,$69,$2B,$F8,$25,$3E,$2D,$F1,$AF,$4B
	.dc.b	$31,$86,$C6,$67,$F8,$64,$E5,$BF,$1B,$FC,$CD,$BE,$27,$28,$C4,$7F
	.dc.b	$84,$25,$18,$8F,$C8,$66,$F8,$C7,$C1,$79,$A3,$0D,$8C,$CF,$C4,$6B
	.dc.b	$F1,$1F,$EB,$CE,$7C,$CE,$04,$7F,$05,$F1,$AF,$82,$FB,$9B,$7F,$88
	.dc.b	$63,$13,$F8,$E6,$BF,$E1,$E3,$CD,$F1,$BC,$1F,$D4,$60,$8F,$9B,$1A
	.dc.b	$3E,$7C,$65,$B1,$A8,$C4,$7F,$37,$6B,$FE,$36,$6B,$FB,$26,$BF,$B3
	.dc.b	$F8,$D7,$C6,$F8,$FA,$D1,$89,$FD,$7B,$5F,$57,$C6,$FB,$3A,$3F,$6B
	.dc.b	$39,$6D,$F3,$8C,$67,$FC,$0A,$B8,$37,$4F,$7C,$7C,$6C,$3F,$C9,$AD
	.dc.b	$74,$7E,$41,$C3,$FC,$3D,$A9,$A6,$BC,$5F,$F2,$4F,$C6,$FC,$F6,$CF
	.dc.b	$F8,$BF,$D8,$4F,$D3,$87,$C1,$3E,$37,$0D,$8F,$D0,$B5,$FE,$CC,$D7
	.dc.b	$FC,$CA,$D7,$F9,$0B,$5F,$E4,$CD,$7B,$C7,$E0,$23,$57,$FE,$53,$E2
	.dc.b	$F4,$79,$9F,$94,$BF,$EC,$D1,$EB,$F2,$9F,$1B,$4F,$DA,$DF,$F6,$16
	.dc.b	$C6,$AB,$FB,$2B,$5F,$F1,$4D,$BE,$37,$B9,$23,$FC,$0D,$F0,$D5,$FE
	.dc.b	$ED,$17,$8F,$FC,$FF,$E8,$AB,$1C,$5F,$E6,$EA,$B7,$FE,$0A,$9F,$A5
	.dc.b	$5B,$8F,$EE,$ED,$7F,$C1,$B1,$EA,$8C,$47,$F9,$6B,$C0,$F8,$64,$7F
	.dc.b	$CD,$1E,$9F,$F1,$76,$C7,$C2,$8C,$47,$F9,$19,$AF,$FA,$A1,$AF,$FA
	.dc.b	$9F,$D4,$7C,$17,$E1,$DB,$1F,$E3,$E8,$FD,$C7,$82,$F8,$D7,$C6,$EE
	.dc.b	$65,$6C,$FF,$20,$CE,$9B,$3C,$A3,$11,$FE,$A2,$AE,$93,$F9,$7F,$63
	.dc.b	$D3,$BF,$FA,$B1,$F3,$D7,$FA,$E5,$BC,$7F,$E6,$1D,$3F,$03,$18,$E1
	.dc.b	$FE,$3B,$7F,$EA,$BF,$07,$F1,$67,$E8,$FD,$AF,$F4,$BE,$3F,$E8,$28
	.dc.b	$CB,$4F,$F4,$A4,$69,$FF,$1B,$7C,$27,$E5,$08,$FF,$A6,$7E,$37,$FC
	.dc.b	$4A,$FF,$A3,$23,$2F,$03,$5E,$BF,$E7,$3C,$C9,$1F,$E9,$27,$F6,$DF
	.dc.b	$1B,$FC,$1A,$3F,$DA,$CF,$EE,$C9,$E0,$FE,$79,$B1,$FE,$40,$6B,$FE
	.dc.b	$84,$7F,$81,$69,$F6,$A3,$FC,$9D,$C7,$FA,$C3,$F8,$27,$C1,$7F,$A0
	.dc.b	$63,$FD,$47,$19,$69,$FE,$B5,$8D,$3F,$E5,$7F,$8C,$FD,$A9,$1F,$F5
	.dc.b	$D7,$C3,$3E,$37,$11,$B3,$FD,$5F,$3F,$C4,$E7,$51,$FE,$B2,$F8,$DF
	.dc.b	$DF,$91,$F0,$69,$E6,$3F,$7B,$87,$FA,$3E,$3F,$CB,$ED,$7F,$D4,$6D
	.dc.b	$7F,$DF,$31,$C0,$8F,$07,$FA,$7F,$A7,$FC,$83,$8F,$FC,$75,$F1,$5F
	.dc.b	$1E,$3F,$68,$DE,$EF,$F1,$DB,$9F,$11,$5C,$36,$FF,$1F,$DF,$3C,$57
	.dc.b	$C5,$D1,$FE,$47,$8E,$B3,$C2,$F9,$CF,$08,$C5,$F0,$8F,$F1,$3E,$1E
	.dc.b	$29,$EE,$D8,$E9,$5C,$B0,$F3,$5F,$1C,$3B,$D7,$86,$F8,$3F,$E3,$2B
	.dc.b	$E1,$87,$82,$3C,$46,$CE,$53,$C2,$BD,$D8,$76,$9D,$9E,$63,$67,$86
	.dc.b	$FA,$BE,$6C,$32,$F0,$AF,$2B,$E3,$8E,$06,$AB,$E5,$46,$EB,$9B,$B3
	.dc.b	$95,$F9,$4F,$34,$6A,$FC,$F8,$7C,$8A,$E8,$FB,$EF,$D4,$76,$23,$92
	.dc.b	$47,$C3,$3B,$9E,$63,$C9,$87,$FA,$0F,$E3,$3C,$E4,$9F,$98,$FC,$73
	.dc.b	$13,$CD,$C4,$76,$8F,$94,$76,$3A,$92,$38,$BA,$78,$BF,$34,$E1,$5E
	.dc.b	$A7,$0F,$BA,$7E,$14,$69,$F0,$5F,$E6,$6B,$AB,$FD,$55,$EB,$3F,$3D
	.dc.b	$CE,$3F,$E9,$E7,$33,$F3,$A7,$E3,$5F,$85,$FA,$CE,$AB,$85,$F8,$AF
	.dc.b	$8E,$3F,$A2,$7A,$DF,$CB,$9F,$92,$79,$AB,$EF,$47,$D1,$BE,$8E,$CF
	.dc.b	$99,$3C,$1F,$89,$5D,$13,$E0,$C7,$F0,$71,$E8,$9F,$87,$87,$8D,$F3
	.dc.b	$12,$39,$BD,$5F,$A4,$E5,$F8,$CF,$9D,$EA,$7C,$27,$86,$3F,$35,$E0
	.dc.b	$FD,$17,$38,$FC,$69,$F5,$9E,$03,$ED,$C7,$DA,$C7,$E7,$57,$D1,$9F
	.dc.b	$09,$C0,$F2,$9C,$4C,$98,$23,$E1,$7A,$9F,$0C,$87,$A4,$E4,$FA,$5E
	.dc.b	$CF,$17,$47,$E6,$A3,$CB,$7F,$B3,$3F,$69,$7F,$6A,$7F,$3B,$3F,$95
	.dc.b	$7B,$D7,$F0,$43,$D8,$49,$E1,$3B,$9E,$F3,$B9,$FE,$03,$87,$E1,$70
	.dc.b	$F8,$C4,$AF,$69,$F0,$AB,$D4,$FC,$2C,$3F,$00,$EA,$3A,$3C,$CC,$5F
	.dc.b	$8A,$BF,$0F,$7F,$EC,$6A,$FE,$85,$43,$F9,$61,$4F,$DB,$14,$FF,$71
	.dc.b	$FC,$6B,$D1,$E4,$D8,$D1,$A8,$C9,$FE,$A8,$5F,$1B,$48,$FF,$AB,$1A
	.dc.b	$FF,$B6,$9A,$F4,$6B,$E3,$6B,$E7,$6B,$EB,$6B,$E0,$3B,$10,$EC,$53
	.dc.b	$B1,$4E,$C5,$3B,$14,$EC,$53,$B1,$D0,$87,$42,$9D,$0A,$74,$29,$D0
	.dc.b	$A7,$43,$91,$0E,$45,$39,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70
	.dc.b	$34,$43,$45,$34,$60,$86,$08,$FD,$42,$3F,$25,$AF,$C0,$6B,$E9,$6B
	.dc.b	$E4,$6B,$E1,$6B,$E0,$6B,$E0,$3B,$10,$EC,$53,$B1,$4E,$C5,$3B,$14
	.dc.b	$EC,$53,$B1,$D0,$87,$42,$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0E,$45
	.dc.b	$39,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60
	.dc.b	$86,$08,$FD,$42,$3F,$25,$AF,$C0,$6B,$E9,$6B,$E4,$6B,$E1,$6B,$E0
	.dc.b	$6B,$E0,$3B,$10,$EC,$53,$B1,$4E,$C5,$3B,$14,$EC,$53,$B1,$D0,$87
	.dc.b	$42,$9D,$0A,$74,$29,$D0,$A7,$43,$91,$0E,$45,$39,$14,$E4,$53,$91
	.dc.b	$FC,$02,$95,$FC,$E7,$EA,$1F,$35,$FE,$86,$8C,$37,$0F,$E8,$D3,$AB
	.dc.b	$FF,$43,$56,$9B,$3A,$AE,$10,$AE,$0F,$F4,$53,$11,$A8,$CD,$FF,$54
	.dc.b	$9F,$EA,$64,$3F,$A9,$94,$FF,$36,$1F,$D5,$5E,$0E,$88,$68,$E0,$E9
	.dc.b	$C1,$A7,$04,$D3,$FD,$3F,$1D,$5A,$F2,$6B,$C5,$AF,$42,$47,$3C,$BF
	.dc.b	$B8,$4E,$A3,$13,$FD,$A1,$AF,$79,$D9,$1D,$4F,$12,$46,$63,$81,$A7
	.dc.b	$0F,$9D,$AF,$9D,$AF,$9A,$39,$BC,$48,$74,$21,$CC,$C9,$4C,$9D,$4C
	.dc.b	$94,$C9,$DE,$3C,$33,$F0,$23,$91,$0E,$05,$38,$1D,$0A,$74,$79,$3C
	.dc.b	$08,$68,$E4,$E9,$C1,$A7,$04,$9F,$87,$1D,$5A,$F2,$6B,$C5,$AF,$42
	.dc.b	$47,$39,$F7,$CE,$A3,$13,$EB,$6B,$DE,$76,$47,$53,$C4,$91,$98,$E0
	.dc.b	$69,$C3,$E7,$6B,$E7,$6B,$E6,$8E,$6F,$12,$1D,$08,$73,$32,$53,$27
	.dc.b	$53,$25,$32,$77,$8F,$0C,$FC,$08,$E4,$43,$81,$4E,$07,$42,$9D,$1E
	.dc.b	$4F,$02,$1A,$39,$3A,$70,$69,$C1,$27,$E1,$C7,$56,$BC,$9A,$F1,$6B
	.dc.b	$D0,$91,$CE,$7D,$F3,$A8,$C4,$FA,$DA,$F7,$9D,$91,$D4,$F1,$24,$66
	.dc.b	$38,$1A,$70,$F9,$DA,$FF,$36,$3F,$88,$63,$FE,$79,$7F,$95,$35,$FE
	.dc.b	$72,$63,$1F,$F3,$DB,$C1,$AE,$4C,$10,$C1,$0E,$64,$3C,$46,$4A,$64
	.dc.b	$FE,$8C,$64,$A6,$6F,$FA,$CC,$78,$A3,$13,$E9,$79,$90,$E2,$59,$E2
	.dc.b	$F9,$CE,$33,$CD,$D4,$F3,$78,$92,$78,$BE,$56,$BC,$8E,$87,$29,$E8
	.dc.b	$F1,$9E,$93,$C2,$34,$70,$35,$3C,$23,$51,$83,$4E,$09,$3D,$09,$3C
	.dc.b	$5B,$3C,$1A,$FA,$5A,$FA,$4C,$47,$12,$3C,$5A,$E8,$C1,$0C,$10,$E8
	.dc.b	$43,$DE,$64,$A6,$4F,$71,$92,$99,$9F,$7C,$78,$A3,$13,$EA,$79,$90
	.dc.b	$E2,$59,$E2,$F9,$CE,$33,$CD,$D4,$F3,$78,$92,$78,$BE,$56,$BC,$8E
	.dc.b	$87,$29,$E8,$F1,$9E,$93,$C2,$34,$70,$35,$3C,$23,$51,$83,$4E,$09
	.dc.b	$3D,$09,$3C,$5B,$3C,$1A,$FA,$5A,$FA,$4C,$47,$12,$3C,$5A,$E8,$C1
	.dc.b	$0C,$10,$E8,$43,$DE,$64,$A6,$4F,$71,$92,$99,$9F,$7C,$78,$A3,$13
	.dc.b	$EA,$79,$90,$E2,$59,$E2,$F9,$CE,$33,$CD,$D4,$F3,$78,$92,$78,$BE
	.dc.b	$56,$BC,$8E,$87,$29,$E8,$F1,$9E,$93,$C2,$34,$70,$35,$3C,$23,$51
	.dc.b	$83,$4E,$09,$3D,$09,$3C,$5B,$3C,$1A,$FA,$5A,$FA,$4C,$47,$12,$3C
	.dc.b	$5A,$E8,$C1,$0C,$10,$E8,$43,$DE,$64,$A6,$4F,$71,$92,$99,$9F,$7C
	.dc.b	$78,$A3,$13,$EA,$79,$90,$E2,$59,$E2,$F9,$CE,$33,$CD,$D4,$F3,$78
	.dc.b	$92,$78,$BE,$56,$BC,$8E,$87,$29,$E8,$F1,$9E,$93,$C2,$34,$70,$35
	.dc.b	$3C,$23,$51,$83,$4E,$09,$3D,$09,$3C,$5B,$3C,$1A,$FA,$5A,$FA,$4C
	.dc.b	$47,$CD,$F8,$F1,$00

;--------------------------------------------------------------------------------
;	FSIN.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsin_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FSIN.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSIN.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsin_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1953+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsin.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSIN.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1953+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsin.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSIN.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsin_data_compressed::
	.dc.b	$00,$02,$FA,$EC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$F3,$FE,$4D,$CD,$F8,$3E,$27,$13,$D5,$F5,$7A,$BB
	.dc.b	$AF,$C3,$52,$F8,$FA,$59,$F4,$B5,$E9,$6F,$D2,$E1,$E9,$71,$F4,$B9
	.dc.b	$7A,$5C,$FD,$2E,$9E,$97,$5F,$4B,$B7,$A5,$DF,$D2,$F0,$7A,$5E,$1F
	.dc.b	$4B,$C5,$E9,$7A,$7F,$58,$FF,$77,$F1,$1F,$F5,$4B,$7E,$37,$5F,$15
	.dc.b	$BF,$80,$E1,$EE,$71,$F5,$F9,$7A,$FC,$FD,$6E,$9E,$B7,$5F,$5B,$B7
	.dc.b	$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$78,$FC,$2F,$27,$AD,$E5
	.dc.b	$F5,$BC,$DE,$B7,$E0,$FD,$AF,$CA,$7C,$7F,$E5,$7C,$7F,$F6,$B7,$C4
	.dc.b	$FA,$FD,$9C,$CE,$BE,$27,$7F,$0B,$C3,$E0,$B8,$FB,$DC,$BC,$DE,$7E
	.dc.b	$6F,$4F,$37,$AF,$9B,$DB,$CD,$EF,$E6,$F8,$3C,$DF,$0F,$9B,$E2,$F3
	.dc.b	$7C,$7E,$6F,$93,$CD,$F2,$F9,$BE,$6F,$37,$CF,$E6,$FA,$3C,$DF,$D6
	.dc.b	$7C,$AF,$FB,$60,$FF,$48,$9F,$E4,$B3,$FC,$06,$7E,$D4,$FD,$A7,$49
	.dc.b	$79,$1C,$BC,$8E,$7E,$47,$4F,$23,$AF,$91,$DB,$C8,$EF,$E4,$78,$3C
	.dc.b	$8F,$0F,$91,$E2,$F2,$3C,$7E,$47,$93,$C8,$F2,$F9,$1F,$9D,$FD,$67
	.dc.b	$FD,$AB,$C0,$F2,$7D,$3E,$3F,$F0,$5F,$97,$7C,$AE,$FE,$27,$87,$C7
	.dc.b	$71,$F8,$CE,$5F,$0F,$CF,$E1,$7A,$7C,$2F,$5F,$85,$ED,$F0,$BD,$FE
	.dc.b	$17,$C1,$F0,$BE,$1F,$85,$F1,$7C,$2F,$8F,$E1,$7C,$9F,$0B,$E5,$F8
	.dc.b	$5F,$37,$C2,$F9,$FE,$17,$D1,$F0,$9E,$9F,$84,$F5,$7B,$DF,$B7,$FD
	.dc.b	$6B,$F8,$E7,$C2,$17,$C0,$67,$E0,$35,$F0,$1B,$F8,$0E,$1F,$01,$C7
	.dc.b	$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7,$C0,$77,$F8,$0F
	.dc.b	$07,$C0,$78,$7E,$03,$C5,$F0,$1F,$27,$C6,$F9,$5E,$37,$97,$E0,$3E
	.dc.b	$67,$8F,$F3,$7C,$7F,$9D,$E3,$FC,$FF,$1F,$E8,$78,$FF,$47,$C7,$FA
	.dc.b	$5E,$3F,$D3,$F1,$FE,$A7,$8F,$F5,$7C,$7F,$AD,$E3,$FD,$7F,$1F,$EC
	.dc.b	$78,$FF,$67,$C7,$FB,$5E,$3F,$DB,$F1,$FE,$E7,$8F,$F7,$7C,$7F,$E8
	.dc.b	$5F,$33,$FA,$F3,$FE,$00,$1F,$EB,$83,$FC,$56,$7F,$65,$3F,$91,$1F
	.dc.b	$C7,$BD,$C2,$F0,$79,$F8,$3D,$3C,$1E,$BE,$0F,$6F,$07,$BF,$83,$E0
	.dc.b	$F0,$7C,$3E,$0F,$8B,$C1,$F1,$F7,$F9,$3E,$07,$CB,$F0,$3E,$6F,$81
	.dc.b	$F3,$FC,$0F,$A3,$E0,$7F,$01,$5F,$D6,$FC,$57,$CA,$5F,$8B,$FC,$26
	.dc.b	$3F,$85,$C7,$F0,$D8,$FE,$1F,$1F,$E7,$D8,$FF,$26,$3F,$92,$8F,$87
	.dc.b	$CF,$C3,$EB,$E1,$F7,$F0,$FC,$3E,$1F,$8F,$C3,$F2,$F8,$7E,$7F,$0F
	.dc.b	$D3,$E1,$FA,$FC,$3F,$6F,$87,$EF,$F0,$FE,$0F,$87,$F0,$FC,$3F,$8B
	.dc.b	$E1,$FC,$7F,$0F,$E4,$F8,$7F,$2F,$C2,$F9,$BE,$17,$FA,$A7,$EB,$7F
	.dc.b	$B4,$F2,$7F,$A7,$2F,$73,$3D,$FA,$F7,$37,$EE,$70,$F7,$38,$FB,$9C
	.dc.b	$BD,$CE,$7E,$E7,$4F,$73,$AF,$B9,$F2,$FC,$0F,$99,$E0,$78,$3D,$CF
	.dc.b	$9D,$E0,$FC,$FF,$07,$E8,$78,$3F,$47,$C1,$FA,$5E,$0F,$D3,$F0,$7E
	.dc.b	$A7,$83,$F5,$7C,$1F,$AD,$E0,$FD,$7F,$07,$EC,$78,$3F,$67,$C1,$FB
	.dc.b	$5E,$0F,$DB,$F0,$7E,$E7,$83,$F7,$7C,$1F,$BD,$E0,$FF,$86,$7E,$6F
	.dc.b	$E1,$78,$47,$FD,$C2,$7F,$A6,$CF,$F2,$99,$FE,$0A,$75,$DB,$C7,$B7
	.dc.b	$97,$6F,$3E,$DE,$9D,$BD,$7B,$7B,$75,$F7,$F7,$7C,$1E,$EF,$87,$DD
	.dc.b	$F1,$7B,$BE,$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$E7,$F7,$7E,$CE
	.dc.b	$5F,$E3,$1F,$9C,$FE,$BC,$7F,$16,$E3,$FF,$1F,$2F,$83,$D7,$C1,$FE
	.dc.b	$03,$6F,$C0,$ED,$F8,$2D,$BF,$07,$B7,$E1,$36,$FC,$2E,$DF,$86,$DB
	.dc.b	$F0,$FB,$7F,$86,$FF,$5C,$FE,$4A,$77,$F8,$73,$DF,$7F,$F0,$C3,$FE
	.dc.b	$F0,$8C,$FC,$2F,$1F,$85,$E5,$F0,$BC,$FE,$13,$A7,$C2,$75,$F8,$4F
	.dc.b	$F2,$67,$EE,$6F,$84,$F0,$7C,$2F,$85,$F1,$7C,$2F,$8F,$E1,$7C,$9F
	.dc.b	$0B,$E5,$F8,$5F,$37,$C2,$F9,$FD,$FF,$47,$C1,$FA,$7E,$0F,$D5,$F0
	.dc.b	$7E,$BF,$83,$FF,$1B,$7E,$BD,$FE,$1E,$3F,$C3,$A7,$EE,$7A,$3F,$C6
	.dc.b	$4B,$DE,$DF,$BD,$F4,$7A,$3E,$97,$47,$D3,$E8,$E7,$EF,$7D,$5E,$9F
	.dc.b	$AD,$D3,$F5,$FA,$7E,$C7,$47,$D9,$E8,$FB,$5D,$1F,$6F,$A3,$EE,$74
	.dc.b	$7D,$DE,$8F,$BD,$D1,$F7,$FA,$3F,$01,$D1,$F8,$1E,$8F,$C1,$74,$7E
	.dc.b	$0F,$A3,$FC,$E9,$F4,$1F,$C6,$BA,$7F,$D0,$0B,$E1,$B3,$F0,$3A,$EA
	.dc.b	$DF,$57,$0E,$7E,$3F,$03,$CB,$E0,$79,$FC,$0F,$4F,$81,$EB,$F0,$3D
	.dc.b	$BE,$07,$BF,$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93
	.dc.b	$E0,$7C,$BF,$03,$E6,$F8,$1F,$3F,$C0,$FF,$A2,$F9,$3F,$66,$7F,$13
	.dc.b	$E7,$FF,$4C,$7E,$37,$FE,$C4,$87,$C2,$EF,$E1,$78,$7C,$2F,$1F,$84
	.dc.b	$E5,$F0,$9C,$FE,$13,$A7,$C2,$75,$F8,$4E,$DF,$09,$DF,$E1,$3C,$1F
	.dc.b	$09,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$FA,$87,$93,$F8,$27,$23,$E0,$7D,$5F,$03,$FE,$A5,$FA,$2F,$F2
	.dc.b	$91,$FB,$5E,$83,$FA,$76,$FC,$1E,$FD,$FE,$1E,$FF,$1F,$7F,$97,$BF
	.dc.b	$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BF,$E1,$F7,$FC
	.dc.b	$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF,$F9,$FD,$FF,$47,$BF
	.dc.b	$FE,$BD,$FA,$3F,$C5,$73,$1F,$C5,$3F,$EC,$B7,$F4,$CE,$7D,$EE,$1E
	.dc.b	$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$7C,$9F,$3F,$B7,$BD,$DF,$DE
	.dc.b	$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$E5,$F7,$BC,$DE
	.dc.b	$F7,$9F,$DE,$F4,$7B,$DE,$9F,$7B,$FD,$AD,$F4,$9F,$CF,$BC,$F7,$B9
	.dc.b	$8F,$89,$CF,$C0,$6B,$E0,$37,$F0,$1C,$39,$78,$DF,$2F,$80,$E7,$F0
	.dc.b	$1D,$3E,$03,$AF,$C0,$76,$F8,$0E,$FF,$01,$E0,$F8,$0F,$0F,$C0,$78
	.dc.b	$BE,$03,$C7,$F0,$1E,$4F,$80,$F2,$FC,$07,$E2,$F5,$FC,$66,$BF,$8D
	.dc.b	$D7,$F1,$DA,$FE,$3F,$5F,$5F,$BF,$EC,$F7,$FF,$DE,$5E,$F7,$FA,$7C
	.dc.b	$FC,$71,$F8,$D0,$6B,$DE,$DF,$BD,$C3,$DE,$E3,$EF,$72,$F7,$B9,$FB
	.dc.b	$DF,$9D,$E6,$EB,$EF,$76,$F7,$BB,$FB,$DF,$A1,$E9,$FD,$17,$4F,$E8
	.dc.b	$FA,$7F,$49,$D3,$FA,$5E,$9F,$D3,$74,$FE,$9F,$93,$F5,$1C,$8F,$52
	.dc.b	$7E,$39,$FE,$B7,$3F,$4C,$7E,$94,$F4,$BA,$E5,$E0,$A3,$FD,$D8,$7F
	.dc.b	$92,$CF,$F2,$53,$D3,$8B,$70,$E2,$BC,$38,$E1,$F9,$3E,$FF,$CA,$F7
	.dc.b	$FE,$5F,$BF,$F3,$3D,$FF,$9B,$EF,$FC,$EF,$7F,$E7,$FB,$FF,$43,$DF
	.dc.b	$FA,$3E,$FF,$93,$E9,$BF,$4E,$7E,$9B,$84,$FE,$98,$FE,$94,$FC,$15
	.dc.b	$EF,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE
	.dc.b	$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$F0,$BC,$8F,$87
	.dc.b	$D5,$F1,$3A,$BE,$2F,$57,$C6,$EA,$F8,$FD,$5F,$23,$AB,$E4,$F5,$7C
	.dc.b	$AE,$AF,$97,$D5,$F4,$BD,$0F,$E0,$A7,$BC,$FE,$06,$7B,$BA,$CF,$B9
	.dc.b	$FC,$03,$FE,$D8,$3F,$DA,$EE,$39,$FA,$73,$F5,$E7,$F9,$9E,$F7,$CD
	.dc.b	$F7,$BE,$77,$BD,$E1,$EA,$FA,$1E,$F7,$E1,$3E,$9F,$FC,$76,$7F,$49
	.dc.b	$CC,$FF,$48,$7F,$20,$BD,$BE,$1E,$DF,$1F,$6F,$97,$B7,$CF,$DB,$E9
	.dc.b	$ED,$F5,$E2,$ED,$ED,$FC,$1F,$0F,$E1,$78,$7F,$0F,$C3,$F8,$9E,$1F
	.dc.b	$C5,$F0,$FC,$9E,$DF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$D1,$EB,$FA,$7D
	.dc.b	$7F,$57,$AF,$EB,$F5,$FD,$9E,$BF,$B7,$D7,$FD,$0F,$D4,$7E,$8B,$AC
	.dc.b	$FF,$4D,$3F,$A6,$1F,$A8,$73,$EC,$70,$F6,$F8,$FB,$7C,$BD,$BF,$B5
	.dc.b	$E9,$74,$F6,$FA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$D6,$F0,$FA,$DE,$2F
	.dc.b	$5B,$C7,$EB,$79,$3D,$6F,$2F,$AD,$E6,$F5,$BC,$FE,$B7,$A3,$D6,$FC
	.dc.b	$47,$A5,$EA,$F5,$BF,$77,$E7,$7F,$4E,$3F,$50,$7F,$4C,$3F,$14,$7E
	.dc.b	$27,$9F,$EC,$7B,$1C,$3D,$8E,$3E,$C7,$2F,$63,$9F,$B1,$D3,$D6,$EB
	.dc.b	$E6,$F6,$F3,$7B,$F9,$BE,$0F,$37,$C3,$E6,$F8,$BB,$BC,$7C,$3E,$4F
	.dc.b	$63,$CB,$EC,$79,$BD,$8F,$E2,$3F,$53,$FE,$88,$3F,$1C,$7E,$37,$8C
	.dc.b	$FE,$35,$FE,$3C,$FF,$A2,$47,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7F,$D7,$69,$E0,$F5,$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F
	.dc.b	$5F,$CB,$EB,$F9,$BD,$7F,$3F,$AF,$E8,$F5,$FD,$3E,$BF,$F2,$AF,$AA
	.dc.b	$FC,$51,$FE,$15,$3F,$C2,$87,$E2,$47,$AF,$BF,$5F,$87,$AF,$FC,$07
	.dc.b	$AB,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F,$07,$AF
	.dc.b	$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB,$F9,$FD
	.dc.b	$7F,$47,$AF,$FD,$0B,$EA,$FF,$C2,$A7,$F2,$F3,$FC,$26,$7F,$11,$3F
	.dc.b	$10,$3D,$7E,$1E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB
	.dc.b	$7A,$FD,$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$79,$DE
	.dc.b	$5F,$03,$CD,$EB,$FF,$58,$F8,$4F,$F6,$79,$FE,$5D,$F8,$C7,$C5,$FF
	.dc.b	$22,$F6,$BF,$D1,$DF,$07,$BF,$73,$87,$BD,$FE,$96,$E6,$FF,$4B,$F3
	.dc.b	$7F,$A6,$39,$BF,$D3,$3C,$DF,$E9,$AE,$6F,$F4,$DF,$37,$FA,$73,$9B
	.dc.b	$FD,$3B,$CD,$FE,$9E,$E6,$FF,$4F,$F3,$7F,$A8,$39,$BF,$D4,$3C,$DF
	.dc.b	$EA,$2E,$6F,$F5,$1F,$37,$F7,$5E,$57,$A0,$F9,$6C,$3E,$93,$F4,$1F
	.dc.b	$97,$3C,$CF,$EE,$AF,$E2,$3A,$17,$CB,$9F,$8F,$3F,$2F,$7B,$9F,$BD
	.dc.b	$D3,$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$7B,$C3,$EF,$78,$BD,$EF
	.dc.b	$1F,$BD,$E4,$F7,$BC,$BE,$F7,$9B,$93,$CF,$EF,$7A,$3D,$EF,$A5,$C1
	.dc.b	$F4,$F8,$3E,$A7,$07,$D5,$E0,$F6,$FB,$DF,$5F,$87,$EC,$70,$FD,$9E
	.dc.b	$1F,$B5,$C3,$F6,$F8,$7E,$E7,$0F,$DD,$E1,$FB,$DC,$3F,$7F,$87,$F0
	.dc.b	$1C,$3F,$81,$E1,$FC,$16,$BF,$E1,$AF,$AD,$FE,$74,$7F,$A1,$4F,$F4
	.dc.b	$21,$FC,$E4,$F9,$5C,$C5,$EE,$F1,$E9,$E5,$EE,$F3,$F7,$7A,$7B,$BD
	.dc.b	$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7,$7C
	.dc.b	$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE,$FF,$8C,$BE,$BB,$FD
	.dc.b	$0A,$7F,$87,$CF,$F4,$19,$FE,$1E,$F2,$1F,$AF,$FF,$57,$8F,$E7,$47
	.dc.b	$EC,$F9,$FF,$0B,$0F,$99,$FC,$82,$F9,$4B,$F1,$58,$F8,$3C,$F5,$6B
	.dc.b	$AB,$7D,$5C,$3A,$B8,$F5,$72,$EA,$E7,$D5,$D3,$AB,$AF,$57,$6E,$AE
	.dc.b	$FD,$5E,$0E,$AF,$0F,$57,$F9,$3B,$F1,$FF,$F5,$E7,$2F,$F9,$5D,$77
	.dc.b	$F9,$BB,$3C,$FD,$9E,$8E,$CF,$4F,$67,$EA,$7F,$60,$FF,$79,$76,$9F
	.dc.b	$F5,$4B,$7E,$4B,$5F,$07,$BF,$8B,$E1,$F0,$DC,$7E,$1B,$97,$C2,$73
	.dc.b	$F8,$4E,$9F,$09,$D7,$E1,$3B,$7C,$27,$7F,$84,$FA,$98,$7D,$5C,$3C
	.dc.b	$5F,$09,$F5,$F1,$FB,$18,$FD,$9C,$7E,$D6,$3F,$6F,$1F,$B9,$7F,$77
	.dc.b	$E1,$3E,$F7,$C2,$7D,$FF,$84,$FC,$07,$C2,$7E,$07,$E1,$3F,$05,$F0
	.dc.b	$9F,$83,$F8,$4F,$F3,$AF,$D9,$7F,$07,$EE,$FF,$3F,$3F,$E3,$33,$FB
	.dc.b	$31,$FD,$95,$2F,$82,$E3,$F0,$5C,$BE,$0B,$9F,$C1,$74,$F8,$2E,$BF
	.dc.b	$05,$DB,$E0,$BB,$FC,$17,$83,$E0,$BC,$3F,$05,$E2,$F8,$2F,$1F,$C1
	.dc.b	$79,$3E,$0B,$CB,$F0,$5E,$6F,$2F,$CF,$F0,$5E,$8F,$82,$FF,$46,$FB
	.dc.b	$9F,$E2,$D3,$FD,$F0,$7F,$BD,$CF,$F1,$61,$FB,$43,$F6,$7C,$BF,$C7
	.dc.b	$BC,$A7,$6F,$2F,$6F,$9F,$B7,$F8,$3E,$FF,$C2,$77,$FE,$17,$BF,$F0
	.dc.b	$DD,$FF,$87,$EF,$FC,$47,$7F,$F3,$1F,$D8,$7E,$57,$9A,$7F,$A4,$1F
	.dc.b	$E5,$B7,$1F,$09,$BF,$07,$87,$83,$C7,$C1,$E5,$E0,$F3,$F0,$7A,$78
	.dc.b	$3D,$7C,$1E,$DE,$0F,$7F,$07,$C1,$E0,$F8,$7C,$1F,$17,$83,$E3,$F0
	.dc.b	$7C,$9E,$0F,$97,$C1,$F3,$78,$3E,$7F,$07,$FA,$47,$4F,$FA,$DB,$94
	.dc.b	$FE,$F7,$C0,$FE,$86,$FE,$F5,$BE,$FF,$0F,$0B,$8F,$BF,$CB,$C1,$E7
	.dc.b	$E0,$F4,$F7,$FE,$A9,$F5,$BD,$FF,$AF,$EF,$F8,$1F,$B3,$EF,$FD,$AF
	.dc.b	$7F,$ED,$FB,$FF,$73,$DF,$FB,$BE,$FF,$DE,$F7,$FE,$FF,$BF,$F8,$0F
	.dc.b	$7B,$F0,$3E,$F7,$E0,$BD,$EF,$C1,$FB,$DF,$ED,$0F,$B4,$7C,$ED,$F9
	.dc.b	$7C,$7C,$0E,$7E,$07,$5F,$03,$BF,$81,$E1,$F0,$3C,$7E,$07,$97,$C0
	.dc.b	$F3,$F8,$1E,$9F,$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7B,$FE,$1F
	.dc.b	$7F,$C5,$EF,$F8,$FC,$8F,$27,$BF,$E5,$F7,$FC,$DE,$FF,$FB,$9B,$ED
	.dc.b	$7F,$65,$E6,$1F,$F5,$21,$FE,$57,$3F,$64,$7E,$C5,$FF,$7A,$3F,$E5
	.dc.b	$66,$FB,$9F,$73,$C7,$E9,$EE,$75,$F7,$3B,$7B,$9F,$80,$F2,$7F,$03
	.dc.b	$E4,$FE,$0B,$C9,$FC,$1F,$93,$F8,$4F,$27,$F0,$BE,$4F,$E1,$BC,$9F
	.dc.b	$C3,$F9,$3F,$88,$F2,$7F,$13,$E4,$FE,$2B,$C9,$FF,$09,$7E,$C5,$FC
	.dc.b	$A8,$FE,$52,$3C,$7C,$F8,$FA,$F1,$F7,$E3,$F0,$F1,$F8,$F8,$FC,$BC
	.dc.b	$7E,$7E,$3F,$4F,$1F,$AF,$8F,$DB,$C7,$EF,$E3,$F8,$3C,$7F,$0F,$8F
	.dc.b	$E2,$F1,$FC,$7E,$3F,$93,$C7,$F2,$F8,$FE,$6F,$1F,$FC,$51,$FB,$1F
	.dc.b	$EE,$0F,$F1,$F1,$FB,$73,$E7,$FB,$DF,$57,$DE,$DF,$C0,$7D,$7F,$7F
	.dc.b	$EC,$7B,$FF,$67,$DF,$FB,$5E,$FF,$DB,$F7,$FE,$E7,$BF,$F7,$7D,$FF
	.dc.b	$BD,$EF,$FD,$FF,$7D,$E9,$CF,$B8,$FF,$63,$1F,$B1,$4F,$EC,$5B,$EF
	.dc.b	$F0,$73,$EF,$D9,$FE,$3B,$1C,$3D,$EC,$7B,$D9,$F7,$B5,$EF,$6F,$D9
	.dc.b	$E1,$EB,$71,$F5,$B9,$7A,$DC,$FD,$6E,$9E,$B7,$5F,$5B,$B7,$AD,$DF
	.dc.b	$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$79,$7B,$FF,$63,$C7,$EE,$5E,$E6
	.dc.b	$7D,$6D,$7A,$DB,$F5,$B8,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7,$AD
	.dc.b	$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B
	.dc.b	$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5,$BE,$67,$DC,$FE,$C8,$7C,$C6
	.dc.b	$3E,$43,$3F,$0B,$AE,$AD,$FC,$2F,$0F,$80,$E3,$F0,$1C,$BE,$03,$9F
	.dc.b	$C0,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$05,$F7,$5F,$83,$F3
	.dc.b	$0F,$E2,$4F,$C4,$0F,$7F,$7E,$FF,$0F,$85,$E3,$EF,$F2,$F7,$F9,$FB
	.dc.b	$FD,$3D,$FE,$BE,$FF,$6F,$7F,$BF,$BF,$E0,$F7,$FC,$3E,$FF,$8B,$DF
	.dc.b	$F1,$FB,$FE,$4F,$7F,$CB,$EF,$F9,$BD,$FF,$3F,$BF,$E8,$F7,$FF,$35
	.dc.b	$F7,$7F,$86,$E7,$3F,$86,$7F,$4A,$FF,$09,$6F,$BD,$C3,$DE,$E3,$EF
	.dc.b	$72,$F7,$B9,$FB,$DD,$3D,$EE,$BE,$F7,$6F,$7B,$BF,$BD,$E0,$F7,$BC
	.dc.b	$3E,$F7,$8B,$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD
	.dc.b	$E8,$F7,$BF,$65,$C5,$FC,$E0,$FC,$3F,$27,$EF,$17,$B9,$AF,$73,$7E
	.dc.b	$E7,$0F,$73,$8F,$B9,$CB,$DC,$E7,$EE,$74,$F7,$3A,$FB,$9D,$BD,$CE
	.dc.b	$FE,$E7,$83,$DC,$F0,$FB,$9E,$2F,$73,$C7,$EE,$79,$3D,$CF,$2F,$B9
	.dc.b	$E6,$F7,$3C,$FE,$E7,$F0,$9F,$BC,$FC,$41,$FC,$3C,$FE,$1C,$7E,$1F
	.dc.b	$9F,$F8,$AB,$FE,$9C,$3F,$86,$9F,$EC,$97,$3E,$C7,$3F,$63,$A7,$B1
	.dc.b	$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1,$7B,$1E,$3F,$63
	.dc.b	$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7,$F2,$1F,$BD,$FC
	.dc.b	$41,$F8,$7E,$23,$F8,$76,$FB,$1B,$F2,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3F,$9C,$7D
	.dc.b	$F7,$F7,$E3,$F0,$FC,$A7,$F0,$EF,$F4,$47,$FC,$E0,$7F,$7F,$75,$EB
	.dc.b	$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF,$AD,$E0,$F5,$BC
	.dc.b	$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79,$BD,$6F,$3F,$AD
	.dc.b	$E8,$F5,$BF,$A8,$7B,$9F,$C3,$CF,$EA,$C7,$F5,$53,$F1,$1C,$7F,$EE
	.dc.b	$0F,$77,$71,$C3,$DD,$E3,$EC,$72,$F6,$39,$FB,$1D,$3D,$8E,$BE,$C7
	.dc.b	$6F,$63,$BF,$B1,$E0,$F6,$3C,$3E,$C7,$8B,$D8,$F1,$FB,$1E,$4F,$63
	.dc.b	$CB,$EC,$79,$BD,$8F,$3F,$B1,$FD,$9F,$EF,$FF,$87,$1F,$87,$3F,$0C
	.dc.b	$3D,$8D,$7B,$1B,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7,$3F,$63,$A7,$B1
	.dc.b	$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1,$7B,$1E,$3F,$63
	.dc.b	$C9,$EC,$79,$7D,$8F,$37,$B1,$FD,$F7,$F0,$0F,$F2,$19,$F8,$63,$F0
	.dc.b	$BE,$41,$FF,$5A,$1F,$85,$3F,$C8,$4E,$7D,$7E,$5E,$BF,$3F,$5F,$A7
	.dc.b	$AF,$D7,$D7,$ED,$EB,$F7,$F5,$FC,$1E,$BF,$87,$D7,$F1,$7A,$FE,$3F
	.dc.b	$5F,$C9,$EB,$F9,$7D,$7F,$37,$AF,$E7,$F5,$FF,$C2,$9F,$80,$FF,$53
	.dc.b	$3F,$C3,$47,$F8,$64,$FC,$28,$F6,$37,$EC,$70,$F6,$38,$FB,$1C,$BD
	.dc.b	$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F,$B1
	.dc.b	$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7F,$8A
	.dc.b	$FD,$9F,$EA,$47,$F0,$B3,$F8,$51,$F8,$5F,$17,$FC,$6E,$BD,$7E,$1E
	.dc.b	$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F
	.dc.b	$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB
	.dc.b	$F9,$FD,$7F,$F2,$17,$E0,$5F,$E8,$83,$F8,$59,$F8,$63,$FD,$F0,$7F
	.dc.b	$09,$3F,$0A,$3D,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6
	.dc.b	$F5,$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC
	.dc.b	$BE,$BF,$9B,$D7,$FF,$2C,$7E,$07,$FE,$17,$3F,$CB,$C7,$F9,$74,$FE
	.dc.b	$12,$7E,$17,$C5,$FF,$32,$BF,$E5,$C3,$FA,$82,$5E,$A7,$4F,$53,$AF
	.dc.b	$A9,$DB,$D4,$EF,$EA,$78,$3D,$4F,$0F,$A9,$E2,$F5,$3C,$7E,$A7,$93
	.dc.b	$D4,$F2,$FA,$9E,$6F,$53,$CF,$EA,$7A,$3D,$4F,$4F,$A9,$FE,$71,$FC
	.dc.b	$13,$FC,$30,$7F,$51,$3F,$A8,$1F,$87,$E3,$3F,$87,$7F,$CF,$EB,$D3
	.dc.b	$E5,$E9,$F3,$F4,$FA,$7A,$7D,$7D,$3E,$DE,$9F,$7F,$4F,$C1,$E9,$F8
	.dc.b	$7D,$3F,$17,$A7,$E3,$F4,$FC,$9E,$9F,$97,$D3,$F3,$7A,$7E,$7F,$4F
	.dc.b	$D1,$E9,$FF,$A2,$78,$7F,$D9,$E7,$F5,$23,$F8,$68,$F8,$1D,$7C,$0E
	.dc.b	$FD,$7E,$1E,$AF,$1F,$57,$97,$AB,$CF,$D5,$E9,$EA,$F5,$F5,$7B,$7A
	.dc.b	$BD,$FD,$5F,$07,$AB,$E1,$F5,$7C,$5E,$AF,$8F,$D5,$F2,$7A,$BE,$5F
	.dc.b	$57,$CD,$EA,$FF,$A7,$FF,$05,$FF,$2F,$1F,$EA,$53,$FD,$48,$7F,$50
	.dc.b	$3F,$11,$E3,$1F,$F5,$19,$F8,$83,$F0,$EF,$FA,$C1,$7A,$9D,$7D,$4E
	.dc.b	$DE,$A7,$7F,$53,$C1,$EA,$78,$7D,$4F,$17,$A9,$E3,$F5,$3C,$9E,$A7
	.dc.b	$97,$D4,$F3,$7A,$9E,$7F,$53,$D1,$EA,$7A,$7D,$4F,$F5,$C7,$E0,$DF
	.dc.b	$E5,$F3,$FC,$30,$7F,$85,$CF,$C4,$1F,$87,$E2,$3F,$87,$7F,$D8,$EB
	.dc.b	$D4,$E7,$EA,$74,$F5,$3A,$FA,$9D,$BD,$4E,$FE,$A7,$83,$D4,$F0,$FA
	.dc.b	$9E,$2F,$53,$C7,$EA,$79,$3D,$4F,$2F,$A9,$E6,$F5,$3C,$FE,$A7,$A3
	.dc.b	$D4,$FF,$68,$7E,$0E,$F9,$EB,$F1,$58,$F8,$7C,$FA,$FA,$F5,$F7,$EB
	.dc.b	$F0,$F5,$F8,$FA,$FC,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF
	.dc.b	$EB,$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE
	.dc.b	$6F,$5F,$FD,$CD,$EF,$7F,$A9,$4F,$F7,$61,$FE,$EB,$3F,$C2,$E7,$E2
	.dc.b	$0F,$F7,$49,$FC,$38,$FC,$3F,$92,$7F,$0E,$FF,$BE,$17,$AD,$DB,$D6
	.dc.b	$EF,$EB,$78,$3D,$6F,$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6,$F2,$FA
	.dc.b	$DE,$6F,$5B,$CF,$EB,$7A,$3D,$6F,$4F,$AD,$FE,$FF,$FC,$23,$FD,$4C
	.dc.b	$7F,$97,$CF,$F2,$F1,$FC,$3C,$FC,$41,$F8,$7E,$13,$F8,$74,$63,$D5
	.dc.b	$E9,$EA,$F5,$F5,$7B,$7A,$BD,$FD,$5F,$07,$AB,$E1,$F5,$7C,$5E,$AF
	.dc.b	$8F,$D5,$F2,$7A,$BE,$5F,$57,$CD,$EA,$F9,$FD,$5F,$47,$AB,$E0,$E0
	.dc.b	$7E,$13,$F8,$97,$C6,$FF,$98,$8F,$F0,$E9,$FC,$45,$CF,$A3,$C6,$3F
	.dc.b	$C3,$A7,$F5,$98,$E3,$E7,$E7,$CF,$D7,$9E,$FE,$01,$85,$AE,$F0,$F9
	.dc.b	$7F,$87,$F2,$0E,$5E,$31,$86,$D2,$74,$F2,$9D,$34,$D4,$F0,$9D,$C7
	.dc.b	$C0,$AF,$03,$A3,$84,$6B,$24,$EA,$30,$DA,$E8,$4A,$7E,$03,$FE,$59
	.dc.b	$E8,$7C,$BF,$DE,$FC,$2A,$D6,$BE,$2D,$D7,$89,$CE,$1C,$6F,$81,$F8
	.dc.b	$67,$FC,$C9,$3B,$8C,$57,$C7,$77,$7D,$71,$D3,$DF,$AE,$E1,$FE,$6D
	.dc.b	$9B,$6E,$BB,$D7,$86,$E3,$D2,$FF,$CE,$9D,$7C,$3A,$F8,$B6,$75,$83
	.dc.b	$AF,$A9,$DB,$AE,$CE,$BA,$FC,$3D,$7E,$26,$C7,$2E,$BF,$3F,$D1,$7F
	.dc.b	$1D,$8C,$1E,$37,$1C,$2E,$3F,$93,$0F,$73,$89,$CB,$DC,$B7,$C7,$DC
	.dc.b	$EA,$76,$EF,$B1,$D7,$DC,$FD,$36,$3D,$3D,$CB,$3F,$4F,$2F,$D0,$67
	.dc.b	$E1,$8C,$5E,$A3,$F0,$72,$F0,$71,$CA,$8C,$4F,$AE,$73,$0F,$06,$F6
	.dc.b	$E3,$DC,$97,$F0,$63,$7E,$E7,$C6,$F7,$3E,$3B,$74,$F8,$DE,$E7,$CA
	.dc.b	$F7,$3E,$5B,$63,$E5,$7B,$9F,$38,$F9,$FE,$E5,$8F,$9D,$EE,$79,$3D
	.dc.b	$C9,$A6,$0F,$A5,$EE,$7A,$3D,$C9,$18,$3F,$11,$F4,$5F,$C3,$DF,$8C
	.dc.b	$7E,$99,$C7,$3F,$E8,$1B,$8F,$E7,$D7,$D1,$7C,$85,$F4,$34,$97,$DB
	.dc.b	$E5,$75,$F1,$D2,$3C,$2F,$1F,$8E,$E5,$F0,$52,$7C,$AF,$5F,$8B,$ED
	.dc.b	$F1,$1F,$A6,$CB,$FA,$5E,$9E,$A8,$C7,$C2,$67,$E1,$24,$FB,$BE,$12
	.dc.b	$E5,$EE,$F0,$18,$BE,$13,$A7,$C2,$75,$6E,$1F,$1B,$E1,$3D,$DF,$09
	.dc.b	$31,$C1,$F2,$BE,$13,$E1,$11,$F9,$BF,$09,$F1,$7E,$12,$47,$D0,$7E
	.dc.b	$47,$C2,$7C,$9F,$84,$F6,$7C,$27,$B5,$B9,$FE,$F7,$E8,$3F,$8B,$B7
	.dc.b	$E6,$7F,$83,$C6,$31,$F6,$28,$CF,$C0,$EB,$D9,$B7,$CD,$E3,$F2,$1F
	.dc.b	$A0,$BF,$EC,$5E,$07,$F1,$E6,$D7,$F1,$D7,$FC,$5D,$19,$F8,$AE,$1F
	.dc.b	$0F,$33,$F3,$47,$3F,$84,$91,$E9,$7B,$7C,$27,$7F,$83,$93,$ED,$7C
	.dc.b	$5F,$09,$E3,$F8,$39,$3F,$05,$F3,$7C,$27,$9F,$E0,$FD,$1F,$09,$2B
	.dc.b	$E3,$47,$AF,$E1,$3E,$2F,$C2,$7C,$66,$F0,$FC,$BF,$83,$F9,$3F,$09
	.dc.b	$F2,$9B,$1F,$3F,$E1,$3F,$99,$F9,$9F,$C6,$B9,$3E,$24,$63,$E2,$73
	.dc.b	$F0,$BA,$F8,$5B,$87,$F3,$3C,$11,$CB,$E1,$79,$FC,$2D,$AD,$3D,$BE
	.dc.b	$12,$47,$1F,$85,$BC,$5F,$97,$C7,$F8,$9B,$7E,$17,$F8,$BC,$63,$6F
	.dc.b	$EA,$AF,$F1,$59,$F1,$7C,$16,$BE,$0E,$46,$0F,$37,$C1,$72,$23,$E9
	.dc.b	$F8,$3E,$BF,$05,$F0,$8F,$87,$F0,$77,$5F,$77,$C1,$FC,$63,$E3,$FC
	.dc.b	$1D,$8F,$85,$F0,$7F,$28,$F9,$7F,$07,$63,$E3,$7C,$1F,$CE,$3E,$7F
	.dc.b	$C1,$D8,$FE,$B9,$F3,$DF,$E3,$26,$FC,$27,$F6,$68,$C6,$DF,$BB,$7F
	.dc.b	$89,$CF,$0F,$84,$D1,$BF,$84,$B1,$D3,$93,$F9,$FD,$3E,$12,$98,$8C
	.dc.b	$FC,$26,$BE,$12,$65,$E4,$8E,$3F,$09,$23,$CE,$F4,$F8,$4E,$BF,$09
	.dc.b	$DB,$E1,$25,$7B,$23,$C3,$F0,$32,$3D,$EF,$93,$E0,$BC,$BF,$05,$27
	.dc.b	$E1,$BE,$8F,$82,$F4,$FC,$14,$9F,$8E,$FB,$3E,$0B,$DB,$F0,$5E,$EF
	.dc.b	$82,$95,$F3,$23,$E0,$FC,$0F,$C2,$F8,$29,$3F,$42,$3F,$B1,$FC,$F7
	.dc.b	$F0,$DF,$1F,$FC,$42,$DE,$8F,$F1,$03,$FC,$3D,$FF,$09,$39,$AF,$F1
	.dc.b	$7B,$FE,$43,$8C,$FC,$17,$F8,$7A,$1F,$05,$63,$1F,$05,$9F,$8F,$B7
	.dc.b	$D1,$E1,$F0,$3E,$4F,$82,$99,$78,$23,$CF,$F0,$5E,$8F,$82,$93,$E4
	.dc.b	$8F,$5F,$C1,$48,$F3,$BE,$EF,$82,$F7,$FC,$14,$9F,$5B,$F0,$BE,$0B
	.dc.b	$E1,$FC,$0C,$9F,$7B,$F1,$BE,$0B,$E3,$FC,$0F,$C8,$F8,$29,$5F,$12
	.dc.b	$3E,$5F,$C0,$C8,$F8,$EF,$CE,$F8,$2F,$9F,$F0,$52,$7E,$5B,$FE,$14
	.dc.b	$F9,$CF,$F4,$14,$7F,$59,$7F,$C5,$4F,$F9,$05,$AE,$7E,$0B,$87,$C1
	.dc.b	$71,$F8,$2F,$F2,$6D,$3E,$0A,$C6,$1C,$FC,$14,$E9,$E9,$F0,$57,$2F
	.dc.b	$07,$C1,$72,$F8,$8E,$6D,$AF,$17,$C1,$F6,$3B,$FC,$1D,$8F,$37,$C1
	.dc.b	$F8,$BE,$0F,$C6,$D8,$F5,$7C,$1F,$98,$F3,$FC,$1D,$8F,$77,$C1,$FC
	.dc.b	$32,$69,$F0,$7E,$0F,$E3,$7C,$1C,$8F,$88,$FC,$9F,$83,$F9,$5F,$07
	.dc.b	$F2,$FE,$0E,$57,$C9,$8F,$F2,$0F,$27,$F8,$7A,$3F,$D0,$EF,$F8,$79
	.dc.b	$FF,$37,$35,$FF,$05,$3B,$F8,$8F,$99,$F0,$7C,$89,$3C,$3E,$0F,$AF
	.dc.b	$C0,$F6,$F8,$29,$3D,$23,$C3,$F0,$3F,$E8,$2C,$7F,$DD,$5C,$1E,$0F
	.dc.b	$82,$C4,$E7,$E0,$B5,$F0,$56,$BC,$AF,$1F,$81,$E5,$F0,$56,$7D,$2F
	.dc.b	$5F,$82,$F7,$93,$4F,$6F,$C1,$7C,$2F,$82,$F8,$7F,$05,$27,$E0,$C7
	.dc.b	$C6,$F8,$29,$1F,$11,$F9,$3F,$05,$F2,$BE,$07,$E5,$FC,$14,$AF,$93
	.dc.b	$1F,$3B,$E0,$7E,$7F,$C1,$49,$F9,$B1,$FE,$72,$F9,$CF,$F5,$9C,$7F
	.dc.b	$8D,$DF,$F0,$E3,$FE,$9F,$6B,$9D,$F8,$7C,$07,$FA,$AA,$7F,$88,$4F
	.dc.b	$1F,$81,$C4,$E7,$E0,$65,$F3,$F8,$1B,$7D,$FE,$03,$91,$CF,$E0,$6D
	.dc.b	$78,$7E,$03,$CD,$F0,$32,$F0,$79,$7E,$07,$D4,$47,$D1,$F0,$3E,$DF
	.dc.b	$81,$91,$EB,$7E,$07,$C0,$FC,$1F,$81,$F5,$7C,$0F,$AD,$B9,$FC,$1F
	.dc.b	$80,$F7,$7C,$0F,$BD,$B1,$F1,$7E,$07,$E1,$7C,$0F,$C3,$6C,$7C,$9F
	.dc.b	$81,$FF,$67,$FC,$E7,$F8,$76,$3F,$DB,$5D,$09,$CF,$C3,$EB,$E0,$F7
	.dc.b	$F0,$76,$FF,$DA,$18,$7F,$87,$9B,$F0,$5D,$0E,$BE,$6D,$8D,$FC,$17
	.dc.b	$FA,$BA,$BF,$D6,$BF,$0B,$FE,$EC,$6C,$FF,$BA,$D4,$EF,$E0,$78,$7C
	.dc.b	$14,$E7,$F0,$C7,$3F,$82,$91,$E4,$7B,$7C,$17,$7F,$81,$F0,$7C,$14
	.dc.b	$AF,$4C,$78,$FE,$0A,$47,$B1,$F3,$7C,$17,$9F,$E0,$BD,$1F,$05,$2B
	.dc.b	$E0,$C7,$AF,$E0,$BE,$5F,$C1,$7C,$C6,$F0,$FC,$6F,$81,$F9,$FF,$05
	.dc.b	$F4,$1B,$1F,$EF,$9F,$9C,$FF,$0D,$F2,$A1,$93,$E0,$D5,$A9,$CC,$63
	.dc.b	$E2,$38,$57,$FA,$76,$B4,$F1,$7F,$D6,$AF,$7F,$83,$FF,$6F,$C1,$BF
	.dc.b	$09,$8F,$83,$95,$E5,$77,$F0,$9C,$3E,$0E,$4F,$A5,$E7,$F0,$9D,$3E
	.dc.b	$13,$AF,$C2,$4A,$F7,$47,$83,$E1,$3C,$3F,$09,$27,$E1,$47,$93,$E1
	.dc.b	$3C,$BF,$09,$27,$E3,$47,$A3,$E0,$FD,$FC,$9F,$03,$8F,$E2,$C6,$39
	.dc.b	$3E,$2F,$1B,$E1,$31,$8F,$C3,$AB,$7F,$33,$E1,$2D,$BE,$12,$C6,$1C
	.dc.b	$FC,$1C,$CB,$EC,$7C,$25,$CB,$E8,$7C,$1F,$2F,$84,$E6,$DA,$FB,$7B
	.dc.b	$79,$FE,$0B,$FC,$4D,$A9,$FF,$15,$61,$FE,$D5,$72,$62,$78,$7C,$1F
	.dc.b	$A2,$3F,$8E,$C7,$3F,$84,$C4,$E6,$75,$F0,$97,$1F,$09,$F8,$88,$47
	.dc.b	$3F,$84,$93,$AC,$7F,$21,$4F,$84,$CC,$79,$7E,$13,$C4,$78,$FE,$12
	.dc.b	$DF,$1D,$7D,$9F,$09,$E7,$3D,$1F,$09,$67,$59,$3E,$13,$D8,$7B,$7E
	.dc.b	$12,$CE,$BE,$0F,$E0,$7C,$27,$C1,$6C,$72,$F8,$3F,$89,$F0,$9F,$15
	.dc.b	$B1,$DB,$E0,$FE,$47,$C2,$7C,$96,$C7,$8B,$E0,$FE,$61,$F8,$4F,$9C
	.dc.b	$FE,$49,$C7,$F0,$A3,$1F,$07,$FB,$48,$75,$7E,$46,$BF,$D7,$F1,$F0
	.dc.b	$5A,$EA,$FF,$6F,$1F,$E3,$69,$E1,$F0,$5C,$DB,$5C,$63,$15,$C3,$E0
	.dc.b	$AC,$FF,$03,$BF,$E4,$AF,$7F,$86,$CF,$32,$7F,$95,$CE,$FE,$0B,$31
	.dc.b	$FE,$C7,$C3,$C7,$F0,$38,$33,$F0,$5A,$F8,$29,$87,$AD,$E3,$F0,$5C
	.dc.b	$BE,$06,$4F,$BD,$EB,$F0,$5D,$BE,$0B,$BF,$C1,$4A,$F8,$91,$E2,$F8
	.dc.b	$1F,$1F,$C1,$49,$F9,$11,$E6,$F8,$19,$1F,$2D,$F4,$FC,$17,$AB,$E0
	.dc.b	$7E,$B7,$C1,$7D,$76,$F0,$FE,$47,$E7,$3F,$90,$F2,$9F,$F6,$2E,$3F
	.dc.b	$4C,$FE,$58,$E8,$CF,$4F,$F0,$B5,$F0,$3F,$C1,$9F,$E4,$D7,$D6,$31
	.dc.b	$F0,$59,$F8,$29,$3E,$1F,$82,$B8,$F7,$CB,$C5,$F0,$5C,$FE,$0B,$A3
	.dc.b	$6F,$D1,$F0,$3D,$CF,$07,$C1,$58,$F6,$7C,$0F,$8C,$F2,$7C,$15,$8F
	.dc.b	$81,$F0,$3E,$73,$D1,$F0,$56,$3E,$27,$C0,$FA,$FE,$0B,$D8,$D8,$F9
	.dc.b	$1F,$03,$EF,$3E,$07,$C0,$D8,$F9,$9F,$03,$F0,$CF,$89,$F0,$36,$3E
	.dc.b	$87,$C0,$7C,$7F,$81,$F9,$0D,$8F,$E0,$9F,$39,$FC,$80,$9F,$1B,$FC
	.dc.b	$C9,$C7,$C1,$5F,$83,$C3,$6B,$8F,$B9,$BF,$83,$E0,$D8,$EB,$F0,$7C
	.dc.b	$FE,$0F,$A3,$63,$C3,$F0,$7D,$FE,$0F,$C0,$D8,$F2,$FC,$1F,$8F,$E0
	.dc.b	$FC,$8D,$8F,$4F,$C1,$FE,$BB,$6F,$E3,$1E,$0E,$0F,$6F,$C1,$E7,$E0
	.dc.b	$B5,$F0,$72,$7E,$0C,$71,$F8,$2E,$5F,$07,$27,$E2,$C7,$5F,$82,$F8
	.dc.b	$9F,$07,$F1,$5B,$A7,$CA,$F8,$3F,$90,$7F,$5A,$F2,$BF,$8B,$F9,$4D
	.dc.b	$BF,$C7,$55,$FC,$CE,$5F,$07,$BF,$87,$E0,$D9,$D7,$C2,$73,$F8,$4E
	.dc.b	$8D,$8E,$5F,$07,$DC,$F0,$7C,$25,$8E,$DF,$07,$E3,$3C,$9F,$07,$63
	.dc.b	$C5,$F0,$7F,$C5,$B4,$34,$C7,$C1,$E7,$E0,$E7,$2F,$A5,$E1,$F0,$7C
	.dc.b	$7E,$0A,$4F,$B5,$E9,$F0,$7D,$7E,$0B,$B7,$C1,$CA,$F8,$51,$E1,$F8
	.dc.b	$3F,$17,$C1,$C9,$F8,$D1,$E5,$F8,$29,$1F,$25,$F4,$7C,$1F,$F2,$EF
	.dc.b	$9C,$FE,$1A,$7F,$08,$E1,$FF,$19,$B8,$F8,$7F,$F0,$6B,$FC,$2A,$3F
	.dc.b	$B2,$36,$FF,$82,$3F,$C3,$1B,$F0,$7C,$3E,$0A,$57,$0E,$4B,$5C,$3E
	.dc.b	$07,$A9,$DB,$E0,$AD,$76,$F8,$2F,$09,$E2,$F8,$2B,$1E,$2F,$82,$F2
	.dc.b	$FC,$17,$99,$B1,$E6,$F8,$2F,$4F,$C1,$7A,$9B,$1E,$AF,$81,$F6,$9E
	.dc.b	$EF,$82,$B1,$EE,$F8,$2F,$82,$7C,$2F,$82,$B1,$F0,$BE,$0B,$E2,$FC
	.dc.b	$17,$C6,$6C,$7C,$6F,$81,$F9,$27,$CA,$F8,$2B,$1F,$E0,$DF,$9C,$FF
	.dc.b	$14,$9F,$E2,$57,$FC,$C6,$E3,$E0,$FF,$C7,$AF,$F8,$9A,$33,$F0,$78
	.dc.b	$F8,$3C,$BA,$F8,$29,$3D,$7E,$0E,$F4,$FF,$39,$C1,$F0,$5D,$3E,$0F
	.dc.b	$AB,$6F,$C9,$F0,$5E,$0F,$83,$F0,$B6,$3D,$1F,$05,$E4,$F8,$3F,$2B
	.dc.b	$63,$D9,$F0,$7E,$8F,$83,$F4,$B6,$3E,$07,$C1,$7F,$7E,$D1,$F0,$74
	.dc.b	$C4,$67,$E0,$A6,$FF,$1D,$E1,$F0,$7C,$7E,$0A,$4F,$CB,$7A,$7C,$1F
	.dc.b	$5F,$83,$93,$F3,$DF,$F0,$6F,$CE,$3E,$2E,$FC,$16,$3E,$0E,$5F,$F9
	.dc.b	$06,$BF,$C3,$DB,$FF,$A5,$97,$C1,$72,$F8,$39,$5B,$8E,$BF,$07,$FE
	.dc.b	$74,$AF,$F6,$0D,$F3,$F8,$3C,$4E,$7E,$0E,$4F,$7F,$83,$B8,$FF,$6F
	.dc.b	$C3,$F8,$6C,$F8,$FE,$0F,$A7,$C1,$F5,$7B,$7C,$1C,$9F,$47,$C1,$DC
	.dc.b	$39,$7C,$1F,$8F,$E0,$FC,$8D,$AF,$77,$C1,$79,$CF,$47,$C1,$D8,$F8
	.dc.b	$5F,$05,$F2,$3E,$0E,$63,$83,$E3,$7C,$1F,$CC,$F8,$39,$18,$3F,$CD
	.dc.b	$5C,$7F,$E4,$F7,$FC,$14,$BE,$0F,$FC,$7E,$D8,$D9,$8F,$82,$CF,$C3
	.dc.b	$48,$E7,$1C,$3E,$0E,$47,$67,$9F,$C1,$F4,$F8,$3E,$BF,$07,$2B,$C7
	.dc.b	$1E,$0F,$83,$FF,$54,$E4,$F8,$3B,$18,$73,$F0,$5A,$F8,$3B,$C7,$EC
	.dc.b	$78,$FC,$17,$2F,$83,$B3,$F0,$1E,$BF,$07,$EE,$26,$9F,$13,$E0,$FE
	.dc.b	$0F,$C1,$FC,$2F,$83,$93,$F2,$23,$E2,$FC,$17,$C6,$F8,$39,$3F,$32
	.dc.b	$3E,$4F,$C1,$FA,$BE,$0F,$D6,$DD,$3F,$D3,$5F,$39,$FE,$B6,$3F,$87
	.dc.b	$5F,$FB,$CD,$C7,$C2,$7F,$B2,$9F,$E1,$D1,$9F,$84,$C7,$C2,$65,$D7
	.dc.b	$C2,$4C,$3A,$FC,$25,$BF,$F2,$36,$1F,$C3,$A7,$C3,$F0,$9D,$7E,$13
	.dc.b	$B3,$DF,$E0,$E4,$F9,$BE,$12,$E1,$CB,$E0,$FC,$9F,$09,$E5,$6D,$7B
	.dc.b	$3E,$0F,$D0,$7A,$7E,$12,$C7,$C0,$F8,$2F,$67,$C1,$FB,$5B,$1F,$13
	.dc.b	$E0,$FF,$D1,$FB,$9A,$63,$E0,$F3,$F0,$7A,$F8,$39,$B7,$CC,$8E,$3F
	.dc.b	$07,$FE,$98,$F9,$CF,$E1,$93,$FE,$C9,$6B,$F1,$23,$5F,$07,$FE,$F8
	.dc.b	$9F,$F0,$DE,$7B,$F8,$3C,$4E,$7E,$0A,$4F,$3F,$83,$B8,$FF,$9C,$F0
	.dc.b	$FE,$1A,$DF,$83,$E9,$E6,$F5,$7B,$7C,$1C,$C3,$C9,$F0,$76,$F9,$7C
	.dc.b	$0F,$8F,$E0,$BC,$8D,$AF,$57,$C1,$79,$CF,$47,$C1,$58,$F7,$7C,$0F
	.dc.b	$AC,$F6,$7C,$15,$8F,$85,$F0,$5E,$F3,$E0,$7C,$15,$8F,$8D,$F0,$5F
	.dc.b	$0F,$E0,$BE,$23,$63,$E5,$7C,$17,$C7,$3E,$47,$C1,$58,$FF,$7B,$7C
	.dc.b	$E7,$FA,$98,$9F,$06,$8C,$F0,$D3,$FD,$35,$3F,$E1,$9F,$8B,$C7,$C5
	.dc.b	$23,$14,$A4,$E7,$E1,$79,$5F,$0B,$DD,$69,$7C,$26,$B0,$E5,$87,$58
	.dc.b	$C5,$6F,$E1,$2C,$E6,$BD,$7F,$07,$8C,$1F,$09,$63,$0E,$7E,$13,$5F
	.dc.b	$09,$6F,$E1,$BC,$7E,$13,$FD,$D7,$83,$E1,$29,$88,$CF,$C2,$6B,$E1
	.dc.b	$26,$DF,$32,$38,$FC,$24,$8F,$9E,$F4,$F8,$4E,$BF,$07,$E7,$3D,$1F
	.dc.b	$09,$75,$FA,$BF,$07,$EB,$F8,$4F,$63,$63,$EC,$FC,$1F,$BF,$E1,$3E
	.dc.b	$03,$63,$E9,$74,$7D,$6F,$80,$FB,$5A,$FF,$B4,$DB,$19,$9F,$B1,$39
	.dc.b	$F8,$4B,$3C,$7E,$0E,$D7,$D4,$B4,$62,$FE,$FC,$61,$F0,$CE,$FE,$13
	.dc.b	$5D,$4A,$7C,$26,$1B,$7C,$0E,$71,$98,$D7,$C2,$7E,$01,$7C,$25,$8C
	.dc.b	$39,$F8,$49,$97,$3C,$7C,$1F,$09,$8C,$BE,$17,$C1,$F3,$3A,$7C,$25
	.dc.b	$BE,$19,$BE,$13,$C1,$F0,$9E,$16,$CE,$BE,$0F,$C9,$F0,$9E,$56,$C7
	.dc.b	$2F,$83,$F4,$1E,$9F,$83,$B1,$DB,$E0,$BE,$17,$C1,$CD,$B0,$78,$BE
	.dc.b	$0F,$E4,$FC,$E7,$F2,$4E,$9F,$D8,$B7,$0F,$A9,$3A,$F8,$39,$7F,$06
	.dc.b	$BF,$C6,$6F,$ED,$17,$C2,$74,$F8,$4F,$D5,$C7,$F8,$DB,$0E,$1F,$09
	.dc.b	$89,$CF,$C1,$C9,$E9,$F0,$77,$16,$9E,$4F,$83,$E6,$74,$F8,$3B,$7E
	.dc.b	$3F,$82,$F5,$FC,$1C,$C3,$07,$9F,$E0,$FD,$E4,$7D,$5F,$07,$F0,$BE
	.dc.b	$0F,$E1,$FC,$1C,$9F,$74,$7C,$6F,$83,$F5,$FC,$1F,$B1,$B9,$FC,$3F
	.dc.b	$82,$F7,$FC,$1F,$C0,$6C,$7C,$7F,$83,$F8,$7F,$07,$F1,$1B,$1F,$8B
	.dc.b	$F9,$CF,$F1,$A7,$1F,$F2,$27,$FC,$65,$A7,$C2,$9F,$E3,$31,$89,$D4
	.dc.b	$FE,$8A,$1F,$0F,$49,$1D,$7E,$13,$5F,$09,$BF,$84,$93,$E1,$8E,$5F
	.dc.b	$09,$23,$C8,$F5,$F8,$4E,$DF,$09,$DF,$E1,$25,$7A,$63,$C5,$F0,$52
	.dc.b	$3D,$8F,$97,$E0,$FC,$DF,$05,$27,$E0,$3E,$9F,$83,$F5,$7C,$1F,$AF
	.dc.b	$E0,$E5,$7C,$58,$F7,$7C,$1C,$8F,$90,$FC,$1F,$83,$F8,$5F,$05,$27
	.dc.b	$E6,$3F,$17,$E0,$FE,$37,$C1,$7C,$7F,$83,$95,$F4,$63,$F8,$5F,$CE
	.dc.b	$7F,$1A,$EA,$FE,$8E,$BE,$13,$26,$A3,$7F,$09,$2F,$F8,$25,$7F,$17
	.dc.b	$7E,$C1,$FD,$61,$7C,$27,$6F,$84,$93,$C1,$F0,$FB,$BE,$2F,$84,$93
	.dc.b	$D1,$F2,$FC,$27,$9B,$E1,$24,$F8,$1F,$4F,$C1,$FA,$BE,$0F,$D7,$F0
	.dc.b	$72,$BC,$B1,$EE,$F8,$29,$1E,$87,$E0,$FC,$1F,$C2,$F8,$2F,$87,$F0
	.dc.b	$72,$BD,$B1,$F1,$BE,$0E,$47,$C0,$7E,$4F,$C1,$FC,$AF,$83,$F9,$7F
	.dc.b	$07,$2B,$E2,$C7,$CE,$F8,$3F,$E7,$3E,$47,$F0,$F9,$FE,$FA,$DF,$84
	.dc.b	$73,$AB,$7F,$09,$2F,$F9,$95,$FF,$12,$B3,$FC,$4A,$33,$F0,$3D,$BE
	.dc.b	$0A,$46,$0E,$3F,$05,$E2,$23,$D3,$E0,$BC,$BF,$05,$23,$BB,$E8,$F8
	.dc.b	$2F,$4F,$C1,$49,$F1,$BE,$CF,$82,$F6,$FC,$14,$9F,$3B,$F0,$3E,$0B
	.dc.b	$E0,$FC,$0C,$9F,$5B,$F1,$3E,$0B,$E2,$FC,$0C,$9F,$7B,$F2,$3E,$0B
	.dc.b	$E4,$FC,$0F,$CA,$F8,$29,$5F,$12,$3E,$6F,$C0,$C8,$F8,$EF,$D0,$F8
	.dc.b	$2F,$ED,$3F,$39,$FC,$43,$CD,$FF,$12,$B6,$BF,$C5,$ED,$E8,$3F,$C4
	.dc.b	$67,$7F,$03,$CB,$E0,$65,$FF,$65,$AF,$E2,$6D,$F8,$0E,$FF,$03,$23
	.dc.b	$06,$FB,$3C,$64,$79,$7C,$0F,$9B,$E0,$24,$75,$7D,$3F,$03,$EA,$F8
	.dc.b	$1F,$5F,$C0,$CA,$F1,$47,$BB,$E0,$7D,$FF,$03,$27,$CD,$1F,$0B,$E0
	.dc.b	$3E,$1F,$C0,$C9,$F5,$47,$C6,$F8,$19,$1E,$D7,$E4,$FC,$0F,$CA,$F8
	.dc.b	$19,$3F,$05,$F9,$BF,$03,$F3,$BE,$03,$E7,$FC,$0C,$AF,$8D,$1F,$E1
	.dc.b	$AF,$9A,$FE,$21,$1F,$E5,$E6,$F5,$1A,$7F,$13,$8F,$F0,$93,$FC,$3F
	.dc.b	$E1,$3F,$CD,$2D,$EB,$FF,$2F,$B6,$7A,$7C,$0F,$7F,$81,$91,$C9,$F1
	.dc.b	$7C,$0F,$8F,$E0,$7C,$9F,$03,$2B,$BC,$79,$FE,$06,$47,$89,$F5,$7C
	.dc.b	$0F,$AF,$E0,$24,$F9,$9F,$77,$C0,$FB,$FE,$07,$E0,$7C,$0C,$AF,$5C
	.dc.b	$7C,$3F,$80,$91,$EE,$7E,$37,$C0,$FC,$7F,$81,$F9,$1F,$03,$2B,$E1
	.dc.b	$C7,$CB,$F8,$0F,$99,$F0,$32,$7E,$3C,$7C,$FF,$80,$91,$F2,$9F,$F2
	.dc.b	$5F,$E1,$7F,$C4,$4F,$E1,$84,$F8,$1C,$FC,$3E,$BE,$0B,$7F,$05,$30
	.dc.b	$FF,$22,$5A,$79,$FC,$14,$9C,$BD,$BE,$0B,$BF,$C1,$78,$3E,$0A,$57
	.dc.b	$28,$F1,$FC,$14,$8E,$AF,$9B,$E0,$BC,$FF,$05,$E8,$F8,$29,$5E,$28
	.dc.b	$F5,$FC,$17,$B3,$E0,$A4,$F9,$A3,$DF,$F0,$5F,$03,$E0,$A4,$FA,$A3
	.dc.b	$E1,$FC,$0C,$8F,$6B,$F1,$BE,$0B,$E3,$FC,$14,$9F,$82,$FC,$AF,$82
	.dc.b	$FF,$3A,$F0,$FF,$09,$3F,$84,$4F,$FA,$E9,$BE,$69,$D7,$C1,$70,$F8
	.dc.b	$29,$5F,$E6,$F9,$FE,$16,$FF,$6C,$AE,$1F,$03,$D8,$91,$BF,$81,$F0
	.dc.b	$FC,$0C,$8E,$4F,$93,$E0,$7C,$BF,$03,$27,$B3,$E8,$F8,$1F,$4F,$C0
	.dc.b	$49,$F1,$3E,$CF,$81,$F6,$FC,$07,$BB,$E0,$65,$79,$E3,$E0,$FC,$07
	.dc.b	$C2,$F8,$19,$3E,$B8,$F8,$BF,$03,$23,$DC,$FC,$8F,$81,$F9,$3F,$03
	.dc.b	$27,$E1,$3F,$33,$E0,$7E,$6F,$C0,$49,$F8,$CF,$D0,$F8,$1F,$F5,$2F
	.dc.b	$E1,$9F,$C2,$8F,$E1,$13,$F2,$5B,$F0,$5F,$EF,$07,$5F,$0B,$C0,$93
	.dc.b	$FE,$9F,$8F,$F0,$CB,$7E,$03,$AF,$C0,$48,$C1,$BF,$7F,$C3,$F0,$12
	.dc.b	$30,$73,$F8,$0F,$29,$1E,$DF,$01,$E8,$F8,$09,$1E,$17,$D7,$F0,$1E
	.dc.b	$CF,$7F,$FD,$EF,$A7,$FA,$AF,$E0,$FC,$DF,$03,$89,$CF,$C0,$EB,$E0
	.dc.b	$6D,$7A,$DE,$3F,$01,$CB,$E0,$6C,$FB,$DE,$BF,$01,$23,$E1,$7C,$0D
	.dc.b	$DB,$E5,$7C,$0A,$7F,$D8,$9F,$86,$BE,$3F,$CB,$39,$AF,$CC,$4E,$6B
	.dc.b	$7F,$07,$9E,$AD,$79,$AA,$4E,$69,$6A,$D7,$C1,$E6,$31,$7C,$2B,$C7
	.dc.b	$F0,$39,$BE,$9F,$07,$6B,$FD,$69,$7E,$48,$C7,$C1,$4C,$3D,$7E,$95
	.dc.b	$AF,$4F,$C0,$F1,$F8,$2E,$4D,$AF,$83,$F0,$5D,$4E,$DF,$05,$63,$E2
	.dc.b	$FC,$0F,$84,$F1,$7C,$15,$8F,$93,$F0,$3E,$53,$CD,$F0,$56,$3E,$6F
	.dc.b	$C0,$FA,$7E,$0B,$D4,$D8,$FA,$3F,$05,$ED,$3D,$DF,$05,$63,$EA,$FC
	.dc.b	$17,$C1,$F8,$2F,$84,$D8,$FF,$7E,$7E,$1D,$F3,$9B,$C5,$F2,$61,$F1
	.dc.b	$B6,$31,$F0,$99,$F8,$4B,$7C,$9E,$1F,$09,$23,$AF,$C2,$5B,$FE,$35
	.dc.b	$8F,$F9,$62,$7C,$3F,$07,$DF,$E1,$3C,$0F,$87,$E1,$24,$F9,$BE,$12
	.dc.b	$E1,$CB,$E1,$3C,$C7,$9F,$E1,$2D,$7B,$3E,$0F,$D4,$7A,$FE,$12,$C7
	.dc.b	$C0,$F8,$3F,$77,$C2,$7B,$DB,$1F,$13,$E1,$3E,$11,$F0,$FE,$12,$C7
	.dc.b	$C8,$F8,$4F,$8C,$7C,$7F,$84,$B1,$F3,$3E,$13,$E5,$1F,$2F,$E1,$2C
	.dc.b	$7E,$73,$D0,$FD,$0F,$C0,$FE,$AE,$30,$7E,$83,$CE,$47,$FA,$CE,$33
	.dc.b	$C5,$FA,$38,$E0,$FF,$45,$74,$DF,$84,$FD,$14,$7F,$AC,$A7,$8D,$74
	.dc.b	$9C,$FC,$36,$BE,$16,$56,$AD,$F0,$B8,$C7,$B4,$74,$F8,$5E,$AD,$AE
	.dc.b	$1F,$0B,$E0,$3C,$3F,$0B,$63,$A7,$C2,$F9,$0F,$2F,$C2,$D8,$F0,$7C
	.dc.b	$1F,$A3,$E1,$3D,$2D,$8F,$27,$C2,$7B,$3E,$13,$DA,$D8,$F4,$7C,$27
	.dc.b	$C0,$F8,$4F,$82,$D8,$F6,$7C,$27,$C4,$3E,$2F,$C2,$58,$F8,$1F,$07
	.dc.b	$FC,$27,$F0,$FF,$E3,$1C,$7F,$36,$31,$F0,$B9,$F8,$59,$7F,$32,$DF
	.dc.b	$0B,$8C,$3F,$BC,$47,$F1,$0E,$CF,$E3,$ED,$AF,$BC,$BE,$0F,$C1,$F0
	.dc.b	$BE,$1F,$84,$95,$D7,$E1,$2E,$3D,$3E,$0F,$CD,$F0,$9E,$76,$D7,$8F
	.dc.b	$E1,$3D,$47,$AF,$E1,$2C,$79,$FE,$13,$DC,$7B,$FE,$12,$C7,$AF,$E1
	.dc.b	$3E,$17,$C2,$7C,$36,$C7,$BF,$E0,$FE,$37,$C2,$7C,$76,$C7,$C3,$F8
	.dc.b	$4F,$95,$F0,$9F,$2D,$B1,$F1,$FE,$0F,$E7,$1F,$3F,$E1,$2C,$7E,$5B
	.dc.b	$F1,$0F,$E2,$9C,$5F,$E2,$17,$F8,$9C,$66,$3F,$9F,$46,$1D,$7C,$26
	.dc.b	$3E,$13,$2D,$AE,$9F,$09,$C3,$E1,$38,$B6,$3C,$1F,$07,$D3,$E0,$FA
	.dc.b	$B6,$3C,$9F,$07,$E0,$3C,$3F,$07,$63,$D1,$F0,$7E,$43,$CB,$F0,$76
	.dc.b	$3D,$9F,$07,$E8,$F8,$3F,$4B,$63,$E0,$7C,$17,$B0,$F6,$FC,$1D,$8F
	.dc.b	$89,$F0,$7F,$03,$E0,$FE,$0B,$63,$E4,$7C,$1F,$C4,$3E,$2F,$C1,$D8
	.dc.b	$F9,$9F,$07,$F2,$3E,$0F,$E4,$B6,$3F,$82,$FE,$23,$FC,$33,$CD,$FE
	.dc.b	$EA,$FF,$0C,$6C,$7F,$6D,$8C,$3F,$C5,$D7,$C1,$E7,$E0,$F4,$D9,$EB
	.dc.b	$F0,$7C,$7E,$0F,$93,$63,$C3,$F0,$5D,$7E,$0F,$B3,$63,$CB,$F0,$7E
	.dc.b	$13,$C5,$F0,$76,$3D,$3F,$07,$E5,$3C,$DF,$07,$63,$DB,$F0,$5E,$9F
	.dc.b	$83,$F5,$36,$3E,$0F,$C1,$7B,$4F,$77,$C1,$D8,$F8,$BF,$05,$F0,$4F
	.dc.b	$85,$F0,$76,$3E,$4F,$C1,$FC,$53,$E3,$7C,$1D,$8F,$9B,$F0,$7F,$E1
	.dc.b	$5F,$1F,$F8,$6F,$47,$CE,$8C,$7C,$1E,$7E,$0E,$5F,$F3,$8B,$FE,$1A
	.dc.b	$D8,$E5,$F0,$12,$35,$F0,$37,$D0,$FF,$44,$5B,$E0,$7C,$07,$87,$E0
	.dc.b	$6D,$F5,$F8,$0F,$27,$C0,$F9,$5B,$1E,$1F,$81,$F4,$7C,$0F,$A5,$B1
	.dc.b	$E5,$F8,$1F,$67,$C0,$FB,$5B,$1E,$9F,$80,$F8,$07,$C1,$F8,$1B,$1E
	.dc.b	$DF,$80,$F8,$9F,$03,$F1,$5B,$1F,$07,$E0,$3E,$47,$C0,$FC,$96,$C7
	.dc.b	$C5,$F8,$1F,$98,$7C,$DF,$81,$B1,$FD,$83,$F1,$2F,$E1,$7D,$7F,$E5
	.dc.b	$07,$F8,$43,$79,$BF,$C9,$11,$89,$FE,$E4,$BE,$03,$3F,$03,$A6,$CF
	.dc.b	$5F,$80,$E2,$72,$F8,$1B,$1E,$1F,$81,$EA,$76,$EB,$B1,$E5,$F8,$1F
	.dc.b	$09,$E2,$F8,$1B,$1E,$9F,$81,$F2,$FC,$0F,$99,$B1,$ED,$F8,$0F,$49
	.dc.b	$EA,$F8,$1B,$1F,$07,$E0,$7D,$A7,$BB,$E0,$6C,$7C,$5F,$80,$F8,$3F
	.dc.b	$03,$F0,$9B,$1F,$27,$E0,$7E,$2F,$C0,$FC,$66,$C7,$CD,$F8,$0F,$93
	.dc.b	$F0,$3F,$29,$B1,$FE,$10,$FC,$4F,$F8,$63,$FE,$1D,$6F,$4F,$F9,$F5
	.dc.b	$FE,$16,$D8,$FF,$3D,$47,$F0,$C7,$97,$C0,$63,$E0,$72,$EB,$E0,$37
	.dc.b	$F0,$36,$BC,$0F,$2F,$80,$E7,$F0,$36,$7C,$8F,$6F,$81,$EF,$F0,$36
	.dc.b	$7D,$0F,$8B,$E0,$7C,$7F,$03,$67,$D8,$F9,$BE,$03,$CF,$F0,$36,$7E
	.dc.b	$03,$EA,$F8,$1F,$5F,$C0,$D9,$F8,$8F,$BB,$E0,$7D,$FF,$03,$67,$E4
	.dc.b	$3F,$0B,$E0,$7F,$D4,$7C,$AF,$2F,$FD,$53,$CC,$F8,$1C,$C6,$1B,$87
	.dc.b	$CF,$BB,$7F,$3F,$E0,$6F,$C3,$3F,$14,$FF,$5D,$4F,$F9,$19,$AF,$7B
	.dc.b	$FB,$15,$C2,$7F,$D6,$F3,$BF,$87,$C1,$9F,$83,$D7,$C1,$FF,$AC,$EB
	.dc.b	$FC,$41,$5F,$EB,$B7,$FC,$41,$87,$79,$D7,$C1,$48,$F1,$7C,$15,$BF
	.dc.b	$11,$C7,$E0,$7A,$1D,$7E,$0A,$D7,$A7,$E0,$7C,$1F,$05,$E1,$6C,$7B
	.dc.b	$7E,$07,$C9,$F0,$5E,$56,$C7,$C1,$F8,$2F,$41,$E9,$F8,$2B,$1F,$17
	.dc.b	$E0,$7D,$87,$B7,$E0,$AC,$7C,$9F,$81,$F8,$1F,$05,$F0,$5B,$1F,$37
	.dc.b	$E0,$BE,$21,$F1,$7E,$0A,$C7,$F9,$DB,$83,$EF,$F0,$FC,$0E,$94,$63
	.dc.b	$B3,$FD,$D2,$DA,$FF,$44,$3F,$EB,$8A,$E5,$F0,$1F,$7A,$3F,$8D,$46
	.dc.b	$0E,$DF,$03,$9F,$81,$91,$E1,$78,$7C,$0F,$1F,$81,$FF,$76,$5B,$E0
	.dc.b	$6C,$61,$CF,$C0,$CC,$FD,$5F,$03,$72,$F5,$60,$F8,$0E,$67,$4F,$81
	.dc.b	$B7,$F0,$7E,$07,$BF,$C0,$F8,$1B,$1F,$17,$E0,$7C,$67,$93,$E0,$6C
	.dc.b	$7C,$9F,$80,$F3,$9E,$8F,$81,$B1,$F3,$7E,$03,$E2,$7C,$0C,$DB,$07
	.dc.b	$D1,$F8,$0F,$90,$47,$EA,$7C,$0F,$EA,$3F,$15,$61,$7E,$46,$57,$FA
	.dc.b	$BA,$72,$DF,$82,$E0,$47,$5F,$07,$CF,$E0,$E4,$71,$7B,$7C,$1F,$7F
	.dc.b	$83,$93,$D5,$F1,$7C,$1F,$8F,$E0,$FC,$9F,$07,$2B,$C5,$1E,$7F,$83
	.dc.b	$F4,$7C,$1C,$9F,$34,$7A,$FE,$0F,$D9,$F0,$72,$7D,$51,$EF,$F8,$3F
	.dc.b	$5F,$91,$EC,$F2,$5C,$4F,$83,$CF,$2F,$FB,$66,$17,$8C,$57,$C2,$F0
	.dc.b	$F2,$F8,$3B,$66,$FC,$1E,$1C,$FC,$0C,$CB,$E7,$7C,$15,$CB,$F4,$D8
	.dc.b	$3E,$0B,$AF,$E2,$DF,$E2,$EB,$F2,$46,$3E,$1F,$3F,$07,$27,$FD,$CB
	.dc.b	$5F,$E2,$EC,$3F,$60,$A7,$9F,$C1,$74,$F8,$7B,$7C,$1E,$FF,$07,$FC
	.dc.b	$B2,$FF,$C6,$18,$74,$8C,$7C,$1E,$7C,$E9,$3E,$07,$87,$C1,$F1,$F8
	.dc.b	$2E,$5F,$07,$2B,$CB,$1D,$7E,$0B,$DA,$7B,$BE,$0E,$EB,$EA,$F8,$3F
	.dc.b	$83,$F0,$7F,$09,$B1,$EE,$F8,$3F,$8A,$7C,$6F,$83,$B1,$F0,$BE,$0B
	.dc.b	$E4,$FC,$1F,$CA,$6C,$7C,$6F,$83,$F9,$A7,$CE,$F8,$3B,$0D,$BF,$8C
	.dc.b	$7E,$2F,$FE,$30,$8C,$78,$08,$C4,$FE,$11,$FF,$18,$34,$FC,$2B,$FC
	.dc.b	$61,$E7,$F0,$7F,$D6,$20,$DF,$84,$C7,$C2,$4A,$F1,$78,$DF,$97,$B4
	.dc.b	$F1,$F8,$4E,$5F,$09,$CF,$E1,$26,$1C,$34,$F3,$E0,$AF,$07,$C2,$4A
	.dc.b	$D1,$E9,$7C,$9F,$09,$E1,$8F,$83,$95,$CF,$C5,$9F,$A7,$E1,$25,$72
	.dc.b	$DF,$E2,$47,$B7,$E1,$3D,$DF,$07,$EF,$F8,$49,$7B,$D9,$3D,$77,$F3
	.dc.b	$E9,$6B,$E6,$67,$F1,$FE,$13,$E4,$7C,$24,$9E,$51,$F2,$FE,$13,$E6
	.dc.b	$7C,$24,$9E,$D1,$F3,$FE,$12,$47,$85,$FA,$5F,$09,$F4,$FE,$12,$4F
	.dc.b	$95,$FA,$DF,$09,$FE,$0E,$F7,$BF,$91,$CF,$F5,$B5,$F0,$99,$35,$BF
	.dc.b	$69,$FC,$4E,$37,$9B,$ED,$65,$CF,$E1,$25,$7C,$9C,$7F,$DA,$2D,$F8
	.dc.b	$4F,$07,$C2,$78,$7E,$12,$4E,$E3,$A6,$7F,$7B,$3B,$3F,$93,$CB,$D1
	.dc.b	$F0,$9E,$9C,$A4,$F7,$8F,$67,$C1,$C8,$F1,$3E,$FF,$84,$F8,$1F,$07
	.dc.b	$F0,$7E,$12,$57,$9E,$3E,$27,$C1,$FC,$5F,$84,$93,$EB,$8F,$91,$F0
	.dc.b	$92,$3D,$CF,$CB,$F8,$4E,$B1,$F0,$70,$EA,$FC,$DC,$3E,$87,$C1,$FD
	.dc.b	$1F,$84,$93,$F1,$E3,$EA,$7C,$24,$8F,$94,$FD,$7F,$84,$EB,$1F,$5F
	.dc.b	$0E,$BD,$1F,$D0,$78,$59,$7D,$DF,$83,$97,$F4,$DF,$F2,$6F,$E3,$1F
	.dc.b	$E4,$AB,$FE,$90,$FF,$92,$27,$FC,$38,$D8,$CF,$C1,$F0,$F8,$7E,$3F
	.dc.b	$09,$2F,$F4,$D6,$8E,$BF,$09,$DB,$E0,$FB,$FC,$1C,$BD,$C7,$8B,$E0
	.dc.b	$FC,$7F,$07,$27,$9C,$75,$7C,$78,$5D,$3C,$F8,$FA,$BE,$0A,$57,$89
	.dc.b	$F6,$FC,$1F,$BB,$E0,$E4,$F9,$9F,$83,$F0,$7C,$FC,$2F,$E4,$3B,$FA
	.dc.b	$6B,$D7,$9F,$A9,$F6,$FC,$17,$C8,$23,$F0,$3E,$0F,$E5,$FC,$1C,$8F
	.dc.b	$86,$FC,$EF,$83,$F9,$FF,$05,$27,$E3,$BF,$4B,$E0,$FE,$9F,$C1,$FD
	.dc.b	$4F,$83,$95,$F3,$23,$FD,$07,$F8,$CF,$F9,$2E,$3F,$BB,$B7,$DE,$34
	.dc.b	$FF,$27,$47,$DA,$C3,$C0,$D3,$EB,$E1,$D3,$E0,$E4,$7F,$1F,$AF,$E7
	.dc.b	$9C,$DF,$E0,$B7,$F9,$DC,$67,$DA,$F1,$FB,$52,$30,$7B,$08,$77,$3F
	.dc.b	$17,$87,$AB,$E0,$BD,$7F,$07,$27,$C1,$1E,$EF,$83,$F7,$FC,$1C,$9F
	.dc.b	$24,$7C,$2F,$83,$F8,$7F,$07,$27,$D1,$1F,$1B,$E0,$7E,$3F,$C1,$C9
	.dc.b	$F6,$47,$CA,$F8,$1F,$97,$F0,$52,$7E,$04,$75,$7E,$16,$1D,$5F,$F3
	.dc.b	$F7,$07,$F9,$95,$F8,$DF,$03,$F5,$3E,$0A,$46,$0F,$95,$ED,$FD,$82
	.dc.b	$3F,$37,$DB,$EC,$F4,$C3,$B7,$2F,$CF,$DB,$EF,$FC,$14,$AF,$A8,$FE
	.dc.b	$0B,$E0,$BF,$D8,$BF,$8D,$7F,$BD,$BA,$CF,$FB,$D9,$BE,$27,$F8,$EA
	.dc.b	$30,$7F,$BD,$EF,$FD,$30,$DC,$7F,$C9,$0B,$E0,$78,$7C,$17,$1F,$81
	.dc.b	$98,$7F,$63,$A7,$C1,$63,$1E,$7F,$05,$DF,$D7,$F0,$36,$F8,$63,$D1
	.dc.b	$E5,$95,$BF,$1F,$C1,$79,$FE,$0B,$D0,$DA,$F0,$7B,$3E,$BF,$6B,$D8
	.dc.b	$D8,$EA,$4B,$F6,$E1,$D7,$7F,$6F,$C0,$FC,$3F,$82,$F8,$8D,$BF,$67
	.dc.b	$C1,$7C,$7F,$82,$F9,$0D,$8F,$81,$F0,$3F,$2C,$F9,$9F,$05,$63,$E2
	.dc.b	$7C,$17,$CF,$3E,$87,$C1,$58,$F9,$1F,$05,$FE,$6C,$F1,$FF,$CC,$1B
	.dc.b	$FF,$9C,$3A,$3F,$CC,$95,$FE,$0F,$9C,$F5,$7D,$CC,$FB,$5F,$D8,$CE
	.dc.b	$F0,$7F,$9F,$EF,$FC,$C5,$87,$1F,$81,$ED,$F0,$5D,$DF,$07,$B1,$E1
	.dc.b	$F7,$AD,$77,$79,$BF,$E4,$2D,$7F,$CF,$CF,$DF,$CB,$39,$FF,$A1,$B2
	.dc.b	$7C,$0F,$AC,$F6,$7C,$15,$BF,$57,$C0,$FB,$CF,$81,$F0,$56,$3D,$DF
	.dc.b	$05,$F0,$CF,$89,$F0,$56,$3E,$17,$C1,$7C,$7F,$82,$F9,$0D,$8F,$8D
	.dc.b	$F0,$3F,$2F,$E0,$7E,$63,$63,$A9,$36,$FF,$7B,$6C,$C7,$B4,$63,$E0
	.dc.b	$7E,$A1,$F5,$7E,$06,$DF,$D3,$F7,$FE,$C1,$F6,$7D,$CB,$1D,$7A,$7E
	.dc.b	$96,$FE,$A8,$C1,$FE,$6C,$CD,$F0,$1F,$81,$3F,$05,$F0,$36,$7F,$C9
	.dc.b	$DF,$8D,$FF,$3B,$3F,$9C,$E1,$FE,$BB,$6C,$7F,$9D,$9D,$7C,$17,$F9
	.dc.b	$EF,$65,$6E,$30,$D8,$EB,$F0,$D8,$BE,$5F,$07,$67,$FD,$83,$87,$F9
	.dc.b	$C9,$F1,$C6,$7C,$14,$BE,$0F,$0F,$F9,$C6,$BD,$3F,$07,$83,$3F,$07
	.dc.b	$AD,$24,$FB,$9E,$39,$F2,$DF,$9F,$C1,$CA,$F8,$71,$DB,$D9,$91,$F1
	.dc.b	$9F,$0F,$B5,$D2,$3E,$0E,$FF,$17,$1C,$70,$7C,$9C,$BC,$FF,$05,$E8
	.dc.b	$F8,$39,$87,$D0,$8F,$5F,$C1,$7B,$3E,$0E,$4F,$D4,$8F,$7F,$C1,$48
	.dc.b	$FA,$EF,$C2,$F8,$2F,$87,$F0,$5F,$13,$E0,$A5,$7D,$C8,$F8,$FF,$05
	.dc.b	$23,$EF,$BF,$2B,$E0,$BA,$47,$A8,$FA,$78,$63,$6F,$C1,$63,$F4,$FF
	.dc.b	$1C,$FF,$3B,$C7,$FB,$FD,$FF,$3B,$D7,$FA,$5D,$AF,$3E,$A4,$70,$E2
	.dc.b	$FF,$5F,$BF,$E7,$49,$D3,$FD,$05,$E5,$ED,$77,$F7,$FC,$55,$F9,$1C
	.dc.b	$AD,$7E,$23,$2F,$27,$C1,$F9,$7E,$0E,$4F,$8E,$3D,$1F,$05,$E9,$F8
	.dc.b	$39,$3E,$78,$F6,$7C,$17,$B7,$E0,$E4,$FA,$E3,$E0,$7C,$14,$8F,$73
	.dc.b	$F0,$FE,$0F,$E2,$7C,$14,$9F,$84,$FC,$7F,$83,$EB,$1F,$17,$0E,$AF
	.dc.b	$C0,$C3,$E6,$7C,$17,$CD,$F8,$39,$3F,$2E,$3E,$87,$C0,$48,$F9,$CF
	.dc.b	$D3,$F6,$FA,$C7,$CA,$C3,$AF,$47,$D5,$C7,$EC,$7C,$1F,$D9,$F8,$39
	.dc.b	$7F,$5E,$3E,$E7,$C1,$C8,$FB,$4F,$EC,$7F,$1D,$FF,$3B,$74,$7F,$BD
	.dc.b	$9F,$F3,$A3,$7E,$0F,$46,$FE,$13,$87,$C2,$4C,$3F,$DB,$57,$FE,$5D
	.dc.b	$7F,$08,$BE,$13,$B7,$C1,$F7,$F8,$39,$5C,$23,$A3,$CB,$1E,$EF,$4C
	.dc.b	$7C,$DF,$07,$E7,$F8,$39,$3E,$08,$F5,$7B,$52,$3C,$6F,$B7,$DF,$EB
	.dc.b	$1E,$4C,$3A,$BE,$6C,$3E,$17,$C1,$48,$F5,$BF,$17,$E0,$FE,$37,$C1
	.dc.b	$49,$F7,$BF,$27,$E0,$FE,$57,$C1,$FC,$BF,$83,$95,$F1,$23,$E7,$7C
	.dc.b	$1F,$CF,$F8,$39,$3F,$22,$3E,$97,$C1,$FD,$3F,$83,$93,$F3,$23,$AB
	.dc.b	$F3,$B0,$EA,$FC,$EC,$3F,$95,$71,$7F,$9D,$C9,$18,$F9,$3C,$FB,$7A
	.dc.b	$F6,$E4,$8F,$DD,$57,$F9,$D3,$AF,$F0,$58,$76,$F0,$11,$F8,$5C,$BC
	.dc.b	$77,$C3,$E1,$3C,$3F,$09,$23,$A5,$F0,$5E,$AB,$2F,$E8,$51,$E0,$F8
	.dc.b	$2F,$49,$1E,$39,$79,$6B,$87,$C1,$FB,$BE,$0B,$DF,$F0,$72,$F6,$76
	.dc.b	$C5,$3F,$13,$E0,$FE,$2F,$C1,$CA,$DE,$3E,$39,$F9,$3F,$05,$27,$38
	.dc.b	$23,$E6,$FC,$1F,$86,$73,$97,$85,$F9,$39,$7D,$2F,$82,$FA,$7F,$07
	.dc.b	$27,$9C,$7D,$6F,$73,$EB,$FB,$92,$7B,$C7,$57,$E9,$E1,$D7,$8B,$E5
	.dc.b	$6D,$F2,$72,$FB,$FF,$05,$F8,$0F,$83,$98,$7A,$23,$F0,$7F,$07,$23
	.dc.b	$D6,$FE,$1B,$E0,$FF,$B8,$7E,$3D,$FC,$DE,$7E,$AB,$FC,$F3,$2E,$ED
	.dc.b	$E6,$FA,$4E,$F3,$3F,$CF,$67,$7F,$05,$D0,$EB,$F0,$7D,$BE,$0B,$BF
	.dc.b	$C1,$CC,$BE,$8E,$09,$F1,$E5,$E4,$F8,$99,$5A,$8F,$3F,$C1,$C8,$E2
	.dc.b	$FA,$BE,$13,$D7,$ED,$FB,$3D,$C9,$5D,$A3,$AB,$E0,$C3,$AB,$FD,$67
	.dc.b	$2F,$60,$F8,$4F,$8A,$48,$F3,$FC,$27,$C8,$F8,$49,$1E,$A7,$E5,$FC
	.dc.b	$27,$CC,$F8,$3F,$9B,$F0,$92,$BD,$F1,$F4,$3E,$13,$E8,$FC,$24,$9F
	.dc.b	$87,$1F,$53,$E0,$E4,$7C,$67,$EB,$FC,$27,$58,$F9,$38,$75,$7D,$58
	.dc.b	$7D,$CF,$84,$91,$F3,$9F,$F1,$77,$E3,$FF,$CF,$A3,$F6,$6F,$F3,$E6
	.dc.b	$FB,$9A,$37,$EE,$F7,$8F,$8D,$87,$7E,$CF,$C2,$E3,$D3,$E1,$65,$7E
	.dc.b	$B3,$1F,$E8,$55,$F9,$87,$FA,$14,$67,$E0,$BC,$7F,$07,$23,$07,$1F
	.dc.b	$83,$F3,$F9,$D2,$30,$75,$F8,$3F,$59,$1F,$07,$C1,$FB,$BE,$0F,$DF
	.dc.b	$F0,$72,$7C,$91,$F0,$B0,$E9,$1F,$E1,$9C,$FF,$8A,$3D,$8C,$7C,$1F
	.dc.b	$C8,$F8,$39,$1E,$D8,$F9,$7E,$E7,$CC,$F7,$24,$FC,$18,$EC,$FC,$3D
	.dc.b	$7B,$F0,$7C,$3C,$BE,$9F,$C1,$7D,$4F,$83,$97,$F2,$63,$EB,$FC,$17
	.dc.b	$D8,$F8,$39,$3F,$36,$3F,$CD,$1F,$90,$7F,$9D,$63,$FC,$64,$FF,$9E
	.dc.b	$9F,$E5,$ED,$F3,$CE,$F2,$FE,$16,$FF,$3F,$9D,$FC,$1F,$43,$AF,$C1
	.dc.b	$F6,$F8,$3F,$E7,$57,$C1,$BE,$77,$F3,$5C,$3F,$CF,$0F,$F2,$C5,$1A
	.dc.b	$C3,$C7,$F0,$57,$2F,$1B,$69,$F0,$56,$31,$EC,$CC,$7A,$7B,$57,$1F
	.dc.b	$7F,$C0,$FC,$2F,$82,$91,$E1,$8F,$8B,$F0,$52,$3C,$8F,$C8,$F8,$2F
	.dc.b	$93,$F0,$32,$7D,$0F,$CC,$F8,$2F,$9B,$F0,$32,$7D,$8F,$D0,$F8,$2F
	.dc.b	$A3,$F0,$3F,$4B,$E0,$A5,$7C,$18,$FA,$BF,$05,$23,$E2,$3F,$63,$E0
	.dc.b	$BF,$D4,$1E,$DF,$F6,$CE,$1F,$EA,$6B,$DB,$F8,$CE,$6F,$FC,$DC,$E6
	.dc.b	$77,$F0,$5A,$37,$F0,$56,$39,$FC,$0F,$23,$9F,$C1,$58,$EF,$F0,$5E
	.dc.b	$22,$63,$E2,$F8,$2F,$2E,$5E,$6F,$82,$93,$E6,$8F,$4F,$C0,$C8,$F4
	.dc.b	$BE,$CF,$81,$F6,$FC,$0F,$AB,$E0,$7D,$6D,$CF,$DF,$EC,$FB,$BD,$AF
	.dc.b	$7B,$63,$E1,$FC,$07,$C2,$3E,$1F,$C0,$D8,$F8,$FF,$03,$F1,$8F,$8F
	.dc.b	$F0,$36,$3A,$9F,$33,$0E,$AD,$3E,$7F,$C0,$FC,$E3,$F2,$9F,$90,$BE
	.dc.b	$3F,$13,$FE,$06,$87,$C1,$E8,$DF,$C1,$D9,$EF,$C5,$F5,$B1,$EE,$D8
	.dc.b	$E5,$F0,$7D,$8E,$FF,$0F,$63,$B7,$C2,$7F,$B8,$F2,$7C,$25,$31,$19
	.dc.b	$F6,$B5,$ED,$CC,$BC,$D1,$C7,$E1,$24,$7A,$5E,$9F,$09,$D7,$E0,$FB
	.dc.b	$7C,$24,$AF,$74,$75,$7E,$06,$1D,$5F,$F5,$F6,$6C,$7C,$DF,$05,$27
	.dc.b	$E3,$3E,$9F,$83,$F5,$7C,$14,$9F,$94,$FB,$7E,$0F,$B4,$74,$C7,$B7
	.dc.b	$BB,$FE,$1E,$CB,$B4,$FD,$1F,$82,$F8,$84,$7E,$A7,$C1,$FC,$7F,$82
	.dc.b	$91,$F5,$DF,$95,$F0,$7F,$E5,$D8,$FF,$6B,$7E,$45,$FC,$87,$9B,$FC
	.dc.b	$98,$FF,$25,$24,$6F,$2F,$F1,$7C,$7F,$25,$F8,$13,$49,$F0,$7B,$BE
	.dc.b	$7F,$07,$70,$FF,$2F,$D7,$F9,$7B,$0E,$73,$9C,$39,$FC,$3D,$AF,$25
	.dc.b	$F0,$F8,$8C,$19,$75,$F0,$71,$FF,$2C,$CA,$39,$7C,$27,$3F,$84,$95
	.dc.b	$A7,$B7,$C2,$73,$8E,$38,$76,$79,$9E,$3F,$83,$F2,$7C,$24,$8E,$F1
	.dc.b	$E7,$F8,$49,$1E,$27,$D5,$F0,$9E,$BF,$84,$F6,$7C,$24,$AF,$3C,$7B
	.dc.b	$FE,$13,$E0,$7C,$24,$9F,$5C,$7C,$3F,$83,$F8,$9F,$09,$27,$DF,$1D
	.dc.b	$5F,$83,$87,$57,$EE,$E1,$F2,$FE,$12,$47,$C6,$7E,$77,$C2,$7D,$1F
	.dc.b	$C8,$FF,$CB,$44,$EA,$C7,$C2,$F5,$BF,$C3,$E3,$D7,$AF,$F0,$79,$71
	.dc.b	$F8,$5E,$5F,$0B,$2F,$FD,$17,$97,$F9,$8F,$2F,$F4,$C3,$FE,$5D,$8C
	.dc.b	$FC,$27,$84,$F1,$7C,$2F,$8F,$E1,$64,$F2,$7C,$DF,$0B,$E7,$F8,$49
	.dc.b	$3D,$9F,$57,$C2,$FA,$FE,$17,$D9,$F0,$B2,$BC,$71,$D1,$70,$7F,$A2
	.dc.b	$B4,$65,$98,$C6,$5F,$17,$E1,$3E,$37,$C2,$CC,$3D,$B1,$F2,$7E,$17
	.dc.b	$E5,$7B,$B2,$7E,$0C,$73,$E1,$F8,$39,$59,$FF,$60,$E6,$CB,$E9,$7C
	.dc.b	$2F,$D3,$F8,$59,$5F,$2A,$3E,$B7,$C2,$C8,$F9,$AF,$D9,$F8,$5F,$B5
	.dc.b	$F0,$B2,$7E,$8B,$FA,$DF,$1F,$FC,$BD,$E6,$9F,$F2,$F3,$7E,$13,$46
	.dc.b	$FE,$13,$87,$C2,$4C,$3F,$D8,$B7,$FE,$5A,$E0,$FF,$69,$3F,$E5,$A8
	.dc.b	$F9,$71,$F7,$B1,$F1,$46,$0F,$7B,$E3,$F8,$5F,$27,$C1,$F9,$7E,$12
	.dc.b	$57,$68,$F4,$7C,$07,$A7,$DC,$93,$E2,$8E,$BD,$5F,$67,$1B,$97,$93
	.dc.b	$2F,$81,$F0,$9F,$07,$E1,$26,$3E,$A8,$F8,$9F,$07,$F1,$7E,$12,$4F
	.dc.b	$BA,$3E,$47,$C2,$48,$F8,$2F,$CB,$F8,$4F,$99,$F0,$92,$7E,$2B,$F3
	.dc.b	$FE,$13,$E8,$7C,$1F,$D1,$F8,$39,$5F,$2A,$3A,$3F,$17,$A7,$EA,$63
	.dc.b	$8E,$1F,$F6,$EF,$33,$3F,$B5,$F0,$73,$0F,$A6,$FD,$DF,$83,$FE,$55
	.dc.b	$F9,$27,$F2,$8C,$8D,$27,$BB,$D7,$97,$EC,$E9,$71,$5F,$E1,$72,$E7
	.dc.b	$F1,$12,$47,$9F,$1F,$E8,$43,$E1,$7C,$3F,$0B,$E2,$F8,$59,$5C,$38
	.dc.b	$3F,$DE,$F7,$FD,$07,$4E,$79,$F9,$BE,$13,$D3,$F0,$93,$0E,$15,$AB
	.dc.b	$FE,$79,$18,$3C,$EB,$E1,$3E,$09,$23,$58,$7C,$2E,$3F,$8B,$F0,$7F
	.dc.b	$1B,$E1,$25,$6B,$04,$FC,$AC,$FC,$53,$EF,$CF,$0E,$FE,$13,$E7,$FC
	.dc.b	$24,$8C,$1C,$FD,$CF,$A6,$47,$B7,$BB,$C9,$E8,$78,$6F,$1C,$1F,$1B
	.dc.b	$E0,$FE,$DF,$C2,$4A,$F3,$47,$DE,$F8,$39,$1E,$97,$F0,$3F,$09,$F8
	.dc.b	$2F,$84,$93,$ED,$7F,$0B,$F0,$9F,$DB,$FF,$06,$73,$DF,$85,$97,$F2
	.dc.b	$27,$2D,$F8,$5E,$04,$75,$F0,$BD,$9F,$C2,$63,$76,$FB,$F9,$77,$F8
	.dc.b	$49,$5D,$5F,$17,$C2,$F8,$FD,$DF,$27,$B9,$2B,$C5,$1C,$F6,$FC,$16
	.dc.b	$56,$7F,$03,$8F,$AF,$E1,$7D,$9F,$0B,$27,$D5,$1E,$FF,$84,$F8,$1F
	.dc.b	$0B,$27,$DD,$1F,$0F,$E1,$24,$7C,$17,$E3,$7C,$2F,$C7,$F8,$4F,$91
	.dc.b	$F0,$B2,$BE,$34,$7C,$BF,$83,$F9,$9F,$09,$27,$E5,$47,$47,$F7,$DC
	.dc.b	$EC,$71,$CB,$F3,$BC,$74,$B3,$FA,$DF,$09,$F5,$FE,$12,$63,$F5,$63
	.dc.b	$ED,$7B,$BF,$E2,$9F,$C1,$BF,$9E,$9F,$CF,$B0,$47,$6E,$6F,$D7,$C7
	.dc.b	$86,$FD,$C7,$DF,$C7,$97,$C2,$F3,$F8,$59,$3F,$AA,$C7,$FA,$39,$FD
	.dc.b	$05,$FE,$0E,$DD,$0D,$3F,$A2,$6B,$ED,$5F,$07,$E7,$F8,$49,$7D,$63
	.dc.b	$D5,$F0,$72,$3C,$0F,$B7,$E1,$3D,$DF,$05,$27,$C8,$FC,$1F,$83,$F8
	.dc.b	$57,$D2,$7F,$96,$63,$FE,$88,$7B,$1E,$BF,$82,$F9,$1F,$07,$23,$07
	.dc.b	$BF,$DC,$F9,$84,$7E,$17,$B9,$D9,$F8,$9A,$F7,$DB,$DD,$97,$D3,$F8
	.dc.b	$39,$5F,$25,$FA,$DF,$07,$F5,$FE,$0B,$EC,$7C,$1C,$AF,$9D,$1F,$6F
	.dc.b	$E0,$E4,$7D,$17,$EF,$7C,$1F,$F9,$9F,$8B,$F9,$F1,$27,$1F,$07,$9F
	.dc.b	$83,$D7,$C1,$C9,$FE,$47,$5F,$E7,$9E,$46,$3D,$A7,$E3,$BD,$9D,$FC
	.dc.b	$1F,$72,$3C,$BE,$0F,$C5,$EF,$48,$EA,$F9,$7D,$AE,$9C,$7F,$81,$C2
	.dc.b	$F8,$3F,$D0,$32,$6F,$EB,$F8,$39,$97,$99,$F7,$7C,$1F,$BF,$E0,$7F
	.dc.b	$89,$E5,$FA,$7D,$FF,$9B,$E5,$EC,$F8,$3C,$BF,$D3,$23,$38,$FF,$6D
	.dc.b	$5F,$05,$A8,$FE,$BF,$18,$9F,$8B,$F0,$59,$F0,$D3,$9A,$E5,$F0,$59
	.dc.b	$87,$C1,$58,$C3,$9F,$81,$D7,$C1,$5C,$BE,$8B,$C7,$39,$1E,$03,$EA
	.dc.b	$BD,$7E,$0B,$B7,$C1,$59,$FB,$2F,$87,$DB,$91,$F7,$3D,$BB,$AF,$BF
	.dc.b	$0E,$AF,$C1,$DE,$F0,$FF,$0F,$FB,$DF,$E8,$38,$7F,$66,$6D,$7F,$84
	.dc.b	$D7,$C1,$FC,$37,$35,$C7,$E0,$F0,$67,$E0,$EE,$1C,$FE,$0F,$87,$C1
	.dc.b	$F1,$6C,$77,$F8,$2E,$87,$5F,$83,$B1,$E3,$F8,$2F,$01,$E1,$F8,$3B
	.dc.b	$1D,$4F,$46,$1D,$5A,$7A,$FE,$0B,$D1,$F0,$7E,$96,$C7,$BF,$DF,$F6
	.dc.b	$1E,$DF,$6E,$C7,$5E,$2F,$85,$87,$56,$C7,$C7,$F8,$2F,$89,$F0,$7F
	.dc.b	$15,$B1,$F2,$FE,$0B,$E4,$1F,$27,$E0,$EC,$7C,$FF,$82,$F9,$9F,$07
	.dc.b	$F3,$5B,$1F,$4F,$E0,$FF,$25,$F9,$2B,$E1,$BE,$94,$61,$CF,$C2,$6B
	.dc.b	$E1,$2D,$FF,$88,$2D,$1D,$63,$EF,$63,$76,$FB,$CF,$6F,$87,$93,$C7
	.dc.b	$E2,$ED,$7F,$95,$70,$FE,$73,$3D,$7D,$EF,$2F,$BF,$E6,$7A,$77,$FE
	.dc.b	$37,$2B,$5F,$E3,$EF,$01,$1E,$CF,$86,$93,$E7,$F8,$6B,$9F,$7F,$3B
	.dc.b	$E0,$FC,$37,$C2,$6D,$7B,$7E,$1B,$E2,$FC,$37,$C6,$6C,$7C,$1F,$86
	.dc.b	$F9,$27,$CA,$F8,$6B,$1F,$17,$E1,$BE,$6F,$C3,$7C,$E6,$C7,$53,$E5
	.dc.b	$61,$D5,$A7,$CD,$F8,$5F,$AB,$F0,$DF,$59,$B1,$F4,$7D,$FF,$B2,$7D
	.dc.b	$AF,$7A,$C7,$5D,$FE,$66,$1D,$5B,$1F,$E5,$8F,$C9,$BF,$CD,$ED,$8F
	.dc.b	$E2,$90,$F8,$7B,$18,$F8,$79,$5C,$7E,$1E,$CF,$FA,$36,$DF,$0F,$C8
	.dc.b	$E7,$F0,$F6,$FC,$1F,$0F,$D8,$EF,$F0,$F6,$3C,$9F,$0D,$E2,$F8,$6F
	.dc.b	$1B,$63,$D1,$87,$3C,$BC,$18,$DE,$EF,$F0,$F6,$7F,$CD,$1B,$F0,$BE
	.dc.b	$CF,$86,$F6,$BE,$EF,$76,$65,$F0,$FD,$DB,$97,$AF,$3E,$93,$FE,$8A
	.dc.b	$C9,$9E,$2B,$FD,$37,$0F,$85,$F9,$3F,$0B,$F2,$9B,$7F,$3F,$E1,$7E
	.dc.b	$6F,$C2,$FC,$E6,$C7,$D3,$F8,$5F,$A3,$F0,$BF,$49,$B1,$F5,$FE,$17
	.dc.b	$EA,$FC,$2F,$D6,$6C,$7F,$A4,$BD,$EF,$E5,$2F,$FA,$81,$BA,$FF,$AC
	.dc.b	$1F,$E5,$F1,$B3,$B1,$F8,$3C,$2C,$FD,$B7,$FA,$B4,$BE,$12,$FC,$26
	.dc.b	$1B,$7E,$1F,$84,$D9,$C3,$DD,$B1,$ED,$3B,$17,$1E,$CF,$A7,$E1,$3B
	.dc.b	$9E,$0F,$84,$B1,$ED,$F8,$4F,$1E,$7E,$46,$C7,$C1,$F8,$3F,$39,$E8
	.dc.b	$F8,$4B,$1F,$17,$E1,$3D,$67,$B3,$E1,$2C,$7C,$9F,$83,$F7,$FC,$27
	.dc.b	$C0,$6C,$7C,$DF,$84,$F8,$67,$C4,$F8,$4B,$1F,$47,$E1,$3E,$3F,$B5
	.dc.b	$F2,$1B,$1F,$57,$E1,$3F,$C0,$1F,$93,$FF,$2F,$8C,$1F,$EC,$A7,$1C
	.dc.b	$7F,$EF,$75,$F0,$9C,$3E,$13,$8B,$6B,$87,$C2,$74,$F8,$4E,$AD,$8E
	.dc.b	$9F,$05,$E0,$3C,$3F,$07,$63,$C1,$F0,$5E,$43,$CB,$F0,$56,$3C,$9F
	.dc.b	$03,$E8,$3D,$3F,$05,$63,$D1,$F0,$5F,$EF,$8D,$FD,$0D,$F4,$BF,$D2
	.dc.b	$1C,$08,$C5,$7F,$BD,$21,$ED,$62,$DF,$0E,$97,$C1,$E6,$31,$F0,$76
	.dc.b	$4D,$9C,$7C,$14,$BF,$98,$EF,$E0,$F8,$7C,$14,$9F,$A0,$F3,$F8,$3E
	.dc.b	$51,$F4,$F0,$EC,$61,$FA,$9F,$07,$E1,$F8,$39,$1F,$66,$3C,$DF,$94
	.dc.b	$7C,$38,$70,$FF,$AC,$63,$0D,$8D,$FC,$27,$0F,$84,$90,$38,$F1,$5E
	.dc.b	$16,$0A,$EF,$F0,$9E,$0F,$73,$C3,$EE,$4C,$79,$67,$E1,$9F,$2F,$C2
	.dc.b	$79,$BE,$12,$56,$B4,$ED,$9A,$BD,$7F,$07,$EC,$F8,$49,$7B,$3E,$BD
	.dc.b	$27,$E0,$FC,$27,$C2,$F8,$49,$5B,$E0,$F4,$CF,$C6,$F8,$4F,$8F,$F0
	.dc.b	$92,$B5,$8F,$F9,$A9,$BF,$05,$F3,$3E,$0E,$46,$0D,$FC,$17,$D0,$F8
	.dc.b	$39,$18,$39,$FC,$1F,$D4,$F8,$19,$18,$39,$1F,$38,$F5,$DF,$67,$1A
	.dc.b	$FD,$0C,$7E,$EF,$C1,$49,$F3,$BF,$80,$F8,$3F,$CB,$7E,$52,$E5,$A4
	.dc.b	$AF,$85,$1F,$E6,$07,$C3,$D1,$F8,$5C,$BC,$0F,$E1,$F1,$E7,$F0,$72
	.dc.b	$B8,$BD,$BE,$13,$BF,$C2,$7F,$A2,$F1,$37,$E1,$30,$E7,$E1,$26,$9E
	.dc.b	$2F,$84,$B9,$7B,$B0,$7C,$1F,$33,$A7,$C2,$5B,$F4,$FC,$27,$73,$C1
	.dc.b	$F0,$96,$3D,$B9,$74,$E0,$F8,$59,$7E,$F3,$26,$79,$BF,$CE,$43,$E1
	.dc.b	$3D,$5F,$09,$EB,$6C,$FC,$8F,$83,$F7,$7C,$27,$BD,$B1,$E0,$26,$3E
	.dc.b	$BC,$7B,$3E,$1F,$84,$F8,$DF,$09,$F1,$DB,$5F,$53,$E0,$FE,$51,$F2
	.dc.b	$FE,$12,$C7,$D8,$F8,$4F,$F5,$C7,$8B,$FC,$DB,$83,$DB,$DC,$E1,$56
	.dc.b	$BE,$23,$7F,$09,$31,$FD,$16,$3F,$E6,$86,$C7,$4F,$83,$91,$AF,$84
	.dc.b	$BC,$1F,$57,$1E,$93,$E3,$39,$E3,$E4,$F8,$4F,$2F,$C2,$59,$EE,$FA
	.dc.b	$3E,$12,$47,$8B,$E1,$2E,$3D,$63,$99,$D5,$FD,$0E,$BE,$F8,$FE,$24
	.dc.b	$FF,$B6,$1F,$5F,$C2,$7C,$4F,$84,$F8,$AF,$C6,$F8,$49,$3F,$03,$E1
	.dc.b	$2E,$1D,$3E,$0F,$E5,$9F,$33,$DF,$B5,$F1,$BE,$0F,$E7,$FC,$1F,$D0
	.dc.b	$6C,$7C,$AF,$83,$FA,$67,$D4,$F8,$3B,$1F,$3B,$0E,$78,$FA,$B1,$BD
	.dc.b	$5F,$63,$77,$C1,$7D,$C3,$EE,$FC,$1D,$C7,$F8,$C7,$E5,$5F,$E6,$6C
	.dc.b	$7F,$81,$BF,$E6,$58,$CC,$7F,$BE,$E3,$0F,$B6,$3C,$47,$E0,$31,$F0
	.dc.b	$BF,$CE,$E7,$F9,$D3,$DB,$E0,$F9,$7C,$1F,$37,$A7,$C1,$75,$F8,$3B
	.dc.b	$5E,$47,$C1,$F0,$5E,$1F,$83,$B3,$E8,$7C,$9F,$05,$23,$D7,$F0,$77
	.dc.b	$5F,$5F,$07,$A7,$E0,$FD,$4D,$AF,$83,$8F,$56,$9D,$4F,$8B,$F0,$7F
	.dc.b	$04,$F8,$5F,$07,$63,$E4,$FB,$7F,$14,$F8,$DE,$E5,$8E,$79,$7C,$CD
	.dc.b	$7D,$2D,$8F,$A3,$F0,$5F,$34,$F9,$DF,$05,$63,$EA,$FC,$17,$D1,$F8
	.dc.b	$2F,$A4,$D8,$FB,$3F,$05,$F5,$4F,$AD,$F0,$56,$3F,$AA,$7E,$57,$FE
	.dc.b	$66,$F4,$4D,$3F,$CC,$F9,$FF,$42,$71,$F0,$7F,$C7,$3B,$FF,$80,$C7
	.dc.b	$F9,$96,$F9,$FC,$1E,$5F,$EF,$11,$9B,$FC,$4E,$5E,$17,$F1,$79,$5C
	.dc.b	$39,$FC,$17,$19,$FF,$11,$DF,$F3,$CC,$3C,$FF,$07,$9D,$3C,$F7,$69
	.dc.b	$EF,$DF,$6F,$0D,$C3,$E1,$93,$1F,$0B,$72,$F5,$FC,$1F,$33,$A7,$C1
	.dc.b	$D9,$F8,$FF,$07,$DF,$E0,$FC,$0D,$8F,$97,$F0,$5E,$33,$C9,$F0,$76
	.dc.b	$3E,$7F,$C1,$79,$CF,$47,$C1,$D8,$FA,$7F,$05,$EB,$F8,$3F,$63,$63
	.dc.b	$A9,$36,$FA,$98,$75,$7E,$B7,$C1,$FC,$33,$E2,$7C,$1D,$AF,$BF,$EF
	.dc.b	$FC,$73,$E5,$7E,$59,$FC,$76,$FE,$7C,$73,$E2,$FB,$BB,$7A,$AF,$C1
	.dc.b	$1F,$D0,$2F,$FC,$EB,$1F,$E1,$D7,$F9,$E5,$FF,$82,$1C,$B7,$E2,$3B
	.dc.b	$FC,$1F,$83,$E0,$ED,$F1,$7C,$7F,$05,$23,$A7,$C1,$5C,$3A,$7C,$17
	.dc.b	$A0,$F4,$FC,$15,$AF,$17,$C0,$FB,$3E,$0B,$DA,$D8,$F3,$63,$CF,$6F
	.dc.b	$56,$56,$7F,$C9,$DB,$7F,$9D,$DF,$67,$C1,$7C,$6F,$82,$F8,$EF,$C8
	.dc.b	$F7,$A4,$FC,$1F,$72,$EF,$F0,$73,$E8,$FF,$9F,$31,$65,$8F,$11,$0F
	.dc.b	$81,$ED,$8F,$CB,$6D,$FC,$BF,$82,$FA,$E7,$D8,$F8,$2B,$5F,$47,$E0
	.dc.b	$BE,$D9,$F7,$3E,$0A,$C7,$D5,$F8,$2F,$BE,$7E,$A7,$93,$F9,$E7,$1F
	.dc.b	$DB,$8C,$7C,$3E,$7E,$06,$DF,$F9,$66,$FF,$9B,$E9,$F7,$72,$EA,$FD
	.dc.b	$FC,$AE,$5F,$E0,$4A,$7C,$17,$73,$C1,$F0,$57,$1E,$9F,$05,$E3,$F7
	.dc.b	$BC,$8D,$8F,$61,$32,$EC,$DC,$7B,$7B,$5E,$BF,$82,$F6,$36,$7D,$1F
	.dc.b	$05,$EF,$F8,$1F,$80,$D8,$F6,$7C,$17,$C3,$F8,$2F,$88,$D8,$F8,$1F
	.dc.b	$05,$F1,$CF,$91,$F0,$56,$3E,$27,$C1,$7C,$BF,$82,$F9,$8D,$8E,$A7
	.dc.b	$C9,$C3,$AB,$4F,$99,$F0,$5F,$4F,$E0,$7E,$A3,$63,$E8,$7C,$17,$D7
	.dc.b	$3E,$C7,$C0,$58,$EB,$C5,$FE,$76,$E7,$63,$DA,$30,$FD,$FF,$80,$95
	.dc.b	$FE,$93,$FC,$B7,$FA,$05,$A7,$F9,$64,$7F,$40,$E5,$FF,$1E,$BF,$CE
	.dc.b	$5B,$F1,$BA,$F8,$4D,$FC,$24,$C7,$A7,$C1,$DD,$3F,$D1,$38,$BE,$13
	.dc.b	$AF,$C1,$76,$6D,$F8,$FE,$0F,$C3,$F0,$7E,$26,$C7,$9F,$2E,$9A,$F9
	.dc.b	$33,$B3,$FE,$C5,$C9,$F0,$7E,$A3,$D7,$F0,$56,$7E,$07,$C1,$FB,$BD
	.dc.b	$BF,$7B,$63,$B1,$37,$F4,$E3,$D9,$FF,$7A,$E2,$F7,$FE,$3F,$C1,$FC
	.dc.b	$86,$DF,$CD,$F8,$3F,$97,$F0,$5F,$31,$B1,$F4,$7E,$0F,$E7,$9F,$43
	.dc.b	$E0,$EC,$7D,$5F,$83,$FA,$7F,$07,$F5,$1B,$1F,$EC,$BF,$CB,$BF,$9C
	.dc.b	$7B,$07,$F9,$7F,$27,$F9,$D1,$C9,$89,$F8,$D8,$F6,$DF,$EA,$E5,$6F
	.dc.b	$87,$C2,$7F,$6D,$C1,$F0,$96,$30,$E7,$E1,$35,$EE,$DC,$FC,$8F,$37
	.dc.b	$B9,$F8,$FC,$71,$93,$4F,$47,$BB,$DC,$F0,$7C,$25,$BF,$77,$C2,$78
	.dc.b	$CF,$27,$C2,$58,$F8,$5E,$2F,$9C,$F4,$7C,$1D,$8F,$8D,$F0,$9E,$B3
	.dc.b	$D9,$F0,$96,$3E,$57,$C2,$7B,$CF,$81,$F0,$76,$3A,$93,$6F,$F7,$1F
	.dc.b	$27,$F4,$56,$E5,$F1,$FE,$0F,$E4,$7C,$1D,$BF,$AC,$FC,$BF,$80,$F9
	.dc.b	$9F,$01,$67,$ED,$3E,$7C,$39,$C7,$DD,$DB,$1D,$3F,$6F,$E0,$FF,$C4
	.dc.b	$1F,$97,$FF,$38,$A5,$FC,$BD,$55,$A8,$DF,$C2,$70,$F8,$4B,$7C,$31
	.dc.b	$FF,$5C,$E3,$FC,$F1,$BF,$09,$D8,$EF,$F0,$96,$B8,$1E,$0E,$A5,$BE
	.dc.b	$0F,$CB,$F0,$9E,$66,$D7,$0D,$3F,$01,$4F,$83,$F5,$9E,$CF,$84,$B5
	.dc.b	$BC,$7A,$E7,$E1,$C1,$A7,$A3,$4C,$3E,$AF,$83,$F8,$BF,$09,$F1,$9B
	.dc.b	$8F,$3D,$1F,$09,$F2,$BD,$CF,$96,$D9,$D6,$7D,$1F,$75,$32,$C6,$7E
	.dc.b	$AF,$84,$FA,$7F,$09,$F5,$1B,$7D,$FE,$0F,$EB,$FC,$27,$D8,$6C,$78
	.dc.b	$FE,$0F,$ED,$9F,$73,$E1,$2C,$79,$FE,$13,$EF,$9F,$80,$F8,$4B,$1E
	.dc.b	$BF,$84,$FC,$19,$FE,$63,$F7,$3F,$CF,$5B,$FF,$3B,$9E,$C4,$DB,$E4
	.dc.b	$BE,$CC,$7B,$C6,$0F,$DA,$57,$F4,$19,$FB,$CF,$F4,$16,$FC,$27,$7F
	.dc.b	$84,$F0,$7C,$07,$87,$DD,$B5,$F6,$6F,$FC,$F1,$87,$5E,$AF,$C3,$EF
	.dc.b	$E0,$EA,$FC,$8E,$97,$0E,$FF,$07,$EB,$3D,$9F,$09,$63,$A7,$C1,$FB
	.dc.b	$FE,$13,$E0,$36,$3C,$1F,$09,$F0,$F5,$F8,$8D,$8F,$27,$C1,$FC,$7F
	.dc.b	$84,$F9,$0D,$8F,$47,$C1,$7C,$BF,$83,$F9,$8D,$8E,$A7,$B7,$0E,$AD
	.dc.b	$3E,$07,$C1,$7D,$3F,$83,$FA,$8D,$8F,$89,$F0,$3F,$5F,$DD,$FB,$0D
	.dc.b	$8E,$BC,$BF,$1F,$01,$D9,$C3,$F7,$FE,$0A,$4F,$CF,$F8,$3B,$5F,$3F
	.dc.b	$F3,$0F,$E8,$A4,$F8,$AF,$F5,$43,$8F,$85,$BF,$0B,$86,$E1,$FB,$19
	.dc.b	$7C,$2F,$0F,$85,$E2,$D9,$D7,$C2,$F4,$F8,$5E,$AD,$8E,$58,$74,$CF
	.dc.b	$C8,$DE,$D4,$FF,$0B,$C9,$F0,$BE,$63,$CF,$F0,$B6,$7C,$9F,$03,$EA
	.dc.b	$F7,$7D,$6D,$8E,$E7,$A7,$7F,$63,$4F,$67,$C2,$FC,$23,$E1,$FC,$2D
	.dc.b	$8F,$81,$F0,$BF,$1B,$E1,$7E,$3B,$63,$E2,$7C,$2F,$CA,$F8,$5F,$96
	.dc.b	$D8,$F9,$1F,$0B,$F3,$BE,$17,$E7,$B6,$3E,$67,$C2,$7D,$23,$FD,$B5
	.dc.b	$F9,$8F,$F3,$BE,$5F,$EC,$9C,$7F,$CE,$AB,$B9,$F4,$F2,$EE,$D3,$F9
	.dc.b	$35,$3E,$17,$9F,$C2,$F4,$6C,$EB,$DD,$EE,$78,$3D,$DB,$1D,$B2,$FC
	.dc.b	$46,$3D,$9B,$1D,$BE,$13,$CF,$F0,$BE,$86,$C7,$8B,$E1,$7F,$A9,$D7
	.dc.b	$EF,$F4,$FE,$7C,$E4,$F3,$FC,$2E,$BB,$3F,$03,$1F,$E0,$39,$7C,$26
	.dc.b	$70,$FE,$9D,$0F,$85,$C5,$6D,$BE,$BA,$78,$CF,$C4,$F8,$4C,$C7,$F4
	.dc.b	$08,$C1,$F2,$3E,$17,$3F,$09,$AF,$85,$93,$F3,$23,$8F,$C2,$72,$F8
	.dc.b	$59,$3F,$42,$3A,$FB,$BD,$BD,$D9,$3F,$52,$3C,$3F,$0B,$E2,$F8,$59
	.dc.b	$3F,$62,$3C,$BF,$0B,$23,$ED,$BE,$CF,$CC,$BE,$2C,$7F,$29,$5F,$0B
	.dc.b	$92,$47,$F6,$78,$7C,$2F,$1F,$85,$93,$97,$A7,$C2,$F5,$F8,$59,$3C
	.dc.b	$5F,$07,$C2,$F8,$7E,$13,$C5,$F0,$B2,$BB,$47,$97,$E1,$7C,$DE,$EC
	.dc.b	$9F,$14,$7A,$7E,$12,$47,$95,$F6,$7C,$2F,$B7,$E1,$24,$FA,$5F,$81
	.dc.b	$F0,$BF,$07,$E1,$3F,$0B,$E1,$FE,$1B,$C0,$71,$A7,$3E,$17,$F8,$4E
	.dc.b	$1F,$0B,$8C,$BF,$C9,$D1,$B9,$F8,$DF,$09,$9C,$39,$FC,$2D,$A4,$D8
	.dc.b	$C7,$C2,$CB,$F9,$DF,$0B,$71,$FE,$B9,$E3,$3E,$33,$F2,$6B,$E1,$A9
	.dc.b	$88,$CF,$C1,$4A,$FF,$0F,$52,$38,$FB,$DC,$BE,$16,$56,$9E,$BF,$0D
	.dc.b	$DB,$E1,$64,$F2,$7C,$3F,$0D,$D6,$3C,$B8,$5E,$5F,$95,$97,$F9,$F2
	.dc.b	$7A,$10,$F5,$E3,$D9,$F2,$FC,$37,$B0,$8F,$A3,$E1,$BA,$70,$7F,$90
	.dc.b	$EF,$D5,$3E,$CC,$7D,$6F,$BB,$4F,$5B,$D8,$C1,$F0,$BE,$17,$E4,$91
	.dc.b	$F8,$BF,$0D,$F3,$3E,$1A,$47,$C8,$7E,$7F,$C3,$7D,$0F,$85,$93,$F3
	.dc.b	$1F,$A7,$F0,$DF,$53,$E1,$A4,$FD,$07,$EB,$FB,$DD,$23,$ED,$E1,$78
	.dc.b	$BE,$3E,$3D,$27,$EF,$E3,$67,$E5,$E3,$F8,$0F,$85,$FC,$0F,$C3,$4B
	.dc.b	$FB,$B1,$CF,$87,$E3,$65,$67,$FD,$45,$C2,$CF,$A5,$7C,$BC,$ED,$6B
	.dc.b	$2F,$C6,$FC,$37,$F2,$0F,$CC,$FF,$97,$CF,$F8,$B5,$BA,$FD,$E9,$D7
	.dc.b	$C2,$EF,$E1,$A5,$FF,$9B,$6F,$FA,$43,$7E,$1B,$A7,$C3,$48,$C1,$BF
	.dc.b	$85,$F0,$11,$E5,$F0,$DD,$9E,$98,$76,$7C,$98,$F9,$F0,$91,$D0,$F4
	.dc.b	$7C,$37,$AF,$E1,$BD,$9F,$0D,$27,$CD,$1D,$5F,$76,$1D,$76,$FE,$09
	.dc.b	$AF,$AA,$7E,$16,$3D,$9A,$7B,$F1,$F9,$1F,$0D,$F2,$79,$24,$FC,$38
	.dc.b	$F9,$9F,$0B,$F3,$7E,$12,$4F,$C7,$8F,$A1,$F0,$BF,$47,$E1,$24,$FC
	.dc.b	$B8,$FA,$9F,$0B,$F5,$7E,$12,$4F,$CF,$8E,$AF,$D1,$C3,$AB,$F6,$70
	.dc.b	$EA,$FD,$5C,$3A,$BF,$77,$0F,$C0,$7C,$2F,$E0,$7E,$12,$4F,$DB,$8E
	.dc.b	$AF,$E0,$F0,$EB,$C1,$F1,$B1,$EB,$3F,$87,$C7,$AC,$FC,$AC,$7F,$19
	.dc.b	$F0,$B2,$BF,$0C,$FE,$3F,$E1,$7F,$B7,$FE,$69,$FD,$52,$FF,$9A,$37
	.dc.b	$BB,$FD,$88,$DB,$FF,$7F,$3F,$D6,$1F,$F5,$E3,$FD,$59,$BF,$09,$D0
	.dc.b	$EB,$F0,$BD,$BE,$17,$BF,$C2,$CA,$E3,$1E,$2C,$24,$74,$7C,$BF,$0D
	.dc.b	$CA,$3A,$90,$FF,$49,$63,$E1,$6F,$C3,$7B,$3E,$16,$47,$9A,$3A,$BC
	.dc.b	$2E,$E9,$F0,$30,$ED,$3D,$34,$F5,$3E,$ED,$3E,$3F,$C3,$7C,$8F,$85
	.dc.b	$93,$F0,$A3,$E5,$FC,$37,$CC,$F8,$69,$3F,$1A,$3E,$7F,$C3,$7D,$0F
	.dc.b	$86,$93,$F2,$A3,$E9,$FC,$37,$D4,$F8,$69,$3F,$3A,$3A,$BF,$43,$0E
	.dc.b	$AF,$C2,$C3,$AB,$F5,$30,$EA,$FC,$6C,$3E,$FF,$C3,$48,$FB,$2F,$E0
	.dc.b	$BE,$13,$AC,$7C,$CC,$3A,$F0,$7E,$13,$1E,$B3,$F4,$31,$EB,$3F,$88
	.dc.b	$C7,$F1,$7F,$09,$F8,$CF,$84,$97,$F8,$68,$FC,$7F,$C2,$7F,$8F,$BF
	.dc.b	$0F,$FE,$DD,$23,$FD,$98,$E1,$FB,$B1,$BF,$85,$E1,$EF,$49,$E1,$97
	.dc.b	$F8,$2E,$FF,$BB,$46,$39,$FA,$7C,$3A,$7F,$B3,$4F,$6F,$85,$F1,$1E
	.dc.b	$3F,$83,$F2,$C7,$11,$5C,$7E,$1B,$D0,$6A,$3C,$F9,$F8,$DF,$4E,$7E
	.dc.b	$77,$95,$78,$7E,$0B,$E0,$7C,$34,$9C,$E6,$9E,$9A,$FA,$F2,$EF,$5F
	.dc.b	$AE,$DB,$A3,$D8,$E3,$96,$2B,$E1,$7C,$37,$CD,$F8,$69,$5D,$A3,$E8
	.dc.b	$7C,$34,$8F,$0B,$F4,$FE,$1B,$EA,$7C,$34,$9F,$2B,$F5,$FE,$0B,$EC
	.dc.b	$7C,$14,$9F,$4B,$F6,$FE,$0B,$A4,$7D,$AC,$2E,$7F,$57,$1E,$93,$F7
	.dc.b	$B1,$B3,$F6,$71,$FC,$27,$C1,$4A,$F8,$AF,$E1,$FE,$1B,$9F,$47,$DC
	.dc.b	$E3,$67,$8B,$FA,$F0,$CF,$AD,$6F,$3B,$5A,$CB,$F2,$5F,$0D,$23,$E9
	.dc.b	$BF,$95,$F8,$5F,$CB,$7C,$37,$E5,$FE,$1A,$57,$D8,$8F,$F4,$BE,$1F
	.dc.b	$DD,$B1,$FC,$7B,$67,$F1,$2D,$E9,$3F,$D7,$78,$3F,$CC,$4B,$E0,$39
	.dc.b	$FB,$F2,$FF,$13,$39,$7C,$07,$E2,$32,$F1,$BF,$A8,$CB,$C6,$F9,$3E
	.dc.b	$1A,$3C,$8F,$3F,$C3,$7A,$3E,$1A,$47,$68,$F7,$BD,$48,$75,$39,$1E
	.dc.b	$49,$EA,$B3,$F5,$BF,$13,$E0,$A3,$EA,$7E,$3F,$C2,$7C,$8F,$85,$F9
	.dc.b	$3F,$07,$2B,$DF,$1F,$33,$E1,$24,$7C,$27,$E7,$FB,$FF,$43,$E0,$3E
	.dc.b	$8F,$BF,$2B,$E3,$C7,$57,$E4,$E1,$D5,$F3,$72,$FD,$6C,$7A,$C6,$0F
	.dc.b	$AF,$F0,$1F,$70,$CE,$BF,$5B,$E0,$FE,$F9,$27,$EA,$FC,$27,$77,$97
	.dc.b	$47,$DE,$CB,$C1,$7F,$7B,$1E,$95,$F7,$31,$AF,$0E,$6F,$F6,$E3,$7E
	.dc.b	$0F,$FD,$39,$06,$FC,$26,$1C,$FC,$26,$BE,$12,$EF,$F8,$87,$FD,$49
	.dc.b	$F9,$AA,$29,$8F,$80,$AF,$F1,$68,$FF,$67,$B7,$DF,$E3,$F0,$1C,$9E
	.dc.b	$BA,$79,$F0,$BB,$FE,$69,$EF,$F0,$93,$0E,$9F,$0B,$6F,$E2,$E4,$F8
	.dc.b	$3F,$2F,$C2,$79,$9B,$7F,$A0,$DB,$D8,$D3,$D8,$79,$FE,$0F,$DA,$7B
	.dc.b	$BE,$12,$C7,$AF,$DF,$F8,$3F,$01,$F0,$9B,$1E,$FF,$7F,$E2,$FC,$07
	.dc.b	$C6,$6C,$7C,$3F,$83,$F9,$3F,$09,$F2,$9B,$1F,$1F,$E1,$3E,$69,$F3
	.dc.b	$BE,$12,$C7,$53,$E6,$61,$D5,$A7,$52,$70,$FC,$4F,$83,$ED,$1F,$4F
	.dc.b	$E1,$3E,$C9,$F6,$BD,$EB,$1D,$6B,$E5,$DF,$56,$C7,$4E,$8F,$F7,$77
	.dc.b	$2B,$4F,$C1,$E5,$DA,$FF,$2F,$F9,$B7,$FB,$7F,$0F,$EA,$ED,$8C,$FC
	.dc.b	$1F,$FB,$B6,$31,$8F,$FB,$E5,$FF,$6F,$57,$0F,$85,$D7,$C0,$ED,$E1
	.dc.b	$F0,$B2,$7A,$FC,$25,$CB,$C3,$F0,$BD,$7E,$0F,$B3,$6B,$C7,$8F,$36
	.dc.b	$9D,$CF,$3E,$1C,$F3,$F2,$63,$67,$FC,$03,$9B,$E1,$3D,$47,$AF,$E0
	.dc.b	$2C,$F4,$D3,$E1,$EB,$EA,$8E,$D1,$F1,$B2,$E8,$FF,$81,$30,$FE,$F1
	.dc.b	$3E,$23,$E4,$FB,$FF,$1C,$F9,$1E,$4D,$8F,$9B,$EF,$FC,$B3,$E6,$7C
	.dc.b	$25,$8F,$A3,$F0,$9F,$3F,$E1,$3E,$83,$63,$EA,$FC,$27,$D3,$F8,$3F
	.dc.b	$A8,$D8,$EA,$4D,$3E,$EE,$1D,$5F,$BF,$7C,$DB,$3F,$80,$C3,$EF,$7C
	.dc.b	$1C,$8F,$C1,$FC,$05,$AF,$C3,$E7,$CF,$93,$F0,$59,$5B,$F9,$F9,$75
	.dc.b	$9F,$C3,$36,$3E,$9F,$C0,$FE,$2C,$FE,$29,$87,$F2,$8F,$CD,$FF,$C8
	.dc.b	$53,$F9,$29,$CD,$EB,$D7,$53,$E2,$73,$7F,$E3,$E8,$FE,$D4,$FF,$8C
	.dc.b	$1D,$37,$E1,$77,$8F,$F8,$F2,$5F,$0B,$8C,$3A,$B6,$1F,$07,$63,$1F
	.dc.b	$09,$31,$ED,$F0,$B7,$1F,$56,$7C,$73,$E4,$DA,$F2,$61,$CD,$F3,$63
	.dc.b	$6F,$E1,$7C,$2F,$84,$F1,$7C,$2D,$AF,$66,$1C,$EB,$BC,$7B,$4E,$A4
	.dc.b	$7E,$65,$F5,$DB,$E5,$6A,$F7,$BD,$C7,$BF,$DF,$B8,$7C,$9F,$85,$F8
	.dc.b	$47,$C3,$F8,$5B,$1F,$37,$E1,$3E,$31,$F1,$FE,$16,$C7,$D1,$F8,$3F
	.dc.b	$95,$F0,$BF,$2D,$B1,$D4,$9B,$7E,$07,$0E,$AF,$E0,$F0,$E6,$FE,$17
	.dc.b	$0B,$AF,$AB,$E1,$7E,$B1,$F5,$FE,$12,$E3,$D2,$7D,$FA,$7A,$63,$07
	.dc.b	$E1,$F1,$E6,$FE,$27,$0B,$8F,$E1,$7E,$17,$F0,$47,$E0,$FD,$DB,$5F
	.dc.b	$8D,$F8,$2F,$C3,$1F,$89,$F7,$3F,$B7,$46,$4F,$F4,$0A,$F8,$5D,$1B
	.dc.b	$F8,$4B,$3A,$F8,$5E,$47,$3F,$85,$B1,$CF,$3F,$D3,$63,$DE,$3C,$27
	.dc.b	$6C,$39,$CF,$F8,$4B,$1F,$EF,$6D,$8F,$37,$C2,$CA,$F2,$FC,$25,$AE
	.dc.b	$B1,$DF,$87,$F6,$18,$7B,$27,$27,$39,$F0,$1E,$ED,$3C,$35,$C3,$E1
	.dc.b	$7E,$21,$F1,$7E,$16,$D7,$C4,$F8,$5F,$91,$F0,$BF,$25,$B1,$F2,$32
	.dc.b	$F9,$9F,$0B,$F3,$5B,$1F,$33,$E0,$3E,$81,$F4,$7D,$FB,$1D,$4F,$A3
	.dc.b	$87,$56,$9D,$4F,$AB,$7D,$5A,$7D,$8F,$85,$FB,$9F,$0B,$F7,$5B,$1D
	.dc.b	$77,$F9,$D7,$D5,$B1,$D5,$FA,$57,$D5,$B1,$F8,$4F,$85,$FC,$47,$C1
	.dc.b	$7E,$25,$B1,$F8,$8F,$85,$FC,$67,$C2,$FE,$35,$B1,$FE,$CC,$FC,$E3
	.dc.b	$FB,$9B,$FE,$DE,$6E,$1F,$ED,$85,$F0,$BF,$C7,$5C,$FA,$27,$FB,$CC
	.dc.b	$FF,$96,$17,$BF,$BF,$83,$E0,$F1,$F8,$0E,$5F,$07,$72,$F1,$3D,$7E
	.dc.b	$13,$B7,$C2,$D9,$F3,$3E,$1F,$85,$FF,$7F,$67,$DF,$E1,$6F,$6F,$F3
	.dc.b	$DD,$7F,$D7,$73,$ED,$AC,$E1,$ED,$F8,$4B,$3E,$D6,$EF,$EB,$6D,$FF
	.dc.b	$AB,$A7,$E3,$7B,$D8,$F7,$A4,$60,$F9,$5E,$F7,$0F,$81,$91,$83,$E7
	.dc.b	$7C,$17,$4F,$83,$91,$83,$E9,$7C,$1F,$80,$8F,$D5,$F8,$3F,$1F,$C1
	.dc.b	$C8,$FB,$0F,$A3,$F3,$9F,$F6,$26,$AC,$7A,$D7,$D3,$C7,$D7,$1D,$23
	.dc.b	$EB,$63,$78,$BE,$46,$5C,$FD,$E9,$43,$FD,$27,$97,$FB,$3F,$8B,$1C
	.dc.b	$48,$FF,$66,$4F,$E0,$72,$F0,$E3,$F4,$F3,$F0,$CF,$B8,$87,$88,$F7
	.dc.b	$1D,$E3,$FB,$B5,$7F,$6C,$8C,$61,$E0,$8F,$ED,$71,$81,$87,$8F,$DE
	.dc.b	$F8,$9F,$03,$F1,$7E,$13,$E3,$7C,$24,$BE,$17,$E9,$9F,$95,$F0,$72
	.dc.b	$73,$8A,$3E,$77,$C1,$FC,$FF,$82,$FA,$1F,$05,$2F,$71,$DD,$E3,$97
	.dc.b	$77,$E4,$E3,$D5,$F9,$78,$5D,$BD,$F8,$FD,$BF,$76,$57,$89,$FB,$DF
	.dc.b	$01,$F7,$FD,$E9,$3E,$67,$F0,$5E,$F7,$E0,$FE,$06,$4F,$A9,$FC,$37
	.dc.b	$C1,$7E,$1F,$E0,$BF,$11,$F0,$52,$BD,$F1,$D1,$FB,$F8,$5D,$3E,$9E
	.dc.b	$3D,$27,$F0,$78,$D9,$FA,$F8,$FE,$4F,$E0,$BF,$29,$F0,$52,$FE,$5C
	.dc.b	$7E,$5F,$E0,$24,$7C,$E7,$F6,$3F,$9D,$7F,$8E,$A3,$FD,$C2,$FF,$5F
	.dc.b	$7E,$BA,$F7,$B6,$70,$F8,$1E,$3F,$07,$CB,$E0,$E5,$F0,$8E,$EF,$E6
	.dc.b	$B0,$BA,$7F,$90,$F1,$F0,$37,$0E,$B5,$E0,$C7,$B3,$E2,$F8,$3F,$39
	.dc.b	$1F,$2F,$C1,$F4,$F0,$3F,$2D,$95,$9F,$CA,$E5,$DA,$7F,$52,$FB,$63
	.dc.b	$DF,$87,$81,$F8,$38,$FC,$4F,$83,$91,$F0,$9F,$8F,$F0,$7F,$23,$E0
	.dc.b	$FE,$4F,$BD,$2B,$E3,$C7,$CC,$F8,$3F,$9B,$EF,$C9,$F9,$71,$F4,$3E
	.dc.b	$0E,$47,$CE,$7E,$9F,$C1,$F5,$8F,$A3,$87,$57,$EA,$E1,$CD,$FA,$F8
	.dc.b	$5E,$2F,$99,$8F,$DC,$F8,$2F,$BB,$F0,$72,$FE,$DC,$7E,$03,$E0,$BF
	.dc.b	$03,$F0,$72,$7E,$FC,$7E,$13,$E0,$BF,$0B,$F0,$72,$7F,$07,$1F,$CE
	.dc.b	$3C,$37,$C5,$5F,$84,$95,$FB,$E9,$CB,$D0,$FC,$5E,$3D,$DF,$C7,$E3
	.dc.b	$DD,$FC,$7E,$3D,$DF,$C9,$E3,$DF,$E1,$24,$75,$7C,$5F,$09,$E3,$F8
	.dc.b	$4F,$27,$C2,$4A,$F1,$47,$9F,$E1,$3D,$1F,$09,$27,$CD,$1E,$BF,$83
	.dc.b	$F6,$7C,$24,$9F,$54,$75,$7D,$98,$75,$5C,$BF,$E9,$DE,$06,$5D,$67
	.dc.b	$3E,$07,$F3,$89,$67,$9B,$E1,$9F,$CB,$F8,$49,$87,$C9,$7E,$77,$C3
	.dc.b	$F8,$63,$99,$FC,$C3,$06,$7E,$3C,$3C,$59,$F5,$CF,$C1,$9D,$C3,$86
	.dc.b	$7F,$67,$E1,$7F,$D5,$19,$3E,$16,$C6,$1C,$FC,$2C,$C7,$EE,$7C,$2D
	.dc.b	$C7,$C7,$EE,$31,$FE,$C5,$5F,$81,$F7,$FA,$FC,$07,$67,$BF,$C1,$F8
	.dc.b	$3E,$0E,$D7,$E1,$DE,$9C,$3F,$2F,$4F,$2E,$98,$AF,$95,$8F,$47,$F1
	.dc.b	$DF,$0B,$75,$F1,$FE,$77,$FE,$D2,$6F,$C2,$FF,$B1,$63,$19,$7F,$B4
	.dc.b	$DF,$F6,$6C,$F0,$F8,$8D,$7C,$2E,$DB,$8F,$F5,$8A,$FF,$65,$B7,$E0
	.dc.b	$7A,$7C,$37,$57,$B7,$C1,$C9,$E1,$F0,$77,$2E,$D1,$CF,$2E,$CD,$3B
	.dc.b	$61,$CD,$A7,$73,$C5,$F0,$7E,$BF,$84,$F6,$36,$3C,$DF,$0B,$EF,$F8
	.dc.b	$5F,$80,$D8,$F5,$74,$7C,$3F,$83,$F8,$8D,$8F,$77,$C1,$FC,$7F,$83
	.dc.b	$F9,$0D,$8E,$A7,$C3,$C3,$AB,$4E,$A7,$C7,$BE,$AD,$3E,$57,$C1,$FD
	.dc.b	$3F,$84,$FA,$8D,$8E,$BB,$FD,$AB,$EA,$D8,$EA,$FD,$EB,$EA,$D8,$FA
	.dc.b	$DF,$0B,$F7,$CF,$C0,$7C,$25,$8F,$B5,$F0,$BF,$83,$3F,$09,$F0,$36
	.dc.b	$3F,$C2,$BF,$9E,$7F,$A9,$FA,$91,$FE,$30,$8C,$37,$E0,$33,$97,$E8
	.dc.b	$A7,$2F,$F8,$91,$FF,$5C,$D6,$BC,$44,$3E,$17,$8C,$E6,$37,$87,$79
	.dc.b	$F2,$5F,$E7,$32,$F3,$CF,$A8,$F3,$9C,$8F,$31,$63,$C5,$8D,$C7,$C3
	.dc.b	$37,$07,$C3,$5F,$86,$C3,$67,$E1,$FC,$36,$CE,$1F,$09,$63,$E3,$FC
	.dc.b	$27,$33,$A7,$C0,$D8,$F9,$7F,$05,$DF,$E0,$7C,$0D,$8E,$A4,$DB,$E5
	.dc.b	$61,$D5,$F9,$98,$73,$7E,$76,$17,$87,$E4,$7C,$17,$AF,$E1,$3D,$8D
	.dc.b	$C7,$ED,$FC,$27,$BF,$E0,$FE,$03,$63,$EF,$FC,$27,$C3,$F8,$1F,$88
	.dc.b	$D8,$FC,$1F,$C1,$7C,$7F,$81,$F9,$0D,$8E,$A7,$E2,$30,$EA,$D3,$A9
	.dc.b	$F8,$CB,$EA,$D3,$F1,$FF,$05,$F4,$CF,$AD,$F9,$EF,$FA,$FB,$9D,$1D
	.dc.b	$38,$3F,$1D,$85,$DF,$F1,$0F,$48,$FC,$96,$36,$7F,$18,$F4,$F8,$4E
	.dc.b	$BF,$09,$73,$FF,$39,$E7,$FE,$CA,$BF,$CB,$3F,$EC,$A8,$CF,$C2,$79
	.dc.b	$3D,$FF,$2B,$E6,$F8,$0F,$3F,$BF,$6B,$9B,$EA,$F8,$0F,$5F,$C1,$D9
	.dc.b	$EE,$F2,$7A,$9E,$2C,$31,$B7,$F4,$CC,$BF,$AE,$CF,$43,$CF,$8F,$81
	.dc.b	$F4,$FC,$27,$C8,$3E,$4F,$C2,$58,$F6,$FC,$27,$CC,$F8,$4F,$9A,$D8
	.dc.b	$F8,$3F,$09,$F4,$3E,$07,$E8,$B6,$3E,$2F,$C0,$FD,$43,$EA,$FC,$0D
	.dc.b	$8E,$A7,$CA,$C3,$AB,$4E,$A7,$CE,$BE,$AD,$3E,$8F,$C2,$7E,$03,$E1
	.dc.b	$3F,$02,$D8,$FA,$BF,$09,$F8,$4F,$84,$FC,$2B,$63,$FD,$57,$87,$F5
	.dc.b	$C7,$FD,$8A,$FF,$5C,$E7,$FE,$20,$E7,$8B,$FD,$6E,$E6,$77,$F0,$7F
	.dc.b	$C1,$60,$DF,$7F,$0F,$82,$7F,$0F,$8D,$BF,$AC,$F1,$79,$7C,$25,$9F
	.dc.b	$2B,$D7,$E0,$3B,$7C,$1D,$9F,$4B,$CF,$3F,$C8,$E1,$67,$FD,$BF,$CE
	.dc.b	$8E,$98,$FF,$84,$34,$F5,$3E,$C3,$C3,$5F,$13,$E0,$FD,$9F,$07,$ED
	.dc.b	$6C,$7C,$8F,$82,$F8,$07,$C1,$F8,$3B,$1F,$33,$E0,$FE,$27,$BD,$F1
	.dc.b	$5B,$1F,$43,$E0,$FE,$41,$F2,$7D,$FB,$1D,$49,$9F,$9F,$0E,$AF,$A6
	.dc.b	$F9,$B6,$7E,$DE,$1F,$4F,$E0,$BE,$A7,$C1,$D9,$FB,$EF,$D7,$F8,$2F
	.dc.b	$B1,$F0,$76,$7F,$06,$FD,$BF,$83,$91,$F8,$6F,$83,$B9,$7C,$9F,$81
	.dc.b	$FC,$07,$C1,$FE,$05,$B5,$F8,$CC,$7A,$B4,$EA,$7E,$43,$0E,$7C,$FF
	.dc.b	$8E,$C6,$CB,$17,$C0,$BF,$3E,$FE,$CA,$DF,$83,$FE,$A9,$18,$0E,$3E
	.dc.b	$1F,$3F,$0B,$70,$E9,$97,$F6,$BB,$FE,$D5,$18,$3C,$18,$77,$F8,$4E
	.dc.b	$A7,$6F,$85,$B3,$BC,$FC,$B6,$F8,$5F,$19,$E4,$F8,$3B,$5B,$C7,$D5
	.dc.b	$A7,$92,$31,$7F,$DE,$30,$E7,$A7,$67,$FC,$CB,$1F,$DB,$A7,$79,$67
	.dc.b	$4F,$57,$C1,$FC,$2F,$83,$F8,$6D,$AE,$B9,$77,$C7,$9B,$6F,$E1,$6D
	.dc.b	$CE,$BE,$0E,$4C,$B1,$87,$BB,$E1,$3E,$7F,$C2,$FD,$06,$DF,$AB,$E1
	.dc.b	$3E,$99,$F5,$3E,$16,$C7,$BB,$E1,$7E,$BF,$C0,$7D,$86,$C7,$C2,$F8
	.dc.b	$5F,$B6,$7D,$CF,$81,$B1,$D4,$99,$FC,$4C,$3A,$BF,$1A,$F9,$B6,$7E
	.dc.b	$6E,$1F,$86,$F8,$49,$1F,$43,$E1,$6D,$7C,$EF,$84,$FF,$0F,$7E,$7F
	.dc.b	$FE,$D4,$AF,$AB,$18,$F8,$8C,$FC,$36,$BE,$1A,$D7,$DD,$BF,$EC,$4F
	.dc.b	$E2,$9F,$EC,$58,$7F,$A7,$9F,$F5,$EC,$6F,$E1,$BB,$FC,$1F,$E6,$B2
	.dc.b	$ED,$3E,$03,$A6,$5D,$B0,$E5,$8D,$E8,$FF,$22,$65,$E0,$9F,$D3,$BF
	.dc.b	$D8,$1F,$07,$C3,$7B,$3E,$13,$DA,$FB,$BE,$1A,$65,$E9,$F8,$6B,$7E
	.dc.b	$7F,$86,$F8,$67,$C4,$F8,$6B,$5E,$FF,$86,$F8,$E7,$C8,$F6,$2C,$7C
	.dc.b	$3C,$79,$EB,$E6,$CA,$CF,$C5,$C7,$9E,$5F,$17,$2B,$7E,$CF,$86,$FA
	.dc.b	$67,$D4,$F8,$4B,$87,$52,$5F,$CD,$C7,$AB,$F0,$B1,$E7,$3F,$4F,$1B
	.dc.b	$3F,$1B,$E1,$BE,$F9,$F8,$0F,$86,$B5,$F6,$FE,$17,$F0,$67,$E1,$3E
	.dc.b	$1A,$C7,$DF,$F8,$6F,$C3,$9F,$88,$F8,$0B,$1F,$BF,$FD,$03,$FB,$47
	.dc.b	$9C,$69,$FD,$41,$C6,$7D,$0F,$C3,$63,$DE,$BC,$47,$72,$47,$E2,$EF
	.dc.b	$BB,$FE,$EA,$BF,$EE,$33,$E1,$F8,$7F,$F3,$C4,$62,$75,$18,$F8,$5C
	.dc.b	$FC,$2D,$BF,$43,$C3,$E0,$E4,$7A,$FE,$12,$E1,$DF,$E1,$3A,$9D,$BE
	.dc.b	$0A,$D7,$C1,$C7,$AB,$4E,$A7,$C5,$C3,$9F,$47,$FA,$77,$46,$58,$AF
	.dc.b	$E6,$70,$F8,$2F,$59,$EC,$F8,$2B,$3F,$3F,$E1,$3D,$E7,$C0,$F8,$4B
	.dc.b	$1F,$4F,$E1,$3E,$1F,$C2,$7C,$46,$C7,$D7,$F8,$1F,$8F,$F0,$9F,$21
	.dc.b	$B1,$D4,$9C,$5F,$3F,$0E,$AF,$D1,$BE,$6D,$9F,$C1,$61,$F4,$BE,$0B
	.dc.b	$A4,$7E,$15,$B3,$F6,$72,$E7,$BF,$C3,$C6,$DF,$C2,$C7,$AC,$FC,$76
	.dc.b	$C7,$C6,$F8,$2F,$BF,$F0,$7F,$80,$6C,$7E,$4F,$E1,$3F,$06,$7F,$AE
	.dc.b	$FD,$8F,$ED,$7C,$48,$C7,$C2,$4A,$FE,$B1,$5F,$D5,$7B,$4F,$F6,$E8
	.dc.b	$FC,$F6,$1D,$DD,$FC,$15,$DF,$F2,$98,$F8,$27,$99,$61,$AF,$FB,$77
	.dc.b	$0E,$91,$89,$FE,$E7,$18,$9D,$7C,$07,$96,$B4,$7F,$81,$A7,$FB,$33
	.dc.b	$8C,$BC,$FE,$F7,$BF,$E0,$3E,$03,$F0,$7E,$13,$E1,$7C,$25,$CF,$DE
	.dc.b	$FC,$5F,$83,$91,$F0,$BE,$12,$E5,$D7,$3E,$FA,$FA,$72,$ED,$7F,$21
	.dc.b	$E9,$5B,$C6,$D7,$BD,$E6,$FC,$1C,$7E,$A6,$38,$C3,$E0,$FC,$17,$D7
	.dc.b	$F8,$3F,$B0,$D8,$FA,$B9,$FD,$BF,$81,$FB,$8D,$8F,$B3,$F0,$3F,$7C
	.dc.b	$FC,$07,$C0,$D8,$FB,$B8,$73,$D7,$D9,$8D,$D3,$EA,$63,$CE,$7E,$03
	.dc.b	$63,$EC,$63,$CE,$7E,$26,$36,$7E,$E7,$C1,$FE,$3C,$E9,$97,$DC,$7C
	.dc.b	$EF,$E3,$B3,$B7,$F3,$5E,$51,$F8,$2C,$3D,$2F,$B0,$FC,$AF,$C1,$FE
	.dc.b	$6F,$E0,$FF,$38,$D8,$FC,$CF,$C1,$FE,$7C,$FD,$07,$C0,$58,$FF,$1A
	.dc.b	$7E,$83,$FD,$FF,$2F,$C4,$2F,$84,$4E,$BE,$0A,$D6,$A6,$ED,$8E,$96
	.dc.b	$1F,$DC,$DB,$F0,$5E,$3B,$EA,$F8,$5F,$C5,$67,$E4,$CB,$CE,$77,$3C
	.dc.b	$B8,$F6,$7F,$CA,$B8,$79,$F1,$F0,$3E,$78,$F5,$FC,$26,$F0,$FF,$34
	.dc.b	$CF,$F8,$4A,$B7,$F0,$99,$CF,$CF,$76,$BF,$C1,$15,$D3,$E0,$B1,$F0
	.dc.b	$79,$75,$F0,$33,$3F,$93,$EF,$DD,$3E,$4F,$BF,$CC,$E9,$EF,$DA,$F9
	.dc.b	$F8,$F3,$69,$F4,$B1,$E5,$97,$B3,$2F,$26,$58,$AF,$83,$F0,$7E,$73
	.dc.b	$D1,$F0,$76,$3E,$DF,$C1,$FA,$CF,$67,$C1,$D8,$FB,$FF,$07,$EF,$F8
	.dc.b	$3F,$80,$D8,$FC,$1F,$C0,$FC,$33,$E2,$7C,$1D,$8E,$A4,$E2,$FC,$3E
	.dc.b	$1D,$5F,$C4,$DF,$36,$CF,$E3,$B0,$F9,$DF,$07,$F3,$FE,$0E,$CF,$E4
	.dc.b	$9E,$5C,$1F,$73,$4F,$4C,$F7,$8F,$BF,$8F,$56,$C7,$53,$FD,$1D,$FA
	.dc.b	$17,$F8,$40,$FF,$07,$17,$E2,$3F,$00,$FF,$80,$E7,$F3,$0B,$E1,$73
	.dc.b	$F1,$3A,$77,$F0,$73,$0E,$BF,$0D,$77,$FE,$5D,$83,$E1,$3A,$FC,$27
	.dc.b	$66,$DF,$93,$2E,$8D,$3C,$F9,$73,$CF,$C7,$9F,$9B,$3C,$57,$FA,$D3
	.dc.b	$27,$C3,$7A,$8F,$5F,$C3,$59,$F8,$3F,$0D,$EE,$3D,$FF,$0D,$63,$E2
	.dc.b	$FC,$17,$C2,$F8,$2F,$86,$D8,$F9,$3F,$05,$F1,$8F,$8F,$F0,$96,$3A
	.dc.b	$93,$8F,$E0,$E1,$D5,$F8,$77,$CD,$B3,$F5,$30,$FA,$3F,$0B,$23,$EB
	.dc.b	$FC,$35,$AF,$97,$F0,$DF,$5B,$E0,$FE,$BB,$6B,$EE,$FC,$37,$DA,$F8
	.dc.b	$3F,$B6,$D8,$FF,$67,$7E,$87,$FE,$E6,$3F,$B4,$3F,$ED,$C7,$1F,$0D
	.dc.b	$D2,$3F,$0F,$85,$C7,$ED,$BD,$23,$F1,$78,$59,$FB,$EF,$F0,$9B,$FF
	.dc.b	$02,$13,$E1,$B1,$F0,$DC,$FA,$7F,$1B,$8D,$C7,$ED,$3C,$E3,$F2,$38
	.dc.b	$59,$FB,$CF,$5F,$86,$ED,$F0,$B6,$7D,$AF,$87,$E1,$BC,$5F,$0D,$67
	.dc.b	$E0,$BE,$5F,$86,$F3,$7C,$35,$9F,$8A,$FA,$7E,$17,$D5,$F0,$D6,$7E
	.dc.b	$4B,$CF,$87,$FC,$15,$C2,$C7,$1B,$7F,$3F,$8C,$5F,$D0,$CB,$BB,$4E
	.dc.b	$E7,$D4,$F8,$6F,$90,$7C,$9F,$84,$B1,$F6,$3E,$13,$E6,$1F,$37,$E1
	.dc.b	$AC,$7D,$CF,$86,$FA,$1F,$0B,$F4,$5B,$1F,$80,$F8,$6F,$A9,$F0,$7F
	.dc.b	$55,$B1,$D4,$9B,$FB,$30,$EA,$FB,$AF,$9B,$67,$F1,$78,$7D,$FF,$86
	.dc.b	$FE,$91,$E0,$7F,$51,$9F,$C6,$46,$3E,$1B,$3F,$0B,$25,$87,$F8,$CE
	.dc.b	$FF,$B2,$1F,$D5,$DC,$7C,$1F,$43,$AF,$C2,$DA,$E3,$C9,$FD,$B6,$FF
	.dc.b	$AA,$C7,$4F,$85,$43,$E0,$7C,$BF,$05,$E6,$7C,$59,$78,$73,$F4,$FC
	.dc.b	$36,$32,$E9,$A7,$7A,$F3,$69,$67,$C1,$F0,$7F,$03,$E0,$FE,$0B,$76
	.dc.b	$F2,$9E,$3D,$7D,$95,$E7,$CF,$BC,$EB,$16,$DE,$89,$ED,$3C,$73,$F8
	.dc.b	$2E,$E7,$DB,$1F,$43,$E1,$7E,$8F,$C2,$DA,$D6,$48,$FA,$BF,$0B,$F5
	.dc.b	$BE,$16,$D6,$9F,$B3,$EF,$FD,$AF,$83,$B3,$C9,$FB,$BF,$01,$23,$AF
	.dc.b	$C1,$DD,$FE,$D6,$3D,$67,$C2,$D8,$FB,$D8,$75,$9F,$2B,$63,$F0,$5F
	.dc.b	$0B,$F8,$A3,$F1,$7F,$0B,$6B,$D7,$F0,$BF,$8E,$F8,$5F,$C7,$B6,$3E
	.dc.b	$27,$E8,$9F,$EE,$C2,$7C,$37,$F8,$6D,$C7,$C2,$53,$1F,$09,$6B,$A9
	.dc.b	$F2,$30,$EA,$D3,$AE,$7C,$CE,$98,$58,$F0,$FC,$27,$7F,$84,$F0,$36
	.dc.b	$3C,$BF,$0B,$E3,$3C,$9F,$0D,$63,$D3,$F0,$BE,$7F,$85,$F4,$36,$3D
	.dc.b	$BF,$0B,$EB,$3D,$9F,$07,$63,$A9,$34,$FF,$4F,$70,$7F,$BA,$23,$1E
	.dc.b	$07,$F9,$4E,$3F,$B3,$C7,$C8,$C7,$BB,$4F,$97,$F0,$7F,$27,$B7,$E5
	.dc.b	$36,$39,$9F,$AF,$AF,$EF,$CF,$D2,$D3,$D9,$8F,$8B,$4F,$53,$67,$B3
	.dc.b	$E6,$F8,$5F,$AD,$F0,$BF,$5D,$B3,$F7,$3E,$17,$ED,$1F,$6F,$E1,$6C
	.dc.b	$7E,$03,$E1,$7E,$F7,$C0,$FD,$F6,$C7,$E1,$3E,$17,$F0,$47,$E0,$FE
	.dc.b	$0A,$C7,$52,$67,$F2,$30,$EA,$FC,$AB,$E6,$D9,$FC,$7E,$1F,$E6,$7F
	.dc.b	$D1,$7F,$D0,$71,$FB,$57,$FB,$6C,$FE,$F1,$CB,$FE,$D1,$73,$E4,$9F
	.dc.b	$F7,$50,$F8,$5E,$9F,$0B,$D5,$ED,$F0,$BD,$FE,$16,$E3,$DD,$F1,$7C
	.dc.b	$2F,$8F,$E0,$AC,$F8,$DE,$6F,$96,$EE,$7E,$37,$9B,$E5,$BB,$D5,$FE
	.dc.b	$99,$CD,$1E,$FF,$81,$98,$7B,$FE,$16,$E1,$ED,$F8,$5F,$88,$7C,$5F
	.dc.b	$85,$B5,$F1,$7E,$13,$E4,$1F,$27,$E1,$2C,$7C,$9F,$84,$F9,$87,$CD
	.dc.b	$F8,$2B,$1F,$37,$0E,$78,$FC,$BC,$6E,$DE,$7C,$B9,$CF,$CF,$CA,$CF
	.dc.b	$AF,$E0,$BE,$C7,$C1,$7D,$96,$E5,$D4,$97,$D5,$B5,$D5,$F5,$1F,$7E
	.dc.b	$FA,$BF,$D4,$78,$1F,$07,$F8,$53,$F0,$DF,$09,$67,$F0,$DF,$09,$FE
	.dc.b	$AF,$FD,$1B,$FB,$46,$5F,$87,$8C,$7C,$2E,$7D,$F9,$7F,$D0,$6F,$FD
	.dc.b	$94,$7F,$4C,$7F,$B0,$3F,$EB,$B6,$FB,$FD,$7E,$0B,$B3,$F9,$7C,$3A
	.dc.b	$CF,$88,$E9,$87,$6C,$5B,$7F,$B6,$71,$ED,$1F,$D5,$9F,$0E,$5E,$9F
	.dc.b	$85,$98,$79,$7E,$16,$DF,$28,$FE,$EF,$F6,$CF,$6F,$FD,$A5,$92,$31
	.dc.b	$3F,$D8,$E3,$1E,$DE,$A7,$37,$CC,$F2,$AF,$6F,$46,$FD,$9B,$87,$C7
	.dc.b	$F8,$4F,$9D,$F0,$BF,$3D,$B1,$F2,$FE,$17,$E9,$7C,$2F,$D3,$6C,$7C
	.dc.b	$FF,$81,$FA,$C7,$D7,$F8,$5B,$1F,$4F,$E0,$BE,$D1,$F6,$FE,$16,$C7
	.dc.b	$53,$EC,$61,$D5,$A7,$53,$EE,$7C,$2F,$4D,$FD,$DE,$73,$9D,$3F,$87
	.dc.b	$F8,$59,$7F,$83,$E5,$B8,$7C,$5E,$5F,$99,$D7,$F2,$4F,$95,$D9,$6F
	.dc.b	$E7,$73,$7F,$8A,$2F,$E5,$F4,$FE,$4F,$A3,$E8,$3F,$95,$E8,$FC,$B7
	.dc.b	$45,$C3,$F2,$4F,$DA,$F3,$3F,$3F,$1F,$65,$77,$67,$BF,$4F,$DB,$EE
	.dc.b	$95,$BE,$FB,$A7,$C0,$E8,$FF,$22,$CF,$F8,$7D,$7C,$1A,$31,$34,$E1
	.dc.b	$1E,$3F,$17,$C9,$E1,$DA,$F2,$3B,$E4,$E1,$D2,$C7,$FD,$C1,$5F,$E4
	.dc.b	$96,$E3,$FE,$34,$7F,$DC,$51,$E3,$F8,$5E,$4E,$FE,$0E,$D6,$FE,$12
	.dc.b	$E1,$D6,$6E,$09,$F9,$1E,$8C,$C3,$E4,$71,$5C,$7E,$17,$C2,$70,$F7
	.dc.b	$B8,$B6,$BE,$27,$BD,$D0,$EB,$F0,$56,$3E,$91,$F5,$7E,$23,$EB,$34
	.dc.b	$F9,$9F,$0B,$E4,$3C,$BF,$0B,$63,$E8,$7C,$27,$A3,$E1,$7D,$2D,$8F
	.dc.b	$DB,$7D,$BB,$E0,$FF,$76,$D8,$CF,$C0,$E2,$7E,$0F,$C2,$7C,$2F,$85
	.dc.b	$B5,$F7,$5F,$8B,$32,$3F,$01,$57,$87,$F3,$98,$34,$EE,$78,$34,$B7
	.dc.b	$F8,$7F,$83,$F9,$E7,$D0,$F8,$3B,$1F,$8B,$F8,$5F,$A6,$7D,$4F,$85
	.dc.b	$B1,$EC,$EA,$F5,$F4,$FB,$1B,$1F,$E7,$1F,$D1,$FF,$BD,$F7,$7E,$29
	.dc.b	$FE,$F6,$62,$3F,$D2,$4E,$3E,$1B,$F0,$6F,$F7,$D8,$E5,$18,$F8,$5F
	.dc.b	$E4,$7F,$73,$FC,$6E,$7F,$91,$2F,$27,$3E,$56,$9E,$7F,$0F,$D3,$E1
	.dc.b	$ED,$F9,$5E,$FE,$17,$83,$C2,$B3,$E9,$7B,$FA,$FE,$0F,$12,$CF,$81
	.dc.b	$F3,$FC,$3C,$8F,$81,$F0,$F7,$AF,$F8,$AF,$0B,$E1,$7D,$BF,$0B,$EE
	.dc.b	$6D,$FF,$3B,$F7,$BF,$57,$CA,$7F,$77,$19,$F2,$FF,$A1,$C6,$31,$EB
	.dc.b	$E6,$FC,$83,$E4,$F9,$B6,$FE,$97,$C3,$FC,$C3,$E6,$FC,$3D,$8F,$AD
	.dc.b	$F0,$DF,$40,$FA,$3F,$0F,$63,$ED,$7A,$DF,$53,$D6,$FA,$AD,$8F,$BD
	.dc.b	$EA,$FE,$EB,$F4,$8F,$EE,$BE,$AB,$91,$39,$F5,$F5,$F0,$F3,$0F,$F5
	.dc.b	$B6,$0F,$8B,$C7,$17,$F3,$C8,$FE,$EC,$FF,$B0,$17,$BB,$DF,$DE,$F0
	.dc.b	$3E,$1F,$72,$61,$D3,$E2,$2E,$1C,$FE,$1F,$CD,$F1,$1E,$76,$D7,$8B
	.dc.b	$A7,$D4,$7A,$FE,$1A,$C7,$9B,$E2,$3D,$DF,$11,$EF,$6C,$7A,$BC,$3F
	.dc.b	$85,$E2,$7C,$36,$C7,$BB,$E1,$FE,$31,$F1,$FE,$1E,$C7,$C2,$F8,$7F
	.dc.b	$94,$7C,$BF,$86,$B1,$F1,$BC,$EF,$9D,$E7,$7C,$F6,$C7,$FB,$7F,$F4
	.dc.b	$9F,$E8,$7C,$5F,$D9,$1F,$C3,$78,$FF,$CE,$9C,$B7,$E1,$FF,$98,$C7
	.dc.b	$F3,$AE,$F3,$FC,$EA,$B9,$79,$1A,$F8,$CD,$FA,$B3,$2F,$07,$AD,$6F
	.dc.b	$C1,$F0,$FD,$0E,$BF,$11,$6B,$CD,$F1,$1E,$03,$C3,$F1,$16,$3D,$5F
	.dc.b	$01,$E4,$F8,$0F,$2B,$63,$DD,$F0,$DE,$8F,$87,$F4,$B6,$3E,$17,$C3
	.dc.b	$B4,$A4,$3E,$27,$C3,$E5,$F8,$DF,$0F,$79,$BE,$37,$C3,$DC,$39,$73
	.dc.b	$7C,$48,$C5,$A3,$E3,$DE,$1F,$93,$F0,$D7,$9D,$F0,$D4,$C4,$67,$E1
	.dc.b	$A6,$1F,$65,$E1,$F0,$DC,$FD,$AF,$EF,$13,$FE,$1E,$5F,$11,$93,$5A
	.dc.b	$EF,$E1,$F8,$7C,$3C,$BF,$85,$C1,$FD,$EB,$05,$75,$F8,$6E,$DF,$0D
	.dc.b	$2F,$78,$FC,$0B,$57,$8F,$E1,$A5,$68,$F8,$52,$8F,$47,$C3,$7A,$7E
	.dc.b	$1A,$56,$F8,$D6,$9F,$77,$C3,$7B,$FE,$16,$5E,$EB,$F9,$43,$7E,$1B
	.dc.b	$E2,$7C,$34,$8D,$3F,$1E,$7E,$47,$C3,$7C,$9F,$86,$95,$CE,$3E,$67
	.dc.b	$C3,$7C,$DF,$86,$93,$DE,$3E,$87,$C2,$C8,$F1,$3F,$4F,$E1,$BE,$A7
	.dc.b	$C2,$FD,$FF,$D2,$9F,$0D,$4C,$DF,$BE,$1F,$0D,$B2,$46,$7E,$1B,$97
	.dc.b	$C2,$F3,$F8,$59,$3C,$63,$B7,$C2,$C8,$E8,$F8,$7E,$17,$C5,$F0,$9E
	.dc.b	$3F,$85,$95,$E1,$8F,$37,$C2,$FE,$63,$F4,$77,$13,$D3,$C3,$6F,$D1
	.dc.b	$5E,$DF,$86,$91,$EB,$7E,$07,$C2,$FC,$1F,$86,$93,$EF,$7E,$27,$C2
	.dc.b	$FE,$9B,$C8,$78,$1F,$85,$AF,$F4,$C3,$5F,$F3,$93,$63,$FC,$A0,$FF
	.dc.b	$BA,$A3,$E5,$7C,$37,$CF,$F8,$69,$18,$3E,$77,$C3,$7D,$32,$3F,$47
	.dc.b	$E1,$BC,$5F,$80,$FE,$CA,$3F,$D5,$0D,$7F,$D0,$4D,$C7,$FC,$C0,$BD
	.dc.b	$2F,$BD,$F0,$FF,$7F,$E1,$A5,$FD,$D8,$FC,$17,$C3,$C8,$FC,$03,$FC
	.dc.b	$0F,$F4,$BF,$E1,$D1,$DB,$DD,$F1,$FB,$D2,$3C,$71,$BF,$88,$91,$FA
	.dc.b	$7A,$FF,$3D,$B7,$E2,$3A,$7C,$44,$8D,$3D,$FE,$23,$DF,$F0,$FF,$03
	.dc.b	$E1,$E5,$7C,$08,$F2,$7C,$44,$8E,$CF,$9E,$3D,$13,$27,$C4,$FA,$FE
	.dc.b	$23,$E5,$C4,$3E,$59,$F3,$B3,$F8,$1D,$F2,$3D,$4F,$C3,$F8,$7F,$89
	.dc.b	$F1,$1F,$17,$79,$5E,$F8,$FC,$46,$3F,$EB,$A8,$FF,$6F,$35,$FF,$5C
	.dc.b	$37,$9B,$FD,$2C,$BA,$3E,$87,$14,$AF,$94,$FD,$3F,$88,$FA,$9F,$11
	.dc.b	$F5,$7E,$1E,$57,$CF,$8F,$0F,$87,$E2,$F0,$E4,$F8,$A3,$EE,$7C,$47
	.dc.b	$DD,$EC,$93,$F5,$E3,$F0,$1F,$0F,$F8,$1F,$87,$93,$F6,$E3,$DB,$E6
	.dc.b	$C8,$F6,$BF,$03,$CE,$FC,$47,$C3,$FF,$4B,$FD,$33,$FA,$0F,$1F,$FA
	.dc.b	$C9,$BF,$09,$FD,$53,$C0,$56,$FE,$22,$65,$FC,$BE,$FF,$A3,$B7,$E1
	.dc.b	$FF,$69,$F8,$23,$DB,$BE,$DC,$AF,$E7,$9A,$2B,$51,$CF,$E2,$7C,$9F
	.dc.b	$11,$23,$07,$7F,$89,$F4,$7C,$4C,$8C,$1D,$2D,$F0,$9E,$2F,$84,$93
	.dc.b	$98,$F8,$1F,$13,$23,$D4,$FA,$3E,$17,$E2,$7C,$44,$9F,$73,$F1,$FD
	.dc.b	$6F,$6E,$52,$7C,$0F,$C0,$BF,$83,$5F,$37,$E2,$25,$79,$63,$E8,$7C
	.dc.b	$44,$8F,$94,$FD,$3C,$BE,$A6,$72,$7E,$73,$F5,$FE,$23,$EC,$69,$F6
	.dc.b	$7D,$D9,$5F,$4E,$3E,$E7,$C4,$48,$FA,$CF,$DF,$E5,$E9,$0E,$4F,$C0
	.dc.b	$63,$87,$F0,$58,$F6,$9E,$84,$38,$75,$FF,$86,$3A,$1C,$FF,$8B,$F8
	.dc.b	$89,$5F,$85,$7F,$C3,$1E,$3F,$F5,$A8,$F5,$90,$E6,$7F,$87,$EB,$FB
	.dc.b	$CB,$F8,$DC,$3C,$4E,$BE,$25,$2D,$B3,$9F,$E6,$72,$EF,$F0,$FE,$0F
	.dc.b	$87,$98,$FF,$81,$73,$FE,$CE,$FE,$4B,$F0,$57,$A1,$FD,$2D,$B1,$98
	.dc.b	$C7,$9B,$9A,$93,$C5,$E1,$E6,$7B,$BC,$6F,$7F,$8D,$2B,$C1,$1F,$0B
	.dc.b	$E2,$24,$78,$DF,$8B,$F1,$1F,$1B,$E1,$E4,$F9,$DF,$93,$F1,$1F,$2B
	.dc.b	$CF,$93,$EB,$7E,$6F,$A1,$D2,$3E,$36,$8C,$71,$C6,$E9,$FF,$27,$CB
	.dc.b	$3E,$F7,$C3,$3B,$7B,$8F,$F2,$95,$33,$FB,$7F,$0F,$27,$E6,$3F,$7B
	.dc.b	$E1,$FE,$FF,$C3,$C9,$FA,$0F,$E0,$BE,$1F,$F0,$7F,$0F,$F8,$4F,$5E
	.dc.b	$57,$D5,$8F,$C3,$FC,$3F,$E2,$3E,$16,$4F,$D9,$8E,$8F,$BB,$0B,$AF
	.dc.b	$9F,$BD,$B3,$3E,$F7,$F8,$1C,$FB,$BF,$1B,$3F,$CA,$FC,$37,$E5,$BE
	.dc.b	$1A,$4F,$E1,$E3,$FC,$E1,$FA,$6F,$F6,$EC,$FB,$37,$93,$FC,$F2,$DA
	.dc.b	$FF,$3C,$3F,$DB,$1F,$CC,$2F,$87,$E6,$48,$FF,$2C,$43,$E1,$FB,$FC
	.dc.b	$3C,$9C,$BE,$2F,$87,$F1,$C4,$38,$9E,$6F,$85,$E8,$F8,$30,$B9,$7A
	.dc.b	$30,$F5,$FC,$3C,$AF,$0B,$EE,$F8,$3E,$91,$E8,$2D,$7C,$0C,$39,$E9
	.dc.b	$F0,$31,$B5,$F0,$31,$F8,$FF,$0F,$F2,$3E,$1E,$4F,$BA,$3E,$5F,$C3
	.dc.b	$C8,$F8,$2F,$CE,$F8,$7F,$9F,$E3,$FD,$0F,$1E,$57,$C6,$8F,$A7,$D1
	.dc.b	$F5,$3C,$79,$3F,$2A,$3A,$3F,$43,$0B,$BF,$D8,$C7,$AC,$FD,$0C,$7A
	.dc.b	$BF,$5B,$1F,$BF,$F0,$FF,$80,$F8,$69,$3F,$5A,$3F,$07,$F0,$FF,$84
	.dc.b	$F8,$69,$3F,$6A,$3F,$0F,$F0,$F2,$3E,$EB,$F8,$AF,$86,$FF,$5D,$7E
	.dc.b	$9D,$FD,$BA,$7F,$D7,$CD,$F0,$4D,$3F,$B6,$C7,$E2,$AF,$BB,$F8,$CC
	.dc.b	$3C,$6F,$62,$1F,$90,$C3,$C0,$FF,$A9,$6B,$FD,$6A,$DF,$86,$F1,$9E
	.dc.b	$4F,$86,$F2,$FC,$37,$9B,$E1,$A5,$71,$8F,$4F,$C2,$7A,$BE,$12,$4F
	.dc.b	$58,$F6,$FC,$1C,$8F,$03,$F0,$3E,$17,$AC,$78,$F0,$EA,$FF,$3D,$CF
	.dc.b	$FD,$86,$F3,$3D,$18,$76,$75,$87,$CA,$F8,$59,$1E,$D7,$E6,$FC,$2F
	.dc.b	$68,$E9,$87,$6E,$8F,$E0,$D8,$F6,$9E,$67,$62,$18,$3C,$9F,$0B,$F6
	.dc.b	$08,$FC,$BF,$82,$FB,$7F,$03,$F7,$3E,$02,$4F,$CF,$8F,$BF,$F0,$B2
	.dc.b	$3E,$93,$F8,$2F,$85,$FC,$1F,$C2,$FE,$13,$E0,$E5,$7D,$78,$EA,$FD
	.dc.b	$9C,$3A,$BF,$1F,$0E,$4F,$42,$1F,$2E,$FF,$1F,$F0,$A8,$FF,$73,$7E
	.dc.b	$9F,$FD,$CD,$FE,$D8,$DD,$BE,$2C,$EE,$73,$F0,$B6,$7F,$97,$53,$E1
	.dc.b	$71,$E6,$75,$EF,$4E,$70,$EF,$86,$94,$EB,$E0,$F0,$D8,$F2,$C6,$29
	.dc.b	$F0,$97,$E0,$A5,$7A,$9D,$7C,$16,$F1,$E1,$9C,$AF,$09,$D0,$87,$C0
	.dc.b	$C7,$B6,$1D,$FE,$0A,$4F,$C6,$7C,$5F,$05,$E3,$F8,$49,$3F,$29,$F3
	.dc.b	$7C,$27,$9F,$E1,$3D,$1F,$09,$2B,$E7,$C7,$AF,$E0,$BD,$9F,$05,$27
	.dc.b	$E9,$C7,$47,$E9,$E1,$79,$3E,$7E,$3D,$67,$EC,$E3,$D5,$FB,$38,$FC
	.dc.b	$7F,$81,$91,$F7,$9F,$95,$F0,$5C,$F9,$BE,$B7,$2B,$3C,$5F,$D4,$86
	.dc.b	$5D,$E3,$F0,$B9,$77,$CF,$F0,$39,$FD,$5F,$83,$95,$F8,$C7,$EC,$7C
	.dc.b	$27,$D9,$F8,$39,$3F,$90,$7E,$E7,$C2,$7E,$03,$87,$FB,$47,$57,$F7
	.dc.b	$7F,$7F,$06,$6F,$5E,$FE,$FD,$F9,$1F,$9B,$D7,$FC,$01,$87,$F6,$D8
	.dc.b	$EC,$43,$F3,$B9,$78,$1E,$1C,$5F,$97,$9F,$09,$FD,$DE,$B9,$E3,$E1
	.dc.b	$AD,$9F,$98,$97,$C1,$FA,$7E,$0E,$4E,$B4,$FF,$08,$4F,$B7,$E0,$FD
	.dc.b	$DF,$07,$2B,$58,$27,$E1,$7C,$17,$C3,$F8,$39,$5A,$8F,$8D,$F0,$32
	.dc.b	$38,$BF,$27,$E0,$7B,$47,$4C,$7B,$3F,$E6,$7C,$7B,$3C,$CF,$0E,$1D
	.dc.b	$9D,$61,$F4,$FE,$03,$EA,$7C,$0C,$9F,$3C,$7D,$7F,$82,$FB,$1F,$07
	.dc.b	$27,$D7,$1F,$6F,$E0,$A4,$7B,$9F,$BD,$F0,$7F,$7F,$E0,$BF,$01,$F0
	.dc.b	$52,$BE,$1C,$75,$7E,$2E,$1D,$5F,$56,$1D,$1F,$66,$17,$9B,$F0,$38
	.dc.b	$FE,$2F,$E0,$A5,$7C,$E7,$F1,$DF,$07,$D6,$3E,$26,$3D,$6F,$F1,$38
	.dc.b	$F3,$E9,$FC,$26,$56,$FF,$01,$97,$F1,$6F,$D4,$3F,$BB,$12,$B1,$F1
	.dc.b	$39,$F8,$5D,$7C,$2C,$C3,$F0,$97,$FD,$C1,$FF,$4D,$3F,$D4,$73,$FF
	.dc.b	$2C,$37,$0F,$E5,$4D,$9F,$E4,$EB,$DE,$F1,$7C,$0F,$8F,$E0,$A5,$76
	.dc.b	$8F,$13,$E0,$C7,$C4,$F8,$32,$F5,$7C,$2C,$8F,$2B,$ED,$C3,$DD,$F0
	.dc.b	$BE,$FF,$85,$95,$EA,$8F,$85,$F0,$BF,$0F,$E1,$64,$FB,$A3,$E3,$7C
	.dc.b	$27,$C7,$F8,$49,$3F,$0A,$3E,$57,$C0,$7C,$BF,$81,$93,$F1,$A3,$AB
	.dc.b	$F2,$30,$EA,$FC,$8C,$3A,$BF,$33,$0E,$AF,$CC,$C3,$EB,$7C,$24,$8F
	.dc.b	$A2,$FD,$9F,$82,$FB,$5F,$07,$F6,$FE,$12,$57,$D6,$8F,$BD,$F0,$92
	.dc.b	$3E,$CB,$FD,$5B,$F5,$1F,$EC,$DD,$1F,$D7,$DF,$F6,$5C,$FF,$B5,$5B
	.dc.b	$D3,$FE,$93,$6C,$EF,$E0,$BC,$35,$F8,$4C,$7C,$2F,$E1,$31,$F0,$3F
	.dc.b	$86,$C6,$E3,$CF,$E3,$FC,$5F,$07,$E3,$F8,$39,$7F,$C4,$74,$4F,$68
	.dc.b	$FC,$7E,$5D,$B0,$FC,$7E,$5D,$34,$FC,$7E,$76,$FF,$BA,$E3,$EE,$6F
	.dc.b	$C3,$7C,$12,$3D,$FE,$1B,$E2,$7C,$2C,$8F,$13,$F1,$FE,$17,$E4,$7C
	.dc.b	$14,$9F,$33,$F2,$FE,$07,$E6,$7C,$07,$CD,$C2,$57,$AE,$3A,$3F,$37
	.dc.b	$0B,$B7,$CD,$C7,$A4,$FD,$1C,$6C,$FD,$1C,$7E,$C7,$C0,$CA,$F8,$CF
	.dc.b	$DB,$F8,$5F,$B9,$EA,$7D,$DF,$85,$95,$F2,$E3,$F0,$1F,$0B,$23,$E7
	.dc.b	$3F,$83,$F8,$4E,$51,$F4,$6E,$BF,$83,$F8,$4E,$AF,$D5,$C3,$AB,$F8
	.dc.b	$7C,$3A,$3F,$89,$C2,$E7,$F8,$9C,$7F,$C5,$0B,$F1,$17,$B5,$86,$DF
	.dc.b	$F5,$69,$D4,$E7,$DD,$D7,$B5,$BF,$6E,$5E,$F0,$4F,$3F,$77,$A7,$BB
	.dc.b	$2B,$51,$FA,$1F,$82,$FD,$17,$C2,$C9,$FC,$84,$7F,$8E,$BF,$52,$FE
	.dc.b	$F6,$7F,$6B,$EE,$FF,$22,$36,$7F,$C8,$2E,$BE,$1B,$87,$C3,$49,$FE
	.dc.b	$D5,$3F,$DE,$DF,$D6,$5F,$85,$EE,$6B,$E0,$FC,$0F,$87,$E1,$A3,$C8
	.dc.b	$F2,$61,$D5,$71,$32,$F2,$E5,$A9,$E8,$76,$3C,$26,$0E,$31,$EC,$7B
	.dc.b	$98,$3D,$B7,$E4,$7D,$77,$E1,$BF,$04,$7C,$2B,$F0,$3C,$63,$EF,$7C
	.dc.b	$0F,$DD,$F8,$29,$3F,$5E,$3F,$01,$F0,$52,$3E,$D3,$F8,$3F,$82,$FC
	.dc.b	$3F,$E2,$5F,$EC,$DC,$0D,$27,$BF,$AF,$6F,$7E,$E4,$C7,$F1,$55,$FD
	.dc.b	$5E,$FC,$E7,$2F,$73,$AF,$B9,$2B,$71,$E0,$F7,$64,$72,$7C,$7E,$F7
	.dc.b	$93,$DD,$F2,$FB,$D2,$BB,$C7,$FA,$15,$7E,$60,$F7,$31,$F8,$0F,$FA
	.dc.b	$2E,$7E,$70,$71,$A7,$CC,$8E,$33,$F0,$88,$FB,$1F,$8B,$EF,$7C,$6F
	.dc.b	$76,$4F,$C0,$7E,$4F,$BD,$E3,$C7,$C5,$3F,$E4,$23,$C1,$EE,$79,$0D
	.dc.b	$61,$F4,$3D,$FF,$A3,$F0,$12,$BE,$5C,$7D,$4F,$80,$91,$F3,$9F,$AF
	.dc.b	$F0,$1C,$A3,$B6,$DE,$CF,$03,$FD,$64,$D8,$FF,$5D,$67,$E4,$7F,$11
	.dc.b	$CC,$F7,$3F,$03,$EE,$C9,$FB,$71,$F8,$4F,$77,$F8,$5F,$E2,$7F,$CF
	.dc.b	$79,$0D,$C1,$3E,$3B,$F9,$F9,$78,$5F,$82,$7C,$3C,$BC,$58,$7E,$1B
	.dc.b	$0F,$EB,$F3,$FA,$E7,$D1,$1D,$B8,$7C,$3F,$19,$E2,$F8,$0F,$1F,$C0
	.dc.b	$CC,$3F,$8D,$61,$FD,$69,$ED,$C9,$F3,$DF,$8B,$87,$97,$1F,$C0,$67
	.dc.b	$B9,$F3,$E3,$F8,$83,$5C,$1F,$E8,$7C,$5F,$01,$F0,$BE,$02,$5F,$82
	.dc.b	$36,$FF,$B9,$F0,$F8,$07,$C1,$DB,$DF,$87,$FB,$95,$B3,$BC,$7C,$0F
	.dc.b	$A0,$87,$E6,$21,$D7,$F8,$D3,$BC,$7E,$60,$F6,$90,$E1,$AE,$B0,$F7
	.dc.b	$69,$F0,$30,$E3,$87,$8A,$FF,$53,$87,$57,$C3,$5F,$A5,$9D,$5F,$B3
	.dc.b	$5E,$38,$FA,$2F,$E4,$9C,$32,$F7,$47,$38,$F5,$63,$C1,$F3,$63,$F4
	.dc.b	$AF,$11,$C8,$91,$E1,$AF,$3D,$73,$8C,$4F,$73,$E8,$57,$DF,$7E,$09
	.dc.b	$93,$F8,$06,$5E,$79,$E3,$1A,$3C,$A7,$E5,$6B,$C7,$1E,$58,$F3,$E1
	.dc.b	$EC,$7A,$63,$F3,$23,$F5,$F7,$C1,$FB,$D7,$E7,$78,$57,$7B,$E8,$FE
	.dc.b	$9B,$0E,$0F,$E0,$CF,$DD,$DE,$1F,$AA,$79,$4F,$66,$5F,$52,$7E,$05
	.dc.b	$7B,$09,$7E,$4A,$E6,$70,$CB,$C8,$FF,$04,$C7,$B6,$5F,$76,$34,$70
	.dc.b	$3E,$9D,$7A,$AF,$AC,$71,$3C,$87,$B6,$34,$7F,$12,$C3,$CA,$E2,$3F
	.dc.b	$89,$E1,$F4,$E3,$CD,$1C,$4C,$9D,$2B,$F2,$58,$79,$AF,$9C,$E8,$E0
	.dc.b	$7D,$E3,$D1,$3E,$67,$F4,$C7,$DE,$23,$9C,$7E,$54,$7D,$E3,$A4,$71
	.dc.b	$C7,$EF,$47,$D5,$3B,$CF,$EA,$4F,$DA,$BE,$7C,$3D,$CF,$82,$BE,$F9
	.dc.b	$23,$33,$FA,$CC,$BE,$15,$78,$27,$38,$FD,$68,$E5,$1D,$30,$FA,$D1
	.dc.b	$E2,$CB,$8B,$F0,$4C,$1E,$33,$EB,$18,$3F,$20,$7B,$CE,$A7,$84,$F7
	.dc.b	$63,$FC,$1D,$FC,$19,$F2,$2B,$F3,$AE,$23,$D8,$FE,$C8,$C4,$7D,$C8
	.dc.b	$E5,$1E,$17,$ED,$BE,$53,$C2,$4A,$F6,$C7,$D5,$AF,$6C,$7D,$29,$DE
	.dc.b	$3F,$47,$0F,$8C,$FD,$3A,$FC,$2C,$FF,$50,$3C,$18,$78,$C8,$FB,$CE
	.dc.b	$75,$CA,$7F,$94,$65,$CA,$FA,$57,$E3,$2F,$8B,$E2,$AF,$0D,$FE,$56
	.dc.b	$3E,$65,$70,$7F,$52,$7C,$CA,$FC,$0D,$70,$8F,$8E,$7C,$93,$EA,$9E
	.dc.b	$83,$81,$0F,$AF,$8F,$F7,$2C,$FF,$A9,$C7,$F9,$0F,$56,$BF,$DE,$27
	.dc.b	$D7,$3E,$4B,$FE,$89,$AF,$E2,$32,$FE,$A7,$8F,$F7,$C8,$FE,$D9,$5C
	.dc.b	$F6,$FE,$E0,$FF,$20,$D7,$E2,$61,$F5,$2B,$93,$FE,$07,$9F,$EE,$7F
	.dc.b	$05,$FE,$08,$30,$FE,$DB,$7F,$BB,$B7,$F0,$3C,$BF,$C1,$7C,$1F,$E0
	.dc.b	$EC,$A5,$FF,$7C,$8F,$F1,$0B,$FE,$18,$E6,$B5,$FE,$4E,$E3,$73,$7B
	.dc.b	$AB,$E4,$CF,$F1,$63,$FC,$B7,$CB,$FE,$1D,$8F,$F1,$0F,$37,$F9,$82
	.dc.b	$3F,$C7,$71,$FE,$24,$EA,$B5,$D7,$AB,$A3,$F9,$E9,$FA,$38,$BA,$7F
	.dc.b	$C7,$11,$FE,$2C,$EA,$FF,$1E,$3F,$E5,$63,$FC,$CB,$D3,$FE,$59,$7F
	.dc.b	$C6,$3D,$9D,$38,$7F,$93,$47,$F5,$BC,$7F,$CE,$3C,$9F,$E5,$28,$FF
	.dc.b	$1B,$BF,$E5,$27,$FC,$DF,$E1,$7F,$9C,$5F,$F1,$DF,$85,$FE,$75,$7F
	.dc.b	$CD,$A7,$F9,$D3,$C2,$FF,$38,$3F,$E4,$0F,$0B,$FC,$DE,$FF,$90,$BC
	.dc.b	$2B,$8F,$CB,$F0,$BF,$D8,$DE,$37,$FA,$2A,$BF,$D1,$DE,$93,$C8,$CC
	.dc.b	$7F,$A2,$88,$6B,$C8,$FF,$46,$3F,$E8,$4D,$FF,$D1,$A4,$7F,$D0,$FE
	.dc.b	$3F,$FB,$07,$E2,$FF,$D8,$5E,$AC,$9F,$F5,$DC,$7F,$B1,$7D,$69,$1F
	.dc.b	$EB,$E7,$FD,$91,$EA,$FF,$B2,$7D,$69,$3F,$EC,$27,$FD,$99,$EA,$F5
	.dc.b	$8F,$F6,$2E,$1D,$5F,$F5,$87,$C5,$BD,$56,$5D,$A7,$31,$FE,$E6,$6E
	.dc.b	$5E,$07,$B1,$0D,$E3,$E0,$7F,$DC,$D1,$FE,$EF,$23,$FE,$E9,$C3,$FD
	.dc.b	$DA,$FF,$BA,$E7,$FD,$DA,$FC,$EC,$BF,$DD,$8F,$FB,$BF,$D7,$FF,$76
	.dc.b	$3F,$4B,$3F,$F7,$44,$48,$FF,$70,$BF,$EE,$BF,$8A,$FF,$76,$7C,$5F
	.dc.b	$FB,$B7,$E2,$A5,$7F,$B9,$A3,$FD,$E1,$F1,$7F,$EF,$1F,$8B,$93,$FE
	.dc.b	$EA,$8F,$F7,$A7,$C5,$C8,$FF,$76,$BD,$DF,$8B,$5D,$1F,$8F,$7E,$07
	.dc.b	$91,$FE,$F4,$BE,$AF,$CC,$BF,$F8,$01,$CF,$FF,$00,$79,$A4,$FF,$BE
	.dc.b	$A0,$FF,$BF,$39,$AB,$F5,$3D,$7B,$D5,$FE,$E4,$E6,$FB,$51,$8E,$8E
	.dc.b	$31,$F7,$6F,$16,$8F,$DC,$4E,$1B,$1E,$0B,$E9,$E2,$7F,$B1,$5B,$D3
	.dc.b	$B8,$FF,$6F,$5F,$66,$D7,$56,$DF,$E3,$73,$ED,$D0,$CA,$E7,$30,$F7
	.dc.b	$1E,$1C,$7A,$1D,$C8,$E4,$ED,$97,$8F,$BE,$46,$72,$FE,$29,$1E,$7E
	.dc.b	$EF,$0E,$3D,$EB,$E5,$1E,$03,$F4,$7A,$7B,$F1,$F0,$C7,$38,$C3,$E0
	.dc.b	$31,$DF,$F0,$C8,$EF,$BF,$E3,$77,$FC,$7F,$0A,$4F,$38,$F9,$5E,$1C
	.dc.b	$8E,$CF,$CD,$F0,$FE,$77,$AF,$F3,$FC,$89,$5E,$38,$EA,$F9,$70,$EA
	.dc.b	$FA,$B0,$E8,$FB,$30,$BC,$BF,$73,$1F,$B5,$E4,$4A,$F7,$3F,$77,$5F
	.dc.b	$BD,$E3,$FD,$FF,$32,$57,$C3,$8F,$C1,$79,$92,$3E,$33,$F8,$5F,$2F
	.dc.b	$F0,$DE,$64,$9F,$94,$FE,$27,$D7,$EB,$1F,$37,$0E,$AF,$D2,$C3,$A3
	.dc.b	$F5,$30,$BA,$7E,$53,$1F,$CC,$7E,$A6,$F4,$A9,$2F,$F0,$51,$FE,$EC
	.dc.b	$EA,$2F,$47,$8F,$A7,$2B,$71,$D3,$D4,$91,$C9,$EF,$EB,$78,$3D,$7F
	.dc.b	$0F,$C4,$CA,$EF,$1E,$4B,$E9,$1F,$CB,$B1,$FF,$34,$E5,$E2,$9E,$B3
	.dc.b	$9C,$6E,$BF,$BB,$BF,$F7,$7C,$FA,$BE,$27,$DF,$F1,$32,$30,$7A,$0F
	.dc.b	$7E,$BE,$0B,$E5,$8F,$4E,$FF,$E0,$78,$B2,$C6,$1F,$C0,$E3,$BB,$E7
	.dc.b	$F8,$9F,$9A,$47,$E5,$FC,$57,$D0,$F8,$A9,$1F,$39,$FA,$7F,$15,$F5
	.dc.b	$3E,$23,$EA,$FC,$4C,$AF,$A7,$1F,$63,$E2,$7E,$CF,$C4,$C9,$FA,$F1
	.dc.b	$D1,$F9,$18,$5D,$FE,$26,$3D,$67,$EE,$E3,$D5,$F9,$D8,$FE,$13,$E2
	.dc.b	$BF,$0B,$F1,$32,$7F,$07,$1F,$88,$F8,$A9,$1F,$86,$7F,$91,$7E,$A9
	.dc.b	$FC,$BA,$3F,$B2,$AF,$8A,$C9,$23,$F8,$74,$3E,$2B,$8F,$C5,$49,$CB
	.dc.b	$D3,$E2,$FB,$C7,$E0,$F0,$BB,$7F,$39,$C3,$C0,$DC,$7B,$57,$89,$F2
	.dc.b	$E3,$E6,$F8,$4F,$3F,$C1,$CA,$F1,$C7,$AB,$E1,$3D,$7F,$07,$27,$CF
	.dc.b	$1E,$EF,$84,$91,$EA,$7E,$0F,$C2,$7C,$2F,$84,$93,$EE,$7E,$2F,$C2
	.dc.b	$75,$8F,$83,$87,$57,$E3,$E1,$D1,$F9,$38,$5D,$7D,$D8,$FC,$EF,$83
	.dc.b	$F9,$FF,$07,$2F,$E5,$C7,$57,$E8,$E3,$D6,$FE,$36,$3C,$FC,$0F,$EB
	.dc.b	$BC,$2C,$F1,$87,$F5,$D8,$67,$F7,$3E,$0A,$4F,$D8,$7E,$FF,$C1,$FE
	.dc.b	$03,$E1,$3F,$03,$F0,$92,$BE,$EC,$7E,$13,$E1,$3F,$0B,$EF,$C9,$FC
	.dc.b	$0C,$7E,$23,$E0,$E4,$7E,$11,$FE,$D9,$FA,$AF,$F4,$8F,$35,$AB,$1E
	.dc.b	$D7,$F8,$EC,$BC,$4F,$43,$F1,$B9,$77,$7F,$27,$97,$4F,$84,$91,$FD
	.dc.b	$7F,$24,$78,$3E,$1F,$C3,$F0,$BE,$2F,$85,$97,$B8,$F2,$FC,$2F,$9B
	.dc.b	$E1,$64,$F3,$8F,$4F,$C1,$48,$EC,$FB,$3E,$0F,$AC,$78,$70,$EA,$FF
	.dc.b	$93,$33,$63,$DA,$3C,$D8,$F6,$75,$8F,$C6,$F8,$3F,$8F,$F0,$72,$7D
	.dc.b	$91,$F2,$BE,$13,$E5,$FC,$2C,$9F,$81,$1F,$3B,$E1,$64,$7C,$37,$E8
	.dc.b	$FC,$2F,$D2,$F8,$2F,$A7,$F0,$72,$BE,$44,$75,$7E,$56,$1D,$5F,$56
	.dc.b	$1D,$1F,$66,$17,$C1,$FF,$18,$6F,$FE,$A1,$9F,$A5,$F0,$7F,$80,$23
	.dc.b	$F5,$7E,$0F,$B3,$F1,$71,$ED,$7C,$B1,$E9,$E1,$7F,$8E,$31,$65,$8C
	.dc.b	$3F,$C7,$11,$DD,$F9,$DF,$0B,$F8,$EF,$85,$91,$F8,$38,$FF,$1E,$78
	.dc.b	$5F,$D9,$63,$FC,$48,$D7,$FD,$02,$DC,$BF,$C9,$4F,$F6,$73,$FB,$34
	.dc.b	$70,$F7,$F9,$9D,$3D,$FE,$BE,$FC,$9E,$AF,$83,$E1,$3C,$3F,$0B,$27
	.dc.b	$C2,$F9,$3E,$27,$94,$75,$21,$E4,$BF,$4F,$C2,$FA,$BE,$1A,$4F,$AA
	.dc.b	$3D,$BF,$0D,$23,$DA,$FC,$0F,$86,$F8,$3F,$0B,$F0,$BE,$16,$57,$C2
	.dc.b	$8F,$8B,$F0,$72,$3E,$2B,$F2,$3E,$0F,$A4,$7C,$6C,$2F,$17,$C6,$C7
	.dc.b	$AC,$FC,$CC,$7A,$BF,$33,$1F,$A3,$F0,$5F,$4B,$E0,$E4,$FD,$28,$FA
	.dc.b	$BF,$09,$F5,$BE,$16,$4F,$D6,$8F,$B3,$F0,$92,$3E,$CB,$F7,$3E,$17
	.dc.b	$EE,$FC,$1C,$9F,$BA,$FF,$9E,$FF,$56,$FE,$97,$DF,$FE,$82,$EB,$63
	.dc.b	$DA,$FF,$03,$97,$89,$EA,$43,$F0,$B9,$77,$7F,$CD,$98,$7F,$63,$6F
	.dc.b	$C1,$F7,$3C,$1F,$09,$CE,$3A,$9F,$90,$BC,$71,$7E,$3B,$0F,$64,$F6
	.dc.b	$3F,$2B,$AF,$A7,$E1,$BD,$5F,$0D,$27,$C3,$1E,$DF,$85,$91,$E4,$7E
	.dc.b	$07,$C3,$7C,$1F,$81,$F8,$5F,$05,$2B,$D3,$1F,$17,$E0,$BE,$37,$C1
	.dc.b	$49,$F6,$C7,$47,$FD,$CB,$C3,$FE,$E4,$7B,$1F,$0B,$0E,$CF,$C4,$C3
	.dc.b	$B3,$C3,$0F,$A5,$F0,$DF,$4F,$E1,$A4,$FC,$A8,$FA,$DF,$0B,$F5,$FE
	.dc.b	$1A,$4F,$CE,$8F,$B5,$CF,$F6,$FE,$1A,$4F,$D2,$8F,$BD,$F0,$72,$3E
	.dc.b	$AB,$F8,$1F,$83,$EB,$1F,$63,$0E,$AF,$B3,$0E,$AF,$DC,$C3,$AB,$F0
	.dc.b	$30,$FC,$57,$C1,$48,$FC,$0B,$F8,$DF,$83,$FC,$77,$C3,$49,$FC,$2B
	.dc.b	$FE,$CB,$FD,$5F,$FB,$01,$FD,$52,$3F,$D6,$8D,$5D,$5F,$ED,$98,$C3
	.dc.b	$7E,$13,$97,$C0,$CA,$E2,$F5,$F8,$2F,$04,$7D,$6C,$2D,$F3,$F0,$BC
	.dc.b	$5F,$0B,$2B,$C2,$F9,$71,$F3,$7C,$17,$9F,$E0,$E5,$79,$A3,$9D,$FF
	.dc.b	$BA,$BA,$99,$62,$BF,$DD,$5A,$FB,$B2,$ED,$1E,$CB,$FF,$78,$CF,$B1
	.dc.b	$B9,$7C,$5F,$85,$F8,$DF,$0B,$2F,$E2,$C7,$FB,$DB,$57,$CA,$3A,$B0
	.dc.b	$DA,$FF,$71,$4E,$A7,$3C,$DA,$E7,$95,$AB,$47,$2E,$6F,$AD,$F0,$B2
	.dc.b	$BE,$AB,$F6,$7E,$1B,$ED,$7C,$07,$DB,$F8,$19,$5F,$6A,$3E,$F7,$C0
	.dc.b	$FD,$FF,$86,$93,$F7,$A3,$F0,$5F,$0B,$F8,$3F,$86,$93,$F8,$28,$EA
	.dc.b	$FE,$13,$0E,$AF,$C9,$3E,$6E,$1D,$9F,$C4,$E1,$D9,$D7,$C4,$39,$BF
	.dc.b	$B3,$C7,$E3,$BE,$1F,$F2,$44,$7F,$23,$F0,$FF,$37,$E5,$9D,$D4,$93
	.dc.b	$F3,$21,$CF,$BE,$84,$CE,$87,$56,$34,$7C,$B3,$93,$77,$E6,$8C,$3D
	.dc.b	$70,$FD,$37,$C3,$FE,$9F,$E2,$24,$FE,$9A,$3C,$DC,$7E,$7E,$69,$3E
	.dc.b	$58,$F5,$72,$C8,$F4,$3E,$DE,$4F,$0E,$DE,$69,$6A,$FD,$63,$FC,$01
	.dc.b	$BF,$A5,$FF,$73,$13,$E1,$BE,$37,$27,$C7,$D6,$61,$F1,$63,$AD,$F1
	.dc.b	$F8,$5B,$3D,$B0,$FF,$6F,$92,$33,$8F,$87,$E2,$2C,$FF,$B4,$EB,$E6
	.dc.b	$5F,$8E,$73,$F1,$1F,$A3,$F3,$DF,$13,$49,$7F,$A2,$8E,$4F,$E1,$30
	.dc.b	$7C,$3F,$62,$47,$63,$CD,$F1,$1E,$2F,$86,$F1,$FC,$34,$9D,$6B,$F1
	.dc.b	$E7,$C5,$3E,$3F,$85,$95,$E1,$8F,$5C,$7B,$3E,$1A,$4F,$2C,$93,$F0
	.dc.b	$3E,$23,$E0,$FC,$44,$AD,$47,$BF,$E2,$24,$7B,$5F,$85,$F0,$DF,$23
	.dc.b	$E2,$24,$F5,$7E,$5C,$FC,$C9,$93,$E1,$7E,$7F,$C2,$FF,$0D,$F9,$9F
	.dc.b	$F0,$2C,$FF,$32,$6F,$59,$A7,$F7,$F8,$F9,$9F,$0D,$F6,$08,$FB,$7E
	.dc.b	$1B,$ED,$FC,$4F,$DE,$CB,$AF,$19,$BC,$09,$F1,$7C,$17,$8F,$9E,$63
	.dc.b	$FC,$73,$34,$F3,$F8,$8E,$9F,$13,$2B,$EA,$E5,$E5,$DB,$F3,$DA,$77
	.dc.b	$F8,$8F,$17,$C4,$CB,$DC,$7C,$1F,$63,$E1,$7B,$12,$7C,$11,$E9,$F8
	.dc.b	$99,$1D,$9F,$66,$BE,$DD,$FD,$DF,$0F,$2B,$C7,$1F,$37,$E1,$7E,$77
	.dc.b	$C2,$49,$F6,$47,$C5,$F8,$8F,$8D,$F1,$32,$7D,$71,$F2,$7E,$26,$47
	.dc.b	$B9,$F9,$9A,$7F,$45,$F7,$1F,$09,$49,$5F,$D1,$3F,$15,$FF,$58,$BF
	.dc.b	$1B,$83,$EA,$11,$F9,$3F,$01,$F5,$FE,$02,$47,$CC,$7E,$D7,$C4,$FD
	.dc.b	$BF,$8A,$FB,$9B,$CA,$FA,$31,$F7,$F8,$3F,$01,$F1,$52,$7E,$AC,$7E
	.dc.b	$0F,$E1,$FF,$80,$FE,$B3,$FE,$E3,$3F,$DC,$9D,$9F,$DB,$2B,$F5,$F1
	.dc.b	$FD,$9D,$FE,$2E,$DE,$B3,$47,$C3,$F4,$F8,$BE,$BF,$17,$2F,$9C,$78
	.dc.b	$38,$E4,$76,$7C,$7F,$09,$E4,$F8,$3F,$2F,$C0,$CA,$F1,$C7,$A3,$9E
	.dc.b	$47,$99,$F5,$FC,$1F,$B3,$E1,$FD,$BF,$03,$2B,$D7,$1F,$03,$E0,$A4
	.dc.b	$7B,$9F,$87,$F0,$3F,$13,$E0,$A4,$FC,$27,$E3,$FC,$47,$C8,$F8,$7F
	.dc.b	$93,$F0,$52,$BE,$3C,$7C,$CF,$83,$91,$F2,$9F,$9F,$F0,$BF,$CC,$3F
	.dc.b	$5A,$FE,$7F,$3F,$D0,$97,$C6,$E4,$D7,$C3,$EF,$E1,$F8,$7C,$24,$BF
	.dc.b	$E4,$77,$FE,$C4,$7F,$B8,$BF,$CC,$3D,$2F,$D8,$AF,$85,$F0,$1E,$1F
	.dc.b	$86,$F1,$7C,$37,$8F,$E1,$A6,$1D,$23,$CD,$F0,$D2,$3B,$BE,$9F,$86
	.dc.b	$F5,$7C,$2F,$AF,$E1,$A5,$79,$23,$DD,$F0,$BE,$FF,$86,$93,$E8,$8F
	.dc.b	$85,$F0,$DF,$D8,$F7,$F3,$B7,$E1,$FF,$BD,$70,$7F,$17,$7D,$F1,$9B
	.dc.b	$F9,$1F,$0F,$7C,$A5,$BE,$1F,$0F,$FA,$BB,$0F,$8F,$18,$F8,$7C,$FC
	.dc.b	$3D,$9F,$96,$F0,$F8,$69,$1F,$D4,$3F,$5B,$FF,$07,$BF,$42,$3C,$10
	.dc.b	$F8,$6A,$63,$E1,$AD,$71,$F8,$5D,$9C,$3E,$1A,$C7,$5F,$86,$E6,$74
	.dc.b	$F8,$6B,$1E,$1F,$85,$EF,$F0,$DE,$06,$C7,$97,$E1,$7C,$67,$93,$E1
	.dc.b	$AC,$7A,$7E,$1B,$CE,$7A,$3E,$1A,$C7,$B7,$E1,$7D,$7F,$0B,$EC,$6C
	.dc.b	$7C,$1F,$85,$F7,$FC,$2F,$C0,$6C,$7C,$5F,$84,$F8,$7F,$0B,$F1,$1B
	.dc.b	$1F,$27,$E1,$3E,$39,$F2,$3E,$16,$C7,$CD,$F8,$5F,$B5,$EB,$7F,$8E
	.dc.b	$A3,$07,$F8,$01,$C7,$83,$FE,$4E,$5F,$1B,$C0,$E3,$F0,$96,$B8,$7C
	.dc.b	$1F,$4F,$84,$EA,$D8,$E9,$F0,$9E,$03,$C3,$F0,$96,$3C,$1F,$07,$E4
	.dc.b	$F8,$4F,$2B,$63,$FC,$31,$F3,$CF,$D8,$9C,$DF,$95,$8C,$36,$3F,$CF
	.dc.b	$F3,$C1,$DB,$ED,$F8,$6F,$82,$7C,$2F,$86,$B1,$F0,$7E,$17,$E2,$FC
	.dc.b	$37,$C6,$6C,$7F,$A4,$3D,$6F,$C9,$3F,$E9,$2F,$A4,$7B,$1F,$A7,$F5
	.dc.b	$15,$AE,$7F,$F3,$E4,$2B,$77,$F3,$FE,$1F,$E9,$FC,$3F,$D4,$6C,$7D
	.dc.b	$3E,$9F,$AE,$7D,$8F,$87,$B1,$E1,$F6,$7C,$18,$1B,$EC,$61,$FB,$FF
	.dc.b	$0D,$2B,$EF,$7C,$3D,$D7,$EF,$7C,$3F,$E1,$3E,$1F,$F0,$AD,$AF,$77
	.dc.b	$AF,$DB,$D7,$EE,$D8,$FF,$23,$FE,$BA,$F8,$7F,$E0,$CF,$F2,$E8,$CC
	.dc.b	$63,$E2,$33,$F1,$12,$78,$FC,$45,$D3,$E3,$FD,$33,$D9,$F9,$8B,$D9
	.dc.b	$B1,$87,$BF,$C4,$4C,$3C,$7F,$13,$70,$FF,$53,$EB,$FE,$64,$9F,$3F
	.dc.b	$C4,$79,$BE,$27,$CE,$F6,$F6,$64,$FF,$B3,$B4,$7B,$38,$C7,$9F,$83
	.dc.b	$EE,$F0,$7D,$ED,$BF,$87,$F1,$1F,$0B,$E2,$7E,$1B,$63,$A7,$B3,$EC
	.dc.b	$3D,$BE,$BD,$8F,$97,$F1,$3F,$28,$F9,$7D,$F6,$3E,$7F,$81,$F3,$8F
	.dc.b	$9F,$F1,$76,$3F,$DC,$BE,$7F,$EE,$BC,$8F,$E8,$0F,$EE,$A3,$27,$E5
	.dc.b	$DC,$6F,$F1,$BE,$2B,$EC,$FC,$47,$DA,$6D,$7D,$EF,$88,$FB,$A7,$DE
	.dc.b	$F0,$6C,$78,$3D,$4E,$87,$5C,$6C,$7E,$1B,$C3,$FC,$29,$FD,$2F,$F5
	.dc.b	$EF,$E7,$F9,$FD,$68,$C7,$C4,$4A,$FF,$49,$D3,$E2,$31,$7F,$AC,$C1
	.dc.b	$E8,$FA,$FC,$FF,$63,$70,$E3,$F1,$1D,$CF,$07,$8B,$63,$AF,$8B,$E3
	.dc.b	$3C,$9F,$11,$63,$E5,$F9,$BF,$1F,$CC,$F9,$0D,$8F,$F5,$67,$33,$EA
	.dc.b	$3F,$45,$18,$FC,$39,$ED,$FF,$05,$9C,$F0,$7F,$04,$9C,$D7,$BB,$E2
	.dc.b	$7E,$27,$C5,$7C,$56,$C7,$6B,$FF,$61,$B7,$CE,$E9,$E7,$75,$7B,$79
	.dc.b	$DD,$FC,$CB,$5C,$5F,$A1,$F1,$52,$3E,$6F,$9B,$75,$6F,$FE,$8C,$EA
	.dc.b	$FD,$0D,$7D,$2F,$8A,$FB,$1E,$6F,$D9,$7E,$D7,$BD,$27,$EB,$FC,$55
	.dc.b	$C7,$9C,$1F,$F7,$9E,$0F,$8A,$FC,$11,$F8,$3F,$32,$D7,$E0,$32,$EC
	.dc.b	$D3,$C4,$7E,$13,$1E,$D9,$F9,$F2,$B3,$DF,$E3,$59,$3C,$DF,$C8,$1F
	.dc.b	$91,$F8,$BB,$5F,$37,$CA,$FA,$C7,$D7,$F2,$6C,$7D,$1F,$27,$ED,$1F
	.dc.b	$6F,$C8,$B1,$F9,$6F,$8B,$FC,$E1,$F9,$DF,$8B,$B1,$F9,$AF,$8B,$FD
	.dc.b	$01,$FE,$2B,$FD,$7F,$F6,$D1,$E7,$D5,$39,$F8,$C9,$3F,$EF,$4A,$FF
	.dc.b	$40,$B6,$39,$7C,$57,$3F,$8B,$B3,$B7,$BB,$E1,$39,$18,$E8,$FA,$A3
	.dc.b	$89,$87,$F0,$4E,$4F,$F6,$6B,$6B,$E6,$3E,$2A,$E3,$5C,$31,$F6,$7C
	.dc.b	$5F,$B7,$D1,$B8,$7A,$5F,$06,$3F,$8D,$D3,$E1,$4E,$0E,$B7,$F0,$31
	.dc.b	$EB,$97,$7B,$F1,$CF,$C9,$F3,$FE,$57,$8F,$70,$F8,$CF,$CD,$F8,$9F
	.dc.b	$9D,$F1,$36,$7E,$53,$F4,$7E,$27,$E9,$7A,$36,$7E,$73,$F5,$7D,$29
	.dc.b	$1F,$47,$E2,$6E,$DE,$4C,$7A,$CF,$D5,$6C,$7A,$30,$EB,$3F,$65,$B1
	.dc.b	$EC,$F8,$9F,$C1,$1F,$83,$F8,$9B,$5F,$7F,$E2,$3F,$0C,$7E,$1F,$D1
	.dc.b	$B1,$F8,$3F,$4B,$F1,$47,$E2,$FE,$22,$C7,$EB,$F8,$CF,$E8,$FC,$7F
	.dc.b	$E2,$31,$8C,$FF,$01,$1F,$DD,$DF,$B6,$EA,$78,$7C,$3F,$82,$7E,$9E
	.dc.b	$5E,$39,$F3,$1E,$32,$47,$D8,$C3,$C6,$F4,$F8,$7F,$F4,$56,$06,$FA
	.dc.b	$18,$73,$E8,$EB,$D1,$B8,$FA,$CF,$53,$CB,$D1,$91,$9C,$7D,$7F,$0D
	.dc.b	$87,$D7,$F0,$9E,$03,$C3,$F0,$B6,$B8,$78,$4D,$59,$F8,$67,$31,$86
	.dc.b	$E7,$D7,$87,$E5,$5F,$1D,$3D,$BA,$66,$FE,$87,$C2,$7C,$0F,$85,$F8
	.dc.b	$2D,$8E,$B9,$F7,$DB,$E8,$B7,$2E,$BA,$7F,$86,$67,$EA,$E5,$DA,$FC
	.dc.b	$AF,$CC,$F8,$59,$1E,$8F,$85,$B5,$F7,$3E,$17,$E8,$9F,$4B,$E1,$6D
	.dc.b	$7B,$BE,$17,$EA,$9F,$5B,$D2,$B1,$F0,$BD,$2F,$B2,$7D,$AF,$4A,$C7
	.dc.b	$C6,$C3,$9E,$3F,$8A,$C6,$ED,$EA,$CB,$9C,$FE,$3B,$2B,$3E,$EF,$4B
	.dc.b	$F0,$A7,$FB,$3F,$F6,$0F,$F0,$2C,$67,$2F,$AF,$1F,$E6,$D8,$FA,$2D
	.dc.b	$8F,$E8,$0F,$F8,$1D,$DF,$C3,$73,$F8,$5E,$9F,$0B,$67,$9B,$DF,$E1
	.dc.b	$64,$76,$F8,$5B,$A7,$F1,$EC,$7B,$3F,$C9,$DB,$D4,$7C,$11,$BC,$BB
	.dc.b	$3C,$71,$B9,$FA,$1F,$77,$C1,$CC,$BD,$BF,$09,$6F,$C1,$F0,$7F,$0C
	.dc.b	$F8,$9F,$07,$6B,$E1,$FC,$0F,$C7,$F8,$3F,$90,$D8,$F8,$FF,$05,$F2
	.dc.b	$CF,$99,$F0,$56,$3E,$5E,$1C,$F1,$F7,$63,$75,$F9,$B8,$F3,$9F,$84
	.dc.b	$D8,$FA,$3F,$07,$F5,$FB,$FE,$C3,$71,$EA,$4B,$EA,$DA,$EA,$FD,$CD
	.dc.b	$3D,$2D,$3E,$FF,$C1,$FE,$0C,$FC,$27,$C1,$D8,$FC,$1F,$C1,$FE,$1C
	.dc.b	$FC,$47,$C1,$D8,$FC,$3F,$BD,$F8,$B3,$F9,$06,$3F,$B8,$FD,$87,$FB
	.dc.b	$8E,$7F,$6E,$73,$5F,$E0,$99,$FE,$C1,$3F,$D9,$5B,$D0,$7F,$AA,$4E
	.dc.b	$FD,$EF,$05,$7E,$1F,$2F,$1C,$F9,$8F,$19,$23,$FB,$4D,$78,$AF,$F1
	.dc.b	$58,$78,$A7,$C1,$F0,$76,$7C,$1F,$07,$6D,$36,$31,$F0,$73,$1F,$E3
	.dc.b	$B4,$F8,$3C,$63,$E0,$F8,$2E,$5F,$05,$CD,$B7,$C3,$DE,$ED,$F0,$5D
	.dc.b	$DB,$1D,$32,$E7,$C3,$F2,$F2,$B3,$F2,$70,$E7,$97,$78,$F1,$E5,$E9
	.dc.b	$F7,$FD,$5F,$05,$67,$CE,$FB,$7E,$0A,$47,$AB,$E0,$AE,$1F,$23,$E0
	.dc.b	$BE,$11,$F0,$FE,$0A,$D7,$C0,$F7,$FE,$37,$C1,$7C,$76,$C7,$C4,$C7
	.dc.b	$9E,$FF,$83,$CA,$CF,$E0,$71,$E7,$97,$D6,$CA,$DF,$E2,$7E,$0B,$E9
	.dc.b	$1F,$4F,$E0,$6E,$1D,$49,$7F,$7B,$1E,$AF,$E2,$F1,$E7,$3F,$8F,$C6
	.dc.b	$CF,$F8,$CF,$F6,$2F,$EE,$4D,$F8,$3F,$C1,$46,$3A,$8F,$F7,$9D,$3F
	.dc.b	$A9,$BF,$DB,$70,$FE,$BA,$FF,$A2,$2F,$FC,$6E,$E5,$BF,$07,$D3,$DC
	.dc.b	$95,$C3,$E0,$2D,$70,$F7,$7C,$3F,$01,$E2,$6D,$76,$F8,$3F,$2E,$3E
	.dc.b	$66,$C7,$8B,$1E,$79,$F9,$B2,$B3,$FE,$77,$CD,$F0,$7E,$EF,$83,$F7
	.dc.b	$B6,$7D,$7F,$05,$F0,$8F,$87,$F0,$76,$3D,$FF,$07,$F1,$8F,$8F,$F0
	.dc.b	$76,$3E,$1F,$C0,$FC,$AF,$81,$F9,$6D,$8E,$A4,$E3,$F8,$F8,$75,$7E
	.dc.b	$4E,$1C,$DF,$97,$85,$DB,$E8,$FC,$0F,$D6,$F8,$1F,$AE,$DC,$7E,$9F
	.dc.b	$C1,$FD,$A3,$ED,$FC,$1D,$8F,$AF,$F0,$7F,$BF,$AF,$EC,$31,$F1,$63
	.dc.b	$19,$67,$DF,$93,$FE,$1A,$AF,$EA,$39,$7F,$85,$1F,$EA,$BC,$6C,$BB
	.dc.b	$61,$F8,$2C,$AD,$FE,$23,$2E,$B8,$F8,$8E,$95,$E3,$F8,$89,$1D,$FE
	.dc.b	$12,$CF,$E3,$31,$E7,$3F,$6F,$1B,$3F,$9B,$C3,$AB,$F9,$26,$E3,$E7
	.dc.b	$F8,$1F,$80,$7C,$1F,$84,$B8,$7B,$7E,$13,$E2,$7C,$1F,$C5,$6C,$7C
	.dc.b	$1F,$83,$F9,$07,$C9,$F8,$1B,$1F,$16,$F9,$E7,$FE,$98,$E1,$46,$33
	.dc.b	$F9,$18,$F4,$AD,$36,$3E,$66,$3D,$27,$96,$36,$7E,$87,$C0,$FD,$93
	.dc.b	$ED,$7C,$0D,$CF,$EB,$7B,$FF,$77,$E0,$FE,$F3,$63,$ED,$7C,$1F,$E0
	.dc.b	$7E,$0B,$F0,$4D,$8F,$BD,$7C,$F3,$F5,$37,$2F,$C0,$61,$CE,$7D,$CD
	.dc.b	$8F,$C2,$5F,$39,$F8,$4D,$8F,$C4,$7C,$1F,$E4,$4F,$C9,$7C,$1D,$C7
	.dc.b	$F1,$DF,$07,$F9,$5F,$81,$FC,$B3,$63,$F2,$5F,$07,$F9,$9F,$82,$FC
	.dc.b	$D3,$63,$F2,$D7,$CF,$3F,$A4,$DC,$BF,$31,$87,$39,$FA,$CD,$8F,$CE
	.dc.b	$5F,$39,$FB,$4D,$8F,$D0,$57,$3B,$57,$FB,$77,$95,$39,$8C,$61,$D6
	.dc.b	$B7,$E7,$2D,$39,$AE,$58,$F7,$BD,$D2,$31,$2C,$7F,$8D,$F0,$7F,$7B
	.dc.b	$7F,$8E,$ED,$7C,$74,$D8,$C6,$F9,$D6,$E7,$FE,$9D,$C3,$F9,$2C,$71
	.dc.b	$D3,$96,$F6,$B5,$92,$3F,$C1,$1F,$1B,$FE,$0D,$3F,$C3,$25,$F8,$4C
	.dc.b	$FC,$7C,$BF,$F5,$1C,$FF,$83,$1B,$1C,$BE,$12,$46,$BE,$06,$F0,$7B
	.dc.b	$E3,$E4,$DF,$57,$99,$E2,$C3,$B4,$EC,$71,$AB,$FC,$57,$18,$7C,$4E
	.dc.b	$E3,$58,$F8,$36,$FF,$58,$D3,$2F,$7E,$59,$CB,$C7,$F0,$3F,$08,$F8
	.dc.b	$7F,$03,$63,$DF,$F0,$3F,$1B,$E0,$BE,$3B,$63,$E1,$FC,$1F,$CA,$F8
	.dc.b	$3F,$96,$D8,$F8,$F8,$73,$DF,$CD,$A3,$2C,$6B,$E6,$CB,$A5,$69,$B1
	.dc.b	$EA,$C7,$A4,$F2,$6C,$7F,$96,$3F,$18,$F7,$65,$F8,$35,$AF,$E3,$5C
	.dc.b	$DF,$CD,$A4,$6A,$B8,$67,$B3,$86,$56,$3A,$65,$CC,$E9,$AD,$8F,$06
	.dc.b	$7D,$F5,$F0,$36,$3C,$9F,$09,$E3,$F8,$2F,$23,$63,$D1,$F0,$1E,$73
	.dc.b	$D1,$F0,$16,$3D,$9F,$01,$EB,$F8,$3F,$63,$63,$FD,$FD,$C2,$FD,$33
	.dc.b	$F3,$D1,$F0,$FE,$17,$FC,$D1,$39,$EA,$F8,$91,$F1,$7E,$16,$57,$C9
	.dc.b	$F8,$5B,$7F,$1F,$F1,$9F,$F1,$7F,$39,$A7,$F8,$B6,$BF,$85,$C6,$7E
	.dc.b	$43,$BE,$7E,$2B,$FA,$1D,$0C,$F3,$A7,$8B,$1F,$DD,$65,$FE,$31,$CB
	.dc.b	$FD,$26,$FF,$8D,$5B,$F0,$7F,$C1,$E3,$13,$E0,$5F,$07,$93,$5F,$07
	.dc.b	$67,$E2,$E9,$E3,$C3,$F1,$19,$78,$A7,$E3,$1E,$2F,$85,$ED,$F0,$DD
	.dc.b	$DB,$1E,$6F,$13,$C5,$F0,$BE,$36,$C7,$6D,$BD,$9C,$1F,$BB,$3E,$84
	.dc.b	$76,$7E,$67,$C2,$FA,$FE,$0F,$D8,$D9,$F8,$7F,$07,$EF,$F8,$3F,$80
	.dc.b	$D8,$EC,$73,$D8,$FF,$3C,$C3,$B9,$F4,$36,$F9,$1F,$09,$27,$8C,$FD
	.dc.b	$D8,$F9,$9F,$0B,$F3,$7D,$2B,$5F,$51,$F8,$D5,$CE,$38,$E1,$F2,$9A
	.dc.b	$7D,$3B,$DC,$7D,$A6,$C7,$97,$8B,$FA,$23,$C2,$BF,$A9,$C6,$1F,$90
	.dc.b	$FD,$78,$FC,$6F,$3F,$E2,$5E,$17,$FB,$AC,$BA,$DF,$F4,$DE,$66,$38
	.dc.b	$AF,$CC,$3A,$9F,$D3,$7A,$57,$2F,$05,$7C,$A9,$F1,$F0,$FD,$A8,$C4
	.dc.b	$7C,$C7,$E0,$3E,$68,$F9,$8D,$CB,$F3,$78,$75,$C2,$47,$A3,$3F,$55
	.dc.b	$6C,$F9,$85,$C3,$E4,$4F,$BB,$2F,$5B,$F1,$E7,$D2,$DC,$BD,$2F,$CD
	.dc.b	$CB,$D0,$D3,$F0,$C5,$9F,$C0,$3D,$72,$F2,$C7,$86,$3E,$B6,$5F,$27
	.dc.b	$2F,$3E,$5F,$0B,$1F,$A9,$7F,$DE,$E7,$37,$D6,$BD,$33,$CC,$FA,$98
	.dc.b	$76,$8E,$58,$FF,$06,$C7,$B6,$73,$0F,$21,$30,$E4,$7C,$02,$E5,$F0
	.dc.b	$1F,$AF,$1E,$E6,$9F,$33,$0E,$F1,$F5,$0E,$8E,$23,$F1,$F7,$E5,$9F
	.dc.b	$51,$F9,$0B,$F2,$4F,$AE,$BF,$17,$3C,$8D,$C7,$E6,$AB,$93,$E2,$9F
	.dc.b	$C3,$9D,$CB,$3C,$1E,$71,$D7,$3F,$BE,$FB,$AB,$9B,$F9,$1A,$F2,$1F
	.dc.b	$8B,$BF,$CF,$57,$0B,$ED,$3F,$B3,$3F,$0D,$7C,$B0,$FB,$33,$FA,$99
	.dc.b	$E4,$7A,$F0,$F4,$3C,$5F,$13,$C8,$DB,$F0,$F3,$FC,$0B,$C0,$FB,$F1
	.dc.b	$CD,$F4,$1F,$15,$E4,$F1,$C7,$E6,$D7,$A2,$38,$9F,$0F,$3F,$46,$7F
	.dc.b	$87,$BF,$0D,$6C,$F5,$56,$E3,$C1,$8F,$E5,$A3,$F1,$13,$D5,$B3,$F4
	.dc.b	$71,$F4,$C7,$9A,$79,$9D,$73,$F9,$95,$F5,$DB,$3E,$F3,$99,$6F,$F2
	.dc.b	$8F,$3B,$EB,$1F,$41,$B8,$7F,$08,$C3,$B6,$3E,$EC,$7F,$AB,$E7,$F1
	.dc.b	$1F,$36,$7F,$7A,$64,$7C,$AA,$F6,$63,$E0,$3B,$57,$1C,$BE,$A4,$FE
	.dc.b	$38,$EF,$5C,$4F,$CC,$C7,$A8,$F4,$4F,$17,$E2,$67,$F0,$1B,$87,$9D
	.dc.b	$FD,$4D,$7D,$43,$CE,$7D,$2C,$BD,$CF,$F7,$7D,$7F,$A7,$3F,$E0,$06
	.dc.b	$CF,$E0,$E3,$93,$E8,$BF,$06,$5F,$98,$7E,$8E,$5F,$E1,$3D,$BF,$B2
	.dc.b	$57,$F8,$5D,$B1,$F8,$5D,$7E,$7E,$1F,$C1,$63,$E1,$3F,$9D,$7E,$1D
	.dc.b	$7F,$87,$F8,$BF,$BD,$B6,$7F,$C2,$BC,$7F,$0F,$2F,$9B,$1F,$E1,$8E
	.dc.b	$2F,$EF,$15,$FD,$F2,$3F,$C3,$7C,$7F,$E1,$18,$FF,$15,$73,$FF,$85
	.dc.b	$27,$FC,$3E,$F3,$DF,$F9,$BC,$FF,$59,$E5,$FF,$07,$BF,$E3,$16,$E9
	.dc.b	$FE,$24,$E9,$FF,$11,$1F,$E2,$3F,$87,$B8,$FC,$8B,$FB,$2F,$F8,$B3
	.dc.b	$AB,$FC,$51,$F0,$FF,$E2,$96,$C7,$F9,$30,$FF,$34,$77,$FF,$93,$1A
	.dc.b	$73,$DF,$DF,$7F,$E3,$AE,$EF,$F2,$2E,$7F,$E7,$1E,$EB,$7F,$DD,$BC
	.dc.b	$4F,$F3,$05,$FF,$90,$1B,$5F,$E7,$7F,$0F,$FC,$C1,$3F,$E4,$46,$C7
	.dc.b	$F9,$EF,$C3,$FF,$30,$4F,$F9,$23,$C3,$B3,$FE,$7F,$F3,$3F,$CF,$9E
	.dc.b	$5F,$F9,$F5,$B9,$7F,$9E,$89,$7F,$E5,$0F,$2B,$FC,$F4,$FF,$94,$BC
	.dc.b	$AF,$F3,$C3,$FE,$54,$F2,$AE,$3F,$EA,$8D,$3F,$D8,$BE,$4F,$FB,$19
	.dc.b	$B8,$FF,$B0,$B5,$FF,$64,$EB,$FE,$CA,$6C,$7F,$B1,$BC,$9F,$F6,$6F
	.dc.b	$1F,$FB,$39,$B1,$FE,$CA,$C3,$9E,$7F,$EA,$4C,$6E,$5F,$EB,$8C,$79
	.dc.b	$CF,$FA,$A1,$B1,$FE,$BC,$C3,$9C,$FF,$AB,$1B,$1F,$EC,$0B,$E7,$87
	.dc.b	$FB,$43,$0B,$7F,$EC,$9C,$39,$CF,$FB,$51,$B1,$FE,$CD,$C3,$9C,$FF
	.dc.b	$B6,$1B,$1F,$EF,$9E,$E7,$A7,$FE,$F5,$AF,$F7,$6F,$A9,$6B,$FD,$5D
	.dc.b	$4F,$53,$FD,$F1,$5F,$EF,$26,$CE,$BD,$3F,$F7,$BC,$FF,$BD,$9B,$1F
	.dc.b	$EC,$8F,$23,$FD,$EF,$D1,$FE,$F8,$6D,$7F,$BD,$3A,$3F,$C7,$1C,$1F
	.dc.b	$E3,$9F,$49,$FB,$1B,$BF,$FD,$FF,$5A,$B6,$5F,$EE,$EA,$D4,$7F,$A6
	.dc.b	$8C,$DF,$2E,$FE,$15,$B6,$C7,$6E,$EB,$7F,$EA,$3A,$FF,$7B,$69,$E6
	.dc.b	$7D,$58,$75,$72,$FA,$4C,$5F,$A3,$2F,$3C,$62,$FC,$B5,$FE,$FF,$7B
	.dc.b	$57,$96,$75,$87,$8B,$1F,$1F,$7C,$9E,$5E,$15,$C7,$97,$9F,$E8,$3D
	.dc.b	$3E,$15,$AF,$07,$9F,$EC,$E4,$F6,$B6,$3C,$98,$FC,$0F,$1F,$A7,$17
	.dc.b	$D1,$DD,$1C,$F3,$EF,$7E,$BC,$B1,$7B,$C7,$A5,$7B,$5B,$1C,$FC,$7F
	.dc.b	$9B,$E3,$FC,$E6,$E1,$F1,$7C,$9F,$A3,$E7,$FD,$26,$C7,$C9,$F3,$7E
	.dc.b	$A9,$F5,$BC,$DB,$1F,$37,$0E,$79,$FC,$DC,$6E,$5E,$BC,$79,$CF,$D1
	.dc.b	$6C,$7B,$F1,$E7,$3F,$57,$1B,$3F,$0F,$0F,$C2,$F9,$FD,$32,$FC,$6E
	.dc.b	$E8,$F4,$47,$DF,$D2,$E3,$F7,$DE,$71,$C3,$2F,$53,$ED,$3F,$0D,$F1
	.dc.b	$1F,$92,$3F,$27,$F0,$96,$3F,$15,$F0,$9F,$96,$3F,$2F,$F0,$96,$3F
	.dc.b	$23,$FB,$27,$E9,$1B,$F1,$3F,$9E,$8C,$66,$DF,$F5,$D0,$9E,$D3,$9A
	.dc.b	$FD,$4C,$BE,$23,$37,$C1,$F0,$57,$E6,$F1,$F1,$DF,$98,$F1,$C7,$87
	.dc.b	$0F,$13,$63,$CF,$F1,$1E,$53,$CD,$F1,$16,$3D,$7F,$0F,$E9,$F8,$7F
	.dc.b	$53,$63,$DF,$F1,$1E,$D3,$DD,$F1,$16,$3E,$1F,$C4,$7C,$13,$A6,$5E
	.dc.b	$D1,$E2,$38,$13,$DA,$BE,$5E,$56,$F8,$5E,$E3,$B4,$62,$D8,$F8,$AF
	.dc.b	$CA,$7D,$4F,$88,$FA,$5F,$0F,$F4,$DB,$1F,$63,$E1,$FE,$B1,$F5,$FE
	.dc.b	$22,$C7,$DC,$F4,$3E,$D1,$F6,$FD,$0B,$1F,$80,$C3,$EF,$7A,$1D,$34
	.dc.b	$F2,$3C,$9F,$36,$3D,$6B,$C0,$75,$24,$7A,$EF,$AB,$CA,$3F,$1B,$5D
	.dc.b	$79,$FF,$12,$F5,$7F,$17,$7D,$6B,$BD,$74,$9F,$C7,$E3,$E9,$8C,$1F
	.dc.b	$8E,$F8,$7F,$CB,$1F,$97,$F4,$6C,$7E,$73,$DF,$FC,$D1,$F9,$BF,$82
	.dc.b	$B1,$FA,$0F,$88,$FE,$77,$EA,$A3,$F4,$11,$87,$9E,$7F,$6A,$5F,$11
	.dc.b	$8D,$3F,$A8,$46,$2B,$C0,$4A,$FB,$D8,$78,$1F,$C0,$61,$D6,$7C,$47
	.dc.b	$F4,$2A,$F1,$37,$E2,$3C,$BF,$17,$E6,$6C,$6F,$E0,$7D,$3F,$05,$EA
	.dc.b	$6C,$73,$F8,$8F,$6F,$C1,$FB,$9B,$1D,$F0,$E7,$9F,$B7,$1B,$AF,$B3
	.dc.b	$1E,$73,$F0,$5B,$1F,$03,$0E,$73,$F1,$5B,$1F,$13,$E0,$FE,$69,$F3
	.dc.b	$BE,$0E,$E5,$EF,$F8,$4F,$A3,$F1,$1F,$49,$B1,$F0,$FE,$0B,$EA,$9F
	.dc.b	$5B,$E0,$EC,$7C,$7C,$3E,$CF,$C1,$F4,$D3,$EB,$3C,$E7,$E6,$63,$67
	.dc.b	$ED,$3C,$9F,$B9,$87,$59,$F0,$1F,$4E,$BF,$0B,$F0,$BD,$A3,$EB,$3C
	.dc.b	$F4,$FC,$0E,$16,$7F,$C1,$3C,$68,$ED,$3F,$86,$C3,$B6,$3E,$0B,$FC
	.dc.b	$05,$FE,$4B,$E1,$7F,$26,$D8,$FC,$27,$C2,$FE,$5B,$E1,$3F,$2E,$D8
	.dc.b	$FC,$47,$C2,$FE,$68,$FF,$0A,$FE,$AD,$FE,$08,$CF,$F2,$B1,$CF,$2F
	.dc.b	$E2,$73,$FE,$0A,$C5,$97,$0F,$84,$EB,$8F,$E4,$F3,$B8,$7E,$85,$ED
	.dc.b	$1E,$03,$FC,$33,$87,$77,$E0,$E3,$E2,$F8,$5F,$1F,$C1,$59,$DB,$E6
	.dc.b	$F8,$59,$1C,$BE,$16,$E5,$E6,$F8,$3F,$5F,$C1,$FB,$1B,$5C,$89,$3E
	.dc.b	$CC,$7B,$C7,$93,$0E,$7C,$1F,$E4,$5D,$19,$62,$BF,$C7,$F0,$C7,$AE
	.dc.b	$3E,$19,$F6,$E5,$F3,$3E,$16,$47,$C0,$F8,$5B,$5F,$33,$E1,$7E,$8F
	.dc.b	$C1,$FD,$26,$D7,$C6,$F8,$5F,$AB,$F0,$5F,$59,$B1,$F2,$BE,$0B,$EC
	.dc.b	$FC,$2F,$DA,$6C,$7C,$EB,$E7,$8F,$78,$FA,$38,$F3,$7E,$9E,$36,$7E
	.dc.b	$FB,$C9,$FC,$0F,$C2,$5E,$1F,$88,$F9,$EB,$99,$DC,$B3,$F8,$8C,$FC
	.dc.b	$ED,$8E,$E7,$E0,$BE,$13,$F2,$27,$E4,$BE,$12,$C7,$E1,$BE,$0B,$F2
	.dc.b	$A7,$E5,$BE,$12,$C7,$F9,$8B,$F5,$6F,$F0,$7E,$7F,$E7,$B7,$FC,$1D
	.dc.b	$19,$C7,$FC,$E3,$18,$7F,$37,$8E,$3E,$17,$BC,$7F,$70,$9F,$F0,$95
	.dc.b	$78,$49,$1F,$A2,$C3,$C2,$EC,$EB,$F0,$BD,$9B,$3E,$5F,$85,$F0,$FC
	.dc.b	$2F,$89,$B1,$E9,$F8,$5F,$29,$E6,$F8,$2B,$1E,$DF,$81,$F4,$FC,$2F
	.dc.b	$A9,$B1,$F0,$70,$F6,$FC,$2F,$4C,$BD,$4F,$27,$D9,$8F,$5A,$F0,$1D
	.dc.b	$4F,$95,$87,$56,$9F,$37,$E0,$FE,$49,$F2,$BE,$0A,$C7,$D1,$F8,$0F
	.dc.b	$9B,$F0,$7F,$39,$B1,$F5,$7E,$07,$E8,$FC,$1F,$D2,$6C,$7D,$9F,$83
	.dc.b	$FA,$A7,$D6,$F8,$3B,$1D,$4F,$BD,$87,$56,$9D,$4F,$C1,$5F,$56,$9F
	.dc.b	$85,$F8,$3F,$C0,$9D,$36,$FB,$CF,$37,$F0,$15,$7A,$7F,$14,$F3,$8F
	.dc.b	$C2,$5D,$9F,$C7,$3F,$8C,$F8,$3F,$C6,$FC,$15,$C7,$F2,$8F,$E4,$3E
	.dc.b	$0A,$47,$FA,$E3,$F5,$4F,$EE,$F8,$FE,$3E,$7F,$D8,$B1,$FE,$03,$8F
	.dc.b	$F2,$0B,$71,$FF,$6B,$2F,$80,$DF,$C2,$78,$30,$E9,$F0,$57,$0E,$2F
	.dc.b	$81,$ED,$1E,$13,$18,$74,$AE,$71,$E5,$BB,$5E,$2F,$82,$F2,$9E,$6F
	.dc.b	$84,$B5,$EB,$F8,$3F,$49,$EA,$F8,$4B,$1E,$FF,$81,$F6,$FC,$27,$B9
	.dc.b	$B1,$F0,$FE,$13,$E0,$FC,$0F,$C2,$6C,$7C,$7F,$81,$F8,$A7,$C6,$F8
	.dc.b	$4B,$1F,$2F,$E1,$3E,$4F,$C2,$7C,$A6,$C7,$CF,$F8,$3F,$9A,$7C,$EF
	.dc.b	$83,$B1,$F4,$FE,$13,$E8,$FC,$1F,$D2,$6C,$7D,$7F,$83,$FA,$A7,$FB
	.dc.b	$EB,$C0,$FE,$B1,$CB,$FE,$7C,$EA,$4E,$7E,$12,$5F,$84,$FF,$7B,$DF
	.dc.b	$FB,$CA,$98,$76,$C3,$EF,$B7,$AB,$EF,$E3,$E0,$C3,$8F,$C4,$5B,$FF
	.dc.b	$42,$E2,$F8,$4F,$27,$C2,$F9,$5B,$7E,$0F,$85,$F4,$1E,$9F,$85,$B1
	.dc.b	$E4,$F8,$4F,$67,$C2,$7B,$5B,$1E,$8F,$84,$F8,$07,$C1,$F8,$4B,$1E
	.dc.b	$CF,$82,$F8,$87,$C5,$F8,$2B,$1F,$03,$E0,$3E,$47,$C1,$7C,$96,$C7
	.dc.b	$C4,$F8,$3F,$99,$F0,$9F,$35,$B1,$F2,$3E,$13,$E8,$1F,$47,$E0,$AC
	.dc.b	$7C,$CF,$84,$FA,$9F,$09,$F5,$5B,$1F,$43,$E0,$BE,$C1,$FE,$77,$E7
	.dc.b	$FF,$56,$FE,$A7,$FD,$22,$51,$FE,$B6,$E8,$56,$FA,$F8,$7C,$43,$0F
	.dc.b	$E9,$51,$86,$E1,$CA,$75,$F0,$DC,$63,$13,$AC,$7F,$D8,$6B,$E1,$35
	.dc.b	$18,$F8,$2B,$87,$79,$B4,$9B,$F0,$33,$0F,$4E,$5E,$3F,$84,$C6,$5E
	.dc.b	$EF,$84,$E2,$72,$F8,$3B,$7C,$0F,$2F,$C2,$76,$F8,$4E,$ED,$8D,$5F
	.dc.b	$A3,$E1,$3C,$67,$93,$E1,$2C,$EB,$83,$E3,$7C,$0F,$A0,$F4,$FC,$25
	.dc.b	$9D,$70,$BE,$0B,$DB,$F0,$9E,$E6,$CE,$B9,$FE,$09,$F0,$BE,$12,$C7
	.dc.b	$2F,$84,$F8,$A7,$C6,$F8,$4B,$1D,$BE,$13,$E5,$FE,$A5,$FC,$C3,$9F
	.dc.b	$EC,$46,$27,$3F,$09,$1F,$55,$7F,$2F,$6C,$7D,$6F,$D0,$5F,$02,$E8
	.dc.b	$C3,$6B,$E3,$F3,$7F,$37,$F2,$33,$5E,$2F,$86,$F1,$FC,$2D,$BE,$EF
	.dc.b	$9B,$E1,$64,$78,$BE,$16,$E2,$DF,$F3,$4F,$F4,$3F,$2F,$C3,$D1,$F8
	.dc.b	$A8,$C5,$FD,$8D,$BC,$F3,$F0,$B6,$F8,$7F,$09,$6B,$DE,$FC,$6F,$85
	.dc.b	$91,$F0,$BE,$12,$E3,$E1,$8F,$D3,$BF,$D4,$FC,$FF,$83,$CF,$E1,$F8
	.dc.b	$64,$62,$BE,$5F,$C3,$7D,$2F,$86,$FA,$6D,$8F,$9F,$F0,$DF,$58,$FA
	.dc.b	$FF,$0B,$63,$E5,$E0,$F4,$FE,$97,$CA,$FD,$5D,$3C,$87,$EE,$A7,$55
	.dc.b	$F0,$67,$5C,$9F,$BA,$EC,$7C,$3F,$E1,$3E,$1F,$F0,$AD,$BF,$C0,$FC
	.dc.b	$3F,$E2,$0F,$C4,$FC,$25,$8F,$1F,$B3,$E2,$E7,$F8,$0D,$8F,$C4,$FC
	.dc.b	$3F,$E4,$3E,$17,$F2,$2D,$8F,$D3,$7E,$91,$FD,$2B,$CD,$34,$7C,$47
	.dc.b	$F0,$27,$FA,$7F,$17,$F8,$07,$DD,$F8,$3D,$1F,$3D,$B8,$74,$F8,$5E
	.dc.b	$1F,$0D,$C5,$B1,$E0,$F8,$6E,$87,$5F,$86,$B1,$F3,$BE,$07,$E6,$F1
	.dc.b	$FD,$B6,$C7,$A3,$E1,$7C,$9F,$0D,$E5,$6C,$7B,$3E,$1B,$D0,$7A,$7E
	.dc.b	$1A,$C7,$DA,$F8,$4F,$D5,$7E,$4F,$EE,$8C,$1F,$97,$71,$C3,$F7,$39
	.dc.b	$3E,$2F,$C3,$FC,$3F,$4B,$E2,$36,$FE,$5F,$9B,$F1,$CF,$91,$E3,$D8
	.dc.b	$FD,$9F,$EF,$EE,$3F,$D3,$36,$33,$C7,$89,$FA,$1E,$2C,$9F,$AF,$DB
	.dc.b	$75,$F2,$75,$3A,$7E,$B1,$F5,$F9,$AD,$FE,$63,$E2,$3D,$FB,$FE,$95
	.dc.b	$B1,$F8,$2E,$3F,$BD,$1F,$7D,$B1,$F8,$6F,$88,$FC,$11,$FD,$12,$FF
	.dc.b	$AC,$64,$E3,$FC,$4C,$FE,$C4,$F9,$BA,$5B,$F9,$8F,$1F,$8B,$E5,$F1
	.dc.b	$36,$7F,$9C,$63,$FE,$A0,$C7,$E8,$3F,$D5,$63,$3F,$13,$E1,$F8,$9F
	.dc.b	$13,$FC,$3F,$F5,$1F,$1F,$AA,$EA,$FE,$D3,$7F,$C8,$A7,$31,$8C,$7A
	.dc.b	$FC,$57,$AF,$E2,$BD,$8D,$BF,$1F,$C5,$7B,$CF,$81,$EE,$D8,$E9,$BF
	.dc.b	$F3,$3F,$CA,$7F,$B2,$46,$0F,$E4,$0E,$31,$F5,$FC,$57,$C9,$F8,$BF
	.dc.b	$94,$DA,$F8,$3F,$17,$F3,$4F,$9D,$F1,$76,$3D,$5C,$9D,$EB,$E0,$36
	.dc.b	$3E,$4F,$C5,$FD,$5F,$7F,$EB,$36,$3E,$6F,$C4,$FD,$93,$ED,$7C,$55
	.dc.b	$8F,$8D,$CD,$EB,$BF,$98,$D8,$FA,$BF,$15,$F8,$1F,$8A,$FC,$13,$63
	.dc.b	$EC,$FC,$07,$E1,$4F,$C3,$7C,$55,$8F,$A5,$D7,$F1,$E7,$EC,$36,$3F
	.dc.b	$BE,$FE,$CA,$FC,$51,$E4,$FC,$DF,$8C,$FF,$0A,$CE,$63,$15,$F8,$89
	.dc.b	$FC,$A1,$FE,$18,$FD,$2F,$FB,$74,$7F,$88,$FC,$14,$FE,$1F,$B7,$FB
	.dc.b	$57,$67,$F3,$4A,$FE,$97,$E0,$27,$3F,$05,$B9,$FE,$B5,$4F,$8C,$C7
	.dc.b	$6F,$F4,$0F,$0F,$FB,$94,$78,$7E,$36,$5F,$F8,$4F,$5E,$D7,$FA,$77
	.dc.b	$FD,$59,$1E,$0F,$0F,$C7,$F1,$7F,$9B,$6C,$7E,$65,$F3,$FB,$F9,$8F
	.dc.b	$1B,$EF,$C6,$47,$7F,$8D,$B8,$FF,$54,$1B,$BC,$6F,$83,$F1,$5F,$A9
	.dc.b	$6D,$FE,$9D,$F8,$BF,$15,$98,$F8,$2F,$CD,$F8,$D9,$1E,$DD,$6E,$3D
	.dc.b	$F1,$FF,$09,$3F,$17,$85,$F1,$5F,$B7,$7E,$8B,$6F,$E6,$FC,$35,$AE
	.dc.b	$0F,$DB,$DE,$4F,$CB,$F8,$DB,$5D,$FC,$7F,$B7,$F0,$FF,$C0,$DB,$5F
	.dc.b	$C0,$5F,$BF,$F0,$D9,$8F,$B6,$FE,$0B,$E1,$E4,$7F,$07,$F3,$6E,$1E
	.dc.b	$B7,$FC,$61,$F9,$5F,$F4,$4F,$8B,$FF,$10,$F0,$27,$3F,$13,$FC,$4D
	.dc.b	$DF,$C3,$DC,$7F,$C4,$58,$7F,$A9,$DF,$E2,$9E,$8F,$4F,$88,$FE,$34
	.dc.b	$D8,$C9,$C3,$DF,$CB,$4C,$E9,$F9,$1B,$FD,$3C,$63,$4F,$F4,$77,$80
	.dc.b	$9C,$D7,$D5,$F4,$FD,$7F,$0D,$FC,$A5,$B8,$7F,$27,$7D,$FF,$0F,$98
	.dc.b	$F4,$B6,$BC,$1F,$0F,$72,$EF,$8A,$3F,$9A,$7A,$BF,$23,$E1,$EE,$3F
	.dc.b	$09,$B8,$7A,$7E,$02,$E3,$C9,$B5,$FE,$A4,$D7,$D5,$F0,$F8,$B4,$62
	.dc.b	$7F,$9F,$7A,$DF,$5B,$E1,$FF,$A2,$36,$32,$7D,$2F,$87,$CB,$4C,$DF
	.dc.b	$CB,$F8,$6C,$B6,$3F,$A5,$7B,$7F,$82,$F8,$0F,$EA,$0D,$8C,$9F,$7B
	.dc.b	$E1,$72,$D3,$35,$F5,$FD,$3C,$B6,$3F,$AB,$7B,$BF,$8E,$F8,$4F,$EB
	.dc.b	$8D,$8C,$9F,$8A,$F8,$5C,$B4,$CD,$7E,$0F,$E1,$32,$D8,$FE,$C5,$EF
	.dc.b	$7F,$A2,$7F,$2B,$FE,$F3,$3F,$C1,$23,$0F,$F6,$B7,$5F,$09,$6B,$FC
	.dc.b	$F5,$5F,$EC,$9A,$E1,$18,$F8,$3D,$47,$F2,$39,$FE,$F5,$CF,$FA,$6F
	.dc.b	$82,$F0,$7C,$1F,$F7,$76,$E1,$FD,$D5,$F2,$7C,$0E,$63,$BB,$70,$E9
	.dc.b	$EE,$DC,$3F,$3B,$8F,$F3,$7B,$FE,$FD,$F0,$5E,$DF,$6F,$FC,$08,$D8
	.dc.b	$C9,$EA,$F6,$32,$D3,$38,$FA,$7D,$5C,$B6,$3F,$C1,$5F,$07,$F2,$7D
	.dc.b	$1F,$F0,$83,$63,$27,$C6,$F3,$72,$D3,$35,$F1,$7C,$BC,$B6,$3F,$C2
	.dc.b	$DF,$01,$F5,$7C,$AF,$F0,$E3,$63,$27,$D2,$F1,$F2,$D3,$35,$F4,$7C
	.dc.b	$2C,$B6,$3F,$D8,$7F,$A9,$7F,$64,$8F,$D4,$F0,$FF,$B8,$67,$27,$E1
	.dc.b	$7C,$0F,$F5,$D5,$6F,$C5,$53,$BB,$37,$F8,$B6,$BF,$86,$EC,$B8,$FE
	.dc.b	$0D,$E7,$F0,$1F,$91,$EA,$B5,$F8,$C6,$D7,$E2,$79,$AF,$47,$DE,$6D
	.dc.b	$7F,$03,$C9,$C9,$8B,$F5,$BE,$8F,$81,$FC,$F7,$15,$BF,$CC,$B6,$BF
	.dc.b	$BC,$7E,$59,$F0,$D7,$07,$F7,$6E,$04,$E6,$30,$FF,$78,$91,$C3,$0F
	.dc.b	$F7,$A4,$66,$39,$FC,$0E,$27,$E5,$B7,$1F,$A6,$F0,$32,$FF,$B9,$EB
	.dc.b	$FC,$1C,$E8,$C7,$C0,$EA,$FD,$2F,$D6,$3C,$3E,$FD,$9E,$4D,$9F,$7F
	.dc.b	$BF,$78,$3F,$A2,$E0,$8C,$5F,$F2,$F8,$7B,$B8,$BF,$E5,$F1,$89,$FC
	.dc.b	$5E,$1F,$03,$DD,$FC,$3B,$63,$27,$AF,$E4,$72,$D3,$37,$F4,$BE,$0B
	.dc.b	$2D,$8F,$CB,$E5,$F3,$3E,$0B,$F2,$6D,$8C,$9F,$1F,$E3,$72,$D3,$35
	.dc.b	$F7,$BE,$47,$2D,$8F,$D1,$EB,$F6,$3E,$13,$F3,$ED,$8C,$9F,$4F,$E1
	.dc.b	$32,$D3,$35,$F8,$AF,$8F,$CB,$63,$F5,$FB,$FE,$13,$E3,$FF,$56,$D8
	.dc.b	$C9,$F7,$FE,$1B,$2D,$33,$5F,$96,$F8,$6C,$B6,$3F,$23,$C8,$7F,$27
	.dc.b	$0F,$A2,$FE,$FF,$3F,$7B,$E1,$F5,$3F,$E4,$39,$CD,$7D,$F9,$CD,$7F
	.dc.b	$82,$29,$78,$AF,$EE,$51,$87,$C2,$62,$B5,$1F,$DD,$0D,$1F,$E6,$5E
	.dc.b	$B7,$CA,$69,$FE,$F2,$D7,$FC,$59,$5F,$DF,$DD,$BF,$C6,$39,$B8,$7C
	.dc.b	$06,$2F,$FC,$B9,$DA,$9C,$E1,$FC,$61,$DF,$C3,$DC,$3C,$F7,$FE,$08
	.dc.b	$C3,$BB,$FB,$AC,$39,$4E,$B1,$FF,$0D,$7C,$3F,$1A,$FF,$3F,$5F,$F8
	.dc.b	$E5,$F0,$F5,$79,$3E,$1F,$F9,$3B,$70,$FE,$5E,$FA,$3E,$1F,$31,$D6
	.dc.b	$FF,$E0,$05,$F2,$8C,$7C,$56,$A3,$BB,$FC,$CB,$B7,$E0,$FC,$45,$AE
	.dc.b	$4D,$AF,$2F,$C4,$5E,$4F,$F2,$0E,$28,$C5,$FF,$8A,$E3,$FC,$6C,$FF
	.dc.b	$44,$C3,$E6,$C7,$F4,$0E,$EF,$9F,$F0,$F6,$BD,$0D,$C7,$E1,$FC,$3D
	.dc.b	$CB,$BB,$6B,$B7,$C3,$DA,$FA,$0F,$F4,$CF,$07,$ED,$FC,$3D,$AF,$88
	.dc.b	$DA,$F9,$FF,$0F,$71,$F5,$B6,$BD,$5F,$0F,$6B,$EE,$3F,$D6,$3C,$4F
	.dc.b	$C3,$FC,$3D,$AF,$A0,$DA,$FB,$7F,$0F,$71,$F8,$ED,$AF,$8D,$F0,$F6
	.dc.b	$BF,$10,$FF,$64,$F1,$BF,$27,$F0,$F6,$BE,$E3,$6B,$F0,$FF,$0F,$71
	.dc.b	$FA,$6D,$AF,$A5,$F0,$F6,$BF,$28,$FF,$70,$F2,$7F,$4D,$F3,$7F,$E3
	.dc.b	$6F,$05,$18,$AD,$5F,$E5,$EB,$FC,$69,$5A,$8C,$7C,$46,$AB,$FD,$AD
	.dc.b	$5F,$E3,$49,$FE,$0F,$5B,$9C,$FC,$47,$F8,$03,$CC,$F0,$FC,$45,$BE
	.dc.b	$CD,$AE,$7F,$11,$76,$FF,$47,$E4,$8C,$5F,$F9,$E6,$3F,$C6,$D5,$FB
	.dc.b	$8C,$3D,$BF,$11,$FE,$0E,$F3,$BD,$FF,$11,$70,$F6,$36,$FD,$3F,$11
	.dc.b	$73,$EE,$DA,$ED,$F1,$16,$BF,$D0,$B9,$FF,$B4,$AF,$F9,$0C,$7F,$7E
	.dc.b	$F4,$3E,$6F,$C4,$7F,$87,$9B,$87,$F8,$99,$FA,$3F,$11,$98,$F9,$CD
	.dc.b	$AF,$97,$F1,$17,$0F,$82,$FF,$8A,$3D,$3F,$B3,$F1,$17,$0F,$AC,$DB
	.dc.b	$FA,$7F,$11,$70,$F9,$2D,$AF,$91,$F1,$16,$BD,$6F,$F2,$E5,$FA,$97
	.dc.b	$F2,$3B,$FA,$3F,$13,$98,$C5,$FD,$18,$C7,$B1,$F8,$CF,$89,$B3,$F8
	.dc.b	$96,$E1,$F8,$6F,$89,$B8,$FD,$C7,$C1,$ED,$BD,$3F,$CA,$FC,$46,$1B
	.dc.b	$8F,$F2,$9B,$7C,$2F,$F3,$FF,$8D,$FF,$25,$46,$27,$F9,$74,$7F,$93
	.dc.b	$4F,$F1,$67,$B8,$7F,$C6,$46,$07,$C3,$7B,$7D,$EF,$D2,$FC,$4D,$BF
	.dc.b	$D1,$37,$66,$3F,$80,$E2,$FE,$70,$DB,$F8,$DE,$FE,$E3,$F3,$3F,$11
	.dc.b	$71,$FC,$C3,$A7,$FA,$4E,$0F,$C1,$FF,$A4,$BF,$4A,$FE,$A4,$63,$1E
	.dc.b	$F5,$FD,$D1,$7C,$0F,$3B,$F2,$CE,$63,$18,$FE,$AB,$7B,$3F,$A9,$7E
	.dc.b	$E4,$FE,$07,$F6,$77,$BD,$8A,$F2,$C6,$1F,$3E,$3E,$07,$D3,$95,$CB
	.dc.b	$C2,$7F,$96,$A7,$19,$7E,$9A,$6D,$79,$4F,$EC,$9F,$3F,$FE,$67,$AC
	.dc.b	$DF,$FB,$CA,$31,$F0,$5E,$5F,$87,$FC,$93,$6F,$C6,$FE,$57,$E0,$F3
	.dc.b	$1E,$66,$DF,$93,$95,$87,$F9,$AF,$1F,$EC,$F9,$FF,$86,$A7,$3F,$1B
	.dc.b	$CF,$1F,$7F,$3D,$C7,$DA,$7E,$99,$C5,$FE,$0F,$E0,$3F,$C4,$7F,$B1
	.dc.b	$FD,$4A,$F8,$5A,$7E,$A6,$73,$C5,$F6,$63,$F5,$9F,$0B,$CB,$2F,$B8
	.dc.b	$FE,$BF,$2F,$A9,$F0,$96,$7E,$9B,$BC,$7E,$76,$3D,$1E,$F3,$F5,$A3
	.dc.b	$11,$FB,$CC,$3B,$CF,$C5,$3E,$DE,$18,$9F,$8D,$BF,$E9,$27,$27,$D6
	.dc.b	$EA,$F0,$4F,$F0,$6C,$3A,$F0,$7F,$90,$23,$C7,$5F,$E9,$AD,$7E,$14
	.dc.b	$7F,$0C,$C7,$FC,$6D,$8F,$FA,$4A,$3E,$5D,$FF,$82,$EF,$F0,$7F,$1B
	.dc.b	$FC,$3F,$E1,$B7,$F1,$B6,$7F,$0E,$FC,$B7,$B7,$07,$E8,$F4,$F1,$ED
	.dc.b	$9C,$9A,$F8,$B3,$FF,$0A,$C7,$82,$BA,$37,$0F,$CE,$FC,$3F,$1E,$2F
	.dc.b	$DF,$36,$3B,$1E,$09,$F7,$9D,$B2,$FB,$91,$F4,$4D,$47,$C7,$70,$FC
	.dc.b	$0D,$3A,$3C,$AF,$FA,$5F,$43,$E1,$B3,$28,$C5,$F2,$D3,$D9,$F0,$FF
	.dc.b	$4B,$4F,$A6,$D9,$FD,$6F,$C6,$76,$E8,$FD,$0B,$FC,$EE,$7E,$9F,$55
	.dc.b	$AF,$D9,$B7,$2F,$61,$FB,$3F,$8C,$EB,$3F,$07,$0F,$D5,$C6,$27,$F1
	.dc.b	$3F,$19,$C7,$6F,$D6,$73,$23,$FA,$57,$C5,$7E,$0B,$E3,$2E,$5F,$C0
	.dc.b	$9B,$87,$80,$FE,$A7,$F1,$7F,$8E,$AF,$C7,$B6,$BF,$84,$FC,$67,$0C
	.dc.b	$BB,$BF,$D5,$E7,$F1,$FE,$15,$AF,$E1,$ED,$C3,$B7,$43,$E3,$3A,$D7
	.dc.b	$D8,$8C,$36,$FF,$33,$DF,$C7,$2F,$6B,$FD,$96,$FF,$3B,$CF,$6F,$F8
	.dc.b	$EB,$71,$F9,$7C,$3F,$C4,$BE,$33,$F5,$37,$FB,$98,$C5,$F8,$63,$F4
	.dc.b	$FF,$13,$C7,$2F,$96,$FF,$73,$F8,$CF,$E8,$BE,$2D,$BF,$E6,$6D,$CB
	.dc.b	$DB,$F1,$9F,$B6,$AF,$DB,$B6,$BF,$98,$FC,$46,$F1,$FA,$EF,$F7,$F9
	.dc.b	$FF,$00,$7C,$65,$AF,$E8,$2D,$C3,$EB,$7E,$A8,$F1,$5F,$15,$F2,$70
	.dc.b	$FE,$23,$39,$8C,$61,$FD,$2B,$E3,$79,$67,$FD,$AD,$EF,$1F,$D9,$3F
	.dc.b	$68,$FE,$6A,$FF,$5F,$CF,$FC,$2D,$18,$CF,$FA,$76,$5D,$B3,$F0,$C7
	.dc.b	$C0,$BF,$D8,$E7,$E1,$8C,$1F,$C4,$31,$F0,$CF,$F6,$B7,$FC,$3C,$7F
	.dc.b	$16,$F8,$8B,$1F,$D5,$9F,$81,$E1,$FF,$1C,$BE,$8F,$F0,$DC,$7C,$5A
	.dc.b	$FE,$27,$0F,$E3,$FF,$0F,$CB,$2F,$E5,$AF,$F8,$83,$CB,$FE,$47,$F0
	.dc.b	$57,$4F,$EC,$6F,$F8,$CF,$3E,$BE,$4F,$FB,$C9,$B8,$7C,$0A,$F6,$ED
	.dc.b	$F0,$35,$FE,$A1,$CF,$FE,$E4,$F2,$BF,$DC,$9E,$D9,$BE,$4E,$1F,$F1
	.dc.b	$8F,$97,$FC,$D3,$E0,$2E,$5F,$DD,$1F,$16,$5D,$75,$F1,$37,$4F,$11
	.dc.b	$F8,$27,$19,$FF,$BD,$3C,$4F,$0F,$83,$E2,$6E,$3F,$E2,$0F,$6B,$C8
	.dc.b	$FF,$B9,$7F,$31,$FF,$26,$E2,$13,$A7,$FB,$D4,$69,$70,$FC,$46,$E5
	.dc.b	$E5,$7F,$CB,$7C,$FF,$51,$FD,$8F,$0B,$B7,$E2,$9F,$1A,$6E,$5E,$5E
	.dc.b	$23,$E5,$F6,$3F,$CA,$F1,$8C,$3C,$71,$F1,$BA,$7E,$99,$F9,$E8,$F8
	.dc.b	$8F,$F2,$7C,$5D,$3F,$50,$FA,$BF,$1F,$71,$F8,$9E,$77,$5E,$FF,$F0
	.dc.b	$A7,$07,$F9,$03,$2F,$89,$C2,$AD,$5F,$C7,$E4,$5E,$EF,$E1,$E9,$E4
	.dc.b	$7F,$43,$D3,$F9,$9E,$2F,$AF,$FB,$3F,$CF,$C7,$E8,$69,$FE,$23,$8C
	.dc.b	$5F,$F8,$A5,$FD,$EC,$E8,$F0,$3A,$9F,$F3,$26,$BE,$4A,$67,$E4,$E7
	.dc.b	$FF,$45,$4E,$63,$18,$7F,$98,$B9,$7E,$D1,$F6,$FC,$87,$2F,$D9,$D7
	.dc.b	$FB,$D6,$1F,$E9,$3F,$90,$E7,$87,$F9,$51,$C3,$FE,$54,$F4,$EE,$9F
	.dc.b	$E3,$8F,$63,$FC,$2C,$FC,$43,$EA,$9F,$13,$5F,$CE,$F0,$FF,$41,$C3
	.dc.b	$FD,$49,$F1,$DE,$DC,$7F,$CE,$11,$8B,$91,$FC,$5F,$C9,$B1,$FE,$46
	.dc.b	$F1,$7F,$A4,$BF,$E4,$8E,$0F,$6F,$27,$E5,$78,$7D,$B8,$7D,$5E,$0F
	.dc.b	$F1,$3E,$7D,$F0,$FD,$16,$DF,$E2,$57,$FC,$A3,$19,$7F,$D1,$1E,$1F
	.dc.b	$9B,$6F,$E4,$BE,$0D,$C7,$F9,$53,$F9,$5A,$F7,$E3,$FC,$5E,$7C,$1C
	.dc.b	$1F,$AF,$96,$3F,$79,$EF,$F1,$FC,$DB,$5F,$E9,$89,$FF,$80,$11,$FB
	.dc.b	$77,$8C,$66,$7E,$4C,$78,$6F,$F2,$18,$FE,$95,$B7,$FD,$FE,$79,$BE
	.dc.b	$2F,$8E,$FF,$2C,$C6,$0F,$F7,$FC,$FC,$67,$FA,$1F,$81,$DB,$1F,$E9
	.dc.b	$0F,$7C,$BE,$DE,$1F,$74,$FE,$19,$C1,$FC,$1E,$FA,$63,$DB,$3F,$BE
	.dc.b	$F8,$3E,$3E,$E5,$EF,$6E,$1F,$6F,$87,$F8,$76,$3F,$97,$E7,$FC,$B6
	.dc.b	$7F,$2F,$3F,$3B,$72,$ED,$97,$C9,$7D,$A7,$F5,$CA,$EB,$D3,$FC,$37
	.dc.b	$27,$27,$E1,$9F,$07,$07,$D8,$7F,$98,$F0,$7C,$A3,$F1,$38,$7C,$8F
	.dc.b	$8F,$EB,$97,$F9,$4B,$44,$66,$3F,$98,$79,$9F,$9E,$8F,$DA,$76,$FF
	.dc.b	$6D,$C7,$F4,$D8,$F5,$DB,$FD,$A1,$B7,$BA,$7F,$01,$7F,$E1,$9B,$75
	.dc.b	$75,$9F,$F0,$FC,$63,$2F,$17,$C7,$F6,$F7,$7F,$5A,$D8,$FE,$05,$C7
	.dc.b	$F3,$0F,$D3,$65,$DF,$0F,$CA,$D7,$E1,$3E,$3F,$F0,$AD,$BF,$74,$7A
	.dc.b	$BA,$7F,$72,$D8,$EB,$C5,$F2,$6F,$FA,$2C,$5E,$15,$7E,$43,$E3,$BA
	.dc.b	$57,$E5,$5C,$57,$F5,$7E,$AF,$E0,$B1,$F2,$1F,$C8,$4F,$E3,$B4,$F4
	.dc.b	$D7,$5A,$F8,$DE,$5F,$2A,$F8,$AF,$FA,$DB,$C4,$79,$3F,$4B,$C9,$FE
	.dc.b	$55,$5F,$D8,$A3,$18,$7A,$A3,$F7,$7C,$BF,$52,$3F,$09,$E2,$3C,$57
	.dc.b	$C6,$7F,$93,$5B,$3F,$E7,$95,$F1,$99,$9F,$D7,$FC,$77,$83,$C4,$FF
	.dc.b	$53,$F8,$1F,$93,$E4,$4E,$78,$FD,$FA,$FE,$1B,$5F,$13,$E4,$3F,$82
	.dc.b	$4F,$EE,$BE,$3B,$AE,$BF,$93,$70,$FF,$0B,$CF,$F9,$7C,$7F,$30,$7A
	.dc.b	$F8,$5F,$AD,$9F,$AD,$3F,$83,$C8,$DA,$F8,$FF,$B4,$BA,$B9,$5F,$F5
	.dc.b	$7C,$BF,$C1,$93,$98,$F9,$5E,$07,$E7,$67,$F5,$7A,$7F,$08,$8F,$E6
	.dc.b	$78,$F7,$3F,$25,$CA,$F3,$FF,$4F,$E6,$7D,$68,$C6,$3F,$E1,$B8,$F4
	.dc.b	$F5,$7E,$0B,$2F,$88,$7C,$73,$F0,$DD,$9F,$0E,$BF,$29,$7F,$D2,$38
	.dc.b	$FF,$C0,$3D,$1D,$7C,$AF,$DA,$BD,$2F,$F9,$56,$DE,$72,$63,$D0,$FE
	.dc.b	$59,$AF,$A3,$C9,$FD,$EE,$DE,$76,$E9,$F0,$AF,$8D,$FE,$90,$F3,$F2
	.dc.b	$7F,$58,$C7,$CD,$1E,$AD,$7F,$8A,$6B,$FE,$16,$D7,$E1,$78,$1F,$C2
	.dc.b	$9F,$35,$7F,$31,$3E,$01,$30,$FE,$65,$AF,$CC,$F5,$3F,$88,$EB,$F2
	.dc.b	$DB,$AF,$96,$FD,$9C,$5F,$D6,$5F,$2F,$0F,$F0,$7D,$BC,$B3,$F6,$E7
	.dc.b	$FC,$93,$E7,$7F,$1C,$D7,$EB,$37,$0E,$93,$E4,$F3,$3F,$97,$5F,$1D
	.dc.b	$7E,$04,$FC,$AA,$FA,$BE,$77,$F2,$6D,$3F,$00,$D8,$E7,$5F,$97,$AF
	.dc.b	$2E,$7F,$98,$9F,$BB,$E7,$7F,$2E,$CF,$F1,$0D,$8E,$3A,$FE,$1F,$3F
	.dc.b	$13,$ED,$C3,$C4,$7E,$2B,$F3,$0F,$F0,$37,$83,$FE,$40,$D3,$F8,$3C
	.dc.b	$63,$C8,$47,$D2,$C3,$E4,$1D,$34,$F2,$61,$CD,$FE,$83,$9F,$26,$E7
	.dc.b	$E4,$7F,$1F,$3F,$4B,$9B,$F0,$FB,$7D,$AA,$FE,$59,$9F,$F8,$13,$4F
	.dc.b	$F8,$01,$7E,$4F,$3F,$FC,$81,$EF,$7F,$86,$70,$F9,$BA,$79,$32,$E7
	.dc.b	$BF,$BE,$3F,$63,$9F,$4E,$FE,$3A,$FB,$9B,$1E,$5C,$FF,$1D,$8F,$4C
	.dc.b	$7F,$37,$5F,$C1,$AB,$CB,$E8,$78,$B4,$F8,$ED,$8E,$59,$78,$F1,$F7
	.dc.b	$10,$FC,$CD,$F3,$7E,$97,$9F,$EC,$8F,$A2,$DA,$E7,$8F,$F2,$4B,$FA
	.dc.b	$17,$FC,$0A,$F9,$47,$E5,$F7,$FE,$EB,$7F,$C5,$F7,$6E,$CF,$EE,$F9
	.dc.b	$DF,$29,$FB,$99,$F5,$3E,$4E,$FF,$73,$2E,$9E,$0F,$D0,$DB,$F0,$AD
	.dc.b	$CF,$CB,$B7,$F2,$03,$DD,$C1,$F2,$5F,$F1,$E7,$57,$F7,$1D,$FF,$B2
	.dc.b	$D7,$97,$6F,$3D,$3B,$3F,$C3,$15,$A8,$C5,$7F,$A5,$6B,$75,$FE,$1D
	.dc.b	$C3,$B5,$7F,$0F,$CF,$E2,$9F,$B9,$E1,$F8,$BB,$7F,$14,$C9,$E5,$7E
	.dc.b	$2B,$5F,$CF,$B7,$4F,$5E,$3C,$EF,$E1,$BF,$3F,$5F,$57,$89,$F9,$2F
	.dc.b	$2F,$F4,$ED,$BE,$3D,$9F,$C0,$B0,$E3,$AF,$8F,$0F,$84,$7E,$6B,$2F
	.dc.b	$D7,$B6,$F8,$F6,$FF,$85,$F7,$F6,$EF,$EA,$D7,$F1,$4F,$F7,$0A,$FF
	.dc.b	$35,$7C,$87,$FA,$3B,$C9,$FF,$47,$BF,$72,$BE,$A1,$FC,$9F,$AB,$F4
	.dc.b	$9A,$FF,$93,$73,$FE,$C5,$5F,$E9,$6F,$90,$E3,$8F,$F9,$D5,$E5,$87
	.dc.b	$C3,$3E,$EE,$1C,$9F,$BB,$3F,$E2,$19,$FF,$00,$78,$BC,$E3,$FB,$44
	.dc.b	$BC,$AF,$F0,$AC,$7B,$8F,$F1,$56,$BE,$EC,$BF,$BA,$7C,$87,$FA,$A3
	.dc.b	$E4,$3F,$D5,$2D,$AE,$18,$7F,$24,$3E,$A3,$FE,$31,$9F,$F5,$7F,$C8
	.dc.b	$7F,$A0,$1B,$1D,$B4,$FE,$49,$B7,$D4,$3F,$7B,$8F,$F8,$0E,$FF,$D0
	.dc.b	$FF,$1F,$FE,$88,$6D,$F1,$ED,$FE,$33,$97,$CB,$7F,$4D,$87,$F0,$6B
	.dc.b	$FF,$7E,$72,$A3,$FD,$49,$E2,$BE,$07,$9E,$0A,$D2,$3F,$DF,$B1,$F8
	.dc.b	$1D,$3C,$F1,$F4,$31,$EE,$FE,$0B,$E3,$FF,$D3,$BF,$1D,$FE,$9E,$6E
	.dc.b	$9E,$79,$FE,$37,$7F,$5D,$FE,$23,$B7,$F8,$86,$FC,$FA,$7F,$AE,$78
	.dc.b	$58,$72,$BF,$34,$E6,$3C,$99,$7C,$2A,$FD,$89,$FC,$E6,$BF,$BF,$CF
	.dc.b	$47,$D9,$AB,$E0,$F1,$0A,$EF,$5F,$07,$6F,$87,$2B,$E1,$3E,$B9,$FE
	.dc.b	$59,$3F,$44,$FC,$09,$FE,$0E,$F8,$EE,$B6,$F8,$EB,$18,$7D,$33,$F1
	.dc.b	$32,$E9,$9F,$E8,$F1,$FC,$9D,$7E,$C4,$DE,$3F,$E0,$3E,$DF,$DE,$61
	.dc.b	$F4,$CF,$0C,$7F,$34,$7F,$DC,$47,$B7,$76,$38,$8F,$F1,$61,$AC,$D3
	.dc.b	$A8,$D4,$62,$78,$47,$36,$C6,$63,$73,$C5,$E8,$D7,$FD,$B2,$62,$78
	.dc.b	$C6,$63,$FD,$B7,$F1,$CF,$7B,$AB,$63,$C6,$D7,$D0,$D7,$F1,$0D,$79
	.dc.b	$C7,$22,$3C,$9A,$F2,$6B,$F9,$1F,$8D,$FD,$13,$F9,$03,$99,$4F,$C4
	.dc.b	$4F,$97,$1F,$C3,$E2,$AE,$31,$89,$D6,$BE,$6C,$3F,$1C,$DC,$3A,$7E
	.dc.b	$AF,$FE,$73,$72,$78,$2F,$95,$62,$3C,$2D,$C5,$F0,$9E,$0C,$3F,$55
	.dc.b	$F1,$D6,$FF,$79,$5E,$57,$86,$79,$7F,$45,$36,$59,$FA,$27,$C5,$7E
	.dc.b	$7C,$BF,$4B,$9F,$9E,$7E,$25,$F0,$CF,$C9,$87,$19,$F3,$9A,$CF,$B7
	.dc.b	$C7,$78,$7F,$6A,$FB,$99,$7E,$C1,$B9,$FF,$03,$9F,$26,$1C,$F2,$ED
	.dc.b	$8F,$82,$3B,$67,$E3,$C7,$E3,$CE,$73,$F1,$9F,$1F,$3F,$77,$97,$E2
	.dc.b	$9F,$95,$96,$1D,$9C,$49,$7F,$03,$D8,$F1,$E1,$A8,$EE,$7E,$0B,$0F
	.dc.b	$95,$F1,$DC,$EF,$E2,$E3,$87,$86,$BF,$2C,$8F,$F3,$1C,$FD,$51,$88
	.dc.b	$F9,$38,$76,$8F,$E4,$D8,$7D,$7F,$8E,$F9,$5F,$03,$2F,$F0,$91,$E6
	.dc.b	$7D,$35,$CF,$0E,$B1,$F7,$08,$76,$8E,$AF,$E3,$7E,$3B,$B3,$D0,$C1
	.dc.b	$EA,$38,$F0,$7D,$4C,$BE,$34,$64,$E8,$6B,$2F,$C7,$77,$FE,$BB,$D0
	.dc.b	$FF,$15,$61,$F9,$13,$35,$FD,$E3,$0E,$98,$FC,$36,$C7,$C7,$E0,$F7
	.dc.b	$65,$F1,$63,$B1,$FA,$5E,$DE,$DD,$7F,$D3,$73,$B3,$FA,$46,$BF,$D0
	.dc.b	$E6,$E1,$DA,$3D,$24,$D7,$E4,$B8,$8D,$13,$5F,$D5,$7C,$77,$4F,$1B
	.dc.b	$FD,$33,$5F,$E1,$26,$BF,$E4,$C3,$F6,$B1,$E0,$E6,$FC,$F1,$39,$7F
	.dc.b	$80,$63,$E0,$C7,$BC,$E6,$FF,$58,$7F,$03,$F8,$EF,$4E,$12,$3E,$54
	.dc.b	$75,$C3,$EA,$5F,$B7,$0E,$98,$73,$DB,$E5,$61,$FD,$97,$3F,$E0,$9F
	.dc.b	$1D,$F1,$32,$93,$F5,$63,$A3,$F8,$E3,$F1,$78,$74,$CB,$BE,$3F,$B9
	.dc.b	$C3,$F8,$87,$A6,$F8,$2F,$A7,$EC,$7F,$A1,$BD,$3B,$D4,$7E,$AD,$6F
	.dc.b	$3F,$C2,$63,$E7,$E6,$FC,$E7,$37,$E0,$9F,$C3,$47,$79,$FE,$51,$4F
	.dc.b	$43,$AF,$91,$27,$FC,$8B,$39,$7A,$ED,$E8,$3E,$8E,$3D,$63,$F3,$F8
	.dc.b	$75,$7F,$A2,$53,$D0,$F4,$ED,$27,$AC,$7C,$6E,$3F,$9B,$C9,$F4,$71
	.dc.b	$E9,$5F,$8B,$C7,$A5,$7F,$84,$B1,$7A,$1F,$13,$59,$7E,$88,$E9,$B7
	.dc.b	$C0,$D7,$CF,$3E,$93,$EA,$1F,$6C,$F8,$7E,$7F,$CE,$D6,$47,$C2,$8E
	.dc.b	$6F,$BA,$79,$BC,$B5,$FC,$8C,$7F,$2B,$8F,$9B,$E7,$FD,$8C,$64,$7C
	.dc.b	$C8,$E6,$FB,$67,$9E,$9F,$0B,$5F,$B5,$3F,$D5,$B9,$3F,$0B,$E7,$FE
	.dc.b	$0B,$39,$3F,$5A,$39,$E7,$E2,$8F,$ED,$3C,$1F,$2A,$31,$8F,$F3,$39
	.dc.b	$FC,$0F,$9F,$F8,$CD,$24,$7E,$02,$39,$E7,$E2,$87,$3F,$E6,$5F,$81
	.dc.b	$1F,$53,$83,$FC,$0B,$C4,$F3,$FF,$2F,$AC,$BF,$C5,$C7,$4E,$0F,$E4
	.dc.b	$58,$F3,$D3,$E0,$47,$C9,$AF,$C4,$F9,$FF,$A0,$D2,$47,$E5,$23,$93
	.dc.b	$F6,$AF,$96,$7F,$7E,$3F,$80,$9F,$E1,$5F,$13,$FC,$D2,$FF,$B7,$72
	.dc.b	$FD,$59,$FC,$7A,$33,$1F,$C4,$4F,$A8,$75,$E0,$F9,$5B,$7C,$3E,$0F
	.dc.b	$F0,$6C,$FF,$84,$69,$E7,$7E,$DB,$39,$3F,$EA,$E9,$CB,$FA,$A3,$E4
	.dc.b	$EB,$F8,$0C,$FD,$1B,$7E,$8B,$8F,$FC,$25,$7F,$E6,$7B,$FF,$49,$78
	.dc.b	$9E,$AE,$C7,$E6,$BE,$88,$C3,$7C,$4F,$F1,$BD,$F1,$BF,$66,$DF,$82
	.dc.b	$CF,$F8,$2C,$7E,$8F,$3F,$03,$EB,$F3,$3F,$88,$E3,$E8,$8F,$E6,$99
	.dc.b	$F8,$37,$F9,$15,$F4,$AF,$F9,$2F,$1F,$D2,$BF,$D3,$1F,$4A,$3F,$D1
	.dc.b	$39,$FF,$8A,$A3,$D7,$A7,$FA,$B7,$1F,$5F,$9F,$E0,$3E,$A7,$17,$5D
	.dc.b	$7C,$F7,$F2,$0F,$D8,$70,$7C,$6F,$37,$CF,$AC,$9F,$A1,$1C,$5F,$E6
	.dc.b	$3C,$9F,$6B,$8F,$F9,$FD,$FC,$EF,$37,$DF,$94,$BF,$B1,$1C,$5F,$EB
	.dc.b	$D7,$C5,$F0,$9F,$81,$F3,$3E,$3E,$32,$3F,$01,$1C,$79,$3F,$8E,$47
	.dc.b	$CF,$D3,$FB,$C6,$5F,$7B,$C3,$F9,$F9,$4A,$FC,$44,$71,$CF,$FC,$77
	.dc.b	$9F,$87,$8B,$EB,$D7,$E7,$7C,$CF,$AF,$A4,$8F,$C8,$47,$1D,$BF,$1D
	.dc.b	$1F,$52,$B8,$9F,$E7,$6F,$8F,$FE,$FF,$E8,$FD,$FE,$9F,$D1,$B7,$97
	.dc.b	$FD,$A5,$3C,$EF,$F8,$69,$FB,$7B,$FD,$CB,$F7,$F8,$3F,$C4,$EF,$E4
	.dc.b	$27,$FC,$12,$7F,$82,$B8,$39,$57,$F8,$92,$78,$1E,$5A,$E3,$1F,$E8
	.dc.b	$AF,$8F,$FF,$09,$FE,$24,$F7,$69,$FE,$14,$8E,$75,$FE,$96,$EA,$FF
	.dc.b	$02,$47,$F9,$E7,$19,$8F,$FA,$16,$3F,$5B,$E8,$FE,$7B,$92,$4F,$ED
	.dc.b	$E3,$9B,$FE,$2B,$3F,$D2,$3D,$4F,$B4,$7B,$1F,$E8,$B8,$C3,$63,$33
	.dc.b	$C1,$7C,$17,$FA,$2F,$E4,$7F,$5B,$CD,$2B,$FC,$79,$1E,$07,$F9,$7F
	.dc.b	$46,$DF,$0F,$82,$CF,$C0,$D8,$FF,$50,$57,$FA,$67,$7F,$F6,$27,$3F
	.dc.b	$F0,$4F,$57,$F7,$67,$F8,$F2,$33,$CF,$FC,$43,$8F,$F5,$BC,$DF,$E9
	.dc.b	$AE,$9F,$06,$5F,$E9,$AC,$BA,$47,$5F,$91,$FE,$0B,$EB,$4A,$FD,$D9
	.dc.b	$FE,$77,$8F,$56,$5F,$E4,$40,$F2,$8F,$F1,$CD,$73,$7F,$C7,$7F,$29
	.dc.b	$FC,$43,$D7,$91,$D0,$FE,$A3,$1D,$38,$7D,$CF,$C2,$F5,$A4,$F2,$3F
	.dc.b	$C8,$BA,$FF,$96,$BE,$29,$F0,$74,$F8,$51,$CE,$76,$71,$3F,$DA,$37
	.dc.b	$F2,$E3,$E6,$7C,$8F,$09,$F2,$6A,$9F,$99,$3C,$E7,$FD,$B3,$1C,$5E
	.dc.b	$A7,$FB,$73,$E4,$3E,$97,$C8,$FE,$C3,$CA,$93,$D2,$39,$71,$FF,$B3
	.dc.b	$2B,$C1,$3D,$0F,$F3,$0B,$F6,$CE,$6F,$13,$CA,$F2,$BF,$D2,$3C,$A7
	.dc.b	$FD,$AF,$1F,$87,$8F,$C2,$7C,$8F,$FA,$03,$F6,$BF,$F2,$65,$FF,$63
	.dc.b	$6F,$47,$F6,$6D,$15,$D6,$BF,$D6,$CF,$8D,$F0,$9F,$19,$FC,$87,$C8
	.dc.b	$75,$F6,$64,$FC,$A8,$E4,$FF,$83,$9F,$DF,$FC,$24,$8D,$1F,$43,$D0
	.dc.b	$FE,$07,$F1,$12,$30,$7F,$6D,$DB,$F8,$99,$FC,$34,$FF,$7A,$CF,$8A
	.dc.b	$39,$9F,$DF,$F2,$47,$E9,$7E,$4B,$FB,$0F,$C0,$CA,$FB,$F1,$E6,$A3
	.dc.b	$FA,$BA,$93,$A3,$FC,$01,$DF,$FD,$AB,$E0,$64,$60,$F3,$71,$7E,$C5
	.dc.b	$FE,$1C,$32,$FC,$62,$F1,$DE,$0B,$E2,$7F,$D0,$AD,$9F,$D8,$CE,$A7
	.dc.b	$3F,$13,$FE,$1D,$F0,$BF,$BD,$F9,$F2,$FF,$C2,$91,$F6,$6F,$AF,$C1
	.dc.b	$F4,$24,$F4,$C5,$F0,$7F,$E2,$7F,$0F,$FC,$0B,$E7,$CA,$FF,$0F,$47
	.dc.b	$3A,$FF,$44,$47,$F8,$21,$BD,$5F,$E1,$96,$9F,$E0,$13,$FD,$0D,$1F
	.dc.b	$E2,$7F,$0F,$FC,$23,$E7,$C8,$C1,$F2,$0F,$E8,$C7,$C8,$E0,$F8,$FA
	.dc.b	$7C,$6F,$80,$F8,$BF,$01,$2B,$D5,$1F,$E3,$FE,$9F,$F0,$EF,$95,$27
	.dc.b	$FC,$6D,$1E,$9F,$7B,$E6,$92,$3E,$17,$BD,$FC,$A7,$E4,$7F,$D6,$BF
	.dc.b	$9B,$7F,$2A,$9F,$AC,$D7,$FC,$CF,$1F,$CB,$FC,$CF,$EF,$1D,$5F,$D5
	.dc.b	$5F,$E6,$D1,$CB,$CE,$95,$FC,$32,$3F,$9B,$7C,$57,$FB,$CB,$E5,$0D
	.dc.b	$33,$3F,$C4,$AF,$AB,$F0,$F0,$EC,$F3,$21,$F9,$F8,$F2,$3F,$C6,$A7
	.dc.b	$9B,$93,$A1,$FE,$17,$3A,$1F,$96,$E5,$FE,$05,$87,$C3,$8D,$9F,$AB
	.dc.b	$F9,$2F,$EE,$FE,$7C,$8F,$E4,$F1,$FD,$44,$FE,$AD,$84,$8F,$F7,$DC
	.dc.b	$7F,$59,$3E,$6F,$9F,$23,$89,$FC,$C3,$E4,$BE,$8E,$72,$30,$43,$F5
	.dc.b	$4F,$F9,$37,$67,$A1,$FD,$DD,$B2,$36,$5F,$1D,$7F,$DA,$CF,$BF,$E7
	.dc.b	$C9,$ED,$B7,$F2,$3F,$63,$F0,$7E,$84,$9F,$E8,$B1,$FD,$BC,$E4,$F7
	.dc.b	$EB,$FD,$EB,$FE,$AC,$C3,$FB,$C4,$C8,$FD,$DE,$5F,$CB,$9B,$E7,$F3
	.dc.b	$7F,$A8,$FA,$FF,$92,$F3,$E4,$60,$87,$B0,$FF,$5C,$DF,$57,$1A,$FF
	.dc.b	$85,$1F,$26,$FF,$E2,$9B,$FC,$EF,$99,$27,$9E,$DF,$D0,$7D,$CF,$D0
	.dc.b	$F9,$D2,$7F,$B1,$C7,$F8,$34,$F9,$DC,$F2,$3A,$9F,$36,$3F,$C2,$C7
	.dc.b	$EA,$BA,$24,$72,$3F,$B6,$FC,$8F,$E8,$73,$91,$82,$1E,$AA,$E6,$F6
	.dc.b	$3F,$DA,$F3,$FE,$23,$3F,$6F,$D1,$23,$8E,$5F,$D6,$FE,$47,$FC,$49
	.dc.b	$52,$7F,$BE,$C5,$7E,$3C,$73,$7B,$7E,$B2,$F4,$F5,$7F,$D9,$6A,$F9
	.dc.b	$1F,$8D,$B5,$D7,$FD,$9C,$DA,$FF,$5E,$72,$BB,$3F,$86,$9C,$63,$FD
	.dc.b	$AB,$3B,$BF,$F2,$F6,$1E,$3C,$3C,$A7,$93,$1E,$13,$E1,$AE,$13,$F9
	.dc.b	$DA,$F2,$BF,$A1,$8C,$BE,$6F,$8F,$FE,$3B,$D1,$23,$FC,$65,$29,$EB
	.dc.b	$1C,$70,$F8,$DD,$3F,$9C,$E6,$63,$F0,$63,$F3,$F8,$7C,$7B,$E7,$87
	.dc.b	$F2,$AE,$79,$18,$3B,$FB,$FF,$C8,$78,$64,$60,$F1,$E7,$E0,$E3,$F8
	.dc.b	$71,$F0,$4C,$9C,$CF,$9F,$87,$A1,$FA,$73,$71,$F2,$47,$D4,$C3,$06
	.dc.b	$66,$47,$E4,$32,$F7,$67,$ED,$C7,$E7,$56,$AB,$B9,$E1,$DF,$EC,$C7
	.dc.b	$7F,$92,$FF,$17,$7C,$64,$AF,$9B,$19,$7D,$27,$E1,$A7,$0F,$E2,$23
	.dc.b	$FA,$8F,$54,$8E,$27,$D6,$F5,$FF,$C7,$5E,$4C,$8C,$10,$63,$FE,$0C
	.dc.b	$7F,$91,$7C,$17,$F8,$2D,$CF,$2F,$E6,$63,$13,$F9,$2B,$F9,$07,$F6
	.dc.b	$2E,$99,$7D,$67,$F2,$FE,$27,$FA,$E3,$E4,$A4,$FE,$2A,$3E,$81,$C9
	.dc.b	$E1,$8F,$E5,$E3,$31,$FC,$8B,$83,$FB,$79,$87,$F5,$3E,$BF,$F7,$2E
	.dc.b	$C9,$1F,$96,$8F,$B8,$6F,$1F,$1C,$64,$87,$B8,$FF,$17,$7E,$13,$FA
	.dc.b	$6F,$3C,$FE,$93,$9F,$E0,$E7,$FB,$BF,$09,$EC,$FE,$8A,$F7,$D5,$70
	.dc.b	$FD,$36,$3F,$E9,$9B,$DB,$FE,$36,$E9,$78,$FC,$57,$8F,$4C,$46,$7C
	.dc.b	$7D,$78,$F2,$F9,$47,$E6,$7D,$9F,$07,$7C,$9F,$DB,$C6,$5F,$F2,$2E
	.dc.b	$7F,$9E,$E1,$FC,$06,$DE,$EE,$DD,$57,$F1,$4B,$F5,$76,$C8,$E6,$7F
	.dc.b	$02,$E8,$F7,$75,$48,$C1,$0F,$99,$87,$CA,$E9,$F8,$87,$EB,$EA,$4F
	.dc.b	$93,$9F,$E4,$18,$7F,$92,$FA,$FF,$2F,$29,$1F,$C4,$A3,$F7,$07,$CF
	.dc.b	$A9,$19,$21,$E2,$D5,$3F,$54,$E3,$1F,$C7,$3D,$7F,$B0,$47,$04,$39
	.dc.b	$7C,$87,$DD,$3F,$84,$44,$8F,$9B,$EE,$7E,$04,$C3,$FC,$FA,$7F,$09
	.dc.b	$F2,$12,$3F,$95,$C7,$F1,$13,$F1,$1E,$1C,$8C,$90,$F5,$78,$5F,$8D
	.dc.b	$38,$3F,$CD,$2F,$F2,$1E,$B4,$8C,$10,$F8,$5F,$1F,$F9,$53,$F9,$04
	.dc.b	$C8,$FC,$2F,$AB,$F9,$93,$0F,$F5,$6C,$BF,$39,$EA,$C8,$FE,$87,$1F
	.dc.b	$CA,$4F,$D0,$7C,$7C,$8C,$90,$FC,$1F,$07,$F0,$F6,$F8,$5C,$5F,$E9
	.dc.b	$BA,$FE,$AE,$5F,$1F,$4F,$E6,$D1,$27,$FD,$3F,$2F,$82,$E0,$66,$3F
	.dc.b	$B6,$CF,$3F,$8F,$91,$D8,$FE,$B7,$C1,$DF,$E3,$24,$60,$87,$A3,$E3
	.dc.b	$FC,$87,$F4,$79,$91,$E2,$F8,$EF,$41,$87,$FB,$EF,$1F,$AF,$E3,$A4
	.dc.b	$7F,$6A,$8F,$EA,$07,$BF,$E3,$A4,$64,$87,$C8,$F8,$DF,$88,$70,$7F
	.dc.b	$B9,$F3,$7C,$7F,$8D,$91,$82,$1F,$43,$E2,$7E,$61,$FD,$7E,$64,$7C
	.dc.b	$AF,$88,$FA,$06,$1F,$F0,$AF,$47,$D3,$F8,$79,$1F,$E0,$28,$FE,$D0
	.dc.b	$7D,$7F,$87,$91,$92,$1F,$80,$F8,$5F,$B8,$70,$7F,$C1,$7D,$7F,$7F
	.dc.b	$D1,$91,$82,$1F,$88,$F8,$9F,$C2,$1F,$DD,$E6,$47,$E0,$BE,$27,$F1
	.dc.b	$06,$1F,$E7,$87,$FB,$F2,$7F,$19,$F1,$32,$3F,$9A,$47,$F3,$9F,$51
	.dc.b	$CB,$49,$3F,$EE,$C8,$72,$EF,$92,$4E,$5E,$5C,$9F,$E0,$BE,$FF,$CD
	.dc.b	$7C,$4C,$AF,$F1,$4C,$65,$FF,$7B,$FE,$6F,$FE,$C9,$F2,$CF,$FB,$27
	.dc.b	$D7,$56,$EB,$F9,$5C,$BE,$2B,$FA,$83,$63,$A9,$FE,$9B,$F4,$7F,$5D
	.dc.b	$EF,$7E,$BD,$FE,$7F,$2F,$89,$E2,$7F,$88,$E7,$5E,$3F,$F9,$5D,$B6
	.dc.b	$CF,$CB,$1E,$03,$84,$C9,$F5,$FC,$4F,$90,$C3,$F4,$A7,$F7,$6F,$C0
	.dc.b	$F6,$3D,$27,$87,$C9,$91,$F1,$3D,$9F,$69,$87,$94,$7C,$0F,$62,$47
	.dc.b	$BF,$1F,$44,$FA,$FC,$AF,$8B,$EC,$CA,$D1,$FE,$BF,$CB,$F8,$43,$80
	.dc.b	$E1,$F9,$FE,$DF,$CE,$3A,$3B,$8D,$70,$7C,$CF,$89,$FA,$66,$A7,$E0
	.dc.b	$79,$9F,$5B,$E2,$64,$71,$8F,$A0,$7D,$AF,$89,$91,$92,$1F,$80,$F8
	.dc.b	$4F,$BE,$70,$7C,$5E,$7F,$E0,$BE,$22,$46,$08,$7E,$23,$E2,$3F,$0E
	.dc.b	$7D,$F9,$91,$F9,$2F,$87,$FC,$59,$87,$EC,$7A,$5F,$8E,$F8,$79,$1F
	.dc.b	$06,$3F,$10,$7E,$4B,$E1,$A4,$64,$87,$E6,$3E,$1B,$F2,$E7,$07,$E5
	.dc.b	$78,$DF,$9A,$F8,$69,$18,$21,$FE,$0A,$F9,$6F,$F1,$A1,$FE,$96,$E4
	.dc.b	$FD,$BC,$6B,$93,$FC,$37,$3F,$E3,$08,$E1,$E4,$7E,$77,$0F,$CF,$79
	.dc.b	$1F,$E3,$7A,$7C,$35,$33,$7F,$76,$75,$F0,$D2,$30,$43,$C9,$F0,$DC
	.dc.b	$CE,$2F,$E7,$FC,$BE,$DF,$0D,$23,$F1,$11,$FA,$F3,$C5,$F0,$D2,$32
	.dc.b	$43,$E0,$7C,$37,$9C,$E0,$FE,$47,$CE,$F5,$7C,$34,$8C,$10,$F9,$1F
	.dc.b	$01,$EF,$3F,$7D,$32,$3F,$47,$E5,$37,$FF,$12,$C6,$1A,$FE,$E6,$7F
	.dc.b	$83,$7B,$BF,$27,$E1,$24,$FE,$8E,$32,$F1,$F8,$4F,$9C,$61,$FE,$05
	.dc.b	$1F,$47,$E1,$24,$71,$3F,$59,$F0,$1F,$57,$E1,$24,$60,$87,$E0,$FE
	.dc.b	$13,$ED,$1F,$C6,$66,$47,$A7,$E1,$3E,$F1,$87,$F8,$97,$C1,$7E,$07
	.dc.b	$E1,$24,$7E,$EE,$3F,$91,$1F,$85,$F8,$49,$19,$21,$F9,$3F,$83,$FC
	.dc.b	$51,$C1,$FE,$09,$F0,$9F,$8D,$F8,$19,$18,$21,$F9,$BF,$73,$F2,$47
	.dc.b	$F3,$19,$91,$F5,$3E,$4B,$F9,$8F,$29,$FE,$67,$1F,$C5,$A7,$F9,$C7
	.dc.b	$C2,$FF,$A4,$67,$F9,$6C,$FF,$9C,$1B,$E1,$FF,$42,$9D,$CF,$5F,$57
	.dc.b	$81,$87,$F9,$B4,$73,$F4,$64,$7F,$1F,$8F,$E8,$DF,$0B,$DF,$CE,$93
	.dc.b	$FC,$8E,$32,$FA,$7C,$CF,$21,$87,$FA,$14,$79,$FC,$79,$1C,$4F,$E5
	.dc.b	$9F,$0B,$EB,$F1,$E4,$60,$87,$B7,$C4,$F8,$07,$F5,$99,$91,$F2,$7B
	.dc.b	$FE,$21,$87,$FA,$97,$C1,$FC,$7E,$E9,$1F,$CE,$E3,$FB,$11,$F2,$FA
	.dc.b	$E4,$64,$87,$CD,$E9,$FA,$07,$07,$FA,$27,$C2,$FD,$3E,$59,$18,$21
	.dc.b	$F5,$78,$7E,$C1,$FD,$C6,$64,$7E,$07,$7F,$B8,$61,$FB,$9A,$7D,$F2
	.dc.b	$3F,$D4,$FE,$17,$FB,$D1,$F8,$3A,$91,$92,$1F,$85,$8F,$C4,$1C,$1F
	.dc.b	$EC,$BF,$4A,$FA,$26,$E9,$CC,$61,$B1,$97,$FD,$D5,$3C,$77,$FF,$06
	.dc.b	$7C,$3F,$FB,$97,$DC,$FF,$0B,$57,$FB,$11,$BD,$3F,$E1,$69,$DC,$FF
	.dc.b	$14,$9F,$F0,$B3,$7E,$13,$2F,$6C,$9C,$5E,$1C,$1F,$01,$E7,$CF,$D5
	.dc.b	$39,$8C,$57,$0B,$F5,$F1,$E5,$F7,$7E,$B3,$FE,$33,$CF,$FD,$EB,$1E
	.dc.b	$3F,$84,$D2,$97,$C9,$61,$E1,$F0,$BE,$F3,$11,$D7,$4F,$99,$CF,$73
	.dc.b	$34,$C7,$3F,$3F,$88,$F9,$1A,$4C,$FF,$C2,$D9,$FC,$8D,$BE,$83,$63
	.dc.b	$FC,$6D,$2F,$83,$FA,$07,$38,$FB,$3D,$3E,$12,$3F,$73,$A7,$D5,$C9
	.dc.b	$F5,$FE,$06,$4F,$82,$32,$FD,$5F,$81,$FB,$86,$1F,$A9,$D9,$ED,$23
	.dc.b	$F8,$8E,$CF,$8D,$ED,$FE,$0F,$DF,$93,$EC,$8C,$BF,$81,$F7,$FF,$10
	.dc.b	$61,$FC,$07,$7F,$C9,$23,$F9,$4E,$DF,$A5,$F0,$FF,$6F,$C9,$7E,$6F
	.dc.b	$F9,$5A,$78,$1F,$86,$7F,$CB,$91,$98,$FF,$6A,$72,$AF,$F0,$9F,$3C
	.dc.b	$FC,$B1,$CB,$D7,$0F,$F0,$4E,$3F,$DA,$9A,$F6,$E8,$FD,$45,$F1,$AF
	.dc.b	$C1,$7E,$B1,$F8,$8A,$FD,$7B,$5F,$D4,$B7,$3F,$37,$72,$3E,$F4,$FE
	.dc.b	$47,$C1,$F1,$7B,$D2,$3F,$0B,$1D,$BA,$3B,$F2,$49,$EF,$1E,$2D,$E4
	.dc.b	$78,$9F,$2D,$7E,$6F,$C2,$F7,$7C,$74,$AF,$C8,$C7,$A8,$F5,$FC,$64
	.dc.b	$8F,$5C,$7B,$BE,$2E,$47,$B9,$F8,$3E,$07,$E9,$EB,$E5,$7B,$12,$BF
	.dc.b	$3B,$1F,$1B,$C1,$F8,$FF,$17,$27,$E3,$C7,$CA,$F8,$B9,$1F,$29,$F9
	.dc.b	$BE,$17,$EC,$F2,$FA,$DE,$B4,$AF,$D4,$C7,$D2,$F0,$FE,$9F,$C5,$C9
	.dc.b	$FA,$71,$F5,$BE,$2E,$47,$D6,$7E,$CF,$87,$FB,$FD,$FF,$1D,$F9,$CB
	.dc.b	$D6,$A6,$6F,$EE,$F8,$BF,$80,$F8,$C9,$18,$3F,$03,$F1,$9F,$84,$23
	.dc.b	$FC,$CB,$F5,$8F,$CA,$78,$5F,$BE,$E3,$EF,$EA,$CA,$FE,$07,$1F,$C3
	.dc.b	$6B,$FC,$C1,$E0,$FE,$21,$FC,$9B,$7E,$33,$5F,$1B,$BF,$8D,$98,$77
	.dc.b	$79,$7C,$6F,$EA,$7F,$65,$74,$52,$57,$F9,$22,$3F,$DA,$AF,$9F,$C4
	.dc.b	$F1,$11,$F5,$7C,$77,$97,$E3,$A4,$7B,$5F,$47,$C7,$7F,$29,$EF,$F9
	.dc.b	$7E,$9C,$AF,$E3,$F1,$ED,$F8,$EF,$77,$8B,$27,$E3,$47,$C1,$F8,$D9
	.dc.b	$1F,$25,$F8,$9F,$1B,$FC,$D7,$C4,$FE,$09,$F9,$DF,$F2,$7F,$50,$FE
	.dc.b	$B3,$D4,$4F,$CB,$F8,$EF,$99,$F1,$D3,$2F,$A9,$1F,$3F,$CB,$91,$F5
	.dc.b	$DF,$A5,$F1,$DF,$D0,$FC,$BE,$FE,$6C,$AF,$E7,$31,$F5,$FE,$3B,$EC
	.dc.b	$7C,$74,$9F,$C0,$47,$DB,$F8,$E9,$1F,$83,$7E,$F7,$5F,$F4,$FF,$37
	.dc.b	$D7,$CD,$2B,$FA,$2C,$7E,$0F,$E3,$BF,$A8,$74,$7F,$BE,$BD,$73,$7C
	.dc.b	$4F,$B3,$7F,$DE,$23,$FC,$8F,$E3,$FF,$69,$6C,$2B,$8F,$C8,$74,$F8
	.dc.b	$EF,$EC,$1E,$97,$CC,$F4,$A5,$FF,$56,$8F,$0F,$9F,$E2,$F8,$E9,$3F
	.dc.b	$E0,$EC,$BF,$50,$DF,$8E,$F3,$91,$D7,$C7,$7F,$6D,$C7,$F9,$E7,$C6
	.dc.b	$FE,$92,$5E,$07,$F8,$4D,$FE,$CB,$3F,$03,$E3,$BE,$0F,$A3,$27,$C1
	.dc.b	$1F,$12,$E4,$78,$DF,$8F,$F1,$DF,$DE,$FD,$8E,$FE,$64,$AF,$EE,$31
	.dc.b	$F3,$3E,$3B,$E6,$FC,$74,$9F,$64,$7D,$0F,$8C,$91,$EF,$7E,$9F,$C6
	.dc.b	$7A,$BC,$D9,$3F,$EB,$2C,$FF,$D5,$FE,$B7,$F5,$2F,$6B,$EC,$FC,$27
	.dc.b	$DA,$F8,$B9,$7F,$26,$3E,$EF,$34,$8F,$98,$FE,$03,$E2,$FF,$03,$F1
	.dc.b	$72,$7E,$83,$F8,$4F,$8B,$FF,$0A,$7E,$B3,$FD,$3F,$C1,$FF,$14,$3F
	.dc.b	$E0,$BF,$04,$D3,$57,$F8,$37,$FC,$6A,$FF,$90,$1B,$AF,$F5,$86,$D7
	.dc.b	$4F,$88,$EF,$F0,$F2,$FF,$C9,$F5,$FD,$BD,$BF,$0F,$E4,$F3,$BC,$BF
	.dc.b	$0F,$27,$71,$E8,$F8,$79,$1C,$9F,$5F,$C3,$FB,$3E,$12,$4F,$67,$DF
	.dc.b	$E6,$FC,$0F,$5F,$E0,$FA,$72,$BC,$71,$F1,$3C,$F9,$1E,$67,$E3,$F9
	.dc.b	$7F,$23,$C6,$93,$EA,$7E,$5F,$85,$F3,$3B,$3E,$6F,$C3,$4A,$F7,$C7
	.dc.b	$D0,$F8,$6F,$F2,$27,$EB,$3F,$9B,$E9,$37,$95,$3A,$F8,$69,$87,$F9
	.dc.b	$DA,$BF,$9F,$F4,$FF,$97,$DF,$E7,$ED,$F8,$6E,$BF,$0D,$2B,$84,$78
	.dc.b	$3C,$99,$1C,$DF,$1F,$C2,$F9,$3E,$16,$4F,$77,$CF,$F0,$BE,$8F,$85
	.dc.b	$F4,$F9,$32,$BC,$91,$EC,$F8,$5F,$6F,$C2,$C9,$F4,$47,$C0,$F8,$29
	.dc.b	$1E,$B7,$E1,$FC,$1F,$C4,$F8,$3F,$8B,$F0,$72,$BE,$04,$7C,$8F,$82
	.dc.b	$91,$F0,$DF,$97,$F0,$7F,$33,$E0,$FE,$6F,$C1,$CA,$F9,$11,$F4,$3E
	.dc.b	$0E,$47,$CB,$7F,$CE,$DF,$B5,$FF,$27,$E8,$61,$D2,$BE,$83,$AF,$83
	.dc.b	$ED,$1F,$4F,$1E,$CF,$FA,$8A,$98,$F6,$8C,$63,$7C,$FF,$F1,$EE,$5F
	.dc.b	$CD,$47,$C1,$F8,$BE,$0E,$5F,$FA,$D2,$72,$DF,$83,$F3,$91,$D7,$C1
	.dc.b	$FA,$BE,$0F,$D7,$F0,$32,$79,$47,$BB,$E0,$A4,$75,$7E,$0F,$BB,$F0
	.dc.b	$BE,$0A,$4F,$85,$F8,$BF,$03,$F1,$BE,$0B,$E3,$FC,$14,$AF,$34,$7C
	.dc.b	$AF,$82,$F9,$7F,$03,$27,$D5,$1F,$3B,$E0,$A4,$7B,$5F,$A3,$F0,$5F
	.dc.b	$4B,$E0,$A4,$FC,$17,$EA,$FC,$17,$D6,$F8,$2E,$D3,$D0,$F8,$CF,$47
	.dc.b	$FD,$A7,$C8,$C3,$1C,$FF,$43,$1E,$B5,$AC,$6C,$FD,$4C,$7F,$D7,$1D
	.dc.b	$BF,$C8,$3D,$1F,$C1,$FC,$8E,$0C,$FC,$86,$BE,$33,$7F,$0B,$33,$FF
	.dc.b	$75,$E1,$FC,$89,$FE,$D4,$FF,$1F,$7F,$DC,$EF,$FB,$39,$BF,$09,$E0
	.dc.b	$F8,$5E,$B1,$CB,$0A,$F0,$F8,$4F,$03,$DB,$0F,$03,$D3,$1E,$4F,$8B
	.dc.b	$0F,$59,$87,$C3,$F0,$BE,$E2,$3E,$8F,$85,$EA,$F9,$AF,$AE,$FF,$E9
	.dc.b	$4D,$3D,$9B,$7E,$9B,$5E,$B7,$DC,$F8,$2F,$CA,$F8,$5F,$97,$F0,$B2
	.dc.b	$7E,$2C,$7C,$EF,$85,$91,$F2,$1F,$A3,$F0,$9F,$4B,$E1,$7E,$9F,$C2
	.dc.b	$CA,$F9,$B1,$F5,$BE,$17,$B4,$74,$3E,$8B,$D1,$F9,$F7,$75,$F5,$61
	.dc.b	$D6,$7E,$BE,$1D,$5F,$AB,$87,$E0,$BE,$17,$F0,$7F,$0B,$27,$EE,$C7
	.dc.b	$E1,$BE,$12,$47,$E0,$1F,$C4,$FC,$2F,$E2,$BE,$11,$33,$3F,$F5,$EF
	.dc.b	$D3,$FF,$41,$3F,$A1,$71,$FF,$40,$9E,$A4,$3F,$0F,$8F,$27,$FD,$9C
	.dc.b	$7F,$8D,$67,$FA,$1B,$F8,$EC,$7C,$6F,$E3,$9F,$1B,$F9,$4C,$39,$61
	.dc.b	$E2,$8B,$F0,$9E,$6E,$46,$1E,$4B,$CD,$77,$F8,$CB,$5A,$F8,$4B,$66
	.dc.b	$CE,$3E,$17,$3F,$0B,$30,$F8,$59,$77,$C1,$5C,$BE,$17,$9F,$C2,$CB
	.dc.b	$D9,$EF,$8E,$FF,$0B,$DA,$33,$A7,$A3,$3C,$C6,$31,$EA,$DC,$FC,$4F
	.dc.b	$CA,$AF,$7E,$5E,$18,$E1,$97,$85,$D5,$B2,$F7,$FC,$24,$9E,$CF,$C2
	.dc.b	$F8,$5F,$81,$1D,$C8,$72,$C3,$B3,$D3,$0E,$DC,$1F,$84,$C3,$E5,$FC
	.dc.b	$24,$AF,$53,$F3,$BE,$17,$E7,$FC,$2C,$9F,$73,$F4,$BE,$17,$E9,$FC
	.dc.b	$27,$D4,$F8,$59,$5F,$0E,$39,$3D,$08,$7B,$AF,$A3,$F0,$2E,$E5,$F9
	.dc.b	$8C,$3A,$CF,$CD,$C3,$AB,$F1,$B0,$FC,$1F,$C2,$FE,$13,$AA,$4F,$D3
	.dc.b	$8F,$C3,$FC,$17,$E2,$3E,$12,$4F,$D7,$8F,$C5,$FC,$1C,$8F,$B4,$FE
	.dc.b	$43,$E7,$3F,$D3,$87,$FA,$5B,$9B,$F2,$F8,$78,$2B,$F0,$73,$9E,$6F
	.dc.b	$EE,$71,$F8,$3C,$7B,$BF,$73,$1E,$CF,$DE,$C6,$EC,$EA,$FD,$26,$3E
	.dc.b	$0F,$84,$F1,$7C,$24,$C3,$BC,$79,$7E,$13,$CD,$F0,$92,$7C,$71,$E9
	.dc.b	$F8,$39,$1E,$67,$D9,$F0,$9E,$DF,$84,$F7,$61,$2B,$D7,$1D,$5F,$6E
	.dc.b	$1D,$5F,$CD,$64,$C7,$AC,$63,$1B,$A7,$AE,$3F,$D7,$94,$CB,$C1,$1D
	.dc.b	$88,$74,$F8,$4F,$03,$DB,$1F,$05,$F3,$E3,$61,$E1,$AF,$61,$0F,$11
	.dc.b	$F5,$30,$FB,$5F,$09,$23,$EB,$BF,$77,$E1,$3E,$F7,$C1,$FD,$FF,$84
	.dc.b	$95,$F7,$23,$F0,$5F,$07,$F8,$3F,$84,$93,$F8,$08,$E4,$F4,$21,$F1
	.dc.b	$AF,$AB,$F8,$6B,$EA,$FC,$AB,$EA,$FE,$2A,$FA,$BF,$3A,$FF,$25,$F0
	.dc.b	$5F,$93,$F8,$29,$3F,$90,$8F,$CB,$7C,$14,$8F,$C9,$BF,$99,$F8,$2F
	.dc.b	$CD,$7C,$27,$F2,$AF,$9E,$FF,$07,$1F,$E0,$FE,$04,$76,$BE,$84,$3F
	.dc.b	$8C,$4F,$FA,$DE,$57,$DE,$7F,$06,$F7,$7F,$47,$5D,$FE,$12,$46,$0F
	.dc.b	$E5,$37,$FD,$F0,$9F,$05,$E5,$F7,$FC,$DE,$FC,$AD,$BE,$9F,$82,$F5
	.dc.b	$7C,$1C,$9E,$6F,$B7,$E0,$F9,$47,$6C,$3E,$01,$87,$D7,$7D,$5A,$79
	.dc.b	$2F,$A3,$F0,$2E,$F4,$FE,$47,$26,$3F,$2F,$E0,$FA,$4E,$73,$C6,$1E
	.dc.b	$82,$1D,$8F,$97,$8F,$4F,$0B,$FB,$16,$2C,$B1,$4E,$B4,$F8,$23,$E4
	.dc.b	$FC,$17,$DB,$23,$F3,$3E,$0F,$EF,$7C,$17,$DF,$F8,$29,$3F,$42,$3F
	.dc.b	$05,$F0,$32,$3E,$9B,$F8,$5F,$82,$FC,$35,$C9,$E8,$7D,$8F,$83,$EA
	.dc.b	$FD,$AC,$3A,$BF,$86,$C3,$A3,$F8,$8C,$2F,$07,$C9,$C7,$F2,$5F,$03
	.dc.b	$F9,$3F,$83,$97,$F8,$48,$FC,$B7,$C1,$48,$FC,$3B,$F9,$9F,$83,$FC
	.dc.b	$D7,$C1,$FF,$81,$23,$FD,$B8,$7F,$B7,$78,$3F,$C3,$95,$F9,$8B,$EE
	.dc.b	$F5,$3F,$BC,$CF,$F8,$6B,$87,$E6,$BD,$E3,$F3,$EF,$77,$F2,$F7,$DD
	.dc.b	$86,$0E,$7F,$03,$E3,$F8,$39,$18,$3B,$FB,$FE,$7F,$82,$91,$83,$C7
	.dc.b	$EF,$FA,$C8,$F9,$BE,$03,$DD,$84,$8E,$87,$F8,$2F,$27,$C1,$FC,$3C
	.dc.b	$7A,$CE,$71,$C3,$D8,$87,$1C,$7E,$56,$3F,$2E,$FB,$4F,$5C,$70,$F9
	.dc.b	$CE,$87,$CA,$BE,$AF,$33,$AF,$17,$F3,$7C,$BF,$C3,$51,$EB,$3E,$8F
	.dc.b	$C1,$FD,$9F,$83,$91,$83,$EA,$FC,$1F,$DD,$23,$F6,$3E,$0F,$F0,$1F
	.dc.b	$07,$F8,$1F,$83,$93,$F7,$23,$F0,$98,$48,$E8,$7C,$7F,$82,$E6,$FC
	.dc.b	$9C,$2E,$7E,$7C,$7A,$CF,$E1,$B1,$EA,$FC,$FC,$7F,$21,$F0,$72,$3F
	.dc.b	$16,$FE,$4F,$E0,$BF,$29,$F0,$7F,$95,$F8,$19,$5F,$90,$8F,$CC,$7C
	.dc.b	$14,$8F,$C9,$BF,$9B,$F8,$2E,$51,$D0,$87,$DF,$BE,$AF,$E6,$AF,$AB
	.dc.b	$F8,$3B,$E8,$FE,$16,$EE,$9F,$4F,$0F,$F3,$6A,$FC,$51,$CD,$86,$DF
	.dc.b	$F9,$8A,$75,$39,$E7,$D7,$3E,$F8,$E5,$EF,$0F,$C9,$B7,$93,$A7,$1C
	.dc.b	$8C,$1F,$E6,$AF,$A4,$FF,$03,$3F,$E6,$56,$E7,$FD,$FD,$BE,$4F,$FA
	.dc.b	$1D,$FF,$19,$0F,$83,$FF,$42,$AF,$84,$A6,$23,$3F,$07,$AF,$81,$98
	.dc.b	$F8,$63,$8E,$12,$3C,$8F,$4F,$89,$E5,$1E,$7C,$3A,$BE,$9C,$3A,$BE
	.dc.b	$BC,$3A,$BE,$BC,$39,$3D,$08,$7B,$EF,$AB,$F0,$6F,$AB,$F0,$EF,$AB
	.dc.b	$F1,$6F,$AE,$5F,$17,$0F,$B1,$C9,$F6,$79,$65,$FD,$38,$FB,$9C,$B2
	.dc.b	$3E,$B3,$F7,$F9,$7F,$07,$F8,$AF,$CE,$96,$28,$C3,$76,$DE,$DC,$35
	.dc.b	$E3,$B4,$C3,$F1,$18,$7D,$56,$EB,$DB,$69,$18,$37,$F0,$BE,$22,$3C
	.dc.b	$BE,$13,$CB,$F0,$DE,$6F,$86,$93,$DA,$3F,$D8,$EB,$F1,$67,$C3,$E3
	.dc.b	$E7,$5E,$E5,$37,$3A,$AF,$3D,$7C,$22,$3E,$AF,$88,$F8,$BF,$0F,$23
	.dc.b	$DA,$FC,$8F,$88,$E7,$5D,$67,$FD,$92,$31,$F0,$B8,$C3,$C0,$FF,$BB
	.dc.b	$39,$7E,$77,$C3,$FD,$2F,$88,$93,$F2,$A3,$EA,$FC,$3C,$8F,$9A,$FD
	.dc.b	$8F,$88,$F0,$C7,$83,$1E,$EF,$FB,$E0,$F1,$6D,$A7,$C1,$CB,$FE,$3A
	.dc.b	$7E,$BF,$C4,$7E,$08,$FC,$1F,$C3,$FE,$13,$E1,$FF,$83,$FE,$2E,$F8
	.dc.b	$8A,$66,$FD,$9C,$3F,$05,$F0,$67,$E2,$E0,$FD,$EF,$0F,$C0,$7F,$D0
	.dc.b	$83,$07,$C5,$F6,$DA,$61,$FC,$46,$72,$DF,$86,$F1,$FC,$4C,$8C,$1E
	.dc.b	$23,$E1,$57,$8F,$0F,$0E,$7F,$83,$B9,$5F,$1D,$F8,$93,$E4,$C1,$F0
	.dc.b	$FF,$03,$E1,$E5,$F8,$E3,$7E,$1A,$26,$DE,$F8,$C4,$7D,$03,$C8,$7D
	.dc.b	$3D,$BE,$14,$6A,$59,$79,$2B,$59,$F6,$7B,$5B,$B3,$F3,$26,$63,$F4
	.dc.b	$13,$EF,$C3,$A6,$B9,$BE,$F5,$E6,$C3,$B6,$BF,$12,$3C,$AF,$D6,$C7
	.dc.b	$AC,$63,$0F,$5D,$F2,$8F,$61,$F7,$B0,$ED,$8F,$4A,$F6,$63,$8A,$FD
	.dc.b	$06,$1D,$1F,$69,$D4,$F9,$37,$E9,$7D,$F7,$ED,$CB,$E2,$C7,$D7,$38
	.dc.b	$1F,$00,$EA,$43,$A9,$32,$FE,$9B,$1F,$56,$3B,$9C,$B2,$F6,$47,$68
	.dc.b	$FC,$CE,$3E,$97,$EC,$63,$E9,$BF,$91,$1D,$72,$F9,$EF,$E4,$0E,$98
	.dc.b	$FC,$37,$E8,$18,$3C,$D3,$F3,$DC,$1F,$2C,$E4,$75,$3E,$19,$F0,$4F
	.dc.b	$29,$C1,$E4,$F9,$8C,$4F,$91,$E8,$73,$3E,$41,$C8,$91,$EC,$24,$7E
	.dc.b	$02,$BE,$03,$F8,$2A,$FA,$99,$7D,$7B,$E7,$7C,$1F,$79,$D2,$B8,$3F
	.dc.b	$3E,$F9,$BE,$C3,$C9,$8F,$D5,$7F,$1B,$5C,$DE,$B5,$F8,$C7,$E1,$46
	.dc.b	$72,$FC,$A1,$FA,$58,$F9,$B3,$E0,$7A,$E1,$E7,$9D,$E3,$F5,$67,$F9
	.dc.b	$B6,$5F,$7F,$0F,$96,$4C,$39,$3E,$3C,$7C,$C7,$9F,$0E,$77,$E4,$AF
	.dc.b	$33,$F3,$E3,$CB,$97,$C0,$8F,$59,$F0,$6F,$F2,$B9,$79,$72,$FD,$06
	.dc.b	$5E,$5C,$3A,$DF,$DD,$25,$FC,$1A,$FA,$59,$78,$DE,$33,$E3,$3B,$65
	.dc.b	$EA,$3E,$E3,$F1,$8C,$90,$FA,$2F,$E2,$6B,$B4,$73,$3E,$39,$F8,$E3
	.dc.b	$F8,$AE,$5E,$DC,$3C,$F3,$EC,$9F,$CD,$BC,$9F,$C2,$1F,$86,$70,$7C
	.dc.b	$F7,$99,$D4,$F1,$9F,$B8,$3E,$65,$FD,$9C,$3F,$71,$3C,$B0,$F6,$D7
	.dc.b	$C6,$7E,$03,$ED,$9F,$B3,$97,$DF,$9F,$AF,$5F,$C1,$B2,$C5,$F8,$8E
	.dc.b	$47,$C5,$C7,$F8,$3D,$FD,$78,$FE,$27,$87,$8B,$0E,$67,$82,$3E,$F9
	.dc.b	$FE,$15,$CB,$E3,$65,$FA,$A8,$F8,$B3,$BC,$3F,$89,$E3,$F8,$1A,$FE
	.dc.b	$75,$7C,$2F,$BE,$3E,$17,$F5,$35,$B7,$FA,$B1,$FD,$A7,$4F,$EB,$2F
	.dc.b	$F8,$3F,$5F,$EB,$AF,$23,$C2,$7B,$70,$F9,$95,$FD,$47,$4F,$EE,$CF
	.dc.b	$F6,$F9,$E5,$B7,$F8,$12,$3D,$51,$FC,$3A,$7F,$4D,$1C,$A3,$F9,$F4
	.dc.b	$7F,$92,$F8,$7F,$C1,$31,$FD,$F7,$87,$FC,$14,$FE,$D8,$F1,$9F,$A6
	.dc.b	$9F,$F0,$37,$0F,$F8,$12,$3D,$5A,$7F,$85,$23,$FC,$49,$1F,$E1,$CE
	.dc.b	$6B,$5F,$E0,$EE,$57,$37,$9A,$3F,$29,$5F,$CB,$5F,$F2,$F7,$2F,$F8
	.dc.b	$86,$3F,$C4,$67,$F9,$69,$FF,$20,$3F,$E2,$8E,$AB,$3D,$7A,$BA,$72
	.dc.b	$FE,$9C,$FC,$7C,$7F,$8F,$2B,$FC,$C7,$D5,$FE,$3C,$7F,$C6,$3D,$5F
	.dc.b	$E5,$C7,$FC,$CF,$CD,$FE,$5C,$7F,$C6,$BD,$9D,$1F,$D0,$5F,$F2,$D3
	.dc.b	$FC,$E9,$CB,$FE,$56,$8F,$F1,$DB,$FE,$7A,$EC,$FF,$3A,$47,$F9,$CB
	.dc.b	$C1,$FF,$3A,$3F,$E4,$0F,$0B,$FC,$FF,$E1,$7F,$9C,$A3,$FC,$87,$E1
	.dc.b	$61,$FF,$39,$90,$FF,$23,$F8,$5F,$E8,$6F,$0B,$FD,$1F,$1F,$EB,$FF
	.dc.b	$1F,$FD,$1E,$FF,$96,$FC,$B7,$91,$FE,$8F,$8C,$79,$17,$7F,$F2,$0D
	.dc.b	$3C,$8F,$F4,$8F,$2C,$BF,$F4,$4B,$FE,$8E,$7F,$D8,$D9,$FF,$AF,$88
	.dc.b	$FF,$B2,$73,$FF,$60,$F1,$7F,$B0,$B8,$E4,$FF,$B3,$63,$FD,$8B,$C7
	.dc.b	$FE,$C6,$E3,$93,$FE,$D1,$8E,$6F,$FB,$47,$0B,$A7,$FB,$47,$1E,$B3
	.dc.b	$FE,$D7,$C7,$AB,$FE,$D7,$C7,$AB,$FE,$DB,$C7,$AB,$FE,$DB,$C7,$FD
	.dc.b	$AB,$CF,$FE,$D6,$E7,$93,$FE,$E1,$8F,$F6,$CF,$34,$8F,$F7,$23,$FE
	.dc.b	$DC,$E7,$FF,$6E,$F5,$C9,$FF,$74,$3F,$EE,$0E,$AE,$B1,$FE,$EB,$C3
	.dc.b	$AB,$FE,$EB,$C3,$A3,$FE,$ED,$C2,$ED,$FE,$ED,$C7,$A4,$FF,$BC,$71
	.dc.b	$B3,$FE,$F1,$C7,$FE,$00,$4F,$FB,$F7,$CF,$FF,$2E,$CF,$DA,$73,$1F
	.dc.b	$81,$F3,$B3,$C5,$FE,$65,$A7,$23,$CE,$CD,$FF,$B3,$A3,$0E,$EB,$84
	.dc.b	$2F,$77,$CB,$CE,$B2,$F3,$A9,$88,$EB,$EE,$3C,$FF,$F4,$F6,$5F,$ED
	.dc.b	$CC,$BF,$CF,$13,$9F,$3B,$74,$6F,$85,$8F,$1F,$9F,$D1,$FF,$A8,$71
	.dc.b	$F3,$0C,$11,$98,$EC,$67,$05,$72,$F1,$B9,$F8,$F2,$B9,$E1,$CA,$7C
	.dc.b	$31,$8B,$F2,$37,$2C,$BD,$AB,$DD,$87,$9B,$C8,$93,$C6,$3C,$0F,$6C
	.dc.b	$BF,$CC,$55,$E8,$CB,$33,$CF,$1E,$ED,$C7,$E2,$9E,$0E,$06,$BF,$2A
	.dc.b	$BD,$B5,$D7,$3F,$0B,$CF,$3E,$B7,$C2,$3E,$46,$7E,$22,$69,$F3,$BC
	.dc.b	$9F,$9F,$E5,$4A,$F4,$47,$D2,$F3,$FE,$9F,$9B,$27,$D9,$1F,$5B,$E2
	.dc.b	$BE,$BF,$C5,$49,$F8,$11,$CD,$F9,$19,$31,$C6,$7E,$5E,$2F,$B7,$9F
	.dc.b	$7B,$F9,$19,$F7,$78,$E7,$DD,$F9,$99,$F7,$7A,$E7,$F8,$8F,$8A,$FC
	.dc.b	$4F,$C5,$49,$FA,$51,$F8,$CF,$88,$FC,$6F,$C4,$C9,$FA,$D1,$F9,$0F
	.dc.b	$89,$91,$F6,$5F,$C9,$FC,$4F,$38,$F5,$E1,$77,$F5,$70,$B2,$ED,$5F
	.dc.b	$81,$CB,$B3,$F0,$B2,$EA,$FC,$4C,$AE,$1C,$B2,$FD,$0F,$C5,$4A,$FC
	.dc.b	$53,$FA,$4F,$87,$FD,$2F,$C5,$49,$FC,$73,$FA,$8F,$83,$FD,$4F,$C5
	.dc.b	$7E,$AB,$E2,$65,$7E,$4E,$3F,$6F,$FB,$67,$FB,$A2,$3F,$38,$DC,$FE
	.dc.b	$A3,$7B,$7F,$73,$C2,$D3,$C1,$7F,$73,$3B,$87,$C2,$CF,$C3,$3F,$A1
	.dc.b	$CF,$C2,$FE,$0B,$3F,$0F,$C8,$78,$BE,$26,$4F,$ED,$35,$4F,$4D,$7C
	.dc.b	$39,$DA,$FC,$DE,$5D,$9F,$C6,$E5,$DB,$3F,$A7,$9F,$BB,$E2,$FD,$FF
	.dc.b	$17,$2F,$B4,$7C,$2F,$8B,$91,$E1,$7E,$2F,$C2,$FC,$6F,$32,$4F,$95
	.dc.b	$F9,$3E,$5F,$CA,$F3,$3E,$5F,$95,$2B,$D5,$1D,$1F,$9B,$85,$D3,$E6
	.dc.b	$E3,$D2,$7E,$8E,$36,$7E,$8E,$3F,$5B,$CB,$FA,$FF,$15,$2F,$E3,$47
	.dc.b	$DA,$F8,$9F,$B7,$F1,$52,$7E,$54,$7D,$EF,$8B,$91,$F3,$5F,$C0,$FC
	.dc.b	$5F,$E0,$BE,$2F,$F0,$7F,$03,$2B,$E9,$47,$57,$EA,$61,$D5,$FC,$3E
	.dc.b	$1D,$1F,$C4,$E1,$74,$FC,$4E,$3F,$8E,$F8,$29,$5F,$75,$FC,$8F,$C1
	.dc.b	$7E,$4B,$E0,$A4,$FE,$05,$FE,$97,$EA,$FF,$19,$8F,$E6,$0B,$E2,$F2
	.dc.b	$48,$FE,$77,$0F,$86,$E3,$F0,$FC,$BE,$1E,$56,$A3,$BB,$FA,$0C,$2E
	.dc.b	$9F,$C4,$F0,$F0,$37,$1E,$D5,$E2,$7C,$B8,$F9,$BE,$1F,$CF,$F0,$F2
	.dc.b	$BC,$71,$D3,$C2,$FE,$C3,$E2,$7F,$8C,$67,$DC,$7A,$23,$AE,$3D,$E3
	.dc.b	$D7,$8F,$67,$DB,$F0,$FF,$13,$E1,$A4,$60,$F8,$3F,$0F,$F2,$08,$FC
	.dc.b	$4F,$86,$F9,$7F,$0F,$F3,$3E,$1A,$4F,$C8,$8F,$9F,$F0,$FF,$43,$E1
	.dc.b	$E4,$FC,$C8,$E6,$FD,$2C,$2E,$9F,$1F,$1E,$B3,$F4,$B1,$EA,$FD,$8C
	.dc.b	$7E,$DF,$C3,$C8,$FA,$EF,$DE,$F8,$7F,$BF,$F0,$F2,$7E,$DB,$F8,$2E
	.dc.b	$5F,$C1,$FC,$2C,$9F,$BE,$FE,$1B,$E0,$FF,$C2,$3F,$B6,$FF,$80,$F5
	.dc.b	$FE,$98,$DE,$4F,$F0,$8B,$6F,$FC,$14,$BE,$13,$C1,$1F,$8E,$CA,$DF
	.dc.b	$E0,$F2,$F0,$4F,$E4,$B2,$B3,$F8,$7C,$BC,$3F,$0F,$E2,$F8,$39,$7F
	.dc.b	$E0,$EC,$FF,$C0,$27,$F6,$17,$1F,$0B,$E8,$3D,$3F,$07,$EA,$F8,$4F
	.dc.b	$5F,$C1,$CA,$E5,$1E,$EF,$84,$91,$D5,$F8,$3F,$07,$CE,$3A,$90,$FF
	.dc.b	$0C,$63,$FE,$D2,$6D,$75,$73,$7E,$17,$D1,$F0,$9F,$30,$8F,$AF,$E0
	.dc.b	$FB,$3D,$33,$BB,$FD,$0D,$39,$4F,$43,$E1,$61,$89,$FA,$BF,$09,$F6
	.dc.b	$09,$3F,$1F,$E1,$3E,$DF,$C2,$48,$F9,$4F,$DE,$F8,$2F,$BF,$F0,$9F
	.dc.b	$80,$F8,$19,$5F,$3E,$3F,$07,$F0,$32,$3E,$93,$F8,$6F,$80,$E9,$1F
	.dc.b	$13,$0B,$9F,$E2,$31,$EB,$3F,$67,$1E,$AF,$CA,$C7,$F1,$FF,$03,$23
	.dc.b	$EF,$3F,$E6,$2F,$DB,$BF,$B4,$47,$F9,$5D,$FE,$B6,$FF,$94,$1A,$FF
	.dc.b	$92,$DB,$87,$F8,$8D,$7C,$27,$3F,$82,$95,$C5,$ED,$F0,$7D,$FE,$06
	.dc.b	$4F,$57,$C5,$F0,$3E,$3F,$84,$93,$E1,$7C,$D8,$74,$8F,$16,$17,$2F
	.dc.b	$16,$3E,$BF,$82,$F6,$7C,$14,$BF,$54,$7B,$FE,$0F,$E0,$7C,$24,$9F
	.dc.b	$74,$7C,$3F,$83,$91,$F0,$5F,$8D,$F0,$9F,$1F,$E0,$64,$FC,$57,$E5
	.dc.b	$7C,$17,$58,$F9,$18,$75,$7E,$46,$1D,$1F,$95,$85,$D3,$E5,$63,$F4
	.dc.b	$FE,$0A,$57,$D1,$7E,$B7,$C1,$75,$8F,$A5,$8F,$5B,$FA,$58,$F3,$F3
	.dc.b	$BF,$D3,$3C,$BF,$EA,$1B,$F5,$1F,$6F,$2F,$C0,$7C,$1F,$E0,$7E,$12
	.dc.b	$4F,$DF,$8F,$C2,$7C,$24,$8F,$C1,$3F,$87,$F8,$4F,$C4,$7C,$14,$9F
	.dc.b	$C3,$3F,$EA,$CF,$DB,$FF,$A9,$E7,$FE,$AA,$7F,$B1,$4F,$FA,$69,$AF
	.dc.b	$FA,$55,$BA,$FF,$9B,$5F,$EC,$91,$E0,$21,$F9,$1C,$3C,$4F,$80,$87
	.dc.b	$E5,$6F,$C5,$7E,$3F,$84,$93,$E1,$7C,$DF,$0B,$E7,$F8,$3F,$47,$C2
	.dc.b	$4A,$F3,$47,$AF,$E0,$BD,$9F,$05,$27,$D5,$1E,$FF,$80,$F8,$1F,$01
	.dc.b	$27,$DD,$1D,$5F,$81,$87,$57,$E0,$61,$D5,$F8,$98,$75,$7E,$26,$1F
	.dc.b	$2F,$DF,$F9,$9F,$01,$27,$E5,$47,$CF,$F8,$19,$1F,$35,$FA,$5F,$05
	.dc.b	$F4,$FE,$0A,$4F,$D1,$7E,$B7,$C1,$7D,$7F,$81,$FB,$1F,$03,$2B,$EB
	.dc.b	$47,$47,$EB,$61,$79,$FE,$B6,$3D,$27,$ED,$63,$67,$ED,$63,$F8,$3F
	.dc.b	$81,$95,$F8,$17,$F0,$DF,$03,$CF,$9B,$F0,$D9,$5A,$FF,$74,$74,$B3
	.dc.b	$E9,$5F,$8B,$CE,$D6,$B2,$FF,$7D,$F4,$FF,$67,$F8,$73,$49,$F0,$6A
	.dc.b	$FF,$7A,$EC,$FC,$59,$E0,$66,$7F,$DE,$B3,$98,$C3,$FE,$D5,$31,$1C
	.dc.b	$A5,$E4,$6B,$4E,$D5,$D7,$C8,$C5,$A3,$0F,$18,$CC,$78,$FC,$7F,$5F
	.dc.b	$C2,$4A,$FF,$80,$3A,$7F,$BD,$07,$C2,$7C,$0F,$7B,$E0,$FB,$D2,$B7
	.dc.b	$1F,$13,$E0,$64,$72,$7E,$3F,$BF,$F2,$3E,$12,$4F,$67,$E5,$FC,$27
	.dc.b	$58,$F0,$E1,$D5,$F5,$6F,$F2,$F9,$3E,$7E,$5D,$A7,$22,$3E,$9E,$59
	.dc.b	$DF,$FD,$D5,$AB,$4F,$B3,$F0,$7F,$6B,$E0,$E6,$3F,$02,$3E,$2F,$E2
	.dc.b	$AF,$1A,$92,$7E,$DC,$3C,$4D,$F8,$BF,$43,$57,$E2,$AF,$27,$0D,$C3
	.dc.b	$8D,$7E,$2B,$E1,$64,$7D,$07,$F1,$BF,$0B,$E2,$F2,$7C,$7E,$24,$AF
	.dc.b	$0C,$79,$BC,$2E,$BA,$49,$F2,$C7,$E5,$BE,$0E,$47,$DC,$7F,$33,$F0
	.dc.b	$7E,$EF,$2B,$DF,$E4,$CA,$F6,$C7,$E7,$BE,$0A,$47,$E1,$1F,$D0,$FC
	.dc.b	$1F,$E8,$BE,$16,$4F,$E2,$1F,$D2,$FC,$2F,$E3,$BD,$DF,$C4,$CF,$E8
	.dc.b	$9A,$FE,$81,$BD,$9F,$9F,$7F,$C8,$51,$F7,$39,$BF,$5B,$F0,$BF,$B0
	.dc.b	$24,$7E,$83,$C0,$FC,$27,$EE,$1F,$E0,$CA,$3F,$E0,$88,$CF,$C2,$7D
	.dc.b	$58,$C5,$7E,$1D,$FF,$05,$BE,$5F,$13,$D1,$DD,$23,$07,$73,$87,$C1
	.dc.b	$F3,$F8,$39,$19,$CB,$C3,$3D,$FE,$1B,$C1,$F0,$D2,$B5,$92,$7E,$37
	.dc.b	$34,$9F,$8C,$FC,$9E,$4F,$47,$04,$9E,$2F,$AF,$E1,$7D,$9F,$05,$27
	.dc.b	$AB,$EF,$F8,$3F,$DF,$FE,$2B,$FE,$30,$E5,$34,$9C,$7F,$17,$E1,$BE
	.dc.b	$37,$AB,$31,$F3,$C7,$C9,$F5,$24,$7A,$9F,$99,$EA,$76,$E2,$EF,$B4
	.dc.b	$AF,$E4,$39,$27,$E9,$7A,$92,$7E,$1B,$F5,$7E,$1B,$EB,$7C,$27,$D7
	.dc.b	$F8,$59,$5F,$22,$3D,$7A,$FB,$34,$93,$DA,$3E,$F7,$A3,$F7,$FE,$16
	.dc.b	$4F,$D0,$8F,$C1,$7C,$2C,$8F,$A6,$FE,$E7,$F7,$1F,$F0,$5C,$7F,$29
	.dc.b	$DB,$F9,$EB,$EA,$AF,$96,$7C,$CC,$F8,$7C,$54,$9F,$C1,$5F,$F8,$3B
	.dc.b	$C1,$FE,$30,$FF,$83,$A3,$F3,$8B,$E1,$FC,$1F,$0F,$27,$8C,$7F,$36
	.dc.b	$F8,$77,$C3,$DF,$5E,$57,$F4,$C9,$CB,$E0,$9F,$54,$C8,$C1,$E4,$F3
	.dc.b	$FD,$DE,$A4,$8C,$1E,$8B,$F8,$44,$7D,$7E,$B7,$C5,$F8,$79,$1E,$E7
	.dc.b	$E4,$7A,$7F,$27,$D4,$93,$F0,$9F,$99,$F0,$9F,$37,$D5,$F9,$DF,$09
	.dc.b	$2B,$E3,$C7,$D1,$F8,$79,$1F,$29,$FA,$9E,$AF,$D5,$F5,$BE,$B7,$C3
	.dc.b	$CA,$F9,$F1,$FC,$C7,$F7,$2F,$EE,$84,$F8,$FC,$7C,$46,$7E,$1F,$5F
	.dc.b	$0F,$30,$FD,$C5,$A7,$97,$C2,$C9,$CB,$D7,$E1,$BB,$7C,$37,$7F,$86
	.dc.b	$95,$CA,$3C,$5F,$0F,$E3,$F7,$64,$F6,$8F,$37,$BB,$23,$C2,$FA,$7E
	.dc.b	$1F,$D5,$F0,$BE,$BF,$85,$95,$E6,$8F,$77,$C2,$C8,$F4,$BF,$07,$DC
	.dc.b	$F8,$5E,$E4,$9F,$6B,$F1,$7E,$17,$E3,$7C,$2C,$9F,$82,$FC,$9F,$85
	.dc.b	$F9,$5F,$0B,$27,$E2,$BF,$D7,$63,$F9,$71,$FC,$AB,$E1,$0D,$27,$C5
	.dc.b	$EF,$E1,$78,$7C,$34,$C3,$F9,$35,$27,$A7,$C2,$F5,$F7,$65,$6A,$3C
	.dc.b	$1E,$F7,$F8,$8F,$D3,$FF,$01,$60,$8C,$57,$F8,$06,$1F,$0B,$8D,$3C
	.dc.b	$B5,$65,$F0,$B7,$E1,$70,$DC,$7C,$BF,$07,$BF,$7B,$83,$63,$D3,$EF
	.dc.b	$F3,$F8,$3E,$8D,$8F,$6F,$C1,$F7,$F8,$3F,$03,$63,$E0,$FC,$1F,$8F
	.dc.b	$DF,$F2,$36,$3E,$2F,$C0,$79,$FE,$0F,$D0,$D8,$F9,$3F,$07,$EB,$3D
	.dc.b	$DF,$B9,$A3,$07,$F5,$35,$F0,$7A,$F8,$FD,$B6,$75,$F0,$5C,$BE,$0F
	.dc.b	$9B,$63,$97,$C1,$F6,$3B,$FC,$1D,$8E,$DF,$07,$E2,$3C,$7F,$07,$63
	.dc.b	$C5,$F0,$7E,$63,$CF,$F0,$76,$3C,$DF,$07,$EA,$F8,$3F,$5B,$63,$D5
	.dc.b	$F0,$7E,$E3,$DF,$F0,$76,$3D,$DF,$05,$F0,$BE,$0F,$E1,$B6,$3E,$17
	.dc.b	$C1,$FC,$63,$E3,$FC,$1D,$8F,$8D,$F0,$5F,$2B,$E0,$FE,$5B,$63,$FB
	.dc.b	$E7,$EE,$8F,$84,$FC,$63,$63,$3F,$09,$89,$CF,$C1,$7E,$53,$E2,$1F
	.dc.b	$8A,$FF,$0D,$BE,$73,$98,$C7,$9F,$FE,$1F,$C7,$FC,$DF,$5E,$0F,$84
	.dc.b	$EF,$F0,$DE,$07,$C3,$F0,$B2,$7C,$BF,$09,$70,$FF,$45,$FC,$23,$E2
	.dc.b	$7C,$4D,$AF,$F4,$9A,$F6,$B3,$3E,$CF,$86,$95,$F0,$3E,$1A,$E1,$E5
	.dc.b	$C3,$E0,$FC,$37,$C2,$6D,$7F,$A7,$B7,$7C,$1F,$B7,$E1,$7D,$CD,$9F
	.dc.b	$97,$F0,$DF,$2B,$E1,$BE,$5B,$63,$E7,$D7,$CE,$AF,$9E,$D8,$ED,$ED
	.dc.b	$7C,$9F,$82,$F9,$4D,$8F,$AF,$F0,$DF,$58,$FA,$FF,$0D,$63,$ED,$FC
	.dc.b	$37,$DA,$F8,$6F,$B6,$D8,$FF,$5B,$FE,$28,$F8,$7F,$B4,$FF,$8D,$E3
	.dc.b	$31,$8F,$85,$FC,$1E,$5F,$84,$CE,$D7,$E2,$1F,$E2,$5F,$BA,$FF,$B0
	.dc.b	$E3,$FC,$8C,$DE,$4F,$F2,$F3,$F5,$5A,$78,$3E,$0E,$4F,$93,$E0,$6C
	.dc.b	$F0,$EA,$EA,$76,$F8,$8B,$5D,$B4,$F0,$E9,$E2,$6C,$7B,$BE,$03,$DA
	.dc.b	$7B,$BD,$EB,$1E,$6F,$88,$F4,$9E,$AF,$86,$B1,$EA,$F8,$8F,$6F,$C3
	.dc.b	$FB,$9B,$1F,$EE,$0F,$63,$F8,$A3,$7D,$DF,$D2,$46,$37,$FF,$70,$BF
	.dc.b	$C5,$A7,$E2,$F0,$FC,$8E,$1F,$92,$D8,$F9,$3F,$0F,$F3,$0F,$9B,$F0
	.dc.b	$D6,$3C,$1E,$E7,$63,$BE,$16,$3E,$8F,$C3,$FD,$43,$EA,$FC,$3D,$8F
	.dc.b	$AB,$C5,$F6,$38,$BE,$CB,$63,$D9,$ED,$FA,$BD,$AF,$5B,$63,$EE,$FC
	.dc.b	$3F,$BB,$DB,$FC,$0B,$63,$F0,$3F,$0D,$F8,$4F,$87,$FC,$2B,$63,$E4
	.dc.b	$7B,$5F,$1B,$D9,$F8,$ED,$8F,$F4,$97,$AE,$FC,$23,$F8,$97,$3F,$FB
	.dc.b	$8A,$31,$EC,$FC,$EF,$6F,$F8,$95,$6A,$7F,$A3,$57,$FB,$9F,$C3,$7C
	.dc.b	$47,$E5,$8F,$ED,$19,$FD,$5A,$DF,$C4,$AB,$79,$F8,$B1,$FE,$2F,$8F
	.dc.b	$85,$B7,$D8,$F3,$7B,$3E,$2F,$6B,$C7,$ED,$79,$3D,$AB,$8F,$37,$C3
	.dc.b	$E0,$F8,$BE,$22,$CF,$FA,$B7,$64,$79,$BE,$1F,$CF,$F1,$16,$B4,$FA
	.dc.b	$BE,$23,$D7,$E0,$D9,$E4,$FB,$BC,$19,$1D,$7E,$22,$F1,$7F,$B4,$73
	.dc.b	$FF,$3B,$4F,$43,$C5,$97,$81,$F2,$63,$D9,$A7,$63,$D1,$F1,$1F,$33
	.dc.b	$E2,$3E,$6B,$63,$E5,$7A,$9F,$58,$FA,$FE,$AD,$8F,$9D,$EA,$7D,$A3
	.dc.b	$ED,$FA,$96,$3E,$27,$C4,$7D,$8F,$13,$EC,$B6,$3E,$47,$8B,$F7,$0F
	.dc.b	$BB,$E2,$D8,$F9,$9F,$0F,$F8,$03,$F0,$3F,$11,$63,$E8,$7C,$47,$E1
	.dc.b	$0F,$C2,$F8,$B6,$3A,$93,$93,$E4,$61,$D5,$F9,$58,$73,$7E,$66,$17
	.dc.b	$BF,$F1,$BC,$A3,$F2,$27,$E4,$BC,$4B,$8F,$59,$DE,$9E,$A8,$C0,$C1
	.dc.b	$E1,$F3,$CB,$B4,$E6,$3B,$65,$8A,$DF,$EC,$FF,$A5,$AD,$36,$33,$F1
	.dc.b	$3F,$E4,$B8,$C5,$9F,$EF,$39,$7E,$99,$FD,$0D,$F2,$F8,$9E,$1E,$1F
	.dc.b	$1F,$13,$97,$89,$6B,$D1,$87,$18,$ED,$E2,$77,$F8,$7B,$5A,$3C,$AF
	.dc.b	$8F,$1E,$B1,$9D,$7D,$71,$85,$87,$97,$1F,$0D,$6F,$4F,$54,$62,$B7
	.dc.b	$3F,$0B,$E1,$FE,$07,$C3,$7C,$16,$C7,$1D,$9F,$0F,$F1,$4F,$8D,$F0
	.dc.b	$B6,$75,$F0,$DF,$27,$E1,$7E,$53,$63,$97,$C3,$7C,$DF,$86,$F9,$CD
	.dc.b	$8E,$D8,$73,$D3,$CA,$7D,$8C,$71,$BF,$D1,$C7,$A4,$EB,$7F,$AD,$18
	.dc.b	$9F,$AD,$F0,$DF,$6F,$C4,$FB,$8D,$CB,$D9,$E2,$7D,$FF,$13,$F0,$0D
	.dc.b	$8F,$81,$E2,$7E,$0C,$FF,$47,$7E,$ED,$FD,$EF,$1F,$C1,$46,$3E,$26
	.dc.b	$57,$CF,$AF,$F1,$83,$63,$AB,$F2,$F0,$BA,$FA,$78,$FF,$D5,$B3,$E0
	.dc.b	$24,$7B,$6F,$C1,$1F,$98,$C3,$C7,$F0,$DE,$4F,$85,$B3,$E1,$7A,$E7
	.dc.b	$F9,$7C,$AC,$FE,$81,$F6,$3D,$A3,$FA,$0E,$1D,$B1,$FD,$3E,$3F,$03
	.dc.b	$E1,$A5,$FB,$7E,$1A,$D7,$97,$E1,$BE,$29,$F1,$BE,$16,$D7,$C3,$F8
	.dc.b	$6F,$93,$F0,$BF,$29,$B1,$F1,$FE,$1B,$E6,$FC,$37,$CE,$6C,$7C,$BC
	.dc.b	$39,$E3,$F0,$F1,$BB,$7A,$31,$E7,$3F,$1D,$B1,$EC,$F8,$6F,$B2,$7D
	.dc.b	$AF,$86,$B8,$FD,$7F,$86,$FB,$A7,$DE,$F8,$5B,$1F,$6F,$E1,$BF,$03
	.dc.b	$F0,$DF,$82,$6C,$7D,$FF,$85,$FC,$2F,$C3,$7E,$19,$B1,$D4,$9A,$FD
	.dc.b	$BC,$3A,$BF,$76,$F9,$B6,$7F,$15,$87,$E4,$3E,$16,$47,$EE,$FF,$77
	.dc.b	$FE,$F0,$FF,$1F,$E8,$FE,$EF,$7F,$A3,$9F,$F1,$73,$7E,$0F,$FD,$CD
	.dc.b	$18,$E0,$FE,$FE,$FF,$6E,$BE,$9E,$FE,$8D,$FC,$1D,$8F,$07,$C0,$72
	.dc.b	$F8,$3E,$6D,$8F,$27,$C2,$76,$CB,$BB,$63,$A9,$33,$F3,$61,$D5,$F4
	.dc.b	$7C,$27,$9B,$E1,$3C,$ED,$AE,$A7,$C1,$C3,$AB,$4E,$9A,$FB,$34,$F4
	.dc.b	$C6,$0F,$E8,$9A,$BE,$0F,$E1,$9F,$13,$E1,$2C,$FC,$DF,$83,$F8,$FF
	.dc.b	$09,$F2,$1B,$1F,$47,$DF,$F9,$7F,$05,$F3,$1B,$1F,$57,$E0,$3E,$79
	.dc.b	$F4,$3E,$02,$C7,$53,$ED,$61,$D5,$A7,$53,$EF,$5F,$56,$9F,$81,$F8
	.dc.b	$4F,$B6,$7D,$CF,$84,$B1,$F8,$5F,$84,$FB,$FF,$09,$F8,$06,$C7,$F4
	.dc.b	$3F,$03,$FA,$E8,$F8,$4F,$F1,$D4,$7F,$69,$8C,$C6,$39,$FF,$B7,$C7
	.dc.b	$F6,$29,$D7,$C0,$67,$D8,$4F,$F6,$4A,$DC,$F8,$6B,$F3,$18,$D9,$FC
	.dc.b	$B3,$C6,$BB,$CF,$97,$E1,$F1,$8E,$7E,$07,$5F,$09,$6B,$D4,$F1,$F8
	.dc.b	$4E,$5F,$09,$67,$DC,$F5,$F8,$39,$1F,$07,$E1,$2E,$BF,$D2,$74,$FE
	.dc.b	$BB,$3F,$17,$E0,$7C,$9F,$03,$E5,$7A,$56,$F1,$BD,$FF,$E6,$8C,$11
	.dc.b	$D2,$31,$97,$B7,$2C,$65,$F3,$7E,$07,$E0,$7C,$27,$C1,$6C,$7D,$5D
	.dc.b	$7D,$6D,$3D,$67,$D9,$BE,$79,$FD,$4C,$2F,$27,$D2,$F8,$3F,$98,$7C
	.dc.b	$DF,$84,$B8,$7E,$07,$E0,$FE,$81,$F4,$7E,$12,$C7,$E1,$7E,$13,$EA
	.dc.b	$1F,$57,$E0,$2C,$7E,$27,$E0,$3F,$01,$FB,$C7,$F4,$7C,$FF,$03,$18
	.dc.b	$D7,$99,$F3,$72,$ED,$7E,$13,$B1,$23,$E9,$61,$D9,$FE,$F9,$87,$FB
	.dc.b	$9F,$8B,$FD,$2A,$FF,$6C,$8C,$FC,$2F,$8B,$E2,$7C,$7F,$0B,$31,$FF
	.dc.b	$10,$5F,$FB,$F5,$B1,$E8,$F8,$4F,$4F,$C2,$59,$DB,$EC,$F8,$19,$1C
	.dc.b	$BE,$06,$E7,$E4,$C3,$B3,$E6,$F8,$5B,$AF,$9B,$2E,$B8,$78,$9B,$5E
	.dc.b	$CF,$82,$F9,$3F,$0B,$F2,$9B,$5E,$8F,$84,$F9,$BF,$0B,$F3,$9B,$1E
	.dc.b	$CF,$85,$FA,$27,$D2,$F8,$5B,$1F,$03,$E1,$7E,$AF,$C2,$7D,$66,$C7
	.dc.b	$53,$E2,$E1,$D5,$A7,$52,$67,$F4,$AF,$AB,$F5,$3E,$13,$F0,$27,$E0
	.dc.b	$BE,$12,$D7,$59,$FB,$37,$D5,$B1,$D3,$7F,$B7,$9F,$A6,$30,$7F,$9E
	.dc.b	$79,$5F,$0B,$F8,$E3,$F1,$FE,$3D,$9F,$BB,$F0,$BF,$C4,$7F,$79,$FE
	.dc.b	$C7,$97,$FA,$13,$0F,$EC,$79,$7F,$B7,$27,$3D,$E7,$FB,$16,$1F,$E8
	.dc.b	$76,$DF,$FA,$31,$FE,$C8,$FF,$93,$5C,$CF,$1F,$7B,$C0,$78,$7E,$0A
	.dc.b	$DF,$90,$FF,$38,$57,$83,$1F,$1C,$77,$3D,$1F,$07,$1D,$FC,$0D,$CB
	.dc.b	$C1,$F0,$9E,$DF,$85,$F7,$36,$BA,$FC,$27,$C1,$F8,$5F,$84,$D8,$F0
	.dc.b	$FC,$27,$C5,$3E,$37,$C2,$58,$F2,$FC,$27,$C9,$3E,$57,$C0,$D8,$EA
	.dc.b	$7A,$B0,$EA,$D3,$A9,$33,$F8,$D7,$D5,$F9,$1F,$09,$F5,$4F,$AD,$F0
	.dc.b	$56,$BE,$2F,$C2,$7D,$9F,$84,$FB,$4D,$8F,$93,$F0,$9F,$74,$FB,$DF
	.dc.b	$09,$63,$E6,$FC,$27,$E0,$4F,$C1,$7C,$1D,$8E,$A7,$D2,$C3,$AB,$4E
	.dc.b	$A4,$CF,$EF,$5F,$57,$FD,$19,$FB,$D7,$FB,$0B,$A1,$87,$F5,$F8,$C3
	.dc.b	$D2,$7F,$0D,$85,$EC,$FC,$83,$CD,$FC,$96,$7E,$AC,$BC,$10,$7F,$DB
	.dc.b	$18,$7F,$73,$7F,$D7,$0D,$BF,$1F,$C2,$F8,$7E,$13,$C5,$F1,$3E,$3F
	.dc.b	$85,$B8,$70,$7C,$DF,$0B,$E7,$F8,$0B,$3D,$1F,$56,$12,$3B,$FC,$1D
	.dc.b	$CB,$E0,$E3,$D6,$7C,$6D,$8F,$8B,$87,$59,$F3,$B6,$3F,$D7,$3A,$3E
	.dc.b	$05,$E6,$7F,$82,$E3,$0D,$87,$9B,$63,$1E,$76,$7C,$EB,$97,$FB,$87
	.dc.b	$0E,$91,$C7,$CE,$E5,$E6,$5A,$D6,$5F,$C2,$9F,$96,$FD,$9F,$81,$FB
	.dc.b	$4D,$8F,$9F,$F0,$9F,$74,$FB,$DF,$0B,$63,$E9,$FC,$27,$E0,$4F,$C1
	.dc.b	$7C,$2D,$8E,$A7,$D8,$C3,$AB,$4E,$A4,$E0,$FC,$6D,$F5,$7F,$1F,$7C
	.dc.b	$DF,$C8,$DD,$D7,$F0,$D1,$FE,$FE,$FC,$0D,$E7,$7C,$B5,$E5,$58,$C3
	.dc.b	$9F,$2E,$67,$F2,$69,$E5,$E3,$0F,$AF,$0F,$23,$A1,$D7,$CB,$B7,$C7
	.dc.b	$C9,$F0,$79,$7E,$16,$C7,$5F,$2F,$C9,$E5,$F9,$5B,$1E,$1F,$2F,$D0
	.dc.b	$7A,$7C,$BB,$1E,$5F,$1F,$D9,$E3,$FB,$5B,$0D,$9C,$7F,$11,$7E,$2B
	.dc.b	$F7,$DF,$5F,$97,$FE,$59,$73,$D9,$F1,$E3,$E4,$79,$93,$0F,$89,$E5
	.dc.b	$DB,$FC,$CF,$E0,$8F,$89,$FC,$C3,$6B,$F2,$4F,$ED,$73,$FA,$A1,$D5
	.dc.b	$79,$A7,$D0,$F8,$9F,$A5,$8F,$97,$3F,$45,$F2,$E6,$4F,$83,$C9,$F0
	.dc.b	$F9,$17,$0F,$0B,$E4,$F2,$E4,$79,$3C,$9B,$9F,$84,$F9,$19,$F6,$9F
	.dc.b	$95,$9B,$3E,$F1,$83,$CF,$3B,$AF,$C8,$67,$F0,$BC,$BF,$87,$E5,$DB
	.dc.b	$F8,$6F,$C6,$F1,$BE,$3F,$8B,$67,$E3,$BE,$FE,$4F,$A5,$1D,$F1,$F2
	.dc.b	$C6,$1F,$A3,$E3,$7D,$0F,$1B,$E8,$B6,$BE,$8F,$8D,$F5,$3C,$6F,$AA
	.dc.b	$D8,$E9,$7F,$3C,$FC,$1B,$8D,$BD,$9C,$1F,$07,$7F,$BB,$F1,$32,$B5
	.dc.b	$A5,$C3,$F0,$7F,$13,$F8,$2F,$86,$FC,$1B,$6B,$67,$DC,$2C,$F8,$0F
	.dc.b	$A4,$78,$F1,$FC,$3F,$43,$1F,$89,$5F,$8C,$6D,$70,$CF,$FD,$5B,$3D
	.dc.b	$2F,$F4,$11,$89,$D6,$5F,$26,$72,$DA,$FB,$6F,$AE,$33,$1D,$CE,$87
	.dc.b	$E0,$3D,$8F,$0C,$76,$6C,$7E,$23,$97,$F0,$F7,$E0,$8F,$80,$D3,$F5
	.dc.b	$38,$F0,$7F,$6F,$97,$BE,$7B,$C7,$E5,$2F,$4F,$EE,$9E,$18,$7D,$59
	.dc.b	$DB,$67,$B3,$F2,$AF,$77,$D6,$73,$5E,$B3,$BB,$D5,$E7,$87,$AA,$FE
	.dc.b	$1C,$FD,$76,$9D,$67,$C3,$5D,$63,$31,$D6,$7B,$9E,$B7,$AE,$BE,$AB
	.dc.b	$ED,$1F,$48,$FA,$13,$C2,$BF,$84,$61,$F0,$B3,$FB,$91,$DA,$7A,$1B
	.dc.b	$BE,$6F,$EA,$2F,$E8,$CE,$CE,$6F,$A2,$FA,$65,$23,$CA,$4A,$F1,$65
	.dc.b	$ED,$8F,$34,$7D,$A7,$E4,$E3,$E4,$73,$87,$E1,$5E,$4F,$87,$1F,$2D
	.dc.b	$7E,$7E,$7F,$87,$DF,$58,$F4,$1D,$4F,$92,$7D,$52,$DF,$77,$77,$FC
	.dc.b	$97,$2F,$98,$7E,$03,$29,$87,$9B,$2F,$75,$FC,$7A,$DC,$FC,$32,$47
	.dc.b	$69,$DD,$FA,$6B,$73,$F1,$2F,$AB,$BB,$EC,$7C,$03,$E1,$9F,$A5,$9D
	.dc.b	$D7,$D3,$C7,$E0,$57,$C3,$3E,$DC,$7D,$D7,$E0,$3E,$1A,$E9,$1F,$01
	.dc.b	$F9,$37,$D1,$EF,$1C,$9A,$F2,$73,$3F,$A7,$3A,$47,$AE,$3E,$0C,$78
	.dc.b	$71,$F9,$D3,$E5,$39,$1E,$13,$FA,$0C,$F9,$1F,$54,$FE,$39,$E8,$7C
	.dc.b	$52,$3E,$02,$47,$92,$FE,$1D,$7F,$32,$6E,$7F,$D8,$F2,$FA,$75,$EA
	.dc.b	$CF,$F1,$77,$FC,$E1,$B5,$FC,$6E,$7A,$65,$C3,$2F,$41,$EA,$3C,$F5
	.dc.b	$C2,$7F,$4E,$70,$3E,$19,$F2,$AF,$E1,$9F,$D8,$67,$93,$E1,$CF,$F0
	.dc.b	$B9,$FD,$DC,$3E,$A5,$78,$2B,$8E,$3E,$BC,$3E,$44,$71,$7B,$CF,$CD
	.dc.b	$BF,$94,$7F,$44,$D3,$FA,$64,$7F,$50,$9F,$F0,$7E,$BE,$07,$E2,$65
	.dc.b	$E0,$9F,$A7,$AF,$EB,$70,$FF,$09,$6D,$FD,$8E,$7F,$C3,$6D,$8F,$A7
	.dc.b	$1E,$1D,$7F,$85,$C7,$8B,$0F,$C7,$9E,$67,$FC,$39,$C7,$FD,$F2,$3F
	.dc.b	$C4,$2D,$AF,$D7,$47,$E6,$27,$FC,$45,$C3,$FD,$DE,$7F,$BD,$CF,$F8
	.dc.b	$A3,$8B,$FC,$2B,$5F,$D8,$B9,$FF,$C5,$BC,$F8,$AF,$69,$CF,$47,$3F
	.dc.b	$F8,$42,$3F,$C6,$2D,$C7,$FC,$6F,$D5,$FE,$36,$3F,$C6,$FE,$85,$AF
	.dc.b	$A3,$C9,$FD,$8E,$BF,$A9,$F5,$FF,$8B,$27,$FC,$76,$D8,$FF,$31,$F7
	.dc.b	$7F,$94,$1B,$3F,$E6,$9E,$FF,$A5,$1F,$27,$93,$FC,$91,$BF,$F8,$E7
	.dc.b	$BF,$FC,$E9,$DF,$8C,$3F,$B4,$78,$9F,$E6,$6C,$3F,$CD,$ED,$BF,$F2
	.dc.b	$07,$87,$FE,$66,$9F,$F3,$9B,$63,$FC,$89,$E1,$FF,$99,$A7,$FC,$EF
	.dc.b	$E1,$D9,$FF,$24,$79,$7F,$E8,$0C,$7F,$CF,$CD,$AF,$F3,$EF,$95,$FE
	.dc.b	$7F,$AF,$F2,$77,$95,$6F,$FB,$E7,$95,$FE,$7E,$BF,$F4,$1F,$95,$6F
	.dc.b	$FD,$23,$EB,$BD,$2F,$E7,$4D,$F0,$FF,$DC,$9E,$2F,$FB,$95,$FF,$73
	.dc.b	$78,$9F,$EE,$7F,$89,$BA,$7F,$B7,$1F,$F7,$57,$C4,$FF,$BA,$FE,$22
	.dc.b	$CF,$FB,$81,$E4,$FF,$B8,$AF,$AB,$4F,$F7,$2D,$F5,$69,$D4,$FF,$74
	.dc.b	$D7,$56,$9D,$4E,$5A,$FF,$BF,$BE,$27,$BB,$63,$9B,$FE,$69,$CF,$E0
	.dc.b	$65,$EA,$8C,$68,$8F,$4C,$63,$3B,$AF,$C8,$A4,$7F,$B7,$70,$FF,$6A
	.dc.b	$FA,$BF,$EE,$1F,$5B,$18,$7F,$BA,$3C,$87,$AD,$FE,$E1,$BF,$F6,$DB
	.dc.b	$67,$5E,$AF,$FB,$82,$7F,$DB,$ED,$8F,$46,$DF,$EA,$79,$FF,$37,$6C
	.dc.b	$8C,$4E,$BE,$27,$FD,$33,$5F,$EA,$DA,$D4,$E0,$3C,$7E,$27,$0F,$3F
	.dc.b	$86,$B8,$26,$C6,$3E,$1A,$61,$E2,$A7,$C3,$63,$1F,$3F,$C4,$F2,$39
	.dc.b	$B6,$B8,$63,$D1,$A7,$3C,$79,$71,$FE,$5B,$16,$5E,$4C,$B3,$7E,$D9
	.dc.b	$7C,$4F,$A3,$E2,$7D,$2D,$9F,$37,$C4,$7B,$0F,$6F,$C4,$D8,$F5,$7C
	.dc.b	$4F,$C0,$3E,$0F,$C4,$D8,$F7,$7C,$4F,$C4,$3A,$70,$7D,$47,$93,$F5
	.dc.b	$B0,$EB,$3E,$03,$A9,$23,$ED,$D7,$57,$EE,$F4,$7D,$0E,$8F,$A2,$DA
	.dc.b	$F9,$DC,$FF,$53,$E0,$FE,$AB,$63,$E9,$74,$7D,$83,$EC,$FA,$16,$3E
	.dc.b	$B7,$B9,$F7,$0F,$BB,$F1,$36,$3A,$9F,$6F,$0E,$AD,$3A,$93,$3F,$C9
	.dc.b	$DF,$57,$F5,$9F,$BD,$A7,$F1,$6D,$7F,$2E,$BE,$23,$31,$DA,$7F,$3B
	.dc.b	$77,$A3,$F0,$0F,$57,$F0,$59,$FB,$2B,$C5,$1F,$C6,$B0,$7C,$47,$83
	.dc.b	$E2,$BC,$2D,$AF,$0F,$99,$E4,$F8,$8F,$2B,$63,$CB,$E6,$FA,$3C,$DF
	.dc.b	$4B,$63,$D3,$E6,$FB,$3C,$DF,$6B,$63,$A9,$EE,$C3,$AB,$4E,$A7,$C2
	.dc.b	$BE,$AD,$3E,$2F,$A3,$F2,$3E,$23,$E4,$B6,$3E,$4F,$A5,$F3,$3E,$23
	.dc.b	$E6,$B6,$3E,$6F,$A7,$F4,$0F,$A3,$E9,$D8,$FA,$3E,$C7,$D4,$3E,$AF
	.dc.b	$B1,$63,$A9,$F5,$B0,$EA,$D3,$A9,$39,$FE,$FD,$F5,$7F,$03,$EC,$7E
	.dc.b	$00,$FC,$0F,$B1,$6B,$F0,$3E,$C7,$F0,$EF,$DF,$1E,$E7,$E1,$A3,$0E
	.dc.b	$7D,$C9,$5F,$C7,$A9,$EE,$63,$2F,$E7,$50,$F8,$9E,$9F,$03,$D5,$B7
	.dc.b	$E0,$39,$63,$DD,$A7,$5C,$7A,$77,$7F,$64,$C3,$BC,$E4,$F1,$7C,$0F
	.dc.b	$A7,$E0,$3D,$4D,$8E,$D9,$F9,$31,$EC,$D8,$E4,$FA,$31,$F8,$5A,$78
	.dc.b	$27,$F9,$B6,$2F,$89,$F8,$C7,$C7,$F6,$AC,$FC,$3F,$89,$F9,$5F,$0B
	.dc.b	$F2,$DB,$1F,$1F,$E1,$7E,$77,$C3,$FC,$F6,$C7,$CB,$F8,$9F,$A5,$F0
	.dc.b	$DF,$4D,$B1,$D4,$FA,$18,$75,$69,$D4,$FA,$97,$D5,$A7,$D7,$F8,$6F
	.dc.b	$BD,$F0,$DF,$7D,$B1,$F6,$FE,$17,$F0,$47,$E0,$FE,$1A,$C7,$DF,$F8
	.dc.b	$5F,$C3,$7B,$DF,$87,$6C,$7F,$60,$FD,$F7,$FA,$FF,$4F,$F8,$69,$B1
	.dc.b	$FD,$7D,$7C,$37,$F5,$C7,$3C,$FF,$D8,$39,$7F,$BF,$57,$8C,$FC,$76
	.dc.b	$5E,$29,$FC,$6E,$3E,$06,$CF,$E4,$F2,$E9,$F0,$92,$3C,$9F,$09,$6B
	.dc.b	$FC,$1D,$97,$F5,$89,$F4,$7C,$27,$8F,$E1,$3C,$8F,$97,$E1,$7C,$DF
	.dc.b	$09,$6B,$DC,$FA,$7E,$17,$D5,$F0,$56,$7E,$13,$C9,$EA,$7C,$5C,$31
	.dc.b	$97,$8B,$0E,$B3,$F2,$5B,$1E,$6F,$85,$F8,$C7,$C7,$F8,$5B,$5D,$27
	.dc.b	$D7,$9F,$A6,$31,$C9,$F4,$70,$E6,$D8,$F4,$9F,$5F,$E1,$3E,$91,$F4
	.dc.b	$FE,$16,$C7,$DB,$F8,$4F,$AD,$F0,$BF,$5D,$B1,$F7,$FE,$17,$ED,$1F
	.dc.b	$6F,$E0,$6C,$7E,$0F,$E0,$7E,$F7,$C1,$FD,$F6,$C7,$53,$F1,$18,$75
	.dc.b	$69,$D4,$9B,$7D,$8B,$EA,$FD,$AF,$85,$FC,$57,$C2,$FE,$2D,$B5,$FE
	.dc.b	$32,$F4,$7F,$AF,$57,$F9,$35,$FE,$BD,$19,$3F,$C5,$AE,$23,$5F,$05
	.dc.b	$8F,$83,$CB,$6B,$A7,$C1,$F0,$38,$FC,$05,$8E,$87,$87,$E1,$7A,$BD
	.dc.b	$9E,$58,$78,$F0,$EB,$3E,$77,$C7,$F0,$92,$3D,$5F,$05,$6B,$D5,$F0
	.dc.b	$7E,$8F,$84,$F4,$B6,$BE,$07,$C1,$FB,$3E,$13,$DA,$D8,$F8,$9F,$09
	.dc.b	$F0,$0F,$83,$F0,$36,$3E,$45,$F3,$C7,$BC,$7C,$BC,$79,$BF,$37,$1B
	.dc.b	$3F,$35,$F9,$7F,$09,$27,$E9,$7C,$15,$AE,$67,$D3,$BF,$4C,$F5,$3D
	.dc.b	$27,$52,$C7,$57,$EE,$7C,$27,$D8,$3E,$CF,$C2,$58,$FC,$07,$C2,$7D
	.dc.b	$CF,$84,$FB,$AD,$8F,$C2,$73,$FE,$00,$FF,$36,$FE,$FD,$FD,$82,$32
	.dc.b	$7F,$9E,$97,$C0,$6B,$E0,$36,$D9,$EE,$7E,$3F,$0E,$ED,$39,$E7,$F9
	.dc.b	$1C,$3B,$FC,$0F,$86,$7B,$7C,$27,$8B,$E1,$7C,$6D,$8F,$17,$C2,$79
	.dc.b	$8F,$3F,$C2,$D8,$F3,$7C,$2F,$AB,$E1,$7D,$6D,$8F,$57,$C2,$FB,$BE
	.dc.b	$0B,$DE,$D8,$EA,$4D,$7F,$D6,$5C,$2C,$7B,$46,$30,$E8,$DA,$F8,$D8
	.dc.b	$FC,$AF,$85,$91,$F2,$7E,$0E,$D7,$4F,$84,$F9,$E7,$D0,$F8,$5B,$5F
	.dc.b	$3F,$E1,$3E,$99,$F5,$3E,$16,$C7,$D3,$F8,$5F,$AE,$7D,$8F,$83,$B1
	.dc.b	$F5,$F0,$E7,$8F,$AB,$1B,$EB,$7F,$BB,$38,$19,$F4,$AF,$7E,$76,$B5
	.dc.b	$F0,$7F,$84,$F8,$3F,$C2,$B7,$3E,$C4,$BE,$CD,$AE,$D8,$73,$8F,$F5
	.dc.b	$F6,$2C,$71,$7F,$EB,$EF,$DF,$FF,$B0,$FC,$41,$FE,$CB,$7F,$EC,$A7
	.dc.b	$26,$27,$FD,$CC,$FF,$B0,$63,$5F,$0B,$FE,$C6,$87,$C2,$D8,$C7,$BD
	.dc.b	$9F,$81,$B7,$D5,$E1,$EF,$C8,$F0,$7C,$25,$CB,$BE,$1C,$E7,$C9,$8D
	.dc.b	$9F,$1E,$3D,$67,$D0,$D8,$F3,$FC,$27,$94,$F3,$7C,$25,$AF,$77,$C2
	.dc.b	$7A,$4F,$57,$C2,$58,$F8,$5F,$09,$ED,$3D,$DF,$09,$63,$E3,$7C,$27
	.dc.b	$C1,$F8,$1F,$84,$D8,$EA,$4C,$FE,$4E,$1D,$5F,$97,$87,$37,$E6,$E1
	.dc.b	$79,$FF,$D5,$11,$F5,$FE,$13,$E7,$7C,$17,$CF,$6E,$29,$57,$01,$F2
	.dc.b	$5F,$E1,$C8,$CC,$EE,$73,$E0,$7F,$C0,$17,$7E,$05,$BE,$11,$CD,$76
	.dc.b	$E4,$D7,$75,$C3,$F1,$1F,$09,$F8,$33,$F0,$9F,$0B,$63,$F1,$9F,$09
	.dc.b	$F8,$73,$F1,$1F,$0B,$63,$C8,$7A,$FB,$DF,$89,$FF,$6A,$D3,$E1,$C6
	.dc.b	$27,$DC,$6E,$7F,$B5,$4E,$9E,$11,$9F,$86,$DF,$02,$FA,$C6,$1E,$FF
	.dc.b	$07,$6F,$FD,$93,$7E,$82,$5F,$E8,$B1,$F4,$3F,$30,$E9,$3E,$06,$CF
	.dc.b	$3F,$84,$E8,$F5,$F8,$68,$FC,$09,$E3,$F0,$D8,$7F,$5F,$AF,$6D,$7E
	.dc.b	$0F,$8A,$CF,$18,$7C,$13,$D1,$F0,$DD,$9E,$9B,$7C,$A8,$F6,$7C,$34
	.dc.b	$9C,$EC,$F8,$3C,$69,$EB,$CF,$F1,$3E,$B7,$E0,$CF,$C2,$7A,$F7,$0F
	.dc.b	$44,$FE,$3B,$D7,$FC,$47,$AB,$F8,$96,$CF,$6F,$80,$F9,$9F,$0D,$F3
	.dc.b	$5B,$1C,$B5,$7B,$FF,$93,$5E,$ED,$8C,$3F,$97,$FB,$DF,$45,$78,$9B
	.dc.b	$B7,$A7,$BF,$FB,$63,$EB,$8C,$EB,$F8,$5F,$85,$FB,$C7,$DF,$F8,$7B
	.dc.b	$87,$8D,$F1,$7C,$3F,$8D,$B1,$DD,$FC,$34,$78,$DB,$1F,$33,$E1,$FF
	.dc.b	$15,$F0,$FF,$8B,$6C,$7B,$F9,$FD,$C7,$BF,$7B,$1F,$53,$E1,$FF,$25
	.dc.b	$F0,$FF,$93,$6C,$7D,$8E,$1F,$CB,$70,$FE,$5D,$B1,$F2,$CF,$95,$E0
	.dc.b	$7C,$B6,$9F,$80,$F8,$5F,$D0,$76,$FC,$6E,$1F,$D4,$46,$3E,$17,$3F
	.dc.b	$0D,$AF,$87,$B8,$7E,$33,$0F,$F7,$1F,$37,$E1,$FE,$59,$E7,$7F,$0B
	.dc.b	$6D,$7E,$19,$79,$F9,$9F,$0F,$81,$2B,$A7,$81,$70,$FD,$67,$87,$F9
	.dc.b	$FF,$03,$CE,$7A,$3C,$1B,$7D,$7D,$5F,$01,$E1,$F4,$EC,$79,$FE,$1B
	.dc.b	$DE,$7C,$0F,$86,$B1,$EB,$F8,$7F,$86,$7C,$4F,$88,$B1,$E9,$F7,$3D
	.dc.b	$9F,$03,$ED,$6C,$7C,$3F,$88,$F9,$67,$CC,$F8,$8B,$1F,$1F,$E2,$3E
	.dc.b	$79,$F4,$3E,$22,$C7,$EA,$BE,$5F,$F9,$6B,$7D,$9F,$E6,$91,$8E,$4F
	.dc.b	$DA,$28,$FB,$3F,$13,$2B,$EA,$7C,$45,$BF,$93,$D9,$FC,$EA,$7F,$96
	.dc.b	$36,$3F,$91,$2F,$77,$F0,$7E,$F7,$83,$49,$7E,$5D,$2E,$1D,$3D,$EF
	.dc.b	$C4,$E9,$F8,$A6,$D7,$F1,$0F,$E0,$0F,$F0,$69,$3E,$37,$F9,$73,$FE
	.dc.b	$F0,$F1,$9C,$0E,$7F,$79,$FD,$4F,$9F,$FC,$21,$3F,$D6,$E7,$58,$75
	.dc.b	$F8,$8E,$7F,$11,$D1,$B1,$E1,$E7,$EF,$D1,$E0,$6C,$7F,$18,$C5,$F1
	.dc.b	$1E,$4F,$8A,$F0,$36,$7D,$5F,$15,$E8,$3D,$31,$63,$DD,$F1,$5E,$C3
	.dc.b	$DB,$E0,$58,$F8,$5E,$07,$C0,$F8,$8F,$82,$D8,$F8,$DF,$11,$F1,$0F
	.dc.b	$8B,$F1,$36,$3E,$57,$C5,$7C,$8F,$2B,$E4,$B6,$3E,$77,$97,$F3,$0F
	.dc.b	$9B,$F1,$56,$3E,$97,$C5,$7D,$0F,$86,$FA,$2D,$8F,$AD,$E1,$FD,$43
	.dc.b	$FB,$17,$F0,$0F,$F5,$9F,$35,$DB,$FD,$3E,$BF,$A5,$CF,$F6,$E7,$FB
	.dc.b	$B3,$7C,$1F,$ED,$8F,$F6,$87,$F7,$2E,$68,$BE,$2B,$BF,$C5,$78,$1B
	.dc.b	$7D,$BE,$2B,$C7,$EF,$79,$1B,$1E,$2F,$7B,$CE,$7A,$3E,$22,$C7,$9B
	.dc.b	$E1,$FD,$67,$B3,$E2,$2C,$7A,$BE,$1B,$DE,$7C,$0F,$85,$B1,$EE,$F8
	.dc.b	$6F,$87,$F0,$BF,$11,$B1,$F0,$BE,$1B,$E3,$FC,$2F,$C8,$6C,$7C,$6F
	.dc.b	$86,$F9,$67,$CC,$F8,$5B,$1F,$2B,$E1,$BE,$7F,$C2,$FD,$06,$C7,$F7
	.dc.b	$8F,$DC,$7F,$9A,$3F,$E2,$37,$FD,$86,$E7,$E1,$BF,$C3,$11,$8E,$9F
	.dc.b	$E4,$EB,$E1,$7F,$C8,$91,$86,$FA,$5F,$E2,$A9,$D4,$F7,$F8,$4C,$E1
	.dc.b	$DF,$E1,$2C,$FF,$89,$E9,$F0,$98,$EE,$46,$27,$D1,$59,$8F,$F5,$D4
	.dc.b	$7A,$FE,$0F,$06,$7E,$13,$5E,$C6,$FD,$39,$5F,$06,$39,$7C,$24,$8F
	.dc.b	$88,$F5,$F8,$4E,$DF,$09,$27,$E4,$3E,$1F,$80,$F1,$7C,$04,$9F,$98
	.dc.b	$F9,$7E,$13,$CD,$F0,$9E,$7F,$83,$95,$F4,$63,$D5,$F0,$9E,$BF,$83
	.dc.b	$93,$F5,$63,$DD,$F0,$9F,$1B,$F7,$DF,$9F,$3F,$E3,$B6,$BF,$52,$34
	.dc.b	$BE,$16,$FC,$1C,$8E,$06,$4B,$F0,$9C,$08,$EB,$E0,$F9,$FC,$24,$8E
	.dc.b	$2F,$6F,$84,$EF,$F0,$9E,$0F,$84,$95,$DA,$3C,$7F,$09,$23,$C2,$F9
	.dc.b	$BE,$13,$CF,$F0,$7E,$8F,$84,$95,$E6,$8F,$5F,$C2,$48,$F4,$BE,$EF
	.dc.b	$84,$F7,$FC,$1F,$C0,$F8,$49,$5E,$E8,$F8,$7F,$09,$23,$E0,$BF,$1B
	.dc.b	$E1,$3E,$3F,$C1,$C9,$F8,$AF,$CA,$F8,$4F,$97,$F0,$9F,$47,$ED,$BF
	.dc.b	$D4,$F5,$FE,$73,$EA,$46,$3C,$FF,$F5,$5C,$7F,$A9,$EB,$87,$C4,$F1
	.dc.b	$F8,$E9,$87,$D9,$C1,$3F,$A2,$C1,$F9,$4F,$F1,$9A,$FF,$30,$FC,$03
	.dc.b	$E0,$F4,$D8,$FF,$32,$CE,$DF,$26,$16,$FC,$6D,$F4,$FF,$D7,$5A,$7A
	.dc.b	$3E,$17,$D9,$F0,$DE,$DC,$FD,$DF,$0D,$32,$F3,$47,$97,$E8,$9C,$77
	.dc.b	$8E,$57,$F9,$DA,$72,$FB,$7E,$1F,$E4,$11,$F8,$1F,$0F,$F2,$F8,$7E
	.dc.b	$67,$C3,$C9,$F8,$91,$DF,$B7,$C1,$D9,$27,$B4,$7D,$3F,$87,$FA,$9F
	.dc.b	$0F,$27,$E6,$47,$D7,$F8,$69,$1F,$3D,$FB,$5F,$0F,$EB,$F1,$BD,$9E
	.dc.b	$2C,$AF,$54,$7D,$FF,$86,$FC,$07,$C3,$C9,$FB,$11,$F8,$3F,$87,$91
	.dc.b	$F6,$DF,$C3,$7C,$3F,$C7,$88,$7C,$53,$E5,$79,$FF,$8B,$F8,$69,$1F
	.dc.b	$83,$7F,$1D,$F0,$FF,$BF,$F0,$1F,$0F,$7E,$32,$5F,$F0,$19,$CB,$FE
	.dc.b	$A6,$D9,$EC,$FE,$2D,$B1,$98,$C7,$B1,$D7,$E2,$3B,$7C,$44,$AE,$91
	.dc.b	$E1,$F8,$69,$1D,$DF,$27,$C4,$75,$F8,$4E,$DF,$07,$2B,$C5,$1E,$9F
	.dc.b	$85,$91,$E7,$7D,$9F,$0D,$ED,$F8,$6F,$75,$4A,$F6,$47,$A7,$39,$1E
	.dc.b	$D7,$D9,$97,$C5,$C7,$E3,$7C,$34,$AF,$89,$1F,$27,$E1,$BE,$57,$C3
	.dc.b	$49,$F9,$11,$FA,$FF,$A5,$FF,$61,$C7,$FB,$39,$FF,$61,$B7,$9F,$5D
	.dc.b	$3F,$4F,$E1,$E4,$FD,$17,$EB,$72,$7D,$7E,$4F,$B1,$F0,$F2,$BE,$B4
	.dc.b	$77,$F0,$FC,$1E,$14,$9F,$04,$78,$FB,$E4,$78,$DF,$37,$85,$F8,$3F
	.dc.b	$86,$93,$F8,$17,$F0,$DF,$0F,$EB,$F4,$24,$FA,$DF,$77,$99,$EF,$F3
	.dc.b	$24,$FB,$DF,$85,$E5,$FE,$3F,$E1,$E4,$FE,$35,$FE,$A1,$F7,$DF,$D4
	.dc.b	$3C,$3F,$E3,$AF,$F4,$99,$FE,$44,$BE,$23,$67,$0F,$23,$8F,$C4,$72
	.dc.b	$F8,$89,$7F,$D6,$6D,$3D,$BE,$23,$BF,$9D,$2B,$51,$CD,$E1,$8F,$67
	.dc.b	$96,$3D,$9E,$97,$5F,$2F,$93,$EA,$F8,$89,$1E,$17,$DB,$F1,$1F,$CC
	.dc.b	$7E,$9D,$E9,$D2,$50,$6E,$F7,$70,$A9,$EA,$F3,$F8,$79,$7A,$BD,$D2
	.dc.b	$7C,$1F,$0D,$F3,$BE,$26,$5F,$C5,$7E,$8F,$C4,$FD,$2C,$FE,$9E,$52
	.dc.b	$BE,$54,$7D,$6C,$24,$7C,$D7,$EC,$E1,$F6,$BE,$26,$4F,$D1,$7E,$EF
	.dc.b	$C4,$F5,$8F,$A9,$87,$57,$EE,$E1,$D1,$FB,$F8,$5D,$7F,$07,$8F,$E1
	.dc.b	$BD,$FF,$C3,$FB,$F2,$FE,$F4,$7E,$2B,$E2,$64,$7E,$05,$FC,$6F,$C4
	.dc.b	$FF,$87,$7D,$BF,$ED,$5C,$5F,$D5,$9B,$4F,$85,$FE,$47,$3B,$F8,$8E
	.dc.b	$1F,$13,$2B,$FC,$49,$68,$E9,$F0,$FE,$09,$FC,$BE,$57,$6F,$CD,$E7
	.dc.b	$E1,$CA,$57,$27,$B3,$D3,$E1,$FC,$C4,$7B,$FC,$3F,$39,$FE,$DB,$E0
	.dc.b	$B1,$C5,$7A,$61,$87,$68,$FF,$05,$69,$EE,$DB,$DB,$1E,$AD,$BE,$27
	.dc.b	$8F,$F1,$7C,$79,$3E,$E8,$F9,$1E,$3F,$C9,$F1,$E4,$FC,$28,$F9,$9F
	.dc.b	$0D,$F3,$7E,$1A,$4F,$C6,$8F,$A1,$E7,$7D,$1F,$3E,$4F,$CA,$8E,$AF
	.dc.b	$CC,$C3,$AB,$E8,$C3,$AB,$F4,$30,$EA,$FB,$30,$FB,$9E,$74,$8F,$AA
	.dc.b	$FD,$FF,$3F,$F0,$1F,$0D,$27,$EC,$BF,$83,$F8,$6F,$C2,$7A,$DF,$85
	.dc.b	$F5,$E5,$7D,$E8,$FC,$47,$C3,$7F,$99,$7F,$0F,$FE,$C7,$3F,$E3,$56
	.dc.b	$F6,$7F,$8D,$39,$95,$DE,$BE,$8E,$56,$B5,$97,$82,$7F,$17,$97,$81
	.dc.b	$FA,$F9,$77,$F8,$7F,$07,$C3,$C9,$FF,$3A,$68,$9F,$27,$C1,$49,$CB
	.dc.b	$E7,$F8,$2F,$47,$C3,$FA,$7E,$1E,$57,$28,$F6,$7C,$3C,$8E,$AF,$BF
	.dc.b	$E1,$BA,$47,$BF,$0B,$BF,$F9,$4B,$36,$5D,$AB,$CB,$97,$67,$E3,$65
	.dc.b	$F2,$7E,$1F,$E5,$7C,$34,$9F,$5C,$74,$F1,$3F,$C9,$38,$B3,$C5,$7E
	.dc.b	$D6,$3B,$BE,$93,$B9,$0E,$E7,$C6,$F8,$7F,$B0,$47,$E4,$FC,$37,$DB
	.dc.b	$F8,$7F,$B9,$F0,$D2,$7E,$6C,$7D,$FF,$86,$91,$F4,$1F,$C1,$7C,$2F
	.dc.b	$E0,$FE,$1A,$4F,$D4,$7F,$0D,$F0,$DD,$63,$EB,$E1,$D5,$FC,$46,$1D
	.dc.b	$1F,$C5,$61,$78,$FE,$66,$3F,$8F,$F8,$69,$5F,$80,$7F,$25,$F0,$DF
	.dc.b	$EB,$7F,$E0,$1F,$ED,$1C,$3F,$E7,$07,$FB,$46,$1F,$E8,$66,$C6,$6F
	.dc.b	$87,$C2,$71,$F8,$39,$1F,$EB,$FA,$FE,$DC,$DF,$87,$ED,$F0,$D2,$30
	.dc.b	$6C,$F1,$7C,$34,$70,$73,$C3,$AB,$FE,$9B,$BF,$EE,$6E,$4F,$07,$C3
	.dc.b	$FA,$FE,$1E,$46,$0F,$27,$C3,$FB,$C8,$F9,$FE,$1F,$E1,$7C,$34,$8F
	.dc.b	$53,$F1,$7E,$1B,$E3,$7C,$1F,$C7,$F8,$49,$5E,$F8,$E6,$F9,$70,$BC
	.dc.b	$5F,$27,$1E,$93,$E9,$C6,$CF,$CD,$C7,$E9,$7C,$27,$D3,$F8,$49,$7F
	.dc.b	$2E,$39,$F2,$7D,$5C,$AC,$FF,$4F,$DD,$9F,$4A,$FB,$59,$DA,$D6,$5F
	.dc.b	$7F,$E1,$BF,$01,$F0,$D2,$7E,$C4,$7E,$0F,$E1,$64,$7D,$B7,$F0,$DF
	.dc.b	$0D,$F8,$7F,$82,$FC,$47,$C0,$CA,$FC,$04,$7F,$C0,$1F,$C3,$3F,$A3
	.dc.b	$13,$4C,$7C,$0F,$4A,$FB,$58,$DE,$3F,$C7,$65,$D6,$7F,$15,$97,$57
	.dc.b	$F0,$19,$75,$F8,$7E,$DF,$0F,$27,$FC,$30,$7F,$71,$D3,$FB,$63,$8F
	.dc.b	$84,$43,$E1,$FC,$DF,$0F,$2B,$53,$96,$FC,$2F,$AF,$E1,$A4,$60,$DF
	.dc.b	$C1,$FB,$FE,$12,$46,$0E,$78,$F5,$7F,$DE,$58,$7F,$B8,$DB,$1F,$1F
	.dc.b	$0E,$B1,$AC,$30,$F9,$3E,$0F,$E6,$11,$F3,$FC,$27,$CF,$F8,$5F,$A1
	.dc.b	$F0,$D2,$7D,$71,$F4,$FE,$1B,$EA,$7C,$34,$9F,$7C,$7D,$7F,$84,$FB
	.dc.b	$1F,$09,$27,$E1,$C7,$57,$E2,$E1,$D5,$F5,$E1,$D5,$F9,$38,$75,$7D
	.dc.b	$F8,$7E,$0F,$E0,$E4,$7C,$E7,$F0,$DF,$09,$D6,$3E,$2E,$1D,$7A,$BF
	.dc.b	$CB,$11,$F6,$F1,$EC,$FC,$AC,$7B,$4E,$B9,$7E,$EE,$5F,$91,$F8,$59
	.dc.b	$7F,$71,$FF,$32,$3F,$E6,$4E,$EF,$F3,$35,$7F,$99,$7E,$1B,$38,$73
	.dc.b	$9F,$EC,$95,$F8,$96,$BC,$4F,$F1,$F3,$88,$E1,$F0,$DC,$A3,$A9,$FD
	.dc.b	$46,$79,$7C,$1E,$5E,$2D,$C3,$FC,$ED,$EF,$DA,$7C,$15,$25,$F8,$FE
	.dc.b	$0F,$5F,$0D,$23,$CC,$F1,$CB,$96,$1C,$F2,$95,$EB,$8E,$DF,$0D,$DF
	.dc.b	$E1,$A4,$FB,$E3,$C5,$F0,$BE,$3F,$86,$93,$F0,$E3,$CD,$F0,$B2,$3E
	.dc.b	$33,$E9,$F8,$5F,$57,$C1,$7A,$FE,$0E,$57,$CB,$8E,$AF,$CD,$C3,$AB
	.dc.b	$F3,$78,$BF,$D3,$18,$F5,$8C,$71,$32,$C7,$1F,$D7,$CF,$E4,$7C,$1F
	.dc.b	$C9,$F8,$39,$87,$D9,$8F,$99,$F0,$B2,$3E,$E3,$F3,$FE,$17,$E8,$7C
	.dc.b	$2F,$D1,$F8,$59,$5F,$81,$8F,$AD,$FC,$03,$FD,$9A,$3F,$CF,$CD,$DC
	.dc.b	$D3,$FB,$3C,$7E,$1B,$0F,$03,$EA,$7C,$0F,$53,$F1,$D7,$E0,$7D,$F7
	.dc.b	$E0,$F8,$29,$1F,$C6,$EF,$FB,$6B,$F1,$DF,$EC,$6F,$53,$DA,$43,$E4
	.dc.b	$E3,$EE,$7B,$90,$EE,$75,$F8,$5F,$79,$1F,$07,$C2,$FC,$2F,$84,$F8
	.dc.b	$7F,$09,$27,$C9,$1F,$1B,$E0,$24,$79,$DF,$93,$EF,$FC,$AF,$7E,$4F
	.dc.b	$AD,$F9,$BE,$FF,$58,$F7,$61,$D5,$F7,$61,$D1,$F8,$18,$5E,$5F,$E0
	.dc.b	$5A,$FF,$62,$9F,$8D,$F0,$7F,$60,$8F,$C9,$F8,$4F,$B7,$F0,$72,$3E
	.dc.b	$63,$F7,$BE,$13,$EF,$FC,$17,$E0,$3E,$0E,$57,$D1,$8F,$C1,$FC,$17
	.dc.b	$E1,$30,$93,$F5,$63,$A3,$F5,$70,$B9,$F9,$B1,$EB,$3F,$6F,$1E,$AF
	.dc.b	$DB,$C7,$F1,$FF,$05,$23,$F0,$0F,$E4,$BE,$07,$F2,$7F,$05,$FC,$27
	.dc.b	$F0,$9F,$F5,$A1,$FE,$B5,$DF,$FD,$73,$7F,$D8,$61,$F0,$9B,$24,$67
	.dc.b	$E1,$3A,$3F,$8D,$C2,$BF,$B3,$A7,$C2,$78,$23,$A7,$8B,$E1,$30,$EF
	.dc.b	$0E,$AD,$3B,$1E,$39,$EB,$BF,$D8,$C3,$D5,$1C,$A3,$F9,$E6,$1E,$CC
	.dc.b	$30,$7F,$38,$7D,$77,$ED,$C3,$BD,$70,$C7,$E2,$7C,$07,$C5,$F8,$49
	.dc.b	$7F,$0E,$3E,$47,$C0,$FC,$9F,$81,$93,$F1,$E3,$E6,$7C,$24,$8F,$94
	.dc.b	$FC,$FF,$84,$EB,$1F,$37,$0E,$AF,$C7,$C3,$AB,$F4,$70,$EA,$FC,$BC
	.dc.b	$3A,$3F,$37,$0B,$97,$C0,$C7,$EE,$7C,$1C,$AF,$B4,$FD,$FF,$83,$FC
	.dc.b	$07,$BD,$27,$EF,$3F,$83,$F7,$FF,$09,$EF,$49,$FC,$13,$F8,$7F,$7F
	.dc.b	$A4,$7D,$DF,$82,$FC,$51,$8C,$FE,$8E,$1D,$67,$F0,$4F,$57,$99,$0F
	.dc.b	$C3,$5F,$57,$F2,$1F,$07,$F9,$5F,$83,$91,$83,$F2,$9F,$07,$F9,$92
	.dc.b	$3F,$97,$F8,$3F,$EC,$9F,$C0,$3F,$EE,$4E,$9F,$EE,$EF,$F6,$F6,$FC
	.dc.b	$0E,$BE,$13,$AD,$FF,$49,$A7,$C1,$72,$33,$B7,$E3,$32,$EF,$5F,$9D
	.dc.b	$7F,$55,$97,$38,$FF,$24,$5F,$75,$EB,$23,$0D,$C7,$CF,$F0,$32,$B8
	.dc.b	$BE,$AF,$82,$EF,$97,$86,$7E,$09,$EF,$3B,$D7,$C0,$D2,$46,$0F,$1F
	.dc.b	$C2,$7C,$42,$3E,$6F,$84,$F8,$FF,$07,$F2,$3E,$12,$4F,$AA,$3E,$5F
	.dc.b	$C0,$C8,$F6,$BF,$3B,$E0,$BE,$7F,$C1,$7D,$0F,$82,$95,$F0,$A3,$AB
	.dc.b	$F1,$30,$EA,$FC,$FC,$3A,$3F,$47,$0B,$AF,$F8,$D7,$85,$97,$DC,$F8
	.dc.b	$39,$7F,$39,$FB,$FF,$09,$F8,$0F,$83,$93,$F4,$9F,$C1,$FC,$27,$E1
	.dc.b	$3E,$12,$4F,$D6,$7F,$0F,$F0,$3F,$88,$F8,$39,$3F,$69,$FC,$5F,$C2
	.dc.b	$74,$8F,$C3,$61,$75,$F3,$63,$D2,$7F,$15,$8D,$9F,$56,$3F,$E3,$7F
	.dc.b	$87,$FF,$00,$12,$B0,$E7,$E1,$64,$7F,$A4,$63,$FB,$BD,$7F,$95,$DF
	.dc.b	$EE,$ED,$F8,$5E,$84,$9E,$1F,$0B,$DF,$E0,$64,$73,$7C,$5F,$05,$DA
	.dc.b	$3F,$C5,$18,$76,$F0,$FF,$CB,$AD,$8F,$CC,$BE,$08,$CE,$1D,$E7,$8B
	.dc.b	$ED,$C3,$DD,$F0,$5E,$FF,$85,$95,$EB,$8F,$06,$7E,$BD,$2E,$3E,$0F
	.dc.b	$88,$E9,$3E,$DB,$F8,$59,$75,$7E,$27,$C3,$7C,$BF,$86,$91,$83,$E4
	.dc.b	$7C,$37,$CF,$F8,$69,$18,$3E,$67,$C2,$FD,$3F,$85,$91,$83,$E8,$7C
	.dc.b	$27,$D7,$23,$F4,$FE,$17,$ED,$61,$D2,$3E,$4E,$18,$79,$1F,$32,$FA
	.dc.b	$BF,$3B,$0F,$C1,$7C,$34,$8F,$BC,$FE,$17,$E1,$BF,$0D,$F0,$92,$7F
	.dc.b	$04,$FE,$27,$E1,$BF,$D0,$1F,$C0,$3F,$E9,$E9,$FF,$30,$3F,$D7,$1B
	.dc.b	$F0,$9A,$E4,$DF,$C2,$70,$F8,$69,$5F,$ED,$DB,$FF,$6F,$BF,$82,$C7
	.dc.b	$C0,$D3,$F2,$58,$F3,$78,$63,$E2,$30,$F8,$BE,$13,$CB,$F0,$B2,$3B
	.dc.b	$47,$A3,$E1,$3D,$3F,$0B,$27,$C5,$1E,$CF,$85,$91,$E5,$7D,$FF,$0B
	.dc.b	$F0,$3E,$16,$4F,$A5,$F8,$7F,$0B,$D6,$3D,$98,$75,$7E,$2E,$1D,$1F
	.dc.b	$8F,$85,$ED,$FF,$61,$6E,$CB,$E6,$FC,$17,$CE,$F8,$59,$87,$C7,$8E
	.dc.b	$CF,$D0,$CB,$B6,$1C,$B2,$E9,$C9,$FE,$C6,$C9,$9E,$30,$FE,$87,$1D
	.dc.b	$DF,$A5,$F0,$7F,$73,$E1,$64,$60,$FA,$DF,$0B,$F8,$0F,$85,$91,$83
	.dc.b	$ED,$7C,$2F,$E1,$08,$FD,$DF,$84,$FC,$3F,$C2,$7F,$B1,$FE,$4B,$FB
	.dc.b	$24,$FF,$54,$6F,$C2,$B4,$57,$E3,$B2,$EF,$7F,$8E,$CB,$0F,$80,$87
	.dc.b	$E4,$B2,$FC,$AE,$5D,$FE,$0A,$48,$FF,$5B,$67,$FD,$88,$FE,$DE,$E7
	.dc.b	$87,$FD,$E9,$1F,$EE,$18,$DF,$C2,$FA,$4F,$57,$C2,$FA,$FE,$13,$D9
	.dc.b	$F0,$92,$B9,$72,$7F,$BD,$B0,$FF,$71,$C6,$0F,$81,$F0,$3F,$0F,$E0
	.dc.b	$A4,$67,$8D,$4A,$F0,$4E,$63,$C9,$A7,$9B,$4F,$1C,$E7,$3F,$13,$EE
	.dc.b	$CF,$CC,$F1,$F8,$2F,$A5,$F0,$32,$30,$75,$F8,$3F,$AD,$F0,$92,$30
	.dc.b	$78,$7E,$0F,$ED,$7C,$1C,$8C,$1E,$5F,$81,$FB,$C4,$7D,$1F,$03,$F8
	.dc.b	$1C,$3A,$47,$D2,$C3,$0F,$23,$EA,$DF,$57,$EB,$E1,$F8,$9F,$81,$FC
	.dc.b	$57,$C0,$C9,$F8,$91,$C9,$E8,$7C,$7B,$C7,$07,$DC,$BE,$8F,$DE,$C3
	.dc.b	$F2,$91,$F9,$5F,$83,$97,$F3,$DF,$CC,$7C,$1F,$E2,$63,$FD,$FD,$CB
	.dc.b	$FE,$E9,$3F,$DC,$E7,$F5,$67,$ED,$39,$F7,$78,$62,$E8,$FC,$94,$FF
	.dc.b	$8D,$3D,$DC,$C6,$23,$FA,$F5,$FF,$8E,$1F,$EC,$47,$F8,$EB,$DD,$DC
	.dc.b	$65,$B1,$98,$F0,$FC,$06,$23,$CD,$F0,$34,$C4,$E7,$2D,$63,$23,$D3
	.dc.b	$1C,$7E,$0F,$97,$C1,$C9,$F6,$C7,$5F,$82,$91,$F0,$1F,$07,$C1,$F8
	.dc.b	$7E,$0F,$C5,$F0,$72,$BE,$2C,$79,$7E,$07,$CD,$F0,$32,$7E,$4C,$74
	.dc.b	$7E,$7E,$17,$8B,$E5,$47,$C5,$C7,$B3,$F4,$31,$EC,$FD,$6C,$7E,$17
	.dc.b	$C0,$48,$FA,$AF,$C5,$F8,$1F,$8D,$F0,$7F,$1F,$E0,$E5,$7D,$A8,$F9
	.dc.b	$5F,$07,$F2,$FE,$0E,$4F,$DE,$8F,$9D,$F0,$3F,$3F,$E0,$64,$FE,$0A
	.dc.b	$3A,$BF,$84,$C3,$AB,$F8,$AC,$3A,$BF,$88,$C3,$AB,$F8,$EC,$3E,$EF
	.dc.b	$9D,$FD,$BF,$7F,$83,$D0,$8F,$EE,$11,$86,$F5,$FE,$39,$E0,$FE,$37
	.dc.b	$1F,$0C,$F2,$A6,$5E,$6C,$7C,$07,$8C,$86,$04,$F7,$F8,$3F,$2F,$C1
	.dc.b	$C9,$FE,$35,$97,$F7,$CA,$F2,$7C,$0F,$A8,$93,$AF,$82,$F6,$FB,$92
	.dc.b	$38,$BF,$03,$DD,$F8,$3E,$EF,$C2,$F8,$09,$5D,$A3,$9B,$E4,$E0,$F8
	.dc.b	$18,$E3,$2F,$89,$97,$5A,$D1,$FE,$A4,$CD,$9E,$6B,$E5,$67,$F4,$3E
	.dc.b	$0B,$E8,$FC,$14,$C3,$DB,$1F,$53,$E0,$64,$7C,$07,$EB,$FC,$17,$D8
	.dc.b	$F8,$29,$3F,$11,$FB,$7F,$05,$F7,$3E,$02,$4F,$C8,$7E,$FF,$C0,$74
	.dc.b	$8F,$3F,$3F,$DF,$C7,$1A,$FD,$FC,$BA,$D6,$B7,$67,$8A,$FC,$26,$7F
	.dc.b	$8A,$F8,$0F,$C5,$FC,$04,$C3,$EB,$C7,$E3,$BE,$07,$F8,$5F,$E0,$7F
	.dc.b	$DD,$0F,$EE,$B9,$7F,$74,$AF,$EA,$D1,$FE,$CA,$3F,$B4,$BE,$66,$E3
	.dc.b	$F9,$67,$5E,$F7,$4F,$82,$95,$CA,$3C,$0F,$4F,$86,$F0,$98,$7A,$61
	.dc.b	$E4,$F8,$79,$1E,$28,$F3,$FC,$0F,$A3,$E0,$64,$F9,$A3,$96,$7E,$3C
	.dc.b	$3A,$CF,$E5,$B4,$F7,$E1,$D2,$3D,$58,$E1,$F7,$FC,$1F,$C4,$23,$F0
	.dc.b	$BE,$13,$E3,$FC,$1F,$C8,$F8,$39,$3F,$1A,$3E,$5F,$C0,$FC,$CF,$7E
	.dc.b	$4F,$CA,$8F,$9F,$EF,$7D,$0F,$7E,$4F,$CE,$8E,$AF,$D0,$C3,$AB,$F3
	.dc.b	$F0,$EA,$FD,$4C,$3A,$BF,$4F,$0F,$B7,$F0,$7F,$73,$E0,$E4,$FD,$A8
	.dc.b	$FB,$FF,$05,$23,$EE,$BF,$82,$F8,$3F,$C1,$FC,$1C,$9F,$C0,$BF,$86
	.dc.b	$F8,$3F,$EA,$3F,$C0,$3F,$D9,$67,$F2,$EF,$F6,$42,$3D,$4F,$C3,$5D
	.dc.b	$E6,$FC,$36,$1D,$67,$8F,$C0,$74,$31,$3F,$C8,$2D,$F0,$3F,$D0,$A3
	.dc.b	$0D,$F3,$7F,$B6,$CB,$E1,$31,$83,$C6,$4F,$2A,$FF,$18,$2F,$84,$DC
	.dc.b	$EB,$E0,$AC,$E6,$B7,$18,$F8,$39,$1C,$BE,$06,$E3,$F0,$BE,$07,$8F
	.dc.b	$C0,$72,$6D,$78,$33,$EB,$8F,$66,$C7,$93,$1E,$7B,$7C,$1C,$AC,$FB
	.dc.b	$BE,$07,$CA,$79,$BE,$0A,$C7,$53,$CE,$5D,$3E,$AE,$5E,$CB,$E7,$1F
	.dc.b	$17,$54,$63,$6F,$8F,$F0,$7F,$0B,$E0,$FE,$1B,$71,$F9,$3F,$07,$F1
	.dc.b	$BD,$FF,$8E,$D8,$F9,$BE,$FF,$CA,$F8,$0F,$96,$D8,$FA,$3F,$03,$F3
	.dc.b	$BE,$0B,$E7,$B6,$3A,$9F,$5B,$0E,$AD,$3A,$9F,$6A,$FA,$B4,$FB,$BF
	.dc.b	$07,$F6,$8F,$BB,$FC,$03,$FE,$33,$8C,$9F,$E7,$15,$F1,$1A,$F8,$5D
	.dc.b	$B6,$75,$F0,$BC,$8E,$7F,$07,$63,$97,$C1,$F6,$3B,$FC,$1D,$8E,$D7
	.dc.b	$CF,$9B,$F9,$EE,$5E,$09,$F1,$65,$D1,$EE,$79,$71,$ED,$5C,$3E,$0F
	.dc.b	$D6,$7B,$3E,$12,$D7,$AF,$E1,$7D,$E7,$C0,$F8,$4B,$1E,$FF,$85,$F8
	.dc.b	$7F,$07,$F1,$1B,$1F,$0F,$E0,$FE,$3F,$C1,$FC,$86,$C7,$52,$67,$EA
	.dc.b	$C3,$AB,$EC,$BE,$6D,$9F,$9D,$87,$D2,$F8,$3F,$A7,$F0,$B6,$7E,$93
	.dc.b	$CB,$C4,$FF,$0F,$F0,$FF,$B9,$67,$EB,$65,$97,$FC,$D7,$0C,$BB,$D7
	.dc.b	$06,$CE,$BE,$17,$F0,$3F,$09,$F8,$26,$C7,$E0,$7E,$17,$F0,$A7,$E1
	.dc.b	$BD,$FB,$1F,$EB,$5E,$67,$C0,$7F,$06,$6C,$67,$E0,$31,$39,$F8,$09
	.dc.b	$3C,$7E,$0E,$F1,$FE,$03,$2E,$AF,$E0,$B2,$BB,$7F,$8F,$B3,$EB,$3F
	.dc.b	$E0,$26,$C6,$73,$EF,$8F,$97,$E1,$3C,$9F,$07,$E5,$6C,$7A,$7E,$13
	.dc.b	$D1,$F0,$9E,$96,$C7,$B7,$E1,$3D,$87,$B7,$E0,$6C,$7C,$1F,$82,$F8
	.dc.b	$1F,$03,$F0,$5B,$1D,$4F,$8D,$87,$56,$9D,$49,$B7,$AE,$FA,$BE,$DF
	.dc.b	$82,$F9,$87,$CD,$F8,$3B,$5F,$47,$E1,$3E,$87,$C1,$FD,$16,$C7,$D5
	.dc.b	$F8,$4F,$A8,$7D,$5F,$82,$B1,$F6,$7E,$0B,$EC,$1F,$67,$E1,$2C,$75
	.dc.b	$3E,$F6,$1D,$5A,$75,$26,$7F,$4E,$FA,$BF,$57,$E1,$3F,$A4,$FF,$00
	.dc.b	$A3,$F0,$D1,$87,$A1,$F6,$EE,$F3,$FD,$F7,$9B,$F8,$1C,$FD,$58,$F8
	.dc.b	$23,$FD,$03,$83,$E1,$FB,$FC,$37,$81,$B5,$E1,$AD,$FC,$37,$90,$F2
	.dc.b	$FC,$35,$9D,$1F,$ED,$4C,$3F,$AF,$67,$FE,$50,$5F,$01,$EC,$F8,$4F
	.dc.b	$6B,$EE,$F8,$4F,$F6,$8C,$FF,$48,$AE,$67,$F6,$F7,$0F,$38,$DF,$C2
	.dc.b	$F0,$F8,$69,$3C,$3E,$09,$6A,$EB,$F0,$DD,$BE,$1A,$5E,$E9,$3E,$2F
	.dc.b	$83,$F1,$FC,$24,$AD,$47,$9B,$E1,$3C,$FF,$03,$27,$94,$7A,$BE,$12
	.dc.b	$47,$57,$DB,$F0,$5E,$EF,$83,$F7,$FC,$24,$AF,$14,$7C,$2F,$84,$FF
	.dc.b	$79,$FF,$00,$FF,$AC,$27,$FC,$20,$DE,$1F,$6C,$EB,$CE,$DF,$C1,$4B
	.dc.b	$F8,$57,$FE,$BA,$78,$FC,$2F,$42,$3A,$F8,$3F,$03,$C3,$AB,$C2,$61
	.dc.b	$FF,$44,$D7,$3C,$8F,$FA,$D6,$7F,$2D,$18,$C3,$C9,$5E,$2F,$85,$F5
	.dc.b	$91,$F2,$FC,$2F,$5E,$FF,$F6,$9E,$3E,$D3,$DD,$CE,$DB,$9D,$F8,$B1
	.dc.b	$F7,$3E,$4D,$BE,$3F,$C2,$C8,$F8,$2F,$CA,$F8,$5F,$97,$F0,$B2,$7E
	.dc.b	$2B,$F3,$BE,$03,$E7,$FB,$F2,$7E,$4B,$F4,$BE,$0F,$E9,$FC,$24,$9F
	.dc.b	$9A,$FD,$6F,$85,$E9,$1F,$1B,$0B,$C1,$EC,$C7,$A4,$FC,$AC,$6C,$FC
	.dc.b	$0C,$7E,$FF,$C2,$CA,$FB,$2F,$E0,$BE,$17,$9F,$37,$CE,$E3,$67,$8B
	.dc.b	$FC,$2E,$7D,$2B,$59,$D9,$FC,$4E,$5F,$8C,$F8,$39,$1F,$86,$7F,$1F
	.dc.b	$F0,$9F,$90,$F8,$19,$3F,$8A,$7F,$27,$F0,$3F,$C4,$E3,$FD,$97,$FC
	.dc.b	$03,$FD,$69,$FD,$83,$5C,$FC,$2E,$F8,$5A,$7E,$1A,$9F,$07,$98,$C7
	.dc.b	$C2,$DC,$3F,$CD,$D4,$F8,$5C,$77,$A3,$13,$E0,$AF,$3B,$F9,$6D,$2E
	.dc.b	$5F,$8C,$D3,$8D,$F9,$63,$C7,$5D,$B4,$C9,$AD,$37,$F0,$92,$7D,$AF
	.dc.b	$2F,$85,$EB,$B7,$D3,$CE,$D7,$ED,$B3,$E6,$FC,$4C,$6B,$F2,$31,$F1
	.dc.b	$FC,$2C,$8F,$92,$F9,$BE,$0B,$CF,$F0,$BE,$8F,$83,$95,$F3,$A3,$D7
	.dc.b	$F0,$92,$3E,$8B,$EE,$F8,$5F,$7F,$C2,$C9,$FA,$AF,$C2,$F8,$5E,$B1
	.dc.b	$F6,$30,$EA,$FD,$CC,$3A,$3F,$7B,$0B,$C5,$F2,$39,$7E,$96,$5F,$33
	.dc.b	$E1,$65,$FE,$09,$F9,$FF,$03,$DA,$3F,$0F,$97,$6C,$39,$73,$33,$EB
	.dc.b	$C9,$F8,$5D,$2E,$1F,$82,$CF,$EC,$FC,$0C,$8F,$C7,$BF,$73,$E0,$FE
	.dc.b	$EF,$C2,$C9,$FC,$9B,$F8,$0F,$85,$FC,$0F,$C2,$FE,$2B,$E1,$7F,$D9
	.dc.b	$DB,$7A,$DB,$A7,$E5,$67,$5F,$0B,$B9,$98,$FF,$4D,$7F,$B4,$3D,$88
	.dc.b	$7E,$93,$1F,$03,$FA,$4A,$F0,$30,$C1,$CF,$E1,$7C,$84,$7B,$7C,$2F
	.dc.b	$AD,$E8,$76,$CB,$E4,$E7,$EC,$8F,$21,$D4,$FD,$AD,$FB,$FE,$16,$47
	.dc.b	$A5,$F8,$5F,$0B,$F0,$FE,$17,$E2,$7C,$3C,$AF,$74,$7C,$7F,$85,$91
	.dc.b	$F0,$5F,$95,$F0,$DF,$2F,$E1,$7E,$67,$C3,$4A,$F8,$D1,$D5,$F9,$18
	.dc.b	$75,$5A,$FD,$1C,$7A,$C6,$38,$FF,$A5,$F1,$7F,$79,$8F,$01,$F4,$72
	.dc.b	$FB,$5F,$0B,$23,$EA,$3F,$77,$E1,$BC,$11,$CC,$FF,$7C,$57,$F8,$2B
	.dc.b	$1F,$C0,$65,$A7,$F0,$5A,$F8,$9B,$5D,$CF,$C0,$7C,$2F,$E2,$48,$FE
	.dc.b	$0F,$E1,$BF,$19,$F0,$DF,$8D,$F8,$69,$3F,$87,$8F,$C8,$7C,$2C,$8F
	.dc.b	$C5,$3F,$93,$F8,$6F,$CA,$7C,$0C,$9F,$C7,$3F,$C8,$3F,$80,$7F,$C0
	.dc.b	$1B,$30,$EB,$5E,$DF,$1D,$2C,$BC,$73,$D8,$86,$E1,$97,$86,$3F,$C5
	.dc.b	$18,$7F,$7D,$3F,$BE,$2B,$46,$1B,$F0,$3D,$A7,$A6,$1E,$4D,$BF,$59
	.dc.b	$97,$BE,$7C,$D8,$3E,$1F,$D8,$66,$BF,$86,$43,$E1,$FE,$07,$C3,$CB
	.dc.b	$F2,$C7,$C3,$F8,$7F,$89,$F0,$B2,$7D,$31,$F1,$FE,$16,$47,$B1,$F9
	.dc.b	$5F,$01,$F2,$FE,$02,$4F,$C0,$7E,$77,$C0,$75,$8F,$87,$87,$57,$D3
	.dc.b	$87,$D3,$F8,$09,$1F,$21,$FA,$DF,$0B,$F5,$FE,$16,$4F,$CC,$7E,$D7
	.dc.b	$C2,$F5,$8F,$87,$87,$5E,$0F,$83,$8F,$DF,$F8,$59,$5F,$51,$FC,$17
	.dc.b	$C2,$7E,$0F,$E1,$3F,$AC,$DF,$C8,$6F,$89,$FE,$32,$E4,$7C,$06,$31
	.dc.b	$FB,$50,$AC,$75,$7D,$CF,$80,$BF,$C0,$3F,$EC,$4C,$79,$46,$3E,$07
	.dc.b	$3F,$03,$27,$FC,$DD,$4F,$84,$C6,$7F,$93,$8F,$EF,$13,$AF,$84,$EB
	.dc.b	$F0,$BD,$9F,$0E,$AC,$FB,$57,$E4,$9B,$87,$3C,$7A,$EF,$FD,$47,$2F
	.dc.b	$05,$F8,$B3,$F4,$FC,$2C,$8F,$2F,$C2,$DC,$7F,$23,$97,$4C,$7F,$1E
	.dc.b	$DA,$FF,$2C,$E3,$D2,$BC,$11,$EE,$F8,$5F,$8B,$9F,$C6,$F7,$FE,$3F
	.dc.b	$C0,$DA,$F8,$6F,$CA,$F8,$0F,$97,$F0,$16,$7E,$3B,$F3,$BE,$17,$E7
	.dc.b	$FC,$2D,$9F,$96,$FD,$2F,$85,$FA,$7F,$0B,$67,$E7,$BC,$F5,$F9,$D8
	.dc.b	$5E,$0F,$88,$F3,$8F,$A5,$85,$9F,$90,$FD,$EF,$84,$98,$7D,$AF,$84
	.dc.b	$B7,$F6,$74,$F4,$CF,$DE,$6C,$7D,$DD,$3A,$4F,$C8,$F8,$3B,$3F,$C1
	.dc.b	$F8,$9F,$03,$F8,$C3,$F1,$BF,$03,$6F,$FD,$6D,$FC,$03,$FE,$DD,$6D
	.dc.b	$FE,$52,$3F,$DA,$91,$83,$7E,$A6,$4D,$7C,$25,$8E,$7F,$09,$C4,$E5
	.dc.b	$F0,$96,$39,$61,$FC,$9F,$0F,$03,$7E,$17,$C0,$F3,$35,$8F,$6A,$F0
	.dc.b	$9E,$8F,$85,$F3,$7C,$1F,$9D,$B1,$DB,$97,$FC,$4D,$8F,$66,$E3,$DA
	.dc.b	$79,$E3,$D9,$F8,$7A,$78,$6B,$8F,$C0,$7C,$3F,$80,$F8,$8D,$AF,$93
	.dc.b	$F0,$7F,$1F,$E1,$3E,$43,$63,$E6,$FC,$2F,$CB,$F8,$5F,$98,$D8,$FA
	.dc.b	$3F,$0B,$F3,$CF,$A1,$F0,$B6,$3A,$9F,$5B,$0E,$AD,$3A,$9F,$6A,$FA
	.dc.b	$B4,$FB,$BF,$0B,$F6,$FE,$03,$EE,$36,$3A,$EF,$F4,$EF,$AB,$63,$AB
	.dc.b	$F5,$EF,$AB,$63,$F1,$3F,$03,$F8,$73,$F1,$1F,$07,$63,$F1,$BF,$0B
	.dc.b	$F8,$BF,$85,$FC,$63,$61,$7F,$DA,$7C,$9F,$F4,$AC,$63,$8F,$FC,$BD
	.dc.b	$0F,$85,$FE,$4E,$EA,$B8,$46,$3E,$23,$3F,$0D,$70,$E4,$7F,$B9,$2F
	.dc.b	$FB,$03,$E4,$3F,$43,$97,$91,$A7,$0C,$3F,$B8,$4F,$7C,$7C,$78,$F8
	.dc.b	$AB,$C6,$F9,$3E,$17,$CB,$F0,$D6,$79,$7B,$1F,$CD,$63,$B6,$3F,$A1
	.dc.b	$CA,$E3,$F9,$67,$A5,$F0,$CB,$E0,$34,$EE,$67,$DD,$F8,$7E,$F7,$C4
	.dc.b	$6C,$71,$F8,$6F,$8E,$7C,$8F,$86,$B1,$D7,$E1,$BE,$5F,$C3,$7C,$C6
	.dc.b	$C7,$87,$E1,$BE,$7F,$C3,$7D,$06,$C7,$52,$6D,$F6,$70,$EA,$FD,$BB
	.dc.b	$E6,$D9,$F6,$61,$F6,$BE,$1A,$47,$BF,$E1,$AD,$7E,$0F,$3E,$7C,$1F
	.dc.b	$82,$CA,$DF,$DB,$CB,$AC,$FE,$19,$B1,$F7,$FE,$1B,$F0,$FF,$0D,$F8
	.dc.b	$86,$C7,$CD,$F8,$5F,$C5,$FC,$37,$E3,$1B,$1F,$47,$DF,$FE,$EB,$FC
	.dc.b	$03,$FD,$CF,$5F,$C1,$C6,$3E,$1F,$3F,$09,$AE,$7B,$87,$E0,$30,$FE
	.dc.b	$B5,$B7,$E6,$B2,$EB,$97,$E7,$DB,$1F,$9E,$CB,$A5,$F8,$4E,$59,$78
	.dc.b	$BE,$1F,$C7,$F0,$F6,$7B,$3E,$97,$F1,$1A,$DC,$BF,$58,$FE,$A3,$0E
	.dc.b	$B1,$FA,$F6,$E3,$FE,$CB,$CD,$F0,$DF,$03,$E1,$BE,$0B,$71,$F6,$7C
	.dc.b	$37,$C4,$F8,$6F,$8A,$D8,$F8,$1F,$0D,$F2,$0F,$93,$F0,$16,$3E,$27
	.dc.b	$C2,$7C,$C3,$E6,$FC,$0D,$8E,$A7,$C9,$C3,$AB,$4E,$A4,$DB,$FC,$E1
	.dc.b	$A3,$0E,$D1,$8F,$84,$FB,$3F,$0D,$F6,$9B,$7D,$A7,$8E,$1D,$9B,$1D
	.dc.b	$77,$F8,$9F,$0D,$D1,$B1,$F7,$7E,$1B,$F0,$BF,$0D,$F8,$66,$C7,$E0
	.dc.b	$7E,$1B,$F1,$27,$E2,$BE,$16,$C7,$E1,$7E,$17,$F1,$A7,$F9,$0B,$EB
	.dc.b	$3F,$AB,$47,$F2,$4F,$35,$C2,$AD,$7C,$04,$AF,$E1,$B7,$FE,$90,$6C
	.dc.b	$76,$DB,$E2,$E5,$E1,$9F,$21,$E1,$24,$7F,$A6,$6B,$C2,$FF,$A3,$F0
	.dc.b	$F0,$4F,$6F,$80,$F3,$1E,$7F,$86,$B1,$D2,$BB,$BF,$A1,$D3,$C5,$BF
	.dc.b	$E3,$74,$F7,$CF,$A2,$A9,$E1,$F8,$6F,$84,$7C,$3F,$84,$B3,$EE,$F8
	.dc.b	$5F,$8C,$7C,$7F,$83,$B1,$F0,$BE,$13,$E5,$7C,$27,$CB,$6C,$7C,$6F
	.dc.b	$7F,$E7,$1F,$3F,$E1,$2C,$75,$26,$7F,$0F,$0E,$AF,$C5,$C3,$9B,$F1
	.dc.b	$F0,$BA,$AF,$EC,$71,$3E,$03,$EE,$7C,$27,$DD,$6C,$F6,$AE,$18,$7F
	.dc.b	$AF,$6D,$AF,$B6,$B4,$74,$CB,$BB,$DB,$2B,$7C,$7E,$13,$FB,$5B,$F6
	.dc.b	$DB,$8F,$6E,$86,$49,$CE,$1E,$8F,$83,$B8,$78,$FE,$12,$E1,$C3,$E1
	.dc.b	$2E,$3F,$ED,$DA,$FC,$8D,$63,$E1,$24,$7E,$51,$DF,$C2,$70,$F7,$FF
	.dc.b	$D3,$FF,$C0,$3F,$DA,$F4,$F8,$F1,$87,$3F,$07,$1F,$ED,$32,$F8,$5C
	.dc.b	$63,$FE,$EF,$8F,$EF,$6D,$F8,$5E,$BF,$0B,$D9,$EF,$F0,$B3,$0E,$7F
	.dc.b	$09,$6F,$97,$C2,$F9,$0F,$2F,$C0,$5A,$F0,$FC,$07,$A3,$E0,$3D,$2D
	.dc.b	$8F,$2F,$C1,$FB,$3E,$13,$DA,$D8,$F4,$FC,$2F,$C0,$3E,$0F,$C2,$D8
	.dc.b	$F6,$FC,$2F,$C4,$3E,$2F,$C2,$58,$F8,$3F,$0B,$F2,$3E,$0F,$E4,$B6
	.dc.b	$3E,$2F,$C0,$7C,$CF,$80,$F9,$AD,$8F,$F0,$8F,$3F,$F5,$9E,$FF,$95
	.dc.b	$F0,$BF,$ED,$08,$C0,$71,$F0,$B2,$B8,$FC,$25,$E6,$78,$BF,$8B,$C1
	.dc.b	$F0,$9C,$FE,$0F,$A3,$70,$F0,$FC,$27,$73,$C1,$F0,$96,$3C,$BE,$F7
	.dc.b	$E3,$B0,$34,$C7,$C2,$67,$E0,$B5,$F0,$93,$5F,$6C,$71,$F8,$4E,$5F
	.dc.b	$07,$27,$E0,$C7,$5F,$84,$91,$F1,$1F,$07,$C2,$75,$8F,$8F,$87,$57
	.dc.b	$DF,$A7,$F3,$3D,$3F,$A7,$47,$CC,$F8,$3F,$47,$C2,$48,$C1,$F4,$3E
	.dc.b	$0F,$D9,$F0,$12,$30,$7A,$0E,$E4,$3B,$8E,$9F,$D0,$6F,$F1,$7E,$03
	.dc.b	$F6,$7F,$91,$7E,$FA,$B7,$48,$C7,$47,$D5,$AD,$FC,$2F,$0F,$88,$98
	.dc.b	$7F,$95,$2D,$3D,$3E,$17,$AF,$C2,$CA,$D4,$78,$3E,$16,$47,$17,$C7
	.dc.b	$F0,$BE,$4F,$80,$F2,$FC,$1C,$AE,$D1,$D5,$F0,$61,$D5,$F4,$E4,$C7
	.dc.b	$DB,$F0,$12,$7C,$CF,$C0,$F8,$1F,$83,$F0,$B2,$7D,$4F,$C4,$F8,$5E
	.dc.b	$D1,$D3,$1E,$DC,$3F,$C8,$72,$ED,$3F,$07,$E1,$3E,$5F,$C2,$C8,$C1
	.dc.b	$F1,$7E,$17,$E7,$FC,$2C,$8C,$1F,$27,$E0,$FE,$9F,$C2,$C8,$C1,$F3
	.dc.b	$7E,$13,$EB,$91,$FA,$1F,$09,$F6,$BE,$13,$F9,$6F,$E3,$FF,$D0,$A7
	.dc.b	$ED,$B7,$6F,$44,$FE,$03,$1E,$D5,$F2,$31,$C3,$FE,$8E,$BF,$F5,$33
	.dc.b	$7C,$2E,$A7,$6F,$85,$EF,$F0,$7E,$0F,$84,$95,$C6,$39,$BD,$4F,$C3
	.dc.b	$61,$8C,$FF,$A5,$62,$F8,$4F,$4F,$C2,$4B,$F1,$47,$B3,$E1,$3D,$BF
	.dc.b	$09,$27,$CD,$1F,$03,$E0,$E4,$7A,$5F,$87,$F0,$9F,$13,$E0,$E4,$FB
	.dc.b	$5F,$8F,$F0,$9F,$23,$E1,$24,$FC,$17,$E5,$FC,$27,$48,$F9,$78,$5D
	.dc.b	$BD,$38,$FD,$0F,$84,$FA,$3F,$09,$2F,$E5,$47,$D4,$F8,$3F,$AB,$F0
	.dc.b	$92,$7E,$74,$73,$F5,$3F,$C6,$DC,$6C,$F1,$5F,$B9,$86,$5F,$7B,$E0
	.dc.b	$E4,$FD,$77,$F0,$3F,$09,$FD,$BB,$F1,$E7,$C4,$DF,$86,$97,$FE,$CA
	.dc.b	$9C,$B7,$E1,$B8,$11,$D7,$C3,$73,$F8,$5E,$9F,$0D,$27,$94,$77,$F8
	.dc.b	$69,$1D,$5F,$17,$C3,$73,$8F,$E1,$B8,$B1,$C7,$37,$F8,$2E,$19,$7A
	.dc.b	$7E,$1B,$D5,$F0,$D3,$0F,$44,$7B,$7D,$DF,$77,$C1,$49,$F6,$47,$4E
	.dc.b	$6F,$F3,$06,$3D,$E7,$DC,$7B,$0F,$85,$F0,$BF,$1F,$E1,$A4,$60,$F8
	.dc.b	$DF,$0D,$F2,$FE,$1A,$46,$0F,$95,$F0,$DF,$3C,$8F,$CD,$F8,$6F,$A5
	.dc.b	$F0,$DF,$4F,$E1,$A4,$FD,$18,$FA,$DF,$0B,$23,$EA,$3F,$E2,$6F,$7F
	.dc.b	$F9,$8E,$DE,$AC,$7A,$D7,$CA,$7E,$F6,$3B,$F8,$4E,$1F,$0B,$28,$7F
	.dc.b	$3C,$C3,$F9,$AB,$8E,$6F,$F1,$D4,$7F,$1A,$8D,$3E,$13,$C5,$EE,$F8
	.dc.b	$27,$EB,$63,$72,$F9,$B9,$79,$FE,$16,$63,$D6,$7F,$D2,$18,$A7,$D9
	.dc.b	$F0,$BE,$DF,$84,$F7,$7C,$2C,$C3,$81,$FD,$5E,$FF,$CB,$2E,$3E,$1B
	.dc.b	$F9,$E7,$C3,$7C,$52,$46,$65,$F0,$DF,$27,$E1,$A4,$E5,$F9,$9F,$0D
	.dc.b	$F3,$7E,$17,$E7,$7C,$2C,$AE,$51,$D1,$F9,$18,$DE,$2F,$8F,$8F,$D5
	.dc.b	$F8,$5F,$AD,$F0,$B2,$FC,$51,$F6,$7E,$0B,$ED,$7B,$D2,$7C,$D1,$CF
	.dc.b	$93,$E2,$E5,$67,$E0,$63,$F8,$1F,$85,$FC,$17,$C2,$C9,$F7,$47,$E1
	.dc.b	$7E,$17,$FC,$D7,$F8,$EF,$F3,$DC,$7F,$D7,$CD,$AF,$95,$3A,$F8,$59
	.dc.b	$1F,$2E,$7F,$9E,$BF,$E0,$07,$FC,$C8,$E7,$E1,$3A,$FC,$2C,$8C,$1C
	.dc.b	$7E,$13,$C3,$F0,$B2,$30,$75,$C3,$BB,$D5,$F3,$FC,$37,$A3,$E1,$64
	.dc.b	$F8,$DF,$5F,$C3,$7B,$3D,$EF,$6F,$BB,$2B,$D1,$1D,$3A,$3F,$D2,$5B
	.dc.b	$7A,$9B,$97,$6A,$F7,$7C,$37,$C6,$F8,$69,$18,$3E,$17,$C3,$7C,$A2
	.dc.b	$3F,$17,$E1,$BE,$6F,$C2,$FC,$EF,$86,$93,$F2,$63,$E8,$FC,$2C,$8F
	.dc.b	$98,$FD,$4F,$86,$FA,$BF,$0B,$F5,$BE,$16,$57,$D1,$8E,$AF,$D3,$C3
	.dc.b	$AB,$F4,$F0,$FF,$53,$FE,$37,$FC,$DA,$3F,$C6,$AF,$F9,$B5,$FD,$23
	.dc.b	$5C,$FC,$37,$0F,$81,$E3,$F0,$32,$77,$1E,$07,$F0,$18,$78,$38,$BE
	.dc.b	$5E,$3E,$0F,$84,$95,$D9,$F1,$FC,$2F,$93,$E1,$7C,$BF,$0B,$2B,$C7
	.dc.b	$1E,$8F,$84,$91,$E6,$7D,$7F,$0B,$EC,$F8,$4F,$6F,$C2,$CA,$F5,$C7
	.dc.b	$C0,$F8,$4F,$83,$F0,$92,$7D,$F1,$D1,$F8,$78,$5D,$BF,$DA,$9C,$6C
	.dc.b	$BE,$4F,$C2,$4B,$F8,$EF,$CC,$F8,$4F,$9B,$EF,$49,$F9,$6F,$D0,$F7
	.dc.b	$BA,$79,$9F,$CA,$BD,$A6,$5F,$E8,$CB,$F6,$1F,$4B,$2F,$AF,$F0,$72
	.dc.b	$3E,$AB,$F6,$BE,$13,$ED,$FC,$24,$9F,$B2,$FF,$B9,$3F,$1A,$FE,$7F
	.dc.b	$1F,$E7,$77,$F9,$BB,$FE,$74,$6E,$FF,$EF,$16,$E7,$E5,$5F,$09,$CF
	.dc.b	$DF,$E9,$F0,$93,$0F,$E5,$55,$FC,$FF,$0E,$5F,$09,$E1,$F8,$49,$5A
	.dc.b	$8F,$27,$C2,$C8,$E2,$F9,$F2,$F4,$7C,$2C,$9E,$AF,$77,$D1,$F0,$BE
	.dc.b	$DF,$76,$47,$8A,$3A,$3E,$DB,$B9,$7B,$70,$F8,$9E,$EC,$AF,$4B,$F1
	.dc.b	$FE,$17,$E4,$7C,$2C,$9F,$6B,$F2,$FE,$13,$E6,$7C,$2C,$9F,$82,$FC
	.dc.b	$FF,$84,$FA,$1F,$0B,$F4,$7E,$12,$57,$C6,$8F,$A9,$F0,$BF,$57,$E1
	.dc.b	$24,$FC,$A8,$E8,$FD,$5C,$2E,$DF,$57,$1F,$B9,$F0,$B2,$BE,$8B,$F7
	.dc.b	$FE,$17,$FB,$AC,$7F,$AC,$3D,$1F,$E7,$1E,$0F,$F8,$8B,$DB,$CE,$2E
	.dc.b	$DF,$C9,$57,$F1,$F6,$9E,$03,$F0,$59,$DC,$3F,$15,$9F,$29,$E3,$F0
	.dc.b	$36,$7F,$D8,$97,$FD,$01,$F1,$46,$72,$FF,$19,$FC,$2D,$A7,$C2,$DF
	.dc.b	$85,$98,$7F,$75,$9C,$B7,$E1,$78,$11,$D7,$C2,$F3,$F8,$59,$1C,$5E
	.dc.b	$DF,$09,$DF,$E1,$64,$F5,$7C,$5F,$09,$CE,$3E,$46,$17,$8B,$E3,$47
	.dc.b	$B3,$1F,$47,$C2,$48,$F3,$3E,$BF,$84,$F6,$7C,$27,$B7,$DD,$95,$EB
	.dc.b	$8E,$9C,$9F,$43,$2B,$3D,$4F,$81,$EF,$7C,$52,$3F,$0F,$E1,$3E,$47
	.dc.b	$C2,$48,$F8,$CF,$CB,$F8,$3F,$99,$F0,$9F,$3F,$F1,$2F,$E7,$DC,$1E
	.dc.b	$AE,$94,$63,$4F,$C0,$D6,$FE,$22,$65,$FE,$25,$AF,$E8,$58,$71,$7F
	.dc.b	$CC,$11,$FE,$FC,$7F,$CD,$AE,$FE,$13,$C0,$78,$7A,$3C,$71,$CF,$E1
	.dc.b	$BC,$86,$1E,$5F,$0B,$E7,$23,$E1,$F8,$4F,$57,$C2,$FA,$FE,$06,$4F
	.dc.b	$96,$39,$67,$E5,$C3,$AC,$FE,$27,$5F,$85,$EE,$48,$F6,$3F,$17,$E1
	.dc.b	$3E,$37,$C2,$FC,$7F,$84,$95,$F0,$63,$E5,$7C,$2C,$8F,$88,$FC,$DF
	.dc.b	$85,$F9,$DF,$0B,$27,$E4,$3F,$47,$E1,$7E,$97,$C2,$C9,$F9,$8F,$D5
	.dc.b	$F8,$4E,$91,$F2,$F0,$BB,$FC,$FC,$7E,$D7,$C2,$4A,$FA,$8F,$DD,$F8
	.dc.b	$4F,$D6,$7F,$00,$FF,$3E,$EC,$34,$9E,$EF,$4E,$DF,$EA,$1D,$8D,$31
	.dc.b	$93,$0F,$C2,$E7,$D3,$DD,$EB,$F0,$B2,$BF,$CD,$19,$FF,$44,$6F,$C2
	.dc.b	$F8,$88,$EB,$E0,$FC,$BF,$09,$E6,$F8,$49,$3C,$A3,$D3,$F0,$72,$3A
	.dc.b	$BE,$CF,$84,$F6,$FC,$24,$9F,$0B,$F0,$3E,$13,$AC,$79,$30,$EA,$B9
	.dc.b	$BF,$41,$A3,$2F,$8F,$F0,$9F,$23,$E1,$25,$FB,$23,$E5,$FB,$B2,$3D
	.dc.b	$EF,$CE,$F7,$7B,$C7,$5E,$7F,$EF,$58,$FF,$A3,$23,$07,$CB,$71,$97
	.dc.b	$8B,$2F,$AF,$F0,$72,$BE,$5B,$F6,$BE,$13,$ED,$FC,$1F,$DC,$F8,$49
	.dc.b	$5F,$42,$3E,$FF,$C2,$48,$FA,$6F,$E0,$BE,$13,$F9,$77,$E1,$7F,$D0
	.dc.b	$0F,$E7,$93,$F7,$5B,$D8,$75,$F0,$9C,$09,$3F,$EA,$58,$7C,$27,$78
	.dc.b	$F6,$D7,$E3,$B0,$C6,$DF,$3B,$1F,$0F,$C2,$78,$BE,$12,$61,$CE,$3C
	.dc.b	$BE,$DC,$8E,$CF,$A3,$E0,$39,$F2,$7C,$CC,$AD,$7C,$BC,$7D,$BF,$09
	.dc.b	$23,$CC,$FC,$0F,$84,$F8,$3F,$09,$FC,$E3,$5F,$E4,$1C,$08,$C6,$5F
	.dc.b	$E3,$18,$7C,$26,$31,$F8,$D5,$78,$3E,$37,$C2,$5B,$7C,$25,$8C,$39
	.dc.b	$F8,$39,$97,$C9,$F8,$4B,$97,$C8,$C1,$F0,$9C,$FE,$13,$A3,$6F,$E8
	.dc.b	$65,$D1,$A7,$D3,$CB,$C5,$F0,$92,$3E,$B7,$C2,$5C,$BB,$7B,$9E,$9E
	.dc.b	$A7,$BB,$E9,$8C,$3E,$9B,$E8,$76,$AF,$36,$3C,$7E,$13,$97,$C4,$5A
	.dc.b	$FF,$74,$62,$8E,$DF,$09,$DF,$E1,$6D,$69,$F1,$7C,$2F,$8F,$E1,$6C
	.dc.b	$F2,$7C,$DF,$09,$E7,$F8,$5B,$3D,$9F,$57,$C2,$C8,$F0,$FC,$2D,$DF
	.dc.b	$F2,$7A,$BE,$13,$E0,$1F,$07,$E1,$6D,$FA,$3D,$DF,$89,$EE,$FC,$56
	.dc.b	$C7,$B3,$E1,$7E,$41,$F2,$7E,$16,$C7,$C0,$F8,$5F,$99,$F0,$BF,$35
	.dc.b	$B1,$F1,$3E,$17,$E8,$7C,$2F,$D1,$6C,$36,$FD,$4F,$F0,$0D,$8C,$79
	.dc.b	$1F,$DD,$63,$13,$AF,$86,$FC,$94,$62,$7F,$BD,$70,$B2,$F0,$DF,$C2
	.dc.b	$6C,$F3,$F8,$5E,$27,$2F,$85,$B9,$78,$0F,$F4,$96,$2F,$85,$FF,$1B
	.dc.b	$37,$1D,$1F,$D1,$5C,$3E,$AF,$85,$DF,$B7,$23,$7E,$62,$BF,$98,$46
	.dc.b	$1F,$E4,$78,$78,$3E,$17,$B9,$27,$5F,$0B,$E2,$F8,$5F,$1F,$C2,$C9
	.dc.b	$E5,$1E,$6F,$83,$F3,$FC,$24,$9E,$D1,$EA,$F8,$4F,$5F,$C2,$49,$F1
	.dc.b	$47,$BB,$E0,$FD,$FF,$07,$27,$CD,$1F,$0B,$E0,$A4,$7A,$5F,$8B,$F0
	.dc.b	$7F,$1B,$DE,$F8,$FE,$F4,$AF,$74,$7F,$97,$FF,$80,$7F,$CE,$27,$F3
	.dc.b	$02,$71,$67,$E1,$25,$7C,$89,$FF,$34,$75,$FF,$8D,$DF,$E6,$31,$9F
	.dc.b	$83,$E8,$47,$87,$C2,$77,$F8,$49,$1C,$DF,$17,$C2,$78,$FE,$13,$C9
	.dc.b	$F0,$92,$BC,$11,$E7,$F8,$39,$1E,$37,$D5,$F0,$9E,$BF,$7E,$4F,$9D
	.dc.b	$F7,$7B,$7D,$63,$D5,$87,$57,$FA,$3E,$AC,$7E,$27,$C2,$7C,$5F,$84
	.dc.b	$95,$F0,$63,$E4,$7C,$17,$C9,$F8,$39,$3F,$16,$3B,$3D,$31,$ED,$C5
	.dc.b	$F1,$F1,$FA,$1F,$07,$F4,$7E,$0E,$5F,$CD,$8F,$A9,$F0,$72,$3E,$83
	.dc.b	$FE,$95,$FE,$01,$FF,$32,$47,$F8,$8D,$FF,$30,$F4,$7F,$BD,$DB,$19
	.dc.b	$F8,$3E,$1F,$07,$C7,$E0,$F9,$7C,$1C,$AF,$D1,$DF,$F9,$99,$FE,$D2
	.dc.b	$BE,$0F,$C1,$F0,$72,$77,$1E,$3C,$64,$72,$7B,$BD,$3E,$0F,$D1,$F0
	.dc.b	$72,$30,$78,$3E,$0F,$D8,$47,$C7,$F0,$7D,$1F,$1D,$DE,$3F,$F4,$36
	.dc.b	$7F,$E7,$E6,$FC,$1F,$C5,$25,$7B,$3E,$0F,$E4,$7C,$14,$8F,$7B,$F2
	.dc.b	$FE,$0F,$E6,$7C,$1F,$CD,$F8,$39,$5F,$12,$3E,$87,$BD,$23,$E3,$BF
	.dc.b	$4F,$E0,$FE,$A7,$C1,$7D,$5F,$83,$95,$F3,$23,$AB,$F3,$B0,$EA,$FC
	.dc.b	$EC,$3F,$DA,$1E,$FF,$F9,$94,$7C,$1E,$0C,$CE,$BE,$0B,$7F,$05,$23
	.dc.b	$F8,$DD,$7F,$97,$DE,$87,$72,$18,$3F,$C4,$AF,$F2,$6E,$1F,$F5,$83
	.dc.b	$7B,$7F,$A1,$7C,$07,$F2,$F8,$F2,$FC,$17,$9B,$E0,$A6,$1E,$07,$D3
	.dc.b	$F0,$FE,$AF,$82,$F5,$FC,$14,$AF,$2C,$7B,$BE,$06,$47,$A1,$F8,$3F
	.dc.b	$05,$F0,$BE,$07,$E1,$FC,$14,$AF,$6C,$75,$7D,$F8,$75,$7D,$78,$7C
	.dc.b	$AF,$81,$91,$F1,$1F,$9B,$F0,$5F,$3B,$E0,$64,$FC,$87,$E8,$FC,$17
	.dc.b	$58,$F8,$B8,$75,$E3,$F8,$B8,$FD,$6F,$82,$FA,$FF,$05,$2F,$E8,$C7
	.dc.b	$DA,$F8,$19,$1F,$51,$FB,$BF,$05,$FD,$B6,$3F,$CA,$BF,$C0,$3F,$EF
	.dc.b	$86,$FC,$56,$72,$FF,$71,$4F,$F9,$88,$D3,$8F,$84,$D7,$57,$D6,$9F
	.dc.b	$E6,$37,$D3,$E1,$B2,$FF,$A9,$E3,$F9,$BC,$EF,$E1,$73,$D8,$94,$61
	.dc.b	$B5,$FD,$99,$7C,$2E,$6F,$59,$CA,$D4,$71,$F8,$5E,$5F,$0B,$27,$94
	.dc.b	$75,$F6,$E4,$75,$7C,$1E,$E7,$58,$F0,$61,$D5,$F8,$F8,$79,$7E,$16
	.dc.b	$47,$95,$F4,$7C,$2F,$A7,$E1,$64,$FA,$5F,$67,$C2,$FB,$7E,$0E,$4F
	.dc.b	$B5,$F8,$1F,$09,$F0,$7E,$0E,$4F,$C1,$7E,$27,$C2,$7C,$5F,$83,$F8
	.dc.b	$DF,$09,$2B,$E3,$47,$47,$EE,$E1,$7A,$7E,$DE,$3F,$37,$E1,$3E,$8F
	.dc.b	$F0,$0F,$F3,$AE,$93,$7A,$13,$AF,$72,$65,$FE,$18,$AF,$E7,$B7,$C3
	.dc.b	$E1,$7B,$72,$7E,$0F,$4B,$7F,$81,$CF,$BF,$C2,$F8,$3E,$16,$4F,$38
	.dc.b	$F1,$FC,$27,$FA,$2F,$3F,$E8,$18,$F7,$F8,$5C,$4E,$7E,$16,$4F,$8F
	.dc.b	$E1,$6E,$3F,$EB,$0C,$1F,$0B,$CC,$E9,$F0,$B6,$FD,$5F,$09,$DF,$E1
	.dc.b	$3C,$0D,$8F,$76,$5D,$39,$91,$FC,$53,$16,$79,$8C,$7C,$27,$A4,$F5
	.dc.b	$7C,$25,$8F,$91,$EF,$FB,$7E,$03,$DC,$D8,$EE,$4D,$FD,$58,$75,$5C
	.dc.b	$FF,$5B,$E1,$3E,$31,$F1,$FE,$12,$DF,$D5,$F8,$3F,$94,$7C,$BF,$84
	.dc.b	$B1,$FE,$50,$FE,$01,$FF,$33,$3F,$CB,$9F,$E7,$EE,$71,$FF,$5E,$46
	.dc.b	$1F,$F6,$DA,$F8,$4C,$9A,$F8,$4B,$5D,$7E,$13,$89,$CB,$E1,$2C,$75
	.dc.b	$3C,$5F,$0B,$DB,$1E,$EF,$83,$E1,$7C,$3F,$0B,$6B,$D0,$F9,$3D,$C9
	.dc.b	$1E,$BF,$76,$E1,$E6,$39,$C7,$77,$D3,$87,$B3,$E1,$64,$FC,$3F,$85
	.dc.b	$B5,$EE,$F8,$4F,$83,$F0,$BF,$09,$B5,$F2,$7E,$17,$E2,$9F,$1B,$E1
	.dc.b	$6C,$7C,$DF,$85,$F9,$3F,$0B,$F2,$9B,$1F,$47,$E1,$3E,$6F,$C2,$FC
	.dc.b	$E6,$C7,$52,$67,$F4,$30,$EA,$FD,$2F,$85,$FA,$BF,$0B,$F5,$9B,$5F
	.dc.b	$E8,$CF,$3B,$F2,$ED,$F7,$7F,$DE,$D1,$8B,$78,$FF,$9C,$7F,$1B,$09
	.dc.b	$EF,$39,$F0,$5D,$AE,$0F,$81,$8F,$9A,$B5,$83,$E1,$7B,$FC,$2F,$81
	.dc.b	$B5,$E5,$F8,$5F,$1F,$C2,$F9,$1B,$1E,$9F,$84,$F3,$9E,$8F,$85,$B1
	.dc.b	$ED,$F8,$5F,$5F,$C2,$FB,$1B,$1F,$07,$E1,$7D,$E7,$C0,$F8,$5B,$1D
	.dc.b	$49,$C7,$F0,$9F,$F0,$7E,$CC,$7B,$CE,$4F,$8B,$F0,$BF,$28,$F9,$7F
	.dc.b	$0B,$6F,$E8,$7B,$BF,$38,$F9,$FE,$ED,$8E,$FC,$7F,$1B,$7F,$64,$60
	.dc.b	$F0,$FC,$2F,$D6,$F8,$5F,$AE,$D8,$FB,$9F,$09,$F6,$8F,$B7,$F0,$B6
	.dc.b	$3F,$01,$F0,$9F,$6F,$1F,$B1,$FC,$02,$F8,$AF,$C0,$CE,$6B,$FC,$5B
	.dc.b	$39,$7F,$DA,$6B,$E1,$B5,$C3,$FC,$42,$1F,$0D,$8E,$F4,$6E,$F8,$CE
	.dc.b	$61,$F0,$D7,$9B,$0D,$AF,$39,$E0,$CF,$CA,$D3,$C7,$F0,$BC,$FE,$1B
	.dc.b	$A3,$63,$CF,$F0,$5D,$FD,$DF,$03,$63,$AE,$BF,$67,$5F,$3E,$3D,$A3
	.dc.b	$1A,$7C,$16,$FC,$37,$A4,$F5,$7C,$35,$C3,$E2,$FC,$2F,$B4,$F7,$7C
	.dc.b	$35,$8F,$93,$F0,$BF,$04,$F8,$5F,$0B,$63,$E6,$FC,$27,$C5,$3E,$37
	.dc.b	$C2,$D8,$FA,$3F,$0B,$F2,$7E,$17,$E5,$36,$3A,$93,$5F,$4E,$1D,$5F
	.dc.b	$5F,$C2,$FD,$13,$E9,$7C,$2D,$A4,$FC,$3F,$E0,$1F,$E8,$2F,$F4,$C7
	.dc.b	$F9,$F7,$6F,$EF,$DC,$F2,$1C,$D7,$E2,$36,$F0,$57,$E3,$B2,$B3,$FE
	.dc.b	$78,$AF,$E8,$CF,$82,$7A,$7C,$37,$E2,$65,$F0,$D6,$31,$F0,$D9,$F8
	.dc.b	$6B,$87,$13,$A3,$C7,$E1,$A4,$67,$5F,$57,$C3,$61,$F7,$7C,$37,$7F
	.dc.b	$86,$F0,$36,$B8,$70,$7B,$7E,$17,$C8,$79,$7E,$16,$CE,$B7,$67,$D7
	.dc.b	$83,$E3,$69,$6B,$E0,$7C,$27,$B7,$E1,$7D,$CD,$8E,$5E,$E7,$C1,$3E
	.dc.b	$17,$C0,$D8,$ED,$8F,$3E,$0F,$5C,$60,$F1,$7C,$27,$C9,$3E,$57,$C2
	.dc.b	$D8,$F3,$7C,$27,$CD,$F8,$5F,$9C,$D8,$F5,$7C,$27,$D1,$3E,$97,$C2
	.dc.b	$D8,$F8,$7F,$C0,$3F,$D1,$78,$7F,$A0,$3F,$D0,$E3,$27,$F0,$17,$19
	.dc.b	$FE,$B5,$FE,$81,$1C,$BE,$13,$5F,$09,$B7,$B5,$7E,$47,$1F,$0D,$79
	.dc.b	$0F,$0F,$C1,$F5,$F8,$4E,$CD,$8F,$2F,$B7,$E1,$F8,$0F,$13,$63,$96
	.dc.b	$1D,$5F,$B1,$AF,$83,$5F,$5F,$C1,$FA,$4F,$57,$C2,$5A,$F8,$3F,$09
	.dc.b	$ED,$3D,$DF,$09,$63,$E2,$FC,$27,$C1,$F8,$4F,$84,$D8,$F9,$3F,$09
	.dc.b	$F1,$7E,$13,$E3,$36,$3E,$6F,$C1,$FC,$9F,$84,$F9,$4D,$8E,$A7,$D2
	.dc.b	$C3,$AB,$4F,$AB,$F0,$9F,$44,$FA,$5F,$09,$63,$EC,$FC,$27,$D5,$3F
	.dc.b	$9B,$72,$FF,$1D,$E1,$FB,$71,$D3,$8F,$EC,$63,$67,$EF,$BC,$3E,$13
	.dc.b	$8F,$C2,$5C,$7F,$4F,$8F,$F4,$26,$FC,$1F,$63,$BF,$C2,$58,$DF,$C1
	.dc.b	$F8,$BE,$13,$C6,$D8,$E7,$F0,$7E,$6F,$84,$F3,$B6,$3B,$FC,$17,$A8
	.dc.b	$F5,$FC,$15,$8E,$A7,$93,$0E,$AD,$3C,$FF,$05,$F0,$8F,$87,$F0,$56
	.dc.b	$3D,$7E,$E7,$C6,$3E,$3F,$B9,$63,$AF,$27,$F4,$9E,$06,$3D,$A3,$0F
	.dc.b	$CE,$F8,$29,$5F,$1B,$E0,$AD,$71,$F8,$1F,$A6,$7D,$4F,$82,$B5,$F3
	.dc.b	$3E,$0B,$EB,$FC,$17,$D8,$6C,$7F,$16,$FE,$01,$FE,$66,$4F,$88,$FE
	.dc.b	$E4,$E3,$E0,$EF,$C2,$61,$B5,$D4,$FB,$38,$75,$69,$D7,$E1,$39,$9D
	.dc.b	$3E,$12,$C7,$87,$D9,$EF,$EF,$78,$1B,$1D,$77,$F8,$D8,$75,$6C,$7A
	.dc.b	$7E,$0F,$CE,$7A,$3E,$12,$C7,$B7,$E0,$FF,$AA,$D7,$F7,$7D,$91,$8F
	.dc.b	$17,$FC,$A5,$18,$9F,$89,$F0,$79,$BF,$EB,$BF,$08,$A7,$C1,$66,$FF
	.dc.b	$D5,$30,$F8,$3C,$56,$DB,$1C,$27,$E8,$7C,$1E,$08,$FD,$3F,$83,$DF
	.dc.b	$C1,$C8,$FA,$CF,$2F,$83,$E7,$97,$4C,$65,$7D,$B8,$EF,$F0,$52,$3E
	.dc.b	$F3,$E2,$F8,$3F,$1F,$C0,$79,$3D,$D9,$5F,$83,$8E,$AF,$E1,$70,$EA
	.dc.b	$FC,$9C,$3D,$BF,$C0,$3F,$E7,$D8,$FF,$11,$37,$8B,$FC,$F8,$DE,$1F
	.dc.b	$92,$FF,$A0,$67,$FB,$2B,$FC,$F1,$E3,$F0,$7F,$E4,$18,$7C,$1D,$8C
	.dc.b	$78,$D9,$F8,$3B,$87,$27,$87,$C1,$48,$EB,$F0,$77,$2F,$2F,$C1,$FA
	.dc.b	$3E,$0E,$47,$8A,$3D,$7F,$07,$23,$CA,$FB,$BE,$0F,$DF,$F0,$72,$7D
	.dc.b	$2F,$C2,$F7,$3E,$1F,$C1,$7A,$4F,$57,$C1,$DD,$3D,$FF,$05,$ED,$F8
	.dc.b	$3F,$73,$63,$E1,$FC,$1F,$C1,$3E,$17,$C1,$D8,$F8,$FF,$07,$FE,$65
	.dc.b	$FE,$01,$FE,$61,$1F,$32,$31,$F1,$19,$F8,$59,$3F,$E1,$2A,$FE,$59
	.dc.b	$BF,$F9,$5D,$FF,$36,$37,$E1,$3A,$1D,$7E,$12,$D7,$1F,$84,$FC,$DD
	.dc.b	$9A,$63,$DC,$CF,$C1,$CC,$7C,$0F,$0F,$84,$E3,$F0,$92,$7C,$8F,$4F
	.dc.b	$84,$EB,$F0,$9E,$DF,$84,$F7,$37,$4F,$57,$C2,$7C,$1F,$84,$F8,$4D
	.dc.b	$8F,$77,$C2,$7C,$53,$E3,$7C,$25,$8F,$85,$F0,$9F,$27,$E1,$3E,$53
	.dc.b	$63,$E3,$7C,$27,$CD,$F7,$7E,$73,$63,$FC,$77,$F0,$1F,$E5,$83,$FC
	.dc.b	$AF,$C0,$F6,$91,$9F,$84,$FF,$4D,$3F,$E5,$69,$FF,$43,$3F,$E5,$4A
	.dc.b	$E7,$F0,$7A,$F8,$4E,$F1,$F5,$F2,$B7,$F5,$DE,$7F,$09,$D3,$0B,$5E
	.dc.b	$47,$BF,$C1,$F8,$3E,$12,$CF,$A1,$E7,$EA,$FF,$BB,$FB,$58,$E2,$BF
	.dc.b	$8E,$C6,$1F,$7F,$C2,$7A,$8F,$5F,$C2,$58,$F8,$7F,$09,$EE,$3D,$FF
	.dc.b	$09,$63,$E3,$FB,$5F,$0B,$DB,$F8,$6D,$8F,$97,$F0,$7F,$18,$F8,$FF
	.dc.b	$07,$63,$E7,$FC,$1F,$CA,$3E,$5F,$C1,$D8,$EA,$7D,$4C,$3A,$B4,$FA
	.dc.b	$FF,$07,$F4,$8F,$A7,$F0,$76,$3E,$DF,$C1,$7D,$63,$EB,$FC,$1D,$8E
	.dc.b	$BD,$1F,$63,$0E,$AD,$8F,$F3,$EF,$F0,$0F,$F9,$6D,$FE,$8A,$FF,$96
	.dc.b	$39,$51,$83,$FD,$B9,$19,$77,$F0,$58,$F8,$3C,$B6,$3A,$FC,$1F,$0F
	.dc.b	$83,$E2,$D8,$F0,$FC,$1F,$43,$AF,$C1,$D8,$F2,$FC,$17,$83,$E0,$FC
	.dc.b	$2D,$8E,$A7,$AB,$0E,$AD,$3D,$BF,$07,$E8,$3D,$3F,$07,$63,$E0,$FC
	.dc.b	$1F,$B0,$F6,$FC,$1D,$8E,$BC,$3F,$EE,$5E,$06,$3D,$A3,$0F,$C4,$F8
	.dc.b	$39,$5F,$33,$E0,$ED,$71,$F8,$2F,$93,$F0,$7F,$29,$B5,$F4,$BE,$0F
	.dc.b	$E6,$FC,$1F,$CE,$6C,$7D,$6F,$83,$FA,$78,$7F,$BD,$FF,$80,$7F,$94
	.dc.b	$BF,$B9,$D5,$3A,$E9,$DF,$A2,$BD,$B7,$E0,$F5,$8F,$F8,$12,$7F,$95
	.dc.b	$D7,$07,$F9,$7B,$67,$C9,$3F,$88,$CE,$E5,$F1,$5E,$6F,$F0,$BB,$7C
	.dc.b	$1E,$31,$F1,$7C,$1D,$9C,$E1,$FE,$CE,$51,$9F,$83,$93,$9F,$83,$B8
	.dc.b	$FD,$3D,$FE,$2B,$67,$C8,$73,$F8,$3E,$DF,$07,$DD,$B1,$DF,$E0,$FC
	.dc.b	$5F,$07,$E3,$6C,$78,$FE,$0B,$CD,$F0,$7E,$76,$C7,$9F,$E0,$BD,$5F
	.dc.b	$07,$EB,$6C,$7A,$FE,$0F,$DD,$F0,$7E,$F6,$C7,$53,$E0,$61,$D5,$A7
	.dc.b	$C3,$F8,$2F,$8C,$7C,$7F,$83,$B1,$F1,$FE,$0F,$E5,$1F,$2F,$DC,$B1
	.dc.b	$D7,$87,$F0,$D8,$75,$6C,$7F,$BF,$3F,$80,$51,$F5,$5F,$E8,$7D,$07
	.dc.b	$33,$BF,$84,$D7,$C2,$48,$E5,$F0,$97,$07,$7F,$E4,$30,$7C,$27,$5F
	.dc.b	$84,$EC,$DC,$3C,$5F,$09,$E1,$3C,$5F,$09,$63,$CD,$F0,$7E,$5F,$83
	.dc.b	$F3,$36,$3D,$57,$CF,$3F,$0C,$7F,$97,$B2,$F0,$BE,$FC,$BD,$DF,$07
	.dc.b	$EF,$F8,$3B,$3F,$0D,$F8,$5F,$07,$23,$E3,$7B,$77,$1E,$A7,$F6,$7C
	.dc.b	$7A,$B8,$E3,$F9,$19,$7B,$AF,$C0,$7C,$FF,$83,$F9,$C7,$CF,$F8,$3B
	.dc.b	$1F,$4F,$E0,$FE,$91,$F4,$FE,$0E,$C7,$D7,$F8,$3F,$AD,$F0,$7F,$5D
	.dc.b	$B1,$F6,$FE,$0F,$ED,$7C,$1F,$DB,$6C,$7D,$EE,$8F,$E0,$7E,$67,$F3
	.dc.b	$FD,$FF,$C0,$0D,$DF,$FC,$49,$F1,$BF,$B7,$8C,$70,$B3,$EF,$7E,$F8
	.dc.b	$C6,$3F,$8B,$D3,$87,$C1,$F1,$F8,$4B,$3E,$13,$FD,$23,$8F,$F9,$91
	.dc.b	$C6,$1E,$2F,$84,$F0,$7B,$9E,$16,$CF,$C1,$3B,$16,$7C,$58,$F9,$FE
	.dc.b	$12,$4F,$4D,$BF,$C7,$DF,$09,$8B,$FA,$D9,$3E,$0F,$DD,$F0,$9E,$F6
	.dc.b	$E1,$C6,$5F,$09,$F0,$CF,$89,$F0,$96,$75,$F0,$9F,$1F,$E1,$3E,$43
	.dc.b	$63,$97,$C2,$7C,$B3,$E6,$7C,$25,$8E,$D8,$73,$D3,$CB,$8D,$E8,$FA
	.dc.b	$FF,$07,$F4,$FE,$13,$EA,$37,$0F,$37,$B9,$F5,$FE,$07,$EC,$36,$3A
	.dc.b	$93,$1E,$AD,$C3,$F0,$7F,$07,$FC,$D7,$F8,$07,$FA,$2F,$0F,$F4,$AA
	.dc.b	$F9,$51,$FD,$20,$FE,$8A,$FF,$B2,$9B,$7F,$E4,$97,$FA,$3C,$67,$E1
	.dc.b	$3A,$7C,$24,$8E,$7F,$07,$6F,$F5,$17,$FE,$73,$9E,$FF,$07,$E3,$C7
	.dc.b	$C8,$F9,$7E,$0B,$CD,$F0,$76,$BC,$AF,$3D,$7F,$AF,$E1,$FD,$33,$1E
	.dc.b	$B9,$66,$FD,$1F,$05,$EF,$F8,$3F,$80,$D8,$F7,$FC,$1F,$C3,$F6,$FE
	.dc.b	$23,$63,$E1,$EF,$CF,$1F,$81,$8D,$EF,$66,$F8,$2F,$98,$7C,$DF,$83
	.dc.b	$B8,$FC,$CF,$82,$FF,$76,$D1,$A6,$3E,$0F,$3F,$07,$33,$FA,$6F,$0F
	.dc.b	$83,$E3,$F0,$72,$7E,$BB,$D3,$E0,$FF,$DF,$3F,$C0,$3F,$CF,$A7,$F4
	.dc.b	$8B,$E1,$32,$6A,$3B,$1F,$53,$0B,$CD,$F3,$31,$E7,$F0,$92,$B9,$BD
	.dc.b	$BE,$13,$BF,$C0,$49,$EE,$F8,$BD,$CE,$7C,$3F,$89,$CA,$D7,$E2,$71
	.dc.b	$F3,$FC,$27,$A3,$E1,$24,$FA,$23,$D7,$F0,$7E,$CF,$84,$93,$EC,$8F
	.dc.b	$7F,$C2,$48,$F7,$BF,$0B,$E1,$3E,$1F,$C2,$49,$F8,$6F,$C6,$F8,$4F
	.dc.b	$8F,$F0,$9F,$23,$E1,$25,$7C,$88,$E8,$FF,$3F,$E6,$63,$8E,$5F,$F1
	.dc.b	$7C,$32,$FA,$3F,$09,$F4,$BE,$12,$61,$F4,$A3,$EA,$FC,$0F,$D6,$F7
	.dc.b	$24,$FD,$68,$E9,$CB,$FD,$DB,$C2,$65,$E0,$AF,$69,$F7,$32,$FB,$FF
	.dc.b	$07,$23,$EF,$BF,$82,$F8,$4F,$C1,$6B,$FE,$44,$FE,$01,$FF,$25,$D7
	.dc.b	$F5,$95,$F0,$BA,$CF,$C1,$18,$9F,$E5,$8F,$0A,$E5,$F0,$9B,$9F,$F6
	.dc.b	$54,$A3,$1A,$23,$13,$CB,$E1,$33,$0F,$83,$BF,$09,$2B,$BC,$6B,$E1
	.dc.b	$37,$F0,$92,$7C,$71,$CB,$29,$1E,$67,$AE,$3D,$BE,$0F,$BF,$C2,$4A
	.dc.b	$F5,$C7,$8B,$E0,$3C,$7F,$01,$27,$DF,$1D,$1F,$8D,$85,$DF,$E5,$63
	.dc.b	$EA,$F8,$39,$5F,$19,$F6,$FC,$27,$BB,$E1,$3F,$CC,$5C,$26,$FC,$26
	.dc.b	$1C,$FC,$1E,$BE,$12,$F3,$7D,$07,$8F,$C1,$C8,$FA,$7F,$09,$74,$FB
	.dc.b	$5F,$09,$FE,$6F,$E8,$FF,$3F,$E9,$FC,$EE,$7F,$C2,$D1,$8C,$BB,$61
	.dc.b	$F7,$1B,$3A,$F8,$4E,$5F,$19,$CD,$B5,$FD,$B6,$FF,$97,$71,$7F,$A3
	.dc.b	$9F,$F3,$CC,$67,$DD,$F1,$7B,$BE,$AC,$7F,$1D,$BD,$C3,$F2,$4F,$9F
	.dc.b	$E1,$7D,$1F,$0B,$71,$EC,$FA,$FE,$17,$D9,$F0,$B6,$7C,$4F,$BF,$E1
	.dc.b	$24,$79,$7E,$16,$E7,$F0,$3E,$13,$E2,$7C,$2F,$C5,$6D,$7A,$FE,$17
	.dc.b	$E4,$1F,$27,$E1,$6C,$7B,$F1,$E7,$BF,$BF,$2B,$3F,$B5,$E2,$FE,$74
	.dc.b	$FC,$5F,$84,$FA,$5F,$09,$F4,$DF,$A9,$EE,$C9,$F9,$5E,$ED,$DF,$EA
	.dc.b	$E7,$D1,$79,$7F,$ED,$0C,$F0,$FF,$B4,$BE,$13,$EF,$7C,$27,$DF,$6E
	.dc.b	$3F,$E2,$1F,$E0,$1F,$F3,$CE,$1F,$E5,$56,$EF,$FE,$C4,$7F,$CF,$E6
	.dc.b	$27,$F8,$72,$F8,$4F,$F5,$C3,$63,$A7,$C2,$62,$73,$F0,$92,$7C,$1F
	.dc.b	$09,$71,$F0,$7C,$1F,$2F,$84,$E6,$DA,$F3,$61,$CD,$EA,$60,$F5,$7C
	.dc.b	$1F,$88,$F1,$FC,$25,$8F,$77,$C0,$79,$8F,$3F,$B7,$63,$96,$5F,$07
	.dc.b	$0F,$5E,$BD,$E3,$E2,$7C,$27,$BB,$E1,$3D,$ED,$BF,$95,$F0,$7F,$08
	.dc.b	$F8,$7F,$09,$63,$E7,$7C,$1F,$C6,$F8,$4F,$8E,$D8,$FA,$5F,$07,$F2
	.dc.b	$8F,$97,$F0,$96,$3E,$B7,$C1,$FC,$E3,$E7,$FC,$1D,$8E,$A7,$DB,$C3
	.dc.b	$AB,$4F,$F2,$F7,$F0,$0F,$F9,$E4,$FE,$72,$E0,$FF,$78,$38,$DB,$FD
	.dc.b	$2C,$FF,$35,$8E,$3E,$E6,$BD,$CD,$BD,$AB,$F1,$17,$7B,$3F,$10,$F4
	.dc.b	$F8,$3E,$BE,$3D,$C7,$C6,$F8,$3E,$0B,$C3,$F0,$76,$7C,$EF,$93,$E0
	.dc.b	$A4,$7A,$BE,$0E,$E7,$E6,$F8,$2F,$4F,$C1,$FA,$9B,$5F,$03,$E0,$FD
	.dc.b	$BF,$07,$EE,$6C,$7C,$4C,$79,$EE,$8F,$ED,$5C,$4C,$F3,$18,$F8,$2F
	.dc.b	$8E,$7C,$8F,$83,$B1,$F3,$BD,$FF,$97,$ED,$FC,$C6,$C7,$72,$6F,$F3
	.dc.b	$B0,$EA,$FE,$EB,$17,$C1,$FD,$43,$EA,$FC,$1D,$BF,$B7,$F0,$7E,$7B
	.dc.b	$FF,$66,$7F,$00,$BE,$37,$ED,$CE,$77,$D7,$A4,$AD,$61,$FE,$A1,$7F
	.dc.b	$CE,$B1,$9F,$83,$E1,$3F,$E0,$B9,$FF,$3B,$E3,$FC,$39,$D9,$FC,$F2
	.dc.b	$35,$F0,$5C,$27,$5F,$07,$70,$F0,$4D,$BF,$F3,$D5,$7A,$3E,$0F,$1F
	.dc.b	$07,$97,$59,$4C,$BC,$DF,$09,$6F,$DF,$F0,$9C,$CE,$9F,$09,$6B,$D9
	.dc.b	$EE,$77,$F7,$3C,$0D,$8F,$81,$8F,$3E,$0F,$8B,$AF,$BB,$7F,$85,$5F
	.dc.b	$17,$E0,$FD,$1F,$09,$E9,$6C,$7C,$9F,$84,$FF,$0C,$DB,$E1,$29,$88
	.dc.b	$CF,$C1,$CE,$0F,$A0,$F0,$F8,$4E,$3F,$09,$27,$EA,$3D,$3E,$13,$AF
	.dc.b	$C1,$C9,$FB,$0F,$83,$E1,$3F,$C5,$3F,$01,$FE,$79,$97,$0F,$E0,$23
	.dc.b	$FC,$F5,$1F,$ED,$F6,$BA,$DB,$FC,$91,$1C,$67,$87,$B9,$D0,$95,$C7
	.dc.b	$DC,$EF,$F0,$7F,$76,$CD,$F8,$3C,$39,$89,$8F,$87,$E0,$EE,$1E,$13
	.dc.b	$FA,$CC,$BE,$0F,$A7,$C1,$F5,$6D,$FA,$7E,$0F,$DF,$F0,$73,$0C,$1E
	.dc.b	$DF,$82,$F8,$64,$7E,$07,$C1,$FC,$6F,$83,$F8,$FF,$07,$27,$E2,$47
	.dc.b	$CA,$F8,$3F,$97,$EE,$C9,$F9,$11,$F3,$BE,$0E,$47,$CB,$7E,$8F,$C1
	.dc.b	$FD,$2F,$83,$93,$F3,$DF,$AB,$F0,$7F,$E5,$EF,$C7,$BF,$87,$70,$2B
	.dc.b	$E1,$78,$A9,$FE,$65,$D5,$FB,$87,$73,$A7,$35,$FE,$7F,$7F,$9B,$46
	.dc.b	$7E,$13,$F6,$11,$8A,$E5,$18,$F8,$3C,$FC,$1D,$C3,$C5,$5D,$E3,$8F
	.dc.b	$C1,$C9,$C9,$DF,$E0,$F1,$97,$8F,$DF,$EE,$78,$3D,$BB,$5C,$27,$FC
	.dc.b	$2B,$A3,$E0,$BC,$A7,$9B,$E0,$ED,$6F,$7F,$85,$F0,$5E,$AF,$83,$F5
	.dc.b	$B6,$75,$87,$F9,$95,$BF,$07,$F0,$3E,$0F,$E0,$BF,$0B,$E0,$BE,$1F
	.dc.b	$C1,$DA,$E2,$FC,$6F,$82,$FA,$9F,$05,$37,$EB,$1F,$5F,$E0,$64,$78
	.dc.b	$1F,$F4,$BF,$E3,$FF,$CB,$E3,$D0,$BD,$9C,$92,$33,$EF,$70,$F8,$2E
	.dc.b	$3F,$05,$27,$8C,$74,$F8,$19,$1D,$1E,$FF,$05,$E0,$F8,$29,$3E,$07
	.dc.b	$C7,$F0,$5E,$4F,$81,$FE,$61,$A1,$BF,$05,$87,$3F,$03,$38,$FD,$3F
	.dc.b	$05,$72,$FC,$9E,$0F,$80,$E6,$74,$F8,$1B,$7F,$03,$E0,$7E,$19,$2F
	.dc.b	$E1,$FC,$0F,$C6,$F8,$1F,$8F,$F0,$32,$7E,$3C,$7C,$AF,$81,$91,$F2
	.dc.b	$9F,$9B,$F0,$3F,$EC,$4F,$C8,$3F,$8A,$65,$FE,$DD,$7F,$8C,$F9,$9F
	.dc.b	$D7,$1B,$19,$BE,$1F,$05,$C7,$E0,$A4,$7E,$36,$BF,$CB,$0D,$F8,$1E
	.dc.b	$C4,$75,$F0,$5E,$1F,$81,$91,$C5,$F2,$7C,$17,$97,$E0,$64,$F5,$7D
	.dc.b	$1F,$05,$E9,$F8,$29,$3E,$17,$D9,$F0,$5E,$DF,$82,$FE,$CD,$A3,$E0
	.dc.b	$AC,$61,$CF,$C0,$EB,$E0,$AF,$17,$AD,$E3,$F0,$3C,$BE,$0A,$CF,$BD
	.dc.b	$EB,$F0,$5F,$2F,$E0,$A6,$DF,$0E,$3E,$77,$C1,$7C,$FF,$82,$93,$F1
	.dc.b	$E3,$FD,$EB,$E4,$FF,$BE,$F1,$FE,$31,$EC,$A3,$FE,$00,$57,$F1,$A6
	.dc.b	$F5,$FF,$84,$1E,$11,$FE,$96,$7F,$DF,$AE,$FE,$0B,$A9,$DB,$E1,$BB
	.dc.b	$FC,$1F,$83,$E0,$E5,$7F,$05,$B4,$F9,$3E,$0F,$CB,$F0,$72,$B5,$1E
	.dc.b	$8F,$83,$91,$C5,$F5,$FC,$1F,$B3,$E0,$E4,$F5,$7D,$FF,$07,$F0,$3E
	.dc.b	$0B,$E0,$FC,$1C,$AF,$14,$7C,$4F,$81,$F8,$BF,$05,$27,$CD,$1F,$23
	.dc.b	$E0,$BE,$4F,$C1,$49,$F5,$47,$CC,$F8,$1F,$9B,$F0,$52,$7D,$D1,$F4
	.dc.b	$3E,$0B,$FA,$34,$7F,$46,$FE,$01,$FE,$20,$DF,$86,$D3,$FC,$A6,$3F
	.dc.b	$89,$37,$E1,$38,$F5,$7F,$90,$A7,$33,$CF,$E1,$33,$DB,$F7,$25,$F0
	.dc.b	$78,$AF,$F6,$6C,$61,$ED,$19,$EE,$4B,$E1,$30,$BE,$12,$92,$FC,$DD
	.dc.b	$BA,$F8,$49,$1E,$97,$8F,$C2,$72,$F8,$39,$3E,$D7,$AF,$C2,$76,$F8
	.dc.b	$49,$3F,$05,$F0,$FC,$27,$8B,$E0,$FC,$7F,$07,$2B,$E3,$47,$9B,$E0
	.dc.b	$FC,$FF,$07,$27,$E5,$47,$AB,$E0,$A4,$7C,$D7,$DB,$F0,$7E,$EF,$83
	.dc.b	$F7,$FC,$1C,$AF,$A5,$1F,$17,$F8,$07,$F8,$CC,$7F,$BD,$9F,$E2,$ED
	.dc.b	$F8,$DD,$3B,$F8,$5E,$1F,$0B,$23,$FB,$25,$7F,$18,$7F,$C0,$EF,$F8
	.dc.b	$C1,$CF,$C2,$77,$3C,$1F,$0B,$E1,$F8,$5F,$17,$C2,$CA,$E7,$1E,$5F
	.dc.b	$85,$91,$D9,$F4,$7C,$2F,$A7,$E1,$64,$F8,$9F,$67,$C2,$FB,$7E,$12
	.dc.b	$4F,$99,$F8,$1F,$09,$F0,$7E,$12,$4F,$A9,$F8,$9F,$09,$F1,$7E,$12
	.dc.b	$4F,$B9,$F9,$1F,$09,$F2,$7E,$0F,$E5,$7C,$24,$AF,$87,$1F,$37,$E0
	.dc.b	$E4,$7C,$67,$E8,$7C,$27,$E5,$7F,$80,$7F,$88,$F9,$87,$F8,$AF,$9D
	.dc.b	$FC,$69,$B6,$9D,$4F,$1F,$84,$93,$FE,$15,$AF,$E3,$3E,$8F,$FA,$B1
	.dc.b	$7C,$2F,$7F,$84,$F0,$7C,$2C,$BE,$11,$E3,$F8,$4F,$27,$C2,$C9,$E9
	.dc.b	$1E,$7F,$85,$91,$DD,$F5,$7C,$2F,$AF,$E1,$7D,$9F,$0B,$2B,$C9,$1E
	.dc.b	$FF,$84,$F8,$1F,$0B,$27,$D1,$1F,$0F,$E1,$3E,$27,$C2,$C9,$F6,$47
	.dc.b	$C7,$F8,$49,$1E,$F7,$E5,$7C,$2F,$CB,$F8,$4F,$99,$F0,$B2,$BE,$24
	.dc.b	$7C,$FF,$85,$91,$F1,$DF,$E0,$DC,$1F,$C6,$39,$BF,$DA,$6F,$F1,$26
	.dc.b	$FC,$26,$8D,$FC,$27,$0F,$84,$98,$7F,$91,$6F,$F8,$93,$FE,$68,$7F
	.dc.b	$88,$39,$F8,$2E,$FF,$05,$23,$07,$1F,$82,$F1,$FC,$14,$8C,$1D,$7E
	.dc.b	$0B,$CF,$F0,$52,$30,$78,$7E,$0B,$D7,$F0,$52,$30,$79,$7E,$07,$DF
	.dc.b	$F0,$52,$30,$7A,$7E,$07,$E1,$91,$F6,$7C,$17,$C6,$F8,$2F,$8F,$F0
	.dc.b	$52,$7E,$04,$7C,$AF,$81,$91,$F0,$DF,$9B,$F0,$5F,$CC,$FF,$80,$7F
	.dc.b	$87,$F9,$A7,$F8,$7B,$7E,$0F,$57,$BF,$83,$E1,$F0,$72,$3F,$CE,$57
	.dc.b	$FC,$3F,$B1,$E8,$7F,$80,$23,$F8,$84,$FF,$A6,$1B,$3A,$8F,$D5,$46
	.dc.b	$27,$87,$C1,$79,$7E,$03,$CD,$F0,$32,$FB,$C7,$A7,$E0,$64,$78,$9F
	.dc.b	$67,$C0,$FB,$7E,$07,$DD,$F0,$32,$BC,$F1,$F0,$7E,$06,$47,$A9,$F8
	.dc.b	$9F,$03,$F1,$7E,$06,$4F,$B9,$F9,$1F,$03,$F2,$7E,$07,$E5,$7C,$0C
	.dc.b	$AF,$87,$1F,$37,$E0,$3E,$77,$C0,$C9,$F8,$F1,$F4,$7E,$06,$47,$CA
	.dc.b	$7E,$A7,$C0,$FF,$67,$FB,$EF,$E2,$64,$E5,$C7,$C7,$E7,$E0,$E6,$1F
	.dc.b	$EA,$3A,$FE,$1F,$5F,$E3,$07,$F8,$7C,$67,$E0,$3A,$7C,$0C,$8C,$1C
	.dc.b	$7D,$0F,$07,$C0,$C8,$C1,$D7,$E0,$7C,$84,$7C,$1F,$03,$E7,$F8,$0F
	.dc.b	$47,$C0,$C9,$F2,$47,$AF,$E0,$24,$79,$DF,$77,$C0,$FB,$FE,$02,$4F
	.dc.b	$AD,$F8,$5F,$03,$F0,$FE,$07,$E2,$7C,$0C,$AF,$81,$1F,$1F,$E0,$24
	.dc.b	$7C,$37,$E5,$7C,$0F,$CB,$F8,$1F,$F0,$97,$DF,$FF,$0D,$E6,$FF,$2A
	.dc.b	$B6,$FF,$9E,$73,$2B,$7F,$05,$27,$FD,$A1,$5F,$C4,$EF,$8F,$C1,$74
	.dc.b	$F8,$2E,$BF,$11,$2F,$71,$E0,$F8,$39,$1C,$9F,$1F,$C1,$F9,$3E,$0E
	.dc.b	$4F,$67,$CF,$F0,$7E,$8F,$83,$93,$E2,$7D,$7F,$07,$EC,$F8,$3F,$6F
	.dc.b	$C1,$CA,$F3,$C7,$C0,$F8,$39,$1E,$A7,$E1,$FC,$1F,$C4,$F8,$3F,$8B
	.dc.b	$F0,$72,$BD,$F1,$F2,$3E,$0A,$47,$C2,$7E,$5F,$C1,$FC,$CF,$83,$F9
	.dc.b	$BF,$07,$2B,$E3,$C7,$F9,$07,$DC,$FE,$1E,$48,$C7,$C1,$67,$E0,$A4
	.dc.b	$7F,$BF,$67,$F8,$67,$4F,$FA,$29,$7C,$17,$3F,$86,$95,$B7,$B7,$C0
	.dc.b	$F7,$F8,$1F,$05,$CA,$E9,$1E,$3F,$81,$91,$DD,$F3,$7C,$0F,$9F,$E0
	.dc.b	$3D,$1F,$03,$2B,$C9,$1E,$BF,$81,$91,$E7,$7D,$DF,$03,$EF,$F8,$1F
	.dc.b	$81,$F0,$32,$BD,$91,$F0,$FE,$02,$47,$BD,$F8,$DF,$03,$F1,$FE,$02
	.dc.b	$4F,$C3,$7E,$57,$C0,$FC,$BF,$81,$F9,$9F,$03,$2B,$E4,$47,$F9,$BF
	.dc.b	$F0,$0F,$F0,$D9,$FE,$18,$F8,$53,$4C,$FC,$17,$F3,$3F,$85,$FE,$69
	.dc.b	$3A,$F8,$6C,$75,$FF,$AE,$AD,$78,$C7,$FC,$45,$1F,$E1,$C9,$43,$0E
	.dc.b	$73,$FE,$B2,$9C,$C6,$3E,$13,$96,$3E,$4F,$84,$B7,$BC,$39,$61,$E1
	.dc.b	$F8,$4C,$9A,$F8,$4B,$1D,$0F,$37,$C2,$72,$F8,$4E,$6D,$8D,$5F,$A3
	.dc.b	$E0,$FB,$9E,$0F,$84,$B3,$AD,$3E,$3F,$C2,$79,$3E,$13,$CA,$D9,$D6
	.dc.b	$AF,$84,$F4,$9E,$AF,$84,$B3,$AF,$83,$F6,$9E,$EF,$84,$B1,$CB,$E1
	.dc.b	$3E,$09,$F0,$BE,$12,$C7,$6F,$83,$F8,$A7,$C8,$FC,$09,$F1,$78,$9F
	.dc.b	$44,$3E,$17,$5F,$09,$B6,$CE,$BE,$0F,$97,$C2,$73,$6C,$72,$F8,$4E
	.dc.b	$C7,$7F,$84,$B1,$DB,$E1,$3C,$47,$8F,$E1,$2C,$78,$BE,$13,$CC,$79
	.dc.b	$FE,$12,$C7,$9B,$E0,$FD,$47,$AF,$E1,$2C,$7A,$BE,$13,$DD,$F0,$9E
	.dc.b	$F6,$C7,$BB,$E0,$FE,$17,$C2,$7C,$36,$C7,$C2,$F8,$4F,$8D,$F0,$9F
	.dc.b	$1D,$B1,$F1,$BE,$0F,$E5,$7C,$27,$CB,$6C,$7E,$07,$F8,$07,$F3,$6D
	.dc.b	$F1,$FF,$2F,$0F,$85,$B1,$8F,$85,$CF,$C2,$DB,$E4,$F0,$F8,$59,$1D
	.dc.b	$7E,$16,$DF,$E7,$8F,$F4,$66,$2F,$84,$EF,$F0,$BE,$06,$DF,$97,$E1
	.dc.b	$7C,$67,$93,$E1,$6C,$7A,$7E,$13,$CF,$F0,$BE,$86,$C7,$B7,$E1,$7D
	.dc.b	$67,$B3,$E1,$6C,$7C,$1F,$84,$F7,$FC,$2F,$C0,$6C,$7C,$5F,$85,$F8
	.dc.b	$7F,$0B,$F1,$1B,$1F,$27,$E1,$7E,$39,$F2,$3E,$16,$C7,$CD,$F8,$4F
	.dc.b	$96,$7C,$CF,$85,$B1,$FB,$1F,$1F,$FC,$96,$DF,$88,$FE,$0F,$18,$E6
	.dc.b	$F9,$1E,$DA,$13,$AF,$85,$97,$CF,$E1,$AE,$1C,$2F,$F8,$77,$C3,$7F
	.dc.b	$20,$6D,$7F,$1A,$C7,$A7,$C2,$F8,$3E,$1B,$C3,$F0,$DE,$2F,$86,$B7
	.dc.b	$E7,$7C,$BF,$0B,$23,$D5,$F0,$D7,$4F,$07,$C3,$7A,$BE,$1B,$D6,$DA
	.dc.b	$F8,$1F,$0B,$EE,$3D,$FF,$0D,$63,$E2,$7C,$2F,$C2,$3E,$1F,$C2,$D8
	.dc.b	$F9,$1F,$0B,$F1,$BE,$17,$E3,$B6,$3E,$67,$C2,$FC,$A3,$E5,$FC,$2D
	.dc.b	$8F,$A1,$F0,$BF,$38,$FE,$6B,$FC,$03,$FC,$3A,$32,$7F,$25,$7F,$87
	.dc.b	$37,$E0,$F7,$F0,$9C,$1E,$3F,$07,$CB,$E1,$2D,$71,$7A,$FC,$27,$6F
	.dc.b	$84,$B3,$D5,$F0,$F6,$F8,$BE,$12,$CF,$85,$F2,$FC,$27,$9B,$E1,$2C
	.dc.b	$F9,$5F,$4F,$C2,$48,$F4,$7C,$25,$EB,$FC,$4F,$03,$E0,$FD,$FF,$09
	.dc.b	$F0,$1B,$7E,$FF,$83,$F8,$67,$C4,$F8,$4B,$1F,$0F,$E0,$FE,$39,$F2
	.dc.b	$3E,$12,$C7,$C7,$F8,$3F,$96,$7C,$CF,$84,$B1,$FD,$3F,$F0,$8F,$E1
	.dc.b	$FA,$FE,$71,$BE,$67,$F7,$47,$F8,$84,$EE,$7F,$B4,$C3,$E1,$29,$8F
	.dc.b	$84,$B5,$D7,$E1,$36,$70,$F8,$4B,$1E,$1F,$82,$E6,$74,$F8,$3B,$1E
	.dc.b	$5F,$83,$EE,$78,$3E,$0E,$C7,$A7,$E0,$BC,$7F,$07,$E4,$6C,$7B,$7E
	.dc.b	$0F,$CE,$7A,$3E,$0E,$C7,$C1,$F8,$2F,$5F,$C1,$FB,$1B,$1F,$17,$E0
	.dc.b	$BD,$FF,$07,$F0,$1B,$1F,$27,$E0,$FE,$19,$F1,$3E,$0E,$C7,$CD,$F8
	.dc.b	$3F,$8F,$F0,$7F,$21,$B1,$FD,$FF,$F0,$9F,$E1,$9D,$5F,$E2,$26,$C7
	.dc.b	$F0,$37,$F8,$6B,$B3,$FC,$32,$FF,$89,$1E,$5F,$05,$8F,$83,$CB,$AF
	.dc.b	$81,$93,$DF,$E0,$AF,$77,$F1,$AC,$1F,$05,$D0,$EB,$F0,$56,$FC,$DF
	.dc.b	$05,$E0,$3C,$3F,$05,$63,$D5,$F0,$3E,$43,$CB,$F0,$56,$3D,$DF,$05
	.dc.b	$E8,$3D,$3F,$05,$63,$E1,$7C,$17,$B3,$E0,$BD,$AD,$8F,$8D,$F0,$3F
	.dc.b	$03,$E0,$BE,$0B,$63,$E5,$7C,$17,$C4,$3E,$2F,$C1,$58,$F9,$DF,$05
	.dc.b	$FE,$40,$E0,$FE,$19,$C7,$F0,$63,$1F,$11,$9F,$82,$97,$FE,$33,$B7
	.dc.b	$C1,$E3,$0F,$E8,$30,$F8,$3E,$87,$5F,$83,$B7,$C7,$E0,$FC,$1F,$07
	.dc.b	$E1,$6C,$75,$F8,$3F,$21,$E5,$F8,$3B,$1E,$1F,$83,$F4,$7C,$1F,$A5
	.dc.b	$B1,$E5,$F8,$3F,$61,$ED,$F8,$3B,$1E,$9F,$81,$F8,$1F,$05,$F0,$5B
	.dc.b	$1E,$DF,$82,$FF,$34,$F0,$FB,$3E,$0A,$F9,$3F,$E7,$4E,$34,$62,$FF
	.dc.b	$AB,$42,$F1,$6F,$84,$4B,$E0,$F3,$18,$F8,$3B,$26,$CE,$3E,$0F,$3F
	.dc.b	$07,$30,$FA,$51,$CB,$F0,$BF,$E2,$B9,$FF,$67,$7F,$8A,$CF,$F7,$47
	.dc.b	$2F,$8D,$CE,$BF,$E8,$E7,$F8,$C4,$FF,$9B,$DB,$3C,$7E,$0B,$FD,$0B
	.dc.b	$18,$AE,$1F,$05,$8F,$82,$CB,$70,$E5,$F0,$5E,$52,$63,$D7,$E0,$BD
	.dc.b	$1F,$05,$23,$C0,$FA,$FE,$0B,$D9,$F0,$32,$7C,$8F,$BF,$E0,$BE,$07
	.dc.b	$C1,$79,$3E,$0B,$CA,$DC,$FD,$5F,$03,$E8,$F8,$2F,$4B,$63,$DD,$F0
	.dc.b	$3E,$C3,$DB,$F0,$56,$3E,$17,$C1,$7C,$DF,$82,$9A,$E0,$F8,$DF,$03
	.dc.b	$F4,$7E,$0A,$46,$0F,$F4,$C7,$E8,$2C,$3E,$FB,$63,$31,$8F,$87,$CF
	.dc.b	$C2,$6B,$E1,$25,$72,$8E,$3F,$09,$23,$AB,$D3,$E1,$3A,$FC,$1C,$9F
	.dc.b	$0B,$E0,$F1,$3C,$3F,$07,$FE,$BE,$D0,$DF,$83,$C3,$9F,$83,$9B,$7A
	.dc.b	$BE,$0E,$E5,$FE,$3E,$C1,$F0,$7C,$CE,$9F,$07,$6F,$E0,$FC,$17,$C0
	.dc.b	$25,$FC,$4F,$83,$F8,$7F,$05,$F1,$3E,$0E,$4F,$C8,$8F,$8F,$F0,$7F
	.dc.b	$23,$E0,$E4,$FC,$C8,$F9,$7F,$07,$FE,$D8,$FD,$07,$F8,$CF,$99,$FE
	.dc.b	$E8,$DD,$39,$F8,$4D,$7C,$26,$FE,$12,$F0,$7F,$B3,$71,$FE,$36,$DF
	.dc.b	$84,$E8,$75,$F8,$4B,$1B,$F8,$3F,$DC,$D1,$A6,$3E,$13,$3F,$09,$AF
	.dc.b	$84,$9C,$3D,$E3,$8F,$C1,$F2,$F8,$49,$3E,$38,$EB,$F0,$92,$3C,$CF
	.dc.b	$83,$E1,$3C,$3F,$07,$27,$D4,$F9,$3E,$13,$CB,$F0,$72,$7D,$CF,$A3
	.dc.b	$E1,$3D,$3F,$05,$27,$E1,$3E,$CF,$83,$F6,$FC,$14,$9F,$8C,$FC,$0F
	.dc.b	$83,$FE,$33,$EE,$FF,$26,$9F,$F4,$0B,$FE,$B3,$7F,$DF,$0D,$72,$70
	.dc.b	$F8,$68,$E9,$E7,$F0,$9D,$3E,$0B,$AF,$C1,$CA,$E7,$1E,$0F,$83,$91
	.dc.b	$D9,$F1,$FC,$1F,$93,$19,$3E,$27,$CF,$F0,$7E,$8F,$83,$93,$E6,$7D
	.dc.b	$7F,$07,$EC,$F8,$2F,$6F,$C1,$CA,$F5,$C7,$C0,$F8,$2F,$DD,$79,$3F
	.dc.b	$BB,$F8,$56,$69,$CF,$87,$FE,$99,$87,$C2,$63,$1F,$E8,$91,$B9,$F8
	.dc.b	$FF,$09,$9C,$39,$FC,$25,$A4,$D8,$C7,$C2,$67,$E1,$2E,$1F,$41,$E1
	.dc.b	$F0,$9F,$5B,$E1,$26,$7F,$52,$3F,$A8,$FE,$81,$FC,$4E,$3F,$D7,$2D
	.dc.b	$DF,$D8,$DC,$3F,$D6,$4F,$FA,$F2,$47,$0F,$83,$E7,$F0,$92,$33,$7C
	.dc.b	$A7,$BF,$C2,$49,$CE,$1D,$23,$C7,$F0,$9E,$4F,$84,$F2,$FC,$24,$BD
	.dc.b	$9F,$DA,$F3,$FF,$1A,$B8,$E3,$47,$F1,$C8,$C3,$5F,$6F,$C1,$FC,$0F
	.dc.b	$84,$F8,$3F,$09,$F0,$BE,$12,$57,$08,$F8,$BF,$09,$F1,$BE,$12,$4F
	.dc.b	$48,$F9,$3F,$09,$23,$BB,$F3,$3E,$13,$E6,$FC,$1F,$DB,$D0,$DF,$84
	.dc.b	$C3,$9F,$83,$D7,$C2,$5E,$5F,$4B,$C7,$E0,$FF,$09,$FA,$07,$F1,$E9
	.dc.b	$FA,$71,$8F,$84,$CF,$C2,$7F,$83,$2B,$FC,$6B,$B7,$C5,$6B,$FE,$F6
	.dc.b	$51,$BF,$83,$E1,$F0,$72,$B8,$47,$3F,$82,$E9,$F0,$52,$7A,$47,$7F
	.dc.b	$81,$F2,$FC,$17,$99,$BA,$F8,$7E,$07,$D2,$7A,$BE,$0A,$C7,$97,$E0
	.dc.b	$BD,$BF,$05,$EE,$6C,$7A,$7E,$0B,$E0,$FC,$17,$C2,$6C,$7B,$7E,$0B
	.dc.b	$E2,$FC,$17,$C6,$6C,$7C,$1F,$82,$F9,$3F,$05,$F2,$9B,$1F,$17,$E0
	.dc.b	$BE,$41,$38,$7E,$47,$C1,$7F,$8D,$BF,$40,$FE,$3D,$E7,$7E,$FD,$FE
	.dc.b	$38,$DF,$8D,$D5,$6F,$E1,$23,$F9,$5A,$FF,$62,$37,$E1,$3A,$7C,$1C
	.dc.b	$8D,$3D,$FE,$13,$C1,$F0,$92,$79,$3E,$3F,$84,$F2,$7B,$B2,$7B,$3E
	.dc.b	$7F,$84,$F4,$7C,$27,$F0,$BC,$CD,$F8,$4C,$39,$F8,$3D,$7C,$25,$E4
	.dc.b	$F4,$3C,$7E,$13,$97,$C2,$59,$F6,$3D,$7E,$13,$B7,$C2,$59,$F8,$0F
	.dc.b	$87,$E1,$24,$7C,$3F,$84,$BC,$0E,$8F,$99,$F0,$7E,$73,$D1,$F0,$96
	.dc.b	$FF,$7D,$E4,$7F,$13,$BF,$F3,$63,$FC,$4D,$E5,$1F,$C5,$3D,$E4,$E7
	.dc.b	$0F,$E6,$CF,$18,$CF,$C1,$FF,$9C,$A1,$F0,$94,$C4,$67,$E1,$35,$F0
	.dc.b	$93,$2F,$1C,$71,$F8,$4E,$5F,$09,$27,$CF,$1D,$7E,$12,$47,$A9,$F0
	.dc.b	$7C,$27,$87,$E0,$FC,$5F,$09,$2B,$DF,$1E,$5F,$83,$F3,$7C,$24,$9F
	.dc.b	$87,$1E,$9F,$84,$91,$F1,$9F,$67,$C2,$7B,$7E,$0B,$DD,$F0,$72,$BE
	.dc.b	$5C,$7C,$1F,$83,$F8,$5F,$07,$27,$E7,$C7,$C5,$F8,$39,$1F,$49,$FF
	.dc.b	$4C,$FE,$32,$F8,$4B,$F0,$72,$BF,$9A,$4F,$F0,$E6,$FC,$3F,$0F,$83
	.dc.b	$91,$A7,$9F,$C2,$74,$F8,$39,$3C,$9E,$FF,$09,$E0,$F8,$4F,$0F,$7C
	.dc.b	$AE,$F1,$E4,$F8,$4F,$2F,$C2,$49,$F1,$C7,$A3,$E0,$E4,$79,$9F,$5F
	.dc.b	$C2,$7B,$3E,$0E,$4F,$A9,$F7,$FC,$27,$C0,$F8,$3F,$EE,$DC,$3F,$EC
	.dc.b	$2F,$27,$DF,$18,$F8,$4C,$FC,$1E,$BE,$12,$D7,$C4,$78,$FC,$24,$8F
	.dc.b	$8F,$F0,$97,$4F,$CD,$9F,$EC,$9F,$C6,$7F,$D7,$3E,$71,$A7,$FA,$EA
	.dc.b	$BF,$40,$E3,$E1,$3F,$C1,$4F,$FA,$DE,$33,$F0,$98,$9C,$FC,$24,$9F
	.dc.b	$ED,$54,$F8,$4C,$63,$E2,$C7,$C1,$F0,$9F,$ED,$8A,$7C,$25,$31,$19
	.dc.b	$F8,$49,$7D,$1E,$1F,$09,$C7,$E0,$A4,$F8,$1E,$9F,$07,$D7,$E0,$FB
	.dc.b	$7C,$1C,$AF,$2C,$78,$7E,$0E,$47,$A1,$F2,$7C,$1F,$97,$E0,$FC,$DF
	.dc.b	$07,$2B,$DB,$1E,$9F,$83,$F8,$A7,$C6,$F8,$3B,$C9,$F0,$BE,$0F,$E4
	.dc.b	$FC,$1F,$CA,$6C,$7F,$83,$FF,$40,$FF,$5E,$37,$E2,$FF,$C7,$11,$8C
	.dc.b	$BE,$42,$8C,$FC,$1F,$FC,$00,$8F,$E1,$F3,$CF,$E1,$31,$39,$F8,$3D
	.dc.b	$7C,$24,$AF,$04,$71,$F8,$3E,$5F,$09,$27,$C9,$1D,$7E,$12,$47,$9D
	.dc.b	$F0,$7C,$27,$87,$E0,$FC,$5F,$09,$2B,$D9,$1E,$5F,$83,$F3,$7C,$24
	.dc.b	$9F,$81,$1E,$9F,$84,$91,$F0,$DF,$67,$C2,$7B,$7E,$0E,$4F,$C7,$7E
	.dc.b	$07,$C2,$7C,$1F,$83,$F8,$5F,$09,$2B,$E6,$47,$C5,$F8,$49,$1F,$3D
	.dc.b	$FF,$27,$BF,$93,$DF,$F8,$83,$FE,$41,$7F,$87,$39,$F8,$3D,$FC,$2F
	.dc.b	$F9,$D6,$72,$DF,$84,$CF,$5F,$F4,$79,$7C,$26,$30,$79,$DF,$EE,$6F
	.dc.b	$84,$C3,$63,$C1,$38,$AF,$F6,$14,$78,$FE,$13,$06,$7E,$13,$5F,$07
	.dc.b	$BF,$84,$95,$E9,$8E,$5F,$07,$23,$D8,$F5,$F8,$4E,$DF,$09,$DF,$E1
	.dc.b	$25,$7C,$18,$F1,$7C,$27,$8F,$E1,$24,$FC,$58,$F3,$7C,$1F,$9F,$E0
	.dc.b	$E4,$FC,$98,$F5,$7C,$1F,$FA,$47,$81,$F0,$76,$30,$E7,$E0,$F5,$F0
	.dc.b	$77,$9F,$E9,$3F,$E9,$9F,$CF,$FF,$8B,$93,$E2,$31,$9E,$7E,$12,$5F
	.dc.b	$F9,$E6,$9F,$0B,$8A,$FE,$21,$0F,$85,$E8,$75,$F8,$5B,$7C,$7E,$0F
	.dc.b	$C0,$78,$7E,$12,$C7,$5F,$84,$F2,$7C,$27,$95,$B1,$E1,$F8,$4F,$41
	.dc.b	$E9,$F8,$4B,$1E,$5F,$83,$F6,$7C,$27,$B5,$B1,$E9,$F8,$4F,$C2,$ED
	.dc.b	$FE,$BE,$DF,$D9,$F0,$98,$9C,$FC,$1E,$BE,$12,$57,$C1,$8E,$3F,$09
	.dc.b	$23,$E2,$3D,$3E,$13,$AF,$C1,$C9,$F9,$0F,$83,$E1,$3F,$2B,$F9,$FF
	.dc.b	$F8,$74,$9A,$E3,$E1,$73,$F0,$7A,$F8,$49,$8F,$FA,$A6,$D3,$CB,$E0
	.dc.b	$F9,$FC,$24,$AD,$47,$6F,$83,$91,$C5,$F0,$FC,$27,$8B,$C6,$93,$D5
	.dc.b	$F2,$FC,$27,$9B,$E0,$FC,$FF,$09,$2B,$C5,$1E,$AF,$83,$F5,$FC,$24
	.dc.b	$9F,$34,$7B,$BE,$12,$47,$A5,$F8,$3F,$09,$F0,$BE,$12,$4F,$B5,$F8
	.dc.b	$BF,$09,$F1,$BE,$0F,$E3,$FC,$24,$AF,$85,$1F,$2B,$E1,$24,$7C,$57
	.dc.b	$F8,$0F,$E5,$FF,$C4,$23,$FB,$23,$FE,$AA,$E9,$FF,$69,$2F,$83,$FF
	.dc.b	$7D,$39,$9D,$C6,$3E,$07,$3F,$07,$6F,$8B,$C3,$E0,$B8,$FC,$15,$9E
	.dc.b	$AF,$4F,$81,$EB,$F0,$56,$7C,$2F,$83,$E0,$A4,$79,$3E,$0A,$E9,$E8
	.dc.b	$F8,$1F,$29,$E6,$F8,$2B,$5E,$AF,$81,$F7,$FC,$14,$BC,$1E,$EF,$81
	.dc.b	$F8,$7F,$05,$23,$07,$C2,$F8,$2F,$8E,$47,$E2,$FC,$17,$CA,$F8,$2F
	.dc.b	$8B,$F0,$5F,$19,$B9,$FF,$BE,$FD,$BF,$F0,$F9,$3E,$09,$19,$C2,$9F
	.dc.b	$E1,$59,$FF,$0F,$3F,$CE,$1F,$E1,$71,$CF,$E0,$B7,$F0,$9C,$3E,$11
	.dc.b	$5A,$BE,$5F,$07,$AC,$3F,$D1,$53,$FE,$23,$7C,$91,$99,$ED,$F0,$76
	.dc.b	$F9,$FC,$1D,$C3,$FC,$4B,$87,$AE,$31,$F0,$79,$F8,$39,$3E,$FA,$B8
	.dc.b	$FA,$FE,$0F,$FA,$7C,$3E,$0E,$98,$8C,$FC,$16,$BE,$0E,$5F,$C9,$8E
	.dc.b	$3F,$03,$23,$E6,$3D,$3E,$0B,$AF,$C0,$F6,$F8,$29,$5F,$46,$3C,$3F
	.dc.b	$05,$E2,$F8,$29,$3F,$56,$3C,$BF,$05,$E6,$F8,$29,$3F,$66,$3F,$B3
	.dc.b	$FE,$7D,$FE,$23,$E8,$FA,$0F,$F8,$8F,$89,$5F,$E4,$D8,$D4,$7F,$AE
	.dc.b	$5D,$46,$7E,$03,$E7,$47,$F1,$69,$E3,$5D,$3E,$0B,$35,$AF,$80,$95
	.dc.b	$A3,$A7,$C0,$E3,$1F,$07,$C0,$F9,$88,$EC,$FF,$00,$E4,$F8,$1F,$5F
	.dc.b	$C0,$C9,$D7,$6A,$93,$F0,$3E,$07,$E0,$FC,$04,$BD,$D2,$3E,$2F,$C0
	.dc.b	$FC,$6F,$80,$F6,$FC,$0F,$B9,$BB,$70,$F8,$1F,$83,$F0,$3F,$09,$B1
	.dc.b	$D3,$E0,$7E,$2F,$C0,$FC,$66,$C7,$83,$E0,$3E,$49,$F2,$BE,$06,$C7
	.dc.b	$93,$E0,$7F,$35,$F8,$BF,$FA,$EE,$30,$7B,$9C,$70,$7F,$A0,$DF,$E3
	.dc.b	$9D,$5F,$AB,$6C,$FF,$BC,$5F,$E4,$4F,$1F,$87,$EB,$F0,$3F,$E2,$A9
	.dc.b	$FF,$17,$DF,$58,$C7,$C0,$E7,$E0,$75,$F0,$32,$BC,$51,$C7,$E0,$79
	.dc.b	$7C,$0C,$9F,$34,$75,$F8,$1F,$5F,$C0,$FB,$1B,$AF,$AF,$E0,$7D,$E7
	.dc.b	$C0,$F8,$1B,$1E,$FF,$81,$F8,$7F,$03,$F1,$1B,$1F,$0F,$E0,$7E,$39
	.dc.b	$F2,$3E,$06,$C7,$C7,$F8,$1F,$96,$7C,$CF,$81,$B1,$F2,$FE,$03,$E7
	.dc.b	$FC,$0F,$D0,$6C,$7E,$33,$F3,$EF,$F2,$FF,$0F,$FA,$C5,$BD,$3F,$C6
	.dc.b	$5B,$3F,$AA,$F2,$10,$F8,$2F,$F2,$BA,$F8,$2B,$F1,$13,$1E,$D1,$AF
	.dc.b	$83,$91,$E1,$78,$FC,$1F,$2F,$83,$E7,$F0,$72,$BC,$D1,$DB,$E0,$E4
	.dc.b	$7A,$5F,$0F,$C1,$F8,$BE,$0B,$C7,$F0,$72,$BD,$D1,$E6,$F8,$2F,$E5
	.dc.b	$5B,$1B,$F0,$78,$73,$F0,$73,$8F,$E3,$7C,$1D,$C7,$E3,$7C,$17,$23
	.dc.b	$9F,$C1,$DA,$F9,$9F,$05,$D8,$EF,$F0,$76,$3E,$87,$C1,$78,$8F,$1F
	.dc.b	$C1,$D8,$FD,$FF,$8D,$FC,$5B,$2F,$F4,$2B,$FC,$5B,$B3,$FA,$63,$9E
	.dc.b	$C3,$9A,$FA,$CF,$F1,$9A,$FF,$76,$2F,$83,$FE,$94,$E6,$F9,$C6,$3E
	.dc.b	$0B,$3F,$05,$70,$F1,$BC,$3E,$0A,$47,$9B,$E0,$AE,$5D,$BE,$0B,$FD
	.dc.b	$33,$9F,$F1,$68,$F5,$C6,$3E,$0B,$3F,$03,$27,$DE,$F0,$F8,$1E,$3F
	.dc.b	$01,$CB,$E0,$65,$7C,$48,$EB,$F0,$1E,$A3,$D7,$F0,$37,$5F,$93,$F0
	.dc.b	$1E,$EF,$81,$F7,$B6,$3E,$6F,$C0,$FC,$2F,$81,$F8,$6D,$8F,$A3,$F0
	.dc.b	$3F,$DB,$FF,$3C,$FE,$27,$D1,$F4,$23,$1F,$1B,$9F,$83,$97,$FC,$DA
	.dc.b	$FF,$89,$EE,$C3,$FC,$05,$1F,$C6,$23,$FD,$7E,$FF,$B0,$A7,$87,$C1
	.dc.b	$77,$F8,$2F,$03,$E1,$F8,$1F,$17,$C1,$5A,$EC,$F9,$7E,$07,$CD,$F0
	.dc.b	$56,$7C,$4F,$A7,$AE,$47,$97,$E0,$AE,$9E,$58,$FD,$E6,$8F,$82,$FF
	.dc.b	$68,$B6,$7D,$91,$8F,$82,$CF,$C0,$EB,$E0,$A5,$7C,$18,$E3,$F0,$52
	.dc.b	$3E,$23,$D3,$E0,$BA,$FC,$14,$9F,$90,$F8,$3E,$0B,$C3,$F0,$32,$7E
	.dc.b	$63,$E4,$F8,$2F,$F7,$2F,$E7,$7F,$C5,$27,$FD,$D2,$DF,$34,$D3,$F8
	.dc.b	$B4,$7F,$6A,$7F,$C7,$31,$AF,$81,$FF,$16,$47,$F8,$D7,$AB,$FC,$70
	.dc.b	$E8,$FF,$62,$71,$71,$F8,$0D,$FC,$0F,$07,$8F,$C0,$49,$ED,$F0,$37
	.dc.b	$3F,$9C,$79,$BE,$07,$D3,$F0,$32,$3C,$91,$EC,$F8,$0F,$6F,$C0,$C9
	.dc.b	$F4,$47,$C0,$F7,$E4,$7A,$DF,$87,$F0,$1F,$13,$E0,$3E,$2F,$C0,$4A
	.dc.b	$F8,$11,$F2,$3E,$03,$E4,$FC,$04,$9F,$89,$1F,$33,$E0,$3E,$6F,$C0
	.dc.b	$49,$F9,$11,$F4,$3E,$03,$E3,$9F,$23,$E0,$2F,$17,$F8,$5B,$F3,$BF
	.dc.b	$E2,$84,$F8,$BF,$F2,$C3,$8F,$88,$BF,$07,$86,$D7,$1F,$83,$DF,$C1
	.dc.b	$F0,$6C,$75,$F8,$3E,$7F,$07,$D1,$B1,$E1,$F8,$3E,$E7,$83,$E0,$EC
	.dc.b	$79,$7E,$0F,$FC,$87,$C3,$F5,$3C,$46,$CF,$83,$CF,$89,$FE,$79,$8C
	.dc.b	$57,$13,$DD,$F0,$5A,$C7,$FB,$24,$BE,$0F,$14,$C7,$AC,$61,$EF,$76
	.dc.b	$9F,$07,$7E,$0E,$61,$F2,$E3,$5F,$05,$F2,$BE,$0F,$E5,$B7,$3F,$A1
	.dc.b	$F0,$5F,$38,$F9,$FF,$07,$63,$EA,$7C,$1F,$D2,$3F,$D1,$5C,$5F,$C7
	.dc.b	$23,$D3,$18,$F8,$2F,$34,$FF,$B3,$73,$FF,$28,$B5,$F9,$0A,$37,$F0
	.dc.b	$32,$76,$F2,$F8,$2E,$7F,$05,$D3,$E0,$A5,$74,$8E,$FF,$05,$E4,$3C
	.dc.b	$BF,$05,$74,$F0,$FC,$17,$A0,$F4,$FC,$15,$8F,$2F,$C0,$FB,$0F,$6F
	.dc.b	$C0,$D8,$F4,$FC,$07,$C0,$F8,$1F,$82,$D8,$F6,$FC,$0F,$C4,$F8,$1F
	.dc.b	$8A,$D8,$F8,$3F,$03,$F2,$0F,$93,$F0,$36,$3E,$2F,$C0,$FC,$8F,$81
	.dc.b	$9C,$58,$3F,$CF,$3F,$C0,$3F,$C5,$4F,$E2,$6F,$E4,$5B,$E2,$7F,$B1
	.dc.b	$DD,$7C,$17,$E1,$DF,$F1,$64,$F2,$8C,$7C,$36,$7E,$0B,$FD,$89,$5F
	.dc.b	$EC,$9A,$E9,$18,$F8,$2C,$FC,$14,$D3,$C7,$F0,$56,$FC,$5F,$03,$CB
	.dc.b	$E0,$B9,$B6,$BD,$3F,$03,$E3,$F8,$29,$78,$3D,$BF,$05,$E7,$23,$F0
	.dc.b	$3E,$0B,$D5,$F0,$32,$3E,$1B,$ED,$F8,$2F,$77,$C0,$FB,$FE,$0A,$57
	.dc.b	$C8,$8F,$85,$F0,$32,$3E,$5B,$F1,$7E,$0B,$E3,$7C,$17,$C7,$F8,$29
	.dc.b	$5F,$42,$3F,$61,$FC,$03,$FE,$FA,$8F,$E0,$0B,$E0,$F2,$6B,$87,$7F
	.dc.b	$05,$2B,$FD,$63,$48,$E7,$F0,$7D,$3E,$0B,$AF,$C1,$CB,$DC,$7F,$02
	.dc.b	$5F,$4B,$FE,$36,$FC,$6B,$FD,$A7,$87,$FA,$D6,$7F,$DA,$B3,$97,$B7
	.dc.b	$C1,$7F,$16,$8C,$4E,$E7,$9F,$9C,$BE,$77,$B8,$C4,$F2,$3A,$B8,$AE
	.dc.b	$57,$D2,$B8,$5F,$2B,$DB,$5E,$ED,$9E,$11,$A3,$B1,$A3,$FD,$16,$53
	.dc.b	$FD,$1A,$FB,$5B,$7E,$66,$CF,$07,$81,$E4,$39,$CF,$17,$07,$03,$33
	.dc.b	$CD,$C8,$FC,$6E,$00

;--------------------------------------------------------------------------------
;	FSINCOS.X FPm,FPc:FPs
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsincos_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FSINCOS.X FPM,FPC:FPS'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSINCOS.X FPm,FPc:FPs',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d2	0=failed,1=successful upper
;	d3	0=failed,1=successful lower
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result upper handle,expected status upper,expected result lower handle,expected status lower,...
;	fp2	source
;	fp4	expected result upper
;	fp5	expected result lower
;	fp6	actual result upper
;	fp7	actual result lower
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsincos_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result upper handle
	addq.l	#4,a0			;expected status upper
	add.l	d0,(a0)+		;expected result lower handle
	addq.l	#4,a0			;expected status lower
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1953+4,a2),a4	;expected result upper handle,expected status upper,expected result lower handle,expected status lower,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn:FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsincos.x	fp7,fp7:fp7	;EXECUTE
					;actual result upper
					;actual result lower
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([8,a4]),fp5		;expected result lower
	move.l	(12,a4),d5		;expected status lower
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result lower
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result lower
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful lower
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSINCOS.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn:FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful lower
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result lower
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result lower
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	lea.l	16(a4),a4		;expected result upper handle,expected status upper,expected result lower handle,expected status lower,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1953+4,a2),a4	;expected result upper handle,expected status upper,expected result lower handle,expected status lower,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPc:FPs
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp6		;fill upper=NaN
	fmove.s	#$7FFFFFFF,fp7		;fill lower=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsincos.x	fp2,fp6:fp7	;EXECUTE
					;actual result upper
					;actual result lower
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp4		;expected result upper
	move.l	(4,a4),d4		;expected status upper
	fmove.x	([8,a4]),fp5		;expected result lower
	move.l	(12,a4),d5		;expected status lower
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	clr.l	-(sp)			;dummy status
	fmove.x	fp4,-(sp)		;expected result upper
	clr.l	-(sp)			;dummy status
	fmove.x	fp6,-(sp)		;actual result upper
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d2			;0=failed,1=successful upper
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result lower
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result lower
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful lower
;
	move.l	d2,d0			;0=failed,1=successful upper
	and.l	d3,d0			;0=failed,1=successful lower
					;0=failed,1=successful
	move.l	d0,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSINCOS.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPc:FPs @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d2,-(sp)		;0=failed,1=successful upper
	clr.l	-(sp)			;dummy status
	fmove.x	fp4,-(sp)		;expected result upper
	clr.l	-(sp)			;dummy status
	fmove.x	fp6,-(sp)		;actual result upper
	jbsr	output_extended
	lea.l	36(sp),sp
	move.l	d3,-(sp)		;0=failed,1=successful lower
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result lower
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result lower
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	lea.l	16(a4),a4		;expected result upper handle,expected status upper,expected result lower handle,expected status lower,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsincos_data_compressed::
	.dc.b	$00,$05,$D7,$4C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$F3,$FE,$4D,$CD,$F8,$3E,$27,$13,$D5,$F5,$7A,$BB
	.dc.b	$AF,$C3,$52,$F8,$FA,$59,$F4,$B5,$E9,$6F,$D2,$E1,$E9,$71,$F4,$B9
	.dc.b	$7A,$5C,$FD,$2E,$9E,$97,$5F,$4B,$B7,$A5,$DF,$D2,$F0,$7A,$5E,$1F
	.dc.b	$4B,$C5,$E9,$7A,$7F,$58,$FF,$77,$F1,$1F,$F5,$4B,$7E,$37,$5F,$15
	.dc.b	$BF,$80,$E1,$EE,$71,$F5,$F9,$7A,$FC,$FD,$6E,$9E,$B7,$5F,$5B,$B7
	.dc.b	$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$78,$FC,$2F,$27,$AD,$E5
	.dc.b	$F5,$BC,$DE,$B7,$E0,$FD,$AF,$CA,$7C,$7F,$E5,$7C,$7F,$F6,$B7,$C4
	.dc.b	$FA,$FD,$9C,$CE,$BE,$27,$7F,$0B,$C3,$E0,$B8,$FB,$DC,$BC,$DE,$7E
	.dc.b	$6F,$4F,$37,$AF,$9B,$DB,$CD,$EF,$E6,$F8,$3C,$DF,$0F,$9B,$E2,$F3
	.dc.b	$7C,$7E,$6F,$93,$CD,$F2,$F9,$BE,$6F,$37,$CF,$E6,$FA,$3C,$DF,$D6
	.dc.b	$7C,$AF,$FB,$60,$FF,$48,$9F,$E4,$B3,$FC,$06,$7E,$D4,$FD,$A7,$49
	.dc.b	$79,$1C,$BC,$8E,$7E,$47,$4F,$23,$AF,$91,$DB,$C8,$EF,$E4,$78,$3C
	.dc.b	$8F,$0F,$91,$E2,$F2,$3C,$7E,$47,$93,$C8,$F2,$F9,$1F,$9D,$FD,$67
	.dc.b	$FD,$AB,$C0,$F2,$7D,$3E,$3F,$F0,$5F,$97,$7C,$AE,$FE,$27,$87,$C7
	.dc.b	$71,$F8,$CE,$5F,$0F,$CF,$E1,$7A,$7C,$2F,$5F,$85,$ED,$F0,$BD,$FE
	.dc.b	$17,$C1,$F0,$BE,$1F,$85,$F1,$7C,$2F,$8F,$E1,$7C,$9F,$0B,$E5,$F8
	.dc.b	$5F,$37,$C2,$F9,$FE,$17,$D1,$F0,$9E,$9F,$84,$F5,$7B,$DF,$B7,$FD
	.dc.b	$6B,$F8,$E7,$C2,$17,$C0,$67,$E0,$35,$F0,$1B,$F8,$0E,$1F,$01,$C7
	.dc.b	$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7,$C0,$77,$F8,$0F
	.dc.b	$07,$C0,$78,$7E,$03,$C5,$F0,$1F,$27,$C6,$F9,$5E,$37,$97,$E0,$3E
	.dc.b	$67,$8F,$F3,$7C,$7F,$9D,$E3,$FC,$FF,$1F,$E8,$78,$FF,$47,$C7,$FA
	.dc.b	$5E,$3F,$D3,$F1,$FE,$A7,$8F,$F5,$7C,$7F,$AD,$E3,$FD,$7F,$1F,$EC
	.dc.b	$78,$FF,$67,$C7,$FB,$5E,$3F,$DB,$F1,$FE,$E7,$8F,$F7,$7C,$7F,$E8
	.dc.b	$5F,$33,$FA,$F3,$FE,$00,$1F,$EB,$83,$FC,$56,$7F,$65,$3F,$91,$1F
	.dc.b	$C7,$BD,$C2,$F0,$79,$F8,$3D,$3C,$1E,$BE,$0F,$6F,$07,$BF,$83,$E0
	.dc.b	$F0,$7C,$3E,$0F,$8B,$C1,$F1,$F7,$F9,$3E,$07,$CB,$F0,$3E,$6F,$81
	.dc.b	$F3,$FC,$0F,$A3,$E0,$7F,$01,$5F,$D6,$FC,$57,$CA,$5F,$8B,$FC,$26
	.dc.b	$3F,$85,$C7,$F0,$D8,$FE,$1F,$1F,$E7,$D8,$FF,$26,$3F,$92,$8F,$87
	.dc.b	$CF,$C3,$EB,$E1,$F7,$F0,$FC,$3E,$1F,$8F,$C3,$F2,$F8,$7E,$7F,$0F
	.dc.b	$D3,$E1,$FA,$FC,$3F,$6F,$87,$EF,$F0,$FE,$0F,$87,$F0,$FC,$3F,$8B
	.dc.b	$E1,$FC,$7F,$0F,$E4,$F8,$7F,$2F,$C2,$F9,$BE,$17,$FA,$A7,$EB,$7F
	.dc.b	$B4,$F2,$7F,$A7,$2F,$73,$3D,$FA,$F7,$37,$EE,$70,$F7,$38,$FB,$9C
	.dc.b	$BD,$CE,$7E,$E7,$4F,$73,$AF,$B9,$F2,$FC,$0F,$99,$E0,$78,$3D,$CF
	.dc.b	$9D,$E0,$FC,$FF,$07,$E8,$78,$3F,$47,$C1,$FA,$5E,$0F,$D3,$F0,$7E
	.dc.b	$A7,$83,$F5,$7C,$1F,$AD,$E0,$FD,$7F,$07,$EC,$78,$3F,$67,$C1,$FB
	.dc.b	$5E,$0F,$DB,$F0,$7E,$E7,$83,$F7,$7C,$1F,$BD,$E0,$FF,$86,$7E,$6F
	.dc.b	$E1,$78,$47,$FD,$C2,$7F,$A6,$CF,$F2,$99,$FE,$0A,$75,$DB,$C7,$B7
	.dc.b	$97,$6F,$3E,$DE,$9D,$BD,$7B,$7B,$75,$F7,$F7,$7C,$1E,$EF,$87,$DD
	.dc.b	$F1,$7B,$BE,$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$E7,$F7,$7E,$CE
	.dc.b	$5F,$E3,$1F,$9C,$FE,$BC,$7F,$16,$E3,$FF,$1F,$2F,$83,$D7,$C1,$FE
	.dc.b	$03,$6F,$C0,$ED,$F8,$2D,$BF,$07,$B7,$E1,$36,$FC,$2E,$DF,$86,$DB
	.dc.b	$F0,$FB,$7F,$86,$FF,$5C,$FE,$4A,$77,$F8,$73,$DF,$7F,$F0,$C3,$FE
	.dc.b	$F0,$8C,$FC,$2F,$1F,$85,$E5,$F0,$BC,$FE,$13,$A7,$C2,$75,$F8,$4F
	.dc.b	$F2,$67,$EE,$6F,$84,$F0,$7C,$2F,$85,$F1,$7C,$2F,$8F,$E1,$7C,$9F
	.dc.b	$0B,$E5,$F8,$5F,$37,$C2,$F9,$FD,$FF,$47,$C1,$FA,$7E,$0F,$D5,$F0
	.dc.b	$7E,$BF,$83,$FF,$1B,$7E,$BD,$FE,$1E,$3F,$C3,$A7,$EE,$7A,$3F,$C6
	.dc.b	$4B,$DE,$DF,$BD,$F4,$7A,$3E,$97,$47,$D3,$E8,$E7,$EF,$7D,$5E,$9F
	.dc.b	$AD,$D3,$F5,$FA,$7E,$C7,$47,$D9,$E8,$FB,$5D,$1F,$6F,$A3,$EE,$74
	.dc.b	$7D,$DE,$8F,$BD,$D1,$F7,$FA,$3F,$01,$D1,$F8,$1E,$8F,$C1,$74,$7E
	.dc.b	$0F,$A3,$FC,$E9,$F4,$1F,$C6,$BA,$7F,$D0,$0B,$E1,$B3,$F0,$3A,$EA
	.dc.b	$DF,$57,$0E,$7E,$3F,$03,$CB,$E0,$79,$FC,$0F,$4F,$81,$EB,$F0,$3D
	.dc.b	$BE,$07,$BF,$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93
	.dc.b	$E0,$7C,$BF,$03,$E6,$F8,$1F,$3F,$C0,$FF,$A2,$F9,$3F,$66,$7F,$13
	.dc.b	$E7,$FF,$4C,$7E,$37,$FE,$C4,$87,$C2,$EF,$E1,$78,$7C,$2F,$1F,$84
	.dc.b	$E5,$F0,$9C,$FE,$13,$A7,$C2,$75,$F8,$4E,$DF,$09,$DF,$E1,$3C,$1F
	.dc.b	$09,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$FA,$87,$93,$F8,$27,$23,$E0,$7D,$5F,$03,$FE,$A5,$FA,$2F,$F2
	.dc.b	$91,$FB,$5E,$83,$FA,$76,$FC,$1E,$FD,$FE,$1E,$FF,$1F,$7F,$97,$BF
	.dc.b	$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BF,$E1,$F7,$FC
	.dc.b	$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF,$F9,$FD,$FF,$47,$BF
	.dc.b	$FE,$BD,$FA,$3F,$C5,$73,$1F,$C5,$3F,$EC,$B7,$F4,$CE,$7D,$EE,$1E
	.dc.b	$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$7C,$9F,$3F,$B7,$BD,$DF,$DE
	.dc.b	$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$E5,$F7,$BC,$DE
	.dc.b	$F7,$9F,$DE,$F4,$7B,$DE,$9F,$7B,$FD,$AD,$F4,$9F,$CF,$BC,$F7,$B9
	.dc.b	$8F,$89,$CF,$C0,$6B,$E0,$37,$F0,$1C,$39,$78,$DF,$2F,$80,$E7,$F0
	.dc.b	$1D,$3E,$03,$AF,$C0,$76,$F8,$0E,$FF,$01,$E0,$F8,$0F,$0F,$C0,$78
	.dc.b	$BE,$03,$C7,$F0,$1E,$4F,$80,$F2,$FC,$07,$E2,$F5,$FC,$66,$BF,$8D
	.dc.b	$D7,$F1,$DA,$FE,$3F,$5F,$5F,$BF,$EC,$F7,$FF,$DE,$5E,$F7,$FA,$7C
	.dc.b	$FC,$71,$F8,$D0,$6B,$DE,$DF,$BD,$C3,$DE,$E3,$EF,$72,$F7,$B9,$FB
	.dc.b	$DF,$9D,$E6,$EB,$EF,$76,$F7,$BB,$FB,$DF,$A1,$E9,$FD,$17,$4F,$E8
	.dc.b	$FA,$7F,$49,$D3,$FA,$5E,$9F,$D3,$74,$FE,$9F,$93,$F5,$1C,$8F,$52
	.dc.b	$7E,$39,$FE,$B7,$3F,$4C,$7E,$94,$F4,$BA,$E5,$E0,$A3,$FD,$D8,$7F
	.dc.b	$92,$CF,$F2,$53,$D3,$8B,$70,$E2,$BC,$38,$E1,$F9,$3E,$FF,$CA,$F7
	.dc.b	$FE,$5F,$BF,$F3,$3D,$FF,$9B,$EF,$FC,$EF,$7F,$E7,$FB,$FF,$43,$DF
	.dc.b	$FA,$3E,$FF,$93,$E9,$BF,$4E,$7E,$9B,$84,$FE,$98,$FE,$94,$FC,$15
	.dc.b	$EF,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE
	.dc.b	$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$F0,$BC,$8F,$87
	.dc.b	$D5,$F1,$3A,$BE,$2F,$57,$C6,$EA,$F8,$FD,$5F,$23,$AB,$E4,$F5,$7C
	.dc.b	$AE,$AF,$97,$D5,$F4,$BD,$0F,$E0,$A7,$BC,$FE,$06,$7B,$BA,$CF,$B9
	.dc.b	$FC,$03,$FE,$D8,$3F,$DA,$EE,$39,$FA,$73,$F5,$E7,$F9,$9E,$F7,$CD
	.dc.b	$F7,$BE,$77,$BD,$E1,$EA,$FA,$1E,$F7,$E1,$3E,$9F,$FC,$76,$7F,$49
	.dc.b	$CC,$FF,$48,$7F,$20,$BD,$BE,$1E,$DF,$1F,$6F,$97,$B7,$CF,$DB,$E9
	.dc.b	$ED,$F5,$E2,$ED,$ED,$FC,$1F,$0F,$E1,$78,$7F,$0F,$C3,$F8,$9E,$1F
	.dc.b	$C5,$F0,$FC,$9E,$DF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$D1,$EB,$FA,$7D
	.dc.b	$7F,$57,$AF,$EB,$F5,$FD,$9E,$BF,$B7,$D7,$FD,$0F,$D4,$7E,$8B,$AC
	.dc.b	$FF,$4D,$3F,$A6,$1F,$A8,$73,$EC,$70,$F6,$F8,$FB,$7C,$BD,$BF,$B5
	.dc.b	$E9,$74,$F6,$FA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$D6,$F0,$FA,$DE,$2F
	.dc.b	$5B,$C7,$EB,$79,$3D,$6F,$2F,$AD,$E6,$F5,$BC,$FE,$B7,$A3,$D6,$FC
	.dc.b	$47,$A5,$EA,$F5,$BF,$77,$E7,$7F,$4E,$3F,$50,$7F,$4C,$3F,$14,$7E
	.dc.b	$27,$9F,$EC,$7B,$1C,$3D,$8E,$3E,$C7,$2F,$63,$9F,$B1,$D3,$D6,$EB
	.dc.b	$E6,$F6,$F3,$7B,$F9,$BE,$0F,$37,$C3,$E6,$F8,$BB,$BC,$7C,$3E,$4F
	.dc.b	$63,$CB,$EC,$79,$BD,$8F,$E2,$3F,$53,$FE,$88,$3F,$1C,$7E,$37,$8C
	.dc.b	$FE,$35,$FE,$3C,$FF,$A2,$47,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7F,$D7,$69,$E0,$F5,$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F
	.dc.b	$5F,$CB,$EB,$F9,$BD,$7F,$3F,$AF,$E8,$F5,$FD,$3E,$BF,$F2,$AF,$AA
	.dc.b	$FC,$51,$FE,$15,$3F,$C2,$87,$E2,$47,$AF,$BF,$5F,$87,$AF,$FC,$07
	.dc.b	$AB,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F,$07,$AF
	.dc.b	$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB,$F9,$FD
	.dc.b	$7F,$47,$AF,$FD,$0B,$EA,$FF,$C2,$A7,$F2,$F3,$FC,$26,$7F,$11,$3F
	.dc.b	$10,$3D,$7E,$1E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB
	.dc.b	$7A,$FD,$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$79,$DE
	.dc.b	$5F,$03,$CD,$EB,$FF,$58,$F8,$4F,$F6,$79,$FE,$5D,$F8,$C7,$C5,$FF
	.dc.b	$22,$F6,$BF,$D1,$DF,$07,$BF,$73,$87,$BD,$FE,$96,$E6,$FF,$4B,$F3
	.dc.b	$7F,$A6,$39,$BF,$D3,$3C,$DF,$E9,$AE,$6F,$F4,$DF,$37,$FA,$73,$9B
	.dc.b	$FD,$3B,$CD,$FE,$9E,$E6,$FF,$4F,$F3,$7F,$A8,$39,$BF,$D4,$3C,$DF
	.dc.b	$EA,$2E,$6F,$F5,$1F,$37,$F7,$5E,$57,$A0,$F9,$6C,$3E,$93,$F4,$1F
	.dc.b	$97,$3C,$CF,$EE,$AF,$E2,$3A,$17,$CB,$9F,$8F,$3F,$2F,$7B,$9F,$BD
	.dc.b	$D3,$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$7B,$C3,$EF,$78,$BD,$EF
	.dc.b	$1F,$BD,$E4,$F7,$BC,$BE,$F7,$9B,$93,$CF,$EF,$7A,$3D,$EF,$A5,$C1
	.dc.b	$F4,$F8,$3E,$A7,$07,$D5,$E0,$F6,$FB,$DF,$5F,$87,$EC,$70,$FD,$9E
	.dc.b	$1F,$B5,$C3,$F6,$F8,$7E,$E7,$0F,$DD,$E1,$FB,$DC,$3F,$7F,$87,$F0
	.dc.b	$1C,$3F,$81,$E1,$FC,$16,$BF,$E1,$AF,$AD,$FE,$74,$7F,$A1,$4F,$F4
	.dc.b	$21,$FC,$E4,$F9,$5C,$C5,$EE,$F1,$E9,$E5,$EE,$F3,$F7,$7A,$7B,$BD
	.dc.b	$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7,$7C
	.dc.b	$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE,$FF,$8C,$BE,$BB,$FD
	.dc.b	$0A,$7F,$87,$CF,$F4,$19,$FE,$1E,$F2,$1F,$AF,$FF,$57,$8F,$E7,$47
	.dc.b	$EC,$F9,$FF,$0B,$0F,$99,$FC,$82,$F9,$4B,$F1,$58,$F8,$3C,$F5,$6B
	.dc.b	$AB,$7D,$5C,$3A,$B8,$F5,$72,$EA,$E7,$D5,$D3,$AB,$AF,$57,$6E,$AE
	.dc.b	$FD,$5E,$0E,$AF,$0F,$57,$F9,$3B,$F1,$FF,$F5,$E7,$2F,$F9,$5D,$77
	.dc.b	$F9,$BB,$3C,$FD,$9E,$8E,$CF,$4F,$67,$EA,$7F,$60,$FF,$79,$76,$9F
	.dc.b	$F5,$4B,$7E,$4B,$5F,$07,$BF,$8B,$E1,$F0,$DC,$7E,$1B,$97,$C2,$73
	.dc.b	$F8,$4E,$9F,$09,$D7,$E1,$3B,$7C,$27,$7F,$84,$FA,$98,$7D,$5C,$3C
	.dc.b	$5F,$09,$F5,$F1,$FB,$18,$FD,$9C,$7E,$D6,$3F,$6F,$1F,$B9,$7F,$77
	.dc.b	$E1,$3E,$F7,$C2,$7D,$FF,$84,$FC,$07,$C2,$7E,$07,$E1,$3F,$05,$F0
	.dc.b	$9F,$83,$F8,$4F,$F3,$AF,$D9,$7F,$07,$EE,$FF,$3F,$3F,$E3,$33,$FB
	.dc.b	$31,$FD,$95,$2F,$82,$E3,$F0,$5C,$BE,$0B,$9F,$C1,$74,$F8,$2E,$BF
	.dc.b	$05,$DB,$E0,$BB,$FC,$17,$83,$E0,$BC,$3F,$05,$E2,$F8,$2F,$1F,$C1
	.dc.b	$79,$3E,$0B,$CB,$F0,$5E,$6F,$2F,$CF,$F0,$5E,$8F,$82,$FF,$46,$FB
	.dc.b	$9F,$E2,$D3,$FD,$F0,$7F,$BD,$CF,$F1,$61,$FB,$43,$F6,$7C,$BF,$C7
	.dc.b	$BC,$A7,$6F,$2F,$6F,$9F,$B7,$F8,$3E,$FF,$C2,$77,$FE,$17,$BF,$F0
	.dc.b	$DD,$FF,$87,$EF,$FC,$47,$7F,$F3,$1F,$D8,$7E,$57,$9A,$7F,$A4,$1F
	.dc.b	$E5,$B7,$1F,$09,$BF,$07,$87,$83,$C7,$C1,$E5,$E0,$F3,$F0,$7A,$78
	.dc.b	$3D,$7C,$1E,$DE,$0F,$7F,$07,$C1,$E0,$F8,$7C,$1F,$17,$83,$E3,$F0
	.dc.b	$7C,$9E,$0F,$97,$C1,$F3,$78,$3E,$7F,$07,$FA,$47,$4F,$FA,$DB,$94
	.dc.b	$FE,$F7,$C0,$FE,$86,$FE,$F5,$BE,$FF,$0F,$0B,$8F,$BF,$CB,$C1,$E7
	.dc.b	$E0,$F4,$F7,$FE,$A9,$F5,$BD,$FF,$AF,$EF,$F8,$1F,$B3,$EF,$FD,$AF
	.dc.b	$7F,$ED,$FB,$FF,$73,$DF,$FB,$BE,$FF,$DE,$F7,$FE,$FF,$BF,$F8,$0F
	.dc.b	$7B,$F0,$3E,$F7,$E0,$BD,$EF,$C1,$FB,$DF,$ED,$0F,$B4,$7C,$ED,$F9
	.dc.b	$7C,$7C,$0E,$7E,$07,$5F,$03,$BF,$81,$E1,$F0,$3C,$7E,$07,$97,$C0
	.dc.b	$F3,$F8,$1E,$9F,$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7B,$FE,$1F
	.dc.b	$7F,$C5,$EF,$F8,$FC,$8F,$27,$BF,$E5,$F7,$FC,$DE,$FF,$FB,$9B,$ED
	.dc.b	$7F,$65,$E6,$1F,$F5,$21,$FE,$57,$3F,$64,$7E,$C5,$FF,$7A,$3F,$E5
	.dc.b	$66,$FB,$9F,$73,$C7,$E9,$EE,$75,$F7,$3B,$7B,$9F,$80,$F2,$7F,$03
	.dc.b	$E4,$FE,$0B,$C9,$FC,$1F,$93,$F8,$4F,$27,$F0,$BE,$4F,$E1,$BC,$9F
	.dc.b	$C3,$F9,$3F,$88,$F2,$7F,$13,$E4,$FE,$2B,$C9,$FF,$09,$7E,$C5,$FC
	.dc.b	$A8,$FE,$52,$3C,$7C,$F8,$FA,$F1,$F7,$E3,$F0,$F1,$F8,$F8,$FC,$BC
	.dc.b	$7E,$7E,$3F,$4F,$1F,$AF,$8F,$DB,$C7,$EF,$E3,$F8,$3C,$7F,$0F,$8F
	.dc.b	$E2,$F1,$FC,$7E,$3F,$93,$C7,$F2,$F8,$FE,$6F,$1F,$FC,$51,$FB,$1F
	.dc.b	$EE,$0F,$F1,$F1,$FB,$73,$E7,$FB,$DF,$57,$DE,$DF,$C0,$7D,$7F,$7F
	.dc.b	$EC,$7B,$FF,$67,$DF,$FB,$5E,$FF,$DB,$F7,$FE,$E7,$BF,$F7,$7D,$FF
	.dc.b	$BD,$EF,$FD,$FF,$7D,$E9,$CF,$B8,$FF,$63,$1F,$B1,$4F,$EC,$5B,$EF
	.dc.b	$F0,$73,$EF,$D9,$FE,$3B,$1C,$3D,$EC,$7B,$D9,$F7,$B5,$EF,$6F,$D9
	.dc.b	$E1,$EB,$71,$F5,$B9,$7A,$DC,$FD,$6E,$9E,$B7,$5F,$5B,$B7,$AD,$DF
	.dc.b	$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$79,$7B,$FF,$63,$C7,$EE,$5E,$E6
	.dc.b	$7D,$6D,$7A,$DB,$F5,$B8,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7,$AD
	.dc.b	$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B
	.dc.b	$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5,$BE,$67,$DC,$FE,$C8,$7C,$C6
	.dc.b	$3E,$43,$3F,$0B,$AE,$AD,$FC,$2F,$0F,$80,$E3,$F0,$1C,$BE,$03,$9F
	.dc.b	$C0,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$05,$F7,$5F,$83,$F3
	.dc.b	$0F,$E2,$4F,$C4,$0F,$7F,$7E,$FF,$0F,$85,$E3,$EF,$F2,$F7,$F9,$FB
	.dc.b	$FD,$3D,$FE,$BE,$FF,$6F,$7F,$BF,$BF,$E0,$F7,$FC,$3E,$FF,$8B,$DF
	.dc.b	$F1,$FB,$FE,$4F,$7F,$CB,$EF,$F9,$BD,$FF,$3F,$BF,$E8,$F7,$FF,$35
	.dc.b	$F7,$7F,$86,$E7,$3F,$86,$7F,$4A,$FF,$09,$6F,$BD,$C3,$DE,$E3,$EF
	.dc.b	$72,$F7,$B9,$FB,$DD,$3D,$EE,$BE,$F7,$6F,$7B,$BF,$BD,$E0,$F7,$BC
	.dc.b	$3E,$F7,$8B,$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD
	.dc.b	$E8,$F7,$BF,$65,$C5,$FC,$E0,$FC,$3F,$27,$EF,$17,$B9,$AF,$73,$7E
	.dc.b	$E7,$0F,$73,$8F,$B9,$CB,$DC,$E7,$EE,$74,$F7,$3A,$FB,$9D,$BD,$CE
	.dc.b	$FE,$E7,$83,$DC,$F0,$FB,$9E,$2F,$73,$C7,$EE,$79,$3D,$CF,$2F,$B9
	.dc.b	$E6,$F7,$3C,$FE,$E7,$F0,$9F,$BC,$FC,$41,$FC,$3C,$FE,$1C,$7E,$1F
	.dc.b	$9F,$F8,$AB,$FE,$9C,$3F,$86,$9F,$EC,$97,$3E,$C7,$3F,$63,$A7,$B1
	.dc.b	$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1,$7B,$1E,$3F,$63
	.dc.b	$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7,$F2,$1F,$BD,$FC
	.dc.b	$41,$F8,$7E,$23,$F8,$76,$FB,$1B,$F2,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3F,$9C,$7D
	.dc.b	$F7,$F7,$E3,$F0,$FC,$A7,$F0,$EF,$F4,$47,$FC,$E0,$7F,$7F,$75,$EB
	.dc.b	$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF,$AD,$E0,$F5,$BC
	.dc.b	$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79,$BD,$6F,$3F,$AD
	.dc.b	$E8,$F5,$BF,$A8,$7B,$9F,$C3,$CF,$EA,$C7,$F5,$53,$F1,$1C,$7F,$EE
	.dc.b	$0F,$77,$71,$C3,$DD,$E3,$EC,$72,$F6,$39,$FB,$1D,$3D,$8E,$BE,$C7
	.dc.b	$6F,$63,$BF,$B1,$E0,$F6,$3C,$3E,$C7,$8B,$D8,$F1,$FB,$1E,$4F,$63
	.dc.b	$CB,$EC,$79,$BD,$8F,$3F,$B1,$FD,$9F,$EF,$FF,$87,$1F,$87,$3F,$0C
	.dc.b	$3D,$8D,$7B,$1B,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7,$3F,$63,$A7,$B1
	.dc.b	$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1,$7B,$1E,$3F,$63
	.dc.b	$C9,$EC,$79,$7D,$8F,$37,$B1,$FD,$F7,$F0,$0F,$F2,$19,$F8,$63,$F0
	.dc.b	$BE,$41,$FF,$5A,$1F,$85,$3F,$C8,$4E,$7D,$7E,$5E,$BF,$3F,$5F,$A7
	.dc.b	$AF,$D7,$D7,$ED,$EB,$F7,$F5,$FC,$1E,$BF,$87,$D7,$F1,$7A,$FE,$3F
	.dc.b	$5F,$C9,$EB,$F9,$7D,$7F,$37,$AF,$E7,$F5,$FF,$C2,$9F,$80,$FF,$53
	.dc.b	$3F,$C3,$47,$F8,$64,$FC,$28,$F6,$37,$EC,$70,$F6,$38,$FB,$1C,$BD
	.dc.b	$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F,$B1
	.dc.b	$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7F,$8A
	.dc.b	$FD,$9F,$EA,$47,$F0,$B3,$F8,$51,$F8,$5F,$17,$FC,$6E,$BD,$7E,$1E
	.dc.b	$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F
	.dc.b	$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB
	.dc.b	$F9,$FD,$7F,$F2,$17,$E0,$5F,$E8,$83,$F8,$59,$F8,$63,$FD,$F0,$7F
	.dc.b	$09,$3F,$0A,$3D,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6
	.dc.b	$F5,$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC
	.dc.b	$BE,$BF,$9B,$D7,$FF,$2C,$7E,$07,$FE,$17,$3F,$CB,$C7,$F9,$74,$FE
	.dc.b	$12,$7E,$17,$C5,$FF,$32,$BF,$E5,$C3,$FA,$82,$5E,$A7,$4F,$53,$AF
	.dc.b	$A9,$DB,$D4,$EF,$EA,$78,$3D,$4F,$0F,$A9,$E2,$F5,$3C,$7E,$A7,$93
	.dc.b	$D4,$F2,$FA,$9E,$6F,$53,$CF,$EA,$7A,$3D,$4F,$4F,$A9,$FE,$71,$FC
	.dc.b	$13,$FC,$30,$7F,$51,$3F,$A8,$1F,$87,$E3,$3F,$87,$7F,$CF,$EB,$D3
	.dc.b	$E5,$E9,$F3,$F4,$FA,$7A,$7D,$7D,$3E,$DE,$9F,$7F,$4F,$C1,$E9,$F8
	.dc.b	$7D,$3F,$17,$A7,$E3,$F4,$FC,$9E,$9F,$97,$D3,$F3,$7A,$7E,$7F,$4F
	.dc.b	$D1,$E9,$FF,$A2,$78,$7F,$D9,$E7,$F5,$23,$F8,$68,$F8,$1D,$7C,$0E
	.dc.b	$FD,$7E,$1E,$AF,$1F,$57,$97,$AB,$CF,$D5,$E9,$EA,$F5,$F5,$7B,$7A
	.dc.b	$BD,$FD,$5F,$07,$AB,$E1,$F5,$7C,$5E,$AF,$8F,$D5,$F2,$7A,$BE,$5F
	.dc.b	$57,$CD,$EA,$FF,$A7,$FF,$05,$FF,$2F,$1F,$EA,$53,$FD,$48,$7F,$50
	.dc.b	$3F,$11,$E3,$1F,$F5,$19,$F8,$83,$F0,$EF,$FA,$C1,$7A,$9D,$7D,$4E
	.dc.b	$DE,$A7,$7F,$53,$C1,$EA,$78,$7D,$4F,$17,$A9,$E3,$F5,$3C,$9E,$A7
	.dc.b	$97,$D4,$F3,$7A,$9E,$7F,$53,$D1,$EA,$7A,$7D,$4F,$F5,$C7,$E0,$DF
	.dc.b	$E5,$F3,$FC,$30,$7F,$85,$CF,$C4,$1F,$87,$E2,$3F,$87,$7F,$D8,$EB
	.dc.b	$D4,$E7,$EA,$74,$F5,$3A,$FA,$9D,$BD,$4E,$FE,$A7,$83,$D4,$F0,$FA
	.dc.b	$9E,$2F,$53,$C7,$EA,$79,$3D,$4F,$2F,$A9,$E6,$F5,$3C,$FE,$A7,$A3
	.dc.b	$D4,$FF,$68,$7E,$0E,$F9,$EB,$F1,$58,$F8,$7C,$FA,$FA,$F5,$F7,$EB
	.dc.b	$F0,$F5,$F8,$FA,$FC,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF
	.dc.b	$EB,$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE
	.dc.b	$6F,$5F,$FD,$CD,$EF,$7F,$A9,$4F,$F7,$61,$FE,$EB,$3F,$C2,$E7,$E2
	.dc.b	$0F,$F7,$49,$FC,$38,$FC,$3F,$92,$7F,$0E,$FF,$BE,$17,$AD,$DB,$D6
	.dc.b	$EF,$EB,$78,$3D,$6F,$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6,$F2,$FA
	.dc.b	$DE,$6F,$5B,$CF,$EB,$7A,$3D,$6F,$4F,$AD,$FE,$FF,$FC,$23,$FD,$4C
	.dc.b	$7F,$97,$CF,$F2,$F1,$FC,$3C,$FC,$41,$F8,$7E,$13,$F8,$74,$63,$D5
	.dc.b	$E9,$EA,$F5,$F5,$7B,$7A,$BD,$FD,$5F,$07,$AB,$E1,$F5,$7C,$5E,$AF
	.dc.b	$8F,$D5,$F2,$7A,$BE,$5F,$57,$CD,$EA,$F9,$FD,$5F,$47,$AB,$E0,$E0
	.dc.b	$7E,$13,$F8,$97,$C6,$FF,$98,$8F,$F0,$E9,$FC,$45,$CF,$A3,$C6,$3F
	.dc.b	$C3,$A7,$F5,$98,$E3,$E7,$E7,$CF,$D7,$9E,$FE,$01,$85,$AE,$F1,$FE
	.dc.b	$E7,$77,$F2,$FF,$13,$E4,$1C,$BC,$A7,$47,$A2,$95,$B9,$D4,$F3,$C9
	.dc.b	$5A,$85,$F4,$7C,$27,$49,$DD,$F3,$8C,$9C,$EB,$95,$71,$8F,$11,$C6
	.dc.b	$33,$7E,$28,$F9,$C7,$9E,$BA,$3C,$0E,$8E,$2F,$CF,$1C,$E3,$15,$F0
	.dc.b	$23,$91,$EA,$AE,$CF,$C8,$30,$7A,$0F,$0D,$7B,$1C,$D3,$E4,$FF,$CA
	.dc.b	$B1,$F8,$3F,$C0,$7F,$CE,$1D,$2F,$97,$FC,$AF,$E1,$56,$BE,$7D,$FD
	.dc.b	$F7,$DB,$7F,$E7,$7E,$BD,$56,$F1,$ED,$EC,$7D,$8A,$EB,$8F,$43,$F4
	.dc.b	$33,$E3,$F6,$35,$F1,$8C,$79,$D6,$B3,$71,$EB,$7C,$EB,$E4,$8C,$57
	.dc.b	$58,$F3,$F1,$F4,$EE,$DE,$9F,$06,$F7,$4D,$FD,$19,$7D,$28,$F6,$E8
	.dc.b	$EE,$E3,$2F,$67,$CF,$DF,$8E,$5F,$4D,$6A,$73,$B7,$8F,$4F,$A1,$EC
	.dc.b	$FB,$BC,$0E,$6F,$C0,$F0,$3A,$FB,$39,$BE,$1A,$2B,$51,$8F,$67,$E3
	.dc.b	$F8,$5E,$4F,$66,$CE,$78,$7F,$59,$D3,$FE,$AB,$F0,$F5,$18,$E3,$61
	.dc.b	$EC,$F6,$7E,$8F,$89,$EE,$F6,$7E,$9F,$8B,$9C,$38,$4E,$5E,$51,$F1
	.dc.b	$3D,$9F,$B3,$E3,$7C,$6F,$67,$F0,$1F,$45,$FD,$57,$7F,$5C,$69,$F6
	.dc.b	$D6,$BD,$EF,$9B,$ED,$F0,$F7,$BE,$7F,$B9,$6F,$E1,$BD,$3D,$ED,$EB
	.dc.b	$EE,$CD,$7F,$7B,$F6,$5F,$D4,$72,$A7,$FC,$55,$8F,$D3,$97,$D1,$79
	.dc.b	$BE,$03,$83,$E8,$F4,$39,$65,$9B,$E1,$F4,$3E,$CF,$37,$F9,$57,$8F
	.dc.b	$F9,$FC,$BF,$DD,$F7,$FA,$BC,$FF,$4D,$39,$BF,$87,$E6,$F0,$C7,$F4
	.dc.b	$38,$70,$C3,$FA,$1E,$BE,$7C,$B5,$1F,$C7,$A7,$FC,$56,$F5,$F2,$73
	.dc.b	$E2,$A7,$33,$CB,$4F,$D3,$BE,$FF,$8D,$FA,$77,$C2,$B7,$2C,$7E,$37
	.dc.b	$99,$9F,$82,$FB,$3E,$5E,$FE,$2B,$ED,$F9,$79,$C3,$E8,$39,$7E,$94
	.dc.b	$F5,$F8,$7F,$C0,$F9,$1F,$4F,$E1,$3F,$07,$E5,$DB,$FB,$13,$F5,$EB
	.dc.b	$3C,$7F,$57,$E1,$3F,$13,$E5,$FD,$BF,$84,$FC,$5F,$99,$71,$FC,$04
	.dc.b	$FD,$FA,$CD,$FD,$DF,$84,$FC,$8F,$95,$F8,$36,$C6,$6F,$F1,$1E,$57
	.dc.b	$E1,$FE,$13,$31,$FC,$0E,$7E,$AF,$C2,$7E,$6B,$C8,$F9,$0F,$E7,$3C
	.dc.b	$9F,$95,$F0,$96,$FF,$19,$3B,$7F,$1D,$F0,$9F,$A2,$F2,$BE,$87,$C2
	.dc.b	$7E,$93,$CB,$FA,$53,$6B,$F2,$93,$B7,$F2,$DF,$09,$FC,$A7,$E8,$FF
	.dc.b	$A5,$F9,$1F,$8F,$F8,$5C,$57,$DB,$AA,$FE,$76,$38,$79,$7F,$7F,$E1
	.dc.b	$79,$79,$7F,$81,$F8,$5B,$7F,$A3,$6B,$F0,$4F,$C8,$4F,$87,$E5,$FF
	.dc.b	$1F,$8F,$F4,$39,$FE,$83,$9F,$F9,$FB,$3F,$E8,$12,$F9,$AF,$47,$C8
	.dc.b	$F0,$6D,$FF,$23,$9E,$4F,$F5,$1B,$79,$7D,$BE,$17,$E0,$79,$16,$BF
	.dc.b	$8D,$D7,$FB,$43,$2E,$F5,$A9,$E9,$ED,$7C,$7F,$27,$CF,$F2,$3F,$27
	.dc.b	$C5,$FE,$83,$A7,$F8,$CE,$FC,$1E,$46,$3E,$3B,$5D,$FE,$5C,$3E,$84
	.dc.b	$F0,$F8,$CF,$A5,$E2,$DC,$3C,$6D,$7E,$1F,$91,$DB,$E2,$B7,$8F,$BF
	.dc.b	$E2,$3E,$D7,$8F,$E2,$F8,$8F,$B9,$E3,$DC,$7D,$B3,$E6,$AC,$DF,$C7
	.dc.b	$F8,$8F,$C1,$78,$BF,$4B,$E2,$3F,$09,$E3,$DC,$7E,$2B,$F5,$AB,$71
	.dc.b	$F3,$FE,$1F,$F1,$5E,$37,$DA,$6C,$64,$F9,$B1,$F7,$BE,$23,$F2,$1E
	.dc.b	$3F,$E0,$3E,$23,$F2,$5D,$1F,$82,$F8,$8B,$87,$DB,$6B,$F8,$9F,$6D
	.dc.b	$2A,$FC,$D7,$95,$F8,$AF,$86,$FC,$E7,$91,$F5,$3E,$1F,$F3,$DE,$55
	.dc.b	$CB,$EF,$CF,$D8,$AC,$F2,$7E,$27,$E1,$FF,$C1,$9F,$43,$FE,$17,$8F
	.dc.b	$C6,$7C,$46,$3C,$BF,$BE,$EA,$3F,$03,$59,$7F,$25,$F1,$1C,$BC,$8F
	.dc.b	$F0,$2F,$EC,$DF,$E6,$3D,$3F,$C3,$B9,$3C,$9C,$FC,$4E,$F1,$E3,$5F
	.dc.b	$DB,$AB,$51,$FD,$FE,$FF,$A9,$CE,$7C,$7F,$E6,$F0,$F8,$9F,$4F,$93
	.dc.b	$86,$D7,$1F,$89,$F6,$DF,$0F,$89,$F7,$F9,$5C,$BE,$27,$E0,$F9,$37
	.dc.b	$0F,$13,$B7,$C9,$3E,$0F,$88,$F8,$FE,$3D,$9F,$43,$5F,$7B,$E6,$F8
	.dc.b	$9F,$9B,$E5,$FA,$3E,$23,$E7,$F8,$F6,$FE,$03,$EC,$F8,$6C,$C7,$C1
	.dc.b	$9F,$AB,$E5,$7C,$0F,$85,$FA,$FE,$55,$AF,$90,$FC,$4F,$86,$CC,$7C
	.dc.b	$99,$FB,$BE,$57,$C8,$F8,$5F,$BF,$E3,$DA,$FA,$1E,$57,$CC,$F8,$6F
	.dc.b	$C1,$B9,$8F,$B3,$3F,$43,$E1,$7F,$0F,$E3,$7D,$E7,$F1,$3E,$4D,$AF
	.dc.b	$B5,$3F,$82,$F8,$6C,$F8,$1F,$6F,$E1,$BF,$1F,$E3,$FE,$19,$FC,$8F
	.dc.b	$91,$70,$FC,$14,$FE,$2B,$E1,$B3,$7F,$83,$F8,$6F,$CB,$F9,$5F,$8E
	.dc.b	$F8,$6F,$F3,$27,$C0,$FF,$89,$71,$FD,$47,$90,$9D,$E1,$C6,$7F,$C7
	.dc.b	$75,$BF,$7B,$FC,$DD,$6F,$86,$E7,$E4,$7E,$6B,$E1,$BA,$F9,$57,$1E
	.dc.b	$33,$F9,$EA,$CE,$5F,$E5,$5F,$D9,$FF,$C4,$F9,$FF,$89,$AF,$F7,$51
	.dc.b	$8F,$27,$FC,$D2,$EB,$CA,$BF,$0F,$EA,$9C,$FC,$3F,$B3,$CA,$DF,$C3
	.dc.b	$DC,$FB,$3F,$03,$CA,$DE,$7E,$CA,$EB,$A7,$C4,$F1,$FF,$C3,$BA,$FF
	.dc.b	$85,$BE,$21,$3F,$E1,$CB,$F2,$FC,$46,$DF,$54,$7C,$CF,$2B,$8F,$C4
	.dc.b	$7C,$EF,$22,$D7,$CA,$F2,$7A,$FC,$3F,$D2,$73,$1F,$E9,$CC,$3F,$D0
	.dc.b	$5D,$5F,$E9,$BD,$1F,$0F,$F6,$3C,$AF,$2B,$6F,$E5,$7C,$3D,$7E,$4C
	.dc.b	$7D,$EF,$2F,$D7,$F0,$FF,$80,$F2,$BE,$5F,$C3,$DE,$8F,$A4,$FE,$13
	.dc.b	$C9,$F9,$F5,$9C,$7E,$9F,$C3,$7E,$23,$CA,$FA,$6D,$C3,$ED,$3F,$8C
	.dc.b	$F2,$3E,$BF,$C3,$E6,$FE,$DF,$C3,$FE,$43,$CA,$FB,$6D,$C3,$F0,$4F
	.dc.b	$E5,$3C,$9F,$BF,$F0,$F9,$BF,$C1,$FC,$3F,$E6,$3C,$7F,$C1,$B7,$0F
	.dc.b	$C5,$3F,$9C,$F2,$BF,$0F,$F0,$F9,$BF,$C5,$FC,$37,$FB,$0F,$E8,$51
	.dc.b	$F8,$C8,$C7,$95,$63,$FD,$33,$3A,$C7,$F2,$DF,$0F,$A7,$F3,$1F,$0D
	.dc.b	$CF,$CB,$FC,$AF,$C3,$D9,$FC,$E3,$DF,$CC,$FC,$CD,$67,$1F,$EA,$DF
	.dc.b	$B4,$7F,$84,$9F,$F3,$8E,$5F,$E1,$BC,$3F,$DA,$B3,$FE,$1B,$9F,$1F
	.dc.b	$93,$FE,$70,$8E,$11,$FE,$69,$AF,$F0,$CB,$FE,$BF,$8E,$0F,$FB,$9A
	.dc.b	$3A,$2F,$89,$4B,$E2,$3E,$27,$95,$D7,$E2,$3E,$37,$95,$DE,$6E,$1D
	.dc.b	$67,$6F,$7F,$85,$F9,$9E,$2F,$97,$E1,$AC,$E4,$F9,$71,$97,$E6,$F8
	.dc.b	$FE,$BF,$85,$FA,$9E,$27,$B7,$E1,$BE,$B7,$8F,$6F,$E9,$BB,$7E,$AC
	.dc.b	$FC,$3F,$86,$FB,$9E,$35,$9F,$B2,$D7,$E1,$3F,$27,$E1,$BF,$05,$E3
	.dc.b	$7C,$BF,$86,$FC,$27,$8D,$6F,$F0,$2D,$7E,$53,$F4,$7E,$1B,$F1,$5E
	.dc.b	$37,$D3,$F8,$6F,$C6,$78,$DF,$57,$E1,$AE,$1F,$4D,$AF,$E0,$27,$F2
	.dc.b	$5E,$37,$DB,$F8,$5F,$CA,$78,$F6,$BF,$23,$E3,$57,$ED,$3F,$81,$F8
	.dc.b	$6D,$C7,$DF,$F8,$5F,$CE,$78,$BF,$85,$7F,$3D,$E3,$D9,$FC,$DB,$F8
	.dc.b	$9F,$86,$DC,$7E,$1F,$E1,$7F,$46,$C3,$FD,$69,$F4,$5F,$E1,$D7,$F2
	.dc.b	$FF,$0F,$97,$FD,$8F,$19,$F6,$5E,$5F,$E6,$2F,$55,$CB,$E2,$B2,$F2
	.dc.b	$39,$7B,$EC,$39,$5F,$E8,$3E,$23,$8C,$E7,$C8,$E3,$97,$2F,$33,$F3
	.dc.b	$35,$AC,$3F,$35,$87,$EF,$2B,$5E,$A7,$F8,$AF,$2F,$6E,$5D,$FC,$AE
	.dc.b	$77,$DE,$7C,$78,$72,$7F,$C9,$57,$FE,$2A,$7E,$25,$F9,$FC,$AE,$31
	.dc.b	$FA,$B9,$FF,$16,$3C,$BC,$AC,$E9,$FD,$AF,$E1,$BC,$D7,$8B,$DD,$F4
	.dc.b	$C5,$86,$7E,$1B,$D9,$E5,$7F,$BB,$AD,$1E,$FF,$2B,$1F,$0D,$71,$F8
	.dc.b	$1B,$7F,$1C,$D7,$FB,$D5,$F2,$9C,$9E,$7F,$85,$F9,$1E,$57,$5F,$86
	.dc.b	$F9,$5E,$57,$79,$B5,$C3,$0F,$F0,$25,$70,$8C,$69,$FD,$8D,$F9,$BB
	.dc.b	$BC,$9F,$47,$C2,$FD,$57,$D5,$F0,$D6,$B8,$CF,$D9,$F2,$BD,$DF,$0D
	.dc.b	$F6,$FC,$8F,$81,$F0,$D6,$FB,$CD,$7E,$AC,$7E,$07,$CA,$F8,$DF,$09
	.dc.b	$F8,$3F,$2A,$D7,$DC,$7E,$57,$C2,$66,$3D,$33,$F8,$9F,$27,$E7,$7C
	.dc.b	$27,$E2,$FC,$9F,$A1,$F0,$9F,$8D,$F2,$73,$87,$E1,$9C,$BF,$88,$9F
	.dc.b	$AD,$F0,$9F,$93,$F2,$6C,$FE,$31,$FB,$5F,$09,$98,$F9,$33,$F9,$9F
	.dc.b	$2F,$EF,$7C,$27,$E6,$FC,$AB,$5F,$94,$7F,$05,$D5,$98,$FA,$33,$FA
	.dc.b	$6F,$A1,$FF,$44,$79,$7F,$A0,$F8,$5C,$79,$5F,$89,$6D,$FD,$8F,$85
	.dc.b	$AF,$E7,$A3,$97,$C4,$7F,$02,$BF,$D7,$7E,$D3,$FB,$AA,$79,$CC,$BF
	.dc.b	$C6,$D5,$AF,$87,$F1,$79,$7F,$A0,$C3,$85,$EF,$CA,$C5,$6B,$CB,$FC
	.dc.b	$65,$3E,$1F,$D5,$E5,$71,$6C,$EB,$E1,$FD,$DE,$56,$76,$FE,$13,$B3
	.dc.b	$0F,$07,$C4,$FC,$3F,$26,$D6,$BC,$9F,$27,$C4,$67,$3F,$0D,$FC,$99
	.dc.b	$F4,$7C,$37,$CB,$F2,$7D,$5F,$0D,$6F,$CF,$3F,$3F,$CA,$CE,$5E,$2B
	.dc.b	$F8,$1F,$0D,$F4,$FC,$8F,$85,$F0,$DF,$57,$CA,$CD,$FA,$27,$2F,$AA
	.dc.b	$3E,$47,$C2,$FD,$BF,$2B,$E5,$7C,$35,$D7,$E3,$CF,$DF,$F2,$F3,$87
	.dc.b	$C2,$BF,$A1,$F0,$DF,$83,$F2,$3E,$D6,$1F,$C0,$FA,$DF,$0D,$9C,$3E
	.dc.b	$8D,$D7,$FD,$D1,$3A,$7E,$6F,$95,$C9,$FF,$04,$43,$E1,$BF,$21,$E4
	.dc.b	$67,$D6,$4E,$6B,$EA,$CE,$61,$F0,$DC,$27,$F0,$37,$F9,$9F,$2F,$5F
	.dc.b	$0D,$F9,$BF,$2B,$87,$C3,$7E,$77,$CC,$CE,$1F,$80,$72,$FE,$0A,$7B
	.dc.b	$7C,$37,$F4,$8F,$2F,$FA,$D7,$99,$F9,$A8,$C7,$95,$E3,$F8,$6B,$87
	.dc.b	$E4,$A7,$83,$E7,$F8,$6E,$5E,$65,$8F,$EA,$F8,$FE,$5D,$FC,$CF,$C3
	.dc.b	$69,$FC,$DD,$78,$7C,$DF,$89,$FB,$57,$F8,$DB,$D0,$61,$FE,$32,$E7
	.dc.b	$4E,$63,$FC,$73,$83,$6F,$EB,$D3,$FE,$3A,$8F,$EB,$33,$A9,$CF,$5F
	.dc.b	$BF,$CA,$EF,$F0,$DF,$07,$C9,$FE,$A3,$8F,$F8,$F7,$1F,$4C,$63,$E1
	.dc.b	$BE,$37,$97,$AB,$E1,$8F,$F3,$FC,$1F,$0D,$F2,$FC,$AE,$7F,$0D,$F3
	.dc.b	$7C,$BE,$B3,$9B,$E1,$39,$79,$7C,$37,$D3,$F2,$7C,$7F,$0D,$F5,$7C
	.dc.b	$BF,$2C,$E6,$BC,$13,$97,$C5,$F0,$DF,$6F,$C8,$F9,$6D,$E2,$C9,$F5
	.dc.b	$23,$E7,$FC,$37,$E0,$7C,$AF,$A3,$F0,$BF,$83,$F1,$ED,$FD,$AF,$22
	.dc.b	$BF,$DF,$3A,$BF,$2A,$FC,$1F,$85,$E0,$FC,$3F,$85,$FC,$67,$93,$F7
	.dc.b	$1B,$19,$3F,$0B,$1F,$80,$F8,$5F,$C9,$79,$5F,$82,$F8,$5F,$CA,$78
	.dc.b	$F6,$FF,$17,$E4,$7E,$1B,$E1,$7F,$30,$E6,$3C,$93,$F8,$AF,$84,$FC
	.dc.b	$E7,$91,$67,$F2,$7E,$4F,$E3,$BE,$17,$F4,$0E,$63,$D9,$3F,$92,$F8
	.dc.b	$5F,$F1,$5F,$D1,$FF,$8C,$A7,$F1,$DD,$5F,$E0,$7F,$0D,$F0,$DA,$F8
	.dc.b	$8E,$15,$FE,$36,$BF,$F4,$5F,$97,$F9,$4A,$E7,$87,$E7,$BE,$1F,$AF
	.dc.b	$91,$F9,$06,$C7,$F8,$27,$F6,$BF,$F1,$96,$3F,$EB,$0F,$3D,$93,$CC
	.dc.b	$E1,$3D,$F1,$FF,$14,$DF,$F8,$C6,$3F,$C7,$35,$F9,$4C,$3A,$7C,$3F
	.dc.b	$B3,$CA,$DF,$C4,$7B,$BC,$BE,$33,$9A,$F1,$7C,$47,$C2,$F2,$BA,$F8
	.dc.b	$56,$73,$9F,$BF,$CA,$CB,$F0,$6B,$C7,$F0,$DF,$2B,$CB,$FE,$41,$BF
	.dc.b	$F5,$1A,$F8,$DE,$56,$3E,$23,$E7,$BA,$BE,$13,$F0,$FE,$1F,$E9,$F9
	.dc.b	$7C,$BE,$1A,$CE,$4F,$A1,$19,$7E,$97,$95,$E0,$F8,$5F,$B7,$E5,$F8
	.dc.b	$BE,$1A,$D7,$CF,$9A,$FD,$E3,$FD,$81,$AA,$7F,$07,$E5,$FA,$BE,$1B
	.dc.b	$F0,$BE,$5F,$D0,$F8,$6F,$C3,$F9,$97,$1F,$C1,$4F,$D4,$AC,$F3,$7D
	.dc.b	$DF,$85,$FC,$6F,$97,$F6,$1F,$C7,$F9,$77,$0F,$C5,$4F,$DC,$F8,$6C
	.dc.b	$DF,$E1,$7E,$1B,$F2,$BE,$5F,$E0,$3E,$1B,$F2,$FE,$4D,$C7,$F2,$53
	.dc.b	$F8,$4A,$CD,$FE,$37,$E1,$BF,$3B,$E5,$7E,$21,$FF,$44,$FD,$23,$E4
	.dc.b	$3F,$17,$F0,$FA,$BE,$19,$7F,$A4,$AB,$51,$F9,$7F,$86,$E3,$E4,$FE
	.dc.b	$4B,$E1,$F9,$F9,$3F,$94,$9C,$D7,$E7,$67,$2F,$F9,$9F,$6F,$F1,$0E
	.dc.b	$1F,$E2,$68,$FF,$3A,$43,$E1,$FC,$BE,$57,$FA,$1A,$37,$E5,$59,$CD
	.dc.b	$73,$F8,$7F,$5F,$8D,$FD,$BE,$DF,$0F,$EE,$F2,$70,$DE,$0E,$07,$B6
	.dc.b	$B8,$7C,$37,$C4,$F2,$79,$7C,$4F,$C6,$F2,$AD,$FC,$37,$6F,$C5,$9F
	.dc.b	$07,$C3,$FC,$CF,$23,$C5,$F0,$F6,$BE,$03,$5F,$C8,$7C,$43,$65,$FD
	.dc.b	$4F,$27,$D5,$F1,$1F,$5B,$C6,$F6,$7C,$47,$D8,$F2,$AE,$5F,$55,$DB
	.dc.b	$F5,$E7,$E1,$7C,$47,$DE,$F2,$7E,$27,$C4,$5A,$FA,$4D,$7C,$7A,$2B
	.dc.b	$F0,$BE,$47,$CB,$F8,$8F,$C3,$F8,$FF,$37,$E2,$3F,$13,$E5,$5C,$7F
	.dc.b	$0C,$ED,$FC,$44,$FD,$3F,$88,$FC,$7F,$93,$F6,$DF,$C8,$F9,$1F,$77
	.dc.b	$E2,$33,$E0,$FE,$16,$72,$FE,$1F,$E2,$3F,$2F,$E4,$7E,$0D,$FC,$CF
	.dc.b	$91,$F8,$5F,$88,$CD,$6F,$4F,$CC,$4F,$E2,$7E,$1F,$FD,$CF,$F4,$89
	.dc.b	$C1,$FE,$EE,$71,$5F,$95,$F8,$9D,$F9,$3A,$AF,$F6,$ED,$69,$F8,$FF
	.dc.b	$11,$D3,$CA,$FC,$87,$C4,$F6,$F2,$BF,$25,$36,$BF,$41,$3B,$7F,$D9
	.dc.b	$1F,$B6,$7F,$88,$70,$FF,$10,$37,$CA,$FF,$24,$3F,$E1,$D9,$FE,$89
	.dc.b	$3A,$D3,$D7,$E5,$6B,$E1,$7D,$BE,$35,$BF,$63,$CB,$E1,$B3,$1E,$09
	.dc.b	$F8,$7E,$4F,$6F,$85,$F8,$BE,$45,$AF,$89,$E4,$57,$E0,$BE,$4F,$85
	.dc.b	$DC,$7A,$BE,$13,$E6,$F8,$BE,$87,$E7,$F8,$F6,$7E,$73,$EC,$F8,$5D
	.dc.b	$C7,$C2,$F8,$5F,$AB,$E3,$FC,$07,$EB,$F8,$F6,$7E,$B3,$F1,$3E,$17
	.dc.b	$71,$F2,$BE,$17,$EE,$F8,$8E,$6F,$9B,$F0,$B8,$75,$5F,$7F,$C6,$D3
	.dc.b	$FB,$2E,$A5,$6E,$31,$F0,$BF,$88,$F1,$B7,$F1,$1C,$3D,$36,$0A,$FC
	.dc.b	$6F,$91,$AC,$FF,$1D,$5A,$F8,$5E,$51,$F8,$0A,$DB,$F8,$2F,$84,$FC
	.dc.b	$AF,$93,$C7,$46,$1D,$A7,$35,$F8,$8F,$85,$FC,$DF,$95,$BA,$D5,$FE
	.dc.b	$6F,$0C,$68,$F8,$4F,$D0,$F9,$18,$F8,$5F,$D1,$F9,$37,$0F,$D1,$4F
	.dc.b	$0A,$CE,$9F,$96,$F5,$3F,$51,$50,$FF,$6F,$F8,$7F,$3E,$75,$19,$33
	.dc.b	$F1,$3E,$2A,$C4,$70,$9F,$2F,$C3,$63,$E2,$7F,$CD,$77,$E8,$F8,$7C
	.dc.b	$62,$F2,$7D,$DF,$B6,$FF,$89,$B2,$F0,$C6,$3E,$23,$7E,$67,$19,$F1
	.dc.b	$65,$FE,$76,$AE,$3F,$11,$D3,$1B,$5A,$5E,$57,$6F,$88,$D6,$FF,$C1
	.dc.b	$FB,$BF,$01,$AF,$F8,$BF,$0F,$F1,$B6,$1C,$2B,$47,$E0,$25,$F0,$FE
	.dc.b	$AF,$2F,$D2,$D9,$DD,$FD,$EA,$FF,$50,$BE,$3F,$2B,$6F,$96,$FE,$0F
	.dc.b	$C2,$FC,$4F,$2A,$CF,$A5,$F8,$BF,$0D,$F2,$3C,$AC,$CF,$D7,$E2,$FF
	.dc.b	$4B,$D7,$CA,$F8,$5F,$9B,$E5,$FC,$CF,$86,$BB,$79,$3C,$26,$EB,$FA
	.dc.b	$7E,$56,$F3,$F9,$18,$FD,$4F,$86,$FA,$FE,$45,$9F,$98,$FD,$8F,$86
	.dc.b	$FB,$7E,$56,$67,$C9,$CD,$FA,$9D,$7C,$1A,$3E,$17,$F0,$3E,$5F,$E0
	.dc.b	$3E,$1B,$F0,$7E,$67,$E0,$A7,$37,$C2,$72,$F2,$F8,$6F,$C4,$F8,$FF
	.dc.b	$88,$F8,$6F,$C5,$F9,$77,$5F,$C0,$4F,$E3,$2B,$37,$E2,$F8,$5F,$C8
	.dc.b	$F9,$5F,$90,$7F,$27,$E5,$5C,$3F,$11,$3F,$94,$F8,$6C,$DF,$AB,$E1
	.dc.b	$BF,$33,$E5,$FE,$61,$B1,$9B,$FC,$AF,$D3,$7F,$8B,$1F,$F0,$1F,$1B
	.dc.b	$E1,$F7,$18,$F8,$7F,$D3,$53,$CB,$FD,$9F,$93,$F9,$4F,$87,$CF,$97
	.dc.b	$8C,$FF,$25,$86,$31,$F9,$B5,$CB,$0D,$47,$F0,$EF,$DB,$BF,$6A,$BC
	.dc.b	$66,$7F,$9A,$8F,$C2,$4E,$A7,$3E,$67,$2F,$15,$3F,$E3,$9A,$F6,$CE
	.dc.b	$63,$85,$FE,$7A,$7F,$C4,$30,$F8,$7F,$6F,$97,$BF,$87,$F7,$F9,$9C
	.dc.b	$66,$DF,$8A,$76,$F9,$3E,$1F,$E2,$F9,$5D,$FE,$1A,$CF,$A1,$AF,$82
	.dc.b	$7E,$5F,$99,$E5,$F8,$6F,$9B,$E5,$F9,$FE,$1A,$DF,$C0,$6B,$E8,$9F
	.dc.b	$A7,$E5,$FB,$7E,$1B,$EA,$F9,$16,$BE,$EB,$5F,$D9,$71,$7E,$75,$F9
	.dc.b	$7F,$0D,$F7,$3C,$9F,$8D,$F0,$DF,$7B,$C7,$B5,$F8,$6F,$2A,$BC,$9F
	.dc.b	$99,$F0,$DB,$8F,$AB,$F0,$DF,$86,$F2,$BE,$83,$F8,$8F,$22,$CF,$E4
	.dc.b	$1F,$A9,$F0,$DB,$8F,$BB,$F0,$DF,$8E,$F2,$7E,$C7,$C3,$59,$FC,$0B
	.dc.b	$5F,$B3,$3F,$94,$F2,$7E,$F7,$C3,$7E,$5B,$C7,$FE,$A9,$E1,$23,$F3
	.dc.b	$3E,$3E,$3E,$1B,$5E,$4F,$E3,$27,$2F,$E3,$BE,$17,$FA,$27,$D3,$FF
	.dc.b	$8B,$FC,$DF,$EE,$18,$A7,$3F,$13,$DB,$E2,$37,$BF,$82,$73,$8F,$E6
	.dc.b	$3E,$1B,$9F,$93,$E4,$7A,$F9,$3E,$6F,$87,$CD,$6F,$0F,$E9,$3A,$FF
	.dc.b	$A2,$A3,$F9,$26,$5F,$E0,$89,$FE,$59,$2F,$23,$FB,$94,$FF,$8D,$E7
	.dc.b	$F5,$53,$9A,$E3,$F0,$FE,$CF,$2B,$71,$EE,$C3,$8F,$C2,$FC,$0F,$23
	.dc.b	$9F,$C3,$7C,$2F,$2A,$E7,$E4,$9D,$BE,$68,$F0,$FB,$BF,$23,$CA,$B3
	.dc.b	$EA,$6B,$E8,$F2,$3C,$FF,$0D,$BD,$3D,$BF,$0B,$F4,$3C,$9F,$5B,$63
	.dc.b	$E0,$FC,$37,$D4,$F2,$33,$8F,$C6,$BF,$83,$F0,$BF,$63,$C8,$B3,$F2
	.dc.b	$9F,$8B,$F0,$DF,$73,$C8,$CC,$FC,$A9,$F9,$3F,$0B,$F8,$0F,$1F,$E5
	.dc.b	$FC,$37,$E0,$BC,$9C,$DF,$D4,$9C,$BF,$5A,$3E,$8F,$C3,$7E,$23,$C6
	.dc.b	$FA,$7F,$0D,$78,$7E,$CB,$F8,$CF,$23,$78,$7D,$EA,$FB,$3F,$0D,$F9
	.dc.b	$0F,$1B,$ED,$FC,$35,$C7,$F0,$2F,$E5,$3C,$3D,$E1,$F8,$6A,$FC,$0F
	.dc.b	$C2,$FE,$63,$C7,$B3,$F8,$AF,$1A,$BF,$88,$8F,$C3,$FC,$37,$E7,$A7
	.dc.b	$F1,$3F,$0B,$FE,$23,$FA,$7F,$F1,$65,$FF,$B1,$B2,$7C,$3E,$75,$FC
	.dc.b	$7B,$BF,$27,$8D,$7F,$88,$2F,$8F,$9F,$FE,$57,$C1,$F0,$FD,$7C,$9F
	.dc.b	$F2,$47,$ED,$FF,$E2,$BF,$85,$60,$F2,$F3,$F1,$3B,$DB,$F4,$38,$F9
	.dc.b	$67,$8F,$C3,$F9,$FC,$BE,$7F,$11,$E9,$F2,$AF,$07,$8E,$76,$F9,$63
	.dc.b	$C3,$F0,$FE,$FF,$2E,$CF,$A7,$CB,$AF,$F8,$67,$64,$FA,$3E,$33,$E3
	.dc.b	$57,$AB,$E2,$3E,$47,$97,$EC,$F8,$8F,$95,$E6,$5C,$BE,$24,$ED,$F8
	.dc.b	$D1,$F0,$BE,$23,$E8,$79,$76,$7E,$57,$97,$5F,$14,$7C,$8F,$88,$FA
	.dc.b	$D3,$F2,$BE,$23,$EC,$79,$7F,$D8,$FA,$DF,$11,$8C,$7F,$C5,$90,$AC
	.dc.b	$43,$BF,$FD,$19,$3C,$23,$ED,$79,$9C,$FE,$23,$F0,$9E,$5F,$5F,$87
	.dc.b	$FC,$37,$99,$9B,$FC,$03,$97,$F0,$53,$E3,$F8,$7F,$C6,$79,$9E,$5F
	.dc.b	$87,$FC,$77,$99,$9B,$FC,$43,$97,$F1,$53,$EB,$F8,$7F,$CA,$79,$9F
	.dc.b	$88,$F8,$7F,$CB,$79,$BF,$8A,$9C,$F9,$1E,$E9,$CB,$F0,$3E,$1F,$F3
	.dc.b	$9E,$4F,$E4,$1B,$1F,$13,$E1,$FF,$D0,$5F,$01,$FE,$23,$E6,$F8,$75
	.dc.b	$AD,$7E,$35,$FF,$AB,$39,$9F,$0F,$C3,$CB,$FC,$DF,$C3,$F2,$F3,$3F
	.dc.b	$3B,$39,$BF,$A3,$39,$7F,$CB,$DF,$B8,$7F,$86,$F4,$FF,$3B,$C6,$3E
	.dc.b	$1F,$C7,$E4,$FF,$AB,$6B,$FC,$39,$A7,$F9,$AE,$F6,$F9,$0F,$F5,$BC
	.dc.b	$73,$9E,$1F,$0F,$EC,$F2,$6D,$7F,$A5,$2D,$18,$8E,$9E,$4F,$7F,$87
	.dc.b	$E1,$9F,$9B,$E1,$BE,$2F,$93,$E3,$6C,$66,$FA,$79,$1E,$7C,$B3,$1E
	.dc.b	$5F,$8A,$6C,$F8,$7F,$9F,$E3,$FF,$8E,$63,$15,$E5,$8C,$4E,$A3,$E3
	.dc.b	$FC,$3F,$D6,$F2,$78,$7C,$3F,$D8,$F2,$6D,$7B,$5A,$FC,$77,$B7,$C3
	.dc.b	$FD,$EF,$2B,$C1,$F0,$DF,$80,$F2,$6D,$FC,$56,$BF,$3D,$F3,$7C,$3F
	.dc.b	$E1,$BC,$AF,$47,$C3,$7E,$23,$C7,$B7,$F3,$7C,$7F,$67,$C3,$FE,$31
	.dc.b	$CC,$7D,$99,$F8,$1F,$0F,$F9,$0F,$1A,$CF,$D5,$F2,$3E,$27,$C3,$FE
	.dc.b	$51,$CC,$7E,$06,$7E,$47,$C3,$7E,$63,$C6,$F9,$5F,$0F,$6B,$F1,$B3
	.dc.b	$5F,$A5,$D4,$9F,$CF,$F8,$FF,$47,$E1,$FF,$DC,$5F,$53,$FE,$1F,$F0
	.dc.b	$58,$7E,$33,$E1,$73,$E5,$7E,$4E,$F7,$3F,$95,$9C,$D7,$E7,$3E,$23
	.dc.b	$9F,$95,$F9,$96,$C7,$E8,$3E,$23,$BF,$93,$9C,$3F,$DC,$F9,$FF,$9C
	.dc.b	$EB,$FD,$85,$FB,$8F,$F8,$7E,$7F,$E0,$0F,$C3,$AB,$FC,$39,$1F,$8D
	.dc.b	$9D,$3E,$89,$D7,$94,$BE,$7F,$13,$EE,$F2,$B1,$7C,$67,$E0,$D6,$3D
	.dc.b	$3C,$F9,$6C,$9F,$13,$F1,$BC,$BD,$61,$F1,$AB,$53,$98,$F4,$FC,$47
	.dc.b	$CC,$F3,$37,$9A,$F6,$FC,$0C,$33,$1D,$FE,$23,$E9,$F9,$98,$CD,$F1
	.dc.b	$18,$CF,$E2,$D7,$D8,$F3,$35,$39,$CB,$E2,$7C,$37,$DC,$F3,$39,$7C
	.dc.b	$37,$DE,$F3,$3A,$4E,$72,$F9,$F3,$97,$E8,$FC,$37,$E1,$3C,$CF,$17
	.dc.b	$C3,$7E,$1B,$CD,$B9,$FD,$89,$F3,$56,$6F,$EC,$FC,$37,$E3,$3C,$CF
	.dc.b	$53,$63,$37,$F8,$0F,$33,$DD,$F0,$D9,$8F,$B9,$F0,$DF,$94,$F2,$FE
	.dc.b	$13,$6F,$F0,$AF,$E6,$3C,$AF,$8D,$F0,$D9,$BF,$C2,$7C,$2F,$E7,$3C
	.dc.b	$AF,$94,$FE,$7B,$CA,$F9,$9F,$0D,$9C,$BF,$1F,$39,$7F,$23,$F0,$BF
	.dc.b	$A4,$F2,$FE,$97,$C3,$59,$CE,$5F,$98,$F3,$32,$FE,$6A,$BE,$C7,$C3
	.dc.b	$7E,$AA,$21,$FE,$DD,$FA,$AF,$F1,$2D,$FE,$27,$C1,$FF,$16,$CE,$8D
	.dc.b	$79,$7C,$9C,$79,$8B,$E3,$1F,$8B,$AE,$2F,$69,$D5,$7E,$27,$E1,$BA
	.dc.b	$4F,$87,$CA,$C3,$D3,$0F,$F1,$86,$3F,$B0,$AD,$E1,$DE,$FF,$2B,$EA
	.dc.b	$7F,$8D,$EF,$78,$7F,$3D,$87,$C2,$F0,$97,$8F,$C3,$E1,$70,$DB,$F0
	.dc.b	$78,$FD,$2F,$3A,$7F,$10,$F3,$BF,$63,$97,$FC,$01,$BF,$66,$5F,$E3
	.dc.b	$2A,$D1,$FB,$19,$FF,$7E,$4F,$F0,$69,$78,$BE,$8F,$85,$F2,$78,$FE
	.dc.b	$AF,$89,$F3,$4F,$B3,$E1,$BD,$1E,$35,$C7,$E2,$79,$3F,$03,$E1,$B3
	.dc.b	$AF,$82,$7F,$75,$AF,$F6,$A7,$71,$9F,$3F,$FC,$0B,$F0,$DF,$0F,$C9
	.dc.b	$F9,$3F,$0D,$F1,$7C,$9F,$97,$39,$BE,$18,$AB,$51,$8F,$86,$F9,$9E
	.dc.b	$4F,$D2,$F8,$6F,$9D,$E4,$FD,$49,$CD,$72,$9C,$BD,$3E,$1B,$EA,$78
	.dc.b	$FF,$93,$E9,$F9,$BF,$0D,$F5,$FC,$9C,$5F,$CB,$8C,$71,$71,$3F,$0D
	.dc.b	$7A,$7E,$0F,$C3,$7D,$FF,$2B,$1C,$9F,$E2,$48,$C4,$FB,$1A,$FE,$1F
	.dc.b	$E1,$BF,$0D,$E5,$F0,$E3,$7C,$2F,$E2,$7C,$AC,$7C,$35,$C3,$E2,$B5
	.dc.b	$FC,$84,$FE,$3F,$E2,$39,$7C,$37,$E4,$7C,$BE,$9F,$0F,$6F,$E6,$CD
	.dc.b	$7F,$31,$1F,$9E,$FA,$A7,$C5,$F8,$F4,$C7,$93,$6F,$F6,$34,$F3,$3D
	.dc.b	$1F,$11,$C5,$D4,$FE,$C6,$7F,$46,$FE,$0F,$F7,$2F,$D8,$AF,$41,$3F
	.dc.b	$9D,$EB,$FE,$9F,$5F,$E3,$EA,$FE,$BF,$2F,$33,$E6,$CF,$2B,$E3,$18
	.dc.b	$F8,$9F,$57,$99,$AF,$89,$F6,$79,$9F,$D3,$30,$7C,$4E,$3C,$0F,$23
	.dc.b	$F0,$7C,$BD,$5E,$75,$FF,$1C,$5B,$E1,$FE,$37,$95,$CF,$E2,$2E,$5E
	.dc.b	$D7,$E5,$79,$7D,$EB,$38,$72,$F8,$6F,$9D,$E5,$78,$DB,$87,$C5,$7E
	.dc.b	$97,$95,$E7,$F8,$7C,$DF,$8B,$E1,$FE,$B7,$99,$EB,$F8,$7B,$8F,$CD
	.dc.b	$9F,$B5,$E6,$66,$FE,$BD,$FC,$1F,$87,$FB,$DE,$55,$9F,$B7,$E5,$7C
	.dc.b	$5F,$87,$CE,$5F,$5E,$FF,$09,$3F,$27,$E1,$BF,$0D,$E4,$DC,$3F,$06
	.dc.b	$FC,$DF,$87,$FC,$57,$97,$99,$FC,$14,$FD,$1F,$87,$FC,$77,$95,$F6
	.dc.b	$9F,$C8,$79,$56,$BF,$1A,$FD,$EF,$87,$DC,$7E,$27,$E1,$FF,$2D,$E5
	.dc.b	$FE,$09,$FC,$C7,$95,$67,$F2,$AF,$E1,$BE,$1F,$71,$F9,$1F,$87,$FE
	.dc.b	$77,$E2,$FF,$52,$F2,$FF,$1F,$18,$F2,$3F,$1A,$DA,$FC,$C7,$C3,$F0
	.dc.b	$7F,$23,$F0,$DC,$BC,$7F,$C9,$FC,$3F,$4F,$2B,$35,$FD,$5B,$3F,$C7
	.dc.b	$BF,$A1,$9E,$0F,$F2,$5F,$DC,$FF,$C6,$EF,$E8,$E7,$F1,$CE,$7A,$FC
	.dc.b	$FE,$37,$F0,$CA,$7C,$3E,$2B,$47,$7F,$27,$5F,$0F,$EE,$F2,$38,$7C
	.dc.b	$4F,$C0,$F2,$33,$7E,$57,$2F,$9E,$7B,$7C,$47,$C6,$F2,$3F,$C6,$79
	.dc.b	$FF,$5A,$AF,$67,$93,$8A,$D6,$DE,$BF,$88,$F9,$B3,$C3,$E1,$FE,$7F
	.dc.b	$8F,$70,$F8,$9E,$45,$7F,$BB,$71,$A7,$BF,$C3,$FD,$6A,$F0,$FC,$3F
	.dc.b	$D8,$F2,$6E,$1F,$37,$C9,$AF,$28,$F3,$FC,$3F,$DE,$9F,$4F,$C3,$FE
	.dc.b	$03,$C8,$F9,$4D,$BF,$A4,$D7,$FD,$6D,$C4,$F8,$7F,$C3,$F9,$1F,$47
	.dc.b	$E1,$FF,$13,$E4,$7D,$3F,$87,$FC,$5F,$91,$71,$FC,$03,$B7,$F0,$53
	.dc.b	$F6,$7E,$1B,$F2,$3E,$3F,$DB,$F8,$7F,$C9,$F9,$16,$FF,$10,$ED,$FC
	.dc.b	$54,$FE,$07,$E1,$FF,$33,$E4,$FD,$4F,$86,$FC,$DF,$91,$F5,$A7,$3D
	.dc.b	$FF,$8D,$9C,$BF,$8F,$F8,$6F,$F0,$DF,$D6,$7F,$8A,$3E,$21,$CB,$FE
	.dc.b	$28,$D3,$FD,$1B,$3F,$8E,$BE,$11,$FE,$30,$8E,$13,$B8,$FF,$47,$46
	.dc.b	$DF,$93,$F0,$FD,$BC,$9F,$C4,$B6,$FF,$C0,$9F,$BA,$7F,$4E,$AE,$9E
	.dc.b	$56,$9E,$D7,$FE,$82,$9F,$E9,$31,$FE,$3B,$C3,$FC,$57,$7C,$BE,$1B
	.dc.b	$D7,$E4,$6F,$E1,$FD,$BE,$4F,$19,$CD,$78,$3E,$13,$E0,$F9,$3D,$5F
	.dc.b	$87,$E3,$F7,$F8,$5C,$D6,$F4,$F8,$13,$E3,$F2,$BE,$4F,$91,$E5,$F8
	.dc.b	$5B,$7E,$C9,$F9,$BE,$46,$70,$F9,$17,$EB,$F8,$4F,$A3,$E3,$D9,$F9
	.dc.b	$9E,$3D,$7D,$D1,$F0,$7E,$13,$EB,$CF,$C3,$F8,$4F,$B3,$E4,$5B,$FA
	.dc.b	$8D,$7E,$37,$91,$F2,$7E,$13,$7B,$7D,$0F,$83,$FC,$0F,$91,$F3,$5F
	.dc.b	$C1,$F9,$17,$0F,$BD,$3F,$47,$E1,$33,$7F,$63,$E1,$3F,$13,$E3,$7F
	.dc.b	$87,$BB,$D1,$8C,$3E,$EF,$C2,$55,$EF,$7F,$B2,$AB,$71,$F8,$BF,$23
	.dc.b	$97,$C1,$FE,$53,$C9,$E9,$F0,$9F,$96,$F8,$9C,$DF,$E4,$5C,$BF,$93
	.dc.b	$9F,$17,$C2,$FE,$73,$CA,$F2,$7C,$2F,$F9,$E3,$EB,$7F,$1D,$7E,$2E
	.dc.b	$1F,$F6,$B6,$AC,$35,$F1,$5C,$2B,$F3,$D8,$FB,$7E,$1F,$97,$95,$F9
	.dc.b	$3F,$87,$B9,$7E,$6A,$7B,$79,$59,$C3,$FD,$2B,$A7,$FA,$0A,$BF,$CC
	.dc.b	$5B,$FF,$85,$E1,$E4,$FF,$B7,$A3,$FC,$3D,$5F,$EF,$79,$CD,$70,$8F
	.dc.b	$57,$93,$BF,$87,$F6,$61,$C7,$E1,$FD,$DE,$4D,$BF,$0F,$95,$D7,$E1
	.dc.b	$F3,$97,$8A,$FE,$24,$F8,$7E,$1F,$E3,$79,$57,$0F,$4B,$5F,$F5,$56
	.dc.b	$BF,$E7,$3C,$3D,$7F,$0D,$C1,$F6,$FC,$2F,$CF,$F2,$7D,$9F,$0D,$F4
	.dc.b	$7C,$8B,$8F,$C4,$9F,$81,$59,$BF,$8B,$F0,$DF,$5F,$C9,$F8,$8D,$8C
	.dc.b	$DF,$CC,$F1,$FE,$47,$C3,$66,$3E,$E0,$F6,$9D,$1F,$D2,$E3,$E9,$F9
	.dc.b	$1F,$3F,$E1,$7F,$08,$D8,$FA,$FE,$4F,$D3,$F8,$6C,$E7,$F6,$3E,$1B
	.dc.b	$F1,$5E,$4F,$D7,$F8,$6B,$3F,$71,$FC,$77,$95,$F6,$EB,$38,$F9,$35
	.dc.b	$F1,$FC,$37,$E4,$FC,$8F,$C0,$3F,$95,$F2,$AD,$7E,$22,$7F,$09,$F0
	.dc.b	$D9,$CF,$F1,$3F,$0B,$F9,$BF,$1F,$F1,$1F,$0D,$67,$F1,$AF,$E7,$FC
	.dc.b	$9F,$C6,$56,$71,$F1,$72,$BE,$17,$FD,$B7,$F5,$A9,$FC,$97,$C3,$E1
	.dc.b	$A7,$FB,$BA,$1E,$57,$E5,$EB,$59,$FE,$73,$E1,$F9,$F9,$1F,$9B,$6C
	.dc.b	$67,$0E,$5E,$4F,$E7,$FE,$1F,$31,$F1,$7F,$01,$FF,$10,$FC,$5B,$4F
	.dc.b	$F3,$76,$1F,$EC,$99,$CC,$62,$B8,$4F,$AB,$C8,$3B,$F8,$AF,$6F,$99
	.dc.b	$8A,$ED,$F1,$1F,$03,$C8,$C7,$B3,$9E,$D6,$8F,$89,$F8,$BE,$6E,$B4
	.dc.b	$F8,$D5,$A9,$E5,$1E,$5C,$76,$F9,$FE,$23,$E6,$F9,$7C,$74,$FC,$A6
	.dc.b	$5F,$E2,$97,$E8,$1F,$47,$87,$E0,$79,$5B,$CF,$C7,$87,$2C,$7E,$17
	.dc.b	$C4,$7D,$8A,$D6,$2F,$88,$FB,$7E,$4F,$FB,$42,$31,$9F,$D9,$F3,$31
	.dc.b	$F1,$1F,$80,$75,$7C,$27,$E7,$7C,$3F,$E1,$3C,$9E,$5F,$11,$F8,$6F
	.dc.b	$33,$A4,$DA,$FA,$93,$B7,$EB,$7C,$47,$E3,$3C,$CF,$17,$C3,$D9,$C9
	.dc.b	$F8,$B8,$CB,$F8,$DF,$2F,$D1,$F0,$FF,$94,$F2,$7D,$AD,$E4,$FC,$17
	.dc.b	$C3,$FE,$63,$CC,$CE,$1F,$95,$BF,$87,$F0,$DF,$9C,$F2,$6C,$FE,$67
	.dc.b	$CB,$F8,$FF,$0F,$9C,$BF,$19,$7F,$A2,$9F,$97,$F0,$FF,$A4,$F2,$EE
	.dc.b	$1F,$A1,$F2,$BE,$7F,$C3,$E7,$1F,$CA,$5F,$EA,$A7,$E9,$FC,$3F,$DC
	.dc.b	$BF,$BB,$E1,$3C,$0F,$C9,$47,$E7,$FD,$0D,$3F,$81,$8C,$F9,$1F,$93
	.dc.b	$F0,$18,$6A,$BF,$DC,$D5,$AC,$FF,$27,$5F,$E3,$19,$CF,$97,$F8,$1F
	.dc.b	$85,$E1,$9F,$87,$C8,$C5,$78,$23,$0F,$F0,$1C,$FF,$45,$FB,$AF,$F6
	.dc.b	$68,$F4,$E6,$BE,$31,$ED,$BF,$67,$95,$F9,$AC,$38,$FC,$37,$83,$0F
	.dc.b	$CF,$46,$2F,$C1,$18,$D7,$F5,$79,$3E,$1B,$87,$95,$CE,$73,$8F,$3B
	.dc.b	$C6,$1F,$CF,$70,$FE,$D2,$F4,$F1,$BF,$51,$18,$8F,$34,$BC,$8C,$FC
	.dc.b	$37,$96,$F7,$F0,$DE,$79,$E3,$F0,$DE,$9F,$22,$E3,$E8,$7A,$FC,$37
	.dc.b	$B7,$C9,$CC,$F6,$9F,$0F,$C3,$7C,$1F,$2A,$CF,$C0,$7C,$BF,$13,$F1
	.dc.b	$7C,$BC,$CF,$9A,$7D,$3F,$0B,$F2,$7C,$AF,$5F,$C3,$7C,$BF,$2E,$DF
	.dc.b	$CA,$9D,$BF,$32,$3E,$0F,$C3,$7D,$1F,$23,$E1,$FC,$37,$D3,$F2,$AD
	.dc.b	$FD,$29,$DB,$F5,$23,$E4,$FC,$37,$D9,$F2,$2C,$FD,$86,$BF,$33,$C9
	.dc.b	$F9,$FF,$0D,$BE,$EF,$C5,$1F,$73,$E1,$BF,$05,$E4,$FD,$8F,$61,$E0
	.dc.b	$27,$F0,$DE,$5E,$61,$F0,$D8,$CF,$99,$F7,$EB,$9B,$8B,$E7,$F0,$BF
	.dc.b	$8F,$F2,$F9,$7C,$37,$E8,$3E,$BF,$FC,$7F,$E5,$7E,$72,$B1,$F1,$3A
	.dc.b	$C7,$F2,$78,$78,$7E,$23,$87,$95,$67,$F4,$D8,$FF,$A4,$31,$FD,$2F
	.dc.b	$C4,$6A,$7C,$5B,$B0,$F0,$79,$BE,$FF,$DD,$BF,$71,$4F,$41,$97,$F8
	.dc.b	$F2,$3F,$C0,$93,$A9,$CF,$97,$FD,$06,$3F,$C8,$73,$C6,$9E,$6E,$63
	.dc.b	$DD,$5B,$F8,$8F,$81,$E6,$59,$F4,$BC,$FE,$23,$E2,$79,$B9,$9F,$DD
	.dc.b	$62,$AF,$07,$C3,$7C,$9F,$33,$C5,$F0,$F7,$BB,$E8,$6C,$AF,$9D,$E6
	.dc.b	$EB,$2F,$8F,$87,$A7,$E1,$FE,$97,$99,$F0,$DF,$A9,$E6,$7C,$5F,$87
	.dc.b	$B7,$D2,$76,$F6,$F8,$7F,$B5,$E6,$7C,$BF,$87,$B3,$93,$EA,$C6,$5F
	.dc.b	$AF,$E6,$FD,$1F,$87,$FC,$17,$99,$67,$ED,$F9,$5F,$57,$E1,$FF,$0C
	.dc.b	$E6,$3E,$1C,$FD,$9F,$87,$FC,$57,$93,$F4,$DF,$C6,$79,$5F,$57,$E1
	.dc.b	$F3,$CF,$F1,$27,$2F,$C6,$F8,$7F,$C9,$79,$5F,$6D,$FC,$A7,$97,$F7
	.dc.b	$7E,$1F,$35,$BD,$3F,$1F,$3F,$81,$F8,$6F,$CD,$78,$FF,$83,$F8,$7F
	.dc.b	$CE,$79,$59,$BF,$CA,$CE,$5F,$CB,$C7,$E2,$7E,$1F,$FA,$0F,$D8,$7F
	.dc.b	$53,$F3,$7F,$C3,$DE,$17,$F9,$06,$FF,$A1,$4F,$F8,$F1,$FB,$B5,$C3
	.dc.b	$E2,$BF,$35,$F1,$3C,$A7,$F3,$9F,$13,$D3,$CB,$B1,$C9,$AF,$F4,$5C
	.dc.b	$11,$FA,$3F,$89,$F1,$F9,$9F,$C0,$EF,$FC,$7F,$5F,$8B,$BF,$DE,$43
	.dc.b	$E2,$3D,$3E,$56,$DB,$3A,$E4,$F6,$F9,$5C,$FE,$23,$DF,$E3,$75,$F8
	.dc.b	$8B,$7D,$1A,$FF,$9C,$F2,$FE,$B8,$FC,$2F,$23,$C9,$F1,$1F,$23,$C9
	.dc.b	$B3,$F1,$BC,$9A,$FB,$5F,$57,$C4,$6E,$3D,$3F,$11,$F4,$3C,$9F,$77
	.dc.b	$C4,$59,$F6,$B5,$F1,$4F,$D6,$F2,$BE,$27,$C3,$FD,$8F,$1B,$E3,$7C
	.dc.b	$45,$BF,$8A,$D7,$D5,$3F,$7B,$C8,$F9,$9F,$0F,$F8,$0F,$23,$E7,$7C
	.dc.b	$47,$E0,$BC,$9B,$87,$E0,$1D,$BF,$82,$9F,$A9,$F1,$1F,$88,$F2,$6C
	.dc.b	$FE,$19,$AF,$DD,$7E,$D7,$C4,$7E,$3B,$CA,$FB,$9F,$11,$F9,$0F,$1F
	.dc.b	$EF,$7C,$45,$C3,$EE,$B5,$FA,$53,$F9,$6F,$1F,$F0,$9F,$11,$F9,$8E
	.dc.b	$DF,$C3,$7C,$47,$E6,$BC,$8B,$87,$E6,$1D,$BF,$9A,$9F,$C6,$7C,$47
	.dc.b	$F8,$83,$EC,$3F,$C4,$D3,$FE,$C3,$9F,$F1,$37,$C1,$27,$F1,$D3,$F9
	.dc.b	$0F,$89,$E1,$E4,$F2,$8F,$F1,$34,$F2,$F8,$96,$0F,$88,$ED,$E4,$FF
	.dc.b	$5D,$FD,$DF,$FC,$51,$87,$F9,$EF,$0F,$C6,$BF,$EC,$18,$CB,$FD,$0E
	.dc.b	$33,$E9,$FF,$A4,$70,$E9,$3E,$0F,$2B,$A7,$C4,$FA,$FC,$AE,$DF,$11
	.dc.b	$6B,$FB,$16,$5F,$E2,$97,$CF,$3C,$1F,$4F,$93,$E5,$F8,$7F,$89,$E4
	.dc.b	$59,$F6,$B5,$FF,$1E,$6B,$FE,$88,$7D,$1F,$11,$F2,$FC,$AF,$77,$C3
	.dc.b	$7C,$DF,$22,$D7,$C6,$6B,$C9,$F8,$9F,$11,$F4,$FC,$AF,$8D,$F0,$DF
	.dc.b	$57,$CA,$B7,$F3,$BC,$AA,$F8,$9F,$99,$F0,$FB,$8F,$7F,$C3,$FD,$DF
	.dc.b	$2B,$E8,$36,$32,$7D,$88,$FA,$9F,$0F,$F8,$3F,$2B,$EB,$7C,$3F,$E1
	.dc.b	$7C,$8B,$7F,$7B,$C9,$FB,$5F,$0F,$F8,$97,$31,$F1,$27,$EF,$7C,$37
	.dc.b	$E3,$7C,$AB,$3F,$86,$7F,$05,$F0,$F9,$8F,$AB,$3F,$93,$F2,$FF,$0D
	.dc.b	$F0,$FF,$95,$F2,$ED,$7E,$3B,$CA,$FC,$57,$C3,$FE,$65,$CC,$7D,$49
	.dc.b	$FC,$77,$C3,$FE,$77,$CA,$FC,$87,$C3,$FF,$9F,$BD,$DF,$F0,$FF,$BF
	.dc.b	$FE,$D6,$F1,$D7,$9F,$8A,$DE,$5F,$9C,$C7,$F3,$1F,$11,$C7,$CB,$B3
	.dc.b	$FE,$85,$CB,$FC,$41,$97,$3F,$88,$D4,$FE,$42,$FC,$1E,$67,$F8,$F7
	.dc.b	$F7,$8F,$F0,$D6,$1F,$E7,$1F,$43,$FD,$D9,$1F,$E1,$AF,$51,$3F,$E1
	.dc.b	$AC,$3F,$C7,$53,$C3,$CA,$E5,$F1,$1E,$BF,$3B,$A7,$C4,$7B,$7C,$CE
	.dc.b	$DF,$11,$71,$E3,$35,$69,$E0,$9F,$89,$E6,$79,$7E,$1B,$E3,$79,$96
	.dc.b	$FD,$CF,$A7,$E1,$F3,$1E,$49,$F9,$9E,$5F,$B7,$83,$E7,$79,$7E,$FF
	.dc.b	$87,$B7,$EA,$9A,$F8,$B9,$CB,$D9,$72,$2F,$EC,$79,$9F,$27,$E1,$FE
	.dc.b	$D7,$99,$70,$FA,$4F,$CD,$F8,$7C,$C7,$C8,$9F,$C0,$79,$BF,$47,$E1
	.dc.b	$FF,$05,$E5,$7D,$3F,$87,$B7,$F3,$A6,$BE,$53,$C9,$A2,$7F,$15,$E6
	.dc.b	$7D,$BF,$87,$FC,$67,$99,$6F,$F0,$8F,$DF,$F8,$7C,$C7,$DA,$9F,$C9
	.dc.b	$79,$BF,$83,$F8,$6F,$CA,$79,$5F,$85,$F8,$7F,$CB,$79,$B9,$C3,$F1
	.dc.b	$CE,$5F,$C8,$4F,$E2,$FE,$1F,$F3,$9E,$4D,$9F,$CA,$3F,$8F,$F8,$7C
	.dc.b	$C7,$E2,$A7,$F4,$5E,$5F,$E4,$FE,$1F,$FD,$C3,$F6,$5F,$E2,$1E,$DF
	.dc.b	$A7,$DB,$F4,$A7,$38,$7E,$67,$E2,$2B,$F9,$48,$E3,$E5,$FE,$77,$E1
	.dc.b	$79,$F9,$5F,$9F,$F8,$8E,$BE,$56,$70,$FF,$76,$68,$8D,$46,$2B,$FD
	.dc.b	$45,$FB,$CB,$5F,$2F,$91,$8F,$89,$F3,$F9,$5A,$9B,$87,$FA,$7E,$FF
	.dc.b	$AB,$BF,$EF,$1C,$BF,$C5,$0F,$DB,$E0,$4F,$F5,$08,$FC,$89,$FE,$91
	.dc.b	$9F,$D8,$BC,$FE,$27,$E2,$F9,$3E,$4F,$89,$B3,$93,$D7,$E5,$7A,$3E
	.dc.b	$27,$E5,$F9,$1E,$AF,$89,$B5,$E5,$9A,$F8,$F3,$4F,$D2,$F2,$BE,$0F
	.dc.b	$C4,$FD,$4F,$26,$DF,$C8,$F2,$3E,$2F,$C4,$7D,$87,$31,$EF,$9F,$93
	.dc.b	$F1,$1F,$73,$C7,$B3,$F4,$3C,$8F,$9B,$F1,$1F,$80,$73,$1F,$1E,$7E
	.dc.b	$8F,$C4,$7E,$13,$C9,$B3,$F6,$3C,$9F,$AB,$F1,$1F,$88,$73,$1F,$3E
	.dc.b	$7E,$CF,$C4,$7E,$33,$C7,$B3,$F8,$07,$EE,$FC,$46,$63,$EC,$4F,$E4
	.dc.b	$BC,$9F,$C0,$FC,$3F,$E5,$3C,$8B,$5F,$88,$7F,$0B,$F1,$19,$8F,$C0
	.dc.b	$4F,$E6,$BC,$AF,$C4,$FC,$47,$E7,$3C,$AB,$5F,$90,$F2,$BF,$1B,$F1
	.dc.b	$1F,$A0,$73,$1F,$87,$9F,$C8,$FC,$3F,$E1,$23,$F0,$BF,$04,$FB,$2F
	.dc.b	$F0,$F7,$81,$FE,$26,$6D,$6E,$BF,$8D,$D6,$FC,$9F,$CD,$FC,$4E,$67
	.dc.b	$FD,$BD,$7F,$96,$F8,$1F,$CF,$47,$F2,$1C,$7F,$C3,$AF,$87,$0F,$07
	.dc.b	$93,$FA,$4F,$87,$E5,$18,$9F,$F7,$F7,$A7,$FE,$28,$8C,$47,$E3,$67
	.dc.b	$78,$BC,$AF,$CF,$E5,$E0,$F3,$58,$BE,$1F,$7E,$56,$B7,$FE,$07,$3F
	.dc.b	$E2,$9C,$FC,$99,$2B,$51,$FC,$E6,$79,$3F,$0F,$4F,$91,$3A,$C7,$BF
	.dc.b	$C3,$F9,$3C,$BC,$60,$F8,$6F,$3F,$97,$8F,$87,$F4,$F9,$9A,$9B,$A7
	.dc.b	$C0,$9D,$BF,$0B,$E1,$FD,$FE,$5F,$4F,$87,$B3,$93,$C9,$C0,$9D,$46
	.dc.b	$3C,$CF,$1F,$C3,$FC,$8F,$2F,$CB,$F0,$F6,$BE,$7C,$D7,$F8,$66,$E9
	.dc.b	$F9,$FE,$67,$B3,$E1,$FE,$8F,$93,$6F,$C3,$E5,$7C,$0F,$87,$FA,$AE
	.dc.b	$63,$E6,$CF,$C4,$F8,$7F,$B3,$E4,$D9,$F4,$BF,$23,$E1,$F3,$1F,$82
	.dc.b	$9F,$BF,$E5,$FC,$CF,$87,$FC,$0F,$97,$6B,$E0,$F9,$7F,$43,$E1,$BF
	.dc.b	$0A,$E6,$3E,$EC,$FD,$4F,$86,$FC,$4F,$97,$F5,$BE,$1B,$F1,$7E,$66
	.dc.b	$6F,$E5,$B9,$7E,$6C,$FD,$CF,$85,$FC,$8F,$97,$F7,$BE,$1A,$D7,$E6
	.dc.b	$26,$BF,$2A,$3F,$3B,$F9,$07,$F8,$DA,$3F,$19,$F0,$F8,$F3,$3F,$0F
	.dc.b	$D7,$67,$F4,$33,$5F,$A7,$1C,$BC,$8F,$C8,$7E,$F5,$FB,$19,$79,$8C
	.dc.b	$7F,$C6,$5E,$6F,$F6,$0A,$D4,$E7,$C7,$E3,$F1,$1E,$47,$9E,$1E,$6F
	.dc.b	$27,$AF,$C3,$FA,$3C,$AE,$FF,$11,$6B,$FC,$3B,$E8,$FE,$BF,$3F,$F7
	.dc.b	$FC,$7E,$97,$55,$F0,$8C,$79,$9E,$8F,$88,$F8,$7E,$4D,$9E,$1E,$4F
	.dc.b	$B3,$E1,$FE,$3B,$98,$FE,$11,$8A,$BE,$0F,$C3,$7C,$CF,$2A,$D7,$85
	.dc.b	$F8,$BF,$0D,$98,$F3,$F2,$FC,$3A,$FA,$7E,$5F,$CA,$F8,$6F,$AB,$E4
	.dc.b	$5B,$F5,$79,$3F,$3B,$E1,$BE,$CB,$98,$F1,$CF,$D2,$F8,$5F,$BB,$E5
	.dc.b	$7D,$4F,$86,$FB,$FE,$5E,$6F,$E2,$39,$7E,$34,$FD,$AF,$85,$FC,$2F
	.dc.b	$95,$F7,$3E,$1B,$F0,$FE,$5E,$6F,$E6,$39,$7E,$74,$FE,$0B,$E1,$7F
	.dc.b	$1B,$E4,$FE,$13,$E1,$AD,$7C,$03,$FC,$A5,$DC,$AC,$C7,$F8,$26,$13
	.dc.b	$F9,$6F,$2B,$F1,$BF,$0D,$F9,$8F,$2A,$DF,$DD,$7F,$23,$F0,$D9,$8E
	.dc.b	$DE,$F3,$15,$FE,$83,$CC,$FC,$BF,$C3,$7F,$3E,$F1,$3F,$C6,$98,$79
	.dc.b	$23,$1E,$5F,$E7,$3E,$1A,$CF,$4C,$55,$C6,$3F,$43,$F0,$BC,$FC,$BB
	.dc.b	$1F,$CD,$71,$FF,$19,$3C,$7E,$1B,$C1,$E6,$6A,$3F,$A9,$4F,$E6,$1F
	.dc.b	$F1,$57,$EF,$7F,$D6,$1C,$F9,$7F,$CF,$63,$11,$FD,$A2,$7F,$C6,$3E
	.dc.b	$6F,$F4,$F9,$FF,$1A,$CE,$6B,$87,$8D,$EF,$F3,$39,$7C,$37,$C1,$F2
	.dc.b	$2D,$F9,$DE,$DF,$0D,$98,$F9,$78,$FF,$8E,$DF,$EE,$39,$FF,$A2,$5C
	.dc.b	$FC,$37,$CB,$F2,$7C,$CD,$8E,$3F,$0D,$5F,$6C,$7D,$1F,$2B,$D9,$F0
	.dc.b	$DF,$4F,$C9,$B5,$F2,$5F,$81,$F0,$D9,$8F,$1C,$FD,$9F,$2B,$E2,$7C
	.dc.b	$2F,$DB,$F2,$7E,$37,$C2,$FD,$DF,$2B,$38,$7D,$37,$2F,$D5,$9F,$99
	.dc.b	$F0,$BF,$83,$F2,$2C,$FD,$9F,$23,$E8,$7C,$2F,$E1,$DC,$C7,$D8,$9F
	.dc.b	$A9,$F0,$9F,$8B,$F1,$BE,$B7,$C2,$DA,$F9,$33,$5F,$BB,$1F,$91,$F2
	.dc.b	$7E,$E7,$C2,$FE,$4F,$C5,$B5,$F8,$97,$F0,$1F,$0B,$98,$FA,$73,$F9
	.dc.b	$9F,$23,$F0,$9F,$09,$F9,$BF,$23,$F0,$DF,$0B,$6F,$EC,$CD,$7F,$1B
	.dc.b	$1F,$A1,$F2,$3F,$19,$F0,$9F,$E2,$0F,$C8,$FF,$21,$E4,$FE,$2E,$31
	.dc.b	$F1,$5F,$91,$F8,$7B,$8F,$E1,$27,$87,$95,$F9,$5F,$87,$E5,$E5,$7F
	.dc.b	$84,$FF,$7C,$FF,$14,$66,$8D,$61,$FE,$19,$8C,$3F,$B8,$9D,$F9,$5C
	.dc.b	$7E,$23,$C9,$E3,$F3,$F8,$8F,$37,$93,$9C,$39,$39,$7A,$4F,$87,$E1
	.dc.b	$BD,$9E,$3F,$8F,$E1,$FD,$DE,$4E,$6F,$C4,$E5,$F2,$4F,$A7,$E1,$FE
	.dc.b	$27,$93,$EB,$F3,$AD,$79,$E6,$BF,$EA,$1D,$D3,$F2,$FC,$AF,$85,$F0
	.dc.b	$BF,$37,$C9,$B7,$F0,$5F,$8D,$F0,$F9,$8F,$85,$3F,$4F,$CA,$F9,$5F
	.dc.b	$0D,$F5,$7C,$8F,$F2,$DF,$77,$F9,$37,$44,$E7,$AF,$E5,$7C,$2F,$DB
	.dc.b	$F2,$77,$1F,$ED,$0A,$DE,$7F,$47,$1C,$AF,$71,$2F,$84,$FC,$27,$91
	.dc.b	$A9,$CF,$C2,$FE,$1F,$CA,$C5,$EF,$0F,$B9,$78,$AD,$65,$F7,$7E,$13
	.dc.b	$F1,$DE,$4F,$07,$F2,$1E,$4F,$2F,$85,$CD,$6F,$1F,$C3,$4F,$6F,$84
	.dc.b	$FC,$B7,$91,$67,$F1,$4F,$8B,$E1,$7F,$35,$E5,$66,$7F,$CF,$DE,$42
	.dc.b	$BC,$FF,$0B,$F9,$FF,$2E,$D7,$F9,$AF,$F2,$57,$C7,$7B,$3E,$1F,$5B
	.dc.b	$7E,$57,$0F,$F4,$2D,$6A,$3F,$31,$F0,$F8,$F3,$7E,$1F,$C3,$D9,$CE
	.dc.b	$5D,$7C,$EC,$BD,$EB,$E6,$69,$FE,$24,$C7,$FC,$ED,$7F,$D6,$A3,$1E
	.dc.b	$5E,$BE,$1F,$C3,$E6,$F0,$C7,$C7,$3C,$BE,$1B,$CB,$E5,$F4,$F8,$6B
	.dc.b	$97,$39,$F4,$F9,$99,$C3,$E0,$5F,$8B,$E1,$BD,$BE,$55,$9F,$89,$E5
	.dc.b	$F9,$BE,$1B,$39,$79,$6F,$E1,$CF,$AB,$E1,$BE,$2F,$97,$EC,$F8,$8B
	.dc.b	$7E,$B9,$F9,$3E,$66,$72,$FA,$17,$F0,$BE,$1B,$E6,$F9,$76,$7E,$A3
	.dc.b	$F1,$BE,$1B,$E8,$F9,$19,$9F,$F5,$D7,$47,$E5,$F3,$F9,$1F,$0D,$F5
	.dc.b	$BC,$8F,$9A,$D8,$F9,$9F,$0D,$F6,$BC,$9C,$E1,$F8,$1B,$FA,$7F,$0B
	.dc.b	$F7,$BC,$9B,$3F,$85,$F2,$7E,$BF,$C3,$67,$2F,$AD,$7F,$84,$9F,$B7
	.dc.b	$F0,$BF,$86,$F1,$AE,$1F,$8D,$7E,$FF,$C3,$7E,$2B,$C8,$CC,$FA,$A7
	.dc.b	$F0,$7F,$0D,$F8,$EF,$26,$CF,$E5,$7C,$9F,$C3,$FC,$36,$75,$FC,$35
	.dc.b	$FE,$52,$7F,$17,$F0,$DF,$ED,$AF,$C9,$98,$60,$FF,$64,$38,$BF,$C7
	.dc.b	$7C,$3E,$FC,$AD,$57,$FB,$96,$B4,$BD,$AF,$F6,$E7,$98,$F8,$7E,$DE
	.dc.b	$57,$E6,$9F,$07,$97,$F9,$CF,$87,$CE,$9F,$9B,$9C,$BF,$E3,$6F,$DF
	.dc.b	$7F,$C3,$79,$7F,$BF,$33,$79,$3F,$C3,$E7,$33,$FE,$8C,$9C,$D7,$1F
	.dc.b	$88,$F7,$79,$7B,$8F,$81,$87,$1F,$88,$F8,$5E,$57,$3F,$88,$F8,$9E
	.dc.b	$66,$73,$F8,$73,$97,$E2,$C7,$87,$E1,$FE,$57,$97,$E3,$F8,$8B,$9F
	.dc.b	$A6,$7E,$77,$97,$9C,$3E,$6D,$FA,$7E,$1F,$E9,$79,$56,$7E,$8B,$ED
	.dc.b	$F8,$8F,$AD,$E5,$E6,$7E,$84,$FC,$1F,$88,$FB,$5E,$45,$9F,$B2,$FC
	.dc.b	$5F,$88,$FB,$DE,$56,$67,$EC,$4F,$C9,$F8,$8F,$C1,$79,$56,$7F,$03
	.dc.b	$E5,$7C,$DF,$88,$CF,$17,$D3,$BF,$C4,$4F,$D1,$F8,$7F,$C5,$79,$17
	.dc.b	$0F,$C4,$BF,$57,$E2,$3F,$1D,$E5,$66,$7F,$11,$3F,$67,$E1,$7F,$25
	.dc.b	$E5,$59,$FC,$8B,$F7,$7E,$1F,$F2,$DE,$5E,$67,$F2,$13,$F8,$1F,$86
	.dc.b	$FC,$D7,$8D,$F8,$3F,$86,$BC,$1F,$89,$9F,$CF,$79,$19,$C3,$F3,$B7
	.dc.b	$F8,$9F,$84,$FD,$0C,$43,$FD,$AB,$D8,$C3,$F1,$DF,$0D,$99,$FD,$3F
	.dc.b	$C6,$29,$E4,$F3,$9D,$4E,$6F,$8C,$62,$B7,$1F,$9D,$F3,$BA,$4F,$83
	.dc.b	$C8,$FC,$E4,$62,$FB,$79,$1F,$DB,$7F,$7E,$FE,$A3,$7C,$EF,$6F,$59
	.dc.b	$D7,$C2,$F1,$C9,$E3,$E2,$BC,$4F,$3F,$86,$D7,$93,$99,$F8,$59,$FB
	.dc.b	$6B,$C1,$F0,$DC,$F0,$B5,$EB,$F2,$3C,$9F,$0D,$9E,$0F,$F5,$6E,$EC
	.dc.b	$3C,$55,$E9,$F8,$5F,$27,$91,$71,$F8,$8F,$B7,$E1,$BD,$1E,$4E,$67
	.dc.b	$C7,$A2,$BE,$17,$C3,$7B,$7C,$6B,$5F,$35,$F8,$DF,$0D,$F0,$7C,$8C
	.dc.b	$CF,$29,$F9,$5F,$0B,$F1,$7C,$8F,$99,$F0,$DF,$1F,$C9,$CD,$FD,$79
	.dc.b	$CB,$F6,$63,$E9,$7C,$37,$CD,$F2,$3E,$A7,$C3,$7C,$FF,$27,$37,$F7
	.dc.b	$E7,$2F,$E0,$63,$ED,$7C,$27,$D5,$F1,$EC,$FE,$17,$C8,$FB,$DF,$0B
	.dc.b	$9E,$BF,$97,$7F,$6E,$7F,$05,$F0,$BF,$77,$C8,$B8,$7E,$35,$FC,$37
	.dc.b	$C2,$FE,$07,$C9,$CC,$FC,$A9,$FC,$57,$C2,$7E,$17,$C9,$B3,$F9,$8F
	.dc.b	$CA,$7F,$C3,$0F,$F3,$BD,$3F,$A2,$FC,$5B,$C3,$7C,$3F,$19,$FB,$F8
	.dc.b	$FE,$5F,$0E,$3F,$09,$FA,$9C,$7F,$C9,$18,$3E,$1F,$5E,$67,$E6,$FE
	.dc.b	$1E,$E5,$F8,$89,$E5,$E6,$E7,$0F,$25,$FF,$93,$7F,$7F,$FD,$3C,$3C
	.dc.b	$D6,$9F,$E2,$48,$F4,$1F,$E5,$79,$7C,$47,$97,$CC,$E5,$3E,$7F,$37
	.dc.b	$A4,$E7,$4F,$EC,$7B,$FA,$78,$98,$FF,$BC,$70,$FE,$0F,$9F,$83,$CC
	.dc.b	$F2,$7C,$3F,$C1,$7C,$DF,$11,$F0,$FC,$EC,$DF,$CD,$9F,$57,$C2,$FC
	.dc.b	$7F,$33,$D9,$F0,$DF,$27,$CD,$CD,$FD,$39,$CB,$F5,$63,$E1,$7C,$37
	.dc.b	$CF,$F3,$2C,$FD,$97,$E3,$7C,$37,$D3,$F3,$73,$3F,$BC,$E7,$FF,$41
	.dc.b	$F2,$7B,$8F,$BB,$C4,$F8,$5F,$B5,$E6,$7C,$FF,$86,$B5,$BC,$3E,$65
	.dc.b	$7D,$FF,$37,$51,$F8,$9B,$FA,$DF,$0B,$F8,$3F,$2B,$EC,$7C,$37,$E1
	.dc.b	$7C,$CC,$DF,$E3,$E7,$2F,$E4,$63,$EF,$7C,$2F,$E2,$FC,$CF,$C0,$7C
	.dc.b	$37,$E3,$7C,$DC,$DF,$E5,$E7,$2F,$E6,$63,$F0,$DF,$0D,$F9,$3F,$2B
	.dc.b	$F1,$1F,$0D,$F9,$5F,$33,$37,$F9,$F9,$CB,$FA,$18,$FC,$77,$C2,$FE
	.dc.b	$6F,$CB,$B3,$FA,$57,$F2,$5F,$0D,$F9,$FF,$37,$33,$F3,$67,$F2,$DF
	.dc.b	$0B,$FC,$E7,$F2,$BF,$F1,$2F,$95,$F8,$BE,$B4,$E7,$CB,$FC,$E7,$C3
	.dc.b	$EF,$1E,$27,$F4,$AC,$3F,$33,$D9,$F4,$7C,$CF,$C9,$FC,$37,$5F,$2F
	.dc.b	$F4,$8F,$7F,$33,$F4,$DF,$0F,$9B,$E1,$AF,$F8,$CF,$2E,$95,$FE,$B7
	.dc.b	$AF,$F1,$4F,$9C,$CF,$FC,$53,$5F,$AA,$CD,$5A,$97,$97,$C7,$E1,$FD
	.dc.b	$DE,$6F,$3D,$3E,$04,$F5,$F8,$7F,$85,$E6,$77,$F8,$7F,$89,$E6,$67
	.dc.b	$5F,$74,$E5,$F8,$11,$E5,$F8,$6F,$95,$E5,$79,$FE,$1E,$E7,$F0,$F7
	.dc.b	$FE,$BB,$87,$C7,$F3,$34,$FC,$9B,$F7,$7C,$2F,$D3,$F2,$EC,$FC,$D7
	.dc.b	$E1,$7C,$3F,$D7,$F3,$33,$3F,$D8,$B9,$15,$F1,$FE,$1B,$EE,$79,$76
	.dc.b	$BE,$B7,$97,$F2,$FE,$1F,$3B,$F9,$6F,$F0,$53,$F3,$FE,$1B,$F0,$9E
	.dc.b	$5F,$D1,$F8,$7B,$7E,$B9,$FC,$47,$99,$9C,$BF,$05,$7F,$5F,$E1,$FF
	.dc.b	$19,$E5,$D9,$FC,$37,$97,$F6,$FE,$1F,$39,$7C,$5B,$FC,$94,$FD,$FF
	.dc.b	$87,$FC,$A7,$95,$70,$FC,$73,$F8,$3F,$87,$FC,$C7,$95,$99,$F8,$53
	.dc.b	$F8,$7F,$87,$FC,$E7,$95,$67,$F2,$DD,$BF,$8B,$F8,$7C,$EB,$F5,$6F
	.dc.b	$F4,$53,$F8,$FF,$86,$FF,$0E,$7E,$59,$FE,$2D,$AF,$B5,$F1,$18,$F2
	.dc.b	$B5,$97,$F8,$AA,$B4,$FE,$4F,$05,$6E,$FF,$05,$F1,$1D,$3C,$9F,$EA
	.dc.b	$1E,$97,$F3,$9F,$E0,$0F,$F1,$3E,$3F,$E1,$CC,$9E,$5F,$E8,$70,$E1
	.dc.b	$8F,$EB,$6F,$77,$FC,$FA,$F7,$39,$F8,$7F,$4F,$95,$DA,$3F,$B2,$DA
	.dc.b	$7D,$BE,$56,$70,$79,$EA,$7C,$47,$2C,$3A,$57,$BE,$F8,$5E,$FE,$23
	.dc.b	$E3,$F9,$78,$C5,$F1,$1F,$2B,$CC,$C5,$6B,$2F,$54,$E5,$F6,$7C,$3F
	.dc.b	$D0,$F2,$39,$36,$3E,$07,$C4,$7D,$4F,$2B,$3B,$7D,$1B,$F0,$7C,$47
	.dc.b	$D8,$F2,$7C,$5F,$11,$71,$F9,$0F,$DC,$F2,$F7,$87,$D9,$AF,$47,$C4
	.dc.b	$7E,$03,$C8,$F5,$7C,$47,$E0,$BC,$AB,$7F,$7E,$76,$FE,$06,$3E,$07
	.dc.b	$C3,$7E,$23,$CA,$B3,$F8,$56,$BF,$77,$CA,$F8,$DF,$0F,$BD,$BE,$E7
	.dc.b	$C3,$FE,$43,$C8,$F9,$5F,$0F,$67,$F0,$0F,$E5,$3C,$AF,$9D,$59,$C7
	.dc.b	$F9,$17,$5B,$E1,$FF,$33,$E5,$FD,$36,$E3,$F8,$97,$F3,$BE,$4F,$D7
	.dc.b	$F8,$7C,$E1,$CB,$E1,$FF,$43,$E4,$7D,$BF,$87,$FF,$3F,$7B,$BF,$E2
	.dc.b	$8C,$BF,$17,$1A,$7F,$29,$5A,$F8,$DF,$C1,$7C,$3F,$0F,$2A,$E5,$FE
	.dc.b	$90,$C7,$F2,$58,$7E,$6E,$BA,$F9,$5F,$89,$F8,$8E,$13,$FD,$2F,$F8
	.dc.b	$07,$FC,$57,$F0,$AC,$3F,$C5,$5E,$8F,$F8,$86,$73,$18,$AF,$F2,$1C
	.dc.b	$63,$C9,$FC,$BB,$FE,$2E,$CF,$C5,$E5,$63,$E2,$3D,$BE,$5D,$9F,$33
	.dc.b	$5F,$F4,$7E,$2F,$2F,$9F,$C3,$F0,$CF,$CD,$F0,$FF,$1B,$CB,$EF,$F0
	.dc.b	$FF,$23,$CC,$B9,$7C,$19,$F1,$D6,$6F,$DD,$F0,$FF,$3B,$CB,$F3,$BF
	.dc.b	$43,$CB,$B8,$7C,$99,$F5,$EB,$9B,$F8,$DF,$0D,$F5,$BC,$AF,$7B,$F6
	.dc.b	$3C,$8B,$87,$D1,$9F,$87,$F0,$F9,$BF,$9D,$F0,$BF,$7B,$C7,$F8,$FF
	.dc.b	$0F,$67,$E9,$3F,$82,$F2,$BE,$5D,$67,$1F,$EE,$5D,$2F,$86,$FC,$3F
	.dc.b	$91,$F4,$3E,$1B,$F1,$3E,$45,$BF,$C1,$4F,$D4,$AC,$E7,$F7,$FE,$17
	.dc.b	$F1,$FE,$3F,$D8,$F8,$6F,$C8,$F9,$37,$1F,$C5,$4F,$DC,$AC,$DF,$E1
	.dc.b	$FE,$1B,$F2,$FE,$47,$E0,$3E,$1A,$CF,$E2,$DF,$CD,$F9,$1F,$84,$AC
	.dc.b	$E3,$EA,$F8,$5F,$F6,$E7,$E5,$FF,$E2,$4C,$3F,$17,$F0,$F8,$F8,$AF
	.dc.b	$C5,$BA,$BF,$C6,$DE,$6B,$F3,$1F,$11,$CB,$C9,$FC,$8F,$C3,$F4,$F2
	.dc.b	$2E,$1F,$EF,$2C,$D5,$F9,$6B,$D6,$1F,$E2,$BF,$C0,$FF,$C4,$35,$FE
	.dc.b	$20,$F3,$BF,$D0,$11,$8F,$1F,$F9,$9B,$AF,$2A,$FC,$4F,$AA,$73,$F1
	.dc.b	$3E,$CF,$2A,$EB,$E1,$6B,$FE,$EC,$CB,$F5,$D9,$F8,$7E,$27,$83,$E3
	.dc.b	$F8,$8F,$8B,$E4,$F8,$1B,$19,$BF,$67,$8F,$E4,$F8,$8C,$C7,$F8,$DF
	.dc.b	$77,$C3,$FC,$EF,$1B,$D3,$F0,$FF,$43,$C9,$B7,$F1,$67,$DB,$59,$CB
	.dc.b	$E2,$7C,$37,$D6,$F1,$7E,$0B,$63,$37,$F3,$7C,$8F,$8B,$F0,$F9,$8F
	.dc.b	$17,$C3,$7D,$EF,$17,$E4,$FC,$3D,$C3,$E8,$3F,$82,$F2,$7E,$6D,$66
	.dc.b	$FD,$5F,$0F,$F8,$6F,$17,$E8,$FC,$3F,$E2,$3C,$8B,$5F,$76,$7E,$AD
	.dc.b	$67,$2F,$B9,$F0,$FF,$8E,$F1,$BE,$CB,$F9,$0F,$1E,$E1,$F8,$59,$FB
	.dc.b	$BF,$0F,$9B,$FC,$27,$C3,$FE,$5B,$C6,$FC,$0F,$C3,$FE,$63,$CA,$B8
	.dc.b	$FE,$36,$7F,$0B,$59,$BF,$C6,$7C,$32,$A1,$FE,$C8,$FC,$C3,$FC,$37
	.dc.b	$E5,$7E,$3A,$73,$39,$AF,$C7,$C6,$23,$97,$C1,$2D,$E5,$7E,$53,$E1
	.dc.b	$F5,$5F,$EE,$39,$EB,$87,$FB,$E3,$C2,$7C,$3E,$B1,$E5,$E5,$FF,$85
	.dc.b	$74,$F3,$7C,$3E,$71,$4E,$72,$FF,$11,$9E,$28,$FF,$11,$BF,$8F,$C3
	.dc.b	$84,$EF,$E1,$F9,$79,$3C,$7E,$1F,$A7,$93,$70,$F6,$69,$EB,$AE,$11
	.dc.b	$8C,$95,$E1,$F8,$6F,$27,$8F,$E3,$F8,$7F,$37,$97,$70,$E1,$3B,$79
	.dc.b	$47,$A7,$E1,$FD,$9E,$4D,$9E,$DE,$45,$7E,$8E,$BF,$36,$7E,$07,$C3
	.dc.b	$FC,$3A,$F8,$5F,$0D,$F1,$7C,$7B,$87,$9F,$C9,$AF,$2E,$04,$FC,$9F
	.dc.b	$87,$F9,$95,$F2,$FE,$1B,$E7,$78,$FF,$37,$E1,$FE,$87,$97,$72,$F8
	.dc.b	$53,$B7,$E2,$47,$D3,$F8,$5F,$AD,$E5,$7D,$5F,$86,$FB,$1E,$4D,$BF
	.dc.b	$95,$3B,$7E,$64,$7D,$BF,$86,$FB,$DE,$3D,$9F,$A1,$E5,$57,$D5,$1F
	.dc.b	$81,$F8,$6F,$C2,$4F,$E0,$FE,$13,$F0,$DE,$57,$E7,$3C,$6F,$F3,$7E
	.dc.b	$1F,$E2,$53,$ED,$79,$39,$F2,$7F,$45,$3F,$A3,$F8,$5F,$C7,$4E,$69
	.dc.b	$3F,$A7,$C3,$AC,$70,$BE,$2B,$E1,$3F,$2D,$E4,$F1,$BF,$D7,$47,$E7
	.dc.b	$7F,$31,$FF,$3F,$C7,$93,$E1,$B1,$E6,$6F,$0E,$47,$E3,$F0,$C7,$C3
	.dc.b	$71,$F3,$35,$F0,$DC,$FC,$DC,$DF,$E6,$33,$FF,$3C,$BE,$53,$D2,$27
	.dc.b	$93,$9C,$FF,$D0,$2F,$97,$CB,$FD,$87,$C5,$7F,$89,$E3,$F2,$37,$FE
	.dc.b	$2D,$7E,$1C,$7F,$8D,$23,$FC,$FF,$BF,$38,$E1,$3F,$C3,$3C,$85,$71
	.dc.b	$C3,$FD,$1F,$F0,$BF,$0F,$CA,$EE,$DC,$79,$7C,$37,$C7,$F2,$B3,$97
	.dc.b	$B6,$FC,$BF,$0B,$F2,$FC,$AB,$3F,$05,$F4,$FC,$37,$CF,$F2,$73,$3E
	.dc.b	$A9,$F6,$FC,$37,$D3,$F2,$7D,$FF,$13,$F5,$7C,$AB,$7F,$2E,$76,$FC
	.dc.b	$D8,$F8,$BF,$0B,$F6,$FC,$9F,$8F,$F0,$BF,$77,$CA,$B7,$F4,$E7,$6F
	.dc.b	$D5,$8F,$9B,$F0,$9F,$83,$F2,$6C,$FD,$96,$BF,$43,$C8,$FA,$7F,$0B
	.dc.b	$BE,$6F,$9D,$F0,$BF,$8B,$F1,$BE,$BB,$63,$37,$F8,$3F,$23,$ED,$FC
	.dc.b	$2E,$63,$F3,$73,$FA,$0E,$F7,$C2,$FE,$57,$C9,$FC,$0F,$C2,$FE,$5F
	.dc.b	$CA,$B8,$7E,$36,$7F,$0B,$59,$C3,$F0,$1F,$09,$F9,$DF,$27,$F1,$3F
	.dc.b	$0B,$67,$F0,$8F,$E8,$7C,$8F,$C6,$D6,$71,$F1,$9F,$8C,$F8,$4F,$E8
	.dc.b	$5D,$DF,$D4,$3C,$6F,$F7,$3E,$EF,$85,$CD,$FE,$B6,$17,$88,$FC,$94
	.dc.b	$F2,$F8,$8D,$7C,$2F,$4F,$26,$D7,$F3,$EC,$BF,$C6,$2F,$E6,$7E,$1B
	.dc.b	$4F,$E6,$EB,$C7,$E5,$7E,$EF,$47,$07,$9B,$C8,$C3,$6F,$F5,$93,$FD
	.dc.b	$89,$ED,$3A,$7C,$1E,$47,$3F,$84,$F7,$F9,$1F,$D7,$EF,$FA,$ED,$79
	.dc.b	$7C,$8C,$7C,$3E,$B2,$EB,$87,$C6,$9E,$1F,$0B,$F2,$3C,$8E,$5F,$0B
	.dc.b	$6F,$C7,$3F,$33,$C9,$CE,$5F,$06,$FC,$1F,$09,$F4,$3C,$8F,$17,$C2
	.dc.b	$5C,$7D,$73,$F5,$3C,$7C,$E1,$F2,$6F,$D1,$F0,$9F,$63,$C7,$F5,$7C
	.dc.b	$27,$DA,$F1,$ED,$FC,$F9,$DB,$F4,$63,$E0,$7C,$1F,$E0,$3C,$5B,$3F
	.dc.b	$55,$AF,$F6,$2E,$C7,$91,$F1,$FE,$0F,$86,$FF,$43,$E0,$FF,$13,$E2
	.dc.b	$FC,$B6,$C6,$70,$FC,$07,$8F,$F3,$FE,$0F,$31,$F9,$2E,$CF,$E7,$FF
	.dc.b	$07,$F9,$2F,$1F,$F0,$1F,$07,$F9,$4F,$23,$F0,$53,$6F,$EF,$CE,$DF
	.dc.b	$C0,$FC,$1F,$E6,$BC,$6F,$C4,$7C,$1D,$9C,$9F,$91,$8C,$BF,$93,$F2
	.dc.b	$3F,$1D,$F0,$7F,$E2,$3F,$CC,$BF,$1F,$3F,$8C,$8C,$7C,$47,$E4,$FE
	.dc.b	$16,$CE,$4F,$F0,$CD,$70,$7D,$9B,$FF,$8B,$9C,$F8,$FF,$9B,$F8,$4E
	.dc.b	$CD,$8C,$9F,$D1,$7F,$32,$FF,$16,$AF,$82,$60,$F2,$FF,$AF,$53,$E1
	.dc.b	$7D,$1E,$56,$3E,$16,$E3,$CD,$AF,$CF,$AE,$0F,$A3,$CC,$E7,$F0,$BF
	.dc.b	$07,$CA,$B3,$EC,$6B,$FE,$24,$CF,$FD,$16,$F9,$FE,$23,$E4,$79,$9E
	.dc.b	$4F,$84,$F9,$5E,$5F,$9B,$E1,$7E,$67,$99,$70,$F8,$EE,$DF,$93,$3E
	.dc.b	$CF,$23,$E9,$79,$56,$7E,$6F,$91,$5F,$13,$F0,$BE,$17,$71,$F2,$7E
	.dc.b	$17,$ED,$79,$3F,$19,$FB,$9E,$3D,$9F,$AE,$FC,$AF,$85,$DC,$7D,$1F
	.dc.b	$85,$FC,$17,$8F,$FE,$63,$E7,$7C,$2F,$E1,$7C,$8C,$56,$BC,$CF,$B1
	.dc.b	$39,$7E,$D7,$C2,$FE,$2F,$CA,$E4,$D8,$FB,$DF,$0B,$F8,$FF,$2F,$3A
	.dc.b	$7E,$2A,$FC,$1F,$0B,$F9,$3F,$2B,$C5,$F0,$B7,$1F,$C3,$4F,$E5,$FC
	.dc.b	$AC,$E1,$F9,$2B,$F4,$7C,$27,$E6,$FC,$AF,$57,$C2,$FF,$9F,$FF,$33
	.dc.b	$7C,$46,$30,$FC,$CD,$70,$7F,$CE,$73,$AC,$FF,$29,$F0,$FC,$7C,$BF
	.dc.b	$89,$F0,$F6,$7F,$30,$F5,$F2,$FE,$45,$67,$0F,$DF,$65,$FE,$94,$7F
	.dc.b	$39,$8F,$F8,$56,$3F,$C3,$31,$FE,$9C,$8F,$F0,$EB,$C3,$E1,$BD,$1E
	.dc.b	$5F,$FA,$56,$B9,$79,$EB,$DF,$A3,$FB,$83,$DB,$F0,$DE,$FF,$2B,$FC
	.dc.b	$89,$7F,$D4,$32,$FF,$3E,$DA,$73,$F0,$DB,$9F,$F1,$B5,$6D,$FF,$16
	.dc.b	$53,$CB,$E7,$F0,$DF,$2F,$CA,$EB,$F0,$D6,$F8,$CD,$7E,$0C,$7D,$1F
	.dc.b	$27,$C7,$F0,$DF,$4F,$C9,$B5,$F4,$3C,$8F,$3F,$C3,$7D,$77,$31,$E8
	.dc.b	$9F,$5F,$C2,$FD,$BF,$1F,$E2,$7C,$35,$EB,$F4,$CF,$DF,$F2,$33,$87
	.dc.b	$DC,$BF,$95,$F0,$BF,$83,$F2,$7E,$67,$C3,$7E,$17,$CA,$CD,$FE,$0A
	.dc.b	$72,$FE,$12,$3E,$97,$C3,$7E,$2F,$C8,$FA,$9F,$0D,$75,$F9,$33,$F8
	.dc.b	$FF,$2B,$38,$7E,$32,$FE,$D7,$C3,$7E,$4F,$C9,$FB,$9F,$0D,$F9,$5F
	.dc.b	$2A,$DF,$E4,$A7,$6F,$E5,$23,$F0,$5F,$0D,$F9,$BF,$1F,$F0,$9F,$0D
	.dc.b	$75,$FB,$2F,$E7,$FC,$AD,$E1,$F9,$CA,$FC,$57,$C3,$7F,$B8,$7F,$34
	.dc.b	$FF,$7A,$61,$F8,$0F,$87,$C5,$FE,$3D,$D7,$99,$C2,$3F,$DA,$F5,$C1
	.dc.b	$FC,$EF,$C3,$F3,$F2,$BF,$1F,$F0,$FD,$7C,$AF,$C8,$CD,$AF,$C6,$CE
	.dc.b	$DF,$F4,$27,$E6,$9F,$D2,$5F,$F3,$D6,$0F,$1F,$FD,$39,$3F,$E1,$D9
	.dc.b	$FF,$6A,$CE,$5E,$3A,$7A,$EB,$7F,$11,$ED,$F2,$38,$FC,$47,$BF,$C8
	.dc.b	$B8,$7A,$BC,$8E,$BF,$11,$F0,$DC,$C7,$69,$F0,$FC,$47,$C7,$F1,$EC
	.dc.b	$FC,$27,$CB,$F1,$19,$8F,$5C,$FC,$DF,$27,$D3,$B7,$96,$7D,$FA,$D7
	.dc.b	$FC,$05,$C1,$F0,$51,$FD,$87,$83,$DC,$EA,$7D,$8F,$C8,$F8,$8F,$B5
	.dc.b	$E4,$7C,$6F,$86,$B3,$93,$EB,$C6,$5F,$B3,$E4,$FC,$CF,$85,$FC,$17
	.dc.b	$8F,$F3,$BE,$1B,$F0,$9E,$4E,$6F,$EF,$B9,$7F,$03,$3E,$C7,$F0,$7D
	.dc.b	$1F,$41,$A7,$D4,$29,$E9,$3F,$13,$D1,$F7,$3E,$1B,$F2,$5E,$45,$9F
	.dc.b	$C6,$F9,$1F,$80,$F8,$6F,$CB,$39,$8F,$C0,$CF,$E1,$3E,$1B,$F3,$5E
	.dc.b	$45,$9F,$CA,$F9,$1F,$88,$F8,$6F,$CF,$39,$8F,$C4,$CF,$E3,$3E,$1B
	.dc.b	$F4,$5E,$47,$E3,$BE,$16,$D7,$E5,$26,$BF,$5A,$3F,$51,$E4,$FE,$53
	.dc.b	$E1,$3F,$55,$E3,$DA,$FD,$2B,$F9,$8F,$85,$CC,$7E,$7A,$7F,$61,$E4
	.dc.b	$7E,$73,$E1,$3E,$FC,$C3,$FC,$41,$D9,$FE,$2B,$F7,$FF,$9E,$78,$2A
	.dc.b	$D5,$7E,$1B,$E0,$17,$FA,$5B,$DF,$C2,$EA,$B9,$F8,$5D,$DE,$37,$C3
	.dc.b	$C4,$F8,$51,$FB,$CE,$2F,$84,$D3,$E1,$14,$F8,$47,$F0,$2E,$0F,$DA
	.dc.b	$FC,$2F,$7B,$F1,$F9,$BE,$CC,$FF,$AA,$7E,$69,$FE,$3C,$9C,$E1,$FC
	.dc.b	$02,$F8,$6B,$FE,$92,$8F,$F3,$25,$6E,$3F,$C8,$11,$FA,$4B,$E7,$F0
	.dc.b	$DC,$3C,$AE,$DA,$E7,$07,$C3,$74,$F3,$33,$8B,$15,$7F,$90,$A3,$E8
	.dc.b	$69,$F2,$AB,$75,$AB,$ED,$5F,$08,$FE,$23,$BF,$C6,$69,$93,$F8,$CB
	.dc.b	$97,$F8,$DE,$FD,$31,$7C,$37,$C0,$F2,$F0,$FC,$2D,$F5,$F1,$17,$0F
	.dc.b	$81,$B7,$B6,$B8,$46,$0F,$66,$1F,$DE,$DF,$BB,$6F,$2B,$C1,$F0,$FF
	.dc.b	$39,$F1,$7C,$2D,$AE,$5B,$FC,$4A,$FA,$7E,$67,$9F,$E1,$7E,$AF,$93
	.dc.b	$6F,$AF,$93,$EB,$F8,$5F,$B2,$E6,$3E,$5E,$8A,$F8,$1F,$09,$F7,$BC
	.dc.b	$9B,$5E,$6F,$27,$E2,$7C,$2F,$E0,$9C,$C7,$29,$F9,$1F,$09,$F8,$6F
	.dc.b	$27,$E5,$7C,$2D,$AF,$4F,$22,$B1,$1F,$8A,$9F,$C6,$F9,$5E,$58,$FD
	.dc.b	$7F,$77,$EA,$35,$C4,$7E,$AF,$66,$FE,$D8,$F7,$9D,$B8,$7E,$1B,$E0
	.dc.b	$F8,$5F,$CC,$F9,$3F,$77,$E1,$7F,$37,$E5,$FD,$F9,$B5,$E6,$9D,$BE
	.dc.b	$8F,$84,$FD,$0F,$95,$F8,$5F,$85,$B3,$93,$ED,$C6,$5F,$BB,$E5,$FA
	.dc.b	$1F,$E0,$FB,$5D,$FF,$81,$ED,$E9,$9F,$A6,$E7,$7F,$A9,$1F,$23,$E1
	.dc.b	$7F,$63,$E5,$7E,$57,$E1,$6C,$E4,$FC,$5C,$65,$FE,$13,$F9,$AF,$F9
	.dc.b	$A9,$FA,$3F,$0B,$FC,$06,$3F,$59,$E0,$FF,$00,$F8,$6C,$79,$DF,$A1
	.dc.b	$AB,$7F,$62,$78,$79,$9F,$B0,$FC,$D7,$FC,$BB,$8F,$E9,$AB,$F2,$B1
	.dc.b	$8F,$23,$5F,$0F,$DD,$E1,$3E,$1F,$2F,$97,$C3,$F8,$FC,$9B,$3E,$8F
	.dc.b	$2B,$B7,$C3,$F9,$DC,$C7,$F0,$2C,$FF,$54,$FF,$BD,$73,$FE,$E2,$FE
	.dc.b	$22,$1E,$07,$9F,$DC,$F8,$1E,$5F,$A7,$DD,$B7,$E8,$9F,$89,$5E,$68
	.dc.b	$F9,$3B,$57,$C0,$FB,$1C,$9F,$00,$F6,$1F,$0B,$E1,$7E,$77,$9B,$F1
	.dc.b	$DB,$1F,$1B,$E1,$AB,$E9,$8F,$A9,$E7,$7C,$DF,$86,$FA,$DE,$5F,$CF
	.dc.b	$F8,$6B,$7F,$3A,$6B,$F0,$63,$EE,$79,$9E,$78,$F8,$59,$D7,$FA,$97
	.dc.b	$07,$D7,$7E,$47,$3F,$D5,$7D,$47,$E2,$39,$FE,$F7,$C2,$FE,$1F,$CA
	.dc.b	$FC,$07,$C2,$DA,$FB,$F3,$5F,$A5,$1F,$8D,$F2,$FF,$0D,$F0,$9F,$8F
	.dc.b	$F3,$2D,$7E,$51,$FC,$57,$C2,$E6,$3F,$13,$3F,$95,$F3,$7F,$1D,$F0
	.dc.b	$9F,$97,$F2,$2D,$7E,$71,$FC,$97,$C2,$E6,$3F,$23,$3F,$9D,$F2,$7F
	.dc.b	$2D,$F0,$9F,$9F,$F1,$ED,$7E,$93,$C7,$FC,$D7,$C2,$FE,$8D,$CC,$7E
	.dc.b	$36,$7F,$3D,$EF,$7F,$67,$FC,$DB,$FC,$01,$E6,$7F,$AE,$7C,$64,$E6
	.dc.b	$FF,$49,$EF,$D7,$F4,$71,$C7,$A7,$D3,$1F,$99,$E0,$AF,$E5,$DF,$5B
	.dc.b	$ED,$3D,$67,$EE,$B6,$FD,$87,$C3,$78,$FC,$CB,$3F,$E0,$6D,$3F,$D9
	.dc.b	$1A,$79,$7E,$1B,$53,$FB,$4B,$F5,$79,$9F,$DA,$3E,$03,$FC,$B5,$2F
	.dc.b	$51,$8F,$F9,$6B,$1F,$F6,$B5,$6A,$73,$E5,$F1,$F8,$5F,$88,$F3,$9F
	.dc.b	$8D,$E6,$FB,$31,$FC,$7E,$F5,$ED,$E1,$A3,$DF,$1F,$0C,$F7,$9F,$BD
	.dc.b	$E0,$F3,$7C,$2F,$D1,$F2,$EC,$FB,$FC,$CF,$57,$C4,$7D,$57,$31,$FE
	.dc.b	$06,$E0,$57,$BF,$E1,$BE,$D7,$99,$F0,$7E,$1B,$EE,$79,$B9,$C3,$E5
	.dc.b	$39,$7E,$64,$FC,$7F,$86,$FC,$17,$95,$F2,$7D,$9F,$C2,$79,$79,$BF
	.dc.b	$A4,$E5,$FA,$93,$F3,$FE,$1B,$F1,$5E,$4D,$9F,$B1,$E5,$7D,$3F,$86
	.dc.b	$FC,$73,$98,$F5,$4F,$D7,$F7,$7F,$25,$E5,$7D,$9F,$77,$F2,$9E,$66
	.dc.b	$6F,$F0,$4E,$5F,$C2,$4F,$95,$F8,$99,$F9,$5F,$8D,$A6,$5F,$ED,$FC
	.dc.b	$88,$D4,$62,$7F,$11,$F0,$BF,$A3,$F3,$6C,$FE,$47,$CD,$FC,$67,$C3
	.dc.b	$7E,$9D,$CC,$7D,$39,$FC,$87,$C3,$7E,$AF,$CB,$FC,$97,$C2,$DA,$FF
	.dc.b	$0E,$F8,$6A,$C4,$7F,$90,$E3,$FC,$5B,$3F,$B3,$F3,$7D,$13,$FE,$6D
	.dc.b	$D9,$AF,$A2,$31,$B6,$9F,$CF,$4F,$B5,$EF,$CF,$F7,$1F,$0F,$C2,$FE
	.dc.b	$FF,$CB,$FD,$33,$63,$CB,$F0,$B5,$F2,$1E,$58,$FE,$0D,$E6,$FE,$B7
	.dc.b	$E1,$7F,$CD,$DF,$9B,$B7,$FD,$5D,$FA,$E7,$3E,$66,$3C,$ED,$7C,$37
	.dc.b	$0B,$C1,$FE,$8B,$BF,$F4,$1E,$0B,$CF,$C3,$76,$F2,$B9,$7E,$6F,$FE
	.dc.b	$79,$A4,$6A,$F8,$F9,$7A,$F8,$CE,$19,$7F,$93,$30,$FF,$37,$57,$F8
	.dc.b	$C3,$07,$95,$D3,$E2,$3D,$6F,$6F,$88,$F6,$F9,$99,$BF,$2C,$F8,$BE
	.dc.b	$03,$E0,$F9,$76,$7D,$3D,$1E,$6F,$80,$CE,$BE,$1B,$F8,$F3,$E7,$AF
	.dc.b	$F3,$FE,$9E,$AA,$F3,$6D,$E8,$9F,$A3,$C9,$F3,$78,$BE,$7F,$0F,$B6
	.dc.b	$7E,$3F,$17,$C4,$F8,$8F,$A7,$E6,$D9,$F9,$7E,$6F,$C8,$F8,$8C,$E9
	.dc.b	$F0,$EF,$EC,$CF,$CC,$F8,$7F,$B7,$E6,$FC,$EF,$88,$FB,$BE,$76,$72
	.dc.b	$FA,$B3,$97,$EB,$C7,$AD,$F9,$9A,$79,$5F,$6B,$97,$FD,$8D,$C2,$DB
	.dc.b	$51,$C3,$4F,$AD,$F1,$1F,$8C,$F3,$7E,$FB,$F8,$EF,$37,$F0,$3F,$11
	.dc.b	$9A,$DD,$FE,$22,$7F,$0B,$F1,$1F,$94,$F3,$2C,$FE,$33,$CC,$FC,$4F
	.dc.b	$C4,$67,$0F,$C2,$5F,$E6,$A7,$F1,$BF,$11,$F9,$CF,$23,$F1,$FF,$11
	.dc.b	$6F,$F1,$53,$FA,$0F,$2B,$39,$7E,$62,$FF,$2B,$F1,$1F,$A4,$F2,$AC
	.dc.b	$FE,$71,$FC,$CF,$C4,$7E,$A3,$CB,$CC,$FE,$0A,$7F,$3B,$EF,$7E,$B3
	.dc.b	$C0,$FC,$FF,$BD,$FE,$DB,$F6,$DE,$26,$B0,$FD,$3D,$69,$FF,$75,$CE
	.dc.b	$B7,$F3,$C7,$95,$FD,$17,$07,$AE,$3F,$1B,$E6,$B8,$7D,$95,$A9,$E0
	.dc.b	$6B,$E1,$FC,$5E,$77,$EC,$BE,$23,$C9,$E7,$FE,$D2,$73,$7F,$AF,$9C
	.dc.b	$BF,$EB,$CF,$CD,$D8,$7A,$FC,$CF,$E0,$91,$88,$F4,$F9,$B8,$F8,$8D
	.dc.b	$5F,$19,$F8,$53,$F5,$9F,$2F,$9B,$FC,$45,$FF,$76,$6C,$DF,$DD,$39
	.dc.b	$11,$EF,$9F,$E1,$FD,$5F,$69,$FD,$9F,$57,$93,$E2,$3E,$87,$99,$E6
	.dc.b	$F8,$8B,$A7,$AE,$7E,$A7,$9D,$9C,$3E,$8D,$FB,$3E,$1B,$EC,$79,$D6
	.dc.b	$7E,$AB,$F0,$3E,$1F,$EE,$79,$F9,$9F,$F8,$03,$D5,$FE,$7C,$CF,$E4
	.dc.b	$7C,$2F,$E0,$7C,$AF,$8F,$F0,$D6,$73,$7F,$80,$F2,$F2,$FE,$0A,$BE
	.dc.b	$6E,$1F,$89,$F2,$2C,$FE,$1B,$C9,$FA,$3F,$0D,$9C,$7E,$B5,$FE,$3E
	.dc.b	$7E,$AF,$BD,$F9,$1F,$0A,$E1,$F8,$EE,$AF,$B3,$EF,$67,$1F,$BD,$7F
	.dc.b	$97,$9F,$BB,$BF,$9E,$FF,$8D,$F7,$38,$31,$D9,$F5,$63,$FC,$37,$3E
	.dc.b	$BC,$7F,$11,$5A,$7F,$13,$4A,$7C,$2F,$E9,$7C,$CF,$C5,$B6,$BF,$21
	.dc.b	$F0,$DF,$A9,$F3,$73,$97,$E9,$AF,$F2,$3F,$0B,$FA,$DF,$33,$F2,$7F
	.dc.b	$0D,$71,$FC,$C4,$FE,$C7,$CD,$CE,$1F,$AE,$BF,$6B,$F3,$EB,$CE,$D3
	.dc.b	$DF,$BF,$EE,$7A,$7E,$AC,$7D,$EE,$9F,$D0,$FC,$2F,$EF,$7C,$AF,$D1
	.dc.b	$FC,$35,$C7,$F5,$13,$FC,$07,$CB,$CE,$1F,$BE,$BF,$D4,$FC,$37,$EA
	.dc.b	$63,$F5,$5F,$10,$FC,$E3,$FC,$F7,$8F,$C8,$E6,$7C,$46,$EF,$F6,$77
	.dc.b	$BF,$3B,$F6,$5D,$3F,$4E,$B5,$5F,$E6,$9C,$55,$AC,$7A,$65,$AC,$7C
	.dc.b	$17,$91,$8F,$D7,$F8,$7D,$63,$FE,$7A,$C3,$C9,$18,$F2,$B8,$56,$EF
	.dc.b	$CD,$87,$AE,$73,$A3,$E1,$F9,$F9,$9C,$3F,$38,$FF,$3D,$69,$FE,$7D
	.dc.b	$D3,$F0,$F1,$FE,$0B,$CB,$9C,$E5,$F2,$C6,$5F,$95,$97,$C9,$F2,$FA
	.dc.b	$5F,$A3,$CA,$ED,$EF,$FA,$BA,$6E,$1D,$3C,$2A,$F2,$CF,$F3,$CF,$E6
	.dc.b	$30,$F1,$FB,$DC,$63,$F9,$27,$17,$C1,$69,$E6,$E1,$F7,$C6,$63,$F5
	.dc.b	$3C,$1F,$01,$E8,$7A,$BE,$1B,$E6,$F9,$5F,$0D,$F9,$FE,$5D,$9F,$82
	.dc.b	$FC,$7F,$87,$DC,$72,$E6,$72,$3E,$1F,$EB,$F9,$5F,$37,$E1,$ED,$EE
	.dc.b	$91,$88,$FF,$0B,$C7,$17,$E7,$F9,$9E,$A7,$D8,$7B,$8F,$8D,$BE,$5A
	.dc.b	$7F,$47,$DB,$DC,$F4,$F8,$7F,$C4,$F9,$9F,$7D,$FC,$5F,$99,$67,$EE
	.dc.b	$BF,$83,$F8,$7D,$C7,$9F,$E1,$FF,$23,$E5,$FE,$1F,$E1,$FF,$27,$E6
	.dc.b	$DA,$FC,$2B,$F8,$BA,$DC,$7B,$FE,$1B,$F3,$3E,$4F,$E3,$FE,$1A,$CF
	.dc.b	$C3,$6B,$F0,$A7,$F3,$FE,$5F,$E5,$7E,$1B,$F4,$3E,$47,$E5,$FE,$1B
	.dc.b	$F4,$7E,$5D,$C3,$F2,$AE,$DF,$CB,$CF,$E7,$7D,$CF,$D4,$F8,$1F,$9F
	.dc.b	$F7,$6D,$7D,$36,$BF,$4A,$7F,$5F,$CD,$FA,$5D,$3C,$BD,$BF,$D8,$FA
	.dc.b	$9B,$E3,$C3,$FD,$3B,$71,$FD,$2D,$FA,$9F,$C0,$7C,$2F,$EF,$FF,$39
	.dc.b	$FE,$E1,$1F,$80,$F8,$7C,$7C,$4F,$EC,$DB,$1C,$0F,$DD,$CF,$0C,$FC
	.dc.b	$5F,$0F,$CF,$CC,$FD,$D3,$D7,$C9,$B7,$C1,$FE,$0B,$E9,$7F,$88,$A3
	.dc.b	$9D,$7B,$5F,$DE,$F4,$FD,$56,$9F,$54,$A7,$D5,$3F,$81,$F4,$F4,$F8
	.dc.b	$7F,$67,$95,$DB,$E1,$FD,$DE,$55,$BF,$4B,$B7,$D7,$3E,$4F,$87,$F8
	.dc.b	$9E,$5F,$9B,$E1,$FE,$37,$97,$6F,$E0,$BB,$7E,$1C,$FB,$3E,$17,$E6
	.dc.b	$79,$36,$7E,$3F,$91,$5F,$E2,$BC,$A8,$F8,$7F,$0D,$C2,$7F,$75,$C6
	.dc.b	$F8,$6F,$AD,$E3,$FC,$86,$CE,$9A,$FF,$A3,$A5,$F0,$DF,$77,$CA,$F9
	.dc.b	$FE,$EF,$DF,$EF,$B7,$F6,$9A,$F9,$1F,$AB,$EF,$7E,$17,$C1,$F4,$4F
	.dc.b	$F5,$4D,$9B,$E2,$3F,$A8,$C7,$AD,$F4,$F0,$7E,$37,$83,$DB,$1F,$88
	.dc.b	$AF,$C0,$FC,$37,$E4,$7C,$CB,$3F,$8C,$F3,$2B,$F1,$9F,$C3,$FC,$36
	.dc.b	$E3,$E2,$FC,$37,$E6,$7C,$AF,$C5,$B6,$32,$7E,$5A,$3F,$1F,$F0,$DF
	.dc.b	$9F,$F3,$3D,$12,$F7,$9B,$FB,$23,$D2,$7D,$82,$9F,$63,$6F,$5B,$F6
	.dc.b	$B6,$FC,$DF,$C2,$FE,$AF,$CB,$FC,$EF,$C3,$7E,$B7,$CB,$B7,$FA,$87
	.dc.b	$6F,$EA,$A7,$F4,$7F,$0B,$FB,$3F,$2F,$F4,$BF,$0D,$6B,$F0,$6D,$7E
	.dc.b	$F4,$FF,$77,$FC,$EB,$F5,$51,$F8,$1F,$86,$C7,$95,$FA,$EF,$87,$D7
	.dc.b	$97,$6B,$FB,$9D,$A3,$F6,$77,$C2,$7F,$A5,$7E,$75,$FE,$08,$7E,$35
	.dc.b	$3C,$AF,$EF,$B3,$9E,$44,$E6,$B8,$9E,$5A,$DF,$89,$E7,$F2,$B8,$FB
	.dc.b	$BE,$9F,$02,$DF,$93,$C9,$AF,$F7,$AC,$51,$E0,$F8,$7E,$13,$FA,$6D
	.dc.b	$7D,$0F,$E8,$77,$F7,$BF,$AA,$6B,$F1,$4F,$83,$C1,$EA,$F8,$7F,$95
	.dc.b	$E5,$7B,$3E,$1F,$E6,$79,$76,$FE,$3B,$B7,$E4,$CF,$C2,$F8,$7F,$A5
	.dc.b	$E5,$FC,$4F,$87,$B5,$F2,$9A,$FF,$8A,$BD,$9F,$F3,$0F,$3F,$F9,$9A
	.dc.b	$3D,$DE,$67,$D2,$7C,$A7,$A4,$FA,$FD,$3F,$65,$F8,$8E,$DF,$8D,$B7
	.dc.b	$D4,$F8,$6F,$C2,$79,$9F,$5B,$E1,$AD,$7D,$C6,$BE,$33,$FC,$7B,$8A
	.dc.b	$BF,$19,$E5,$FD,$EF,$86,$FC,$77,$99,$F8,$0F,$86,$B8,$7E,$19,$AF
	.dc.b	$3F,$71,$8A,$FF,$2D,$E6,$FE,$23,$E1,$BF,$31,$E5,$7E,$2B,$E1,$BF
	.dc.b	$35,$E5,$DC,$BF,$2A,$ED,$FC,$BC,$FE,$43,$E1,$BF,$41,$E4,$7E,$4B
	.dc.b	$E1,$BF,$45,$E5,$5B,$FC,$EB,$B7,$F3,$F3,$F9,$8F,$7B,$F5,$1E,$17
	.dc.b	$E6,$BD,$EF,$D5,$78,$76,$FF,$4A,$ED,$FD,$3C,$F8,$DF,$CA,$67,$E7
	.dc.b	$7F,$2D,$A5,$7F,$CF,$3E,$32,$38,$46,$27,$F5,$3F,$0B,$FE,$6F,$F0
	.dc.b	$FF,$CF,$B3,$F4,$BE,$1B,$0F,$EB,$9A,$FE,$D2,$78,$79,$9F,$B2,$F8
	.dc.b	$6E,$5E,$65,$AF,$F3,$7E,$0F,$2B,$11,$FE,$64,$8C,$3F,$BE,$F8,$6E
	.dc.b	$2F,$F0,$0E,$7F,$AA,$FF,$02,$E7,$FA,$AD,$3E,$A9,$B3,$EA,$9F,$E5
	.dc.b	$1F,$43,$FC,$16,$FF,$A6,$AF,$FC,$ED,$D9,$FC,$BF,$C6,$F8,$1E,$5F
	.dc.b	$FA,$1A,$F7,$7F,$E9,$AB,$4E,$7E,$17,$E3,$79,$BB,$F8,$4F,$91,$E6
	.dc.b	$DC,$3D,$EF,$3F,$85,$CC,$78,$E7,$E7,$79,$BD,$FE,$13,$E8,$79,$36
	.dc.b	$BE,$3F,$95,$E3,$F8,$5F,$A8,$E6,$3E,$44,$F9,$FE,$13,$EC,$79,$16
	.dc.b	$7E,$7F,$95,$EB,$F8,$5F,$B8,$E6,$3E,$84,$FB,$FD,$BF,$C0,$74,$59
	.dc.b	$FA,$EF,$C3,$F7,$33,$1F,$2E,$7F,$0D,$D3,$E5,$8F,$BB,$9F,$95,$F9
	.dc.b	$DA,$65,$FB,$CE,$5F,$99,$1F,$3F,$E1,$7F,$21,$E6,$7D,$1F,$85,$FC
	.dc.b	$97,$9B,$9B,$FC,$4B,$97,$F1,$73,$F5,$FE,$17,$F3,$1E,$65,$9F,$C7
	.dc.b	$F9,$9F,$6F,$E1,$7F,$38,$E6,$3F,$21,$3E,$57,$F0,$99,$F9,$5F,$EE
	.dc.b	$5D,$CD,$75,$1F,$83,$75,$DF,$FE,$1C,$AF,$C4,$7C,$2F,$EA,$7C,$8F
	.dc.b	$C5,$7C,$2F,$EA,$FC,$AC,$E3,$FA,$27,$2F,$E9,$27,$F2,$1F,$0B,$FB
	.dc.b	$1F,$2A,$CF,$EA,$1F,$CA,$7C,$2E,$63,$F3,$93,$FB,$7F,$27,$F3,$1F
	.dc.b	$07,$FE,$E7,$FC,$EF,$FC,$C5,$E5,$2F,$FD,$CF,$BA,$B5,$E4,$FE,$7F
	.dc.b	$E1,$6D,$7E,$9E,$79,$5F,$E8,$FE,$17,$A7,$91,$6B,$FD,$D3,$92,$3F
	.dc.b	$DD,$BF,$9D,$FF,$80,$27,$C0,$7F,$86,$E7,$4F,$EC,$EB,$CD,$E5,$70
	.dc.b	$F7,$BD,$1E,$55,$AE,$BE,$57,$4F,$7F,$D8,$E6,$3F,$DD,$78,$AB,$E0
	.dc.b	$C7,$93,$4A,$EB,$7F,$73,$E8,$72,$F9,$B8,$17,$F2,$21,$E8,$36,$FF
	.dc.b	$05,$47,$83,$CC,$F7,$FC,$37,$CF,$7E,$0F,$C3,$59,$FF,$70,$57,$07
	.dc.b	$B6,$1F,$ED,$C7,$3F,$0D,$F5,$FC,$AF,$92,$FD,$9F,$2F,$E5,$FC,$35
	.dc.b	$AD,$9F,$42,$BD,$2F,$DD,$D7,$D3,$BE,$23,$E8,$6F,$F0,$1F,$84,$7D
	.dc.b	$D3,$E3,$1E,$6F,$86,$FC,$47,$99,$F6,$BC,$5B,$39,$3E,$FC,$65,$FC
	.dc.b	$0F,$99,$F8,$0F,$86,$FC,$87,$97,$67,$F0,$AF,$E1,$3E,$1B,$31,$F1
	.dc.b	$27,$F2,$DE,$67,$E2,$3E,$1B,$F3,$1E,$4D,$AF,$C6,$F9,$3F,$8C,$F8
	.dc.b	$6F,$CE,$39,$8F,$C0,$CF,$E4,$3E,$1B,$F4,$1E,$3F,$E4,$BE,$17,$F4
	.dc.b	$5E,$4E,$6F,$F2,$EE,$5F,$CC,$CF,$E6,$3D,$FF,$D4,$79,$1F,$9A,$F8
	.dc.b	$0B,$5F,$6A,$6B,$FE,$0B,$EE,$FE,$D4,$FE,$8B,$CA,$F2,$3F,$7B,$4A
	.dc.b	$FD,$5E,$F6,$DE,$B8,$F6,$9F,$86,$E0,$F8,$2F,$E2,$3E,$13,$F7,$5E
	.dc.b	$57,$EB,$3E,$16,$CE,$4F,$D8,$C6,$5F,$D7,$9F,$83,$FC,$F3,$FC,$E3
	.dc.b	$E3,$BB,$FF,$81,$3F,$9B,$F8,$5D,$D6,$FC,$CF,$DC,$FC,$36,$23,$FD
	.dc.b	$25,$82,$7F,$7B,$59,$F0,$D3,$9A,$FC,$FD,$E6,$3F,$CF,$51,$ED,$3F
	.dc.b	$83,$F0,$7C,$47,$F8,$26,$FF,$0D,$FD,$DB,$9B,$FD,$9C,$FF,$0A,$CB
	.dc.b	$D5,$5F,$B4,$CB,$FC,$FD,$87,$ED,$27,$27,$8E,$7D,$B7,$E6,$F2,$B9
	.dc.b	$69,$F0,$B0,$7C,$2F,$9E,$73,$7D,$4F,$57,$95,$8C,$7E,$17,$C2,$FB
	.dc.b	$7C,$BC,$62,$8F,$81,$E5,$63,$E1,$6E,$3F,$43,$4F,$05,$71,$8C,$1E
	.dc.b	$BF,$84,$F9,$1E,$57,$4F,$85,$F9,$5E,$5F,$69,$B5,$C3,$5F,$A9,$5C
	.dc.b	$23,$1A,$7F,$9E,$5F,$E3,$9C,$4F,$1F,$CF,$F0,$9F,$55,$F4,$FC,$2D
	.dc.b	$AE,$33,$F6,$7C,$AF,$6F,$B7,$F6,$FB,$FD,$FE,$E7,$DD,$F2,$F3,$87
	.dc.b	$CE,$CD,$1A,$8C,$57,$A5,$D7,$0F,$B5,$E1,$C3,$97,$F3,$3D,$5F,$8C
	.dc.b	$8D,$46,$30,$57,$D0,$F8,$3F,$C7,$79,$96,$BC,$5E,$67,$D4,$F8,$4F
	.dc.b	$C9,$39,$8E,$53,$F6,$3E,$13,$F2,$DE,$57,$DA,$F8,$4F,$CC,$79,$99
	.dc.b	$BF,$63,$97,$DD,$3E,$77,$E6,$E7,$E6,$7E,$7E,$B9,$7F,$64,$E5,$FD
	.dc.b	$A4,$7E,$23,$E0,$FF,$4D,$E5,$59,$F9,$5E,$57,$E3,$3E,$13,$F5,$4E
	.dc.b	$63,$E1,$4F,$E4,$3E,$0F,$F5,$DE,$5F,$E4,$BE,$13,$F6,$1E,$65,$BF
	.dc.b	$A8,$ED,$FA,$D3,$F9,$8F,$83,$FD,$D7,$E7,$BF,$E7,$EF,$0F,$F7,$D1
	.dc.b	$8F,$33,$57,$FA,$AA,$FD,$3B,$F8,$EF,$85,$E2,$FE,$8B,$E1,$79,$F9
	.dc.b	$1F,$A4,$F8,$5B,$3F,$94,$6B,$FE,$B3,$F1,$D5,$E2,$F2,$FF,$5D,$F9
	.dc.b	$EF,$F9,$C2,$9E,$0B,$37,$97,$FB,$EF,$33,$F8,$DD,$EE,$B5,$39,$BF
	.dc.b	$EF,$32,$F7,$BD,$FE,$6E,$78,$93,$C7,$2E,$B9,$7E,$B6,$3F,$C2,$F8
	.dc.b	$7E,$52,$5E,$67,$19,$CC,$FF,$82,$67,$35,$C7,$E1,$7E,$6F,$9B,$BF
	.dc.b	$86,$F9,$F8,$F1,$F8,$4F,$A3,$E7,$5C,$BE,$73,$D7,$E1,$7E,$AF,$9F
	.dc.b	$99,$F1,$4F,$87,$E1,$7E,$CF,$9B,$F2,$3E,$1F,$ED,$F9,$DF,$2A,$6D
	.dc.b	$FA,$E7,$6F,$B7,$E1,$7F,$03,$E6,$FD,$06,$C7,$C1,$F8,$6A,$FF,$BA
	.dc.b	$79,$13,$F8,$8F,$3B,$EB,$FC,$37,$E2,$BC,$BB,$7F,$87,$7E,$DF,$C3
	.dc.b	$66,$3E,$64,$FE,$43,$CC,$FB,$FF,$0D,$F9,$2F,$2E,$D7,$E3,$DA,$FD
	.dc.b	$17,$F0,$BF,$0D,$F9,$8F,$33,$F0,$FE,$F7,$E6,$BC,$BF,$C4,$FB,$FF
	.dc.b	$9C,$F2,$EE,$3F,$99,$76,$FE,$6E,$7F,$1F,$F0,$BF,$A2,$F2,$7F,$23
	.dc.b	$F0,$D6,$BF,$04,$D7,$E1,$4F,$EA,$3C,$CF,$CB,$FC,$37,$F6,$3E,$5F
	.dc.b	$ED,$DE,$2F,$E2,$7E,$17,$1E,$57,$E1,$EB,$53,$F8,$99,$CB,$F8,$FF
	.dc.b	$84,$E5,$E3,$7E,$35,$B1,$FC,$FF,$F3,$DF,$F3,$66,$9F,$E3,$FC,$9E
	.dc.b	$4E,$E3,$FA,$FC,$7F,$9C,$EB,$87,$A1,$E5,$F2,$3F,$41,$87,$F6,$5B
	.dc.b	$FF,$30,$5F,$F9,$A6,$FB,$7C,$47,$AF,$C9,$DE,$3C,$67,$C5,$F0,$DE
	.dc.b	$FF,$1B,$FB,$26,$4F,$86,$C6,$7A,$3C,$73,$A8,$F8,$DE,$3F,$0F,$86
	.dc.b	$F9,$1A,$DB,$F4,$F8,$FD,$3E,$1B,$E6,$39,$8F,$E8,$B3,$F9,$27,$E2
	.dc.b	$7B,$9F,$47,$C5,$F1,$B6,$3E,$47,$BB,$5F,$F3,$DE,$FF,$E2,$2D,$BF
	.dc.b	$A5,$C7,$C8,$F1,$FD,$7F,$0D,$F6,$FC,$7F,$8F,$F0,$DF,$77,$C7,$B7
	.dc.b	$F3,$A7,$E5,$D6,$71,$FA,$DF,$0B,$F8,$3F,$1F,$E7,$FC,$37,$E1,$7C
	.dc.b	$8B,$8F,$D6,$9F,$A7,$59,$BF,$BD,$F0,$9F,$8B,$F1,$7E,$BF,$C2,$D9
	.dc.b	$FC,$13,$F8,$FF,$1F,$ED,$D6,$71,$F6,$9F,$E9,$8E,$17,$C2,$7E,$57
	.dc.b	$C6,$FC,$0B,$71,$FC,$63,$F9,$9F,$1B,$F0,$BF,$0B,$9C,$39,$FB,$6C
	.dc.b	$5B,$7B,$71,$FC,$A7,$0F,$A6,$F7,$5F,$E0,$27,$DD,$3F,$11,$FC,$AF
	.dc.b	$0F,$E4,$BE,$13,$F5,$1E,$47,$E5,$3E,$16,$DF,$E7,$DF,$F2,$77,$E7
	.dc.b	$CF,$91,$E1,$8F,$E7,$AF,$F3,$7F,$0D,$AF,$23,$F3,$BF,$0F,$72,$FD
	.dc.b	$43,$CB,$CB,$DE,$1F,$E4,$4C,$BF,$CB,$95,$E5,$8F,$C6,$F5,$7D,$67
	.dc.b	$F1,$FB,$57,$F4,$F1,$F6,$1F,$5C,$FF,$8C,$FF,$3E,$FF,$34,$47,$F5
	.dc.b	$8B,$79,$7F,$D3,$E7,$FB,$4C,$FF,$87,$A7,$35,$C7,$E2,$3D,$FE,$5E
	.dc.b	$EB,$E0,$E1,$C7,$E2,$3E,$1F,$97,$FE,$53,$C9,$F1,$1F,$1B,$CC,$C3
	.dc.b	$70,$F2,$5F,$07,$CD,$F0,$FF,$33,$CA,$E4,$FC,$EF,$2A,$CF,$C5,$7B
	.dc.b	$7C,$3E,$E3,$DD,$F0,$BF,$53,$C8,$F1,$7C,$35,$9F,$84,$D7,$F1,$3A
	.dc.b	$FF,$94,$DF,$F6,$B7,$1B,$CC,$F5,$7C,$2F,$DE,$F0,$BD,$9F,$0D,$F8
	.dc.b	$0F,$0E,$E1,$F5,$DD,$BF,$66,$7C,$EF,$CB,$E0,$AF,$FB,$EF,$16,$F8
	.dc.b	$8F,$0B,$EC,$7F,$05,$3F,$2F,$E1,$BF,$1F,$E5,$59,$FC,$33,$5F,$B0
	.dc.b	$FD,$1F,$86,$FC,$AF,$9B,$F4,$FE,$1B,$F2,$FE,$65,$BF,$C7,$35,$FC
	.dc.b	$03,$F6,$7E,$17,$F3,$BE,$67,$96,$7F,$11,$9D,$F8,$86,$DE,$A9,$FB
	.dc.b	$27,$D5,$3F,$21,$D1,$F5,$5F,$9F,$F0,$DF,$A8,$F3,$3F,$0C,$D8,$FC
	.dc.b	$B3,$5F,$AB,$F0,$DF,$AE,$F3,$7F,$19,$F0,$DF,$B0,$F3,$3F,$1D,$F0
	.dc.b	$D6,$FF,$3C,$D7,$EE,$CF,$ED,$BC,$CF,$CA,$7C,$2F,$FB,$17,$F3,$FF
	.dc.b	$F2,$67,$7B,$E1,$F1,$E5,$FE,$65,$B8,$FE,$A6,$F8,$3F,$9D,$F8,$7E
	.dc.b	$5E,$3F,$E7,$FE,$1E,$CF,$EB,$DA,$FE,$3A,$7C,$1E,$4F,$FA,$DF,$E1
	.dc.b	$FF,$C8,$9A,$7F,$92,$2B,$FD,$51,$18,$F2,$B5,$E7,$F9,$DE,$11,$E9
	.dc.b	$E7,$F5,$D7,$FB,$2A,$FD,$7A,$FA,$23,$F3,$AD,$F5,$FF,$DA,$3C,$0D
	.dc.b	$FD,$75,$E0,$F8,$7F,$89,$E6,$F9,$5B,$19,$3B,$47,$A7,$E1,$FE,$57
	.dc.b	$9B,$EB,$F8,$7F,$99,$E6,$FB,$7E,$1E,$E1,$EC,$6B,$FC,$E3,$81,$5F
	.dc.b	$4F,$CE,$F5,$CF,$FA,$BB,$83,$D6,$FB,$7A,$3E,$A3,$4F,$A8,$7C,$13
	.dc.b	$E6,$7C,$3F,$DD,$F3,$2C,$7C,$5F,$2E,$BD,$5F,$A5,$F0,$FB,$8F,$A3
	.dc.b	$F0,$DF,$85,$F3,$3E,$B7,$C3,$D9,$FA,$AD,$7D,$53,$F8,$BF,$33,$EE
	.dc.b	$7C,$3F,$E3,$7C,$8F,$BD,$F0,$FF,$8F,$F2,$6E,$1F,$5D,$DB,$F6,$67
	.dc.b	$F0,$9F,$0D,$F9,$5F,$1E,$CF,$DD,$F2,$2B,$F1,$5F,$C5,$7C,$3E,$E3
	.dc.b	$F1,$3E,$EF,$E7,$7B,$FF,$1C,$FE,$7F,$C2,$B3,$F8,$77,$F2,$5E,$F6
	.dc.b	$E3,$F2,$3B,$7A,$5F,$C9,$ED,$E9,$69,$E9,$36,$7A,$4F,$CC,$FC,$3F
	.dc.b	$EB,$7C,$9F,$CF,$7C,$3D,$9F,$CE,$B5,$FC,$14,$FE,$CF,$CB,$FD,$27
	.dc.b	$C3,$A6,$13,$D3,$7E,$80,$F3,$77,$3A,$38,$4E,$A3,$31,$FE,$F8,$9E
	.dc.b	$0F,$EB,$BE,$23,$83,$C2,$3F,$D6,$4F,$B0,$FC,$97,$37,$D8,$7F,$29
	.dc.b	$CF,$5F,$27,$82,$8F,$B7,$1C,$0F,$DD,$FC,$37,$67,$FD,$8F,$7F,$EB
	.dc.b	$8F,$37,$C6,$7F,$B0,$67,$59,$FB,$BC,$AF,$E0,$D1,$8A,$FF,$35,$DF
	.dc.b	$C5,$BE,$3F,$A0,$7F,$9C,$23,$AF,$9B,$F2,$27,$3F,$0F,$C7,$C9,$C7
	.dc.b	$07,$1C,$3F,$CD,$D8,$7C,$1F,$2F,$2B,$04,$EA,$3D,$33,$A9,$CF,$C3
	.dc.b	$F5,$9F,$25,$F5,$F8,$6F,$47,$95,$BC,$7F,$CE,$55,$F4,$38,$3E,$DC
	.dc.b	$E7,$E1,$F7,$3C,$8F,$15,$7F,$85,$9E,$1B,$BC,$9E,$9F,$0F,$F1,$7C
	.dc.b	$8B,$5A,$6B,$F8,$17,$C5,$F0,$FF,$2F,$CA,$F2,$7B,$DF,$37,$C2,$F3
	.dc.b	$7C,$1F,$CF,$F0,$EE,$3E,$07,$6F,$8A,$7D,$9C,$3E,$88,$FF,$29,$ED
	.dc.b	$EE,$7F,$21,$C1,$ED,$7D,$2F,$07,$D7,$B7,$C5,$F8,$7F,$BD,$E6,$59
	.dc.b	$F7,$F9,$55,$FC,$D3,$F2,$FE,$1F,$71,$F1,$FA,$3E,$BF,$C3,$7E,$1F
	.dc.b	$CB,$FA,$1F,$0F,$F8,$9F,$2E,$DF,$CA,$7E,$A5,$6E,$39,$6D,$F3,$DF
	.dc.b	$49,$8E,$9F,$76,$3F,$90,$6F,$5F,$DF,$9F,$59,$CB,$E1,$FF,$31,$E5
	.dc.b	$7E,$07,$E1,$FF,$35,$E6,$5A,$FB,$6F,$E1,$6B,$71,$E2,$F8,$7F,$D0
	.dc.b	$79,$9F,$89,$6C,$64,$FC,$1C,$7E,$37,$E1,$FF,$4D,$E6,$FE,$3F,$E1
	.dc.b	$BF,$51,$E5,$7E,$47,$E1,$AE,$1E,$E6,$BF,$62,$7F,$63,$ED,$FF,$9E
	.dc.b	$E3,$E1,$FC,$36,$3C,$CF,$CD,$36,$32,$7E,$6E,$53,$F9,$FF,$86,$E7
	.dc.b	$E5,$FE,$9F,$F4,$0F,$F3,$8F,$98,$BF,$9D,$CF,$FC,$0A,$73,$18,$E6
	.dc.b	$FD,$F4,$63,$C3,$E5,$EF,$79,$BA,$7D,$DA,$FE,$93,$87,$DC,$D3,$DC
	.dc.b	$53,$DC,$79,$FE,$27,$C9,$F1,$1F,$03,$CE,$B3,$EB,$F3,$6B,$FC,$B7
	.dc.b	$7F,$44,$7A,$FE,$23,$84,$FE,$47,$5F,$F0,$D3,$BD,$FF,$DB,$2F,$FA
	.dc.b	$DA,$32,$FF,$3B,$8F,$F2,$43,$9F,$88,$FA,$3E,$6F,$C7,$F8,$8F,$A7
	.dc.b	$E7,$FB,$AB,$F9,$D6,$FE,$E6,$9E,$E2,$9E,$E3,$C5,$BF,$D1,$F8,$7F
	.dc.b	$BB,$E6,$7D,$3F,$87,$FB,$FE,$5D,$BF,$B4,$ED,$FB,$93,$F6,$7E,$1F
	.dc.b	$F0,$BE,$4F,$DB,$DF,$F0,$FE,$5D,$BF,$C1,$3B,$7F,$09,$3F,$81,$F8
	.dc.b	$6F,$C6,$F8,$F6,$7F,$10,$D7,$E4,$3F,$87,$F8,$6F,$C9,$F9,$5F,$89
	.dc.b	$F8,$6F,$CA,$F9,$5F,$8B,$F8,$6B,$87,$D1,$6B,$F3,$E7,$F3,$7E,$57
	.dc.b	$E4,$7E,$03,$F3,$BE,$4F,$E4,$FD,$FF,$CF,$F9,$77,$0F,$CD,$3B,$7F
	.dc.b	$39,$3E,$57,$EC,$6B,$5F,$ED,$FC,$28,$C4,$3C,$0F,$CF,$47,$AE,$3F
	.dc.b	$4D,$B7,$E8,$FE,$1B,$F5,$FE,$67,$E9,$7E,$1A,$D7,$E2,$DA,$FE,$26
	.dc.b	$7F,$6B,$E6,$7E,$AF,$E1,$BF,$B3,$FE,$83,$FE,$01,$F3,$7F,$55,$18
	.dc.b	$BD,$1F,$D3,$E7,$5E,$6F,$EB,$3E,$1F,$8F,$9B,$E8,$DF,$CF,$7F,$9B
	.dc.b	$EA,$FA,$ED,$3E,$B9,$F9,$FE,$AF,$AA,$FE,$87,$E1,$FC,$7E,$6F,$EF
	.dc.b	$FE,$1E,$CF,$E9,$5A,$FE,$8E,$7D,$3E,$67,$F6,$BF,$D0,$7F,$C2,$B3
	.dc.b	$FA,$FD,$BF,$57,$C1,$FE,$06,$9C,$CF,$F8,$8E,$5E,$57,$1F,$89,$F8
	.dc.b	$6F,$3E,$1F,$8B,$E6,$F5,$F8,$7F,$8F,$E5,$5C,$7E,$27,$95,$E1,$F8
	.dc.b	$8F,$96,$E6,$3C,$53,$E5,$F8,$8F,$9F,$E5,$79,$FE,$22,$D7,$F5,$1D
	.dc.b	$FF,$CF,$6F,$CF,$9D,$3F,$47,$C1,$F7,$7C,$1F,$D8,$E8,$F8,$1F,$07
	.dc.b	$F6,$BA,$73,$7F,$5D,$CB,$F6,$67,$C8,$FC,$1D,$2B,$FE,$40,$E4,$6D
	.dc.b	$EB,$8F,$69,$E5,$E0,$F8,$2F,$9F,$E1,$FF,$13,$E5,$7D,$17,$F1,$7E
	.dc.b	$5F,$D3,$F8,$7B,$5B,$3F,$15,$5F,$5F,$E1,$BF,$23,$E6,$7D,$9F,$87
	.dc.b	$B5,$F0,$E6,$BF,$E2,$BE,$9F,$F1,$C7,$2A,$BF,$33,$E6,$FA,$AB,$F5
	.dc.b	$1B,$B8,$31,$E1,$23,$DB,$3F,$3F,$AB,$EB,$BE,$D3,$F3,$D5,$F8,$BF
	.dc.b	$87,$FD,$2F,$91,$F8,$DF,$87,$FD,$3F,$95,$9B,$FD,$23,$97,$F4,$D3
	.dc.b	$F9,$3F,$87,$FD,$6F,$95,$F9,$5F,$87,$B5,$F7,$E6,$BE,$D3,$FC,$91
	.dc.b	$C0,$9F,$DA,$F9,$7F,$9D,$F8,$6F,$DB,$F9,$3F,$9F,$F8,$6B,$87,$E2
	.dc.b	$66,$BC,$FC,$96,$2A,$FD,$FF,$97,$FA,$7F,$86,$FF,$33,$FE,$67,$FA
	.dc.b	$8B,$FD,$34,$63,$CC,$FD,$67,$C3,$DC,$BF,$2D,$3C,$3A,$FF,$61,$EF
	.dc.b	$72,$F0,$ED,$7F,$95,$72,$FF,$3B,$57,$F9,$BF,$D0,$FF,$37,$76,$27
	.dc.b	$35,$F0,$30,$F1,$F5,$FB,$E3,$F8,$87,$17,$BD,$A7,$BC,$A7,$BF,$87
	.dc.b	$DC,$FF,$17,$E0,$FF,$37,$65,$FE,$71,$CB,$C1,$E6,$E3,$E1,$FE,$0B
	.dc.b	$AF,$87,$B8,$7F,$90,$EF,$86,$AF,$3F,$9F,$C3,$FC,$8F,$3B,$AF,$C3
	.dc.b	$FC,$AF,$3F,$38,$FC,$17,$2F,$C3,$9F,$63,$FC,$57,$83,$D8,$D3,$D8
	.dc.b	$53,$D9,$D3,$F6,$1F,$27,$C3,$FD,$8F,$37,$DA,$FD,$AF,$33,$DF,$F0
	.dc.b	$F6,$B6,$7D,$3A,$F8,$7F,$0B,$F8,$0F,$33,$E2,$FC,$37,$E0,$BC,$DC
	.dc.b	$DF,$D9,$72,$FD,$B9,$F9,$7F,$09,$F8,$8F,$2F,$E6,$FC,$2D,$AF,$91
	.dc.b	$35,$FC,$97,$32,$7F,$1F,$E6,$7D,$4F,$84,$FC,$8F,$95,$F5,$BE,$17
	.dc.b	$F2,$7E,$5E,$71,$FC,$53,$97,$F1,$93,$F7,$3E,$07,$F3,$3E,$5D,$9F
	.dc.b	$C8,$73,$7E,$03,$E0,$7F,$3A,$E6,$3F,$C6,$5D,$CA,$F5,$C7,$E5,$F8
	.dc.b	$3D,$6D,$3D,$65,$3D,$7B,$7A,$9F,$C3,$7C,$2F,$EA,$BC,$AF,$C7,$BF
	.dc.b	$AC,$F2,$FF,$23,$F0,$B6,$B6,$7E,$96,$BF,$2B,$F0,$9F,$B2,$F2,$EC
	.dc.b	$FF,$AF,$7F,$33,$FF,$35,$3F,$2B,$E1,$BF,$DB,$B1,$8F,$47,$FD,$C3
	.dc.b	$D0,$C3,$95,$FA,$0F,$7E,$DE,$C6,$99,$07,$FC,$00,$E0,$46,$E3,$27
	.dc.b	$F8,$4E,$1D,$1F,$AC,$F8,$6F,$1F,$97,$6B,$D1,$E5,$7E,$C3,$E1,$BC
	.dc.b	$EE,$63,$FC,$C5,$AF,$F9,$F9,$FF,$58,$7E,$7F,$FE,$7C,$8F,$F3,$C4
	.dc.b	$7F,$BA,$63,$1E,$5F,$FB,$9D,$7C,$3F,$C2,$8C,$4F,$C4,$F3,$35,$C1
	.dc.b	$F1,$BC,$9E,$1F,$0F,$F2,$3C,$9C,$DF,$CD,$72,$FC,$F9,$ED,$F0,$DF
	.dc.b	$3B,$C9,$B3,$F4,$DF,$17,$C3,$E6,$3D,$33,$F5,$3C,$AF,$37,$BB,$F5
	.dc.b	$BC,$9B,$5F,$6D,$F5,$7C,$1E,$63,$E0,$CF,$DC,$F0,$BC,$B1,$F1,$B3
	.dc.b	$F2,$BF,$23,$4C,$BF,$53,$D4,$72,$27,$73,$98,$57,$C9,$F8,$7F,$C4
	.dc.b	$F9,$9F,$2F,$E1,$FF,$17,$E6,$E7,$0F,$C8,$39,$7F,$25,$3F,$47,$E1
	.dc.b	$BF,$23,$E6,$59,$FC,$B3,$F5,$7E,$1F,$31,$F7,$27,$F2,$FE,$6F,$A2
	.dc.b	$3F,$1F,$B7,$A3,$4F,$3C,$63,$5F,$53,$F8,$4E,$8F,$B0,$FE,$1B,$A3
	.dc.b	$EE,$3F,$88,$F8,$6F,$D1,$F9,$9F,$85,$6C,$7E,$33,$E1,$EB,$FD,$8F
	.dc.b	$BD,$3F,$AA,$F3,$3F,$19,$F0,$FF,$AC,$F2,$6D,$FE,$C7,$CB,$FC,$87
	.dc.b	$C3,$FE,$C1,$CC,$7E,$6A,$7F,$29,$F0,$BF,$B4,$F2,$6C,$FE,$E5,$FC
	.dc.b	$C7,$C3,$66,$3F,$3F,$3F,$BA,$F2,$7F,$39,$F0,$DF,$AE,$9F,$B9,$E1
	.dc.b	$7F,$9B,$9F,$D1,$7C,$36,$5F,$93,$19,$F4,$1C,$2B,$D5,$7F,$A2,$6B
	.dc.b	$52,$BC,$F4,$27,$F5,$75,$FB,$2F,$73,$30,$E7,$FD,$84,$63,$9F,$F6
	.dc.b	$5F,$03,$6F,$81,$E5,$C3,$A5,$FC,$28,$FD,$FF,$05,$7F,$28,$FC,$47
	.dc.b	$E3,$1F,$10,$FE,$41,$B7,$F1,$0D,$9F,$09,$FC,$59,$7E,$7F,$FE,$7A
	.dc.b	$AF,$87,$E5,$6E,$B5,$2F,$87,$F5,$CE,$69,$7C,$AF,$31,$D5,$E3,$5F
	.dc.b	$0B,$CC,$E1,$5E,$0F,$87,$F8,$BE,$5E,$32,$46,$31,$E4,$76,$87,$C3
	.dc.b	$E6,$3F,$C6,$91,$C1,$F9,$1C,$7F,$E5,$47,$F2,$DC,$FF,$87,$69,$F1
	.dc.b	$8F,$CD,$74,$67,$2F,$8B,$1F,$C9,$E7,$F1,$8F,$8A,$3E,$3F,$C3,$7D
	.dc.b	$CF,$2F,$D1,$F0,$D6,$B5,$97,$FA,$41,$FC,$97,$2F,$E1,$6B,$71,$E6
	.dc.b	$F3,$3E,$07,$C2,$FE,$1B,$C9,$F0,$E7,$31,$4E,$79,$3A,$FC,$37,$E3
	.dc.b	$3C,$BD,$C7,$D6,$9D,$65,$F1,$3E,$17,$F2,$3E,$3E,$A3,$E5,$46,$2B
	.dc.b	$CF,$F0,$DF,$96,$F2,$75,$A7,$E1,$2B,$57,$F3,$BE,$17,$F3,$7E,$4E
	.dc.b	$B6,$FF,$44,$4F,$D0,$F2,$71,$5A,$CF,$E1,$FC,$37,$E8,$A7,$87,$BF
	.dc.b	$FA,$4E,$8E,$5E,$F7,$E9,$BA,$B3,$97,$DA,$9C,$BF,$72,$3C,$1A,$F9
	.dc.b	$36,$F8,$CD,$CF,$F2,$FC,$1E,$B7,$E9,$C6,$6F,$F1,$5F,$0B,$FB,$4F
	.dc.b	$2F,$D5,$F0,$DF,$B6,$F2,$FD,$93,$9C,$3E,$DC,$E5,$FB,$BF,$0B,$FB
	.dc.b	$CF,$33,$E1,$7C,$35,$9C,$E5,$F9,$2F,$33,$2F,$E5,$2B,$D2,$FE,$5B
	.dc.b	$A6,$BF,$86,$F0,$FF,$95,$DF,$D4,$8F,$CD,$F5,$69,$FE,$37,$5F,$47
	.dc.b	$E1,$7F,$88,$FE,$59,$FB,$CB,$FC,$8C,$61,$FA,$DF,$0F,$AF,$2F,$86
	.dc.b	$1F,$C3,$6F,$83,$FA,$AF,$87,$E7,$E5,$7D,$D6,$E1,$FA,$1F,$CF,$BF
	.dc.b	$D1,$F7,$FB,$28,$FC,$FC,$EA,$FF,$D4,$D5,$FE,$85,$A7,$95,$F8,$C8
	.dc.b	$7C,$37,$A3,$CB,$C7,$C3,$FA,$AB,$53,$9C,$BC,$1F,$0D,$EE,$F1,$F9
	.dc.b	$36,$3C,$9F,$0D,$F0,$BC,$AC,$EB,$ED,$BF,$07,$BB,$F1,$BB,$FC,$5F
	.dc.b	$05,$71,$F6,$4F,$CA,$F0,$73,$87,$C5,$BF,$2B,$F1,$F5,$F2,$BF,$27
	.dc.b	$7C,$BF,$BE,$E1,$F7,$E3,$F1,$F7,$F5,$4F,$C4,$F8,$6F,$AF,$E5,$7C
	.dc.b	$37,$EC,$F9,$5F,$17,$E1,$B3,$87,$CF,$9C,$BF,$47,$B7,$EF,$F9,$5F
	.dc.b	$2D,$B1,$F5,$7E,$17,$F0,$7E,$5E,$73,$FB,$D7,$E6,$7F,$C3,$DA,$F9
	.dc.b	$9F,$AF,$AF,$9B,$C9,$FE,$23,$D5,$F6,$78,$7D,$F5,$ED,$E1,$FB,$5F
	.dc.b	$09,$F9,$0F,$23,$EE,$7C,$2F,$E4,$BC,$9C,$DF,$E3,$67,$2F,$E3,$E3
	.dc.b	$F0,$5F,$0B,$F9,$8F,$2B,$F0,$9F,$0B,$78,$FF,$1D,$3F,$9C,$F2,$F3
	.dc.b	$87,$E5,$EF,$F1,$5F,$09,$FA,$0F,$23,$F1,$9F,$09,$FA,$2F,$27,$37
	.dc.b	$F9,$D9,$CB,$F9,$F8,$FC,$97,$C1,$FE,$A3,$C8,$FC,$A7,$C2,$7F,$7A
	.dc.b	$FD,$09,$F1,$9A,$C3,$F4,$D5,$A7,$FB,$A4,$EB,$7F,$D2,$7C,$1F,$1F
	.dc.b	$33,$F3,$DF,$07,$CF,$B3,$F4,$13,$9A,$FD,$54,$E5,$FD,$67,$07,$AD
	.dc.b	$FD,$77,$07,$AD,$CE,$5E,$B7,$2F,$AD,$FE,$A9,$87,$F8,$5A,$1E,$FB
	.dc.b	$37,$9D,$FC,$F2,$B5,$39,$8C,$7C,$4F,$C1,$F3,$F5,$87,$C3,$9E,$1F
	.dc.b	$0B,$F1,$7C,$DB,$3F,$09,$E9,$F0,$DF,$27,$CF,$CC,$FF,$60,$D3,$FC
	.dc.b	$CF,$AF,$AA,$3F,$4D,$D1,$F5,$1F,$D4,$74,$7D,$87,$33,$EE,$72,$FB
	.dc.b	$BE,$1B,$EB,$79,$7E,$B6,$C7,$C2,$F8,$6F,$B5,$E6,$E6,$FE,$BD,$FC
	.dc.b	$1F,$86,$FB,$DE,$77,$C3,$F8,$9F,$C0,$79,$F9,$BF,$BB,$39,$7E,$FC
	.dc.b	$7C,$9F,$86,$FC,$37,$93,$67,$F0,$7E,$4F,$CD,$F8,$6C,$ED,$F5,$2F
	.dc.b	$F1,$93,$F4,$7E,$1B,$F1,$DE,$47,$D3,$F8,$6B,$7F,$6A,$7F,$25,$E5
	.dc.b	$67,$2F,$C7,$DF,$D9,$F7,$3F,$2D,$E1,$7D,$BF,$77,$F3,$1E,$1E,$6F
	.dc.b	$F2,$B3,$97,$F2,$F1,$E7,$7E,$A6,$F5,$FF,$50,$75,$36,$F5,$47,$E7
	.dc.b	$A7,$4F,$FA,$3A,$75,$AF,$E4,$3E,$1B,$F5,$1E,$5F,$E3,$5F,$D5,$79
	.dc.b	$7F,$8F,$F8,$5C,$D6,$F0,$FD,$4C,$FE,$4F,$E1,$7F,$61,$E5,$59,$FD
	.dc.b	$6B,$F9,$7F,$85,$FD,$A7,$97,$99,$FC,$64,$F9,$1F,$CF,$E8,$D7,$1E
	.dc.b	$57,$E8,$77,$F5,$D7,$C0,$D3,$EA,$3F,$61,$FD,$4F,$C2,$FF,$9E,$BF
	.dc.b	$43,$FF,$7C,$C7,$EB,$3E,$1B,$15,$FA,$9A,$D7,$9B,$FA,$B9,$CD,$7E
	.dc.b	$CB,$E1,$79,$79,$9F,$AF,$6C,$7E,$DB,$E1,$BB,$79,$99,$CB,$FC,$FD
	.dc.b	$9F,$FA,$16,$BF,$C9,$DF,$9F,$FF,$9C,$E7,$F7,$72,$F2,$BF,$C6,$B3
	.dc.b	$9D,$D3,$9B,$E3,$F0,$FE,$BF,$2F,$71,$ED,$C3,$8F,$C3,$FB,$FC,$AE
	.dc.b	$7F,$13,$6B,$C3,$3F,$0F,$CB,$CE,$BE,$6B,$F0,$FB,$BF,$1F,$C9,$F1
	.dc.b	$FB,$BF,$27,$CB,$CD,$FB,$27,$2F,$BA,$3C,$6F,$C0,$D2,$BF,$E6,$6D
	.dc.b	$FD,$4F,$FA,$33,$4F,$F4,$B7,$0F,$FA,$46,$3E,$3E,$FC,$1F,$F7,$C5
	.dc.b	$70,$F7,$FF,$CD,$35,$FE,$68,$9F,$9B,$E5,$FC,$8F,$88,$FB,$DE,$6F
	.dc.b	$CA,$F8,$8F,$C0,$4E,$6B,$E9,$CF,$CE,$F8,$7F,$C2,$79,$BF,$43,$E2
	.dc.b	$2E,$DF,$5A,$7F,$11,$E7,$67,$0F,$B7,$7E,$57,$FD,$EB,$B7,$96,$FD
	.dc.b	$67,$D3,$3E,$BC,$7B,$1F,$C0,$99,$7E,$11,$FE,$AE,$D5,$F1,$1F,$97
	.dc.b	$F3,$3F,$06,$FE,$67,$CC,$FC,$2F,$C4,$67,$0F,$C5,$CE,$5F,$C6,$FC
	.dc.b	$3F,$E7,$FC,$DF,$C5,$FC,$47,$E8,$7C,$EF,$C6,$CE,$6B,$F2,$73,$97
	.dc.b	$F2,$BF,$0D,$FA,$7F,$2B,$F2,$7F,$0D,$FA,$9F,$2B,$F2,$B3,$9A,$FC
	.dc.b	$DC,$E5,$FC,$EF,$C2,$FE,$BF,$CA,$FC,$DF,$C3,$7E,$C7,$CB,$FC,$EC
	.dc.b	$E6,$BF,$47,$39,$7F,$4B,$EF,$7F,$BC,$3A,$BF,$CB,$3D,$FF,$80,$F7
	.dc.b	$B1,$5F,$A6,$AD,$79,$7F,$A8,$9C,$CF,$AB,$A7,$F6,$3C,$1E,$A7,$F6
	.dc.b	$7C,$1E,$A7,$33,$EA,$72,$FE,$DB,$E1,$7C,$5E,$5F,$FA,$A3,$F3,$FF
	.dc.b	$F0,$4D,$7F,$BE,$75,$FF,$33,$3F,$6B,$57,$63,$0E,$38,$FF,$B3,$70
	.dc.b	$E3,$7F,$EC,$8C,$38,$E1,$FE,$C4,$97,$C3,$7C,$1F,$2F,$C1,$F0,$D7
	.dc.b	$1E,$53,$F1,$7C,$CC,$E5,$E3,$BF,$63,$FA,$FD,$FE,$03,$FB,$1D,$F2
	.dc.b	$FE,$B2,$32,$FE,$BA,$3D,$DF,$0B,$F4,$7C,$8F,$81,$F0,$D7,$3F,$54
	.dc.b	$FD,$5F,$2B,$38,$7C,$3B,$F8,$DF,$0D,$F6,$7C,$9B,$3F,$1F,$8A,$BF
	.dc.b	$06,$3E,$67,$C3,$7D,$F9,$F9,$DF,$0B,$F8,$1F,$16,$DF,$CF,$6B,$F2
	.dc.b	$7C,$6F,$A9,$F0,$DB,$DB,$E7,$7C,$37,$E2,$7C,$6F,$B0,$D8,$CD,$FD
	.dc.b	$9F,$1B,$EE,$7C,$36,$63,$ED,$75,$7F,$4F,$7E,$F7,$8F,$F8,$1F,$73
	.dc.b	$F2,$9E,$45,$9C,$E3,$F7,$7D,$DC,$BF,$7F,$E1,$7F,$35,$E4,$7E,$25
	.dc.b	$B1,$9A,$FC,$57,$91,$F8,$DF,$86,$CC,$78,$BE,$1B,$F4,$5E,$57,$E4
	.dc.b	$7E,$1B,$F4,$9E,$55,$AF,$C9,$4F,$E5,$6B,$38,$FE,$3F,$E1,$BF,$55
	.dc.b	$E4,$FE,$67,$E1,$98,$42,$7C,$03,$F4,$4F,$F3,$04,$7C,$29,$E3,$87
	.dc.b	$38,$E0,$72,$AE,$1F,$11,$FA,$5F,$87,$CC,$E7,$0F,$CD,$DF,$E8,$67
	.dc.b	$33,$FE,$5A,$9F,$F6,$3C,$E6,$38,$DF,$8F,$CC,$FD,$87,$C4,$70,$8F
	.dc.b	$37,$9C,$BF,$5F,$9F,$70,$2F,$CF,$FF,$CC,$79,$7F,$99,$AB,$67,$08
	.dc.b	$7C,$47,$59,$79,$BC,$EE,$56,$B0,$F2,$D7,$37,$F8,$04,$76,$AE,$18
	.dc.b	$75,$F8,$8E,$9E,$5E,$2F,$CB,$6F,$88,$CD,$F9,$27,$31,$E3,$F2,$F8
	.dc.b	$67,$EB,$C7,$75,$E0,$F8,$3F,$47,$97,$AC,$BF,$C1,$53,$E6,$8C,$4F
	.dc.b	$B7,$C9,$D7,$C1,$F0,$D3,$C1,$97,$C7,$F8,$6F,$85,$E5,$73,$6C,$EB
	.dc.b	$34,$7C,$7F,$27,$C1,$F0,$FA,$D7,$E1,$FC,$3F,$CB,$F1,$FF,$1D,$9F
	.dc.b	$CA,$8C,$4F,$36,$BF,$2B,$E1,$FE,$97,$95,$C3,$4F,$AD,$F0,$FF,$57
	.dc.b	$C9,$C5,$FC,$AF,$87,$FB,$1E,$4E,$33,$FC,$04,$62,$7D,$17,$C5,$F5
	.dc.b	$7C,$37,$DF,$F1,$F8,$71,$3E,$1F,$15,$EF,$6B,$F7,$AB,$F0,$DE,$5F
	.dc.b	$0F,$86,$FC,$47,$95,$CB,$E1,$AD,$FC,$76,$BF,$86,$9F,$C7,$79,$5E
	.dc.b	$0D,$3C,$93,$F8,$AD,$7D,$0D,$3F,$1F,$BE,$DF,$C8,$F0,$7A,$3E,$1B
	.dc.b	$F3,$1E,$55,$9F,$CA,$F9,$35,$FD,$BF,$93,$FB,$D8,$F7,$FC,$37,$09
	.dc.b	$FB,$9E,$FF,$EB,$7F,$45,$FD,$EF,$3F,$E8,$23,$13,$F1,$BE,$02,$BC
	.dc.b	$0F,$DB,$D7,$E9,$5F,$D1,$FC,$3F,$3F,$1F,$E6,$FC,$3D,$9F,$C5,$4D
	.dc.b	$7F,$53,$1E,$1F,$2B,$ED,$7C,$37,$F9,$87,$0F,$CD,$D7,$F6,$D8,$FF
	.dc.b	$23,$3E,$C8,$78,$FC,$3B,$FD,$4F,$2F,$87,$F6,$79,$5D,$30,$F7,$79
	.dc.b	$1D,$BE,$27,$E0,$79,$17,$3F,$5B,$B7,$DB,$3E,$4F,$86,$F8,$DE,$45
	.dc.b	$9F,$83,$E4,$57,$C8,$FA,$BE,$1F,$71,$FA,$EC,$BF,$CB,$0F,$A7,$BF
	.dc.b	$DF,$F0,$BF,$47,$C0,$B3,$F3,$1F,$87,$F0,$DF,$56,$BE,$2F,$C3,$7D
	.dc.b	$7F,$22,$DF,$D2,$F2,$7E,$4F,$C3,$7D,$B7,$31,$F1,$E7,$E6,$FC,$37
	.dc.b	$DF,$F2,$3E,$7F,$C3,$7E,$07,$CA,$CD,$FD,$C7,$2F,$DE,$9F,$AB,$F0
	.dc.b	$DF,$87,$F2,$3E,$BF,$C3,$7E,$27,$C6,$CD,$FE,$11,$CB,$F8,$69,$FB
	.dc.b	$BF,$0B,$F8,$FF,$1B,$EF,$FC,$2F,$E4,$7C,$7C,$DF,$E3,$1C,$BF,$8E
	.dc.b	$9F,$2B,$F8,$1C,$FC,$AF,$E0,$F4,$CB,$FD,$73,$C3,$46,$A3,$13,$F8
	.dc.b	$EF,$85,$FD,$07,$8F,$F9,$0F,$85,$B5,$F5,$E6,$BF,$91,$8F,$D3,$79
	.dc.b	$1F,$96,$F8,$4F,$EA,$9F,$A3,$7F,$51,$AF,$CB,$46,$3C,$9F,$CD,$FC
	.dc.b	$35,$9F,$C1,$4F,$0F,$27,$CF,$1F,$DA,$F6,$6B,$E7,$8C,$6D,$A5,$E7
	.dc.b	$AB,$DB,$1F,$8D,$E8,$FB,$8F,$E3,$FE,$1B,$CB,$E1,$FF,$25,$FD,$07
	.dc.b	$FC,$A3,$3F,$E0,$0D,$D3,$FB,$0A,$DC,$FF,$44,$AD,$BF,$DF,$E9,$E5
	.dc.b	$F3,$F8,$6F,$83,$E5,$7F,$56,$B7,$C3,$63,$8F,$95,$7C,$6F,$27,$53
	.dc.b	$9C,$BF,$CD,$16,$F8,$9F,$97,$E5,$73,$6E,$3E,$3F,$85,$F9,$FE,$56
	.dc.b	$70,$FF,$03,$EC,$C3,$C5,$F0,$BF,$53,$CA,$F2,$7C,$2D,$CB,$D9,$3F
	.dc.b	$63,$CB,$CE,$3C,$AF,$D5,$F0,$BF,$73,$CA,$B3,$D9,$F7,$7C,$2F,$E0
	.dc.b	$3A,$33,$3F,$CB,$B5,$FF,$5C,$E7,$F2,$78,$3D,$33,$EE,$6F,$B1,$FE
	.dc.b	$0D,$E2,$FF,$07,$CF,$AB,$87,$D8,$F3,$E1,$F2,$BD,$7C,$CF,$9D,$F0
	.dc.b	$BF,$90,$F3,$73,$3F,$6E,$72,$FC,$48,$FA,$9F,$0B,$F9,$6F,$33,$EB
	.dc.b	$7C,$2F,$E6,$3C,$DC,$DF,$CA,$9C,$BF,$32,$3C,$6F,$E1,$F2,$AF,$E0
	.dc.b	$DF,$C0,$74,$FD,$4E,$0F,$81,$3F,$52,$7C,$6F,$C5,$F2,$AB,$F9,$1F
	.dc.b	$2B,$D2,$FE,$36,$32,$FE,$3E,$3F,$19,$F0,$BF,$AE,$F2,$7F,$1D,$EF
	.dc.b	$DC,$7F,$2D,$3F,$B2,$F2,$F3,$87,$E1,$2F,$F2,$9F,$01,$FB,$6F,$1A
	.dc.b	$CF,$F9,$3B,$D5,$FF,$2F,$F4,$7D,$3F,$85,$D6,$7F,$A1,$C3,$FC,$A1
	.dc.b	$5A,$8F,$F2,$7F,$4B,$E1,$33,$E3,$FE,$87,$E1,$2D,$7E,$A6,$78,$F8
	.dc.b	$F9,$D3,$C1,$7F,$A7,$F8,$4E,$BD,$F6,$7C,$8F,$EA,$FE,$13,$C3,$D1
	.dc.b	$99,$FE,$F7,$AF,$F9,$47,$83,$F1,$5B,$FA,$27,$F5,$6F,$A3,$E2,$3F
	.dc.b	$69,$C3,$9A,$DD,$7C,$19,$FD,$C7,$C2,$FB,$BC,$AF,$F2,$7F,$E8,$3F
	.dc.b	$E7,$6A,$FE,$1F,$93,$CC,$DC,$7C,$9A,$DB,$F1,$A3,$8F,$C4,$FC,$8F
	.dc.b	$2F,$9F,$C3,$DE,$6F,$F2,$7E,$5F,$E0,$7C,$3F,$D2,$39,$3C,$BD,$C7
	.dc.b	$D5,$C3,$EB,$BF,$AD,$EA,$AF,$0D,$BD,$AF,$F0,$28,$CB,$FE,$95,$BF
	.dc.b	$B0,$FF,$A1,$30,$FF,$38,$3F,$81,$F2,$3D,$FF,$0F,$F7,$9F,$83,$F0
	.dc.b	$DF,$80,$F2,$33,$5B,$D7,$D3,$F0,$FF,$84,$F2,$7E,$3F,$C3,$59,$CD
	.dc.b	$7E,$37,$C4,$CB,$F8,$FA,$F9,$BF,$0F,$F8,$CF,$1B,$E7,$FC,$3D,$C3
	.dc.b	$E2,$CF,$E4,$3B,$33,$87,$E5,$EF,$EA,$FC,$3F,$E5,$3B,$6C,$FE,$6F
	.dc.b	$C4,$FB,$3C,$99,$CB,$E7,$DF,$E6,$A7,$EE,$FC,$3F,$E7,$3C,$6F,$BF
	.dc.b	$F0,$D6,$FE,$AC,$FE,$83,$C5,$CE,$5F,$A7,$BF,$4B,$E3,$AF,$B7,$BF
	.dc.b	$A6,$3F,$D6,$31,$FE,$B4,$E8,$70,$7A,$E3,$73,$C4,$D9,$F5,$7E,$1F
	.dc.b	$F5,$FE,$2F,$E4,$BE,$02,$D7,$E2,$27,$F6,$7E,$26,$6F,$F7,$57,$F9
	.dc.b	$8F,$80,$FF,$63,$7E,$8F,$FE,$DE,$C3,$F1,$FF,$0F,$8F,$17,$58,$7F
	.dc.b	$B3,$AB,$4F,$8B,$C8,$6C,$BE,$18,$7D,$4A,$F4,$57,$FB,$16,$9B,$FC
	.dc.b	$18,$C6,$FE,$FA,$E9,$5A,$78,$DF,$C5,$F8,$8F,$3F,$8F,$FB,$06,$C7
	.dc.b	$E9,$BE,$1F,$D7,$E2,$66,$FD,$17,$FB,$6F,$88,$F7,$F8,$96,$7D,$9E
	.dc.b	$1F,$EE,$BE,$23,$39,$7E,$C2,$FE,$2C,$FE,$FB,$E2,$3E,$3F,$8B,$FE
	.dc.b	$CE,$FD,$0B,$FC,$35,$88,$FF,$7F,$72,$7D,$AC,$7F,$73,$7C,$1F,$F6
	.dc.b	$4D,$70,$AD,$F0,$F3,$F8,$9F,$A7,$C9,$D7,$E1,$EE,$1C,$27,$EB,$F8
	.dc.b	$59,$C3,$EA,$5F,$8F,$E1,$FE,$DF,$89,$E5,$F8,$7F,$BB,$E1,$E6,$FE
	.dc.b	$D4,$E5,$FB,$91,$ED,$7E,$F7,$07,$B7,$3F,$C2,$EB,$99,$FF,$52,$F0
	.dc.b	$7F,$B7,$27,$D6,$68,$F6,$FC,$3F,$E3,$3C,$5F,$8C,$D8,$F8,$3E,$EF
	.dc.b	$E4,$3C,$3C,$E1,$F8,$DB,$F9,$9F,$05,$F9,$4F,$17,$E7,$7C,$35,$C7
	.dc.b	$E4,$CF,$E6,$3C,$4C,$E1,$F9,$5B,$F3,$3F,$4C,$FF,$20,$E9,$E7,$7F
	.dc.b	$31,$CF,$F4,$5C,$ED,$C1,$FB,$3B,$FC,$17,$F4,$11,$F7,$FE,$1F,$F4
	.dc.b	$FE,$25,$9F,$D2,$3F,$83,$F8,$7F,$D5,$F8,$79,$9F,$8F,$3F,$87,$F8
	.dc.b	$7F,$D7,$F8,$96,$7F,$59,$E0,$7E,$2F,$E1,$B3,$C5,$F8,$6B,$FD,$AC
	.dc.b	$FE,$3F,$E1,$BF,$6F,$E1,$FE,$47,$E1,$AD,$FE,$32,$7F,$77,$E1,$E7
	.dc.b	$2F,$DC,$5F,$E5,$FE,$1B,$E2,$47,$C5,$F8,$97,$E9,$0C,$75,$3A,$3F
	.dc.b	$CD,$31,$98,$FC,$CF,$C4,$66,$73,$E3,$70,$7F,$DD,$35,$B7,$E7,$FC
	.dc.b	$46,$F0,$EF,$E4,$7E,$A6,$31,$5F,$E6,$CC,$7F,$4B,$F0,$F8,$F1,$F5
	.dc.b	$8F,$8F,$1E,$35,$EF,$8F,$6B,$FA,$CE,$0F,$87,$1F,$7B,$83,$E1,$CE
	.dc.b	$63,$67,$E0,$BE,$23,$C3,$E4,$7E,$ED,$B3,$FB,$6F,$7B,$CB,$E3,$E6
	.dc.b	$FC,$5A,$B0,$F9,$11,$FC,$8F,$CF,$FF,$40,$BF,$23,$C8,$CF,$A3,$27
	.dc.b	$3B,$7D,$0F,$1B,$3A,$7F,$09,$A5,$3E,$1B,$E2,$E1,$AC,$55,$AF,$80
	.dc.b	$F9,$38,$FE,$16,$7D,$B3,$E0,$F1,$5C,$5F,$12,$31,$C5,$F1,$2B,$F4
	.dc.b	$73,$9D,$3C,$5A,$7F,$A5,$27,$DF,$E4,$F8,$70,$F2,$FC,$37,$D8,$8C
	.dc.b	$56,$EB,$F0,$3E,$2E,$2B,$58,$7B,$4E,$BF,$0D,$F8,$0F,$1B,$8F,$C3
	.dc.b	$7E,$0B,$C3,$E7,$39,$AE,$18,$FC,$0C,$97,$B9,$CE,$9F,$23,$E1,$BF
	.dc.b	$17,$E2,$F9,$1B,$3A,$B7,$C2,$FE,$43,$87,$59,$7D,$CC,$3D,$7F,$0D
	.dc.b	$F9,$4E,$2B,$3F,$80,$7D,$FF,$0D,$F9,$8F,$17,$33,$FA,$09,$F8,$7F
	.dc.b	$0D,$F9,$CF,$1B,$E2,$FC,$2F,$E7,$BC,$5C,$DF,$E2,$A7,$2F,$E3,$23
	.dc.b	$C6,$FE,$AB,$3A,$FE,$DF,$C0,$47,$D0,$DB,$D5,$C1,$ED,$E1,$F7,$DF
	.dc.b	$E4,$FE,$1B,$F5,$BE,$37,$D6,$6D,$7C,$3F,$85,$FD,$8F,$8D,$9B,$FC
	.dc.b	$ED,$FD,$CF,$86,$FD,$AF,$8F,$67,$F4,$2F,$E0,$3E,$1B,$F7,$9F,$A4
	.dc.b	$FF,$92,$3A,$BF,$33,$7C,$33,$F2,$61,$F3,$5A,$FB,$5F,$A9,$B7,$AA
	.dc.b	$3F,$55,$B7,$AA,$FF,$51,$F0,$FD,$BC,$7F,$C7,$BE,$0F,$88,$FC,$8F
	.dc.b	$C3,$E7,$2F,$BD,$39,$7F,$01,$F1,$1E,$6F,$13,$F2,$FF,$0F,$E8,$F0
	.dc.b	$FF,$33,$39,$AF,$C3,$4E,$5F,$C4,$7C,$47,$BB,$C4,$FD,$17,$E8,$9F
	.dc.b	$E7,$FF,$39,$C9,$F9,$4B,$DF,$17,$F1,$78,$FF,$41,$47,$E5,$AB,$84
	.dc.b	$6E,$7F,$C5,$51,$FE,$7D,$7F,$C1,$55,$D3,$8F,$BF,$C3,$FC,$E7,$C3
	.dc.b	$F1,$3F,$43,$C6,$CD,$F7,$9F,$2F,$91,$F5,$3C,$7B,$3E,$3F,$17,$D3
	.dc.b	$F1,$19,$DB,$C9,$7F,$6A,$7F,$77,$C1,$EF,$9F,$45,$D3,$F8,$F4,$7E
	.dc.b	$03,$AB,$E1,$C6,$7A,$BF,$8F,$F0,$3E,$23,$F0,$BE,$3F,$C8,$F7,$EE
	.dc.b	$5F,$0E,$7F,$13,$E2,$67,$1F,$91,$7F,$3B,$DF,$FC,$6F,$8B,$67,$E6
	.dc.b	$3F,$4B,$E1,$FF,$23,$E1,$66,$7F,$AE,$F4,$7D,$4A,$F5,$C7,$D3,$EA
	.dc.b	$F3,$B7,$4F,$AE,$F0,$EA,$FA,$EF,$D4,$3D,$91,$FC,$D7,$77,$C4,$7E
	.dc.b	$83,$C4,$FC,$1B,$EA,$9F,$BD,$F1,$19,$7F,$01,$19,$BF,$BF,$F1,$1F
	.dc.b	$A8,$F0,$FF,$16,$DA,$FC,$33,$FA,$CF,$13,$F1,$FF,$0F,$9B,$FC,$3F
	.dc.b	$C4,$7E,$C3,$C4,$FC,$9F,$C3,$DC,$7F,$1C,$FE,$D3,$C5,$FC,$BD,$66
	.dc.b	$FF,$1F,$F1,$1F,$B8,$EE,$FC,$DF,$C3,$5C,$7F,$2C,$FF,$77,$F0,$3F
	.dc.b	$CF,$FD,$FF,$72,$F5,$87,$E6,$3E,$1F,$5E,$3F,$E9,$1E,$1E,$45,$9F
	.dc.b	$EF,$18,$2B,$F5,$3F,$0F,$AC,$FF,$49,$AF,$97,$3F,$C4,$EF,$67,$E8
	.dc.b	$3E,$B7,$78,$FE,$B7,$83,$F6,$7F,$0F,$E7,$F1,$BF,$6B,$F0,$D6,$FF
	.dc.b	$60,$FA,$FC,$8D,$E1,$E4,$AF,$DD,$FC,$3F,$BF,$C7,$B3,$E8,$F2,$2B
	.dc.b	$F8,$CE,$1F,$CE,$65,$FD,$43,$F4,$4F,$F2,$57,$88,$9F,$F2,$77,$07
	.dc.b	$E8,$B8,$7D,$58,$7F,$78,$F0,$5C,$58,$A4,$62,$7F,$84,$75,$7D,$87
	.dc.b	$DD,$9F,$7F,$88,$FA,$9E,$47,$F7,$3D,$51,$F5,$FC,$8C,$3A,$AF,$37
	.dc.b	$C4,$65,$F4,$7C,$47,$DD,$F1,$39,$3F,$7F,$C5,$E9,$F0,$F9,$AD,$E9
	.dc.b	$F6,$A7,$C1,$F1,$1F,$85,$F1,$AC,$FD,$EF,$17,$C9,$F1,$19,$C3,$E3
	.dc.b	$5F,$E2,$E7,$D1,$F0,$FF,$8D,$E6,$B8,$7E,$19,$F6,$7C,$37,$E4,$79
	.dc.b	$F3,$3F,$46,$7E,$07,$C3,$7E,$57,$C7,$B3,$F8,$EF,$23,$E2,$7C,$36
	.dc.b	$75,$FA,$D7,$F9,$B9,$F3,$D7,$DB,$DE,$BF,$E6,$3E,$C6,$DE,$A8,$FC
	.dc.b	$CC,$E9,$FE,$6B,$3A,$C3,$F0,$5F,$0D,$FA,$7F,$2B,$EA,$3F,$A9,$F2
	.dc.b	$3E,$B7,$C0,$66,$B7,$87,$E9,$27,$ED,$7C,$37,$EB,$FC,$9B,$3F,$A8
	.dc.b	$7E,$F7,$C2,$FE,$CF,$CA,$CC,$FE,$4E,$7C,$8B,$C5,$FC,$96,$D8,$7F
	.dc.b	$29,$BF,$AE,$7E,$06,$7F,$51,$FB,$0F,$E6,$FE,$1B,$F8,$07,$93,$F8
	.dc.b	$D6,$C7,$E7,$FE,$1B,$FC,$F5,$FA,$57,$F9,$F7,$C9,$FF,$2E,$F8,$CB
	.dc.b	$DE,$3F,$E5,$6C,$3F,$CB,$92,$F8,$7E,$3E,$37,$E6,$7E,$17,$9F,$8F
	.dc.b	$F9,$B9,$CD,$FE,$B6,$72,$FE,$BF,$E1,$BC,$3E,$3F,$E8,$7E,$1A,$CE
	.dc.b	$73,$FF,$43,$6D,$FE,$75,$8F,$DB,$4E,$A7,$C8,$7F,$9A,$BF,$45,$FF
	.dc.b	$39,$D3,$C6,$FF,$2F,$47,$F9,$C6,$79,$F8,$F8,$F8,$AF,$83,$5A,$F8
	.dc.b	$8F,$87,$3C,$3E,$23,$E2,$F9,$5C,$BE,$23,$E3,$F9,$79,$CB,$CB,$39
	.dc.b	$7C,$F1,$FB,$BE,$0F,$37,$3F,$CC,$DE,$F0,$7C,$A9,$FA,$3E,$6F,$C1
	.dc.b	$C3,$DF,$C5,$E8,$F8,$8F,$AB,$E4,$D9,$F8,$6F,$B3,$E0,$3E,$CF,$97
	.dc.b	$99,$FE,$CF,$CC,$AF,$83,$F0,$BF,$7B,$C9,$B5,$F3,$3C,$9F,$8B,$F0
	.dc.b	$F9,$DF,$FC,$E3,$D4,$C3,$F0,$B5,$EB,$BF,$2D,$7D,$26,$9F,$4B,$89
	.dc.b	$DE,$E2,$F6,$D6,$EB,$CB,$C3,$F5,$3E,$23,$F2,$3E,$55,$9F,$BF,$E5
	.dc.b	$7D,$8F,$88,$CE,$1E,$4B,$FC,$BC,$FD,$CF,$3B,$F3,$3E,$2F,$DE,$F8
	.dc.b	$7F,$CD,$F8,$F9,$CB,$F1,$33,$97,$F1,$71,$F8,$4F,$88,$FD,$0F,$8D
	.dc.b	$67,$F1,$EF,$E2,$3E,$1F,$F4,$BE,$3E,$67,$E6,$CF,$E3,$3E,$23,$F5
	.dc.b	$3C,$56,$7F,$2F,$E0,$7E,$43,$E1,$F3,$C3,$F3,$2F,$F5,$F3,$F9,$4F
	.dc.b	$87,$FF,$75,$7E,$97,$FE,$6F,$EB,$7C,$46,$3C,$AD,$67,$FE,$E7,$AD
	.dc.b	$4E,$72,$FA,$FC,$1E,$A6,$D7,$A9,$FB,$7C,$1E,$A7,$31,$EA,$7E,$FF
	.dc.b	$C2,$F8,$BC,$AF,$D3,$FC,$07,$93,$CB,$FD,$4C,$E6,$BF,$0B,$39,$7F
	.dc.b	$0F,$F1,$1E,$AF,$27,$FD,$99,$7F,$E6,$BD,$7F,$CC,$9E,$0F,$FB,$A3
	.dc.b	$81,$7B,$C3,$FD,$7F,$7B,$F2,$7E,$76,$A3,$FC,$D3,$E4,$7F,$0F,$AF
	.dc.b	$4C,$64,$FE,$4D,$19,$7F,$25,$83,$7F,$81,$18,$E0,$F8,$CF,$E5,$F1
	.dc.b	$FF,$3B,$3F,$CB,$F4,$FF,$23,$E9,$FE,$FE,$97,$C4,$7D,$6F,$1B,$DB
	.dc.b	$F1,$1F,$63,$C8,$F7,$CE,$76,$F5,$FC,$3F,$DC,$F1,$3E,$1B,$63,$3A
	.dc.b	$7D,$BF,$17,$E3,$FC,$46,$63,$C9,$F1,$1F,$84,$F1,$BE,$5B,$6F,$E3
	.dc.b	$BF,$88,$F1,$BE,$7F,$C4,$66,$FD,$9F,$0F,$F8,$CE,$4F,$A6,$DC,$3E
	.dc.b	$7B,$F9,$0E,$5F,$AF,$F0,$F9,$BF,$89,$F0,$FF,$94,$F1,$3E,$DF,$C3
	.dc.b	$DC,$7E,$BB,$F9,$8F,$1B,$EF,$D6,$6F,$D2,$7E,$5F,$7F,$4B,$6F,$D2
	.dc.b	$FE,$6F,$6F,$4B,$93,$EA,$7B,$FF,$A4,$F1,$BF,$16,$FE,$9B,$C5,$B3
	.dc.b	$FA,$39,$FC,$7F,$BD,$9C,$3F,$17,$F0,$DF,$AC,$F1,$BF,$26,$FE,$BB
	.dc.b	$C6,$B8,$7E,$AE,$7F,$2F,$F0,$F9,$BF,$2C,$FF,$A4,$3C,$2F,$F2,$14
	.dc.b	$62,$3F,$21,$F0,$29,$CC,$7C,$03,$86,$DE,$C7,$96,$FF,$A4,$F8,$7F
	.dc.b	$DF,$F9,$16,$7F,$34,$7F,$B1,$BF,$4C,$FF,$3A,$E5,$FA,$1F,$19,$E3
	.dc.b	$30,$E3,$3F,$A8,$F8,$8E,$38,$71,$3F,$63,$F1,$18,$7F,$C1,$58,$FF
	.dc.b	$9D,$EB,$F5,$FF,$11,$9D,$D3,$F9,$E9,$CD,$74,$9C,$F9,$1D,$23,$15
	.dc.b	$B7,$F5,$5F,$0F,$CE,$1E,$37,$F0,$28,$C3,$67,$8F,$8B,$FC,$1B,$E2
	.dc.b	$33,$97,$F0,$2F,$88,$E5,$E3,$FF,$0A,$7A,$78,$F7,$0F,$0C,$AB,$E3
	.dc.b	$1F,$1F,$F4,$7F,$F3,$CE,$9F,$E7,$88,$C7,$C4,$F9,$7C,$8E,$18,$7F
	.dc.b	$57,$B7,$C4,$E6,$F9,$CB,$9D,$5A,$8F,$83,$E4,$72,$99,$3F,$E7,$89
	.dc.b	$F8,$B8,$7B,$DF,$F6,$CE,$5F,$13,$0F,$E3,$FC,$3F,$02,$7E,$49,$F1
	.dc.b	$23,$F9,$1E,$F9,$C7,$AF,$BD,$F3,$7C,$6F,$05,$7B,$FE,$13,$E8,$79
	.dc.b	$38,$C7,$D1,$AA,$78,$57,$B3,$1F,$5E,$3C,$23,$15,$FE,$27,$BF,$BB
	.dc.b	$E4,$73,$BF,$83,$F0,$FF,$77,$C7,$C5,$A3,$F0,$15,$8F,$88,$D6,$3E
	.dc.b	$EC,$7F,$0F,$D3,$F0,$DF,$D4,$F0,$FC,$37,$E0,$9F,$1B,$A7,$F0,$8F
	.dc.b	$0A,$F2,$FC,$47,$E3,$7C,$9F,$0F,$C4,$7E,$3F,$C8,$B7,$F9,$49,$F2
	.dc.b	$D6,$6F,$DB,$F0,$FF,$95,$F1,$3D,$2F,$E5,$FC,$4B,$87,$E7,$27,$DB
	.dc.b	$F1,$19,$BF,$8B,$F1,$1F,$9D,$F1,$BE,$0B,$63,$37,$FA,$4F,$1B,$E2
	.dc.b	$FC,$46,$63,$F1,$DF,$0F,$FA,$5E,$FF,$92,$DB,$FA,$2F,$EA,$7B,$FE
	.dc.b	$6F,$C3,$E6,$FF,$2D,$F0,$FF,$AD,$F2,$3E,$8F,$C3,$DC,$7E,$CB,$FB
	.dc.b	$BD,$FF,$CF,$51,$FA,$3A,$D5,$FE,$7F,$6F,$27,$7F,$E6,$1B,$5F,$A9
	.dc.b	$7F,$75,$BF,$AA,$BF,$7F,$C1,$F8,$0F,$86,$ED,$E4,$59,$FE,$0D,$9F
	.dc.b	$EA,$B1,$FC,$5D,$69,$FC,$6F,$C3,$F9,$7C,$9F,$C4,$FC,$3F,$9F,$C5
	.dc.b	$B8,$F2,$F1,$AB,$FC,$93,$CE,$66,$AF,$C8,$FC,$3F,$BE,$FE,$9D,$F9
	.dc.b	$B3,$FA,$6F,$EC,$36,$CB,$FB,$2D,$FD,$8E,$67,$F4,$9F,$A3,$BE,$2B
	.dc.b	$E5,$78,$DF,$A1,$8C,$57,$BA,$31,$3F,$3F,$C8,$D7,$C4,$F0,$C3,$9F
	.dc.b	$C4,$FD,$3F,$0F,$97,$A9,$F5,$7C,$7B,$9F,$C8,$9E,$D5,$9B,$F1,$FC
	.dc.b	$4F,$DB,$F1,$BC,$5F,$13,$F7,$79,$6E,$3F,$42,$7C,$D5,$9B,$F5,$FC
	.dc.b	$3F,$E0,$FB,$7D,$5F,$0F,$F8,$5F,$1E,$E3,$F6,$27,$DD,$59,$BF,$87
	.dc.b	$F0,$DF,$8B,$F1,$FE,$17,$C3,$7E,$37,$C7,$B8,$FE,$02,$7E,$35,$66
	.dc.b	$FD,$33,$E5,$7E,$6E,$FE,$58,$FF,$43,$74,$23,$DD,$3C,$27,$F0,$1C
	.dc.b	$1F,$80,$8F,$AB,$EE,$7E,$77,$C6,$FA,$9E,$EF,$E7,$FC,$8B,$87,$E4
	.dc.b	$A7,$EC,$56,$6F,$EE,$FC,$37,$E9,$7C,$7F,$B9,$F0,$DF,$A7,$F2,$2E
	.dc.b	$3F,$9A,$9F,$C0,$56,$6F,$C9,$3F,$E0,$0D,$3F,$D5,$91,$EC,$8F,$D0
	.dc.b	$C7,$D4,$7D,$3D,$1F,$51,$CC,$FE,$3B,$E1,$BF,$6D,$E4,$7E,$37,$E1
	.dc.b	$AC,$FE,$49,$FD,$D7,$8F,$F9,$1A,$CD,$FC,$8F,$85,$FE,$F1,$FA,$6F
	.dc.b	$E5,$F0,$FC,$CC,$63,$C5,$FC,$C3,$AF,$13,$F3,$5F,$0F,$9A,$FC,$FF
	.dc.b	$C3,$72,$F0,$FF,$3C,$F4,$F1,$AE,$7F,$E0,$1C,$D5,$FA,$3F,$87,$D6
	.dc.b	$1F,$A8,$F8,$6F,$1F,$8D,$FA,$77,$CB,$DB,$71,$E5,$3F,$D2,$FF,$48
	.dc.b	$FF,$0A,$43,$D0,$67,$FE,$86,$C7,$FC,$41,$5F,$E7,$49,$CF,$8B,$C7
	.dc.b	$E2,$7E,$0F,$93,$71,$F3,$4F,$5F,$88,$CE,$3E,$E9,$F5,$6F,$EC,$7F
	.dc.b	$86,$67,$FE,$2C,$9D,$46,$27,$DC,$FC,$1E,$0F,$47,$BF,$F4,$3C,$6B
	.dc.b	$3F,$15,$AF,$4F,$27,$DD,$EF,$6F,$1F,$EE,$1C,$0F,$87,$FB,$3E,$2F
	.dc.b	$C3,$F8,$8F,$B7,$E4,$DC,$BE,$84,$FC,$7A,$CE,$1E,$69,$FF,$28,$EF
	.dc.b	$F5,$23,$11,$FE,$A7,$E4,$F9,$7B,$FB,$E3,$7C,$1E,$C7,$E0,$C7,$D3
	.dc.b	$F8,$7F,$C5,$F8,$BF,$57,$E2,$2E,$3E,$87,$F1,$FE,$4E,$F0,$FC,$25
	.dc.b	$7D,$BF,$88,$FC,$9F,$87,$F7,$7E,$23,$F2,$BE,$35,$BF,$C5,$4E,$DF
	.dc.b	$C6,$47,$E0,$FE,$1B,$F3,$7E,$2D,$9F,$C8,$78,$95,$FA,$F1,$F8,$9F
	.dc.b	$88,$FD,$0C,$FE,$2F,$E1,$7F,$47,$E3,$5B,$FC,$C7,$8D,$5F,$BF,$1F
	.dc.b	$91,$F8,$6F,$D4,$CF,$E4,$FE,$17,$F5,$7E,$45,$BF,$D0,$35,$FE,$A5
	.dc.b	$FA,$77,$F9,$D1,$FF,$22,$77,$BE,$1F,$8F,$3F,$DE,$E0,$F5,$B6,$BD
	.dc.b	$6D,$7D,$6F,$E1,$37,$F5,$BF,$86,$F7,$FF,$CE,$98,$FF,$9D,$EB,$FC
	.dc.b	$A5,$3F,$A4,$9C,$F9,$3F,$AB,$77,$E3,$FE,$B7,$DF,$CD,$7E,$43,$E1
	.dc.b	$F9,$F9,$3F,$B1,$6C,$67,$1F,$F3,$F6,$4F,$1F,$F6,$DF,$0F,$A9,$F0
	.dc.b	$9F,$03,$6F,$6B,$8F,$86,$70,$A7,$D7,$39,$8C,$3F,$C2,$FA,$7E,$D3
	.dc.b	$F1,$0D,$7C,$3F,$C0,$F1,$FF,$83,$3F,$0B,$CA,$B3,$E6,$9F,$F2,$A7
	.dc.b	$C2,$FF,$9E,$A3,$EA,$CE,$3E,$1F,$E4,$F8,$C9,$FF,$30,$57,$F9,$E5
	.dc.b	$FB,$71,$97,$D5,$1F,$E7,$97,$27,$0F,$87,$FA,$7E,$3E,$F6,$FF,$40
	.dc.b	$57,$F8,$4E,$BE,$4F,$8B,$9E,$3F,$F5,$C4,$E6,$7F,$13,$E2,$E5,$FF
	.dc.b	$46,$47,$F9,$D5,$FA,$31,$9B,$FF,$0F,$4F,$E0,$AF,$0F,$F9,$D6,$BF
	.dc.b	$27,$C3,$8F,$86,$FC,$3B,$AF,$86,$B3,$F0,$27,$F1,$7D,$BC,$BE,$17
	.dc.b	$F1,$BE,$2D,$AF,$C0,$F8,$95,$FC,$FB,$E0,$F8,$6D,$C7,$CA,$F7,$3F
	.dc.b	$2B,$E2,$F9,$3D,$DF,$CB,$F8,$B6,$BF,$16,$FA,$2B,$71,$F4,$BE,$17
	.dc.b	$F3,$BE,$2F,$B1,$FC,$FF,$89,$67,$F2,$6F,$C0,$F8,$6D,$C7,$DA,$F8
	.dc.b	$5F,$D2,$F8,$BF,$11,$FD,$3F,$87,$67,$F3,$6F,$C8,$F8,$6D,$C7,$E0
	.dc.b	$BE,$13,$F5,$BE,$1F,$CC,$6C,$64,$FD,$1C,$7D,$0F,$85,$FF,$73,$7E
	.dc.b	$9F,$F8,$FE,$DF,$C0,$FC,$2E,$3C,$4F,$AA,$DC,$3F,$1E,$D7,$F1,$BF
	.dc.b	$0D,$CB,$C4,$FB,$7F,$0D,$D3,$C1,$FB,$BF,$0D,$6F,$F2,$ED,$7F,$2B
	.dc.b	$3E,$2F,$88,$FC,$37,$E8,$7F,$E5,$49,$FF,$33,$6B,$F8,$5D,$7F,$DA
	.dc.b	$93,$99,$FC,$F7,$BA,$AD,$79,$1F,$A5,$B5,$EE,$FF,$A5,$56,$9F,$F7
	.dc.b	$14,$3E,$1F,$E1,$F8,$DE,$2F,$7B,$E2,$F8,$B6,$FF,$58,$72,$F0,$F0
	.dc.b	$F9,$9F,$4F,$BD,$C2,$3F,$DA,$98,$BE,$1B,$E7,$F8,$5E,$EF,$87,$FA
	.dc.b	$3E,$2D,$BE,$A7,$F8,$27,$6F,$F4,$2B,$CA,$F8,$BD,$3E,$1B,$EC,$F8
	.dc.b	$9F,$23,$E1,$EC,$F8,$1A,$FF,$1B,$CD,$5F,$80,$F1,$3E,$7F,$C3,$FE
	.dc.b	$0B,$C4,$FA,$3F,$0F,$70,$F4,$B5,$E5,$3F,$88,$F1,$3E,$BF,$C3,$7E
	.dc.b	$2B,$C5,$FF,$76,$66,$8C,$75,$7C,$2F,$87,$FC,$7F,$8B,$A9,$CE,$5E
	.dc.b	$5F,$85,$FC,$9F,$8B,$C9,$FC,$AF,$8B,$D3,$E1,$B3,$97,$CC,$9C,$BF
	.dc.b	$3B,$E1,$BF,$37,$E2,$78,$BE,$1A,$CE,$72,$F9,$BE,$27,$E1,$FC,$6D
	.dc.b	$46,$3A,$D8,$7A,$BD,$CF,$A7,$3F,$E3,$BF,$5B,$F2,$9E,$3F,$E4,$A7
	.dc.b	$38,$FF,$45,$F1,$FF,$33,$18,$AF,$C6,$5F,$E0,$BE,$1B,$1F,$11,$F9
	.dc.b	$EF,$87,$D7,$93,$FA,$0F,$87,$B7,$F8,$86,$BF,$66,$7A,$79,$1F,$A6
	.dc.b	$F8,$7E,$DE,$37,$F2,$DF,$D0,$BF,$CA,$3B,$FF,$57,$8C,$61,$A3,$F2
	.dc.b	$75,$E5,$F2,$78,$7C,$3F,$9F,$C7,$E5,$F1,$16,$FF,$55,$D1,$FA,$BC
	.dc.b	$D3,$99,$FE,$CD,$1F,$E5,$BE,$27,$93,$ED,$9F,$9F,$1F,$E6,$E7,$D8
	.dc.b	$63,$87,$E0,$34,$FD,$3F,$07,$B9,$F3,$9C,$B6,$F5,$B4,$F2,$8E,$66
	.dc.b	$3F,$0B,$0F,$D8,$4F,$23,$FD,$E0,$FF,$AD,$5E,$33,$C3,$E2,$3E,$C7
	.dc.b	$8F,$F2,$DB,$3A,$B4,$62,$3E,$0F,$C3,$FD,$FF,$27,$E0,$47,$6D,$3D
	.dc.b	$AD,$3E,$09,$4F,$B8,$7C,$6E,$5F,$B8,$FC,$8E,$3F,$7B,$E8,$D6,$BF
	.dc.b	$C8,$35,$F6,$BF,$BD,$E4,$FB,$8F,$9F,$E1,$BF,$25,$E2,$FE,$17,$E1
	.dc.b	$AC,$FA,$DA,$FD,$A9,$FC,$C7,$91,$F8,$BF,$86,$FC,$D7,$8B,$6B,$ED
	.dc.b	$B5,$F9,$6F,$E4,$7E,$1B,$F4,$1E,$3F,$E4,$FE,$17,$F4,$5E,$47,$E5
	.dc.b	$7E,$1A,$E1,$F2,$DA,$FE,$2A,$7F,$51,$E4,$FE,$6F,$83,$F5,$5E,$2F
	.dc.b	$E7,$7E,$16,$DF,$D3,$6B,$F9,$29,$FD,$87,$8D,$E5,$8F,$8D,$AD,$7F
	.dc.b	$D1,$52,$8C,$47,$EF,$5F,$53,$F9,$8D,$3D,$4F,$E6,$B4,$F5,$34,$F5
	.dc.b	$14,$F5,$1F,$A0,$CB,$F6,$9F,$0B,$FC,$0F,$BE,$CF,$E9,$3C,$6A,$FE
	.dc.b	$51,$FD,$D7,$C2,$EE,$3F,$21,$C5,$F4,$9A,$7D,$22,$9F,$48,$FC,$B7
	.dc.b	$17,$D2,$7C,$A7,$E8,$32,$F4,$B4,$C9,$F8,$B7,$DE,$FE,$9B,$2F,$E1
	.dc.b	$9F,$0B,$FC,$6F,$A2,$CF,$F2,$AF,$D4,$3F,$B1,$F8,$6F,$1B,$F8,$35
	.dc.b	$FF,$24,$9C,$C7,$F0,$9F,$86,$E5,$7F,$AB,$F8,$6D,$79,$5F,$C6,$5E
	.dc.b	$1E,$4D,$9E,$CF,$F1,$DF,$86,$DC,$7E,$CF,$E1,$BB,$79,$1F,$C8,$5F
	.dc.b	$07,$91,$67,$CC,$FF,$25,$F8,$6D,$C7,$EE,$FE,$1B,$CD,$E1,$FF,$95
	.dc.b	$7F,$43,$FF,$38,$4F,$F8,$33,$83,$FD,$B8,$FF,$06,$8C,$BF,$ED,$38
	.dc.b	$CB,$FC,$8A,$BE,$07,$8B,$E9,$8F,$F0,$3F,$A0,$A7,$06,$67,$B6,$F5
	.dc.b	$E8,$FB,$DF,$9D,$1E,$B7,$9C,$FF,$12,$7E,$11,$F4,$F6,$D3,$E9,$7E
	.dc.b	$03,$F5,$FA,$7D,$DF,$0F,$F5,$7A,$AC,$FD,$BF,$16,$BD,$9F,$89,$F0
	.dc.b	$FB,$8F,$B0,$79,$0A,$79,$33,$F5,$2D,$3E,$04,$7A,$E3,$D3,$A7,$99
	.dc.b	$F9,$59,$FD,$1F,$87,$FA,$B1,$FE,$B1,$E0,$FF,$47,$3E,$B3,$DE,$7C
	.dc.b	$2D,$7E,$C7,$C3,$FE,$3F,$C9,$FB,$5F,$11,$F9,$1F,$1E,$DF,$E5,$DD
	.dc.b	$BF,$99,$9F,$C0,$7C,$47,$E5,$FC,$AF,$C1,$7C,$47,$E6,$7C,$8B,$7F
	.dc.b	$9F,$76,$FE,$86,$7F,$11,$F1,$1F,$9F,$F1,$FF,$15,$F1,$1F,$A1,$F1
	.dc.b	$2D,$FE,$9D,$DB,$FA,$99,$FC,$87,$C4,$7E,$9F,$C4,$FC,$97,$C4,$7E
	.dc.b	$A7,$C5,$B7,$FA,$F7,$6F,$EC,$67,$C6,$FE,$13,$3F,$3B,$F8,$6D,$2B
	.dc.b	$F9,$B7,$6F,$E7,$63,$CA,$FE,$33,$1F,$D1,$7C,$47,$A6,$3F,$40,$D7
	.dc.b	$EA,$69,$EF,$7F,$49,$F1,$1F,$BF,$F1,$FF,$54,$FF,$01,$F1,$EC,$FF
	.dc.b	$07,$7F,$5D,$F1,$1B,$8F,$D6,$71,$FD,$17,$F5,$DC,$7F,$45,$A7,$D1
	.dc.b	$36,$7D,$13,$CA,$7C,$B3,$F6,$DF,$11,$E7,$F2,$FF,$8A,$47,$F8,$17
	.dc.b	$AD,$39,$F2,$FF,$D0,$D0,$DB,$D9,$5F,$91,$F8,$7F,$E3,$5E,$5F,$F0
	.dc.b	$5F,$87,$B3,$FC,$0E,$6B,$F9,$78,$FE,$41,$E5,$FF,$0B,$F8,$7F,$F2
	.dc.b	$2F,$EA,$3F,$EA,$DC,$FB,$FC,$36,$1D,$61,$FE,$54,$9D,$3E,$79,$E3
	.dc.b	$E6,$7F,$15,$F8,$8E,$7E,$5D,$AE,$1E,$3F,$FB,$B7,$D1,$FF,$4F,$47
	.dc.b	$EC,$71,$FF,$7B,$71,$FC,$21,$C4,$C7,$9D,$7F,$88,$AF,$FC,$AE,$FF
	.dc.b	$18,$BF,$F7,$D3,$FE,$2D,$D3,$FD,$59,$3A,$8F,$F5,$8C,$FF,$AA,$F0
	.dc.b	$4F,$F8,$FE,$79,$78,$DE,$5F,$87,$F8,$7E,$47,$C0,$9E,$D8,$FF,$31
	.dc.b	$DF,$E2,$47,$F2,$FE,$0A,$FF,$03,$8F,$86,$F9,$FE,$03,$D2,$F9,$4F
	.dc.b	$F2,$16,$48,$FA,$3B,$FC,$5F,$7E,$D7,$13,$E8,$70,$FC,$9F,$80,$FA
	.dc.b	$FD,$76,$7E,$A7,$5F,$CD,$F8,$6F,$B6,$E6,$3E,$34,$FA,$DF,$87,$B7
	.dc.b	$AD,$FB,$8E,$51,$F5,$5D,$3F,$3F,$9B,$ED,$7C,$37,$B3,$93,$EC,$6F
	.dc.b	$FE,$6C,$AF,$BE,$7D,$7F,$86,$F5,$BF,$06,$7E,$8D,$71,$C3,$EB,$FC
	.dc.b	$2F,$E4,$BC,$DF,$C3,$7C,$27,$E5,$3C,$6F,$C4,$4D,$C7,$F0,$53,$B7
	.dc.b	$F0,$9F,$09,$F9,$AF,$3B,$F1,$CF,$E7,$3C,$EF,$C8,$7C,$15,$AD,$9F
	.dc.b	$9B,$AF,$CA,$7C,$17,$E8,$BC,$6B,$3F,$9F,$F2,$BF,$31,$F0,$1F,$A6
	.dc.b	$73,$1F,$91,$9F,$CE,$7C,$07,$EA,$BC,$BB,$3F,$A7,$F2,$FF,$41,$F0
	.dc.b	$9F,$AE,$73,$1F,$99,$9F,$3B,$E3,$3C,$A7,$E7,$75,$F4,$BF,$B1,$76
	.dc.b	$FE,$87,$4F,$2B,$FB,$5C,$7C,$AF,$E8,$32,$CB,$FA,$77,$2F,$DE,$F6
	.dc.b	$DE,$3F,$F0,$D8,$FD,$77,$C2,$7F,$04,$F3,$7F,$70,$D8,$FD,$97,$C2
	.dc.b	$57,$EE,$76,$27,$F8,$5F,$9D,$E6,$9F,$D8,$E9,$5F,$06,$2D,$7D,$B1
	.dc.b	$EF,$3E,$09,$C0,$FE,$01,$C5,$F5,$5F,$E0,$5C,$7F,$4D,$C9,$F5,$CC
	.dc.b	$9F,$0C,$FE,$11,$F0,$9F,$C7,$BC,$9F,$E2,$4F,$F2,$0F,$27,$F8,$A7
	.dc.b	$C2,$5A,$D9,$FC,$7E,$BF,$46,$FE,$67,$F4,$2F,$F6,$CE,$7F,$C5,$63
	.dc.b	$FD,$6B,$1F,$E2,$87,$FD,$F3,$FA,$5F,$FB,$CB,$C0,$FD,$5F,$84,$AF
	.dc.b	$F7,$0C,$E6,$BF,$D8,$91,$C5,$FE,$39,$E8,$78,$2B,$F9,$24,$FF,$B7
	.dc.b	$27,$AF,$9F,$9C,$3A,$CF,$F8,$BE,$7F,$C7,$78,$7F,$1D,$BE,$3F,$03
	.dc.b	$EA,$9F,$5F,$42,$B5,$87,$2F,$80,$D5,$7E,$16,$7F,$97,$E1,$E2,$E7
	.dc.b	$CC,$FF,$8A,$67,$F8,$D3,$F0,$BE,$1B,$38,$F5,$F4,$7C,$D3,$EE,$BC
	.dc.b	$DF,$08,$F9,$97,$8F,$43,$78,$FC,$79,$FE,$AB,$BE,$36,$F8,$D1,$EA
	.dc.b	$C3,$E5,$56,$E3,$E8,$FC,$4E,$FE,$0F,$8C,$7D,$6E,$0F,$0F,$0F,$3E
	.dc.b	$0F,$1F,$C4,$75,$9B,$5F,$72,$76,$FD,$EF,$88,$F4,$FA,$7E,$37,$D7
	.dc.b	$E9,$F9,$7E,$16,$D6,$CF,$C1,$D7,$A5,$F1,$9F,$8B,$CF,$E1,$E9,$98
	.dc.b	$F4,$9F,$8C,$8D,$EB,$F6,$A3,$F0,$27,$94,$FC,$7E,$9E,$96,$99,$3F
	.dc.b	$07,$AF,$C9,$F8,$7F,$A3,$E8,$D9,$FC,$BF,$A5,$F3,$7E,$1F,$EA,$B9
	.dc.b	$8F,$C5,$CF,$D1,$F8,$6F,$B3,$E9,$59,$FC,$FB,$F5,$7E,$1F,$31,$FA
	.dc.b	$49,$FB,$FE,$9F,$D9,$F8,$3F,$C0,$FA,$96,$BF,$4E,$FD,$DF,$84,$CC
	.dc.b	$7E,$B2,$7F,$0F,$EA,$FE,$07,$E1,$3F,$13,$E4,$FE,$0F,$E1,$ED,$FE
	.dc.b	$CA,$6B,$FB,$78,$FC,$7F,$97,$E7,$8F,$D2,$ED,$5F,$C5,$3E,$97,$D6
	.dc.b	$7A,$4F,$DD,$E9,$E7,$7F,$5B,$A5,$7F,$24,$FA,$5F,$59,$E9,$3F,$81
	.dc.b	$E5,$F9,$9F,$86,$FD,$0F,$B1,$F9,$BF,$87,$B5,$FC,$2A,$6B,$FC,$3E
	.dc.b	$3F,$4F,$EC,$F9,$E3,$F6,$99,$57,$F6,$4F,$A5,$F6,$9E,$93,$F8,$1E
	.dc.b	$5E,$77,$F7,$99,$57,$F7,$4F,$A5,$F6,$9E,$93,$F8,$5E,$5F,$B1,$F8
	.dc.b	$7F,$DC,$FA,$9F,$B3,$F8,$7B,$5F,$C8,$A6,$BF,$C9,$E3,$F7,$FE,$B7
	.dc.b	$EE,$7E,$1F,$F8,$0F,$A9,$FB,$BF,$87,$FE,$07,$EA,$67,$0F,$E5,$2E
	.dc.b	$5F,$E5,$73,$FC,$13,$F4,$2F,$F6,$EC,$7F,$AE,$75,$FF,$7C,$AF,$09
	.dc.b	$2F,$81,$FE,$27,$5A,$9C,$FA,$98,$BE,$95,$AF,$17,$B7,$C2,$70,$F3
	.dc.b	$6D,$7F,$24,$F5,$59,$7F,$2F,$CB,$FC,$B1,$9F,$FB,$0A,$BF,$B6,$4E
	.dc.b	$73,$FE,$83,$2C,$7A,$57,$F4,$8D,$FE,$1B,$FD,$2B,$7F,$86,$E4,$F8
	.dc.b	$66,$4F,$86,$7F,$4C,$BE,$A7,$AD,$C1,$F0,$EF,$E1,$3E,$9C,$3F,$CC
	.dc.b	$2E,$B1,$F8,$6E,$7D,$4F,$86,$EC,$FA,$1F,$0D,$F2,$BD,$5B,$1C,$FD
	.dc.b	$0F,$A9,$F0,$FF,$39,$CC,$7F,$4F,$DF,$E3,$D7,$DD,$8F,$EA,$BC,$7F
	.dc.b	$75,$FE,$AF,$C9,$97,$F9,$E3,$97,$F9,$F4,$7E,$07,$6F,$74,$7F,$92
	.dc.b	$F6,$F6,$BD,$F6,$F6,$B9,$9F,$C0,$EB,$EC,$8D,$1F,$8B,$F8,$5F,$C2
	.dc.b	$FA,$3F,$8D,$F8,$6B,$3F,$17,$15,$62,$3C,$13,$F8,$CF,$3B,$F2,$9F
	.dc.b	$0D,$F8,$EF,$3E,$D7,$CF,$F3,$3F,$31,$F0,$DF,$92,$73,$1E,$79,$FC
	.dc.b	$E7,$C0,$7E,$5B,$CF,$B3,$F5,$DF,$D0,$7C,$06,$63,$C9,$3F,$9C,$F3
	.dc.b	$FF,$49,$F0,$9F,$9E,$F4,$6D,$7D,$FF,$43,$F5,$1F,$09,$FA,$27,$31
	.dc.b	$F1,$E7,$C6,$FF,$81,$39,$9A,$62,$3F,$37,$AF,$A9,$F6,$1F,$17,$7F
	.dc.b	$80,$F9,$0F,$41,$4F,$A9,$E8,$FA,$9F,$C9,$3E,$E7,$EC,$6F,$FB,$DF
	.dc.b	$85,$FD,$BF,$A1,$FB,$FF,$86,$B5,$F4,$E6,$BF,$81,$8F,$DE,$FA,$3E
	.dc.b	$78,$FE,$09,$9D,$EF,$FD,$BE,$9E,$99,$FC,$33,$9D,$7D,$F1,$E7,$3E
	.dc.b	$16,$9E,$56,$9F,$83,$E4,$FE,$17,$19,$8F,$DF,$EB,$FC,$57,$E1,$BF
	.dc.b	$87,$F9,$BF,$ED,$BE,$EF,$E0,$31,$87,$F1,$9F,$0D,$FC,$53,$CE,$D6
	.dc.b	$1F,$B7,$AD,$3F,$9D,$F8,$5F,$E3,$3E,$56,$BC,$9F,$E3,$BF,$0D,$FC
	.dc.b	$73,$CB,$C7,$86,$BD,$E9,$F9,$9A,$D3,$F9,$BF,$83,$FF,$1A,$FE,$A4
	.dc.b	$8E,$31,$87,$B7,$E8,$7F,$EF,$6E,$65,$7F,$2B,$9E,$5A,$FF,$23,$8E
	.dc.b	$51,$FC,$2E,$BF,$DE,$13,$FE,$8D,$9F,$E4,$13,$9F,$7F,$C3,$E8,$76
	.dc.b	$F7,$FC,$7F,$13,$E0,$9C,$D7,$EC,$75,$F5,$E3,$FB,$FD,$7D,$6F,$C5
	.dc.b	$3F,$8E,$70,$FC,$28,$CF,$0F,$F3,$4D,$FD,$FE,$97,$BA,$FD,$0F,$BB
	.dc.b	$E0,$7E,$1F,$A7,$99,$FE,$31,$C0,$AF,$87,$F1,$1F,$23,$CE,$F8,$BF
	.dc.b	$11,$F2,$BC,$FC,$E1,$F0,$E7,$2F,$C5,$8F,$AB,$BF,$AB,$5F,$E8,$FB
	.dc.b	$79,$B8,$3F,$A2,$CE,$5F,$E8,$F1,$9B,$FE,$75,$BF,$D8,$F1,$FC,$D1
	.dc.b	$FE,$7A,$D9,$1E,$B8,$FA,$9E,$76,$9F,$A5,$1F,$73,$E1,$3F,$03,$E7
	.dc.b	$D9,$FB,$5E,$1F,$E0,$3E,$17,$3A,$7D,$BB,$FC,$3C,$FE,$13,$E1,$3F
	.dc.b	$13,$E8,$5C,$3F,$05,$E8,$7E,$23,$E1,$73,$8F,$E0,$EF,$F1,$F3,$F8
	.dc.b	$CF,$81,$FC,$8F,$9F,$70,$FC,$57,$9F,$F9,$0F,$81,$CE,$3F,$8B,$BF
	.dc.b	$CB,$CF,$E5,$3C,$DF,$CC,$F9,$17,$0F,$C9,$3F,$98,$F8,$1F,$CE,$F9
	.dc.b	$79,$9F,$B7,$3E,$77,$EE,$EF,$5F,$95,$AF,$A5,$FC,$E4,$65,$FC,$D4
	.dc.b	$79,$DF,$C2,$EB,$5F,$A4,$FA,$5F,$D2,$74,$E5,$FD,$14,$7E,$B3,$E1
	.dc.b	$7F,$63,$D5,$67,$F5,$4F,$EC,$3E,$17,$F6,$BD,$79,$9F,$C9,$CF,$8D
	.dc.b	$79,$5F,$B4,$CF,$0F,$ED,$B3,$F5,$4F,$BB,$8B,$E9,$3F,$59,$F9,$8F
	.dc.b	$EE,$73,$F3,$3D,$3B,$11,$EB,$8F,$D1,$EB,$AC,$3F,$81,$CF,$F0,$6F
	.dc.b	$85,$FF,$5F,$FE,$A6,$F1,$B1,$87,$F8,$9E,$31,$97,$F0,$5F,$86,$DF
	.dc.b	$97,$AA,$FF,$03,$56,$9F,$F1,$D5,$7F,$BA,$3C,$D7,$C3,$76,$F3,$3F
	.dc.b	$8A,$BE,$0F,$27,$F8,$BF,$C3,$66,$FF,$89,$4E,$5F,$E2,$9F,$01,$E6
	.dc.b	$F2,$FF,$8E,$FC,$0D,$9C,$E7,$E7,$F1,$72,$FA,$6B,$F9,$81,$FE,$E0
	.dc.b	$7E,$8E,$AF,$1B,$31,$F0,$BC,$4D,$FC,$2F,$C4,$9C,$CF,$C3,$9F,$4B
	.dc.b	$F1,$77,$AF,$C2,$DF,$D6,$E7,$3F,$5B,$F2,$E3,$F9,$AE,$BE,$98,$F8
	.dc.b	$E7,$F9,$87,$81,$B6,$7C,$BF,$E7,$D3,$9A,$FA,$FB,$FB,$A3,$E8,$EF
	.dc.b	$EE,$F8,$5F,$B7,$E4,$59,$FB,$4F,$C2,$F8,$7F,$BF,$E4,$66,$7E,$CC
	.dc.b	$FA,$91,$3A,$DC,$8E,$1D,$4E,$61,$C1,$F0,$23,$F0,$73,$A7,$E7,$C7
	.dc.b	$C0,$7E,$A7,$0F,$F4,$FE,$0F,$73,$F8,$7D,$FD,$AE,$67,$67,$C6,$E0
	.dc.b	$6F,$FE,$46,$8F,$39,$F7,$3C,$9F,$BF,$F0,$9F,$9C,$7F,$03,$F0,$BF
	.dc.b	$9E,$F2,$B3,$5F,$9D,$9F,$C2,$FC,$37,$E8,$BC,$BF,$C3,$FC,$35,$D3
	.dc.b	$FC,$DD,$9A,$BF,$4F,$E5,$6B,$1F,$D3,$61,$F8,$EF,$86,$FD,$5F,$97
	.dc.b	$67,$F5,$5E,$27,$E4,$BE,$07,$39,$F4,$BF,$D8,$CF,$E5,$BE,$07,$F6
	.dc.b	$7E,$2D,$C3,$F6,$5E,$1F,$E6,$BE,$03,$38,$F9,$2F,$F7,$33,$EA,$AF
	.dc.b	$1C,$FA,$38,$3D,$51,$F2,$78,$3D,$4F,$EF,$77,$F5,$3F,$BA,$DF,$D0
	.dc.b	$FE,$F3,$7A,$FC,$EE,$BF,$C1,$ED,$EB,$8F,$E0,$D3,$A7,$F0,$91,$FB
	.dc.b	$0F,$86,$FE,$1B,$E3,$59,$FE,$18,$FE,$D3,$E1,$7F,$88,$F5,$66,$7F
	.dc.b	$85,$CF,$AD,$F2,$6D,$EB,$7D,$A7,$AF,$97,$E9,$C6,$63,$67,$BF,$93
	.dc.b	$E7,$B4,$FA,$71,$F7,$39,$73,$D5,$FA,$E3,$F1,$B5,$F5,$27,$F0,$3F
	.dc.b	$0D,$FC,$9B,$B3,$F8,$5B,$63,$F0,$BF,$09,$F9,$2C,$7E,$37,$EA,$9F
	.dc.b	$AA,$9E,$31,$FC,$AE,$FF,$89,$FC,$3E,$E3,$F5,$D4,$F2,$F3,$1F,$ED
	.dc.b	$78,$78,$FF,$C6,$BE,$1F,$84,$FF,$4A,$9E,$18,$FE,$5B,$0C,$F3,$27
	.dc.b	$8C,$FE,$6B,$E1,$F3,$0F,$27,$F9,$2F,$C3,$E1,$FE,$4F,$3A,$F1,$73
	.dc.b	$97,$8E,$7F,$95,$FC,$1F,$2F,$1A,$CF,$9D,$FE,$5F,$F0,$FD,$B9,$F3
	.dc.b	$3F,$EC,$BC,$D5,$EE,$8C,$70,$D7,$E4,$F0,$FB,$DF,$85,$19,$7E,$3C
	.dc.b	$79,$9F,$26,$DF,$B6,$DB,$D4,$FB,$E3,$E5,$6B,$EA,$7C,$5A,$FF,$9C
	.dc.b	$FF,$43,$FE,$5B,$1E,$5F,$0B,$1F,$13,$F1,$9D,$61,$F2,$3A,$33,$5F
	.dc.b	$56,$7C,$8F,$D3,$D2,$F1,$FD,$1D,$7D,$53,$EE,$C3,$EB,$3F,$76,$3C
	.dc.b	$8F,$E2,$33,$AF,$E1,$1F,$73,$F2,$74,$CB,$F8,$18,$F5,$7C,$47,$DC
	.dc.b	$EB,$F6,$7C,$45,$C7,$F9,$DF,$67,$E7,$78,$17,$F8,$2F,$1F,$79,$7E
	.dc.b	$3F,$1F,$89,$F0,$FF,$86,$F2,$6C,$FE,$4D,$F9,$1F,$11,$F8,$AF,$13
	.dc.b	$33,$F7,$67,$E6,$7C,$47,$E3,$BC,$5B,$3F,$9B,$7E,$87,$C4,$7E,$4B
	.dc.b	$C4,$CC,$FE,$16,$7E,$A7,$C3,$FE,$5B,$C5,$FA,$DF,$11,$F9,$8F,$13
	.dc.b	$37,$FA,$59,$CB,$FA,$78,$F1,$BF,$B2,$CE,$BF,$AE,$7D,$2F,$EB,$74
	.dc.b	$CB,$FA,$B8,$F1,$BF,$BA,$C6,$BF,$B6,$7D,$2F,$ED,$72,$CB,$FB,$38
	.dc.b	$FC,$67,$C4,$7E,$BB,$C6,$FC,$77,$C4,$5E,$7F,$A6,$7D,$4E,$65,$7E
	.dc.b	$D3,$C7,$DE,$3F,$C0,$70,$F2,$F6,$FE,$EF,$4B,$E2,$1F,$5D,$7E,$B7
	.dc.b	$6C,$C7,$92,$BF,$81,$6B,$6C,$FA,$AF,$F6,$9B,$66,$3F,$49,$F0,$7F
	.dc.b	$C1,$BC,$4F,$D3,$7C,$1D,$E0,$F4,$E0,$ED,$70,$B0,$FE,$1B,$E4,$F0
	.dc.b	$AF,$E3,$58,$FE,$BF,$E0,$7F,$88,$F9,$5F,$B1,$F8,$8B,$9F,$4C,$F9
	.dc.b	$D7,$F1,$AE,$6F,$F6,$AE,$3F,$E0,$1D,$18,$70,$8C,$7C,$3E,$FC,$AF
	.dc.b	$DE,$7C,$45,$CF,$A6,$0A,$E9,$E2,$6B,$0F,$E6,$5C,$AC,$FF,$DC,$D1
	.dc.b	$DF,$E1,$7C,$5E,$37,$F0,$7F,$87,$B9,$F4,$9F,$37,$8B,$9B,$E7,$83
	.dc.b	$0F,$81,$1F,$CE,$F8,$BE,$03,$70,$F2,$F0,$FB,$DC,$C7,$F2,$79,$FE
	.dc.b	$2D,$F1,$1E,$B8,$FF,$11,$7E,$85,$FE,$FD,$C7,$CF,$C9,$FB,$DA,$F6
	.dc.b	$E1,$F2,$EB,$5C,$7F,$D7,$63,$F8,$1C,$7F,$BF,$B0,$FF,$49,$E0,$F1
	.dc.b	$F5,$1F,$76,$7F,$C8,$0F,$B2,$33,$8A,$79,$70,$7C,$9F,$1F,$31,$C2
	.dc.b	$FF,$B2,$4F,$CD,$BF,$87,$1E,$DC,$7E,$AF,$37,$DB,$C7,$F8,$B6,$DF
	.dc.b	$0E,$73,$5F,$1C,$F8,$3E,$F7,$C3,$AF,$EF,$19,$7F,$91,$27,$E8,$EB
	.dc.b	$E2,$AF,$87,$1F,$23,$DA,$E1,$87,$C2,$F8,$3F,$C7,$79,$5E,$8F,$86
	.dc.b	$B3,$F2,$5F,$C9,$79,$7E,$CA,$CE,$1F,$4B,$E1,$BF,$2D,$E6,$7C,$07
	.dc.b	$F3,$1E,$55,$9F,$C1,$CF,$C4,$F8,$4C,$E5,$F5,$7E,$17,$F3,$DC,$FF
	.dc.b	$23,$E0,$BF,$41,$E2,$DC,$7F,$17,$3F,$32,$B3,$7F,$77,$E0,$FF,$4D
	.dc.b	$E3,$FD,$0F,$80,$B3,$F8,$17,$F5,$5D,$1F,$52,$B3,$8F,$A4,$F2,$D7
	.dc.b	$E1,$F7,$F2,$C6,$4F,$B3,$93,$6F,$54,$7E,$7B,$7F,$23,$E6,$BF,$63
	.dc.b	$FA,$2D,$72,$F2,$BF,$C2,$FC,$0F,$AA,$3F,$37,$F0,$3F,$BF,$EB,$FC
	.dc.b	$4B,$FC,$07,$C6,$B3,$FA,$C9,$FC,$6F,$C0,$E7,$2F,$44,$FA,$8F,$41
	.dc.b	$FE,$57,$C7,$F9,$6C,$6A,$3D,$7A,$7A,$9F,$6E,$9E,$67,$85,$57,$FD
	.dc.b	$44,$FF,$1D,$AF,$6C,$7A,$DF,$D7,$67,$EB,$ED,$FC,$9F,$C1,$FF,$19
	.dc.b	$F2,$BF,$47,$F0,$7F,$C6,$FC,$8B,$8F,$F0,$69,$FD,$3D,$67,$1F,$DC
	.dc.b	$7C,$17,$F2,$1F,$2B,$F5,$7F,$07,$FC,$8F,$C8,$B8,$FF,$0E,$9F,$D7
	.dc.b	$D6,$6F,$F8,$07,$C1,$FF,$90,$BF,$55,$FF,$7A,$6D,$FE,$5D,$FD,$0F
	.dc.b	$FD,$C5,$E5,$27,$37,$FE,$06,$9C,$C3,$E2,$7F,$77,$87,$3F,$8A,$FE
	.dc.b	$03,$5F,$E5,$7D,$3F,$94,$DF,$FB,$17,$0F,$E7,$53,$99,$FE,$03,$3F
	.dc.b	$EC,$69,$CF,$C1,$F9,$7E,$23,$9F,$C2,$79,$FC,$CE,$BF,$07,$E9,$F3
	.dc.b	$7B,$FC,$27,$AF,$CB,$BA,$78,$27,$6F,$8A,$3C,$BF,$0F,$F0,$7C,$EB
	.dc.b	$3E,$6F,$1A,$BE,$48,$F5,$FC,$47,$C7,$9F,$6F,$C3,$FC,$9F,$3F,$FD
	.dc.b	$9B,$D1,$F2,$63,$E6,$79,$D8,$E3,$F9,$1F,$11,$8C,$FE,$0D,$71,$7E
	.dc.b	$1F,$C3,$FD,$3F,$47,$84,$7F,$69,$E8,$7C,$46,$63,$67,$CC,$9D,$E6
	.dc.b	$AD,$CF,$D0,$F3,$F3,$1F,$E0,$38,$FA,$7E,$8E,$3E,$13,$F0,$5E,$4E
	.dc.b	$BE,$12,$D7,$D6,$9F,$C3,$57,$2F,$82,$FC,$47,$A5,$6B,$EE,$FA,$1D
	.dc.b	$BE,$0F,$F1,$8E,$63,$EE,$CF,$8B,$E2,$3F,$21,$E6,$59,$FC,$2F,$9F
	.dc.b	$E6,$F8,$8F,$CA,$39,$8F,$C2,$CF,$AB,$E1,$FF,$31,$E6,$59,$FC,$6B
	.dc.b	$EE,$F8,$8C,$C7,$E3,$A7,$F3,$DE,$87,$C2,$F8,$6F,$D0,$79,$BF,$13
	.dc.b	$E1,$ED,$FE,$52,$6B,$FE,$19,$EE,$FD,$04,$FF,$AA,$3F,$56,$FF,$5D
	.dc.b	$75,$7E,$9F,$83,$D7,$7F,$9C,$7D,$2F,$E7,$B8,$2B,$CC,$FB,$91,$C1
	.dc.b	$C4,$FD,$9D,$3D,$31,$8E,$94,$62,$7F,$6D,$B5,$CF,$F6,$93,$EE,$BF
	.dc.b	$D6,$FC,$27,$9F,$CE,$FC,$2F,$C3,$7F,$A6,$A7,$FD,$ED,$AF,$EC,$72
	.dc.b	$FE,$01,$7B,$9F,$F1,$75,$6D,$0F,$4D,$7F,$ED,$08,$F8,$07,$EF,$B9
	.dc.b	$3E,$C3,$C0,$E1,$E8,$B9,$BE,$F4,$7E,$00,$E2,$7D,$E3,$EC,$1F,$70
	.dc.b	$A7,$F0,$6E,$6F,$BC,$E4,$F8,$DA,$FC,$17,$E5,$E4,$F8,$2F,$B1,$E7
	.dc.b	$FC,$2F,$87,$FB,$5E,$87,$C4,$9B,$7F,$02,$76,$FC,$2F,$87,$FC,$07
	.dc.b	$9F,$F2,$B1,$FC,$17,$9D,$F3,$26,$D7,$C8,$9D,$BF,$2B,$E1,$BF,$11
	.dc.b	$E5,$FD,$2F,$87,$B3,$93,$FC,$9B,$C6,$9D,$46,$3C,$DF,$B3,$F0,$DF
	.dc.b	$91,$F3,$AC,$F0,$F3,$BE,$EF,$C3,$7E,$55,$CC,$7F,$8B,$31,$57,$9E
	.dc.b	$3B,$EF,$EB,$7C,$3B,$E5,$E0,$E5,$E5,$1F,$8A,$D7,$D3,$1F,$3B,$5B
	.dc.b	$D6,$ED,$FF,$65,$F0,$7F,$91,$EB,$DB,$C1,$EC,$7C,$9C,$1F,$94,$F8
	.dc.b	$1F,$D6,$79,$BF,$96,$F8,$1B,$5F,$92,$9A,$FD,$C8,$FD,$97,$9D,$F9
	.dc.b	$CF,$81,$FD,$A7,$9D,$6B,$E5,$BF,$A0,$F8,$4C,$C7,$E7,$27,$F7,$5E
	.dc.b	$7F,$E9,$3E,$1B,$F7,$9E,$4D,$AF,$A6,$FE,$A3,$E1,$B3,$1F,$A4,$9F
	.dc.b	$E0,$5E,$5F,$EB,$3E,$17,$F8,$27,$87,$FA,$EF,$86,$FE,$0D,$E6,$67
	.dc.b	$0F,$BA,$E5,$FB,$F3,$E7,$7E,$B6,$D5,$FD,$AB,$E9,$7D,$67,$A4,$FC
	.dc.b	$3E,$9E,$77,$EF,$69,$5F,$DE,$BE,$97,$D6,$7A,$4F,$C7,$E5,$FC,$13
	.dc.b	$E1,$7F,$8E,$7A,$1F,$C1,$BE,$1B,$F8,$F7,$A3,$9B,$FC,$AB,$97,$F2
	.dc.b	$F3,$FC,$33,$E1,$7F,$92,$7A,$16,$7F,$36,$FF,$10,$F8,$6C,$C7,$F0
	.dc.b	$C9,$FE,$08,$FF,$05,$F4,$5F,$AB,$FE,$26,$3F,$D7,$DF,$A1,$BE,$1F
	.dc.b	$71,$FC,$4A,$B7,$E7,$7F,$60,$AF,$F4,$6D,$BC,$EC,$61,$C3,$0F,$E3
	.dc.b	$13,$97,$AF,$C4,$67,$1E,$71,$E0,$F8,$8C,$57,$FA,$D6,$30,$7F,$B0
	.dc.b	$97,$89,$E1,$F8,$7D,$BE,$3A,$E3,$E8,$FC,$3B,$FD,$97,$05,$7F,$98
	.dc.b	$BF,$15,$F8,$E7,$C5,$3B,$78,$FE,$A7,$C1,$BF,$BF,$4F,$2E,$D9,$9F
	.dc.b	$F5,$4E,$9E,$D7,$FA,$5E,$FF,$12,$FE,$37,$C3,$FB,$3C,$BB,$5E,$B7
	.dc.b	$E5,$7C,$3E,$63,$E8,$CF,$C2,$F3,$3D,$11,$F4,$7D,$06,$D8,$8F,$E8
	.dc.b	$91,$87,$FA,$4F,$37,$E0,$1F,$E6,$7C,$DF,$33,$83,$D1,$1F,$81,$DB
	.dc.b	$F9,$B3,$F5,$0F,$9D,$BE,$9F,$C0,$47,$B7,$CD,$FC,$30,$E0,$F7,$4F
	.dc.b	$E2,$3E,$1F,$EC,$FA,$9F,$85,$F8,$7B,$39,$3E,$D4,$65,$FB,$9E,$AF
	.dc.b	$E2,$FE,$1F,$F0,$3E,$7F,$E3,$7E,$1F,$F0,$7E,$86,$6F,$F0,$4E,$5F
	.dc.b	$C2,$4F,$E4,$FE,$1F,$F1,$3E,$7D,$9F,$C4,$7A,$1F,$97,$F8,$7F,$C6
	.dc.b	$B9,$8F,$B3,$3F,$9B,$F8,$6F,$C8,$FA,$3F,$9D,$F8,$7B,$5F,$A4,$9A
	.dc.b	$FD,$E8,$FC,$BF,$A7,$E5,$8F,$C1,$E7,$E5,$7F,$0B,$A6,$5F,$A0,$E5
	.dc.b	$FA,$51,$E3,$7E,$A6,$55,$FA,$2F,$A5,$F5,$9F,$B6,$D3,$DE,$FE,$E3
	.dc.b	$E1,$BF,$53,$EA,$7E,$D7,$E1,$FF,$57,$EA,$FE,$DE,$6D,$7E,$FA,$76
	.dc.b	$FF,$00,$F8,$3F,$D8,$FA,$1F,$BD,$7F,$67,$E8,$7E,$FF,$E0,$AD,$6C
	.dc.b	$FD,$A5,$7F,$03,$F8,$2F,$DC,$F8,$3F,$C1,$7E,$1F,$F7,$7E,$1E,$6F
	.dc.b	$F7,$4E,$5F,$DE,$4F,$F0,$BF,$87,$FE,$09,$ED,$FF,$11,$9F,$E2,$71
	.dc.b	$8F,$37,$FC,$DF,$FA,$17,$FA,$9A,$BF,$D4,$55,$A3,$F8,$BF,$A6,$8E
	.dc.b	$51,$FB,$4D,$FD,$AF,$A4,$FD,$CE,$FE,$D7,$F8,$73,$F0,$5F,$69,$FC
	.dc.b	$93,$5F,$6B,$FC,$9B,$5F,$6B,$93,$DA,$64,$F6,$9F,$E3,$AF,$17,$F5
	.dc.b	$39,$7F,$AC,$63,$F9,$96,$2F,$47,$DB,$F1,$3F,$09,$F7,$FC,$45,$AE
	.dc.b	$5E,$5F,$3C,$3F,$C8,$51,$FC,$F3,$6F,$81,$EA,$FC,$6F,$87,$F9,$7E
	.dc.b	$8F,$C8,$F8,$7F,$9B,$E9,$E7,$0F,$8F,$B2,$35,$18,$AF,$9F,$F0,$9F
	.dc.b	$53,$CE,$B3,$C3,$CE,$FA,$7F,$0B,$F6,$1C,$C7,$8F,$15,$7D,$8F,$85
	.dc.b	$FB,$BE,$77,$B2,$BC,$3E,$76,$5A,$7B,$87,$33,$8B,$E4,$47,$F5,$3C
	.dc.b	$57,$EB,$8F,$06,$BE,$A7,$C5,$B6,$5F,$9E,$E5,$FA,$31,$E8,$7E,$9E
	.dc.b	$95,$EC,$FB,$9F,$80,$79,$71,$F2,$6D,$F1,$63,$19,$3E,$17,$F3,$7E
	.dc.b	$37,$B6,$7E,$23,$F9,$A8,$CC,$70,$DB,$F4,$3E,$7F,$9E,$3D,$D9,$F9
	.dc.b	$DF,$81,$A6,$5F,$F4,$26,$DF,$65,$F0,$F1,$7E,$01,$F1,$E9,$F5,$5A
	.dc.b	$7A,$8F,$F5,$0E,$8E,$2D,$46,$27,$D8,$EB,$E1,$BF,$6B,$E7,$7F,$42
	.dc.b	$E8,$FC,$94,$7E,$E3,$CE,$C5,$7D,$8C,$7F,$DC,$6F,$E0,$EF,$3C,$5F
	.dc.b	$6E,$F8,$61,$F2,$3E,$1B,$F8,$0F,$93,$AC,$7F,$DD,$18,$FE,$2E,$31
	.dc.b	$B7,$F0,$6F,$27,$57,$C3,$0F,$A3,$F0,$9F,$C2,$BC,$4E,$4F,$F0,$CF
	.dc.b	$2B,$A7,$C2,$66,$B7,$9F,$E8,$A7,$C1,$AF,$F1,$2F,$16,$CF,$E9,$BC
	.dc.b	$5F,$27,$C3,$E7,$0F,$BF,$7F,$C6,$27,$C7,$5F,$AC,$DE,$BE,$9A,$FC
	.dc.b	$2E,$FE,$97,$E0,$46,$DF,$1B,$FA,$9D,$7D,$2F,$ED,$9F,$4B,$FB,$8F
	.dc.b	$27,$2E,$E7,$F2,$3F,$0B,$FE,$2D,$FD,$63,$FD,$93,$18,$9F,$E4,$B0
	.dc.b	$F1,$FE,$77,$C3,$6E,$7E,$85,$F1,$9F,$A5,$F0,$FC,$FC,$BB,$5C,$BC
	.dc.b	$8F,$AD,$F1,$19,$CF,$F4,$37,$E1,$9F,$B5,$F0,$FE,$3F,$02,$E1,$E2
	.dc.b	$EF,$FB,$DF,$11,$9C,$7F,$53,$7E,$99,$FC,$17,$C1,$7A,$FC,$2F,$C4
	.dc.b	$FE,$87,$FE,$E5,$C7,$FD,$C9,$C5,$F9,$9C,$7F,$63,$7F,$07,$C7,$DC
	.dc.b	$71,$3C,$B5,$F0,$B8,$3D,$6F,$F2,$9B,$6D,$EC,$8D,$57,$07,$51,$EA
	.dc.b	$78,$6D,$5F,$F3,$65,$34,$F6,$C7,$D0,$9D,$3F,$E6,$B9,$D5,$FF,$0B
	.dc.b	$CF,$FC,$67,$3E,$3F,$23,$DF,$F0,$DF,$6F,$AF,$E0,$FC,$1F,$DD,$9C
	.dc.b	$D7,$DC,$9F,$8B,$F0,$FF,$81,$EE,$B3,$F8,$0F,$0F,$E4,$FC,$46,$74
	.dc.b	$F0,$5F,$E1,$E7,$E6,$FC,$47,$E2,$7B,$EE,$1F,$88,$F1,$7E,$8F,$C4
	.dc.b	$67,$1F,$45,$FE,$3E,$7E,$AF,$C2,$FE,$47,$C7,$B8,$7E,$43,$C0,$FB
	.dc.b	$3F,$0B,$9C,$7E,$05,$FE,$5E,$7C,$F5,$F4,$F7,$AF,$CC,$D7,$D2,$FE
	.dc.b	$6A,$32,$FC,$B8,$F3,$BF,$6F,$5A,$FD,$4C,$FD,$2F,$E8,$A3,$2F,$D3
	.dc.b	$8F,$C6,$FC,$0F,$EA,$7C,$3F,$C7,$FC,$17,$EA,$FC,$2C,$DF,$EA,$A7
	.dc.b	$2F,$EB,$23,$C6,$FE,$53,$2B,$EC,$BA,$9A,$7A,$AB,$DD,$BF,$D2,$7E
	.dc.b	$B3,$F3,$1F,$33,$5F,$CB,$69,$EA,$7D,$D1,$97,$DC,$FE,$2F,$E1,$7F
	.dc.b	$81,$78,$BF,$A8,$F8,$3B,$39,$AF,$E0,$7E,$2E,$5F,$E0,$B5,$FA,$EF
	.dc.b	$85,$FE,$15,$E3,$FE,$C3,$E1,$6E,$1F,$97,$9F,$E1,$DE,$3E,$70,$FE
	.dc.b	$1B,$7F,$B6,$F8,$5F,$F5,$F7,$EB,$2C,$31,$5F,$E3,$68,$C4,$7E,$87
	.dc.b	$E1,$B7,$E2,$6A,$BF,$C3,$95,$A7,$FC,$07,$5F,$E2,$BF,$15,$F0,$DD
	.dc.b	$BC,$6F,$E0,$BF,$07,$E0,$F1,$3F,$83,$CE,$6F,$F5,$D3,$97,$D5,$87
	.dc.b	$A1,$FD,$97,$07,$BA,$3B,$9F,$E1,$2C,$5C,$1F,$02,$B7,$3C,$4D,$FA
	.dc.b	$EA,$6F,$F0,$AB,$D9,$5F,$EF,$1D,$7D,$97,$FE,$AC,$B6,$FF,$0A,$FF
	.dc.b	$82,$61,$C5,$FE,$0D,$F0,$BF,$33,$C7,$FE,$47,$F0,$DF,$3B,$C7,$FE
	.dc.b	$4B,$39,$AF,$E1,$93,$97,$FC,$31,$E6,$7F,$99,$35,$F1,$79,$3F,$DD
	.dc.b	$23,$11,$FE,$FF,$D9,$39,$F8,$6F,$B7,$E4,$6F,$E2,$BE,$EC,$F1,$F8
	.dc.b	$7F,$BF,$E5,$F3,$F8,$7B,$A7,$82,$7F,$07,$E5,$67,$4F,$C1,$5F,$87
	.dc.b	$E0,$7F,$0F,$E5,$D9,$FC,$37,$89,$E5,$F8,$1C,$E5,$EA,$BF,$C6,$CF
	.dc.b	$8E,$BF,$19,$A5,$7F,$12,$FA,$5F,$C8,$6D,$97,$E0,$DF,$95,$F1,$BF
	.dc.b	$1B,$5F,$4C,$7D,$CE,$1F,$F3,$B3,$F9,$9D,$7D,$4F,$E5,$B5,$F9,$5F
	.dc.b	$0F,$FA,$0F,$1A,$CF,$E7,$FB,$3E,$77,$C2,$E7,$3F,$A7,$7F,$A6,$9F
	.dc.b	$A5,$F0,$DF,$A8,$EF,$B8,$7E,$9F,$B3,$EB,$7C,$3E,$71,$FB,$77,$FA
	.dc.b	$E9,$FB,$5F,$0F,$FB,$0E,$FB,$87,$EB,$FC,$3F,$BD,$F0,$F9,$C7,$F0
	.dc.b	$77,$FB,$69,$FC,$17,$C3,$7E,$E3,$C5,$B8,$7E,$DF,$C1,$FC,$37,$C3
	.dc.b	$E7,$1F,$C5,$DF,$EF,$A7,$CF,$5F,$B8,$DE,$BF,$9F,$D7,$D2,$FF,$01
	.dc.b	$8C,$BF,$8F,$8F,$3B,$FC,$03,$5A,$FE,$9F,$3F,$4B,$FC,$26,$32,$FE
	.dc.b	$5E,$3F,$35,$F0,$5F,$C4,$3C,$3F,$CE,$7C,$35,$D3,$F4,$32,$BF,$F6
	.dc.b	$3F,$EB,$4F,$0F,$71,$FC,$57,$0F,$2A,$F5,$1E,$33,$6C,$C6,$36,$F6
	.dc.b	$D7,$C1,$CF,$EA,$BF,$65,$F9,$EF,$A5,$AF,$F1,$ED,$7D,$AF,$C1,$8C
	.dc.b	$BF,$05,$FD,$CF,$C2,$7A,$63,$FD,$81,$A3,$C7,$FD,$D4,$E6,$75,$8A
	.dc.b	$BF,$D9,$D7,$FC,$0B,$E1,$F5,$5F,$0E,$B5,$87,$28,$F8,$BE,$47,$F0
	.dc.b	$5F,$87,$DC,$7C,$3C,$5E,$1F,$F0,$9F,$87,$CE,$3D,$E3,$85,$7C,$EF
	.dc.b	$23,$4F,$D0,$8F,$E2,$1F,$0F,$C3,$0F,$F6,$96,$BF,$EA,$DF,$D0,$FB
	.dc.b	$DF,$7F,$0B,$6F,$D8,$F2,$FF,$D1,$73,$AC,$FD,$D5,$8F,$88,$E9,$E2
	.dc.b	$DC,$3E,$FF,$87,$C3,$E1,$B3,$9F,$7B,$F1,$4F,$A2,$BE,$6E,$DE,$86
	.dc.b	$D7,$F2,$AE,$1F,$47,$06,$67,$E7,$EF,$F1,$2B,$F9,$7F,$07,$B7,$83
	.dc.b	$CF,$5E,$FD,$BD,$6F,$DA,$8C,$BF,$94,$8F,$6F,$C3,$7C,$5F,$1B,$DF
	.dc.b	$F0,$D7,$1F,$8D,$3F,$27,$C6,$CE,$1F,$9C,$BF,$8B,$E6,$7C,$DF,$22
	.dc.b	$CF,$E8,$3C,$6F,$93,$F1,$19,$CB,$E8,$5F,$D3,$9F,$9B,$F1,$1F,$57
	.dc.b	$C9,$B8,$7E,$A3,$C6,$FA,$3F,$11,$9C,$7E,$C5,$FD,$B9,$FA,$BF,$11
	.dc.b	$F7,$7C,$9F,$AF,$F0,$76,$FE,$F4,$FE,$07,$C4,$CE,$5F,$B4,$BF,$3B
	.dc.b	$F9,$ED,$EB,$F3,$B5,$F4,$BF,$BA,$8C,$BF,$9B,$E7,$4F,$A2,$3F,$4F
	.dc.b	$B5,$7E,$BE,$9E,$A7,$F8,$1C,$65,$E5,$1F,$8E,$F8,$2F,$CB,$78,$BF
	.dc.b	$90,$F8,$6F,$CC,$76,$66,$FF,$85,$CE,$5F,$E1,$B1,$F9,$6F,$83,$FC
	.dc.b	$F7,$8B,$F9,$8F,$88,$FD,$07,$76,$6F,$F8,$9C,$E5,$FE,$2B,$1F,$9E
	.dc.b	$F8,$8F,$D6,$ED,$FE,$BC,$9F,$F0,$7F,$8E,$F8,$8C,$F8,$BB,$9F,$DE
	.dc.b	$F9,$1F,$09,$FD,$3C,$EB,$A3,$F5,$DF,$11,$D3,$C8,$FD,$5F,$C3,$59
	.dc.b	$CD,$F2,$CF,$FD,$4B,$1F,$B5,$9D,$4F,$8B,$83,$DB,$3F,$B9,$E0,$F6
	.dc.b	$BF,$BB,$8C,$CF,$A2,$72,$FA,$B6,$F6,$BE,$63,$C1,$AF,$AD,$FE,$39
	.dc.b	$AF,$AF,$0F,$E0,$DA,$FA,$E7,$E1,$FC,$3F,$C6,$F1,$7F,$82,$FC,$37
	.dc.b	$C8,$F0,$BF,$83,$CE,$70,$FE,$1D,$39,$7C,$D7,$F4,$9F,$53,$F4,$9C
	.dc.b	$BF,$CC,$39,$3E,$B3,$FC,$CB,$93,$CC,$FF,$30,$E1,$FA,$DF,$07,$FE
	.dc.b	$61,$82,$E5,$FB,$11,$EE,$C3,$F9,$07,$07,$C2,$7D,$B1,$93,$EF,$FC
	.dc.b	$3F,$E0,$BC,$5F,$E4,$2F,$E1,$3C,$4E,$D8,$CF,$D0,$BF,$DC,$DC,$7F
	.dc.b	$A2,$BD,$E1,$F6,$EF,$78,$FF,$63,$C1,$87,$FB,$4A,$BD,$5E,$36,$9F
	.dc.b	$4C,$3E,$1F,$F2,$73,$9A,$E3,$F1,$1F,$96,$F1,$F1,$7B,$C3,$F1,$57
	.dc.b	$8F,$88,$FC,$DF,$93,$67,$F1,$DE,$1F,$0F,$87,$CE,$BE,$53,$CD,$8B
	.dc.b	$0F,$D1,$D7,$6F,$87,$FD,$2F,$8B,$71,$FC,$C3,$E2,$F8,$5F,$D4,$F8
	.dc.b	$39,$9F,$C5,$F4,$2B,$CF,$C1,$EC,$D3,$C1,$5F,$03,$8B,$0F,$F9,$FB
	.dc.b	$5F,$8C,$FF,$69,$C1,$C3,$F0,$2B,$7C,$5F,$4F,$7F,$5D,$F1,$7D,$38
	.dc.b	$F6,$9F,$A3,$C1,$A7,$85,$7C,$49,$D6,$4F,$85,$FE,$07,$E1,$7C,$EF
	.dc.b	$84,$FE,$0B,$D5,$F4,$27,$37,$C2,$72,$F2,$F8,$5F,$E1,$7E,$1F,$D6
	.dc.b	$F8,$7F,$E1,$BE,$47,$D8,$9C,$D7,$82,$72,$F8,$BE,$1F,$F8,$9F,$91
	.dc.b	$F7,$BE,$1E,$CE,$77,$FE,$11,$E3,$65,$FE,$15,$5F,$84,$F8,$7F,$E3
	.dc.b	$7E,$3F,$E1,$BE,$07,$F8,$EF,$8B,$9B,$FE,$21,$39,$7F,$89,$47,$95
	.dc.b	$FE,$29,$BF,$A5,$FE,$2D,$BE,$5F,$BB,$19,$7E,$FC,$7E,$53,$5F,$3C
	.dc.b	$7D,$BD,$6F,$2F,$EA,$67,$F9,$7E,$BE,$9A,$FC,$3E,$BF,$9C,$F8,$1F
	.dc.b	$F0,$FF,$EB,$7F,$EB,$AE,$0F,$CA,$FC,$36,$3B,$B5,$87,$F3,$2A,$D3
	.dc.b	$FA,$A9,$D5,$7D,$CF,$83,$E7,$DB,$FA,$8F,$88,$EB,$DD,$FA,$A9,$CD
	.dc.b	$7E,$0A,$72,$FE,$13,$E2,$3C,$7D,$FF,$B0,$F8,$7B,$39,$CF,$C5,$E3
	.dc.b	$65,$F2,$57,$EE,$7F,$43,$FF,$60,$C7,$F9,$7F,$6F,$F6,$0D,$FE,$56
	.dc.b	$7F,$C9,$D3,$EB,$F1,$F8,$7C,$4F,$C2,$F0,$79,$7C,$47,$C4,$9F,$45
	.dc.b	$7C,$1E,$1F,$63,$F0,$F8,$72,$FF,$3C,$CF,$F8,$66,$7E,$D8,$D3,$E9
	.dc.b	$75,$A7,$A5,$B1,$97,$FC,$BF,$2D,$75,$18,$AF,$77,$C4,$7D,$8F,$0E
	.dc.b	$CF,$D5,$F0,$FE,$17,$C3,$67,$2F,$81,$7F,$7A,$7C,$F5,$DB,$4F,$5B
	.dc.b	$E9,$D3,$2F,$F9,$AF,$4F,$5B,$FC,$1F,$93,$EB,$BF,$C2,$79,$3E,$AB
	.dc.b	$FC,$2F,$92,$F6,$7F,$39,$9F,$C6,$F0,$7A,$AB,$A7,$07,$D7,$F8,$6F
	.dc.b	$C8,$F8,$BF,$67,$E1,$BF,$27,$DF,$9B,$FC,$84,$E5,$FC,$94,$7D,$FF
	.dc.b	$86,$FC,$CF,$83,$67,$F2,$CF,$E0,$FE,$23,$F3,$BE,$46,$67,$EB,$4F
	.dc.b	$E1,$FE,$23,$F4,$3E,$4F,$E2,$7E,$1B,$F4,$7E,$26,$6F,$F4,$13,$97
	.dc.b	$F4,$51,$F8,$FF,$87,$FD,$4F,$89,$67,$F4,$CF,$E4,$FE,$17,$F5,$BE
	.dc.b	$2E,$67,$F0,$D3,$E3,$7E,$FE,$75,$FA,$0F,$A5,$FD,$86,$99,$7F,$39
	.dc.b	$1E,$57,$F2,$19,$79,$5F,$C9,$65,$97,$F1,$71,$97,$F1,$B1,$FA,$7F
	.dc.b	$47,$F8,$0F,$4D,$9F,$DF,$3F,$AB,$F8,$4F,$E0,$BD,$19,$9F,$CF,$4F
	.dc.b	$EB,$FE,$13,$F8,$4F,$5F,$EC,$7E,$1B,$FD,$53,$FA,$E7,$FB,$4F,$C3
	.dc.b	$FD,$85,$69,$FD,$94,$EB,$C9,$FD,$C7,$C4,$70,$F2,$7F,$75,$F0,$FC
	.dc.b	$BC,$7C,$EB,$FE,$42,$CF,$FD,$8D,$97,$EE,$EB,$4F,$EF,$7E,$1F,$C3
	.dc.b	$E4,$FF,$03,$F8,$3F,$1F,$87,$FC,$16,$73,$5F,$C0,$E7,$2F,$9B,$1F
	.dc.b	$E3,$DC,$1E,$B7,$C1,$B7,$BD,$CC,$FA,$9C,$BF,$C3,$76,$F5,$3E,$58
	.dc.b	$F4,$6B,$EA,$7F,$C8,$3A,$FA,$9F,$F1,$5C,$FF,$86,$7C,$E6,$DE,$C9
	.dc.b	$C9,$FE,$30,$87,$C1,$FC,$EF,$17,$F9,$03,$F4,$3C,$3F,$F0,$D4,$7F
	.dc.b	$B2,$F1,$FF,$69,$DF,$2F,$21,$87,$28,$F9,$97,$C3,$E1,$BE,$CF,$8B
	.dc.b	$CB,$E1,$EE,$BF,$DF,$73,$57,$DE,$F1,$B5,$8F,$D5,$C3,$C3,$F1,$3F
	.dc.b	$82,$F1,$EC,$FD,$97,$CB,$F1,$1F,$86,$F2,$33,$3F,$12,$7D,$3F,$11
	.dc.b	$F8,$AF,$27,$D7,$F0,$D7,$6F,$14,$FE,$3B,$C4,$CE,$1F,$85,$BF,$53
	.dc.b	$E3,$AF,$47,$07,$AA,$3E,$17,$07,$A9,$FC,$66,$FE,$A7,$EA,$EF,$E4
	.dc.b	$44,$F8,$77,$2B,$6D,$4E,$63,$19,$7C,$7E,$0F,$73,$F2,$78,$7E,$23
	.dc.b	$F2,$FE,$1F,$F4,$FE,$37,$D8,$6C,$7C,$FF,$87,$FD,$5F,$8B,$9C,$BF
	.dc.b	$45,$7E,$E7,$CF,$5E,$E7,$E0,$1E,$EE,$6F,$AD,$19,$8D,$9F,$0B,$1F
	.dc.b	$1B,$4F,$AD,$1E,$97,$F6,$1C,$DE,$E5,$B7,$C2,$DB,$F1,$FF,$07,$FC
	.dc.b	$03,$C6,$B5,$FB,$77,$F2,$7F,$0F,$FC,$13,$C0,$CC,$FE,$8A,$7F,$2F
	.dc.b	$F0,$FF,$C2,$3C,$0F,$CC,$FC,$35,$D7,$F2,$13,$FC,$33,$C6,$CE,$1F
	.dc.b	$C0,$EF,$F3,$FF,$0F,$FC,$43,$C7,$B3,$FC,$1D,$FD,$1F,$C3,$7F,$14
	.dc.b	$F1,$73,$3F,$B2,$9F,$D3,$FC,$37,$DE,$8F,$F5,$AF,$EB,$BF,$EA,$0D
	.dc.b	$3F,$D1,$9D,$FF,$E9,$E9,$C9,$C0,$FF,$6A,$CF,$E9,$27,$F4,$DF,$0B
	.dc.b	$CE,$7A,$72,$AB,$A4,$7F,$AA,$AD,$BF,$BE,$3F,$2B,$C1,$7C,$8F,$F5
	.dc.b	$1D,$6A,$7F,$91,$70,$FC,$27,$F8,$5F,$0F,$B1,$EF,$DE,$DF,$11,$F9
	.dc.b	$57,$E1,$BF,$CA,$F8,$32,$FD,$C9,$E1,$BF,$EF,$FE,$17,$D5,$3F,$A1
	.dc.b	$D9,$E1,$FF,$0E,$F8,$4C,$CF,$E1,$27,$FD,$DA,$FE,$BA,$B3,$C3,$F0
	.dc.b	$6F,$84,$FC,$8E,$CF,$E2,$FF,$0B,$AC,$1E,$1E,$2B,$E6,$F9,$1F,$C7
	.dc.b	$3E,$1F,$53,$99,$FA,$F3,$FC,$83,$E1,$F9,$79,$7F,$EB,$BF,$D0,$FF
	.dc.b	$D9,$38,$A3,$57,$F6,$3C,$8D,$70,$F0,$3F,$D3,$57,$FE,$A4,$9F,$67
	.dc.b	$91,$CF,$E2,$3C,$CF,$5F,$84,$F4,$78,$B9,$AF,$C1,$CF,$95,$FC,$0E
	.dc.b	$7E,$CC,$7F,$95,$70,$7B,$F6,$F6,$57,$E2,$B7,$E0,$FC,$F8,$F2,$BF
	.dc.b	$8E,$DA,$BF,$96,$7D,$6F,$E6,$36,$CB,$E0,$8F,$85,$F0,$9F,$37,$C6
	.dc.b	$B3,$F9,$87,$E3,$7C,$17,$D1,$F0,$73,$3F,$A2,$9F,$95,$F0,$7F,$57
	.dc.b	$C2,$F9,$9F,$0F,$77,$F8,$F3,$F6,$7C,$7C,$E1,$FA,$4B,$FA,$5F,$11
	.dc.b	$F7,$7C,$8F,$A9,$F1,$1F,$7F,$C9,$CD,$FE,$AA,$72,$FE,$B2,$3E,$D7
	.dc.b	$C4,$7E,$17,$CA,$B3,$FB,$0F,$0B,$EF,$7C,$3E,$76,$FB,$37,$F8,$B9
	.dc.b	$F4,$D7,$8E,$7E,$EE,$FE,$98,$FE,$C9,$DD,$FE,$6F,$7F,$77,$BF,$A9
	.dc.b	$FE,$07,$BF,$99,$FE,$0B,$AF,$99,$FE,$0F,$AE,$5F,$BF,$19,$7F,$03
	.dc.b	$1F,$93,$F8,$7F,$CF,$78,$B6,$7F,$84,$EF,$F9,$7F,$84,$CE,$BF,$92
	.dc.b	$BF,$65,$79,$B8,$FD,$1E,$6A,$BF,$AA,$AD,$9B,$EA,$73,$1E,$DB,$FE
	.dc.b	$3B,$BF,$B5,$FC,$AE,$FE,$97,$F7,$DB,$FA,$5F,$E0,$1B,$E5,$FD,$F4
	.dc.b	$65,$FE,$01,$1F,$AF,$F8,$5F,$DD,$77,$59,$FE,$4B,$E2,$FE,$CF,$E1
	.dc.b	$73,$A7,$EB,$AF,$F8,$04,$FE,$DF,$E1,$7F,$81,$78,$FF,$B9,$F8,$6B
	.dc.b	$7F,$B4,$9F,$E1,$7C,$9F,$ED,$2E,$2F,$81,$5A,$CF,$ED,$FC,$3E,$BC
	.dc.b	$9F,$E0,$4D,$C7,$F7,$BF,$0D,$CB,$CA,$CD,$FF,$1E,$F3,$59,$B1,$FE
	.dc.b	$15,$F0,$FE,$8A,$FC,$3B,$74,$FA,$50,$7E,$97,$C2,$79,$BC,$6E,$33
	.dc.b	$FD,$1B,$93,$DC,$FF,$48,$E1,$F7,$37,$6F,$73,$E3,$CF,$DD,$D2,$E1
	.dc.b	$F7,$67,$E0,$3F,$01,$C3,$ED,$BF,$82,$71,$DB,$F9,$17,$0F,$BF,$2F
	.dc.b	$E3,$B5,$BB,$E5,$FA,$17,$FB,$A3,$87,$E0,$F8,$FD,$6F,$F1,$B3,$FE
	.dc.b	$E8,$CB,$FA,$E5,$7C,$6E,$DD,$47,$C1,$9D,$3F,$D8,$36,$61,$AC,$55
	.dc.b	$DE,$7F,$DB,$D8,$FE,$68,$EF,$F0,$9F,$7B,$BB,$85,$23,$F0,$3E,$36
	.dc.b	$1D,$57,$1C,$BC,$5F,$0F,$A8,$C6,$5E,$0F,$87,$FC,$4F,$91,$D1,$B3
	.dc.b	$AC,$91,$F8,$EF,$1F,$C3,$EC,$EB,$2F,$F3,$55,$3E,$17,$F2,$7C,$1E
	.dc.b	$67,$F2,$BC,$36,$BE,$D4,$FA,$BE,$0B,$39,$F7,$DF,$D3,$5D,$1F,$43
	.dc.b	$4F,$C1,$70,$FC,$2E,$0C,$F0,$7D,$8D,$FC,$8D,$AF,$C4,$70,$7B,$5F
	.dc.b	$C5,$70,$7A,$9C,$E5,$EF,$F8,$3F,$D5,$F9,$9F,$3B,$DF,$FD,$6F,$95
	.dc.b	$6F,$F2,$53,$F4,$AB,$37,$F1,$FE,$03,$F6,$7C,$FF,$5B,$E1,$3F,$6B
	.dc.b	$E3,$DC,$7F,$35,$3F,$6A,$B3,$7F,$3F,$E1,$7F,$77,$E4,$FD,$E6,$C6
	.dc.b	$6F,$F4,$5E,$47,$E0,$BE,$13,$31,$F4,$BE,$17,$C6,$FD,$4F,$83,$BC
	.dc.b	$9F,$9C,$7D,$B8,$7D,$DA,$CB,$F9,$CC,$FD,$2F,$E0,$76,$F2,$E1,$F7
	.dc.b	$9C,$BF,$80,$9C,$E1,$FA,$8C,$FC,$6D,$9F,$DC,$69,$EB,$7D,$38,$7E
	.dc.b	$37,$3D,$BF,$AC,$F8,$4F,$E3,$7E,$2F,$E7,$9B,$1F,$93,$7C,$F5,$FA
	.dc.b	$DF,$83,$CB,$E3,$3E,$6C,$FA,$5F,$E1,$1C,$3F,$45,$FD,$BF,$0F,$F1
	.dc.b	$C9,$FB,$11,$A7,$E7,$3A,$CB,$D0,$D8,$CB,$FB,$BE,$1F,$A4,$FB,$A3
	.dc.b	$F5,$1F,$09,$FC,$D3,$C5,$FD,$BB,$FC,$DB,$C6,$B5,$FC,$6E,$7F,$77
	.dc.b	$F0,$99,$BF,$D8,$7C,$27,$F3,$CF,$3F,$F8,$37,$E8,$5F,$E4,$CC,$3F
	.dc.b	$D1,$FA,$A7,$31,$FD,$0B,$D0,$FE,$0B,$7C,$70,$FD,$FE,$1F,$EB,$DA
	.dc.b	$FE,$2F,$FA,$F3,$E2,$7A,$FC,$47,$F9,$52,$3F,$DE,$5C,$FF,$CA,$EF
	.dc.b	$85,$FF,$9E,$27,$86,$1D,$3E,$1F,$87,$C5,$79,$BE,$23,$97,$93,$E8
	.dc.b	$F8,$5B,$7E,$27,$DD,$8F,$FB,$C7,$25,$EA,$3F,$85,$ED,$E9,$74,$5D
	.dc.b	$3C,$8F,$BF,$0F,$8E,$FF,$45,$E0,$F5,$BF,$D6,$75,$F3,$BD,$F6,$F3
	.dc.b	$D7,$C4,$7E,$06,$FF,$3B,$7C,$D7,$97,$E1,$7E,$27,$95,$F4,$BE,$12
	.dc.b	$E3,$F3,$1F,$91,$E5,$FD,$6A,$CD,$FB,$7E,$17,$E6,$7A,$7F,$69,$F9
	.dc.b	$DE,$95,$9F,$34,$FD,$EF,$87,$CE,$5F,$63,$E2,$3E,$A7,$A5,$F8,$2F
	.dc.b	$86,$B3,$F7,$1F,$B1,$E6,$FE,$1A,$B3,$8F,$CD,$F8,$7F,$B9,$E8,$FE
	.dc.b	$29,$B8,$7E,$11,$FC,$07,$99,$F8,$EF,$83,$CD,$FA,$4F,$2D,$7E,$2B
	.dc.b	$7F,$2C,$64,$FC,$5E,$DE,$97,$E8,$6F,$E3,$7E,$96,$DE,$96,$DF,$A5
	.dc.b	$FA,$DA,$ED,$FC,$AC,$FE,$83,$E1,$3F,$29,$E7,$FE,$8B,$E1,$2E,$1F
	.dc.b	$9C,$7F,$31,$E8,$6F,$0F,$C0,$57,$EA,$3E,$13,$F3,$9E,$8D,$9F,$C2
	.dc.b	$35,$FD,$1F,$A5,$FA,$EF,$85,$DE,$5F,$A8,$F8,$7F,$D2,$7A,$BF,$B2
	.dc.b	$F8,$7F,$D3,$7A,$B7,$1F,$C7,$4F,$ED,$AB,$37,$FB,$0F,$84,$FD,$67
	.dc.b	$AD,$FB,$AF,$87,$B3,$FB,$47,$F6,$1E,$A7,$EF,$AB,$38,$FE,$97,$7F
	.dc.b	$1B,$4F,$4D,$FE,$F2,$3F,$71,$B7,$A6,$72,$7E,$B7,$2F,$4D,$7F,$02
	.dc.b	$D7,$CB,$3F,$B1,$72,$EC,$FD,$46,$BF,$C4,$3E,$17,$F8,$27,$A7,$FE
	.dc.b	$17,$F8,$0F,$F1,$ED,$7F,$0A,$F3,$71,$8E,$8F,$E1,$F7,$FC,$2E,$3E
	.dc.b	$74,$FF,$09,$E3,$A8,$7C,$1B,$CC,$4E,$A7,$30,$DB,$DD,$1F,$C6,$A7
	.dc.b	$E8,$3F,$4B,$1F,$73,$EA,$6B,$FC,$83,$93,$6F,$F2,$2D,$7D,$9F,$09
	.dc.b	$FE,$F4,$FD,$7F,$C9,$7F,$E2,$6D,$BF,$DC,$1A,$25,$E7,$70,$C7,$FC
	.dc.b	$CD,$1C,$1F,$F3,$05,$FF,$A4,$AD,$F1,$1D,$6F,$E5,$7C,$47,$7F,$3E
	.dc.b	$E1,$CB,$D3,$AF,$F9,$33,$24,$FD,$1F,$85,$F3,$57,$D3,$CB,$D1,$EB
	.dc.b	$5C,$3C,$6D,$79,$79,$FF,$67,$E1,$77,$B7,$B7,$E1,$FE,$07,$A7,$F7
	.dc.b	$7E,$1B,$E1,$79,$97,$1F,$6C,$FE,$06,$B3,$7F,$17,$83,$D4,$DA,$F5
	.dc.b	$35,$F5,$6F,$E9,$72,$7C,$63,$FD,$0D,$AF,$FA,$D9,$F4,$CF,$D1,$8F
	.dc.b	$3C,$6F,$D8,$47,$FA,$E6,$3F,$14,$7E,$2B,$85,$B7,$BE,$39,$9C,$65
	.dc.b	$F0,$DF,$7F,$D5,$FC,$E7,$C2,$DC,$BF,$00,$FE,$0F,$D5,$FD,$05,$67
	.dc.b	$0E,$5F,$0D,$F8,$7F,$5F,$F4,$8D,$C3,$F1,$0F,$E2,$FD,$6F,$D4,$7C
	.dc.b	$06,$6F,$C5,$F0,$BF,$8F,$F5,$FF,$58,$FE,$47,$D3,$B3,$F8,$C9,$FD
	.dc.b	$87,$C2,$E7,$2F,$CA,$7C,$2F,$E5,$FD,$5F,$DA,$36,$33,$7F,$94,$F3
	.dc.b	$BF,$71,$F0,$D9,$8F,$49,$F9,$8D,$FD,$2D,$AF,$4B,$F9,$CD,$BD,$2E
	.dc.b	$4F,$49,$E5,$8F,$D3,$6B,$E5,$8C,$9F,$99,$D3,$D2,$FE,$9B,$5F,$E1
	.dc.b	$5F,$0B,$FA,$FF,$57,$F8,$67,$C3,$5C,$7F,$60,$FE,$CF,$D2,$DE,$1F
	.dc.b	$AE,$AF,$E2,$5F,$05,$FB,$7F,$4F,$F8,$A7,$C3,$7E,$E7,$D5,$B7,$FB
	.dc.b	$49,$DB,$FB,$68,$FF,$58,$7E,$85,$FE,$BD,$EF,$7B,$18,$F8,$4F,$E0
	.dc.b	$5E,$BD,$C3,$F7,$B3,$C3,$E1,$F3,$C1,$FE,$AA,$C7,$FD,$65,$5F,$EE
	.dc.b	$7F,$D8,$3F,$DE,$51,$88,$FF,$62,$C7,$FB,$A6,$78,$7C,$47,$E6,$27
	.dc.b	$F3,$3E,$9A,$B8,$FC,$3E,$67,$F9,$34,$E7,$D5,$F5,$C7,$F0,$5E,$0F
	.dc.b	$7B,$FC,$1F,$82,$BD,$38,$13,$C2,$31,$3E,$D7,$F8,$76,$D5,$E1,$BD
	.dc.b	$7C,$37,$F1,$B5,$F8,$1B,$FC,$29,$FD,$A6,$FF,$EC,$67,$E1,$7A,$BF
	.dc.b	$3B,$E1,$3D,$51,$FC,$FB,$83,$F9,$1E,$1F,$26,$73,$BF,$C1,$F3,$7E
	.dc.b	$AF,$C3,$E2,$F7,$87,$EE,$79,$BF,$01,$8F,$F3,$5E,$6F,$C0,$36,$7F
	.dc.b	$00,$FF,$39,$9C,$C7,$9B,$9B,$F9,$F7,$0F,$B6,$3F,$84,$70,$DD,$3F
	.dc.b	$9F,$4F,$DF,$9F,$E7,$B3,$F8,$8F,$85,$F2,$7A,$FF,$8A,$F8,$4F,$37
	.dc.b	$9D,$6F,$F0,$53,$B7,$F0,$91,$F9,$0F,$85,$F6,$7A,$FF,$92,$F8,$7F
	.dc.b	$77,$AB,$6F,$F1,$53,$B7,$F1,$91,$F9,$8F,$87,$F8,$9E,$CF,$E6,$BE
	.dc.b	$1B,$E3,$7A,$56,$FF,$25,$3B,$7F,$29,$1F,$A0,$F8,$7F,$99,$EA,$FE
	.dc.b	$8B,$E0,$7E,$77,$AB,$6F,$F3,$53,$B7,$F3,$91,$E3,$7F,$37,$A7,$9D
	.dc.b	$FC,$EE,$95,$FC,$DF,$52,$78,$46,$27,$C8,$FE,$93,$3F,$43,$FA,$6C
	.dc.b	$EB,$CA,$36,$F4,$8F,$DC,$FC,$17,$E1,$7D,$6B,$3F,$B1,$6B,$FB,$0F
	.dc.b	$3F,$F7,$FF,$05,$BE,$DF,$DC,$7C,$17,$E3,$7D,$8F,$E0,$6D,$8C,$DF
	.dc.b	$EE,$FD,$7F,$E0,$FF,$0F,$98,$FC,$FF,$C3,$FE,$57,$D8,$FE,$16,$FE
	.dc.b	$5F,$D4,$B3,$FC,$0E,$7F,$87,$FC,$3E,$71,$FE,$11,$F0,$FF,$9D,$F5
	.dc.b	$BF,$89,$BF,$9F,$F1,$EE,$1F,$C2,$E5,$5F,$E2,$7F,$80,$B2,$5F,$A1
	.dc.b	$7F,$B1,$7C,$8F,$E1,$58,$7B,$2B,$F8,$A7,$0D,$7F,$89,$B5,$FA,$FC
	.dc.b	$1E,$E7,$78,$FF,$1D,$DF,$DC,$D8,$FE,$3D,$BF,$B9,$FE,$3B,$BF,$B9
	.dc.b	$CC,$F9,$CF,$46,$7F,$EC,$A8,$FE,$37,$D4,$F2,$BD,$7B,$4A,$D5,$7F
	.dc.b	$91,$3D,$9F,$F5,$CC,$E7,$E1,$FF,$82,$5F,$0C,$7A,$FC,$1F,$F0,$AB
	.dc.b	$DC,$78,$3E,$1F,$1E,$A6,$F2,$F9,$19,$BE,$1F,$37,$E7,$9C,$C7,$3F
	.dc.b	$6F,$86,$BF,$3B,$1D,$D7,$F8,$4F,$0F,$F0,$7B,$FC,$DF,$95,$9B,$DB
	.dc.b	$E3,$5A,$AF,$E8,$15,$A9,$CF,$C2,$FA,$7D,$4D,$FC,$37,$AE,$78,$FC
	.dc.b	$3F,$B7,$D8,$E7,$F0,$FE,$FF,$56,$E7,$E1,$D5,$5C,$23,$13,$E2,$DB
	.dc.b	$D3,$BF,$CF,$E1,$BC,$5F,$59,$F8,$0F,$C3,$8F,$81,$7D,$78,$FC,$EF
	.dc.b	$7D,$6B,$EC,$CF,$E0,$5F,$1E,$2F,$81,$3F,$D7,$A5,$C1,$EE,$8E,$55
	.dc.b	$BF,$87,$FB,$5E,$B7,$C8,$F8,$5F,$B9,$EA,$DB,$F7,$4F,$CC,$AC,$DF
	.dc.b	$7D,$FD,$75,$F6,$8F,$44,$78,$F7,$F6,$C7,$D9,$9D,$EF,$FA,$AE,$5F
	.dc.b	$A1,$7E,$96,$D7,$DF,$7E,$B4,$7B,$79,$73,$5F,$99,$F8,$5F,$C9,$7A
	.dc.b	$DF,$80,$7F,$29,$EB,$5C,$7E,$B4,$FE,$13,$E1,$73,$8F,$C7,$E2,$FC
	.dc.b	$D7,$AF,$F8,$87,$F3,$9E,$BD,$C3,$EF,$4F,$E3,$3E,$1F,$37,$F3,$FE
	.dc.b	$1F,$F4,$5E,$CF,$E4,$1F,$D2,$7A,$77,$0F,$C3,$4F,$E5,$3E,$1B,$37
	.dc.b	$F5,$FE,$1F,$F5,$5E,$9F,$E6,$1B,$19,$BF,$C7,$7A,$7F,$9C,$F8,$6C
	.dc.b	$C7,$E8,$F7,$F4,$D7,$AE,$3F,$29,$BF,$A5,$FC,$36,$FE,$97,$33,$E9
	.dc.b	$7C,$AD,$7F,$19,$AF,$A5,$FC,$76,$BE,$97,$F4,$11,$FA,$EF,$86,$FE
	.dc.b	$05,$EA,$D9,$FD,$23,$5F,$CB,$79,$FF,$B4,$F8,$1D,$E3,$F9,$7F,$81
	.dc.b	$FE,$15,$EB,$FF,$00,$FD,$0B,$FD,$BF,$87,$E6,$7D,$8C,$5E,$8E,$11
	.dc.b	$FB,$18,$E1,$F1,$1F,$E1,$CF,$D8,$7F,$D8,$9D,$DF,$C7,$B0,$FF,$61
	.dc.b	$DF,$F0,$79,$CE,$1F,$CF,$67,$33,$FC,$1A,$7F,$D9,$F3,$FB,$39,$FF
	.dc.b	$69,$C7,$09,$FE,$85,$0F,$86,$F0,$7A,$5E,$9F,$85,$B8,$71,$7C,$9E
	.dc.b	$27,$B6,$B3,$87,$F1,$BD,$BD,$8F,$FA,$BF,$1F,$82,$FF,$94,$B9,$DB
	.dc.b	$7C,$28,$FE,$21,$BF,$F1,$EB,$FE,$65,$3F,$0A,$7D,$C7,$CB,$F8,$7F
	.dc.b	$88,$D3,$E1,$7A,$DB,$7E,$01,$F4,$7E,$1B,$E5,$79,$1F,$4F,$E1,$3E
	.dc.b	$67,$93,$6B,$E5,$4E,$DF,$99,$1F,$67,$E1,$3E,$97,$95,$67,$E8,$35
	.dc.b	$F9,$DE,$B7,$DF,$F8,$4D,$F2,$FC,$FF,$84,$FB,$5E,$C7,$E0,$FE,$1B
	.dc.b	$EE,$7A,$17,$1F,$B5,$3F,$87,$AC,$DF,$D7,$F8,$6F,$C1,$7A,$9F,$8B
	.dc.b	$6C,$66,$FF,$05,$E9,$7E,$3F,$E1,$B3,$1F,$47,$7F,$1B,$4F,$4D,$7E
	.dc.b	$0E,$3F,$1D,$B7,$A6,$72,$7D,$9C,$BD,$35,$F8,$9D,$7C,$B3,$F7,$5C
	.dc.b	$BB,$3F,$29,$AF,$E8,$7E,$1B,$F3,$5E,$97,$E8,$FE,$13,$F3,$9D,$96
	.dc.b	$FF,$35,$3B,$7F,$39,$1F,$35,$FF,$4E,$F8,$4E,$4C,$47,$FA,$72,$31
	.dc.b	$9F,$AE,$3F,$49,$AF,$AD,$FD,$15,$7C,$D7,$E7,$CF,$AD,$F3,$B5,$FD
	.dc.b	$47,$1E,$DF,$D5,$69,$FB,$BF,$84,$FD,$A7,$A9,$67,$F6,$5E,$A5,$7F
	.dc.b	$61,$1F,$C0,$7E,$0F,$F7,$53,$FC,$0F,$E1,$3F,$79,$EB,$5B,$FD,$D3
	.dc.b	$5F,$DC,$7A,$BF,$C2,$7E,$1B,$7A,$FE,$E7,$E1,$BF,$82,$7A,$DF,$C3
	.dc.b	$7E,$17,$F8,$37,$A7,$71,$FE,$09,$3F,$C4,$6B,$37,$FD,$CF,$89,$F0
	.dc.b	$FF,$C3,$7D,$2F,$F0,$AC,$63,$0F,$E1,$91,$89,$FE,$25,$E5,$6B,$E0
	.dc.b	$B8,$61,$FC,$2F,$83,$CB,$87,$EF,$DB,$E9,$7F,$17,$7F,$8D,$70,$7A
	.dc.b	$E7,$DF,$5E,$B7,$CE,$FF,$15,$DF,$CF,$1F,$C2,$5F,$7C,$6E,$3F,$8E
	.dc.b	$6F,$E8,$F8,$6F,$F6,$0F,$EC,$5F,$E5,$EA,$F7,$7C,$26,$3D,$6F,$6B
	.dc.b	$A8,$E1,$87,$F8,$BE,$BF,$DD,$8F,$F9,$8E,$3F,$CC,$DC,$3F,$ED,$38
	.dc.b	$FF,$17,$CF,$FB,$2B,$5F,$F2,$7C,$FF,$89,$A7,$97,$A7,$F2,$BE,$1B
	.dc.b	$C7,$E8,$FC,$CF,$84,$F2,$CF,$CE,$F8,$6F,$3F,$B1,$F4,$3E,$13,$D3
	.dc.b	$EA,$5C,$BC,$93,$B7,$CD,$1F,$5B,$E1,$7D,$FE,$AD,$9F,$57,$9D,$5F
	.dc.b	$3C,$7D,$CF,$84,$F8,$B3,$E3,$9F,$81,$AF,$9D,$F4,$F1,$7E,$17,$7F
	.dc.b	$4B,$F8,$7D,$FD,$2F,$E2,$76,$F4,$BF,$8B,$DB,$D2,$DB,$F4,$BB,$7D
	.dc.b	$2F,$E4,$7E,$17,$EB,$FA,$BF,$90,$F8,$2B,$3F,$95,$7E,$DF,$A5,$F9
	.dc.b	$4A,$CD,$FD,$53,$FD,$31,$C8,$F8,$3F,$C0,$FA,$7F,$9A,$F8,$3B,$97
	.dc.b	$E7,$DF,$C2,$FA,$FF,$9E,$AC,$E1,$CF,$A7,$F8,$67,$C1,$FE,$2B,$D7
	.dc.b	$FD,$1B,$71,$FD,$43,$F8,$EF,$5F,$F4,$FF,$0B,$9C,$39,$6A,$F8,$5F
	.dc.b	$C9,$FB,$3F,$AC,$F8,$5B,$97,$EC,$5F,$CB,$FA,$9F,$B0,$AC,$E1,$CB
	.dc.b	$6F,$2E,$DF,$8E,$E0,$B8,$FE,$51,$F5,$BF,$BB,$9C,$D7,$C4,$DB,$C9
	.dc.b	$7F,$96,$6D,$7E,$71,$F5,$3F,$C0,$F5,$CD,$7C,$CF,$85,$FD,$5F,$A9
	.dc.b	$FC,$11,$FD,$6F,$97,$67,$F5,$13,$FC,$23,$E0,$B3,$87,$8E,$7F,$59
	.dc.b	$CB,$F4,$DB,$1F,$4D,$FD,$87,$2F,$D3,$72,$7C,$AC,$FC,$7A,$7D,$48
	.dc.b	$FD,$B7,$2F,$D4,$7F,$71,$CB,$F4,$9C,$E1,$FC,$77,$E0,$BF,$81,$FA
	.dc.b	$9F,$EB,$0F,$D0,$BF,$DB,$9E,$4A,$31,$7F,$C2,$3D,$4F,$E4,$37,$C3
	.dc.b	$0F,$F4,$C6,$1F,$EC,$AF,$88,$63,$FE,$CF,$9C,$FC,$3B,$1F,$F5,$C7
	.dc.b	$EC,$7F,$EC,$87,$FD,$D9,$82,$7F,$8A,$FA,$5E,$3F,$86,$E1,$5F,$EB
	.dc.b	$AB,$E1,$C0,$C1,$F1,$1A,$AD,$E8,$B7,$AD,$ED,$AD,$57,$8E,$B5,$39
	.dc.b	$8F,$2F,$A9,$F0,$FE,$1F,$70,$AD,$ED,$F0,$3E,$1F,$31,$DF,$D5,$F8
	.dc.b	$B5,$EE,$6B,$E9,$F3,$3E,$75,$70,$D3,$FD,$73,$3F,$AD,$D7,$E5,$71
	.dc.b	$7C,$98,$C7,$16,$FD,$AF,$AC,$FC,$6D,$3E,$8F,$16,$5F,$A7,$C3,$D7
	.dc.b	$87,$D9,$87,$C9,$6B,$F1,$B3,$70,$FB,$E3,$EA,$CF,$E0,$FC,$7F,$37
	.dc.b	$9D,$F8,$5F,$82,$B9,$7E,$01,$F5,$79,$FB,$C3,$EE,$D7,$E2,$FE,$0B
	.dc.b	$E4,$F3,$BF,$26,$79,$7F,$8F,$F8,$3C,$C6,$33,$FC,$5C,$63,$C7,$E5
	.dc.b	$97,$D8,$C3,$96,$1F,$97,$F8,$8E,$3E,$37,$E6,$7E,$22,$E5,$F9,$47
	.dc.b	$AF,$81,$BC,$79,$68,$BF,$D0,$7C,$47,$D3,$F5,$6D,$7E,$5B,$DA,$AF
	.dc.b	$E4,$63,$F4,$DF,$11,$F6,$67,$F5,$1F,$0F,$F6,$FD,$BF,$D5,$7C,$3F
	.dc.b	$DD,$F5,$EE,$3F,$A0,$9D,$BF,$A2,$8F,$D8,$7C,$3F,$E0,$FD,$AF,$D9
	.dc.b	$7C,$25,$E2,$FD,$6B,$F8,$7F,$5B,$78,$7E,$AA,$BC,$6F,$EB,$3D,$7F
	.dc.b	$3B,$F7,$5A,$FE,$7F,$C1,$7A,$5C,$23,$F5,$93,$E4,$7F,$5D,$A7,$A1
	.dc.b	$FD,$86,$95,$FC,$44,$6D,$FC,$54,$79,$1F,$DB,$61,$EA,$7F,$7F,$EB
	.dc.b	$D7,$3A,$7F,$08,$D7,$F4,$53,$FC,$47,$E1,$FF,$49,$EC,$FF,$13,$F8
	.dc.b	$5B,$7F,$C3,$9F,$D4,$76,$6F,$1F,$E1,$75,$FC,$67,$E1,$FF,$17,$E8
	.dc.b	$59,$FB,$ED,$7F,$83,$77,$FE,$4F,$D0,$FE,$41,$F0,$FC,$33,$FE,$2D
	.dc.b	$26,$FE,$85,$FE,$B4,$B4,$FE,$6F,$C6,$D4,$7F,$65,$9D,$4E,$6B,$99
	.dc.b	$C8,$FE,$01,$E2,$70,$89,$2F,$88,$FD,$1E,$19,$BF,$0F,$C1,$FF,$07
	.dc.b	$F0,$B1,$97,$8B,$0F,$F6,$93,$FC,$B3,$C4,$CE,$49,$CC,$FF,$29,$AC
	.dc.b	$C7,$FB,$82,$3F,$81,$56,$3E,$1F,$F6,$FF,$0B,$FD,$07,$2F,$4D,$77
	.dc.b	$F8,$7C,$FB,$7C,$9D,$FB,$9C,$67,$F6,$D6,$AE,$FF,$0F,$FB,$EF,$5E
	.dc.b	$D7,$EC,$9A,$FD,$4A,$FE,$41,$B7,$C4,$78,$65,$E2,$8F,$67,$C3,$FF
	.dc.b	$30,$F0,$7F,$DE,$91,$EF,$F8,$7C,$7B,$1E,$E7,$58,$FC,$0B,$CE,$1E
	.dc.b	$C8,$7D,$93,$E1,$3E,$0F,$C1,$FC,$E9,$7E,$2B,$FC,$99,$F6,$FC,$26
	.dc.b	$63,$11,$E9,$9E,$17,$CE,$37,$87,$A7,$9B,$11,$FD,$5F,$E0,$9B,$FF
	.dc.b	$61,$AD,$67,$ED,$F7,$FE,$BB,$EF,$F7,$FE,$CF,$C2,$E6,$BA,$69,$F0
	.dc.b	$FE,$13,$EE,$BF,$17,$E1,$2E,$9F,$D8,$34,$EF,$5F,$82,$E0,$D6,$1F
	.dc.b	$0F,$7F,$2D,$FF,$70,$E4,$72,$B9,$FE,$07,$C1,$7E,$19,$B5,$F8,$09
	.dc.b	$F8,$3C,$19,$C7,$E0,$7C,$4F,$D7,$F8,$2F,$C8,$FC,$4F,$D9,$F8,$3B
	.dc.b	$8F,$97,$9B,$FD,$AA,$FE,$AE,$F5,$87,$C9,$F8,$8F,$C0,$7C,$1F,$E6
	.dc.b	$9B,$19,$AE,$5F,$07,$F9,$EF,$89,$CC,$7F,$98,$CF,$BF,$AF,$E7,$AB
	.dc.b	$D3,$1F,$91,$F2,$6B,$F0,$77,$FF,$53,$3C,$78,$FE,$2B,$FC,$1A,$3F
	.dc.b	$D4,$7C,$7F,$BA,$F0,$FF,$29,$F0,$9F,$AE,$F1,$2E,$5F,$81,$7F,$31
	.dc.b	$F0,$9F,$B2,$AC,$E3,$FC,$27,$E2,$7F,$39,$F0,$9F,$BA,$FD,$0B,$FD
	.dc.b	$9D,$3F,$87,$E5,$7C,$26,$7C,$9D,$E9,$F8,$DC,$7F,$49,$3F,$4A,$FF
	.dc.b	$B4,$7E,$58,$F6,$FD,$B1,$FC,$7A,$31,$7F,$A3,$F2,$3F,$37,$3F,$81
	.dc.b	$97,$C1,$73,$8F,$E5,$31,$C1,$FF,$42,$69,$FE,$2E,$7E,$D7,$C2,$7A
	.dc.b	$7C,$1F,$DB,$7C,$25,$9C,$E5,$FB,$BB,$E1,$1E,$D9,$F8,$1E,$C7,$EF
	.dc.b	$74,$F8,$5E,$C5,$C3,$83,$FC,$07,$4D,$E1,$F8,$9A,$F3,$3E,$9E,$6F
	.dc.b	$7B,$FC,$E3,$54,$63,$9D,$0E,$6E,$35,$F1,$6B,$F7,$D9,$FF,$0F,$F7
	.dc.b	$6C,$FE,$5D,$AF,$D5,$F7,$78,$BF,$E4,$7E,$64,$FD,$BF,$63,$F8,$C7
	.dc.b	$B7,$F7,$7E,$22,$F0,$FC,$37,$F8,$E7,$B9,$BC,$7F,$49,$5D,$37,$F3
	.dc.b	$F9,$9E,$9E,$07,$9D,$63,$51,$8C,$BF,$79,$E1,$FF,$95,$BF,$64,$FC
	.dc.b	$DE,$1F,$D6,$32,$FE,$3B,$C3,$9F,$13,$D7,$1F,$E6,$AA,$DE,$1F,$E7
	.dc.b	$2F,$13,$F4,$33,$FE,$63,$9F,$D0,$CE,$71,$ED,$EE,$7F,$22,$F5,$FC
	.dc.b	$1F,$1B,$70,$FE,$25,$82,$BE,$3F,$A4,$85,$7B,$FF,$2E,$FF,$1B,$65
	.dc.b	$F9,$5F,$03,$FC,$97,$0F,$77,$ED,$3A,$CB,$F6,$DA,$6A,$7A,$E2,$F7
	.dc.b	$7D,$B5,$FE,$81,$87,$BB,$88,$F6,$4F,$08,$E3,$EA,$FA,$67,$F9,$C5
	.dc.b	$7F,$24,$CF,$F9,$46,$FF,$36,$BB,$F9,$BE,$9F,$86,$F2,$F7,$79,$9F
	.dc.b	$63,$E4,$F7,$7C,$1C,$5F,$05,$B9,$31,$D6,$5F,$6B,$3F,$93,$BF,$8D
	.dc.b	$FA,$DE,$D5,$7D,$91,$F0,$23,$F1,$1F,$0D,$FD,$36,$7F,$91,$79,$7F
	.dc.b	$BE,$7F,$C3,$F0,$F7,$FF,$48,$EF,$AF,$F0,$DD,$3F,$80,$51,$BC,$BC
	.dc.b	$97,$F1,$3D,$CF,$49,$EA,$F7,$3E,$8B,$F4,$BD,$9F,$34,$FE,$25,$BC
	.dc.b	$3F,$81,$F6,$F5,$87,$4A,$FC,$97,$B3,$FB,$AA,$FC,$33,$78,$9E,$C7
	.dc.b	$DE,$8F,$BF,$EB,$FC,$AF,$63,$E5,$BF,$82,$7F,$07,$EB,$78,$3F,$65
	.dc.b	$FF,$69,$65,$FE,$02,$E7,$FD,$34,$FE,$52,$73,$C7,$FA,$0D,$11,$FE
	.dc.b	$CC,$D7,$F2,$98,$7F,$6D,$8C,$3F,$3B,$C1,$FF,$62,$C7,$FA,$BA,$7D
	.dc.b	$93,$FE,$D1,$8D,$F2,$7C,$EE,$8D,$F8,$BF,$84,$97,$A7,$63,$C8,$F9
	.dc.b	$7C,$DF,$51,$EB,$F3,$71,$97,$6C,$BF,$DB,$18,$3E,$2F,$E5,$2F,$8B
	.dc.b	$B1,$EF,$7E,$07,$5F,$84,$F1,$7C,$4E,$30,$E9,$C1,$F0,$BE,$27,$E5
	.dc.b	$7C,$5F,$CB,$7E,$43,$F2,$7C,$3E,$FE,$27,$81,$F9,$8F,$CD,$F8,$AF
	.dc.b	$71,$EF,$F8,$9C,$63,$DB,$6F,$F6,$26,$FF,$DC,$7A,$7E,$5F,$C5,$7D
	.dc.b	$9F,$8A,$FA,$CF,$D7,$F8,$AF,$6F,$9D,$EE,$7E,$D3,$FC,$3B,$F6,$66
	.dc.b	$1F,$B0,$7F,$34,$FE,$93,$A7,$FD,$F7,$1A,$CF,$DE,$EB,$AF,$EB,$57
	.dc.b	$F8,$AE,$3F,$4D,$3F,$C9,$73,$E3,$F1,$3F,$9C,$9C,$D7,$83,$BB,$F1
	.dc.b	$50,$9F,$27,$AD,$F9,$D8,$37,$D7,$CE,$3E,$2F,$89,$F4,$1E,$9F,$89
	.dc.b	$F6,$3E,$DF,$89,$F0,$1E,$1F,$89,$C5,$7A,$BE,$03,$B7,$C0,$F7,$7E
	.dc.b	$23,$F1,$7E,$27,$E2,$7C,$47,$C5,$7E,$43,$F2,$7E,$23,$FC,$89,$7E
	.dc.b	$FE,$1F,$85,$18,$7D,$DC,$5F,$2A,$31,$1F,$E3,$03,$E4,$7C,$3E,$38
	.dc.b	$FE,$1F,$C3,$6B,$D7,$FE,$29,$4F,$0D,$CB,$FE,$12,$8F,$D7,$9F,$E6
	.dc.b	$97,$0F,$F4,$03,$EA,$57,$5F,$2F,$ED,$FC,$2F,$5E,$AF,$BB,$F0,$D6
	.dc.b	$FE,$D3,$B7,$EE,$4D,$7B,$7C,$2F,$9F,$7F,$C2,$B6,$3F,$04,$ED,$FC
	.dc.b	$69,$E9,$C9,$F0,$D9,$8F,$C8,$E8,$F6,$BF,$B9,$77,$7F,$73,$DF,$E0
	.dc.b	$CE,$7E,$17,$E2,$7A,$5D,$30,$EB,$39,$7F,$67,$87,$2E,$5F,$EE,$51
	.dc.b	$FE,$7F,$8F,$ED,$CF,$F8,$F7,$83,$93,$E2,$8F,$F6,$2E,$5E,$19,$CC
	.dc.b	$FD,$19,$CD,$74,$8F,$F4,$C4,$74,$7E,$81,$8F,$82,$F8,$8E,$BE,$0A
	.dc.b	$BE,$89,$FE,$DE,$FA,$27,$E2,$3D,$FE,$0B,$DE,$FA,$5A,$F8,$1D,$E5
	.dc.b	$F2,$7E,$0A,$DB,$E0,$BE,$E5,$F9,$9A,$F8,$5D,$F1,$F0,$E6,$F2,$CF
	.dc.b	$C3,$6B,$F0,$5E,$BF,$05,$DB,$E0,$7C,$73,$EB,$6B,$F2,$DF,$1E,$1E
	.dc.b	$4F,$81,$F0,$CF,$CF,$F8,$1A,$FC,$D7,$4F,$DA,$F7,$FC,$3A,$FE,$17
	.dc.b	$E0,$2E,$1F,$D7,$AB,$47,$E0,$3E,$03,$C2,$FD,$76,$D7,$E3,$5F,$8B
	.dc.b	$F0,$1F,$1B,$E0,$3C,$32,$FD,$9F,$FD,$1B,$D1,$F8,$1C,$3F,$D8,$D1
	.dc.b	$8E,$7F,$C0,$47,$D9,$BF,$A3,$3F,$E7,$8A,$FF,$02,$F1,$7E,$23,$0E
	.dc.b	$9F,$1F,$EA,$AD,$63,$CB,$1F,$D0,$C7,$FB,$A3,$0F,$E5,$71,$FE,$E3
	.dc.b	$87,$4F,$A9,$FE,$C3,$1F,$C0,$0F,$D0,$B8,$7F,$22,$62,$3C,$BF,$03
	.dc.b	$EE,$93,$4F,$4F,$C0,$FA,$BE,$07,$CD,$F0,$33,$0F,$EB,$B7,$FD,$9D
	.dc.b	$BF,$03,$F3,$BF,$2F,$72,$BD,$0C,$36,$B6,$72,$F8,$2F,$11,$1E,$BF
	.dc.b	$05,$F2,$BE,$0B,$E3,$12,$3C,$3F,$05,$FA,$4E,$1F,$F2,$17,$85,$F7
	.dc.b	$9B,$3F,$0C,$FE,$B6,$E1,$F8,$66,$39,$BF,$5E,$E0,$FE,$59,$B3,$E0
	.dc.b	$BE,$37,$C1,$49,$F8,$31,$F7,$BE,$0B,$ED,$12,$3E,$2F,$C0,$79,$7F
	.dc.b	$38,$7A,$3E,$56,$CF,$94,$66,$8C,$C7,$8E,$31,$E8,$FE,$33,$E0,$BE
	.dc.b	$C7,$C1,$4A,$FA,$71,$F9,$0F,$82,$FC,$61,$23,$FC,$27,$FB,$47,$F2
	.dc.b	$5C,$3F,$17,$1F,$CC,$9F,$07,$B7,$E5,$F6,$E4,$FF,$89,$67,$2F,$2F
	.dc.b	$87,$FD,$2D,$FF,$98,$E3,$F9,$41,$FC,$A3,$C4,$46,$63,$E6,$46,$3E
	.dc.b	$3F,$FB,$15,$BE,$3B,$F2,$35,$C6,$FF,$C0,$0F,$FA,$FC,$C7,$C3,$70
	.dc.b	$97,$87,$FA,$BA,$DD,$7F,$BB,$BC,$4C,$7C,$37,$D1,$F8,$4F,$A1,$F0
	.dc.b	$B2,$BD,$9C,$A9,$FE,$4F,$87,$FB,$1A,$7D,$67,$F4,$BD,$BD,$7F,$0D
	.dc.b	$F0,$0D,$D7,$8F,$C9,$F1,$63,$E5,$F8,$8B,$87,$FB,$C8,$F6,$46,$DF
	.dc.b	$77,$C3,$FE,$15,$FC,$21,$0F,$2F,$1F,$CE,$F1,$E4,$79,$0F,$EB,$BA
	.dc.b	$7E,$AD,$F9,$AF,$FB,$E1,$FC,$FF,$99,$FA,$0F,$90,$79,$FF,$1F,$E1
	.dc.b	$FE,$57,$C8,$7C,$9E,$BB,$87,$D1,$7E,$AD,$F0,$8F,$A9,$5F,$AF,$D5
	.dc.b	$E2,$7E,$BD,$B3,$F9,$BE,$A4,$62,$3F,$34,$BC,$3F,$B5,$E8,$FD,$9E
	.dc.b	$EF,$B9,$EF,$DB,$F4,$9F,$A4,$ED,$FE,$FF,$C2,$AE,$33,$F8,$7F,$88
	.dc.b	$F3,$F9,$FF,$83,$F8,$8B,$3F,$8B,$F8,$7A,$FD,$BC,$11,$EE,$F6,$78
	.dc.b	$4F,$E7,$F6,$FF,$5C,$BE,$B5,$E9,$7B,$CF,$85,$EA,$7C,$6F,$63,$F2
	.dc.b	$5F,$21,$F2,$3D,$9F,$CA,$79,$B7,$1F,$CE,$3B,$7F,$3D,$3F,$3B,$DD
	.dc.b	$FC,$D7,$9F,$67,$F4,$5F,$11,$5F,$43,$F5,$3D,$CD,$C7,$9F,$C5,$FA
	.dc.b	$44,$7D,$5E,$2F,$DA,$F7,$BF,$4D,$F0,$FF,$73,$DF,$FE,$A5,$FB,$4F
	.dc.b	$F8,$8E,$FF,$88,$F8,$49,$CF,$8D,$C6,$7F,$69,$7F,$84,$F8,$3E,$3F
	.dc.b	$0F,$F8,$6F,$6E,$D7,$D8,$F1,$19,$7F,$BD,$63,$F8,$26,$6B,$E5,$18
	.dc.b	$F9,$0F,$E7,$5F,$9E,$BA,$3F,$9D,$2E,$8A,$6A,$3F,$2B,$F0,$DE,$AF
	.dc.b	$91,$B8,$76,$F8,$9A,$FE,$E7,$34,$7E,$73,$E2,$38,$4F,$9F,$5F,$F1
	.dc.b	$B3,$FE,$48,$AF,$E8,$7B,$38,$BF,$55,$8F,$C8,$F8,$2F,$D2,$B6,$F9
	.dc.b	$6F,$F3,$3E,$2B,$F5,$3A,$7F,$91,$B1,$70,$3B,$3F,$5B,$59,$8F,$F2
	.dc.b	$F6,$1E,$38,$C7,$57,$28,$FF,$29,$DF,$2B,$FD,$AE,$7F,$6B,$E2,$AD
	.dc.b	$7C,$E6,$BF,$05,$FD,$D9,$DF,$A3,$F9,$8F,$CB,$3C,$0F,$C1,$F4,$E3
	.dc.b	$BE,$E1,$F6,$EB,$FA,$9B,$F1,$67,$8D,$7E,$2B,$AB,$9F,$7F,$AF,$3B
	.dc.b	$7E,$8E,$04,$62,$3F,$A9,$46,$1E,$DD,$9F,$0F,$5F,$E2,$3C,$BF,$17
	.dc.b	$3B,$5F,$8A,$E9,$AF,$DD,$7F,$8B,$F6,$6E,$3F,$25,$E5,$FF,$85,$5F
	.dc.b	$C1,$75,$FF,$49,$F8,$BF,$D0,$B6,$3F,$2A,$FF,$4D,$F8,$BF,$D2,$F2
	.dc.b	$FF,$4F,$F1,$3F,$4F,$D1,$FD,$47,$C5,$B8,$76,$6B,$FD,$EB,$6F,$F0
	.dc.b	$7E,$7F,$C2,$A3,$F2,$5E,$17,$F5,$9F,$8A,$FD,$8F,$55,$9F,$D2,$F9
	.dc.b	$95,$FC,$C3,$FD,$7F,$E2,$B7,$1E,$CE,$0F,$DC,$F9,$5F,$D8,$FE,$2A
	.dc.b	$CF,$C0,$6B,$E3,$3F,$95,$4F,$F0,$0F,$1B,$D9,$1F,$B4,$DF,$D8,$D3
	.dc.b	$D8,$53,$97,$81,$FD,$DF,$87,$F8,$4E,$DE,$79,$FD,$46,$B5,$FF,$5C
	.dc.b	$F8,$BF,$C2,$DF,$A7,$BF,$A9,$E7,$C0,$DF,$F8,$9F,$8F,$F4,$E7,$F7
	.dc.b	$1E,$C7,$B5,$FE,$12,$D7,$27,$C3,$3F,$C2,$1E,$87,$F9,$CD,$FE,$91
	.dc.b	$A7,$D4,$D7,$F7,$51,$FD,$2F,$97,$1C,$BF,$59,$B1,$C0,$FE,$2F,$BF
	.dc.b	$AD,$FF,$77,$6D,$EB,$7D,$3B,$7A,$F0,$EA,$7F,$86,$DC,$61,$F8,$2D
	.dc.b	$7D,$B5,$FA,$BF,$37,$FD,$05,$A7,$FA,$02,$3F,$C8,$3E,$27,$F6,$87
	.dc.b	$D7,$E9,$7E,$DA,$B6,$FE,$E3,$DC,$E3,$EA,$7E,$EB,$DC,$FE,$1D,$F9
	.dc.b	$F3,$DA,$FD,$EF,$B9,$86,$E5,$FC,$2A,$74,$FE,$F6,$B8,$FB,$1F,$C1
	.dc.b	$3E,$43,$9E,$DF,$C1,$BD,$CB,$7F,$C0,$DA,$FE,$13,$7F,$F0,$83,$FE
	.dc.b	$4D,$C5,$ED,$7F,$8D,$FE,$2F,$DF,$EC,$5A,$FA,$FE,$C5,$7F,$C5,$D9
	.dc.b	$7F,$A5,$F9,$FE,$5F,$77,$F8,$26,$7F,$DF,$73,$FE,$A0,$7F,$7B,$EE
	.dc.b	$7C,$AF,$84,$FF,$22,$36,$32,$7E,$12,$3F,$C9,$1F,$17,$F4,$3E,$1B
	.dc.b	$FC,$95,$F1,$5F,$4B,$C7,$FF,$26,$7C,$5D,$C3,$F8,$4B,$5F,$CB,$CF
	.dc.b	$D8,$F3,$7F,$CA,$5F,$17,$F6,$BC,$5B,$5F,$92,$6B,$EC,$7D,$CF,$AB
	.dc.b	$C5,$F3,$BE,$53,$CE,$7C,$2D,$32,$D3,$F8,$DE,$9E,$97,$F8,$EE,$3E
	.dc.b	$5E,$0F,$8F,$9D,$F0,$BF,$88,$B6,$BF,$C5,$55,$E9,$7F,$BB,$6C,$7F
	.dc.b	$27,$E1,$FF,$9D,$23,$11,$FA,$C8,$C3,$A3,$8F,$C2,$FE,$6B,$C4,$E1
	.dc.b	$F1,$7F,$9C,$F3,$79,$7C,$5D,$BE,$ED,$7F,$85,$E2,$AF,$D1,$F5,$78
	.dc.b	$7E,$2F,$F4,$BF,$0F,$E3,$F8,$9F,$D3,$F4,$DC,$7F,$7E,$ED,$FE,$03
	.dc.b	$3E,$9F,$23,$F5,$BC,$9E,$BF,$2B,$F5,$FC,$56,$FF,$83,$BB,$7F,$84
	.dc.b	$CF,$A9,$FE,$17,$BF,$A9,$A7,$A8,$A7,$AB,$5F,$4B,$F3,$34,$F4,$BF
	.dc.b	$3B,$4F,$4B,$4F,$49,$B3,$D2,$7D,$4F,$1B,$F8,$2F,$27,$D3,$F2,$3F
	.dc.b	$83,$F5,$DA,$FE,$3E,$FD,$7A,$DC,$7D,$CF,$1B,$EC,$BF,$6B,$E2,$9F
	.dc.b	$B5,$7F,$B0,$7A,$BF,$DC,$7C,$DF,$B8,$AD,$4E,$5F,$F6,$9C,$65,$FC
	.dc.b	$35,$72,$7F,$0D,$E4,$76,$73,$5F,$EB,$FE,$CF,$E4,$77,$FC,$3A,$7F
	.dc.b	$D5,$53,$F7,$DF,$C8,$F9,$5C,$56,$1E,$6E,$1F,$C9,$4E,$63,$13,$F9
	.dc.b	$8F,$2B,$17,$FE,$C0,$C7,$BF,$89,$8E,$3F,$CD,$36,$FF,$3E,$D7,$F2
	.dc.b	$3E,$47,$2E,$4F,$D0,$70,$74,$9A,$7F,$B6,$F1,$F8,$98,$FF,$B4,$27
	.dc.b	$53,$FA,$AD,$FC,$7D,$FE,$88,$E1,$BD,$7E,$0B,$5F,$9E,$FE,$BF,$5F
	.dc.b	$5F,$77,$EC,$7E,$27,$DB,$D9,$6F,$C4,$D7,$E5,$BF,$B7,$F8,$9F,$87
	.dc.b	$DB,$E8,$9F,$9E,$53,$EE,$94,$FA,$E7,$DA,$3D,$D9,$79,$5F,$B7,$95
	.dc.b	$7E,$AE,$DF,$65,$FD,$CF,$27,$D2,$9F,$54,$7C,$7C,$FF,$85,$7C,$1F
	.dc.b	$D6,$E8,$B3,$F2,$FA,$AB,$F8,$47,$F8,$87,$C2,$6E,$3F,$8A,$FC,$1F
	.dc.b	$DE,$E3,$FE,$28,$D8,$C9,$F5,$63,$F8,$C7,$C0,$FE,$13,$93,$F8,$CF
	.dc.b	$2F,$F7,$4F,$D1,$BF,$C0,$37,$F9,$2D,$F3,$DB,$FC,$86,$95,$FE,$95
	.dc.b	$AF,$C0,$73,$E9,$4E,$BD,$FF,$C8,$CE,$63,$8D,$FE,$53,$0C,$61,$FE
	.dc.b	$8C,$C3,$F4,$BD,$78,$F8,$8F,$CC,$F6,$59,$FC,$73,$C3,$E2,$7F,$3B
	.dc.b	$5F,$02,$7D,$C7,$C0,$3F,$57,$BE,$5A,$7F,$3C,$DF,$E1,$3F,$EB,$3C
	.dc.b	$7E,$0B,$FB,$EC,$BC,$BB,$7A,$A3,$4D,$7F,$C7,$F3,$C1,$FD,$EC,$FB
	.dc.b	$3E,$27,$F6,$3D,$BE,$EF,$89,$FD,$9F,$6D,$BF,$D4,$3B,$7F,$55,$3F
	.dc.b	$13,$E2,$3F,$73,$D3,$F1,$BE,$1F,$F7,$7D,$16,$FF,$60,$ED,$FD,$94
	.dc.b	$FC,$CF,$87,$FE,$03,$D3,$67,$F6,$CD,$7F,$87,$BF,$4B,$E2,$7F,$83
	.dc.b	$F0,$FD,$4F,$88,$FE,$19,$FB,$5F,$FB,$87,$AB,$F0,$31,$8E,$AD,$63
	.dc.b	$FC,$77,$C7,$5E,$F9,$5F,$13,$CB,$9F,$CD,$7F,$C7,$B7,$F5,$3F,$C8
	.dc.b	$38,$2B,$FC,$B9,$DB,$FC,$C2,$3D,$0F,$F3,$2D,$EB,$FC,$B1,$AF,$EC
	.dc.b	$9F,$63,$FC,$57,$6F,$C7,$FC,$4F,$B3,$B7,$F2,$3F,$0D,$6B,$F0,$4D
	.dc.b	$7E,$EC,$FC,$2E,$DF,$CB,$FC,$1F,$C4,$EB,$B5,$F0,$78,$3F,$DC,$7C
	.dc.b	$5F,$A7,$E1,$5F,$CA,$EB,$FC,$FF,$C3,$71,$C3,$F2,$1F,$13,$F3,$BA
	.dc.b	$FF,$46,$FD,$0E,$BB,$3D,$5F,$D3,$FC,$4E,$E3,$F3,$1F,$13,$F5,$BB
	.dc.b	$7F,$57,$F1,$36,$7F,$38,$D7,$F2,$F3,$F7,$3B,$BD,$31,$FD,$8B,$B3
	.dc.b	$EC,$F1,$62,$3F,$B2,$71,$7B,$1F,$DE,$F0,$FB,$1F,$81,$C1,$EC,$7E
	.dc.b	$15,$7A,$1E,$95,$FE,$D9,$7F,$4F,$B7,$B5,$F8,$27,$B4,$FD,$96,$BE
	.dc.b	$D7,$F5,$BF,$13,$F9,$2E,$EF,$E1,$5F,$A3,$BE,$3B,$15,$FC,$57,$9B
	.dc.b	$F8,$B4,$E6,$3E,$A5,$7F,$0E,$BF,$CD,$F8,$1E,$D9,$FE,$EF,$9B,$4C
	.dc.b	$47,$F6,$E8,$F8,$1E,$0F,$F2,$0E,$8F,$B5,$3F,$C1,$FA,$3E,$D3,$EC
	.dc.b	$3F,$05,$D1,$F0,$5F,$93,$D1,$5E,$EF,$DA,$76,$7A,$4F,$E5,$3A,$BE
	.dc.b	$27,$F6,$3D,$7F,$01,$B3,$BC,$95,$B8,$FC,$95,$7C,$6F,$8A,$FD,$CF
	.dc.b	$6F,$C8,$F8,$8F,$DD,$EB,$F2,$BE,$23,$F7,$BB,$5C,$7F,$34,$ED,$FC
	.dc.b	$E4,$FD,$0F,$83,$FE,$07,$D1,$F4,$BD,$9B,$5E,$3D,$51,$88,$FF,$71
	.dc.b	$C2,$BF,$85,$F4,$FD,$AF,$87,$FE,$1B,$DF,$F7,$3E,$1F,$F8,$7F,$83
	.dc.b	$71,$FD,$63,$B7,$F5,$D3,$E5,$7F,$1B,$A7,$A5,$FC,$7E,$B5,$FC,$CB
	.dc.b	$B7,$F3,$71,$EB,$7F,$71,$8F,$AD,$FD,$D6,$B5,$FC,$CB,$B7,$F3,$71
	.dc.b	$F9,$2F,$87,$FE,$4B,$D7,$F9,$4F,$87,$FF,$19,$7E,$D9,$FE,$ED,$E7
	.dc.b	$FA,$B1,$8F,$03,$86,$3F,$C1,$EF,$F3,$7F,$11,$C3,$AA,$CF,$F2,$BC
	.dc.b	$5D,$38,$8F,$DB,$46,$1F,$A1,$F1,$1C,$5F,$A5,$F0,$7E,$2E,$AF,$D3
	.dc.b	$FC,$45,$9F,$AC,$D7,$EA,$CF,$A3,$AF,$F5,$BF,$0F,$EA,$EA,$FD,$7F
	.dc.b	$C4,$5B,$FB,$CD,$7E,$EC,$FC,$0E,$DF,$34,$7F,$9A,$F4,$FF,$7C,$71
	.dc.b	$FE,$5B,$83,$E2,$EF,$AA,$FF,$33,$D7,$AD,$FC,$3E,$FE,$B7,$CA,$7C
	.dc.b	$3D,$FD,$6D,$3D,$65,$36,$7C,$6D,$7F,$83,$FC,$3F,$D4,$ED,$B3,$F3
	.dc.b	$5A,$FC,$C7,$F8,$6F,$C4,$7D,$AE,$EF,$E1,$FF,$11,$F7,$3A,$7F,$88
	.dc.b	$FC,$45,$C3,$F4,$0D,$7F,$3F,$3F,$82,$EA,$FF,$10,$7E,$8F,$FE,$79
	.dc.b	$9F,$F3,$CE,$EE,$1C,$C7,$E2,$38,$2B,$F8,$17,$8F,$C2,$FE,$32,$79
	.dc.b	$FC,$27,$E3,$BA,$7A,$FC,$27,$E4,$3A,$EE,$3F,$89,$76,$FE,$2E,$7D
	.dc.b	$6F,$E3,$77,$F5,$B4,$F5,$94,$F5,$9F,$93,$D7,$CA,$FE,$4B,$1A,$FD
	.dc.b	$FC,$EB,$F7,$1F,$4B,$F9,$B8,$F8,$7F,$13,$FA,$6E,$AB,$3F,$9F,$6B
	.dc.b	$F9,$E7,$E4,$FC,$4F,$EB,$3B,$3D,$33,$FA,$2C,$BD,$2D,$3D,$25,$3D
	.dc.b	$27,$EA,$32,$F2,$BF,$4B,$1B,$E9,$B8,$5A,$61,$B5,$FC,$BF,$89,$C5
	.dc.b	$F4,$E3,$1F,$13,$FC,$03,$A7,$EF,$FC,$4D,$9E,$0D,$7F,$CC,$BE,$9F
	.dc.b	$F9,$AF,$05,$FF,$08,$EA,$FC,$3F,$C4,$7F,$0A,$E9,$B8,$7F,$03,$6B
	.dc.b	$FC,$09,$FC,$6F,$C4,$7F,$10,$EB,$FC,$7F,$C2,$7F,$B0,$FA,$BF,$D6
	.dc.b	$79,$FF,$B5,$70,$FF,$64,$C7,$FB,$47,$4F,$15,$6F,$7F,$CB,$FC,$47
	.dc.b	$1B,$FC,$CF,$C2,$F3,$EE,$FC,$DF,$C4,$75,$EF,$B5,$FE,$41,$CF,$FD
	.dc.b	$9D,$9F,$B7,$0E,$13,$F2,$78,$3D,$8D,$3D,$85,$3D,$87,$CF,$DF,$D8
	.dc.b	$FD,$1D,$7C,$8D,$3C,$1A,$FC,$06,$9F,$5F,$5D,$BF,$40,$FF,$4A,$EC
	.dc.b	$F8,$7F,$8D,$DF,$FB,$67,$E4,$77,$5A,$F6,$3F,$BA,$F8,$8D,$C7,$E0
	.dc.b	$FE,$1F,$E7,$73,$FE,$FB,$E2,$3E,$87,$4D,$AF,$88,$FF,$02,$AD,$C7
	.dc.b	$E2,$FE,$1B,$EB,$70,$FF,$06,$7E,$C7,$0D,$9F,$98,$FF,$0A,$F8,$7D
	.dc.b	$C7,$E4,$FE,$1F,$EF,$71,$FF,$0E,$F8,$7B,$3F,$97,$6B,$F0,$7C,$D7
	.dc.b	$1A,$FF,$0D,$C9,$E5,$AF,$B5,$C1,$E9,$7F,$C8,$9D,$0E,$0C,$40,$9E
	.dc.b	$12,$75,$1F,$14,$E6,$7B,$CF,$C3,$70,$7B,$5E,$BB,$D7,$A1,$F9,$67
	.dc.b	$E0,$3C,$0F,$F1,$07,$BB,$FE,$AF,$7F,$37,$D3,$FE,$CD,$8C,$61,$E3
	.dc.b	$EC,$C7,$5F,$E8,$DD,$7C,$47,$E9,$6F,$87,$C4,$7E,$9F,$AF,$97,$C4
	.dc.b	$7E,$A7,$AE,$E9,$F9,$D7,$6F,$E7,$E7,$C1,$F0,$5F,$AF,$EB,$F1,$7C
	.dc.b	$2D,$AF,$3B,$5F,$A5,$BF,$F9,$3D,$FD,$47,$6F,$A6,$3F,$71,$DB,$67
	.dc.b	$F5,$9D,$B5,$FA,$6F,$BF,$E0,$B7,$1F,$13,$6F,$53,$F1,$B6,$F5,$34
	.dc.b	$F5,$1B,$3D,$47,$CC,$D3,$CD,$D8,$E3,$FF,$80,$19,$B5,$CC,$61,$B7
	.dc.b	$FF,$00,$27,$D8,$FE,$23,$5F,$A9,$F1,$1F,$C4,$FB,$3E,$B7,$C4,$5A
	.dc.b	$FB,$6D,$7E,$2C,$FF,$19,$EF,$F9,$F1,$F9,$2E,$3F,$A6,$FE,$53,$92
	.dc.b	$BF,$88,$E8,$FC,$94,$70,$8C,$75,$FE,$FD,$F7,$1F,$C5,$F9,$BE,$BB
	.dc.b	$4F,$AE,$53,$EB,$F0,$FF,$2B,$D7,$F1,$DF,$0F,$F9,$69,$F0,$7E,$DB
	.dc.b	$FE,$E5,$E0,$4E,$67,$F4,$B3,$98,$C4,$33,$4F,$1F,$88,$CC,$63,$AF
	.dc.b	$F3,$9F,$0F,$8A,$74,$63,$0F,$F7,$84,$3A,$31,$1F,$EB,$B8,$C3,$FA
	.dc.b	$8F,$87,$E5,$5F,$A7,$F8,$3E,$9B,$DA,$E6,$D7,$93,$FA,$DF,$84,$F1
	.dc.b	$74,$FE,$BF,$E1,$3C,$9C,$3F,$B1,$F8,$4F,$37,$75,$C7,$CA,$ED,$F3
	.dc.b	$CF,$99,$F9,$39,$D7,$E3,$ED,$5F,$8D,$B7,$A9,$FD,$4E,$DE,$A7,$F8
	.dc.b	$1E,$DE,$A6,$9E,$A2,$9E,$A3,$F8,$4E,$BE,$A7,$F8,$5F,$C2,$FC,$DE
	.dc.b	$BF,$E1,$5F,$0D,$F3,$FB,$6D,$7C,$E7,$F8,$75,$6E,$3F,$89,$FC,$2F
	.dc.b	$D5,$E8,$FE,$33,$E0,$FE,$A6,$4B,$F4,$7F,$F6,$A4,$7D,$5E,$AD,$F8
	.dc.b	$5F,$CA,$EB,$5F,$0F,$F7,$E7,$F8,$E6,$3D,$6F,$FD,$81,$87,$FA,$FD
	.dc.b	$FB,$DD,$5C,$9E,$F1,$FE,$6B,$7F,$2B,$C3,$9C,$7C,$5F,$09,$F8,$B8
	.dc.b	$C5,$73,$AF,$CD,$DE,$3E,$13,$F2,$1C,$7A,$F8,$4F,$C9,$72,$5B,$FC
	.dc.b	$8B,$B7,$F2,$73,$E8,$7F,$2B,$C1,$DB,$7F,$64,$7A,$9A,$FF,$24,$DB
	.dc.b	$DC,$FF,$57,$CF,$C8,$FE,$AB,$3F,$3E,$BE,$98,$F8,$6D,$78,$7C,$3F
	.dc.b	$E9,$E7,$DB,$F0,$FF,$A9,$ED,$F7,$FC,$3D,$BF,$96,$D7,$C1,$3F,$AF
	.dc.b	$EE,$F8,$BF,$0F,$FB,$1E,$7F,$8F,$F0,$FF,$B3,$EA,$B8,$7E,$C9,$DB
	.dc.b	$FB,$49,$F9,$BF,$09,$FB,$9E,$0B,$3F,$B8,$6B,$FB,$77,$E9,$FC,$27
	.dc.b	$F0,$2F,$DB,$BF,$D9,$5C,$5F,$E8,$7A,$FE,$19,$E3,$A7,$FD,$5B,$39
	.dc.b	$9F,$E2,$93,$9E,$2F,$BB,$F0,$BC,$DF,$BF,$F0,$BD,$79,$7C,$F7,$FC
	.dc.b	$1F,$5F,$5B,$FC,$27,$6A,$FF,$1E,$76,$FF,$20,$8F,$4B,$FC,$8B,$5A
	.dc.b	$FF,$20,$6B,$FB,$A7,$DA,$FA,$B4,$F6,$FC,$47,$BF,$B3,$F2,$7E,$6F
	.dc.b	$C1,$ED,$B7,$B7,$F2,$F5,$B8,$FD,$1F,$27,$D5,$7F,$4B,$C9,$F5,$5A
	.dc.b	$7D,$53,$67,$D5,$3F,$57,$C9,$F5,$5A,$7D,$52,$9F,$54,$FD,$8F,$27
	.dc.b	$D5,$EA,$FD,$4F,$C4,$7D,$7E,$8F,$D5,$FC,$45,$BF,$DB,$B5,$FC,$E4
	.dc.b	$FD,$DE,$AF,$D8,$FC,$3F,$DF,$D3,$F6,$7F,$0F,$6F,$F7,$ED,$7F,$49
	.dc.b	$3F,$85,$D7,$F7,$3F,$0F,$F8,$7D,$AD,$7C,$D6,$BF,$84,$E3,$47,$F0
	.dc.b	$0F,$87,$FC,$77,$07,$F0,$5F,$8E,$FF,$6B,$76,$FE,$37,$B7,$18,$F9
	.dc.b	$CF,$E2,$DE,$47,$F1,$7A,$DE,$4E,$0E,$5F,$0D,$F9,$9C,$3D,$B8,$7D
	.dc.b	$FE,$1F,$6B,$4F,$69,$4F,$69,$F8,$5D,$FD,$2F,$F7,$DC,$3D,$2F,$F2
	.dc.b	$2D,$FE,$03,$4F,$80,$7F,$26,$DB,$DA,$FA,$0F,$E6,$79,$BE,$23,$F5
	.dc.b	$DD,$FF,$05,$B3,$BF,$25,$5B,$8F,$CC,$57,$C7,$F8,$8F,$DB,$75,$7C
	.dc.b	$9F,$88,$FD,$C7,$6F,$CB,$F8,$8F,$DD,$75,$5C,$7F,$40,$ED,$FD,$14
	.dc.b	$FD,$1F,$85,$FE,$01,$D9,$67,$F4,$CD,$7E,$D3,$F5,$FE,$17,$F8,$37
	.dc.b	$67,$D9,$F8,$5F,$1D,$7A,$77,$7C,$2E,$23,$FD,$05,$18,$7F,$65,$B7
	.dc.b	$17,$F6,$9D,$FE,$57,$F1,$1A,$7A,$5F,$C5,$6B,$5F,$A8,$ED,$FA,$D1
	.dc.b	$E7,$7E,$C6,$95,$FF,$43,$F3,$23,$11,$FE,$23,$8C,$3E,$9D,$7D,$AF
	.dc.b	$AF,$E2,$3F,$C5,$3C,$8F,$8F,$FF,$47,$CF,$B3,$E2,$33,$18,$9F,$EE
	.dc.b	$47,$DA,$8C,$DF,$E7,$EF,$7C,$5F,$A1,$F8,$7E,$3C,$57,$2F,$F1,$D6
	.dc.b	$3D,$F8,$B1,$1F,$E4,$18,$C3,$F3,$FE,$23,$8B,$F4,$7E,$13,$C7,$97
	.dc.b	$EB,$5F,$2E,$96,$7A,$EA,$8F,$D9,$7C,$27,$09,$FB,$5F,$09,$FD,$0F
	.dc.b	$BB,$F6,$CF,$F4,$5E,$EB,$3F,$CA,$DF,$DD,$7C,$26,$E3,$F0,$5F,$11
	.dc.b	$FE,$4C,$FD,$BF,$D1,$E3,$FE,$E6,$31,$E3,$7F,$01,$6D,$7E,$27,$E2
	.dc.b	$78,$3F,$C1,$7E,$27,$97,$8B,$63,$FA,$9E,$3F,$F0,$01,$FD,$6C,$F7
	.dc.b	$F1,$3F,$86,$7C,$4F,$09,$FD,$67,$C9,$7F,$AD,$BB,$5C,$3F,$E1,$58
	.dc.b	$C3,$76,$F5,$57,$FB,$66,$5F,$15,$EB,$F1,$FD,$13,$E2,$29,$FD,$8B
	.dc.b	$47,$06,$23,$FD,$0D,$18,$7D,$C7,$13,$F5,$1A,$F9,$DF,$D3,$BE,$B7
	.dc.b	$27,$C0,$D7,$DA,$FC,$2D,$BD,$5F,$0D,$E5,$8F,$7B,$EA,$EB,$FC,$9E
	.dc.b	$9F,$E7,$A9,$FE,$C5,$B7,$C1,$F8,$6E,$11,$FB,$6E,$3F,$A2,$FD,$A3
	.dc.b	$E7,$EB,$EB,$7F,$DE,$3A,$FA,$DF,$DE,$F1,$F0,$7D,$27,$94,$F4,$9F
	.dc.b	$17,$5F,$B4,$D3,$F8,$47,$1F,$C1,$D3,$CF,$F1,$5F,$AC,$CB,$EB,$7C
	.dc.b	$57,$EB,$B1,$B8,$7E,$69,$FB,$55,$B8,$FB,$FF,$13,$FB,$4E,$2F,$BC
	.dc.b	$D8,$C9,$FA,$28,$FC,$17,$C5,$7E,$EB,$3F,$C2,$7C,$4F,$F0,$0F,$93
	.dc.b	$FF,$36,$5F,$E0,$E3,$1D,$1A,$8F,$DB,$CE,$B6,$F9,$FF,$15,$C7,$A7
	.dc.b	$F1,$DF,$0D,$CF,$A3,$F2,$1F,$15,$D7,$9E,$E5,$C9,$DB,$D2,$7F,$2D
	.dc.b	$F0,$FE,$3E,$2F,$CC,$7C,$57,$97,$96,$DF,$89,$DB,$E4,$9F,$CF,$7C
	.dc.b	$4F,$AF,$97,$F4,$1F,$15,$6B,$F4,$AD,$7F,$41,$3F,$07,$97,$F4,$DF
	.dc.b	$15,$F0,$F9,$6D,$7C,$0E,$2A,$FF,$15,$E4,$47,$EB,$7E,$2B,$84,$FE
	.dc.b	$CB,$E1,$BF,$DA,$6B,$E6,$BF,$DB,$2F,$ED,$BE,$23,$3F,$32,$F8,$AF
	.dc.b	$DB,$46,$1F,$DC,$5F,$2B,$E3,$7F,$B7,$F8,$AF,$B1,$D3,$FE,$91,$F5
	.dc.b	$BF,$D4,$33,$FE,$A6,$D7,$7D,$9F,$CD,$A7,$33,$FE,$A3,$9C,$F6,$FF
	.dc.b	$A9,$7C,$66,$3F,$83,$ED,$E9,$F1,$7F,$85,$EC,$B8,$FE,$09,$AF,$C3
	.dc.b	$7C,$5B,$FE,$2F,$B3,$C9,$F1,$7E,$9D,$ED,$FF,$BD,$B8,$FE,$03,$FE
	.dc.b	$03,$DD,$5F,$06,$31,$F1,$7C,$5D,$7C,$57,$E5,$FA,$FE,$03,$F9,$9E
	.dc.b	$CB,$3F,$96,$7E,$27,$C3,$EE,$3B,$7C,$37,$E7,$FB,$7E,$47,$C3,$59
	.dc.b	$F1,$35,$FF,$66,$F3,$7F,$B2,$75,$5F,$E9,$FB,$3E,$87,$C5,$79,$6F
	.dc.b	$67,$FB,$52,$9F,$15,$EB,$8C,$F5,$A7,$31,$F0,$F9,$79,$A2,$13,$B3
	.dc.b	$EB,$BD,$7E,$27,$F7,$1D,$DF,$81,$7F,$75,$DF,$67,$F6,$EF,$E1,$7E
	.dc.b	$27,$71,$F3,$FE,$1B,$F8,$07,$7F,$E2,$5F,$F0,$0F,$CA,$7F,$BA,$B5
	.dc.b	$FC,$44,$63,$E3,$3F,$1D,$F0,$DC,$30,$FB,$1F,$0F,$F0,$6F,$F1,$CF
	.dc.b	$BD,$B1,$ED,$3E,$F7,$C3,$EF,$9B,$E6,$7C,$57,$83,$C4,$FC,$D3,$E2
	.dc.b	$E3,$B7,$FE,$06,$D7,$F0,$EF,$CA,$E4,$7C,$4F,$18,$FC,$57,$B3,$EA
	.dc.b	$F0,$BF,$49,$EF,$7B,$3C,$3B,$5C,$DF,$D4,$56,$E3,$FC,$A9,$F2,$FF
	.dc.b	$B8,$7F,$71,$E3,$7F,$BF,$63,$1E,$37,$8B,$8B,$FC,$43,$18,$DB,$5E
	.dc.b	$DF,$B7,$05,$2B,$FC,$BD,$1F,$70,$FB,$D7,$FA,$07,$C3,$ED,$7F,$04
	.dc.b	$D7,$C0,$D8,$F2,$FB,$9D,$70,$E9,$EE,$FD,$8F,$27,$CC,$FD,$AF,$22
	.dc.b	$DF,$C9,$7D,$5F,$19,$B8,$F7,$FC,$5F,$3D,$7F,$C7,$F1,$E5,$D7,$DD
	.dc.b	$D3,$F9,$3C,$FD,$5C,$DF,$76,$31,$6C,$39,$7C,$57,$3F,$8A,$F4,$65
	.dc.b	$FA,$0A,$FF,$32,$3F,$04,$65,$F0,$5D,$57,$8F,$E2,$7C,$9F,$0F,$FC
	.dc.b	$87,$D6,$FA,$4D,$AF,$AF,$3E,$9F,$8B,$F5,$7C,$5F,$96,$7E,$8E,$48
	.dc.b	$C6,$FF,$4E,$BE,$07,$C5,$CB,$FE,$29,$CF,$F7,$E3,$D3,$5F,$7B,$E2
	.dc.b	$BE,$37,$C5,$7C,$7F,$8A,$F1,$61,$ED,$6D,$7E,$12,$7E,$67,$C4,$7C
	.dc.b	$DF,$88,$F0,$EA,$6C,$7D,$1F,$88,$98,$F8,$8F,$C9,$BE,$2E,$6F,$9F
	.dc.b	$F1,$1F,$81,$FD,$17,$FB,$BD,$7E,$2F,$D4,$FE,$0B,$A2,$73,$3F,$C2
	.dc.b	$79,$7F,$85,$CF,$F9,$E2,$73,$3F,$C3,$EB,$FA,$DB,$FE,$B5,$8F,$F2
	.dc.b	$1B,$F5,$CF,$D1,$99,$F8,$7F,$17,$C4,$7C,$C9,$36,$3C,$DF,$0F,$E7
	.dc.b	$F8,$8F,$55,$FE,$92,$D1,$8C,$7D,$15,$ED,$F8,$89,$7E,$1C,$BC,$AF
	.dc.b	$F2,$AC,$13,$F4,$35,$FC,$B4,$3E,$23,$E3,$7C,$47,$C7,$DB,$CC,$79
	.dc.b	$DF,$95,$E5,$4C,$7E,$81,$D5,$F2,$5B,$E1,$BE,$87,$C3,$7D,$1F,$86
	.dc.b	$F0,$5F,$AD,$BB,$FD,$19,$FA,$DF,$0D,$F5,$FE,$1B,$E3,$1F,$1D,$FB
	.dc.b	$5F,$0D,$F6,$FE,$13,$B9,$E0,$7E,$F7,$C1,$7D,$FF,$81,$EB,$A9,$B1
	.dc.b	$F8,$3F,$80,$FE,$09,$FA,$2F,$F7,$ED,$BF,$86,$FA,$BF,$80,$E5,$FD
	.dc.b	$D4,$E6,$7F,$63,$39,$7E,$A5,$72,$F7,$39,$CF,$D5,$AF,$E6,$35,$FC
	.dc.b	$17,$C6,$E1,$C2,$95,$F8,$A6,$57,$D9,$E6,$DD,$7E,$FE,$3F,$31,$18
	.dc.b	$F6,$3D,$3E,$A7,$AB,$D3,$FD,$BD,$9B,$1E,$EF,$86,$F7,$FC,$3F,$94
	.dc.b	$F3,$3F,$0B,$E1,$E6,$5E,$13,$F8,$D7,$07,$C2,$8F,$AF,$8F,$F7,$29
	.dc.b	$FE,$29,$DF,$FA,$A8,$7C,$37,$CC,$F8,$6F,$19,$E4,$7E,$7F,$C3,$7D
	.dc.b	$0F,$85,$F8,$C7,$C7,$7E,$9F,$C2,$FD,$4F,$85,$EE,$78,$1F,$AF,$F0
	.dc.b	$BF,$63,$E1,$3D,$E7,$C0,$7E,$DF,$C1,$CD,$7F,$2B,$D1,$FE,$F0,$D7
	.dc.b	$FD,$7B,$18,$E0,$FB,$BF,$09,$FD,$17,$8B,$FA,$D7,$0F,$F0,$46,$F3
	.dc.b	$FD,$86,$DF,$C4,$3F,$AB,$BF,$E9,$F7,$F8,$67,$8D,$BF,$82,$E8,$7F
	.dc.b	$2A,$9F,$EB,$73,$FD,$1F,$B3,$A7,$C1,$F8,$7E,$0F,$C5,$F0,$7F,$BD
	.dc.b	$B3,$63,$CD,$F0,$5E,$7F,$83,$FD,$4D,$7F,$4B,$C7,$EC,$1F,$E2,$87
	.dc.b	$FA,$AB,$FE,$23,$CF,$D5,$F0,$3E,$FF,$83,$F2,$3E,$57,$E1,$7C,$0F
	.dc.b	$C3,$F8,$2F,$13,$E3,$7E,$37,$C0,$CC,$3B,$F3,$39,$7F,$B0,$67,$FD
	.dc.b	$26,$73,$18,$AF,$9D,$F0,$32,$7F,$C3,$C7,$9F,$E0,$7E,$34,$7D,$3F
	.dc.b	$80,$FA,$9F,$03,$F0,$DF,$88,$FD,$7F,$7F,$EC,$7C,$07,$C2,$7E,$1B
	.dc.b	$F6,$FE,$02,$63,$DB,$5F,$33,$F6,$3E,$03,$EA,$4F,$F8,$17,$F4,$4F
	.dc.b	$E7,$B3,$FC,$E0,$FF,$28,$B8,$F8,$4F,$EE,$17,$C8,$47,$F4,$09,$FE
	.dc.b	$ED,$3A,$DF,$FC,$05,$3D,$3B,$BF,$A7,$BF,$EE,$FA,$FD,$4F,$37,$F0
	.dc.b	$48,$FE,$87,$1F,$20,$FF,$11,$BB,$7F,$80,$9F,$91,$C3,$C3,$F0,$5E
	.dc.b	$9F,$80,$93,$EF,$C7,$C3,$F0,$3F,$DE,$B0,$57,$BF,$E0,$65,$FE,$8C
	.dc.b	$EF,$F0,$3F,$D3,$6B,$FA,$3C,$7F,$4B,$CF,$CD,$F0,$3F,$1C,$FE,$73
	.dc.b	$3F,$D3,$A7,$CA,$7C,$88,$F9,$93,$23,$C8,$7B,$3E,$07,$C9,$1F,$47
	.dc.b	$E0,$24,$F8,$F6,$F9,$BF,$03,$E3,$9F,$AD,$F0,$12,$BE,$35,$7D,$2F
	.dc.b	$81,$F1,$CF,$DB,$F8,$09,$5E,$0A,$FA,$FF,$03,$F0,$E7,$F0,$1E,$FC
	.dc.b	$AF,$06,$DF,$D1,$DB,$1F,$85,$F7,$FF,$C9,$5F,$41,$FD,$0F,$9F,$FB
	.dc.b	$6E,$5F,$83,$C3,$FD,$87,$87,$92,$B5,$BF,$F4,$07,$F9,$FC,$7C,$B3
	.dc.b	$ED,$38,$7F,$13,$E6,$6C,$FF,$18,$C6,$63,$F8,$69,$C3,$DF,$F1,$FB
	.dc.b	$FF,$E0,$39,$FE,$8D,$3F,$EA,$2E,$BF,$C9,$3F,$E6,$C8,$4F,$AB,$DE
	.dc.b	$F5,$FB,$FF,$DC,$A5,$16,$3D,$FE,$3C,$BF,$95,$4F,$7F,$CB,$3F,$13
	.dc.b	$DF,$95,$F6,$6B,$FA,$33,$FD,$84,$F7,$C7,$CA,$F7,$A4,$7C,$7F,$7F
	.dc.b	$C5,$1F,$3B,$DE,$93,$E0,$F7,$FE,$1C,$7D,$2F,$7B,$E9,$FB,$FF,$05
	.dc.b	$F8,$4F,$D6,$F7,$A5,$FD,$3F,$7F,$DD,$1F,$6B,$DF,$FB,$7E,$F7,$CE
	.dc.b	$7E,$7B,$FE,$87,$FA,$1F,$E8,$7D,$1F,$DE,$1F,$E7,$D8,$7C,$DE,$7C
	.dc.b	$3F,$E3,$38,$CB,$FE,$45,$9E,$1E,$F7,$FA,$26,$3F,$DB,$73,$FD,$09
	.dc.b	$FF,$43,$7C,$47,$F7,$77,$F7,$90,$F7,$FF,$D2,$91,$8A,$E5,$DB,$FC
	.dc.b	$46,$9E,$F7,$9F,$DF,$FF,$3B,$C6,$2F,$F4,$0F,$F4,$A7,$F2,$07,$8B
	.dc.b	$DF,$F7,$1E,$9F,$7F,$D4,$FC,$1F,$7F,$FD,$7D,$3E,$8F,$7F,$FD,$88
	.dc.b	$F9,$A2,$B9,$3F,$DA,$1B,$FC,$7F,$7F,$E5,$1E,$4F,$7F,$CA,$FC,$DF
	.dc.b	$7F,$E4,$1F,$27,$E1,$ED,$F9,$B7,$F7,$7B,$FF,$EE,$98,$F9,$2F,$FB
	.dc.b	$B5,$F8,$FF,$01,$5C,$9F,$EA,$79,$7C,$67,$ED,$23,$E8,$FC,$0F,$DD
	.dc.b	$3E,$3F,$C0,$FC,$87,$F0,$1F,$03,$FE,$E5,$9F,$AE,$FF,$BB,$5F,$AB
	.dc.b	$F0,$35,$C9,$FE,$EF,$F8,$4F,$C3,$7C,$0F,$FA,$07,$2F,$0F,$A9,$F8
	.dc.b	$9F,$82,$C6,$88,$DE,$1E,$78,$C3,$F8,$29,$DF,$AD,$FE,$48,$F7,$7F
	.dc.b	$A5,$78,$17,$E3,$7F,$D6,$B1,$FD,$57,$D4,$C6,$1F,$27,$E0,$BF,$DA
	.dc.b	$53,$FD,$26,$3F,$DA,$99,$3D,$9E,$9F,$11,$E4,$F8,$7E,$DF,$0B,$6F
	.dc.b	$FA,$76,$1C,$5F,$F7,$AE,$1C,$5C,$FC,$27,$B3,$E0,$3C,$DF,$0B,$67
	.dc.b	$27,$B6,$32,$FB,$FE,$0B,$D9,$F0,$BF,$EF,$5C,$7F,$A7,$CF,$F2,$23
	.dc.b	$C2,$FB,$A3,$5F,$0B,$F2,$BE,$0F,$E2,$3F,$33,$E1,$3E,$37,$C3,$5C
	.dc.b	$3D,$33,$B7,$D7,$CB,$F4,$A3,$E6,$7C,$37,$D4,$AF,$9D,$36,$BE,$0C
	.dc.b	$ED,$F8,$7A,$79,$F4,$FA,$2F,$D2,$7D,$59,$7D,$EB,$FB,$19,$5B,$F9
	.dc.b	$73,$57,$57,$F8,$2F,$55,$7E,$1B,$2F,$C0,$75,$7E,$23,$92,$E1,$F8
	.dc.b	$7E,$3F,$C2,$74,$FE,$31,$CC,$0F,$F6,$CF,$27,$F8,$3F,$3E,$DC,$DF
	.dc.b	$92,$E5,$FC,$67,$37,$E5,$39,$BF,$1D,$37,$0F,$C0,$CE,$DF,$C1,$FC
	.dc.b	$37,$E6,$B9,$FF,$D3,$7F,$47,$FE,$A9,$CB,$F5,$F1,$F9,$79,$D5,$7F
	.dc.b	$85,$EB,$59,$FF,$B3,$E3,$1D,$DF,$A7,$F9,$17,$83,$D3,$B7,$1E,$0F
	.dc.b	$6A,$B7,$CE,$78,$57,$8B,$CA,$E5,$E1,$5A,$FE,$05,$8A,$3C,$FF,$0B
	.dc.b	$A9,$F2,$57,$87,$C3,$F5,$FC,$2F,$ED,$BC,$3F,$6F,$C2,$DB,$F5,$4D
	.dc.b	$7E,$CC,$FC,$FE,$37,$0F,$F3,$B9,$FA,$47,$EC,$3C,$5F,$8F,$F0,$98
	.dc.b	$8C,$FC,$9B,$E4,$7F,$9B,$47,$43,$F0,$B7,$FD,$BD,$E0,$27,$87,$A7
	.dc.b	$D2,$33,$18,$7E,$87,$C3,$F0,$78,$5F,$6C,$BF,$CB,$58,$3E,$1B,$58
	.dc.b	$78,$FD,$6F,$BB,$18,$B4,$7E,$03,$E4,$30,$EB,$1F,$0C,$BD,$6F,$C2
	.dc.b	$FC,$37,$1F,$63,$F0,$FF,$0F,$6F,$F0,$93,$5F,$81,$AF,$CA,$9F,$07
	.dc.b	$B1,$F8,$EF,$86,$F1,$7B,$9F,$90,$F8,$7F,$27,$1E,$71,$F0,$B9,$7C
	.dc.b	$73,$E5,$7E,$87,$0F,$B5,$FC,$A7,$0F,$A9,$FC,$B7,$0D,$7C,$7C,$5F
	.dc.b	$15,$F6,$F0,$7B,$1F,$7F,$07,$B1,$A7,$B0,$A7,$2D,$BD,$6F,$E8,$FE
	.dc.b	$1F,$F4,$1C,$7F,$AA,$6C,$7E,$9F,$E1,$EB,$E2,$8F,$D3,$6D,$FB,$0F
	.dc.b	$86,$FD,$46,$56,$BF,$3F,$97,$ED,$3E,$1F,$F5,$8E,$63,$F2,$1C,$8F
	.dc.b	$07,$EB,$47,$ED,$7E,$1F,$F6,$5C,$DF,$BC,$7F,$69,$F0,$3F,$BE,$F8
	.dc.b	$7B,$5B,$3F,$5B,$5F,$9E,$C3,$FD,$49,$FA,$0F,$E3,$E7,$F1,$F7,$9E
	.dc.b	$FF,$D3,$4E,$79,$FF,$41,$39,$AF,$E4,$73,$FE,$2A,$9C,$C7,$FA,$7E
	.dc.b	$38,$76,$3E,$0F,$86,$29,$FF,$16,$57,$E7,$BE,$0F,$31,$8C,$3F,$5D
	.dc.b	$F0,$78,$F8,$7F,$CE,$3A,$F8,$7F,$CF,$5F,$0F,$87,$FD,$07,$C3,$72
	.dc.b	$F0,$3F,$45,$F0,$B9,$D3,$F3,$EE,$5F,$D0,$CF,$83,$AB,$F5,$1F,$0D
	.dc.b	$67,$F4,$AF,$93,$BB,$31,$ED,$9F,$D7,$71,$F8,$E3,$F5,$79,$D7,$F3
	.dc.b	$3C,$6D,$BD,$51,$FB,$07,$4F,$F0,$C8,$FF,$79,$3F,$AC,$E1,$FF,$22
	.dc.b	$3C,$B3,$F4,$3D,$34,$F6,$34,$FA,$1A,$FC,$27,$E9,$75,$FF,$03,$F3
	.dc.b	$FE,$6F,$5F,$F0,$5F,$4B,$E7,$CD,$AF,$B1,$3B,$7E,$D7,$4F,$F0,$BF
	.dc.b	$43,$EA,$F0,$7F,$0D,$F3,$BE,$BC,$DA,$FC,$04,$ED,$FC,$17,$57,$F1
	.dc.b	$6F,$81,$FF,$7D,$C7,$B6,$31,$D9,$F8,$0E,$4B,$3C,$0F,$E4,$B5,$FE
	.dc.b	$F7,$EC,$FC,$BF,$2F,$3F,$77,$F0,$FC,$96,$FF,$21,$35,$F9,$51,$ED
	.dc.b	$7E,$66,$9E,$57,$E7,$69,$E9,$72,$7F,$24,$D7,$DE,$FF,$26,$DB,$F2
	.dc.b	$BB,$79,$A3,$D4,$79,$B6,$F5,$34,$F5,$1F,$C6,$B5,$F5,$47,$EB,$39
	.dc.b	$3E,$27,$B9,$FA,$06,$D7,$EA,$7A,$AB,$F8,$38,$F9,$5E,$E7,$D0,$8F
	.dc.b	$21,$F4,$0F,$41,$F6,$0A,$7B,$8D,$9E,$83,$C9,$7E,$C7,$D0,$7E,$47
	.dc.b	$8B,$DD,$C5,$FC,$1B,$3F,$B5,$E8,$7E,$DB,$9F,$EE,$7B,$97,$0F,$AE
	.dc.b	$FE,$EB,$A3,$31,$FC,$0E,$7F,$05,$ED,$7E,$FB,$87,$F0,$9F,$03,$FC
	.dc.b	$03,$8F,$F0,$DF,$0D,$9C,$3F,$02,$E5,$FC,$1C,$FF,$0A,$FD,$07,$F1
	.dc.b	$71,$FD,$D3,$97,$FD,$FB,$C3,$FD,$F6,$5B,$7E,$47,$E1,$F8,$4D,$BF
	.dc.b	$F0,$13,$B9,$FC,$B7,$C2,$F5,$EC,$FC,$C7,$C2,$77,$E9,$FC,$D7,$C3
	.dc.b	$E7,$1F,$CA,$B9,$7F,$2F,$3E,$97,$F0,$1A,$FA,$5F,$C1,$6D,$97,$D0
	.dc.b	$E5,$F5,$47,$99,$FD,$0E,$9E,$FC,$FD,$11,$FD,$C3,$5F,$82,$FF,$92
	.dc.b	$7C,$96,$FF,$0A,$3F,$CA,$F3,$F0,$9F,$D5,$ED,$F2,$7A,$7F,$69,$F0
	.dc.b	$F6,$7F,$5E,$FC,$DE,$7C,$C7,$C6,$39,$CF,$EE,$F0,$FA,$5C,$BF,$BD
	.dc.b	$F8,$7B,$5F,$B8,$F8,$7F,$AD,$BF,$F0,$17,$31,$FA,$39,$FB,$5C,$1F
	.dc.b	$C1,$7E,$1E,$CF,$F0,$07,$EF,$70,$E6,$3C,$5D,$8E,$05,$FF,$0D,$F8
	.dc.b	$7F,$C2,$70,$FF,$0F,$F8,$7B,$87,$F0,$A7,$F1,$1C,$B9,$8E,$78,$2B
	.dc.b	$F8,$B7,$C3,$F9,$A7,$F1,$9A,$D7,$E7,$70,$FF,$B0,$1F,$A5,$C1,$ED
	.dc.b	$79,$F0,$FC,$37,$AE,$BE,$97,$EC,$37,$67,$57,$E5,$F8,$30,$FE,$66
	.dc.b	$7D,$95,$E6,$E8,$FF,$1F,$FE,$E0,$F9,$3F,$D1,$46,$1F,$D2,$74,$5A
	.dc.b	$E0,$7F,$93,$A9,$F1,$1F,$A9,$DB,$9F,$C4,$5A,$D7,$C3,$FE,$B7,$87
	.dc.b	$BB,$98,$FF,$19,$E0,$AF,$D9,$74,$F9,$3E,$23,$F6,$9D,$56,$FE,$6C
	.dc.b	$D7,$F3,$11,$EA,$F8,$7F,$DD,$75,$FB,$3E,$23,$F7,$9D,$9E,$EF,$88
	.dc.b	$CE,$1E,$97,$2F,$AE,$7C,$8F,$F9,$D3,$4F,$23,$F7,$37,$F4,$BF,$7B
	.dc.b	$7A,$FE,$A6,$3E,$0B,$E6,$3E,$36,$BE,$A6,$99,$3E,$0B,$97,$E1,$E7
	.dc.b	$FC,$4F,$AB,$EA,$FC,$47,$F8,$7B,$D2,$FA,$D5,$F4,$7E,$17,$4D,$3F
	.dc.b	$37,$1A,$7E,$6C,$FB,$1F,$DE,$67,$EC,$7D,$07,$D6,$3D,$85,$32,$7D
	.dc.b	$6E,$2F,$AA,$FA,$8F,$A2,$5E,$BF,$C8,$BF,$C1,$B8,$FE,$B4,$60,$FD
	.dc.b	$77,$1F,$D7,$9F,$3F,$2F,$E4,$BE,$17,$DC,$D8,$F5,$E1,$5F,$F4,$2E
	.dc.b	$7F,$90,$7F,$25,$F1,$1F,$17,$9F,$F3,$7F,$0B,$F1,$F8,$AD,$7C,$49
	.dc.b	$AF,$28,$FD,$0F,$C4,$7C,$DE,$3F,$D1,$FC,$47,$CF,$E7,$FD,$2F,$C3
	.dc.b	$67,$0F,$D0,$39,$7F,$45,$3F,$57,$A7,$F5,$7F,$0F,$F5,$F9,$BF,$5B
	.dc.b	$F0,$F9,$BF,$D4,$39,$7F,$55,$3E,$97,$F5,$9B,$7A,$5A,$7A,$4A,$7A
	.dc.b	$4F,$D9,$69,$E5,$7F,$57,$8F,$95,$FD,$6E,$59,$7F,$38,$E5,$FC,$F4
	.dc.b	$7E,$37,$A3,$F8,$1F,$C3,$FE,$3F,$97,$F8,$2F,$C3,$66,$FF,$80,$39
	.dc.b	$7F,$81,$4F,$E5,$78,$FF,$85,$FC,$35,$9F,$E0,$CF,$E6,$74,$CC,$7E
	.dc.b	$5A,$7F,$D9,$FF,$B8,$FF,$A5,$3C,$6F,$F7,$57,$3B,$7C,$FC,$A7,$E8
	.dc.b	$77,$B8,$FE,$7A,$6B,$F8,$48,$E7,$F0,$FF,$A7,$E1,$EB,$F1,$16,$BF
	.dc.b	$DC,$B9,$A3,$F5,$9C,$7A,$9F,$D4,$D7,$93,$E2,$3D,$51,$FC,$5B,$82
	.dc.b	$BF,$C1,$5F,$63,$EE,$3D,$87,$63,$F7,$19,$F9,$BB,$FF,$45,$B3,$5C
	.dc.b	$6F,$F0,$2A,$DA,$FF,$DD,$55,$EC,$8F,$DE,$F3,$7C,$8F,$84,$FE,$08
	.dc.b	$FC,$AF,$87,$FE,$0D,$CB,$6B,$67,$C0,$AF,$E1,$5C,$DF,$43,$E2,$2C
	.dc.b	$FC,$4D,$7F,$87,$6D,$F5,$1C,$C7,$FB,$A3,$8F,$FE,$00,$BF,$DA,$3F
	.dc.b	$41,$FF,$5D,$43,$E1,$FF,$D6,$71,$FE,$B9,$8C,$4E,$CF,$BF,$3A,$E7
	.dc.b	$FC,$0F,$C2,$F0,$E8,$FC,$1F,$C2,$DB,$FA,$EF,$4E,$9C,$C7,$86,$7F
	.dc.b	$13,$F0,$5E,$68,$1F,$E1,$9C,$BC,$EF,$F0,$8D,$FD,$8F,$F2,$6D,$AB
	.dc.b	$93,$F0,$BB,$FA,$9F,$39,$EF,$D2,$BF,$57,$9F,$BF,$33,$83,$E1,$F0
	.dc.b	$7B,$6B,$F1,$FC,$1F,$13,$A3,$F3,$FF,$07,$67,$F2,$7F,$0F,$F2,$39
	.dc.b	$7F,$46,$E6,$3E,$84,$FB,$5F,$66,$B5,$7B,$AE,$1F,$F7,$DC,$7F,$26
	.dc.b	$E0,$F6,$ED,$F9,$0E,$1F,$C0,$4F,$D3,$3F,$47,$B7,$9D,$FB,$CF,$C1
	.dc.b	$72,$7C,$13,$D6,$7E,$07,$A3,$F7,$3F,$0F,$F8,$06,$C7,$E1,$73,$AF
	.dc.b	$C7,$3F,$39,$C2,$9F,$E0,$3F,$0D,$F8,$8E,$9F,$E0,$7F,$0F,$6F,$F6
	.dc.b	$FF,$0D,$F8,$CC,$FF,$83,$B9,$8F,$CA,$4F,$E4,$39,$7F,$85,$FC,$3F
	.dc.b	$E4,$BB,$3F,$86,$FC,$3E,$6F,$F8,$0B,$97,$F8,$1C,$FE,$63,$AF,$F8
	.dc.b	$9F,$C3,$FE,$6B,$9E,$D7,$E7,$E6,$BE,$BE,$4F,$C0,$4F,$E4,$22,$1F
	.dc.b	$53,$F7,$2F,$F6,$AC,$FF,$61,$E2,$FA,$B3,$BE,$8F,$B1,$B7,$AE,$31
	.dc.b	$D0,$E0,$F8,$11,$F0,$8F,$D8,$71,$7C,$67,$F6,$5C,$3F,$01,$FD,$A7
	.dc.b	$0F,$C0,$72,$7C,$03,$27,$C0,$3F,$75,$D3,$E8,$7F,$D0,$1A,$FF,$B3
	.dc.b	$1F,$AF,$19,$7D,$9F,$0D,$FB,$EE,$8C,$6A,$BE,$1B,$FE,$8F,$0C,$C3
	.dc.b	$E2,$3F,$82,$F3,$E3,$E2,$2E,$7E,$1C,$79,$7C,$3F,$F0,$AE,$2E,$2E
	.dc.b	$A7,$E6,$CF,$7A,$FE,$1F,$C3,$DB,$E2,$3F,$88,$F4,$78,$3E,$13,$38
	.dc.b	$75,$D5,$1A,$8C,$57,$F1,$8E,$AF,$3F,$C2,$FF,$1A,$EB,$B5,$FC,$7E
	.dc.b	$6B,$F8,$1E,$94,$FB,$BE,$13,$CD,$3F,$77,$26,$B8,$8F,$9B,$BF,$B5
	.dc.b	$F2,$39,$7F,$11,$1E,$87,$DF,$2C,$FD,$11,$8D,$34,$F7,$72,$F8,$63
	.dc.b	$F2,$9C,$9F,$E3,$1F,$D0,$7F,$D7,$B3,$F9,$5F,$86,$CF,$8D,$FC,$6A
	.dc.b	$73,$C7,$F6,$FE,$1F,$37,$FD,$4E,$7F,$D8,$0F,$92,$33,$82,$78,$E7
	.dc.b	$FE,$C7,$8C,$73,$FE,$0B,$E2,$77,$5F,$13,$0F,$A9,$F1,$18,$AF,$C3
	.dc.b	$BA,$DF,$84,$FF,$AD,$A9,$C1,$F8,$CF,$87,$E7,$C1,$6B,$53,$5F,$DB
	.dc.b	$57,$C9,$7F,$05,$F0,$9E,$2E,$2F,$CA,$FC,$2F,$93,$A3,$F2,$FF,$07
	.dc.b	$9B,$FC,$43,$97,$F1,$53,$E7,$7F,$19,$BF,$9D,$F5,$9E,$7D,$FD,$6F
	.dc.b	$A7,$6F,$5B,$E4,$D3,$CD,$C1,$F5,$76,$BC,$9F,$4E,$6D,$7D,$19,$FD
	.dc.b	$5E,$DF,$2B,$9B,$F5,$BF,$07,$6B,$F3,$CF,$CE,$E7,$CC,$7C,$99,$FD
	.dc.b	$9F,$C3,$FA,$63,$F1,$BB,$7A,$5F,$C7,$F2,$E5,$FC,$D3,$97,$F3,$91
	.dc.b	$F6,$B3,$F2,$EB,$F8,$1D,$2F,$27,$D6,$BF,$D8,$72,$FD,$67,$E8,$9F
	.dc.b	$6F,$9F,$F8,$2F,$C3,$FE,$13,$2F,$E0,$FF,$0F,$F8,$69,$B5,$F8,$19
	.dc.b	$DB,$F8,$3D,$3F,$88,$F8,$7F,$E9,$FD,$3F,$3D,$18,$F8,$8F,$C7,$E7
	.dc.b	$6F,$81,$FC,$3E,$FF,$D1,$4F,$E1,$B8,$F9,$FC,$37,$E5,$BA,$3A,$FC
	.dc.b	$37,$E6,$26,$D7,$E4,$E7,$6F,$E5,$7A,$BC,$7F,$0D,$F9,$EE,$8B,$39
	.dc.b	$3C,$11,$97,$C5,$F0,$DF,$A4,$DB,$CD,$1F,$98,$6E,$FF,$B9,$DA,$F5
	.dc.b	$7E,$DE,$7E,$0F,$C3,$F9,$6F,$D9,$BF,$A5,$F7,$6F,$97,$F9,$4F,$02
	.dc.b	$35,$18,$9F,$DA,$F4,$FC,$CF,$87,$B3,$F1,$FE,$1F,$F7,$3D,$1F,$41
	.dc.b	$CC,$74,$9F,$DE,$F4,$7D,$4F,$87,$B3,$F3,$DF,$E0,$FF,$A0,$7F,$B5
	.dc.b	$63,$FB,$16,$89,$DC,$79,$BE,$23,$7B,$7D,$EF,$88,$B3,$F6,$7E,$1F
	.dc.b	$9E,$FF,$82,$73,$1E,$E9,$EF,$C1,$F8,$6F,$88,$B3,$F8,$17,$C7,$CF
	.dc.b	$98,$FF,$1D,$F2,$7F,$8F,$36,$F8,$D1,$F8,$7F,$86,$F2,$BE,$8D,$AB
	.dc.b	$FE,$50,$F2,$9C,$1E,$B8,$E6,$7F,$9E,$EF,$E0,$3B,$8D,$3F,$C5,$B8
	.dc.b	$7F,$37,$C3,$EA,$8F,$8B,$B5,$78,$6F,$F0,$1F,$84,$76,$E8,$63,$F9
	.dc.b	$C8,$E2,$65,$74,$FE,$A3,$E1,$BE,$9B,$FA,$AF,$87,$FA,$BC,$D6,$F8
	.dc.b	$4E,$DE,$5C,$FF,$B0,$F8,$8F,$B6,$FE,$CB,$E2,$3E,$EF,$15,$AD,$9F
	.dc.b	$AF,$AF,$C0,$EF,$FB,$8F,$88,$FC,$1F,$05,$AF,$34,$D7,$FC,$5D,$B2
	.dc.b	$7F,$7F,$F0,$5F,$8A,$E1,$FE,$03,$F0,$7E,$8B,$FD,$E3,$96,$9F,$07
	.dc.b	$7F,$F0,$BF,$EE,$7F,$EC,$F8,$FA,$9B,$FA,$DF,$39,$FA,$6D,$FD,$6F
	.dc.b	$F0,$77,$DE,$FA,$CF,$39,$FC,$2F,$5F,$5B,$4C,$9F,$B0,$72,$FE,$CB
	.dc.b	$2F,$D1,$F5,$79,$3E,$1E,$CF,$F8,$8F,$17,$C4,$7E,$A3,$AB,$D2,$EA
	.dc.b	$7F,$C4,$72,$AF,$2C,$7E,$FB,$8A,$BF,$AE,$CF,$D6,$FB,$4F,$79,$FB
	.dc.b	$28,$FA,$AF,$A4,$FA,$3C,$1F,$4F,$9D,$D7,$FA,$5E,$4C,$3F,$A7,$9F
	.dc.b	$AB,$5F,$8F,$E8,$F9,$FF,$13,$FC,$03,$A7,$E8,$FC,$4F,$F0,$29,$B5
	.dc.b	$F9,$59,$DB,$F9,$7C,$FE,$BF,$C4,$FF,$08,$6C,$7E,$6F,$5A,$FE,$1A
	.dc.b	$3E,$EF,$C4,$7F,$0E,$E5,$FB,$FF,$13,$6B,$E6,$7C,$2F,$F1,$2E,$7F
	.dc.b	$C1,$B9,$8F,$97,$3F,$C5,$BA,$BF,$0F,$F1,$3E,$79,$FA,$9F,$0D,$96
	.dc.b	$9E,$F3,$EC,$47,$9D,$FB,$5A,$D7,$EB,$BF,$C8,$B5,$F4,$F2,$FC,$C9
	.dc.b	$FE,$0D,$B7,$A9,$FD,$C6,$BE,$A6,$9E,$A3,$F7,$9A,$FA,$9C,$9E,$A3
	.dc.b	$F8,$27,$7F,$F9,$E5,$FF,$63,$F4,$3E,$1B,$FC,$73,$39,$F1,$12,$F8
	.dc.b	$9C,$F5,$FE,$A1,$DE,$1F,$AA,$C3,$8F,$67,$EB,$3E,$27,$9F,$57,$EB
	.dc.b	$BE,$1F,$37,$F9,$B7,$2F,$E7,$67,$C3,$D7,$FB,$4F,$1E,$CF,$E8,$7E
	.dc.b	$1F,$CB,$CB,$FB,$87,$31,$FA,$09,$F2,$BF,$A7,$F8,$7A,$FE,$91,$F6
	.dc.b	$F6,$7F,$00,$73,$1F,$EB,$1E,$0F,$F5,$73,$F1,$B5,$F5,$3F,$23,$5F
	.dc.b	$53,$93,$D4,$64,$F5,$1F,$3B,$3F,$53,$4F,$51,$F4,$B3,$F5,$39,$3D
	.dc.b	$47,$D6,$EB,$FE,$2B,$F0,$DF,$53,$B6,$CE,$4F,$E0,$91,$97,$F8,$37
	.dc.b	$C3,$7D,$AE,$CF,$E3,$BF,$0B,$67,$F8,$53,$F7,$BB,$B3,$1F,$84,$9F
	.dc.b	$F5,$CF,$EE,$9F,$ED,$6F,$11,$2E,$FF,$EE,$9D,$1A,$F8,$9F,$C3,$F5
	.dc.b	$5A,$FC,$6C,$D7,$FD,$C9,$DB,$F8,$97,$FD,$79,$8F,$FB,$05,$F5,$1F
	.dc.b	$EB,$A7,$FD,$04,$F0,$DF,$C7,$F0,$7F,$94,$E1,$B3,$F9,$A9,$F3,$FC
	.dc.b	$0F,$B6,$3F,$90,$F0,$57,$EE,$3E,$F7,$E0,$9E,$F3,$C7,$B7,$E8,$33
	.dc.b	$F4,$6D,$FB,$13,$FC,$25,$AE,$27,$E1,$C7,$C2,$AF,$5E,$DF,$09,$F6
	.dc.b	$FC,$27,$9D,$F7,$FC,$15,$7D,$DF,$0D,$FA,$EE,$FF,$9E,$E6,$3E,$78
	.dc.b	$F0,$9C,$5F,$E9,$78,$F9,$7C,$9F,$61,$A7,$B0,$F9,$FC,$B9,$D3,$D5
	.dc.b	$CA,$AF,$BB,$B7,$BE,$FD,$27,$F0,$4D,$FD,$CF,$DF,$AF,$E0,$8E,$8F
	.dc.b	$B7,$1F,$86,$DB,$F8,$3F,$83,$F8,$8F,$86,$B5,$F7,$BE,$17,$F8,$5F
	.dc.b	$85,$F8,$C7,$31,$F7,$E7,$F8,$7F,$87,$F0,$23,$FD,$AB,$F3,$BF,$87
	.dc.b	$7F,$13,$E1,$E5,$FA,$31,$9D,$FD,$B5,$99,$FC,$B4,$E7,$6F,$AF,$B3
	.dc.b	$0D,$60,$A7,$27,$F2,$0F,$8A,$CC,$36,$FE,$47,$DF,$87,$59,$F8,$F0
	.dc.b	$E5,$1A,$8C,$57,$F9,$01,$FE,$5F,$DF,$FA,$CF,$88,$FE,$5A,$D8,$FE
	.dc.b	$67,$D9,$5F,$D3,$C7,$EC,$BE,$1F,$FD,$D7,$FA,$0F,$FB,$CB,$A5,$F1
	.dc.b	$18,$E2,$FD,$BF,$C1,$6B,$BE,$E5,$FE,$41,$C3,$85,$F2,$E9,$FD,$EF
	.dc.b	$C1,$F4,$F1,$7F,$7F,$F0,$F9,$C3,$F6,$AE,$5F,$DB,$CF,$A1,$FD,$CF
	.dc.b	$C3,$D7,$F7,$1F,$0F,$FD,$23,$87,$F8,$4B,$98,$FC,$DF,$85,$F9,$AA
	.dc.b	$F4,$47,$F0,$2C,$EB,$FC,$07,$6F,$53,$FC,$19,$CB,$CB,$83,$E0,$BE
	.dc.b	$DD,$7D,$6F,$BF,$5F,$5B,$93,$D6,$64,$F5,$9F,$17,$C6,$FE,$3F,$F0
	.dc.b	$FF,$3B,$C6,$B3,$93,$F8,$BC,$65,$FE,$33,$F0,$FF,$53,$C7,$FC,$5F
	.dc.b	$CE,$9C,$4F,$E1,$63,$0F,$D3,$F0,$F7,$3A,$8F,$D0,$76,$A9,$5F,$80
	.dc.b	$F1,$3A,$96,$BF,$45,$C4,$FA,$77,$5F,$E1,$BB,$7F,$45,$39,$8E,$15
	.dc.b	$DB,$16,$1F,$8C,$9F,$E7,$FF,$13,$F8,$EE,$5F,$E8,$3F,$0B,$9C,$3F
	.dc.b	$9B,$B9,$7F,$9C,$CF,$E5,$39,$BE,$09,$F9,$6E,$AF,$86,$66,$7E,$13
	.dc.b	$97,$E2,$47,$E7,$3A,$FE,$4F,$4D,$9F,$90,$FE,$80,$CB,$F9,$B8,$F9
	.dc.b	$E7,$E9,$3A,$BD,$FF,$BA,$FF,$BF,$FB,$FF,$9A,$F7,$3A,$33,$F1,$5B
	.dc.b	$C7,$E4,$DF,$FB,$FB,$9F,$F9,$F5,$EE,$73,$F1,$7F,$CC,$A7,$FC,$65
	.dc.b	$0F,$CB,$9E,$37,$E0,$3B,$3F,$6A,$7E,$0B,$8B,$86,$BF,$CE,$AB,$77
	.dc.b	$F8,$39,$DB,$F8,$5A,$FD,$EC,$7A,$BE,$32,$CF,$9B,$E3,$3F,$80,$F2
	.dc.b	$FB,$9C,$C7,$F2,$7E,$A6,$EB,$F0,$CF,$F7,$7F,$D0,$BF,$D9,$AF,$F4
	.dc.b	$B9,$7C,$6E,$B7,$56,$B0,$E7,$39,$8F,$F6,$45,$FF,$B6,$EF,$8B,$FE
	.dc.b	$CD,$8E,$11,$FE,$CA,$8F,$F6,$E9,$F1,$F7,$FE,$5F,$EB,$7F,$BA,$23
	.dc.b	$FC,$15,$C1,$F5,$B7,$CE,$BE,$89,$E0,$7F,$2D,$AE,$1A,$7D,$6A,$E7
	.dc.b	$9F,$97,$5F,$C0,$FC,$57,$9F,$1B,$8F,$96,$6B,$E0,$3E,$24,$7E,$23
	.dc.b	$E2,$BD,$D7,$F8,$AF,$46,$D7,$E1,$BE,$23,$E1,$7C,$6F,$E3,$9C,$C7
	.dc.b	$D1,$E0,$FF,$7B,$EB,$F8,$6E,$5F,$BF,$3F,$BC,$DD,$7F,$80,$8C,$7C
	.dc.b	$4F,$BE,$B7,$1F,$8F,$AF,$F6,$E3,$FC,$E7,$83,$E4,$DF,$D3,$E0,$FC
	.dc.b	$FE,$1F,$57,$6F,$D0,$FC,$66,$70,$FC,$EB,$97,$F3,$F3,$F6,$F6,$FD
	.dc.b	$3F,$C6,$7D,$DC,$7F,$53,$F1,$99,$BF,$D2,$B9,$7F,$4F,$3E,$7E,$5F
	.dc.b	$B7,$F1,$7E,$B9,$F4,$39,$7E,$C3,$97,$FC,$4F,$B3,$E2,$7F,$19,$9F
	.dc.b	$EE,$3E,$33,$F1,$DA,$DB,$FC,$64,$D7,$EC,$71,$7E,$5A,$7F,$7F,$8F
	.dc.b	$E5,$73,$FE,$03,$8F,$E5,$F4,$B8,$7E,$56,$6B,$CB,$A5,$3F,$C2,$3E
	.dc.b	$2F,$D1,$3F,$CC,$BE,$87,$F9,$9F,$5F,$E5,$63,$FB,$F4,$61,$FF,$31
	.dc.b	$F3,$7D,$87,$FC,$87,$19,$D5,$E0,$39,$FF,$C1,$F3,$FE,$F3,$AF,$D4
	.dc.b	$70,$FF,$18,$F8,$CF,$D7,$36,$3F,$53,$8F,$57,$F7,$13,$A7,$F8,$CF
	.dc.b	$C6,$7E,$D7,$87,$F9,$0F,$C6,$FE,$DF,$8B,$FA,$0C,$78,$F8,$4E,$2F
	.dc.b	$47,$F2,$2A,$F8,$2F,$F1,$FC,$DF,$19,$F6,$63,$FD,$01,$0E,$6D,$C7
	.dc.b	$C0,$75,$C0,$9E,$33,$EA,$E7,$F1,$FC,$6F,$A6,$3F,$8C,$64,$F8,$CE
	.dc.b	$71,$FE,$5F,$8C,$3F,$18,$FE,$BE,$FC,$97,$F7,$47,$F0,$63,$9F,$C6
	.dc.b	$7E,$AF,$DE,$F8,$3F,$19,$D3,$C6,$F8,$2E,$6B,$9D,$78,$34,$F4,$7C
	.dc.b	$3F,$AB,$5F,$F2,$1D,$7A,$4F,$C0,$FC,$26,$67,$A6,$6F,$8C,$F6,$57
	.dc.b	$F8,$7A,$30,$F1,$3D,$F1,$A7,$FD,$57,$9F,$D4,$E1,$E7,$C3,$E5,$9F
	.dc.b	$8B,$F1,$9D,$8C,$3D,$63,$D1,$9F,$FB,$5A,$3F,$4B,$D7,$F8,$F7,$F2
	.dc.b	$FC,$7A,$3D,$9F,$0D,$E8,$F8,$AF,$14,$FC,$FF,$8B,$F5,$98,$D3,$F0
	.dc.b	$F3,$EE,$F8,$CF,$0D,$7A,$37,$FF,$41,$BF,$4B,$B7,$F2,$D7,$F8,$EA
	.dc.b	$DC,$FD,$BE,$1F,$9D,$EE,$FF,$A6,$79,$3F,$9D,$4F,$F9,$A6,$31,$E0
	.dc.b	$FF,$67,$DB,$FC,$EB,$18,$9F,$F6,$86,$FE,$D7,$FC,$4F,$F4,$3F,$EC
	.dc.b	$6C,$3F,$98,$5F,$E8,$39,$BD,$5B,$7D,$AB,$F3,$4E,$5F,$C6,$1F,$92
	.dc.b	$F8,$AF,$B7,$F1,$BE,$8D,$FF,$4E,$E2,$7F,$67,$D7,$FA,$CF,$8D,$FC
	.dc.b	$0F,$C6,$FE,$96,$FF,$5B,$CD,$F8,$43,$19,$FF,$A3,$F0,$FC,$11,$F8
	.dc.b	$5C,$FC,$B8,$FE,$28,$F3,$BF,$8C,$F8,$BF,$17,$EE,$DF,$EF,$1A,$FE
	.dc.b	$1F,$3F,$E3,$C8,$C4,$7F,$84,$EB,$F7,$F5,$F9,$F9,$7C,$6F,$95,$B5
	.dc.b	$E2,$3F,$22,$E1,$FC,$4F,$93,$D7,$E3,$39,$7C,$67,$E0,$E9,$F1,$9D
	.dc.b	$4C,$47,$7F,$8C,$F3,$97,$0F,$8E,$7F,$95,$AB,$FC,$C2,$FD,$23,$1F
	.dc.b	$17,$E5,$F8,$9F,$41,$E7,$7D,$5D,$5E,$12,$CF,$D8,$3F,$D1,$75,$FE
	.dc.b	$60,$7D,$87,$FB,$9D,$FF,$42,$BF,$95,$DB,$E4,$7C,$37,$C5,$3E,$6F
	.dc.b	$C4,$7C,$7F,$88,$F8,$C5,$AF,$EE,$07,$A7,$E1,$FE,$67,$C4,$7C,$32
	.dc.b	$C7,$8C,$F9,$7F,$11,$F4,$7E,$23,$EC,$F1,$FD,$33,$ED,$BF,$57,$E1
	.dc.b	$BE,$A1,$6B,$B9,$FE,$3B,$D5,$F0,$DF,$C6,$7F,$77,$FF,$2E,$4F,$F8
	.dc.b	$EF,$D9,$FF,$3E,$C7,$F9,$67,$C6,$FD,$44,$FF,$A1,$23,$0E,$CC,$7C
	.dc.b	$47,$3F,$88,$FC,$F4,$FF,$8B,$2B,$8D,$7F,$92,$E3,$0F,$F5,$EE,$3E
	.dc.b	$F5,$E3,$F8,$7F,$D0,$57,$F8,$FE,$BE,$E1,$C7,$E1,$F3,$3E,$9F,$86
	.dc.b	$FE,$19,$4F,$86,$F6,$18,$8F,$77,$C3,$79,$0B,$87,$E2,$8F,$27,$C3
	.dc.b	$7C,$3F,$86,$F0,$FC,$37,$C5,$3C,$6F,$C7,$F8,$6E,$FF,$0D,$F2,$4F
	.dc.b	$0B,$F2,$F8,$9C,$3D,$71,$FF,$6A,$EA,$9C,$C7,$B0,$CF,$91,$8F,$43
	.dc.b	$FC,$7D,$C7,$F4,$AB,$18,$7C,$AB,$E7,$F0,$78,$7F,$8C,$F5,$7D,$8F
	.dc.b	$86,$C4,$7F,$59,$37,$5F,$82,$7C,$71,$AF,$5F,$F9,$6E,$8F,$86,$E9
	.dc.b	$A7,$F9,$E3,$3F,$C1,$FC,$36,$FC,$EF,$C2,$D6,$BE,$1B,$D9,$9F,$E1
	.dc.b	$6B,$4F,$E1,$FE,$1B,$B7,$07,$E3,$1C,$C7,$FA,$37,$F7,$83,$A3,$86
	.dc.b	$FF,$B0,$F4,$D2,$F8,$7F,$2F,$B7,$FC,$DE,$FF,$E0,$0C,$E6,$F8,$3F
	.dc.b	$E0,$17,$E0,$47,$F6,$48,$D3,$F4,$4C,$7C,$2F,$AF,$7F,$F6,$EC,$A3
	.dc.b	$CF,$3D,$BE,$1B,$4F,$FB,$9A,$75,$7E,$0F,$86,$F5,$4F,$B5,$E2,$F8
	.dc.b	$7E,$1B,$2F,$A2,$32,$B4,$F7,$FC,$2F,$BB,$A3,$C7,$F0,$DA,$9C,$E3
	.dc.b	$EC,$BE,$19,$2B,$4B,$E1,$3C,$F3,$F3,$DC,$FC,$2E,$23,$E4,$C7,$0F
	.dc.b	$85,$F2,$C7,$C7,$79,$FC,$2E,$23,$BC,$76,$F8,$4F,$34,$7C,$3F,$84
	.dc.b	$C6,$7E,$5D,$3E,$CC,$79,$3E,$13,$C1,$5F,$4F,$E1,$31,$7F,$C7,$E3
	.dc.b	$D7,$1E,$AF,$84,$F0,$61,$F8,$2F,$84,$C5,$FE,$B6,$FE,$E4,$7C,$0E
	.dc.b	$FF,$1C,$7E,$35,$F8,$7D,$98,$8F,$C3,$C7,$C6,$F8,$4F,$1C,$7D,$1F
	.dc.b	$63,$F7,$38,$7E,$2A,$73,$8A,$73,$5F,$89,$38,$7C,$27,$9A,$3F,$B2
	.dc.b	$DF,$E6,$57,$C2,$66,$7E,$D4,$3E,$0F,$C9,$7F,$E3,$88,$FE,$C7,$7C
	.dc.b	$67,$33,$E5,$F8,$4E,$3F,$09,$E4,$B3,$D3,$E3,$B0,$F6,$F8,$4F,$27
	.dc.b	$C2,$DC,$BE,$46,$3E,$F8,$FC,$8F,$D8,$3F,$7F,$74,$78,$23,$0E,$A1
	.dc.b	$D1,$97,$78,$7B,$FE,$1F,$C8,$7B,$BE,$3A,$C7,$C3,$F8,$7F,$8D,$F1
	.dc.b	$1F,$09,$B1,$FD,$BB,$D0,$FF,$6D,$F9,$9F,$01,$7D,$13,$DC,$FD,$3C
	.dc.b	$6A,$BF,$44,$EA,$3E,$6F,$C4,$FC,$5F,$89,$B5,$F8,$DE,$BF,$B5,$1F
	.dc.b	$4B,$9F,$EC,$1F,$53,$E2,$6D,$78,$F2,$FF,$5D,$BD,$9B,$CB,$F8,$1C
	.dc.b	$FC,$BE,$87,$9E,$73,$1F,$E4,$89,$FE,$3A,$4C,$BF,$15,$F1,$3F,$64
	.dc.b	$FC,$1F,$99,$71,$FC,$47,$C4,$FE,$2F,$E2,$3F,$0E,$D8,$F2,$7A,$DD
	.dc.b	$FD,$6F,$87,$A2,$31,$3F,$CA,$7F,$7A,$7C,$4F,$E6,$97,$C8,$7F,$0B
	.dc.b	$73,$3F,$15,$CB,$FC,$D3,$7F,$EA,$9E,$D7,$FA,$FB,$83,$D6,$7F,$A3
	.dc.b	$A9,$F1,$5F,$D0,$FE,$03,$33,$98,$FE,$8D,$F0,$38,$8F,$E3,$D4,$F9
	.dc.b	$0E,$11,$C3,$91,$3D,$33,$FC,$27,$D1,$3D,$DF,$C2,$3F,$EA,$58,$CC
	.dc.b	$63,$DC,$F6,$7C,$5F,$BC,$F7,$7C,$65,$AF,$F7,$C6,$5D,$F1,$7C,$5F
	.dc.b	$F5,$BF,$88,$F3,$69,$E2,$8F,$EB,$DF,$11,$8A,$FF,$07,$61,$F1,$4F
	.dc.b	$8D,$C5,$FF,$00,$31,$E9,$7F,$96,$D9,$EF,$7A,$0C,$7B,$97,$0F,$A1
	.dc.b	$E2,$7D,$5F,$8B,$FA,$8D,$8E,$D8,$B9,$D9,$FE,$8E,$3D,$5F,$15,$FD
	.dc.b	$CF,$0F,$37,$C5,$FF,$75,$C7,$D1,$39,$AF,$81,$3E,$5F,$7B,$E5,$9E
	.dc.b	$7F,$7A,$E7,$E0,$CF,$BE,$1F,$07,$87,$FC,$01,$85,$AD,$68,$AF,$8D
	.dc.b	$C7,$AC,$FE,$76,$1F,$E0,$7E,$5F,$95,$E2,$D9,$FA,$4F,$F8,$2F,$9B
	.dc.b	$E7,$75,$66,$7F,$1F,$CF,$F8,$4A,$FE,$03,$EA,$7F,$32,$DF,$FD,$7F
	.dc.b	$5F,$94,$51,$FC,$CA,$3F,$6C,$EF,$6F,$CE,$7C,$3F,$F8,$26,$7F,$3D
	.dc.b	$3B,$AF,$CF,$F4,$7D,$FE,$EF,$F0,$F7,$C6,$7E,$07,$E3,$3F,$C4,$13
	.dc.b	$9C,$BE,$0C,$E5,$F8,$7E,$17,$E1,$FE,$2B,$FC,$4C,$D8,$F8,$FE,$1F
	.dc.b	$E2,$FC,$4C,$EB,$F9,$3B,$F6,$FB,$5F,$C8,$76,$3F,$01,$C3,$FE,$31
	.dc.b	$F1,$7F,$29,$E3,$FF,$8C,$FC,$6B,$9F,$D4,$9F,$CD,$FE,$F7,$FE,$2E
	.dc.b	$8F,$AF,$3A,$CB,$FD,$FB,$E6,$EB,$D1,$FF,$16,$F9,$D6,$FF,$05,$CC
	.dc.b	$AE,$7F,$23,$AC,$3F,$71,$97,$F8,$B3,$9F,$F9,$04,$7D,$3D,$DE,$A7
	.dc.b	$AF,$D3,$FB,$2D,$CF,$AF,$9F,$E6,$F8,$CF,$F1,$EB,$63,$C3,$E8,$FA
	.dc.b	$BD,$0C,$E5,$E0,$BF,$F2,$1F,$AD,$EE,$F4,$7F,$C8,$BE,$B5,$C7,$D3
	.dc.b	$3F,$0B,$E3,$33,$87,$A2,$FF,$C9,$7A,$7F,$15,$DF,$F1,$3E,$AF,$F9
	.dc.b	$3B,$7C,$70,$38,$37,$8F,$F1,$EB,$DE,$9F,$1A,$F7,$5F,$E5,$6F,$5F
	.dc.b	$A7,$D1,$7F,$2E,$C3,$F7,$57,$F9,$29,$F0,$7E,$7C,$F7,$FF,$CB,$BE
	.dc.b	$D7,$0F,$0F,$96,$1D,$EB,$58,$7D,$4F,$81,$FB,$BF,$1D,$FE,$64,$F8
	.dc.b	$2F,$BF,$F1,$DF,$E6,$69,$CD,$7D,$A9,$CB,$F7,$3C,$4F,$2F,$85,$FE
	.dc.b	$6C,$6C,$7E,$03,$8F,$D3,$E8,$67,$4F,$81,$C3,$FC,$E6,$BF,$0B,$C1
	.dc.b	$EE,$F4,$3F,$CE,$AF,$C0,$F2,$FF,$CE,$FE,$FE,$6B,$78,$72,$9F,$F4
	.dc.b	$57,$9B,$F9,$6F,$8D,$FF,$46,$79,$76,$FF,$86,$CF,$E6,$BD,$DC,$E1
	.dc.b	$F9,$0B,$FF,$49,$79,$BF,$9E,$F7,$7F,$D2,$9E,$77,$E8,$3E,$37,$37
	.dc.b	$F9,$69,$CB,$F9,$88,$FF,$4F,$FE,$91,$FD,$56,$3F,$CE,$5E,$57,$F8
	.dc.b	$71,$FA,$E7,$F3,$18,$FF,$49,$3F,$A0,$F8,$5E,$2F,$EB,$BE,$17,$9F
	.dc.b	$A3,$FB,$09,$F6,$CF,$E9,$BE,$3B,$D2,$FF,$1C,$C1,$1A,$8C,$7C,$06
	.dc.b	$9F,$F0,$44,$EB,$87,$F9,$7E,$DE,$F7,$F9,$8E,$DE,$F7,$35,$EF,$72
	.dc.b	$FB,$DF,$E7,$9A,$7B,$FD,$96,$1E,$FF,$1F,$FD,$A5,$3E,$EB,$F3,$69
	.dc.b	$EE,$7F,$7F,$B7,$B9,$F8,$58,$FF,$47,$F8,$1F,$F2,$2C,$7F,$04,$F8
	.dc.b	$1F,$F6,$5C,$62,$3E,$04,$63,$7F,$E2,$BF,$1D,$AC,$3F,$8B,$CF,$0E
	.dc.b	$0F,$D4,$F3,$72,$8B,$97,$79,$FD,$6F,$3E,$72,$FC,$AD,$F8,$31,$FD
	.dc.b	$8F,$17,$8B,$E3,$6E,$3E,$79,$FD,$AF,$1E,$70,$FC,$ED,$FE,$13,$D5
	.dc.b	$FE,$53,$85,$9F,$E3,$EF,$E2,$3D,$3F,$E5,$BF,$1D,$99,$FB,$93,$F8
	.dc.b	$CF,$57,$F9,$8F,$C7,$7E,$3B,$D3,$BA,$FD,$29,$FF,$07,$FE,$F9,$FD
	.dc.b	$57,$0F,$C7,$56,$9F,$C8,$F4,$3A,$1E,$AF,$0F,$8E,$FC,$CB,$72,$FB
	.dc.b	$BE,$77,$4F,$8F,$CE,$3F,$E2,$DC,$FF,$AF,$57,$F8,$F7,$17,$0F,$8B
	.dc.b	$E3,$FF,$47,$E8,$79,$3D,$0F,$66,$3F,$E5,$9C,$7F,$D3,$18,$FE,$2A
	.dc.b	$B4,$FA,$E7,$FA,$1F,$0F,$BD,$B1,$97,$FB,$DD,$7B,$F7,$F6,$70,$7B
	.dc.b	$DF,$C8,$EF,$EB,$AF,$83,$AF,$C0,$7D,$5C,$1F,$B8,$E0,$F7,$65,$F9
	.dc.b	$CB,$E1,$87,$E6,$B4,$F9,$BE,$87,$B3,$1F,$85,$E7,$FB,$11,$FE,$F1
	.dc.b	$D7,$FD,$57,$8F,$B6,$36,$FA,$5F,$D9,$ED,$ED,$7F,$74,$FD,$F7,$E7
	.dc.b	$46,$5D,$BE,$57,$D2,$D7,$E1,$E9,$EB,$7E,$B7,$17,$D8,$73,$3F,$B8
	.dc.b	$F1,$FF,$0F,$E7,$FF,$18,$F2,$7F,$13,$E6,$7F,$1A,$9C,$D7,$EF,$A7
	.dc.b	$2F,$F0,$0F,$2B,$F1,$FE,$67,$F2,$07,$F2,$3E,$CF,$F2,$2F,$33,$35
	.dc.b	$BC,$FF,$11,$3F,$C9,$BC,$9F,$CB,$FB,$5F,$CA,$3C,$8F,$CC,$FB,$59
	.dc.b	$BF,$C7,$4E,$5F,$C8,$47,$9D,$FC,$97,$B3,$FC,$C3,$CB,$FD,$0B,$98
	.dc.b	$FC,$2F,$B1,$97,$79,$F8,$DF,$C4,$ED,$E9,$7C,$B1,$FC,$73,$5F,$4C
	.dc.b	$7E,$23,$6F,$6C,$FF,$20,$D7,$D2,$FF,$22,$D7,$D2,$E6,$BD,$2E,$5F
	.dc.b	$4B,$FC,$A3,$CB,$FD,$BF,$B1,$FD,$21,$B1,$FC,$B3,$CD,$FD,$DF,$B9
	.dc.b	$9B,$FD,$85,$FF,$4E,$F3,$3F,$7F,$EF,$7F,$50,$F2,$6E,$3F,$CD,$27
	.dc.b	$F8,$1F,$B9,$9C,$3F,$71,$7F,$D5,$BC,$BF,$E0,$FE,$E5,$9F,$DE,$3F
	.dc.b	$D6,$BB,$FF,$85,$FB,$B9,$9F,$D7,$CB,$2F,$F3,$E7,$6D,$9F,$F5,$87
	.dc.b	$07,$F8,$57,$C5,$C1,$FE,$0E,$8C,$FC,$07,$AD,$FF,$39,$CE,$EF,$FD
	.dc.b	$8D,$7C,$E7,$35,$EF,$7D,$33,$FD,$43,$7F,$7C,$7E,$F3,$7F,$7C,$E6
	.dc.b	$36,$7F,$00,$D7,$DF,$1F,$D6,$1B,$1F,$C1,$35,$F7,$CE,$63,$E1,$BE
	.dc.b	$EF,$0B,$FD,$77,$9F,$FB,$2E,$7F,$95,$E6,$9C,$FC,$07,$C0,$BF,$5F
	.dc.b	$87,$C3,$DD,$C6,$1F,$EB,$1C,$38,$3F,$22,$F3,$BF,$F8,$E7,$CE,$EF
	.dc.b	$EE,$E2,$DE,$7F,$8B,$DC,$C5,$6B,$4F,$A7,$39,$7E,$AF,$97,$E8,$F7
	.dc.b	$39,$79,$96,$73,$AF,$D8,$F8,$1C,$BF,$6A,$BC,$1E,$67,$C0,$F6,$BC
	.dc.b	$5A,$DC,$3F,$03,$3F,$13,$E2,$B3,$87,$E0,$AF,$CA,$FE,$13,$7F,$56
	.dc.b	$BE,$58,$F9,$3B,$7A,$5F,$97,$19,$E0,$FC,$4E,$6D,$FD,$15,$F4,$37
	.dc.b	$B5,$A9,$FA,$BB,$FA,$AB,$F1,$BB,$FC,$9F,$1B,$EC,$FC,$45,$9F,$CB
	.dc.b	$FC,$57,$A1,$FC,$9C,$65,$FC,$AC,$7A,$9F,$CB,$ED,$EA,$7F,$3F,$C7
	.dc.b	$97,$FA,$08,$F7,$DD,$2A,$E1,$3F,$A0,$AF,$B9,$A7,$AB,$97,$F4,$3B
	.dc.b	$DA,$FC,$34,$FB,$27,$F5,$F3,$9D,$FF,$CF,$DE,$7F,$E5,$3E,$23,$F0
	.dc.b	$CD,$AF,$DA,$FA,$1F,$98,$F8,$9C,$E5,$FB,$6B,$FC,$67,$A3,$F9,$CF
	.dc.b	$88,$FC,$77,$A3,$71,$FD,$EC,$FE,$83,$E2,$73,$87,$EF,$AF,$F2,$9E
	.dc.b	$97,$E9,$3E,$27,$F2,$DE,$6D,$C7,$F8,$2C,$FE,$A3,$E0,$B3,$87,$F0
	.dc.b	$6B,$FC,$E7,$9F,$FA,$CF,$83,$FC,$F7,$9F,$71,$FE,$1B,$3F,$B0,$F8
	.dc.b	$2C,$E1,$FC,$3A,$FD,$2F,$8E,$BF,$89,$EF,$E9,$8F,$DA,$EF,$E9,$9C
	.dc.b	$C6,$CF,$DC,$E9,$E7,$7F,$77,$AD,$E4,$FC,$8C,$FA,$6B,$F8,$FC,$E5
	.dc.b	$FE,$2F,$E8,$FF,$04,$F8,$3F,$DB,$36,$3F,$92,$FA,$3F,$84,$C7,$F0
	.dc.b	$BD,$AF,$DF,$7F,$DD,$51,$FC,$2A,$F8,$3F,$C3,$3D,$3E,$0F,$F8,$32
	.dc.b	$B8,$7C,$37,$F0,$39,$CC,$FE,$3E,$FF,$DD,$18,$FF,$84,$B5,$7E,$90
	.dc.b	$F8,$AD,$E5,$E1,$F8,$5E,$58,$F6,$8E,$B8,$71,$39,$5B,$D7,$E3,$93
	.dc.b	$E1,$FF,$8B,$CB,$D1,$CF,$C3,$E2,$F7,$5A,$9C,$E1,$D7,$26,$BE,$6F
	.dc.b	$4F,$AF,$C3,$F4,$6D,$6E,$7D,$7E,$47,$8B,$E1,$75,$87,$A7,$66,$3E
	.dc.b	$C8,$FA,$1B,$FB,$1F,$98,$6F,$8F,$D3,$C5,$EC,$E2,$E1,$3F,$5B,$0F
	.dc.b	$96,$F9,$B8,$BE,$13,$DF,$83,$D8,$F8,$78,$BE,$03,$9D,$FD,$15,$F1
	.dc.b	$63,$51,$8F,$27,$E5,$FC,$37,$CA,$F3,$EC,$E6,$BD,$9F,$0D,$97,$DD
	.dc.b	$5F,$43,$D1,$FA,$7F,$0B,$67,$E1,$3F,$53,$D6,$FA,$FF,$0B,$99,$F8
	.dc.b	$7B,$AB,$EC,$FB,$1F,$73,$E0,$ED,$7C,$BF,$84,$FB,$BE,$B6,$76,$F5
	.dc.b	$DF,$E0,$A7,$F0,$3E,$B7,$E1,$3E,$12,$E1,$F4,$DF,$C2,$FA,$3F,$88
	.dc.b	$F8,$4C,$CF,$8A,$7C,$8F,$CF,$D2,$BF,$9F,$D7,$CD,$9F,$CA,$DF,$D6
	.dc.b	$FC,$CE,$0F,$82,$F9,$2B,$F0,$5B,$FA,$9F,$7E,$BE,$A7,$33,$B7,$2F
	.dc.b	$D6,$F4,$FF,$3B,$F0,$BF,$9C,$7F,$3F,$EC,$FE,$7B,$D4,$CD,$6E,$FF
	.dc.b	$21,$3E,$87,$F6,$5A,$DE,$FF,$DF,$3F,$4E,$3F,$2D,$CB,$9D,$FD,$B1
	.dc.b	$FE,$B4,$EA,$72,$FC,$F8,$C3,$F3,$F2,$FC,$5F,$36,$7C,$3F,$F1,$6D
	.dc.b	$F0,$BF,$B1,$5F,$90,$F5,$7F,$6F,$F0,$BF,$B6,$F5,$6C,$E6,$BF,$51
	.dc.b	$F0,$99,$7F,$55,$5F,$BC,$F5,$FF,$7F,$F0,$BF,$BE,$F5,$6E,$1F,$9C
	.dc.b	$9F,$E0,$7F,$0B,$9C,$3F,$65,$7F,$C1,$3D,$7F,$E0,$FF,$0B,$67,$F6
	.dc.b	$CF,$F0,$8F,$53,$F8,$9F,$EF,$DF,$EB,$9E,$05,$F0,$CF,$F8,$07,$E9
	.dc.b	$1F,$C7,$B1,$FE,$D9,$7F,$ED,$07,$F8,$4C,$67,$A1,$39,$E3,$4B,$CD
	.dc.b	$EF,$F0,$DF,$E2,$E8,$E1,$5B,$CD,$C1,$EE,$CB,$EA,$70,$DC,$BE,$7C
	.dc.b	$FC,$0D,$3F,$7B,$39,$7F,$93,$EB,$E9,$BF,$B8,$DB,$FA,$FB,$7C,$0A
	.dc.b	$F8,$95,$FB,$4D,$FC,$9E,$6F,$C4,$F8,$5F,$37,$9D,$F1,$BE,$17,$37
	.dc.b	$FC,$DF,$7F,$F7,$4B,$FE,$FC,$E0,$FE,$68,$E6,$7D,$DE,$7F,$CE,$F8
	.dc.b	$3F,$81,$E9,$5D,$7E,$6C,$FD,$2F,$83,$CE,$1D,$6F,$E3,$7A,$9F,$5B
	.dc.b	$E1,$BE,$47,$A3,$F6,$3E,$1B,$37,$E3,$9C,$BE,$58,$F9,$DE,$9F,$DE
	.dc.b	$F8,$6F,$A1,$E4,$7E,$03,$E1,$B3,$7E,$B9,$CB,$ED,$8F,$4B,$E3,$DB
	.dc.b	$F0,$7B,$FA,$63,$E0,$6F,$E9,$9C,$C6,$CF,$89,$AF,$A6,$3F,$17,$AF
	.dc.b	$A5,$FC,$6C,$66,$7E,$54,$E5,$F9,$9E,$4F,$E5,$BE,$1B,$F1,$1E,$87
	.dc.b	$E6,$3E,$0F,$F1,$53,$9C,$7F,$2F,$39,$7F,$33,$E8,$FE,$7B,$E1,$3F
	.dc.b	$21,$EA,$D9,$CE,$5F,$6F,$E1,$32,$FD,$DA,$FC,$B7,$AF,$FA,$6F,$84
	.dc.b	$FC,$C7,$AD,$70,$FD,$2C,$FE,$AB,$E1,$73,$87,$E1,$6F,$F3,$DE,$AF
	.dc.b	$EB,$BE,$1B,$F4,$1E,$87,$EC,$3E,$17,$37,$F8,$B9,$CB,$F8,$D8,$FD
	.dc.b	$36,$FE,$78,$FC,$76,$F7,$C8,$FF,$5E,$F8,$4A,$FD,$DF,$0F,$AA,$FF
	.dc.b	$B8,$F8,$6E,$2F,$5C,$79,$78,$7F,$7B,$C5,$EB,$8F,$CD,$71,$7A,$E7
	.dc.b	$31,$B3,$F3,$DE,$97,$F0,$8F,$86,$FD,$D7,$A5,$6B,$F8,$3C,$FF,$0C
	.dc.b	$F8,$6C,$DF,$E9,$EF,$D6,$FC,$DA,$FE,$1B,$C1,$EB,$7F,$C5,$1B,$FA
	.dc.b	$E7,$EA,$46,$9E,$07,$98,$F2,$70,$7C,$2D,$3D,$4F,$C4,$D3,$D5,$3F
	.dc.b	$B5,$9C,$D7,$6F,$53,$FC,$67,$FB,$FF,$FB,$86,$BF,$8D,$F9,$B8,$F8
	.dc.b	$7F,$E2,$D5,$AF,$84,$FE,$31,$39,$C3,$FC,$49,$F5,$BF,$EE,$7C,$3F
	.dc.b	$C9,$B8,$FF,$B5,$74,$FF,$78,$CE,$67,$FC,$95,$2F,$87,$FF,$67,$CF
	.dc.b	$FB,$4A,$7F,$C4,$F3,$C6,$FF,$C9,$53,$C6,$7F,$D3,$93,$C6,$7C,$3E
	.dc.b	$BF,$AB,$E1,$F9,$FA,$56,$73,$7F,$E4,$DB,$7C,$3E,$A3,$17,$E2,$F4
	.dc.b	$BE,$1F,$C3,$F9,$3D,$2F,$8B,$F0,$99,$BE,$53,$97,$A4,$7A,$B2,$F9
	.dc.b	$7F,$11,$67,$C1,$BF,$BB,$E2,$73,$5F,$20,$FA,$31,$F0,$B7,$F4,$57
	.dc.b	$F9,$52,$7E,$0E,$F8,$E1,$F9,$75,$F6,$37,$F5,$5F,$C8,$DF,$E5,$FA
	.dc.b	$DF,$73,$E1,$EC,$FC,$1F,$84,$F9,$FE,$C6,$74,$FB,$37,$F8,$29,$FA
	.dc.b	$7E,$C7,$E1,$3E,$12,$E1,$F2,$5F,$AF,$EB,$FE,$23,$E2,$73,$3F,$62
	.dc.b	$7E,$DF,$B1,$F8,$CF,$89,$FB,$BE,$AD,$D3,$F0,$F3,$F9,$0F,$89,$CE
	.dc.b	$1F,$56,$FF,$07,$EB,$7E,$53,$E2,$6C,$FD,$9F,$88,$FC,$3F,$9B,$9C
	.dc.b	$BF,$23,$7F,$9A,$9F,$1D,$7E,$4B,$4A,$FE,$41,$F4,$BF,$83,$D7,$2F
	.dc.b	$E3,$23,$D2,$F8,$EF,$F3,$BA,$FA,$63,$EA,$1F,$E8,$8F,$05,$B7,$AE
	.dc.b	$B7,$3C,$4D,$F4,$29,$E7,$FE,$C3,$E2,$3F,$43,$EB,$7E,$CB,$E1,$7F
	.dc.b	$47,$39,$C7,$F5,$B3,$97,$F5,$FE,$C7,$EE,$3E,$17,$F5,$3E,$BD,$9C
	.dc.b	$E9,$F9,$FF,$85,$CB,$FA,$1A,$FD,$7F,$B1,$FC,$03,$E1,$7F,$63,$EA
	.dc.b	$7F,$AD,$36,$FF,$61,$F0,$7E,$E2,$B5,$97,$EF,$AB,$58,$FE,$E3,$D5
	.dc.b	$E1,$F0,$EA,$FF,$53,$7C,$1F,$FA,$FE,$BF,$D7,$99,$7F,$A6,$B2,$F8
	.dc.b	$58,$77,$F8,$6E,$15,$F1,$AF,$83,$FE,$A8,$BE,$1C,$8E,$1F,$74,$FC
	.dc.b	$C8,$70,$E3,$2F,$E1,$F5,$FC,$73,$87,$E0,$DF,$F0,$CE,$1F,$82,$FF
	.dc.b	$03,$E1,$F8,$2F,$98,$C9,$FC,$49,$CB,$E3,$3C,$F7,$33,$FF,$59,$C7
	.dc.b	$9F,$4C,$ED,$FE,$E1,$9F,$F5,$84,$E6,$3E,$2F,$A9,$F3,$FE,$1B,$73
	.dc.b	$FE,$D2,$AD,$BF,$EF,$6D,$D8,$7F,$26,$D7,$DD,$8F,$F2,$4D,$AD,$FE
	.dc.b	$4E,$7E,$04,$FF,$34,$9C,$E7,$F9,$1E,$3F,$B7,$7F,$27,$8F,$ED,$DF
	.dc.b	$F3,$A9,$CD,$7F,$38,$9C,$D7,$D5,$F2,$FF,$0D,$F0,$9E,$8A,$FF,$5D
	.dc.b	$79,$38,$C7,$F0,$35,$F8,$BF,$85,$D6,$EA,$D6,$3F,$7F,$CB,$FC,$87
	.dc.b	$C2,$E9,$FC,$97,$C2,$F0,$F4,$F3,$9F,$E3,$27,$2F,$E3,$BD,$3F,$CC
	.dc.b	$7C,$37,$6F,$42,$CE,$72,$FC,$8E,$99,$7F,$27,$5E,$4F,$3B,$CD,$5F
	.dc.b	$98,$F1,$7C,$BC,$3F,$76,$73,$7F,$95,$9C,$D7,$DD,$CF,$CB,$3E,$B3
	.dc.b	$F3,$FA,$FA,$5F,$D0,$ED,$ED,$73,$5E,$57,$F3,$D9,$79,$5B,$19,$7F
	.dc.b	$05,$9E,$5F,$C2,$4F,$CC,$F1,$BF,$71,$F0,$D6,$7F,$5E,$FD,$0F,$33
	.dc.b	$F7,$9F,$0D,$99,$FD,$64,$FD,$4F,$3B,$F8,$07,$C2,$7D,$6F,$3F,$F8
	.dc.b	$17,$C0,$E6,$FF,$73,$39,$7F,$77,$1F,$73,$CE,$FE,$11,$F0,$1F,$7B
	.dc.b	$C2,$FE,$15,$F0,$19,$BF,$E0,$33,$97,$F8,$1C,$7E,$13,$CE,$F2,$F1
	.dc.b	$7F,$08,$7D,$6F,$F0,$AF,$23,$35,$FC,$16,$72,$FF,$07,$D7,$D3,$5F
	.dc.b	$C4,$35,$F4,$BF,$C4,$A3,$33,$FC,$36,$72,$FF,$0F,$CF,$D3,$5F,$C6
	.dc.b	$33,$F4,$BF,$C6,$A3,$33,$FC,$56,$72,$FF,$17,$F2,$7F,$94,$7C,$37
	.dc.b	$E7,$9B,$7F,$C8,$BC,$AF,$E5,$9F,$0D,$9B,$FE,$47,$7F,$A9,$FD,$1F
	.dc.b	$FD,$91,$8B,$E1,$71,$5F,$CC,$FE,$1B,$5E,$8F,$F3,$59,$CC,$FF,$2C
	.dc.b	$9E,$5E,$6F,$F3,$9F,$85,$B3,$FC,$C3,$E1,$3B,$7A,$19,$DD,$3F,$D5
	.dc.b	$34,$FE,$1A,$EE,$33,$E2,$FF,$1E,$97,$99,$EA,$9C,$9A,$8C,$D7,$B2
	.dc.b	$32,$A9,$2F,$2F,$94,$F8,$6F,$E3,$BE,$33,$DD,$7C,$AF,$8C,$7C,$38
	.dc.b	$C5,$FF,$BA,$E3,$FA,$6E,$BF,$11,$FE,$91,$AF,$C2,$7F,$A5,$69,$5F
	.dc.b	$E5,$9A,$7C,$47,$FA,$AC,$65,$F8,$BF,$12,$E0,$57,$D7,$F1,$3D,$3A
	.dc.b	$7C,$57,$D2,$FC,$7F,$17,$3A,$FC,$59,$CB,$F1,$F1,$F6,$9E,$F3,$ED
	.dc.b	$3E,$CE,$2C,$DF,$D1,$9F,$80,$7D,$7E,$2F,$B4,$D3,$EC,$1F,$6F,$8B
	.dc.b	$33,$F6,$67,$FB,$E6,$5F,$8F,$F1,$3F,$03,$70,$FA,$DE,$4B,$BB,$51
	.dc.b	$93,$3D,$FF,$97,$F1,$B1,$1C,$27,$6F,$3C,$96,$BF,$E8,$DF,$1C,$F0
	.dc.b	$38,$D6,$BA,$B3,$18,$B7,$5F,$0F,$07,$15,$AC,$BC,$98,$76,$AE,$BD
	.dc.b	$DC,$7A,$AD,$6B,$17,$57,$8B,$BB,$5C,$DF,$8D,$D9,$8F,$86,$BF,$69
	.dc.b	$E2,$7B,$BA,$DF,$2B,$FE,$1D,$C3,$F2,$3E,$36,$61,$F0,$7C,$35,$F9
	.dc.b	$17,$C2,$FF,$7D,$7B,$AF,$E0,$1E,$3F,$1F,$82,$B8,$70,$C5,$1F,$C1
	.dc.b	$7C,$8E,$DF,$07,$AA,$FB,$19,$F9,$2B,$F8,$57,$47,$8F,$E1,$BF,$86
	.dc.b	$75,$5D,$BF,$4F,$3E,$7F,$86,$CE,$3D,$EF,$F8,$97,$4F,$AE,$FF,$8A
	.dc.b	$74,$79,$DF,$D7,$C6,$5F,$D8,$C6,$6B,$F8,$56,$FE,$B9,$FD,$AB,$63
	.dc.b	$F8,$76,$DE,$A9,$CC,$7B,$F4,$F4,$BE,$38,$FD,$EE,$9E,$98,$FE,$2D
	.dc.b	$A7,$A6,$73,$1B,$3F,$8D,$74,$7D,$3F,$84,$FE,$59,$CB,$6B,$F8,$3C
	.dc.b	$FD,$7F,$85,$CD,$FC,$FB,$FE,$65,$E3,$FD,$BF,$85,$FE,$69,$E1,$7D
	.dc.b	$DF,$85,$CD,$FD,$59,$CB,$F5,$E3,$FC,$45,$F2,$1F,$ED,$5E,$6F,$C3
	.dc.b	$FC,$1E,$35,$FC,$37,$C2,$6B,$AF,$F1,$13,$9C,$7F,$01,$3C,$B8,$FD
	.dc.b	$11,$FE,$88,$C5,$D5,$8F,$0B,$FC,$A3,$1F,$86,$9F,$C4,$70,$F8,$AB
	.dc.b	$59,$FF,$35,$E0,$F6,$3F,$CA,$38,$3D,$8F,$F2,$A8,$CD,$FF,$43,$9C
	.dc.b	$BE,$19,$F3,$9F,$CC,$F0,$F7,$76,$34,$F6,$57,$E6,$75,$F6,$3F,$96
	.dc.b	$9D,$5F,$1E,$BF,$E0,$3F,$22,$7E,$49,$8E,$DF,$96,$69,$CD,$FF,$00
	.dc.b	$94,$FF,$0C,$F7,$39,$F3,$5A,$D3,$FC,$43,$DC,$EF,$C3,$99,$FE,$2D
	.dc.b	$82,$BF,$8A,$FB,$1E,$4E,$8F,$E2,$FE,$E5,$D7,$F8,$34,$FA,$3A,$73
	.dc.b	$8F,$8E,$FF,$03,$87,$B3,$82,$CF,$9F,$87,$F0,$BA,$67,$2F,$95,$7F
	.dc.b	$0A,$7F,$CD,$52,$F7,$BE,$2F,$C3,$63,$DF,$F8,$FF,$0F,$A9,$CE,$3F
	.dc.b	$D9,$6D,$5A,$8F,$E3,$9F,$15,$F3,$BE,$23,$F2,$BB,$59,$CE,$BF,$2F
	.dc.b	$E2,$32,$FC,$DA,$FC,$DF,$07,$D6,$F8,$5B,$3F,$47,$E1,$7F,$3F,$C5
	.dc.b	$9C,$7F,$09,$7F,$72,$7D,$7C,$1F,$D1,$AD,$F3,$3F,$CC,$AF,$D1,$C3
	.dc.b	$F8,$2E,$1D,$D7,$F4,$CA,$D3,$FE,$E8,$E1,$56,$EF,$F2,$5C,$5F,$8A
	.dc.b	$E2,$FD,$8E,$56,$73,$87,$E1,$FE,$27,$2F,$E2,$6B,$F6,$F8,$7F,$52
	.dc.b	$F9,$37,$27,$EE,$B8,$B0,$EA,$FF,$A7,$CB,$E2,$75,$18,$9F,$B3,$CD
	.dc.b	$EF,$DF,$D3,$E1,$FF,$B1,$B8,$3F,$1D,$7F,$99,$E1,$FB,$F1,$BF,$23
	.dc.b	$8C,$E7,$83,$F5,$DC,$FE,$5F,$8B,$FE,$1B,$D1,$E7,$EE,$FE,$1F,$39
	.dc.b	$AF,$DA,$4E,$5F,$DB,$73,$FB,$7B,$BF,$8A,$E9,$67,$39,$7B,$3B,$32
	.dc.b	$FB,$AB,$F2,$73,$E5,$3F,$1D,$E3,$FD,$19,$FA,$91,$97,$EB,$46,$77
	.dc.b	$FE,$2B,$CF,$F3,$7B,$BF,$92,$BF,$3F,$E2,$BF,$93,$F5,$67,$1F,$E1
	.dc.b	$93,$97,$FC,$81,$FA,$37,$D4,$CB,$E8,$FA,$FF,$EC,$F8,$C4,$7F,$83
	.dc.b	$FA,$9F,$17,$9E,$BF,$B9,$7B,$EB,$FB,$D3,$FE,$D7,$F9,$57,$A9,$E9
	.dc.b	$E0,$FF,$7D,$C6,$32,$FA,$47,$F9,$67,$4F,$F3,$1C,$7F,$85,$23,$F2
	.dc.b	$DD,$5C,$EF,$E0,$57,$F9,$A3,$5F,$F7,$CD,$7F,$9A,$B5,$7C,$67,$A7
	.dc.b	$AF,$F2,$1F,$19,$EB,$EC,$FC,$94,$E6,$FF,$1F,$3E,$FE,$7F,$CB,$7C
	.dc.b	$65,$9F,$C9,$FC,$67,$C3,$EA,$CE,$BF,$22,$FF,$39,$3E,$FF,$6B,$DD
	.dc.b	$3E,$A7,$CF,$7F,$E7,$5E,$17,$B5,$A8,$C4,$EB,$8D,$C2,$EB,$FD,$47
	.dc.b	$C6,$7D,$36,$E5,$F6,$3B,$39,$5F,$09,$CE,$3F,$2C,$FF,$42,$61,$FE
	.dc.b	$46,$7F,$5B,$F1,$9F,$77,$A3,$F6,$9F,$13,$F7,$FA,$FF,$6D,$39,$AF
	.dc.b	$D9,$CF,$DB,$F8,$BE,$FC,$1F,$E8,$D9,$FB,$39,$FD,$C9,$CF,$C6,$7A
	.dc.b	$F3,$FA,$75,$F6,$B6,$FC,$7F,$37,$9B,$04,$7D,$BE,$7F,$5D,$AB,$53
	.dc.b	$F7,$F9,$F3,$CB,$FC,$07,$D6,$FC,$B4,$63,$2F,$E3,$91,$D1,$F8,$59
	.dc.b	$3E,$5F,$FC,$A3,$1E,$33,$2B,$E0,$FE,$93,$A8,$FF,$3A,$79,$5F,$C3
	.dc.b	$B2,$FF,$5B,$CE,$E7,$E0,$C7,$E4,$F9,$FD,$3C,$7E,$89,$FA,$31,$87
	.dc.b	$F8,$EB,$FA,$DE,$3F,$7F,$C2,$FA,$EF,$F6,$71,$8A,$F5,$C7,$FA,$D7
	.dc.b	$05,$FE,$01,$6D,$E8,$8C,$3F,$47,$E1,$7E,$5F,$C2,$78,$F6,$FC,$6C
	.dc.b	$63,$1F,$1E,$3F,$C0,$B6,$FA,$27,$8E,$3E,$4F,$47,$F4,$D7,$FA,$6F
	.dc.b	$87,$FC,$16,$75,$87,$CF,$3F,$82,$FC,$57,$E5,$2F,$FD,$91,$7F,$E0
	.dc.b	$2F,$2B,$58,$FF,$B8,$27,$FD,$C7,$3F,$3B,$1C,$BF,$ED,$B8,$E2,$FF
	.dc.b	$AA,$B0,$7A,$FF,$DE,$6F,$FD,$B1,$4F,$8A,$F0,$75,$FB,$E3,$18,$7D
	.dc.b	$FC,$3C,$D1,$FE,$EA,$8F,$E3,$9C,$1E,$09,$FE,$3B,$D7,$E9,$3D,$B8
	.dc.b	$F9,$FE,$2F,$17,$F5,$0F,$8B,$F1,$9E,$FF,$8B,$FD,$D6,$48,$C4,$FE
	.dc.b	$18,$F6,$7C,$5F,$C5,$3C,$DE,$25,$8F,$A6,$7D,$0F,$17,$E5,$7C,$4F
	.dc.b	$8A,$3E,$24,$61,$F1,$6B,$EC,$F8,$9F,$A1,$F1,$1E,$26,$CF,$86,$7F
	.dc.b	$9C,$FC,$4F,$D5,$F3,$FE,$33,$67,$C3,$3F,$1B,$E2,$BE,$D7,$C5,$78
	.dc.b	$63,$FA,$0C,$62,$7C,$3C,$5F,$6F,$A3,$F0,$1E,$27,$C2,$6C,$F8,$78
	.dc.b	$BF,$DF,$AF,$D3,$A2,$C1,$FA,$27,$EE,$63,$EA,$CF,$DE,$F8,$8F,$E3
	.dc.b	$F3,$99,$FC,$27,$A7,$FA,$4A,$E7,$7F,$93,$AF,$F6,$3D,$77,$3B,$38
	.dc.b	$7F,$7B,$97,$E0,$63,$FC,$FD,$1F,$93,$AE,$67,$48,$E1,$1F,$EF,$BF
	.dc.b	$1D,$5C,$7E,$27,$CA,$7F,$3B,$AF,$F3,$85,$78,$70,$F6,$CE,$7E,$1B
	.dc.b	$D7,$EF,$FE,$D6,$FF,$CD,$F7,$B7,$FC,$36,$FC,$23,$F8,$DA,$F7,$FE
	.dc.b	$19,$F5,$57,$C3,$62,$7E,$A9,$F0,$3E,$17,$E4,$9E,$5F,$86,$B1,$F8
	.dc.b	$5A,$FF,$47,$3E,$1F,$85,$F2,$7C,$37,$D0,$7C,$DF,$0D,$F4,$BE,$1B
	.dc.b	$C0,$DA,$EF,$BF,$DA,$F8,$6F,$AF,$F0,$DF,$0D,$B3,$DF,$E1,$BE,$DF
	.dc.b	$99,$F0,$5B,$1F,$92,$E6,$FA,$1F,$09,$F8,$03,$DD,$F0,$B6,$7B,$FE
	.dc.b	$87,$FE,$16,$7E,$A7,$6F,$EE,$E3,$FC,$95,$1F,$A1,$74,$FE,$40,$FE
	.dc.b	$37,$7F,$79,$FC,$94,$70,$7F,$8C,$46,$74,$4F,$F8,$42,$7F,$33,$D5
	.dc.b	$FD,$15,$E3,$F0,$9F,$CB,$E5,$F0,$9F,$DD,$9E,$16,$ED,$FC,$C5,$FF
	.dc.b	$56,$97,$C1,$EC,$FD,$83,$88,$FC,$52,$F8,$4C,$FC,$1F,$5F,$84,$FF
	.dc.b	$05,$49,$F0,$46,$3E,$13,$C5,$E2,$79,$4F,$23,$E7,$F8,$3F,$31,$E0
	.dc.b	$3D,$1F,$05,$E2,$F8,$2F,$51,$D8,$F6,$7B,$DE,$0F,$5F,$DC,$7C,$13
	.dc.b	$E0,$7C,$17,$C3,$F8,$1F,$85,$F0,$5E,$C3,$E2,$3E,$EF,$80,$F8,$DF
	.dc.b	$03,$EA,$3E,$43,$EC,$F8,$1F,$94,$7C,$D3,$FC,$43,$FA,$1F,$FA,$87
	.dc.b	$9F,$E6,$65,$FC,$EE,$3F,$A4,$F2,$FD,$FE,$AF,$AD,$49,$FE,$A3,$3F
	.dc.b	$BB,$F3,$F8,$7C,$7F,$4F,$83,$FE,$41,$5F,$D2,$AB,$F8,$21,$FE,$0C
	.dc.b	$7F,$A5,$3F,$D5,$7E,$0F,$FA,$A3,$9F,$82,$F2,$FC,$1F,$F8,$F6,$BF
	.dc.b	$C5,$13,$CB,$2F,$ED,$11,$88,$FE,$37,$A7,$A7,$E0,$FD,$BF,$07,$FD
	.dc.b	$E6,$BF,$A6,$57,$F0,$B6,$BF,$C6,$BC,$3F,$CB,$46,$23,$F4,$71,$8F
	.dc.b	$83,$F5,$FC,$17,$C8,$3C,$87,$CA,$F8,$2F,$37,$C0,$FC,$C3,$C0,$7C
	.dc.b	$EF,$82,$F1,$7C,$0F,$D0,$3E,$29,$F4,$BE,$0B,$E3,$FC,$0F,$D4,$3E
	.dc.b	$01,$F5,$BE,$0B,$E1,$7C,$0F,$D8,$3D,$87,$DA,$F8,$2F,$77,$9F,$F7
	.dc.b	$3A,$BE,$89,$F7,$9F,$A7,$F0,$5F,$80,$3E,$61,$FE,$62,$F7,$7F,$A2
	.dc.b	$71,$7F,$42,$EA,$FA,$7D,$1F,$E1,$59,$FE,$73,$3F,$83,$BF,$F0,$04
	.dc.b	$7F,$40,$8F,$A3,$C9,$FB,$F8,$C4,$6E,$31,$F1,$3F,$E2,$B9,$7C,$1F
	.dc.b	$87,$E2,$BF,$B3,$47,$F4,$0B,$FB,$D7,$FD,$52,$37,$1F,$C9,$9F,$E7
	.dc.b	$EE,$8F,$EC,$4F,$F4,$07,$F1,$9E,$C7,$9B,$E0,$FD,$DF,$09,$FE,$87
	.dc.b	$AF,$E8,$55,$FD,$35,$AF,$F4,$7D,$3E,$07,$C2,$7C,$5F,$84,$F4,$57
	.dc.b	$C7,$3C,$E7,$C9,$F8,$4F,$EF,$DF,$40,$E0,$FF,$46,$B7,$1F,$F0,$53
	.dc.b	$5F,$46,$28,$72,$79,$7E,$13,$EA,$1F,$1C,$FA,$DF,$09,$F2,$7E,$0F
	.dc.b	$EC,$7C,$27,$C8,$3E,$D3,$E2,$EE,$F0,$F7,$7F,$A9,$71,$3E,$48,$F8
	.dc.b	$5F,$07,$F8,$13,$EA,$1F,$83,$F8,$4F,$AD,$F0,$9F,$E9,$CE,$7F,$B9
	.dc.b	$CF,$FE,$A0,$6F,$07,$FA,$91,$FC,$04,$E6,$3F,$D1,$39,$3C,$7F,$87
	.dc.b	$F0,$DE,$33,$E2,$BF,$4F,$E1,$7F,$24,$7E,$14,$FC,$A7,$C2,$FF,$9C
	.dc.b	$7F,$42,$FE,$A3,$87,$F8,$7F,$E1,$1E,$DF,$F9,$46,$73,$CF,$FD,$17
	.dc.b	$DB,$C7,$C5,$FF,$B4,$30,$FF,$4D,$D7,$F5,$57,$FC,$43,$3F,$E5,$9D
	.dc.b	$7F,$CE,$DE,$B7,$F1,$E7,$F9,$27,$8B,$FA,$2E,$1F,$F1,$2C,$6B,$B3
	.dc.b	$87,$AB,$FE,$D1,$CB,$FA,$AE,$DF,$D5,$35,$FF,$3E,$46,$3D,$3F,$37
	.dc.b	$C3,$FF,$B4,$23,$FA,$EE,$DF,$DD,$CF,$F1,$D3,$FD,$5D,$FE,$DC,$7B
	.dc.b	$FE,$03,$DD,$F0,$D8,$8F,$F7,$C7,$C1,$61,$C6,$3F,$D9,$6B,$E2,$7C
	.dc.b	$36,$E7,$E4,$57,$F9,$DF,$A1,$F0,$3F,$EE,$26,$CF,$F9,$7D,$AF,$E5
	.dc.b	$97,$C0,$7B,$FE,$13,$A6,$5F,$DC,$FE,$13,$E9,$7C,$2E,$38,$7F,$DA
	.dc.b	$5C,$BF,$E0,$1E,$0F,$ED,$99,$F9,$A5,$7F,$67,$1E,$1F,$0B,$F6,$F8
	.dc.b	$39,$70,$DC,$38,$35,$F8,$BA,$7F,$80,$F8,$7F,$05,$C3,$C3,$5F,$F0
	.dc.b	$87,$02,$3D,$DA,$7A,$39,$FE,$E6,$3F,$E6,$9A,$FF,$77,$BE,$FA,$FB
	.dc.b	$95,$EB,$D7,$F1,$CF,$B7,$82,$D7,$AE,$7F,$25,$B6,$72,$F8,$1D,$7F
	.dc.b	$0F,$AF,$F2,$CD,$8C,$DF,$C3,$F8,$CF,$CD,$76,$66,$3F,$DE,$3A,$FE
	.dc.b	$19,$FA,$5F,$11,$E9,$F0,$BE,$D3,$F6,$78,$FF,$25,$D5,$F4,$7A,$BF
	.dc.b	$49,$D5,$F4,$EA,$E1,$F4,$67,$F5,$15,$9D,$3E,$B7,$57,$C5,$F0,$FF
	.dc.b	$59,$F0,$FF,$1F,$C4,$B8,$FE,$3B,$A1,$5F,$B1,$BD,$61,$FE,$3F,$FA
	.dc.b	$36,$FF,$91,$C3,$FC,$8E,$7F,$A5,$E7,$FD,$1F,$2F,$13,$D9,$1C,$5E
	.dc.b	$1E,$1F,$F8,$42,$79,$D7,$FA,$CF,$C0,$7C,$3F,$DB,$BF,$E0,$4D,$C3
	.dc.b	$F2,$0F,$DF,$F1,$7F,$83,$7C,$46,$70,$FE,$07,$F1,$1F,$83,$F4,$7F
	.dc.b	$85,$3F,$85,$F4,$2C,$FB,$E7,$FD,$BD,$FA,$0F,$F0,$49,$FF,$70,$6A
	.dc.b	$F4,$7F,$18,$63,$D0,$FC,$71,$70,$F9,$11,$C0,$CE,$BF,$1F,$CE,$FC
	.dc.b	$A7,$A3,$D1,$B1,$9A,$FA,$11,$E0,$9C,$C7,$CB,$3F,$D8,$59,$7F,$8E
	.dc.b	$1F,$F6,$27,$33,$D8,$F4,$F9,$BF,$A1,$F5,$ED,$7D,$B9,$F6,$FC,$4E
	.dc.b	$73,$FB,$5E,$CF,$E9,$CF,$79,$5C,$D7,$E0,$FD,$9F,$86,$65,$F2,$79
	.dc.b	$EE,$1F,$F5,$6C,$7E,$27,$D9,$F9,$27,$EC,$CA,$FE,$37,$D9,$F9,$A6
	.dc.b	$70,$FF,$45,$6F,$FE,$4B,$9F,$F4,$97,$C9,$3E,$43,$E9,$56,$BE,$2F
	.dc.b	$F9,$04,$E7,$9B,$EC,$C6,$2B,$77,$E6,$C3,$A3,$FE,$EC,$BE,$18,$78
	.dc.b	$3E,$2B,$87,$2F,$E0,$9E,$5C,$D6,$FF,$9A,$5F,$8E,$BF,$0F,$ED,$6B
	.dc.b	$0F,$4F,$99,$E1,$F1,$7F,$16,$F8,$FC,$6B,$8F,$2D,$7F,$5B,$7F,$03
	.dc.b	$E2,$74,$FC,$2F,$89,$F5,$70,$FE,$53,$E2,$6C,$FC,$67,$DD,$DD,$F9
	.dc.b	$8A,$CE,$3E,$FE,$1F,$5D,$FB,$63,$C1,$C3,$EB,$7F,$A8,$F0,$FA,$E7
	.dc.b	$E8,$F0,$7A,$E7,$E4,$6F,$EB,$9F,$6C,$7A,$77,$F5,$BC,$B7,$F5,$BF
	.dc.b	$3F,$7F,$5B,$F4,$7E,$27,$CB,$D0,$C1,$3F,$9C,$F7,$F3,$18,$97,$2A
	.dc.b	$7D,$F3,$F8,$6B,$DD,$FE,$93,$7F,$7B,$F0,$4F,$7C,$FC,$17,$CD,$CB
	.dc.b	$F6,$1C,$9E,$E3,$EA,$72,$FB,$1F,$C8,$E8,$D7,$11,$F8,$98,$F7,$E9
	.dc.b	$EB,$8F,$DA,$EB,$F0,$23,$07,$AB,$E0,$3F,$B5,$FB,$AF,$89,$CE,$9E
	.dc.b	$FF,$A7,$74,$6A,$73,$93,$15,$6A,$3F,$C0,$78,$3E,$2B,$F8,$CE,$5C
	.dc.b	$72,$FF,$7A,$5F,$C9,$BD,$FC,$57,$4C,$79,$F1,$F9,$3E,$2B,$17,$BB
	.dc.b	$FE,$E3,$9B,$0C,$E5,$BC,$BA,$E9,$EC,$F8,$6E,$2F,$BB,$E1,$6E,$3E
	.dc.b	$69,$EB,$B6,$6F,$F7,$F7,$FB,$7F,$4F,$F9,$8C,$FE,$E7,$D1,$B8,$7E
	.dc.b	$9E,$7F,$9A,$D6,$6F,$F8,$3E,$7F,$BF,$F4,$3F,$9C,$EF,$FC,$13,$F8
	.dc.b	$01,$EF,$E3,$2F,$D8,$57,$F3,$EB,$D4,$FF,$40,$D7,$87,$BD,$FD,$0B
	.dc.b	$7B,$9F,$F1,$47,$A7,$B5,$BC,$BF,$81,$E6,$BF,$E9,$3F,$17,$E5,$9F
	.dc.b	$E2,$1F,$17,$79,$FF,$07,$E0,$23,$E0,$67,$FC,$82,$F4,$FE,$86,$7E
	.dc.b	$85,$FA,$23,$F9,$2F,$07,$A1,$B1,$93,$F8,$F6,$F9,$7F,$90,$6D,$EA
	.dc.b	$7F,$91,$6D,$73,$F1,$BE,$CB,$FE,$63,$59,$7F,$65,$FA,$17,$FB,$22
	.dc.b	$DE,$23,$77,$BD,$F8,$BA,$D4,$E6,$31,$7F,$4F,$E0,$35,$F1,$9F,$56
	.dc.b	$78,$7C,$AF,$D7,$F8,$AB,$9F,$CC,$6B,$FC,$5B,$37,$C3,$F7,$F8,$DE
	.dc.b	$19,$FF,$07,$3C,$FE,$BF,$E0,$7E,$27,$C8,$FE,$0F,$C1,$B8,$FD,$89
	.dc.b	$F4,$7B,$39,$BE,$53,$FA,$FE,$1F,$F6,$73,$FD,$1F,$3F,$7C,$EB,$91
	.dc.b	$5F,$EC,$CA,$FE,$AF,$5F,$E5,$3A,$D4,$E6,$79,$6F,$F1,$FD,$8F,$CB
	.dc.b	$7C,$5F,$BA,$78,$F0,$57,$C0,$D7,$E8,$E3,$F0,$B1,$EB,$BF,$C2,$9F
	.dc.b	$EE,$79,$7F,$BB,$5F,$1E,$BF,$01,$F9,$5B,$57,$37,$E6,$DB,$E1,$47
	.dc.b	$2F,$5F,$F5,$5F,$17,$F7,$7D,$8F,$D6,$7C,$5D,$AF,$D0,$CF,$E0,$6B
	.dc.b	$38,$7C,$0F,$80,$FD,$97,$C5,$FE,$17,$E0,$7F,$69,$F1,$77,$1F,$D4
	.dc.b	$CF,$E2,$6B,$37,$F2,$3D,$BF,$DD,$7C,$57,$E3,$7D,$DB,$3F,$31,$FD
	.dc.b	$F7,$C5,$FE,$46,$B3,$8F,$C2,$F6,$3C,$6F,$C4,$F7,$6F,$67,$F3,$4E
	.dc.b	$64,$7B,$B2,$FA,$F7,$A7,$F4,$3A,$7A,$9F,$B7,$BF,$9B,$1F,$9E,$E5
	.dc.b	$FA,$33,$9C,$7B,$65,$E4,$7C,$1A,$5C,$3F,$5A,$FA,$9F,$81,$7F,$C3
	.dc.b	$35,$FD,$67,$B3,$E3,$8F,$CA,$7B,$3F,$C7,$7E,$2B,$F6,$0E,$71,$FC
	.dc.b	$9D,$67,$0F,$A5,$7F,$39,$F4,$BF,$C6,$38,$BE,$93,$FA,$DE,$2F,$A4
	.dc.b	$E6,$7C,$6F,$EC,$72,$FA,$4F,$96,$3E,$93,$FC,$8B,$8B,$E9,$39,$3F
	.dc.b	$27,$E9,$7F,$33,$F8,$BF,$E0,$CD,$AF,$D2,$BF,$CD,$FE,$27,$F8,$57
	.dc.b	$AB,$9B,$FC,$DF,$A9,$FC,$EF,$E2,$7F,$87,$37,$0F,$D6,$BF,$CF,$FE
	.dc.b	$27,$FA,$E7,$B9,$FD,$02,$FF,$9A,$DE,$3D,$5F,$F2,$3F,$F0,$0F,$F8
	.dc.b	$43,$1F,$DD,$5F,$F3,$9B,$FF,$1E,$CB,$E2,$78,$46,$0E,$BE,$C7,$28
	.dc.b	$EF,$E5,$F9,$F9,$BF,$BF,$E1,$FE,$FD,$BF,$F2,$CE,$0F,$89,$F2,$FA
	.dc.b	$3B,$CF,$FC,$3D,$7F,$EC,$1B,$DD,$7B,$9F,$4C,$FB,$9A,$7C,$43,$FA
	.dc.b	$8E,$DF,$03,$DA,$F8,$91,$FE,$74,$BF,$63,$FE,$3E,$C7,$E0,$39,$DB
	.dc.b	$D5,$87,$C2,$7E,$2C,$7B,$8F,$87,$AF,$CC,$F3,$3E,$4F,$C8,$7C,$EF
	.dc.b	$46,$E1,$F2,$9F,$9B,$F1,$1B,$C3,$E5,$D7,$D4,$F3,$7E,$8F,$C2,$7D
	.dc.b	$6F,$3B,$E9,$FC,$3D,$BF,$A3,$3B,$7E,$9C,$7D,$CF,$4B,$EC,$FC,$3D
	.dc.b	$9F,$AE,$D7,$DD,$F0,$BF,$82,$F3,$F7,$B7,$DE,$F4,$3F,$03,$F1,$1F
	.dc.b	$86,$6C,$66,$FC,$EF,$E1,$3C,$5C,$C7,$CB,$DB,$D3,$3F,$88,$DB,$D2
	.dc.b	$D9,$F4,$BF,$8C,$9C,$C7,$D3,$D3,$C6,$D3,$D3,$3F,$92,$8F,$CA,$E7
	.dc.b	$E9,$9C,$9F,$6F,$C6,$FC,$CF,$3F,$E7,$9F,$CD,$FC,$45,$9F,$CC,$CF
	.dc.b	$E8,$BC,$7C,$E3,$F9,$EF,$1F,$F4,$3F,$0F,$FA,$66,$C6,$6F,$F4,$3F
	.dc.b	$0D,$FA,$AF,$33,$31,$F8,$BF,$2B,$F5,$3F,$0B,$FA,$EF,$2A,$E1,$FA
	.dc.b	$A7,$F5,$BF,$0D,$FB,$2A,$CD,$FE,$4F,$C9,$F1,$BF,$95,$D7,$F6,$7F
	.dc.b	$05,$E7,$8F,$D8,$F0,$7E,$EA,$B3,$97,$96,$7F,$35,$BF,$A5,$FD,$BB
	.dc.b	$5F,$CF,$6D,$E9,$73,$3F,$BE,$CB,$C7,$E4,$7E,$DB,$5B,$3F,$B2,$7D
	.dc.b	$2E,$EB,$F8,$1E,$DF,$C3,$3C,$6F,$1C,$7A,$CF,$E0,$FA,$F9,$DF,$4D
	.dc.b	$7F,$0C,$9D,$E1,$FC,$13,$87,$E5,$B6,$BC,$A7,$ED,$E3,$F8,$AE,$7E
	.dc.b	$98,$F7,$DF,$8D,$FD,$D6,$3F,$45,$F2,$C7,$D1,$7F,$8C,$F0,$7D,$17
	.dc.b	$27,$FB,$DB,$F4,$1F,$F5,$7B,$FC,$7B,$E1,$3F,$D0,$51,$88,$FE,$0D
	.dc.b	$18,$7F,$93,$FC,$1E,$BC,$6F,$E5,$33,$C3,$CA,$FE,$57,$F0,$96,$FF
	.dc.b	$94,$B5,$FE,$21,$F0,$DD,$BC,$6D,$E9,$FE,$29,$CD,$E3,$FF,$34,$F8
	.dc.b	$7F,$1B,$FE,$FC,$F5,$3F,$D8,$D5,$B8,$D3,$C7,$C9,$D3,$E5,$CB,$D8
	.dc.b	$F5,$DB,$97,$C2,$E6,$7D,$B3,$E1,$F2,$B2,$F8,$F7,$F2,$CF,$C7,$6F
	.dc.b	$37,$FA,$EB,$5F,$E7,$53,$FD,$1B,$7F,$63,$F0,$2F,$CC,$FF,$4A,$DB
	.dc.b	$D6,$FC,$D3,$FC,$D3,$B7,$B1,$FA,$9B,$7B,$1F,$AD,$B7,$B2,$FE,$37
	.dc.b	$95,$F0,$FE,$17,$EB,$BF,$17,$E0,$AD,$F5,$F8,$66,$4B,$FB,$BE,$36
	.dc.b	$F1,$FA,$1E,$67,$CD,$F8,$6F,$C0,$BF,$3F,$E1,$2E,$5C,$F1,$57,$E1
	.dc.b	$BC,$8D,$61,$F6,$7C,$6F,$AD,$F0,$DF,$8A,$7E,$C7,$C5,$DC,$79,$4F
	.dc.b	$E3,$BC,$7C,$DF,$E0,$75,$F2,$57,$E5,$78,$6F,$3F,$CF,$7E,$05,$6E
	.dc.b	$7C,$9C,$5E,$67,$F3,$1C,$5F,$03,$6C,$47,$E7,$A3,$E1,$F0,$62,$07
	.dc.b	$D2,$9F,$63,$EF,$8F,$43,$D3,$81,$AE,$23,$A4,$36,$C4,$74,$C3,$E8
	.dc.b	$D7,$BE,$3E,$54,$FE,$BB,$C2,$FC,$E7,$C2,$D9,$F9,$CD,$7C,$1F,$0D
	.dc.b	$FB,$4F,$13,$7B,$7E,$9F,$C4,$FD,$27,$C1,$7E,$E1,$FD,$37,$C1,$DC
	.dc.b	$3E,$C4,$FE,$F3,$C9,$CD,$FE,$BF,$C6,$FD,$67,$C1,$FF,$00,$F2,$2C
	.dc.b	$FE,$CD,$FD,$87,$C1,$FF,$04,$AC,$E3,$FC,$27,$4F,$1B,$FC,$2F,$7F
	.dc.b	$DB,$7C,$27,$A6,$3E,$2F,$07,$A5,$FC,$54,$79,$5F,$99,$AD,$7E,$5E
	.dc.b	$D5,$F9,$5C,$89,$F5,$47,$E4,$E7,$C8,$FE,$57,$6F,$53,$76,$F5,$3F
	.dc.b	$99,$D7,$6F,$DB,$9F,$23,$F7,$71,$AF,$DC,$F8,$5A,$F8,$A3,$F9,$2F
	.dc.b	$97,$FC,$4E,$7C,$D3,$FB,$AC,$AF,$37,$EC,$33,$BA,$7A,$74,$FE,$3D
	.dc.b	$C5,$F4,$B0,$FE,$65,$C3,$F3,$3A,$FF,$86,$71,$DB,$FC,$AC,$3C,$64
	.dc.b	$FD,$49,$FB,$57,$EF,$3F,$9E,$78,$DF,$C9,$F2,$FE,$76,$FF,$29,$F8
	.dc.b	$5B,$3F,$BB,$9F,$E7,$FE,$3E,$6F,$FD,$B5,$FA,$07,$FB,$E2,$5F,$0C
	.dc.b	$9C,$9F,$CC,$FE,$1B,$18,$35,$A3,$0E,$57,$8B,$F5,$C3,$CD,$E5,$1E
	.dc.b	$0F,$27,$F9,$F7,$C3,$6F,$1E,$96,$F1,$33,$7E,$67,$F9,$54,$B9,$5F
	.dc.b	$C0,$2F,$84,$E5,$AF,$8F,$3E,$37,$FD,$87,$83,$E0,$3F,$20,$FE,$51
	.dc.b	$6F,$83,$F8,$F1,$FC,$07,$8B,$E3,$BA,$8E,$BA,$FB,$B5,$F8,$B3,$FC
	.dc.b	$9F,$87,$31,$FD,$67,$87,$7C,$3F,$02,$BE,$44,$FC,$4D,$3C,$BA,$FC
	.dc.b	$AC,$7E,$05,$7F,$31,$E1,$EF,$C3,$F0,$38,$7D,$FC,$3E,$2C,$78,$E5
	.dc.b	$F3,$BC,$7D,$70,$BE,$1F,$CF,$E4,$63,$E1,$EE,$BF,$02,$7D,$77,$9C
	.dc.b	$3F,$07,$E4,$F2,$F8,$7F,$7F,$95,$67,$F0,$EF,$6F,$86,$F8,$75,$9C
	.dc.b	$7D,$A7,$DA,$F1,$FC,$7F,$0D,$F2,$1F,$2F,$C1,$59,$FA,$33,$F3,$3C
	.dc.b	$DC,$E5,$F9,$4D,$3C,$95,$D3,$9F,$F1,$D1,$87,$E4,$3F,$03,$6F,$AB
	.dc.b	$E2,$E6,$FF,$23,$A7,$96,$FE,$76,$D6,$7F,$2A,$FA,$DF,$83,$8F,$E1
	.dc.b	$37,$F3,$3F,$D3,$B8,$9A,$E2,$3F,$A7,$43,$6C,$47,$F4,$38,$DD,$FE
	.dc.b	$BB,$6F,$6C,$FA,$FA,$1E,$2F,$D8,$F8,$3F,$C6,$B6,$FF,$6A,$FD,$CF
	.dc.b	$83,$FC,$8F,$97,$9C,$39,$79,$9F,$80,$F8,$6F,$CA,$BF,$82,$F8,$6B
	.dc.b	$3F,$9E,$9F,$CC,$F9,$D9,$CB,$F8,$0F,$9B,$F8,$8F,$86,$FC,$EF,$93
	.dc.b	$F8,$AF,$86,$B8,$FE,$9A,$7F,$43,$59,$BF,$E1,$3A,$7E,$43,$E1,$7D
	.dc.b	$37,$FA,$CF,$81,$AF,$C9,$8F,$4B,$9A,$FE,$23,$B7,$A5,$B1,$FB,$4D
	.dc.b	$BD,$2F,$D1,$DB,$D2,$E6,$7D,$2F,$95,$FE,$37,$AF,$96,$3E,$73,$ED
	.dc.b	$8D,$C7,$DB,$D7,$C6,$FD,$DC,$BC,$EF,$DC,$F8,$6A,$FD,$B8,$DB,$9A
	.dc.b	$F3,$BE,$56,$BF,$84,$C7,$D2,$FE,$1B,$1F,$4B,$F8,$B8,$F9,$6F,$A4
	.dc.b	$F3,$9F,$C5,$30,$FA,$2F,$E4,$76,$DB,$E9,$BF,$E7,$3E,$67,$F0,$0F
	.dc.b	$85,$FE,$2B,$E5,$D9,$FE,$7A,$FF,$04,$F8,$5F,$E3,$35,$9B,$FC,$97
	.dc.b	$9B,$FC,$23,$E1,$BF,$8E,$F9,$B7,$1F,$E8,$AF,$F0,$CF,$86,$FE,$43
	.dc.b	$59,$BF,$DA,$7E,$83,$FC,$C5,$FF,$29,$7F,$00,$FF,$8B,$B0,$FE,$B6
	.dc.b	$7F,$5C,$9F,$E5,$D3,$AF,$87,$FE,$6F,$18,$8F,$F1,$DD,$71,$7A,$F9
	.dc.b	$FF,$C8,$61,$5E,$A9,$FE,$73,$BF,$C0,$7C,$3F,$0D,$5F,$DC,$46,$DE
	.dc.b	$B1,$EA,$7F,$A0,$EB,$F0,$1B,$97,$C0,$7F,$A2,$E9,$B7,$F8,$14,$FA
	.dc.b	$32,$F2,$47,$AB,$D4,$F3,$65,$63,$E1,$9B,$3E,$06,$1F,$11,$FE,$D5
	.dc.b	$A7,$F3,$77,$E4,$FC,$47,$C4,$F3,$3E,$07,$C3,$D9,$CD,$F1,$F2,$D8
	.dc.b	$3D,$1D,$CE,$61,$E5,$FC,$9F,$87,$F9,$FE,$95,$AD,$3F,$37,$E2,$3E
	.dc.b	$9D,$67,$2F,$BB,$E7,$7D,$1F,$83,$FA,$FE,$7F,$D3,$F8,$4B,$5F,$80
	.dc.b	$9F,$B7,$59,$CB,$C5,$AF,$92,$BF,$0F,$C1,$F6,$FE,$23,$D9,$1F,$80
	.dc.b	$E1,$FC,$1D,$66,$FD,$5A,$78,$F8,$7F,$C1,$FC,$2E,$0C,$57,$E3,$5F
	.dc.b	$53,$C3,$1F,$C9,$F0,$F9,$1F,$34,$7C,$5E,$0F,$34,$7E,$69,$F7,$47
	.dc.b	$C9,$DB,$38,$F8,$3C,$EF,$C9,$7C,$47,$E6,$1B,$5F,$3D,$FC,$B7,$C4
	.dc.b	$7E,$73,$CF,$CE,$5F,$A6,$F1,$FF,$35,$F1,$1F,$A0,$F2,$3F,$39,$F0
	.dc.b	$D6,$BF,$53,$3F,$A4,$AC,$E5,$F6,$FC,$8F,$D1,$7C,$3F,$EA,$1F,$D2
	.dc.b	$7C,$45,$C3,$F6,$33,$FA,$CF,$2F,$37,$F8,$3D,$3F,$55,$F0,$FE,$9B
	.dc.b	$FD,$BB,$5F,$D8,$FC,$3F,$A5,$CD,$7E,$2F,$3F,$4B,$F8,$DD,$BC,$B3
	.dc.b	$FC,$09,$CB,$EF,$3F,$81,$ED,$E5,$7F,$7B,$9D,$7F,$D6,$DE,$43,$5C
	.dc.b	$47,$FA,$AE,$13,$EB,$8F,$E1,$B3,$E5,$7F,$83,$FC,$37,$A5,$E6,$D7
	.dc.b	$96,$9B,$73,$97,$AD,$F9,$EF,$AD,$F3,$B5,$FE,$27,$C1,$F4,$DC,$9F
	.dc.b	$C8,$38,$3E,$9C,$FD,$58,$F2,$B5,$F7,$BF,$B3,$E0,$FA,$71,$FC,$A3
	.dc.b	$D3,$FE,$39,$F1,$1F,$C8,$9F,$E3,$DF,$0F,$67,$F9,$5C,$FF,$26,$F4
	.dc.b	$B3,$87,$EF,$7D,$3F,$E4,$9F,$0F,$FC,$A9,$B1,$9B,$FF,$6F,$9F,$EE
	.dc.b	$3F,$E0,$1F,$F3,$E4,$7F,$05,$F4,$37,$3F,$CE,$BD,$4F,$F2,$75,$7F
	.dc.b	$9D,$EB,$F8,$54,$63,$E2,$3F,$9B,$BA,$F8,$6F,$E7,$3E,$8E,$6B,$F8
	.dc.b	$86,$9E,$6C,$7D,$C7,$88,$FF,$73,$63,$FE,$7D,$9F,$E2,$DB,$EF,$C0
	.dc.b	$FD,$ED,$F8,$BE,$23,$D9,$87,$F4,$CD,$EB,$FD,$19,$AF,$B7,$0F,$E9
	.dc.b	$7B,$7B,$30,$FE,$0F,$AF,$96,$BD,$C7,$8C,$FF,$45,$67,$EF,$72,$79
	.dc.b	$FE,$21,$0D,$BD,$F3,$C6,$73,$BF,$F5,$AF,$53,$E4,$FC,$35,$AD,$60
	.dc.b	$8C,$47,$58,$7C,$47,$FA,$63,$C0,$FF,$4E,$CF,$92,$B1,$EA,$7D,$4F
	.dc.b	$86,$D7,$B1,$78,$7F,$A1,$60,$8F,$B3,$F0,$DC,$31,$F0,$DF,$5F,$5F
	.dc.b	$EE,$FC,$45,$9F,$2F,$C3,$57,$C9,$1E,$3F,$53,$F0,$73,$E5,$F4,$3F
	.dc.b	$0B,$F1,$16,$FD,$BF,$0D,$5F,$64,$7A,$FD,$BF,$C5,$CF,$B7,$DA,$FC
	.dc.b	$6F,$C2,$7B,$FD,$CB,$C5,$EA,$7F,$23,$F0,$9B,$CB,$E4,$D7,$C5,$F7
	.dc.b	$3F,$2B,$F0,$DF,$1F,$D7,$FC,$BF,$C3,$5B,$F9,$F3,$B7,$E8,$C7,$CD
	.dc.b	$F5,$FF,$3B,$F0,$D6,$7E,$AB,$5F,$A9,$F0,$BF,$4F,$D0,$DE,$DF,$3B
	.dc.b	$D6,$FD,$2F,$C3,$7D,$77,$F4,$FF,$0B,$70,$FB,$F3,$F6,$FD,$7C,$DF
	.dc.b	$D6,$F5,$FF,$5B,$F0,$DF,$7F,$D7,$B3,$F6,$9F,$F5,$F7,$B7,$FE,$DB
	.dc.b	$8E,$18,$FE,$2A,$BC,$8F,$E4,$67,$FD,$B1,$3F,$82,$DF,$8F,$C2,$FA
	.dc.b	$EB,$F8,$07,$07,$07,$F3,$97,$C1,$FF,$56,$78,$8D,$FD,$B8,$FB,$E7
	.dc.b	$F3,$3B,$FB,$5F,$CC,$6F,$ED,$73,$3E,$67,$F3,$9A,$7B,$5F,$44,$7B
	.dc.b	$5F,$47,$C3,$7E,$7F,$7F,$83,$1E,$2C,$FE,$0F,$C3,$7E,$8D,$F8,$7F
	.dc.b	$0D,$67,$E0,$4F,$E9,$FD,$DC,$E3,$FA,$2F,$6B,$E4,$FC,$37,$EA,$FD
	.dc.b	$CB,$3F,$A6,$7E,$6F,$C2,$FE,$BE,$B3,$8F,$C2,$F6,$3E,$8F,$C2,$7E
	.dc.b	$CD,$FA,$7F,$0B,$67,$E8,$4B,$1F,$F1,$97,$E8,$36,$7E,$A5,$5F,$6A
	.dc.b	$4B,$D7,$FB,$BF,$0D,$9C,$18,$2D,$7B,$8F,$BD,$5B,$75,$19,$7E,$DF
	.dc.b	$C3,$66,$B8,$FB,$7D,$B0,$F2,$FB,$DF,$8C,$EC,$D6,$3F,$C4,$F1,$FC
	.dc.b	$54,$63,$D8,$FC,$8F,$C3,$6A,$F8,$61,$EE,$F6,$BF,$2B,$F0,$BC,$BD
	.dc.b	$8F,$CB,$FC,$2D,$CF,$F2,$93,$DA,$B3,$7F,$1B,$DA,$FC,$EF,$C1,$F8
	.dc.b	$9F,$CF,$FC,$1D,$C3,$F3,$93,$E6,$F6,$F3,$7F,$3B,$DA,$FD,$2F,$C2
	.dc.b	$FA,$9B,$19,$BF,$D2,$7C,$27,$BB,$DB,$CC,$7C,$D9,$F9,$FC,$DF,$3B
	.dc.b	$DA,$FD,$7F,$C2,$7C,$47,$F6,$3F,$0B,$6F,$F5,$D3,$F2,$3D,$9C,$E1
	.dc.b	$F8,$0F,$6B,$F6,$FF,$09,$F3,$1B,$19,$BF,$E8,$DF,$C0,$3F,$E9,$BC
	.dc.b	$FF,$07,$ED,$6A,$7C,$87,$DC,$F6,$BF,$AA,$2F,$86,$5F,$FA,$83,$0F
	.dc.b	$C7,$D6,$A7,$3F,$0B,$F7,$23,$15,$FD,$0E,$BF,$2B,$EB,$62,$BF,$03
	.dc.b	$EE,$6B,$E1,$2E,$5E,$6F,$84,$AF,$F6,$1C,$53,$F8,$8F,$5F,$AD,$7E
	.dc.b	$2B,$D8,$EF,$F0,$9F,$8C,$F6,$6E,$BF,$A4,$7C,$7F,$09,$BC,$FE,$1D
	.dc.b	$7E,$5F,$F4,$1F,$F6,$44,$77,$F8,$5C,$7B,$9E,$A6,$C7,$CA,$F8,$5E
	.dc.b	$1E,$FE,$73,$FF,$51,$61,$FE,$A9,$AF,$47,$C2,$F5,$F7,$3E,$1F,$C2
	.dc.b	$D9,$FA,$73,$E1,$F7,$B3,$87,$2F,$77,$E4,$FC,$17,$95,$B1,$9B,$FB
	.dc.b	$7F,$0B,$E9,$F7,$B3,$1F,$6B,$BD,$AB,$DC,$FA,$7F,$07,$EF,$6E,$3E
	.dc.b	$87,$EB,$FC,$2F,$C3,$F7,$B3,$8F,$3C,$1E,$EF,$DC,$F8,$4F,$90,$FD
	.dc.b	$EF,$85,$B5,$F8,$E9,$F9,$9E,$FE,$73,$F8,$BE,$EF,$E1,$3E,$17,$E8
	.dc.b	$7B,$B6,$7E,$4B,$F8,$8F,$85,$FA,$95,$9C,$7C,$5E,$DF,$E3,$3E,$13
	.dc.b	$EC,$37,$0F,$A2,$FE,$43,$E1,$7E,$E7,$BD,$9B,$F5,$7B,$7F,$94,$F8
	.dc.b	$5F,$C0,$3F,$99,$FE,$01,$FF,$29,$CE,$8F,$D3,$CF,$E1,$7D,$ED,$65
	.dc.b	$F8,$7F,$77,$87,$C3,$7E,$27,$DD,$B3,$F8,$27,$A7,$C3,$6F,$1F,$F1
	.dc.b	$B6,$5F,$E3,$DE,$DF,$E6,$9D,$2F,$6F,$F5,$FF,$AE,$3E,$09,$E5,$7E
	.dc.b	$56,$73,$18,$D3,$F1,$F1,$FE,$36,$CD,$D1,$FC,$D7,$47,$D2,$7E,$4D
	.dc.b	$FF,$38,$E5,$D7,$2F,$D0,$7B,$FF,$07,$E1,$AD,$79,$BE,$1E,$BF,$E5
	.dc.b	$1E,$0F,$F3,$33,$F8,$3F,$80,$F9,$13,$FC,$FF,$F4,$2F,$F2,$3F,$BE
	.dc.b	$BF,$F2,$5F,$07,$F1,$E3,$F8,$FC,$BD,$DF,$A1,$F0,$FC,$5F,$A5,$F1
	.dc.b	$1C,$F3,$CD,$F1,$7F,$AA,$FE,$0A,$F8,$7F,$81,$E0,$FB,$F8,$FD,$D5
	.dc.b	$FD,$6B,$8F,$E5,$79,$79,$F5,$77,$97,$3A,$65,$CD,$FF,$39,$5F,$09
	.dc.b	$F0,$75,$B7,$FF,$2E,$57,$DA,$EA,$F8,$1F,$17,$F8,$9F,$8A,$B3,$9C
	.dc.b	$38,$DF,$F1,$88,$FC,$0C,$FC,$7F,$7F,$F2,$1F,$15,$F2,$7D,$DB,$87
	.dc.b	$07,$F2,$9C,$5B,$C3,$F1,$35,$F7,$76,$F4,$47,$FA,$27,$B9,$E3,$FB
	.dc.b	$E3,$D7,$B3,$C9,$C4,$6E,$79,$63,$F8,$C9,$FB,$5E,$F7,$E8,$FE,$2A
	.dc.b	$DF,$E6,$3E,$2A,$BF,$83,$8F,$C0,$7B,$FF,$A9,$9F,$C1,$78,$7F,$AB
	.dc.b	$F8,$AF,$C2,$7B,$D7,$8B,$E1,$BF,$AF,$F8,$AD,$E5,$FA,$4A,$F1,$BF
	.dc.b	$D4,$38,$1E,$7F,$A2,$3F,$33,$F4,$BF,$EA,$1C,$7E,$D7,$77,$E0,$27
	.dc.b	$FC,$33,$7C,$27,$EF,$72,$7C,$69,$E3,$3F,$97,$F8,$1F,$E0,$1F,$17
	.dc.b	$67,$F6,$BF,$17,$5F,$D0,$47,$E7,$7E,$0B,$F8,$34,$FE,$7F,$E0,$7F
	.dc.b	$84,$7C,$5F,$E8,$7E,$06,$E9,$F8,$07,$F8,$67,$C4,$EF,$2F,$E0,$35
	.dc.b	$EE,$7E,$C7,$AF,$ED,$7F,$31,$C1,$FC,$07,$0F,$6F,$8D,$FB,$DA,$E1
	.dc.b	$5F,$0B,$BB,$C0,$FF,$0C,$F8,$9A,$FE,$DF,$E2,$6B,$FE,$F0,$F3,$FF
	.dc.b	$69,$3F,$B7,$F2,$BF,$90,$57,$EE,$7E,$03,$FB,$D7,$F0,$0F,$F9,$FF
	.dc.b	$0F,$E9,$5E,$4A,$73,$F1,$5B,$9F,$F6,$85,$7F,$9E,$7C,$CF,$F2,$75
	.dc.b	$BE,$0F,$7C,$1F,$4B,$C7,$FE,$2D,$EE,$FB,$F1,$79,$EE,$47,$8B,$99
	.dc.b	$FF,$6F,$57,$C0,$7E,$2C,$78,$5F,$F7,$26,$8E,$5C,$47,$BE,$1F,$13
	.dc.b	$88,$FF,$2E,$F2,$27,$86,$9F,$B6,$F8,$3F,$83,$7F,$C6,$FE,$0F,$E1
	.dc.b	$FC,$4D,$C7,$DE,$D7,$C3,$F1,$3F,$C8,$7C,$ED,$E3,$EF,$9F,$CD,$7C
	.dc.b	$0F,$BF,$87,$DD,$DB,$ED,$F1,$53,$ED,$AF,$87,$C6,$AE,$D1,$EB,$F8
	.dc.b	$9A,$FA,$B8,$1F,$13,$88,$FF,$1B,$47,$F7,$FC,$FF,$88,$FC,$37,$DC
	.dc.b	$AF,$E6,$DF,$0D,$F7,$BE,$26,$E1,$F6,$9A,$FC,$CF,$89,$FF,$3D,$75
	.dc.b	$7F,$A2,$E3,$E7,$CF,$9D,$6F,$FD,$67,$2F,$CB,$C7,$F2,$AF,$C0,$7F
	.dc.b	$09,$1F,$E8,$C8,$FE,$A7,$D3,$FD,$5F,$E1,$FE,$07,$1A,$F7,$5F,$E9
	.dc.b	$1E,$E6,$BF,$94,$F8,$AF,$2C,$FD,$DF,$88,$AF,$C3,$8F,$3F,$C5,$7E
	.dc.b	$6A,$7D,$3A,$FE,$73,$E0,$7D,$7F,$11,$77,$FE,$7B,$C2,$8F,$D0,$FB
	.dc.b	$BC,$33,$FC,$ED,$FB,$FF,$61,$77,$7E,$55,$B8,$FC,$0A,$F7,$1F,$EA
	.dc.b	$A8,$77,$71,$AF,$97,$F1,$1F,$AE,$F6,$3E,$6F,$C3,$DC,$7C,$6F,$EC
	.dc.b	$BD,$0D,$E5,$FA,$EA,$FA,$7F,$0D,$FB,$6F,$26,$CF,$EC,$9A,$FE,$93
	.dc.b	$C3,$FB,$3E,$8E,$F2,$FC,$A4,$FF,$47,$F8,$AF,$CA,$75,$3A,$3F,$29
	.dc.b	$1F,$52,$31,$5F,$1B,$2F,$E0,$BD,$9D,$B0,$FD,$75,$D7,$F2,$91,$F8
	.dc.b	$8F,$86,$FE,$17,$3F,$8A,$F8,$4F,$E1,$BF,$15,$6F,$F8,$4B,$5F,$93
	.dc.b	$E6,$FE,$91,$FC,$8F,$57,$0C,$FD,$31,$F3,$BA,$D7,$EA,$E3,$FD,$63
	.dc.b	$E9,$7A,$4F,$F2,$1E,$1F,$F4,$C3,$FE,$65,$C1,$97,$FA,$66,$7F,$D2
	.dc.b	$D3,$98,$C7,$C4,$FE,$8B,$E1,$75,$F1,$3F,$A4,$F4,$78,$7C,$6D,$C3
	.dc.b	$F0,$27,$FA,$9F,$1F,$F4,$BE,$DF,$8E,$AE,$DF,$15,$FA,$CF,$84,$E3
	.dc.b	$3E,$CB,$FD,$37,$E9,$7F,$EB,$58,$FD,$47,$C2,$FF,$5D,$F2,$7F,$4D
	.dc.b	$C5,$FC,$96,$B5,$7F,$E2,$4E,$1F,$F2,$4F,$4F,$C4,$C7,$C1,$3E,$4E
	.dc.b	$5F,$E1,$FF,$17,$E7,$8F,$9F,$BB,$E2,$F1,$1F,$1E,$7F,$80,$FC,$47
	.dc.b	$C5,$9F,$E0,$7F,$0D,$F1,$FE,$2E,$DF,$5C,$11,$88,$EB,$F1,$7F,$C2
	.dc.b	$BE,$1F,$86,$9E,$F9,$0D,$FE,$AF,$BF,$FC,$3F,$8F,$E0,$7E,$06,$E6
	.dc.b	$F8,$1D,$7F,$C2,$23,$18,$FE,$EB,$6F,$B5,$F1,$9E,$1C,$FF,$01,$F1
	.dc.b	$75,$FE,$83,$C0,$9F,$E3,$BE,$E7,$E0,$6B,$F8,$FF,$C4,$7E,$0F,$E2
	.dc.b	$EE,$1E,$D6,$BF,$23,$E2,$FF,$92,$FC,$56,$F4,$F6,$4F,$AF,$C1,$FE
	.dc.b	$81,$B7,$79,$FE,$0D,$3F,$C1,$B8,$FF,$1F,$C7,$FC,$B7,$E3,$3F,$C2
	.dc.b	$D3,$FE,$59,$BE,$0F,$E5,$F4,$FC,$B7,$C6,$78,$A3,$F3,$9F,$19,$5F
	.dc.b	$6C,$7F,$38,$F8,$CF,$CE,$4B,$3F,$F2,$DF,$E8,$9F,$ED,$EB,$FD,$27
	.dc.b	$C6,$E5,$FF,$64,$46,$5F,$8F,$19,$D2,$4E,$67,$F3,$DF,$19,$9B,$FB
	.dc.b	$74,$A7,$C8,$7E,$B7,$0F,$E9,$1B,$FE,$93,$15,$FD,$FF,$82,$FF,$50
	.dc.b	$EE,$D3,$F3,$DE,$57,$DF,$7B,$5F,$0C,$FC,$5E,$27,$F6,$AF,$8D,$CE
	.dc.b	$5E,$1F,$8E,$FD,$E7,$C6,$E2,$F7,$DA,$BC,$6B,$AE,$1F,$8B,$F1,$DF
	.dc.b	$C0,$FE,$3B,$83,$63,$3A,$FE,$7F,$E2,$FA,$7B,$F9,$8F,$5C,$FF,$5F
	.dc.b	$EA,$FB,$71,$F9,$EE,$4F,$CB,$BA,$3F,$43,$E2,$FE,$7B,$2F,$DE,$78
	.dc.b	$BF,$9D,$7C,$33,$FC,$33,$F5,$E7,$E8,$7F,$D3,$32,$F1,$CF,$F5,$0F
	.dc.b	$45,$D3,$FD,$A2,$73,$3F,$DC,$6F,$8F,$C1,$7C,$33,$E7,$FC,$4F,$FA
	.dc.b	$A7,$87,$F9,$13,$6B,$F7,$8D,$7F,$9A,$FF,$00,$FF,$B9,$BA,$7F,$07
	.dc.b	$F2,$1C,$32,$FD,$0F,$0B,$D2,$FF,$10,$79,$FF,$E7,$7C,$7E,$01,$FE
	.dc.b	$02,$3F,$93,$9F,$C0,$63,$FC,$07,$1B,$C3,$F8,$16,$7A,$AD,$61,$FE
	.dc.b	$80,$8F,$F7,$3D,$7F,$B5,$2B,$75,$D3,$0F,$15,$FF,$0F,$DF,$FD,$D7
	.dc.b	$F1,$3F,$83,$F8,$FF,$0F,$7B,$6F,$F7,$A5,$FE,$6B,$B7,$FC,$5B,$97
	.dc.b	$F2,$99,$CB,$F9,$1B,$E3,$1E,$7A,$E1,$F1,$D8,$F8,$DF,$CF,$FA,$BF
	.dc.b	$0B,$E0,$FF,$9C,$3E,$E7,$35,$EB,$C7,$FD,$A3,$C1,$FE,$81,$7E,$3F
	.dc.b	$5F,$F9,$69,$F6,$5F,$C6,$F8,$5F,$CE,$7C,$7F,$8B,$82,$D7,$9D,$AF
	.dc.b	$D8,$F8,$AF,$D1,$7C,$3E,$F3,$FA,$FE,$D7,$D6,$F7,$7F,$A6,$3F,$49
	.dc.b	$FD,$BB,$F8,$1B,$FF,$1D,$72,$B6,$7C,$0E,$67,$FA,$CD,$7D,$A8,$FD
	.dc.b	$87,$C7,$7C,$6F,$8B,$B5,$F2,$DA,$FE,$33,$8B,$F6,$DF,$1B,$BD,$3F
	.dc.b	$17,$97,$E2,$BE,$03,$FB,$03,$F8,$67,$F8,$5B,$E1,$BF,$F4,$5A,$F7
	.dc.b	$B1,$1F,$D3,$63,$F1,$B1,$FC,$0F,$E3,$BE,$CF,$25,$AF,$BC,$D7,$F3
	.dc.b	$FF,$15,$FC,$27,$E3,$77,$97,$E7,$BE,$07,$F3,$BE,$DF,$F7,$37,$F3
	.dc.b	$2F,$F1,$E7,$C9,$C7,$F1,$FD,$DF,$1F,$85,$FC,$F3,$C3,$FE,$81,$CF
	.dc.b	$F9,$A9,$FE,$3B,$A3,$EA,$BF,$C4,$4F,$EF,$78,$5E,$E6,$A7,$33,$F5
	.dc.b	$27,$3A,$FF,$32,$3F,$CE,$13,$A9,$FF,$06,$FC,$0F,$6F,$47,$FC,$43
	.dc.b	$1F,$90,$C3,$F9,$AC,$FF,$A5,$FF,$05,$FF,$39,$CF,$ED,$CF,$F0,$EE
	.dc.b	$A9,$FF,$3C,$CE,$AF,$F8,$FD,$7F,$7B,$9F,$83,$3F,$EA,$E8,$FC,$DF
	.dc.b	$93,$FB,$BB,$E5,$97,$ED,$6B,$5B,$7C,$4B,$FD,$D7,$C1,$79,$FE,$2B
	.dc.b	$F8,$3F,$0A,$FF,$D7,$B8,$7F,$BF,$A7,$A5,$79,$32,$FE,$9E,$FA,$FD
	.dc.b	$5F,$F1,$43,$E8,$73,$5F,$B1,$C7,$F9,$56,$1F,$CB,$5F,$39,$E2,$8F
	.dc.b	$37,$A1,$FD,$6B,$E0,$FC,$93,$FE,$40,$CF,$C9,$E8,$62,$3E,$3C,$FC
	.dc.b	$68,$E1,$7F,$E0,$29,$FF,$10,$D7,$AF,$CD,$FF,$1E,$5F,$CC,$77,$3E
	.dc.b	$23,$CB,$E6,$7E,$2A,$7E,$8B,$FD,$B3,$E0,$BC,$13,$FB,$6C,$DE,$56
	.dc.b	$23,$FC,$9B,$3F,$62,$38,$65,$E3,$9F,$F7,$9D,$7F,$A0,$1F,$E2,$3E
	.dc.b	$CF,$D7,$BF,$BF,$E4,$F0,$9F,$21,$DF,$C7,$FC,$EC,$FE,$11,$FE,$FF
	.dc.b	$EF,$78,$67,$AF,$8D,$5F,$F2,$04,$7E,$2A,$37,$97,$8F,$7F,$BD,$E2
	.dc.b	$7F,$97,$EB,$F1,$EE,$AF,$C2,$7C,$AD,$1E,$17,$EB,$AB,$F2,$AF,$F8
	.dc.b	$43,$E0,$3C,$58,$7E,$27,$E2,$FC,$5D,$69,$FC,$DC,$6F,$3F,$16,$2F
	.dc.b	$8F,$FF,$38,$57,$E8,$1D,$67,$E2,$3E,$4F,$BF,$FB,$E9,$FD,$2B,$FE
	.dc.b	$20,$F8,$4F,$0D,$FF,$B9,$EB,$FB,$1F,$81,$FB,$27,$F5,$B3,$FA,$CF
	.dc.b	$8F,$E1,$97,$8B,$E3,$FF,$D0,$13,$FB,$07,$35,$E1,$DB,$F4,$9F,$19
	.dc.b	$FC,$3A,$BF,$6A,$FF,$8C,$3E,$0F,$E0,$FE,$09,$FE,$88,$B6,$7E,$17
	.dc.b	$FC,$A9,$3F,$EF,$49,$D4,$FE,$F3,$87,$FC,$51,$9F,$D5,$9F,$F3,$BB
	.dc.b	$FE,$96,$E9,$FD,$E5,$F1,$F8,$BF,$04,$FE,$B7,$1F,$E9,$37,$FE,$60
	.dc.b	$8F,$F4,$EF,$87,$C1,$FE,$ED,$19,$9F,$F4,$E1,$FD,$EF,$E1,$BD,$3F
	.dc.b	$11,$F6,$1F,$33,$FE,$4C,$7F,$94,$FB,$7F,$EA,$7A,$FF,$17,$F9,$1F
	.dc.b	$E4,$DB,$E5,$87,$6F,$86,$E0,$F9,$63,$F6,$75,$FE,$08,$7E,$E1,$FE
	.dc.b	$95,$C3,$E2,$7C,$0F,$F9,$6B,$E1,$BE,$5F,$BD,$F8,$A7,$E4,$39,$C3
	.dc.b	$E3,$7B,$9F,$EB,$79,$F9,$8E,$70,$FB,$18,$FF,$6B,$AF,$14,$7F,$91
	.dc.b	$EB,$E9,$7A,$3F,$E6,$8F,$87,$FB,$3E,$AF,$E6,$1F,$AC,$E6,$FE,$A7
	.dc.b	$A3,$FE,$C5,$9F,$B4,$E7,$0E,$FE,$87,$F5,$29,$FB,$CF,$F9,$D3,$E1
	.dc.b	$3F,$0B,$E6,$FE,$91,$FC,$13,$9C,$3F,$01,$E5,$7F,$B4,$67,$F0,$CE
	.dc.b	$70,$F5,$F8,$FF,$D7,$A7,$F1,$4F,$F4,$FF,$5B,$FC,$75,$FA,$6F,$F9
	.dc.b	$27,$0F,$F3,$37,$A9,$FC,$F3,$87,$F6,$51,$FC,$EF,$1F,$F2,$2D,$6D
	.dc.b	$FC,$19,$FE,$5F,$9D,$BF,$AA,$BE,$B5,$F8,$2E,$DE,$9E,$0D,$AE,$73
	.dc.b	$F9,$D8,$F2,$46,$23,$E8,$F8,$4C,$3F,$3F,$3F,$82,$C7,$C4,$FE,$A3
	.dc.b	$1F,$47,$7F,$28,$FF,$80,$33,$FE,$46,$7F,$5B,$9F,$B7,$A7,$F9,$04
	.dc.b	$E7,$C1,$FF,$42,$E5,$DE,$79,$5F,$FB,$90,$FF,$72,$D7,$3E,$DF,$CE
	.dc.b	$74,$7A,$E7,$BF,$C1,$7E,$DE,$38,$C7,$1A,$E1,$E0,$7C,$DE,$7F,$F1
	.dc.b	$1C,$78,$76,$C7,$07,$FB,$62,$3F,$CB,$57,$FE,$F4,$D7,$F8,$23,$FA
	.dc.b	$4E,$4F,$4F,$C8,$FF,$05,$7B,$C7,$9A,$3C,$8F,$F9,$1B,$F0,$5F,$8F
	.dc.b	$B3,$97,$F0,$35,$C3,$85,$5F,$33,$1F,$F0,$4E,$5E,$2C,$58,$FF,$98
	.dc.b	$5F,$F4,$F5,$7B,$1F,$C2,$E1,$CE,$3D,$91,$F8,$8F,$84,$FE,$7F,$1C
	.dc.b	$F0,$F5,$CF,$CC,$9E,$15,$EA,$86,$1F,$E8,$BD,$3F,$8E,$47,$F7,$DC
	.dc.b	$30,$FF,$AE,$33,$FC,$DE,$1E,$13,$95,$EA,$7D,$3B,$FF,$86,$2D,$F0
	.dc.b	$9F,$4A,$FD,$0F,$03,$CF,$F0,$7F,$86,$7D,$8E,$72,$FE,$C9,$4F,$83
	.dc.b	$F9,$F8,$FD,$FB,$F8,$51,$B8,$FE,$89,$3F,$E8,$BF,$17,$FD,$4B,$1F
	.dc.b	$96,$AF,$91,$F0,$3E,$8C,$7E,$3F,$C0,$FE,$59,$F9,$8E,$72,$F0,$F4
	.dc.b	$F8,$30,$FC,$AD,$7D,$18,$D4,$78,$39,$7F,$1F,$3F,$55,$F8,$BA,$7D
	.dc.b	$4F,$84,$FD,$1B,$F6,$5C,$E1,$EE,$F9,$2E,$D8,$7D,$4A,$FB,$D1,$A8
	.dc.b	$F6,$FC,$97,$E7,$27,$F0,$4F,$D2,$D3,$F0,$3F,$0D,$FB,$07,$F0,$CE
	.dc.b	$70,$F9,$7F,$0D,$DB,$93,$E3,$47,$FB,$E7,$4F,$B1,$3B,$CB,$E5,$FC
	.dc.b	$8F,$EA,$B1,$FC,$83,$F8,$0E,$0F,$C7,$FC,$8F,$EF,$9F,$CA,$39,$C7
	.dc.b	$EC,$FC,$47,$79,$FC,$C3,$9C,$3E,$BF,$C4,$7E,$DA,$7F,$38,$FE,$57
	.dc.b	$EC,$BF,$CA,$9C,$CC,$3F,$94,$DF,$E5,$7E,$23,$FC,$59,$5A,$9F,$54
	.dc.b	$E5,$FF,$71,$2B,$FB,$5D,$E8,$C4,$FF,$8C,$5F,$F4,$59,$FC,$B4,$FF
	.dc.b	$2F,$1A,$30,$7F,$17,$11,$3E,$43,$FD,$42,$9B,$F2,$3F,$C0,$63,$15
	.dc.b	$FC,$47,$A7,$FC,$6B,$F0,$D8,$C3,$FC,$4B,$1F,$D2,$0F,$E1,$11,$9D
	.dc.b	$3F,$D0,$D3,$FC,$C2,$7D,$73,$FE,$98,$F0,$BF,$88,$4F,$F4,$A9,$E3
	.dc.b	$8F,$ED,$9F,$F5,$AD,$7F,$A4,$30,$FF,$09,$D6,$DF,$F5,$53,$FD,$40
	.dc.b	$FF,$70,$C6,$DC,$1F,$CB,$F9,$3D,$3D,$FF,$CF,$9F,$3B,$FC,$11,$F2
	.dc.b	$FC,$57,$E3,$DF,$5B,$9B,$FE,$A3,$7F,$66,$FE,$A6,$5F,$C8,$5F,$9D
	.dc.b	$CE,$C3,$E1,$FC,$57,$F1,$0E,$DF,$93,$F1,$5F,$D3,$9F,$8E,$E7,$2F
	.dc.b	$F3,$66,$4F,$8A,$FD,$05,$7C,$C7,$59,$7F,$92,$E4,$7F,$48,$9F,$A2
	.dc.b	$FF,$1E,$EF,$FA,$FF,$15,$FD,$71,$FA,$AE,$70,$EF,$F1,$5F,$AD,$9F
	.dc.b	$B2,$E7,$0E,$DF,$15,$FD,$52,$7E,$EB,$FC,$AB,$C1,$FC,$1F,$C5,$7F
	.dc.b	$68,$7F,$02,$E7,$0F,$5F,$C5,$7E,$E6,$7F,$0A,$E7,$0F,$57,$C5,$7F
	.dc.b	$60,$9F,$C4,$BF,$CD,$BC,$4F,$C7,$FC,$57,$F7,$47,$F1,$AE,$70,$F8
	.dc.b	$FF,$15,$FC,$16,$7F,$22,$E7,$0F,$8D,$F1,$3F,$DB,$27,$F2,$AF,$F4
	.dc.b	$2F,$1F,$F3,$7F,$11,$FE,$00,$7F,$32,$E7,$0F,$A7,$F1,$1F,$C4,$67
	.dc.b	$F3,$AE,$70,$FA,$5F,$11,$FD,$E2,$7F,$42,$FF,$4E,$F2,$7F,$D3,$7F
	.dc.b	$53,$FE,$76,$F0,$3F,$C5,$B1,$FD,$B6,$3F,$CF,$3D,$98,$CB,$FD,$59
	.dc.b	$1F,$E7,$8C,$7F,$AD,$1F,$E2,$EA,$E1,$F1,$1D,$6B,$FC,$47,$5F,$E7
	.dc.b	$3A,$FE,$B6,$6E,$3F,$4A,$FF,$88,$CF,$F7,$14,$6F,$D4,$FE,$F7,$E6
	.dc.b	$79,$BE,$23,$FC,$42,$F9,$5F,$EC,$EF,$F1,$29,$FF,$3B,$DF,$F7,$0B
	.dc.b	$FF,$15,$CF,$AF,$E2,$37,$3E,$4F,$88,$FF,$0D,$CF,$BD,$FE,$E3,$E7
	.dc.b	$FC,$4F,$88,$FF,$17,$BF,$0D,$CD,$7C,$2F,$88,$FE,$81,$3F,$1D,$CE
	.dc.b	$3F,$EA,$BC,$BF,$CE,$8F,$F9,$06,$7E,$67,$C4,$7F,$7E,$F3,$FE,$8F
	.dc.b	$C4,$7F,$8F,$1F,$A0,$E7,$1F,$9F,$F1,$1F,$D3,$67,$EA,$39,$C3,$B7
	.dc.b	$C4,$7F,$8D,$27,$EC,$3F,$E0,$AF,$53,$EE,$FC,$47,$F9,$21,$FB,$8E
	.dc.b	$70,$FB,$7F,$11,$FD,$66,$7F,$00,$E7,$0F,$57,$C4,$7F,$90,$27,$F0
	.dc.b	$8F,$F8,$7F,$F4,$DF,$E5,$57,$FD,$2A,$BF,$9D,$7A,$CE,$0F,$F3,$6F
	.dc.b	$AD,$FD,$3E,$7F,$C8,$9D,$9F,$CF,$DF,$F2,$B6,$1C,$9F,$E4,$67,$F4
	.dc.b	$2A,$E3,$87,$F5,$28,$E1,$3B,$7F,$C7,$BF,$01,$CD,$E1,$F1,$3F,$E0
	.dc.b	$09,$D7,$5F,$F9,$9B,$04,$FA,$7D,$9F,$F5,$FF,$B1,$C7,$5F,$47,$C4
	.dc.b	$63,$2E,$99,$72,$F0,$75,$97,$F9,$F3,$7F,$E9,$35,$FD,$67,$DB,$7A
	.dc.b	$DF,$AC,$F8,$8F,$ED,$6F,$EA,$DF,$F2,$C5,$3E,$17,$FB,$4F,$C8,$FF
	.dc.b	$97,$B8,$BC,$53,$9E,$3F,$D2,$F1,$FF,$9F,$AB,$FC,$BD,$AF,$9D,$FF
	.dc.b	$2D,$69,$F4,$1F,$F3,$07,$B9,$FC,$FE,$3F,$BA,$70,$A7,$FC,$BB,$1B
	.dc.b	$F8,$6F,$B5,$EF,$FF,$02,$F8,$8F,$AD,$A7,$D7,$AF,$F4,$89,$F0,$A7
	.dc.b	$F8,$24,$F0,$D3,$F7,$39,$7F,$66,$AF,$AB,$AF,$D1,$BF,$C3,$3B,$8F
	.dc.b	$D2,$7C,$47,$CD,$7E,$DC,$FF,$0F,$8D,$DF,$FA,$E6,$7F,$0D,$F8,$FA
	.dc.b	$BC,$B3,$F8,$F9,$FF,$04,$72,$FF,$6B,$7F,$85,$1C,$27,$9A,$CF,$FD
	.dc.b	$F4,$FE,$2B,$9F,$F0,$D9,$7E,$6D,$DC,$FF,$97,$A7,$FD,$55,$BF,$E6
	.dc.b	$F8,$BD,$11,$E0,$BF,$E4,$D1,$C7,$2F,$44,$FF,$86,$FF,$7C,$7B,$FE
	.dc.b	$88,$FD,$09,$F8,$CC,$BF,$D1,$13,$AA,$FB,$FC,$7F,$9B,$A6,$AB,$FF
	.dc.b	$2D,$ED,$6F,$E8,$E7,$BA,$C5,$FF,$8B,$0F,$F1,$65,$63,$5F,$CB,$C7
	.dc.b	$EC,$30,$F5,$BF,$86,$FA,$FF,$F5,$7C,$73,$3F,$D5,$CF,$E4,$23,$F4
	.dc.b	$7F,$09,$EB,$7F,$0A,$7E,$EF,$4F,$F4,$66,$7F,$0A,$3F,$7B,$3E,$EF
	.dc.b	$87,$FE,$10,$FE,$D7,$E1,$3F,$6C,$FF,$0B,$73,$5F,$EE,$AC,$3F,$3B
	.dc.b	$7F,$7A,$BC,$7F,$19,$BA,$ED,$3F,$E5,$1D,$7F,$63,$18,$8F,$93,$87
	.dc.b	$F1,$7E,$BE,$11,$F8,$39,$FE,$8C,$FE,$2B,$E0,$BF,$8B,$7E,$BB,$FE
	.dc.b	$C7,$E8,$FC,$E4,$FE,$3B,$0F,$E4,$15,$AD,$FF,$8A,$4E,$70,$FD,$1C
	.dc.b	$7D,$3F,$86,$F9,$B3,$E5,$BF,$E5,$11,$E3,$AF,$37,$03,$E1,$3E,$67
	.dc.b	$77,$FB,$3B,$0F,$C0,$EB,$F8,$69,$F4,$F1,$FE,$6A,$7D,$18,$7F,$9C
	.dc.b	$76,$F4,$72,$FE,$0F,$4F,$3C,$E7,$5F,$3D,$FF,$30,$9F,$2B,$F8,$ED
	.dc.b	$6F,$07,$F9,$26,$3F,$56,$7E,$BA,$3D,$37,$F4,$3A,$5C,$9F,$EE,$5F
	.dc.b	$17,$E0,$4F,$E6,$AF,$F9,$F5,$6B,$BB,$D3,$3F,$42,$7F,$A0,$F1,$FC
	.dc.b	$B7,$FC,$16,$7F,$3E,$CF,$D2,$B6,$FF,$7D,$71,$7F,$96,$63,$F9,$AE
	.dc.b	$9F,$3B,$2F,$1F,$C6,$FA,$A3,$F3,$2F,$F5,$7F,$87,$F9,$BA,$FF,$3F
	.dc.b	$DB,$FD,$D7,$AF,$E1,$F7,$FF,$2B,$BE,$9C,$BF,$AD,$CE,$9F,$D6,$1E
	.dc.b	$88,$F9,$8F,$A0,$FA,$67,$9F,$83,$DC,$FE,$06,$3A,$3A,$D7,$D7,$D7
	.dc.b	$FB,$9C,$BE,$8F,$07,$D2,$7E,$0F,$1F,$EE,$2F,$F3,$DF,$0F,$EA,$7F
	.dc.b	$AA,$D5,$7F,$B2,$C7,$0C,$7E,$FC,$7E,$96,$7E,$73,$FD,$A0,$E3,$1E
	.dc.b	$A7,$07,$D1,$7C,$EF,$BC,$F6,$1F,$DD,$8F,$E2,$B1,$E8,$8F,$D0,$C7
	.dc.b	$F6,$8B,$F1,$3B,$D1,$5E,$78,$FE,$4D,$F0,$95,$ED,$8F,$BD,$74,$27
	.dc.b	$F0,$7A,$7E,$7A,$7F,$BA,$FC,$3F,$B3,$AB,$F6,$F1,$E7,$9F,$9A,$EB
	.dc.b	$4F,$77,$2F,$E2,$63,$FC,$17,$9F,$E2,$78,$5F,$17,$EB,$9F,$BE,$7F
	.dc.b	$B5,$23,$37,$F1,$23,$F8,$86,$5E,$97,$5D,$D5,$FE,$73,$1F,$CA,$A3
	.dc.b	$FC,$50,$F9,$FA,$BF,$25,$F1,$76,$BC,$C7,$CE,$3F,$62,$7F,$35,$7E
	.dc.b	$CE,$7E,$98,$FE,$C3,$5E,$9F,$8B,$FD,$CB,$FC,$E1,$FF,$1B,$7C,$57
	.dc.b	$97,$AF,$F6,$FD,$3E,$5E,$54,$FF,$3E,$8D,$EB,$F8,$3E,$2F,$97,$97
	.dc.b	$F2,$DF,$8A,$F3,$3F,$E3,$2B,$F1,$F4,$FF,$51,$9B,$3F,$E2,$89,$F9
	.dc.b	$93,$FC,$31,$FE,$96,$FF,$92,$67,$C9,$C3,$FD,$86,$BF,$DE,$E7,$F8
	.dc.b	$C2,$3F,$A8,$F8,$9B,$C7,$C9,$C3,$FD,$7F,$E2,$BC,$77,$FE,$02,$BF
	.dc.b	$0B,$FE,$46,$8A,$E6,$BF,$A6,$78,$3F,$C7,$AB,$FA,$EB,$FE,$5B,$AF
	.dc.b	$1F,$47,$C5,$7F,$B0,$B9,$AF,$0D,$FF,$1C,$C9,$F1,$1E,$28,$FF,$2D
	.dc.b	$DF,$81,$FE,$D9,$15,$FE,$AD,$1F,$8F,$F8,$5F,$E5,$EF,$F6,$E7,$FC
	.dc.b	$D9,$F1,$5F,$89,$F2,$1E,$47,$D7,$8F,$E8,$BC,$1F,$E0,$59,$CE,$7F
	.dc.b	$93,$C7,$FC,$0F,$5E,$37,$FB,$27,$C5,$57,$33,$FA,$2B,$FF,$04,$4F
	.dc.b	$96,$72,$7F,$65,$07,$FA,$E0,$FE,$47,$F0,$DF,$D1,$AF,$FC,$0A,$FF
	.dc.b	$A0,$E7,$F1,$9F,$15,$F8,$D7,$FD,$0C,$E7,$0F,$93,$7F,$E8,$1F,$D3
	.dc.b	$DE,$3B,$E2,$33,$3F,$E3,$CC,$3C,$95,$A9,$F4,$4F,$F9,$D2,$7F,$AB
	.dc.b	$FC,$5F,$F5,$87,$FD,$28,$F8,$30,$FF,$01,$7E,$F9,$F1,$AB,$F1,$A7
	.dc.b	$FA,$2B,$2F,$96,$FF,$A2,$63,$5E,$3F,$F1,$1E,$D6,$3F,$27,$3F,$15
	.dc.b	$FE,$D7,$B3,$F4,$AF,$F8,$DA,$7C,$FA,$FA,$E3,$F7,$5E,$0F,$ED,$DF
	.dc.b	$F4,$FC,$FA,$34,$FE,$4F,$1F,$E9,$C8,$F4,$C7,$DB,$C5,$AF,$E9,$BE
	.dc.b	$23,$F7,$0F,$F8,$D1,$F9,$97,$E3,$EC,$FF,$24,$69,$E6,$F1,$FF,$93
	.dc.b	$4F,$5C,$7E,$4B,$F4,$74,$F3,$4F,$F2,$68,$FF,$56,$E3,$FC,$8F,$E1
	.dc.b	$7C,$D1,$FC,$8B,$C2,$FE,$FF,$1F,$22,$3F,$D7,$5E,$C3,$CE,$FD,$DF
	.dc.b	$C1,$7F,$21,$8C,$67,$FC,$9F,$0F,$F0,$3C,$F1,$BF,$C6,$9F,$8E,$D3
	.dc.b	$D1,$E7,$FF,$68,$84,$61,$FE,$DC,$7C,$AB,$F5,$BC,$8D,$EB,$EC,$AF
	.dc.b	$EC,$38,$3D,$C7,$A1,$FE,$07,$6E,$1F,$4B,$0F,$F3,$FC,$7F,$56,$8F
	.dc.b	$43,$A8,$FE,$86,$33,$FF,$03,$BF,$8C,$3F,$D0,$FE,$13,$DF,$F8,$5E
	.dc.b	$8F,$F9,$A9,$F8,$13,$F0,$35,$F6,$3F,$DC,$0F,$53,$8E,$8F,$C6,$F3
	.dc.b	$7E,$4B,$3F,$7D,$71,$DF,$DC,$FF,$BA,$36,$F7,$79,$3E,$DF,$23,$DC
	.dc.b	$F8,$25,$3F,$0E,$71,$1E,$BE,$0F,$E8,$DF,$80,$7F,$BA,$3C,$BF,$F4
	.dc.b	$97,$0F,$C1,$ED,$FF,$09,$62,$AF,$F7,$26,$1F,$EF,$1D,$FD,$1D,$5F
	.dc.b	$92,$D7,$CB,$3E,$D3,$EC,$1F,$A4,$F0,$F5,$E0,$FB,$23,$CB,$CC,$F0
	.dc.b	$FE,$B1,$F5,$FB,$FD,$5A,$7D,$B8,$F4,$ED,$F5,$EF,$E9,$1F,$EE,$0D
	.dc.b	$7F,$D9,$1C,$7F,$E8,$39,$D6,$FE,$A9,$CD,$7D,$CD,$FD,$2F,$AF,$6F
	.dc.b	$4F,$2F,$D2,$D3,$F7,$BB,$79,$F5,$FC,$67,$6F,$E2,$CF,$C6,$7C,$6F
	.dc.b	$9B,$83,$F7,$3B,$FF,$35,$DD,$E7,$7C,$EE,$3F,$D3,$3E,$CE,$0F,$F6
	.dc.b	$D7,$85,$FC,$1F,$83,$F4,$9C,$1E,$2B,$F8,$F1,$FB,$F3,$F9,$16,$3F
	.dc.b	$25,$FD,$0E,$DE,$5F,$D8,$7F,$9C,$CF,$E7,$67,$F9,$DE,$E9,$CC,$7F
	.dc.b	$3F,$3F,$9F,$1F,$EE,$07,$F9,$A4,$FE,$E5,$DF,$1F,$F2,$5D,$FF,$0F
	.dc.b	$DD,$FE,$0B,$7D,$4F,$E3,$39,$3E,$5E,$1F,$C3,$F2,$FE,$7A,$FE,$0B
	.dc.b	$C2,$FE,$20,$FE,$83,$5F,$C3,$ED,$F0,$63,$FD,$C3,$8F,$FA,$53,$E3
	.dc.b	$BF,$D2,$AF,$E6,$3B,$51,$4F,$ED,$1E,$B7,$F8,$75,$FC,$EF,$93,$FE
	.dc.b	$9B,$C3,$A6,$9E,$4C,$7C,$1A,$79,$1F,$EC,$FC,$1F,$2E,$7F,$6B,$5E
	.dc.b	$3C,$BB,$CF,$CB,$AF,$EC,$BF,$1D,$F1,$23,$F8,$DF,$57,$FB,$DE,$7F
	.dc.b	$8A,$1F,$ED,$5C,$FF,$5F,$5B,$8F,$F3,$53,$5F,$F7,$59,$FC,$6F,$C8
	.dc.b	$79,$BF,$E4,$B9,$D6,$5F,$16,$73,$8F,$F8,$07,$3F,$66,$1F,$C1,$74
	.dc.b	$F6,$3F,$D5,$67,$D9,$C1,$F3,$BB,$BC,$B3,$EA,$3F,$D8,$7B,$FF,$38
	.dc.b	$7F,$09,$B7,$AD,$F5,$E9,$E2,$AF,$F0,$97,$8F,$EC,$CB,$FC,$C1,$A7
	.dc.b	$F2,$4E,$56,$9F,$CF,$B6,$FF,$6E,$63,$F8,$CC,$BF,$95,$DE,$EB,$FD
	.dc.b	$2F,$1F,$9A,$3F,$3C,$7E,$24,$A7,$F8,$15,$F5,$E1,$FA,$1A,$E9,$87
	.dc.b	$82,$BF,$C3,$39,$FA,$5F,$E8,$D2,$F9,$CF,$F7,$0B,$C3,$0F,$0B,$97
	.dc.b	$FC,$58,$D9,$FE,$1A,$7E,$01,$FA,$93,$F7,$1F,$D0,$6D,$FB,$79,$FE
	.dc.b	$15,$8F,$D2,$E2,$FE,$75,$1F,$C8,$4C,$94,$F9,$B7,$E4,$F8,$EF,$F4
	.dc.b	$E7,$3F,$D6,$96,$3E,$3D,$3F,$D4,$CA,$BF,$C7,$4E,$B3,$FF,$10,$F8
	.dc.b	$3F,$89,$8F,$E7,$FB,$7F,$2D,$9F,$C4,$6B,$E7,$C7,$E7,$74,$7F,$82
	.dc.b	$1F,$9C,$79,$4F,$D3,$94,$FE,$F3,$7E,$3A,$FD,$0E,$7E,$78,$F2,$47
	.dc.b	$FA,$2E,$31,$A7,$E8,$B3,$F3,$FC,$7F,$E7,$5F,$CF,$39,$AF,$2E,$FF
	.dc.b	$9A,$E5,$FC,$EE,$5F,$B6,$3C,$47,$F2,$68,$FC,$15,$FC,$88,$FC,$EE
	.dc.b	$5E,$27,$FC,$71,$97,$8B,$0F,$BB,$5F,$20,$FC,$CF,$85,$E3,$8F,$F3
	.dc.b	$66,$DF,$63,$83,$FD,$A9,$9F,$8B,$1F,$CB,$6D,$F1,$DF,$E3,$99,$7E
	.dc.b	$6A,$7F,$84,$F2,$FC,$3C,$FC,$95,$FA,$AF,$8F,$AF,$E5,$23,$EC,$C4
	.dc.b	$CF,$FC,$E3,$97,$E6,$EB,$F8,$29,$FD,$77,$D0,$FE,$09,$5F,$D2,$5F
	.dc.b	$C8,$EF,$F1,$F8,$BE,$E5,$7D,$AD,$7F,$97,$BF,$C1,$4F,$92,$7E,$32
	.dc.b	$FC,$6F,$F6,$9E,$BA,$FF,$1E,$8F,$F4,$D7,$77,$F8,$06,$7F,$1F,$F1
	.dc.b	$FF,$E0,$8F,$7B,$F8,$B3,$FD,$6D,$F1,$DF,$EE,$0F,$06,$1F,$1D,$FB
	.dc.b	$4F,$91,$FB,$47,$93,$91,$5E,$28,$FD,$0F,$C7,$D9,$CE,$FF,$C0,$63
	.dc.b	$F4,$27,$F8,$6B,$B3,$F9,$43,$FD,$B5,$F1,$E1,$E1,$BF,$B3,$C9,$FE
	.dc.b	$4F,$EA,$FD,$AC,$F8,$E7,$27,$F3,$6B,$F0,$7C,$7F,$85,$FD,$7B,$31
	.dc.b	$FA,$E7,$F1,$FF,$27,$F9,$6E,$FF,$E2,$7F,$3B,$F9,$D5,$7F,$7F,$7C
	.dc.b	$79,$FF,$97,$FA,$3F,$19,$A7,$8A,$3F,$15,$C7,$FC,$DF,$4F,$CD,$E1
	.dc.b	$F5,$31,$FF,$19,$79,$7E,$4D,$3F,$C2,$9E,$3F,$F9,$5F,$7F,$57,$93
	.dc.b	$FE,$FE,$9F,$2E,$3F,$EB,$CF,$09,$C1,$E6,$C3,$FC,$4B,$5E,$9A,$FE
	.dc.b	$8B,$7C,$5F,$F4,$14,$E9,$E5,$D9,$FE,$14,$C7,$E6,$70,$7A,$BB,$BF
	.dc.b	$C4,$F5,$FB,$6D,$7F,$C2,$1F,$1B,$E9,$F2,$0D,$F8,$DF,$C6,$C6,$35
	.dc.b	$FE,$79,$F1,$FF,$EE,$3E,$6F,$EC,$4F,$F7,$C7,$E4,$E3,$FE,$84,$3F
	.dc.b	$13,$A7,$DD,$E6,$F9,$51,$E6,$3F,$97,$C7,$DD,$E9,$5F,$E2,$FE,$3F
	.dc.b	$F1,$8F,$F2,$A6,$63,$F2,$2F,$F6,$27,$F8,$03,$4F,$EE,$93,$FD,$F1
	.dc.b	$F1,$C7,$F9,$7C,$F1,$F4,$7E,$37,$7F,$87,$97,$86,$FF,$39,$C1,$FE
	.dc.b	$7F,$AF,$E5,$11,$A7,$FA,$9E,$1F,$D6,$FC,$1F,$E0,$5C,$DF,$63,$2F
	.dc.b	$EF,$3F,$BE,$7E,$89,$FE,$AA,$7F,$46,$AF,$F3,$EC,$66,$BF,$8C,$F5
	.dc.b	$7F,$04,$AF,$D9,$E1,$E5,$D7,$FB,$5C,$7E,$C7,$2F,$2D,$7F,$98,$AB
	.dc.b	$CB,$5A,$D3,$F1,$BE,$8F,$FA,$23,$CF,$FF,$60,$3F,$E5,$B7,$E1,$F9
	.dc.b	$5F,$EE,$6D,$FF,$09,$7C,$70,$FC,$EF,$89,$FD,$17,$4F,$AB,$1F,$D2
	.dc.b	$9F,$E7,$3A,$7C,$37,$ED,$1F,$ED,$0F,$33,$FD,$2F,$E9,$7F,$B3,$DF
	.dc.b	$F3,$53,$E0,$F4,$3F,$29,$7E,$DA,$FE,$3F,$97,$F3,$47,$B9,$FE,$06
	.dc.b	$8D,$B4,$F0,$E5,$E7,$7C,$27,$FB,$3F,$BB,$FB,$53,$FE,$04,$EC,$FF
	.dc.b	$52,$6F,$FE,$E1,$F3,$7F,$CF,$8F,$8E,$7B,$9F,$85,$3F,$C6,$87,$93
	.dc.b	$8B,$F7,$CF,$F7,$2C,$7E,$13,$4F,$21,$FE,$32,$DB,$E2,$2E,$6F,$F3
	.dc.b	$46,$5F,$D5,$F3,$FD,$D4,$FF,$B9,$73,$FF,$5C,$6F,$FE,$F5,$E7,$FF
	.dc.b	$48,$3E,$59,$EC,$79,$7A,$BF,$52,$F5,$EE,$F8,$AD,$3D,$7C,$9F,$E2
	.dc.b	$D7,$F1,$67,$98,$FF,$17,$E9,$E6,$D3,$CB,$EC,$7F,$B3,$B8,$7F,$39
	.dc.b	$AF,$9A,$FE,$64,$FF,$A8,$E7,$C9,$1F,$E1,$0E,$7F,$DE,$6B,$CF,$3F
	.dc.b	$1E,$7E,$1A,$FD,$79,$E1,$3F,$C2,$B9,$ED,$FE,$37,$E9,$FF,$B7,$AF
	.dc.b	$C3,$1F,$C8,$FC,$2F,$F6,$FC,$FF,$AC,$E7,$C3,$3E,$C3,$E2,$3C,$DF
	.dc.b	$8F,$7E,$17,$FA,$39,$FE,$AA,$3C,$3A,$7E,$25,$F9,$07,$FA,$13,$85
	.dc.b	$E9,$7F,$BC,$36,$F1,$CF,$F0,$E1,$3A,$AF,$E2,$5B,$FF,$B1,$78,$BF
	.dc.b	$DE,$57,$E6,$BF,$82,$79,$AB,$C1,$5F,$EA,$0E,$4F,$DD,$4F,$D8,$E4
	.dc.b	$FD,$DB,$F9,$1E,$9F,$F6,$56,$5F,$E9,$E8,$FC,$E7,$87,$FE,$DA,$7F
	.dc.b	$86,$C6,$7B,$3F,$DD,$5C,$1E,$19,$F8,$3B,$FF,$B9,$F4,$F0,$F1,$FD
	.dc.b	$6C,$E5,$7E,$B0,$F0,$1F,$E1,$2F,$BA,$FF,$4E,$F1,$FF,$53,$E2,$F8
	.dc.b	$7C,$1F,$C7,$EB,$FD,$57,$5F,$E9,$17,$FD,$5B,$C4,$DF,$E3,$47,$FB
	.dc.b	$EF,$87,$F8,$F3,$F1,$35,$F2,$3E,$0E,$5F,$EF,$DD,$5F,$CC,$63,$E2
	.dc.b	$1E,$58,$CD,$7E,$D2,$04,$18,$FF,$A2,$5F,$C5,$1E,$4C,$3E,$36,$FF
	.dc.b	$9D,$8E,$13,$B7,$F0,$67,$3C,$BD,$5E,$9F,$CB,$CB,$F9,$3C,$F9,$F7
	.dc.b	$FE,$4E,$FF,$96,$8F,$39,$FC,$0F,$0F,$9E,$FF,$AF,$30,$FF,$3A,$3E
	.dc.b	$43,$F8,$17,$C0,$7C,$E7,$5E,$9F,$CE,$E2,$FF,$0B,$CF,$CD,$9F,$CE
	.dc.b	$F7,$7E,$B8,$FF,$49,$76,$7F,$04,$E2,$FC,$67,$47,$E7,$78,$FF,$45
	.dc.b	$B7,$CB,$C3,$C8,$7A,$4F,$E5,$27,$C2,$F1,$3E,$EF,$37,$82,$EC,$FA
	.dc.b	$67,$C8,$FD,$EC,$FE,$46,$3E,$08,$FF,$26,$1F,$77,$3F,$2F,$5F,$ED
	.dc.b	$63,$F8,$47,$0F,$F2,$A6,$9F,$C1,$4F,$F7,$CF,$66,$DF,$AB,$E9,$F8
	.dc.b	$9F,$8F,$C3,$5F,$C0,$1F,$80,$8F,$2B,$AE,$3F,$F4,$BB,$E4,$3F,$47
	.dc.b	$8F,$92,$3E,$AC,$7F,$08,$C3,$F4,$6F,$C6,$E8,$FC,$7E,$DF,$8E,$7F
	.dc.b	$CA,$5C,$9F,$E9,$9E,$BF,$F0,$0F,$0F,$D6,$E1,$F2,$61,$F8,$CF,$4B
	.dc.b	$C6,$FE,$2F,$8B,$FD,$84,$FE,$8F,$2F,$F4,$21,$FE,$D3,$E1,$F8,$11
	.dc.b	$FA,$5D,$3C,$6F,$F3,$8D,$7C,$EF,$C0,$D1,$E8,$F8,$21,$D2,$C9,$C3
	.dc.b	$99,$FD,$2D,$79,$1F,$D4,$D6,$E7,$F3,$DD,$7F,$B6,$9F,$EC,$7C,$3F
	.dc.b	$09,$FC,$41,$FC,$AC,$F3,$1F,$B5,$CF,$F4,$71,$F1,$BC,$FF,$26,$1E
	.dc.b	$3D,$FC,$5A,$7F,$88,$75,$FD,$C7,$37,$F2,$9D,$BE,$64,$72,$CB,$FC
	.dc.b	$E5,$A7,$8F,$A7,$F3,$DD,$5F,$D8,$30,$FC,$86,$AB,$FC,$66,$DF,$D1
	.dc.b	$7A,$BF,$A1,$72,$3A,$7E,$6C,$FF,$06,$F3,$BC,$BB,$7C,$AD,$BF,$9D
	.dc.b	$F3,$FF,$BF,$78,$15,$F1,$EF,$E6,$DE,$DF,$B4,$7E,$37,$AF,$F8,$2E
	.dc.b	$BF,$1C,$9A,$F9,$27,$E3,$9F,$8D,$7D,$CF,$EC,$8F,$E3,$78,$3C,$EF
	.dc.b	$2F,$2F,$C8,$7C,$5C,$BF,$76,$3E,$BF,$4F,$E2,$76,$FC,$2E,$1F,$07
	.dc.b	$1F,$C4,$BF,$D8,$0F,$B3,$C4,$E3,$F7,$79,$5F,$0B,$83,$EA,$BD,$B8
	.dc.b	$3E,$0E,$3F,$AA,$DF,$FD,$17,$CD,$F4,$FA,$BE,$5F,$17,$F0,$DD,$3F
	.dc.b	$76,$7E,$14,$72,$FD,$1C,$3C,$B1,$EE,$E5,$FA,$3C,$7F,$84,$7C,$0F
	.dc.b	$B2,$3F,$71,$CF,$F8,$07,$FC,$6F,$C1,$F8,$0E,$1F,$99,$1F,$D7,$9F
	.dc.b	$ED,$3A,$FD,$08,$F9,$87,$8B,$CD,$FD,$16,$6F,$91,$B1,$F7,$67,$CF
	.dc.b	$1E,$1A,$FF,$0E,$E9,$E8,$9E,$C7,$F3,$E3,$F6,$87,$CC,$E3,$FE,$26
	.dc.b	$FF,$53,$C3,$B4,$FE,$32,$32,$7C,$4F,$98,$F1,$E1,$E1,$A4,$78,$72
	.dc.b	$FF,$33,$63,$EB,$BF,$EA,$A7,$DD,$3C,$47,$DD,$75,$9F,$DD,$8E,$F7
	.dc.b	$D3,$1F,$14,$79,$BD,$6F,$F6,$57,$7F,$F1,$A8,$DE,$5F,$0E,$73,$5F
	.dc.b	$96,$9F,$F3,$37,$0F,$E7,$B3,$C6,$5F,$4E,$7F,$9D,$1F,$18,$FC,$F1
	.dc.b	$F4,$A7,$FC,$FC,$FC,$6C,$9F,$25,$F5,$3E,$4B,$EA,$BF,$0D,$F8,$53
	.dc.b	$FE,$CB,$3F,$89,$E1,$E0,$DF,$F3,$B1,$FA,$C3,$E9,$ED,$F0,$A3,$E7
	.dc.b	$C7,$E4,$AB,$F0,$3F,$22,$D3,$F9,$E6,$1F,$E0,$3A,$FF,$05,$E4,$BF
	.dc.b	$A9,$7F,$E7,$CE,$FF,$F5,$47,$3F,$E8,$71,$FC,$26,$BF,$A1,$C7,$EA
	.dc.b	$1F,$00,$FA,$E7,$F9,$0B,$6F,$A8,$FC,$1C,$3C,$4B,$5F,$E3,$4F,$F8
	.dc.b	$14,$FE,$BB,$9B,$E4,$37,$3A,$9F,$F2,$A4,$AB,$3A,$FF,$2C,$DB,$E0
	.dc.b	$CF,$1C,$FF,$C0,$9E,$27,$F8,$27,$0F,$C9,$F3,$7E,$B5,$FF,$47,$E9
	.dc.b	$F6,$DF,$DD,$6B,$ED,$7F,$BA,$DF,$E8,$F0,$E9,$1E,$9C,$BE,$52,$F2
	.dc.b	$5F,$0B,$E4,$8F,$61,$FA,$5D,$F0,$FC,$63,$84,$7C,$FD,$3C,$D1,$FA
	.dc.b	$CD,$3C,$CF,$F0,$ED,$3C,$DC,$3F,$E1,$EE,$1F,$F3,$5C,$FD,$19,$F5
	.dc.b	$1F,$41,$F9,$7F,$21,$E0,$8F,$F2,$D4,$7C,$EF,$8F,$C7,$8B,$F2,$2B
	.dc.b	$5C,$5F,$2A,$7C,$95,$FE,$AE,$CF,$FC,$14,$FF,$01,$3F,$66,$7E,$4C
	.dc.b	$FD,$A6,$FF,$97,$7F,$67,$3D,$9F,$E0,$7C,$5F,$93,$E2,$FC,$9C,$7C
	.dc.b	$63,$F2,$73,$E2,$77,$1E,$27,$FC,$86,$7D,$AE,$EF,$E6,$7A,$FC,$28
	.dc.b	$FE,$67,$B7,$F6,$2E,$EF,$E6,$F7,$E1,$9F,$E2,$FC,$6F,$87,$C4,$7F
	.dc.b	$8B,$33,$F2,$3F,$74,$FE,$57,$3B,$AF,$C5,$C7,$C3,$3C,$4F,$E9,$31
	.dc.b	$F1,$47,$E9,$0F,$DA,$E9,$F0,$DF,$E4,$F6,$F8,$FF,$0C,$7F,$9E,$E1
	.dc.b	$F1,$F8,$8F,$F3,$FC,$6F,$0F,$8F,$1F,$89,$8D,$C7,$92,$3E,$29,$F8
	.dc.b	$9D,$7E,$2F,$8B,$C3,$6F,$C4,$37,$2F,$E9,$27,$96,$BE,$16,$3E,$17
	.dc.b	$FD,$97,$C9,$F8,$87,$CD,$1D,$1F,$F0,$16,$9F,$C1,$5F,$C2,$1F,$E4
	.dc.b	$D3,$F0,$A7,$F6,$4D,$F8,$3F,$E5,$AA,$FF,$35,$37,$8B,$F8,$CE,$AB
	.dc.b	$65,$FD,$2A,$F7,$7F,$E9,$23,$81,$F8,$C9,$DB,$FD,$52,$36,$FF,$33
	.dc.b	$8C,$BC,$0F,$A6,$E1,$E8,$43,$84,$F8,$6F,$CB,$F2,$8F,$82,$C5,$7E
	.dc.b	$72,$30,$F8,$67,$73,$A8,$FC,$0C,$E9,$F1,$1F,$91,$70,$FF,$1B,$3A
	.dc.b	$47,$17,$4D,$7F,$11,$1C,$1F,$EA,$0D,$7F,$0C,$DC,$3F,$82,$47,$09
	.dc.b	$E4,$FE,$3A,$75,$5F,$4B,$E3,$BF,$D9,$DD,$BF,$CC,$70,$FC,$15,$EF
	.dc.b	$0F,$C9,$C6,$21,$9B,$1F,$87,$7F,$ED,$EC,$39,$ED,$ED,$BF,$87,$8F
	.dc.b	$E5,$79,$BE,$3C,$7F,$93,$70,$E3,$5B,$8F,$F5,$0F,$EA,$4F,$63,$9C
	.dc.b	$7C,$0C,$3F,$38,$FC,$83,$FD,$AC,$76,$3F,$78,$7E,$A7,$E2,$38,$BF
	.dc.b	$AB,$F8,$FB,$B7,$AF,$66,$B8,$EC,$F8,$97,$F0,$38,$3F,$2F,$87,$07
	.dc.b	$D3,$C1,$FD,$5F,$95,$C3,$8E,$0F,$F7,$C4,$63,$0F,$35,$7C,$DB,$FE
	.dc.b	$F3,$C1,$ED,$9E,$C6,$AF,$FC,$83,$A7,$CD,$DF,$DF,$3E,$DC,$39,$3F
	.dc.b	$CB,$B7,$F2,$3F,$CC,$3E,$3F,$D4,$FF,$8D,$3F,$7C,$FC,$95,$7D,$DD
	.dc.b	$FF,$C1,$31,$8C,$BE,$85,$7F,$15,$BE,$6F,$78,$F5,$ED,$ED,$8F,$F0
	.dc.b	$C6,$7E,$57,$E7,$F0,$FE,$41,$0D,$7D,$CE,$B6,$E4,$FF,$72,$E3,$FC
	.dc.b	$0B,$FE,$2E,$BE,$9A,$FB,$A3,$F9,$5E,$DE,$47,$F9,$D7,$99,$EE,$7F
	.dc.b	$97,$EB,$5F,$E7,$CE,$DF,$E6,$11,$5F,$CE,$1F,$91,$E0,$FC,$A3,$F4
	.dc.b	$4F,$F3,$6E,$7E,$8C,$3C,$51,$FB,$43,$F1,$E5,$7F,$05,$B7,$B3,$5F
	.dc.b	$85,$3F,$A6,$CB,$C0,$FE,$A3,$83,$CC,$F9,$0F,$D6,$BF,$C9,$32,$DE
	.dc.b	$88,$BA,$FC,$6A,$FF,$2B,$B7,$87,$F1,$D1,$F5,$CF,$4F,$27,$0D,$BF
	.dc.b	$13,$1E,$7D,$3F,$1B,$7C,$F4,$F3,$E4,$D3,$C3,$C1,$E6,$C3,$ED,$F1
	.dc.b	$3E,$0F,$10,$8C,$3F,$9A,$3F,$B0,$CF,$85,$F9,$C4,$E0,$F9,$6F,$E1
	.dc.b	$8F,$C3,$6B,$B8,$F9,$71,$F4,$CF,$C3,$6B,$D5,$FE,$F5,$8F,$C6,$F9
	.dc.b	$0F,$1E,$7F,$E5,$2B,$FC,$F5,$7F,$84,$B1,$D3,$C4,$FE,$48,$78,$F6
	.dc.b	$FC,$EB,$DF,$5F,$B3,$C5,$F8,$19,$FD,$2C,$FE,$22,$FC,$AF,$CF,$3C
	.dc.b	$A7,$E7,$4F,$DA,$1D,$CF,$EF,$BE,$1F,$BF,$CB,$FF,$4A,$47,$F8,$1B
	.dc.b	$3F,$49,$F8,$09,$D7,$77,$82,$7F,$97,$F1,$7D,$FE,$7F,$D4,$F2,$7E
	.dc.b	$02,$7E,$E1,$F8,$03,$C0,$70,$E1,$FC,$0B,$F0,$33,$F4,$BC,$0F,$E4
	.dc.b	$C7,$0E,$0F,$F1,$43,$FE,$AC,$3F,$C5,$19,$79,$1F,$F1,$D1,$FD,$09
	.dc.b	$C3,$FE,$17,$F0,$6F,$07,$E3,$33,$F4,$4F,$DB,$CB,$D0,$F8,$F7,$F7
	.dc.b	$70,$7C,$F8,$FE,$7C,$7E,$B8,$D1,$FC,$16,$BF,$31,$C3,$FE,$0A,$E2
	.dc.b	$FD,$3F,$C8,$78,$39,$FF,$D4,$BE,$5B,$DB,$C6,$3F,$ED,$68,$C3,$FC
	.dc.b	$5B,$9B,$FD,$2E,$FF,$8B,$EF,$D3,$8F,$CE,$E4,$93,$F8,$58,$FE,$73
	.dc.b	$CF,$F9,$7D,$BF,$57,$CB,$F3,$AB,$F0,$D1,$F6,$79,$3F,$59,$1F,$37
	.dc.b	$E4,$3C,$6F,$F9,$1B,$2A,$FF,$95,$5F,$0B,$FE,$6A,$8A,$FE,$3F,$8B
	.dc.b	$E4,$BD,$8F,$E6,$7C,$3F,$11,$FF,$79,$F1,$7C,$B7,$FC,$0B,$97,$95
	.dc.b	$FC,$21,$F3,$4E,$47,$E0,$8F,$CD,$1F,$E5,$AF,$90,$F2,$BF,$E5,$9C
	.dc.b	$2B,$FE,$58,$7C,$AF,$F9,$B6,$2B,$F6,$0F,$E2,$F3,$FC,$B9,$FD,$47
	.dc.b	$27,$EB,$B5,$F9,$51,$F7,$67,$CD,$A7,$F4,$2D,$7E,$57,$27,$F2,$AF
	.dc.b	$5D,$F0,$BE,$2F,$61,$DE,$F5,$B0,$DC,$FE,$F7,$6F,$EA,$F0,$E3,$18
	.dc.b	$3F,$1F,$DB,$FD,$EF,$9F,$EF,$C7,$9B,$9F,$EF,$BF,$A5,$E5,$FE,$61
	.dc.b	$B7,$EC,$8F,$E8,$13,$F9,$39,$FD,$8E,$0F,$57,$C7,$E3,$C9,$F0,$1F
	.dc.b	$EC,$59,$D3,$8E,$4F,$D4,$71,$FC,$78,$FD,$B7,$47,$E0,$1F,$B3,$AF
	.dc.b	$F2,$A7,$FB,$7F,$47,$DF,$BE,$31,$F7,$4D,$E9,$F9,$67,$EE,$9F,$CA
	.dc.b	$8F,$3F,$AB,$F2,$39,$24,$78,$F6,$FE,$6C,$CD,$15,$F7,$A3,$F1,$39
	.dc.b	$F9,$79,$BE,$C4,$7E,$53,$5F,$E2,$7A,$79,$71,$F9,$13,$FC,$0F,$3F
	.dc.b	$92,$FF,$28,$3C,$A7,$F4,$5C,$9E,$AF,$DC,$E0,$93,$E5,$3C,$A7,$F8
	.dc.b	$42,$76,$E4,$FD,$47,$17,$DC,$7F,$90,$6F,$F9,$17,$EE,$67,$F9,$0E
	.dc.b	$2F,$BB,$1F,$BF,$3C,$9C,$3F,$3D,$60,$F5,$7F,$1F,$C3,$27,$C0,$7C
	.dc.b	$78,$F8,$CF,$D2,$3E,$21,$FD,$44,$FE,$70,$7C,$03,$E5,$EB,$F1,$9F
	.dc.b	$C9,$1E,$27,$F3,$5E,$AF,$E7,$B6,$93,$E0,$3E,$44,$78,$38,$7E,$9C
	.dc.b	$7E,$C8,$F0,$0D,$7E,$13,$E1,$3C,$C7,$F5,$CC,$FF,$A3,$BF,$BB,$8F
	.dc.b	$F0,$3F,$4F,$EC,$FD,$5F,$D7,$E7,$2B,$C4,$7D,$38,$F1,$6D,$FD,$0A
	.dc.b	$3E,$99,$F3,$4F,$B3,$C7,$F8,$57,$F0,$84,$3E,$09,$FE,$42,$3F,$5B
	.dc.b	$EA,$7E,$FB,$49,$1E,$03,$EA,$C7,$83,$87,$F5,$51,$FC,$4F,$97,$F2
	.dc.b	$D8,$7F,$27,$8F,$0F,$67,$F1,$27,$E0,$70,$FE,$9F,$7F,$C0,$70,$7F
	.dc.b	$3A,$9F,$8D,$C6,$F4,$FF,$87,$EF,$27,$C5,$39,$7C,$95,$F4,$B3,$F0
	.dc.b	$E9,$E4,$9F,$E6,$3C,$5F,$4B,$2E,$FA,$FE,$4A,$7F,$A0,$69,$FA,$67
	.dc.b	$E1,$1E,$2F,$4F,$F8,$FE,$F2,$3C,$31,$F0,$5F,$BF,$97,$81,$FB,$DC
	.dc.b	$3F,$3D,$EA,$7C,$F3,$F4,$BD,$1F,$C6,$DF,$C2,$79,$5F,$D9,$BC,$3F
	.dc.b	$F0,$CC,$7F,$8E,$7C,$0F,$E6,$3C,$7F,$CC,$0F,$CF,$D6,$FB,$BF,$B1
	.dc.b	$6D,$F8,$1B,$FE,$37,$D7,$EB,$7F,$D7,$FC,$DF,$47,$57,$07,$E8,$67
	.dc.b	$C6,$7E,$84,$FF,$46,$F1,$7F,$9B,$B2,$7A,$5F,$D0,$38,$65,$7E,$0E
	.dc.b	$72,$FD,$2C,$FC,$4F,$E0,$F3,$F1,$3F,$E5,$3E,$1F,$88,$FF,$30,$CB
	.dc.b	$C4,$FE,$87,$87,$E2,$71,$FF,$03,$C5,$E8,$FF,$53,$FC,$0B,$E0,$5F
	.dc.b	$9B,$C9,$FE,$2A,$B9,$13,$A8,$FD,$96,$3A,$7F,$CD,$D7,$F6,$71,$FE
	.dc.b	$A9,$1F,$E2,$9D,$3C,$EF,$F4,$0E,$2F,$91,$A7,$9A,$3F,$17,$87,$9A
	.dc.b	$3F,$C9,$5A,$3D,$2F,$37,$0C,$BF,$8D,$1D,$EB,$FC,$C7,$1F,$B6,$3F
	.dc.b	$62,$7D,$63,$F9,$39,$FD,$37,$A7,$F3,$91,$E4,$8F,$F5,$01,$FE,$52
	.dc.b	$B7,$0F,$C5,$D6,$4F,$D5,$8F,$03,$FC,$93,$5F,$7B,$E0,$3E,$A6,$9E
	.dc.b	$E7,$F2,$C7,$D3,$3F,$08,$7C,$A3,$E7,$DF,$FB,$F1,$FF,$2D,$1F,$ED
	.dc.b	$4D,$FE,$8E,$7F,$5C,$FC,$BF,$A3,$E2,$9F,$E4,$D8,$78,$9F,$C8,$9F
	.dc.b	$C1,$8F,$82,$FF,$24,$7F,$9B,$9F,$C6,$CF,$ED,$87,$EA,$3D,$1F,$C2
	.dc.b	$65,$23,$F2,$91,$DF,$9B,$F6,$95,$DE,$7D,$87,$73,$F9,$A6,$DF,$39
	.dc.b	$F8,$27,$DE,$3E,$0F,$37,$E8,$FD,$1F,$C9,$65,$27,$F4,$51,$DF,$8F
	.dc.b	$FB,$14,$76,$3E,$E1,$DB,$3F,$E2,$0F,$E8,$4F,$61,$FE,$B7,$C3,$F9
	.dc.b	$16,$3F,$A1,$F1,$FF,$41,$A4,$AF,$D8,$47,$77,$F3,$07,$23,$F3,$07
	.dc.b	$CA,$E2,$FE,$5C,$47,$F7,$BC,$9F,$99,$69,$F7,$39,$BF,$5D,$A4,$8F
	.dc.b	$DF,$47,$77,$F3,$07,$D8,$3D,$B8,$FF,$2C,$ED,$FE,$FC,$48,$F9,$BE
	.dc.b	$17,$FB,$C3,$E4,$7F,$CB,$B0,$F5,$7F,$84,$AE,$FF,$E4,$2F,$FB,$66
	.dc.b	$33,$1F,$E8,$AE,$3F,$F7,$5D,$78,$EB,$F8,$F9,$F0,$4F,$F6,$F9,$FE
	.dc.b	$AF,$BF,$F7,$16,$3F,$8B,$8F,$EE,$3C,$9F,$C8,$9E,$1C,$3F,$43,$7F
	.dc.b	$F3,$1C,$7E,$70,$08,$61,$EA,$6C,$7F,$52,$C3,$C8,$FF,$21,$3C,$9D
	.dc.b	$9F,$C8,$1F,$BD,$1E,$4B,$FA,$91,$FD,$90,$FA,$67,$F5,$2C,$7F,$B2
	.dc.b	$FC,$8F,$C8,$FC,$9D,$F0,$14,$FF,$09,$47,$82,$BF,$C8,$FC,$FF,$C1
	.dc.b	$F2,$F0,$47,$AB,$35,$78,$63,$DB,$AF,$C3,$7C,$19,$FC,$05,$19,$7F
	.dc.b	$63,$9F,$91,$FE,$25,$C9,$FC,$D7,$96,$47,$F4,$58,$F2,$3F,$43,$3F
	.dc.b	$27,$67,$86,$3F,$86,$1F,$E6,$2E,$B7,$E1,$0F,$67,$59,$F8,$6B,$F8
	.dc.b	$74,$E5,$FF,$62,$57,$0A,$E9,$C8,$CB,$FA,$2C,$F4,$F8,$5C,$C7,$F4
	.dc.b	$3E,$07,$C9,$7F,$4D,$EA,$99,$7F,$83,$23,$D3,$3F,$E0,$DD,$BD,$33
	.dc.b	$FE,$FB,$9F,$2F,$0F,$6F,$0F,$C1,$3E,$46,$BF,$C3,$A7,$C2,$FF,$B2
	.dc.b	$BA,$FF,$AE,$7B,$32,$3F,$C5,$DC,$C9,$F0,$F7,$FF,$6C,$9F,$E2,$87
	.dc.b	$F1,$9E,$6F,$F3,$45,$6E,$3F,$D7,$B8,$79,$1F,$EC,$BA,$7B,$9F,$F4
	.dc.b	$3E,$1E,$27,$FC,$8B,$F2,$DF,$DB,$7D,$A9,$1E,$28,$F0,$CF,$F8,$AE
	.dc.b	$3F,$CB,$3C,$7F,$E2,$BD,$BF,$0C,$FB,$A3,$FD,$37,$92,$FC,$51,$FE
	.dc.b	$52,$BA,$E9,$F1,$3C,$22,$BF,$E0,$7D,$FF,$C0,$1E,$D4,$8F,$8B,$1E
	.dc.b	$47,$FC,$DF,$87,$8D,$FD,$49,$E2,$8E,$EF,$F5,$5E,$07,$B1,$88,$FF
	.dc.b	$04,$46,$1F,$F5,$5C,$F8,$9F,$F6,$0E,$7F,$C6,$3E,$31,$F0,$B4,$FF
	.dc.b	$48,$EB,$FA,$9F,$0B,$F8,$B7,$37,$F4,$2C,$3F,$66,$7F,$02,$EC,$E8
	.dc.b	$AB,$F6,$1C,$9F,$C5,$56,$7E,$59,$D3,$C2,$37,$9F,$D0,$C3,$FC,$39
	.dc.b	$0F,$92,$FE,$5A,$F8,$FC,$DF,$AB,$D3,$AB,$F3,$D7,$C9,$CB,$FB,$26
	.dc.b	$5F,$8C,$3F,$9D,$CF,$57,$FC,$5B,$87,$81,$FE,$0F,$1C,$DF,$E4,$B3
	.dc.b	$E9,$8F,$37,$C8,$FC,$4F,$92,$FE,$7F,$E3,$FF,$8E,$AF,$F9,$61,$FD
	.dc.b	$07,$CD,$D3,$FE,$49,$E8,$FC,$77,$57,$EB,$E3,$E4,$1F,$ED,$EC,$3F
	.dc.b	$8A,$71,$37,$FE,$2D,$E1,$7F,$05,$AF,$AE,$79,$34,$C9,$E7,$7F,$CE
	.dc.b	$B7,$F6,$0F,$8F,$A4,$8F,$D8,$C7,$C5,$9F,$F1,$BC,$7F,$5F,$3F,$C4
	.dc.b	$E7,$8C,$FE,$8B,$87,$8D,$F8,$A7,$C8,$8F,$B5,$B7,$C9,$8F,$E3,$E7
	.dc.b	$E9,$AB,$F1,$3F,$25,$FD,$93,$F7,$C7,$B7,$4C,$D7,$C5,$3F,$D1,$FC
	.dc.b	$BF,$C4,$9F,$F4,$7D,$7C,$4D,$7D,$F1,$F6,$30,$C3,$E4,$3F,$D8,$6F
	.dc.b	$47,$FB,$84,$7D,$F7,$FD,$8F,$F2,$5E,$3F,$72,$47,$F1,$48,$F8,$8F
	.dc.b	$F9,$2F,$4E,$0F,$E2,$0F,$D8,$9F,$E7,$F8,$F0,$BF,$D2,$A3,$FC,$E5
	.dc.b	$E5,$BE,$1B,$3C,$3F,$66,$93,$97,$C0,$7F,$24,$F4,$FF,$B1,$7C,$54
	.dc.b	$8C,$1F,$17,$A3,$F3,$6F,$F4,$F3,$F1,$27,$FA,$D2,$7B,$E3,$FA,$8C
	.dc.b	$AC,$FE,$79,$FD,$DC,$FE,$F3,$E5,$BF,$D2,$1F,$07,$2B,$F9,$AC,$7C
	.dc.b	$27,$FC,$B7,$7F,$09,$FC,$29,$FE,$C2,$8E,$AF,$FB,$CA,$AB,$FC,$99
	.dc.b	$EC,$FF,$AA,$FC,$9F,$F4,$F7,$C1,$C8,$C1,$ED,$BF,$C9,$BF,$E0,$D3
	.dc.b	$FC,$F4,$7F,$90,$06,$5F,$DE,$FC,$5F,$E6,$91,$FE,$75,$E1,$7C,$5F
	.dc.b	$8A,$30,$D7,$FD,$55,$1A,$7F,$3A,$FB,$5C,$FC,$5F,$FB,$4B,$CA,$FF
	.dc.b	$5A,$7A,$B2,$7F,$D8,$91,$F1,$78,$FF,$CE,$39,$F1,$9F,$F1,$A9,$E6
	.dc.b	$7A,$79,$FF,$BB,$F8,$5F,$13,$67,$C2,$7E,$64,$FF,$72,$64,$9F,$C6
	.dc.b	$E9,$EE,$9F,$4F,$C2,$FF,$B9,$BC,$DF,$F6,$47,$AF,$27,$FD,$B1,$1E
	.dc.b	$5E,$DF,$D5,$70,$FC,$7F,$27,$FC,$0F,$9B,$B3,$C6,$DC,$3C,$67,$E6
	.dc.b	$CF,$F1,$31,$F3,$63,$C7,$3C,$BE,$17,$FD,$EF,$E7,$FF,$B6,$7D,$79
	.dc.b	$7F,$EE,$98,$F1,$3F,$96,$C7,$C4,$F4,$3E,$9F,$C2,$78,$BE,$12,$47
	.dc.b	$8B,$83,$E0,$71,$7F,$23,$AF,$0C,$FF,$AD,$BB,$7F,$DD,$1E,$A4,$8F
	.dc.b	$F7,$C4,$78,$1F,$C4,$9F,$53,$1F,$0B,$FE,$33,$DD,$F0,$5F,$08,$91
	.dc.b	$E1,$DF,$E3,$FC,$17,$F6,$2F,$93,$FA,$7E,$97,$F8,$C3,$F0,$4F,$F0
	.dc.b	$A5,$FF,$0E,$3F,$DF,$07,$B7,$64,$6B,$CD,$FE,$D3,$3F,$D4,$2F,$FB
	.dc.b	$6F,$A9,$FE,$42,$E7,$FE,$D9,$3F,$E1,$2C,$FF,$B0,$CF,$F6,$5A,$FC
	.dc.b	$4E,$1F,$E1,$1A,$FA,$FD,$BF,$D1,$72,$EF,$EC,$79,$E7,$FC,$5F,$F2
	.dc.b	$9F,$E9,$4F,$96,$34,$FE,$FA,$78,$A3,$FA,$AF,$B7,$FE,$33,$7C,$47
	.dc.b	$F6,$6C,$FF,$BB,$1F,$AB,$7F,$BF,$67,$7D,$23,$CE,$3E,$21,$FD,$10
	.dc.b	$FE,$13,$BF,$FA,$89,$FF,$48,$CF,$91,$FE,$D2,$D7,$8F,$47,$E8,$5F
	.dc.b	$D6,$CF,$CA,$7F,$C8,$DE,$17,$F0,$87,$FC,$26,$7E,$02,$BF,$D0,$AF
	.dc.b	$0E,$5F,$E9,$2F,$E4,$63,$CE,$FC,$A9,$FF,$07,$E1,$E9,$BE,$D3,$FE
	.dc.b	$40,$F9,$5F,$F4,$B7,$B5,$23,$FC,$B7,$1F,$E2,$63,$ED,$3F,$1C,$FF
	.dc.b	$26,$61,$E2,$11,$FE,$1F,$3E,$DF,$24,$9E,$67,$F9,$9B,$E5,$7F,$1E
	.dc.b	$FE,$2F,$93,$FC,$40,$FE,$38,$FF,$46,$38,$7E,$71,$F3,$23,$FC,$72
	.dc.b	$79,$1D,$FB,$75,$FF,$51,$72,$7E,$5A,$3F,$C6,$53,$27,$E2,$1E,$1F
	.dc.b	$5F,$F2,$27,$47,$FC,$F7,$EF,$F9,$9F,$1F,$AB,$5E,$B8,$A3,$FC,$97
	.dc.b	$32,$7E,$2D,$F8,$DF,$F5,$A6,$88,$FF,$22,$4C,$9F,$19,$F0,$3D,$6F
	.dc.b	$D1,$9C,$DF,$C3,$F5,$7F,$92,$3D,$FF,$14,$7F,$B7,$B5,$FD,$9F,$A9
	.dc.b	$88,$46,$1F,$F4,$AC,$FF,$98,$0F,$23,$F9,$6C,$2B,$FE,$5D,$7F,$CB
	.dc.b	$11,$23,$CA,$7C,$0F,$4F,$F6,$E7,$37,$FD,$43,$EE,$F8,$6F,$F8,$2F
	.dc.b	$AB,$77,$3F,$E7,$09,$95,$FE,$3F,$3F,$DB,$7C,$5F,$C4,$79,$BC,$B8
	.dc.b	$7F,$0E,$7F,$CD,$55,$2B,$C6,$7C,$1C,$3F,$59,$CD,$FE,$3B,$8F,$E7
	.dc.b	$C7,$E2,$BB,$3B,$BF,$28,$FF,$32,$FC,$6F,$F0,$9E,$99,$1F,$EB,$A8
	.dc.b	$FF,$42,$9F,$13,$91,$8D,$3F,$CF,$91,$27,$C6,$7D,$0C,$5D,$9F,$C7
	.dc.b	$0E,$71,$FE,$C6,$F9,$3F,$F3,$FD,$48,$C1,$E1,$3F,$5A,$F8,$4F,$F4
	.dc.b	$8F,$EA,$9F,$E2,$D8,$F5,$1F,$BF,$8F,$E2,$C4,$7F,$9F,$1F,$D4,$C4
	.dc.b	$7F,$60,$D3,$F8,$0C,$09,$19,$ED,$FF,$48,$C7,$ED,$7B,$3E,$D1,$FD
	.dc.b	$59,$FE,$BD,$3B,$7F,$C9,$27,$F2,$0B,$F4,$3E,$A3,$EF,$1F,$BA,$3C
	.dc.b	$7E,$DF,$37,$FB,$9C,$F3,$7E,$C9,$FE,$D4,$9F,$63,$F0,$0D,$9C,$4F
	.dc.b	$51,$EE,$F9,$1F,$C8,$75,$C8,$F8,$67,$A2,$38,$3F,$02,$FF,$CB,$7C
	.dc.b	$9F,$EB,$4B,$F4,$CF,$F8,$7C,$F6,$9E,$83,$99,$E8,$AF,$EB,$3D,$12
	.dc.b	$32,$79,$31,$4F,$82,$3E,$9F,$C3,$7F,$59,$DE,$47,$03,$E9,$C6,$EB
	.dc.b	$F0,$1B,$7E,$77,$C5,$FF,$67,$D7,$E6,$8F,$9C,$79,$CF,$9D,$3C,$37
	.dc.b	$FF,$1D,$36,$3F,$C6,$87,$97,$1F,$EC,$11,$E6,$8F,$11,$EF,$3E,$81
	.dc.b	$EB,$D7,$F8,$73,$EA,$3F,$D8,$B0,$C7,$D1,$32,$7F,$BD,$EB,$F3,$5F
	.dc.b	$3F,$93,$FE,$18,$7F,$03,$3B,$7F,$B6,$E5,$FE,$8D,$C3,$C9,$18,$E5
	.dc.b	$FC,$E3,$F7,$CF,$3C,$7B,$5F,$83,$F2,$9F,$EB,$5F,$90,$91,$A8,$F9
	.dc.b	$87,$E0,$1F,$D3,$4F,$1B,$F9,$B1,$F9,$EC,$36,$F1,$3F,$7F,$D5,$FE
	.dc.b	$22,$ED,$FF,$0C,$9A,$8F,$8E,$7E,$B3,$DE,$FF,$63,$79,$F2,$3C,$F1
	.dc.b	$F7,$4F,$13,$FE,$49,$F7,$FF,$CF,$4F,$C2,$DF,$FD,$63,$EE,$EE,$38
	.dc.b	$9E,$2E,$BF,$E2,$AD,$EC,$D3,$E2,$3F,$80,$F9,$1F,$63,$07,$CA,$53
	.dc.b	$F1,$D1,$27,$EB,$9F,$13,$2F,$2B,$FB,$67,$4F,$5E,$04,$FF,$8E,$BA
	.dc.b	$E4,$FD,$48,$FC,$D6,$7F,$D0,$FB,$A4,$FD,$88,$CB,$E0,$3F,$86,$CF
	.dc.b	$81,$FD,$53,$A7,$C1,$1F,$E9,$0E,$2F,$E1,$07,$F9,$1C,$FE,$15,$E5
	.dc.b	$7F,$8B,$71,$F9,$7D,$7F,$C8,$2F,$F9,$07,$88,$F7,$FF,$AC,$61,$F8
	.dc.b	$EA,$FF,$1D,$72,$A3,$1D,$BF,$EA,$3B,$E1,$87,$ED,$EB,$FD,$73,$E5
	.dc.b	$79,$E3,$FB,$8F,$67,$F8,$15,$F6,$1F,$E1,$47,$0F,$E1,$A7,$8B,$CE
	.dc.b	$1E,$4F,$FA,$B1,$79,$14,$FE,$33,$5F,$EC,$89,$FE,$0D,$EF,$FA,$BC
	.dc.b	$09,$87,$E9,$23,$2F,$88,$FF,$04,$69,$F9,$FE,$AF,$E8,$D8,$7B,$FB
	.dc.b	$77,$5F,$76,$FE,$77,$74,$8E,$87,$EC,$3A,$7E,$4F,$4C,$8C,$10,$FC
	.dc.b	$D1,$FE,$89,$CB,$F0,$EF,$8C,$FF,$0B,$B8,$7F,$19,$3F,$C8,$4F,$BD
	.dc.b	$CD,$23,$A7,$1F,$DB,$7F,$AE,$FB,$BF,$6B,$29,$1F,$C1,$23,$F9,$51
	.dc.b	$F5,$74,$FB,$98,$5C,$7F,$47,$5B,$21,$FC,$3C,$F4,$57,$E3,$CE,$6F
	.dc.b	$F1,$0F,$77,$F1,$24,$70,$43,$E0,$9F,$ED,$0B,$FC,$67,$C0,$F8,$4F
	.dc.b	$E7,$D3,$27,$F8,$6D,$FF,$B6,$5F,$11,$C5,$C6,$5F,$07,$A1,$5F,$D1
	.dc.b	$EF,$F3,$7E,$DC,$C3,$F9,$2C,$65,$FA,$18,$FE,$77,$C3,$F8,$46,$63
	.dc.b	$F7,$93,$FA,$DF,$0A,$47,$23,$F9,$7F,$B9,$FA,$8F,$6A,$46,$08,$7C
	.dc.b	$03,$FC,$E7,$AB,$E4,$7E,$81,$FD,$76,$A4,$FC,$0F,$67,$F7,$86,$1F
	.dc.b	$E1,$DA,$7E,$DB,$D9,$91,$FD,$02,$3F,$B3,$1F,$6F,$E4,$24,$64,$87
	.dc.b	$CA,$3F,$D0,$BC,$7F,$E4,$7E,$2F,$ED,$9E,$57,$E8,$A3,$1E,$0F,$F6
	.dc.b	$8C,$3F,$B9,$70,$FF,$97,$65,$F2,$14,$CE,$3F,$D4,$A7,$F1,$D2,$F8
	.dc.b	$8A,$6A,$7F,$94,$4F,$F8,$63,$2F,$0C,$7F,$BB,$69,$F2,$19,$CB,$FD
	.dc.b	$7B,$3E,$88,$E5,$5F,$81,$E3,$F2,$7B,$52,$BF,$B1,$C7,$F8,$24,$F1
	.dc.b	$FC,$84,$8C,$90,$F8,$87,$FA,$2F,$17,$C8,$78,$63,$FC,$0E,$7F,$A3
	.dc.b	$23,$2F,$1C,$3F,$25,$CD,$F1,$7E,$42,$4F,$F7,$48,$FF,$0D,$1F,$13
	.dc.b	$E4,$24,$64,$87,$88,$ED,$F0,$3E,$17,$FC,$86,$79,$5E,$6F,$F8,$03
	.dc.b	$A3,$EA,$7B,$B2,$30,$43,$F2,$47,$B7,$E2,$BE,$A9,$FE,$2D,$99,$1F
	.dc.b	$DF,$33,$7C,$4F,$87,$B1,$D1,$F8,$58,$DD,$7F,$61,$EB,$FC,$27,$C4
	.dc.b	$48,$FF,$09,$C7,$F8,$E4,$FC,$1F,$C3,$C8,$C9,$0F,$81,$E0,$7F,$74
	.dc.b	$F8,$6F,$06,$3F,$93,$79,$4F,$F8,$7B,$BF,$F2,$5E,$8C,$8C,$10,$F7
	.dc.b	$9F,$5F,$E2,$BF,$26,$7F,$92,$E6,$47,$BF,$E2,$BB,$BF,$91,$72,$FF
	.dc.b	$82,$7D,$35,$7E,$87,$E2,$A4,$FF,$91,$A7,$2F,$F6,$3F,$77,$F0,$AF
	.dc.b	$8B,$4F,$F3,$0F,$5F,$F3,$19,$CF,$5F,$F7,$39,$FF,$32,$F5,$FF,$5C
	.dc.b	$97,$57,$93,$0F,$E5,$11,$87,$FC,$62,$D7,$FB,$52,$EA,$FF,$34,$78
	.dc.b	$BF,$B7,$F8,$A9,$5F,$E4,$68,$CB,$FC,$8B,$F0,$47,$C7,$7F,$45,$6C
	.dc.b	$6C,$87,$D6,$5F,$19,$E5,$8F,$F4,$8E,$89,$CC,$70,$D7,$FA,$B7,$37
	.dc.b	$F0,$C8,$FF,$58,$7B,$5E,$77,$FD,$32,$7F,$BE,$EB,$FC,$A3,$5F,$E1
	.dc.b	$26,$F8,$DF,$E9,$EA,$78,$3F,$E6,$2C,$3A,$61,$FE,$AD,$6F,$77,$F9
	.dc.b	$6A,$3C,$B8,$7E,$69,$F8,$13,$FE,$93,$38,$DF,$48,$FF,$64,$47,$23
	.dc.b	$FC,$0B,$2F,$8B,$F8,$13,$E7,$79,$63,$CF,$CA,$FD,$C6,$1E,$DA,$FE
	.dc.b	$C1,$6F,$6F,$3C,$5D,$1E,$13,$FE,$90,$CB,$F9,$6B,$F5,$7E,$2F,$EA
	.dc.b	$1A,$7F,$CB,$91,$A7,$EA,$7B,$5E,$45,$C1,$F4,$E3,$87,$07,$D2,$F3
	.dc.b	$7C,$35,$F4,$0D,$E7,$FE,$2E,$7F,$DC,$93,$F4,$7D,$BF,$8D,$52,$7F
	.dc.b	$11,$D7,$F8,$73,$0F,$92,$3C,$58,$FE,$0F,$E2,$EE,$27,$78,$FE,$13
	.dc.b	$CE,$F1,$E1,$F1,$BE,$2E,$BF,$61,$FA,$D1,$23,$F3,$1F,$17,$F9,$73
	.dc.b	$0F,$C4,$8E,$FF,$0F,$23,$E6,$47,$DF,$F4,$7A,$C7,$E3,$FE,$2A,$BF
	.dc.b	$09,$D3,$F4,$E3,$F4,$9F,$15,$23,$04,$3F,$49,$F1,$5E,$E3,$8B,$F5
	.dc.b	$BD,$3E,$BB,$FE,$97,$E2,$AF,$01,$FC,$94,$CA,$FD,$A7,$C4,$FE,$CC
	.dc.b	$C3,$CC,$FF,$58,$D3,$E2,$3E,$69,$AE,$1F,$D7,$F9,$1D,$BE,$22,$5F
	.dc.b	$E3,$63,$F4,$07,$F9,$0F,$E5,$FF,$D1,$3F,$0D,$FE,$62,$7F,$D1,$3E
	.dc.b	$27,$EB,$67,$75,$FE,$C0,$9F,$F4,$1B,$FD,$93,$64,$62,$3F,$65,$3F
	.dc.b	$AD,$F2,$FE,$0C,$FF,$B9,$67,$F1,$BE,$1F,$EE,$FC,$1F,$07,$C3,$EE
	.dc.b	$7F,$45,$7E,$5F,$87,$91,$82,$1F,$D2,$30,$7C,$3F,$E4,$1B,$1A,$DF
	.dc.b	$F8,$8F,$9D,$E2,$F8,$79,$3F,$AF,$8F,$E0,$47,$C3,$F8,$99,$19,$21
	.dc.b	$DF,$E2,$3B,$1C,$1F,$DC,$6B,$EE,$F8,$89,$18,$21,$FD,$0F,$2F,$F3
	.dc.b	$7B,$7D,$DF,$E2,$55,$23,$E7,$7C,$47,$AC,$C3,$FC,$03,$93,$E5,$FC
	.dc.b	$3C,$8F,$E1,$11,$FC,$6C,$FB,$9F,$0F,$23,$24,$3E,$47,$C3,$FC,$73
	.dc.b	$83,$FC,$47,$E0,$FE,$BF,$C3,$C8,$C1,$0F,$57,$C3,$FE,$24,$FE,$55
	.dc.b	$32,$3F,$0D,$F0,$FF,$4C,$C3,$FC,$63,$E1,$7F,$07,$F0,$F2,$3F,$90
	.dc.b	$47,$F3,$33,$F2,$9F,$0F,$23,$24,$3F,$01,$F0,$DF,$7C,$E0,$FF,$29
	.dc.b	$F8,$7F,$C7,$FC,$24,$8C,$10,$FA,$5E,$FF,$E8,$4F,$E8,$53,$23,$F2
	.dc.b	$A7,$E7,$7E,$4F,$FA,$13,$FC,$9F,$C3,$47,$F4,$8E,$2F,$EA,$55,$FD
	.dc.b	$2F,$E2,$3F,$7F,$3F,$CD,$7B,$BF,$DD,$6D,$F0,$FF,$AD,$5F,$F5,$9A
	.dc.b	$E1,$8F,$FA,$52,$BF,$CD,$0D,$F5,$72,$FF,$40,$9F,$E5,$B3,$FE,$66
	.dc.b	$8F,$39,$FE,$46,$3F,$A0,$B9,$78,$1F,$D3,$7E,$23,$CB,$E8,$48,$C1
	.dc.b	$0F,$F0,$C4,$BC,$DF,$F6,$43,$63,$59,$FF,$7C,$BF,$1F,$91,$27,$3A
	.dc.b	$7B,$9F,$ED,$7F,$11,$F1,$3C,$89,$1F,$D7,$E3,$FB,$91,$DB,$C5,$91
	.dc.b	$92,$1F,$E4,$6C,$BF,$B6,$65,$FE,$88,$8F,$F6,$64,$63,$0F,$04,$7F
	.dc.b	$5D,$F6,$D8,$7F,$7F,$F8,$7F,$AB,$C3,$2B,$FB,$84,$65,$FA,$5D,$7F
	.dc.b	$00,$C3,$FD,$C2,$3E,$8F,$4C,$8E,$27,$F7,$9F,$88,$FC,$0F,$2C,$8C
	.dc.b	$10,$F9,$BC,$9F,$30,$FF,$09,$4C,$8F,$B7,$F0,$7F,$E1,$77,$DB,$C7
	.dc.b	$F8,$03,$4B,$81,$BA,$27,$D1,$7F,$4D,$7F,$31,$C3,$F0,$9F,$C3,$C6
	.dc.b	$5F,$93,$C1,$FE,$8F,$8C,$47,$F8,$F2,$78,$3D,$EB,$F9,$C3,$FE,$26
	.dc.b	$F8,$9F,$F1,$51,$F8,$6C,$E4,$7F,$86,$E3,$2F,$B3,$8B,$FC,$20,$FE
	.dc.b	$5A,$3F,$36,$69,$EC,$7B,$7E,$91,$ED,$F0,$FA,$57,$D2,$FD,$FA,$D4
	.dc.b	$E6,$31,$B2,$F8,$D7,$F8,$7E,$F8,$F9,$5F,$EB,$DA,$D7,$B9,$C7,$A3
	.dc.b	$FC,$89,$F1,$7F,$EA,$1F,$5B,$CF,$E0,$BE,$0B,$FC,$BB,$9F,$F9,$F5
	.dc.b	$B8,$FF,$97,$AF,$8C,$FB,$CF,$E1,$13,$FE,$5A,$7F,$B4,$35,$FF,$2C
	.dc.b	$47,$8F,$47,$C0,$FA,$FB,$70,$DA,$FE,$2B,$87,$B6,$31,$1F,$22,$3E
	.dc.b	$07,$C0,$FC,$3E,$DF,$F2,$9E,$BF,$42,$9F,$AA,$7F,$9F,$9F,$57,$C3
	.dc.b	$FB,$DF,$3B,$C1,$FA,$F1,$BC,$3F,$1F,$1F,$E7,$AA,$E9,$39,$E8,$FF
	.dc.b	$7B,$CF,$E4,$F7,$F9,$1F,$0F,$FE,$44,$E0,$3D,$67,$FC,$A5,$7C,$5F
	.dc.b	$C0,$4F,$DB,$F0,$3E,$23,$63,$E0,$F0,$7D,$37,$EE,$73,$2B,$E5,$7C
	.dc.b	$2F,$E0,$FC,$1E,$D5,$FE,$8E,$C7,$EB,$E9,$FA,$78,$78,$3F,$E8,$77
	.dc.b	$D1,$9F,$57,$EF,$3C,$9E,$15,$C9,$F8,$47,$C6,$F8,$BF,$76,$32,$3F
	.dc.b	$9A,$FC,$4F,$FA,$35,$F7,$CF,$DF,$F8,$9F,$CC,$78,$96,$7D,$6D,$7F
	.dc.b	$D4,$BD,$BF,$A0,$8F,$C3,$7C,$4F,$E7,$FC,$5F,$25,$7F,$01,$CF,$F1
	.dc.b	$9A,$FA,$E3,$FD,$07,$AF,$E8,$9F,$CC,$78,$1C,$1F,$C1,$F7,$5B,$E3
	.dc.b	$E1,$7E,$82,$DE,$07,$D9,$35,$5F,$E8,$FD,$7F,$67,$E3,$5E,$6F,$F7
	.dc.b	$1E,$5F,$BA,$3F,$AA,$4E,$9F,$B3,$D1,$FB,$7F,$13,$C1,$9F,$E4,$67
	.dc.b	$FD,$87,$D5,$FA,$FF,$86,$C4,$7F,$B3,$B5,$FE,$31,$DB,$F9,$28,$73
	.dc.b	$FC,$D7,$EF,$61,$FE,$A3,$7F,$3E,$FE,$47,$E1,$A4,$7D,$4E,$EF,$CC
	.dc.b	$CF,$EC,$4F,$E1,$5E,$35,$9F,$C6,$B5,$FA,$7B,$A3,$F7,$1F,$17,$FC
	.dc.b	$47,$C6,$F9,$4F,$A2,$BF,$45,$E4,$FF,$9B,$9F,$F0,$53,$FE,$6E,$7F
	.dc.b	$5B,$1D,$5F,$CA,$7D,$03,$F5,$37,$6E,$67,$9F,$C3,$57,$FD,$7F,$87
	.dc.b	$F9,$B5,$FE,$05,$E2,$FF,$B3,$DF,$F5,$F6,$1C,$EF,$F8,$77,$C7,$7F
	.dc.b	$2B,$FE,$01,$FD,$87,$1F,$F2,$78,$C1,$AB,$FE,$51,$3A,$C3,$F9,$0C
	.dc.b	$7E,$A7,$A5,$3F,$A3,$8F,$A7,$C5,$FC,$13,$C8,$F8,$51,$E6,$F8,$6B
	.dc.b	$D2,$D9,$1F,$91,$AF,$05,$7E,$83,$0F,$E5,$1A,$79,$63,$C7,$CB,$A7
	.dc.b	$C5,$3F,$D9,$E7,$F8,$A9,$FD,$07,$B7,$DC,$7F,$3F,$CD,$D7,$F0,$79
	.dc.b	$30,$DC,$3F,$96,$CE,$9F,$83,$3C,$79,$BE,$46,$7F,$A7,$EC,$F5,$EC
	.dc.b	$F8,$1A,$67,$0E,$79,$7C,$FA,$EF,$1F,$33,$C5,$AF,$BD,$F3,$74,$EE
	.dc.b	$3C,$7C,$DF,$5F,$C5,$F5,$36,$32,$7D,$68,$F7,$78,$BF,$77,$BF,$F7
	.dc.b	$98,$79,$BD,$D9,$7F,$A1,$3F,$9E,$C7,$47,$EF,$F9,$15,$FA,$5E,$75
	.dc.b	$7E,$7B,$F2,$FC,$4F,$C5,$4F,$CD,$F3,$7F,$19,$E7,$5B,$FC,$4B,$5F
	.dc.b	$B8,$FD,$3F,$3B,$F2,$5E,$57,$F1,$1D,$BE,$5F,$B3,$2F,$CC,$7F,$46
	.dc.b	$8E,$4F,$E0,$3C,$AA,$FF,$88,$7C,$1F,$C9,$FC,$6E,$21,$18,$7E,$EF
	.dc.b	$A1,$FA,$1A,$FC,$27,$A5,$FA,$3F,$8D,$B5,$FA,$06,$BF,$93,$7F,$15
	.dc.b	$DF,$FA,$9F,$3B,$F9,$47,$37,$E7,$3F,$05,$FF,$38,$78,$7F,$C5,$7D
	.dc.b	$1F,$E6,$5E,$42,$73,$E9,$FF,$95,$B3,$FE,$A9,$87,$67,$F3,$7E,$85
	.dc.b	$7D,$9E,$47,$FB,$FD,$FC,$AE,$DF,$B9,$F8,$DF,$CE,$7A,$BF,$BB,$BF
	.dc.b	$CF,$7B,$3F,$BD,$F8,$DB,$87,$EE,$9A,$FE,$B5,$FD,$27,$B5,$FE,$2D
	.dc.b	$FE,$01,$FE,$29,$3F,$DD,$3B,$FD,$9E,$DC,$8F,$A6,$7F,$6B,$8E,$6F
	.dc.b	$F0,$39,$7C,$96,$23,$EA,$47,$F1,$5E,$4F,$DA,$F3,$33,$F0,$7A,$DF
	.dc.b	$B5,$CB,$C5,$8F,$ED,$FD,$FF,$27,$C7,$FE,$E7,$E4,$6E,$9F,$B6,$6B
	.dc.b	$FE,$21,$CB,$FD,$52,$FF,$68,$D1,$F2,$1F,$D9,$BA,$5F,$B0,$BC,$3F
	.dc.b	$E2,$6D,$BF,$F3,$8F,$AD,$FE,$40,$8C,$47,$F5,$BF,$7B,$FA,$73,$FE
	.dc.b	$1D,$87,$C2,$7F,$0D,$F8,$4F,$95,$ED,$FF,$0F,$F8,$4B,$7F,$C3,$1A
	.dc.b	$F9,$E7,$E8,$7C,$8F,$F9,$4F,$EC,$39,$CF,$F9,$5F,$77,$C9,$67,$2F
	.dc.b	$AC,$D9,$D3,$FC,$77,$C3,$FB,$5F,$25,$FD,$EF,$D3,$FC,$A7,$A7,$2F
	.dc.b	$EC,$9F,$E1,$A8,$E4,$FC,$7F,$91,$AF,$DB,$F7,$2B,$F6,$9F,$E5,$5F
	.dc.b	$29,$F8,$79,$F4,$7C,$67,$E2,$7E,$4A,$DF,$A5,$AF,$D3,$7D,$DF,$1B
	.dc.b	$F8,$FF,$92,$FF,$08,$7B,$3F,$D1,$3F,$07,$FE,$95,$E8,$FE,$A1,$FE
	.dc.b	$99,$D5,$F4,$31,$FF,$0F,$57,$47,$F2,$1F,$27,$5F,$C7,$FC,$9D,$7F
	.dc.b	$80,$F3,$23,$E7,$77,$7E,$7E,$BE,$87,$BF,$FA,$12,$DF,$D1,$6B,$F8
	.dc.b	$B7,$EB,$7C,$9F,$E9,$CF,$F1,$47,$BB,$E9,$F8,$49,$5F,$38,$FF,$1A
	.dc.b	$47,$27,$F4,$DF,$27,$5F,$D2,$F1,$D7,$CE,$FE,$0E,$3F,$68,$7E,$17
	.dc.b	$87,$F6,$DC,$5F,$87,$CE,$DF,$FA,$13,$B5,$18,$8F,$F4,$24,$2B,$F7
	.dc.b	$DA,$FF,$8F,$FD,$BF,$A1,$D3,$2F,$D0,$7F,$A1,$A3,$B3,$F9,$3E,$EA
	.dc.b	$FF,$95,$3A,$FF,$C6,$5E,$7F,$F7,$0C,$BC,$33,$FE,$AF,$9C,$F7,$7F
	.dc.b	$B5,$69,$5F,$4E,$BA,$1E,$9E,$EF,$F5,$F5,$BB,$7F,$49,$DB,$86,$D7
	.dc.b	$57,$F5,$15,$C3,$CB,$FF,$2D,$7B,$DF,$8B,$F7,$25,$FB,$3E,$45,$3F
	.dc.b	$AF,$F0,$BB,$F9,$B6,$BD,$1E,$35,$7D,$CF,$ED,$7C,$4D,$C7,$CF,$F3
	.dc.b	$3C,$FE,$47,$EE,$7C,$9B,$3F,$4D,$AF,$D2,$9F,$6F,$A5,$FE,$6F,$D7
	.dc.b	$FC,$47,$EC,$FF,$91,$6B,$FC,$C2,$FF,$32,$7F,$D3,$11,$FC,$17,$CB
	.dc.b	$F8,$BE,$AD,$9F,$9D,$E9,$D7,$EA,$3F,$C2,$FC,$DD,$C7,$E1,$4F,$9B
	.dc.b	$87,$F0,$FF,$4E,$C7,$E2,$5A,$FE,$20,$FA,$7E,$97,$FA,$17,$DD,$F3
	.dc.b	$ED,$27,$E6,$47,$F1,$9F,$57,$EC,$FB,$36,$7F,$05,$F0,$55,$FC,$43
	.dc.b	$FC,$7F,$D7,$DC,$7F,$B7,$3E,$D5,$C3,$F8,$2F,$83,$FF,$7B,$43,$2C
	.dc.b	$56,$DA,$FF,$B8,$63,$8B,$F9,$1F,$81,$E3,$EE,$FF,$BE,$B8,$3F,$C6
	.dc.b	$D3,$DE,$31,$D7,$A8,$FF,$62,$1F,$ED,$99,$DB,$FE,$9B,$F6,$3F,$D5
	.dc.b	$8F,$FB,$AA,$BC,$BE,$F7,$E6,$3E,$1A,$CF,$E8,$7E,$1A,$BF,$A7,$7D
	.dc.b	$7F,$01,$B8,$F8,$1F,$01,$FA,$0F,$23,$DF,$E7,$59,$F8,$8D,$7D,$D3
	.dc.b	$FA,$6F,$88,$F8,$BE,$E7,$EA,$3E,$2E,$D7,$EC,$5A,$FE,$DD,$F9,$7E
	.dc.b	$C7,$EB,$BE,$27,$E6,$FA,$7F,$02,$BF,$C5,$5F,$C0,$3F,$E8,$4E,$6F
	.dc.b	$F6,$0C,$67,$E2,$13,$FE,$E1,$F4,$D5,$AA,$FF,$6B,$CE,$63,$F5,$F1
	.dc.b	$C6,$3A,$7C,$6F,$78,$FF,$1A,$C7,$F9,$FA,$53,$9B,$EF,$F0,$BF,$82
	.dc.b	$8C,$56,$FE,$36,$FC,$77,$E1,$BE,$03,$35,$72,$F0,$35,$F2,$BF,$8C
	.dc.b	$F8,$DE,$77,$F8,$EF,$8D,$EB,$76,$FD,$15,$5F,$6B,$F9,$4F,$8C,$DC
	.dc.b	$7F,$1D,$F7,$3F,$8F,$7D,$C7,$FB,$16,$3E,$06,$07,$43,$81,$2E,$2F
	.dc.b	$6D,$6E,$B5,$7D,$27,$31,$CA,$F8,$F2,$7C,$4C,$7F,$4F,$F2,$3F,$1B
	.dc.b	$A2,$DF,$D1,$6B,$F5,$9F,$D6,$FC,$AF,$CC,$EA,$FD,$7F,$27,$CE,$E9
	.dc.b	$B7,$F6,$7A,$6B,$F7,$9F,$DA,$F2,$EE,$3C,$9C,$2F,$1B,$EB,$F5,$7A
	.dc.b	$F8,$3F,$D6,$B3,$9A,$FC,$24,$E7,$8B,$E4,$DF,$09,$E7,$E2,$FE,$03
	.dc.b	$D9,$D4,$7C,$9C,$3F,$DC,$AF,$E3,$7C,$8C,$E5,$F3,$3C,$2F,$C3,$C6
	.dc.b	$2B,$75,$F9,$68,$C7,$85,$F8,$CF,$23,$5E,$17,$E3,$BC,$5B,$7F,$99
	.dc.b	$F1,$6B,$F9,$E7,$A7,$89,$B8,$F8,$3E,$97,$E5,$BC,$6F,$03,$F9,$BF
	.dc.b	$E0,$1F,$E1,$BB,$FC,$28,$C5,$79,$7D,$3E,$0F,$CA,$F2,$B8,$7A,$5E
	.dc.b	$96,$C6,$4F,$DA,$61,$FE,$70,$7E,$8F,$A5,$DF,$D1,$F8,$1E,$8F,$87
	.dc.b	$D0,$B5,$C3,$D6,$AF,$F2,$0F,$39,$8F,$F9,$A2,$3E,$DF,$A5,$C5,$FB
	.dc.b	$BE,$F7,$AF,$D4,$F9,$8D,$8C,$9E,$48,$FA,$1E,$EF,$C1,$F5,$FE,$97
	.dc.b	$B9,$F0,$FD,$8F,$A9,$EE,$5C,$3F,$12,$D7,$F4,$27,$DF,$9F,$95,$EF
	.dc.b	$7D,$BF,$73,$E6,$7B,$FF,$77,$DD,$F9,$DE,$ED,$C3,$E2,$BB,$7E,$3C
	.dc.b	$FE,$1B,$D1,$7C,$37,$D5,$F7,$F0,$FD,$7F,$84,$B7,$F3,$9E,$1F,$0F
	.dc.b	$B8,$FD,$95,$BE,$1F,$EF,$7C,$3F,$5F,$87,$B5,$A6,$BE,$CF,$0D,$7C
	.dc.b	$63,$ED,$FC,$57,$97,$E2,$3F,$11,$F1,$56,$7E,$FF,$C5,$57,$EC,$76
	.dc.b	$A3,$F3,$DE,$AF,$09,$F2,$EF,$F9,$1E,$0F,$81,$C3,$67,$D2,$D7,$C9
	.dc.b	$A7,$F9,$1D,$FC,$6F,$0F,$C7,$E0,$FC,$D7,$0D,$9F,$C8,$F0,$D7,$C6
	.dc.b	$FC,$DE,$0D,$C7,$C6,$DF,$F4,$5C,$1F,$47,$7B,$3F,$29,$AF,$8A,$7F
	.dc.b	$A1,$7F,$00,$FF,$92,$A3,$F9,$46,$8D,$7F,$59,$AF,$B2,$7C,$E7,$B0
	.dc.b	$A7,$B3,$7E,$7D,$DE,$78,$FE,$5D,$E0,$7E,$83,$7C,$47,$F3,$78,$FD
	.dc.b	$1E,$F8,$84,$61,$FC,$0E,$FE,$E7,$D0,$7E,$D3,$7F,$73,$4F,$71,$4D
	.dc.b	$9F,$B1,$D7,$DD,$C9,$F8,$4E,$FF,$83,$DF,$F9,$77,$E1,$F6,$5B,$FE
	.dc.b	$E9,$68,$FC,$E7,$7F,$09,$FC,$AF,$7F,$CA,$EF,$FD,$07,$7D,$9F,$CC
	.dc.b	$B5,$FC,$0C,$FD,$0E,$FF,$D3,$76,$FD,$2E,$FF,$D4,$77,$FD,$4F,$26
	.dc.b	$E1,$E2,$76,$F9,$27,$F9,$9F,$DE,$BC,$BF,$B7,$E5,$E1,$B5,$FD,$6E
	.dc.b	$74,$FB,$27,$F6,$CF,$E0,$BC,$AF,$DC,$79,$7F,$84,$F2,$FF,$75,$E5
	.dc.b	$DA,$FD,$9B,$5F,$CD,$CF,$E2,$BC,$BF,$E0,$1E,$5F,$E3,$3C,$AB,$5F
	.dc.b	$2B,$D7,$AF,$D3,$7F,$83,$79,$9B,$8F,$37,$AD,$F9,$4F,$5F,$DB,$EC
	.dc.b	$7E,$5B,$D8,$B5,$F5,$1F,$83,$5B,$8F,$77,$AF,$F9,$CF,$5F,$E2,$B6
	.dc.b	$32,$7D,$C8,$F9,$3E,$BF,$E8,$BD,$7F,$97,$EB,$FE,$93,$D6,$B7,$F8
	.dc.b	$26,$BF,$8B,$7E,$8F,$AF,$FA,$AF,$81,$F2,$CF,$E3,$FE,$0A,$BF,$A2
	.dc.b	$F1,$9C,$18,$8F,$D1,$46,$1F,$82,$7E,$42,$7D,$2F,$9C,$F4,$9F,$9B
	.dc.b	$D3,$2D,$3E,$F6,$9E,$B7,$F0,$19,$FA,$DA,$7A,$CA,$7A,$CF,$C3,$65
	.dc.b	$EB,$7F,$11,$F0,$5F,$E4,$7F,$49,$F0,$5F,$90,$F8,$3C,$36,$BF,$CB
	.dc.b	$53,$A7,$89,$8F,$83,$FC,$BF,$AF,$CF,$E0,$BF,$33,$F0,$7D,$7E,$06
	.dc.b	$DF,$27,$6F,$49,$FC,$FF,$C1,$F8,$FE,$1E,$CF,$83,$E1,$EB,$FD,$BF
	.dc.b	$34,$7E,$9B,$E1,$78,$4F,$90,$D6,$9E,$96,$9F,$A7,$DB,$E0,$47,$EB
	.dc.b	$B7,$F8,$5F,$0F,$E4,$AE,$DA,$7A,$1F,$06,$B5,$FE,$43,$9A,$38,$46
	.dc.b	$27,$CE,$EB,$4A,$FF,$25,$AF,$4B,$FB,$9D,$7E,$96,$BE,$B8,$F9,$F5
	.dc.b	$E9,$7B,$FC,$3D,$79,$35,$FE,$5B,$68,$F6,$47,$D8,$D7,$CA,$F8,$71
	.dc.b	$F5,$3E,$3C,$AB,$C9,$DB,$D2,$3D,$2F,$6C,$2F,$B6,$C7,$D7,$3F,$DD
	.dc.b	$7C,$B6,$9E,$C8,$FE,$2D,$C1,$F5,$9D,$65,$F9,$1D,$FF,$8F,$E7,$F9
	.dc.b	$3E,$2B,$5F,$90,$7F,$91,$D6,$E3,$F9,$06,$9F,$99,$E2,$FE,$4F,$C1
	.dc.b	$67,$F9,$23,$5F,$24,$FE,$7F,$3F,$E5,$BC,$3F,$A1,$D3,$F9,$7F,$0F
	.dc.b	$E8,$FC,$1B,$87,$E8,$1D,$BF,$A2,$9F,$2B,$F9,$3F,$02,$BF,$84,$DE
	.dc.b	$BF,$7D,$F4,$BF,$AA,$8F,$E7,$B9,$79,$7D,$97,$C2,$FF,$A7,$D7,$17
	.dc.b	$CC,$BF,$C7,$EF,$EE,$7F,$66,$F2,$7F,$6B,$B7,$99,$FD,$36,$7E,$A7
	.dc.b	$F5,$1A,$57,$F3,$2E,$DF,$CD,$C7,$FA,$0B,$EF,$9E,$17,$F0,$5F,$0F
	.dc.b	$1E,$1F,$F0,$7E,$EB,$87,$F0,$47,$85,$6E,$3F,$D5,$D7,$FE,$BB,$7F
	.dc.b	$71,$DD,$D7,$C0,$FE,$20,$D8,$FE,$1A,$F8,$7C,$3F,$E2,$9E,$07,$F6
	.dc.b	$2F,$03,$F8,$B7,$9F,$FD,$93,$C1,$FF,$6F,$7D,$77,$FA,$13,$9B,$E4
	.dc.b	$72,$A7,$39,$F1,$3F,$8C,$DF,$99,$FE,$37,$C3,$EA,$7F,$86,$35,$FD
	.dc.b	$EB,$B7,$FC,$FD,$87,$FA,$1D,$FF,$29,$6D,$ED,$7F,$81,$EB,$EB,$7D
	.dc.b	$E7,$C2,$DB,$DE,$FC,$4D,$BD,$AF,$F7,$9D,$7D,$AF,$F7,$BC,$FD,$2D
	.dc.b	$3F,$DB,$87,$F8,$B7,$1F,$80,$FF,$B8,$F4,$FE,$68,$FF,$25,$D3,$ED
	.dc.b	$79,$3F,$1B,$CD,$FB,$9E,$67,$C8,$F3,$2D,$F4,$3F,$C6,$B8,$A3,$8C
	.dc.b	$67,$C5,$4A,$FF,$0D,$E6,$7D,$2F,$32,$DE,$E9,$ED,$E2,$3E,$4B,$F8
	.dc.b	$DF,$33,$84,$7E,$1B,$CD,$FB,$7E,$D7,$E4,$7C,$DF,$BB,$EC,$5A,$E8
	.dc.b	$FE,$56,$B7,$1F,$8E,$E0,$FC,$1F,$B3,$EC,$8F,$AF,$B5,$7F,$CA,$FA
	.dc.b	$7A,$5E,$07,$9E,$7C,$CE,$B5,$AF,$CF,$D6,$BF,$19,$F5,$3E,$F8,$F2
	.dc.b	$3F,$07,$4F,$43,$F8,$8C,$EB,$E2,$76,$FE,$32,$3C,$CF,$DD,$F6,$2B
	.dc.b	$F5,$1B,$C1,$F4,$9F,$AB,$1F,$94,$F6,$3C,$B3,$F5,$B0,$B9,$FD,$2A
	.dc.b	$FD,$9F,$0F,$EA,$33,$F4,$C7,$E7,$2B,$C6,$FC,$6C,$BC,$EF,$E3,$F3
	.dc.b	$AF,$E1,$9D,$BF,$93,$8F,$E1,$1E,$D7,$ED,$3E,$0E,$CF,$E0,$DA,$FE
	.dc.b	$A9,$FE,$1D,$EB,$7E,$EB,$E0,$FF,$88,$7B,$5F,$BC,$F8,$6F,$E2,$5E
	.dc.b	$E5,$C3,$F8,$5B,$5F,$D8,$CF,$F0,$2F,$81,$FE,$31,$EE,$FF,$04,$F8
	.dc.b	$5B,$5F,$93,$F8,$6A,$FE,$E9,$FE,$3D,$EF,$EE,$3F,$8B,$ED,$E3,$7F
	.dc.b	$8C,$FC,$07,$AD,$A7,$AC,$D9,$F9,$FD,$7D,$2F,$E8,$75,$F4,$B4,$F4
	.dc.b	$94,$F4,$9F,$A7,$CF,$CA,$FE,$5B,$2A,$FE,$FF,$2A,$FE,$D9,$F4,$BF
	.dc.b	$AF,$83,$FD,$27,$CA,$FE,$95,$F8,$03,$E1,$3F,$93,$FC,$46,$AB,$FD
	.dc.b	$D5,$2F,$86,$FE,$57,$F0,$39,$F6,$3F,$A9,$D1,$56,$E3,$F7,$D3,$BC
	.dc.b	$7B,$E3,$DA,$7F,$A0,$C7,$F3,$6F,$87,$FE,$98,$FF,$38,$C6,$CF,$F0
	.dc.b	$77,$FA,$87,$C4,$6E,$3F,$A4,$FC,$47,$F3,$EC,$3F,$AA,$7C,$45,$9F
	.dc.b	$E9,$AD,$7F,$99,$CF,$F1,$B7,$FD,$9F,$F4,$9F,$EC,$4D,$3D,$11,$FC
	.dc.b	$AB,$C2,$4B,$1F,$54,$FB,$EB,$6F,$F0,$8D,$F3,$BF,$C2,$D3,$DD,$3F
	.dc.b	$2B,$4F,$06,$5D,$31,$F6,$CF,$F4,$8D,$AB,$FC,$4A,$EB,$FC,$ED,$F8
	.dc.b	$2F,$F2,$98,$FE,$E5,$A7,$B7,$1F,$A6,$3D,$4F,$F6,$C5,$7F,$2F,$CB
	.dc.b	$E2,$B4,$F9,$27,$F3,$9D,$7E,$17,$1F,$99,$61,$EC,$D9,$39,$DF,$0D
	.dc.b	$BF,$C6,$1F,$17,$3E,$57,$C7,$8B,$A7,$3D,$AF,$07,$3D,$7E,$86,$68
	.dc.b	$F0,$72,$F0,$9E,$BB,$FE,$3B,$8F,$CD,$D5,$E5,$E6,$B5,$AE,$0F,$F2
	.dc.b	$83,$ED,$9F,$55,$7B,$7A,$FD,$33,$F0,$7A,$FD,$DD,$3E,$88,$E4,$D7
	.dc.b	$F4,$7C,$5F,$16,$7D,$11,$FA,$7E,$8F,$C3,$F0,$62,$3F,$5F,$1F,$8A
	.dc.b	$DF,$11,$F9,$07,$DA,$FD,$19,$F2,$BE,$73,$D5,$AF,$B1,$FD,$9B,$76
	.dc.b	$FB,$FB,$35,$F6,$57,$A0,$F6,$14,$F9,$B9,$7B,$9C,$9E,$C3,$F0,$5D
	.dc.b	$7E,$6C,$FD,$71,$F2,$F8,$BF,$09,$A6,$78,$BE,$B4,$FC,$FE,$2F,$AC
	.dc.b	$FF,$08,$CB,$CD,$D5,$FC,$3B,$6B,$3F,$A9,$6E,$DF,$6E,$3F,$25,$C5
	.dc.b	$F8,$FF,$1B,$F2,$5D,$BF,$91,$F0,$2D,$7D,$D6,$BF,$C5,$A7,$F3,$1E
	.dc.b	$6F,$E5,$FC,$7F,$CD,$79,$36,$BF,$3D,$E3,$57,$F1,$EF,$E7,$7C,$8D
	.dc.b	$C7,$E2,$7C,$4F,$D1,$79,$9F,$A1,$7F,$49,$E5,$D9,$FD,$43,$FA,$5F
	.dc.b	$1B,$71,$E5,$3F,$59,$9F,$A5,$FD,$76,$D5,$FC,$FB,$EF,$76,$7E,$67
	.dc.b	$4F,$1B,$FC,$C9,$FD,$A6,$7E,$98,$FC,$87,$2A,$3D,$51,$FB,$BD,$7C
	.dc.b	$CF,$EB,$34,$AF,$0C,$EB,$F9,$6B,$47,$AE,$3F,$82,$4F,$F0,$1F,$89
	.dc.b	$FE,$09,$E6,$D9,$FE,$11,$EC,$57,$F7,$2F,$F0,$7F,$2B,$71,$FB,$3F
	.dc.b	$3B,$F8,$67,$B5,$FC,$2D,$B1,$93,$F8,$94,$7F,$0F,$F4,$3F,$89,$7A
	.dc.b	$BF,$C4,$7C,$FF,$E2,$9E,$B5,$BF,$F2,$87,$D3,$FF,$97,$3B,$3F,$47
	.dc.b	$5F,$CB,$A3,$1D,$DF,$D7,$BD,$2E,$37,$FC,$14,$F2,$9F,$D7,$8B,$7F
	.dc.b	$AC,$6C,$FF,$66,$E0,$E9,$EB,$7A,$67,$CE,$7A,$4F,$E3,$3A,$E5,$A7
	.dc.b	$F1,$3D,$7D,$6F,$F1,$5C,$FC,$6F,$F6,$E7,$D1,$EB,$FB,$63,$FA,$96
	.dc.b	$9E,$A7,$FA,$86,$9C,$34,$FE,$7F,$DA,$F5,$BE,$07,$C4,$7F,$52,$87
	.dc.b	$B7,$8C,$3F,$94,$C6,$1F,$93,$3B,$AF,$91,$F0,$FC,$7E,$0F,$E5,$7C
	.dc.b	$1D,$BF,$9E,$D7,$CA,$F7,$F7,$BE,$87,$C3,$F8,$7E,$03,$E9,$7C,$3F
	.dc.b	$8F,$E0,$6E,$1E,$26,$BF,$E0,$9E,$8F,$F6,$E6,$CB,$FB,$5F,$0F,$EB
	.dc.b	$E0,$F2,$DF,$AB,$E1,$3D,$8F,$13,$FC,$25,$68,$CF,$17,$C6,$AE,$35
	.dc.b	$B8,$53,$7F,$7C,$FC,$66,$BB,$8D,$3F,$08,$F9,$7B,$7C,$07,$E6,$E5
	.dc.b	$EC,$BF,$AB,$A5,$D3,$DA,$DA,$F3,$D7,$B9,$FA,$BF,$09,$F9,$9F,$89
	.dc.b	$FB,$4F,$A3,$0F,$BE,$DC,$3E,$3E,$BB,$BF,$41,$E5,$3D,$1A,$7A,$AF
	.dc.b	$F0,$ED,$AF,$9F,$9F,$A6,$BD,$D9,$79,$6F,$E2,$E7,$74,$36,$7D,$15
	.dc.b	$F1,$7E,$1F,$F6,$1F,$11,$F9,$16,$C7,$E4,$1A,$FC,$BF,$88,$FD,$B7
	.dc.b	$C4,$FE,$5F,$8B,$F7,$15,$F9,$9E,$1F,$DD,$55,$BF,$DF,$BB,$7F,$80
	.dc.b	$CF,$E7,$F0,$FE,$01,$C3,$67,$F8,$2B,$5F,$D5,$BF,$A5,$C7,$F8,$37
	.dc.b	$0F,$E9,$F9,$3C,$75,$FA,$6E,$4A,$FE,$05,$BC,$1F,$B0,$9F,$E1,$DC
	.dc.b	$7E,$3A,$FD,$BE,$7E,$77,$F7,$3A,$57,$FC,$0D,$D4,$8E,$11,$89,$F3
	.dc.b	$3A,$D2,$BF,$E0,$AA,$46,$21,$EC,$A7,$31,$FB,$ED,$7D,$AF,$F0,$0E
	.dc.b	$8F,$E4,$9C,$BF,$C1,$7A,$3F,$93,$72,$5A,$FE,$56,$FF,$09,$AD,$C7
	.dc.b	$F8,$C3,$F0,$27,$87,$FC,$B7,$9F,$FC,$BF,$18,$9F,$E6,$71,$89,$D4
	.dc.b	$7F,$10,$EE,$FE,$69,$D7,$FC,$57,$C2,$FE,$6D,$DB,$6B,$FD,$20,$7F
	.dc.b	$8A,$30,$BC,$DF,$C7,$AB,$FC,$D3,$18,$ED,$FE,$0F,$DD,$CA,$FC,$DE
	.dc.b	$17,$A1,$FD,$F6,$A8,$C6,$3F,$E6,$68,$C4,$FF,$4A,$E1,$F5,$F8,$9D
	.dc.b	$BC,$4F,$54,$FB,$0F,$85,$C1,$EC,$7A,$B7,$0F,$E6,$7C,$1E,$E9,$FE
	.dc.b	$5F,$B7,$95,$FE,$14,$FA,$9F,$80,$7F,$A3,$F6,$F6,$3F,$C6,$B5,$E1
	.dc.b	$7E,$7F,$23,$E0,$F8,$DF,$43,$C8,$F8,$7E,$3D,$C3,$E4,$9F,$84,$C5
	.dc.b	$1F,$5A,$F8,$CF,$DB,$F3,$FE,$5F,$91,$F6,$BD,$0F,$9B,$E4,$DA,$E7
	.dc.b	$DD,$FC,$02,$3E,$FD,$F0,$9F,$C2,$79,$DF,$53,$CB,$FC,$1F,$9F,$67
	.dc.b	$F1,$0D,$7E,$9C,$FD,$AF,$53,$F1,$3E,$77,$92,$7E,$BF,$9F,$78,$FE
	.dc.b	$B3,$7B,$7F,$DB,$30,$EC,$5F,$E1,$3D,$4F,$2E,$1E,$8E,$47,$17,$AE
	.dc.b	$3E,$77,$05,$7F,$CF,$12,$8E,$31,$F4,$E7,$F3,$BA,$FA,$72,$FC,$46
	.dc.b	$D7,$4F,$1E,$9F,$E3,$1C,$3C,$1B,$F0,$7C,$5B,$FE,$9B,$DC,$F3,$47
	.dc.b	$EA,$FD,$CF,$63,$FE,$E9,$B4,$63,$2F,$75,$70,$AF,$CD,$F1,$7D,$17
	.dc.b	$F3,$BC,$9F,$61,$FD,$B3,$5D,$9F,$0F,$6F,$23,$F8,$FC,$FE,$BB,$E8
	.dc.b	$3E,$B9,$4F,$AE,$7A,$8F,$E0,$5F,$03,$FB,$2F,$57,$F8,$0F,$B1,$FB
	.dc.b	$4F,$56,$D7,$D1,$7F,$82,$D6,$E3,$F8,$57,$C1,$7E,$EB,$D9,$FE,$12
	.dc.b	$D8,$C9,$F6,$63,$F8,$6F,$C2,$7F,$00,$F6,$BF,$87,$FC,$0F,$F0,$2F
	.dc.b	$83,$B7,$F7,$FE,$0E,$BF,$AB,$7F,$8A,$FC,$0E,$E3,$F8,$D7,$47,$8D
	.dc.b	$FE,$0A,$DE,$5F,$E0,$8D,$3E,$EF,$C4,$7F,$10,$F8,$3F,$3D,$79,$4F
	.dc.b	$39,$FB,$9D,$72,$D3,$F9,$46,$BE,$97,$F9,$56,$9E,$37,$F8,$AB,$FC
	.dc.b	$73,$3F,$4C,$7E,$51,$F4,$BF,$9B,$D0,$FF,$B8,$BF,$04,$7C,$4F,$F2
	.dc.b	$4F,$85,$CC,$FF,$87,$65,$F1,$19,$8E,$0D,$75,$5F,$CB,$3E,$1B,$8F
	.dc.b	$03,$83,$F5,$F8,$31,$FE,$63,$F0,$DA,$9C,$DF,$F3,$4F,$88,$C6,$4C
	.dc.b	$58,$FF,$B5,$63,$F6,$93,$A9,$CE,$3D,$27,$D9,$8F,$F3,$EC,$B5,$83
	.dc.b	$0C,$57,$C1,$6B,$EF,$AF,$E8,$B9,$F0,$D7,$D9,$3F,$1F,$87,$E0,$67
	.dc.b	$ED,$8F,$36,$29,$CE,$BF,$02,$F8,$D7,$CB,$D3,$CC,$FC,$DD,$7E,$03
	.dc.b	$F4,$DA,$EC,$FE,$25,$9F,$A8,$F3,$E5,$E1,$3E,$2E,$98,$C7,$E7,$B6
	.dc.b	$BC,$B5,$EC,$7F,$01,$B7,$F3,$E7,$E7,$7E,$3E,$F0,$BE,$67,$E0,$DF
	.dc.b	$E5,$C9,$FE,$87,$92,$B5,$1F,$CA,$B2,$FE,$57,$5B,$8F,$ED,$F3,$A7
	.dc.b	$F1,$58,$77,$BE,$5D,$BD,$BA,$31,$6F,$23,$18,$78,$21,$DD,$88,$EA
	.dc.b	$F9,$B1,$E3,$5E,$8E,$1E,$7D,$B6,$F9,$74,$D7,$CA,$FB,$B9,$37,$1F
	.dc.b	$A3,$8F,$89,$E5,$BE,$BF,$EF,$4F,$E8,$F5,$CC,$63,$7F,$A5,$CE,$9C
	.dc.b	$F1,$7E,$43,$1D,$F6,$7E,$4B,$C9,$E3,$DF,$73,$FE,$CB,$9B,$BB,$11
	.dc.b	$FE,$1A,$D7,$C5,$1F,$9A,$E5,$E3,$3F,$B9,$E7,$F1,$F7,$7E,$79,$F2
	.dc.b	$F7,$D9,$EA,$FE,$8B,$97,$71,$FC,$07,$9F,$CE,$FF,$03,$E5,$AF,$F0
	.dc.b	$16,$F2,$7B,$A7,$D7,$3F,$C2,$77,$F1,$BF,$C1,$DF,$8B,$A7,$AA,$3F
	.dc.b	$85,$EB,$EA,$7F,$88,$6B,$C3,$0F,$93,$A7,$92,$7F,$88,$3E,$A7,$DC
	.dc.b	$7A,$8F,$E3,$19,$EE,$7E,$8F,$9D,$E4,$9F,$E3,$1E,$1F,$A3,$93,$F8
	.dc.b	$8B,$63,$F6,$F3,$BA,$FB,$3A,$78,$EF,$F8,$BB,$EA,$AF,$71,$F5,$B4
	.dc.b	$FA,$4F,$D8,$E2,$F9,$CF,$D7,$E2,$71,$62,$3E,$0C,$38,$B1,$1F,$CC
	.dc.b	$B8,$3F,$1D,$1F,$56,$3F,$07,$5F,$C6,$3D,$5F,$D2,$F6,$D9,$FD,$27
	.dc.b	$75,$7F,$5E,$FF,$1E,$F4,$77,$1F,$B5,$3F,$95,$7E,$10,$F1,$7F,$5F
	.dc.b	$E2,$FF,$29,$8C,$4F,$EB,$E3,$13,$A8,$E5,$C4,$A7,$8D,$FB,$8F,$1B
	.dc.b	$97,$93,$FB,$DF,$C8,$3F,$49,$CD,$FE,$2E,$B4,$E6,$37,$8F,$F0,$CA
	.dc.b	$DB,$E6,$F2,$F9,$77,$FF,$3D,$FC,$29,$EE,$75,$E8,$C3,$DF,$A6,$E1
	.dc.b	$CD,$FE,$7E,$EC,$F4,$E8,$F4,$3C,$DC,$FF,$07,$D0,$F4,$74,$DB,$F2
	.dc.b	$3F,$16,$B7,$1C,$BD,$3F,$77,$2F,$C9,$6C,$64,$F6,$47,$CD,$F4,$FE
	.dc.b	$27,$2F,$AF,$DB,$F8,$DC,$BE,$DF,$7F,$E4,$72,$5C,$7E,$23,$B7,$E3
	.dc.b	$4F,$D7,$F7,$7F,$9C,$43,$E6,$DE,$AF,$DB,$F7,$B3,$F3,$0F,$73,$17
	.dc.b	$FC,$FE,$30,$FC,$4C,$39,$E1,$CA,$3E,$FF,$4F,$FB,$D9,$FA,$D8,$FD
	.dc.b	$3F,$6B,$EF,$36,$32,$70,$F8,$0F,$C1,$7B,$5F,$67,$E1,$FF,$09,$CD
	.dc.b	$6B,$F0,$0D,$7F,$D6,$BD,$FF,$55,$F0,$FC,$37,$E2,$FA,$BF,$BF,$79
	.dc.b	$EF,$83,$F3,$E9,$87,$DF,$C5,$70,$F7,$3C,$3E,$13,$71,$EC,$F8,$BF
	.dc.b	$CB,$75,$FE,$37,$E2,$7F,$31,$DD,$6B,$F2,$8F,$E4,$6B,$71,$F1,$3E
	.dc.b	$37,$F3,$DF,$07,$F9,$56,$C6,$4F,$CE,$47,$E6,$7E,$37,$F4,$9F,$0B
	.dc.b	$EA,$7C,$73,$FE,$1D,$F2,$BF,$D2,$7C,$0F,$89,$F5,$CF,$EC,$23,$0F
	.dc.b	$B0,$C5,$72,$F3,$7F,$C0,$5D,$7A,$9F,$97,$DB,$FA,$9C,$3F,$69,$DF
	.dc.b	$67,$F6,$0F,$EB,$7E,$2F,$F7,$15,$FA,$FE,$3F,$DD,$7C,$55,$BF,$F0
	.dc.b	$0F,$E4,$2F,$8B,$C4,$7F,$BB,$23,$0F,$E2,$3E,$37,$8B,$E9,$E6,$F3
	.dc.b	$70,$FB,$9F,$6C,$7C,$13,$FC,$7D,$C2,$8C,$F0,$7F,$87,$61,$E4,$72
	.dc.b	$AF,$F2,$27,$0B,$E3,$BF,$C1,$99,$BE,$37,$F8,$47,$C7,$E3,$B2,$E1
	.dc.b	$EA,$7F,$86,$5F,$0A,$FE,$1D,$BF,$2E,$DB,$7F,$03,$8A,$BE,$F7,$F8
	.dc.b	$A5,$6E,$3F,$92,$FE,$1B,$FB,$47,$F6,$9D,$FF,$E5,$58,$C7,$93,$FB
	.dc.b	$09,$F8,$1C,$29,$CE,$9B,$F6,$FD,$D2,$E0,$E1,$7F,$E7,$41,$87,$E8
	.dc.b	$DF,$DB,$DF,$F0,$0D,$3F,$94,$36,$3F,$77,$1D,$3C,$C7,$6F,$C8,$E7
	.dc.b	$FF,$35,$43,$A7,$E5,$F1,$61,$B8,$7F,$42,$9E,$2F,$E1,$2B,$DD,$F0
	.dc.b	$5E,$8D,$FF,$C8,$13,$E7,$D3,$E0,$6D,$FE,$57,$9C,$CF,$FA,$3A,$73
	.dc.b	$D3,$FE,$F4,$95,$F9,$FB,$25,$FF,$19,$DB,$CF,$1E,$D8,$F9,$79,$BA
	.dc.b	$FD,$DE,$27,$BF,$C4,$FE,$3B,$3E,$E6,$D7,$AE,$7D,$1F,$0D,$E9,$F8
	.dc.b	$AF,$C8,$EB,$E3,$6D,$7F,$A4,$F5,$FF,$1F,$3F,$E6,$9F,$3B,$F9,$43
	.dc.b	$E2,$F1,$FE,$71,$E5,$E1,$FF,$3B,$68,$8C,$63,$FE,$B2,$9F,$F7,$BF
	.dc.b	$8B,$F5,$6A,$5F,$57,$EA,$9E,$87,$DF,$D5,$F3,$FD,$D9,$1F,$C1,$4F
	.dc.b	$C6,$E7,$FE,$9F,$71,$87,$D1,$F7,$BF,$05,$E7,$49,$F9,$87,$DD,$7C
	.dc.b	$AF,$FA,$C7,$1F,$E6,$8F,$E4,$4F,$CE,$9F,$83,$F6,$3F,$19,$EB,$F9
	.dc.b	$70,$F7,$B7,$5F,$75,$FF,$97,$6F,$E4,$69,$F9,$0F,$5F,$F2,$9E,$BF
	.dc.b	$CE,$C5,$7F,$E1,$4B,$F0,$75,$7E,$19,$F2,$5F,$E6,$FD,$79,$1E,$A3
	.dc.b	$F0,$6F,$AA,$F9,$FA,$DF,$AC,$F9,$EF,$89,$5E,$BE,$5F,$B3,$5F,$EA
	.dc.b	$B9,$FF,$71,$CF,$D1,$3F,$C4,$2E,$1E,$19,$7F,$B4,$63,$9C,$7E,$D6
	.dc.b	$B9,$EB,$F7,$67,$FD,$D3,$3F,$00,$E7,$EE,$FB,$E3,$31,$EE,$F0,$FF
	.dc.b	$67,$5F,$B7,$F7,$7F,$71,$EE,$FA,$F2,$36,$3D,$9D,$53,$0F,$41,$FC
	.dc.b	$3B,$14,$7D,$38,$FF,$76,$43,$E2,$FE,$27,$C5,$C9,$FC,$3E,$5E,$98
	.dc.b	$F3,$7C,$7F,$C9,$F8,$F9,$5F,$3F,$83,$E4,$47,$CF,$C5,$D5,$F3,$FD
	.dc.b	$29,$7E,$4B,$F9,$D1,$E3,$55,$FE,$C1,$9F,$8C,$7A,$7E,$2B,$EC,$1F
	.dc.b	$1C,$F9,$1F,$17,$F6,$FE,$32,$4F,$8F,$5F,$A1,$1E,$3F,$0F,$F0,$1E
	.dc.b	$7F,$E0,$7E,$1B,$C1,$91,$B1,$F8,$5F,$8A,$FC,$37,$C5,$FD,$53,$EB
	.dc.b	$3F,$89,$F8,$B9,$9F,$73,$E5,$BF,$07,$CE,$FE,$73,$F4,$5F,$D6,$27
	.dc.b	$D7,$C1,$FE,$F7,$87,$C3,$FC,$16,$CF,$C5,$DF,$F8,$24,$71,$8F,$CB
	.dc.b	$78,$87,$FC,$E7,$1F,$E3,$57,$26,$3E,$1B,$F5,$50,$6F,$C3,$F9,$3E
	.dc.b	$26,$57,$A7,$25,$FF,$9B,$A3,$1C,$3F,$D6,$FC,$2F,$CD,$C2,$B7,$F0
	.dc.b	$3E,$E3,$D4,$7A,$FE,$07,$E0,$FC,$44,$AF,$C5,$9C,$71,$47,$9B,$E1
	.dc.b	$BE,$3F,$C3,$4A,$F2,$52,$3F,$74,$FF,$81,$63,$E5,$F0,$7C,$6F,$7F
	.dc.b	$E7,$9F,$2C,$F9,$9E,$D7,$D2,$F7,$3E,$9F,$C0,$FC,$43,$E2,$BF,$5B
	.dc.b	$E1,$BE,$BF,$C3,$76,$3B,$BF,$DC,$BE,$8D,$EF,$53,$F3,$0F,$F4,$FC
	.dc.b	$FE,$C1,$F1,$5D,$BE,$63,$E0,$7C,$27,$E1,$3E,$13,$EA,$BF,$59,$FC
	.dc.b	$3F,$A1,$27,$E0,$FC,$0F,$C0,$8F,$17,$BF,$E3,$F8,$0F,$0B,$E2,$7C
	.dc.b	$DF,$01,$E7,$F8,$0F,$E5,$59,$FF,$4F,$C7,$E8,$F5,$21,$3E,$DF,$83
	.dc.b	$F7,$7C,$1F,$F2,$79,$FE,$FB,$7F,$CC,$34,$FC,$83,$FE,$18,$8F,$F0
	.dc.b	$33,$ED,$C3,$A7,$B5,$F1,$CF,$8B,$ED,$7C,$67,$E5,$7B,$32,$BC,$77
	.dc.b	$FD,$A1,$FF,$03,$6F,$DB,$07,$4F,$F6,$5A,$FE,$87,$5D,$D7,$C0,$7D
	.dc.b	$5F,$89,$F4,$3E,$97,$EC,$7B,$52,$BE,$AE,$1F,$07,$DB,$FB,$33,$F7
	.dc.b	$7E,$06,$57,$93,$E0,$7E,$4C,$7E,$07,$DF,$FC,$17,$C0,$FC,$57,$E3
	.dc.b	$3F,$E3,$6E,$D7,$B7,$7D,$BF,$E3,$91,$8C,$7C,$2F,$F2,$47,$E4,$3F
	.dc.b	$D2,$1F,$9B,$E5,$7E,$BB,$54,$FF,$4F,$9F,$8A,$7F,$BC,$9E,$0F,$F0
	.dc.b	$3F,$81,$FE,$C9,$3C,$3E,$2F,$C9,$F0,$5F,$99,$E4,$FE,$C5,$87,$F4
	.dc.b	$EC,$3F,$C6,$9D,$3F,$3E,$7E,$71,$EB,$F6,$FD,$9E,$DF,$A9,$F5,$BE
	.dc.b	$FF,$82,$F8,$1F,$05,$FD,$DF,$1F,$E9,$B8,$FE,$39,$FE,$98,$FB,$BE
	.dc.b	$0B,$E3,$7C,$17,$89,$F1,$BF,$27,$DB,$93,$F1,$BD,$BF,$93,$1F,$37
	.dc.b	$E0,$BE,$77,$C7,$77,$7C,$0F,$D1,$F8,$39,$7F,$C6,$F2,$7C,$1F,$C0
	.dc.b	$9F,$F3,$5F,$D4,$7F,$1D,$AF,$E4,$E7,$F9,$C9,$C3,$FD,$E7,$C4,$F8
	.dc.b	$8F,$F1,$FB,$FE,$45,$18,$E6,$57,$0A,$FB,$3E,$FE,$6D,$F0,$7F,$85
	.dc.b	$3E,$37,$C2,$7C,$77,$F1,$1F,$07,$F8,$9F,$84,$FA,$AF,$D6,$7C,$FF
	.dc.b	$15,$E8,$F8,$5F,$EC,$D9,$FF,$43,$CF,$CC,$7F,$92,$7A,$BF,$C9,$51
	.dc.b	$FD,$2A,$3D,$BA,$FE,$A9,$FE,$71,$F8,$77,$F2,$B7,$F7,$0B,$DF,$FF
	.dc.b	$18,$39,$8F,$F0,$91,$D4,$E3,$1F,$2B,$E1,$7E,$5F,$C3,$7F,$8B,$70
	.dc.b	$FE,$A5,$5F,$8C,$38,$61,$F1,$A7,$E8,$FC,$17,$D2,$C3,$D6,$FB,$1F
	.dc.b	$F4,$FE,$3F,$D5,$EF,$CB,$F0,$DE,$4B,$F2,$BF,$6B,$E0,$FE,$DF,$C6
	.dc.b	$78,$5F,$13,$F7,$BD,$FF,$F4,$E5,$F8,$B4,$FD,$5B,$FE,$DA,$F8,$0F
	.dc.b	$B5,$09,$FF,$4E,$CF,$E0,$F9,$B8,$BF,$EB,$6F,$C3,$DC,$9F,$EB,$45
	.dc.b	$DD,$FE,$57,$73,$3F,$CB,$9C,$BF,$C9,$70,$FF,$7A,$3F,$E4,$5E,$9F
	.dc.b	$0F,$C3,$7C,$9F,$87,$F9,$4F,$E5,$BE,$1B,$FD,$99,$5F,$95,$7F,$DB
	.dc.b	$A7,$A3,$E1,$AC,$7B,$63,$7C,$DF,$7C,$65,$FA,$29,$FE,$6C,$D7,$EE
	.dc.b	$9E,$DF,$27,$D5,$E0,$FE,$9B,$87,$EE,$1F,$77,$E1,$AD,$7A,$FE,$17
	.dc.b	$FD,$FA,$FE,$A5,$FF,$78,$9F,$33,$E1,$6C,$7D,$18,$DE,$5E,$5C,$9E
	.dc.b	$47,$CB,$3E,$57,$A7,$F2,$DF,$DB,$FB,$BF,$EF,$5C,$3E,$D3,$FE,$FB
	.dc.b	$7E,$E7,$9F,$5F,$DD,$46,$DF,$F7,$6F,$D5,$FF,$5A,$7F,$A6,$7C,$7F
	.dc.b	$F6,$07,$F2,$B1,$FD,$69,$FF,$76,$64,$9D,$47,$2F,$83,$FF,$50,$3F
	.dc.b	$E2,$18,$CC,$FC,$FA,$E5,$3D,$AB,$F8,$37,$C0,$FE,$C6,$3F,$67,$E2
	.dc.b	$FF,$AF,$E1,$EE,$FF,$A7,$DD,$4F,$F3,$E3,$FC,$5D,$3B,$9F,$0C,$7F
	.dc.b	$AC,$0F,$6F,$C4,$79,$31,$F4,$7C,$27,$F1,$5C,$BF,$C2,$35,$EC,$CE
	.dc.b	$D7,$9F,$E3,$78,$46,$32,$FF,$69,$5F,$0B,$F8,$F8,$73,$F8,$9D,$FC
	.dc.b	$1F,$5F,$89,$B5,$F2,$A6,$BF,$42,$3A,$C7,$C2,$F8,$0F,$4F,$BF,$FD
	.dc.b	$77,$83,$FD,$79,$7E,$FD,$BF,$31,$03,$E2,$57,$0F,$07,$D9,$F1,$3E
	.dc.b	$8C,$BD,$DF,$13,$EA,$8C,$E3,$FE,$D5,$C7,$ED,$46,$A3,$1C,$0B,$F8
	.dc.b	$DE,$B7,$C2,$F8,$9B,$5A,$8F,$95,$97,$C6,$73,$1F,$2E,$7E,$EE,$FE
	.dc.b	$58,$FE,$3F,$AB,$7F,$B9,$3F,$75,$D4,$FE,$53,$E2,$3E,$8E,$3F,$5E
	.dc.b	$AC,$FE,$52,$6B,$F9,$A8,$FA,$F1,$F7,$77,$FB,$3F,$11,$6B,$DB,$F0
	.dc.b	$FF,$81,$CB,$EE,$B9,$8F,$C1,$4F,$E5,$38,$FF,$03,$F1,$1F,$96,$E8
	.dc.b	$FC,$1F,$C4,$66,$FE,$3B,$97,$E4,$CF,$F6,$1F,$C1,$9E,$17,$E2,$B4
	.dc.b	$FC,$87,$0D,$BF,$E8,$35,$A7,$E8,$CE,$9F,$A7,$AF,$E5,$F0,$FC,$D7
	.dc.b	$CA,$FF,$89,$67,$F7,$31,$8F,$8F,$FC,$E6,$76,$7F,$73,$3C,$3E,$2F
	.dc.b	$C9,$E1,$72,$F8,$AB,$5F,$83,$C7,$FD,$F1,$D7,$FC,$97,$57,$99,$BA
	.dc.b	$F3,$E1,$E2,$F8,$AF,$77,$9D,$FA,$6F,$0B,$E0,$79,$96,$FD,$B3,$5F
	.dc.b	$DB,$9F,$EA,$5C,$53,$FA,$EF,$17,$F6,$9D,$1F,$C2,$BE,$4D,$F2,$8E
	.dc.b	$6F,$DC,$75,$E6,$31,$97,$F1,$6B,$C5,$6A,$3F,$CE,$67,$FB,$0E,$F9
	.dc.b	$F3,$FF,$85,$7E,$B7,$FC,$91,$39,$C1,$D3,$9E,$0E,$DE,$0E,$F9,$AD
	.dc.b	$FF,$78,$C3,$FC,$8C,$F1,$BF,$E1,$9C,$5E,$5F,$02,$CF,$88,$69,$EC
	.dc.b	$EF,$FE,$25,$C3,$EB,$77,$3E,$89,$F8,$35,$E7,$E9,$F6,$CF,$87,$B2
	.dc.b	$BF,$33,$A5,$A7,$E1,$63,$FD,$27,$87,$CA,$E2,$F8,$17,$F2,$78,$BD
	.dc.b	$95,$E5,$AF,$9F,$B7,$C1,$8C,$71,$7F,$A3,$1E,$D8,$FC,$87,$D7,$19
	.dc.b	$CB,$FB,$64,$FC,$49,$CD,$7B,$1F,$1D,$7F,$9E,$7C,$7F,$E7,$71,$F4
	.dc.b	$76,$F8,$6F,$81,$E0,$FD,$5D,$BE,$C7,$97,$F8,$1F,$0B,$ED,$79,$9F
	.dc.b	$83,$F1,$33,$7E,$87,$2F,$AA,$7F,$01,$F1,$BF,$89,$F1,$AC,$FB,$9F
	.dc.b	$C2,$7C,$6E,$63,$EF,$4F,$E3,$FC,$9F,$C4,$7C,$6F,$E4,$7C,$BF,$C5
	.dc.b	$7C,$6D,$BF,$C2,$4D,$7E,$3C,$7E,$5F,$CF,$FC,$87,$A9,$E9,$9F,$C5
	.dc.b	$7A,$BE,$97,$F1,$91,$5F,$9F,$1B,$7E,$8E,$5E,$56,$9E,$93,$F2,$99
	.dc.b	$E7,$9F,$E9,$CF,$A6,$7C,$A7,$DC,$CF,$D2,$D3,$27,$E4,$79,$7F,$BC
	.dc.b	$F7,$21,$7F,$8E,$D3,$D9,$1F,$A3,$F8,$BF,$E2,$BF,$1B,$FA,$D7,$F8
	.dc.b	$BF,$C6,$FE,$BF,$E3,$6D,$6C,$FE,$21,$5F,$C9,$FE,$27,$F8,$ED,$D9
	.dc.b	$FE,$29,$5F,$CA,$FE,$27,$F9,$0B,$98,$FF,$09,$F9,$EA,$FE,$61,$F0
	.dc.b	$DF,$C9,$B8,$ED,$7F,$1D,$E3,$FE,$69,$F0,$DF,$CA,$9C,$C7,$29,$FE
	.dc.b	$09,$A7,$F0,$DF,$5A,$CF,$E8,$BD,$6F,$E1,$1B,$FF,$11,$73,$1F,$1E
	.dc.b	$7F,$86,$3F,$C5,$7D,$9A,$FE,$B7,$EC,$FF,$9F,$3F,$AA,$F8,$DF,$E3
	.dc.b	$D1,$8E,$4F,$D4,$CF,$13,$F8,$77,$C7,$6B,$DF,$FE,$30,$F0,$F7,$BF
	.dc.b	$8D,$7C,$75,$AD,$9D,$EB,$F8,$F7,$C3,$7A,$E3,$F8,$C3,$E9,$7F,$8D
	.dc.b	$7C,$35,$7F,$7B,$1B,$7F,$7E,$78,$CA,$7A,$FD,$FF,$E5,$5A,$7A,$9F
	.dc.b	$7C,$7E,$27,$5F,$3B,$E3,$3C,$A7,$F1,$DD,$7F,$DE,$7E,$6B,$3F,$54
	.dc.b	$FF,$0E,$D7,$E0,$3F,$EA,$6E,$9F,$F5,$03,$F9,$5E,$1F,$F6,$74,$62
	.dc.b	$3F,$86,$C6,$39,$BD,$9C,$3F,$E7,$FA,$D5,$7F,$3F,$E2,$F3,$BF,$C5
	.dc.b	$34,$AF,$F4,$7D,$3E,$A7,$6F,$E8,$70,$FE,$5B,$BF,$BE,$BF,$A1,$1F
	.dc.b	$16,$3F,$90,$E9,$E8,$8F,$51,$F8,$1E,$3F,$AE,$D3,$E1,$9F,$CD,$B0
	.dc.b	$71,$FB,$3B,$3F,$0B,$EE,$5A,$FC,$57,$BB,$F0,$3B,$BF,$12,$E6,$3E
	.dc.b	$64,$FC,$4F,$8C,$FC,$6F,$BD,$F1,$BE,$2F,$F1,$FE,$F6,$6F,$F2,$8E
	.dc.b	$5F,$CB,$4F,$CC,$F8,$BF,$CA,$FC,$37,$CE,$F8,$EF,$CB,$FC,$66,$6F
	.dc.b	$F3,$8E,$5F,$CF,$4F,$D4,$F8,$9F,$2C,$7D,$96,$BF,$D5,$FA,$5F,$1D
	.dc.b	$8E,$6F,$AD,$5F,$02,$BF,$01,$B7,$A9,$FC,$16,$DE,$A7,$27,$A8,$C9
	.dc.b	$EA,$3F,$11,$A7,$A9,$FC,$56,$9E,$A7,$27,$A8,$C9,$EA,$3F,$21,$F1
	.dc.b	$BF,$B6,$3F,$1F,$F1,$9F,$B8,$3F,$22,$D8,$FC,$B4,$6D,$FC,$C6,$FF
	.dc.b	$BE,$CB,$F2,$EF,$F0,$0D,$3F,$33,$C1,$6B,$67,$F0,$7A,$FC,$EE,$FF
	.dc.b	$C1,$B5,$FC,$FF,$05,$AF,$D2,$4D,$7E,$64,$7F,$0C,$DB,$F4,$BC,$3E
	.dc.b	$59,$FA,$1C,$57,$87,$F4,$93,$7A,$BF,$3D,$C7,$FE,$40,$BF,$E3,$1C
	.dc.b	$3E,$87,$C8,$79,$8F,$DD,$6F,$EA,$7F,$8F,$3B,$7F,$7D,$B7,$A9,$FE
	.dc.b	$45,$B7,$A9,$A7,$A8,$A7,$A8,$FE,$51,$A7,$EF,$B9,$BF,$93,$F2,$7F
	.dc.b	$00,$E8,$B5,$FC,$16,$6B,$F8,$D8,$FE,$5B,$CB,$FD,$66,$3F,$C4,$DF
	.dc.b	$1F,$FE,$D8,$7F,$61,$D1,$9E,$17,$07,$EF,$EB,$FD,$C9,$5F,$C0,$EB
	.dc.b	$53,$9F,$3B,$F9,$CC,$62,$7F,$B2,$D7,$07,$FC,$7F,$D4,$EA,$D6,$2A
	.dc.b	$FF,$71,$D7,$F0,$FF,$2B,$31,$FE,$CD,$8F,$E2,$5E,$06,$3C,$0F,$E8
	.dc.b	$EE,$BC,$1F,$E9,$37,$C3,$C1,$FF,$34,$7F,$00,$FF,$BD,$B7,$FB,$7E
	.dc.b	$2E,$32,$D1,$FE,$3F,$9D,$3F,$DF,$3A,$BF,$AA,$BF,$C4,$F9,$7C,$7C
	.dc.b	$5E,$E8,$FE,$45,$87,$C5,$F4,$B1,$1F,$32,$6F,$07,$CA,$8F,$87,$3F
	.dc.b	$D7,$B8,$7D,$51,$FD,$8F,$5F,$53,$FD,$97,$7C,$BF,$D1,$5C,$BF,$D1
	.dc.b	$E3,$E2,$FA,$BE,$EF,$17,$E3,$FA,$BF,$03,$C6,$CD,$FB,$5C,$BE,$F9
	.dc.b	$F5,$BF,$DC,$F9,$FE,$7E,$7E,$8D,$BE,$E4,$E4,$F7,$9E,$A3,$FB,$D6
	.dc.b	$9E,$67,$F9,$C6,$7E,$67,$F9,$D6,$99,$7E,$C9,$CB,$5F,$C0,$72,$7D
	.dc.b	$88,$C7,$33,$D6,$FB,$3E,$47,$E0,$9F,$B7,$E5,$7E,$13,$D7,$B7,$C2
	.dc.b	$76,$F2,$F7,$7F,$03,$E6,$FE,$29,$B1,$DB,$E1,$6B,$FA,$28,$FC,$3F
	.dc.b	$A1,$F9,$0F,$73,$F1,$3E,$8F,$E4,$BD,$DB,$7E,$69,$AF,$EA,$A3,$F1
	.dc.b	$FE,$AF,$E6,$3E,$0B,$F2,$3D,$F6,$BF,$1D,$EA,$FE,$73,$E0,$7F,$2A
	.dc.b	$E6,$3F,$D3,$DE,$12,$BD,$11,$F9,$5D,$EB,$A7,$F4,$BB,$FA,$B9,$BF
	.dc.b	$1C,$7E,$AF,$83,$D7,$5E,$73,$F3,$DB,$FA,$DA,$64,$F0,$B9,$7C,$7A
	.dc.b	$FE,$CB,$E2,$BF,$57,$EC,$7E,$D3,$E2,$FF,$5B,$EE,$E6,$FF,$54,$E5
	.dc.b	$FD,$64,$FE,$EB,$BB,$F6,$7E,$F5,$9F,$D8,$7B,$FF,$BE,$F8,$6F,$DB
	.dc.b	$B9,$8F,$87,$3F,$C0,$BE,$27,$F7,$7F,$01,$FC,$13,$E2,$AD,$7E,$22
	.dc.b	$6B,$FC,$72,$3F,$80,$FC,$27,$F0,$EF,$B3,$FF,$5B,$7C,$3B,$A1,$F0
	.dc.b	$D9,$F8,$BF,$E0,$FF,$15,$BF,$8B,$B9,$FE,$5F,$1F,$F5,$D6,$3C,$70
	.dc.b	$F3,$4F,$FB,$33,$16,$D8,$8F,$F7,$74,$37,$F7,$47,$C0,$3F,$3F,$C3
	.dc.b	$F1,$1F,$51,$FC,$4F,$83,$DC,$D3,$27,$E1,$9C,$BF,$88,$DB,$DD,$AF
	.dc.b	$F2,$0F,$8B,$F8,$05,$9F,$54,$57,$FD,$CD,$AA,$7F,$93,$ED,$F1,$F9
	.dc.b	$3F,$94,$F0,$7C,$9D,$AE,$1F,$0E,$6B,$CA,$3F,$DF,$8F,$FB,$F3,$D7
	.dc.b	$FE,$40,$F6,$DB,$3C,$29,$72,$7D,$5D,$F3,$18,$BF,$A1,$78,$AD,$47
	.dc.b	$FB,$D6,$5C,$9F,$77,$83,$8F,$27,$DF,$E4,$E7,$CF,$9C,$38,$39,$79
	.dc.b	$4F,$BD,$FE,$85,$BF,$BD,$A7,$BC,$A7,$BC,$F1,$72,$7E,$37,$4F,$34
	.dc.b	$7E,$1F,$5A,$FC,$13,$D1,$AF,$A9,$FE,$E4,$7C,$7D,$3F,$2D,$DF,$F0
	.dc.b	$BA,$3F,$31,$D1,$F1,$3C,$1C,$DF,$C1,$72,$FC,$39,$F5,$3F,$C2,$B4
	.dc.b	$F5,$34,$F5,$14,$F5,$1E,$2E,$2F,$A9,$AF,$AE,$3C,$E7,$AC,$A7,$AC
	.dc.b	$E1,$C5,$F5,$DF,$C5,$EB,$F6,$B4,$FD,$87,$57,$DC,$F0,$7F,$65,$E0
	.dc.b	$FD,$EE,$EC,$E1,$F6,$DC,$BF,$76,$7F,$71,$E3,$FE,$13,$C1,$FD,$D7
	.dc.b	$85,$6B,$F6,$93,$5F,$CD,$C7,$E2,$BC,$4F,$E0,$1E,$2F,$E3,$3C,$5F
	.dc.b	$E0,$5E,$27,$E3,$BC,$6C,$E1,$F8,$B7,$2F,$E3,$67,$F8,$47,$A3,$F9
	.dc.b	$4F,$23,$F8,$57,$93,$F9,$6F,$4F,$37,$F9,$37,$2F,$E5,$67,$D2,$FE
	.dc.b	$5F,$6F,$4B,$4F,$49,$4F,$49,$F9,$DD,$3C,$AF,$CE,$C7,$CA,$FD,$0C
	.dc.b	$B2,$FF,$8E,$F9,$FF,$59,$1A,$8C,$73,$2B,$F9,$14,$7E,$BB,$D3,$AF
	.dc.b	$EA,$DF,$E4,$DE,$9E,$63,$F9,$01,$FB,$4F,$53,$FC,$6F,$F6,$CF,$57
	.dc.b	$F6,$FE,$AE,$3D,$5B,$7F,$DF,$6B,$4F,$ED,$E7,$4F,$EE,$73,$E7,$EA
	.dc.b	$FF,$00,$F5,$2C,$FE,$F5,$EF,$EA,$E6,$3A,$4F,$F0,$6F,$63,$C7,$EB
	.dc.b	$FF,$0C,$FE,$01,$FF,$6A,$D3,$D7,$68,$F6,$F5,$ED,$70,$AF,$D8,$8E
	.dc.b	$66,$1C,$B0,$E3,$7D,$7D,$DF,$83,$C5,$EE,$C3,$B7,$91,$FE,$E3,$CF
	.dc.b	$FC,$B9,$1E,$F3,$E1,$4E,$F0,$FE,$EB,$3A,$9F,$D4,$F0,$FC,$07,$F5
	.dc.b	$7C,$35,$FB,$2F,$C2,$7E,$21,$F0,$8F,$E4,$5B,$FD,$4F,$7F,$E1,$7B
	.dc.b	$D6,$7D,$BB,$FB,$3E,$0F,$EC,$FC,$07,$C7,$75,$3F,$C0,$78,$57,$F7
	.dc.b	$BE,$1B,$E6,$7C,$2D,$AE,$75,$FE,$A3,$7E,$09,$F3,$F6,$7C,$37,$08
	.dc.b	$FC,$15,$FD,$5F,$88,$FC,$4F,$C3,$FD,$7F,$88,$B5,$D2,$BF,$1B,$F1
	.dc.b	$3F,$6D,$CC,$79,$67,$F2,$3F,$11,$F7,$F1,$B3,$E4,$7F,$2B,$F1,$39
	.dc.b	$8F,$C9,$4F,$E1,$71,$F5,$47,$BF,$5A,$FE,$A7,$5F,$67,$37,$D6,$C7
	.dc.b	$D5,$3E,$E8,$FC,$5E,$7E,$B8,$FC,$F6,$1E,$B6,$9E,$B3,$F4,$5A,$7A
	.dc.b	$DC,$9E,$B3,$F4,$DA,$FE,$66,$3F,$5B,$AD,$9C,$9F,$42,$32,$FD,$2D
	.dc.b	$3D,$28,$FC,$D6,$7E,$AE,$8F,$CA,$65,$EA,$CF,$D9,$5F,$6B,$5E,$2F
	.dc.b	$0C,$FE,$7B,$F0,$F3,$AF,$5E,$3F,$D2,$E1,$F5,$A3,$DC,$7C,$5E,$06
	.dc.b	$DF,$0E,$31,$C5,$FB,$5E,$1F,$E1,$2D,$8F,$E0,$D9,$D7,$67,$E3,$B9
	.dc.b	$3F,$D8,$BE,$8B,$9B,$F7,$DC,$D8,$6D,$7F,$0F,$EC,$AF,$D7,$AE,$0F
	.dc.b	$E6,$3B,$7F,$8C,$F6,$7F,$07,$ED,$FE,$37,$D5,$FC,$27,$B7,$37,$F9
	.dc.b	$E7,$2F,$E8,$27,$C7,$DB,$FC,$3F,$C1,$B3,$FA,$47,$CF,$DD,$98,$FF
	.dc.b	$4F,$67,$FE,$DC,$7E,$AF,$7F,$A5,$FD,$5E,$F5,$FD,$93,$FC,$BB,$6F
	.dc.b	$5F,$57,$F9,$F2,$FD,$75,$FB,$7D,$FC,$EF,$EE,$74,$F3,$BF,$BB,$D7
	.dc.b	$2F,$E7,$5C,$BF,$9F,$8F,$9B,$E1,$7F,$29,$F0,$7E,$7F,$85,$6B,$D3
	.dc.b	$35,$F9,$91,$FE,$3A,$FE,$01,$FF,$73,$E6,$CF,$FD,$71,$5F,$C4,$25
	.dc.b	$E2,$7D,$AF,$2B,$83,$F7,$3C,$8B,$5F,$1A,$7A,$79,$BF,$80,$F3,$3B
	.dc.b	$79,$BF,$82,$F2,$AD,$FC,$C9,$AF,$DF,$8F,$27,$A3,$F8,$8F,$37,$CD
	.dc.b	$E8,$DA,$F2,$3F,$8C,$F3,$B3,$1F,$5A,$7D,$9E,$8F,$96,$3D,$5B,$7A
	.dc.b	$5A,$7D,$CD,$BD,$2F,$E3,$F6,$F4,$BE,$53,$E1,$6B,$E9,$69,$93,$FC
	.dc.b	$61,$BA,$35,$18,$D3,$F4,$3E,$87,$CF,$F4,$6C,$FC,$DF,$53,$F4,$BC
	.dc.b	$7F,$4D,$CC,$74,$9F,$D4,$FB,$DF,$5F,$D5,$FD,$5F,$BF,$F6,$7D,$DC
	.dc.b	$DF,$D7,$72,$FD,$99,$FD,$8F,$B7,$F7,$FD,$DF,$D9,$FB,$B6,$BF,$3B
	.dc.b	$35,$FD,$84,$7E,$17,$E0,$FF,$73,$ED,$FE,$1F,$E0,$FC,$F5,$F8,$57
	.dc.b	$2D,$3F,$4F,$BF,$E3,$7E,$07,$CB,$1F,$1B,$3F,$2B,$F2,$34,$CB,$FE
	.dc.b	$21,$E4,$46,$A3,$13,$EA,$7F,$29,$AF,$A9,$A7,$A8,$A7,$A8,$FC,$D6
	.dc.b	$7F,$C4,$BE,$07,$F4,$1F,$01,$67,$F3,$CF,$F1,$6F,$80,$CC,$7F,$00
	.dc.b	$9F,$D3,$7C,$1F,$9A,$3F,$95,$67,$FE,$FD,$7F,$82,$F1,$7D,$37,$F8
	.dc.b	$3F,$1D,$7F,$8D,$47,$D9,$7C,$C7,$E2,$33,$F5,$34,$C9,$F0,$DC,$BF
	.dc.b	$17,$3F,$E5,$7E,$EF,$EE,$7E,$0E,$CF,$ED,$FE,$1F,$FA,$53,$FD,$7F
	.dc.b	$EF,$1A,$7E,$FE,$77,$3F,$98,$AD,$F8,$EC,$3F,$A9,$2C,$17,$FC,$1F
	.dc.b	$E1,$F7,$5A,$9F,$F6,$86,$1F,$C1,$E7,$2F,$F8,$E6,$38,$B9,$3A,$CF
	.dc.b	$19,$FE,$25,$F0,$FC,$30,$F0,$3F,$C5,$7E,$1F,$18,$7F,$C0,$0E,$04
	.dc.b	$60,$D6,$3D,$0F,$05,$7F,$B5,$BF,$80,$7F,$DB,$38,$7B,$67,$C9,$AE
	.dc.b	$78,$BA,$B6,$39,$4F,$C1,$DB,$11,$F2,$E7,$A7,$1F,$93,$7F,$89,$3C
	.dc.b	$35,$F8,$6F,$F0,$4D,$B4,$FC,$13,$EA,$3F,$01,$E1,$A7,$A9,$FB,$0D
	.dc.b	$CB,$D0,$7D,$3D,$71,$BF,$D1,$9F,$73,$F1,$B9,$BE,$07,$27,$C8,$E5
	.dc.b	$B8,$F9,$66,$BF,$85,$8F,$8D,$CB,$F3,$B9,$7E,$47,$65,$AF,$F2,$8F
	.dc.b	$23,$B3,$E9,$F3,$FC,$D7,$53,$FE,$D1,$95,$7D,$8E,$BF,$A5,$D9,$F6
	.dc.b	$BA,$FE,$A7,$66,$70,$EA,$E5,$EF,$3F,$80,$EC,$FB,$5D,$FE,$D9,$F1
	.dc.b	$F8,$D9,$69,$ED,$3C,$F1,$E4,$7F,$D5,$99,$F9,$1F,$AD,$B7,$A5,$FB
	.dc.b	$1B,$57,$F5,$11,$F0,$9F,$B9,$A7,$A9,$FB,$DA,$7A,$9C,$9E,$A3,$27
	.dc.b	$A8,$FC,$27,$81,$F9,$3F,$1F,$F3,$6F,$E5,$7C,$5F,$CE,$F9,$76,$B6
	.dc.b	$7D,$0A,$FD,$0F,$89,$F9,$BF,$1B,$F4,$7E,$4D,$AF,$C8,$4D,$7F,$80
	.dc.b	$47,$E8,$7C,$BF,$D4,$F9,$5F,$A3,$F2,$ED,$7D,$C7,$F5,$BE,$4E,$63
	.dc.b	$F3,$53,$FA,$9F,$37,$CB,$1F,$82,$F3,$7D,$2F,$E1,$3D,$7C,$BF,$81
	.dc.b	$72,$FE,$0E,$3D,$2F,$E2,$B4,$F4,$B4,$F4,$94,$F4,$9F,$90,$CB,$CE
	.dc.b	$FE,$37,$2A,$FD,$C7,$D2,$FA,$CF,$49,$F9,$8C,$3C,$AF,$E0,$BD,$6A
	.dc.b	$FF,$98,$F9,$91,$FC,$3B,$CE,$FE,$10,$EA,$7F,$D9,$93,$FE,$38,$F1
	.dc.b	$BF,$67,$E4,$38,$7E,$84,$63,$86,$F8,$BF,$C9,$66,$E1,$FC,$8E,$7E
	.dc.b	$A3,$E6,$3D,$85,$3D,$87,$F0,$BE,$2F,$A8,$FF,$3B,$E2,$FA,$8F,$F1
	.dc.b	$6F,$88,$74,$7F,$B8,$A3,$FD,$C5,$E4,$FF,$B8,$67,$31,$FE,$CA,$8C
	.dc.b	$78,$BB,$F8,$34,$6E,$F0,$BD,$BF,$E6,$98,$7F,$4A,$C3,$FD,$DF,$B3
	.dc.b	$E7,$DE,$06,$BD,$0F,$14,$E6,$38,$FA,$18,$AD,$DF,$A6,$F1,$E0,$F6
	.dc.b	$F3,$F5,$AF,$83,$5B,$7E,$F9,$AF,$FA,$EF,$4F,$E5,$EF,$F5,$AC,$5B
	.dc.b	$FF,$4B,$FB,$F7,$AF,$EC,$F7,$71,$EE,$DC,$3F,$9F,$D6,$9F,$8D,$3A
	.dc.b	$7E,$47,$C0,$F3,$F7,$BE,$2F,$C0,$F5,$F6,$FE,$3F,$C3,$E6,$FE,$73
	.dc.b	$97,$E8,$4F,$F5,$A7,$E6,$FC,$3F,$F5,$CF,$76,$C7,$F4,$92,$9F,$DE
	.dc.b	$5F,$A7,$F0,$FE,$BD,$BE,$AF,$0F,$96,$BF,$BF,$9F,$80,$74,$FD,$F7
	.dc.b	$4F,$F8,$13,$AF,$EF,$FC,$37,$9E,$38,$C7,$E0,$B2,$F4,$46,$39,$5A
	.dc.b	$6E,$39,$46,$1F,$93,$AF,$C4,$7D,$47,$E3,$F5,$F7,$34,$C9,$D5,$CB
	.dc.b	$DF,$4F,$F0,$17,$E0,$0F,$67,$F2,$7F,$17,$8F,$67,$F2,$BF,$17,$A9
	.dc.b	$B8,$7F,$78,$BF,$E7,$CF,$A6,$F9,$FB,$7F,$94,$FA,$0F,$EB,$53,$B8
	.dc.b	$C1,$E0,$F6,$EC,$64,$FC,$FC,$A3,$CB,$EE,$73,$2C,$FD,$21,$C4,$FC
	.dc.b	$7D,$EE,$7B,$3D,$CF,$D8,$46,$2B,$D1,$7C,$DF,$55,$7E,$CB,$E2,$FE
	.dc.b	$1F,$B3,$67,$27,$EE,$E3,$2F,$EF,$7E,$2F,$E4,$FB,$9E,$F3,$E5,$FB
	.dc.b	$3F,$04,$CC,$FB,$1C,$BE,$E8,$FA,$3E,$CF,$C7,$F6,$6C,$FB,$76,$F6
	.dc.b	$6F,$F5,$A7,$D6,$EA,$7F,$27,$E7,$38,$55,$F7,$3D,$9F,$E1,$BF,$15
	.dc.b	$F7,$BD,$BB,$87,$D9,$9A,$B0,$FF,$10,$F3,$AB,$F8,$AF,$C4,$FF,$93
	.dc.b	$A9,$F8,$A7,$F0,$27,$FC,$B3,$9B,$DE,$FC,$59,$F8,$1F,$77,$85,$6E
	.dc.b	$BB,$57,$F4,$27,$F1,$55,$F8,$73,$F2,$75,$FC,$93,$E2,$AC,$FF,$2B
	.dc.b	$7F,$2F,$F1,$99,$8F,$C9,$4F,$F2,$AF,$27,$F8,$97,$C6,$7F,$2C,$F4
	.dc.b	$2D,$7F,$33,$7F,$8B,$7B,$39,$8F,$E2,$73,$FC,$CB,$CE,$F3,$F9,$3F
	.dc.b	$CB,$3D,$B7,$A1,$EC,$BF,$E6,$71,$87,$FC,$C4,$7C,$49,$E1,$3B,$8F
	.dc.b	$AB,$2F,$37,$F9,$37,$A3,$FD,$03,$CD,$FE,$51,$76,$FF,$92,$CD,$7E
	.dc.b	$A7,$BD,$FE,$86,$F4,$FC,$11,$FC,$2B,$F8,$07,$F0,$33,$9C,$7F,$A7
	.dc.b	$43,$D4,$FE,$69,$18,$7F,$9B,$7A,$56,$FF,$99,$CF,$0F,$83,$F8,$31
	.dc.b	$EE,$F5,$3D,$4F,$FA,$63,$67,$9F,$A8,$F6,$BA,$7F,$D3,$70,$9F,$E8
	.dc.b	$9E,$A7,$93,$E0,$FF,$A3,$7A,$BE,$6F,$81,$CE,$1F,$9F,$DB,$FB,$F3
	.dc.b	$FD,$1A,$78,$3F,$D2,$3D,$7F,$6F,$C0,$FF,$4F,$CB,$DF,$F0,$3F,$D4
	.dc.b	$66,$D7,$F4,$E9,$DB,$F5,$BC,$2F,$C9,$78,$CD,$9E,$27,$A6,$BF,$45
	.dc.b	$18,$78,$75,$7E,$0E,$75,$5C,$7C,$2F,$EC,$1C,$5F,$43,$0F,$EC,$5E
	.dc.b	$C5,$C7,$FA,$FC,$D7,$ED,$9F,$AF,$E9,$4F,$D8,$F8,$4F,$ED,$5E,$97
	.dc.b	$DA,$F8,$6F,$ED,$9E,$9D,$C3,$FB,$4C,$D7,$9E,$8C,$55,$F8,$2F,$87
	.dc.b	$FA,$D5,$FE,$D0,$FA,$1F,$C3,$BF,$BF,$F3,$7F,$70,$FF,$83,$A3,$2F
	.dc.b	$E6,$CC,$1F,$01,$FC,$63,$FD,$EB,$CF,$FC,$7F,$C4,$7F,$80,$5F,$C8
	.dc.b	$FC,$37,$82,$47,$F3,$6E,$BF,$F0,$9B,$E8,$CD,$5C,$A3,$FC,$0B,$C9
	.dc.b	$F9,$AF,$84,$FF,$06,$72,$D9,$C9,$F8,$78,$CB,$F8,$9F,$84,$F0,$3F
	.dc.b	$C6,$F4,$75,$FA,$79,$F3,$C4,$9F,$C1,$57,$F1,$23,$E1,$57,$2F,$59
	.dc.b	$B7,$F8,$7A,$7E,$01,$F9,$7F,$82,$D4,$66,$3D,$9B,$7E,$CB,$E1,$B3
	.dc.b	$F0,$DF,$C8,$BE,$2B,$F6,$BF,$0F,$C0,$C5,$FE,$86,$7C,$BA,$7C,$6D
	.dc.b	$7F,$99,$FA,$DF,$E3,$E8,$70,$E8,$FA,$9C,$BE,$2E,$FF,$A5,$8F,$EA
	.dc.b	$F3,$F2,$98,$7F,$84,$47,$A3,$E2,$FC,$31,$FC,$2D,$C3,$FA,$79,$FF
	.dc.b	$2C,$7F,$00,$FF,$BF,$3C,$EF,$B7,$3F,$85,$3F,$A8,$E5,$FE,$F2,$D3
	.dc.b	$CF,$1F,$8B,$C7,$D6,$7E,$27,$6F,$99,$83,$E3,$BC,$95,$FE,$27,$8C
	.dc.b	$3F,$B0,$3F,$B1,$38,$7E,$69,$8F,$8E,$E7,$F1,$DE,$A8,$E5,$F1,$5D
	.dc.b	$8C,$3F,$92,$E7,$5A,$BF,$07,$F1,$3F,$E1,$AD,$3D,$37,$FD,$B2,$3F
	.dc.b	$C3,$2F,$EE,$0F,$80,$7C,$28,$46,$9E,$31,$98,$E9,$B7,$C8,$F2,$3F
	.dc.b	$B1,$63,$E2,$F8,$BF,$83,$7E,$5C,$B9,$5E,$F7,$FA,$3F,$17,$AA,$FF
	.dc.b	$0F,$D7,$47,$F5,$B8,$F4,$8F,$F1,$44,$7D,$93,$E4,$FC,$57,$D0,$3E
	.dc.b	$1B,$CB,$E1,$7E,$9E,$9F,$C8,$74,$F8,$AE,$27,$F4,$87,$F9,$62,$FF
	.dc.b	$C5,$4F,$C3,$C3,$C1,$F1,$3E,$5B,$F9,$AF,$DE,$31,$4F,$88,$FC,$0F
	.dc.b	$C4,$F8,$AB,$E9,$B8,$C3,$D0,$77,$F8,$9F,$C3,$FC,$4F,$97,$07,$C4
	.dc.b	$FE,$28,$C4,$7E,$33,$E2,$7B,$96,$FF,$6E,$7A,$7B,$FF,$23,$F0,$FE
	.dc.b	$1C,$3F,$D9,$DC,$9F,$3E,$BF,$1F,$F1,$19,$AF,$E0,$5F,$C0,$2F,$86
	.dc.b	$F8,$CD,$AF,$CB,$1F,$E7,$C7,$0F,$F3,$4F,$43,$F9,$6B,$9F,$87,$E7
	.dc.b	$F0,$9F,$1E,$BF,$BD,$F6,$33,$F1,$5F,$3F,$85,$D5,$F8,$BE,$0F,$E4
	.dc.b	$61,$FE,$0B,$AF,$09,$FE,$8F,$70,$FD,$5F,$17,$FD,$A1,$1F,$E8,$98
	.dc.b	$FB,$1D,$D8,$F8,$2F,$5F,$C0,$7B,$8F,$6B,$F0,$3D,$FF,$37,$BF,$F0
	.dc.b	$8F,$43,$F1,$3D,$FF,$19,$70,$FB,$27,$8F,$DF,$F9,$3E,$EF,$C8,$F7
	.dc.b	$7E,$59,$F2,$9F,$9B,$E9,$FC,$32,$D7,$D8,$3D,$9E,$D7,$D2,$F6,$BE
	.dc.b	$09,$63,$B9,$FB,$DD,$FF,$D9,$8F,$B8,$FF,$35,$2F,$88,$FB,$27,$DC
	.dc.b	$3E,$DF,$C4,$7D,$EF,$88,$FA,$85,$AF,$11,$FC,$26,$3F,$B2,$F9,$1F
	.dc.b	$BA,$7F,$B1,$C7,$CF,$3F,$CF,$0E,$3A,$7E,$E4,$7E,$DB,$93,$FD,$30
	.dc.b	$FF,$A6,$27,$5D,$3F,$C1,$79,$B3,$E3,$25,$F0,$DF,$D4,$D7,$C3,$7E
	.dc.b	$DD,$FF,$13,$E1,$EF,$3F,$D9,$AE,$32,$FE,$27,$1E,$99,$E5,$F0,$BF
	.dc.b	$B7,$AF,$EC,$B3,$F8,$A3,$DD,$EF,$66,$7C,$1E,$FF,$AB,$0F,$5D,$7F
	.dc.b	$A0,$B3,$F3,$E1,$FE,$2B,$D3,$94,$FB,$7E,$17,$84,$7B,$63,$E2,$E1
	.dc.b	$87,$E2,$1F,$23,$E1,$71,$8B,$C3,$FD,$4C,$FF,$BA,$F1,$7C,$2F,$18
	.dc.b	$FE,$66,$7D,$5B,$E3,$5C,$23,$EC,$4E,$F8,$BE,$C7,$C2,$F6,$F8,$5F
	.dc.b	$F0,$8D,$7D,$8F,$7F,$5C,$BE,$CA,$D7,$5F,$BA,$73,$5F,$7B,$E0,$3A
	.dc.b	$71,$7C,$2F,$85,$CC,$FD,$29,$CB,$F8,$48,$CB,$EB,$F8,$5F,$3F,$9F
	.dc.b	$F2,$BE,$13,$33,$F8,$F9,$CB,$F8,$C8,$CB,$F8,$9F,$84,$F2,$BF,$90
	.dc.b	$F8,$2C,$F6,$FE,$26,$73,$3F,$84,$9C,$F2,$B8,$3C,$B3,$F8,$8D,$73
	.dc.b	$18,$8F,$A1,$1C,$3E,$13,$C7,$1F,$8E,$7B,$FF,$00,$FF,$19,$EA,$FE
	.dc.b	$8F,$3F,$5E,$73,$F0,$9E,$28,$FF,$04,$E8,$B7,$C2,$6B,$0F,$D7,$D7
	.dc.b	$FB,$F2,$73,$1F,$8C,$AE,$1F,$07,$E5,$AF,$EF,$94,$F8,$4F,$1C,$E6
	.dc.b	$79,$47,$9B,$E1,$3D,$31,$E2,$A4,$7A,$BE,$13,$33,$D2,$3D,$BF,$07
	.dc.b	$E9,$8E,$EF,$C0,$F8,$4C,$47,$9E,$3E,$1F,$C2,$78,$73,$F7,$3F,$1B
	.dc.b	$E3,$31,$97,$F8,$23,$5F,$62,$F8,$5F,$0E,$3F,$07,$E1,$7E,$6C,$63
	.dc.b	$1F,$91,$1F,$43,$E1,$7C,$31,$F2,$DF,$A7,$F0,$B8,$C1,$1F,$5B,$E1
	.dc.b	$3C,$3F,$0B,$73,$FE,$53,$9F,$CF,$8F,$B7,$F0,$BD,$F0,$3F,$7B,$BF
	.dc.b	$0F,$E0,$3E,$0F,$B7,$C2,$5C,$7E,$CE,$BF,$76,$3F,$27,$FC,$03,$FD
	.dc.b	$A6,$7B,$CA,$31,$9F,$F1,$93,$F0,$AB,$E1,$3E,$53,$AF,$3B,$F8,$8F
	.dc.b	$97,$FE,$2A,$87,$C1,$FC,$89,$7D,$16,$32,$FC,$45,$5C,$7F,$BA,$6D
	.dc.b	$FC,$CA,$F7,$87,$9B,$E1,$7E,$5D,$FF,$BD,$F0,$E3,$18,$F8,$5F,$5F
	.dc.b	$C5,$F9,$0F,$6F,$C2,$DA,$FE,$43,$8F,$FB,$87,$E3,$3C,$9F,$0D,$FE
	.dc.b	$F3,$9C,$BF,$CF,$1F,$D3,$EB,$FE,$A6,$AE,$11,$E8,$F8,$7F,$19,$F2
	.dc.b	$BE,$1E,$C7,$B2,$FE,$2F,$C3,$FC,$E6,$C7,$8B,$87,$FD,$5A,$F6,$C5
	.dc.b	$8F,$C6,$F8,$6F,$E1,$6F,$C8,$F5,$7E,$55,$FF,$B2,$AF,$E4,$FC,$3F
	.dc.b	$C4,$F8,$7F,$B6,$D8,$F9,$B8,$7D,$6B,$FB,$ED,$8F,$17,$A1,$D3,$E0
	.dc.b	$FB,$E3,$FD,$76,$7C,$99,$7D,$89,$FC,$3F,$C3,$FD,$23,$F1,$3F,$0F
	.dc.b	$6B,$FD,$83,$AF,$FA,$48,$76,$7E,$2F,$E3,$7F,$9E,$57,$F6,$AE,$9F
	.dc.b	$F0,$05,$F0,$BF,$C6,$D6,$9E,$47,$29,$DB,$F9,$33,$EA,$7D,$0B,$E3
	.dc.b	$BF,$1C,$BE,$17,$E4,$B9,$C3,$D6,$E5,$FC,$8E,$FF,$8C,$D7,$F4,$6D
	.dc.b	$8F,$F7,$DE,$1E,$2D,$1F,$C0,$3F,$EE,$89,$FC,$21,$F8,$5F,$81,$FF
	.dc.b	$7E,$D6,$E7,$4F,$FB,$AF,$97,$FC,$13,$1F,$8F,$8F,$F7,$77,$1E,$1F
	.dc.b	$3F,$C1,$7A,$8F,$4F,$C1,$5C,$7F,$73,$7F,$95,$AF,$F5,$79,$F3,$F2
	.dc.b	$F2,$FC,$4F,$F4,$E9,$FF,$09,$D7,$08,$FE,$D9,$D7,$E2,$E1,$FE,$DD
	.dc.b	$D5,$E4,$E2,$CE,$5E,$99,$CB,$EB,$8F,$97,$EF,$FC,$E3,$E6,$FB,$F7
	.dc.b	$2F,$87,$5F,$9A,$E1,$76,$FC,$0F,$88,$FE,$F9,$DD,$6F,$7D,$8B,$DE
	.dc.b	$1F,$27,$E1,$F6,$FC,$BC,$3F,$C0,$9E,$27,$CA,$F8,$8B,$3F,$3F,$E2
	.dc.b	$3F,$C1,$1E,$36,$73,$EB,$97,$F9,$C2,$BF,$AE,$F2,$AF,$F5,$BE,$C7
	.dc.b	$F6,$59,$FE,$22,$7E,$43,$54,$7F,$B8,$F0,$FE,$12,$EE,$3F,$C7,$33
	.dc.b	$FB,$88,$D3,$FB,$A9,$FC,$27,$8F,$F8,$0E,$AF,$F0,$E7,$8D,$F8,$2E
	.dc.b	$BF,$F0,$F4,$E6,$BC,$F3,$97,$D3,$F1,$9F,$88,$F8,$8F,$F1,$23,$63
	.dc.b	$DB,$E7,$7E,$33,$E2,$33,$9F,$E4,$6F,$DB,$EF,$7F,$31,$D4,$FC,$07
	.dc.b	$0F,$F8,$BF,$CE,$FC,$AF,$C4,$7F,$8C,$7C,$FB,$9F,$CA,$9F,$CC,$FC
	.dc.b	$46,$72,$FC,$F5,$FF,$8F,$3D,$7F,$CE,$F7,$FF,$8F,$BD,$6F,$CF,$F8
	.dc.b	$19,$BF,$D2,$4E,$5F,$D3,$47,$D2,$F6,$FC,$DE,$D5,$D3,$F8,$8E,$FF
	.dc.b	$D9,$9F,$C3,$FC,$4B,$D9,$FD,$5F,$C4,$FF,$93,$1F,$E9,$14,$F7,$7F
	.dc.b	$C9,$FE,$DE,$B0,$E3,$5A,$9F,$F2,$9F,$BF,$FB,$4F,$89,$FF,$2A,$FC
	.dc.b	$07,$ED,$BE,$27,$3C,$1F,$BB,$9C,$BF,$BD,$8F,$F2,$DE,$FE,$0F,$7B
	.dc.b	$FC,$BB,$C1,$E2,$CF,$37,$F8,$BD,$95,$A8,$C4,$FF,$99,$34,$FE,$C1
	.dc.b	$F4,$1F,$DA,$F6,$FE,$A1,$C3,$EB,$BF,$CA,$4E,$5F,$E8,$F0,$F8,$2F
	.dc.b	$83,$C3,$CE,$77,$DF,$FE,$70,$F8,$6F,$E2,$3F,$15,$FE,$72,$F8,$7B
	.dc.b	$9F,$CC,$9F,$F4,$EE,$BF,$9D,$8F,$0C,$EB,$0F,$F7,$47,$7F,$9B,$BF
	.dc.b	$FC,$F6,$DA,$FA,$BC,$3E,$AF,$37,$37,$E8,$BF,$F4,$0F,$C3,$FB,$BC
	.dc.b	$DB,$3E,$CF,$1F,$FD,$0B,$F0,$F9,$CB,$EF,$DF,$C4,$9F,$F4,$8F,$AE
	.dc.b	$EA,$F3,$FC,$56,$1F,$4F,$C5,$6B,$8F,$38,$FF,$0E,$BE,$19,$FF,$3B
	.dc.b	$E2,$FF,$2C,$BF,$B7,$3E,$5C,$FC,$2F,$8A,$FF,$4B,$BF,$13,$93,$FD
	.dc.b	$33,$9E,$63,$F2,$66,$4F,$CA,$E5,$F2,$B9,$3F,$D3,$F7,$F3,$39,$3F
	.dc.b	$D4,$33,$9A,$FC,$DC,$E5,$F3,$6B,$F4,$B9,$3D,$99,$F9,$DF,$63,$E9
	.dc.b	$E0,$CB,$FD,$07,$5F,$E2,$F7,$E4,$8F,$76,$1E,$A7,$E0,$6B,$EA,$73
	.dc.b	$3B,$79,$47,$99,$F8,$DA,$7A,$9F,$91,$A6,$5F,$04,$65,$F1,$47,$D6
	.dc.b	$F1,$BF,$17,$C5,$67,$E8,$78,$3F,$6B,$C7,$CE,$5F,$56,$FF,$23,$3F
	.dc.b	$7B,$C5,$FD,$1F,$C1,$FE,$03,$C6,$B7,$F6,$E7,$F4,$FF,$19,$9C,$BF
	.dc.b	$4B,$7F,$86,$F1,$FF,$57,$F0,$3F,$88,$F1,$EE,$3F,$83,$9F,$D7,$FC
	.dc.b	$16,$70,$FD,$6D,$FE,$3B,$D9,$FD,$1F,$95,$67,$F1,$1E,$57,$E4,$BC
	.dc.b	$FC,$E5,$F8,$DB,$FD,$4C,$FE,$5B,$D5,$FD,$5F,$99,$70,$FC,$83,$F9
	.dc.b	$AF,$53,$F5,$FE,$66,$67,$ED,$4F,$E7,$BD,$4F,$D9,$F9,$7F,$A0,$F4
	.dc.b	$AE,$9F,$9B,$9F,$DB,$FA,$39,$C3,$F3,$97,$E7,$7F,$09,$E8,$FE,$A3
	.dc.b	$D8,$FD,$EB,$98,$FF,$5F,$F8,$5F,$C5,$31,$FD,$35,$69,$FD,$46,$DE
	.dc.b	$A6,$D7,$A9,$FD,$66,$DE,$A7,$31,$EA,$7C,$CF,$91,$FD,$96,$9E,$A8
	.dc.b	$FD,$87,$AC,$F2,$5B,$7B,$2F,$75,$C4,$DC,$B8,$BF,$8B,$E9,$ED,$9F
	.dc.b	$DE,$3E,$D4,$7F,$94,$97,$A1,$B8,$FD,$44,$7D,$17,$E1,$F0,$D7,$8F
	.dc.b	$C0,$7F,$4F,$8F,$B1,$1F,$C2,$78,$F4,$F5,$8F,$B0,$F9,$F4,$67,$FE
	.dc.b	$89,$CD,$AF,$C1,$9C,$D7,$FA,$02,$7E,$0B,$FC,$57,$8F,$EE,$3F,$BD
	.dc.b	$E3,$FE,$43,$F1,$3F,$E6,$48,$FF,$36,$FF,$00,$FF,$43,$C3,$E5,$4E
	.dc.b	$75,$4E,$71,$7B,$99,$8F,$F5,$D6,$9F,$C9,$6F,$1E,$A7,$F2,$EF,$82
	.dc.b	$D7,$B3,$FC,$C3,$E3,$78,$7A,$D9,$DB,$FC,$C9,$97,$FA,$E7,$6F,$59
	.dc.b	$F6,$EB,$79,$FF,$31,$F8,$2F,$07,$B1,$FC,$E9,$F1,$7B,$1F,$CF,$3E
	.dc.b	$0B,$35,$BC,$BA,$CF,$F4,$07,$D1,$EC,$D7,$C3,$EF,$7C,$47,$C7,$19
	.dc.b	$7F,$CC,$9A,$27,$D1,$18,$DB,$D8,$FA,$B6,$CB,$F8,$83,$E6,$C6,$9E
	.dc.b	$71,$E9,$7E,$0E,$95,$EF,$8F,$AD,$CE,$5E,$B7,$2F,$F9,$83,$F0,$B7
	.dc.b	$8F,$F4,$F9,$7E,$4C,$FF,$B3,$E5,$F1,$9F,$5F,$E1,$73,$87,$FB,$1A
	.dc.b	$78,$BF,$4A,$F3,$C6,$9E,$33,$F5,$AB,$30,$F8,$AF,$C1,$7C,$36,$1B
	.dc.b	$97,$97,$CE,$FC,$37,$C3,$67,$1F,$BF,$7C,$BD,$0F,$C5,$7C,$45,$9F
	.dc.b	$C1,$FC,$47,$6F,$43,$39,$7B,$EF,$F2,$13,$E2,$E0,$FC,$91,$E4,$F8
	.dc.b	$BF,$CA,$19,$C3,$F1,$B1,$97,$F1,$F1,$E3,$7E,$EE,$55,$FB,$1E,$0F
	.dc.b	$E8,$23,$DF,$AF,$AB,$2F,$A1,$B7,$BA,$FF,$88,$EE,$D3,$CF,$18,$7C
	.dc.b	$F5,$F5,$F5,$CF,$0F,$E7,$AF,$85,$FC,$0A,$FB,$9B,$FE,$B7,$1F,$9E
	.dc.b	$FE,$BF,$2F,$A3,$BE,$6B,$77,$FA,$A9,$FA,$BD,$1F,$B5,$D2,$CF,$EB
	.dc.b	$B4,$FB,$3C,$79,$C3,$F1,$57,$FB,$B9,$FB,$BC,$7F,$BD,$E4,$FB,$FC
	.dc.b	$16,$FF,$21,$3F,$C1,$7F,$80,$7F,$DA,$BC,$3E,$BA,$D6,$7F,$13,$97
	.dc.b	$5E,$0F,$E2,$3B,$2E,$5F,$99,$9E,$5E,$06,$6F,$FD,$CD,$97,$FB,$77
	.dc.b	$2F,$8D,$B7,$9E,$7E,$7B,$E7,$C3,$F4,$5B,$E6,$B7,$5E,$F9,$F3,$BF
	.dc.b	$C3,$B6,$F5,$B6,$32,$FF,$B8,$70,$F5,$C7,$EA,$75,$F5,$DF,$D7,$E9
	.dc.b	$F8,$1E,$07,$E8,$9F,$85,$E0,$7A,$71,$E5,$39,$7D,$71,$F4,$9F,$E1
	.dc.b	$7C,$75,$FC,$7F,$7B,$93,$EA,$C7,$F2,$49,$D3,$FD,$2A,$75,$9F,$A2
	.dc.b	$38,$EB,$EC,$7F,$8B,$BF,$69,$F9,$66,$4D,$C7,$F0,$9E,$EF,$B5,$E0
	.dc.b	$7E,$F1,$B1,$FC,$37,$BF,$EF,$78,$59,$BF,$AB,$7F,$C1,$BF,$0D,$FE
	.dc.b	$C7,$1F,$63,$CB,$C7,$97,$F8,$57,$F8,$3F,$8B,$F8,$7F,$3F,$35,$BC
	.dc.b	$FF,$8D,$7C,$4F,$E2,$FD,$0F,$E1,$AF,$E3,$7D,$0F,$E1,$FE,$4E,$6B
	.dc.b	$77,$F8,$69,$F1,$7C,$4F,$E4,$FD,$1B,$3F,$8A,$F3,$BC,$DE,$8E,$70
	.dc.b	$FE,$F7,$D9,$FD,$3B,$87,$FD,$57,$7B,$9C,$9A,$8F,$6C,$7E,$17,$83
	.dc.b	$D2,$D9,$FE,$65,$C1,$ED,$78,$C2,$D7,$CA,$3F,$86,$F7,$7F,$84,$6B
	.dc.b	$D7,$1F,$91,$E0,$F8,$2F,$B6,$3F,$9F,$6F,$F0,$63,$F8,$7F,$07,$C5
	.dc.b	$9E,$F8,$F6,$F3,$FF,$69,$EA,$7D,$5F,$4E,$D6,$B4,$57,$EE,$7D,$BD
	.dc.b	$65,$FB,$0C,$3E,$E7,$AB,$FB,$DF,$72,$CF,$ED,$1F,$C0,$7A,$BF,$C0
	.dc.b	$7D,$AC,$CF,$EA,$67,$F0,$9E,$BF,$F0,$5F,$6F,$F0,$DE,$C7,$F0,$7F
	.dc.b	$6B,$37,$FB,$E9,$CB,$FC,$02,$3F,$19,$EE,$7F,$0D,$F7,$3F,$1D,$EE
	.dc.b	$7F,$0F,$F8,$4C,$DF,$F0,$69,$CB,$FC,$22,$3C,$6F,$CE,$CE,$BF,$E3
	.dc.b	$2E,$94,$7A,$A3,$F8,$7E,$BA,$5E,$DA,$B7,$CB,$F2,$77,$F6,$3F,$2F
	.dc.b	$7F,$63,$9B,$F6,$39,$7D,$8F,$D1,$F8,$3F,$E5,$1F,$0F,$FA,$96,$C7
	.dc.b	$D5,$F8,$0F,$F1,$87,$F0,$0F,$FB,$A6,$FF,$15,$D2,$BD,$C7,$F9,$82
	.dc.b	$1F,$0F,$C3,$E2,$3F,$69,$F0,$FC,$A3,$F6,$C6,$72,$FC,$14,$65,$FC
	.dc.b	$27,$C3,$F8,$23,$F7,$9F,$0F,$67,$38,$FF,$86,$CF,$F2,$DE,$6B,$DC
	.dc.b	$64,$F9,$30,$BF,$E0,$DF,$0F,$EC,$C3,$F8,$47,$C4,$7B,$A3,$37,$C7
	.dc.b	$8D,$82,$F7,$39,$85,$7B,$23,$1A,$7C,$07,$5A,$65,$F4,$61,$D2,$75
	.dc.b	$18,$BF,$F7,$CF,$85,$FC,$0F,$4F,$81,$3F,$37,$2B,$39,$8F,$14,$FC
	.dc.b	$07,$96,$9F,$C8,$F5,$FB,$3A,$D9,$F4,$6F,$FC,$9F,$3C,$D7,$EC,$CF
	.dc.b	$BF,$1E,$79,$FF,$08,$E6,$AF,$3C,$63,$5D,$3F,$4A,$32,$FD,$48,$FE
	.dc.b	$69,$97,$A3,$0F,$E3,$5A,$5A,$FC,$5C,$FE,$38,$F5,$3E,$0C,$FF,$C8
	.dc.b	$3F,$88,$3A,$3F,$27,$C5,$8E,$3B,$97,$F1,$0A,$D6,$1F,$57,$AF,$4F
	.dc.b	$D7,$BE,$7D,$1F,$9C,$EB,$B3,$F6,$FB,$FB,$F7,$67,$2E,$97,$FA,$29
	.dc.b	$F1,$F7,$7E,$93,$BF,$CB,$DF,$FA,$6E,$DC,$E5,$F8,$59,$CB,$F8,$78
	.dc.b	$F5,$F8,$3F,$AC,$EE,$B3,$F8,$B7,$DF,$E0,$7E,$C3,$C2,$CC,$FE,$16
	.dc.b	$7C,$EF,$E1,$F7,$AF,$D6,$D3,$D2,$FE,$56,$32,$FE,$9A,$3C,$AF,$E3
	.dc.b	$F2,$F2,$BF,$91,$CB,$2F,$E0,$23,$2F,$E0,$A3,$E9,$F9,$5F,$C1,$3C
	.dc.b	$4F,$AB,$E2,$5E,$1F,$A5,$3F,$C2,$3C,$CC,$E1,$FA,$7B,$FB,$7E,$77
	.dc.b	$F0,$CF,$32,$CF,$EA,$FC,$FF,$BF,$E5,$E7,$2F,$B9,$7F,$C4,$A7,$F0
	.dc.b	$7E,$7F,$FB,$0F,$B7,$FD,$8D,$D9,$FE,$CE,$C7,$F5,$B1,$FE,$C2,$7F
	.dc.b	$DE,$70,$9E,$1E,$6F,$E3,$3C,$FE,$5E,$6F,$E3,$BD,$1E,$9E,$7D,$C7
	.dc.b	$F8,$16,$5F,$EC,$3D,$7F,$BB,$E7,$FE,$F7,$9C,$FA,$7B,$7C,$37,$EC
	.dc.b	$7F,$79,$C1,$5F,$D5,$ED,$EE,$7F,$85,$46,$5F,$E3,$71,$EE,$7C,$3A
	.dc.b	$F9,$DF,$1E,$99,$7D,$F1,$97,$E0,$C7,$EA,$3D,$8F,$8D,$EA,$59,$F5
	.dc.b	$FB,$5F,$AC,$F5,$73,$B7,$EA,$6F,$E6,$4F,$FA,$C3,$8F,$FC,$85,$7F
	.dc.b	$0B,$E0,$31,$ED,$7D,$1F,$7B,$5E,$CF,$D3,$9C,$D7,$C8,$9F,$DC,$BF
	.dc.b	$5F,$DE,$FD,$DF,$C0,$DD,$3F,$71,$3F,$6F,$E0,$B2,$FD,$0A,$FE,$03
	.dc.b	$F0,$3F,$7F,$E0,$6C,$FD,$47,$F8,$2F,$C0,$FE,$0F,$E0,$B3,$3F,$82
	.dc.b	$9F,$1B,$F8,$4D,$2B,$FE,$AC,$EB,$FF,$75,$67,$FE,$18,$E0,$FF,$68
	.dc.b	$CE,$E3,$1C,$1E,$F7,$CF,$1E,$57,$F8,$86,$DE,$B8,$FC,$79,$EB,$8C
	.dc.b	$BB,$3F,$C6,$4F,$F9,$02,$9F,$07,$F9,$BF,$85,$FA,$0F,$E7,$7E,$17
	.dc.b	$E9,$7C,$2E,$6F,$CD,$39,$7C,$B8,$7D,$56,$BF,$A2,$E3,$FA,$AE,$4F
	.dc.b	$09,$FE,$43,$C7,$EE,$BB,$9F,$6B,$FA,$AE,$2F,$A6,$D8,$CB,$F2,$78
	.dc.b	$F2,$FC,$B9,$FC,$47,$C4,$7E,$D7,$E2,$6C,$FE,$A9,$FC,$67,$C4,$7E
	.dc.b	$E6,$73,$3F,$B8,$9F,$C8,$72,$7E,$F6,$AC,$FE,$C9,$FC,$A6,$FF,$C0
	.dc.b	$75,$CC,$FF,$00,$9F,$CC,$6F,$FC,$17,$7F,$CD,$63,$75,$FB,$D3,$FC
	.dc.b	$27,$87,$38,$7E,$FA,$FF,$41,$B7,$F0,$DE,$7F,$D1,$72,$FF,$82,$EB
	.dc.b	$FC,$67,$FC,$03,$FE,$B4,$7E,$55,$6F,$2F,$C5,$56,$FE,$11,$7B,$F8
	.dc.b	$47,$23,$8B,$DD,$87,$E4,$5B,$97,$D6,$AD,$F1,$7B,$EB,$53,$F0,$DF
	.dc.b	$5B,$E9,$7F,$35,$C1,$EF,$8E,$58,$38,$3E,$05,$6A,$78,$1A,$E5,$F4
	.dc.b	$67,$FE,$DF,$9F,$CB,$72,$E3,$AB,$F8,$25,$6B,$9F,$F8,$34,$E6,$BF
	.dc.b	$55,$E2,$72,$EB,$FE,$15,$DF,$67,$39,$FF,$8B,$73,$77,$6A,$31,$7F
	.dc.b	$C4,$7B,$BC,$7D,$FF,$C4,$FB,$FC,$BE,$36,$6F,$94,$E5,$E9,$1F,$C6
	.dc.b	$7C,$1F,$5F,$C4,$59,$F0,$3F,$C7,$7C,$6F,$7F,$2E,$67,$FC,$6B,$AF
	.dc.b	$FA,$8D,$F9,$3C,$1E,$97,$E5,$F0,$5E,$87,$0F,$EE,$EB,$E4,$F0,$FA
	.dc.b	$EF,$9F,$0F,$9D,$EB,$BF,$9D,$EF,$BE,$51,$3A,$BF,$13,$3B,$8C,$9F
	.dc.b	$92,$9F,$0B,$F8,$7F,$C4,$BF,$D7,$8F,$1D,$5E,$5F,$F3,$BF,$1B,$EE
	.dc.b	$F9,$7C,$37,$98,$70,$C3,$F3,$B7,$B7,$EB,$56,$F7,$FF,$6B,$D3,$D0
	.dc.b	$FC,$3F,$99,$9B,$FF,$67,$4E,$5F,$BD,$E7,$E7,$24,$BD,$4E,$38,$FF
	.dc.b	$B2,$23,$85,$FD,$2B,$E1,$1F,$EC,$68,$FC,$47,$A3,$8F,$43,$F2,$EE
	.dc.b	$BD,$5F,$CC,$FA,$F9,$AD,$E3,$D7,$D3,$FC,$EF,$AF,$D3,$D3,$FC,$FF
	.dc.b	$B5,$DA,$73,$5E,$39,$CB,$E5,$E1,$F3,$E1,$F3,$DF,$D3,$ED,$E6,$DF
	.dc.b	$3C,$1F,$32,$73,$7E,$DD,$7C,$EF,$BF,$7B,$8F,$D9,$9F,$4C,$E6,$3D
	.dc.b	$27,$DD,$F5,$FF,$6B,$ED,$FC,$46,$CF,$C9,$F6,$3F,$73,$ED,$E6,$FF
	.dc.b	$5D,$7F,$2B,$2F,$2D,$FD,$A7,$CB,$3F,$89,$D3,$2F,$DA,$8C,$E9,$F8
	.dc.b	$6C,$FC,$EF,$E2,$34,$B8,$7E,$12,$7D,$37,$F9,$39,$CB,$F8,$4F,$77
	.dc.b	$F8,$6F,$B9,$F7,$1B,$7F,$81,$F7,$3F,$88,$FB,$59,$BF,$E0,$D7,$F8
	.dc.b	$2F,$87,$FE,$2B,$EF,$D9,$FF,$5B,$7F,$00,$FF,$B5,$75,$FC,$6F,$C4
	.dc.b	$6B,$3F,$C6,$61,$FC,$CE,$B5,$D9,$FE,$D9,$C3,$8F,$C3,$7E,$3B,$E2
	.dc.b	$34,$FE,$43,$E0,$F8,$7C,$3E,$6B,$79,$7E,$5F,$E1,$BA,$7C,$37,$E5
	.dc.b	$9B,$1F,$9B,$F8,$7F,$04,$66,$BC,$F5,$E9,$7F,$87,$F0,$57,$F7,$BE
	.dc.b	$5B,$6F,$3F,$07,$AE,$7F,$7D,$C1,$EB,$CF,$C9,$1F,$EC,$CA,$FF,$5B
	.dc.b	$47,$EA,$F7,$F5,$BF,$AD,$DB,$D6,$E6,$BD,$67,$F0,$9D,$BE,$36,$DF
	.dc.b	$B0,$D7,$E4,$65,$FB,$29,$CD,$FE,$DE,$72,$FF,$18,$E1,$FE,$19,$F8
	.dc.b	$AF,$FA,$CB,$1F,$26,$BF,$C0,$61,$BF,$D3,$D7,$1C,$9F,$56,$B5,$39
	.dc.b	$C7,$F8,$21,$F6,$76,$FE,$07,$C5,$63,$39,$7E,$0B,$9B,$2F,$E1,$2B
	.dc.b	$C1,$D9,$F8,$1E,$7B,$3F,$88,$7C,$9C,$FF,$85,$EE,$CC,$FC,$B9,$F2
	.dc.b	$BF,$8E,$D3,$CA,$FE,$43,$6C,$BE,$EC,$FE,$11,$F0,$F7,$F7,$C6,$E3
	.dc.b	$CE,$FD,$7C,$FC,$ED,$8C,$BF,$82,$E5,$6B,$A8,$C5,$7C,$BF,$0B,$F3
	.dc.b	$9D,$F6,$7F,$46,$FC,$FE,$FF,$D0,$77,$66,$7F,$0D,$3F,$4F,$C7,$FD
	.dc.b	$27,$81,$F5,$7C,$7F,$D3,$78,$39,$BF,$D6,$CE,$5F,$D7,$C7,$DB,$F0
	.dc.b	$BF,$59,$E1,$59,$FD,$9B,$F7,$FC,$7F,$D8,$78,$B9,$9F,$CB,$4F,$E0
	.dc.b	$FC,$7F,$DA,$78,$DF,$85,$F1,$EF,$2F,$CA,$B7,$80,$E7,$61,$FB,$BF
	.dc.b	$87,$E1,$5F,$C0,$B1,$F4,$3F,$A3,$DF,$D0,$FE,$97,$7C,$BF,$36,$32
	.dc.b	$FC,$F8,$F3,$3F,$47,$4A,$FF,$73,$E8,$47,$B6,$3F,$88,$ED,$A5,$E3
	.dc.b	$7E,$E6,$7F,$41,$E9,$7F,$90,$FF,$80,$5E,$67,$E8,$FC,$CF,$57,$17
	.dc.b	$BF,$8B,$DD,$7B,$AD,$65,$FC,$7B,$83,$D2,$F8,$F7,$F4,$BE,$5D,$F2
	.dc.b	$F8,$B7,$FE,$DD,$97,$5E,$2F,$F7,$34,$FB,$E3,$1C,$BF,$69,$B1,$F6
	.dc.b	$9E,$1C,$BF,$69,$CC,$7D,$A7,$A7,$9D,$ED,$F5,$BF,$7C,$D8,$F0,$7A
	.dc.b	$DF,$07,$D6,$CD,$FF,$86,$8F,$F0,$F7,$1F,$FB,$DF,$8B,$F8,$4F,$B9
	.dc.b	$F1,$FD,$6F,$E1,$0F,$C9,$F5,$FF,$85,$7A,$D9,$CF,$D9,$39,$7D,$DE
	.dc.b	$BF,$CF,$F6,$7F,$88,$7B,$1F,$47,$D9,$FE,$25,$39,$AF,$89,$39,$7E
	.dc.b	$36,$BE,$76,$BE,$73,$EE,$EA,$DB,$F8,$E7,$B7,$F7,$7D,$CF,$E3,$D5
	.dc.b	$C4,$FF,$3D,$73,$35,$F4,$46,$1F,$43,$AD,$B3,$C3,$F0,$A3,$C1,$DE
	.dc.b	$CD,$C3,$EA,$9C,$C3,$87,$DF,$18,$E1,$D3,$F3,$EF,$94,$EA,$31,$4A
	.dc.b	$FF,$1B,$7C,$83,$E2,$7F,$D4,$19,$3D,$EF,$D8,$F1,$7E,$5E,$B5,$8F
	.dc.b	$C5,$AD,$78,$BC,$BE,$23,$8E,$0F,$1F,$AE,$58,$AE,$FF,$13,$AF,$3A
	.dc.b	$E5,$CF,$07,$9D,$E4,$F8,$8D,$6D,$E1,$C3,$AC,$FA,$3E,$27,$BF,$B1
	.dc.b	$71,$E5,$EC,$7B,$3E,$27,$38,$FA,$6F,$CB,$3F,$03,$E1,$3F,$6D,$A7
	.dc.b	$C2,$F8,$5F,$DC,$7C,$4E,$72,$F9,$F3,$97,$E8,$C7,$C8,$3F,$7D,$F1
	.dc.b	$36,$3E,$AB,$F3,$0F,$E0,$5F,$13,$98,$FA,$31,$F4,$23,$F8,$37,$C4
	.dc.b	$FD,$22,$E9,$F3,$E3,$F8,$56,$59,$BF,$C0,$D7,$D8,$F0,$7D,$37,$F5
	.dc.b	$72,$F4,$BF,$5E,$33,$5F,$EC,$DE,$9F,$B8,$FA,$A7,$EE,$67,$EA,$6C
	.dc.b	$7A,$9F,$C0,$67,$EA,$73,$1E,$A7,$CC,$F9,$1F,$C3,$63,$EA,$8F,$06
	.dc.b	$7E,$A9,$CC,$6C,$F2,$79,$3F,$90,$F9,$07,$C4,$7E,$53,$C6,$C7,$C4
	.dc.b	$7E,$5A,$B5,$39,$BF,$C9,$5A,$BF,$37,$E5,$73,$F8,$8B,$5A,$7F,$3F
	.dc.b	$E2,$F7,$F8,$1C,$CF,$F1,$5F,$09,$8A,$FF,$4B,$E2,$F9,$7E,$07,$F4
	.dc.b	$FE,$2D,$E0,$FD,$24,$FA,$7E,$03,$39,$79,$2F,$FD,$45,$9F,$FA,$6E
	.dc.b	$3B,$65,$FD,$FB,$0F,$E8,$53,$99,$FF,$5F,$E7,$E2,$F8,$2E,$3E,$DF
	.dc.b	$C5,$F8,$2E,$7E,$BF,$C7,$9E,$BE,$EF,$C9,$F8,$3E,$FE,$FF,$CB,$F8
	.dc.b	$2C,$E7,$F1,$A7,$2F,$C8,$8F,$2F,$BB,$F4,$7E,$0F,$F8,$1F,$BF,$78
	.dc.b	$3C,$B3,$F5,$7D,$FC,$E1,$F4,$2F,$DB,$B7,$D9,$DB,$DF,$B7,$DB,$DF
	.dc.b	$37,$F5,$A7,$2F,$D8,$8F,$8B,$AF,$F2,$F8,$FF,$71,$FC,$B3,$C8,$F9
	.dc.b	$3B,$FE,$17,$87,$75,$BA,$D6,$FF,$2A,$B4,$FC,$CD,$7F,$19,$C3,$F4
	.dc.b	$B7,$B3,$9C,$3F,$13,$C1,$97,$F1,$75,$F6,$3B,$7F,$91,$FC,$9F,$F7
	.dc.b	$AB,$FA,$FD,$D8,$E4,$D6,$1F,$C8,$EB,$FA,$E5,$7F,$99,$79,$BF,$29
	.dc.b	$39,$AF,$C1,$F6,$7B,$E1,$BF,$A9,$FF,$10,$DF,$19,$F5,$C6,$0F,$F1
	.dc.b	$2D,$8E,$55,$C7,$9F,$FD,$FD,$1F,$99,$9F,$D2,$F0,$FE,$4B,$A7,$D7
	.dc.b	$E0,$59,$F4,$77,$FE,$5B,$B3,$3A,$7E,$52,$FE,$0C,$FF,$BE,$FE,$53
	.dc.b	$FC,$FD,$7F,$96,$DF,$FA,$AD,$FF,$86,$E7,$FC,$C3,$3F,$AB,$9C,$BE
	.dc.b	$9D,$13,$CA,$3F,$D1,$53,$FD,$6E,$B9,$7C,$47,$D0,$F0,$7F,$53,$1E
	.dc.b	$EE,$9F,$D5,$FC,$47,$C4,$CB,$DF,$39,$3F,$57,$19,$7E,$97,$C5,$FD
	.dc.b	$AF,$0F,$F6,$6F,$DC,$F2,$BF,$6B,$F1,$79,$AF,$D9,$CE,$5F,$67,$AD
	.dc.b	$F8,$2F,$37,$E1,$3F,$84,$F3,$3E,$27,$B1,$9A,$DF,$0F,$E0,$E7,$E4
	.dc.b	$7A,$FE,$58,$F8,$5C,$68,$F3,$47,$F9,$82,$3F,$B0,$BF,$9D,$E0,$E2
	.dc.b	$FD,$AB,$F9,$DE,$A7,$E5,$3C,$FF,$A8,$FE,$5B,$CE,$FA,$D7,$9A,$DD
	.dc.b	$FE,$56,$7E,$D7,$C3,$FE,$73,$D0,$FB,$9F,$11,$F9,$EF,$4F,$37,$F9
	.dc.b	$B9,$CB,$F9,$D8,$FC,$17,$C5,$7A,$FA,$19,$7F,$B0,$F6,$4F,$B2,$7E
	.dc.b	$FF,$C5,$6E,$B5,$39,$A7,$C4,$FE,$BB,$D7,$FC,$73,$6B,$F1,$9F,$13
	.dc.b	$C7,$3F,$8D,$39,$7E,$5F,$C1,$FE,$DB,$DF,$FC,$B3,$FB,$8F,$80,$FC
	.dc.b	$C7,$C1,$E7,$1F,$CB,$4E,$5F,$CC,$7C,$3F,$5D,$FF,$2B,$F0,$FE,$58
	.dc.b	$C6,$C8,$F4,$7C,$4E,$0F,$15,$7C,$2D,$3E,$D7,$C4,$E5,$FE,$09,$C1
	.dc.b	$99,$9F,$2C,$F8,$5F,$03,$6B,$F8,$E7,$C6,$6F,$5F,$F7,$74,$BE,$1F
	.dc.b	$77,$FC,$02,$B5,$7F,$EA,$8A,$7C,$2F,$4F,$86,$FA,$51,$89,$FC,$66
	.dc.b	$3F,$E9,$18,$4F,$D3,$9F,$7F,$C5,$FB,$2B,$EE,$46,$0F,$5B,$C7,$BB
	.dc.b	$D2,$78,$DE,$9D,$98,$9F,$5C,$77,$EC,$F7,$9E,$4D,$EC,$FE,$74,$FE
	.dc.b	$0D,$97,$F9,$39,$FF,$39,$7A,$BF,$C7,$6F,$E1,$F4,$79,$67,$E0,$74
	.dc.b	$FC,$A8,$C7,$13,$0F,$F3,$35,$7C,$03,$AF,$97,$E7,$7B,$FC,$2F,$D2
	.dc.b	$8C,$57,$D6,$F3,$3E,$A9,$F2,$BC,$AB,$7E,$79,$FB,$7E,$47,$DA,$3C
	.dc.b	$38,$BC,$7C,$57,$C8,$3F,$91,$7E,$64,$F7,$7F,$03,$ED,$78,$38,$15
	.dc.b	$FE,$E2,$AD,$3E,$9C,$BF,$13,$5B,$8F,$C0,$7C,$3F,$E2,$8F,$BF,$EC
	.dc.b	$58,$F3,$61,$F8,$AF,$5B,$F1,$FE,$A7,$99,$B3,$E5,$C7,$F2,$D9,$3E
	.dc.b	$03,$F4,$8D,$9F,$B6,$7A,$BE,$03,$33,$E3,$A5,$7F,$86,$EB,$D6,$7F
	.dc.b	$5E,$F3,$FF,$2F,$3A,$F8,$4F,$E0,$D5,$FE,$EE,$AF,$76,$7F,$0A,$B5
	.dc.b	$1E,$A3,$E2,$7B,$9E,$43,$EF,$D3,$E0,$71,$3E,$BB,$E5,$2F,$73,$D6
	.dc.b	$79,$D7,$BB,$8A,$F6,$E3,$F0,$BE,$1B,$E0,$9E,$6F,$86,$B3,$E3,$9F
	.dc.b	$7F,$C2,$FC,$63,$C7,$F0,$B6,$7C,$47,$DB,$F7,$7E,$5F,$BD,$E1,$6C
	.dc.b	$78,$35,$F3,$FC,$2F,$D0,$3E,$2F,$C2,$D9,$F0,$6D,$FD,$F5,$FD,$2F
	.dc.b	$1F,$CC,$8C,$47,$F7,$5F,$1B,$1F,$09,$F6,$BE,$17,$F4,$BE,$1B,$DF
	.dc.b	$FE,$40,$FF,$8B,$EF,$C4,$38,$51,$B8,$FA,$A6,$5E,$0F,$ED,$BD,$0F
	.dc.b	$BB,$1F,$DF,$63,$FC,$77,$9E,$27,$FA,$24,$FF,$81,$A7,$F0,$A7,$E1
	.dc.b	$76,$FC,$01,$F8,$18,$E5,$7F,$E7,$4E,$8F,$74,$FE,$50,$FC,$0F,$C1
	.dc.b	$DB,$F9,$F5,$E8,$F6,$FD,$A7,$E5,$70,$7B,$98,$AF,$7D,$7B,$BE,$0F
	.dc.b	$E1,$9E,$6F,$83,$B3,$F7,$EF,$FB,$C3,$F0,$AB,$E2,$FC,$4F,$CC,$F8
	.dc.b	$4F,$96,$FC,$EF,$6F,$E6,$FB,$7E,$16,$DF,$83,$07,$C2,$7D,$2F,$84
	.dc.b	$F8,$AD,$9F,$07,$C1,$FD,$6F,$84,$FB,$2D,$8F,$7D,$FF,$92,$DF,$94
	.dc.b	$7F,$8B,$DC,$7A,$EE,$5F,$AB,$3F,$80,$F6,$FE,$27,$C1,$FE,$08,$FA
	.dc.b	$A7,$E1,$3E,$13,$EB,$FC,$27,$E1,$BE,$0F,$F1,$27,$E2,$1F,$E5,$FF
	.dc.b	$99,$FF,$37,$C3,$ED,$E5,$FC,$DD,$FC,$54,$3D,$CF,$E0,$CD,$8F,$94
	.dc.b	$F1,$7F,$83,$CF,$9F,$AF,$F0,$71,$CB,$E0,$FF,$BB,$57,$F6,$38,$FD
	.dc.b	$93,$5F,$D1,$CF,$0E,$74,$FF,$6D,$9F,$F5,$8F,$6F,$F1,$28,$FE,$C9
	.dc.b	$1C,$0F,$F2,$B3,$87,$F5,$55,$CE,$31,$1E,$6E,$EE,$5F,$05,$F0,$BD
	.dc.b	$BF,$29,$EE,$3C,$FE,$D7,$C0,$F8,$2F,$4F,$C1,$7F,$89,$71,$3E,$C8
	.dc.b	$C7,$C1,$7B,$BC,$8F,$9A,$7C,$07,$E7,$FB,$1F,$08,$F9,$67,$C4,$F6
	.dc.b	$7E,$6F,$C0,$FC,$6F,$81,$EC,$7C,$87,$C1,$D9,$F2,$BE,$07,$EC,$9F
	.dc.b	$31,$FE,$D3,$F9,$63,$93,$FC,$48,$FF,$2D,$E3,$FD,$3B,$5F,$91,$CB
	.dc.b	$FE,$1D,$8E,$11,$AF,$81,$FA,$FF,$03,$EF,$39,$BF,$07,$E0,$7E,$DF
	.dc.b	$C0,$FE,$28,$FB,$AF,$8B,$D9,$F0,$FB,$3F,$D3,$71,$3E,$48,$C7,$C0
	.dc.b	$79,$BE,$07,$FC,$A1,$47,$D3,$18,$F8,$1F,$5F,$C4,$FB,$8F,$6B,$F0
	.dc.b	$3D,$7F,$79,$DC,$F8,$3E,$C7,$87,$E0,$7E,$1F,$C0,$F6,$3E,$2B,$E0
	.dc.b	$F8,$1F,$8E,$7C,$A3,$E4,$FC,$0F,$CC,$F1,$7E,$59,$EB,$3E,$6F,$B7
	.dc.b	$ED,$F8,$0F,$9F,$F0,$3E,$A3,$E8,$BE,$CF,$81,$FA,$7F,$03,$F5,$8F
	.dc.b	$AA,$FF,$88,$3E,$77,$F8,$AF,$57,$D4,$C7,$FB,$54,$7F,$A3,$23,$F9
	.dc.b	$B4,$7E,$96,$31,$1F,$B8,$F2,$FF,$BA,$C7,$28,$FD,$DF,$A9,$8F,$81
	.dc.b	$FF,$33,$4F,$F4,$09,$FE,$FB,$DD,$FC,$D7,$0E,$98,$6A,$FF,$D0,$57
	.dc.b	$FD,$0E,$FF,$8A,$DF,$F9,$8E,$3F,$9E,$FC,$53,$C8,$F0,$E5,$DB,$E2
	.dc.b	$BD,$9E,$C7,$83,$D9,$FF,$07,$60,$7C,$71,$8F,$83,$F2,$FC,$1F,$80
	.dc.b	$F3,$BE,$2F,$83,$F4,$9F,$1C,$F5,$FC,$1F,$C9,$F6,$7F,$D0,$DB,$FF
	.dc.b	$B5,$F2,$47,$FA,$1E,$30,$EB,$7F,$57,$B5,$97,$D9,$F0,$1F,$1B,$E0
	.dc.b	$7E,$09,$F2,$1F,$87,$F0,$1F,$2B,$E0,$7E,$C1,$F3,$1F,$F2,$CC,$FF
	.dc.b	$0E,$C7,$FB,$97,$27,$F0,$6E,$7F,$E3,$93,$FE,$78,$9F,$F5,$A4,$63
	.dc.b	$DB,$FF,$5E,$3A,$F8,$8F,$F5,$E3,$F7,$E7,$EC,$9F,$50,$FB,$7F,$05
	.dc.b	$F5,$BE,$0B,$F1,$5F,$05,$27,$FA,$F7,$0A,$7C,$BE,$D7,$FB,$7E,$7F
	.dc.b	$DF,$99,$A3,$FD,$C9,$8F,$99,$D7,$07,$FA,$86,$7E,$F5,$7F,$B9,$6B
	.dc.b	$D7,$F1,$1F,$EB,$19,$FE,$85,$3F,$E1,$1F,$8B,$FE,$25,$5A,$8F,$EE
	.dc.b	$13,$FD,$25,$FE,$77,$1F,$13,$E0,$BF,$DC,$11,$E4,$7F,$DC,$91,$98
	.dc.b	$FF,$07,$9F,$33,$E0,$7F,$CA,$75,$FD,$26,$3F,$37,$C2,$F7,$F3,$ED
	.dc.b	$70,$9F,$A7,$EF,$70,$F7,$31,$B3,$04,$E6,$BB,$CE,$7D,$DE,$71,$8A
	.dc.b	$DE,$9F,$73,$E0,$BE,$16,$BF,$EE,$9D,$7F,$D1,$57,$F0,$27,$1A,$F8
	.dc.b	$30,$7B,$FE,$A8,$C6,$D6,$FB,$4F,$BB,$E0,$73,$9F,$E2,$BD,$CE,$5E
	.dc.b	$FF,$C2,$6C,$66,$FC,$D1,$F1,$BD,$EC,$C7,$E0,$3E,$07,$EC,$1E,$73
	.dc.b	$ED,$71,$7A,$78,$7C,$D5,$F3,$9F,$44,$DA,$F8,$53,$F4,$B3,$CE,$3F
	.dc.b	$DE,$7E,$7F,$FD,$15,$5F,$83,$F6,$BF,$D1,$31,$8E,$47,$3B,$DB,$CD
	.dc.b	$7D,$48,$3C,$23,$07,$2F,$82,$FF,$80,$17,$FD,$6F,$0F,$E3,$77,$F9
	.dc.b	$07,$FC,$C5,$8F,$FA,$4E,$75,$1F,$E0,$18,$F3,$57,$73,$CB,$F0,$5E
	.dc.b	$1F,$89,$FC,$6F,$15,$AF,$DC,$BE,$59,$DD,$FE,$1B,$1F,$F4,$D5,$7C
	.dc.b	$1F,$6B,$D4,$7C,$0F,$6B,$D8,$5A,$E1,$1F,$10,$CC,$FC,$9F,$82,$F8
	.dc.b	$47,$C8,$F8,$3F,$88,$5A,$F0,$47,$CC,$33,$3F,$47,$E0,$7E,$51,$F4
	.dc.b	$3E,$0A,$C7,$D5,$7E,$71,$F5,$23,$35,$F2,$0F,$F4,$56,$EF,$6B,$EA
	.dc.b	$1F,$6B,$DB,$FA,$C5,$9F,$85,$1F,$78,$CE,$1F,$83,$F8,$2F,$B8,$7E
	.dc.b	$08,$AE,$67,$E5,$7C,$17,$E1,$8C,$BE,$3E,$66,$3F,$EC,$38,$FA,$1F
	.dc.b	$03,$F8,$C3,$F1,$05,$7E,$A7,$C1,$7E,$40,$CE,$1F,$95,$F8,$0F,$E0
	.dc.b	$9F,$1E,$F8,$FF,$EB,$53,$9B,$FE,$1F,$39,$F2,$FF,$33,$7C,$2F,$75
	.dc.b	$E5,$C9,$F0,$79,$BF,$F7,$3F,$D0,$FF,$AC,$6B,$FD,$4F,$8F,$7F,$89
	.dc.b	$CF,$C2,$73,$F3,$ED,$78,$33,$FF,$38,$6B,$FE,$63,$C7,$FB,$C4,$86
	.dc.b	$3E,$9B,$EF,$F0,$7E,$5F,$8E,$B5,$CF,$D7,$AF,$8B,$24,$F9,$BE,$13
	.dc.b	$D9,$5F,$B6,$CF,$D4,$7E,$EC,$B7,$EF,$7E,$17,$C2,$FE,$F3,$E1,$F3
	.dc.b	$CC,$9F,$32,$DB,$E1,$5F,$A4,$F4,$E7,$EE,$7E,$F4,$69,$F7,$6D,$ED
	.dc.b	$BF,$85,$BF,$D1,$E0,$F9,$BF,$0F,$F2,$67,$E3,$70,$7C,$16,$E1,$F0
	.dc.b	$5A,$FC,$1E,$0F,$80,$E4,$F0,$FC,$67,$AA,$7D,$39,$FF,$8A,$E7,$C7
	.dc.b	$A3,$DD,$FB,$B7,$B9,$FF,$78,$57,$F8,$D1,$DC,$FD,$C7,$E8,$67,$5F
	.dc.b	$85,$AF,$B5,$F8,$3A,$FB,$72,$FC,$06,$BE,$D9,$F8,$BA,$FB,$5A,$7A
	.dc.b	$76,$FA,$0F,$8A,$35,$18,$3E,$97,$17,$DE,$BF,$C6,$7C,$3F,$F6,$2F
	.dc.b	$65,$F0,$BF,$98,$F8,$8C,$37,$2F,$EC,$72,$F8,$6F,$CE,$DE,$B0,$FC
	.dc.b	$B7,$C3,$F3,$E3,$FD,$0B,$D7,$E3,$2E,$3C,$A7,$F4,$BF,$13,$9B,$FC
	.dc.b	$F6,$7E,$3E,$1F,$D4,$BE,$5E,$1B,$87,$8A,$7F,$5B,$A6,$6F,$F4,$DF
	.dc.b	$0B,$FB,$CE,$5F,$D3,$47,$EA,$7E,$1B,$18,$7E,$D1,$D7,$C3,$7E,$DA
	.dc.b	$F3,$5F,$AF,$F8,$9E,$5F,$0B,$FB,$AF,$89,$E9,$F1,$17,$2F,$E0,$B9
	.dc.b	$FF,$8B,$70,$E1,$18,$9F,$E0,$3F,$13,$E3,$C6,$CF,$06,$BF,$C1,$2D
	.dc.b	$87,$F0,$8F,$86,$E1,$8F,$F0,$1C,$BD,$98,$FF,$0C,$6C,$67,$0F,$19
	.dc.b	$FC,$43,$2C,$BF,$AC,$CB,$D7,$3F,$C2,$F1,$F5,$B6,$7D,$77,$FC,$62
	.dc.b	$B3,$1F,$B4,$C3,$C6,$D3,$D5,$5F,$C5,$63,$E8,$5F,$A6,$72,$7E,$F0
	.dc.b	$FA,$98,$FF,$24,$7E,$B6,$75,$F9,$71,$FC,$A3,$4C,$E1,$FC,$28,$FE
	.dc.b	$17,$F9,$AB,$C5,$FB,$DB,$FE,$76,$73,$E1,$61,$E1,$8B,$B5,$87,$F9
	.dc.b	$1E,$3F,$86,$F0,$BA,$B9,$46,$E7,$B3,$F7,$B1,$CD,$F2,$F1,$7F,$1D
	.dc.b	$DB,$8B,$FF,$40,$78,$1F,$C1,$63,$13,$F9,$3E,$DD,$67,$C3,$4F,$2E
	.dc.b	$3E,$4F,$1B,$F3,$1D,$9C,$FC,$2B,$5A,$D3,$FC,$D1,$B7,$F0,$8C,$5C
	.dc.b	$9E,$1C,$77,$3F,$D9,$EB,$D2,$FE,$37,$7F,$37,$0F,$C5,$8D,$3F,$A7
	.dc.b	$E1,$F5,$71,$67,$6F,$E1,$18,$7B,$33,$FF,$5C,$69,$E3,$E0,$F5,$65
	.dc.b	$DD,$F7,$47,$9E,$73,$7C,$BE,$27,$F6,$9D,$DF,$1F,$C7,$FD,$B7,$6D
	.dc.b	$CB,$F5,$73,$F2,$EB,$38,$FC,$1F,$85,$FD,$E7,$83,$F3,$FC,$3F,$DF
	.dc.b	$78,$77,$1F,$D9,$CF,$D3,$AC,$DF,$C9,$F8,$8F,$E0,$9E,$1F,$D7,$6C
	.dc.b	$66,$FF,$77,$E2,$7D,$BF,$89,$CC,$7C,$BF,$88,$FE,$19,$E4,$7D,$F6
	.dc.b	$DF,$D5,$7F,$88,$79,$1F,$83,$F8,$8C,$DF,$D3,$D3,$D3,$7F,$6F,$7F
	.dc.b	$2C,$FD,$77,$2E,$CF,$E1,$BB,$F9,$DF,$E1,$FB,$57,$F8,$43,$5F,$C2
	.dc.b	$6B,$E9,$77,$7F,$89,$F8,$6F,$F0,$FF,$F0,$0F,$F7,$58,$FC,$67,$C2
	.dc.b	$E1,$FC,$D3,$AF,$3F,$F3,$95,$99,$FA,$57,$F3,$9F,$53,$E6,$6B,$FC
	.dc.b	$6F,$8F,$E9,$39,$3F,$29,$C7,$F3,$3A,$3F,$1A,$DA,$FD,$EB,$FC,$E7
	.dc.b	$3F,$55,$7B,$A7,$F4,$5F,$0F,$EA,$F4,$FF,$61,$F1,$1E,$CF,$4E,$DF
	.dc.b	$F8,$E3,$75,$7E,$D6,$F5,$87,$EC,$3D,$87,$C1,$7C,$4F,$6B,$F7,$5F
	.dc.b	$07,$F1,$BD,$CB,$9F,$39,$FE,$03,$7C,$23,$B4,$FF,$01,$F8,$9F,$9B
	.dc.b	$3F,$C0,$FE,$27,$E7,$FB,$D6,$FC,$93,$B7,$CD,$1D,$3E,$1F,$EA,$FB
	.dc.b	$FD,$BE,$1F,$EB,$FC,$05,$BF,$64,$ED,$F7,$47,$93,$6F,$56,$FF,$C1
	.dc.b	$F7,$F5,$36,$BD,$4F,$C1,$9C,$C7,$F0,$7D,$7D,$33,$FC,$3F,$5F,$4B
	.dc.b	$67,$D2,$FF,$13,$9C,$C7,$F9,$D7,$5F,$F5,$FF,$AC,$E3,$7C,$2F,$C8
	.dc.b	$F8,$8F,$C8,$FC,$3D,$C3,$EB,$4F,$CC,$F8,$9C,$E7,$F5,$7E,$1F,$F2
	.dc.b	$E7,$D0,$3F,$33,$F1,$16,$FE,$F4,$7D,$48,$CC,$FD,$DF,$87,$FC,$F9
	.dc.b	$F6,$0A,$E6,$BF,$0D,$F1,$1F,$70,$CB,$EB,$C7,$F4,$F8,$FE,$01,$B5
	.dc.b	$F8,$97,$F5,$79,$FE,$13,$3C,$DF,$C3,$CF,$D3,$7E,$B8,$FC,$A6,$7E
	.dc.b	$97,$FD,$07,$A2,$BD,$51,$AA,$F3,$3A,$D3,$D4,$FE,$75,$AF,$2D,$3D
	.dc.b	$4E,$67,$F3,$DC,$9F,$C0,$39,$BF,$33,$CB,$67,$F4,$4F,$F0,$4E,$6F
	.dc.b	$CE,$D6,$6F,$EE,$75,$FF,$08,$E6,$FD,$0F,$67,$FA,$A3,$F8,$05,$D3
	.dc.b	$8B,$FD,$85,$7E,$9A,$F5,$3F,$A8,$EC,$E1,$D3,$FA,$AE,$CB,$B7,$EC
	.dc.b	$DE,$9D,$7B,$CB,$FD,$63,$9A,$FF,$63,$D7,$E1,$ED,$B5,$AE,$EA,$FF
	.dc.b	$AB,$AD,$3F,$B8,$EE,$F4,$57,$96,$BC,$F9,$FF,$03,$E0,$F3,$B6,$32
	.dc.b	$7C,$23,$C7,$B7,$95,$ED,$9F,$AD,$F3,$B5,$F1,$69,$B7,$FC,$FF,$9F
	.dc.b	$AE,$FF,$89,$1F,$2F,$C3,$FE,$1A,$FC,$DF,$16,$D7,$C2,$9F,$E2,$3E
	.dc.b	$2E,$6B,$F8,$D7,$17,$D5,$7D,$15,$F0,$73,$FA,$8F,$E3,$B8,$FE,$03
	.dc.b	$99,$FE,$45,$C7,$F4,$A3,$F7,$5C,$77,$C4,$FF,$44,$72,$23,$EA,$DF
	.dc.b	$0A,$F7,$BF,$E1,$AF,$CE,$5E,$77,$E1,$7C,$BF,$F1,$3C,$3C,$DC,$56
	.dc.b	$DF,$C5,$79,$9A,$BC,$E1,$FE,$FF,$BF,$F5,$83,$FE,$DE,$E3,$79,$3D
	.dc.b	$3C,$CF,$C9,$79,$F6,$BF,$0F,$3E,$0F,$33,$3A,$79,$7C,$DF,$CC,$79
	.dc.b	$FE,$47,$F3,$5E,$7D,$C3,$F1,$F3,$E8,$F3,$F3,$7E,$DF,$47,$F4,$1E
	.dc.b	$8F,$B1,$FD,$17,$A7,$70,$FC,$BC,$FC,$0F,$4F,$37,$F1,$74,$FD,$46
	.dc.b	$BE,$7B,$FC,$B6,$D5,$FC,$2C,$7A,$5F,$D0,$C7,$95,$FC,$E6,$95,$FC
	.dc.b	$5E,$B5,$FF,$73,$F0,$27,$D5,$1F,$AA,$9F,$A7,$EF,$FE,$E7,$DC,$FA
	.dc.b	$BE,$FF,$EE,$FD,$FB,$7F,$B0,$9D,$BF,$B2,$8F,$B7,$EF,$FF,$01,$F7
	.dc.b	$2C,$FE,$D9,$AF,$EC,$7D,$EF,$C0,$FB,$FB,$E1,$FC,$37,$BF,$FC,$27
	.dc.b	$DE,$FC,$2B,$FC,$2F,$DE,$B8,$7F,$00,$9F,$C4,$FC,$06,$6F,$F1,$DF
	.dc.b	$08,$A8,$7F,$BA,$BF,$80,$7F,$D7,$EF,$E2,$7E,$13,$2E,$4F,$C9,$B8
	.dc.b	$79,$7C,$12,$5F,$09,$F9,$8B,$D5,$2B,$4F,$B8,$FD,$DE,$F5,$FA,$2D
	.dc.b	$7F,$31,$BF,$C2,$7E,$36,$5F,$09,$F7,$35,$FE,$31,$AF,$C2,$7F,$8D
	.dc.b	$6B,$F0,$9F,$E4,$71,$FA,$CF,$86,$F2,$E3,$E9,$F8,$7C,$57,$B6,$31
	.dc.b	$87,$ED,$6B,$7A,$BE,$23,$F6,$BF,$11,$8C,$7F,$D3,$BF,$9E,$7F,$A6
	.dc.b	$A9,$78,$CF,$8B,$73,$ED,$3F,$ED,$8C,$BF,$D3,$77,$C6,$BF,$80,$D7
	.dc.b	$83,$7A,$F0,$E1,$AF,$D6,$D7,$E9,$4F,$F0,$8E,$1F,$3D,$7F,$0A,$D3
	.dc.b	$D3,$CB,$FC,$33,$7F,$5F,$2D,$CB,$C7,$3B,$7C,$B1,$EC,$E1,$F4,$ED
	.dc.b	$E6,$E0,$B5,$FE,$C5,$8F,$C6,$79,$CE,$05,$7C,$8E,$2E,$57,$F0,$38
	.dc.b	$7D,$2F,$E4,$78,$2B,$F8,$CE,$27,$06,$23,$F1,$50,$9F,$74,$7C,$3C
	.dc.b	$FE,$4F,$27,$D6,$E6,$FA,$6D,$8C,$CF,$D0,$E6,$FA,$FC,$D9,$8F,$51
	.dc.b	$FE,$DF,$DB,$E9,$57,$BE,$3D,$0D,$7E,$86,$5F,$65,$E3,$BF,$C1,$E0
	.dc.b	$F8,$AA,$BE,$2F,$91,$FA,$4A,$F5,$EB,$ED,$E7,$FD,$56,$F7,$2F,$76
	.dc.b	$1F,$D1,$79,$FE,$DB,$F8,$6E,$9F,$C9,$F5,$FE,$43,$AA,$CF,$E2,$9F
	.dc.b	$CB,$F5,$FE,$52,$B3,$A7,$9B,$83,$F8,$F3,$F9,$2F,$17,$F3,$3D,$DF
	.dc.b	$9E,$F0,$EC,$E7,$2F,$CA,$F7,$E5,$FC,$BF,$89,$FA,$4F,$13,$F4,$3E
	.dc.b	$2D,$9F,$CD,$BF,$A8,$F2,$3F,$4B,$59,$C3,$C5,$E3,$7E,$B3,$C8,$FD
	.dc.b	$4F,$8F,$FA,$EF,$22,$D7,$E9,$A7,$F5,$B5,$9C,$BF,$4F,$B7,$ED,$37
	.dc.b	$F3,$DF,$D4,$DE,$BF,$C7,$63,$F6,$BB,$FA,$63,$E1,$69,$E3,$EB,$FA
	.dc.b	$D1,$F8,$1D,$F1,$97,$E7,$38,$51,$EB,$8F,$83,$3F,$BA,$E1,$FE,$03
	.dc.b	$E6,$7F,$06,$F2,$BF,$81,$F9,$7F,$C2,$3C,$CB,$7F,$C0,$27,$6F,$F0
	.dc.b	$28,$FE,$13,$ED,$FF,$14,$F4,$7F,$D9,$D3,$F5,$FD,$AC,$7A,$BF,$C3
	.dc.b	$9D,$69,$C2,$7F,$8C,$52,$BF,$89,$FA,$DC,$FD,$6F,$E2,$BE,$BF,$5F
	.dc.b	$6A,$E1,$C2,$76,$F2,$8F,$E3,$7E,$D7,$8F,$D8,$B3,$D9,$AF,$F1,$1C
	.dc.b	$BF,$C3,$D9,$FF,$1A,$F6,$B8,$3F,$C7,$36,$F3,$BF,$C7,$B6,$F5,$B6
	.dc.b	$FD,$2F,$AD,$E9,$B7,$91,$ED,$9F,$A1,$B3,$FC,$A3,$5F,$73,$FC,$AB
	.dc.b	$6F,$73,$FC,$50,$FE,$2B,$F9,$F7,$FB,$46,$BD,$1E,$D7,$F3,$8F,$6F
	.dc.b	$E8,$36,$3E,$5F,$B9,$F8,$4E,$A9,$7C,$67,$FB,$4F,$99,$FE,$E2,$BD
	.dc.b	$63,$DB,$E0,$BE,$D7,$BD,$8B,$62,$C5,$5F,$7F,$E0,$35,$39,$8E,$38
	.dc.b	$7A,$6F,$8D,$BE,$0B,$84,$E6,$FA,$7C,$4F,$E2,$7E,$27,$9D,$FF,$B3
	.dc.b	$39,$DF,$11,$8A,$D6,$1C,$69,$F1,$3F,$91,$AE,$3F,$0D,$F9,$3F,$87
	.dc.b	$E7,$F0,$F7,$0E,$0F,$E5,$FE,$27,$7B,$FE,$4A,$BC,$3C,$1E,$A8,$FE
	.dc.b	$4B,$9F,$C0,$7F,$2F,$C3,$A7,$F3,$3C,$5E,$8E,$2F,$6C,$70,$DF,$CE
	.dc.b	$D3,$F3,$FB,$7A,$DF,$D0,$F0,$FC,$0E,$0F,$56,$FE,$DD,$7F,$59,$AF
	.dc.b	$C4,$DA,$CF,$C1,$7F,$61,$9F,$C8,$AC,$E3,$E8,$BF,$57,$17,$E0,$9F
	.dc.b	$2B,$63,$D3,$87,$CB,$BC,$BF,$90,$D3,$D3,$5F,$8A,$7C,$B5,$EB,$3F
	.dc.b	$55,$9F,$A5,$F6,$D7,$D5,$8F,$E0,$93,$F7,$1B,$19,$9F,$E0,$39,$FE
	.dc.b	$03,$4C,$C7,$D0,$E3,$FE,$19,$C5,$F8,$4E,$5F,$E1,$DC,$76,$BF,$84
	.dc.b	$CF,$E2,$2B,$38,$FE,$17,$87,$F8,$A7,$27,$E3,$1B,$19,$BF,$F1,$47
	.dc.b	$F0,$0F,$FB,$72,$3F,$19,$C9,$A9,$FC,$0F,$7F,$F8,$C2,$BF,$D7,$51
	.dc.b	$F8,$4F,$89,$C7,$81,$F9,$87,$5E,$0F,$E6,$AF,$38,$FE,$5F,$4F,$5B
	.dc.b	$F9,$9D,$FC,$F8,$FE,$39,$CB,$F0,$4F,$E4,$BB,$FA,$5F,$E4,$FB,$57
	.dc.b	$F8,$D3,$5F,$DE,$E9,$EB,$7F,$33,$AF,$AF,$93,$F8,$1F,$85,$EB,$F0
	.dc.b	$BF,$5E,$FB,$7C,$2B,$7F,$12,$7F,$67,$E0,$E7,$5F,$DC,$78,$3F,$EC
	.dc.b	$0A,$F9,$3E,$0F,$F0,$28,$C4,$7F,$8F,$74,$4E,$7C,$4F,$95,$5B,$F1
	.dc.b	$3E,$64,$FF,$00,$C3,$E7,$78,$DF,$C0,$BC,$5F,$A1,$E3,$5C,$FE,$BC
	.dc.b	$ED,$FB,$31,$E1,$F4,$3E,$B7,$A1,$67,$EE,$FA,$55,$F2,$47,$9F,$D2
	.dc.b	$FB,$93,$E5,$9F,$5E,$5E,$ED,$BC,$AD,$8F,$C3,$ED,$E9,$76,$7F,$06
	.dc.b	$8F,$49,$FC,$57,$4F,$4B,$63,$D2,$FF,$19,$32,$FF,$8C,$A7,$FD,$C3
	.dc.b	$E4,$B8,$3F,$D7,$95,$F4,$3C,$BF,$C9,$7A,$5F,$3D,$FC,$A7,$A1,$F4
	.dc.b	$7D,$3C,$D7,$D6,$F8,$1F,$CC,$7A,$5F,$55,$B1,$9D,$3F,$43,$E8,$FD
	.dc.b	$9F,$80,$CC,$78,$FD,$9F,$D0,$7A,$FF,$75,$B7,$F8,$27,$F4,$9E,$CF
	.dc.b	$E0,$7D,$9C,$DF,$AF,$D8,$FD,$47,$C0,$FE,$15,$B8,$7E,$29,$FD,$67
	.dc.b	$C0,$FE,$27,$DB,$CD,$FC,$3D,$BC,$7E,$2F,$F5,$BE,$74,$63,$1F,$F3
	.dc.b	$CC,$62,$BD,$71,$F9,$6D,$F5,$1F,$D6,$A5,$B7,$9B,$0E,$8D,$BE,$6F
	.dc.b	$B1,$FC,$FE,$B9,$BE,$5F,$0D,$FC,$0F,$E0,$BF,$41,$F0,$DF,$C1,$7E
	.dc.b	$1A,$D7,$F0,$C9,$FD,$25,$67,$0F,$A1,$3F,$A5,$E5,$FA,$8F,$F1,$26
	.dc.b	$BF,$A9,$E5,$FA,$8E,$67,$E8,$3F,$C6,$39,$7E,$A3,$63,$2F,$EC,$79
	.dc.b	$7E,$A3,$93,$F1,$DF,$07,$FE,$CA,$FE,$01,$FF,$74,$F0,$7C,$DF,$84
	.dc.b	$C7,$C1,$FE,$ED,$D5,$FE,$F6,$F3,$8F,$F0,$2F,$89,$E4,$7F,$01,$F8
	.dc.b	$9B,$1F,$C1,$9E,$C7,$F0,$58,$CD,$7E,$7F,$E2,$3C,$47,$F0,$9F,$89
	.dc.b	$F2,$16,$3F,$DB,$9A,$7F,$B5,$30,$FE,$23,$5A,$7F,$89,$CF,$AF,$1F
	.dc.b	$E2,$4D,$8C,$BC,$B2,$FE,$2D,$9E,$63,$F6,$19,$79,$34,$FC,$76,$97
	.dc.b	$0F,$C6,$BE,$A7,$E0,$57,$92,$BF,$90,$65,$EA,$9F,$E4,$BA,$79,$6B
	.dc.b	$F7,$8E,$5F,$80,$7B,$1F,$F1,$4F,$DF,$FF,$B4,$23,$DF,$C5,$8E,$4F
	.dc.b	$AC,$D8,$F8,$7C,$BC,$38,$F3,$8F,$F8,$AF,$07,$17,$DD,$EA,$EA,$D9
	.dc.b	$D3,$F8,$1E,$AF,$0F,$76,$72,$FF,$61,$E0,$ED,$FC,$37,$5F,$9B,$AA
	.dc.b	$E5,$DD,$FC,$57,$5F,$AA,$B3,$87,$2D,$7F,$1D,$DD,$E7,$8F,$B7,$DD
	.dc.b	$51,$FE,$54,$EB,$4F,$C4,$EC,$FC,$B5,$7A,$AB,$F0,$7C,$35,$E2,$72
	.dc.b	$F1,$18,$7F,$06,$D7,$DF,$3F,$8A,$E1,$F7,$BC,$A3,$E4,$69,$E8,$EA
	.dc.b	$FF,$34,$70,$B8,$31,$5E,$69,$4E,$6F,$F2,$BC,$1F,$05,$FC,$BF,$0F
	.dc.b	$F9,$93,$ED,$9D,$BF,$AF,$F0,$F0,$FE,$C7,$C2,$B7,$F9,$E9,$E1,$E0
	.dc.b	$67,$4F,$F3,$36,$0E,$AF,$7D,$FC,$0D,$5F,$36,$E9,$F0,$75,$EA,$BD
	.dc.b	$13,$9C,$79,$CE,$67,$FC,$FB,$3F,$7E,$7B,$E1,$9C,$3A,$E1,$EB,$E6
	.dc.b	$C5,$FC,$5B,$F2,$F2,$63,$8F,$E0,$3A,$EC,$E1,$9F,$8F,$2F,$3D,$7E
	.dc.b	$83,$CE,$FE,$25,$E6,$FE,$8B,$C4,$B9,$7E,$6D,$FE,$2D,$E6,$6F,$1F
	.dc.b	$E0,$75,$FA,$8F,$37,$F8,$D7,$99,$FA,$AF,$36,$E3,$FA,$37,$F8,$F7
	.dc.b	$A1,$BC,$3F,$85,$D7,$EC,$3C,$CF,$85,$E0,$3F,$23,$FF,$6C,$F0,$7E
	.dc.b	$0A,$73,$18,$CB,$F0,$5D,$E9,$CF,$9F,$FB,$AF,$3B,$75,$E7,$CF,$EB
	.dc.b	$E5,$D7,$D5,$F7,$47,$D5,$8C,$4F,$BE,$31,$93,$D3,$E5,$1F,$EB,$88
	.dc.b	$5F,$AE,$3A,$F8,$AD,$BE,$14,$7B,$B0,$F8,$4F,$F2,$AD,$F8,$3F,$AF
	.dc.b	$9F,$84,$F8,$26,$BE,$8C,$3E,$03,$FE,$C2,$DB,$E0,$3E,$6D,$7E,$03
	.dc.b	$C6,$BF,$89,$72,$FC,$8F,$6F,$D3,$3F,$B5,$C1,$18,$86,$DF,$47,$DD
	.dc.b	$FE,$47,$EA,$71,$C7,$F8,$E7,$2F,$DC,$BF,$E2,$BC,$B7,$97,$F6,$EF
	.dc.b	$DE,$AD,$CF,$C1,$3F,$8D,$F2,$FD,$48,$FE,$4C,$FD,$CF,$73,$DF,$3F
	.dc.b	$36,$7D,$6F,$C1,$8F,$DD,$73,$FE,$0B,$DC,$FD,$E7,$4D,$BF,$DA,$BF
	.dc.b	$86,$F7,$F7,$87,$DA,$AF,$E0,$5C,$BF,$A8,$F6,$6C,$FE,$8F,$C0,$46
	.dc.b	$23,$F7,$BB,$7E,$F3,$DA,$FE,$15,$D3,$C7,$4F,$F7,$47,$DD,$1F,$D9
	.dc.b	$FB,$7F,$EF,$28,$C6,$1D,$7D,$AC,$56,$8E,$3D,$7F,$BA,$8E,$1D,$DF
	.dc.b	$9E,$F7,$39,$78,$1F,$A0,$F8,$4B,$8F,$E5,$27,$6F,$E5,$A3,$C1,$DB
	.dc.b	$FC,$1F,$8D,$F0,$5E,$3E,$DC,$7C,$05,$C3,$F8,$15,$6B,$6F,$5E,$1A
	.dc.b	$7D,$BD,$5F,$C4,$7E,$E9,$D3,$ED,$8C,$37,$0F,$95,$F1,$9F,$07,$1C
	.dc.b	$E1,$F1,$AB,$F7,$3F,$0B,$F1,$77,$B3,$F2,$9F,$DE,$FC,$57,$C9,$AC
	.dc.b	$E2,$E9,$F9,$6B,$EE,$DC,$5E,$FB,$F9,$F5,$AF,$B8,$7C,$47,$C3,$F8
	.dc.b	$DC,$E9,$FD,$5A,$73,$97,$C2,$C3,$2F,$F6,$2C,$BA,$BF,$1A,$FE,$E6
	.dc.b	$5F,$0E,$AC,$FF,$06,$9A,$FE,$23,$85,$3F,$83,$DB,$F8,$CD,$7E,$16
	.dc.b	$3F,$8D,$FC,$6D,$BF,$E1,$FF,$1B,$5F,$E0,$11,$F8,$BA,$FE,$43,$3E
	.dc.b	$69,$F8,$58,$FD,$FE,$3C,$69,$EC,$9F,$CC,$F1,$27,$3B,$FE,$16,$39
	.dc.b	$6D,$EE,$AD,$D7,$E6,$73,$FE,$5F,$F1,$BF,$9B,$CF,$F9,$8F,$C6,$DB
	.dc.b	$FE,$4B,$3B,$7F,$93,$C7,$E8,$63,$F9,$BF,$4D,$9F,$E5,$7D,$15,$FE
	.dc.b	$39,$1F,$A7,$AF,$E7,$B3,$E4,$DB,$F3,$5C,$7E,$8B,$FC,$BF,$6D,$7D
	.dc.b	$B1,$B7,$2F,$ED,$30,$FE,$8F,$F1,$9F,$B3,$C3,$FA,$6F,$83,$FE,$F3
	.dc.b	$E7,$FC,$3F,$80,$9D,$E1,$C7,$1F,$E8,$97,$B8,$FF,$08,$53,$0E,$7F
	.dc.b	$1B,$FE,$48,$EC,$EB,$CB,$71,$E3,$3F,$C0,$EB,$38,$7F,$86,$4F,$F1
	.dc.b	$3F,$27,$D1,$AF,$F0,$57,$D0,$3D,$EF,$74,$61,$AF,$F8,$86,$1E,$FF
	.dc.b	$C1,$8F,$E2,$F1,$C6,$FF,$CA,$58,$FC,$0F,$1F,$FC,$BB,$9D,$9F,$F2
	.dc.b	$B3,$F1,$3C,$7F,$F3,$0D,$67,$0F,$F2,$8D,$FC,$8F,$1F,$FC,$CB,$57
	.dc.b	$0F,$F2,$F3,$F3,$3C,$7F,$F3,$4D,$66,$FF,$71,$5E,$EE,$6F,$F0,$D6
	.dc.b	$7F,$E4,$39,$C9,$FE,$50,$ED,$FC,$C3,$F0,$BD,$AF,$7B,$FC,$06,$5E
	.dc.b	$D7,$BE,$35,$87,$CB,$79,$63,$F7,$BD,$4F,$D1,$B6,$75,$2C,$7F,$9F
	.dc.b	$BF,$7B,$D2,$E8,$FF,$96,$79,$D8,$7F,$A1,$37,$FC,$4F,$A5,$FE,$86
	.dc.b	$DE,$E5,$FE,$7F,$7F,$1B,$E9,$6F,$1F,$C4,$D7,$47,$D5,$F1,$5E,$CE
	.dc.b	$C7,$03,$E2,$B0,$FF,$AB,$6B,$E2,$9F,$BE,$3F,$C7,$0E,$F2,$F5,$F4
	.dc.b	$FE,$26,$7D,$1E,$A7,$C0,$7F,$32,$D7,$FC,$43,$BB,$D9,$F2,$7A,$C8
	.dc.b	$FF,$15,$61,$D3,$9F,$F4,$FE,$D7,$F0,$6F,$3E,$CF,$81,$FD,$5F,$C4
	.dc.b	$F2,$B5,$7B,$F7,$7D,$07,$F9,$57,$4F,$5F,$47,$DE,$7F,$CF,$50,$F7
	.dc.b	$BE,$23,$BA,$FE,$7F,$C1,$ED,$AF,$E8,$53,$BD,$3C,$D1,$FB,$6F,$89
	.dc.b	$F7,$3F,$E9,$2D,$FD,$91,$88,$FB,$9F,$17,$E7,$8F,$47,$C0,$FB,$67
	.dc.b	$F7,$ED,$7E,$3F,$C4,$EB,$0E,$F1,$FE,$04,$C7,$DF,$1F,$8C,$D7,$D9
	.dc.b	$3E,$6D,$BD,$7C,$1F,$0E,$F7,$3F,$73,$5F,$66,$BF,$47,$E0,$BC,$D3
	.dc.b	$FE,$32,$E1,$F7,$3F,$23,$BF,$F8,$CC,$FC,$FE,$2F,$B3,$FC,$03,$FE
	.dc.b	$B8,$DB,$EA,$5F,$F9,$6E,$30,$F7,$E4,$FF,$36,$C3,$E3,$F8,$CF,$97
	.dc.b	$C1,$FA,$27,$E5,$7B,$75,$87,$C0,$9F,$F4,$F6,$EA,$F0,$FC,$77,$A3
	.dc.b	$76,$4F,$8F,$C2,$9F,$3C,$7A,$3B,$FE,$76,$3F,$5E,$BF,$38,$F9,$CF
	.dc.b	$54,$7B,$BB,$FF,$19,$A7,$FA,$C7,$4F,$A1,$E4,$7B,$3E,$2B,$D1,$F1
	.dc.b	$1F,$11,$F8,$BF,$17,$E2,$3C,$71,$8A,$F6,$65,$FE,$D7,$CD,$F1,$5E
	.dc.b	$48,$C3,$5F,$F3,$2C,$7F,$19,$F0,$3F,$2B,$E0,$FC,$CA,$FA,$7E,$4F
	.dc.b	$F1,$8E,$2F,$87,$F1,$97,$0F,$E3,$39,$78,$32,$E7,$3F,$6B,$E3,$7E
	.dc.b	$39,$F2,$3E,$3F,$15,$BC,$3E,$6F,$C7,$F8,$DF,$9F,$E9,$5C,$93,$F8
	.dc.b	$48,$FC,$2F,$A5,$E1,$3C,$5F,$17,$8C,$3A,$5B,$E2,$FE,$BF,$C5,$FD
	.dc.b	$87,$F1,$AF,$E3,$BC,$3E,$C7,$87,$F8,$07,$FD,$27,$C1,$FC,$54,$FF
	.dc.b	$56,$DF,$F0,$BC,$3D,$12,$F7,$3E,$23,$63,$EC,$D7,$FA,$C6,$3F,$D5
	.dc.b	$71,$C2,$7F,$8A,$C7,$FA,$C6,$3F,$5D,$E2,$FF,$B1,$A3,$FD,$33,$18
	.dc.b	$AE,$5F,$15,$C3,$E2,$78,$FC,$57,$9F,$E1,$3D,$0F,$47,$AF,$C2,$F9
	.dc.b	$0F,$2F,$C5,$63,$1F,$4F,$93,$F9,$DC,$CD,$EC,$F2,$BE,$6F,$89,$F8
	.dc.b	$87,$C5,$F8,$9C,$5F,$C2,$F8,$9F,$83,$F1,$3F,$09,$F6,$BE,$EF,$89
	.dc.b	$ED,$F1,$3D,$DF,$82,$FC,$2F,$89,$FA,$07,$D1,$F8,$9C,$61,$F3,$BE
	.dc.b	$27,$E6,$FC,$3F,$CE,$7E,$4B,$FC,$FB,$F8,$07,$FC,$35,$CF,$FD,$98
	.dc.b	$FF,$70,$2F,$88,$FD,$D3,$A8,$FB,$87,$F2,$B7,$57,$F9,$37,$FC,$9D
	.dc.b	$E4,$FF,$36,$E1,$FF,$7C,$C6,$27,$FC,$3B,$18,$F3,$57,$C7,$AE,$BF
	.dc.b	$C8,$F9,$38,$FC,$46,$32,$FF,$3D,$5F,$4F,$86,$C6,$5F,$E7,$98,$C6
	.dc.b	$1E,$3B,$CB,$FE,$79,$C3,$C7,$F0,$D8,$8F,$F0,$06,$5E,$6F,$87,$C7
	.dc.b	$BF,$FE,$02,$9F,$19,$FE,$E2,$DB,$FC,$02,$FB,$EF,$97,$BB,$FC,$E2
	.dc.b	$7F,$91,$4F,$F3,$03,$8F,$96,$87,$A9,$C3,$87,$E2,$BC,$B4,$F8,$FE
	.dc.b	$CD,$C3,$E9,$FC,$3F,$93,$83,$E5,$B6,$32,$7D,$1F,$86,$E0,$F8,$32
	.dc.b	$FE,$F8,$FE,$8C,$C7,$C3,$7B,$B8,$7E,$AB,$C2,$3E,$E7,$C1,$57,$EA
	.dc.b	$ED,$E9,$F8,$2F,$44,$7E,$19,$AF,$77,$83,$F1,$FE,$16,$CF,$C1,$F8
	.dc.b	$6F,$44,$F9,$7E,$1A,$BF,$8B,$77,$9A,$6F,$C3,$79,$EB,$F2,$BF,$0D
	.dc.b	$5F,$73,$A7,$F2,$5F,$0D,$E4,$7F,$17,$FB,$EF,$DA,$7F,$A5,$1F,$71
	.dc.b	$C7,$77,$F3,$88,$FD,$0C,$FA,$A7,$F2,$BF,$0F,$E6,$9F,$CB,$C7,$E6
	.dc.b	$3E,$1B,$10,$8C,$3F,$CE,$DE,$2D,$3F,$B1,$1F,$4C,$F2,$98,$3F,$BE
	.dc.b	$1E,$77,$F4,$BF,$0D,$E3,$87,$C3,$5F,$86,$F7,$47,$90,$FF,$24,$1F
	.dc.b	$91,$3A,$BC,$7E,$1B,$97,$C3,$7A,$30,$FD,$8F,$C3,$5B,$3A,$7D,$1F
	.dc.b	$0D,$F3,$DF,$3B,$5F,$03,$E5,$F8,$9F,$37,$C3,$F9,$27,$E2,$0F,$87
	.dc.b	$C6,$1F,$C0,$1D,$C7,$C5,$F8,$7F,$11,$1F,$91,$F0,$FF,$07,$E1,$FC
	.dc.b	$11,$D3,$E1,$EE,$1F,$C2,$9D,$47,$CF,$F8,$7E,$EF,$FA,$90,$F8,$87
	.dc.b	$C7,$7E,$67,$C3,$FC,$DF,$87,$F8,$44,$9F,$B1,$F0,$FF,$47,$E1,$FB
	.dc.b	$FE,$2C,$F4,$E9,$9A,$FC,$04,$7F,$5D,$7B,$9F,$85,$5F,$1D,$F0,$9B
	.dc.b	$1C,$3E,$23,$F5,$3D,$CF,$A2,$7C,$4F,$E3,$A3,$FA,$7F,$C4,$5F,$4C
	.dc.b	$F0,$AD,$E1,$E2,$BF,$39,$1F,$2F,$C4,$FE,$77,$2E,$31,$E6,$E3,$43
	.dc.b	$E2,$7C,$C6,$F1,$F2,$FD,$13,$8F,$C8,$FF,$81,$F0,$FE,$F2,$7F,$28
	.dc.b	$D5,$19,$8F,$E7,$87,$EC,$1E,$4F,$C8,$F8,$AF,$9A,$47,$E5,$FC,$57
	.dc.b	$9F,$7F,$91,$EB,$C9,$F9,$F1,$E4,$EC,$F1,$F6,$49,$FF,$16,$65,$E2
	.dc.b	$9F,$81,$F1,$3F,$6B,$E2,$65,$7D,$88,$FE,$8F,$4F,$93,$7C,$0F,$F0
	.dc.b	$9A,$FB,$5F,$15,$A9,$7B,$F9,$E5,$F9,$33,$98,$C7,$2A,$38,$61,$8F
	.dc.b	$8A,$F9,$7E,$4F,$CA,$F2,$66,$7E,$8E,$54,$FD,$7F,$8A,$FC,$8F,$C5
	.dc.b	$4A,$FC,$7C,$78,$FD,$CE,$FF,$15,$E5,$F8,$6B,$5F,$EF,$43,$F8,$66
	.dc.b	$C9,$E1,$18,$F8,$AF,$BA,$FD,$C2,$1E,$3F,$53,$F4,$3F,$15,$23,$F8
	.dc.b	$D1,$FC,$92,$D1,$F8,$DF,$8A,$F9,$7F,$20,$7E,$37,$3B,$E4,$3E,$51
	.dc.b	$8F,$8C,$B8,$FF,$21,$9F,$F0,$93,$FE,$FB,$57,$CE,$35,$5F,$A8,$E3
	.dc.b	$7C,$0F,$E9,$DB,$3F,$94,$3F,$C1,$4E,$1F,$C9,$9F,$97,$7F,$A8,$3D
	.dc.b	$FA,$BC,$2F,$F4,$7C,$3F,$C3,$49,$FC,$37,$C3,$FB,$38,$BF,$08,$D9
	.dc.b	$E5,$9F,$C0,$8F,$87,$C9,$F0,$B8,$ED,$7B,$DA,$F7,$7E,$4F,$1F,$C8
	.dc.b	$E8,$F7,$7C,$4F,$B7,$E2,$25,$FE,$7B,$6F,$F0,$6B,$F1,$F8,$5C,$DF
	.dc.b	$43,$B3,$E9,$F2,$DA,$D3,$5F,$67,$9D,$FE,$34,$E2,$F9,$F7,$F6,$3E
	.dc.b	$33,$ED,$F8,$3F,$6B,$E2,$AE,$1F,$5F,$6A,$FC,$77,$F0,$3E,$06,$E3
	.dc.b	$CD,$CD,$F5,$48,$FA,$78,$FF,$0F,$E0,$7E,$1B,$8B,$F1,$3D,$DF,$88
	.dc.b	$F8,$AB,$7F,$85,$76,$FE,$1E,$7F,$1F,$E4,$7E,$3B,$E2,$BF,$23,$E4
	.dc.b	$7E,$43,$E2,$AD,$FE,$35,$DB,$F8,$F9,$FE,$63,$F4,$1F,$22,$3F,$10
	.dc.b	$7C,$37,$F3,$AE,$0F,$C1,$38,$7F,$16,$7F,$95,$DE,$4F,$CC,$3F,$11
	.dc.b	$D0,$F2,$FF,$47,$F1,$7F,$A6,$6C,$FD,$96,$BC,$79,$FF,$43,$E5,$7E
	.dc.b	$AB,$C3,$FD,$6F,$AB,$FA,$CF,$88,$FD,$7F,$A9,$FA,$EF,$8A,$B8,$FE
	.dc.b	$AD,$DB,$FA,$D9,$FD,$AF,$AB,$FC,$77,$26,$CE,$CF,$DC,$FA,$99,$8F
	.dc.b	$E3,$58,$7E,$36,$31,$D7,$CA,$3F,$C5,$73,$C5,$FC,$97,$AB,$C7,$A7
	.dc.b	$F8,$1B,$63,$27,$08,$FE,$0F,$EA,$77,$EB,$FE,$67,$F2,$DF,$D1,$AB
	.dc.b	$F6,$5F,$15,$8F,$73,$F8,$5B,$63,$F8,$43,$C3,$D8,$FE,$1F,$F1,$5C
	.dc.b	$BD,$DF,$56,$96,$FC,$ED,$7F,$AC,$6B,$FD,$85,$FE,$21,$93,$DD,$F8
	.dc.b	$5A,$FF,$1C,$F2,$7E,$27,$83,$6F,$E0,$B5,$E6,$FF,$22,$F7,$7E,$57
	.dc.b	$85,$EC,$F7,$3F,$91,$FC,$57,$BB,$DD,$B8,$78,$1A,$FE,$07,$45,$7F
	.dc.b	$4F,$FC,$67,$FB,$76,$7F,$B7,$97,$C1,$E7,$DC,$F9,$3F,$07,$BF,$8B
	.dc.b	$B9,$7F,$93,$71,$47,$CE,$BE,$13,$F0,$3E,$13,$B7,$C5,$FD,$2F,$85
	.dc.b	$F0,$7C,$5D,$AE,$4F,$D6,$AD,$C7,$A4,$ED,$F1,$7E,$D7,$C1,$F0,$95
	.dc.b	$FF,$39,$67,$FD,$79,$F9,$FC,$1E,$DA,$F2,$47,$9F,$6F,$4B,$C3,$5A
	.dc.b	$FF,$9F,$AF,$D6,$FD,$8D,$7D,$6F,$C9,$CB,$CA,$FC,$BD,$2F,$89,$FE
	.dc.b	$3B,$8F,$E7,$9C,$1F,$E4,$77,$CB,$BF,$B5,$F9,$1E,$DF,$E5,$37,$FA
	.dc.b	$5F,$17,$F9,$68,$B3,$F8,$86,$BF,$66,$7E,$C7,$C5,$FE,$72,$BF,$61
	.dc.b	$F4,$1F,$E6,$5A,$FC,$7C,$E3,$9B,$F4,$2D,$8F,$C9,$D7,$0E,$3F,$D2
	.dc.b	$FC,$67,$2E,$0F,$D3,$FC,$65,$BF,$CC,$B5,$FC,$54,$F8,$39,$FF,$5B
	.dc.b	$CD,$F8,$B8,$FF,$10,$EA,$E9,$C5,$FF,$71,$8C,$3F,$89,$9E,$2F,$E2
	.dc.b	$F0,$FD,$BF,$3F,$E5,$76,$FD,$CF,$3F,$E5,$F5,$B7,$F9,$17,$6F,$E4
	.dc.b	$E7,$F7,$FC,$5F,$9D,$DA,$CF,$E5,$DA,$FE,$21,$FE,$0B,$D1,$FA,$3D
	.dc.b	$FD,$3B,$7E,$97,$9F,$F8,$4D,$D3,$F7,$65,$3F,$96,$6F,$EC,$7F,$27
	.dc.b	$CF,$E4,$7F,$2B,$9F,$A1,$FC,$BE,$95,$FC,$83,$B7,$F2,$51,$E6,$7F
	.dc.b	$29,$9D,$7F,$D9,$3D,$1F,$B6,$7F,$86,$EB,$EF,$7F,$87,$EB,$EB,$7E
	.dc.b	$87,$89,$FB,$FF,$13,$EA,$BF,$C0,$7C,$5B,$3F,$BB,$7E,$CF,$8B,$B8
	.dc.b	$FB,$1E,$AF,$F0,$7F,$1B,$EE,$FA,$D6,$7E,$E3,$5F,$87,$3F,$C3,$7C
	.dc.b	$5F,$C1,$FA,$BF,$E5,$DF,$87,$FD,$35,$7C,$88,$C7,$87,$A8,$FF,$6C
	.dc.b	$4E,$BC,$3F,$D6,$1F,$E5,$AF,$53,$97,$8D,$F8,$EF,$47,$A7,$A9,$F9
	.dc.b	$0F,$4E,$E3,$F8,$F6,$BF,$5A,$7C,$5E,$A7,$E5,$B6,$F4,$65,$DB,$B1
	.dc.b	$C1,$8D,$BF,$C3,$51,$8C,$7F,$90,$70,$F1,$7F,$91,$70,$FA,$51,$0F
	.dc.b	$F7,$E7,$07,$AE,$F9,$F0,$FC,$09,$FD,$23,$5F,$C5,$F0,$FB,$9F,$D4
	.dc.b	$7B,$1F,$27,$D5,$FD,$6F,$B3,$67,$F5,$8D,$7F,$27,$3F,$3F,$D5,$F3
	.dc.b	$47,$F2,$6D,$3E,$CB,$EA,$3E,$C9,$4F,$B2,$7B,$0F,$29,$F5,$32,$F7
	.dc.b	$3E,$DF,$8B,$FF,$39,$46,$23,$E0,$C6,$1F,$E0,$19,$7A,$DF,$D3,$FB
	.dc.b	$DF,$83,$F5,$3F,$83,$BF,$85,$F4,$EC,$FE,$09,$FE,$17,$EF,$6E,$3F
	.dc.b	$85,$7B,$BF,$8B,$F5,$7F,$87,$FC,$1F,$E3,$7D,$6B,$5F,$8A,$7F,$89
	.dc.b	$D6,$E3,$F8,$97,$C1,$FE,$4F,$DE,$FE,$2F,$F0,$3F,$95,$F7,$ED,$7E
	.dc.b	$49,$FE,$37,$5B,$8F,$E3,$5F,$0D,$F9,$BF,$80,$FE,$3F,$F0,$5F,$9D
	.dc.b	$F8,$1B,$5F,$9A,$7F,$91,$D6,$E3,$D2,$7E,$7B,$6F,$3B,$F8,$73,$F0
	.dc.b	$FA,$E1,$F9,$DD,$88,$F5,$C7,$13,$D6,$7E,$A3,$5F,$4B,$CF,$97,$FD
	.dc.b	$4D,$AE,$23,$9B,$EC,$78,$1F,$CD,$6F,$F6,$DF,$01,$FC,$E9,$B1,$93
	.dc.b	$F6,$B1,$FC,$FB,$E1,$FF,$79,$F0,$5F,$EE,$0F,$D3,$3F,$D8,$95,$FB
	.dc.b	$AF,$88,$C7,$C4,$7F,$01,$6C,$64,$FF,$5A,$4F,$F9,$79,$F7,$CF,$3F
	.dc.b	$88,$FE,$11,$F1,$1D,$7E,$2B,$F8,$54,$5A,$FE,$0E,$ED,$FE,$13,$3E
	.dc.b	$3C,$7F,$88,$55,$9F,$9C,$36,$7E,$4B,$FE,$CB,$8F,$1D,$7F,$BC,$E3
	.dc.b	$11,$FE,$69,$C9,$9F,$48,$F4,$E7,$ED,$69,$ED,$38,$39,$79,$65,$FC
	.dc.b	$8E,$3D,$11,$F5,$F6,$FF,$3F,$F9,$5F,$C8,$30,$FF,$6B,$4E,$75,$F8
	.dc.b	$0D,$3E,$4B,$F1,$9F,$97,$CB,$E9,$DF,$EA,$70,$7A,$F9,$6D,$7C,$2E
	.dc.b	$0F,$83,$1F,$66,$F8,$4F,$D4,$ED,$F8,$5C,$FF,$75,$B1,$93,$97,$0A
	.dc.b	$7F,$05,$CF,$F2,$7A,$FF,$09,$D1,$F2,$FB,$3F,$0D,$D3,$F3,$7A,$EE
	.dc.b	$5C,$9D,$BD,$27,$F1,$9D,$5F,$4F,$B2,$CF,$81,$AF,$E7,$F4,$47,$E4
	.dc.b	$FB,$3E,$D7,$77,$A6,$BC,$FB,$57,$83,$5F,$E3,$97,$EB,$7F,$2B,$B7
	.dc.b	$AD,$FC,$BE,$BE,$B7,$F3,$3A,$FA,$DA,$7A,$CD,$9E,$B3,$F3,$FE,$1F
	.dc.b	$E3,$7B,$FF,$51,$E0,$7E,$3F,$C4,$B5,$F3,$1F,$D6,$56,$E3,$E8,$1E
	.dc.b	$82,$9E,$8C,$FD,$4F,$F3,$2C,$FD,$4F,$E6,$73,$F5,$3F,$9B,$CF,$D4
	.dc.b	$FE,$77,$3F,$53,$4F,$51,$B3,$D4,$7E,$8F,$CC,$FD,$37,$8D,$FC,$07
	.dc.b	$C7,$B3,$FB,$C6,$BB,$3F,$0F,$E6,$7F,$65,$EF,$FE,$FB,$FA,$6F,$FA
	.dc.b	$96,$7F,$17,$E6,$FF,$0B,$F2,$7F,$65,$E5,$D9,$FC,$7B,$C6,$3F,$B7
	.dc.b	$DA,$F9,$57,$E0,$EF,$F8,$A7,$A3,$FB,$BF,$2F,$F8,$B7,$A3,$6B,$F8
	.dc.b	$8B,$5F,$DF,$4F,$F0,$1F,$43,$B4,$31,$ED,$E9,$FF,$05,$F4,$B3,$18
	.dc.b	$9F,$D1,$C6,$33,$56,$E7,$CF,$A7,$A2,$5E,$87,$F0,$FF,$53,$87,$A5
	.dc.b	$FC,$47,$7F,$81,$87,$E7,$78,$2B,$F9,$67,$6F,$C8,$3F,$95,$ED,$EB
	.dc.b	$D7,$F8,$56,$F7,$4F,$8F,$E6,$23,$17,$ED,$85,$FC,$18,$F3,$9F,$07
	.dc.b	$5F,$F4,$E3,$FA,$6F,$53,$DD,$EE,$7F,$27,$6C,$7F,$04,$7E,$17,$AD
	.dc.b	$FC,$AF,$DE,$F8,$9E,$EF,$F3,$2A,$FE,$65,$FC,$03,$FE,$D2,$7B,$46
	.dc.b	$7C,$F4,$BE,$0B,$8D,$7F,$19,$8C,$3E,$6E,$0F,$C1,$7B,$DC,$67,$A4
	.dc.b	$7E,$17,$E0,$71,$F0,$5F,$55,$D7,$BF,$70,$E8,$FD,$9F,$82,$DC,$75
	.dc.b	$E1,$7C,$1F,$5F,$81,$FB,$CF,$7F,$85,$B5,$E5,$7F,$05,$F0,$9B,$8E
	.dc.b	$5C,$3E,$57,$D6,$7F,$77,$DB,$D6,$FE,$3F,$A7,$F0,$D3,$EB,$E1,$FC
	.dc.b	$67,$0F,$1A,$F2,$6F,$E6,$78,$1F,$5D,$F6,$BA,$3D,$A7,$AF,$6E,$1C
	.dc.b	$3F,$B2,$F8,$7F,$91,$F0,$9F,$99,$F8,$6F,$95,$F1,$37,$0F,$A0,$FE
	.dc.b	$76,$B7,$1F,$17,$93,$E8,$7C,$47,$E8,$79,$2C,$FC,$96,$BE,$6C,$BF
	.dc.b	$71,$5F,$5E,$FF,$51,$C7,$F6,$72,$B7,$F7,$DA,$FC,$77,$F5,$D9,$FD
	.dc.b	$FE,$4F,$D8,$69,$F8,$1E,$4F,$D9,$6B,$F8,$3E,$2B,$8F,$E2,$5D,$BF
	.dc.b	$8B,$9F,$DC,$69,$E6,$D7,$F8,$AE,$D7,$8F,$F7,$0D,$BF,$D2,$57,$A5
	.dc.b	$FC,$36,$DE,$96,$9E,$92,$9E,$93,$F1,$9A,$FA,$5F,$C7,$71,$FE,$67
	.dc.b	$8F,$F8,$53,$F9,$BE,$3B,$3F,$A1,$7F,$87,$73,$6E,$3E,$91,$FA,$5D
	.dc.b	$FE,$93,$4F,$A4,$53,$E9,$73,$33,$F5,$46,$33,$F5,$34,$F5,$14,$F5
	.dc.b	$1C,$B3,$F5,$3F,$77,$F5,$0F,$F7,$23,$FE,$87,$E7,$73,$FF,$28,$9C
	.dc.b	$CF,$FA,$22,$5D,$99,$EA,$FD,$D3,$FC,$AE,$7F,$79,$7C,$7C,$7F,$DF
	.dc.b	$76,$59,$FE,$08,$D7,$F5,$8F,$7F,$1F,$F8,$27,$8F,$E1,$F1,$BF,$83
	.dc.b	$78,$F6,$FF,$86,$78,$F5,$FD,$A3,$E7,$F0,$B7,$1F,$C3,$F7,$78,$DF
	.dc.b	$C5,$FF,$26,$FF,$5F,$4F,$F4,$79,$CB,$FE,$2F,$8C,$F6,$A7,$33,$FE
	.dc.b	$77,$9C,$C3,$C9,$E9,$E3,$7A,$EB,$E5,$6F,$5F,$90,$D7,$F9,$5B,$EF
	.dc.b	$7F,$93,$6D,$EB,$7E,$96,$B5,$FA,$1A,$D7,$F9,$9B,$EF,$7F,$97,$47
	.dc.b	$D7,$F2,$BD,$DE,$4F,$D9,$F2,$BE,$07,$9B,$6F,$DC,$ED,$F8,$13,$F7
	.dc.b	$FD,$7F,$8D,$E6,$FE,$07,$D7,$F9,$1E,$AD,$BF,$8C,$ED,$F9,$13,$F8
	.dc.b	$7F,$5B,$E7,$7A,$FF,$89,$F4,$BE,$87,$BD,$6F,$E7,$3B,$7E,$84,$FE
	.dc.b	$3F,$DD,$FA,$DE,$BF,$E4,$7D,$BF,$B1,$EC,$5B,$FA,$CE,$DF,$B1,$3E
	.dc.b	$37,$EB,$67,$E7,$7E,$C6,$95,$FF,$72,$75,$23,$84,$62,$7C,$8F,$DF
	.dc.b	$CB,$D0,$FE,$07,$3A,$F2,$76,$F4,$8F,$D4,$7B,$DF,$8F,$F6,$7F,$55
	.dc.b	$EF,$5A,$FC,$A3,$5F,$D9,$4F,$E5,$7D,$EF,$F0,$BE,$4F,$83,$FC,$C7
	.dc.b	$BF,$8F,$89,$FC,$D7,$C0,$DC,$7F,$30,$F0,$AD,$C7,$E8,$7E,$17,$F4
	.dc.b	$1F,$09,$FB,$B7,$F4,$5F,$13,$67,$F4,$0F,$EF,$FE,$1B,$71,$FA,$9F
	.dc.b	$86,$FD,$47,$C4,$FF,$03,$7F,$55,$F1,$16,$7F,$50,$FF,$07,$F8,$8D
	.dc.b	$C7,$F7,$FD,$FD,$4F,$EB,$77,$F4,$3F,$AB,$DA,$BF,$51,$F5,$3B,$3D
	.dc.b	$47,$ED,$75,$F4,$3F,$B3,$D2,$BF,$71,$F5,$3B,$3C,$58,$FF,$01,$F8
	.dc.b	$9F,$9A,$58,$F3,$35,$FE,$C1,$8F,$FB,$99,$FE,$01,$BF,$91,$FE,$05
	.dc.b	$97,$A1,$1F,$ED,$7E,$86,$79,$8C,$3C,$5F,$C8,$CF,$A1,$E3,$F0,$EA
	.dc.b	$69,$99,$E9,$A5,$AD,$3E,$E8,$FE,$29,$3F,$88,$8F,$E3,$DC,$56,$7F
	.dc.b	$8E,$35,$FE,$34,$FE,$3B,$83,$FC,$75,$F0,$3F,$EE,$AF,$38,$FF,$9E
	.dc.b	$36,$72,$EB,$93,$F2,$AD,$C7,$EF,$35,$F9,$3B,$74,$E4,$FC,$DF,$07
	.dc.b	$6E,$4B,$5F,$ED,$DC,$9C,$D8,$8F,$96,$7C,$48,$CB,$F8,$BE,$0E,$4F
	.dc.b	$E3,$7B,$7D,$3D,$1F,$A8,$E9,$B3,$F9,$16,$BF,$72,$7D,$FD,$FE,$A8
	.dc.b	$F1,$6F,$5F,$08,$E5,$F8,$71,$98,$F2,$C6,$1F,$CF,$6F,$EF,$7F,$41
	.dc.b	$B7,$99,$FC,$5B,$F3,$B5,$F7,$47,$2D,$7D,$AF,$E4,$35,$E1,$C5,$F4
	.dc.b	$A3,$FD,$39,$C4,$EC,$FA,$FD,$FF,$C1,$BB,$FE,$CF,$87,$6F,$E6,$3F
	.dc.b	$C2,$AB,$71,$FB,$3F,$17,$EF,$F8,$9F,$C3,$9F,$C0,$F8,$96,$7E,$A3
	.dc.b	$FC,$4B,$C3,$DC,$7E,$EF,$C3,$FC,$3F,$89,$FC,$5B,$5B,$3F,$BF,$6B
	.dc.b	$FA,$A9,$FC,$6F,$8F,$FC,$73,$CC,$FC,$7F,$91,$6B,$F0,$5E,$7D,$7E
	.dc.b	$8F,$47,$E4,$A3,$F9,$1F,$93,$C2,$7D,$47,$E2,$77,$F4,$3C,$39,$9B
	.dc.b	$E2,$3E,$73,$EB,$78,$1E,$73,$F2,$5B,$7A,$DA,$64,$E8,$FF,$34,$D7
	.dc.b	$D7,$1F,$E9,$BF,$D4,$FF,$DB,$6F,$DB,$F3,$FF,$9C,$E1,$FA,$CF,$52
	.dc.b	$CF,$E7,$DF,$F6,$9D,$70,$8F,$CF,$56,$3D,$FF,$D9,$FA,$B6,$7F,$4C
	.dc.b	$D7,$DC,$F2,$F5,$BF,$73,$EB,$74,$F5,$FF,$77,$EB,$F6,$F5,$BF,$7B
	.dc.b	$EC,$5C,$7F,$60,$ED,$FD,$94,$F9,$3D,$7F,$F5,$E7,$F0,$0F,$FA,$D3
	.dc.b	$6F,$F7,$0D,$7E,$90,$FD,$24,$67,$D7,$DC,$71,$35,$D0,$9F,$F5,$3F
	.dc.b	$87,$F9,$B1,$6E,$0F,$7D,$FC,$B6,$BF,$E9,$18,$D3,$F0,$8F,$A1,$BF
	.dc.b	$C0,$7E,$96,$DF,$01,$FA,$9B,$7C,$06,$9F,$00,$D9,$F0,$0F,$B5,$F0
	.dc.b	$BF,$03,$E0,$7E,$AF,$C1,$D9,$FB,$CD,$7C,$93,$F1,$BE,$03,$EA,$47
	.dc.b	$EE,$F9,$2B,$FC,$65,$AF,$F0,$BE,$74,$7D,$98,$FE,$45,$CB,$F4,$DF
	.dc.b	$CD,$72,$7D,$77,$F3,$9C,$B5,$FE,$10,$ED,$FE,$15,$1F,$8C,$F8,$7F
	.dc.b	$B3,$F0,$B6,$7E,$9F,$22,$31,$1F,$AE,$8C,$3F,$9B,$F8,$6F,$C0,$FC
	.dc.b	$1F,$E5,$BE,$1F,$F0,$7F,$09,$6B,$9C,$57,$93,$F9,$CF,$86,$DC,$7E
	.dc.b	$97,$E2,$7F,$17,$1F,$A0,$F8,$9B,$3F,$A9,$6B,$F7,$E7,$F2,$27,$E9
	.dc.b	$B0,$FC,$9C,$FE,$A3,$0B,$5F,$B1,$6B,$F8,$79,$FC,$CE,$9E,$58,$FE
	.dc.b	$6F,$9A,$B1,$1F,$DC,$21,$96,$23,$F3,$AF,$AD,$F9,$33,$FB,$8C,$7C
	.dc.b	$FD,$7F,$9B,$EA,$67,$8E,$3F,$F0,$1C,$FE,$7A,$31,$82,$3F,$43,$7E
	.dc.b	$E8,$FE,$15,$C9,$FB,$0D,$FF,$82,$F0,$59,$FE,$1C,$D7,$F4,$93,$FB
	.dc.b	$6E,$5F,$EC,$F8,$7E,$B7,$F5,$5F,$F5,$BD,$7E,$DB,$9F,$F8,$7B,$FB
	.dc.b	$C6,$BF,$84,$78,$E9,$E8,$9E,$3D,$BF,$B9,$ED,$FE,$2F,$F1,$5F,$C1
	.dc.b	$3B,$AE,$3F,$8C,$EE,$AF,$E2,$DF,$E3,$BD,$BB,$8F,$87,$97,$5E,$CF
	.dc.b	$E1,$BD,$DE,$09,$F8,$9A,$27,$D9,$1C,$0C,$B8,$7E,$55,$79,$FB,$39
	.dc.b	$C6,$F7,$5F,$AB,$A5,$C1,$F0,$2B,$F4,$0D,$7F,$0A,$F8,$EB,$8F,$6F
	.dc.b	$F2,$2E,$0F,$6D,$7E,$2B,$7A,$FE,$27,$7A,$FF,$24,$7E,$0B,$FB,$08
	.dc.b	$E9,$93,$C2,$FE,$5B,$E0,$E1,$B5,$E9,$9D,$3F,$B8,$9E,$3E,$07,$F3
	.dc.b	$4F,$07,$9F,$8B,$FC,$DB,$E1,$FA,$F8,$DF,$CD,$70,$FF,$0C,$7E,$4D
	.dc.b	$FE,$DA,$C9,$39,$E9,$4E,$70,$E5,$7E,$1F,$37,$52,$F3,$7D,$31,$86
	.dc.b	$CF,$F3,$E9,$D3,$E9,$F3,$78,$F9,$9F,$03,$CD,$F4,$4F,$B7,$CB,$AF
	.dc.b	$DA,$3F,$02,$E3,$9F,$F3,$53,$E2,$F4,$3D,$F5,$FC,$6B,$83,$DE,$F8
	.dc.b	$7B,$2B,$FE,$08,$CF,$D8,$EC,$F6,$1E,$7D,$3D,$2E,$DA,$FF,$8F,$23
	.dc.b	$D6,$F9,$71,$FC,$5E,$9E,$C8,$FA,$DE,$97,$95,$FB,$1E,$7F,$A7,$1F
	.dc.b	$D5,$36,$3F,$3B,$3B,$7A,$71,$FF,$85,$9F,$31,$8E,$3F,$A3,$7E,$C8
	.dc.b	$F2,$3F,$09,$FE,$71,$A7,$A1,$FF,$44,$E7,$E8,$CF,$F8,$17,$0F,$D7
	.dc.b	$9F,$82,$7B,$CF,$AC,$7D,$9E,$1F,$CA,$FB,$5F,$80,$F4,$AC,$FD,$DF
	.dc.b	$4A,$BF,$35,$FC,$DF,$AB,$B8,$FC,$C7,$BF,$F8,$8F,$53,$F3,$EF,$E2
	.dc.b	$BD,$5B,$3F,$87,$7F,$47,$EF,$EE,$3F,$41,$EC,$FE,$43,$D8,$FD,$3B
	.dc.b	$63,$27,$E3,$E3,$F5,$7E,$FF,$E5,$BD,$EF,$D6,$FC,$47,$8E,$BE,$6F
	.dc.b	$BF,$F9,$AD,$BF,$62,$D8,$FD,$74,$EF,$8F,$F9,$46,$BE,$39,$FA,$AF
	.dc.b	$A5,$F6,$9E,$93,$F6,$DA,$6E,$7F,$98,$69,$E7,$9F,$BA,$D8,$CC,$7A
	.dc.b	$4F,$DF,$67,$E9,$7F,$8C,$7E,$AD,$FE,$E6,$7F,$35,$F0,$3F,$C4,$E3
	.dc.b	$11,$F9,$58,$C7,$C2,$7E,$D9,$D7,$C3,$FE,$E3,$E1,$38,$7C,$37,$EE
	.dc.b	$BE,$17,$97,$C3,$5C,$7B,$CD,$7F,$3D,$92,$7F,$80,$FC,$37,$87,$E2
	.dc.b	$7F,$81,$FC,$3D,$BF,$E0,$0F,$97,$E2,$73,$1E,$A9,$FE,$13,$8F,$A6
	.dc.b	$FC,$D3,$EE,$7F,$86,$E1,$ED,$C2,$CE,$4F,$E1,$9A,$FA,$9F,$E1,$D9
	.dc.b	$FA,$9A,$7A,$8A,$7A,$AB,$D2,$F8,$CF,$39,$4F,$E0,$D5,$E9,$72,$7C
	.dc.b	$2C,$7D,$2F,$D4,$D3,$F9,$1C,$7D,$37,$F9,$2E,$7F,$57,$0B,$5B,$3F
	.dc.b	$92,$57,$D9,$E6,$FE,$57,$B7,$DB,$EA,$B5,$F8,$09,$AF,$D2,$8F,$E6
	.dc.b	$3C,$1F,$81,$EA,$FE,$71,$FC,$03,$FE,$F5,$F0,$FF,$C1,$BC,$6E,$CC
	.dc.b	$F5,$EF,$2F,$F3,$8D,$A3,$F1,$53,$CB,$AF,$F1,$9D,$9E,$8B,$FC,$0B
	.dc.b	$DB,$B3,$F2,$1C,$D6,$7F,$2B,$3B,$EA,$FF,$3D,$66,$E0,$F4,$C6,$38
	.dc.b	$2E,$DE,$69,$B8,$7F,$88,$6D,$5E,$E8,$FD,$0E,$FE,$57,$F4,$7A,$FB
	.dc.b	$1C,$9E,$CD,$BD,$6F,$BB,$6F,$D3,$79,$5F,$1B,$B6,$CF,$C2,$7F,$55
	.dc.b	$DF,$98,$FD,$7C,$FC,$CE,$EF,$AD,$1F,$D6,$F9,$7E,$B3,$4F,$AC,$53
	.dc.b	$EB,$1F,$D8,$79,$7E,$83,$FE,$D8,$E5,$FA,$CF,$BF,$4F,$53,$FA,$ED
	.dc.b	$6F,$17,$CA,$9F,$81,$3F,$B2,$F2,$BE,$FF,$7F,$F0,$16,$C7,$F0,$6F
	.dc.b	$0A,$BB,$3E,$FF,$85,$FE,$19,$A7,$A3,$F8,$8F,$0F,$0F,$E2,$BC,$9D
	.dc.b	$7A,$17,$0F,$F1,$45,$F0,$7F,$13,$7F,$C4,$7A,$BF,$23,$E5,$D9,$FC
	.dc.b	$6B,$DF,$C9,$CC,$7F,$1A,$9F,$CB,$F9,$FE,$3F,$53,$F3,$3E,$67,$97
	.dc.b	$DB,$FC,$DF,$99,$9C,$3F,$2E,$E5,$FC,$CC,$FA,$1F,$19,$E6,$DF,$D0
	.dc.b	$FF,$75,$D7,$FD,$AA,$FD,$1D,$F8,$3F,$4F,$6F,$33,$E8,$3C,$C5,$39
	.dc.b	$6B,$EA,$7F,$91,$EB,$EA,$7E,$2F,$AD,$FB,$2F,$43,$E7,$7A,$D6,$72
	.dc.b	$7E,$C2,$32,$FE,$CB,$D2,$FA,$9F,$0F,$FB,$AF,$53,$EB,$7C,$15,$AF
	.dc.b	$A3,$35,$F6,$47,$F0,$0F,$63,$EE,$7C,$07,$F0,$2F,$66,$D7,$EF,$9F
	.dc.b	$C0,$7B,$79,$8F,$B7,$3F,$C2,$3D,$EF,$C2,$7C,$1F,$F0,$AF,$7E,$D7
	.dc.b	$F0,$67,$F1,$1F,$09,$98,$FC,$1C,$FF,$10,$F7,$FD,$31,$FC,$33,$6F
	.dc.b	$4B,$4F,$49,$4F,$49,$FC,$4B,$4F,$3B,$FA,$5D,$2A,$27,$BE,$F1,$FF
	.dc.b	$DC,$B1,$F9,$4D,$BD,$8F,$DF,$DB,$D8,$FE,$63,$E1,$FF,$CB,$9F,$07
	.dc.b	$CE,$30,$53,$AB,$7C,$87,$47,$FA,$DB,$B1,$F1,$3C,$AF,$F5,$58,$74
	.dc.b	$C3,$F5,$87,$6F,$84,$FD,$71,$67,$F5,$31,$5F,$CA,$47,$93,$E1,$7F
	.dc.b	$68,$79,$BE,$1A,$CF,$94,$6E,$BF,$F7,$9C,$7E,$76,$78,$3F,$B6,$AF
	.dc.b	$7F,$C4,$FE,$FC,$F8,$3F,$13,$FC,$04,$B3,$FB,$C8,$AF,$E9,$E3,$E3
	.dc.b	$E5,$EC,$8F,$16,$3E,$1C,$BD,$B1,$8C,$91,$B8,$F9,$93,$B7,$F8,$4C
	.dc.b	$7B,$DF,$31,$C7,$2F,$7B,$C2,$BD,$BB,$7D,$28,$FF,$72,$6B,$F0,$1F
	.dc.b	$B5,$87,$C3,$BE,$31,$8B,$67,$FE,$B7,$FD,$6B,$FD,$AF,$1F,$6A,$7F
	.dc.b	$D8,$51,$8E,$7F,$E4,$91,$F8,$67,$87,$3F,$E2,$38,$3D,$93,$D7,$5F
	.dc.b	$82,$F9,$DC,$BF,$8A,$72,$F8,$F8,$3C,$AF,$E3,$B3,$AF,$DC,$CB,$DF
	.dc.b	$C3,$F8,$F9,$FB,$73,$FC,$66,$BE,$DB,$E8,$EA,$FC,$DF,$4F,$B7,$AB
	.dc.b	$F3,$BD,$9E,$F9,$B5,$EE,$9D,$BF,$03,$B7,$F4,$7D,$BF,$17,$B7,$F4
	.dc.b	$BD,$3F,$1E,$6D,$7C,$69,$DB,$F2,$3B,$7F,$57,$D7,$F3,$7B,$7F,$5B
	.dc.b	$E0,$7C,$F9,$B5,$F3,$A7,$6F,$D0,$F1,$BF,$67,$E0,$FD,$56,$C7,$D4
	.dc.b	$F0,$6B,$FE,$AD,$EA,$4F,$EE,$BC,$8F,$34,$FE,$47,$7F,$53,$F9,$3D
	.dc.b	$F2,$FE,$2D,$CB,$F8,$D8,$F3,$3F,$99,$D7,$D4,$FE,$6F,$5C,$BF,$93
	.dc.b	$72,$FE,$56,$3F,$19,$E4,$FF,$0C,$F2,$6C,$FE,$8D,$FC,$87,$93,$98
	.dc.b	$FC,$7C,$FF,$12,$F2,$7F,$29,$E6,$7F,$14,$F3,$2D,$7E,$AD,$FC,$C7
	.dc.b	$95,$98,$FC,$BC,$FF,$1A,$F2,$BF,$39,$E5,$FF,$BD,$FF,$80,$7F,$4D
	.dc.b	$5F,$1F,$CD,$C7,$9F,$A8,$FD,$CC,$FF,$B0,$7C,$2F,$E5,$9E,$5B,$9D
	.dc.b	$87,$39,$FD,$47,$A9,$D7,$D0,$FD,$57,$9B,$DF,$D5,$CE,$7D,$1C,$BD
	.dc.b	$A7,$D6,$FE,$EF,$86,$BF,$B5,$7D,$AF,$BC,$F6,$9F,$C2,$F7,$F4,$3F
	.dc.b	$C2,$B5,$F6,$EB,$FC,$0C,$F4,$69,$EE,$9F,$DD,$EF,$5F,$84,$7C,$0D
	.dc.b	$7F,$81,$7B,$FF,$1F,$D7,$B3,$F1,$3D,$DF,$E0,$DE,$CF,$CB,$73,$1E
	.dc.b	$19,$FE,$91,$97,$87,$F5,$D7,$C0,$FD,$1F,$7B,$39,$FC,$C9,$7C,$1F
	.dc.b	$D5,$F6,$F3,$93,$15,$72,$8F,$AB,$3A,$9C,$D7,$59,$FA,$F7,$D7,$DD
	.dc.b	$FC,$07,$B9,$BC,$5E,$EE,$2F,$67,$78,$FF,$62,$BF,$DA,$B8,$FF,$D0
	.dc.b	$8E,$B2,$F1,$D7,$E2,$FE,$07,$F9,$47,$BF,$F8,$DF,$82,$FE,$55,$F0
	.dc.b	$37,$0F,$E5,$33,$5F,$DB,$47,$E4,$FE,$13,$C9,$C5,$E9,$9F,$81,$BF
	.dc.b	$C5,$E0,$F8,$91,$E1,$DB,$E1,$3F,$7F,$87,$D9,$3C,$76,$FA,$1C,$3E
	.dc.b	$DA,$E6,$7B,$77,$C4,$EA,$3F,$3F,$C3,$F1,$67,$D9,$1D,$FE,$23,$F5
	.dc.b	$1F,$0B,$F2,$7E,$1F,$F5,$5F,$11,$F2,$E6,$D7,$96,$76,$F9,$F7,$F2
	.dc.b	$6F,$EB,$8F,$B2,$64,$A7,$9B,$3F,$B6,$FE,$CB,$9B,$EC,$6B,$E8,$AF
	.dc.b	$C4,$69,$6B,$EC,$4D,$BF,$AB,$1F,$BA,$E5,$FC,$06,$9F,$C0,$7E,$23
	.dc.b	$F0,$47,$F0,$3D,$73,$5F,$C0,$1C,$BF,$C0,$A3,$F1,$15,$FC,$27,$5F
	.dc.b	$C5,$73,$7F,$0B,$DB,$37,$FC,$21,$CB,$FC,$2A,$7F,$21,$C9,$FC,$47
	.dc.b	$8B,$F2,$5C,$36,$BE,$E4,$D7,$E1,$72,$27,$F8,$C6,$FF,$99,$E4,$FE
	.dc.b	$35,$C3,$F9,$BE,$4F,$F1,$06,$1F,$E2,$0F,$9D,$FF,$53,$F1,$FE,$A2
	.dc.b	$F3,$F0,$A9,$CE,$7F,$9C,$BF,$E1,$18,$FA,$9F,$E1,$5B,$7A,$9F,$E1
	.dc.b	$9B,$E5,$FF,$05,$F4,$FE,$4B,$3F,$DF,$70,$FC,$59,$F5,$9F,$CA,$B8
	.dc.b	$7D,$ED,$32,$7F,$17,$72,$FF,$19,$DF,$F6,$DD,$7E,$8D,$7F,$DE,$13
	.dc.b	$FA,$18,$C7,$4F,$EE,$FA,$6C,$E4,$FF,$18,$D3,$B7,$F8,$07,$8F,$CF
	.dc.b	$BB,$F8,$17,$87,$6F,$F5,$93,$5F,$A3,$1E,$1F,$1B,$F8,$47,$8B,$E3
	.dc.b	$F1,$AD,$76,$F2,$3F,$86,$78,$7E,$77,$31,$FE,$3C,$D1,$5F,$C5,$FF
	.dc.b	$5E,$78,$FE,$DF,$2B,$0F,$BF,$C7,$FE,$2D,$E3,$5C,$7F,$88,$5F,$07
	.dc.b	$DB,$7E,$C7,$DF,$C1,$EC,$69,$EC,$29,$EC,$3E,$2E,$DE,$67,$FC,$8B
	.dc.b	$9F,$99,$FE,$1D,$B7,$AD,$FE,$21,$B5,$7F,$2B,$1F,$0D,$F2,$F9,$7F
	.dc.b	$5F,$C9,$F7,$79,$5F,$67,$CC,$F8,$13,$6B,$D7,$3B,$7D,$BE,$97,$DF
	.dc.b	$F3,$BE,$37,$A1,$F8,$1F,$53,$E4,$4D,$AF,$87,$3B,$7E,$2F,$A5,$F8
	.dc.b	$7F,$57,$E7,$7A,$DF,$89,$F6,$BE,$84,$DA,$F9,$73,$B7,$E6,$FA,$FF
	.dc.b	$8F,$F6,$BE,$B7,$AF,$67,$27,$E3,$23,$2F,$E3,$BD,$AF,$2B,$F9,$0D
	.dc.b	$BD,$2F,$E4,$B6,$CB,$F8,$37,$2F,$E1,$63,$CA,$FE,$63,$4F,$4B,$F9
	.dc.b	$AD,$32,$FE,$2D,$CB,$F8,$D8,$FC,$67,$BB,$FA,$7F,$6F,$F1,$DE,$DF
	.dc.b	$EA,$7D,$DC,$DF,$E9,$1C,$BF,$A6,$9F,$2B,$F8,$2C,$7C,$AF,$E1,$32
	.dc.b	$CB,$FE,$CC,$E9,$FF,$26,$BF,$48,$F8,$07,$D2,$3D,$47,$CE,$3D,$46
	.dc.b	$4F,$51,$93,$D4,$7E,$7B,$E1,$7F,$80,$7B,$DF,$A7,$F7,$EC,$E4,$FD
	.dc.b	$EC,$65,$FD,$FF,$C2,$FE,$B7,$E1,$BF,$8F,$FA,$FF,$ED,$5F,$13,$FC
	.dc.b	$73,$E2,$7F,$A7,$AF,$FD,$6B,$3F,$EC,$57,$F5,$BF,$11,$C3,$E1,$79
	.dc.b	$46,$8F,$DC,$7C,$47,$4F,$87,$B1,$CB,$E1,$BF,$79,$F1,$1E,$07,$31
	.dc.b	$FE,$BC,$C7,$FD,$7E,$FE,$EA,$3C,$BF,$11,$FC,$0C,$B1,$93,$CA,$E5
	.dc.b	$F3,$EB,$E4,$7A,$14,$FF,$5E,$D7,$AD,$FE,$D5,$A3,$4F,$64,$7F,$0A
	.dc.b	$AF,$8A,$F9,$8E,$39,$54,$7F,$BA,$AD,$8F,$BA,$3E,$01,$FC,$53,$3F
	.dc.b	$88,$FF,$16,$D3,$E8,$E3,$FC,$7D,$FA,$73,$FC,$87,$0B,$5B,$3E,$9D
	.dc.b	$7F,$8E,$F1,$73,$7D,$AD,$71,$C1,$6F,$FC,$65,$5A,$7E,$DC,$E9,$FB
	.dc.b	$BC,$5C,$F8,$7F,$07,$C5,$FC,$C6,$7F,$0B,$D1,$97,$F0,$6E,$5F,$C2
	.dc.b	$C7,$8F,$AF,$F1,$7D,$36,$7F,$13,$E1,$79,$FA,$BF,$1E,$E6,$3F,$01
	.dc.b	$3E,$BD,$BD,$11,$FE,$1D,$D5,$B6,$3D,$9F,$F3,$B4,$7A,$E7,$F1,$1B
	.dc.b	$FA,$DF,$C5,$6F,$E7,$7F,$19,$AF,$9D,$FC,$76,$D9,$7F,$06,$E5,$FC
	.dc.b	$2C,$7C,$DE,$DF,$D2,$F8,$7F,$3F,$BE,$D7,$CC,$9A,$F9,$E3,$F5,$7E
	.dc.b	$1F,$D5,$F3,$7F,$5B,$E2,$5A,$FD,$5F,$89,$F6,$7C,$8F,$D8,$B9,$8F
	.dc.b	$D1,$4F,$DD,$F2,$3F,$6B,$E3,$59,$FD,$9F,$91,$F8,$1F,$1B,$F7,$2E
	.dc.b	$63,$F5,$53,$F8,$5F,$1B,$F7,$BE,$45,$9F,$DD,$F9,$DF,$89,$F2,$7F
	.dc.b	$80,$B9,$8F,$D9,$4F,$8D,$FD,$3E,$75,$FC,$FF,$8A,$1A,$FA,$E3,$DA
	.dc.b	$7E,$43,$7F,$82,$FE,$4B,$6F,$5B,$4F,$59,$F9,$6D,$BD,$6E,$4F,$59
	.dc.b	$F9,$AF,$43,$FC,$6B,$EA,$FF,$AE,$3C,$AF,$F6,$7F,$5A,$73,$E9,$FE
	.dc.b	$97,$D1,$B8,$71,$3F,$6E,$4D,$58,$F3,$8F,$F5,$56,$0D,$3D,$31,$F9
	.dc.b	$37,$EB,$BF,$AC,$E5,$AE,$4F,$E4,$FB,$FD,$37,$F8,$7F,$25,$7E,$4E
	.dc.b	$2E,$5F,$B5,$1F,$08,$FD,$07,$07,$C6,$7F,$45,$E8,$7C,$0F,$4F,$F8
	.dc.b	$03,$F0,$BD,$0F,$E0,$5E,$9D,$AF,$DF,$CE,$DF,$E0,$3F,$0F,$F2,$3D
	.dc.b	$3F,$E1,$0D,$8F,$E0,$BC,$35,$FD,$84,$7C,$EF,$5F,$F8,$77,$AD,$F4
	.dc.b	$3D,$7F,$E2,$1E,$BD,$BF,$E1,$B3,$5F,$DC,$47,$D6,$F8,$0F,$F6,$87
	.dc.b	$EC,$3F,$EC,$69,$FC,$46,$5F,$E6,$BF,$2B,$FD,$E1,$3F,$EC,$E9,$FA
	.dc.b	$A7,$E3,$A5,$F0,$BC,$BD,$BF,$C0,$BD,$3E,$07,$F0,$7F,$07,$EC,$BE
	.dc.b	$27,$6E,$0C,$3D,$39,$FE,$FC,$7F,$29,$C7,$E0,$67,$FC,$33,$87,$E0
	.dc.b	$5F,$B0,$E7,$7F,$86,$E0,$F8,$31,$8B,$47,$C6,$8C,$3A,$7C,$10,$DB
	.dc.b	$E1,$FC,$0F,$E6,$BE,$03,$E2,$FC,$07,$E7,$3E,$17,$38,$75,$72,$F7
	.dc.b	$9F,$97,$F1,$1F,$A2,$F8,$6F,$9B,$F0,$F6,$BF,$DE,$1C,$AA,$C4,$7F
	.dc.b	$B1,$E1,$3F,$AA,$9F,$AB,$F1,$1F,$AC,$8B,$5E,$D7,$EC,$FC,$3E,$63
	.dc.b	$AC,$FE,$C8,$FB,$B9,$7E,$D2,$7E,$FE,$36,$BC,$73,$5F,$14,$7E,$EB
	.dc.b	$4F,$5C,$7C,$B8,$F5,$B4,$F5,$E3,$EA,$E2,$FC,$3E,$5E,$99,$F1,$9E
	.dc.b	$72,$9F,$41,$F4,$B9,$3E,$16,$5E,$97,$E3,$F3,$7F,$0C,$DB,$F2,$FB
	.dc.b	$7F,$0E,$E7,$FC,$CC,$DA,$F9,$B3,$B7,$E7,$F6,$FF,$14,$E8,$FC,$FB
	.dc.b	$FE,$CA,$FC,$A3,$FD,$9B,$C4,$E8,$C7,$46,$B0,$FC,$56,$1F,$A6,$EC
	.dc.b	$E1,$CF,$67,$F2,$38,$FF,$AD,$5F,$BB,$D1,$D7,$B3,$51,$F9,$D9,$FF
	.dc.b	$60,$BF,$7F,$A7,$C5,$D9,$FB,$27,$C9,$DB,$FB,$4E,$DB,$5F,$88,$9D
	.dc.b	$BE,$C3,$CA,$53,$B6,$BF,$BC,$DB,$D7,$19,$3F,$96,$69,$FB,$B7,$F2
	.dc.b	$1A,$FB,$1F,$CA,$BE,$97,$D6,$7B,$B4,$F5,$BF,$CD,$74,$F5,$F3,$FF
	.dc.b	$06,$F2,$3E,$67,$7F,$F0,$C7,$E7,$79,$3F,$C3,$BC,$1B,$87,$E8,$A7
	.dc.b	$6F,$94,$FE,$19,$C7,$F4,$5F,$59,$FA,$8D,$7D,$6F,$E9,$F9,$7E,$E3
	.dc.b	$E8,$3E,$9E,$DE,$67,$FB,$1E,$9F,$55,$FD,$77,$27,$B5,$F8,$27,$03
	.dc.b	$FC,$8F,$AB,$CA,$BF,$B1,$FF,$B5,$E7,$ED,$79,$7F,$CA,$27,$F1,$4F
	.dc.b	$F5,$8A,$E1,$25,$5C,$63,$E4,$4F,$17,$F1,$3E,$57,$1B,$EF,$E6,$FE
	.dc.b	$57,$CB,$C5,$73,$C3,$FB,$DB,$F9,$0F,$23,$39,$76,$F2,$BF,$39,$18
	.dc.b	$AD,$D7,$E4,$6F,$1E,$57,$E8,$7C,$DB,$3F,$9A,$F4,$78,$79,$FF,$A5
	.dc.b	$73,$1F,$99,$9E,$9E,$97,$EA,$7D,$2B,$3F,$A2,$F5,$3C,$1E,$8F,$EB
	.dc.b	$5C,$C7,$E8,$67,$C9,$C1,$ED,$8F,$9F,$C3,$F0,$1F,$61,$83,$E0,$70
	.dc.b	$7B,$1C,$70,$FC,$CA,$E5,$1F,$3E,$BE,$BE,$DE,$D8,$DF,$07,$F7,$B7
	.dc.b	$DC,$63,$7F,$73,$A9,$E0,$F3,$3E,$87,$B1,$FC,$1F,$D5,$F9,$8F,$F0
	.dc.b	$9F,$7F,$E7,$7B,$16,$BB,$CE,$DF,$0F,$B3,$FC,$3F,$DF,$FA,$8D,$8F
	.dc.b	$2F,$BD,$5F,$06,$FF,$88,$9F,$E2,$DE,$F7,$DB,$F7,$FF,$8C,$7B,$F6
	.dc.b	$FF,$88,$BF,$7F,$DE,$CC,$7C,$09,$FE,$3D,$F0,$1E,$88,$FE,$CF,$FC
	.dc.b	$03,$FE,$CB,$7D,$A6,$3E,$17,$D6,$D3,$D6,$70,$3D,$27,$F2,$0D,$6B
	.dc.b	$FC,$75,$FC,$86,$BE,$BD,$7F,$46,$7A,$E7,$F9,$3E,$9E,$97,$F9,$26
	.dc.b	$95,$FF,$20,$63,$FE,$CE,$7E,$1F,$23,$5F,$F0,$4C,$7F,$97,$67,$58
	.dc.b	$76,$A6,$BF,$0A,$B8,$FC,$17,$0F,$80,$61,$FE,$C2,$E7,$FC,$2F,$C2
	.dc.b	$E6,$31,$7C,$0F,$8F,$F1,$1E,$AA,$FE,$07,$5B,$F8,$7C,$57,$ED,$5E
	.dc.b	$5F,$09,$8D,$7F,$80,$E1,$D7,$E1,$B1,$B7,$FB,$4B,$0F,$A7,$78,$AD
	.dc.b	$65,$F9,$88,$F2,$65,$C1,$B1,$F9,$CC,$3D,$1A,$E7,$0F,$BB,$7D,$8F
	.dc.b	$67,$3F,$83,$97,$DD,$D1,$9A,$FC,$1C,$E5,$FC,$2C,$79,$B6,$F5,$E5
	.dc.b	$FA,$AD,$3C,$B5,$F5,$27,$2F,$D6,$8C,$ED,$F4,$F5,$F3,$CF,$DA,$6C
	.dc.b	$7D,$7D,$7D,$38,$7B,$EB,$D2,$FE,$E3,$3F,$A9,$D1,$F2,$3E,$1E,$CE
	.dc.b	$6B,$F3,$7C,$F9,$7F,$3B,$5E,$97,$E6,$CF,$A5,$FD,$0E,$19,$7F,$CA
	.dc.b	$1C,$4C,$FE,$94,$6A,$BC,$8E,$B8,$7E,$73,$C3,$87,$ED,$E1,$EA,$8F
	.dc.b	$71,$EA,$3F,$77,$FB,$2B,$AB,$F1,$DC,$DF,$85,$8C,$4F,$ED,$39,$FF
	.dc.b	$0A,$EB,$2F,$E3,$F3,$A8,$F3,$CF,$1E,$BF,$CC,$78,$96,$7F,$7B,$E2
	.dc.b	$75,$ED,$CE,$3C,$AF,$F3,$D3,$E1,$ED,$FD,$07,$83,$E3,$EE,$FD,$17
	.dc.b	$85,$9C,$BF,$83,$CE,$5F,$E1,$31,$E9,$F1,$3F,$51,$E1,$7A,$FC,$2F
	.dc.b	$D5,$78,$79,$BF,$E1,$F3,$97,$F8,$8C,$79,$DF,$A3,$BD,$7F,$80,$E9
	.dc.b	$E9,$7F,$8B,$C6,$5F,$F6,$EF,$42,$7C,$D1,$F6,$B3,$F3,$3F,$73,$3C
	.dc.b	$BF,$C3,$23,$2F,$F0,$E8,$FA,$5E,$37,$F0,$1F,$88,$FA,$9E,$35,$E2
	.dc.b	$FA,$33,$FC,$27,$F8,$07,$FC,$09,$1F,$52,$75,$87,$DE,$F2,$F5,$E9
	.dc.b	$FD,$DF,$2F,$87,$A9,$F7,$E7,$38,$7D,$C9,$CB,$F7,$BC,$DE,$DE,$97
	.dc.b	$E1,$5B,$1F,$82,$F3,$BC,$5E,$9E,$73,$FF,$50,$EA,$C3,$F1,$9E,$AF
	.dc.b	$9F,$D8,$FC,$77,$A5,$72,$FC,$5C,$FA,$FD,$3C,$E3,$CA,$FC,$CF,$4E
	.dc.b	$0F,$53,$DB,$83,$2F,$FA,$DF,$34,$EA,$31,$3E,$57,$5A,$57,$FD,$F9
	.dc.b	$5E,$7C,$BD,$91,$EA,$9D,$21,$E9,$AB,$84,$7E,$B3,$DC,$FA,$9E,$DF
	.dc.b	$EA,$DB,$87,$EA,$3D,$9F,$B1,$EE,$66,$FE,$3D,$FF,$35,$CB,$FD,$C1
	.dc.b	$97,$5F,$80,$C7,$C0,$FD,$F7,$5E,$E7,$E0,$7E,$07,$38,$FD,$09,$FD
	.dc.b	$D7,$BF,$F8,$5F,$88,$FD,$E7,$BF,$F8,$7F,$87,$CD,$FD,$69,$CB,$F6
	.dc.b	$23,$F8,$17,$C0,$FE,$37,$E1,$EC,$FD,$CF,$87,$FE,$0D,$F0,$99,$E4
	.dc.b	$FE,$07,$7F,$93,$9F,$1D,$7D,$3D,$2B,$F3,$9F,$4B,$F8,$6D,$72,$FC
	.dc.b	$DE,$0F,$81,$87,$F0,$ED,$BD,$4F,$F1,$0D,$BD,$4E,$6B,$D4,$E5,$F5
	.dc.b	$3F,$E3,$4E,$17,$C4,$7E,$9F,$E2,$3E,$77,$C4,$7E,$A7,$E2,$7E,$84
	.dc.b	$E6,$F8,$4E,$5E,$5C,$BF,$45,$B8,$FC,$F7,$B7,$2F,$D3,$73,$1F,$05
	.dc.b	$F3,$3F,$95,$E5,$F9,$EF,$E5,$F9,$7E,$9B,$99,$F5,$39,$7D,$59,$FE
	.dc.b	$FF,$2F,$C3,$67,$FC,$07,$3F,$C4,$4E,$6B,$E0,$4E,$5F,$85,$8F,$F0
	.dc.b	$7C,$FF,$1C,$D8,$F8,$D9,$7F,$0B,$CF,$39,$FE,$EA,$FF,$29,$CB,$FE
	.dc.b	$17,$EC,$70,$62,$BF,$48,$4E,$1F,$D8,$C7,$E6,$F9,$38,$71,$6E,$B6
	.dc.b	$7E,$AA,$F9,$E9,$F5,$F9,$3B,$72,$FE,$8F,$96,$CE,$6F,$89,$C7,$9B
	.dc.b	$4E,$34,$FE,$03,$7E,$A8,$FE,$09,$B5,$7F,$7E,$FE,$BF,$83,$E0,$61
	.dc.b	$F1,$A7,$E0,$3F,$89,$DF,$E0,$6F,$E9,$9F,$76,$BE,$87,$FC,$CB,$CC
	.dc.b	$E0,$F7,$E1,$C6,$7C,$95,$BF,$82,$FF,$28,$DB,$A3,$E6,$75,$FF,$01
	.dc.b	$7E,$77,$5F,$F0,$3E,$BC,$E5,$F9,$B9,$CB,$F9,$DF,$0B,$EA,$78,$3F
	.dc.b	$C2,$71,$FA,$DE,$17,$F0,$B9,$CD,$7E,$8E,$72,$FE,$97,$C1,$FB,$9E
	.dc.b	$37,$F1,$1F,$0E,$CE,$78,$3D,$3E,$36,$5F,$5D,$7F,$17,$E1,$FC,$27
	.dc.b	$8D,$FC,$67,$C5,$FC,$37,$8D,$9B,$F8,$33,$97,$E1,$C7,$95,$F8,$BB
	.dc.b	$FA,$5F,$8F,$BE,$5F,$C3,$F5,$FF,$3C,$9D,$46,$31,$57,$F2,$9D,$BD
	.dc.b	$33,$F5,$36,$BC,$BF,$42,$7F,$35,$B7,$AE,$B9,$6D,$FE,$65,$FD,$AB
	.dc.b	$FD,$C9,$1F,$C4,$F2,$79,$3F,$CD,$78,$7F,$44,$EF,$CA,$FD,$27,$95
	.dc.b	$9B,$FB,$F3,$CF,$CF,$FD,$47,$A7,$67,$F0,$7E,$9F,$7F,$43,$39,$7F
	.dc.b	$B2,$B4,$61,$FA,$FA,$F2,$7A,$1F,$B1,$F5,$AE,$3F,$8C,$7D,$1E,$8F
	.dc.b	$ED,$7D,$7C,$CF,$C5,$9F,$67,$AB,$FB,$9F,$5F,$DD,$EB,$5D,$7C,$13
	.dc.b	$FB,$DF,$5F,$38,$7E,$62,$FD,$0F,$CF,$E0,$AF,$E8,$B5,$F5,$3F,$9E
	.dc.b	$8C,$BF,$26,$3C,$6F,$EA,$32,$AF,$CD,$7D,$2F,$E9,$B3,$CB,$FC,$16
	.dc.b	$3E,$A7,$B5,$FC,$47,$DB,$B3,$FB,$FF,$E0,$1F,$EE,$9A,$FC,$9F,$6F
	.dc.b	$4F,$CB,$BF,$F5,$74,$EA,$BF,$DB,$7C,$7F,$39,$F4,$57,$D1,$E3,$FA
	.dc.b	$0F,$D3,$E4,$FA,$B3,$C2,$72,$F2,$E4,$FA,$95,$F6,$5B,$1D,$B9,$3E
	.dc.b	$A4,$E6,$3E,$C3,$F7,$FD,$EF,$2F,$C2,$7E,$45,$F3,$FC,$27,$E4,$FE
	.dc.b	$07,$35,$BB,$F8,$13,$F9,$7F,$88,$F6,$FC,$3F,$E6,$7E,$22,$DF,$E2
	.dc.b	$E7,$E0,$FC,$46,$70,$F9,$17,$F9,$FF,$87,$F8,$BF,$0F,$FA,$1F,$88
	.dc.b	$F8,$FF,$11,$9B,$F9,$D3,$97,$E8,$47,$E9,$F7,$F9,$BB,$FF,$42,$F2
	.dc.b	$3F,$DD,$BC,$A9,$FA,$3C,$39,$DF,$FC,$55,$4A,$D6,$3F,$6E,$B4,$A7
	.dc.b	$5C,$3D,$A3,$F4,$DB,$7D,$CE,$0D,$D7,$3E,$0F,$BF,$C3,$8C,$91,$8C
	.dc.b	$38,$9F,$89,$BD,$6F,$F8,$6E,$1B,$3F,$8D,$E1,$AF,$F1,$2E,$64,$75
	.dc.b	$E2,$E1,$3E,$CE,$0F,$6F,$ED,$8E,$BF,$C8,$F4,$F9,$A1,$D9,$F9,$4E
	.dc.b	$BC,$37,$0E,$47,$C1,$9E,$4F,$E7,$EB,$97,$7F,$E7,$3A,$FA,$78,$1F
	.dc.b	$9E,$ED,$B7,$FA,$57,$6F,$E9,$E7,$C5,$DD,$FA,$4E,$EF,$8D,$34,$FE
	.dc.b	$24,$53,$E9,$6D,$EC,$9F,$D5,$F7,$7A,$FC,$0F,$D6,$F8,$36,$FF,$69
	.dc.b	$E1,$57,$E4,$3F,$07,$C2,$DC,$7B,$BB,$5C,$3F,$CE,$A3,$F7,$9E,$2F
	.dc.b	$C7,$F2,$FF,$BF,$C7,$F9,$C3,$F8,$05,$B6,$63,$FC,$01,$18,$7A,$5B
	.dc.b	$C9,$FE,$03,$87,$D1,$F2,$B9,$F9,$CC,$3F,$D0,$F8,$70,$9C,$C3,$D1
	.dc.b	$E1,$3D,$9F,$B5,$E5,$EB,$CC,$FB,$9E,$87,$0F,$4F,$EF,$79,$D7,$0F
	.dc.b	$1B,$5F,$96,$7D,$D9,$EF,$E9,$FA,$E3,$F1,$DA,$D7,$FC,$59,$9F,$B5
	.dc.b	$E3,$C5,$F2,$38,$BC,$FC,$5F,$8D,$E1,$F7,$D7,$AC,$F8,$BA,$FA,$9F
	.dc.b	$ED,$78,$7B,$5D,$18,$DF,$DA,$F0,$3E,$5F,$AD,$F1,$3D,$1F,$CD,$7A
	.dc.b	$76,$7E,$7B,$5F,$2F,$2A,$BE,$5F,$A7,$EB,$9F,$F3,$F7,$90,$E0,$C4
	.dc.b	$7D,$DD,$AF,$3F,$D5,$7D,$91,$F8,$43,$CE,$7C,$4D,$BF,$59,$EA,$7D
	.dc.b	$8D,$AC,$FD,$73,$67,$FA,$37,$7F,$BE,$F9,$BE,$07,$EF,$FA,$DF,$B5
	.dc.b	$F7,$AC,$FE,$2A,$6B,$EC,$8F,$C2,$FB,$1F,$C4,$3F,$6E,$FF,$20,$4F
	.dc.b	$E4,$BE,$0B,$1F,$0D,$F8,$A7,$5E,$ED,$9F,$D0,$52,$BF,$1F,$F0,$7F
	.dc.b	$C1,$7E,$0F,$F2,$3F,$0B,$FC,$1F,$E0,$2E,$1F,$9C,$9A,$FC,$E8,$FC
	.dc.b	$BF,$C1,$F8,$FE,$1F,$F3,$3F,$09,$E5,$F8,$5B,$7E,$09,$AF,$E2,$B4
	.dc.b	$FF,$67,$F1,$FA,$A3,$1B,$79,$DD,$6D,$7C,$2F,$D7,$0B,$E1,$3F,$96
	.dc.b	$E0,$F8,$2F,$ED,$38,$3E,$37,$C3,$7A,$2F,$8F,$7F,$FA,$2A,$BF,$73
	.dc.b	$C5,$A6,$9F,$BA,$F8,$6F,$6B,$F2,$38,$7F,$6B,$A7,$D1,$E1,$B3,$93
	.dc.b	$F8,$1C,$65,$FE,$0B,$F0,$FF,$49,$FA,$C7,$D2,$8F,$67,$67,$F5,$4D
	.dc.b	$FD,$73,$F8,$7D,$FD,$6F,$EA,$37,$FC,$06,$1E,$BB,$E1,$F1,$16,$BF
	.dc.b	$AC,$60,$AC,$6F,$F1,$EB,$D7,$5F,$47,$0F,$E1,$F1,$F8,$BD,$AC,$FE
	.dc.b	$1A,$6B,$F5,$E3,$F8,$AF,$1F,$E4,$7C,$7F,$E2,$FC,$B6,$BF,$91,$BF
	.dc.b	$95,$F1,$73,$1F,$90,$9F,$E3,$BC,$9F,$99,$F0,$FF,$91,$FF,$00,$FD
	.dc.b	$3A,$FC,$24,$63,$97,$F3,$DD,$D6,$7F,$33,$3C,$3B,$BF,$45,$DB,$CB
	.dc.b	$BB,$F4,$9C,$DD,$3B,$B3,$87,$FA,$D7,$34,$6A,$31,$5E,$97,$F8,$AE
	.dc.b	$F5,$FE,$42,$FA,$DF,$69,$EB,$3F,$A0,$EB,$EB,$7B,$6B,$5F,$F4,$4E
	.dc.b	$1E,$DE,$4F,$CC,$F3,$7F,$AC,$67,$67,$B0,$F3,$6B,$FB,$DE,$FF,$8B
	.dc.b	$DF,$FB,$FF,$2E,$D7,$EE,$A6,$BF,$AB,$8F,$97,$E3,$7A,$63,$F8,$86
	.dc.b	$B5,$FE,$B4,$FA,$DF,$79,$EB,$3F,$9A,$EB,$E9,$7F,$DD,$9A,$7A,$7A
	.dc.b	$FF,$2D,$C7,$F4,$E7,$F7,$5C,$97,$7F,$6C,$FD,$59,$FE,$1F,$5F,$7B
	.dc.b	$C7,$FE,$31,$E2,$59,$C9,$F5,$63,$2F,$D7,$F2,$7F,$C5,$BC,$0F,$5B
	.dc.b	$F0,$FE,$56,$1F,$C4,$F9,$DA,$F5,$6E,$1F,$DF,$6F,$83,$F8,$1B,$E7
	.dc.b	$E9,$7E,$43,$CA,$B3,$F8,$57,$BF,$A9,$98,$E5,$3F,$96,$F3,$FC,$7E
	.dc.b	$9F,$E6,$3D,$1B,$5F,$8D,$7C,$FE,$86,$63,$C5,$3F,$9E,$F4,$BD,$11
	.dc.b	$E3,$3C,$DB,$FA,$1F,$ED,$BA,$B6,$C7,$2F,$F2,$08,$F5,$CF,$E3,$37
	.dc.b	$F3,$BE,$93,$CE,$53,$9E,$DE,$B7,$AE,$DE,$B7,$E3,$7B,$7F,$B2,$F5
	.dc.b	$7E,$7B,$63,$E5,$7A,$F5,$F2,$C7,$EE,$3D,$6F,$5C,$78,$CF,$59,$FA
	.dc.b	$FE,$4F,$A8,$D3,$26,$CF,$29,$E7,$D3,$DD,$8F,$9D,$F8,$19,$5E,$8F
	.dc.b	$F3,$CE,$AA,$F5,$5F,$C8,$F7,$BF,$85,$FB,$BF,$88,$6C,$7E,$0F,$DC
	.dc.b	$AE,$CF,$E0,$9E,$EF,$E3,$BD,$EF,$E2,$7E,$F7,$E4,$3E,$26,$D7,$E2
	.dc.b	$E6,$BF,$5A,$3F,$8C,$FB,$FF,$96,$F8,$8F,$E3,$7F,$01,$68,$42,$FB
	.dc.b	$EF,$88,$C4,$7E,$03,$E1,$3F,$C8,$B1,$8E,$4F,$C0,$61,$D2,$BF,$39
	.dc.b	$F0,$FA,$F8,$9F,$D2,$BC,$2F,$F4,$FF,$0B,$6B,$67,$61,$97,$86,$FE
	.dc.b	$04,$7F,$25,$DF,$E0,$34,$F8,$05,$3E,$01,$FC,$AF,$5F,$73,$FC,$7B
	.dc.b	$5A,$FB,$EF,$DC,$FE,$EB,$5F,$6B,$A3,$E1,$F9,$6A,$FD,$FF,$C3,$FC
	.dc.b	$2F,$89,$B5,$E9,$C3,$E4,$47,$F0,$4F,$88,$D4,$FE,$FE,$BE,$49,$EC
	.dc.b	$7F,$98,$E7,$EC,$69,$EC,$29,$EC,$3F,$9B,$E7,$E5,$7E,$3E,$5E,$57
	.dc.b	$F8,$67,$07,$A9,$FE,$1D,$A5,$7F,$73,$1F,$5D,$FD,$DE,$FF,$73,$2F
	.dc.b	$E3,$6D,$8F,$E2,$D9,$D7,$67,$BE,$1C,$7F,$EC,$EF,$DC,$BF,$D7,$33
	.dc.b	$9E,$4C,$72,$FE,$23,$2D,$6D,$6B,$FD,$3D,$7C,$1F,$C7,$F0,$7E,$47
	.dc.b	$0E,$9B,$FE,$43,$6E,$DC,$36,$F8,$CD,$7F,$D0,$38,$7C,$37,$9E,$AE
	.dc.b	$4F,$E6,$F1,$F9,$AE,$4A,$FA,$F9,$FF,$9D,$F2,$FE,$79,$CC,$72,$9F
	.dc.b	$5B,$ED,$3D,$3C,$1E,$B7,$C3,$BD,$E6,$FF,$21,$E7,$ED,$9F,$C8,$66
	.dc.b	$DF,$DB,$1F,$2F,$7F,$6B,$4F,$69,$4F,$6E,$FF,$B2,$DB,$E8,$77,$7E
	.dc.b	$D3,$A2,$D7,$D5,$7E,$A7,$4E,$63,$E7,$CF,$EE,$BB,$FD,$71,$F1,$B2
	.dc.b	$AF,$9B,$97,$EB,$BF,$7D,$CB,$FC,$1B,$4F,$53,$E6,$3E,$7E,$9E,$A6
	.dc.b	$99,$3E,$0B,$97,$E1,$E7,$F8,$9F,$17,$F8,$6F,$83,$67,$F1,$CF,$E3
	.dc.b	$7C,$2C,$C7,$E2,$27,$F8,$9F,$8B,$F9,$1F,$2F,$F8,$AF,$8D,$6B,$F2
	.dc.b	$DE,$27,$E5,$7C,$6F,$E3,$2E,$63,$F0,$B3,$F9,$9F,$1B,$F8,$EF,$89
	.dc.b	$67,$F3,$CF,$E7,$7C,$BC,$C7,$E6,$27,$F9,$1F,$91,$FA,$1F,$2F,$F9
	.dc.b	$2F,$95,$6B,$F4,$CF,$E9,$7C,$BC,$C7,$E8,$27,$FD,$DF,$FC,$03,$FE
	.dc.b	$D9,$EE,$FC,$3E,$FE,$AA,$FD,$2B,$E8,$7F,$4F,$BD,$7F,$3F,$1B,$7F
	.dc.b	$43,$9F,$99,$A7,$A8,$FD,$7E,$99,$BF,$CA,$CF,$AA,$7F,$67,$E9,$F9
	.dc.b	$3C,$FF,$DE,$3E,$6F,$57,$F7,$DE,$85,$AD,$9F,$C8,$74,$FF,$64,$E3
	.dc.b	$EA,$8F,$E0,$DC,$5F,$3D,$FD,$0F,$1F,$D3,$7E,$A9,$EE,$36,$7C,$D3
	.dc.b	$D0,$7A,$8F,$E0,$B4,$F5,$7E,$A4,$78,$5F,$B0,$FF,$00,$E4,$FE,$29
	.dc.b	$ED,$7C,$EF,$77,$F8,$B7,$A5,$F4,$3D,$EC,$DF,$A5,$CB,$EB,$9F,$EC
	.dc.b	$9A,$7F,$AF,$3F,$75,$7C,$07,$D8,$F5,$F3,$EB,$FF,$6A,$A5,$3D,$AF
	.dc.b	$BB,$EC,$EB,$25,$6B,$1C,$C7,$1C,$39,$61,$C6,$3E,$66,$3C,$6F,$AF
	.dc.b	$B5,$F8,$8F,$80,$C5,$A3,$17,$FC,$B3,$DE,$AF,$83,$1E,$B3,$F9,$0F
	.dc.b	$80,$FE,$71,$EF,$7E,$4B,$E0,$BF,$9D,$7C,$0D,$C3,$F9,$AC,$D7,$F9
	.dc.b	$6C,$7E,$63,$E0,$BC,$3B,$FC,$39,$F5,$1E,$ED,$3D,$4F,$D8,$E0,$DB
	.dc.b	$F8,$77,$D3,$C3,$C6,$3C,$C7,$1D,$3F,$4B,$F0,$FE,$E6,$C7,$B4,$FC
	.dc.b	$84,$65,$FC,$97,$C3,$FC,$4F,$86,$FD,$6F,$C4,$D9,$FC,$B3,$F2,$3E
	.dc.b	$1F,$31,$E8,$E5,$FB,$35,$FB,$4F,$88,$F3,$CF,$E4,$39,$FC,$EF,$A4
	.dc.b	$E1,$B7,$E1,$1C,$1E,$AB,$FC,$1E,$BE,$D7,$F1,$5A,$FF,$00,$C7,$D1
	.dc.b	$5F,$9C,$7F,$B5,$73,$66,$3A,$CF,$F0,$6C,$FE,$FE,$9F,$C2,$38,$1C
	.dc.b	$DE,$2D,$B0,$F9,$27,$F8,$6E,$9A,$C7,$C9,$AA,$F7,$18,$E7,$FE,$27
	.dc.b	$9E,$F6,$FD,$04,$7F,$8D,$67,$FC,$6B,$FC,$02,$DB,$5A,$7E,$63,$A3
	.dc.b	$8B,$F1,$2F,$87,$C1,$AF,$84,$FC,$8E,$9C,$C3,$A7,$9E,$5F,$89,$9C
	.dc.b	$C6,$33,$FD,$6F,$47,$0E,$9C,$70,$2F,$79,$7D,$4E,$8E,$BD,$BA,$7B
	.dc.b	$F6,$F0,$EA,$CD,$6F,$5F,$F1,$D6,$CA,$EB,$C1,$EC,$9F,$8C,$6B,$8B
	.dc.b	$C3,$C5,$9E,$0F,$8F,$39,$C7,$DC,$F4,$DF,$CF,$B7,$9A,$3E,$9F,$07
	.dc.b	$AE,$B7,$1F,$E2,$2D,$3E,$24,$7E,$3B,$AB,$E3,$F8,$BF,$03,$B2,$CE
	.dc.b	$6F,$D3,$E2,$65,$F5,$D7,$A9,$FE,$6D,$CF,$5E,$DB,$FA,$9C,$DF,$D9
	.dc.b	$7D,$47,$E2,$34,$F3,$34,$F6,$47,$E8,$B4,$FB,$7E,$4F,$BA,$72,$7E
	.dc.b	$43,$BF,$EF,$F9,$5F,$61,$FC,$0F,$93,$F6,$BC,$2C,$E3,$FA,$C9,$CB
	.dc.b	$FB,$5F,$DE,$1E,$7F,$E2,$3C,$CF,$C2,$43,$CF,$FC,$5F,$9D,$8A,$D5
	.dc.b	$F1,$9C,$BF,$B8,$3F,$23,$E7,$FE,$2B,$CF,$FC,$9F,$A3,$F8,$C9,$C9
	.dc.b	$FB,$E8,$CB,$FC,$03,$D0,$FC,$CF,$A5,$F9,$27,$F3,$7E,$D7,$E5,$3C
	.dc.b	$FC,$D6,$F8,$FF,$1B,$3E,$57,$F1,$FB,$FA,$B4,$F2,$C7,$F0,$9D,$FD
	.dc.b	$EE,$AF,$D4,$FE,$5B,$7F,$33,$F9,$8D,$BD,$4F,$E6,$B6,$CB,$F8,$B8
	.dc.b	$CB,$F8,$D8,$F9,$3E,$8F,$EC,$BD,$BF,$97,$E8,$DC,$7E,$84,$FE,$DB
	.dc.b	$E0,$33,$87,$E9,$AF,$D4,$FE,$5F,$5F,$53,$E5,$BF,$B1,$C7,$F4,$A3
	.dc.b	$D9,$B7,$BB,$C3,$FE,$D9,$5F,$3A,$7F,$8A,$F2,$59,$CC,$7E,$D3,$E0
	.dc.b	$3D,$2F,$FA,$E7,$C3,$70,$7E,$0B,$D6,$FE,$17,$F0,$9F,$84,$F5,$EE
	.dc.b	$7F,$89,$9F,$E1,$FF,$01,$9C,$7F,$7F,$7F,$8A,$F6,$BF,$8B,$FF,$00
	.dc.b	$FF,$B9,$30,$FC,$84,$62,$BF,$1F,$F0,$7A,$F8,$5E,$11,$FE,$C1,$BE
	.dc.b	$0F,$D9,$F8,$5E,$7F,$01,$F9,$66,$E3,$F9,$BF,$84,$EF,$F0,$F9,$BE
	.dc.b	$57,$F9,$CF,$85,$F1,$FC,$57,$E7,$BE,$0E,$E3,$FA,$39,$F3,$FC,$4E
	.dc.b	$70,$F1,$5F,$91,$FF,$55,$E9,$E4,$E7,$FE,$79,$E1,$7F,$A8,$E7,$D0
	.dc.b	$7B,$0F,$56,$FC,$5F,$F7,$35,$71,$CB,$FC,$3D,$4D,$FD,$75,$EF,$3D
	.dc.b	$33,$ED,$7E,$26,$DE,$D7,$C7,$B7,$EE,$3E,$1B,$E7,$FC,$4D,$9F,$95
	.dc.b	$F1,$1F,$BC,$F8,$8C,$E3,$FC,$06,$FE,$AC,$F9,$68,$3E,$05,$AF,$9E
	.dc.b	$3F,$C5,$FA,$FB,$DF,$86,$7B,$CF,$8B,$B7,$37,$94,$F1,$DD,$5E,$F9
	.dc.b	$F6,$9E,$F3,$D7,$97,$BD,$F9,$FB,$7B,$DF,$93,$B7,$F1,$6D,$BF,$1B
	.dc.b	$AD,$9F,$C4,$6D,$FC,$6B,$6C,$E3,$FC,$7E,$FF,$27,$3F,$C7,$AB,$F2
	.dc.b	$BC,$DF,$C8,$36,$FC,$BF,$36,$67,$F2,$53,$97,$F2,$91,$FC,$9B,$3F
	.dc.b	$CE,$F0,$FF,$28,$D7,$F3,$FC,$D9,$BF,$CD,$4E,$5F,$CE,$47,$F2,$EE
	.dc.b	$3F,$D2,$F2,$FF,$30,$E5,$BB,$FF,$35,$9F,$D4,$F3,$67,$0F,$D2,$5F
	.dc.b	$95,$FC,$9E,$BE,$57,$F2,$BA,$E5,$FC,$7C,$65,$FC,$8C,$79,$5F,$CD
	.dc.b	$E7,$E5,$7F,$3B,$9E,$5F,$CB,$C6,$5F,$CC,$C7,$F8,$27,$17,$3F,$EF
	.dc.b	$BA,$70,$FF,$00,$EA,$D7,$56,$76,$FF,$22,$5F,$0C,$7F,$7D,$5E,$77
	.dc.b	$F2,$3D,$5F,$90,$D7,$D1,$18,$EA,$6D,$B8,$FC,$6C,$E9,$FC,$7C,$7A
	.dc.b	$5E,$1B,$7A,$5E,$5B,$65,$FC,$AC,$65,$FC,$BC,$7B,$7C,$3F,$E3,$3E
	.dc.b	$25,$9F,$E2,$8F,$C1,$EF,$FE,$3B,$E2,$E6,$7F,$82,$CF,$C5,$F1,$FF
	.dc.b	$90,$F8,$96,$7F,$8E,$78,$9F,$27,$C2,$CF,$1F,$C4,$BF,$E4,$F3,$F3
	.dc.b	$7C,$BF,$F1,$FF,$B3,$FE,$54,$AF,$93,$E4,$E3,$C7,$D6,$5F,$B8,$27
	.dc.b	$2A,$F8,$47,$F9,$A7,$9D,$7C,$70,$FA,$7E,$57,$6F,$33,$EE,$3E,$0F
	.dc.b	$33,$EF,$79,$99,$AD,$E3,$DC,$FA,$F9,$7F,$B2,$63,$D5,$1F,$CE,$78
	.dc.b	$7E,$03,$FC,$A3,$7F,$80,$E6,$76,$E5,$FC,$2E,$FF,$01,$FC,$76,$9E
	.dc.b	$79,$F5,$4E,$67,$E1,$C7,$8C,$FE,$31,$83,$7F,$CB,$79,$FF,$2B,$CE
	.dc.b	$FC,$C7,$9D,$F3,$3C,$FC,$E1,$C7,$8F,$E4,$56,$A3,$18,$AB,$D7,$1F
	.dc.b	$CE,$B6,$AF,$E5,$76,$F6,$BF,$D1,$A3,$2F,$F2,$28,$F5,$BD,$34,$AF
	.dc.b	$CB,$C7,$EC,$69,$F7,$23,$18,$FD,$AF,$3B,$FD,$15,$C9,$F6,$A7,$CB
	.dc.b	$5F,$B3,$F5,$BF,$0F,$EC,$D9,$F4,$FB,$3F,$B7,$F6,$73,$A7,$EC,$AF
	.dc.b	$F1,$B3,$FE,$83,$FD,$EF,$FD,$61,$7E,$0F,$73,$1F,$09,$F9,$27,$5F
	.dc.b	$0D,$F9,$4F,$7B,$35,$F1,$A7,$F8,$27,$B5,$F9,$8F,$72,$CF,$CA,$F7
	.dc.b	$3F,$84,$7B,$99,$DB,$F8,$1D,$FE,$7A,$7F,$86,$7C,$27,$E8,$3E,$0A
	.dc.b	$E1,$F4,$BE,$0B,$F8,$87,$3E,$71,$FE,$17,$7F,$A6,$9F,$2D,$78,$E7
	.dc.b	$FD,$A5,$C1,$FE,$E8,$8F,$D2,$DB,$5D,$C7,$CE,$DB,$D6,$FC,$1B,$FE
	.dc.b	$2F,$F0,$7E,$78,$FD,$53,$E7,$8F,$E3,$7F,$07,$9E,$9F,$AB,$39,$AF
	.dc.b	$F6,$C6,$4F,$88,$FD,$DF,$C3,$FC,$FF,$88,$B9,$79,$A7,$F7,$FF,$11
	.dc.b	$9C,$3F,$23,$7E,$67,$D3,$AF,$99,$B7,$EE,$E6,$FA,$8F,$C0,$E7,$FB
	.dc.b	$0F,$C2,$F8,$7F,$33,$FD,$FB,$66,$A9,$F4,$47,$E3,$BE,$1F,$78,$A3
	.dc.b	$73,$AB,$E5,$6E,$3F,$E2,$FB,$7E,$2D,$FE,$33,$BD,$C7,$F5,$73,$F8
	.dc.b	$FD,$B3,$87,$D6,$ED,$FE,$3F,$B7,$E4,$FA,$FF,$90,$EF,$71,$FD,$9C
	.dc.b	$FE,$5E,$B3,$7F,$7B,$9F,$F9,$3E,$DF,$9B,$E2,$FF,$73,$FF,$00,$FF
	.dc.b	$B9,$B1,$FA,$71,$AB,$E1,$1F,$C0,$67,$5C,$3F,$EF,$7E,$1F,$F7,$14
	.dc.b	$E7,$A3,$F4,$DD,$3D,$3A,$6C,$E7,$1F,$C7,$F6,$65,$F6,$47,$A1,$FC
	.dc.b	$9F,$0F,$A2,$3F,$0A,$FC,$08,$DC,$7F,$00,$E1,$FD,$A7,$57,$9E,$3F
	.dc.b	$17,$C1,$EC,$E7,$FD,$C3,$9D,$BD,$15,$9C,$3F,$43,$E2,$FC,$2E,$AF
	.dc.b	$DF,$78,$BF,$13,$B2,$E7,$F0,$67,$F8,$15,$66,$FC,$73,$FA,$8E,$7F
	.dc.b	$A4,$F9,$63,$D6,$FB,$4F,$CF,$F3,$FB,$73,$F9,$5D,$9E,$5A,$FC,$3E
	.dc.b	$9E,$5B,$FD,$2B,$97,$F1,$DC,$BF,$B5,$7E,$A6,$FF,$C5,$7C,$1F,$B7
	.dc.b	$E0,$59,$FB,$1E,$0D,$7F,$95,$74,$A7,$F8,$E7,$9B,$F8,$2A,$67,$FA
	.dc.b	$13,$F4,$5F,$BF,$3D,$AF,$C3,$F8,$7A,$CB,$F4,$15,$A9,$FC,$5F,$89
	.dc.b	$9E,$19,$87,$1B,$FC,$56,$19,$C3,$AF,$A5,$C1,$F0,$4F,$EA,$E9,$E9
	.dc.b	$FE,$5B,$CC,$FE,$65,$E8,$5B,$FE,$32,$FE,$6B,$CD,$E9,$87,$1B,$D3
	.dc.b	$E6,$3A,$79,$9F,$A0,$F4,$BF,$9E,$61,$4F,$E4,$67,$E9,$3D,$3E,$7B
	.dc.b	$34,$FF,$80,$18,$7E,$8E,$F3,$1E,$D8,$F9,$BB,$63,$83,$D9,$1E,$1B
	.dc.b	$E5,$1F,$2B,$1F,$7C,$75,$DF,$6F,$C8,$D3,$D2,$DB,$FD,$96,$FE,$09
	.dc.b	$F8,$17,$CB,$2F,$F7,$CD,$7D,$EA,$F2,$FB,$FF,$BE,$F4,$AC,$FE,$EF
	.dc.b	$D2,$AF,$E1,$23,$D7,$EF,$7F,$04,$9F,$54,$FD,$8C,$7B,$6B,$E6,$7F
	.dc.b	$01,$19,$3D,$D9,$35,$D4,$7D,$33,$D3,$3E,$7D,$3F,$88,$FA,$BE,$B9
	.dc.b	$C9,$CB,$D4,$CD,$7B,$7D,$8F,$E2,$FE,$B7,$D0,$7F,$8C,$FB,$16,$FF
	.dc.b	$8B,$4F,$D4,$F7,$F3,$7F,$17,$E1,$BF,$8F,$FB,$5F,$61,$B1,$9B,$FE
	.dc.b	$3D,$F0,$1F,$73,$E1,$F3,$1E,$CF,$7F,$F9,$3F,$C0,$FE,$01,$B7,$F4
	.dc.b	$5F,$F1,$74,$FF,$B1,$7F,$0A,$FF,$5E,$71,$BE,$1F,$73,$F1,$BE,$0B
	.dc.b	$7F,$02,$B7,$C3,$FE,$33,$E0,$F3,$18,$CF,$EF,$46,$3E,$1F,$F2,$38
	.dc.b	$6A,$FD,$98,$FF,$37,$E1,$AF,$F0,$06,$BF,$E6,$2F,$2F,$FD,$2B,$8F
	.dc.b	$E2,$78,$3D,$FE,$0F,$EE,$36,$F3,$BF,$BA,$E0,$B5,$FC,$11,$F2,$E1
	.dc.b	$FA,$2A,$CB,$FB,$9F,$86,$F4,$FC,$3F,$E9,$BE,$22,$E1,$F9,$77,$DB
	.dc.b	$F0,$FF,$AA,$AC,$DF,$E2,$B9,$FC,$77,$FC,$49,$B1,$FC,$15,$FE,$B3
	.dc.b	$CF,$F5,$E3,$E0,$4F,$E8,$F2,$F1,$E9,$FB,$9D,$2F,$37,$FB,$D6,$FF
	.dc.b	$D0,$7C,$BF,$5A,$F3,$B7,$D0,$D3,$F7,$AD,$CB,$F5,$CF,$D4,$DB,$F8
	.dc.b	$0E,$D9,$C7,$F7,$DF,$BF,$B8,$BE,$CF,$07,$F0,$A8,$FF,$5E,$4F,$F9
	.dc.b	$6F,$44,$E7,$83,$EF,$F0,$EF,$0E,$3A,$7F,$06,$C5,$D1,$F8,$4D,$FF
	.dc.b	$88,$E7,$F8,$6E,$1A,$FF,$9D,$F4,$FC,$86,$1C,$6F,$4F,$F0,$7E,$4F
	.dc.b	$C6,$F2,$FF,$1A,$E6,$B3,$FC,$2D,$FC,$8F,$3F,$F1,$EA,$CE,$3E,$ED
	.dc.b	$7C,$AF,$C0,$DE,$F3,$7F,$41,$DF,$DF,$3F,$3B,$83,$D8,$F9,$78,$7E
	.dc.b	$26,$DE,$48,$FE,$12,$5B,$FE,$92,$FB,$F5,$F5,$3E,$DE,$0F,$99,$D5
	.dc.b	$FA,$6E,$CB,$3F,$07,$AE,$BF,$3E,$3E,$97,$67,$EB,$27,$C7,$3F,$4F
	.dc.b	$3F,$A4,$F9,$71,$FA,$1B,$7D,$47,$E9,$6F,$EE,$7F,$C8,$DB,$FD,$4A
	.dc.b	$F2,$47,$F3,$37,$D4,$FF,$6A,$7D,$47,$E1,$B6,$CB,$FE,$49,$97,$97
	.dc.b	$FC,$03,$C6,$FF,$6C,$70,$7C,$B8,$C5,$7E,$0F,$C9,$AF,$E4,$23,$F8
	.dc.b	$3F,$8D,$C3,$2F,$2F,$95,$FC,$2B,$C6,$C5,$F7,$8F,$E1,$BE,$36,$32
	.dc.b	$69,$FE,$A9,$AF,$CB,$4E,$61,$5C,$E7,$F2,$9E,$3F,$F1,$6F,$1F,$5E
	.dc.b	$3D,$9C,$9F,$93,$8C,$BF,$95,$F1,$FA,$7B,$5F,$C7,$BC,$CF,$D3,$7A
	.dc.b	$3F,$C9,$3F,$80,$7F,$D7,$9C,$1F,$3A,$34,$FE,$8E,$B4,$FE,$B7,$D1
	.dc.b	$E1,$E9,$FE,$BF,$D0,$B3,$FA,$97,$A7,$A7,$BD,$3F,$DB,$99,$7F,$B8
	.dc.b	$EB,$F5,$DE,$9F,$87,$D3,$FD,$C7,$A7,$67,$F6,$4F,$97,$DA,$FD,$E5
	.dc.b	$67,$0F,$EF,$D7,$FE,$AB,$43,$D3,$63,$FE,$E7,$C7,$FC,$03,$5F,$E8
	.dc.b	$DA,$F1,$FB,$9F,$E1,$08,$C4,$7F,$B8,$69,$ED,$67,$DE,$F8,$D7,$FC
	.dc.b	$37,$D8,$F9,$1E,$FD,$C3,$DA,$D7,$97,$B9,$FC,$4F,$E0,$F7,$A7,$F0
	.dc.b	$DF,$82,$FA,$1E,$F7,$F1,$77,$E9,$7C,$15,$C3,$E3,$CF,$F1,$BF,$81
	.dc.b	$CD,$FA,$BE,$27,$EC,$7C,$27,$A5,$FB,$5F,$09,$70,$F9,$F3,$ED,$F8
	.dc.b	$7C,$DF,$C2,$F8,$6F,$C0,$61,$F0,$5B,$19,$BF,$AF,$5F,$17,$E2,$33
	.dc.b	$1F,$43,$E2,$7F,$11,$7F,$25,$B7,$F3,$9F,$C6,$57,$CD,$F8,$8C,$DF
	.dc.b	$D8,$E2,$FC,$85,$7D,$17,$F2,$53,$67,$F0,$F3,$F5,$78,$B3,$87,$DA
	.dc.b	$DF,$F3,$1C,$BF,$67,$87,$F3,$5F,$13,$71,$FC,$7C,$FD,$DA,$CD,$FE
	.dc.b	$0B,$83,$FC,$07,$DD,$FE,$AE,$8F,$F6,$97,$52,$73,$CD,$F8,$57,$7C
	.dc.b	$FF,$87,$E5,$CD,$FE,$33,$C0,$E7,$D1,$F8,$BC,$AC,$E7,$5F,$CA,$93
	.dc.b	$B1,$2E,$9E,$13,$A3,$F3,$73,$AB,$FC,$B6,$BE,$8F,$0A,$D6,$CF,$D0
	.dc.b	$61,$FE,$8F,$C7,$F3,$FD,$FB,$7F,$43,$7F,$03,$C3,$FD,$07,$6F,$C2
	.dc.b	$F0,$EE,$3C,$F4,$F2,$C7,$E9,$7B,$FE,$3F,$87,$AA,$F3,$56,$B2,$FF
	.dc.b	$36,$FD,$BF,$FA,$2A,$79,$6A,$F0,$FF,$90,$4E,$67,$8C,$E5,$E3,$E3
	.dc.b	$FD,$4A,$DF,$8F,$F5,$BC,$6E,$3F,$01,$F6,$3D,$1B,$87,$6F,$4A,$BD
	.dc.b	$1E,$FE,$FE,$E3,$CB,$E5,$FE,$03,$CE,$F1,$BF,$82,$F8,$9B,$3E,$87
	.dc.b	$F9,$B7,$3F,$FA,$C6,$3C,$F5,$EA,$F5,$7F,$13,$E8,$59,$F7,$B5,$F6
	.dc.b	$BF,$03,$D5,$FC,$7F,$A3,$E6,$9F,$C1,$71,$FF,$52,$7E,$4E,$FF,$94
	.dc.b	$DB,$51,$F8,$29,$47,$B2,$3E,$6E,$FF,$33,$D5,$FC,$DF,$BB,$67,$E8
	.dc.b	$FB,$95,$F9,$EF,$D2,$F5,$B7,$1F,$63,$E0,$FF,$47,$ED,$7D,$6F,$7F
	.dc.b	$F4,$BE,$E5,$AF,$B6,$FD,$AA,$DC,$7A,$0F,$19,$4F,$A4,$53,$DF,$A7
	.dc.b	$A5,$FC,$0E,$9F,$84,$F7,$FF,$67,$EE,$D9,$FC,$97,$E9,$3F,$A3,$F7
	.dc.b	$BF,$DB,$DE,$02,$FF,$CF,$73,$9F,$07,$FA,$0D,$29,$F0,$1D,$30,$FC
	.dc.b	$AF,$C4,$70,$F8,$3F,$C9,$FC,$3F,$2F,$85,$B5,$E0,$7F,$2F,$5B,$8F
	.dc.b	$F6,$4F,$F0,$0F,$F9,$3D,$F2,$FC,$4F,$F7,$B8,$C4,$7E,$93,$57,$C4
	.dc.b	$E7,$9B,$CE,$EF,$97,$D3,$87,$E9,$3E,$1F,$D7,$97,$E9,$BE,$22,$E5
	.dc.b	$FA,$C6,$BF,$ED,$DE,$0F,$F2,$DB,$F1,$B6,$FD,$6E,$FF,$13,$0F,$D7
	.dc.b	$EF,$F1,$B0,$B7,$F3,$1D,$BF,$3A,$7C,$CF,$B0,$F7,$EB,$E8,$69,$F5
	.dc.b	$35,$F5,$3B,$3C,$9A,$7D,$4E,$7F,$DF,$B6,$3F,$82,$B5,$F5,$4F,$DA
	.dc.b	$E8,$F9,$5D,$7F,$73,$BA,$D7,$E0,$9A,$FC,$D7,$E8,$76,$7E,$0B,$C7
	.dc.b	$F4,$CF,$D1,$C6,$BF,$CB,$F9,$3E,$9B,$EA,$3E,$B6,$5E,$A7,$81,$F7
	.dc.b	$BA,$BF,$1B,$D5,$F7,$5F,$C7,$F4,$59,$FC,$AB,$FC,$71,$D9,$F8,$6F
	.dc.b	$13,$F2,$BE,$27,$E1,$7C,$6F,$CB,$F8,$96,$BF,$3A,$FE,$26,$B7,$1F
	.dc.b	$8E,$F4,$FF,$3B,$E1,$FE,$37,$D2,$FC,$FF,$89,$6B,$F4,$AF,$E4,$6B
	.dc.b	$71,$F9,$6F,$1F,$F4,$BE,$6F,$E5,$7C,$8B,$3F,$9A,$6A,$F1,$1E,$2F
	.dc.b	$EA,$BC,$56,$1F,$AE,$F3,$3F,$3D,$E6,$FE,$C3,$CC,$FD,$07,$9B,$FE
	.dc.b	$EC,$FD,$2B,$F9,$36,$5C,$B3,$E9,$92,$B5,$9F,$29,$FD,$EE,$1E,$97
	.dc.b	$F7,$FC,$3E,$D7,$F5,$5C,$15,$FB,$4E,$DF,$D7,$47,$EC,$7C,$BF,$17
	.dc.b	$AF,$FB,$3F,$37,$C9,$E7,$5B,$FE,$B9,$B7,$F7,$07,$F7,$77,$C1,$FD
	.dc.b	$EF,$AD,$EB,$F4,$3F,$79,$EB,$FB,$7D,$7B,$5C,$9F,$E0,$15,$B8,$FB
	.dc.b	$07,$A8,$FE,$0F,$D3,$F6,$1A,$7D,$83,$67,$A8,$FE,$1B,$EB,$7C,$BF
	.dc.b	$5F,$F8,$67,$AD,$F3,$7D,$7B,$5E,$A7,$F8,$85,$6E,$3F,$90,$F2,$7F
	.dc.b	$22,$FC,$FB,$FC,$C9,$1F,$E6,$3D,$DF,$09,$FE,$91,$9C,$FB,$29,$7C
	.dc.b	$16,$7E,$03,$EE,$3B,$F8,$1F,$BD,$87,$1F,$7F,$F0,$1F,$03,$73,$F9
	.dc.b	$AD,$7F,$C0,$B8,$A3,$C1,$EF,$FE,$1F,$E0,$BF,$94,$7C,$0F,$E2,$70
	.dc.b	$B8,$7D,$66,$BC,$9F,$E5,$DF,$05,$F8,$FB,$F5,$7C,$37,$E4,$7E,$23
	.dc.b	$D9,$F0,$F7,$0F,$86,$7E,$77,$89,$19,$8F,$D0,$71,$FC,$7B,$FC,$D7
	.dc.b	$C4,$7A,$67,$FC,$6D,$BF,$A5,$E7,$E3,$B1,$FF,$23,$47,$C0,$32,$E1
	.dc.b	$F8,$3C,$1F,$01,$FC,$8F,$07,$D1,$F8,$7F,$D4,$65,$67,$F2,$BA,$57
	.dc.b	$E4,$BF,$5F,$E2,$37,$1D,$B0,$FD,$85,$FD,$B7,$F6,$53,$67,$F3,$EF
	.dc.b	$DF,$C3,$71,$E6,$E1,$FB,$2D,$3D,$47,$C4,$3E,$43,$A7,$19,$FC,$1E
	.dc.b	$0F,$B5,$1F,$03,$83,$F8,$17,$37,$E3,$5B,$19,$3F,$63,$1F,$91,$E0
	.dc.b	$FE,$11,$CB,$F9,$3E,$BF,$F2,$D5,$7F,$AE,$7F,$4A,$73,$7E,$5F,$AF
	.dc.b	$3D,$49,$74,$66,$FF,$CB,$11,$87,$E9,$E1,$C2,$39,$C7,$DA,$F0,$3F
	.dc.b	$CA,$4F,$F9,$CE,$1D,$FF,$EF,$59,$CC,$FF,$9C,$A7,$2F,$1F,$3F,$A5
	.dc.b	$6F,$CD,$ED,$7F,$AD,$ED,$F0,$7A,$1F,$AF,$EF,$F1,$78,$57,$2F,$3B
	.dc.b	$B7,$D3,$3F,$B5,$F2,$3D,$1E,$17,$ED,$FC,$7B,$5F,$8E,$6B,$F5,$3C
	.dc.b	$15,$7B,$FC,$4F,$56,$FE,$6A,$F4,$1F,$1B,$7C,$B4,$F8,$9B,$FA,$9F
	.dc.b	$CD,$F8,$DF,$27,$CE,$FE,$11,$E3,$FC,$BF,$36,$D7,$D0,$7F,$86,$56
	.dc.b	$E3,$E7,$79,$7F,$47,$CA,$F9,$9E,$67,$D3,$F3,$2D,$7D,$87,$E8,$56
	.dc.b	$E3,$CE,$7F,$A3,$BA,$3E,$8B,$F9,$2E,$8F,$AA,$FE,$1F,$6F,$5B,$B7
	.dc.b	$EA,$3F,$84,$8F,$B9,$E6,$FE,$0F,$D1,$B3,$F8,$8F,$4E,$BD,$DF,$E4
	.dc.b	$DE,$7E,$E3,$F1,$1E,$BF,$E2,$FD,$2F,$C3,$3F,$8D,$F4,$EC,$FE,$49
	.dc.b	$FC,$57,$AD,$B8,$FC,$87,$C4,$7E,$4F,$DF,$FC,$73,$63,$27,$E6,$A3
	.dc.b	$F2,$5F,$0F,$F9,$9F,$7F,$F2,$9E,$D7,$E6,$FE,$0E,$DF,$E8,$3E,$26
	.dc.b	$BF,$1D,$FC,$D7,$B9,$B8,$FD,$07,$C2,$7E,$8F,$DE,$FC,$F7,$C2,$59
	.dc.b	$FD,$23,$5F,$9D,$3F,$B0,$F8,$2F,$F3,$74,$7F,$81,$F9,$DB,$FA,$63
	.dc.b	$EA,$37,$D7,$FF,$01,$1F,$ED,$D7,$1C,$BF,$E7,$6A,$70,$7B,$6F,$F6
	.dc.b	$9F,$09,$DF,$E2,$3F,$64,$F8,$7E,$22,$CF,$FB,$57,$34,$7E,$DF,$E1
	.dc.b	$B8,$4F,$EF,$67,$D0,$7E,$EC,$AE,$4E,$4F,$EF,$FE,$17,$DC,$7D,$48
	.dc.b	$FA,$47,$D4,$3F,$25,$CD,$96,$9F,$C2,$F2,$F5,$3F,$95,$F8,$7F,$91
	.dc.b	$B7,$F0,$B6,$C7,$F1,$16,$BF,$9B,$F8,$8F,$9D,$48,$FE,$9F,$FA,$33
	.dc.b	$93,$E9,$5F,$F1,$59,$FA,$97,$B7,$F9,$AD,$6D,$F8,$D1,$B7,$E4,$4E
	.dc.b	$F8,$BE,$AD,$3B,$FE,$F7,$26,$79,$3F,$D5,$73,$9C,$3E,$84,$E5,$FF
	.dc.b	$24,$43,$BF,$8C,$7A,$2F,$F1,$1C,$DF,$CA,$F9,$7F,$15,$CB,$FC,$B7
	.dc.b	$96,$DF,$F3,$39,$AF,$EF,$63,$F2,$1C,$BD,$F5,$66,$E3,$FC,$A7,$76
	.dc.b	$61,$D9,$F9,$7E,$DC,$55,$CB,$E5,$CF,$07,$CC,$7B,$0A,$7C,$B1,$EC
	.dc.b	$3A,$5C,$5D,$F8,$BD,$F5,$F4,$38,$BD,$EF,$D8,$ED,$FD,$47,$77,$97
	.dc.b	$BB,$F5,$5D,$DE,$79,$B5,$F7,$A7,$6F,$E0,$3C,$7F,$D8,$78,$BE,$DF
	.dc.b	$1B,$F6,$5E,$3F,$BE,$6D,$7E,$1A,$76,$FA,$8F,$C2,$6F,$E9,$DB,$E3
	.dc.b	$6B,$E8,$E5,$FC,$0E,$F7,$AF,$EF,$CF,$AB,$0F,$C5,$F8,$9F,$C0,$7C
	.dc.b	$DF,$9C,$FF,$03,$F3,$7E,$87,$8D,$6B,$F2,$F3,$B7,$F3,$3E,$97,$F0
	.dc.b	$9F,$23,$EB,$3F,$DE,$BF,$49,$FF,$43,$E7,$F8,$EF,$43,$0E,$AF,$F7
	.dc.b	$D7,$F8,$1F,$D2,$7F,$CA,$F7,$FD,$7A,$7F,$C6,$70,$F7,$7A,$7A,$7B
	.dc.b	$F7,$ED,$6A,$7C,$1E,$B7,$E1,$FD,$5F,$17,$A7,$6B,$F0,$07,$E0,$32
	.dc.b	$7A,$BF,$8D,$F5,$7D,$0E,$E7,$F0,$DD,$2A,$5F,$E4,$FD,$4F,$77,$B7
	.dc.b	$F9,$5F,$57,$E0,$7B,$B9,$C7,$BE,$3C,$63,$51,$8C,$57,$EB,$8C,$70
	.dc.b	$7C,$3D,$FD,$B1,$FE,$34,$E1,$F8,$B3,$98,$E5,$C1,$FA,$5F,$67,$E8
	.dc.b	$7B,$96,$7B,$3F,$A9,$F6,$F3,$1F,$C4,$27,$EB,$7B,$9F,$47,$E1,$BE
	.dc.b	$C7,$C3,$7D,$3F,$86,$FB,$5F,$0D,$9C,$3D,$0E,$5F,$54,$FA,$5F,$8F
	.dc.b	$C1,$F6,$5F,$46,$FE,$AF,$13,$FA,$3F,$07,$AA,$7E,$C1,$EA,$3F,$91
	.dc.b	$FC,$17,$E2,$7E,$1B,$F8,$07,$C1,$FE,$2F,$E1,$BF,$81,$4D,$AF,$E5
	.dc.b	$33,$B7,$F0,$FF,$0D,$F9,$13,$F1,$85,$7F,$17,$F0,$D5,$F9,$91,$F9
	.dc.b	$78,$FC,$97,$3F,$E6,$7E,$22,$D7,$D7,$F8,$8F,$CB,$75,$7E,$75,$CC
	.dc.b	$7C,$09,$FC,$D1,$FA,$1E,$4B,$1F,$7D,$FC,$F4,$E6,$3F,$3B,$3F,$A7
	.dc.b	$E4,$FD,$17,$1F,$EA,$7C,$2B,$5F,$87,$E3,$FD,$37,$2F,$EB,$5C,$C7
	.dc.b	$D0,$9F,$1B,$FC,$13,$95,$8E,$23,$FA,$87,$67,$E6,$F3,$F5,$C7,$B4
	.dc.b	$FD,$7E,$DF,$05,$FD,$8F,$27,$F9,$D3,$97,$FC,$DB,$1F,$93,$8C,$71
	.dc.b	$FE,$E7,$96,$CF,$03,$F4,$35,$C3,$9B,$F0,$3D,$FC,$FB,$FF,$80,$3D
	.dc.b	$7B,$FF,$81,$78,$17,$0F,$E0,$33,$B7,$CA,$7E,$A7,$7F,$63,$FB,$BD
	.dc.b	$BD,$6F,$E4,$9C,$EF,$FB,$99,$FE,$1D,$DD,$EB,$ED,$B7,$E6,$7F,$89
	.dc.b	$76,$E6,$3F,$88,$CF,$C1,$EE,$FE,$AD,$C8,$F0,$FE,$27,$8D,$87,$E3
	.dc.b	$78,$D7,$0F,$82,$FF,$1D,$F1,$B7,$1F,$EC,$EB,$7B,$7F,$37,$C7,$EB
	.dc.b	$ED,$5A,$D3,$5F,$F5,$AD,$70,$7E,$6F,$91,$FC,$A7,$C8,$FA,$DE,$8D
	.dc.b	$9F,$A3,$E8,$D7,$F8,$DF,$5A,$3F,$98,$79,$5C,$27,$C9,$E9,$FD,$FF
	.dc.b	$43,$DE,$FE,$07,$D1,$B3,$F7,$1F,$87,$E9,$EE,$3D,$9C,$1E,$7F,$03
	.dc.b	$FC,$8B,$BB,$83,$1C,$5F,$E2,$19,$FE,$59,$18,$B4,$7F,$2F,$BF,$C8
	.dc.b	$70,$FD,$0F,$4B,$F2,$5E,$87,$D2,$F4,$AE,$1F,$2D,$AF,$C6,$9F,$CC
	.dc.b	$7A,$5F,$63,$DF,$FC,$D7,$B7,$6B,$F2,$AD,$7E,$2B,$F7,$BD,$DF,$D0
	.dc.b	$7B,$9F,$80,$D3,$CD,$AF,$A2,$FF,$3B,$D5,$F6,$1A,$7C,$23,$EA,$9F
	.dc.b	$7F,$DB,$FD,$57,$B5,$F8,$AF,$6E,$CF,$E0,$DA,$FD,$E9,$FD,$87,$B7
	.dc.b	$F9,$0F,$81,$FD,$97,$C1,$DA,$FD,$6B,$F9,$4F,$83,$CC,$7E,$36,$7F
	.dc.b	$DD,$7F,$01,$FE,$7C,$D3,$FC,$3D,$78,$F8,$2F,$CD,$DD,$3E,$A7,$4F
	.dc.b	$D5,$F8,$1F,$F2,$74,$7F,$62,$C1,$5C,$23,$FA,$EC,$7F,$9D,$9F,$CE
	.dc.b	$7C,$0F,$82,$BF,$4E,$F8,$BE,$1B,$F5,$3F,$09,$6B,$67,$2A,$FD,$69
	.dc.b	$E8,$F8,$6B,$1D,$BE,$1B,$F6,$37,$EC,$73,$1E,$59,$FF,$25,$57,$A2
	.dc.b	$3F,$C4,$B8,$FA,$1F,$D8,$1E,$F3,$F1,$F9,$FB,$F4,$C6,$BF,$98,$9F
	.dc.b	$93,$F0,$DF,$C0,$31,$F9,$7B,$FF,$02,$CE,$D7,$EE,$E6,$BF,$B6,$8F
	.dc.b	$73,$D8,$FC,$EE,$6E,$4F,$AB,$C3,$E8,$E2,$FA,$FC,$3E,$A9,$B7,$F5
	.dc.b	$0E,$33,$C1,$C6,$9F,$71,$FA,$7C,$5F,$55,$FB,$0F,$C4,$DB,$DB,$BF
	.dc.b	$EC,$7C,$1F,$F1,$1D,$7F,$44,$E1,$FB,$4F,$F0,$EB,$FC,$4F,$1F,$F2
	.dc.b	$0E,$2B,$3F,$C6,$66,$BF,$C5,$23,$F1,$FD,$7F,$4B,$BF,$F2,$3D,$76
	.dc.b	$BF,$0E,$FD,$6E,$DC,$C7,$AB,$8D,$5F,$98,$EB,$FB,$7E,$6F,$E6,$BC
	.dc.b	$2B,$7F,$90,$7E,$FF,$99,$98,$E5,$3F,$A0,$EC,$FC,$1F,$81,$FA,$2E
	.dc.b	$EB,$5F,$98,$F1,$BF,$0F,$E0,$FE,$99,$CC,$7E,$7E,$7F,$17,$E5,$FE
	.dc.b	$AB,$C6,$FC,$6F,$95,$FA,$CF,$17,$37,$FA,$27,$2F,$E9,$27,$CE,$FE
	.dc.b	$A7,$6A,$FE,$51,$F4,$BE,$B3,$D2,$7E,$B3,$4F,$CD,$F9,$3F,$BA,$F1
	.dc.b	$EC,$FE,$C1,$FC,$FF,$93,$98,$F9,$53,$FC,$03,$CD,$FD,$1F,$A3,$F7
	.dc.b	$BD,$37,$AF,$FE,$74,$BF,$C9,$FA,$39,$8F,$F1,$C7,$89,$FE,$FC,$EA
	.dc.b	$61,$9F,$47,$F5,$B7,$BF,$13,$D7,$8F,$E2,$78,$2B,$FE,$AC,$F0,$FF
	.dc.b	$A0,$3F,$77,$AF,$EB,$3F,$A3,$CF,$FD,$61,$D7,$F5,$E3,$F0,$5E,$77
	.dc.b	$8F,$D0,$FD,$DB,$63,$F0,$DE,$85,$79,$6E,$9F,$57,$AD,$FC,$0B,$E1
	.dc.b	$BD,$9E,$C5,$BF,$F3,$9F,$1F,$F9,$C1,$FC,$87,$B7,$F0,$7D,$DD,$47
	.dc.b	$FA,$02,$7F,$D0,$AF,$EB,$BE,$1D,$F0,$FF,$1F,$E1,$BF,$DB,$10,$F8
	.dc.b	$7C,$56,$EA,$BE,$F9,$D3,$E2,$F6,$FF,$8B,$FB,$FF,$47,$DD,$FE,$33
	.dc.b	$95,$AF,$D2,$4D,$7D,$F1,$F5,$FE,$17,$C3,$F0,$FF,$67,$E1,$AD,$7B
	.dc.b	$9F,$2F,$C3,$E6,$3D,$53,$F7,$FE,$17,$D3,$AF,$9B,$A9,$DD,$FD,$D7
	.dc.b	$7F,$3E,$1F,$0F,$87,$6F,$CA,$7E,$0F,$0F,$E2,$78,$7E,$1F,$C3,$7E
	.dc.b	$2F,$E1,$AD,$FD,$07,$E3,$FC,$3E,$63,$E6,$4F,$E4,$74,$F9,$7A,$FE
	.dc.b	$4F,$1B,$5F,$63,$E1,$BE,$7E,$3F,$97,$73,$1F,$5E,$7C,$AF,$9C,$FB
	.dc.b	$0F,$B7,$5F,$66,$16,$7F,$C1,$FA,$FB,$9F,$F7,$5F,$1B,$5F,$BB,$8F
	.dc.b	$E9,$7E,$1E,$D7,$E2,$9F,$C0,$99,$7F,$0D,$1F,$AB,$E1,$FC,$2D,$7E
	.dc.b	$B7,$8B,$F0,$FC,$D6,$FF,$19,$35,$FC,$34,$7E,$CF,$81,$1F,$CF,$BE
	.dc.b	$27,$FC,$A0,$FE,$6B,$8B,$F2,$3B,$FE,$F7,$E2,$7F,$CD,$CF,$F8,$B6
	.dc.b	$B7,$5F,$ED,$A9,$CB,$F0,$27,$F8,$05,$7E,$6B,$9F,$97,$55,$AF,$F6
	.dc.b	$CD,$A3,$F3,$FD,$1A,$9F,$D2,$57,$87,$A7,$F4,$7D,$BE,$3E,$3F,$D2
	.dc.b	$F6,$79,$7C,$8C,$E1,$D1,$CB,$DA,$7D,$EF,$F0,$5D,$BD,$2F,$D4,$D7
	.dc.b	$DE,$F9,$3B,$B7,$DB,$F9,$88,$FF,$24,$D7,$07,$FD,$3F,$BF,$C0,$97
	.dc.b	$46,$E7,$F7,$D8,$7E,$EB,$BB,$E5,$78,$F6,$7E,$0B,$FB,$EE,$FC,$C7
	.dc.b	$F0,$39,$FA,$1E,$27,$8B,$97,$E8,$7A,$1F,$4F,$C6,$C5,$F9,$23,$EB
	.dc.b	$78,$78,$D5,$BB,$CE,$CD,$FD,$FB,$E4,$FE,$07,$CC,$FA,$EF,$F1,$3C
	.dc.b	$BF,$01,$C5,$CF,$0F,$81,$2F,$31,$C7,$F0,$78,$FE,$04,$FC,$B3,$F8
	.dc.b	$E7,$99,$F8,$9F,$27,$F8,$DF,$95,$67,$27,$E0,$A3,$2F,$E1,$3D,$2F
	.dc.b	$07,$A1,$F9,$1F,$4B,$C5,$E9,$FE,$4F,$D1,$CD,$FE,$29,$CB,$F8,$C9
	.dc.b	$F4,$7C,$3F,$E6,$7D,$3F,$57,$C2,$DA,$FD,$04,$D7,$F0,$91,$F9,$FF
	.dc.b	$6B,$E0,$7A,$FF,$A1,$F6,$ED,$7E,$61,$F8,$9E,$C6,$63,$F5,$53,$FA
	.dc.b	$7F,$6B,$E4,$7B,$BF,$A9,$F6,$FE,$57,$BB,$6F,$F6,$13,$5F,$CA,$47
	.dc.b	$EB,$FE,$13,$D3,$1F,$A6,$DB,$D2,$D3,$D2,$53,$D2,$7E,$B3,$4F,$B1
	.dc.b	$EF,$7E,$E7,$E1,$3E,$D7,$BF,$6B,$F8,$04,$D7,$F4,$91,$FB,$FF,$84
	.dc.b	$FC,$07,$C3,$FF,$7D,$F8,$AF,$F3,$F7,$80,$F8,$8C,$61,$A3,$F5,$A7
	.dc.b	$F0,$CF,$05,$5C,$27,$97,$C2,$FB,$2B,$F8,$46,$DE,$C6,$9E,$C2,$9E
	.dc.b	$C3,$F8,$76,$DF,$93,$F8,$8F,$27,$C3,$59,$F0,$9F,$95,$C5,$F0,$BF
	.dc.b	$99,$F8,$7F,$53,$B9,$F3,$E9,$E6,$BF,$CF,$67,$EF,$F8,$9F,$D0,$45
	.dc.b	$AF,$E4,$D3,$5F,$E1,$51,$F1,$77,$FD,$87,$C5,$FF,$93,$A7,$C7,$87
	.dc.b	$C2,$DF,$3B,$7C,$BE,$ED,$F6,$DB,$FF,$37,$E1,$C5,$F2,$EF,$F3,$DF
	.dc.b	$E6,$DB,$7D,$3E,$EF,$DA,$36,$3F,$9D,$6F,$5F,$E3,$D1,$F6,$78,$7C
	.dc.b	$BD,$1F,$6F,$8A,$D7,$9B,$14,$7A,$BA,$35,$3E,$4A,$FC,$17,$57,$AA
	.dc.b	$38,$70,$57,$CF,$BF,$B1,$E8,$E5,$FE,$63,$C3,$F0,$DF,$7F,$3F,$E3
	.dc.b	$7A,$BF,$87,$47,$E3,$FA,$BE,$5C,$D8,$FE,$AC,$6C,$F8,$FE,$07,$E5
	.dc.b	$7A,$FE,$8F,$55,$9C,$9E,$D8,$CB,$EF,$F0,$BC,$8F,$E5,$F8,$7E,$73
	.dc.b	$FE,$49,$E6,$FA,$4F,$A2,$1A,$FB,$63,$DE,$7D,$8E,$8F,$D3,$6F,$F8
	.dc.b	$2E,$EF,$D4,$78,$5F,$84,$EF,$B7,$F7,$A6,$BD,$E3,$F5,$DE,$07,$E2
	.dc.b	$BC,$3F,$D8,$78,$5F,$8C,$F0,$FF,$65,$E6,$67,$0F,$B0,$E5,$FB,$53
	.dc.b	$F9,$2F,$5B,$F7,$1E,$75,$9F,$BC,$FE,$5B,$D7,$CC,$7E,$42,$7F,$7D
	.dc.b	$E7,$FE,$6B,$C9,$FF,$3E,$7C,$67,$F9,$EF,$D1,$73,$27,$3E,$6F,$E8
	.dc.b	$3D,$4B,$97,$E6,$A7,$8F,$A5,$FA,$4F,$53,$9F,$A1,$6B,$F4,$B9,$A3
	.dc.b	$F5,$3E,$A6,$A7,$F4,$75,$E2,$F6,$BD,$51,$FA,$4E,$1A,$FD,$57,$D8
	.dc.b	$FB,$8F,$61,$F9,$DD,$FC,$CF,$83,$DA,$F7,$3E,$2F,$5F,$2F,$F0,$4C
	.dc.b	$FF,$CF,$4F,$E0,$F5,$F8,$75,$FB,$EF,$7B,$E2,$FB,$3F,$C0,$3D,$CB
	.dc.b	$7F,$BA,$9A,$FE,$2E,$3E,$5F,$C0,$FA,$23,$F2,$1A,$57,$F8,$C3,$EA
	.dc.b	$7D,$C7,$A8,$FE,$07,$A7,$F0,$EF,$73,$EA,$FC,$0F,$A2,$7F,$5B,$AA
	.dc.b	$35,$1F,$39,$D3,$FC,$D6,$1E,$FF,$F8,$1F,$8D,$F0,$5F,$80,$F8,$1F
	.dc.b	$F1,$07,$54,$9C,$E1,$F6,$27,$3C,$FF,$E4,$AA,$FB,$5F,$05,$F8,$8F
	.dc.b	$89,$CD,$A7,$EF,$63,$D6,$F8,$1C,$71,$63,$F9,$0F,$89,$E3,$7F,$85
	.dc.b	$C7,$F2,$7F,$11,$8B,$DE,$5F,$8A,$BC,$56,$B1,$EB,$3E,$1A,$7C,$2F
	.dc.b	$E7,$3E,$27,$AF,$80,$F8,$6C,$61,$C3,$4F,$CA,$5F,$E8,$BE,$23,$75
	.dc.b	$AC,$BF,$D9,$D9,$7F,$9F,$5F,$99,$F1,$3E,$5A,$FD,$4F,$C4,$F9,$F4
	.dc.b	$FD,$5C,$E6,$BF,$43,$3E,$BE,$0F,$81,$A7,$87,$16,$DE,$BB,$EF,$5A
	.dc.b	$7C,$31,$9D,$7E,$B6,$3F,$B8,$E1,$F8,$D9,$5C,$BA,$4F,$EF,$38,$B3
	.dc.b	$8F,$EC,$AF,$E6,$72,$FF,$00,$E0,$B3,$FD,$6B,$C0,$F8,$D8,$F9,$79
	.dc.b	$74,$F9,$EF,$FB,$8D,$7F,$91,$6B,$6F,$A0,$FA,$D3,$ED,$D7,$D0,$FE
	.dc.b	$07,$6F,$66,$DF,$8F,$9C,$D7,$D9,$E4,$EB,$ED,$31,$FF,$44,$63,$F1
	.dc.b	$F9,$B3,$0E,$CF,$C3,$5E,$3B,$7F,$11,$5A,$9F,$C5,$72,$70,$F1,$BF
	.dc.b	$19,$87,$2F,$1E,$E9,$F0,$67,$6F,$C3,$8F,$C9,$78,$3E,$0E,$DB,$3F
	.dc.b	$1F,$C6,$AF,$F8,$3B,$04,$FE,$67,$C4,$F3,$D7,$E6,$FC,$3F,$4F,$8B
	.dc.b	$70,$FA,$1E,$35,$79,$47,$E8,$7C,$5F,$7C,$FE,$9B,$E3,$5E,$4F,$C2
	.dc.b	$F2,$30,$DE,$6F,$CE,$4B,$CA,$F8,$FE,$66,$B3,$FB,$D8,$73,$F2,$BE
	.dc.b	$5F,$99,$D7,$CB,$B9,$72,$9F,$9F,$EC,$67,$0F,$C3,$5F,$ED,$6B,$E9
	.dc.b	$FB,$1F,$B7,$F3,$BE,$AF,$A9,$9B,$FC,$64,$E5,$FC,$74,$7A,$FD,$8F
	.dc.b	$B7,$F0,$9E,$DF,$66,$E3,$EA,$7E,$FF,$AB,$BC,$3F,$2D,$5F,$0F,$E0
	.dc.b	$3F,$07,$EB,$7C,$5F,$7B,$F0,$BE,$FD,$BF,$CE,$4E,$DF,$CF,$47,$CB
	.dc.b	$F6,$7F,$17,$F0,$16,$7F,$45,$EF,$57,$EE,$47,$D1,$F6,$FF,$23,$3F
	.dc.b	$4F,$E1,$3F,$2F,$F1,$DF,$EE,$EE,$0F,$A3,$18,$C3,$47,$F9,$1A,$5F
	.dc.b	$11,$F6,$FE,$13,$57,$F6,$37,$F5,$BF,$6B,$7F,$5B,$70,$F5,$BB,$7D
	.dc.b	$6F,$E0,$3E,$13,$C7,$F1,$1F,$87,$F8,$5B,$3F,$84,$7C,$FF,$0D,$F8
	.dc.b	$BA,$CD,$FF,$39,$AF,$E7,$DE,$13,$0F,$6F,$C3,$FF,$B3,$E3,$FC,$DF
	.dc.b	$5E,$B8,$C6,$1A,$8F,$C8,$D7,$C4,$F8,$8E,$15,$F1,$BE,$1A,$D7,$C3
	.dc.b	$6B,$FE,$76,$E0,$FF,$1A,$FB,$3F,$D5,$32,$61,$F3,$BE,$1F,$C5,$87
	.dc.b	$D0,$E3,$B8,$FC,$D6,$BC,$DF,$44,$FD,$6E,$2F,$F7,$47,$C7,$BA,$BE
	.dc.b	$CF,$26,$3A,$BE,$DF,$2D,$CB,$EC,$3C,$2B,$71,$FD,$26,$DD,$3F,$82
	.dc.b	$E5,$EA,$FE,$13,$9A,$D7,$E0,$5F,$0F,$46,$E3,$F6,$E7,$E2,$BB,$BE
	.dc.b	$77,$4F,$E3,$3B,$AC,$FE,$25,$FA,$4E,$CF,$17,$85,$F9,$2E,$DF,$6F
	.dc.b	$87,$F9,$4E,$EB,$5F,$91,$7E,$0D,$6E,$3D,$5E,$17,$E6,$BC,$9F,$8A
	.dc.b	$D8,$C9,$F9,$98,$F9,$3E,$1F,$E8,$3C,$9F,$97,$E1,$FE,$8B,$C6,$F9
	.dc.b	$BE,$2F,$E9,$3C,$7B,$8F,$E8,$5D,$BF,$A3,$9F,$A7,$E8,$7F,$40,$F9
	.dc.b	$0F,$F2,$97,$47,$F2,$AA,$FF,$56,$C3,$C9,$DC,$61,$FB,$7E,$87,$1F
	.dc.b	$2B,$EE,$F9,$FC,$FD,$6B,$7C,$9A,$FF,$B1,$30,$FF,$2A,$3F,$6F,$CF
	.dc.b	$F1,$7A,$7F,$86,$F3,$FC,$9E,$A5,$AF,$1B,$5E,$4F,$E3,$3D,$1F,$57
	.dc.b	$A9,$F8,$EF,$67,$D9,$EA,$FE,$43,$D8,$B8,$7E,$26,$6B,$FB,$13,$FC
	.dc.b	$55,$C5,$FD,$75,$F8,$3E,$B1,$FF,$0C,$7C,$8F,$F8,$4D,$7B,$8E,$0F
	.dc.b	$F2,$24,$6B,$E1,$FE,$67,$C0,$F1,$8E,$DE,$B3,$0E,$58,$71,$AD,$4F
	.dc.b	$F9,$12,$57,$F5,$7E,$07,$F5,$3E,$EF,$D7,$F8,$1F,$D5,$FB,$D7,$3F
	.dc.b	$D2,$4D,$7B,$EC,$9F,$BD,$F0,$3D,$BA,$FF,$C9,$99,$FF,$94,$2F,$EA
	.dc.b	$7C,$0E,$5F,$F2,$64,$3E,$13,$F0,$D3,$9D,$BD,$5F,$09,$8B,$F8,$B5
	.dc.b	$F8,$BF,$83,$D6,$1E,$EE,$3F,$8B,$5F,$0A,$3D,$AF,$E3,$B8,$B7,$F0
	.dc.b	$FF,$22,$3E,$51,$F9,$2E,$2E,$7F,$0B,$F9,$79,$FE,$17,$F1,$1F,$99
	.dc.b	$E5,$CD,$FE,$61,$CB,$F9,$A9,$F1,$EF,$F9,$FF,$87,$F2,$EB,$FA,$1F
	.dc.b	$88,$CD,$FE,$81,$CB,$FA,$29,$F2,$BE,$33,$CA,$7E,$6F,$4C,$B4,$C9
	.dc.b	$FE,$F4,$DB,$E0,$F0,$FD,$DA,$FD,$87,$0F,$C6,$C3,$F6,$5C,$3F,$22
	.dc.b	$6E,$1F,$90,$9D,$BF,$92,$E0,$FD,$C7,$07,$CE,$CE,$CE,$4F,$DC,$C6
	.dc.b	$5F,$C9,$1F,$9B,$E8,$7C,$97,$F9,$D6,$3F,$3B,$C9,$FE,$42,$87,$5F
	.dc.b	$DA,$8C,$3F,$73,$C7,$B5,$C0,$EA,$7E,$92,$BF,$54,$FD,$FE,$BE,$9D
	.dc.b	$7F,$84,$7B,$75,$7E,$1B,$AA,$D7,$EB,$A7,$6F,$EC,$3C,$1F,$27,$91
	.dc.b	$F8,$CF,$07,$CD,$E4,$7E,$3A,$6D,$7E,$DA,$76,$F9,$4F,$83,$BF,$C0
	.dc.b	$DB,$D9,$1E,$88,$C3,$FB,$ED,$BD,$AF,$E5,$F7,$F8,$9E,$0F,$E7,$3C
	.dc.b	$2F,$8D,$E4,$7E,$7B,$C3,$F9,$1E,$3E,$70,$E9,$C8,$8D,$46,$2B,$F5
	.dc.b	$58,$3C,$9F,$A1,$E3,$E1,$B5,$FE,$A7,$97,$95,$88,$FE,$17,$1C,$5E
	.dc.b	$FE,$47,$B1,$F7,$E5,$F6,$9F,$84,$7F,$13,$DF,$DE,$FE,$CF,$7F,$7F
	.dc.b	$81,$FC,$47,$83,$DF,$3F,$C8,$3C,$AF,$C2,$7A,$3F,$BC,$F2,$EC,$E4
	.dc.b	$F7,$C6,$5F,$83,$E9,$FB,$BD,$4F,$C6,$7A,$76,$7E,$2B,$F0,$BD,$4C
	.dc.b	$C7,$BE,$7F,$25,$E9,$FC,$6F,$82,$FC,$A7,$A7,$6B,$E6,$FA,$7F,$2B
	.dc.b	$E0,$7F,$30,$E6,$3E,$0C,$FC,$EF,$5B,$F3,$9E,$F7,$D0,$F5,$ED,$7C
	.dc.b	$D9,$AF,$CC,$8F,$D1,$7C,$07,$D6,$F7,$FF,$49,$EF,$FD,$8F,$7F,$F4
	.dc.b	$DE,$FE,$70,$FB,$EE,$5F,$C0,$CF,$8D,$FA,$1D,$6D,$31,$1F,$B3,$86
	.dc.b	$BE,$B8,$F6,$9F,$81,$DF,$E0,$BF,$83,$F7,$FF,$D6,$FF,$0F,$FB,$CE
	.dc.b	$B7,$5F,$F5,$77,$FD,$6D,$3A,$F8,$0F,$C7,$FC,$05,$C3,$91,$FA,$1C
	.dc.b	$39,$61,$C7,$C1,$7C,$27,$6F,$88,$FC,$BF,$C2,$F8,$3E,$27,$F3,$33
	.dc.b	$71,$FC,$B4,$ED,$F4,$1F,$A2,$E1,$F7,$BF,$C1,$78,$3D,$CF,$E5,$9C
	.dc.b	$F8,$9F,$C5,$27,$F4,$BF,$0D,$EE,$D7,$F4,$FF,$0D,$70,$FD,$24,$D7
	.dc.b	$F4,$71,$F1,$3E,$1F,$F5,$BA,$FC,$6D,$2D,$7B,$DF,$D8,$EB,$98,$FD
	.dc.b	$74,$FC,$CD,$7F,$AF,$7C,$9F,$D7,$9F,$8D,$CB,$8D,$FE,$8F,$7E,$B7
	.dc.b	$B5,$F4,$78,$BE,$96,$1F,$72,$38,$9F,$BB,$E0,$FB,$3C,$1F,$C0,$9B
	.dc.b	$1F,$BF,$DE,$BF,$BE,$8F,$BF,$C1,$E4,$E8,$FC,$0F,$5F,$9B,$AB,$F0
	.dc.b	$7D,$79,$C3,$EE,$39,$7E,$F4,$F9,$DF,$19,$FE,$14,$CF,$CC,$FF,$26
	.dc.b	$C5,$B6,$E3,$0F,$C1,$78,$F5,$7E,$41,$3F,$D0,$1E,$5D,$5F,$C7,$9F
	.dc.b	$CB,$BF,$C8,$3A,$AC,$7F,$1D,$3F,$37,$D5,$F4,$3C,$3F,$CE,$F8,$5F
	.dc.b	$4B,$C2,$B5,$F3,$4E,$FA,$2B,$31,$EB,$9F,$D2,$F8,$7E,$98,$FE,$41
	.dc.b	$AB,$A3,$D5,$1E,$C3,$A6,$AC,$7F,$01,$19,$3E,$37,$47,$EC,$2B,$F0
	.dc.b	$BE,$1F,$EC,$BC,$3F,$C3,$F8,$5F,$B4,$F0,$F3,$87,$EB,$5C,$BF,$AF
	.dc.b	$9F,$C6,$F9,$9F,$BA,$F3,$2C,$FE,$CD,$FC,$8F,$97,$98,$F2,$E8,$AF
	.dc.b	$E0,$3E,$67,$E5,$BD,$1F,$EE,$35,$FE,$54,$F9,$4F,$A8,$FD,$DF,$23
	.dc.b	$3E,$42,$7F,$CF,$93,$D2,$B3,$18,$8E,$07,$F6,$57,$9F,$17,$DF,$3E
	.dc.b	$37,$97,$C3,$CD,$FD,$3F,$99,$CB,$CD,$FD,$4C,$DC,$3D,$33,$B7,$D7
	.dc.b	$EB,$78,$3D,$5F,$D7,$FA,$D6,$72,$78,$63,$2F,$8F,$D6,$F8,$0F,$97
	.dc.b	$CB,$FD,$BF,$07,$AB,$D2,$CC,$FF,$80,$37,$F6,$6F,$E9,$E4,$FC,$B4
	.dc.b	$FF,$88,$E7,$DB,$87,$EE,$3D,$3F,$87,$EB,$7F,$02,$7E,$2F,$AD,$FC
	.dc.b	$13,$D6,$B8,$7D,$49,$DB,$F8,$AF,$95,$7B,$9F,$33,$DE,$FF,$34,$46
	.dc.b	$27,$E6,$47,$F0,$EF,$88,$D4,$71,$AF,$A7,$EF,$FA,$A3,$FB,$C6,$BE
	.dc.b	$67,$F4,$3C,$3E,$F7,$ED,$1F,$5F,$87,$E1,$BF,$14,$FC,$4F,$BF,$F7
	.dc.b	$BD,$DF,$E3,$AD,$8F,$C6,$FB,$F5,$FE,$09,$1F,$84,$F7,$FD,$3F,$03
	.dc.b	$F8,$6F,$86,$F5,$FC,$15,$BF,$81,$35,$FF,$4F,$F0,$A7,$F1,$BF,$0D
	.dc.b	$F0,$B8,$BF,$1F,$F0,$F6,$FF,$1C,$FC,$6E,$1C,$C7,$CB,$9F,$CA,$FC
	.dc.b	$47,$CA,$F8,$7F,$CB,$E7,$6B,$F2,$CF,$CE,$F8,$7C,$C7,$D3,$9F,$CE
	.dc.b	$E7,$F4,$B5,$FC,$FC,$7D,$4D,$3F,$43,$79,$BF,$D0,$39,$7F,$45,$3E
	.dc.b	$77,$F1,$9A,$57,$EA,$BE,$97,$D6,$7A,$4F,$D5,$65,$F8,$28,$FD,$7F
	.dc.b	$2F,$E1,$23,$F6,$BF,$2D,$FE,$73,$AF,$93,$58,$E7,$D4,$7E,$CE,$FF
	.dc.b	$17,$D7,$C3,$C7,$B3,$FE,$5B,$C3,$FC,$80,$FD,$2E,$CD,$3F,$95,$AE
	.dc.b	$FD,$3E,$98,$FD,$16,$55,$FE,$14,$FA,$DF,$79,$EB,$3F,$59,$9F,$E7
	.dc.b	$3A,$3D,$3C,$FF,$9E,$E7,$F5,$F4,$66,$FC,$8E,$5F,$34,$FE,$93,$C7
	.dc.b	$F8,$3E,$27,$F0,$DE,$FF,$F0,$D7,$5A,$73,$1F,$AD,$F1,$FE,$3F,$8B
	.dc.b	$BC,$7F,$A8,$DE,$DF,$85,$86,$7C,$37,$CB,$BD,$8F,$9F,$E1,$E6,$31
	.dc.b	$7F,$26,$31,$EB,$7D,$4F,$17,$57,$C3,$5F,$DF,$79,$1F,$63,$CD,$ED
	.dc.b	$AB,$C9,$C5,$6B,$3F,$A9,$E7,$6A,$73,$1F,$DA,$27,$FC,$71,$AF,$F9
	.dc.b	$EE,$3F,$D2,$FD,$8F,$33,$CD,$E7,$FE,$1B,$CD,$F4,$79,$FF,$88,$9C
	.dc.b	$DF,$E0,$67,$CE,$FA,$F1,$FE,$BB,$8F,$BA,$3A,$F1,$7B,$9F,$C4,$70
	.dc.b	$F0,$73,$5F,$C6,$CF,$CA,$FB,$1E,$EF,$36,$C6,$67,$F2,$1E,$C6,$5F
	.dc.b	$C9,$57,$CD,$F6,$FF,$3B,$F0,$96,$7F,$2D,$EC,$FD,$1F,$6F,$38,$F9
	.dc.b	$AF,$F4,$73,$E3,$AF,$F4,$C7,$13,$5C,$47,$FB,$6E,$31,$7E,$EE,$9F
	.dc.b	$B0,$FC,$0E,$9F,$B1,$5F,$3C,$FF,$0B,$F0,$3D,$6F,$D8,$FB,$3F,$83
	.dc.b	$F6,$2E,$1F,$22,$7F,$6B,$EC,$E7,$0F,$D7,$5F,$E2,$7E,$13,$F7,$3F
	.dc.b	$05,$67,$F6,$4F,$E3,$7E,$0F,$F7,$BF,$07,$99,$FB,$B3,$F9,$1F,$86
	.dc.b	$FE,$EB,$F3,$09,$FC,$A7,$C4,$63,$E0,$F5,$1F,$ED,$79,$D6,$DF,$72
	.dc.b	$B4,$FD,$EF,$85,$E7,$87,$E7,$5B,$1F,$82,$F8,$6E,$F5,$9C,$3A,$5F
	.dc.b	$E8,$FE,$27,$C7,$F0,$D6,$7C,$0F,$E9,$FE,$27,$CF,$F1,$39,$9F,$F7
	.dc.b	$6E,$8A,$F3,$C6,$0F,$59,$E9,$D7,$2F,$FB,$4E,$7D,$6F,$D5,$F3,$3F
	.dc.b	$6F,$9F,$B2,$31,$CF,$FE,$39,$7E,$16,$7F,$B9,$F8,$9F,$93,$A7,$EE
	.dc.b	$FE,$27,$E5,$CE,$6B,$E4,$4F,$EF,$DF,$9F,$35,$F9,$8F,$F0,$3C,$FE
	.dc.b	$9D,$E6,$7D,$B3,$E8,$7E,$B7,$C4,$B4,$F4,$E9,$FE,$25,$CD,$A7,$A6
	.dc.b	$31,$BF,$DE,$C3,$E1,$9F,$23,$3F,$E2,$5B,$7E,$0B,$9A,$CF,$DF,$7F
	.dc.b	$8B,$70,$7E,$1B,$9B,$33,$F4,$27,$F8,$F6,$0E,$7F,$C5,$F2,$E1,$FC
	.dc.b	$6F,$26,$BA,$33,$CD,$FE,$60,$BE,$18,$FE,$36,$B9,$F9,$DF,$94,$F0
	.dc.b	$7A,$F9,$B7,$0E,$13,$F9,$8F,$0F,$38,$7E,$56,$FC,$7D,$7F,$9C,$EF
	.dc.b	$B3,$F9,$9F,$03,$CF,$DD,$9C,$BC,$57,$FA,$29,$F5,$F8,$BF,$A4,$F0
	.dc.b	$2E,$1F,$A1,$F0,$7D,$FE,$36,$71,$F5,$5F,$EA,$A7,$D3,$5E,$39,$F7
	.dc.b	$6F,$E9,$8F,$C7,$70,$FF,$9A,$DF,$D6,$6F,$EA,$7F,$37,$BF,$CC,$F1
	.dc.b	$BF,$6B,$E6,$59,$FD,$97,$99,$F4,$3C,$8C,$E7,$F2,$EF,$F7,$73,$F5
	.dc.b	$3C,$8F,$F3,$8F,$AB,$FE,$F6,$AF,$A1,$E5,$E3,$CB,$D6,$5F,$C0,$49
	.dc.b	$E3,$2F,$84,$7A,$2B,$85,$FD,$28,$F7,$1E,$CE,$0F,$77,$87,$FE,$77
	.dc.b	$E1,$F7,$4F,$ED,$B8,$7D,$CF,$EC,$78,$7F,$13,$E6,$F9,$BD,$4F,$C5
	.dc.b	$F9,$DE,$8F,$53,$37,$E5,$3F,$43,$B7,$F7,$5D,$7F,$15,$5B,$7F,$19
	.dc.b	$EA,$FC,$0F,$4F,$F2,$B5,$F0,$BD,$2F,$CB,$CE,$6B,$F2,$53,$97,$FD
	.dc.b	$67,$F3,$2F,$86,$F9,$3E,$EF,$F7,$B8,$FF,$26,$67,$E3,$1C,$BF,$1E
	.dc.b	$75,$5F,$43,$DE,$E1,$F0,$DC,$B1,$FD,$07,$B3,$F5,$3E,$1F,$F5,$2F
	.dc.b	$D6,$F8,$8F,$D5,$FB,$99,$AD,$E9,$DF,$4F,$A9,$5E,$3F,$83,$FB,$BE
	.dc.b	$FD,$AD,$70,$23,$F6,$BF,$09,$F8,$2F,$80,$D5,$7C,$BC,$7E,$F5,$7A
	.dc.b	$23,$8F,$0F,$A1,$B9,$FC,$4E,$1F,$63,$F1,$B8,$3D,$93,$F9,$CF,$84
	.dc.b	$FC,$7F,$C4,$7F,$04,$6D,$7E,$FF,$E1,$7F,$27,$F1,$19,$BF,$3D,$FC
	.dc.b	$DF,$86,$FC,$B9,$F3,$FE,$1A,$E1,$F5,$23,$F3,$71,$9A,$F7,$C7,$A5
	.dc.b	$F6,$70,$B5,$F5,$46,$1D,$3F,$9E,$87,$3E,$A3,$13,$F7,$FE,$1F,$F5
	.dc.b	$3B,$D9,$F9,$BB,$FE,$0F,$E2,$33,$A7,$E2,$2F,$F5,$F3,$F8,$7E,$1F
	.dc.b	$D8,$C7,$E2,$78,$ED,$7E,$3A,$7F,$6B,$79,$C3,$EC,$DF,$E3,$FC,$0F
	.dc.b	$DC,$F3,$59,$FC,$17,$CD,$39,$FF,$29,$E0,$EB,$3F,$CC,$E1,$FE,$F0
	.dc.b	$AD,$47,$E7,$39,$31,$CF,$F9,$BE,$6D,$73,$FE,$76,$73,$5F,$A2,$9C
	.dc.b	$BF,$A4,$E3,$E9,$DD,$FA,$3E,$5E,$DC,$DF,$A5,$9C,$D7,$EA,$A7,$2F
	.dc.b	$93,$7F,$CA,$F0,$7A,$DF,$C9,$6B,$E6,$7F,$B9,$77,$38,$3D,$91,$8E
	.dc.b	$0F,$67,$0F,$EC,$FB,$BD,$DD,$D6,$BE,$2F,$77,$ED,$FB,$B3,$AF,$EF
	.dc.b	$2F,$E2,$4F,$EE,$E3,$E3,$79,$56,$FE,$6F,$93,$FB,$FF,$2B,$35,$FC
	.dc.b	$10,$F9,$91,$E9,$9F,$C5,$70,$57,$FB,$AF,$07,$AF,$C9,$5F,$F8,$B5
	.dc.b	$70,$FB,$6B,$E1,$CF,$C5,$7E,$1E,$FF,$C3,$FC,$4F,$B5,$EB,$FF,$11
	.dc.b	$F1,$6D,$FF,$16,$9F,$BD,$E5,$E7,$0F,$C2,$DF,$FA,$47,$E7,$3F,$C9
	.dc.b	$D1,$F8,$8F,$27,$1E,$A7,$E1,$5B,$1F,$8C,$F2,$B8,$7A,$59,$CF,$D0
	.dc.b	$7A,$70,$7B,$BF,$8D,$F5,$3B,$36,$77,$83,$85,$7C,$23,$F3,$3E,$AF
	.dc.b	$97,$DE,$E1,$1F,$E3,$19,$FF,$49,$3F,$D4,$B9,$1E,$A7,$AF,$D4,$FC
	.dc.b	$E3,$67,$F4,$7E,$AF,$BF,$D6,$CE,$9E,$2D,$3C,$3E,$A7,$E8,$FD,$DF
	.dc.b	$89,$EB,$7E,$97,$DD,$B9,$7E,$B2,$7E,$45,$67,$0E,$5B,$7A,$6B,$F2
	.dc.b	$FC,$5F,$01,$FD,$9B,$5F,$CD,$F0,$FB,$5C,$D7,$2F,$73,$F6,$9E,$DF
	.dc.b	$D5,$F6,$EC,$F6,$7F,$71,$EE,$7D,$9A,$CD,$FC,$5F,$85,$FE,$A0,$FF
	.dc.b	$50,$F9,$CF,$F2,$0E,$1F,$57,$E1,$F3,$EF,$25,$F0,$9F,$83,$9C,$C7
	.dc.b	$0B,$FA,$DB,$FF,$97,$5F,$E8,$91,$C3,$0F,$83,$D7,$F5,$B9,$7F,$37
	.dc.b	$5C,$23,$13,$F8,$57,$F8,$87,$07,$C1,$9F,$8F,$3F,$2B,$E0,$BC,$3F
	.dc.b	$03,$F9,$5F,$82,$B3,$F3,$9F,$2F,$C2,$FE,$66,$B3,$7F,$85,$F8,$5F
	.dc.b	$49,$F9,$D3,$D7,$F0,$B6,$3F,$B3,$6B,$FE,$83,$C7,$EC,$CE,$9F,$C2
	.dc.b	$7C,$EF,$F9,$42,$BD,$FF,$15,$FE,$F7,$87,$67,$C6,$8C,$3F,$23,$E2
	.dc.b	$B5,$DB,$9C,$3F,$0B,$5F,$33,$9F,$F6,$15,$F3,$B9,$6E,$5E,$49,$FD
	.dc.b	$A5,$66,$FF,$1B,$3F,$53,$8F,$F7,$0D,$8C,$BE,$CF,$13,$F7,$9C,$F9
	.dc.b	$8F,$F4,$C6,$BE,$B7,$FC,$73,$BB,$1F,$4C,$7C,$46,$BF,$7B,$4F,$5B
	.dc.b	$9C,$7F,$3F,$C9,$F8,$6E,$2F,$E1,$4F,$E2,$39,$2D,$7C,$E9,$FE,$1D
	.dc.b	$C9,$9B,$F9,$BE,$17,$E3,$BB,$FE,$63,$63,$37,$F5,$BC,$0F,$A1,$E1
	.dc.b	$66,$3D,$7C,$BF,$35,$F6,$F3,$5E,$AF,$C8,$8E,$2F,$AF,$3F,$80,$E0
	.dc.b	$F5,$BF,$73,$2F,$BF,$E0,$7E,$83,$BB,$EF,$36,$33,$3F,$88,$F0,$3F
	.dc.b	$05,$E0,$66,$3E,$6F,$8F,$FA,$8F,$23,$F0,$DE,$47,$EA,$BC,$8B,$5F
	.dc.b	$90,$9F,$C5,$56,$71,$FC,$7F,$B9,$FB,$0F,$1F,$F1,$DE,$E5,$9F,$C9
	.dc.b	$BF,$B4,$F2,$7F,$25,$59,$C7,$EE,$F9,$3F,$BC,$F9,$EF,$F3,$9F,$27
	.dc.b	$C2,$E4,$79,$19,$F5,$3F,$34,$EF,$4F,$CE,$61,$97,$F4,$3E,$AF,$3F
	.dc.b	$6B,$F4,$1E,$B7,$5F,$6E,$EB,$FE,$4F,$D7,$FC,$DC,$FE,$3A,$F5,$87
	.dc.b	$B2,$3F,$5F,$C3,$EC,$6C,$7B,$1F,$D9,$F0,$7B,$1C,$9F,$98,$F4,$7D
	.dc.b	$9E,$9F,$EC,$5F,$77,$A9,$67,$C5,$3F,$B5,$F5,$33,$87,$EE,$BE,$07
	.dc.b	$E2,$7B,$DF,$B9,$7E,$37,$BD,$70,$F5,$4F,$EF,$7E,$07,$37,$E9,$9F
	.dc.b	$E3,$FA,$FA,$5B,$1E,$97,$F9,$1E,$BE,$97,$27,$EC,$3D,$FF,$A9,$EF
	.dc.b	$7F,$09,$6D,$7F,$0E,$7E,$C7,$BF,$FC,$37,$DF,$CD,$FF,$65,$AF,$F0
	.dc.b	$27,$4B,$E0,$BE,$F7,$C2,$7E,$7A,$73,$18,$7F,$AE,$4E,$EB,$EB,$7C
	.dc.b	$2E,$DF,$EE,$10,$F8,$5E,$11,$DC,$CE,$1F,$8B,$F8,$5E,$1E,$02,$D8
	.dc.b	$7E,$3F,$E1,$73,$18,$C3,$F0,$97,$8C,$35,$97,$93,$E1,$7F,$2D,$F1
	.dc.b	$1C,$1F,$CC,$7C,$4D,$C3,$F1,$93,$FC,$BF,$E1,$B3,$7E,$C8,$FC,$F7
	.dc.b	$C4,$F8,$1F,$D0,$7C,$4D,$BF,$CA,$4F,$93,$0C,$DF,$C4,$CB,$D8,$FC
	.dc.b	$6D,$BC,$B7,$F0,$77,$F7,$46,$A3,$1B,$7F,$A9,$DF,$D0,$6F,$F0,$32
	.dc.b	$FD,$87,$0F,$C2,$D2,$E5,$F4,$9F,$DA,$70,$EF,$0F,$D4,$57,$C8,$E5
	.dc.b	$FD,$C7,$0F,$CA,$E4,$B8,$FD,$A7,$FB,$BF,$6F,$F9,$1E,$3F,$07,$5A
	.dc.b	$C3,$CB,$C1,$E7,$E0,$FE,$43,$BF,$D3,$9F,$F5,$D7,$82,$E0,$FA,$FC
	.dc.b	$3B,$CB,$F1,$B8,$FC,$2C,$78,$F2,$F8,$3B,$7E,$EB,$E2,$EB,$B9,$FF
	.dc.b	$B3,$F6,$57,$E0,$B3,$D6,$5F,$92,$9F,$4F,$91,$F8,$6F,$02,$CF,$E5
	.dc.b	$9F,$6F,$93,$F8,$AA,$CE,$3F,$4F,$E7,$FF,$CA,$AF,$FB,$4E,$1D,$BF
	.dc.b	$94,$97,$9B,$F1,$BB,$B0,$FC,$89,$D7,$9B,$9C,$FF,$4B,$E1,$7C,$CF
	.dc.b	$1B,$F3,$2F,$CE,$F2,$2E,$9E,$D9,$FC,$EF,$8B,$9B,$FD,$69,$F5,$3C
	.dc.b	$7F,$D0,$B4,$CD,$7C,$5F,$1F,$F4,$BD,$79,$8F,$66,$7E,$5A,$FF,$29
	.dc.b	$6F,$EA,$9F,$99,$BF,$A9,$F9,$DB,$7A,$9C,$D7,$EF,$BC,$9F,$C2,$F9
	.dc.b	$BF,$B2,$6C,$66,$7E,$B7,$9D,$FB,$6F,$2F,$31,$F3,$7D,$2F,$C6,$F9
	.dc.b	$BF,$2F,$D1,$B8,$79,$1F,$C8,$F9,$FF,$3E,$B3,$7E,$A3,$CB,$5E,$AE
	.dc.b	$7F,$9D,$1F,$65,$F7,$3B,$3E,$D6,$DF,$67,$CF,$FC,$EF,$B1,$F6,$FD
	.dc.b	$0B,$87,$C4,$7F,$43,$EB,$EF,$0F,$C8,$57,$E0,$7D,$BF,$D2,$FB,$16
	.dc.b	$7F,$28,$D7,$F1,$1E,$C7,$E1,$FD,$CD,$E5,$F4,$3E,$1B,$F5,$BE,$CF
	.dc.b	$E2,$DF,$D7,$FB,$37,$0F,$CF,$4F,$E3,$FE,$1B,$37,$F6,$3D,$DF,$F3
	.dc.b	$5F,$D0,$FF,$99,$FC,$E7,$72,$73,$8F,$E5,$DD,$FC,$1F,$E6,$7D,$FC
	.dc.b	$E3,$F8,$2F,$86,$E7,$F0,$5F,$9D,$F8,$7E,$BF,$07,$72,$10,$BF,$12
	.dc.b	$D9,$7F,$A5,$C3,$8E,$3E,$FB,$F5,$B5,$FD,$67,$0F,$BD,$FD,$77,$0F
	.dc.b	$BD,$FD,$F4,$7E,$BF,$E2,$3D,$FC,$76,$7C,$47,$8B,$2F,$F3,$63,$FB
	.dc.b	$99,$F8,$BC,$7F,$B7,$F8,$8E,$39,$7E,$83,$4F,$93,$F0,$FF,$BB,$DF
	.dc.b	$E5,$FC,$3D,$CB,$A8,$E0,$F7,$57,$F0,$1B,$DE,$1E,$BA,$F8,$94,$F4
	.dc.b	$9C,$1E,$D9,$CC,$38,$31,$1F,$2A,$37,$7F,$B4,$D7,$EF,$CF,$F1,$3C
	.dc.b	$BE,$FF,$0F,$F1,$1D,$2D,$FE,$E9,$FC,$1F,$17,$F1,$5A,$CD,$FF,$4D
	.dc.b	$FA,$1F,$1B,$EC,$D7,$EC,$70,$FD,$BE,$8F,$F2,$CD,$79,$6B,$5C,$7F
	.dc.b	$EC,$D9,$CE,$FF,$90,$BE,$1A,$7E,$12,$58,$76,$C3,$EA,$E3,$F8,$5E
	.dc.b	$3D,$C7,$FB,$C6,$5E,$57,$E6,$BB,$73,$87,$2F,$27,$F3,$BD,$98,$B4
	.dc.b	$63,$0F,$F7,$54,$72,$7C,$F7,$CB,$B3,$FD,$A7,$1F,$67,$7F,$D2,$3F
	.dc.b	$C8,$3A,$FF,$53,$D9,$B6,$C7,$F2,$43,$F5,$BD,$9E,$1E,$FF,$D7,$F8
	.dc.b	$B6,$7E,$2B,$5F,$7E,$08,$F4,$78,$3F,$B6,$F1,$FD,$5C,$3E,$FB,$F4
	.dc.b	$EF,$EB,$7E,$B3,$78,$7E,$9C,$EE,$7E,$D7,$85,$FC,$03,$CC,$F8,$9E
	.dc.b	$1D,$9F,$86,$D7,$67,$DA,$F1,$FE,$57,$9F,$FC,$23,$C7,$F9,$9E,$7F
	.dc.b	$F0,$AF,$22,$DF,$E0,$1D,$BF,$82,$9F,$4B,$F3,$8F,$49,$F8,$6E,$9F
	.dc.b	$AD,$B7,$A7,$B3,$F0,$9D,$5F,$56,$BE,$2F,$57,$DB,$F2,$FF,$5B,$1F
	.dc.b	$AD,$FA,$2F,$F7,$CB,$F9,$7F,$33,$3D,$E9,$CF,$2F,$E5,$A7,$31,$8C
	.dc.b	$BB,$E1,$8F,$43,$F1,$1E,$BE,$BD,$0F,$F2,$5E,$AF,$5B,$18,$7E,$2A
	.dc.b	$B9,$FA,$3A,$9C,$ED,$FA,$9F,$4B,$BF,$9B,$C9,$F0,$FA,$FD,$3D,$3C
	.dc.b	$E5,$F9,$49,$CB,$F9,$6F,$57,$CF,$EB,$F8,$9B,$1F,$9A,$F5,$FD,$7E
	.dc.b	$BE,$73,$FD,$B7,$1B,$0F,$F3,$1F,$D1,$FF,$9B,$67,$F5,$B2,$F8,$0F
	.dc.b	$D4,$CF,$F9,$B6,$78,$FC,$0E,$3E,$0F,$E3,$BA,$F8,$4F,$93,$7C,$3E
	.dc.b	$13,$E5,$FC,$15,$C3,$C2,$FE,$C7,$E2,$73,$1F,$BF,$DF,$F8,$06,$6B
	.dc.b	$FA,$7F,$09,$FB,$7F,$82,$FA,$BF,$09,$70,$F5,$BF,$BB,$F8,$4C,$C7
	.dc.b	$EE,$A7,$ED,$FC,$27,$AA,$BE,$EF,$C4,$FC,$06,$E7,$E1,$B7,$86,$D1
	.dc.b	$5F,$84,$F8,$9F,$8F,$8F,$0B,$FF,$37,$53,$1F,$C4,$FC,$4F,$CC,$BF
	.dc.b	$C5,$FC,$4D,$BF,$A3,$3F,$42,$B3,$9F,$87,$3F,$04,$FE,$4B,$97,$EA
	.dc.b	$E1,$6B,$59,$23,$F2,$FC,$9F,$6A,$F5,$97,$93,$9F,$F3,$7C,$5F,$7B
	.dc.b	$A3,$F4,$3F,$49,$FE,$71,$AF,$7E,$5F,$83,$7F,$0B,$5B,$7F,$9E,$4E
	.dc.b	$F5,$F8,$9C,$FC,$B8,$FF,$12,$DD,$BC,$BD,$1D,$B9,$73,$8F,$2B,$FC
	.dc.b	$7F,$47,$8B,$C2,$B3,$D9,$FC,$9F,$47,$9B,$C1,$CC,$FF,$40,$D5,$5F
	.dc.b	$98,$CF,$D7,$DD,$F9,$66,$D7,$C6,$F1,$6B,$F5,$A3,$E0,$F8,$1F,$9C
	.dc.b	$F1,$7E,$1F,$83,$6B,$DE,$FE,$83,$C4,$CC,$7D,$09,$F9,$3E,$17,$E9
	.dc.b	$3C,$4F,$97,$E7,$DA,$F8,$EF,$EA,$3C,$5C,$C7,$D8,$9F,$A3,$E7,$FF
	.dc.b	$54,$E4,$FE,$79,$3F,$3B,$D4,$FF,$6A,$C3,$D5,$FA,$FE,$76,$1F,$B3
	.dc.b	$3A,$F5,$F3,$C3,$F8,$9B,$FF,$46,$E1,$FE,$08,$E5,$79,$9D,$3D,$5F
	.dc.b	$C0,$F9,$D6,$BE,$EC,$F8,$3D,$4C,$E3,$CF,$D3,$FC,$3F,$B3,$E4,$F5
	.dc.b	$6C,$F7,$7F,$17,$EC,$7A,$2B,$38,$FF,$B4,$75,$7A,$BF,$90,$F6,$3D
	.dc.b	$BE,$BF,$E4,$BD,$8F,$7C,$E7,$3F,$54,$E5,$F6,$7C,$4F,$E6,$3E,$1F
	.dc.b	$E2,$FC,$3F,$E6,$BE,$0F,$E3,$CE,$6B,$E1,$4E,$5F,$89,$F0,$1F,$D0
	.dc.b	$23,$FC,$B9,$A7,$F7,$8D,$7F,$C1,$70,$F8,$0F,$A3,$F0,$1B,$9F,$A7
	.dc.b	$7C,$67,$EA,$FC,$3F,$3F,$88,$FA,$FF,$0F,$D7,$E2,$73,$87,$F8,$8F
	.dc.b	$2F,$F2,$CE,$9F,$56,$B4,$FB,$27,$F5,$BB,$FB,$1F,$D7,$EF,$EC,$73
	.dc.b	$3E,$C7,$2F,$DF,$F8,$7F,$67,$07,$E2,$1F,$76,$3F,$8A,$F8,$7C,$D6
	.dc.b	$EF,$C9,$3F,$8E,$D3,$E2,$7C,$45,$9F,$47,$C4,$FE,$4B,$6C,$E1,$F8
	.dc.b	$DB,$F9,$53,$E5,$AF,$DA,$6B,$E9,$7F,$6D,$AE,$5F,$C9,$46,$5F,$CA
	.dc.b	$47,$E7,$B1,$FA,$9B,$FE,$82,$EE,$3F,$9B,$9F,$B1,$BE,$70,$F9,$96
	.dc.b	$AF,$E5,$1F,$0F,$FE,$45,$9F,$F3,$3E,$AD,$F5,$1F,$CF,$2B,$FC,$99
	.dc.b	$5F,$E5,$38,$FC,$0C,$EB,$D8,$93,$CE,$7F,$D2,$76,$E0,$D6,$1C,$B5
	.dc.b	$FC,$66,$38,$BF,$E9,$38,$7F,$85,$5F,$1F,$3E,$72,$FF,$40,$78,$9F
	.dc.b	$93,$E8,$C5,$FF,$94,$AF,$F1,$17,$8A,$D6,$5F,$BD,$9F,$CD,$73,$7F
	.dc.b	$06,$6C,$7F,$01,$E6,$FC,$F7,$3E,$67,$F2,$55,$DB,$BB,$F4,$5E,$27
	.dc.b	$83,$BA,$E3,$F2,$A7,$F4,$DE,$2E,$70,$FC,$D5,$F8,$DF,$2D,$7D,$0C
	.dc.b	$FC,$B1,$F3,$4F,$83,$E3,$7E,$BF,$5E,$11,$FA,$ED,$FD,$FD,$FF,$B3
	.dc.b	$F1,$7E,$0F,$81,$72,$FB,$73,$FB,$7F,$0F,$38,$7E,$AE,$FE,$3F,$95
	.dc.b	$FB,$BF,$0E,$CF,$EB,$DF,$97,$E4,$7E,$FF,$C5,$CC,$FE,$FA,$7C,$6F
	.dc.b	$D4,$CB,$D4,$FE,$DF,$4F,$A7,$B7,$A6,$73,$3E,$97,$F1,$FE,$37,$F9
	.dc.b	$2B,$E9,$7F,$25,$1F,$92,$F2,$31,$E8,$FD,$CA,$D4,$FD,$E9,$CB,$F9
	.dc.b	$8F,$4B,$97,$9F,$F8,$27,$A7,$A1,$F8,$4F,$4B,$35,$BC,$FF,$D5,$D9
	.dc.b	$AB,$F1,$9F,$4D,$FE,$39,$9F,$F6,$3C,$FF,$B1,$EB,$FC,$19,$39,$9F
	.dc.b	$F7,$6C,$BD,$1E,$3F,$01,$EB,$F4,$B9,$FC,$17,$B6,$7F,$2B,$EA,$FB
	.dc.b	$FE,$03,$F2,$FE,$B5,$D3,$F6,$33,$F0,$FE,$03,$3A,$79,$AF,$F3,$BE
	.dc.b	$B7,$C7,$F7,$AC,$FA,$BD,$EF,$D0,$FB,$D9,$CB,$F7,$77,$F3,$67,$C7
	.dc.b	$5F,$03,$6A,$FF,$B0,$B8,$3D,$57,$EE,$E0,$F5,$3F,$03,$83,$D4,$FF
	.dc.b	$06,$F6,$FE,$C7,$BD,$FB,$06,$C7,$F0,$AF,$77,$EE,$7B,$D9,$CB,$E7
	.dc.b	$DF,$CE,$F8,$1F,$C0,$7C,$35,$9F,$A6,$FD,$2F,$83,$FC,$27,$C3,$66
	.dc.b	$7C,$D3,$E8,$7E,$AE,$F5,$FA,$7D,$1F,$49,$F7,$69,$E9,$7E,$FC,$7D
	.dc.b	$EF,$83,$FC,$87,$C2,$D9,$FC,$1B,$FC,$2B,$E1,$FF,$29,$F0,$F9,$9F
	.dc.b	$8D,$3F,$86,$CB,$F3,$1A,$D9,$FC,$5E,$BF,$8A,$D7,$3A,$7E,$22,$FF
	.dc.b	$3D,$3F,$8E,$D3,$F4,$1C,$5F,$90,$DB,$F4,$5C,$79,$CB,$F2,$B3,$97
	.dc.b	$F2,$F1,$F9,$6D,$3F,$51,$BF,$E6,$36,$FF,$59,$FD,$47,$F9,$37,$BB
	.dc.b	$FC,$D5,$87,$E7,$AB,$48,$FF,$80,$33,$BE,$AE,$3D,$3F,$A2,$6F,$07
	.dc.b	$E8,$3A,$7A,$F5,$67,$0E,$44,$F6,$58,$B1,$F5,$4F,$E3,$77,$F5,$3F
	.dc.b	$8F,$DF,$2F,$EB,$A3,$2F,$EC,$23,$F6,$7C,$FE,$CE,$AF,$DA,$F4,$FB
	.dc.b	$BA,$B3,$7E,$23,$D9,$9F,$8A,$B7,$19,$11,$E6,$9F,$DF,$F8,$7F,$23
	.dc.b	$C3,$B5,$BA,$F5,$C7,$F0,$4F,$0B,$E6,$F8,$7A,$AF,$93,$3E,$87,$F8
	.dc.b	$1E,$B5,$EF,$DE,$DB,$DF,$1A,$E3,$FE,$17,$D3,$F8,$07,$F8,$6F,$81
	.dc.b	$F6,$BC,$2F,$E2,$6F,$DC,$F0,$FF,$8A,$F8,$79,$AD,$E1,$E7,$E6,$57
	.dc.b	$F6,$5F,$A8,$79,$5F,$84,$F3,$70,$FE,$1B,$CD,$B8,$71,$9E,$1E,$6E
	.dc.b	$72,$FF,$55,$61,$FE,$50,$AE,$DE,$9F,$5F,$3B,$F1,$FE,$75,$9C,$DF
	.dc.b	$0F,$4F,$2F,$F2,$83,$F2,$BE,$77,$F2,$E7,$F2,$FE,$75,$AF,$54,$FF
	.dc.b	$32,$F3,$F3,$5E,$2F,$53,$F3,$BE,$B7,$B7,$D6,$FC,$FF,$AF,$71,$F8
	.dc.b	$53,$F0,$6B,$37,$E3,$9F,$29,$FD,$B7,$DB,$70,$37,$F3,$CE,$63,$0F
	.dc.b	$09,$FF,$0E,$47,$C3,$8F,$D4,$EF,$F3,$FD,$5F,$D8,$FA,$FF,$47,$D7
	.dc.b	$B9,$7C,$A7,$F6,$BE,$C6,$F0,$FB,$15,$F5,$FD,$CF,$DC,$FC,$27,$D9
	.dc.b	$F7,$EE,$3F,$49,$FD,$EF,$BF,$BC,$3F,$01,$5F,$7F,$9F,$F8,$0F,$47
	.dc.b	$D3,$9F,$2E,$1F,$6B,$87,$2F,$C3,$8C,$9E,$FF,$77,$F4,$CF,$E9,$BE
	.dc.b	$A7,$F8,$34,$FE,$0B,$DF,$CF,$98,$9F,$F3,$F4,$E6,$BF,$51,$1C,$1E
	.dc.b	$86,$3E,$03,$F2,$9F,$09,$CA,$7F,$CF,$73,$F3,$FE,$17,$1F,$0D,$F9
	.dc.b	$97,$5F,$0F,$C2,$7A,$CF,$F9,$07,$0F,$F2,$64,$79,$FE,$23,$1C,$5D
	.dc.b	$7E,$1A,$CF,$92,$78,$71,$E7,$5F,$CE,$FC,$3F,$8F,$6F,$D4,$7C,$45
	.dc.b	$9F,$D0,$BE,$7D,$BF,$59,$59,$C7,$F0,$DA,$FA,$E3,$F6,$0F,$B6,$AC
	.dc.b	$FC,$49,$FD,$A6,$D9,$C3,$F9,$97,$77,$F9,$9F,$7F,$5F,$0F,$CD,$DB
	.dc.b	$D3,$8F,$F1,$D8,$FA,$1B,$E2,$3F,$98,$70,$AB,$DB,$1F,$4A,$7F,$81
	.dc.b	$47,$CF,$E8,$FE,$09,$1F,$47,$A2,$D7,$D8,$9D,$BF,$6A,$3E,$77,$4F
	.dc.b	$D7,$E9,$FA,$1D,$57,$6F,$49,$EA,$D5,$1F,$77,$A7,$8E,$3F,$84,$BF
	.dc.b	$2B,$EB,$D7,$D1,$87,$28,$6F,$FE,$C9,$E7,$FA,$D1,$FB,$7C,$BF,$D1
	.dc.b	$CF,$C0,$DF,$9B,$F0,$B7,$FE,$40,$FE,$37,$A6,$BF,$94,$6B,$E1,$E9
	.dc.b	$FE,$4D,$7B,$CB,$C5,$97,$FB,$67,$C4,$FC,$B7,$81,$F8,$F6,$CE,$A9
	.dc.b	$1F,$9B,$F0,$FF,$29,$E2,$EB,$2F,$47,$A7,$F9,$FF,$07,$F3,$0F,$E8
	.dc.b	$7C,$1B,$3F,$A8,$9F,$CE,$7A,$99,$CF,$B7,$8B,$FA,$7F,$2F,$F4,$1E
	.dc.b	$3F,$EA,$7C,$9B,$8F,$EC,$27,$F4,$95,$9B,$F3,$79,$5F,$B8,$D3,$FC
	.dc.b	$DB,$7F,$2E,$31,$3F,$AB,$75,$E5,$7E,$B7,$CB,$CC,$FB,$EF,$D4,$DB
	.dc.b	$FD,$FE,$DE,$A7,$EE,$6D,$EA,$73,$3F,$1F,$C9,$F1,$7A,$DF,$B9,$7C
	.dc.b	$9E,$B5,$AF,$EE,$18,$AB,$F7,$DE,$66,$B0,$FA,$1E,$C7,$AF,$D4,$FE
	.dc.b	$05,$EC,$7B,$7E,$0E,$E5,$CA,$7F,$83,$56,$6F,$D1,$3E,$66,$BF,$AA
	.dc.b	$DB,$D4,$FE,$B3,$6F,$53,$F9,$78,$FE,$21,$EB,$7C,$BF,$5A,$CF,$A1
	.dc.b	$AF,$ED,$F6,$7A,$DF,$C5,$FD,$5E,$19,$7E,$3B,$EA,$9F,$03,$F5,$7E
	.dc.b	$07,$FD,$B9,$18,$C7,$E1,$FC,$0E,$2B,$57,$F9,$1F,$89,$FB,$93,$C3
	.dc.b	$E1,$FE,$F7,$C0,$F2,$F8,$8B,$7D,$DF,$C1,$7C,$16,$F2,$F9,$F5,$FC
	.dc.b	$AF,$E0,$FF,$0D,$87,$F2,$DF,$84,$FC,$46,$16,$FE,$AC,$ED,$FA,$F1
	.dc.b	$E8,$F8,$9F,$C7,$7C,$47,$AB,$E2,$6E,$9E,$F7,$F2,$5F,$13,$BC,$3E
	.dc.b	$FD,$78,$DF,$91,$A7,$A5,$F2,$DF,$A5,$FC,$2E,$BB,$7F,$CA,$1C,$AA
	.dc.b	$F9,$3F,$11,$F9,$FC,$FE,$5F,$C3,$FE,$87,$4B,$87,$E3,$A7,$6F,$E4
	.dc.b	$23,$E8,$DF,$E9,$EF,$E9,$E3,$FA,$9C,$2D,$FE,$5A,$76,$FE,$62,$3C
	.dc.b	$8F,$F9,$4B,$1F,$26,$5E,$98,$F6,$B5,$F4,$DF,$07,$FC,$AB,$3C,$1F
	.dc.b	$F0,$B7,$27,$F9,$E9,$FC,$67,$27,$E1,$B6,$FE,$E7,$F5,$6E,$5C,$57
	.dc.b	$EA,$AB,$F1,$7C,$1A,$9F,$C6,$F4,$F0,$E5,$FC,$7F,$4D,$DB,$F2,$8F
	.dc.b	$4E,$6D,$E5,$FE,$63,$CF,$FC,$F9,$5F,$C9,$BE,$B3,$FC,$9B,$39,$EE
	.dc.b	$4E,$7C,$AF,$2C,$62,$7F,$C6,$15,$C1,$FE,$B9,$6F,$0B,$5E,$9C,$A7
	.dc.b	$99,$CB,$2F,$D4,$77,$7C,$0E,$5F,$D3,$BF,$0B,$CE,$B9,$F9,$E7,$F5
	.dc.b	$7D,$F9,$BF,$D8,$78,$FF,$23,$C2,$FD,$7B,$F2,$BC,$2B,$87,$BE,$7F
	.dc.b	$67,$E4,$66,$FD,$93,$EF,$6B,$EA,$E1,$F4,$C7,$F7,$CD,$BD,$D3,$FE
	.dc.b	$A5,$E1,$F7,$4F,$C9,$DF,$D9,$3F,$7B,$C6,$FB,$7E,$67,$F0,$27,$EE
	.dc.b	$F9,$76,$7E,$94,$FF,$06,$F2,$73,$87,$D3,$F2,$7F,$07,$E5,$F8,$AF
	.dc.b	$FB,$5E,$8F,$2B,$F0,$FE,$5E,$7A,$13,$9A,$FB,$D7,$98,$61,$C3,$1F
	.dc.b	$7D,$7A,$F8,$35,$D5,$EA,$C7,$EC,$3F,$94,$EA,$E5,$C3,$9C,$3F,$C3
	.dc.b	$FE,$1F,$E0,$F1,$FC,$0F,$A3,$FC,$83,$CD,$FC,$DB,$63,$F2,$1E,$87
	.dc.b	$F2,$4B,$C9,$F9,$1F,$57,$F4,$3E,$87,$FB,$4B,$5F,$C4,$7A,$B8,$C3
	.dc.b	$47,$E6,$67,$57,$F5,$67,$F5,$3E,$96,$2B,$E3,$FC,$37,$EB,$3D,$EC
	.dc.b	$63,$F1,$BE,$1B,$15,$FA,$1A,$AF,$E7,$23,$F6,$7E,$FF,$0E,$67,$B3
	.dc.b	$FB,$6F,$87,$C7,$B3,$70,$FD,$54,$D7,$F4,$B1,$FB,$CF,$77,$97,$C1
	.dc.b	$7F,$9C,$3E,$B7,$FD,$BD,$AF,$E9,$63,$1E,$EF,$7F,$83,$B8,$FE,$D6
	.dc.b	$78,$3E,$3D,$FC,$F7,$FA,$DD,$EF,$1F,$E5,$E6,$D7,$E2,$27,$D6,$FE
	.dc.b	$FF,$E0,$BC,$5F,$11,$ED,$6C,$7F,$03,$F8,$4A,$FE,$FA,$3D,$1F,$0D
	.dc.b	$F0,$FE,$13,$D4,$59,$FF,$77,$6F,$FE,$59,$7F,$86,$7C,$26,$9F,$E1
	.dc.b	$D3,$F0,$63,$D5,$1F,$C4,$F9,$FE,$76,$7E,$8D,$BF,$37,$AD,$E6,$4F
	.dc.b	$AA,$FF,$88,$FC,$3F,$CB,$AF,$AC,$FC,$D9,$FB,$1F,$11,$6B,$F8,$F4
	.dc.b	$ED,$FE,$40,$7D,$37,$EF,$3F,$56,$3F,$01,$C9,$63,$67,$BA,$7F,$17
	.dc.b	$F5,$DF,$E4,$88,$F8,$3C,$98,$EF,$FB,$CE,$BC,$0B,$3F,$E5,$4A,$57
	.dc.b	$E0,$F9,$FF,$1D,$CD,$F8,$5E,$5F,$C8,$72,$DC,$3F,$9B,$4D,$7F,$98
	.dc.b	$C7,$E2,$F9,$BC,$7C,$7F,$8D,$F0,$3C,$BD,$BF,$8F,$EF,$CE,$1F,$59
	.dc.b	$CB,$F6,$27,$CC,$FA,$8F,$36,$9E,$87,$FD,$DD,$AF,$BD,$FF,$4D,$74
	.dc.b	$B8,$3E,$04,$7E,$0B,$7F,$8B,$DD,$F9,$FE,$EF,$8F,$DB,$FA,$1E,$EC
	.dc.b	$DF,$E2,$1C,$BF,$8A,$9F,$9B,$E4,$7E,$9F,$C8,$B3,$F8,$EF,$2B,$E8
	.dc.b	$F9,$5F,$AB,$73,$1E,$49,$F5,$BE,$6D,$3C,$AD,$3E,$97,$47,$B7,$BF
	.dc.b	$FA,$BF,$07,$D4,$9F,$51,$F5,$FC,$4F,$DC,$79,$3F,$80,$6C,$7D,$BF
	.dc.b	$16,$BF,$1A,$3F,$7D,$E4,$7E,$1B,$D1,$FF,$77,$FD,$7F,$F2,$7C,$BB
	.dc.b	$FA,$78,$F2,$B5,$88,$FF,$63,$52,$3F,$1F,$3C,$BC,$AF,$E3,$67,$F9
	.dc.b	$36,$B7,$EA,$7E,$53,$DD,$EE,$EB,$DF,$B3,$F9,$19,$F1,$FA,$DF,$9A
	.dc.b	$F4,$FC,$BE,$AD,$AF,$11,$3E,$29,$AB,$D5,$FD,$0F,$AD,$EC,$78,$57
	.dc.b	$A2,$FF,$4B,$EB,$7C,$0F,$7E,$CF,$43,$DD,$8F,$4F,$7B,$F5,$7E,$C7
	.dc.b	$C6,$77,$3F,$0A,$BD,$4F,$11,$5E,$4E,$0F,$81,$19,$8C,$3F,$EC,$2E
	.dc.b	$1F,$8C,$FA,$F5,$FF,$4A,$3F,$0C,$E0,$B7,$FD,$EF,$B7,$F6,$3D,$DB
	.dc.b	$5A,$F7,$3F,$80,$FB,$BF,$71,$CC,$7D,$89,$FA,$BF,$09,$F8,$0F,$85
	.dc.b	$B3,$DB,$E1,$3E,$CF,$C2,$FE,$11,$CC,$7E,$02,$7D,$EF,$DC,$3C,$C5
	.dc.b	$3F,$DD,$BA,$FF,$BE,$3A,$7E,$EC,$7E,$DB,$A7,$EE,$BF,$C3,$FE,$0F
	.dc.b	$F2,$3F,$13,$FC,$61,$B1,$FC,$4F,$E1,$2B,$FB,$E8,$FC,$BF,$C4,$FE
	.dc.b	$3B,$E1,$BF,$32,$7F,$83,$F2,$7C,$3E,$1F,$9D,$1F,$9E,$8D,$19,$BF
	.dc.b	$39,$E8,$E8,$FD,$1F,$C4,$F3,$EA,$FD,$2F,$C4,$F5,$9C,$E3,$F6,$67
	.dc.b	$2F,$DB,$AF,$D5,$F1,$F8,$DB,$1F,$7E,$FF,$5F,$C7,$9C,$7E,$ED,$FA
	.dc.b	$78,$BF,$67,$C5,$EB,$E3,$FC,$65,$7E,$33,$BF,$FC,$E5,$97,$E2,$6F
	.dc.b	$6F,$E2,$EB,$7D,$7F,$A3,$E5,$F4,$61,$F8,$AD,$7D,$8F,$E3,$35,$CB
	.dc.b	$FB,$C8,$CB,$FB,$E8,$F9,$9C,$BD,$33,$FF,$3A,$CF,$C8,$F1,$3F,$CE
	.dc.b	$73,$9E,$6F,$A5,$CB,$BE,$6E,$37,$FA,$1C,$15,$F5,$FC,$0E,$9E,$07
	.dc.b	$D9,$9E,$DE,$06,$6F,$84,$E5,$E5,$1E,$87,$F5,$1A,$FA,$1F,$D5,$6B
	.dc.b	$97,$F7,$7D,$E9,$D4,$62,$7F,$11,$D7,$ED,$EE,$B3,$E7,$EE,$FC,$67
	.dc.b	$76,$79,$BF,$71,$7F,$0E,$7F,$21,$1F,$17,$C5,$B7,$EF,$F1,$3F,$29
	.dc.b	$E2,$E6,$BF,$80,$1F,$2E,$3F,$CC,$DF,$60,$F6,$7E,$77,$8F,$86,$D7
	.dc.b	$F3,$1E,$9F,$F4,$CD,$3D,$AF,$A9,$E4,$6F,$3F,$9D,$8F,$E8,$FC,$8F
	.dc.b	$B1,$E9,$D9,$FA,$5E,$9F,$E9,$FC,$AC,$E9,$FC,$4E,$FE,$F4,$F9,$3D
	.dc.b	$2F,$C0,$7A,$37,$0F,$B4,$FA,$3D,$3F,$C2,$7B,$59,$9F,$E5,$5E,$2F
	.dc.b	$F3,$3A,$F2,$46,$3B,$DB,$62,$3F,$9D,$46,$34,$F7,$1E,$FE,$1F,$74
	.dc.b	$FE,$27,$83,$83,$FC,$DA,$B8,$3F,$93,$9F,$F5,$1B,$9F,$43,$E6,$FA
	.dc.b	$5F,$99,$7E,$7F,$A3,$F9,$BF,$4B,$35,$F9,$59,$FA,$7F,$01,$F9,$FF
	.dc.b	$76,$CF,$E6,$5F,$AF,$EF,$7E,$8F,$DE,$CC,$FB,$67,$CA,$FE,$DF,$7B
	.dc.b	$C3,$FD,$55,$FB,$FD,$3F,$63,$87,$DF,$4E,$A7,$07,$E0,$E3,$ED,$CE
	.dc.b	$A7,$3E,$CF,$EC,$FD,$CF,$C4,$BF,$B5,$F7,$3F,$17,$EE,$66,$F8,$63
	.dc.b	$FE,$86,$C7,$F6,$55,$F9,$0F,$81,$FE,$0F,$F6,$3F,$DF,$6F,$7F,$09
	.dc.b	$8B,$D1,$FD,$66,$5F,$09,$F9,$89,$D5,$76,$AE,$7F,$09,$F9,$C6,$C7
	.dc.b	$8A,$BB,$FC,$26,$67,$A5,$7E,$8B,$E0,$BC,$7F,$13,$FA,$4F,$84,$B8
	.dc.b	$FA,$A7,$CF,$F1,$39,$C3,$C9,$7F,$AA,$3D,$7F,$11,$FA,$C9,$F6,$FC
	.dc.b	$46,$6B,$D5,$39,$7D,$91,$E3,$7F,$33,$B3,$3C,$47,$F8,$3E,$31,$97
	.dc.b	$AE,$3E,$26,$9A,$7F,$C1,$33,$A7,$F2,$98,$7C,$ED,$FF,$7E,$DB,$FA
	.dc.b	$B8,$FD,$2D,$F3,$7F,$3E,$FF,$DD,$5F,$64,$E5,$FA,$FB,$E3,$8A,$E5
	.dc.b	$F8,$5A,$D6,$1F,$5F,$87,$4F,$D9,$BF,$33,$FE,$16,$D7,$CD,$CD,$FC
	.dc.b	$32,$3E,$B9,$FA,$BE,$3F,$AF,$8F,$DD,$8F,$AE,$FE,$33,$87,$F1,$3D
	.dc.b	$BF,$C6,$9F,$C5,$F6,$FF,$1C,$E3,$CD,$6E,$FF,$15,$3E,$CE,$EF,$C8
	.dc.b	$F9,$9E,$EF,$02,$DF,$AE,$7F,$2B,$D9,$9C,$3F,$25,$7F,$13,$CF,$FC
	.dc.b	$CF,$6D,$9F,$CB,$76,$FC,$8F,$3B,$39,$7C,$5B,$FC,$FC,$FC,$CF,$07
	.dc.b	$F4,$3E,$57,$CE,$F0,$ED,$FC,$B9,$FD,$2F,$95,$9C,$BF,$45,$7F,$53
	.dc.b	$CE,$FD,$4F,$8F,$F5,$BC,$CB,$8F,$D3,$9F,$D6,$F9,$19,$C3,$F5,$57
	.dc.b	$E7,$7F,$1F,$BD,$7F,$1D,$AF,$A5,$FD,$84,$65,$FC,$6C,$7E,$13,$CB
	.dc.b	$FE,$FF,$F6,$6C,$BF,$0F,$E6,$E3,$D0,$D6,$1F,$E5,$1A,$D5,$7E,$1E
	.dc.b	$B4,$FE,$27,$D4,$E7,$E8,$7E,$41,$B1,$F8,$DF,$4B,$BF,$A5,$9C,$7A
	.dc.b	$5F,$95,$FC,$66,$B7,$D0,$78,$6D,$FD,$95,$F0,$32,$FA,$8F,$D8,$7F
	.dc.b	$35,$E8,$7B,$3D,$AF,$D0,$BE,$EF,$4B,$F4,$7E,$9E,$6B,$77,$ED,$9F
	.dc.b	$D3,$FB,$9F,$13,$E1,$BF,$53,$EE,$FC,$6F,$86,$CD,$FC,$39,$CB,$F1
	.dc.b	$63,$E5,$3F,$E1,$DF,$B5,$7C,$4F,$CE,$F6,$B3,$E7,$49,$FF,$27,$D7
	.dc.b	$CD,$9C,$E9,$FD,$DE,$78,$CF,$D1,$F6,$F2,$FF,$3F,$8E,$2F,$F6,$8E
	.dc.b	$A5,$FF,$00,$F8,$0F,$BB,$F0,$B6,$BE,$D7,$C2,$FF,$04,$F8,$2C,$E7
	.dc.b	$FC,$06,$FF,$07,$3D,$B4,$66,$F8,$7F,$C3,$FC,$2E,$63,$19,$7E,$17
	.dc.b	$E1,$71,$F1,$1A,$CF,$FA,$85,$6A,$BC,$D5,$C7,$87,$E2,$56,$BE,$26
	.dc.b	$D7,$C9,$9F,$CA,$F1,$77,$9C,$E9,$F2,$7E,$27,$F3,$35,$FC,$7B,$E2
	.dc.b	$6E,$3E,$09,$FC,$ED,$E7,$1F,$CD,$5F,$A7,$5F,$D0,$FC,$45,$9F,$CF
	.dc.b	$7C,$47,$B7,$5C,$E5,$EA,$BF,$D3,$CF,$A6,$BE,$C4,$FA,$5A,$7A,$4F
	.dc.b	$B0,$E5,$FA,$DC,$BF,$27,$0F,$D8,$F0,$D9,$FD,$73,$F3,$6F,$F6,$BC
	.dc.b	$39,$9F,$BD,$3F,$46,$FF,$73,$AF,$D3,$E0,$FF,$3C,$FD,$AF,$F9,$FA
	.dc.b	$FF,$25,$5A,$CF,$E9,$D6,$B8,$3E,$D4,$70,$EB,$FB,$9E,$4D,$8F,$B2
	.dc.b	$74,$EC,$CE,$3F,$ED,$3C,$BF,$CF,$7C,$9F,$9B,$E9,$F0,$F4,$7E,$15
	.dc.b	$F1,$F4,$FE,$1F,$A7,$39,$7E,$12,$72,$FE,$1B,$C0,$F4,$F9,$1F,$8D
	.dc.b	$6C,$7E,$2B,$BF,$DB,$E4,$67,$5F,$15,$FA,$9F,$27,$07,$99,$FF,$25
	.dc.b	$ED,$E6,$C3,$F5,$87,$AC,$FC,$AF,$07,$BE,$7F,$71,$E0,$7C,$AF,$4F
	.dc.b	$F3,$DE,$0D,$BF,$CD,$CF,$CE,$F5,$33,$7F,$0E,$FF,$51,$CB,$FE,$4A
	.dc.b	$8F,$F7,$DE,$5F,$E4,$EE,$4F,$E0,$32,$F3,$BE,$BF,$89,$C3,$CE,$FB
	.dc.b	$3E,$37,$29,$CE,$BF,$4B,$2F,$6B,$E8,$69,$FC,$3D,$F6,$3F,$C4,$77
	.dc.b	$F6,$65,$FC,$3F,$CB,$FC,$2F,$9F,$FB,$97,$F0,$FE,$87,$EE,$FC,$EC
	.dc.b	$E3,$FB,$89,$CB,$FE,$49,$DB,$FC,$99,$97,$CC,$F4,$0E,$7D,$6F,$C8
	.dc.b	$FA,$38,$8F,$F5,$F5,$7F,$9C,$9F,$A7,$19,$D3,$E1,$FA,$DF,$98,$C3
	.dc.b	$18,$FF,$A5,$F0,$FF,$28,$61,$F6,$BD,$7C,$BF,$E5,$F8,$FE,$88,$FE
	.dc.b	$4E,$33,$82,$73,$3C,$4F,$57,$A7,$FA,$6F,$63,$72,$F5,$7F,$53,$EB
	.dc.b	$E3,$D7,$B8,$7C,$16,$BE,$C9,$FD,$7F,$B1,$CB,$E0,$FF,$63,$F0,$7D
	.dc.b	$3E,$12,$DF,$C9,$9A,$FC,$48,$FD,$BF,$C2,$78,$BE,$0B,$FD,$E3,$F0
	.dc.b	$DF,$26,$BE,$4F,$C2,$63,$E1,$35,$19,$1F,$00,$AF,$F3,$6C,$7D,$7A
	.dc.b	$E6,$FB,$BE,$13,$AF,$C2,$7C,$0F,$83,$B3,$F7,$66,$BF,$62,$3C,$7F
	.dc.b	$09,$F1,$AB,$CB,$F1,$1F,$9E,$6F,$57,$E1,$67,$D3,$F1,$19,$C3,$CD
	.dc.b	$A7,$97,$0F,$D2,$9E,$ED,$3F,$4F,$57,$0F,$C7,$4F,$C2,$F8,$8C,$E1
	.dc.b	$CA,$FD,$D5,$F9,$37,$E3,$F1,$7E,$C1,$F9,$3C,$5F,$B2,$BC,$D6,$EF
	.dc.b	$94,$FF,$5D,$FB,$67,$47,$D0,$E3,$FC,$4C,$39,$BE,$9F,$1E,$1B,$87
	.dc.b	$FA,$AE,$78,$BE,$5A,$E3,$CD,$F6,$B9,$39,$F3,$5A,$E0,$D7,$FD,$71
	.dc.b	$A7,$FB,$D7,$0F,$F3,$8C,$7F,$B8,$B6,$72,$FE,$5A,$3F,$0B,$DF,$F9
	.dc.b	$8E,$7F,$C3,$F8,$36,$7E,$33,$B7,$E4,$4F,$B3,$C5,$FC,$6F,$77,$BB
	.dc.b	$C5,$FC,$7F,$C3,$E6,$FE,$73,$97,$E8,$4F,$C4,$F0,$FF,$2B,$DD,$67
	.dc.b	$EA,$3F,$23,$C4,$CC,$7C,$59,$FC,$DF,$81,$F3,$3C,$4F,$CE,$F9,$5F
	.dc.b	$3B,$C4,$B7,$F2,$E6,$BF,$10,$FF,$6E,$71,$A7,$F4,$BE,$4F,$D6,$F4
	.dc.b	$BF,$45,$5F,$A2,$FB,$8F,$F2,$AB,$C7,$C8,$60,$F5,$37,$E3,$2D,$E6
	.dc.b	$7E,$0E,$73,$1C,$AF,$FC,$6F,$6C,$73,$7F,$89,$F4,$37,$E6,$5C,$FF
	.dc.b	$3F,$8B,$CD,$C4,$7E,$EF,$4F,$05,$7E,$47,$D1,$EF,$5F,$93,$F3,$FC
	.dc.b	$3E,$B5,$C3,$9F,$AD,$5E,$59,$27,$F3,$5E,$97,$A2,$BF,$C7,$1C,$0F
	.dc.b	$7B,$D7,$EA,$E1,$BE,$2F,$F3,$89,$FF,$29,$9B,$CB,$D3,$EB,$7E,$97
	.dc.b	$D5,$DC,$7C,$BE,$57,$AF,$F1,$BD,$EF,$D5,$7B,$B7,$2F,$D2,$3F,$2B
	.dc.b	$DE,$FD,$75,$67,$0E,$5E,$F7,$CE,$F7,$3F,$64,$DC,$3F,$58,$FD,$2F
	.dc.b	$77,$F6,$DE,$F6,$6F,$EE,$6F,$FE,$64,$7E,$1F,$C1,$FC,$0F,$86,$FB
	.dc.b	$3F,$0B,$F0,$BE,$1B,$ED,$CE,$6B,$EA,$CF,$C6,$F8,$6F,$BF,$5F,$23
	.dc.b	$E1,$AE,$BF,$14,$F8,$D9,$2B,$F0,$B9,$6F,$1F,$BF,$87,$D0,$F8,$5F
	.dc.b	$C4,$D5,$9F,$C1,$E3,$F5,$3E,$1B,$39,$75,$96,$53,$0F,$C8,$DF,$DA
	.dc.b	$E4,$FC,$9C,$7F,$5C,$97,$1F,$E5,$B1,$C3,$70,$E9,$3D,$2F,$8C,$62
	.dc.b	$1C,$5F,$9E,$E8,$E8,$D9,$D3,$5F,$9F,$C5,$FA,$8F,$BA,$FE,$D9,$1C
	.dc.b	$F8,$B1,$CB,$E5,$6C,$64,$FE,$83,$3F,$88,$7C,$5C,$7C,$F9,$BD,$9E
	.dc.b	$17,$5E,$7B,$5C,$39,$FE,$07,$83,$E1,$73,$1F,$AB,$C5,$5F,$17,$BF
	.dc.b	$CD,$E3,$7C,$7E,$FB,$7F,$06,$6B,$F8,$A8,$F6,$78,$FF,$37,$C0,$F7
	.dc.b	$78,$DF,$3F,$C1,$B7,$F2,$66,$BF,$92,$8F,$89,$E2,$FF,$84,$3E,$F3
	.dc.b	$FC,$15,$3F,$03,$C6,$FF,$55,$43,$CD,$F9,$51,$8F,$43,$E6,$78,$FA
	.dc.b	$9C,$F8,$5F,$6A,$FF,$C9,$5E,$4B,$8D,$EA,$74,$F3,$FE,$A3,$DB,$CD
	.dc.b	$FA,$DE,$A6,$76,$FA,$33,$FC,$13,$2F,$B5,$E8,$D9,$FA,$AF,$F0,$8F
	.dc.b	$3F,$EF,$7A,$59,$9F,$9D,$3E,$AF,$6F,$F0,$5E,$9F,$BD,$B3,$E7,$6B
	.dc.b	$FC,$C7,$7F,$EB,$2F,$E0,$BD,$3F,$8D,$EB,$7E,$2F,$E0,$3E,$47,$B1
	.dc.b	$F8,$DF,$7E,$DF,$E2,$1D,$BF,$8A,$9F,$9D,$EC,$7E,$4F,$DE,$FA,$1E
	.dc.b	$D5,$AF,$90,$D7,$C5,$3F,$99,$F7,$BE,$B7,$BD,$FD,$0B,$EF,$3F,$C5
	.dc.b	$DE,$97,$FB,$D3,$77,$C1,$E7,$DF,$FB,$8D,$CB,$EA,$61,$C5,$FC,$07
	.dc.b	$C2,$73,$F8,$8F,$C1,$7C,$2F,$5F,$88,$B5,$FE,$22,$D1,$1C,$23,$15
	.dc.b	$F8,$9F,$85,$F2,$7C,$3D,$9E,$0D,$7F,$A5,$5A,$3F,$21,$F0,$FE,$BF
	.dc.b	$88,$FC,$95,$7B,$74,$FC,$A6,$D6,$3F,$18,$53,$EB,$3F,$0F,$6F,$CD
	.dc.b	$67,$F1,$76,$B5,$E9,$7F,$3D,$AE,$63,$F2,$D3,$F2,$F6,$FD,$16,$9F
	.dc.b	$37,$6B,$5F,$05,$FD,$36,$99,$8F,$CF,$4F,$D3,$D7,$FC,$F3,$0E,$BF
	.dc.b	$AD,$D6,$EF,$FE,$3B,$39,$3E,$67,$C4,$65,$FF,$00,$1F,$E0,$73,$E8
	.dc.b	$4E,$79,$24,$71,$3E,$5C,$E6,$BF,$B2,$47,$F8,$F0,$F9,$FD,$99,$7F
	.dc.b	$9B,$99,$3F,$D3,$DE,$33,$95,$DB,$BC,$7F,$D4,$BD,$BF,$8F,$C3,$18
	.dc.b	$71,$3F,$D5,$38,$71,$BC,$F6,$7E,$57,$AF,$7D,$7F,$97,$EC,$CE,$1F
	.dc.b	$8B,$9C,$BF,$8D,$8E,$BE,$57,$E7,$7C,$AF,$97,$E5,$7E,$7F,$CA,$B7
	.dc.b	$F9,$37,$E7,$D6,$E3,$E6,$F8,$DF,$A5,$F2,$7E,$9F,$91,$FA,$7F,$26
	.dc.b	$D7,$E6,$DF,$AF,$5B,$8F,$AB,$E3,$FF,$8B,$BC,$FF,$A9,$1F,$52,$31
	.dc.b	$E5,$7D,$E6,$C7,$DC,$F2,$B8,$74,$FE,$0B,$D3,$E5,$EB,$7E,$13,$D5
	.dc.b	$E9,$EB,$5C,$3F,$D1,$B9,$FF,$8B,$DF,$C2,$DF,$07,$F0,$FE,$B7,$8F
	.dc.b	$D6,$FC,$6B,$E5,$F5,$AC,$F2,$7F,$23,$7B,$8F,$C7,$FA,$DE,$BF,$5B
	.dc.b	$F2,$AF,$B7,$D6,$B3,$E2,$7F,$33,$EB,$6E,$3F,$BF,$7A,$6F,$BF,$7C
	.dc.b	$17,$C5,$F8,$0F,$F6,$DC,$E6,$31,$5C,$2B,$6F,$BA,$B7,$0F,$82,$F9
	.dc.b	$DF,$05,$8F,$82,$FA,$1F,$05,$FA,$AF,$88,$B8,$FE,$A1,$AF,$F9,$B7
	.dc.b	$3F,$25,$7D,$7F,$83,$FD,$8F,$C1,$FD,$9F,$83,$FD,$9F,$C1,$7D,$BF
	.dc.b	$83,$B8,$FD,$07,$6F,$D2,$9F,$29,$F8,$1F,$86,$F3,$CF,$E0,$FE,$1E
	.dc.b	$D7,$D8,$76,$FD,$A9,$F6,$9F,$89,$F8,$7B,$1F,$7B,$E1,$EB,$FE,$6B
	.dc.b	$E9,$47,$C4,$BE,$13,$F2,$6B,$F2,$5C,$BF,$22,$BF,$29,$C7,$6B,$F1
	.dc.b	$2F,$CC,$AD,$C7,$D1,$E9,$FC,$D7,$07,$D0,$E8,$FF,$44,$FE,$15,$FE
	.dc.b	$22,$AF,$95,$CE,$9C,$F2,$FD,$6C,$38,$D7,$DB,$E8,$E3,$C9,$F6,$9B
	.dc.b	$19,$3F,$DC,$F8,$FF,$89,$DE,$BD,$1E,$0E,$5F,$C0,$F4,$78,$BB,$BF
	.dc.b	$07,$D1,$6F,$F1,$0D,$7C,$B3,$E8,$F0,$3F,$13,$E1,$7A,$BB,$ED,$78
	.dc.b	$1A,$FF,$A5,$F3,$FF,$50,$3F,$93,$F1,$7E,$0F,$4F,$E4,$BC,$3F,$87
	.dc.b	$E0,$5A,$F4,$B5,$E4,$FE,$63,$C3,$F9,$3E,$0F,$E6,$BC,$3F,$97,$E5
	.dc.b	$5B,$F8,$2D,$7C,$4F,$E8,$3C,$4F,$A3,$E6,$7E,$A3,$F0,$27,$A1,$F5
	.dc.b	$3C,$AC,$7A,$17,$1F,$D8,$D2,$31,$1F,$A1,$8E,$2F,$D0,$F3,$7A,$7A
	.dc.b	$1F,$7B,$CD,$ED,$E8,$7E,$03,$CE,$B7,$F5,$9D,$BF,$62,$7C,$9E,$87
	.dc.b	$E1,$BD,$BF,$37,$A3,$F8,$8F,$88,$B7,$F7,$9D,$BF,$80,$9F,$67,$C0
	.dc.b	$7E,$3B,$E2,$3D,$DE,$FD,$AF,$5B,$5F,$54,$FE,$53,$DB,$F8,$9E,$FF
	.dc.b	$E5,$BD,$9F,$8D,$EF,$FE,$63,$DA,$B8,$7E,$39,$DB,$F9,$09,$F9,$9E
	.dc.b	$EF,$FB,$9B,$F0,$BF,$F1,$2F,$8C,$A7,$B8,$E8,$FE,$BF,$19,$3F,$B1
	.dc.b	$D3,$E2,$3E,$B7,$BD,$CD,$EB,$3B,$3F,$B0,$CF,$67,$AD,$7F,$87,$5D
	.dc.b	$0F,$7D,$2F,$85,$FC,$17,$C2,$79,$FE,$1A,$DF,$0B,$DC,$7E,$1F,$E1
	.dc.b	$7D,$97,$F8,$9F,$89,$F7,$45,$BE,$18,$B1,$C4,$7F,$BE,$63,$0F,$E3
	.dc.b	$3E,$27,$8B,$F8,$EF,$89,$F9,$18,$7E,$55,$F9,$53,$67,$AB,$F9,$9F
	.dc.b	$89,$DC,$7F,$46,$FC,$11,$C9,$F4,$6F,$FA,$2C,$62,$7C,$B1,$89,$D4
	.dc.b	$71,$E5,$FB,$17,$FA,$58,$FB,$5D,$35,$F6,$B5,$F6,$3F,$AB,$E2,$FC
	.dc.b	$07,$57,$EB,$78,$3F,$05,$D3,$6F,$E2,$B5,$F8,$8F,$EC,$F8,$BF,$11
	.dc.b	$D3,$E8,$F0,$3F,$15,$D5,$EA,$F0,$6E,$1E,$F6,$BF,$D4,$F8,$55,$F9
	.dc.b	$1E,$9F,$83,$DD,$F9,$3F,$1E,$DF,$D6,$6B,$F5,$5F,$8F,$E0,$7E,$67
	.dc.b	$C6,$F9,$3E,$0F,$E6,$FC,$5F,$97,$E1,$7E,$77,$C5,$B8,$FE,$01,$DB
	.dc.b	$F8,$29,$FA,$3E,$77,$E8,$FC,$6F,$A7,$E6,$DA,$FB,$0D,$7D,$53,$FA
	.dc.b	$9F,$1F,$EC,$F9,$9F,$9A,$9F,$CD,$7E,$1B,$FE,$24,$F8,$AF,$F2,$C7
	.dc.b	$92,$E5,$79,$BB,$F2,$56,$F4,$FF,$07,$39,$8C,$63,$F8,$9C,$B1,$E6
	.dc.b	$7E,$23,$CE,$D7,$A5,$6B,$F4,$96,$8C,$47,$EC,$F2,$EF,$1F,$90,$F3
	.dc.b	$FB,$7A,$5F,$92,$F7,$7C,$1E,$9D,$C3,$9F,$A7,$F9,$6F,$6F,$C8,$E6
	.dc.b	$3A,$66,$AF,$CD,$FB,$7E,$9F,$7B,$F3,$BE,$D7,$AF,$E0,$33,$87,$99
	.dc.b	$CB,$E8,$9F,$D1,$FB,$7F,$07,$DF,$B3,$EC,$7F,$4F,$ED,$E6,$3F,$59
	.dc.b	$3F,$1F,$DF,$FF,$17,$F6,$3E,$0F,$E5,$7B,$F8,$7E,$67,$C0,$6B,$E1
	.dc.b	$6E,$3F,$E3,$AB,$E0,$FC,$AB,$E7,$F0,$DF,$4F,$E2,$BA,$FC,$3F,$D5
	.dc.b	$F8,$AC,$DF,$D0,$72,$FD,$29,$F1,$FC,$47,$DB,$F8,$AF,$DF,$FC,$47
	.dc.b	$DD,$F8,$AC,$DF,$D8,$72,$FD,$A9,$F5,$E7,$F8,$3D,$FD,$BB,$DA,$F4
	.dc.b	$4D,$7F,$C8,$1B,$27,$F1,$5B,$FC,$4D,$3F,$19,$BD,$BF,$C3,$EF,$F2
	.dc.b	$37,$FC,$83,$98,$FB,$B3,$F3,$34,$FC,$A6,$F6,$7F,$1E,$FD,$0D,$B3
	.dc.b	$1F,$2E,$7F,$3D,$C9,$FE,$2B,$F2,$3F,$47,$B7,$E1,$EB,$F9,$D4,$FF
	.dc.b	$88,$E7,$3D,$7C,$3B,$7E,$DF,$5D,$BF,$AD,$3D,$3B,$7E,$FE,$1D,$BB
	.dc.b	$2D,$74,$7F,$07,$D7,$98,$FB,$D3,$E4,$EC,$FC,$3F,$67,$9B,$AE,$D7
	.dc.b	$93,$B3,$F1,$7D,$9E,$A7,$31,$FA,$BC,$3F,$C5,$2F,$F9,$BF,$F0,$A7
	.dc.b	$99,$F0,$3C,$BF,$F4,$0C,$62,$7D,$FE,$66,$39,$3E,$2B,$AB,$E1,$3F
	.dc.b	$93,$F1,$FE,$4F,$91,$F9,$C7,$E5,$F9,$3F,$9E,$F2,$AD,$6C,$F9,$75
	.dc.b	$FA,$2F,$23,$E8,$F9,$5F,$A4,$F2,$3E,$9F,$93,$9B,$FA,$2E,$5F,$A7
	.dc.b	$3E,$8F,$63,$EC,$FA,$FE,$AF,$5F,$ED,$FB,$19,$BF,$B2,$E5,$FB,$73
	.dc.b	$F0,$3D,$7F,$C0,$FB,$16,$7E,$FB,$F1,$3D,$7C,$C7,$FA,$AB,$97,$FC
	.dc.b	$6E,$FC,$DF,$63,$E4,$FA,$BF,$8A,$F6,$2C,$FE,$23,$D8,$F9,$BE,$C7
	.dc.b	$E3,$9C,$C7,$D5,$9F,$A3,$F0,$7F,$92,$F8,$3F,$A7,$F0,$56,$BC,$13
	.dc.b	$5F,$B5,$1F,$98,$F8,$3F,$B3,$F0,$7F,$CD,$FF,$10,$FF,$0D,$7B,$CD
	.dc.b	$DF,$07,$9F,$84,$DE,$5F,$E1,$5B,$FF,$13,$3E,$D9,$E4,$FE,$13,$E0
	.dc.b	$FA,$7C,$27,$E1,$BE,$0F,$B7,$C2,$66,$B9,$39,$7A,$4F,$E3,$0F,$27
	.dc.b	$C4,$FE,$3A,$6C,$FC,$B9,$AF,$E5,$63,$D5,$F1,$3F,$94,$AF,$67,$C4
	.dc.b	$7E,$5A,$3D,$DF,$11,$9B,$F5,$39,$7D,$93,$F9,$C3,$E2,$71,$D8,$F8
	.dc.b	$0F,$E8,$27,$31,$F7,$67,$E5,$71,$FF,$B9,$3F,$0C,$73,$7C,$DE,$4C
	.dc.b	$3F,$3F,$83,$5C,$F7,$1F,$F7,$7D,$F0,$7E,$75,$F3,$E7,$FA,$DC,$5D
	.dc.b	$79,$ED,$70,$9A,$FF,$B9,$6D,$3F,$77,$8F,$F6,$F3,$F7,$FC,$1F,$DC
	.dc.b	$F3,$58,$FC,$B1,$4F,$E0,$0F,$E1,$7C,$1F,$67,$85,$F8,$7F,$03,$DD
	.dc.b	$E1,$FE,$27,$C0,$CE,$1F,$85,$72,$FE,$1E,$7E,$27,$85,$F8,$FF,$02
	.dc.b	$CF,$E2,$FC,$0F,$91,$E1,$FE,$4D,$CC,$7F,$7D,$EF,$57,$CD,$F0,$FF
	.dc.b	$31,$E5,$5A,$FC,$A3,$F4,$7C,$4C,$C7,$CC,$9F,$CF,$79,$9F,$57,$D1
	.dc.b	$FF,$0B,$7E,$09,$FE,$1B,$F3,$7F,$D0,$DE,$43,$A7,$F9,$0E,$3F,$E8
	.dc.b	$5F,$89,$5E,$E3,$FD,$17,$2F,$4B,$9E,$1D,$67,$FC,$F5,$5F,$E1,$E7
	.dc.b	$EA,$7A,$3E,$0F,$37,$F0,$CD,$8F,$C0,$FA,$35,$D9,$CB,$CE,$FC,$67
	.dc.b	$9F,$E8,$F6,$BF,$1D,$E9,$7A,$BD,$AC,$DF,$81,$CB,$E2,$9F,$F4,$F7
	.dc.b	$B9,$FE,$38,$FC,$35,$F0,$1F,$0B,$D7,$CF,$07,$C2,$9C,$E1,$EA,$F6
	.dc.b	$FF,$39,$F1,$3F,$21,$DC,$7F,$AB,$27,$6F,$F8,$CF,$89,$87,$E8,$FD
	.dc.b	$FF,$9F,$EE,$5A,$F8,$AF,$E9,$FD,$FC,$C7,$E8,$A7,$EA,$FB,$BF,$AB
	.dc.b	$F7,$FE,$BF,$C4,$5A,$F9,$BF,$11,$FA,$FF,$81,$FB,$6E,$63,$C5,$3E
	.dc.b	$6E,$E6,$CF,$82,$FC,$0F,$C3,$E6,$31,$7F,$62,$31,$F0,$9A,$8F,$F6
	.dc.b	$57,$37,$C9,$BF,$C4,$FC,$3F,$1F,$89,$FC,$5C,$73,$F8,$9B,$7C,$3A
	.dc.b	$3E,$65,$62,$18,$7D,$3A,$FC,$9E,$3E,$3F,$88,$FC,$AD,$F9,$7E,$27
	.dc.b	$F2,$F5,$9C,$3F,$1A,$E5,$FC,$7C,$FA,$F9,$BF,$3B,$76,$7F,$26,$FB
	.dc.b	$F9,$33,$1E,$53,$ED,$4F,$E9,$30,$F8,$9C,$3F,$A6,$EB,$B5,$FE,$D5
	.dc.b	$FC,$1B,$FC,$41,$B7,$E1,$F9,$3F,$D0,$11,$89,$FC,$25,$F1,$7B,$72
	.dc.b	$7D,$2E,$3D,$F3,$FD,$16,$CE,$B9,$5C,$B8,$8F,$8F,$9F,$E5,$6B,$BF
	.dc.b	$47,$D9,$F1,$3C,$3D,$3F,$6F,$C2,$F1,$F5,$67,$1F,$43,$97,$D5,$3F
	.dc.b	$81,$EF,$F4,$F9,$1F,$83,$F0,$6D,$78,$E6,$BE,$23,$F4,$11,$F0,$3C
	.dc.b	$6F,$C5,$77,$FC,$2F,$1A,$D7,$C7,$7F,$1D,$E1,$66,$3D,$D3,$F2,$3C
	.dc.b	$6F,$CB,$74,$FF,$1E,$9F,$9D,$E3,$63,$CD,$F9,$AE,$BC,$FB,$3F,$32
	.dc.b	$B8,$3F,$EA,$3E,$BF,$F5,$9B,$9F,$3B,$EA,$FA,$BD,$9F,$AF,$EA,$F8
	.dc.b	$3C,$EB,$5B,$3E,$F5,$7E,$97,$AB,$EE,$FA,$76,$7F,$04,$FE,$A7,$E2
	.dc.b	$B3,$1F,$66,$7F,$07,$EA,$7B,$3D,$DF,$C2,$FA,$76,$BF,$15,$EA,$7C
	.dc.b	$0F,$6F,$F1,$2E,$63,$D5,$3F,$13,$DB,$FC,$6F,$C0,$D9,$FC,$93,$F2
	.dc.b	$3D,$BC,$C7,$C3,$9F,$C9,$FC,$07,$CC,$F6,$FF,$2B,$EF,$DA,$FC,$D7
	.dc.b	$BF,$F4,$3D,$EF,$CC,$B9,$8F,$95,$3F,$53,$E0,$FF,$DB,$9F,$09,$FE
	.dc.b	$26,$EA,$38,$AD,$03,$FD,$A8,$EC,$E3,$F0,$3F,$7B,$E1,$79,$FC,$4D
	.dc.b	$AE,$07,$F4,$FA,$47,$E1,$3E,$1B,$73,$F7,$EB,$C7,$F1,$3F,$88,$F8
	.dc.b	$6F,$2F,$C4,$DA,$E6,$78,$FE,$1B,$CF,$1F,$8E,$F8,$8D,$CF,$E2,$6B
	.dc.b	$DB,$F1,$3F,$92,$8F,$7E,$F5,$E6,$30,$F8,$3A,$FE,$63,$6F,$8A,$EE
	.dc.b	$7E,$0D,$7E,$73,$6F,$93,$AF,$E7,$B5,$B5,$F9,$99,$AF,$EB,$7D,$9F
	.dc.b	$EA,$B3,$F4,$37,$FD,$2E,$FF,$4B,$6B,$7E,$4D,$7E,$9E,$FF,$AA,$DF
	.dc.b	$EB,$BA,$9F,$A7,$5F,$D0,$FF,$0E,$75,$7D,$CE,$DC,$36,$BF,$8B,$CB
	.dc.b	$B3,$11,$E5,$8E,$2F,$87,$89,$DB,$D7,$B7,$F0,$FD,$96,$B5,$DB,$E1
	.dc.b	$ED,$FC,$5B,$98,$FC,$3C,$F9,$7A,$BF,$1F,$DB,$E7,$EC,$B5,$E4,$9A
	.dc.b	$F8,$E3,$F2,$BD,$BE,$DF,$23,$F2,$FE,$4D,$AF,$33,$F0,$7E,$23,$31
	.dc.b	$EE,$9F,$CE,$F9,$5F,$17,$C6,$FC,$FF,$93,$F1,$FC,$9F,$D0,$F9,$39
	.dc.b	$C3,$E0,$39,$7E,$14,$FC,$DF,$27,$F0,$F1,$F8,$7F,$C2,$3F,$C5,$1E
	.dc.b	$13,$A1,$5A,$EF,$57,$F8,$9F,$E1,$9E,$57,$F8,$D2,$DE,$56,$EB,$53
	.dc.b	$9C,$30,$FD,$DF,$57,$5E,$C7,$2D,$3E,$C7,$AF,$C7,$D8,$C6,$99,$DB
	.dc.b	$C9,$87,$0C,$7F,$03,$7B,$7F,$07,$EC,$78,$7D,$8E,$31,$E5,$F5,$F0
	.dc.b	$EB,$3F,$CA,$6C,$F6,$37,$0C,$FE,$15,$F0,$8F,$C6,$7B,$1E,$DF,$5F
	.dc.b	$75,$F1,$E3,$18,$FE,$53,$D7,$F8,$5E,$C6,$B9,$55,$AC,$7F,$55,$F8
	.dc.b	$7B,$E1,$3E,$47,$C1,$F4,$8C,$4F,$9B,$E1,$31,$F0,$9A,$D3,$FB,$BD
	.dc.b	$6A,$BF,$85,$F5,$FD,$6B,$E5,$ED,$FD,$3F,$83,$F1,$7C,$55,$C7,$8C
	.dc.b	$FD,$7F,$85,$CE,$3F,$12,$FD,$1F,$0B,$F6,$FE,$0F,$D5,$F0,$7F,$77
	.dc.b	$E1,$33,$7F,$2A,$72,$FC,$C8,$F5,$47,$E0,$FE,$27,$D9,$17,$3F,$4C
	.dc.b	$FE,$1F,$E2,$73,$7F,$52,$FE,$11,$F8,$BF,$88,$B1,$F6,$1F,$8D,$3F
	.dc.b	$8F,$F8,$9C,$CF,$C2,$EE,$57,$CB,$3F,$29,$CD,$F3,$6A,$E5,$F2,$A7
	.dc.b	$F3,$1C,$B9,$C3,$F0,$B7,$F4,$FB,$3F,$39,$C5,$67,$F1,$2F,$D7,$EA
	.dc.b	$FD,$37,$E1,$4E,$5D,$57,$8E,$BE,$E7,$4E,$B9,$3E,$F7,$3F,$0E,$5C
	.dc.b	$DF,$F2,$7C,$55,$A8,$C4,$FE,$17,$9B,$BF,$87,$67,$87,$87,$F8,$9E
	.dc.b	$9C,$F1,$FE,$1A,$FC,$B3,$F8,$D8,$F3,$F8,$76,$FC,$0F,$E4,$7E,$23
	.dc.b	$D7,$E1,$66,$7F,$1B,$AA,$BF,$2D,$E3,$FC,$0F,$02,$D7,$A5,$FC,$D7
	.dc.b	$8F,$F1,$3C,$3C,$CF,$29,$FC,$F7,$8D,$F2,$3C,$FB,$3F,$05,$FD,$17
	.dc.b	$8F,$F3,$3D,$0C,$CF,$8A,$7F,$63,$F8,$83,$D6,$FA,$3E,$8E,$1B,$C5
	.dc.b	$FE,$5C,$97,$A9,$F5,$BD,$2D,$63,$F3,$F0,$E7,$EA,$7D,$AF,$42,$CF
	.dc.b	$D3,$F4,$7B,$FA,$99,$CF,$A5,$FE,$02,$7C,$7E,$9F,$E0,$BD,$EB,$87
	.dc.b	$DB,$F7,$BC,$FE,$B6,$71,$F2,$5F,$E2,$27,$D7,$F0,$1F,$8A,$F7,$7D
	.dc.b	$BF,$03,$F8,$CF,$7B,$39,$7E,$16,$72,$FE,$1E,$3E,$1F,$C0,$FE,$4B
	.dc.b	$DD,$F8,$BF,$01,$F9,$4F,$7B,$37,$F8,$D9,$CB,$F8,$F8,$F9,$7E,$FF
	.dc.b	$F3,$FF,$C5,$3F,$C4,$DB,$FC,$9F,$81,$C5,$E8,$FE,$F9,$2F,$8A,$FA
	.dc.b	$73,$AA,$FA,$1F,$0B,$CF,$E2,$3E,$BB,$D7,$E2,$3E,$CF,$C3,$E6,$B7
	.dc.b	$97,$49,$FB,$BF,$15,$E3,$BB,$3E,$07,$F0,$3F,$11,$E7,$C3,$33,$FC
	.dc.b	$8B,$45,$7E,$1B,$E2,$BD,$98,$5A,$F4,$E1,$F8,$AF,$8A,$CE,$7F,$87
	.dc.b	$BF,$85,$3F,$CF,$B8,$1C,$3F,$16,$71,$16,$BF,$D5,$2E,$B0,$F8,$B8
	.dc.b	$69,$F8,$F5,$F9,$87,$E7,$73,$D7,$E5,$F3,$FE,$73,$8B,$38,$FE,$66
	.dc.b	$7E,$A4,$FE,$83,$87,$EB,$72,$FE,$8B,$8A,$DF,$E7,$E7,$ED,$72,$E7
	.dc.b	$0F,$AF,$7E,$BE,$FF,$BD,$CD,$67,$ED,$F3,$7B,$FB,$73,$97,$B2,$FF
	.dc.b	$09,$3F,$0F,$B7,$F0,$DE,$27,$C5,$EC,$FC,$47,$87,$9C,$BF,$0B,$39
	.dc.b	$7F,$0F,$1F,$2F,$B7,$F1,$DE,$07,$CD,$EE,$BA,$7C,$A9,$FC,$97,$83
	.dc.b	$9C,$3F,$1F,$7F,$4F,$C9,$FC,$B7,$83,$F5,$7C,$AB,$8F,$D2,$9F,$CD
	.dc.b	$78,$79,$C3,$F2,$F7,$F6,$FC,$7F,$F0,$E5,$7F,$86,$AF,$EC,$C6,$3D
	.dc.b	$0F,$C0,$79,$3A,$9E,$18,$7F,$96,$ED,$5F,$85,$F1,$F9,$F9,$B6,$B4
	.dc.b	$FE,$27,$C9,$EF,$E6,$E6,$7F,$BB,$E0,$AF,$C7,$7A,$DE,$4F,$3B,$F2
	.dc.b	$1E,$B5,$D3,$F1,$B3,$E8,$F3,$F3,$8F,$8E,$FF,$2D,$EB,$7B,$3D,$CF
	.dc.b	$CC,$7A,$BE,$EF,$73,$37,$E9,$9C,$BE,$B8,$FC,$F7,$A9,$F1,$3D,$AF
	.dc.b	$D0,$7A,$F7,$5F,$CE,$CF,$C8,$F6,$B3,$87,$C3,$BF,$F1,$1F,$E2,$5F
	.dc.b	$DC,$23,$E3,$7B,$58,$F8,$0F,$9F,$EE,$6B,$E0,$7E,$8C,$E6,$BE,$64
	.dc.b	$F2,$F8,$0F,$AB,$F0,$BD,$3E,$0A,$ED,$FE,$B9,$F8,$BF,$F2,$26,$9F
	.dc.b	$E2,$1C,$7F,$D6,$BF,$11,$FE,$2E,$F8,$8E,$11,$F5,$F1,$F2,$FC,$27
	.dc.b	$E0,$3E,$22,$CF,$DB,$7F,$7D,$F0,$DF,$84,$F8,$8C,$CF,$D6,$9F,$6F
	.dc.b	$07,$E2,$38,$AC,$FE,$0F,$8B,$E0,$F1,$67,$83,$CC,$7F,$B3,$FE,$2D
	.dc.b	$C0,$C7,$F2,$37,$F2,$36,$FC,$9F,$1D,$CB,$F1,$CF,$CC,$E0,$FC,$BF
	.dc.b	$1E,$67,$F1,$33,$F4,$36,$FC,$DF,$0D,$9F,$CB,$3F,$53,$83,$FC,$F5
	.dc.b	$F8,$AF,$F8,$57,$7F,$2D,$7F,$97,$F6,$61,$CA,$73,$BF,$A3,$B7,$8F
	.dc.b	$7F,$DE,$EF,$B5,$AC,$15,$DB,$C0,$D6,$5F,$EC,$1C,$FF,$C3,$77,$BE
	.dc.b	$EF,$1F,$81,$F8,$87,$CB,$E0,$7E,$2B,$C0,$CD,$6F,$0E,$53,$F8,$EE
	.dc.b	$DF,$5F,$7D,$9E,$DE,$07,$E4,$BC,$0C,$E1,$E5,$BF,$83,$3F,$E7,$8F
	.dc.b	$C5,$DE,$57,$C4,$F1,$31,$E5,$FC,$6F,$33,$53,$9C,$7E,$1D,$AB,$51
	.dc.b	$F0,$4F,$9B,$E6,$7E,$8B,$CB,$F9,$FE,$67,$E9,$27,$27,$C7,$8C,$BF
	.dc.b	$27,$C9,$FA,$BE,$67,$EA,$9B,$1F,$37,$CB,$FB,$3E,$66,$78,$BE,$65
	.dc.b	$FA,$FD,$8F,$BB,$EB,$FB,$7D,$7F,$BF,$EC,$66,$FE,$94,$E5,$FA,$91
	.dc.b	$F0,$FD,$6F,$C2,$FB,$16,$7E,$C3,$F1,$FD,$7F,$C4,$FB,$19,$9F,$F4
	.dc.b	$7F,$81,$FE,$F6,$DF,$E2,$FA,$BF,$8E,$F5,$FE,$73,$63,$E4,$FA,$FF
	.dc.b	$92,$F6,$33,$87,$E1,$6F,$EA,$7C,$07,$E5,$BE,$0E,$CF,$E2,$5F,$B1
	.dc.b	$F0,$5F,$9A,$F8,$3C,$CF,$8A,$7E,$E7,$C1,$FF,$B5,$FF,$11,$FF,$0C
	.dc.b	$CB,$DE,$FF,$65,$78,$0F,$83,$D4,$F0,$07,$FA,$AA,$BF,$C3,$9B,$FE
	.dc.b	$02,$76,$FE,$0B,$E0,$3B,$7C,$17,$E2,$DB,$1F,$86,$F8,$2F,$17,$C1
	.dc.b	$E7,$1E,$A7,$F1,$6C,$58,$7E,$4C,$F4,$7C,$3F,$E5,$6A,$E1,$F9,$09
	.dc.b	$F6,$7C,$4E,$70,$E6,$7A,$3E,$15,$8F,$F8,$82,$FF,$2F,$8F,$C3,$F8
	.dc.b	$8F,$D0,$36,$3F,$37,$87,$C7,$F8,$8C,$E3,$D0,$63,$D3,$0F,$D4,$1F
	.dc.b	$37,$97,$F5,$55,$70,$FD,$2C,$FD,$1E,$5C,$E1,$CF,$0F,$2E,$1F,$EE
	.dc.b	$8F,$C6,$5C,$FF,$5F,$97,$1D,$1F,$67,$93,$53,$9D,$3F,$89,$DA,$B5
	.dc.b	$18,$E7,$FC,$07,$2F,$56,$C7,$0E,$8F,$C2,$73,$67,$6F,$27,$23,$0F
	.dc.b	$E0,$91,$F8,$9F,$0F,$F8,$37,$C4,$FE,$2F,$C4,$CC,$F0,$9C,$BC,$A3
	.dc.b	$D9,$E2,$FE,$47,$C3,$B3,$DB,$C2,$F8,$1E,$36,$76,$F5,$DF,$E5,$E7
	.dc.b	$E2,$78,$DF,$99,$F0,$AE,$1E,$67,$E4,$78,$DF,$9D,$F0,$F3,$3F,$97
	.dc.b	$9F,$99,$E3,$7E,$87,$CD,$B3,$EE,$F3,$7E,$87,$91,$9D,$7E,$5D,$FE
	.dc.b	$9E,$7E,$A7,$A7,$F8,$1B,$FC,$0F,$E2,$7F,$E2,$57,$F5,$57,$FD,$6F
	.dc.b	$AF,$FD,$85,$1F,$76,$75,$E4,$AB,$57,$CF,$C1,$57,$F8,$A2,$7E,$DF
	.dc.b	$A5,$98,$E3,$1D,$7D,$2C,$79,$DF,$87,$75,$E7,$5C,$FC,$13,$F8,$BF
	.dc.b	$4B,$37,$F8,$6F,$4B,$A7,$B5,$F8,$F6,$C6,$6F,$D1,$EE,$7E,$4F,$D4
	.dc.b	$CC,$7B,$87,$BD,$FD,$1B,$E0,$7C,$DE,$DF,$E6,$7E,$2A,$E3,$F9,$47
	.dc.b	$D5,$EE,$7C,$1C,$39,$61,$C5,$F8,$D7,$FA,$1F,$80,$F8,$1E,$DF,$E8
	.dc.b	$FD,$FB,$97,$E7,$9F,$89,$EE,$EF,$0F,$9D,$5F,$A9,$F7,$FE,$47,$C4
	.dc.b	$59,$FA,$4D,$7F,$82,$F4,$FD,$0F,$89,$FD,$87,$C0,$F0,$CF,$FB,$EF
	.dc.b	$85,$FE,$00,$BF,$B5,$F1,$1E,$9E,$FF,$57,$0F,$F6,$6F,$85,$FA,$D3
	.dc.b	$9C,$3F,$C3,$98,$27,$ED,$7C,$46,$71,$F9,$33,$D5,$F1,$F7,$B0,$D6
	.dc.b	$3F,$DC,$BE,$2B,$F0,$5F,$0F,$8C,$1F,$15,$F8,$5C,$F0,$EB,$4F,$4D
	.dc.b	$E5,$F5,$FC,$4F,$E2,$E3,$97,$C5,$7E,$36,$AE,$BF,$97,$9E,$D5,$9B
	.dc.b	$F8,$7C,$BF,$93,$8F,$17,$35,$9F,$8E,$FE,$5F,$1F,$35,$67,$0F,$BF
	.dc.b	$9F,$F8,$F5,$FD,$27,$67,$AF,$93,$F4,$5D,$7F,$E3,$19,$DC,$7E,$86
	.dc.b	$7E,$07,$3E,$BB,$3E,$17,$1F,$0E,$BB,$87,$F2,$BC,$7F,$C5,$6F,$82
	.dc.b	$7A,$F5,$FC,$9E,$7E,$1C,$1F,$7B,$C2,$F0,$F6,$7C,$DF,$0E,$CF,$E0
	.dc.b	$9F,$2F,$67,$D1,$AC,$E5,$EC,$F0,$BD,$3E,$37,$D5,$7D,$7E,$35,$9F
	.dc.b	$24,$FD,$9F,$0F,$39,$7E,$3B,$C2,$F8,$3E,$37,$DD,$E1,$B3,$F9,$27
	.dc.b	$E2,$F8,$DF,$81,$AC,$E3,$FA,$AF,$C6,$DE,$67,$CA,$F2,$3F,$13,$18
	.dc.b	$9F,$89,$E4,$62,$B5,$97,$F9,$6A,$9E,$6F,$D2,$AE,$3D,$BF,$53,$D1
	.dc.b	$B8,$FC,$D6,$BF,$AA,$D5,$E9,$78,$3C,$DE,$19,$FE,$A8,$FB,$BE,$8F
	.dc.b	$E6,$7E,$2B,$EF,$FA,$57,$1F,$B1,$3F,$9D,$AC,$9E,$5F,$63,$F0,$BE
	.dc.b	$97,$F8,$1B,$D0,$FD,$8F,$0F,$F5,$5C,$3F,$01,$E9,$E7,$0F,$F4,$64
	.dc.b	$BD,$AF,$C6,$CE,$6B,$F8,$CC,$F1,$7D,$58,$67,$D1,$4B,$D8,$FC,$AF
	.dc.b	$BD,$98,$C5,$FE,$3F,$DE,$C7,$B5,$F9,$A7,$57,$C2,$7E,$77,$B5,$F9
	.dc.b	$EF,$7B,$97,$B7,$FA,$0F,$7B,$A4,$DA,$FA,$93,$B7,$EB,$7C,$1F,$F4
	.dc.b	$7F,$C5,$BF,$C7,$3C,$7F,$0A,$31,$EF,$FE,$03,$E1,$75,$7F,$82,$9C
	.dc.b	$F2,$FD,$FF,$84,$E5,$F1,$1F,$86,$F8,$6B,$39,$BF,$F0,$A6,$6F,$89
	.dc.b	$D4,$62,$FF,$1B,$F0,$FE,$3F,$89,$B3,$C3,$E2,$7F,$23,$F1,$39,$CB
	.dc.b	$F1,$D7,$E9,$9F,$CA,$D7,$AF,$5F,$98,$D3,$F2,$9B,$D7,$F3,$C7,$C1
	.dc.b	$DF,$E9,$67,$F0,$F7,$B3,$E8,$7E,$B6,$B9,$8F,$CF,$4F,$C9,$DF,$ED
	.dc.b	$6D,$F2,$F4,$FB,$9A,$5B,$FD,$24,$D7,$F5,$D1,$F4,$76,$FC,$37,$E3
	.dc.b	$AE,$CF,$A9,$DF,$FC,$E2,$31,$7F,$2B,$BF,$1D,$FA,$E3,$FF,$64,$D6
	.dc.b	$AB,$FA,$9F,$27,$E5,$E7,$3D,$DF,$7F,$AF,$A7,$67,$E0,$7B,$7B,$4E
	.dc.b	$6B,$9C,$E5,$EB,$DB,$F8,$7E,$CF,$23,$63,$C3,$DB,$F8,$BE,$DC,$E9
	.dc.b	$F8,$0B,$F5,$79,$7F,$8F,$F2,$FD,$FE,$57,$E4,$7C,$BB,$7F,$86,$7E
	.dc.b	$1D,$6E,$3D,$BE,$57,$E5,$FC,$BF,$8F,$E5,$7E,$67,$CA,$B5,$F8,$E7
	.dc.b	$E5,$D6,$E3,$E2,$F9,$1F,$E2,$5F,$C6,$BF,$C1,$11,$FE,$08,$E7,$4E
	.dc.b	$78,$7E,$9B,$67,$E6,$F9,$5C,$7C,$BF,$AF,$EB,$F3,$F5,$6D,$7F,$9C
	.dc.b	$72,$7A,$FF,$73,$D7,$F0,$3A,$9F,$F0,$3C,$AB,$F0,$3E,$AF,$97,$D7
	.dc.b	$B5,$CD,$FC,$2F,$AF,$98,$FB,$F3,$EB,$F5,$FF,$13,$EB,$FB,$7D,$6B
	.dc.b	$5E,$37,$F1,$BE,$BE,$63,$F0,$F3,$F0,$FD,$7F,$E2,$DF,$8F,$3E,$13
	.dc.b	$E3,$7C,$17,$F8,$22,$1F,$07,$8E,$7F,$F5,$CF,$B3,$FE,$18,$9E,$19
	.dc.b	$7C,$1F,$84,$E1,$7B,$8F,$F1,$AC,$BE,$0F,$E9,$7C,$1F,$56,$E3,$F9
	.dc.b	$AF,$83,$FA,$DF,$09,$9C,$3E,$65,$FE,$8F,$E1,$3E,$D7,$C2,$59,$FA
	.dc.b	$0F,$E9,$FE,$13,$EF,$7C,$26,$67,$E7,$CF,$AC,$FC,$17,$C4,$7C,$06
	.dc.b	$C7,$B3,$3F,$F6,$AF,$C2,$B6,$61,$AA,$F5,$DF,$E2,$FE,$23,$E4,$E3
	.dc.b	$F8,$DF,$87,$B5,$F8,$3F,$88,$F9,$B5,$F9,$17,$31,$F8,$29,$FA,$27
	.dc.b	$E5,$79,$AC,$7E,$2F,$8F,$EA,$D7,$E6,$5C,$C7,$E2,$A7,$EC,$F5,$7E
	.dc.b	$77,$93,$EC,$3F,$E8,$3F,$C7,$3F,$C4,$5C,$FF,$5F,$A7,$5C,$5E,$FC
	.dc.b	$7F,$CE,$3C,$2C,$38,$73,$FE,$07,$A7,$97,$37,$E0,$FA,$7A,$73,$E7
	.dc.b	$1F,$F6,$86,$BF,$E2,$6D,$7F,$09,$5A,$D3,$B7,$3F,$CB,$E8,$F2,$78
	.dc.b	$9F,$8D,$EA,$B5,$AE,$0F,$F1,$E6,$7F,$E6,$CC,$DE,$26,$E3,$BE,$1F
	.dc.b	$95,$BF,$77,$7F,$E6,$9B,$3C,$DA,$FE,$4B,$C5,$F8,$9E,$0F,$E8,$3C
	.dc.b	$3F,$8D,$E0,$7E,$8B,$C5,$F9,$1E,$15,$C3,$DA,$ED,$F7,$CF,$EA,$3C
	.dc.b	$4F,$9D,$E6,$D9,$F8,$7E,$7D,$7D,$4F,$EB,$BC,$6D,$C7,$C4,$F0,$1E
	.dc.b	$8F,$D7,$F3,$3F,$C1,$10,$F4,$B1,$5B,$6B,$FE,$A0,$8E,$2F,$D2,$F3
	.dc.b	$F8,$FA,$5F,$81,$F3,$B9,$FA,$56,$BC,$0D,$79,$4F,$E1,$FC,$FF,$0F
	.dc.b	$A5,$F8,$9F,$76,$D7,$DE,$6B,$F5,$5F,$3F,$A7,$F8,$FF,$6F,$D3,$F0
	.dc.b	$3F,$91,$F6,$7D,$7E,$FD,$C3,$E0,$35,$F5,$4F,$E5,$FD,$AF,$83,$EF
	.dc.b	$FE,$67,$DB,$F8,$7E,$FD,$BF,$90,$D7,$E1,$4F,$E7,$FD,$CF,$93,$EF
	.dc.b	$7F,$B8,$FE,$15,$58,$29,$3E,$2D,$0F,$88,$D4,$FF,$AF,$23,$0F,$D7
	.dc.b	$F8,$1E,$8F,$D6,$F8,$4E,$DF,$0D,$63,$A9,$E0,$C3,$AC,$66,$3A,$3F
	.dc.b	$7B,$E1,$7C,$DF,$0F,$F8,$0F,$89,$F4,$61,$F8,$2F,$89,$F5,$55,$C3
	.dc.b	$A8,$C7,$D3,$1C,$63,$31,$EC,$AF,$C5,$E7,$F0,$F0,$B5,$A8,$46,$23
	.dc.b	$E0,$3F,$92,$F8,$9F,$95,$8F,$F9,$6F,$F2,$07,$1F,$CD,$C3,$1C,$37
	.dc.b	$1F,$DF,$D6,$9E,$D3,$A7,$C1,$87,$E7,$9F,$AD,$CB,$FA,$0E,$1B,$1F
	.dc.b	$A5,$29,$F9,$F7,$EE,$73,$7E,$9B,$87,$EF,$72,$5A,$F6,$72,$FE,$AB
	.dc.b	$8F,$F0,$4E,$63,$EE,$4F,$AF,$BF,$F0,$DC,$DF,$EE,$6D,$BF,$C0,$75
	.dc.b	$F1,$63,$1D,$BF,$8B,$E7,$D5,$F0,$F3,$FD,$FD,$BF,$8F,$F1,$79,$76
	.dc.b	$FE,$47,$C4,$E9,$39,$AF,$8B,$39,$7E,$3F,$77,$E5,$FC,$1F,$17,$7F
	.dc.b	$E6,$7C,$2F,$24,$E6,$BE,$6C,$E5,$F9,$FE,$57,$E7,$FC,$3F,$AC,$FE
	.dc.b	$87,$C3,$BC,$5F,$7D,$FB,$5E,$5E,$E3,$EB,$F9,$5F,$8C,$7F,$19,$F8
	.dc.b	$F7,$F8,$9B,$DA,$65,$FD,$93,$9D,$7B,$F1,$97,$BA,$D4,$E6,$31,$83
	.dc.b	$D9,$60,$F3,$75,$E9,$7E,$37,$CD,$E1,$E8,$DC,$3F,$37,$8F,$F8,$B1
	.dc.b	$FC,$6C,$F5,$F4,$7F,$29,$E7,$70,$9F,$C8,$FB,$9E,$1F,$4B,$F3,$1E
	.dc.b	$E5,$9C,$9D,$23,$2F,$6F,$4F,$F3,$DE,$B7,$A7,$DF,$B3,$E2,$F8,$0F
	.dc.b	$D1,$7B,$3E,$D7,$31,$FC,$03,$9B,$E1,$D7,$E9,$FD,$9F,$85,$EF,$FE
	.dc.b	$A7,$DA,$B7,$FA,$69,$AB,$AB,$E7,$BF,$1F,$9D,$5F,$2F,$DF,$FF,$66
	.dc.b	$7E,$42,$F8,$5F,$9D,$F0,$38,$F8,$5F,$A1,$F0,$5A,$9B,$97,$F7,$2B
	.dc.b	$57,$08,$C7,$C2,$FD,$7F,$88,$EB,$F0,$5F,$67,$E2,$7B,$CD,$AE,$53
	.dc.b	$B7,$A7,$C3,$FD,$FF,$89,$F2,$B6,$3F,$81,$FC,$45,$7D,$DE,$3F,$F8
	.dc.b	$BB,$4F,$F2,$24,$7D,$EF,$89,$F6,$F0,$7E,$27,$7F,$87,$C9,$F1,$3C
	.dc.b	$76,$1F,$E0,$6C,$3F,$9D,$78,$9F,$0F,$2F,$76,$3F,$91,$E2,$E3,$1D
	.dc.b	$F8,$FA,$3F,$0F,$8F,$F2,$DC,$5A,$C7,$F2,$D9,$7F,$90,$1F,$C8,$F1
	.dc.b	$67,$4F,$0F,$1F,$0C,$79,$56,$9F,$9D,$C7,$FA,$1E,$3E,$18,$B8,$BF
	.dc.b	$51,$F9,$07,$F8,$D2,$FE,$14,$6D,$E3,$87,$F3,$A9,$D3,$E4,$F0,$38
	.dc.b	$F8,$5F,$82,$79,$F8,$56,$78,$3F,$86,$F0,$B7,$1F,$81,$EE,$F0,$F8
	.dc.b	$5F,$8A,$7C,$7E,$15,$9F,$03,$F8,$EF,$0B,$71,$F8,$9F,$07,$D3,$E0
	.dc.b	$FE,$49,$F5,$F8,$16,$7D,$0F,$E5,$BC,$0D,$C7,$F9,$0B,$F2,$27,$9D
	.dc.b	$F0,$BC,$EF,$99,$0F,$3B,$E2,$F9,$D8,$AD,$72,$F1,$9C,$BF,$9B,$3E
	.dc.b	$5F,$9D,$F3,$9A,$7E,$7F,$CD,$F9,$FE,$76,$73,$F9,$95,$F5,$3C,$EF
	.dc.b	$A7,$E7,$7D,$6F,$3B,$EA,$F9,$D9,$BF,$A5,$39,$7E,$A4,$7A,$3D,$AF
	.dc.b	$B7,$EC,$59,$FB,$1E,$D5,$7F,$82,$77,$FF,$B3,$B6,$F8,$5E,$D7,$E0
	.dc.b	$AB,$E0,$FB,$5F,$84,$F6,$6E,$1F,$81,$F6,$6B,$CA,$3E,$3F,$B5,$F8
	.dc.b	$A9,$F9,$3E,$D7,$E3,$3D,$9B,$7F,$89,$F6,$6B,$E2,$8F,$9F,$ED,$7E
	.dc.b	$4A,$7E,$8F,$C1,$FE,$53,$E1,$7F,$C4,$BE,$C7,$EB,$FB,$9F,$09,$9C
	.dc.b	$BE,$D5,$D7,$F7,$67,$F3,$D9,$DB,$F9,$9F,$85,$E7,$C7,$FE,$AC,$97
	.dc.b	$C2,$FF,$49,$FC,$8B,$FC,$4F,$97,$78,$C7,$C3,$70,$C7,$A7,$C3,$61
	.dc.b	$FC,$54,$F1,$75,$EB,$B1,$7C,$1F,$5F,$86,$CC,$62,$BF,$C4,$B9,$A7
	.dc.b	$3F,$0B,$B9,$FC,$B5,$F9,$7E,$1B,$8D,$79,$FE,$27,$9C,$5A,$FC,$E3
	.dc.b	$5F,$87,$97,$8E,$BD,$DF,$15,$E2,$BF,$81,$F1,$5E,$48,$B7,$FA,$56
	.dc.b	$BC,$B4,$57,$C7,$F8,$AF,$5D,$FC,$9E,$4B,$7E,$C6,$BF,$D4,$F6,$FF
	.dc.b	$2E,$BF,$B1,$AF,$A1,$CB,$F2,$39,$9D,$3F,$4F,$8B,$F8,$F4,$3A,$31
	.dc.b	$CF,$FE,$BA,$8D,$E3,$F2,$B9,$77,$5A,$8F,$83,$3F,$E4,$17,$C5,$C9
	.dc.b	$D2,$3F,$01,$CB,$DB,$A3,$F0,$53,$9A,$FA,$B3,$FC,$11,$FC,$37,$87
	.dc.b	$FC,$1B,$A3,$F1,$1E,$0E,$67,$ED,$CE,$5F,$BB,$1E,$AF,$17,$F1,$DE
	.dc.b	$0F,$B3,$C5,$FC,$87,$83,$6F,$F0,$73,$B7,$F0,$B1,$F0,$BC,$5F,$CB
	.dc.b	$78,$36,$7F,$12,$D7,$E7,$78,$3F,$23,$C4,$DF,$27,$C7,$F1,$3F,$3D
	.dc.b	$E6,$7C,$CF,$16,$CF,$CB,$7F,$45,$E6,$7D,$0A,$CE,$3F,$2B,$D2,$FF
	.dc.b	$17,$79,$DF,$E2,$AF,$1B,$FC,$E3,$18,$F3,$BE,$D7,$A3,$6F,$EB,$4F
	.dc.b	$0C,$FE,$F7,$A1,$CB,$CE,$FC,$06,$5D,$3C,$EC,$E1,$FE,$88,$8F,$F1
	.dc.b	$73,$F6,$9E,$0F,$E0,$7C,$FF,$1F,$B3,$F8,$9F,$47,$CB,$EC,$7E,$2E
	.dc.b	$6D,$7E,$1E,$76,$FF,$AC,$BD,$1F,$EF,$5F,$91,$FF,$80,$23,$FC,$51
	.dc.b	$8B,$D8,$FD,$DC,$E6,$7F,$D6,$D2,$F5,$F3,$EF,$7C,$47,$F3,$5F,$13
	.dc.b	$F1,$B0,$FC,$E7,$BB,$F2,$3D,$AF,$CF,$7B,$B7,$2F,$CD,$35,$FC,$BC
	.dc.b	$FC,$EF,$6F,$F4,$9E,$EF,$D0,$F8,$7F,$D3,$7B,$DF,$4B,$E1,$AE,$1F
	.dc.b	$1D,$DB,$F2,$67,$D3,$F0,$5F,$63,$E1,$AC,$FC,$DF,$84,$F6,$FC,$1F
	.dc.b	$DC,$73,$1F,$0A,$7E,$0F,$C4,$FE,$02,$FE,$1F,$C4,$5A,$F9,$5D,$FF
	.dc.b	$ED,$CF,$0D,$D1,$FE,$43,$9F,$B5,$7C,$1F,$B9,$9F,$CC,$F8,$8F,$C5
	.dc.b	$E5,$F3,$BE,$23,$F1,$B9,$E6,$FF,$04,$E5,$FC,$24,$FD,$4E,$5F,$C9
	.dc.b	$DF,$D6,$E3,$FC,$AE,$16,$FF,$14,$ED,$FC,$64,$FD,$CE,$0F,$CD,$F4
	.dc.b	$D9,$FC,$83,$5F,$C1,$BF,$82,$E2,$FF,$47,$7E,$44,$E9,$FC,$2F,$1E
	.dc.b	$39,$EE,$1F,$EE,$0A,$46,$23,$FC,$E9,$18,$7E,$49,$FE,$93,$E2,$FF
	.dc.b	$1F,$3A,$E8,$FC,$8F,$85,$E0,$E9,$B3,$D5,$FC,$AF,$85,$98,$E7,$CF
	.dc.b	$F4,$AB,$CF,$D5,$F9,$AF,$0B,$D3,$E2,$7E,$73,$C3,$B8,$70,$D1,$58
	.dc.b	$8F,$E9,$30,$9F,$83,$E2,$FE,$93,$BF,$E1,$F8,$BF,$A6,$F0,$7E,$2F
	.dc.b	$85,$9C,$7E,$13,$97,$E2,$4F,$F6,$6F,$C9,$1E,$77,$CC,$F1,$F1,$E7
	.dc.b	$7C,$EF,$23,$53,$70,$F6,$DA,$B8,$46,$3C,$DF,$AB,$E9,$F5,$6C,$70
	.dc.b	$F3,$6B,$FD,$92,$FF,$C8,$2F,$D7,$F4,$FF,$7F,$E5,$7D,$EF,$46,$CF
	.dc.b	$DB,$7F,$81,$F9,$D9,$8F,$91,$3F,$84,$F4,$7D,$9E,$C7,$E1,$BD,$1F
	.dc.b	$F3,$B6,$5F,$DF,$EF,$F0,$3E,$96,$3D,$8D,$63,$EC,$C3,$F1,$B3,$C3
	.dc.b	$D7,$FC,$7F,$BD,$70,$FC,$67,$B9,$D3,$D9,$CE,$3F,$12,$FF,$2B,$3E
	.dc.b	$0F,$67,$F2,$FE,$E7,$8B,$DA,$B7,$F2,$A7,$F3,$7E,$E6,$72,$FC,$C5
	.dc.b	$FA,$3E,$0F,$FD,$D9,$F9,$22,$70,$52,$7C,$72,$35,$8F,$D6,$AE,$3E
	.dc.b	$FF,$C2,$F8,$3E,$78,$7C,$4F,$85,$EB,$F1,$36,$3A,$1D,$F2,$E9,$19
	.dc.b	$8F,$F6,$E6,$4F,$8A,$F9,$BF,$0D,$CB,$4F,$C2,$FC,$37,$A3,$E2,$3E
	.dc.b	$8B,$63,$38,$76,$19,$7A,$7E,$2B,$EB,$FC,$46,$E7,$F0,$C7,$C2,$E0
	.dc.b	$FC,$C7,$07,$C4,$E0,$B3,$CF,$0F,$85,$1F,$9D,$8E,$07,$E6,$37,$F9
	.dc.b	$7C,$1F,$A1,$6C,$64,$E5,$C1,$FE,$26,$7F,$2D,$C1,$F4,$B8,$3F,$51
	.dc.b	$C1,$F5,$38,$2D,$70,$6B,$F1,$9F,$D7,$6F,$F6,$B7,$FF,$2A,$FE,$4A
	.dc.b	$EE,$FB,$BD,$98,$EC,$FB,$FD,$BA,$9B,$97,$F0,$CB,$57,$08,$C7,$6F
	.dc.b	$E1,$BB,$3A,$B6,$38,$76,$57,$FC,$A7,$7F,$E3,$87,$ED,$F6,$F9,$3A
	.dc.b	$BF,$1F,$D9,$E6,$EC,$FC,$8F,$6E,$6F,$E2,$B9,$7E,$3C,$FB,$3C,$9F
	.dc.b	$CB,$F9,$1F,$EE,$5C,$9E,$4E,$3C,$0F,$55,$7E,$73,$C9,$D4,$E7,$1F
	.dc.b	$2F,$93,$FA,$0F,$27,$97,$91,$FA,$2F,$27,$A4,$E7,$2F,$89,$39,$7E
	.dc.b	$37,$93,$F8,$DC,$3F,$1B,$F9,$3A,$7C,$9E,$4E,$7C,$64,$BC,$AF,$46
	.dc.b	$19,$8E,$17,$F4,$6F,$15,$ED,$F6,$35,$EB,$7D,$C7,$87,$B1,$73,$FC
	.dc.b	$A6,$3F,$E2,$AF,$8A,$F1,$43,$D9,$72,$BD,$9E,$73,$F7,$FD,$9F,$17
	.dc.b	$B3,$F8,$8F,$5F,$C9,$EC,$5A,$EA,$FE,$32,$B7,$1F,$87,$F5,$7D,$5E
	.dc.b	$C7,$E4,$1F,$67,$B1,$67,$CA,$FE,$53,$D8,$DC,$7F,$5A,$ED,$7C,$17
	.dc.b	$C3,$F8,$5F,$F5,$6C,$3E,$13,$15,$B6,$BF,$EA,$08,$E2,$FC,$2F,$85
	.dc.b	$E3,$F0,$9F,$37,$E1,$7F,$47,$F0,$9F,$3F,$E1,$6D,$FC,$87,$6F,$CA
	.dc.b	$9F,$D4,$FC,$27,$D5,$F8,$5F,$D5,$FC,$27,$D7,$F8,$AB,$7F,$41,$DB
	.dc.b	$F4,$A7,$D2,$7D,$DF,$8A,$F5,$CD,$9F,$80,$D7,$D5,$3F,$83,$F8,$AF
	.dc.b	$83,$5F,$85,$F8,$9F,$87,$16,$BE,$43,$5F,$85,$3F,$8B,$F8,$AF,$93
	.dc.b	$5F,$8D,$E3,$B5,$F8,$6E,$2A,$FF,$32,$F2,$3F,$2B,$1F,$42,$F8,$4F
	.dc.b	$D5,$E8,$FC,$B7,$17,$D4,$6C,$7D,$9E,$7A,$FD,$38,$FD,$07,$37,$F8
	.dc.b	$A6,$3E,$F3,$8E,$1F,$BA,$D8,$FC,$17,$2D,$7E,$E4,$72,$E2,$FC,$1F
	.dc.b	$1F,$4F,$0F,$F0,$BC,$BD,$BB,$B3,$87,$F3,$5C,$3F,$3C,$FE,$36,$78
	.dc.b	$3F,$8F,$F1,$FC,$BD,$9F,$8E,$F1,$AC,$FE,$4D,$AF,$E3,$67,$D7,$DF
	.dc.b	$F9,$4F,$1B,$DB,$DF,$F9,$6F,$1B,$DF,$E0,$5C,$3C,$4E,$DF,$24,$FE
	.dc.b	$73,$C2,$F8,$BE,$75,$9F,$43,$5F,$A9,$B2,$3F,$47,$E1,$FC,$CF,$3B
	.dc.b	$FB,$A7,$8F,$FB,$EF,$1D,$F9,$43,$D1,$FA,$3E,$76,$B8,$BF,$CF,$13
	.dc.b	$9D,$3E,$37,$9D,$97,$F8,$34,$3D,$2F,$B3,$39,$C7,$AE,$3D,$2F,$FC
	.dc.b	$A3,$7F,$E3,$C7,$E8,$79,$FB,$8C,$47,$AB,$CF,$C7,$9B,$F8,$47,$5E
	.dc.b	$97,$E1,$AF,$87,$A5,$F8,$8F,$6A,$E1,$F6,$FD,$AE,$9E,$A7,$E3,$1C
	.dc.b	$C7,$C6,$9F,$07,$BF,$F9,$0F,$6B,$F9,$0F,$27,$C5,$D9,$3F,$94,$F6
	.dc.b	$B3,$C9,$F1,$23,$83,$EC,$F7,$B4,$FB,$BD,$EF,$CD,$FB,$7C,$34,$F8
	.dc.b	$F1,$F9,$EF,$6F,$1C,$DF,$4A,$37,$7F,$1B,$DE,$D3,$F2,$3D,$EF,$E8
	.dc.b	$9F,$95,$7F,$8A,$B1,$6F,$E7,$F7,$F3,$1D,$38,$11,$C3,$E2,$31,$86
	.dc.b	$B1,$FA,$7F,$05,$D3,$E1,$B8,$7C,$1D,$9F,$AE,$F8,$3E,$17,$A5,$67
	.dc.b	$3F,$0D,$7E,$1B,$E2,$7C,$B3,$E1,$F8,$9B,$3F,$80,$7D,$38,$79,$6B
	.dc.b	$37,$C8,$FC,$5F,$C4,$FB,$A3,$D4,$DB,$FC,$4B,$F0,$AF,$DD,$F1,$19
	.dc.b	$BE,$5C,$8F,$CA,$9C,$FF,$23,$0F,$9F,$0E,$7F,$97,$86,$2B,$59,$7F
	.dc.b	$A9,$AB,$58,$FF,$8E,$79,$57,$F3,$63,$E9,$F4,$D7,$4F,$D1,$E3,$FA
	.dc.b	$FD,$39,$9F,$E5,$17,$F8,$CC,$7F,$47,$C7,$F7,$3A,$7E,$C7,$27,$DE
	.dc.b	$E9,$FB,$53,$9A,$FD,$4C,$E5,$F3,$78,$1F,$84,$E7,$FD,$86,$EF,$0B
	.dc.b	$F0,$FD,$38,$6E,$BC,$0F,$65,$70,$EE,$FC,$6F,$89,$67,$E0,$78,$D5
	.dc.b	$F4,$BD,$BC,$1D,$C7,$C7,$F0,$7F,$2B,$E2,$78,$9B,$19,$3E,$54,$79
	.dc.b	$BC,$2F,$CD,$F8,$BE,$8F,$37,$F3,$BE,$37,$AB,$CE,$B8,$7D,$39,$AF
	.dc.b	$DE,$E1,$4F,$E9,$3C,$7F,$83,$E6,$7F,$8A,$7F,$2C,$FF,$0D,$DE,$E3
	.dc.b	$1C,$DF,$1B,$CC,$B3,$F7,$E7,$87,$A7,$F2,$BC,$DE,$5E,$97,$CC,$F3
	.dc.b	$BA,$7A,$59,$C3,$FC,$F9,$9F,$F8,$A7,$3F,$F5,$17,$93,$F9,$1B,$E3
	.dc.b	$5F,$8B,$F6,$FC,$9E,$9F,$D6,$7C,$DE,$97,$D8,$F6,$AD,$6C,$EB,$5F
	.dc.b	$73,$D9,$F6,$7B,$96,$7C,$2F,$E0,$3D,$AC,$C7,$E6,$67,$E1,$7B,$BF
	.dc.b	$84,$F6,$BE,$27,$B3,$6B,$D2,$FE,$23,$DA,$CC,$7E,$86,$7E,$57,$BB
	.dc.b	$FC,$EF,$F2,$C7,$C2,$7C,$DF,$7F,$1F,$03,$F3,$FE,$02,$E3,$F1,$1E
	.dc.b	$15,$B8,$FF,$63,$7C,$07,$F5,$1C,$3F,$C6,$31,$FE,$0B,$E8,$7C,$3F
	.dc.b	$7F,$85,$FB,$5F,$11,$6B,$E7,$BE,$3F,$85,$FB,$D5,$E5,$F8,$5F,$C0
	.dc.b	$7C,$3F,$E8,$BE,$26,$E1,$FB,$B6,$BF,$C1,$B3,$57,$E1,$FE,$27,$DD
	.dc.b	$C3,$F8,$9D,$FE,$07,$05,$C3,$D3,$E0,$7D,$EC,$57,$99,$E9,$5F,$91
	.dc.b	$E2,$F9,$1A,$FE,$4F,$82,$D7,$E2,$37,$F9,$9C,$1F,$97,$73,$1F,$0E
	.dc.b	$7E,$87,$07,$E6,$F7,$B3,$F9,$07,$EA,$6F,$98,$F2,$9F,$E7,$9D,$55
	.dc.b	$FE,$8B,$FC,$AB,$FC,$59,$3F,$07,$B3,$1D,$FF,$D4,$AD,$1B,$EE,$C3
	.dc.b	$AE,$2E,$BE,$0A,$9E,$06,$E7,$30,$F0,$3C,$1D,$FD,$5F,$17,$81,$74
	.dc.b	$FF,$72,$6A,$AF,$17,$81,$AC,$79,$F8,$1E,$9E,$FF,$33,$EB,$F0,$2E
	.dc.b	$3C,$A7,$D5,$E0,$66,$FC,$FC,$CF,$33,$E1,$79,$9F,$DC,$21,$E5,$7C
	.dc.b	$5F,$2F,$15,$AB,$E3,$39,$7D,$A7,$CB,$F2,$BE,$43,$4F,$83,$E5,$7C
	.dc.b	$FF,$2F,$3A,$FB,$AB,$E7,$79,$5F,$4F,$CB,$B3,$F0,$9F,$A5,$E5,$FD
	.dc.b	$7F,$2F,$33,$FE,$6F,$E6,$57,$A7,$D5,$FB,$9E,$C7,$F8,$BF,$85,$1F
	.dc.b	$7F,$D7,$C7,$B1,$71,$F8,$73,$C1,$F8,$BE,$BF,$E1,$7D,$8E,$4F,$E1
	.dc.b	$FD,$7E,$9E,$C5,$AD,$9F,$62,$BC,$1E,$BF,$E3,$7D,$8B,$3F,$71,$F2
	.dc.b	$7A,$F9,$8F,$A7,$3F,$93,$F6,$3D,$1F,$09,$F9,$5F,$83,$FA,$9F,$07
	.dc.b	$F9,$7F,$84,$B8,$7E,$1A,$7E,$C5,$67,$B3,$EE,$FC,$27,$FB,$9F,$0F
	.dc.b	$F1,$27,$B9,$FE,$5D,$E5,$7C,$1E,$72,$FC,$03,$BF,$84,$FC,$16,$19
	.dc.b	$C7,$F0,$FF,$07,$CF,$E0,$BF,$0D,$F0,$7D,$7E,$0E,$E4,$3F,$D4,$5A
	.dc.b	$AB,$F1,$97,$BC,$3F,$23,$1E,$5F,$88,$FC,$86,$1E,$7F,$88,$B8,$F3
	.dc.b	$27,$C5,$7F,$83,$AB,$F2,$D7,$BC,$3F,$37,$1E,$FF,$89,$FC,$D6,$1F
	.dc.b	$07,$E2,$6E,$3C,$CF,$36,$3F,$E1,$8A,$FD,$05,$EF,$0F,$D2,$C7,$C9
	.dc.b	$E3,$FD,$26,$16,$7F,$52,$FC,$DD,$7F,$51,$59,$C7,$FC,$5F,$F9,$6B
	.dc.b	$AB,$E9,$70,$FF,$A0,$63,$13,$E3,$1A,$7F,$8A,$B8,$F5,$D1,$F6,$5E
	.dc.b	$18,$72,$AF,$E1,$F7,$FE,$3A,$79,$5F,$F8,$CB,$8B,$B7,$47,$E0,$79
	.dc.b	$3C,$1D,$16,$F8,$CF,$E1,$6B,$F4,$D1,$F8,$7E,$FF,$D4,$73,$D8,$FE
	.dc.b	$0E,$53,$FC,$97,$D0,$8F,$C7,$78,$1E,$DF,$0F,$F2,$1E,$0F,$BF,$C5
	.dc.b	$FC,$97,$83,$71,$F3,$F3,$7F,$A9,$5F,$5D,$F0,$7D,$BE,$27,$E6,$7C
	.dc.b	$0F,$90,$FE,$6F,$C1,$B3,$C9,$F9,$9E,$26,$E3,$E2,$F8,$7F,$A1,$F3
	.dc.b	$3E,$87,$8B,$FA,$3F,$32,$D7,$89,$FA,$95,$B8,$F9,$BE,$87,$D5,$7E
	.dc.b	$AF,$E5,$EF,$0B,$FD,$2B,$D8,$AD,$76,$29,$E7,$6B,$D2,$CC,$71,$C3
	.dc.b	$AE,$3F,$62,$37,$1F,$73,$D1,$DF,$9D,$DB,$D1,$B3,$9B,$E1,$E7,$E5
	.dc.b	$E5,$5E,$4F,$43,$BF,$B7,$E6,$F4,$AE,$1F,$88,$9F,$1F,$B5,$9C,$3C
	.dc.b	$57,$FE,$B0,$FC,$BF,$FC,$03,$1E,$5F,$6B,$FD,$B1,$18,$8F,$D5,$70
	.dc.b	$FB,$E7,$F3,$9F,$13,$EE,$77,$F0,$5F,$03,$DA,$FD,$07,$C0,$7C,$2F
	.dc.b	$6F,$F4,$5F,$01,$72,$FC,$DC,$D7,$F5,$31,$F2,$3D,$CF,$D4,$7B,$BF
	.dc.b	$2B,$E2,$3F,$55,$EF,$DB,$FD,$1C,$D7,$F6,$31,$F4,$3E,$23,$D3,$F0
	.dc.b	$7F,$4B,$E1,$FD,$CF,$D4,$F8,$7F,$81,$F0,$D9,$E8,$FE,$F7,$CB,$FE
	.dc.b	$06,$CB,$EB,$57,$C5,$F8,$9F,$B7,$85,$9F,$B5,$87,$C9,$F8,$9C,$E3
	.dc.b	$CA,$FF,$03,$3F,$37,$D4,$FC,$1C,$7C,$FF,$89,$FC,$2D,$E7,$0F,$C2
	.dc.b	$4E,$5F,$C3,$47,$D5,$E4,$FC,$5D,$D9,$FC,$53,$5F,$AB,$E1,$30,$FB
	.dc.b	$9C,$7C,$33,$F6,$F1,$FE,$53,$9F,$F0,$0F,$E6,$BF,$2E,$75,$63,$2F
	.dc.b	$CB,$57,$E1,$79,$35,$3F,$87,$E4,$E1,$D1,$6F,$F8,$9E,$08,$C4,$7E
	.dc.b	$22,$1D,$1F,$8F,$E5,$E3,$A7,$D0,$F0,$3C,$5D,$3F,$90,$6E,$1F,$53
	.dc.b	$C1,$AF,$FB,$73,$A5,$3E,$9E,$AF,$CB,$F7,$7A,$FC,$6F,$CC,$F8,$1E
	.dc.b	$DF,$17,$38,$F9,$9C,$BE,$89,$FC,$FF,$81,$F0,$FC,$5F,$D0,$F8,$3F
	.dc.b	$17,$C5,$CD,$FB,$9C,$BF,$02,$7F,$AC,$7C,$3F,$F0,$78,$7C,$53,$F3
	.dc.b	$07,$A5,$F3,$FC,$8D,$5E,$EB,$53,$F4,$FC,$9C,$F3,$7F,$AA,$B0,$7A
	.dc.b	$3C,$63,$FC,$A3,$1D,$32,$FA,$35,$C1,$F0,$C7,$F8,$FD,$FA,$BE,$9E
	.dc.b	$71,$F2,$7A,$3F,$81,$8C,$57,$F9,$0A,$BE,$DF,$AD,$8A,$D6,$3F,$A6
	.dc.b	$3F,$11,$EA,$7F,$09,$7F,$15,$EA,$7F,$0B,$F4,$33,$1B,$C3,$F0,$B3
	.dc.b	$DB,$DB,$FC,$87,$A7,$E0,$F7,$AD,$FB,$27,$F2,$9E,$B6,$70,$FC,$6D
	.dc.b	$F9,$BD,$CF,$CC,$7B,$F6,$7F,$22,$FA,$BD,$CF,$CE,$7C,$0E,$67,$F2
	.dc.b	$53,$EE,$F7,$7F,$41,$F0,$1F,$03,$DE,$FD,$17,$C0,$66,$FF,$37,$39
	.dc.b	$7F,$3B,$1F,$1B,$E0,$FF,$A0,$FE,$63,$FE,$0D,$ED,$FB,$5D,$0F,$85
	.dc.b	$CE,$1F,$84,$E3,$FB,$58,$62,$3E,$CC,$F2,$F8,$3D,$47,$D8,$F8,$6E
	.dc.b	$BF,$13,$8E,$EF,$C5,$7C,$2F,$83,$E2,$B1,$A2,$F1,$87,$E1,$2B,$8B
	.dc.b	$F8,$6F,$88,$F3,$FC,$4F,$0F,$89,$B3,$93,$FC,$17,$C2,$9D,$46,$3E
	.dc.b	$2B,$BB,$F0,$38,$3C,$3B,$FC,$2E,$0C,$CF,$27,$2F,$49,$F3,$F0,$7C
	.dc.b	$8D,$6C,$F8,$1F,$5E,$F9,$8F,$CE,$CF,$CE,$DB,$DF,$9F,$D0,$DA,$D7
	.dc.b	$A3,$6F,$87,$AF,$D4,$73,$1F,$D6,$36,$57,$CB,$E8,$75,$FD,$AE,$DF
	.dc.b	$D2,$46,$2F,$E1,$C6,$3B,$7E,$FF,$76,$AF,$87,$37,$FB,$0B,$0F,$F1
	.dc.b	$EE,$1D,$FB,$B9,$F6,$FE,$19,$EB,$3F,$88,$EE,$CD,$7D,$09,$F0,$F6
	.dc.b	$7E,$33,$BA,$CF,$D4,$7C,$BD,$BF,$90,$EE,$CC,$FB,$67,$D3,$E4,$7E
	.dc.b	$53,$C8,$F5,$F8,$F7,$6F,$34,$FE,$63,$C8,$CE,$1F,$80,$BF,$83,$E4
	.dc.b	$7E,$73,$C7,$B3,$F8,$47,$E2,$F9,$1F,$A0,$F8,$AC,$CF,$CD,$9F,$93
	.dc.b	$E4,$FF,$89,$32,$FF,$10,$4B,$C7,$FE,$3D,$C8,$F2,$35,$E4,$F0,$AF
	.dc.b	$F2,$DD,$F0,$BD,$D6,$B7,$F9,$DE,$B7,$5F,$53,$EB,$BD,$FD,$6F,$B3
	.dc.b	$EA,$E6,$B7,$97,$49,$FB,$BE,$97,$97,$D5,$FB,$FE,$A7,$9F,$D5,$CD
	.dc.b	$F8,$A7,$2F,$92,$3F,$0B,$E9,$7B,$7D,$5F,$C3,$FA,$97,$3F,$C1,$4F
	.dc.b	$C1,$F5,$73,$87,$B2,$FF,$98,$7E,$70,$F8,$4F,$8D,$F0,$BF,$BD,$8F
	.dc.b	$EF,$15,$F0,$A3,$13,$A8,$FF,$57,$72,$7F,$47,$AD,$C7,$F8,$D3,$95
	.dc.b	$F0,$7C,$FE,$0B,$E9,$7C,$1D,$AF,$9A,$FE,$87,$E0,$B3,$1D,$37,$FF
	.dc.b	$5D,$7B,$8C,$BF,$C8,$13,$FE,$B2,$C9,$F0,$7F,$A9,$F8,$2F,$BD,$F0
	.dc.b	$7F,$AB,$F8,$2F,$C0,$7C,$16,$70,$FB,$6E,$5F,$BB,$3E,$F3,$F0,$DF
	.dc.b	$13,$F1,$26,$E9,$E3,$AF,$ED,$39,$A7,$F1,$9F,$13,$F2,$B0,$DD,$7E
	.dc.b	$36,$FD,$37,$FE,$39,$8F,$89,$F1,$3F,$43,$1F,$CA,$7C,$4D,$9F,$C7
	.dc.b	$CF,$D4,$C7,$38,$78,$77,$F8,$53,$F9,$BE,$5F,$B2,$FE,$77,$E2,$2E
	.dc.b	$1F,$98,$9F,$BB,$8E,$70,$E5,$9B,$A3,$F4,$9C,$BF,$82,$E9,$FF,$44
	.dc.b	$FE,$67,$FE,$30,$BD,$C6,$9E,$35,$AE,$6F,$C4,$F5,$70,$E5,$BA,$FF
	.dc.b	$B0,$B1,$FF,$18,$61,$E0,$AE,$BC,$BF,$90,$E9,$E1,$3E,$4E,$9F,$0F
	.dc.b	$87,$F9,$46,$C6,$71,$E9,$E1,$FE,$63,$AB,$31,$F3,$CF,$AB,$3E,$AF
	.dc.b	$0B,$F3,$BE,$37,$B3,$C3,$FC,$FC,$E4,$F8,$71,$97,$E2,$F8,$BF,$0B
	.dc.b	$C1,$FD,$2F,$8B,$F1,$3C,$3F,$D3,$CE,$6B,$E5,$CE,$5F,$9B,$E2,$FC
	.dc.b	$AF,$0F,$F5,$BE,$2D,$9C,$EB,$F1,$7C,$FC,$BF,$1E,$BF,$C4,$5F,$9C
	.dc.b	$BD,$1F,$A7,$E7,$FF,$6F,$8C,$57,$CD,$8C,$4E,$A3,$F0,$D3,$FE,$DB
	.dc.b	$7F,$D4,$BD,$2F,$3F,$97,$A1,$F7,$BC,$EB,$5F,$5F,$CD,$AF,$07,$C1
	.dc.b	$E8,$EE,$3A,$FA,$1F,$86,$F6,$3C,$9E,$95,$9F,$0B,$5F,$A9,$E6,$35
	.dc.b	$FF,$20,$47,$93,$DA,$F6,$7B,$9F,$90,$F6,$BD,$DE,$F5,$AF,$5B,$5E
	.dc.b	$77,$FD,$A1,$FC,$77,$AD,$F1,$7D,$DF,$CC,$FB,$56,$7F,$24,$D7,$E1
	.dc.b	$BF,$2F,$DD,$FC,$FF,$B7,$F3,$7D,$CF,$F7,$4F,$E6,$BF,$E2,$9E,$DF
	.dc.b	$E2,$39,$A7,$38,$EC,$1F,$ED,$2A,$E0,$FF,$9B,$A5,$EF,$74,$F8,$8F
	.dc.b	$B5,$F0,$5D,$BE,$13,$EF,$B7,$2F,$B3,$F0,$7E,$2F,$85,$CE,$5D,$89
	.dc.b	$F1,$AD,$18,$FE,$23,$E2,$3D,$39,$5B,$E0,$79,$AF,$86,$3F,$8E,$F8
	.dc.b	$9D,$E7,$F8,$8C,$3E,$0C,$FE,$4B,$E2,$7E,$1C,$5C,$39,$8C,$3C,$51
	.dc.b	$F9,$8F,$89,$F9,$38,$EE,$BE,$24,$FF,$9F,$7F,$3C,$72,$FC,$EC,$7F
	.dc.b	$C9,$50,$E4,$FA,$38,$E1,$B8,$7F,$21,$9E,$2F,$8B,$2F,$85,$7F,$A5
	.dc.b	$8F,$B3,$D5,$5D,$70,$3A,$71,$1F,$5D,$FD,$67,$1F,$08,$FD,$4F,$17
	.dc.b	$E0,$BA,$3F,$61,$C9,$67,$F5,$AD,$7F,$27,$3F,$88,$E8,$F5,$F8,$5F
	.dc.b	$8A,$E9,$B5,$E2,$E8,$AF,$E1,$DF,$83,$E1,$6E,$3D,$DE,$07,$E4,$BC
	.dc.b	$6F,$8A,$FE,$53,$E2,$6C,$FB,$1F,$93,$E0,$6E,$3E,$37,$81,$F9,$AF
	.dc.b	$17,$E6,$F8,$3F,$9C,$F1,$6D,$7C,$47,$E8,$D6,$E3,$E7,$7A,$1F,$A2
	.dc.b	$F1,$BE,$C3,$7C,$2F,$A5,$E8,$7E,$9B,$C7,$CD,$FC,$EB,$FB,$DE,$5F
	.dc.b	$E1,$A3,$F0,$DF,$9B,$FF,$8A,$70,$FE,$F7,$E0,$2B,$5E,$02,$B5,$7F
	.dc.b	$ED,$6A,$79,$BA,$97,$A5,$DA,$BF,$05,$1C,$3C,$B5,$BC,$EE,$1E,$97
	.dc.b	$E4,$5E,$5E,$97,$E4,$FC,$FC,$ED,$F8,$D9,$CB,$F8,$FF,$73,$C1,$E9
	.dc.b	$FE,$67,$DB,$F1,$7A,$97,$6F,$0C,$FE,$76,$B3,$7F,$97,$F6,$BD,$1E
	.dc.b	$F7,$E8,$7D,$AF,$57,$BF,$71,$F4,$CF,$E9,$6B,$37,$F9,$FF,$6B,$E0
	.dc.b	$7B,$DF,$A9,$6C,$66,$FE,$0F,$BD,$FA,$DF,$73,$31,$FC,$C3,$C2,$7C
	.dc.b	$2F,$C9,$F7,$FF,$D7,$10,$F8,$3F,$99,$F0,$18,$6D,$FF,$83,$E7,$8B
	.dc.b	$F3,$AB,$8F,$C2,$7D,$3F,$88,$E7,$F0,$DF,$57,$E2,$6D,$FD,$47,$6F
	.dc.b	$D6,$9F,$0F,$C3,$7D,$BF,$89,$FE,$03,$F0,$FF,$77,$E2,$6D,$FD,$C7
	.dc.b	$6F,$DE,$9F,$4E,$7F,$83,$DD,$C5,$E7,$E0,$FC,$37,$06,$30,$F8,$51
	.dc.b	$B9,$F6,$5E,$9F,$76,$9F,$8D,$DF,$87,$C5,$71,$F0,$DC,$4A,$D5,$F0
	.dc.b	$CB,$F2,$3C,$5C,$23,$FC,$AF,$5F,$E4,$59,$FC,$AE,$39,$58,$74,$DB
	.dc.b	$F3,$9C,$39,$8E,$FC,$18,$C7,$E9,$3F,$A6,$E4,$FF,$1A,$61,$F7,$70
	.dc.b	$D3,$F7,$FB,$35,$D9,$F7,$BB,$78,$77,$5A,$FE,$51,$82,$3F,$07,$7C
	.dc.b	$27,$F0,$1D,$9D,$FB,$3F,$0E,$D8,$C9,$CA,$3F,$17,$DB,$E5,$EE,$FC
	.dc.b	$6F,$77,$9F,$BB,$F1,$F9,$FA,$7B,$AE,$3E,$27,$6F,$92,$7F,$AF,$74
	.dc.b	$7D,$1F,$CF,$BF,$C0,$F3,$FD,$82,$DE,$6F,$C7,$97,$99,$F1,$63,$0F
	.dc.b	$C7,$F3,$75,$E6,$E7,$9F,$F3,$87,$CB,$F3,$7E,$87,$67,$F0,$CC,$FE
	.dc.b	$2C,$63,$CB,$FA,$1E,$66,$AF,$81,$FA,$5F,$2B,$EA,$79,$1C,$9B,$19
	.dc.b	$C3,$E7,$F9,$3D,$BC,$AC,$C7,$F8,$93,$83,$FC,$18,$FD,$4F,$63,$D9
	.dc.b	$EC,$7D,$FF,$5F,$DD,$EC,$7E,$06,$73,$5F,$6A,$7E,$17,$B1,$F8,$5F
	.dc.b	$63,$E2,$7B,$1F,$87,$F6,$33,$7F,$80,$9C,$BF,$82,$8F,$95,$C9,$F8
	.dc.b	$DF,$5B,$E6,$7A,$F7,$87,$FD,$21,$BF,$E8,$B0,$FC,$4F,$B1,$A7,$F1
	.dc.b	$77,$F4,$FE,$07,$F2,$DF,$09,$F6,$B0,$FF,$5C,$E0,$F8,$4F,$CD,$7C
	.dc.b	$26,$71,$FD,$1C,$E5,$FC,$AC,$67,$CF,$4B,$1E,$B5,$E4,$F8,$4F,$E9
	.dc.b	$3F,$9C,$FF,$8C,$2F,$7F,$09,$8F,$84,$E5,$8F,$F8,$07,$0F,$F0,$45
	.dc.b	$A7,$38,$72,$9D,$FC,$27,$4F,$83,$B1,$C1,$AF,$F8,$B3,$E0,$98,$7F
	.dc.b	$8C,$A3,$E3,$FC,$1F,$93,$E1,$3C,$37,$E6,$F8,$9F,$CC,$47,$A3,$E2
	.dc.b	$7F,$35,$39,$E3,$F9,$F3,$97,$E8,$D7,$BB,$E2,$3F,$40,$D8,$FA,$B7
	.dc.b	$F0,$BE,$27,$39,$7B,$AF,$F4,$C7,$C6,$E3,$FD,$44,$FC,$8E,$5C,$D7
	.dc.b	$C4,$9C,$BF,$1A,$3F,$8A,$7E,$7E,$E8,$F9,$FC,$5F,$56,$1D,$3F,$4B
	.dc.b	$9B,$0D,$C3,$FD,$B1,$3C,$5F,$A3,$5C,$7A,$7E,$CF,$2F,$3E,$9B,$5C
	.dc.b	$1A,$FF,$A2,$AF,$FC,$82,$FD,$1E,$57,$2F,$DE,$8F,$C1,$F8,$7F,$80
	.dc.b	$E8,$B1,$F9,$A6,$BF,$B1,$3F,$13,$E1,$FA,$FC,$7F,$C5,$F8,$56,$7F
	.dc.b	$0F,$E1,$7B,$FC,$6F,$C7,$B9,$8F,$14,$FC,$3F,$17,$F2,$7E,$17,$C5
	.dc.b	$F1,$6D,$7C,$29,$AF,$AE,$3F,$33,$E1,$FC,$BF,$17,$F3,$7E,$5F,$CD
	.dc.b	$F1,$BF,$3B,$E6,$E7,$0F,$CC,$B9,$7F,$37,$3F,$4F,$D1,$FF,$15,$FE
	.dc.b	$77,$FD,$86,$7E,$47,$A7,$8F,$3B,$51,$FE,$69,$9F,$E4,$0E,$8C,$78
	.dc.b	$BF,$77,$D1,$E7,$E7,$59,$E0,$FE,$07,$D1,$CC,$7D,$E9,$F0,$F9,$FF
	.dc.b	$85,$F4,$7C,$7E,$CF,$E1,$FD,$2F,$2F,$B7,$9C,$3C,$4E,$5F,$24,$FF
	.dc.b	$05,$5F,$83,$7F,$91,$DF,$F2,$63,$FE,$49,$F5,$3F,$D6,$3A,$3E,$13
	.dc.b	$FD,$4D,$3B,$AF,$F4,$BB,$B5,$3B,$7F,$D7,$31,$B7,$FD,$4E,$EE,$35
	.dc.b	$1E,$2C,$BF,$CB,$8D,$BF,$0D,$F8,$27,$37,$C6,$31,$5E,$89,$FF,$00
	.dc.b	$BE,$28,$CB,$E6,$8C,$DF,$82,$BC,$93,$D6,$FB,$BE,$8A,$EE,$70,$BE
	.dc.b	$71,$EA,$8C,$3B,$3C,$53,$C9,$F6,$18,$3C,$86,$CF,$23,$E9,$74,$75
	.dc.b	$29,$E0,$3C,$C7,$AC,$C9,$CC,$F2,$3C,$DE,$A6,$CA,$75,$29,$D4,$F8
	.dc.b	$0F,$17,$99,$C4,$A7,$FB,$D3,$F0,$51,$00

;--------------------------------------------------------------------------------
;	FSINH.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsinh_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FSINH.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSINH.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsinh_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsinh.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSINH.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsinh.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSINH.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsinh_data_compressed::
	.dc.b	$00,$02,$84,$2C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CF,$ED,$1F,$8A,$7F,$87,$A7,$FC,$D0,$5F,$8A,$E1
	.dc.b	$F0,$FC,$7D,$FE,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$1F
	.dc.b	$07,$9F,$DB,$F4,$7B,$7F,$DF,$7E,$6F,$FA,$B0,$F8,$8C,$7C,$46,$7E
	.dc.b	$23,$5F,$11,$BF,$88,$E1,$F1,$1C,$7E,$17,$97,$C0,$F3,$F8,$1E,$9F
	.dc.b	$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03
	.dc.b	$E3,$F8,$1F,$27,$C0,$F9,$7E,$07,$CD,$F0,$3E,$7F,$81,$FF,$0A,$7C
	.dc.b	$E7,$E1,$BE,$13,$FC,$40,$BE,$07,$3E,$66,$BE,$07,$7F,$03,$C3,$E0
	.dc.b	$78,$FC,$0F,$2F,$81,$E7,$F0,$3F,$63,$D2,$EB,$F0,$3D,$BE,$07,$BF
	.dc.b	$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93,$E0,$7C,$BF
	.dc.b	$03,$E6,$F8,$1F,$3F,$C0,$FA,$3E,$07,$FC,$59,$F3,$BF,$88,$E9,$FF
	.dc.b	$1C,$7C,$07,$F0,$A5,$1F,$EF,$63,$FD,$60,$7F,$9A,$0F,$EB,$47,$F1
	.dc.b	$23,$81,$FC,$41,$FF,$21,$2F,$7B,$BF,$BD,$E0,$F7,$BC,$3E,$F7,$8B
	.dc.b	$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD,$E8,$F7,$BF
	.dc.b	$3B,$EF,$FE,$7B,$5F,$CF,$EB,$FA,$0D,$7F,$43,$AF,$BB,$DE,$F7,$FB
	.dc.b	$DF,$E4,$BF,$6F,$FC,$48,$7E,$3F,$8B,$FC,$B0,$BD,$CD,$3B,$F7,$38
	.dc.b	$7B,$9C,$7D,$CE,$5E,$E7,$3F,$73,$A7,$B9,$FB,$0E,$8E,$DE,$E7,$7F
	.dc.b	$73,$C1,$EE,$7E,$D3,$AF,$F6,$BD,$7F,$B6,$EB,$FD,$BF,$5F,$EE,$3A
	.dc.b	$FF,$73,$D7,$FB,$AE,$6F,$DD,$F3,$7F,$98,$FF,$15,$FF,$1E,$1E,$9E
	.dc.b	$73,$E9,$7F,$CE,$4F,$FB,$38,$FF,$44,$1F,$DE,$0F,$E6,$07,$F2,$F7
	.dc.b	$8F,$0F,$5E,$1E,$DC,$1D,$F8,$3E,$47,$BB,$F2,$7D,$DF,$95,$EE,$FC
	.dc.b	$BF,$77,$E6,$7B,$BF,$37,$DD,$F9,$DE,$EF,$CF,$F7,$7E,$87,$BB,$FE
	.dc.b	$7B,$F9,$FF,$D4,$70,$1F,$D4,$3F,$E8,$95,$EE,$6F,$DC,$E1,$EE,$71
	.dc.b	$F7,$39,$7B,$9C,$FD,$CE,$9E,$E7,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B
	.dc.b	$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$F3,$7C,$9F,$9D,$D9,$F3
	.dc.b	$FB,$3E,$87,$67,$D1,$EC,$FA,$5D,$9F,$4F,$B3,$EA,$76,$7D,$5E,$CF
	.dc.b	$AD,$D9,$FE,$99,$F4,$FD,$E7,$BB,$B4,$FB,$9F,$F5,$42,$EB,$E1,$D7
	.dc.b	$C7,$AF,$97,$5F,$3E,$BE,$9D,$7D,$7A,$FE,$67,$B7,$F3,$7D,$BF,$9D
	.dc.b	$ED,$F8,$7B,$BE,$87,$B7,$FE,$AB,$FA,$0F,$E0,$1C,$05,$F0,$99,$F8
	.dc.b	$AD,$7C,$AE,$FE,$17,$87,$C0,$F1,$F8,$6E,$5F,$01,$CF,$E0,$3A,$74
	.dc.b	$75,$F8,$0F,$81,$E4,$FC,$1F,$27,$E1,$79,$3F,$0F,$C9,$F8,$9E,$4F
	.dc.b	$8F,$E0,$3C,$9E,$D7,$97,$DA,$F3,$7B,$5E,$7F,$6B,$D1,$ED,$7A,$7D
	.dc.b	$AF,$57,$B5,$EB,$F6,$BD,$9E,$D7,$FB,$1F,$E8,$7F,$D4,$07,$E9,$4F
	.dc.b	$D2,$76,$97,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1F,$67,$D9
	.dc.b	$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BB,$E1,$F7,$7C,$5E,$EF
	.dc.b	$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE,$F9,$FD,$DF,$47,$BB,$F8,$7F
	.dc.b	$57,$D5,$EE,$FF,$B8,$7C,$FF,$D3,$9F,$89,$3F,$11,$D2,$7F,$10,$FF
	.dc.b	$BC,$5F,$F7,$61,$FE,$A5,$3F,$CB,$AE,$FD,$9E,$9E,$CF,$5F,$67,$B7
	.dc.b	$AF,$DF,$CC,$F0,$79,$9E,$1F,$33,$C5,$E6,$78,$FC,$CF,$27,$67,$97
	.dc.b	$7F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$FB,$E7,$E8,$BF,$1E,$7E,$3B,$84
	.dc.b	$17,$B1,$BE,$3E,$1E,$C7,$1F,$63,$97,$B1,$CF,$D8,$E9,$EC,$75,$F6
	.dc.b	$3B,$7B,$1F,$AF,$DF,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E
	.dc.b	$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7A,$7D,$8F,$3F,$3B,$E8
	.dc.b	$FF,$DD,$87,$F3,$13,$F1,$A7,$E3,$1F,$6A,$F5,$F8,$BF,$EA,$83,$FC
	.dc.b	$C1,$C2,$EB,$CF,$AB,$AF,$57,$7E,$AF,$0F,$57,$8F,$AB,$CB,$D5,$E7
	.dc.b	$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E,$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F
	.dc.b	$57,$C7,$EA,$FB,$3E,$93,$F9,$99,$FC,$54,$FC,$69,$F8,$CE,$62,$F7
	.dc.b	$F8,$7B,$FC,$7D,$FE,$5E,$CF,$3F,$67,$A7,$B3,$D7,$D9,$ED,$EC,$F7
	.dc.b	$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9,$EC,$F9,$7C,$FF
	.dc.b	$37,$85,$E7,$F6,$7E,$9F,$C0,$7F,$1A,$3F,$19,$CB,$FC,$A7,$D9,$CF
	.dc.b	$B3,$AF,$67,$7E,$CF,$0F,$67,$8F,$B3,$CB,$D9,$E7,$E6,$74,$F6,$7A
	.dc.b	$FB,$3D,$BD,$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F,$67,$C7,$EC,$F9
	.dc.b	$3D,$9F,$2F,$B3,$E6,$F6,$7F,$13,$F4,$BF,$8A,$E6,$3F,$DE,$4F,$E2
	.dc.b	$A7,$E2,$5C,$FB,$5C,$3D,$DE,$3E,$EF,$2F,$6F,$9F,$B7,$D3,$DB,$EB
	.dc.b	$ED,$F6,$F6,$FB,$FB,$7E,$0F,$6F,$C3,$ED,$F8,$BD,$BF,$1F,$B7,$E4
	.dc.b	$F6,$FC,$BE,$DF,$9B,$DB,$F3,$FB,$7E,$8F,$6F,$F4,$3F,$4D,$FD,$E0
	.dc.b	$FE,$B4,$7F,$12,$3F,$0C,$7E,$17,$97,$F7,$AB,$DA,$E3,$ED,$72,$F6
	.dc.b	$B9,$FB,$5D,$3D,$AE,$BE,$D7,$6F,$6B,$BF,$B5,$E0,$F6,$BC,$3E,$D7
	.dc.b	$8B,$DA,$F1,$FB,$5E,$4F,$6B,$CB,$ED,$79,$BD,$AF,$3F,$B5,$FB,$8F
	.dc.b	$A7,$FE,$BE,$7E,$18,$FE,$1F,$C9,$FC,$49,$FE,$54,$DF,$63,$87,$B1
	.dc.b	$C7,$D8,$E5,$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1
	.dc.b	$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E
	.dc.b	$7F,$63,$F8,$7F,$AF,$F8,$73,$FC,$7C,$7F,$5A,$3F,$0C,$7F,$0D,$E4
	.dc.b	$FE,$56,$BD,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6,$F5
	.dc.b	$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC,$BE
	.dc.b	$BF,$9B,$D7,$F3,$FA,$FE,$8F,$5F,$F9,$3F,$D4,$7F,$8F,$8F,$F1,$01
	.dc.b	$FD,$60,$FE,$18,$7E,$1F,$94,$FE,$1D,$FE,$8A,$FF,$58,$1E,$AF,$4F
	.dc.b	$57,$AF,$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E
	.dc.b	$AF,$93,$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FA,$3D,$5F,$E7,$FF,$53
	.dc.b	$FE,$26,$3F,$86,$9F,$88,$E2,$3F,$88,$6F,$AB,$C3,$D5,$E3,$EA,$F2
	.dc.b	$F5,$79,$FA,$BD,$3D,$5E,$BE,$AF,$6F,$57,$BF,$AB,$E0,$F5,$7C,$3E
	.dc.b	$AF,$8B,$D5,$F1,$FA,$BE,$4F,$57,$CB,$EA,$F9,$BD,$5F,$3F,$AB,$FD
	.dc.b	$5B,$EA,$BF,$87,$1F,$E8,$53,$FC,$42,$7F,$0D,$3F,$11,$C8,$7F,$10
	.dc.b	$FF,$74,$5E,$A7,$3F,$53,$A7,$A9,$D7,$D4,$ED,$EA,$77,$F5,$3C,$1E
	.dc.b	$A7,$87,$D4,$F1,$7A,$9E,$3F,$53,$C9,$EA,$79,$7D,$4F,$37,$A9,$E7
	.dc.b	$F5,$3D,$1E,$A7,$A7,$D4,$FE,$DF,$E2,$7F,$A1,$8F,$F3,$41,$FE,$20
	.dc.b	$3F,$AA,$1F,$89,$3F,$11,$C8,$7F,$C3,$E7,$E2,$1F,$F0,$7A,$F4,$BA
	.dc.b	$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA,$5E,$2F,$4B,$C7,$E9,$79
	.dc.b	$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$A3,$D2,$FF,$03,$FD,$5F,$F9
	.dc.b	$B8,$FE,$AA,$7E,$24,$FC,$41,$E7,$E1,$FF,$11,$AF,$4B,$8F,$A5,$CB
	.dc.b	$D2,$E7,$E9,$74,$F4,$BA,$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA
	.dc.b	$5E,$2F,$4B,$C7,$E9,$79,$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$F8
	.dc.b	$7B,$EB,$3F,$AB,$1F,$EC,$D3,$FC,$D2,$7F,$55,$3F,$12,$7E,$23,$90
	.dc.b	$FE,$21,$FF,$1E,$BF,$EC,$D1,$E8,$75,$F4,$3B,$7A,$1D,$FD,$0F,$07
	.dc.b	$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F,$43,$CD,$E8,$79
	.dc.b	$FD,$0F,$47,$A1,$E9,$F4,$3F,$C6,$FF,$5B,$FE,$CE,$3F,$D6,$07,$F9
	.dc.b	$A0,$FF,$0E,$1F,$C4,$4F,$C4,$9F,$E6,$63,$F1,$1C,$47,$F1,$0F,$F9
	.dc.b	$61,$79,$FD,$BC,$FE,$FE,$7F,$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7
	.dc.b	$F9,$3C,$FF,$2F,$9F,$E6,$F3,$FC,$FE,$7F,$A3,$CF,$FF,$27,$FC,$17
	.dc.b	$FA,$D8,$FF,$0E,$9F,$C4,$4F,$E2,$07,$E2,$78,$3F,$C6,$9E,$7F,$0F
	.dc.b	$3F,$8F,$9F,$CB,$CF,$E7,$E7,$F4,$F3,$FA,$F9,$FD,$BC,$FE,$FE,$7F
	.dc.b	$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7,$F9,$3C,$FF,$2F,$9F,$E6,$F3
	.dc.b	$FF,$CC,$BF,$5D,$FE,$1D,$1F,$27,$8F,$8F,$CF,$C4,$EB,$DE,$DF,$BD
	.dc.b	$C3,$DA,$E3,$ED,$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$FE,$7C,$FA,$F7,$CB,$DF,$94,$C7,$C8,$67,$E2,$B5
	.dc.b	$E7,$6F,$E1,$B8,$7C,$37,$1F,$84,$E5,$F0,$3C,$FD,$EE,$9E,$F7,$5F
	.dc.b	$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27
	.dc.b	$BD,$E5,$F7,$BC,$DE,$F7,$FA,$4B,$EC,$3F,$DF,$27,$F9,$80,$FE,$A8
	.dc.b	$7F,$53,$3F,$08,$7F,$0B,$F7,$FF,$D5,$EB,$DE,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$FD,$53,$EC,$FF,$98
	.dc.b	$4F,$C2,$F1,$FF,$B0,$9F,$F7,$A9,$FE,$61,$3F,$AA,$9F,$C3,$4F,$E1
	.dc.b	$8F,$F8,$77,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8E,$FE,$C7,$83
	.dc.b	$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79,$3D,$8F,$2F,$B1,$E6,$F6,$3C
	.dc.b	$FE,$C7,$A3,$D8,$FF,$60,$FD,$8F,$E1,$B8,$BF,$DB,$2B,$D7,$CF,$AF
	.dc.b	$AF,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D
	.dc.b	$7F,$2F,$AF,$E6,$F5,$FC,$FE,$BF,$FB,$63,$EC,$9F,$35,$7E,$4F,$1F
	.dc.b	$1F,$9F,$8F,$D7,$C4,$EF,$E1,$B8,$7B,$BC,$7D,$DE,$5E,$EF,$3F,$77
	.dc.b	$A7,$BB,$D7,$DD,$ED,$EE,$F7,$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE
	.dc.b	$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$FE,$EE,$FB,$3F,$F5,$31,$F8
	.dc.b	$6F,$38,$BE,$17,$5F,$0B,$BF,$85,$E1,$F0,$3C,$7E,$17,$97,$C1,$73
	.dc.b	$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F,$82,$F0,$7C,$17,$87,$E0
	.dc.b	$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E,$0B,$CD,$F0,$5E,$7F,$82
	.dc.b	$F4,$7C,$13,$9E,$78,$FF,$87,$7F,$BE,$7C,$0E,$5F,$EF,$DF,$03,$67
	.dc.b	$FD,$DC,$7F,$98,$CF,$F0,$D8,$F7,$B5,$EF,$6F,$DE,$E1,$EF,$71,$F7
	.dc.b	$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE
	.dc.b	$1F,$7B,$C5,$EF,$78,$FD,$EF,$37,$DA,$7E,$23,$93,$D8,$FF,$BB,$8F
	.dc.b	$F3,$11,$FE,$61,$3F,$A9,$1F,$C3,$D3,$FC,$3D,$E3,$EC,$74,$F6,$3A
	.dc.b	$FB,$1D,$BD,$8E,$FE,$C7,$83,$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79
	.dc.b	$3D,$8F,$2F,$B1,$E6,$F6,$3C,$FE,$C7,$A3,$D8,$F9,$DF,$6B,$FE,$EF
	.dc.b	$3F,$14,$7E,$27,$87,$EB,$3F,$EE,$F3,$FC,$C6,$7F,$86,$CF,$E1,$E3
	.dc.b	$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD,$3E,$FE,$9F,$83,$D3,$F0,$FA
	.dc.b	$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F,$A7,$E6,$F4,$FC,$FE,$9F,$A3
	.dc.b	$D3,$FC,$37,$DB,$7E,$24,$FE,$1D,$C7,$F8,$D5,$E9,$EB,$E1,$37,$F0
	.dc.b	$9C,$3E,$13,$8F,$AB,$CB,$D5,$E7,$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E
	.dc.b	$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F,$57,$C7,$EA,$F9,$3D,$5F,$2F,$AB
	.dc.b	$E6,$F5,$7C,$FE,$AF,$E7,$78,$7F,$11,$CD,$FF,$00,$7E,$0B,$1F,$05
	.dc.b	$9F,$82,$D7,$C1,$6F,$DA,$E1,$F0,$5C,$7D,$AE,$5E,$D7,$3F,$6B,$A7
	.dc.b	$B5,$D7,$DA,$ED,$ED,$77,$F6,$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F
	.dc.b	$6B,$C9,$ED,$79,$7D,$AF,$37,$B5,$FB,$3F,$B7,$FD,$A7,$39,$FE,$20
	.dc.b	$7E,$1D,$C7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$D1,$EF,$7F,$0B,$FB
	.dc.b	$8F,$E1,$C7,$E1,$B9,$8F,$E1,$9F,$E2,$8B,$DD,$E1,$EE,$F1,$F3,$B9
	.dc.b	$7B,$BC,$FD,$DE,$9E,$EF,$5F,$77,$B7,$BB,$DF,$DD,$F0,$7B,$BE,$1F
	.dc.b	$77,$C5,$EE,$F8,$FD,$DF,$27,$BB,$E5,$F7,$7C,$DE,$EF,$9F,$DD,$F4
	.dc.b	$7B,$BF,$C8,$FE,$E7,$F0,$E3,$DE,$C7,$BD,$9F,$7B,$5E,$F6,$FD,$EE
	.dc.b	$1E,$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$75,$F7,$BB,$7B,$DD,$FD
	.dc.b	$EF,$07,$BD,$E1,$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD
	.dc.b	$EF,$7F,$39,$F6,$BF,$0E,$7F,$55,$3F,$0C,$3D,$DD,$7B,$BB,$F7,$78
	.dc.b	$4F,$1F,$77,$97,$BB,$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7B,$BD,$FD,$DF
	.dc.b	$07,$BB,$E1,$F7,$7C,$5E,$EF,$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE
	.dc.b	$F9,$FD,$DF,$EA,$1F,$75,$FD,$4C,$FC,$31,$F8,$5F,$78,$FE,$15,$FE
	.dc.b	$B4,$BD,$CE,$3E,$E7,$2F,$73,$9F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B
	.dc.b	$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7
	.dc.b	$9B,$DC,$F3,$FB,$9E,$8F,$73,$FB,$47,$DD,$FF,$0C,$3F,$0D,$CB,$FD
	.dc.b	$BD,$7B,$7A,$F6,$F7,$ED,$F0,$F6,$F8,$FB,$7C,$BD,$BE,$7E,$DF,$4F
	.dc.b	$6D,$FC,$03,$0B,$A7,$F0,$4F,$B5,$FE,$0A,$D2,$9F,$DF,$DA,$D2,$B8
	.dc.b	$4D,$23,$A2,$1D,$1C,$B4,$7D,$87,$F9,$89,$FF,$56,$FC,$7F,$F8,$3E
	.dc.b	$31,$1D,$8D,$FA,$3F,$C7,$3E,$47,$FC,$B1,$3A,$BE,$13,$FE,$13,$FB
	.dc.b	$AF,$F1,$66,$18,$C3,$DD,$7F,$E7,$09,$CC,$74,$33,$5F,$E1,$BF,$BB
	.dc.b	$FF,$46,$5F,$19,$FF,$0F,$61,$67,$31,$C4,$C9,$4F,$59,$4F,$56,$1E
	.dc.b	$FC,$7D,$38,$7D,$2B,$ED,$8F,$92,$BD,$97,$D6,$31,$5B,$8F,$06,$1E
	.dc.b	$2A,$D9,$FE,$8C,$71,$3B,$8F,$06,$3E,$BA,$FB,$91,$87,$CD,$87,$8B
	.dc.b	$E3,$7C,$B1,$F1,$0F,$F2,$16,$3E,$27,$03,$1E,$EE,$1A,$FC,$4B,$EC
	.dc.b	$79,$B1,$FB,$58,$F8,$E7,$E2,$C7,$D9,$C3,$C3,$1C,$63,$13,$BC,$7E
	.dc.b	$96,$FA,$AF,$5D,$7D,$98,$C4,$74,$3F,$CA,$D9,$E2,$BE,$13,$63,$E7
	.dc.b	$61,$71,$FC,$33,$70,$E9,$8F,$86,$BE,$8E,$39,$7A,$C6,$6F,$DF,$5F
	.dc.b	$44,$F4,$CF,$CE,$3B,$CE,$E3,$AC,$66,$7E,$D5,$75,$8F,$F3,$96,$15
	.dc.b	$FC,$6C,$FC,$B7,$EA,$E1,$BC,$FE,$B4,$7C,$93,$FC,$F5,$9F,$D4,$7E
	.dc.b	$39,$8C,$3A,$61,$32,$F2,$4F,$46,$AB,$EA,$F2,$3F,$05,$5E,$7B,$ED
	.dc.b	$1F,$44,$F4,$D7,$27,$D3,$97,$C5,$7B,$1F,$05,$F8,$F9,$7B,$63,$AC
	.dc.b	$F8,$AB,$AE,$7F,$6A,$3A,$65,$F0,$73,$91,$BC,$BC,$87,$B7,$0E,$C7
	.dc.b	$D8,$29,$ED,$9E,$2E,$30,$EE,$FE,$02,$BB,$91,$E3,$9F,$C4,$7E,$03
	.dc.b	$71,$FC,$BC,$F8,$CE,$19,$7B,$6B,$E1,$9F,$AC,$AE,$0F,$E1,$8F,$16
	.dc.b	$3E,$AA,$E3,$9F,$C9,$9F,$9A,$78,$B2,$F4,$E1,$F2,$72,$F4,$BF,$8C
	.dc.b	$BE,$35,$FB,$A3,$C2,$7E,$1C,$F8,$4F,$E0,$31,$F1,$CF,$19,$F1,$CF
	.dc.b	$27,$81,$C8,$FC,$24,$F0,$C3,$ED,$9E,$7B,$F1,$BF,$A0,$BF,$85,$9F
	.dc.b	$D5,$7F,$01,$5C,$6F,$F0,$B1,$EB,$CB,$E1,$47,$D8,$3C,$07,$E3,$CB
	.dc.b	$9F,$F1,$BC,$BD,$59,$F0,$BE,$51,$E9,$AF,$CD,$1C,$0F,$A1,$1E,$29
	.dc.b	$E1,$97,$A7,$0F,$33,$F4,$6B,$85,$7F,$08,$BE,$E7,$DF,$8F,$8E,$79
	.dc.b	$F2,$E2,$FD,$FC,$3E,$BD,$F5,$AE,$17,$F1,$1F,$35,$7E,$32,$B6,$FF
	.dc.b	$2F,$C3,$CA,$FD,$BC,$3A,$BF,$6F,$2F,$94,$FD,$C3,$C7,$87,$C2,$7C
	.dc.b	$98,$76,$CB,$E4,$47,$E2,$E3,$C7,$97,$96,$36,$4A,$F8,$D3,$F0,$23
	.dc.b	$AC,$EE,$7D,$51,$B8,$FC,$C5,$F9,$0F,$2E,$7F,$84,$9F,$E2,$97,$E3
	.dc.b	$7F,$B6,$63,$F0,$1F,$66,$7F,$01,$FC,$C5,$71,$7F,$47,$8F,$80,$F2
	.dc.b	$BE,$B3,$CA,$7C,$9B,$ED,$1F,$CE,$23,$EA,$47,$AE,$3E,$31,$31,$FC
	.dc.b	$06,$1F,$3A,$BF,$3F,$87,$95,$FE,$49,$87,$8C,$F2,$3F,$86,$AF,$EC
	.dc.b	$2F,$8E,$FC,$A7,$57,$C8,$7C,$73,$EE,$61,$E3,$9F,$93,$7F,$06,$7B
	.dc.b	$E7,$E3,$7F,$45,$8F,$8D,$FC,$45,$FD,$A7,$AE,$1F,$17,$0F,$5C,$F5
	.dc.b	$9F,$82,$7C,$63,$E6,$DF,$5C,$FF,$98,$47,$A6,$BC,$75,$D6,$36,$FC
	.dc.b	$C8,$F7,$DF,$B9,$DC,$F5,$3D,$33,$B7,$F9,$A4,$ED,$F1,$BF,$15,$EA
	.dc.b	$7C,$53,$F8,$8E,$1E,$FB,$FC,$CC,$EE,$7A,$9B,$3F,$7F,$3B,$7F,$7F
	.dc.b	$7F,$97,$23,$F4,$0E,$07,$CA,$3E,$01,$F2,$9D,$BE,$2B,$F5,$3F,$AF
	.dc.b	$AF,$53,$E8,$7C,$8F,$D6,$8D,$E7,$FA,$3C,$BE,$E4,$FF,$69,$CF,$F1
	.dc.b	$38,$FD,$18,$F3,$9F,$13,$1F,$0B,$EE,$C7,$EE,$3E,$97,$C2,$FE,$5B
	.dc.b	$1F,$AC,$EE,$7F,$02,$FB,$B2,$F9,$31,$F5,$AB,$CF,$1F,$D6,$A7,$47
	.dc.b	$ED,$2B,$F1,$0F,$ED,$1E,$AF,$A7,$3F,$45,$7E,$B6,$34,$7C,$A3,$E2
	.dc.b	$5F,$A9,$F8,$87,$68,$FE,$58,$EC,$F4,$1F,$44,$F7,$47,$AA,$3D,$F1
	.dc.b	$F5,$DF,$4B,$FD,$E4,$E9,$1F,$94,$3D,$C4,$7A,$1E,$43,$A9,$0F,$4C
	.dc.b	$76,$BF,$88,$F6,$3D,$D1,$F8,$EB,$F8,$AF,$D1,$3F,$40,$7C,$57,$F0
	.dc.b	$EF,$F0,$53,$FA,$FD,$78,$DF,$11,$0E,$87,$A8,$D9,$F8,$33,$D2,$7C
	.dc.b	$63,$F2,$86,$CF,$CA,$3D,$5D,$10,$FA,$33,$F8,$A7,$DF,$1B,$7D,$07
	.dc.b	$AA,$BF,$81,$CE,$E3,$DC,$FC,$57,$E8,$3F,$16,$7E,$84,$7C,$B3,$6F
	.dc.b	$59,$F9,$6F,$E0,$DE,$AF,$DD,$35,$1F,$A2,$3E,$21,$E4,$9F,$13,$EA
	.dc.b	$8D,$C7,$E2,$8F,$14,$7F,$4E,$3E,$21,$F4,$4F,$8A,$7C,$B3,$AC,$F9
	.dc.b	$1F,$41,$D4,$D1,$EA,$3E,$89,$E3,$9F,$43,$C2,$3F,$5A,$F0,$34,$70
	.dc.b	$7D,$EE,$CF,$71,$D4,$F1,$9F,$32,$3A,$BF,$7A,$7E,$81,$EF,$8F,$95
	.dc.b	$1B,$3A,$47,$E5,$9F,$D3,$47,$C6,$3F,$C6,$87,$08,$D4,$F9,$27,$F0
	.dc.b	$67,$A5,$F8,$8F,$92,$36,$FF,$05,$3F,$70,$76,$3D,$C7,$8C,$F4,$4F
	.dc.b	$A9,$F7,$10,$F8,$A7,$57,$F1,$2F,$03,$C4,$7C,$63,$67,$E6,$E7,$D0
	.dc.b	$FE,$A2,$7E,$2C,$FF,$95,$23,$C7,$1F,$E5,$D3,$F1,$33,$C2,$3D,$07
	.dc.b	$C0,$3C,$47,$E2,$63,$6F,$BD,$EA,$FE,$D0,$D9,$FE,$9A,$3F,$1F,$87
	.dc.b	$F5,$18,$FA,$D1,$E2,$AF,$A2,$FD,$67,$DE,$FC,$3A,$FC,$4B,$D3,$1F
	.dc.b	$E8,$B8,$7F,$4A,$CB,$A5,$F2,$AF,$09,$F9,$D3,$F3,$D9,$FF,$4F,$AF
	.dc.b	$E6,$F8,$7D,$A9,$F8,$D5,$F4,$DF,$82,$78,$4F,$B0,$7C,$D3,$E0,$3B
	.dc.b	$8F,$2C,$6F,$1F,$99,$3F,$8B,$3D,$71,$FA,$93,$E6,$1B,$7D,$0F,$58
	.dc.b	$FC,$59,$23,$EC,$47,$03,$EF,$92,$7E,$C1,$1F,$C2,$3E,$47,$47,$A8
	.dc.b	$F1,$1F,$00,$FC,$5C,$FD,$47,$E7,$67,$FD,$6A,$BD,$35,$CC,$93,$CD
	.dc.b	$F9,$4F,$42,$3A,$AE,$91,$ED,$9F,$7B,$A3,$E5,$9E,$27,$DE,$EC,$FC
	.dc.b	$83,$F3,$DD,$9F,$0E,$7B,$47,$47,$C6,$7C,$39,$F9,$53,$F9,$53,$E8
	.dc.b	$1D,$0F,$C9,$1F,$E3,$29,$ED,$1E,$88,$FC,$19,$F2,$9F,$BA,$FE,$5C
	.dc.b	$E0,$7A,$4E,$C6,$E3,$C8,$FA,$0F,$A0,$7A,$4F,$AE,$7C,$B2,$47,$A1
	.dc.b	$FF,$3E,$C7,$F1,$77,$AC,$FB,$9D,$C7,$D6,$7B,$1E,$88,$F8,$AF,$BC
	.dc.b	$FD,$2E,$7F,$75,$E6,$78,$AB,$DA,$EB,$0F,$D7,$1F,$DF,$1F,$61,$CC
	.dc.b	$F5,$90,$F0,$1C,$EF,$E0,$CF,$48,$D5,$FD,$59,$FB,$F1,$C2,$74,$F6
	.dc.b	$3E,$09,$F3,$CF,$EA,$E7,$C9,$77,$3F,$0D,$F9,$11,$FC,$03,$2F,$E0
	.dc.b	$AF,$9C,$E8,$7F,$05,$C3,$2F,$43,$F2,$07,$F1,$99,$FE,$3E,$7D,$3C
	.dc.b	$BF,$85,$47,$F8,$0C,$FA,$27,$E0,$8F,$E6,$C7,$97,$0F,$6E,$1E,$57
	.dc.b	$E1,$9F,$5E,$BA,$61,$E2,$CF,$FD,$55,$9F,$D5,$C3,$E0,$9C,$CF,$AF
	.dc.b	$5F,$01,$F0,$9A,$3E,$69,$F4,$4D,$DF,$81,$F8,$58,$7D,$17,$D0,$7E
	.dc.b	$90,$F8,$47,$D1,$BF,$E5,$EF,$F8,$AF,$1F,$83,$7F,$E0,$C8,$F8,$D7
	.dc.b	$FA,$83,$D8,$7C,$07,$C4,$68,$FD,$79,$E7,$3F,$81,$47,$EB,$EF,$F2
	.dc.b	$F1,$B3,$F2,$47,$87,$1F,$8D,$8F,$FA,$22,$7D,$53,$CD,$E3,$87,$FB
	.dc.b	$BA,$75,$3F,$A4,$9D,$CF,$AB,$0F,$E6,$D3,$F1,$0D,$13,$0F,$23,$EA
	.dc.b	$C7,$EA,$E7,$FE,$5A,$CF,$F1,$38,$77,$8F,$E9,$B5,$C7,$0F,$86,$FD
	.dc.b	$22,$1E,$D3,$F1,$E7,$E4,$E3,$F5,$D8,$FC,$2C,$3B,$CF,$F6,$A9,$CB
	.dc.b	$E3,$CB,$DB,$97,$E8,$32,$F1,$E7,$8C,$BF,$01,$7F,$14,$F2,$1E,$2C
	.dc.b	$FD,$0F,$E6,$EF,$96,$3D,$9F,$82,$7E,$3C,$95,$EC,$9E,$91,$A8,$FC
	.dc.b	$03,$A8,$F9,$8F,$47,$FC,$4B,$9F,$B7,$2F,$34,$7E,$48,$FC,$1D,$72
	.dc.b	$23,$C9,$F5,$BF,$1C,$E6,$7C,$3B,$FA,$37,$DC,$FF,$0A,$C7,$E7,$2F
	.dc.b	$F6,$4F,$33,$C0,$FE,$13,$1F,$C6,$D7,$2B,$F7,$57,$F0,$B8,$F8,$91
	.dc.b	$F3,$33,$FD,$2E,$1F,$C1,$A3,$CF,$5D,$CF,$DA,$CE,$A3,$F4,$D1,$F4
	.dc.b	$6F,$E0,$BF,$83,$3E,$09,$E4,$3F,$53,$97,$D6,$BF,$D8,$47,$C1,$7D
	.dc.b	$B1,$9A,$FC,$DC,$FD,$47,$EB,$47,$E7,$E3,$F6,$37,$E3,$7F,$96,$46
	.dc.b	$5F,$F0,$0E,$1F,$9D,$3D,$EF,$E6,$27,$F2,$77,$EF,$8F,$33,$F8,$17
	.dc.b	$59,$7B,$1F,$36,$1F,$D6,$5F,$89,$3F,$85,$7E,$06,$5F,$BA,$7E,$71
	.dc.b	$F7,$23,$57,$E7,$8F,$E5,$A7,$F1,$7C,$7E,$03,$FC,$30,$EE,$7F,$73
	.dc.b	$3B,$DF,$D7,$C3,$F9,$64,$76,$BF,$C2,$DF,$B2,$BE,$AC,$FC,$93,$EB
	.dc.b	$63,$F8,$37,$F1,$06,$49,$87,$C9,$7F,$A4,$D7,$9D,$FE,$67,$7F,$15
	.dc.b	$F1,$9C,$A3,$4F,$E0,$0C,$91,$F2,$3F,$D8,$4F,$45,$FC,$B8,$CC,$7C
	.dc.b	$43,$95,$78,$DF,$E7,$58,$7C,$0C,$FF,$82,$E3,$EB,$9F,$A2,$7F,$04
	.dc.b	$3D,$79,$FE,$89,$FE,$09,$8F,$E8,$AB,$8C,$60,$F8,$67,$D0,$3F,$AB
	.dc.b	$1E,$9B,$FC,$14,$7F,$5B,$C3,$C9,$9F,$EC,$AB,$C3,$9F,$ED,$F0,$FC
	.dc.b	$35,$7D,$C9,$FE,$83,$3F,$8F,$7E,$2C,$65,$FE,$53,$7D,$71,$F4,$57
	.dc.b	$E9,$A3,$19,$7E,$6A,$BB,$E1,$FC,$2A,$73,$1F,$8F,$8F,$C3,$4F,$E8
	.dc.b	$63,$07,$E7,$1F,$90,$FB,$CF,$C9,$5F,$EE,$B3,$FB,$79,$7F,$0B,$AF
	.dc.b	$04,$7D,$6A,$FC,$1D,$FE,$D2,$FC,$EF,$EE,$23,$A1,$FE,$23,$7F,$85
	.dc.b	$CF,$18,$FC,$31,$F8,$23,$F3,$CE,$6F,$F5,$58,$7D,$6C,$7F,$BB,$4F
	.dc.b	$92,$BE,$F6,$5F,$49,$FE,$2B,$9F,$E1,$73,$F8,$D3,$F3,$67,$FC,$4F
	.dc.b	$87,$C6,$C3,$F4,$F7,$F7,$2F,$FC,$8B,$5F,$18,$F9,$A7,$E1,$2B,$86
	.dc.b	$5F,$97,$AE,$13,$FD,$EE,$BE,$2C,$62,$FE,$F6,$1E,$57,$F8,$0C,$79
	.dc.b	$6B,$B4,$F9,$72,$FC,$54,$7C,$87,$FC,$03,$9F,$D3,$AE,$17,$FC,$9A
	.dc.b	$31,$87,$DD,$8C,$3F,$A3,$CF,$E3,$47,$F8,$D2,$3C,$A7,$E0,$33,$F5
	.dc.b	$E3,$F4,$30,$F4,$67,$FC,$23,$3F,$43,$F8,$A9,$FE,$F0,$7A,$32,$F0
	.dc.b	$B8,$C7,$C2,$FE,$AC,$F9,$27,$F0,$D3,$EF,$1F,$AC,$CB,$EF,$3F,$B7
	.dc.b	$32,$7F,$8E,$5C,$C7,$EA,$E3,$F9,$15,$FC,$77,$FC,$8E,$7C,$73,$FA
	.dc.b	$6D,$79,$27,$F5,$D1,$D0,$FC,$91,$FC,$5F,$0F,$97,$87,$ED,$AB,$83
	.dc.b	$FC,$A2,$78,$67,$FD,$F7,$3F,$90,$7E,$1B,$0F,$C4,$B8,$9F,$C1,$1F
	.dc.b	$89,$CB,$D3,$7F,$C3,$E7,$F2,$78,$76,$AF,$E4,$59,$FE,$76,$BF,$AF
	.dc.b	$DF,$D8,$7F,$59,$7F,$62,$BE,$61,$EF,$C3,$F9,$96,$7F,$0D,$FE,$A5
	.dc.b	$19,$AF,$EF,$33,$97,$F8,$35,$7A,$5F,$CD,$3E,$9A,$FC,$BE,$3F,$BA
	.dc.b	$BF,$EA,$58,$FE,$19,$FB,$19,$FE,$19,$FE,$DF,$3B,$7F,$8E,$E1,$F8
	.dc.b	$98,$C4,$7F,$1C,$C3,$C0,$FE,$3D,$F0,$4E,$DF,$76,$1F,$C7,$DD,$CF
	.dc.b	$81,$DB,$E0,$77,$3F,$C3,$F0,$F3,$4F,$E2,$70,$F3,$3F,$D2,$F3,$FD
	.dc.b	$5D,$FE,$96,$7F,$03,$3F,$8C,$C7,$EA,$3B,$3F,$86,$BE,$A7,$F8,$4C
	.dc.b	$F9,$1E,$CF,$91,$E1,$8F,$D4,$9E,$71,$FC,$98,$FF,$06,$E3,$F5,$23
	.dc.b	$31,$FC,$8B,$1F,$C1,$3F,$E1,$CC,$3C,$AF,$F5,$2C,$BE,$D3,$FC,$4F
	.dc.b	$1F,$13,$F9,$C3,$F5,$F9,$FE,$25,$EE,$7F,$80,$0F,$DA,$39,$7F,$B2
	.dc.b	$C7,$CA,$7E,$FB,$D3,$3F,$D0,$47,$E4,$30,$F9,$AF,$7A,$FC,$EB,$D8
	.dc.b	$E6,$7E,$87,$0F,$A2,$FC,$13,$FA,$FE,$3F,$02,$FF,$9E,$E1,$FA,$43
	.dc.b	$EA,$1F,$C6,$F2,$FD,$A3,$8A,$FE,$1D,$18,$7D,$77,$F1,$9F,$EB,$33
	.dc.b	$BC,$FF,$80,$E5,$F3,$27,$FC,$86,$72,$3F,$01,$3B,$BF,$E0,$13,$B9
	.dc.b	$FF,$10,$DF,$57,$E7,$67,$F4,$9F,$E3,$A7,$EB,$CF,$9D,$87,$F5,$69
	.dc.b	$E1,$8F,$E4,$27,$CF,$8F,$B5,$F0,$98,$3C,$87,$DF,$CB,$E3,$3F,$C9
	.dc.b	$71,$F2,$3F,$E2,$29,$E4,$ED,$F5,$3F,$C3,$1F,$98,$F5,$C3,$E1,$C6
	.dc.b	$1F,$33,$CA,$3E,$49,$BC,$BE,$1C,$6D,$FE,$2B,$8F,$B2,$7F,$97,$CF
	.dc.b	$81,$FD,$0C,$FF,$18,$7F,$05,$1F,$A5,$7F,$05,$97,$F4,$C7,$F9,$94
	.dc.b	$78,$5F,$9B,$8F,$C0,$C3,$FA,$04,$7B,$4F,$21,$ED,$3F,$A9,$4F,$85
	.dc.b	$FC,$71,$FC,$E0,$FF,$55,$1F,$99,$7F,$38,$E2,$3F,$87,$3F,$CE,$1F
	.dc.b	$E7,$6E,$CC,$5F,$F2,$67,$11,$FC,$9A,$7F,$51,$1F,$C8,$0F,$E2,$87
	.dc.b	$C0,$37,$3F,$C7,$5C,$4F,$80,$E4,$E2,$3F,$8D,$4F,$9A,$7F,$CD,$D9
	.dc.b	$FE,$BA,$BF,$B5,$65,$F8,$2C,$FF,$CA,$13,$FB,$B9,$F8,$15,$C3,$3F
	.dc.b	$D8,$C7,$F9,$AB,$2F,$E3,$57,$FE,$0E,$C3,$F7,$6F,$EF,$AF,$C9,$97
	.dc.b	$9D,$F8,$66,$2F,$DA,$E2,$3C,$F1,$F1,$8F,$E7,$87,$C0,$36,$FC,$98
	.dc.b	$E5,$1F,$86,$3F,$4F,$1F,$8C,$CB,$F0,$55,$FA,$7A,$FC,$AD,$7E,$90
	.dc.b	$FC,$AB,$83,$F3,$6F,$E0,$8F,$C2,$9D,$4F,$E2,$6F,$9D,$FE,$3A,$79
	.dc.b	$CF,$E9,$87,$9C,$E1,$9F,$E5,$5F,$E5,$79,$7E,$7E,$7F,$9E,$5F,$E3
	.dc.b	$E7,$F9,$C3,$FA,$57,$F9,$C3,$EA,$7F,$4A,$79,$0F,$CA,$E3,$F6,$5F
	.dc.b	$F2,$BC,$6E,$7A,$C6,$CE,$A7,$B5,$F8,$18,$FE,$D2,$30,$FC,$97,$E3
	.dc.b	$1F,$AA,$9F,$36,$3F,$CB,$A3,$67,$98,$DB,$F3,$E7,$95,$7C,$C3,$FC
	.dc.b	$76,$78,$27,$F6,$0E,$23,$67,$8D,$C4,$75,$79,$47,$F3,$77,$94,$7D
	.dc.b	$53,$F2,$F1,$FE,$84,$C3,$F2,$F5,$FE,$8F,$9F,$D2,$1F,$62,$BF,$4E
	.dc.b	$7D,$B7,$F6,$07,$D8,$CB,$F3,$AF,$F9,$13,$1F,$CE,$BF,$BF,$C3,$F2
	.dc.b	$F5,$C9,$FC,$A1,$83,$F4,$06,$23,$C0,$E1,$FC,$EC,$7E,$69,$FE,$37
	.dc.b	$5E,$08,$FF,$18,$D7,$81,$FD,$96,$3F,$C8,$6B,$F6,$4F,$F0,$87,$FC
	.dc.b	$48,$7E,$4C,$FF,$05,$9F,$93,$3A,$BB,$7F,$C2,$26,$CF,$EE,$58,$FC
	.dc.b	$87,$FC,$2F,$1B,$C7,$F7,$CF,$F0,$BC,$7E,$73,$E3,$8F,$9D,$7F,$2E
	.dc.b	$BC,$6F,$C5,$7D,$F5,$F1,$5F,$DD,$98,$3E,$29,$83,$F5,$71,$F1,$5F
	.dc.b	$79,$F6,$8F,$AE,$7B,$0C,$1E,$08,$F9,$CF,$27,$C0,$FD,$DA,$F6,$3F
	.dc.b	$75,$F6,$3F,$ED,$B3,$E4,$1E,$09,$F6,$3F,$21,$F6,$3D,$4F,$49,$D4
	.dc.b	$F1,$9B,$9F,$1B,$F5,$23,$D9,$5F,$22,$7F,$0D,$3E,$C3,$C0,$E1,$F6
	.dc.b	$1F,$14,$E4,$FB,$1F,$85,$3E,$69,$FA,$F1,$F7,$5F,$E6,$93,$B8,$EA
	.dc.b	$7D,$13,$C1,$1E,$37,$07,$80,$F6,$4F,$57,$E4,$1D,$4F,$D2,$1F,$08
	.dc.b	$F1,$CF,$A5,$F8,$47,$A4,$FC,$7D,$7E,$25,$FC,$79,$F2,$2B,$D3,$5F
	.dc.b	$3A,$3D,$E7,$53,$D2,$7D,$13,$D2,$72,$7C,$C7,$EA,$A7,$EE,$C3,$2F
	.dc.b	$F1,$86,$7F,$D7,$F2,$FE,$55,$7F,$EE,$CB,$FE,$55,$8F,$F9,$B5,$FE
	.dc.b	$43,$87,$F8,$A5,$FE,$6C,$FF,$90,$A7,$F8,$AB,$FE,$32,$7F,$8A,$BF
	.dc.b	$CD,$A3,$6F,$F9,$09,$EA,$FF,$58,$3F,$8F,$4F,$F2,$13,$F8,$F1,$FC
	.dc.b	$91,$FE,$4E,$E1,$D9,$83,$C0,$ED,$FE,$49,$5E,$C8,$FF,$78,$57,$9A
	.dc.b	$7F,$C4,$4F,$99,$FE,$8A,$FF,$30,$7D,$91,$E3,$7A,$9E,$03,$AD,$7B
	.dc.b	$1F,$F2,$51,$FD,$08,$FE,$AF,$3D,$71,$FE,$90,$F5,$8F,$84,$6D,$FE
	.dc.b	$AE,$6D,$F9,$C7,$F4,$E3,$C6,$3E,$D1,$F8,$DB,$C5,$60,$A6,$78,$78
	.dc.b	$56,$E7,$53,$9B,$E7,$E2,$EE,$31,$5B,$BE,$77,$F4,$5F,$1D,$F8,$BC
	.dc.b	$5E,$13,$E1,$8F,$2F,$8B,$8A,$F0,$D5,$9C,$D7,$88,$E9,$3F,$13,$36
	.dc.b	$BC,$F1,$78,$BC,$57,$8B,$63,$0F,$7B,$FB,$B5,$F8,$9C,$7E,$3F,$8B
	.dc.b	$C4,$E5,$E2,$DC,$FC,$57,$F7,$8F,$6E,$4F,$17,$C0,$78,$7C,$5B,$5C
	.dc.b	$2B,$B9,$F3,$BC,$5F,$31,$E7,$F1,$6C,$F6,$CF,$EC,$F8,$BE,$B3,$D9
	.dc.b	$E2,$D9,$E9,$87,$F8,$DB,$6F,$85,$E2,$FC,$9F,$C7,$1F,$25,$F0,$E3
	.dc.b	$0E,$7E,$23,$5E,$3D,$C7,$C1,$68,$F8,$D8,$F5,$3F,$0B,$87,$2A,$F3
	.dc.b	$70,$A3,$13,$F8,$23,$F2,$56,$F1,$FC,$87,$97,$C7,$B8,$7A,$AB,$FC
	.dc.b	$54,$3A,$3E,$EB,$CB,$4E,$BD,$7F,$90,$CA,$DF,$E2,$F2,$78,$FF,$04
	.dc.b	$F8,$5E,$15,$9E,$DE,$07,$F8,$C7,$2F,$2F,$8F,$F1,$CF,$91,$E3,$DA
	.dc.b	$F8,$7E,$3F,$CB,$F8,$8F,$98,$D8,$E7,$7F,$BA,$E5,$F1,$61,$C6,$3E
	.dc.b	$17,$43,$56,$59,$DF,$F5,$90,$F8,$8F,$B1,$F1,$1F,$65,$B9,$F6,$E3
	.dc.b	$FE,$03,$A7,$68,$FA,$FA,$7A,$1A,$7D,$AF,$88,$FC,$17,$C4,$7E,$0D
	.dc.b	$B1,$CE,$BF,$9C,$6D,$F3,$DF,$95,$F1,$1F,$89,$3F,$15,$F1,$16,$BF
	.dc.b	$0B,$F1,$1F,$8D,$3F,$1D,$F1,$16,$3A,$61,$FC,$DF,$8B,$E0,$61,$F9
	.dc.b	$3F,$88,$E5,$3F,$13,$85,$1F,$97,$F8,$89,$87,$E8,$7E,$0B,$F8,$B3
	.dc.b	$F9,$48,$FC,$9D,$FE,$1B,$87,$F0,$B9,$76,$AF,$CC,$E1,$DB,$93,$E6
	.dc.b	$BF,$C2,$F1,$7A,$14,$C7,$A5,$6F,$D3,$97,$F6,$18,$FE,$49,$97,$E4
	.dc.b	$A3,$F9,$40,$EA,$FE,$55,$19,$F4,$95,$C3,$8F,$F8,$F6,$4F,$88,$F1
	.dc.b	$1E,$3F,$88,$B9,$FB,$FE,$23,$CD,$7E,$76,$C7,$93,$1E,$AF,$82,$7C
	.dc.b	$7A,$7F,$3B,$D3,$0F,$F3,$DA,$DB,$F7,$4F,$E8,$99,$78,$EF,$E7,$65
	.dc.b	$E3,$69,$F4,$7E,$23,$E4,$7C,$47,$C9,$6C,$78,$F5,$F7,$61,$E2,$CB
	.dc.b	$C2,$FC,$99,$F1,$F1,$32,$F0,$4F,$E9,$17,$C4,$7C,$28,$C3,$F5,$E2
	.dc.b	$57,$E0,$FE,$22,$D7,$F0,$D3,$F9,$5D,$F5,$F8,$8F,$BC,$7D,$FF,$88
	.dc.b	$B7,$F8,$BF,$88,$FC,$17,$C4,$7E,$0D,$B1,$EB,$3D,$38,$7C,$B3,$E4
	.dc.b	$1E,$B3,$E5,$61,$DE,$3F,$2D,$87,$76,$9F,$99,$F8,$8F,$C8,$7C,$47
	.dc.b	$E4,$5B,$1D,$EF,$A9,$DF,$9F,$F1,$5C,$1F,$33,$1E,$D3,$F1,$FE,$23
	.dc.b	$D6,$DB,$FD,$37,$C4,$7E,$74,$FC,$F7,$C4,$58,$ED,$8F,$E6,$6B,$B7
	.dc.b	$47,$E6,$BA,$3F,$0B,$3F,$AA,$E3,$7C,$47,$E9,$CF,$EE,$5F,$8E,$FF
	.dc.b	$82,$31,$FD,$A4,$63,$E4,$A4,$7F,$57,$97,$C7,$E3,$3F,$AC,$7F,$49
	.dc.b	$DF,$F0,$58,$75,$BE,$05,$9F,$E4,$F6,$F8,$DF,$0F,$C5,$F8,$9B,$5E
	.dc.b	$CD,$BF,$43,$7F,$B3,$3E,$BE,$DE,$C8,$FD,$8E,$3D,$7C,$3F,$F0,$D6
	.dc.b	$5E,$BB,$FC,$94,$FB,$BE,$2F,$DF,$F1,$37,$2F,$63,$FC,$43,$BB,$FA
	.dc.b	$46,$CC,$5F,$13,$F1,$8F,$8F,$F1,$17,$0F,$89,$F1,$3F,$2B,$E2,$7E
	.dc.b	$5B,$63,$D5,$AF,$EA,$4F,$4F,$57,$E0,$71,$EB,$3F,$37,$1E,$AD,$3E
	.dc.b	$87,$C4,$FD,$6F,$89,$FA,$ED,$8E,$B8,$7F,$18,$D7,$F6,$BC,$7F,$16
	.dc.b	$3F,$C8,$D8,$F6,$7E,$8E,$3D,$9B,$7F,$77,$E2,$7F,$07,$F1,$3F,$84
	.dc.b	$6C,$76,$9F,$C6,$9F,$C4,$74,$F5,$7C,$4F,$E2,$BE,$27,$F1,$6D,$AF
	.dc.b	$C4,$7C,$4F,$E3,$BE,$27,$F1,$ED,$8E,$98,$7D,$DE,$FF,$D2,$C7,$F0
	.dc.b	$CD,$58,$73,$AF,$C4,$B7,$A7,$EF,$FC,$4F,$E6,$BE,$0F,$F3,$6D,$C3
	.dc.b	$AF,$17,$C4,$DB,$CF,$1F,$9B,$DB,$CE,$D3,$FC,$69,$F8,$F7,$F8,$B1
	.dc.b	$BF,$15,$FE,$57,$8C,$69,$FE,$00,$51,$F8,$98,$FF,$03,$1F,$E7,$58
	.dc.b	$FF,$17,$4F,$4F,$88,$E5,$F1,$5C,$DE,$9F,$13,$D7,$E2,$2D,$78,$9E
	.dc.b	$9C,$5F,$E0,$4C,$7A,$61,$C6,$BC,$ED,$8F,$27,$C4,$79,$CF,$47,$C1
	.dc.b	$DA,$F8,$73,$F9,$27,$F9,$CE,$7E,$68,$F6,$67,$E6,$9F,$41,$F1,$3E
	.dc.b	$1B,$E1,$1F,$0F,$E1,$EC,$74,$CF,$EC,$DF,$F2,$AC,$1F,$11,$F2,$7E
	.dc.b	$1F,$E5,$36,$FE,$7F,$C4,$7C,$D3,$E7,$7C,$3D,$8E,$9E,$57,$F1,$BE
	.dc.b	$5F,$E5,$18,$FF,$8A,$E7,$EB,$E3,$D1,$A7,$DA,$F8,$8F,$B0,$7D,$9F
	.dc.b	$85,$B1,$D3,$4F,$CD,$3F,$0A,$BC,$91,$F7,$74,$BD,$7F,$EA,$5C,$5F
	.dc.b	$11,$F8,$3F,$88,$FC,$23,$71,$E9,$87,$F4,$BE,$0F,$F0,$CD,$BE,$23
	.dc.b	$F1,$47,$E2,$FE,$1A,$DF,$E4,$3E,$1F,$F1,$C7,$FA,$97,$F1,$FF,$F1
	.dc.b	$2E,$3F,$8F,$8F,$C1,$F1,$7F,$8F,$32,$EB,$8F,$2A,$FC,$A6,$37,$5F
	.dc.b	$F5,$E6,$3F,$E2,$4C,$BF,$D2,$0D,$AF,$F2,$CB,$FE,$0A,$9D,$FC,$47
	.dc.b	$69,$FF,$1F,$69,$E9,$BF,$D2,$E3,$E9,$69,$D3,$E1,$BD,$1F,$11,$E9
	.dc.b	$6C,$7C,$4F,$33,$F8,$A7,$A4,$D1,$F1,$1E,$FE,$0F,$80,$DB,$F3,$FC
	.dc.b	$3F,$C3,$F8,$8F,$88,$D8,$E9,$D5,$FD,$AF,$0E,$71,$EE,$C3,$9B,$4F
	.dc.b	$83,$F1,$1F,$34,$F9,$DE,$F5,$8E,$67,$F8,$72,$B9,$F0,$FC,$6B,$E6
	.dc.b	$D9,$F9,$7F,$11,$F5,$CF,$B1,$F0,$B6,$39,$CF,$F8,$32,$DF,$11,$F7
	.dc.b	$7E,$1F,$EF,$36,$BE,$AF,$C4,$7E,$07,$E1,$FF,$04,$D8,$E7,$D7,$FE
	.dc.b	$64,$C3,$9C,$7D,$CC,$39,$B4,$FB,$FF,$11,$F8,$BF,$88,$FC,$63,$63
	.dc.b	$9E,$DF,$EF,$7B,$E7,$A7,$E1,$30,$E6,$DA,$FC,$4F,$C4,$7E,$57,$E2
	.dc.b	$3F,$2C,$D8,$E7,$2E,$7F,$BD,$F0,$9F,$EF,$FF,$13,$FB,$A7,$81,$F9
	.dc.b	$48,$C5,$E7,$E2,$35,$F1,$16,$3F,$DC,$D7,$FE,$10,$7A,$E1,$FC,$31
	.dc.b	$E2,$7E,$72,$EE,$3F,$EB,$4A,$FF,$07,$4F,$2F,$87,$F1,$7C,$47,$8D
	.dc.b	$FC,$F5,$7E,$DF,$83,$FC,$1F,$EC,$7F,$47,$BF,$F2,$EE,$1E,$85,$5E
	.dc.b	$D8,$C3,$EF,$E1,$FF,$16,$CF,$A0,$F8,$1F,$11,$1F,$5F,$C4,$5A,$FF
	.dc.b	$14,$9F,$E9,$FC,$7A,$EB,$FC,$A3,$E0,$FE,$41,$F2,$7E,$22,$E1,$F1
	.dc.b	$3E,$1F,$E6,$1F,$37,$E2,$2C,$7C,$6D,$7F,$C0,$1C,$88,$EB,$0F,$87
	.dc.b	$E7,$A7,$C7,$7E,$B7,$C4,$4A,$FA,$3F,$0F,$6B,$CF,$C5,$FE,$26,$C3
	.dc.b	$AD,$74,$C3,$9E,$3D,$EF,$E3,$47,$E0,$7E,$22,$57,$DC,$F8,$5B,$5D
	.dc.b	$0F,$ED,$3C,$1F,$D2,$3E,$0F,$F1,$1F,$0F,$F8,$96,$DF,$E1,$3E,$23
	.dc.b	$F1,$9F,$0D,$F8,$D6,$C7,$C1,$9D,$E1,$F8,$36,$E9,$E5,$C7,$FA,$AF
	.dc.b	$C3,$F9,$E3,$1F,$90,$7F,$7E,$75,$BA,$75,$8F,$D6,$AD,$3F,$88,$C7
	.dc.b	$D3,$4C,$7B,$CF,$D0,$6C,$FE,$96,$BF,$C1,$33,$F8,$CC,$3C,$3F,$11
	.dc.b	$DA,$B5,$F0,$D7,$36,$1C,$27,$E4,$78,$FF,$E0,$59,$CF,$C2,$6E,$8D
	.dc.b	$F8,$8C,$39,$F8,$89,$A7,$84,$F6,$7C,$46,$32,$EB,$DD,$FD,$3C,$F8
	.dc.b	$17,$CA,$B8,$4F,$AF,$0C,$2D,$BE,$5F,$C4,$78,$CF,$27,$C4,$5B,$FD
	.dc.b	$DE,$9E,$67,$9E,$BE,$A6,$C3,$1E,$B8,$7D,$6F,$88,$F7,$1E,$FE,$3B
	.dc.b	$7D,$3E,$13,$E1,$7C,$47,$C3,$6C,$74,$F8,$8F,$8C,$7C,$7F,$88,$B1
	.dc.b	$C4,$F0,$D7,$16,$9E,$4F,$88,$F9,$DF,$0F,$F3,$DB,$1C,$78,$7E,$13
	.dc.b	$F8,$7A,$E5,$18,$3F,$04,$FD,$7F,$88,$FB,$1F,$09,$6B,$E0,$BF,$6F
	.dc.b	$E1,$E4,$7C,$4F,$88,$B8,$73,$E3,$7E,$43,$FE,$08,$E2,$FC,$F3,$6F
	.dc.b	$F8,$6B,$FE,$01,$8C,$D7,$E3,$63,$FC,$05,$3A,$F8,$1C,$7C,$36,$5D
	.dc.b	$7C,$3C,$9F,$AF,$5F,$D9,$B0,$EF,$3C,$FA,$E5,$6B,$E1,$2E,$5E,$7C
	.dc.b	$1F,$0B,$E1,$3C,$5F,$0B,$6F,$AF,$C2,$F9,$7E,$17,$CC,$D8,$F0,$FC
	.dc.b	$2F,$A7,$E1,$7D,$4D,$8F,$2F,$C0,$FB,$7E,$13,$DC,$D8,$F4,$FC,$2F
	.dc.b	$C1,$3E,$17,$C2,$58,$F6,$FC,$2F,$C5,$F8,$4F,$8C,$D8,$F8,$3F,$0B
	.dc.b	$F2,$4F,$95,$F0,$76,$3A,$93,$8B,$E0,$61,$D5,$F8,$57,$CD,$B3,$F3
	.dc.b	$30,$FE,$01,$F9,$17,$F4,$EE,$A3,$47,$C3,$6A,$BA,$74,$FE,$03,$5F
	.dc.b	$5C,$78,$63,$F0,$79,$77,$6C,$77,$3F,$63,$7F,$DB,$F2,$FC,$63,$67
	.dc.b	$F9,$43,$FD,$B2,$33,$F0,$DE,$6F,$86,$97,$CB,$E1,$2D,$72,$F8,$4F
	.dc.b	$59,$EC,$F8,$4B,$5E,$0F,$84,$F7,$9F,$03,$E1,$2C,$79,$31,$E7,$A7
	.dc.b	$8B,$2B,$3F,$DA,$34,$65,$D3,$3F,$4E,$77,$0F,$4F,$C2,$7C,$C3,$E6
	.dc.b	$FC,$0D,$C7,$E0,$FC,$27,$D0,$3E,$8F,$C2,$58,$F8,$BF,$07,$F5,$3E
	.dc.b	$0F,$EA,$B6,$3E,$4F,$BF,$F6,$3E,$0F,$EC,$B6,$3A,$93,$5F,$9B,$87
	.dc.b	$57,$E7,$DF,$36,$CF,$D4,$C3,$F0,$7F,$05,$F8,$4F,$83,$B3,$F6,$1F
	.dc.b	$C3,$FC,$1C,$8F,$E5,$7B,$7F,$56,$C3,$FC,$69,$C5,$FD,$6E,$73,$F0
	.dc.b	$7F,$D3,$6B,$FA,$67,$0F,$F6,$57,$FD,$88,$D8,$D7,$BF,$BF,$72,$DF
	.dc.b	$67,$C1,$8F,$E2,$32,$E9,$F1,$B8,$C7,$95,$F8,$0F,$2F,$C2,$78,$72
	.dc.b	$F1,$7C,$1C,$9F,$4F,$C1,$DC,$BA,$47,$E5,$B2,$EA,$D3,$A4,$7F,$23
	.dc.b	$D7,$D7,$1F,$E2,$C7,$D6,$DC,$3E,$07,$C1,$C8,$F8,$FF,$07,$6F,$97
	.dc.b	$C1,$7C,$5F,$82,$F8,$CD,$AF,$9B,$F0,$5F,$27,$DE,$F9,$4D,$8F,$A3
	.dc.b	$F0,$5F,$34,$F9,$DF,$05,$63,$EA,$DF,$3C,$7B,$C7,$D8,$C7,$9B,$F6
	.dc.b	$B1,$B3,$ED,$7E,$C7,$C0,$FD,$9F,$82,$B5,$F8,$07,$EE,$7C,$17,$DD
	.dc.b	$F8,$2B,$3F,$84,$7F,$01,$F0,$5F,$81,$F8,$2B,$3F,$88,$7F,$C2,$1F
	.dc.b	$91,$FF,$49,$CF,$FC,$F8,$DF,$44,$FF,$4B,$AF,$F2,$CB,$9E,$56,$5E
	.dc.b	$19,$FC,$9B,$6F,$EB,$63,$E1,$9F,$CB,$B6,$3F,$AE,$5F,$F5,$47,$C3
	.dc.b	$F0,$7E,$4F,$83,$F2,$BE,$6F,$83,$98,$7F,$82,$AF,$FA,$A3,$E4,$8F
	.dc.b	$5F,$C1,$FB,$3E,$0E,$CE,$DF,$7F,$C1,$7C,$0F,$82,$B3,$CD,$EB,$9F
	.dc.b	$93,$0F,$8B,$F1,$38,$F0,$BF,$C9,$99,$B3,$ED,$8F,$8D,$B8,$7B,$3E
	.dc.b	$0B,$E6,$FC,$1F,$CE,$6D,$7A,$72,$E8,$FC,$CC,$AD,$FD,$2C,$39,$E5
	.dc.b	$E0,$6D,$7D,$2F,$82,$FB,$3F,$07,$F6,$9B,$1F,$17,$E0,$BE,$EF,$C1
	.dc.b	$FD,$E6,$C7,$C9,$F7,$BF,$02,$7E,$0B,$E0,$6C,$7C,$DF,$7F,$F0,$BF
	.dc.b	$03,$F8,$66,$C7,$53,$E9,$61,$D5,$A7,$52,$6B,$F7,$AF,$AB,$F8,$0F
	.dc.b	$83,$FC,$8F,$C1,$FE,$49,$B5,$FE,$46,$FC,$93,$FB,$21,$FD,$89,$C1
	.dc.b	$FE,$67,$71,$5F,$E2,$57,$FA,$D4,$7F,$AB,$9F,$EB,$5C,$1F,$E8,$37
	.dc.b	$33,$FE,$D9,$73,$3B,$F8,$0E,$9F,$01,$D5,$B1,$E0,$3C,$9F,$09,$E1
	.dc.b	$7C,$4F,$8F,$E1,$3C,$9F,$15,$67,$D4,$F9,$FE,$0E,$47,$B7,$E0,$EE
	.dc.b	$7E,$9F,$85,$F6,$1E,$DF,$85,$B5,$F0,$FC,$6F,$81,$F0,$BF,$05,$B1
	.dc.b	$F1,$FE,$17,$E2,$7C,$1F,$C5,$6C,$7C,$BC,$39,$E3,$F1,$F1,$BC,$5F
	.dc.b	$17,$1E,$73,$F2,$DB,$1F,$27,$E0,$FE,$87,$C1,$FD,$16,$E3,$D4,$97
	.dc.b	$D5,$B5,$D5,$FB,$9A,$7A,$5A,$7D,$FF,$84,$FB,$9F,$0B,$F7,$5B,$1F
	.dc.b	$83,$F8,$5F,$C0,$1F,$81,$F8,$5B,$1F,$87,$F7,$BF,$D5,$5F,$92,$FF
	.dc.b	$63,$F4,$9D,$7F,$D8,$F5,$FF,$35,$4E,$4F,$E9,$2A,$51,$FD,$2A,$30
	.dc.b	$EA,$33,$F0,$5E,$09,$FC,$86,$5E,$3C,$7C,$C7,$8C,$FF,$4F,$DF,$8F
	.dc.b	$0F,$16,$DF,$ED,$58,$F1,$7C,$2F,$A3,$E1,$7D,$2D,$BE,$3F,$09,$EC
	.dc.b	$3D,$BF,$09,$63,$AF,$C1,$FC,$0F,$83,$F8,$2D,$8F,$0F,$C0,$7C,$43
	.dc.b	$E2,$FC,$05,$8E,$A4,$DF,$E2,$E1,$D5,$F8,$F7,$CD,$B3,$EC,$C3,$E7
	.dc.b	$FC,$1F,$D0,$F8,$1B,$3F,$01,$FA,$7F,$07,$F5,$3E,$0E,$CF,$C4,$7E
	.dc.b	$BF,$C1,$C8,$F8,$FF,$07,$72,$FA,$FF,$03,$F7,$3E,$0F,$EE,$B6,$BE
	.dc.b	$6E,$3D,$5A,$75,$3E,$8E,$1C,$F9,$7E,$86,$36,$58,$FD,$EF,$82,$FC
	.dc.b	$49,$F8,$AF,$83,$B3,$D2,$7F,$13,$A7,$A9,$B1,$D8,$8F,$E3,$AF,$B3
	.dc.b	$F8,$DF,$77,$FB,$61,$3E,$3B,$FE,$00,$38,$F8,$4A,$63,$E1,$2E,$1F
	.dc.b	$EE,$89,$7B,$FC,$0E,$3E,$FD,$9D,$7B,$FD,$3E,$13,$AB,$63,$96,$3D
	.dc.b	$33,$FF,$66,$62,$CB,$1C,$0B,$F8,$BE,$1A,$BC,$31,$E3,$D3,$C2,$D3
	.dc.b	$CD,$F0,$9E,$CF,$84,$F6,$B6,$3D,$5F,$07,$F0,$0F,$83,$76,$3D,$DF
	.dc.b	$05,$F1,$0F,$8B,$F0,$76,$3E,$17,$C1,$FC,$83,$E4,$FC,$05,$8E,$A4
	.dc.b	$DF,$DB,$87,$57,$DF,$87,$37,$E0,$E1,$76,$F8,$11,$F5,$FE,$0F,$EA
	.dc.b	$FC,$0F,$D6,$6E,$3F,$4F,$E0,$BE,$CF,$C1,$FD,$A6,$C7,$D7,$F8,$2F
	.dc.b	$BA,$7D,$EF,$83,$B1,$F9,$0C,$BB,$F2,$3F,$26,$FF,$61,$13,$E0,$BC
	.dc.b	$73,$A8,$FF,$63,$70,$32,$EF,$B7,$DA,$6D,$74,$96,$7E,$0A,$FB,$F9
	.dc.b	$DA,$E9,$F0,$5D,$B4,$FB,$F9,$7F,$66,$D3,$C9,$F0,$59,$C7,$F9,$34
	.dc.b	$E7,$3F,$15,$FE,$57,$6F,$84,$D3,$E1,$12,$BC,$2D,$CB,$D9,$F0,$7D
	.dc.b	$72,$F8,$5F,$09,$72,$4D,$8C,$7C,$26,$7E,$12,$E3,$F2,$CF,$4B,$C7
	.dc.b	$E0,$A4,$67,$3F,$4F,$C1,$61,$69,$F3,$7E,$0B,$C1,$F0,$1E,$16,$DF
	.dc.b	$1D,$BE,$B6,$7E,$18,$C6,$DF,$ED,$46,$E9,$D7,$1F,$B1,$A7,$AF,$4C
	.dc.b	$57,$CE,$F8,$3F,$77,$C2,$7B,$DB,$1D,$3E,$0F,$E1,$1F,$0F,$E1,$2C
	.dc.b	$78,$3E,$0F,$E3,$1F,$1F,$E1,$2C,$79,$3E,$07,$E5,$1F,$2F,$E1,$2C
	.dc.b	$75,$27,$27,$C2,$7F,$03,$87,$68,$C1,$F8,$0C,$3A,$B8,$E2,$FC,$C6
	.dc.b	$38,$DF,$F2,$DF,$09,$F6,$FF,$27,$FE,$E5,$C7,$FD,$83,$3F,$E1,$71
	.dc.b	$FD,$D4,$FE,$D8,$4F,$85,$E1,$F0,$FC,$7E,$0E,$57,$D1,$A7,$C2,$63
	.dc.b	$5F,$14,$3E,$13,$C1,$F0,$5E,$16,$DF,$1C,$7A,$BC,$F2,$B8,$72,$CB
	.dc.b	$A6,$BF,$B4,$CB,$BD,$F8,$67,$D9,$F0,$52,$3C,$9F,$05,$70,$F1,$E5
	.dc.b	$D3,$1E,$59,$5B,$F8,$38,$F6,$7E,$1E,$B7,$0F,$F0,$96,$6F,$84,$F9
	.dc.b	$5F,$09,$F2,$DB,$7F,$0F,$E1,$3E,$71,$F3,$FE,$12,$C7,$C7,$F8,$2F
	.dc.b	$A4,$7D,$3F,$7E,$C7,$CB,$F8,$2F,$AD,$F0,$1F,$5D,$B1,$D4,$9A,$7C
	.dc.b	$EC,$3A,$BF,$43,$0E,$6F,$D2,$C2,$F0,$7C,$9F,$83,$FC,$11,$F8,$3F
	.dc.b	$84,$B8,$FD,$BF,$84,$FC,$31,$F8,$7F,$84,$B1,$FA,$FF,$CA,$3F,$DA
	.dc.b	$47,$FB,$47,$2F,$F1,$B3,$8F,$85,$FE,$10,$FF,$B2,$7B,$CD,$3F,$AE
	.dc.b	$57,$3F,$82,$EB,$7F,$88,$C7,$C1,$3E,$33,$C1,$1D,$30,$EA,$D8,$F2
	.dc.b	$7C,$0F,$80,$F0,$FC,$2D,$8F,$47,$C2,$79,$0F,$2F,$C2,$D8,$F6,$7C
	.dc.b	$1F,$A3,$E1,$3D,$2D,$8F,$81,$F0,$9E,$C3,$DB,$EC,$58,$EA,$4C,$FF
	.dc.b	$65,$BB,$1E,$D1,$8C,$3A,$36,$BE,$66,$3F,$23,$E0,$A4,$7C,$FF,$82
	.dc.b	$B5,$D3,$4E,$7C,$DF,$D5,$B3,$ED,$7E,$18,$C6,$5D,$9B,$1D,$8F,$B5
	.dc.b	$F0,$9F,$5B,$E1,$3E,$BB,$63,$EF,$7C,$1F,$DA,$3E,$DF,$C2,$58,$FC
	.dc.b	$17,$C1,$FD,$E3,$EF,$FB,$F6,$3F,$0D,$F0,$7F,$82,$3F,$07,$F0,$76
	.dc.b	$3A,$93,$7F,$8F,$87,$57,$E4,$E1,$CD,$F9,$78,$5D,$3F,$BA,$F8,$9F
	.dc.b	$E2,$3C,$3F,$A1,$3F,$ED,$08,$CC,$7F,$59,$8C,$3F,$E2,$87,$FB,$51
	.dc.b	$3E,$0B,$5F,$07,$BF,$83,$E1,$F0,$76,$FA,$BC,$FE,$07,$A7,$C0,$D9
	.dc.b	$F0,$BC,$9F,$1D,$F8,$70,$C6,$9F,$CD,$72,$FE,$D3,$5D,$0F,$4E,$5E
	.dc.b	$07,$D7,$F0,$3E,$93,$D5,$F0,$36,$3D,$FF,$05,$ED,$F8,$3F,$73,$63
	.dc.b	$E1,$FC,$17,$C1,$F8,$3F,$84,$D8,$F8,$FF,$07,$F1,$4F,$8D,$F0,$36
	.dc.b	$3A,$9F,$33,$0E,$AD,$3A,$9F,$42,$FA,$B4,$FA,$7F,$03,$F4,$7E,$07
	.dc.b	$E9,$36,$3A,$EF,$F0,$EF,$AB,$63,$AB,$F1,$EF,$AB,$63,$EF,$FC,$1F
	.dc.b	$DD,$3E,$F7,$C1,$D8,$FC,$1F,$C1,$FE,$04,$FF,$0D,$7E,$53,$FE,$D1
	.dc.b	$E3,$FE,$C1,$87,$FB,$36,$3F,$D9,$47,$FB,$37,$D0,$38,$F7,$B8,$FB
	.dc.b	$DC,$BD,$EB,$97,$F8,$46,$BF,$A9,$E1,$F7,$72,$F0,$4E,$7E,$2A,$CF
	.dc.b	$F7,$5A,$E9,$3C,$7E,$03,$CB,$97,$9B,$E0,$E4,$F5,$F8,$4B,$87,$1F
	.dc.b	$84,$F5,$FC,$27,$B1,$B5,$E3,$F8,$2F,$7F,$C1,$FC,$06,$C7,$9F,$8B
	.dc.b	$E1,$9F,$13,$E0,$6C,$7A,$EF,$9E,$3D,$E3,$DD,$8F,$37,$E0,$63,$67
	.dc.b	$D8,$FC,$EF,$81,$F9,$FF,$03,$6B,$E3,$3F,$4B,$E0,$BE,$9F,$C1,$D9
	.dc.b	$F9,$4F,$D6,$F8,$3F,$AF,$F0,$76,$7E,$73,$F6,$BE,$07,$ED,$FC,$1D
	.dc.b	$9F,$A4,$F3,$D3,$E7,$61,$7C,$EF,$F3,$DF,$3A,$3A,$4F,$D3,$C6,$D6
	.dc.b	$9F,$F3,$07,$E5,$5F,$ED,$2C,$7F,$C9,$CD,$AF,$F1,$73,$FD,$87,$B4
	.dc.b	$E6,$FB,$EB,$EE,$BF,$C3,$36,$FF,$05,$AF,$BA,$7F,$14,$D8,$EB,$F0
	.dc.b	$9E,$23,$C7,$F1,$16,$BC,$7F,$09,$E6,$3C,$FF,$09,$63,$CF,$EF,$7A
	.dc.b	$BE,$0B,$D6,$D8,$F5,$FB,$DE,$E3,$DF,$F0,$56,$3A,$93,$3F,$7E,$1D
	.dc.b	$5F,$83,$7C,$DB,$3F,$1B,$0F,$93,$F0,$92,$3E,$4F,$C2,$5A,$F9,$3F
	.dc.b	$09,$F3,$8F,$9F,$F0,$96,$BE,$7F,$C1,$7D,$23,$E9,$FC,$1D,$8F,$A7
	.dc.b	$F0,$7F,$58,$FA,$FF,$03,$63,$EB,$E1,$CF,$1F,$AF,$8D,$E6,$FF,$5E
	.dc.b	$F3,$B3,$E9,$5F,$73,$3B,$5A,$F8,$1F,$C1,$9F,$84,$F8,$1B,$9F,$E1
	.dc.b	$3E,$0F,$F0,$FF,$07,$F8,$86,$C7,$FA,$57,$F2,$BF,$EA,$D9,$7F,$9D
	.dc.b	$5F,$EA,$6B,$B5,$1A,$CB,$FD,$A2,$EA,$33,$F0,$3F,$EB,$38,$FF,$61
	.dc.b	$0F,$81,$C4,$78,$70,$FC,$BE,$57,$1F,$CB,$BC,$63,$97,$C4,$D7,$CC
	.dc.b	$F5,$F8,$29,$1E,$9F,$85,$B7,$FA,$0D,$7D,$AD,$9E,$E7,$BC,$EA,$53
	.dc.b	$A9,$F0,$FE,$17,$D4,$7A,$FE,$16,$C7,$C7,$F8,$5F,$77,$C2,$FB,$DB
	.dc.b	$1F,$2F,$E1,$7E,$11,$F0,$FE,$02,$C7,$CF,$F8,$5F,$8D,$F0,$3F,$1D
	.dc.b	$B1,$D4,$FA,$98,$75,$69,$D4,$99,$FD,$7B,$EA,$FD,$9F,$85,$FA,$47
	.dc.b	$D3,$F8,$5B,$5F,$7F,$E1,$3E,$B7,$C2,$FD,$76,$C7,$E0,$FE,$17,$ED
	.dc.b	$7C,$2F,$DB,$6C,$7E,$1F,$E0,$FE,$F1,$FE,$F1,$CD,$F0,$78,$9E,$C7
	.dc.b	$E3,$71,$EC,$D3,$B1,$F9,$1C,$3B,$34,$FF,$74,$53,$E0,$FB,$1D,$FE
	.dc.b	$0E,$CE,$BE,$17,$C4,$78,$FE,$16,$C7,$2F,$85,$F3,$7C,$2F,$9D,$B1
	.dc.b	$DB,$E0,$BD,$5F,$05,$EB,$6C,$75,$27,$27,$FB,$67,$4F,$ED,$7E,$2A
	.dc.b	$53,$FA,$79,$67,$DE,$7D,$79,$5A,$E4,$FC,$9F,$85,$93,$F0,$3E,$0E
	.dc.b	$D7,$7B,$F5,$BE,$1D,$6E,$5E,$18,$F0,$6B,$EC,$7C,$CD,$8D,$78,$ED
	.dc.b	$1F,$0B,$F6,$4F,$B5,$F0,$B7,$4F,$A3,$F0,$BF,$74,$FB,$DF,$0B,$63
	.dc.b	$EA,$FC,$07,$E0,$7E,$0B,$F0,$4D,$8F,$B3,$F0,$3F,$85,$3F,$0D,$F0
	.dc.b	$56,$3A,$93,$7F,$99,$87,$57,$E7,$5F,$36,$CF,$E1,$30,$FC,$87,$C2
	.dc.b	$FE,$47,$E1,$6C,$FE,$21,$F8,$9C,$1F,$03,$F2,$CF,$EC,$13,$F5,$B8
	.dc.b	$51,$FD,$94,$CE,$3F,$C7,$DC,$7C,$2F,$2A,$E1,$1F,$EC,$49,$5F,$5A
	.dc.b	$D4,$71,$F8,$5B,$87,$6F,$85,$B8,$1B,$F0,$78,$73,$7A,$F8,$8B,$9F
	.dc.b	$E3,$33,$F5,$47,$2C,$F9,$E7,$6B,$47,$99,$EF,$F0,$9E,$0F,$84,$B3
	.dc.b	$AC,$7D,$D1,$E4,$F8,$69,$39,$DB,$E2,$FC,$36,$2F,$DF,$3F,$0B,$E1
	.dc.b	$BD,$67,$B3,$E1,$AC,$71,$E1,$7C,$27,$C0,$F8,$4F,$82,$D9,$D6,$7D
	.dc.b	$78,$7E,$B6,$76,$7B,$1C,$F2,$EC,$D3,$B7,$C1,$FC,$C3,$E6,$FC,$1D
	.dc.b	$8E,$A4,$D7,$D2,$7D,$DB,$EC,$FD,$FB,$F4,$BA,$DB,$F0,$BB,$63,$2F
	.dc.b	$C5,$7C,$37,$DA,$3E,$DF,$C3,$5C,$BE,$07,$C2,$FD,$E3,$EF,$FC,$2D
	.dc.b	$8F,$89,$EF,$7E,$0B,$E0,$BF,$06,$D8,$FA,$1F,$96,$FF,$6D,$7F,$2C
	.dc.b	$DD,$7F,$B9,$3F,$DB,$CC,$6B,$F9,$AC,$7B,$4F,$A7,$95,$96,$35,$FD
	.dc.b	$16,$5D,$EB,$4D,$8F,$E0,$B8,$BE,$13,$F3,$2F,$F7,$9C,$FC,$B1,$8F
	.dc.b	$85,$CF,$C2,$6B,$E1,$2D,$7A,$9E,$3F,$09,$23,$DB,$F0,$97,$4F,$9D
	.dc.b	$93,$E1,$3B,$9E,$0F,$81,$B7,$F1,$32,$EC,$D3,$B1,$F2,$31,$E9,$AF
	.dc.b	$AB,$2B,$3F,$2B,$E1,$3D,$7E,$87,$B1,$B1,$F4,$3E,$0F,$DF,$F0,$9F
	.dc.b	$01,$B1,$F5,$3E,$13,$E1,$FC,$27,$C4,$6C,$7D,$8F,$84,$F8,$E7,$C8
	.dc.b	$F8,$1B,$1D,$4F,$BB,$87,$56,$9D,$4F,$C0,$DF,$56,$9F,$84,$F8,$4F
	.dc.b	$A6,$7D,$4F,$82,$B1,$D7,$93,$E5,$DF,$56,$C7,$57,$E7,$DF,$56,$C7
	.dc.b	$E4,$3E,$13,$F8,$BF,$E5,$DF,$DE,$A3,$07,$F0,$37,$13,$FA,$57,$FB
	.dc.b	$A4,$67,$E0,$F8,$FC,$27,$27,$9F,$BF,$27,$97,$BF,$70,$E5,$EF,$F8
	.dc.b	$0F,$0F,$C1,$5A,$F0,$63,$CD,$A7,$73,$C9,$87,$3E,$7F,$EB,$B9,$FF
	.dc.b	$74,$9F,$E7,$CF,$F7,$27,$D5,$F0,$7E,$E3,$DF,$F0,$96,$3D,$DF,$09
	.dc.b	$F0,$8F,$87,$F0,$96,$3E,$17,$C2,$7C,$6F,$84,$F8,$ED,$8F,$8D,$F0
	.dc.b	$9F,$2B,$E0,$BE,$5B,$63,$A9,$38,$7E,$56,$1D,$5F,$99,$7C,$DB,$3F
	.dc.b	$47,$0F,$AB,$F0,$5F,$5B,$E0,$AC,$FD,$57,$EC,$FC,$27,$DA,$F8,$4B
	.dc.b	$3F,$65,$FB,$BF,$09,$F7,$BE,$12,$CF,$DD,$7F,$AB,$FB,$9F,$D8,$BD
	.dc.b	$3F,$E7,$3F,$05,$8D,$72,$F4,$7F,$05,$8D,$E6,$FA,$2F,$48,$FC,$36
	.dc.b	$16,$7E,$AB,$DB,$E0,$BB,$FC,$15,$CF,$FA,$86,$7F,$DB,$6F,$ED,$EB
	.dc.b	$EB,$9F,$C9,$37,$1F,$C9,$6B,$EB,$9F,$CB,$36,$3F,$9D,$5F,$F6,$F9
	.dc.b	$EB,$F0,$7E,$EF,$84,$F7,$BF,$03,$E1,$24,$F8,$BE,$12,$E1,$CB,$E0
	.dc.b	$FE,$2F,$C0,$FC,$66,$D7,$A3,$E0,$7E,$49,$F2,$BE,$06,$C7,$B3,$1E
	.dc.b	$7B,$FF,$8C,$B5,$67,$8A,$FF,$0A,$43,$1E,$B9,$F8,$67,$E2,$E7,$F5
	.dc.b	$AA,$47,$C8,$F8,$4B,$5F,$03,$E1,$3E,$D9,$F7,$3E,$12,$D7,$CE,$F8
	.dc.b	$4F,$BE,$7E,$03,$E1,$2C,$7D,$2F,$83,$FC,$1F,$C1,$7E,$11,$B1,$F5
	.dc.b	$B0,$E7,$8F,$CE,$C6,$EF,$ED,$CB,$9C,$FD,$2C,$AC,$FC,$1F,$82,$FC
	.dc.b	$79,$F9,$0F,$84,$B9,$7F,$84,$7F,$2F,$FE,$BA,$4F,$87,$FF,$15,$B8
	.dc.b	$F8,$5A,$63,$E1,$6D,$71,$F8,$1D,$9C,$3E,$16,$C7,$53,$99,$D3,$0A
	.dc.b	$F8,$6F,$9E,$BF,$E0,$3C,$7F,$B1,$36,$3C,$7F,$05,$2B,$D1,$F0,$56
	.dc.b	$B8,$47,$47,$96,$17,$6F,$F3,$9E,$3F,$DB,$EB,$C0,$7C,$1D,$3C,$33
	.dc.b	$AF,$84,$F8,$3F,$09,$F0,$9B,$5F,$1F,$E0,$FE,$2F,$C2,$7C,$66,$C7
	.dc.b	$CB,$F8,$4F,$93,$EF,$FC,$A6,$C7,$CF,$F8,$4F,$9A,$7C,$EF,$80,$B1
	.dc.b	$D4,$FA,$98,$75,$69,$D4,$FB,$17,$D5,$A7,$DB,$F8,$4F,$B3,$F0,$9F
	.dc.b	$69,$B1,$F7,$FE,$13,$EE,$FC,$27,$DE,$6C,$7E,$0F,$E1,$3F,$02,$7F
	.dc.b	$9B,$3F,$30,$FE,$A7,$C3,$F2,$A3,$1F,$03,$9F,$85,$B7,$FE,$5B,$B4
	.dc.b	$75,$8F,$C7,$63,$79,$FE,$F3,$DA,$3C,$07,$1F,$88,$C4,$FF,$91,$6F
	.dc.b	$FA,$63,$7E,$0F,$C9,$F0,$DE,$57,$CD,$F0,$D3,$0F,$17,$C3,$5B,$E5
	.dc.b	$F0,$BE,$B3,$D9,$F0,$B6,$BD,$1F,$0B,$EF,$F8,$3F,$80,$D8,$F6,$63
	.dc.b	$CF,$8B,$FD,$7B,$AB,$3C,$57,$FA,$7E,$19,$75,$CF,$D5,$9D,$C3,$8F
	.dc.b	$C1,$FC,$DF,$83,$F9,$CD,$C7,$B9,$2F,$E1,$65,$D9,$7C,$0B,$3E,$F5
	.dc.b	$F4,$73,$F6,$B4,$FA,$9F,$09,$F6,$CF,$B9,$F0,$B6,$3E,$C7,$C2,$FD
	.dc.b	$F3,$F0,$1F,$0B,$63,$EE,$7C,$07,$E0,$FE,$13,$F0,$8D,$8F,$C0,$7C
	.dc.b	$0F,$E1,$CF,$C4,$7C,$25,$8E,$A4,$DB,$EE,$61,$D5,$FB,$D7,$CD,$B3
	.dc.b	$F8,$BC,$3F,$D7,$BF,$98,$FF,$67,$27,$C3,$62,$B3,$F0,$DA,$F8,$6B
	.dc.b	$87,$FA,$B6,$FF,$AF,$3F,$ED,$37,$FA,$B5,$7F,$A4,$57,$C3,$76,$F8
	.dc.b	$6E,$EF,$83,$E1,$3C,$3F,$09,$70,$E8,$F6,$C3,$96,$37,$AB,$FE,$00
	.dc.b	$E5,$E0,$9F,$AE,$F8,$23,$38,$F7,$C3,$CC,$FB,$BE,$12,$47,$A7,$E1
	.dc.b	$2E,$1E,$7F,$86,$F8,$67,$C4,$F8,$6B,$5E,$FF,$86,$F8,$E7,$C8,$F8
	.dc.b	$6B,$1F,$0F,$E1,$3E,$59,$F3,$3E,$12,$C7,$C7,$C3,$9E,$3F,$17,$1B
	.dc.b	$AF,$B3,$2E,$73,$F2,$72,$B3,$F0,$3E,$13,$EB,$9F,$63,$E1,$2E,$5D
	.dc.b	$49,$7D,$5B,$5D,$5F,$51,$F6,$6F,$AB,$F6,$7E,$17,$F0,$7F,$0D,$F8
	.dc.b	$46,$C7,$DF,$F8,$6F,$C3,$FC,$37,$E2,$1B,$1F,$EE,$FF,$23,$FD,$76
	.dc.b	$4F,$81,$46,$4B,$9F,$FB,$56,$7F,$A4,$37,$E1,$33,$A6,$BC,$85,$7F
	.dc.b	$58,$8C,$3C,$1F,$C1,$E3,$E2,$BF,$C8,$36,$37,$87,$41,$3F,$BE,$AF
	.dc.b	$34,$79,$4F,$57,$C3,$6E,$74,$DC,$3C,$7F,$0D,$6D,$37,$E1,$A6,$1F
	.dc.b	$0F,$E1,$AE,$3E,$FF,$83,$E0,$71,$F8,$4B,$5F,$27,$E1,$3A,$7C,$0F
	.dc.b	$56,$C7,$CD,$C7,$9E,$DF,$0B,$2B,$3E,$CD,$BE,$3E,$5D,$33,$F9,$D9
	.dc.b	$DC,$3E,$4F,$C2,$7A,$7E,$0B,$D4,$DC,$7E,$D7,$C2,$7B,$7E,$13,$DC
	.dc.b	$D8,$FB,$DF,$07,$F0,$4F,$85,$F0,$96,$3F,$05,$F0,$9F,$14,$F8,$DF
	.dc.b	$03,$63,$A9,$F8,$7C,$3A,$B4,$EA,$4D,$7F,$0D,$7D,$5F,$C4,$7C,$27
	.dc.b	$D1,$3E,$A7,$E6,$5F,$DA,$AB,$E1,$73,$A7,$AD,$7E,$3F,$0B,$BF,$E2
	.dc.b	$9E,$8F,$E3,$34,$F5,$E1,$E1,$8F,$C9,$63,$FE,$C6,$B4,$63,$E1,$7C
	.dc.b	$3F,$0B,$E2,$7C,$7F,$07,$30,$FC,$7D,$72,$F8,$4C,$61,$CA,$7E,$C6
	.dc.b	$2F,$82,$F5,$FC,$07,$B1,$B3,$C8,$DF,$BF,$F0,$0F,$83,$F0,$56,$35
	.dc.b	$87,$F5,$AC,$3F,$BE,$E5,$EA,$CF,$E3,$E7,$E2,$8F,$6E,$5D,$AF,$C8
	.dc.b	$71,$71,$1F,$0B,$E0,$FE,$79,$F4,$3E,$12,$E5,$E0,$EC,$66,$F8,$4F
	.dc.b	$AB,$CB,$F5,$9B,$5B,$97,$C1,$FD,$A3,$ED,$FC,$1D,$9D,$7C,$1F,$DE
	.dc.b	$F8,$1F,$BE,$D8,$E5,$87,$4D,$FD,$39,$F8,$A3,$1B,$7D,$AC,$BA,$56
	.dc.b	$B3,$67,$8A,$FB,$FF,$03,$F8,$CF,$81,$FC,$6B,$73,$F4,$7C,$1F,$E4
	.dc.b	$0F,$C8,$FC,$1D,$8F,$85,$F9,$9F,$FB,$44,$9F,$09,$FC,$25,$C7,$C2
	.dc.b	$5F,$81,$C3,$6B,$A9,$36,$F3,$61,$D5,$F6,$DF,$57,$A6,$1D,$5E,$DF
	.dc.b	$09,$DF,$E0,$6E,$1E,$47,$96,$DF,$9C,$D3,$D3,$3D,$E3,$F9,$86,$3D
	.dc.b	$63,$D7,$8F,$57,$DB,$F0,$9E,$B3,$D9,$F0,$96,$3E,$0F,$C1,$FB,$FE
	.dc.b	$0F,$E0,$36,$3E,$2F,$C0,$FC,$33,$E2,$7C,$0D,$8F,$93,$F0,$3F,$1F
	.dc.b	$DF,$F9,$0D,$8E,$A7,$CE,$C3,$AB,$4E,$A4,$E1,$F8,$F7,$D5,$F9,$3F
	.dc.b	$07,$F4,$CF,$A9,$F0,$76,$BE,$CF,$C1,$FD,$73,$EC,$7C,$1D,$8F,$BB
	.dc.b	$F0,$5F,$6C,$FB,$9F,$07,$63,$F0,$3F,$01,$F7,$CF,$C0,$7C,$1D,$8E
	.dc.b	$A4,$CF,$ED,$E1,$D5,$FB,$B8,$73,$7E,$FE,$17,$9B,$FB,$27,$E6,$9F
	.dc.b	$DB,$30,$FD,$83,$6B,$F4,$2B,$E1,$3F,$AC,$39,$AE,$3F,$09,$83,$3F
	.dc.b	$11,$70,$E7,$F0,$9C,$0E,$3F,$0B,$63,$BF,$C1,$F4,$F8,$3E,$AD,$8F
	.dc.b	$1E,$1C,$F4,$F1,$E3,$72,$F4,$E5,$CE,$7C,$F9,$59,$F6,$FC,$1F,$A0
	.dc.b	$F4,$FC,$1D,$CB,$A9,$2F,$AB,$6B,$AB,$EA,$3E,$2D,$F5,$7F,$B3,$6A
	.dc.b	$F8,$4F,$8A,$7C,$6F,$85,$B3,$F3,$3E,$13,$E4,$9F,$2B,$E1,$6C,$7D
	.dc.b	$0F,$85,$F9,$A7,$CE,$F8,$1B,$1F,$53,$E0,$7E,$8F,$C1,$FD,$26,$C7
	.dc.b	$52,$6D,$F6,$30,$EA,$FD,$AC,$39,$BF,$73,$0B,$B7,$E0,$BE,$17,$EE
	.dc.b	$FC,$2F,$DE,$6E,$3F,$84,$F8,$5F,$C0,$9F,$E1,$9D,$7F,$B3,$63,$FE
	.dc.b	$9E,$E0,$FE,$CD,$5F,$CE,$67,$FA,$F6,$3F,$E3,$17,$FD,$9F,$19,$F8
	.dc.b	$4E,$5F,$0B,$CF,$E0,$BA,$7C,$15,$AE,$4F,$81,$F1,$1D,$7E,$23,$19
	.dc.b	$FF,$36,$C7,$BC,$FF,$86,$5E,$F1,$9C,$7B,$CF,$97,$E0,$FD,$47,$AF
	.dc.b	$E1,$AC,$7A,$7E,$1B,$DC,$7B,$FE,$1A,$C7,$B7,$E1,$BE,$11,$F0,$FE
	.dc.b	$1A,$C7,$C1,$F8,$2F,$8C,$7C,$7F,$86,$B1,$D4,$9A,$7A,$70,$EA,$FA
	.dc.b	$F0,$E6,$FB,$70,$BC,$5F,$2F,$E0,$FE,$97,$C1,$7D,$36,$E3,$D2,$7E
	.dc.b	$86,$9E,$98,$C1,$F1,$71,$E6,$FC,$7C,$2E,$3F,$ED,$8E,$07,$C2,$FD
	.dc.b	$FF,$86,$FC,$03,$6F,$EF,$7C,$1F,$E0,$FE,$13,$F0,$8D,$8F,$C1,$7B
	.dc.b	$FF,$87,$3F,$11,$F0,$56,$3F,$BF,$7E,$6B,$FE,$6A,$6F,$C0,$7F,$99
	.dc.b	$63,$1A,$7F,$72,$51,$DA,$7F,$1D,$8D,$AF,$C1,$3D,$5F,$C2,$61,$E0
	.dc.b	$9F,$19,$E0,$F8,$4E,$C7,$7F,$85,$B1,$E4,$F8,$5F,$17,$C2,$F8,$DB
	.dc.b	$1E,$8F,$84,$F3,$1E,$7F,$84,$B1,$EC,$F8,$4F,$51,$EB,$F8,$1B,$1D
	.dc.b	$4F,$83,$87,$56,$9D,$49,$A7,$C1,$BE,$AF,$C3,$F8,$4F,$8C,$7C,$7F
	.dc.b	$82,$B5,$F3,$3E,$0F,$E5,$1F,$2F,$E1,$2C,$7D,$0F,$84,$F9,$C7,$CF
	.dc.b	$F8,$4B,$1F,$53,$E1,$3E,$97,$C1,$7D,$36,$C7,$52,$67,$F5,$70,$EA
	.dc.b	$FD,$7C,$39,$BF,$67,$0B,$E0,$7F,$AC,$FA,$1F,$05,$F7,$CF,$C0,$7C
	.dc.b	$15,$CB,$AD,$6F,$5F,$54,$60,$FF,$4B,$5B,$2E,$B1,$8C,$6E,$4F,$87
	.dc.b	$5B,$E1,$7F,$1C,$7E,$3F,$E1,$6E,$1F,$E5,$1F,$CD,$BF,$AF,$3F,$EA
	.dc.b	$37,$FB,$13,$93,$FD,$74,$E3,$0F,$F2,$3B,$FE,$CB,$8E,$1F,$07,$AF
	.dc.b	$83,$DB,$C3,$E0,$E4,$F7,$F8,$1B,$87,$28,$EB,$F0,$DD,$9A,$79,$6F
	.dc.b	$9B,$E7,$C2,$D7,$8B,$E1,$7C,$BF,$0D,$E6,$6D,$7B,$7E,$1B,$D2,$7A
	.dc.b	$BE,$1A,$C7,$C1,$F8,$6F,$6F,$C3,$7B,$9B,$1F,$17,$CE,$F8,$3F,$09
	.dc.b	$F0,$9B,$1D,$49,$9F,$C4,$C3,$AB,$F1,$AF,$9B,$67,$E8,$61,$F3,$3E
	.dc.b	$12,$47,$D3,$F8,$4B,$5F,$3B,$E1,$7E,$8F,$C3,$7D,$26,$D7,$D9,$F8
	.dc.b	$5F,$AA,$7D,$6F,$86,$B1,$F7,$7E,$1B,$FD,$E5,$F9,$BF,$F5,$AE,$D7
	.dc.b	$57,$F5,$AC,$FF,$D0,$33,$98,$65,$DA,$7C,$58,$7E,$1F,$3E,$B1,$F8
	.dc.b	$2C,$FC,$13,$E3,$3F,$D3,$98,$7F,$56,$9F,$CA,$37,$13,$FD,$63,$C1
	.dc.b	$57,$FB,$CE,$35,$97,$8A,$3C,$07,$E5,$F2,$F2,$57,$0C,$3F,$41,$4C
	.dc.b	$7C,$B3,$9C,$7D,$FA,$63,$1E,$DF,$0B,$F1,$3E,$17,$E2,$B6,$3D,$3F
	.dc.b	$07,$F2,$0F,$93,$F0,$76,$3D,$BE,$F7,$CC,$3E,$6F,$C0,$D8,$F8,$3E
	.dc.b	$FF,$D0,$F7,$FE,$8B,$63,$A9,$38,$3E,$8E,$1D,$5F,$A7,$87,$37,$EA
	.dc.b	$E1,$76,$F6,$71,$7F,$B1,$F0,$F9,$FF,$07,$F7,$BE,$0F,$EF,$BF,$80
	.dc.b	$F8,$2F,$C0,$FC,$1D,$AF,$AA,$FE,$13,$E0,$BF,$0B,$F0,$76,$7E,$CB
	.dc.b	$F8,$8F,$83,$FC,$4F,$C0,$D9,$FB,$AF,$3D,$BF,$13,$85,$D7,$CA,$F3
	.dc.b	$8F,$C6,$E1,$67,$D2,$FE,$53,$E0,$7F,$2B,$F0,$77,$1F,$C4,$BF,$76
	.dc.b	$BE,$D7,$AD,$FD,$99,$B8,$66,$8F,$F2,$19,$FE,$D3,$1F,$EE,$D7,$83
	.dc.b	$AE,$A4,$3E,$0B,$95,$71,$F8,$2B,$5A,$9B,$49,$BF,$05,$8F,$7E,$E3
	.dc.b	$DD,$DE,$32,$3C,$5F,$07,$63,$FD,$DF,$87,$86,$7C,$D8,$76,$C7,$BB
	.dc.b	$E0,$F8,$3F,$0F,$C0,$DA,$F6,$3C,$F5,$F8,$71,$D8,$F8,$18,$F3,$7A
	.dc.b	$9F,$0F,$0E,$AF,$C5,$F8,$3F,$6F,$C1,$FB,$9B,$1F,$27,$E0,$FE,$09
	.dc.b	$F0,$BE,$0E,$C7,$CD,$F7,$FE,$29,$F1,$B6,$B1,$F4,$7D,$FF,$93,$EF
	.dc.b	$FC,$A6,$C7,$52,$69,$F1,$30,$EA,$FC,$6C,$39,$BF,$23,$0B,$AF,$E1
	.dc.b	$3E,$0B,$EA,$9F,$5B,$E0,$EE,$3F,$81,$F8,$3F,$B2,$7D,$AF,$83,$B1
	.dc.b	$F8,$5F,$82,$FB,$A7,$DE,$F8,$2B,$1F,$8C,$C7,$F8,$57,$E7,$1F,$D9
	.dc.b	$B4,$FE,$DC,$DE,$B3,$FD,$96,$77,$5D,$AB,$EF,$E5,$E1,$8F,$21,$E1
	.dc.b	$24,$7E,$13,$0F,$0A,$AF,$0F,$C0,$7E,$53,$07,$C0,$58,$C3,$9F,$83
	.dc.b	$98,$FA,$4F,$D0,$53,$E0,$F3,$8F,$59,$CF,$C0,$F5,$3B,$7C,$15,$8E
	.dc.b	$53,$F0,$3E,$0B,$C4,$78,$FE,$02,$CE,$B8,$3E,$1E,$7D,$B6,$F9,$39
	.dc.b	$DA,$EE,$6F,$46,$9E,$08,$C3,$EF,$F8,$0F,$81,$F0,$16,$78,$3C,$F5
	.dc.b	$FA,$79,$7C,$5C,$71,$9F,$D4,$C7,$D5,$5E,$F3,$EC,$ED,$8D,$3E,$EF
	.dc.b	$C0,$FC,$DF,$82,$F9,$CD,$D3,$CD,$F0,$5F,$44,$FA,$5F,$05,$63,$D5
	.dc.b	$F0,$1F,$57,$E0,$7E,$B3,$63,$DD,$F0,$1F,$64,$FB,$5F,$01,$63,$A9
	.dc.b	$35,$F3,$6F,$F8,$CC,$7B,$46,$39,$3F,$B2,$37,$2E,$F7,$F2,$DF,$C3
	.dc.b	$7C,$17,$F1,$AF,$CE,$7F,$C0,$17,$FD,$FF,$83,$FB,$24,$7F,$80,$CF
	.dc.b	$EE,$84,$F8,$8E,$1F,$0F,$C7,$E0,$F9,$7C,$25,$CF,$EF,$5A,$3B,$7C
	.dc.b	$24,$9C,$FC,$15,$DB,$D7,$7D,$B1,$F2,$65,$DA,$F9,$E5,$D9,$A7,$6C
	.dc.b	$BD,$5F,$05,$EB,$F8,$2B,$3E,$27,$DD,$F0,$7E,$FF,$84,$B3,$E6,$7E
	.dc.b	$17,$C1,$48,$F4,$FC,$1D,$DB,$CB,$F0,$7F,$1F,$E0,$3E,$43,$6B,$DF
	.dc.b	$8F,$56,$9D,$4F,$87,$87,$3E,$3F,$9B,$8D,$9F,$EA,$1C,$3F,$EC,$E7
	.dc.b	$E4,$FC,$07,$D5,$F8,$3F,$AC,$FA,$34,$F8,$FA,$DC,$7E,$BB,$CE,$3F
	.dc.b	$CF,$D8,$7F,$80,$5F,$9B,$8E,$7A,$92,$F8,$3F,$C1,$7C,$1F,$E0,$DB
	.dc.b	$A7,$DA,$F8,$3F,$C3,$1F,$87,$F8,$3B,$1F,$BB,$FC,$EB,$FB,$4E,$7F
	.dc.b	$E8,$D7,$FB,$6B,$F9,$67,$39,$1C,$EB,$FE,$31,$7F,$AF,$4E,$FE,$07
	.dc.b	$FB,$6C,$62,$75,$1E,$09,$FC,$3E,$3E,$3C,$3C,$C7,$8E,$38,$FC,$17
	.dc.b	$26,$C7,$9F,$E1,$3A,$9D,$BE,$12,$C7,$AF,$E1,$3C,$27,$8B,$E1,$2C
	.dc.b	$7B,$FE,$0B,$CB,$F0,$7E,$66,$C7,$C3,$F8,$3F,$4F,$C0,$FA,$9B,$1D
	.dc.b	$49,$9F,$C5,$C3,$AB,$F1,$EF,$9B,$67,$E7,$61,$F1,$3E,$06,$47,$D1
	.dc.b	$F8,$1B,$5F,$3F,$E0,$BE,$4F,$C1,$FC,$A6,$D7,$D7,$F8,$3F,$9B,$F0
	.dc.b	$7F,$39,$B1,$F6,$FE,$07,$E8,$FC,$1F,$D2,$6C,$7D,$FB,$E7,$8F,$78
	.dc.b	$FC,$16,$3C,$DF,$C2,$63,$67,$F0,$4F,$DC,$F8,$3F,$BB,$F0,$76,$BF
	.dc.b	$14,$F9,$DE,$67,$72,$D7,$E2,$F3,$F3,$B6,$3B,$9F,$D3,$7C,$6F,$EE
	.dc.b	$07,$F6,$F7,$FC,$78,$E3,$E1,$FF,$C7,$2F,$FB,$56,$7F,$DC,$8F,$F6
	.dc.b	$89,$E5,$F0,$9A,$F8,$4D,$FC,$07,$0F,$80,$B5,$DD,$E7,$F0,$1D,$3E
	.dc.b	$02,$CF,$8D,$EF,$87,$38,$EA,$79,$F0,$E7,$97,$F3,$CC,$BF,$B9,$4F
	.dc.b	$AF,$2F,$03,$ED,$F8,$4F,$49,$EA,$F8,$4B,$1F,$07,$E1,$3D,$BF,$09
	.dc.b	$EE,$6C,$7C,$5F,$83,$F8,$3F,$09,$F0,$9B,$1F,$27,$E1,$3E,$29,$F1
	.dc.b	$BE,$06,$C7,$53,$E7,$61,$D5,$A7,$53,$E9,$5F,$56,$9F,$57,$E1,$3E
	.dc.b	$89,$F4,$BE,$0A,$C7,$D9,$F8,$4F,$AA,$7D,$6F,$84,$B1,$F7,$7E,$13
	.dc.b	$FC,$E3,$F9,$DF,$FB,$5F,$8B,$FA,$C4,$FF,$9E,$63,$FD,$B4,$7F,$B2
	.dc.b	$CF,$F6,$53,$9F,$82,$E3,$F0,$5D,$39,$7E,$86,$5D,$EB,$C4,$77,$3F
	.dc.b	$C3,$17,$DF,$1F,$0F,$C5,$78,$BE,$0F,$C7,$F0,$76,$F6,$F5,$F5,$3F
	.dc.b	$D9,$DA,$FF,$B4,$67,$F2,$DB,$65,$FC,$AE,$3D,$1F,$06,$9F,$DF,$B3
	.dc.b	$F5,$FC,$2F,$C0,$F8,$5F,$82,$DA,$F2,$FC,$2F,$C4,$3E,$2F,$C2,$D8
	.dc.b	$F4,$FC,$27,$C8,$F8,$1F,$92,$D8,$F6,$FC,$27,$CC,$F8,$2F,$9A,$D8
	.dc.b	$EA,$4D,$3C,$98,$75,$7C,$D7,$CD,$B3,$F2,$30,$FA,$FB,$C8,$F9,$7F
	.dc.b	$0B,$6B,$DD,$F0,$BF,$73,$E1,$7E,$EB,$6B,$E8,$FC,$1F,$E0,$3E,$13
	.dc.b	$F0,$2D,$8F,$AB,$F0,$1F,$84,$3F,$0B,$F0,$96,$3E,$CD,$F3,$C7,$BC
	.dc.b	$7D,$CC,$79,$BF,$7B,$1B,$3F,$45,$FF,$5F,$FE,$79,$FD,$68,$FE,$B3
	.dc.b	$3F,$EC,$77,$1F,$05,$AF,$88,$93,$FE,$5F,$9F,$F6,$23,$63,$9F,$C1
	.dc.b	$C8,$D7,$C1,$DC,$7F,$C7,$B6,$F8,$3F,$0F,$C0,$F8,$9B,$7C,$C9,$5C
	.dc.b	$3E,$13,$C2,$F2,$C7,$A7,$7A,$BD,$19,$61,$F5,$7C,$17,$BB,$E0,$BD
	.dc.b	$ED,$8E,$B5,$C7,$07,$C3,$35,$FE,$ED,$5F,$0F,$3F,$0D,$EE,$B5,$9F
	.dc.b	$7C,$38,$E0,$CF,$15,$CF,$E1,$3E,$79,$F4,$3E,$16,$E7,$F2,$FE,$17
	.dc.b	$E9,$9F,$53,$E1,$6C,$7C,$FF,$83,$FA,$E7,$D8,$F8,$3B,$1F,$4F,$E0
	.dc.b	$3E,$DF,$C1,$FD,$C6,$C7,$52,$6D,$F4,$B0,$EA,$FD,$4C,$39,$BF,$5B
	.dc.b	$0B,$C1,$EF,$DD,$F0,$9F,$88,$F8,$5F,$C4,$B7,$2F,$C2,$7C,$2F,$E3
	.dc.b	$0F,$C6,$FC,$2D,$8F,$F5,$37,$E7,$BF,$EC,$93,$FA,$D6,$7F,$41,$C7
	.dc.b	$C3,$7D,$3C,$FE,$74,$7F,$5B,$E4,$3F,$D6,$EB,$FD,$0E,$E4,$C4,$F9
	.dc.b	$F1,$F0,$4F,$E4,$31,$B3,$D3,$1E,$C2,$7C,$58,$E1,$F2,$45,$C7,$C3
	.dc.b	$37,$17,$C0,$D3,$1F,$0D,$6F,$DF,$F0,$DB,$38,$7C,$35,$8F,$87,$F0
	.dc.b	$9C,$FE,$13,$A3,$63,$E3,$FC,$27,$7F,$81,$F0,$36,$3A,$93,$6F,$87
	.dc.b	$87,$57,$E2,$DF,$36,$CF,$D2,$C3,$D5,$F0,$9E,$BF,$82,$B3,$F5,$9E
	.dc.b	$5C,$3F,$4B,$4F,$4C,$F7,$8F,$AB,$8F,$56,$C7,$53,$EF,$FC,$27,$C7
	.dc.b	$3E,$47,$C2,$58,$FC,$1F,$C2,$7C,$B3,$E6,$7C,$25,$8F,$C3,$FC,$17
	.dc.b	$CF,$3E,$87,$BF,$63,$F1,$7E,$FF,$D5,$D5,$F0,$5F,$56,$30,$F3,$E1
	.dc.b	$FC,$3E,$5D,$AF,$C2,$76,$24,$7E,$33,$0E,$CF,$FB,$53,$07,$C2,$77
	.dc.b	$3C,$1F,$09,$6F,$F1,$5B,$FF,$05,$A7,$C2,$79,$4F,$37,$C2,$5A,$D9
	.dc.b	$F9,$1E,$D5,$3E,$0B,$D8,$7B,$7E,$0E,$D7,$0D,$3F,$8B,$D3,$E0,$FE
	.dc.b	$17,$C0,$FC,$36,$D6,$F1,$EB,$9F,$87,$0F,$66,$77,$4F,$46,$9D,$A7
	.dc.b	$E0,$69,$67,$D9,$F0,$3F,$3F,$E0,$7E,$83,$73,$ED,$B7,$FA,$D6,$BE
	.dc.b	$17,$C1,$FD,$5F,$83,$FA,$CF,$D7,$F8,$3F,$B1,$F0,$77,$0E,$2F,$DB
	.dc.b	$F8,$09,$1D,$3E,$02,$ED,$F4,$F1,$E7,$3E,$06,$C7,$D7,$C3,$AC,$F9
	.dc.b	$1B,$1F,$6F,$E0,$FF,$10,$7E,$27,$E0,$6D,$73,$3D,$79,$FA,$5A,$73
	.dc.b	$3C,$E5,$AF,$C3,$67,$F9,$3F,$83,$FC,$A7,$C1,$DA,$F8,$AF,$F0,$BF
	.dc.b	$CF,$BF,$D9,$24,$F8,$4C,$61,$9F,$84,$97,$F6,$6B,$FB,$7C,$B8,$7F
	.dc.b	$A9,$47,$F5,$68,$FE,$B2,$BD,$FE,$C7,$7F,$7E,$DF,$3A,$EE,$F5,$C6
	.dc.b	$C7,$58,$FC,$A5,$F5,$EB,$34,$F1,$E3,$E2,$CB,$D9,$F0,$72,$FD,$1F
	.dc.b	$09,$71,$E8,$79,$BE,$0F,$E1,$7C,$27,$C3,$6D,$7B,$FE,$0F,$E3,$7C
	.dc.b	$1F,$C7,$6C,$7C,$3F,$83,$F9,$5F,$03,$F2,$DB,$1F,$1E,$F9,$E3,$E0
	.dc.b	$DF,$D3,$18,$DB,$E4,$E3,$D2,$B5,$BA,$31,$3F,$3B,$E0,$7E,$C1,$F6
	.dc.b	$7E,$06,$E5,$F5,$7E,$0B,$EE,$1F,$77,$E0,$EC,$7D,$9F,$81,$FC,$01
	.dc.b	$F8,$1F,$82,$B1,$F7,$7E,$0B,$F0,$9F,$05,$F8,$56,$C7,$52,$6D,$E6
	.dc.b	$C3,$AB,$E8,$C3,$9B,$EA,$C2,$E9,$FE,$17,$FC,$FF,$FC,$4D,$87,$F0
	.dc.b	$37,$FB,$8F,$09,$CD,$FC,$77,$3C,$0D,$7B,$E1,$F9,$0C,$EE,$3F,$86
	.dc.b	$CF,$BB,$F8,$8C,$EE,$5F,$E2,$CC,$BF,$BD,$1F,$DD,$A7,$F9,$03,$8F
	.dc.b	$87,$FE,$DC,$FF,$6C,$8F,$37,$C1,$63,$E0,$B2,$EB,$DD,$93,$EB,$F8
	.dc.b	$0B,$8F,$A7,$DE,$E6,$74,$CA,$D7,$42,$4F,$BF,$1E,$B1,$F1,$70,$E7
	.dc.b	$B7,$C9,$C6,$CF,$A3,$E0,$3C,$E7,$A3,$E0,$6C,$7C,$DF,$82,$F5,$9E
	.dc.b	$CF,$82,$B1,$F4,$7D,$0F,$79,$F0,$3E,$0A,$C7,$D5,$F8,$2F,$86,$7C
	.dc.b	$4F,$7A,$C7,$52,$6F,$F5,$F0,$EA,$FD,$9C,$39,$BF,$6F,$0B,$AF,$E0
	.dc.b	$FE,$0B,$E7,$9F,$43,$E0,$AE,$3F,$85,$F8,$0F,$A6,$7D,$4F,$82,$B1
	.dc.b	$FD,$2F,$F4,$0F,$ED,$ED,$F8,$3F,$F1,$34,$7F,$66,$8C,$1B,$F8,$2C
	.dc.b	$FC,$06,$9B,$1D,$C9,$AF,$E4,$70,$EE,$FE,$E6,$FB,$B7,$0E,$FF,$07
	.dc.b	$E0,$7C,$3F,$03,$30,$F3,$65,$6F,$F3,$99,$F4,$DB,$F4,$99,$DB,$FE
	.dc.b	$E5,$9F,$69,$FF,$3F,$36,$33,$9F,$77,$E1,$7C,$0F,$C0,$3E,$0F,$C0
	.dc.b	$D8,$F8,$DF,$05,$F1,$3E,$0F,$E2,$B6,$3E,$57,$C1,$7C,$83,$E4,$FC
	.dc.b	$1D,$8F,$9D,$F0,$3F,$33,$DE,$F9,$AD,$8E,$A7,$D3,$C3,$AB,$4E,$A4
	.dc.b	$E0,$F9,$57,$D5,$F9,$9F,$03,$F6,$0F,$B3,$EF,$DA,$FB,$DF,$07,$F7
	.dc.b	$3E,$0F,$EE,$B6,$3F,$05,$F0,$7F,$80,$F8,$3F,$C0,$B6,$3F,$0D,$F0
	.dc.b	$7F,$84,$F8,$3F,$C2,$B6,$3A,$9F,$8B,$C3,$AB,$4E,$A7,$E3,$EF,$AB
	.dc.b	$4F,$F0,$E7,$AF,$FE,$79,$6F,$C0,$FF,$9D,$E3,$1A,$FF,$41,$7F,$DB
	.dc.b	$AF,$FA,$F9,$FE,$B8,$4F,$87,$E1,$F0,$5C,$7E,$12,$61,$DB,$E1,$2E
	.dc.b	$1C,$F1,$ED,$F0,$9D,$DB,$5E,$4C,$39,$BE,$6C,$6D,$F8,$F1,$E7,$DB
	.dc.b	$FE,$CB,$CF,$FD,$91,$7E,$C9,$E7,$5E,$9F,$84,$B7,$B7,$9B,$C7,$1F
	.dc.b	$81,$8E,$30,$FF,$26,$63,$FE,$03,$7A,$1E,$B2,$E1,$F1,$7E,$03,$E4
	.dc.b	$7C,$27,$C9,$6C,$FC,$EF,$81,$F9,$9F,$09,$F3,$5B,$1F,$4B,$E0,$FE
	.dc.b	$81,$F4,$7E,$12,$C7,$D6,$CB,$EA,$7C,$27,$D5,$6C,$75,$3E,$DE,$1D
	.dc.b	$5A,$75,$26,$7F,$6E,$FA,$BF,$77,$E0,$3F,$01,$F0,$1F,$81,$6D,$7E
	.dc.b	$1B,$E0,$3F,$09,$F0,$1F,$85,$6C,$7E,$2B,$E0,$FF,$11,$F0,$9F,$89
	.dc.b	$6C,$7F,$9A,$7F,$41,$FE,$CB,$CB,$FF,$00,$5F,$ED,$26,$23,$FC,$F4
	.dc.b	$FF,$6C,$DB,$FD,$CF,$1F,$96,$C7,$C0,$D3,$C0,$7E,$6B,$0F,$03,$4F
	.dc.b	$07,$C5,$7F,$B4,$6F,$FD,$C4,$7F,$69,$D5,$F0,$C8,$CF,$C2,$6C,$E1
	.dc.b	$F0,$57,$0F,$57,$C1,$73,$3A,$7C,$15,$8F,$77,$C2,$77,$F8,$3F,$03
	.dc.b	$63,$E1,$61,$CF,$3F,$F1,$66,$4C,$B1,$C1,$F1,$73,$E9,$5A,$CE,$CF
	.dc.b	$C9,$F8,$3F,$67,$C1,$FB,$5B,$9F,$62,$5F,$6C,$7A,$C7,$CE,$C7,$9B
	.dc.b	$F4,$31,$B5,$F1,$32,$7C,$0F,$C9,$3E,$57,$C0,$DB,$FB,$9F,$03,$F3
	.dc.b	$4F,$9D,$F0,$76,$3F,$01,$F0,$7F,$47,$E0,$FE,$93,$63,$F0,$9F,$07
	.dc.b	$F5,$7D,$FF,$AC,$D8,$EA,$4D,$3E,$3E,$1D,$5F,$93,$7C,$DB,$3F,$8F
	.dc.b	$C3,$F0,$1E,$FC,$8F,$F5,$D7,$E8,$5F,$D4,$EB,$DD,$29,$CF,$BF,$67
	.dc.b	$FC,$C1,$4F,$84,$C7,$36,$7A,$93,$FD,$DE,$B7,$5A,$C3,$FD,$A9,$3F
	.dc.b	$ED,$A7,$C3,$F0,$59,$AF,$C8,$CB,$E0,$F1,$87,$4A,$B3,$FE,$DB,$BF
	.dc.b	$4F,$C0,$E3,$E0,$B2,$EB,$4D,$E7,$6B,$DE,$F2,$C2,$47,$C2,$CA,$E7
	.dc.b	$EE,$F8,$2E,$FE,$FF,$81,$B5,$F2,$3E,$0B,$C6,$79,$3E,$02,$C7,$CC
	.dc.b	$F8,$1F,$39,$E8,$F8,$2B,$1F,$43,$E0,$BD,$67,$4C,$BE,$6B,$C9,$F9
	.dc.b	$F8,$75,$9F,$01,$D4,$91,$F5,$2B,$AB,$F5,$BE,$0B,$E3,$9D,$A7,$EF
	.dc.b	$BD,$79,$3E,$C7,$1B,$0E,$D7,$E0,$BE,$B5,$AC,$BD,$51,$83,$ED,$FC
	.dc.b	$17,$D4,$F7,$FE,$AB,$63,$F1,$BF,$05,$F6,$3E,$03,$EC,$B6,$3F,$23
	.dc.b	$F0,$3F,$70,$FB,$BF,$05,$63,$F2,$B8,$7E,$0B,$F4,$3F,$EE,$39,$FA
	.dc.b	$23,$1A,$74,$AF,$CD,$E5,$63,$F2,$EF,$37,$F3,$38,$F6,$9F,$08,$8F
	.dc.b	$F5,$E6,$1F,$DF,$3A,$CF,$FB,$5C,$C1,$7D,$DF,$27,$C0,$79,$5B,$7C
	.dc.b	$F3,$FD,$25,$7F,$72,$C3,$AF,$C1,$7A,$FE,$0F,$D8,$FB,$7E,$0A,$61
	.dc.b	$C4,$F2,$67,$E2,$F8,$2C,$E1,$E6,$CF,$BB,$6B,$83,$FC,$EB,$2F,$EE
	.dc.b	$CF,$B3,$2E,$F3,$AC,$3D,$B1,$89,$F8,$3F,$03,$F3,$CF,$A1,$F0,$56
	.dc.b	$F8,$61,$FE,$DB,$6F,$C1,$7D,$5F,$80,$FA,$CF,$D7,$F8,$2F,$B1,$F0
	.dc.b	$36,$B8,$BF,$6F,$DD,$FB,$9F,$05,$67,$AB,$C9,$EA,$78,$30,$C6,$9F
	.dc.b	$5B,$0E,$B3,$E4,$6C,$7D,$AA,$FC,$47,$C0,$F6,$AF,$4B,$CB,$93,$EA
	.dc.b	$E7,$E9,$9E,$F7,$D5,$F3,$9D,$4A,$FE,$1B,$E0,$BF,$28,$7E,$57,$E0
	.dc.b	$AC,$FC,$6F,$81,$FC,$C1,$F9,$9F,$82,$B1,$F2,$BE,$0B,$F8,$EF,$89
	.dc.b	$FD,$FF,$1F,$B1,$18,$9C,$FC,$54,$7E,$8E,$56,$BF,$2F,$8F,$47,$F3
	.dc.b	$38,$DD,$3F,$23,$8F,$3C,$3C,$07,$E0,$72,$F0,$73,$97,$C1,$F9,$0F
	.dc.b	$2F,$C3,$DA,$E1,$F0,$5E,$8F,$80,$F4,$B6,$3A,$7C,$07,$B3,$E0,$3D
	.dc.b	$AD,$8F,$07,$C1,$FC,$03,$A6,$ED,$BF,$A5,$68,$9E,$B5,$E7,$CA,$DE
	.dc.b	$E3,$F4,$B5,$FE,$07,$7C,$27,$B7,$1F,$0B,$4F,$81,$F0,$9F,$43,$E0
	.dc.b	$FE,$8B,$63,$E2,$7C,$27,$D4,$3E,$AF,$C1,$D8,$F9,$1F,$07,$F6,$3E
	.dc.b	$0F,$EC,$B6,$3E,$66,$1F,$73,$E0,$7A,$69,$E7,$3F,$C5,$BA,$23,$A4
	.dc.b	$67,$93,$E9,$67,$E0,$C3,$C6,$78,$0F,$B1,$97,$81,$A7,$DB,$BF,$03
	.dc.b	$E2,$30,$72,$BE,$D8,$2F,$B4,$61,$E8,$EA,$EF,$07,$D4,$7F,$2F,$F0
	.dc.b	$33,$0F,$C5,$7C,$0D,$BE,$BF,$05,$F9,$CF,$83,$FC,$EB,$6B,$F2,$1F
	.dc.b	$07,$FA,$03,$FB,$EF,$E8,$9F,$E1,$1D,$FF,$25,$18,$F8,$29,$5F,$97
	.dc.b	$CE,$E5,$DA,$3F,$35,$97,$66,$9D,$89,$1F,$37,$1E,$DB,$B1,$F0,$FC
	.dc.b	$14,$BF,$E4,$95,$FE,$15,$BE,$13,$E6,$F8,$1F,$3F,$C2,$5B,$DB,$EA
	.dc.b	$F8,$49,$1C,$BE,$0A,$E1,$E1,$F8,$3F,$79,$DD,$C4,$F8,$30,$EA,$D3
	.dc.b	$A9,$E4,$BE,$7C,$9F,$E2,$FC,$D8,$E2,$BF,$A3,$C7,$F7,$B7,$D7,$EF
	.dc.b	$FC,$EF,$83,$F9,$EF,$D0,$F8,$0F,$A3,$F0,$56,$BE,$0B,$F5,$3E,$0F
	.dc.b	$EA,$F1,$59,$F8,$AF,$D8,$F8,$29,$1D,$8F,$92,$F3,$DF,$E8,$5D,$D7
	.dc.b	$D0,$F3,$8F,$A9,$76,$7D,$8F,$48,$FB,$1F,$01,$67,$E0,$3C,$F0,$F7
	.dc.b	$DD,$AF,$F4,$CE,$A8,$E9,$3F,$13,$0B,$5A,$7F,$23,$F0,$52,$7F,$03
	.dc.b	$F0,$36,$BF,$11,$F0,$3F,$96,$3F,$2F,$F0,$56,$BF,$B5,$7E,$8B,$FE
	.dc.b	$14,$7F,$BB,$B6,$BF,$D5,$6F,$F8,$2C,$C7,$69,$7C,$07,$F9,$79,$B5
	.dc.b	$F9,$2C,$B1,$3D,$EF,$C7,$3F,$96,$C3,$1B,$FF,$9C,$70,$EA,$4F,$87
	.dc.b	$E9,$7D,$7E,$06,$61,$E8,$F8,$4B,$3F,$C4,$6F,$FC,$2D,$87,$3F,$82
	.dc.b	$F2,$7C,$27,$95,$F3,$7C,$27,$9F,$E1,$2E,$5F,$09,$F5,$7C,$27,$AF
	.dc.b	$DD,$B3,$F1,$9F,$77,$C0,$FB,$FE,$06,$CF,$CA,$7E,$17,$C1,$FC,$3F
	.dc.b	$83,$B3,$F3,$9F,$8D,$F0,$72,$3E,$8F,$C1,$DE,$1F,$57,$C0,$FC,$BF
	.dc.b	$83,$F9,$8D,$AF,$AF,$F0,$5F,$3C,$FA,$1F,$05,$63,$ED,$FC,$17,$D3
	.dc.b	$3A,$72,$FC,$D7,$9C,$7E,$03,$0B,$8F,$C9,$79,$3F,$2F,$0E,$B3,$E0
	.dc.b	$3F,$C6,$1F,$A3,$7F,$70,$27,$C2,$7F,$A7,$9C,$7C,$25,$31,$F0,$96
	.dc.b	$B8,$FC,$26,$FE,$0B,$83,$63,$AF,$C2,$73,$F8,$3E,$8D,$8F,$F2,$2E
	.dc.b	$8F,$DF,$1E,$F7,$FA,$C6,$30,$D8,$7C,$15,$8F,$37,$C2,$CB,$F5,$FC
	.dc.b	$2D,$C3,$FA,$D6,$EF,$84,$F6,$7C,$1F,$B5,$B7,$E2,$CD,$EF,$F7,$3C
	.dc.b	$1E,$FD,$9F,$8F,$F0,$BF,$17,$6F,$8C,$D8,$F9,$7B,$7C,$9F,$85,$F9
	.dc.b	$4D,$8F,$7D,$FF,$B2,$DB,$EF,$7B,$E3,$FD,$95,$B7,$F7,$53,$FA,$24
	.dc.b	$7D,$3F,$82,$95,$F6,$3E,$16,$D7,$BB,$E1,$7E,$C6,$BF,$65,$B5,$E0
	.dc.b	$F7,$FB,$1D,$FD,$EB,$1F,$82,$CF,$FD,$BB,$A7,$F6,$1E,$3F,$EA,$93
	.dc.b	$FE,$75,$8F,$EC,$87,$FB,$AF,$C0,$FF,$77,$3F,$EF,$18,$D7,$C2,$F2
	.dc.b	$F8,$5F,$67,$BD,$27,$E0,$FB,$17,$7F,$F0,$76,$1F,$D7,$71,$E7,$C5
	.dc.b	$E2,$F8,$8F,$1B,$E4,$F8,$69,$8F,$8F,$E1,$6E,$3F,$1F,$F7,$D7,$B5
	.dc.b	$F8,$96,$D7,$E3,$9F,$E3,$73,$B9,$F7,$FC,$37,$C0,$F8,$7B,$7E,$F7
	.dc.b	$E1,$FC,$3C,$8F,$85,$CD,$70,$FA,$DE,$43,$9F,$F3,$0D,$AF,$07,$B1
	.dc.b	$89,$F9,$A7,$CE,$F8,$4B,$3F,$35,$FA,$3F,$0F,$23,$E8,$7C,$2D,$C7
	.dc.b	$BD,$76,$3D,$DE,$BD,$AF,$77,$AD,$F6,$BE,$1F,$E1,$36,$3E,$D7,$57
	.dc.b	$DE,$3E,$FF,$55,$8F,$D9,$EC,$D3,$D6,$7C,$BF,$56,$CF,$CB,$F4,$FE
	.dc.b	$77,$A7,$F3,$DB,$1F,$87,$F8,$6F,$C5,$9F,$8C,$F8,$6B,$1D,$B8,$7E
	.dc.b	$39,$FB,$FF,$DF,$BF,$D2,$1C,$5F,$E8,$08,$C7,$AD,$9D,$6D,$FF,$08
	.dc.b	$BF,$F4,$2B,$7D,$4E,$47,$3F,$5A,$C7,$B3,$8F,$EA,$9D,$FD,$6B,$1C
	.dc.b	$FD,$5F,$17,$AD,$E3,$6C,$77,$F5,$BC,$C7,$9F,$D6,$B1,$F2,$39,$FF
	.dc.b	$02,$7A,$FD,$5B,$1E,$7F,$5B,$DD,$EA,$7B,$DB,$1E,$BF,$57,$E1,$1F
	.dc.b	$0F,$CC,$B1,$F5,$3B,$3F,$1A,$7C,$7F,$56,$C7,$C3,$F5,$7E,$57,$AB
	.dc.b	$F2,$DB,$1F,$1F,$D4,$F9,$C7,$CF,$F5,$6C,$7E,$03,$B7,$F3,$27,$D3
	.dc.b	$F5,$6C,$7F,$36,$FD,$FD,$F1,$DF,$CD,$5F,$E0,$50,$EF,$91,$A8,$E1
	.dc.b	$EB,$6F,$D6,$E1,$EA,$D9,$E8,$F3,$F5,$64,$77,$F5,$6E,$3F,$C8,$F1
	.dc.b	$7A,$BE,$1F,$4F,$C4,$DB,$F3,$7A,$7E,$53,$CD,$E9,$D8,$FC,$EF,$93
	.dc.b	$FB,$63,$D5,$E9,$D8,$F7,$7A,$5E,$DF,$4B,$DC,$D8,$F8,$5E,$87,$C1
	.dc.b	$3E,$17,$9B,$63,$F5,$3F,$0F,$FC,$08,$F8,$DE,$4D,$8F,$95,$E2,$7C
	.dc.b	$9E,$FF,$94,$D8,$F9,$DD,$5F,$34,$F9,$DA,$D8,$FE,$CF,$F8,$CF,$F8
	.dc.b	$EF,$90,$DE,$46,$7F,$DA,$63,$5E,$DC,$F5,$FF,$B6,$37,$0F,$EE,$2F
	.dc.b	$F2,$4F,$65,$1F,$6F,$14,$E6,$30,$BE,$23,$9F,$C4,$F7,$C2,$57,$4F
	.dc.b	$89,$B8,$7F,$5B,$C3,$FC,$77,$3C,$BE,$23,$C7,$F1,$1E,$47,$CB,$EE
	.dc.b	$C9,$F2,$FC,$45,$C3,$FB,$26,$DE,$37,$F2,$0D,$7C,$FF,$11,$F0,$75
	.dc.b	$F7,$FC,$47,$C0,$F8,$8B,$8F,$C0,$7E,$1F,$C0,$C8,$F8,$7F,$11,$71
	.dc.b	$EF,$C5,$F2,$CF,$93,$F0,$56,$BE,$4F,$C4,$7C,$CF,$88,$F9,$AD,$8F
	.dc.b	$9B,$F0,$9F,$40,$FA,$3F,$07,$63,$ED,$F3,$7D,$73,$EA,$FC,$25,$8F
	.dc.b	$AB,$F1,$17,$4F,$94,$DE,$AF,$AF,$1F,$B1,$E9,$FA,$F5,$F7,$1F,$F1
	.dc.b	$57,$F0,$07,$F2,$C7,$33,$F2,$BA,$BF,$0C,$6B,$E1,$AD,$FE,$3A,$2C
	.dc.b	$FF,$84,$EF,$F9,$7B,$C6,$7F,$C7,$30,$F8,$3E,$E7,$83,$E1,$2C,$FE
	.dc.b	$63,$B7,$F2,$87,$93,$E1,$2C,$64,$98,$77,$F8,$3C,$BE,$1F,$84,$F5
	.dc.b	$9E,$CF,$84,$B5,$FA,$4F,$07,$F4,$07,$C0,$F8,$4B,$19,$25,$FA,$FE
	.dc.b	$13,$2F,$B7,$E0,$FE,$39,$F2,$3E,$12,$D7,$EC,$3C,$5F,$D6,$1F,$33
	.dc.b	$E1,$2C,$64,$97,$F1,$FE,$13,$2F,$C9,$F8,$4F,$A6,$7D,$4F,$83,$B5
	.dc.b	$FB,$CF,$23,$F7,$07,$D8,$F8,$4B,$19,$25,$FD,$3F,$84,$CB,$FA,$4F
	.dc.b	$27,$F8,$01,$F8,$0F,$84,$B5,$FC,$22,$2C,$FD,$BF,$84,$BD,$FF,$8A
	.dc.b	$8F,$E1,$73,$8A,$FE,$05,$FC,$03,$FC,$EE,$7B,$47,$F1,$2F,$33,$F8
	.dc.b	$71,$FE,$77,$9F,$E8,$73,$FB,$48,$CC,$EC,$FF,$3F,$CF,$F9,$1E,$7F
	.dc.b	$77,$3A,$6C,$EB,$E0,$B1,$87,$97,$CF,$79,$DF,$C7,$4F,$07,$C1,$5B
	.dc.b	$FE,$4B,$17,$0F,$F3,$56,$5F,$E6,$CC,$39,$4E,$5B,$5E,$4F,$80,$C3
	.dc.b	$FC,$87,$17,$A1,$FC,$AC,$F7,$7C,$0D,$9F,$E6,$71,$72,$EF,$F0,$37
	.dc.b	$1F,$44,$DA,$F9,$7F,$03,$6F,$F8,$CF,$A5,$FC,$DC,$F9,$5F,$03,$6B
	.dc.b	$F9,$FC,$59,$F5,$FC,$05,$C7,$E2,$4D,$AF,$AF,$F0,$36,$FF,$92,$FA
	.dc.b	$9F,$D0,$CF,$AD,$F0,$36,$BF,$A5,$C5,$9F,$8F,$F0,$37,$1F,$A1,$36
	.dc.b	$BF,$07,$F0,$36,$FF,$98,$FB,$1F,$D3,$CF,$C1,$7C,$05,$AF,$EA,$F1
	.dc.b	$67,$E9,$FC,$0D,$C7,$EE,$4D,$AF,$C7,$FC,$0D,$BF,$E7,$BE,$DF,$F5
	.dc.b	$B3,$F1,$DF,$03,$6B,$FB,$1C,$59,$FB,$FF,$01,$71,$FC,$44,$DA,$FE
	.dc.b	$B9,$F7,$5F,$DA,$7A,$53,$FD,$9F,$DE,$95,$FE,$D4,$AF,$F4,$4E,$3F
	.dc.b	$D4,$27,$2D,$9D,$7C,$06,$31,$FC,$B5,$59,$EF,$F0,$16,$FF,$AB,$FB
	.dc.b	$FF,$DD,$4F,$F6,$F6,$2F,$80,$C5,$FF,$7E,$9B,$5F,$EC,$3A,$FE,$D2
	.dc.b	$F2,$9D,$11,$D7,$B3,$C3,$2F,$D4,$7C,$07,$F8,$0C,$EF,$EB,$5B,$FF
	.dc.b	$05,$C5,$BE,$DE,$95,$BE,$35,$A3,$E2,$79,$D9,$7F,$4D,$F0,$1F,$E0
	.dc.b	$F3,$D7,$E5,$D8,$FF,$0B,$C5,$C3,$D5,$E5,$DC,$3F,$79,$36,$BE,$AF
	.dc.b	$91,$6F,$FC,$99,$F8,$DF,$FA,$10,$FF,$42,$EF,$FB,$4A,$FF,$28,$61
	.dc.b	$FC,$66,$75,$1F,$D9,$23,$0F,$CE,$F0,$B0,$FF,$87,$67,$6F,$23,$07
	.dc.b	$D4,$EF,$DE,$1E,$28,$AA,$5F,$03,$86,$CF,$E2,$7B,$71,$32,$3B,$57
	.dc.b	$B2,$3F,$03,$D3,$98,$FF,$80,$17,$C9,$B3,$F8,$6E,$8D,$67,$EF,$F8
	.dc.b	$9F,$01,$F8,$AE,$7B,$97,$6C,$3C,$D3,$87,$F1,$5C,$B6,$F9,$56,$F2
	.dc.b	$FE,$FB,$1F,$D0,$F7,$FF,$03,$BA,$BF,$2F,$C0,$7B,$CF,$8B,$F9,$77
	.dc.b	$FA,$07,$0C,$1C,$E2,$CA,$BF,$C2,$F5,$FE,$87,$D7,$31,$F6,$6A,$DF
	.dc.b	$FA,$2E,$BF,$D1,$99,$E3,$0F,$C1,$FC,$0F,$CF,$3C,$2F,$F8,$46,$BD
	.dc.b	$51,$96,$9F,$42,$BF,$D2,$D1,$99,$F9,$ED,$C3,$CB,$0F,$81,$C5,$FE
	.dc.b	$1D,$FB,$C7,$83,$E0,$6C,$7C,$98,$B3,$EF,$8F,$52,$F8,$1C,$DF,$E5
	.dc.b	$EA,$D6,$BE,$06,$DF,$E8,$30,$FC,$61,$EE,$F8,$1B,$5F,$5E,$2C,$F7
	.dc.b	$19,$FF,$AC,$FE,$07,$39,$7E,$96,$AD,$F9,$3E,$06,$E1,$FA,$ED,$3F
	.dc.b	$34,$7C,$CF,$81,$B5,$F8,$58,$B3,$DF,$2F,$F6,$07,$C0,$E3,$2F,$D9
	.dc.b	$55,$BF,$83,$F0,$37,$0F,$DD,$ED,$FA,$73,$EC,$FC,$0D,$AF,$C9,$45
	.dc.b	$9E,$C3,$E0,$71,$8F,$EF,$E6,$D7,$CE,$F8,$1B,$7F,$ED,$3F,$C7,$3F
	.dc.b	$C0,$13,$FE,$9B,$6C,$7C,$61,$F0,$5F,$BC,$8C,$ED,$FE,$DB,$F8,$2F
	.dc.b	$C6,$9F,$8E,$F8,$2B,$87,$E2,$E7,$FC,$49,$9F,$E7,$1B,$1F,$99,$7F
	.dc.b	$C4,$67,$D4,$F8,$AF,$E0,$DC,$D2,$79,$7C,$4D,$D3,$F8,$0C,$DC,$3A
	.dc.b	$7C,$25,$BF,$E3,$13,$6B,$EF,$E2,$F8,$4C,$47,$F1,$5E,$9F,$E2,$87
	.dc.b	$ED,$A7,$E0,$C6,$6F,$F8,$04,$E9,$F6,$7C,$1D,$9F,$E2,$73,$72,$F7
	.dc.b	$7C,$1D,$BF,$E5,$5D,$5F,$C8,$4E,$DF,$07,$63,$F8,$64,$5C,$3F,$11
	.dc.b	$93,$E0,$B1,$8F,$F3,$9A,$B7,$F2,$7E,$0A,$E1,$FD,$0B,$B7,$F9,$71
	.dc.b	$EB,$F8,$2B,$5F,$C6,$62,$CF,$6F,$82,$B8,$FF,$45,$9B,$5F,$57,$E0
	.dc.b	$AD,$FF,$4E,$F0,$3F,$9D,$1F,$1F,$E0,$AD,$7F,$25,$8B,$3E,$AF,$82
	.dc.b	$B8,$FF,$51,$9B,$5F,$81,$F8,$2B,$7F,$D6,$BC,$2F,$E8,$C7,$D3,$F8
	.dc.b	$2B,$5F,$CC,$62,$CF,$C6,$F8,$2B,$8F,$F5,$D9,$B5,$F8,$DF,$82,$B7
	.dc.b	$FD,$9B,$C5,$FE,$A4,$7D,$FF,$82,$B5,$FC,$F6,$2C,$FE,$E7,$0F,$E7
	.dc.b	$0F,$F5,$08,$CC,$7F,$6B,$7F,$9E,$BF,$E2,$23,$13,$FD,$1F,$C6,$FE
	.dc.b	$C0,$6B,$E0,$6C,$64,$97,$DF,$E0,$72,$FF,$5E,$AB,$7F,$D1,$0F,$F1
	.dc.b	$15,$70,$7F,$BE,$78,$FF,$DB,$4F,$1F,$C3,$59,$FE,$B1,$16,$FD,$9F
	.dc.b	$03,$6F,$8D,$68,$FE,$84,$BE,$07,$51,$FE,$03,$F2,$BF,$BD,$1F,$03
	.dc.b	$E0,$6C,$7F,$65,$8B,$87,$C9,$F8,$1B,$87,$F8,$46,$6D,$7F,$59,$87
	.dc.b	$C0,$E2,$FF,$C3,$7E,$67,$F8,$1C,$F9,$BF,$03,$6F,$FB,$94,$59,$FA
	.dc.b	$DF,$01,$70,$FF,$0F,$4D,$AE,$DF,$09,$6F,$FC,$4F,$E7,$FF,$84,$CF
	.dc.b	$B3,$F0,$36,$BF,$BF,$45,$9F,$C1,$7C,$0D,$C7,$FC,$55,$36,$BD,$5F
	.dc.b	$03,$6C,$3F,$1C,$FE,$3E,$E1,$BE,$57,$F7,$FF,$82,$FF,$13,$C6,$A7
	.dc.b	$87,$A9,$83,$F1,$BF,$05,$71,$FF,$0C,$56,$CE,$79,$BE,$0B,$58,$7F
	.dc.b	$5F,$1E,$BB,$D3,$FE,$74,$7E,$63,$E0,$AE,$3F,$80,$F8,$3F,$F0,$BF
	.dc.b	$DD,$23,$F3,$B1,$87,$3F,$05,$AF,$81,$B1,$FE,$09,$AF,$F1,$2B,$F0
	.dc.b	$B2,$F8,$5F,$19,$EE,$3A,$FC,$1D,$9F,$5C,$59,$FE,$D1,$4F,$83,$C6
	.dc.b	$FF,$EF,$1E,$0F,$F7,$4D,$74,$C3,$C1,$F0,$7A,$7C,$13,$FE,$E9,$AE
	.dc.b	$1F,$01,$EC,$F8,$3F,$9E,$CA,$F9,$6F,$1C,$3E,$B3,$97,$E5,$5F,$39
	.dc.b	$FF,$18,$65,$FE,$17,$78,$9F,$53,$E0,$7E,$D9,$FE,$62,$FD,$AF,$FC
	.dc.b	$4D,$19,$25,$F4,$9E,$D9,$7B,$5E,$33,$A8,$FF,$21,$DB,$E0,$F3,$7E
	.dc.b	$4F,$82,$FC,$49,$F6,$3E,$0E,$E1,$F8,$68,$B3,$F3,$3E,$0E,$E3,$F2
	.dc.b	$E6,$D7,$B3,$E0,$2D,$FE,$5F,$E3,$7F,$2A,$7A,$FE,$16,$D7,$E4,$A2
	.dc.b	$CF,$8E,$30,$47,$E4,$C7,$E3,$3D,$FD,$E3,$F6,$FA,$FF,$42,$62,$FE
	.dc.b	$CB,$6F,$F5,$1F,$05,$F9,$35,$F7,$4E,$1F,$BF,$97,$EA,$AB,$53,$9A
	.dc.b	$F5,$61,$FA,$8F,$86,$E9,$1F,$B5,$7F,$62,$48,$FC,$D7,$BF,$67,$D3
	.dc.b	$1F,$B0,$BE,$B8,$F8,$23,$07,$F0,$0C,$3C,$15,$D6,$3C,$07,$C3,$CF
	.dc.b	$F3,$58,$FF,$00,$F8,$69,$3F,$A9,$F7,$6D,$FE,$43,$2E,$EF,$C2,$9E
	.dc.b	$D9,$78,$63,$51,$87,$F8,$75,$F2,$C7,$F1,$3E,$DD,$AF,$D2,$FC,$37
	.dc.b	$F1,$03,$F2,$DE,$D5,$8F,$5E,$1F,$3B,$1F,$D5,$E5,$E2,$8F,$E3,$9B
	.dc.b	$FC,$FA,$F0,$47,$F1,$F9,$C3,$FA,$8C,$EE,$7F,$B9,$9B,$5E,$23,$BC
	.dc.b	$65,$E7,$D4,$9F,$03,$FE,$34,$31,$1A,$8E,$77,$F0,$23,$99,$FC,$0B
	.dc.b	$E1,$BF,$96,$9F,$B3,$F6,$AC,$7F,$2A,$8E,$A7,$6A,$B1,$E2,$8D,$D7
	.dc.b	$F1,$4D,$2E,$3F,$C4,$FE,$1B,$F9,$C9,$FC,$A3,$1B,$87,$F3,$68,$F4
	.dc.b	$9F,$6B,$CD,$76,$FF,$9B,$27,$DF,$3A,$78,$1F,$CA,$FD,$3E,$79,$7F
	.dc.b	$12,$F8,$6F,$E8,$E7,$F3,$6F,$46,$D7,$F4,$48,$F8,$E7,$39,$FB,$47
	.dc.b	$F5,$17,$31,$B3,$FD,$81,$E7,$FF,$4B,$8F,$EA,$7F,$0D,$27,$3E,$45
	.dc.b	$BF,$E4,$73,$F5,$0E,$6F,$63,$FB,$03,$99,$FE,$AE,$FF,$B8,$AF,$F9
	.dc.b	$54,$FF,$63,$8F,$EC,$3F,$0D,$2B,$C7,$CD,$6B,$F9,$8C,$FE,$08,$F9
	.dc.b	$57,$FE,$7B,$AF,$C3,$C6,$5F,$EC,$AF,$7E,$6E,$5C,$3F,$D2,$7E,$1B
	.dc.b	$FB,$79,$FC,$2D,$FE,$19,$83,$F6,$CF,$ED,$93,$AC,$3F,$26,$F5,$9F
	.dc.b	$77,$3A,$3F,$2A,$6E,$3F,$D8,$19,$7F,$6F,$AF,$EE,$F3,$FD,$FB,$E1
	.dc.b	$E4,$60,$FE,$F6,$F1,$EE,$FE,$13,$F8,$EB,$CE,$7C,$7F,$8A,$FF,$5F
	.dc.b	$39,$8C,$65,$FD,$5B,$E2,$7C,$5B,$7F,$67,$7F,$C1,$A4,$BF,$BF,$E1
	.dc.b	$5A,$FE,$BB,$A7,$EF,$71,$E9,$1E,$E8,$C6,$5E,$28,$F8,$38,$71,$C7
	.dc.b	$E6,$63,$FE,$14,$8F,$F6,$36,$32,$3F,$1B,$DF,$6F,$FC,$29,$97,$6D
	.dc.b	$3D,$F9,$75,$7F,$94,$EF,$F3,$70,$F7,$CE,$78,$FF,$C5,$BD,$6E,$1F
	.dc.b	$F5,$35,$7E,$6B,$C0,$FC,$77,$81,$F8,$F7,$FD,$B1,$AC,$9F,$CF,$F8
	.dc.b	$17,$6F,$1E,$7D,$5F,$F1,$CE,$DE,$5D,$3F,$C1,$7C,$5F,$0B,$8F,$F9
	.dc.b	$76,$DE,$1E,$DF,$CE,$75,$FE,$75,$BA,$7F,$90,$B9,$7F,$1B,$EA,$F9
	.dc.b	$B4,$FF,$1A,$36,$3F,$AE,$61,$EB,$EB,$FF,$77,$70,$FE,$57,$0F,$F1
	.dc.b	$56,$DE,$3C,$7C,$BB,$F9,$DF,$F2,$76,$DE,$57,$F7,$F8,$FC,$5E,$BF
	.dc.b	$D8,$1F,$B1,$9B,$87,$94,$FF,$80,$1C,$7F,$EF,$D3,$F6,$DD,$36,$35
	.dc.b	$87,$83,$87,$E9,$9F,$90,$E7,$FB,$2F,$C4,$E6,$74,$FE,$FC,$FE,$01
	.dc.b	$F1,$37,$0E,$78,$7F,$95,$39,$FD,$CF,$F9,$A3,$91,$1F,$C1,$FE,$27
	.dc.b	$33,$BA,$E3,$C5,$F0,$E0,$DE,$AF,$C4,$47,$F1,$38,$F9,$1C,$4F,$CC
	.dc.b	$3F,$D7,$7C,$FF,$41,$B8,$7E,$31,$FF,$5D,$36,$FF,$CE,$BA,$48,$F8
	.dc.b	$77,$CF,$E2,$B1,$5F,$8A,$E0,$F8,$AF,$EA,$6A,$F0,$FE,$73,$0D,$E5
	.dc.b	$E1,$F8,$AB,$5F,$E5,$8D,$BF,$D0,$67,$C4,$CF,$D1,$F1,$B8,$9F,$83
	.dc.b	$D1,$F8,$1D,$FE,$07,$3F,$E0,$74,$FD,$4E,$3B,$D7,$FC,$B3,$AD,$D7
	.dc.b	$F8,$FE,$BF,$02,$7F,$D2,$ED,$7F,$91,$77,$FD,$08,$C6,$5F,$3C,$EF
	.dc.b	$A7,$B3,$BB,$FC,$FD,$3C,$4E,$5C,$6C,$F9,$63,$CF,$86,$61,$F2,$B8
	.dc.b	$DF,$17,$8A,$F9,$47,$D9,$E1,$FC,$F6,$3F,$73,$83,$F1,$F3,$FD,$3F
	.dc.b	$3F,$AE,$FA,$B2,$F9,$18,$BE,$2F,$A7,$07,$C0,$E5,$F6,$1F,$8C,$F0
	.dc.b	$BF,$65,$B7,$AA,$3F,$1F,$C3,$EF,$DB,$C3,$C9,$F3,$78,$B5,$7F,$9A
	.dc.b	$D9,$87,$68,$CE,$DF,$EB,$19,$1F,$2C,$F0,$C7,$17,$BE,$1F,$6C,$EF
	.dc.b	$1F,$36,$7E,$AD,$7D,$3E,$1F,$F1,$95,$F8,$2B,$C5,$18,$E5,$FB,$07
	.dc.b	$B7,$96,$CF,$E2,$38,$7A,$63,$FC,$22,$2E,$1F,$49,$FF,$1D,$B6,$3E
	.dc.b	$76,$3E,$19,$CE,$36,$7D,$D8,$DD,$FF,$B5,$ED,$F2,$72,$FE,$0D,$AF
	.dc.b	$F1,$AD,$7F,$8C,$69,$D7,$8B,$FC,$CE,$D9,$F9,$38,$7E,$5B,$3F,$7F
	.dc.b	$0F,$E4,$CF,$B1,$C9,$76,$F6,$6D,$F8,$7D,$3B,$BF,$C5,$CF,$C3,$57
	.dc.b	$29,$F4,$9E,$E7,$E5,$FC,$5F,$E7,$CF,$C4,$96,$3A,$E9,$F9,$03,$FA
	.dc.b	$AF,$17,$E0,$EF,$F5,$35,$CA,$BF,$A2,$F2,$3E,$2F,$15,$F9,$77,$F5
	.dc.b	$87,$E5,$BE,$2E,$D7,$6D,$FE,$36,$9E,$AB,$FE,$3B,$C3,$F2,$30,$FC
	.dc.b	$09,$D3,$4F,$92,$FE,$13,$1F,$DC,$1F,$A2,$F8,$AB,$97,$5C,$3E,$25
	.dc.b	$FA,$A3,$F0,$35,$C5,$FC,$0D,$DA,$FF,$1E,$76,$3C,$4F,$E6,$6D,$AF
	.dc.b	$DD,$7C,$56,$27,$F3,$7A,$7E,$3B,$AF,$F7,$1B,$7B,$31,$FE,$AB,$BB
	.dc.b	$E2,$BF,$CC,$8D,$CB,$E4,$FC,$56,$27,$F3,$FC,$1F,$C4,$CF,$DE,$FC
	.dc.b	$55,$AF,$66,$1F,$1B,$3F,$91,$D5,$FD,$17,$A9,$87,$2B,$FA,$BF,$15
	.dc.b	$70,$FE,$15,$C1,$FC,$7C,$FE,$9F,$FB,$6F,$EE,$2B,$F7,$D1,$EC,$9F
	.dc.b	$C7,$1F,$CF,$4B,$CB,$F9,$A8,$FD,$A9,$D5,$C4,$FE,$B3,$5F,$E5,$87
	.dc.b	$77,$F8,$53,$D6,$3F,$1D,$C1,$FE,$10,$9F,$DA,$D7,$E3,$EF,$F8,$5B
	.dc.b	$78,$DE,$5F,$D4,$3F,$90,$FC,$5D,$BF,$8B,$BF,$F3,$8C,$3C,$F8,$FD
	.dc.b	$3C,$BB,$C7,$F4,$2C,$BA,$79,$9F,$78,$FE,$57,$F1,$77,$1E,$7D,$5F
	.dc.b	$C5,$AF,$99,$F8,$CC,$7A,$1F,$D1,$30,$F3,$79,$9F,$88,$3F,$9A,$FC
	.dc.b	$5D,$C3,$9F,$2F,$D8,$C7,$F9,$13,$F0,$23,$F8,$BF,$07,$F7,$CF,$0B
	.dc.b	$FC,$57,$DF,$FC,$FB,$E2,$FF,$26,$FC,$9D,$3C,$3A,$8F,$E3,$C7,$98
	.dc.b	$FC,$31,$7A,$1B,$7C,$BF,$37,$F3,$A7,$F4,$AF,$8B,$BA,$79,$AF,$F0
	.dc.b	$1C,$7F,$21,$E6,$DC,$39,$CF,$A3,$CD,$FD,$39,$FD,$53,$E2,$EC,$73
	.dc.b	$AF,$D9,$5F,$97,$9B,$F9,$43,$F9,$49,$F8,$3E,$57,$EC,$0F,$F0,$27
	.dc.b	$E6,$28,$C1,$CF,$6F,$C4,$57,$3E,$4F,$96,$F3,$3E,$17,$1F,$F7,$5A
	.dc.b	$FF,$19,$B6,$1C,$7F,$E5,$4C,$1F,$19,$FC,$03,$C9,$F9,$D5,$F7,$CF
	.dc.b	$9D,$E2,$7E,$F9,$C1,$F1,$33,$FA,$6F,$E6,$32,$FE,$A3,$F9,$0F,$F8
	.dc.b	$3C,$FF,$09,$7A,$0C,$9F,$1B,$FE,$40,$9F,$F0,$8E,$DE,$AE,$1F,$8E
	.dc.b	$7D,$1E,$1F,$8E,$FE,$BB,$2F,$BD,$5F,$E0,$44,$FF,$81,$B1,$ED,$7D
	.dc.b	$7A,$BD,$75,$E7,$3D,$7D,$3F,$AF,$C7,$AE,$BF,$9B,$D7,$D3,$E7,$F9
	.dc.b	$4F,$A7,$F1,$B6,$3E,$96,$1C,$9F,$ED,$F7,$CA,$3F,$5D,$1F,$0B,$A3
	.dc.b	$DC,$7D,$CF,$8D,$B1,$CA,$7F,$33,$BF,$BF,$5F,$85,$9F,$8E,$7D,$FE
	.dc.b	$8F,$C7,$3F,$0B,$F1,$B7,$0E,$5C,$9F,$56,$BD,$C4,$D3,$F9,$37,$77
	.dc.b	$F0,$DE,$3F,$F0,$07,$47,$FB,$91,$B8,$7F,$AC,$39,$9F,$0D,$F5,$7D
	.dc.b	$3F,$93,$8F,$F0,$DE,$0F,$37,$37,$EE,$74,$FA,$7E,$9F,$DB,$3F,$33
	.dc.b	$F1,$77,$3F,$2F,$47,$CF,$E4,$F6,$DF,$C6,$3E,$DF,$A9,$F8,$33,$F4
	.dc.b	$3F,$17,$6F,$89,$F3,$AB,$8D,$7E,$3C,$FC,$1F,$A9,$F8,$B3,$F5,$3F
	.dc.b	$17,$6B,$8C,$FE,$9F,$4F,$46,$BF,$BD,$E2,$7A,$DF,$93,$3F,$63,$F1
	.dc.b	$57,$0E,$35,$F0,$30,$FE,$19,$CD,$F2,$FD,$3F,$CD,$1F,$B8,$F8,$AB
	.dc.b	$7C,$3B,$3F,$4B,$CB,$F8,$5C,$3D,$18,$F8,$71,$FD,$11,$FC,$03,$E2
	.dc.b	$6E,$1C,$6B,$F5,$F7,$C7,$BF,$F8,$8C,$7F,$1D,$F5,$FF,$CF,$47,$F5
	.dc.b	$CF,$4A,$DF,$19,$FC,$17,$27,$C8,$BE,$39,$FF,$1B,$F5,$FF,$D0,$47
	.dc.b	$0A,$FF,$2E,$3F,$80,$AF,$E8,$53,$F0,$4F,$E4,$5A,$7F,$57,$B7,$AF
	.dc.b	$FE,$8B,$39,$E3,$FE,$69,$79,$72,$7F,$41,$3E,$D1,$FE,$52,$AE,$59
	.dc.b	$FF,$A4,$CF,$F6,$1F,$D3,$3D,$5C,$61,$EB,$DF,$E5,$1F,$9C,$7F,$CB
	.dc.b	$BE,$AF,$FA,$6C,$DE,$FF,$E7,$57,$C2,$FF,$3C,$E2,$FA,$7B,$FC,$3C
	.dc.b	$FD,$5E,$A7,$F9,$C0,$FF,$38,$DF,$FA,$95,$FC,$C6,$BE,$9C,$3F,$14
	.dc.b	$7F,$2F,$E0,$FE,$89,$EA,$7F,$9E,$0E,$59,$FF,$9B,$9E,$33,$E1,$3E
	.dc.b	$DE,$5E,$19,$FF,$3A,$FA,$7F,$E8,$03,$FA,$04,$AD,$A7,$F5,$F8,$FF
	.dc.b	$78,$F4,$6E,$FE,$5C,$75,$3E,$FE,$5F,$9E,$AF,$E6,$BC,$6F,$4B,$FD
	.dc.b	$1A,$71,$AF,$F5,$73,$E6,$C7,$F9,$8D,$7C,$6B,$E5,$5F,$92,$C3,$CB
	.dc.b	$5E,$48,$FE,$E5,$C1,$FD,$5E,$7F,$D1,$B8,$6B,$0F,$44,$E6,$BE,$4D
	.dc.b	$F6,$7E,$1E,$1D,$9F,$E3,$1E,$87,$17,$AB,$57,$A7,$CE,$BC,$50,$BF
	.dc.b	$EB,$79,$7C,$49,$ED,$3F,$10,$FD,$CC,$7F,$38,$F4,$3A,$57,$F5,$48
	.dc.b	$FF,$06,$57,$48,$E7,$5F,$4F,$2F,$5D,$FC,$77,$FB,$9D,$7E,$95,$F9
	.dc.b	$FC,$7F,$80,$AF,$B0,$7F,$85,$0B,$6E,$06,$9A,$9F,$F4,$E1,$A8,$CB
	.dc.b	$99,$E1,$3B,$8E,$87,$F9,$C4,$E0,$E6,$37,$1C,$67,$9B,$A3,$A7,$9C
	.dc.b	$F7,$B8,$36,$33,$18,$23,$82,$1E,$C8,$E0,$D8,$DB,$4D,$9F,$84,$F3
	.dc.b	$FF,$39,$9F,$CF,$6E,$79,$9F,$A6,$FE,$2B,$CE,$CC,$7E,$43,$0F,$35
	.dc.b	$FD,$76,$D7,$E4,$1B,$3F,$E4,$7F,$C8,$7F,$CD,$6F,$43,$96,$35,$FC
	.dc.b	$54,$F0,$6C,$67,$37,$C8,$78,$A7,$F2,$19,$59,$34,$D7,$0B,$C5,$D6
	.dc.b	$3D,$68,$D3,$C1,$97,$9B,$0F,$C8,$1E,$62,$9E,$42,$CF,$2C,$BC,$D9
	.dc.b	$78,$FC,$D9,$5E,$97,$CD,$E5,$F8,$AB,$D9,$8F,$85,$F8,$58,$F8,$76
	.dc.b	$F9,$91,$A2,$61,$E0,$C7,$9B,$F1,$70,$AF,$93,$C9,$EC,$F9,$B1,$EC
	.dc.b	$FA,$B1,$ED,$BF,$D7,$C7,$AE,$5E,$0A,$EC,$7D,$1C,$3E,$71,$1F,$B1
	.dc.b	$FB,$73,$E3,$BB,$D7,$12,$61,$FC,$A7,$1E,$F7,$F2,$E7,$2D,$9F,$E2
	.dc.b	$FA,$7D,$BD,$E4,$FE,$05,$FB,$DD,$7D,$AB,$D1,$3E,$DD,$3E,$4F,$B7
	.dc.b	$98,$E8,$43,$F8,$B6,$1C,$9F,$C6,$5D,$7E,$DF,$91,$FC,$FF,$D2,$FE
	.dc.b	$87,$1F,$7F,$3F,$3B,$72,$F6,$63,$DE,$30,$77,$E3,$FC,$BB,$F9,$9C
	.dc.b	$3A,$75,$FF,$34,$6C,$7F,$3A,$6B,$FC,$AF,$C9,$B5,$F8,$D8,$F7,$13
	.dc.b	$1F,$E6,$3B,$FC,$C8,$F0,$92,$78,$4F,$F2,$BD,$7F,$4D,$7D,$09,$5F
	.dc.b	$EB,$BD,$3D,$FE,$2B,$DA,$F0,$F2,$FE,$23,$1F,$83,$9F,$7C,$FA,$EF
	.dc.b	$F4,$A7,$E2,$C7,$C0,$CB,$F4,$BF,$1D,$EA,$23,$DF,$7F,$EA,$BF,$1D
	.dc.b	$DB,$8F,$F2,$38,$F3,$3E,$6D,$60,$F9,$99,$FE,$E7,$83,$F8,$05,$FC
	.dc.b	$62,$4F,$53,$FB,$2F,$C7,$75,$7E,$45,$F5,$7F,$8A,$71,$FD,$9D,$3D
	.dc.b	$71,$F6,$70,$FE,$19,$7F,$48,$91,$D4,$73,$BE,$17,$B6,$FF,$0B,$4F
	.dc.b	$E1,$79,$75,$7E,$56,$7F,$9E,$C3,$F9,$07,$A2,$F8,$0F,$C0,$B6,$3B
	.dc.b	$35,$FF,$64,$F0,$7F,$A1,$F8,$7B,$DF,$AE,$7F,$03,$C5,$F5,$71,$EF
	.dc.b	$1E,$B3,$BF,$A1,$D8,$8F,$C7,$F7,$3F,$D8,$DC,$7D,$23,$E7,$F2,$7E
	.dc.b	$B1,$FA,$47,$E3,$6F,$E9,$3E,$AF,$47,$D2,$47,$A7,$B2,$E4,$E9,$1C
	.dc.b	$6F,$A3,$F5,$0E,$E7,$F1,$03,$F9,$7E,$FF,$E5,$E7,$E1,$FA,$5F,$18
	.dc.b	$FD,$4E,$7F,$ED,$DF,$8E,$EB,$D1,$F6,$AB,$E1,$F2,$7D,$1C,$7D,$33
	.dc.b	$F9,$5E,$5F,$F0,$CE,$4F,$4F,$E8,$93,$0F,$4F,$4F,$4E,$0F,$97,$0E
	.dc.b	$3F,$9D,$97,$49,$F8,$73,$E8,$F4,$BE,$D9,$1F,$87,$F1,$BC,$EB,$E5
	.dc.b	$DF,$29,$F8,$47,$EE,$70,$FE,$BD,$93,$D3,$FC,$29,$27,$E6,$FC,$6F
	.dc.b	$37,$8D,$F3,$D7,$F2,$58,$FF,$54,$9F,$C7,$FA,$9F,$8E,$23,$F8,$1F
	.dc.b	$CC,$BF,$C1,$3E,$17,$F2,$0E,$1F,$94,$4E,$0F,$CD,$D7,$BF,$0F,$97
	.dc.b	$B7,$9B,$D5,$FC,$C9,$27,$A7,$C7,$79,$74,$FE,$13,$C9,$F4,$E7,$DF
	.dc.b	$97,$D3,$ED,$FF,$37,$71,$7F,$99,$E7,$F4,$3E,$AF,$E9,$0F,$BF,$BB
	.dc.b	$E4,$BA,$5F,$E8,$78,$BE,$03,$FA,$3C,$7D,$3B,$7F,$81,$FE,$37,$F9
	.dc.b	$DF,$17,$F3,$C6,$F8,$5F,$E7,$D6,$E8,$EB,$69,$FE,$5E,$BF,$89,$C9
	.dc.b	$F8,$2C,$BF,$1D,$A7,$83,$0F,$CC,$F0,$FF,$2F,$7F,$C3,$99,$3D,$7F
	.dc.b	$E7,$AD,$8F,$E1,$4D,$7F,$7A,$78,$72,$7C,$8F,$BA,$3F,$AD,$61,$DD
	.dc.b	$FD,$51,$FA,$7C,$3F,$CE,$76,$F6,$3B,$92,$7E,$F7,$C8,$73,$7F,$4A
	.dc.b	$7D,$B3,$E9,$F0,$7C,$7F,$5F,$CE,$48,$FC,$37,$A3,$C7,$6F,$83,$CF
	.dc.b	$F6,$F3,$FE,$83,$87,$CF,$E0,$F7,$93,$0F,$C7,$7C,$77,$18,$FC,$5C
	.dc.b	$7E,$09,$F9,$93,$F5,$FD,$8F,$8E,$49,$FC,$B7,$C8,$71,$ED,$FE,$13
	.dc.b	$8F,$E7,$A7,$D2,$7D,$0F,$67,$E7,$91,$FC,$F7,$0F,$17,$FC,$45,$A7
	.dc.b	$A7,$A7,$F1,$19,$FE,$1F,$D8,$FA,$E4,$AF,$D3,$7C,$87,$1C,$BF,$B8
	.dc.b	$E1,$C6,$7F,$97,$1F,$E4,$6C,$5E,$CF,$E0,$09,$1F,$AF,$F4,$79,$67
	.dc.b	$FD,$7A,$3E,$16,$BF,$AC,$7F,$13,$EC,$FE,$20,$8F,$ED,$FE,$3F,$8B
	.dc.b	$FD,$68,$FB,$59,$FD,$27,$FC,$D1,$E5,$FF,$82,$88,$FF,$59,$E5,$FE
	.dc.b	$ED,$DD,$F9,$CC,$78,$E5,$E3,$C3,$F3,$57,$E4,$3F,$1D,$1F,$D6,$5F
	.dc.b	$CB,$9F,$E2,$C3,$A7,$0F,$F8,$6B,$D4,$67,$E6,$9F,$E5,$1C,$1F,$E7
	.dc.b	$7A,$91,$FC,$54,$FF,$35,$F9,$5F,$E7,$B3,$FC,$42,$47,$F8,$B4,$F3
	.dc.b	$ED,$FF,$09,$4E,$E7,$FD,$0C,$FF,$A5,$BC,$1F,$DC,$FB,$FF,$C0,$7D
	.dc.b	$BF,$D8,$92,$BF,$90,$7C,$6F,$3C,$78,$78,$6F,$85,$A4,$AF,$F4,$C6
	.dc.b	$3F,$C1,$3D,$9F,$F4,$87,$9D,$FE,$36,$24,$FF,$2B,$F7,$BA,$3F,$E9
	.dc.b	$98,$DB,$4F,$F3,$FF,$27,$FA,$6F,$7F,$E6,$3E,$4F,$F9,$9E,$3F,$C8
	.dc.b	$71,$89,$FE,$09,$F1,$BD,$1F,$F3,$D4,$F1,$23,$FE,$8A,$F3,$BA,$1C
	.dc.b	$88,$FF,$3F,$F8,$CE,$5D,$1F,$E4,$A9,$DC,$F1,$3D,$F8,$FF,$BA,$A7
	.dc.b	$FC,$9F,$E6,$F9,$09,$3F,$E2,$3F,$8E,$7C,$57,$4E,$BF,$F4,$05,$7B
	.dc.b	$49,$5F,$E9,$E8,$F8,$07,$B8,$8F,$F8,$B7,$5F,$E4,$2F,$E5,$FC,$8F
	.dc.b	$E6,$07,$F0,$8C,$7F,$CC,$35,$CA,$7A,$1F,$EA,$AF,$2B,$E4,$F9,$D2
	.dc.b	$3A,$3F,$95,$E7,$F3,$47,$FB,$56,$B9,$3F,$EA,$57,$EA,$57,$D2,$F3
	.dc.b	$64,$F9,$5F,$AB,$1C,$A3,$FC,$08,$FA,$9E,$47,$8C,$F6,$3C,$1A,$7F
	.dc.b	$92,$FB,$DD,$9C,$F4,$FF,$56,$4E,$8F,$D7,$B8,$6F,$99,$F8,$62,$3F
	.dc.b	$25,$FF,$12,$F0,$BE,$0F,$A4,$6A,$78,$9F,$E4,$19,$7C,$6F,$E4,$3C
	.dc.b	$D9,$3F,$4D,$FF,$4A,$FE,$DE,$F7,$7C,$D1,$A9,$F2,$BF,$98,$F3,$24
	.dc.b	$7D,$B6,$BF,$D7,$BB,$D8,$BE,$1F,$E0,$56,$BC,$8F,$F5,$8E,$C8,$FD
	.dc.b	$1F,$9D,$31,$FC,$33,$5E,$98,$3E,$3B,$E5,$4E,$70,$E6,$FC,$4E,$BF
	.dc.b	$EE,$7F,$0F,$30,$FE,$E6,$FC,$8F,$8D,$FC,$2E,$BE,$6F,$8D,$F9,$64
	.dc.b	$AF,$91,$F1,$BF,$DF,$3E,$2A,$47,$F7,$C7,$E8,$FC,$6F,$2D,$9F,$1B
	.dc.b	$F5,$09,$5F,$47,$E3,$7F,$C0,$BF,$1B,$23,$FC,$0A,$FD,$AF,$8C,$E5
	.dc.b	$F1,$7F,$70,$93,$F6,$7A,$3F,$C1,$BF,$19,$23,$FC,$1A,$FE,$0B,$E2
	.dc.b	$7C,$5F,$13,$F8,$4F,$89,$95,$F8,$18,$FF,$0A,$FC,$64,$8F,$F0,$AB
	.dc.b	$F8,$AF,$8A,$F5,$7C,$57,$E3,$09,$3F,$89,$F8,$AF,$E2,$FF,$03,$23
	.dc.b	$F7,$4F,$E4,$BE,$27,$F8,$DF,$99,$FD,$9E,$7F,$A1,$81,$E7,$7F,$B7
	.dc.b	$2B,$94,$7E,$81,$E2,$FF,$05,$F7,$FF,$92,$91,$FE,$11,$5A,$7F,$AC
	.dc.b	$57,$F7,$67,$FD,$1F,$E0,$7E,$32,$FA,$BF,$8E,$AE,$0D,$9E,$67,$ED
	.dc.b	$7D,$CE,$97,$CE,$3F,$C1,$27,$A5,$FF,$79,$F9,$7F,$CE,$EF,$9F,$4F
	.dc.b	$F7,$DC,$3F,$3F,$7F,$E4,$DF,$CC,$FF,$C9,$EF,$FA,$2E,$FF,$9F,$91
	.dc.b	$FF,$5A,$47,$F9,$16,$7F,$47,$3F,$D1,$C9,$3F,$D5,$27,$FC,$80,$F8
	.dc.b	$FD,$0F,$98,$F5,$3F,$96,$4F,$57,$F6,$52,$F8,$8F,$E9,$F9,$C9,$FE
	.dc.b	$64,$FF,$AB,$A5,$E9,$E6,$7E,$BE,$1D,$BC,$95,$FF,$5B,$3F,$AE,$7B
	.dc.b	$B6,$B9,$90,$F0,$FC,$47,$F5,$E8,$91,$FD,$05,$F0,$7C,$46,$23,$B9
	.dc.b	$F8,$7B,$FE,$EC,$FB,$CC,$4F,$F6,$BF,$7F,$C3,$1C,$C8,$7C,$0F,$88
	.dc.b	$FE,$DB,$12,$3F,$A8,$3E,$FF,$88,$C4,$78,$0F,$F6,$04,$75,$7E,$87
	.dc.b	$C4,$57,$F9,$AC,$70,$E1,$FB,$FE,$53,$81,$F0,$9F,$DF,$3E,$0E,$61
	.dc.b	$FD,$7D,$FA,$F7,$F8,$68,$7C,$97,$FA,$7D,$D1,$EF,$3F,$DA,$53,$FE
	.dc.b	$09,$3F,$C1,$5F,$07,$63,$67,$F4,$88,$FB,$5E,$87,$F8,$32,$64,$FF
	.dc.b	$6F,$7F,$C1,$FF,$05,$88,$F8,$1C,$29,$EB,$1E,$1F,$C9,$7F,$9C,$3C
	.dc.b	$48,$7F,$56,$8F,$5F,$57,$E4,$A7,$F2,$73,$FD,$6A,$FB,$BF,$DF,$DF
	.dc.b	$F2,$47,$1F,$87,$B1,$D1,$F8,$DC,$7C,$37,$F4,$F0,$F0,$BD,$CF,$0E
	.dc.b	$DF,$D2,$27,$C3,$1F,$C3,$67,$C3,$3F,$35,$B9,$79,$67,$62,$32,$FE
	.dc.b	$DF,$CF,$F2,$61,$E3,$8F,$A5,$A7,$C5,$7E,$27,$AF,$CA,$F8,$C7,$33
	.dc.b	$07,$EA,$FC,$DD,$7B,$92,$3F,$91,$67,$E5,$8F,$D8,$79,$7C,$F0,$F0
	.dc.b	$D7,$E8,$75,$FD,$EE,$DF,$1A,$3D,$D3,$CD,$FF,$3A,$E5,$FE,$01,$7A
	.dc.b	$ED,$F4,$F1,$E9,$18,$3E,$37,$B7,$E9,$23,$93,$E2,$5F,$07,$E8,$1F
	.dc.b	$86,$C7,$C7,$B7,$DF,$9D,$69,$EB,$AE,$59,$78,$DF,$2C,$EB,$2E,$11
	.dc.b	$A8,$FF,$45,$74,$66,$3D,$A6,$4E,$7E,$C6,$5F,$29,$93,$F8,$EF,$95
	.dc.b	$F4,$E2,$47,$03,$F9,$6F,$C7,$61,$F8,$F1,$72,$FD,$CF,$53,$E2,$F1
	.dc.b	$9F,$E6,$EA,$DF,$F3,$BF,$8A,$FC,$07,$B5,$23,$DE,$FF,$8A,$3D,$0C
	.dc.b	$47,$D4,$8B,$97,$E4,$FD,$AB,$8F,$E9,$E6,$D7,$F4,$7F,$8A,$FC,$65
	.dc.b	$48,$F9,$6F,$F3,$5F,$89,$C4,$7E,$02,$2E,$1F,$9F,$F7,$2E,$3F,$B3
	.dc.b	$9B,$4F,$0F,$FB,$8F,$C4,$FE,$66,$E4,$FD,$87,$F5,$73,$FD,$DB,$E2
	.dc.b	$73,$3F,$8E,$9B,$8F,$F2,$0E,$BF,$E3,$13,$93,$91,$0F,$F2,$13,$F9
	.dc.b	$97,$FD,$97,$F0,$0E,$4F,$D9,$B6,$3F,$24,$D7,$8F,$C3,$F9,$29,$F0
	.dc.b	$F7,$E1,$E5,$7E,$77,$4F,$F0,$6C,$EF,$E1,$B8,$7C,$2C,$AD,$60,$9F
	.dc.b	$DD,$7C,$24,$9F,$CC,$BD,$F3,$C4,$7E,$BA,$2E,$BF,$07,$E1,$6F,$1F
	.dc.b	$F2,$19,$B5,$F3,$F1,$7B,$1F,$C1,$FE,$16,$4F,$E9,$9F,$67,$B5,$88
	.dc.b	$FD,$DC,$5C,$BE,$77,$B5,$72,$FE,$5D,$36,$BB,$7B,$5F,$C4,$FE,$16
	.dc.b	$47,$EC,$9F,$91,$EC,$E2,$3F,$83,$C5,$C7,$ED,$79,$F7,$1F,$E7,$53
	.dc.b	$6B,$D5,$E2,$FF,$1F,$F8,$59,$1F,$BE,$7E,$A7,$A7,$88,$FE,$27,$17
	.dc.b	$1F,$C3,$79,$77,$1F,$E8,$D3,$6B,$E3,$7A,$5F,$CA,$FE,$16,$47,$F0
	.dc.b	$A7,$F2,$7F,$99,$FF,$2E,$F3,$51,$A9,$FE,$42,$EB,$D0,$C6,$7F,$D0
	.dc.b	$62,$E5,$FE,$51,$ED,$63,$FE,$A2,$C3,$F8,$BF,$C3,$7F,$3B,$23,$F8
	.dc.b	$B8,$FE,$A3,$87,$F4,$B9,$CC,$FF,$30,$7C,$BE,$5E,$27,$53,$D0,$E5
	.dc.b	$E4,$FF,$47,$F8,$69,$1F,$C9,$9F,$6F,$91,$88,$FE,$77,$16,$BF,$C1
	.dc.b	$B8,$BC,$6C,$69,$FD,$BE,$AD,$FE,$F6,$9D,$DF,$D5,$7E,$1A,$4F,$F3
	.dc.b	$47,$E5,$F5,$E2,$3F,$A4,$C5,$CB,$BF,$45,$CB,$FB,$E4,$DA,$ED,$CB
	.dc.b	$FD,$87,$E1,$A4,$7F,$40,$7E,$BF,$1E,$23,$FA,$AC,$5C,$7D,$7C,$57
	.dc.b	$1F,$F0,$44,$DA,$F5,$6D,$FD,$B7,$CD,$91,$FD,$31,$FC,$1E,$78,$8F
	.dc.b	$EC,$31,$71,$F8,$F7,$70,$FF,$0A,$4D,$AF,$8C,$FF,$79,$F8,$69,$1F
	.dc.b	$D6,$1F,$C7,$98,$7F,$B6,$B6,$FE,$9F,$C3,$5B,$FF,$10,$4D,$AF,$A5
	.dc.b	$F0,$DF,$E0,$5A,$91,$FD,$91,$FC,$DF,$C3,$62,$3F,$BC,$C5,$C3,$FC
	.dc.b	$97,$D3,$FE,$88,$EA,$4E,$72,$FF,$17,$3F,$CD,$A7,$FD,$11,$1F,$E3
	.dc.b	$1C,$7F,$C2,$84,$C3,$F9,$24,$FF,$A2,$1F,$F0,$3C,$65,$A6,$BE,$1B
	.dc.b	$18,$FF,$8A,$66,$DF,$EA,$6F,$FD,$0B,$7F,$CD,$BD,$4F,$E7,$24,$7B
	.dc.b	$FC,$2F,$97,$E1,$64,$78,$9F,$47,$C2,$FA,$7E,$16,$4F,$99,$F6,$7C
	.dc.b	$2F,$F8,$B7,$82,$4F,$F8,$31,$F8,$1F,$0B,$88,$FF,$10,$C5,$DB,$E4
	.dc.b	$FC,$25,$DB,$FC,$A9,$36,$BF,$B1,$6C,$F8,$4F,$F2,$1F,$C0,$7F,$AD
	.dc.b	$89,$7F,$37,$E0,$F2,$CB,$FF,$13,$4F,$F9,$E7,$2F,$AF,$F0,$79,$94
	.dc.b	$7F,$9F,$67,$9C,$7D,$2F,$81,$EF,$F1,$D2,$39,$BF,$6F,$DC,$C4,$72
	.dc.b	$C7,$FD,$7B,$7F,$7A,$3F,$DD,$4D,$F2,$9A,$FF,$86,$67,$99,$0E,$47
	.dc.b	$0C,$3F,$11,$F0,$3F,$03,$E3,$A4,$FB,$1F,$C6,$7C,$16,$23,$89,$FE
	.dc.b	$76,$8C,$6B,$F9,$6F,$82,$BA,$79,$A6,$D7,$F8,$3F,$F3,$37,$C1,$FC
	.dc.b	$EE,$E9,$3F,$29,$FF,$07,$CB,$E0,$F3,$3E,$0E,$0F,$F6,$C3,$B3,$FB
	.dc.b	$F4,$FF,$84,$33,$F9,$D7,$C1,$B3,$AA,$FB,$BF,$13,$23,$EC,$BE,$0F
	.dc.b	$84,$C4,$77,$8B,$5F,$EB,$CC,$3F,$C1,$F3,$C0,$86,$CF,$1F,$C2,$7E
	.dc.b	$2B,$C0,$91,$F8,$67,$DB,$F0,$58,$8F,$64,$5B,$ED,$F0,$77,$83,$EF
	.dc.b	$CD,$AF,$95,$F0,$7F,$96,$F0,$E4,$7E,$49,$F9,$3F,$07,$88,$F9,$11
	.dc.b	$71,$F5,$7C,$15,$C7,$F1,$73,$6B,$EB,$7C,$17,$E8,$BC,$69,$1F,$9E
	.dc.b	$7E,$AF,$C1,$62,$3E,$A4,$5C,$7E,$37,$C1,$5C,$7F,$2F,$36,$BF,$05
	.dc.b	$F0,$5F,$AE,$F2,$64,$7E,$A9,$FC,$0F,$C1,$62,$3F,$01,$17,$1F,$A5
	.dc.b	$F0,$57,$1F,$D1,$CD,$AF,$C7,$7C,$17,$EE,$BC,$C9,$1F,$B6,$7F,$1B
	.dc.b	$F0,$58,$8F,$C6,$45,$C7,$EF,$7C,$15,$C7,$F5,$F3,$6B,$FC,$D1,$F9
	.dc.b	$9F,$FA,$FB,$B7,$FD,$FB,$1F,$C0,$BC,$FF,$E1,$27,$F9,$A2,$7F,$D4
	.dc.b	$18,$FF,$01,$8C,$CE,$CF,$F2,$B4,$FF,$84,$5F,$CF,$46,$9A,$6B,$E1
	.dc.b	$FF,$8A,$FA,$12,$3F,$88,$BD,$FE,$13,$11,$FA,$78,$B8,$33,$FF,$10
	.dc.b	$62,$F8,$4C,$C7,$F0,$CA,$B7,$EE,$F8,$4F,$E4,$7E,$94,$8F,$E3,$EF
	.dc.b	$B7,$E1,$31,$1F,$B5,$8B,$97,$78,$F8,$FF,$07,$8C,$7F,$8C,$D5,$9F
	.dc.b	$97,$F0,$7F,$CC,$3D,$59,$1F,$CB,$1F,$95,$F0,$78,$8F,$E0,$51,$71
	.dc.b	$ED,$E9,$B7,$7C,$1E,$73,$FE,$4F,$77,$0F,$B3,$F0,$7F,$CF,$BD,$89
	.dc.b	$1F,$CE,$9F,$B1,$F0,$78,$8F,$E2,$11,$73,$EF,$93,$E0,$F1,$97,$F3
	.dc.b	$4A,$B7,$F8,$6F,$83,$FE,$97,$ED,$C8,$FE,$8E,$FE,$17,$E0,$F1,$1F
	.dc.b	$C7,$62,$E5,$DB,$E0,$EE,$3F,$D0,$26,$D7,$E4,$BE,$0F,$FA,$BF,$BD
	.dc.b	$23,$FA,$9B,$F9,$1F,$82,$C4,$7F,$29,$8B,$8F,$AB,$E0,$AE,$3F,$D3
	.dc.b	$26,$D7,$E7,$BE,$07,$FB,$1F,$C1,$48,$FE,$BE,$FF,$B1,$3E,$93,$FB
	.dc.b	$33,$98,$FE,$6C,$E7,$E0,$B1,$97,$F6,$18,$B8,$FF,$9C,$6F,$FC,$25
	.dc.b	$3F,$D9,$7E,$0B,$FB,$89,$30,$FF,$5D,$4F,$F4,$07,$FA,$14,$65,$A6
	.dc.b	$BD,$BC,$57,$38,$FE,$EB,$3E,$7F,$5F,$FB,$EF,$C1,$49,$FE,$F2,$99
	.dc.b	$5F,$99,$5D,$0E,$3E,$11,$FD,$4E,$AD,$6E,$9C,$58,$D3,$FB,$BD,$5B
	.dc.b	$ED,$5F,$E3,$A7,$DB,$EA,$FC,$B2,$3F,$03,$D2,$FF,$09,$FC,$14,$8F
	.dc.b	$F0,$7B,$F4,$7D,$0C,$47,$F6,$58,$B9,$FE,$93,$81,$E6,$FD,$7F,$32
	.dc.b	$5F,$CC,$7E,$D7,$93,$F0,$63,$FB,$EF,$E5,$1F,$DD,$9F,$87,$BE,$63
	.dc.b	$11,$AF,$82,$FC,$17,$91,$C7,$3F,$91,$5A,$6C,$E6,$F9,$3F,$89,$F1
	.dc.b	$3F,$15,$E0,$CC,$7E,$FB,$F8,$DE,$FE,$11,$F5,$71,$FF,$02,$3E,$43
	.dc.b	$1F,$05,$A5,$1A,$7F,$C3,$FE,$3B,$81,$DF,$BA,$D6,$BE,$8C,$30,$7E
	.dc.b	$77,$B7,$F3,$DD,$93,$2F,$C9,$BF,$A1,$E9,$E5,$F1,$F2,$7C,$87,$89
	.dc.b	$FD,$37,$36,$63,$E6,$C5,$C7,$F8,$CE,$3F,$1F,$8B,$FD,$A8,$F9,$1A
	.dc.b	$FC,$2D,$BF,$C4,$8D,$3F,$C4,$11,$F5,$7D,$1F,$61,$30,$F8,$D5,$FE
	.dc.b	$2B,$8C,$47,$DE,$6D,$F8,$BE,$0B,$A7,$C4,$CA,$E3,$8F,$F8,$81,$F9
	.dc.b	$6B,$E0,$BE,$55,$49,$F8,$3C,$08,$F3,$7C,$17,$9F,$E0,$B5,$5F,$8E
	.dc.b	$9C,$3D,$BE,$0B,$D8,$47,$C3,$F0,$5F,$5F,$39,$1D,$9F,$85,$F0,$5F
	.dc.b	$0F,$E0,$BE,$27,$C1,$4A,$F4,$C7,$C7,$F8,$29,$1E,$C7,$E5,$7C,$17
	.dc.b	$E0,$F7,$93,$EA,$7E,$77,$C1,$7C,$FF,$82,$FA,$1F,$05,$2B,$E2,$C7
	.dc.b	$D3,$F8,$29,$1F,$21,$FA,$DF,$05,$F8,$FE,$39,$3F,$19,$FB,$9F,$99
	.dc.b	$FE,$C6,$3F,$2E,$FF,$82,$3A,$BF,$8A,$75,$23,$FC,$0D,$D3,$FD,$81
	.dc.b	$B3,$A7,$85,$2B,$87,$C1,$F6,$24,$7E,$12,$5F,$07,$FA,$3E,$99,$3F
	.dc.b	$69,$F2,$FC,$1F,$9B,$E0,$FC,$FF,$07,$2B,$94,$7A,$BE,$0E,$47,$57
	.dc.b	$DB,$F0,$7F,$AF,$ED,$93,$F8,$67,$E0,$FC,$1F,$C2,$F8,$3F,$87,$F0
	.dc.b	$72,$BC,$D1,$F1,$BD,$79,$1E,$97,$E4,$FC,$17,$EE,$FC,$29,$3F,$92
	.dc.b	$7E,$6F,$C1,$7C,$EF,$82,$F9,$FE,$E4,$AF,$85,$1F,$4B,$DC,$FA,$7F
	.dc.b	$05,$27,$E3,$47,$D6,$F8,$29,$1F,$25,$FB,$3F,$05,$FB,$5F,$C2,$FF
	.dc.b	$D8,$5D,$9F,$61,$FF,$5A,$11,$D7,$C3,$CB,$FD,$AC,$FF,$62,$9E,$1F
	.dc.b	$0F,$FC,$57,$CB,$95,$FA,$C7,$B7,$C2,$77,$F8,$4F,$07,$C2,$4A,$E7
	.dc.b	$1E,$3F,$84,$91,$D9,$F3,$7C,$27,$F2,$1F,$3E,$4F,$EE,$1F,$57,$C2
	.dc.b	$7A,$FE,$13,$D9,$F0,$92,$BC,$F1,$EF,$F8,$49,$1E,$A7,$E1,$7C,$1F
	.dc.b	$F2,$DF,$56,$4F,$F0,$47,$E3,$7C,$1F,$C7,$F8,$3F,$91,$F0,$52,$BE
	.dc.b	$1C,$7C,$BF,$81,$91,$F1,$9F,$9D,$F0,$5F,$CE,$7D,$A9,$3F,$C4,$1F
	.dc.b	$A5,$F0,$5F,$4F,$E0,$BF,$94,$7E,$67,$FC,$B6,$BF,$48,$DF,$51,$D8
	.dc.b	$AF,$E9,$FF,$08,$D9,$EC,$61,$B8,$FF,$51,$9D,$4F,$0F,$D4,$BF,$75
	.dc.b	$EC,$7F,$62,$F7,$FF,$A7,$12,$FF,$99,$E8,$F8,$4F,$2F,$C1,$79,$BE
	.dc.b	$0A,$56,$A3,$A7,$C5,$48,$F0,$65,$FB,$ED,$0F,$F0,$28,$FE,$CD,$F0
	.dc.b	$5F,$D7,$48,$F8,$3D,$BF,$87,$EB,$7C,$4F,$46,$4F,$92,$3D,$5E,$F4
	.dc.b	$8E,$EF,$B7,$DD,$FE,$D1,$F0,$32,$7F,$9F,$BF,$3B,$BB,$E7,$F1,$7C
	.dc.b	$3F,$7A,$57,$C0,$8F,$8D,$EE,$C8,$F5,$BF,$27,$DE,$FE,$E9,$F0,$52
	.dc.b	$7F,$A5,$BF,$68,$F9,$DE,$FF,$CF,$F7,$E4,$FC,$48,$FA,$5E,$FC,$8F
	.dc.b	$8E,$FD,$5F,$7B,$FB,$57,$C4,$7C,$7F,$04,$D2,$70,$FD,$BF,$80,$FB
	.dc.b	$9F,$01,$31,$FA,$31,$F7,$FE,$03,$F0,$1F,$01,$27,$EA,$C7,$6D,$24
	.dc.b	$7F,$72,$C5,$1E,$2D,$7F,$11,$EF,$FE,$27,$E0,$25,$FD,$E8,$FC,$67
	.dc.b	$BF,$F8,$DF,$80,$93,$F8,$28,$F5,$E5,$23,$B3,$EE,$CF,$F2,$9F,$01
	.dc.b	$F9,$5F,$80,$95,$F8,$A8,$FC,$FF,$EA,$7F,$C6,$23,$FC,$4C,$FC,$47
	.dc.b	$F3,$CD,$5E,$47,$FA,$F7,$C3,$4F,$1E,$4F,$97,$7F,$38,$97,$F0,$70
	.dc.b	$EF,$F0,$32,$3A,$BE,$2F,$80,$F1,$FB,$FE,$4F,$7A,$57,$8A,$3F,$C4
	.dc.b	$5F,$9A,$9A,$48,$FF,$0F,$C2,$7D,$BF,$01,$EE,$F8,$09,$5E,$C8,$F8
	.dc.b	$3F,$01,$F0,$BD,$F9,$3F,$02,$3B,$FB,$F2,$3A,$BE,$2F,$7F,$E4,$92
	.dc.b	$3E,$3B,$F3,$23,$E6,$FB,$F2,$7E,$5B,$F4,$3D,$FF,$A3,$EB,$C9,$F9
	.dc.b	$EF,$D4,$F7,$BE,$AF,$BF,$F5,$BD,$E9,$5F,$52,$3E,$CF,$BF,$23,$EB
	.dc.b	$BF,$73,$DE,$FB,$BE,$EF,$F1,$CF,$D5,$3F,$91,$76,$1B,$D9,$FC,$87
	.dc.b	$0F,$E3,$91,$FE,$51,$7F,$CD,$0D,$9E,$1F,$13,$CF,$E2,$BB,$57,$E1
	.dc.b	$B1,$AF,$E1,$BE,$17,$FD,$59,$7D,$FD,$FF,$1C,$F6,$F8,$BF,$21,$87
	.dc.b	$81,$E0,$C7,$D0,$47,$CB,$8F,$AF,$DF,$F6,$7C,$04,$9F,$4C,$7B,$FE
	.dc.b	$2B,$E0,$7C,$4C,$9F,$6C,$7C,$3F,$7F,$E2,$7B,$D2,$7E,$0C,$7C,$7F
	.dc.b	$7E,$47,$C4,$7E,$57,$C4,$7C,$BF,$88,$93,$F2,$1F,$9D,$EF,$FC,$FF
	.dc.b	$88,$FA,$1E,$FC,$AF,$9B,$1F,$4F,$E0,$E4,$7D,$07,$EB,$7C,$17,$D7
	.dc.b	$F7,$E4,$FD,$47,$ED,$7C,$57,$DB,$F8,$89,$3F,$61,$FE,$B3,$EA,$7F
	.dc.b	$52,$EA,$FA,$3D,$8C,$7A,$DF,$DE,$77,$F0,$FD,$A3,$E8,$F6,$7E,$17
	.dc.b	$3C,$7A,$BF,$BF,$C1,$A7,$82,$F7,$1F,$D9,$1B,$A7,$89,$F1,$FC,$97
	.dc.b	$93,$E0,$64,$7F,$69,$DB,$FA,$C6,$67,$FC,$A7,$19,$F8,$6F,$59,$EC
	.dc.b	$F8,$7F,$6F,$C3,$FB,$BE,$1E,$57,$38,$EE,$F5,$CA,$BE,$CC,$3F,$98
	.dc.b	$BE,$1B,$EA,$F7,$30,$73,$3C,$D7,$D9,$E5,$7F,$3B,$E0,$A4,$7A,$DF
	.dc.b	$A3,$F0,$5E,$78,$E8,$78,$72,$C3,$C8,$EA,$43,$07,$9F,$E1,$FE,$D9
	.dc.b	$1F,$91,$F0,$FF,$7B,$E1,$BE,$FF,$C3,$C9,$F9,$91,$F8,$2F,$82,$91
	.dc.b	$F3,$DF,$C2,$ED,$F8,$6B,$93,$D0,$FA,$9F,$0B,$D5,$FA,$D8,$75,$7E
	.dc.b	$56,$1D,$1F,$99,$85,$E6,$FE,$89,$8F,$DB,$9F,$BD,$F0,$DF,$94,$23
	.dc.b	$F8,$1F,$85,$FC,$BF,$C3,$7E,$63,$E1,$64,$FE,$16,$3F,$37,$F0,$B2
	.dc.b	$3F,$10,$FF,$8D,$BF,$55,$FE,$B1,$D1,$F8,$9C,$3F,$42,$4A,$FF,$1B
	.dc.b	$4F,$F5,$9C,$FC,$B8,$76,$9F,$C2,$63,$E2,$7A,$90,$FC,$46,$3D,$FC
	.dc.b	$0F,$07,$C3,$49,$EA,$F8,$FE,$1B,$C9,$F0,$D2,$7C,$2F,$9F,$E1,$7D
	.dc.b	$1F,$09,$E9,$F8,$59,$5E,$68,$E4,$F4,$21,$FB,$4D,$98,$75,$8C,$61
	.dc.b	$76,$FE,$E9,$5F,$D8,$67,$A1,$F0,$B0,$EE,$F3,$C3,$E5,$7C,$27,$82
	.dc.b	$3B,$63,$87,$D2,$43,$B9,$F3,$32,$F5,$BD,$48,$75,$3E,$8F,$C2,$FD
	.dc.b	$72,$3F,$53,$E1,$7E,$D7,$C2,$FD,$BF,$85,$93,$F6,$23,$EF,$7C,$2F
	.dc.b	$DF,$F8,$59,$3F,$72,$3F,$05,$72,$3A,$1F,$80,$F8,$5E,$AF,$E0,$B0
	.dc.b	$EA,$FC,$BC,$3A,$BF,$86,$C3,$AB,$F3,$F0,$FC,$77,$C2,$48,$FC,$5B
	.dc.b	$F9,$1F,$85,$FC,$97,$C2,$C9,$FC,$7B,$F9,$5F,$83,$FC,$B7,$C2,$FF
	.dc.b	$A3,$3F,$56,$FE,$E2,$7F,$72,$E5,$FE,$E1,$7F,$77,$0E,$8F,$E6,$1E
	.dc.b	$1F,$07,$D6,$3F,$05,$8D,$E7,$FB,$F9,$76,$9F,$F4,$A6,$3D,$BE,$13
	.dc.b	$11,$FC,$F2,$1F,$09,$E3,$F8,$59,$3C,$23,$CD,$F0,$92,$39,$BE,$9F
	.dc.b	$85,$F5,$7C,$27,$AF,$E1,$65,$78,$23,$DD,$F0,$9E,$FC,$24,$F9,$23
	.dc.b	$A3,$E3,$C2,$EB,$FE,$4B,$CD,$97,$6A,$F5,$E5,$D9,$F5,$65,$F2,$FE
	.dc.b	$16,$47,$C0,$7E,$77,$C2,$F4,$F1,$3F,$C2,$38,$B3,$C5,$BC,$94,$F8
	.dc.b	$23,$D4,$78,$08,$78,$0F,$9B,$F0,$BF,$6C,$8F,$D0,$F8,$5F,$BD,$F0
	.dc.b	$BF,$7F,$E1,$24,$FD,$48,$FC,$17,$C2,$FE,$0F,$E1,$64,$FD,$88,$FC
	.dc.b	$37,$C2,$FE,$1F,$09,$3F,$72,$3A,$BF,$7B,$0E,$AF,$DD,$C3,$AB,$F8
	.dc.b	$2C,$3A,$BF,$81,$C3,$F2,$5F,$09,$F9,$3F,$84,$93,$F8,$88,$FC,$B7
	.dc.b	$C1,$FF,$B8,$BF,$57,$FF,$05,$4F,$C7,$6E,$FF,$22,$75,$F0,$D2,$BF
	.dc.b	$DC,$95,$FE,$D5,$3F,$BE,$8F,$85,$F0,$C7,$73,$6F,$81,$E2,$78,$4C
	.dc.b	$1F,$E1,$48,$E9,$3D,$E3,$1F,$0B,$E7,$F8,$59,$1E,$27,$D5,$F0,$3E
	.dc.b	$BF,$84,$93,$E6,$7D,$DF,$0B,$EF,$F8,$59,$3E,$A7,$E1,$7C,$2F,$28
	.dc.b	$F6,$DD,$7C,$9F,$0B,$D1,$F3,$61,$7C,$1F,$F5,$36,$AC,$BA,$D7,$AF
	.dc.b	$2B,$5A,$CB,$E8,$7C,$2F,$D1,$CA,$5F,$CC,$8E,$9D,$BF,$E6,$CC,$99
	.dc.b	$E2,$BE,$B6,$7D,$2B,$59,$D9,$FB,$59,$7D,$DF,$83,$FB,$DF,$09,$27
	.dc.b	$EC,$C7,$E0,$7E,$13,$F0,$5F,$09,$27,$EE,$C7,$E1,$7E,$16,$47,$E0
	.dc.b	$1F,$C4,$7C,$2F,$E2,$7E,$16,$4F,$E1,$1F,$C6,$7C,$2F,$58,$FC,$3E
	.dc.b	$1D,$5F,$BD,$87,$47,$F0,$18,$5E,$3F,$97,$8F,$E5,$7E,$17,$F2,$DF
	.dc.b	$0B,$2F,$F2,$30,$65,$7F,$B5,$FB,$FF,$C0,$E4,$E1,$D4,$FF,$98,$23
	.dc.b	$FC,$0D,$1F,$E9,$85,$F0,$BA,$AE,$5F,$03,$6B,$4D,$BF,$E1,$97,$C1
	.dc.b	$F0,$CF,$3F,$85,$D7,$80,$AF,$F6,$D5,$68,$E3,$3A,$8F,$EF,$51,$E5
	.dc.b	$AC,$11,$E1,$5C,$FE,$2B,$86,$5C,$71,$95,$AC,$3C,$8F,$BF,$E1,$7B
	.dc.b	$7C,$34,$8D,$47,$87,$E1,$A4,$71,$7C,$9F,$0B,$E5,$F8,$59,$3D,$5F
	.dc.b	$47,$C1,$FA,$7E,$13,$D5,$F0,$72,$BC,$51,$D1,$FA,$18,$5E,$2F,$AD
	.dc.b	$8F,$49,$FA,$98,$D9,$FB,$58,$FC,$5F,$84,$95,$ED,$7E,$47,$C2,$F3
	.dc.b	$E3,$FB,$59,$5A,$FA,$B9,$73,$9F,$BD,$95,$9F,$B3,$8F,$D1,$F8,$5F
	.dc.b	$A5,$F0,$72,$7E,$54,$7D,$5F,$82,$FA,$DF,$05,$27,$E7,$47,$D9,$F8
	.dc.b	$5F,$BF,$FA,$C7,$FB,$C0,$FF,$78,$F2,$FF,$80,$38,$5A,$FF,$B9,$A7
	.dc.b	$FD,$DC,$7F,$76,$C3,$EE,$B5,$FF,$5A,$BF,$EF,$3A,$61,$E1,$9F,$CD
	.dc.b	$63,$E5,$7F,$01,$8F,$8B,$E1,$24,$60,$EB,$F0,$9E,$6F,$76,$46,$0F
	.dc.b	$0F,$C1,$7A,$BE,$06,$46,$0F,$2F,$C1,$7B,$88,$FA,$3E,$0F,$E0,$FC
	.dc.b	$1F,$C2,$F8,$39,$3E,$C8,$F8,$B8,$74,$8F,$6E,$18,$7A,$90,$F8,$38
	.dc.b	$7C,$CC,$3E,$6F,$C1,$7C,$EF,$82,$95,$F2,$23,$E8,$FC,$04,$8F,$96
	.dc.b	$FD,$4F,$81,$FA,$BC,$72,$7E,$7B,$F6,$3E,$07,$EC,$FC,$17,$69,$E8
	.dc.b	$7D,$47,$A3,$F5,$2E,$F4,$7D,$C8,$FF,$07,$73,$B1,$ED,$1F,$6B,$1B
	.dc.b	$5B,$F2,$16,$CB,$F1,$3F,$03,$E2,$C3,$C0,$7E,$11,$EF,$D5,$F8,$9C
	.dc.b	$EC,$F8,$33,$EB,$3F,$8D,$CE,$CF,$93,$0F,$CA,$FC,$17,$E5,$BE,$0A
	.dc.b	$4F,$E4,$23,$F8,$BF,$EB,$3F,$E0,$01,$F0,$58,$33,$3A,$F8,$2D,$FC
	.dc.b	$1C,$8F,$E2,$B5,$FE,$FE,$3F,$DF,$AB,$87,$E0,$46,$1B,$F0,$7E,$0F
	.dc.b	$83,$EB,$5C,$B1,$F0,$3D,$0F,$01,$DB,$0F,$03,$D7,$6F,$E8,$5C,$6C
	.dc.b	$BC,$5F,$01,$EC,$F8,$09,$7E,$68,$F7,$FC,$07,$C0,$F8,$09,$3E,$A8
	.dc.b	$F8,$7F,$07,$23,$DA,$FC,$6F,$83,$F8,$FF,$03,$27,$E0,$BF,$2B,$E0
	.dc.b	$BA,$C7,$C4,$C3,$AB,$E7,$3F,$CA,$99,$B1,$ED,$19,$D3,$D5,$96,$38
	.dc.b	$BE,$4E,$7F,$5B,$DF,$97,$F4,$1F,$B3,$F0,$1F,$6B,$E0,$24,$FD,$47
	.dc.b	$EE,$FC,$07,$DE,$F8,$29,$3F,$61,$FC,$0F,$C1,$7E,$0B,$E0,$79,$CF
	.dc.b	$DC,$C3,$0F,$32,$1E,$7D,$D8,$F6,$8E,$84,$35,$87,$67,$F0,$FF,$05
	.dc.b	$F9,$02,$3F,$D7,$BF,$5A,$FF,$7F,$DF,$48,$EF,$B7,$E2,$B0,$ED,$1D
	.dc.b	$0E,$D1,$F8,$EC,$3B,$47,$CB,$BF,$EF,$EF,$FA,$0D,$BC,$06,$9F,$DA
	.dc.b	$2B,$F6,$CF,$FB,$CA,$7C,$9F,$01,$FD,$F2,$31,$1D,$7E,$0F,$1F,$13
	.dc.b	$23,$D4,$EF,$E0,$78,$7C,$14,$9F,$73,$CF,$DE,$E9,$8C,$9E,$E7,$C3
	.dc.b	$C3,$A3,$EB,$C2,$E5,$E6,$C7,$C9,$EF,$4A,$F9,$4F,$9F,$DE,$F4,$7B
	.dc.b	$D2,$7E,$73,$EB,$F8,$1F,$67,$C1,$49,$FA,$4F,$BF,$E0,$BE,$07,$C1
	.dc.b	$73,$9F,$AD,$86,$1E,$64,$3E,$86,$1D,$5E,$64,$3E,$A5,$F5,$7F,$01
	.dc.b	$F0,$3F,$37,$E0,$A4,$60,$FC,$27,$C1,$7D,$12,$3F,$87,$F8,$2F,$A9
	.dc.b	$F0,$52,$3F,$14,$FF,$8C,$B3,$7C,$15,$F8,$8C,$7C,$44,$BF,$F1,$6D
	.dc.b	$27,$AC,$7E,$27,$1B,$C7,$F8,$4C,$BB,$4F,$E3,$72,$B3,$FC,$F7,$0E
	.dc.b	$F3,$D3,$E1,$3C,$44,$7B,$FC,$27,$4F,$3B,$F8,$CF,$67,$F7,$E9,$F7
	.dc.b	$9E,$D3,$D6,$43,$C0,$7A,$36,$F7,$7C,$17,$BF,$E0,$A4,$FB,$23,$E1
	.dc.b	$7C,$17,$C3,$F8,$29,$3F,$02,$3E,$37,$C1,$FC,$7F,$84,$93,$F1,$23
	.dc.b	$E5,$7C,$1C,$8F,$8E,$FC,$DF,$84,$EB,$1F,$2B,$0E,$AF,$F1,$BE,$06
	.dc.b	$3D,$A3,$E7,$E3,$D9,$D6,$3F,$5F,$E0,$64,$7D,$47,$ED,$7C,$0F,$DB
	.dc.b	$F8,$19,$3F,$61,$FB,$DF,$03,$F7,$FE,$12,$4F,$DC,$7F,$05,$F0,$7F
	.dc.b	$83,$F8,$2F,$C2,$7C,$1C,$AF,$C0,$C7,$47,$D1,$85,$E9,$FB,$98,$F4
	.dc.b	$9F,$66,$36,$7F,$01,$8F,$FA,$03,$F5,$BF,$F7,$49,$2B,$1F,$03,$9F
	.dc.b	$84,$D7,$C1,$49,$FF,$3E,$D2,$79,$7C,$27,$3F,$84,$95,$A8,$ED,$F0
	.dc.b	$72,$38,$BE,$1F,$84,$ED,$1D,$31,$EC,$FF,$4D,$C7,$B3,$CC,$F0,$E1
	.dc.b	$D9,$D6,$1E,$BF,$81,$F6,$7C,$14,$9F,$3C,$76,$7A,$61,$DB,$6F,$F0
	.dc.b	$F6,$3D,$A7,$99,$D8,$86,$0F,$27,$C1,$7C,$9F,$82,$91,$F1,$23,$E6
	.dc.b	$7C,$24,$8F,$8E,$FC,$FF,$84,$FA,$1F,$09,$27,$E5,$BF,$4F,$E0,$7E
	.dc.b	$A7,$BF,$27,$E7,$BF,$5F,$E1,$3A,$47,$C5,$C2,$E9,$EA,$C7,$A4,$FC
	.dc.b	$9C,$6C,$FB,$B1,$FC,$07,$C0,$7E,$07,$E1,$25,$FD,$C8,$FC,$27,$C2
	.dc.b	$48,$FB,$EF,$E1,$FE,$13,$FD,$8F,$FA,$E7,$F9,$2E,$7F,$C0,$4D,$C7
	.dc.b	$FC,$90,$DF,$3B,$FC,$3E,$FF,$5A,$3F,$AC,$38,$F8,$7E,$67,$4F,$82
	.dc.b	$F0,$5F,$E0,$71,$B8,$7D,$6C,$7C,$38,$4A,$FF,$67,$E3,$DD,$7C,$2A
	.dc.b	$3C,$1F,$07,$E8,$F8,$39,$5C,$23,$D7,$F0,$72,$39,$BE,$EF,$85,$F7
	.dc.b	$FC,$2F,$C0,$F8,$59,$5E,$08,$F8,$7F,$07,$F1,$3E,$0E,$4F,$92,$3A
	.dc.b	$3F,$1B,$0B,$BF,$C6,$C7,$AC,$FA,$B1,$EA,$FC,$CC,$7E,$7F,$C2,$C8
	.dc.b	$F7,$BF,$4B,$E1,$39,$F0,$7C,$EC,$AD,$7C,$2D,$D9,$76,$8F,$AF,$97
	.dc.b	$6C,$FE,$BE,$7F,$73,$E1,$7E,$EF,$AB,$2F,$E6,$C7,$E0,$3E,$16,$47
	.dc.b	$D0,$7F,$07,$F0,$BF,$84,$F8,$3F,$C2,$FC,$1C,$AF,$AB,$1F,$88,$F8
	.dc.b	$34,$4C,$FF,$D1,$DF,$AE,$FF,$52,$8F,$F2,$63,$63,$F1,$73,$E0,$9F
	.dc.b	$C7,$E3,$67,$F1,$F8,$F8,$67,$F0,$78,$F8,$5F,$CA,$E3,$E0,$7A,$FC
	.dc.b	$2D,$BF,$0F,$C3,$5E,$3F,$F3,$AE,$9F,$ED,$D9,$F2,$D6,$6F,$F1,$B2
	.dc.b	$F8,$6C,$75,$A3,$13,$ED,$C3,$9D,$6A,$5F,$0D,$7E,$0E,$5E,$CE,$AE
	.dc.b	$FE,$13,$87,$C0,$C9,$D6,$5D,$E3,$A7,$C1,$75,$CA,$56,$B2,$F1,$BF
	.dc.b	$33,$2E,$EF,$CE,$CA,$EB,$F4,$B3,$F3,$7C,$0F,$9F,$E0,$A5,$F3,$8F
	.dc.b	$57,$C2,$48,$EC,$FB,$7E,$13,$DD,$F0,$92,$7C,$4F,$C1,$F8,$4F,$85
	.dc.b	$F0,$5F,$0F,$E0,$A5,$79,$E3,$A3,$F8,$2C,$2E,$9F,$86,$C7,$A4,$FE
	.dc.b	$1B,$1B,$3F,$8A,$C7,$E7,$7C,$17,$CF,$F8,$29,$7F,$0E,$39,$F3,$7D
	.dc.b	$EC,$AC,$FD,$CC,$B9,$CF,$E0,$B2,$B3,$F8,$0C,$7E,$D7,$C1,$C8,$F9
	.dc.b	$CF,$DD,$F8,$4F,$C1,$FB,$1F,$8F,$9F,$F5,$33,$FE,$D6,$7E,$CB,$5C
	.dc.b	$9C,$3E,$0F,$8F,$BB,$23,$71,$D3,$E0,$24,$72,$7B,$FC,$0F,$82,$3C
	.dc.b	$3F,$09,$1E,$E7,$53,$C3,$7D,$5F,$2D,$FA,$3E,$13,$D3,$F0,$92,$7C
	.dc.b	$F1,$EC,$F8,$39,$1E,$A7,$DF,$F0,$9F,$03,$E0,$BE,$0F,$C1,$CA,$F7
	.dc.b	$C7,$C4,$F8,$2F,$8B,$F0,$52,$7E,$1C,$74,$7E,$3E,$17,$A9,$CD,$FE
	.dc.b	$E8,$E2,$67,$DE,$FE,$5E,$7D,$DF,$9F,$9F,$D2,$F8,$2F,$A7,$F0,$52
	.dc.b	$7E,$8C,$7D,$6C,$24,$7D,$47,$EC,$FC,$1F,$DA,$F8,$2F,$B7,$F0,$72
	.dc.b	$BE,$CC,$7D,$EF,$81,$FB,$FF,$03,$27,$EE,$C7,$47,$F0,$38,$5D,$FD
	.dc.b	$71,$FD,$5A,$18,$F7,$8F,$C2,$E3,$DD,$FC,$4E,$3F,$C3,$BF,$5E,$FE
	.dc.b	$BD,$E7,$1A,$4F,$8A,$ED,$C5,$F1,$F2,$B8,$77,$CB,$B3,$F9,$4C,$BB
	.dc.b	$67,$E4,$CB,$FD,$CB,$9F,$7F,$85,$97,$FC,$43,$4F,$EF,$2F,$F6,$E7
	.dc.b	$FD,$9B,$7C,$57,$C1,$F9,$FE,$12,$57,$18,$F5,$7C,$07,$AF,$DF,$93
	.dc.b	$D6,$3D,$DE,$F4,$8F,$03,$F0,$7D,$FF,$85,$7D,$25,$97,$87,$1C,$C7
	.dc.b	$33,$D1,$8F,$67,$7C,$3F,$D4,$B6,$F9,$59,$FC,$DF,$83,$F9,$DF,$09
	.dc.b	$2F,$E0,$47,$D1,$F8,$4F,$A5,$F0,$92,$7E,$24,$7D,$5F,$82,$FA,$DF
	.dc.b	$07,$27,$E4,$47,$D9,$F8,$2F,$B5,$F0,$52,$7E,$64,$75,$7E,$76,$1D
	.dc.b	$5F,$39,$F3,$F0,$EC,$FD,$3C,$3B,$3A,$D3,$F0,$98,$FE,$1F,$E0,$7F
	.dc.b	$11,$F0,$52,$BE,$D4,$7E,$2F,$E0,$FF,$AC,$FE,$BF,$FD,$BF,$87,$E2
	.dc.b	$36,$BE,$F4,$EB,$E0,$F7,$F0,$92,$7F,$A9,$D7,$FB,$4D,$BF,$05,$D3
	.dc.b	$E0,$64,$60,$FC,$56,$1E,$0F,$88,$91,$83,$9E,$3D,$5F,$66,$1F,$DC
	.dc.b	$F2,$F2,$CF,$A3,$E0,$A5,$78,$9F,$5F,$C1,$7B,$63,$CB,$87,$67,$E3
	.dc.b	$77,$B2,$F6,$47,$AB,$5F,$07,$67,$E8,$F4,$F6,$4F,$8B,$E1,$7E,$3F
	.dc.b	$C2,$C8,$F8,$31,$F2,$BE,$0E,$47,$C4,$7E,$6F,$C2,$7C,$EF,$80,$F9
	.dc.b	$FF,$01,$2B,$E4,$C7,$D2,$F8,$09,$1F,$31,$FA,$BF,$01,$D2,$3E,$1E
	.dc.b	$17,$4F,$36,$3D,$67,$E9,$E3,$D5,$F9,$38,$FD,$EF,$83,$FB,$FF,$09
	.dc.b	$27,$EC,$C7,$E0,$BE,$0E,$47,$DC,$7F,$0B,$F0,$9F,$86,$F8,$4F,$C3
	.dc.b	$FC,$24,$AF,$C0,$C7,$F8,$8B,$F6,$0F,$EC,$1A,$FF,$0B,$7F,$AF,$37
	.dc.b	$E0,$F5,$5D,$8F,$BB,$8D,$9F,$A1,$8F,$79,$FC,$6E,$3D,$DF,$C1,$E3
	.dc.b	$DF,$E0,$BC,$1F,$07,$27,$FC,$4B,$9F,$FA,$EF,$1F,$17,$C2,$F9,$7E
	.dc.b	$16,$56,$A3,$D1,$F0,$92,$38,$BE,$BF,$85,$F6,$7C,$1F,$B7,$E0,$E5
	.dc.b	$76,$8F,$81,$7D,$23,$F6,$59,$31,$CC,$74,$21,$AC,$7B,$3E,$8F,$82
	.dc.b	$F9,$5F,$07,$23,$07,$9C,$F6,$D7,$7D,$3D,$B1,$D1,$FA,$17,$EB,$8F
	.dc.b	$1E,$9F,$57,$E1,$24,$7C,$67,$EC,$7C,$2F,$D9,$F8,$49,$3F,$29,$FB
	.dc.b	$9F,$0B,$F7,$7E,$0B,$EF,$7C,$14,$AF,$9F,$1F,$81,$F8,$1F,$C1,$7C
	.dc.b	$14,$9F,$A7,$1D,$1F,$8B,$85,$D7,$F0,$D8,$F5,$9F,$B3,$8F,$57,$E5
	.dc.b	$E3,$F8,$DF,$84,$91,$F7,$9F,$C8,$7C,$2F,$F9,$CB,$BF,$FB,$29,$3D
	.dc.b	$0C,$7C,$2E,$7E,$17,$5F,$11,$31,$FF,$35,$DF,$F5,$CF,$64,$FF,$5A
	.dc.b	$8C,$FC,$1F,$53,$B7,$C2,$77,$CA,$4F,$27,$BB,$D3,$1E,$CF,$F3,$5B
	.dc.b	$FE,$CF,$3F,$D4,$5F,$00,$F8,$3F,$57,$C2,$4A,$F2,$C7,$B7,$E1,$A4
	.dc.b	$7A,$1F,$81,$F0,$DF,$07,$E1,$64,$FB,$1F,$89,$F0,$DF,$17,$E0,$E4
	.dc.b	$FC,$07,$E4,$7C,$27,$48,$F3,$61,$7A,$3F,$C0,$9A,$32,$EB,$5E,$BC
	.dc.b	$AD,$6B,$2F,$A5,$F0,$7F,$4F,$E1,$25,$FC,$E8,$E9,$BF,$C1,$CE,$CF
	.dc.b	$FB,$E3,$3E,$CF,$A4,$87,$73,$EB,$E1,$F7,$FE,$16,$47,$DA,$7F,$05
	.dc.b	$F0,$DF,$83,$F8,$59,$3F,$79,$FC,$37,$C3,$7E,$1F,$DF,$93,$F8,$27
	.dc.b	$FD,$77,$FB,$0F,$F6,$1E,$93,$FD,$0B,$83,$FC,$4C,$DD,$BF,$C1,$4D
	.dc.b	$B5,$70,$F8,$0F,$0C,$7E,$07,$2B,$87,$CC,$CB,$C3,$3F,$85,$CA,$CF
	.dc.b	$6F,$82,$F1,$FC,$3F,$93,$E1,$E5,$FF,$B0,$B4,$FF,$5E,$DF,$93,$E1
	.dc.b	$BD,$24,$9D,$7C,$37,$B3,$E0,$E4,$71,$7D,$FF,$09,$F0,$3E,$0B,$E0
	.dc.b	$FC,$14,$AE,$D1,$CD,$FF,$07,$64,$C7,$19,$7C,$4C,$BA,$D6,$B2,$B3
	.dc.b	$F2,$32,$F9,$BF,$03,$F3,$BE,$0A,$5F,$AE,$3E,$8F,$C2,$48,$F7,$3F
	.dc.b	$53,$E1,$3E,$AF,$C2,$7D,$6E,$C9,$5F,$0E,$3E,$CF,$C1,$FD,$AF,$83
	.dc.b	$93,$F1,$E3,$A3,$EA,$C2,$E9,$F7,$31,$EB,$3F,$37,$1E,$AF,$C0,$C7
	.dc.b	$F0,$BF,$05,$F8,$6F,$83,$93,$F4,$E3,$9F,$6F,$E1,$32,$B3,$F8,$4E
	.dc.b	$36,$5D,$A3,$E5,$E5,$DB,$3F,$C6,$E6,$9F,$F7,$FF,$EC,$5F,$DB,$32
	.dc.b	$FF,$2C,$3F,$DB,$1B,$F0,$BB,$9F,$F2,$14,$ED,$FC,$9B,$FD,$8E,$78
	.dc.b	$FC,$26,$F0,$FF,$40,$4A,$31,$3D,$BE,$16,$DF,$2F,$83,$BE,$13,$0F
	.dc.b	$EB,$D3,$E6,$F8,$0C,$C2,$AF,$C0,$4A,$F5,$BA,$F8,$0D,$E3,$C3,$39
	.dc.b	$5F,$02,$39,$DF,$4C,$64,$FC,$48,$EF,$F0,$9E,$0F,$85,$93,$F2,$23
	.dc.b	$C7,$F0,$92,$3E,$5B,$E6,$F8,$4F,$3F,$C2,$49,$F9,$EF,$AB,$E1,$3D
	.dc.b	$7F,$05,$EC,$F8,$29,$5F,$52,$3A,$3F,$43,$0B,$C9,$F3,$31,$E9,$3F
	.dc.b	$53,$1B,$3F,$43,$1F,$8F,$F0,$32,$BE,$FB,$F2,$BE,$0B,$E5,$FC,$1F
	.dc.b	$CC,$F8,$49,$5F,$84,$8F,$9F,$F0,$92,$3F,$0E,$FD,$2F,$84,$FA,$FF
	.dc.b	$B1,$FF,$B2,$07,$C2,$E0,$CF,$67,$6B,$FC,$7E,$3D,$9F,$C5,$E3,$D5
	.dc.b	$FC,$6E,$37,$5F,$C4,$E5,$DB,$E0,$A5,$7E,$66,$7F,$D9,$99,$FF,$6D
	.dc.b	$3F,$B0,$C7,$E5,$30,$ED,$3F,$92,$C7,$C5,$3D,$8F,$59,$0F,$CE,$E1
	.dc.b	$EC,$F8,$3F,$6F,$C2,$49,$F0,$69,$FE,$BD,$CB,$FB,$D4,$7F,$A9,$5F
	.dc.b	$ED,$CE,$8F,$E6,$EE,$1D,$7C,$1F,$C8,$F8,$4F,$93,$EF,$7C,$AF,$82
	.dc.b	$95,$C7,$C3,$60,$AF,$9F,$F0,$5F,$43,$E0,$25,$EE,$93,$F5,$32,$93
	.dc.b	$97,$C0,$EE,$FB,$BE,$9C,$7C,$6F,$3F,$84,$FB,$A4,$7B,$7C,$27,$E0
	.dc.b	$3E,$0F,$F0,$3F,$09,$27,$C5,$1F,$84,$F8,$49,$1E,$57,$F0,$FF,$09
	.dc.b	$F8,$8F,$83,$93,$E9,$7F,$17,$F0,$7D,$63,$D9,$87,$57,$E7,$61,$D1
	.dc.b	$FA,$18,$5E,$8F,$C1,$E3,$F9,$4F,$83,$95,$F1,$5F,$E1,$B5,$FD,$C4
	.dc.b	$FE,$D9,$C6,$69,$3E,$13,$7F,$07,$2F,$EA,$4F,$F6,$4E,$7F,$F7,$F2
	.dc.b	$F8,$3E,$BF,$03,$DB,$E0,$65,$F0,$8F,$0F,$C4,$C8,$E6,$F9,$32,$ED
	.dc.b	$1F,$EE,$FC,$3B,$3F,$E0,$06,$E7,$E7,$C7,$D7,$F0,$52,$BC,$AF,$BB
	.dc.b	$E0,$BB,$47,$3C,$6F,$03,$5F,$4E,$7E,$08,$C5,$77,$DB,$15,$EE,$CB
	.dc.b	$E4,$FC,$1C,$AF,$86,$FC,$CF,$84,$F9,$BF,$07,$F3,$BE,$12,$57,$C8
	.dc.b	$8F,$A3,$F0,$12,$3E,$5B,$F5,$3D,$FF,$AB,$EF,$49,$F9,$EF,$D8,$F7
	.dc.b	$FA,$C7,$D2,$C3,$AB,$F1,$F0,$E8,$FC,$9C,$2E,$DE,$CE,$3F,$BD,$97
	.dc.b	$E0,$BE,$0F,$F0,$7F,$09,$30,$FB,$B1,$F8,$6F,$83,$FC,$3F,$C2,$49
	.dc.b	$FC,$0C,$7F,$56,$FD,$93,$FB,$16,$26,$93,$E1,$35,$F0,$1B,$F8,$19
	.dc.b	$87,$EF,$6B,$FB,$43,$F7,$70,$EC,$FE,$07,$1F,$13,$B3,$C0,$43,$96
	.dc.b	$1E,$3F,$80,$F2,$7C,$0C,$9E,$D1,$E7,$F8,$39,$1E,$17,$D5,$F0,$7E
	.dc.b	$BF,$82,$93,$E5,$7D,$DF,$07,$EF,$F8,$19,$3E,$97,$E1,$7C,$0F,$48
	.dc.b	$72,$FC,$FD,$D9,$66,$31,$97,$6A,$DF,$77,$F8,$4F,$8F,$E5,$69,$9B
	.dc.b	$E5,$A7,$CF,$F8,$19,$87,$C9,$7E,$97,$C0,$F5,$C7,$96,$9E,$0A,$F7
	.dc.b	$1F,$3F,$3E,$93,$AC,$AC,$FF,$9B,$29,$8F,$DD,$F8,$2F,$BD,$F0,$72
	.dc.b	$BE,$C4,$7E,$07,$E0,$E4,$7D,$B7,$F0,$9F,$07,$F8,$5F,$80,$FC,$37
	.dc.b	$C0,$4A,$FC,$04,$7E,$27,$E0,$BF,$15,$F0,$32,$7F,$09,$1D,$1F,$8A
	.dc.b	$79,$70,$C7,$17,$CD,$C7,$B4,$FE,$2F,$1E,$CF,$0E,$3F,$CA,$65,$FE
	.dc.b	$2C,$FD,$97,$F7,$B3,$97,$FB,$BB,$FD,$3D,$AE,$78,$78,$7C,$1C,$8D
	.dc.b	$3C,$FE,$23,$A7,$C2,$49,$E4,$F7,$F8,$4F,$07,$C0,$49,$EC,$F8,$FE
	.dc.b	$13,$A4,$78,$F0,$BB,$FE,$DF,$3F,$EF,$7A,$9F,$ED,$F1,$9C,$BC,$17
	.dc.b	$EB,$CB,$DD,$F0,$32,$3D,$8F,$C1,$F8,$1F,$85,$F0,$52,$7E,$03,$F1
	.dc.b	$7E,$13,$E3,$7C,$1C,$9F,$88,$FC,$9F,$83,$F9,$5F,$01,$F2,$FE,$06
	.dc.b	$57,$C9,$8E,$8F,$CD,$C2,$EB,$EB,$C7,$A4,$FD,$1C,$6C,$FB,$F1,$FA
	.dc.b	$DE,$EC,$AF,$A8,$FD,$9F,$81,$E7,$CD,$FA,$7E,$1F,$EF,$F5,$EA,$3E
	.dc.b	$D6,$5D,$27,$59,$59,$FF,$64,$D3,$1F,$C2,$7C,$1C,$9F,$C1,$3F,$87
	.dc.b	$F8,$3F,$C4,$7C,$1C,$9F,$C3,$3F,$E7,$6F,$D9,$BF,$DB,$47,$F6,$A2
	.dc.b	$7C,$06,$7E,$03,$5F,$01,$2F,$FC,$39,$5F,$D6,$8F,$ED,$0F,$42,$1F
	.dc.b	$95,$BE,$DF,$13,$DF,$09,$3C,$9E,$CF,$4F,$84,$F2,$7C,$2C,$8C,$1E
	.dc.b	$0F,$85,$F4,$7C,$2C,$8C,$1E,$4F,$84,$F6,$11,$F3,$FC,$27,$BF,$E0
	.dc.b	$64,$7A,$9F,$85,$F0,$5F,$0F,$0E,$93,$FB,$1D,$98,$E6,$3B,$10,$D6
	.dc.b	$3F,$2F,$1F,$99,$F0,$5F,$37,$E0,$A5,$7C,$88,$FA,$1F,$09,$23,$E5
	.dc.b	$BF,$4F,$E1,$3E,$A7,$C1,$C9,$F9,$EF,$D7,$F8,$4F,$B1,$F0,$5F,$67
	.dc.b	$E0,$A5,$7D,$48,$E8,$FA,$70,$BE,$17,$F6,$9D,$D9,$75,$AF,$76,$56
	.dc.b	$B5,$97,$FA,$EB,$D1,$FE,$CA,$7F,$B2,$E9,$E5,$A3,$0D,$8E,$FE,$0F
	.dc.b	$F4,$3C,$99,$E2,$78,$C7,$89,$F6,$90,$F1,$9F,$8E,$C3,$C5,$F0,$9E
	.dc.b	$3F,$84,$93,$FE,$6D,$CB,$FB,$8E,$5E,$3F,$82,$F4,$7C,$1C,$AD,$47
	.dc.b	$AF,$E0,$64,$71,$7D,$DE,$FF,$BF,$DF,$93,$D5,$F8,$5E,$F7,$C3,$BE
	.dc.b	$93,$FC,$D7,$2F,$F6,$EB,$63,$AB,$9C,$2F,$0F,$F8,$66,$98,$FC,$DF
	.dc.b	$83,$97,$EB,$7E,$87,$C1,$FD,$1F,$83,$93,$EF,$7E,$A7,$C1,$FD,$5F
	.dc.b	$81,$FA,$DF,$05,$2B,$E2,$47,$D9,$F8,$09,$1F,$1D,$FB,$99,$F4,$8F
	.dc.b	$56,$17,$4F,$36,$3D,$67,$E7,$63,$D5,$F8,$18,$FE,$17,$DF,$91,$F4
	.dc.b	$DF,$C4,$7C,$07,$E2,$7E,$07,$F1,$5F,$05,$2B,$EC,$41,$FF,$53,$7E
	.dc.b	$CF,$FD,$73,$3F,$F1,$F3,$7B,$7F,$C1,$8D,$9F,$C8,$BF,$D7,$A7,$7F
	.dc.b	$01,$CE,$7F,$B4,$4A,$31,$3A,$9E,$FF,$07,$E5,$7F,$15,$8F,$95,$FB
	.dc.b	$98,$F0,$C7,$C2,$58,$F0,$B7,$A9,$97,$95,$BF,$11,$98,$FE,$BF,$3E
	.dc.b	$AF,$81,$C1,$9F,$82,$F1,$61,$F8,$BC,$BC,$5A,$7E,$1F,$3E,$58,$4A
	.dc.b	$F8,$EF,$5C,$BB,$7C,$27,$7F,$84,$95,$F3,$23,$C5,$F0,$92,$3E,$7B
	.dc.b	$E5,$F8,$4F,$37,$C1,$49,$FA,$6F,$A7,$E0,$3D,$5F,$01,$27,$EB,$BE
	.dc.b	$DF,$7F,$A4,$7D,$0C,$2E,$DF,$6B,$1E,$93,$F5,$31,$B3,$F7,$B1,$F8
	.dc.b	$DF,$07,$2B,$F0,$6F,$C9,$F8,$4F,$95,$F0,$9F,$2F,$E1,$25,$7E,$22
	.dc.b	$3E,$77,$C1,$C8,$FC,$5B,$F4,$7E,$0F,$E9,$7C,$0F,$D7,$FD,$A1,$F0
	.dc.b	$54,$CD,$F3,$3F,$27,$8F,$67,$F1,$78,$F5,$7F,$1B,$8D,$D7,$F2,$F9
	.dc.b	$75,$F8,$29,$5F,$F0,$07,$2F,$F6,$A8,$F8,$4F,$17,$C2,$49,$D3,$E5
	.dc.b	$F8,$8F,$37,$C2,$C9,$E4,$FA,$7E,$17,$D5,$F0,$72,$7B,$3E,$DF,$83
	.dc.b	$E9,$1F,$E4,$9C,$BF,$D9,$E4,$C3,$A4,$67,$0B,$AF,$FA,$B6,$98,$FC
	.dc.b	$7F,$83,$97,$EB,$7E,$57,$C1,$F6,$BE,$1A,$5F,$03,$FD,$05,$8F,$F8
	.dc.b	$13,$0E,$C7,$A8,$87,$97,$1F,$A9,$F0,$BF,$57,$E1,$64,$FC,$98,$FB
	.dc.b	$1F,$09,$F6,$7E,$16,$4F,$CD,$8F,$B9,$F0,$3F,$77,$E0,$E4,$FD,$18
	.dc.b	$FC,$07,$C1,$C8,$FA,$8F,$E0,$FE,$0B,$AC,$7D,$7C,$3A,$BF,$27,$0E
	.dc.b	$AF,$DB,$C3,$AB,$F3,$70,$FC,$67,$C2,$48,$FC,$03,$F8,$FF,$85,$FE
	.dc.b	$0D,$FB,$4F,$FB,$38,$93,$8F,$89,$CF,$C3,$6B,$E1,$A4,$FF,$25,$E5
	.dc.b	$FF,$70,$5F,$F5,$E1,$F0,$7D,$09,$3B,$E5,$FF,$59,$4F,$F5,$4A,$E3
	.dc.b	$3E,$22,$1B,$3F,$BD,$C7,$F7,$A3,$B5,$F8,$A3,$FD,$25,$1E,$1F,$83
	.dc.b	$F5,$FC,$1C,$9E,$5E,$BA,$BF,$D7,$11,$C3,$E1,$7E,$09,$1D,$C7,$FA
	.dc.b	$D9,$BF,$09,$F1,$88,$EB,$E1,$3E,$4F,$C1,$7C,$AF,$82,$93,$CA,$3A
	.dc.b	$3D,$CE,$B8,$7D,$0F,$85,$E7,$1E,$AC,$2E,$DF,$33,$1F,$AB,$F0,$52
	.dc.b	$BC,$AF,$D8,$F8,$3E,$B5,$F4,$B1,$BC,$3F,$3B,$2E,$6F,$D0,$CA,$E3
	.dc.b	$F3,$32,$FC,$0F,$C2,$FE,$0B,$E1,$64,$FC,$28,$FC,$2F,$C2,$FE,$1B
	.dc.b	$E1,$64,$FC,$68,$FC,$4F,$BF,$23,$E4,$BF,$8C,$F8,$0F,$EA,$5E,$17
	.dc.b	$F5,$13,$FB,$26,$9F,$EF,$36,$D7,$F8,$E1,$FE,$AD,$C8,$C3,$BC,$FD
	.dc.b	$EC,$7C,$6F,$62,$1F,$82,$C7,$C0,$FE,$02,$FF,$B7,$37,$E1,$7C,$67
	.dc.b	$93,$E1,$7C,$BF,$07,$E6,$F8,$49,$5C,$63,$D3,$F0,$5E,$AF,$83,$93
	.dc.b	$D6,$3D,$BF,$03,$EE,$F8,$19,$3E,$18,$EA,$F8,$F0,$EA,$FF,$BA,$73
	.dc.b	$FE,$E2,$FC,$1B,$EA,$E7,$0F,$0B,$EB,$F8,$1F,$96,$47,$DD,$F0,$3F
	.dc.b	$3B,$E0,$E4,$7C,$17,$E8,$FC,$1F,$D2,$F8,$39,$3F,$15,$FA,$BF,$07
	.dc.b	$F5,$BE,$02,$4F,$C9,$7E,$CF,$C0,$F4,$8F,$46,$17,$C0,$74,$FD,$BC
	.dc.b	$BA,$D7,$B7,$2B,$5A,$8F,$BD,$97,$E1,$3E,$06,$47,$D7,$7F,$0F,$F0
	.dc.b	$3D,$71,$FC,$36,$56,$BF,$0D,$97,$49,$FC,$56,$56,$7A,$9F,$81,$F8
	.dc.b	$3F,$F1,$17,$ED,$5F,$DB,$23,$F8,$1B,$5F,$A5,$1A,$F8,$39,$1A,$78
	.dc.b	$FC,$1F,$2F,$76,$4F,$27,$AF,$C0,$F6,$F8,$19,$3D,$9F,$0F,$BD,$D2
	.dc.b	$1C,$5F,$0F,$CF,$68,$CF,$53,$98,$FE,$E9,$5D,$C8,$72,$DB,$FC,$AD
	.dc.b	$7C,$B3,$F1,$CF,$BB,$E0,$FE,$09,$1F,$83,$F0,$7F,$13,$E0,$E4,$7C
	.dc.b	$47,$E3,$FC,$1F,$C8,$F8,$1F,$93,$F0,$52,$BE,$4C,$7C,$CF,$80,$91
	.dc.b	$F3,$1F,$9F,$F0,$3D,$23,$CC,$7F,$AD,$36,$63,$9D,$FD,$D9,$77,$AF
	.dc.b	$AB,$97,$77,$8E,$DF,$1B,$3F,$B7,$F0,$1F,$73,$E0,$65,$7D,$C8,$FB
	.dc.b	$FE,$27,$E0,$3E,$0A,$4F,$E0,$23,$F0,$7F,$05,$F8,$4F,$82,$93,$F8
	.dc.b	$48,$FF,$34,$7E,$D7,$FE,$77,$8F,$0B,$FE,$CA,$23,$D8,$FC,$5D,$F6
	.dc.b	$7C,$FC,$0C,$3B,$C7,$E4,$30,$EE,$EB,$0E,$FF,$05,$E0,$F8,$19,$3F
	.dc.b	$D6,$32,$FF,$64,$E3,$CE,$3B,$90,$EF,$8F,$67,$C3,$8F,$6E,$1F,$C5
	.dc.b	$65,$EC,$F8,$2F,$6F,$C4,$4B,$ED,$1F,$03,$E0,$E4,$78,$5F,$87,$F0
	.dc.b	$7F,$13,$DF,$F8,$BE,$FC,$AF,$34,$7C,$8F,$7E,$47,$A5,$F9,$7E,$FF
	.dc.b	$48,$F9,$78,$5D,$3F,$45,$C1,$FE,$D0,$9E,$67,$C1,$C7,$B3,$F4,$B1
	.dc.b	$FA,$BF,$05,$23,$E3,$3F,$63,$E0,$FE,$CF,$C1,$FD,$AF,$83,$95,$F2
	.dc.b	$E3,$EE,$FC,$14,$8F,$9C,$FE,$03,$E0,$FF,$03,$F0,$3F,$82,$F8,$19
	.dc.b	$5F,$4E,$3A,$BF,$57,$0E,$AF,$E1,$B0,$E8,$FE,$23,$0B,$B7,$BB,$1F
	.dc.b	$C6,$FC,$0C,$AF,$BC,$FE,$43,$E0,$FF,$D7,$1F,$B6,$7F,$AD,$CF,$EC
	.dc.b	$BE,$11,$A6,$7E,$0F,$7F,$07,$30,$FF,$0F,$CF,$F6,$07,$F5,$2F,$F4
	.dc.b	$EA,$E5,$F0,$7D,$8E,$FF,$03,$E0,$F3,$FC,$3F,$11,$30,$E7,$1E,$4C
	.dc.b	$BB,$47,$4C,$B0,$F8,$3E,$0B,$D3,$F0,$52,$30,$72,$3C,$B8,$75,$7C
	.dc.b	$B8,$72,$7A,$1E,$AB,$C7,$07,$F2,$7D,$5F,$07,$F1,$89,$5F,$03,$E1
	.dc.b	$3E,$4F,$C2,$7C,$AF,$84,$93,$F1,$23,$E6,$FB,$F2,$3E,$3B,$F4,$3D
	.dc.b	$FF,$A3,$EF,$FD,$2F,$82,$95,$F3,$23,$AB,$F3,$B0,$EA,$FC,$EC,$3A
	.dc.b	$3F,$43,$0B,$A7,$A7,$1F,$BB,$F0,$92,$BE,$BB,$F8,$0F,$84,$FC,$0F
	.dc.b	$C2,$49,$FB,$6F,$E1,$3E,$13,$F0,$BF,$09,$FE,$FD,$AF,$EB,$58,$FF
	.dc.b	$4B,$6F,$1F,$F4,$4E,$25,$6F,$E0,$26,$5F,$E6,$4B,$FE,$A0,$DF,$84
	.dc.b	$F0,$3F,$88,$CE,$E5,$F4,$73,$F0,$CF,$1F,$88,$F1,$18,$7F,$9B,$E0
	.dc.b	$F8,$2F,$31,$23,$C1,$F0,$BE,$9F,$85,$91,$E3,$7D,$9F,$0B,$ED,$F8
	.dc.b	$59,$3E,$77,$E0,$7C,$2F,$C1,$F8,$3F,$85,$F0,$72,$BD,$91,$D1,$F1
	.dc.b	$E1,$76,$F8,$D8,$F4,$9F,$3E,$36,$7E,$56,$3F,$37,$E0,$BE,$77,$C1
	.dc.b	$CB,$F9,$11,$CF,$9B,$FB,$8E,$EC,$F1,$4F,$1D,$3F,$ED,$88,$F5,$10
	.dc.b	$F0,$9F,$4F,$1F,$B9,$F0,$B2,$3E,$B3,$F7,$FE,$17,$F0,$1F,$09,$F8
	.dc.b	$1F,$85,$95,$F6,$E3,$F0,$9F,$03,$F8,$5F,$85,$93,$F7,$E3,$F1,$1F
	.dc.b	$03,$F8,$9F,$82,$93,$F8,$38,$EA,$FE,$17,$0E,$AF,$DC,$C3,$AB,$F8
	.dc.b	$9C,$3A,$BF,$80,$C0,$FF,$AC,$3F,$6D,$FE,$D5,$C7,$FE,$1F,$6F,$0F
	.dc.b	$F8,$61,$B3,$F1,$D7,$C4,$72,$8F,$F0,$0C,$3E,$1F,$17,$CD,$B7,$CD
	.dc.b	$B5,$CF,$E1,$EF,$1B,$0F,$F6,$AC,$F9,$7E,$1B,$31,$FE,$D3,$C3,$C9
	.dc.b	$F0,$D8,$F8,$39,$7E,$B7,$7F,$0D,$C3,$39,$3E,$F7,$9E,$7D,$2F,$AE
	.dc.b	$52,$BE,$24,$78,$3E,$0A,$47,$C7,$7C,$7F,$07,$E4,$F8,$69,$3F,$2D
	.dc.b	$F3,$FC,$37,$A3,$E1,$7D,$3F,$0B,$2B,$E8,$47,$B3,$E0,$BD,$BF,$07
	.dc.b	$27,$EA,$47,$47,$E6,$E1,$79,$3E,$B6,$3D,$67,$ED,$63,$D5,$FA,$78
	.dc.b	$FC,$8F,$83,$91,$F7,$DF,$97,$F0,$7C,$F9,$3E,$F6,$56,$BE,$F6,$3D
	.dc.b	$5F,$BB,$8F,$5C,$BF,$0D,$97,$D4,$F8,$49,$5F,$8B,$7E,$BF,$C2,$FD
	.dc.b	$8F,$85,$FB,$3F,$0B,$2B,$F2,$11,$F8,$0F,$DB,$BF,$D2,$11,$FE,$4A
	.dc.b	$7F,$B5,$93,$5D,$7C,$24,$9F,$CC,$E7,$FE,$75,$AF,$ED,$4D,$F8,$1F
	.dc.b	$04,$FE,$77,$1F,$03,$F9,$6C,$7C,$18,$F8,$7E,$0A,$4F,$48,$FF,$3B
	.dc.b	$E3,$E4,$C7,$B7,$C3,$79,$FE,$1A,$5E,$CF,$E1,$55,$FD,$E4,$7C,$37
	.dc.b	$B7,$E1,$A4,$EF,$C2,$56,$A3,$0F,$F6,$86,$FC,$37,$C5,$3E,$37,$C3
	.dc.b	$7C,$7F,$83,$93,$C1,$F9,$5F,$09,$F2,$F2,$EF,$3F,$2F,$2C,$3D,$08
	.dc.b	$7C,$FC,$3A,$BE,$2F,$84,$FA,$84,$7C,$BF,$09,$F5,$FE,$17,$EC,$7C
	.dc.b	$34,$9F,$4C,$7D,$BF,$86,$FB,$9E,$B4,$9F,$6C,$7D,$FF,$82,$91,$F0
	.dc.b	$1F,$C1,$7C,$17,$58,$F8,$78,$75,$7F,$09,$87,$57,$E3,$E1,$D5,$FC
	.dc.b	$46,$1F,$8B,$F8,$59,$1F,$31,$FE,$1B,$FB,$7F,$F7,$08,$FA,$F7,$D7
	.dc.b	$9F,$F1,$58,$F8,$67,$99,$D8,$86,$0F,$F4,$FD,$7F,$74,$6F,$C3,$76
	.dc.b	$3B,$FC,$37,$83,$E1,$64,$FE,$06,$BF,$B8,$37,$E1,$7C,$BE,$FC,$8D
	.dc.b	$3E,$8F,$83,$F4,$FC,$1C,$9E,$4F,$B3,$E0,$7A,$C7,$93,$1B,$AF,$FB
	.dc.b	$67,$26,$5D,$6B,$D3,$95,$AD,$65,$F1,$BE,$12,$57,$9D,$F9,$3F,$0B
	.dc.b	$F2,$BE,$12,$4F,$AD,$F9,$BF,$0B,$F3,$BE,$17,$E7,$FC,$2C,$AF,$81
	.dc.b	$1F,$4B,$E0,$A4,$7C,$37,$EA,$FC,$1F,$48,$F9,$B8,$5D,$3D,$58,$F5
	.dc.b	$9F,$95,$8F,$57,$E9,$E3,$F7,$BE,$0A,$47,$CF,$7F,$03,$F0,$7F,$82
	.dc.b	$F8,$4F,$C1,$FC,$2C,$AF,$A9,$1F,$D4,$BD,$3F,$EB,$FE,$51,$FE,$BF
	.dc.b	$3F,$B9,$6B,$FE,$E0,$76,$7F,$61,$75,$F0,$1C,$CE,$9F,$03,$E1,$8F
	.dc.b	$C1,$61,$E1,$7F,$17,$87,$85,$F1,$7C,$2C,$7C,$07,$97,$E0,$7C,$DF
	.dc.b	$05,$23,$C9,$1D,$5F,$CA,$DF,$5E,$1F,$C1,$61,$D6,$7D,$58,$75,$7D
	.dc.b	$47,$C1,$F8,$48,$FB,$DF,$89,$F0,$9F,$17,$E1,$3E,$37,$C2,$4A,$F8
	.dc.b	$91,$F2,$7E,$03,$E5,$7C,$04,$9F,$91,$1F,$37,$DF,$91,$F2,$DF,$A1
	.dc.b	$F0,$1D,$63,$E7,$61,$D5,$F9,$D8,$75,$7E,$96,$1D,$5F,$A5,$87,$D9
	.dc.b	$F8,$3F,$B5,$F0,$92,$7E,$C4,$7D,$DF,$84,$91,$F6,$DF,$C0,$7C,$27
	.dc.b	$E0,$79,$FF,$05,$F0,$92,$BF,$01,$1F,$85,$F8,$19,$1F,$83,$7F,$11
	.dc.b	$F0,$5D,$23,$F0,$98,$5E,$57,$C0,$FF,$8C,$3B,$3F,$19,$9F,$7B,$FC
	.dc.b	$66,$7D,$DF,$C6,$67,$FE,$24,$FB,$CF,$EA,$58,$71,$7F,$D9,$E4,$75
	.dc.b	$F0,$BB,$F8,$59,$7F,$D3,$EB,$FB,$2D,$71,$F8,$4E,$9F,$09,$2B,$51
	.dc.b	$DF,$E0,$24,$71,$7C,$5F,$01,$DA,$3A,$63,$D9,$6F,$EE,$7F,$CD,$99
	.dc.b	$78,$1B,$87,$77,$79,$7A,$F2,$F7,$4F,$BF,$E0,$25,$7A,$5F,$85,$F0
	.dc.b	$7E,$18,$E2,$76,$CB,$C5,$97,$87,$85,$9F,$85,$B8,$F8,$67,$52,$CB
	.dc.b	$C3,$87,$C7,$F8,$3F,$A1,$F0,$72,$30,$7C,$BF,$83,$FA,$84,$7E,$77
	.dc.b	$C1,$FD,$7F,$7B,$EC,$7B,$D2,$7E,$94,$7D,$BF,$80,$FB,$9F,$03,$27
	.dc.b	$EB,$47,$37,$DB,$BF,$AB,$1C,$6D,$E8,$CB,$B5,$7D,$DC,$BB,$3C,$36
	.dc.b	$67,$F8,$9F,$83,$93,$F8,$37,$F1,$9F,$07,$F8,$DF,$83,$93,$F8,$77
	.dc.b	$FC,$E7,$F6,$1F,$D9,$F5,$3F,$D7,$F4,$F9,$CD,$E7,$FC,$E3,$6F,$FD
	.dc.b	$58,$BE,$0F,$9F,$C0,$F4,$F8,$09,$5F,$E2,$3A,$FE,$C6,$FA,$72,$F1
	.dc.b	$B4,$D7,$C4,$79,$3E,$16,$47,$17,$CF,$8F,$A3,$E0,$A4,$F5,$7D,$7F
	.dc.b	$05,$EC,$F8,$59,$3E,$17,$DF,$F0,$BF,$03,$E1,$3E,$0F,$C2,$4A,$F3
	.dc.b	$47,$C4,$F8,$29,$1E,$97,$E3,$FC,$17,$48,$FB,$3B,$B1,$C7,$1F,$C8
	.dc.b	$CB,$B5,$7C,$1C,$BB,$3C,$32,$FA,$3F,$03,$23,$E3,$3F,$53,$E0,$B9
	.dc.b	$F2,$FD,$0C,$AD,$7D,$0C,$7A,$BE,$2C,$7A,$E5,$F6,$32,$FB,$BF,$09
	.dc.b	$2B,$E9,$3F,$80,$F8,$4F,$C0,$FC,$24,$9F,$AC,$FF,$AE,$2D,$F8,$DF
	.dc.b	$F9,$66,$7D,$FF,$0B,$F8,$8F,$80,$91,$F7,$63,$F1,$7F,$03,$23,$F0
	.dc.b	$0F,$FB,$03,$ED,$3F,$B7,$F1,$B2,$F0,$57,$CA,$7C,$18,$F7,$CB,$F2
	.dc.b	$38,$F6,$AF,$9F,$97,$E6,$23,$A7,$C1,$75,$F8,$59,$5F,$E7,$1C,$FF
	.dc.b	$BC,$E1,$FC,$1D,$7C,$37,$8C,$93,$FE,$7D,$97,$E3,$8F,$6F,$D1,$F0
	.dc.b	$D2,$BA,$3E,$BF,$87,$F6,$7C,$3F,$B7,$E1,$E5,$78,$63,$AB,$FB,$EC
	.dc.b	$FC,$2F,$97,$3F,$9D,$97,$78,$EA,$43,$86,$3D,$DF,$67,$C2,$7C,$B2
	.dc.b	$3E,$FF,$85,$F9,$DF,$09,$23,$E1,$3D,$8F,$A5,$F0,$B2,$3E,$33,$F5
	.dc.b	$7E,$1F,$EB,$7C,$3F,$D7,$F8,$69,$5F,$2E,$3A,$3E,$BB,$BC,$5F,$D0
	.dc.b	$CF,$92,$F5,$8F,$81,$85,$9F,$A5,$87,$E0,$FE,$1F,$F0,$9F,$0D,$2F
	.dc.b	$EC,$47,$4F,$0F,$FB,$57,$0B,$2C,$57,$F5,$2E,$0C,$3E,$A2,$1E,$03
	.dc.b	$F0,$58,$FE,$47,$E1,$A4,$7E,$15,$FC,$A7,$C2,$FE,$57,$E1,$A4,$FE
	.dc.b	$25,$FA,$BD,$3F,$99,$E9,$93,$F8,$D7,$F3,$9F,$0E,$8F,$F6,$07,$67
	.dc.b	$F7,$A5,$EA,$21,$F1,$5B,$9F,$E8,$31,$87,$8F,$07,$89,$FC,$46,$57
	.dc.b	$8F,$E7,$6F,$FE,$51,$9F,$C6,$67,$E4,$69,$FA,$1C,$FB,$57,$F6,$7C
	.dc.b	$7F,$B3,$0F,$83,$CE,$5F,$E1,$78,$C5,$FA,$BE,$23,$F8,$17,$0B,$C7
	.dc.b	$DC,$FB,$29,$F0,$98,$E0,$F7,$3C,$2B,$13,$FD,$C6,$BE,$2F,$C2,$60
	.dc.b	$CF,$C4,$6B,$E2,$37,$F1,$12,$BE,$5C,$73,$7E,$6E,$3D,$0C,$3F,$2A
	.dc.b	$FA,$3D,$8F,$A7,$7E,$2F,$88,$91,$F5,$9F,$2F,$C4,$7B,$BD,$7F,$3F
	.dc.b	$C4,$4A,$FE,$09,$BA,$7D,$7F,$0D,$27,$EF,$BE,$EF,$86,$F7,$FC,$24
	.dc.b	$9F,$C1,$BF,$0B,$E1,$BB,$47,$E1,$B1,$EC,$FF,$87,$B9,$D9,$77,$8E
	.dc.b	$A4,$3F,$0F,$8F,$77,$F1,$FF,$0F,$F3,$BE,$22,$46,$0F,$4E,$9D,$B0
	.dc.b	$FC,$16,$3E,$29,$E6,$75,$8F,$C3,$61,$D6,$3F,$2B,$F0,$FF,$68,$8F
	.dc.b	$E7,$BE,$1F,$FA,$37,$E3,$A6,$92,$3F,$96,$42,$7F,$07,$F0,$D2,$7F
	.dc.b	$4E,$FE,$27,$ED,$BF,$C2,$47,$F8,$37,$AC,$FF,$83,$67,$F5,$D9,$7F
	.dc.b	$90,$63,$11,$FE,$2A,$7F,$C2,$90,$C3,$C9,$3E,$2C,$7C,$8F,$CF,$8F
	.dc.b	$23,$E2,$21,$FB,$4C,$3C,$9F,$0F,$E5,$F8,$9F,$37,$C4,$CA,$F1,$F8
	.dc.b	$1F,$E7,$FC,$FF,$C3,$B9,$FF,$23,$C9,$BF,$C6,$25,$FC,$4E,$7F,$81
	.dc.b	$F0,$F2,$39,$9F,$C2,$30,$47,$C5,$E2,$F8,$DF,$07,$2B,$7C,$3F,$1A
	.dc.b	$3E,$57,$C1,$F6,$9F,$76,$57,$4F,$26,$7D,$E7,$E4,$65,$67,$E6,$65
	.dc.b	$F4,$FE,$0A,$57,$6D,$11,$F5,$FE,$0F,$F0,$3B,$FD,$9F,$87,$97,$F8
	.dc.b	$08,$FB,$9F,$0F,$23,$93,$F7,$FE,$1F,$F0,$1F,$0F,$F8,$1F,$22,$57
	.dc.b	$78,$EA,$F8,$70,$EA,$FD,$9C,$3A,$3F,$81,$C2,$EB,$F8,$5C,$7F,$19
	.dc.b	$F0,$FF,$8D,$F8,$59,$7E,$B8,$EA,$FE,$27,$1E,$B7,$F6,$71,$E6,$7D
	.dc.b	$DC,$6D,$7D,$5C,$7F,$31,$E5,$48,$F8,$CF,$E6,$FE,$13,$FC,$55,$F8
	.dc.b	$F3,$DD,$A4,$AF,$F0,$84,$3D,$AF,$D1,$7C,$47,$F2,$BF,$B7,$FE,$8D
	.dc.b	$5F,$E8,$66,$EB,$FA,$19,$D7,$C3,$EF,$D9,$97,$F8,$4B,$FF,$10,$EF
	.dc.b	$FA,$8C,$BC,$53,$E0,$CF,$C4,$FF,$11,$8F,$13,$F7,$F1,$F1,$30,$C1
	.dc.b	$D7,$E1,$3C,$DF,$09,$23,$07,$B3,$E1,$BE,$09,$1F,$7F,$C3,$7B,$7E
	.dc.b	$1F,$DC,$C8,$F4,$47,$C1,$9F,$85,$F0,$F2,$7D,$91,$CD,$F4,$37,$83
	.dc.b	$FC,$25,$9B,$0E,$D5,$F0,$DE,$CF,$CB,$C3,$E7,$7C,$3F,$CF,$F8,$39
	.dc.b	$3F,$26,$3F,$CC,$5F,$8F,$FF,$95,$E3,$FC,$8C,$D7,$F9,$C3,$71,$FF
	.dc.b	$11,$3F,$DF,$E3,$E9,$7C,$3F,$DB,$F4,$E4,$60,$FA,$DF,$11,$F7,$F8
	.dc.b	$A4,$60,$EF,$DD,$E2,$23,$E2,$EA,$FC,$36,$1D,$23,$F0,$D8,$61,$FC
	.dc.b	$17,$4F,$E2,$C8,$FE,$17,$E2,$39,$BE,$AC,$2E,$9E,$AC,$79,$4F,$43
	.dc.b	$F1,$98,$62,$7F,$DE,$5D,$0F,$87,$FC,$C7,$C3,$CB,$FC,$94,$7C,$AF
	.dc.b	$17,$F3,$9F,$0D,$27,$E5,$C7,$E7,$FE,$17,$F4,$1F,$0B,$27,$F3,$51
	.dc.b	$FE,$0D,$FD,$C3,$FC,$59,$E8,$9F,$F1,$44,$FF,$BA,$9B,$A3,$C0,$FF
	.dc.b	$1A,$56,$DF,$15,$7E,$97,$2F,$13,$F6,$B2,$F1,$3F,$A9,$CB,$C4,$FE
	.dc.b	$C7,$2F,$17,$C2,$F8,$FE,$1A,$4F,$F3,$2D,$13,$F8,$8F,$8D,$F4,$7C
	.dc.b	$2C,$AF,$C4,$C7,$AF,$E1,$64,$71,$7D,$DF,$0B,$EF,$F8,$6F,$81,$F0
	.dc.b	$F2,$BB,$47,$57,$C1,$87,$57,$F2,$78,$74,$7E,$26,$17,$8B,$E2,$47
	.dc.b	$F9,$53,$3F,$F5,$77,$0F,$C6,$5F,$0F,$F3,$FD,$79,$5E,$D8,$FF,$62
	.dc.b	$7B,$0F,$82,$A4,$9F,$F5,$54,$3E,$1F,$EC,$3F,$66,$64,$FC,$78,$FB
	.dc.b	$93,$F7,$7E,$1E,$4F,$CB,$8E,$8F,$F8,$96,$7C,$0F,$6A,$F1,$CF,$68
	.dc.b	$FA,$51,$D9,$FC,$24,$FE,$27,$E1,$FF,$15,$F0,$F2,$7E,$C4,$76,$D7
	.dc.b	$FD,$35,$5D,$A7,$C1,$7D,$9F,$41,$D8,$87,$63,$F0,$5F,$0F,$F9,$72
	.dc.b	$3F,$85,$F8,$7F,$93,$C5,$F9,$BF,$87,$93,$F2,$23,$F3,$DF,$0F,$FE
	.dc.b	$6B,$A7,$C3,$D3,$35,$FE,$00,$8F,$F1,$31,$FE,$22,$CF,$EE,$F7,$FF
	.dc.b	$8B,$FE,$37,$97,$3F,$82,$FF,$2D,$9F,$81,$F9,$99,$F4,$58,$72,$C7
	.dc.b	$C3,$1D,$F2,$D4,$E7,$E2,$3C,$C4,$9F,$27,$C4,$7F,$A1,$A5,$F9,$0B
	.dc.b	$C5,$F6,$7C,$34,$BF,$F7,$84,$E5,$F6,$7C,$37,$C1,$F8,$89,$18,$3E
	.dc.b	$07,$C4,$7C,$52,$3F,$0F,$C7,$F0,$5F,$38,$F5,$F7,$FF,$02,$DF,$FC
	.dc.b	$57,$1F,$27,$2D,$E7,$F3,$F3,$EF,$3C,$0F,$F5,$CD,$33,$D4,$7D,$1F
	.dc.b	$87,$FA,$C4,$7E,$A7,$C2,$78,$BD,$C9,$1E,$E7,$CB,$EB,$FD,$DF,$84
	.dc.b	$93,$F6,$DF,$C0,$7C,$27,$E0,$7E,$1B,$F0,$5F,$0F,$2B,$F0,$11,$D1
	.dc.b	$FF,$44,$70,$B1,$C7,$27,$FB,$CE,$19,$76,$AF,$67,$67,$F9,$CF,$3C
	.dc.b	$5F,$DF,$CF,$F2,$1F,$0F,$F9,$1F,$5A,$61,$F8,$F8,$EB,$CB,$FE,$72
	.dc.b	$CF,$C1,$3E,$D3,$ED,$E7,$D2,$7E,$86,$56,$7E,$5E,$3F,$9D,$F5,$BF
	.dc.b	$3D,$F0,$F2,$7F,$39,$1F,$69,$87,$E1,$8F,$BB,$F0,$FF,$A5,$F8,$6F
	.dc.b	$F6,$87,$EE,$3F,$E2,$4C,$BF,$9F,$B6,$BE,$CC,$EA,$A4,$7F,$9A,$27
	.dc.b	$FC,$4A,$FF,$BD,$9F,$F7,$66,$FF,$26,$3C,$27,$CE,$3C,$27,$EC,$6B
	.dc.b	$C2,$C3,$07,$5F,$87,$F2,$91,$F0,$7C,$3F,$E9,$7F,$23,$6B,$49,$3F
	.dc.b	$E8,$2C,$12,$E5,$F7,$FC,$2C,$AF,$53,$F0,$BE,$1B,$E1,$FC,$44,$9F
	.dc.b	$73,$F1,$BA,$3A,$C7,$FB,$3F,$0F,$0B,$DD,$AE,$70,$EC,$FE,$23,$66
	.dc.b	$38,$CB,$F8,$7C,$7F,$8B,$E7,$E6,$72,$7D,$42,$3F,$3F,$E1,$FE,$07
	.dc.b	$89,$23,$C0,$FC,$3F,$17,$ED,$FC,$27,$DC,$F8,$29,$5F,$5E,$3E,$FF
	.dc.b	$C3,$48,$FB,$4F,$E0,$BE,$1B,$A4,$7B,$70,$BA,$7F,$60,$3E,$5E,$3D
	.dc.b	$A7,$F0,$58,$F6,$7D,$D8,$FE,$33,$E1,$BF,$1B,$F0,$D2,$7F,$11,$1D
	.dc.b	$3A,$BF,$DA,$7C,$4C,$F1,$5F,$CA,$A3,$FD,$87,$3F,$94,$C3,$B6,$7F
	.dc.b	$23,$2F,$15,$FC,$DF,$85,$FC,$D9,$1F,$CB,$79,$3F,$CA,$FF,$24,$FF
	.dc.b	$6B,$C7,$F1,$5B,$47,$FB,$7A,$30,$D8,$F8,$0F,$07,$EB,$FC,$37,$EA
	.dc.b	$3E,$16,$57,$E8,$E3,$B7,$BD,$23,$F2,$56,$8F,$17,$B3,$E3,$F6,$65
	.dc.b	$69,$F3,$7B,$9E,$7F,$76,$4F,$27,$D5,$EE,$7A,$FD,$7F,$67,$C0,$4A
	.dc.b	$EF,$1E,$FF,$83,$91,$E2,$7E,$17,$C1,$7C,$3F,$83,$93,$E6,$7E,$37
	.dc.b	$C1,$7C,$7F,$84,$F9,$1F,$0B,$2B,$D7,$07,$FD,$DE,$3F,$33,$7B,$DF
	.dc.b	$D8,$A1,$F6,$BF,$9D,$5F,$9A,$BD,$EE,$51,$AF,$36,$CE,$63,$B5,$71
	.dc.b	$C3,$AF,$56,$2F,$C4,$BD,$B7,$5F,$F6,$95,$F1,$34,$C4,$7E,$0F,$C1
	.dc.b	$99,$7D,$67,$F0,$DF,$23,$E4,$C3,$FC,$7F,$3E,$DD,$BD,$9A,$7A,$F5
	.dc.b	$F3,$4F,$A4,$F8,$8E,$23,$D1,$3C,$34,$F3,$E9,$E6,$BF,$E2,$CF,$99
	.dc.b	$BA,$79,$9C,$F7,$FA,$C8,$F0,$EF,$F3,$6B,$F2,$B7,$F8,$B3,$E3,$3E
	.dc.b	$7F,$07,$CE,$D3,$D0,$DC,$F8,$69,$E5,$AF,$F0,$CE,$9E,$57,$F4,$CF
	.dc.b	$95,$BA,$79,$5C,$F8,$DF,$3C,$8F,$B3,$C1,$F2,$BF,$87,$CF,$CB,$9F
	.dc.b	$8A,$3E,$D1,$F2,$33,$F2,$3B,$CB,$C9,$C1,$F7,$7C,$2F,$B8,$49,$F9
	.dc.b	$3E,$4F,$DF,$F2,$64,$7C,$C7,$F0,$5E,$47,$8A,$3E,$D5,$F8,$1F,$8A
	.dc.b	$5E,$4F,$B1,$C3,$F1,$27,$CA,$78,$67,$EB,$E3,$E1,$7E,$2E,$3F,$8F
	.dc.b	$F2,$A4,$7D,$C7,$F2,$5E,$3F,$87,$7F,$83,$8F,$E7,$B6,$F7,$DF,$88
	.dc.b	$FB,$13,$A7,$E9,$E3,$DF,$2F,$E6,$5C,$68,$FC,$F6,$7F,$9F,$F3,$FF
	.dc.b	$41,$E6,$4C,$3F,$1B,$1E,$1E,$1F,$BD,$1D,$F8,$BE,$93,$E0,$8F,$9F
	.dc.b	$8B,$D2,$FD,$57,$9F,$2B,$F2,$F1,$FA,$DD,$FF,$5D,$E8,$C9,$FC,$DC
	.dc.b	$77,$DB,$E6,$70,$7B,$B8,$FE,$6D,$7D,$D9,$F9,$B9,$77,$8F,$D1,$E5
	.dc.b	$DD,$FC,$8E,$5F,$BB,$F8,$E9,$1F,$A8,$7F,$7D,$EA,$77,$8F,$D0,$63
	.dc.b	$D9,$F7,$67,$F6,$DF,$DB,$65,$FC,$1F,$0E,$CF,$CC,$C7,$F8,$44,$7F
	.dc.b	$09,$98,$7E,$D8,$FE,$19,$EB,$77,$DF,$E4,$93,$0E,$CF,$F7,$BC,$1F
	.dc.b	$93,$BD,$CF,$F3,$E2,$F6,$29,$88,$CF,$B3,$32,$FF,$0D,$52,$38,$FB
	.dc.b	$5E,$1C,$BF,$15,$5F,$38,$FC,$E7,$4F,$E2,$72,$F0,$E3,$DA,$79,$E5
	.dc.b	$DD,$EB,$EC,$F9,$08,$F8,$3D,$FE,$CF,$ED,$4E,$C7,$E1,$EF,$F4,$BA
	.dc.b	$7B,$63,$CF,$87,$57,$FB,$74,$3D,$EF,$83,$EE,$C9,$F6,$BF,$13,$DD
	.dc.b	$EF,$1F,$60,$F0,$1D,$CF,$B3,$4F,$84,$F9,$64,$8F,$8F,$D7,$F3,$BE
	.dc.b	$07,$E7,$F5,$49,$F9,$71,$DB,$A3,$F0,$C7,$E1,$F8,$7E,$46,$3D,$6B
	.dc.b	$E8,$E3,$D5,$F5,$E3,$F6,$BA,$FE,$DF,$C1,$49,$FA,$F1,$D4,$FD,$01
	.dc.b	$FB,$EC,$3A,$D7,$3E,$0F,$F1,$66,$3D,$EF,$A9,$F8,$0F,$7B,$F0,$FF
	.dc.b	$09,$23,$07,$F7,$0C,$3F,$83,$3F,$1C,$8F,$AB,$E1,$7F,$1F,$F0,$F2
	.dc.b	$7F,$17,$1F,$92,$F8,$8F,$C9,$FC,$1C,$9F,$C7,$C7,$A5,$FB,$FD,$BF
	.dc.b	$D6,$B4,$FB,$F7,$CE,$7E,$EE,$1F,$9C,$8F,$F6,$C7,$E5,$1F,$E2,$9F
	.dc.b	$43,$FB,$AB,$6B,$FD,$CA,$DB,$2F,$8A,$F0,$4F,$EE,$B8,$7E,$47,$0F
	.dc.b	$F4,$2E,$1F,$F7,$5C,$F8,$0F,$D2,$E7,$E1,$7F,$D0,$18,$3E,$2F,$C6
	.dc.b	$48,$CF,$C5,$76,$DB,$F7,$11,$FC,$4C,$FE,$0B,$4F,$82,$F5,$FC,$14
	.dc.b	$9E,$B1,$EE,$F8,$2F,$7F,$C6,$49,$F0,$C7,$5F,$13,$F6,$B1,$F9,$9C
	.dc.b	$7F,$82,$63,$D6,$3C,$F8,$F5,$7F,$CA,$91,$96,$FC,$2F,$CC,$F8,$59
	.dc.b	$1E,$E8,$EC,$FE,$9A,$FE,$B3,$EA,$CF,$A7,$1F,$F9,$E7,$06,$3F,$5A
	.dc.b	$7E,$BF,$C0,$4B,$F9,$2F,$DA,$F8,$6E,$B1,$F8,$CE,$3F,$EE,$78,$3E
	.dc.b	$1B,$EF,$92,$7E,$8F,$C2,$FE,$0B,$E1,$64,$7D,$47,$F0,$BF,$0B,$D3
	.dc.b	$C3,$FD,$86,$3D,$30,$E5,$7F,$09,$FC,$5E,$1F,$8C,$F8,$5F,$C6,$FC
	.dc.b	$34,$AF,$BF,$1C,$FA,$FF,$0F,$C3,$FD,$E3,$5F,$35,$7E,$13,$1E,$B3
	.dc.b	$F0,$F1,$FC,$BF,$C2,$4A,$FC,$4B,$F9,$AF,$86,$E9,$1F,$CB,$34,$FF
	.dc.b	$09,$FE,$53,$FE,$25,$C3,$FD,$48,$DF,$03,$FC,$A6,$DA,$4A,$1F,$0B
	.dc.b	$FE,$24,$8C,$7C,$2D,$9F,$F1,$74,$74,$8F,$D5,$DF,$8A,$FF,$15,$D7
	.dc.b	$DB,$E1,$F8,$12,$3F,$DC,$18,$76,$6F,$C2,$F5,$23,$AF,$87,$E9,$D9
	.dc.b	$F8,$83,$FA,$C5,$71,$9F,$F5,$33,$EB,$CF,$D5,$87,$9F,$4F,$47,$C2
	.dc.b	$CB,$F0,$BE,$BF,$87,$EB,$A7,$F6,$43,$E2,$FC,$2F,$BF,$E1,$65,$FA
	.dc.b	$23,$E1,$7C,$2F,$C3,$F8,$79,$3E,$C8,$E6,$FE,$DF,$6F,$3D,$72,$8F
	.dc.b	$3D,$FC,$B8,$F9,$9F,$0D,$F3,$7E,$1E,$57,$C6,$8E,$6F,$F9,$10,$FF
	.dc.b	$02,$F0,$2B,$A7,$07,$DE,$C3,$A5,$7C,$EF,$86,$FB,$1F,$0F,$23,$07
	.dc.b	$F3,$5E,$7F,$F1,$7F,$81,$FD,$3B,$BB,$F2,$B3,$DB,$E1,$BF,$01,$F0
	.dc.b	$B2,$3E,$CC,$7E,$0F,$E1,$BF,$09,$C9,$27,$EE,$C7,$A3,$D6,$FE,$FD
	.dc.b	$C0,$BE,$55,$F6,$30,$FC,$64,$7E,$37,$E1,$BF,$1D,$F0,$F2,$BF,$0F
	.dc.b	$1C,$F5,$FC,$ED,$72,$E1,$FC,$EE,$1C,$AF,$FD,$5F,$DD,$FE,$B0,$E3
	.dc.b	$FF,$0D,$47,$E4,$0F,$CD,$1F,$9D,$72,$FE,$8A,$3F,$04,$FF,$17,$DB
	.dc.b	$F9,$2F,$37,$E4,$5E,$9F,$05,$C4,$93,$E0,$3F,$57,$7F,$E0,$B7,$18
	.dc.b	$FD,$7F,$87,$F0,$7C,$3C,$9C,$D7,$F8,$A5,$FA,$95,$FD,$73,$0F,$09
	.dc.b	$E4,$7F,$01,$87,$82,$7D,$07,$81,$F6,$7C,$27,$B7,$E1,$BD,$DF,$0D
	.dc.b	$2B,$BC,$7C,$BB,$FF,$44,$E3,$FE,$25,$9E,$71,$D9,$F2,$E1,$D1,$F3
	.dc.b	$FC,$37,$C9,$F8,$69,$18,$3D,$27,$F6,$53,$D4,$7E,$8F,$3F,$A9,$C5
	.dc.b	$F4,$49,$1F,$0B,$E1,$3E,$A7,$C2,$C8,$F8,$AF,$D7,$F8,$6F,$44,$7F
	.dc.b	$99,$B4,$FF,$7D,$BF,$2A,$F9,$BE,$A8,$FF,$0C,$68,$E3,$FC,$27,$5F
	.dc.b	$E0,$49,$5F,$4B,$E1,$BA,$F2,$FF,$99,$70,$EB,$3C,$F1,$EB,$8F,$38
	.dc.b	$FB,$FA,$7E,$2F,$E1,$3F,$19,$F0,$D2,$7E,$EC,$74,$DD,$1F,$CC,$3F
	.dc.b	$2A,$FF,$06,$3F,$82,$F8,$7F,$CA,$1F,$CC,$63,$31,$EB,$F8,$3F,$CC
	.dc.b	$7C,$3C,$8F,$C4,$C7,$57,$A7,$C3,$71,$23,$F8,$EF,$87,$E9,$F0,$B2
	.dc.b	$3F,$2B,$82,$3C,$1F,$1D,$CF,$74,$F8,$9F,$80,$63,$0E,$B7,$F3,$F0
	.dc.b	$F4,$7C,$2F,$A7,$E1,$A4,$F6,$8E,$98,$3E,$17,$DD,$F0,$F2,$BC,$71
	.dc.b	$F0,$7E,$17,$E1,$7C,$34,$9F,$3C,$72,$9F,$8C,$48,$F5,$FC,$3F,$C9
	.dc.b	$F8,$79,$1E,$E7,$E6,$7C,$3F,$18,$EE,$43,$F4,$39,$2B,$E9,$7C,$3C
	.dc.b	$9F,$8E,$FD,$5F,$87,$F5,$FC,$0F,$D7,$F8,$39,$5F,$32,$3E,$D7,$C3
	.dc.b	$C8,$F9,$EF,$DD,$F8,$5E,$31,$F4,$AB,$8B,$E2,$AF,$C1,$78,$9F,$83
	.dc.b	$F8,$59,$3F,$62,$38,$BB,$3C,$07,$DC,$F8,$7F,$EF,$5F,$95,$FF,$65
	.dc.b	$3F,$BC,$CF,$E4,$BA,$51,$8A,$FF,$01,$BF,$E0,$57,$FA,$87,$8B,$FE
	.dc.b	$52,$7F,$C5,$D3,$F9,$A8,$FF,$65,$4B,$E0,$7B,$92,$BF,$90,$C3,$E1
	.dc.b	$BC,$7F,$0D,$E4,$F8,$59,$5A,$8F,$C8,$57,$38,$E3,$7C,$DF,$E2,$B1
	.dc.b	$CD,$EB,$F0,$9E,$E2,$3E,$0F,$85,$E7,$5C,$73,$FC,$CD,$F5,$AE,$79
	.dc.b	$7C,$6F,$82,$91,$E8,$7E,$4F,$C2,$F9,$3E,$13,$E5,$92,$7D,$BF,$07
	.dc.b	$F3,$BE,$17,$E7,$FA,$12,$7E,$0C,$7B,$27,$84,$7A,$AB,$EA,$C7,$D6
	.dc.b	$F3,$24,$FC,$87,$EC,$FC,$2F,$18,$F9,$75,$CB,$5F,$5D,$FD,$EF,$81
	.dc.b	$95,$F4,$1F,$C0,$FC,$07,$CC,$F8,$0F,$C1,$92,$7E,$AF,$C2,$7E,$19
	.dc.b	$87,$D8,$3F,$13,$F0,$5F,$C3,$5F,$C1,$E9,$FC,$69,$2C,$31,$D1,$89
	.dc.b	$FE,$3F,$39,$7F,$CA,$D9,$FE,$26,$3F,$25,$F9,$63,$E0,$BF,$CA,$0F
	.dc.b	$F7,$49,$FE,$1E,$DE,$BD,$CF,$8F,$E1,$7F,$D9,$B7,$D7,$06,$3E,$17
	.dc.b	$CB,$97,$F8,$72,$3C,$99,$32,$FE,$ED,$5F,$D8,$1F,$EE,$4D,$F8,$4F
	.dc.b	$1F,$C2,$77,$F6,$FF,$CC,$F3,$DB,$2F,$84,$78,$71,$F0,$34,$FE,$F1
	.dc.b	$83,$E1,$3D,$A4,$AF,$F0,$2C,$7F,$71,$CB,$FA,$64,$7C,$E3,$FC,$24
	.dc.b	$43,$04,$AE,$CF,$CE,$CF,$DD,$1F,$E1,$A8,$68,$F7,$7F,$DF,$C6,$A3
	.dc.b	$FB,$A1,$D7,$E1,$BF,$D3,$F8,$7D,$68,$F7,$1D,$4F,$B0,$53,$FC,$FE
	.dc.b	$72,$3B,$CF,$67,$96,$7F,$E2,$97,$B7,$BF,$E1,$7D,$59,$7E,$23,$2F
	.dc.b	$96,$E7,$26,$BF,$43,$0F,$B9,$1E,$BB,$F1,$9E,$03,$BD,$FC,$1C,$FF
	.dc.b	$CB,$91,$83,$CD,$1F,$21,$FC,$29,$93,$E6,$CF,$57,$E0,$9D,$4F,$79
	.dc.b	$F4,$A7,$AB,$88,$FC,$9E,$1E,$B8,$F8,$D7,$F0,$5E,$B9,$7C,$57,$86
	.dc.b	$3F,$E7,$2C,$B8,$47,$F2,$E3,$BE,$5E,$73,$D1,$8F,$C4,$CF,$D0,$F9
	.dc.b	$F1,$F8,$B8,$FC,$9C,$7A,$61,$FE,$32,$3D,$AF,$CB,$C3,$E7,$C7,$CA
	.dc.b	$7A,$C6,$63,$EE,$5F,$59,$FE,$00,$D8,$CC,$F6,$7F,$C8,$07,$B3,$0E
	.dc.b	$99,$7B,$2F,$AC,$7A,$30,$F8,$58,$FC,$5C,$7D,$98,$72,$8F,$EA,$11
	.dc.b	$87,$E5,$5F,$C3,$C3,$DD,$87,$E0,$AB,$C9,$7F,$5E,$FA,$63,$E5,$79
	.dc.b	$BE,$D8,$FC,$89,$E6,$3E,$5B,$99,$E6,$47,$07,$3C,$3F,$CE,$B3,$F0
	.dc.b	$AF,$93,$F8,$88,$F7,$BF,$85,$C7,$DF,$1F,$B5,$BF,$63,$CA,$7F,$01
	.dc.b	$87,$BB,$1F,$90,$7B,$A3,$F2,$75,$E4,$BC,$57,$F7,$5C,$FF,$73,$5F
	.dc.b	$88,$BE,$2F,$B3,$1F,$8D,$7E,$28,$F4,$1F,$3F,$1F,$16,$3E,$48,$FD
	.dc.b	$3D,$F2,$7E,$E1,$E2,$AE,$4F,$E3,$2F,$F1,$D8,$78,$B1,$F5,$4F,$C4
	.dc.b	$7E,$9C,$F8,$A2,$1D,$23,$D5,$9F,$B3,$2F,$CA,$CF,$CB,$C7,$0F,$F9
	.dc.b	$AB,$2F,$84,$FB,$4F,$51,$FC,$9D,$F5,$3F,$78,$E4,$6E,$FF,$4E,$F8
	.dc.b	$9F,$2B,$C8,$FA,$75,$FC,$AC,$FC,$83,$C5,$FD,$1E,$5F,$0A,$BB,$9E
	.dc.b	$0C,$BF,$82,$DF,$43,$FA,$2E,$1F,$2D,$C5,$F2,$C3,$C4,$FE,$17,$3F
	.dc.b	$C7,$47,$AD,$FC,$8E,$7F,$2E,$7E,$2D,$72,$AF,$89,$97,$8E,$F7,$1E
	.dc.b	$BC,$7E,$A5,$FA,$F1,$F1,$3D,$F1,$F0,$D7,$23,$DC,$7F,$38,$29,$FC
	.dc.b	$A0,$F9,$67,$D9,$8F,$53,$F8,$53,$75,$F0,$A3,$B4,$70,$C7,$E1,$BD
	.dc.b	$32,$FB,$6F,$33,$C0,$53,$C1,$1C,$1C,$47,$08,$F8,$50,$AF,$03,$F9
	.dc.b	$E8,$E0,$4A,$F1,$BF,$82,$8F,$41,$DE,$78,$BF,$9D,$3D,$13,$F4,$0E
	.dc.b	$27,$C0,$9E,$47,$8A,$78,$D7,$E1,$31,$FE,$21,$5F,$8A,$BF,$4E,$5F
	.dc.b	$5F,$1F,$EA,$58,$70,$BF,$B3,$8F,$9E,$FF,$A5,$E1,$E0,$7E,$B1,$E8
	.dc.b	$8F,$06,$7F,$D2,$AF,$EF,$CF,$EE,$8F,$C4,$1F,$0E,$BD,$04,$70,$7D
	.dc.b	$0A,$F9,$71,$E7,$7E,$6E,$5F,$3E,$3C,$53,$F3,$EF,$C0,$7E,$D9,$FA
	.dc.b	$F1,$F1,$DE,$A7,$F3,$57,$C0,$FC,$DA,$FE,$02,$FE,$3A,$3C,$07,$6A
	.dc.b	$F0,$47,$2B,$F8,$D8,$FF,$79,$AD,$CF,$6A,$DB,$D6,$3F,$97,$57,$C4
	.dc.b	$8F,$E0,$97,$F3,$5D,$1F,$31,$FB,$73,$EE,$9E,$F3,$FC,$B4,$F9,$67
	.dc.b	$03,$D8,$71,$8E,$18,$7C,$48,$F6,$67,$FB,$07,$CD,$9F,$E0,$5F,$F0
	.dc.b	$6D,$F6,$C3,$F9,$4C,$7E,$40,$D9,$FB,$63,$C1,$1B,$C3,$F7,$91,$E4
	.dc.b	$C7,$F2,$19,$6D,$FE,$B4,$FE,$3B,$2F,$24,$FE,$B6,$BE,$4B,$EB,$AF
	.dc.b	$92,$72,$7F,$85,$4F,$DD,$9D,$CE,$70,$F8,$F5,$F0,$4F,$AD,$5D,$A3
	.dc.b	$F1,$79,$7C,$19,$ED,$97,$C1,$9F,$F1,$36,$5E,$B9,$FC,$C1,$F0,$4D
	.dc.b	$9E,$BC,$FF,$4B,$97,$DC,$C7,$C9,$8F,$5C,$BC,$93,$F8,$E7,$E0,$DF
	.dc.b	$CE,$AF,$24,$FE,$68,$F5,$DF,$49,$F5,$1F,$70,$F9,$F3,$C1,$FC,$DC
	.dc.b	$F0,$CF,$FC,$39,$1E,$57,$E7,$3F,$E3,$1B,$F1,$BF,$BA,$21,$EC,$3F
	.dc.b	$75,$3F,$88,$78,$5F,$8D,$FD,$59,$C2,$79,$BE,$57,$DB,$97,$CA,$C3
	.dc.b	$CD,$3F,$95,$BF,$33,$FE,$13,$3C,$C7,$C3,$3B,$E1,$FA,$BC,$7E,$56
	.dc.b	$3E,$1C,$7F,$06,$F8,$F2,$FA,$D7,$FA,$AC,$3F,$7C,$FA,$F2,$FC,$1B
	.dc.b	$F2,$0E,$A7,$EA,$4F,$B9,$7E,$A3,$AE,$3F,$22,$3F,$B9,$47,$BA,$BD
	.dc.b	$35,$FE,$47,$CF,$F1,$11,$FE,$7F,$CF,$DF,$5B,$CF,$D3,$8F,$D4,$C7
	.dc.b	$E6,$9F,$94,$3C,$78,$FE,$B8,$FE,$05,$5E,$3C,$7A,$CF,$FA,$77,$2F
	.dc.b	$7C,$FF,$9F,$A3,$F5,$D9,$FC,$5A,$DC,$7F,$A1,$72,$F1,$CF,$D9,$CF
	.dc.b	$D2,$7F,$90,$6B,$77,$FE,$07,$AD,$E5,$F4,$70,$EA,$FA,$72,$F9,$75
	.dc.b	$F8,$0C,$BC,$79,$7F,$4C,$CF,$DF,$87,$DD,$3D,$E7,$D1,$3E,$5C,$78
	.dc.b	$E3,$77,$F2,$E7,$F0,$6F,$E2,$5F,$8B,$9F,$F0,$C7,$EB,$D7,$BD,$DC
	.dc.b	$7D,$17,$67,$EE,$27,$E5,$CE,$E3,$C7,$87,$A7,$2F,$7C,$7E,$0E,$B7
	.dc.b	$97,$BF,$0F,$DC,$61,$F8,$FB,$DD,$7D,$79,$FC,$7E,$1F,$5F,$0F,$19
	.dc.b	$F5,$EB,$73,$E9,$AE,$B3,$F1,$5F,$97,$7D,$63,$D2,$7E,$6C,$FD,$73
	.dc.b	$E9,$8D,$E1,$F9,$57,$FD,$16,$FD,$79,$F8,$AF,$A5,$FE,$13,$87,$BD
	.dc.b	$EA,$FE,$25,$D9,$E3,$BF,$7C,$F5,$7F,$CE,$77,$F5,$E7,$F0,$66,$CE
	.dc.b	$AF,$A5,$FB,$B3,$E3,$BF,$79,$F7,$63,$E8,$CE,$E3,$D3,$1F,$46,$BE
	.dc.b	$5C,$FB,$DF,$1C,$7F,$05,$3A,$9B,$3E,$2D,$7B,$DF,$E2,$53,$D6,$7D
	.dc.b	$33,$E3,$7A,$BB,$BF,$E2,$F1,$F8,$9A,$EB,$7F,$75,$EA,$7B,$CE,$B5
	.dc.b	$F1,$5F,$F3,$23,$D5,$DC,$7B,$CF,$AE,$FA,$5F,$79,$D5,$FE,$56,$7F
	.dc.b	$17,$9F,$7C,$7E,$86,$3C,$71,$F8,$93,$D2,$F8,$EB,$D2,$FE,$3C,$FB
	.dc.b	$B7,$FB,$37,$EB,$D7,$8E,$BA,$9E,$F7,$C6,$F5,$8F,$4B,$FC,$15,$FC
	.dc.b	$D9,$FC,$2C,$EB,$7F,$E1,$08,$FF,$66,$CF,$59,$FC,$1C,$FC,$57,$FD
	.dc.b	$32,$7E,$CE,$7D,$33,$B8,$F9,$67,$C5,$8F,$AE,$FD,$D3,$E2,$DF,$EB
	.dc.b	$5F,$C4,$D7,$5C,$BF,$C5,$53,$F9,$BB,$FC,$7D,$FA,$6F,$67,$53,$EB
	.dc.b	$9D,$5F,$49,$B8,$FA,$E7,$D1,$3E,$2C,$7A,$4F,$8A,$7F,$B0,$70,$FF
	.dc.b	$7B,$67,$FC,$A9,$FF,$53,$C7,$C0,$3F,$74,$7A,$9F,$CE,$3F,$18,$F2
	.dc.b	$1B,$BF,$55,$FC,$63,$C8,$7D,$88,$FB,$C7,$E1,$0F,$21,$D7,$1F,$CE
	.dc.b	$5F,$F9,$87,$1F,$E6,$B1,$F5,$F3,$FF,$32,$61,$FD,$7A,$FC,$35,$F2
	.dc.b	$B1,$FC,$44,$7F,$04,$CB,$EB,$57,$E3,$A3,$F0,$47,$84,$F6,$E5,$F2
	.dc.b	$9F,$89,$8F,$F8,$FF,$2F,$16,$7F,$C8,$A7,$CF,$5B,$57,$FE,$D8,$AF
	.dc.b	$95,$7F,$EF,$AB,$FF,$15,$D7,$E6,$AF,$E1,$BF,$27,$0F,$AD,$1F,$88
	.dc.b	$7F,$58,$F5,$C3,$A4,$FE,$69,$FE,$BC,$FE,$68,$FF,$1E,$46,$F1,$FB
	.dc.b	$71,$B7,$ED,$E5,$F1,$71,$F9,$53,$FE,$C7,$CB,$CE,$FF,$95,$0F,$F3
	.dc.b	$69,$E7,$3E,$19,$FD,$41,$F8,$6F,$87,$1F,$F4,$06,$1E,$1A,$F9,$13
	.dc.b	$FE,$D8,$CB,$E4,$3F,$3B,$3F,$B3,$87,$83,$0F,$F4,$BC,$FF,$9A,$67
	.dc.b	$FB,$5E,$1F,$1A,$3F,$5B,$3F,$19,$FD,$0B,$F1,$A7,$9C,$7B,$F0,$FA
	.dc.b	$0F,$A8,$FC,$15,$73,$CF,$F0,$47,$CA,$8F,$C1,$3F,$98,$7F,$4A,$7D
	.dc.b	$CC,$FE,$5B,$FB,$2C,$3C,$4F,$E7,$E3,$F1,$F1,$FB,$88,$F8,$EF,$F8
	.dc.b	$3B,$1F,$8F,$1D,$A7,$77,$FA,$58,$DB,$FE,$6D,$9D,$D7,$EE,$27,$E4
	.dc.b	$1F,$5C,$F7,$1F,$5C,$F9,$07,$D7,$3D,$18,$7D,$27,$FD,$E5,$97,$D2
	.dc.b	$C7,$DD,$1F,$AF,$3F,$99,$5F,$83,$0F,$C9,$1F,$CA,$E3,$58,$7E,$AA
	.dc.b	$75,$1F,$4A,$74,$7E,$05,$FC,$31,$F8,$13,$F0,$CF,$F9,$70,$D4,$FC
	.dc.b	$58,$D1,$F1,$70,$FC,$AB,$FE,$FC,$3D,$F1,$F8,$63,$FD,$42,$FB,$EF
	.dc.b	$F4,$AF,$D5,$AF,$E8,$87,$A9,$FD,$BE,$5F,$86,$74,$7F,$12,$AF,$9D
	.dc.b	$5F,$12,$BC,$A7,$D3,$9F,$2C,$EB,$2F,$B3,$1F,$86,$AF,$0C,$FF,$30
	.dc.b	$AF,$0C,$FC,$53,$C2,$7D,$93,$C2,$7F,$12,$8E,$98,$7C,$05,$3F,$EF
	.dc.b	$7C,$7F,$4C,$FF,$AD,$6B,$F3,$93,$F8,$F7,$EF,$9F,$40,$FB,$E7,$D0
	.dc.b	$3F,$D4,$B1,$EC,$7F,$C5,$73,$F6,$9F,$CD,$57,$86,$7C,$11,$F4,$1F
	.dc.b	$03,$F8,$E3,$F5,$67,$E6,$0E,$67,$C6,$7E,$83,$F9,$83,$ED,$1F,$02
	.dc.b	$FF,$61,$7F,$02,$3F,$8C,$E3,$F5,$63,$E5,$47,$D5,$AE,$75,$F5,$6B
	.dc.b	$51,$FD,$2E,$7E,$53,$AC,$7E,$AC,$7E,$62,$3F,$80,$63,$EF,$CF,$F9
	.dc.b	$14,$F8,$0F,$A9,$8F,$81,$E6,$7D,$93,$D4,$FD,$47,$E0,$1F,$41,$FC
	.dc.b	$D3,$F4,$0F,$CD,$1D,$4F,$82,$F5,$C7,$E0,$DF,$9F,$2F,$A1,$5D,$27
	.dc.b	$F1,$73,$ED,$77,$1E,$37,$E1,$3E,$88,$D6,$3F,$CF,$B3,$FE,$9F,$3F
	.dc.b	$58,$F3,$4F,$E6,$DF,$36,$1F,$44,$FE,$73,$9F,$CC,$8F,$EA,$93,$F9
	.dc.b	$A7,$F0,$A7,$AE,$7E,$13,$FE,$3F,$36,$7D,$73,$CD,$7F,$4A,$7F,$21
	.dc.b	$1F,$EE,$DC,$FE,$96,$1E,$6B,$FC,$D6,$3F,$02,$B9,$9F,$48,$F0,$C7
	.dc.b	$99,$F8,$06,$A3,$E2,$BA,$BF,$E1,$D8,$FA,$E3,$FA,$4C,$7A,$F3,$FC
	.dc.b	$0D,$F3,$9F,$AB,$3F,$2B,$3F,$E9,$31,$F9,$F9,$F2,$E7,$FD,$09,$FB
	.dc.b	$4F,$CF,$8F,$D9,$D7,$F3,$B7,$E7,$9F,$A4,$3E,$7B,$F7,$DF,$94,$FE
	.dc.b	$B9,$FD,$24,$73,$7C,$11,$CF,$1F,$89,$1E,$A3,$E7,$1E,$F3,$F7,$B8
	.dc.b	$FB,$E3,$FA,$ED,$73,$BF,$E4,$F5,$F8,$39,$E6,$6B,$0F,$7C,$E8,$F7
	.dc.b	$BF,$D2,$5F,$BF,$1F,$EB,$19,$FC,$EC,$7C,$93,$35,$FE,$44,$8F,$31
	.dc.b	$F6,$71,$F9,$F8,$FD,$48,$FB,$26,$AF,$EC,$CE,$A3,$D6,$6A,$3C,$CF
	.dc.b	$F5,$FB,$F3,$57,$C5,$BF,$A8,$FD,$99,$D2,$7F,$59,$1F,$93,$C7,$E3
	.dc.b	$67,$F9,$29,$FB,$27,$F3,$B9,$E6,$FE,$35,$F9,$55,$CE,$3E,$E9,$FC
	.dc.b	$11,$FB,$57,$F8,$4C,$BE,$5C,$6F,$15,$FF,$B4,$6F,$E9,$3D,$9F,$84
	.dc.b	$F6,$7C,$F5,$FC,$11,$FD,$71,$A7,$AB,$A3,$F1,$65,$BF,$E2,$67,$D4
	.dc.b	$8F,$F4,$76,$7F,$AF,$C3,$E4,$47,$DD,$7C,$D7,$FE,$CF,$BF,$DC,$CF
	.dc.b	$87,$0F,$F4,$C6,$1E,$D9,$F9,$A7,$7B,$FB,$8F,$EA,$8F,$91,$5E,$67
	.dc.b	$E3,$D7,$F1,$87,$47,$AC,$E6,$6A,$FE,$B3,$F8,$1C,$7E,$B3,$FA,$43
	.dc.b	$EB,$57,$A5,$F1,$1F,$7F,$3F,$F4,$3B,$F1,$0F,$93,$7F,$8A,$C7,$FC
	.dc.b	$15,$8F,$C0,$BF,$F7,$05,$7F,$47,$3F,$99,$61,$FD,$C1,$E9,$97,$F8
	.dc.b	$A7,$2F,$2C,$7A,$B1,$E9,$87,$F6,$6C,$7F,$2B,$3E,$63,$FB,$A3,$F4
	.dc.b	$DF,$F4,$BD,$FD,$37,$51,$F6,$27,$C9,$1A,$3F,$2A,$79,$4F,$97,$87
	.dc.b	$95,$65,$F0,$2B,$F6,$95,$F9,$43,$F4,$8F,$C0,$9F,$8E,$7F,$BE,$23
	.dc.b	$F9,$AE,$1F,$EA,$B9,$FF,$2E,$D7,$EC,$EB,$59,$FE,$A1,$FE,$FC,$7E
	.dc.b	$18,$F9,$D5,$F8,$6B,$EE,$7C,$33,$D7,$7F,$E7,$79,$FB,$87,$23,$F8
	.dc.b	$D1,$D8,$C9,$FC,$A0,$EC,$FB,$1F,$4C,$F9,$E7,$F9,$E6,$3F,$2E,$7E
	.dc.b	$34,$FF,$1E,$C7,$F1,$37,$F5,$27,$39,$7E,$C6,$BE,$CC,$7D,$E3,$FC
	.dc.b	$50,$78,$CF,$BD,$5C,$B0,$F5,$9F,$E6,$0C,$7E,$24,$F9,$CF,$5C,$23
	.dc.b	$FD,$37,$1F,$8C,$3F,$C5,$87,$73,$F9,$5C,$7E,$AB,$0F,$C3,$1F,$E6
	.dc.b	$F8,$F8,$D9,$7D,$39,$E9,$1F,$E6,$1C,$7D,$58,$79,$63,$EE,$CE,$9F
	.dc.b	$E1,$F9,$FE,$9D,$FD,$73,$FB,$5C,$BF,$8A,$BF,$A2,$9F,$C1,$D7,$E3
	.dc.b	$70,$FC,$64,$79,$4F,$C6,$5F,$EF,$8F,$DC,$DF,$25,$8F,$DF,$7E,$49
	.dc.b	$EA,$CF,$EB,$57,$68,$F8,$67,$E6,$A3,$F5,$D8,$FF,$11,$CB,$FA,$66
	.dc.b	$1F,$5A,$B4,$7D,$6B,$F7,$61,$FD,$62,$3F,$5F,$1F,$59,$FE,$CF,$8F
	.dc.b	$C9,$CF,$E9,$C7,$EA,$8F,$0C,$7B,$9F,$EA,$27,$A0,$FB,$8F,$17,$E8
	.dc.b	$63,$F2,$F0,$FE,$C9,$87,$F8,$AF,$0F,$8D,$5F,$96,$CF,$E5,$67,$EC
	.dc.b	$8F,$D3,$E7,$EC,$9F,$B1,$1F,$50,$CE,$3F,$D6,$70,$FC,$14,$78,$AF
	.dc.b	$E4,$E7,$F9,$09,$FC,$3E,$3F,$A2,$AF,$E0,$EF,$CD,$7F,$C1,$B7,$F7
	.dc.b	$F1,$FE,$A5,$5F,$1C,$FE,$21,$3F,$C1,$33,$F8,$4F,$E8,$63,$AB,$F9
	.dc.b	$93,$99,$F8,$09,$E7,$7F,$D3,$23,$E8,$9C,$F1,$F8,$15,$FB,$98,$FE
	.dc.b	$67,$97,$C0,$CF,$F1,$13,$F8,$73,$BC,$FD,$53,$BB,$FC,$65,$F8,$CF
	.dc.b	$38,$FE,$21,$1F,$93,$8D,$47,$4B,$FD,$B4,$FF,$0C,$9E,$B3,$FE,$10
	.dc.b	$BF,$97,$39,$AF,$C4,$3F,$E1,$98,$CB,$F8,$3A,$E3,$87,$F7,$78,$F1
	.dc.b	$D7,$E0,$1F,$69,$F9,$F3,$E9,$E7,$F5,$E3,$AE,$1E,$19,$C9,$E1,$3E
	.dc.b	$41,$F9,$D9,$FC,$BC,$7F,$1B,$8E,$87,$F6,$33,$F8,$DE,$7F,$93,$BF
	.dc.b	$EF,$51,$FC,$4A,$B8,$C7,$F0,$68,$E3,$87,$F2,$3C,$3F,$0C,$FF,$7D
	.dc.b	$3E,$4C,$FD,$C9,$FE,$AB,$3F,$E8,$99,$F8,$37,$F8,$B3,$F8,$45,$FB
	.dc.b	$6B,$D2,$7F,$1C,$3F,$57,$3C,$1F,$C2,$9F,$20,$FE,$44,$FE,$A5,$DD
	.dc.b	$7F,$0E,$7F,$C0,$D1,$E4,$AE,$F1,$F1,$CE,$47,$F0,$43,$C0,$7E,$66
	.dc.b	$79,$61,$FC,$27,$3F,$CE,$C7,$F0,$83,$F3,$17,$FA,$2A,$E7,$3A,$C3
	.dc.b	$E4,$E1,$F3,$4F,$F1,$5C,$FD,$83,$F0,$97,$F7,$31,$FC,$26,$3F,$91
	.dc.b	$9F,$F8,$02,$68,$F8,$B8,$FB,$DF,$31,$FC,$1C,$FE,$27,$97,$EA,$2F
	.dc.b	$F1,$95,$F9,$E9,$F8,$71,$FD,$B5,$FE,$7F,$8F,$99,$F9,$B3,$FC,$52
	.dc.b	$7F,$D2,$17,$FA,$7C,$3F,$88,$57,$E0,$A7,$D1,$8F,$F2,$F3,$F8,$83
	.dc.b	$FE,$E8,$8D,$1D,$8F,$E4,$27,$69,$FC,$7E,$5F,$C9,$63,$FA,$79,$9A
	.dc.b	$F8,$EF,$E5,$EB,$F1,$E7,$F8,$19,$FC,$0B,$E6,$3D,$F5,$E6,$8F,$E4
	.dc.b	$07,$CF,$BF,$C5,$1F,$5F,$0F,$74,$7F,$81,$AB,$8D,$FA,$A3,$F8,$C5
	.dc.b	$7E,$76,$7E,$44,$7E,$21,$FC,$13,$F9,$E3,$FD,$71,$1D,$A3,$75,$F6
	.dc.b	$E7,$FD,$00,$7D,$73,$F8,$AC,$6A,$FF,$51,$1F,$B4,$3E,$11,$FE,$3B
	.dc.b	$3E,$16,$3F,$A7,$7F,$1F,$97,$DF,$7F,$42,$75,$C3,$D0,$FC,$63,$DC
	.dc.b	$68,$FE,$1E,$7C,$63,$A9,$FA,$5C,$3E,$2B,$FC,$0C,$FB,$31,$F2,$6F
	.dc.b	$EC,$D7,$F1,$E8,$F8,$13,$F0,$B0,$FC,$2B,$FC,$FF,$1F,$D8,$BF,$9B
	.dc.b	$3D,$27,$F1,$73,$AC,$7F,$07,$38,$CF,$EE,$63,$FC,$0B,$1F,$CD,$1E
	.dc.b	$58,$79,$B2,$FC,$55,$7F,$55,$AE,$55,$FA,$87,$F0,$CF,$C1,$AF,$D2
	.dc.b	$1F,$BD,$C3,$E5,$47,$EB,$0F,$8B,$5E,$C7,$F8,$33,$FB,$1A,$FC,$0C
	.dc.b	$7D,$78,$F8,$98,$7E,$61,$FE,$4D,$97,$D7,$C3,$CD,$87,$C7,$C3,$F8
	.dc.b	$6E,$7F,$D6,$F3,$FC,$0C,$7E,$7B,$2F,$EA,$8F,$E1,$72,$F9,$39,$78
	.dc.b	$F1,$FF,$2D,$57,$F8,$5B,$1E,$8F,$E0,$6F,$F4,$CF,$A1,$F8,$2F,$B6
	.dc.b	$7E,$0E,$3E,$1C,$3A,$D7,$92,$BF,$72,$7E,$7F,$2F,$E0,$D8,$FB,$A3
	.dc.b	$F2,$8F,$B9,$FC,$46,$5F,$84,$7F,$61,$5C,$DF,$C7,$4E,$DF,$B3,$3B
	.dc.b	$7F,$22,$62,$FE,$55,$7E,$08,$FD,$AC,$7E,$08,$FD,$AB,$EE,$CF,$F9
	.dc.b	$84,$7E,$28,$FC,$91,$F8,$13,$F4,$27,$C1,$AF,$E0,$D7,$F7,$67,$8C
	.dc.b	$FF,$12,$8F,$71,$FE,$9C,$CF,$E6,$C7,$E1,$2F,$C8,$FE,$B4,$FD,$31
	.dc.b	$C5,$FC,$8B,$F9,$E3,$FA,$6B,$F8,$83,$EF,$3F,$CD,$0E,$55,$F9,$A8
	.dc.b	$FE,$9A,$72,$CF,$F1,$33,$F3,$0F,$BC,$7E,$18,$E3,$5E,$68,$FB,$B1
	.dc.b	$F6,$9F,$E2,$47,$FB,$AE,$FF,$15,$9F,$F9,$3A,$7F,$64,$7A,$4F,$92
	.dc.b	$7A,$63,$F4,$19,$FE,$0B,$0F,$26,$5F,$66,$7E,$E1,$D2,$BF,$A9,$9D
	.dc.b	$1F,$DA,$3D,$30,$F5,$BF,$80,$8E,$0F,$F2,$43,$D6,$7B,$CF,$A0,$7D
	.dc.b	$B3,$E7,$BD,$14,$7E,$91,$FF,$19,$1C,$5F,$86,$FE,$4A,$BE,$CB,$F8
	.dc.b	$3C,$FF,$A0,$3F,$E4,$9A,$FE,$52,$FE,$A4,$FC,$4C,$FD,$9C,$3F,$79
	.dc.b	$87,$C1,$8E,$66,$67,$F4,$77,$F9,$A7,$FB,$11,$F3,$4F,$3E,$3F,$45
	.dc.b	$FC,$2D,$FC,$D7,$E3,$5F,$57,$8E,$1E,$FC,$7D,$C7,$F9,$0E,$3E,$56
	.dc.b	$3F,$AF,$9F,$CE,$65,$FC,$72,$3C,$93,$FA,$D3,$F7,$2F,$A3,$0F,$04
	.dc.b	$79,$0F,$B2,$79,$0F,$D6,$19,$3E,$B1,$F1,$4F,$A7,$1F,$98,$7D,$39
	.dc.b	$7C,$17,$F8,$51,$CB,$0F,$4E,$3F,$E2,$03,$F9,$0D,$FC,$CA,$E1,$3F
	.dc.b	$98,$94,$7F,$83,$EB,$F8,$11,$F5,$EF,$F8,$16,$1E,$FC,$3D,$78,$76
	.dc.b	$BF,$F2,$85,$F0,$8F,$D1,$5F,$C7,$3F,$07,$9F,$F0,$2C,$BE,$76,$7F
	.dc.b	$55,$FD,$0D,$FA,$63,$15,$FC,$B0,$F9,$B3,$F9,$A3,$F1,$35,$CE,$3E
	.dc.b	$25,$7C,$99,$CE,$5E,$78,$E8,$7D,$08,$EE,$FE,$1F,$0F,$53,$FD,$AC
	.dc.b	$F5,$67,$F1,$1F,$F6,$09,$8C,$BF,$CA,$0F,$DE,$3F,$13,$87,$8A,$B9
	.dc.b	$DF,$F6,$D3,$FD,$55,$1F,$0E,$7F,$5D,$8F,$F1,$37,$FC,$3C,$7A,$63
	.dc.b	$E7,$BF,$BD,$7A,$65,$F1,$23,$F4,$F9,$FD,$28,$C4,$7E,$18,$E7,$87
	.dc.b	$8A,$BF,$AF,$47,$A4,$F5,$E1,$FE,$DA,$3F,$36,$F7,$38,$D7,$4C,$3C
	.dc.b	$D1,$FD,$CE,$BF,$8A,$E3,$F8,$AC,$BF,$1B,$3F,$B4,$3E,$21,$F9,$43
	.dc.b	$E7,$5F,$3C,$FF,$95,$4F,$F4,$6B,$FC,$2C,$7D,$7A,$D9,$F5,$E3,$F0
	.dc.b	$59,$FB,$E3,$07,$C2,$3E,$E9,$FA,$8C,$7F,$08,$FA,$8D,$9F,$14,$F5
	.dc.b	$1F,$AA,$3C,$F9,$FE,$8F,$2F,$D1,$67,$F1,$70,$F2,$D7,$C5,$38,$9F
	.dc.b	$5C,$FE,$8C,$7E,$DE,$FE,$3C,$7F,$52,$BF,$F1,$14,$F1,$8C,$67,$FE
	.dc.b	$05,$8F,$A9,$7F,$E8,$3C,$3F,$93,$DF,$FB,$36,$7F,$96,$1F,$8E,$3D
	.dc.b	$87,$BF,$2F,$91,$5F,$AC,$8F,$7E,$3E,$2C,$BF,$80,$47,$BB,$3F,$AE
	.dc.b	$FF,$17,$AF,$EE,$4F,$F3,$B8,$FA,$8F,$E5,$AF,$EB,$E3,$E4,$9F,$F6
	.dc.b	$C4,$F9,$0F,$BD,$5E,$17,$C8,$7B,$EF,$C3,$1D,$A3,$F9,$59,$FD,$B8
	.dc.b	$FE,$EC,$7C,$9B,$EE,$FC,$4C,$FF,$1E,$FA,$70,$FF,$15,$BC,$72,$F9
	.dc.b	$AF,$F2,$B3,$F1,$27,$88,$C5,$F9,$4F,$E0,$AF,$94,$FC,$C3,$E5,$3F
	.dc.b	$B7,$D7,$ED,$E3,$FB,$B3,$F9,$27,$F1,$F1,$C9,$EB,$1F,$B6,$7E,$8D
	.dc.b	$FA,$A7,$F1,$51,$F2,$CF,$9D,$7F,$2C,$30,$96,$C7,$F8,$94,$64,$FE
	.dc.b	$E6,$E1,$DB,$5E,$71,$C6,$78,$46,$27,$4F,$28,$CB,$4D,$CF,$16,$CF
	.dc.b	$08,$C4,$78,$DA,$FA,$3D,$F7,$DD,$74,$94,$62,$7A,$B5,$E4,$D7,$94
	.dc.b	$F1,$8C,$47,$B1,$AF,$F8,$51,$B3,$DA,$78,$12,$7B,$BF,$27,$F0,$F1
	.dc.b	$00

;--------------------------------------------------------------------------------
;	FSMOVE.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsmove_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FSMOVE.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSMOVE.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsmove_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsmove.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSMOVE.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsmove.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSMOVE.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsmove_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$BB,$F6,$7F,$79,$A5,$3F,$9A,$B5,$A5,$70,$9A,$47
	.dc.b	$44,$3A,$39,$68,$FA,$EF,$F3,$0F,$DC,$7F,$5C,$F8,$FF,$E8,$B1,$FE
	.dc.b	$64,$9F,$0B,$AF,$5F,$FC,$26,$E6,$3C,$47,$3F,$53,$F9,$CD,$F3,$8C
	.dc.b	$1E,$E9,$E5,$7D,$AB,$E1,$61,$6B,$FA,$AD,$F1,$BE,$D3,$99,$4E,$6B
	.dc.b	$8C,$76,$33,$87,$9A,$33,$7E,$09,$D1,$EC,$8F,$3C,$EE,$3C,$D1,$8C
	.dc.b	$3C,$11,$B3,$CB,$3B,$7F,$AB,$E1,$E5,$C3,$C9,$1F,$73,$E2,$3C,$07
	.dc.b	$A9,$FE,$B5,$87,$56,$9B,$71,$1B,$38,$BE,$E6,$C6,$CF,$1E,$1F,$DF
	.dc.b	$23,$B6,$3E,$F2,$D7,$76,$D7,$03,$E1,$9F,$69,$E1,$1F,$20,$E0,$58
	.dc.b	$E1,$1C,$9F,$9F,$5F,$19,$E5,$3C,$1F,$96,$70,$79,$3C,$0F,$0E,$1F
	.dc.b	$27,$1F,$99,$1F,$3C,$D5,$F7,$C7,$E6,$C7,$D0,$3F,$00,$E1,$E0,$D7
	.dc.b	$C6,$78,$4F,$19,$D2,$FC,$71,$BB,$F1,$61,$EB,$6D,$79,$8E,$F8,$79
	.dc.b	$9E,$A7,$33,$AD,$73,$77,$5C,$DA,$73,$C3,$E7,$BC,$2B,$A5,$F8,$27
	.dc.b	$E0,$46,$27,$8C,$62,$BA,$3C,$A3,$E2,$35,$F8,$73,$C6,$31,$1E,$8B
	.dc.b	$F3,$46,$2F,$C1,$87,$C4,$BE,$0F,$DB,$BE,$AD,$7A,$E1,$E2,$8C,$47
	.dc.b	$59,$E2,$FD,$13,$89,$4E,$27,$C3,$8E,$35,$E2,$8F,$89,$5D,$5E,$07
	.dc.b	$57,$0F,$D1,$3A,$94,$EB,$7F,$1D,$EC,$72,$9E,$CF,$0C,$3E,$1D,$F8
	.dc.b	$A3,$E7,$BF,$0E,$FB,$CE,$A7,$91,$A3,$CE,$76,$3C,$47,$9C,$C3,$C5
	.dc.b	$CD,$78,$9F,$3D,$77,$77,$87,$CA,$76,$77,$3D,$0E,$6B,$A1,$1C,$12
	.dc.b	$FD,$2D,$9E,$B1,$82,$38,$26,$1F,$24,$97,$E5,$6C,$7C,$06,$BE,$49
	.dc.b	$E7,$18,$8F,$69,$B9,$E4,$47,$04,$39,$1E,$27,$4E,$23,$A3,$F3,$5B
	.dc.b	$87,$CB,$8C,$4F,$07,$C2,$71,$70,$E8,$F0,$1A,$21,$93,$81,$A3,$06
	.dc.b	$A3,$C0,$47,$EB,$19,$29,$C0,$8E,$E7,$F6,$51,$A8,$C4,$FE,$4E,$72
	.dc.b	$47,$04,$9F,$C9,$C6,$5B,$19,$7F,$13,$19,$6C,$67,$0F,$C4,$5F,$D8
	.dc.b	$8F,$8B,$7E,$2A,$E6,$43,$89,$6B,$8B,$F5,$CF,$8D,$1C,$5D,$92,$37
	.dc.b	$7F,$62,$3E,$01,$1F,$34,$6A,$BC,$73,$97,$D1,$87,$E6,$EB,$91,$C1
	.dc.b	$C4,$70,$8C,$61,$ED,$AF,$29,$CC,$F5,$57,$07,$33,$C1,$F6,$1F,$82
	.dc.b	$21,$F7,$5D,$1E,$2A,$F2,$D7,$79,$D9,$F8,$03,$C0,$7C,$3A,$F0,$3A
	.dc.b	$C3,$E4,$61,$F0,$63,$C7,$7D,$1F,$19,$E1,$71,$87,$86,$3D,$78,$78
	.dc.b	$5D,$18,$38,$9E,$38,$F4,$CF,$C7,$39,$96,$38,$47,$57,$D1,$3D,$5E
	.dc.b	$53,$C1,$F3,$9C,$1E,$4F,$03,$C3,$7E,$EC,$3E,$0C,$79,$8D,$5F,$7C
	.dc.b	$3E,$14,$7C,$53,$E4,$B8,$78,$35,$F1,$9E,$13,$C6,$74,$BF,$1C,$6E
	.dc.b	$FC,$58,$7A,$DF,$DD,$63,$FE,$30,$BF,$F0,$83,$5F,$F3,$AB,$5F,$F3
	.dc.b	$1B,$63,$FC,$69,$7F,$E3,$C3,$FC,$74,$59,$F1,$DF,$85,$F5,$E3,$F0
	.dc.b	$9F,$49,$E7,$36,$E6,$BC,$F1,$F2,$8E,$38,$7A,$5A,$67,$1F,$BA,$F2
	.dc.b	$8D,$46,$23,$E6,$35,$F9,$75,$DD,$E7,$1F,$4E,$30,$EE,$3C,$F7,$EB
	.dc.b	$8C,$5F,$99,$E1,$7E,$0C,$3F,$0D,$7E,$06,$BE,$08,$D3,$F6,$4D,$14
	.dc.b	$D1,$ED,$8E,$2F,$8B,$0F,$D5,$E1,$FE,$93,$FB,$95,$FF,$9C,$5B,$5F
	.dc.b	$DE,$1A,$FF,$95,$5B,$3C,$67,$FB,$A4,$62,$B8,$B6,$14,$8E,$6F,$F8
	.dc.b	$79,$D3,$98,$F1,$1D,$0A,$FF,$9D,$63,$8B,$FE,$09,$78,$92,$38,$47
	.dc.b	$F9,$EC,$F0,$94,$FF,$47,$18,$3F,$CE,$05,$AF,$F5,$B3,$6B,$BB,$A3
	.dc.b	$89,$0E,$84,$34,$7F,$AE,$3D,$9E,$32,$69,$2B,$FD,$BD,$19,$6C,$66
	.dc.b	$7F,$DC,$93,$96,$C6,$5F,$43,$5F,$F6,$7F,$B3,$E0,$25,$76,$6C,$74
	.dc.b	$6B,$F0,$5A,$F2,$8E,$04,$7C,$7E,$CF,$B0,$91,$EA,$6C,$6A,$30,$47
	.dc.b	$04,$3E,$84,$70,$7E,$96,$5F,$20,$93,$F1,$9B,$1F,$11,$AF,$08,$D1
	.dc.b	$1D,$35,$FC,$1A,$FD,$F5,$EE,$7E,$62,$3F,$2E,$3F,$7E,$B5,$F9,$53
	.dc.b	$F4,$F5,$C2,$7F,$3B,$AF,$EA,$1E,$13,$CC,$DD,$D8,$C9,$C0,$EF,$5C
	.dc.b	$E7,$B4,$69,$B1,$9A,$ED,$7E,$47,$C1,$19,$AF,$4C,$E5,$EB,$7E,$18
	.dc.b	$C5,$F9,$27,$C4,$FF,$02,$8C,$C7,$C5,$9E,$13,$E9,$AE,$CE,$CC,$46
	.dc.b	$A7,$99,$C4,$B5,$E0,$96,$8F,$67,$E3,$46,$1F,$91,$EE,$BF,$32,$7C
	.dc.b	$0E,$AF,$86,$3E,$8C,$38,$36,$33,$1F,$5E,$F9,$57,$1C,$25,$ED,$EF
	.dc.b	$87,$68,$C1,$2F,$F6,$33,$FB,$A2,$47,$EE,$1B,$1A,$8C,$11,$C1,$0F
	.dc.b	$E1,$D1,$C3,$4F,$E4,$FF,$BA,$BD,$9F,$2D,$7E,$3E,$53,$9B,$F2,$47
	.dc.b	$0A,$DB,$63,$27,$E5,$27,$85,$7E,$64,$F8,$E7,$57,$33,$FC,$F2,$FF
	.dc.b	$30,$F8,$2F,$AD,$F3,$6D,$FF,$3B,$BE,$33,$FC,$BB,$3F,$A3,$7E,$3A
	.dc.b	$F0,$46,$27,$91,$E6,$8E,$57,$F7,$49,$1F,$94,$CD,$F0,$6F,$2F,$37
	.dc.b	$D2,$EE,$3F,$D6,$2F,$75,$8C,$3C,$95,$6D,$5A,$9F,$C4,$4E,$67,$F2
	.dc.b	$AF,$E4,$B0,$F0,$4E,$89,$3E,$17,$CD,$1A,$8C,$13,$0F,$EC,$F5,$CF
	.dc.b	$0E,$D8,$FD,$4A,$FE,$A1,$3F,$D1,$89,$87,$F4,$46,$C7,$F4,$26,$BC
	.dc.b	$23,$44,$74,$DB,$F8,$9A,$3E,$0B,$E1,$67,$FA,$D6,$C6,$6B,$DB,$5C
	.dc.b	$27,$24,$8C,$9F,$DF,$2B,$93,$8B,$F9,$B7,$CD,$F8,$51,$B8,$C5,$FC
	.dc.b	$48,$FB,$D1,$87,$EB,$98,$BF,$05,$72,$3A,$B8,$AE,$51,$FB,$FA,$E4
	.dc.b	$FF,$84,$31,$FC,$AB,$C2,$FC,$D8,$7A,$23,$A3,$6F,$FC,$3B,$EC,$7F
	.dc.b	$78,$25,$7F,$76,$6D,$7F,$74,$6B,$FE,$09,$6B,$FA,$36,$BE,$37,$FC
	.dc.b	$07,$EC,$49,$FF,$00,$B6,$7F,$BF,$B5,$FF,$0A,$35,$FD,$73,$5F,$73
	.dc.b	$5F,$E5,$BE,$C7,$F8,$40,$93,$FE,$0E,$6C,$7F,$83,$1A,$FF,$88,$5A
	.dc.b	$FE,$ED,$AF,$CB,$7F,$C4,$7F,$1D,$FE,$48,$9F,$F1,$54,$FF,$8F,$A3
	.dc.b	$0E,$1A,$FF,$8B,$DA,$FF,$8D,$1A,$F1,$8E,$A4,$3A,$A8,$DB,$63,$4F
	.dc.b	$77,$CB,$F2,$51,$E3,$28,$C4,$F9,$5A,$F9,$5A,$F9,$5A,$FB,$5A,$F8
	.dc.b	$FE,$37,$E1,$92,$BB,$B6,$3E,$13,$5F,$84,$D7,$E1,$35,$F9,$4F,$CE
	.dc.b	$F6,$A4,$FA,$9B,$3F,$35,$AF,$CD,$6B,$F3,$5A,$FD,$56,$BF,$2F,$DA
	.dc.b	$FB,$64,$9F,$8E,$D8,$FB,$4D,$7E,$D3,$5F,$B4,$D7,$F0,$4F,$E1,$BD
	.dc.b	$A9,$3F,$49,$B3,$F8,$56,$BF,$85,$6B,$F8,$56,$BF,$8D,$6B,$F8,$3F
	.dc.b	$3B,$FC,$13,$91,$FF,$0F,$6F,$86,$5F,$F8,$1E,$74,$FF,$85,$27,$35
	.dc.b	$FD,$96,$72,$DA,$E7,$5C,$5D,$EF,$FE,$A4,$AD,$BF,$E3,$4B,$CD,$7E
	.dc.b	$76,$AD,$FF,$8B,$66,$BE,$99,$E8,$78,$CA,$F8,$5B,$5D,$5A,$75,$2C
	.dc.b	$75,$6B,$D4,$E4,$7F,$93,$3A,$7F,$B0,$C6,$B0,$F4,$BC,$E7,$57,$E6
	.dc.b	$9C,$C7,$EF,$70,$F8,$6F,$4C,$3E,$14,$7B,$AF,$FC,$D1,$7C,$63,$18
	.dc.b	$78,$2F,$85,$77,$BE,$2F,$C1,$8C,$9D,$DF,$69,$D0,$F4,$57,$4B,$ED
	.dc.b	$1A,$BE,$D7,$D2,$F9,$9F,$0D,$F9,$4E,$30,$F6,$B9,$AF,$F3,$4D,$F4
	.dc.b	$8F,$1C,$78,$5E,$18,$78,$63,$C6,$78,$67,$85,$F8,$4E,$E5,$8F,$80
	.dc.b	$4A,$FB,$07,$47,$91,$D1,$F0,$E1,$EC,$8E,$11,$E7,$3E,$31,$F6,$1E
	.dc.b	$13,$DE,$3A,$B5,$E0,$71,$8F,$64,$67,$0F,$86,$7E,$02,$32,$78,$1C
	.dc.b	$E1,$F8,$29,$E0,$FC,$3A,$F5,$1E,$8C,$3E,$1D,$FB,$CE,$71,$E8,$AE
	.dc.b	$31,$F8,$16,$C7,$18,$F5,$9C,$70,$F8,$51,$C6,$3C,$97,$E3,$69,$F2
	.dc.b	$A3,$C6,$7F,$51,$8F,$EB,$24,$8F,$E8,$AD,$8F,$EB,$0D,$78,$46,$88
	.dc.b	$E9,$B5,$FE,$B4,$F0,$3F,$3E,$48,$FC,$A3,$67,$ED,$B6,$7E,$B4,$FF
	.dc.b	$B3,$E3,$0F,$BD,$B3,$EF,$3F,$D6,$CE,$1F,$B5,$68,$C6,$3F,$66,$31
	.dc.b	$5E,$22,$57,$89,$B7,$E2,$9D,$CE,$49,$19,$BE,$B8,$7F,$B2,$DB,$7F
	.dc.b	$EB,$E6,$D7,$F0,$F6,$D7,$C5,$9E,$E7,$B4,$B3,$ED,$7E,$11,$87,$DB
	.dc.b	$3C,$A3,$FD,$BA,$D8,$F8,$71,$BC,$3E,$04,$E6,$39,$CF,$A9,$F0,$34
	.dc.b	$F0,$61,$F0,$2F,$DD,$3C,$F0,$F7,$34,$F7,$16,$BD,$CD,$9F,$74,$75
	.dc.b	$3D,$31,$93,$D2,$57,$D2,$6F,$0F,$4C,$64,$F4,$CF,$92,$75,$5E,$46
	.dc.b	$CF,$91,$A7,$90,$B3,$E4,$10,$89,$F7,$8F,$03,$FD,$E6,$D9,$FF,$7C
	.dc.b	$F8,$38,$99,$1F,$EF,$19,$D5,$7F,$4D,$AC,$BF,$EC,$F8,$CD,$F8,$23
	.dc.b	$BD,$D8,$C5,$B2,$61,$CA,$78,$5B,$0E,$77,$E9,$8C,$61,$CE,$3D,$87
	.dc.b	$3B,$E4,$EE,$B8,$D7,$8A,$3D,$A7,$AE,$B7,$19,$3D,$91,$92,$3E,$D8
	.dc.b	$F0,$3F,$33,$0F,$4C,$EB,$0F,$4D,$F8,$2B,$CF,$3E,$07,$93,$B3,$CD
	.dc.b	$7D,$1F,$51,$98,$F2,$57,$49,$D5,$F8,$27,$99,$27,$55,$DD,$B1,$98
	.dc.b	$EE,$64,$AE,$4E,$EF,$CC,$9E,$D3,$C8,$91,$C1,$FB,$71,$A8,$C1,$27
	.dc.b	$99,$F1,$F0,$F9,$EF,$C4,$3C,$11,$F0,$1F,$11,$9A,$EE,$F1,$8F,$88
	.dc.b	$E5,$EC,$71,$8D,$C6,$4F,$0C,$64,$F2,$39,$34,$E5,$F3,$99,$C3,$F1
	.dc.b	$31,$98,$C6,$1F,$88,$8C,$4F,$DF,$24,$7D,$6B,$F7,$C6,$27,$E8,$57
	.dc.b	$68,$C3,$F9,$03,$E0,$B8,$BF,$9B,$5C,$AF,$C1,$87,$CF,$BE,$51,$89
	.dc.b	$D6,$1F,$89,$3C,$B1,$89,$ED,$7F,$1B,$0E,$D5,$F7,$EF,$53,$F4,$23
	.dc.b	$89,$F4,$27,$E1,$E1,$E1,$BF,$87,$3D,$E3,$13,$D2,$7E,$8C,$63,$0E
	.dc.b	$B5,$DC,$E8,$FA,$5A,$73,$C3,$75,$C9,$F5,$3F,$C0,$CF,$C5,$16,$FF
	.dc.b	$8A,$B6,$BF,$85,$BF,$9E,$31,$5F,$EC,$E9,$DC,$64,$87,$F0,$33,$67
	.dc.b	$FB,$81,$DB,$EC,$9F,$A6,$49,$F8,$4D,$8F,$CE,$DF,$C3,$9F,$E4,$13
	.dc.b	$9C,$3E,$74,$E4,$91,$99,$FC,$EC,$61,$EA,$4A,$EB,$3B,$7C,$91,$9A
	.dc.b	$EB,$39,$9F,$CF,$D7,$37,$8D,$6C,$EF,$3B,$BF,$BE,$D3,$F8,$89,$63
	.dc.b	$FA,$65,$FF,$02,$8C,$57,$FA,$45,$B1,$99,$ED,$3E,$63,$C2,$48,$F3
	.dc.b	$C6,$A3,$CF,$3C,$A3,$F8,$A3,$63,$F9,$93,$63,$E6,$9C,$6B,$D2,$D8
	.dc.b	$F4,$B4,$F4,$9D,$0C,$4F,$A5,$B5,$E9,$8F,$34,$7C,$63,$27,$17,$27
	.dc.b	$B6,$3D,$6E,$0F,$5C,$70,$7D,$2D,$3D,$2E,$EB,$D2,$DA,$F4,$BD,$A7
	.dc.b	$FB,$57,$83,$FD,$10,$FE,$8B,$E0,$D8,$F6,$CF,$F5,$79,$C3,$FE,$C5
	.dc.b	$9B,$87,$F3,$A9,$B5,$FD,$06,$6D,$71,$8E,$67,$80,$B8,$A7,$71,$99
	.dc.b	$FF,$69,$57,$92,$36,$EA,$8E,$67,$73,$FE,$0F,$E5,$FD,$CC,$7F,$86
	.dc.b	$1A,$FF,$81,$5B,$91,$A4,$9F,$DF,$36,$7F,$74,$D7,$F8,$43,$5F,$F0
	.dc.b	$47,$2F,$42,$4F,$F0,$76,$C7,$F0,$36,$BF,$C4,$5A,$F2,$8E,$04,$7F
	.dc.b	$8C,$0F,$BC,$FF,$1A,$C6,$23,$0F,$BA,$33,$1C,$1A,$FF,$21,$6B,$FE
	.dc.b	$25,$6B,$FE,$3F,$6B,$D2,$3B,$46,$08,$F7,$3C,$1C,$D1,$DB,$67,$C4
	.dc.b	$D7,$C4,$D7,$C4,$D7,$C4,$D7,$E0,$73,$7B,$49,$3E,$46,$C7,$BD,$AF
	.dc.b	$BD,$AF,$BD,$AF,$BD,$F9,$1C,$D2,$7D,$ED,$9F,$94,$D7,$E5,$35,$F9
	.dc.b	$4D,$7E,$53,$5F,$B1,$CD,$F5,$49,$3F,$31,$B1,$F5,$DA,$FD,$76,$BF
	.dc.b	$5D,$AF,$D7,$7F,$01,$CD,$27,$EB,$B6,$7F,$04,$D7,$F0,$4D,$7F,$04
	.dc.b	$D7,$F0,$4D,$7F,$21,$EE,$7E,$34,$93,$F8,$46,$C7,$E3,$DA,$FE,$3D
	.dc.b	$AF,$E3,$DA,$FE,$3D,$FC,$C7,$B9,$27,$F1,$ED,$9F,$CD,$35,$FC,$D3
	.dc.b	$5F,$CD,$35,$FC,$D3,$5F,$D4,$7B,$9F,$A5,$24,$FE,$71,$B1,$FA,$76
	.dc.b	$BF,$A7,$6B,$FA,$76,$BF,$A7,$7F,$61,$EE,$49,$FD,$3B,$67,$F6,$4D
	.dc.b	$7F,$64,$D7,$F6,$4D,$7F,$64,$D7,$F8,$07,$B9,$FB,$D2,$4F,$ED,$1B
	.dc.b	$1F,$BF,$6B,$FB,$F6,$BF,$BF,$6B,$FB,$F7,$F8,$47,$83,$27,$F7,$ED
	.dc.b	$9F,$E1,$4D,$7F,$85,$35,$FE,$14,$D7,$F8,$53,$5F,$E3,$1E,$E7,$F1
	.dc.b	$52,$4F,$F0,$C6,$C7,$F1,$76,$BF,$C5,$DA,$FF,$17,$6B,$FC,$5D,$FE
	.dc.b	$41,$EE,$49,$FE,$2E,$D9,$FE,$44,$D7,$F9,$13,$5F,$E4,$4D,$7F,$91
	.dc.b	$35,$FE,$61,$EE,$7F,$2D,$24,$FF,$24,$6C,$7F,$2F,$6B,$FC,$BD,$AF
	.dc.b	$F2,$F6,$BF,$CB,$DF,$E7,$1E,$E4,$9F,$E5,$ED,$9F,$E7,$4D,$7F,$9D
	.dc.b	$35,$FE,$74,$D7,$F9,$D3,$5F,$E9,$1E,$E7,$F4,$52,$4F,$F3,$C6,$C7
	.dc.b	$F4,$76,$BF,$D1,$DA,$FF,$47,$6B,$FD,$1D,$FE,$DB,$0F,$47,$FD,$BF
	.dc.b	$28,$C1,$FE,$00,$F8,$7C,$B8,$87,$6F,$FB,$EA,$30,$DA,$FF,$6C,$D7
	.dc.b	$06,$4F,$F8,$5A,$BB,$46,$A3,$07,$21,$1F,$F0,$02,$38,$BE,$08,$FF
	.dc.b	$6C,$4E,$2B,$C1,$1C,$AB,$A5,$6C,$E5,$19,$23,$95,$D8,$BE,$91,$E2
	.dc.b	$9F,$3B,$C2,$FF,$DC,$73,$A9,$EA,$77,$36,$75,$AE,$95,$C4,$E8,$7B
	.dc.b	$E3,$8D,$74,$8D,$46,$1D,$DF,$B2,$7D,$F3,$DE,$79,$3E,$08,$E4,$E6
	.dc.b	$7C,$CE,$67,$E4,$9D,$9C,$1D,$CF,$C1,$57,$A1,$E5,$1B,$3D,$07,$C8
	.dc.b	$BE,$AF,$88,$CC,$F1,$BF,$39,$23,$D6,$7E,$34,$E9,$7F,$2D,$E8,$60
	.dc.b	$F8,$66,$2F,$C6,$49,$E8,$FA,$2F,$C7,$5C,$EF,$C5,$7E,$0A,$E0,$DB
	.dc.b	$F8,$E7,$28,$F8,$67,$C4,$8F,$04,$65,$F7,$57,$D3,$79,$12,$7E,$53
	.dc.b	$C0,$C3,$D1,$F8,$B7,$E6,$7D,$B7,$E6,$9F,$A4,$6E,$BB,$38,$9F,$09
	.dc.b	$F2,$5F,$59,$98,$DB,$C6,$3A,$D7,$78,$DC,$79,$5E,$0E,$8F,$A0,$EA
	.dc.b	$BC,$8F,$33,$13,$CE,$FE,$8D,$73,$9D,$46,$2B,$8C,$FA,$27,$8B,$DA
	.dc.b	$75,$5C,$E3,$E5,$1C,$CE,$13,$FB,$33,$F5,$A5,$7F,$8A,$36,$FF,$8E
	.dc.b	$92,$3F,$CE,$06,$A3,$F0,$46,$DC,$C6,$E3,$FC,$F3,$1B,$7E,$DF,$C3
	.dc.b	$BC,$EF,$CB,$36,$7F,$2A,$D9,$F9,$AD,$C3,$E2,$B6,$7D,$AD,$9E,$B5
	.dc.b	$D0,$93,$D1,$B1,$D1,$B5,$D1,$B3,$D1,$B3,$D2,$39,$11,$E4,$D9,$E4
	.dc.b	$D3,$91,$63,$91,$C0,$91,$C1,$B1,$C1,$A7,$08,$D1,$23,$4D,$8D,$46
	.dc.b	$08,$E0,$87,$EC,$EB,$E4,$36,$BD,$ED,$9F,$43,$4F,$19,$6B,$C0,$D8
	.dc.b	$EE,$D3,$BB,$FC,$7A,$BB,$B6,$BB,$B4,$EE,$59,$EE,$DA,$EE,$D3,$B9
	.dc.b	$63,$BD,$FF,$71,$7F,$AA,$5F,$82,$31,$59,$A4,$E5,$FF,$7B,$B6,$B8
	.dc.b	$DF,$9A,$F8,$D7,$F6,$C3,$8C,$65,$E2,$FC,$AF,$BC,$79,$7F,$E0,$46
	.dc.b	$CF,$F7,$4C,$7C,$59,$F8,$A3,$FB,$E4,$EB,$1F,$EC,$4D,$AF,$64,$23
	.dc.b	$13,$E3,$AE,$53,$C2,$FC,$04,$9F,$01,$BC,$B1,$3B,$8F,$0C,$FB,$A3
	.dc.b	$FA,$CC,$EB,$0F,$0E,$1E,$09,$FF,$06,$B6,$FF,$C0,$35,$B7,$FC,$01
	.dc.b	$1B,$9F,$45,$F5,$9D,$9E,$02,$47,$8C,$FF,$00,$57,$37,$BC,$D9,$E3
	.dc.b	$1B,$BF,$2C,$E7,$0F,$87,$18,$AE,$AD,$AE,$B1,$83,$CF,$5C,$6F,$C1
	.dc.b	$3E,$89,$D3,$F1,$1A,$E4,$F1,$3F,$34,$CE,$1F,$04,$F8,$D8,$79,$23
	.dc.b	$B9,$AB,$F8,$91,$9A,$F7,$CE,$4F,$B2,$F8,$4E,$E6,$8A,$FC,$4C,$3D
	.dc.b	$C4,$7B,$C7,$C0,$38,$9E,$98,$D3,$D5,$AE,$63,$8C,$7E,$63,$0F,$8B
	.dc.b	$87,$B7,$0F,$4C,$61,$F9,$55,$E9,$BF,$2E,$1E,$D7,$A1,$E9,$C3,$C3
	.dc.b	$3E,$93,$B9,$23,$4F,$CD,$7D,$38,$78,$63,$E2,$9C,$63,$64,$3B,$B9
	.dc.b	$7C,$24,$8F,$AB,$87,$A6,$7E,$F9,$B8,$ED,$3E,$A9,$D3,$67,$DD,$5E
	.dc.b	$AB,$F7,$5F,$08,$C5,$F0,$6D,$70,$3D,$C7,$7C,$34,$FB,$E3,$E9,$D7
	.dc.b	$38,$C1,$F1,$70,$F5,$57,$D2,$8E,$91,$F0,$9E,$87,$EB,$9F,$F0,$14
	.dc.b	$FF,$0F,$6C,$FF,$0E,$7F,$0E,$62,$7F,$C1,$0F,$E3,$CC,$18,$9E,$4F
	.dc.b	$02,$1B,$29,$F4,$B0,$7C,$47,$ED,$C9,$5F,$42,$32,$D8,$CC,$FE,$29
	.dc.b	$B8,$7C,$F6,$CF,$CA,$AE,$35,$B8,$E4,$47,$04,$AE,$51,$96,$C6,$63
	.dc.b	$A5,$76,$8F,$99,$5D,$AB,$64,$AD,$9C,$CB,$3C,$CE,$35,$CD,$F8,$8D
	.dc.b	$BF,$E0,$ED,$C3,$F8,$2B,$6B,$F3,$F8,$78,$5E,$B8,$78,$70,$EB,$1E
	.dc.b	$16,$9E,$13,$BB,$96,$99,$C3,$C3,$19,$6C,$67,$0F,$1C,$E5,$B1,$98
	.dc.b	$F2,$E1,$E7,$8F,$F5,$D3,$6F,$FB,$0B,$67,$FA,$D3,$67,$FD,$52,$D8
	.dc.b	$FF,$66,$B7,$0F,$81,$3E,$CA,$F8,$0F,$A9,$B8,$7A,$1B,$87,$99,$B8
	.dc.b	$79,$8F,$F7,$67,$A6,$E7,$FD,$16,$1F,$EE,$A8,$C4,$FE,$8A,$11,$87
	.dc.b	$F4,$AD,$7F,$CD,$CD,$7F,$DC,$4D,$7A,$4F,$67,$9B,$63,$51,$89,$F0
	.dc.b	$46,$09,$CF,$B2,$46,$9B,$3E,$46,$BE,$46,$BE,$46,$BE,$F7,$C5,$CF
	.dc.b	$27,$C2,$D9,$F7,$B5,$F7,$B5,$F7,$B5,$F9,$4D,$7E,$4F,$3F,$BC,$93
	.dc.b	$EE,$6C,$7C,$C6,$BF,$31,$AF,$CC,$6B,$F5,$DF,$95,$CF,$27,$E4,$B6
	.dc.b	$7E,$BB,$5F,$AE,$D7,$EB,$B5,$FC,$13,$5F,$C0,$F3,$FD,$72,$4F,$D6
	.dc.b	$6C,$7E,$11,$AF,$E1,$1A,$FE,$11,$AF,$E3,$DF,$C2,$7E,$76,$F7,$FF
	.dc.b	$C6,$92,$FB,$D7,$B9,$8A,$D6,$1F,$E0,$6A,$D4,$E6,$F8,$DF,$0A,$D5
	.dc.b	$F0,$B6,$FF,$E4,$EA,$D4,$62,$8C,$AE,$15,$BC,$3B,$57,$86,$F8,$CE
	.dc.b	$63,$13,$AB,$E8,$74,$AD,$CF,$68,$F4,$BC,$0F,$3E,$1E,$AA,$91,$99
	.dc.b	$E6,$30,$F2,$C7,$3B,$EB,$3D,$EB,$8B,$C2,$B6,$E6,$BA,$61,$E9,$8D
	.dc.b	$61,$C2,$30,$49,$F4,$57,$2A,$F4,$9E,$4C,$3C,$2E,$2B,$FC,$F7,$3B
	.dc.b	$9C,$93,$0F,$8D,$3D,$6B,$C8,$73,$8C,$DF,$B5,$F2,$34,$F8,$E7,$A8
	.dc.b	$C4,$71,$72,$D2,$4F,$77,$E5,$61,$D1,$F0,$C6,$70,$F4,$5E,$67,$D9
	.dc.b	$7F,$0A,$F1,$87,$AE,$3B,$E1,$D6,$31,$5F,$32,$7C,$CF,$67,$CF,$3F
	.dc.b	$1A,$B9,$DF,$7C,$3C,$35,$C2,$FC,$27,$E5,$6B,$71,$99,$DC,$FB,$A3
	.dc.b	$A9,$E7,$39,$61,$E4,$C3,$C1,$18,$C3,$C1,$87,$4B,$E4,$77,$9D,$1D
	.dc.b	$E3,$55,$F1,$DB,$1E,$09,$E0,$72,$3D,$24,$C3,$CF,$5F,$1E,$FE,$FC
	.dc.b	$FC,$52,$1B,$BE,$AE,$2B,$8D,$F7,$8D,$3E,$67,$C1,$1A,$79,$E1,$F9
	.dc.b	$68,$C4,$79,$23,$15,$DA,$39,$9E,$98,$FE,$0E,$FE,$E4,$FD,$09,$5C
	.dc.b	$BF,$92,$31,$7F,$D8,$63,$31,$F9,$E3,$37,$FD,$BE,$33,$1D,$0E,$44
	.dc.b	$3D,$3E,$AF,$E4,$09,$3F,$71,$B1,$E7,$6E,$3F,$93,$6C,$FC,$7A,$46
	.dc.b	$2B,$AC,$F3,$23,$CD,$B5,$CD,$A7,$39,$E5,$3F,$2D,$B5,$FC,$01,$B7
	.dc.b	$FC,$D9,$B5,$FC,$AD,$B5,$FC,$2D,$B7,$FC,$DD,$B3,$FC,$7D,$B3,$FC
	.dc.b	$B5,$EE,$75,$8F,$2C,$65,$EC,$61,$FE,$9D,$1B,$7D,$C7,$86,$36,$E2
	.dc.b	$7C,$51,$87,$E1,$46,$C8,$E0,$DB,$88,$F8,$C7,$13,$D4,$78,$EF,$E9
	.dc.b	$B9,$8F,$A6,$71,$3D,$47,$8D,$E2,$E4,$FA,$0F,$17,$DA,$F5,$AF,$75
	.dc.b	$79,$63,$8B,$E6,$31,$3D,$9C,$47,$A2,$3D,$C7,$E7,$E7,$F3,$84,$8F
	.dc.b	$CD,$36,$3F,$D5,$75,$EC,$6C,$65,$E5,$3C,$08,$E9,$FE,$FD,$F9,$E3
	.dc.b	$CD,$F4,$C7,$F6,$98,$C3,$EF,$6C,$FC,$36,$E1,$EA,$8F,$EF,$B1,$89
	.dc.b	$F8,$27,$3A,$FF,$04,$F4,$48,$FF,$07,$7E,$A9,$FD,$66,$31,$7F,$D7
	.dc.b	$DF,$F0,$BC,$66,$7F,$C0,$91,$AC,$11,$FD,$FE,$75,$5C,$63,$F8,$05
	.dc.b	$7F,$05,$23,$82,$47,$CA,$8C,$B6,$32,$FF,$11,$8C,$B6,$3F,$C4,$DE
	.dc.b	$83,$E3,$3C,$91,$FE,$11,$8C,$3E,$83,$71,$A9,$FF,$11,$CE,$9F,$41
	.dc.b	$FE,$12,$71,$87,$F8,$AE,$FF,$C7,$31,$8B,$FF,$0A,$46,$2F,$A5,$AB
	.dc.b	$4F,$F8,$96,$75,$5E,$38,$ED,$5C,$2B,$71,$FE,$45,$AD,$BE,$4B,$F9
	.dc.b	$EF,$A2,$B8,$12,$FA,$4E,$A7,$C5,$39,$79,$4F,$C5,$8C,$4F,$B6,$3E
	.dc.b	$29,$C8,$D1,$67,$C0,$D7,$AB,$6F,$D3,$A3,$0F,$4D,$F5,$BF,$7D,$EF
	.dc.b	$0F,$0B,$F2,$8F,$0C,$79,$8E,$E7,$02,$4F,$33,$DD,$87,$C8,$25,$74
	.dc.b	$9E,$33,$A8,$C4,$7A,$0E,$75,$E1,$69,$C1,$F8,$2F,$8C,$E1,$3C,$E3
	.dc.b	$CE,$F8,$DD,$B9,$8E,$67,$DB,$8D,$9A,$23,$F8,$57,$2F,$84,$87,$13
	.dc.b	$9D,$7B,$A3,$E5,$1C,$23,$CA,$FA,$E7,$C4,$47,$BC,$73,$8C,$3D,$A7
	.dc.b	$EC,$DF,$D8,$AD,$C6,$1E,$CF,$86,$7E,$23,$6B,$9C,$75,$9E,$37,$CE
	.dc.b	$3D,$C7,$A9,$C3,$B3,$C4,$FF,$BA,$EB,$F6,$91,$FE,$13,$6B,$F5,$0C
	.dc.b	$46,$C9,$3F,$EF,$D9,$DC,$FE,$28,$DC,$66,$77,$1F,$E1,$F3,$E4,$9F
	.dc.b	$C1,$3E,$23,$CB,$46,$92,$BE,$1B,$67,$EA,$B6,$FE,$1D,$71,$AD,$92
	.dc.b	$B6,$D9,$DD,$71,$8F,$A2,$D9,$FA,$CF,$F1,$22,$47,$13,$F9,$03,$88
	.dc.b	$EA,$7F,$1B,$71,$3D,$A7,$F9,$4C,$61,$EE,$60,$93,$CA,$7F,$5F,$3B
	.dc.b	$8F,$1C,$E9,$FA,$D1,$F6,$49,$1F,$4D,$B1,$F3,$EB,$E2,$36,$32,$F2
	.dc.b	$9E,$04,$74,$FF,$B9,$7E,$FD,$E4,$F8,$E5,$18,$9F,$4E,$1F,$C8,$E3
	.dc.b	$13,$F0,$E7,$95,$71,$9C,$D7,$BA,$72,$72,$3F,$DB,$91,$A7,$E7,$9A
	.dc.b	$AE,$87,$FB,$6A,$34,$FE,$2E,$35,$3D,$63,$55,$E1,$3F,$9E,$C7,$9D
	.dc.b	$B3,$E0,$74,$F2,$30,$79,$67,$FD,$FB,$5E,$07,$E1,$CF,$27,$2F,$27
	.dc.b	$C3,$3C,$DF,$31,$82,$57,$8E,$3B,$47,$47,$83,$4F,$83,$1B,$78,$C7
	.dc.b	$99,$CB,$63,$44,$7E,$5C,$70,$3C,$EF,$6C,$3F,$0F,$3B,$9C,$9F,$B9
	.dc.b	$BF,$BA,$FF,$A6,$5A,$F8,$4C,$5F,$F6,$57,$04,$39,$9A,$BF,$E0,$8F
	.dc.b	$B4,$C1,$DE,$39,$3E,$2A,$FE,$E1,$EC,$3E,$1B,$E8,$4F,$C5,$9C,$C6
	.dc.b	$30,$FA,$D5,$BC,$3E,$5B,$6F,$E1,$07,$17,$F0,$63,$17,$CF,$CA,$F1
	.dc.b	$C7,$E8,$9B,$81,$A4,$C3,$DB,$19,$6C,$66,$7E,$63,$5F,$98,$D7,$F4
	.dc.b	$7E,$57,$62,$4F,$46,$C7,$D5,$6B,$F5,$5A,$F2,$8E,$04,$7E,$E7,$95
	.dc.b	$EA,$24,$7A,$1B,$1A,$8C,$11,$C1,$0F,$C2,$47,$07,$F0,$DE,$57,$C6
	.dc.b	$24,$FC,$46,$C7,$E3,$5A,$F0,$8D,$11,$D3,$51,$17,$84,$F6,$FF,$69
	.dc.b	$18,$E1,$B3,$98,$E3,$AF,$F9,$01,$D1,$83,$A4,$6E,$D3,$B8,$FF,$28
	.dc.b	$CF,$82,$FA,$DF,$66,$C7,$3B,$FF,$3D,$D7,$3B,$E5,$7B,$AD,$1E,$2A
	.dc.b	$EE,$F8,$AF,$9C,$60,$E6,$E2,$BC,$8D,$9E,$07,$4B,$F3,$9E,$17,$DB
	.dc.b	$7F,$A7,$EE,$FC,$D9,$23,$F3,$2D,$C3,$F8,$13,$5F,$E0,$4D,$7F,$9C
	.dc.b	$B5,$FE,$4E,$FE,$9B,$BA,$4F,$E9,$1B,$3F,$C3,$5A,$FF,$0D,$6B,$FD
	.dc.b	$11,$AF,$F3,$26,$BF,$D5,$3B,$7F,$80,$1F,$CC,$F8,$FF,$78,$E2,$BF
	.dc.b	$AE,$36,$7E,$54,$EA,$72,$48,$FE,$9D,$3C,$F8,$8F,$1D,$3F,$B0,$3F
	.dc.b	$C0,$89,$5F,$C0,$1B,$5F,$CA,$5A,$FF,$29,$6B,$FD,$81,$AF,$F4,$E7
	.dc.b	$8C,$FF,$A5,$2F,$F8,$B7,$97,$E1,$BF,$F5,$6E,$08,$C6,$1F,$EA,$58
	.dc.b	$C4,$FA,$EF,$94,$62,$7D,$B3,$F9,$5B,$EA,$EC,$EB,$7F,$CA,$FC,$A9
	.dc.b	$1B,$B7,$97,$8C,$3F,$BA,$55,$C7,$BC,$DB,$FE,$FB,$36,$BC,$D3,$5F
	.dc.b	$F0,$6C,$7F,$2C,$24,$FF,$2A,$6C,$7F,$53,$6B,$FD,$4D,$AF,$68,$E0
	.dc.b	$47,$83,$E9,$8F,$16,$3F,$D0,$5B,$81,$A4,$9F,$F1,$23,$F8,$33,$FA
	.dc.b	$77,$2C,$8F,$89,$E5,$DC,$3F,$C2,$D3,$72,$FA,$33,$6B,$FC,$3F,$36
	.dc.b	$BE,$C4,$D7,$FC,$5F,$1F,$D3,$89,$3F,$D3,$1B,$1F,$DD,$5A,$FF,$75
	.dc.b	$6B,$DA,$38,$11,$E0,$FD,$BF,$87,$F1,$65,$FD,$8D,$B9,$1A,$49,$FF
	.dc.b	$23,$5F,$E9,$63,$FB,$9F,$DF,$BB,$A9,$FA,$99,$B5,$FC,$8F,$15,$6E
	.dc.b	$30,$D8,$FD,$AE,$1F,$DD,$27,$F9,$FC,$70,$BE,$6F,$F2,$33,$11,$F9
	.dc.b	$5B,$CE,$7F,$EF,$7C,$3A,$57,$23,$67,$FA,$55,$CB,$D9,$AF,$F4,$E6
	.dc.b	$BC,$0E,$4E,$09,$7F,$E0,$66,$DF,$F8,$79,$B3,$EC,$F5,$FC,$B3,$FB
	.dc.b	$A6,$C7,$EE,$5B,$9F,$F8,$55,$B9,$1A,$7D,$0F,$86,$97,$FB,$5F,$12
	.dc.b	$E1,$F1,$71,$56,$33,$FE,$07,$56,$FE,$04,$DB,$FE,$15,$35,$FA,$F1
	.dc.b	$ED,$24,$FB,$5B,$1F,$2D,$AF,$C6,$6B,$DA,$38,$11,$E0,$FF,$13,$FD
	.dc.b	$67,$F0,$57,$F6,$6B,$F9,$E4,$62,$79,$57,$F3,$A8,$C3,$F5,$E7,$7A
	.dc.b	$7A,$F4,$56,$AB,$D7,$7D,$EB,$F3,$3F,$0D,$23,$C5,$F0,$76,$BE,$54
	.dc.b	$DC,$BF,$9A,$60,$AC,$5F,$E2,$6A,$DB,$1F,$0D,$62,$3F,$43,$3F,$87
	.dc.b	$25,$7E,$1D,$B1,$F9,$56,$BF,$90,$6B,$DE,$38,$11,$E0,$FC,$BF,$88
	.dc.b	$F1,$DF,$C2,$6E,$06,$92,$7F,$67,$7F,$42,$3F,$7B,$E9,$49,$FA,$DA
	.dc.b	$5C,$7F,$1D,$37,$3F,$8B,$36,$BF,$59,$36,$BE,$26,$3F,$94,$9F,$E0
	.dc.b	$33,$CE,$7F,$5A,$FE,$AC,$91,$FA,$B6,$C7,$EE,$5A,$F0,$8D,$11,$D3
	.dc.b	$6F,$EC,$FC,$47,$8F,$3F,$C2,$B7,$03,$49,$5E,$63,$F2,$CF,$F2,$0F
	.dc.b	$94,$91,$F9,$9F,$2E,$E3,$FB,$19,$B8,$FE,$22,$6D,$7F,$0F,$9B,$5F
	.dc.b	$12,$6B,$FC,$9A,$3F,$86,$12,$7F,$86,$36,$3F,$8D,$35,$FE,$2A,$D7
	.dc.b	$B4,$70,$23,$C1,$FD,$86,$BE,$2C,$BF,$4E,$DC,$8D,$24,$FF,$38,$BF
	.dc.b	$DA,$C7,$F4,$2F,$92,$93,$FB,$BF,$8A,$B8,$FF,$13,$9B,$9F,$EA,$A6
	.dc.b	$D7,$F3,$19,$B5,$F8,$69,$AF,$F4,$88,$FE,$5C,$49,$FE,$5C,$D8,$FE
	.dc.b	$78,$D7,$F9,$BB,$5E,$D1,$C0,$8F,$07,$F8,$97,$C5,$78,$B3,$FE,$12
	.dc.b	$DC,$CD,$24,$FF,$5A,$BF,$E2,$F1,$FD,$93,$E4,$A4,$FF,$1D,$DE,$E3
	.dc.b	$FC,$D6,$6E,$7F,$C3,$26,$D7,$F5,$39,$B5,$FA,$89,$AF,$F6,$A8,$FE
	.dc.b	$A0,$49,$FE,$A0,$D8,$FE,$BC,$D7,$FA,$CB,$5E,$D1,$C0,$8F,$07,$F9
	.dc.b	$A7,$CB,$78,$B3,$FE,$56,$DC,$CD,$24,$FF,$7C,$BF,$E7,$31,$FE,$F8
	.dc.b	$FC,$03,$FD,$71,$87,$F7,$96,$E5,$FC,$B3,$E5,$F5,$33,$0F,$F0,$45
	.dc.b	$69,$FF,$60,$E5,$FC,$F2,$B5,$5C,$CB,$7C,$EB,$8C,$68,$FF,$01,$5F
	.dc.b	$89,$A7,$FB,$F9,$C3,$FE,$10,$6B,$FE,$D0,$6A,$69,$1E,$44,$8F,$F0
	.dc.b	$83,$C4,$C1,$FD,$67,$83,$CF,$1E,$CC,$D1,$89,$FE,$4E,$DC,$7F,$AD
	.dc.b	$B6,$77,$5E,$E3,$FB,$54,$7C,$FF,$85,$91,$FD,$B7,$E5,$EE,$1F,$46
	.dc.b	$6E,$5F,$D7,$26,$D7,$C1,$9B,$5F,$CE,$26,$BF,$06,$3E,$41,$27,$DF
	.dc.b	$19,$6E,$49,$D4,$FC,$48,$C4,$EA,$73,$3F,$59,$AF,$8E,$34,$7F,$81
	.dc.b	$FE,$5F,$CB,$9F,$F7,$E6,$E6,$69,$27,$F1,$77,$FE,$0C,$8F,$CA,$7C
	.dc.b	$34,$9F,$F0,$97,$15,$C7,$F2,$D3,$74,$FF,$00,$CD,$AF,$C3,$4D,$AF
	.dc.b	$EB,$B3,$5F,$C3,$47,$E3,$49,$3F,$11,$B1,$F0,$DA,$FE,$59,$AF,$68
	.dc.b	$E0,$47,$83,$FE,$29,$EE,$F1,$67,$FE,$20,$6E,$66,$92,$7F,$5B,$7F
	.dc.b	$E2,$E8,$FD,$B7,$BD,$27,$FC,$69,$EB,$DC,$7F,$71,$37,$3F,$F1,$1C
	.dc.b	$DA,$FD,$44,$DA,$FE,$FF,$35,$FD,$44,$7E,$BC,$93,$F8,$66,$C7,$E1
	.dc.b	$5A,$FE,$E1,$AF,$68,$E0,$47,$83,$FE,$4B,$E3,$F1,$67,$FE,$42,$6E
	.dc.b	$66,$92,$7F,$87,$DF,$F9,$42,$3F,$8C,$7C,$BC,$9F,$F2,$B7,$C6,$DC
	.dc.b	$7F,$8D,$4D,$CF,$FC,$8B,$36,$BF,$85,$4D,$AF,$F1,$0C,$D7,$F8,$54
	.dc.b	$7F,$11,$24,$FE,$A1,$B1,$FA,$76,$BF,$C6,$9A,$F6,$8E,$04,$78,$3F
	.dc.b	$E6,$DF,$5B,$C5,$9F,$F9,$91,$B9,$9A,$49,$FE,$63,$7F,$E7,$28,$FE
	.dc.b	$75,$EF,$C9,$60,$C9,$1F,$CD,$BF,$5A,$FE,$71,$3F,$CA,$6B,$83,$FD
	.dc.b	$05,$DB,$C6,$78,$3A,$AC,$BF,$CB,$E7,$13,$A3,$FA,$2B,$D2,$3F,$9E
	.dc.b	$BF,$CE,$C9,$1F,$C4,$9B,$1F,$C4,$5A,$F0,$8D,$11,$D3,$67,$CF,$92
	.dc.b	$FF,$CF,$D3,$E2,$3E,$1C,$EA,$33,$5E,$F9,$C9,$F1,$6B,$D7,$1F,$E8
	.dc.b	$79,$DD,$7C,$0E,$46,$1F,$1A,$3A,$5F,$03,$C4,$7F,$6D,$F9,$59,$8F
	.dc.b	$6C,$7A,$7C,$66,$1F,$EE,$13,$6F,$CD,$36,$BF,$B0,$4D,$AE,$31,$FD
	.dc.b	$CA,$39,$47,$F6,$A7,$FB,$41,$23,$F9,$FB,$63,$F9,$F3,$5E,$11,$A2
	.dc.b	$3A,$6D,$7C,$9E,$6F,$16,$1F,$3D,$B9,$FF,$48,$BF,$5D,$FB,$EB,$4F
	.dc.b	$9C,$C4,$EE,$FE,$66,$1F,$2E,$7F,$2D,$87,$F8,$57,$E2,$24,$70,$F8
	.dc.b	$CB,$97,$F8,$5E,$6E,$1F,$76,$6D,$7F,$81,$E6,$D7,$DB,$9A,$FF,$81
	.dc.b	$E3,$FC,$1C,$49,$FE,$CA,$D8,$FE,$C8,$D7,$FC,$2E,$D7,$B4,$70,$23
	.dc.b	$C1,$F7,$FC,$67,$8B,$3F,$CB,$37,$33,$49,$3F,$E2,$EB,$F8,$97,$F9
	.dc.b	$39,$AF,$FB,$7F,$DC,$FD,$9B,$FC,$00,$C5,$7F,$80,$3C,$74,$EA,$7F
	.dc.b	$82,$C6,$27,$53,$99,$E2,$7F,$89,$A7,$4D,$8D,$46,$23,$C0,$70,$2B
	.dc.b	$97,$C8,$D7,$C3,$F1,$3C,$89,$5E,$76,$C7,$85,$AF,$46,$BD,$1A,$F4
	.dc.b	$8F,$81,$EE,$79,$88,$E0,$91,$ED,$8C,$B6,$32,$FA,$A3,$2D,$8F,$99
	.dc.b	$EE,$7C,$22,$38,$24,$7C,$98,$CB,$63,$2F,$C6,$8C,$B6,$3E,$C7,$B9
	.dc.b	$F3,$88,$E0,$91,$F5,$63,$2D,$8C,$BF,$4A,$32,$D8,$FC,$27,$B9,$F6
	.dc.b	$88,$E0,$91,$F8,$18,$CB,$63,$2F,$DE,$8C,$B6,$3F,$21,$F0,$9F,$96
	.dc.b	$C8,$FE,$6B,$7F,$E6,$DD,$F9,$89,$5F,$C9,$6D,$7C,$23,$F8,$3D,$6A
	.dc.b	$7F,$AA,$1F,$9A,$AC,$71,$7F,$3D,$9B,$5F,$90,$9F,$CB,$91,$F2,$92
	.dc.b	$BF,$3D,$19,$6C,$65,$FC,$DC,$65,$B1,$C6,$BF,$A6,$7C,$4F,$82,$7F
	.dc.b	$A4,$37,$1F,$EA,$8D,$E2,$34,$97,$F0,$72,$7C,$2F,$C9,$3E,$45,$3E
	.dc.b	$27,$15,$B9,$B8,$FC,$4C,$7F,$B1,$56,$32,$FE,$EF,$8F,$4C,$38,$1F
	.dc.b	$CC,$A7,$4F,$ED,$B0,$FD,$E9,$1F,$E1,$4D,$C3,$F8,$23,$5F,$DF,$B5
	.dc.b	$E5,$1C,$08,$F0,$7F,$A0,$F8,$1E,$08,$FE,$DC,$DC,$7F,$BD,$37,$23
	.dc.b	$49,$5F,$1F,$D9,$FC,$71,$F1,$BE,$02,$C6,$49,$87,$C5,$B4,$EA,$3F
	.dc.b	$16,$7E,$42,$B1,$A7,$F8,$1E,$6D,$FF,$1A,$9F,$E4,$24,$7F,$97,$36
	.dc.b	$BF,$94,$35,$FE,$46,$D7,$94,$70,$23,$C1,$FE,$CB,$EF,$F8,$23,$FC
	.dc.b	$24,$DC,$BF,$C3,$CD,$C4,$D2,$57,$E2,$FD,$9F,$D9,$1F,$8A,$F0,$6C
	.dc.b	$64,$98,$7C,$58,$CB,$F1,$37,$56,$34,$FF,$15,$55,$BF,$E7,$B3,$FD
	.dc.b	$10,$8F,$F5,$16,$DF,$F4,$B6,$BF,$D1,$9A,$F2,$8E,$04,$78,$3F,$E0
	.dc.b	$8F,$98,$F0,$47,$F8,$D5,$B9,$FF,$90,$5B,$99,$A4,$AF,$D7,$7A,$FF
	.dc.b	$C5,$4F,$D6,$FC,$C5,$8C,$93,$0F,$C4,$46,$5F,$89,$37,$3F,$F2,$64
	.dc.b	$DA,$FE,$BF,$3F,$D9,$88,$FF,$73,$6D,$7F,$6D,$6B,$FD,$A1,$AF,$28
	.dc.b	$E0,$47,$83,$FE,$2A,$F0,$BC,$11,$FE,$57,$6E,$5F,$E6,$36,$E5,$FE
	.dc.b	$58,$AF,$66,$3F,$E6,$39,$C9,$FE,$0C,$FC,$05,$EC,$53,$F8,$8F,$CC
	.dc.b	$DB,$FE,$6D,$3A,$6C,$E9,$A4,$CB,$F5,$B3,$B7,$F6,$55,$CD,$CD,$7F
	.dc.b	$84,$67,$51,$FE,$0C,$6B,$FE,$1D,$6B,$FE,$9E,$6A,$69,$23,$A1,$2B
	.dc.b	$FC,$2F,$3C,$63,$0E,$27,$CD,$87,$FA,$1F,$E1,$BC,$CD,$BF,$63,$6B
	.dc.b	$D3,$1F,$E5,$C9,$DB,$FE,$AF,$6C,$7F,$A2,$DF,$9D,$EC,$49,$FF,$31
	.dc.b	$F8,$D6,$BE,$3C,$DB,$FF,$4C,$4D,$AF,$85,$36,$BF,$D2,$D3,$5F,$85
	.dc.b	$1F,$2C,$93,$F0,$23,$2D,$C1,$3A,$9F,$84,$D7,$E9,$35,$F0,$CF,$18
	.dc.b	$E1,$7F,$EA,$DF,$91,$F2,$6D,$FE,$B3,$6E,$C6,$92,$FF,$17,$7F,$D7
	.dc.b	$A3,$F2,$5E,$BC,$9F,$F4,$3F,$C8,$5C,$BF,$17,$37,$2F,$F5,$E4,$DA
	.dc.b	$FC,$34,$DA,$FF,$5D,$4D,$7F,$0D,$1F,$8F,$24,$FC,$36,$C7,$E0,$DA
	.dc.b	$FE,$4D,$AF,$68,$E0,$47,$83,$FE,$9E,$F1,$FC,$59,$FF,$B5,$5B,$99
	.dc.b	$A4,$9F,$D6,$DF,$F8,$02,$3F,$69,$EB,$49,$FF,$56,$7E,$B7,$FE,$FC
	.dc.b	$7F,$67,$19,$C7,$F5,$EF,$F8,$BE,$75,$18,$9D,$91,$FF,$04,$46,$B3
	.dc.b	$FF,$7E,$EC,$BE,$95,$FB,$37,$F6,$C4,$BF,$C5,$36,$3F,$54,$D7,$84
	.dc.b	$68,$8E,$9B,$5F,$EF,$AB,$FF,$47,$BE,$2A,$F5,$79,$1E,$D8,$D3,$69
	.dc.b	$6D,$18,$72,$9F,$E1,$18,$72,$7E,$06,$1C,$9F,$E3,$7E,$AF,$F2,$13
	.dc.b	$E4,$79,$36,$32,$4C,$7E,$3C,$65,$FE,$3B,$36,$BD,$53,$6B,$F8,$C4
	.dc.b	$FF,$1D,$23,$FB,$16,$D7,$F1,$16,$BF,$C8,$5A,$F2,$8E,$04,$78,$3F
	.dc.b	$84,$F2,$7C,$11,$F0,$F1,$46,$32,$F8,$AD,$8C,$D7,$B7,$1F,$A5,$39
	.dc.b	$3F,$A2,$7A,$72,$7F,$23,$E5,$5C,$7F,$9E,$CD,$BF,$B5,$36,$BF,$9C
	.dc.b	$4D,$AF,$24,$D7,$F9,$C4,$7F,$41,$24,$FF,$15,$6C,$7F,$35,$6B,$FD
	.dc.b	$15,$AF,$68,$E0,$47,$83,$FA,$8F,$2B,$C5,$97,$E1,$9B,$91,$A4,$9F
	.dc.b	$EB,$F7,$FA,$B8,$FE,$CD,$E8,$49,$FD,$9F,$97,$71,$FE,$BF,$37,$3F
	.dc.b	$CE,$4D,$AF,$EB,$53,$6B,$EB,$4D,$7F,$AD,$47,$F6,$32,$4F,$F3,$76
	.dc.b	$C7,$F5,$76,$BF,$D9,$DA,$F6,$8E,$04,$78,$3F,$C2,$BE,$3F,$C5,$9F
	.dc.b	$EA,$1B,$99,$A4,$9F,$F0,$0D,$FF,$0D,$84,$7F,$AE,$7E,$5D,$E8,$E3
	.dc.b	$1F,$F3,$6C,$62,$7F,$8C,$F9,$B9,$D3,$54,$AD,$3C,$2F,$57,$FB,$E8
	.dc.b	$D4,$7F,$1D,$BC,$C7,$F0,$2A,$E3,$7D,$1B,$7C,$5B,$3F,$A5,$AE,$2F
	.dc.b	$49,$F5,$5F,$92,$38,$35,$EC,$D7,$81,$FE,$0C,$AF,$4B,$E7,$32,$57
	.dc.b	$2F,$49,$FE,$E9,$18,$8E,$B1,$89,$ED,$39,$9E,$F7,$FC,$9F,$E2,$7D
	.dc.b	$33,$EC,$D3,$F9,$0C,$62,$FF,$A1,$37,$0F,$E9,$2D,$C3,$FA,$0B,$70
	.dc.b	$E9,$3F,$6F,$D0,$EA,$7F,$4B,$F8,$9B,$19,$25,$FF,$4A,$8C,$BF,$80
	.dc.b	$9B,$5F,$CF,$66,$D7,$D3,$9F,$69,$1F,$86,$DA,$F9,$8D,$7E,$53,$5E
	.dc.b	$51,$C0,$8F,$07,$FB,$3F,$9D,$E0,$8F,$E4,$8D,$CB,$FA,$3B,$77,$FF
	.dc.b	$73,$4F,$A6,$B8,$9E,$EC,$7F,$AB,$57,$D0,$F3,$E4,$FF,$74,$F8,$9B
	.dc.b	$7F,$A0,$9B,$7F,$D7,$66,$D7,$E4,$E6,$D7,$F3,$79,$AF,$E8,$E3,$F0
	.dc.b	$24,$9F,$C3,$B6,$3F,$20,$D7,$F1,$CD,$7B,$47,$02,$3C,$1F,$F0,$5F
	.dc.b	$C4,$F8,$B2,$FE,$D2,$DC,$8D,$24,$FE,$FE,$FF,$C1,$D1,$F9,$2F,$3A
	.dc.b	$4F,$F8,$5B,$CF,$B8,$FF,$02,$9B,$9F,$F7,$F9,$B5,$FB,$59,$B5,$FD
	.dc.b	$66,$6B,$FC,$16,$3F,$46,$49,$FD,$4B,$63,$F6,$4D,$7F,$60,$D7,$B4
	.dc.b	$70,$23,$C1,$FF,$16,$FC,$4F,$8B,$3F,$F0,$5B,$73,$34,$93,$FC,$86
	.dc.b	$FF,$C6,$11,$FB,$4F,$32,$4F,$F8,$E3,$E2,$6E,$3F,$C9,$26,$E7,$FE
	.dc.b	$21,$9B,$5F,$C5,$E6,$D7,$F7,$B9,$AF,$F2,$78,$FE,$0A,$49,$FE,$16
	.dc.b	$D8,$FE,$28,$D7,$F8,$93,$5E,$D1,$C0,$8F,$07,$FC,$9F,$E7,$F8,$B3
	.dc.b	$FF,$16,$B7,$33,$49,$3F,$D0,$EF,$FC,$A5,$1F,$C5,$BC,$A9,$3F,$E3
	.dc.b	$8C,$BF,$A3,$7C,$93,$F0,$65,$6B,$3F,$E9,$2D,$AD,$CE,$AF,$FA,$84
	.dc.b	$E5,$FE,$7C,$FF,$44,$BE,$8E,$6F,$FA,$33,$FC,$B0,$93,$FC,$C9,$B1
	.dc.b	$FC,$ED,$AF,$F3,$96,$BD,$A3,$81,$1E,$0F,$B4,$FF,$2A,$CB,$F0,$7F
	.dc.b	$F3,$74,$71,$96,$1F,$0A,$73,$1E,$69,$E4,$AB,$94,$7A,$DA,$FC,$23
	.dc.b	$6E,$23,$E4,$C6,$27,$AC,$7F,$6F,$F2,$FF,$A6,$92,$7F,$A9,$B6,$3F
	.dc.b	$AE,$35,$FE,$B4,$D7,$FB,$63,$5F,$EE,$4F,$F5,$8F,$2E,$4F,$F5,$E6
	.dc.b	$CF,$F6,$76,$BF,$D9,$5A,$FF,$77,$6B,$FD,$F5,$AF,$F8,$1F,$C9,$FE
	.dc.b	$CA,$49,$FE,$D6,$D8,$FE,$E8,$D7,$FB,$93,$5F,$F0,$23,$5F,$F0,$53
	.dc.b	$FD,$F7,$F0,$4F,$F4,$54,$7F,$A3,$9B,$C1,$FE,$8E,$F8,$DF,$F0,$4C
	.dc.b	$6A,$4D,$37,$C2,$9D,$F9,$53,$1F,$BB,$7C,$23,$FD,$71,$39,$8F,$C2
	.dc.b	$CF,$F8,$86,$D5,$E5,$24,$7F,$BF,$63,$2D,$8C,$CF,$F8,$4C,$ED,$19
	.dc.b	$23,$99,$E2,$D9,$E2,$FF,$86,$E3,$E0,$FE,$79,$FE,$2A,$8E,$67,$98
	.dc.b	$FF,$59,$39,$8D,$F7,$FA,$F8,$3F,$C8,$99,$AB,$58,$7F,$8E,$A3,$15
	.dc.b	$BC,$7F,$D7,$B1,$87,$F4,$63,$E3,$DE,$BF,$F9,$0A,$31,$EB,$D9,$CC
	.dc.b	$70,$51,$99,$D3,$67,$4D,$D3,$D9,$1D,$0E,$53,$C1,$A7,$02,$C7,$0A
	.dc.b	$EF,$18,$CF,$FC,$C8,$F3,$9D,$57,$36,$C7,$36,$9C,$EF,$D9,$CF,$27
	.dc.b	$D9,$EB,$5A,$F0,$4D,$C3,$BE,$2A,$C5,$FF,$BA,$0F,$95,$58,$BF,$F3
	.dc.b	$DD,$6E,$7E,$44,$E5,$D9,$E0,$32,$49,$F0,$B9,$69,$2F,$C7,$3F,$56
	.dc.b	$FC,$13,$FA,$8F,$17,$F5,$67,$E0,$F8,$5F,$1F,$8C,$7F,$4F,$C2,$AC
	.dc.b	$5F,$E4,$76,$FF,$80,$37,$8C,$FF,$D1,$B7,$BC,$38,$E1,$D6,$B5,$1F
	.dc.b	$AD,$8F,$E0,$04,$7F,$58,$DC,$3F,$4C,$D7,$F4,$8D,$79,$47,$02,$3C
	.dc.b	$1F,$E7,$78,$31,$7B,$1E,$2A,$FF,$7C,$36,$74,$D9,$DD,$7A,$8E,$6E
	.dc.b	$2B,$F6,$47,$CB,$8C,$CF,$E8,$FD,$8F,$D2,$75,$59,$FD,$23,$6B,$F3
	.dc.b	$0D,$D3,$F1,$2D,$9F,$C4,$6F,$FD,$C6,$BF,$3D,$7D,$27,$F5,$CD,$8C
	.dc.b	$C7,$63,$25,$72,$78,$1C,$B4,$CD,$F8,$A3,$F7,$5E,$BF,$EF,$4F,$DF
	.dc.b	$74,$D8,$C9,$2B,$F6,$51,$97,$F6,$93,$6B,$F4,$B3,$6B,$DB,$36,$BF
	.dc.b	$47,$36,$7F,$88,$4D,$C1,$36,$30,$F1,$AF,$E8,$7D,$FF,$D1,$8F,$E3
	.dc.b	$DD,$EF,$8E,$C6,$3F,$9C,$9B,$5F,$9B,$AD,$4E,$49,$1F,$7B,$8B,$FC
	.dc.b	$A7,$1F,$8C,$9D,$CE,$6B,$9B,$FC,$E5,$FE,$B8,$49,$FE,$90,$D8,$FE
	.dc.b	$84,$D7,$84,$68,$8E,$9B,$9F,$F3,$CA,$7B,$5E,$A8,$C3,$67,$F9,$DC
	.dc.b	$E9,$F5,$CE,$9A,$6A,$78,$53,$85,$7C,$63,$F9,$77,$6F,$F3,$FE,$DF
	.dc.b	$E8,$0D,$C3,$FA,$03,$67,$71,$FD,$0E,$76,$D3,$65,$8D,$9F,$CE,$E3
	.dc.b	$6D,$8D,$B4,$DE,$3F,$CC,$27,$6D,$9D,$B6,$7F,$A6,$FE,$FF,$FE,$40
	.dc.b	$93,$4F,$F0,$8B,$83,$FA,$F3,$96,$F6,$61,$87,$F6,$68,$E5,$18,$9E
	.dc.b	$F3,$6B,$FA,$9D,$76,$3A,$47,$13,$A9,$49,$3C,$9E,$D5,$CD,$44,$8F
	.dc.b	$AD,$49,$CC,$FF,$86,$B3,$FD,$1F,$B7,$98,$FE,$33,$7C,$2B,$6D,$8E
	.dc.b	$95,$C2,$7F,$C5,$33,$99,$E9,$18,$9D,$47,$47,$B5,$F7,$8D,$9F,$E2
	.dc.b	$5A,$FC,$63,$E6,$32,$57,$2F,$92,$79,$36,$BF,$C5,$CF,$03,$07,$58
	.dc.b	$DB,$EA,$8D,$CF,$48,$C1,$B8,$FD,$E7,$C7,$FA,$1E,$77,$FA,$F8,$C4
	.dc.b	$F0,$8C,$4F,$C1,$38,$96,$3F,$86,$4F,$07,$E0,$CF,$07,$F2,$7E,$BF
	.dc.b	$E3,$BD,$6F,$C7,$B6,$FF,$18,$D9,$DC,$7E,$3E,$76,$D3,$65,$8D,$9F
	.dc.b	$A0,$8D,$B6,$36,$D3,$75,$F9,$69,$DB,$67,$6D,$9F,$D5,$FA,$DF,$A6
	.dc.b	$3F,$4F,$EA,$DA,$FD,$24,$59,$FD,$8C,$DA,$FD,$1C,$DA,$FF,$13,$6A
	.dc.b	$AC,$5F,$E6,$E1,$58,$AF,$DC,$D5,$9C,$C7,$80,$FD,$A4,$66,$74,$F4
	.dc.b	$AF,$F3,$EF,$B5,$F5,$74,$3F,$D2,$3E,$3F,$0C,$C3,$F4,$2E,$5A,$4A
	.dc.b	$FD,$5C,$65,$B1,$97,$F3,$91,$F9,$47,$C3,$1F,$94,$8C,$57,$8A,$30
	.dc.b	$49,$FD,$44,$E8,$91,$98,$EA,$EF,$2F,$E8,$DE,$EF,$A6,$7F,$AF,$36
	.dc.b	$33,$3E,$96,$CF,$A0,$DC,$64,$FE,$97,$3F,$C9,$3D,$8F,$E4,$BE,$BD
	.dc.b	$9F,$E4,$4D,$AF,$E5,$ED,$BF,$E4,$2D,$9F,$BE,$D9,$FB,$CD,$8F,$E6
	.dc.b	$AD,$D4,$D2,$47,$F3,$1B,$F1,$B4,$F1,$9F,$CD,$7C,$1F,$E7,$67,$F3
	.dc.b	$CF,$5A,$C6,$49,$87,$F4,$68,$CB,$FD,$22,$6E,$3F,$CE,$A6,$D7,$E6
	.dc.b	$26,$D7,$E5,$E6,$CF,$F4,$E9,$B8,$A6,$C6,$1E,$35,$FE,$D0,$F8,$9F
	.dc.b	$CA,$1F,$E1,$AF,$8F,$B8,$7F,$32,$8B,$3F,$E1,$99,$B8,$FF,$2F,$D9
	.dc.b	$58,$BF,$F0,$1D,$5B,$FE,$91,$3F,$90,$24,$FF,$26,$6C,$7F,$37,$6B
	.dc.b	$FC,$D5,$AF,$F4,$86,$BD,$23,$91,$FE,$27,$F8,$FF,$07,$5F,$F8,$31
	.dc.b	$BD,$66,$92,$7D,$C7,$13,$FB,$77,$B3,$FD,$D3,$D8,$FE,$EA,$D8,$C9
	.dc.b	$FD,$CE,$32,$D3,$38,$7F,$80,$A3,$2D,$8C,$E9,$FD,$DA,$72,$D8,$C9
	.dc.b	$FC,$5A,$32,$D3,$33,$FC,$5E,$32,$D8,$FF,$03,$7B,$1F,$E0,$C3,$FC
	.dc.b	$1B,$EB,$D8,$C9,$27,$FC,$2F,$19,$7F,$C3,$33,$6F,$FC,$17,$36,$BF
	.dc.b	$94,$CD,$AF,$E5,$13,$67,$FC,$3F,$37,$85,$36,$30,$F1,$A7,$E0,$DF
	.dc.b	$E4,$3B,$FF,$07,$E1,$FD,$EE,$31,$3F,$E6,$9E,$6F,$EB,$D1,$89,$D6
	.dc.b	$D8,$3F,$CE,$73,$6F,$99,$FE,$16,$AE,$0D,$9D,$BA,$BF,$F0,$FC,$E9
	.dc.b	$B3,$FC,$AE,$3C,$71,$FE,$0B,$6B,$CE,$7F,$C0,$F1,$88,$E3,$18,$9E
	.dc.b	$44,$70,$48,$FF,$0B,$1F,$12,$34,$FF,$A2,$7E,$43,$D8,$D9,$F4,$3B
	.dc.b	$7F,$D1,$11,$AC,$3F,$CF,$4D,$8F,$F3,$BB,$67,$FC,$E5,$BA,$3F,$CA
	.dc.b	$7E,$CF,$F9,$53,$C3,$B7,$FE,$51,$6D,$FF,$97,$9B,$87,$F9,$41,$B3
	.dc.b	$FD,$B9,$B3,$FD,$B5,$B1,$FE,$64,$6E,$DF,$EA,$09,$FF,$34,$DF,$86
	.dc.b	$BF,$CC,$0D,$8C,$C7,$8C,$C9,$5F,$F3,$8E,$3F,$E6,$C3,$FC,$DB,$EC
	.dc.b	$59,$FF,$35,$45,$9F,$F3,$BC,$DC,$7F,$CD,$33,$6B,$FC,$0F,$36,$BF
	.dc.b	$C0,$D3,$67,$FC,$FB,$37,$0F,$F4,$DC,$76,$3F,$CF,$11,$99,$D3,$CE
	.dc.b	$BF,$43,$ED,$7E,$70,$FF,$65,$FC,$85,$BF,$BB,$B2,$71,$5F,$EC,$BA
	.dc.b	$B9,$7E,$9E,$6D,$FF,$AD,$26,$D7,$E7,$27,$F3,$44,$9F,$CB,$B6,$3F
	.dc.b	$23,$7D,$E7,$24,$8C,$CE,$DF,$D1,$35,$FD,$73,$FE,$E4,$F8,$1F,$0E
	.dc.b	$FF,$EB,$E6,$E6,$69,$2F,$DF,$5C,$5F,$F5,$07,$B1,$FE,$A6,$F4,$BF
	.dc.b	$D4,$ED,$8C,$9F,$EA,$58,$CB,$4C,$E3,$FE,$B4,$8C,$B6,$33,$A7,$FA
	.dc.b	$A2,$72,$D8,$C9,$FE,$48,$8C,$B4,$CC,$FF,$92,$63,$2D,$8F,$F5,$C7
	.dc.b	$A3,$FE,$BE,$3F,$CF,$AB,$C9,$FF,$7E,$DF,$FA,$E2,$75,$3C,$0F,$E2
	.dc.b	$86,$FC,$CC,$DF,$FB,$3A,$6D,$7F,$AB,$EB,$47,$19,$D3,$63,$57,$E0
	.dc.b	$8F,$F3,$1D,$72,$69,$E0,$8D,$57,$F9,$96,$74,$D9,$D5,$E2,$3F,$9D
	.dc.b	$FA,$7F,$CD,$8F,$F4,$DE,$DF,$EF,$EF,$90,$C5,$FE,$72,$6D,$32,$FF
	.dc.b	$64,$5E,$EB,$44,$9D,$3F,$CE,$EB,$AE,$EC,$39,$E1,$FD,$26,$7F,$9C
	.dc.b	$12,$BF,$9A,$B6,$FF,$3E,$D7,$F9,$E3,$5E,$D1,$C0,$8F,$07,$FA,$46
	.dc.b	$0C,$1F,$21,$E2,$9F,$F7,$4C,$E6,$31,$5F,$CC,$1B,$19,$9F,$76,$1F
	.dc.b	$0A,$F9,$C6,$0E,$6E,$2F,$E1,$9F,$EF,$3D,$3F,$DD,$93,$F8,$DF,$33
	.dc.b	$F1,$F0,$F3,$6C,$49,$FB,$CE,$A3,$15,$D2,$79,$1A,$3F,$09,$1A,$C3
	.dc.b	$F1,$75,$65,$5E,$3A,$C3,$E5,$9B,$3F,$A1,$9B,$7E,$82,$38,$F3,$7D
	.dc.b	$3E,$4F,$A9,$B5,$F9,$56,$E1,$E7,$B4,$63,$1F,$CF,$B6,$BC,$A7,$EC
	.dc.b	$5C,$57,$C9,$6C,$64,$FD,$15,$F8,$5A,$78,$4B,$3E,$19,$E3,$7E,$29
	.dc.b	$CD,$9F,$8B,$4F,$C1,$B1,$1F,$DE,$9B,$1F,$73,$4F,$C6,$7C,$27,$FB
	.dc.b	$F6,$73,$7C,$EB,$E7,$C6,$2F,$AB,$63,$FB,$ED,$71,$98,$77,$3F,$52
	.dc.b	$E2,$3F,$16,$79,$63,$FD,$95,$3E,$48,$F4,$99,$8F,$F0,$BB,$4F,$F0
	.dc.b	$CC,$6D,$F1,$C7,$22,$1C,$23,$51,$EA,$8C,$BC,$9B,$5F,$A3,$F1,$3C
	.dc.b	$CD,$8F,$51,$FB,$2D,$BF,$D2,$6E,$2B,$E5,$46,$1F,$6D,$FC,$5C,$7F
	.dc.b	$71,$87,$4B,$FD,$6B,$FA,$8F,$2E,$47,$F2,$3F,$1A,$E1,$F8,$F9,$B9
	.dc.b	$7F,$2C,$9B,$5F,$8E,$D1,$58,$BF,$DB,$55,$AF,$E8,$53,$69,$36,$30
	.dc.b	$D9,$FE,$69,$1F,$BC,$8E,$1E,$37,$EF,$BC,$2F,$DF,$B6,$BF,$9D,$37
	.dc.b	$0F,$D3,$37,$1F,$E8,$6D,$9F,$43,$67,$F8,$63,$63,$FA,$C3,$71,$FF
	.dc.b	$56,$47,$F4,$7B,$F0,$B4,$F0,$96,$BC,$2D,$BF,$CD,$47,$E8,$FC,$BF
	.dc.b	$CF,$1F,$CF,$3E,$46,$D7,$E9,$62,$CF,$DC,$9B,$5C,$B6,$FF,$7C,$DE
	.dc.b	$CF,$C2,$4E,$9F,$D2,$DF,$E7,$08,$FE,$4D,$B7,$F9,$A6,$BF,$A0,$AE
	.dc.b	$AE,$67,$81,$1C,$12,$FF,$9D,$7C,$2F,$86,$3F,$CD,$ED,$DF,$FA,$43
	.dc.b	$78,$0D,$25,$FF,$73,$F1,$7F,$9D,$9F,$CF,$3C,$1B,$19,$26,$1F,$C8
	.dc.b	$E3,$2F,$F2,$59,$B8,$FF,$80,$A6,$D7,$E5,$E6,$D7,$F4,$B9,$B3,$FE
	.dc.b	$17,$9B,$B2,$6C,$61,$CF,$81,$FD,$43,$AF,$FA,$8B,$6F,$FC,$18,$DC
	.dc.b	$3F,$9C,$37,$1F,$F0,$AB,$67,$F5,$4D,$9F,$EB,$CD,$8F,$F1,$53,$71
	.dc.b	$FF,$26,$47,$F8,$66,$FC,$2D,$3C,$25,$AF,$0B,$6F,$F1,$5E,$27,$F2
	.dc.b	$F3,$FC,$1B,$F2,$36,$7F,$9B,$45,$AF,$E1,$B3,$6B,$8D,$68,$FE,$29
	.dc.b	$19,$7F,$9A,$D7,$F2,$C2,$3F,$C8,$DB,$5F,$CA,$9A,$FE,$45,$AF,$28
	.dc.b	$E0,$47,$83,$FC,$8F,$E4,$7C,$11,$FE,$BE,$6E,$FF,$E1,$76,$EE,$69
	.dc.b	$2B,$FC,$85,$DB,$FE,$0A,$3F,$C1,$7D,$36,$32,$4C,$3F,$B9,$46,$5F
	.dc.b	$EE,$93,$73,$FF,$28,$CD,$AF,$E4,$D3,$6B,$FC,$35,$36,$7F,$CC,$D3
	.dc.b	$75,$4D,$8C,$39,$E9,$FF,$0F,$F2,$FF,$88,$1B,$7F,$E5,$B6,$E1,$FE
	.dc.b	$07,$6E,$3F,$E6,$46,$CF,$F3,$96,$CF,$F8,$B5,$B1,$FE,$79,$6E,$3F
	.dc.b	$D3,$6B,$FC,$DB,$7E,$1A,$FF,$35,$36,$FF,$C2,$CD,$9D,$CF,$FB,$EB
	.dc.b	$F0,$6F,$F3,$7D,$1F,$F3,$8E,$BF,$E4,$DF,$88,$FF,$00,$5F,$F2,$89
	.dc.b	$D9,$3C,$1F,$F3,$1C,$EB,$0F,$F0,$7C,$E5,$ED,$3A,$69,$30,$FE,$FA
	.dc.b	$78,$63,$F9,$AB,$C9,$AF,$F8,$FA,$51,$88,$E1,$3F,$E0,$49,$CB,$67
	.dc.b	$51,$89,$EE,$D7,$FC,$1A,$D7,$C1,$3C,$48,$FF,$89,$7F,$5D,$FF,$1D
	.dc.b	$BF,$E1,$89,$F8,$F1,$8C,$3E,$35,$FF,$A4,$A3,$17,$D2,$F9,$1F,$1C
	.dc.b	$F6,$4E,$DC,$1E,$C3,$71,$FE,$A5,$E7,$FF,$36,$9F,$E6,$EE,$3B,$19
	.dc.b	$24,$FF,$95,$63,$2F,$F9,$5E,$6E,$1F,$EB,$49,$B5,$FE,$09,$9B,$5F
	.dc.b	$E7,$D9,$B3,$FE,$C9,$9B,$AA,$6C,$61,$CF,$0F,$FA,$0B,$4F,$F4,$1B
	.dc.b	$6F,$FD,$7C,$DC,$3F,$CD,$4D,$C7,$FD,$8E,$D9,$FF,$0E,$36,$7F,$D2
	.dc.b	$0D,$8F,$F6,$E3,$71,$FE,$19,$1F,$EC,$BB,$F0,$B4,$F0,$96,$BC,$2D
	.dc.b	$BF,$CF,$75,$FD,$2D,$7F,$C1,$F5,$F8,$1D,$B1,$32,$FF,$54,$FD,$F8
	.dc.b	$C4,$FE,$16,$2D,$FF,$B3,$B6,$56,$2F,$F4,$F5,$F9,$62,$3F,$9B,$6D
	.dc.b	$FE,$61,$AF,$E6,$1A,$F2,$8E,$04,$78,$3F,$EC,$8A,$7C,$17,$86,$7C
	.dc.b	$71,$89,$F5,$9B,$9D,$5A,$75,$5E,$DA,$D1,$C1,$FD,$35,$7E,$76,$1F
	.dc.b	$C0,$1F,$A2,$9F,$F4,$A3,$63,$37,$B8,$C1,$27,$11,$96,$4F,$FB,$03
	.dc.b	$1F,$F4,$A3,$DA,$36,$E4,$FF,$58,$D7,$37,$11,$E2,$36,$7F,$B4,$63
	.dc.b	$4D,$33,$3D,$A3,$2D,$8C,$CF,$67,$D5,$7E,$BF,$92,$B3,$E0,$3E,$1B
	.dc.b	$89,$F0,$36,$BB,$36,$7D,$2D,$9F,$92,$D8,$E6,$75,$BF,$AB,$5C,$A3
	.dc.b	$17,$E3,$69,$BA,$E5,$46,$9F,$A1,$E4,$93,$F8,$98,$B8,$7F,$21,$9B
	.dc.b	$87,$E7,$66,$D7,$F4,$29,$B5,$F8,$E9,$AF,$F4,$F8,$FE,$6A,$49,$FE
	.dc.b	$74,$D8,$FE,$6E,$D7,$F9,$BB,$5E,$D1,$C0,$8F,$07,$F4,$1F,$25,$E2
	.dc.b	$CB,$F3,$4D,$CC,$D2,$4F,$F6,$5B,$FE,$75,$1F,$95,$24,$7C,$FF,$8B
	.dc.b	$B8,$7E,$26,$6E,$5F,$87,$9B,$5F,$92,$9B,$5F,$A2,$9B,$3F,$A6,$9B
	.dc.b	$9A,$6C,$61,$B3,$F7,$E3,$F5,$11,$C3,$E4,$BF,$55,$F2,$5F,$AB,$6D
	.dc.b	$7E,$15,$B8,$7E,$7D,$B8,$FE,$75,$B3,$FA,$86,$CF,$EE,$1B,$1F,$BC
	.dc.b	$6E,$26,$92,$3F,$29,$7E,$36,$9F,$B2,$AF,$E2,$67,$F6,$0D,$BF,$83
	.dc.b	$6B,$FC,$3F,$2C,$4C,$AE,$15,$AD,$7E,$34,$62,$74,$33,$FF,$5F,$D6
	.dc.b	$EF,$31,$89,$FF,$05,$4F,$F7,$C4,$43,$CB,$5F,$E1,$66,$9F,$E1,$88
	.dc.b	$E0,$ED,$AF,$26,$BC,$0E,$85,$38,$1E,$22,$47,$78,$F7,$65,$E8,$8F
	.dc.b	$2E,$3F,$CC,$E3,$15,$FC,$E9,$B1,$99,$F8,$87,$F4,$18,$CE,$3F,$5E
	.dc.b	$3F,$84,$79,$BF,$CC,$4F,$E6,$5F,$25,$63,$24,$C3,$F9,$34,$65,$FE
	.dc.b	$51,$37,$2F,$E4,$D3,$6B,$F9,$8C,$DA,$FE,$83,$36,$7F,$A2,$CD,$D5
	.dc.b	$36,$30,$E7,$E4,$BF,$A3,$7C,$7F,$F4,$76,$DF,$F2,$16,$E1,$FC,$ED
	.dc.b	$B8,$FF,$39,$6C,$FF,$44,$6C,$FF,$54,$6C,$7F,$58,$6E,$26,$92,$3F
	.dc.b	$98,$5F,$8D,$A7,$E5,$B7,$FC,$E1,$FE,$43,$2B,$92,$5F,$F9,$36,$32
	.dc.b	$FE,$66,$3F,$19,$18,$C3,$FC,$14,$D8,$C9,$F8,$77,$F3,$64,$70,$49
	.dc.b	$FC,$C4,$65,$B1,$97,$F2,$BB,$27,$51,$87,$FC,$B9,$9F,$86,$3F,$C1
	.dc.b	$2D,$DB,$FC,$0C,$DD,$4D,$26,$1F,$D5,$FE,$2F,$FC,$00,$7F,$80,$7E
	.dc.b	$2E,$C6,$49,$87,$F7,$58,$CB,$FD,$DE,$6E,$7F,$DD,$66,$D7,$F8,$02
	.dc.b	$6D,$7F,$84,$26,$CF,$F8,$52,$6E,$C9,$B1,$87,$3F,$17,$FE,$15,$F4
	.dc.b	$3F,$C2,$CD,$BF,$EE,$0D,$C3,$FC,$14,$DC,$7F,$C1,$0D,$9F,$F0,$9B
	.dc.b	$67,$FC,$46,$D8,$FF,$13,$B7,$13,$49,$1F,$DF,$EF,$C6,$D3,$F9,$2C
	.dc.b	$7F,$2B,$3F,$D2,$5C,$36,$32,$4B,$FF,$4E,$46,$5F,$C8,$1F,$C6,$67
	.dc.b	$18,$7F,$96,$66,$D7,$F1,$F9,$FE,$52,$47,$F8,$C3,$6B,$F9,$33,$5F
	.dc.b	$C6,$B5,$E5,$1C,$08,$F0,$7F,$CB,$1C,$3E,$08,$FF,$2B,$B7,$6F,$F2
	.dc.b	$A3,$76,$FD,$B5,$7B,$71,$FF,$38,$4E,$4F,$F2,$77,$C4,$C9,$FF,$14
	.dc.b	$7C,$55,$C7,$FC,$85,$36,$FF,$C8,$13,$6B,$FC,$97,$36,$BF,$CB,$73
	.dc.b	$67,$FC,$BF,$37,$54,$D8,$C3,$67,$FC,$71,$1F,$E6,$18,$E1,$F1,$3F
	.dc.b	$E6,$3F,$89,$FF,$32,$36,$BF,$C7,$CD,$C3,$FC,$B0,$DC,$7F,$CA,$CD
	.dc.b	$9F,$F3,$0B,$67,$FC,$E2,$D8,$FF,$3A,$B7,$13,$49,$1F,$E4,$DB,$F1
	.dc.b	$B4,$FE,$E7,$DD,$FD,$E4,$FF,$70,$73,$58,$C9,$30,$FF,$75,$46,$5F
	.dc.b	$C7,$9F,$E9,$49,$C6,$3F,$EA,$79,$B5,$FD,$B6,$7F,$BB,$91,$FE,$CC
	.dc.b	$DA,$FE,$E8,$D7,$F8,$DB,$5E,$51,$C0,$8F,$07,$FD,$4D,$C7,$E0,$8F
	.dc.b	$F5,$2B,$77,$FF,$51,$37,$73,$49,$5F,$E7,$3F,$88,$FF,$4E,$1F,$E9
	.dc.b	$DC,$2C,$64,$98,$7F,$A4,$63,$2F,$FA,$4E,$6E,$5F,$E9,$19,$B5,$FE
	.dc.b	$9C,$9B,$5F,$EA,$C9,$B3,$FE,$B0,$9B,$AA,$6C,$61,$DF,$F0,$0A,$BF
	.dc.b	$D7,$0B,$E0,$AC,$61,$AF,$FA,$6E,$34,$D3,$45,$35,$1C,$30,$FE,$1B
	.dc.b	$09,$E3,$18,$7C,$51,$C0,$E5,$19,$9E,$58,$FF,$B0,$2B,$8B,$6B,$C3
	.dc.b	$5C,$DB,$19,$7A,$47,$E4,$3F,$09,$7C,$37,$C1,$8C,$36,$BF,$DF,$1D
	.dc.b	$16,$FF,$C8,$97,$C1,$FF,$2D,$36,$F8,$C7,$FB,$BE,$78,$47,$F9,$32
	.dc.b	$75,$5D,$E1,$18,$AE,$B7,$C3,$7F,$F7,$86,$1C,$AF,$73,$6F,$C1,$1F
	.dc.b	$E4,$C6,$BC,$AF,$FC,$B1,$18,$8F,$14,$F8,$5D,$1F,$E5,$43,$AC,$F0
	.dc.b	$8D,$47,$08,$C1,$1E,$13,$D5,$E2,$7F,$4A,$B6,$FF,$D6,$AB,$E3,$DF
	.dc.b	$B6,$31,$5F,$20,$E0,$E3,$1F,$8D,$87,$C5,$96,$1D,$39,$BE,$16,$1C
	.dc.b	$DB,$8F,$FB,$FA,$74,$FE,$5F,$E2,$BF,$3A,$7E,$7B,$E2,$AC,$64,$98
	.dc.b	$7E,$4E,$32,$FE,$56,$6E,$5F,$91,$9B,$5F,$E8,$83,$F3,$93,$8A,$FD
	.dc.b	$66,$AA,$C5,$68,$EF,$7C,$DA,$73,$95,$F0,$79,$46,$AF,$AF,$C5,$7E
	.dc.b	$E3,$E2,$BF,$72,$D9,$FD,$B3,$67,$F5,$CD,$C3,$F5,$4D,$9F,$03,$67
	.dc.b	$AC,$7E,$FE,$BA,$B4,$EA,$58,$EA,$DA,$EA,$D3,$A9,$63,$F4,$9F,$0B
	.dc.b	$F9,$B3,$F0,$9D,$B6,$7F,$5B,$16,$7F,$0B,$36,$BF,$0B,$36,$BF,$D4
	.dc.b	$BC,$0A,$C5,$F6,$9F,$D1,$D7,$E7,$23,$F6,$2D,$E0,$34,$93,$C0,$93
	.dc.b	$FA,$08,$CB,$63,$2F,$EC,$E3,$F3,$DD,$7E,$2D,$FF,$46,$DC,$4D,$25
	.dc.b	$FC,$03,$F3,$51,$A7,$F9,$87,$0F,$F3,$63,$F9,$BF,$AF,$63,$24,$C7
	.dc.b	$F9,$64,$65,$FE,$5D,$37,$4F,$E5,$53,$6B,$F2,$13,$6B,$F3,$33,$67
	.dc.b	$F9,$C4,$DD,$D3,$63,$0E,$7E,$2B,$FA,$5F,$C5,$7F,$4C,$6D,$FF,$49
	.dc.b	$6E,$1F,$D0,$5B,$8F,$F3,$D6,$CF,$E8,$DB,$3F,$AC,$6C,$7F,$4B,$6E
	.dc.b	$26,$92,$3F,$AC,$5F,$8D,$A7,$F3,$AF,$81,$FE,$58,$7F,$15,$EF,$B1
	.dc.b	$92,$61,$FC,$5A,$32,$FF,$4B,$9B,$8F,$F4,$BD,$BF,$90,$D6,$2F,$FA
	.dc.b	$4D,$7E,$3C,$8F,$F3,$66,$DF,$F2,$C6,$BF,$CF,$5A,$F2,$8E,$04,$78
	.dc.b	$3F,$88,$EF,$F0,$47,$F2,$8D,$BF,$C7,$B3,$EA,$C7,$CD,$18,$9F,$54
	.dc.b	$F1,$34,$7E,$B2,$B9,$E3,$FC,$F1,$FF,$03,$FC,$54,$9F,$F0,$27,$C5
	.dc.b	$5B,$FE,$F9,$37,$1F,$EF,$13,$6B,$F8,$F4,$DA,$FE,$4F,$36,$7F,$C0
	.dc.b	$D3,$76,$4D,$8C,$36,$7F,$C2,$91,$FE,$1A,$8E,$1F,$15,$FE,$1C,$E3
	.dc.b	$FF,$0E,$B6,$BF,$C3,$4D,$C3,$FC,$24,$DC,$7F,$C1,$CD,$9F,$E6,$8D
	.dc.b	$9F,$E7,$AD,$8F,$F0,$E3,$71,$34,$91,$FE,$29,$BF,$1B,$45,$7F,$99
	.dc.b	$3F,$09,$7C,$07,$F8,$05,$B3,$FC,$13,$C4,$C4,$C8,$E3,$97,$F8,$36
	.dc.b	$B7,$7F,$DB,$2F,$4F,$5B,$B5,$F8,$29,$B7,$E2,$A5,$F2,$70,$62,$FF
	.dc.b	$95,$46,$63,$83,$5F,$5B,$5E,$13,$FE,$0F,$69,$FE,$11,$8D,$12,$7A
	.dc.b	$92,$78,$47,$B2,$FF,$33,$E0,$FA,$2F,$F3,$CD,$DF,$FC,$22,$DE,$03
	.dc.b	$49,$87,$F9,$6F,$DA,$FF,$30,$9F,$E6,$2F,$8B,$B1,$92,$61,$FE,$57
	.dc.b	$8C,$BF,$E5,$99,$B8,$FF,$95,$66,$D7,$F7,$B9,$B5,$FE,$04,$9B,$3F
	.dc.b	$E6,$39,$BB,$A6,$C6,$1C,$FB,$7F,$E7,$6F,$8A,$FF,$3B,$B6,$FF,$CE
	.dc.b	$8D,$C3,$FC,$D8,$DC,$7F,$CD,$0D,$9F,$F0,$73,$67,$FC,$2E,$D8,$FF
	.dc.b	$3B,$37,$13,$49,$1F,$E8,$3B,$F1,$B4,$FC,$99,$F8,$4F,$73,$F3,$47
	.dc.b	$F9,$07,$C6,$B1,$AC,$3F,$0F,$38,$7F,$C8,$73,$71,$FC,$E4,$DA,$FE
	.dc.b	$FF,$36,$B8,$C7,$E1,$A7,$F2,$A4,$70,$48,$FC,$84,$65,$B1,$97,$F3
	.dc.b	$FB,$A7,$51,$87,$FC,$ED,$FB,$07,$FA,$62,$3F,$71,$18,$AF,$27,$02
	.dc.b	$B5,$8F,$AA,$7F,$CA,$D1,$87,$E3,$18,$38,$B9,$BF,$F2,$5F,$C5,$7F
	.dc.b	$AD,$8F,$F5,$BF,$C5,$5A,$FF,$5A,$45,$9F,$F5,$64,$DB,$FF,$54,$4D
	.dc.b	$AF,$F2,$74,$DA,$EB,$1C,$8F,$F5,$F4,$65,$FF,$5C,$4D,$D9,$36,$30
	.dc.b	$E7,$AB,$FD,$43,$4F,$A6,$7B,$DF,$EC,$E9,$CC,$62,$7F,$D9,$91,$89
	.dc.b	$D4,$7E,$14,$E4,$E3,$1F,$F6,$5D,$71,$7F,$D9,$55,$C6,$B9,$9C,$67
	.dc.b	$59,$7F,$B2,$67,$55,$C2,$B5,$7D,$6B,$FD,$89,$1E,$03,$9C,$F8,$27
	.dc.b	$51,$FE,$81,$BE,$AF,$F2,$2F,$6F,$F9,$C1,$F8,$6F,$32,$C7,$E7,$EF
	.dc.b	$F9,$6C,$E1,$FF,$54,$6B,$FE,$DC,$9E,$46,$F8,$15,$BC,$BF,$C1,$D5
	.dc.b	$BB,$FE,$5F,$3C,$DF,$CC,$BF,$CD,$89,$7F,$CA,$9B,$1F,$CC,$5A,$F0
	.dc.b	$8D,$11,$D3,$6B,$FD,$D1,$F8,$5B,$EB,$1E,$F7,$96,$BF,$AD,$CE,$63
	.dc.b	$16,$8C,$4F,$B7,$0F,$F7,$3C,$62,$7F,$B1,$46,$2F,$AD,$6C,$FE,$11
	.dc.b	$5B,$C3,$EB,$E8,$F8,$DF,$C6,$FC,$6F,$E3,$9B,$7F,$83,$6E,$5F,$8A
	.dc.b	$6D,$F2,$6C,$FD,$D6,$CA,$3F,$39,$18,$69,$F8,$FA,$F0,$34,$F0,$16
	.dc.b	$7C,$0D,$AF,$03,$4F,$C6,$7C,$67,$E9,$8F,$D3,$FC,$6D,$8F,$CE,$45
	.dc.b	$AF,$D1,$CD,$AF,$87,$36,$BF,$2D,$36,$BD,$58,$AB,$15,$FB,$7A,$B3
	.dc.b	$98,$EE,$7E,$AE,$33,$3A,$7A,$53,$F0,$B7,$BB,$F1,$E1,$8D,$75,$1F
	.dc.b	$CA,$BD,$2D,$4E,$6B,$8E,$3F,$A7,$6D,$F2,$8F,$F0,$B4,$62,$7F,$47
	.dc.b	$5C,$A7,$37,$6F,$FC,$5B,$7B,$9C,$9F,$CC,$AB,$53,$E4,$8F,$05,$7F
	.dc.b	$B1,$EB,$C3,$3F,$E2,$26,$BC,$8F,$F1,$29,$C4,$AE,$9F,$0C,$73,$24
	.dc.b	$FF,$88,$E7,$53,$C4,$ED,$3A,$7F,$82,$FA,$7E,$69,$F5,$43,$AB,$FA
	.dc.b	$74,$E6,$BF,$AA,$B6,$FF,$A7,$B6,$B9,$C7,$B8,$C9,$FC,$9F,$E3,$BF
	.dc.b	$9A,$F8,$9F,$CD,$9B,$19,$3F,$96,$46,$5A,$66,$7F,$9A,$C6,$5B,$19
	.dc.b	$CF,$F6,$F3,$96,$C6,$4F,$E5,$91,$96,$99,$9F,$CC,$C6,$5B,$1F,$CF
	.dc.b	$FE,$0F,$FA,$69,$FD,$3B,$E3,$AC,$64,$93,$FD,$2A,$32,$FF,$4C,$9B
	.dc.b	$7F,$C3,$26,$D7,$F3,$F9,$B5,$FA,$A9,$B3,$FD,$7E,$6F,$02,$6C,$61
	.dc.b	$E3,$5F,$A7,$F6,$3E,$E1,$FE,$0C,$F5,$EE,$1F,$A1,$8B,$3F,$D1,$A6
	.dc.b	$E3,$FA,$79,$B5,$FE,$8A,$DB,$FC,$21,$58,$BF,$D0,$D7,$E6,$C9,$5F
	.dc.b	$91,$6C,$7E,$61,$AF,$E7,$DA,$FE,$C1,$AF,$48,$E4,$7F,$87,$3D,$7F
	.dc.b	$05,$FF,$88,$5B,$D4,$69,$27,$DC,$71,$3F,$BC,$7C,$6F,$F8,$1B,$E3
	.dc.b	$7F,$C0,$ED,$8C,$9F,$DF,$63,$2D,$33,$87,$F8,$1A,$32,$D8,$CE,$7F
	.dc.b	$D2,$27,$2D,$8C,$9F,$DF,$63,$2D,$33,$3F,$CA,$A3,$2D,$8F,$F0,$87
	.dc.b	$C6,$FF,$87,$8F,$F0,$FF,$C6,$D8,$C9,$27,$FC,$37,$19,$7F,$C3,$B3
	.dc.b	$6F,$FA,$EC,$DA,$FF,$08,$4D,$AF,$E7,$73,$67,$FC,$61,$37,$81,$36
	.dc.b	$30,$F1,$AF,$E7,$1E,$67,$F0,$D3,$FC,$BB,$EC,$5C,$3F,$99,$45,$9F
	.dc.b	$F0,$BC,$DC,$7F,$9C,$4D,$AF,$C9,$6A,$AC,$5F,$F3,$2A,$FE,$58,$4A
	.dc.b	$FE,$44,$D8,$FE,$52,$D7,$F9,$83,$5F,$E8,$2D,$7A,$47,$23,$FC,$DB
	.dc.b	$EC,$78,$37,$FF,$39,$37,$03,$49,$3E,$E3,$89,$FE,$49,$F2,$7F,$CA
	.dc.b	$FF,$0F,$FE,$58,$6C,$64,$FF,$27,$46,$5A,$67,$0F,$F2,$BC,$65,$B1
	.dc.b	$9C,$FF,$C3,$33,$96,$C6,$4F,$F2,$74,$65,$A6,$67,$FB,$DC,$65,$B1
	.dc.b	$FE,$61,$F8,$DF,$F3,$79,$FE,$70,$F8,$DB,$19,$24,$FF,$9B,$23,$2F
	.dc.b	$F9,$BA,$6D,$FF,$8B,$A6,$D7,$F9,$86,$6D,$7F,$83,$26,$CF,$FA,$06
	.dc.b	$6F,$02,$6C,$61,$E3,$4F,$03,$F8,$35,$FF,$99,$DB,$1F,$E6,$6F,$6F
	.dc.b	$FB,$BC,$63,$1F,$F0,$63,$69,$52,$30,$FF,$67,$D5,$7C,$B9,$65,$F7
	.dc.b	$7F,$D7,$B1,$8A,$FF,$08,$1C,$E3,$FC,$34,$F6,$3D,$0E,$1D,$4F,$F8
	.dc.b	$11,$CB,$6F,$D7,$1C,$4C,$9F,$E0,$A7,$2F,$56,$BD,$0D,$4F,$77,$89
	.dc.b	$CE,$36,$F4,$C3,$FD,$0F,$ED,$FA,$DB,$7E,$F3,$FD,$A7,$19,$7F,$72
	.dc.b	$D9,$F9,$0D,$E0,$FF,$6E,$3F,$EB,$AF,$8D,$FF,$5D,$FC,$6D,$BF,$F5
	.dc.b	$83,$6F,$FD,$6E,$DC,$3F,$CE,$ED,$9F,$F5,$53,$67,$FC,$AA,$D8,$FF
	.dc.b	$65,$37,$73,$49,$1F,$EB,$FB,$F1,$B4,$F1,$9F,$EC,$3F,$27,$FD,$A2
	.dc.b	$7F,$B4,$BC,$9B,$19,$26,$1F,$EC,$E8,$CB,$FE,$D0,$9B,$8F,$FA,$42
	.dc.b	$6D,$7F,$B0,$E6,$D7,$F9,$BA,$6C,$FF,$B8,$E6,$E3,$FE,$BC,$8E,$F7
	.dc.b	$FE,$9A,$9C,$D7,$2B,$FD,$27,$B3,$FA,$03,$FC,$B9,$A3,$81,$ED,$E7
	.dc.b	$2F,$D4,$55,$BF,$F6,$6C,$DC,$BF,$4F,$36,$BF,$9B,$CD,$AF,$D4,$4F
	.dc.b	$E3,$77,$34,$95,$F9,$B6,$CF,$E6,$5A,$FE,$4D,$AF,$E6,$DA,$F6,$8F
	.dc.b	$DB,$DF,$FB,$73,$DB,$F1,$65,$ED,$6E,$46,$92,$BD,$E7,$23,$FA,$8B
	.dc.b	$FE,$81,$FA,$4F,$DB,$46,$70,$47,$F8,$9E,$30,$EA,$3F,$7E,$E5,$A7
	.dc.b	$F0,$C8,$DB,$FE,$71,$6B,$DC,$C1,$C0,$C5,$E5,$B1,$C0,$E7,$19,$69
	.dc.b	$98,$F0,$C6,$5B,$19,$C3,$FD,$5D,$5E,$76,$CF,$9C,$D9,$EA,$F8,$AF
	.dc.b	$31,$E7,$F8,$DB,$7E,$C8,$B3,$F0,$26,$D7,$49,$B5,$FC,$BA,$FF,$97
	.dc.b	$D6,$2F,$E6,$8A,$C4,$EA,$BB,$12,$7C,$8F,$CC,$AD,$C7,$18,$D5,$FB
	.dc.b	$27,$B4,$FF,$47,$F5,$3F,$A1,$9F,$A4,$CF,$F2,$F7,$FD,$26,$32,$D3
	.dc.b	$F0,$59,$7E,$A7,$4C,$DF,$F3,$E3,$A4,$ED,$FE,$83,$3B,$8C,$63,$FD
	.dc.b	$42,$FF,$4C,$48,$FE,$7C,$DA,$FE,$78,$D7,$F9,$B3,$5F,$E7,$CD,$7F
	.dc.b	$AC,$3C,$67,$FD,$D5,$5F,$BF,$F7,$BC,$F1,$8C,$51,$8C,$3F,$C6,$2D
	.dc.b	$C3,$FB,$03,$6F,$94,$FA,$8D,$9F,$9C,$F8,$DF,$CD,$7C,$6F,$E6,$DB
	.dc.b	$19,$3F,$43,$19,$69,$9B,$FD,$3C,$65,$B1,$9C,$7F,$2B,$39,$6C,$64
	.dc.b	$FC,$64,$A7,$51,$86,$C7,$E3,$9B,$AF,$F8,$01,$FD,$97,$99,$2B,$F6
	.dc.b	$FF,$1B,$6F,$F7,$73,$70,$FD,$3C,$DA,$F0,$CD,$AF,$CD,$CD,$C3,$FC
	.dc.b	$1B,$36,$53,$63,$0D,$AF,$E0,$F1,$FC,$16,$3A,$4B,$F0,$B7,$9A,$BF
	.dc.b	$E2,$D3,$98,$FC,$E7,$C0,$66,$3F,$2D,$5C,$2B,$6D,$8C,$D7,$09,$CB
	.dc.b	$C7,$2E,$53,$D6,$72,$7E,$93,$76,$1B,$8F,$F0,$F6,$1D,$6F,$8B,$DE
	.dc.b	$FF,$C3,$4D,$77,$87,$F8,$56,$31,$1F,$E3,$46,$BE,$13,$FC,$44,$53
	.dc.b	$9C,$71,$25,$7F,$8A,$E7,$89,$23,$F8,$07,$C0,$79,$EB,$B4,$62,$BD
	.dc.b	$B1,$FB,$AC,$55,$B8,$F1,$61,$FA,$B8,$C6,$1E,$FB,$FF,$4D,$C6,$30
	.dc.b	$E9,$7F,$D1,$3E,$3B,$FA,$07,$C7,$7F,$41,$6D,$FF,$46,$6C,$EE,$3F
	.dc.b	$A4,$CE,$DA,$6C,$B1,$B3,$FA,$8C,$6D,$B1,$B6,$9B,$C7,$FA,$0C,$ED
	.dc.b	$B3,$B6,$CF,$F5,$CF,$8E,$FE,$B0,$7F,$59,$F8,$EB,$5F,$D7,$A2,$CF
	.dc.b	$F6,$29,$B5,$FD,$3A,$6D,$7E,$E2,$6D,$7F,$42,$9B,$C1,$FE,$74,$9B
	.dc.b	$29,$B1,$86,$D7,$84,$FC,$EF,$97,$F5,$0F,$BD,$F3,$0F,$C3,$FF,$D7
	.dc.b	$73,$A2,$61,$FA,$6A,$DC,$EA,$31,$3A,$CB,$F3,$75,$AF,$2F,$B4,$F2
	.dc.b	$AF,$CE,$CF,$E6,$C9,$1F,$93,$6C,$7E,$6B,$83,$FD,$DF,$39,$24,$66
	.dc.b	$76,$F4,$8C,$DF,$EF,$7E,$A1,$F6,$CF,$6F,$C7,$7F,$E9,$29,$CC,$62
	.dc.b	$B8,$56,$EF,$D8,$7F,$A8,$AB,$6E,$2D,$18,$7E,$19,$D6,$79,$3F,$E2
	.dc.b	$0F,$92,$FF,$0E,$1F,$E1,$DF,$56,$DF,$F8,$8A,$2C,$FF,$89,$A6,$E1
	.dc.b	$FE,$12,$9B,$5F,$CF,$26,$D7,$F8,$1A,$6F,$0F,$FA,$E2,$6C,$A6,$C6
	.dc.b	$1B,$5E,$13,$FC,$67,$F2,$3F,$E3,$1F,$91,$FF,$19,$36,$32,$7F,$8E
	.dc.b	$23,$2D,$33,$7F,$E3,$E8,$CB,$63,$38,$FF,$8B,$27,$2D,$8C,$9F,$D6
	.dc.b	$23,$2D,$33,$3F,$E2,$48,$CB,$63,$F9,$CF,$91,$FC,$40,$FF,$0D,$7C
	.dc.b	$5D,$8C,$92,$7F,$8D,$46,$5F,$E8,$73,$6F,$FC,$09,$36,$BF,$9C,$CF
	.dc.b	$F3,$32,$3F,$CA,$DB,$5F,$99,$6B,$FC,$F1,$AF,$28,$E0,$47,$83,$FD
	.dc.b	$27,$E2,$FC,$11,$FE,$EF,$6F,$83,$FE,$97,$6F,$61,$A4,$AF,$F3,$6F
	.dc.b	$C8,$7F,$9A,$CF,$F3,$67,$C8,$58,$C9,$30,$FF,$3A,$C6,$5F,$F3,$BC
	.dc.b	$DC,$FF,$CC,$73,$6B,$FC,$1B,$36,$BF,$CA,$F3,$78,$10,$9C,$29,$B1
	.dc.b	$87,$3F,$21,$FE,$81,$F9,$0F,$F4,$13,$70,$FF,$43,$B6,$FF,$D1,$6D
	.dc.b	$C7,$FC,$EE,$D9,$FF,$10,$B6,$7F,$CD,$6D,$C7,$CA,$D8,$34,$93,$FE
	.dc.b	$96,$BF,$1B,$4F,$F0,$6F,$E1,$9F,$E2,$D6,$F8,$BF,$E6,$9F,$8E,$FE
	.dc.b	$C9,$39,$C3,$FC,$C8,$DC,$BF,$C1,$2D,$BF,$8F,$BF,$FB,$4E,$F8,$E1
	.dc.b	$27,$A9,$FE,$0D,$8C,$61,$D2,$7F,$C2,$6F,$D0,$17,$E1,$70,$62,$36
	.dc.b	$FD,$68,$D3,$C5,$AE,$63,$6F,$1A,$FF,$07,$B4,$FF,$08,$CF,$07,$07
	.dc.b	$81,$C1,$27,$FD,$1D,$E4,$FA,$B1,$FF,$39,$36,$FE,$1F,$02,$31,$83
	.dc.b	$2F,$F7,$C4,$61,$B5,$F1,$A7,$FC,$2D,$3F,$EC,$BF,$6E,$4F,$FB,$43
	.dc.b	$E4,$2D,$7F,$B5,$26,$E3,$FE,$C0,$9B,$5F,$E6,$A9,$B5,$FE,$B4,$9B
	.dc.b	$BF,$E7,$A6,$CA,$6C,$61,$B5,$FE,$E0,$8F,$F6,$F4,$72,$FC,$37,$FC
	.dc.b	$B1,$3F,$EE,$A5,$F1,$DF,$EF,$17,$37,$FE,$EC,$73,$3B,$35,$3B,$69
	.dc.b	$93,$A3,$96,$99,$9E,$91,$96,$C6,$71,$FF,$72,$D7,$82,$7C,$53,$D9
	.dc.b	$AF,$46,$9C,$8F,$D0,$F4,$FE,$90,$FF,$67,$7C,$85,$9F,$D5,$C5,$9F
	.dc.b	$F2,$74,$DB,$FD,$3C,$DA,$FC,$2C,$DA,$FC,$D4,$FE,$5B,$83,$FB,$05
	.dc.b	$7E,$59,$B3,$F9,$69,$D1,$27,$4D,$8D,$46,$08,$E0,$97,$FE,$CC,$FD
	.dc.b	$85,$F2,$3E,$3E,$04,$62,$FD,$A7,$FB,$A1,$C5,$FC,$48,$C5,$71,$9C
	.dc.b	$DF,$C0,$9C,$9F,$85,$F8,$CF,$C2,$1F,$85,$F8,$BB,$19,$26,$1F,$87
	.dc.b	$8C,$BF,$89,$9B,$5F,$72,$6D,$7F,$38,$E1,$6C,$BC,$E0,$9C,$DE,$1D
	.dc.b	$CD,$A4,$D8,$C3,$9F,$88,$FC,$E7,$BF,$F9,$D6,$E3,$FA,$16,$E3,$F9
	.dc.b	$F6,$E5,$F9,$26,$CF,$9F,$14,$62,$BC,$D1,$89,$FD,$65,$F7,$69,$DC
	.dc.b	$B3,$DD,$B7,$DD,$A7,$7B,$ED,$3F,$CB,$BA,$3F,$9F,$9F,$EA,$9F,$0D
	.dc.b	$F6,$CF,$6B,$38,$7F,$48,$AB,$7F,$E5,$3A,$D4,$E4,$91,$FD,$22,$7A
	.dc.b	$4E,$A3,$13,$A7,$F9,$74,$FE,$84,$91,$FA,$16,$C7,$F3,$66,$BC,$23
	.dc.b	$44,$74,$DC,$7F,$DC,$1F,$8F,$BB,$3D,$31,$86,$CF,$F8,$02,$13,$B8
	.dc.b	$F5,$C6,$DA,$6E,$35,$5E,$C1,$3B,$7F,$90,$7C,$3F,$F2,$3F,$2F,$F9
	.dc.b	$23,$6F,$F9,$43,$67,$71,$FC,$AE,$76,$D3,$65,$8D,$9F,$CB,$63,$6D
	.dc.b	$8D,$B4,$DD,$7F,$24,$9D,$B6,$76,$D9,$FE,$71,$F0,$7F,$CE,$CF,$E7
	.dc.b	$9F,$07,$6B,$FA,$04,$59,$FE,$7D,$36,$BF,$98,$CD,$AF,$D0,$4D,$AF
	.dc.b	$CF,$CD,$E4,$FD,$CC,$DE,$24,$D8,$C3,$6F,$C2,$7D,$93,$FC,$15,$F8
	.dc.b	$7F,$EA,$70,$FA,$71,$8E,$23,$FC,$6F,$E4,$6C,$E8,$FF,$20,$D7,$1C
	.dc.b	$F3,$08,$C5,$73,$39,$1E,$18,$E3,$7D,$2B,$8D,$70,$9C,$D7,$07,$C0
	.dc.b	$75,$AE,$11,$D6,$3F,$C4,$26,$5F,$F0,$CC,$FB,$63,$FC,$4A,$D7,$81
	.dc.b	$89,$FF,$10,$C6,$63,$C2,$74,$70,$F5,$31,$1C,$09,$3F,$D0,$78,$BC
	.dc.b	$B3,$CA,$FF,$A5,$46,$23,$94,$62,$7E,$24,$61,$FF,$23,$F3,$FF,$03
	.dc.b	$9F,$8B,$7D,$3A,$7F,$A8,$E1,$CF,$1F,$91,$18,$C3,$FC,$1D,$F0,$3F
	.dc.b	$E0,$DF,$81,$FF,$07,$36,$32,$7F,$84,$E3,$2D,$33,$9F,$F8,$4E,$32
	.dc.b	$D8,$CE,$5F,$E0,$89,$CB,$63,$27,$F3,$98,$CB,$4C,$CF,$F3,$A8,$CB
	.dc.b	$63,$FC,$4D,$EF,$FF,$89,$4F,$F1,$37,$BF,$63,$24,$9F,$F1,$44,$65
	.dc.b	$FF,$15,$4D,$BF,$F0,$EC,$DA,$FE,$99,$36,$BF,$A5,$CD,$E2,$FE,$CB
	.dc.b	$37,$89,$36,$30,$F1,$AF,$D0,$63,$F4,$CF,$E7,$7E,$D5,$C7,$EE,$EA
	.dc.b	$9C,$57,$F4,$5A,$B9,$FE,$8A,$6D,$FF,$AB,$BA,$BF,$A7,$D6,$2F,$F6
	.dc.b	$35,$F9,$82,$57,$E8,$1B,$1F,$99,$6B,$FA,$26,$BF,$C0,$9A,$F4,$8E
	.dc.b	$47,$F8,$6F,$E4,$7C,$1B,$FF,$88,$1B,$C4,$69,$27,$DF,$7F,$97,$9D
	.dc.b	$47,$F9,$87,$DD,$FF,$30,$1F,$E6,$1F,$76,$C6,$49,$8F,$F9,$8E,$32
	.dc.b	$FF,$99,$66,$E9,$FE,$58,$9B,$5F,$E0,$79,$B5,$FE,$06,$9B,$C3,$FE
	.dc.b	$20,$9B,$C2,$9B,$18,$73,$E8,$7F,$9D,$7C,$FF,$F3,$B3,$71,$FF,$3D
	.dc.b	$37,$1F,$F3,$C3,$72,$FF,$35,$B6,$7F,$C3,$4D,$9F,$F0,$CB,$72,$FF
	.dc.b	$19,$B7,$23,$49,$5F,$E8,$9C,$3C,$6D,$3F,$1B,$DB,$FC,$0C,$FA,$1E
	.dc.b	$AD,$8C,$93,$1F,$E7,$5C,$1F,$E5,$A9,$D4,$7F,$2E,$E3,$FE,$EF,$A7
	.dc.b	$F3,$4C,$39,$E1,$C6,$30,$FF,$2E,$8F,$E5,$24,$70,$4A,$FE,$55,$19
	.dc.b	$6C,$65,$FE,$6B,$19,$6C,$7F,$47,$1F,$8E,$BD,$CF,$3D,$FF,$54,$9C
	.dc.b	$C6,$15,$7F,$5A,$AE,$11,$E9,$32,$70,$31,$1B,$9E,$4D,$8F,$F5,$C5
	.dc.b	$FF,$AD,$FC,$DF,$F5,$C3,$63,$27,$FA,$F6,$32,$D3,$33,$FE,$BD,$8C
	.dc.b	$B6,$33,$87,$FA,$CA,$72,$D8,$C9,$FE,$63,$8C,$B4,$CC,$FF,$99,$23
	.dc.b	$2D,$8F,$E9,$BE,$33,$B7,$FD,$FE,$D8,$FF,$7F,$B6,$7F,$B5,$B5,$F9
	.dc.b	$67,$CC,$70,$E1,$AE,$67,$A9,$27,$AB,$63,$AB,$FE,$08,$7F,$CF,$DE
	.dc.b	$57,$F6,$8E,$2F,$F2,$2F,$91,$FE,$E5,$8C,$65,$FE,$A8,$C3,$67,$F8
	.dc.b	$2E,$B6,$FF,$BB,$67,$75,$C6,$35,$5C,$6F,$FD,$13,$5A,$9E,$06,$AB
	.dc.b	$84,$FF,$87,$EF,$AB,$FE,$F5,$AE,$0D,$AE,$07,$27,$85,$F7,$AF,$F5
	.dc.b	$7D,$70,$8F,$F1,$3E,$2D,$FE,$99,$30,$FA,$6D,$C3,$E9,$B5,$FA,$67
	.dc.b	$FA,$E5,$C3,$F4,$DA,$FD,$77,$EF,$F4,$C9,$FB,$ED,$9F,$BE,$D7,$AD
	.dc.b	$23,$11,$F7,$DA,$FE,$0D,$AF,$FB,$2F,$E4,$7F,$1A,$4A,$FC,$6B,$63
	.dc.b	$F1,$AD,$7B,$B5,$FC,$63,$5F,$C8,$3F,$98,$F9,$19,$3F,$98,$6C,$FE
	.dc.b	$61,$AF,$A9,$AF,$E5,$DA,$FE,$6D,$AF,$FB,$8F,$E4,$7F,$4A,$49,$FD
	.dc.b	$2B,$63,$F4,$AD,$7E,$3B,$5F,$D2,$35,$FD,$43,$FB,$0F,$91,$93,$FB
	.dc.b	$06,$CF,$EC,$1A,$FD,$26,$BF,$AF,$6B,$FB,$36,$BF,$EF,$BF,$91,$FD
	.dc.b	$E9,$27,$F7,$AD,$8F,$DE,$B5,$FB,$ED,$7F,$78,$D7,$F8,$03,$FC,$23
	.dc.b	$E4,$64,$FF,$08,$6C,$FF,$08,$6B,$F8,$A6,$BF,$C1,$DA,$FF,$0B,$6A
	.dc.b	$CD,$EF,$7F,$16,$25,$7F,$16,$6C,$FF,$16,$6B,$F9,$86,$BF,$C5,$5A
	.dc.b	$FF,$19,$7F,$90,$FC,$8C,$9F,$E4,$2D,$9F,$E4,$2D,$7F,$46,$D7,$F9
	.dc.b	$03,$5F,$E4,$8D,$7D,$5F,$23,$FC,$B8,$93,$FC,$B9,$B1,$FC,$B9,$AF
	.dc.b	$EC,$1A,$FF,$2D,$6B,$FC,$C5,$FE,$73,$F2,$32,$7F,$9C,$B6,$7F,$9C
	.dc.b	$B5,$FD,$DB,$5F,$E7,$0D,$7F,$9E,$35,$FA,$5F,$23,$FD,$18,$93,$FD
	.dc.b	$19,$B1,$FD,$19,$AF,$F0,$86,$BF,$D1,$5A,$FF,$49,$7F,$A8,$FC,$84
	.dc.b	$9F,$EA,$2D,$9F,$EA,$2D,$7F,$89,$B5,$FE,$A0,$D7,$FA,$A3,$5F,$C5
	.dc.b	$7C,$8F,$F5,$E2,$4F,$F5,$E6,$C7,$F5,$E6,$BF,$C8,$1A,$FF,$5D,$6B
	.dc.b	$FD,$85,$FE,$D3,$F0,$12,$7F,$B4,$B6,$7F,$B4,$B5,$FE,$56,$D7,$FB
	.dc.b	$43,$5F,$ED,$8D,$7F,$45,$F2,$1F,$DD,$89,$3F,$DD,$9B,$1F,$DD,$9A
	.dc.b	$FF,$38,$6B,$FD,$D5,$AF,$F7,$97,$FB,$83,$FE,$26,$FC,$43,$FC,$46
	.dc.b	$FF,$BA,$9A,$FF,$4E,$6B,$FE,$02,$6B,$C4,$E6,$1C,$B4,$CB,$82,$38
	.dc.b	$24,$FF,$BE,$9E,$5E,$0F,$02,$46,$65,$18,$9F,$43,$5F,$43,$5F,$43
	.dc.b	$5F,$80,$F9,$3E,$46,$57,$56,$CF,$C3,$6B,$F0,$DA,$FC,$36,$BF,$2D
	.dc.b	$AF,$C3,$F9,$1F,$82,$49,$F5,$36,$3E,$83,$5F,$A0,$D7,$E8,$35,$FB
	.dc.b	$0F,$CC,$F9,$19,$3F,$15,$B3,$F6,$DA,$FD,$B6,$BF,$6D,$AF,$E0,$DA
	.dc.b	$FD,$BF,$82,$FB,$24,$9F,$A4,$D8,$FC,$43,$5F,$C4,$35,$FC,$43,$5F
	.dc.b	$C8,$3F,$84,$F9,$19,$3F,$75,$B3,$F9,$36,$BF,$93,$6B,$F9,$36,$BF
	.dc.b	$9B,$6B,$F9,$3F,$91,$FC,$89,$27,$F1,$4D,$8F,$D0,$35,$FD,$03,$5F
	.dc.b	$D0,$35,$FD,$43,$F9,$CF,$91,$93,$F9,$56,$CF,$EA,$DA,$FE,$AD,$AF
	.dc.b	$EA,$DA,$FE,$CD,$AF,$EA,$FE,$47,$F5,$24,$9F,$D1,$36,$3F,$70,$D7
	.dc.b	$F7,$0D,$7F,$70,$D7,$F8,$03,$FB,$4F,$91,$93,$FA,$D6,$CF,$F0,$36
	.dc.b	$BF,$C0,$DA,$FF,$03,$6B,$FC,$2D,$AF,$F0,$3F,$23,$F8,$09,$27,$F7
	.dc.b	$4D,$8F,$E2,$0D,$7F,$88,$35,$FE,$20,$D7,$F8,$C3,$FC,$33,$E0,$A4
	.dc.b	$FF,$05,$6C,$FF,$1B,$6B,$FC,$6D,$AF,$F1,$B6,$BF,$C8,$DA,$FF,$1B
	.dc.b	$F8,$DF,$E3,$24,$9F,$E2,$4D,$8F,$E5,$0D,$7F,$94,$35,$FE,$50,$D7
	.dc.b	$F9,$83,$FC,$93,$E3,$64,$FF,$1D,$6C,$FF,$33,$6B,$FC,$CD,$AF,$F3
	.dc.b	$36,$BF,$CE,$DA,$FF,$33,$F8,$DF,$E6,$24,$9F,$E5,$4D,$8F,$E8,$0D
	.dc.b	$7F,$A0,$35,$FE,$80,$D7,$FA,$43,$FC,$F3,$E3,$64,$FF,$35,$6C,$FF
	.dc.b	$4B,$6B,$FD,$2D,$AF,$F4,$B6,$BF,$D4,$DA,$FF,$4B,$F8,$DF,$E9,$24
	.dc.b	$9F,$E8,$4D,$8F,$EB,$0D,$7F,$AC,$35,$FE,$B0,$D7,$FB,$03,$FD,$53
	.dc.b	$E3,$64,$FF,$4D,$6C,$FF,$63,$6B,$FD,$8D,$AF,$F6,$36,$BF,$DA,$DA
	.dc.b	$FF,$63,$F8,$5F,$EC,$24,$9F,$EB,$4D,$8F,$EE,$0D,$7F,$B8,$35,$FE
	.dc.b	$E0,$D7,$FB,$C3,$FD,$B3,$E3,$24,$FF,$65,$6C,$FF,$7B,$6B,$FD,$ED
	.dc.b	$AF,$F7,$B6,$BF,$E0,$36,$BF,$DE,$FE,$33,$FB,$C9,$27,$FB,$93,$63
	.dc.b	$FC,$10,$D7,$FC,$10,$D7,$FC,$10,$D7,$FC,$20,$FF,$82,$3F,$10,$4D
	.dc.b	$21,$FE,$03,$72,$D3,$33,$FE,$E2,$11,$A7,$0D,$7F,$DC,$ED,$44,$70
	.dc.b	$D8,$EA,$F0,$F8,$DF,$31,$23,$C8,$D8,$EE,$D7,$84,$68,$8E,$9A,$F9
	.dc.b	$3C,$5F,$84,$48,$F8,$0D,$8F,$5B,$5E,$11,$A2,$3A,$6B,$F0,$3E,$33
	.dc.b	$E7,$12,$3E,$63,$63,$E3,$B5,$E1,$1A,$23,$A6,$BF,$33,$E1,$7E,$D1
	.dc.b	$23,$EC,$36,$3E,$9B,$5E,$11,$A2,$3A,$6B,$F6,$3E,$33,$F0,$C4,$8F
	.dc.b	$C2,$36,$3E,$FB,$5E,$11,$A2,$3A,$6B,$F8,$4F,$8A,$FC,$91,$23,$F2
	.dc.b	$0D,$8F,$C5,$B5,$E1,$1A,$23,$A6,$BF,$90,$F8,$AF,$CF,$12,$3F,$38
	.dc.b	$D8,$FC,$BB,$5E,$11,$A2,$3A,$6B,$F9,$CF,$89,$FD,$51,$23,$F5,$0D
	.dc.b	$8F,$D1,$B5,$E1,$1A,$23,$A6,$BF,$A8,$F8,$AF,$DB,$12,$3F,$68,$D8
	.dc.b	$FD,$7B,$5E,$11,$A2,$3A,$6B,$FB,$4F,$89,$FE,$04,$48,$FE,$00,$D8
	.dc.b	$FD,$DB,$5E,$11,$A2,$3A,$6B,$FC,$03,$E1,$7F,$87,$12,$3F,$86,$36
	.dc.b	$3F,$83,$B5,$E1,$1A,$23,$A6,$BF,$C3,$3E,$27,$F8,$D1,$23,$F8,$C3
	.dc.b	$63,$F8,$9B,$5E,$11,$A2,$3A,$6B,$FC,$63,$E2,$7F,$93,$12,$3F,$92
	.dc.b	$36,$3F,$8F,$B5,$E1,$1A,$23,$A6,$BF,$C9,$3E,$1B,$F9,$91,$23,$F9
	.dc.b	$83,$63,$F9,$5B,$5E,$11,$A2,$3A,$6B,$FC,$C3,$E1,$BF,$9F,$12,$3F
	.dc.b	$9E,$36,$3F,$9B,$B5,$E1,$1A,$23,$A6,$BF,$CF,$3E,$17,$FA,$51,$23
	.dc.b	$FA,$43,$63,$FA,$1B,$5E,$11,$A2,$3A,$6B,$FD,$23,$E1,$7F,$AB,$12
	.dc.b	$3F,$AA,$36,$3F,$A7,$B5,$E1,$1A,$23,$A6,$BF,$D5,$3E,$0F,$FB,$11
	.dc.b	$23,$FB,$03,$63,$FA,$DB,$5E,$11,$A2,$3A,$6B,$FD,$83,$E0,$FF,$B7
	.dc.b	$12,$3F,$B6,$36,$3F,$B3,$B5,$E1,$1A,$23,$A6,$BF,$DB,$3E,$03,$FB
	.dc.b	$D1,$23,$FB,$C3,$63,$FB,$9B,$5E,$11,$A2,$3A,$6B,$FD,$D8,$FE,$EF
	.dc.b	$F8,$83,$E0,$FF,$C0,$F1,$FE,$01,$8C,$4F,$F8,$0E,$11,$87,$53,$C1
	.dc.b	$E8,$69,$C3,$DE,$38,$3A,$6B,$98,$DB,$E4,$6B,$CE,$7C,$53,$C2,$35
	.dc.b	$18,$F8,$4D,$C9,$A4,$8D,$12,$7D,$D1,$96,$C6,$5F,$34,$65,$F4,$7C
	.dc.b	$5F,$62,$4F,$91,$B1,$F2,$5A,$F0,$8D,$11,$D3,$5F,$89,$F1,$5E,$A2
	.dc.b	$47,$C0,$6C,$7D,$56,$BC,$23,$44,$74,$D7,$E8,$7C,$27,$C6,$24,$7C
	.dc.b	$C6,$C7,$E0,$5A,$F0,$8D,$11,$D3,$5F,$B9,$F0,$9F,$48,$91,$F6,$1B
	.dc.b	$1F,$8D,$6B,$C2,$34,$47,$4D,$74,$3E,$C9,$7F,$8D,$F8,$1F,$CC,$C6
	.dc.b	$27,$F3,$B7,$8A,$D5,$FF,$95,$AB,$55,$F9,$CA,$D3,$FA,$4B,$AE,$6B
	.dc.b	$85,$F2,$8F,$13,$F9,$DA,$DE,$6B,$B1,$86,$DF,$2B,$EC,$D7,$B1,$D0
	.dc.b	$C9,$4C,$E1,$E8,$BE,$D3,$C5,$5F,$78,$C3,$6B,$94,$EA,$78,$5E,$6F
	.dc.b	$A3,$E6,$AD,$B9,$C3,$E0,$4F,$87,$0C,$57,$9A,$38,$DF,$80,$EA,$48
	.dc.b	$E6,$E5,$A4,$8E,$2F,$9A,$7A,$BF,$C4,$2E,$E1,$E9,$8E,$18,$7D,$3C
	.dc.b	$3F,$86,$57,$F1,$A6,$9F,$E9,$42,$CF,$F1,$4A,$F9,$2D,$9E,$0F,$32
	.dc.b	$1F,$4A,$78,$91,$C9,$FC,$93,$E0,$78,$BF,$CB,$1A,$E6,$3C,$4F,$F2
	.dc.b	$38,$CD,$1A,$49,$DC,$F3,$6B,$CC,$F5,$46,$DB,$1A,$97,$C0,$F6,$24
	.dc.b	$65,$B3,$F0,$DB,$1F,$CD,$5B,$1F,$CE,$5A,$FF,$48,$7D,$3F,$03,$27
	.dc.b	$C0,$D9,$F9,$CD,$8F,$E8,$0D,$8F,$E8,$8D,$7F,$A9,$B5,$FE,$9F,$F0
	.dc.b	$3F,$18,$93,$ED,$6C,$7D,$B6,$C7,$F4,$D6,$C7,$F5,$16,$BF,$D8,$1F
	.dc.b	$A3,$F0,$32,$7E,$43,$67,$F0,$CD,$8F,$EB,$0D,$8F,$EB,$8D,$7F,$B5
	.dc.b	$B5,$FE,$CF,$F0,$3F,$78,$93,$F5,$5B,$1F,$93,$6C,$7F,$65,$6C,$7F
	.dc.b	$69,$6B,$FD,$E1,$E3,$B7,$F1,$AF,$C4,$1E,$2B,$DD,$FE,$09,$5F,$C6
	.dc.b	$A7,$31,$8C,$3F,$88,$46,$30,$FE,$0B,$5B,$BE,$51,$8A,$DC,$63,$0F
	.dc.b	$E2,$55,$BC,$3A,$5F,$F1,$7C,$3A,$56,$B0,$E3,$5A,$8F,$46,$1C,$F8
	.dc.b	$3D,$D5,$AC,$3C,$07,$39,$D3,$F0,$63,$55,$CA,$ED,$F2,$AE,$11,$84
	.dc.b	$30,$F2,$CE,$EB,$94,$F0,$23,$E9,$6C,$7F,$8A,$67,$8C,$74,$6B,$F8
	.dc.b	$0B,$F0,$3D,$F0,$F4,$B7,$0F,$85,$7C,$E7,$37,$C8,$E7,$B3,$2F,$9D
	.dc.b	$87,$88,$F0,$57,$49,$D6,$3E,$0B,$E5,$39,$BE,$39,$7B,$89,$8F,$C1
	.dc.b	$3C,$59,$71,$AE,$18,$58,$F5,$E3,$DB,$2F,$11,$EB,$2B,$9C,$3A,$C7
	.dc.b	$9A,$78,$E3,$E2,$C7,$BC,$F0,$8C,$3B,$31,$87,$85,$A7,$83,$2F,$6D
	.dc.b	$FC,$A8,$E1,$87,$81,$F2,$3F,$64,$C4,$F0,$8F,$C6,$C6,$32,$FA,$27
	.dc.b	$C0,$8E,$CE,$2B,$B4,$74,$C7,$D3,$5D,$F2,$F8,$11,$82,$4F,$91,$F9
	.dc.b	$CD,$21,$CC,$FC,$E1,$E0,$AF,$8F,$8F,$C1,$BE,$D9,$7C,$6A,$E0,$4C
	.dc.b	$3C,$2F,$4C,$BD,$D5,$F0,$71,$F1,$CF,$09,$FC,$6D,$F3,$77,$3E,$69
	.dc.b	$D6,$5F,$05,$F8,$F1,$CA,$77,$1E,$8B,$F9,$59,$7A,$23,$D4,$7D,$88
	.dc.b	$CB,$6B,$95,$F5,$9F,$34,$FA,$63,$13,$E3,$BE,$F3,$EF,$C3,$D1,$5A
	.dc.b	$8F,$15,$74,$C3,$C5,$87,$80,$F9,$39,$7C,$B9,$F8,$79,$7C,$8B,$F7
	.dc.b	$DF,$09,$E5,$1E,$73,$C9,$19,$69,$93,$89,$72,$F3,$9F,$3A,$FC,$11
	.dc.b	$F8,$0B,$F0,$4E,$B1,$FC,$86,$3E,$7C,$7C,$B9,$7B,$B2,$F0,$63,$DA
	.dc.b	$BD,$6F,$CE,$69,$C5,$ED,$3C,$5F,$25,$FC,$77,$E5,$65,$E4,$69,$CF
	.dc.b	$0F,$0C,$67,$1F,$54,$E4,$EB,$19,$AF,$4D,$7C,$8C,$3C,$6D,$39,$D7
	.dc.b	$13,$BC,$79,$DB,$1B,$7A,$9B,$BF,$0C,$ED,$B3,$B9,$E2,$7D,$49,$F0
	.dc.b	$B6,$3C,$31,$C6,$BC,$F3,$93,$BC,$67,$1F,$DF,$DF,$D5,$6D,$7C,$98
	.dc.b	$F0,$D7,$13,$C2,$7D,$5B,$EE,$79,$63,$71,$93,$BC,$7E,$39,$FF,$74
	.dc.b	$C7,$F7,$36,$CF,$F7,$26,$D7,$FB,$CE,$75,$1F,$CA,$8C,$9F,$EF,$23
	.dc.b	$F9,$A1,$82,$1B,$21,$FD,$67,$C6,$73,$75,$6C,$73,$94,$62,$3C,$ED
	.dc.b	$7F,$B2,$35,$E9,$3C,$C8,$F3,$8D,$3F,$DC,$DB,$3D,$A7,$73,$D9,$FF
	.dc.b	$0B,$CF,$93,$1F,$F1,$03,$71,$FE,$FB,$4F,$C4,$1C,$18,$BD,$E1,$C8
	.dc.b	$95,$FE,$AD,$AD,$B8,$9D,$BF,$E1,$BC,$3B,$3F,$E0,$A9,$DB,$9A,$E1
	.dc.b	$87,$C2,$7F,$D7,$93,$AB,$E6,$49,$D5,$78,$70,$F3,$46,$27,$53,$F2
	.dc.b	$8D,$1D,$8C,$3A,$7E,$04,$F2,$73,$3C,$C8,$FA,$63,$4F,$D1,$BE,$8D
	.dc.b	$3B,$9E,$D2,$57,$7B,$F6,$5F,$49,$E1,$39,$BE,$B8,$FD,$B9,$E3,$5B
	.dc.b	$BE,$31,$E6,$21,$CA,$3E,$DC,$6A,$30,$6A,$7A,$5F,$FB,$EA,$75,$18
	.dc.b	$AF,$23,$6B,$E0,$1E,$E7,$0F,$43,$E4,$99,$C7,$EF,$B9,$AE,$B1,$89
	.dc.b	$E2,$72,$8C,$E3,$C4,$FC,$0E,$3F,$0A,$B8,$12,$33,$7B,$3E,$29,$E6
	.dc.b	$29,$E5,$BE,$D5,$CE,$78,$63,$F0,$63,$15,$C2,$B6,$EB,$1F,$A1,$3A
	.dc.b	$AF,$C4,$56,$AB,$86,$1E,$57,$F0,$33,$A8,$C3,$DE,$3A,$1F,$63,$1F
	.dc.b	$82,$F2,$C7,$DA,$D3,$D7,$1A,$9E,$93,$A6,$CE,$B0,$F3,$63,$F6,$B0
	.dc.b	$FA,$78,$71,$8C,$61,$C5,$B7,$C4,$F2,$9F,$66,$F9,$BF,$89,$BE,$71
	.dc.b	$88,$F9,$D1,$87,$8E,$3F,$17,$1F,$29,$F9,$D8,$CC,$ED,$F7,$46,$5B
	.dc.b	$7F,$41,$DE,$1E,$77,$C4,$FE,$C3,$1F,$64,$60,$FA,$33,$EC,$AE,$87
	.dc.b	$89,$E9,$1D,$8F,$A5,$8F,$AF,$1F,$CA,$E3,$E5,$BE,$15,$A3,$C6,$76
	.dc.b	$3C,$0F,$6B,$F0,$1D,$9E,$31,$F2,$CE,$2F,$E0,$DF,$11,$92,$D7,$C9
	.dc.b	$9E,$71,$E1,$33,$5E,$58,$CC,$71,$AE,$F5,$C6,$3C,$B3,$E3,$37,$19
	.dc.b	$8F,$B3,$8F,$CB,$C3,$E1,$61,$F9,$8B,$FE,$17,$18,$9D,$57,$AA,$72
	.dc.b	$D3,$F6,$58,$78,$AF,$CD,$1E,$23,$D5,$19,$AF,$B9,$1E,$23,$F2,$47
	.dc.b	$B9,$CE,$1F,$51,$F0,$C6,$4F,$A4,$71,$9F,$A4,$F7,$3E,$25,$FC,$0B
	.dc.b	$F1,$C6,$E3,$E2,$3E,$87,$E5,$E3,$FC,$A4,$FC,$6C,$72,$34,$75,$75
	.dc.b	$8F,$C5,$8F,$1C,$70,$C7,$E7,$36,$FC,$2F,$F6,$2D,$3F,$B8,$D7,$F2
	.dc.b	$A0,$E2,$35,$1F,$38,$C9,$63,$57,$FD,$E9,$B1,$FE,$F8,$6B,$C0,$FE
	.dc.b	$4C,$E0,$8F,$33,$FA,$4F,$07,$36,$9C,$CA,$73,$3F,$C1,$0E,$DF,$EF
	.dc.b	$CD,$72,$71,$32,$53,$27,$3A,$47,$16,$C6,$DA,$EE,$35,$C1,$23,$4D
	.dc.b	$7A,$36,$BE,$63,$63,$FB,$3B,$5F,$F0,$B3,$5F,$F1,$26,$FE,$32,$3E
	.dc.b	$36,$BE,$96,$CF,$D9,$6C,$7F,$74,$6B,$FE,$23,$7D,$DF,$07,$23,$DC
	.dc.b	$D7,$E2,$36,$7F,$08,$D8,$FE,$FE,$D7,$FC,$5C,$D7,$FC,$79,$F0,$7F
	.dc.b	$2C,$8F,$CB,$6B,$F4,$5B,$3F,$91,$6C,$7F,$83,$1A,$FF,$8E,$DF,$AD
	.dc.b	$F0,$72,$3E,$B3,$5F,$B8,$D9,$FC,$E3,$63,$FC,$2E,$D7,$FC,$8C,$D7
	.dc.b	$FC,$A8,$FF,$8C,$32,$34,$FF,$19,$FE,$20,$F8,$1C,$57,$F9,$2A,$6A
	.dc.b	$E1,$70,$FF,$98,$2F,$84,$EC,$87,$F9,$76,$BA,$DF,$2B,$FF,$14,$45
	.dc.b	$BE,$53,$6B,$D2,$72,$AE,$E8,$5F,$81,$EC,$6E,$BE,$1B,$B9,$E2,$F0
	.dc.b	$AF,$1B,$B3,$B9,$ED,$9F,$1B,$C5,$AF,$62,$4F,$4A,$F1,$36,$BD,$53
	.dc.b	$F4,$0E,$47,$A6,$77,$5D,$2B,$E1,$CF,$8C,$F3,$1C,$0A,$78,$23,$DA
	.dc.b	$F9,$CE,$95,$EC,$74,$7B,$27,$C8,$F3,$AF,$90,$FF,$28,$6B,$EE,$31
	.dc.b	$3F,$EB,$B7,$A9,$63,$6E,$0F,$F6,$03,$87,$CE,$72,$38,$47,$37,$27
	.dc.b	$E2,$7E,$07,$12,$69,$23,$C0,$D7,$BB,$5F,$53,$5E,$87,$03,$D1,$F0
	.dc.b	$3D,$88,$E0,$87,$A9,$CB,$4C,$9F,$1D,$CB,$4F,$89,$F0,$3E,$A2,$38
	.dc.b	$21,$F1,$9C,$B4,$C9,$F4,$DC,$B4,$FA,$1F,$03,$F1,$88,$E0,$87,$D2
	.dc.b	$72,$D3,$27,$DF,$72,$D3,$EE,$7C,$0F,$D2,$23,$82,$1F,$79,$CB,$4C
	.dc.b	$9F,$8B,$72,$D3,$F1,$1F,$03,$F7,$88,$E0,$87,$E2,$9C,$B4,$C9,$F9
	.dc.b	$77,$2D,$3F,$29,$F0,$3F,$8A,$23,$82,$1F,$96,$72,$D3,$27,$E8,$DC
	.dc.b	$B4,$FD,$07,$C0,$FE,$58,$8E,$08,$7E,$89,$CB,$4C,$9F,$AF,$72,$D3
	.dc.b	$F5,$9F,$03,$FA,$22,$38,$21,$FA,$E7,$2D,$32,$7E,$ED,$CB,$4F,$DC
	.dc.b	$7C,$0F,$EB,$88,$E0,$87,$EE,$9C,$B4,$C9,$FC,$1D,$CB,$4F,$E0,$9F
	.dc.b	$03,$FB,$A2,$38,$21,$FC,$19,$CB,$4C,$9F,$C4,$DC,$B4,$FE,$21,$F0
	.dc.b	$3F,$C1,$88,$E0,$87,$F1,$27,$2D,$32,$7F,$1F,$72,$D3,$F8,$E7,$C0
	.dc.b	$FF,$12,$23,$82,$1F,$C7,$9C,$B4,$C9,$FC,$AD,$CB,$4F,$E5,$1F,$03
	.dc.b	$FC,$78,$8E,$08,$7F,$2A,$72,$D3,$27,$F3,$77,$2D,$3F,$9A,$7C,$0F
	.dc.b	$F2,$A2,$38,$21,$FC,$D9,$CB,$4C,$9F,$D0,$DC,$B4,$FE,$81,$F0,$3F
	.dc.b	$CD,$88,$E0,$87,$F4,$27,$2D,$32,$7F,$4F,$72,$D3,$FA,$67,$C0,$FF
	.dc.b	$42,$23,$82,$1F,$D3,$9C,$B4,$C9,$FD,$6D,$CB,$4F,$EB,$1F,$03,$FD
	.dc.b	$38,$8E,$08,$7F,$5A,$72,$D3,$27,$F6,$77,$2D,$3F,$B3,$74,$7F,$9E
	.dc.b	$DF,$F0,$B1,$FE,$26,$EB,$FE,$C1,$19,$8F,$F0,$D3,$5D,$1C,$A3,$64
	.dc.b	$70,$43,$FD,$3E,$70,$2C,$73,$1F,$01,$E0,$24,$79,$9A,$F0,$6B,$FE
	.dc.b	$AA,$8E,$2E,$C8,$6C,$A6,$FE,$03,$DA,$47,$E1,$B5,$F2,$B5,$E0,$68
	.dc.b	$86,$8A,$78,$FC,$8F,$92,$47,$E7,$B5,$F8,$2D,$78,$1A,$21,$A2,$9E
	.dc.b	$FF,$17,$EA,$91,$FB,$6D,$7E,$6B,$5E,$06,$88,$68,$A7,$CB,$F8,$0F
	.dc.b	$C0,$91,$FC,$3B,$5F,$B2,$D7,$81,$A2,$1A,$29,$F5,$FE,$03,$F1,$A4
	.dc.b	$7F,$26,$D7,$F0,$AD,$78,$1A,$21,$A2,$9F,$83,$F8,$0F,$CC,$91,$FC
	.dc.b	$FB,$5F,$C8,$B5,$E0,$68,$86,$8A,$7E,$3F,$E0,$3F,$4A,$47,$F5,$6D
	.dc.b	$7F,$3A,$D7,$81,$A2,$1A,$29,$F9,$BF,$80,$FD,$89,$1F,$DB,$B5,$FD
	.dc.b	$4B,$5E,$06,$88,$68,$A7,$E9,$FE,$03,$F7,$A4,$7F,$81,$B5,$FD,$AB
	.dc.b	$5E,$06,$88,$68,$A7,$EC,$FE,$03,$F8,$49,$1F,$E1,$ED,$7F,$80,$B5
	.dc.b	$E0,$68,$86,$8A,$7E,$FF,$E0,$3F,$8A,$91,$FE,$36,$D7,$F8,$6B,$5E
	.dc.b	$06,$88,$68,$A7,$F0,$BF,$80,$FE,$42,$47,$F9,$3B,$5F,$E3,$2D,$78
	.dc.b	$1A,$21,$A2,$9F,$C5,$FE,$03,$F9,$69,$1F,$E6,$6D,$7F,$92,$B5,$E0
	.dc.b	$68,$86,$8A,$7F,$23,$F8,$0F,$E7,$24,$7F,$9F,$B5,$FE,$62,$D7,$81
	.dc.b	$A2,$1A,$29,$FC,$BF,$E0,$3F,$A2,$91,$FE,$96,$D7,$F9,$EB,$5E,$06
	.dc.b	$88,$68,$A7,$F3,$BF,$80,$FE,$A2,$47,$FA,$BB,$5F,$E9,$2D,$78,$1A
	.dc.b	$21,$A2,$9F,$D1,$FE,$03,$FA,$E9,$1F,$EC,$6D,$7F,$AA,$B5,$E0,$68
	.dc.b	$86,$8A,$7F,$53,$F8,$0F,$ED,$24,$7F,$B7,$B5,$FE,$C2,$D7,$81,$A2
	.dc.b	$1A,$29,$FD,$7F,$E0,$3F,$BA,$91,$FE,$F6,$D7,$FB,$6B,$5E,$06,$88
	.dc.b	$68,$A7,$F8,$08,$7E,$20,$F6,$3F,$DC,$4D,$8F,$F7,$73,$5F,$EF,$AD
	.dc.b	$79,$35,$FF,$06,$B5,$1D,$A3,$A9,$1E,$AD,$7A,$BE,$2F,$81,$91,$E2
	.dc.b	$6B,$E2,$6B,$E2,$6B,$E2,$6B,$E2,$94,$62,$3E,$0F,$C0,$FC,$02,$3F
	.dc.b	$01,$AF,$C0,$6B,$F0,$1A,$FC,$06,$BD,$DF,$97,$D5,$23,$E5,$B5,$F9
	.dc.b	$6D,$7E,$5B,$5F,$96,$D7,$D4,$D7,$EC,$FC,$0F,$D8,$23,$F6,$1A,$FD
	.dc.b	$86,$BF,$61,$AF,$D8,$6B,$F1,$DF,$C1,$FC,$0C,$8F,$C1,$B5,$FC,$1B
	.dc.b	$5F,$C1,$B5,$FC,$1B,$5F,$A4,$D7,$F2,$3F,$13,$F9,$02,$3F,$90,$6B
	.dc.b	$F9,$06,$BF,$90,$6B,$F9,$06,$BF,$7D,$FC,$DF,$C4,$C8,$FC,$DB,$5F
	.dc.b	$CD,$B5,$FC,$DB,$5F,$CD,$B5,$FC,$53,$5F,$D4,$FC,$4F,$EA,$08,$FE
	.dc.b	$A1,$AF,$EA,$1A,$FE,$A1,$AF,$EA,$1A,$FE,$5D,$FD,$9F,$C4,$C8,$FD
	.dc.b	$9B,$5F,$D9,$B5,$FD,$9B,$5F,$D9,$B5,$FD,$13,$5F,$E0,$3F,$0B,$FC
	.dc.b	$00,$8F,$F0,$06,$BF,$C0,$1A,$FF,$00,$6B,$FC,$01,$AF,$EB,$DF,$E1
	.dc.b	$7F,$0B,$23,$F8,$5B,$5F,$E1,$6D,$7F,$85,$B5,$FE,$16,$D7,$F7,$4D
	.dc.b	$7F,$8C,$FC,$2F,$F1,$82,$3F,$C6,$1A,$FF,$18,$6B,$FC,$61,$AF,$F1
	.dc.b	$86,$BF,$C1,$DF,$E4,$7F,$05,$23,$F9,$1B,$5F,$E4,$6D,$7F,$91,$B5
	.dc.b	$FE,$46,$D7,$F8,$93,$5F,$E6,$3F,$05,$FC,$C0,$8F,$F3,$06,$BF,$CC
	.dc.b	$1A,$FF,$30,$6B,$FC,$C1,$AF,$F1,$F7,$F9,$DF,$C1,$48,$FE,$76,$D7
	.dc.b	$F9,$DB,$5F,$E7,$6D,$7F,$9D,$B5,$FE,$54,$D7,$FA,$4F,$C1,$7F,$48
	.dc.b	$23,$FD,$21,$AF,$F4,$86,$BF,$D2,$1A,$FF,$48,$6B,$FC,$DD,$FE,$A7
	.dc.b	$F0,$52,$3F,$A9,$B5,$FE,$A6,$D7,$FA,$9B,$5F,$EA,$6D,$7F,$A1,$35
	.dc.b	$FE,$C3,$F0,$5F,$D8,$08,$FF,$60,$6B,$FD,$81,$AF,$F6,$06,$BF,$D8
	.dc.b	$1A,$FF,$4F,$7F,$B5,$FC,$0C,$8F,$ED,$6D,$7F,$B5,$B5,$FE,$D6,$D7
	.dc.b	$FB,$5B,$5F,$EB,$4D,$7F,$BC,$FC,$4F,$F7,$82,$3F,$DE,$1A,$FF,$78
	.dc.b	$6B,$FD,$E1,$AF,$F7,$86,$BF,$D9,$DF,$F7,$DB,$FD,$EB,$F1,$0A,$D2
	.dc.b	$B8,$4D,$23,$A2,$1D,$1C,$B4,$C9,$FE,$15,$39,$FC,$37,$62,$3C,$5A
	.dc.b	$EE,$11,$88,$FF,$0F,$46,$1E,$04,$38,$46,$DF,$F0,$B9,$EA,$F8,$9F
	.dc.b	$69,$1F,$43,$5F,$03,$5E,$ED,$79,$CA,$31,$1F,$2D,$F9,$3F,$09,$23
	.dc.b	$E2,$35,$F6,$35,$F5,$B5,$EA,$D7,$EB,$35,$FA,$7F,$07,$F5,$88,$FD
	.dc.b	$16,$BF,$25,$AF,$C8,$6B,$EA,$6B,$F8,$37,$F0,$7F,$92,$BD,$2F,$93
	.dc.b	$F1,$8F,$53,$E5,$CB,$D4,$CC,$61,$B8,$7F,$91,$A7,$93,$FE,$48,$AE
	.dc.b	$4E,$6F,$6A,$7A,$3B,$BE,$8D,$38,$14,$E1,$3C,$A7,$8C,$2B,$BB,$D6
	.dc.b	$F5,$52,$39,$57,$0A,$DB,$6F,$DF,$5B,$AE,$91,$E4,$69,$EF,$BD,$5F
	.dc.b	$2A,$F2,$44,$9E,$F1,$F0,$33,$7C,$1F,$EC,$09,$3F,$AB,$6B,$FA,$36
	.dc.b	$BF,$A2,$6B,$F9,$46,$BF,$BD,$7F,$7B,$F0,$4F,$67,$E7,$AF,$66,$98
	.dc.b	$89,$5F,$36,$74,$D9,$CB,$F0,$04,$75,$74,$43,$4E,$5B,$3E,$07,$59
	.dc.b	$7D,$69,$D3,$67,$53,$CA,$BC,$33,$98,$E3,$9F,$EF,$AF,$CF,$1D,$8F
	.dc.b	$C3,$62,$F8,$1F,$83,$EE,$49,$CC,$7D,$33,$0F,$A2,$30,$E8,$F8,$8E
	.dc.b	$B0,$FC,$6D,$75,$9D,$1F,$C6,$6B,$F9,$59,$1F,$E4,$CD,$74,$60,$86
	.dc.b	$08,$7F,$20,$78,$4F,$81,$F2,$C7,$CE,$F7,$AC,$E6,$3C,$6F,$CD,$8C
	.dc.b	$E3,$FC,$6A,$AD,$7F,$AF,$31,$57,$E1,$A7,$F0,$FF,$03,$2B,$F0,$91
	.dc.b	$97,$F0,$D1,$97,$EF,$C6,$5F,$C0,$C6,$70,$FB,$D3,$97,$F0,$11,$9C
	.dc.b	$3E,$D4,$E5,$FB,$91,$9A,$F3,$4E,$5F,$44,$65,$FC,$9C,$7E,$7B,$59
	.dc.b	$1F,$9B,$7F,$43,$F0,$78,$8F,$CA,$C5,$C3,$F2,$53,$70,$FC,$64,$DA
	.dc.b	$F8,$D3,$5F,$D0,$C5,$C9,$36,$30,$D9,$F1,$1F,$AD,$7F,$81,$7B,$FF
	.dc.b	$B4,$F8,$59,$3F,$98,$8C,$BF,$B3,$8C,$BF,$B1,$D5,$3A,$8F,$D7,$CE
	.dc.b	$B2,$FC,$AC,$E6,$7F,$B4,$BF,$DF,$88,$FF,$79,$6B,$FD,$C5,$AF,$03
	.dc.b	$44,$34,$58,$EE,$FA,$4F,$E1,$5C,$BE,$58,$CC,$7A,$67,$F8,$9C,$E7
	.dc.b	$3F,$ED,$B1,$67,$F2,$93,$FD,$1B,$F2,$57,$4D,$24,$FF,$17,$29,$FC
	.dc.b	$56,$36,$FF,$47,$6B,$FC,$50,$DE,$1F,$D1,$A3,$2D,$9F,$E4,$F3,$AC
	.dc.b	$3F,$C1,$75,$A6,$CE,$A7,$B3,$B4,$27,$83,$FA,$F3,$46,$62,$EB,$FE
	.dc.b	$18,$8F,$57,$6F,$C1,$ED,$91,$EB,$8F,$88,$61,$F8,$11,$87,$47,$C1
	.dc.b	$75,$8F,$BE,$75,$3C,$0D,$1C,$0F,$09,$83,$46,$09,$1F,$26,$7A,$E3
	.dc.b	$F1,$AB,$FB,$EF,$E2,$4F,$81,$FB,$5E,$0C,$C7,$FB,$D4,$E5,$B3,$97
	.dc.b	$ED,$4F,$28,$91,$F6,$E3,$58,$FF,$7E,$AE,$53,$FD,$55,$AF,$CB,$6B
	.dc.b	$FE,$44,$6B,$F5,$9A,$9A,$4A,$FF,$33,$B6,$3F,$CD,$6D,$7F,$CB,$CD
	.dc.b	$9F,$24,$FA,$27,$F2,$9E,$1D,$9C,$C7,$C0,$3F,$2D,$19,$C7,$CD,$56
	.dc.b	$7F,$17,$AA,$BF,$49,$E3,$C9,$FC,$F3,$FA,$7F,$1F,$11,$FA,$38,$B8
	.dc.b	$FE,$7E,$6E,$3F,$9C,$9B,$5D,$A6,$BF,$99,$8B,$82,$6C,$61,$B3,$E2
	.dc.b	$3F,$60,$FE,$E7,$CB,$FD,$D7,$97,$27,$F6,$91,$97,$F6,$D1,$97,$F7
	.dc.b	$51,$97,$F7,$91,$9C,$7F,$71,$39,$7F,$75,$19,$C7,$F6,$F3,$97,$F7
	.dc.b	$31,$9A,$F9,$73,$97,$E6,$C6,$5F,$DC,$47,$E6,$BD,$BF,$E2,$5E,$74
	.dc.b	$9F,$CC,$C6,$5F,$E1,$D1,$97,$F2,$51,$97,$F8,$A4,$67,$0F,$CC,$4E
	.dc.b	$67,$F2,$CF,$E3,$88,$FE,$39,$AF,$E3,$9A,$F0,$34,$43,$45,$8E,$EF
	.dc.b	$A0,$FE,$49,$E8,$F9,$63,$31,$E8,$32,$6C,$C1,$1F,$E5,$11,$FE,$7A
	.dc.b	$FC,$95,$EB,$5F,$5A,$5F,$F9,$DE,$72,$D9,$CB,$FE,$4B,$84,$F3,$99
	.dc.b	$3C,$63,$77,$FE,$7C,$AE,$73,$C6,$B3,$A7,$FB,$02,$79,$CF,$72,$1D
	.dc.b	$EF,$6F,$F3,$68,$E7,$7D,$6F,$32,$9E,$6F,$73,$D9,$ED,$7C,$12,$3E
	.dc.b	$FF,$6A,$BE,$D8,$B7,$E9,$9B,$8F,$9E,$6D,$78,$E6,$BE,$18,$B6,$9B
	.dc.b	$18,$6C,$F9,$0F,$11,$FC,$BB,$0F,$AD,$EE,$48,$C1,$0F,$A4,$E5,$FE
	.dc.b	$55,$1F,$CE,$8C,$63,$F4,$67,$13,$C4,$FE,$5D,$3F,$C9,$88,$FF,$26
	.dc.b	$6B,$FC,$99,$AF,$03,$44,$34,$59,$FE,$7D,$1E,$08,$FC,$47,$BB,$67
	.dc.b	$31,$EB,$7F,$0D,$19,$CB,$FA,$05,$5A,$FC,$04,$FE,$67,$E0,$7F,$35
	.dc.b	$F0,$32,$7F,$31,$19,$7F,$35,$19,$7F,$31,$19,$7F,$35,$19,$C3,$F2
	.dc.b	$D3,$97,$F3,$11,$9C,$3F,$2D,$39,$7F,$31,$19,$AF,$CA,$4E,$5F,$CB
	.dc.b	$46,$5F,$CA,$47,$EC,$7C,$A9,$1F,$AF,$7F,$69,$E5,$E2,$3F,$5D,$17
	.dc.b	$0F,$D5,$4D,$C3,$F5,$13,$6B,$F4,$93,$5F,$D1,$45,$DB,$FD,$CD,$1F
	.dc.b	$A9,$8C,$1E,$07,$C4,$43,$C2,$7F,$03,$9F,$F8,$03,$D2,$F6,$3F,$C4
	.dc.b	$F3,$FE,$CD,$FC,$95,$F0,$98,$CF,$FD,$F7,$18,$9F,$F1,$6C,$7F,$BF
	.dc.b	$AF,$94,$63,$3E,$71,$88,$E9,$7C,$E7,$BD,$F3,$9F,$F6,$E5,$6E,$FA
	.dc.b	$61,$B9,$FF,$1D,$D6,$E7,$C7,$87,$AA,$A5,$7F,$8E,$4F,$F2,$0D,$F8
	.dc.b	$5F,$F1,$E3,$5F,$2C,$6D,$DB,$5F,$F1,$43,$63,$84,$68,$E0,$E1,$D9
	.dc.b	$88,$E0,$47,$E2,$1C,$4F,$7B,$ED,$57,$FE,$F2,$C3,$CB,$1C,$0E,$A7
	.dc.b	$03,$11,$B7,$A6,$1F,$21,$E1,$7F,$2E,$BE,$7F,$9D,$DE,$FF,$D9,$F5
	.dc.b	$CB,$E1,$70,$67,$CD,$91,$92,$1F,$85,$72,$FE,$1E,$38,$BE,$48,$CE
	.dc.b	$3F,$88,$9C,$CF,$13,$47,$48,$D3,$83,$47,$53,$58,$7E,$22,$BD,$5E
	.dc.b	$64,$9F,$86,$FB,$7C,$DC,$47,$E6,$E2,$DF,$E6,$A6,$DF,$E4,$A6,$D7
	.dc.b	$E3,$A6,$BF,$52,$2E,$29,$B1,$86,$CF,$88,$FA,$AF,$EB,$BD,$9F,$A1
	.dc.b	$E3,$C9,$FA,$B1,$97,$ED,$C6,$5F,$D1,$C6,$5F,$DA,$46,$71,$FD,$84
	.dc.b	$E6,$7F,$3A,$FE,$5C,$8F,$E5,$9A,$FE,$97,$54,$62,$38,$9B,$21,$B7
	.dc.b	$8C,$F8,$23,$D2,$7E,$4B,$CE,$F3,$46,$63,$D2,$64,$D9,$82,$3F,$C3
	.dc.b	$23,$F2,$9E,$57,$E5,$7D,$29,$5F,$9E,$8C,$BF,$A0,$8C,$BF,$C5,$E3
	.dc.b	$2F,$F1,$98,$CD,$7F,$18,$9C,$BF,$C6,$A3,$3A,$FF,$12,$9C,$BF,$C5
	.dc.b	$23,$35,$FC,$42,$72,$FF,$12,$8C,$BF,$B8,$8F,$DD,$7C,$1F,$FB,$3A
	.dc.b	$0D,$31,$B7,$FB,$37,$06,$1F,$CC,$27,$5E,$67,$23,$F6,$D7,$4F,$E0
	.dc.b	$51,$A7,$F8,$09,$91,$5D,$1F,$F7,$0C,$6D,$FE,$78,$6E,$34,$F9,$A7
	.dc.b	$15,$FC,$D2,$29,$CA,$2B,$FC,$58,$FE,$93,$E6,$7C,$0F,$5E,$47,$F1
	.dc.b	$08,$CB,$E8,$8C,$BF,$CF,$A3,$2F,$A2,$BE,$14,$EA,$FF,$A6,$56,$AB
	.dc.b	$F9,$CB,$FC,$CC,$8F,$F3,$26,$BF,$9D,$6B,$C0,$D1,$0D,$16,$3B,$BE
	.dc.b	$93,$E7,$F9,$BE,$68,$CC,$7A,$4C,$9B,$30,$47,$E6,$E2,$9F,$C2,$F9
	.dc.b	$72,$BE,$DB,$F8,$8F,$67,$11,$F0,$CF,$B3,$18,$BE,$73,$72,$FC,$44
	.dc.b	$DA,$FB,$F3,$5F,$C8,$C5,$C1,$36,$30,$D9,$F1,$9F,$8E,$7F,$37,$E5
	.dc.b	$FE,$73,$CB,$93,$F9,$28,$CB,$F9,$48,$CB,$E9,$3F,$43,$1A,$71,$7F
	.dc.b	$ED,$97,$81,$CC,$86,$A7,$9E,$3F,$2A,$B9,$3F,$32,$33,$8F,$EA,$67
	.dc.b	$2F,$EA,$E3,$35,$FA,$29,$FE,$DD,$F8,$A3,$83,$F6,$FE,$EC,$AF,$ED
	.dc.b	$D3,$96,$CE,$5F,$4C,$F2,$89,$1F,$B4,$8D,$63,$FA,$CA,$E5,$3F,$0D
	.dc.b	$AF,$F8,$B8,$38,$7F,$C6,$2D,$46,$AB,$FC,$29,$19,$8C,$1F,$E1,$97
	.dc.b	$04,$8D,$92,$34,$D7,$C8,$FA,$9F,$AB,$E6,$5A,$F5,$4D,$CB,$F8,$B4
	.dc.b	$DA,$F2,$DF,$BA,$F7,$1F,$C5,$A7,$F9,$1F,$95,$23,$F8,$CB,$FC,$9B
	.dc.b	$DE,$C4,$7E,$12,$2D,$7E,$A2,$6E,$1F,$C9,$66,$D7,$F1,$C9,$AF,$F2
	.dc.b	$E8,$BA,$A6,$C6,$1B,$3E,$23,$F9,$5B,$FE,$31,$FC,$97,$FD,$A3,$7F
	.dc.b	$E6,$27,$FD,$9F,$87,$F2,$66,$C6,$7C,$8E,$07,$1F,$32,$61,$FE,$37
	.dc.b	$AE,$6F,$DA,$6A,$79,$B9,$8E,$6F,$19,$E6,$FE,$AE,$76,$F1,$8D,$B9
	.dc.b	$9D,$DF,$9E,$72,$FA,$63,$37,$E9,$AF,$CF,$FC,$57,$C1,$F8,$59,$18
	.dc.b	$21,$E1,$72,$FE,$6E,$3F,$56,$62,$FE,$34,$E2,$78,$9F,$9F,$D9,$5F
	.dc.b	$9A,$24,$7E,$65,$AF,$F9,$E6,$36,$E4,$86,$63,$6F,$57,$8D,$7E,$C5
	.dc.b	$F1,$47,$C9,$F3,$AC,$E6,$3D,$AF,$DD,$8C,$E1,$FB,$2A,$B5,$F6,$67
	.dc.b	$F0,$DE,$64,$8F,$BC,$FE,$27,$CF,$C4,$7D,$78,$B8,$7E,$26,$6E,$1F
	.dc.b	$8D,$9B,$5F,$85,$9A,$FE,$16,$2E,$A9,$B1,$86,$CF,$88,$FC,$83,$F9
	.dc.b	$BF,$33,$F3,$9E,$6C,$9F,$CA,$46,$5F,$CB,$46,$5F,$C7,$46,$5F,$C8
	.dc.b	$46,$71,$FD,$14,$E5,$FD,$24,$67,$1F,$D4,$4E,$5F,$D5,$46,$6B,$F4
	.dc.b	$53,$97,$F4,$91,$97,$F4,$91,$FC,$DB,$D9,$FD,$A7,$87,$27,$F9,$AC
	.dc.b	$65,$FD,$74,$65,$FE,$5D,$19,$7F,$51,$19,$C3,$F9,$A4,$E6,$7F,$28
	.dc.b	$FF,$2C,$23,$FC,$AD,$AF,$E4,$DA,$F0,$34,$43,$45,$8E,$EF,$A0,$FE
	.dc.b	$09,$E6,$F9,$63,$31,$E8,$32,$6C,$C1,$1F,$E2,$51,$FE,$C7,$FC,$97
	.dc.b	$FD,$81,$26,$93,$DB,$D7,$99,$27,$FD,$95,$39,$6C,$65,$FE,$F1,$19
	.dc.b	$6C,$66,$FF,$C8,$F3,$96,$C6,$74,$FE,$AF,$5E,$38,$F3,$5E,$63,$FA
	.dc.b	$DD,$F9,$A7,$31,$E4,$F2,$FD,$9E,$64,$8C,$10,$F2,$B9,$7C,$F1,$F0
	.dc.b	$CC,$63,$E6,$9C,$4F,$18,$C3,$F0,$A3,$2F,$23,$04,$30,$48,$E2,$F8
	.dc.b	$8E,$98,$FC,$0F,$C5,$9E,$87,$D5,$F8,$C9,$5F,$EF,$99,$CB,$67,$2F
	.dc.b	$D0,$9F,$93,$7F,$E7,$4A,$DC,$FF,$8A,$DB,$1F,$4A,$78,$BB,$AE,$37
	.dc.b	$E3,$7F,$C1,$D7,$4A,$2F,$AB,$FE,$13,$6B,$B3,$8B,$87,$FC,$26,$D7
	.dc.b	$94,$74,$69,$DA,$34,$48,$D3,$EA,$9E,$D5,$E9,$AF,$CA,$F9,$FE,$A7
	.dc.b	$BD,$D7,$A9,$F0,$DF,$D1,$46,$6F,$A4,$F0,$3E,$71,$D0,$FD,$0F,$99
	.dc.b	$FA,$92,$3F,$A4,$F7,$6B,$FA,$28,$B8,$7E,$82,$6E,$3F,$9E,$D9,$58
	.dc.b	$BF,$CE,$D5,$7F,$69,$16,$53,$63,$0D,$9F,$29,$E3,$3F,$6F,$E1,$7E
	.dc.b	$FF,$CD,$91,$82,$1F,$B8,$72,$FE,$EA,$3F,$83,$18,$C7,$F7,$33,$89
	.dc.b	$E3,$18,$7F,$82,$C6,$5E,$46,$08,$60,$91,$C5,$F1,$1D,$31,$FD,$27
	.dc.b	$47,$F1,$7F,$80,$93,$FA,$38,$FB,$06,$1F,$E2,$31,$87,$47,$F1,$87
	.dc.b	$55,$FA,$59,$D4,$FE,$59,$FC,$69,$1F,$C9,$35,$D1,$82,$18,$21,$FA
	.dc.b	$37,$84,$77,$8F,$19,$FE,$9D,$F9,$BF,$F7,$F6,$AA,$F1,$D6,$EB,$5E
	.dc.b	$76,$31,$F8,$71,$CA,$7F,$DE,$38,$2B,$F3,$CF,$FB,$1E,$7F,$D0,$33
	.dc.b	$AF,$33,$26,$BC,$DB,$3A,$C3,$F9,$8D,$F1,$8F,$35,$6D,$D1,$FE,$8C
	.dc.b	$8D,$5F,$58,$C8,$8C,$D7,$4B,$EF,$3D,$0D,$CE,$9E,$91,$AA,$E9,$1B
	.dc.b	$3E,$44,$66,$BC,$F7,$67,$89,$D3,$CD,$F8,$7E,$6C,$8C,$10,$FA,$4E
	.dc.b	$5F,$A9,$1F,$28,$C6,$1F,$3A,$71,$3C,$63,$0F,$AA,$32,$F2,$30,$43
	.dc.b	$04,$8E,$2F,$88,$E9,$7F,$D0,$FE,$27,$ED,$FA,$32,$7F,$A1,$47,$F1
	.dc.b	$73,$0F,$E2,$E3,$0E,$8F,$9A,$EA,$BF,$A2,$4E,$A7,$F9,$AB,$FC,$B0
	.dc.b	$8F,$F3,$16,$BA,$30,$43,$04,$3F,$A1,$3C,$23,$BC,$78,$CF,$D0,$79
	.dc.b	$DE,$18,$CC,$7F,$55,$9F,$CC,$CE,$6F,$E1,$4F,$1A,$FC,$FC,$FE,$7B
	.dc.b	$CD,$93,$F8,$A7,$F4,$3E,$D6,$23,$F5,$D1,$6B,$F4,$F3,$6F,$F3,$F3
	.dc.b	$6B,$F3,$93,$5F,$CD,$45,$D9,$36,$30,$D9,$F1,$1F,$9C,$7F,$67,$E6
	.dc.b	$FE,$D3,$CD,$93,$FA,$08,$CB,$FA,$28,$CB,$FC,$16,$32,$FF,$07,$8C
	.dc.b	$E3,$FC,$02,$72,$FF,$02,$8C,$E3,$FB,$A9,$CB,$FB,$C8,$CD,$7E,$E6
	.dc.b	$72,$FE,$EE,$32,$FE,$E6,$3F,$B4,$7E,$2C,$F6,$FF,$B5,$68,$F8,$AA
	.dc.b	$6A,$25,$FF,$02,$9D,$36,$72,$B2,$65,$FE,$40,$C3,$B4,$EC,$86,$E3
	.dc.b	$58,$72,$8F,$F2,$38,$9E,$2F,$F8,$61,$AE,$9A,$F2,$3F,$C3,$6E,$11
	.dc.b	$CE,$3F,$C2,$91,$98,$C1,$C2,$32,$47,$07,$F2,$28,$F1,$D7,$BA,$3A
	.dc.b	$C6,$2B,$FC,$BF,$87,$FB,$F2,$31,$3A,$BF,$8F,$7C,$61,$5B,$9C,$DF
	.dc.b	$AE,$7E,$31,$EA,$97,$A1,$F6,$3C,$C9,$3F,$55,$FB,$7E,$E6,$23,$89
	.dc.b	$DA,$31,$8F,$86,$D5,$8C,$3D,$35,$6F,$C1,$35,$E4,$77,$3F,$08,$E6
	.dc.b	$93,$63,$0D,$9F,$C5,$74,$7E,$43,$CC,$91,$82,$1E,$8B,$FF,$6F,$3F
	.dc.b	$91,$8F,$CB,$CC,$8C,$10,$E0,$E5,$E5,$1D,$9C,$9D,$B1,$F8,$51,$97
	.dc.b	$E2,$4F,$17,$33,$D5,$CB,$F8,$3F,$53,$F5,$3F,$19,$23,$04,$3D,$4E
	.dc.b	$5F,$A7,$B2,$7F,$66,$66,$BE,$A5,$62,$B9,$1F,$9F,$9F,$CC,$91,$FC
	.dc.b	$FB,$5F,$D3,$35,$E0,$68,$86,$8B,$3F,$93,$AF,$24,$7A,$63,$F0,$DE
	.dc.b	$6F,$8E,$3D,$35,$67,$31,$BD,$7F,$80,$4E,$5F,$D1,$D7,$F0,$EF,$2F
	.dc.b	$F8,$7F,$8D,$2F,$F8,$64,$65,$FE,$1D,$19,$7F,$05,$19,$7F,$09,$19
	.dc.b	$C3,$F3,$53,$97,$F3,$91,$9B,$FD,$9C,$E5,$FD,$AC,$66,$BF,$59,$39
	.dc.b	$7F,$5D,$19,$7F,$6B,$1F,$E8,$FF,$C9,$7F,$DD,$A4,$F2,$F1,$E6,$65
	.dc.b	$95,$FE,$28,$9D,$CE,$23,$F6,$71,$C5,$D4,$67,$1F,$E0,$93,$99,$DC
	.dc.b	$E7,$1F,$F7,$5C,$E6,$B7,$39,$AF,$E3,$F3,$9A,$DC,$FF,$40,$EF,$F5
	.dc.b	$79,$B2,$FF,$8E,$46,$5F,$44,$65,$FC,$BC,$65,$F4,$C6,$70,$FC,$44
	.dc.b	$E6,$7F,$9A,$3F,$CB,$88,$FF,$34,$6B,$FC,$ED,$AF,$03,$44,$34,$58
	.dc.b	$EE,$FA,$0F,$9D,$E6,$F9,$63,$31,$E8,$32,$6C,$C1,$1F,$9F,$1F,$77
	.dc.b	$CB,$FB,$DE,$D4,$AF,$B5,$19,$7E,$E4,$65,$FB,$71,$97,$EE,$C6,$6B
	.dc.b	$ED,$CE,$5F,$BB,$19,$CB,$ED,$CE,$5F,$BB,$19,$AF,$B7,$39,$7E,$EC
	.dc.b	$65,$66,$9F,$CB,$79,$72,$7F,$24,$FE,$67,$CB,$C4,$7E,$4A,$2E,$3F
	.dc.b	$90,$9B,$8F,$E3,$A6,$D7,$E3,$26,$BE,$08,$B8,$7F,$B6,$A3,$B9,$0E
	.dc.b	$C7,$E8,$E7,$73,$C1,$FF,$7D,$79,$4F,$86,$FD,$77,$C2,$4B,$FF,$48
	.dc.b	$CE,$5B,$39,$7F,$57,$3C,$A2,$47,$EC,$63,$59,$7F,$6F,$AE,$53,$FD
	.dc.b	$91,$AF,$F8,$41,$AF,$F8,$20,$38,$7F,$C1,$6D,$73,$5F,$E1,$87,$84
	.dc.b	$64,$91,$98,$E2,$D7,$89,$98,$DB,$63,$D6,$F9,$DF,$E2,$5E,$6E,$2B
	.dc.b	$D9,$17,$3F,$E0,$B3,$6B,$B9,$F2,$0F,$E2,$8F,$B6,$BF,$C2,$5F,$92
	.dc.b	$FF,$B3,$E7,$F8,$C1,$FC,$7C,$D5,$7C,$38,$D4,$7F,$83,$1B,$3F,$C6
	.dc.b	$23,$81,$FE,$C8,$8C,$CF,$1F,$33,$33,$FC,$88,$EA,$7F,$8D,$A7,$6D
	.dc.b	$8D,$D7,$58,$E9,$1C,$14,$F0,$7F,$46,$70,$AF,$03,$C0,$F6,$90,$F5
	.dc.b	$4F,$39,$FF,$6C,$47,$4E,$1C,$19,$F3,$2C,$7B,$62,$BF,$0E,$2E,$1F
	.dc.b	$32,$6E,$1F,$26,$6D,$7B,$E6,$BE,$A8,$B7,$E2,$23,$89,$CB,$FA,$8F
	.dc.b	$77,$D1,$F0,$92,$FF,$4F,$19,$7E,$94,$65,$FD,$04,$65,$FB,$51,$9C
	.dc.b	$7F,$4D,$39,$9F,$D2,$3F,$97,$D4,$D2,$47,$E6,$5A,$FE,$8E,$36,$E4
	.dc.b	$86,$63,$6F,$63,$D2,$43,$F1,$DE,$77,$9E,$BC,$B1,$97,$F3,$31,$9C
	.dc.b	$FF,$63,$56,$BF,$19,$3F,$77,$C1,$FB,$DD,$B2,$7F,$2D,$19,$7F,$31
	.dc.b	$19,$7F,$3D,$19,$7F,$41,$19,$C3,$F5,$33,$97,$F5,$71,$9C,$3F,$51
	.dc.b	$39,$7F,$55,$19,$AF,$D0,$4E,$5F,$D1,$46,$5F,$CE,$C7,$E5,$7C,$A9
	.dc.b	$1F,$AD,$7F,$31,$E5,$62,$3F,$69,$17,$0F,$DE,$CD,$C3,$F7,$53,$6B
	.dc.b	$F6,$13,$5F,$D5,$C5,$DF,$FD,$CB,$1D,$C8,$76,$3F,$7D,$3B,$9E,$0F
	.dc.b	$F3,$CF,$4F,$F5,$9E,$44,$AF,$E7,$71,$97,$F8,$34,$65,$FE,$69,$19
	.dc.b	$7F,$87,$46,$71,$FE,$75,$39,$9F,$E7,$0F,$E5,$48,$FF,$2B,$6B,$F9
	.dc.b	$36,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$8E,$79,$5E,$58,$CC,$7A,$0C
	.dc.b	$9B,$30,$47,$F9,$7C,$7F,$19,$F9,$AF,$F5,$DC,$FF,$B5,$A7,$FD,$89
	.dc.b	$0F,$6F,$46,$FC,$B9,$5B,$D7,$FA,$B5,$6E,$30,$7F,$B8,$1C,$3F,$EC
	.dc.b	$27,$BB,$67,$AB,$C0,$47,$0B,$F1,$CF,$82,$79,$5F,$07,$91,$C6,$BA
	.dc.b	$D7,$AA,$74,$F2,$9D,$56,$2F,$D3,$5D,$6B,$E0,$F9,$5F,$14,$97,$D7
	.dc.b	$CB,$AF,$AE,$2C,$FC,$89,$B7,$F1,$26,$D7,$C3,$9A,$FB,$A2,$E3,$FE
	.dc.b	$E2,$8F,$54,$60,$FA,$D3,$63,$27,$8C,$F9,$1F,$8C,$3C,$6F,$B9,$F0
	.dc.b	$92,$73,$1F,$D5,$E7,$51,$F1,$E7,$4F,$FB,$DA,$32,$D9,$CE,$5F,$84
	.dc.b	$AE,$51,$27,$FC,$19,$5C,$1F,$F0,$C0,$70,$FF,$86,$DA,$8D,$47,$F8
	.dc.b	$4E,$33,$18,$3A,$C6,$48,$E0,$87,$F8,$86,$3D,$34,$CB,$C1,$5E,$7A
	.dc.b	$FC,$97,$99,$6B,$51,$E3,$BE,$37,$C2,$3F,$19,$5B,$8C,$47,$C5,$32
	.dc.b	$7E,$47,$CA,$FD,$67,$B3,$23,$04,$3F,$44,$E5,$FD,$24,$7E,$CC,$C6
	.dc.b	$1F,$B2,$9C,$4F,$18,$C3,$F9,$F8,$CB,$C8,$C1,$0C,$12,$38,$BE,$23
	.dc.b	$A5,$7E,$EB,$1F,$E0,$84,$8F,$D6,$79,$75,$FD,$D4,$5A,$FE,$0F,$36
	.dc.b	$B9,$57,$EC,$EB,$F8,$74,$E6,$3F,$82,$45,$7F,$7F,$17,$6F,$A3,$1A
	.dc.b	$3C,$27,$32,$1E,$23,$27,$E6,$7D,$EF,$F1,$A7,$88,$F8,$3A,$7E,$86
	.dc.b	$24,$FF,$0F,$9D,$36,$72,$F3,$30,$43,$04,$3F,$3F,$B2,$BF,$34,$48
	.dc.b	$FD,$2B,$5F,$CB,$35,$E2,$D7,$F4,$CD,$7F,$28,$F9,$33,$FC,$BC,$7E
	.dc.b	$B6,$3D,$FE,$6E,$74,$E3,$3F,$C1,$27,$2D,$21,$EA,$AE,$78,$AF,$E4
	.dc.b	$F9,$7F,$2B,$9E,$5F,$C7,$8C,$BF,$26,$32,$AF,$9C,$EA,$31,$8F,$FA
	.dc.b	$6F,$05,$F0,$BD,$D7,$53,$A1,$0E,$27,$F0,$79,$5F,$59,$F4,$D7,$3B
	.dc.b	$ED,$39,$7C,$11,$9B,$E5,$3F,$86,$F2,$A4,$7E,$0D,$FC,$4F,$97,$88
	.dc.b	$F0,$8C,$7F,$DB,$CE,$CF,$C5,$46,$F1,$F9,$75,$71,$F8,$13,$6B,$D5
	.dc.b	$35,$E9,$16,$FA,$C7,$81,$B7,$F5,$A7,$55,$B7,$F9,$E7,$BF,$FA,$0F
	.dc.b	$0A,$61,$FD,$12,$32,$FE,$7E,$32,$FF,$3F,$8C,$BD,$E3,$37,$F8,$ED
	.dc.b	$95,$AA,$FC,$FB,$FC,$DC,$8F,$F4,$26,$BF,$CC,$DA,$F0,$34,$43,$45
	.dc.b	$9F,$04,$79,$1F,$CA,$F9,$77,$3F,$E9,$53,$6B,$F0,$93,$5F,$EA,$11
	.dc.b	$73,$4F,$F0,$4F,$27,$F8,$2F,$8B,$2F,$F8,$14,$65,$FE,$09,$19,$7E
	.dc.b	$94,$65,$FA,$91,$9C,$7F,$53,$39,$7F,$57,$19,$C7,$F4,$13,$97,$F4
	.dc.b	$51,$9A,$FC,$E4,$E5,$FC,$F4,$65,$FC,$84,$7F,$B3,$FF,$25,$FF,$67
	.dc.b	$13,$CA,$FE,$3F,$39,$63,$FE,$2B,$8F,$F6,$A6,$5F,$DD,$31,$FE,$13
	.dc.b	$1C,$67,$55,$C9,$A4,$3F,$43,$1C,$9E,$A6,$22,$CF,$F0,$A8,$B8,$FF
	.dc.b	$7D,$55,$DD,$FC,$E1,$FE,$E9,$FC,$65,$EA,$FB,$7E,$0E,$4F,$F9,$0E
	.dc.b	$72,$D9,$CB,$E4,$9E,$51,$23,$E0,$46,$B0,$FE,$4F,$5C,$A7,$F7,$6D
	.dc.b	$7F,$C2,$ED,$7D,$0D,$7F,$C2,$CD,$4D,$25,$7F,$88,$23,$8B,$5E,$2D
	.dc.b	$78,$B6,$38,$C6,$DF,$34,$FA,$27,$D9,$3F,$4B,$CD,$B3,$98,$F8,$87
	.dc.b	$DF,$8C,$E3,$E2,$8C,$9C,$4F,$C2,$79,$7F,$8B,$F6,$A4,$60,$87,$E0
	.dc.b	$1C,$BF,$82,$8F,$C9,$18,$BF,$C5,$CE,$27,$8C,$61,$FC,$94,$65,$E4
	.dc.b	$60,$86,$09,$1C,$5F,$11,$D2,$BF,$19,$9B,$CB,$FD,$21,$27,$F4,$1E
	.dc.b	$65,$7F,$33,$16,$FF,$43,$36,$F9,$CD,$AF,$F2,$A5,$AB,$11,$F9,$A9
	.dc.b	$B3,$98,$FD,$74,$60,$FD,$B4,$D8,$C9,$E4,$3F,$39,$1F,$BB,$F8,$39
	.dc.b	$18,$21,$FB,$07,$2F,$E6,$23,$F5,$06,$30,$FD,$DC,$E2,$78,$9F,$9C
	.dc.b	$9F,$C7,$11,$FC,$73,$5F,$CC,$6B,$F2,$DC,$90,$CC,$6D,$E8,$66,$BF
	.dc.b	$5B,$1E,$18,$FE,$0F,$E6,$59,$CC,$7B,$1F,$E2,$31,$9C,$BF,$5F,$56
	.dc.b	$BE,$8C,$FF,$1F,$F2,$BF,$90,$6F,$27,$F8,$E4,$65,$FE,$3D,$19,$7F
	.dc.b	$8B,$C6,$5F,$E3,$31,$9C,$3F,$90,$CE,$5F,$E4,$71,$9C,$3F,$1B,$39
	.dc.b	$7F,$1F,$19,$AF,$C2,$CE,$5F,$C3,$C6,$5F,$E3,$F1,$FC,$86,$BF,$C6
	.dc.b	$FF,$92,$FF,$B5,$2F,$F9,$23,$6B,$FD,$44,$D4,$FF,$B8,$63,$7E,$5F
	.dc.b	$36,$CF,$FB,$9E,$78,$1B,$79,$CE,$DA,$6D,$C3,$DB,$CB,$B3,$CA,$69
	.dc.b	$88,$B3,$C4,$87,$F8,$92,$7C,$0F,$A2,$76,$E5,$E4,$FF,$3B,$F8,$EE
	.dc.b	$FD,$52,$30,$43,$A3,$97,$F9,$BC,$7F,$47,$31,$8F,$CC,$9C,$4F,$13
	.dc.b	$F9,$DC,$FF,$2D,$23,$FC,$B5,$AF,$E8,$1A,$F0,$34,$43,$45,$9F,$E9
	.dc.b	$71,$E0,$8F,$85,$E6,$59,$CC,$7A,$DF,$C0,$C6,$72,$FE,$9B,$56,$BF
	.dc.b	$03,$3F,$4B,$CA,$FA,$7E,$CC,$9F,$BD,$19,$7F,$01,$19,$7E,$94,$65
	.dc.b	$FA,$91,$9C,$3F,$23,$39,$7F,$27,$19,$C3,$F2,$73,$97,$F2,$B1,$9A
	.dc.b	$FC,$AC,$E5,$FC,$BC,$65,$FC,$BC,$7E,$2B,$CB,$91,$F9,$57,$F1,$BE
	.dc.b	$5E,$23,$F1,$11,$70,$FD,$2C,$DC,$3F,$4B,$36,$BF,$4B,$35,$FD,$2C
	.dc.b	$5D,$53,$63,$0D,$9F,$11,$FA,$77,$F4,$3F,$8D,$3E,$03,$F4,$3F,$07
	.dc.b	$28,$7F,$A0,$27,$4E,$32,$FF,$58,$C6,$23,$F4,$53,$C5,$E1,$5D,$8C
	.dc.b	$57,$F0,$29,$C4,$F3,$CB,$FC,$4D,$7C,$E3,$FC,$50,$1C,$3A,$BF,$F0
	.dc.b	$54,$62,$3C,$ED,$78,$46,$88,$FF,$84,$E3,$44,$8C,$C7,$F8,$66,$3D
	.dc.b	$97,$EE,$9F,$3C,$FE,$FF,$CE,$B3,$98,$F5,$1E,$29,$FE,$17,$3A,$BF
	.dc.b	$44,$E8,$F2,$9F,$C6,$63,$F8,$9F,$E4,$BF,$ED,$18,$D2,$F6,$F2,$6B
	.dc.b	$CC,$93,$A8,$CB,$C2,$32,$FF,$AE,$27,$93,$B3,$44,$35,$3C,$9E,$13
	.dc.b	$CB,$1F,$F7,$7D,$72,$9D,$9C,$63,$33,$C5,$DD,$71,$C3,$FC,$C5,$7F
	.dc.b	$ED,$8B,$F6,$F9,$7F,$00,$F8,$3E,$65,$8C,$90,$F5,$39,$7D,$91,$70
	.dc.b	$F2,$CD,$BE,$F3,$6B,$F3,$77,$FE,$F0,$7A,$5F,$B2,$75,$09,$D3,$D8
	.dc.b	$DD,$74,$7E,$AD,$FE,$AB,$C5,$FB,$3F,$09,$27,$EF,$6D,$FE,$E7,$6C
	.dc.b	$E5,$FA,$93,$C5,$C9,$F8,$07,$59,$7E,$BA,$75,$3F,$95,$7F,$42,$47
	.dc.b	$F3,$8D,$74,$60,$86,$08,$7E,$BA,$BC,$CE,$67,$C0,$F9,$0F,$C5,$79
	.dc.b	$BE,$28,$CC,$7F,$05,$AF,$CC,$CE,$71,$FD,$DC,$59,$FA,$33,$F9,$EF
	.dc.b	$2B,$F3,$F3,$27,$F3,$F1,$97,$F4,$31,$97,$F3,$51,$97,$F3,$91,$9B
	.dc.b	$FC,$AC,$E5,$FC,$BC,$66,$FF,$23,$39,$7F,$27,$19,$AF,$C2,$CE,$5F
	.dc.b	$C3,$C6,$5F,$C2,$47,$ED,$BC,$99,$1F,$B6,$7F,$73,$E5,$62,$3F,$61
	.dc.b	$17,$0F,$D4,$CD,$C3,$F4,$73,$6B,$F2,$B3,$5F,$C9,$45,$D5,$36,$30
	.dc.b	$D9,$F1,$1F,$C2,$1F,$E7,$BB,$7F,$0C,$E2,$93,$F9,$48,$CB,$FC,$3A
	.dc.b	$32,$FF,$3F,$8C,$BF,$C2,$63,$38,$FF,$40,$9C,$CF,$E2,$5F,$E6,$64
	.dc.b	$7F,$97,$35,$FE,$58,$D7,$81,$A2,$1A,$2C,$77,$7D,$07,$F2,$9F,$79
	.dc.b	$E5,$F9,$A7,$53,$E9,$35,$19,$CF,$FA,$6D,$D9,$FE,$BB,$7C,$9E,$9E
	.dc.b	$57,$F9,$12,$1E,$CD,$31,$12,$7C,$51,$97,$89,$06,$2C,$3F,$B5,$C2
	.dc.b	$FA,$BC,$8E,$13,$B7,$CB,$7D,$5E,$75,$CA,$B8,$C6,$0D,$9C,$A3,$57
	.dc.b	$C6,$FA,$4F,$86,$FD,$4F,$13,$C9,$3B,$AE,$D3,$F0,$FC,$B9,$5F,$45
	.dc.b	$F8,$DE,$5E,$23,$96,$29,$C6,$3E,$08,$56,$30,$F5,$55,$BF,$A7,$35
	.dc.b	$F8,$D1,$6B,$8C,$78,$48,$78,$23,$C6,$E2,$FC,$11,$C6,$BF,$CF,$FE
	.dc.b	$37,$C2,$9F,$27,$C1,$7D,$F3,$38,$CC,$3F,$1D,$19,$7F,$BC,$57,$19
	.dc.b	$FB,$6D,$8F,$C5,$CE,$DF,$99,$19,$4F,$D2,$7F,$C3,$AD,$79,$CF,$F8
	.dc.b	$81,$A5,$38,$B5,$A6,$63,$93,$FE,$38,$6B,$A3,$A3,$82,$47,$37,$CC
	.dc.b	$47,$F4,$5E,$47,$C3,$9F,$5C,$5B,$FB,$93,$63,$CF,$3A,$3F,$6D,$1F
	.dc.b	$AB,$AF,$D6,$7B,$12,$7F,$75,$19,$7F,$79,$19,$7F,$17,$19,$7F,$1B
	.dc.b	$19,$BF,$C6,$4E,$5F,$C7,$46,$6F,$F4,$D3,$97,$F5,$11,$9A,$FE,$03
	.dc.b	$39,$7F,$81,$C6,$5F,$DB,$47,$F0,$3F,$22,$47,$F1,$17,$F8,$37,$93
	.dc.b	$88,$FD,$0C,$5C,$3F,$41,$37,$0F,$DC,$4D,$AF,$E2,$33,$5F,$E0,$91
	.dc.b	$75,$4D,$8C,$36,$7C,$47,$F2,$37,$F3,$B3,$FE,$DC,$FC,$95,$F0,$34
	.dc.b	$CD,$7E,$6A,$70,$FF,$94,$E7,$6E,$48,$7F,$05,$72,$D9,$E6,$FE,$7A
	.dc.b	$7F,$14,$47,$F2,$8D,$7F,$22,$D7,$F4,$2D,$7F,$56,$D7,$F3,$4F,$1C
	.dc.b	$7F,$8C,$E5,$FC,$A6,$7C,$B1,$FC,$72,$DE,$57,$92,$37,$77,$3F,$E6
	.dc.b	$35,$AB,$E3,$1E,$06,$9F,$EA,$08,$D3,$F0,$7C,$8F,$99,$EC,$C8,$F9
	.dc.b	$11,$F3,$CC,$3F,$1E,$30,$E8,$F9,$CE,$B0,$F4,$4E,$A7,$81,$A3,$81
	.dc.b	$E1,$30,$68,$C1,$23,$E1,$4F,$5B,$F7,$D7,$E0,$FC,$89,$7F,$7D,$FC
	.dc.b	$37,$93,$88,$FB,$D1,$70,$F9,$93,$6F,$E4,$CD,$AF,$8D,$35,$62,$9C
	.dc.b	$29,$B1,$86,$CF,$8C,$FC,$93,$FC,$DB,$1F,$CD,$65,$27,$F9,$8C,$65
	.dc.b	$FC,$C4,$65,$FE,$5F,$19,$7F,$31,$19,$C7,$F1,$33,$99,$FE,$62,$FF
	.dc.b	$21,$23,$FC,$AD,$AF,$F2,$86,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$5D
	.dc.b	$E4,$79,$63,$31,$E8,$32,$6C,$C1,$1F,$CF,$47,$EF,$7C,$69,$3F,$B9
	.dc.b	$7F,$80,$7B,$18,$8F,$DC,$45,$AF,$D3,$4D,$D3,$F4,$73,$6B,$F4,$13
	.dc.b	$5F,$B7,$17,$44,$D8,$C3,$67,$C4,$7F,$0B,$7F,$C3,$F7,$FE,$10,$FC
	.dc.b	$97,$FD,$B9,$3F,$E0,$F7,$FD,$95,$19,$F1,$F7,$8F,$F8,$9A,$77,$3C
	.dc.b	$7C,$8C,$D8,$4F,$2A,$E1,$7A,$9C,$D5,$AE,$75,$A9,$FF,$79,$47,$17
	.dc.b	$A3,$A6,$CF,$1A,$91,$C2,$3A,$18,$7D,$91,$87,$47,$67,$57,$EB,$8F
	.dc.b	$53,$FE,$34,$FC,$75,$CD,$E7,$F8,$29,$5A,$CB,$FC,$93,$5A,$8C,$1F
	.dc.b	$0D,$C3,$F3,$67,$8B,$FE,$3D,$8C,$DF,$7B,$CB,$F5,$6B,$E4,$4E,$AF
	.dc.b	$BD,$7F,$83,$E3,$AD,$FF,$96,$83,$87,$67,$F8,$5D,$C3,$E5,$8E,$44
	.dc.b	$74,$D7,$FC,$41,$1C,$1E,$B1,$9B,$F4,$47,$C0,$32,$7D,$AE,$FF,$4C
	.dc.b	$C8,$D5,$7A,$DF,$BB,$39,$7E,$24,$DA,$FC,$B4,$FE,$1F,$C7,$95,$F8
	.dc.b	$67,$F1,$5E,$C6,$23,$F3,$51,$6B,$F3,$D3,$71,$FC,$D4,$DA,$FC,$6C
	.dc.b	$D7,$F3,$31,$75,$4D,$8C,$36,$7C,$47,$E6,$1F,$CE,$78,$FF,$9D,$F2
	.dc.b	$24,$FE,$72,$32,$FE,$7A,$32,$FE,$C6,$32,$FE,$CE,$33,$8F,$ED,$E7
	.dc.b	$2F,$EE,$63,$38,$FE,$D6,$72,$FE,$DE,$33,$5F,$AA,$9C,$BF,$AC,$8C
	.dc.b	$BF,$B8,$8F,$C7,$FC,$AF,$ED,$7E,$0A,$4F,$D8,$8C,$BF,$B6,$8C,$BF
	.dc.b	$9B,$8C,$BF,$C2,$E3,$38,$7E,$AE,$73,$3F,$97,$D9,$1F,$9B,$24,$7E
	.dc.b	$5D,$AF,$E4,$9A,$F1,$6B,$FB,$26,$BF,$B1,$7C,$73,$E1,$8F,$E1,$BE
	.dc.b	$4E,$32,$E1,$1E,$D3,$25,$3D,$85,$9F,$E5,$13,$FC,$6F,$F2,$5F,$F6
	.dc.b	$B4,$FF,$AD,$DB,$3F,$EB,$C6,$D7,$F2,$99,$E1,$E4,$47,$F5,$13,$A7
	.dc.b	$F8,$F9,$93,$A4,$5A,$FE,$71,$16,$BF,$DB,$35,$C6,$0D,$DF,$F8,$25
	.dc.b	$78,$EF,$85,$7F,$37,$AD,$38,$38,$C5,$AE,$37,$DB,$C7,$F8,$1E,$44
	.dc.b	$BF,$4C,$7C,$33,$0F,$C3,$8C,$3A,$3D,$EE,$B1,$F8,$D3,$A9,$E0,$68
	.dc.b	$E0,$7C,$D7,$85,$7C,$0A,$E1,$5C,$AB,$81,$C8,$E0,$7F,$10,$F8,$2F
	.dc.b	$B0,$47,$04,$3E,$D3,$97,$F9,$B3,$FC,$C8,$C4,$7D,$D7,$11,$C4,$FC
	.dc.b	$E3,$FC,$D4,$8F,$F3,$16,$BF,$CB,$1A,$F0,$34,$43,$45,$8F,$EA,$91
	.dc.b	$E0,$8F,$C8,$79,$36,$73,$1E,$B7,$F2,$51,$9B,$FE,$AF,$56,$BF,$05
	.dc.b	$3F,$9B,$F1,$E4,$7E,$51,$FC,$F7,$B1,$88,$FC,$F4,$5C,$3F,$3D,$37
	.dc.b	$0F,$C9,$CD,$AF,$C7,$4D,$7F,$2B,$17,$54,$D8,$C3,$67,$C4,$7E,$99
	.dc.b	$FD,$87,$8B,$FB,$1F,$22,$4F,$EA,$A3,$2F,$EB,$23,$2F,$ED,$63,$2F
	.dc.b	$ED,$E3,$38,$FE,$DE,$72,$FE,$E6,$33,$8F,$EB,$A7,$2F,$EC,$23,$35
	.dc.b	$FA,$B9,$CB,$FA,$D8,$CB,$FB,$48,$FF,$11,$7E,$3D,$FA,$C8,$FD,$97
	.dc.b	$C1,$7F,$0B,$32,$C3,$F8,$2B,$97,$FB,$74,$71,$4D,$CD,$3B,$96,$2C
	.dc.b	$7F,$C6,$D1,$C0,$EC,$7F,$82,$4C,$14,$A6,$6F,$0F,$F8,$FD,$AE,$8F
	.dc.b	$F0,$93,$82,$3C,$1A,$F5,$6B,$E6,$8F,$2C,$7B,$4F,$67,$5F,$9A,$BD
	.dc.b	$55,$E1,$31,$5F,$C8,$A6,$C6,$4F,$8E,$7F,$37,$7F,$CA,$5C,$9F,$EC
	.dc.b	$AA,$E7,$3F,$D4,$BC,$8C,$9A,$F6,$2C,$7F,$87,$6B,$86,$1F,$29,$B1
	.dc.b	$FE,$61,$53,$C2,$3D,$86,$6B,$17,$E0,$AD,$CF,$59,$CD,$F5,$7C,$51
	.dc.b	$92,$1D,$0F,$41,$E5,$BE,$2F,$73,$77,$DE,$3E,$0F,$93,$F0,$BC,$69
	.dc.b	$5E,$D8,$CB,$EF,$8C,$BD,$F1,$4E,$A3,$13,$AB,$F6,$4E,$5F,$74,$67
	.dc.b	$1F,$91,$39,$7E,$54,$66,$BE,$54,$E5,$F9,$91,$97,$FC,$0F,$1F,$AA
	.dc.b	$EB,$FB,$7F,$05,$27,$F4,$B1,$97,$EC,$46,$5F,$91,$AA,$75,$1E,$D9
	.dc.b	$D6,$3F,$84,$AD,$4F,$43,$F3,$15,$F9,$B2,$38,$21,$FA,$77,$2D,$32
	.dc.b	$7E,$99,$CB,$4E,$C7,$99,$FC,$BF,$95,$E6,$AF,$24,$5C,$BF,$25,$36
	.dc.b	$BF,$69,$35,$FF,$16,$C7,$E8,$3C,$89,$1F,$9B,$7F,$47,$EB,$62,$3E
	.dc.b	$F4,$5C,$3F,$2D,$37,$0F,$D0,$4D,$AF,$D0,$4D,$7F,$C7,$71,$75,$4D
	.dc.b	$8C,$36,$7C,$47,$EB,$5F,$DA,$F9,$1F,$B6,$F2,$24,$FE,$C6,$32,$FE
	.dc.b	$CE,$32,$FE,$56,$32,$FE,$5E,$33,$8F,$EB,$E7,$2F,$EC,$63,$38,$FE
	.dc.b	$EE,$72,$FE,$F6,$33,$5F,$BD,$9C,$BF,$BF,$8C,$BF,$E5,$A8,$FE,$7B
	.dc.b	$F3,$9F,$ED,$FF,$C9,$5F,$07,$4C,$D7,$F3,$A9,$C3,$FE,$A9,$9D,$B9
	.dc.b	$21,$FB,$F7,$2D,$9E,$6F,$E7,$27,$F9,$71,$1F,$E5,$CD,$7F,$9D,$35
	.dc.b	$FE,$6E,$D7,$F9,$8B,$5F,$E7,$CF,$1C,$7F,$48,$F8,$1F,$E5,$35,$FE
	.dc.b	$F5,$9F,$EA,$73,$9C,$79,$C7,$B7,$CB,$C3,$D0,$F6,$DF,$28,$F9,$1E
	.dc.b	$4C,$9F,$88,$FC,$BF,$63,$11,$F0,$E2,$D7,$8A,$6D,$FF,$5E,$C5,$58
	.dc.b	$B0,$85,$66,$31,$36,$73,$1E,$22,$1E,$13,$EC,$4E,$E7,$83,$F8,$1F
	.dc.b	$2B,$F0,$5E,$54,$AF,$BD,$19,$7F,$01,$19,$7E,$F4,$65,$FC,$04,$67
	.dc.b	$1F,$9F,$39,$7E,$8C,$67,$1F,$81,$39,$7E,$14,$66,$BE,$0E,$2A,$D4
	.dc.b	$62,$74,$FC,$58,$FF,$0B,$E9,$F9,$BF,$82,$93,$FE,$12,$8C,$BF,$99
	.dc.b	$8C,$BF,$CB,$A3,$2F,$E6,$63,$38,$FF,$2D,$9C,$CF,$F8,$21,$FF,$03
	.dc.b	$11,$FF,$03,$35,$FF,$09,$35,$E0,$68,$86,$8B,$1D,$DF,$41,$FA,$FE
	.dc.b	$4F,$2C,$66,$3D,$06,$4D,$98,$23,$FA,$08,$FD,$F7,$91,$27,$F7,$4F
	.dc.b	$F0,$1F,$5B,$11,$FB,$98,$B5,$FA,$89,$BA,$7E,$5E,$6D,$7D,$69,$AF
	.dc.b	$E5,$E2,$E8,$9B,$18,$6C,$F8,$8F,$E1,$8F,$F9,$02,$FF,$C3,$FF,$92
	.dc.b	$FF,$B1,$EF,$FC,$94,$7F,$92,$DF,$F6,$51,$9A,$FE,$34,$71,$C3,$FC
	.dc.b	$93,$1A,$F2,$B3,$3F,$C6,$4D,$1F,$AE,$8D,$39,$21,$FE,$19,$72,$FF
	.dc.b	$1D,$39,$99,$8E,$6A,$3C,$8F,$2B,$F5,$CE,$BF,$20,$FE,$9F,$3F,$BE
	.dc.b	$9F,$7F,$C1,$F2,$B3,$BE,$79,$5B,$C7,$DB,$3B,$33,$1F,$E2,$FC,$3A
	.dc.b	$BE,$98,$D3,$C4,$FE,$1B,$1A,$C3,$FC,$9A,$F0,$8F,$F3,$88,$70,$FF
	.dc.b	$9D,$1A,$F2,$BF,$F3,$7C,$62,$38,$46,$08,$FF,$9D,$E3,$44,$8C,$C7
	.dc.b	$F9,$F6,$3C,$97,$DE,$7D,$B5,$F8,$4F,$2E,$CE,$6F,$D8,$78,$0B,$8A
	.dc.b	$77,$1E,$73,$78,$76,$7F,$05,$1F,$57,$E1,$7D,$64,$8F,$43,$5F,$33
	.dc.b	$5F,$8F,$28,$C4,$7C,$76,$BF,$41,$F8,$FD,$92,$3E,$23,$5F,$84,$D7
	.dc.b	$AB,$5F,$A4,$D7,$ED,$B5,$FB,$1E,$AF,$D4,$23,$F4,$5A,$FC,$F6,$BE
	.dc.b	$A6,$BF,$7D,$AF,$E2,$1F,$BF,$F3,$DF,$97,$F3,$64,$FC,$C8,$FC,$F1
	.dc.b	$87,$F5,$51,$87,$47,$D6,$75,$A7,$E4,$A7,$53,$F8,$37,$BB,$F8,$96
	.dc.b	$B9,$30,$43,$04,$3E,$AB,$C1,$A7,$88,$FD,$43,$FB,$1F,$37,$C3,$3E
	.dc.b	$39,$D3,$FC,$52,$33,$AF,$E8,$EA,$D7,$F0,$B9,$FD,$F7,$C1,$7E,$EF
	.dc.b	$DA,$93,$F9,$08,$CB,$F9,$68,$CB,$FB,$98,$CB,$FC,$46,$33,$87,$ED
	.dc.b	$A7,$33,$F9,$07,$F6,$24,$7F,$5C,$D7,$F5,$8D,$78,$1A,$21,$A2,$C7
	.dc.b	$77,$D0,$7E,$B3,$CF,$F2,$C6,$63,$D0,$64,$D9,$82,$3F,$CF,$23,$F9
	.dc.b	$17,$91,$FC,$7F,$E1,$25,$7E,$D2,$32,$FE,$EA,$32,$FF,$1D,$8C,$BF
	.dc.b	$CD,$63,$35,$FC,$6A,$73,$3F,$B4,$7F,$8A,$91,$FE,$24,$D7,$F8,$83
	.dc.b	$5E,$06,$88,$68,$B1,$DD,$F4,$1F,$C4,$3C,$EF,$2C,$66,$3D,$06,$4D
	.dc.b	$98,$23,$FD,$82,$3F,$A5,$FC,$FB,$CD,$FE,$89,$E7,$4B,$FE,$3F,$39
	.dc.b	$6C,$E5,$FE,$B1,$3C,$A2,$47,$F2,$48,$D5,$B7,$FF,$27,$CA,$FA,$3F
	.dc.b	$EF,$96,$BF,$D1,$1A,$F0,$6B,$FE,$58,$6B,$CE,$FF,$A5,$BB,$6B,$B6
	.dc.b	$BB,$33,$1C,$89,$1A,$6C,$7C,$02,$1F,$DD,$E3,$DE,$FA,$5B,$5F,$E0
	.dc.b	$C6,$D7,$8E,$78,$CE,$5F,$9B,$F2,$BF,$DF,$3F,$25,$7B,$B4,$FA,$B1
	.dc.b	$27,$FA,$8C,$E9,$B3,$97,$E5,$4F,$17,$27,$62,$1F,$1B,$14,$7B,$C9
	.dc.b	$1F,$11,$AE,$DA,$FD,$56,$BF,$55,$AF,$D5,$7C,$97,$F8,$18,$C1,$EA
	.dc.b	$F4,$73,$8F,$18,$F2,$CE,$D6,$2A,$DC,$72,$3D,$E7,$33,$F2,$FB,$7C
	.dc.b	$BE,$09,$18,$20,$BE,$D6,$9D,$CF,$E4,$EB,$F4,$A6,$2B,$73,$89,$E6
	.dc.b	$7E,$5E,$7F,$10,$47,$F1,$AD,$7E,$83,$5E,$06,$88,$68,$B3,$F0,$F2
	.dc.b	$4F,$86,$7D,$58,$FE,$B3,$07,$A1,$9C,$38,$5D,$AD,$4D,$BF,$D5,$4D
	.dc.b	$7E,$7C,$7F,$04,$F5,$3F,$29,$E8,$C9,$FE,$07,$19,$7F,$27,$19,$7F
	.dc.b	$75,$19,$7C,$A7,$FB,$3B,$14,$EF,$1F,$65,$6E,$BF,$80,$3F,$B1,$23
	.dc.b	$FB,$66,$BF,$99,$6B,$C0,$D1,$0D,$16,$3B,$BE,$A3,$F5,$7D,$3E,$78
	.dc.b	$CC,$7A,$8C,$9B,$30,$47,$F1,$31,$FC,$AB,$9B,$F7,$5E,$EC,$AF,$E5
	.dc.b	$31,$97,$F7,$31,$97,$F9,$04,$65,$F8,$E7,$E8,$23,$55,$F8,$19,$D4
	.dc.b	$FF,$25,$7F,$8B,$11,$FE,$36,$D7,$F7,$CD,$78,$1A,$21,$A2,$C7,$77
	.dc.b	$D2,$7F,$10,$F4,$3C,$D1,$98,$F4,$99,$36,$60,$8F,$EB,$A3,$FA,$7F
	.dc.b	$C3,$7F,$1F,$F8,$59,$5F,$D3,$A3,$2F,$F1,$E8,$CB,$FD,$16,$32,$FC
	.dc.b	$6D,$13,$AB,$FD,$35,$6A,$BF,$A5,$3F,$CE,$C8,$FF,$40,$6B,$FC,$8D
	.dc.b	$AF,$03,$44,$34,$58,$EE,$FA,$4F,$E6,$BE,$8F,$9A,$33,$1E,$93,$26
	.dc.b	$CC,$11,$FE,$27,$1F,$DD,$3C,$4F,$F5,$FF,$92,$F2,$A9,$9B,$FE,$EB
	.dc.b	$38,$7F,$D2,$B3,$B7,$24,$3F,$1A,$E5,$B3,$CD,$FE,$E5,$3F,$D8,$48
	.dc.b	$FF,$66,$6B,$FD,$25,$AF,$F7,$76,$BF,$DD,$DA,$FF,$77,$78,$DF,$F1
	.dc.b	$B7,$C0,$FF,$3F,$A7,$A1,$8E,$4E,$57,$E9,$AD,$4E,$63,$A1,$99,$FF
	.dc.b	$06,$D7,$28,$FF,$5B,$FE,$43,$FE,$CE,$BE,$DE,$F7,$CC,$3F,$DD,$71
	.dc.b	$9B,$FF,$0A,$B6,$82,$D5,$C6,$3A,$9C,$8C,$1F,$39,$C3,$A3,$FC,$27
	.dc.b	$3F,$E1,$24,$D2,$3F,$E1,$66,$BC,$0C,$10,$C1,$0F,$F1,$0B,$E9,$7A
	.dc.b	$9E,$23,$F0,$3E,$95,$C3,$FB,$74,$EC,$E1,$69,$D6,$1F,$12,$E9,$F1
	.dc.b	$27,$4F,$CE,$F3,$BF,$2F,$F0,$D2,$30,$43,$F2,$2E,$5F,$E3,$F8,$27
	.dc.b	$E9,$99,$C7,$F0,$95,$8A,$E4,$7F,$92,$8F,$93,$3F,$1C,$8F,$CE,$6B
	.dc.b	$F1,$1A,$F1,$AF,$03,$B2,$1B,$2D,$7C,$69,$F1,$47,$EA,$BD,$1B,$39
	.dc.b	$8F,$73,$F9,$B8,$CE,$3F,$8C,$AB,$5F,$26,$7F,$23,$F4,$5F,$BE,$F4
	.dc.b	$A4,$FE,$62,$32,$FE,$F2,$32,$FC,$88,$CB,$FB,$C8,$CE,$1F,$72,$73
	.dc.b	$3E,$F7,$F1,$84,$7F,$26,$D7,$F1,$2D,$78,$1A,$21,$A2,$C7,$77,$D0
	.dc.b	$7F,$12,$CF,$CB,$19,$8F,$41,$93,$66,$08,$FF,$03,$8F,$DA,$D7,$F2
	.dc.b	$2F,$7E,$57,$EF,$23,$2F,$F2,$08,$CB,$F8,$C8,$CB,$FC,$82,$33,$5F
	.dc.b	$A0,$9C,$CF,$E0,$DF,$D8,$11,$FD,$BB,$5F,$D6,$B5,$E0,$68,$86,$8B
	.dc.b	$1D,$DF,$41,$FC,$DB,$1F,$2C,$66,$3D,$06,$4D,$98,$23,$FC,$9E,$3F
	.dc.b	$8C,$FA,$9F,$D1,$BE,$16,$57,$F2,$08,$CB,$FD,$12,$32,$FE,$C2,$32
	.dc.b	$FF,$44,$8C,$D7,$F0,$49,$CC,$FE,$9D,$FE,$28,$47,$F8,$DB,$5F,$E2
	.dc.b	$2D,$78,$1A,$21,$A2,$C7,$77,$D0,$7F,$30,$7F,$D0,$5F,$92,$BD,$4F
	.dc.b	$3C,$6E,$7D,$46,$F6,$FF,$66,$D6,$E3,$99,$A2,$4F,$4A,$FE,$83,$C3
	.dc.b	$FE,$D4,$8E,$FC,$97,$0E,$6E,$AF,$FA,$5D,$65,$FE,$8B,$82,$71,$08
	.dc.b	$C3,$C0,$87,$F8,$F9,$CB,$63,$B3,$FC,$92,$7F,$A1,$11,$FE,$92,$D7
	.dc.b	$F9,$FB,$5F,$E3,$8D,$7F,$AD,$B5,$FE,$B7,$C3,$FC,$E1,$E2,$79,$30
	.dc.b	$FE,$87,$3E,$AB,$7A,$5E,$49,$CD,$6F,$1F,$9F,$39,$7B,$C6,$4F,$8B
	.dc.b	$1F,$DB,$BE,$77,$A7,$A1,$27,$FB,$AC,$65,$FB,$47,$CC,$8D,$3F,$D1
	.dc.b	$A3,$4F,$C9,$8C,$DF,$F5,$A9,$CC,$FF,$3D,$7F,$B5,$11,$FE,$E2,$D7
	.dc.b	$FB,$3B,$5E,$06,$88,$68,$B1,$DD,$F4,$9E,$83,$F0,$1D,$DE,$68,$D4
	.dc.b	$7A,$4D,$62,$9D,$63,$FE,$10,$BB,$5F,$8C,$9F,$F0,$AF,$4F,$D3,$F3
	.dc.b	$E4,$FF,$87,$23,$2F,$48,$CB,$FD,$B2,$32,$FE,$3E,$33,$8F,$F8,$02
	.dc.b	$73,$3F,$D8,$DF,$F0,$81,$1F,$F0,$BB,$5F,$F0,$6B,$5E,$06,$88,$68
	.dc.b	$B1,$DD,$F4,$1F,$0B,$CF,$F2,$C6,$63,$D0,$64,$D9,$82,$3F,$C1,$23
	.dc.b	$F5,$FF,$91,$3B,$7F,$C0,$95,$FE,$38,$11,$97,$26,$22,$B4,$CC,$6D
	.dc.b	$E6,$D7,$FC,$20,$D7,$B3,$5F,$F2,$3B,$5E,$87,$28,$EF,$19,$1E,$DE
	.dc.b	$09,$3E,$26,$BE,$26,$BE,$26,$BE,$26,$BE,$27,$BF,$81,$23,$DA,$D7
	.dc.b	$DA,$D7,$DA,$D7,$DA,$D7,$DA,$D7,$E6,$F9,$3E,$C2,$3F,$29,$AF,$CA
	.dc.b	$6B,$F2,$9A,$FC,$A6,$BF,$29,$F8,$FE,$1C,$8F,$AA,$D7,$EA,$B5,$FA
	.dc.b	$AD,$7E,$AB,$5F,$AA,$D7,$F0,$BE,$BF,$D4,$23,$F8,$26,$BF,$82,$6B
	.dc.b	$F8,$26,$BF,$82,$6B,$F8,$27,$EF,$FA,$BF,$CA,$FC,$B9,$3F,$C1,$F2
	.dc.b	$4F,$E6,$8C,$C7,$F4,$59,$C3,$B8,$CE,$BF,$D9,$27,$32,$87,$17,$E5
	.dc.b	$2B,$C3,$3F,$9C,$6B,$F9,$C6,$BF,$9C,$6B,$C4,$E0,$43,$81,$67,$BB
	.dc.b	$FA,$08,$F1,$C7,$F7,$8F,$4A,$D7,$3A,$F3,$36,$75,$1E,$D2,$1A,$3A
	.dc.b	$1F,$CA,$27,$F7,$FE,$5F,$F5,$5F,$4A,$4F,$CB,$8C,$BF,$CB,$A3,$2F
	.dc.b	$33,$F8,$1C,$69,$FE,$E5,$1A,$AF,$5C,$68,$E8,$7F,$04,$9F,$DC,$91
	.dc.b	$C1,$0F,$E0,$EE,$5A,$64,$FE,$12,$E5,$A7,$63,$CD,$92,$3F,$B9,$7A
	.dc.b	$5E,$7A,$F2,$CE,$6B,$84,$7F,$89,$A7,$58,$FF,$14,$BB,$5F,$E1,$69
	.dc.b	$FE,$4F,$EF,$7F,$85,$BF,$25,$7A,$94,$CD,$7E,$56,$70,$FF,$5D,$9D
	.dc.b	$B9,$21,$F3,$5C,$B6,$79,$BE,$C9,$FE,$42,$47,$F9,$7B,$5F,$E5,$ED
	.dc.b	$7F,$97,$B5,$FE,$5E,$D7,$F9,$7B,$C7,$1F,$8A,$F8,$1F,$B7,$4F,$4F
	.dc.b	$19,$72,$BF,$4D,$6A,$73,$1D,$0C,$CF,$F4,$BA,$E5,$1F,$B9,$9F,$95
	.dc.b	$E9,$C8,$C1,$0F,$88,$E5,$FB,$B1,$FD,$60,$C5,$F9,$A7,$13,$C4,$FE
	.dc.b	$9F,$3F,$D2,$08,$FF,$56,$6B,$FD,$59,$AF,$03,$44,$34,$59,$FC,$5C
	.dc.b	$78,$23,$ED,$7A,$76,$73,$1E,$B7,$E5,$C6,$70,$FE,$C3,$56,$83,$05
	.dc.b	$7F,$76,$FA,$5F,$C8,$6B,$2B,$F9,$14,$65,$FC,$74,$65,$FD,$1C,$65
	.dc.b	$FC,$74,$67,$1F,$D1,$CE,$67,$FB,$AB,$FD,$B8,$8F,$F7,$C6,$BF,$DF
	.dc.b	$1A,$F0,$34,$43,$45,$8E,$EF,$A0,$FD,$27,$A7,$E5,$8C,$C7,$A0,$C9
	.dc.b	$B3,$04,$7F,$2B,$1F,$E8,$8F,$B5,$70,$FE,$D7,$D4,$97,$FE,$4D,$9C
	.dc.b	$B6,$72,$FE,$C6,$79,$44,$8F,$DC,$46,$AF,$F8,$ED,$72,$9F,$E4,$8D
	.dc.b	$7F,$C5,$01,$C3,$FE,$2B,$6B,$F7,$9A,$EE,$BF,$C5,$71,$82,$38,$24
	.dc.b	$7F,$8C,$A3,$2D,$8F,$43,$E4,$7F,$8B,$73,$E2,$BD,$31,$76,$FD,$A4
	.dc.b	$DA,$F2,$4D,$7F,$25,$1F,$37,$E0,$FF,$90,$75,$49,$F9,$91,$97,$F8
	.dc.b	$F4,$65,$F8,$B1,$97,$F8,$F4,$67,$0F,$97,$39,$9F,$93,$8A,$3E,$19
	.dc.b	$23,$E0,$35,$FF,$2B,$C6,$DC,$90,$CC,$6D,$EA,$F1,$9F,$0C,$7A,$8F
	.dc.b	$E6,$DD,$9E,$78,$CC,$7A,$8C,$9B,$30,$47,$F9,$3C,$7E,$4F,$B7,$FD
	.dc.b	$B7,$F6,$EE,$FA,$66,$FF,$29,$38,$7F,$D4,$53,$B7,$24,$3F,$B9,$39
	.dc.b	$6C,$F3,$7F,$27,$3F,$8A,$23,$F8,$76,$BF,$21,$AF,$D0,$6B,$F9,$76
	.dc.b	$BF,$97,$78,$DF,$E8,$9F,$03,$FE,$00,$A7,$A9,$8D,$B9,$5F,$A6,$B5
	.dc.b	$39,$8E,$86,$67,$FD,$21,$5C,$A3,$F7,$1F,$4D,$F2,$BD,$59,$18,$21
	.dc.b	$F1,$1C,$BF,$B4,$8F,$E0,$86,$30,$F3,$4E,$27,$89,$F9,$09,$FD,$89
	.dc.b	$1F,$D7,$35,$FC,$7B,$5E,$06,$88,$68,$B3,$FC,$22,$3C,$11,$F6,$BC
	.dc.b	$9B,$39,$8F,$5B,$F2,$E3,$38,$7F,$09,$AB,$5F,$D1,$B0,$57,$F2,$1F
	.dc.b	$23,$F1,$FE,$AC,$AF,$E4,$11,$97,$F1,$B1,$97,$F8,$CC,$65,$FC,$6C
	.dc.b	$67,$1F,$E3,$F3,$99,$FD,$AB,$FC,$58,$8F,$F1,$36,$BF,$B4,$6B,$C0
	.dc.b	$D1,$0D,$16,$3B,$BE,$83,$F4,$7E,$AF,$96,$33,$1E,$83,$26,$CC,$11
	.dc.b	$FC,$A4,$7F,$45,$F8,$6F,$D9,$FA,$72,$BF,$A2,$46,$5F,$D8,$C6,$5F
	.dc.b	$E7,$B1,$97,$F6,$31,$9A,$FE,$87,$39,$9F,$E3,$2F,$F3,$A2,$3F,$CD
	.dc.b	$DA,$FF,$18,$6B,$C0,$D1,$0D,$16,$3B,$BE,$83,$F8,$3F,$AB,$E5,$8C
	.dc.b	$C7,$A0,$C9,$B3,$04,$7F,$71,$1F,$DA,$7D,$1F,$E2,$FE,$AC,$AF,$ED
	.dc.b	$11,$97,$F8,$AC,$65,$FE,$C3,$19,$7F,$8A,$C6,$6B,$FB,$3C,$E6,$7F
	.dc.b	$9E,$BF,$D7,$88,$FF,$5B,$6B,$FC,$F1,$AF,$03,$44,$34,$58,$EE,$FA
	.dc.b	$0F,$E5,$7E,$AF,$96,$33,$1E,$83,$26,$CC,$11,$FE,$39,$1F,$E3,$48
	.dc.b	$FF,$51,$53,$E5,$BF,$9E,$FA,$D2,$7F,$C7,$37,$FE,$A2,$AD,$46,$38
	.dc.b	$B9,$46,$23,$F9,$DD,$71,$7A,$DF,$7B,$E4,$E5,$FE,$5D,$39,$BE,$73
	.dc.b	$6B,$FC,$93,$7D,$5F,$F0,$9B,$5D,$9F,$E1,$63,$71,$5A,$58,$E0,$E0
	.dc.b	$E0,$E1,$34,$91,$CE,$32,$D8,$F8,$17,$EC,$9F,$88,$75,$7F,$D9,$3E
	.dc.b	$63,$2F,$64,$F2,$36,$53,$E3,$98,$8E,$0E,$AB,$BE,$1F,$02,$36,$7F
	.dc.b	$B6,$E7,$6F,$DC,$9E,$F0,$F5,$69,$E2,$8F,$7C,$F5,$72,$F8,$63,$51
	.dc.b	$23,$04,$3B,$1C,$4F,$9C,$FC,$52,$38,$21,$F4,$1C,$B4,$C9,$F7,$9C
	.dc.b	$B4,$EC,$78,$0F,$81,$EA,$79,$5C,$63,$F9,$79,$F8,$33,$98,$C4,$F2
	.dc.b	$7E,$AC,$7C,$3F,$8A,$F8,$FC,$52,$BE,$84,$65,$FA,$91,$97,$F5,$31
	.dc.b	$97,$E9,$46,$6B,$F5,$13,$99,$FC,$CB,$F8,$92,$3F,$95,$6B,$F9,$46
	.dc.b	$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$15,$E9,$F9,$63,$31,$E8,$32,$6C
	.dc.b	$C1,$1F,$CC,$C7,$E1,$F8,$FF,$17,$E9,$CA,$FC,$A4,$65,$FC,$C4,$65
	.dc.b	$FE,$1B,$19,$7F,$2D,$19,$AF,$E1,$93,$99,$FD,$EB,$FA,$D2,$3F,$B9
	.dc.b	$6B,$FB,$86,$BC,$0D,$10,$D1,$63,$BB,$E8,$3F,$5D,$D7,$E5,$8C,$C7
	.dc.b	$A0,$C9,$B3,$04,$7F,$7B,$1F,$AB,$F8,$5F,$D7,$FA,$52,$BF,$71,$19
	.dc.b	$7F,$79,$19,$7F,$98,$C6,$5F,$DD,$46,$6B,$F9,$84,$E6,$7F,$90,$BF
	.dc.b	$C4,$48,$FF,$1D,$6B,$FC,$71,$AF,$03,$44,$34,$58,$EE,$FA,$0F,$E2
	.dc.b	$5E,$8F,$96,$33,$1E,$83,$26,$CC,$11,$FE,$43,$1F,$C3,$FB,$7F,$89
	.dc.b	$F8,$B2,$BF,$8E,$46,$5F,$E4,$11,$97,$FA,$AC,$65,$FE,$3D,$19,$AF
	.dc.b	$EA,$93,$99,$FE,$8A,$FF,$35,$23,$FD,$05,$AF,$F4,$06,$BC,$0D,$10
	.dc.b	$D1,$63,$BB,$E8,$3F,$9B,$7A,$1E,$58,$CC,$7A,$0C,$9B,$30,$47,$FA
	.dc.b	$2C,$7F,$33,$F0,$3F,$A6,$12,$7F,$AC,$35,$FE,$AE,$D7,$FB,$1B,$5F
	.dc.b	$EE,$6D,$7F,$B4,$BF,$D5,$FE,$2E,$47,$F6,$36,$BF,$D8,$9A,$FF,$6E
	.dc.b	$6B,$FD,$F9,$AF,$F7,$46,$AF,$C9,$5E,$7F,$F5,$E5,$F0,$14,$C4,$49
	.dc.b	$FE,$C3,$3A,$6C,$E5,$62,$E9,$FF,$0C,$57,$69,$D9,$0D,$C6,$AF,$95
	.dc.b	$71,$91,$E7,$BF,$23,$E2,$9D,$D7,$69,$CC,$F4,$38,$DF,$85,$C1,$F0
	.dc.b	$6B,$77,$DE,$3B,$1D,$4F,$EF,$FF,$92,$3C,$7F,$F0,$63,$63,$FC,$30
	.dc.b	$D7,$FC,$38,$D7,$FC,$70,$1C,$3F,$E3,$B6,$BE,$E6,$BB,$8E,$E4,$7B
	.dc.b	$B5,$EF,$EC,$79,$08,$F9,$1A,$E8,$C1,$0C,$10,$F3,$CA,$38,$C6,$3C
	.dc.b	$BF,$82,$47,$E0,$B5,$F8,$2D,$78,$1A,$21,$A2,$9E,$0F,$63,$E6,$91
	.dc.b	$F9,$AD,$7E,$6B,$5E,$06,$88,$68,$A7,$B3,$BB,$F3,$DE,$A4,$8F,$BB
	.dc.b	$1F,$E3,$EC,$0E,$63,$F0,$FA,$2B,$53,$27,$F4,$71,$BC,$FF,$15,$7C
	.dc.b	$A7,$A9,$F8,$6A,$FC,$49,$1C,$10,$FC,$4B,$96,$99,$3D,$45,$3F,$14
	.dc.b	$F8,$DF,$CE,$C7,$85,$BF,$05,$9A,$F2,$D5,$DF,$F2,$53,$6B,$F2,$53
	.dc.b	$5F,$CC,$45,$BE,$97,$CA,$7F,$4B,$39,$AD,$36,$3E,$09,$E2,$2C,$F1
	.dc.b	$7C,$F3,$F4,$67,$84,$60,$EA,$70,$BF,$7B,$4F,$D4,$47,$17,$07,$EB
	.dc.b	$E7,$0F,$84,$E8,$4A,$FA,$F3,$A7,$C8,$4A,$F8,$B3,$E2,$38,$3E,$27
	.dc.b	$A4,$72,$7C,$37,$F1,$A7,$8C,$69,$E1,$3D,$63,$D2,$75,$3F,$63,$5F
	.dc.b	$C3,$88,$FF,$0E,$6B,$FC,$39,$AF,$03,$44,$34,$58,$F3,$3E,$33,$F8
	.dc.b	$86,$FF,$06,$BC,$F3,$A3,$F3,$71,$99,$F3,$9C,$4F,$E3,$8E,$5F,$C3
	.dc.b	$9F,$1E,$38,$B6,$7A,$91,$F0,$99,$3A,$C7,$86,$3F,$93,$4F,$ED,$4E
	.dc.b	$D6,$BC,$5F,$F3,$69,$CC,$62,$3F,$47,$18,$74,$43,$D5,$3C,$DC,$5F
	.dc.b	$A1,$CD,$73,$8C,$1E,$58,$F3,$DF,$99,$E2,$FB,$5F,$86,$78,$8C,$9C
	.dc.b	$08,$7C,$A3,$27,$08,$EC,$E4,$E4,$7F,$4E,$BF,$EA,$64,$7F,$A9,$B5
	.dc.b	$D1,$82,$18,$21,$FD,$01,$E1,$1E,$E9,$F6,$4F,$F6,$8F,$C9,$1F,$05
	.dc.b	$F0,$E7,$35,$BA,$FE,$EF,$5A,$7D,$78,$FF,$81,$DF,$4C,$FF,$6F,$9D
	.dc.b	$46,$2B,$99,$1E,$77,$DD,$FC,$39,$B3,$27,$6C,$34,$AB,$AD,$F9,$63
	.dc.b	$E5,$36,$33,$7D,$E7,$2D,$8C,$C7,$82,$79,$61,$EB,$7D,$58,$78,$23
	.dc.b	$17,$C6,$31,$87,$93,$0E,$78,$76,$3A,$56,$DF,$51,$F8,$28,$DD,$F4
	.dc.b	$8D,$90,$F7,$5F,$67,$83,$5F,$8C,$68,$EA,$62,$7A,$DE,$F0,$F5,$CF
	.dc.b	$F8,$C1,$AF,$2B,$F1,$3C,$CF,$C2,$5F,$1C,$3F,$C6,$F1,$9A,$E2,$7C
	.dc.b	$8B,$CB,$89,$DE,$1E,$2A,$F4,$9E,$18,$CD,$F5,$8F,$BE,$78,$0F,$5D
	.dc.b	$73,$72,$F7,$79,$18,$9E,$6F,$06,$BC,$CF,$75,$F9,$A3,$33,$D6,$7E
	.dc.b	$C1,$2F,$ED,$9E,$97,$E6,$5F,$17,$C0,$7D,$7C,$3E,$4C,$F0,$9F,$64
	.dc.b	$FC,$7B,$EE,$7E,$4E,$B7,$3C,$E7,$47,$3C,$31,$7C,$E3,$FC,$9E,$7C
	.dc.b	$63,$FC,$A2,$FE,$B9,$B3,$A7,$47,$C2,$32,$66,$38,$39,$21,$C0,$87
	.dc.b	$90,$8F,$E7,$BF,$0D,$72,$79,$AA,$57,$D4,$38,$4E,$9C,$57,$AD,$FA
	.dc.b	$99,$7F,$9D,$67,$51,$FC,$4F,$BB,$E2,$43,$9E,$9F,$12,$24,$F7,$9D
	.dc.b	$36,$72,$F3,$30,$43,$04,$3E,$5C,$FC,$97,$F8,$73,$6C,$D2,$47,$CB
	.dc.b	$6B,$F5,$A3,$6E,$48,$66,$37,$7F,$C9,$DF,$3B,$FE,$97,$FC,$91,$E4
	.dc.b	$63,$6F,$D5,$D5,$BF,$F5,$4D,$69,$F9,$B1,$97,$D2,$7D,$38,$D3,$F0
	.dc.b	$2F,$C0,$F0,$35,$7C,$70,$F4,$C6,$F0,$FE,$03,$7C,$AF,$8C,$E4,$FE
	.dc.b	$8B,$87,$E6,$E7,$55,$E8,$9E,$C6,$C4,$69,$E2,$6A,$BA,$9F,$0F,$0F
	.dc.b	$25,$FB,$6B,$85,$7B,$8E,$98,$7A,$2F,$84,$73,$C3,$C4,$F8,$6B,$8D
	.dc.b	$6E,$73,$5E,$07,$DF,$87,$82,$F9,$1F,$41,$C3,$CD,$B3,$C6,$34,$7C
	.dc.b	$C3,$89,$E7,$7C,$AF,$FB,$62,$FC,$78,$7B,$6B,$E5,$1D,$2F,$C7,$5C
	.dc.b	$CF,$7C,$E9,$C5,$F8,$5E,$D3,$C1,$ED,$5C,$DF,$29,$83,$CB,$87,$C3
	.dc.b	$73,$5E,$63,$A4,$66,$3E,$CC,$F8,$5F,$19,$E2,$AF,$05,$70,$BD,$3C
	.dc.b	$E3,$E7,$92,$7B,$BC,$0C,$3E,$A3,$24,$30,$4C,$3F,$06,$7C,$A7,$E7
	.dc.b	$B8,$BF,$81,$1E,$83,$CC,$7E,$40,$FB,$46,$0E,$B8,$7B,$E7,$94,$7C
	.dc.b	$6C,$3D,$77,$F3,$63,$FB,$29,$D4,$FE,$6F,$3F,$B0,$72,$D7,$81,$0F
	.dc.b	$A6,$6A,$FF,$89,$46,$E3,$26,$CE,$C6,$CF,$43,$FA,$4F,$C3,$DE,$3F
	.dc.b	$9A,$7E,$44,$E6,$31,$1E,$7C,$BE,$A2,$15,$C9,$EB,$7C,$23,$11,$F8
	.dc.b	$92,$4F,$27,$8D,$76,$BF,$93,$09,$61,$FE,$0C,$BE,$71,$C6,$B6,$EA
	.dc.b	$BC,$93,$AA,$FC,$34,$7B,$30,$FE,$D5,$39,$8C,$1F,$6D,$C3,$EF,$9E
	.dc.b	$2D,$3A,$E1,$E7,$AF,$73,$63,$93,$5F,$7C,$FA,$63,$17,$F0,$6F,$BE
	.dc.b	$1E,$AA,$EF,$18,$C3,$71,$C7,$0F,$9A,$76,$29,$D8,$F2,$98,$BE,$D8
	.dc.b	$72,$8F,$D2,$5F,$57,$55,$D5,$E4,$7B,$C9,$1B,$73,$87,$E1,$5E,$27
	.dc.b	$4A,$EC,$F0,$6C,$6E,$38,$E1,$F2,$1E,$26,$CB,$1B,$C3,$EF,$4E,$E7
	.dc.b	$89,$F0,$4A,$7A,$2F,$CD,$3C,$23,$FC,$1D,$18,$C3,$ED,$DF,$56,$DF
	.dc.b	$5B,$E1,$7C,$F0,$FC,$39,$EC,$C3,$F0,$A7,$38,$F8,$07,$68,$D1,$FD
	.dc.b	$77,$3F,$DA,$92,$3F,$C6,$18,$FF,$84,$9F,$E0,$31,$96,$9A,$70,$FF
	.dc.b	$04,$6B,$FD,$B9,$FE,$03,$EF,$C8,$EA,$D7,$F8,$6B,$5F,$E1,$AD,$7F
	.dc.b	$BB,$B5,$FF,$03,$B5,$FE,$E3,$BF,$F0,$E2,$3E,$A6,$BF,$C6,$9A,$FF
	.dc.b	$1A,$6B,$FE,$04,$6B,$FE,$14,$7F,$8C,$F5,$C8,$F8,$AD,$7F,$92,$B5
	.dc.b	$FE,$4A,$D7,$FC,$1E,$D7,$FC,$3E,$D7,$FC,$13,$E9,$7F,$26,$23,$F4
	.dc.b	$9A,$FF,$32,$6B,$FC,$C9,$AF,$F8,$71,$AF,$F8,$B1,$FF,$21,$7A,$7F
	.dc.b	$C4,$63,$FC,$E2,$D7,$F9,$FB,$5F,$F5,$2B,$5F,$E5,$4D,$4D,$26,$DD
	.dc.b	$89,$1F,$E7,$98,$CB,$7D,$CF,$E3,$BE,$34,$8F,$F3,$74,$A7,$FD,$14
	.dc.b	$D8,$FD,$2C,$ED,$FE,$61,$5E,$6A,$D9,$9D,$FD,$DB,$AF,$37,$DA,$B5
	.dc.b	$7F,$CD,$E7,$33,$D0,$F7,$CE,$9F,$E6,$F1,$AB,$F9,$58,$79,$67,$81
	.dc.b	$FD,$3A,$78,$3F,$1A,$B8,$5F,$78,$F5,$9E,$32,$1C,$B0,$F4,$3A,$3A
	.dc.b	$DF,$7C,$3C,$37,$DA,$B7,$5A,$8E,$EF,$A0,$E5,$3C,$1E,$EF,$C6,$C3
	.dc.b	$CF,$7D,$70,$F9,$95,$C1,$F4,$3D,$4E,$2E,$2B,$94,$60,$F9,$27,$13
	.dc.b	$C4,$71,$23,$C1,$F8,$15,$E8,$7E,$41,$C4,$F9,$87,$53,$27,$03,$55
	.dc.b	$E9,$7C,$26,$49,$3F,$05,$F0,$D7,$A4,$8F,$08,$D1,$E5,$24,$7C,$48
	.dc.b	$E2,$7B,$23,$2F,$18,$D4,$62,$3D,$27,$03,$D2,$E7,$0F,$B2,$F8,$27
	.dc.b	$89,$E4,$9E,$31,$97,$B1,$E1,$8D,$57,$C1,$8E,$B1,$C6,$BA,$BD,$23
	.dc.b	$89,$D1,$E3,$87,$D1,$9D,$CF,$53,$75,$D6,$33,$1C,$4C,$CF,$E4,$F0
	.dc.b	$FB,$55,$B7,$F1,$B1,$D5,$B3,$F6,$EB,$99,$D2,$BD,$47,$18,$FB,$B5
	.dc.b	$C5,$A7,$1A,$E6,$FC,$DA,$E7,$1E,$29,$F2,$4F,$E2,$6B,$A3,$F8,$4B
	.dc.b	$F8,$8F,$E3,$0F,$E4,$73,$FC,$A0,$8F,$F1,$92,$1F,$CA,$9C,$B4,$C9
	.dc.b	$C8,$A7,$69,$F6,$C7,$F8,$FB,$C0,$F3,$4F,$C3,$6B,$EB,$6E,$3F,$81
	.dc.b	$8F,$F2,$DC,$62,$FD,$38,$A3,$13,$F7,$A3,$0F,$C1,$BF,$BD,$3A,$8F
	.dc.b	$C3,$9F,$E5,$47,$18,$7D,$F6,$CF,$9E,$79,$1E,$72,$9E,$72,$E1,$E7
	.dc.b	$BF,$0C,$7E,$86,$FE,$0B,$FE,$20,$BE,$AD,$38,$E1,$EC,$8C,$C7,$F2
	.dc.b	$56,$DF,$F3,$E6,$CF,$F9,$C5,$B3,$EB,$39,$9F,$0D,$CB,$F2,$8C,$9F
	.dc.b	$08,$C1,$E7,$C3,$E3,$BA,$BF,$88,$D8,$F8,$8E,$CF,$88,$53,$E2,$16
	.dc.b	$FE,$25,$78,$B0,$F8,$B1,$F7,$6F,$E2,$D7,$BC,$EC,$7C,$53,$96,$1F
	.dc.b	$15,$C9,$B3,$27,$BC,$CC,$FC,$76,$C7,$C7,$69,$F1,$CE,$27,$C3,$AF
	.dc.b	$2C,$7F,$AA,$E1,$F7,$5F,$E9,$2F,$15,$ED,$E7,$0F,$F0,$74,$62,$BF
	.dc.b	$D6,$51,$FE,$11,$AF,$F4,$C9,$C6,$B8,$5F,$5B,$E1,$7F,$DB,$2B,$7C
	.dc.b	$7A,$8F,$F5,$E4,$EB,$0E,$F8,$66,$7C,$B5,$6F,$B4,$78,$B0,$ED,$59
	.dc.b	$A9,$39,$3A,$C7,$3C,$3B,$57,$C1,$BF,$1D,$6C,$4E,$9E,$26,$63,$6F
	.dc.b	$F8,$1A,$34,$F0,$AF,$0B,$EF,$BF,$04,$48,$E5,$1A,$C3,$DF,$1E,$1A
	.dc.b	$E7,$7D,$A7,$33,$DE,$F8,$61,$E9,$C3,$A4,$E4,$EA,$43,$04,$AF,$B1
	.dc.b	$1C,$A3,$DA,$66,$7A,$3C,$67,$C2,$48,$CC,$7B,$4F,$45,$78,$A7,$AC
	.dc.b	$71,$9F,$AA,$E8,$C1,$FE,$FB,$BF,$B6,$F8,$4C,$5F,$95,$F9,$87,$13
	.dc.b	$C0,$4A,$F2,$3E,$A7,$C2,$E6,$BC,$2F,$53,$F1,$67,$88,$EA,$66,$BE
	.dc.b	$84,$70,$9E,$B3,$C0,$E8,$77,$8D,$1F,$12,$3C,$27,$0C,$3E,$DD,$FA
	.dc.b	$DF,$C3,$98,$AF,$13,$8C,$3E,$06,$1F,$12,$FC,$2F,$92,$BC,$04,$3C
	.dc.b	$91,$E0,$C3,$C7,$19,$9E,$0F,$BA,$79,$E1,$E3,$BF,$B1,$7E,$92,$57
	.dc.b	$5C,$3C,$CF,$C0,$39,$47,$67,$C2,$F0,$72,$78,$6B,$B3,$F6,$6F,$CC
	.dc.b	$FD,$E3,$13,$CF,$0F,$3D,$FC,$39,$FA,$D1,$F0,$E7,$57,$D2,$7E,$C3
	.dc.b	$FB,$07,$F8,$E1,$F5,$CB,$28,$CC,$7E,$68,$CB,$87,$C8,$6A,$38,$39
	.dc.b	$6B,$98,$E0,$F7,$3A,$10,$E7,$7F,$01,$F8,$E7,$F2,$0F,$0B,$EF,$CF
	.dc.b	$E1,$5B,$1F,$84,$6B,$F4,$70,$46,$2B,$A4,$62,$7E,$3D,$73,$69,$CC
	.dc.b	$A7,$3A,$E5,$87,$D9,$AE,$53,$F7,$F0,$FA,$75,$A8,$F5,$D7,$F3,$D8
	.dc.b	$C4,$7B,$0F,$E5,$2E,$27,$FA,$63,$6F,$F9,$93,$67,$F9,$03,$5F,$E5
	.dc.b	$8D,$C3,$FA,$8B,$6F,$C1,$3E,$A9,$F8,$33,$98,$F9,$86,$67,$D9,$19
	.dc.b	$8F,$29,$93,$D6,$70,$3E,$05,$7C,$96,$9F,$24,$A7,$C9,$39,$1F,$07
	.dc.b	$0F,$93,$7F,$7F,$0F,$93,$3A,$7E,$1B,$6B,$E1,$B4,$F8,$65,$3E,$19
	.dc.b	$67,$E1,$B7,$0F,$86,$F9,$23,$E1,$B9,$3D,$8F,$8E,$7D,$8D,$8F,$63
	.dc.b	$5F,$61,$FE,$E2,$FC,$91,$EA,$7E,$79,$B1,$FA,$76,$BF,$AC,$6B,$FB
	.dc.b	$26,$BF,$E0,$F6,$A0,$E1,$B1,$DC,$8F,$76,$BD,$FC,$8F,$21,$1F,$23
	.dc.b	$5D,$18,$21,$82,$1E,$77,$83,$E9,$F5,$7E,$01,$1F,$80,$D7,$E0,$35
	.dc.b	$E0,$68,$86,$8A,$7C,$5F,$8E,$F9,$84,$7E,$63,$5F,$98,$D7,$81,$A2
	.dc.b	$1A,$29,$F4,$7A,$3E,$C1,$1F,$B0,$D7,$EC,$35,$E0,$68,$86,$8A,$7D
	.dc.b	$DE,$8F,$C2,$11,$FC,$23,$5F,$C2,$35,$E0,$68,$86,$8A,$7E,$27,$A3
	.dc.b	$F2,$04,$7F,$20,$D7,$F2,$0D,$78,$1A,$21,$A2,$9F,$95,$E8,$FC,$E1
	.dc.b	$1F,$CE,$35,$FC,$E3,$5E,$06,$88,$68,$A7,$E8,$7A,$3F,$50,$47,$F5
	.dc.b	$0D,$7F,$50,$D7,$81,$A2,$1A,$29,$FD,$7B,$F2,$47,$07,$F5,$26,$C7
	.dc.b	$F5,$B6,$BF,$D8,$1A,$F1,$8E,$04,$78,$35,$E0,$D7,$81,$D7,$8F,$C0
	.dc.b	$47,$04,$3C,$0E,$5A,$64,$F5,$39,$69,$E9,$F2,$3D,$84,$70,$43,$D8
	.dc.b	$E5,$A6,$4F,$8C,$E5,$A7,$C5,$F2,$3E,$41,$1C,$10,$F9,$0E,$5A,$64
	.dc.b	$FA,$4E,$5A,$7D,$1F,$23,$EA,$11,$C1,$0F,$A8,$E5,$A6,$4F,$BC,$E5
	.dc.b	$A7,$DD,$F6,$FF,$00,$47,$04,$3F,$00,$E5,$A6,$4F,$C5,$39,$69,$F8
	.dc.b	$9F,$1F,$F1,$84,$70,$43,$F1,$8E,$5A,$64,$FC,$B3,$96,$9F,$95,$F7
	.dc.b	$3F,$30,$47,$04,$3F,$30,$E5,$A6,$4F,$D1,$39,$69,$FA,$1F,$83,$FD
	.dc.b	$21,$1C,$10,$FD,$23,$96,$99,$3F,$5C,$E5,$A7,$EB,$7D,$EF,$D8,$11
	.dc.b	$C1,$0F,$D8,$39,$69,$93,$F7,$4E,$5A,$7F,$94,$FF,$25,$70,$FF,$B8
	.dc.b	$DB,$1F,$ED,$76,$AE,$9F,$DF,$C3,$8B,$85,$4A,$E0,$F3,$8C,$47,$1E
	.dc.b	$0F,$EE,$8A,$BA,$DF,$4B,$E1,$3C,$88,$71,$3F,$BE,$4F,$0A,$E6,$F8
	.dc.b	$E3,$75,$CD,$B3,$AB,$F1,$CE,$67,$A1,$E3,$BF,$0B,$8D,$1E,$87,$C5
	.dc.b	$E5,$93,$FE,$FB,$9C,$B6,$72,$F4,$C9,$5F,$03,$0F,$E4,$F7,$C2,$BB
	.dc.b	$5F,$09,$F9,$B5,$BC,$BE,$3D,$EE,$78,$D6,$E7,$E0,$D6,$EF,$FC,$41
	.dc.b	$7B,$9E,$35,$B9,$F9,$35,$B5,$5B,$78,$CE,$DF,$AD,$3B,$9D,$47,$09
	.dc.b	$D6,$1E,$E8,$F8,$44,$3E,$15,$73,$AE,$27,$29,$E1,$3A,$38,$61,$EE
	.dc.b	$7F,$25,$5C,$6F,$C4,$78,$1B,$1C,$30,$F8,$6F,$03,$07,$B0,$C1,$DC
	.dc.b	$FA,$26,$8F,$79,$AA,$F1,$3A,$38,$3F,$59,$E1,$87,$AE,$BC,$D1,$CF
	.dc.b	$0F,$6D,$6E,$39,$3D,$0C,$12,$FE,$BC,$71,$AE,$6F,$08,$F6,$BE,$27
	.dc.b	$6E,$4F,$BA,$7A,$5C,$CF,$77,$8B,$DE,$3E,$21,$C4,$F8,$27,$53,$2F
	.dc.b	$17,$A9,$93,$F0,$47,$9C,$F0,$1C,$23,$C2,$78,$08,$E4,$87,$82,$3C
	.dc.b	$8F,$43,$C9,$5F,$2A,$78,$C6,$67,$C8,$F1,$8C,$DF,$E3,$B0,$FC,$34
	.dc.b	$FD,$66,$BE,$F8,$FB,$0F,$A0,$E9,$3E,$87,$DF,$5E,$49,$E9,$1B,$BF
	.dc.b	$23,$EA,$8E,$6F,$C5,$76,$FC,$78,$DB,$D8,$F8,$2E,$0C,$C7,$09,$F8
	.dc.b	$71,$8B,$FC,$FD,$73,$C2,$D7,$E5,$A3,$F9,$69,$F2,$4B,$7F,$C3,$DB
	.dc.b	$5A,$7F,$12,$60,$C1,$C4,$C9,$0F,$C3,$9B,$39,$47,$57,$E4,$1E,$D2
	.dc.b	$5B,$9F,$CF,$8E,$AB,$F8,$14,$62,$BE,$55,$FF,$53,$8C,$47,$39,$DB
	.dc.b	$F6,$CF,$D7,$46,$A3,$E4,$57,$F4,$48,$C5,$72,$BF,$15,$F1,$3F,$1B
	.dc.b	$3A,$79,$99,$C3,$F1,$B3,$C2,$3F,$11,$5C,$EB,$84,$72,$38,$5F,$C2
	.dc.b	$AE,$17,$F9,$6C,$3C,$B5,$FD,$6A,$F8,$CF,$66,$DF,$16,$CF,$E1,$9A
	.dc.b	$FC,$C6,$D7,$C4,$8F,$29,$EA,$29,$E5,$2D,$77,$7C,$E7,$52,$9D,$4B
	.dc.b	$1D,$63,$D5,$87,$C1,$9C,$BF,$5D,$A7,$F8,$26,$BC,$B1,$98,$F0,$DF
	.dc.b	$B5,$D9,$FD,$9E,$76,$F1,$33,$5F,$E1,$08,$CC,$6C,$CC,$FD,$F6,$9F
	.dc.b	$34,$A7,$F8,$3E,$38,$CF,$38,$CE,$1F,$5A,$BF,$C2,$BF,$82,$BE,$A7
	.dc.b	$FC,$49,$D7,$2C,$7F,$82,$27,$4E,$30,$FF,$51,$C6,$23,$FC,$05,$3C
	.dc.b	$5E,$15,$FD,$C6,$75,$5F,$E2,$CA,$D3,$CA,$74,$62,$FF,$C3,$F3,$8B
	.dc.b	$F0,$47,$A2,$75,$1F,$E2,$69,$D3,$E0,$3F,$C7,$8E,$95,$F8,$AB,$94
	.dc.b	$68,$87,$12,$5F,$9D,$CB,$E7,$7C,$67,$C6,$9E,$8F,$8C,$DC,$F6,$79
	.dc.b	$BE,$C7,$27,$1B,$F4,$D7,$78,$F9,$67,$63,$D3,$1A,$9E,$51,$A3,$A9
	.dc.b	$CC,$EA,$7F,$94,$6F,$83,$93,$85,$77,$3D,$8F,$23,$81,$CA,$FD,$AF
	.dc.b	$88,$EB,$82,$BA,$C6,$4E,$B7,$E6,$BE,$A4,$70,$48,$F2,$46,$5F,$94
	.dc.b	$72,$33,$1F,$2E,$3D,$0F,$18,$F0,$47,$17,$D2,$71,$33,$1C,$4E,$15
	.dc.b	$E9,$8F,$9D,$7E,$E3,$DF,$19,$8E,$51,$9C,$3F,$19,$39,$BF,$04,$F0
	.dc.b	$C3,$F2,$0E,$EF,$E3,$C7,$C6,$8D,$47,$CD,$BF,$93,$7C,$F0,$F6,$57
	.dc.b	$1A,$F5,$4E,$5F,$86,$43,$DF,$87,$C4,$71,$3D,$9F,$4D,$F6,$C3,$CA
	.dc.b	$D8,$F1,$1F,$09,$F6,$C7,$9D,$B1,$92,$1E,$48,$D3,$EE,$33,$5F,$4E
	.dc.b	$FD,$11,$E6,$C3,$E9,$3E,$98,$C9,$AC,$3E,$04,$7F,$7E,$7F,$87,$37
	.dc.b	$0F,$EF,$2F,$E3,$CC,$1F,$DF,$23,$6F,$EC,$4C,$9F,$E0,$37,$2F,$E8
	.dc.b	$CC,$90,$F2,$9F,$88,$3C,$DE,$A7,$B9,$FD,$35,$23,$18,$7B,$DB,$87
	.dc.b	$29,$D5,$7C,$48,$EB,$7F,$8E,$AD,$CF,$63,$47,$DA,$C3,$FD,$B1,$1A
	.dc.b	$8F,$21,$FE,$CD,$71,$3C,$E3,$89,$A7,$0F,$EF,$5F,$C6,$61,$FE,$DE
	.dc.b	$9E,$35,$DA,$B8,$D6,$AF,$FD,$A3,$5A,$9F,$F7,$43,$5F,$F6,$13,$6F
	.dc.b	$A4,$72,$BF,$CE,$CE,$6B,$9C,$E4,$F3,$E1,$F1,$9F,$35,$70,$74,$7D
	.dc.b	$0A,$E4,$FF,$B9,$9A,$FC,$33,$77,$F4,$DB,$7F,$4E,$77,$7D,$EB,$F7
	.dc.b	$37,$DE,$FC,$27,$92,$B6,$47,$99,$F5,$CD,$9E,$53,$27,$D6,$AE,$D7
	.dc.b	$F8,$0A,$ED,$7E,$0A,$EC,$70,$C3,$EB,$C6,$63,$A1,$D9,$F6,$36,$BD
	.dc.b	$8F,$F4,$7F,$C9,$5C,$FF,$A6,$6C,$7F,$98,$1A,$FF,$99,$5A,$FF,$9B
	.dc.b	$1A,$FF,$80,$83,$87,$FC,$0A,$D7,$31,$DC,$8F,$76,$BD,$F9,$FC,$84
	.dc.b	$7C,$8D,$74,$60,$86,$08,$79,$DE,$0F,$A7,$9F,$E0,$11,$F8,$0D,$7E
	.dc.b	$03,$5E,$06,$88,$68,$A7,$C5,$E0,$F9,$84,$7E,$63,$5F,$98,$D7,$81
	.dc.b	$A2,$1A,$29,$F4,$79,$FE,$C1,$1F,$B0,$D7,$EC,$35,$E0,$68,$86,$8A
	.dc.b	$7D,$DE,$7F,$C2,$11,$FC,$23,$5F,$C2,$35,$E0,$68,$86,$8A,$7E,$27
	.dc.b	$9F,$F2,$04,$7F,$20,$D7,$F2,$0D,$78,$1A,$21,$A2,$9F,$95,$E7,$FC
	.dc.b	$E1,$1F,$CE,$35,$FC,$E3,$5E,$06,$88,$68,$A7,$E8,$7E,$3F,$F5,$04
	.dc.b	$7F,$50,$D7,$F5,$0D,$78,$1A,$21,$A2,$9F,$AD,$F1,$FF,$59,$BB,$7F
	.dc.b	$D7,$71,$3E,$03,$31,$86,$DF,$EC,$2E,$BF,$B4,$6D,$71,$8F,$F2,$43
	.dc.b	$55,$F7,$79,$1F,$C1,$E7,$8D,$48,$C1,$0E,$37,$E3,$7A,$C7,$8B,$85
	.dc.b	$54,$E5,$3C,$0F,$F4,$5E,$1E,$3C,$3C,$51,$E9,$3B,$38,$8D,$61,$DE
	.dc.b	$79,$1B,$2D,$F8,$1F,$19,$0C,$DF,$8D,$C5,$F5,$C3,$11,$A3,$37,$DE
	.dc.b	$7C,$B7,$D6,$74,$7F,$27,$BF,$E5,$84,$7F,$96,$35,$D1,$82,$18,$21
	.dc.b	$F8,$33,$33,$E0,$8F,$2B,$F0,$FD,$FB,$7F,$CC,$EF,$C9,$3B,$BF,$BF
	.dc.b	$5A,$7E,$E9,$E8,$8E,$4F,$CE,$AE,$57,$B8,$62,$BC,$61,$DA,$F7,$5D
	.dc.b	$8E,$47,$6C,$3C,$4F,$D6,$BE,$4F,$F5,$3F,$87,$FC,$AF,$BB,$23,$04
	.dc.b	$38,$39,$7E,$EC,$7E,$34,$C5,$7E,$16,$71,$3C,$63,$0F,$E8,$23,$2F
	.dc.b	$23,$04,$30,$48,$E2,$F8,$8E,$97,$F3,$EB,$FB,$51,$23,$FB,$53,$5F
	.dc.b	$ED,$4D,$7F,$B5,$35,$E4,$70,$21,$C0,$EE,$7F,$6C,$7F,$77,$E7,$7A
	.dc.b	$AB,$F2,$71,$6B,$F8,$3C,$DA,$E0,$72,$3E,$DB,$FC,$1B,$F2,$37,$8F
	.dc.b	$FC,$1A,$61,$FE,$CD,$8C,$B6,$32,$FD,$B3,$93,$1F,$E1,$A6,$B0,$FD
	.dc.b	$DC,$F2,$8F,$DB,$B6,$3F,$79,$1B,$BF,$F0,$64,$E5,$B5,$99,$FE,$0E
	.dc.b	$77,$39,$B9,$7F,$2E,$66,$BF,$85,$FE,$4A,$F6,$FF,$D0,$AD,$9F,$F0
	.dc.b	$D1,$FE,$1B,$70,$E0,$38,$76,$D7,$84,$74,$23,$C4,$E2,$E1,$E9,$1C
	.dc.b	$51,$A3,$82,$F9,$1F,$39,$23,$2D,$79,$B5,$F2,$B5,$F6,$35,$EF,$1D
	.dc.b	$5E,$FE,$07,$C4,$23,$82,$1E,$87,$2D,$32,$7C,$47,$2D,$3D,$7E,$7F
	.dc.b	$D0,$23,$82,$1F,$11,$CB,$4C,$9F,$41,$CB,$4F,$8F,$E6,$FD,$C2,$38
	.dc.b	$21,$F4,$1C,$B4,$C9,$F7,$1C,$B4,$E3,$8F,$E2,$A3,$FB,$6D,$9F,$D5
	.dc.b	$7B,$BF,$DA,$A9,$F8,$B3,$E4,$3F,$19,$5C,$09,$3F,$AD,$AE,$15,$B8
	.dc.b	$FA,$B3,$A7,$FA,$ED,$F7,$78,$61,$F8,$8A,$E1,$1F,$96,$AE,$13,$E7
	.dc.b	$8D,$9E,$4C,$3C,$6F,$E4,$23,$4F,$69,$F5,$C7,$59,$DC,$E4,$E1,$B6
	.dc.b	$27,$F7,$B8,$FA,$AB,$F7,$2D,$AF,$59,$CC,$A7,$8C,$FD,$0E,$E8,$E1
	.dc.b	$08,$C5,$71,$8F,$71,$72,$F8,$8F,$33,$85,$72,$AE,$31,$99,$DE,$5F
	.dc.b	$21,$B8,$7A,$30,$ED,$7C,$2B,$59,$7F,$A7,$DB,$86,$B1,$EC,$7A,$31
	.dc.b	$E7,$1E,$E9,$FA,$B7,$CF,$1E,$D3,$E3,$C3,$A4,$7C,$EA,$DE,$1D,$31
	.dc.b	$F7,$56,$AF,$A5,$F2,$BD,$D6,$A6,$1F,$0F,$0F,$03,$F3,$B0,$F3,$61
	.dc.b	$CE,$3E,$59,$B8,$F0,$3E,$53,$9D,$F6,$75,$78,$C3,$D0,$78,$C9,$1E
	.dc.b	$62,$1F,$52,$3E,$41,$C5,$CE,$1E,$67,$47,$84,$95,$E1,$25,$78,$23
	.dc.b	$8D,$FA,$27,$EA,$C7,$09,$F1,$BD,$23,$24,$8C,$D7,$87,$1F,$B3,$5F
	.dc.b	$12,$BB,$BF,$27,$0F,$43,$9B,$E3,$87,$CC,$8C,$E1,$EB,$3A,$63,$F3
	.dc.b	$F1,$EB,$19,$AE,$15,$E7,$3B,$1F,$38,$C1,$DC,$CE,$3E,$47,$04,$9E
	.dc.b	$98,$F9,$AF,$BD,$FC,$5C,$7E,$45,$EE,$30,$FD,$69,$DD,$7D,$68,$F0
	.dc.b	$C6,$1F,$26,$1F,$25,$FA,$33,$B8,$C4,$75,$8C,$5F,$82,$B9,$57,$A5
	.dc.b	$F9,$E6,$23,$EB,$B8,$C7,$F2,$D8,$FB,$30,$F1,$C6,$31,$F4,$C7,$C2
	.dc.b	$AE,$4F,$C6,$BF,$05,$6F,$1F,$67,$CE,$FE,$18,$93,$FB,$47,$F3,$C6
	.dc.b	$2B,$F8,$BB,$88,$D1,$FC,$91,$D4,$7E,$61,$F6,$57,$DF,$3C,$11,$F2
	.dc.b	$27,$C6,$FC,$43,$CC,$7D,$18,$DE,$5C,$A3,$F1,$78,$FB,$B1,$F8,$91
	.dc.b	$C4,$FC,$E5,$F4,$7B,$1D,$0F,$89,$7D,$23,$E6,$BE,$58,$E5,$97,$E5
	.dc.b	$9E,$F9,$75,$BE,$F1,$D0,$EE,$79,$CE,$38,$7E,$6B,$1F,$C0,$DF,$D7
	.dc.b	$7F,$07,$5F,$52,$B8,$9E,$43,$8D,$FE,$5F,$0F,$34,$7A,$67,$C3,$5F
	.dc.b	$9F,$BF,$34,$FE,$91,$ED,$8F,$C1,$9D,$63,$E3,$8F,$CF,$CE,$AB,$D0
	.dc.b	$7A,$DE,$75,$E8,$AF,$9D,$87,$C5,$AD,$47,$37,$F3,$E7,$C3,$C3,$CB
	.dc.b	$87,$F0,$2C,$7D,$53,$EC,$3F,$2E,$5C,$3E,$1E,$1E,$E8,$F8,$77,$C9
	.dc.b	$FB,$18,$F9,$30,$F4,$E3,$E4,$9F,$AA,$78,$6F,$CD,$87,$C0,$C7,$C3
	.dc.b	$3E,$86,$C6,$63,$C6,$64,$AE,$4F,$2B,$E7,$BF,$98,$7B,$5D,$B6,$37
	.dc.b	$1C,$DA,$73,$8E,$31,$EC,$BF,$94,$FF,$05,$3D,$47,$B2,$7F,$40,$FD
	.dc.b	$D8,$F8,$0D,$AF,$CA,$5F,$C0,$BF,$11,$DA,$7F,$1B,$87,$C9,$7E,$BE
	.dc.b	$1D,$0F,$BD,$7C,$A3,$C1,$5F,$A4,$69,$F9,$1B,$EC,$FA,$67,$E5,$CF
	.dc.b	$82,$3F,$17,$5E,$F6,$C6,$5F,$CF,$1F,$15,$F3,$5F,$E4,$E3,$33,$C5
	.dc.b	$B3,$C7,$1F,$C2,$47,$B3,$1F,$93,$1F,$AB,$BF,$74,$FC,$BB,$F7,$5F
	.dc.b	$E2,$30,$F7,$4F,$C9,$3E,$79,$67,$6F,$C6,$9F,$84,$E5,$B1,$F2,$23
	.dc.b	$8B,$4E,$2F,$86,$36,$78,$5D,$B6,$37,$7E,$9C,$3D,$78,$FE,$8F,$1F
	.dc.b	$5C,$E7,$1F,$9F,$8F,$D1,$3F,$22,$5C,$3F,$1F,$8F,$C7,$C3,$EB,$CF
	.dc.b	$C7,$AE,$2D,$38,$96,$B8,$BC,$EB,$EA,$E3,$F9,$09,$F7,$47,$B2,$7E
	.dc.b	$31,$F8,$53,$F0,$E7,$EE,$AF,$C3,$8F,$DD,$C7,$EC,$B6,$33,$87,$5B
	.dc.b	$F8,$87,$5B,$F9,$47,$53,$E9,$17,$0F,$1B,$63,$8B,$4F,$29,$E3,$3C
	.dc.b	$B7,$F3,$2F,$E7,$5F,$D0,$9F,$DF,$1F,$98,$3E,$11,$4F,$8D,$5F,$13
	.dc.b	$1F,$C5,$E1,$E2,$9E,$C7,$98,$AE,$70,$F0,$47,$63,$F3,$A7,$62,$9D
	.dc.b	$8F,$03,$F1,$DF,$4D,$F9,$DF,$D5,$1A,$7E,$53,$F0,$23,$BB,$A3,$AB
	.dc.b	$CD,$F3,$3A,$9F,$D8,$63,$F5,$AB,$9D,$7D,$23,$D7,$87,$CE,$69,$DE
	.dc.b	$3C,$95,$CB,$1F,$BB,$8F,$C6,$BE,$18,$F7,$7F,$2D,$87,$B2,$7F,$2B
	.dc.b	$3E,$5B,$F9,$18,$74,$BD,$DF,$A2,$30,$FC,$9C,$7D,$18,$77,$8F,$23
	.dc.b	$D2,$7D,$73,$FA,$B9,$FD,$0E,$1D,$EF,$CB,$5D,$27,$E9,$1F,$21,$E3
	.dc.b	$8F,$E2,$30,$E8,$77,$9F,$8D,$7F,$EA,$EC,$BF,$C4,$78,$FF,$AD,$1F
	.dc.b	$F1,$44,$FF,$97,$1C,$DA,$24,$69,$FF,$15,$63,$FE,$2D,$23,$C5,$FD
	.dc.b	$35,$7A,$F2,$FD,$CC,$FA,$23,$E1,$BF,$C4,$0F,$86,$7C,$0C,$BD,$6F
	.dc.b	$F7,$53,$C4,$7F,$9C,$0A,$7F,$A4,$8B,$8F,$FB,$29,$B1,$FE,$FF,$6E
	.dc.b	$5C,$1E,$47,$FA,$F8,$C9,$99,$E4,$E4,$87,$FB,$8B,$DA,$7E,$01,$CE
	.dc.b	$3F,$DC,$51,$87,$93,$5F,$2B,$5F,$5C,$A3,$11,$D6,$7A,$4E,$5B,$3A
	.dc.b	$9E,$F1,$82,$3E,$07,$E2,$FE,$0E,$20

;--------------------------------------------------------------------------------
;	FSMUL.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsmul_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FSMUL.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSMUL.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsmul_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsmul.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSMUL.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsmul.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSMUL.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsmul_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3A,$F9,$AF,$EC,$ED,$2D,$FC,$E8,$CC,$76,$B4,$62
	.dc.b	$75,$2F,$A0,$F5,$BF,$03,$D8,$FE,$E9,$18,$86,$BE,$98,$C5,$72,$6E
	.dc.b	$1D,$AB,$A6,$9F,$33,$0E,$6D,$8F,$F0,$0D,$70,$C6,$CF,$C6,$BE,$5F
	.dc.b	$3F,$E1,$CB,$E4,$37,$1F,$F0,$43,$C7,$13,$71,$D4,$74,$3D,$85,$3E
	.dc.b	$26,$37,$2F,$56,$5E,$59,$F4,$9C,$0A,$70,$2C,$70,$75,$5E,$07,$39
	.dc.b	$7C,$38,$F2,$57,$57,$46,$27,$ED,$38,$8D,$1F,$0E,$FC,$D5,$CE,$33
	.dc.b	$5E,$07,$4F,$82,$7F,$C5,$B8,$78,$C9,$3A,$7C,$13,$F7,$31,$FF,$15
	.dc.b	$FC,$17,$C1,$BF,$D5,$65,$ED,$F9,$C7,$E2,$DF,$E3,$69,$FE,$49,$18
	.dc.b	$C3,$88,$70,$FF,$8F,$A3,$F9,$57,$CB,$3F,$45,$7E,$99,$C2,$F7,$7F
	.dc.b	$E4,$26,$BF,$E3,$BC,$3F,$36,$F5,$6C,$70,$C3,$AF,$E5,$3F,$B5,$6F
	.dc.b	$C9,$75,$AF,$46,$9F,$E5,$0F,$A0,$FE,$37,$C3,$E8,$6E,$5E,$79,$D7
	.dc.b	$95,$FC,$62,$30,$3E,$8D,$FB,$5B,$E3,$B9,$56,$EB,$FC,$B7,$87,$46
	.dc.b	$C6,$5F,$F2,$5B,$5E,$F8,$FC,$AE,$1E,$9E,$05,$C3,$F8,$D5,$78,$B0
	.dc.b	$E9,$B7,$F5,$51,$FB,$7F,$F9,$3E,$7C,$F1,$8E,$FF,$AA,$D8,$F5,$E2
	.dc.b	$E1,$7C,$0E,$B0,$E5,$5F,$7D,$AF,$F9,$C1,$AF,$DD,$6B,$FC,$39,$E3
	.dc.b	$7E,$55,$F1,$3D,$4E,$D8,$FE,$22,$BF,$9B,$F8,$37,$C4,$34,$D7,$DF
	.dc.b	$7F,$53,$C7,$F9,$F4,$EA,$BF,$8F,$B5,$FE,$4A,$D7,$FD,$1A,$D7,$FA
	.dc.b	$01,$F7,$04,$BE,$57,$B9,$E1,$8D,$4B,$DF,$F3,$12,$3E,$5F,$E5,$CF
	.dc.b	$7F,$11,$E2,$9F,$CC,$D7,$18,$F3,$C6,$0F,$D4,$FD,$07,$E9,$3A,$6F
	.dc.b	$6F,$F5,$BD,$3D,$F5,$FA,$EE,$A7,$ED,$B9,$BF,$F5,$5C,$61,$D4,$67
	.dc.b	$1F,$CC,$3F,$E1,$9B,$FE,$CD,$E9,$FF,$81,$27,$FA,$3B,$5F,$E9,$AD
	.dc.b	$7F,$D6,$AD,$7F,$D7,$3C,$DF,$A1,$F2,$FF,$2C,$48,$FC,$B7,$E4,$8F
	.dc.b	$77,$F3,$5C,$D2,$7F,$23,$D7,$F8,$46,$FC,$7F,$E0,$F2,$7C,$97,$F0
	.dc.b	$06,$CF,$9A,$30,$7F,$5E,$CF,$F6,$5D,$3F,$A7,$24,$F6,$E9,$FC,$77
	.dc.b	$DB,$FF,$87,$E3,$FB,$1F,$0F,$FB,$46,$31,$C1,$FE,$09,$6C,$7F,$B5
	.dc.b	$7C,$0F,$F1,$F5,$70,$3F,$DB,$33,$CD,$FF,$6C,$B5,$E9,$E7,$7F,$18
	.dc.b	$8C,$47,$08,$FD,$04,$FF,$B3,$A3,$F2,$4D,$7B,$09,$F1,$BF,$6B,$E5
	.dc.b	$3F,$83,$7F,$00,$7F,$A6,$63,$F7,$6D,$7C,$DC,$DF,$86,$9F,$C1,$9C
	.dc.b	$3E,$82,$4F,$F1,$C9,$D3,$CB,$E1,$FA,$91,$D7,$43,$E2,$31,$1F,$0F
	.dc.b	$85,$97,$F9,$66,$39,$1C,$27,$F7,$DF,$6F,$FB,$9D,$FF,$E0,$03,$64
	.dc.b	$1C,$2F,$96,$FF,$15,$BF,$12,$3F,$CA,$8F,$03,$8B,$87,$AF,$CF,$FF
	.dc.b	$48,$8C,$46,$A3,$F8,$84,$FF,$8C,$F8,$BF,$C9,$0D,$BE,$35,$F0,$39
	.dc.b	$7E,$7F,$A3,$23,$04,$3E,$16,$7C,$56,$9F,$12,$7F,$67,$5F,$B4,$F7
	.dc.b	$FE,$CF,$B3,$2F,$E6,$C6,$5F,$B7,$19,$7E,$DE,$FF,$92,$D3,$FC,$BD
	.dc.b	$F1,$DF,$C2,$E7,$F0,$ED,$8C,$F8,$3F,$CB,$7F,$1B,$FC,$24,$62,$BF
	.dc.b	$C8,$B5,$C5,$F9,$47,$FA,$72,$BF,$B7,$C6,$63,$FD,$36,$7F,$27,$AF
	.dc.b	$BC,$FE,$59,$B5,$D6,$3E,$1F,$4F,$E5,$88,$FE,$2B,$E2,$6B,$A3,$F3
	.dc.b	$31,$FC,$62,$3F,$B1,$7E,$9A,$F9,$FB,$F2,$B2,$FF,$AC,$CE,$5B,$39
	.dc.b	$7F,$29,$09,$E7,$3F,$D3,$FE,$5F,$FD,$21,$5F,$AE,$6E,$3F,$E8,$96
	.dc.b	$F2,$7E,$37,$E5,$FF,$D4,$15,$FD,$1E,$78,$BF,$21,$AF,$FB,$3F,$E8
	.dc.b	$3F,$C2,$F1,$88,$FF,$6A,$47,$F5,$F9,$7C,$C5,$3C,$91,$FE,$01,$FD
	.dc.b	$47,$FD,$39,$5F,$DD,$9F,$1B,$F8,$C6,$F3,$9D,$C7,$EC,$DB,$97,$F8
	.dc.b	$C1,$D9,$3E,$1B,$FA,$2F,$CE,$FF,$0A,$F3,$A5,$FE,$FA,$32,$FF,$0C
	.dc.b	$8C,$BF,$92,$8F,$EF,$1F,$3F,$FC,$DB,$C3,$FD,$CF,$D5,$7F,$A8,$A9
	.dc.b	$39,$51,$98,$C3,$BF,$9F,$7C,$C7,$EE,$E3,$5D,$DF,$C7,$5B,$87,$6F
	.dc.b	$99,$FE,$01,$87,$27,$F8,$16,$1C,$B1,$D1,$FC,$32,$74,$DC,$3A,$57
	.dc.b	$26,$C7,$F1,$2A,$E4,$FA,$27,$57,$D2,$3D,$B5,$C1,$B1,$9A,$E3,$87
	.dc.b	$A2,$79,$1E,$19,$E5,$87,$92,$15,$C3,$0E,$F1,$C0,$EA,$59,$E1,$3B
	.dc.b	$8E,$86,$70,$F2,$4E,$B1,$F9,$97,$D3,$1F,$21,$1F,$93,$1A,$BE,$73
	.dc.b	$96,$CE,$9F,$4D,$F6,$26,$3F,$06,$F9,$4F,$A2,$F9,$36,$F9,$3E,$7A
	.dc.b	$E6,$47,$E0,$63,$ED,$78,$1E,$22,$1F,$22,$FB,$E3,$F2,$C9,$8F,$D5
	.dc.b	$AE,$51,$8B,$ED,$8F,$E4,$6F,$8D,$FA,$B0,$F2,$1E,$28,$CD,$70,$C7
	.dc.b	$E5,$1E,$AB,$F6,$E1,$D6,$F8,$1B,$3C,$EF,$D3,$70,$7B,$09,$8F,$9B
	.dc.b	$1F,$15,$7C,$88,$C6,$1C,$2F,$95,$F9,$AB,$C3,$7E,$FC,$7C,$F3,$9C
	.dc.b	$7C,$EF,$E7,$63,$2F,$F4,$0B,$FD,$B3,$6B,$FC,$14,$FE,$48,$C6,$1F
	.dc.b	$D1,$9F,$C7,$98,$31,$3C,$9E,$E7,$B5,$F3,$E1,$E6,$9F,$83,$8F,$B2
	.dc.b	$BC,$C7,$DC,$71,$3B,$8F,$96,$68,$E9,$7E,$F7,$E2,$3F,$23,$1F,$A8
	.dc.b	$78,$B0,$FA,$95,$F7,$EB,$78,$78,$B1,$F8,$F3,$DC,$D4,$F7,$AE,$37
	.dc.b	$D2,$3E,$DE,$1D,$DF,$98,$74,$3C,$D8,$FA,$A7,$C5,$3E,$79,$E8,$F1
	.dc.b	$7F,$34,$71,$BF,$C0,$61,$E4,$AE,$38,$FE,$B6,$B5,$1F,$3B,$0E,$D1
	.dc.b	$E4,$9D,$3D,$9F,$73,$C0,$F9,$D5,$D5,$E0,$7C,$EB,$F8,$57,$D6,$75
	.dc.b	$3F,$42,$BA,$D6,$A3,$99,$FC,$C3,$C7,$FF,$4A,$E7,$FD,$23,$7F,$F8
	.dc.b	$03,$E1,$7F,$C0,$0A,$FE,$9C,$43,$FC,$53,$B6,$F9,$BF,$D3,$7D,$FF
	.dc.b	$E6,$3B,$FE,$B5,$C3,$FD,$97,$92,$61,$FD,$96,$3F,$B4,$18,$7F,$B1
	.dc.b	$46,$1F,$EF,$0F,$F6,$F7,$FD,$81,$CD,$FD,$C9,$C9,$0F,$CF,$73,$FF
	.dc.b	$79,$4D,$24,$FF,$78,$8E,$16,$C7,$FC,$11,$F2,$BF,$E1,$09,$FF,$0B
	.dc.b	$F5,$7F,$85,$6B,$51,$FE,$46,$EB,$FF,$0D,$3F,$E1,$8E,$BD,$61,$FE
	.dc.b	$AE,$9F,$F0,$D4,$FF,$87,$8F,$F1,$39,$0F,$F1,$51,$FE,$46,$EC,$FF
	.dc.b	$16,$DF,$D0,$ED,$B7,$F3,$27,$FD,$E1,$DF,$FE,$47,$8F,$F3,$57,$7F
	.dc.b	$F8,$E9,$BE,$46,$BC,$19,$97,$F9,$21,$FF,$22,$91,$C1,$FE,$EB,$BE
	.dc.b	$8F,$F9,$F1,$AA,$FF,$CA,$77,$F7,$BC,$3F,$F2,$F6,$7F,$86,$F0,$FF
	.dc.b	$CC,$57,$FE,$60,$9C,$D7,$FB,$43,$9D,$5A,$BF,$97,$D3,$FE,$E4,$EF
	.dc.b	$FF,$3B,$63,$FE,$74,$F1,$3F,$CE,$2F,$F9,$DE,$BF,$27,$1A,$78,$F4
	.dc.b	$FF,$B3,$A7,$FD,$1D,$E3,$7F,$A1,$AA,$47,$FA,$5A,$BE,$CB,$FE,$97
	.dc.b	$9F,$F4,$E9,$FE,$8F,$23,$FC,$9C,$D1,$7C,$7F,$F4,$E9,$23,$FD,$27
	.dc.b	$F2,$3F,$EA,$CA,$FD,$3F,$15,$DB,$F4,$7E,$4D,$C7,$FD,$1B,$BF,$D7
	.dc.b	$6D,$7F,$C0,$09,$FF,$5F,$C7,$FB,$13,$C1,$FF,$5F,$BF,$E8,$AF,$2B
	.dc.b	$2F,$FA,$62,$33,$8F,$C2,$9F,$F7,$3F,$3F,$FB,$5B,$C1,$97,$F0,$73
	.dc.b	$FF,$6E,$B4,$FF,$6E,$1E,$FF,$2F,$FD,$BE,$FF,$B6,$6B,$FD,$C4,$47
	.dc.b	$81,$92,$C7,$FB,$79,$79,$9F,$E9,$FA,$FF,$62,$F9,$5F,$EF,$E9,$F0
	.dc.b	$35,$FF,$69,$7A,$2F,$3F,$FD,$89,$A7,$EB,$A9,$2F,$3F,$38,$7F,$B2
	.dc.b	$7D,$0F,$C0,$63,$FE,$A2,$6B,$9E,$8F,$B4,$A3,$2F,$FA,$FA,$DE,$97
	.dc.b	$48,$C3,$71,$F0,$B6,$5E,$95,$F5,$65,$FB,$25,$1C,$FD,$F7,$AC,$F5
	.dc.b	$BB,$5F,$8F,$C5,$F2,$E8,$69,$23,$FB,$C7,$80,$F0,$3F,$33,$C5,$FA
	.dc.b	$0E,$1F,$44,$E5,$B2,$FA,$0E,$FE,$CF,$EF,$3D,$97,$A3,$CA,$FF,$DF
	.dc.b	$31,$87,$7E,$4C,$CF,$F4,$57,$FE,$CB,$7F,$DF,$ED,$4D,$3F,$83,$92
	.dc.b	$3E,$57,$B6,$F6,$FC,$7D,$3F,$0B,$DD,$7B,$B4,$CE,$3F,$8E,$C7,$F0
	.dc.b	$0F,$8A,$BE,$DC,$62,$3F,$C3,$8D,$7F,$DC,$6E,$78,$3F,$7E,$DC,$D3
	.dc.b	$EA,$AF,$17,$8F,$F8,$1B,$7B,$DF,$93,$AF,$D3,$46,$3A,$F5,$3F,$88
	.dc.b	$9C,$BB,$53,$FD,$18,$E7,$DD,$FC,$2D,$B1,$FE,$33,$6B,$FE,$FA,$6B
	.dc.b	$FA,$29,$E9,$1C,$DB,$8F,$F1,$B6,$CB,$B5,$F0,$0F,$80,$FC,$97,$C1
	.dc.b	$7E,$D2,$BF,$2D,$F0,$5F,$9B,$33,$E1,$C8,$FF,$23,$77,$AB,$84,$7C
	.dc.b	$83,$E4,$C6,$67,$F9,$33,$63,$8C,$F0,$D3,$FB,$3C,$E6,$7E,$2E,$9C
	.dc.b	$AB,$4F,$D4,$EB,$7C,$1F,$F0,$0A,$FE,$2D,$F0,$CF,$85,$F8,$D7,$FC
	.dc.b	$33,$67,$C3,$7E,$DD,$B7,$C2,$B6,$AF,$93,$FC,$F5,$AF,$F9,$95,$AF
	.dc.b	$BC,$E1,$1F,$C2,$A3,$3A,$7F,$80,$9B,$87,$E3,$1B,$5F,$4F,$E1,$FF
	.dc.b	$C5,$0F,$C0,$3F,$23,$F1,$1F,$C9,$AF,$F2,$7F,$16,$F8,$BF,$9D,$F1
	.dc.b	$8F,$5F,$F8,$3B,$70,$FE,$63,$87,$15,$7C,$5E,$BC,$7F,$92,$5F,$19
	.dc.b	$95,$7C,$DC,$E3,$E0,$C7,$EF,$56,$F1,$EC,$DC,$3A,$63,$CF,$1E,$4D
	.dc.b	$BF,$C3,$3B,$CB,$DB,$8F,$E2,$71,$F4,$4E,$F2,$C9,$FB,$4C,$39,$D6
	.dc.b	$F1,$EE,$DC,$37,$5E,$AC,$7B,$CF,$A6,$33,$8F,$68,$E1,$1F,$19,$D3
	.dc.b	$4F,$04,$F2,$7E,$4D,$72,$BF,$0C,$68,$F0,$E1,$E6,$76,$73,$18,$78
	.dc.b	$5E,$06,$63,$83,$A9,$EA,$DA,$FE,$2B,$97,$C5,$8F,$AA,$72,$7E,$63
	.dc.b	$83,$81,$D7,$0F,$53,$F1,$4F,$54,$C7,$47,$AA,$37,$87,$9E,$39,$9F
	.dc.b	$A1,$7F,$1C,$DB,$FC,$63,$6B,$FC,$4D,$2F,$41,$F5,$D8,$6D,$7F,$B4
	.dc.b	$67,$53,$93,$8D,$F4,$9F,$AE,$75,$38,$39,$7E,$79,$92,$1C,$08,$68
	.dc.b	$A7,$B0,$B3,$EC,$6C,$75,$74,$71,$21,$82,$47,$17,$2D,$73,$1C,$E3
	.dc.b	$04,$70,$4B,$F8,$24,$9F,$7B,$5F,$69,$DA,$38,$3A,$21,$A3,$C3,$1C
	.dc.b	$5F,$6B,$5E,$4D,$76,$72,$36,$53,$65,$36,$77,$9F,$64,$73,$6C,$6D
	.dc.b	$AE,$E7,$D2,$D9,$E1,$18,$8F,$09,$93,$93,$97,$C6,$64,$87,$02,$1A
	.dc.b	$29,$E8,$2C,$F9,$1B,$1D,$5D,$1C,$48,$60,$91,$C5,$CB,$5C,$C7,$38
	.dc.b	$C1,$1C,$12,$7E,$09,$27,$DE,$D7,$DA,$76,$8E,$0E,$88,$68,$F0,$C7
	.dc.b	$17,$DA,$D7,$93,$5D,$9C,$8D,$94,$D9,$4D,$9D,$E7,$D9,$1C,$DB,$1B
	.dc.b	$6B,$B9,$F4,$B6,$78,$46,$23,$C2,$64,$E4,$E5,$F1,$99,$21,$C0,$86
	.dc.b	$8A,$7A,$0B,$3E,$46,$C7,$57,$47,$12,$18,$24,$71,$72,$D7,$31,$CE
	.dc.b	$30,$47,$04,$9F,$82,$49,$F7,$B5,$F6,$9D,$A3,$83,$A2,$1F,$B3,$BF
	.dc.b	$E5,$FE,$9B,$C5,$FE,$E1,$18,$6E,$3F,$E6,$29,$D5,$7F,$85,$2B,$4D
	.dc.b	$9D,$57,$08,$57,$07,$FA,$FD,$70,$6D,$70,$7F,$C5,$5F,$0D,$FE,$B2
	.dc.b	$4D,$23,$FE,$B6,$6B,$FE,$F1,$6B,$FE,$33,$38,$9B,$21,$B2,$9B,$78
	.dc.b	$BE,$23,$25,$38,$10,$E0,$68,$FF,$30,$3C,$9A,$64,$FF,$45,$B9,$69
	.dc.b	$93,$CE,$75,$3C,$07,$63,$07,$42,$9C,$8A,$72,$38,$10,$E6,$43,$91
	.dc.b	$92,$99,$3A,$19,$29,$93,$B1,$5F,$63,$83,$99,$0E,$25,$34,$60,$86
	.dc.b	$08,$F4,$23,$F1,$DA,$FC,$06,$BE,$73,$89,$B2,$1B,$29,$B7,$8B,$EB
	.dc.b	$32,$53,$81,$0E,$06,$8F,$63,$C9,$A6,$4F,$6B,$96,$99,$3C,$E7,$53
	.dc.b	$C0,$76,$30,$74,$29,$C8,$A7,$23,$81,$0E,$64,$39,$19,$29,$93,$A1
	.dc.b	$92,$99,$3B,$15,$F6,$38,$39,$90,$E2,$53,$46,$08,$60,$8F,$42,$3F
	.dc.b	$1D,$AF,$C0,$6B,$E7,$38,$9B,$21,$B2,$9B,$78,$BE,$B3,$25,$38,$10
	.dc.b	$E0,$68,$F6,$3C,$9A,$64,$F6,$B9,$69,$93,$CE,$75,$3C,$07,$63,$07
	.dc.b	$42,$9C,$8A,$72,$38,$10,$E6,$43,$91,$92,$99,$3A,$1F,$C1,$72,$FF
	.dc.b	$39,$BF,$E7,$5A,$FF,$3B,$BF,$E8,$16,$BF,$D0,$4F,$EA,$67,$F9,$23
	.dc.b	$7F,$F4,$B1,$37,$FF,$57,$BF,$E9,$63,$11,$FE,$A2,$71,$1A,$3F,$D6
	.dc.b	$6E,$A3,$FD,$36,$E9,$E1,$1A,$38,$3F,$EB,$08,$E1,$BF,$FA,$B2,$FF
	.dc.b	$D3,$47,$FA,$6D,$FF,$57,$99,$D3,$FD,$5F,$19,$DB,$FD,$4F,$39,$DB
	.dc.b	$FD,$4D,$7D,$27,$FD,$4B,$7C,$0F,$F5,$BD,$74,$8C,$D7,$0C,$7C,$D8
	.dc.b	$FA,$0F,$33,$E8,$7F,$D6,$87,$FA,$C8,$FF,$64,$63,$E8,$C3,$AD,$F9
	.dc.b	$0E,$B3,$A8,$CB,$FE,$B9,$3C,$07,$63,$C0,$76,$C7,$E5,$3F,$32,$FC
	.dc.b	$AE,$5F,$25,$FC,$38,$E8,$77,$C3,$AB,$B3,$AC,$79,$E3,$75,$F6,$63
	.dc.b	$13,$ED,$8C,$5F,$C9,$AD,$DF,$8A,$F7,$5F,$3E,$FB,$C7,$8D,$E5,$3E
	.dc.b	$6B,$F8,$31,$83,$E0,$1E,$D3,$E0,$E3,$F6,$DF,$51,$B3,$D4,$6C,$8F
	.dc.b	$DB,$7C,$31,$DA,$FD,$38,$7C,$2A,$D5,$74,$9F,$9B,$18,$7E,$7E,$1E
	.dc.b	$97,$8C,$EF,$0F,$A3,$5E,$59,$E3,$8F,$DF,$AF,$0E,$1F,$19,$F9,$F8
	.dc.b	$FC,$97,$D5,$7C,$70,$FC,$63,$E1,$7E,$35,$71,$9F,$7E,$3F,$3A,$7C
	.dc.b	$38,$7E,$33,$1F,$0D,$6E,$B4,$FC,$A9,$D5,$76,$7C,$D1,$EB,$3E,$8E
	.dc.b	$1E,$17,$E9,$D7,$C1,$AF,$BD,$8F,$A7,$0F,$BD,$5C,$4F,$A3,$87,$48
	.dc.b	$F9,$33,$F4,$6B,$8E,$3F,$46,$B5,$1E,$EB,$F8,$71,$E6,$7C,$6E,$9F
	.dc.b	$1D,$73,$7E,$5D,$73,$7E,$04,$79,$6B,$E2,$C7,$D2,$3A,$96,$FD,$27
	.dc.b	$9A,$BE,$76,$3F,$4E,$7C,$27,$48,$EC,$FE,$47,$1F,$9A,$FE,$4A,$78
	.dc.b	$3F,$08,$E0,$7C,$57,$E6,$3F,$97,$8F,$98,$FA,$8D,$63,$EF,$75,$8F
	.dc.b	$BD,$FC,$0B,$ED,$7E,$E1,$EB,$C7,$DE,$FD,$3A,$EA,$F0,$C3,$D3,$8F
	.dc.b	$E6,$8F,$C3,$C7,$AE,$FA,$3E,$BC,$3E,$93,$EB,$35,$5C,$B0,$FB,$71
	.dc.b	$E8,$3D,$F7,$D0,$ED,$08,$9E,$9B,$3C,$42,$DC,$78,$AD,$4F,$19,$CB
	.dc.b	$CA,$B8,$C6,$21,$93,$DC,$CE,$1C,$EE,$D7,$8E,$9E,$AE,$23,$8B,$71
	.dc.b	$F0,$C7,$96,$F9,$3E,$37,$D3,$5C,$A3,$07,$5A,$E5,$7D,$7D,$77,$CD
	.dc.b	$D2,$57,$01,$F1,$CF,$94,$CC,$78,$4F,$95,$0F,$53,$D0,$D8,$E5,$7D
	.dc.b	$9B,$7A,$A9,$18,$21,$D4,$F0,$BA,$71,$3C,$DD,$7C,$0F,$FB,$1F,$7F
	.dc.b	$A0,$DD,$7E,$DB,$67,$F0,$1F,$02,$F7,$3E,$74,$E6,$52,$F7,$7E,$93
	.dc.b	$63,$E2,$4E,$CE,$11,$98,$F4,$37,$0F,$1F,$5F,$DB,$FA,$17,$C3,$FE
	.dc.b	$03,$B3,$BF,$C2,$DC,$0E,$E7,$87,$C5,$67,$E2,$A6,$7F,$57,$14,$FE
	.dc.b	$1F,$E3,$9F,$07,$ED,$AF,$F6,$E4,$63,$83,$FD,$BD,$08,$C3,$A9,$E0
	.dc.b	$FE,$88,$D7,$13,$DA,$E3,$48,$C4,$F6,$6B,$C6,$35,$3E,$0F,$B9,$7A
	.dc.b	$DF,$A8,$E2,$FD,$4F,$0B,$F1,$9F,$5F,$37,$E4,$6F,$6F,$E6,$A7,$5B
	.dc.b	$71,$8F,$DB,$CB,$D5,$C6,$8B,$E9,$F8,$6E,$21,$5B,$BE,$98,$AB,$5F
	.dc.b	$56,$FA,$3E,$7A,$79,$73,$CC,$62,$F9,$65,$F9,$7B,$E1,$8F,$EC,$30
	.dc.b	$E0,$FE,$AE,$33,$3F,$A9,$9C,$BC,$6F,$AD,$7E,$E3,$CD,$7C,$F7,$E4
	.dc.b	$E3,$94,$FA,$FE,$B5,$F4,$1F,$02,$DF,$83,$3E,$3B,$53,$98,$C6,$1F
	.dc.b	$BD,$BE,$2F,$78,$CE,$5F,$B1,$9C,$BF,$BF,$C3,$AE,$1D,$2F,$83,$67
	.dc.b	$57,$CB,$3F,$E4,$58,$72,$6E,$3F,$5B,$EC,$5F,$85,$FF,$20,$CB,$C9
	.dc.b	$3F,$5F,$EE,$DF,$49,$F7,$3D,$D7,$D0,$BE,$8B,$53,$98,$C6,$1F,$C6
	.dc.b	$6F,$8B,$FC,$8A,$F8,$DF,$08,$C5,$70,$DB,$F8,$85,$F0,$6E,$1F,$8C
	.dc.b	$EB,$7E,$BC,$F8,$DE,$55,$FC,$5C,$FE,$99,$EC,$7F,$4D,$F5,$5E,$AF
	.dc.b	$F4,$CE,$67,$BD,$FD,$22,$73,$7F,$AF,$BE,$17,$BA,$D3,$FB,$39,$4E
	.dc.b	$9F,$EB,$57,$D3,$6C,$47,$F5,$A9,$B3,$98,$61,$E0,$74,$62,$FC,$17
	.dc.b	$CA,$3D,$66,$4E,$F8,$79,$1F,$41,$91,$7D,$EF,$B4,$72,$7C,$11,$9B
	.dc.b	$F6,$61,$E6,$BE,$97,$E4,$BD,$C6,$23,$95,$79,$9F,$01,$83,$CB,$87
	.dc.b	$A5,$F8,$E7,$97,$0F,$03,$E8,$BE,$B3,$BC,$3C,$4D,$9E,$18,$78,$5E
	.dc.b	$75,$C1,$F0,$98,$C3,$C9,$1F,$02,$30,$79,$0D,$DF,$66,$9D,$AF,$C1
	.dc.b	$3D,$4E,$53,$B6,$CE,$CE,$2E,$EB,$E5,$C6,$63,$66,$5F,$81,$1E,$63
	.dc.b	$71,$98,$FA,$11,$97,$67,$A6,$39,$C6,$63,$66,$5E,$66,$5F,$A8,$7E
	.dc.b	$12,$FE,$55,$7A,$E7,$73,$C4,$EE,$6E,$FE,$1C,$73,$8F,$96,$64,$F8
	.dc.b	$2E,$67,$A4,$72,$C3,$EF,$BE,$63,$86,$1E,$A7,$C0,$66,$36,$FA,$1A
	.dc.b	$F8,$1B,$3D,$A3,$86,$1E,$C7,$E8,$5F,$B1,$EE,$DA,$E3,$1A,$3C,$D3
	.dc.b	$C6,$FE,$94,$78,$63,$15,$ED,$8E,$07,$83,$0F,$8A,$FA,$8E,$0E,$1F
	.dc.b	$09,$D9,$C5,$F8,$AF,$BC,$FB,$6F,$BB,$CC,$C5,$70,$8C,$C7,$43,$C4
	.dc.b	$7D,$23,$86,$1F,$15,$F1,$99,$38,$18,$AF,$A4,$E0,$EE,$6E,$FC,$D8
	.dc.b	$7C,$19,$F3,$D7,$5A,$DC,$72,$AD,$C7,$88,$F8,$E7,$7A,$FA,$57,$EA
	.dc.b	$9D,$CF,$86,$FA,$11,$E2,$76,$AE,$38,$7C,$32,$47,$9E,$FA,$12,$33
	.dc.b	$87,$96,$31,$3B,$9C,$D7,$48,$C1,$2B,$84,$EA,$3E,$D1,$99,$E1,$3A
	.dc.b	$8C,$1A,$70,$48,$D9,$23,$4D,$7E,$DC,$6D,$C9,$0C,$C6,$DF,$98,$64
	.dc.b	$A6,$8C,$1F,$24,$B7,$F3,$67,$ED,$C7,$CE,$8C,$4F,$DF,$24,$7D,$B9
	.dc.b	$DB,$FE,$5A,$9F,$F2,$DB,$FD,$58,$C1,$FD,$9F,$D8,$FF,$5E,$78,$1F
	.dc.b	$E6,$06,$C6,$7D,$3F,$F2,$33,$67,$5E,$AF,$FB,$06,$39,$35,$FF,$32
	.dc.b	$57,$FA,$F3,$A3,$64,$8F,$F7,$D9,$FE,$DF,$9F,$F7,$37,$B3,$FE,$F2
	.dc.b	$AF,$F7,$0B,$63,$3E,$CF,$FB,$72,$72,$D8,$CD,$7F,$BA,$CF,$F7,$69
	.dc.b	$D6,$FF,$BF,$DF,$FB,$EA,$FF,$E0,$0B,$6B,$FD,$D2,$D9,$FF,$73,$57
	.dc.b	$FB,$F7,$3F,$F7,$26,$9F,$EE,$86,$C6,$63,$FD,$C7,$87,$92,$3C,$47
	.dc.b	$FB,$CC,$B7,$D5,$FF,$78,$1F,$EF,$42,$D7,$FB,$C1,$FF,$78,$4F,$37
	.dc.b	$D4,$7C,$88,$FF,$0A,$47,$AB,$1F,$F7,$9C,$71,$9F,$F8,$00,$D9,$E4
	.dc.b	$F0,$C7,$DF,$5C,$30,$F6,$1E,$A8,$F9,$77,$E0,$AF,$F1,$16,$1E,$97
	.dc.b	$E0,$C2,$3E,$B6,$3F,$3B,$0E,$93,$F1,$4F,$5D,$F4,$BF,$15,$7C,$E9
	.dc.b	$E5,$5F,$6E,$39,$44,$3C,$58,$1F,$BB,$5C,$9F,$93,$1E,$A7,$BC,$F1
	.dc.b	$9F,$85,$7E,$1A,$F7,$4F,$87,$0F,$4D,$FC,$D9,$CB,$D0,$F5,$1F,$31
	.dc.b	$EC,$D3,$38,$FC,$E8,$E3,$87,$C6,$AE,$AF,$CA,$C7,$E2,$DF,$C0,$3F
	.dc.b	$0A,$E2,$77,$5E,$F9,$FC,$51,$F0,$8F,$82,$F6,$23,$83,$E0,$9E,$93
	.dc.b	$B1,$C4,$E7,$5F,$24,$F3,$57,$D1,$8F,$84,$72,$8C,$BE,$63,$E6,$4F
	.dc.b	$81,$F8,$47,$68,$E9,$19,$AF,$C4,$36,$BD,$C7,$79,$F8,$95,$D0,$F0
	.dc.b	$4E,$DC,$0B,$FD,$03,$86,$BF,$AB,$9F,$3B,$EF,$BF,$0B,$D0,$F8,$C7
	.dc.b	$3C,$7F,$46,$F3,$6B,$CE,$FC,$4D,$39,$1F,$29,$F8,$57,$F8,$18,$CB
	.dc.b	$5E,$D8,$FE,$3F,$1F,$81,$5F,$77,$0F,$2D,$7C,$A5,$3F,$86,$BF,$E5
	.dc.b	$53,$FE,$89,$F6,$1A,$7E,$17,$16,$DF,$87,$CD,$EC,$66,$30,$DC,$3F
	.dc.b	$55,$87,$27,$F5,$97,$C9,$CE,$1B,$58,$74,$77,$87,$46,$9C,$0A,$2F
	.dc.b	$B2,$BE,$CF,$4C,$3B,$5F,$58,$D5,$FE,$86,$F8,$DF,$0F,$29,$EB,$79
	.dc.b	$7D,$77,$AF,$4F,$44,$EB,$1F,$1B,$5E,$A7,$F0,$B9,$E2,$FA,$1A,$E8
	.dc.b	$47,$1B,$F1,$5F,$79,$E1,$7E,$08,$91,$82,$1C,$EB,$96,$3F,$BD,$F5
	.dc.b	$9E,$B7,$CF,$6D,$7C,$A6,$BE,$3F,$59,$E9,$6E,$BE,$B4,$3D,$7A,$66
	.dc.b	$B9,$4E,$1F,$AB,$3B,$72,$7D,$8F,$51,$EB,$7E,$1F,$A5,$4F,$5F,$0D
	.dc.b	$BE,$15,$F8,$7B,$7A,$7A,$F5,$E5,$75,$8F,$C6,$7A,$CF,$5F,$D7,$3F
	.dc.b	$EF,$98,$C6,$DF,$EF,$B8,$46,$1D,$4F,$07,$F9,$31,$AE,$17,$89,$C6
	.dc.b	$BE,$CB,$63,$33,$DD,$AF,$2F,$59,$EB,$FE,$8F,$9F,$F4,$9C,$2F,$57
	.dc.b	$DB,$CD,$F9,$EB,$DB,$FA,$29,$D6,$DC,$63,$F9,$D4,$BD,$6C,$64,$BE
	.dc.b	$9E,$BE,$21,$5B,$BE,$98,$AB,$5E,$BB,$D6,$E7,$A7,$97,$3C,$C6,$2F
	.dc.b	$96,$5F,$9B,$BE,$18,$FE,$E3,$0E,$0F,$EE,$23,$33,$FA,$69,$CB,$C6
	.dc.b	$FA,$D7,$F0,$0F,$25,$EB,$FE,$46,$39,$4F,$AF,$D5,$7A,$7F,$02,$DE
	.dc.b	$B3,$D7,$D4,$E6,$31,$87,$F0,$AB,$E2,$F7,$8C,$E5,$FB,$59,$CB,$FB
	.dc.b	$CC,$3A,$E1,$D2,$F8,$36,$75,$7C,$B3,$FE,$27,$87,$27,$F9,$16,$1C
	.dc.b	$B0,$F1,$63,$E5,$C7,$EB,$7A,$4F,$4B,$ED,$FA,$AF,$51,$EA,$EA,$73
	.dc.b	$18,$C7,$F9,$2D,$F1,$7F,$94,$DF,$1B,$E1,$18,$AE,$1B,$7F,$13,$BE
	.dc.b	$0F,$F1,$AB,$E1,$5D,$30,$EF,$87,$E2,$EB,$FD,$1E,$43,$FD,$1C,$7F
	.dc.b	$A4,$48,$67,$D2,$FF,$46,$BF,$E8,$BF,$47,$FD,$1A,$FF,$B0,$4F,$F4
	.dc.b	$91,$3D,$2F,$F4,$A9,$23,$7B,$7F,$B2,$09,$E9,$FF,$AF,$27,$FD,$51
	.dc.b	$5E,$03,$C6,$78,$4E,$95,$CA,$78,$3F,$EA,$B8,$D9,$2F,$CE,$F9,$6B
	.dc.b	$C4,$FA,$48,$66,$7A,$65,$FE,$C7,$8F,$15,$72,$C3,$CF,$39,$9F,$25
	.dc.b	$F4,$79,$1D,$27,$6F,$FB,$30,$E6,$7A,$8D,$61,$EA,$7F,$D7,$A6,$4E
	.dc.b	$64,$8E,$73,$F1,$A3,$C3,$87,$A1,$EC,$60,$EB,$7E,$77,$E2,$1D,$6B
	.dc.b	$9C,$FC,$9C,$3D,$D8,$78,$EB,$CF,$87,$B6,$F7,$1E,$68,$C4,$EE,$3C
	.dc.b	$07,$9D,$C3,$E0,$8D,$E1,$E3,$BE,$ED,$9F,$25,$70,$9C,$C7,$B0,$E5
	.dc.b	$87,$BE,$FE,$14,$60,$F7,$9B,$9E,$31,$A6,$CF,$BE,$7B,$DF,$86,$BC
	.dc.b	$B1,$C4,$C9,$5C,$9C,$DC,$B4,$CC,$F5,$6C,$7C,$78,$F4,$19,$2B,$FD
	.dc.b	$5A,$BF,$DE,$69,$A4,$7F,$DD,$C4,$3F,$DD,$7E,$87,$FB,$C9,$BD,$7F
	.dc.b	$EE,$B7,$FD,$DF,$E7,$FF,$BC,$5F,$F4,$C9,$8E,$3F,$F7,$86,$3F,$E6
	.dc.b	$E8,$FF,$2D,$1F,$E4,$FF,$4B,$FD,$01,$ED,$E4,$99,$FF,$A8,$9F,$F3
	.dc.b	$5D,$7F,$9F,$2B,$FC,$C3,$EA,$7F,$96,$3D,$4F,$F2,$F1,$FE,$72,$9F
	.dc.b	$F5,$7E,$3F,$E7,$89,$FF,$56,$DF,$4B,$FF,$5D,$47,$F9,$79,$FF,$5E
	.dc.b	$47,$FA,$EC,$98,$7F,$99,$71,$FF,$51,$47,$19,$FF,$35,$87,$CB,$3F
	.dc.b	$E7,$2C,$BF,$D7,$D3,$FE,$72,$7F,$CA,$0E,$E3,$A3,$E0,$6B,$93,$FD
	.dc.b	$53,$3D,$1E,$58,$7A,$EF,$E3,$D7,$0A,$F2,$65,$FE,$76,$78,$5F,$9B
	.dc.b	$2B,$3F,$E7,$A7,$DB,$1F,$0A,$FD,$B7,$CC,$FA,$07,$FA,$42,$7E,$5C
	.dc.b	$FF,$AD,$0E,$91,$A7,$91,$EB,$29,$EB,$3E,$3D,$F8,$B1,$FF,$4E,$D7
	.dc.b	$48,$F9,$99,$7D,$3A,$E1,$8F,$AF,$0F,$11,$FE,$E0,$7E,$FE,$1F,$33
	.dc.b	$0F,$5D,$7F,$B3,$B0,$F8,$73,$F4,$1C,$E1,$D5,$CE,$5E,$7A,$F2,$E3
	.dc.b	$F0,$5F,$4C,$F4,$C7,$C5,$18,$9D,$3C,$0F,$AE,$E2,$FB,$3F,$16,$3F
	.dc.b	$3C,$06,$3F,$9B,$79,$1E,$98,$EB,$5E,$08,$E2,$7F,$B2,$F2,$FF,$6C
	.dc.b	$BF,$EE,$33,$F5,$71,$F1,$31,$FF,$5F,$46,$6F,$E7,$5F,$96,$FE,$FC
	.dc.b	$FA,$0F,$D5,$BD,$1E,$C7,$FB,$4E,$B9,$B8,$3B,$E5,$F9,$E7,$F0,$67
	.dc.b	$23,$81,$EE,$AF,$85,$8F,$C4,$8F,$94,$78,$4F,$6E,$1F,$0F,$1F,$54
	.dc.b	$71,$3A,$1E,$0C,$3F,$44,$FB,$6F,$F0,$98,$78,$EB,$F3,$D1,$E0,$C7
	.dc.b	$DF,$5F,$34,$F0,$4F,$6C,$3F,$29,$87,$B9,$F1,$61,$CC,$FC,$BE,$1D
	.dc.b	$A7,$F1,$18,$FA,$F0,$E8,$7D,$AC,$3A,$57,$E6,$23,$F4,$78,$74,$BE
	.dc.b	$38,$F9,$30,$E5,$87,$DF,$3F,$00,$F5,$33,$8F,$E6,$F1,$E9,$8A,$7F
	.dc.b	$8E,$C7,$DD,$8F,$CE,$57,$3A,$F5,$61,$ED,$AF,$E2,$CF,$CA,$AF,$9C
	.dc.b	$7D,$A3,$F2,$77,$E1,$C3,$E4,$E1,$CE,$B8,$63,$F9,$9F,$0F,$FD,$97
	.dc.b	$87,$E7,$5A,$FF,$B6,$3D,$0F,$F7,$37,$B3,$FE,$D4,$8F,$F5,$39,$93
	.dc.b	$FD,$5D,$3F,$EE,$57,$FD,$56,$D7,$F5,$06,$CF,$F7,$27,$A7,$FE,$E6
	.dc.b	$C3,$FD,$F6,$DE,$2D,$D7,$8B,$D2,$FF,$6C,$D7,$FC,$00,$6C,$FF,$B8
	.dc.b	$7D,$1F,$F6,$CF,$99,$FE,$D7,$C7,$CB,$5F,$EE,$07,$FD,$7A,$D7,$81
	.dc.b	$FE,$C5,$79,$BF,$BD,$F4,$3F,$DB,$8F,$FB,$9F,$D2,$E9,$8F,$87,$0F
	.dc.b	$F7,$14,$E4,$FF,$70,$9F,$EF,$FF,$9B,$7D,$57,$C0,$A6,$7F,$0A,$73
	.dc.b	$5F,$C2,$E7,$2F,$F0,$D8,$CA,$8C,$AF,$45,$F1,$3D,$6F,$A6,$1F,$ED
	.dc.b	$FA,$61,$C9,$FF,$6E,$B5,$D9,$FE,$E5,$7B,$E1,$DB,$07,$C5,$F5,$AF
	.dc.b	$E4,$11,$87,$D1,$87,$27,$CC,$FD,$EF,$53,$FD,$F3,$1F,$EF,$16,$BF
	.dc.b	$88,$6B,$FE,$F0,$6B,$C6,$7D,$51,$EF,$8E,$51,$83,$DD,$E9,$FF,$BE
	.dc.b	$E3,$FD,$F2,$D7,$F1,$ED,$7F,$9C,$35,$FF,$7F,$7A,$7F,$EF,$B2,$3F
	.dc.b	$EF,$B7,$F3,$AF,$FB,$AB,$E3,$7E,$8B,$FE,$F2,$6B,$FD,$0F,$D2,$46
	.dc.b	$23,$86,$3C,$A7,$FD,$FB,$1F,$A8,$6B,$FA,$43,$FD,$EF,$EA,$71,$8C
	.dc.b	$C7,$EA,$3E,$C5,$F5,$5F,$F0,$03,$17,$EB,$EF,$A1,$FF,$80,$15,$AC
	.dc.b	$BE,$AE,$0B,$DD,$6B,$0F,$35,$EE,$BF,$76,$D7,$C8,$D7,$C5,$1C,$6F
	.dc.b	$FD,$F1,$E9,$BE,$9B,$15,$FE,$FD,$9F,$0F,$17,$FC,$00,$3F,$33,$EA
	.dc.b	$BD,$5F,$CF,$C4,$F5,$9E,$BE,$2F,$F8,$3D,$ED,$E7,$3A,$7D,$B8,$27
	.dc.b	$4F,$E4,$8F,$D4,$4E,$DF,$21,$FC,$62,$36,$FD,$16,$BE,$0C,$3E,$D4
	.dc.b	$74,$C3,$D3,$1E,$59,$79,$CC,$DD,$99,$9F,$25,$5F,$59,$EB,$7F,$C0
	.dc.b	$1D,$3D,$31,$87,$EB,$E2,$F5,$DE,$BF,$1B,$FC,$65,$EF,$1F,$E6,$2D
	.dc.b	$7F,$14,$D7,$F9,$7B,$BD,$3F,$84,$B7,$1F,$F0,$D7,$C1,$BE,$0D,$F0
	.dc.b	$3D,$E7,$E6,$E1,$FE,$B6,$D1,$EA,$FF,$16,$F3,$1E,$87,$F4,$2F,$57
	.dc.b	$FC,$65,$5A,$C3,$FC,$69,$87,$F8,$9E,$73,$87,$F1,$EF,$4F,$FD,$45
	.dc.b	$87,$FA,$72,$BF,$D9,$2D,$78,$1F,$E5,$1C,$3A,$3F,$E5,$33,$33,$E2
	.dc.b	$7F,$C6,$91,$A8,$EB,$18,$AF,$45,$79,$E3,$A4,$78,$E3,$13,$CE,$7E
	.dc.b	$EC,$E5,$FF,$58,$1C,$30,$F6,$BF,$EB,$13,$07,$94,$FF,$67,$1F,$ED
	.dc.b	$D7,$FD,$B1,$3D,$6F,$CD,$7E,$03,$FB,$2B,$8A,$FF,$61,$1F,$ED,$6C
	.dc.b	$7D,$6F,$33,$FD,$9D,$3F,$1D,$FF,$7B,$D7,$5C,$3D,$11,$FD,$C3,$1F
	.dc.b	$A3,$8F,$A1,$FF,$6E,$C7,$63,$A6,$1D,$B1,$FF,$7C,$C7,$D4,$3C,$58
	.dc.b	$FF,$B8,$F1,$ED,$87,$39,$F8,$31,$C7,$1F,$F7,$A4,$7F,$B8,$2B,$FD
	.dc.b	$B4,$F2,$3C,$17,$D9,$FB,$F8,$74,$7F,$C3,$67,$DE,$9F,$8B,$5E,$C8
	.dc.b	$C1,$F7,$70,$F1,$57,$FB,$AA,$3D,$07,$B2,$BC,$2D,$3C,$35,$E6,$9F
	.dc.b	$01,$F8,$18,$EA,$7E,$0A,$BD,$D8,$79,$4F,$73,$FE,$83,$9E,$13,$F8
	.dc.b	$39,$D5,$FD,$D8,$F3,$1F,$85,$C7,$E9,$B4,$F9,$31,$B7,$C0,$66,$79
	.dc.b	$BF,$92,$6B,$F7,$4E,$26,$63,$E3,$DF,$A1,$B1,$DE,$FE,$33,$FB,$DC
	.dc.b	$3F,$4A,$F1,$3C,$F7,$EC,$C3,$CE,$FC,$08,$FC,$89,$EC,$AD,$3E,$23
	.dc.b	$C9,$1E,$67,$B6,$1F,$1F,$0F,$A7,$8F,$EF,$2F,$8D,$EE,$FE,$55,$7A
	.dc.b	$23,$C1,$32,$3E,$35,$7E,$9E,$BB,$C6,$CE,$B8,$FD,$D7,$E0,$D7,$67
	.dc.b	$53,$EF,$3E,$7D,$F5,$7F,$0B,$87,$9D,$D6,$1D,$DF,$C2,$B5,$F3,$9F
	.dc.b	$22,$BC,$B8,$7E,$5E,$3F,$81,$CF,$E7,$DB,$1E,$57,$F4,$87,$E6,$23
	.dc.b	$E2,$4F,$2A,$FB,$53,$D4,$F2,$9F,$48,$FC,$3E,$1F,$49,$FF,$0C,$E3
	.dc.b	$FE,$43,$7F,$8B,$18,$3F,$CF,$3E,$83,$D5,$7A,$BF,$E2,$08,$C3,$63
	.dc.b	$33,$C1,$AF,$F8,$66,$5E,$AF,$F9,$5E,$3A,$B5,$D3,$5F,$F4,$1C,$70
	.dc.b	$E2,$F0,$4F,$FA,$7A,$7F,$D3,$73,$FE,$B2,$7F,$D9,$38,$FF,$35,$F5
	.dc.b	$5E,$AF,$FA,$32,$3F,$D6,$93,$FE,$A2,$9C,$BF,$ED,$A8,$FF,$62,$BD
	.dc.b	$2F,$B4,$60,$9A,$FF,$AD,$2B,$8D,$7F,$B9,$E7,$C1,$7D,$EB,$FD,$63
	.dc.b	$3D,$E3,$FD,$E1,$3E,$23,$71,$E1,$9E,$93,$C6,$07,$FA,$DF,$0F,$73
	.dc.b	$FE,$EA,$9E,$B5,$CA,$3F,$DD,$07,$73,$A1,$F1,$23,$FD,$B8,$E4,$DC
	.dc.b	$71,$24,$74,$3F,$D6,$77,$EE,$8C,$CF,$92,$BF,$DE,$2F,$53,$FD,$C3
	.dc.b	$5C,$5A,$43,$D5,$87,$CC,$8F,$7C,$7F,$BF,$4D,$4F,$47,$DC,$77,$9F
	.dc.b	$F7,$AB,$A6,$BF,$EE,$5C,$3E,$43,$DB,$1F,$F6,$CD,$F8,$E3,$47,$C4
	.dc.b	$3A,$9B,$3E,$26,$1E,$E7,$0D,$C7,$E1,$4F,$36,$CF,$8E,$BF,$DE,$0F
	.dc.b	$9D,$AF,$29,$F9,$77,$F8,$27,$93,$63,$81,$F9,$19,$F4,$B8,$BF,$C8
	.dc.b	$BF,$3C,$F5,$9F,$91,$AF,$BB,$1E,$7A,$F8,$78,$FC,$B3,$F1,$24,$C7
	.dc.b	$E9,$5F,$3C,$7D,$6F,$9C,$FB,$0E,$27,$94,$F0,$9F,$93,$18,$C7,$CF
	.dc.b	$1C,$4F,$2E,$1F,$2E,$75,$1F,$3B,$1F,$39,$F1,$1F,$A9,$1E,$5A,$E6
	.dc.b	$F0,$AF,$64,$FE,$1C,$EB,$5F,$29,$E3,$7E,$5A,$E5,$1D,$CF,$AC,$7C
	.dc.b	$D3,$07,$23,$E7,$4F,$E5,$1F,$98,$70,$AF,$A2,$FC,$0A,$FC,$8E,$1F
	.dc.b	$15,$B8,$FD,$06,$CF,$9E,$75,$8F,$F4,$78,$7A,$4F,$57,$FD,$D5,$18
	.dc.b	$6C,$66,$3F,$D3,$7E,$A3,$D5,$CD,$70,$87,$C0,$53,$35,$FE,$A2,$BE
	.dc.b	$4F,$FA,$84,$4F,$07,$AC,$F0,$75,$EA,$BD,$4F,$2B,$63,$FD,$5A,$D7
	.dc.b	$A7,$AA,$F5,$B7,$5E,$C8,$FF,$50,$CF,$FA,$D1,$AF,$68,$DB,$9F,$0B
	.dc.b	$81,$87,$46,$48,$60,$F6,$7A,$8F,$57,$E5,$B6,$3F,$D8,$6D,$7E,$17
	.dc.b	$A6,$F5,$B7,$5F,$4A,$1E,$B5,$33,$5C,$A7,$0F,$FB,$3A,$76,$E4,$FF
	.dc.b	$68,$FA,$CF,$77,$F9,$DC,$7F,$B4,$7B,$1A,$AF,$3C,$AE,$EF,$E8,$35
	.dc.b	$A9,$FC,$2E,$3C,$F0,$EB,$1C,$2B,$78,$7F,$A5,$70,$E5,$5F,$DE,$23
	.dc.b	$13,$E2,$BF,$04,$FF,$B9,$BD,$77,$BD,$F6,$63,$CE,$7F,$52,$8C,$F1
	.dc.b	$62,$31,$3B,$77,$3A,$74,$FE,$7B,$D4,$E1,$08,$C3,$AE,$74,$EA,$73
	.dc.b	$3E,$2F,$59,$EC,$7F,$BB,$77,$7A,$2F,$A0,$FF,$76,$56,$B6,$FE,$D3
	.dc.b	$C1,$F0,$AF,$75,$AF,$67,$96,$1F,$56,$F6,$7D,$6C,$13,$AE,$3F,$F7
	.dc.b	$C7,$AE,$F5,$FF,$DF,$DB,$BD,$97,$B3,$A9,$CC,$63,$2E,$BC,$1F,$DF
	.dc.b	$30,$E5,$18,$C3,$96,$1C,$63,$17,$C6,$FF,$71,$7C,$1B,$87,$DE,$DF
	.dc.b	$EF,$F9,$6F,$59,$EC,$F3,$BF,$E0,$1E,$C3,$D8,$E1,$A7,$C6,$E2,$7B
	.dc.b	$38,$BF,$F0,$8D,$EE,$BD,$39,$FD,$0A,$D5,$2B,$4A,$7C,$97,$D2,$FE
	.dc.b	$0C,$EF,$0E,$D5,$CE,$73,$7C,$A3,$8E,$1D,$A3,$13,$E7,$8C,$54,$8F
	.dc.b	$9B,$9F,$C1,$F0,$B1,$8F,$F1,$FF,$59,$EB,$FB,$69,$EC,$BD,$9F,$5D
	.dc.b	$6B,$1E,$73,$97,$D3,$19,$C3,$DD,$1F,$E3,$6B,$E7,$7C,$B0,$ED,$87
	.dc.b	$1B,$F1,$46,$6F,$F0,$7E,$CB,$D9,$7B,$39,$BF,$E6,$C7,$FA,$90,$FF
	.dc.b	$3D,$65,$FE,$85,$F5,$FF,$DE,$B1,$FE,$83,$F6,$3F,$D0,$CF,$FB,$64
	.dc.b	$C6,$3F,$D3,$DE,$5E,$AF,$FA,$65,$C9,$D2,$78,$DF,$58,$FF,$7A,$1F
	.dc.b	$E8,$A9,$E0,$F2,$F5,$BF,$D5,$37,$D6,$3F,$D2,$ED,$C3,$CB,$1B,$39
	.dc.b	$4F,$07,$C1,$3B,$7B,$9E,$4A,$EE,$F4,$36,$47,$35,$DA,$35,$18,$79
	.dc.b	$9E,$13,$A5,$78,$63,$47,$FB,$7A,$38,$C6,$48,$FA,$08,$72,$29,$D4
	.dc.b	$87,$23,$25,$38,$10,$DC,$FA,$48,$FA,$0E,$06,$8C,$3A,$9F,$43,$67
	.dc.b	$FD,$B2,$7A,$27,$B1,$C8,$91,$B7,$FD,$96,$75,$9E,$6F,$F3,$63,$FD
	.dc.b	$86,$7F,$3F,$8F,$F6,$31,$FE,$BF,$F0,$64,$7F,$AF,$DA,$FF,$AE,$B9
	.dc.b	$7F,$D6,$E4,$7F,$D7,$ED,$AF,$F1,$DC,$7F,$8B,$8F,$F1,$91,$FE,$49
	.dc.b	$F5,$3F,$CA,$67,$F8,$A1,$FF,$1C,$47,$03,$FD,$55,$87,$F9,$26,$3F
	.dc.b	$D3,$6D,$8F,$F2,$87,$AB,$FE,$32,$9E,$D5,$CE,$7F,$C7,$7E,$BF,$F9
	.dc.b	$19,$E9,$7F,$E4,$BA,$E6,$7F,$98,$BD,$7F,$F1,$E6,$3F,$E7,$CB,$FF
	.dc.b	$2A,$57,$73,$FC,$B7,$5B,$3F,$C8,$15,$C1,$F4,$C6,$DF,$F2,$63,$FE
	.dc.b	$4F,$BF,$5C,$71,$3C,$F3,$C6,$3C,$F8,$FF,$97,$49,$87,$F9,$99,$F7
	.dc.b	$47,$2A,$F9,$78,$FB,$F0,$EC,$7B,$67,$E2,$DF,$AE,$F8,$CE,$4F,$5D
	.dc.b	$F2,$C3,$C5,$1C,$A3,$FC,$CE,$7F,$A2,$63,$2D,$3E,$09,$0F,$59,$6F
	.dc.b	$CD,$87,$DE,$9E,$33,$FE,$6D,$C3,$C8,$47,$E4,$9F,$16,$B8,$CF,$FA
	.dc.b	$A6,$3F,$10,$DC,$7E,$AC,$6F,$0F,$63,$F4,$0F,$31,$E4,$25,$74,$7F
	.dc.b	$D1,$77,$F3,$4F,$14,$7D,$27,$11,$E4,$3E,$31,$ED,$9E,$EF,$E7,$2B
	.dc.b	$C6,$E8,$EF,$1A,$C3,$EB,$47,$92,$FC,$4F,$E4,$8E,$0E,$0F,$B2,$F8
	.dc.b	$63,$E5,$3E,$13,$EA,$9F,$2C,$D6,$1F,$17,$1F,$BF,$1A,$F2,$3F,$D0
	.dc.b	$B9,$7F,$92,$23,$FD,$0B,$EA,$7F,$A2,$DF,$F4,$57,$AD,$FE,$85,$9D
	.dc.b	$7A,$DF,$E8,$99,$FF,$46,$36,$39,$5F,$4E,$0F,$F3,$0B,$68,$76,$3F
	.dc.b	$D2,$27,$63,$FD,$1B,$EA,$FF,$A0,$9F,$F3,$FD,$79,$7D,$6F,$F4,$83
	.dc.b	$67,$53,$FE,$92,$EC,$FF,$44,$4E,$63,$8E,$1F,$E7,$2F,$47,$FD,$23
	.dc.b	$39,$9F,$F4,$53,$F0,$63,$33,$BA,$E6,$FF,$9E,$BD,$6F,$F4,$55,$7C
	.dc.b	$76,$C6,$7D,$4F,$F4,$57,$AD,$FE,$88,$C7,$FD,$1F,$EB,$7F,$A2,$1F
	.dc.b	$F4,$BC,$E7,$0F,$A1,$EB,$7F,$A2,$A7,$67,$FA,$2C,$FF,$5C,$7A,$DF
	.dc.b	$E8,$17,$FD,$15,$A7,$F9,$E9,$B3,$FE,$86,$F3,$FF,$D0,$EF,$FA,$16
	.dc.b	$3F,$D0,$5D,$17,$AB,$FD,$22,$EA,$3F,$D0,$5E,$AF,$F9,$ED,$FF,$4B
	.dc.b	$FA,$BF,$E7,$B7,$FC,$FB,$19,$7F,$D0,$3E,$B7,$F9,$DF,$97,$FC,$F3
	.dc.b	$1F,$E7,$BF,$4B,$04,$7F,$D1,$1E,$A7,$F9,$D5,$FF,$51,$79,$BF,$E7
	.dc.b	$9E,$4F,$F2,$2C,$E7,$87,$FC,$F1,$39,$AF,$F6,$C4,$FF,$9D,$2B,$FC
	.dc.b	$D7,$AF,$FA,$0B,$7F,$F3,$E4,$62,$FF,$CF,$97,$FE,$66,$8F,$F3,$5F
	.dc.b	$A7,$FE,$63,$C7,$FD,$4B,$97,$F9,$C2,$7F,$CA,$BE,$97,$F9,$B6,$73
	.dc.b	$E9,$FF,$9C,$5F,$D2,$46,$67,$F3,$35,$C7,$0E,$C4,$AF,$F3,$9F,$A5
	.dc.b	$FE,$72,$AE,$BE,$9F,$F9,$52,$7F,$CB,$B1,$FE,$60,$F4,$3F,$CC,$7A
	.dc.b	$7F,$99,$F0,$FF,$3B,$C6,$3D,$1E,$15,$FE,$61,$38,$1F,$E7,$4A,$E1
	.dc.b	$7F,$B9,$F4,$BF,$CC,$F8,$7F,$9C,$E7,$35,$FE,$D2,$9C,$D7,$F9,$C6
	.dc.b	$BB,$CF,$3B,$FF,$29,$4F,$FA,$33,$D0,$FF,$2D,$3F,$C1,$7D,$2F,$F3
	.dc.b	$13,$9F,$4B,$77,$FE,$63,$AD,$BF,$EE,$0A,$DB,$99,$DE,$0F,$37,$FC
	.dc.b	$D3,$3F,$E6,$EF,$4B,$FC,$D3,$5F,$E8,$EA,$FF,$0B,$75,$FF,$87,$9F
	.dc.b	$F1,$6F,$9D,$FE,$1B,$7F,$CC,$06,$8F,$F0,$E9,$FE,$20,$3F,$C5,$24
	.dc.b	$F3,$7F,$C3,$E4,$9E,$91,$FE,$28,$9E,$98,$FF,$98,$EB,$53,$D2,$73
	.dc.b	$E7,$FF,$9A,$09,$97,$F9,$6E,$B7,$5E,$43,$C4,$E2,$FA,$61,$E0,$9F
	.dc.b	$41,$BC,$3A,$DF,$39,$D4,$F2,$6C,$EA,$3F,$CC,$91,$FE,$41,$79,$D7
	.dc.b	$67,$D0,$7C,$18,$E2,$D3,$37,$E5,$C3,$D7,$1D,$0F,$9E,$6C,$F3,$16
	.dc.b	$BB,$38,$9E,$58,$7A,$63,$AD,$F8,$27,$31,$D4,$FA,$07,$3B,$F0,$C7
	.dc.b	$F9,$DA,$78,$12,$79,$90,$F0,$39,$69,$9B,$F2,$92,$BA,$CF,$C7,$9F
	.dc.b	$F2,$F9,$FD,$50,$FE,$AB,$87,$F9,$81,$FF,$32,$3F,$E5,$E7,$5E,$7B
	.dc.b	$D8,$75,$E6,$7F,$AE,$CE,$63,$FC,$C5,$1F,$E6,$53,$FC,$C6,$43,$19
	.dc.b	$7F,$94,$9F,$EF,$33,$FE,$55,$C3,$FC,$91,$1F,$E0,$BF,$3B,$FB,$CB
	.dc.b	$FE,$20,$3F,$C2,$E2,$B9,$39,$3F,$C1,$5E,$87,$F8,$19,$FF,$08,$9D
	.dc.b	$A7,$FC,$96,$F1,$AE,$F3,$FE,$14,$8F,$F0,$55,$7F,$88,$5D,$1F,$E2
	.dc.b	$D3,$97,$17,$F8,$92,$34,$78,$B0,$FF,$10,$D7,$86,$3F,$C3,$57,$FE
	.dc.b	$4E,$8C,$3F,$E2,$3F,$43,$FC,$87,$5F,$07,$0E,$C7,$AE,$FF,$C4,$27
	.dc.b	$6A,$F3,$D7,$0C,$7D,$73,$E1,$7E,$3D,$F8,$A7,$FC,$45,$8F,$F8,$F2
	.dc.b	$FF,$CB,$98,$75,$7C,$93,$A8,$FF,$14,$9F,$3B,$0E,$95,$C6,$3F,$C7
	.dc.b	$15,$E9,$7F,$C6,$47,$33,$FC,$A0,$7C,$A7,$BD,$F8,$E3,$A4,$7A,$9E
	.dc.b	$31,$97,$EC,$DF,$8E,$FE,$04,$7C,$49,$D4,$7C,$4A,$EB,$87,$C0,$26
	.dc.b	$1E,$7B,$F1,$1E,$83,$EC,$9F,$70,$FF,$27,$CF,$AD,$FC,$39,$EB,$3F
	.dc.b	$25,$1C,$EF,$EB,$47,$03,$9D,$FC,$67,$E4,$9D,$71,$FC,$67,$C1,$7E
	.dc.b	$A6,$3F,$C8,$77,$FE,$51,$C5,$E9,$7F,$86,$65,$18,$77,$C3,$FB,$7B
	.dc.b	$E5,$0B,$FF,$2D,$46,$1E,$97,$CE,$F9,$56,$EF,$96,$DD,$F0,$E1,$1F
	.dc.b	$C0,$EB,$78,$72,$BF,$09,$FE,$5A,$AD,$B6,$FB,$61,$23,$62,$76,$BB
	.dc.b	$FF,$CE,$16,$C7,$FC,$CB,$BB,$F2,$0D,$5F,$F8,$81,$AF,$A4,$FF,$35
	.dc.b	$DF,$2A,$FE,$34,$D8,$FE,$2B,$3C,$2F,$B4,$62,$7E,$71,$B5,$E3,$3C
	.dc.b	$67,$8D,$FC,$92,$BF,$C8,$B1,$8A,$72,$79,$2F,$85,$EE,$7F,$CF,$FF
	.dc.b	$48,$FA,$4F,$E5,$D7,$BC,$39,$36,$F8,$5F,$8A,$73,$8F,$F3,$5C,$39
	.dc.b	$4E,$B1,$EE,$F8,$A7,$CB,$79,$BE,$33,$E2,$C7,$C3,$5A,$8E,$07,$F9
	.dc.b	$70,$A7,$86,$BC,$78,$78,$23,$38,$78,$31,$F4,$46,$31,$EE,$AB,$84
	.dc.b	$E8,$98,$78,$5F,$B7,$1A,$8C,$12,$7C,$93,$AC,$7D,$31,$89,$D3,$E7
	.dc.b	$3D,$35,$CD,$D1,$0D,$57,$37,$E3,$63,$1F,$53,$F0,$2F,$BC,$70,$BE
	.dc.b	$F8,$FC,$17,$EA,$4E,$A3,$13,$DD,$AF,$C3,$8D,$63,$FE,$8B,$8F,$41
	.dc.b	$FE,$8E,$BF,$EE,$4F,$D5,$2D,$3C,$8F,$F4,$7C,$61,$AF,$FA,$61,$AF
	.dc.b	$FB,$3A,$38,$78,$9F,$E9,$3C,$BF,$2D,$5F,$E9,$87,$F5,$7F,$66,$F7
	.dc.b	$7F,$4B,$1A,$AF,$C7,$56,$F4,$ED,$1F,$96,$BE,$1C,$1A,$89,$97,$E1
	.dc.b	$AF,$83,$FE,$E0,$AD,$B9,$9D,$E1,$DE,$3F,$23,$3A,$8C,$13,$0F,$0C
	.dc.b	$EB,$1F,$3C,$62,$75,$3F,$07,$0F,$0A,$BE,$95,$B2,$46,$F0,$F6,$61
	.dc.b	$D2,$30,$79,$C9,$87,$C8,$8D,$47,$F0,$8C,$3B,$E5,$F4,$23,$78,$7F
	.dc.b	$8B,$71,$F0,$C7,$4C,$3C,$EF,$1C,$7C,$39,$7C,$4B,$F2,$CE,$4F,$81
	.dc.b	$19,$79,$46,$AF,$AC,$6E,$FF,$93,$F9,$5F,$D3,$2F,$FD,$93,$F8,$E3
	.dc.b	$E7,$FF,$91,$F1,$7F,$24,$C1,$7B,$7F,$97,$72,$2F,$86,$1F,$ED,$8E
	.dc.b	$6E,$2F,$F3,$88,$CD,$7F,$B6,$27,$2F,$F9,$1A,$33,$87,$7F,$84,$7C
	.dc.b	$1E,$BE,$11,$F0,$AF,$85,$E3,$8A,$F7,$9F,$F5,$2B,$DD,$6A,$73,$5F
	.dc.b	$D1,$67,$35,$FD,$4A,$73,$8F,$F4,$29,$CD,$7F,$4B,$3B,$63,$E3,$C7
	.dc.b	$C5,$87,$18,$C6,$3D,$DB,$19,$3C,$93,$C2,$BA,$1F,$E9,$4F,$0F,$FC
	.dc.b	$E2,$FF,$A3,$3C,$0F,$F4,$19,$FE,$7F,$7F,$CF,$11,$A7,$FD,$0B,$1A
	.dc.b	$F0,$BF,$CF,$1C,$5F,$EA,$18,$E1,$3F,$E8,$02,$78,$9F,$E7,$FC,$7F
	.dc.b	$D2,$51,$8C,$3A,$3E,$38,$CC,$EE,$3F,$D2,$86,$4A,$7F,$AC,$4F,$2C
	.dc.b	$70,$9E,$71,$93,$B4,$64,$8F,$2B,$F4,$61,$EA,$8F,$F5,$04,$EB,$0F
	.dc.b	$5E,$1E,$36,$E1,$DB,$0E,$37,$F2,$70,$FF,$50,$5E,$DE,$97,$F1,$E7
	.dc.b	$47,$A0,$F1,$1E,$02,$BB,$7E,$7C,$74,$BF,$04,$6E,$7C,$78,$79,$EB
	.dc.b	$F9,$BC,$FF,$AA,$8F,$EF,$A7,$FA,$A0,$87,$F3,$CF,$0F,$FD,$48,$F0
	.dc.b	$6B,$9F,$07,$FD,$56,$E5,$AF,$FA,$A4,$87,$FA,$CB,$C2,$FF,$20,$3F
	.dc.b	$EA,$B3,$FC,$C3,$C1,$FE,$55,$8F,$F4,$1D,$FF,$9B,$67,$FC,$B7,$7F
	.dc.b	$E6,$88,$FF,$3F,$C7,$F9,$C8,$FF,$36,$4F,$F9,$A1,$FF,$2C,$E3,$FE
	.dc.b	$6D,$7F,$CD,$07,$F9,$8F,$C3,$FF,$53,$5F,$F9,$89,$FF,$4D,$C7,$F9
	.dc.b	$8A,$FC,$15,$FE,$A0,$8F,$F4,$1D,$6F,$C4,$FF,$3E,$57,$A2,$7B,$BE
	.dc.b	$47,$FD,$02,$79,$4F,$F3,$71,$E1,$8F,$F5,$06,$3F,$EB,$19,$E5,$7F
	.dc.b	$E8,$6A,$DD,$FB,$DF,$0E,$3F,$11,$F7,$DF,$C7,$39,$3E,$E9,$E0,$FF
	.dc.b	$A0,$0F,$29,$CA,$36,$FF,$A3,$09,$1E,$D7,$D1,$8F,$C4,$7D,$07,$FA
	.dc.b	$7A,$BE,$75,$F9,$E7,$55,$E0,$7E,$6C,$F9,$EB,$D6,$F0,$C3,$DE,$48
	.dc.b	$F4,$46,$5B,$87,$B5,$FF,$49,$63,$F4,$1C,$9E,$C3,$DD,$8F,$8E,$35
	.dc.b	$1E,$CA,$F1,$61,$E6,$7B,$47,$F9,$F7,$2F,$F4,$03,$FC,$AE,$3F,$D1
	.dc.b	$39,$FF,$A3,$A3,$FD,$07,$E2,$7F,$A0,$A7,$30,$3F,$D3,$FC,$DF,$E9
	.dc.b	$3B,$91,$FE,$8A,$9D,$BA,$39,$C6,$9F,$04,$F8,$B3,$FF,$51,$CF,$F9
	.dc.b	$FB,$0E,$D5,$FD,$73,$C4,$FF,$42,$47,$FA,$41,$B1,$FE,$8C,$8E,$0F
	.dc.b	$FA,$47,$B3,$8D,$FF,$A2,$BA,$7F,$D2,$D1,$A9,$FE,$95,$DD,$FE,$40
	.dc.b	$27,$85,$FE,$64,$F1,$3F,$CA,$93,$FE,$4B,$3F,$C8,$D1,$9F,$13,$FC
	.dc.b	$8B,$E2,$7F,$94,$FA,$3F,$CC,$F5,$FE,$79,$31,$8F,$F9,$4A,$F8,$46
	.dc.b	$0E,$75,$FE,$59,$70,$77,$26,$1E,$2C,$7C,$B3,$A8,$C5,$7F,$9E,$A7
	.dc.b	$35,$FE,$6C,$C7,$CA,$FF,$99,$89,$8F,$F9,$AD,$B8,$F9,$71,$F1,$46
	.dc.b	$0F,$29,$CC,$F2,$DF,$6A,$E4,$7F,$A3,$23,$FC,$C8,$7F,$2E,$8F,$F4
	.dc.b	$FB,$83,$FC,$C5,$E1,$7F,$32,$7F,$C4,$C6,$0E,$1E,$1F,$F9,$91,$FF
	.dc.b	$12,$61,$FE,$27,$F1,$3F,$C7,$37,$FE,$37,$C3,$FC,$66,$D9,$FF,$1C
	.dc.b	$4E,$AF,$FC,$7A,$4C,$3F,$C6,$F8,$7F,$8F,$8F,$F1,$99,$0F,$F2,$84
	.dc.b	$F1,$8F,$F1,$C4,$78,$8F,$F3,$24,$65,$E4,$64,$E3,$BF,$0C,$3C,$91
	.dc.b	$D7,$C4,$FF,$1F,$C6,$23,$83,$63,$38,$79,$E7,$AC,$63,$0E,$93,$FE
	.dc.b	$66,$8C,$4F,$8A,$B9,$61,$E5,$CB,$E2,$DE,$E3,$19,$7E,$B3,$C3,$FF
	.dc.b	$24,$70,$FE,$97,$C3,$FF,$1F,$EB,$9E,$5F,$F2,$1F,$1F,$F9,$2F,$C4
	.dc.b	$FF,$1E,$E1,$C2,$BF,$C9,$51,$8A,$DD,$F2,$DB,$F8,$37,$89,$FE,$3A
	.dc.b	$BF,$E1,$1E,$27,$F8,$EB,$3F,$F2,$67,$87,$FE,$34,$BF,$F1,$E5,$ED
	.dc.b	$56,$EF,$F8,$1C,$E6,$31,$3D,$AB,$95,$F4,$BF,$F1,$EF,$89,$FE,$41
	.dc.b	$69,$FE,$41,$AF,$F2,$87,$89,$FE,$41,$AF,$F3,$64,$7F,$8A,$FC,$0F
	.dc.b	$F0,$4B,$FE,$10,$F0,$7F,$C1,$D7,$FE,$0B,$8F,$F0,$B9,$FE,$2A,$34
	.dc.b	$7F,$86,$67,$89,$27,$FC,$3E,$72,$8D,$9C,$63,$4F,$4F,$13,$FC,$4C
	.dc.b	$DF,$13,$FC,$3B,$3F,$E2,$18,$F3,$57,$47,$C0,$D7,$A3,$6B,$9C,$7F
	.dc.b	$8A,$0D,$9D,$6F,$D8,$E4,$E2,$64,$F3,$C7,$27,$C2,$66,$BD,$2F,$18
	.dc.b	$D3,$FE,$4F,$6F,$37,$F9,$12,$35,$87,$B3,$0F,$89,$7C,$A7,$A9,$AC
	.dc.b	$3C,$AF,$F5,$1B,$FE,$A5,$2A,$FE,$8F,$83,$06,$1A,$9F,$F2,$0D,$FF
	.dc.b	$11,$C3,$91,$FE,$40,$32,$7F,$8E,$F8,$FF,$8B,$78,$9F,$E4,$08,$FE
	.dc.b	$A6,$D7,$FB,$13,$52,$F0,$7F,$C7,$F1,$FE,$42,$36,$E1,$FF,$22,$69
	.dc.b	$DA,$3F,$BB,$F8,$9F,$DE,$AB,$FC,$29,$5F,$E2,$C2,$6B,$FE,$3D,$7F
	.dc.b	$BF,$5F,$F8,$72,$77,$39,$6E,$5F,$E3,$58,$C6,$1D,$AF,$FC,$3F,$87
	.dc.b	$78,$C4,$76,$AE,$04,$7F,$C5,$47,$F8,$50,$E2,$43,$81,$AA,$FA,$8F
	.dc.b	$F8,$E0,$FF,$0C,$78,$9F,$E2,$6C,$7D,$17,$E0,$AF,$2C,$60,$95,$DC
	.dc.b	$EA,$7F,$88,$2F,$FC,$54,$FC,$23,$53,$C9,$B1,$E9,$78,$1F,$E2,$A8
	.dc.b	$FF,$14,$3E,$97,$81,$E8,$7F,$C7,$33,$FE,$33,$8C,$5F,$F8,$AE,$3F
	.dc.b	$19,$3F,$E2,$38,$61,$FE,$10,$D3,$FC,$33,$18,$9F,$F1,$45,$7F,$BE
	.dc.b	$63,$11,$FE,$DB,$CB,$F1,$13,$D2,$13,$A8,$E1,$7D,$67,$C0,$D7,$F2
	.dc.b	$39,$7F,$A3,$B1,$F3,$CE,$6F,$96,$3E,$48,$E1,$8F,$83,$05,$FD,$8F
	.dc.b	$B9,$FE,$63,$E7,$77,$7F,$99,$FB,$77,$94,$D9,$97,$EE,$CB,$FC,$8D
	.dc.b	$7C,$23,$01,$C2,$D7,$F6,$EE,$70,$5F,$BE,$5F,$B8,$BE,$DF,$0D,$BE
	.dc.b	$B3,$A9,$F8,$DF,$30,$FC,$D3,$AD,$77,$CF,$E9,$E1,$C6,$30,$78,$DC
	.dc.b	$61,$E4,$3A,$47,$AA,$FA,$57,$CD,$EF,$93,$EE,$3E,$CB,$F4,$7C,$CC
	.dc.b	$C7,$63,$26,$F8,$BF,$CB,$6F,$E0,$23,$78,$7E,$1E,$AD,$FE,$32,$7F
	.dc.b	$05,$AB,$E4,$1F,$4B,$86,$E3,$F4,$27,$53,$F1,$3F,$4B,$7C,$76,$E2
	.dc.b	$56,$EF,$8B,$F1,$35,$5F,$08,$D7,$C5,$F7,$23,$F1,$7E,$A9,$F6,$F8
	.dc.b	$8F,$9B,$9F,$9E,$B1,$BF,$E0,$31,$61,$C2,$FF,$D1,$5E,$2F,$FA,$17
	.dc.b	$7F,$F4,$1F,$5F,$FA,$33,$D3,$FF,$4A,$4E,$6B,$FD,$29,$39,$F4,$FF
	.dc.b	$2D,$F6,$12,$7F,$23,$D5,$FE,$0C,$8F,$C2,$FD,$4B,$E4,$75,$5F,$82
	.dc.b	$AD,$3C,$38,$17,$C2,$BF,$65,$E1,$3F,$26,$7D,$C6,$1B,$87,$E8,$27
	.dc.b	$53,$FB,$7F,$C1,$BF,$A3,$4E,$6B,$6F,$E4,$AB,$6E,$6B,$86,$BF,$BF
	.dc.b	$F4,$BF,$CE,$5D,$9F,$EA,$65,$EB,$FF,$A3,$AF,$FC,$F4,$7F,$A2,$E3
	.dc.b	$3E,$B7,$F9,$FB,$C9,$FF,$4B,$CE,$6B,$84,$62,$BA,$CF,$FA,$6C,$E6
	.dc.b	$E1,$7B,$1F,$E9,$2B,$FF,$42,$DF,$2C,$3F,$D1,$95,$E2,$AD,$3C,$A7
	.dc.b	$4F,$FA,$3B,$0F,$1C,$E9,$F1,$4E,$A7,$CD,$7F,$EA,$1F,$63,$FD,$3B
	.dc.b	$5A,$8F,$66,$1E,$08,$E7,$3A,$8C,$1F,$EC,$CC,$3B,$38,$BF,$75,$73
	.dc.b	$7D,$15,$C5,$F3,$4F,$18,$C4,$79,$0F,$F5,$31,$FE,$AF,$3F,$D4,$07
	.dc.b	$D5,$C1,$87,$FA,$87,$1F,$F5,$2C,$FD,$9F,$5B,$FD,$41,$06,$93,$D8
	.dc.b	$E0,$FF,$A1,$72,$65,$FE,$A0,$E9,$FF,$42,$D7,$FA,$63,$D8,$FF,$42
	.dc.b	$D7,$FA,$0A,$3F,$D4,$2D,$C7,$FD,$1B,$EC,$7F,$A1,$1F,$F5,$16,$5F
	.dc.b	$E8,$AC,$39,$C6,$9B,$5E,$08,$FF,$4E,$57,$17,$FD,$37,$97,$FA,$3E
	.dc.b	$B7,$97,$FA,$E5,$B3,$C1,$AF,$FA,$A7,$2F,$3C,$FC,$3C,$BD,$58,$75
	.dc.b	$CB,$FD,$50,$79,$6F,$FD,$41,$8F,$C3,$BF,$85,$18,$BE,$38,$F6,$BE
	.dc.b	$19,$7C,$13,$D5,$5E,$1C,$7B,$63,$CE,$3F,$D7,$F7,$FE,$4F,$C7,$C9
	.dc.b	$0B,$F8,$5C,$2F,$C7,$7F,$D3,$0D,$F5,$7F,$A4,$60,$B9,$F8,$3D,$8F
	.dc.b	$E0,$97,$C0,$95,$C2,$B6,$7F,$AC,$A3,$FD,$49,$D4,$FC,$03,$FC,$1F
	.dc.b	$7F,$C7,$1B,$5F,$CD,$FF,$80,$5F,$31,$CE,$F8,$7A,$96,$F7,$E2,$FF
	.dc.b	$70,$9C,$E3,$FD,$77,$0E,$2F,$5B,$FF,$56,$BF,$EB,$AF,$63,$FD,$50
	.dc.b	$FF,$AD,$34,$CE,$7F,$EB,$89,$CB,$7D,$9F,$F5,$93,$FE,$B0,$7D,$DE
	.dc.b	$0B,$EA,$7E,$24,$FF,$46,$6E,$9F,$D1,$7C,$77,$89,$C2,$73,$9F,$FB
	.dc.b	$31,$B7,$FE,$C8,$6F,$91,$FD,$32,$B8,$4F,$1F,$66,$5F,$F7,$BF,$31
	.dc.b	$F3,$98,$AF,$15,$5C,$FF,$AB,$56,$AF,$FB,$66,$5F,$E8,$F7,$FC,$01
	.dc.b	$ED,$BE,$43,$ED,$62,$8C,$4E,$BE,$FB,$31,$89,$FF,$02,$37,$1F,$F0
	.dc.b	$BB,$6B,$FD,$3B,$5C,$2E,$47,$81,$E7,$18,$BF,$F0,$DD,$BF,$73,$FF
	.dc.b	$15,$DF,$F8,$99,$B3,$FD,$F7,$D8,$7D,$F7,$E4,$6F,$87,$B7,$6F,$71
	.dc.b	$FE,$23,$6E,$1F,$9D,$E1,$FF,$17,$5E,$EF,$F1,$DF,$80,$7E,$3F,$D1
	.dc.b	$B7,$F9,$76,$E7,$FD,$BF,$B1,$18,$CF,$F2,$91,$8A,$F9,$BC,$5F,$20
	.dc.b	$F9,$B7,$FE,$BA,$F7,$BF,$CF,$1C,$3F,$4E,$3E,$5C,$EE,$BF,$C6,$91
	.dc.b	$BF,$80,$FF,$63,$CE,$63,$73,$A7,$F7,$C7,$F9,$92,$77,$19,$F0,$FF
	.dc.b	$DB,$53,$FE,$D6,$21,$FE,$DF,$AF,$F6,$E3,$FE,$EE,$3F,$DB,$47,$FB
	.dc.b	$A9,$CB,$FE,$EE,$33,$87,$FB,$85,$E5,$5F,$EE,$06,$9C,$5C,$57,$8A
	.dc.b	$33,$3C,$63,$04,$8F,$51,$E3,$3C,$EE,$63,$E4,$D7,$93,$0F,$F7,$BD
	.dc.b	$70,$3C,$C7,$03,$0E,$A7,$8D,$F9,$27,$35,$CC,$FF,$3E,$CF,$FA,$B6
	.dc.b	$3E,$91,$FC,$96,$BE,$C4,$66,$3E,$8B,$D1,$F1,$1F,$EF,$83,$F9,$61
	.dc.b	$0F,$AE,$53,$47,$FB,$F0,$D7,$C0,$7F,$BD,$C9,$A3,$C4,$FF,$7F,$B6
	.dc.b	$79,$4F,$FA,$95,$FF,$7D,$1F,$EA,$ED,$7F,$DF,$04,$7F,$D9,$E7,$FB
	.dc.b	$87,$4F,$F5,$73,$FE,$BB,$F7,$7F,$DC,$0B,$E7,$FF,$D6,$AF,$F4,$56
	.dc.b	$B9,$9F,$6D,$FF,$AD,$67,$FD,$AF,$3F,$EE,$2A,$E7,$97,$03,$FD,$B3
	.dc.b	$7E,$09,$D7,$83,$F1,$A7,$FD,$B9,$87,$FA,$FD,$BE,$8F,$FA,$C6,$BC
	.dc.b	$07,$FB,$30,$F2,$5F,$3C,$77,$96,$27,$FD,$D5,$3F,$EF,$2C,$3C,$CE
	.dc.b	$C9,$1F,$D8,$63,$2F,$02,$1F,$2C,$F0,$9F,$EE,$0C,$7E,$ED,$FC,$43
	.dc.b	$C3,$1A,$9F,$55,$7B,$F1,$F8,$33,$E9,$3E,$99,$D5,$E3,$5C,$E7,$CB
	.dc.b	$3E,$18,$F9,$B3,$F2,$E7,$E8,$E3,$FE,$EB,$6E,$3F,$06,$FC,$C5,$CB
	.dc.b	$FD,$F7,$87,$6A,$F8,$97,$C6,$BC,$98,$75,$7E,$24,$65,$F1,$4F,$92
	.dc.b	$F9,$BE,$08,$F8,$13,$B9,$65,$F7,$6B,$E1,$1B,$BF,$8E,$77,$BF,$39
	.dc.b	$FB,$BC,$BE,$4B,$E2,$CB,$D5,$7F,$42,$FE,$2D,$7D,$8A,$F9,$51,$D9
	.dc.b	$FC,$F1,$EC,$70,$E8,$E8,$7D,$CC,$3D,$6F,$D5,$CB,$E3,$3F,$82,$AF
	.dc.b	$1B,$E6,$C7,$D1,$18,$3C,$55,$CA,$7D,$75,$C9,$F0,$1C,$6F,$C7,$5C
	.dc.b	$72,$F9,$18,$FC,$7A,$F4,$9B,$2E,$1F,$31,$FD,$C4,$6D,$EB,$5E,$69
	.dc.b	$D4,$7E,$B8,$FF,$1F,$63,$F4,$30,$FC,$0E,$5C,$A6,$1F,$47,$1F,$C0
	.dc.b	$61,$CF,$1E,$D8,$FE,$F3,$0F,$96,$DB,$F9,$C7,$9A,$76,$79,$6B,$94
	.dc.b	$7F,$93,$F0,$6F,$FC,$0D,$B1,$FA,$56,$BF,$22,$FD,$6F,$F2,$B6,$BF
	.dc.b	$C2,$CC,$5F,$F2,$D7,$95,$7F,$32,$97,$CF,$FF,$34,$BE,$B1,$8A,$F2
	.dc.b	$57,$8C,$FF,$32,$78,$AF,$05,$CD,$C6,$73,$0F,$82,$C5,$AF,$84,$61
	.dc.b	$E5,$76,$B8,$57,$FB,$E6,$B0,$66,$79,$E2,$F8,$17,$C1,$61,$B7,$E8
	.dc.b	$13,$B9,$F1,$F5,$BB,$7E,$5C,$FC,$E3,$55,$C6,$5F,$4D,$FE,$5A,$8C
	.dc.b	$35,$FF,$1B,$E1,$E0,$7B,$EB,$F3,$7C,$DF,$83,$0D,$FF,$D3,$38,$FC
	.dc.b	$E9,$CC,$3E,$0E,$C7,$D7,$AE,$0D,$9F,$F5,$0C,$7D,$F8,$C1,$CB,$E6
	.dc.b	$A4,$7C,$13,$89,$F0,$F8,$1F,$04,$F8,$3F,$AB,$18,$6C,$66,$7F,$CD
	.dc.b	$DF,$09,$89,$FA,$DF,$55,$27,$A5,$6D,$F8,$27,$E1,$57,$C2,$74,$24
	.dc.b	$7C,$DF,$35,$F0,$B8,$8E,$3D,$1F,$8D,$FB,$CF,$C2,$19,$E1,$FD,$0F
	.dc.b	$73,$E1,$7E,$33,$6F,$F3,$CD,$7F,$C1,$8D,$7F,$C2,$0E,$67,$F5,$BC
	.dc.b	$FF,$E9,$6F,$84,$FC,$2F,$A2,$7E,$CD,$6F,$9F,$F3,$BF,$0B,$F9,$72
	.dc.b	$63,$F8,$7B,$E0,$FE,$47,$4F,$D3,$4E,$A3,$DB,$C8,$8E,$F8,$FE,$9B
	.dc.b	$C6,$7C,$23,$E1,$3F,$71,$18,$6C,$67,$0F,$F1,$44,$E5,$B1,$97,$BF
	.dc.b	$C4,$FE,$B4,$ED,$19,$DF,$F8,$5F,$BE,$F8,$37,$C1,$FF,$BA,$70,$FF
	.dc.b	$71,$4E,$63,$10,$8C,$3F,$EE,$DE,$DF,$F1,$EF,$C2,$BE,$17,$FC,$7F
	.dc.b	$E8,$FF,$BE,$2F,$6D,$C7,$BE,$1C,$4E,$EE,$23,$FD,$FF,$7C,$E7,$C3
	.dc.b	$5C,$B0,$FF,$25,$B6,$38,$E3,$E6,$AE,$11,$D4,$F4,$61,$D6,$F8,$B6
	.dc.b	$38,$9B,$9E,$38,$79,$EF,$C3,$5B,$AE,$71,$DF,$1F,$7E,$1D,$23,$64
	.dc.b	$9E,$93,$A7,$BD,$F5,$7C,$D8,$FA,$D6,$3E,$2C,$3A,$46,$B1,$F1,$56
	.dc.b	$EB,$53,$D2,$B7,$7D,$4F,$97,$87,$86,$7D,$27,$3C,$7C,$D3,$A8,$B5
	.dc.b	$C8,$88,$F2,$D7,$3C,$3D,$04,$C7,$E3,$63,$E4,$AE,$F8,$78,$F0,$E7
	.dc.b	$1C,$8E,$E7,$A4,$86,$4F,$31,$23,$93,$E3,$9E,$D7,$E1,$C7,$9C,$F1
	.dc.b	$AD,$BE,$2B,$F0,$DF,$0C,$3E,$59,$2F,$EE,$57,$D9,$3F,$D3,$18,$7E
	.dc.b	$81,$B3,$F9,$A6,$CF,$F6,$F9,$7C,$13,$E1,$30,$DA,$FF,$50,$4E,$A7
	.dc.b	$24,$8F,$C1,$CE,$F2,$65,$FE,$AA,$F8,$57,$C2,$FF,$81,$A7,$F5,$D5
	.dc.b	$BA,$FD,$2B,$6F,$F7,$6D,$57,$CB,$A3,$F5,$B3,$9C,$BC,$95,$E2,$CB
	.dc.b	$F8,$0D,$EE,$30,$4C,$7C,$33,$AC,$BC,$99,$78,$A3,$A9,$2B,$F6,$97
	.dc.b	$CE,$F8,$C7,$3A,$E0,$EA,$F9,$BE,$AB,$E5,$5C,$0F,$8B,$1F,$95,$BE
	.dc.b	$97,$CA,$74,$D3,$E5,$5F,$4A,$FE,$39,$8F,$C1,$26,$1E,$BC,$BB,$DF
	.dc.b	$1A,$F1,$1E,$58,$CB,$8C,$3B,$BE,$3C,$3B,$65,$F3,$27,$33,$E0,$C3
	.dc.b	$B3,$F0,$4E,$58,$77,$7E,$BD,$F2,$C7,$CB,$1B,$33,$8F,$89,$E8,$67
	.dc.b	$2F,$86,$F4,$21,$AC,$7E,$4B,$C8,$E9,$97,$9F,$1F,$15,$72,$9E,$11
	.dc.b	$E5,$9D,$47,$BE,$FE,$1F,$13,$E1,$1F,$05,$FE,$2F,$9F,$F2,$5F,$B1
	.dc.b	$FE,$52,$7F,$C7,$15,$B9,$D5,$FF,$92,$6B,$4F,$F8,$3A,$75,$5F,$D8
	.dc.b	$AF,$AB,$6F,$FC,$CF,$5F,$C5,$EB,$6F,$F1,$93,$C3,$E8,$F6,$EB,$63
	.dc.b	$FE,$B6,$C3,$FC,$85,$F0,$6F,$84,$D4,$E7,$0F,$F0,$BB,$79,$FF,$CE
	.dc.b	$F8,$72,$C7,$D3,$E0,$31,$E5,$97,$F8,$17,$0E,$0F,$F2,$BE,$E7,$C2
	.dc.b	$3E,$13,$8D,$F0,$D3,$FC,$5B,$5A,$9C,$EB,$F1,$A3,$18,$7B,$6B,$A6
	.dc.b	$1C,$E3,$11,$FE,$79,$8C,$3B,$9E,$2D,$F8,$2F,$E7,$D5,$E7,$FB,$8E
	.dc.b	$D1,$89,$F8,$31,$86,$4F,$F8,$F7,$C3,$B1,$E8,$A7,$BD,$F2,$BC,$5E
	.dc.b	$6D,$BF,$F2,$B4,$E9,$B1,$F2,$BE,$11,$F0,$BC,$27,$FC,$BD,$F0,$0F
	.dc.b	$85,$FC,$3E,$48,$7B,$D8,$75,$5C,$23,$15,$F8,$6A,$FE,$B9,$5A,$7E
	.dc.b	$3B,$5F,$98,$D7,$F1,$4E,$F9,$FF,$D6,$AD,$BC,$FC,$2F,$67,$FD,$29
	.dc.b	$2F,$85,$FF,$54,$C7,$FA,$3F,$DB,$FF,$54,$3A,$8C,$E1,$FE,$6C,$E8
	.dc.b	$FF,$4C,$7B,$DF,$A4,$C3,$97,$BD,$6F,$71,$FE,$5C,$C5,$F0,$B9,$AF
	.dc.b	$CC,$FC,$24,$CF,$F2,$5F,$0B,$74,$FC,$DB,$71,$FC,$DF,$6A,$31,$5F
	.dc.b	$03,$93,$E6,$7C,$E7,$CD,$6C,$7D,$47,$E6,$61,$F4,$DB,$C3,$F7,$63
	.dc.b	$E6,$C6,$0E,$53,$FE,$99,$8F,$F0,$77,$8F,$FE,$F6,$E4,$F0,$4E,$65
	.dc.b	$39,$7E,$CD,$7F,$BA,$3E,$17,$F0,$31,$FA,$39,$D3,$E3,$AA,$F6,$13
	.dc.b	$C1,$D1,$8B,$FD,$4D,$72,$C3,$D5,$5B,$AE,$4F,$18,$CD,$70,$9E,$47
	.dc.b	$9A,$FC,$2F,$A8,$C9,$E0,$AE,$8E,$5E,$98,$7B,$6B,$A4,$62,$B9,$E1
	.dc.b	$E5,$AD,$5F,$C1,$3A,$E1,$E4,$71,$7C,$5F,$09,$EA,$C3,$E7,$B8,$3C
	.dc.b	$4E,$23,$CB,$7D,$F0,$F0,$D7,$6C,$3C,$2F,$FA,$08,$C4,$6A,$BC,$07
	.dc.b	$28,$C9,$DE,$BA,$34,$CC,$6C,$E3,$7F,$08,$F2,$16,$3E,$71,$C7,$E1
	.dc.b	$BD,$C7,$72,$C6,$E3,$E8,$37,$1F,$8D,$1E,$9A,$EE,$FC,$89,$EE,$FC
	.dc.b	$DA,$E6,$D8,$F9,$31,$E7,$77,$1C,$DA,$FC,$F3,$CF,$3B,$6C,$7E,$26
	.dc.b	$7A,$C7,$CB,$6B,$C4,$F4,$9E,$53,$89,$4C,$9E,$13,$27,$38,$F3,$9D
	.dc.b	$CA,$F1,$AF,$86,$F7,$6B,$F8,$43,$99,$9C,$7F,$02,$FE,$B6,$BC,$EF
	.dc.b	$E7,$DB,$87,$DE,$C3,$E0,$4F,$98,$E5,$8F,$E1,$F1,$F8,$F1,$C5,$F2
	.dc.b	$BE,$93,$EF,$DF,$2C,$7C,$B8,$75,$AF,$B9,$8F,$92,$FD,$07,$73,$A1
	.dc.b	$CA,$3E,$13,$4F,$85,$1C,$67,$A4,$63,$0F,$33,$F0,$8E,$67,$F0,$AC
	.dc.b	$3F,$21,$7E,$88,$F6,$9F,$9D,$C7,$F1,$AF,$D5,$3C,$97,$D9,$F7,$1C
	.dc.b	$E7,$CA,$D3,$C4,$73,$7E,$3C,$71,$8E,$F3,$F1,$70,$F5,$47,$EB,$2B
	.dc.b	$E2,$E1,$EA,$8F,$09,$93,$AD,$71,$8D,$63,$F3,$A5,$8F,$DC,$6D,$FB
	.dc.b	$DF,$9F,$3B,$BF,$7C,$79,$27,$C2,$FE,$FC,$FF,$0B,$9F,$C9,$4B,$8E
	.dc.b	$1F,$E7,$07,$F8,$7F,$E1,$9B,$BC,$5E,$31,$86,$C6,$67,$FC,$47,$39
	.dc.b	$6F,$63,$87,$8B,$63,$31,$82,$3F,$CE,$3E,$05,$E2,$FF,$93,$74,$FE
	.dc.b	$95,$18,$9F,$EA,$38,$30,$E9,$5F,$D3,$E3,$18,$7F,$56,$BE,$17,$D7
	.dc.b	$0E,$9A,$7F,$96,$6F,$73,$FD,$21,$BB,$78,$2B,$FC,$BC,$D9,$FF,$7C
	.dc.b	$3D,$F2,$F6,$37,$1F,$05,$76,$AF,$F7,$F6,$3E,$2C,$B5,$1D,$F0,$F4
	.dc.b	$E5,$FE,$6B,$AD,$36,$F9,$53,$2F,$3E,$1C,$8F,$24,$7B,$4F,$8B,$87
	.dc.b	$44,$71,$70,$B0,$EE,$F9,$AF,$BD,$F6,$C3,$E3,$5F,$0B,$DD,$F1,$BD
	.dc.b	$CF,$4B,$F9,$37,$E9,$AF,$8E,$D7,$F3,$67,$92,$36,$D9,$F9,$73,$BC
	.dc.b	$BD,$A7,$C8,$3E,$03,$CF,$26,$F4,$F3,$BF,$81,$F8,$2F,$F4,$B3,$AC
	.dc.b	$3E,$9C,$62,$7E,$83,$61,$D4,$FD,$15,$F8,$0E,$67,$E4,$57,$EA,$76
	.dc.b	$34,$91,$FB,$08,$CB,$F4,$B1,$F1,$3F,$47,$CD,$E9,$3F,$EC,$89,$CB
	.dc.b	$C2,$7F,$D2,$FF,$8A,$7F,$AA,$5E,$59,$FF,$A7,$E3,$11,$AF,$23,$D5
	.dc.b	$F5,$CF,$B6,$7D,$27,$B7,$CA,$61,$C3,$0F,$F5,$D4,$63,$0F,$E0,$ED
	.dc.b	$8F,$F4,$FB,$5F,$F4,$F4,$F8,$33,$FF,$5B,$FD,$43,$EA,$7E,$57,$A0
	.dc.b	$8C,$5F,$A0,$61,$CE,$FE,$A3,$E4,$FB,$EF,$36,$26,$9E,$CD,$3F,$CE
	.dc.b	$77,$CE,$B8,$3E,$6A,$E0,$DA,$F5,$C6,$23,$D3,$1E,$6C,$3F,$D4,$71
	.dc.b	$E5,$C7,$D1,$E5,$78,$23,$B5,$BD,$FF,$29,$23,$D9,$BF,$CF,$F2,$71
	.dc.b	$1F,$1A,$7C,$24,$7C,$CF,$E3,$F5,$7D,$9B,$DC,$C3,$6B,$C3,$5F,$67
	.dc.b	$B5,$F5,$94,$93,$E2,$E1,$79,$38,$8F,$F6,$A5,$70,$7E,$DB,$5C,$D7
	.dc.b	$E0,$23,$19,$7D,$CE,$4F,$8C,$73,$7E,$BC,$79,$6B,$FD,$D6,$D8,$CE
	.dc.b	$7F,$5F,$E2,$BD,$0F,$52,$57,$89,$F0,$7C,$54,$8F,$2B,$5F,$4F,$DC
	.dc.b	$79,$48,$FA,$9F,$47,$C3,$C8,$F6,$B5,$F8,$3D,$3E,$D2,$3F,$09,$FE
	.dc.b	$C5,$FA,$33,$E8,$7F,$7C,$D9,$FC,$C0,$70,$FF,$B2,$F7,$FF,$74,$3A
	.dc.b	$8C,$BF,$CC,$0F,$E6,$2E,$1C,$35,$CC,$F8,$8F,$C2,$7E,$E7,$FA,$88
	.dc.b	$FE,$D5,$DB,$FC,$26,$31,$B7,$F0,$AB,$F1,$BF,$C7,$79,$BF,$80,$7C
	.dc.b	$5F,$E7,$7E,$DA,$63,$F8,$58,$FD,$51,$87,$A8,$EB,$FD,$C4,$EA,$3F
	.dc.b	$49,$D8,$FB,$FF,$D4,$12,$3F,$21,$F1,$B5,$E7,$7F,$C0,$67,$19,$79
	.dc.b	$7A,$57,$FC,$AF,$F7,$6F,$E5,$B8,$7C,$E6,$BE,$40,$E2,$BF,$9D,$79
	.dc.b	$0F,$CE,$3F,$D9,$D1,$FC,$B2,$78,$D7,$D1,$38,$F9,$1F,$E1,$F8,$CC
	.dc.b	$7D,$56,$BF,$D9,$9E,$DC,$5F,$DA,$30,$ED,$ED,$FF,$45,$6D,$79,$EF
	.dc.b	$F8,$A7,$C7,$7F,$0B,$F9,$09,$3F,$B3,$8C,$BF,$2A,$32,$FD,$0E,$3F
	.dc.b	$E6,$D3,$FC,$4F,$C6,$93,$F4,$1F,$E4,$1F,$75,$88,$F9,$BC,$7F,$C6
	.dc.b	$F6,$FE,$63,$5A,$9F,$CB,$FC,$03,$E4,$FF,$AC,$36,$33,$1F,$D6,$FE
	.dc.b	$ED,$ED,$EA,$3F,$C8,$95,$BA,$F9,$41,$C3,$F3,$5A,$FE,$7D,$E1,$87
	.dc.b	$F5,$EB,$FD,$2C,$FD,$B6,$C7,$F6,$29,$DF,$0F,$F8,$5B,$9F,$FA,$47
	.dc.b	$D8,$7F,$35,$26,$1F,$C7,$7E,$5A,$BF,$0A,$2E,$1F,$8D,$9F,$AD,$F9
	.dc.b	$E3,$ED,$6F,$C5,$4C,$3F,$A9,$73,$2B,$51,$86,$BF,$A4,$8D,$4F,$18
	.dc.b	$C4,$EB,$0F,$57,$57,$CE,$24,$FF,$08,$D9,$F3,$38,$8F,$97,$BF,$E8
	.dc.b	$9F,$F3,$AB,$5C,$D7,$D2,$3E,$B7,$D8,$48,$FB,$B1,$F9,$3F,$CF,$7F
	.dc.b	$C0,$D8,$FE,$91,$7D,$06,$4D,$7B,$32,$FE,$0C,$E9,$B3,$97,$F2,$9E
	.dc.b	$03,$EA,$FF,$B3,$78,$F2,$7F,$A7,$47,$F7,$33,$0F,$E5,$23,$0F,$F2
	.dc.b	$8F,$B3,$7D,$97,$FA,$3D,$B1,$FE,$94,$6B,$F8,$E6,$BF,$B9,$73,$9F
	.dc.b	$F8,$E7,$E1,$1F,$66,$DF,$FC,$71,$18,$7F,$C6,$D5,$C3,$AB,$FC,$C3
	.dc.b	$ED,$B9,$1F,$37,$FE,$23,$6E,$7F,$E0,$B6,$BF,$C9,$31,$E8,$E6,$7F
	.dc.b	$C4,$B3,$A8,$FF,$14,$B5,$EE,$D7,$FC,$3E,$BE,$87,$89,$23,$FC,$5B
	.dc.b	$19,$6C,$7F,$85,$7D,$99,$1E,$58,$FF,$29,$7C,$C7,$F8,$1E,$FF,$CF
	.dc.b	$BF,$12,$F8,$2F,$EA,$13,$9B,$4E,$5F,$F4,$1C,$66,$31,$0F,$9D,$7C
	.dc.b	$F7,$F5,$99,$CD,$6E,$FA,$FB,$5F,$22,$72,$D8,$FE,$C1,$7C,$B0,$F2
	.dc.b	$C7,$0C,$6C,$78,$AB,$83,$63,$2F,$8F,$1F,$15,$71,$8F,$5B,$70,$F0
	.dc.b	$D7,$0C,$3C,$15,$A3,$A1,$46,$3E,$B7,$35,$F0,$A7,$8B,$D2,$38,$4E
	.dc.b	$DE,$38,$FA,$1F,$31,$E8,$8D,$12,$35,$5D,$30,$F4,$12,$BC,$2F,$C8
	.dc.b	$C7,$DF,$18,$21,$E0,$C7,$DD,$1A,$8C,$1E,$6B,$F2,$61,$E7,$C3,$C1
	.dc.b	$3C,$A3,$C9,$7C,$A3,$D1,$5C,$09,$19,$8E,$8E,$88,$FC,$BA,$EA,$F8
	.dc.b	$B1,$F9,$04,$C7,$D9,$1C,$23,$0F,$B2,$78,$E1,$E4,$8E,$E7,$5C,$7D
	.dc.b	$D1,$9A,$E3,$3F,$40,$DB,$88,$F8,$13,$C3,$1F,$D9,$47,$E8,$EB,$FC
	.dc.b	$08,$D7,$F1,$86,$3E,$71,$F3,$DF,$D8,$23,$F0,$F1,$87,$FB,$1C,$A3
	.dc.b	$11,$C6,$78,$4E,$5F,$C2,$4F,$1C,$3F,$5F,$F3,$AF,$9E,$FF,$6A,$47
	.dc.b	$EB,$A7,$5A,$7E,$E9,$B8,$FE,$E1,$AA,$F9,$71,$F6,$C6,$4F,$E6,$2B
	.dc.b	$75,$A8,$C4,$EB,$0F,$03,$FB,$09,$D4,$62,$76,$47,$D3,$8F,$A3,$1F
	.dc.b	$35,$F3,$6D,$7E,$BE,$F9,$C6,$2F,$A1,$1F,$F3,$34,$6A,$7D,$D5,$CD
	.dc.b	$A7,$3B,$F0,$38,$BE,$4D,$BE,$73,$A8,$FB,$35,$CD,$A7,$3C,$71,$97
	.dc.b	$D7,$8F,$75,$F2,$9C,$E5,$F4,$2B,$58,$FC,$29,$CE,$1D,$72,$F5,$63
	.dc.b	$DA,$BC,$AD,$C7,$A3,$D8,$F4,$10,$FB,$99,$7A,$71,$EF,$3F,$24,$92
	.dc.b	$37,$8F,$89,$E8,$60,$B3,$C9,$ED,$87,$9A,$31,$C7,$FD,$97,$F5,$15
	.dc.b	$FF,$B8,$CF,$CF,$1F,$D9,$5C,$DF,$F9,$FA,$32,$FF,$A2,$A7,$94,$7F
	.dc.b	$5C,$9D,$CF,$F6,$66,$BF,$D9,$9A,$FF,$66,$C6,$D7,$F7,$76,$F1,$FE
	.dc.b	$66,$7F,$B7,$C6,$0F,$1F,$47,$92,$B8,$61,$F1,$63,$FC,$11,$F6,$3F
	.dc.b	$E1,$28,$FE,$FD,$5C,$1F,$F5,$B7,$D6,$7F,$99,$DE,$07,$CD,$E7,$57
	.dc.b	$1C,$7F,$D2,$95,$B7,$FC,$0D,$19,$C3,$DF,$39,$9F,$F4,$93,$5E,$FE
	.dc.b	$67,$C0,$51,$95,$E9,$3C,$0E,$D7,$F4,$3C,$DF,$AB,$CD,$E9,$BD,$D6
	.dc.b	$B4,$FF,$35,$09,$D6,$1C,$F2,$E1,$87,$2A,$F7,$E1,$CA,$73,$1C,$4F
	.dc.b	$F2,$3F,$D7,$77,$F2,$DE,$2F,$F8,$1A,$72,$7F,$93,$3A,$B5,$0A,$B1
	.dc.b	$86,$CF,$FB,$B3,$E5,$3F,$D9,$8F,$8B,$A9,$FA,$0B,$CD,$F7,$68,$BD
	.dc.b	$BF,$ED,$8A,$DC,$EA,$73,$3F,$EC,$16,$BF,$EF,$36,$BF,$A9,$6B,$D7
	.dc.b	$DB,$E4,$7C,$DA,$FE,$6D,$F0,$18,$C7,$FD,$12,$BE,$7B,$E2,$7E,$01
	.dc.b	$FA,$89,$CC,$61,$F9,$13,$B7,$FD,$23,$3B,$C3,$FD,$1D,$39,$7F,$D2
	.dc.b	$11,$9E,$5F,$C0,$53,$E1,$BE,$A9,$F7,$39,$9F,$4D,$89,$FD,$C4,$DC
	.dc.b	$3F,$CF,$73,$6F,$F6,$7A,$FE,$5E,$FF,$4D,$FB,$B7,$CC,$9C,$F9,$9F
	.dc.b	$8D,$51,$99,$F9,$53,$F9,$9A,$FD,$26,$1F,$3E,$30,$49,$FB,$CF,$CB
	.dc.b	$E1,$7D,$D7,$D7,$8F,$C7,$FC,$4C,$AF,$C9,$F5,$3F,$75,$FF,$3B,$C7
	.dc.b	$F1,$39,$D6,$3F,$C4,$5F,$CE,$CE,$A3,$13,$CA,$1F,$2F,$F9,$CF,$8B
	.dc.b	$DB,$67,$4D,$AF,$F7,$A5,$70,$9F,$E4,$55,$F9,$2E,$9F,$9F,$1F,$EC
	.dc.b	$AF,$CE,$FE,$A3,$FA,$AB,$FF,$34,$F0,$FE,$9E,$73,$0E,$1F,$E4,$77
	.dc.b	$CA,$3F,$DD,$AD,$7F,$9E,$1F,$CF,$5C,$38,$6B,$9A,$F2,$1D,$FB,$EE
	.dc.b	$1F,$CE,$7E,$33,$F7,$67,$C6,$F3,$EC,$64,$97,$F0,$23,$2F,$C2,$F2
	.dc.b	$FF,$81,$FD,$07,$CE,$6D,$7C,$D6,$CE,$E3,$FA,$44,$FE,$E3,$F4,$B6
	.dc.b	$1F,$8B,$E7,$7C,$B6,$63,$0D,$BF,$F1,$85,$70,$6C,$FF,$27,$8C,$61
	.dc.b	$F1,$8F,$EC,$4E,$36,$FE,$9E,$DA,$FE,$CD,$E1,$BB,$5F,$19,$DF,$0F
	.dc.b	$EB,$1F,$39,$FE,$CB,$BE,$90,$F3,$13,$98,$F0,$46,$3D,$7F,$EA,$F3
	.dc.b	$99,$FB,$33,$C3,$AF,$24,$BE,$B1,$96,$D3,$E6,$39,$92,$32,$F6,$DE
	.dc.b	$47,$06,$E1,$E4,$A7,$CA,$79,$09,$5D,$9B,$3F,$01,$FD,$4F,$BD,$FD
	.dc.b	$2F,$1F,$7C,$5B,$7C,$5F,$F4,$4F,$97,$C3,$71,$FE,$93,$7F,$1A,$3E
	.dc.b	$19,$23,$DA,$DC,$7E,$7B,$FB,$8F,$80,$E7,$3C,$B0,$7B,$FF,$C4,$7F
	.dc.b	$76,$FE,$85,$39,$8C,$79,$BE,$57,$E8,$92,$3E,$5B,$71,$FB,$AF,$F0
	.dc.b	$6B,$E1,$3E,$8F,$C8,$BF,$97,$3F,$DA,$39,$3E,$14,$62,$DE,$97,$F8
	.dc.b	$92,$73,$5F,$E1,$6A,$FE,$75,$5F,$80,$6B,$FE,$35,$D5,$18,$8F,$C6
	.dc.b	$3D,$63,$33,$FC,$6B,$EA,$3A,$BE,$5F,$77,$FB,$9F,$CB,$F9,$1B,$19
	.dc.b	$3F,$17,$7F,$E3,$2C,$3C,$53,$E5,$24,$EE,$3F,$8E,$FC,$FF,$38,$FE
	.dc.b	$E5,$F4,$FF,$DA,$7E,$53,$FB,$53,$63,$27,$FA,$8F,$CF,$47,$F4,$12
	.dc.b	$46,$5E,$2F,$F8,$33,$B5,$F3,$FF,$D3,$F1,$FE,$E1,$5F,$DF,$1F,$F2
	.dc.b	$D7,$C1,$7F,$96,$FE,$85,$F4,$3F,$DD,$FE,$8F,$54,$AD,$4E,$72,$FE
	.dc.b	$61,$3C,$8F,$E7,$F1,$96,$9F,$C1,$E7,$FC,$D1,$E8,$76,$8F,$F0,$15
	.dc.b	$FF,$97,$23,$11,$FC,$CA,$77,$5C,$E7,$33,$DD,$DE,$8B,$E5,$5F,$C2
	.dc.b	$AF,$CF,$7D,$B0,$AF,$51,$8F,$2B,$FF,$36,$BD,$A3,$E0,$1A,$BF,$64
	.dc.b	$66,$F9,$E1,$E1,$C3,$AD,$FC,$3A,$E9,$18,$8E,$71,$87,$A1,$F1,$4A
	.dc.b	$7F,$8B,$F0,$F1,$E1,$EB,$8C,$DF,$77,$67,$78,$F2,$9B,$AE,$C7,$9E
	.dc.b	$3E,$99,$D6,$FE,$6B,$E1,$C7,$C8,$6E,$78,$D7,$33,$AC,$F9,$E3,$2D
	.dc.b	$8C,$BF,$34,$FF,$7D,$63,$FE,$FB,$7F,$0C,$60,$FE,$7F,$F3,$CF,$A3
	.dc.b	$7C,$17,$FC,$01,$8C,$36,$33,$3C,$1A,$FF,$53,$97,$0F,$F7,$F8,$EA
	.dc.b	$D7,$4D,$7F,$BF,$C7,$0F,$23,$C1,$3C,$72,$FC,$9F,$D1,$FF,$52,$BF
	.dc.b	$C4,$B5,$FA,$F7,$C5,$B5,$C7,$3F,$F0,$5D,$6E,$F9,$F5,$FF,$83,$9F
	.dc.b	$CF,$E1,$C5,$B7,$C7,$0F,$F6,$75,$F1,$C3,$9E,$1C,$5F,$4D,$6E,$98
	.dc.b	$73,$9D,$C7,$A2,$B8,$3F,$ED,$99,$D3,$6B,$95,$F7,$8F,$0D,$70,$7A
	.dc.b	$CF,$7A,$E3,$0A,$E0,$DA,$E1,$3A,$37,$3A,$36,$E6,$BC,$33,$DE,$73
	.dc.b	$1D,$CC,$D7,$AC,$FD,$DD,$71,$C7,$11,$FE,$3F,$DB,$E0,$B4,$F7,$47
	.dc.b	$0A,$E7,$19,$3D,$07,$F9,$2E,$BC,$FE,$1F,$F9,$12,$BF,$DD,$78,$23
	.dc.b	$18,$7D,$E6,$DB,$93,$F9,$8F,$CD,$E2,$3E,$34,$DB,$79,$D4,$C4,$49
	.dc.b	$FF,$2E,$4E,$9B,$39,$7F,$E0,$07,$D1,$FF,$84,$1F,$F0,$1B,$5C,$F8
	.dc.b	$DE,$4F,$63,$FD,$B9,$E7,$79,$BC,$05,$EF,$6F,$F6,$F7,$CD,$7F,$88
	.dc.b	$F1,$FF,$6A,$F8,$68,$C4,$74,$3B,$E5,$F0,$27,$5E,$5D,$AD,$DB,$87
	.dc.b	$FB,$A6,$3D,$DB,$F3,$5E,$38,$34,$F4,$7B,$CF,$4F,$1A,$79,$31,$FF
	.dc.b	$2D,$B7,$CE,$CB,$F1,$2F,$C5,$12,$3E,$2C,$77,$F8,$F9,$1F,$21,$EF
	.dc.b	$E7,$E2,$3F,$11,$A3,$E8,$FF,$DF,$91,$FE,$F8,$8C,$7A,$FB,$7D,$D1
	.dc.b	$FE,$59,$D5,$87,$E1,$27,$73,$C6,$F9,$BF,$6A,$7E,$A7,$53,$E8,$E9
	.dc.b	$27,$FD,$6B,$BF,$F9,$AD,$FF,$20,$CE,$DF,$89,$C7,$F4,$4E,$0F,$E8
	.dc.b	$79,$3F,$82,$DE,$63,$67,$27,$C3,$A4,$8E,$AF,$D2,$F2,$F1,$1C,$F0
	.dc.b	$FF,$72,$3E,$2F,$81,$CE,$BF,$0B,$1F,$E2,$71,$F8,$5E,$47,$CE,$D3
	.dc.b	$DF,$1F,$03,$E0,$3E,$0F,$BF,$31,$F6,$C5,$7D,$5F,$92,$3E,$77,$F3
	.dc.b	$4D,$8F,$87,$5F,$2B,$B3,$C1,$F3,$B2,$7E,$3C,$7E,$CF,$E8,$FF,$CB
	.dc.b	$F1,$FE,$C9,$6B,$FC,$0D,$BB,$7D,$CF,$13,$FD,$35,$E8,$F1,$F5,$BE
	.dc.b	$E7,$D2,$7E,$7B,$85,$8F,$FB,$3E,$30,$E3,$1E,$73,$FE,$CE,$6C,$6F
	.dc.b	$C1,$E6,$75,$7E,$FF,$D2,$7E,$AF,$DF,$91,$82,$1F,$C7,$7C,$4F,$DB
	.dc.b	$46,$A3,$F9,$07,$8A,$ED,$5F,$ED,$7E,$32,$57,$EA,$DF,$D7,$FC,$E6
	.dc.b	$23,$A7,$8C,$9C,$71,$7E,$7B,$1E,$97,$FC,$77,$E8,$DE,$65,$26,$3F
	.dc.b	$E5,$3F,$A4,$FF,$1E,$E1,$FA,$68,$57,$07,$5C,$9F,$CA,$B9,$56,$8E
	.dc.b	$73,$FC,$DE,$B8,$3F,$CF,$6B,$73,$C6,$B8,$4F,$D4,$BF,$05,$F0,$7D
	.dc.b	$5B,$3E,$93,$F8,$C7,$C1,$4B,$FD,$7C,$7F,$12,$30,$FC,$48,$C3,$FC
	.dc.b	$4F,$CA,$FE,$40,$47,$F6,$FE,$05,$7E,$5C,$5C,$3C,$33,$FD,$17,$E8
	.dc.b	$1F,$29,$49,$7E,$9E,$9F,$C6,$AF,$98,$7D,$26,$A3,$FA,$D4,$7C,$06
	.dc.b	$F0,$67,$2F,$A1,$8F,$69,$E8,$7E,$62,$B8,$4F,$D5,$6E,$1E,$0A,$E1
	.dc.b	$5F,$84,$F0,$BF,$9F,$11,$FE,$2B,$E0,$57,$47,$F3,$78,$FE,$77,$1F
	.dc.b	$77,$E8,$DF,$49,$7E,$1A,$5F,$FA,$66,$72,$D9,$CB,$F7,$E1,$3C,$E7
	.dc.b	$E5,$74,$4A,$F6,$8C,$51,$FD,$CF,$37,$D2,$6E,$BE,$4F,$C5,$E4,$C6
	.dc.b	$DF,$22,$F1,$3D,$2D,$5A,$DB,$F8,$0F,$4B,$E6,$7F,$C0,$D7,$FE,$B3
	.dc.b	$5F,$17,$F9,$D6,$C6,$7D,$AC,$19,$99,$19,$3F,$A0,$FD,$27,$F6,$DF
	.dc.b	$22,$46,$08,$7E,$E1,$CB,$FC,$0B,$B1,$3F,$E1,$3F,$A3,$7C,$77,$F8
	.dc.b	$63,$E0,$DF,$49,$EC,$C3,$F0,$B1,$8D,$F8,$54,$9C,$D6,$DF,$F1,$75
	.dc.b	$6D,$FA,$37,$D9,$FC,$7F,$A5,$F9,$9F,$A4,$FF,$09,$71,$1A,$78,$F0
	.dc.b	$E9,$C9,$FD,$2E,$75,$7F,$E1,$4F,$C4,$DF,$1F,$FE,$11,$6C,$76,$9D
	.dc.b	$BF,$D8,$57,$CA,$7F,$85,$7D,$C9,$3B,$8F,$19,$87,$F9,$0C,$61,$FC
	.dc.b	$5F,$D1,$BD,$BF,$1F,$C5,$7F,$8D,$65,$39,$D7,$FD,$DB,$F0,$9F,$B8
	.dc.b	$8C,$4F,$FB,$AE,$3F,$77,$F4,$AF,$98,$FF,$70,$46,$F1,$EA,$DC,$39
	.dc.b	$DF,$F4,$EC,$39,$E3,$E1,$C7,$94,$E8,$F1,$E1,$C9,$E0,$D7,$37,$B9
	.dc.b	$E3,$1E,$53,$71,$CE,$FB,$E1,$E3,$9D,$0A,$E3,$87,$C3,$AD,$CF,$23
	.dc.b	$6E,$1E,$18,$FC,$07,$53,$C7,$1F,$35,$6E,$FA,$E1,$E1,$AD,$63,$E5
	.dc.b	$BF,$65,$6B,$0E,$77,$B2,$3F,$DD,$A3,$4F,$C2,$3C,$B3,$E0,$39,$E3
	.dc.b	$E5,$8D,$61,$E0,$77,$87,$C9,$9D,$63,$ED,$25,$F2,$AF,$35,$47,$89
	.dc.b	$F2,$B0,$F1,$D7,$09,$D6,$1F,$77,$0E,$B8,$79,$67,$38,$FE,$2A,$73
	.dc.b	$18,$C7,$E0,$9E,$F8,$CB,$E0,$BF,$51,$D5,$E0,$7B,$AB,$E3,$BF,$58
	.dc.b	$CE,$1E,$57,$D1,$5D,$1F,$EA,$A7,$CB,$2C,$7E,$81,$B8,$7E,$D1,$FD
	.dc.b	$27,$C9,$3E,$7F,$8C,$7E,$76,$30,$E2,$91,$88,$EA,$D7,$9C,$FE,$5E
	.dc.b	$38,$E3,$FB,$5F,$A5,$7D,$27,$F9,$DA,$3F,$2D,$3A,$CF,$8E,$1F,$96
	.dc.b	$8C,$3F,$E3,$CB,$E3,$E0,$F5,$C6,$4E,$DF,$D8,$4E,$A3,$04,$C3,$BC
	.dc.b	$EB,$1F,$16,$3E,$09,$CD,$6E,$3F,$BF,$DF,$39,$C9,$23,$81,$AA,$E7
	.dc.b	$5F,$26,$FC,$B5,$B8,$F5,$4E,$9F,$E2,$F8,$FB,$9B,$5C,$2F,$A6,$36
	.dc.b	$7A,$1E,$53,$9B,$97,$89,$91,$8F,$A6,$7C,$F1,$CB,$1F,$23,$AC,$7C
	.dc.b	$99,$7D,$EC,$7A,$36,$7F,$87,$E1,$CD,$E5,$87,$47,$C3,$87,$DF,$C3
	.dc.b	$9E,$5E,$DC,$39,$4F,$83,$0E,$57,$C3,$0E,$55,$A7,$E0,$C7,$AE,$30
	.dc.b	$F8,$0F,$82,$66,$FB,$1F,$2E,$BA,$CF,$09,$EC,$71,$3E,$06,$5F,$37
	.dc.b	$9B,$FC,$79,$F4,$BF,$EC,$AE,$6F,$F6,$5F,$B5,$FE,$E9,$CB,$17,$FE
	.dc.b	$8E,$BD,$BF,$EF,$63,$F4,$A7,$F6,$1B,$DE,$0B,$E5,$39,$38,$5E,$AF
	.dc.b	$FC,$99,$AF,$0A,$73,$3E,$B7,$5C,$FF,$9D,$D3,$CB,$39,$CB,$FB,$86
	.dc.b	$1A,$3C,$3D,$6F,$A6,$FF,$07,$CE,$8F,$F1,$AF,$D3,$7F,$98,$1F,$F0
	.dc.b	$7C,$6B,$66,$3F,$E8,$7A,$D4,$7F,$83,$A7,$53,$9F,$47,$BC,$62,$FF
	.dc.b	$CC,$F7,$C2,$BF,$DE,$B8,$78,$FB,$FE,$2F,$B3,$FE,$A5,$C7,$E0,$61
	.dc.b	$FD,$F4,$FF,$77,$B8,$E8,$DF,$87,$CE,$7F,$D7,$FC,$27,$86,$BF,$E2
	.dc.b	$FB,$E1,$1F,$E4,$1F,$A6,$FA,$7E,$4F,$F9,$17,$B2,$F9,$29,$B1,$86
	.dc.b	$D7,$13,$FC,$95,$1F,$73,$07,$CB,$FF,$AD,$67,$FC,$F7,$DC,$FA,$4F
	.dc.b	$F3,$2C,$7F,$98,$A7,$55,$F3,$23,$FD,$AF,$5B,$F4,$3A,$56,$B8,$FD
	.dc.b	$97,$FC,$D7,$0E,$6F,$13,$FD,$BF,$3B,$C3,$9C,$EA,$BF,$D7,$D3,$9A
	.dc.b	$DF,$BB,$F9,$4C,$7F,$99,$FC,$EF,$82,$3E,$2A,$FA,$6F,$8F,$F4,$DF
	.dc.b	$37,$3F,$F4,$BF,$02,$B5,$18,$9D,$63,$FE,$C3,$9C,$BF,$EC,$58,$CE
	.dc.b	$7F,$EF,$B9,$CB,$FE,$FD,$8C,$D7,$DC,$1F,$17,$F8,$43,$F0,$3F,$47
	.dc.b	$63,$F6,$11,$70,$FF,$37,$4D,$BF,$83,$D1,$FB,$25,$F4,$7F,$0E,$BE
	.dc.b	$CC,$E6,$31,$BF,$E6,$67,$F4,$31,$89,$D6,$DF,$59,$99,$7E,$03,$81
	.dc.b	$F4,$D8,$C3,$EB,$73,$FE,$57,$E3,$BF,$3A,$7E,$83,$E9,$BF,$CE,$18
	.dc.b	$7F,$09,$8C,$B4,$FF,$62,$F6,$27,$77,$D2,$55,$CF,$E7,$FF,$53,$F1
	.dc.b	$F6,$7F,$8A,$36,$B8,$37,$83,$F8,$EC,$A3,$F3,$9D,$2E,$2F,$D3,$FD
	.dc.b	$37,$D2,$AF,$F5,$DC,$E5,$FC,$1C,$67,$0F,$CA,$1F,$E6,$BA,$E3,$C9
	.dc.b	$D1,$B5,$C9,$C3,$CB,$BE,$D7,$B7,$E4,$3F,$81,$9F,$1F,$D2,$B5,$FC
	.dc.b	$BE,$2C,$FE,$8F,$83,$E3,$56,$35,$FD,$4C,$7C,$9F,$3B,$F8,$6F,$CD
	.dc.b	$7D,$16,$E3,$F3,$B8,$91,$8C,$7A,$46,$27,$F9,$F5,$7F,$01,$FA,$66
	.dc.b	$7F,$89,$F9,$BC,$36,$BF,$DC,$33,$A7,$FD,$FF,$9F,$E4,$AB,$71,$8E
	.dc.b	$A4,$63,$2F,$97,$7C,$5F,$D6,$7C,$86,$70,$FE,$0B,$E7,$3E,$7B,$FD
	.dc.b	$C7,$92,$FF,$23,$EA,$3E,$4B,$FA,$75,$6A,$73,$18,$EE,$FE,$A5,$F9
	.dc.b	$2B,$E9,$31,$8F,$FB,$22,$30,$FF,$B2,$B0,$E8,$DA,$EB,$7C,$30,$FE
	.dc.b	$2D,$97,$97,$6F,$EB,$99,$FB,$B0,$E5,$18,$3C,$BF,$37,$E6,$6C,$68
	.dc.b	$F0,$46,$2F,$EA,$FC,$1D,$DB,$E1,$E8,$F8,$FF,$E9,$7F,$3D,$F0,$1B
	.dc.b	$8F,$A5,$B8,$79,$AB,$84,$FF,$7B,$AE,$0E,$CF,$C3,$7C,$2E,$2B,$87
	.dc.b	$C2,$FF,$0D,$FA,$7F,$E6,$15,$FE,$05,$C1,$39,$F3,$FE,$8B,$67,$E5
	.dc.b	$36,$F9,$C6,$0F,$CA,$57,$07,$FD,$F3,$F4,$FF,$C9,$A7,$FC,$19,$1F
	.dc.b	$CC,$CC,$9F,$DA,$BC,$DF,$F0,$64,$EF,$C8,$FE,$CD,$18,$9D,$1C,$09
	.dc.b	$5F,$A0,$F8,$3D,$47,$83,$DD,$FE,$EB,$F3,$BD,$1B,$19,$3F,$C4,$51
	.dc.b	$96,$99,$AF,$C1,$E3,$FE,$26,$7F,$5B,$F4,$5B,$7F,$84,$FA,$1F,$8E
	.dc.b	$6F,$C8,$FF,$64,$ED,$FE,$CA,$DC,$3F,$B6,$36,$77,$1F,$DB,$E7,$1C
	.dc.b	$BF,$DA,$BD,$37,$D3,$FF,$3D,$BF,$EF,$35,$FD,$ED,$FD,$47,$5F,$EA
	.dc.b	$7E,$95,$F4,$BF,$AB,$9F,$EF,$FF,$4F,$B5,$5B,$9D,$6A,$AD,$4E,$6B
	.dc.b	$FB,$BC,$61,$5F,$67,$3A,$7E,$E7,$4F,$F0,$0C,$78,$6F,$93,$E2,$9A
	.dc.b	$FA,$0E,$78,$F8,$B0,$FF,$04,$E1,$D6,$79,$1A,$3B,$94,$CB,$BC,$7D
	.dc.b	$57,$DA,$7D,$E6,$E3,$2F,$13,$C5,$3C,$6B,$C3,$7F,$EC,$CC,$3D,$37
	.dc.b	$C2,$BC,$97,$C2,$BD,$18,$77,$BD,$E1,$CA,$BF,$DB,$B1,$89,$D4,$78
	.dc.b	$E7,$4F,$AA,$B9,$BE,$53,$D3,$7E,$5A,$E2,$D8,$D5,$F5,$69,$D6,$77
	.dc.b	$5D,$2B,$D3,$3D,$23,$07,$BD,$FF,$75,$9F,$54,$B5,$87,$07,$E3,$7E
	.dc.b	$89,$F4,$EF,$80,$E3,$18,$6C,$66,$7F,$DE,$F3,$96,$F0,$FF,$9E,$1B
	.dc.b	$AF,$28,$C4,$7E,$6C,$4A,$33,$1E,$37,$96,$5F,$89,$FA,$7F,$E9,$D3
	.dc.b	$F9,$26,$BF,$89,$AE,$57,$F9,$66,$C7,$E3,$67,$75,$CF,$A7,$F5,$8F
	.dc.b	$E5,$6B,$6D,$C3,$C3,$A7,$EC,$6F,$8B,$6F,$8B,$E7,$C7,$C1,$5C,$31
	.dc.b	$FE,$C5,$7C,$1B,$7C,$1F,$EC,$D8,$F8,$30,$E5,$1E,$79,$D2,$AE,$4F
	.dc.b	$C8,$BE,$CD,$C3,$E2,$57,$0A,$DE,$1D,$6B,$BD,$F0,$9E,$66,$9E,$73
	.dc.b	$B9,$F0,$57,$33,$C7,$19,$35,$5C,$A3,$35,$D7,$C0,$FE,$31,$87,$F0
	.dc.b	$AA,$46,$36,$F9,$8D,$C5,$C7,$FC,$9B,$E7,$F1,$1F,$06,$6D,$F2,$8F
	.dc.b	$E2,$D7,$C9,$FF,$42,$F2,$2F,$A7,$1B,$F0,$7F,$C8,$67,$FE,$9C,$8C
	.dc.b	$5E,$A7,$38,$7F,$86,$73,$7E,$0E,$FC,$2B,$57,$0F,$09,$FE,$B2,$BE
	.dc.b	$2F,$AA,$78,$3A,$C3,$BB,$FE,$9A,$FB,$6F,$E6,$CD,$E7,$E8,$FF,$86
	.dc.b	$3E,$AD,$F4,$3F,$17,$CA,$46,$27,$C3,$18,$9F,$95,$EA,$BE,$21,$EA
	.dc.b	$71,$C3,$86,$1D,$6F,$07,$0A,$DC,$74,$9D,$F0,$FF,$B0,$BC,$57,$D2
	.dc.b	$FF,$91,$AB,$E0,$C6,$3C,$EE,$71,$E4,$FA,$97,$C3,$FA,$5B,$19,$8F
	.dc.b	$3F,$B1,$FC,$42,$30,$43,$FC,$73,$3C,$1F,$F3,$1F,$83,$CE,$BF,$11
	.dc.b	$87,$F6,$2F,$42,$F8,$9F,$E3,$1B,$E5,$A7,$F9,$5B,$C1,$7D,$57,$CB
	.dc.b	$6C,$FC,$A6,$BF,$E7,$BE,$0F,$F6,$73,$99,$E0,$47,$FD,$4B,$1A,$7F
	.dc.b	$C9,$F3,$CA,$BE,$46,$EF,$4B,$CE,$DA,$E8,$7C,$7F,$4F,$E3,$11,$FF
	.dc.b	$42,$7D,$65,$7E,$1C,$5B,$FF,$46,$F0,$AB,$1D,$FF,$E7,$1D,$FF,$C0
	.dc.b	$BC,$BF,$10,$CE,$2E,$8F,$E6,$3F,$09,$F0,$FE,$0E,$4F,$C3,$8A,$FB
	.dc.b	$22,$E7,$F2,$E7,$EE,$7D,$1C,$AF,$C0,$3A,$CB,$FD,$05,$83,$9D,$F3
	.dc.b	$3F,$83,$25,$FF,$A8,$7E,$DF,$F3,$D8,$FF,$84,$32,$FF,$46,$B7,$1E
	.dc.b	$6D,$AF,$56,$1D,$1F,$D1,$C3,$F0,$8E,$64,$8F,$CB,$FE,$28,$FA,$1F
	.dc.b	$E1,$BF,$46,$F9,$3A,$4A,$FD,$E8,$F9,$87,$D3,$E6,$3F,$D9,$F5,$C5
	.dc.b	$DB,$5F,$F6,$9C,$F0,$73,$5C,$75,$FF,$16,$D6,$DF,$F1,$7D,$6E,$BF
	.dc.b	$D8,$3F,$29,$F9,$CE,$29,$1F,$EC,$78,$FD,$49,$87,$FD,$5F,$18,$7F
	.dc.b	$6B,$D2,$8F,$C8,$92,$3F,$33,$F4,$F5,$61,$F9,$89,$C3,$FE,$DA,$ED
	.dc.b	$57,$F3,$78,$7E,$EF,$E7,$5C,$7F,$E4,$98,$C3,$76,$FE,$7D,$E7,$7E
	.dc.b	$9F,$2E,$F3,$97,$FD,$AB,$5C,$1F,$1B,$67,$85,$F6,$9C,$DF,$8E,$B8
	.dc.b	$4F,$92,$3B,$FB,$DF,$CD,$2D,$F5,$0F,$A8,$F1,$AF,$82,$A6,$22,$65
	.dc.b	$FC,$FA,$B8,$CE,$A3,$15,$CD,$FE,$47,$E2,$FF,$AC,$1B,$F0,$5F,$C5
	.dc.b	$0F,$E3,$FF,$25,$63,$24,$3F,$DF,$EF,$E5,$39,$E4,$7E,$31,$FE,$C7
	.dc.b	$CA,$FA,$8C,$CF,$E4,$A6,$EF,$F8,$A9,$BC,$3F,$E8,$9E,$7F,$D0,$5F
	.dc.b	$33,$F3,$B8,$73,$7D,$3D,$DF,$CA,$88,$FF,$5E,$F3,$DE,$2B,$E8,$33
	.dc.b	$3F,$EF,$EA,$B9,$FF,$14,$9F,$BF,$F3,$7F,$DD,$EB,$EF,$9F,$DF,$97
	.dc.b	$C4,$68,$DF,$B9,$27,$E8,$4E,$9B,$39,$7F,$63,$F1,$3F,$BA,$23,$F5
	.dc.b	$FE,$69,$F2,$98,$8D,$9F,$B6,$9F,$C9,$CE,$A3,$04,$36,$7A,$3A,$11
	.dc.b	$B8,$FE,$8D,$F5,$1F,$CF,$39,$A4,$7F,$44,$8F,$ED,$86,$1F,$83,$E6
	.dc.b	$7F,$6A,$9C,$C7,$CD,$E0,$FF,$03,$BA,$E8,$7C,$E7,$F5,$C3,$FB,$3F
	.dc.b	$BD,$27,$F1,$F1,$97,$94,$7F,$8A,$3E,$9D,$F5,$14,$93,$FE,$18,$F8
	.dc.b	$8F,$F6,$4F,$6E,$23,$EA,$8A,$E2,$FF,$22,$6B,$F6,$A7,$8B,$FC,$C3
	.dc.b	$E3,$3F,$9C,$91,$ED,$F5,$1F,$EC,$1E,$8F,$E7,$73,$99,$FC,$1F,$A2
	.dc.b	$F3,$D3,$FE,$43,$C1,$87,$1E,$FF,$E8,$3E,$0F,$F4,$62,$63,$F2,$7E
	.dc.b	$8A,$BD,$31,$FC,$FB,$67,$FC,$1F,$3F,$E4,$CF,$A5,$7C,$DB,$E9,$FF
	.dc.b	$DE,$3F,$37,$FE,$EF,$9C,$C6,$30,$08,$CB,$FE,$F2,$9F,$F7,$B7,$D4
	.dc.b	$38,$BF,$DD,$11,$CF,$0E,$D5,$B3,$9D,$F2,$E6,$56,$F1,$F0,$53,$1E
	.dc.b	$8E,$2F,$85,$6E,$3A,$1E,$5C,$3B,$E1,$D2,$3D,$37,$C9,$DD,$F2,$7C
	.dc.b	$75,$C2,$FA,$5F,$19,$F1,$63,$EC,$77,$8F,$8E,$F8,$46,$0F,$1E,$3D
	.dc.b	$EF,$96,$3D,$E7,$CD,$7C,$4F,$7D,$6F,$1F,$09,$27,$A5,$F8,$E7,$51
	.dc.b	$89,$D3,$D0,$F0,$E3,$E7,$71,$5E,$2C,$7C,$F1,$8A,$FB,$35,$BC,$7C
	.dc.b	$C4,$AD,$E1,$E0,$BE,$07,$84,$87,$C6,$3D,$87,$CE,$8E,$F3,$F0,$67
	.dc.b	$93,$D7,$1D,$E3,$ED,$C3,$C5,$7C,$09,$3F,$1A,$32,$F0,$C7,$E0,$E1
	.dc.b	$D2,$B7,$8F,$86,$B8,$3B,$3C,$F8,$F8,$F0,$E7,$1B,$26,$3E,$37,$D6
	.dc.b	$7F,$BE,$67,$F4,$2D,$BF,$CE,$B6,$FF,$DF,$CF,$E7,$BE,$53,$FC,$F4
	.dc.b	$FE,$B5,$AE,$7E,$07,$6F,$EB,$63,$2D,$8F,$D0,$47,$FA,$2B,$0F,$CA
	.dc.b	$7D,$43,$5F,$F0,$24,$7E,$3A,$75,$AF,$7B,$E2,$7E,$B2,$F8,$70,$63
	.dc.b	$1F,$D1,$DF,$07,$C8,$7E,$B2,$76,$EA,$B8,$61,$DD,$E9,$8F,$8B,$1F
	.dc.b	$05,$6C,$95,$E8,$9D,$47,$68,$E4,$EA,$7F,$87,$AB,$F0,$CE,$89,$3D
	.dc.b	$9E,$98,$F9,$23,$04,$AF,$E3,$35,$CA,$31,$1F,$34,$F0,$1A,$BF,$2B
	.dc.b	$9B,$EE,$F9,$E7,$CB,$5C,$E3,$04,$38,$46,$AB,$9E,$8C,$BE,$75,$68
	.dc.b	$E1,$7C,$DC,$12,$78,$65,$F3,$B2,$F7,$4F,$79,$CB,$F5,$F0,$EF,$87
	.dc.b	$69,$CE,$5E,$AA,$F9,$95,$AC,$39,$C7,$83,$1F,$15,$F0,$CB,$E3,$B6
	.dc.b	$7E,$36,$5C,$23,$11,$C7,$0E,$8F,$F3,$FE,$47,$E1,$2F,$97,$FF,$2D
	.dc.b	$73,$FF,$96,$E7,$31,$8C,$7F,$A7,$F9,$3F,$E3,$0F,$A3,$C6,$3F,$EB
	.dc.b	$9C,$79,$57,$F5,$4C,$79,$46,$27,$FB,$6B,$63,$FD,$8F,$5C,$B1,$EF
	.dc.b	$7C,$4F,$F5,$BD,$3E,$1B,$4D,$8F,$F0,$47,$D3,$FF,$B6,$DE,$18,$F8
	.dc.b	$EB,$64,$8D,$FC,$25,$9F,$F0,$6F,$D3,$FF,$8D,$27,$FD,$5A,$D8,$FF
	.dc.b	$06,$66,$FA,$7F,$7C,$27,$53,$FE,$81,$8C,$79,$5A,$9C,$DF,$F7,$6E
	.dc.b	$EF,$F4,$DE,$19,$AF,$93,$BF,$F9,$6B,$D5,$F9,$7E,$3B,$DF,$FF,$09
	.dc.b	$63,$FE,$33,$9C,$C6,$33,$FF,$2C,$DF,$1A,$FF,$65,$F8,$1D,$3C,$7F
	.dc.b	$F4,$64,$60,$FF,$34,$7B,$7F,$EC,$D7,$E3,$E5,$FE,$A6,$EF,$F1,$CE
	.dc.b	$71,$FF,$22,$D7,$F8,$50,$E3,$EF,$F9,$9C,$47,$8A,$78,$70,$FF,$A8
	.dc.b	$F0,$E9,$5F,$ED,$8F,$8F,$79,$7F,$29,$B1,$FE,$6C,$DF,$13,$23,$83
	.dc.b	$F8,$EF,$39,$E8,$FB,$33,$FF,$22,$E1,$C9,$B5,$C5,$FF,$33,$4E,$F2
	.dc.b	$FA,$D5,$B7,$FD,$71,$7D,$3E,$03,$B5,$7F,$89,$F1,$FF,$3F,$C6,$3C
	.dc.b	$6C,$B6,$32,$7F,$A1,$23,$EA,$FC,$24,$8F,$F3,$57,$B7,$73,$FF,$37
	.dc.b	$F2,$AB,$19,$FF,$B2,$79,$7F,$D3,$77,$8C,$3F,$97,$EB,$F9,$17,$E5
	.dc.b	$73,$FE,$C3,$C5,$FE,$9D,$EE,$7C,$B8,$93,$A3,$F4,$73,$A7,$1A,$7C
	.dc.b	$89,$DB,$FE,$A2,$9D,$39,$9F,$C2,$E8,$FA,$2F,$C8,$9F,$8D,$FA,$9B
	.dc.b	$3A,$8B,$5F,$94,$E9,$56,$36,$F9,$F9,$FF,$A2,$B0,$FF,$52,$1F,$75
	.dc.b	$C5,$FF,$A7,$7D,$5F,$F7,$CC,$FF,$B7,$E7,$35,$F6,$27,$35,$FE,$A3
	.dc.b	$9C,$D7,$F9,$93,$93,$FD,$4A,$FF,$16,$B7,$EF,$1F,$B4,$B4,$63,$B7
	.dc.b	$C5,$CF,$BA,$D4,$E6,$FF,$86,$B6,$FC,$15,$BB,$95,$AB,$ED,$83,$1F
	.dc.b	$C8,$4F,$F2,$59,$F1,$74,$A1,$F4,$DF,$AD,$3F,$67,$85,$8F,$DB,$9A
	.dc.b	$2C,$69,$A7,$FB,$97,$1F,$F4,$2B,$F6,$BE,$47,$F8,$7C,$63,$3F,$C9
	.dc.b	$C6,$1F,$E1,$AD,$9E,$07,$EE,$7E,$87,$F7,$07,$E4,$7E,$A5,$E4,$E2
	.dc.b	$74,$4B,$FC,$64,$EA,$31,$BF,$EC,$AF,$85,$6A,$EC,$FF,$48,$E8,$FE
	.dc.b	$7E,$7F,$24,$E6,$E5,$5F,$D2,$A3,$FA,$47,$D1,$FA,$EB,$FA,$C3,$71
	.dc.b	$FF,$52,$F9,$48,$7C,$AF,$0C,$38,$B2,$BE,$46,$36,$7F,$D8,$74,$FA
	.dc.b	$0E,$E7,$F5,$7F,$9E,$6B,$8C,$3F,$A8,$E4,$9C,$C6,$1B,$87,$FB,$39
	.dc.b	$7C,$97,$3B,$E8,$DA,$E0,$DA,$F1,$BF,$D0,$BE,$03,$E0,$B6,$7C,$39
	.dc.b	$74,$6C,$7E,$03,$4D,$C6,$09,$7F,$D2,$B8,$BE,$2C,$63,$6F,$F0,$16
	.dc.b	$6F,$95,$FE,$81,$9F,$C1,$6D,$7B,$5C,$3F,$DC,$BC,$17,$6B,$E0,$B3
	.dc.b	$9F,$C1,$BB,$97,$F2,$1A,$D6,$1C,$CF,$97,$F9,$47,$EF,$A7,$FA,$67
	.dc.b	$83,$FA,$B9,$C9,$FD,$93,$B7,$FC,$15,$5B,$8F,$C0,$B7,$16,$7F,$D4
	.dc.b	$BE,$2F,$75,$0E,$67,$EA,$5C,$4F,$F0,$68,$FC,$5F,$D1,$F3,$EE,$FD
	.dc.b	$C6,$3F,$E1,$98,$C4,$F9,$6B,$80,$E6,$4F,$8D,$B3,$FD,$DF,$D5,$E4
	.dc.b	$C3,$F1,$73,$E9,$CB,$E5,$CE,$6F,$AF,$8C,$F9,$C7,$CC,$FF,$67,$FA
	.dc.b	$FB,$97,$F8,$92,$B5,$39,$3F,$C7,$2E,$23,$55,$C2,$FD,$3E,$2B,$E7
	.dc.b	$7F,$63,$C7,$FC,$5E,$58,$71,$7C,$77,$C6,$75,$39,$7F,$B7,$D7,$2C
	.dc.b	$17,$CA,$F8,$47,$F6,$0E,$EF,$F0,$31,$FE,$01,$F9,$FF,$D4,$E1,$FD
	.dc.b	$EA,$33,$3F,$E0,$D7,$5D,$F8,$C3,$E8,$9F,$E9,$8F,$7F,$BF,$D1,$BE
	.dc.b	$8F,$FB,$6F,$81,$EB,$8C,$57,$FA,$78,$57,$0A,$DE,$36,$75,$35,$F8
	.dc.b	$77,$CE,$BF,$D9,$F3,$9D,$BF,$5B,$4C,$3B,$BE,$53,$15,$D9,$A7,$62
	.dc.b	$9D,$B0,$F0,$3F,$2A,$FB,$61,$E1,$8E,$17,$DB,$0F,$0B,$E0,$8E,$07
	.dc.b	$48,$CE,$1E,$88,$C5,$73,$7C,$D5,$CE,$31,$1C,$A3,$18,$79,$23,$E4
	.dc.b	$9C,$9C,$61,$E6,$6C,$64,$F6,$9C,$EF,$E1,$57,$9E,$39,$BD,$9D,$98
	.dc.b	$8D,$9E,$97,$FD,$F3,$5F,$EF,$A7,$DE,$60,$FB,$C3,$E7,$7F,$D1,$DE
	.dc.b	$CF,$FC,$00,$6C,$69,$AF,$E1,$5A,$FF,$B9,$17,$D1,$FF,$83,$23,$83
	.dc.b	$5E,$0D,$78,$35,$FD,$1C,$72,$DB,$F1,$9F,$47,$FD,$2E,$6F,$87,$FE
	.dc.b	$F2,$CF,$F1,$37,$B9,$FC,$2B,$63,$F1,$B5,$C6,$7F,$12,$D8,$E3,$87
	.dc.b	$9D,$B8,$7E,$D3,$EA,$F8,$E3,$FF,$00,$1B,$86,$74,$F4,$D5,$8F,$4E
	.dc.b	$1C,$E7,$F7,$B0,$C7,$C1,$18,$6C,$70,$AF,$D5,$65,$E9,$AD,$63,$EF
	.dc.b	$6D,$6B,$0E,$CF,$13,$DB,$3B,$7F,$86,$E1,$D9,$FB,$D3,$C1,$B8,$7C
	.dc.b	$C6,$1F,$2A,$35,$87,$8E,$39,$DF,$57,$D8,$D8,$D4,$74,$18,$FC,$86
	.dc.b	$E3,$EA,$69,$E9,$8F,$AF,$F2,$3F,$97,$C8,$D2,$4F,$E6,$0F,$98,$FC
	.dc.b	$EC,$BF,$DA,$52,$F8,$5C,$6F,$F3,$DB,$7F,$12,$EF,$0F,$F4,$98,$BC
	.dc.b	$BF,$C8,$71,$FE,$72,$DC,$3F,$CD,$9C,$CF,$8F,$DD,$F8,$F0,$E4,$FE
	.dc.b	$B1,$AF,$F2,$B6,$BF,$E9,$CB,$E8,$BA,$1F,$47,$FE,$36,$8F,$F7,$DC
	.dc.b	$63,$9B,$BC,$7F,$8B,$BE,$8D,$F4,$3E,$CF,$05,$18,$9D,$4F,$C0,$F0
	.dc.b	$1E,$FB,$D3,$E1,$7B,$BF,$E8,$18,$72,$C3,$E7,$36,$3F,$83,$B5,$DC
	.dc.b	$F1,$E1,$FF,$70,$7C,$F7,$F7,$29,$FF,$76,$B5,$FF,$01,$D7,$19,$FE
	.dc.b	$D4,$BE,$81,$E9,$7A,$DB,$5F,$E4,$C6,$BF,$EA,$DF,$87,$D4,$F9,$A7
	.dc.b	$33,$FB,$0A,$E9,$39,$24,$7F,$C0,$0F,$03,$FC,$78,$DF,$3B,$9B,$FE
	.dc.b	$5A,$ED,$7D,$1F,$CD,$6C,$66,$3E,$67,$07,$FA,$76,$30,$7C,$69,$DB
	.dc.b	$F3,$1A,$F9,$5A,$FF,$9D,$9D,$E9,$F2,$BE,$23,$F2,$FE,$63,$74,$78
	.dc.b	$F1,$F8,$DE,$A4,$C7,$E3,$3F,$7F,$D1,$7D,$1E,$67,$EE,$F0,$FF,$9C
	.dc.b	$1F,$B6,$D7,$35,$F2,$CF,$0E,$FF,$77,$BB,$FC,$2F,$A7,$F8,$13,$97
	.dc.b	$A6,$DF,$6B,$7F,$F7,$E6,$3F,$27,$E0,$7E,$41,$F2,$7E,$22,$C7,$C5
	.dc.b	$8F,$C6,$CF,$08,$F8,$D1,$F6,$FE,$86,$47,$99,$AF,$AB,$E6,$FF,$00
	.dc.b	$47,$D6,$FE,$62,$FF,$C6,$D1,$F9,$C6,$BF,$85,$6B,$FE,$A6,$6B,$FE
	.dc.b	$57,$5F,$45,$CC,$91,$FA,$28,$CB,$FC,$4F,$E7,$9F,$17,$FC,$3A,$11
	.dc.b	$87,$FC,$7F,$9F,$92,$31,$1F,$E3,$6F,$9E,$7B,$1C,$67,$33,$FE,$33
	.dc.b	$6B,$FD,$31,$AF,$F9,$29,$AF,$82,$FC,$54,$FA,$5A,$75,$8F,$C3,$7C
	.dc.b	$5F,$DE,$D6,$5F,$FA,$67,$9D,$3A,$8F,$D4,$4E,$9F,$F3,$EF,$85,$FE
	.dc.b	$9E,$9F,$CB,$F5,$AA,$92,$BF,$D4,$9D,$8F,$A3,$C4,$7F,$A1,$B9,$BF
	.dc.b	$3D,$58,$E0,$FF,$37,$65,$FC,$62,$FF,$9C,$C3,$E8,$FE,$3E,$1F,$33
	.dc.b	$A3,$E6,$C6,$23,$FD,$23,$3B,$E2,$FE,$83,$E3,$7E,$DE,$78,$09,$D3
	.dc.b	$FE,$24,$6B,$FD,$0E,$BA,$3F,$D3,$DA,$F8,$2B,$AF,$D1,$BE,$2E,$9E
	.dc.b	$8A,$F2,$5F,$FA,$27,$05,$3E,$43,$8F,$C3,$4C,$BF,$A0,$CE,$5B,$39
	.dc.b	$7F,$D2,$B0,$9F,$0C,$FE,$E3,$E1,$65,$7F,$1F,$7F,$82,$7C,$7E,$23
	.dc.b	$E8,$78,$09,$C6,$5F,$ED,$3A,$FC,$97,$CD,$CB,$F8,$E7,$E7,$1F,$C6
	.dc.b	$FD,$1E,$63,$FC,$97,$D8,$9F,$F2,$83,$63,$F4,$D1,$83,$F2,$BF,$41
	.dc.b	$23,$AC,$ED,$F2,$F8,$1F,$C4,$88,$FE,$57,$E8,$EB,$A3,$F9,$E4,$7F
	.dc.b	$21,$8F,$BF,$F2,$9F,$E0,$A9,$FE,$D6,$FF,$BC,$DB,$F1,$5A,$F6,$24
	.dc.b	$60,$87,$FB,$0A,$79,$3F,$EA,$CC,$1F,$17,$FB,$72,$47,$F7,$8F,$92
	.dc.b	$7C,$8E,$23,$E2,$CF,$C4,$23,$FC,$25,$FF,$0A,$FA,$3F,$E0,$DE,$BB
	.dc.b	$ED,$7F,$1A,$E3,$FE,$11,$5F,$D4,$3E,$8F,$F5,$FC,$B3,$1F,$D6,$C6
	.dc.b	$5F,$EE,$31,$97,$F6,$FE,$12,$7F,$A3,$61,$FC,$FB,$E8,$E5,$7F,$01
	.dc.b	$8C,$BF,$D9,$E3,$2F,$F0,$0F,$1B,$FC,$44,$FE,$13,$E8,$DC,$DF,$E8
	.dc.b	$F6,$C7,$F8,$DD,$AF,$FA,$41,$AF,$FB,$3D,$CE,$9F,$EE,$DF,$86,$7D
	.dc.b	$17,$FB,$1A,$30,$DD,$FF,$CD,$B5,$C1,$FE,$75,$F4,$7F,$E2,$3F,$38
	.dc.b	$D3,$FC,$83,$87,$4F,$85,$C6,$5F,$CE,$E7,$FA,$41,$1F,$E7,$DD,$9F
	.dc.b	$D2,$C8,$E0,$87,$F1,$9F,$92,$FA,$D8,$FF,$66,$6C,$7F,$4B,$9F,$E8
	.dc.b	$9F,$3E,$F9,$B7,$D1,$7F,$B6,$7E,$1F,$FD,$0F,$39,$8C,$58,$46,$5F
	.dc.b	$F7,$AC,$FF,$A3,$BE,$8D,$E2,$7F,$BF,$E3,$9E,$1D,$AB,$67,$3B,$E5
	.dc.b	$E2,$FF,$99,$E3,$13,$B8,$C4,$30,$EA,$F9,$0F,$1C,$EE,$FB,$E3,$9C
	.dc.b	$3A,$C7,$97,$1F,$25,$F6,$AD,$5F,$18,$C1,$E6,$3C,$13,$C1,$EF,$3C
	.dc.b	$08,$EF,$0F,$2B,$C7,$0F,$0B,$D3,$1F,$5C,$62,$3A,$9C,$A3,$38,$78
	.dc.b	$E3,$B9,$D7,$0F,$15,$71,$24,$EE,$FC,$37,$CB,$0E,$98,$FC,$4C,$79
	.dc.b	$18,$3C,$8E,$67,$E7,$D7,$28,$C1,$27,$67,$E0,$E3,$4E,$0F,$3C,$F8
	.dc.b	$5E,$E7,$C3,$BF,$66,$1E,$8C,$3E,$D5,$F2,$AE,$38,$77,$69,$DF,$1F
	.dc.b	$83,$87,$22,$5F,$28,$CB,$ED,$9E,$27,$CE,$3C,$4E,$70,$FA,$58,$78
	.dc.b	$5F,$24,$E8,$F4,$E3,$EB,$7F,$DD,$67,$D4,$2C,$7E,$8D,$B8,$7E,$7D
	.dc.b	$CF,$D1,$FF,$A1,$1F,$D0,$35,$FD,$5B,$64,$D3,$FD,$23,$5F,$88,$6E
	.dc.b	$5F,$EF,$CF,$A3,$7D,$1B,$7F,$D0,$56,$B1,$EF,$DF,$83,$F5,$17,$C2
	.dc.b	$B1,$3F,$97,$AE,$0F,$92,$B7,$7D,$63,$51,$C3,$1F,$23,$AB,$C6,$3F
	.dc.b	$C2,$B0,$E7,$97,$B2,$B5,$87,$22,$61,$C9,$FD,$86,$5E,$57,$38,$73
	.dc.b	$AE,$B1,$FC,$4E,$75,$5C,$49,$5B,$9E,$79,$7A,$67,$31,$F1,$0E,$A7
	.dc.b	$38,$D4,$F1,$9D,$CF,$C3,$39,$E5,$E2,$71,$1A,$BF,$1D,$71,$24,$F7
	.dc.b	$9F,$E3,$73,$A8,$C4,$E9,$E6,$FD,$DB,$F1,$C6,$88,$68,$F4,$5F,$8F
	.dc.b	$1F,$9D,$1E,$C1,$7D,$DE,$04,$8D,$E1,$E4,$BE,$F3,$9C,$BE,$C9,$F8
	.dc.b	$4B,$E5,$7C,$31,$F4,$36,$BC,$67,$CC,$F8,$37,$D1,$BE,$5F,$F1,$5D
	.dc.b	$5F,$8B,$F5,$9C,$AF,$9E,$CF,$07,$F4,$FC,$78,$E1,$8C,$98,$70,$BC
	.dc.b	$56,$AF,$F6,$B4,$8C,$3F,$B8,$6B,$FE,$F5,$CB,$C9,$3C,$27,$CF,$F0
	.dc.b	$FF,$EB,$46,$C6,$65,$39,$6C,$66,$3C,$71,$8A,$FF,$3E,$4F,$F6,$9E
	.dc.b	$A7,$D1,$FF,$BA,$A3,$FD,$03,$87,$6A,$D9,$FD,$BF,$E8,$FF,$C7,$33
	.dc.b	$FE,$86,$BE,$3F,$11,$F0,$E9,$5A,$78,$CE,$AB,$FC,$CF,$39,$9F,$E3
	.dc.b	$2D,$7F,$BE,$1C,$BC,$9F,$F6,$8C,$66,$1E,$D3,$C4,$FF,$43,$5F,$FA
	.dc.b	$63,$7F,$B3,$E0,$FD,$5A,$D7,$A7,$F5,$E7,$35,$FE,$BB,$8C,$71,$AF
	.dc.b	$86,$1D,$31,$FF,$0D,$47,$3F,$27,$FD,$B7,$87,$73,$FC,$30,$7F,$AE
	.dc.b	$BD,$EE,$CE,$2B,$FC,$9B,$18,$EE,$E4,$D9,$FF,$44,$FC,$7F,$FA,$3D
	.dc.b	$FF,$58,$F7,$BF,$48,$FF,$0F,$46,$BC,$DF,$C8,$4E,$71,$FA,$17,$C5
	.dc.b	$FC,$7B,$5F,$E8,$AD,$7E,$6B,$5F,$F1,$53,$C7,$AF,$F9,$64,$23,$17
	.dc.b	$FE,$2F,$F6,$5E,$97,$C5,$BD,$D6,$AD,$5A,$F8,$0F,$D3,$3C,$3C,$7B
	.dc.b	$87,$FB,$AB,$1F,$F4,$8C,$62,$B2,$D8,$C9,$F7,$78,$9F,$09,$F6,$0F
	.dc.b	$AF,$ED,$59,$D4,$5C,$BF,$CE,$93,$74,$F3,$FD,$27,$CA,$FA,$2F,$F6
	.dc.b	$36,$1F,$4A,$31,$5F,$31,$F8,$77,$F2,$1A,$FC,$86,$BF,$1D,$BB,$1E
	.dc.b	$55,$F7,$3C,$87,$CF,$E2,$FF,$CC,$DF,$49,$F9,$23,$F2,$3F,$03,$6F
	.dc.b	$E0,$45,$9F,$F4,$74,$DD,$3F,$DD,$16,$F5,$7E,$7F,$B3,$FD,$C7,$1F
	.dc.b	$F7,$DF,$27,$82,$73,$1F,$52,$B8,$3F,$EB,$53,$11,$A7,$FC,$DD,$7F
	.dc.b	$C9,$BE,$93,$F7,$8F,$83,$DF,$FE,$44,$A3,$33,$F8,$7B,$56,$A7,$F7
	.dc.b	$55,$BF,$0F,$E5,$B6,$FF,$93,$B7,$4F,$F0,$EB,$4F,$14,$FA,$7A,$7A
	.dc.b	$3F,$B0,$F2,$A5,$FA,$0F,$F4,$37,$86,$EF,$F2,$E3,$FC,$E2,$7F,$9F
	.dc.b	$1F,$BB,$AE,$31,$FE,$C4,$BB,$8F,$F8,$3F,$06,$3F,$B4,$AF,$F4,$F7
	.dc.b	$C9,$7F,$0F,$8C,$57,$F4,$BE,$94,$E6,$BF,$87,$B7,$2F,$CE,$3F,$BA
	.dc.b	$F9,$CF,$97,$F5,$5F,$89,$9F,$E0,$B1,$8E,$7D,$47,$F4,$CB,$FD,$B4
	.dc.b	$FE,$40,$91,$9E,$84,$70,$9F,$E7,$7D,$EE,$B7,$D2,$E7,$5F,$F1,$67
	.dc.b	$97,$FD,$9B,$B7,$FA,$61,$33,$FE,$A0,$FF,$53,$FA,$5F,$85,$97,$F3
	.dc.b	$8A,$46,$23,$FB,$43,$5F,$EC,$4D,$8D,$FC,$BF,$32,$5F,$F6,$19,$47
	.dc.b	$7F,$A0,$93,$96,$DF,$F6,$BF,$9E,$F7,$CF,$F6,$96,$C6,$76,$FE,$D5
	.dc.b	$18,$8F,$43,$AC,$FE,$0B,$6B,$FC,$00,$BE,$53,$FB,$EB,$9A,$F5,$C7
	.dc.b	$F4,$AF,$84,$FF,$00,$E7,$BB,$FE,$31,$18,$9F,$90,$48,$F7,$36,$BF
	.dc.b	$B7,$E9,$F2,$A3,$F1,$9A,$71,$AF,$EE,$B8,$BE,$63,$FA,$7E,$3E,$46
	.dc.b	$CE,$9F,$79,$1F,$88,$DC,$7B,$D7,$39,$FD,$44,$7F,$6E,$F3,$5F,$11
	.dc.b	$D6,$7F,$C0,$BF,$4B,$FC,$36,$BF,$DF,$DE,$0F,$F8,$82,$31,$BF,$E8
	.dc.b	$63,$15,$FE,$19,$9E,$15,$F8,$76,$BF,$E2,$D6,$BF,$E2,$77,$9C,$64
	.dc.b	$FD,$77,$C7,$74,$7F,$C2,$FF,$4B,$AA,$FD,$E5,$FF,$8C,$A3,$1E,$9F
	.dc.b	$FA,$4E,$31,$5F,$E2,$B9,$EF,$C2,$69,$23,$67,$C4,$EF,$FF,$08,$7B
	.dc.b	$58,$C7,$FC,$4B,$D6,$F9,$D7,$D2,$7F,$60,$FA,$5B,$8F,$CC,$AD,$4E
	.dc.b	$4F,$F2,$53,$88,$FE,$F3,$5C,$2F,$FB,$CB,$5F,$EF,$2D,$7F,$BC,$BB
	.dc.b	$E0,$FF,$52,$57,$39,$FF,$1C,$F9,$4F,$A0,$FD,$FE,$1E,$3C,$38,$3F
	.dc.b	$E0,$4E,$CF,$F0,$5F,$CB,$FF,$83,$1B,$1F,$E1,$67,$FA,$E9,$1F,$E1
	.dc.b	$CF,$F6,$2C,$03,$E1,$2F,$D1,$3E,$2B,$FC,$23,$DB,$FE,$B9,$8C,$57
	.dc.b	$FB,$6A,$B7,$5A,$C2,$CE,$47,$D2,$70,$7F,$C2,$D8,$7F,$B0,$E3,$11
	.dc.b	$E2,$BF,$1C,$6F,$6F,$E1,$77,$8C,$3F,$CE,$B5,$6F,$FC,$47,$87,$39
	.dc.b	$CC,$78,$1F,$23,$4F,$26,$1D,$67,$84,$62,$77,$3E,$48,$F8,$87,$13
	.dc.b	$D3,$7E,$17,$FC,$C3,$77,$0F,$3C,$6C,$F6,$1C,$5C,$D7,$59,$E1,$87
	.dc.b	$8A,$B9,$8C,$7C,$D8,$7C,$C8,$CC,$61,$D6,$3F,$03,$0C,$1D,$63,$C1
	.dc.b	$7D,$5F,$1C,$6A,$78,$1F,$00,$F9,$31,$FD,$5F,$0F,$EB,$0F,$A8,$C1
	.dc.b	$FE,$A6,$F9,$CF,$F4,$DF,$C0,$7F,$5D,$6C,$67,$E1,$13,$96,$C7,$1F
	.dc.b	$A4,$71,$F1,$6C,$66,$30,$47,$E9,$7D,$23,$C1,$FF,$48,$E9,$F8,$D8
	.dc.b	$C4,$FD,$5C,$98,$74,$AF,$C9,$46,$30,$FB,$57,$C2,$FA,$E1,$D3,$4F
	.dc.b	$F4,$ED,$EE,$7F,$04,$DC,$3C,$15,$B9,$FD,$5D,$F3,$9E,$F8,$F8,$70
	.dc.b	$E6,$DA,$DD,$F1,$57,$C6,$B7,$8F,$AE,$F8,$56,$AF,$85,$EE,$3F,$77
	.dc.b	$8F,$97,$0E,$31,$EC,$3D,$F5,$DE,$75,$1C,$2F,$AD,$F3,$8E,$C7,$87
	.dc.b	$08,$F3,$26,$3E,$E8,$CB,$F1,$23,$B9,$8C,$7F,$D8,$F4,$F9,$5F,$F6
	.dc.b	$4D,$FF,$A6,$A1,$F1,$38,$E1,$DB,$70,$F9,$B7,$5F,$8E,$DA,$E3,$1F
	.dc.b	$D9,$0F,$97,$39,$9E,$5C,$5F,$D0,$67,$A7,$3B,$E9,$3F,$C3,$31,$FE
	.dc.b	$72,$6B,$FD,$23,$9F,$FC,$BB,$97,$F3,$A8,$E3,$3A,$CD,$3F,$EB,$18
	.dc.b	$C3,$B9,$FE,$74,$70,$E4,$8E,$0F,$F7,$DF,$D2,$3E,$87,$E0,$78,$08
	.dc.b	$C4,$EA,$7E,$27,$7F,$F9,$9F,$D2,$DD,$6A,$BF,$5B,$8F,$F3,$EF,$0D
	.dc.b	$F3,$FF,$D8,$27,$FD,$E4,$D7,$FC,$37,$5C,$75,$FF,$26,$B6,$F9,$56
	.dc.b	$E7,$DF,$4B,$EB,$E0,$78,$B3,$F4,$FD,$13,$D2,$A4,$BF,$EE,$DE,$53
	.dc.b	$B7,$11,$E8,$3A,$77,$FF,$8A,$5F,$F0,$FE,$5F,$D5,$FC,$BD,$E9,$FE
	.dc.b	$62,$CB,$B5,$6A,$F8,$3C,$EF,$83,$FE,$28,$E2,$57,$08,$FF,$1D,$F7
	.dc.b	$3E,$93,$E8,$B6,$38,$35,$F3,$72,$7F,$A5,$FC,$9D,$D7,$E5,$F1,$F9
	.dc.b	$FF,$19,$78,$CE,$B0,$E7,$59,$89,$18,$3C,$D9,$7F,$96,$DB,$E7,$FC
	.dc.b	$C3,$F0,$3C,$2F,$A4,$C4,$FC,$1E,$4F,$17,$07,$D9,$87,$07,$C0,$C3
	.dc.b	$71,$F1,$67,$5A,$FF,$80,$EF,$3C,$3F,$E8,$EE,$1F,$F7,$DD,$FC,$2F
	.dc.b	$84,$F8,$A7,$C6,$F8,$BB,$1E,$8F,$A2,$FC,$09,$87,$FC,$CF,$BA,$3E
	.dc.b	$6F,$C2,$C9,$E0,$7C,$0F,$82,$FD,$0D,$FE,$3A,$3F,$CD,$ED,$CB,$F0
	.dc.b	$13,$AE,$BF,$A1,$EE,$BE,$5A,$99,$C3,$F0,$18,$3E,$93,$7F,$5F,$27
	.dc.b	$31,$FC,$37,$E7,$DF,$1F,$FC,$3F,$E9,$5E,$A5,$35,$12,$FF,$88,$4E
	.dc.b	$9B,$39,$7F,$38,$2B,$AB,$A2,$1A,$72,$FE,$3B,$A9,$F1,$DF,$73,$8A
	.dc.b	$4F,$EA,$A3,$F0,$66,$1F,$B9,$18,$7F,$7E,$FE,$4C,$8F,$F9,$9F,$E2
	.dc.b	$1F,$4B,$88,$FD,$47,$81,$F5,$DE,$26,$E3,$F9,$A7,$23,$E9,$5E,$EF
	.dc.b	$F2,$CF,$A5,$7C,$7D,$37,$3A,$C3,$FD,$33,$7C,$6F,$51,$FE,$9C,$F4
	.dc.b	$18,$72,$8D,$DF,$17,$F8,$35,$F1,$78,$4E,$DF,$F3,$5B,$63,$F8,$B9
	.dc.b	$B8,$CF,$CA,$F9,$FE,$1A,$47,$28,$F5,$18,$79,$C6,$1F,$E1,$DF,$17
	.dc.b	$FA,$E2,$3F,$C7,$3E,$4A,$BF,$AD,$8B,$7F,$CA,$3C,$B7,$1A,$FF,$21
	.dc.b	$D7,$F8,$FF,$A1,$98,$FF,$5B,$E6,$FC,$E3,$67,$2F,$E3,$3A,$15,$CA
	.dc.b	$34,$7E,$51,$DE,$5F,$AB,$BE,$73,$FC,$07,$BF,$F8,$39,$1F,$E6,$BF
	.dc.b	$49,$5D,$1F,$CB,$63,$F8,$8C,$7F,$6A,$F9,$5F,$EF,$93,$F6,$5F,$F6
	.dc.b	$9B,$7E,$23,$5E,$D4,$8C,$10,$FF,$50,$CF,$27,$F4,$5F,$13,$FE,$01
	.dc.b	$F5,$1F,$47,$FC,$07,$C6,$FF,$4C,$C6,$1F,$DC,$4C,$AF,$F7,$2C,$E5
	.dc.b	$B5,$C5,$FF,$50,$E8,$AE,$11,$FD,$46,$B8,$36,$B8,$3F,$B9,$9F,$EE
	.dc.b	$3D,$CF,$A4,$FE,$7D,$CD,$2B,$F5,$91,$A3,$0F,$EB,$23,$0F,$F6,$2B
	.dc.b	$FE,$6F,$F2,$B2,$30,$43,$F9,$B3,$97,$F8,$24,$7F,$88,$FE,$8F,$F0
	.dc.b	$31,$FE,$BC,$E2,$FE,$EB,$F0,$CF,$A4,$FF,$59,$46,$1B,$BF,$FB,$0A
	.dc.b	$97,$C9,$FE,$2E,$DA,$E1,$F0,$9F,$E3,$3C,$3A,$3E,$1E,$6F,$E6,$FF
	.dc.b	$49,$FD,$87,$85,$F4,$94,$CE,$5F,$CC,$A7,$0F,$F7,$C9,$DB,$93,$F9
	.dc.b	$BF,$57,$F3,$CD,$24,$60,$87,$F7,$DF,$3B,$EE,$7A,$87,$EF,$7A,$EF
	.dc.b	$43,$F3,$FB,$FF,$4B,$FA,$17,$CF,$78,$3D,$8F,$F2,$4E,$49,$CE,$FF
	.dc.b	$ED,$5F,$87,$F8,$D1,$89,$FF,$74,$47,$C9,$FA,$4F,$DD,$46,$CF,$F7
	.dc.b	$6C,$EE,$FA,$B6,$B9,$E1,$E1,$79,$78,$9F,$EC,$C7,$38,$F3,$C3,$C9
	.dc.b	$5B,$7A,$B5,$63,$E2,$7C,$A6,$A7,$96,$1E,$18,$C1,$A3,$8E,$1E,$17
	.dc.b	$38,$FA,$30,$ED,$5C,$67,$67,$48,$C9,$2F,$D1,$87,$8E,$F9,$46,$C9
	.dc.b	$5E,$19,$D6,$3F,$02,$3C,$98,$77,$3C,$13,$A8,$F2,$CE,$EF,$AC,$73
	.dc.b	$C7,$D4,$FC,$8C,$3B,$CE,$8F,$BF,$7C,$E7,$47,$83,$1F,$53,$CC,$E3
	.dc.b	$87,$72,$23,$E1,$9C,$6F,$6F,$98,$D5,$77,$C7,$D9,$7D,$B0,$F1,$E3
	.dc.b	$E8,$9C,$E1,$CB,$0E,$B5,$E9,$C7,$C3,$1E,$88,$E2,$EE,$B8,$BE,$93
	.dc.b	$C8,$48,$E4,$F9,$2F,$C1,$87,$EC,$A3,$F8,$21,$F1,$0B,$5F,$A1,$6C
	.dc.b	$6D,$FD,$2F,$D2,$7F,$9A,$DB,$C5,$C9,$FD,$1C,$65,$B1,$97,$F6,$11
	.dc.b	$FE,$7A,$C3,$F4,$ED,$C7,$F8,$B7,$D2,$3E,$83,$FD,$23,$5A,$C3,$F3
	.dc.b	$B7,$BC,$3A,$9F,$9D,$BE,$1C,$7A,$A9,$3F,$8E,$AE,$0A,$B6,$E6,$77
	.dc.b	$7D,$E3,$A6,$1E,$2C,$7D,$F1,$B2,$4F,$F0,$CB,$E6,$F1,$24,$F0,$75
	.dc.b	$5C,$EB,$D3,$1A,$6C,$FC,$29,$D4,$7C,$8B,$F2,$CF,$33,$58,$E2,$78
	.dc.b	$90,$F7,$57,$2C,$D9,$7C,$26,$E1,$C6,$B7,$76,$73,$1E,$EC,$3B,$B4
	.dc.b	$ED,$1A,$C7,$DB,$5C,$70,$EC,$7D,$0C,$BE,$2E,$1E,$9C,$38,$CE,$72
	.dc.b	$F3,$3C,$32,$F2,$5F,$0A,$DC,$62,$B6,$F7,$3F,$95,$4E,$E7,$E4,$1E
	.dc.b	$72,$E3,$E6,$7D,$5C,$2F,$A4,$7C,$AF,$F9,$2B,$4F,$F7,$67,$23,$E7
	.dc.b	$F1,$87,$FA,$BA,$F7,$5F,$ED,$8A,$D3,$FE,$D1,$8C,$CF,$F5,$9A,$E3
	.dc.b	$87,$7C,$39,$F3,$7F,$5B,$9C,$DF,$FB,$9F,$0F,$ED,$1F,$0F,$83,$7A
	.dc.b	$E7,$2F,$F7,$56,$5E,$99,$C9,$BA,$E2,$7F,$81,$CF,$F0,$17,$D1,$BE
	.dc.b	$83,$D7,$DC,$FA,$4F,$F6,$CD,$6F,$1F,$E4,$17,$B8,$C5,$6F,$E1,$FE
	.dc.b	$1D,$2B,$4F,$FB,$F6,$75,$5C,$E3,$58,$78,$30,$EF,$1F,$EA,$1F,$4F
	.dc.b	$E1,$F0,$FF,$7B,$BF,$17,$A5,$FE,$6A,$9C,$F7,$7D,$28,$C4,$FD,$18
	.dc.b	$C7,$A4,$AD,$CF,$FA,$87,$DE,$F9,$EF,$F8,$10,$C1,$EF,$E4,$39,$CB
	.dc.b	$FE,$00,$F9,$7F,$EC,$68,$D1,$9F,$07,$FC,$FD,$19,$94,$72,$55,$D3
	.dc.b	$BB,$E3,$FC,$8B,$CB,$FC,$5F,$3B,$E9,$3F,$A7,$67,$F1,$EB,$4F,$F9
	.dc.b	$7A,$33,$8F,$FB,$3A,$72,$FF,$96,$E3,$39,$7F,$BA,$E7,$2F,$8E,$FB
	.dc.b	$57,$8B,$0F,$0D,$F0,$C3,$FC,$8D,$87,$07,$FC,$E3,$19,$A7,$B2,$F4
	.dc.b	$7F,$DA,$F5,$FE,$83,$F1,$3F,$4B,$39,$F7,$ED,$7F,$A8,$6B,$FD,$11
	.dc.b	$18,$AC,$B6,$3F,$55,$F0,$BF,$82,$3E,$DF,$B3,$63,$24,$C7,$FD,$D1
	.dc.b	$19,$7E,$27,$D0,$FC,$CF,$A1,$4F,$D0,$8C,$36,$BF,$D1,$33,$A7,$E8
	.dc.b	$4E,$9A,$6A,$78,$57,$FA,$26,$B8,$3F,$E5,$5F,$86,$FC,$C9,$F9,$2F
	.dc.b	$80,$B7,$FC,$16,$2C,$FF,$BF,$E6,$F1,$7E,$1F,$67,$C6,$BD,$6F,$97
	.dc.b	$F4,$9F,$94,$CB,$FD,$D1,$39,$7F,$DD,$51,$9C,$7F,$23,$1F,$E1,$7B
	.dc.b	$E7,$7C,$A3,$EA,$DF,$1A,$D9,$F7,$2F,$96,$1F,$EE,$AF,$A4,$FD,$DB
	.dc.b	$E5,$A7,$BF,$FC,$8A,$13,$AC,$3C,$15,$A7,$FE,$00,$C6,$70,$FC,$4C
	.dc.b	$E5,$FF,$7E,$C6,$72,$F8,$38,$75,$8C,$56,$F0,$F2,$78,$5E,$AA,$FD
	.dc.b	$DF,$93,$32,$FE,$0F,$D9,$DF,$B3,$F7,$73,$9D,$3D,$F1,$8A,$DE,$19
	.dc.b	$C7,$FD,$F1,$D5,$F9,$FA,$FF,$47,$7C,$9F,$F0,$F8,$C7,$86,$8C,$63
	.dc.b	$FC,$61,$B8,$FF,$43,$7F,$7F,$F4,$5F,$1F,$E9,$1D,$18,$AF,$E5,$D9
	.dc.b	$A7,$31,$FC,$06,$B8,$56,$F0,$FA,$F7,$B8,$C3,$FC,$3C,$C5,$71,$3A
	.dc.b	$DF,$4A,$FE,$8F,$E6,$7E,$62,$BF,$8B,$68,$F0,$3F,$C4,$71,$FC,$67
	.dc.b	$CC,$FE,$C9,$D7,$FD,$43,$D0,$FE,$B3,$F4,$7F,$16,$DF,$49,$8C,$FF
	.dc.b	$AA,$BA,$F9,$8C,$6D,$FD,$2F,$26,$09,$CD,$FF,$5A,$87,$D0,$7F,$66
	.dc.b	$F9,$F7,$BB,$E0,$8C,$39,$E3,$C6,$5D,$69,$18,$C3,$FB,$2C,$62,$75
	.dc.b	$3D,$7E,$57,$94,$79,$71,$47,$F5,$2F,$84,$F7,$B6,$74,$DC,$3D,$0D
	.dc.b	$E1,$F9,$17,$CA,$B6,$7F,$5A,$E8,$F8,$0E,$2F,$CB,$F3,$1F,$D5,$31
	.dc.b	$F8,$0D,$9F,$63,$67,$71,$FD,$4A,$76,$E8,$FE,$CF,$E7,$BE,$23,$35
	.dc.b	$FE,$06,$FA,$4F,$DE,$D7,$F8,$73,$C1,$FF,$09,$C6,$27,$FA,$BF,$6F
	.dc.b	$F1,$2A,$FB,$13,$9E,$9F,$F7,$85,$F0,$9F,$EE,$ED,$76,$7E,$9F,$E3
	.dc.b	$B1,$87,$F8,$5B,$E9,$39,$4F,$58,$C3,$FD,$D2,$BF,$C4,$B3,$A6,$C6
	.dc.b	$4F,$C5,$47,$9F,$9C,$FA,$7D,$56,$BF,$E3,$0E,$2F,$CD,$4F,$B7,$D1
	.dc.b	$7C,$F3,$E7,$3F,$AB,$FD,$25,$CF,$FC,$41,$5A,$9C,$9F,$E4,$07,$11
	.dc.b	$FE,$25,$F2,$95,$C7,$6F,$E5,$75,$A7,$F9,$6C,$EA,$F9,$61,$BC,$BF
	.dc.b	$A0,$61,$71,$FF,$67,$63,$C6,$73,$7F,$E0,$8E,$CF,$F0,$4F,$CB,$5B
	.dc.b	$FF,$03,$37,$3F,$F0,$23,$6F,$FC,$04,$FF,$8D,$A3,$FD,$B5,$F0,$1F
	.dc.b	$ED,$BF,$A1,$7C,$57,$F8,$03,$C1,$FF,$0C,$46,$2B,$FD,$8B,$5B,$AD
	.dc.b	$63,$67,$22,$B8,$2D,$3F,$3E,$E6,$BF,$D8,$D7,$C6,$35,$87,$82,$73
	.dc.b	$1B,$3E,$1C,$65,$EF,$7D,$AF,$AB,$E0,$8F,$69,$C4,$EB,$87,$77,$31
	.dc.b	$C6,$BA,$47,$FB,$F4,$61,$E5,$BE,$B8,$79,$2F,$86,$1C,$EF,$67,$7A
	.dc.b	$DB,$E6,$AF,$8E,$F8,$CF,$0B,$8A,$F4,$47,$27,$AB,$B3,$11,$B3,$C4
	.dc.b	$57,$B1,$F1,$AB,$C9,$7E,$9C,$3D,$B3,$F1,$0F,$F7,$94,$7F,$BC,$DC
	.dc.b	$1F,$40,$F8,$43,$E6,$DF,$04,$F8,$0F,$F7,$D4,$61,$B1,$99,$E2,$71
	.dc.b	$70,$F4,$97,$D1,$7F,$99,$A3,$71,$8D,$BA,$CE,$67,$83,$5F,$AD,$3B
	.dc.b	$7E,$F7,$D1,$3C,$5F,$B5,$B7,$E4,$A3,$13,$F8,$FA,$E1,$3F,$96,$6E
	.dc.b	$DF,$81,$AD,$D7,$4B,$E7,$3E,$0C,$3D,$18,$71,$6C,$71,$9D,$D0,$5F
	.dc.b	$4B,$E1,$8F,$8F,$0E,$35,$BA,$D6,$1C,$E3,$55,$1F,$21,$FA,$99,$DB
	.dc.b	$DF,$1F,$5B,$6F,$AB,$E2,$C7,$CA,$D3,$A6,$3E,$47,$A9,$1C,$13,$1F
	.dc.b	$24,$65,$F8,$98,$F2,$73,$3F,$46,$BB,$E1,$E4,$E9,$7D,$0B,$E1,$7F
	.dc.b	$61,$8F,$F8,$3E,$31,$97,$F8,$1D,$AF,$FA,$CB,$D9,$DF,$17,$CD,$C3
	.dc.b	$9E,$18,$A4,$61,$E5,$19,$EC,$7A,$8C,$38,$5F,$93,$E7,$EE,$46,$9C
	.dc.b	$6B,$FC,$81,$8F,$96,$32,$D3,$FC,$A3,$E9,$E1,$FF,$7A,$D7,$37,$47
	.dc.b	$FB,$32,$3C,$3F,$87,$FE,$92,$3F,$C4,$CD,$7F,$DF,$AD,$D7,$FD,$99
	.dc.b	$9B,$F0,$FF,$F4,$35,$70,$E3,$E6,$47,$FD,$D3,$1A,$79,$C7,$C7,$EF
	.dc.b	$FF,$47,$7C,$8D,$F7,$0F,$1A,$FF,$42,$EE,$FA,$77,$CC,$FC,$86,$CF
	.dc.b	$F8,$F5,$AF,$F6,$0F,$A5,$F8,$0F,$28,$CE,$BF,$E4,$D6,$F2,$FD,$B6
	.dc.b	$CF,$FB,$3A,$B8,$D6,$EF,$EF,$2F,$C0,$1F,$4F,$86,$DF,$8E,$BC,$DE
	.dc.b	$CC,$8F,$33,$F8,$3F,$4B,$FC,$9F,$18,$21,$E7,$8D,$78,$BC,$63,$FD
	.dc.b	$13,$C1,$FC,$9B,$2F,$F3,$4B,$75,$F5,$5F,$E3,$3F,$2A,$EB,$7F,$29
	.dc.b	$B7,$FE,$6D,$6D,$7F,$07,$E4,$64,$F1,$B9,$67,$31,$FF,$35,$61,$C1
	.dc.b	$5F,$0A,$D4,$79,$BD,$9F,$D2,$79,$FF,$EF,$87,$39,$FF,$A7,$BC,$2F
	.dc.b	$37,$AF,$2F,$FC,$AD,$15,$F4,$45,$9E,$11,$E2,$F0,$3D,$C4,$7C,$B3
	.dc.b	$F0,$7D,$D9,$1E,$87,$E8,$73,$7C,$5F,$5A,$4F,$B1,$DB,$F1,$F8,$3E
	.dc.b	$51,$1F,$F5,$3F,$DB,$FF,$B4,$9F,$E5,$D9,$38,$BF,$BD,$7C,$EB,$E9
	.dc.b	$7F,$82,$52,$30,$F0,$9D,$CF,$F0,$2F,$CE,$BF,$D6,$B3,$93,$F9,$D5
	.dc.b	$70,$7F,$DC,$E7,$39,$EB,$19,$BF,$F5,$45,$72,$AD,$CE,$6F,$BC,$75
	.dc.b	$39,$1F,$EA,$59,$E1,$5F,$ED,$2F,$AE,$FB,$3D,$1F,$EA,$7A,$F2,$71
	.dc.b	$A3,$11,$FE,$FA,$E5,$FC,$4D,$7E,$74,$E3,$E7,$C9,$FF,$49,$47,$4F
	.dc.b	$5E,$47,$4B,$47,$E9,$7D,$EC,$CF,$E8,$22,$E5,$FE,$E6,$9F,$9B,$F2
	.dc.b	$0F,$95,$A4,$BF,$F3,$37,$E8,$6E,$EC,$47,$CA,$AE,$0F,$C3,$6B,$FE
	.dc.b	$68,$9E,$0E,$6B,$8E,$1F,$7E,$3F,$B7,$57,$08,$C6,$4B,$E3,$5F,$17
	.dc.b	$93,$FD,$8A,$FB,$FB,$BE,$69,$FC,$0F,$A6,$46,$48,$71,$7E,$9F,$6C
	.dc.b	$8F,$90,$FF,$0B,$F8,$1C,$47,$F0,$68,$B9,$7F,$38,$C3,$FD,$9B,$3F
	.dc.b	$81,$FB,$FF,$1B,$F8,$C6,$CF,$B6,$11,$8E,$6F,$F2,$B7,$D2,$72,$9E
	.dc.b	$19,$7E,$7A,$B5,$1F,$E1,$16,$A9,$E1,$7D,$67,$24,$AF,$D3,$CE,$AF
	.dc.b	$FC,$35,$CD,$F8,$FE,$FF,$C9,$93,$1F,$F6,$E7,$B3,$5F,$E4,$B1,$72
	.dc.b	$F8,$13,$FD,$3F,$F4,$77,$75,$F4,$26,$1F,$EA,$B9,$CB,$67,$2F,$FA
	.dc.b	$AA,$3F,$A1,$E5,$FE,$2D,$E3,$47,$09,$FD,$AF,$C5,$3E,$97,$F8,$0B
	.dc.b	$63,$F6,$0D,$7F,$77,$B7,$EC,$30,$FF,$41,$4E,$74,$FA,$4D,$8C,$DF
	.dc.b	$C6,$F0,$FA,$B3,$0E,$AE,$5F,$F6,$37,$C8,$3C,$6F,$EE,$0F,$EE,$31
	.dc.b	$FC,$BB,$70,$F6,$3B,$AE,$35,$C3,$BE,$46,$0F,$E2,$DE,$07,$F4,$8F
	.dc.b	$1B,$F8,$3F,$8D,$27,$3D,$1F,$E2,$AC,$FF,$25,$59,$9F,$EB,$75,$FD
	.dc.b	$8E,$3F,$C4,$7F,$97,$7F,$80,$A7,$F9,$17,$9A,$8C,$47,$F8,$11,$B7
	.dc.b	$F9,$DF,$B4,$E3,$E7,$CA,$E3,$5A,$DF,$FA,$F6,$6B,$EF,$F7,$32,$F6
	.dc.b	$FF,$20,$F1,$F1,$1D,$E2,$E5,$FD,$C2,$7C,$FF,$57,$2B,$BB,$FE,$1D
	.dc.b	$F9,$97,$CE,$66,$7D,$13,$6F,$F8,$3D,$6A,$7F,$9A,$7D,$13,$E7,$1F
	.dc.b	$47,$FE,$E8,$FA,$3F,$F6,$34,$E6,$31,$5F,$E7,$F8,$C3,$FE,$CA,$9F
	.dc.b	$F6,$67,$D2,$3E,$7B,$FD,$9D,$F1,$09,$E7,$8F,$6B,$E0,$73,$71,$3B
	.dc.b	$8C,$63,$E0,$58,$75,$71,$8F,$8A,$B8,$42,$31,$87,$82,$F9,$5F,$1A
	.dc.b	$F0,$9B,$C3,$C3,$5B,$8C,$DF,$27,$FD,$3D,$5C,$70,$EF,$39,$9E,$58
	.dc.b	$79,$30,$E7,$1C,$31,$F8,$78,$F5,$C3,$8D,$6A,$B8,$9E,$4C,$7B,$56
	.dc.b	$C9,$1F,$10,$D5,$F4,$71,$7D,$30,$F1,$1F,$02,$75,$3D,$A7,$4F,$BB
	.dc.b	$0F,$16,$1E,$08,$E7,$87,$81,$E0,$FD,$EB,$E7,$8F,$C1,$25,$F2,$8C
	.dc.b	$BE,$D8,$CE,$3E,$3B,$F2,$9E,$16,$E3,$F6,$B1,$F1,$4F,$80,$F1,$90
	.dc.b	$FC,$1E,$1D,$AB,$8E,$3E,$7A,$E3,$87,$82,$FB,$1F,$20,$D9,$CD,$FB
	.dc.b	$51,$C3,$1F,$41,$2F,$D1,$5C,$8F,$F7,$01,$F4,$CA,$7D,$A2,$D7,$E9
	.dc.b	$9B,$F4,$9F,$C7,$DB,$D7,$C1,$FD,$64,$65,$FD,$8B,$5C,$C7,$E8,$23
	.dc.b	$F9,$46,$7F,$A0,$FA,$47,$CB,$FF,$81,$E3,$F2,$B3,$AD,$39,$DF,$13
	.dc.b	$F5,$97,$C3,$9B,$18,$FE,$82,$F8,$3E,$03,$F4,$F3,$B7,$55,$C3,$0E
	.dc.b	$EF,$4C,$7C,$58,$F8,$2B,$64,$AF,$44,$EA,$3B,$47,$27,$53,$F3,$AF
	.dc.b	$C1,$39,$24,$75,$79,$E3,$E3,$8C,$12,$82,$B9,$C6,$63,$F8,$46,$3F
	.dc.b	$01,$B8,$FA,$EF,$A4,$79,$AB,$85,$F2,$8C,$10,$E2,$7C,$8C,$7D,$CE
	.dc.b	$31,$F4,$BB,$9E,$4E,$6B,$C9,$93,$2F,$25,$6A,$31,$97,$92,$F8,$1E
	.dc.b	$E9,$D5,$7A,$EB,$73,$F3,$0D,$63,$E4,$8C,$E3,$E4,$9E,$67,$C7,$9D
	.dc.b	$57,$32,$61,$F1,$67,$A1,$C2,$F9,$E3,$F6,$6F,$8D,$7F,$4A,$E9,$7E
	.dc.b	$5E,$F8,$8F,$F2,$F7,$1F,$F9,$7E,$73,$18,$DB,$FC,$45,$F4,$2F,$A0
	.dc.b	$C6,$AB,$DD,$7F,$7A,$C7,$A3,$70,$DF,$90,$C7,$94,$E4,$E1,$F1,$BD
	.dc.b	$30,$F2,$BD,$BC,$EF,$F4,$97,$8F,$9C,$93,$49,$7F,$E8,$D9,$B9,$B5
	.dc.b	$F6,$5B,$ED,$79,$1E,$AF,$17,$FD,$C7,$78,$C3,$D5,$37,$0F,$F3,$E4
	.dc.b	$DC,$3F,$CD,$DE,$67,$97,$6F,$F2,$7E,$1F,$E2,$28,$C1,$D7,$C2,$CE
	.dc.b	$1F,$EF,$C8,$C1,$A9,$FF,$1C,$FB,$3F,$1B,$C9,$7A,$36,$30,$D9,$FF
	.dc.b	$1F,$B6,$FD,$D1,$FE,$3C,$8F,$AB,$0F,$BF,$7A,$D8,$6D,$7F,$B4,$47
	.dc.b	$F0,$07,$FA,$D1,$FF,$64,$46,$AB,$FD,$A6,$DD,$7F,$D1,$0D,$9F,$DE
	.dc.b	$9C,$BD,$CE,$D8,$48,$E6,$FE,$21,$D1,$83,$FD,$C5,$B3,$D4,$7C,$DF
	.dc.b	$E2,$17,$CC,$58,$C3,$6B,$8C,$E9,$B3,$A7,$FD,$A1,$EB,$7D,$73,$EA
	.dc.b	$15,$C9,$34,$FF,$5D,$46,$5F,$F7,$9F,$88,$F7,$3E,$6B,$6B,$E5,$6F
	.dc.b	$FE,$71,$7F,$DE,$11,$9A,$E3,$5C,$0F,$85,$5B,$8C,$D7,$09,$D3,$D3
	.dc.b	$BE,$CF,$DE,$1D,$7F,$EF,$5F,$B1,$FC,$91,$F8,$F2,$D7,$FB,$BD,$B3
	.dc.b	$FE,$D5,$6F,$17,$FB,$1F,$9F,$F3,$47,$E5,$CB,$5F,$E7,$C6,$D7,$FB
	.dc.b	$75,$B3,$FE,$C9,$FD,$53,$E8,$4F,$94,$38,$8F,$F7,$3F,$E8,$AF,$89
	.dc.b	$FC,$AC,$E7,$0F,$EF,$6D,$BF,$E8,$ED,$E3,$5F,$1A,$FE,$97,$87,$87
	.dc.b	$8F,$99,$23,$F1,$73,$BC,$3C,$98,$75,$D1,$8B,$E9,$3F,$74,$7E,$DC
	.dc.b	$B8,$76,$3F,$DA,$B1,$B7,$17,$F4,$A3,$0F,$FC,$01,$9E,$2F,$D7,$EE
	.dc.b	$6A,$F8,$CF,$E1,$07,$F0,$52,$E1,$FE,$AA,$6D,$71,$9E,$13,$E0,$C7
	.dc.b	$D3,$5F,$92,$F6,$E4,$FF,$AA,$BF,$40,$FE,$3C,$FE,$4A,$33,$87,$E6
	.dc.b	$9F,$E5,$D3,$A8,$C4,$EC,$8F,$E3,$FC,$44,$EF,$81,$5F,$E3,$2E,$66
	.dc.b	$1F,$CE,$7E,$11,$ED,$D3,$FA,$3D,$BE,$93,$F8,$46,$7F,$D2,$6B,$78
	.dc.b	$FF,$B0,$97,$AF,$CC,$E9,$F0,$96,$74,$E1,$FF,$6A,$FC,$57,$91,$C9
	.dc.b	$CF,$DE,$F2,$1E,$5F,$72,$C7,$77,$0F,$2F,$B8,$B9,$7D,$9B,$46,$34
	.dc.b	$F8,$32,$E5,$56,$AF,$A1,$EB,$E6,$FE,$B5,$CB,$E5,$BF,$99,$B7,$E2
	.dc.b	$31,$E7,$3F,$BC,$BF,$95,$71,$F6,$FC,$5D,$C3,$8C,$DB,$F1,$E5,$E2
	.dc.b	$AC,$5F,$F8,$C3,$95,$F9,$A7,$FB,$3E,$3F,$15,$97,$9F,$2F,$BD,$39
	.dc.b	$7F,$87,$47,$F0,$ED,$7F,$BA,$4E,$72,$FF,$03,$36,$FF,$C7,$B9,$7C
	.dc.b	$2F,$5B,$7E,$CC,$9F,$F7,$ED,$6F,$2F,$91,$BF,$F8,$B6,$73,$5D,$BC
	.dc.b	$5F,$C3,$FD,$E7,$EE,$73,$EF,$18,$F0,$75,$7D,$23,$FC,$1B,$F8,$2B
	.dc.b	$EC,$73,$9F,$B2,$6E,$1F,$63,$C0,$5E,$F0,$E4,$79,$BF,$03,$FA,$FC
	.dc.b	$9F,$01,$B1,$92,$63,$F8,$5D,$75,$18,$C7,$E2,$D7,$F8,$37,$6F,$81
	.dc.b	$C9,$F4,$A3,$FC,$3D,$F4,$7F,$83,$AF,$F1,$83,$6F,$E5,$FD,$23,$A3
	.dc.b	$71,$0F,$E2,$38,$2B,$86,$7F,$93,$BE,$15,$A9,$CD,$6E,$39,$57,$F1
	.dc.b	$5A,$DB,$FB,$FA,$D9,$F5,$5C,$72,$7F,$89,$E5,$F2,$FE,$A8,$97,$DD
	.dc.b	$D4,$62,$BF,$A2,$1F,$E8,$2F,$7F,$FD,$07,$F3,$EF,$8E,$FF,$5D,$FD
	.dc.b	$13,$E3,$3F,$5B,$39,$BF,$F5,$ED,$F0,$BD,$D6,$9F,$F5,$D4,$66,$7F
	.dc.b	$D1,$F7,$CF,$2C,$47,$FB,$22,$6C,$E6,$18,$77,$74,$62,$FB,$DF,$28
	.dc.b	$F5,$19,$3B,$E1,$E3,$7C,$E6,$45,$F7,$BE,$D1,$C9,$F0,$46,$6F,$D7
	.dc.b	$3C,$63,$0F,$49,$DD,$77,$8F,$3C,$62,$FC,$D0,$C3,$DF,$5B,$8F,$06
	.dc.b	$1E,$C7,$CC,$60,$F5,$D7,$2A,$E2,$D8,$F2,$57,$37,$B1,$88,$D5,$F8
	.dc.b	$E3,$A9,$A3,$FD,$E0,$53,$DE,$62,$B3,$F4,$1F,$E7,$EF,$A2,$BE,$F9
	.dc.b	$A4,$AF,$F7,$D4,$FE,$53,$8B,$8E,$BF,$E8,$19,$CB,$5F,$F6,$99,$FE
	.dc.b	$D4,$9E,$0E,$A7,$F0,$FF,$45,$FE,$A5,$8B,$F0,$5C,$33,$FC,$35,$F0
	.dc.b	$6E,$1D,$DF,$C5,$5F,$1B,$FC,$13,$63,$78,$7E,$52,$FC,$3D,$BC,$31
	.dc.b	$F2,$B7,$0F,$F7,$4B,$C1,$FF,$15,$6F,$EA,$AF,$13,$EC,$C7,$B3,$6F
	.dc.b	$D0,$EF,$2F,$33,$71,$F6,$E1,$0F,$CC,$61,$D1,$F8,$D5,$C1,$FF,$7D
	.dc.b	$E1,$D9,$B8,$76,$77,$7E,$DA,$E1,$0A,$E0,$DA,$E0,$AF,$C0,$DA,$E3
	.dc.b	$5C,$0E,$33,$A3,$E4,$E1,$EF,$AF,$8E,$D7,$E3,$9E,$42,$F2,$3E,$81
	.dc.b	$E3,$7E,$52,$30,$D8,$CC,$FE,$20,$FE,$D8,$E1,$C4,$F0,$7F,$C6,$DF
	.dc.b	$72,$F2,$3F,$1F,$AF,$F7,$E8,$C4,$FF,$3A,$BE,$53,$FE,$F1,$5E,$C7
	.dc.b	$0E,$4E,$ED,$AF,$F7,$84,$E9,$FF,$29,$F8,$AF,$A2,$7D,$16,$F4,$F7
	.dc.b	$61,$EB,$25,$FA,$CE,$6E,$1F,$F0,$0E,$7F,$EB,$37,$D7,$ED,$F3,$CB
	.dc.b	$C1,$1F,$E4,$5F,$A2,$FE,$86,$FF,$54,$6B,$9E,$3F,$46,$8F,$A2,$FF
	.dc.b	$31,$4E,$E7,$E8,$57,$08,$FD,$AE,$68,$C4,$7F,$86,$DA,$A3,$53,$F1
	.dc.b	$7C,$6F,$F3,$77,$CA,$DF,$58,$FC,$3C,$3C,$BC,$5F,$E3,$AC,$3F,$CA
	.dc.b	$FF,$41,$F1,$FD,$43,$4E,$9F,$45,$F0,$2B,$84,$EB,$4F,$95,$7C,$9B
	.dc.b	$19,$7F,$0A,$D7,$DA,$E6,$7F,$C9,$63,$E8,$BF,$D4,$D3,$FE,$8E,$8C
	.dc.b	$76,$F8,$E3,$FC,$03,$4F,$9F,$F4,$12,$3F,$07,$ED,$FF,$C0,$07,$71
	.dc.b	$F9,$D9,$E1,$1E,$68,$C1,$1F,$F4,$76,$FF,$9F,$C3,$FC,$AA,$DE,$9F
	.dc.b	$F2,$65,$F2,$AF,$F4,$37,$D1,$77,$6F,$C2,$7E,$46,$31,$87,$F9,$5B
	.dc.b	$27,$CF,$E6,$BF,$61,$5C,$6F,$F0,$2D,$7F,$CB,$4D,$7F,$00,$F4,$C7
	.dc.b	$FD,$6B,$F0,$7F,$A5,$F6,$3F,$D7,$AF,$5A,$FF,$13,$D3,$E2,$3C,$84
	.dc.b	$8F,$8F,$E1,$57,$F7,$73,$B7,$EE,$B5,$CC,$FA,$3C,$29,$1A,$6B,$CB
	.dc.b	$E3,$3D,$A4,$79,$BF,$03,$E3,$24,$76,$6B,$E2,$DB,$E2,$91,$F1,$BF
	.dc.b	$DA,$97,$D1,$3D,$6F,$E0,$D1,$86,$C6,$63,$FC,$93,$F4,$4C,$33,$5C
	.dc.b	$EB,$55,$FC,$15,$AF,$F2,$16,$BF,$DA,$9D,$F1,$7F,$9B,$7F,$01,$7A
	.dc.b	$5F,$B3,$6D,$7F,$BA,$0F,$21,$F9,$6F,$81,$FB,$1C,$FF,$E4,$18,$FC
	.dc.b	$8F,$A5,$FC,$23,$5F,$F6,$1C,$6C,$C9,$0F,$F7,$35,$72,$F8,$C9,$1F
	.dc.b	$9A,$79,$7C,$76,$23,$FD,$3B,$82,$71,$9F,$E5,$6B,$E4,$FC,$67,$F9
	.dc.b	$C6,$FD,$0D,$53,$FE,$C2,$51,$98,$FF,$47,$7D,$1F,$E9,$5D,$47,$19
	.dc.b	$FE,$94,$D7,$FC,$26,$D4,$74,$F6,$BE,$F4,$75,$EA,$CE,$4C,$1F,$2B
	.dc.b	$FC,$85,$B8,$7A,$30,$F9,$5C,$FF,$3B,$B2,$4F,$F0,$68,$CB,$FC,$02
	.dc.b	$32,$FB,$B1,$FF,$5E,$EF,$FA,$1D,$3F,$81,$F6,$FD,$82,$57,$F1,$0F
	.dc.b	$46,$BF,$3A,$2C,$FF,$01,$9F,$CA,$FD,$1F,$3C,$3F,$C6,$6D,$BF,$82
	.dc.b	$D5,$F4,$5F,$DA,$1F,$21,$AC,$FF,$45,$3F,$B5,$8F,$EE,$CD,$7A,$C6
	.dc.b	$8F,$D2,$CF,$F7,$0A,$FF,$21,$B6,$7A,$1F,$C8,$BB,$BF,$2A,$47,$F9
	.dc.b	$6F,$9F,$5F,$C5,$C5,$CB,$F8,$F4,$FF,$53,$FA,$37,$75,$F4,$A6,$1F
	.dc.b	$7A,$FF,$DA,$95,$A8,$C6,$5F,$7A,$31,$1F,$03,$91,$1F,$D4,$E7,$57
	.dc.b	$DF,$0F,$E0,$3F,$16,$FA,$3A,$4C,$7F,$65,$AF,$F0,$17,$F6,$93,$B7
	.dc.b	$FC,$EA,$D7,$FC,$0C,$D7,$FC,$44,$D7,$FC,$D6,$EF,$3F,$D2,$4F,$96
	.dc.b	$7F,$36,$D5,$3B,$C7,$F3,$7F,$2A,$F1,$7F,$BD,$3F,$C0,$EB,$E0,$37
	.dc.b	$0F,$F7,$D3,$B8,$F8,$E3,$54,$66,$38,$F6,$F4,$3A,$DF,$F0,$DF,$13
	.dc.b	$F8,$8F,$8F,$2B,$FB,$4C,$65,$FE,$C7,$19,$7F,$23,$E9,$BA,$3F,$11
	.dc.b	$E6,$3E,$8F,$FC,$1F,$5F,$E0,$46,$BF,$E4,$66,$BF,$E1,$E6,$BF,$C0
	.dc.b	$23,$97,$A3,$FC,$93,$C3,$93,$FF,$00,$2B,$9E,$FF,$E1,$95,$F4,$7F
	.dc.b	$E3,$F6,$CE,$E3,$8B,$FA,$8C,$3C,$D1,$FE,$5F,$9D,$CF,$96,$30,$67
	.dc.b	$E8,$A4,$7A,$DA,$FA,$DF,$E7,$3E,$24,$8F,$34,$7F,$52,$F9,$E7,$C4
	.dc.b	$3E,$4B,$F8,$EF,$C2,$3E,$7B,$FD,$67,$5A,$9C,$C6,$2B,$86,$1F,$ED
	.dc.b	$B8,$FF,$5C,$FD,$03,$E8,$7F,$DF,$91,$BB,$E3,$5F,$CA,$B0,$E8,$F1
	.dc.b	$31,$87,$4C,$7C,$51,$FC,$C3,$1E,$8F,$21,$5C,$A7,$71,$8A,$E5,$87
	.dc.b	$86,$3B,$0B,$E6,$E2,$F9,$4E,$EF,$9C,$62,$7C,$58,$FB,$1F,$3E,$3E
	.dc.b	$39,$D4,$78,$63,$CD,$87,$6B,$E7,$8F,$92,$B8,$9E,$17,$0F,$97,$0E
	.dc.b	$CF,$33,$57,$DE,$77,$7D,$71,$F4,$12,$BA,$E1,$F1,$27,$51,$89,$D3
	.dc.b	$D6,$75,$87,$83,$1F,$59,$30,$EF,$19,$7C,$F1,$9C,$7D,$F3,$97,$CC
	.dc.b	$4B,$FA,$38,$F8,$9F,$06,$3E,$2B,$F9,$33,$97,$A4,$FC,$03,$8D,$F5
	.dc.b	$71,$8F,$CC,$C3,$C3,$1D,$E3,$0E,$B0,$FC,$AD,$71,$9C,$92,$33,$3C
	.dc.b	$5F,$26,$3F,$02,$3D,$47,$FB,$3F,$0F,$CC,$B6,$3F,$46,$DC,$3F,$D7
	.dc.b	$4F,$E5,$BC,$EF,$E4,$0F,$EB,$1A,$FE,$8D,$AE,$F9,$F8,$DF,$F8,$E1
	.dc.b	$FC,$74,$FF,$2D,$C7,$F3,$FF,$43,$FE,$0E,$EC,$CD,$4C,$3F,$39,$7B
	.dc.b	$CF,$FC,$33,$0B,$E0,$EA,$B8,$61,$FA,$0B,$E1,$3F,$E2,$0A,$DC,$E6
	.dc.b	$B7,$7E,$B7,$F6,$93,$A8,$C5,$75,$25,$7A,$3C,$C6,$3C,$A7,$E1,$61
	.dc.b	$5F,$34,$E9,$E3,$7D,$2F,$C5,$1F,$BC,$C3,$C0,$F2,$36,$58,$F0,$61
	.dc.b	$D7,$2F,$99,$3A,$3B,$C6,$72,$F8,$31,$F4,$CE,$D8,$F8,$EF,$C2,$4C
	.dc.b	$7C,$87,$CE,$C7,$C7,$5C,$67,$4F,$8E,$75,$5C,$DE,$31,$92,$1E,$43
	.dc.b	$CD,$8F,$A2,$B9,$C6,$5B,$19,$9E,$73,$96,$C6,$63,$AD,$F9,$A3,$11
	.dc.b	$E7,$9E,$33,$9C,$3E,$1D,$68,$DC,$67,$2F,$E7,$BC,$6F,$9F,$7C,$37
	.dc.b	$FB,$AB,$5F,$F3,$4F,$AF,$FE,$03,$F9,$6C,$6A,$C3,$85,$E2,$B5,$7F
	.dc.b	$E0,$A9,$CB,$FE,$FC,$C3,$AB,$6B,$FD,$01,$7F,$DD,$E7,$2F,$E8,$0E
	.dc.b	$CB,$E2,$B7,$18,$6D,$FA,$5B,$3B,$BF,$04,$7A,$27,$51,$FE,$F4,$F5
	.dc.b	$F1,$5B,$BF,$EF,$8B,$E7,$FD,$9E,$27,$FA,$F6,$FD,$93,$97,$91,$89
	.dc.b	$F0,$91,$F8,$0F,$A3,$C0,$FF,$0F,$79,$1D,$0F,$96,$E2,$7F,$C6,$3A
	.dc.b	$FF,$8C,$67,$3E,$0F,$F9,$82,$73,$8D,$CB,$FC,$53,$37,$BF,$F6,$18
	.dc.b	$71,$BF,$F1,$DD,$FF,$8E,$BB,$7E,$CB,$FE,$3C,$F3,$6F,$8C,$9B,$18
	.dc.b	$6D,$72,$9D,$36,$7F,$07,$93,$E7,$DE,$97,$E1,$A9,$F3,$F8,$7F,$DB
	.dc.b	$13,$6B,$FD,$23,$1A,$AE,$73,$99,$FC,$13,$5F,$E4,$6D,$7F,$D7,$AD
	.dc.b	$7C,$51,$C8,$F1,$57,$FB,$56,$BC,$31,$F0,$BC,$F7,$CA,$5F,$9E,$C3
	.dc.b	$6B,$E4,$CE,$9B,$3A,$69,$F5,$3D,$0F,$C3,$9F,$60,$AE,$49,$8F,$D8
	.dc.b	$8C,$BF,$EB,$23,$FC,$79,$EC,$7A,$3C,$EF,$9B,$E2,$7F,$B9,$2F,$FC
	.dc.b	$EA,$7C,$C8,$CC,$FE,$92,$B8,$CF,$F5,$AA,$E1,$1F,$AB,$32,$57,$3C
	.dc.b	$7F,$EF,$16,$CF,$FB,$A0,$FB,$3D,$DD,$9F,$C8,$FC,$FF,$E9,$0F,$CA
	.dc.b	$95,$C9,$2F,$F2,$B1,$97,$F2,$FC,$DF,$AC,$3F,$3A,$59,$FF,$48,$F4
	.dc.b	$7F,$A0,$E3,$13,$C2,$7F,$43,$1F,$90,$F9,$E6,$DF,$9C,$F9,$F7,$C4
	.dc.b	$E6,$30,$DB,$FE,$FB,$5C,$1B,$3F,$26,$31,$87,$C7,$6E,$BF,$D6,$E1
	.dc.b	$18,$A4,$7F,$A9,$F0,$7C,$27,$78,$FF,$76,$F5,$FF,$AA,$FE,$77,$F8
	.dc.b	$59,$FB,$C2,$DF,$EF,$1B,$3F,$BC,$6E,$3F,$DA,$F5,$FF,$53,$7C,$57
	.dc.b	$F1,$43,$F8,$29,$6F,$E4,$EC,$8C,$61,$FC,$19,$B7,$F7,$FD,$6F,$E0
	.dc.b	$7F,$A2,$B8,$7F,$1F,$18,$6E,$3F,$8B,$8F,$C7,$71,$FF,$B4,$23,$04
	.dc.b	$C7,$F3,$51,$97,$FD,$4D,$F4,$1F,$C9,$DC,$D7,$F3,$3F,$80,$7B,$1F
	.dc.b	$CE,$9B,$87,$E5,$FC,$74,$62,$FF,$A2,$46,$2B,$87,$AF,$2F,$85,$A3
	.dc.b	$17,$C2,$1E,$FF,$72,$57,$07,$C5,$ED,$48,$E6,$DA,$E5,$E6,$F9,$C9
	.dc.b	$3E,$06,$C7,$FB,$16,$D1,$FC,$FB,$E6,$9F,$3F,$F0,$63,$E1,$73,$7F
	.dc.b	$69,$E7,$FF,$16,$E3,$FD,$A7,$04,$E6,$36,$7B,$9F,$90,$47,$07,$17
	.dc.b	$C1,$F0,$FC,$32,$FF,$6C,$7C,$FF,$FB,$2A,$BF,$A7,$E1,$E4,$7F,$BC
	.dc.b	$46,$67,$F3,$33,$97,$8F,$B1,$FD,$C2,$75,$1F,$E5,$6E,$14,$62,$3F
	.dc.b	$DE,$4D,$7F,$C1,$0F,$5E,$7F,$F2,$34,$F0,$AF,$8D,$EA,$FF,$77,$AD
	.dc.b	$63,$CB,$C6,$FF,$08,$E0,$F9,$FF,$D7,$CF,$A2,$73,$18,$CB,$DF,$1E
	.dc.b	$72,$4F,$E3,$6F,$F9,$E4,$FE,$31,$B7,$E4,$6C,$FE,$FB,$CE,$7C,$F3
	.dc.b	$E7,$FC,$D8,$F9,$FE,$7F,$ED,$70,$5C,$3E,$6B,$73,$ED,$3F,$24,$93
	.dc.b	$E3,$75,$9F,$EF,$F9,$7E,$0D,$7C,$89,$FF,$0D,$7C,$FF,$E1,$A3,$F8
	.dc.b	$6D,$FF,$13,$7F,$C6,$91,$9B,$FF,$01,$CE,$78,$53,$C6,$BE,$C3,$5F
	.dc.b	$E3,$71,$A9,$FF,$16,$CE,$67,$AC,$78,$8F,$F0,$CF,$CF,$FE,$D9,$FF
	.dc.b	$20,$4F,$85,$E6,$B8,$7F,$C8,$D0,$F8,$4F,$5C,$61,$B1,$FC,$06,$78
	.dc.b	$36,$7C,$67,$F5,$FB,$FE,$63,$EF,$7F,$32,$F1,$9F,$11,$FE,$A8,$F9
	.dc.b	$F7,$9F,$FE,$93,$9C,$DF,$F9,$36,$F8,$5E,$EB,$4F,$F6,$78,$FE,$7B
	.dc.b	$F3,$DC,$29,$87,$4C,$B3,$7B,$95,$6E,$B5,$18,$9D,$4E,$70,$FF,$53
	.dc.b	$3D,$67,$D6,$6C,$FE,$9B,$5C,$A3,$50,$C3,$C2,$EA,$B9,$DF,$63,$D9
	.dc.b	$86,$A7,$FD,$97,$57,$0E,$95,$B9,$E5,$5B,$7C,$78,$79,$E3,$11,$E1
	.dc.b	$3C,$C7,$C4,$C3,$C2,$FC,$D3,$FB,$1D,$78,$9A,$67,$1E,$8E,$0F,$29
	.dc.b	$D3,$0F,$2B,$CF,$0D,$9F,$0B,$1F,$56,$1E,$09,$E3,$3A,$C3,$EA,$47
	.dc.b	$FB,$F2,$BF,$DF,$AF,$CE,$30,$7F,$89,$7E,$77,$FD,$1D,$F3,$FF,$E2
	.dc.b	$78,$46,$1D,$7B,$E9,$D4,$E6,$7F,$D3,$0F,$F8,$4B,$8B,$3E,$89,$A4
	.dc.b	$9F,$04,$65,$F9,$7F,$3F,$FE,$B2,$7E,$C3,$5F,$9F,$5D,$5B,$3C,$73
	.dc.b	$FC,$B5,$6D,$BF,$03,$FE,$16,$AD,$E1,$F8,$3C,$39,$56,$A7,$BD,$6E
	.dc.b	$3F,$41,$3A,$D3,$93,$6B,$FC,$43,$87,$79,$EC,$D8,$FF,$12,$CE,$DF
	.dc.b	$97,$8F,$9E,$B7,$87,$76,$CF,$96,$74,$F2,$6A,$BE,$CD,$AE,$4E,$F1
	.dc.b	$E4,$7B,$18,$7C,$22,$63,$F0,$5F,$A3,$3C,$9F,$3C,$F2,$72,$72,$9F
	.dc.b	$8B,$7E,$6E,$57,$CE,$3C,$6F,$A1,$87,$FA,$D3,$E7,$7F,$C2,$38,$FF
	.dc.b	$AF,$9A,$FE,$17,$93,$F9,$8D,$F1,$8F,$F7,$53,$5F,$F2,$5C,$F0,$AE
	.dc.b	$9E,$47,$6B,$E3,$85,$BF,$F7,$B6,$0F,$63,$1E,$2B,$27,$B3,$86,$DF
	.dc.b	$82,$B9,$69,$EC,$75,$52,$74,$43,$CA,$E5,$F6,$1E,$28,$E7,$F3,$9F
	.dc.b	$E7,$68,$FF,$6A,$FA,$7F,$D0,$DB,$DB,$AE,$09,$18,$21,$D3,$15,$F3
	.dc.b	$8C,$37,$81,$BF,$F7,$9C,$7B,$57,$93,$C5,$FF,$33,$65,$DA,$3A,$79
	.dc.b	$CF,$9C,$F8,$2D,$8F,$F3,$FE,$3F,$E1,$88,$C4,$73,$F9,$CF,$9C,$EB
	.dc.b	$2F,$ED,$73,$AA,$F1,$CF,$F0,$3F,$9C,$FF,$43,$BB,$31,$E0,$78,$31
	.dc.b	$91,$9A,$57,$08,$F9,$B1,$8A,$DC,$7A,$BE,$73,$D2,$47,$FD,$0B,$EF
	.dc.b	$D7,$FD,$2B,$AF,$F0,$56,$99,$BF,$75,$7E,$22,$75,$1E,$AE,$54,$63
	.dc.b	$4F,$7D,$7E,$23,$E7,$18,$D2,$63,$FE,$96,$F9,$7A,$F8,$BE,$03,$B1
	.dc.b	$8C,$FF,$C8,$B7,$CA,$7E,$36,$7E,$0A,$DC,$7C,$9E,$45,$6E,$3F,$0C
	.dc.b	$BE,$13,$C8,$48,$FF,$53,$F8,$15,$DF,$67,$A4,$C6,$5E,$27,$F5,$39
	.dc.b	$3D,$8A,$6E,$BC,$58,$BE,$2B,$E1,$7C,$44,$9C,$C5,$7E,$CF,$9D,$FE
	.dc.b	$7E,$7E,$F3,$5F,$4C,$FC,$BF,$7F,$C3,$F4,$92,$7B,$C7,$EF,$D7,$CE
	.dc.b	$3D,$2F,$E0,$5C,$EF,$9C,$FF,$19,$57,$E3,$E3,$1E,$87,$1B,$E1,$3F
	.dc.b	$93,$39,$DF,$2B,$E3,$1F,$A9,$6F,$0F,$FB,$96,$B8,$CF,$F0,$03,$89
	.dc.b	$DE,$77,$D3,$FE,$6B,$6C,$7F,$BA,$A7,$8D,$7F,$B7,$F2,$FF,$5D,$4F
	.dc.b	$E0,$32,$FF,$1D,$B7,$CF,$FE,$3D,$F0,$6A,$DC,$FF,$10,$8C,$7B,$FF
	.dc.b	$99,$A9,$3F,$EE,$68,$FC,$E7,$CE,$48,$EC,$F4,$B7,$C6,$66,$7F,$4F
	.dc.b	$37,$2F,$34,$FD,$1F,$89,$FF,$6B,$D7,$CD,$F9,$C7,$C9,$FD,$26,$C7
	.dc.b	$CD,$9D,$BF,$1D,$B7,$FC,$1B,$55,$70,$AF,$E3,$0D,$9F,$E1,$70,$AE
	.dc.b	$3C,$1F,$DF,$DB,$1F,$B6,$AE,$13,$FE,$9B,$6C,$0E,$5F,$05,$F0,$7B
	.dc.b	$7F,$80,$78,$32,$7F,$8B,$70,$FC,$F9,$D4,$7D,$59,$D3,$F4,$77,$FD
	.dc.b	$83,$F6,$7C,$1F,$E1,$44,$79,$C7,$F5,$BC,$BC,$5D,$BF,$C3,$2B,$55
	.dc.b	$FC,$59,$FA,$9E,$87,$E7,$FE,$73,$AC,$FE,$3F,$E6,$BF,$B3,$70,$7C
	.dc.b	$46,$E3,$FE,$21,$AD,$CF,$C8,$6C,$7F,$89,$67,$75,$F0,$DA,$A7,$6F
	.dc.b	$C7,$6A,$8D,$9C,$23,$78,$FF,$26,$F0,$BF,$95,$12,$BE,$0F,$A1,$5F
	.dc.b	$E6,$71,$71,$FD,$14,$FE,$EB,$E7,$1D,$F7,$D1,$98,$7E,$EE,$72,$D9
	.dc.b	$CB,$FD,$7E,$3F,$AA,$B6,$3F,$84,$78,$8F,$9C,$FE,$15,$AE,$B2,$F9
	.dc.b	$58,$EA,$66,$1F,$C1,$6B,$5B,$7C,$9A,$CC,$FF,$01,$C3,$A3,$B8,$EA
	.dc.b	$6B,$0F,$CF,$7C,$93,$C9,$FE,$2E,$FE,$FA,$BF,$BB,$36,$77,$3A,$37
	.dc.b	$DF,$2B,$F8,$E5,$73,$7F,$1B,$E4,$7F,$59,$F3,$A4,$60,$87,$F6,$27
	.dc.b	$2F,$E6,$63,$FC,$43,$F3,$9F,$E0,$8C,$7F,$B0,$77,$7F,$0E,$6F,$85
	.dc.b	$AF,$46,$61,$FC,$5E,$3F,$B4,$18,$7F,$C3,$16,$F9,$CF,$E9,$7F,$73
	.dc.b	$27,$FA,$D4,$E5,$B3,$97,$FC,$57,$D0,$15,$E6,$9F,$3F,$CE,$FF,$77
	.dc.b	$F1,$65,$F4,$8C,$BF,$CA,$E3,$2F,$92,$3F,$A5,$7C,$D3,$E2,$DF,$21
	.dc.b	$FE,$EE,$F9,$BF,$F4,$54,$E6,$31,$7F,$C2,$63,$0F,$FB,$56,$7F,$D2
	.dc.b	$1F,$3A,$F9,$DF,$F6,$8C,$72,$C3,$AD,$6C,$E4,$E3,$E1,$93,$BC,$7B
	.dc.b	$D3,$0E,$8E,$31,$F0,$DF,$08,$E7,$59,$C3,$A6,$1E,$4C,$7C,$75,$A3
	.dc.b	$C4,$30,$E7,$3E,$18,$CE,$3E,$07,$73,$A7,$96,$3E,$17,$96,$1E,$07
	.dc.b	$F9,$75,$71,$9D,$57,$19,$DE,$3E,$73,$DF,$5B,$C3,$C5,$7C,$30,$E9
	.dc.b	$7C,$0F,$2E,$3E,$C7,$13,$BC,$3A,$46,$8B,$19,$7D,$75,$D3,$0F,$04
	.dc.b	$7B,$30,$F0,$91,$F5,$E1,$DF,$1F,$51,$2F,$94,$65,$F3,$C6,$71,$F6
	.dc.b	$CE,$5F,$1D,$F5,$BF,$BD,$39,$7C,$B3,$A7,$C2,$53,$DF,$3C,$70,$F4
	.dc.b	$C6,$0F,$51,$0E,$A7,$1B,$F2,$3F,$14,$EE,$4C,$7E,$31,$F7,$67,$AC
	.dc.b	$FE,$82,$7F,$CD,$4D,$7F,$10,$63,$1F,$CF,$BA,$F4,$7F,$8B,$37,$A7
	.dc.b	$FC,$E4,$FE,$AE,$32,$FE,$C1,$AE,$63,$F4,$11,$FC,$7F,$0F,$C8,$7C
	.dc.b	$EF,$F8,$43,$B7,$35,$2F,$F2,$F7,$BC,$FF,$C0,$10,$BE,$0E,$AB,$86
	.dc.b	$1F,$96,$BE,$13,$FE,$07,$AD,$CE,$6B,$77,$E7,$7F,$6F,$3A,$8C,$57
	.dc.b	$52,$57,$A3,$CE,$FF,$52,$BC,$E3,$33,$C2,$35,$8F,$69,$F8,$46,$8F
	.dc.b	$25,$72,$AE,$04,$8D,$D7,$27,$81,$FB,$FC,$7D,$AD,$C7,$D3,$7C,$E3
	.dc.b	$8D,$F3,$74,$7F,$08,$C7,$DC,$E8,$86,$B0,$F0,$1E,$D2,$4F,$39,$E3
	.dc.b	$3B,$8E,$D9,$B1,$F2,$D6,$A3,$18,$FC,$3B,$E0,$7C,$2C,$B1,$3A,$8F
	.dc.b	$7D,$71,$BF,$7C,$E7,$1F,$1D,$71,$3E,$2C,$67,$1F,$36,$5F,$0E,$7B
	.dc.b	$92,$74,$F1,$AE,$7E,$BB,$E7,$5F,$09,$FE,$E2,$EA,$FE,$D7,$F2,$77
	.dc.b	$6F,$F2,$95,$EE,$B1,$8F,$F5,$FE,$C5,$E2,$B5,$7F,$E8,$58,$C6,$3D
	.dc.b	$A3,$07,$FB,$15,$7C,$4F,$4E,$A7,$A1,$FE,$FF,$AD,$4E,$63,$19,$34
	.dc.b	$FF,$7C,$E5,$EB,$AD,$7B,$FE,$06,$E1,$DF,$E7,$7B,$6B,$FE,$C3,$8C
	.dc.b	$1F,$E9,$BF,$73,$CF,$3A,$6C,$64,$FF,$32,$F9,$5C,$8F,$87,$A7,$FA
	.dc.b	$7D,$C4,$68,$FF,$53,$3A,$F4,$7F,$C8,$F1,$83,$FC,$4B,$DD,$F3,$31
	.dc.b	$78,$DF,$EE,$59,$C9,$FE,$2A,$F4,$F5,$1F,$EA,$28,$C1,$AE,$0F,$F1
	.dc.b	$AC,$FD,$65,$F3,$BF,$F0,$02,$7F,$D1,$9C,$27,$FD,$1C,$3E,$77,$55
	.dc.b	$FE,$C7,$AB,$5B,$8E,$13,$8A,$FE,$2F,$56,$73,$EA,$F8,$63,$13,$D6
	.dc.b	$30,$7E,$29,$C1,$FE,$DD,$8D,$D7,$3C,$9E,$8B,$E4,$7E,$32,$F9,$DB
	.dc.b	$18,$6E,$1F,$EE,$B9,$D3,$67,$4F,$FB,$67,$CF,$FC,$19,$F5,$4A,$E4
	.dc.b	$99,$FF,$AA,$63,$2F,$FB,$AF,$C5,$FF,$5F,$4F,$CC,$3E,$5E,$4F,$3F
	.dc.b	$E9,$4E,$AB,$F3,$15,$A9,$CF,$B5,$F0,$EF,$84,$FF,$BF,$EF,$86,$FF
	.dc.b	$46,$3E,$B5,$F0,$7C,$18,$7D,$48,$C3,$D8,$C4,$7F,$BD,$FE,$6B,$F4
	.dc.b	$27,$E4,$CA,$E4,$91,$FE,$C7,$8C,$BF,$EC,$9E,$4F,$D4,$9F,$9B,$2C
	.dc.b	$FF,$B2,$1B,$1B,$8F,$2F,$63,$81,$5F,$95,$F9,$DF,$AB,$7E,$79,$7C
	.dc.b	$EB,$E2,$F5,$C7,$FA,$58,$C6,$5F,$02,$D8,$72,$D7,$FD,$ED,$5A,$9F
	.dc.b	$84,$6A,$F9,$61,$D3,$0F,$93,$8F,$82,$31,$5E,$3B,$E5,$87,$FA,$3B
	.dc.b	$E6,$BF,$84,$9F,$BB,$2E,$1E,$EC,$11,$8B,$FF,$78,$36,$BF,$55,$D7
	.dc.b	$F0,$BE,$2F,$F8,$99,$FC,$14,$B8,$7F,$BC,$9B,$7F,$EF,$96,$D7,$EC
	.dc.b	$E3,$FD,$07,$EB,$FF,$09,$F9,$D7,$0F,$E4,$23,$0D,$C3,$DF,$3A,$7F
	.dc.b	$1D,$3A,$69,$A9,$E1,$5F,$56,$B8,$3F,$8D,$F0,$DF,$00,$F6,$7F,$9D
	.dc.b	$7C,$53,$E0,$7F,$A2,$4E,$72,$FC,$1C,$E5,$FC,$CD,$F2,$F6,$7F,$A6
	.dc.b	$EA,$FA,$2F,$E1,$93,$9A,$FC,$FC,$E7,$C1,$EB,$F0,$5D,$B0,$B5,$D3
	.dc.b	$24,$78,$BD,$DF,$2F,$C0,$DA,$F1,$BE,$6F,$3B,$1C,$1E,$DC,$3F,$AE
	.dc.b	$D7,$3B,$D3,$D6,$BF,$A0,$FC,$4B,$E7,$73,$87,$1E,$8F,$EB,$9D,$FF
	.dc.b	$CB,$F0,$F2,$46,$1F,$8A,$D7,$27,$EA,$23,$84,$F3,$8C,$BF,$BD,$E9
	.dc.b	$60,$F8,$9E,$75,$F3,$3E,$73,$FD,$90,$F3,$CB,$FB,$B4,$63,$0F,$F6
	.dc.b	$CE,$3E,$9A,$E1,$ED,$FA,$2F,$6D,$8C,$92,$BC,$58,$FF,$88,$E3,$1C
	.dc.b	$7F,$DD,$31,$FE,$71,$1F,$63,$E7,$5C,$FE,$2F,$27,$EF,$C6,$27,$D1
	.dc.b	$18,$9D,$47,$16,$E3,$F9,$1B,$EA,$FF,$09,$F9,$D7,$CE,$EB,$2F,$17
	.dc.b	$CE,$FD,$2E,$0F,$82,$DC,$BE,$33,$70,$E8,$DA,$E3,$3F,$4E,$B8,$B4
	.dc.b	$FC,$CF,$CE,$7D,$69,$FE,$CD,$C3,$FD,$B7,$C6,$FF,$12,$4E,$69,$8A
	.dc.b	$B5,$1C,$8F,$F1,$11,$FC,$6C,$FF,$18,$CF,$19,$E1,$19,$5D,$89,$D5
	.dc.b	$74,$9F,$DB,$46,$2B,$85,$6F,$DE,$EF,$F0,$FE,$66,$CF,$B1,$B7,$C9
	.dc.b	$B9,$7F,$11,$8F,$5F,$B9,$EC,$F9,$87,$6F,$FA,$6B,$E7,$3F,$AD,$D2
	.dc.b	$72,$FF,$A7,$85,$71,$AE,$1E,$77,$F8,$3E,$B7,$5F,$EA,$4A,$D3,$F2
	.dc.b	$AF,$AC,$EA,$3C,$67,$7B,$E7,$97,$08,$E4,$64,$E5,$87,$79,$FF,$5F
	.dc.b	$E1,$BC,$3A,$C6,$2B,$80,$8C,$BE,$B9,$E1,$5C,$E3,$07,$09,$DB,$E2
	.dc.b	$C3,$DB,$18,$8F,$4C,$62,$79,$46,$EF,$D6,$F8,$8D,$B8,$9F,$85,$1F
	.dc.b	$30,$F0,$4F,$27,$67,$C0,$C3,$E2,$38,$3D,$91,$B7,$94,$F8,$E7,$71
	.dc.b	$8A,$F0,$1F,$15,$FF,$73,$9E,$12,$E1,$87,$EA,$9F,$81,$F9,$BF,$F4
	.dc.b	$0F,$CE,$70,$6C,$7F,$BC,$1A,$EF,$DD,$4E,$FC,$FF,$F4,$14,$ED,$AF
	.dc.b	$56,$BF,$EE,$CA,$FF,$06,$6B,$C4,$FC,$27,$CE,$7F,$A8,$A6,$FC,$0F
	.dc.b	$FB,$CB,$5F,$C8,$DF,$09,$FC,$0B,$63,$F2,$B5,$C6,$7F,$0A,$D8,$E3
	.dc.b	$87,$99,$B8,$76,$A8,$74,$9D,$69,$E7,$6C,$7F,$08,$BE,$93,$FC,$2A
	.dc.b	$11,$87,$F8,$6E,$1E,$59,$E5,$7D,$B0,$F3,$E3,$ED,$9C,$DF,$86,$B5
	.dc.b	$1D,$70,$E9,$2C,$3A,$3D,$B1,$F2,$BE,$2C,$3A,$3E,$76,$BC,$63,$73
	.dc.b	$CA,$38,$1E,$03,$A6,$1F,$B7,$8D,$5F,$88,$F1,$BE,$2E,$27,$CE,$3D
	.dc.b	$8F,$89,$7F,$EC,$48,$C6,$1F,$E0,$E6,$BF,$DA,$7E,$89,$F4,$5C,$38
	.dc.b	$BF,$DB,$F5,$AB,$E5,$86,$70,$FE,$5F,$39,$AF,$E7,$7E,$2B,$1E,$75
	.dc.b	$FE,$1C,$FB,$D7,$C5,$71,$D1,$18,$9D,$4F,$5B,$7D,$07,$FB,$92,$72
	.dc.b	$43,$FC,$61,$5C,$5F,$F0,$14,$6D,$CD,$77,$9D,$69,$FE,$E6,$6E,$5E
	.dc.b	$8B,$E1,$3F,$EE,$AF,$49,$F0,$DE,$76,$E3,$FE,$36,$6B,$FE,$77,$F9
	.dc.b	$EF,$84,$FF,$52,$8C,$D7,$3A,$E4,$7C,$9F,$65,$7C,$35,$FF,$24,$4E
	.dc.b	$B0,$EC,$EE,$7A,$DB,$F0,$67,$C9,$7D,$38,$C3,$63,$37,$FE,$7C,$A7
	.dc.b	$A1,$9A,$93,$FE,$4A,$AD,$BE,$B9,$CB,$FE,$A9,$BF,$07,$C9,$75,$32
	.dc.b	$FF,$B8,$F2,$7C,$EB,$D7,$FB,$CD,$BF,$F2,$EB,$5F,$F1,$C7,$C0,$7F
	.dc.b	$99,$5C,$CF,$02,$3F,$EB,$CE,$4F,$F5,$3F,$B5,$88,$FF,$5E,$79,$1F
	.dc.b	$EA,$BE,$6F,$CA,$36,$39,$9F,$E6,$CF,$61,$F1,$FF,$91,$6C,$66,$3F
	.dc.b	$21,$EA,$3D,$ED,$44,$9D,$1E,$39,$D3,$83,$C7,$E9,$BE,$1F,$C6,$D8
	.dc.b	$D7,$0F,$C7,$E0,$7C,$5F,$C6,$F8,$B9,$5F,$1A,$2B,$FE,$8D,$94,$E2
	.dc.b	$B9,$57,$E0,$71,$FF,$57,$DF,$FA,$11,$AF,$D9,$6C,$F8,$6B,$E1,$78
	.dc.b	$D2,$BE,$33,$F0,$BE,$77,$E3,$7B,$D2,$7E,$53,$B7,$FD,$41,$BF,$FB
	.dc.b	$55,$F9,$DE,$47,$CD,$3C,$99,$FF,$07,$87,$CC,$FC,$37,$FD,$A6,$DC
	.dc.b	$FF,$D7,$4D,$F0,$BF,$9E,$7C,$BB,$E0,$79,$4E,$6F,$FC,$92,$D7,$AF
	.dc.b	$17,$7A,$E1,$1F,$EE,$59,$D1,$27,$55,$F8,$0F,$B4,$FA,$EC,$AF,$BD
	.dc.b	$19,$7F,$D6,$F1,$97,$FC,$FF,$1F,$90,$F0,$FF,$C6,$B9,$7F,$98,$1B
	.dc.b	$97,$F9,$7B,$9F,$FE,$00,$79,$BF,$C7,$7B,$5F,$19,$C6,$26,$5F,$EE
	.dc.b	$97,$51,$FB,$DF,$B2,$7D,$55,$25,$7F,$BE,$3E,$73,$FD,$BE,$FF,$B1
	.dc.b	$27,$6D,$9D,$AD,$19,$FF,$9C,$B4,$61,$CA,$BF,$80,$61,$C9,$B8,$72
	.dc.b	$C3,$8C,$2F,$8F,$C4,$FF,$1D,$C3,$F2,$F3,$E5,$F3,$F0,$67,$AA,$46
	.dc.b	$48,$79,$73,$F2,$57,$28,$FD,$6E,$9F,$AC,$23,$FB,$3C,$EB,$A3,$FD
	.dc.b	$DB,$1F,$BC,$8F,$C9,$77,$BE,$72,$92,$BF,$D2,$7A,$FB,$17,$37,$ED
	.dc.b	$9C,$E5,$FE,$0F,$BE,$51,$83,$FD,$36,$E1,$F5,$18,$BF,$CA,$73,$F7
	.dc.b	$8C,$9F,$C3,$E3,$18,$78,$70,$F0,$4F,$F0,$CF,$88,$FE,$16,$47,$F8
	.dc.b	$97,$A5,$5D,$1E,$3E,$5F,$F6,$6C,$FF,$1C,$9F,$EA,$1F,$25,$FD,$82
	.dc.b	$BE,$C2,$F6,$B2,$6B,$C5,$93,$ED,$E5,$56,$E3,$0D,$7D,$F1,$A9,$FE
	.dc.b	$01,$F8,$E7,$F8,$0A,$7D,$ED,$D3,$E7,$37,$83,$F2,$5F,$A5,$3E,$CF
	.dc.b	$8E,$93,$0F,$9C,$2F,$83,$F9,$1A,$E0,$EA,$32,$EE,$79,$3E,$69,$DB
	.dc.b	$FD,$4B,$5F,$DD,$7A,$72,$3F,$A8,$76,$27,$F7,$06,$63,$FA,$8E,$AA
	.dc.b	$CC,$7F,$45,$F8,$8F,$E8,$3E,$84,$8C,$10,$FE,$8E,$E5,$FC,$3F,$9A
	.dc.b	$F2,$55,$FE,$27,$FC,$E5,$F3,$74,$97,$FD,$4F,$E1,$6B,$F9,$0F,$49
	.dc.b	$5C,$23,$F2,$D1,$8A,$FE,$BF,$5F,$D8,$D9,$3F,$D9,$DF,$EB,$3D,$78
	.dc.b	$8F,$E7,$F1,$79,$7F,$DF,$BE,$7F,$E5,$9F,$F0,$17,$E1,$BF,$ED,$B6
	.dc.b	$D7,$F8,$FE,$31,$CF,$FC,$3F,$F4,$BF,$E7,$59,$A7,$35,$FE,$FA,$6C
	.dc.b	$66,$3F,$C5,$13,$FC,$FF,$E7,$5F,$1E,$F9,$7F,$F7,$8F,$AB,$FE,$83
	.dc.b	$9C,$C6,$2C,$23,$2F,$FB,$C2,$7F,$D1,$5F,$3E,$FA,$0F,$F7,$74,$73
	.dc.b	$C3,$B5,$6C,$E7,$7C,$BE,$23,$FC,$17,$1D,$67,$71,$8C,$3A,$4F,$8C
	.dc.b	$D6,$36,$F9,$61,$DE,$35,$5C,$2B,$C0,$70,$2D,$70,$7C,$78,$78,$2F
	.dc.b	$D7,$8F,$AA,$73,$87,$7A,$D4,$60,$EB,$87,$7C,$7D,$78,$F5,$9F,$04
	.dc.b	$62,$B8,$C2,$B8,$63,$E9,$8F,$2C,$EA,$F9,$56,$A7,$C1,$7C,$F0,$F2
	.dc.b	$1F,$03,$1F,$44,$F8,$31,$E9,$5B,$8F,$3C,$EB,$0F,$31,$31,$F2,$E3
	.dc.b	$DE,$B9,$1E,$0C,$39,$9F,$19,$F0,$9E,$8C,$6E,$3F,$67,$1E,$D8,$7A
	.dc.b	$C9,$3E,$7A,$DD,$F4,$7B,$1F,$90,$C3,$C5,$8F,$96,$B7,$39,$C3,$AC
	.dc.b	$62,$3C,$E7,$CD,$8C,$BE,$52,$63,$E5,$78,$57,$57,$E2,$9E,$C3,$FD
	.dc.b	$CC,$7E,$24,$AF,$E8,$9B,$8F,$E7,$9C,$FC,$57,$FA,$B5,$FD,$63,$5F
	.dc.b	$D0,$B5,$FD,$63,$64,$FF,$AF,$6B,$F1,$6D,$CB,$FD,$F1,$F4,$1F,$C7
	.dc.b	$37,$FD,$14,$E7,$0E,$BE,$16,$0F,$D4,$56,$E3,$F4,$33,$B7,$C3,$3B
	.dc.b	$7F,$4F,$3B,$C3,$C3,$39,$6C,$EB,$8F,$FC,$37,$18,$26,$3F,$02,$BC
	.dc.b	$33,$A8,$C5,$78,$A7,$A4,$7B,$27,$53,$EA,$34,$4A,$FE,$0F,$7E,$5A
	.dc.b	$E7,$1E,$09,$E0,$EA,$B9,$C7,$B8,$E5,$3C,$0F,$06,$3E,$28,$E3,$19
	.dc.b	$7F,$90,$DF,$77,$24,$3A,$5F,$7A,$E5,$18,$AE,$4A,$B9,$39,$39,$65
	.dc.b	$F8,$0C,$7E,$3D,$F2,$8C,$4F,$A0,$F1,$E3,$EC,$8C,$E3,$F0,$67,$A1
	.dc.b	$DE,$75,$97,$DF,$AD,$E5,$E4,$CB,$C7,$18,$BF,$1D,$7E,$16,$72,$F1
	.dc.b	$26,$5E,$69,$F2,$70,$7F,$53,$F5,$BF,$CB,$BC,$7F,$D7,$1B,$AF,$F9
	.dc.b	$9F,$F4,$C7,$85,$8C,$3F,$AC,$55,$C3,$95,$FF,$70,$E5,$4E,$FB,$D8
	.dc.b	$F2,$AD,$4F,$1F,$CF,$7F,$AB,$C7,$F7,$86,$E3,$FE,$CE,$F9,$7D,$32
	.dc.b	$74,$F8,$FC,$97,$52,$B5,$3F,$E0,$5C,$FF,$C3,$97,$E5,$F0,$FD,$5E
	.dc.b	$0F,$F8,$47,$E3,$FD,$55,$AB,$E1,$58,$37,$E1,$2B,$18,$7F,$B9,$7E
	.dc.b	$4B,$C5,$9B,$E6,$B7,$18,$6C,$FF,$95,$E7,$4F,$13,$11,$E4,$F1,$FF
	.dc.b	$DC,$31,$FE,$5D,$E2,$B9,$FF,$8A,$26,$F2,$FF,$AC,$7D,$3E,$07,$C3
	.dc.b	$AC,$DF,$F8,$0B,$75,$E3,$07,$0F,$FA,$6F,$0F,$B9,$0F,$9F,$F9,$58
	.dc.b	$7F,$94,$9B,$1F,$E6,$0F,$37,$FD,$F7,$39,$E3,$E3,$5A,$D3,$FD,$18
	.dc.b	$D9,$FF,$58,$57,$0B,$FF,$41,$CE,$5F,$F0,$BB,$5F,$15,$FE,$26,$78
	.dc.b	$5F,$48,$C7,$D0,$67,$DC,$B8,$FF,$9D,$9B,$87,$9D,$B7,$FE,$BA,$DB
	.dc.b	$FC,$FB,$1F,$89,$FC,$AB,$FD,$8B,$7F,$E8,$B6,$DF,$F9,$5B,$91,$F2
	.dc.b	$FB,$A8,$6C,$FC,$2B,$89,$F6,$FC,$37,$C4,$3F,$51,$2F,$A0,$7D,$3E
	.dc.b	$63,$E7,$4D,$AF,$F5,$5C,$6E,$71,$A7,$CC,$AB,$3F,$A8,$BE,$6F,$23
	.dc.b	$D5,$F3,$BE,$D3,$F3,$3F,$83,$58,$C9,$31,$F7,$46,$5F,$F4,$EF,$B9
	.dc.b	$FE,$74,$9F,$F5,$22,$F4,$3F,$78,$E6,$77,$18,$F4,$B2,$DA,$E2,$FF
	.dc.b	$09,$CD,$FC,$01,$FE,$A1,$9F,$F6,$64,$63,$D8,$EB,$E3,$6E,$B5,$8F
	.dc.b	$FB,$3A,$B5,$7C,$27,$31,$F7,$72,$FE,$1D,$5B,$9C,$D7,$FB,$81,$B0
	.dc.b	$AE,$6F,$C0,$BF,$06,$1F,$AB,$F7,$E6,$1F,$EA,$66,$D7,$FA,$91,$B6
	.dc.b	$9F,$D8,$76,$27,$31,$FB,$3E,$37,$C6,$7E,$47,$E7,$3F,$86,$B6,$7F
	.dc.b	$DD,$DD,$FF,$EA,$58,$C6,$5F,$93,$6E,$5F,$EF,$9F,$4D,$F4,$9F,$BA
	.dc.b	$6E,$1F,$EF,$73,$F2,$1E,$66,$66,$47,$FB,$3A,$77,$87,$E4,$39,$97
	.dc.b	$A8,$FF,$7C,$CE,$B0,$F0,$61,$F9,$CB,$DE,$1C,$E3,$F9,$E7,$95,$23
	.dc.b	$EE,$FD,$0B,$B7,$19,$7F,$AE,$7C,$B7,$81,$FD,$0C,$FE,$89,$36,$FF
	.dc.b	$A2,$38,$7E,$2F,$0B,$E9,$F1,$87,$E1,$71,$76,$FE,$3D,$B7,$FE,$C4
	.dc.b	$F8,$CF,$EA,$AF,$92,$7F,$AD,$46,$2B,$FD,$98,$DD,$FE,$E4,$3C,$F4
	.dc.b	$BE,$2B,$8E,$FC,$9B,$8F,$76,$DF,$BF,$CE,$FE,$21,$E8,$FF,$61,$6D
	.dc.b	$7F,$61,$70,$FB,$BE,$4A,$D7,$18,$D1,$92,$BE,$8F,$23,$FB,$27,$99
	.dc.b	$FC,$6B,$5F,$4E,$BE,$89,$CD,$FC,$06,$3F,$33,$AE,$E1,$CE,$3E,$44
	.dc.b	$FF,$62,$9C,$DF,$F8,$23,$E8,$9F,$75,$FB,$D7,$E2,$DF,$EF,$1C,$B7
	.dc.b	$83,$53,$23,$FD,$AF,$5F,$CD,$7E,$8B,$EE,$47,$19,$D6,$5F,$C0,$70
	.dc.b	$E6,$FE,$E6,$7F,$DB,$FE,$62,$F8,$CE,$EF,$FC,$2D,$F4,$7F,$C5,$F1
	.dc.b	$F8,$6D,$C3,$8D,$FF,$3E,$7C,$67,$FB,$AB,$54,$EF,$0F,$C1,$36,$33
	.dc.b	$E4,$7F,$87,$EF,$95,$79,$4D,$C7,$88,$97,$DA,$19,$7A,$7C,$CC,$C7
	.dc.b	$1C,$1F,$73,$F6,$EF,$C8,$D9,$78,$7F,$DF,$BC,$DF,$F1,$A4,$E6,$F8
	.dc.b	$DE,$0F,$F0,$74,$D7,$C5,$1A,$6C,$F5,$AE,$91,$83,$FC,$81,$F4,$9D
	.dc.b	$4D,$7C,$DF,$E0,$31,$F4,$46,$67,$67,$F8,$C2,$75,$BF,$E5,$27,$D5
	.dc.b	$42,$7B,$D7,$E8,$5F,$25,$99,$7D,$23,$E3,$3F,$6B,$5A,$9F,$F6,$0D
	.dc.b	$F1,$BE,$17,$BC,$F1,$5A,$17,$C5,$EB,$87,$4B,$FF,$65,$C3,$0E,$2E
	.dc.b	$AE,$BE,$83,$9C,$F8,$2B,$94,$77,$BE,$D8,$78,$E3,$66,$6B,$85,$F4
	.dc.b	$8F,$01,$A3,$C0,$53,$C1,$5D,$2F,$BC,$7C,$7C,$3C,$F8,$77,$AF,$74
	.dc.b	$F0,$C3,$BB,$CC,$EF,$1C,$F0,$F1,$5F,$2C,$3A,$D7,$13,$EC,$61,$D7
	.dc.b	$0F,$8F,$5B,$BE,$35,$E9,$AD,$3E,$F3,$8D,$F4,$7C,$37,$E4,$C3,$CF
	.dc.b	$1F,$08,$FA,$65,$3B,$98,$30,$7E,$2F,$E8,$5F,$48,$FA,$4E,$31,$86
	.dc.b	$C6,$67,$EF,$CE,$5B,$F3,$3F,$E0,$76,$ED,$CA,$31,$1F,$95,$12,$8C
	.dc.b	$C7,$E6,$63,$13,$F8,$AF,$A4,$7C,$E7,$E7,$72,$FC,$94,$62,$FF,$1F
	.dc.b	$7C,$1F,$C1,$36,$77,$5C,$73,$FD,$35,$F1,$BD,$E3,$E9,$BF,$CF,$61
	.dc.b	$CF,$E0,$38,$D7,$EC,$5A,$FF,$AB,$30,$F4,$5F,$C1,$9E,$2F,$38,$EF
	.dc.b	$8F,$86,$BF,$75,$8F,$87,$1D,$E5,$EA,$C3,$85,$6A,$F8,$2B,$E1,$87
	.dc.b	$82,$F8,$1F,$1E,$B7,$3A,$AF,$BD,$7D,$1F,$A2,$D8,$D4,$75,$9E,$F9
	.dc.b	$7C,$9C,$3B,$4F,$EB,$6B,$67,$09,$D4,$76,$C3,$E3,$57,$C0,$8A,$7A
	.dc.b	$CA,$78,$A3,$3C,$AF,$A4,$FF,$22,$4F,$F0,$C8,$C7,$A9,$F9,$C6,$C6
	.dc.b	$67,$6F,$E3,$63,$2F,$F4,$DF,$0D,$F6,$2F,$6B,$F9,$36,$3F,$E1,$89
	.dc.b	$CC,$63,$97,$FA,$E6,$3D,$2B,$FC,$0B,$F4,$9F,$E4,$0C,$7F,$A1,$B5
	.dc.b	$FE,$A4,$D8,$FF,$11,$F5,$3A,$BF,$5B,$5C,$3D,$6F,$55,$F1,$8F,$E7
	.dc.b	$1A,$A3,$11,$FD,$B5,$AF,$F8,$B5,$ED,$97,$F4,$3D,$3F,$98,$7C,$17
	.dc.b	$A6,$BF,$DB,$3D,$68,$C6,$1D,$71,$BF,$26,$FA,$4A,$4C,$3F,$B8,$78
	.dc.b	$4F,$A4,$F8,$5A,$7F,$7F,$BE,$15,$F5,$5A,$FD,$16,$BE,$C6,$BC,$6F
	.dc.b	$CD,$C0,$FB,$07,$66,$1B,$5F,$DC,$2F,$AD,$7F,$93,$87,$D2,$BC,$EE
	.dc.b	$6D,$AF,$F5,$3B,$5F,$F0,$76,$9E,$BF,$97,$DD,$7F,$26,$BE,$6D,$BE
	.dc.b	$17,$F8,$66,$BF,$E3,$27,$A4,$F8,$DF,$F4,$BC,$78,$3D,$EE,$AF,$F8
	.dc.b	$23,$E9,$3F,$CD,$ED,$F5,$BE,$9C,$62,$3F,$C4,$5D,$0F,$80,$7D,$2E
	.dc.b	$AF,$F4,$17,$BB,$FC,$3B,$5F,$F1,$93,$5F,$F3,$93,$D7,$2F,$CF,$6B
	.dc.b	$F8,$89,$74,$7F,$8B,$39,$7B,$47,$F8,$7B,$D8,$7C,$97,$E2,$1B,$19
	.dc.b	$8F,$C7,$6A,$F7,$1F,$3D,$B9,$FD,$B5,$F0,$BF,$DA,$35,$FC,$FB,$5F
	.dc.b	$F2,$53,$BC,$FF,$69,$CD,$FE,$A3,$9F,$F2,$16,$7F,$86,$F7,$BB,$47
	.dc.b	$CD,$EC,$F9,$64,$7E,$87,$CC,$D7,$F8,$27,$07,$E8,$DF,$E1,$8D,$73
	.dc.b	$3F,$83,$D5,$ED,$52,$4F,$F8,$7B,$3F,$81,$1E,$EF,$26,$4F,$F1,$7D
	.dc.b	$7F,$A6,$3E,$E8,$F8,$7E,$CC,$8D,$35,$F8,$BE,$5F,$C8,$23,$CD,$FE
	.dc.b	$F1,$F4,$9F,$0E,$3F,$CD,$01,$C6,$3F,$EB,$28,$C6,$DF,$C0,$D7,$CF
	.dc.b	$7F,$AC,$DF,$F3,$61,$A3,$FC,$0F,$5D,$5F,$EE,$CD,$7F,$C9,$8D,$7F
	.dc.b	$CA,$4F,$83,$C1,$FF,$28,$4B,$DD,$F0,$36,$BB,$D6,$DF,$F3,$9E,$CF
	.dc.b	$A4,$FE,$31,$C7,$FC,$0B,$E0,$2F,$2F,$9E,$FF,$4B,$7B,$7F,$DB,$B7
	.dc.b	$AA,$91,$83,$F0,$FD,$5F,$E5,$7D,$BF,$43,$E7,$76,$8C,$6D,$FE,$6B
	.dc.b	$EA,$FD,$87,$99,$9A,$98,$72,$7F,$3F,$DF,$FD,$26,$3F,$D2,$98,$BE
	.dc.b	$91,$F4,$5F,$C7,$1B,$3A,$6B,$FC,$4B,$E8,$DF,$49,$BA,$FE,$25,$5A
	.dc.b	$AF,$EA,$AD,$7F,$D8,$2D,$57,$E1,$9F,$E8,$1A,$BE,$89,$F8,$66,$1B
	.dc.b	$7E,$3A,$DE,$5F,$A1,$C3,$FD,$F3,$82,$3F,$9F,$D7,$FA,$C1,$B1,$9E
	.dc.b	$EE,$27,$2F,$52,$57,$F4,$8A,$E9,$C3,$FD,$57,$97,$E4,$4F,$F7,$36
	.dc.b	$E3,$FB,$EE,$1F,$E1,$7C,$BF,$1A,$31,$5F,$C0,$2B,$77,$CE,$F7,$A7
	.dc.b	$E7,$7C,$37,$D2,$D2,$61,$F9,$EE,$A7,$87,$88,$FF,$4B,$D7,$07,$FC
	.dc.b	$CE,$D7,$35,$FE,$67,$8C,$67,$FA,$DB,$F8,$B5,$FA,$CD,$BA,$4F,$F1
	.dc.b	$8E,$CF,$CC,$92,$7F,$DB,$5D,$1F,$C9,$FE,$43,$11,$F0,$E6,$D7,$F8
	.dc.b	$23,$81,$5F,$73,$E5,$BF,$9C,$72,$1F,$F5,$E6,$1F,$DD,$9A,$FF,$5E
	.dc.b	$76,$4F,$77,$91,$CF,$C8,$FE,$09,$F4,$BD,$AB,$F8,$1B,$5F,$E1,$AD
	.dc.b	$CB,$F0,$BD,$0F,$92,$E3,$9F,$FA,$2E,$B8,$5F,$F9,$F5,$AF,$FB,$4D
	.dc.b	$AF,$FB,$65,$E9,$CD,$FE,$DB,$D1,$F7,$0F,$8B,$FC,$DB,$6F,$CB,$87
	.dc.b	$EC,$B8,$3F,$75,$E9,$C9,$F9,$BE,$07,$F0,$F7,$FB,$5B,$76,$FE,$C3
	.dc.b	$3B,$9C,$D7,$E4,$3E,$1F,$F8,$7F,$A9,$27,$EB,$F2,$27,$51,$F9,$39
	.dc.b	$D3,$FC,$52,$3F,$C5,$BF,$4C,$FA,$2F,$F2,$B3,$67,$FC,$B4,$D7,$FC
	.dc.b	$AC,$D7,$FC,$82,$E7,$2F,$EC,$3D,$9F,$C8,$8E,$B5,$FD,$CC,$95,$FD
	.dc.b	$DD,$D3,$4F,$EA,$4F,$F7,$22,$3F,$E6,$BF,$9F,$FF,$61,$7B,$66,$92
	.dc.b	$3F,$CD,$CF,$FA,$DE,$FF,$B9,$F8,$7F,$CD,$0F,$F2,$5C,$67,$91,$53
	.dc.b	$E8,$9F,$46,$F9,$8C,$FB,$FF,$E0,$39,$CC,$63,$D7,$E3,$69,$CA,$AF
	.dc.b	$F0,$57,$D2,$BE,$97,$FD,$F1,$7C,$2F,$75,$D7,$1F,$06,$1C,$6B,$51
	.dc.b	$89,$E7,$7C,$AB,$D1,$39,$AD,$E3,$E5,$9C,$B5,$63,$E3,$71,$8F,$8A
	.dc.b	$B8,$C7,$84,$E2,$53,$8E,$1E,$17,$E0,$61,$E1,$77,$8F,$AF,$1F,$06
	.dc.b	$3D,$E7,$96,$3D,$DD,$E3,$E0,$BF,$6C,$79,$30,$E8,$F6,$C3,$07,$4C
	.dc.b	$7C,$58,$73,$9D,$1D,$AB,$8E,$3E,$1C,$3C,$93,$9C,$39,$BE,$D9,$D6
	.dc.b	$3E,$32,$47,$98,$FC,$3C,$69,$C1,$E0,$AE,$A7,$A9,$EC,$47,$33,$F8
	.dc.b	$89,$E5,$3C,$63,$BB,$4E,$F8,$FB,$B0,$F8,$78,$FB,$CF,$81,$7C,$9C
	.dc.b	$46,$8F,$83,$3C,$71,$F7,$E3,$E2,$C3,$A9,$8C,$7C,$38,$74,$BE,$13
	.dc.b	$E9,$9F,$2E,$3D,$F1,$FC,$BC,$FE,$88,$FA,$25,$BF,$D1,$B6,$36,$FE
	.dc.b	$87,$E5,$3F,$D8,$B2,$69,$3C,$0E,$93,$FB,$58,$FF,$6D,$65,$FB,$36
	.dc.b	$E3,$FB,$2F,$A5,$FE,$35,$AF,$E7,$27,$37,$C7,$0C,$1F,$A3,$AD,$C7
	.dc.b	$E5,$27,$6F,$EE,$DB,$8F,$96,$B8,$71,$30,$E7,$7C,$66,$E5,$FE,$89
	.dc.b	$C7,$AB,$E7,$8C,$E5,$F0,$5E,$77,$8B,$F3,$1C,$CF,$D8,$5F,$2A,$E1
	.dc.b	$19,$BE,$8E,$B0,$F2,$39,$9D,$CE,$8F,$24,$E9,$C4,$68,$93,$FC,$92
	.dc.b	$BC,$0F,$92,$76,$D9,$FA,$15,$CD,$E3,$1F,$27,$0F,$62,$BE,$6E,$4F
	.dc.b	$31,$CC,$F0,$57,$37,$06,$EF,$C2,$F3,$8D,$39,$BF,$0E,$3F,$5E,$3E
	.dc.b	$26,$3E,$28,$C5,$F2,$C3,$C7,$18,$BE,$39,$7E,$0A,$72,$FE,$08,$99
	.dc.b	$7D,$8C,$7D,$27,$84,$F3,$CF,$08,$FE,$65,$C5,$FD,$5F,$D9,$FA,$BC
	.dc.b	$7F,$EC,$6E,$BB,$AF,$F5,$2A,$D3,$FE,$49,$8C,$E2,$9C,$BF,$AF,$6B
	.dc.b	$FD,$49,$AF,$3E,$36,$1F,$EA,$D8,$ED,$F3,$3F,$D8,$30,$FD,$6D,$A3
	.dc.b	$13,$D5,$AF,$56,$BC,$8F,$2C,$71,$7C,$FE,$AB,$CE,$FF,$69,$DF,$F6
	.dc.b	$EC,$FF,$2B,$7B,$8C,$1E,$7E,$DE,$35,$C3,$0F,$F4,$3C,$E4,$FF,$36
	.dc.b	$7B,$7E,$0D,$3F,$D7,$11,$FE,$05,$BF,$75,$7C,$79,$CB,$F1,$DA,$FF
	.dc.b	$78,$F2,$5D,$EF,$A3,$F0,$FD,$1F,$D0,$9C,$C6,$31,$FF,$16,$B6,$FF
	.dc.b	$D7,$AD,$D7,$F6,$34,$E1,$76,$FC,$4F,$1A,$E3,$FE,$40,$8C,$1D,$E7
	.dc.b	$38,$FC,$01,$C2,$BC,$E3,$DA,$AE,$1F,$E1,$AA,$D6,$1F,$88,$C1,$F4
	.dc.b	$7F,$2F,$0F,$F5,$3F,$8C,$F3,$F5,$39,$C7,$FC,$B9,$39,$7F,$DD,$F1
	.dc.b	$F4,$EF,$F9,$F3,$5F,$F1,$0B,$5F,$F1,$5B,$5F,$0F,$17,$63,$F3,$7A
	.dc.b	$7F,$A0,$FC,$2C,$5F,$FA,$0F,$E7,$33,$E7,$3D,$CB,$18,$6C,$FF,$A2
	.dc.b	$5B,$8F,$FA,$25,$FC,$8F,$D1,$FF,$35,$9F,$C6,$CF,$CA,$8C,$3F,$EF
	.dc.b	$4F,$97,$DC,$C3,$7C,$2A,$DF,$07,$FA,$76,$B5,$7F,$54,$F9,$BF,$07
	.dc.b	$FB,$6C,$1F,$3E,$8F,$9A,$65,$C4,$7C,$C8,$FC,$BC,$F2,$8F,$EB,$ED
	.dc.b	$3F,$B0,$F0,$70,$6C,$F4,$9F,$69,$E0,$F9,$8C,$4F,$E2,$73,$F5,$7D
	.dc.b	$1F,$C4,$3F,$39,$D4,$FA,$3C,$61,$F0,$36,$56,$30,$FC,$97,$4F,$E9
	.dc.b	$AF,$1B,$FF,$BC,$3D,$BF,$F4,$76,$3F,$3B,$95,$0F,$A3,$FB,$EE,$B1
	.dc.b	$FC,$E4,$EA,$73,$E8,$69,$B7,$CE,$F9,$3F,$6B,$07,$D1,$FF,$AE,$E3
	.dc.b	$C1,$EC,$3E,$1F,$F9,$3D,$6B,$D9,$FC,$7C,$E6,$B9,$5F,$1D,$3F,$D0
	.dc.b	$CD,$7F,$CF,$77,$C6,$7F,$C2,$2D,$8D,$DF,$6A,$E5,$3F,$0F,$D7,$C4
	.dc.b	$FE,$FF,$DB,$98,$7F,$AF,$9B,$87,$D5,$D9,$18,$C3,$F7,$8D,$AF,$F4
	.dc.b	$AF,$07,$F1,$6F,$8B,$FC,$EF,$B5,$F6,$9B,$3F,$EC,$06,$E3,$F6,$39
	.dc.b	$51,$8C,$7F,$C0,$B3,$FE,$9D,$FA,$3F,$CA,$7B,$BF,$C1,$FD,$94,$63
	.dc.b	$1F,$D4,$D6,$DF,$CA,$CE,$9B,$3F,$9A,$8C,$4F,$F8,$E1,$B7,$E0,$6D
	.dc.b	$7F,$51,$ED,$7A,$1E,$1C,$3F,$A5,$F9,$FF,$D6,$2B,$FA,$AB,$71,$FE
	.dc.b	$A9,$E8,$7F,$3E,$F4,$3F,$AC,$36,$5E,$25,$25,$F2,$75,$E2,$48,$EA
	.dc.b	$DC,$3F,$D9,$1F,$45,$FA,$F7,$F0,$6D,$7F,$AD,$35,$6C,$8C,$65,$F9
	.dc.b	$3C,$BC,$DD,$5E,$9B,$FE,$EC,$D7,$FC,$00,$D7,$FB,$F3,$EF,$BF,$EC
	.dc.b	$ED,$C7,$7F,$0F,$FD,$C2,$31,$5F,$E0,$5D,$3F,$AF,$DF,$F9,$0B,$C2
	.dc.b	$FE,$35,$EA,$7C,$CF,$4B,$F9,$2B,$72,$F9,$B1,$F2,$48,$E0,$FE,$53
	.dc.b	$A7,$F8,$16,$3F,$84,$FC,$7F,$1C,$BF,$AB,$4F,$EC,$7C,$CF,$F0,$0E
	.dc.b	$3F,$CA,$30,$F6,$C6,$1F,$43,$F1,$08,$FF,$84,$B4,$FE,$95,$5C,$E3
	.dc.b	$C3,$1F,$D5,$6D,$D5,$FD,$E3,$AB,$9D,$7F,$05,$FA,$27,$D1,$7F,$02
	.dc.b	$7F,$80,$5F,$F6,$BD,$D1,$AA,$E1,$8E,$EF,$FC,$13,$5A,$6D,$6E,$FF
	.dc.b	$D3,$D3,$9B,$FE,$0B,$39,$8F,$F1,$EB,$61,$3F,$DE,$7E,$85,$F4,$7D
	.dc.b	$E7,$E6,$FD,$1F,$F2,$5C,$38,$FB,$3F,$E2,$C8,$C4,$FF,$8A,$71,$FF
	.dc.b	$16,$3C,$3C,$9F,$D1,$4F,$F2,$DC,$BF,$C8,$4D,$7F,$C9,$8D,$7F,$C9
	.dc.b	$4F,$46,$C7,$F7,$4C,$9E,$87,$86,$3F,$C7,$BF,$47,$F6,$EB,$FC,$27
	.dc.b	$F7,$7F,$CD,$1F,$C0,$C6,$79,$13,$9D,$7F,$C7,$F3,$FD,$C2,$BF,$CA
	.dc.b	$CD,$7F,$CC,$0D,$7F,$CB,$CD,$E3,$3E,$16,$E9,$FE,$27,$F4,$3C,$AF
	.dc.b	$F9,$4B,$E9,$3C,$4F,$84,$FF,$36,$38,$F3,$B5,$1F,$C2,$63,$B1,$1F
	.dc.b	$EE,$78,$7D,$8F,$80,$FB,$3C,$4F,$90,$FB,$72,$FA,$47,$91,$FE,$B0
	.dc.b	$AD,$4F,$FB,$8E,$F8,$DF,$0B,$DE,$18,$AD,$0F,$A2,$E2,$F5,$C7,$FD
	.dc.b	$77,$18,$8F,$21,$BB,$E9,$3A,$7F,$DC,$F8,$7E,$42,$B9,$E1,$E0,$BE
	.dc.b	$07,$82,$33,$7C,$63,$17,$D3,$0E,$4F,$48,$E2,$79,$63,$2F,$03,$F3
	.dc.b	$F7,$DE,$FD,$10,$C3,$6F,$13,$99,$EC,$8F,$1D,$6A,$31,$7E,$37,$67
	.dc.b	$33,$FD,$1C,$E7,$1E,$F7,$F3,$67,$B4,$61,$F4,$E1,$E5,$AF,$66,$1E
	.dc.b	$09,$EB,$7E,$6A,$D4,$7B,$4F,$44,$F3,$7F,$C6,$F3,$FE,$38,$7E,$A9
	.dc.b	$83,$F0,$E7,$F0,$2F,$A1,$7D,$1B,$E9,$3F,$C7,$F1,$86,$C6,$67,$88
	.dc.b	$9D,$38,$5F,$31,$FD,$4A,$30,$D7,$C3,$1A,$E3,$EF,$39,$9F,$CD,$CE
	.dc.b	$5F,$BB,$F4,$8F,$9A,$FC,$FE,$7F,$94,$8C,$4F,$DA,$AE,$13,$F9,$86
	.dc.b	$EB,$FA,$6A,$DD,$74,$BE,$73,$E0,$C3,$D1,$87,$16,$C7,$19,$DD,$7F
	.dc.b	$2A,$BE,$77,$BC,$7C,$58,$70,$AD,$5F,$05,$7C,$30,$F0,$5F,$03,$F2
	.dc.b	$71,$99,$E2,$64,$FC,$FE,$3E,$C7,$27,$AC,$D6,$3E,$37,$04,$70,$4B
	.dc.b	$F6,$47,$83,$0F,$35,$75,$8F,$BD,$8F,$AD,$F4,$CF,$AF,$0F,$34,$6B
	.dc.b	$1F,$D8,$F2,$3E,$8F,$FC,$97,$3F,$E9,$AF,$A4,$FF,$39,$B7,$DA,$FF
	.dc.b	$2E,$F1,$FF,$54,$BE,$1E,$AE,$29,$18,$78,$10,$FE,$FF,$87,$67,$FC
	.dc.b	$03,$F4,$7F,$E3,$A7,$FD,$53,$C7,$FE,$E0,$6F,$C2,$F5,$B5,$6E,$78
	.dc.b	$FA,$7E,$1E,$EF,$E0,$D7,$70,$E5,$5A,$3F,$C2,$77,$C7,$6F,$F0,$F4
	.dc.b	$E9,$B5,$DE,$75,$A7,$BF,$E4,$DF,$45,$EF,$B7,$D1,$3A,$F0,$DC,$FF
	.dc.b	$C4,$37,$C1,$B7,$AA,$91,$82,$1F,$E0,$FA,$E4,$D8,$F8,$55,$CE,$3F
	.dc.b	$C2,$9F,$44,$F3,$BC,$8F,$F9,$E2,$3F,$D4,$0D,$77,$1A,$37,$ED,$48
	.dc.b	$DC,$65,$E3,$19,$78,$9F,$E3,$CF,$A1,$FF,$6B,$37,$E1,$7E,$B4,$62
	.dc.b	$3F,$D8,$3F,$16,$FA,$2C,$D7,$D7,$F8,$0F,$9B,$4E,$44,$63,$99,$8F
	.dc.b	$3E,$17,$17,$E4,$63,$0F,$F8,$9A,$F8,$F3,$75,$3B,$E3,$FE,$52,$F6
	.dc.b	$5F,$23,$F8,$B6,$C6,$63,$F3,$5E,$AB,$E0,$5F,$1B,$B9,$FC,$77,$33
	.dc.b	$DE,$A6,$70,$E4,$7F,$8E,$36,$FF,$35,$4E,$E3,$1A,$F9,$E3,$11,$FE
	.dc.b	$62,$CF,$C6,$FF,$99,$36,$FC,$A5,$7B,$FA,$FE,$69,$F3,$BE,$32,$D7
	.dc.b	$FB,$77,$57,$B5,$F9,$56,$C7,$D7,$AC,$4F,$D6,$E6,$93,$C6,$3E,$27
	.dc.b	$89,$F1,$7D,$99,$3F,$26,$2B,$FE,$63,$C1,$8F,$73,$3B,$7D,$69,$FE
	.dc.b	$07,$AF,$F4,$E7,$E0,$9F,$7F,$E2,$BE,$6B,$8E,$54,$62,$3E,$27,$8F
	.dc.b	$FC,$35,$7D,$17,$C8,$C3,$FD,$A7,$F3,$6F,$7B,$4D,$AF,$F6,$8D,$6F
	.dc.b	$8B,$FC,$62,$D8,$FF,$36,$CE,$DF,$F5,$9C,$FF,$8D,$A3,$17,$FE,$E0
	.dc.b	$AE,$15,$E2,$E5,$EF,$84,$8E,$F1,$F8,$AD,$9F,$45,$F8,$8F,$82,$95
	.dc.b	$F0,$BA,$BF,$1D,$3B,$33,$1E,$23,$F4,$9D,$0A,$D4,$7C,$7E,$DF,$F1
	.dc.b	$FE,$9F,$EB,$76,$ED,$FA,$36,$CF,$FA,$83,$9F,$F9,$DB,$D7,$8B,$F5
	.dc.b	$FE,$77,$EC,$C9,$1E,$59,$DE,$1F,$6C,$7D,$0B,$E7,$FF,$A0,$62,$F8
	.dc.b	$D7,$D0,$E1,$B8,$FE,$03,$0E,$37,$FC,$A7,$0E,$53,$F8,$4C,$39,$61
	.dc.b	$C4,$D5,$49,$D6,$1D,$1F,$8D,$5C,$1E,$D7,$D1,$CF,$D0,$FE,$A9,$FE
	.dc.b	$1A,$FF,$51,$C7,$F2,$D4,$8C,$6B,$FE,$B0,$8C,$77,$7B,$A7,$94,$6F
	.dc.b	$CF,$C1,$99,$FE,$0B,$E8,$4C,$7F,$39,$C5,$F2,$B1,$FF,$5D,$F2,$7F
	.dc.b	$0D,$DB,$55,$F4,$EB,$4F,$E0,$2F,$F1,$53,$FA,$1F,$0D,$F4,$3F,$9E
	.dc.b	$CE,$5F,$EB,$0F,$E2,$F1,$B3,$2F,$E2,$C7,$5F,$B9,$FF,$66,$D7,$47
	.dc.b	$93,$5E,$11,$B7,$8D,$75,$7D,$BD,$5F,$C9,$08,$FE,$77,$A5,$F1,$F8
	.dc.b	$8F,$74,$DE,$2F,$E0,$DB,$7F,$2F,$AF,$C5,$FC,$A7,$F4,$4E,$23,$FE
	.dc.b	$A6,$C3,$F0,$CD,$7F,$02,$EC,$9E,$DF,$23,$9F,$91,$FC,$27,$E8,$7C
	.dc.b	$75,$F8,$FF,$96,$7C,$57,$E6,$9B,$1F,$2E,$76,$FC,$96,$E9,$F8,$89
	.dc.b	$DC,$FE,$49,$B1,$FC,$86,$13,$C3,$9B,$FD,$82,$D8,$FE,$BF,$5C,$27
	.dc.b	$F2,$0D,$8F,$F8,$03,$3B,$D5,$D9,$FC,$52,$FF,$7D,$E9,$7F,$6B,$3F
	.dc.b	$AC,$C6,$6F,$F8,$07,$06,$5F,$E1,$13,$A8,$FE,$2D,$F0,$9F,$D7,$3E
	.dc.b	$86,$57,$D9,$8C,$BF,$9A,$8C,$BF,$C5,$71,$FE,$CD,$3F,$E3,$5F,$A1
	.dc.b	$7C,$F7,$F6,$6F,$2E,$5F,$F4,$0F,$45,$5A,$8C,$35,$FD,$14,$F0,$7F
	.dc.b	$91,$4F,$0A,$FE,$F0,$7F,$8F,$FC,$B7,$D2,$D3,$35,$FD,$FE,$70,$FF
	.dc.b	$4D,$9D,$B9,$3C,$9F,$3B,$FE,$AE,$E3,$FE,$EF,$E1,$7F,$7E,$3F,$CA
	.dc.b	$11,$9E,$2F,$F3,$03,$6B,$F9,$AC,$ED,$E6,$7F,$57,$F9,$F7,$D0,$3E
	.dc.b	$61,$3F,$ED,$BF,$5A,$AA,$DC,$EB,$DF,$FF,$01,$C6,$27,$81,$CB,$E8
	.dc.b	$5F,$45,$CA,$B7,$8F,$7A,$D9,$D5,$C4,$6B,$0E,$D8,$F8,$B0,$E7,$78
	.dc.b	$3C,$F8,$F8,$6F,$84,$74,$8C,$57,$1C,$7C,$4F,$B2,$33,$3C,$B1,$F0
	.dc.b	$CE,$5B,$1E,$BA,$E2,$EF,$1F,$13,$E9,$6D,$F4,$86,$3E,$87,$C9,$7D
	.dc.b	$1E,$06,$2B,$9C,$EF,$1F,$59,$31,$F2,$4E,$B1,$F1,$DF,$2A,$D3,$E5
	.dc.b	$3E,$6D,$F3,$BE,$31,$DC,$93,$FE,$3C,$9D,$47,$18,$E6,$F5,$9F,$94
	.dc.b	$FB,$B0,$F5,$63,$EF,$3E,$05,$F2,$9E,$07,$3A,$F0,$DF,$7C,$3D,$37
	.dc.b	$C7,$0E,$67,$C5,$C3,$94,$67,$1F,$01,$F0,$71,$A7,$AB,$1E,$A7,$99
	.dc.b	$C1,$E1,$21,$F5,$30,$F0,$E1,$E0,$AE,$78,$F9,$6A,$DF,$47,$8E,$3E
	.dc.b	$4B,$FC,$D4,$7E,$98,$FB,$65,$C7,$F4,$8D,$8D,$BF,$91,$F9,$5F,$F6
	.dc.b	$33,$7B,$B9,$3F,$9C,$8C,$B6,$32,$FE,$CE,$3F,$DA,$98,$FE,$C9,$B8
	.dc.b	$FE,$E3,$E8,$BF,$8F,$EF,$F9,$49,$CD,$F3,$3F,$4C,$E1,$4F,$07,$38
	.dc.b	$71,$23,$FB,$88,$D6,$1E,$19,$CB,$67,$5C,$5F,$E6,$F8,$C1,$31,$F8
	.dc.b	$13,$AC,$78,$C6,$27,$53,$FC,$53,$0E,$F3,$92,$47,$AE,$7F,$D2,$B3
	.dc.b	$A8,$C4,$E9,$F9,$18,$7E,$E2,$75,$18,$3D,$E4,$AE,$78,$7A,$6F,$96
	.dc.b	$6C,$BE,$43,$71,$E9,$5B,$C3,$C2,$43,$B6,$1D,$32,$F7,$C6,$6F,$93
	.dc.b	$F0,$CF,$16,$3E,$0C,$7E,$B6,$5E,$8A,$DE,$1F,$1A,$FA,$DF,$E0,$71
	.dc.b	$F1,$46,$0E,$D8,$72,$8C,$DF,$2A,$ED,$97,$AB,$2F,$2E,$1C,$49,$97
	.dc.b	$87,$1F,$82,$4A,$E1,$3E,$81,$C8,$E1,$7A,$FF,$E7,$7C,$7F,$D9,$DD
	.dc.b	$B6,$F7,$5A,$9F,$F2,$AC,$E6,$F9,$5E,$FD,$27,$9B,$FE,$10,$BF,$EC
	.dc.b	$F3,$98,$C6,$3F,$DD,$BE,$5F,$FA,$ED,$7F,$AE,$FE,$27,$FB,$03,$FE
	.dc.b	$01,$8C,$CF,$FA,$46,$72,$F1,$3C,$0E,$3A,$3F,$D9,$59,$2B,$8C,$7F
	.dc.b	$B3,$72,$56,$F0,$FE,$EB,$E0,$79,$7D,$CF,$66,$DF,$EC,$59,$FF,$39
	.dc.b	$C6,$27,$51,$FE,$02,$6E,$7F,$E1,$DF,$29,$F4,$3E,$FF,$A2,$F9,$5E
	.dc.b	$1A,$73,$87,$F9,$FA,$73,$18,$BF,$F3,$E4,$62,$F9,$57,$FA,$B7,$BB
	.dc.b	$E7,$79,$67,$AF,$25,$9F,$F5,$F4,$60,$D9,$5C,$9F,$E8,$1F,$A2,$FF
	.dc.b	$80,$2F,$F9,$77,$D0,$FC,$06,$49,$CB,$FE,$57,$8C,$C6,$38,$BF,$CB
	.dc.b	$91,$8D,$FF,$11,$93,$CA,$4E,$71,$EC,$78,$6B,$73,$A8,$FF,$3B,$FC
	.dc.b	$E7,$2F,$39,$EE,$58,$C3,$67,$FC,$FA,$DD,$3F,$CF,$AF,$E4,$FC,$64
	.dc.b	$D3,$E7,$FC,$B5,$7F,$C7,$C5,$AF,$B5,$17,$5F,$F3,$A6,$CF,$83,$FD
	.dc.b	$25,$BE,$7B,$0F,$CB,$8C,$C6,$30,$FF,$55,$55,$BF,$1C,$79,$27,$17
	.dc.b	$FE,$3C,$AB,$39,$BF,$9E,$74,$2C,$FF,$91,$2F,$9B,$E3,$8F,$81,$F4
	.dc.b	$52,$3F,$D5,$9F,$45,$6F,$FD,$F5,$9A,$B1,$87,$FA,$B3,$9D,$78,$C3
	.dc.b	$FD,$B9,$ED,$3D,$0F,$BA,$BE,$5A,$C6,$1B,$3F,$EF,$59,$D3,$67,$4F
	.dc.b	$FB,$F7,$E8,$9F,$45,$F8,$8F,$80,$7C,$47,$F1,$1D,$15,$AB,$FF,$7D
	.dc.b	$CE,$63,$1B,$FF,$BF,$2F,$8D,$75,$C3,$A5,$ED,$B1,$93,$99,$F2,$CF
	.dc.b	$F3,$7F,$86,$EB,$7B,$9F,$B0,$3F,$5E,$59,$E1,$8A,$B8,$4E,$63,$FD
	.dc.b	$41,$DD,$CA,$B7,$5A,$7F,$DC,$1F,$17,$F9,$7F,$6F,$F8,$1B,$63,$27
	.dc.b	$FA,$4E,$32,$D3,$FD,$F5,$EA,$3E,$8B,$F2,$3E,$EF,$EE,$A3,$13,$E6
	.dc.b	$3F,$31,$19,$9D,$C7,$26,$E5,$FF,$00,$6B,$8A,$E1,$FC,$7F,$5B,$0E
	.dc.b	$58,$75,$8C,$3F,$CE,$7C,$DF,$E6,$BC,$3F,$D1,$FC,$BF,$E9,$58,$FF
	.dc.b	$3B,$8C,$78,$1F,$7F,$E8,$BF,$7F,$87,$F4,$18,$C7,$83,$C4,$DC,$7F
	.dc.b	$45,$6D,$F9,$63,$F4,$95,$FA,$AF,$37,$F1,$7E,$EF,$5F,$07,$FA,$C6
	.dc.b	$48,$BA,$67,$0F,$F6,$84,$F9,$B7,$FD,$B8,$C7,$A4,$66,$7F,$D6,$B8
	.dc.b	$FE,$62,$FA,$FC,$3F,$49,$FE,$B5,$E6,$FF,$13,$F4,$BE,$13,$63,$27
	.dc.b	$FB,$3E,$32,$D3,$3A,$FF,$67,$9F,$E8,$FF,$1D,$A8,$FF,$6F,$E5,$FC
	.dc.b	$CE,$BF,$33,$D1,$FC,$81,$F5,$F9,$5F,$DD,$63,$15,$C2,$31,$3F,$11
	.dc.b	$AF,$02,$47,$F5,$DE,$8D,$47,$E8,$37,$7D,$03,$E8,$BF,$70,$FE,$CA
	.dc.b	$FE,$A1,$F8,$97,$55,$C5,$D5,$7C,$E7,$33,$C2,$3A,$E5,$D2,$FA,$B6
	.dc.b	$F9,$36,$B8,$5F,$49,$FA,$BF,$43,$FC,$65,$F0,$7A,$BF,$E1,$98,$C5
	.dc.b	$70,$8C,$78,$9E,$A6,$D7,$03,$F7,$FC,$6B,$E3,$9F,$A6,$B7,$87,$3B
	.dc.b	$DF,$CF,$FD,$0A,$FF,$09,$7D,$07,$F2,$4A,$F8,$B8,$AB,$57,$FC,$4A
	.dc.b	$B5,$7F,$E1,$C9,$CD,$FF,$0F,$9C,$D7,$E4,$A7,$3C,$09,$CD,$7F,$90
	.dc.b	$27,$31,$8C,$3C,$17,$FE,$39,$8C,$7C,$FF,$28,$A7,$F8,$F4,$A7,$F1
	.dc.b	$A2,$DF,$DE,$C3,$FB,$7F,$93,$FE,$0A,$80,$F7,$AF,$B2,$F8,$EC,$CB
	.dc.b	$E7,$FF,$84,$BF,$E3,$68,$FF,$58,$D7,$07,$73,$AA,$E5,$5F,$EB,$FA
	.dc.b	$D3,$E0,$9D,$57,$28,$E8,$67,$C3,$FE,$E7,$7C,$A3,$C7,$7C,$9F,$24
	.dc.b	$EA,$B9,$56,$EF,$AD,$78,$4E,$A7,$0C,$B9,$3F,$0B,$0E,$93,$C2,$3B
	.dc.b	$9C,$1C,$3B,$31,$87,$85,$ED,$7E,$18,$C4,$72,$8C,$5F,$82,$3E,$81
	.dc.b	$C9,$C5,$F8,$AB,$8C,$F7,$AE,$2F,$A8,$C6,$1F,$42,$77,$1D,$27,$67
	.dc.b	$C9,$76,$7C,$DB,$EE,$FF,$B9,$8F,$19,$FE,$E8,$BC,$38,$3F,$07,$F3
	.dc.b	$AF,$9E,$7C,$FF,$FB,$BA,$30,$D8,$CC,$FF,$BD,$A7,$2D,$F9,$5F,$EB
	.dc.b	$0D,$E3,$E5,$18,$8F,$C8,$89,$46,$63,$C6,$F2,$CB,$F0,$DF,$3E,$F9
	.dc.b	$7F,$CC,$C6,$1B,$B7,$0C,$7F,$15,$8F,$47,$F1,$97,$C7,$1B,$7C,$1F
	.dc.b	$D3,$D6,$DF,$C5,$B7,$0E,$D1,$C0,$EB,$8F,$97,$1F,$04,$7E,$4B,$0E
	.dc.b	$CD,$C3,$B3,$EF,$AE,$0F,$81,$AA,$78,$DF,$82,$3C,$73,$AC,$3C,$98
	.dc.b	$7E,$EA,$F6,$DC,$3B,$CE,$AF,$B6,$3E,$97,$C8,$7A,$98,$E0,$98,$FE
	.dc.b	$D6,$79,$47,$A6,$79,$4F,$94,$D6,$3E,$CE,$47,$CF,$3B,$BF,$87,$DF
	.dc.b	$FB,$1A,$31,$87,$F3,$C6,$BF,$D6,$3E,$05,$F5,$3F,$E6,$18,$D5,$F2
	.dc.b	$E2,$FF,$00,$DF,$0C,$3A,$4E,$69,$87,$2E,$CC,$65,$F8,$FF,$66,$ED
	.dc.b	$FE,$73,$BF,$E9,$AD,$7C,$5A,$7F,$9B,$DC,$CF,$02,$3F,$E5,$2C,$3C
	.dc.b	$FC,$F8,$8F,$E9,$B7,$CF,$3F,$F7,$2B,$75,$FE,$6B,$7C,$2B,$FA,$EF
	.dc.b	$90,$FB,$3F,$3B,$63,$31,$ED,$97,$A5,$FE,$A5,$9C,$90,$FE,$BB,$5C
	.dc.b	$5F,$F3,$1F,$97,$D2,$E4,$6F,$2F,$F5,$2E,$1D,$2B,$E3,$FC,$DF,$5A
	.dc.b	$FE,$CE,$DA,$FF,$7F,$B7,$2F,$5E,$4F,$9C,$E3,$84,$C3,$FB,$CD,$ED
	.dc.b	$FA,$B5,$B7,$33,$B7,$BB,$EF,$8D,$47,$F8,$70,$7D,$97,$F9,$C5,$FF
	.dc.b	$41,$98,$F5,$3F,$04,$D3,$3F,$2C,$FA,$3C,$D7,$19,$E8,$DD,$3F,$D8
	.dc.b	$F1,$C4,$F0,$57,$17,$BC,$EF,$4F,$ED,$73,$E6,$AE,$2D,$8F,$ED,$F3
	.dc.b	$BA,$FE,$EA,$BE,$0B,$C6,$48,$F3,$F0,$57,$CB,$17,$0E,$B3,$6F,$C3
	.dc.b	$EA,$7C,$0C,$4E,$70,$FD,$1F,$C0,$78,$5F,$93,$F1,$FF,$08,$FC,$64
	.dc.b	$67,$34,$66,$7D,$EE,$B5,$FD,$5F,$CC,$FA,$FA,$4D,$39,$4E,$BE,$6E
	.dc.b	$5F,$F8,$EA,$51,$E0,$3E,$9F,$A9,$27,$E6,$BC,$1F,$F1,$BF,$E8,$4F
	.dc.b	$73,$F7,$6D,$8F,$F2,$2B,$5F,$F1,$EF,$DB,$BE,$67,$F0,$7F,$95,$3D
	.dc.b	$9A,$4A,$FF,$65,$FC,$13,$ED,$9E,$D6,$67,$F2,$17,$C5,$FC,$7B,$5F
	.dc.b	$F3,$7D,$71,$9E,$1D,$1F,$8A,$AD,$37,$0F,$E2,$D5,$BB,$FE,$15,$F4
	.dc.b	$7E,$9A,$74,$7F,$B9,$3C,$AF,$EE,$95,$FE,$DA,$6B,$FC,$9E,$B8,$4F
	.dc.b	$F0,$C6,$C7,$FB,$96,$77,$5D,$23,$1E,$77,$F0,$BA,$DF,$A6,$B7,$E1
	.dc.b	$7F,$CA,$E3,$0D,$8C,$DF,$A3,$D8,$C4,$C8,$F3,$3D,$63,$F6,$7E,$D7
	.dc.b	$FB,$06,$7F,$6E,$D7,$F7,$8D,$CB,$CD,$A3,$ED,$FB,$D6,$BC,$0E,$64
	.dc.b	$7E,$CC,$6A,$BF,$73,$7C,$F6,$FF,$2F,$B6,$BF,$CC,$AD,$C3,$B6,$5F
	.dc.b	$0F,$E8,$FD,$DF,$6B,$2F,$EA,$C6,$5F,$A5,$19,$7E,$94,$7C,$5F,$BB
	.dc.b	$91,$F5,$DF,$9D,$F3,$58,$8F,$F3,$C6,$E9,$C6,$9F,$CC,$B8,$57,$FC
	.dc.b	$6B,$F2,$07,$DE,$D2,$63,$FC,$83,$D7,$FE,$98,$FF,$A0,$27,$6F,$F2
	.dc.b	$46,$BF,$CA,$5B,$87,$F1,$E9,$E3,$29,$CF,$07,$FA,$19,$B1,$D6,$FF
	.dc.b	$96,$D7,$83,$93,$F1,$1F,$41,$F8,$52,$57,$E7,$3E,$9A,$BF,$02,$2E
	.dc.b	$5E,$F9,$FD,$1F,$CB,$BE,$52,$FC,$7C,$C3,$F6,$B3,$96,$CE,$5F,$D8
	.dc.b	$C2,$79,$CF,$F4,$2F,$88,$FE,$5D,$5F,$D1,$9A,$FF,$4B,$6E,$1F,$D0
	.dc.b	$3E,$13,$F5,$33,$FE,$ED,$F3,$F9,$11,$FE,$9F,$1A,$AE,$3C,$1F,$EA
	.dc.b	$C6,$CF,$FA,$C9,$AA,$7F,$6B,$D0,$FB,$47,$CD,$7F,$04,$5F,$25,$4C
	.dc.b	$44,$BF,$E2,$F3,$A6,$CE,$5F,$E0,$98,$7E,$DF,$EB,$24,$7F,$0E,$8F
	.dc.b	$E0,$06,$1F,$C9,$C6,$1F,$E5,$7C,$3F,$CB,$88,$FF,$16,$E3,$AF,$F3
	.dc.b	$48,$BB,$7E,$66,$7F,$9A,$FE,$15,$2B,$F9,$AB,$FC,$24,$C3,$FC,$F1
	.dc.b	$B3,$FC,$57,$93,$F8,$94,$7F,$3E,$F9,$59,$5F,$CF,$9F,$E3,$1F,$1D
	.dc.b	$88,$FE,$8B,$16,$FF,$55,$3F,$E0,$2F,$A2,$FE,$B7,$E6,$FF,$B0,$FA
	.dc.b	$FF,$DC,$31,$8C,$3F,$DC,$D1,$8F,$47,$FA,$67,$C5,$7E,$CA,$7F,$DD
	.dc.b	$7F,$48,$DF,$FD,$AF,$7B,$C7,$A5,$7F,$BC,$31,$E9,$87,$18,$46,$2B
	.dc.b	$FD,$E3,$87,$39,$E2,$6B,$1F,$0C,$F5,$8F,$29,$D6,$32,$F9,$31,$F1
	.dc.b	$B9,$C7,$9E,$1D,$E3,$07,$21,$87,$85,$C4,$6D,$F1,$4F,$2B,$EF,$1C
	.dc.b	$2F,$BB,$A3,$8C,$6A,$B9,$5F,$5C,$7E,$14,$75,$24,$F4,$9D,$63,$EF
	.dc.b	$AD,$CF,$33,$58,$78,$2F,$8E,$3E,$53,$E4,$E1,$CB,$1F,$7E,$1C,$A7
	.dc.b	$51,$F0,$67,$4F,$A6,$74,$FA,$A7,$9D,$F7,$C3,$C5,$1C,$AB,$EE,$5F
	.dc.b	$2C,$3C,$78,$71,$3B,$61,$C6,$F8,$63,$ED,$3F,$03,$19,$9E,$18,$78
	.dc.b	$8E,$B1,$97,$13,$C6,$7F,$00,$74,$3D,$EE,$5F,$AC,$64,$F7,$39,$C3
	.dc.b	$D1,$8F,$F8,$9E,$7F,$1C,$7F,$BD,$31,$FD,$23,$63,$4F,$E7,$7D,$17
	.dc.b	$D1,$7F,$BE,$A3,$0D,$7F,$3F,$1A,$75,$48,$C4,$77,$75,$8F,$E1,$9B
	.dc.b	$8A,$67,$D1,$BE,$93,$FC,$35,$7B,$C7,$F3,$F8,$70,$BF,$19,$F9,$FB
	.dc.b	$E1,$D7,$AA,$93,$FA,$0A,$E0,$AB,$6E,$67,$77,$DE,$3A,$61,$E2,$C7
	.dc.b	$F8,$5E,$1C,$B2,$F7,$61,$C2,$7D,$57,$D6,$F8,$3E,$6C,$3A,$12,$77
	.dc.b	$8F,$8A,$31,$3B,$9E,$F7,$FC,$22,$B5,$18,$9F,$41,$2B,$F8,$94,$EA
	.dc.b	$39,$10,$FA,$F5,$CE,$FC,$B1,$83,$95,$F5,$BE,$73,$AC,$3D,$F5,$A3
	.dc.b	$A6,$1D,$DE,$55,$EA,$6D,$FB,$AF,$96,$3F,$24,$98,$F8,$F2,$F9,$57
	.dc.b	$E7,$CB,$BE,$3E,$29,$CC,$F4,$3C,$38,$74,$BC,$DF,$0C,$7C,$17,$8C
	.dc.b	$7B,$E1,$F7,$23,$17,$D3,$2F,$56,$1C,$69,$EB,$BF,$43,$7C,$6F,$F9
	.dc.b	$B7,$8F,$FC,$DD,$39,$8C,$65,$FE,$75,$F8,$DB,$87,$2C,$38,$CF,$E7
	.dc.b	$F0,$E3,$18,$94,$62,$F8,$FC,$2B,$C8,$63,$D3,$2F,$06,$5C,$CF,$D5
	.dc.b	$3E,$7F,$97,$D6,$67,$C7,$D4,$FC,$BB,$FD,$CD,$7C,$75,$F3,$57,$02
	.dc.b	$46,$6B,$87,$C1,$E2,$BC,$F3,$E0,$C1,$F3,$5E,$03,$D3,$A3,$D6,$C5
	.dc.b	$77,$D3,$E6,$56,$30,$FF,$4F,$55,$E0,$FF,$33,$7A,$5E,$8D,$9F,$0D
	.dc.b	$D2,$30,$DC,$BC,$D3,$A7,$FD,$5F,$3A,$7F,$D8,$3F,$6F,$FE,$C0,$6F
	.dc.b	$3F,$D6,$F2,$13,$9C,$FF,$D7,$EF,$F1,$BA,$E3,$ED,$EE,$31,$3A,$8C
	.dc.b	$5F,$E0,$21,$F8,$93,$EA,$B0,$DC,$3E,$80,$F9,$FF,$AC,$FD,$3A,$E3
	.dc.b	$5C,$09,$1C,$27,$73,$F4,$8F,$E5,$55,$D5,$C1,$CA,$7A,$46,$76,$FF
	.dc.b	$63,$47,$87,$9B,$AD,$7F,$B7,$BD,$5F,$F6,$2C,$FF,$BA,$17,$B1,$F1
	.dc.b	$DC,$CE,$E3,$1B,$E5,$B5,$C5,$FB,$5F,$19,$23,$FD,$C7,$16,$BD,$5D
	.dc.b	$3F,$95,$AC,$65,$FE,$A7,$AB,$7F,$CB,$8F,$C6,$1F,$1C,$B7,$F0,$DF
	.dc.b	$99,$EE,$7D,$37,$35,$FE,$E5,$9F,$F7,$8F,$2E,$4F,$DB,$CF,$03,$A3
	.dc.b	$96,$9F,$46,$B1,$3E,$8E,$4F,$DD,$5F,$F9,$75,$B5,$FE,$F7,$D3,$F3
	.dc.b	$5D,$B2,$3F,$2E,$FE,$9A,$31,$7F,$EF,$FF,$1B,$F4,$18,$FE,$A9,$B7
	.dc.b	$FB,$7E,$EF,$E0,$91,$8A,$E1,$3F,$A2,$AF,$C9,$7E,$B9,$F9,$6C,$FE
	.dc.b	$03,$67,$FB,$87,$43,$E4,$FF,$51,$87,$E2,$2B,$4F,$F7,$56,$A7,$33
	.dc.b	$C7,$AB,$BF,$44,$C3,$F3,$55,$AC,$3A,$E3,$E2,$AF,$DE,$72,$CB,$F6
	.dc.b	$76,$BC,$EC,$69,$F3,$6A,$E1,$FE,$E9,$9B,$7F,$E0,$9D,$FF,$93,$79
	.dc.b	$BF,$C1,$FE,$2F,$F8,$8B,$67,$AB,$71,$FD,$45,$70,$9F,$E1,$35,$FB
	.dc.b	$9F,$B4,$FF,$67,$47,$EC,$DB,$5F,$AF,$E0,$7D,$DB,$E3,$F8,$7B,$53
	.dc.b	$0E,$0F,$F0,$D9,$D4,$60,$97,$FB,$6C,$3A,$EA,$9F,$AB,$5C,$2B,$57
	.dc.b	$FD,$07,$F0,$EF,$E2,$D7,$FD,$49,$B7,$FD,$0B,$F0,$27,$46,$FC,$3F
	.dc.b	$E2,$9E,$93,$D1,$A7,$3E,$6B,$87,$F3,$5C,$7F,$03,$E4,$1F,$C1,$F1
	.dc.b	$FF,$53,$F4,$38,$56,$F0,$F2,$1D,$B0,$B1,$FC,$6F,$1F,$C4,$9C,$BA
	.dc.b	$1E,$A6,$E7,$F0,$2D,$E1,$FA,$B3,$A9,$E5,$7F,$CB,$7E,$77,$F8,$D4
	.dc.b	$7F,$92,$7D,$17,$D5,$3E,$57,$5E,$67,$F7,$29,$CC,$62,$BF,$97,$5F
	.dc.b	$2C,$BF,$CB,$FB,$3E,$5F,$51,$FC,$AA,$74,$9E,$18,$7F,$95,$E7,$7C
	.dc.b	$DF,$C3,$67,$FB,$1E,$1F,$DB,$61,$F4,$9C,$A7,$38,$7F,$39,$9C,$E7
	.dc.b	$ED,$F4,$26,$1F,$DD,$72,$F6,$F0,$63,$0F,$25,$5C,$3F,$11,$1F,$E7
	.dc.b	$8D,$BE,$43,$AC,$3F,$B8,$7B,$2F,$9B,$FF,$05,$36,$3F,$86,$7C,$B7
	.dc.b	$E2,$AB,$FC,$43,$39,$7F,$A7,$74,$23,$18,$FF,$88,$9B,$7F,$B3,$BE
	.dc.b	$55,$BC,$3B,$92,$7B,$F4,$DC,$7A,$FD,$17,$E5,$67,$FC,$0C,$D8,$FD
	.dc.b	$BE,$1F,$94,$8C,$65,$FE,$32,$6D,$7F,$88,$9B,$4E,$95,$F1,$E9,$FF
	.dc.b	$15,$56,$B3,$3C,$2B,$7C,$FA,$AF,$F1,$87,$BA,$F8,$1A,$7F,$8D,$BE
	.dc.b	$F1,$F0,$18,$D7,$FC,$3D,$5B,$6D,$6E,$75,$AF,$CC,$AD,$36,$75,$5C
	.dc.b	$0E,$7E,$27,$88,$F0,$55,$8C,$92,$BF,$8E,$E0,$9D,$47,$F8,$10,$FF
	.dc.b	$58,$7B,$FF,$EB,$1F,$8A,$FF,$2E,$C7,$FA,$CF,$A9,$F4,$3F,$8C,$9C
	.dc.b	$CF,$FA,$7A,$B8,$7A,$5F,$EA,$0B,$DE,$58,$AF,$F5,$18,$AC,$3F,$A9
	.dc.b	$C3,$A5,$FF,$B3,$67,$37,$CA,$98,$78,$1F,$39,$8A,$F0,$5F,$28,$EF
	.dc.b	$7C,$A3,$11,$E7,$38,$1E,$D7,$2F,$4A,$F6,$DF,$85,$F0,$47,$A4,$E2
	.dc.b	$7A,$70,$F5,$BC,$70,$F1,$C6,$23,$84,$F4,$BE,$F1,$8A,$DB,$D0,$E4
	.dc.b	$79,$6B,$96,$1E,$B6,$C7,$8B,$0F,$35,$7C,$98,$C5,$F6,$BE,$11,$D4
	.dc.b	$F0,$1D,$4C,$3F,$EF,$B7,$E6,$9F,$EF,$D7,$0F,$C1,$3E,$DF,$CE,$7F
	.dc.b	$9E,$FE,$87,$EE,$C7,$FB,$E6,$30,$E1,$AF,$0F,$59,$3C,$21,$3F,$E9
	.dc.b	$CE,$7C,$36,$39,$7C,$07,$FA,$91,$B3,$A8,$C6,$BF,$94,$FA,$1F,$CF
	.dc.b	$57,$E3,$9A,$FD,$8B,$E9,$3F,$9A,$6E,$9F,$EF,$8A,$DC,$F3,$DF,$FD
	.dc.b	$F6,$EF,$0E,$F5,$FE,$FE,$BF,$37,$C9,$6F,$10,$70,$D9,$E1,$5E,$B9
	.dc.b	$E9,$97,$9F,$1E,$F5,$E3,$6D,$FC,$DB,$F0,$B6,$B8,$30,$F7,$46,$9F
	.dc.b	$9A,$DC,$BD,$B7,$D6,$7F,$67,$8F,$C2,$7A,$1F,$9E,$9D,$BE,$C6,$BD
	.dc.b	$2F,$8D,$F5,$C3,$CF,$2A,$DD,$F5,$8E,$07,$C3,$3A,$4F,$C8,$70,$71
	.dc.b	$71,$8F,$C6,$8E,$1D,$2F,$A1,$FE,$B5,$1F,$92,$8C,$76,$FE,$35,$B1
	.dc.b	$99,$DB,$FC,$52,$32,$FF,$71,$F2,$3F,$C4,$8F,$F7,$76,$BF,$D8,$27
	.dc.b	$8E,$9F,$F0,$03,$27,$D0,$7A,$63,$0D,$E5,$FF,$79,$CE,$AF,$FC,$F1
	.dc.b	$F4,$3F,$E0,$D7,$FD,$C5,$19,$7C,$8D,$7E,$26,$5F,$A6,$BF,$64,$E6
	.dc.b	$7A,$C7,$75,$EF,$BD,$27,$CB,$66,$30,$EF,$0E,$95,$FE,$F3,$F7,$3F
	.dc.b	$C5,$0D,$F3,$B5,$18,$8F,$F3,$3F,$6F,$FB,$C9,$D1,$F4,$63,$4F,$A1
	.dc.b	$AF,$E9,$29,$18,$84,$62,$BC,$DE,$AF,$F4,$8D,$FF,$C3,$79,$23,$1B
	.dc.b	$F3,$C7,$C9,$5F,$E2,$0C,$1F,$43,$E5,$CB,$FC,$2E,$D9,$D3,$6F,$FD
	.dc.b	$11,$9B,$E0,$7E,$24,$F1,$BF,$A2,$D7,$D4,$D7,$D2,$DC,$3C,$3E,$F7
	.dc.b	$AB,$CE,$95,$FC,$13,$2F,$0C,$FD,$78,$7D,$03,$E8,$30,$D8,$E1,$A7
	.dc.b	$E9,$63,$11,$FE,$FC,$FA,$17,$C5,$70,$BF,$E7,$55,$AA,$FE,$E2,$D7
	.dc.b	$94,$68,$F1,$57,$C1,$8E,$E7,$F1,$A7,$0B,$1F,$F7,$45,$78,$9B,$F2
	.dc.b	$5E,$82,$4F,$21,$5F,$80,$71,$49,$F3,$92,$3F,$A2,$46,$5B,$1E,$6F
	.dc.b	$7A,$47,$28,$F8,$FE,$5C,$8E,$AF,$E8,$FC,$DF,$F1,$FE,$59,$9F,$EA
	.dc.b	$F8,$7E,$F1,$F8,$B5,$F2,$0F,$E1,$1F,$43,$F1,$E3,$F2,$AD,$E0,$34
	.dc.b	$9F,$43,$C0,$93,$FC,$4E,$11,$F5,$3C,$F9,$3E,$E6,$BF,$C4,$FE,$85
	.dc.b	$EE,$7F,$19,$6C,$7F,$1D,$6B,$B8,$FE,$F1,$F3,$DF,$BE,$8F,$A0,$D7
	.dc.b	$EC,$71,$FF,$2D,$8C,$71,$7F,$A1,$BA,$9F,$3E,$F5,$BF,$29,$5C,$70
	.dc.b	$FF,$1C,$B5,$E5,$DF,$E0,$23,$F8,$8B,$E7,$E0,$FF,$13,$6C,$79,$27
	.dc.b	$EA,$78,$BE,$AC,$BF,$7B,$E6,$3C,$EF,$E2,$CD,$BF,$C9,$F3,$AF,$86
	.dc.b	$DF,$EF,$F6,$CF,$F2,$0A,$E0,$FC,$03,$17,$E0,$DF,$AF,$C6,$49,$FC
	.dc.b	$FF,$87,$FE,$4E,$8F,$2F,$95,$99,$E5,$39,$77,$82,$B7,$1F,$9F,$F9
	.dc.b	$EF,$CF,$3E,$10,$E3,$5F,$F5,$AC,$63,$C5,$F8,$DF,$3F,$CE,$39,$47
	.dc.b	$F8,$CE,$B9,$C7,$F9,$A9,$AF,$F6,$76,$BF,$E1,$07,$B6,$DF,$E1,$2E
	.dc.b	$15,$BC,$FE,$E7,$F1,$1A,$DD,$FF,$A9,$78,$1F,$3F,$F4,$BE,$3E,$4F
	.dc.b	$D2,$9C,$B6,$72,$FC,$6D,$95,$F7,$A7,$ED,$FC,$2C,$AF,$92,$FE,$0B
	.dc.b	$E5,$B1,$1C,$62,$EF,$FE,$DA,$E5,$FC,$05,$7F,$27,$F9,$F7,$CF,$D2
	.dc.b	$61,$FC,$3F,$D4,$FE,$D5,$87,$CE,$AD,$CF,$F9,$3D,$AF,$F9,$61,$AF
	.dc.b	$FA,$59,$AF,$F8,$4D,$DC,$7D,$89,$FE,$F1,$3F,$75,$B1,$DE,$BF,$33
	.dc.b	$F2,$32,$BF,$0E,$FE,$83,$E7,$F1,$1F,$1E,$2D,$7B,$F7,$57,$EA,$7E
	.dc.b	$7D,$F3,$F7,$E3,$A6,$3F,$AF,$9C,$B6,$72,$FE,$87,$99,$2A,$E9,$5F
	.dc.b	$D4,$7E,$0B,$F9,$75,$FE,$5D,$B9,$7E,$59,$BD,$1F,$E3,$5F,$9F,$FD
	.dc.b	$7D,$7F,$B7,$BC,$DF,$F4,$45,$72,$8F,$F5,$FB,$5F,$F1,$E3,$5F,$F3
	.dc.b	$2B,$D1,$FF,$33,$CF,$E8,$5F,$D4,$36,$FC,$55,$FC,$2F,$E7,$FF,$DE
	.dc.b	$15,$FC,$3D,$7C,$FE,$4D,$7C,$2C,$9D,$46,$5E,$11,$FC,$4B,$1F,$E1
	.dc.b	$FF,$29,$27,$F3,$D1,$97,$F7,$F1,$97,$F8,$67,$3F,$F1,$39,$FF,$15
	.dc.b	$79,$4E,$4F,$E7,$04,$AF,$E7,$0D,$7F,$4D,$F2,$9F,$C7,$8E,$31,$C9
	.dc.b	$91,$FD,$CB,$CF,$52,$78,$4E,$63,$0F,$F4,$58,$FE,$90,$47,$FA,$47
	.dc.b	$C6,$FF,$4B,$23,$82,$1F,$B9,$F9,$1F,$E6,$27,$B2,$3F,$C2,$DF,$3A
	.dc.b	$F4,$FF,$A9,$FC,$3B,$BF,$FC,$F1,$39,$F2,$FF,$D4,$33,$9C,$7F,$D4
	.dc.b	$73,$95,$19,$5F,$3F,$F2,$5E,$37,$D3,$0E,$ED,$C3,$33,$D2,$72,$FF
	.dc.b	$AC,$2B,$93,$DA,$B8,$34,$E1,$5C,$96,$1E,$5C,$3C,$97,$C9,$F3,$46
	.dc.b	$6F,$9C,$63,$0F,$0D,$FA,$B6,$FF,$4D,$4E,$71,$ED,$87,$1A,$F0,$9E
	.dc.b	$33,$81,$AB,$E7,$1B,$8F,$2C,$76,$C7,$C9,$8F,$88,$93,$D3,$0F,$4E
	.dc.b	$3E,$29,$CC,$F2,$C7,$B1,$F0,$6F,$9C,$71,$9D,$E3,$E1,$9C,$CF,$13
	.dc.b	$D1,$1E,$53,$57,$DE,$39,$10,$E1,$87,$91,$F0,$9C,$B1,$F7,$9E,$D9
	.dc.b	$E5,$1C,$30,$F1,$CE,$63,$8D,$7C,$D7,$96,$3E,$BB,$E3,$87,$96,$FA
	.dc.b	$CE,$6B,$8E,$1D,$23,$11,$C8,$98,$75,$3D,$0E,$27,$83,$F6,$71,$F8
	.dc.b	$B8,$7E,$76,$3F,$DB,$47,$C7,$2B,$97,$F2,$66,$3D,$37,$B9,$C2,$3F
	.dc.b	$4F,$18,$71,$28,$C4,$71,$9E,$0F,$E6,$A3,$86,$1F,$98,$6E,$1F,$EE
	.dc.b	$DF,$9F,$6C,$EB,$FC,$AD,$6B,$0F,$0E,$38,$3F,$4D,$7C,$2B,$13,$F8
	.dc.b	$2A,$E0,$F8,$6B,$77,$D6,$35,$1C,$31,$F2,$3A,$BC,$63,$FC,$2B,$0E
	.dc.b	$79,$7B,$EB,$58,$72,$26,$1C,$9F,$D5,$65,$E5,$73,$87,$3A,$EB,$01
	.dc.b	$3B,$AE,$44,$AE,$13,$D3,$2F,$54,$E6,$3E,$29,$D8,$E9,$5C,$EB,$8C
	.dc.b	$EE,$7E,$21,$CF,$2F,$1B,$88,$D5,$71,$8E,$B1,$C7,$1F,$85,$1C,$4F
	.dc.b	$01,$C6,$FB,$9F,$77,$0F,$3B,$88,$E6,$D7,$E0,$C6,$9C,$90,$E8,$7B
	.dc.b	$1D,$38,$24,$F9,$2B,$F9,$D5,$FA,$67,$39,$7D,$A3,$F0,$B1,$9A,$E1
	.dc.b	$87,$8D,$B5,$DF,$1F,$AF,$C4,$F9,$F7,$C4,$FE,$2F,$B3,$F1,$9E,$83
	.dc.b	$E7,$1F,$19,$9C,$7E,$F6,$3C,$70,$C6,$4C,$38,$5E,$2B,$57,$FB,$16
	.dc.b	$D7,$EE,$9A,$FE,$24,$38,$7B,$E4,$F9,$AF,$F3,$EC,$61,$AF,$08,$D5
	.dc.b	$77,$6C,$77,$6E,$7F,$A9,$BF,$85,$7E,$AC,$7E,$07,$C0,$FA,$6F,$78
	.dc.b	$F8,$BE,$53,$5A,$AB,$51,$86,$CF,$86,$3C,$B8,$FC,$F6,$D9,$F4,$F6
	.dc.b	$FF,$AC,$7D,$5E,$BF,$07,$F3,$23,$18,$7F,$AA,$B2,$4E,$7C,$BF,$F5
	.dc.b	$34,$E6,$31,$7C,$A3,$FD,$91,$F3,$FF,$ED,$16,$F3,$7D,$8F,$2D,$39
	.dc.b	$E2,$FF,$2E,$3F,$C8,$EB,$8F,$BB,$B8,$C4,$EA,$31,$7F,$82,$87,$CF
	.dc.b	$3E,$5F,$F0,$B8,$3E,$7D,$F0,$DB,$AD,$4F,$F0,$9E,$7F,$E3,$8E,$E7
	.dc.b	$57,$F7,$B0,$EA,$DB,$E3,$1F,$CF,$5A,$FF,$46,$73,$7F,$EC,$46,$DF
	.dc.b	$F1,$73,$C9,$CF,$D9,$FF,$77,$7A,$BF,$32,$FF,$CF,$CF,$FA,$C2,$33
	.dc.b	$BF,$C9,$8C,$4E,$A3,$1E,$BE,$5B,$5F,$67,$E2,$7F,$12,$7D,$99,$B1
	.dc.b	$92,$57,$F9,$8B,$8F,$AE,$7F,$B4,$E4,$61,$F3,$EF,$FD,$F8,$7C,$93
	.dc.b	$D9,$9F,$FA,$E9,$C6,$1E,$FF,$7F,$E9,$CE,$63,$FD,$D5,$3C,$9F,$EC
	.dc.b	$CD,$7F,$B7,$B5,$FD,$B3,$5E,$8F,$67,$FB,$04,$FE,$03,$81,$EF,$79
	.dc.b	$E3,$F5,$3D,$D2,$FF,$DC,$1B,$DC,$1A,$7B,$EB,$0F,$FA,$E6,$6D,$7F
	.dc.b	$4F,$E8,$6D,$FE,$85,$8F,$D3,$65,$FA,$5F,$23,$F6,$ED,$8C,$9F,$EE
	.dc.b	$F8,$CB,$E9,$F9,$FF,$CE,$E3,$FD,$B7,$A5,$F3,$EF,$A5,$FC,$F5,$6A
	.dc.b	$FF,$B0,$56,$A7,$3D,$3F,$97,$AF,$D3,$EF,$FE,$18,$6B,$FC,$7D,$AF
	.dc.b	$F6,$97,$A3,$63,$F6,$D8,$7F,$00,$C3,$CD,$1F,$C4,$B8,$A4,$7D,$ED
	.dc.b	$BE,$1F,$A1,$8C,$3E,$5C,$FF,$75,$F0,$BF,$9E,$5E,$75,$FF,$72,$5D
	.dc.b	$AF,$99,$C1,$FE,$A3,$7F,$88,$75,$FF,$0F,$F3,$BF,$90,$B6,$32,$78
	.dc.b	$67,$86,$5E,$03,$FA,$27,$47,$F0,$09,$FC,$37,$C7,$FF,$B9,$DF,$CA
	.dc.b	$F1,$3E,$81,$F2,$9F,$C2,$EB,$58,$7F,$BB,$E7,$37,$CB,$DA,$FE,$09
	.dc.b	$5A,$AF,$F2,$2B,$5F,$E9,$8D,$7F,$C2,$CF,$4D,$BF,$C5,$50,$8C,$57
	.dc.b	$DC,$B7,$C8,$78,$F0,$FE,$B9,$F3,$FF,$C9,$F0,$FE,$B1,$E4,$23,$11
	.dc.b	$FD,$99,$AF,$F5,$C6,$AF,$4B,$FD,$81,$8F,$F9,$39,$B1,$FE,$57,$6B
	.dc.b	$FE,$60,$6B,$D7,$9B,$FA,$6D,$7E,$2A,$FC,$51,$83,$3E,$9C,$C3,$FB
	.dc.b	$95,$72,$BF,$24,$FF,$2D,$C3,$F2,$71,$E5,$BC,$12,$FC,$CF,$F3,$F9
	.dc.b	$FC,$BF,$3F,$EC,$BD,$4C,$4F,$F4,$3F,$98,$FE,$5F,$87,$F9,$CB,$E7
	.dc.b	$5F,$29,$9F,$27,$FC,$13,$EB,$27,$3A,$7F,$0B,$C3,$F9,$DE,$1F,$E0
	.dc.b	$56,$BF,$E4,$F6,$A6,$9E,$07,$17,$DB,$D6,$7C,$FF,$86,$FB,$7A,$1F
	.dc.b	$E1,$AE,$77,$CF,$F1,$CB,$FB,$B5,$7F,$2A,$9C,$E3,$ED,$BF,$EA,$95
	.dc.b	$AC,$BF,$D1,$CD,$7F,$D2,$ED,$7D,$51,$C6,$FF,$9A,$E9,$F0,$77,$C5
	.dc.b	$7E,$A7,$91,$F3,$FF,$A0,$8F,$D6,$78,$5F,$E2,$FC,$3F,$C5,$D3,$9D
	.dc.b	$3F,$C7,$33,$FE,$91,$9F,$E8,$37,$CB,$AF,$FC,$27,$7C,$F0,$FC,$CB
	.dc.b	$5F,$F1,$EB,$5F,$F1,$9B,$D6,$B3,$3F,$E1,$5F,$9F,$FF,$04,$63,$FE
	.dc.b	$2B,$AD,$5F,$F8,$EB,$E7,$FF,$37,$3E,$4F,$91,$FF,$22,$46,$31,$FE
	.dc.b	$29,$9A,$73,$D3,$FE,$28,$AF,$CB,$68,$69,$23,$77,$FC,$87,$3F,$E5
	.dc.b	$35,$B9,$FF,$29,$7C,$4B,$E0,$7F,$C7,$D7,$FC,$F0,$F8,$B1,$97,$FB
	.dc.b	$3E,$5F,$EB,$9F,$23,$8D,$7F,$8E,$AB,$5A,$FF,$96,$5A,$FF,$02,$6B
	.dc.b	$FE,$66,$77,$AF,$F9,$3D,$B8,$7F,$59,$F9,$F7,$C1,$76,$8F,$1F,$83
	.dc.b	$F9,$3F,$8E,$FD,$1E,$3F,$83,$C3,$F8,$9D,$7E,$4E,$73,$8F,$5B,$F3
	.dc.b	$12,$7F,$CE,$15,$C2,$FF,$DB,$63,$E0,$7F,$D7,$1F,$23,$FE,$82,$8F
	.dc.b	$D6,$7B,$FF,$EC,$D8,$C4,$7F,$9D,$67,$7E,$A7,$EC,$EB,$4F,$FB,$B6
	.dc.b	$3F,$D8,$5F,$1F,$C3,$B3,$FD,$13,$96,$2B,$58,$2B,$58,$7F,$B6,$DE
	.dc.b	$F1,$FE,$CB,$C3,$83,$E4,$BA,$F8,$CE,$57,$E3,$8C,$C7,$13,$27,$12
	.dc.b	$9C,$6B,$C3,$1C,$23,$E2,$1A,$C3,$D2,$F0,$3A,$4F,$07,$07,$07,$07
	.dc.b	$A9,$C5,$79,$5D,$DF,$CA,$7C,$98,$EE,$31,$1F,$C6,$F1,$F9,$AF,$FA
	.dc.b	$94,$E6,$7A,$EF,$83,$E7,$8F,$95,$7F,$3A,$7C,$6D,$3C,$4F,$09,$E8
	.dc.b	$7C,$72,$9F,$08,$B5,$8F,$9B,$7A,$EF,$9F,$C3,$6B,$9C,$EA,$72,$71
	.dc.b	$F9,$2F,$F1,$EC,$E9,$AF,$B6,$51,$88,$FF,$62,$46,$38,$39,$1F,$82
	.dc.b	$F9,$F7,$07,$FB,$27,$3F,$CB,$C6,$2F,$F0,$F7,$C5,$FC,$0B,$67,$75
	.dc.b	$C7,$4F,$F6,$95,$F1,$BD,$E3,$E8,$BF,$D1,$E1,$C6,$F7,$87,$1B,$0C
	.dc.b	$39,$5F,$0C,$39,$5F,$B1,$B5,$C2,$B7,$3D,$23,$D3,$3A,$58,$78,$5E
	.dc.b	$18,$78,$5F,$03,$63,$51,$C2,$79,$5F,$68,$E2,$76,$C6,$3F,$A6,$8D
	.dc.b	$3D,$0C,$E3,$EB,$9F,$A5,$87,$86,$B7,$87,$80,$F0,$BC,$8C,$E1,$F0
	.dc.b	$3A,$5F,$3F,$FE,$40,$CB,$FB,$77,$A4,$EB,$F9,$71,$FE,$4B,$CF,$FC
	.dc.b	$FF,$18,$7F,$DF,$55,$C6,$7F,$B4,$36,$1C,$9F,$EC,$DB,$E9,$5D,$AF
	.dc.b	$94,$FD,$8C,$3C,$7E,$06,$29,$1F,$43,$BF,$1D,$BF,$E4,$7F,$96,$BE
	.dc.b	$89,$E1,$5B,$BF,$EE,$3C,$1F,$D7,$5F,$E9,$B5,$B6,$D7,$09,$D4,$7F
	.dc.b	$80,$5B,$AF,$AA,$18,$78,$5E,$27,$6C,$7C,$3E,$DF,$F8,$BA,$3F,$AD
	.dc.b	$FA,$7F,$C2,$6F,$95,$A3,$1A,$FC,$1B,$E1,$7F,$E1,$16,$DF,$0A,$DC
	.dc.b	$F8,$9B,$A7,$FA,$FB,$CD,$F0,$9F,$1A,$37,$8F,$7A,$C9,$FE,$14,$F9
	.dc.b	$E7,$C2,$EF,$C9,$93,$98,$F9,$73,$A8,$F5,$CE,$9F,$F1,$34,$65,$B3
	.dc.b	$9C,$BD,$15,$CA,$3E,$E4,$BE,$71,$F3,$B8,$F9,$EF,$F7,$6C,$FF,$B6
	.dc.b	$DA,$FF,$A6,$CE,$11,$CA,$78,$C6,$FE,$0F,$91,$99,$91,$92,$1C,$E7
	.dc.b	$8B,$7E,$4B,$A1,$1F,$17,$05,$7C,$58,$3D,$CF,$41,$9D,$3F,$D3,$78
	.dc.b	$30,$CD,$FA,$EB,$5C,$5F,$AC,$AF,$D1,$CF,$8F,$A3,$D8,$7E,$87,$D7
	.dc.b	$79,$BE,$6C,$BA,$CF,$E7,$BE,$7B,$18,$FE,$A5,$B8,$7F,$22,$C7,$AD
	.dc.b	$FE,$67,$99,$5A,$E4,$7B,$AF,$4F,$F5,$AD,$C7,$FC,$B3,$5D,$2F,$8F
	.dc.b	$CF,$7D,$1F,$6A,$56,$A2,$BF,$E6,$9F,$A2,$7D,$17,$ED,$1B,$1F,$CD
	.dc.b	$B4,$F8,$F7,$FB,$8F,$94,$FF,$5A,$4F,$F9,$0B,$3F,$F8,$00,$DF,$86
	.dc.b	$D7,$BF,$F8,$AF,$9F,$FD,$D4,$FF,$B7,$FB,$DF,$41,$FB,$8E,$4F,$E8
	.dc.b	$71,$8F,$0B,$87,$89,$2F,$FD,$C3,$7B,$7F,$68,$D5,$87,$6D,$BF,$DA
	.dc.b	$ED,$C7,$AD,$71,$C7,$B5,$FB,$67,$F9,$4D,$6E,$1D,$2F,$1F,$DF,$1F
	.dc.b	$E8,$CF,$33,$FB,$EB,$FE,$62,$6B,$FE,$FB,$9E,$1A,$FE,$16,$B8,$E7
	.dc.b	$FC,$32,$F6,$DF,$3F,$55,$FC,$A2,$5F,$1B,$49,$8F,$9C,$F2,$E0,$F4
	.dc.b	$F3,$1E,$59,$B8,$FF,$23,$CB,$D5,$5F,$BC,$F5,$9E,$1F,$F3,$9D,$DF
	.dc.b	$41,$FE,$0C,$CB,$B4,$63,$CF,$E1,$52,$73,$5B,$7E,$E3,$72,$FE,$65
	.dc.b	$0C,$3B,$BE,$93,$15,$E1,$AE,$1C,$3F,$82,$6C,$FF,$81,$AB,$8B,$F3
	.dc.b	$75,$7D,$07,$C7,$F8,$F9,$3F,$17,$6F,$B1,$3F,$44,$CC,$7D,$99,$C3
	.dc.b	$F6,$7E,$3B,$E8,$11,$E5,$6F,$96,$C4,$7E,$06,$6E,$BF,$36,$B5,$3E
	.dc.b	$DF,$A0,$7D,$07,$F2,$AF,$5F,$FD,$D7,$3F,$05,$AF,$F2,$79,$DB,$C6
	.dc.b	$78,$7C,$3C,$8C,$10,$FE,$5D,$5C,$B0,$F9,$ED,$E0,$FF,$01,$E1,$D2
	.dc.b	$BF,$31,$F2,$3F,$8D,$23,$EE,$FA,$0A,$E8,$FC,$E4,$7E,$66,$3F,$0D
	.dc.b	$F4,$0F,$A0,$BF,$1F,$2F,$F0,$F3,$96,$CE,$5F,$E0,$10,$9E,$73,$FD
	.dc.b	$3B,$E1,$3F,$99,$57,$E1,$7E,$83,$F4,$2D,$EC,$D7,$A7,$23,$04,$3F
	.dc.b	$A8,$57,$26,$C7,$F5,$AA,$E7,$1F,$D5,$5B,$BF,$E5,$AB,$83,$FE,$09
	.dc.b	$EC,$7D,$07,$FB,$8E,$FF,$46,$BE,$83,$F5,$0D,$8C,$FC,$4E,$0C,$CC
	.dc.b	$8C,$9F,$8C,$C3,$F7,$DF,$2B,$23,$04,$3F,$87,$39,$7F,$41,$1F,$CC
	.dc.b	$78,$7F,$83,$7C,$AC,$9F,$E0,$9E,$8B,$CD,$56,$E7,$F8,$B5,$6D,$6B
	.dc.b	$FD,$3E,$B7,$1C,$94,$FF,$8E,$BC,$BF,$E6,$1E,$A1,$FE,$65,$8F,$79
	.dc.b	$FE,$88,$6E,$25,$F3,$8F,$E9,$1F,$1F,$FC,$77,$E4,$A4,$F7,$8C,$BF
	.dc.b	$CA,$E3,$2F,$9B,$AB,$FC,$A1,$92,$BF,$C2,$BF,$3D,$FE,$61,$F2,$FF
	.dc.b	$DC,$5E,$17,$F9,$8A,$31,$87,$FB,$E6,$31,$EC,$FF,$58,$F8,$AF,$DB
	.dc.b	$CB,$E8,$3E,$0B,$B3,$FD,$2B,$7C,$A7,$6D,$BE,$78,$78,$2B,$FD,$37
	.dc.b	$7C,$AF,$78,$77,$BD,$D7,$51,$87,$37,$8F,$17,$F9,$EA,$F8,$E3,$D6
	.dc.b	$77,$87,$37,$CA,$D7,$D5,$5C,$70,$EF,$7C,$A7,$C1,$87,$77,$67,$57
	.dc.b	$04,$BE,$98,$FA,$EF,$99,$2B,$C3,$8F,$96,$75,$18,$3B,$D7,$96,$FC
	.dc.b	$13,$A9,$F5,$D7,$86,$B4,$78,$31,$F5,$4F,$3C,$7D,$44,$C7,$FD,$3F
	.dc.b	$87,$3C,$7D,$18,$F3,$3C,$98,$72,$C3,$B9,$E4,$8D,$3F,$4E,$B9,$63
	.dc.b	$E2,$9C,$E3,$F8,$69,$CC,$63,$1F,$85,$7F,$28,$E8,$4A,$F3,$3F,$70
	.dc.b	$E0,$7A,$E3,$8D,$74,$8E,$11,$8C,$3D,$18,$FB,$27,$27,$33,$B9,$FE
	.dc.b	$C0,$3D,$A5,$7F,$42,$DC,$7F,$C3,$2F,$E9,$7D,$57,$C1,$71,$8F,$CF
	.dc.b	$C6,$1C,$52,$31,$1D,$5A,$F3,$9F,$C5,$47,$1C,$BF,$5B,$F4,$0E,$0F
	.dc.b	$F5,$A4,$7E,$AA,$75,$9F,$1C,$3F,$2B,$18,$42,$F9,$70,$76,$C6,$4F
	.dc.b	$07,$F4,$D3,$A8,$C1,$30,$F0,$4E,$B1,$F1,$E3,$E1,$9C,$D6,$E3,$FD
	.dc.b	$D7,$7C,$E7,$24,$8E,$06,$AB,$9D,$7C,$AB,$F2,$D6,$E3,$D7,$7C,$EF
	.dc.b	$8C,$6C,$F0,$1A,$BE,$6F,$F0,$6C,$7E,$23,$70,$EB,$5E,$1C,$6C,$F3
	.dc.b	$C7,$C0,$FC,$9C,$39,$E5,$F6,$F0,$DC,$F3,$3F,$8A,$65,$E5,$79,$63
	.dc.b	$E6,$75,$87,$DC,$C3,$9E,$5E,$4C,$7A,$57,$A8,$F0,$65,$DA,$33,$7E
	.dc.b	$4C,$7B,$57,$4C,$7B,$56,$9F,$64,$73,$8C,$3E,$43,$E3,$5F,$5C,$3C
	.dc.b	$47,$C6,$BE,$B5,$C2,$BB,$1C,$4F,$29,$F4,$38,$DF,$3E,$F8,$C6,$3F
	.dc.b	$E9,$DF,$8E,$B7,$FD,$4C,$5F,$0A,$CF,$46,$BA,$3F,$D6,$F7,$8C,$3F
	.dc.b	$37,$7B,$7A,$1F,$DE,$31,$F0,$38,$C7,$B4,$F9,$3E,$6B,$94,$30,$FE
	.dc.b	$E5,$E0,$E2,$B5,$7F,$E8,$59,$CB,$D6,$F8,$C6,$2F,$AE,$09,$CF,$57
	.dc.b	$FB,$8E,$FC,$9F,$01,$C3,$DC,$7D,$07,$7C,$3F,$DA,$55,$FE,$C1,$8C
	.dc.b	$56,$DB,$19,$3F,$CD,$7E,$2F,$3F,$88,$F5,$FA,$7F,$EA,$D8,$C4,$E8
	.dc.b	$FF,$66,$4E,$BD,$0F,$F5,$1C,$60,$FF,$4A,$FD,$03,$E8,$3E,$2F,$92
	.dc.b	$8F,$E3,$13,$A9,$CC,$63,$83,$FD,$A4,$F1,$9E,$7D,$3A,$AD,$BF,$42
	.dc.b	$9F,$40,$F9,$AF,$A7,$6F,$A0,$7C,$56,$EB,$53,$FE,$B7,$35,$19,$AE
	.dc.b	$35,$C2,$7F,$D4,$95,$C2,$FF,$1D,$1C,$0C,$D7,$0B,$E5,$1D,$6F,$95
	.dc.b	$7F,$BB,$7D,$7F,$8E,$DF,$63,$E2,$C6,$3B,$3F,$CB,$CA,$33,$C3,$A6
	.dc.b	$E1,$FE,$C2,$9D,$3F,$EF,$FF,$8D,$FC,$21,$F5,$66,$C6,$49,$87,$FA
	.dc.b	$DE,$32,$FE,$A4,$FD,$6F,$C0,$FC,$AE,$AF,$F5,$3D,$7F,$B1,$44,$67
	.dc.b	$DA,$FA,$51,$83,$E6,$4F,$0A,$E9,$19,$7E,$4C,$67,$3F,$F6,$9C,$E5
	.dc.b	$FD,$DB,$FA,$1A,$EB,$7D,$E3,$F5,$5D,$5F,$B4,$C3,$F8,$07,$62,$31
	.dc.b	$87,$FB,$07,$4F,$D3,$78,$13,$0E,$2F,$E6,$63,$18,$7F,$B6,$7C,$8F
	.dc.b	$D6,$7B,$DF,$9D,$6D,$FF,$B4,$1B,$3B,$8F,$1C,$FE,$4F,$E8,$3F,$35
	.dc.b	$8F,$FA,$5F,$E6,$9F,$0F,$9E,$8F,$D1,$C6,$30,$FF,$78,$5F,$07,$FB
	.dc.b	$6F,$5A,$75,$97,$D5,$AD,$29,$D6,$1D,$6F,$FB,$AE,$3E,$18,$C5,$7F
	.dc.b	$25,$AD,$D7,$CA,$E2,$FE,$14,$7E,$EF,$D3,$B1,$92,$57,$FB,$F6,$32
	.dc.b	$FF,$BF,$FD,$2F,$E2,$5D,$3F,$C0,$DB,$5F,$EF,$66,$CE,$E3,$EC,$CF
	.dc.b	$EE,$7E,$4F,$FD,$A1,$1E,$83,$E0,$FD,$03,$E5,$75,$EE,$7F,$0A,$8C
	.dc.b	$4E,$F0,$FF,$09,$DF,$27,$DD,$3A,$7F,$39,$19,$DB,$FD,$95,$39,$7F
	.dc.b	$A4,$CA,$75,$5E,$2C,$3C,$2F,$7D,$58,$71,$8F,$E9,$3F,$41,$FC,$5F
	.dc.b	$0F,$C9,$FE,$60,$F6,$7F,$A1,$4E,$73,$FE,$A5,$39,$BF,$55,$BD,$77
	.dc.b	$AD,$FD,$27,$E8,$DD,$5F,$C7,$E7,$37,$FE,$CC,$9C,$BF,$EC,$EC,$BF
	.dc.b	$0B,$87,$3C,$3C,$3D,$1C,$5F,$19,$FD,$43,$27,$C1,$E3,$0F,$EA,$D1
	.dc.b	$87,$CB,$97,$E4,$3D,$06,$5E,$2C,$BF,$84,$4E,$63,$1A,$FC,$C8,$F1
	.dc.b	$9F,$98,$F3,$1E,$5F,$FB,$1F,$1E,$BA,$FF,$3B,$F9,$CF,$E5,$D8,$7F
	.dc.b	$7C,$F6,$5F,$44,$F9,$9D,$79,$9F,$DC,$27,$31,$8C,$7F,$09,$18,$7E
	.dc.b	$6E,$1D,$1F,$F2,$CC,$67,$6F,$E4,$13,$97,$FC,$BA,$D8,$FF,$33,$7C
	.dc.b	$3F,$59,$DF,$A7,$FE,$00,$CD,$F4,$5C,$65,$87,$F7,$27,$FC,$29,$3A
	.dc.b	$CF,$DD,$7C,$6F,$85,$EE,$BF,$C2,$F1,$89,$D5,$70,$CB,$F4,$15,$CD
	.dc.b	$FD,$17,$13,$E8,$BF,$29,$97,$F8,$4F,$CA,$FF,$13,$D6,$A7,$31,$FE
	.dc.b	$17,$8F,$F0,$B4,$62,$37,$3F,$E1,$E8,$C4,$EB,$B7,$FC,$2F,$5A,$C3
	.dc.b	$FC,$63,$87,$86,$73,$5F,$E2,$1F,$A2,$FC,$1B,$FE,$88,$BF,$F1,$37
	.dc.b	$B1,$9E,$AE,$6D,$9F,$F1,$34,$A3,$19,$FF,$3E,$8C,$4F,$C0,$BE,$B3
	.dc.b	$EF,$AD,$E5,$FE,$3F,$BD,$CF,$F8,$E3,$A9,$F0,$9F,$E3,$0C,$7F,$97
	.dc.b	$1F,$2A,$32,$FF,$8D,$CC,$CF,$19,$E1,$E4,$FF,$8F,$6B,$73,$FD,$4A
	.dc.b	$B7,$87,$C4,$9C,$BF,$E5,$18,$C9,$FE,$50,$8C,$BF,$14,$CC,$FF,$96
	.dc.b	$A1,$F4,$3F,$8F,$F8,$DF,$D1,$57,$9E,$31,$CF,$E3,$6C,$64,$FC,$3C
	.dc.b	$F0,$BF,$EF,$C7,$FA,$EB,$E0,$7F,$D7,$7F,$23,$FE,$6D,$8F,$D0,$FB
	.dc.b	$FF,$E6,$C8,$C4,$7F,$B0,$04,$F0,$F4,$71,$D2,$BE,$11,$FE,$D9,$9D
	.dc.b	$57,$4C,$6B,$CE,$2D,$F1,$9C,$CF,$94,$61,$E2,$9D,$46,$2F,$B3,$4E
	.dc.b	$C5,$3B,$16,$38,$CF,$B2,$3B,$3D,$70,$F4,$C7,$FB,$E7,$0F,$2C,$62
	.dc.b	$39,$CE,$F0,$F3,$C7,$C4,$3B,$4E,$DE,$B8,$7C,$68,$C4,$6E,$B9,$3E
	.dc.b	$73,$67,$A9,$F3,$3B,$9E,$55,$E0,$BF,$2B,$67,$C2,$DA,$EA,$72,$9F
	.dc.b	$B5,$3F,$20,$A7,$CF,$29,$EB,$37,$F3,$DF,$E4,$0F,$A1,$CB,$63,$8B
	.dc.b	$5D,$FA,$A9,$DF,$C9,$38,$69,$2B,$8C,$65,$B1,$F8,$6F,$A1,$FE,$AB
	.dc.b	$17,$8B,$F2,$59,$FE,$53,$26,$1C,$63,$0D,$D7,$F3,$18,$72,$C3,$F0
	.dc.b	$8D,$AD,$E3,$F9,$4C,$39,$BD,$EA,$1F,$8B,$8C,$B7,$0E,$C8,$5F,$5C
	.dc.b	$3C,$38,$FA,$1F,$D5,$5F,$2C,$7C,$37,$E4,$AD,$3C,$EF,$93,$D6,$B8
	.dc.b	$5F,$2C,$3C,$33,$E0,$BE,$2D,$3D,$91,$AA,$E3,$1D,$67,$77,$DA,$13
	.dc.b	$A6,$D7,$69,$D1,$B9,$D1,$B7,$35,$E2,$8E,$FD,$5D,$39,$FF,$85,$67
	.dc.b	$FD,$22,$7F,$9F,$63,$FD,$69,$AF,$FB,$EF,$C4,$7D,$1B,$C1,$E3,$C5
	.dc.b	$FE,$F8,$BD,$E1,$CF,$1C,$E2,$CB,$C0,$FF,$91,$F2,$E7,$8E,$A5,$F7
	.dc.b	$3F,$E7,$6A,$FE,$D6,$D7,$FD,$FE,$DC,$19,$7D,$FC,$1F,$48,$F7,$BA
	.dc.b	$5E,$EB,$F0,$98,$F6,$F5,$F2,$79,$B0,$E3,$78,$CB,$C3,$35,$F8,$58
	.dc.b	$F5,$7F,$BF,$FD,$1B,$E1,$3D,$4D,$8C,$C7,$5E,$C7,$C0,$6A,$24,$E8
	.dc.b	$F8,$93,$A7,$07,$ED,$F4,$F4,$DF,$58,$C1,$F3,$5C,$6D,$F1,$21,$A7
	.dc.b	$F5,$8C,$7E,$77,$B4,$F8,$AA,$4A,$FA,$9F,$42,$FA,$EC,$47,$6E,$15
	.dc.b	$F1,$8F,$F0,$FB,$5F,$C3,$57,$17,$37,$CB,$2F,$F0,$AD,$6D,$F3,$64
	.dc.b	$BE,$17,$FE,$E1,$FA,$27,$C5,$FE,$2A,$11,$87,$FC,$EE,$D7,$FD,$D1
	.dc.b	$F4,$6F,$82,$E1,$39,$C7,$FD,$1A,$DE,$0F,$ED,$4D,$BF,$E1,$17,$C5
	.dc.b	$EB,$87,$22,$3F,$E8,$68,$D3,$E7,$F1,$FB,$91,$FF,$4B,$7E,$14,$FA
	.dc.b	$3C,$46,$CF,$4C,$FE,$9B,$DE,$56,$E7,$27,$F9,$AB,$E0,$BF,$D4,$BA
	.dc.b	$7F,$9C,$DB,$A7,$AB,$0F,$D5,$7D,$27,$1B,$F9,$5E,$C7,$C2,$33,$9D
	.dc.b	$8F,$93,$15,$FB,$D3,$C7,$1F,$F2,$5F,$D1,$BE,$93,$F6,$CD,$AF,$F2
	.dc.b	$FB,$5E,$E6,$3D,$6D,$91,$FE,$01,$ED,$BF,$0C,$FA,$BF,$1F,$27,$ED
	.dc.b	$BB,$7F,$21,$7F,$E4,$67,$BF,$A9,$E2,$3E,$47,$D2,$35,$FC,$37,$D0
	.dc.b	$3E,$3A,$92,$FF,$D7,$BE,$67,$F1,$6E,$3F,$F1,$55,$6E,$7F,$D8,$2D
	.dc.b	$7F,$5A,$1C,$70,$7F,$B5,$AB,$94,$FF,$4F,$9C,$D7,$4B,$F0,$4F,$2A
	.dc.b	$DF,$67,$FA,$C3,$D7,$FE,$4B,$EC,$3E,$0F,$FD,$11,$7F,$EC,$0F,$77
	.dc.b	$F8,$A7,$1F,$FA,$BE,$3E,$4D,$F0,$AD,$C7,$FB,$7A,$B7,$3D,$3D,$47
	.dc.b	$C6,$D2,$57,$D2,$FA,$0A,$FF,$0A,$CB,$FC,$7B,$38,$E0,$FD,$AE,$6B
	.dc.b	$FE,$77,$ED,$BE,$BD,$F3,$78,$6E,$5F,$DC,$70,$FE,$25,$5B,$AF,$CF
	.dc.b	$7D,$0B,$EF,$78,$44,$BE,$1C,$9F,$EC,$0C,$79,$C7,$F1,$EF,$0D,$87
	.dc.b	$18,$D9,$FE,$BE,$9D,$E5,$E0,$6D,$F3,$C3,$C3,$8F,$8B,$E8,$BC,$7F
	.dc.b	$21,$31,$E7,$9F,$69,$D4,$63,$35,$6E,$3C,$90,$9F,$A9,$EE,$49,$FF
	.dc.b	$7C,$BF,$6B,$A3,$11,$F3,$F9,$BF,$C4,$EE,$8F,$D2,$C7,$F1,$FF,$8E
	.dc.b	$FE,$29,$06,$98,$FA,$27,$C6,$6F,$CB,$95,$F1,$0E,$0F,$F2,$99,$DC
	.dc.b	$66,$B9,$CF,$08,$FC,$FC,$E9,$F3,$7C,$17,$E3,$C8,$FF,$79,$FB,$CF
	.dc.b	$F2,$CF,$27,$F4,$FE,$9F,$CD,$56,$AB,$89,$B3,$F8,$F7,$D1,$FE,$ED
	.dc.b	$FB,$8B,$E8,$F2,$6B,$C4,$93,$F5,$27,$4D,$9C,$BF,$B6,$D3,$F9,$1B
	.dc.b	$FE,$41,$6B,$F2,$79,$FF,$C2,$7F,$37,$F2,$3C,$CF,$9D,$12,$7E,$23
	.dc.b	$AE,$0F,$E4,$4F,$FA,$BE,$F9,$91,$D5,$7E,$DA,$3F,$61,$F2,$F2,$FF
	.dc.b	$8F,$C6,$5F,$CB,$EB,$FE,$AA,$9D,$47,$F8,$B7,$BB,$EF,$3F,$37,$E4
	.dc.b	$3F,$7D,$F0,$12,$30,$43,$F1,$4E,$5F,$F1,$B7,$12,$7F,$C5,$9F,$73
	.dc.b	$FC,$96,$7F,$4F,$ED,$23,$11,$FD,$F5,$BC,$9F,$DD,$2F,$FD,$95,$3C
	.dc.b	$27,$3B,$7F,$8A,$9B,$1F,$E3,$D6,$F8,$EA,$E3,$3D,$7D,$BF,$19,$1F
	.dc.b	$87,$D4,$F9,$BC,$46,$CF,$04,$FF,$1F,$8F,$5F,$DD,$C9,$F2,$3F,$E3
	.dc.b	$9B,$7C,$FE,$67,$D9,$37,$4F,$64,$FF,$82,$FE,$85,$F2,$5F,$A4,$F8
	.dc.b	$FF,$F0,$D4,$63,$A3,$FD,$E9,$F0,$3F,$EE,$F8,$C6,$5F,$EF,$08,$FF
	.dc.b	$79,$7D,$27,$F2,$E8,$D9,$FE,$91,$9D,$DF,$4C,$3B,$5F,$36,$D7,$FA
	.dc.b	$5E,$F9,$3D,$4E,$51,$A5,$7D,$2B,$96,$1E,$09,$5F,$2C,$3C,$13,$9B
	.dc.b	$E1,$5D,$23,$85,$6E,$F9,$CF,$84,$EF,$87,$86,$F8,$C7,$6F,$37,$FD
	.dc.b	$5C,$EE,$32,$49,$F4,$5F,$59,$F0,$1B,$8E,$86,$AB,$9E,$3E,$89,$D6
	.dc.b	$3E,$9A,$D4,$58,$8E,$8F,$45,$F7,$7D,$31,$AA,$E8,$7F,$95,$E7,$58
	.dc.b	$79,$E7,$58,$FD,$CC,$7C,$6F,$C4,$BE,$2D,$35,$87,$03,$DB,$1A,$7E
	.dc.b	$E4,$F2,$73,$3C,$5A,$68,$E4,$7A,$0F,$64,$F9,$E7,$94,$66,$3B,$36
	.dc.b	$7C,$6F,$07,$B3,$93,$B1,$93,$C2,$64,$FB,$87,$CC,$3E,$E9,$E3,$2D
	.dc.b	$FE,$89,$B8,$FF,$47,$7F,$4D,$E9,$FF,$11,$78,$9F,$98,$70,$F2,$6B
	.dc.b	$C3,$9B,$6E,$DF,$CC,$47,$3C,$3F,$D1,$DF,$46,$FA,$4F,$F4,$F4,$7E
	.dc.b	$E2,$75,$97,$3C,$3F,$29,$18,$42,$F9,$73,$F6,$C6,$4F,$07,$F5,$93
	.dc.b	$A8,$C1,$30,$F0,$4E,$B1,$F1,$E3,$E1,$9C,$D6,$E3,$F4,$F7,$CE,$72
	.dc.b	$48,$E0,$6A,$B9,$D7,$F0,$BC,$7D,$4F,$C5,$3A,$1E,$CB,$E9,$5C,$A3
	.dc.b	$86,$1E,$CC,$3E,$4E,$1D,$27,$94,$78,$5D,$9F,$C0,$6F,$9B,$88,$D3
	.dc.b	$C2,$FC,$D8,$F5,$7C,$11,$93,$C9,$7D,$F3,$FF,$3B,$63,$B9,$E9,$5C
	.dc.b	$9F,$59,$C4,$98,$F8,$9C,$57,$27,$71,$98,$C6,$3E,$66,$D7,$58,$C6
	.dc.b	$5F,$2A,$73,$7E,$6C,$BC,$8F,$69,$F9,$F8,$75,$78,$13,$1F,$C0,$57
	.dc.b	$AF,$A3,$FB,$67,$99,$FE,$F1,$E2,$FF,$79,$46,$27,$FC,$51,$F1,$19
	.dc.b	$7F,$B5,$55,$CB,$FB,$53,$6B,$84,$7F,$7C,$AC,$75,$7F,$9A,$6F,$FB
	.dc.b	$A7,$D5,$E0,$E5,$DF,$FE,$8B,$F6,$F1,$3A,$26,$1F,$E5,$18,$CB,$FE
	.dc.b	$FD,$F9,$CF,$F2,$5C,$FF,$81,$9B,$1F,$E0,$CF,$88,$F2,$46,$35,$F6
	.dc.b	$B6,$BF,$5B,$A2,$31,$6E,$FF,$F0,$26,$1C,$BA,$3E,$04,$7F,$81,$FE
	.dc.b	$4F,$D3,$E0,$3D,$8C,$F7,$A7,$31,$86,$E9,$FE,$53,$6E,$5F,$E2,$BF
	.dc.b	$0B,$E1,$FD,$13,$93,$BF,$88,$9C,$C6,$1B,$87,$FB,$59,$B7,$FE,$C2
	.dc.b	$EB,$78,$5F,$5E,$11,$8A,$FF,$49,$F9,$0F,$AD,$FF,$28,$57,$1B,$E1
	.dc.b	$7B,$8C,$5F,$0A,$DF,$2A,$3F,$B1,$DF,$02,$3B,$F2,$2D,$FF,$A9,$7E
	.dc.b	$95,$F0,$3F,$8B,$F5,$D1,$FE,$E5,$BF,$F4,$EB,$A6,$F8,$1B,$8C,$5F
	.dc.b	$38,$C5,$E5,$FA,$BE,$9C,$BF,$F4,$17,$C5,$5B,$FF,$4E,$4D,$D3,$FC
	.dc.b	$95,$36,$BF,$C0,$FB,$BE,$0F,$E5,$36,$3F,$D7,$63,$E9,$5F,$1B,$AA
	.dc.b	$93,$FE,$9D,$AE,$19,$31,$F9,$0F,$C9,$C3,$8D,$6E,$F8,$CF,$FB,$F2
	.dc.b	$F8,$DF,$0A,$F0,$61,$A8,$FF,$4E,$7E,$63,$FD,$22,$7F,$CF,$2D,$BF
	.dc.b	$F3,$7D,$7E,$C2,$31,$5C,$27,$07,$E6,$A6,$D7,$FB,$3B,$E8,$BF,$24
	.dc.b	$DF,$43,$EC,$46,$2B,$FD,$C0,$FF,$99,$67,$87,$36,$AB,$6D,$8F,$F7
	.dc.b	$1F,$C0,$3A,$BF,$83,$61,$F1,$67,$FB,$0C,$E7,$C8,$EF,$EC,$7F,$11
	.dc.b	$8C,$4F,$2B,$E3,$A7,$FB,$E9,$B1,$FD,$AB,$99,$A3,$0E,$55,$AC,$98
	.dc.b	$F4,$7C,$46,$23,$57,$CA,$B8,$E5,$FB,$8F,$0E,$65,$FE,$FE,$F0,$9D
	.dc.b	$DF,$0F,$2E,$7E,$3F,$BE,$F7,$18,$CD,$1F,$C0,$38,$33,$BF,$FA,$77
	.dc.b	$E9,$7E,$C7,$97,$FC,$39,$BA,$7F,$B7,$5B,$8F,$FA,$B1,$B5,$F8,$E8
	.dc.b	$FD,$B6,$DF,$C0,$EB,$E1,$1F,$9A,$E7,$74,$66,$64,$FF,$BD,$4D,$D7
	.dc.b	$C4,$7F,$41,$3A,$8F,$8E,$76,$8E,$13,$FD,$1D,$D9,$83,$F9,$C7,$27
	.dc.b	$F3,$63,$F8,$17,$D2,$3E,$FF,$13,$FE,$22,$CB,$D5,$E4,$3D,$6F,$E3
	.dc.b	$8D,$9F,$E8,$5B,$E2,$7F,$A0,$B2,$3E,$A7,$90,$FA,$2C,$6B,$FD,$2F
	.dc.b	$AB,$FA,$39,$D3,$D4,$B7,$E1,$3F,$D1,$39,$7F,$AC,$9C,$57,$F8,$DA
	.dc.b	$30,$FE,$23,$CD,$7D,$2F,$F1,$B8,$F8,$B3,$AD,$7F,$B7,$CE,$5F,$6E
	.dc.b	$1F,$DA,$E3,$18,$78,$23,$F9,$D7,$6B,$E5,$BF,$B9,$73,$BE,$0F,$F7
	.dc.b	$77,$FE,$9E,$AD,$3E,$69,$D6,$7C,$AE,$4F,$AB,$BD,$F8,$6B,$D2,$CD
	.dc.b	$FD,$CC,$79,$B7,$0F,$E7,$15,$BC,$3E,$1C,$FE,$AA,$7C,$FE,$2C,$9F
	.dc.b	$17,$B0,$F9,$BC,$5F,$5A,$BC,$1F,$C9,$35,$FF,$34,$56,$30,$F6,$55
	.dc.b	$B5,$6A,$7C,$39,$FF,$26,$7F,$BE,$FD,$9B,$EC,$3F,$C2,$CD,$8F,$DB
	.dc.b	$7C,$CB,$EC,$F3,$52,$7E,$EF,$32,$F8,$65,$FC,$13,$C7,$63,$C6,$F7
	.dc.b	$5A,$DB,$FC,$56,$D8,$F1,$E3,$FE,$0A,$AE,$18,$7E,$1E,$B4,$B1,$7C
	.dc.b	$7B,$E9,$FF,$C5,$CD,$C7,$D3,$E8,$62,$64,$7A,$27,$57,$E7,$C7,$FB
	.dc.b	$3E,$7F,$E3,$E9,$CB,$6F,$8D,$6F,$0B,$3F,$E4,$4F,$C2,$DF,$3F,$FE
	.dc.b	$3F,$6E,$5F,$C7,$F9,$5F,$4D,$9A,$93,$F2,$AB,$78,$FC,$9B,$DC,$E6
	.dc.b	$B7,$3F,$A9,$AD,$E1,$FD,$9B,$83,$F4,$18,$EA,$3F,$CA,$FF,$28,$F8
	.dc.b	$BF,$F2,$93,$72,$FE,$65,$F4,$1F,$E1,$67,$FC,$7F,$19,$CB,$F6,$CD
	.dc.b	$BE,$77,$FE,$3E,$BE,$15,$F4,$EB,$57,$FE,$C6,$F6,$FF,$D6,$BF,$2B
	.dc.b	$FE,$3F,$AF,$EA,$F9,$7F,$7C,$8C,$47,$FA,$E8,$FF,$18,$87,$3C,$FF
	.dc.b	$0A,$31,$EF,$75,$69,$D3,$0E,$F1,$9A,$E3,$7D,$E3,$07,$F0,$17,$15
	.dc.b	$FE,$E4,$C3,$8B,$CC,$EA,$79,$5C,$BF,$D8,$A3,$8B,$93,$78,$79,$EF
	.dc.b	$DB,$1F,$ED,$6A,$E3,$39,$BF,$1C,$F7,$AD,$BF,$09,$F4,$E1,$E0,$8E
	.dc.b	$47,$6C,$3C,$2E,$27,$6F,$A0,$C5,$7B,$EB,$94,$75,$8C,$3B,$30,$78
	.dc.b	$DF,$2B,$EF,$8E,$38,$7A,$AF,$C3,$3F,$60,$FF,$7F,$C7,$FC,$00,$7C
	.dc.b	$E6,$0F,$BC,$7F,$BE,$3E,$8F,$FC,$1D,$F4,$D8,$8C,$79,$C8,$C4,$27
	.dc.b	$6E,$DA,$F4,$5F,$25,$FE,$49,$8D,$35,$D3,$5F,$C8,$47,$0E,$0E,$ED
	.dc.b	$8F,$C0,$FD,$33,$7F,$CB,$6D,$F9,$88,$C4,$FE,$12,$B8,$4F,$E7,$1B
	.dc.b	$B7,$E7,$6B,$75,$D2,$F9,$CF,$8B,$1F,$46,$1C,$5F,$57,$D8,$F7,$AF
	.dc.b	$D3,$36,$C1,$E1,$96,$1B,$1D,$CF,$CC,$E3,$E2,$C3,$24,$C7,$CD,$87
	.dc.b	$92,$BA,$D6,$E7,$CD,$87,$BF,$1F,$3C,$70,$AE,$B7,$C5,$FB,$13,$A7
	.dc.b	$F0,$0D,$57,$D5,$EF,$3C,$72,$F9,$31,$C3,$0E,$AC,$3F,$6B,$1A,$7B
	.dc.b	$C7,$C1,$C3,$C9,$1F,$15,$B9,$7C,$43,$F8,$A7,$5F,$E3,$7E,$9B,$F1
	.dc.b	$ED,$AF,$CA,$35,$FC,$CB,$5F,$E3,$0E,$6F,$FD,$2F,$EC,$3E,$B3,$FB
	.dc.b	$65,$7F,$88,$A3,$1E,$07,$F1,$FA,$FF,$4E,$DF,$1E,$4F,$F7,$7F,$D3
	.dc.b	$FF,$4A,$C3,$FB,$6B,$5F,$F1,$43,$6B,$FA,$07,$3F,$FB,$86,$77,$D7
	.dc.b	$EA,$AE,$11,$E6,$6B,$DA,$34,$78,$EF,$F1,$79,$7F,$B8,$BE,$53,$14
	.dc.b	$8C,$3D,$8F,$F0,$0F,$D3,$FF,$BC,$DB,$F1,$1C,$34,$4E,$67,$FA,$E7
	.dc.b	$B9,$98,$C4,$7C,$A6,$BF,$9C,$9E,$35,$F5,$27,$33,$FE,$01,$7A,$6B
	.dc.b	$F4,$3C,$DF,$F0,$FC,$FC,$F6,$BF,$E4,$F9,$DE,$7D,$7E,$01,$F2,$1E
	.dc.b	$76,$D7,$F8,$71,$AF,$F8,$17,$D8,$D4,$FE,$1A,$73,$3F,$82,$6B,$D6
	.dc.b	$34,$7C,$AD,$7F,$6D,$EB,$5E,$6F,$F3,$75,$77,$AF,$9E,$BE,$9D,$F3
	.dc.b	$38,$6C,$7F,$92,$9A,$FF,$A3,$3E,$9F,$FD,$2A,$E6,$7F,$99,$4E,$A3
	.dc.b	$F9,$8B,$5F,$EE,$0D,$7F,$85,$BD,$37,$FF,$52,$53,$E0,$7C,$AD,$AE
	.dc.b	$C7,$8F,$E9,$7C,$C4,$7F,$CD,$1F,$4E,$F9,$AC,$47,$F3,$FC,$55,$E6
	.dc.b	$24,$7F,$40,$8C,$BF,$E7,$5F,$A7,$FF,$5B,$65,$FE,$61,$6F,$17,$AA
	.dc.b	$FE,$37,$C6,$4B,$F8,$CF,$E4,$FE,$9F,$C5,$18,$3B,$47,$C6,$23,$FC
	.dc.b	$03,$E9,$9F,$4F,$FC,$11,$B1,$98,$C1,$9F,$66,$47,$16,$BE,$FF,$4F
	.dc.b	$E9,$91,$F7,$3D,$7D,$09,$1E,$16,$BF,$D9,$7E,$99,$F2,$1F,$86,$6C
	.dc.b	$7F,$A3,$1A,$FF,$B3,$BC,$A7,$91,$BA,$FE,$E9,$5A,$AF,$EE,$6D,$7F
	.dc.b	$C6,$4D,$7F,$A6,$BD,$38,$7F,$DC,$D7,$FE,$95,$9F,$F4,$DB,$5F,$F6
	.dc.b	$EC,$EF,$83,$FA,$AF,$A7,$FE,$FE,$F9,$5F,$D1,$7C,$B7,$F9,$E2,$FF
	.dc.b	$92,$35,$FF,$44,$7C,$7A,$7F,$88,$D7,$19,$DF,$B5,$C8,$CD,$6E,$3F
	.dc.b	$97,$7A,$4F,$9C,$FD,$83,$63,$31,$C3,$E9,$B2,$C8,$C9,$FE,$D8,$C1
	.dc.b	$6A,$FE,$47,$EC,$BE,$99,$F4,$D8,$6E,$1F,$EE,$09,$D4,$FF,$2A,$F2
	.dc.b	$BF,$DA,$5F,$29,$B8,$FF,$0E,$CE,$AB,$FC,$44,$D7,$FC,$4E,$D7,$FD
	.dc.b	$B1,$7D,$97,$1B,$87,$FC,$07,$8B,$E8,$FF,$C5,$F3,$9A,$4E,$5F,$11
	.dc.b	$FE,$DD,$F8,$97,$D2,$F9,$D7,$D3,$53,$11,$27,$84,$65,$E5,$1F,$7B
	.dc.b	$DC,$91,$F7,$9F,$AF,$D3,$88,$F4,$E7,$FD,$43,$9F,$FC,$73,$1B,$3E
	.dc.b	$97,$17,$E2,$DF,$F0,$FF,$CD,$FF,$15,$AF,$F4,$43,$63,$3E,$57,$F2
	.dc.b	$AF,$A6,$7C,$87,$F2,$CA,$FE,$3D,$46,$92,$BF,$CB,$0F,$38,$57,$F6
	.dc.b	$D7,$CE,$7F,$8C,$1C,$70,$7C,$D8,$E9,$5F,$9E,$F8,$19,$1F,$9E,$7F
	.dc.b	$37,$F2,$5F,$53,$2F,$81,$1A,$9E,$BE,$93,$54,$F0,$9F,$C5,$72,$7F
	.dc.b	$AE,$DF,$F2,$0E,$5F,$77,$93,$F0,$1E,$CE,$8D,$F8,$72,$B7,$9F,$E1
	.dc.b	$B0,$56,$E7,$27,$E1,$A3,$B3,$F0,$0F,$91,$3F,$2C,$8F,$0F,$9B,$F9
	.dc.b	$FE,$76,$23,$FC,$F9,$C3,$F1,$5F,$F4,$0B,$5F,$B3,$5F,$40,$FF,$07
	.dc.b	$D7,$D8,$E6,$BC,$7F,$94,$AF,$29,$E2,$3F,$0B,$7F,$BE,$FA,$69,$18
	.dc.b	$21,$F3,$3C,$14,$6A,$3F,$87,$CF,$D4,$F8,$DF,$E2,$BF,$31,$27,$F8
	.dc.b	$17,$2F,$F8,$2B,$8B,$FD,$DC,$DD,$7F,$31,$5B,$AF,$C2,$F8,$4A,$D4
	.dc.b	$3D,$C7,$D1,$3C,$FF,$F2,$7B,$67,$FC,$9E,$D7,$FC,$9E,$D7,$FC,$9E
	.dc.b	$E7,$2F,$F2,$7F,$43,$E5,$FF,$96,$9F,$E3,$19,$D6,$9F,$E2,$B5,$EE
	.dc.b	$64,$98,$7F,$32,$AA,$F8,$23,$CA,$47,$CB,$F3,$FE,$82,$38,$21,$F9
	.dc.b	$AC,$1F,$31,$FD,$00,$F2,$4F,$F8,$7B,$E9,$1F,$31,$FB,$6B,$7C,$6F
	.dc.b	$F8,$C2,$73,$CF,$FE,$E9,$F7,$BF,$D3,$F4,$9C,$BC,$23,$3F,$4B,$FC
	.dc.b	$FE,$30,$7F,$A9,$A7,$77,$D7,$0F,$03,$70,$ED,$5F,$EA,$CB,$E5,$39
	.dc.b	$8E,$67,$86,$B8,$E1,$E4,$C3,$AB,$E2,$F3,$FC,$F7,$C4,$91,$BA,$E3
	.dc.b	$7D,$EB,$73,$9C,$7C,$73,$DD,$B8,$76,$8E,$06,$2F,$9B,$B3,$CF,$1B
	.dc.b	$58,$FB,$2F,$B4,$6A,$B8,$D6,$EB,$53,$F2,$67,$24,$BF,$2C,$EA,$7C
	.dc.b	$18,$FB,$9E,$13,$AC,$7D,$B1,$83,$C1,$1E,$93,$4D,$32,$49,$FF,$39
	.dc.b	$D7,$7C,$3D,$84,$C3,$E0,$C6,$5F,$35,$71,$C7,$EA,$3D,$6F,$CC,$FA
	.dc.b	$23,$2E,$CF,$34,$6A,$BA,$61,$C6,$BE,$4C,$F9,$31,$F6,$DF,$33,$8C
	.dc.b	$E8,$EB,$3E,$0C,$7D,$17,$F0,$2B,$DE,$7E,$78,$FB,$45,$8F,$CF,$B7
	.dc.b	$1F,$D4,$39,$E2,$FF,$4F,$BF,$93,$6B,$FA,$16,$BF,$A9,$6C,$FF,$13
	.dc.b	$E7,$FC,$6C,$7F,$AA,$B0,$FF,$50,$FD,$2B,$E6,$3F,$C1,$51,$FB,$59
	.dc.b	$D6,$7D,$EF,$89,$F9,$8B,$E1,$C9,$8C,$BF,$27,$7C,$1F,$25,$6F,$1E
	.dc.b	$B8,$73,$9D,$CF,$1C,$7C,$75,$A9,$E2,$68,$91,$AC,$38,$5F,$38,$F4
	.dc.b	$5F,$79,$CD,$72,$9C,$BF,$BE,$C7,$D2,$F4,$BE,$44,$42,$37,$3C,$89
	.dc.b	$3B,$9D,$61,$E9,$7E,$96,$1E,$26,$9E,$18,$D6,$5F,$56,$3C,$47,$9E
	.dc.b	$BF,$91,$E1,$E0,$7B,$4E,$F0,$F0,$D7,$9A,$B4,$F1,$21,$E0,$8F,$90
	.dc.b	$F6,$31,$5E,$99,$E3,$18,$38,$D7,$79,$F1,$DF,$A3,$2F,$BD,$5B,$25
	.dc.b	$F2,$BE,$11,$CA,$77,$39,$78,$CE,$A7,$8D,$75,$C7,$F9,$BF,$37,$F6
	.dc.b	$CF,$43,$FC,$BD,$CF,$FD,$93,$D0,$BB,$7F,$61,$9B,$8B,$C7,$FF,$17
	.dc.b	$56,$30,$E5,$7F,$DC,$AF,$FC,$2D,$18,$BF,$EE,$71,$8E,$97,$D2,$D2
	.dc.b	$5F,$F7,$A6,$D7,$F9,$22,$FF,$C0,$31,$E2,$C7,$FC,$7D,$F3,$7E,$1C
	.dc.b	$39,$E1,$FE,$32,$B7,$D2,$FF,$96,$A3,$FD,$4F,$5E,$78,$C6,$3F,$DB
	.dc.b	$E3,$1C,$BE,$89,$CC,$FF,$68,$BE,$73,$93,$FC,$65,$DD,$FE,$17,$F9
	.dc.b	$0F,$27,$67,$FA,$C7,$2F,$F1,$0C,$63,$D7,$F9,$35,$B9,$FF,$5C,$C6
	.dc.b	$3C,$05,$6E,$7F,$CF,$3E,$0F,$BF,$E8,$9F,$17,$CB,$D1,$4E,$63,$0D
	.dc.b	$AE,$27,$F9,$53,$89,$7C,$A3,$FC,$87,$DA,$F0,$7E,$EF,$92,$FA,$5F
	.dc.b	$F6,$15,$6A,$73,$18,$E3,$FF,$2E,$5F,$1A,$FE,$61,$1F,$BC,$8C,$38
	.dc.b	$6B,$FC,$D5,$AF,$F7,$EA,$FC,$6B,$FE,$1A,$7B,$F8,$F8,$9E,$7D,$7F
	.dc.b	$E7,$7F,$A5,$7B,$FF,$94,$F4,$51,$FE,$45,$CB,$FC,$E7,$18,$9D,$4E
	.dc.b	$72,$FF,$49,$CE,$6B,$73,$9B,$ED,$7C,$3C,$2F,$BB,$EA,$CB,$FF,$2E
	.dc.b	$FD,$1D,$CB,$FC,$71,$37,$0F,$F2,$FC,$DA,$FF,$18,$EE,$F8,$1F,$D5
	.dc.b	$C3,$E9,$1F,$17,$F0,$AB,$53,$FE,$8C,$35,$19,$AF,$F4,$16,$4A,$DB
	.dc.b	$F4,$9A,$FF,$63,$6B,$E2,$8E,$55,$D6,$DF,$1D,$8A,$FF,$55,$FD,$2F
	.dc.b	$F3,$6A,$FC,$53,$63,$FC,$EF,$D7,$C2,$73,$1F,$B5,$9D,$69,$FE,$E9
	.dc.b	$E7,$46,$2B,$FA,$0F,$6F,$F9,$FF,$1E,$18,$7D,$BF,$3F,$F3,$B5,$F6
	.dc.b	$9F,$F3,$B3,$7A,$B5,$E8,$ED,$B1,$93,$84,$7F,$18,$F8,$0F,$D0,$F9
	.dc.b	$1F,$87,$F5,$DF,$23,$9E,$F4,$E6,$30,$FF,$AB,$EB,$8E,$1F,$9E,$6E
	.dc.b	$7F,$E0,$F6,$BF,$E6,$E7,$3C,$09,$FE,$BF,$18,$7F,$B0,$F5,$FF,$9F
	.dc.b	$BD,$BE,$F1,$FB,$7C,$BC,$AD,$CB,$F8,$06,$9F,$C0,$D9,$19,$3F,$DF
	.dc.b	$F1,$8C,$BE,$AF,$3F,$D3,$F3,$7F,$87,$36,$BF,$D6,$AD,$9D,$C7,$DE
	.dc.b	$9F,$D9,$BF,$EE,$A3,$F2,$5C,$2E,$AC,$E7,$F9,$A8,$C4,$FF,$B3,$8E
	.dc.b	$17,$FC,$11,$AF,$F8,$D5,$AF,$FA,$51,$AF,$E3,$E7,$AC,$7E,$57,$AA
	.dc.b	$CF,$F4,$3F,$27,$F0,$AF,$F8,$FF,$C9,$FE,$7F,$C9,$FC,$E4,$97,$FC
	.dc.b	$52,$5E,$4F,$FA,$EA,$30,$D8,$FE,$32,$D7,$FC,$94,$D7,$FC,$2F,$5F
	.dc.b	$D3,$36,$FE,$8C,$4A,$FE,$43,$19,$6E,$1F,$EB,$A6,$F3,$7F,$4D,$23
	.dc.b	$E2,$F4,$BF,$77,$8F,$FA,$F2,$30,$4C,$7C,$B3,$FC,$9A,$75,$1F,$D9
	.dc.b	$7E,$8F,$F8,$C5,$7F,$AF,$A7,$3C,$DF,$91,$AF,$C8,$FC,$EF,$F7,$AE
	.dc.b	$27,$C1,$7E,$CE,$73,$87,$F8,$D2,$F8,$56,$A7,$F9,$AF,$43,$0F,$E2
	.dc.b	$2D,$BF,$9C,$D9,$FD,$34,$E8,$E2,$7F,$37,$FA,$37,$9F,$8B,$F0,$1F
	.dc.b	$A9,$F0,$BC,$DF,$59,$FD,$AF,$06,$7E,$B5,$97,$F2,$08,$F3,$CE,$DD
	.dc.b	$47,$F2,$09,$FD,$8D,$7A,$09,$7E,$87,$DA,$7C,$5F,$51,$F4,$98,$9F
	.dc.b	$F0,$67,$D1,$BE,$6F,$FC,$25,$F1,$1F,$43,$17,$3A,$B5,$7F,$DC,$6B
	.dc.b	$53,$9B,$FF,$12,$D6,$AF,$F3,$4D,$7F,$C2,$0D,$7F,$22,$EF,$4F,$B9
	.dc.b	$5C,$F0,$F9,$34,$F1,$1F,$46,$FA,$4F,$F1,$B6,$3C,$6F,$F3,$7F,$18
	.dc.b	$FA,$4F,$F1,$67,$95,$FC,$5E,$7F,$C5,$53,$97,$FC,$15,$19,$CB,$FC
	.dc.b	$69,$39,$7F,$64,$D7,$FC,$50,$D7,$BD,$7F,$91,$23,$86,$9E,$A6,$DF
	.dc.b	$F4,$4C,$BF,$BE,$BE,$2A,$FF,$27,$7D,$23,$E7,$7F,$C8,$5E,$B3,$E5
	.dc.b	$B3,$18,$6D,$FE,$D2,$B8,$36,$7F,$CA,$B1,$8C,$FF,$85,$37,$3F,$9A
	.dc.b	$DA,$FE,$AF,$D4,$FA,$4E,$B7,$FE,$55,$F9,$37,$C6,$FF,$95,$6F,$C8
	.dc.b	$7F,$94,$E3,$2F,$F8,$CC,$CC,$F1,$9F,$F3,$3C,$EA,$3F,$8E,$35,$FF
	.dc.b	$29,$35,$FF,$16,$0F,$9C,$FF,$24,$BF,$D3,$AF,$F8,$04,$62,$3F,$D5
	.dc.b	$E3,$D9,$FF,$58,$C6,$7B,$3D,$98,$E3,$DE,$EB,$F1,$79,$C7,$F8,$53
	.dc.b	$CF,$1F,$F1,$F0,$75,$5F,$EB,$C7,$51,$DC,$D5,$7F,$65,$AE,$6F,$63
	.dc.b	$15,$C5,$C4,$FA,$E3,$0F,$F9,$24,$E6,$64,$A7,$BB,$1D,$BF,$C1,$AB
	.dc.b	$BB,$C6,$FC,$EF,$03,$D6,$E0,$F4,$CF,$2C,$3E,$2C,$EC,$F6,$63,$C8
	.dc.b	$DB,$F4,$F1,$F5,$B4,$F0,$38,$BD,$9D,$DF,$7C,$79,$F1,$F4,$BE,$48
	.dc.b	$E0,$7E,$01,$FE,$B9,$87,$F5,$D7,$E9,$98,$3F,$CE,$7F,$44,$F3,$5F
	.dc.b	$49,$FD,$8E,$30,$D8,$CC,$F0,$6B,$F8,$81,$2F,$93,$FE,$67,$1D,$9A
	.dc.b	$EE,$35,$C5,$DE,$73,$3F,$92,$9C,$BF,$63,$E9,$1C,$1F,$96,$D7,$F3
	.dc.b	$11,$89,$FB,$75,$C2,$7F,$38,$DD,$3F,$3B,$5B,$AE,$97,$CE,$7C,$18
	.dc.b	$7A,$30,$E2,$D8,$E3,$3B,$A0,$BE,$97,$C3,$1F,$1E,$1C,$6B,$75,$AC
	.dc.b	$39,$C6,$AA,$3E,$43,$F2,$B3,$B7,$BE,$3E,$B6,$DF,$57,$C5,$8F,$95
	.dc.b	$A7,$4C,$7C,$8F,$52,$38,$26,$3E,$48,$CB,$F1,$31,$E4,$E6,$7E,$8D
	.dc.b	$77,$C3,$C9,$DD,$F8,$2F,$98,$FE,$17,$87,$F9,$F7,$E9,$3F,$B2,$65
	.dc.b	$FE,$CF,$6B,$FE,$F9,$B5,$70,$E0,$FF,$15,$C6,$27,$87,$83,$95,$DC
	.dc.b	$C3,$9E,$3F,$D8,$2F,$85,$7F,$81,$30,$FF,$1A,$FD,$1F,$F4,$58,$FF
	.dc.b	$70,$71,$BE,$57,$A6,$5F,$C8,$5E,$DA,$AF,$85,$F3,$3F,$BD,$B8,$CB
	.dc.b	$FD,$35,$87,$39,$EB,$7C,$AB,$58,$F8,$BA,$FB,$DE,$EF,$A6,$5E,$17
	.dc.b	$FB,$BE,$5E,$1A,$F1,$FD,$1F,$FB,$FA,$7F,$DF,$1E,$DB,$CD,$F4,$B6
	.dc.b	$38,$CE,$DF,$45,$A3,$1A,$FF,$AC,$6B,$85,$78,$9B,$1F,$E0,$A9,$DE
	.dc.b	$7F,$E5,$E6,$D7,$D2,$AD,$CF,$E9,$32,$46,$27,$FC,$43,$5C,$30,$FF
	.dc.b	$65,$FC,$87,$D0,$EA,$7D,$1D,$3E,$8F,$B2,$9A,$67,$0F,$F0,$DE,$5F
	.dc.b	$D7,$6B,$31,$F2,$2B,$83,$AB,$EA,$D9,$DB,$F6,$C7,$D0,$BE,$67,$F1
	.dc.b	$D0,$FA,$3F,$F7,$8B,$FE,$2A,$31,$EF,$6A,$B8,$46,$5A,$F5,$9E,$2D
	.dc.b	$9E,$15,$23,$A4,$EB,$83,$FC,$B4,$DE,$0F,$EE,$B8,$74,$9F,$F2,$F3
	.dc.b	$FE,$1D,$AF,$F1,$EF,$0B,$E8,$BF,$CC,$55,$FE,$76,$6D,$F2,$F1,$FD
	.dc.b	$07,$EA,$EB,$84,$59,$DD,$FF,$1E,$C3,$B3,$F1,$DB,$87,$5A,$E0,$74
	.dc.b	$9F,$85,$F1,$1F,$14,$8F,$E7,$3E,$89,$F3,$98,$8F,$CA,$FD,$13,$E8
	.dc.b	$BF,$62,$D8,$E6,$7E,$91,$C3,$F9,$CF,$AF,$FF,$5C,$BF,$A8,$6B,$D2
	.dc.b	$79,$FB,$1F,$5F,$E3,$A4,$F3,$8A,$FE,$F7,$AD,$F5,$2D,$3F,$88,$46
	.dc.b	$1D,$ED,$FE,$6E,$CF,$FC,$7B,$7F,$94,$F8,$CF,$F2,$13,$7C,$EF,$5F
	.dc.b	$9F,$F8,$CF,$A4,$FF,$73,$4F,$E2,$FC,$9F,$F4,$5B,$7D,$BD,$7C,$4C
	.dc.b	$8C,$10,$FC,$C5,$72,$6C,$7E,$62,$B9,$C7,$E7,$5B,$9F,$FA,$96,$B8
	.dc.b	$3F,$EC,$1F,$6D,$F2,$AF,$96,$FE,$47,$F2,$EF,$72,$9F,$CC,$6B,$73
	.dc.b	$6B,$86,$5F,$B4,$BE,$0D,$BD,$57,$F2,$EF,$45,$F3,$B4,$95,$FE,$96
	.dc.b	$C7,$C7,$F4,$98,$8F,$E4,$B5,$C1,$FE,$02,$D7,$F9,$A4,$F0,$7F,$D3
	.dc.b	$FE,$CB,$E0,$5F,$39,$F6,$BE,$4F,$FC,$31,$BF,$FB,$87,$A5,$18,$F0
	.dc.b	$7F,$BE,$E1,$C6,$FF,$A8,$37,$1F,$F5,$5D,$F0,$7A,$1A,$C7,$BB,$E2
	.dc.b	$6B,$F8,$DC,$3B,$3F,$C9,$9B,$5D,$63,$F2,$5F,$10,$FA,$2F,$C8,$AF
	.dc.b	$A4,$A6,$22,$4F,$92,$74,$D9,$CB,$FD,$3B,$8F,$FC,$6C,$FD,$7F,$73
	.dc.b	$EF,$9F,$6F,$A2,$C6,$48,$7C,$EE,$6F,$C4,$EF,$FE,$88,$F9,$AF,$E2
	.dc.b	$8F,$FA,$2B,$E9,$1F,$27,$FC,$A9,$B7,$FC,$A6,$B6,$FF,$8A,$9A,$A3
	.dc.b	$6F,$F3,$56,$F1,$FF,$BF,$67,$67,$08,$CD,$F8,$6F,$F9,$CC,$F6,$F4
	.dc.b	$DA,$7F,$97,$F1,$FC,$B7,$C0,$FE,$78,$FC,$E7,$CC,$7D,$4A,$DC,$66
	.dc.b	$76,$7E,$56,$7E,$CF,$2F,$E2,$3C,$0F,$C6,$7B,$92,$BF,$0B,$CD,$FC
	.dc.b	$11,$FC,$04,$E5,$FC,$14,$FE,$4E,$3C,$FA,$FF,$87,$7E,$69,$F2,$1F
	.dc.b	$81,$BF,$F7,$CC,$63,$9F,$71,$F2,$EB,$85,$6E,$3F,$C2,$D5,$B7,$44
	.dc.b	$33,$1B,$E2,$57,$F9,$72,$B8,$57,$F6,$5E,$0F,$E8,$7D,$7F,$C1,$9F
	.dc.b	$DD,$7D,$24,$9F,$DC,$72,$A7,$F8,$21,$98,$FD,$CC,$E1,$F9,$5E,$97
	.dc.b	$F0,$EE,$59,$18,$21,$F9,$77,$2F,$EF,$B8,$BF,$C1,$9E,$DB,$E7,$FF
	.dc.b	$9B,$4F,$F1,$1F,$97,$FE,$32,$7F,$8E,$A3,$31,$FE,$22,$6E,$FF,$02
	.dc.b	$3F,$C9,$7D,$0A,$E3,$1F,$E3,$2A,$DF,$07,$FA,$03,$0F,$07,$D2,$7F
	.dc.b	$92,$3C,$B7,$BB,$4F,$04,$4C,$3F,$A2,$46,$5E,$91,$E5,$F9,$FF,$E5
	.dc.b	$FF,$31,$27,$C1,$19,$7F,$96,$46,$5F,$54,$7F,$85,$BE,$89,$F3,$3F
	.dc.b	$D8,$BE,$17,$FC,$43,$49,$CF,$4F,$FA,$62,$73,$18,$9F,$F8,$01,$F0
	.dc.b	$7F,$E9,$C5,$F4,$9F,$CB,$6B,$78,$75,$AD,$5F,$19,$CE,$1D,$67,$95
	.dc.b	$F1,$77,$7C,$70,$FF,$54,$E1,$C6,$73,$7C,$67,$FD,$5B,$7C,$5D,$DF
	.dc.b	$17,$D8,$DB,$E5,$1C,$3C,$DF,$F7,$F6,$1D,$23,$DA,$72,$C3,$B3,$C9
	.dc.b	$B8,$77,$8C,$1C,$0B,$8F,$B9,$C4,$6F,$0F,$0C,$6A,$FC,$15,$F4,$AB
	.dc.b	$58,$F9,$C9,$87,$83,$1F,$3E,$1E,$53,$C1,$8F,$43,$AD,$F2,$AF,$4C
	.dc.b	$66,$FE,$24,$E6,$B8,$C7,$67,$CF,$7D,$9F,$31,$DF,$0F,$61,$E7,$8D
	.dc.b	$5F,$5C,$3B,$DF,$69,$CB,$D6,$B8,$E3,$F8,$09,$CD,$70,$26,$1E,$22
	.dc.b	$D7,$07,$C5,$5C,$5C,$BF,$28,$8F,$DD,$32,$70,$21,$F9,$8A,$F0,$BF
	.dc.b	$9B,$3E,$61,$70,$FD,$03,$70,$FF,$23,$BF,$9C,$F5,$7F,$D4,$6D,$E8
	.dc.b	$E4,$FE,$62,$32,$D8,$CB,$FA,$F8,$FF,$59,$63,$FA,$E6,$E3,$FA,$DF
	.dc.b	$A4,$7C,$BF,$F1,$1A,$D6,$1F,$72,$F7,$87,$53,$F2,$57,$C3,$93,$55
	.dc.b	$27,$F0,$F5,$C1,$56,$DC,$CE,$EF,$BC,$74,$C3,$C5,$8F,$BE,$36,$49
	.dc.b	$FE,$17,$7C,$DE,$24,$9E,$0E,$AB,$9D,$7A,$63,$4D,$9F,$85,$3A,$8F
	.dc.b	$91,$7E,$59,$E6,$6B,$1C,$4F,$12,$1E,$EA,$E5,$9F,$F9,$DE,$F7,$3D
	.dc.b	$0F,$75,$62,$75,$1E,$DB,$EC,$D3,$AC,$6B,$0F,$65,$71,$CB,$E7,$4E
	.dc.b	$6F,$8D,$7B,$AF,$C0,$DC,$BE,$44,$FB,$27,$AD,$FD,$79,$CC,$63,$1F
	.dc.b	$4C,$62,$39,$D7,$16,$83,$E1,$63,$E5,$AE,$39,$7C,$4C,$3F,$17,$C9
	.dc.b	$FD,$9F,$E8,$3F,$B0,$74,$FF,$61,$F1,$FF,$C4,$BE,$7E,$32,$FE,$BD
	.dc.b	$D2,$BC,$69,$FD,$9A,$AE,$3C,$EF,$FD,$01,$18,$BE,$11,$83,$FB,$57
	.dc.b	$D2,$72,$C5,$D4,$F1,$B8,$4E,$63,$15,$C0,$FF,$56,$4E,$9A,$7C,$1D
	.dc.b	$3F,$BD,$64,$F9,$EF,$F1,$4C,$FF,$A4,$FE,$73,$CF,$18,$D7,$FC,$C5
	.dc.b	$18,$E8,$F2,$77,$A3,$17,$C4,$FF,$20,$7C,$96,$BD,$6F,$87,$E3,$BC
	.dc.b	$34,$E6,$B6,$7F,$A8,$EB,$73,$A8,$C1,$FE,$28,$F1,$3E,$27,$D0,$BE
	.dc.b	$93,$AF,$93,$FE,$6A,$AF,$71,$F7,$F9,$55,$B8,$E4,$6B,$74,$EA,$F9
	.dc.b	$E1,$D9,$FB,$72,$EE,$78,$7F,$7B,$CA,$7C,$CF,$FA,$DB,$2E,$97,$BA
	.dc.b	$D4,$E6,$1C,$DF,$E2,$B7,$55,$FE,$58,$9D,$4E,$70,$EF,$39,$F1,$3C
	.dc.b	$71,$8B,$FF,$2C,$DF,$0A,$FF,$0A,$4E,$6F,$FC,$71,$8F,$9B,$0F,$CA
	.dc.b	$F9,$EF,$A4,$7C,$1E,$1F,$F7,$D6,$3F,$EB,$87,$2D,$F0,$75,$3B,$6C
	.dc.b	$64,$FF,$6E,$7A,$9F,$86,$3E,$DF,$D1,$58,$C9,$36,$FF,$2B,$46,$5F
	.dc.b	$8B,$1F,$E2,$DC,$DF,$09,$FA,$C8,$7D,$23,$E3,$FE,$AD,$6A,$7E,$E1
	.dc.b	$A8,$CD,$71,$AE,$13,$F5,$AB,$85,$F8,$0F,$F3,$EC,$E9,$C0,$8D,$EF
	.dc.b	$FE,$63,$AE,$C7,$3B,$ED,$87,$ED,$6E,$4F,$FA,$8F,$D0,$B1,$DF,$9B
	.dc.b	$F7,$AE,$30,$FF,$37,$76,$7E,$9E,$B1,$7F,$C6,$78,$5E,$7F,$E7,$F0
	.dc.b	$F4,$46,$3D,$1F,$BB,$39,$C7,$FD,$F0,$FF,$9D,$EF,$FD,$E6,$E5,$E8
	.dc.b	$E6,$F9,$B9,$57,$D6,$7F,$51,$EE,$7E,$B3,$CA,$FC,$5F,$AE,$FA,$4F
	.dc.b	$E4,$33,$9A,$FF,$5E,$C6,$3B,$BF,$8F,$E3,$FE,$FD,$8C,$61,$C9,$B5
	.dc.b	$C0,$EF,$5C,$33,$FF,$49,$64,$CB,$C1,$5F,$BB,$DF,$F7,$F8,$7F,$06
	.dc.b	$3F,$82,$EB,$67,$58,$F0,$CF,$E9,$37,$2F,$A3,$DF,$F0,$DF,$E8,$11
	.dc.b	$9A,$FE,$95,$C1,$FC,$6F,$87,$F9,$97,$9D,$FC,$61,$B3,$FE,$E1,$6E
	.dc.b	$3F,$EB,$36,$E1,$FC,$06,$5E,$CF,$E7,$78,$DD,$1F,$C4,$27,$27,$E7
	.dc.b	$23,$2F,$F0,$E3,$38,$7E,$65,$FF,$56,$CF,$1A,$E8,$D8,$E4,$D9,$E1
	.dc.b	$5D,$1F,$D1,$79,$3F,$D1,$7C,$2F,$EA,$5F,$99,$3E,$6B,$FA,$2C,$61
	.dc.b	$FE,$4B,$E5,$3D,$7F,$E1,$BF,$52,$FA,$9C,$C6,$1F,$E9,$BC,$B8,$DB
	.dc.b	$F9,$96,$7F,$D4,$E3,$1A,$78,$BC,$EF,$CF,$D6,$BE,$9F,$F8,$9C,$E7
	.dc.b	$BB,$FB,$6C,$63,$8F,$F9,$66,$BF,$EB,$EB,$EB,$87,$2C,$3F,$77,$39
	.dc.b	$79,$4F,$F3,$88,$C5,$72,$F6,$7A,$F8,$9E,$E7,$FA,$1F,$80,$FA,$0F
	.dc.b	$EF,$FD,$2F,$83,$FF,$07,$65,$FD,$0A,$B4,$FF,$4F,$9D,$67,$CA,$F8
	.dc.b	$DF,$08,$FE,$1F,$7C,$1F,$E2,$95,$C2,$FA,$5F,$09,$E5,$1D,$6F,$96
	.dc.b	$5F,$D3,$70,$FE,$2F,$3F,$E2,$1C,$1F,$4C,$F4,$3D,$7E,$2F,$AA,$31
	.dc.b	$5E,$98,$C4,$EA,$3D,$53,$AC,$BC,$F5,$A9,$E0,$6A,$BF,$C4,$3F,$4A
	.dc.b	$FA,$0F,$F0,$E7,$C3,$FF,$4E,$C3,$F1,$F3,$9B,$FD,$D4,$E7,$4F,$F0
	.dc.b	$BC,$E6,$BF,$94,$4E,$6B,$FC,$79,$CC,$AD,$5F,$F8,$DE,$B5,$6A,$D6
	.dc.b	$1E,$2E,$1F,$2B,$71,$FC,$9F,$C5,$3E,$93,$FC,$65,$E3,$A7,$FC,$77
	.dc.b	$39,$7F,$88,$46,$5E,$33,$C1,$DB,$5C,$9C,$0C,$94,$FD,$1F,$D2,$3E
	.dc.b	$7F,$FC,$85,$EA,$3E,$6B,$FC,$97,$39,$AF,$E4,$13,$97,$8D,$70,$AD
	.dc.b	$B6,$32,$70,$8C,$B4,$FD,$9F,$C8,$3E,$2B,$FC,$A5,$1D,$77,$FE,$0B
	.dc.b	$39,$8F,$E6,$06,$78,$7F,$C9,$B1,$99,$FE,$BD,$39,$3E,$BC,$66,$7F
	.dc.b	$CD,$91,$99,$FF,$69,$7A,$FF,$E7,$7F,$8E,$FF,$23,$C7,$F9,$E7,$A5
	.dc.b	$F2,$5F,$4E,$13,$A7,$FC,$42,$63,$DD,$FF,$40,$2A,$E3,$87,$59,$FF
	.dc.b	$6B,$E1,$6B,$B6,$1D,$70,$E9,$1F,$E8,$99,$D3,$CE,$75,$5C,$A3,$C0
	.dc.b	$72,$BD,$BC,$4C,$9C,$5C,$BF,$E4,$D3,$38,$79,$9D,$9C,$70,$F7,$46
	.dc.b	$5F,$2C,$71,$73,$7E,$48,$F9,$67,$47,$0F,$C6,$8D,$D7,$8B,$0F,$6D
	.dc.b	$F5,$C3,$D9,$5A,$BF,$19,$DE,$3E,$6C,$EE,$3B,$CE,$CF,$2D,$74,$C3
	.dc.b	$C3,$3E,$BB,$E3,$18,$8E,$87,$C9,$3B,$CF,$DA,$7E,$71,$4F,$98,$53
	.dc.b	$E5,$1B,$FA,$1F,$ED,$BF,$49,$96,$C7,$16,$BB,$F3,$93,$BF,$90,$FF
	.dc.b	$3B,$CE,$DA,$F5,$6B,$F8,$EA,$46,$23,$FD,$A1,$18,$D7,$F1,$3F,$48
	.dc.b	$DF,$F2,$BA,$FE,$66,$31,$7F,$8E,$BE,$0F,$E2,$1B,$3B,$AE,$3A,$FE
	.dc.b	$7A,$F8,$DE,$F1,$F4,$DF,$E9,$70,$E3,$7B,$C3,$8D,$FF,$B9,$70,$E3
	.dc.b	$7B,$C3,$8D,$FB,$5B,$5B,$BE,$2A,$F8,$D6,$F1,$F5,$5F,$0A,$D5,$F0
	.dc.b	$BD,$C7,$C6,$9D,$63,$E6,$8E,$C7,$C8,$AE,$F1,$98,$36,$B9,$C7,$63
	.dc.b	$C3,$71,$FD,$8C,$68,$DC,$F3,$39,$57,$94,$EF,$7F,$03,$AB,$BF,$87
	.dc.b	$FC,$33,$1F,$F7,$5D,$A3,$1C,$3F,$D5,$23,$19,$7F,$B3,$FD,$27,$D5
	.dc.b	$BE,$F7,$73,$FD,$56,$B6,$FF,$8C,$9B,$8F,$6C,$77,$87,$2B,$E3,$97
	.dc.b	$C2,$FC,$12,$E1,$FE,$D6,$C7,$FC,$5A,$D7,$A4,$6B,$E0,$A4,$69,$E3
	.dc.b	$18,$8F,$14,$5C,$BD,$F3,$E2,$F2,$3F,$98,$47,$F3,$3C,$1F,$4A,$FA
	.dc.b	$4F,$2B,$71,$F8,$57,$C3,$7F,$EF,$33,$99,$FF,$28,$B7,$0E,$8D,$9F
	.dc.b	$F3,$8D,$71,$AF,$EF,$93,$96,$DF,$80,$F3,$71,$F8,$BC,$7F,$1F,$5C
	.dc.b	$AF,$F2,$8D,$FF,$B3,$9B,$19,$7C,$6F,$17,$24,$32,$72,$9F,$AB,$F5
	.dc.b	$0F,$8C,$FB,$B2,$FA,$2F,$B7,$A7,$F5,$46,$C3,$8D,$7C,$60,$7F,$B4
	.dc.b	$A7,$6F,$3E,$4E,$B1,$CA,$B8,$E1,$89,$95,$CB,$7F,$ED,$0F,$13,$F2
	.dc.b	$98,$77,$7F,$D0,$3F,$44,$FA,$2E,$BE,$4C,$9F,$F3,$BC,$E5,$E9,$A7
	.dc.b	$AE,$75,$1E,$77,$F2,$B3,$B8,$C1,$C0,$F6,$E8,$9E,$38,$FE,$2E,$B8
	.dc.b	$3F,$EE,$4E,$8F,$4B,$70,$FD,$1C,$61,$FF,$5A,$BF,$E9,$7C,$3B,$0F
	.dc.b	$C0,$1C,$DF,$51,$B1,$A6,$AB,$EC,$E6,$7F,$5F,$F9,$63,$E3,$A9,$CE
	.dc.b	$3F,$6B,$F1,$DF,$E3,$9A,$FF,$4E,$D7,$F9,$BD,$BE,$86,$BC,$8F,$17
	.dc.b	$B5,$27,$C4,$F6,$9F,$DE,$FD,$13,$EA,$3C,$3F,$05,$2F,$FD,$53,$39
	.dc.b	$7E,$67,$C4,$23,$E2,$47,$E2,$BF,$0A,$7C,$9F,$E5,$1B,$1C,$27,$6A
	.dc.b	$32,$FC,$3E,$3F,$F1,$5C,$63,$87,$F1,$6D,$8F,$CB,$E1,$E1,$7B,$61
	.dc.b	$2F,$E9,$D6,$D5,$6D,$FE,$4D,$BB,$EF,$7E,$5F,$AD,$27,$31,$F3,$8C
	.dc.b	$3E,$53,$F5,$53,$FE,$8F,$76,$74,$FA,$6F,$AA,$47,$BF,$7D,$79,$77
	.dc.b	$A9,$59,$D7,$FC,$A1,$39,$BF,$DE,$F0,$FF,$9D,$70,$FC,$7E,$0F,$87
	.dc.b	$FE,$02,$D8,$FC,$9C,$ED,$F4,$1F,$EB,$3F,$4F,$75,$23,$71,$D8,$C3
	.dc.b	$C2,$BA,$3C,$AB,$A6,$3E,$69,$FF,$3D,$4F,$CC,$F0,$BF,$2A,$7E,$5B
	.dc.b	$82,$C6,$48,$7C,$29,$FA,$D9,$FC,$3F,$83,$FE,$AF,$EC,$3E,$1F,$F8
	.dc.b	$CB,$7C,$4F,$D0,$CC,$C3,$F7,$D3,$96,$D7,$F1,$7C,$15,$FC,$9B,$E8
	.dc.b	$5D,$1F,$95,$FB,$19,$87,$EE,$A7,$2F,$E2,$A3,$2F,$CA,$E7,$47,$3A
	.dc.b	$CE,$FF,$96,$8B,$D1,$FE,$84,$C3,$9E,$1F,$89,$D3,$FD,$1B,$F2,$BE
	.dc.b	$EF,$60,$D3,$FB,$7F,$E9,$7F,$E2,$EC,$BF,$D2,$55,$C2,$FD,$F7,$C1
	.dc.b	$B7,$C3,$D4,$FE,$FF,$D8,$FD,$3D,$F3,$D8,$6D,$FE,$E9,$AF,$E7,$7D
	.dc.b	$26,$8F,$6B,$85,$CA,$E0,$F9,$A3,$FA,$8F,$E6,$7B,$9F,$E4,$7F,$9A
	.dc.b	$3E,$CB,$FA,$AB,$63,$F9,$3C,$ED,$FC,$AC,$65,$B3,$C0,$EB,$56,$3A
	.dc.b	$C5,$7F,$3B,$5C,$F8,$3F,$D7,$1A,$3F,$51,$7D,$E7,$C4,$6D,$FF,$1A
	.dc.b	$6B,$FB,$DF,$81,$60,$FA,$8E,$17,$2B,$83,$FE,$2C,$AD,$CF,$DB,$FA
	.dc.b	$39,$5F,$59,$FB,$7E,$7E,$23,$F0,$5C,$A9,$C6,$9F,$E1,$DE,$1F,$E8
	.dc.b	$15,$FD,$67,$89,$F2,$9F,$C5,$E3,$0D,$7E,$C3,$6B,$F6,$9E,$E7,$F8
	.dc.b	$E2,$31,$D5,$FD,$FA,$F9,$F9,$59,$3F,$A5,$D7,$F8,$FB,$F0,$CF,$CE
	.dc.b	$E3,$DB,$CB,$E4,$67,$BA,$46,$48,$7E,$26,$78,$BF,$C4,$3F,$17,$FF
	.dc.b	$15,$26,$93,$E7,$FF,$C3,$2E,$3D,$EA,$62,$3F,$C3,$BF,$51,$FE,$61
	.dc.b	$F9,$2F,$F7,$37,$77,$FB,$C2,$31,$83,$E2,$7F,$37,$18,$7F,$3B,$F5
	.dc.b	$3F,$A7,$9D,$57,$1F,$2D,$87,$49,$FF,$11,$E1,$C6,$B5,$39,$95,$F3
	.dc.b	$C3,$BE,$1C,$A7,$A9,$C9,$C4,$6A,$B9,$57,$1B,$EB,$7C,$06,$1E,$17
	.dc.b	$95,$70,$F1,$3F,$C6,$F1,$B9,$D3,$C3,$1E,$F7,$D3,$1F,$83,$7C,$1F
	.dc.b	$15,$71,$AD,$BD,$8F,$16,$3E,$57,$13,$C0,$97,$E2,$9D,$4F,$87,$1F
	.dc.b	$8D,$8F,$90,$F2,$CE,$9C,$61,$E0,$AE,$07,$83,$1F,$36,$1E,$92,$47
	.dc.b	$89,$F0,$46,$A3,$04,$39,$E3,$F0,$1F,$1D,$77,$BF,$0E,$3F,$2E,$72
	.dc.b	$C7,$E2,$39,$69,$2B,$E0,$BF,$E6,$8C,$7D,$97,$D4,$95,$C6,$B7,$5E
	.dc.b	$4C,$7D,$6F,$E2,$B1,$F5,$39,$3D,$58,$FC,$88,$CB,$E8,$8F,$14,$FC
	.dc.b	$27,$D0,$7E,$10,$F9,$65,$8F,$D2,$B7,$0F,$CB,$39,$F7,$5F,$25,$C6
	.dc.b	$3F,$43,$18,$71,$5F,$E6,$99,$D4,$7E,$5A,$74,$F3,$8F,$C0,$C7,$1C
	.dc.b	$7F,$A6,$FD,$4B,$EA,$5C,$FA,$B9,$87,$E6,$F0,$E1,$9F,$E7,$B0,$E1
	.dc.b	$7A,$C7,$F7,$B5,$AB,$F3,$E1,$FB,$5B,$DC,$73,$8E,$18,$7F,$07,$AE
	.dc.b	$12,$AD,$DF,$4B,$DC,$F4,$3F,$69,$3A,$BE,$53,$A9,$F8,$4F,$87,$1F
	.dc.b	$6C,$78,$49,$7D,$5E,$71,$A8,$C1,$2B,$AC,$EB,$1E,$D9,$7D,$18,$E7
	.dc.b	$3F,$16,$F9,$3C,$2F,$94,$FA,$5F,$A3,$3A,$8C,$4E,$C8,$FF,$B0,$B1
	.dc.b	$F7,$3A,$21,$EE,$3C,$87,$88,$CD,$F9,$6F,$BE,$1E,$8C,$BE,$7C,$E7
	.dc.b	$0F,$44,$E7,$2F,$95,$8F,$83,$1E,$D8,$FA,$9B,$7C,$1F,$01,$83,$CB
	.dc.b	$3C,$70,$ED,$1E,$4C,$7E,$FC,$E5,$BE,$07,$D6,$EA,$FF,$6B,$CF,$F8
	.dc.b	$1B,$3F,$F5,$DB,$FE,$A3,$6C,$ED,$FC,$A3,$6F,$E4,$35,$FC,$67,$B3
	.dc.b	$6B,$A4,$7F,$65,$F4,$B8,$75,$BE,$69,$F4,$98,$E9,$FF,$54,$E7,$FE
	.dc.b	$01,$9C,$BF,$E0,$38,$CE,$1F,$E3,$79,$CB,$D6,$78,$57,$69,$71,$D3
	.dc.b	$D1,$EA,$7F,$A4,$7E,$4B,$17,$DE,$6E,$1F,$E7,$59,$BB,$FF,$94,$BF
	.dc.b	$36,$FE,$0F,$7F,$E7,$06,$D7,$F9,$27,$F1,$3F,$1D,$6F,$9B,$E3,$7A
	.dc.b	$2B,$DC,$E7,$0F,$F0,$F5,$F0,$58,$2F,$86,$1F,$3F,$CC,$FF,$1E,$7B
	.dc.b	$37,$CD,$3F,$4F,$CC,$FF,$1C,$7D,$8E,$62,$65,$FE,$3F,$75,$E6,$A7
	.dc.b	$39,$FF,$A9,$B2,$7C,$0F,$DB,$6E,$1F,$EF,$EF,$94,$FF,$44,$BF,$F0
	.dc.b	$02,$33,$5C,$6B,$81,$F8,$3A,$DC,$66,$B9,$7A,$FA,$69,$CE,$BD,$DE
	.dc.b	$AF,$FC,$01,$8F,$F3,$B7,$A3,$F8,$8E,$1F,$44,$60,$98,$7F,$B5,$78
	.dc.b	$3F,$D0,$15,$BD,$39,$BF,$4B,$8D,$F4,$74,$E9,$F6,$D7,$1F,$B9,$17
	.dc.b	$1F,$B9,$37,$97,$F8,$EC,$37,$FF,$33,$5F,$C5,$E1,$FD,$17,$E3,$AF
	.dc.b	$8C,$F8,$51,$27,$47,$DE,$C3,$9D,$6E,$33,$87,$FB,$62,$30,$FF,$B6
	.dc.b	$B0,$EE,$E7,$0E,$8F,$FB,$2F,$F0,$AF,$C7,$FC,$03,$E4,$BF,$3F,$F0
	.dc.b	$29,$CC,$61,$B9,$78,$CF,$C8,$F7,$7F,$BD,$63,$38,$7F,$BE,$3A,$7F
	.dc.b	$AC,$D6,$E1,$5F,$92,$8F,$D6,$CB,$E8,$EC,$43,$84,$ED,$B3,$B6,$9F
	.dc.b	$C2,$FE,$C7,$FD,$0C,$F7,$B4,$77,$7F,$DF,$AE,$9B,$3C,$7D,$AE,$4D
	.dc.b	$AE,$55,$C7,$83,$ED,$BC,$5B,$3B,$BE,$75,$E6,$F0,$FD,$A7,$BB,$CA
	.dc.b	$B1,$92,$4F,$C7,$8C,$BF,$27,$C3,$F8,$BF,$15,$F1,$9B,$5F,$09,$B3
	.dc.b	$B8,$F8,$B3,$FC,$B7,$CE,$7C,$B5,$3F,$82,$78,$FF,$76,$BE,$34,$E9
	.dc.b	$B3,$B7,$EC,$77,$7F,$31,$E2,$C6,$3F,$A9,$C3,$92,$DB,$F3,$F8,$F3
	.dc.b	$8D,$57,$09,$FE,$FF,$E4,$BE,$1D,$F2,$7F,$CC,$7E,$07,$F9,$FC,$63
	.dc.b	$4F,$E7,$91,$FD,$1F,$EE,$BF,$90,$65,$FC,$81,$B8,$7F,$45,$C3,$5C
	.dc.b	$5F,$8C,$F8,$FF,$E7,$91,$71,$FE,$8C,$F3,$E8,$FE,$81,$D9,$67,$FA
	.dc.b	$53,$72,$F6,$47,$FA,$2B,$2F,$C8,$47,$E7,$7A,$B5,$B7,$F8,$6B,$AB
	.dc.b	$F9,$6C,$7E,$70,$FC,$EF,$C9,$5C,$3F,$29,$EC,$7F,$4E,$6F,$47,$F5
	.dc.b	$EF,$A3,$75,$7F,$65,$AD,$67,$E2,$3F,$9F,$B8,$CF,$F8,$6B,$70,$FF
	.dc.b	$6C,$5F,$27,$A9,$FE,$E5,$71,$5C,$CF,$F5,$DE,$1E,$1C,$3C,$0F,$85
	.dc.b	$95,$FB,$2F,$A8,$FD,$03,$FE,$AC,$C3,$9E,$3E,$2A,$DE,$1E,$7D,$18
	.dc.b	$70,$8F,$4C,$FF,$6C,$8C,$47,$FB,$32,$31,$DD,$CA,$FF,$0F,$F3,$EF
	.dc.b	$8F,$FF,$01,$B7,$3F,$F8,$00,$FF,$4C,$9D,$CE,$89,$3A,$F3,$6E,$3F
	.dc.b	$2F,$9D,$87,$1A,$FF,$80,$0F,$59,$D3,$95,$3C,$2F,$C7,$F0,$52,$FF
	.dc.b	$8B,$FC,$BF,$E1,$B2,$F4,$CE,$70,$E6,$FF,$83,$27,$55,$E0,$8C,$73
	.dc.b	$6E,$FF,$19,$5A,$BF,$0E,$4B,$FF,$12,$FC,$63,$E6,$69,$8D,$6E,$7F
	.dc.b	$CC,$75,$56,$E3,$0D,$C7,$E2,$3E,$F9,$DD,$F3,$AD,$B6,$B6,$F8,$BE
	.dc.b	$4E,$4F,$F3,$8F,$BD,$7D,$AE,$2F,$AD,$5D,$BF,$51,$5A,$BF,$1C,$FF
	.dc.b	$91,$6B,$FC,$3B,$E1,$7F,$87,$BE,$7D,$F0,$FF,$EF,$CF,$65,$F1,$DF
	.dc.b	$E8,$A9,$CD,$FF,$AF,$6F,$85,$EE,$B4,$FF,$AE,$63,$33,$FE,$2A,$BE
	.dc.b	$78,$E2,$3F,$C5,$A2,$71,$3A,$86,$1E,$07,$73,$AC,$3B,$C7,$8C,$DE
	.dc.b	$1E,$3C,$3B,$47,$B8,$C9,$E3,$9E,$2F,$43,$23,$0F,$16,$1E,$18,$E4
	.dc.b	$F8,$E3,$37,$EF,$9E,$31,$87,$C9,$3B,$AE,$F1,$CA,$BC,$CF,$80,$C1
	.dc.b	$E5,$BF,$4B,$F1,$4E,$B3,$C9,$C5,$75,$9D,$E1,$F1,$5B,$3C,$1B,$3B
	.dc.b	$8F,$01,$D4,$FF,$75,$CF,$FB,$B1,$F9,$06,$0F,$F7,$A7,$CE,$BE,$45
	.dc.b	$F4,$1F,$EF,$28,$C3,$63,$33,$C1,$AF,$E0,$04,$BE,$7F,$FA,$F4,$76
	.dc.b	$6B,$B8,$D6,$FD,$E7,$33,$F8,$A9,$CB,$F6,$BE,$83,$F6,$6F,$E3,$DA
	.dc.b	$FD,$E8,$DE,$5F,$8D,$6E,$BF,$94,$AD,$CF,$E4,$5B,$1D,$EF,$CF,$3C
	.dc.b	$5F,$4F,$C3,$6F,$97,$F3,$AD,$C7,$FD,$16,$EF,$2B,$87,$37,$CD,$5F
	.dc.b	$A4,$C3,$85,$F9,$F0,$EF,$7B,$8F,$75,$6E,$3D,$0D,$8F,$F4,$CE,$5E
	.dc.b	$C6,$D7,$27,$FD,$3B,$97,$B1,$B5,$C5,$FA,$B3,$B6,$E3,$E6,$61,$E5
	.dc.b	$7F,$79,$7D,$E1,$7D,$5B,$7D,$67,$47,$D8,$CB,$DE,$4C,$7D,$CF,$F1
	.dc.b	$0E,$5F,$C3,$F8,$3F,$89,$C3,$FD,$6B,$18,$70,$D7,$F1,$CD,$7F,$87
	.dc.b	$B9,$CB,$FC,$4F,$83,$D0,$FE,$27,$18,$6F,$0F,$69,$FF,$6C,$DF,$F3
	.dc.b	$F7,$FC,$F8,$DF,$17,$6F,$F5,$A8,$CB,$FD,$8D,$79,$3A,$C7,$F4,$35
	.dc.b	$EB,$AE,$0D,$8C,$C7,$B2,$B9,$BE,$A7,$AE,$3F,$E4,$0F,$BB,$7C,$CF
	.dc.b	$56,$D7,$94,$FE,$AD,$F4,$3F,$D6,$DB,$D9,$E3,$8C,$47,$F8,$37,$E4
	.dc.b	$FF,$98,$61,$FE,$4C,$9C,$CF,$CA,$6B,$F4,$9A,$FF,$7A,$6E,$47,$AC
	.dc.b	$F3,$EB,$F0,$1E,$1C,$3C,$5F,$37,$30,$F3,$BE,$5F,$9C,$C4,$7D,$CC
	.dc.b	$13,$F1,$DB,$1F,$82,$6B,$B8,$FA,$B8,$3C,$37,$C4,$E1,$B5,$E0,$3F
	.dc.b	$D3,$DF,$43,$F8,$07,$FC,$78,$DA,$FF,$68,$B7,$8F,$FD,$6F,$5F,$E7
	.dc.b	$0A,$E3,$CB,$FC,$AE,$B9,$47,$F2,$16,$BF,$D8,$9A,$83,$D5,$FF,$12
	.dc.b	$F9,$DD,$5F,$F1,$73,$70,$F1,$D7,$E4,$A7,$8F,$95,$6F,$FC,$C3,$1E
	.dc.b	$C8,$FC,$95,$FF,$B9,$63,$07,$E4,$E7,$6F,$EB,$DA,$FA,$67,$D8,$7E
	.dc.b	$6F,$EE,$3F,$C5,$11,$FE,$23,$6B,$F2,$23,$C9,$F4,$3F,$A2,$8F,$F6
	.dc.b	$53,$5F,$F3,$43,$71,$FF,$40,$7D,$17,$FB,$F9,$FF,$7F,$C6,$E7,$FA
	.dc.b	$53,$5F,$F0,$5B,$5F,$94,$DE,$37,$D1,$3E,$1F,$F5,$CD,$AF,$E1,$2D
	.dc.b	$7F,$84,$B5,$FE,$12,$BD,$EE,$24,$8E,$2F,$F6,$9B,$FE,$19,$1F,$BD
	.dc.b	$6B,$FC,$65,$AF,$AB,$B7,$C7,$F1,$72,$3C,$6D,$7F,$90,$7D,$0B,$E4
	.dc.b	$7E,$F3,$63,$FD,$E8,$D7,$FD,$39,$F0,$0F,$A2,$F9,$5C,$DF,$E3,$0A
	.dc.b	$D5,$7F,$73,$6B,$9B,$FF,$01,$46,$23,$F2,$8F,$5D,$D5,$FE,$3A,$FC
	.dc.b	$7F,$F7,$AD,$EC,$F2,$E1,$FC,$93,$D8,$7C,$2B,$BF,$F9,$3F,$C9,$3E
	.dc.b	$1D,$F0,$38,$6E,$DF,$A3,$D7,$FD,$85,$87,$18,$C6,$9F,$A7,$CB,$AE
	.dc.b	$5C,$71,$95,$A1,$3A,$71,$97,$87,$A7,$C0,$48,$E1,$3E,$8F,$5F,$11
	.dc.b	$FE,$C7,$AE,$0F,$2D,$3F,$4F,$A7,$9A,$BF,$65,$A7,$FA,$B6,$BC,$47
	.dc.b	$C2,$F6,$B2,$7E,$DF,$9F,$FD,$F1,$1A,$9E,$0F,$F5,$F6,$BF,$E2,$96
	.dc.b	$BF,$E6,$A6,$BF,$C2,$5E,$37,$F5,$A3,$1E,$67,$90,$F0,$BE,$E9,$FE
	.dc.b	$4F,$3B,$EB,$FC,$47,$A3,$FD,$5F,$9B,$FC,$81,$7B,$F1,$3E,$E6,$1C
	.dc.b	$38,$FF,$B9,$D6,$AB,$FD,$84,$DC,$7E,$CD,$FD,$DF,$0A,$57,$C9,$79
	.dc.b	$F9,$38,$8F,$9D,$C1,$FC,$87,$87,$F0,$D5,$A9,$FC,$FF,$CE,$FE,$88
	.dc.b	$8F,$CE,$FA,$37,$6E,$23,$F5,$33,$F9,$C2,$3F,$DC,$7B,$11,$FC,$A7
	.dc.b	$E8,$9F,$55,$4E,$5B,$7C,$BC,$FF,$31,$E0,$7E,$5C,$98,$FE,$33,$D8
	.dc.b	$7C,$EE,$23,$D7,$36,$F9,$56,$A7,$F2,$DE,$13,$E5,$5F,$39,$FE,$0C
	.dc.b	$F7,$1C,$4F,$5B,$0D,$C3,$8D,$F0,$6D,$EA,$A4,$60,$FC,$9F,$D2,$3E
	.dc.b	$2B,$FA,$43,$63,$31,$F0,$FB,$32,$FF,$A1,$67,$53,$FE,$33,$6B,$FE
	.dc.b	$7D,$6B,$FE,$BE,$77,$AF,$E4,$E1,$0F,$8E,$F0,$9D,$A3,$E1,$FC,$CB
	.dc.b	$E7,$3F,$25,$E7,$49,$CC,$7F,$8C,$38,$15,$B9,$FF,$7A,$77,$7F,$3B
	.dc.b	$CB,$F3,$8D,$EF,$FE,$DD,$7C,$30,$FE,$27,$C6,$F8,$7F,$F1,$B7,$CD
	.dc.b	$FF,$A5,$F1,$FD,$23,$70,$FE,$ED,$D1,$FE,$E6,$AE,$27,$0C,$E1,$F8
	.dc.b	$37,$FC,$81,$F4,$6F,$A3,$FF,$1F,$36,$7F,$CB,$2D,$7F,$CB,$2D,$7F
	.dc.b	$CB,$2E,$71,$F1,$F6,$F9,$8E,$B5,$DB,$C9,$97,$FC,$87,$C6,$61,$FD
	.dc.b	$D3,$0E,$33,$98,$FE,$EE,$F6,$AF,$29,$1F,$F2,$B7,$D1,$FF,$B8,$DF
	.dc.b	$F7,$CB,$5F,$F0,$C3,$53,$4F,$F2,$5E,$9C,$3E,$6B,$FC,$2D,$06,$F9
	.dc.b	$3F,$E1,$9F,$A0,$FC,$FF,$C7,$FF,$88,$FE,$27,$F3,$98,$A7,$2A,$33
	.dc.b	$18,$77,$F4,$5F,$AD,$9D,$77,$FF,$8B,$5E,$5E,$0B,$0E,$D3,$9C,$39
	.dc.b	$36,$75,$5C,$8F,$F1,$B5,$F5,$C3,$C3,$3A,$6E,$1D,$2B,$FC,$79,$87
	.dc.b	$47,$8F,$89,$E3,$AD,$C7,$53,$CB,$5C,$5A,$67,$0F,$06,$3E,$59,$D6
	.dc.b	$1D,$6B,$77,$F0,$B0,$ED,$7B,$9E,$D5,$B3,$86,$1D,$23,$CF,$8F,$AE
	.dc.b	$B8,$61,$E6,$3A,$4F,$82,$74,$F0,$C7,$07,$A6,$FA,$47,$92,$F9,$91
	.dc.b	$F8,$67,$03,$CB,$5D,$1E,$06,$8C,$D6,$EF,$C9,$5F,$76,$FB,$E1,$ED
	.dc.b	$C3,$93,$EE,$C7,$E1,$E3,$E7,$39,$93,$0F,$6B,$70,$F3,$3E,$21,$87
	.dc.b	$6C,$3A,$93,$1F,$81,$87,$6A,$E1,$8F,$8E,$73,$1E,$9C,$3A,$CF,$96
	.dc.b	$30,$79,$71,$F6,$BF,$05,$FC,$3B,$F9,$B3,$F0,$A5,$8F,$D2,$B7,$0F
	.dc.b	$CC,$B9,$F4,$5F,$41,$FE,$D1,$A4,$62,$7F,$DD,$D1,$89,$D4,$E6,$79
	.dc.b	$47,$E4,$23,$85,$FE,$DB,$E8,$5F,$45,$FE,$11,$8F,$DA,$CE,$B2,$FD
	.dc.b	$73,$70,$FD,$2B,$55,$F2,$E1,$ED,$8C,$9F,$CC,$D6,$EB,$51,$89,$D6
	.dc.b	$1E,$07,$F5,$53,$A8,$C4,$EC,$8F,$A7,$1F,$46,$3E,$6B,$E6,$DA,$FD
	.dc.b	$B5,$F3,$8C,$5F,$42,$3F,$C3,$E3,$53,$EE,$AE,$6D,$39,$DF,$81,$C5
	.dc.b	$F2,$6D,$F3,$9D,$47,$F2,$3C,$3E,$46,$5F,$5E,$3D,$75,$BC,$BB,$C6
	.dc.b	$F1,$B9,$79,$23,$07,$03,$B9,$FE,$D3,$75,$7D,$72,$F2,$4F,$07,$E4
	.dc.b	$5F,$69,$D6,$1F,$10,$F1,$57,$53,$83,$97,$B6,$5E,$C9,$E3,$1E,$23
	.dc.b	$66,$AB,$B6,$5F,$00,$FC,$15,$71,$71,$97,$BA,$30,$FE,$0B,$B3,$FC
	.dc.b	$C9,$CB,$FE,$13,$D3,$FC,$5F,$87,$F9,$D9,$FF,$08,$37,$3D,$BF,$6E
	.dc.b	$BF,$C3,$59,$58,$FD,$93,$F1,$E3,$2D,$8C,$F3,$3E,$6B,$FD,$3F,$18
	.dc.b	$6C,$FE,$EE,$75,$EB,$72,$6D,$6E,$3C,$72,$9C,$AC,$BF,$C2,$FE,$C6
	.dc.b	$A7,$31,$FE,$7D,$C3,$FC,$97,$87,$F8,$17,$BB,$FD,$98,$FF,$81,$E3
	.dc.b	$38,$73,$83,$DE,$76,$E4,$F1,$4F,$07,$FB,$75,$FB,$F8,$BD,$53,$99
	.dc.b	$E7,$39,$BF,$93,$DC,$FA,$0F,$E2,$B1,$E5,$F8,$6F,$7C,$E7,$87,$D9
	.dc.b	$9A,$73,$7C,$31,$F2,$62,$AD,$63,$FE,$C0,$AD,$69,$FE,$3C,$C3,$E5
	.dc.b	$F8,$5F,$E4,$AF,$5F,$E4,$7C,$D7,$FB,$7B,$0F,$F2,$9C,$61,$DF,$8C
	.dc.b	$8C,$46,$8F,$F1,$84,$F1,$D3,$E3,$1F,$CC,$3E,$4F,$EC,$BF,$CC,$B8
	.dc.b	$AE,$E9,$B1,$87,$F1,$73,$C6,$FF,$1E,$D3,$F2,$35,$C6,$32,$D8,$F0
	.dc.b	$B7,$4F,$F3,$97,$CE,$FF,$A9,$AB,$FD,$AB,$39,$7F,$CC,$BE,$7B,$CA
	.dc.b	$FC,$7F,$A0,$FD,$01,$F3,$DF,$22,$B5,$7F,$EE,$FA,$D4,$E6,$F8,$DF
	.dc.b	$0A,$FF,$4B,$C6,$27,$EE,$FB,$2F,$8F,$A7,$DC,$F1,$ED,$FF,$BB,$E2
	.dc.b	$ED,$FE,$F3,$E6,$FF,$25,$4E,$4F,$F7,$7E,$CF,$A0,$F9,$EF,$F9,$F6
	.dc.b	$33,$18,$F3,$B2,$FC,$E9,$E1,$3F,$39,$B7,$FB,$E6,$BF,$DC,$5C,$F0
	.dc.b	$FC,$CA,$FB,$BF,$40,$F9,$EE,$B8,$7E,$AA,$BF,$D9,$5A,$BE,$4B,$E0
	.dc.b	$9F,$BC,$F8,$37,$C1,$3E,$83,$38,$71,$3F,$D2,$37,$C5,$C6,$AF,$05
	.dc.b	$39,$8F,$DB,$61,$D2,$F9,$C6,$3B,$3F,$2D,$9F,$FC,$01,$EC,$7D,$BF
	.dc.b	$8E,$73,$9F,$EE,$27,$2F,$83,$1F,$D1,$3F,$A8,$AE,$0D,$3C,$B8,$7E
	.dc.b	$22,$FF,$43,$F0,$9F,$C2,$B1,$FF,$62,$73,$F2,$AF,$D3,$CE,$6F,$FC
	.dc.b	$F1,$39,$9F,$F1,$03,$5C,$DF,$E9,$A3,$11,$FC,$61,$E9,$7D,$A7,$F6
	.dc.b	$31,$F8,$39,$FD,$AB,$AA,$F9,$BF,$1D,$F9,$43,$E1,$78,$F6,$3E,$7E
	.dc.b	$FF,$3A,$71,$7F,$2E,$AE,$2B,$FE,$53,$A7,$F0,$7F,$7F,$F3,$FA,$FC
	.dc.b	$D6,$CF,$CB,$D1,$18,$C7,$EB,$37,$0F,$F7,$57,$9F,$FE,$E9,$BF,$9B
	.dc.b	$CA,$FA,$1F,$E1,$B3,$9C,$3F,$88,$CE,$6B,$E9,$CE,$67,$FC,$94,$D7
	.dc.b	$FD,$29,$EE,$7F,$10,$8C,$47,$FA,$CE,$3A,$9F,$CA,$3C,$AF,$CB,$57
	.dc.b	$48,$FE,$A1,$0F,$7B,$FA,$09,$27,$FB,$03,$FC,$73,$E8,$7F,$B0,$DF
	.dc.b	$F9,$6D,$AF,$FA,$DD,$AF,$F5,$26,$F9,$9F,$DA,$F3,$FE,$9A,$4B,$FE
	.dc.b	$D8,$FF,$50,$EB,$91,$FD,$BD,$B7,$FD,$E2,$DF,$3F,$E3,$25,$7F,$2E
	.dc.b	$8C,$B6,$3F,$9C,$C6,$27,$F5,$5D,$98,$9F,$E7,$BE,$AF,$F6,$AE,$0F
	.dc.b	$C1,$57,$F7,$7F,$AC,$7C,$26,$BD,$1F,$EE,$D1,$8C,$FF,$4F,$87,$F6
	.dc.b	$7B,$FE,$E2,$D7,$FD,$B1,$5F,$D1,$23,$07,$79,$FF,$6F,$36,$7F,$C0
	.dc.b	$DF,$45,$FE,$7E,$9E,$B5,$E6,$F5,$FF,$C0,$D3,$BA,$FF,$72,$46,$27
	.dc.b	$55,$FD,$42,$F9,$57,$F8,$81,$AF,$FA,$4D,$AF,$8A,$B8,$BF,$C4,$BE
	.dc.b	$8B,$F6,$19,$FC,$48,$EF,$7F,$C6,$B8,$5F,$44,$F8,$DF,$F0,$FD,$BF
	.dc.b	$02,$7D,$B7,$F8,$76,$B5,$87,$F2,$2A,$D4,$E6,$FF,$C5,$55,$AC,$7F
	.dc.b	$C5,$CD,$7F,$C2,$ED,$7F,$C2,$AE,$EB,$EC,$E3,$E4,$AF,$07,$A7,$D7
	.dc.b	$1F,$45,$BE,$0B,$C9,$F2,$1E,$56,$CE,$9F,$19,$1F,$F2,$2C,$F1,$6C
	.dc.b	$71,$7F,$97,$7C,$C7,$E3,$F2,$F1,$D6,$F3,$FF,$0F,$FC,$F3,$E4,$FF
	.dc.b	$C8,$1F,$55,$F1,$B1,$F9,$93,$97,$E7,$46,$6B,$8D,$7F,$93,$EB,$51
	.dc.b	$FE,$53,$6B,$FE,$3C,$6B,$FE,$3A,$77,$87,$C6,$6D,$E7,$4F,$F2,$94
	.dc.b	$62,$BF,$CB,$0B,$4C,$FC,$4E,$9B,$7E,$87,$C8,$47,$FC,$D7,$3C,$5F
	.dc.b	$F1,$1D,$7F,$95,$BC,$1F,$F2,$BF,$C7,$BE,$23,$FC,$B3,$2F,$69,$F1
	.dc.b	$9F,$B4,$AD,$4F,$F2,$20,$E1,$E1,$1B,$AF,$F3,$FD,$FF,$99,$3E,$8B
	.dc.b	$58,$74,$C7,$15,$AC,$15,$AC,$3F,$81,$BD,$63,$D0,$6A,$7C,$F7,$CE
	.dc.b	$30,$79,$67,$95,$F6,$8C,$57,$0B,$EA,$F8,$6B,$73,$C8,$F8,$78,$71
	.dc.b	$7D,$B7,$5F,$1C,$F0,$3C,$CA,$78,$61,$E5,$9C,$C6,$30,$F0,$D7,$CB
	.dc.b	$C7,$C2,$69,$F9,$B8,$F9,$F1,$FA,$EF,$9C,$C4,$79,$CF,$1B,$8C,$39
	.dc.b	$1E,$57,$84,$7F,$A5,$B1,$F2,$47,$D1,$8F,$F3,$3E,$3F,$86,$7D,$47
	.dc.b	$E2,$1F,$F3,$5F,$D0,$3E,$45,$F4,$78,$6D,$7F,$AF,$A7,$53,$93,$8F
	.dc.b	$D1,$FF,$AC,$67,$4D,$7F,$22,$25,$19,$8F,$CA,$C6,$38,$B9,$9F,$57
	.dc.b	$E8,$DC,$1F,$9A,$D7,$F2,$91,$8B,$FB,$57,$C5,$FB,$CD,$9D,$D7,$1D
	.dc.b	$3F,$47,$7C,$6F,$78,$FA,$6F,$F4,$18,$73,$DF,$8D,$7D,$63,$FD,$87
	.dc.b	$7C,$63,$77,$F0,$5B,$97,$A7,$2F,$05,$F0,$6B,$F0,$EB,$C1,$97,$C2
	.dc.b	$6C,$6C,$E4,$7B,$9C,$BC,$6B,$2F,$E7,$70,$F6,$65,$E2,$8F,$6C,$EA
	.dc.b	$3D,$71,$82,$38,$25,$F8,$E7,$C0,$71,$7E,$4E,$1E,$AC,$BE,$74,$7F
	.dc.b	$B7,$1F,$F1,$CD,$FF,$C0,$06,$BF,$EF,$D6,$BF,$EE,$66,$BC,$7C,$DF
	.dc.b	$E2,$F7,$FE,$08,$F4,$FF,$C9,$5F,$75,$8E,$2F,$F7,$9C,$62,$FF,$B4
	.dc.b	$61,$DA,$FF,$AB,$FD,$23,$E6,$F8,$53,$BD,$39,$8F,$EE,$0D,$7F,$DD
	.dc.b	$37,$FE,$95,$73,$5F,$ED,$0A,$E2,$DA,$E1,$3F,$F0,$01,$AF,$FB,$F9
	.dc.b	$E9,$7F,$DA,$AD,$EB,$79,$30,$FD,$23,$6B,$79,$30,$F4,$7D,$1F,$F4
	.dc.b	$77,$FC,$0C,$D7,$FD,$FE,$DC,$7F,$BB,$BF,$D8,$E3,$7D,$3E,$98,$E1
	.dc.b	$1F,$D1,$31,$46,$23,$C1,$18,$9E,$07,$CA,$D3,$FD,$5D,$B5,$7E,$86
	.dc.b	$3C,$30,$F1,$3F,$EB,$FE,$F7,$D1,$FD,$2C,$D1,$8C,$FF,$C3,$0D,$7D
	.dc.b	$7E,$D6,$E3,$11,$F8,$46,$BD,$6B,$F8,$6B,$4B,$5D,$A3,$E7,$60,$F8
	.dc.b	$CA,$75,$8F,$A5,$BF,$FA,$26,$FF,$C6,$4D,$57,$FE,$22,$8C,$3F,$E3
	.dc.b	$9F,$A3,$79,$7C,$BD,$3F,$C3,$57,$1B,$F9,$CD,$7E,$6B,$5F,$CA,$3D
	.dc.b	$35,$FC,$EF,$D9,$7D,$99,$F0,$C6,$3C,$4E,$D1,$FE,$47,$FA,$4F,$F1
	.dc.b	$83,$7C,$8F,$C6,$46,$23,$89,$F8,$BC,$DE,$6E,$AB,$F3,$95,$BA,$FC
	.dc.b	$23,$5F,$C1,$B5,$FD,$53,$D7,$5F,$F2,$2F,$DE,$3C,$FF,$CE,$36,$BC
	.dc.b	$87,$FB,$8B,$E8,$DF,$13,$FA,$1F,$A4,$7C,$65,$3F,$6F,$4A,$DB,$FE
	.dc.b	$65,$9D,$BF,$80,$EC,$46,$23,$FA,$7C,$62,$BF,$71,$39,$9F,$F1,$1F
	.dc.b	$C9,$3E,$19,$F0,$BF,$C1,$A3,$0D,$8C,$CF,$86,$FC,$B8,$FE,$82,$3F
	.dc.b	$6C,$D7,$D1,$DD,$ED,$F8,$2F,$DF,$7C,$F7,$F8,$B2,$7D,$0D,$7C,$ED
	.dc.b	$7F,$68,$D7,$F8,$FF,$C4,$3E,$23,$F6,$4D,$8F,$F7,$9B,$5F,$BF,$B7
	.dc.b	$C4,$FA,$2F,$97,$F4,$9F,$8B,$E4,$FF,$48,$37,$AB,$FC,$8A,$DB,$FE
	.dc.b	$21,$F4,$6F,$95,$E3,$D9,$FE,$4E,$AF,$CA,$57,$F7,$C6,$BC,$A3,$47
	.dc.b	$CB,$FB,$27,$CC,$D3,$9C,$FA,$3E,$3B,$D7,$F5,$72,$30,$43,$F3,$3E
	.dc.b	$53,$81,$3B,$9C,$C2,$BE,$2F,$81,$F1,$BE,$32,$57,$C6,$8A,$F0,$8B
	.dc.b	$C1,$FD,$3E,$FF,$4B,$3F,$CC,$63,$FD,$6D,$F2,$1F,$E6,$36,$FA,$5F
	.dc.b	$96,$8C,$71,$7F,$9B,$9F,$F5,$C5,$71,$8C,$4E,$C9,$19,$3F,$9F,$4F
	.dc.b	$07,$F6,$E1,$EA,$EA,$38,$BA,$8E,$15,$FE,$B6,$E6,$FF,$1D,$4F,$FB
	.dc.b	$2D,$AF,$E1,$1B,$1F,$EF,$4F,$90,$ED,$E7,$7E,$5B,$DA,$FC,$6F,$E1
	.dc.b	$93,$0E,$51,$F9,$0F,$33,$F2,$3F,$59,$27,$F2,$31,$97,$F2,$71,$97
	.dc.b	$EA,$C7,$EF,$3E,$9F,$FE,$00,$C7,$C1,$0E,$1F,$82,$DC,$BF,$DC,$DD
	.dc.b	$4F,$A6,$E5,$F1,$9F,$EA,$0A,$E1,$7F,$E4,$E6,$BF,$EA,$06,$BF,$DA
	.dc.b	$DE,$9C,$3F,$EA,$4F,$D0,$7F,$3D,$3E,$36,$BE,$2A,$FD,$7F,$85,$FB
	.dc.b	$0F,$1A,$4F,$EC,$22,$BF,$92,$8B,$9F,$E2,$E7,$FC,$23,$F0,$5F,$D1
	.dc.b	$BE,$02,$ED,$FC,$06,$99,$FF,$9E,$63,$0F,$F8,$07,$C9,$7C,$E3,$E8
	.dc.b	$38,$E1,$FC,$5F,$AF,$FB,$46,$3C,$B5,$60,$9C,$E3,$D6,$FF,$96,$FD
	.dc.b	$46,$6F,$E9,$35,$FC,$8B,$70,$E9,$08,$7C,$AF,$2C,$3F,$DA,$31,$C6
	.dc.b	$7F,$CF,$ED,$7F,$DA,$2D,$7F,$C3,$8F,$5C,$33,$E7,$F9,$0E,$93,$F9
	.dc.b	$9F,$37,$F9,$49,$1F,$D4,$66,$FA,$0F,$E1,$B1,$B3,$3E,$EB,$E0,$7F
	.dc.b	$9E,$36,$34,$7F,$26,$FA,$1F,$F4,$AE,$DF,$27,$CE,$46,$34,$FF,$11
	.dc.b	$B6,$9E,$47,$FB,$AB,$E7,$7F,$C7,$53,$98,$FE,$2D,$97,$F3,$63,$35
	.dc.b	$AE,$CF,$F2,$55,$7E,$6E,$3F,$AD,$35,$FE,$B2,$D7,$FC,$6C,$E7,$2F
	.dc.b	$37,$D4,$7F,$71,$CB,$FC,$6E,$DA,$FF,$2F,$56,$E7,$FA,$4F,$DB,$FF
	.dc.b	$48,$24,$FF,$11,$D1,$7C,$73,$FE,$5D,$5A,$AF,$E5,$AF,$F9,$5E,$7B
	.dc.b	$43,$EA,$5F,$45,$FE,$32,$8F,$F2,$C4,$61,$C3,$5F,$F1,$BB,$5F,$F6
	.dc.b	$2B,$9C,$FF,$AA,$FD,$87,$F5,$83,$FC,$07,$19,$CF,$FC,$C9,$4F,$A8
	.dc.b	$FF,$18,$FB,$9F,$E6,$7F,$8F,$FD,$34,$FF,$9A,$67,$2F,$FB,$E7,$B3
	.dc.b	$8C,$62,$77,$2F,$AA,$7D,$57,$69,$CD,$6E,$F8,$F8,$1F,$E1,$D9,$CB
	.dc.b	$63,$CB,$7C,$B0,$F2,$47,$0C,$6C,$79,$6B,$83,$63,$2F,$8F,$1F,$15
	.dc.b	$71,$8F,$51,$B2,$9F,$0E,$B9,$BC,$4E,$78,$FA,$5E,$18,$FA,$30,$EF
	.dc.b	$8F,$9E,$79,$D6,$DE,$93,$CA,$FA,$D6,$EF,$95,$70,$9F,$0D,$F1,$AD
	.dc.b	$61,$E9,$BD,$91,$F8,$F8,$FB,$B0,$F2,$C6,$C8,$FF,$A8,$EB,$9B,$A2
	.dc.b	$1E,$63,$E1,$BA,$71,$5D,$DD,$C6,$6F,$DC,$4C,$3D,$B8,$78,$EB,$8E
	.dc.b	$1E,$26,$9E,$38,$D5,$71,$C7,$E3,$1F,$32,$33,$7F,$10,$91,$9A,$D9
	.dc.b	$F2,$5B,$8F,$E5,$31,$F3,$57,$0A,$D1,$EE,$BE,$B7,$C3,$1F,$35,$F0
	.dc.b	$8F,$DA,$63,$FA,$2B,$FD,$49,$F1,$CA,$E5,$FD,$2F,$A2,$FA,$8E,$2D
	.dc.b	$8F,$F5,$7C,$FF,$A2,$A7,$24,$8F,$C4,$CE,$F3,$FD,$6F,$D4,$BE,$AB
	.dc.b	$FA,$B4,$7E,$B2,$75,$9F,$83,$0F,$CB,$C6,$30,$FF,$71,$61,$C7,$6C
	.dc.b	$D4,$9E,$EF,$EA,$27,$51,$82,$61,$DE,$75,$8F,$8B,$1F,$05,$69,$B5
	.dc.b	$E8,$9D,$4F,$F0,$BC,$7C,$8F,$F1,$16,$E5,$F0,$E3,$F6,$78,$F8,$AF
	.dc.b	$C1,$1D,$4E,$58,$F8,$B2,$F5,$4E,$89,$97,$F1,$EB,$E4,$F9,$09,$7C
	.dc.b	$8E,$F9,$79,$9F,$26,$1D,$27,$70,$AE,$6F,$1B,$EE,$D9,$F0,$5F,$3C
	.dc.b	$7D,$95,$AA,$EA,$78,$31,$EC,$F1,$21,$DE,$B9,$65,$F1,$89,$97,$C1
	.dc.b	$9F,$88,$74,$BE,$D9,$79,$A3,$19,$7A,$B2,$F1,$BE,$6C,$BC,$33,$A3
	.dc.b	$C3,$5C,$6B,$71,$E4,$6B,$E5,$9E,$33,$FE,$2D,$7F,$CA,$9F,$90,$59
	.dc.b	$FF,$39,$CD,$E2,$FF,$11,$CD,$AF,$AD,$87,$F8,$9A,$EC,$7F,$63,$BE
	.dc.b	$51,$FB,$06,$A3,$FC,$17,$F2,$CF,$17,$FC,$75,$8F,$F8,$EF,$A1,$39
	.dc.b	$C3,$FB,$C7,$D2,$FF,$A7,$DF,$F4,$54,$EA,$F8,$C7,$58,$C3,$AA,$E5
	.dc.b	$87,$FB,$1A,$F9,$3F,$EA,$E8,$CE,$1F,$E2,$3E,$BF,$F6,$15,$F8,$3D
	.dc.b	$8E,$B7,$FE,$0E,$F7,$AF,$6B,$2E,$78,$F1,$9F,$89,$5B,$C3,$FC,$3F
	.dc.b	$18,$BF,$F1,$7B,$71,$7C,$6F,$6D,$3E,$0F,$B9,$DA,$73,$87,$AE,$31
	.dc.b	$52,$3E,$1F,$D5,$7F,$19,$7F,$C3,$D5,$C3,$2F,$F7,$2F,$C1,$7F,$9C
	.dc.b	$27,$FC,$97,$F4,$EF,$95,$F5,$CE,$69,$B7,$F9,$B7,$C9,$5E,$F0,$FF
	.dc.b	$15,$5E,$EB,$FC,$94,$D8,$FF,$3D,$79,$2F,$85,$EB,$1F,$2B,$1F,$E5
	.dc.b	$DC,$17,$27,$D2,$67,$B3,$E8,$D7,$FB,$12,$73,$7F,$EA,$99,$CC,$FF
	.dc.b	$3C,$6B,$FE,$26,$6B,$F8,$F6,$BE,$08,$E5,$7D,$FE,$73,$C0,$FD,$EF
	.dc.b	$4D,$F5,$6F,$5B,$F0,$39,$3F,$51,$78,$5B,$AD,$65,$FE,$76,$9C,$BF
	.dc.b	$ED,$FC,$39,$E1,$B3,$9F,$AC,$F4,$B1,$5F,$97,$F0,$7F,$05,$DF,$FA
	.dc.b	$FD,$FF,$CE,$3F,$08,$F5,$33,$5F,$87,$8C,$5F,$13,$F4,$AE,$2B,$F3
	.dc.b	$D3,$F4,$7E,$B1,$EA,$7D,$3F,$19,$E5,$66,$30,$DC,$7F,$CB,$D5,$C1
	.dc.b	$B3,$F9,$EC,$13,$97,$FB,$79,$88,$E4,$FF,$BF,$A1,$F5,$98,$EC,$FE
	.dc.b	$01,$9B,$EA,$7E,$0F,$D8,$FB,$5B,$97,$B1,$B8,$FE,$5A,$3F,$80,$57
	.dc.b	$17,$F8,$05,$7E,$73,$EB,$7E,$5E,$36,$7F,$82,$36,$FF,$4B,$3F,$BE
	.dc.b	$F8,$6F,$CC,$56,$72,$FF,$61,$FC,$37,$E5,$1B,$D7,$F9,$BE,$6D,$E3
	.dc.b	$E3,$3F,$8F,$56,$EF,$FC,$F2,$D8,$CD,$71,$EF,$FD,$24,$E6,$4D,$31
	.dc.b	$7D,$B0,$F0,$E1,$FC,$77,$C3,$FC,$39,$F8,$3F,$97,$B1,$F8,$26,$E1
	.dc.b	$E5,$F2,$5C,$09,$CE,$1F,$4B,$6F,$C8,$FD,$2F,$E3,$5B,$8F,$E3,$1B
	.dc.b	$8F,$3C,$11,$8B,$FF,$74,$F8,$CF,$99,$FD,$4F,$86,$F8,$9F,$E0,$53
	.dc.b	$9C,$BF,$DB,$B3,$97,$FA,$35,$F2,$BF,$C7,$56,$F2,$FC,$6B,$5F,$F1
	.dc.b	$EB,$5F,$F5,$3B,$B9,$E5,$5D,$0F,$F4,$F7,$C0,$F8,$AF,$F5,$FF,$45
	.dc.b	$27,$FA,$7F,$C1,$DC,$7F,$94,$D7,$F3,$9F,$2F,$B4,$3D,$04,$61,$B3
	.dc.b	$FA,$5F,$AC,$FE,$8B,$87,$FA,$E9,$B7,$C7,$D2,$F3,$4E,$7D,$3F,$E0
	.dc.b	$3F,$17,$23,$FB,$0F,$D6,$DB,$FE,$B4,$FF,$4B,$B3,$49,$3F,$ED,$5C
	.dc.b	$3F,$94,$D7,$08,$F8,$4F,$F7,$38,$CE,$E9,$FE,$ED,$9F,$F5,$DB,$FE
	.dc.b	$BF,$F5,$AF,$89,$FE,$C3,$F5,$BE,$AA,$4E,$6F,$FD,$DD,$39,$8C,$65
	.dc.b	$FE,$72,$AD,$DF,$F6,$39,$CC,$FF,$B7,$A7,$37,$FC,$29,$B1,$FA,$D6
	.dc.b	$E3,$E7,$8F,$F0,$67,$8C,$F8,$0F,$3E,$3E,$67,$FB,$67,$DE,$FF,$59
	.dc.b	$AE,$59,$EF,$2F,$F0,$5C,$E7,$1E,$99,$2B,$57,$E9,$6B,$E7,$6B,$F2
	.dc.b	$1A,$FE,$15,$E3,$3D,$1B,$7F,$EF,$CC,$BF,$A3,$D7,$93,$1F,$F0,$7D
	.dc.b	$BE,$B5,$F5,$BF,$E1,$0E,$D7,$C2,$EA,$73,$97,$DC,$D1,$5A,$8F,$BF
	.dc.b	$87,$3C,$3F,$C2,$AD,$8F,$E6,$ED,$CB,$FD,$4E,$D7,$FC,$70,$7F,$81
	.dc.b	$30,$EE,$FF,$1F,$8E,$07,$C8,$B7,$D6,$F9,$70,$F4,$7C,$67,$F8,$77
	.dc.b	$E0,$1F,$2B,$8C,$FF,$C7,$71,$87,$D5,$82,$BF,$C5,$0D,$8F,$C2,$37
	.dc.b	$2F,$C6,$35,$FC,$53,$9C,$BC,$95,$F2,$7E,$83,$C1,$5F,$8B,$FA,$DF
	.dc.b	$5D,$7E,$03,$C7,$7D,$CF,$F8,$FA,$73,$97,$E9,$27,$2F,$CF,$8F,$87
	.dc.b	$7F,$99,$6B,$FA,$A6,$BF,$C1,$DB,$C0,$7A,$C6,$76,$FF,$26,$DB,$E6
	.dc.b	$FC,$51,$8F,$AE,$FF,$23,$7C,$49,$C6,$5F,$E6,$48,$C3,$F8,$3B,$B8
	.dc.b	$7E,$C5,$AF,$EB,$DA,$FB,$6B,$FB,$E3,$FE,$77,$AF,$F3,$C7,$D6,$7D
	.dc.b	$88,$FF,$3D,$7B,$BF,$E7,$6C,$33,$C7,$FE,$00,$BB,$5A,$F6,$FF,$6D
	.dc.b	$87,$18,$FC,$5C,$E9,$61,$CD,$4E,$DF,$11,$8B,$FC,$9C,$EE,$75,$2A
	.dc.b	$E9,$39,$AF,$24,$F1,$3C,$98,$78,$5E,$B7,$E5,$AE,$11,$FA,$03,$A1
	.dc.b	$C4,$F3,$E1,$EC,$BE,$B5,$C6,$31,$3C,$70,$F1,$C6,$23,$E0,$9C,$EF
	.dc.b	$BB,$8A,$E5,$87,$96,$3C,$3C,$98,$7F,$D5,$18,$FB,$23,$07,$C2,$C7
	.dc.b	$BD,$71,$6D,$7A,$4F,$7D,$7C,$AC,$7C,$91,$83,$E2,$9B,$3E,$09,$FE
	.dc.b	$3C,$29,$F3,$0C,$4E,$7E,$AD,$E1,$3E,$BB,$0D,$AF,$F2,$2C,$EA,$72
	.dc.b	$71,$F2,$3F,$D2,$33,$A6,$BF,$4E,$51,$88,$FE,$2B,$18,$E5,$E4,$7E
	.dc.b	$37,$EB,$BF,$D3,$D3,$78,$7F,$8D,$EF,$F9,$3B,$E1,$3F,$75,$B1,$F9
	.dc.b	$7A,$E3,$3F,$81,$6C,$71,$C3,$CE,$DC,$3F,$DB,$DB,$F1,$C7,$F9,$4B
	.dc.b	$70,$CE,$DE,$8C,$3D,$F3,$C6,$7B,$57,$9A,$BF,$98,$E3,$E0,$78,$61
	.dc.b	$FA,$9C,$BC,$F1,$AA,$E1,$7F,$EE,$FB,$E0,$FF,$BC,$6B,$6D,$C7,$D3
	.dc.b	$7C,$63,$E3,$1D,$63,$2E,$F0,$CB,$E6,$34,$C7,$04,$CB,$D8,$4A,$ED
	.dc.b	$19,$7D,$6F,$ED,$79,$58,$7F,$A7,$E3,$FC,$04,$D7,$FC,$14,$D7,$FC
	.dc.b	$10,$D7,$FC,$32,$EE,$FF,$BD,$BF,$E8,$D8,$FE,$CC,$D7,$FC,$18,$EF
	.dc.b	$83,$FC,$43,$F5,$AF,$AB,$FE,$C3,$7F,$E0,$48,$C5,$FF,$6F,$6B,$FE
	.dc.b	$81,$F6,$1F,$66,$FA,$3E,$38,$F4,$C7,$F5,$59,$B1,$E3,$8F,$F8,$85
	.dc.b	$AE,$CF,$F1,$13,$DB,$2F,$F0,$AF,$C1,$7F,$88,$67,$FC,$48,$D8,$F3
	.dc.b	$5F,$82,$7F,$CA,$DF,$5A,$FB,$3C,$62,$8C,$4F,$F8,$25,$AF,$AF,$E4
	.dc.b	$3F,$C1,$EE,$6B,$D3,$3A,$8F,$F1,$C3,$5F,$F2,$4B,$5F,$0C,$78,$23
	.dc.b	$FD,$3B,$1D,$FD,$CC,$4F,$93,$C4,$FD,$7E,$5F,$E3,$26,$EB,$FE,$54
	.dc.b	$6D,$B8,$1F,$5F,$F6,$63,$FC,$E9,$3F,$B8,$7F,$C9,$8D,$79,$73,$FD
	.dc.b	$28,$C4,$70,$8F,$17,$8B,$23,$FC,$59,$3E,$0D,$BF,$D5,$1E,$77,$CA
	.dc.b	$F6,$FF,$CE,$35,$FE,$31,$6B,$F8,$3C,$6E,$9F,$E4,$C9,$CC,$FF,$97
	.dc.b	$DA,$FF,$9C,$1A,$FF,$A3,$5B,$7E,$08,$FF,$3C,$8F,$AE,$7C,$67,$DF
	.dc.b	$8C,$3E,$5C,$7E,$FF,$D7,$7D,$CC,$7F,$CB,$AD,$DF,$FC,$FE,$DB,$F9
	.dc.b	$9F,$62,$F8,$BE,$3C,$5F,$98,$AF,$B9,$A1,$A4,$8F,$16,$8F,$5E,$9C
	.dc.b	$B1,$FF,$48,$F1,$BE,$A1,$EF,$7E,$53,$EB,$9F,$0B,$4F,$CD,$56,$E7
	.dc.b	$17,$FE,$85,$AD,$B6,$BF,$5B,$18,$8F,$EF,$43,$EB,$5F,$4B,$F1,$DB
	.dc.b	$1F,$31,$AF,$FA,$3E,$79,$71,$7F,$A3,$FE,$4D,$F5,$CF,$A5,$FD,$74
	.dc.b	$61,$B1,$9B,$FF,$57,$CF,$0F,$A4,$FD,$8F,$DA,$BC,$CF,$E0,$6D,$AF
	.dc.b	$F5,$03,$5F,$F5,$43,$5F,$F0,$A6,$7F,$E9,$F7,$FD,$77,$F1,$6F,$45
	.dc.b	$F4,$BF,$C0,$63,$0F,$BF,$DD,$FC,$37,$D9,$BE,$C2,$93,$0F,$3E,$5F
	.dc.b	$EF,$97,$F1,$F3,$B7,$FC,$04,$D7,$E6,$B5,$FF,$14,$35,$FF,$44,$BB
	.dc.b	$E0,$FE,$31,$F5,$8F,$A9,$FC,$C3,$6B,$BE,$1F,$80,$E5,$7C,$7F,$AF
	.dc.b	$6F,$E5,$95,$F0,$1A,$FF,$31,$9D,$B9,$F3,$F8,$11,$FC,$8C,$69,$FF
	.dc.b	$66,$69,$FB,$D2,$79,$3E,$1F,$8B,$FA,$EF,$87,$E9,$E5,$7F,$AB,$EB
	.dc.b	$F8,$9F,$33,$E8,$3F,$C2,$B3,$F8,$0F,$9C,$7C,$BF,$E1,$9B,$1F,$84
	.dc.b	$9D,$AE,$2F,$C7,$46,$1F,$C8,$4F,$09,$D7,$90,$58,$27,$55,$CB,$D9
	.dc.b	$F0,$63,$2F,$C8,$FE,$63,$1E,$D8,$FF,$C0,$0C,$BF,$C7,$35,$ED,$F7
	.dc.b	$3F,$2E,$7D,$3F,$AF,$B1,$ED,$F9,$DF,$61,$87,$CA,$3A,$FF,$29,$F8
	.dc.b	$0F,$D9,$FC,$02,$4F,$D9,$8A,$F0,$9F,$CD,$4E,$36,$FF,$7F,$47,$29
	.dc.b	$FE,$43,$F5,$FE,$67,$B8,$71,$3F,$D9,$5B,$A7,$FB,$6B,$EB,$DE,$CF
	.dc.b	$F2,$79,$E5,$7F,$E7,$F6,$BF,$EE,$06,$BF,$3D,$AF,$7D,$7F,$98,$7D
	.dc.b	$7B,$E5,$E9,$D6,$3F,$7F,$AC,$AF,$C4,$3F,$9D,$F9,$8C,$47,$C1,$EC
	.dc.b	$7D,$67,$E7,$CC,$EC,$DB,$FD,$63,$A3,$E7,$FF,$BF,$7C,$4F,$C1,$6C
	.dc.b	$67,$1F,$C7,$71,$E2,$7F,$01,$F5,$52,$7F,$15,$8F,$57,$FD,$67,$7C
	.dc.b	$5F,$B7,$83,$83,$F8,$24,$61,$B5,$C6,$75,$F3,$7F,$E1,$1C,$DF,$AE
	.dc.b	$FF,$82,$B0,$E1,$3F,$E8,$BA,$DC,$7F,$36,$6B,$FE,$B1,$6B,$FE,$C0
	.dc.b	$C3,$BD,$FE,$86,$3C,$9E,$47,$D2,$3E,$1C,$EA,$1E,$C3,$AB,$FA,$E4
	.dc.b	$7E,$E7,$D9,$FE,$66,$7F,$0F,$F9,$EB,$1F,$C3,$7E,$93,$F8,$81,$87
	.dc.b	$F3,$B6,$FC,$2B,$FD,$8D,$BF,$C9,$8C,$4F,$F8,$BA,$73,$A7,$ED,$FD
	.dc.b	$7F,$EB,$BF,$65,$9A,$FE,$39,$9C,$7E,$9D,$7E,$F7,$0E,$71,$83,$D7
	.dc.b	$87,$F2,$77,$FD,$AD,$1E,$3F,$3B,$FC,$9B,$39,$9F,$8D,$39,$7F,$B1
	.dc.b	$B5,$F5,$FB,$1F,$CA,$B3,$F5,$37,$0F,$15,$F7,$9C,$F9,$9F,$CC,$FC
	.dc.b	$C9,$87,$97,$1F,$EF,$73,$FE,$4E,$C7,$94,$FF,$03,$EB,$FE,$D7,$97
	.dc.b	$F1,$9F,$0B,$FA,$C6,$3B,$C3,$FC,$35,$F6,$2F,$93,$FF,$68,$62,$8C
	.dc.b	$3F,$ED,$36,$BF,$DF,$9A,$FF,$BE,$5C,$E9,$FE,$67,$7D,$58,$FE,$25
	.dc.b	$B3,$FC,$AD,$DC,$7E,$37,$EC,$1C,$EF,$9C,$FE,$61,$F2,$8F,$03,$53
	.dc.b	$9C,$53,$97,$F9,$B5,$F2,$BE,$37,$F9,$9F,$B1,$7A,$1F,$9B,$C3,$86
	.dc.b	$0C,$38,$3D,$2F,$BE,$1C,$A3,$AE,$3E,$1B,$E3,$8F,$49,$D1,$CE,$75
	.dc.b	$7C,$A3,$A1,$CA,$F7,$5C,$63,$27,$18,$E7,$87,$93,$0F,$3B,$6F,$D9
	.dc.b	$8F,$AD,$ED,$3A,$C7,$C3,$87,$73,$AE,$3D,$E7,$58,$F6,$AD,$D7,$84
	.dc.b	$DE,$1D,$5E,$55,$C1,$F3,$E3,$E3,$7F,$B3,$61,$D7,$1F,$19,$23,$8B
	.dc.b	$E8,$7C,$18,$C7,$9E,$3E,$2A,$E5,$1C,$08,$79,$70,$FA,$38,$F9,$63
	.dc.b	$18,$77,$69,$E4,$8D,$63,$E7,$8C,$92,$FB,$B8,$8D,$1B,$8C,$B8,$C3
	.dc.b	$E3,$B9,$AE,$B1,$82,$63,$F7,$A3,$2F,$BF,$1F,$94,$7A,$B0,$C6,$3E
	.dc.b	$9C,$3A,$46,$0F,$82,$E1,$F8,$F5,$FE,$70,$7E,$C1,$FC,$DC,$A7,$E4
	.dc.b	$4C,$47,$FA,$6B,$EB,$78,$36,$3F,$CF,$2D,$77,$F0,$E9,$DB,$63,$F0
	.dc.b	$2D,$C3,$F5,$7F,$63,$FD,$B3,$5F,$CC,$CE,$72,$FD,$F3,$6F,$F6,$ED
	.dc.b	$55,$C5,$EB,$87,$22,$3F,$A6,$8D,$61,$FA,$4A,$E5,$8F,$AE,$31,$7F
	.dc.b	$B4,$AD,$E1,$E2,$8E,$73,$AC,$3A,$78,$DF,$DD,$63,$04,$AE,$13,$AC
	.dc.b	$B8,$C6,$27,$53,$FE,$47,$C3,$BB,$EE,$C3,$6D,$39,$1E,$E2,$4F,$27
	.dc.b	$E1,$E1,$E3,$BE,$C4,$8E,$53,$F4,$27,$51,$89,$D3,$FC,$77,$2F,$9D
	.dc.b	$87,$92,$31,$87,$91,$A7,$93,$0F,$16,$5F,$2F,$0E,$77,$C6,$31,$5D
	.dc.b	$C8,$7A,$32,$E9,$1A,$8E,$B7,$E7,$9C,$E3,$F6,$0F,$2E,$5E,$AA,$DE
	.dc.b	$3E,$AC,$39,$63,$E1,$BE,$19,$7A,$63,$0F,$DF,$F4,$3F,$C5,$DC,$3F
	.dc.b	$E2,$FC,$7F,$1B,$E3,$E3,$4F,$F0,$E6,$1F,$E8,$87,$FD,$FE,$FF,$8F
	.dc.b	$DB,$7C,$6B,$6D,$7E,$F3,$5F,$CC,$B9,$E9,$7E,$A1,$FD,$B6,$BF,$D5
	.dc.b	$5E,$C7,$86,$FF,$C4,$F3,$97,$FC,$A9,$19,$AE,$3C,$DB,$E9,$FF,$68
	.dc.b	$4E,$6F,$86,$9F,$E7,$8E,$AF,$F6,$76,$5F,$E0,$5E,$CF,$F3,$8B,$FE
	.dc.b	$07,$AD,$BF,$E0,$EF,$3B,$96,$5D,$EB,$7A,$7F,$87,$9A,$FF,$87,$1A
	.dc.b	$F1,$AF,$F0,$EC,$7F,$BD,$9F,$F7,$2F,$D9,$7F,$A9,$DF,$17,$C0,$7A
	.dc.b	$63,$15,$FE,$3C,$AE,$0D,$9F,$89,$E1,$BB,$DF,$55,$E5,$9F,$89,$F7
	.dc.b	$4F,$87,$E9,$CF,$86,$E9,$FE,$25,$9D,$36,$3A,$E0,$F8,$4E,$13,$FE
	.dc.b	$3F,$F3,$3F,$CF,$DE,$A7,$D5,$D7,$FD,$7F,$D7,$F5,$E7,$35,$FE,$45
	.dc.b	$9C,$BC,$7A,$FF,$D2,$8F,$F9,$B5,$B1,$FE,$6D,$6B,$FE,$6B,$9E,$6E
	.dc.b	$6F,$B4,$F0,$E6,$FF,$43,$5B,$F4,$9F,$F0,$CB,$DE,$B7,$EC,$7E,$2F
	.dc.b	$DC,$FF,$72,$6B,$8A,$FF,$70,$37,$2F,$F4,$6E,$4B,$E1,$ED,$3C,$CD
	.dc.b	$69,$FE,$6D,$E1,$7C,$3F,$D9,$EC,$FA,$2D,$D3,$E7,$38,$7F,$DF,$FF
	.dc.b	$0D,$74,$FF,$63,$FD,$7F,$E5,$1F,$F6,$4F,$33,$CE,$F9,$D3,$9C,$FF
	.dc.b	$6D,$39,$C3,$E7,$4F,$1B,$FE,$B8,$D7,$FD,$28,$D7,$FD,$EC,$D4,$F5
	.dc.b	$8D,$F8,$1F,$77,$F6,$37,$C7,$F8,$A3,$F2,$DE,$8C,$AF,$CD,$F8,$37
	.dc.b	$1F,$F7,$47,$83,$FA,$A7,$FD,$E0,$AB,$F3,$D3,$BD,$3E,$57,$27,$D5
	.dc.b	$F8,$1F,$9D,$8F,$E5,$63,$1C,$9F,$9C,$64,$7F,$A4,$7C,$1F,$D1,$75
	.dc.b	$63,$4F,$F7,$FF,$53,$17,$A9,$D4,$FB,$7E,$7D,$C7,$FD,$35,$4E,$AF
	.dc.b	$C4,$E1,$CB,$0E,$35,$A9,$CE,$BF,$ED,$5B,$ED,$8F,$4A,$E7,$7C,$30
	.dc.b	$F1,$E1,$C2,$50,$BE,$31,$FB,$3F,$53,$F8,$17,$DB,$7E,$E9,$B1,$92
	.dc.b	$53,$2F,$F7,$36,$BF,$C0,$FD,$9E,$15,$FC,$3B,$E7,$3F,$86,$1F,$BC
	.dc.b	$E6,$B1,$F2,$7C,$0F,$BA,$EC,$C4,$6B,$1E,$D1,$FE,$DD,$FA,$ED,$C7
	.dc.b	$E6,$3C,$1F,$E0,$51,$8C,$FF,$A5,$75,$FF,$45,$AE,$27,$C0,$9E,$2F
	.dc.b	$F8,$C5,$AF,$5C,$3F,$A3,$46,$23,$84,$76,$3F,$75,$ED,$3E,$C7,$19
	.dc.b	$7F,$0F,$FA,$E7,$8F,$FC,$AF,$E5,$FF,$8B,$46,$27,$F9,$8C,$7F,$49
	.dc.b	$F0,$BF,$D5,$51,$FE,$D2,$6E,$5F,$ED,$A6,$BF,$ED,$86,$BF,$EC,$7E
	.dc.b	$F7,$D3,$7F,$39,$3F,$9D,$7D,$2D,$BF,$27,$99,$E2,$3F,$AC,$7D,$7B
	.dc.b	$9B,$13,$FE,$D7,$8C,$3F,$EE,$46,$E3,$FE,$EC,$6B,$FE,$EA,$6B,$FE
	.dc.b	$FB,$E5,$7C,$FF,$6B,$FE,$76,$DB,$FF,$72,$7D,$7B,$E0,$7F,$9D,$7B
	.dc.b	$DF,$C6,$A7,$31,$8F,$33,$FB,$75,$6E,$7F,$DE,$ED,$CB,$FD,$FE,$D7
	.dc.b	$FD,$FA,$D7,$F3,$FE,$C2,$38,$75,$2B,$31,$87,$F4,$BD,$AF,$96,$70
	.dc.b	$F8,$B3,$FF,$00,$7D,$6F,$DB,$CF,$F1,$A7,$FB,$8F,$E0,$73,$D3,$EC
	.dc.b	$6E,$5F,$E7,$F9,$6A,$BE,$51,$FB,$43,$2E,$1F,$4B,$5C,$DF,$47,$F3
	.dc.b	$B9,$BE,$6F,$FC,$E9,$B7,$B6,$77,$8F,$F0,$08,$C6,$2F,$BE,$F6,$CE
	.dc.b	$4F,$8E,$FE,$31,$C5,$7A,$23,$6F,$EB,$27,$97,$3F,$F8,$8E,$73,$3E
	.dc.b	$B9,$ED,$B7,$F8,$BE,$73,$3D,$A3,$07,$13,$FC,$1B,$F3,$CF,$4F,$18
	.dc.b	$F8,$B0,$7C,$FF,$F5,$1C,$3C,$4B,$B7,$FC,$68,$E7,$4F,$7B,$9B,$E4
	.dc.b	$F1,$AF,$33,$5F,$C3,$B5,$FA,$ED,$BF,$0F,$6F,$63,$CF,$54,$F3,$7D
	.dc.b	$53,$D6,$FF,$23,$3F,$E2,$CC,$7E,$8B,$99,$DE,$3A,$AF,$F2,$A4,$E6
	.dc.b	$3F,$D8,$D8,$23,$11,$F3,$5A,$FE,$79,$E1,$7D,$AB,$FC,$6F,$EF,$7F
	.dc.b	$B1,$70,$F2,$C7,$F9,$67,$E5,$3E,$34,$7E,$99,$FB,$B5,$F8,$67,$38
	.dc.b	$7E,$61,$CC,$EF,$2F,$F3,$54,$7C,$52,$3F,$E6,$F9,$FE,$9F,$F0,$FF
	.dc.b	$D4,$3E,$9D,$E5,$FF,$8B,$72,$FC,$B4,$62,$BF,$DC,$95,$BA,$D6,$36
	.dc.b	$72,$2B,$82,$F7,$7E,$44,$62,$3C,$26,$EB,$9B,$4E,$71,$CA,$78,$CF
	.dc.b	$ED,$EF,$C1,$5D,$70,$F3,$CE,$63,$91,$9C,$36,$F8,$30,$E9,$5C,$23
	.dc.b	$BD,$78,$A7,$CD,$0B,$EF,$87,$8A,$36,$7F,$78,$70,$FF,$07,$8D,$DF
	.dc.b	$58,$C0,$F9,$CF,$0E,$3F,$E0,$07,$BD,$5C,$7E,$1C,$62,$BD,$78,$76
	.dc.b	$7D,$53,$E7,$3E,$06,$3D,$23,$E0,$9A,$3C,$D8,$F8,$DF,$A2,$7F,$21
	.dc.b	$3D,$C5,$29,$9C,$35,$F4,$EF,$94,$7D,$4E,$1B,$5F,$CA,$E7,$53,$93
	.dc.b	$8F,$CC,$7F,$85,$67,$4D,$7F,$97,$09,$46,$63,$F9,$94,$63,$83,$99
	.dc.b	$F8,$2F,$A9,$70,$FF,$37,$D3,$F2,$31,$8B,$FC,$3D,$F1,$7F,$0A,$D9
	.dc.b	$DD,$71,$D7,$FA,$05,$F1,$BD,$E3,$E9,$BF,$CE,$E1,$CF,$7E,$35,$F6
	.dc.b	$4F,$F1,$DD,$F1,$8D,$DF,$C1,$6E,$5E,$9C,$BC,$17,$C1,$AF,$C3,$AF
	.dc.b	$06,$5F,$09,$B1,$B3,$91,$EE,$72,$F1,$AC,$BF,$A4,$C3,$D9,$97,$8A
	.dc.b	$3D,$B3,$A8,$F5,$C6,$08,$E0,$97,$E3,$9F,$01,$F0,$65,$5F,$06,$7E
	.dc.b	$5E,$1F,$DA,$F1,$F0,$1F,$C3,$79,$3F,$01,$E6,$FE,$4D,$B8,$7F,$75
	.dc.b	$6B,$FC,$C3,$4F,$F1,$F3,$96,$BF,$D8,$7E,$A1,$F7,$9F,$DE,$74,$FE
	.dc.b	$DB,$18,$9F,$EB,$15,$C2,$7F,$B8,$F9,$3F,$DD,$67,$FB,$53,$63,$3C
	.dc.b	$BC,$CF,$F0,$CD,$BE,$A7,$17,$FE,$0A,$BE,$4F,$F6,$F6,$BF,$E2,$06
	.dc.b	$BB,$AF,$81,$1D,$DF,$F4,$1F,$C6,$7F,$AC,$BD,$7E,$AD,$AF,$EE,$15
	.dc.b	$B7,$FC,$33,$F5,$2F,$A9,$F1,$EC,$8C,$4F,$F7,$76,$BF,$E0,$75,$F4
	.dc.b	$FC,$1C,$C7,$F8,$1D,$AF,$F8,$E1,$AF,$82,$BE,$C4,$74,$D7,$FC,$3C
	.dc.b	$3C,$2F,$8D,$5F,$E1,$88,$C6,$BD,$EF,$CF,$E0,$FF,$82,$AB,$FC,$18
	.dc.b	$D7,$FC,$38,$DA,$FF,$13,$BF,$B5,$6F,$83,$F8,$98,$E1,$3F,$E1,$F6
	.dc.b	$BF,$E5,$06,$BF,$E4,$07,$A6,$FF,$E3,$3C,$7E,$1C,$F9,$DA,$F8,$AB
	.dc.b	$EF,$AF,$A7,$FF,$46,$BF,$E3,$43,$1E,$4E,$A3,$31,$FE,$35,$FA,$9F
	.dc.b	$F7,$4B,$B8,$CB,$FC,$89,$AF,$F9,$45,$AF,$FB,$19,$AF,$63,$FC,$65
	.dc.b	$F5,$2F,$AC,$FC,$0B,$63,$B5,$F8,$FD,$79,$1E,$17,$7C,$38,$8F,$F3
	.dc.b	$26,$09,$FE,$88,$D8,$FE,$90,$D7,$71,$FA,$2F,$80,$7D,$CF,$C0,$E0
	.dc.b	$FF,$72,$37,$7F,$65,$7C,$AF,$AA,$7A,$D4,$98,$FF,$93,$7E,$AD,$ED
	.dc.b	$62,$3F,$CF,$15,$C1,$FE,$B4,$D7,$35,$FC,$2F,$1F,$EE,$11,$FC,$2D
	.dc.b	$AF,$F8,$E9,$AF,$F9,$95,$AF,$FA,$5B,$DE,$AF,$FA,$19,$BD,$26,$9A
	.dc.b	$85,$BD,$6F,$ED,$8D,$8F,$3C,$60,$FE,$41,$E6,$BE,$8A,$92,$7F,$DA
	.dc.b	$4F,$F7,$5F,$A9,$7C,$77,$E4,$9B,$3F,$91,$6B,$FE,$A2,$FA,$A7,$C8
	.dc.b	$6E,$B9,$12,$7F,$C0,$55,$C1,$FE,$E6,$F2,$E3,$FE,$49,$F4,$8F,$B2
	.dc.b	$FD,$0B,$6B,$C1,$7F,$C5,$BD,$3C,$47,$FB,$FB,$C5,$7C,$23,$6F,$E3
	.dc.b	$6D,$9F,$F4,$DC,$FF,$9D,$71,$E4,$6A,$A4,$EA,$FA,$E5,$FE,$AB,$B7
	.dc.b	$D9,$FE,$06,$30,$DE,$9F,$E5,$6D,$B5,$7F,$A0,$FE,$3F,$F9,$6C,$E6
	.dc.b	$39,$F3,$E0,$F6,$4F,$F3,$EF,$5D,$FA,$0B,$E4,$F0,$DC,$7F,$D6,$B3
	.dc.b	$A9,$FD,$27,$C8,$3E,$7B,$71,$FE,$31,$9D,$57,$FB,$4D,$AF,$F8,$91
	.dc.b	$AF,$FB,$4D,$CE,$BF,$EC,$BF,$01,$F5,$0F,$B3,$FD,$93,$6F,$C7,$9F
	.dc.b	$95,$5F,$1C,$DF,$47,$ED,$EC,$FD,$2C,$FF,$B5,$F3,$FC,$D7,$AD,$EA
	.dc.b	$A9,$87,$BA,$F7,$C7,$FE,$63,$6C,$35,$65,$DA,$3E,$FF,$3B,$EC,$7F
	.dc.b	$06,$48,$FB,$DF,$5B,$5F,$C5,$4E,$DE,$33,$F8,$F9,$FE,$67,$F5,$4F
	.dc.b	$99,$A4,$AF,$E6,$9D,$BF,$E5,$97,$F9,$B4,$ED,$FF,$31,$35,$FF,$29
	.dc.b	$87,$0F,$F9,$59,$A8,$8F,$1E,$0F,$F0,$53,$F5,$23,$BF,$17,$8D,$F5
	.dc.b	$FD,$97,$E8,$88,$FE,$77,$EA,$2B,$A3,$E3,$C7,$EB,$63,$F9,$77,$CD
	.dc.b	$7F,$07,$9F,$F7,$97,$91,$F7,$71,$71,$79,$7C,$8F,$E6,$7D,$3C,$3C
	.dc.b	$69,$87,$08,$FF,$03,$E1,$C6,$FD,$9C,$8C,$BB,$5F,$E2,$27,$E0,$3F
	.dc.b	$17,$97,$E8,$91,$FD,$0E,$EF,$93,$C4,$7C,$C8,$FE,$44,$D9,$FF,$48
	.dc.b	$46,$23,$E7,$47,$F4,$BF,$90,$FD,$44,$7C,$FD,$BF,$41,$18,$8F,$21
	.dc.b	$F4,$A7,$F8,$97,$CD,$FF,$88,$BE,$EB,$FD,$1B,$7F,$12,$9C,$09,$CE
	.dc.b	$7F,$E3,$0F,$11,$87,$2B,$F0,$4F,$CB,$AD,$C7,$26,$F8,$6B,$E3,$3F
	.dc.b	$C5,$F8,$7F,$8F,$FA,$92,$30,$43,$F3,$AE,$5F,$E5,$13,$F4,$DF,$3F
	.dc.b	$8C,$DB,$FA,$39,$23,$FC,$CD,$19,$6C,$7F,$33,$F3,$A4,$7A,$8E,$71
	.dc.b	$FE,$46,$97,$8B,$49,$3F,$CB,$EA,$BF,$EB,$6D,$BF,$9D,$4F,$F5,$82
	.dc.b	$47,$FB,$2F,$EA,$FF,$A5,$3F,$86,$6B,$FA,$76,$A6,$9F,$D6,$3E,$2B
	.dc.b	$F8,$34,$3E,$AF,$FC,$02,$FE,$BE,$33,$18,$8F,$F7,$B7,$87,$DF,$CC
	.dc.b	$FF,$79,$3F,$96,$FA,$BF,$EE,$5F,$31,$C6,$31,$CE,$8C,$5F,$EE,$EB
	.dc.b	$6F,$43,$F7,$D8,$7C,$67,$62,$37,$8F,$A1,$B8,$76,$BD,$DF,$1B,$E8
	.dc.b	$71,$C3,$C4,$E2,$F8,$BE,$38,$CE,$1F,$03,$3F,$6C,$E7,$0F,$1D,$F0
	.dc.b	$8E,$91,$E0,$9D,$C6,$0E,$D7,$D3,$1F,$34,$79,$31,$F3,$63,$EB,$AD
	.dc.b	$4F,$73,$8E,$1D,$5F,$0C,$69,$ED,$3C,$67,$51,$C0,$91,$97,$C1,$87
	.dc.b	$9B,$1F,$51,$2F,$A6,$1E,$4C,$7C,$F8,$71,$23,$FC,$52,$35,$87,$9E
	.dc.b	$72,$D9,$D6,$1D,$EF,$D1,$87,$7C,$7C,$F8,$70,$C3,$D7,$87,$80,$E2
	.dc.b	$FC,$09,$8F,$B2,$32,$D3,$C0,$43,$F3,$98,$78,$67,$8D,$77,$C7,$D3
	.dc.b	$5E,$03,$86,$1F,$76,$32,$F9,$E3,$B1,$D0,$F1,$9F,$CD,$8F,$C0,$15
	.dc.b	$FC,$EB,$71,$FC,$8B,$9F,$61,$F4,$7F,$D8,$A3,$F3,$D1,$87,$FB,$2D
	.dc.b	$23,$11,$C6,$78,$4E,$5F,$CA,$CF,$1C,$3F,$7B,$F5,$6F,$07,$F9,$A4
	.dc.b	$7E,$A6,$75,$AF,$EE,$5B,$8F,$ED,$5A,$AF,$97,$4F,$6C,$64,$FE,$AE
	.dc.b	$B7,$5A,$8C,$4E,$B0,$F0,$3F,$BC,$9D,$46,$27,$64,$7D,$38,$FA,$31
	.dc.b	$F3,$5F,$36,$D7,$EF,$AF,$9C,$62,$FA,$11,$FE,$D7,$1A,$9F,$75,$73
	.dc.b	$69,$CE,$FC,$0E,$2F,$93,$6F,$9C,$EA,$3E,$CD,$73,$69,$CF,$1C,$65
	.dc.b	$F5,$E3,$DD,$7C,$A7,$39,$7D,$0A,$D6,$3F,$0A,$73,$87,$5C,$BD,$58
	.dc.b	$F6,$AF,$2B,$71,$E8,$F6,$3D,$04,$3E,$E6,$5E,$9C,$7B,$CF,$C9,$24
	.dc.b	$FF,$8C,$6F,$BC,$EA,$39,$35,$E7,$8F,$99,$F0,$18,$2C,$BE,$83,$E4
	.dc.b	$2F,$B2,$B1,$86,$D7,$E9,$67,$83,$67,$FC,$1D,$5F,$B1,$9C,$BF,$DA
	.dc.b	$7E,$B3,$FC,$0D,$5F,$E4,$EE,$44,$62,$75,$D1,$C9,$FF,$1D,$46,$5F
	.dc.b	$F1,$8C,$F0,$9F,$F0,$E7,$77,$F8,$76,$31,$7F,$EA,$7F,$5B,$2F,$2A
	.dc.b	$B7,$FE,$22,$C1,$F4,$B8,$EE,$FF,$0D,$61,$AC,$7C,$B3,$9C,$3F,$D9
	.dc.b	$78,$FF,$82,$E7,$2F,$AF,$0E,$6D,$AF,$56,$3D,$DB,$8F,$FB,$12,$3C
	.dc.b	$1E,$6B,$E5,$2C,$61,$B7,$FE,$88,$6C,$7F,$A1,$1E,$3C,$9E,$7F,$07
	.dc.b	$95,$7F,$A3,$A3,$13,$A8,$F4,$4E,$AB,$FD,$1B,$5A,$9F,$F1,$34,$7F
	.dc.b	$A6,$79,$BE,$3F,$8D,$FE,$D9,$FB,$2F,$93,$39,$AF,$F1,$64,$E5,$E2
	.dc.b	$FF,$25,$7F,$95,$36,$3F,$D1,$AD,$7F,$9B,$35,$FF,$36,$BB,$E1,$FC
	.dc.b	$5B,$67,$FD,$21,$E9,$7E,$5F,$CE,$F1,$CE,$6F,$FD,$61,$E4,$BC,$2F
	.dc.b	$B5,$F6,$5F,$E8,$4C,$7C,$F3,$97,$ED,$46,$6B,$FC,$D7,$7F,$E8,$1B
	.dc.b	$E3,$5F,$0A,$F8,$57,$F9,$C3,$91,$87,$1C,$FF,$DE,$9C,$2F,$88,$FB
	.dc.b	$A7,$DE,$E6,$B9,$7F,$B0,$A2,$E3,$FE,$BE,$9B,$7F,$AA,$E1,$78,$BF
	.dc.b	$41,$7D,$B5,$8C,$37,$1F,$EB,$13,$C1,$B3,$F9,$2C,$3F,$AE,$B6,$75
	.dc.b	$1F,$E8,$A3,$33,$F8,$2F,$A6,$FC,$FC,$66,$77,$87,$BB,$EB,$7F,$32
	.dc.b	$7E,$6B,$C7,$B8,$7F,$B7,$A2,$DF,$FB,$6A,$6D,$7F,$BC,$7C,$BF,$D9
	.dc.b	$F2,$BE,$E5,$F5,$FF,$90,$F6,$15,$A9,$CE,$7F,$90,$9C,$BF,$E7,$DB
	.dc.b	$E5,$7C,$6F,$83,$DB,$DC,$7D,$47,$E7,$3D,$7F,$F7,$96,$1C,$67,$2F
	.dc.b	$E7,$23,$3A,$FF,$AB,$B1,$56,$A3,$FC,$1D,$18,$BF,$DF,$B6,$FF,$DD
	.dc.b	$6D,$9D,$D7,$1D,$3F,$D6,$93,$97,$FD,$6D,$D4,$D3,$FD,$67,$87,$97
	.dc.b	$1F,$E2,$DE,$FB,$E4,$29,$FC,$2B,$EA,$AD,$7F,$A7,$FC,$35,$6E,$30
	.dc.b	$DD,$7F,$D9,$32,$EA,$5F,$29,$CC,$3E,$03,$F9,$37,$E6,$1F,$9F,$C5
	.dc.b	$4C,$39,$56,$A7,$F8,$EB,$5F,$25,$F3,$7C,$D7,$CE,$F9,$F4,$7F,$48
	.dc.b	$DF,$FA,$5B,$F0,$A3,$F2,$CE,$6B,$FD,$E6,$E6,$77,$E5,$7E,$7E,$3F
	.dc.b	$A8,$D7,$F9,$0D,$AF,$08,$D1,$FC,$72,$3A,$47,$B3,$EA,$FE,$B5,$EC
	.dc.b	$FE,$81,$E4,$BD,$7F,$E7,$BE,$17,$F5,$A6,$C6,$67,$FA,$24,$7E,$C7
	.dc.b	$EA,$FF,$8F,$61,$F2,$1A,$9A,$6A,$FF,$80,$FA,$BD,$BC,$0F,$F5,$C6
	.dc.b	$1F,$EE,$08,$C1,$E2,$F1,$1E,$DD,$24,$FF,$6F,$8C,$BF,$E5,$DC,$1F
	.dc.b	$17,$E5,$24,$65,$B9,$7F,$05,$9D,$CF,$F6,$1F,$AB,$E7,$39,$AF,$E8
	.dc.b	$B8,$BE,$1B,$FA,$4D,$3E,$67,$31,$86,$E1,$FE,$6F,$AE,$0D,$9F,$E3
	.dc.b	$11,$8C,$3F,$CF,$87,$FB,$66,$38,$BF,$CC,$FE,$AD,$EF,$63,$97,$FC
	.dc.b	$27,$6F,$6F,$FB,$E5,$3E,$27,$CF,$18,$6E,$5E,$38,$F0,$91,$C1,$FD
	.dc.b	$9A,$BA,$3F,$E0,$3F,$AA,$FE,$A5,$97,$FB,$B6,$BF,$C1,$5F,$4E,$FA
	.dc.b	$8F,$F0,$8E,$6F,$B7,$7D,$4E,$A7,$31,$8C,$BF,$C1,$D7,$C5,$FF,$14
	.dc.b	$37,$1F,$F4,$F7,$3A,$31,$1F,$E2,$66,$BF,$B6,$8D,$5F,$7A,$EB,$85
	.dc.b	$AF,$EB,$9F,$30,$F9,$AF,$1F,$B9,$FE,$23,$8C,$5F,$F0,$0D,$3F,$C6
	.dc.b	$3A,$7F,$88,$30,$E3,$87,$E9,$2F,$71,$E6,$6B,$FE,$3A,$6B,$E5,$77
	.dc.b	$5F,$E3,$B6,$DE,$7A,$BC,$57,$FE,$35,$FA,$8F,$E2,$33,$FC,$7F,$D3
	.dc.b	$FE,$04,$FB,$E3,$38,$FF,$1E,$9C,$AD,$93,$A8,$FF,$26,$B5,$E4,$7E
	.dc.b	$F9,$C3,$C5,$EB,$97,$F9,$45,$B7,$E1,$F5,$7C,$D5,$FE,$4B,$FA,$2F
	.dc.b	$B8,$FF,$82,$6B,$E4,$5F,$F9,$66,$72,$F3,$DB,$86,$FF,$C3,$A7,$8D
	.dc.b	$FF,$98,$9A,$FD,$D6,$BF,$EE,$1F,$85,$FE,$6B,$F2,$2F,$A3,$FE,$6F
	.dc.b	$2F,$47,$FC,$E1,$39,$8E,$01,$C3,$BD,$1A,$F4,$BF,$D9,$E1,$FD,$0B
	.dc.b	$E9,$77,$8F,$F9,$ED,$E3,$87,$F4,$69,$D5,$78,$AB,$4F,$29,$57,$2C
	.dc.b	$3B,$C6,$27,$53,$9C,$39,$CF,$A2,$30,$EC,$C4,$F7,$AE,$73,$E4,$96
	.dc.b	$1C,$9F,$75,$D7,$D0,$73,$32,$53,$33,$E0,$3A,$4F,$3C,$3D,$B3,$F3
	.dc.b	$31,$F2,$9A,$7E,$76,$3E,$5C,$3E,$64,$7A,$0F,$F2,$FE,$3E,$89,$F8
	.dc.b	$26,$E3,$C4,$7A,$8F,$35,$F8,$1F,$BF,$87,$8E,$B9,$47,$DE,$3F,$95
	.dc.b	$9C,$4B,$18,$7E,$09,$F8,$81,$F4,$6F,$A8,$7D,$47,$28,$C3,$63,$33
	.dc.b	$FC,$E2,$72,$DF,$A8,$70,$F1,$6C,$66,$30,$47,$E9,$7D,$43,$83,$F3
	.dc.b	$7C,$1F,$96,$8C,$4F,$DB,$C1,$87,$4A,$FC,$DC,$63,$0F,$C0,$5F,$0B
	.dc.b	$EB,$87,$4D,$BF,$4F,$7B,$9F,$C1,$37,$83,$C1,$5F,$AC,$6C,$FD,$A3
	.dc.b	$E3,$E5,$EC,$8F,$0B,$E3,$BF,$D9,$E1,$E1,$CB,$8D,$F6,$AD,$3F,$E2
	.dc.b	$EC,$7C,$53,$A2,$5F,$9A,$B7,$5E,$0C,$BD,$B9,$7A,$E3,$FC,$6F,$8F
	.dc.b	$7B,$E1,$5F,$1E,$F8,$CE,$6B,$77,$C5,$E8,$79,$6B,$AB,$0C,$13,$0F
	.dc.b	$2B,$C8,$ED,$1F,$B8,$E4,$FC,$5F,$D1,$7E,$49,$B3,$FD,$F9,$AF,$F5
	.dc.b	$EC,$D3,$B6,$C7,$F9,$7F,$89,$F5,$3F,$C7,$B0,$FE,$13,$E9,$FF,$18
	.dc.b	$8C,$4F,$F6,$3B,$E3,$7F,$EB,$BF,$5B,$FC,$46,$FF,$BF,$5A,$E7,$C9
	.dc.b	$E2,$FF,$97,$FD,$07,$D3,$E3,$5F,$F3,$A5,$7A,$EF,$8D,$FF,$30,$9C
	.dc.b	$BE,$D8,$CB,$FD,$D5,$EF,$A7,$A3,$0F,$F5,$1C,$FF,$9E,$DA,$F8,$CF
	.dc.b	$EE,$59,$7F,$7D,$FA,$77,$C5,$F8,$DB,$3A,$3F,$B0,$38,$7F,$63,$3F
	.dc.b	$DE,$FE,$9B,$FD,$B1,$1B,$C9,$5B,$8F,$53,$5F,$05,$FD,$98,$C4,$76
	.dc.b	$8F,$8F,$BF,$FA,$17,$3A,$FF,$A6,$67,$B5,$78,$F3,$FE,$7F,$87,$F8
	.dc.b	$47,$93,$E6,$78,$1F,$BA,$BF,$23,$5F,$49,$FC,$2B,$37,$D2,$F0,$AF
	.dc.b	$C6,$56,$EB,$E9,$35,$FF,$12,$35,$FF,$02,$3D,$70,$FC,$4F,$E0,$1F
	.dc.b	$E5,$B9,$FF,$56,$35,$F0,$E3,$E7,$7E,$47,$DF,$3C,$BF,$C2,$36,$33
	.dc.b	$1F,$87,$FA,$8F,$F4,$54,$60,$FD,$04,$ED,$FF,$1E,$B5,$FC,$53,$5F
	.dc.b	$F1,$E3,$BC,$3F,$3F,$F7,$EF,$1F,$E8,$B6,$BB,$1E,$8B,$7D,$57,$8C
	.dc.b	$91,$F8,$7F,$83,$7C,$F3,$E4,$B3,$3F,$E5,$0B,$E2,$FE,$DD,$AE,$6F
	.dc.b	$F7,$91,$8C,$BF,$67,$F8,$3B,$C7,$FC,$2B,$6F,$F2,$15,$BE,$9F,$F2
	.dc.b	$C6,$5F,$E1,$39,$FC,$A7,$D5,$3E,$5B,$F6,$0D,$8F,$CC,$4E,$DF,$AB
	.dc.b	$19,$7F,$CC,$ED,$7F,$68,$D7,$FC,$CC,$D7,$FD,$B9,$F0,$CF,$77,$F3
	.dc.b	$6D,$8F,$F7,$5C,$78,$AF,$F8,$2C,$3E,$A7,$FD,$20,$FF,$BC,$F4,$7C
	.dc.b	$CF,$E9,$1B,$1F,$B5,$38,$7A,$FF,$C2,$F9,$1F,$45,$FC,$95,$B3,$FC
	.dc.b	$95,$AF,$FA,$09,$AF,$F8,$FB,$91,$F2,$4F,$97,$FE,$33,$18,$6C,$66
	.dc.b	$7F,$00,$FF,$B2,$9F,$27,$DD,$3E,$57,$DA,$D8,$CC,$7C,$0F,$A5,$FF
	.dc.b	$56,$46,$0F,$F0,$6C,$ED,$FF,$5C,$B5,$FF,$06,$35,$FF,$2A,$BB,$E7
	.dc.b	$FF,$20,$C3,$ED,$9E,$DF,$CE,$6C,$77,$3F,$81,$F8,$CF,$8C,$FE,$57
	.dc.b	$D2,$94,$E7,$EF,$FF,$8A,$D1,$BE,$64,$AF,$E5,$15,$B6,$D6,$DB,$3F
	.dc.b	$84,$F9,$5E,$7E,$04,$9E,$3D,$7F,$CE,$35,$FE,$A6,$DC,$FC,$99,$79
	.dc.b	$EF,$85,$FA,$AB,$4B,$8D,$F2,$AF,$91,$FB,$7F,$4F,$F5,$EB,$F9,$1D
	.dc.b	$23,$19,$FF,$9A,$A3,$8D,$70,$F8,$D9,$7F,$BF,$BD,$E9,$F1,$EF,$73
	.dc.b	$9A,$DB,$DA,$B6,$F9,$71,$F2,$3F,$C2,$5F,$E9,$6F,$91,$B8,$F8,$DB
	.dc.b	$87,$D5,$F7,$BF,$DE,$DE,$DF,$E5,$A7,$37,$D7,$E6,$B0,$7F,$72,$AD
	.dc.b	$C7,$D4,$F4,$BE,$D9,$1F,$B3,$F2,$6F,$A9,$C4,$7E,$1B,$97,$FA,$93
	.dc.b	$6B,$FA,$BD,$6E,$3F,$69,$F5,$1A,$D7,$F2,$6D,$D7,$FD,$56,$D9,$FC
	.dc.b	$73,$F5,$9E,$37,$FE,$6B,$8E,$F1,$FE,$6B,$6B,$CA,$34,$7F,$0A,$FA
	.dc.b	$FF,$C0,$C1,$B3,$DF,$97,$F3,$DF,$41,$F9,$AF,$76,$5F,$E5,$A3,$F3
	.dc.b	$A6,$1F,$81,$18,$7F,$01,$D7,$F8,$A2,$3F,$CA,$BE,$72,$BF,$7E,$2E
	.dc.b	$9F,$DD,$BC,$C5,$7D,$3F,$81,$FE,$5D,$3F,$EB,$8E,$24,$63,$7F,$BC
	.dc.b	$DC,$7F,$39,$F2,$0F,$9C,$E5,$E6,$FF,$A5,$30,$E2,$DB,$FE,$69,$18
	.dc.b	$BF,$F4,$A1,$FE,$DB,$8E,$9D,$7F,$A1,$FB,$9F,$D2,$9D,$27,$F6,$7B
	.dc.b	$7F,$3A,$23,$FE,$1F,$FA,$2F,$F4,$F6,$BF,$B6,$9C,$CF,$EC,$DF,$F1
	.dc.b	$F4,$EB,$5F,$DC,$F3,$FD,$FA,$FC,$67,$57,$F0,$C3,$F1,$7F,$31,$63
	.dc.b	$F8,$4C,$57,$F8,$BF,$4F,$F8,$09,$FF,$05,$F9,$6F,$A8,$FE,$80,$48
	.dc.b	$FE,$80,$D7,$F3,$5F,$07,$FC,$68,$FE,$91,$1F,$C7,$3C,$59,$3F,$C6
	.dc.b	$1D,$34,$87,$E8,$DC,$BF,$E2,$C1,$F5,$0F,$A4,$FE,$55,$1F,$CD,$0C
	.dc.b	$B8,$D7,$FD,$8D,$3B,$9F,$E6,$11,$89,$D4,$72,$9D,$3F,$EF,$2F,$A7
	.dc.b	$C7,$4F,$FB,$61,$B1,$FC,$E7,$E3,$91,$C6,$F5,$3F,$CF,$7E,$9D,$EA
	.dc.b	$7F,$8B,$EF,$73,$98,$C5,$FA,$AB,$7E,$FF,$58,$F2,$6B,$D2,$D5,$A7
	.dc.b	$CC,$73,$AE,$0D,$32,$7C,$89,$E5,$3C,$70,$F1,$CF,$4C,$3C,$6F,$18
	.dc.b	$CE,$7F,$81,$8E,$A6,$F0,$B7,$CF,$1F,$56,$1C,$AB,$C0,$69,$C4,$EE
	.dc.b	$3B,$4F,$B9,$D5,$47,$D3,$3B,$79,$98,$AE,$64,$9F,$2D,$F9,$E7,$51
	.dc.b	$CF,$0F,$7D,$F3,$AD,$46,$2F,$95,$F0,$3E,$4C,$EB,$0F,$6E,$1C,$DB
	.dc.b	$3A,$C7,$E6,$DF,$2C,$3A,$5F,$92,$72,$F1,$AF,$82,$7F,$80,$F0,$EC
	.dc.b	$E0,$F9,$C7,$8E,$BE,$9D,$72,$8C,$61,$E2,$9C,$C7,$53,$8B,$E2,$7C
	.dc.b	$C6,$71,$F9,$87,$C3,$AE,$93,$B2,$57,$D5,$AE,$E7,$96,$B8,$3F,$5C
	.dc.b	$FD,$99,$E8,$2B,$F9,$76,$E3,$FE,$87,$7F,$37,$E6,$3E,$4F,$F9,$4C
	.dc.b	$61,$B5,$FE,$7F,$6C,$EE,$7F,$97,$CE,$5F,$C3,$C6,$70,$FC,$EF,$D3
	.dc.b	$BE,$A1,$C1,$F9,$8A,$D6,$7F,$B1,$6E,$1F,$B9,$6A,$BE,$55,$C7,$29
	.dc.b	$3E,$07,$F4,$B3,$A8,$C1,$31,$FC,$8D,$F4,$9D,$46,$2B,$9B,$6B,$F7
	.dc.b	$33,$AC,$3C,$D5,$A6,$CE,$B1,$F6,$BD,$E3,$37,$CC,$F8,$73,$A7,$11
	.dc.b	$A2,$4F,$F9,$56,$FC,$0F,$C3,$BF,$03,$4F,$06,$38,$7E,$C1,$E1,$39
	.dc.b	$61,$DE,$FA,$C6,$B0,$F0,$BA,$AE,$59,$7D,$B9,$CC,$FD,$1B,$E4,$F8
	.dc.b	$63,$33,$BB,$FE,$87,$8F,$A2,$BB,$47,$82,$75,$1D,$CF,$31,$A3,$A4
	.dc.b	$72,$BE,$F1,$92,$4F,$27,$C9,$87,$A3,$2F,$C3,$CE,$70,$F1,$3B,$24
	.dc.b	$7E,$06,$FA,$C6,$FD,$AE,$F4,$F0,$AC,$61,$B8,$7F,$5A,$C7,$DF,$3F
	.dc.b	$80,$8F,$D8,$D7,$FA,$F1,$D4,$FD,$B9,$D3,$FE,$7E,$FC,$FD,$F4,$7F
	.dc.b	$88,$EE,$FC,$4F,$90,$9C,$E1,$FE,$90,$5F,$23,$B8,$E5,$77,$0F,$F4
	.dc.b	$DD,$70,$7F,$D4,$2D,$C3,$AB,$CA,$BA,$57,$1F,$86,$7D,$17,$FB,$CE
	.dc.b	$7C,$18,$E6,$31,$8F,$48,$C7,$BF,$FE,$01,$85,$F2,$8C,$36,$FF,$C6
	.dc.b	$AD,$9F,$E2,$F9,$FB,$FD,$9F,$53,$6B,$FC,$EF,$F1,$D8,$D3,$FD,$7F
	.dc.b	$1F,$0E,$B0,$70,$F0,$15,$89,$FF,$77,$7C,$07,$4C,$FF,$DD,$31,$FE
	.dc.b	$3D,$9C,$4C,$8D,$99,$2E,$4D,$FF,$CB,$5F,$21,$FE,$33,$9F,$F6,$54
	.dc.b	$7F,$AA,$FE,$8D,$F2,$BF,$4F,$CD,$5E,$F0,$FF,$59,$D6,$B0,$E2,$7F
	.dc.b	$A7,$E7,$53,$FE,$2C,$AE,$3D,$DF,$35,$B5,$FE,$5E,$6C,$FF,$40,$6E
	.dc.b	$3E,$6A,$FC,$27,$98,$F1,$9E,$97,$97,$2F,$25,$FA,$79,$77,$A7,$E1
	.dc.b	$73,$F8,$1A,$7F,$B9,$F9,$5F,$0B,$FC,$EA,$3D,$33,$FE,$D3,$AE,$18
	.dc.b	$73,$9C,$E8,$9C,$C7,$3B,$E5,$7F,$13,$1F,$BD,$F3,$DF,$7F,$B2,$DF
	.dc.b	$E8,$5B,$5F,$9C,$6E,$5F,$9C,$7F,$45,$1F,$EE,$5F,$80,$BD,$EF,$85
	.dc.b	$CE,$BF,$EF,$DC,$7D,$37,$FD,$56,$73,$7F,$9B,$8E,$53,$FD,$92,$B8
	.dc.b	$47,$FA,$39,$AF,$F7,$07,$38,$FF,$42,$8E,$27,$F9,$A7,$C4,$7C,$27
	.dc.b	$6B,$FD,$E7,$27,$E7,$4F,$CD,$7C,$2D,$8C,$92,$FF,$D0,$B1,$97,$FD
	.dc.b	$FB,$F8,$03,$C5,$FB,$ED,$AF,$AB,$D0,$FA,$87,$9D,$AB,$95,$F5,$AF
	.dc.b	$86,$1F,$80,$18,$71,$BD,$C7,$E6,$BC,$8F,$C3,$9F,$8E,$F4,$3F,$DD
	.dc.b	$98,$7E,$03,$A3,$FC,$EA,$D9,$DB,$F8,$49,$DD,$73,$8C,$7B,$DA,$8F
	.dc.b	$17,$12,$B7,$C5,$FA,$FC,$3F,$78,$7F,$05,$FA,$CB,$8F,$F2,$76,$E1
	.dc.b	$FC,$89,$B5,$FC,$89,$FE,$19,$E8,$7F,$0D,$F4,$6D,$FF,$2F,$6D,$FF
	.dc.b	$2A,$6D,$FF,$2A,$7F,$1F,$F8,$03,$D2,$FC,$8F,$EC,$AF,$B5,$CC,$61
	.dc.b	$B8,$FF,$8B,$AB,$83,$67,$F2,$11,$8C,$7F,$DB,$6D,$C7,$F1,$8D,$AF
	.dc.b	$F5,$8F,$CC,$B4,$EB,$7F,$D0,$3D,$DF,$E5,$07,$F2,$CF,$23,$FA,$C6
	.dc.b	$3F,$D0,$A3,$F9,$E7,$B1,$FC,$DB,$D0,$34,$99,$7F,$91,$3E,$A7,$F0
	.dc.b	$B7,$FD,$5B,$CE,$7C,$B7,$EC,$27,$34,$CD,$E9,$2F,$73,$FC,$22,$1E
	.dc.b	$83,$E5,$FF,$9B,$1F,$D4,$FE,$A7,$F7,$15,$B9,$D3,$FC,$43,$CD,$4E
	.dc.b	$B1,$FF,$30,$C6,$3C,$8F,$EA,$33,$99,$FF,$6B,$CE,$6B,$FA,$9F,$9A
	.dc.b	$F8,$BF,$06,$1F,$D8,$E3,$13,$DF,$0E,$DB,$FF,$8F,$B0,$7D,$3F,$F7
	.dc.b	$FC,$BF,$51,$AB,$E3,$BF,$81,$E3,$FC,$A6,$B5,$87,$F7,$E9,$CD,$FF
	.dc.b	$68,$9E,$4F,$99,$AF,$1F,$67,$FB,$5C,$62,$38,$47,$F7,$98,$EB,$7F
	.dc.b	$47,$C1,$F8,$BF,$53,$9A,$FE,$FD,$F5,$0F,$91,$FE,$DF,$F8,$0F,$F8
	.dc.b	$17,$0F,$E7,$D3,$97,$EB,$46,$72,$FE,$FB,$39,$7E,$7B,$5F,$9A,$D7
	.dc.b	$9D,$7F,$80,$A3,$86,$9F,$25,$B7,$F1,$7D,$4F,$2D,$7F,$84,$FE,$A7
	.dc.b	$F0,$F8,$7C,$CF,$A2,$FF,$00,$BF,$C5,$78,$53,$AC,$BF,$C5,$35,$A7
	.dc.b	$EB,$47,$2A,$FC,$0B,$5F,$EB,$8D,$7F,$C3,$2D,$AE,$BD,$5E,$33,$FC
	.dc.b	$5D,$4F,$A8,$7D,$36,$74,$FC,$A7,$BE,$F9,$6F,$F0,$DF,$CA,$BE,$AF
	.dc.b	$FC,$5D,$39,$AF,$D1,$4E,$5E,$35,$FE,$28,$9D,$6B,$F9,$C6,$BF,$9A
	.dc.b	$6B,$FD,$D5,$DE,$5F,$95,$6D,$7F,$85,$B2,$7D,$47,$8B,$0F,$F2,$07
	.dc.b	$E4,$6F,$3B,$FC,$63,$ED,$FF,$AA,$72,$FC,$F4,$E7,$1F,$DC,$4E,$5F
	.dc.b	$F0,$24,$67,$3F,$D8,$B5,$FD,$7B,$5D,$DF,$F9,$12,$38,$5F,$68,$F0
	.dc.b	$F9,$3E,$47,$FC,$A9,$EF,$BE,$AF,$FC,$8B,$F1,$5F,$E1,$BC,$BF,$81
	.dc.b	$4E,$5F,$E1,$91,$9C,$BF,$0D,$39,$7F,$82,$B5,$FE,$06,$D7,$FC,$3C
	.dc.b	$D7,$FD,$D7,$1F,$C6,$F9,$9F,$53,$FE,$35,$F1,$BF,$D7,$F1,$8A,$D8
	.dc.b	$70,$FF,$6A,$8D,$CF,$FB,$DA,$15,$D6,$3A,$61,$67,$9D,$F2,$F6,$7F
	.dc.b	$CE,$2F,$48,$FE,$5D,$7B,$7B,$55,$72,$78,$EF,$AB,$4E,$AE,$2B,$A4
	.dc.b	$79,$0D,$1D,$0A,$74,$7C,$98,$7B,$23,$99,$FD,$07,$0F,$3D,$7B,$F0
	.dc.b	$F3,$46,$27,$67,$5B,$F0,$CF,$F4,$9E,$4E,$B8,$DC,$39,$5A,$72,$D8
	.dc.b	$CD,$E0,$E7,$1E,$AC,$3C,$0F,$F5,$53,$DD,$1B,$9E,$27,$C3,$3B,$C7
	.dc.b	$D7,$3E,$21,$70,$C3,$F3,$8E,$87,$FA,$03,$E1,$9F,$55,$C7,$E1,$7F
	.dc.b	$D0,$AD,$8C,$CF,$E0,$E7,$2D,$F2,$FF,$D2,$0D,$E2,$E5,$18,$8E,$F3
	.dc.b	$FD,$A6,$72,$7E,$0F,$E5,$3F,$D8,$F3,$78,$3E,$06,$9F,$94,$BE,$13
	.dc.b	$F7,$9B,$1F,$98,$AE,$33,$F8,$26,$C7,$1C,$3C,$AD,$C3,$B5,$43,$A4
	.dc.b	$EB,$4F,$A1,$87,$67,$E9,$61,$D1,$B8,$F9,$9D,$D7,$1B,$FF,$06,$5F
	.dc.b	$17,$FC,$1D,$7C,$5B,$7C,$5D,$D7,$17,$FD,$66,$D9,$E1,$06,$D7,$58
	.dc.b	$DE,$1E,$BC,$7E,$26,$1D,$67,$27,$BA,$75,$8F,$AA,$3F,$69,$7C,$CE
	.dc.b	$98,$78,$0E,$F8,$F9,$7F,$03,$FF,$6C,$D7,$FA,$51,$AF,$F9,$CC,$38
	.dc.b	$C1,$C1,$FD,$FF,$EF,$3F,$A7,$57,$1F,$8B,$FC,$F5,$7F,$A4,$EF,$85
	.dc.b	$FE,$85,$B4,$F2,$18,$72,$8F,$07,$DC,$FF,$89,$5F,$F5,$3B,$5C,$FA
	.dc.b	$1D,$B0,$FD,$67,$E1,$8F,$B3,$C5,$7F,$90,$EF,$86,$7F,$EA,$56,$DF
	.dc.b	$EC,$DB,$3F,$E4,$9B,$E4,$F8,$30,$E6,$A3,$CD,$7F,$EA,$F6,$DF,$3A
	.dc.b	$DD,$FF,$87,$EF,$D3,$E3,$E0,$F2,$F8,$CF,$A6,$7D,$66,$6B,$64,$3F
	.dc.b	$C2,$11,$A7,$19,$FD,$27,$E5,$7C,$EF,$FA,$CD,$FF,$70,$B5,$F9,$78
	.dc.b	$7A,$BF,$1C,$FF,$04,$C7,$FB,$49,$AF,$DF,$6D,$7B,$7E,$4B,$87,$95
	.dc.b	$2B,$83,$F1,$37,$FF,$7F,$46,$0F,$F6,$EA,$FA,$4D,$1C,$63,$F1,$9F
	.dc.b	$8A,$65,$FF,$71,$B5,$CC,$7D,$AE,$46,$9F,$2E,$72,$70,$9F,$AB,$5C
	.dc.b	$72,$F8,$97,$C3,$4C,$D7,$3C,$1F,$23,$4C,$E3,$FE,$F4,$FC,$39,$F0
	.dc.b	$1F,$80,$6C,$66,$3F,$07,$DD,$CA,$30,$43,$F5,$13,$C1,$FD,$54,$6D
	.dc.b	$CC,$F2,$E8,$91,$C7,$4F,$4E,$3F,$EF,$B7,$FD,$F9,$F7,$7F,$EF,$A9
	.dc.b	$FF,$4F,$46,$3E,$4F,$D9,$D3,$F9,$6A,$D5,$7E,$E5,$B5,$B8,$FC,$2C
	.dc.b	$FE,$73,$AF,$EF,$BF,$B5,$E3,$C1,$CF,$0F,$D0,$FD,$E7,$AF,$0F,$C8
	.dc.b	$60,$8C,$5F,$F5,$D6,$C7,$EF,$7E,$F7,$F5,$AB,$2F,$D6,$BE,$5E,$DF
	.dc.b	$41,$DB,$6F,$D7,$FD,$57,$A3,$1F,$ED,$0D,$AF,$F5,$F3,$57,$D5,$BD
	.dc.b	$8F,$DB,$FE,$64,$F2,$29,$2B,$E4,$E1,$FD,$CF,$C4,$7D,$C6,$67,$F9
	.dc.b	$05,$F1,$76,$7F,$B3,$5C,$3F,$C9,$67,$8B,$A1,$1A,$E0,$F8,$ED,$BF
	.dc.b	$F0,$13,$67,$FD,$B1,$5C,$9F,$E6,$1F,$93,$7F,$BF,$5E,$8F,$F8,$45
	.dc.b	$BE,$77,$F1,$48,$D4,$70,$F8,$DC,$4C,$8E,$31,$BF,$9B,$91,$D1,$E5
	.dc.b	$EA,$62,$3C,$31,$73,$FB,$F8,$7F,$BF,$E7,$DD,$FA,$0F,$EC,$E7,$F0
	.dc.b	$6D,$D3,$FC,$3C,$D9,$FF,$7A,$7E,$0D,$FC,$19,$FE,$99,$19,$4D,$31
	.dc.b	$1B,$AE,$BE,$0F,$F3,$29,$D3,$FC,$D2,$75,$7F,$E2,$DF,$D3,$3E,$55
	.dc.b	$FE,$59,$7D,$87,$F3,$A6,$E3,$F5,$DD,$D7,$1E,$1D,$54,$AD,$47,$F4
	.dc.b	$2F,$AE,$F9,$F8,$7E,$89,$B7,$FE,$3D,$6D,$03,$F6,$5F,$ED,$8E,$9C
	.dc.b	$EF,$D4,$ED,$C1,$FD,$43,$E7,$7E,$84,$FF,$73,$6E,$1F,$B2,$6E,$1E
	.dc.b	$1E,$0E,$1C,$32,$FF,$29,$5A,$7C,$99,$2B,$51,$D6,$AE,$1E,$D9,$FE
	.dc.b	$C3,$F6,$FF,$3A,$BD,$8D,$4D,$1F,$37,$C3,$CB,$93,$FA,$59,$D4,$FF
	.dc.b	$6C,$FC,$63,$E8,$CF,$C7,$7E,$AB,$FE,$11,$6E,$1F,$C4,$1D,$C1,$A6
	.dc.b	$B4,$FF,$37,$36,$3A,$6F,$FD,$EB,$F3,$D7,$C8,$52,$4F,$EE,$7E,$27
	.dc.b	$EA,$4E,$A3,$F7,$8D,$7F,$C4,$8D,$C7,$F9,$1C,$F0,$AE,$75,$A7,$FC
	.dc.b	$13,$1F,$7C,$FA,$11,$DF,$6C,$9A,$DA,$46,$48,$7F,$8C,$E7,$8B,$FE
	.dc.b	$35,$F7,$FE,$91,$F4,$1F,$E2,$4D,$75,$BF,$F8,$6A,$37,$B5,$31,$1F
	.dc.b	$E1,$EF,$C0,$BF,$D9,$39,$FF,$3E,$6E,$FF,$E9,$96,$E1,$F9,$D5,$F0
	.dc.b	$9C,$7D,$C9,$3F,$C8,$E7,$69,$DB,$FE,$4F,$9F,$EA,$44,$78,$7C,$45
	.dc.b	$7F,$CA,$5E,$5F,$06,$CF,$9E,$7F,$C6,$BF,$58,$FC,$0F,$FD,$5D,$F0
	.dc.b	$4F,$9F,$FF,$58,$4E,$70,$FF,$6C,$C6,$2B,$6F,$ED,$FE,$0D,$F8,$97
	.dc.b	$F8,$3F,$0E,$57,$FB,$BC,$39,$3F,$EE,$AB,$E4,$F4,$35,$87,$47,$FD
	.dc.b	$67,$7C,$E7,$FD,$85,$F5,$AF,$AE,$FE,$7D,$8F,$6B,$4E,$7C,$0E,$91
	.dc.b	$8A,$DD,$FB,$3E,$A7,$B4,$E6,$B7,$7C,$63,$CC,$7F,$13,$71,$87,$7C
	.dc.b	$D7,$CA,$73,$87,$9B,$1E,$F8,$71,$6A,$BE,$2F,$82,$33,$5C,$A3,$86
	.dc.b	$3E,$AA,$DC,$79,$0F,$41,$67,$83,$8C,$7C,$D8,$F9,$70,$EA,$4C,$3A
	.dc.b	$46,$5F,$11,$2B,$B6,$1E,$0B,$F9,$D8,$76,$26,$1E,$C9,$D6,$3F,$07
	.dc.b	$0E,$57,$C2,$B5,$8F,$7C,$7D,$11,$E3,$7B,$C6,$67,$85,$72,$BF,$B1
	.dc.b	$54,$F8,$A7,$28,$F5,$B9,$3E,$14,$78,$DE,$A6,$71,$F4,$CF,$08,$E4
	.dc.b	$4A,$F0,$BC,$0C,$57,$38,$C9,$F3,$A3,$A6,$1E,$B8,$C4,$70,$3C,$EF
	.dc.b	$09,$F5,$C7,$80,$FE,$C0,$7B,$0A,$7D,$E2,$DF,$E5,$5B,$E0,$3E,$33
	.dc.b	$71,$F9,$78,$C3,$89,$46,$23,$8C,$F0,$7F,$53,$1C,$2F,$EF,$B6,$FC
	.dc.b	$BF,$5D,$FD,$87,$97,$F4,$F3,$9B,$F0,$61,$83,$F6,$35,$B8,$FC,$B4
	.dc.b	$ED,$EF,$3B,$7F,$39,$3B,$C3,$C3,$39,$6C,$EB,$AF,$FD,$75,$18,$26
	.dc.b	$3F,$C1,$B0,$ED,$87,$38,$C6,$1C,$DB,$87,$3C,$70,$FA,$8F,$3C,$FA
	.dc.b	$B1,$EF,$5C,$31,$EF,$5B,$7F,$89,$61,$DE,$31,$3B,$23,$FE,$1D,$8D
	.dc.b	$3F,$08,$EE,$49,$E7,$5E,$08,$E6,$43,$84,$6B,$0F,$4C,$62,$FB,$AB
	.dc.b	$EF,$97,$E0,$B1,$DB,$67,$E1,$1E,$1A,$E3,$19,$C3,$B4,$E4,$F1,$46
	.dc.b	$67,$F0,$B8,$F9,$E7,$C4,$6B,$2F,$A7,$19,$AE,$13,$9C,$3E,$76,$5E
	.dc.b	$4B,$FC,$35,$6A,$BD,$57,$C6,$7E,$EF,$E1,$BD,$F3,$3F,$1F,$89,$B3
	.dc.b	$ED,$75,$33,$3F,$CA,$BB,$9D,$46,$0F,$E9,$3F,$47,$FD,$8D,$FF,$4B
	.dc.b	$37,$37,$57,$FA,$F3,$2C,$3F,$EB,$E9,$B5,$FB,$C6,$D7,$F9,$75,$AF
	.dc.b	$D5,$DD,$3D,$9B,$1D,$A3,$F6,$3F,$45,$8B,$F0,$36,$3F,$6B,$E3,$DA
	.dc.b	$FC,$74,$D7,$F0,$4D,$C4,$D9,$C3,$C7,$99,$B3,$ED,$30,$BF,$01,$B1
	.dc.b	$86,$E5,$FE,$C0,$AE,$31,$8C,$BD,$47,$F8,$1A,$F9,$56,$DF,$F2,$6B
	.dc.b	$7D,$EE,$FE,$2F,$81,$B1,$93,$84,$65,$FF,$65,$7E,$74,$F8,$DF,$80
	.dc.b	$D8,$FF,$69,$FC,$8B,$E0,$FE,$0C,$70,$AD,$91,$DC,$6B,$2F,$48,$BE
	.dc.b	$8E,$1B,$C3,$FE,$FE,$6D,$7F,$A6,$F0,$F0,$6F,$D6,$FE,$1E,$DF,$3A
	.dc.b	$BE,$6F,$C2,$7C,$E6,$9F,$14,$B3,$C3,$6F,$C3,$C7,$C9,$9F,$B1,$2F
	.dc.b	$BE,$A6,$3E,$0A,$D7,$D5,$9D,$36,$74,$D3,$FD,$E9,$F9,$CB,$DC,$FC
	.dc.b	$53,$63,$FD,$FB,$F4,$AF,$97,$7D,$4E,$AE,$57,$27,$F2,$95,$B9,$CE
	.dc.b	$3F,$33,$0E,$8D,$AD,$E1,$D3,$0E,$47,$A7,$CC,$F7,$9F,$03,$CA,$B1
	.dc.b	$92,$5F,$CE,$3F,$CD,$EE,$38,$FF,$5B,$C1,$FD,$C2,$3E,$27,$89,$F2
	.dc.b	$AB,$E5,$B6,$32,$7C,$98,$CB,$FE,$D7,$AF,$DB,$C7,$EB,$3D,$3F,$D8
	.dc.b	$71,$9F,$DA,$5B,$F3,$EF,$F6,$EE,$1C,$4F,$F0,$0D,$6D,$C5,$FF,$B8
	.dc.b	$E3,$0F,$E4,$9B,$87,$F0,$6B,$E9,$EB,$F8,$31,$98,$72,$C3,$C8,$F4
	.dc.b	$AD,$B9,$C3,$9B,$FE,$F1,$1F,$A1,$3E,$D3,$F8,$64,$61,$B1,$F6,$32
	.dc.b	$FE,$2F,$39,$3D,$25,$C3,$F8,$0B,$63,$73,$C6,$3C,$0F,$77,$8A,$FB
	.dc.b	$7E,$FF,$1B,$E0,$6C,$E9,$B7,$DA,$B8,$4F,$2A,$FE,$45,$F7,$2F,$88
	.dc.b	$A7,$D4,$DD,$D5,$8B,$F9,$15,$B6,$D6,$E7,$58,$FF,$04,$AD,$36,$75
	.dc.b	$5C,$0F,$E9,$3E,$47,$F3,$3F,$23,$FA,$AC,$7D,$5F,$AE,$7B,$79,$AF
	.dc.b	$E7,$11,$EA,$C3,$F8,$8B,$77,$FC,$7F,$D9,$3F,$07,$DF,$0F,$F3,$F8
	.dc.b	$C6,$5F,$81,$6E,$1F,$CC,$BE,$4D,$F1,$FF,$AD,$8F,$E3,$B3,$AC,$7A
	.dc.b	$FC,$DF,$33,$FA,$57,$AA,$FB,$3C,$65,$F8,$D8,$C3,$63,$FA,$96,$4F
	.dc.b	$AD,$7D,$BF,$29,$CC,$62,$B8,$1F,$A0,$AE,$15,$B8,$FC,$D0,$F8,$FF
	.dc.b	$C9,$FD,$3F,$D0,$CF,$F6,$33,$93,$F2,$31,$97,$F7,$D3,$FD,$8D,$D6
	.dc.b	$BF,$94,$FC,$5F,$F8,$2B,$FC,$FA,$33,$7D,$2B,$81,$FA,$2A,$DC,$66
	.dc.b	$7F,$5D,$E4,$7F,$67,$7D,$59,$BF,$00,$FE,$E5,$39,$C7,$E9,$4E,$6B
	.dc.b	$8F,$A9,$A3,$E2,$FA,$36,$32,$4A,$E3,$19,$7F,$AA,$F3,$B8,$BC,$DF
	.dc.b	$4B,$FE,$01,$8C,$5F,$ED,$35,$4E,$6A,$4E,$7C,$3B,$87,$49,$BB,$7E
	.dc.b	$FB,$0F,$F4,$85,$FF,$0D,$C7,$AD,$F8,$BC,$4F,$BB,$F5,$BE,$16,$D6
	.dc.b	$DC,$3F,$C7,$BE,$89,$F5,$DF,$C0,$6F,$94,$EC,$8F,$F8,$2A,$77,$E3
	.dc.b	$72,$8F,$EE,$BE,$9B,$E7,$3F,$C4,$5A,$BE,$A9,$F4,$9F,$E2,$3A,$D6
	.dc.b	$5F,$CB,$AB,$53,$9B,$93,$FC,$9F,$F1,$5F,$DA,$BF,$CA,$EF,$8E,$3E
	.dc.b	$0A,$DB,$FE,$1E,$E2,$5F,$08,$D5,$FE,$CF,$CD,$FD,$B1,$FA,$6F,$6D
	.dc.b	$F6,$98,$9D,$12,$BF,$71,$3A,$8C,$69,$FD,$42,$43,$C1,$BF,$54,$F4
	.dc.b	$FF,$D9,$BB,$7E,$03,$0C,$C7,$FB,$22,$AD,$FF,$85,$FA,$7F,$CA,$38
	.dc.b	$71,$8F,$F6,$9C,$E9,$61,$CD,$4E,$DF,$F6,$DC,$EE,$75,$39,$9F,$3C
	.dc.b	$E6,$BB,$DF,$8E,$79,$BE,$88,$CC,$78,$8C,$9E,$7C,$3C,$8F,$43,$27
	.dc.b	$9A,$38,$BE,$43,$38,$79,$2B,$C1,$87,$A6,$31,$3D,$E3,$15,$CA,$3B
	.dc.b	$46,$2F,$C3,$1F,$44,$EB,$3B,$71,$3B,$7C,$A7,$2B,$F7,$57,$2C,$3E
	.dc.b	$23,$67,$9B,$6B,$91,$F2,$E3,$C0,$7F,$BE,$E7,$FD,$F8,$FC,$33,$07
	.dc.b	$E1,$89,$F5,$5F,$CA,$FD,$BC,$46,$3C,$1F,$F4,$0B,$81,$3B,$70,$D7
	.dc.b	$8A,$F9,$CF,$F5,$2C,$69,$AE,$9A,$F8,$9A,$FE,$6A,$39,$6B,$F8,$FF
	.dc.b	$AA,$7A,$7E,$9C,$FF,$1D,$18,$BF,$CA,$5F,$07,$F0,$2D,$9D,$D7,$1D
	.dc.b	$7E,$15,$F1,$BD,$E3,$E8,$BF,$CD,$61,$CF,$E0,$B8,$D7,$EB,$9A,$9D
	.dc.b	$D7,$C0,$6E,$5E,$9C,$BC,$15,$FB,$6C,$7C,$18,$EA,$3C,$35,$B6,$18
	.dc.b	$24,$78,$09,$97,$C9,$9F,$83,$8F,$A6,$3D,$B8,$78,$2D,$87,$82,$78
	.dc.b	$47,$47,$6F,$83,$0F,$21,$E5,$C7,$DE,$49,$E2,$FE,$FF,$E3,$FF,$D6
	.dc.b	$15,$F8,$70,$E1,$FC,$53,$5F,$C7,$B5,$FC,$A3,$BC,$BF,$DB,$DF,$83
	.dc.b	$BE,$4B,$F3,$11,$86,$F0,$F0,$F1,$2E,$1F,$C8,$BE,$AF,$FB,$13,$FE
	.dc.b	$61,$6B,$FE,$BA,$6D,$FF,$9E,$3E,$A5,$F4,$9C,$A9,$5C,$1F,$D2,$B5
	.dc.b	$FD,$53,$5F,$07,$8F,$FB,$08,$EB,$8F,$EB,$ED,$F2,$9D,$5B,$7E,$5B
	.dc.b	$F0,$5B,$E7,$E9,$27,$FB,$37,$D5,$FF,$5A,$7D,$91,$DC,$8F,$D1,$9E
	.dc.b	$4D,$8F,$8B,$E8,$BE,$A1,$F2,$38,$6D,$74,$3F,$D1,$1F,$57,$FD,$AD
	.dc.b	$BE,$2F,$C5,$8C,$47,$FB,$93,$4F,$AC,$E8,$FC,$14,$69,$FE,$1A,$D7
	.dc.b	$F8,$93,$5F,$EB,$2F,$4E,$0F,$A9,$F5,$5F,$1C,$F0,$5F,$D5,$FA,$B4
	.dc.b	$52,$5F,$F9,$27,$1F,$82,$FF,$BE,$27,$6F,$E5,$5A,$FF,$1D,$6B,$FC
	.dc.b	$89,$AF,$F2,$77,$76,$E1,$FE,$DF,$A7,$CB,$F8,$1F,$94,$78,$AD,$5F
	.dc.b	$EC,$4F,$8E,$FF,$48,$47,$FA,$CD,$AE,$AF,$F9,$A4,$62,$7F,$D9,$7D
	.dc.b	$BF,$E9,$4A,$E1,$F0,$3F,$CF,$2B,$8C,$7F,$38,$6B,$CA,$34,$7E,$6E
	.dc.b	$5E,$35,$3F,$3D,$5D,$F0,$FE,$89,$F5,$4F,$8E,$FD,$13,$71,$FF,$65
	.dc.b	$35,$FF,$45,$6D,$FE,$E0,$CF,$FD,$C3,$39,$9F,$E9,$6D,$7A,$D7,$F5
	.dc.b	$48,$C4,$7F,$86,$3B,$7F,$04,$FF,$08,$E1,$FE,$B3,$18,$33,$F5,$7E
	.dc.b	$E9,$FF,$0F,$B5,$34,$C5,$7E,$F2,$3F,$8A,$35,$FE,$B8,$D7,$CF,$D7
	.dc.b	$FD,$A3,$EA,$FC,$B0,$69,$31,$FE,$DB,$F5,$5F,$E9,$39,$FF,$77,$35
	.dc.b	$FB,$87,$F8,$D1,$C6,$7F,$89,$8F,$F1,$BF,$8A,$F5,$7F,$88,$57,$19
	.dc.b	$FE,$E6,$D7,$FB,$C3,$5E,$9E,$47,$88,$FE,$31,$4F,$90,$A7,$48,$FE
	.dc.b	$39,$F5,$4F,$66,$F8,$F3,$0F,$F2,$4E,$7E,$55,$5B,$73,$C9,$FE,$09
	.dc.b	$8C,$47,$B8,$DD,$73,$71,$1A,$AE,$8B,$E6,$BC,$64,$9E,$A7,$F9,$6A
	.dc.b	$3F,$DC,$6E,$0F,$1C,$6D,$EA,$FB,$3E,$23,$F9,$C7,$D5,$7F,$00,$BF
	.dc.b	$53,$71,$FF,$10,$36,$FF,$CC,$FF,$55,$FC,$61,$F5,$47,$F0,$87,$FC
	.dc.b	$46,$D7,$8F,$89,$FE,$2C,$8C,$47,$F9,$C1,$E8,$A7,$F8,$3D,$7F,$8B
	.dc.b	$DB,$1B,$AF,$E9,$DF,$55,$F8,$18,$F7,$2F,$AA,$C9,$AE,$39,$3A,$8C
	.dc.b	$BC,$23,$FA,$C7,$AF,$F3,$F1,$FF,$1F,$36,$7F,$D0,$AD,$4F,$F8,$65
	.dc.b	$F9,$CE,$49,$C1,$D3,$83,$FB,$27,$8F,$F5,$2B,$F0,$0D,$DB,$F1,$0D
	.dc.b	$C3,$C3,$C1,$C3,$83,$FC,$A7,$3C,$6F,$FC,$9E,$D7,$FC,$A8,$D7,$FD
	.dc.b	$36,$F3,$7D,$3F,$71,$EE,$3A,$3F,$85,$FA,$BF,$B5,$8F,$E6,$5B,$8F
	.dc.b	$F9,$7D,$B3,$FE,$A6,$5F,$4F,$FD,$FD,$CC,$1A,$48,$FF,$02,$7D,$5F
	.dc.b	$D7,$9F,$D0,$3F,$5D,$FC,$B3,$5C,$EF,$C0,$E3,$C1,$23,$71,$FE,$0F
	.dc.b	$FA,$B7,$CF,$52,$4F,$E8,$7E,$AF,$EC,$4E,$A3,$FC,$5C,$D7,$FC,$EC
	.dc.b	$D7,$FC,$F6,$D7,$FD,$00,$EF,$4F,$DF,$4B,$E6,$3F,$C4,$0D,$AF,$D8
	.dc.b	$56,$F5,$FF,$44,$4F,$EE,$67,$C9,$C1,$C0,$E3,$C3,$23,$24,$3F,$83
	.dc.b	$4F,$17,$F8,$47,$D5,$FD,$03,$E9,$BF,$BF,$6B,$AF,$5B,$FC,$69,$1B
	.dc.b	$E0,$A6,$23,$FC,$71,$F5,$7F,$EA,$3A,$FF,$71,$B5,$34,$9E,$CF,$0F
	.dc.b	$5E,$4F,$F0,$F9,$FF,$6D,$6F,$FC,$8F,$C8,$7C,$E7,$F9,$15,$B5,$FC
	.dc.b	$55,$AF,$F2,$3A,$FF,$1E,$BF,$C9,$63,$2F,$F9,$82,$7F,$AE,$91,$FF
	.dc.b	$34,$7D,$53,$EA,$FF,$D6,$DC,$3F,$EB,$48,$C4,$FF,$BF,$A3,$13,$A8
	.dc.b	$FF,$5E,$4E,$9F,$E7,$41,$C7,$C0,$BE,$9B,$7F,$57,$FE,$F6,$6D,$7F
	.dc.b	$BD,$1E,$77,$98,$0F,$AA,$7D,$5F,$F4,$5C,$3A,$4E,$A7,$3E,$0F,$38
	.dc.b	$C5,$6F,$0E,$F0,$BE,$2F,$59,$D3,$C2,$B9,$39,$6E,$4C,$BC,$D3,$93
	.dc.b	$A6,$1F,$ED,$FC,$7C,$71,$83,$A1,$4F,$8D,$3C,$A3,$78,$F9,$1B,$5C
	.dc.b	$F1,$F2,$DF,$03,$CB,$3A,$7A,$D7,$19,$DE,$1E,$79,$DC,$72,$AE,$11
	.dc.b	$E9,$AF,$87,$5B,$8F,$34,$E8,$95,$D3,$0F,$43,$CC,$98,$FC,$68,$CB
	.dc.b	$E0,$25,$7C,$1B,$ED,$87,$9E,$72,$D9,$D5,$72,$7C,$77,$D9,$C9,$C3
	.dc.b	$1F,$8D,$76,$3F,$13,$87,$69,$CC,$72,$9D,$3F,$2C,$C1,$30,$F1,$1F
	.dc.b	$13,$0F,$13,$8A,$E6,$E6,$79,$E3,$F1,$AF,$E1,$46,$31,$F4,$56,$E3
	.dc.b	$B5,$F2,$23,$E8,$C3,$E0,$3F,$A2,$3F,$12,$53,$F2,$05,$C3,$F1,$ED
	.dc.b	$F0,$1F,$15,$FE,$E5,$9F,$F5,$8B,$63,$33,$CC,$FD,$73,$87,$94,$7E
	.dc.b	$3A,$37,$87,$EA,$3E,$A7,$FC,$59,$B6,$6A,$5F,$E2,$2F,$79,$7A,$6F
	.dc.b	$73,$9A,$DE,$1F,$9F,$BD,$CF,$F0,$0A,$DC,$E6,$B7,$7F,$AD,$BD,$DA
	.dc.b	$F7,$39,$AD,$CF,$F0,$5A,$DD,$6A,$31,$3A,$7E,$0B,$FC,$02,$35,$18
	.dc.b	$3D,$04,$BF,$85,$D3,$FE,$6D,$7C,$D1,$9C,$3C,$4F,$EF,$EB,$9E,$5F
	.dc.b	$4A,$34,$78,$8D,$5F,$3A,$E3,$7D,$08,$FC,$B8,$D4,$F1,$9C,$B6,$75
	.dc.b	$97,$B6,$FC,$15,$AA,$F9,$17,$C9,$F7,$46,$67,$77,$CA,$7C,$E6,$9F
	.dc.b	$29,$99,$E1,$97,$D4,$AE,$35,$F6,$EF,$B4,$62,$39,$1E,$73,$58,$79
	.dc.b	$EB,$91,$F6,$2B,$8B,$8C,$BE,$C6,$5F,$32,$3E,$EF,$D5,$7A,$E3,$EC
	.dc.b	$87,$15,$97,$FD,$49,$7F,$AB,$9D,$3F,$D7,$F0,$FC,$C4,$E9,$B1,$97
	.dc.b	$FA,$AF,$D0,$7F,$5D,$8F,$D5,$42,$30,$FE,$B5,$AF,$EE,$5A,$FE,$F5
	.dc.b	$DF,$2F,$FA,$8D,$B3,$FA,$9D,$7F,$D7,$53,$96,$C7,$EB,$7B,$AD,$FE
	.dc.b	$A6,$6B,$FA,$58,$B8,$F4,$8F,$D7,$47,$F7,$86,$D7,$AA,$7F,$C6,$CD
	.dc.b	$9D,$7D,$23,$E2,$9F,$2D,$E0,$8C,$7D,$55,$9C,$DF,$FA,$BA,$72,$D8
	.dc.b	$CE,$3F,$ED,$89,$CB,$63,$39,$FF,$B6,$67,$2D,$8F,$F2,$CD,$73,$F7
	.dc.b	$3C,$1E,$27,$85,$B1,$93,$C1,$19,$7F,$D7,$DF,$54,$F8,$9F,$93,$87
	.dc.b	$FB,$4B,$ED,$7E,$4C,$E6,$BF,$D8,$F3,$97,$8D,$7C,$9C,$D5,$B7,$F9
	.dc.b	$63,$5F,$E7,$0D,$7F,$9E,$3B,$AF,$F7,$13,$6B,$FC,$43,$F2,$0F,$80
	.dc.b	$FA,$77,$FE,$E1,$AD,$2D,$7E,$A5,$FD,$1F,$83,$FA,$4D,$3E,$81,$67
	.dc.b	$E7,$37,$1F,$9D,$0F,$AB,$FB,$8D,$9F,$B4,$BD,$CC,$CC,$9C,$BC,$23
	.dc.b	$13,$FE,$F7,$FA,$B7,$D5,$FE,$5B,$57,$E7,$8F,$AD,$FC,$BD,$6B,$2F
	.dc.b	$F7,$AD,$6A,$73,$7F,$3E,$B5,$87,$F5,$96,$BF,$D9,$5A,$FF,$69,$77
	.dc.b	$9B,$0C,$57,$EA,$3E,$81,$F5,$AF,$80,$EF,$5F,$03,$CD,$F8,$67,$C4
	.dc.b	$F2,$AC,$64,$99,$FE,$6C,$FF,$3C,$B8,$CB,$E6,$C7,$CC,$F0,$FE,$6B
	.dc.b	$67,$E6,$B4,$E4,$59,$F4,$3F,$E6,$4E,$D7,$A9,$FC,$02,$DF,$58,$FB
	.dc.b	$DF,$E0,$35,$AC,$7F,$CD,$78,$F2,$BD,$D6,$B0,$FC,$D5,$EF,$1F,$F0
	.dc.b	$23,$5F,$F0,$83,$5F,$F0,$A3,$BB,$F2,$C7,$FA,$F6,$7C,$98,$FF,$C0
	.dc.b	$0C,$7C,$27,$08,$FE,$2E,$BE,$B9,$F3,$F8,$6C,$70,$BF,$B1,$1F,$6A
	.dc.b	$3F,$D2,$8E,$A1,$87,$F3,$C8,$C3,$6B,$A7,$0F,$57,$B7,$D5,$F7,$F8
	.dc.b	$EB,$8F,$73,$C8,$E3,$1E,$B1,$E0,$8C,$61,$DD,$FE,$5F,$F5,$CF,$90
	.dc.b	$FE,$67,$F4,$4F,$95,$7D,$86,$A7,$31,$8C,$7D,$D9,$30,$E5,$1F,$D3
	.dc.b	$1B,$1F,$E3,$76,$BF,$E4,$36,$BF,$E4,$77,$79,$7D,$46,$CF,$FA,$B3
	.dc.b	$8F,$F0,$FD,$FE,$38,$F0,$F9,$8F,$53,$F8,$E3,$63,$FA,$EB,$71,$FE
	.dc.b	$B9,$3F,$C8,$7E,$BD,$C1,$49,$7F,$D9,$23,$2F,$F6,$6F,$AD,$7D,$7F
	.dc.b	$F2,$96,$C6,$63,$06,$7E,$76,$47,$16,$E3,$FE,$59,$F5,$9F,$5D,$FC
	.dc.b	$CD,$B5,$FD,$D1,$B3,$B9,$D1,$FD,$9A,$33,$C1,$FA,$EF,$99,$7D,$5B
	.dc.b	$EB,$B5,$7F,$DB,$31,$7C,$9F,$E6,$BE,$BB,$E6,$E7,$FE,$E7,$8C,$73
	.dc.b	$ED,$FE,$8D,$5C,$0F,$F0,$39,$67,$FC,$E4,$D7,$FC,$FC,$E7,$3F,$D6
	.dc.b	$B4,$FF,$7A,$7D,$6F,$F0,$E9,$FC,$A9,$E1,$8F,$C5,$F8,$FF,$DA,$9F
	.dc.b	$F7,$5E,$BE,$CF,$AE,$FF,$02,$CE,$69,$39,$AE,$3E,$C7,$F5,$DA,$FE
	.dc.b	$EF,$5F,$E9,$16,$BF,$E9,$56,$BF,$EF,$87,$A2,$7B,$57,$CC,$F4,$FC
	.dc.b	$35,$FE,$16,$E7,$71,$FF,$04,$CD,$F5,$19,$8C,$37,$3F,$F1,$7D,$70
	.dc.b	$6C,$FF,$73,$8C,$67,$FE,$9D,$6E,$BF,$C5,$9B,$5F,$2B,$AF,$A5,$7C
	.dc.b	$A3,$F4,$5E,$27,$F8,$B3,$5F,$C1,$FD,$77,$9E,$31,$CC,$8C,$61,$E4
	.dc.b	$9F,$11,$1C,$1F,$BE,$C7,$F9,$42,$C1,$E0,$BE,$BB,$F8,$CC,$FF,$B0
	.dc.b	$6B,$75,$FE,$04,$FA,$0F,$F0,$BF,$D7,$3E,$BB,$FC,$7F,$E9,$A7,$31
	.dc.b	$87,$FC,$2D,$19,$C3,$FC,$AE,$DB,$FF,$61,$35,$FF,$67,$39,$E0,$FE
	.dc.b	$7A,$D9,$FE,$15,$F5,$CF,$91,$ED,$1F,$C6,$BC,$1F,$E8,$18,$7E,$53
	.dc.b	$06,$6F,$8F,$FF,$2A,$D6,$A7,$FC,$BB,$39,$BE,$5F,$31,$FC,$14,$8F
	.dc.b	$F0,$97,$FC,$61,$87,$C9,$F0,$BE,$57,$CC,$3D,$4F,$F7,$46,$DF,$8A
	.dc.b	$C3,$31,$FC,$1E,$D5,$B7,$FC,$5F,$D5,$FD,$BA,$31,$1F,$EF,$1C,$3E
	.dc.b	$B7,$D5,$E3,$1E,$92,$8D,$BE,$33,$18,$78,$B0,$FE,$F3,$3A,$95,$F4
	.dc.b	$AD,$63,$67,$FD,$FD,$35,$61,$E0,$BE,$53,$98,$C5,$FB,$67,$CD,$87
	.dc.b	$82,$30,$76,$BF,$29,$D8,$A7,$68,$E0,$E6,$79,$E1,$E3,$8F,$74,$61
	.dc.b	$E2,$7C,$1C,$70,$79,$9F,$9F,$7E,$18,$C4,$F1,$2C,$79,$5B,$3F,$52
	.dc.b	$BC,$0F,$C8,$3C,$F8,$7C,$07,$D4,$7F,$3F,$3F,$D6,$66,$CC,$1F,$AF
	.dc.b	$3F,$61,$F5,$BF,$CE,$BE,$13,$FD,$72,$D8,$D1,$AF,$0F,$FD,$40,$E0
	.dc.b	$46,$DD,$AF,$A0,$FF,$09,$C6,$E3,$1C,$1D,$A7,$33,$C1,$AF,$96,$76
	.dc.b	$FD,$4F,$AD,$7A,$5E,$9D,$3F,$2B,$18,$9F,$C3,$57,$09,$FC,$CB,$75
	.dc.b	$F8,$15,$BA,$E9,$7C,$E7,$C1,$87,$A3,$0E,$2D,$8E,$33,$BA,$FE,$31
	.dc.b	$7C,$EF,$78,$F8,$B0,$E1,$5A,$BE,$0A,$F8,$61,$E0,$BE,$07,$E7,$E3
	.dc.b	$33,$C4,$C9,$F9,$6C,$7D,$8E,$4F,$59,$AC,$7C,$6E,$08,$E0,$97,$EC
	.dc.b	$8F,$06,$1E,$6A,$EB,$1F,$7B,$1F,$5B,$E9,$9F,$5E,$1E,$68,$D6,$3F
	.dc.b	$B7,$F9,$2F,$F6,$64,$7F,$B2,$3E,$B5,$F2,$DF,$91,$AF,$F7,$DC,$62
	.dc.b	$7F,$C8,$35,$C1,$FE,$A4,$D9,$DD,$71,$E0,$FF,$3D,$DF,$1B,$DF,$8F
	.dc.b	$97,$FA,$67,$D6,$7F,$6C,$7F,$C9,$AD,$7F,$AF,$37,$0F,$F3,$CF,$D6
	.dc.b	$3E,$9F,$95,$2B,$58,$FE,$A6,$9F,$33,$A6,$D7,$6F,$27,$8D,$F0,$8F
	.dc.b	$EB,$D3,$A7,$FB,$73,$71,$FF,$44,$CE,$E7,$C1,$F4,$5C,$88,$FF,$A0
	.dc.b	$FC,$57,$D6,$FF,$61,$8F,$6C,$FC,$1F,$29,$F5,$6F,$96,$C3,$6B,$64
	.dc.b	$3F,$C5,$33,$C2,$76,$78,$23,$FC,$09,$EE,$7F,$6D,$7F,$CE,$ED,$73
	.dc.b	$E3,$FC,$3E,$4F,$93,$39,$D7,$E6,$FD,$6F,$C7,$33,$52,$FF,$D0,$3B
	.dc.b	$A7,$51,$DA,$7E,$9F,$D6,$E6,$3F,$D0,$ED,$7F,$C3,$0D,$BF,$F2,$D6
	.dc.b	$8D,$78,$E1,$F0,$BD,$D9,$87,$28,$CB,$FE,$AF,$DB,$E4,$CE,$A3,$AC
	.dc.b	$FE,$13,$E4,$3F,$D3,$D1,$FE,$98,$6B,$FE,$29,$6E,$3F,$EC,$4E,$B5
	.dc.b	$F1,$F5,$FF,$1B,$0F,$93,$A6,$70,$E9,$38,$7F,$D4,$B3,$B7,$27,$FA
	.dc.b	$A3,$EB,$1F,$1F,$F9,$86,$C6,$63,$F0,$3D,$59,$7F,$3B,$DD,$F5,$63
	.dc.b	$FA,$23,$67,$FC,$69,$5C,$9F,$D9,$F1,$7E,$37,$E1,$1F,$11,$4E,$71
	.dc.b	$FA,$8F,$AD,$F8,$38,$FF,$4F,$C3,$F5,$98,$79,$FB,$F5,$AF,$9D,$FD
	.dc.b	$7F,$D8,$3E,$22,$99,$AF,$F6,$2F,$D7,$78,$F8,$FF,$DD,$FE,$F3,$EC
	.dc.b	$BF,$6E,$DA,$DF,$17,$EE,$BE,$BF,$FD,$2F,$1F,$EC,$E6,$BF,$E5,$F6
	.dc.b	$DF,$FB,$EB,$C9,$7B,$9C,$7C,$EF,$E0,$70,$F9,$0A,$67,$0E,$93,$87
	.dc.b	$FD,$B7,$3B,$72,$7F,$B7,$BE,$BF,$FD,$9C,$DF,$4F,$F8,$7C,$7F,$8D
	.dc.b	$23,$5E,$EE,$0C,$CC,$8C,$9C,$FE,$7F,$89,$1E,$DE,$E5,$7F,$CF,$96
	.dc.b	$9C,$68,$E7,$FF,$02,$4F,$F9,$6E,$BF,$DE,$7F,$5F,$FA,$96,$F8,$9F
	.dc.b	$C9,$23,$11,$FE,$87,$FA,$FF,$E6,$58,$73,$8F,$F3,$E4,$EA,$31,$09
	.dc.b	$FD,$B6,$1F,$E1,$36,$C7,$0B,$EC,$7F,$32,$8C,$C7,$1F,$13,$2F,$FA
	.dc.b	$26,$FA,$FD,$7F,$D8,$8C,$61,$FE,$1F,$87,$D6,$7F,$3C,$6C,$E8,$FF
	.dc.b	$49,$3C,$5F,$F5,$13,$6B,$AC,$F1,$39,$71,$FF,$46,$F7,$7E,$DD,$FF
	.dc.b	$AA,$5B,$5F,$01,$B8,$7F,$8C,$5F,$BD,$3B,$E1,$E4,$47,$C8,$23,$75
	.dc.b	$F2,$BC,$EF,$AD,$3F,$EB,$76,$C6,$78,$7C,$9C,$3F,$D6,$A7,$38,$F8
	.dc.b	$7E,$BF,$D0,$66,$A6,$1E,$0C,$53,$A8,$F1,$DF,$27,$FD,$71,$F5,$FF
	.dc.b	$80,$C3,$FC,$98,$D8,$CF,$A1,$FD,$AA,$3F,$D4,$11,$EC,$74,$9A,$4F
	.dc.b	$82,$FE,$E3,$F5,$FF,$46,$FF,$CB,$0F,$D8,$AF,$F6,$B3,$75,$FC,$6B
	.dc.b	$B8,$FF,$66,$35,$3B,$9E,$7C,$1D,$B8,$3F,$BF,$FD,$7B,$E8,$E9,$30
	.dc.b	$FA,$7F,$5F,$F6,$9F,$F6,$C4,$ED,$CF,$95,$FE,$08,$9F,$E4,$B1,$F6
	.dc.b	$A7,$4F,$FB,$B5,$B9,$FE,$92,$77,$3F,$EF,$16,$C7,$EA,$27,$79,$7F
	.dc.b	$9E,$67,$F7,$55,$E4,$E2,$F0,$1E,$1E,$39,$19,$21,$D5,$FF,$0E,$FD
	.dc.b	$7F,$D2,$3E,$AC,$7F,$C0,$16,$E3,$FB,$87,$5C,$3C,$38,$25,$7E,$5A
	.dc.b	$7E,$DF,$D7,$FF,$B2,$2F,$F7,$EF,$F8,$77,$3F,$F4,$83,$6B,$FE,$00
	.dc.b	$3B,$8C,$FC,$0F,$F8,$C2,$35,$26,$9B,$43,$CF,$7C,$F7,$F5,$C2,$47
	.dc.b	$EA,$A7,$9C,$EA,$3F,$BF,$7E,$C9,$FC,$53,$1F,$E3,$1E,$B2,$31,$97
	.dc.b	$F6,$4C,$7F,$C8,$7F,$5E,$FB,$0F,$F7,$27,$D6,$7F,$AA,$E1,$F5,$F8
	.dc.b	$CF,$CB,$18,$7F,$DE,$D3,$BA,$D4,$27,$A3,$A3,$8D,$6A,$39,$47,$FB
	.dc.b	$A2,$B9,$3B,$1F,$5E,$FB,$0F,$E9,$BF,$1C,$AF,$F6,$04,$E7,$C2,$E0
	.dc.b	$DB,$E8,$78,$0C,$DF,$2A,$E2,$D8,$FA,$33,$C6,$FB,$47,$13,$99,$E4
	.dc.b	$9E,$0D,$BE,$A7,$A7,$0F,$6D,$72,$9D,$1C,$AF,$B6,$1E,$2B,$E1,$8F
	.dc.b	$CA,$9C,$D7,$93,$1E,$F1,$88,$F9,$D3,$B9,$C9,$2B,$9C,$F9,$0F,$04
	.dc.b	$65,$F5,$E3,$E6,$C3,$C5,$1A,$AE,$18,$FA,$6B,$53,$E3,$9C,$D7,$0A
	.dc.b	$D6,$1E,$39,$CD,$6C,$EF,$19,$7B,$12,$78,$BC,$0F,$8E,$FC,$9C,$7E
	.dc.b	$06,$1E,$29,$CE,$1E,$12,$3F,$5E,$B8,$63,$EF,$6E,$3F,$12,$7C,$15
	.dc.b	$C1,$E6,$71,$71,$87,$CF,$3C,$38,$FA,$6F,$9D,$72,$38,$1E,$F7,$24
	.dc.b	$AF,$D6,$4F,$C1,$9F,$DF,$9F,$14,$A7,$E1,$8B,$5F,$9B,$6F,$80,$F9
	.dc.b	$6D,$C7,$E6,$23,$0E,$27,$FD,$4D,$3A,$8F,$D5,$CE,$9E,$71,$FA,$88
	.dc.b	$E3,$87,$E5,$FE,$BF,$FC,$59,$BE,$6A,$5F,$E6,$6F,$79,$7A,$AF,$73
	.dc.b	$9A,$DE,$1F,$66,$F7,$3F,$C7,$AB,$73,$9A,$DD,$FE,$F2,$F7,$6B,$DC
	.dc.b	$E6,$B7,$3F,$05,$FE,$11,$3A,$8C,$1E,$32,$5F,$B7,$A7,$FC,$78,$F8
	.dc.b	$63,$37,$D5,$FD,$FD,$73,$7E,$4D,$F8,$1A,$6A,$F9,$D7,$1B,$E8,$47
	.dc.b	$E3,$46,$A7,$CB,$39,$6C,$EB,$1E,$D9,$7D,$A9,$F2,$9F,$26,$F9,$3E
	.dc.b	$2B,$E4,$DB,$E4,$F0,$30,$76,$39,$5F,$6C,$BE,$7E,$1D,$EF,$E3,$4E
	.dc.b	$71,$F8,$19,$7C,$3A,$D6,$3E,$F2,$63,$D0,$F0,$46,$9F,$01,$F6,$F2
	.dc.b	$F4,$C6,$63,$CA,$4A,$ED,$97,$E0,$97,$D8,$7C,$07,$EA,$18,$CB,$F0
	.dc.b	$38,$66,$74,$D8,$C9,$FD,$37,$E8,$BF,$A5,$3F,$DA,$3C,$1F,$F3,$F7
	.dc.b	$37,$F9,$F2,$73,$92,$3F,$B5,$F2,$FF,$AA,$AF,$FC,$B7,$5A,$9E,$1A
	.dc.b	$F2,$9F,$D2,$36,$FF,$BE,$D6,$E7,$F5,$1E,$06,$67,$F3,$D5,$74,$FF
	.dc.b	$69,$D7,$F9,$A1,$E1,$3F,$A9,$AD,$56,$DB,$5F,$E8,$1F,$AD,$7C,$8B
	.dc.b	$D5,$F1,$46,$3E,$C2,$CE,$6B,$82,$8C,$D6,$9B,$3A,$78,$F8,$BC,$BE
	.dc.b	$0A,$CF,$26,$E7,$F2,$63,$FC,$3B,$8F,$92,$73,$B7,$FA,$B6,$31,$3F
	.dc.b	$E8,$6F,$AF,$7C,$7F,$C9,$A7,$A6,$FB,$0F,$91,$5A,$BF,$F5,$05,$6A
	.dc.b	$73,$7C,$6F,$85,$75,$8C,$4E,$A3,$07,$FA,$A7,$E5,$5F,$11,$F4,$B4
	.dc.b	$F5,$1F,$E3,$6C,$BC,$0E,$33,$F9,$F7,$F3,$3E,$23,$E6,$B4,$F9,$A5
	.dc.b	$9E,$1C,$3F,$ED,$78,$FA,$33,$F5,$A5,$F6,$14,$C7,$BD,$6B,$EF,$4E
	.dc.b	$9B,$3A,$69,$FE,$B0,$FA,$E7,$C1,$7E,$4F,$57,$D8,$3E,$C3,$F2,$55
	.dc.b	$AB,$F8,$53,$97,$95,$F1,$8C,$63,$F1,$05,$71,$7E,$3E,$1D,$B0,$EB
	.dc.b	$73,$0F,$F6,$B7,$D1,$BE,$C1,$F0,$79,$C7,$E6,$0F,$43,$E2,$1F,$17
	.dc.b	$CD,$B8,$7C,$28,$B3,$C2,$6F,$47,$E3,$63,$FC,$DB,$5F,$23,$C7,$F9
	.dc.b	$D1,$F3,$DB,$19,$3E,$6C,$65,$FF,$6C,$F7,$7F,$01,$8F,$C1,$CB,$EB
	.dc.b	$DF,$61,$FB,$BA,$D6,$0C,$3F,$CF,$B7,$BA,$D7,$8B,$FC,$16,$31,$3D
	.dc.b	$A3,$18,$74,$8C,$13,$1F,$0C,$7F,$BA,$E3,$18,$77,$9F,$07,$D8,$3E
	.dc.b	$43,$F8,$95,$23,$0F,$FA,$42,$B8,$39,$E1,$DC,$EA,$FF,$43,$3C,$6F
	.dc.b	$26,$AA,$C7,$33,$E7,$D3,$EB,$B9,$FC,$97,$46,$CE,$F0,$F3,$C6,$1E
	.dc.b	$F5,$C6,$74,$7A,$6B,$F9,$47,$D7,$BE,$43,$F9,$77,$CC,$BE,$61,$F6
	.dc.b	$1A,$9C,$C6,$30,$F0,$64,$C3,$94,$78,$F0,$E5,$87,$18,$C5,$F1,$BF
	.dc.b	$D0,$5F,$06,$E1,$F6,$3B,$F8,$47,$E9,$7C,$D7,$AB,$FC,$FF,$EE,$FF
	.dc.b	$3F,$97,$F4,$A8,$C7,$57,$F4,$3F,$B1,$F8,$18,$7E,$B5,$7C,$0E,$78
	.dc.b	$B4,$FF,$49,$FA,$67,$D8,$E7,$2F,$6C,$61,$F7,$DF,$4E,$7F,$C3,$62
	.dc.b	$A7,$D0,$7F,$57,$F7,$DF,$08,$FC,$0B,$B4,$E6,$31,$97,$CA,$F7,$5F
	.dc.b	$63,$A8,$C3,$72,$FE,$0D,$3A,$9C,$9F,$81,$C5,$F5,$CF,$B2,$FE,$C9
	.dc.b	$5B,$BF,$E1,$D6,$F9,$7F,$CD,$FD,$97,$D1,$DF,$D7,$39,$3F,$0F,$19
	.dc.b	$7F,$CD,$E6,$6B,$F3,$71,$97,$F8,$2D,$74,$E9,$FE,$E9,$5B,$BE,$B3
	.dc.b	$FA,$AF,$B1,$FF,$7A,$CF,$E2,$8D,$1F,$E7,$CF,$2F,$FB,$66,$1F,$D2
	.dc.b	$35,$F6,$FD,$97,$F7,$F9,$CE,$3F,$EF,$6D,$93,$9F,$77,$E9,$46,$31
	.dc.b	$FC,$BD,$F0,$58,$B1,$EB,$1A,$BE,$4F,$CD,$BF,$EA,$F8,$FE,$03,$0E
	.dc.b	$58,$72,$EA,$FC,$0D,$FE,$BB,$37,$D6,$FF,$85,$67,$35,$FC,$0A,$73
	.dc.b	$5F,$AA,$9C,$E1,$FC,$2A,$73,$5F,$D1,$67,$35,$FC,$CA,$73,$5F,$92
	.dc.b	$9C,$D7,$8F,$93,$E7,$79,$3E,$17,$C5,$E4,$FC,$4F,$B2,$B9,$7B,$5B
	.dc.b	$7F,$3F,$54,$63,$2F,$9C,$DA,$FF,$58,$E5,$FC,$D9,$F9,$7E,$2B,$EC
	.dc.b	$BF,$90,$5F,$97,$2F,$F1,$B7,$A4,$FA,$5F,$F0,$A7,$D8,$3E,$CB,$31
	.dc.b	$86,$E5,$FE,$14,$9D,$3F,$C4,$2B,$93,$99,$DE,$DF,$DA,$E7,$2F,$F4
	.dc.b	$EF,$B2,$FE,$03,$5C,$70,$FE,$01,$F3,$5E,$2F,$94,$B3,$DD,$B7,$FE
	.dc.b	$18,$D1,$9A,$73,$1F,$E6,$08,$C5,$7F,$24,$AF,$C6,$F6,$FE,$3B,$EC
	.dc.b	$BF,$B9,$47,$E4,$34,$FB,$B1,$88,$FF,$6D,$CE,$FD,$7F,$F6,$ED,$69
	.dc.b	$FF,$6B,$C6,$67,$97,$47,$F5,$7B,$E5,$1E,$00,$F5,$BE,$51,$FE,$E6
	.dc.b	$AE,$4A,$32,$F8,$8E,$37,$E2,$BE,$CF,$8A,$32,$78,$B0,$E6,$F8,$70
	.dc.b	$F0,$57,$08,$F0,$1C,$0F,$03,$97,$C1,$1C,$27,$9C,$30,$F3,$E1,$E6
	.dc.b	$8E,$D8,$79,$6B,$75,$AC,$3A,$56,$8F,$31,$E1,$30,$78,$A7,$8B,$E5
	.dc.b	$BE,$CF,$84,$C4,$F4,$71,$87,$9D,$E8,$6C,$FE,$1A,$7C,$02,$9F,$34
	.dc.b	$A7,$D7,$33,$F6,$5F,$CD,$7D,$FC,$C2,$30,$F0,$6B,$C3,$BF,$FD,$2E
	.dc.b	$E7,$E8,$FF,$D6,$0D,$DB,$9D,$27,$33,$E0,$6B,$E8,$8D,$4F,$E1,$BE
	.dc.b	$C9,$EA,$7B,$32,$FC,$C4,$62,$FF,$17,$7C,$1F,$BC,$D9,$DD,$71,$DF
	.dc.b	$E2,$DF,$1B,$DE,$3E,$8B,$FD,$26,$1C,$6F,$78,$71,$BF,$E7,$38,$71
	.dc.b	$BD,$E1,$C6,$FD,$4D,$AD,$DF,$15,$7C,$6B,$78,$FA,$AF,$85,$6A,$F8
	.dc.b	$5E,$E3,$FA,$4E,$3E,$56,$D7,$09,$D4,$71,$3A,$5C,$7F,$45,$1A,$37
	.dc.b	$5C,$E3,$F0,$15,$CE,$FC,$27,$A6,$74,$6E,$B9,$46,$67,$E0,$7E,$20
	.dc.b	$FC,$5F,$FC,$25,$87,$F9,$1F,$EF,$1E,$13,$ED,$BF,$2F,$A7,$F9,$86
	.dc.b	$73,$18,$E3,$FE,$CF,$8F,$3F,$33,$18,$A3,$0F,$5C,$7A,$36,$F7,$3A
	.dc.b	$8F,$F6,$1A,$FC,$5F,$FD,$5B,$1E,$5E,$6F,$EA,$2D,$F0,$78,$CE,$A3
	.dc.b	$8B,$74,$F4,$57,$09,$FF,$67,$E9,$FC,$26,$F8,$FB,$72,$7A,$10,$E6
	.dc.b	$F9,$BE,$2D,$E6,$7B,$AF,$FC,$0D,$D3,$7C,$AF,$EF,$B8,$7F,$9C,$75
	.dc.b	$FF,$77,$DF,$13,$55,$27,$58,$74,$6C,$E5,$F3,$DF,$47,$E2,$CE,$DF
	.dc.b	$8F,$F6,$FF,$EB,$E7,$ED,$7D,$57,$F9,$39,$BE,$8F,$CE,$9D,$47,$0F
	.dc.b	$5F,$D5,$5B,$8F,$8A,$DE,$0F,$F5,$06,$8B,$E3,$18,$6F,$2F,$F8,$A3
	.dc.b	$65,$F1,$8C,$37,$0F,$C2,$57,$07,$F0,$DF,$89,$7F,$9E,$9F,$C9,$76
	.dc.b	$FF,$94,$DB,$DD,$F8,$69,$D4,$7E,$19,$B9,$7E,$56,$B8,$4F,$DE,$6C
	.dc.b	$7A,$A7,$78,$FE,$61,$B4,$CB,$FB,$2E,$1F,$77,$0E,$37,$BF,$2B,$C5
	.dc.b	$EF,$7E,$07,$EF,$1F,$5D,$4C,$E5,$F8,$1C,$FF,$D4,$EF,$EB,$67,$6D
	.dc.b	$9D,$BE,$BA,$E4,$E4,$E4,$FE,$9B,$F1,$3F,$F0,$74,$7E,$35,$7D,$5E
	.dc.b	$4D,$79,$16,$7F,$79,$F7,$EF,$AE,$FC,$C3,$63,$95,$62,$73,$F2,$D2
	.dc.b	$78,$C7,$ED,$B1,$FF,$79,$C7,$F0,$7E,$BF,$F7,$13,$7E,$1F,$5F,$13
	.dc.b	$FA,$59,$FD,$97,$0F,$E1,$7F,$00,$FC,$17,$29,$B3,$AC,$3E,$97,$C3
	.dc.b	$F9,$5F,$DB,$AF,$87,$FD,$83,$63,$F8,$C3,$BE,$8F,$F5,$8F,$CB,$FD
	.dc.b	$C9,$3F,$94,$7F,$DA,$AD,$51,$C3,$9F,$F8,$CC,$70,$6C,$EE,$7A,$F5
	.dc.b	$E2,$65,$74,$8D,$98,$7F,$8D,$F9,$3E,$5D,$7F,$D1,$BE,$0B,$F4,$1F
	.dc.b	$C8,$B7,$0F,$E4,$75,$BB,$FE,$54,$D9,$D4,$7F,$AF,$6B,$F3,$3D,$FF
	.dc.b	$CE,$F9,$3F,$8A,$7C,$8B,$E8,$EF,$D0,$CC,$3F,$9F,$4E,$5B,$39,$7F
	.dc.b	$92,$C2,$79,$CF,$C4,$FD,$09,$F2,$F4,$97,$FF,$00,$3E,$C3,$EA,$3F
	.dc.b	$33,$7F,$F0,$FD,$70,$8F,$F5,$6F,$0F,$FB,$9A,$31,$1B,$3F,$D6,$2E
	.dc.b	$23,$FD,$D7,$5D,$1E,$17,$F5,$A3,$CD,$5E,$1C,$7C,$6F,$AB,$C6,$F8
	.dc.b	$64,$7F,$9A,$7E,$6F,$FC,$03,$A5,$AF,$F1,$36,$D6,$EB,$FA,$9D,$6E
	.dc.b	$9E,$D7,$CE,$24,$F4,$E5,$AF,$CA,$8B,$A7,$F2,$AB,$FC,$56,$DF,$E4
	.dc.b	$2F,$B0,$7D,$BF,$E4,$17,$E0,$34,$C4,$4C,$7E,$4D,$71,$9D,$46,$0F
	.dc.b	$F6,$CD,$74,$70,$1C,$3F,$3C,$C4,$72,$71,$1A,$C3,$FC,$01,$9B,$E3
	.dc.b	$3F,$1B,$E1,$4B,$FC,$5C,$7E,$40,$C3,$F8,$A8,$C3,$E5,$E8,$7D,$4F
	.dc.b	$F8,$F7,$06,$7F,$A4,$8C,$37,$D9,$FF,$05,$D7,$0F,$26,$D6,$B0,$FF
	.dc.b	$7B,$36,$7A,$D6,$AB,$F7,$7F,$5E,$F9,$CB,$F2,$93,$1F,$8B,$39,$6C
	.dc.b	$E5,$F8,$D0,$9E,$73,$27,$8C,$6F,$3F,$5C,$05,$75,$7F,$5B,$F1,$BF
	.dc.b	$B6,$23,$F9,$4F,$9A,$AE,$8F,$DB,$C7,$F8,$9B,$91,$3F,$BE,$FA,$29
	.dc.b	$5F,$BC,$75,$F4,$B8,$8F,$DD,$C5,$C7,$F2,$53,$CF,$3F,$E0,$FF,$1D
	.dc.b	$2F,$F4,$B1,$97,$F8,$3C,$65,$FD,$1C,$7F,$0E,$F9,$6F,$F1,$17,$D1
	.dc.b	$7E,$AC,$FD,$33,$99,$FF,$26,$46,$7C,$7C,$47,$E4,$7E,$AD,$F5,$9F
	.dc.b	$ED,$EC,$BF,$D4,$B1,$F8,$69,$FF,$73,$C6,$39,$75,$1F,$36,$78,$BA
	.dc.b	$31,$5A,$51,$CD,$FF,$64,$57,$37,$81,$A9,$E1,$F5,$9F,$CA,$70,$CB
	.dc.b	$67,$89,$F7,$5C,$7C,$0A,$FA,$4E,$84,$EA,$3F,$D8,$2D,$8E,$5E,$3F
	.dc.b	$F8,$86,$B9,$47,$B7,$2F,$1C,$E6,$F8,$B6,$78,$E1,$EA,$CB,$DD,$87
	.dc.b	$29,$F0,$D6,$DB,$5E,$C8,$C5,$70,$BE,$B3,$92,$57,$AE,$74,$F8,$0F
	.dc.b	$76,$3E,$B7,$A9,$BC,$7D,$35,$CB,$1F,$25,$6E,$F9,$DF,$09,$CE,$3E
	.dc.b	$7A,$D6,$1C,$AB,$BC,$E4,$F4,$C6,$6B,$8E,$1D,$F1,$FA,$71,$8B,$E3
	.dc.b	$39,$7E,$71,$4F,$14,$FC,$27,$D4,$67,$0F,$71,$E4,$71,$87,$8E,$BE
	.dc.b	$96,$1E,$3B,$E5,$87,$84,$FA,$98,$FC,$0C,$79,$E1,$C7,$1E,$E7,$98
	.dc.b	$95,$C6,$FD,$67,$42,$47,$D3,$8E,$B8,$7C,$78,$FE,$24,$7D,$72,$9F
	.dc.b	$58,$B7,$F9,$D6,$F8,$8F,$94,$DC,$7E,$7E,$30,$E2,$7F,$C0,$D3,$A8
	.dc.b	$FC,$FC,$E9,$E7,$1F,$AA,$8E,$37,$F9,$5F,$A9,$7D,$5F,$F2,$88,$FC
	.dc.b	$2C,$EB,$2F,$D3,$B7,$0F,$16,$1F,$B8,$C3,$8F,$6F,$58,$93,$F9,$4A
	.dc.b	$E0,$AB,$6E,$67,$78,$78,$1E,$98,$F8,$F1,$F0,$C6,$9B,$5F,$BD,$BE
	.dc.b	$77,$C6,$3A,$11,$F7,$5F,$38,$F4,$99,$AE,$73,$AC,$7B,$4E,$A3,$A1
	.dc.b	$97,$81,$99,$E2,$7D,$EB,$F1,$B8,$8D,$10,$F8,$57,$E3,$7E,$BE,$1E
	.dc.b	$06,$9E,$D8,$D6,$5F,$7A,$3C,$26,$B0,$F2,$1E,$6B,$F0,$46,$5F,$25
	.dc.b	$72,$CB,$E9,$DF,$3B,$E1,$7F,$D9,$F1,$F4,$BE,$79,$D6,$1D,$27,$33
	.dc.b	$D4,$D5,$FD,$E8,$CD,$70,$9C,$E3,$F7,$32,$F7,$4E,$67,$91,$F1,$23
	.dc.b	$E1,$53,$EB,$3E,$33,$E0,$31,$1C,$E7,$F0,$53,$AC,$F6,$D9,$FE,$A9
	.dc.b	$E4,$3F,$40,$FD,$64,$7F,$8D,$E3,$17,$FE,$08,$E7,$FF,$04,$C6,$F3
	.dc.b	$C5,$7F,$BB,$6F,$FC,$18,$F1,$79,$DF,$F9,$A6,$B9,$CE,$6B,$8C,$FF
	.dc.b	$B7,$6B,$8B,$FE,$5C,$ED,$CD,$F9,$7D,$4D,$F9,$EE,$DF,$F1,$06,$5D
	.dc.b	$E7,$31,$8C,$3C,$18,$A7,$38,$F4,$8C,$5F,$9F,$C7,$C1,$CB,$C2,$B1
	.dc.b	$92,$61,$FE,$70,$8C,$BD,$BE,$3B,$A1,$E2,$F0,$ED,$7C,$38,$B3,$FE
	.dc.b	$78,$9B,$BF,$F8,$D3,$EA,$5F,$20,$FB,$2F,$8B,$18,$F3,$EC,$E7,$2F
	.dc.b	$F3,$64,$E5,$B1,$9C,$7F,$D3,$13,$96,$C6,$70,$F0,$2B,$EB,$87,$66
	.dc.b	$E1,$D1,$FA,$FE,$43,$F5,$77,$C3,$E1,$B5,$FE,$6B,$5F,$03,$99,$93
	.dc.b	$97,$84,$63,$3F,$AF,$F2,$BF,$57,$EC,$BE,$B3,$6F,$F1,$8D,$9D,$C7
	.dc.b	$E3,$E7,$E3,$C3,$E3,$EF,$CC,$61,$B5,$F2,$27,$4D,$9D,$34,$D4,$F0
	.dc.b	$9F,$96,$70,$2C,$FE,$2F,$09,$5F,$EB,$AF,$93,$B5,$F9,$29,$BC,$1F
	.dc.b	$EA,$F9,$B5,$FA,$9B,$FC,$A1,$F9,$AF,$8B,$B1,$FA,$B8,$B8,$7F,$B0
	.dc.b	$26,$DF,$F9,$F2,$3F,$CE,$3E,$9F,$EB,$3B,$9F,$50,$F5,$3F,$5F,$5A
	.dc.b	$CB,$FD,$69,$5A,$9C,$FA,$5F,$EA,$B9,$FF,$5A,$B6,$BF,$DC,$55,$D3
	.dc.b	$EB,$B7,$8C,$BD,$BE,$1C,$53,$9B,$FD,$9F,$C9,$7E,$C7,$DD,$FD,$93
	.dc.b	$70,$FE,$10,$D9,$DC,$7F,$0B,$9F,$DE,$FE,$0B,$FB,$9F,$91,$B3,$FC
	.dc.b	$3D,$B5,$FC,$01,$BB,$7F,$18,$F0,$51,$FC,$42,$9F,$55,$FC,$1E,$BF
	.dc.b	$D4,$B1,$8F,$07,$53,$98,$FE,$3A,$A3,$3E,$1E,$9B,$3F,$ED,$F3,$91
	.dc.b	$67,$96,$3F,$EC,$C7,$94,$66,$B9,$3F,$CC,$7F,$0A,$7C,$67,$F2,$25
	.dc.b	$F6,$F6,$30,$FF,$26,$F8,$B7,$D6,$FF,$28,$F8,$BF,$F6,$CE,$1F,$CE
	.dc.b	$E7,$2F,$F3,$D8,$CE,$6D,$7F,$D8,$BE,$67,$83,$DB,$E9,$E5,$7F,$30
	.dc.b	$D1,$18,$C7,$FA,$37,$C8,$3E,$3B,$F9,$BF,$9C,$F3,$7F,$9C,$CE,$6B
	.dc.b	$FA,$64,$E5,$FF,$69,$67,$E0,$C5,$F4,$1C,$31,$F0,$36,$B6,$7F,$B6
	.dc.b	$74,$F8,$BE,$D6,$A3,$3B,$27,$31,$87,$E5,$D7,$F4,$AF,$92,$79,$1F
	.dc.b	$D3,$BF,$11,$FE,$AF,$A7,$C2,$9C,$E1,$FD,$8E,$72,$F8,$63,$38,$FF
	.dc.b	$5D,$9C,$BE,$18,$CD,$7F,$68,$9C,$E2,$9C,$DF,$FB,$CF,$D4,$56,$A7
	.dc.b	$38,$27,$3E,$1F,$F6,$2F,$95,$FE,$FF,$3F,$DB,$7F,$12,$DE,$49,$CE
	.dc.b	$7F,$DE,$27,$31,$8C,$FC,$11,$8F,$5F,$FB,$7B,$63,$FC,$04,$DC,$3B
	.dc.b	$6B,$FD,$CB,$E6,$1F,$0F,$FD,$DB,$F1,$37,$C8,$E6,$30,$DC,$7F,$C1
	.dc.b	$93,$A7,$EA,$57,$27,$33,$BC,$3F,$C1,$15,$B7,$F2,$D3,$F8,$1A,$E1
	.dc.b	$1A,$AF,$F0,$2F,$86,$FA,$AF,$F0,$4F,$AD,$70,$FC,$14,$DB,$FF,$0A
	.dc.b	$4D,$DF,$F4,$33,$6B,$EA,$4D,$AF,$C6,$4D,$E2,$4E,$78,$7F,$86,$E3
	.dc.b	$E0,$3C,$65,$C3,$C3,$87,$F8,$73,$EB,$3F,$48,$FB,$0E,$A6,$7A,$3D
	.dc.b	$6D,$9E,$27,$F8,$D6,$78,$E7,$E8,$9D,$1C,$F0,$FE,$3E,$E2,$3F,$7B
	.dc.b	$F3,$CF,$83,$FF,$16,$E1,$C8,$FD,$64,$65,$FF,$21,$19,$AF,$D6,$46
	.dc.b	$5F,$F1,$E4,$66,$D3,$9B,$FF,$23,$CF,$1A,$FF,$5F,$FC,$21,$7C,$E3
	.dc.b	$C2,$FF,$61,$FD,$1F,$F8,$D2,$73,$1F,$EB,$EA,$E1,$5B,$9E,$55,$B8
	.dc.b	$C4,$76,$38,$9A,$F0,$18,$72,$C3,$A4,$E8,$5F,$38,$CD,$F4,$8C,$4F
	.dc.b	$FB,$5B,$0E,$4F,$5B,$AF,$86,$B8,$61,$E0,$8E,$C7,$C1,$AD,$DF,$48
	.dc.b	$E1,$87,$C2,$AE,$17,$EC,$9E,$07,$9E,$B8,$38,$8E,$F8,$78,$A3,$17
	.dc.b	$CE,$31,$0C,$3C,$38,$76,$AF,$79,$B9,$F3,$9B,$3C,$11,$B7,$E1,$18
	.dc.b	$AE,$B5,$D1,$B1,$93,$9C,$7F,$BC,$DF,$59,$FE,$F5,$70,$F7,$3F,$0C
	.dc.b	$7E,$1F,$E8,$DE,$B7,$F9,$9A,$3F,$E0,$0C,$63,$E0,$3F,$DF,$CD,$8C
	.dc.b	$CE,$DE,$62,$1E,$CF,$F6,$18,$E6,$D7,$71,$AD,$7A,$CE,$67,$CB,$39
	.dc.b	$7F,$15,$F4,$7F,$E0,$57,$F0,$6D,$7F,$1D,$1B,$D7,$F2,$ED,$D7,$DF
	.dc.b	$5B,$9F,$CD,$B6,$3B,$DF,$9E,$78,$B6,$38,$FA,$1F,$22,$78,$B6,$38
	.dc.b	$E3,$F9,$6C,$3C,$18,$FE,$CF,$1E,$CD,$BE,$0F,$D1,$C7,$B3,$6F,$DB
	.dc.b	$3A,$85,$72,$7E,$75,$70,$6D,$70,$61,$F1,$23,$57,$E4,$8E,$77,$D6
	.dc.b	$FA,$4E,$4D,$3D,$CC,$D7,$23,$C0,$E6,$7C,$D3,$F0,$38,$5F,$34,$FA
	.dc.b	$3F,$C7,$E5,$FE,$DE,$FA,$3F,$F4,$56,$7F,$EE,$36,$BF,$CB,$6B,$83
	.dc.b	$FE,$0D,$6B,$FE,$DF,$8D,$CF,$3F,$7F,$AF,$17,$F8,$66,$F7,$3F,$CD
	.dc.b	$3E,$8F,$38,$7F,$7F,$E3,$7B,$38,$ED,$FD,$C3,$AD,$7F,$97,$DD,$E8
	.dc.b	$3B,$BE,$38,$7F,$BE,$DA,$FF,$45,$9D,$BF,$E2,$56,$BF,$EF,$78,$DE
	.dc.b	$3F,$F0,$01,$B9,$FF,$46,$BF,$17,$73,$E8,$FC,$DD,$FF,$EE,$FC,$7F
	.dc.b	$C6,$7A,$FF,$66,$EA,$DD,$49,$CD,$6D,$FE,$E7,$39,$7F,$AD,$5F,$57
	.dc.b	$71,$D8,$D6,$7F,$E5,$8F,$07,$FD,$B5,$3E,$0F,$9B,$FF,$6A,$B7,$E1
	.dc.b	$3E,$44,$EA,$3F,$B3,$36,$BF,$9B,$66,$BE,$35,$FD,$D6,$31,$7C,$BE
	.dc.b	$1A,$4F,$52,$1D,$1F,$AD,$E5,$3E,$47,$ED,$77,$7F,$1A,$AF,$8A,$D7
	.dc.b	$FB,$C4,$ED,$FF,$2C,$35,$F1,$46,$F2,$F3,$B7,$87,$FC,$C9,$5B,$9F
	.dc.b	$F3,$DB,$63,$FC,$9D,$3B,$AF,$F2,$63,$79,$3F,$C9,$18,$7A,$3D,$1F
	.dc.b	$C8,$7D,$1F,$F9,$EE,$FE,$DB,$5F,$3C,$6D,$CF,$A9,$C0,$C3,$A3,$24
	.dc.b	$30,$7F,$85,$FD,$07,$C3,$7E,$75,$7D,$1D,$31,$16,$7F,$D2,$3E,$67
	.dc.b	$E9,$7D,$6F,$F2,$8C,$7F,$9F,$EF,$8B,$A3,$FD,$2F,$F0,$DB,$38,$57
	.dc.b	$EB,$79,$DF,$47,$49,$5F,$E8,$9D,$3F,$D3,$CF,$FA,$6B,$B3,$F6,$9E
	.dc.b	$DF,$E1,$B9,$7F,$12,$DE,$4F,$F3,$E6,$3F,$AC,$9F,$C6,$FB,$7E,$6C
	.dc.b	$3F,$18,$BE,$8F,$F7,$CD,$8F,$F5,$B6,$9F,$C1,$17,$D1,$FD,$B7,$FC
	.dc.b	$54,$FE,$5D,$FF,$50,$9A,$EC,$E1,$1B,$8E,$3C,$D8,$99,$1C,$27,$44
	.dc.b	$8C,$C7,$F9,$54,$E2,$7F,$1F,$AF,$E2,$FD,$EF,$A3,$F2,$37,$0F,$E2
	.dc.b	$DE,$2F,$BF,$4F,$F6,$6B,$76,$FF,$21,$56,$F1,$FF,$5C,$B5,$FE,$5D
	.dc.b	$8F,$F2,$1F,$9E,$7B,$D7,$E5,$A5,$7F,$0E,$9C,$B6,$72,$FE,$E6,$13
	.dc.b	$CE,$7E,$67,$D1,$BE,$26,$92,$FE,$5F,$D1,$FD,$37,$E1,$4E,$DC,$DF
	.dc.b	$D5,$9E,$B1,$F4,$A7,$4F,$0A,$D5,$48,$D4,$7B,$7C,$39,$1E,$B7,$F9
	.dc.b	$F7,$CA,$FD,$FF,$19,$9F,$F1,$7D,$BF,$D9,$0D,$BF,$F6,$33,$7C,$35
	.dc.b	$F1,$87,$D1,$FC,$B2,$57,$FB,$96,$7F,$14,$FC,$78,$CB,$4D,$72,$FE
	.dc.b	$4F,$E8,$9F,$47,$7E,$62,$57,$E4,$A7,$2D,$9C,$BF,$8B,$84,$F3,$99
	.dc.b	$3C,$63,$7C,$DF,$E6,$98,$57,$47,$EE,$7C,$97,$E1,$7C,$29,$1F,$83
	.dc.b	$8F,$C4,$18,$7F,$05,$18,$7F,$1F,$F3,$9F,$E3,$5B,$FF,$3B,$74,$23
	.dc.b	$1C,$1F,$96,$6D,$7F,$B0,$1F,$E4,$55,$C1,$CD,$71,$F5,$FF,$4B,$18
	.dc.b	$37,$7F,$E5,$CF,$A3,$7D,$0F,$EF,$57,$C8,$53,$11,$32,$FD,$EC,$E9
	.dc.b	$B3,$97,$F6,$D1,$D1,$C9,$0C,$CF,$EA,$FE,$1A,$47,$EA,$5F,$C7,$FC
	.dc.b	$B6,$23,$F5,$11,$74,$FF,$4D,$73,$7E,$32,$BF,$69,$F4,$72,$FF,$64
	.dc.b	$FE,$63,$E1,$31,$1F,$B1,$8B,$7C,$B9,$FF,$DA,$53,$F9,$1C,$BF,$7D
	.dc.b	$F4,$72,$30,$43,$F7,$4E,$5F,$F6,$F4,$FB,$DF,$DF,$7C,$87,$F8,$9F
	.dc.b	$E8,$FF,$50,$7E,$85,$FE,$14,$67,$C5,$93,$FC,$26,$3F,$97,$7D,$13
	.dc.b	$E8,$FF,$DA,$FA,$FF,$A4,$A3,$F0,$95,$FE,$B9,$8C,$72,$6A,$3F,$DB
	.dc.b	$93,$C5,$D1,$8A,$D2,$8E,$6F,$FA,$CE,$B9,$BC,$0D,$4F,$0F,$A3,$FE
	.dc.b	$43,$8E,$5B,$3C,$4F,$F7,$63,$8F,$7B,$E6,$C7,$7C,$79,$78,$FF,$E0
	.dc.b	$D9,$CD,$6F,$1F,$F5,$7B,$6F,$D7,$7C,$30,$F0,$61,$C6,$F6,$DA,$D9
	.dc.b	$A9,$8F,$94,$5F,$57,$C8,$6B,$0F,$1D,$73,$8D,$B7,$2F,$8F,$3A,$BE
	.dc.b	$18,$78,$9F,$91,$3A,$8C,$5F,$87,$0F,$01,$E1,$9D,$47,$C8,$C3,$C1
	.dc.b	$3A,$7C,$18,$FC,$19,$E5,$1A,$32,$F8,$8E,$58,$57,$DA,$70,$71,$5E
	.dc.b	$1A,$E9,$8F,$AE,$B4,$7C,$7C,$3C,$AF,$C2,$BE,$38,$78,$23,$75,$C4
	.dc.b	$98,$FE,$2A,$32,$FC,$28,$CC,$7D,$63,$27,$A0,$A7,$A1,$DB,$E1,$33
	.dc.b	$7E,$4C,$7E,$36,$1E,$F8,$C6,$1F,$E5,$7A,$FA,$27,$FB,$C2,$BF,$2E
	.dc.b	$D8,$D3,$F9,$8F,$15,$F2,$5F,$EF,$68,$C3,$5F,$C4,$C6,$9D,$57,$F7
	.dc.b	$C9,$D3,$63,$F0,$13,$BC,$7F,$47,$F4,$6F,$A3,$75,$FE,$AE,$B5,$AF
	.dc.b	$E8,$9B,$87,$88,$FC,$D5,$F1,$AE,$19,$49,$EE,$FE,$E6,$75,$18,$26
	.dc.b	$3F,$B0,$BE,$B0,$9D,$CE,$6B,$A3,$67,$C3,$8F,$A1,$E7,$8F,$9E,$B6
	.dc.b	$A2,$5F,$48,$D4,$70,$3D,$97,$F1,$2F,$A3,$DC,$93,$C1,$D6,$3D,$AB
	.dc.b	$8C,$E4,$F0,$3E,$BA,$E7,$19,$8D,$90,$D4,$F3,$7E,$EE,$17,$0F,$44
	.dc.b	$6F,$2F,$B9,$8F,$92,$3D,$B8,$F6,$75,$97,$CC,$AD,$E5,$E4,$C3,$9D
	.dc.b	$76,$3E,$11,$70,$F6,$B8,$AE,$2F,$9E,$78,$C6,$23,$85,$7B,$5F,$59
	.dc.b	$C6,$FC,$47,$DE,$AE,$2E,$31,$F7,$63,$E9,$AE,$67,$D8,$F9,$4C,$5F
	.dc.b	$A0,$38,$9C,$B7,$3F,$C8,$36,$BF,$A2,$78,$EF,$9B,$FD,$6D,$7F,$BD
	.dc.b	$1F,$C3,$37,$1F,$C6,$74,$7F,$B8,$A7,$37,$FE,$E1,$78,$B4,$DD,$7E
	.dc.b	$3E,$FB,$5E,$25,$1E,$17,$1D,$7F,$E5,$4E,$ED,$4F,$97,$D3,$CF,$4F
	.dc.b	$F8,$8B,$1F,$34,$62,$B7,$92,$31,$3A,$3D,$9E,$2E,$8E,$7E,$0D,$8C
	.dc.b	$93,$3F,$F0,$4C,$65,$F7,$7C,$67,$63,$C7,$EE,$DA,$F9,$11,$67,$FC
	.dc.b	$23,$37,$7F,$DA,$7D,$1B,$E3,$DF,$47,$F0,$E3,$1E,$7D,$9C,$E5,$FB
	.dc.b	$7B,$E3,$09,$D5,$ED,$B5,$B3,$F7,$D7,$D1,$C6,$28,$F1,$57,$19,$CC
	.dc.b	$61,$B7,$DD,$FF,$58,$FD,$1B,$E1,$7E,$D7,$94,$8C,$4E,$BE,$07,$31
	.dc.b	$8C,$9A,$7F,$9A,$A3,$0D,$C3,$75,$C1,$FB,$7F,$47,$F7,$3E,$62,$E3
	.dc.b	$F9,$56,$E3,$F9,$16,$E1,$F9,$57,$E6,$43,$E3,$6F,$D1,$E1,$B7,$F0
	.dc.b	$27,$4D,$9D,$34,$D4,$F0,$9F,$5C,$7C,$E7,$89,$83,$E7,$09,$FC,$AE
	.dc.b	$52,$7F,$D5,$9F,$27,$6B,$F5,$53,$71,$FF,$31,$4D,$AF,$D5,$73,$F3
	.dc.b	$7F,$6D,$87,$E9,$3E,$53,$F4,$2C,$8C,$9F,$EB,$98,$C6,$3F,$EF,$7F
	.dc.b	$3D,$E9,$FE,$CF,$BD,$F4,$6F,$53,$F6,$D6,$BD,$E3,$F8,$AB,$DD,$6A
	.dc.b	$73,$5F,$06,$73,$5F,$90,$8E,$58,$77,$C3,$B4,$E6,$F8,$EB,$FE,$BB
	.dc.b	$6C,$7F,$57,$C3,$CB,$F4,$7D,$F0,$FD,$FF,$C9,$7F,$00,$F7,$2D,$FF
	.dc.b	$19,$6E,$7F,$C5,$5B,$97,$F1,$97,$F8,$5F,$D1,$FF,$08,$F9,$1B,$7F
	.dc.b	$C8,$5B,$7F,$C7,$5B,$7F,$C8,$5F,$E3,$52,$FA,$3F,$E2,$95,$F8,$18
	.dc.b	$C7,$83,$A9,$CD,$FF,$75,$9C,$9F,$B4,$39,$78,$9B,$AE,$0D,$8D,$57
	.dc.b	$27,$F9,$AA,$9D,$C6,$9F,$E5,$3F,$46,$F8,$BB,$F4,$78,$6D,$7F,$04
	.dc.b	$F2,$5F,$16,$FA,$3F,$E6,$3F,$17,$FE,$22,$AD,$4E,$5E,$11,$9D,$BF
	.dc.b	$41,$7C,$BC,$9F,$E7,$36,$7F,$9E,$E2,$F6,$B3,$1F,$CF,$BE,$41,$F3
	.dc.b	$DF,$D0,$3C,$E7,$9B,$98,$C3,$CB,$E8,$F1,$B7,$F1,$CF,$47,$DD,$18
	.dc.b	$DB,$F6,$51,$F0,$7D,$9E,$15,$BD,$51,$8A,$DF,$C9,$3C,$9F,$EA,$3F
	.dc.b	$47,$FD,$73,$AB,$FD,$BB,$39,$C7,$E4,$4E,$5F,$F1,$66,$49,$D6,$1F
	.dc.b	$12,$B4,$A7,$58,$76,$F0,$7F,$B0,$DF,$0C,$7F,$C7,$5C,$8C,$78,$E1
	.dc.b	$C3,$2F,$0E,$1F,$DA,$3E,$5F,$FC,$03,$1F,$6F,$E8,$F9,$3E,$48,$CE
	.dc.b	$9F,$5A,$72,$F8,$E3,$39,$78,$27,$38,$27,$37,$FD,$EE,$73,$18,$BF
	.dc.b	$EF,$71,$8F,$5F,$FB,$CB,$CB,$E6,$5F,$0F,$FD,$F3,$E8,$DF,$23,$FE
	.dc.b	$00,$9C,$E5,$F9,$39,$CB,$F7,$63,$38,$FE,$3A,$72,$FD,$E8,$CE,$9F
	.dc.b	$22,$72,$FF,$98,$F4,$4E,$AF,$FC,$29,$5A,$7E,$CC,$66,$FF,$C2,$B3
	.dc.b	$97,$EC,$78,$6F,$A3,$FF,$0A,$FA,$D7,$0F,$C9,$4D,$AF,$D1,$CD,$CB
	.dc.b	$C5,$36,$BF,$C4,$93,$6B,$AD,$F4,$3F,$C4,$75,$9E,$25,$6F,$34,$E6
	.dc.b	$30,$FF,$8D,$31,$FC,$4F,$D1,$FE,$B3,$2F,$2E,$5F,$E2,$C8,$C3,$FA
	.dc.b	$6A,$E1,$3F,$E3,$79,$CD,$7A,$2B,$87,$57,$91,$B5,$D0,$FE,$15,$5C
	.dc.b	$1F,$D1,$FC,$F3,$E0,$BF,$C7,$71,$9A,$FD,$14,$E6,$3F,$89,$19,$CD
	.dc.b	$19,$9F,$E1,$B3,$98,$C6,$3F,$A6,$8E,$93,$FE,$8E,$F8,$4F,$F4,$F7
	.dc.b	$CE,$FF,$11,$8F,$F5,$07,$D1,$FF,$91,$6B,$3E,$17,$FB,$2E,$ED,$6B
	.dc.b	$C1,$FF,$55,$61,$C6,$3A,$07,$18,$73,$8C,$5F,$36,$9C,$F0,$EA,$F2
	.dc.b	$AE,$37,$DA,$B8,$46,$2B,$75,$FE,$B7,$C3,$8E,$1D,$EB,$C9,$39,$3B
	.dc.b	$88,$D5,$F4,$3F,$D6,$2E,$23,$FD,$66,$2B,$AE,$1E,$38,$CC,$63,$0F
	.dc.b	$1C,$6C,$EF,$5C,$DF,$65,$F6,$BE,$51,$83,$E1,$1E,$63,$D1,$87,$BD
	.dc.b	$C1,$D5,$C1,$EE,$71,$3D,$6F,$C7,$87,$96,$78,$BE,$C7,$D2,$53,$E3
	.dc.b	$15,$C7,$D1,$3D,$5F,$F3,$A4,$F1,$8C,$7B,$DA,$9C,$CF,$C8,$9C,$BB
	.dc.b	$F6,$7F,$CB,$AD,$D7,$AC,$62,$38,$35,$FF,$71,$CA,$38,$E7,$F8,$FF
	.dc.b	$A3,$FE,$F9,$5F,$88,$6B,$F7,$AB,$85,$FE,$31,$B1,$F8,$29,$DD,$73
	.dc.b	$F3,$BF,$DE,$2F,$E6,$EB,$6D,$C3,$C3,$AF,$FB,$DA,$F8,$B6,$F8,$BE
	.dc.b	$89,$DE,$01,$87,$49,$E1,$1E,$9A,$E2,$EE,$74,$DA,$E6,$AF,$C2,$F5
	.dc.b	$6B,$DA,$37,$3C,$A3,$85,$76,$69,$F1,$67,$67,$6A,$E6,$74,$AE,$47
	.dc.b	$37,$35,$CE,$33,$3C,$63,$B4,$F1,$8F,$87,$CE,$FA,$2F,$F4,$06,$1F
	.dc.b	$E2,$D6,$BF,$E2,$66,$E7,$FD,$57,$E6,$9F,$87,$7C,$BA,$D7,$B9,$6B
	.dc.b	$FC,$07,$7C,$38,$BF,$DD,$8D,$AF,$F0,$66,$1C,$9F,$07,$AF,$C6,$B6
	.dc.b	$D8,$FF,$78,$D2,$31,$CE,$B7,$DB,$6F,$4F,$65,$5B,$E1,$87,$63,$FC
	.dc.b	$65,$7C,$1C,$1E,$F9,$E0,$76,$8D,$47,$F6,$6F,$97,$ED,$DC,$69,$E4
	.dc.b	$EE,$FF,$1A,$4F,$F9,$5E,$74,$D9,$DB,$FD,$CB,$CA,$AF,$53,$46,$4E
	.dc.b	$5C,$BE,$02,$3F,$DE,$BE,$02,$BF,$DE,$A2,$F0,$7F,$7A,$9B,$C1,$FD
	.dc.b	$EA,$6D,$7F,$81,$A7,$E4,$73,$BD,$9A,$49,$FF,$2C,$79,$2F,$98,$C4
	.dc.b	$7F,$9A,$AB,$83,$FE,$75,$6B,$FE,$5C,$9E,$0E,$6B,$8E,$5F,$E4,$8A
	.dc.b	$DB,$FE,$21,$AD,$D7,$5F,$91,$F2,$91,$F1,$FC,$AD,$74,$78,$63,$F0
	.dc.b	$B2,$9D,$57,$79,$BA,$7F,$93,$E7,$F1,$7E,$7F,$FB,$C7,$0F,$F3,$B3
	.dc.b	$5F,$F3,$7F,$C2,$6B,$A7,$F1,$DE,$AB,$E3,$E9,$2B,$8D,$57,$FD,$1F
	.dc.b	$3F,$ED,$A7,$FD,$34,$D7,$33,$F9,$FF,$A9,$7E,$13,$FC,$AB,$56,$9F
	.dc.b	$C6,$67,$3E,$0E,$EB,$53,$FE,$7C,$F5,$9F,$07,$E6,$F6,$26,$7E,$07
	.dc.b	$F4,$DE,$4B,$8B,$D7,$F8,$67,$EA,$7D,$DF,$65,$FF,$AA,$37,$CB,$FE
	.dc.b	$E5,$FA,$17,$C9,$7E,$D6,$11,$87,$51,$98,$7D,$8D,$89,$3A,$3A,$46
	.dc.b	$17,$17,$C0,$7F,$9C,$57,$57,$3B,$FF,$11,$8C,$4E,$E7,$C7,$5F,$C0
	.dc.b	$FE,$7F,$FD,$53,$1F,$EC,$97,$DA,$FF,$B0,$9B,$3F,$E9,$A7,$71,$9F
	.dc.b	$7B,$96,$52,$39,$47,$F1,$5F,$09,$F4,$4F,$8A,$C3,$6B,$FA,$86,$3E
	.dc.b	$BA,$DD,$7F,$0D,$C3,$6C,$AD,$E9,$FD,$5E,$B7,$1F,$C6,$2B,$E8,$7D
	.dc.b	$0B,$E3,$E9,$27,$89,$FD,$7F,$E8,$9F,$53,$98,$F9,$27,$F6,$3A,$E4
	.dc.b	$F0,$0E,$1D,$89,$E6,$EE,$35,$9F,$D1,$97,$1F,$E1,$5B,$5B,$C3,$F9
	.dc.b	$CF,$CC,$3E,$9B,$F9,$C7,$C0,$FD,$9C,$7F,$D9,$6D,$DB,$FD,$AF,$5B
	.dc.b	$C3,$FD,$D4,$D7,$FD,$F7,$3B,$5E,$87,$FA,$C1,$FF,$76,$CE,$EF,$AE
	.dc.b	$DE,$02,$4F,$4F,$A6,$AF,$48,$B3,$FE,$B5,$A5,$7E,$87,$E9,$1F,$2D
	.dc.b	$49,$87,$CB,$87,$99,$88,$F9,$71,$C1,$D4,$F3,$9D,$61,$F9,$BA,$D5
	.dc.b	$70,$AD,$3F,$7B,$2F,$F6,$AF,$37,$CA,$24,$7C,$6E,$CA,$FC,$68,$B8
	.dc.b	$7A,$E7,$F3,$5C,$8F,$88,$BD,$D3,$0F,$C5,$F3,$FE,$2A,$76,$66,$3F
	.dc.b	$17,$1F,$8E,$9C,$C7,$E2,$FE,$93,$FC,$90,$DF,$67,$F8,$44,$62,$3F
	.dc.b	$1A,$FF,$97,$5D,$10,$CC,$70,$F7,$38,$B2,$35,$9F,$F0,$1A,$E5,$1E
	.dc.b	$4F,$D8,$1F,$4F,$FB,$77,$FD,$F7,$1E,$C6,$BB,$8D,$1B,$E6,$91,$B8
	.dc.b	$FC,$CF,$D1,$FE,$63,$E1,$A4,$FE,$52,$32,$FD,$5A,$4E,$A3,$F2,$73
	.dc.b	$FA,$0F,$99,$91,$B7,$F4,$BF,$1B,$88,$FC,$EC,$5C,$FF,$25,$3F,$A8
	.dc.b	$F9,$8F,$D4,$F1,$4B,$F0,$46,$5F,$14,$65,$FD,$3C,$7F,$39,$FA,$26
	.dc.b	$3F,$ED,$3F,$89,$FF,$1E,$47,$F4,$5A,$FF,$5E,$46,$39,$75,$1F,$E6
	.dc.b	$B9,$E2,$E8,$C5,$69,$47,$37,$FD,$75,$5C,$DE,$06,$A7,$87,$D1,$39
	.dc.b	$DF,$47,$A8,$C3,$63,$35,$D0,$F2,$63,$CA,$33,$F1,$3F,$EF,$D9,$FF
	.dc.b	$41,$5F,$27,$80,$71,$1B,$8F,$4D,$73,$BF,$01,$23,$8E,$5E,$DA,$E1
	.dc.b	$7C,$B2,$FF,$33,$36,$BC,$38,$F9,$27,$E1,$47,$8A,$74,$F8,$E7,$4F
	.dc.b	$0C,$3E,$3E,$1D,$EF,$AD,$6A,$B8,$57,$97,$0F,$55,$EE,$79,$1D,$EB
	.dc.b	$86,$1E,$09,$D4,$F2,$35,$7C,$F0,$F3,$E1,$EE,$C3,$89,$27,$CB,$8F
	.dc.b	$C7,$BE,$24,$AF,$1C,$65,$E5,$3D,$67,$EC,$94,$F1,$46,$9F,$9F,$3C
	.dc.b	$6F,$E0,$D6,$F0,$F4,$E3,$F2,$89,$8F,$C5,$8C,$BE,$7C,$3C,$97,$EC
	.dc.b	$39,$B8,$3E,$3E,$3E,$87,$C9,$7D,$30,$F7,$C6,$2F,$A5,$6E,$3E,$09
	.dc.b	$FE,$A7,$3E,$61,$6B,$F2,$4D,$9F,$CD,$B9,$F0,$3F,$9A,$3C,$4F,$CB
	.dc.b	$38,$75,$D5,$97,$2D,$AF,$F2,$B3,$FA,$99,$E7,$7F,$9D,$FA,$3F,$F1
	.dc.b	$DF,$85,$9A,$98,$7E,$4E,$F7,$9F,$FA,$FE,$F7,$39,$AD,$E1,$FA,$0B
	.dc.b	$DC,$FF,$BB,$EB,$73,$9A,$DD,$FE,$92,$F7,$6B,$DC,$E6,$B7,$3F,$B8
	.dc.b	$AD,$D6,$A3,$13,$A7,$E0,$9F,$BC,$8D,$38,$BF,$13,$E9,$35,$E6,$FF
	.dc.b	$93,$E3,$24,$C3,$C5,$3A,$C7,$B6,$5F,$32,$39,$57,$CB,$BE,$57,$C2
	.dc.b	$39,$4F,$98,$E4,$70,$8D,$11,$F9,$13,$C9,$E1,$3C,$9F,$54,$FD,$39
	.dc.b	$D4,$60,$DC,$73,$7E,$64,$6A,$BD,$15,$CD,$B5,$CE,$FD,$B9,$7D,$49
	.dc.b	$CE,$1E,$49,$CE,$5F,$3F,$2F,$9B,$5A,$AF,$11,$23,$37,$CC,$95,$FD
	.dc.b	$02,$32,$DE,$2F,$5C,$7D,$80,$E3,$3C,$B7,$1E,$07,$38,$F9,$1D,$CF
	.dc.b	$CF,$FF,$C6,$F3,$FE,$27,$7F,$CA,$0D,$F0,$FF,$D1,$DC,$DF,$E8,$FA
	.dc.b	$E1,$E2,$E2,$9A,$FF,$93,$30,$E6,$FF,$9A,$AB,$73,$CC,$ED,$5C,$67
	.dc.b	$87,$25,$AF,$23,$9F,$87,$78,$DF,$AC,$CB,$C5,$97,$FA,$D7,$2F,$45
	.dc.b	$F1,$8C,$63,$FE,$5C,$D3,$13,$CF,$0F,$F5,$7D,$FB,$F9,$F1,$E7,$71
	.dc.b	$6C,$64,$FF,$35,$C6,$5F,$F1,$27,$C3,$FA,$7C,$57,$82,$FA,$7C,$29
	.dc.b	$B1,$86,$D7,$23,$FC,$5D,$1F,$0F,$E8,$9F,$1A,$FA,$8C,$36,$BF,$D3
	.dc.b	$33,$AF,$0F,$35,$27,$35,$BE,$0F,$F4,$DD,$EE,$73,$5B,$9C,$D6,$F8
	.dc.b	$BF,$D9,$97,$B9,$CD,$7B,$FC,$27,$BD,$7E,$E7,$0D,$CB,$FC,$C7,$3A
	.dc.b	$6C,$E9,$A7,$E3,$FD,$AF,$B0,$7D,$5F,$9D,$B1,$92,$63,$FE,$96,$8C
	.dc.b	$BE,$C1,$5F,$25,$B3,$F1,$97,$D0,$E6,$64,$E5,$E1,$18,$9E,$91,$74
	.dc.b	$F8,$71,$CA,$B8,$1F,$E8,$DF,$CD,$FF,$D0,$8D,$F0,$3F,$1D,$97,$E6
	.dc.b	$E7,$35,$FA,$36,$DF,$EA,$DB,$7F,$EB,$AA,$E1,$7F,$97,$FB,$4E,$5E
	.dc.b	$85,$AF,$03,$71,$E8,$DC,$7A,$3F,$AD,$8F,$F7,$87,$D5,$FE,$BF,$07
	.dc.b	$D2,$FF,$B8,$30,$FF,$6E,$7B,$4A,$D4,$E6,$FF,$DB,$B3,$9A,$DD,$6B
	.dc.b	$CD,$ED,$39,$AD,$CE,$0D,$57,$03,$E4,$7E,$1B,$F1,$4F,$DF,$7E,$4B
	.dc.b	$FC,$2A,$32,$4A,$F9,$13,$A7,$84,$7F,$AF,$7E,$A1,$F2,$9F,$C2,$5F
	.dc.b	$AD,$3F,$C4,$5C,$CE,$F8,$F5,$5B,$6C,$64,$FE,$37,$B9,$F2,$7D,$3F
	.dc.b	$F2,$4C,$13,$97,$51,$98,$C4,$A3,$F9,$1C,$61,$D4,$74,$F2,$3F,$91
	.dc.b	$3B,$6C,$70,$8D,$1E,$42,$D7,$89,$B5,$D7,$C9,$E3,$7F,$CA,$F8,$9E
	.dc.b	$77,$F2,$DF,$A7,$FE,$43,$8F,$F3,$59,$CE,$3F,$EE,$79,$CE,$3F,$CF
	.dc.b	$FE,$29,$F5,$5F,$CD,$5B,$5F,$EE,$DF,$A7,$70,$E6,$BF,$9E,$7C,$63
	.dc.b	$F1,$E7,$CD,$E1,$B8,$7A,$F5,$7A,$0E,$DF,$C5,$6B,$F3,$71,$E5,$8F
	.dc.b	$87,$C3,$F6,$7C,$75,$CB,$FD,$FD,$C2,$8C,$67,$FD,$47,$83,$F8,$E5
	.dc.b	$7F,$56,$D3,$B7,$B9,$AD,$3E,$D4,$62,$7C,$8D,$BF,$9C,$79,$B5,$7A
	.dc.b	$DF,$D5,$56,$1F,$D6,$7E,$97,$E1,$E1,$ED,$F5,$51,$8C,$39,$E2,$F9
	.dc.b	$3F,$C1,$C2,$B7,$C9,$CC,$98,$79,$2B,$73,$A8,$ED,$87,$5D,$3F,$BB
	.dc.b	$36,$3E,$05,$70,$8C,$61,$DD,$7D,$8F,$A4,$FE,$E1,$3F,$EE,$0A,$EB
	.dc.b	$19,$69,$E3,$F4,$F1,$7F,$DF,$EF,$9C,$EA,$3D,$F7,$CD,$D5,$FF,$7F
	.dc.b	$F4,$BF,$6F,$8F,$F8,$35,$B7,$F5,$FE,$6E,$DF,$C5,$9B,$99,$A6,$B5
	.dc.b	$56,$3B,$FB,$13,$0F,$BB,$19,$7E,$B5,$3E,$DF,$C8,$7F,$84,$67,$13
	.dc.b	$A8,$CB,$4E,$6E,$31,$FF,$10,$57,$06,$C7,$33,$59,$7F,$8A,$E9,$F2
	.dc.b	$3F,$E2,$16,$9F,$92,$F9,$D7,$D5,$66,$64,$FE,$34,$DE,$1F,$8C,$AD
	.dc.b	$C6,$67,$73,$FE,$33,$8D,$DF,$E8,$2B,$73,$9A,$F1,$74,$CB,$FD,$1F
	.dc.b	$D1,$BC,$9C,$5F,$3A,$B8,$FE,$AA,$B5,$7E,$59,$FE,$A5,$5F,$E4,$6F
	.dc.b	$81,$FF,$23,$FC,$D3,$E7,$7F,$C9,$32,$FA,$37,$7F,$F2,$DA,$D4,$FF
	.dc.b	$B1,$6F,$8D,$F0,$BD,$E3,$8A,$D0,$BE,$2F,$5C,$3A,$5F,$FB,$1F,$0E
	.dc.b	$13,$9B,$AF,$9C,$E5,$3D,$EB,$8C,$76,$BE,$B8,$78,$A3,$66,$6B,$85
	.dc.b	$F4,$86,$1E,$0A,$E1,$7E,$69,$F0,$3D,$30,$EF,$1F,$1A,$B8,$46,$23
	.dc.b	$78,$76,$7C,$26,$CB,$7D,$A3,$E0,$46,$27,$93,$E4,$31,$87,$9E,$FB
	.dc.b	$B6,$32,$76,$3C,$27,$FA,$0A,$FB,$3F,$04,$C5,$76,$AF,$F7,$94,$7C
	.dc.b	$D3,$FD,$E8,$E1,$F1,$1F,$84,$1F,$46,$F8,$DF,$F3,$2C,$7F,$BF,$E3
	.dc.b	$1E,$D7,$FB,$F5,$B1,$99,$DB,$CE,$1F,$1B,$FD,$EA,$39,$35,$D3,$5F
	.dc.b	$CA,$C7,$0D,$FB,$4F,$1D,$7F,$1B,$F4,$7F,$D4,$6F,$F1,$6D,$7E,$9D
	.dc.b	$F1,$6D,$71,$DB,$F4,$35,$BB,$E7,$E2,$7E,$91,$FC,$CE,$1C,$5B,$7C
	.dc.b	$70,$FF,$58,$DF,$1C,$39,$E1,$C5,$F4,$D6,$EB,$E5,$B6,$B8,$47,$2A
	.dc.b	$E0,$FF,$AF,$30,$F0,$5F,$2C,$7D,$B5,$AA,$5F,$27,$AC,$FE,$B2,$F8
	.dc.b	$C7,$53,$89,$9C,$3C,$91,$A8,$E7,$7D,$6F,$A4,$E4,$D3,$DC,$CD,$72
	.dc.b	$3D,$93,$C4,$F3,$4F,$8B,$AD,$F4,$7F,$E9,$AC,$3F,$BB,$35,$FF,$6C
	.dc.b	$37,$1F,$E4,$DE,$37,$F1,$67,$FD,$C5,$19,$E2,$FF,$70,$DF,$4F,$73
	.dc.b	$51,$89,$D6,$1F,$E1,$EE,$8F,$F1,$1B,$E0,$F4,$76,$FF,$80,$67,$FC
	.dc.b	$B6,$E4,$FE,$85,$F0,$1A,$C7,$D3,$56,$B7,$5E,$88,$E7,$E0,$30,$FF
	.dc.b	$69,$4F,$33,$45,$CF,$FA,$16,$3D,$7B,$9F,$47,$49,$5F,$DA,$3E,$8E
	.dc.b	$BF,$D7,$A2,$DF,$F8,$AB,$C1,$56,$36,$F8,$B8,$74,$C3,$FB,$7E,$1E
	.dc.b	$6B,$DC,$7F,$72,$E5,$F1,$11,$FE,$EF,$F2,$95,$FE,$DB,$17,$2F,$04
	.dc.b	$DC,$7B,$CD,$AF,$F2,$6D,$AB,$E5,$73,$BE,$8A,$92,$BF,$C5,$DF,$47
	.dc.b	$5F,$F2,$0F,$1F,$F4,$B9,$C6,$7F,$E4,$8F,$1D,$78,$D3,$F0,$78,$F5
	.dc.b	$C7,$FC,$84,$D7,$FC,$47,$3E,$2F,$89,$91,$A7,$EC,$DB,$E8,$F3,$3E
	.dc.b	$2B,$56,$31,$F1,$55,$C7,$C3,$36,$FF,$CD,$B8,$73,$7F,$CB,$BF,$46
	.dc.b	$FA,$3F,$C3,$B6,$33,$18,$E9,$FC,$87,$A5,$FE,$60,$9F,$F2,$E3,$5F
	.dc.b	$F3,$77,$A5,$CA,$B7,$EF,$C9,$DB,$CA,$7F,$3F,$EA,$3E,$8F,$F9,$36
	.dc.b	$FF,$EA,$BE,$3F,$F0,$0E,$DF,$9F,$F1,$FD,$F9,$7F,$A1,$F0,$E1,$E1
	.dc.b	$79,$3C,$C9,$3E,$47,$85,$FA,$3E,$77,$F5,$1F,$45,$E7,$CF,$B3,$79
	.dc.b	$3F,$73,$C2,$AF,$D9,$E2,$FA,$27,$C8,$E1,$B1,$FE,$99,$35,$19,$F9
	.dc.b	$0D,$32,$35,$3C,$1F,$F4,$59,$D6,$74,$E2,$35,$97,$FB,$36,$3A,$9C
	.dc.b	$A7,$F8,$17,$D1,$3E,$5E,$F2,$4A,$FF,$48,$4E,$5B,$39,$7F,$37,$1F
	.dc.b	$05,$FF,$48,$3B,$7F,$D3,$FE,$13,$E8,$BF,$C8,$11,$FE,$B7,$8C,$7A
	.dc.b	$9F,$C6,$1B,$1F,$E5,$7C,$31,$5C,$23,$9B,$72,$FF,$74,$57,$09,$FF
	.dc.b	$2C,$57,$D0,$FA,$27,$D0,$D2,$4F,$CF,$FA,$27,$C9,$62,$3E,$5C,$7D
	.dc.b	$97,$5C,$39,$CB,$E5,$61,$D2,$72,$43,$E9,$B9,$7F,$D6,$BF,$2E,$F8
	.dc.b	$8F,$E6,$1F,$05,$F4,$EF,$F2,$CD,$E6,$FF,$66,$56,$EF,$FD,$92,$D7
	.dc.b	$F8,$4C,$ED,$FF,$5F,$37,$CF,$4E,$EF,$AE,$DE,$02,$5F,$4F,$89,$AF
	.dc.b	$48,$B3,$FE,$DA,$A5,$7E,$6B,$E8,$9F,$45,$49,$87,$E6,$63,$F9,$23
	.dc.b	$63,$2F,$E5,$A3,$8B,$BF,$33,$38,$7E,$56,$73,$3B,$AF,$05,$7C,$8E
	.dc.b	$79,$3F,$EE,$3D,$3F,$1D,$87,$9F,$B7,$33,$F1,$AA,$CE,$63,$F2,$1C
	.dc.b	$6F,$91,$FC,$83,$FE,$86,$8C,$F8,$58,$99,$18,$21,$F9,$C9,$E4,$FE
	.dc.b	$E7,$E8,$BF,$C8,$6D,$F6,$7F,$80,$C6,$23,$F6,$EF,$F9,$45,$D1,$0C
	.dc.b	$C7,$0F,$6F,$8B,$23,$58,$FE,$EE,$B9,$47,$EB,$BE,$69,$F4,$5F,$AB
	.dc.b	$5B,$53,$11,$27,$F6,$53,$A6,$CE,$5F,$CC,$FC,$47,$E6,$3E,$6A,$47
	.dc.b	$E5,$23,$FC,$0B,$CA,$73,$1F,$96,$9C,$C7,$E8,$3E,$8A,$4F,$D9,$78
	.dc.b	$FC,$F6,$23,$F3,$B1,$74,$FC,$94,$FE,$A3,$CE,$FD,$4F,$04,$BF,$C1
	.dc.b	$46,$5F,$C2,$46,$5F,$D3,$C7,$F3,$2F,$A2,$F3,$47,$FB,$27,$E4,$FF
	.dc.b	$D8,$F1,$FD,$12,$3F,$D5,$71,$8E,$6F,$F5,$33,$AC,$EC,$85,$74,$78
	.dc.b	$99,$7A,$7C,$2A,$BA,$4F,$84,$D5,$74,$5F,$43,$FC,$CF,$E8,$B5,$18
	.dc.b	$E9,$46,$27,$53,$E0,$3C,$79,$73,$8C,$E3,$E1,$9F,$44,$63,$1F,$F7
	.dc.b	$3B,$63,$D1,$5C,$70,$EF,$1D,$67,$4F,$8F,$0F,$2E,$5E,$EC,$39,$DF
	.dc.b	$0C,$7C,$17,$B3,$94,$66,$7E,$1C,$62,$FD,$93,$98,$F3,$B5,$F6,$9A
	.dc.b	$88,$F1,$3E,$15,$F8,$DE,$31,$AB,$F1,$E1,$EA,$AD,$57,$96,$73,$7D
	.dc.b	$EB,$58,$FC,$3C,$39,$63,$EE,$7E,$0C,$67,$0F,$16,$1C,$A7,$31,$C0
	.dc.b	$F1,$9A,$AF,$83,$3B,$8C,$1A,$7E,$4E,$3E,$6C,$3E,$89,$31,$F8,$91
	.dc.b	$97,$D0,$4C,$3C,$37,$D0,$F7,$61,$E9,$7E,$3E,$3E,$9C,$7C,$44,$9F
	.dc.b	$29,$B7,$13,$E5,$8C,$B4,$87,$63,$F1,$67,$D6,$3F,$DE,$58,$FE,$61
	.dc.b	$B8,$7E,$39,$B8,$7F,$BD,$DF,$CB,$F7,$7F,$2E,$78,$B5,$FC,$CB,$5D
	.dc.b	$F4,$F1,$7F,$57,$1F,$CD,$B1,$FC,$0B,$71,$FD,$D4,$BE,$8B,$FC,$6B
	.dc.b	$E0,$FE,$B6,$B5,$3D,$0F,$D1,$38,$78,$79,$A9,$E1,$8C,$9D,$BF,$BA
	.dc.b	$BE,$B1,$82,$63,$E4,$9D,$5F,$39,$CE,$3E,$E9,$E9,$5F,$AC,$BE,$6F
	.dc.b	$1A,$D9,$E7,$75,$7C,$DD,$12,$FA,$4F,$F0,$9A,$E7,$18,$9D,$4F,$38
	.dc.b	$F8,$F3,$A8,$C1,$0F,$9F,$7E,$AC,$3C,$F5,$CC,$93,$FC,$A6,$B9,$3C
	.dc.b	$0F,$49,$6F,$D3,$5C,$A3,$B6,$8C,$BE,$95,$F2,$8C,$63,$BA,$E5,$8F
	.dc.b	$A3,$2F,$84,$F8,$A3,$27,$A6,$F9,$61,$F7,$70,$E7,$3E,$49,$CE,$1D
	.dc.b	$67,$39,$79,$AB,$D3,$18,$C3,$B5,$F0,$9C,$E1,$E8,$C7,$C7,$1E,$1E
	.dc.b	$2A,$FE,$33,$3F,$C4,$46,$2F,$E3,$87,$15,$96,$CF,$F4,$DF,$4D,$F4
	.dc.b	$5F,$E3,$A8,$FE,$B1,$CB,$FE,$A5,$8F,$F2,$A4,$60,$FE,$B9,$3C,$3C
	.dc.b	$3F,$C9,$56,$FC,$5B,$5A,$BE,$99,$FF,$B7,$1E,$6D,$BD,$AC,$3C,$77
	.dc.b	$87,$CB,$C7,$6F,$FD,$A7,$87,$F9,$83,$3C,$4C,$8F,$36,$1C,$EB,$6F
	.dc.b	$F9,$BA,$B4,$FF,$9C,$3C,$EE,$9C,$DE,$2F,$1B,$FC,$CB,$18,$C3,$FC
	.dc.b	$DA,$DA,$E1,$1D,$B2,$F9,$37,$EA,$F2,$DD,$EF,$82,$C2,$FA,$2B,$18
	.dc.b	$6E,$1F,$E2,$E6,$E3,$FE,$37,$7E,$2F,$D1,$3E,$2D,$F1,$B8,$6D,$FC
	.dc.b	$09,$D7,$85,$9A,$93,$9A,$DE,$0E,$5F,$E1,$58,$70,$AD,$5F,$0A,$D4
	.dc.b	$7F,$86,$E7,$C5,$3E,$1C,$35,$7F,$EA,$EF,$05,$EE,$7C,$35,$F4,$36
	.dc.b	$30,$DC,$7F,$CD,$ED,$CB,$FC,$D0,$FD,$FF,$96,$93,$FE,$9E,$FA,$2B
	.dc.b	$5F,$8B,$9B,$8F,$FA,$4E,$6D,$79,$EB,$C2,$FC,$56,$9E,$C1,$A7,$F8
	.dc.b	$DF,$E5,$3E,$7C,$6A,$FE,$2B,$6B,$A6,$3A,$AD,$B2,$3B,$39,$69,$F8
	.dc.b	$7F,$A2,$FF,$46,$37,$C3,$FC,$87,$92,$D3,$F4,$15,$AC,$7C,$DC,$E8
	.dc.b	$C6,$7F,$D0,$63,$18,$72,$BE,$2F,$4F,$A2,$E7,$E8,$5C,$FC,$4D,$CB
	.dc.b	$AE,$48,$C6,$5D,$63,$15,$FA,$CA,$7B,$2F,$A1,$FD,$CE,$4F,$A2,$FF
	.dc.b	$73,$DE,$EB,$4F,$ED,$27,$5C,$9F,$C0,$30,$E7,$87,$2C,$38,$C7,$D5
	.dc.b	$33,$7C,$63,$39,$7F,$61,$9C,$9B,$C7,$BB,$F7,$71,$F2,$79,$BD,$E7
	.dc.b	$35,$ED,$F9,$BF,$9E,$7F,$01,$FA,$2F,$F3,$C6,$5F,$5E,$32,$D3,$F8
	.dc.b	$6F,$82,$9D,$E5,$F2,$AB,$F8,$C7,$D1,$3E,$4E,$F3,$E1,$B8,$FF,$09
	.dc.b	$9D,$36,$74,$D3,$F8,$D8,$FA,$2F,$E5,$39,$FF,$80,$63,$1E,$47,$F2
	.dc.b	$A7,$5E,$4D,$AD,$C7,$B1,$B9,$FF,$84,$DB,$1C,$0E,$EE,$09,$5C,$EB
	.dc.b	$A3,$FC,$D7,$97,$FD,$31,$5F,$C8,$FE,$89,$E3,$E7,$CE,$FE,$75,$18
	.dc.b	$9F,$E5,$1F,$08,$FA,$1F,$E7,$CD,$8F,$E5,$9F,$44,$E2,$CD,$7F,$46
	.dc.b	$F8,$A7,$D1,$3E,$6B,$FA,$35,$EF,$0F,$E8,$35,$A9,$CE,$1E,$7F,$8C
	.dc.b	$95,$E2,$F2,$6F,$1F,$EB,$B4,$FE,$4B,$C9,$FA,$D6,$D7,$FA,$EF,$A7
	.dc.b	$F6,$DE,$E7,$A3,$0F,$E9,$1E,$1F,$FB,$AA,$BF,$0A,$DC,$7F,$87,$5F
	.dc.b	$EF,$7D,$6D,$E1,$FD,$77,$E8,$BD,$15,$FB,$E6,$E7,$FE,$3D,$B7,$C8
	.dc.b	$FD,$8F,$1D,$5A,$CF,$FB,$9B,$70,$FE,$D4,$DB,$E3,$D3,$DF,$82,$5F
	.dc.b	$F5,$9A,$D6,$3D,$5E,$58,$F6,$FA,$2F,$31,$FD,$96,$7F,$DF,$55,$C6
	.dc.b	$32,$D3,$F4,$B1,$AC,$BD,$93,$A6,$CE,$BD,$4C,$69,$FC,$6F,$D4,$77
	.dc.b	$FF,$80,$5B,$5F,$B0,$F9,$7B,$87,$F2,$49,$B7,$C6,$F8,$1F,$DE,$EB
	.dc.b	$37,$FE,$18,$AB,$7F,$E0,$F6,$E6,$73,$BA,$F3,$13,$E8,$BC,$87,$52
	.dc.b	$E1,$FE,$03,$84,$63,$0F,$C0,$D6,$2B,$53,$9C,$BE,$03,$CE,$32,$7F
	.dc.b	$83,$47,$C8,$FF,$88,$DA,$7F,$12,$F9,$C7,$D1,$66,$64,$F1,$7F,$C3
	.dc.b	$93,$A8,$C1,$31,$FE,$9A,$F4,$7F,$2F,$1B,$73,$3E,$1E,$79,$5F,$C8
	.dc.b	$BE,$8B,$FA,$1B,$E7,$8C,$D7,$47,$FC,$59,$3A,$8C,$4F,$F4,$B4,$3E
	.dc.b	$06,$FC,$CB,$E7,$3F,$D6,$3F,$45,$FD,$4A,$93,$97,$FD,$6A,$71,$33
	.dc.b	$3C,$3C,$7F,$F5,$34,$EE,$7F,$D7,$77,$DA,$B5,$85,$9E,$73,$56,$1E
	.dc.b	$1B,$E5,$1E,$89,$D4,$E6,$FF,$DA,$91,$8C,$3C,$31,$8A,$F2,$9F,$02
	.dc.b	$B8,$3E,$03,$27,$82,$78,$B9,$BF,$54,$F0,$8F,$59,$C1,$C3,$E5,$31
	.dc.b	$3C,$9D,$CF,$28,$C4,$6E,$31,$87,$C1,$8C,$5F,$C6,$7C,$A6,$0F,$55
	.dc.b	$72,$7E,$95,$72,$76,$62,$79,$47,$CA,$8C,$1C,$AB,$C0,$FD,$B3,$FD
	.dc.b	$ED,$87,$FB,$DD,$F6,$18,$3F,$02,$3E,$85,$F1,$BF,$E6,$78,$CC,$63
	.dc.b	$DD,$FF,$80,$0D,$8C,$CE,$DF,$C5,$C3,$E3,$7F,$CC,$D1,$C9,$AE,$9A
	.dc.b	$FE,$68,$F1,$38,$7F,$3B,$1C,$F8,$FF,$19,$F4,$4F,$13,$86,$9F,$95
	.dc.b	$8C,$5F,$E4,$6F,$83,$F8,$46,$CE,$EB,$8E,$BE,$1B,$E3,$7B,$C7,$D3
	.dc.b	$7F,$A1,$C3,$8D,$EF,$0E,$37,$FB,$2C,$38,$DE,$F0,$E3,$7E,$DC,$38
	.dc.b	$DE,$F0,$E2,$B0,$E3,$5B,$BE,$33,$F0,$9A,$EE,$B8,$D7,$08,$D6,$3F
	.dc.b	$06,$B7,$39,$9E,$06,$66,$3F,$A8,$8D,$1B,$C3,$CF,$1F,$52,$7E,$8D
	.dc.b	$F8,$0F,$4C,$E8,$DD,$72,$8C,$DF,$A3,$F3,$B7,$6F,$DF,$C3,$FC,$A5
	.dc.b	$18,$C7,$FC,$CC,$D7,$FD,$1D,$D3,$FE,$18,$CF,$87,$07,$E1,$EB,$57
	.dc.b	$CB,$0C,$E1,$FC,$92,$73,$5F,$C9,$B0,$E5,$3F,$ED,$EB,$7E,$74,$FA
	.dc.b	$2E,$31,$86,$C6,$67,$FD,$CB,$4F,$8F,$FE,$53,$7F,$EF,$AA,$D5,$7F
	.dc.b	$A9,$63,$93,$DE,$33,$0C,$3C,$95,$C9,$FE,$77,$1C,$2B,$E0,$FE,$74
	.dc.b	$F7,$BD,$B8,$A3,$13,$D7,$2F,$82,$FF,$A3,$7E,$B6,$FB,$A6,$9C,$BF
	.dc.b	$22,$FF,$54,$4F,$FA,$56,$74,$D9,$DB,$F0,$27,$6F,$FA,$36,$76,$FD
	.dc.b	$68,$7C,$57,$53,$8C,$7C,$3F,$03,$FD,$E2,$FD,$16,$B9,$EE,$F6,$FE
	.dc.b	$65,$FE,$81,$8D,$47,$D4,$9D,$ED,$F2,$67,$FD,$E5,$18,$DF,$EF,$5F
	.dc.b	$19,$E8,$7D,$8F,$CB,$BF,$D6,$2D,$F7,$BA,$63,$F2,$3D,$E7,$D8,$D2
	.dc.b	$57,$C1,$FA,$8A,$FF,$AB,$67,$6F,$FA,$F5,$AF,$C3,$8D,$E9,$E7,$E3
	.dc.b	$6F,$FC,$96,$BE,$67,$56,$B6,$FE,$B9,$F5,$DF,$E8,$29,$FC,$1C,$63
	.dc.b	$DD,$FC,$EF,$8A,$AD,$57,$E9,$BF,$26,$7C,$7E,$A7,$3A,$7C,$7B,$E1
	.dc.b	$86,$4F,$C5,$DF,$0C,$3B,$9F,$DE,$2F,$8E,$3F,$A9,$FC,$9F,$F2,$D3
	.dc.b	$FE,$3F,$6E,$1D,$23,$FD,$BD,$EC,$3E,$BB,$F2,$7F,$55,$FB,$29,$FF
	.dc.b	$0B,$B5,$FF,$7C,$46,$BC,$29,$1F,$EE,$79,$D5,$73,$8C,$78,$7D,$08
	.dc.b	$EE,$BF,$4B,$F9,$27,$CD,$8F,$E1,$3F,$92,$3E,$B7,$F5,$2D,$8F,$DD
	.dc.b	$CE,$F8,$7E,$37,$E4,$AF,$93,$F9,$ED,$AF,$8B,$B2,$31,$1D,$B1,$FD
	.dc.b	$AF,$C6,$7E,$0B,$0F,$C0,$7D,$E7,$FB,$FE,$BF,$84,$B5,$FE,$3B,$3B
	.dc.b	$E2,$FE,$C5,$F8,$83,$E0,$7F,$76,$DA,$FA,$A1,$C3,$FE,$FF,$FC,$41
	.dc.b	$F4,$DC,$2A,$4F,$0A,$DE,$5F,$F0,$06,$F7,$39,$AD,$A9,$FD,$73,$DE
	.dc.b	$3F,$86,$CF,$08,$D6,$1F,$DE,$7E,$A3,$FD,$23,$1E,$66,$BA,$3F,$CA
	.dc.b	$EE,$2F,$AF,$D5,$3E,$9B,$F8,$CC,$E6,$FA,$7C,$0E,$0F,$E3,$B5,$BC
	.dc.b	$3B,$7D,$5F,$02,$3F,$1A,$1F,$2F,$88,$F1,$0C,$15,$9C,$BC,$75,$E2
	.dc.b	$F3,$E5,$FA,$DF,$E5,$BE,$47,$E2,$63,$9E,$1F,$CF,$33,$F8,$55,$C2
	.dc.b	$7B,$D6,$F0,$F9,$77,$B9,$FF,$3C,$CF,$F8,$42,$7E,$7B,$6F,$9C,$FE
	.dc.b	$03,$E2,$7F,$66,$FF,$A0,$1A,$9B,$3F,$CC,$A3,$87,$85,$FD,$12,$76
	.dc.b	$6B,$CE,$C7,$91,$2B,$B3,$FD,$2B,$EA,$BF,$29,$1D,$E3,$4F,$CA,$A4
	.dc.b	$6F,$CB,$ED,$F2,$52,$BD,$0F,$3F,$84,$C4,$71,$8B,$AF,$E3,$B1,$FF
	.dc.b	$64,$4F,$02,$1B,$3F,$42,$E9,$FB,$7F,$87,$FF,$19,$6F,$95,$FD,$8A
	.dc.b	$31,$5F,$9A,$FC,$41,$F4,$19,$A9,$7F,$AC,$BD,$F0,$FF,$94,$EF,$73
	.dc.b	$9F,$27,$A1,$1F,$D9,$FD,$5B,$EA,$71,$1E,$49,$B9,$FE,$AA,$B5,$3D
	.dc.b	$F8,$7C,$C4,$79,$FC,$C5,$7D,$B1,$6B,$94,$DC,$BF,$CD,$19,$FE,$06
	.dc.b	$FA,$1C,$8F,$E0,$FF,$A2,$7F,$63,$BF,$DD,$B6,$BF,$CE,$0D,$9F,$F6
	.dc.b	$5F,$DB,$7E,$55,$46,$5B,$F3,$BF,$E0,$E8,$C4,$ED,$FE,$09,$7D,$BC
	.dc.b	$3D,$1F,$E1,$49,$7D,$26,$30,$F2,$CD,$7F,$6B,$3A,$7F,$97,$FB,$5F
	.dc.b	$8E,$6F,$47,$F6,$28,$C4,$7F,$91,$BF,$64,$7D,$56,$6A,$4F,$03,$FC
	.dc.b	$73,$3A,$6F,$C9,$FF,$8B,$3E,$C1,$F7,$54,$FF,$17,$D2,$69,$9A,$FF
	.dc.b	$21,$D7,$07,$99,$FC,$DB,$EC,$38,$3F,$ED,$FE,$CF,$F4,$F4,$3E,$B7
	.dc.b	$FD,$D7,$18,$6C,$FF,$BF,$27,$8B,$FE,$FD,$9E,$37,$A8,$4F,$37,$B1
	.dc.b	$C6,$B5,$1C,$BE,$41,$7F,$EA,$FF,$AF,$7D,$87,$E5,$27,$38,$04,$EB
	.dc.b	$B3,$AC,$62,$B8,$5F,$A6,$33,$5B,$7D,$07,$37,$17,$DA,$39,$5F,$4C
	.dc.b	$3D,$51,$C0,$C9,$6B,$8D,$FA,$70,$F0,$CF,$BC,$F0,$B8,$78,$E1,$DE
	.dc.b	$B8,$C6,$CE,$93,$B8,$CC,$70,$AF,$63,$AC,$3A,$91,$F3,$61,$E8,$C3
	.dc.b	$B6,$1C,$C9,$5E,$DC,$7E,$C6,$3D,$89,$3E,$7B,$E5,$8F,$B6,$FB,$61
	.dc.b	$D1,$E5,$8E,$27,$4F,$94,$F1,$E3,$DB,$0F,$0C,$66,$FE,$86,$1D,$A3
	.dc.b	$EA,$4E,$A7,$B4,$72,$3B,$1A,$C2,$BF,$90,$BE,$D1,$8C,$7E,$4E,$1D
	.dc.b	$6B,$E5,$1C,$4F,$16,$3F,$33,$0E,$E4,$7D,$6F,$9B,$0E,$B3,$C0,$E5
	.dc.b	$19,$7C,$18,$FA,$6B,$A1,$F1,$23,$F3,$47,$84,$A7,$E3,$0B,$87,$E5
	.dc.b	$DB,$D4,$FA,$7D,$C7,$E6,$63,$0E,$27,$FB,$84,$EA,$3F,$4F,$3A,$79
	.dc.b	$C7,$E9,$23,$8E,$1F,$A1,$FB,$0F,$F5,$3F,$56,$6A,$5F,$E7,$EF,$78
	.dc.b	$FA,$AF,$73,$9A,$DE,$1F,$9D,$BD,$CF,$EE,$6B,$73,$9A,$DD,$FF,$01
	.dc.b	$BD,$DA,$F7,$39,$AD,$CF,$C1,$7F,$80,$CE,$A3,$07,$8C,$97,$ED,$F1
	.dc.b	$7F,$D6,$CF,$86,$33,$7D,$5F,$DF,$57,$37,$E4,$DF,$81,$A6,$AF,$9D
	.dc.b	$71,$BE,$84,$7E,$34,$6A,$7C,$B3,$96,$CE,$B1,$ED,$97,$DA,$9F,$29
	.dc.b	$F2,$6F,$93,$E2,$BE,$4D,$BE,$4F,$03,$07,$63,$95,$F6,$CB,$E7,$E1
	.dc.b	$DE,$FE,$34,$E7,$1F,$81,$97,$C3,$AD,$63,$EF,$26,$3D,$0F,$04,$69
	.dc.b	$F0,$1F,$6F,$2F,$4C,$66,$3C,$A4,$AE,$D9,$7E,$09,$FE,$D3,$8F,$DE
	.dc.b	$BF,$83,$18,$AF,$A5,$3F,$82,$E4,$CD,$69,$B1,$FD,$4B,$E0,$DC,$4F
	.dc.b	$C0,$7F,$DF,$1D,$7F,$EE,$3F,$C4,$1F,$4F,$8C,$3F,$A8,$61,$C2,$F1
	.dc.b	$9A,$17,$C2,$BF,$B8,$D6,$A7,$37,$71,$F0,$57,$FB,$72,$72,$FF,$AE
	.dc.b	$BE,$CD,$F5,$FD,$BD,$C4,$62,$7F,$BE,$FA,$19,$8C,$69,$FE,$9B,$AD
	.dc.b	$DE,$4F,$05,$5B,$FF,$5B,$FD,$63,$EC,$7D,$78,$23,$13,$FE,$F0,$E6
	.dc.b	$7E,$8B,$FE,$D7,$8E,$37,$C2,$B7,$7F,$E9,$A6,$F0,$AB,$FB,$C6,$1C
	.dc.b	$BD,$EE,$18,$71,$F9,$49,$5F,$ED,$AF,$C4,$5F,$81,$63,$3E,$95,$72
	.dc.b	$E9,$0D,$D7,$C3,$0F,$6D,$F9,$7B,$64,$7A,$F6,$7D,$EE,$31,$F4,$55
	.dc.b	$CB,$AE,$DF,$02,$B1,$87,$F3,$8D,$FF,$C2,$AF,$CA,$BF,$C1,$D6,$E3
	.dc.b	$07,$FB,$93,$F0,$57,$AB,$F3,$9B,$1E,$FF,$92,$C4,$C8,$E0,$FE,$07
	.dc.b	$64,$E7,$4F,$F6,$3D,$FE,$12,$FB,$72,$F8,$BE,$CD,$F2,$BD,$E2,$4E
	.dc.b	$8F,$B9,$3A,$71,$7D,$3E,$AF,$CC,$73,$E5,$B3,$EB,$8B,$3F,$EE,$9D
	.dc.b	$57,$F8,$0A,$FE,$BB,$92,$F7,$C3,$FE,$F8,$8C,$4F,$8B,$1C,$BF,$D9
	.dc.b	$3E,$E9,$F1,$FF,$94,$6C,$7A,$3F,$02,$C4,$C8,$FF,$22,$61,$F8,$4D
	.dc.b	$17,$C1,$F3,$66,$C3,$84,$7F,$BC,$25,$EB,$77,$3F,$7B,$F8,$03,$F1
	.dc.b	$07,$E0,$D9,$CF,$B9,$E7,$AC,$60,$CF,$F2,$B8,$71,$BC,$C7,$97,$1F
	.dc.b	$CB,$57,$93,$C1,$F7,$76,$FD,$96,$E5,$DA,$7F,$CC,$15,$C2,$FF,$2B
	.dc.b	$AB,$EC,$F7,$5E,$BC,$6D,$6A,$6E,$1F,$E1,$0E,$67,$CE,$FF,$00,$6C
	.dc.b	$7E,$13,$E3,$B1,$32,$3C,$58,$FE,$4A,$3F,$1D,$39,$8F,$81,$3D,$63
	.dc.b	$A1,$F9,$3F,$7A,$C7,$2B,$4E,$2B,$EA,$D5,$C7,$ED,$53,$EA,$FC,$C7
	.dc.b	$E7,$7E,$1A,$E3,$CE,$2E,$1F,$7A,$6D,$F2,$9B,$5F,$C4,$B5,$56,$2F
	.dc.b	$F9,$D6,$5F,$AB,$7D,$67,$F3,$DF,$C3,$7F,$D2,$B7,$FC,$DE,$75,$8F
	.dc.b	$F1,$2F,$21,$5A,$BF,$E6,$5E,$37,$EB,$5B,$DD,$FD,$0F,$E5,$5F,$11
	.dc.b	$FC,$F6,$B5,$8F,$3F,$95,$FF,$59,$69,$FD,$16,$F6,$FF,$47,$8C,$E3
	.dc.b	$E6,$CD,$5A,$8F,$49,$FD,$2B,$E3,$BF,$9E,$F9,$87,$F9,$FF,$47,$E9
	.dc.b	$72,$FD,$44,$E7,$2F,$BF,$39,$7F,$CA,$ED,$BF,$F5,$93,$DE,$BF,$60
	.dc.b	$E6,$3F,$B6,$E9,$DB,$BB,$BB,$6F,$F3,$9E,$5B,$EC,$9F,$11,$FD,$9F
	.dc.b	$E6,$BF,$82,$E3,$FD,$6E,$73,$7F,$D9,$E7,$39,$FF,$42,$C3,$96,$1F
	.dc.b	$D9,$EF,$78,$7C,$5F,$01,$F5,$59,$A9,$97,$8A,$F7,$8F,$DB,$BD,$CE
	.dc.b	$6B,$FB,$2F,$E5,$EF,$63,$FB,$E7,$6B,$E1,$73,$18,$6D,$FE,$A2,$B8
	.dc.b	$3F,$DF,$EB,$FC,$03,$A7,$D6,$9C,$B6,$F8,$BF,$76,$77,$97,$F1,$DA
	.dc.b	$DB,$FE,$18,$FC,$CB,$E9,$61,$FE,$16,$FA,$37,$D2,$7F,$85,$A9,$5A
	.dc.b	$C7,$FC,$31,$5A,$BF,$E1,$D5,$AE,$2E,$17,$BE,$8F,$F0,$8F,$C7,$E2
	.dc.b	$65,$F0,$AD,$79,$EA,$CE,$5F,$92,$FC,$E1,$EB,$FF,$88,$FF,$25,$7D
	.dc.b	$27,$F8,$42,$73,$97,$F8,$8E,$72,$FF,$10,$8C,$DF,$3B,$E5,$EC,$71
	.dc.b	$65,$6D,$CB,$4F,$6F,$E7,$BF,$AE,$7F,$C7,$70,$FB,$4F,$8D,$39,$BF
	.dc.b	$F1,$E4,$E6,$B8,$F8,$9A,$3F,$C7,$91,$FD,$EA,$7F,$83,$F0,$AF,$93
	.dc.b	$6B,$85,$FF,$8F,$EF,$84,$FF,$94,$2B,$C1,$6F,$86,$A7,$8F,$89,$F8
	.dc.b	$F7,$F8,$42,$7F,$7B,$3A,$CB,$FC,$72,$EE,$75,$87,$F6,$38,$EB,$41
	.dc.b	$F0,$F7,$EC,$5F,$4B,$FE,$D4,$F0,$5F,$67,$FE,$2E,$9C,$DF,$FB,$5A
	.dc.b	$F8,$5E,$EB,$4F,$FB,$52,$33,$3D,$EF,$9E,$F8,$8F,$F7,$0C,$D9,$CC
	.dc.b	$30,$EE,$E8,$C5,$F7,$BE,$51,$EB,$BE,$58,$78,$A3,$CC,$70,$3C,$4E
	.dc.b	$5E,$97,$DB,$0F,$0B,$E0,$8E,$87,$13,$FC,$9D,$3C,$5C,$C7,$18,$C4
	.dc.b	$7C,$C3,$84,$F4,$C3,$E2,$46,$27,$77,$E8,$C3,$D7,$3C,$70,$E9,$18
	.dc.b	$9D,$4F,$8C,$F0,$DF,$C3,$70,$71,$71,$5E,$B8,$C1,$F0,$AB,$A3,$F8
	.dc.b	$33,$CA,$53,$E8,$14,$F8,$66,$7E,$CB,$F9,$5F,$D9,$E6,$11,$87,$83
	.dc.b	$5E,$1E,$22,$78,$4F,$F0,$2D,$6F,$3F,$F8,$A6,$F8,$B5,$F0,$35,$F4
	.dc.b	$4F,$17,$EB,$7D,$9B,$EC,$FD,$99,$FE,$2A,$31,$3F,$91,$AE,$13,$F8
	.dc.b	$E6,$E9,$F0,$EB,$75,$D2,$F9,$CF,$83,$0F,$36,$1C,$5F,$47,$3F,$7A
	.dc.b	$F9,$8D,$BF,$D8,$3B,$CA,$E1,$CD,$FD,$26,$3E,$0C,$78,$5F,$E4,$72
	.dc.b	$F2,$BF,$B9,$6A,$9E,$17,$E9,$C7,$CF,$3B,$CB,$CF,$3D,$E3,$D8,$78
	.dc.b	$27,$6F,$DA,$9D,$BC,$71,$D3,$0F,$24,$69,$E1,$19,$C7,$D2,$7B,$E3
	.dc.b	$71,$9B,$F3,$47,$D3,$AE,$5F,$66,$FA,$BF,$C6,$4A,$30,$FE,$3D,$AF
	.dc.b	$E5,$5A,$FE,$65,$CD,$FF,$8D,$7E,$F7,$FD,$25,$3F,$EE,$F6,$BF,$E2
	.dc.b	$69,$DF,$07,$FA,$2B,$ED,$1F,$15,$FA,$5C,$3F,$D9,$E2,$33,$7F,$DA
	.dc.b	$1A,$FF,$8C,$FA,$FF,$55,$3C,$27,$33,$CD,$AF,$7A,$FD,$54,$F5,$6C
	.dc.b	$F4,$8F,$F2,$77,$D9,$BE,$D3,$C9,$8F,$F3,$BE,$C4,$E7,$1F,$26,$3A
	.dc.b	$7F,$A0,$7B,$9F,$E0,$47,$FC,$F6,$D7,$FC,$8E,$DC,$7F,$CB,$9F,$67
	.dc.b	$FE,$B2,$9D,$F9,$1F,$52,$B8,$47,$F0,$76,$BF,$C2,$DA,$F8,$23,$C5
	.dc.b	$8A,$9F,$25,$86,$D7,$63,$FC,$27,$E0,$7F,$9D,$9B,$ED,$FA,$F2,$FF
	.dc.b	$6E,$BD,$3E,$71,$F4,$19,$BE,$17,$F4,$9A,$FD,$B6,$BF,$EC,$86,$BF
	.dc.b	$EB,$97,$8D,$FF,$9F,$7C,$CF,$F7,$8C,$FF,$AC,$F3,$46,$23,$FC,$03
	.dc.b	$5C,$38,$3F,$96,$7D,$A3,$E9,$BE,$95,$23,$0F,$FA,$B1,$AF,$F9,$EF
	.dc.b	$57,$CF,$F0,$9C,$CF,$F3,$56,$BD,$AB,$F4,$F1,$88,$FE,$FC,$F4,$CF
	.dc.b	$ED,$F5,$7C,$F3,$A5,$7E,$17,$C5,$7D,$9F,$FA,$53,$0F,$F3,$AB,$5F
	.dc.b	$F5,$1C,$6B,$D0,$FB,$7F,$69,$FE,$78,$9D,$C7,$F4,$39,$D4,$7F,$49
	.dc.b	$6B,$FD,$35,$AF,$F8,$3E,$BB,$63,$FE,$AB,$F6,$7F,$53,$B7,$E7,$5B
	.dc.b	$4B,$E7,$5F,$EF,$FF,$69,$EA,$FC,$EF,$B4,$FD,$DD,$FF,$87,$DA,$A7
	.dc.b	$6F,$FB,$72,$3F,$84,$BF,$D8,$DA,$F1,$F0,$BF,$89,$C6,$20,$FF,$85
	.dc.b	$AD,$97,$1F,$0F,$C4,$78,$EF,$F4,$3F,$01,$FE,$60,$C3,$FD,$79,$F0
	.dc.b	$3F,$A4,$F8,$13,$AE,$3F,$A9,$F6,$6F,$B4,$F9,$8D,$BF,$98,$D7,$E6
	.dc.b	$35,$FE,$F9,$D6,$FB,$4F,$F7,$FC,$7E,$65,$AF,$FB,$25,$BB,$7F,$B6
	.dc.b	$FE,$1B,$F1,$1B,$FF,$81,$7E,$D1,$F6,$9F,$9A,$6E,$1F,$E4,$BB,$F8
	.dc.b	$2F,$F9,$36,$32,$FF,$B4,$BE,$D3,$FD,$4C,$EE,$FA,$5F,$F3,$EA,$DC
	.dc.b	$FF,$84,$5A,$E8,$FF,$07,$3D,$2F,$BC,$2F,$DC,$F5,$3F,$CB,$CE,$1F
	.dc.b	$F7,$1C,$F1,$BF,$F1,$07,$D8,$7F,$B8,$E7,$80,$70,$FF,$BA,$9B,$7E
	.dc.b	$0F,$B3,$7D,$A7,$EF,$27,$37,$FE,$2B,$9F,$F6,$35,$F4,$6B,$CF,$E2
	.dc.b	$F8,$7D,$A4,$8F,$0D,$A3,$F8,$3D,$3E,$D3,$55,$FE,$FC,$87,$C6,$68
	.dc.b	$CE,$7F,$13,$0F,$E9,$19,$7B,$BC,$EF,$31,$FC,$4B,$CB,$FC,$7E,$3E
	.dc.b	$08,$FE,$DF,$5F,$D5,$23,$FD,$0C,$D7,$E2,$46,$8E,$5E,$B4,$8E,$47
	.dc.b	$CA,$F8,$8F,$DD,$E1,$FE,$4D,$6C,$7F,$A4,$5A,$B0,$FE,$82,$B1,$FB
	.dc.b	$B3,$A9,$FC,$F4,$7E,$01,$B5,$D7,$C4,$FE,$5B,$5F,$CA,$2F,$C5,$E4
	.dc.b	$60,$FE,$63,$F6,$9F,$9C,$8F,$14,$63,$EB,$E4,$70,$3D,$5F,$55,$D0
	.dc.b	$8F,$E1,$A9,$F4,$38,$8F,$F3,$36,$0A,$FF,$0C,$B6,$3F,$CC,$CD,$77
	.dc.b	$1E,$4F,$AF,$91,$E0,$3F,$27,$F6,$5F,$CB,$5F,$89,$A7,$F9,$DE,$31
	.dc.b	$B7,$E8,$9B,$1F,$E7,$BF,$B4,$FF,$79,$57,$2F,$91,$FE,$A3,$1F,$EF
	.dc.b	$28,$FC,$FB,$5E,$A7,$FB,$16,$B9,$B7,$CA,$FE,$B3,$1F,$CA,$23,$61
	.dc.b	$C6,$1F,$B4,$FB,$47,$D8,$70,$BF,$EC,$31,$FD,$9E,$75,$F3,$59,$C3
	.dc.b	$DF,$B7,$62,$3F,$B4,$A7,$CA,$E2,$3F,$D2,$99,$7E,$B1,$FF,$4C,$35
	.dc.b	$F8,$D5,$FE,$4A,$8F,$D8,$E1,$E5,$FB,$1F,$19,$CD,$FE,$17,$F6,$9F
	.dc.b	$E0,$AB,$FE,$06,$D7,$DE,$D8,$FE,$23,$C9,$FA,$E9,$FF,$76,$4E,$78
	.dc.b	$3F,$D5,$93,$FE,$4E,$AD,$FD,$37,$F8,$0A,$B5,$7C,$3E,$63,$04,$7F
	.dc.b	$89,$DB,$ED,$31,$1F,$42,$3F,$87,$BF,$EF,$B6,$AA,$FE,$5E,$FF,$78
	.dc.b	$F7,$BF,$48,$FF,$71,$6B,$FE,$FF,$6B,$FE,$F7,$6B,$FD,$BD,$E3,$8F
	.dc.b	$F3,$EF,$29,$F6,$8D,$7F,$C7,$AD,$BF,$E6,$5E,$B3,$0E,$3B,$7F,$95
	.dc.b	$EB,$55,$F9,$0C,$3F,$8D,$F4,$7F,$91,$BC,$23,$D3,$2C,$32,$BF,$C6
	.dc.b	$53,$A7,$FC,$69,$7D,$33,$FE,$8B,$F6,$9D,$BE,$43,$F6,$0B,$E4,$BF
	.dc.b	$DC,$11,$86,$F6,$FF,$B8,$E7,$83,$63,$38,$04,$F3,$71,$5C,$DB,$5C
	.dc.b	$F3,$E3,$0F,$B4,$7D,$6F,$E9,$A1,$7F,$C1,$67,$3D,$BC,$A7,$C1,$18
	.dc.b	$77,$87,$95,$D1,$D6,$FA,$5F,$36,$CF,$FB,$FE,$B9,$4F,$9E,$B9,$9D
	.dc.b	$AB,$6D,$BE,$B5,$C2,$FC,$13,$C0,$F0,$4E,$AB,$8C,$60,$D6,$3F,$1D
	.dc.b	$E1,$7C,$DE,$C6,$27,$8C,$EF,$0F,$55,$F5,$25,$74,$CB,$E6,$E1,$CE
	.dc.b	$3C,$15,$C1,$D1,$C1,$C1,$31,$F5,$B6,$3C,$6D,$7E,$FD,$76,$C7,$DB
	.dc.b	$7D,$CF,$06,$3E,$7C,$3E,$74,$E8,$98,$FD,$69,$D6,$1D,$AB,$73,$E5
	.dc.b	$7C,$66,$23,$E2,$46,$5F,$C9,$4F,$27,$DF,$3B,$C7,$D1,$39,$C3,$CD
	.dc.b	$87,$8A,$FE,$19,$D0,$95,$C3,$1F,$76,$3D,$E7,$A9,$E0,$8C,$BC,$31
	.dc.b	$F0,$1E,$87,$ED,$9F,$A7,$29,$EB,$30,$7E,$14,$B8,$3A,$5F,$59,$A9
	.dc.b	$FD,$0C,$61,$C4,$EE,$75,$1F,$AD,$9D,$3D,$1E,$18,$FE,$B5,$B7,$E5
	.dc.b	$FB,$4F,$F6,$9F,$1F,$E7,$A7,$37,$E0,$C3,$07,$EB,$AB,$71,$F9,$A9
	.dc.b	$DB,$E1,$9D,$BF,$A5,$9D,$E1,$E1,$9C,$B6,$75,$F0,$FF,$E6,$A8,$C1
	.dc.b	$31,$FD,$C6,$1D,$B0,$E7,$18,$C3,$9B,$70,$E7,$8E,$1F,$51,$E7,$9F
	.dc.b	$56,$3D,$EB,$86,$3D,$EB,$6F,$F8,$5F,$0E,$F1,$89,$D9,$1F,$E4,$91
	.dc.b	$A7,$E1,$1D,$C9,$3C,$EB,$C1,$1C,$C8,$70,$8D,$61,$E9,$8C,$5F,$75
	.dc.b	$7D,$F2,$FC,$14,$76,$9D,$4E,$72,$FA,$B5,$AC,$3D,$D7,$CB,$1F,$1B
	.dc.b	$D8,$F1,$CE,$F0,$FC,$34,$EA,$31,$3A,$BF,$A9,$39,$7B,$93,$2F,$9F
	.dc.b	$97,$AE,$F8,$DF,$73,$C5,$1E,$FC,$BE,$51,$3E,$D3,$CA,$D8,$E7,$3F
	.dc.b	$8D,$9C,$E3,$F3,$39,$5F,$68,$FB,$3F,$F5,$AF,$27,$F7,$98,$C6,$3F
	.dc.b	$DA,$BE,$D3,$2F,$F5,$DB,$E3,$78,$C9,$1F,$E9,$0A,$C3,$AA,$E4,$7F
	.dc.b	$AC,$5C,$4F,$F9,$33,$A9,$F4,$2F,$A5,$E3,$D4,$C7,$B5,$6A,$73,$18
	.dc.b	$AE,$18,$FF,$03,$AD,$36,$33,$87,$F0,$C9,$CB,$FC,$27,$EC,$DF,$63
	.dc.b	$EB,$B7,$E4,$6F,$0F,$DD,$6B,$DD,$7F,$B2,$CD,$CE,$A3,$33,$FE,$62
	.dc.b	$6F,$07,$FB,$8B,$95,$F6,$BF,$EF,$CC,$3F,$D9,$B5,$A7,$F9,$22,$F9
	.dc.b	$7F,$9D,$5C,$3A,$7B,$46,$2F,$FD,$19,$7C,$A3,$2F,$F8,$BB,$ED,$5F
	.dc.b	$5F,$8C,$FC,$9D,$FC,$7C,$CF,$43,$6F,$C6,$FF,$BA,$A7,$C6,$D7,$FC
	.dc.b	$3D,$A7,$F8,$5A,$7F,$96,$B6,$3E,$64,$F4,$9F,$F7,$86,$9F,$03,$E7
	.dc.b	$F1,$5F,$E8,$6F,$B5,$FF,$53,$D7,$FB,$D7,$77,$E0,$FF,$0E,$BF,$C5
	.dc.b	$95,$A7,$FD,$27,$3A,$E3,$FA,$B3,$98,$FF,$80,$2D,$BF,$F1,$A7,$23
	.dc.b	$76,$DC,$3D,$3E,$3C,$DF,$2F,$C4,$B8,$7A,$39,$3F,$D2,$EF,$F8,$F7
	.dc.b	$3E,$B5,$AD,$7F,$D4,$BC,$0F,$AF,$7C,$B6,$6F,$C7,$E7,$FD,$A9,$7C
	.dc.b	$D7,$A2,$30,$DC,$BF,$AB,$F0,$A7,$E8,$64,$8C,$4E,$A7,$31,$89,$F9
	.dc.b	$91,$DF,$8B,$A4,$7F,$AD,$3E,$C9,$F2,$5E,$3F,$63,$F3,$B1,$8C,$7E
	.dc.b	$BE,$69,$CD,$49,$CF,$DB,$5C,$3A,$7A,$DF,$B6,$FC,$63,$F3,$91,$8A
	.dc.b	$FF,$3F,$C6,$37,$47,$F8,$07,$3F,$F5,$CB,$F3,$4C,$5F,$F0,$DF,$B6
	.dc.b	$7D,$BB,$ED,$FC,$F3,$E8,$F0,$3E,$67,$8F,$70,$F4,$B7,$4F,$F0,$36
	.dc.b	$5F,$A2,$9F,$F0,$1B,$63,$35,$F3,$A3,$19,$F9,$A7,$FC,$F9,$18,$AF
	.dc.b	$F3,$F7,$37,$DF,$FB,$7E,$D7,$BB,$FC,$6F,$37,$DF,$F8,$7F,$E1,$3D
	.dc.b	$BF,$80,$C7,$E6,$C6,$26,$47,$D0,$F8,$EB,$7F,$EE,$AB,$47,$E2,$7C
	.dc.b	$5E,$6D,$AF,$91,$5F,$E9,$1A,$DC,$62,$3F,$3F,$3F,$91,$F7,$F8,$CE
	.dc.b	$69,$F6,$FF,$95,$FB,$3F,$3B,$67,$4F,$E4,$A7,$71,$FE,$2C,$6E,$1F
	.dc.b	$E3,$36,$BF,$E3,$A7,$3A,$7C,$56,$CF,$CA,$EA,$7C,$37,$82,$3F,$A3
	.dc.b	$7D,$BF,$F1,$CC,$3F,$D8,$4D,$C3,$FD,$9D,$A7,$F4,$8A,$DF,$97,$FD
	.dc.b	$1B,$E5,$BF,$63,$1F,$D0,$DB,$87,$F4,$16,$BF,$E5,$1F,$21,$18,$8F
	.dc.b	$EB,$DF,$31,$FE,$C3,$AF,$EB,$1F,$30,$F8,$AF,$EB,$B3,$9A,$FF,$6B
	.dc.b	$4E,$6B,$E8,$7A,$6F,$90,$FE,$92,$D8,$FA,$9D,$9F,$AF,$B6,$DF,$89
	.dc.b	$AD,$46,$2B,$FC,$CD,$5B,$8F,$F3,$2B,$5C,$F8,$7F,$C1,$A3,$11,$FE
	.dc.b	$D3,$7C,$8F,$33,$11,$FD,$72,$33,$7F,$99,$7C,$8F,$E8,$75,$7D,$BF
	.dc.b	$F0,$3C,$3F,$BD,$7C,$EF,$F0,$FA,$D4,$E7,$1F,$F7,$C4,$E6,$BF,$98
	.dc.b	$CE,$67,$E8,$35,$CF,$8D,$FD,$37,$0E,$ED,$9E,$98,$76,$8F,$F0,$6F
	.dc.b	$82,$FA,$FF,$F0,$2D,$6B,$1F,$F4,$67,$DB,$FF,$A1,$DF,$F0,$87,$67
	.dc.b	$F7,$B7,$F4,$91,$9C,$FE,$27,$B2,$AD,$47,$C7,$13,$B7,$FD,$2A,$D7
	.dc.b	$BF,$C0,$7F,$36,$8C,$40,$E3,$1F,$E0,$A8,$F0,$69,$FC,$36,$FC,$1E
	.dc.b	$D6,$6B,$FC,$59,$2F,$B7,$FA,$B9,$7F,$8C,$7E,$99,$F5,$1A,$9C,$E7
	.dc.b	$FE,$32,$9C,$D7,$F6,$B9,$FC,$0C,$FF,$AB,$EF,$A7,$4F,$F7,$B6,$D7
	.dc.b	$FA,$AD,$B3,$FE,$B2,$69,$FE,$46,$AE,$F9,$FF,$8E,$3C,$2F,$16,$1F
	.dc.b	$E2,$6F,$B6,$FF,$0C,$63,$FE,$3F,$FB,$77,$D0,$E7,$C6,$FF,$22,$C6
	.dc.b	$27,$F8,$B5,$70,$D3,$FD,$78,$D7,$FD,$86,$D7,$FD,$90,$D7,$E7,$3C
	.dc.b	$74,$FE,$64,$D9,$FF,$2C,$53,$DC,$F1,$47,$F8,$3F,$ED,$BF,$82,$DF
	.dc.b	$D3,$C5,$F6,$FF,$4A,$73,$5F,$E6,$19,$CB,$C7,$D8,$F9,$53,$FC,$27
	.dc.b	$3F,$DD,$B7,$03,$4E,$71,$A9,$FF,$33,$DF,$F8,$5E,$31,$7F,$CE,$7E
	.dc.b	$2B,$7C,$0F,$B7,$FE,$53,$5F,$8D,$9C,$C6,$21,$1E,$36,$E1,$FE,$0A
	.dc.b	$8C,$19,$8F,$CB,$7C,$5F,$E5,$FE,$CD,$F4,$FF,$C9,$7E,$DF,$FB,$14
	.dc.b	$62,$BF,$DE,$B5,$BA,$D6,$76,$72,$2B,$82,$F0,$58,$74,$BF,$F7,$D6
	.dc.b	$1C,$6B,$53,$97,$9D,$F2,$BE,$38,$6E,$7D,$27,$7C,$3C,$6D,$38,$B8
	.dc.b	$AF,$1C,$70,$7A,$5F,$78,$FD,$C0,$C3,$C9,$87,$82,$3A,$1A,$8C,$CF
	.dc.b	$18,$E9,$7D,$CF,$6B,$C2,$B9,$5F,$8A,$3B,$1F,$95,$9F,$4F,$0E,$70
	.dc.b	$FE,$14,$F9,$B1,$B3,$D2,$F1,$7E,$43,$C2,$7F,$98,$31,$F2,$3D,$A3
	.dc.b	$67,$57,$C9,$1A,$C3,$E0,$C7,$A0,$A7,$CE,$29,$E8,$37,$F6,$DF,$CC
	.dc.b	$FE,$DF,$2D,$8E,$2D,$77,$E0,$A7,$7C,$FF,$D8,$A7,$6D,$7A,$B5,$FA
	.dc.b	$B5,$FC,$33,$93,$89,$F8,$9F,$B7,$7C,$FF,$DA,$D3,$F2,$31,$8B,$FC
	.dc.b	$75,$F1,$7E,$DB,$67,$75,$C7,$5F,$C1,$DF,$1B,$DE,$3E,$7B,$FC,$EE
	.dc.b	$1C,$6F,$78,$71,$BF,$E8,$58,$71,$BD,$E1,$C6,$FD,$6D,$AD,$DF,$15
	.dc.b	$7C,$6B,$78,$FA,$AF,$85,$6A,$F8,$5E,$E3,$FA,$96,$3E,$56,$D7,$09
	.dc.b	$D4,$71,$3A,$5C,$7F,$3F,$1A,$37,$5C,$E0,$D8,$D5,$FA,$63,$46,$5F
	.dc.b	$29,$9B,$F4,$FD,$B3,$EB,$3F,$AB,$70,$7F,$69,$E6,$7D,$BF,$FA,$8F
	.dc.b	$3F,$F2,$FC,$63,$E1,$37,$1F,$DB,$F0,$E4,$FF,$98,$5A,$FF,$7E,$AE
	.dc.b	$4D,$9E,$2A,$77,$AF,$FA,$04,$7D,$BB,$E9,$3B,$C6,$1B,$19,$AF,$2F
	.dc.b	$5F,$F6,$9C,$FF,$D9,$13,$99,$FF,$07,$FD,$B3,$ED,$FD,$3A,$A3,$13
	.dc.b	$FE,$E3,$AE,$19,$78,$DB,$87,$C5,$AD,$D7,$4C,$3C,$79,$7B,$3D,$DF
	.dc.b	$F0,$3C,$FF,$A0,$DA,$FF,$87,$5B,$8F,$F8,$4F,$ED,$BF,$D8,$B3,$BF
	.dc.b	$23,$CB,$0F,$97,$A7,$58,$93,$F5,$27,$4D,$9C,$BF,$EF,$8F,$03,$FC
	.dc.b	$FE,$FF,$BF,$DA,$E7,$DC,$F8,$7F,$3E,$FB,$7F,$9F,$5A,$CB,$FC,$51
	.dc.b	$E6,$7F,$BE,$B0,$FC,$1B,$5E,$47,$5C,$3E,$6C,$66,$78,$FC,$07,$49
	.dc.b	$FC,$0F,$DB,$BE,$DE,$93,$0F,$67,$D5,$57,$FC,$8F,$B7,$F8,$71,$F8
	.dc.b	$1D,$99,$CB,$FD,$85,$87,$39,$C9,$0F,$F2,$63,$97,$FC,$E9,$E3,$BE
	.dc.b	$DB,$FD,$1D,$1F,$EB,$C8,$C7,$9F,$F9,$E6,$C7,$E8,$BE,$DF,$FD,$EC
	.dc.b	$EE,$B8,$BF,$E7,$23,$F1,$BE,$D7,$E4,$DF,$EE,$51,$AC,$7F,$CF,$CD
	.dc.b	$EA,$FD,$85,$6F,$B0,$F1,$C7,$C3,$87,$F9,$97,$DC,$FF,$34,$3F,$EC
	.dc.b	$FF,$B7,$FD,$4E,$1F,$DA,$9B,$1F,$E6,$39,$DB,$FE,$6D,$8C,$A7,$FC
	.dc.b	$0F,$97,$FB,$AD,$B5,$CF,$C3,$FD,$BC,$62,$BF,$CD,$CD,$79,$78,$FA
	.dc.b	$9F,$D6,$FC,$07,$FA,$FA,$7F,$D7,$7F,$05,$83,$F6,$3F,$05,$F6,$E3
	.dc.b	$E9,$FD,$B3,$ED,$FF,$6B,$F0,$DF,$5F,$5F,$F6,$8F,$4F,$D3,$AD,$5F
	.dc.b	$F1,$4F,$B7,$78,$7F,$BC,$6C,$66,$39,$D7,$F1,$0F,$B5,$7D,$BD,$25
	.dc.b	$7F,$BE,$7E,$DD,$F4,$58,$8F,$F4,$A4,$31,$FE,$32,$D8,$FF,$80,$17
	.dc.b	$D1,$B7,$AA,$91,$83,$FD,$2D,$9B,$93,$FC,$43,$F6,$9F,$ED,$99,$CC
	.dc.b	$7F,$89,$3E,$D5,$F0,$7F,$CA,$1B,$3A,$3F,$D4,$15,$C5,$FF,$55,$36
	.dc.b	$FB,$47,$FA,$96,$78,$BF,$EA,$F6,$C6,$A1,$3C,$6F,$CB,$F6,$DE,$1F
	.dc.b	$8B,$95,$E1,$37,$1F,$CF,$28,$EA,$3B,$4E,$63,$5E,$6F,$F4,$2F,$1B
	.dc.b	$FB,$F6,$3E,$16,$E3,$F1,$27,$78,$77,$9C,$BF,$EB,$F6,$BE,$88,$DB
	.dc.b	$FE,$CB,$8C,$9A,$C7,$F9,$DC,$7F,$B1,$F3,$FD,$93,$CF,$97,$C8,$7F
	.dc.b	$56,$8C,$C7,$D2,$F0,$B2,$F8,$E7,$51,$BF,$1B,$FA,$EF,$DB,$3E,$CA
	.dc.b	$99,$BF,$91,$82,$B3,$1F,$2E,$B8,$3A,$8F,$17,$C8,$48,$E2,$F6,$FA
	.dc.b	$3C,$47,$FB,$A6,$9F,$63,$E2,$33,$C1,$F7,$71,$61,$9B,$F0,$56,$B8
	.dc.b	$3F,$CE,$7F,$69,$FC,$7A,$7F,$DE,$9F,$6C,$F6,$3F,$BC,$B6,$3F,$05
	.dc.b	$3B,$CB,$F1,$4D,$BF,$F3,$ED,$6E,$B9,$DF,$F8,$0A,$B5,$1F,$EF,$97
	.dc.b	$F8,$BD,$FE,$4F,$ED,$BF,$A5,$61,$FE,$86,$6D,$6F,$E4,$39,$1F,$E0
	.dc.b	$C8,$CC,$1B,$7D,$2B,$DB,$A7,$84,$95,$F9,$BA,$7C,$96,$21,$8F,$8F
	.dc.b	$EC,$FB,$19,$7D,$F7,$89,$E5,$39,$C3,$FD,$3B,$F6,$BF,$DB,$67,$F5
	.dc.b	$1C,$7F,$E2,$4C,$3F,$60,$DA,$DF,$AF,$FE,$26,$BE,$17,$E2,$9F,$F5
	.dc.b	$3D,$6E,$07,$E3,$FD,$FE,$17,$C7,$C2,$FF,$17,$D7,$EE,$EB,$6D,$58
	.dc.b	$7F,$20,$7F,$78,$D7,$C0,$70,$F9,$7E,$67,$57,$FB,$47,$B5,$F9,$B7
	.dc.b	$FC,$8F,$F6,$AD,$7F,$C5,$ED,$8D,$F4,$FF,$6D,$AE,$11,$DB,$D3,$56
	.dc.b	$E3,$0D,$DF,$F8,$DF,$A4,$BE,$31,$FC,$7A,$DE,$17,$E8,$B0,$FF,$2C
	.dc.b	$B5,$EC,$D9,$E1,$87,$5E,$1F,$F3,$14,$E5,$BD,$3F,$E4,$88,$C5,$B0
	.dc.b	$F3,$E9,$FE,$5C,$FB,$57,$C6,$FF,$B5,$FB,$3F,$DB,$B1,$8B,$FF,$80
	.dc.b	$31,$89,$D4,$7D,$A9,$F8,$1F,$17,$FB,$68,$C4,$E9,$4F,$37,$FD,$D1
	.dc.b	$5C,$DE,$A7,$3B,$E1,$F6,$AF,$AF,$7D,$6E,$A9,$39,$8C,$36,$3A,$9E
	.dc.b	$23,$33,$C9,$C9,$C9,$C6,$5E,$98,$C5,$77,$C7,$D1,$39,$BE,$CD,$BE
	.dc.b	$38,$FE,$73,$0E,$27,$6B,$E3,$19,$BE,$2D,$AE,$35,$FA,$49,$CB,$E9
	.dc.b	$3C,$EE,$1E,$B7,$E1,$C3,$CD,$97,$C8,$BE,$75,$AC,$3B,$56,$A3,$E1
	.dc.b	$B1,$EB,$3C,$31,$F3,$C7,$8E,$39,$E3,$E1,$3C,$53,$A9,$EC,$6E,$BA
	.dc.b	$63,$F2,$E7,$55,$C6,$F9,$9D,$F1,$F4,$5F,$0C,$7C,$F8,$FC,$99,$CB
	.dc.b	$E6,$9D,$3F,$17,$1F,$84,$E4,$8F,$86,$78,$BA,$25,$7E,$46,$76,$F8
	.dc.b	$23,$31,$D6,$30,$47,$DC,$79,$4F,$19,$F0,$89,$8F,$C0,$C3,$DF,$86
	.dc.b	$A3,$0F,$E1,$E7,$DE,$5C,$7F,$34,$DA,$FC,$C3,$9E,$A7,$D1,$71,$8F
	.dc.b	$C4,$46,$1C,$57,$F8,$CA,$75,$1F,$A8,$9D,$3C,$E3,$F5,$31,$C7,$2F
	.dc.b	$D1,$7D,$AF,$FA,$EF,$87,$35,$2F,$F2,$77,$BD,$3F,$41,$7B,$9C,$D6
	.dc.b	$F0,$FD,$45,$EE,$7F,$A0,$D6,$E7,$35,$BB,$FD,$35,$EE,$D7,$B9,$CD
	.dc.b	$6E,$7E,$0B,$FC,$16,$75,$18,$3C,$64,$BF,$6F,$C2,$7F,$93,$9F,$0C
	.dc.b	$66,$FA,$BF,$C1,$AB,$9B,$F2,$6F,$C0,$D3,$57,$CE,$B8,$DF,$42,$3F
	.dc.b	$1A,$35,$3E,$59,$CB,$67,$58,$F6,$CB,$ED,$4F,$94,$F9,$37,$C9,$F1
	.dc.b	$5F,$26,$DF,$27,$81,$83,$B1,$CA,$FB,$65,$F3,$F0,$EF,$7F,$1A,$73
	.dc.b	$8F,$C0,$CB,$E1,$D6,$B1,$F7,$93,$1E,$87,$82,$34,$F8,$0F,$B7,$97
	.dc.b	$A6,$33,$1E,$52,$57,$6C,$BF,$04,$FD,$22,$7D,$AF,$C1,$71,$74,$C4
	.dc.b	$5C,$BF,$A6,$F0,$BE,$D5,$F6,$BF,$E8,$EE,$7F,$EE,$51,$8A,$FE,$E3
	.dc.b	$F3,$B9,$7F,$C2,$F7,$C6,$F1,$A7,$F6,$BB,$E0,$FF,$80,$DB,$87,$F7
	.dc.b	$DE,$66,$3C,$30,$E5,$8F,$FB,$E7,$EC,$1F,$45,$9B,$DC,$E6,$31,$7E
	.dc.b	$0E,$C5,$F0,$AF,$39,$B2,$E7,$FE,$22,$1F,$68,$FB,$0F,$4C,$BE,$D7
	.dc.b	$FD,$92,$FF,$A9,$67,$4F,$FA,$DC,$D7,$89,$F0,$2F,$9D,$F2,$BE,$31
	.dc.b	$DE,$F8,$BF,$E3,$2B,$E2,$75,$F2,$FE,$4F,$43,$EC,$FF,$CF,$B3,$93
	.dc.b	$FD,$8D,$19,$E9,$F0,$61,$FE,$7B,$9C,$D7,$0C,$3C,$18,$74,$8C,$71
	.dc.b	$7F,$B1,$6F,$FC,$15,$7C,$3E,$33,$9F,$66,$DB,$19,$3E,$14,$65,$A4
	.dc.b	$C7,$FC,$9B,$9F,$A7,$E7,$38,$C7,$CF,$FB,$37,$A5,$F8,$6C,$DF,$6B
	.dc.b	$F1,$1F,$F6,$C4,$EB,$1F,$F7,$24,$6B,$87,$E7,$4E,$70,$FF,$2D,$CE
	.dc.b	$6B,$FD,$B5,$39,$8F,$CD,$5F,$59,$FF,$09,$4F,$E1,$9F,$8F,$E0,$F4
	.dc.b	$E4,$F8,$CD,$8C,$9D,$23,$2F,$6F,$37,$EB,$FD,$53,$E0,$7C,$B1,$86
	.dc.b	$CF,$D8,$9D,$3C,$4C,$46,$B3,$F8,$D5,$D3,$4C,$69,$FE,$FB,$FA,$E7
	.dc.b	$C5,$F8,$3E,$AB,$F2,$11,$8C,$BB,$46,$27,$51,$FE,$FD,$9D,$63,$F5
	.dc.b	$B3,$5F,$2F,$4F,$F5,$5F,$66,$F8,$9C,$63,$FE,$94,$CF,$FC,$EF,$39
	.dc.b	$8E,$95,$C2,$B7,$8F,$FA,$3B,$B9,$7B,$8F,$D8,$66,$FB,$0F,$93,$E1
	.dc.b	$7A,$9B,$97,$9D,$B5,$C1,$B9,$FC,$36,$CF,$F9,$36,$FA,$3F,$3E,$FF
	.dc.b	$69,$D5,$F4,$30,$F8,$1F,$5F,$8E,$3F,$B2,$DB,$FA,$CD,$9D,$C7,$C1
	.dc.b	$3F,$CB,$54,$F8,$AF,$C0,$F7,$F6,$6C,$EE,$BF,$D7,$75,$B8,$FC,$15
	.dc.b	$6D,$A6,$E5,$2F,$AF,$FC,$6F,$D5,$FA,$9B,$5B,$9D,$3F,$8D,$9D,$34
	.dc.b	$D4,$E2,$BF,$01,$F0,$77,$7F,$F5,$77,$D7,$FF,$04,$BF,$E7,$6D,$AF
	.dc.b	$F5,$26,$7F,$A7,$7F,$D7,$91,$9E,$2F,$C3,$7C,$D7,$F3,$9A,$D6,$1F
	.dc.b	$EB,$CE,$44,$E7,$AF,$F9,$F7,$C8,$3E,$12,$9F,$D0,$7E,$43,$FD,$5F
	.dc.b	$AF,$C9,$F3,$55,$B8,$F8,$DE,$6B,$DD,$FE,$60,$D8,$F9,$5D,$1F,$9E
	.dc.b	$D5,$1F,$E9,$19,$D7,$0F,$F4,$1A,$D4,$E6,$FF,$9D,$46,$27,$FD,$E7
	.dc.b	$5C,$2F,$9C,$7F,$BD,$A7,$C8,$F3,$EF,$E3,$A7,$27,$FA,$A7,$D7,$FE
	.dc.b	$BB,$2F,$67,$CA,$BC,$3C,$F8,$3F,$D8,$63,$13,$F8,$9E,$55,$F1,$CB
	.dc.b	$FD,$D9,$5A,$7F,$5D,$19,$D3,$FD,$D1,$39,$7F,$DD,$B1,$9B,$52,$FA
	.dc.b	$E1,$FD,$C7,$C1,$7D,$3F,$92,$FF,$29,$F5,$EF,$63,$FB,$B7,$C3,$FF
	.dc.b	$5C,$D3,$E1,$E6,$AD,$5F,$D9,$AD,$3F,$EC,$18,$CC,$AB,$FD,$FD,$5A
	.dc.b	$8F,$2C,$EA,$73,$E9,$FF,$6D,$8C,$5F,$3C,$3C,$7E,$A6,$5F,$E9,$9F
	.dc.b	$5F,$F5,$30,$FA,$FD,$0F,$9E,$7C,$FE,$B9,$BF,$C2,$11,$8C,$BF,$A7
	.dc.b	$E1,$C5,$FF,$0E,$C6,$75,$FE,$9F,$39,$58,$75,$C3,$A5,$F2,$C3,$FC
	.dc.b	$39,$D8,$F0,$F3,$87,$F8,$47,$EB,$FF,$BC,$3F,$0E,$FF,$C5,$7F,$5E
	.dc.b	$FA,$0F,$F0,$65,$6B,$6F,$5C,$E6,$BF,$84,$4E,$5E,$35,$C3,$C8,$FF
	.dc.b	$16,$56,$F0,$F0,$46,$6B,$F5,$D2,$F5,$31,$5F,$DB,$3E,$BB,$F7,$35
	.dc.b	$FB,$3B,$7D,$7F,$F7,$99,$CD,$7A,$4F,$87,$3A,$AF,$EF,$33,$AC,$FE
	.dc.b	$9C,$E6,$BB,$78,$BF,$E4,$18,$C4,$FF,$94,$6B,$85,$FF,$51,$6D,$F6
	.dc.b	$F5,$BC,$58,$7F,$8C,$7E,$17,$3C,$6F,$AF,$FE,$4B,$7F,$E3,$F9,$CC
	.dc.b	$63,$2F,$CC,$C6,$3C,$DE,$58,$7F,$94,$EB,$59,$FF,$B2,$FE,$17,$F5
	.dc.b	$3F,$5A,$FA,$0F,$D5,$E7,$FA,$BF,$AF,$FE,$0B,$39,$8F,$F6,$D5,$70
	.dc.b	$AD,$F8,$8B,$DC,$75,$A5,$F1,$7A,$61,$C5,$E6,$6A,$79,$D7,$28,$CD
	.dc.b	$71,$BE,$D7,$C6,$7C,$F8,$59,$CC,$73,$38,$E1,$D2,$B8,$47,$F0,$8A
	.dc.b	$E1,$87,$78,$E6,$71,$3C,$E5,$38,$19,$BE,$F8,$7B,$63,$2F,$9E,$7A
	.dc.b	$46,$30,$F1,$3C,$C5,$F8,$B0,$F2,$C6,$0E,$D5,$C9,$C5,$F5,$9D,$D7
	.dc.b	$19,$E8,$D9,$F2,$47,$0B,$F0,$C6,$27,$EF,$CF,$D7,$29,$EE,$29,$E3
	.dc.b	$37,$F5,$BF,$CB,$3E,$BF,$2D,$8E,$2D,$77,$E2,$27,$7D,$1F,$C9,$E7
	.dc.b	$6D,$7A,$B5,$FC,$9D,$7F,$09,$E1,$E2,$7E,$07,$EB,$DF,$5F,$F9,$8C
	.dc.b	$FF,$1D,$18,$BF,$C3,$5F,$17,$EE,$36,$77,$5C,$74,$FD,$0D,$F1,$BD
	.dc.b	$E3,$E7,$BF,$CD,$61,$C6,$F7,$87,$1B,$FE,$A7,$87,$1B,$DE,$1C,$6F
	.dc.b	$D6,$DA,$DD,$F1,$57,$C6,$B7,$8F,$AA,$F8,$56,$AF,$85,$EE,$3F,$B1
	.dc.b	$E3,$E5,$6D,$70,$9D,$47,$13,$A5,$C7,$F4,$11,$A3,$75,$CE,$0D,$8D
	.dc.b	$5F,$A6,$34,$65,$F2,$99,$BF,$81,$F7,$4F,$BA,$79,$FF,$C4,$32,$FF
	.dc.b	$09,$FD,$DB,$ED,$BF,$DC,$FA,$7F,$94,$63,$1E,$96,$F8,$FF,$C0,$39
	.dc.b	$73,$C7,$16,$8C,$3C,$70,$E6,$FF,$3D,$C3,$9B,$FE,$5D,$9D,$BF,$D0
	.dc.b	$7A,$9F,$60,$F9,$FE,$B1,$8F,$52,$CF,$FA,$6A,$BF,$C7,$CE,$DB,$5A
	.dc.b	$99,$18,$21,$DE,$79,$36,$3A,$CF,$B2,$FF,$D9,$FF,$45,$EE,$C7,$FB
	.dc.b	$BF,$9B,$72,$3B,$F0,$15,$4D,$44,$CB,$E0,$4E,$9B,$39,$7E,$35,$74
	.dc.b	$72,$7F,$96,$FE,$63,$E5,$F0,$FF,$85,$1F,$F2,$D4,$67,$D7,$F8,$F1
	.dc.b	$AE,$1F,$F1,$BD,$A3,$15,$FE,$10,$78,$BF,$45,$B7,$D7,$8B,$99,$D3
	.dc.b	$09,$5E,$38,$F8,$DF,$79,$FE,$90,$8F,$C4,$75,$3E,$F7,$E4,$36,$3F
	.dc.b	$01,$3B,$7F,$C0,$1C,$08,$C6,$3F,$94,$AE,$15,$F8,$96,$C7,$F9,$AA
	.dc.b	$B8,$7C,$07,$88,$EF,$84,$BE,$F1,$F8,$AF,$AF,$FD,$0C,$7F,$A1,$57
	.dc.b	$DC,$E4,$D7,$44,$9D,$46,$5E,$11,$97,$F3,$13,$C9,$D9,$A3,$F3,$DF
	.dc.b	$5F,$F4,$31,$34,$9B,$7E,$3D,$FA,$BF,$7F,$FE,$11,$F0,$BF,$5A,$DB
	.dc.b	$E7,$C1,$8F,$1B,$FC,$DF,$1F,$EC,$79,$D3,$FE,$2B,$6E,$FF,$1F,$0E
	.dc.b	$BF,$85,$70,$AD,$36,$7F,$D2,$B4,$8C,$72,$FF,$00,$EB,$79,$9F,$B4
	.dc.b	$F8,$07,$CA,$7F,$02,$F2,$9F,$25,$FE,$43,$D3,$FD,$24,$D7,$F7,$D5
	.dc.b	$C1,$FF,$5B,$7C,$2F,$F9,$11,$BE,$8F,$F0,$08,$C4,$7F,$91,$6F,$AF
	.dc.b	$A5,$FC,$42,$1F,$55,$FE,$DF,$7F,$DC,$4B,$E6,$34,$D8,$CF,$A5,$83
	.dc.b	$33,$23,$24,$3F,$CA,$87,$67,$4E,$27,$C0,$FF,$1D,$F8,$C7,$DD,$7F
	.dc.b	$07,$6C,$7F,$2B,$6B,$FA,$0F,$C1,$5F,$27,$BA,$93,$B7,$F8,$94,$EA
	.dc.b	$3F,$94,$3F,$A2,$E3,$FF,$59,$F9,$2F,$AE,$FF,$55,$F4,$7F,$3C,$6C
	.dc.b	$FF,$38,$AE,$17,$FC,$DD,$B7,$C6,$7F,$8F,$57,$0A,$D7,$87,$F3,$FE
	.dc.b	$0B,$FD,$E1,$1F,$EC,$2C,$7F,$D7,$31,$88,$F8,$ED,$8F,$F5,$EF,$C7
	.dc.b	$3E,$77,$97,$C0,$CB,$E5,$97,$F9,$8B,$0E,$0F,$F9,$92,$F8,$4E,$8F
	.dc.b	$F6,$8C,$E9,$FF,$69,$C7,$F4,$17,$EB,$B7,$0F,$2D,$FF,$3B,$F4,$BF
	.dc.b	$2B,$5C,$FC,$C7,$E1,$3F,$E1,$1B,$FF,$5F,$B7,$2F,$EB,$F7,$C2,$FF
	.dc.b	$8B,$B6,$B7,$EA,$FF,$56,$BE,$15,$FD,$8E,$B5,$7F,$DA,$3E,$D9,$F3
	.dc.b	$5F,$D2,$FE,$83,$FD,$79,$7F,$DB,$DA,$FF,$24,$9D,$BF,$D9,$DA,$FF
	.dc.b	$6B,$8D,$E5,$FD,$E9,$B7,$DE,$FF,$4D,$F5,$9F,$96,$AF,$F7,$7B,$FA
	.dc.b	$3B,$FC,$EB,$63,$3D,$5C,$0E,$3D,$52,$BE,$90,$AE,$98,$7F,$A6,$9B
	.dc.b	$03,$53,$C5,$EF,$5F,$DF,$3E,$57,$D7,$87,$FB,$B9,$FF,$4E,$4F,$D5
	.dc.b	$6B,$FE,$05,$77,$87,$F5,$C6,$C7,$FB,$A9,$DC,$F3,$F8,$0E,$D9,$7F
	.dc.b	$82,$BF,$3F,$DF,$09,$DE,$3F,$D9,$9B,$8F,$FB,$D1,$D9,$2F,$91,$CE
	.dc.b	$FF,$89,$7D,$67,$E5,$AB,$F8,$3B,$F9,$54,$D3,$58,$B0,$FF,$60,$C6
	.dc.b	$1B,$C7,$D0,$EB,$CD,$2F,$E8,$D7,$39,$C4,$7F,$8A,$7F,$42,$7C,$57
	.dc.b	$F8,$A1,$7D,$95,$31,$12,$7F,$45,$3A,$6C,$E5,$FF,$1A,$7E,$C7,$F9
	.dc.b	$A5,$BA,$3E,$FB,$78,$7F,$C6,$F1,$B8,$D7,$91,$89,$95,$FB,$1A,$E9
	.dc.b	$E9,$B9,$FE,$8F,$7B,$A1,$CD,$FA,$73,$FC,$69,$3F,$5A,$B8,$72,$E7
	.dc.b	$1F,$D5,$56,$A3,$F5,$DF,$7F,$27,$F5,$CE,$DF,$D5,$39,$69,$C8,$FC
	.dc.b	$FF,$D4,$BE,$2B,$FD,$C3,$E6,$FF,$BE,$23,$13,$FE,$DD,$8C,$4E,$A3
	.dc.b	$F4,$52,$FA,$AC,$42,$38,$BF,$ED,$4A,$E2,$F3,$38,$D7,$09,$E5,$19
	.dc.b	$1F,$55,$FD,$C3,$C8,$C3,$63,$A1,$FB,$37,$1F,$17,$FD,$C6,$3C,$57
	.dc.b	$E4,$9C,$E1,$FB,$99,$EB,$C7,$FD,$46,$F8,$56,$A7,$CA,$78,$27,$55
	.dc.b	$C6,$75,$3C,$B0,$F4,$63,$F0,$30,$E7,$87,$0A,$F3,$35,$E0,$7A,$8F
	.dc.b	$39,$DE,$A3,$9B,$EF,$7D,$70,$F1,$5F,$39,$F1,$9A,$C7,$E0,$46,$49
	.dc.b	$87,$9A,$32,$F6,$C7,$CE,$E7,$1C,$61,$DA,$2C,$7C,$43,$C9,$7D,$1B
	.dc.b	$1C,$5A,$4A,$EA,$FB,$F1,$F5,$3F,$12,$7A,$BF,$12,$7E,$D4,$EA,$31
	.dc.b	$7E,$37,$A3,$F8,$C6,$9C,$CA,$7C,$C2,$E3,$F1,$B0,$FA,$55,$D6,$31
	.dc.b	$87,$8A,$31,$3E,$DC,$3D,$27,$93,$1F,$93,$87,$49,$E7,$18,$8F,$C7
	.dc.b	$1F,$78,$B8,$7E,$65,$B3,$FE,$21,$73,$E2,$FF,$62,$7F,$1C,$D7,$F3
	.dc.b	$CD,$7F,$4A,$D9,$3F,$DB,$2F,$F3,$6D,$CB,$F4,$1C,$3F,$E1,$7E,$CF
	.dc.b	$D5,$CE,$70,$EB,$DD,$83,$F6,$55,$B8,$FC,$94,$ED,$F0,$CE,$DF,$CD
	.dc.b	$4E,$F0,$F0,$CE,$5B,$3A,$ED,$FE,$8C,$F4,$9D,$63,$F0,$1F,$0C,$EA
	.dc.b	$B8,$92,$75,$5C,$F1,$E2,$FA,$8E,$86,$4B,$3C,$5D,$3F,$C1,$F0,$F4
	.dc.b	$57,$48,$D4,$FB,$8E,$4E,$32,$F9,$F8,$F8,$27,$F9,$E6,$3E,$7B,$E4
	.dc.b	$48,$EE,$FD,$28,$D4,$60,$87,$7B,$F2,$46,$A3,$13,$CD,$A7,$F8,$4F
	.dc.b	$0F,$4B,$CB,$0F,$2B,$F0,$27,$C1,$8F,$9E,$73,$87,$BA,$F8,$DF,$42
	.dc.b	$4E,$AF,$8C,$73,$3B,$65,$F8,$3A,$E3,$5B,$9C,$E5,$F6,$B1,$F8,$D5
	.dc.b	$D8,$9F,$55,$ED,$71,$9D,$31,$16,$7F,$A4,$73,$3E,$D1,$3F,$8A,$F0
	.dc.b	$3F,$17,$39,$8C,$57,$FA,$2A,$9D,$F8,$CD,$7B,$AF,$F3,$93,$61,$87
	.dc.b	$4B,$EF,$5A,$79,$B7,$0E,$9D,$1F,$DF,$E7,$B6,$1E,$7C,$73,$D3,$E9
	.dc.b	$C3,$94,$62,$B6,$74,$CB,$FC,$CC,$E2,$3D,$1D,$9F,$E1,$5F,$A5,$BD
	.dc.b	$8F,$A2,$B1,$86,$E3,$FE,$13,$6E,$3F,$E2,$07,$9F,$E0,$5F,$EA,$58
	.dc.b	$F7,$CF,$F9,$77,$04,$FF,$0C,$8F,$F4,$63,$BD,$FF,$73,$3A,$E4,$E7
	.dc.b	$E6,$74,$6D,$F0,$7E,$87,$E1,$4F,$91,$7C,$E6,$1F,$F7,$8D,$7F,$A8
	.dc.b	$E7,$53,$9F,$67,$4D,$9D,$57,$26,$CE,$9A,$6A,$78,$69,$EE,$3F,$CC
	.dc.b	$5F,$03,$F3,$3C,$A7,$D9,$D8,$C3,$6F,$FD,$59,$3A,$6C,$E9,$FF,$3E
	.dc.b	$6F,$F5,$BC,$67,$C0,$D8,$C3,$67,$FD,$04,$DC,$7F,$D0,$4F,$C3,$8F
	.dc.b	$89,$F8,$25,$7D,$ED,$9F,$56,$DF,$2E,$31,$C5,$FC,$B2,$31,$5F,$18
	.dc.b	$E6,$53,$99,$FE,$B1,$F8,$DF,$CB,$F8,$1F,$E9,$D8,$C1,$FA,$47,$18
	.dc.b	$7F,$AE,$67,$6F,$E7,$1B,$3F,$EB,$BF,$A3,$E7,$D5,$FA,$06,$C6,$4E
	.dc.b	$91,$96,$9F,$EB,$2F,$65,$F4,$1F,$B7,$F7,$5F,$50,$FA,$6F,$DD,$66
	.dc.b	$BD,$E1,$FE,$EB,$BD,$D6,$A7,$35,$F8,$13,$86,$1F,$C0,$FE,$51,$F8
	.dc.b	$1B,$DF,$E1,$33,$2F,$23,$B9,$D4,$60,$FF,$7D,$78,$AF,$7D,$F8,$07
	.dc.b	$F0,$2F,$C2,$9E,$AF,$F0,$69,$CD,$FF,$10,$9C,$BC,$A3,$3A,$7E,$5F
	.dc.b	$C9,$56,$A3,$F3,$58,$F7,$C7,$B6,$3F,$C4,$3E,$BF,$F7,$39,$7F,$1C
	.dc.b	$6D,$F0,$C9,$F1,$BB,$EC,$97,$B7,$CD,$18,$AF,$C7,$9F,$82,$EA,$FC
	.dc.b	$73,$FF,$00,$2B,$F7,$0E,$5F,$F7,$37,$57,$FB,$9F,$CF,$FE,$03,$39
	.dc.b	$11,$99,$E3,$1A,$9D,$B6,$32,$76,$8F,$E4,$7F,$94,$7F,$15,$9F,$DE
	.dc.b	$79,$08,$C6,$1F,$E0,$EE,$87,$E5,$2F,$91,$E3,$E4,$FF,$2C,$F8,$B7
	.dc.b	$CC,$FE,$3E,$B5,$AF,$FB,$FF,$1F,$E3,$F5,$A8,$FC,$74,$EB,$4F,$E1
	.dc.b	$9F,$1E,$FC,$1B,$F8,$DC,$61,$B8,$7F,$88,$39,$DF,$1E,$F4,$7F,$AA
	.dc.b	$61,$C7,$CE,$BE,$73,$E9,$29,$FD,$3A,$FF,$DD,$DA,$F9,$46,$8C,$3F
	.dc.b	$96,$D6,$E7,$89,$F9,$6A,$E3,$5E,$8B,$ED,$39,$38,$1B,$F2,$AC,$FC
	.dc.b	$EF,$05,$F7,$1F,$D5,$17,$C6,$D8,$C3,$6F,$CD,$5F,$EE,$EA,$DC,$63
	.dc.b	$A3,$F7,$F7,$C2,$31,$3D,$9B,$7C,$EF,$C4,$FF,$70,$F9,$9F,$F6,$D4
	.dc.b	$FF,$7A,$FA,$C7,$E0,$19,$F8,$1F,$EE,$71,$89,$FD,$BD,$70,$C3,$F5
	.dc.b	$D3,$97,$F5,$B1,$9E,$4F,$F0,$6C,$E5,$FB,$51,$9A,$FA,$3E,$32,$B5
	.dc.b	$1F,$53,$95,$5B,$C3,$FC,$23,$5A,$5B,$3E,$27,$FA,$E4,$7E,$71,$FC
	.dc.b	$4E,$1F,$DD,$5C,$E5,$FD,$C9,$CC,$F2,$73,$3B,$E5,$FF,$04,$C6,$27
	.dc.b	$51,$8E,$2C,$B6,$BE,$A5,$74,$CD,$F3,$D2,$7F,$C4,$F9,$5C,$7F,$89
	.dc.b	$CD,$CB,$FC,$67,$36,$BE,$14,$DA,$FF,$18,$CD,$BE,$B3,$FE,$28,$8C
	.dc.b	$3B,$D7,$E2,$BF,$E2,$E3,$11,$CC,$64,$FB,$8F,$F1,$3F,$E6,$9F,$9E
	.dc.b	$CF,$FC,$2F,$39,$CF,$F2,$13,$97,$FC,$41,$19,$CF,$F8,$5C,$E5,$FA
	.dc.b	$B1,$9A,$FF,$10,$CE,$70,$F0,$E1,$E0,$C3,$FC,$6F,$7B,$7F,$C5,$95
	.dc.b	$CF,$37,$E1,$2F,$96,$FF,$23,$D7,$F4,$CC,$FF,$9E,$CE,$63,$FC,$AA
	.dc.b	$67,$3F,$E1,$51,$99,$FE,$8D,$39,$3F,$21,$19,$9F,$F3,$2C,$66,$7F
	.dc.b	$DA,$BF,$0D,$FC,$23,$E7,$DF,$4D,$FC,$2A,$5E,$2F,$F4,$F9,$CC,$70
	.dc.b	$0E,$1D,$F9,$BF,$EB,$C8,$C4,$7F,$14,$BE,$D7,$D6,$FA,$3F,$EE,$69
	.dc.b	$4E,$9F,$E3,$73,$A7,$C1,$5D,$E3,$11,$E7,$9E,$0F,$8B,$0A,$F8,$4F
	.dc.b	$3D,$F8,$70,$F3,$3E,$1A,$E3,$7D,$2B,$86,$1D,$EF,$71,$E2,$BE,$58
	.dc.b	$76,$8F,$49,$D8,$F4,$CF,$07,$CC,$64,$F3,$14,$EF,$3D,$1C,$1D,$2B
	.dc.b	$C5,$5D,$F0,$F3,$DF,$5A,$E2,$78,$27,$4F,$7C,$3D,$8F,$3C,$3B,$61
	.dc.b	$F5,$A3,$D2,$7D,$32,$9F,$14,$C1,$83,$FA,$7F,$D2,$FF,$82,$BE,$03
	.dc.b	$83,$63,$EE,$35,$FC,$60,$70,$FE,$3D,$7D,$2F,$FA,$C2,$38,$B5,$E3
	.dc.b	$1A,$D3,$BC,$E6,$79,$3C,$6F,$F2,$9F,$4C,$F9,$9F,$CE,$46,$1B,$97
	.dc.b	$0C,$3F,$13,$8F,$5C,$7E,$B6,$1C,$5F,$C6,$B5,$FB,$51,$BC,$3F,$44
	.dc.b	$DB,$FD,$55,$6E,$7C,$15,$DB,$1F,$35,$78,$63,$BE,$3E,$47,$FB,$DD
	.dc.b	$F1,$9D,$5F,$6B,$E3,$5A,$BE,$0A,$F8,$3F,$5A,$76,$FD,$16,$C6,$A3
	.dc.b	$84,$F2,$C3,$D5,$1C,$71,$F8,$33,$DA,$B8,$56,$AB,$A5,$69,$87,$C6
	.dc.b	$BF,$0C,$EE,$BC,$07,$28,$F5,$F0,$BE,$63,$FC,$6D,$7F,$EF,$4F,$63
	.dc.b	$FD,$56,$DF,$80,$FE,$2D,$97,$F8,$AE,$B7,$3F,$E6,$46,$C3,$83,$FB
	.dc.b	$AD,$F2,$AF,$F3,$63,$67,$97,$9D,$D2,$BF,$DC,$5F,$4D,$FE,$44,$BF
	.dc.b	$F0,$B3,$FE,$38,$6F,$8F,$B7,$FB,$CC,$6D,$FF,$29,$B6,$FF,$DC,$EF
	.dc.b	$18,$CC,$F1,$EB,$4E,$EF,$C3,$E7,$66,$38,$61,$CA,$F1,$3E,$1D,$7F
	.dc.b	$BA,$FD,$37,$8F,$BF,$FC,$17,$F0,$37,$33,$BF,$01,$54,$D4,$4D,$7F
	.dc.b	$D3,$73,$A6,$CE,$5F,$F4,$A5,$74,$72,$7F,$8E,$FE,$9B,$E6,$6D,$FE
	.dc.b	$54,$94,$3E,$17,$E0,$36,$B8,$71,$E0,$CC,$C8,$C9,$0F,$F3,$FC,$F1
	.dc.b	$6C,$71,$7F,$C7,$BF,$4D,$F0,$FA,$7F,$CA,$B1,$FE,$52,$D7,$FA,$CF
	.dc.b	$95,$F2,$7C,$85,$D3,$84,$4A,$E6,$43,$FD,$69,$7D,$1B,$1C,$5F,$F2
	.dc.b	$E7,$D2,$FE,$4B,$5F,$84,$BE,$9B,$F0,$0D,$9D,$F3,$E0,$CC,$C8,$C9
	.dc.b	$0F,$F4,$FC,$F1,$6C,$71,$7F,$DA,$7F,$4B,$F4,$97,$22,$30,$DE,$0F
	.dc.b	$C5,$47,$CB,$E7,$FF,$4B,$3C,$7D,$27,$D3,$7F,$B2,$30,$FF,$3D,$C6
	.dc.b	$3D,$CE,$38,$7E,$8E,$31,$3F,$3F,$F0,$A7,$C5,$FE,$55,$B1,$CB,$0E
	.dc.b	$8F,$FA,$AB,$E9,$EE,$7F,$EB,$0F,$97,$FD,$37,$B8,$7B,$4E,$B8,$BF
	.dc.b	$1D,$39,$BF,$F4,$7F,$92,$F9,$8F,$F1,$74,$61,$AF,$FB,$22,$78,$3F
	.dc.b	$EC,$AF,$A6,$7D,$3F,$ED,$5B,$1F,$E9,$76,$AC,$FF,$D9,$D7,$D1,$FE
	.dc.b	$21,$1F,$EB,$4F,$A7,$7D,$07,$F0,$78,$43,$E9,$30,$D8,$FF,$53,$9A
	.dc.b	$9C,$9A,$99,$1A,$9E,$AD,$9C,$BE,$5D,$3F,$C7,$93,$D6,$30,$79,$67
	.dc.b	$F8,$47,$C7,$3E,$3A,$92,$7F,$D7,$3D,$0F,$8A,$7C,$FE,$67,$9F,$84
	.dc.b	$C3,$94,$7E,$BD,$AF,$82,$F9,$57,$F1,$9B,$FD,$56,$BF,$23,$07,$CB
	.dc.b	$FC,$AC,$3A,$B6,$7F,$DD,$55,$C3,$5F,$F5,$73,$69,$7C,$A7,$F9,$25
	.dc.b	$F1,$77,$E0,$7D,$3F,$88,$7C,$D5,$26,$1F,$4B,$E9,$DF,$3D,$88,$F9
	.dc.b	$22,$3F,$9C,$3F,$EF,$26,$AA,$EA,$FD,$C6,$BC,$A3,$67,$08,$CE,$9F
	.dc.b	$D7,$79,$3F,$A3,$7D,$3F,$F8,$46,$3F,$1B,$E5,$A3,$1A,$7F,$B3,$FD
	.dc.b	$3F,$D0,$65,$F6,$1A,$FE,$02,$B8,$7A,$FD,$0F,$E9,$53,$AC,$3F,$33
	.dc.b	$C6,$F9,$E7,$C7,$FF,$4D,$FA,$07,$C6,$53,$FA,$94,$EB,$3F,$C3,$B5
	.dc.b	$FC,$2C,$F0,$7F,$99,$35,$FC,$24,$6F,$0F,$C0,$7D,$3F,$E7,$27,$E7
	.dc.b	$BF,$99,$8C,$F6,$FE,$AE,$35,$1C,$3B,$31,$32,$BF,$57,$5A,$DF,$FD
	.dc.b	$3C,$D8,$07,$B4,$6A,$7B,$57,$F7,$7F,$1B,$CD,$3F,$D7,$57,$BF,$93
	.dc.b	$5A,$CC,$7A,$E5,$F6,$2B,$71,$8B,$E5,$7B,$8F,$AD,$F3,$BD,$97,$12
	.dc.b	$38,$C6,$A7,$FC,$15,$1B,$9D,$63,$89,$95,$DC,$E7,$5D,$57,$D3,$FE
	.dc.b	$79,$FE,$22,$FE,$4B,$0F,$F6,$23,$63,$E3,$BB,$7F,$46,$DC,$BF,$70
	.dc.b	$EE,$3A,$F4,$FF,$21,$8D,$57,$4E,$9C,$4C,$AF,$0C,$7F,$89,$7E,$9D
	.dc.b	$F1,$37,$E9,$E5,$7A,$F6,$56,$A3,$0D,$7F,$40,$A7,$8C,$62,$BF,$C6
	.dc.b	$5F,$47,$FA,$1C,$BF,$C3,$0F,$E1,$70,$34,$D5,$7D,$46,$C6,$B8,$79
	.dc.b	$9D,$3A,$FF,$4B,$C1,$FA,$9E,$29,$87,$E9,$A3,$2F,$E9,$23,$2F,$E9
	.dc.b	$23,$F6,$1F,$4F,$23,$F5,$2E,$B8,$7F,$94,$73,$A3,$53,$FB,$07,$5A
	.dc.b	$7F,$BF,$7E,$99,$F2,$1F,$EE,$9F,$3B,$0D,$BF,$D1,$CE,$9F,$F7,$D4
	.dc.b	$F1,$FA,$7D,$42,$74,$FF,$BE,$A7,$9B,$C0,$E7,$5C,$25,$5C,$BE,$93
	.dc.b	$FB,$B7,$91,$BB,$46,$1E,$87,$ED,$9C,$7C,$5F,$F7,$8B,$F0,$BF,$DE
	.dc.b	$A3,$38,$FE,$F2,$B9,$F0,$7F,$58,$79,$98,$C3,$C1,$87,$2A,$E1,$7C
	.dc.b	$AB,$C1,$7C,$70,$EC,$DA,$F7,$CF,$4A,$E0,$F5,$31,$3B,$7E,$19,$EF
	.dc.b	$8D,$92,$35,$1F,$37,$2F,$8D,$18,$9D,$63,$E8,$9C,$BC,$E7,$47,$2C
	.dc.b	$7D,$97,$EA,$9D,$11,$F8,$71,$A7,$D1,$8F,$BA,$74,$FC,$FB,$EF,$3E
	.dc.b	$C9,$CE,$3F,$0A,$B8,$4E,$48,$F1,$76,$78,$CC,$C7,$13,$E8,$E1,$EA
	.dc.b	$AD,$CF,$C3,$2E,$3F,$09,$F4,$E1,$E4,$C3,$C2,$4C,$3A,$4F,$78,$C3
	.dc.b	$63,$2F,$C0,$7C,$37,$E0,$C7,$E2,$61,$DE,$F8,$DF,$8C,$F7,$1F,$8F
	.dc.b	$3F,$00,$58,$FC,$D3,$71,$FC,$93,$9F,$13,$FB,$33,$F9,$06,$BF,$9F
	.dc.b	$6B,$FA,$46,$CF,$F8,$37,$B3,$F1,$51,$FD,$C6,$FF,$CB,$7B,$FF,$86
	.dc.b	$36,$CD,$4C,$3F,$2D,$7B,$D3,$F4,$D7,$B9,$CD,$6F,$0F,$D6,$5E,$E7
	.dc.b	$FB,$3D,$6E,$73,$5B,$BF,$53,$FB,$B9,$D4,$62,$BA,$92,$BC,$FD,$3F
	.dc.b	$E6,$87,$9C,$66,$78,$46,$B1,$ED,$3F,$17,$0A,$B0,$EE,$E4,$E4,$5A
	.dc.b	$E4,$E0,$F9,$93,$CA,$78,$C6,$0D,$65,$F3,$DF,$E1,$31,$A8,$C5,$7A
	.dc.b	$09,$87,$CC,$C7,$D8,$FA,$89,$7E,$37,$58,$FA,$AF,$9C,$6A,$7D,$95
	.dc.b	$C9,$E0,$43,$CD,$7E,$C3,$C2,$E6,$B9,$BF,$62,$79,$3C,$2B,$B4,$EA
	.dc.b	$3B,$DF,$8F,$2F,$C0,$46,$CE,$51,$9C,$BE,$46,$1D,$2F,$96,$5F,$28
	.dc.b	$9F,$4F,$E0,$71,$34,$C4,$5B,$FE,$91,$E4,$BE,$91,$3F,$EA,$CE,$4F
	.dc.b	$F4,$34,$62,$BF,$DF,$8F,$F9,$E9,$B3,$B7,$FC,$7B,$5C,$E3,$19,$FF
	.dc.b	$A1,$AB,$93,$FE,$9A,$79,$36,$37,$E2,$7F,$98,$E7,$8D,$79,$7B,$B3
	.dc.b	$E1,$F9,$F1,$F1,$C6,$2B,$67,$4C,$BF,$CD,$B5,$BC,$BC,$15,$E4,$EB
	.dc.b	$FF,$07,$FC,$CD,$ED,$7D,$3D,$8C,$37,$3F,$F1,$14,$E9,$B3,$A7,$FC
	.dc.b	$4D,$F3,$FF,$EA,$3C,$7F,$D0,$4F,$FA,$56,$33,$E6,$F8,$63,$13,$A8
	.dc.b	$C7,$2E,$5B,$5F,$E4,$BF,$9F,$7C,$8B,$E5,$FE,$6C,$63,$E5,$AC,$E6
	.dc.b	$B8,$28,$CD,$69,$B3,$A6,$E3,$FE,$D8,$BE,$93,$CA,$B8,$3F,$E4,$3F
	.dc.b	$25,$F0,$FF,$2D,$7D,$3D,$8C,$36,$78,$37,$0F,$F3,$8B,$F4,$FC,$66
	.dc.b	$D7,$E0,$30,$DB,$FF,$29,$4E,$9B,$3A,$69,$F0,$0F,$87,$F3,$BF,$11
	.dc.b	$A6,$4F,$8A,$E5,$A6,$70,$F8,$5B,$27,$51,$87,$F2,$7F,$1B,$F9,$6E
	.dc.b	$FF,$F4,$ED,$FF,$AD,$23,$15,$A8,$FF,$42,$75,$FF,$8C,$23,$07,$FB
	.dc.b	$F2,$B8,$BF,$EB,$6F,$A7,$E9,$D9,$FA,$06,$DF,$FB,$01,$B3,$B8,$EB
	.dc.b	$5F,$9F,$EC,$FF,$7E,$7B,$2F,$9D,$FD,$9F,$C0,$3E,$9D,$F4,$FF,$B6
	.dc.b	$C9,$7B,$CB,$FD,$35,$7B,$AD,$4E,$69,$8A,$FF,$D7,$B7,$B7,$FD,$45
	.dc.b	$3D,$31,$F0,$E3,$E0,$9C,$E1,$CF,$0F,$3D,$7F,$08,$F9,$57,$D3,$BE
	.dc.b	$03,$C3,$3F,$C1,$BE,$01,$F4,$DF,$C1,$BE,$9F,$FD,$8B,$9F,$FA,$C7
	.dc.b	$C8,$56,$A3,$E8,$CE,$B6,$F4,$C7,$FB,$52,$B7,$7C,$FD,$5E,$CD,$9F
	.dc.b	$B3,$F4,$FF,$91,$BE,$23,$99,$19,$DF,$FD,$FB,$9B,$E3,$3F,$8D,$CE
	.dc.b	$70,$FE,$4E,$DC,$39,$37,$3F,$4D,$70,$ED,$FD,$FD,$7E,$1B,$1F,$A0
	.dc.b	$BC,$2F,$F7,$6F,$9F,$A8,$DB,$72,$FE,$15,$08,$C3,$FB,$F3,$11,$A7
	.dc.b	$8B,$4E,$31,$FC,$A7,$E9,$FF,$8C,$DF,$8B,$CB,$46,$2B,$F0,$7F,$4E
	.dc.b	$F9,$2E,$1E,$57,$F3,$1F,$8B,$7C,$CF,$E5,$2B,$57,$FE,$CF,$9C,$BF
	.dc.b	$92,$8C,$E9,$F4,$3E,$41,$F4,$FF,$C7,$6B,$86,$1F,$D4,$27,$3E,$6F
	.dc.b	$78,$C7,$9A,$F8,$6F,$E9,$BD,$4F,$B5,$CC,$61,$FE,$A7,$F3,$9F,$9D
	.dc.b	$CB,$FD,$F9,$97,$F3,$1A,$CC,$70,$3F,$21,$3B,$D3,$F0,$BD,$08,$C7
	.dc.b	$27,$D8,$8C,$79,$5C,$AF,$FD,$DD,$DE,$FA,$8F,$EA,$CB,$E3,$AC,$61
	.dc.b	$B3,$C1,$B9,$7E,$53,$A3,$F8,$04,$63,$2F,$AD,$87,$3B,$EF,$87,$37
	.dc.b	$F5,$1F,$35,$FE,$D8,$7F,$8B,$FD,$3B,$EA,$3F,$BA,$CE,$73,$FD,$0F
	.dc.b	$84,$AD,$5F,$2F,$84,$FE,$F3,$18,$AF,$87,$7C,$73,$FC,$CE,$1C,$71
	.dc.b	$EB,$87,$06,$E3,$E2,$7F,$BC,$7C,$47,$F5,$DC,$7D,$CF,$E0,$5B,$CD
	.dc.b	$AE,$8D,$B6,$D3,$A6,$E5,$F9,$9B,$E4,$DA,$E1,$5B,$3B,$DF,$F8,$5B
	.dc.b	$E7,$C7,$F8,$5B,$86,$E3,$FC,$46,$2D,$7C,$69,$B8,$7E,$06,$6D,$7F
	.dc.b	$8A,$26,$D3,$8F,$AD,$75,$3C,$05,$AE,$CD,$9F,$C0,$60,$FA,$8F,$F0
	.dc.b	$F7,$D4,$7E,$42,$FE,$0E,$3F,$E0,$39,$CC,$7F,$88,$EB,$83,$F1,$0D
	.dc.b	$D7,$1B,$5F,$19,$D4,$62,$75,$39,$8F,$C9,$46,$2B,$AC,$62,$BF,$C7
	.dc.b	$FF,$4E,$FA,$8F,$F1,$FC,$7F,$3F,$3F,$9D,$46,$5F,$ED,$26,$67,$F5
	.dc.b	$B1,$97,$F2,$71,$9D,$3F,$5D,$39,$7F,$1F,$19,$3F,$D9,$52,$F8,$7F
	.dc.b	$C2,$FD,$03,$EA,$3F,$0F,$2F,$53,$FA,$44,$E6,$3F,$DA,$55,$C2,$B7
	.dc.b	$5A,$7A,$C6,$67,$FB,$2F,$81,$FC,$E2,$F9,$47,$69,$D3,$DE,$75,$5C
	.dc.b	$A1,$5C,$2F,$B4,$78,$8C,$CF,$8A,$B8,$C6,$6B,$84,$62,$3C,$07,$03
	.dc.b	$CB,$87,$89,$FD,$3E,$1E,$4C,$3B,$C7,$0B,$E7,$5B,$8F,$94,$73,$3D
	.dc.b	$4E,$6B,$95,$F8,$67,$A9,$BB,$EA,$E2,$FA,$BC,$CC,$61,$ED,$C3,$A4
	.dc.b	$71,$BE,$8F,$8A,$B8,$9F,$39,$E2,$7B,$8E,$87,$A0,$FC,$19,$4F,$A0
	.dc.b	$63,$0C,$FD,$47,$F7,$BF,$6E,$F9,$BF,$E7,$4A,$FC,$5C,$E5,$B1,$97
	.dc.b	$F1,$62,$1F,$4F,$FE,$2A,$8E,$2D,$77,$1A,$D7,$C3,$39,$9F,$CA,$CE
	.dc.b	$5F,$A1,$F5,$0F,$99,$FC,$CE,$7F,$8E,$8C,$4F,$E3,$AB,$84,$FE,$49
	.dc.b	$BB,$7E,$82,$B7,$5D,$2F,$9C,$F8,$30,$F4,$61,$C5,$B1,$C6,$77,$5F
	.dc.b	$CC,$2F,$9D,$EF,$1F,$16,$1C,$2B,$57,$C1,$5F,$0C,$3C,$17,$C0,$FD
	.dc.b	$3C,$66,$78,$99,$3F,$25,$8F,$B1,$C9,$EB,$35,$8F,$8D,$C1,$1C,$12
	.dc.b	$FD,$91,$E0,$C3,$CD,$5D,$63,$EE,$63,$EB,$7D,$33,$EB,$C3,$CD,$1A
	.dc.b	$C7,$F8,$27,$63,$D7,$FF,$55,$CF,$FA,$19,$AF,$F6,$E6,$CF,$F8,$27
	.dc.b	$E6,$5F,$47,$FC,$5A,$3F,$B5,$CE,$B8,$BF,$DE,$6D,$BF,$F2,$B3,$7A
	.dc.b	$D8,$72,$9F,$F4,$AF,$17,$F5,$C6,$F5,$F5,$C3,$FD,$FB,$39,$9F,$E6
	.dc.b	$3D,$F9,$7F,$CC,$51,$9C,$BF,$BE,$B6,$B8,$DE,$F1,$73,$B1,$E5,$AF
	.dc.b	$F3,$3A,$B8,$71,$CB,$AC,$71,$BE,$11,$FE,$AC,$EB,$7B,$DC,$9B,$19
	.dc.b	$8F,$05,$3E,$C7,$51,$27,$47,$F7,$98,$C3,$FD,$E6,$2E,$BF,$DE,$67
	.dc.b	$CF,$D6,$F0,$29,$2F,$FD,$25,$F8,$9B,$E7,$B1,$1F,$E2,$8A,$E0,$FF
	.dc.b	$B1,$1A,$FF,$92,$27,$83,$9A,$E3,$87,$F8,$A2,$B6,$FF,$92,$EB,$75
	.dc.b	$FE,$34,$EA,$FF,$60,$3F,$ED,$06,$B9,$F1,$FE,$97,$8A,$F9,$2F,$A3
	.dc.b	$5A,$C3,$FC,$6A,$DE,$5F,$F1,$4F,$1F,$F5,$2C,$38,$DF,$06,$63,$DD
	.dc.b	$CB,$FE,$E0,$FC,$5B,$F8,$FB,$7B,$3E,$FC,$62,$3F,$C8,$BE,$E6,$26
	.dc.b	$46,$CF,$F2,$AC,$61,$FF,$28,$CF,$1C,$FF,$C9,$6D,$CB,$FD,$79,$A7
	.dc.b	$FB,$1A,$BF,$D8,$9F,$4C,$F8,$9F,$C7,$FD,$33,$E5,$A9,$F9,$59,$D5
	.dc.b	$7F,$BE,$27,$4D,$9D,$BD,$2A,$BF,$E6,$49,$7C,$D7,$63,$39,$FF,$98
	.dc.b	$3C,$77,$CA,$FE,$81,$7C,$D5,$31,$1F,$A4,$FC,$37,$EF,$ED,$F4,$9B
	.dc.b	$97,$FB,$45,$BC,$8C,$39,$CF,$F8,$DA,$1F,$89,$3E,$87,$F5,$ED,$9D
	.dc.b	$35,$FF,$3E,$DF,$67,$F3,$5F,$3B,$FB,$F6,$9F,$EF,$EE,$C4,$65,$FF
	.dc.b	$1E,$46,$63,$04,$78,$75,$D7,$FD,$D6,$74,$2C,$F4,$76,$7A,$FE,$2F
	.dc.b	$F3,$FB,$FF,$A2,$9B,$AF,$E5,$9B,$3F,$E6,$BF,$88,$F1,$BF,$E9,$88
	.dc.b	$CE,$A6,$9C,$3C,$65,$70,$C3,$FD,$2F,$F2,$AF,$83,$FE,$2E,$DC,$7E
	.dc.b	$73,$5F,$F6,$0D,$7E,$A5,$62,$9D,$5F,$4B,$E6,$7F,$25,$BE,$17,$DF
	.dc.b	$D9,$FA,$13,$FE,$C5,$8F,$F3,$89,$92,$FE,$09,$FE,$B7,$73,$1B,$7E
	.dc.b	$6B,$6D,$A7,$F9,$7A,$30,$FF,$98,$2B,$95,$70,$BF,$C0,$43,$F0,$3A
	.dc.b	$79,$23,$F9,$E7,$CE,$3E,$27,$F9,$DF,$CF,$7F,$BB,$F0,$FC,$93,$5F
	.dc.b	$F7,$2C,$ED,$FF,$45,$EE,$8C,$47,$FB,$3A,$78,$69,$FE,$C8,$6E,$7F
	.dc.b	$EB,$CB,$E1,$87,$F4,$DF,$3D,$F4,$1F,$D5,$7C,$D7,$D2,$BE,$D3,$0D
	.dc.b	$D3,$C3,$7C,$1B,$7A,$A9,$18,$3E,$8F,$D3,$3E,$6F,$F4,$8D,$8C,$C6
	.dc.b	$3C,$1C,$B2,$32,$72,$8C,$3F,$A2,$9E,$4D,$9E,$4F,$EE,$FE,$63,$FD
	.dc.b	$13,$C1,$FE,$EA,$6E,$FF,$CD,$9B,$3F,$EC,$BE,$1B,$D0,$9A,$4F,$6F
	.dc.b	$EC,$9F,$1F,$F6,$1E,$3C,$27,$F0,$79,$7E,$23,$1F,$F6,$1C,$63,$2F
	.dc.b	$F6,$6B,$C0,$9F,$09,$CC,$E9,$87,$F0,$9F,$A7,$FE,$69,$7F,$C0,$1B
	.dc.b	$87,$58,$46,$30,$FF,$59,$F4,$7F,$9A,$AB,$53,$9A,$FF,$7E,$B5,$E3
	.dc.b	$E6,$FF,$17,$AE,$17,$FE,$BC,$AD,$12,$75,$5F,$9A,$C7,$FC,$27,$A7
	.dc.b	$E7,$E3,$E8,$61,$97,$F2,$FB,$7C,$17,$F7,$B3,$B5,$97,$D5,$AE,$11
	.dc.b	$FA,$0F,$C2,$BA,$FA,$B2,$39,$C7,$EA,$8C,$3C,$B4,$FB,$93,$98,$FC
	.dc.b	$17,$D3,$F9,$88,$F9,$2E,$BC,$4F,$2C,$63,$0F,$B5,$D5,$FA,$FA,$F7
	.dc.b	$FE,$29,$2F,$F1,$CE,$F1,$FC,$63,$98,$E9,$C7,$F8,$B8,$D4,$07,$D4
	.dc.b	$7F,$8D,$9F,$F7,$77,$AB,$FE,$EC,$8C,$7B,$9F,$E1,$19,$E1,$18,$9D
	.dc.b	$79,$DF,$EE,$57,$87,$BB,$6B,$6D,$7B,$B6,$BF,$B0,$C7,$80,$F2,$70
	.dc.b	$FF,$7F,$F2,$BF,$DB,$B7,$FE,$FE,$8C,$4E,$CF,$D3,$7A,$5F,$D4,$72
	.dc.b	$F3,$47,$63,$89,$D2,$35,$5C,$DB,$19,$BD,$BC,$4E,$D7,$E0,$BE,$AD
	.dc.b	$BE,$55,$C3,$0F,$89,$7C,$CE,$38,$76,$7C,$F3,$BC,$7E,$26,$1C,$E7
	.dc.b	$86,$3D,$AF,$86,$1D,$27,$89,$D3,$0F,$3D,$F1,$8F,$45,$F1,$9C,$DF
	.dc.b	$86,$B5,$7C,$DE,$04,$8E,$F8,$79,$1E,$04,$AF,$46,$1D,$18,$F8,$CF
	.dc.b	$A9,$7D,$67,$84,$67,$0F,$AB,$39,$AE,$31,$E7,$9E,$37,$E1,$9C,$E3
	.dc.b	$F3,$A7,$26,$0F,$8E,$E6,$BA,$D7,$19,$F0,$4E,$4F,$CA,$BC,$4F,$97
	.dc.b	$3E,$13,$8B,$9C,$7E,$25,$FC,$08,$E6,$78,$08,$7D,$0A,$F3,$3F,$AD
	.dc.b	$3F,$04,$53,$EC,$96,$7F,$36,$DF,$1B,$FB,$5B,$F9,$F6,$BF,$A4,$6B
	.dc.b	$FD,$15,$B1,$FE,$1F,$F0,$FF,$07,$1F,$DD,$F0,$FD,$1F,$88,$FA,$8F
	.dc.b	$E8,$91,$F9,$A9,$D6,$3D,$EF,$89,$FA,$2B,$E1,$C3,$8C,$BF,$47,$7C
	.dc.b	$1F,$25,$6F,$1E,$B8,$73,$9D,$CF,$1C,$7C,$75,$A9,$E2,$68,$91,$AC
	.dc.b	$38,$5F,$38,$F4,$5F,$79,$CD,$72,$9C,$BF,$C1,$71,$F4,$BD,$2F,$91
	.dc.b	$1F,$E4,$51,$A9,$E2,$49,$D4,$A7,$51,$8C,$BE,$84,$F5,$39,$E3,$ED
	.dc.b	$7E,$B6,$3E,$66,$9E,$58,$D4,$F2,$C3,$C1,$3E,$BB,$FC,$56,$1E,$08
	.dc.b	$C1,$0E,$73,$F1,$DF,$11,$89,$F4,$CF,$18,$C1,$C6,$BC,$13,$DA,$FC
	.dc.b	$79,$7D,$DA,$D9,$2F,$95,$F0,$8E,$53,$B9,$CB,$C6,$75,$3C,$6B,$AE
	.dc.b	$3F,$66,$3E,$16,$5F,$8F,$F4,$7E,$46,$3E,$72,$5E,$23,$2F,$E2,$78
	.dc.b	$BF,$AC,$FD,$5F,$F8,$7F,$9B,$FC,$41,$18,$C3,$FD,$59,$C3,$FE,$BE
	.dc.b	$6D,$6D,$FF,$73,$C2,$31,$97,$FA,$0A,$31,$3C,$BC,$3C,$DF,$5A,$FE
	.dc.b	$DB,$5D,$5B,$3C,$B9,$FF,$CC,$DF,$65,$C2,$B5,$97,$99,$B8,$F8,$AF
	.dc.b	$87,$62,$F8,$46,$3A,$F9,$7C,$C5,$BF,$6B,$73,$F1,$B7,$2E,$CF,$4F
	.dc.b	$7B,$C5,$F7,$76,$FE,$2B,$6F,$D6,$DB,$F3,$3F,$13,$E6,$DF,$15,$F1
	.dc.b	$2B,$FC,$31,$18,$E1,$D4,$FF,$9B,$F0,$FF,$07,$61,$FE,$53,$AD,$46
	.dc.b	$0F,$F0,$93,$8B,$E7,$5D,$6B,$83,$CE,$33,$7D,$89,$7C,$E1,$7E,$6F
	.dc.b	$23,$31,$FE,$2D,$F2,$3F,$CC,$DF,$5D,$F7,$9F,$F3,$47,$C7,$DF,$29
	.dc.b	$36,$30,$DC,$7B,$57,$26,$D2,$BF,$0B,$F9,$0B,$D7,$A7,$DC,$F1,$ED
	.dc.b	$FF,$BF,$E7,$4D,$9D,$34,$F6,$69,$F4,$3D,$5F,$9A,$7C,$EF,$02,$CE
	.dc.b	$B8,$3F,$DB,$0D,$9D,$34,$9B,$7C,$E8,$CB,$63,$FC,$CF,$E3,$FA,$BC
	.dc.b	$2F,$C9,$B6,$32,$7A,$63,$2D,$3F,$CD,$FF,$4D,$FA,$6F,$0F,$FC,$9D
	.dc.b	$18,$3F,$D4,$97,$F9,$E9,$D3,$63,$27,$FB,$A3,$D7,$7C,$B7,$ED,$3C
	.dc.b	$57,$E0,$0E,$AF,$DB,$E0,$BD,$E1,$F8,$3B,$DD,$6A,$73,$5F,$E6,$F3
	.dc.b	$86,$1D,$70,$7C,$66,$E2,$63,$E3,$75,$18,$AF,$F3,$6F,$D1,$7F,$BA
	.dc.b	$EB,$F2,$5E,$2A,$31,$3F,$E7,$DF,$15,$ED,$F8,$BB,$FF,$8B,$D6,$9B
	.dc.b	$B7,$F0,$E8,$C6,$3F,$EE,$9A,$E0,$FF,$9E,$71,$ED,$5F,$C5,$3E,$89
	.dc.b	$E0,$D3,$F8,$95,$BE,$1F,$18,$79,$AB,$6D,$AD,$CE,$B4,$FD,$E8,$F5
	.dc.b	$BF,$81,$36,$BF,$2A,$BD,$4C,$CC,$9C,$BC,$23,$13,$D2,$2E,$BF,$C1
	.dc.b	$63,$95,$70,$3E,$5F,$E2,$4F,$47,$F9,$8B,$63,$E7,$FC,$07,$EC,$5D
	.dc.b	$CF,$F3,$68,$C5,$7F,$A8,$9B,$7F,$D1,$97,$D4,$6F,$C3,$E1,$D3,$C6
	.dc.b	$2E,$3D,$9F,$E8,$3F,$0A,$FA,$5F,$E8,$7F,$0C,$FA,$8C,$C6,$1B,$7F
	.dc.b	$84,$AE,$0F,$F1,$7E,$87,$C4,$F9,$23,$EE,$4E,$B6,$FC,$44,$E5,$74
	.dc.b	$27,$5B,$FF,$B1,$AB,$53,$9C,$BF,$AB,$FE,$2F,$FB,$2C,$3F,$D8,$6D
	.dc.b	$BE,$1E,$EE,$A6,$65,$F9,$AA,$D3,$FE,$C7,$CF,$CF,$5A,$AE,$65,$BE
	.dc.b	$75,$C6,$3F,$B3,$FC,$E7,$E1,$DF,$F6,$9B,$67,$FD,$BB,$F1,$8F,$8E
	.dc.b	$DF,$3C,$C3,$F5,$77,$B7,$C5,$5B,$73,$3B,$C3,$AB,$FD,$C6,$75,$18
	.dc.b	$9F,$DA,$FE,$1E,$F9,$FF,$EF,$6D,$8F,$A9,$F2,$4F,$5B,$35,$27,$8B
	.dc.b	$FD,$F6,$75,$18,$25,$FF,$BC,$2F,$A3,$FC,$06,$76,$E6,$BF,$C1,$7E
	.dc.b	$9B,$E4,$29,$FE,$09,$F9,$4C,$63,$FC,$32,$6D,$FF,$BF,$1B,$BF,$F8
	.dc.b	$81,$B3,$F5,$EA,$D7,$E2,$26,$F2,$1A,$6B,$89,$58,$C7,$FC,$3F,$F8
	.dc.b	$BB,$E4,$29,$FE,$1D,$F9,$67,$C4,$E3,$4F,$E4,$35,$B6,$D6,$E7,$5A
	.dc.b	$35,$FF,$1B,$DE,$E3,$17,$FC,$76,$31,$28,$FC,$96,$1E,$07,$A9,$FC
	.dc.b	$C0,$7C,$9F,$F8,$D9,$A7,$EB,$E1,$F0,$F9,$99,$3C,$5F,$64,$63,$0F
	.dc.b	$E8,$C0,$7C,$3D,$F9,$5F,$E4,$91,$9F,$9C,$78,$3E,$78,$4E,$9F,$F6
	.dc.b	$25,$71,$E9,$FF,$63,$5F,$0C,$B1,$5F,$EC,$7A,$AA,$74,$F4,$AF,$F6
	.dc.b	$63,$E4,$30,$78,$F0,$F1,$61,$E1,$8F,$55,$F4,$BE,$73,$99,$E4,$F9
	.dc.b	$E3,$37,$DA,$B8,$5F,$48,$C5,$F8,$0F,$34,$65,$F5,$19,$AE,$6E,$2F
	.dc.b	$B4,$EE,$3E,$29,$B7,$0F,$A0,$C6,$1F,$11,$C4,$F2,$8C,$47,$AA,$FE
	.dc.b	$6D,$72,$8C,$1F,$16,$77,$87,$B1,$B1,$D8,$F6,$1D,$AB,$9C,$6A,$3E
	.dc.b	$89,$F4,$4B,$38,$7D,$07,$FC,$01,$F9,$B7,$CE,$7F,$9A,$A3,$EC,$C6
	.dc.b	$3D,$AE,$6D,$8C,$CE,$DE,$5E,$67,$EA,$5E,$2D,$73,$C1,$CA,$31,$1F
	.dc.b	$99,$15,$FE,$03,$9F,$C6,$FC,$DB,$E7,$35,$B7,$E5,$E3,$17,$F9,$2B
	.dc.b	$E0,$FD,$F6,$CE,$EB,$8E,$BD,$EF,$8D,$EF,$1F,$45,$FE,$8F,$0E,$37
	.dc.b	$BC,$38,$DF,$EB,$F0,$E3,$7B,$C3,$8D,$FA,$F0,$E3,$7B,$C3,$8A,$C3
	.dc.b	$8D,$6E,$F8,$CF,$EE,$EF,$8B,$BB,$E3,$7C,$23,$55,$C1,$F9,$2D,$8D
	.dc.b	$47,$13,$E3,$46,$EB,$AC,$E8,$DB,$E0,$33,$5E,$03,$D2,$E6,$7B,$46
	.dc.b	$6B,$AE,$3F,$C1,$BA,$9F,$38,$FA,$DF,$BD,$3F,$EE,$28,$C5,$7F,$95
	.dc.b	$1A,$FF,$94,$FE,$5F,$FC,$BE,$E6,$B8,$F0,$FF,$93,$6B,$77,$CE,$73
	.dc.b	$4E,$C6,$1C,$A3,$F9,$C9,$FE,$16,$E2,$FE,$96,$FF,$99,$1A,$EB,$2F
	.dc.b	$E5,$7D,$9D,$2D,$D1,$FE,$3F,$AE,$13,$E0,$AD,$C6,$0F,$E7,$9E,$7E
	.dc.b	$70,$EB,$9F,$F3,$1A,$DD,$77,$9C,$EB,$FD,$2A,$FA,$77,$3E,$76,$92
	.dc.b	$FF,$C6,$D9,$3E,$2B,$11,$FD,$C6,$6E,$7F,$CF,$F5,$FE,$D3,$58,$CB
	.dc.b	$FC,$ED,$8F,$5C,$3F,$BD,$61,$EF,$C7,$A4,$7F,$A2,$3E,$13,$FC,$F2
	.dc.b	$DF,$57,$E0,$46,$23,$FC,$15,$E0,$E2,$64,$6C,$FF,$0C,$C6,$1F,$9D
	.dc.b	$3C,$74,$F8,$8D,$CB,$FC,$79,$E2,$7F,$7D,$AF,$F4,$CF,$57,$F8,$CD
	.dc.b	$BE,$4F,$D1,$8C,$47,$FA,$6F,$C8,$7C,$06,$6A,$4F,$03,$FC,$BD,$3A
	.dc.b	$70,$7E,$06,$B9,$E5,$F5,$DB,$8F,$F8,$FB,$0F,$C0,$7C,$EF,$FA,$66
	.dc.b	$FF,$D5,$AD,$7F,$14,$DB,$FF,$25,$F8,$DC,$3B,$25,$70,$7F,$1D,$97
	.dc.b	$F8,$62,$30,$43,$9C,$6A,$7F,$29,$E7,$FF,$8D,$67,$FC,$F5,$F2,$7F
	.dc.b	$93,$F4,$4E,$BC,$AC,$32,$76,$7F,$9F,$EB,$93,$FA,$1A,$E2,$FF,$A1
	.dc.b	$7E,$37,$A1,$C5,$FD,$37,$8E,$F9,$4F,$D1,$36,$3F,$D9,$ED,$77,$F2
	.dc.b	$7F,$A7,$F6,$7E,$F4,$7F,$A2,$1B,$E0,$1B,$5C,$78,$3F,$51,$F3,$BF
	.dc.b	$E2,$AB,$FF,$6D,$B5,$FD,$C3,$63,$FD,$55,$D5,$FC,$12,$7F,$C8,$DE
	.dc.b	$6D,$EB,$FF,$17,$C7,$F8,$A6,$BF,$C4,$8E,$8C,$73,$FF,$55,$97,$5F
	.dc.b	$1B,$95,$A3,$FC,$65,$3D,$1F,$F7,$D3,$C1,$D1,$D4,$CC,$72,$77,$3F
	.dc.b	$C2,$FE,$13,$F4,$11,$FC,$4D,$B8,$7F,$09,$6E,$1F,$EF,$1A,$E1,$84
	.dc.b	$BF,$F7,$FD,$69,$FE,$49,$EF,$F8,$DF,$E3,$DF,$10,$F8,$9F,$E3,$4D
	.dc.b	$8F,$67,$02,$31,$1F,$CA,$BC,$45,$FE,$92,$3F,$D8,$33,$AD,$3F,$8F
	.dc.b	$37,$4F,$F2,$B6,$1C,$F1,$FE,$59,$87,$1C,$3C,$5F,$39,$F5,$E4,$D3
	.dc.b	$ED,$7C,$EF,$FB,$EE,$7E,$5C,$E9,$F9,$B0,$F9,$8D,$9C,$63,$75,$87
	.dc.b	$71,$97,$D9,$F2,$AF,$86,$FE,$79,$F3,$0F,$A2,$A7,$F4,$39,$D5,$7F
	.dc.b	$B9,$DA,$FF,$49,$9E,$0F,$CA,$6B,$DA,$BF,$A6,$79,$EF,$9D,$FE,$B3
	.dc.b	$E6,$3E,$79,$F3,$D8,$6E,$1F,$EF,$5B,$E0,$DB,$D5,$48,$C1,$F9,$0F
	.dc.b	$9E,$7C,$CF,$E8,$DB,$19,$8C,$78,$59,$64,$64,$E5,$18,$7F,$B5,$89
	.dc.b	$E6,$E1,$BC,$FF,$AC,$53,$FD,$9F,$E4,$FF,$D2,$13,$FD,$8D,$BC,$FF
	.dc.b	$86,$6D,$FF,$7B,$E0,$BC,$E9,$A4,$F6,$FE,$A1,$F2,$BE,$63,$8F,$09
	.dc.b	$FC,$46,$1F,$E0,$28,$C7,$C7,$6E,$A4,$FE,$EE,$72,$DA,$E3,$9F,$FA
	.dc.b	$A6,$7A,$57,$F0,$DF,$9E,$FE,$71,$7F,$11,$B5,$FC,$31,$BA,$7F,$97
	.dc.b	$F9,$FF,$CC,$15,$BF,$33,$91,$1F,$E0,$F1,$AB,$E5,$39,$9E,$06,$8F
	.dc.b	$CA,$67,$F3,$6A,$47,$E5,$63,$2F,$EF,$34,$FD,$33,$67,$2F,$E8,$BE
	.dc.b	$23,$EB,$7A,$F2,$3F,$3D,$1F,$A1,$30,$FF,$89,$EB,$FC,$0E,$F1,$3B
	.dc.b	$7C,$CF,$DF,$23,$FA,$7B,$AF,$18,$B9,$7F,$85,$F8,$7C,$B3,$D7,$E5
	.dc.b	$FF,$10,$47,$A9,$EB,$70,$EC,$C9,$F9,$6F,$9D,$FF,$1C,$3F,$EE,$6F
	.dc.b	$57,$FD,$21,$18,$F7,$3F,$C0,$52,$8C,$4E,$BC,$FF,$F7,$2B,$C3,$DC
	.dc.b	$B5,$FE,$F1,$AD,$B6,$32,$F3,$AB,$AF,$F7,$4F,$2F,$FD,$BF,$87,$FA
	.dc.b	$86,$31,$5E,$53,$FD,$F8,$E3,$1F,$2E,$3D,$A7,$F4,$5E,$9F,$F5,$A9
	.dc.b	$F4,$D6,$EF,$A4,$E5,$F4,$57,$80,$EB,$19,$6E,$19,$7A,$B7,$0F,$05
	.dc.b	$F5,$9D,$9E,$73,$65,$8D,$D7,$58,$CE,$1E,$68,$C4,$EA,$FC,$17,$D6
	.dc.b	$F7,$87,$A1,$ED,$19,$25,$F2,$CB,$E6,$BD,$E3,$37,$E9,$9C,$BD,$A7
	.dc.b	$47,$1A,$C4,$47,$BC,$65,$B1,$9C,$7D,$4F,$58,$C6,$1F,$42,$2C,$7E
	.dc.b	$0C,$A7,$B0,$F9,$33,$F7,$2F,$CB,$1D,$89,$19,$78,$E1,$EC,$7E,$74
	.dc.b	$F0,$9F,$1C,$70,$BF,$91,$18,$23,$E5,$8E,$27,$4C,$3E,$1E,$1E,$B8
	.dc.b	$F1,$1F,$C5,$B0,$FC,$AB,$67,$FC,$3C,$F3,$35,$E3,$7F,$62,$7F,$38
	.dc.b	$D7,$F3,$6D,$7F,$9D,$36,$7F,$C2,$5E,$17,$E0,$A3,$FB,$6E,$3F,$B4
	.dc.b	$F0,$DF,$3C,$E8,$D5,$CC,$7F,$0B,$87,$0D,$11,$30,$E3,$7B,$C7,$F4
	.dc.b	$F7,$BA,$F2,$E1,$FA,$1B,$DC,$74,$8E,$18,$7B,$1D,$63,$E3,$9D,$13
	.dc.b	$0E,$B3,$AC,$78,$CF,$63,$57,$E6,$58,$78,$2F,$8C,$68,$E0,$FC,$3C
	.dc.b	$3C,$0E,$89,$5E,$59,$FD,$A5,$73,$8C,$4E,$A7,$9C,$78,$67,$51,$82
	.dc.b	$1F,$5B,$0F,$53,$E3,$30,$58,$E2,$E2,$BB,$36,$BB,$5F,$A1,$F1,$92
	.dc.b	$3B,$D7,$2C,$BF,$07,$1D,$DB,$7F,$07,$1F,$97,$3D,$4E,$38,$75,$7D
	.dc.b	$35,$C3,$0E,$D1,$F3,$A7,$4F,$12,$4E,$DF,$C4,$46,$A3,$11,$AC,$BF
	.dc.b	$10,$F3,$BF,$1E,$3F,$08,$9E,$97,$67,$18,$53,$11,$67,$F9,$A7,$77
	.dc.b	$F5,$1F,$9D,$FC,$67,$2F,$E3,$63,$13,$FE,$74,$E2,$FF,$15,$B6,$B6
	.dc.b	$FF,$9B,$A1,$18,$C9,$5C,$95,$F5,$AE,$71,$8B,$E5,$7C,$6F,$FB,$27
	.dc.b	$81,$6F,$78,$7F,$77,$F9,$EC,$4C,$C3,$CC,$F8,$7C,$D7,$42,$B5,$9F
	.dc.b	$F9,$C3,$B3,$A7,$C6,$EE,$91,$8C,$7F,$CD,$CD,$AE,$11,$F0,$EB,$87
	.dc.b	$A5,$E5,$F9,$BB,$3F,$1D,$B5,$ED,$3E,$1E,$A8,$CE,$BF,$0D,$F9,$1F
	.dc.b	$38,$F8,$AF,$97,$87,$FA,$FE,$31,$C3,$A9,$FF,$36,$CE,$5D,$E3,$DE
	.dc.b	$B7,$39,$AE,$1E,$4F,$16,$46,$4E,$51,$8A,$FF,$2C,$79,$0F,$9E,$FB
	.dc.b	$9E,$4F,$F9,$77,$E5,$EC,$E7,$0F,$F2,$C4,$62,$B2,$D8,$C9,$FE,$BE
	.dc.b	$F9,$EF,$F6,$EB,$7D,$8F,$A5,$E3,$7E,$06,$73,$97,$C9,$E9,$46,$38
	.dc.b	$7F,$2C,$DC,$3F,$98,$F4,$7E,$03,$0E,$58,$FC,$7F,$5A,$57,$C7,$EE
	.dc.b	$B9,$FC,$7E,$3F,$F6,$84,$FC,$BA,$F9,$D5,$AC,$FF,$98,$71,$FD,$59
	.dc.b	$C5,$7F,$40,$8F,$B1,$58,$C3,$FD,$15,$E3,$FA,$FC,$1F,$C7,$36,$FF
	.dc.b	$D7,$4D,$9F,$F6,$1B,$71,$FF,$2A,$7C,$C7,$E8,$FC,$2F,$E9,$B7,$FE
	.dc.b	$C1,$8C,$5F,$E6,$5B,$19,$3F,$CE,$11,$FB,$31,$EC,$3E,$77,$F6,$DE
	.dc.b	$33,$E7,$BF,$29,$7F,$D6,$B2,$FE,$9D,$87,$0B,$DD,$6A,$3F,$D6,$78
	.dc.b	$73,$BE,$07,$F6,$2C,$3A,$76,$78,$67,$55,$FA,$39,$CD,$70,$8F,$ED
	.dc.b	$33,$AC,$7C,$2D,$9E,$98,$3E,$7B,$A6,$5F,$C1,$FE,$7B,$FD,$D5,$5C
	.dc.b	$FA,$11,$8C,$BF,$65,$CE,$F7,$3A,$47,$F1,$69,$D6,$9F,$C8,$5B,$87
	.dc.b	$F7,$CC,$39,$EB,$FC,$37,$0E,$31,$89,$F0,$7C,$F3,$C2,$FE,$34,$D8
	.dc.b	$F7,$67,$E1,$F8,$5C,$D4,$9E,$2F,$58,$C6,$9F,$BB,$1E,$B7,$F0,$66
	.dc.b	$D7,$F0,$55,$EA,$66,$64,$E5,$E1,$18,$9E,$91,$76,$FC,$EC,$72,$AE
	.dc.b	$07,$CB,$F9,$E7,$A5,$FC,$D5,$B1,$F3,$FE,$0D,$E1,$66,$BF,$9D,$46
	.dc.b	$27,$53,$8F,$9E,$CB,$63,$94,$70,$E9,$C6,$3F,$C0,$7E,$15,$F3,$BF
	.dc.b	$D1,$7E,$19,$F3,$DF,$D1,$E7,$38,$FE,$06,$72,$FE,$0E,$33,$7C,$EF
	.dc.b	$FA,$37,$C4,$79,$63,$15,$F5,$A3,$1B,$FF,$63,$BE,$2F,$F8,$C7,$91
	.dc.b	$3A,$AF,$C3,$56,$9F,$C4,$7C,$F7,$EE,$30,$FE,$34,$DC,$39,$5B,$DD
	.dc.b	$FE,$C1,$39,$BE,$37,$83,$CF,$36,$3F,$3D,$3B,$9F,$ED,$ED,$C7,$CB
	.dc.b	$5B,$9F,$E5,$5F,$3D,$F8,$86,$F2,$7F,$67,$F8,$C7,$C7,$FF,$71,$AD
	.dc.b	$69,$FE,$03,$6E,$7F,$A3,$6E,$0C,$98,$72,$AF,$D9,$56,$BA,$CD,$38
	.dc.b	$D7,$0C,$7C,$19,$7F,$80,$7E,$79,$F3,$D4,$FE,$F9,$F2,$4F,$5F,$19
	.dc.b	$FE,$FE,$B6,$DA,$DC,$EB,$5F,$E9,$55,$A6,$CE,$AB,$81,$FC,$27,$D4
	.dc.b	$7C,$87,$F8,$4D,$B1,$FB,$AF,$93,$B7,$FD,$56,$6F,$0F,$F8,$A5,$B5
	.dc.b	$FC,$25,$B3,$F1,$EA,$F0,$26,$93,$0F,$F0,$D4,$DB,$FE,$39,$F3,$CF
	.dc.b	$91,$FF,$13,$36,$3F,$86,$FC,$B3,$E2,$73,$52,$78,$BF,$E2,$59,$D4
	.dc.b	$60,$99,$FF,$6F,$BE,$8F,$F2,$89,$DB,$9A,$FF,$1B,$53,$E7,$A9,$D7
	.dc.b	$E2,$2C,$7F,$33,$34,$58,$D3,$4F,$E8,$55,$FE,$95,$F8,$7F,$F5,$5F
	.dc.b	$CB,$7F,$20,$8F,$F5,$6F,$CF,$3C,$0F,$F5,$14,$E6,$78,$07,$13,$BB
	.dc.b	$E7,$CF,$FE,$A1,$AF,$F5,$BE,$18,$7A,$E1,$D2,$75,$09,$D2,$BE,$8F
	.dc.b	$79,$DC,$EA,$73,$87,$96,$BB,$9B,$3B,$E1,$D1,$CC,$7A,$30,$F1,$47
	.dc.b	$9E,$FA,$46,$23,$DF,$3B,$AE,$51,$E7,$3B,$1E,$79,$E0,$F6,$C3,$FD
	.dc.b	$85,$87,$99,$F4,$C7,$08,$C3,$E1,$31,$3C,$9E,$D7,$E1,$7B,$9D,$4F
	.dc.b	$69,$83,$75,$E0,$9E,$58,$7B,$30,$F4,$B6,$B8,$3F,$49,$FB,$A5,$70
	.dc.b	$FC,$C3,$EE,$7C,$E3,$E7,$BF,$CD,$51,$F6,$A3,$1E,$E7,$36,$C6,$67
	.dc.b	$6F,$2F,$33,$FD,$2E,$F1,$6B,$9E,$7E,$51,$88,$FF,$76,$52,$38,$E9
	.dc.b	$F9,$5F,$9E,$7C,$F7,$FB,$C6,$30,$DD,$B8,$D7,$06,$E3,$F9,$0B,$E3
	.dc.b	$8D,$BE,$08,$63,$E1,$AF,$CB,$B6,$3F,$07,$5C,$9B,$5C,$9E,$38,$78
	.dc.b	$5E,$95,$C7,$1F,$3C,$78,$4E,$B3,$B7,$8E,$19,$7A,$9A,$63,$82,$4F
	.dc.b	$EE,$6B,$B6,$3F,$02,$72,$FB,$56,$3E,$9A,$E1,$87,$86,$3D,$98,$7F
	.dc.b	$96,$1B,$AF,$16,$DF,$B2,$B8,$65,$D6,$F8,$36,$F8,$36,$B8,$4F,$F0
	.dc.b	$2F,$CA,$9F,$65,$99,$46,$1F,$E3,$6D,$7F,$D8,$43,$F2,$97,$99,$C6
	.dc.b	$75,$7F,$E7,$06,$E3,$F0,$5B,$7F,$7E,$F8,$BD,$BE,$37,$91,$1F,$E5
	.dc.b	$71,$FE,$44,$74,$43,$C7,$29,$F5,$4E,$E7,$31,$8F,$93,$91,$C7,$0F
	.dc.b	$F4,$2F,$E4,$5F,$EE,$D8,$FE,$72,$D7,$DE,$D8,$FF,$45,$DF,$FB,$BA
	.dc.b	$77,$F1,$FF,$26,$B5,$97,$DF,$6C,$FB,$FE,$3B,$55,$E6,$F8,$89,$3C
	.dc.b	$8F,$E9,$3F,$31,$FE,$44,$6F,$A5,$F5,$63,$11,$F4,$B2,$7D,$76,$6A
	.dc.b	$4F,$03,$E4,$C6,$1E,$1F,$45,$F1,$8C,$6B,$FE,$A6,$FC,$5F,$FB,$84
	.dc.b	$FF,$5A,$6B,$F8,$06,$C7,$FA,$B3,$93,$FC,$3B,$3B,$F1,$3F,$15,$5A
	.dc.b	$E0,$FD,$03,$67,$F7,$D5,$C1,$E9,$5F,$91,$87,$D2,$D3,$9C,$7E,$57
	.dc.b	$EA,$BE,$F5,$7F,$6E,$6B,$F9,$46,$E3,$FE,$C0,$FC,$A1,$E2,$71,$E6
	.dc.b	$97,$C5,$FD,$04,$EA,$30,$7F,$7A,$5F,$45,$A3,$8C,$7E,$8A,$3D,$8B
	.dc.b	$F2,$8F,$CD,$BF,$A6,$6E,$1E,$3C,$BF,$5D,$18,$DB,$F8,$23,$6B,$F9
	.dc.b	$13,$55,$F6,$FC,$A3,$FC,$BB,$1F,$E0,$E6,$BF,$B4,$6E,$1F,$EE,$0A
	.dc.b	$E5,$1F,$D2,$E7,$87,$85,$FB,$F8,$C4,$7E,$F3,$F2,$87,$99,$FC,$0F
	.dc.b	$F2,$C7,$DC,$D2,$57,$23,$FC,$3F,$1F,$E2,$96,$F9,$7C,$4E,$5F,$35
	.dc.b	$FC,$37,$E4,$DF,$77,$49,$5F,$E2,$8F,$3D,$F3,$F8,$8F,$E1,$F5,$C1
	.dc.b	$FF,$16,$B5,$FE,$27,$3C,$1C,$D7,$1E,$0F,$F8,$03,$5B,$5C,$5F,$CC
	.dc.b	$EF,$85,$FF,$1C,$FA,$9F,$A0,$3E,$13,$F9,$23,$63,$FC,$7A,$EB,$C8
	.dc.b	$CC,$C9,$CC,$70,$F1,$64,$70,$79,$FC,$56,$23,$C1,$17,$2F,$2C,$7F
	.dc.b	$63,$C3,$FD,$E4,$FC,$0F,$96,$F9,$44,$7C,$7C,$75,$F8,$F1,$6B,$9C
	.dc.b	$2B,$EE,$BC,$C8,$70,$39,$9F,$E5,$AF,$A5,$78,$DF,$D1,$5F,$BD,$1F
	.dc.b	$D0,$9B,$5D,$9D,$C7,$19,$E1,$C7,$23,$07,$5F,$9E,$E8,$47,$C1,$DD
	.dc.b	$5E,$D1,$70,$ED,$3F,$D5,$FF,$15,$7D,$85,$25,$FF,$9C,$BE,$9F,$F7
	.dc.b	$8F,$F5,$89,$DB,$FE,$76,$6B,$99,$FE,$C5,$3C,$2F,$E4,$D6,$9B,$3D
	.dc.b	$AB,$FB,$1D,$6F,$93,$FB,$F3,$FE,$29,$8F,$17,$A3,$D0,$EB,$DD,$63
	.dc.b	$24,$39,$BE,$2E,$9F,$1F,$4C,$9F,$64,$65,$F7,$46,$5F,$1C,$7F,$7B
	.dc.b	$FC,$57,$FD,$83,$1F,$E9,$46,$BF,$DE,$9B,$9F,$E1,$BD,$0F,$E2,$11
	.dc.b	$8C,$FF,$C3,$0D,$AF,$F1,$B3,$54,$F0,$F7,$7C,$04,$8F,$F5,$1C,$7F
	.dc.b	$1B,$3F,$B4,$3F,$E0,$A7,$F3,$AD,$E8,$FF,$09,$46,$B0,$F2,$74,$62
	.dc.b	$65,$74,$8D,$F4,$70,$FA,$19,$3E,$08,$CB,$E2,$8C,$BE,$18,$FF,$10
	.dc.b	$7D,$0F,$F2,$F2,$75,$E3,$B3,$2C,$AF,$F5,$EC,$EE,$7F,$C5,$1F,$46
	.dc.b	$F9,$9F,$F6,$07,$77,$F8,$FE,$31,$7F,$EB,$68,$C4,$EA,$3F,$CC,$17
	.dc.b	$FD,$23,$DD,$56,$A7,$81,$FE,$65,$BE,$6F,$FA,$EE,$34,$D3,$9E,$3D
	.dc.b	$54,$BE,$4F,$0D,$8E,$A6,$A3,$3F,$51,$A6,$46,$A7,$83,$DC,$EB,$3D
	.dc.b	$FE,$7D,$F5,$1F,$DA,$F2,$F5,$D7,$F9,$F6,$73,$E7,$F8,$1B,$1C,$F0
	.dc.b	$EB,$5C,$21,$3A,$8C,$37,$3F,$F3,$63,$67,$E3,$65,$E3,$C7,$C5,$5A
	.dc.b	$6C,$F8,$DA,$FC,$13,$C8,$E2,$BA,$47,$8A,$37,$87,$9A,$BA,$63,$C0
	.dc.b	$95,$EB,$C3,$C5,$18,$C7,$D3,$87,$29,$CD,$7C,$BC,$7B,$9E,$0C,$7B
	.dc.b	$47,$9F,$0E,$58,$F8,$23,$B1,$83,$47,$3C,$3B,$39,$3D,$53,$6B,$AC
	.dc.b	$60,$E1,$19,$C3,$CB,$8F,$A2,$31,$3C,$F1,$F3,$E1,$DA,$38,$92,$B8
	.dc.b	$3E,$63,$18,$FA,$63,$27,$DF,$C7,$C7,$7D,$30,$EF,$7C,$0F,$6D,$70
	.dc.b	$BF,$D7,$63,$E6,$7A,$12,$7F,$06,$7D,$43,$FD,$64,$7B,$CB,$7F,$92
	.dc.b	$6E,$3F,$D2,$1F,$C6,$F8,$CF,$9B,$FF,$7D,$52,$31,$3D,$0F,$CB,$B8
	.dc.b	$7A,$B5,$E7,$3F,$A1,$8E,$38,$FE,$7F,$EA,$3F,$C4,$3D,$59,$A9,$7F
	.dc.b	$85,$BD,$E5,$FE,$CB,$BD,$CE,$6B,$78,$7E,$9A,$F7,$3F,$EE,$DA,$DC
	.dc.b	$E6,$B7,$7E,$4A,$FD,$E5,$6A,$31,$5B,$9E,$C6,$BA,$FF,$C2,$F1,$92
	.dc.b	$4F,$03,$CD,$1A,$71,$3F,$C0,$F0,$F2,$BF,$23,$0F,$E2,$78,$78,$63
	.dc.b	$51,$E6,$69,$E1,$C7,$0F,$13,$CB,$8F,$CB,$C7,$C1,$1C,$23,$CA,$DC
	.dc.b	$7C,$B5,$CB,$2F,$B9,$87,$4C,$7C,$57,$D7,$0E,$2D,$C3,$8B,$E0,$7D
	.dc.b	$AE,$4F,$A1,$8F,$89,$C1,$30,$EE,$FA,$CE,$47,$94,$82,$FC,$71,$C2
	.dc.b	$FC,$6D,$BF,$1C,$63,$2F,$9B,$39,$BF,$1E,$5E,$F9,$C9,$F4,$B0,$EA
	.dc.b	$DC,$7D,$11,$F0,$87,$D4,$62,$25,$7D,$87,$71,$8B,$FC,$5F,$A5,$C8
	.dc.b	$E7,$E9,$DB,$FB,$F1,$67,$F2,$33,$73,$FD,$23,$FE,$64,$EB,$FF,$2D
	.dc.b	$E4,$8C,$61,$FA,$3F,$66,$E1,$F9,$F9,$AF,$C9,$6D,$71,$A4,$AF,$AD
	.dc.b	$EA,$F9,$FD,$23,$E1,$7F,$D4,$78,$70,$BF,$F3,$BC,$E7,$E0,$77,$5A
	.dc.b	$C7,$A5,$F6,$AD,$3C,$4C,$4E,$EF,$9A,$BE,$7E,$97,$47,$AF,$D0,$76
	.dc.b	$FA,$3B,$87,$3C,$99,$79,$E7,$31,$E2,$33,$8A,$33,$97,$81,$F8,$FF
	.dc.b	$3F,$FE,$F2,$C3,$A5,$A3,$0F,$3F,$8A,$7B,$DF,$2E,$3B,$1B,$F7,$79
	.dc.b	$92,$FC,$D5,$C3,$7C,$57,$0C,$3A,$E5,$FE,$98,$6D,$FD,$6B,$E0,$F9
	.dc.b	$23,$35,$EC,$F3,$BE,$8E,$3F,$49,$B5,$F1,$75,$FB,$51,$8A,$F8,$91
	.dc.b	$89,$D6,$BD,$23,$F0,$1E,$B7,$DC,$3E,$EF,$D6,$5A,$DC,$5B,$E7,$6A
	.dc.b	$C6,$9F,$EB,$2F,$C4,$3F,$DD,$B8,$7C,$C6,$DF,$CB,$DD,$F8,$27,$E6
	.dc.b	$27,$37,$C7,$C8,$C1,$F9,$EE,$0F,$A3,$58,$C7,$A4,$DC,$7A,$4D,$BF
	.dc.b	$C6,$FD,$47,$E5,$8F,$CB,$F6,$DC,$3E,$0C,$5B,$F5,$CD,$AF,$CA,$D7
	.dc.b	$E7,$BE,$4F,$F3,$ED,$3E,$49,$6B,$E1,$B6,$7E,$17,$D8,$FB,$4F,$DB
	.dc.b	$F8,$4F,$A7,$FD,$15,$7E,$D2,$DB,$2F,$84,$FC,$97,$8D,$70,$AF,$DD
	.dc.b	$D6,$EB,$44,$8F,$D1,$5F,$7E,$BE,$71,$F8,$21,$F7,$BF,$C0,$1B,$5F
	.dc.b	$89,$5F,$0F,$99,$93,$97,$84,$62,$7A,$7D,$57,$3F,$89,$E8,$DB,$E8
	.dc.b	$D9,$DC,$77,$9F,$E3,$3F,$54,$F9,$CA,$7F,$1E,$FB,$37,$DE,$3E,$9B
	.dc.b	$38,$7E,$E2,$F8,$36,$F8,$56,$E3,$04,$C7,$F7,$35,$C1,$FE,$11,$87
	.dc.b	$6C,$3A,$C7,$EF,$BE,$6D,$C9,$FC,$B9,$B8,$7F,$08,$E0,$F8,$EF,$F0
	.dc.b	$A8,$CD,$7F,$06,$97,$B1,$FC,$D9,$B5,$FC,$43,$E6,$DF,$1F,$9A,$FE
	.dc.b	$77,$18,$9D,$4F,$F3,$EF,$CC,$9F,$84,$6A,$37,$F0,$32,$BA,$79,$17
	.dc.b	$3F,$E2,$F8,$FF,$19,$EB,$57,$E4,$63,$0F,$E4,$FF,$11,$EA,$F2,$3F
	.dc.b	$92,$B8,$AE,$77,$FD,$42,$31,$7D,$27,$FA,$7D,$F4,$7F,$A2,$72,$FF
	.dc.b	$58,$C1,$F9,$37,$FB,$B1,$FE,$63,$E3,$A7,$53,$9A,$FE,$5F,$39,$AD
	.dc.b	$D6,$BE,$9F,$FB,$0C,$E6,$B7,$38,$35,$5C,$0F,$E7,$E3,$E1,$7C,$97
	.dc.b	$F3,$EB,$F9,$EC,$E4,$FE,$A0,$23,$59,$7B,$E7,$E5,$D7,$F7,$2A,$E1
	.dc.b	$1F,$D4,$DF,$7F,$D7,$66,$64,$78,$9F,$7C,$62,$7F,$A6,$FE,$72,$F8
	.dc.b	$2F,$86,$7F,$7E,$71,$7E,$4D,$2D,$70,$AD,$1F,$2B,$E0,$2D,$7F,$5E
	.dc.b	$BD,$E5,$E9,$6C,$7F,$81,$8F,$AB,$3C,$2F,$FA,$FF,$BD,$FD,$DE,$7C
	.dc.b	$1F,$43,$FE,$10,$8C,$4F,$BF,$24,$E7,$D5,$97,$E7,$E9,$B7,$D2,$6E
	.dc.b	$7F,$89,$AF,$C2,$D6,$1F,$E4,$DE,$67,$DD,$9F,$C7,$4B,$F2,$57,$C2
	.dc.b	$7F,$88,$AB,$57,$FE,$20,$AD,$4E,$78,$E4,$FF,$82,$E5,$EE,$FF,$6E
	.dc.b	$CD,$87,$27,$FB,$A5,$F0,$7F,$96,$CE,$E7,$51,$C3,$E7,$3C,$47,$F4
	.dc.b	$0C,$9D,$58,$CB,$FC,$1F,$6A,$C5,$FF,$38,$BD,$D6,$A7,$D9,$B4,$8F
	.dc.b	$F0,$BD,$7F,$8F,$3D,$CF,$F1,$EF,$D0,$3E,$87,$E3,$DE,$EB,$FC,$2F
	.dc.b	$97,$D4,$C3,$8C,$30,$E3,$7C,$27,$31,$89,$FF,$0A,$5F,$EF,$70,$ED
	.dc.b	$1F,$EC,$E9,$D3,$F2,$27,$57,$FE,$C7,$9C,$E1,$E1,$A5,$F2,$7F,$B1
	.dc.b	$4E,$AF,$8C,$75,$37,$7E,$56,$99,$AE,$57,$D5,$E9,$1F,$14,$D9,$E5
	.dc.b	$C3,$CE,$E6,$38,$DF,$82,$3C,$67,$0A,$E9,$7E,$B8,$C4,$ED,$E5,$1E
	.dc.b	$5A,$F0,$C7,$CD,$37,$7E,$07,$11,$E0,$8C,$61,$F0,$5B,$3F,$EE,$F3
	.dc.b	$D4,$58,$C3,$83,$FC,$15,$F4,$5F,$C6,$FD,$7E,$0D,$8F,$F7,$C3,$5F
	.dc.b	$F0,$80,$70,$F2,$F8,$EF,$F4,$33,$C1,$AF,$08,$46,$23,$87,$1F,$82
	.dc.b	$38,$E3,$F7,$7E,$89,$F3,$DF,$E1,$F8,$C3,$74,$E3,$87,$D6,$C7,$AE
	.dc.b	$3F,$8C,$C3,$8C,$DB,$E0,$FF,$A7,$2B,$6F,$1C,$7D,$3F,$67,$DA,$3F
	.dc.b	$D4,$2D,$4F,$E4,$67,$8B,$67,$D3,$8F,$F8,$D7,$0F,$0D,$FE,$29,$B9
	.dc.b	$6F,$2F,$D0,$E3,$E2,$8E,$C4,$BF,$6B,$BB,$E5,$97,$A9,$61,$D2,$B6
	.dc.b	$48,$FF,$5C,$C6,$5B,$3C,$6F,$C1,$5E,$78,$FF,$26,$61,$FD,$56,$FD
	.dc.b	$2D,$8C,$EB,$D7,$0F,$0C,$7E,$32,$BD,$F1,$F8,$66,$BD,$4F,$CA,$7D
	.dc.b	$1F,$FA,$E5,$BD,$1F,$94,$8C,$47,$E3,$83,$87,$F2,$2E,$B0,$FE,$D5
	.dc.b	$F4,$3F,$E6,$19,$FF,$59,$B5,$FF,$57,$CE,$F1,$FF,$34,$2F,$96,$F5
	.dc.b	$CF,$FB,$7A,$30,$FF,$AF,$1A,$FF,$AE,$29,$E0,$70,$8F,$2D,$7A,$48
	.dc.b	$FB,$9A,$F1,$F9,$2E,$14,$79,$5F,$6C,$7F,$C0,$9F,$47,$FC,$F6,$FF
	.dc.b	$D9,$AD,$7F,$D9,$2D,$8F,$83,$E6,$78,$E3,$53,$FB,$66,$BF,$E8,$76
	.dc.b	$BF,$EF,$C6,$A7,$A6,$7F,$E8,$DC,$5F,$47,$E8,$3B,$74,$FC,$1F,$9B
	.dc.b	$7C,$4D,$26,$3F,$4B,$BE,$BF,$6E,$76,$FF,$15,$6B,$99,$FE,$2B,$18
	.dc.b	$D3,$EE,$FB,$1F,$20,$E6,$FF,$8A,$7E,$0B,$FC,$C3,$7F,$EF,$26,$BF
	.dc.b	$EE,$E6,$C7,$F8,$BF,$97,$FC,$E7,$3B,$F3,$FF,$91,$D7,$08,$FE,$3A
	.dc.b	$D7,$FD,$5C,$D4,$7F,$BF,$23,$F0,$52,$F8,$6A,$7E,$16,$BC,$3B,$FF
	.dc.b	$AD,$FD,$0F,$BC,$DE,$5F,$C5,$46,$31,$EB,$F4,$6F,$27,$35,$FC,$E2
	.dc.b	$BF,$27,$5F,$CC,$9A,$F3,$3F,$A0,$3C,$D5,$3E,$33,$F3,$0D,$8F,$5E
	.dc.b	$BF,$E5,$3A,$E1,$F4,$6F,$33,$F3,$CD,$AD,$E7,$FE,$D1,$8C,$42,$FF
	.dc.b	$CB,$98,$7C,$08,$F8,$F3,$A7,$FA,$4B,$5E,$57,$FD,$52,$31,$1F,$3B
	.dc.b	$E8,$FF,$D7,$8F,$F5,$B6,$BF,$C1,$5A,$FF,$54,$6B,$FE,$E5,$7F,$CE
	.dc.b	$CF,$1F,$43,$FB,$0C,$62,$39,$47,$F6,$4F,$A2,$7C,$0D,$24,$FF,$6B
	.dc.b	$8C,$B7,$E7,$F8,$11,$C1,$FD,$C7,$E1,$BF,$C7,$2F,$C8,$3F,$12,$E1
	.dc.b	$F9,$1A,$FF,$BE,$E3,$1A,$FF,$A1,$7E,$2F,$FC,$9B,$5C,$44,$F1,$7F
	.dc.b	$B9,$B5,$E5,$E7,$FF,$7E,$8C,$47,$82,$3F,$8C,$4B,$EF,$A9,$FC,$6E
	.dc.b	$BC,$38,$7E,$7A,$DF,$45,$F7,$B1,$FF,$4C,$3F,$E8,$D8,$F2,$35,$F5
	.dc.b	$9F,$A6,$78,$4F,$36,$C6,$7D,$7F,$E5,$51,$C6,$BA,$F0,$E3,$C3,$97
	.dc.b	$D9,$DF,$CE,$62,$3B,$9E,$68,$C5,$79,$A7,$D9,$F2,$D2,$BE,$56,$68
	.dc.b	$F0,$FC,$76,$67,$FA,$84,$E5,$FB,$53,$F2,$C8,$FC,$0F,$B0,$F4,$9C
	.dc.b	$70,$FF,$5E,$7D,$1B,$C2,$FE,$8C,$B8,$E9,$88,$97,$F7,$67,$4D,$9C
	.dc.b	$BE,$E8,$7D,$1F,$42,$47,$1E,$DA,$EE,$2E,$1E,$49,$FE,$B1,$F4,$6F
	.dc.b	$94,$A4,$BF,$C7,$7D,$1F,$F1,$17,$F2,$13,$B7,$FC,$98,$D7,$FB,$B3
	.dc.b	$5F,$F2,$2B,$5F,$C5,$3B,$BF,$F7,$04,$BE,$53,$FB,$53,$6B,$BD,$E3
	.dc.b	$D6,$91,$E4,$77,$DB,$88,$FC,$0D,$7F,$93,$1D,$1E,$A8,$EB,$EE,$F6
	.dc.b	$EC,$93,$EE,$8C,$BF,$02,$32,$F4,$8F,$EF,$DF,$46,$FA,$3A,$49,$FD
	.dc.b	$A7,$A7,$F9,$C5,$3B,$7F,$CE,$4D,$7F,$C3,$8D,$7F,$CD,$6D,$7F,$60
	.dc.b	$EF,$57,$6F,$EC,$EB,$FC,$39,$18,$AF,$EA,$71,$89,$E3,$7F,$E1,$58
	.dc.b	$75,$5E,$A9,$7E,$88,$D1,$87,$F5,$72,$9C,$C7,$BB,$AB,$A7,$D1,$C8
	.dc.b	$C1,$0E,$2E,$5E,$71,$FE,$27,$FA,$3F,$E7,$B0,$69,$3B,$7F,$D2,$AF
	.dc.b	$F4,$CF,$7B,$F6,$52,$FA,$2F,$F7,$BB,$FE,$D9,$EE,$FE,$2D,$18,$8F
	.dc.b	$F6,$A4,$7F,$19,$AE,$4E,$BE,$13,$FD,$8E,$D9,$E4,$7F,$7C,$AE,$57
	.dc.b	$E0,$BE,$33,$9B,$FE,$49,$BF,$FA,$8A,$FC,$0D,$8C,$CF,$53,$EB,$1D
	.dc.b	$7E,$8B,$2D,$3F,$C1,$73,$CE,$33,$1E,$F6,$D1,$F4,$C7,$F8,$4B,$E8
	.dc.b	$5F,$45,$FB,$7C,$BE,$3C,$63,$CF,$F6,$D7,$32,$C6,$5F,$8E,$73,$AE
	.dc.b	$5A,$7F,$41,$9D,$5F,$D2,$C7,$C5,$7C,$AF,$6F,$1C,$30,$78,$27,$71
	.dc.b	$D8,$5F,$4C,$3B,$5F,$29,$CB,$6B,$8E,$1E,$C8,$D5,$F1,$AF,$01,$D6
	.dc.b	$B7,$7D,$30,$F1,$CF,$86,$B8,$3F,$0C,$C5,$F4,$24,$FA,$71,$F9,$F3
	.dc.b	$E4,$8E,$78,$F8,$DD,$46,$30,$F0,$3D,$23,$2D,$24,$F1,$7D,$15,$DB
	.dc.b	$0F,$29,$30,$F8,$51,$97,$CB,$3C,$B0,$FD,$0C,$E7,$0F,$14,$FC,$4C
	.dc.b	$3E,$C4,$E6,$38,$19,$AE,$2F,$03,$DC,$68,$F8,$D3,$E7,$24,$7B,$9E
	.dc.b	$B1,$C7,$0F,$45,$FC,$09,$DC,$F7,$21,$D2,$B9,$11,$CE,$3F,$69,$FB
	.dc.b	$A7,$F8,$4C,$F2,$96,$FF,$34,$DC,$7F,$AC,$39,$F1,$1F,$0F,$FE,$CB
	.dc.b	$8F,$C8,$C6,$1F,$F6,$75,$7F,$8B,$27,$4D,$8C,$CE,$F1,$FD,$6B,$71
	.dc.b	$FF,$13,$7D,$0B,$A7,$F9,$A5,$6B,$1F,$D7,$DE,$F0,$FD,$C3,$55,$F1
	.dc.b	$DB,$75,$27,$F3,$15,$BA,$D4,$62,$75,$7D,$E3,$F5,$D3,$A8,$C1,$31
	.dc.b	$FD,$E5,$F3,$C7,$CF,$18,$C3,$9B,$70,$E7,$3A,$7D,$33,$A6,$9A,$C7
	.dc.b	$0F,$CE,$8B,$8F,$8E,$FA,$63,$E3,$AE,$2F,$F3,$E9,$D4,$62,$76,$47
	.dc.b	$FD,$63,$1A,$7B,$1F,$0C,$93,$CF,$0F,$44,$6A,$3C,$0D,$AF,$B3,$3A
	.dc.b	$CB,$F0,$15,$A6,$CE,$B1,$F6,$C6,$2F,$CC,$79,$2F,$96,$5F,$76,$75
	.dc.b	$3D,$CD,$57,$29,$D1,$F4,$DC,$65,$F2,$2F,$9B,$8A,$F9,$C7,$BB,$0F
	.dc.b	$23,$4E,$A5,$3F,$1B,$3F,$E0,$7A,$FD,$56,$5F,$8A,$6C,$66,$79,$C6
	.dc.b	$BD,$1C,$47,$D0,$6B,$F9,$86,$D7,$FB,$36,$7F,$AD,$CF,$E9,$27,$E8
	.dc.b	$CE,$63,$F2,$BE,$8E,$27,$2D,$AF,$EA,$F4,$8C,$57,$09,$FD,$D4,$E5
	.dc.b	$B3,$FD,$BA,$7F,$66,$D5,$F1,$8F,$86,$FF,$5E,$D7,$FA,$CE,$31,$F0
	.dc.b	$3F,$E2,$BE,$DF,$F1,$64,$6F,$D1,$C5,$7F,$C0,$08,$DB,$C6,$F9,$E1
	.dc.b	$FE,$A0,$9C,$B6,$33,$7D,$16,$1D,$27,$CB,$0F,$A1,$C5,$F9,$FE,$86
	.dc.b	$CE,$71,$F0,$46,$2B,$2D,$8C,$9F,$EC,$7F,$A1,$7B,$3D,$F8,$9F,$8B
	.dc.b	$7F,$A8,$6B,$53,$98,$C6,$7F,$EA,$CB,$E3,$5F,$CA,$5B,$1F,$CC,$C3
	.dc.b	$87,$F9,$B3,$5F,$F6,$53,$C3,$5F,$F1,$1B,$65,$2F,$8C,$7B,$7F,$76
	.dc.b	$3C,$58,$FC,$2F,$2F,$E1,$B6,$FE,$93,$67,$E8,$37,$1F,$A0,$FC,$AF
	.dc.b	$5E,$57,$D9,$FA,$2B,$7F,$5E,$6D,$FD,$79,$B5,$FE,$E3,$F9,$F7,$89
	.dc.b	$F6,$69,$F8,$B7,$DE,$AD,$4E,$63,$18,$FB,$AF,$8D,$7F,$53,$6C,$7C
	.dc.b	$B6,$BF,$D7,$DA,$F4,$8F,$8F,$7F,$E4,$C9,$E2,$FF,$A0,$F5,$F9,$3E
	.dc.b	$87,$E3,$6D,$5A,$BF,$CD,$72,$3E,$8F,$F1,$A7,$E3,$BB,$AC,$FE,$76
	.dc.b	$2D,$FE,$6E,$6E,$5F,$9B,$AF,$CA,$FC,$D7,$E5,$9A,$7E,$94,$B5,$FA
	.dc.b	$36,$CF,$A7,$E8,$FF,$29,$F4,$7F,$A4,$9F,$69,$F3,$E3,$3D,$7E,$BD
	.dc.b	$11,$FB,$DC,$3D,$AD,$8C,$D7,$CD,$8C,$1C,$4E,$FD,$96,$3F,$81,$53
	.dc.b	$E8,$FF,$62,$D8,$CF,$C3,$62,$64,$70,$77,$18,$9F,$E0,$43,$E8,$F9
	.dc.b	$FC,$4F,$46,$DF,$81,$B3,$C1,$BA,$FF,$A2,$3E,$85,$F4,$7F,$C1,$F9
	.dc.b	$5F,$82,$BE,$33,$EF,$56,$B2,$F9,$39,$7E,$AE,$B5,$18,$84,$ED,$FD
	.dc.b	$D3,$5F,$F1,$73,$5F,$DA,$9C,$30,$FA,$11,$C7,$1F,$34,$7F,$1B,$FA
	.dc.b	$37,$D2,$3E,$93,$CB,$3F,$93,$FA,$47,$3F,$D9,$6E,$7F,$E3,$56,$E7
	.dc.b	$FC,$35,$AF,$F9,$07,$C5,$FF,$6D,$CF,$F8,$ED,$B1,$FC,$4D,$AE,$FD
	.dc.b	$AF,$C3,$7D,$0B,$E4,$69,$2F,$FC,$89,$19,$6F,$BF,$C0,$8E,$0D,$F9
	.dc.b	$BD,$08,$E0,$98,$38,$3F,$D8,$37,$FC,$6D,$FF,$61,$FD,$13,$E8,$FF
	.dc.b	$6D,$39,$BF,$9F,$E7,$BA,$7F,$A3,$43,$E8,$5F,$01,$FB,$DA,$D4,$FE
	.dc.b	$00,$D4,$66,$BF,$4D,$3A,$CF,$F9,$C3,$5F,$F3,$63,$5F,$0C,$71,$AE
	.dc.b	$96,$FA,$3C,$57,$B6,$7F,$DC,$7F,$09,$E6,$9F,$95,$39,$3F,$DB,$B1
	.dc.b	$9E,$0F,$E2,$94,$8F,$DD,$34,$FF,$3B,$16,$7F,$A7,$35,$FF,$41,$39
	.dc.b	$DB,$FD,$DA,$D3,$FD,$D7,$83,$E6,$FB,$DF,$F2,$3F,$A3,$7C,$8F,$CC
	.dc.b	$AB,$7F,$A5,$9B,$8F,$A6,$3E,$85,$71,$3F,$88,$4E,$A7,$FD,$28,$D7
	.dc.b	$F4,$ED,$7F,$AE,$BD,$31,$56,$AB,$C5,$F0,$7D,$63,$F6,$DF,$01,$FC
	.dc.b	$8F,$0F,$16,$2F,$A2,$FE,$79,$39,$C7,$FC,$79,$39,$BE,$3E,$4F,$F2
	.dc.b	$5A,$E5,$87,$FA,$A5,$AF,$EF,$1A,$FF,$72,$7A,$4E,$67,$E3,$74,$F8
	.dc.b	$1F,$E9,$9E,$1F,$E2,$F1,$F5,$E1,$FE,$AE,$FA,$37,$C2,$6B,$93,$FC
	.dc.b	$21,$18,$CB,$F8,$96,$1F,$8F,$BF,$F0,$0B,$5F,$F0,$15,$7F,$3D,$8C
	.dc.b	$1D,$E7,$F8,$9B,$67,$FD,$7B,$EF,$7F,$82,$DE,$75,$E6,$F9,$AF,$EE
	.dc.b	$DF,$92,$EE,$7F,$C1,$D1,$89,$D6,$1F,$EC,$5C,$13,$E1,$24,$67,$83
	.dc.b	$C8,$47,$04,$BF,$D0,$1F,$E1,$B8,$FE,$E4,$FF,$87,$BE,$0F,$FC,$3F
	.dc.b	$F4,$6F,$A2,$F5,$D6,$A9,$5A,$9C,$DD,$9F,$F7,$14,$D7,$B4,$57,$B0
	.dc.b	$BE,$CF,$CE,$BF,$AF,$18,$8F,$E7,$B5,$CF,$1F,$C5,$47,$1C,$3F,$C7
	.dc.b	$D3,$9B,$E3,$18,$9D,$1D,$70,$F1,$3E,$A3,$27,$98,$A7,$9B,$0F,$16
	.dc.b	$1E,$07,$C9,$1E,$AF,$A4,$CE,$1F,$8E,$73,$1D,$5C,$DF,$F9,$29,$CE
	.dc.b	$3E,$78,$F1,$B6,$7C,$27,$96,$F8,$B8,$8D,$47,$28,$FF,$2F,$1F,$30
	.dc.b	$B1,$87,$07,$F6,$6F,$A4,$FE,$39,$EB,$F0,$6C,$7F,$99,$DA,$FF,$6C
	.dc.b	$0E,$1E,$5F,$23,$FD,$25,$E0,$D7,$84,$23,$11,$C3,$8F,$C1,$1C,$70
	.dc.b	$FB,$DF,$48,$FA,$0F,$EF,$71,$86,$EF,$C7,$0F,$AF,$8F,$5C,$7F,$1D
	.dc.b	$87,$19,$B7,$C1,$FF,$3C,$D6,$DE,$38,$F8,$70,$FF,$04,$DF,$19,$FC
	.dc.b	$1B,$71,$FC,$95,$F1,$6D,$F1,$AF,$F4,$2D,$6E,$F9,$E1,$E1,$9F,$93
	.dc.b	$5B,$6E,$3E,$66,$1E,$47,$F5,$18,$F7,$BE,$73,$AC,$7C,$35,$AB,$E9
	.dc.b	$8F,$86,$BC,$78,$F8,$71,$E5,$18,$3F,$C5,$D3,$EF,$BF,$E4,$B0,$8C
	.dc.b	$69,$F3,$F1,$FA,$53,$C3,$C6,$FF,$17,$DF,$0B,$FF,$71,$4F,$4C,$73
	.dc.b	$87,$F8,$D3,$E9,$3F,$D9,$73,$FE,$3D,$FA,$3F,$F3,$2B,$7D,$8D,$BF
	.dc.b	$EA,$E9,$DB,$FC,$59,$B8,$F9,$AB,$8C,$FF,$1A,$6C,$7A,$A7,$77,$E1
	.dc.b	$F9,$7E,$B6,$94,$61,$B1,$E5,$F0,$F1,$3C,$30,$FF,$0D,$CF,$A7,$C7
	.dc.b	$CB,$FE,$57,$BE,$2F,$8E,$38,$1A,$7F,$97,$FD,$27,$F3,$87,$FC,$BC
	.dc.b	$D7,$3D,$1D,$FC,$DF,$3C,$E7,$4F,$81,$F4,$9E,$33,$35,$2F,$E6,$46
	.dc.b	$5E,$B1,$F0,$3E,$77,$FC,$E5,$1F,$E6,$D6,$BF,$D0,$9B,$5F,$37,$B3
	.dc.b	$87,$B1,$F0,$7D,$A9,$7C,$63,$2F,$DB,$8C,$BC,$E3,$E7,$FC,$47,$F9
	.dc.b	$6E,$3F,$CF,$8D,$7F,$A8,$B6,$FF,$23,$C9,$FE,$70,$9D,$F9,$9F,$62
	.dc.b	$1F,$0F,$4E,$B1,$27,$FD,$11,$3A,$6C,$E5,$FE,$BF,$D1,$F5,$AF,$F3
	.dc.b	$ED,$E6,$FF,$44,$B6,$7F,$DF,$FF,$48,$F2,$7F,$13,$39,$94,$BE,$3F
	.dc.b	$F1,$AD,$8F,$F4,$D4,$EC,$E1,$19,$8F,$EE,$2D,$C3,$C7,$D1,$F9,$6F
	.dc.b	$A4,$79,$B4,$97,$FE,$A5,$BF,$91,$9F,$FA,$82,$B7,$3F,$EA,$9F,$A4
	.dc.b	$FF,$59,$B7,$DA,$E2,$FF,$80,$23,$2F,$F8,$0B,$EA,$5C,$9F,$A6,$6C
	.dc.b	$71,$3A,$7A,$7F,$AA,$FA,$57,$C2,$53,$F5,$B8,$7F,$99,$EB,$FD,$75
	.dc.b	$3A,$6F,$D4,$F1,$38,$3A,$FA,$3F,$DA,$FE,$5E,$FA,$EA,$77,$8F,$DC
	.dc.b	$FC,$5B,$E0,$29,$2F,$FD,$91,$F2,$2F,$57,$11,$FE,$1E,$87,$D4,$7F
	.dc.b	$02,$6C,$7F,$B3,$EF,$A3,$6F,$55,$23,$07,$FB,$57,$EA,$3E,$7B,$FE
	.dc.b	$2A,$7F,$CE,$EE,$7D,$DF,$E1,$F1,$AE,$AF,$E5,$AD,$E7,$FF,$6B,$BC
	.dc.b	$2B,$95,$70,$E9,$E3,$E4,$49,$F0,$3D,$3E,$87,$11,$E1,$8B,$8F,$69
	.dc.b	$F8,$FF,$37,$2B,$DC,$F9,$3E,$53,$11,$E9,$FA,$8F,$39,$8B,$FE,$87
	.dc.b	$BA,$F3,$5F,$06,$B5,$AF,$FB,$C3,$EA,$1E,$57,$F2,$E5,$C9,$4C,$44
	.dc.b	$BF,$F7,$F4,$E9,$B3,$97,$FC,$AB,$0F,$A8,$E8,$48,$E3,$E1,$57,$71
	.dc.b	$74,$FE,$B7,$5F,$B7,$FA,$87,$CD,$7F,$42,$6D,$8F,$F3,$1B,$87,$FC
	.dc.b	$CB,$F5,$1F,$C0,$5D,$5F,$F6,$69,$D3,$FD,$A2,$5F,$33,$FD,$39,$B5
	.dc.b	$C3,$5E,$CA,$F8,$C6,$1B,$1E,$39,$E2,$F8,$AB,$D1,$87,$5F,$6E,$4F
	.dc.b	$53,$83,$E1,$F0,$B3,$1E,$A8,$B9,$79,$E7,$CD,$D9,$E7,$EB,$97,$CC
	.dc.b	$65,$FA,$B7,$0D,$7E,$2C,$70,$7E,$3C,$FF,$6E,$FA,$87,$D4,$52,$4F
	.dc.b	$5F,$5B,$F4,$4F,$D1,$9D,$B9,$F6,$BF,$BB,$CF,$F7,$CE,$0F,$93,$5A
	.dc.b	$AF,$F4,$CB,$74,$FF,$0C,$D6,$E7,$EE,$35,$4E,$EB,$E9,$CB,$AF,$BF
	.dc.b	$5C,$BF,$34,$78,$8C,$3F,$EA,$89,$4E,$63,$D9,$D7,$E8,$FA,$89,$18
	.dc.b	$21,$C5,$CB,$CE,$3F,$C2,$BF,$51,$FC,$33,$E0,$BF,$61,$3F,$95,$FA
	.dc.b	$7F,$F6,$BB,$FE,$B5,$EE,$FC,$5C,$62,$3F,$D6,$71,$F8,$EA,$E4,$DF
	.dc.b	$03,$FD,$52,$F0,$F8,$9B,$5D,$5A,$AB,$8B,$93,$C0,$53,$B9,$70,$F3
	.dc.b	$BC,$6F,$9E,$1E,$59,$CB,$FC,$C2,$78,$87,$11,$AB,$E3,$F5,$1E,$4F
	.dc.b	$A7,$7D,$47,$EC,$EB,$BC,$63,$D6,$F2,$E3,$CE,$F8,$B8,$9D,$3E,$06
	.dc.b	$D7,$92,$7B,$5F,$5D,$3F,$6F,$5A,$6C,$66,$FE,$A5,$70,$C3,$B4,$E6
	.dc.b	$57,$C5,$D9,$A9,$8F,$8E,$78,$64,$BE,$93,$9A,$DE,$3E,$78,$B8,$F9
	.dc.b	$26,$C7,$63,$81,$27,$E3,$E1,$DA,$FA,$C6,$2B,$8D,$7C,$92,$E3,$E9
	.dc.b	$7D,$77,$C5,$CB,$EF,$8D,$63,$E9,$9C,$D7,$48,$C4,$78,$F1,$F4,$34
	.dc.b	$F4,$61,$E3,$70,$78,$2B,$EE,$D7,$29,$D1,$2B,$9B,$F1,$6B,$8D,$FA
	.dc.b	$4F,$2E,$3F,$02,$B6,$49,$F0,$B5,$F0,$1E,$2B,$F0,$46,$6B,$A6,$3F
	.dc.b	$0E,$BB,$BF,$0C,$C6,$1E,$B8,$FF,$75,$4F,$E4,$9B,$3F,$D3,$5B,$8F
	.dc.b	$FB,$EA,$5E,$2B,$E3,$B0,$DA,$FF,$7A,$1F,$8E,$70,$F6,$6B,$F9,$A7
	.dc.b	$86,$1F,$A8,$6E,$3F,$D5,$3E,$A3,$FC,$51,$CB,$FA,$79,$CD,$FF,$65
	.dc.b	$C7,$C7,$8E,$4F,$D7,$DF,$09,$FD,$65,$70,$7C,$93,$B7,$F7,$F3,$BC
	.dc.b	$7C,$53,$96,$CE,$BA,$98,$E7,$2F,$55,$DA,$FE,$F1,$87,$39,$CC,$78
	.dc.b	$A7,$4A,$74,$FA,$F1,$F3,$65,$EC,$8E,$07,$87,$0E,$31,$9C,$BC,$A7
	.dc.b	$F1,$EB,$E7,$5E,$E3,$8C,$6B,$1F,$99,$97,$AA,$73,$1E,$69,$D4,$7A
	.dc.b	$A7,$51,$88,$E8,$DA,$EF,$7E,$3B,$E4,$4A,$E5,$87,$A3,$1F,$69,$F4
	.dc.b	$EF,$8B,$8C,$BE,$2C,$61,$E0,$7D,$CC,$BC,$78,$75,$C3,$C6,$DA,$E1
	.dc.b	$1E,$C3,$E6,$57,$1C,$39,$C6,$4F,$A5,$7D,$1D,$9F,$3C,$D5,$74,$CB
	.dc.b	$E7,$4E,$E7,$C2,$7B,$8D,$FB,$19,$A7,$E0,$56,$30,$DC,$3F,$2F,$1B
	.dc.b	$F6,$71,$7F,$94,$6E,$3F,$75,$B9,$FE,$95,$B3,$F9,$06,$FC,$A3,$E2
	.dc.b	$FF,$D4,$78,$FF,$8C,$E3,$1F,$23,$F8,$1E,$EF,$C1,$46,$FD,$7C,$57
	.dc.b	$F9,$46,$36,$F1,$BE,$78,$AF,$8C,$30,$E3,$2B,$E9,$7C,$67,$31,$E5
	.dc.b	$FA,$9B,$6F,$A9,$B1,$86,$DF,$85,$B8,$78,$63,$CD,$5C,$5F,$F2,$37
	.dc.b	$D4,$7F,$BD,$1F,$F2,$97,$1B,$EA,$7F,$D1,$11,$AF,$81,$F3,$CE,$78
	.dc.b	$3D,$38,$B0,$E4,$F7,$3F,$90,$61,$C9,$C1,$BB,$E7,$19,$DB,$F5,$D3
	.dc.b	$9C,$7C,$07,$F9,$97,$E4,$DF,$0A,$E2,$F0,$E5,$E0,$C3,$E6,$F9,$BF
	.dc.b	$1B,$3F,$8E,$DC,$7E,$A3,$72,$F3,$70,$FE,$06,$31,$7F,$5B,$D9,$F9
	.dc.b	$E7,$D0,$FA,$9B,$87,$DD,$8B,$87,$2E,$15,$62,$FD,$7F,$4C,$F1,$7E
	.dc.b	$FC,$3E,$A7,$F0,$AE,$A3,$31,$8C,$BF,$D1,$F5,$C5,$FF,$43,$57,$1E
	.dc.b	$8F,$A3,$39,$CB,$E8,$4E,$6B,$FD,$05,$39,$8E,$D1,$F2,$F8,$7F,$D3
	.dc.b	$EE,$B1,$FC,$6F,$D4,$C8,$FC,$D7,$7D,$C7,$DD,$37,$0F,$CD,$CD,$AF
	.dc.b	$C6,$9F,$96,$FA,$1F,$CB,$B4,$FD,$21,$6F,$E4,$36,$BE,$3F,$9D,$F8
	.dc.b	$EF,$A9,$FD,$34,$FA,$A3,$1D,$BF,$01,$B1,$93,$F3,$33,$C2,$FF,$1F
	.dc.b	$AA,$B8,$56,$B6,$B5,$F8,$21,$F5,$3F,$B0,$6C,$FF,$03,$5F,$0F,$99
	.dc.b	$93,$97,$84,$62,$7A,$7D,$4F,$3F,$8F,$E8,$DB,$E8,$D9,$DC,$77,$9F
	.dc.b	$D6,$FD,$43,$E2,$FF,$83,$F7,$3E,$A5,$F2,$3A,$9C,$C6,$30,$FF,$61
	.dc.b	$5F,$17,$F5,$51,$9C,$7F,$67,$39,$7F,$71,$87,$5C,$3A,$5F,$2C,$3F
	.dc.b	$8B,$FD,$4B,$EA,$5F,$53,$AC,$7F,$7B,$F5,$2E,$7F,$E0,$BE,$77,$E6
	.dc.b	$E7,$FD,$ED,$39,$CB,$FD,$AD,$39,$9F,$F1,$37,$B5,$FC,$33,$C9,$F9
	.dc.b	$F5,$FE,$DB,$7F,$D7,$CD,$E3,$FE,$23,$F4,$EF,$96,$E1,$78,$6D,$7F
	.dc.b	$B8,$E5,$EF,$FF,$18,$FC,$A9,$EA,$F2,$8C,$3D,$3C,$CC,$5F,$16,$E9
	.dc.b	$FC,$6B,$0F,$E2,$53,$93,$FD,$E1,$F5,$4F,$AA,$E5,$1F,$EF,$3F,$31
	.dc.b	$D7,$FD,$02,$1F,$50,$F8,$1F,$E5,$75,$A9,$65,$FC,$AD,$DC,$EA,$F9
	.dc.b	$5F,$19,$ED,$18,$9D,$46,$0E,$7E,$2F,$F8,$6D,$F0,$FD,$56,$B0,$FE
	.dc.b	$6B,$0F,$86,$F3,$4F,$CD,$8E,$27,$92,$7F,$C7,$B3,$AE,$2F,$E8,$73
	.dc.b	$C4,$DC,$FF,$90,$AB,$93,$DA,$B8,$4E,$A3,$9B,$4E,$4F,$C1,$C9,$F4
	.dc.b	$3D,$AB,$FA,$8F,$D5,$3E,$4B,$E1,$98,$2E,$1D,$63,$4F,$90,$FA,$31
	.dc.b	$B7,$17,$AC,$3F,$AB,$D5,$CF,$FA,$AB,$6B,$EA,$9F,$6B,$E0,$F9,$57
	.dc.b	$F5,$EE,$4F,$E2,$57,$F3,$F0,$7D,$47,$F6,$89,$CE,$1C,$A7,$2F,$C3
	.dc.b	$8C,$DF,$F9,$AA,$72,$F5,$F3,$3F,$03,$7C,$D9,$7E,$AE,$BB,$5F,$DC
	.dc.b	$3C,$4F,$C6,$3F,$DD,$61,$F5,$4F,$85,$FE,$F7,$5A,$D3,$FC,$C1,$5A
	.dc.b	$9C,$F2,$FF,$80,$AF,$86,$3F,$67,$0E,$0F,$E6,$2F,$83,$9A,$E2,$FF
	.dc.b	$83,$BB,$FF,$BC,$3F,$8B,$AE,$39,$7E,$81,$7D,$07,$F6,$8F,$AA,$CC
	.dc.b	$A7,$35,$FD,$92,$73,$18,$9F,$EC,$71,$8A,$E5,$87,$BB,$6F,$7F,$0D
	.dc.b	$C7,$D4,$DC,$3E,$44,$FF,$87,$9F,$D9,$FC,$07,$ED,$3E,$9D,$F5,$3E
	.dc.b	$9A,$D5,$2B,$53,$9C,$2C,$FE,$A4,$D0,$BE,$6F,$C1,$BF,$F4,$84,$62
	.dc.b	$3F,$61,$7C,$9F,$85,$3A,$BF,$F6,$5C,$E7,$4F,$C3,$DF,$1B,$F4,$9D
	.dc.b	$EB,$C8,$D3,$C9,$5C,$6F,$A4,$60,$D1,$E0,$C3,$C4,$F4,$BF,$36,$1E
	.dc.b	$0C,$3C,$73,$C2,$30,$F3,$9D,$DF,$68,$F6,$9B,$70,$F5,$AF,$3C,$FA
	.dc.b	$E3,$E4,$BC,$AF,$C7,$18,$9D,$9E,$27,$FD,$DE,$7D,$02,$E1,$87,$E2
	.dc.b	$9F,$57,$EA,$7F,$91,$7B,$7C,$1B,$1F,$EF,$86,$BB,$E9,$FF,$45,$39
	.dc.b	$F9,$1F,$F4,$03,$79,$79,$46,$23,$B3,$5F,$E5,$94,$8E,$39,$FE,$07
	.dc.b	$EA,$5F,$4B,$FC,$C6,$30,$DD,$38,$D7,$06,$E3,$F8,$7B,$E3,$8D,$BE
	.dc.b	$0F,$F4,$EC,$7C,$15,$F8,$D6,$C7,$E0,$2B,$8B,$6B,$8A,$9D,$E3,$F9
	.dc.b	$76,$E1,$FD,$1E,$F8,$CF,$A2,$BB,$63,$E7,$7C,$4D,$53,$C2,$B9,$61
	.dc.b	$E0,$C3,$AD,$70,$C7,$D8,$C3,$F2,$F1,$A7,$CD,$87,$5C,$3A,$47,$F6
	.dc.b	$E6,$BA,$AF,$DB,$FD,$7B,$F0,$47,$C1,$FF,$61,$97,$07,$F7,$6A,$D7
	.dc.b	$C1,$E6,$31,$68,$C6,$0A,$E9,$86,$1B,$1F,$DA,$1E,$6D,$AE,$4D,$9E
	.dc.b	$1E,$93,$E5,$29,$27,$FB,$8F,$D9,$BF,$00,$FF,$64,$47,$5A,$E0,$DA
	.dc.b	$E0,$FF,$AE,$6B,$87,$C5,$79,$79,$9F,$29,$49,$5D,$BE,$C9,$F6,$58
	.dc.b	$8F,$F6,$16,$DF,$C8,$EB,$1C,$3F,$22,$59,$30,$E3,$1F,$07,$1E,$ED
	.dc.b	$BE,$3C,$3F,$E2,$4B,$E1,$97,$C9,$F5,$BF,$93,$65,$D7,$34,$62,$FE
	.dc.b	$B3,$71,$F0,$FC,$33,$E5,$F9,$74,$C9,$FA,$55,$BC,$FF,$D7,$55,$99
	.dc.b	$F6,$CF,$3C,$7B,$DF,$DE,$BD,$D7,$DB,$FB,$07,$D6,$FD,$F6,$C6,$63
	.dc.b	$87,$DA,$7F,$B2,$A3,$04,$3E,$5C,$F0,$7E,$B4,$6D,$CC,$F5,$23,$F8
	.dc.b	$28,$D6,$BF,$17,$B5,$F3,$3F,$91,$6D,$7E,$45,$AF,$E0,$FE,$DB,$FD
	.dc.b	$1E,$E6,$78,$11,$FC,$7C,$69,$FC,$0C,$F2,$AF,$05,$6E,$7A,$7C,$17
	.dc.b	$E8,$7F,$03,$FF,$28,$47,$E8,$5A,$FE,$29,$B5,$F8,$9D,$1F,$73,$FD
	.dc.b	$AE,$3F,$01,$3A,$E3,$FD,$A3,$6F,$F6,$0D,$BE,$31,$E1,$F0,$E3,$E1
	.dc.b	$7E,$D6,$1D,$6F,$F7,$DE,$2F,$AD,$B9,$FE,$E6,$31,$7F,$C1,$FF,$0D
	.dc.b	$FE,$4F,$9F,$E4,$FE,$ED,$F5,$1F,$BF,$6D,$7E,$FD,$AF,$EA,$2F,$9F
	.dc.b	$2A,$73,$7C,$FE,$B3,$91,$CF,$0F,$E1,$BE,$1B,$27,$DC,$61,$B9,$FF
	.dc.b	$9D,$27,$53,$F9,$7F,$4D,$F5,$FF,$C4,$FE,$F3,$F1,$D9,$FE,$AD,$BC
	.dc.b	$1F,$C7,$DB,$8F,$F1,$BD,$DE,$9F,$F1,$BF,$C1,$3F,$1F,$7F,$C9,$57
	.dc.b	$CD,$64,$D7,$CD,$7F,$24,$C1,$F5,$EF,$8A,$C3,$63,$FD,$0C,$D5,$A7
	.dc.b	$F0,$B7,$8F,$87,$C2,$A4,$70,$79,$C6,$23,$8E,$5F,$E9,$BB,$E9,$7C
	.dc.b	$DB,$19,$AF,$E6,$DF,$21,$F1,$AF,$F9,$B3,$FD,$11,$BE,$76,$B8,$24
	.dc.b	$60,$87,$F2,$49,$E4,$FF,$26,$F9,$97,$D2,$7F,$43,$6C,$7F,$43,$6B
	.dc.b	$FC,$D7,$EF,$FF,$0B,$18,$D3,$F9,$03,$73,$FE,$A8,$DA,$FE,$5F,$5F
	.dc.b	$96,$5F,$57,$FD,$11,$F8,$67,$FB,$41,$C5,$FC,$D6,$DF,$FB,$07,$EB
	.dc.b	$3E,$ED,$6F,$D0,$FB,$55,$A8,$C1,$FE,$BD,$9D,$BE,$09,$FC,$04,$7C
	.dc.b	$98,$F0,$87,$17,$E1,$9F,$ED,$3E,$5B,$EB,$5F,$01,$87,$FD,$5B,$87
	.dc.b	$FB,$42,$75,$D1,$9A,$93,$9A,$DB,$F0,$2B,$6E,$67,$FB,$A7,$D4,$7F
	.dc.b	$3D,$27,$9D,$FC,$F7,$F0,$2B,$5F,$DD,$E6,$BF,$DA,$63,$FA,$72,$FA
	.dc.b	$DF,$F1,$E3,$FE,$DF,$6B,$F0,$1B,$A7,$E6,$DF,$E3,$11,$C7,$DB,$FD
	.dc.b	$34,$6B,$83,$FD,$E4,$D9,$1D,$FC,$9D,$D4,$8F,$03,$CA,$3E,$5F,$E0
	.dc.b	$9F,$C9,$75,$34,$FE,$4D,$ED,$3E,$C7,$75,$30,$F9,$D5,$A9,$FA,$BF
	.dc.b	$92,$BE,$66,$F4,$4B,$FA,$93,$96,$CE,$5F,$A5,$09,$E7,$3F,$C1,$7E
	.dc.b	$47,$FE,$00,$D7,$BB,$7F,$E0,$11,$88,$FC,$4B,$63,$3F,$21,$C7,$E4
	.dc.b	$25,$71,$E1,$F8,$EA,$F8,$46,$15,$70,$9E,$B5,$6F,$AC,$7F,$8B,$BA
	.dc.b	$FF,$C3,$2F,$EE,$FC,$3F,$DD,$1F,$E0,$38,$CC,$FF,$59,$6E,$3F,$B6
	.dc.b	$9D,$D7,$3F,$6F,$F1,$DF,$11,$FE,$45,$FA,$F7,$D6,$53,$35,$FE,$3E
	.dc.b	$E8,$56,$63,$FC,$95,$5C,$1D,$47,$E6,$7E,$87,$FD,$97,$F5,$7F,$E5
	.dc.b	$9A,$FE,$FF,$F0,$48,$C7,$AF,$AC,$3F,$C6,$F8,$B0,$C7,$B2,$8C,$4F
	.dc.b	$FB,$9A,$F8,$B4,$FF,$30,$63,$DA,$3F,$C7,$53,$C2,$7F,$DD,$11,$C1
	.dc.b	$A7,$83,$E7,$7B,$C9,$F2,$7B,$2B,$E9,$5D,$B2,$DD,$48,$F1,$56,$E7
	.dc.b	$D9,$0F,$AF,$E2,$BC,$1F,$E6,$7F,$29,$E9,$8C,$79,$48,$C5,$70,$38
	.dc.b	$D7,$0A,$F4,$9C,$27,$77,$E0,$9D,$6F,$F3,$EB,$F7,$31,$8D,$7F,$2A
	.dc.b	$D8,$FD,$EE,$1D,$B1,$F1,$47,$69,$D3,$DC,$F0,$CE,$F1,$F5,$4E,$8D
	.dc.b	$C6,$67,$95,$74,$C3,$C9,$1C,$DB,$3B,$73,$5C,$EB,$86,$1F,$AC,$AD
	.dc.b	$3E,$A8,$CD,$FB,$9A,$FD,$1C,$7E,$24,$62,$FA,$63,$F2,$2B,$D3,$39
	.dc.b	$24,$66,$3D,$2E,$88,$FC,$2A,$EB,$18,$C7,$D3,$5E,$2C,$7C,$F7,$E5
	.dc.b	$AD,$63,$EB,$C3,$8C,$F6,$26,$1F,$12,$32,$F4,$AE,$33,$F6,$27,$35
	.dc.b	$F0,$E7,$33,$F0,$4B,$8F,$C0,$7D,$06,$31,$F7,$B8,$3C,$35,$CD,$A7
	.dc.b	$73,$85,$76,$C7,$EF,$47,$F2,$03,$F0,$C5,$BF,$CC,$36,$FF,$1A,$E7
	.dc.b	$EA,$3F,$B7,$BC,$4F,$D3,$38,$75,$E3,$65,$CB,$69,$3C,$B3,$FC,$4B
	.dc.b	$72,$FE,$67,$F5,$FF,$E4,$AE,$7F,$CB,$CE,$71,$FD,$2B,$5E,$A7,$EB
	.dc.b	$A7,$83,$FA,$38,$D6,$1E,$0C,$79,$9F,$BB,$AD,$E3,$E2,$8C,$DF,$17
	.dc.b	$5E,$AF,$FB,$5E,$75,$39,$23,$FA,$6A,$E7,$97,$EE,$31,$F0,$93,$0E
	.dc.b	$79,$76,$BE,$64,$AF,$83,$8F,$77,$95,$6D,$B8,$FB,$E7,$4F,$C4,$9D
	.dc.b	$34,$D5,$E1,$F8,$F7,$DD,$FF,$03,$E3,$E8,$78,$46,$5D,$E1,$E4,$AE
	.dc.b	$59,$7C,$B9,$F8,$33,$96,$9C,$9F,$3D,$7C,$5C,$7C,$F3,$D0,$D1,$06
	.dc.b	$1E,$37,$44,$BF,$05,$73,$C3,$C7,$97,$CD,$AF,$79,$E2,$38,$3F,$87
	.dc.b	$BE,$97,$CB,$1F,$29,$2B,$E1,$E3,$E6,$9D,$46,$30,$F9,$57,$CB,$0E
	.dc.b	$97,$B6,$CF,$D2,$DF,$FB,$57,$5F,$D6,$E4,$FE,$CD,$F1,$3F,$62,$73
	.dc.b	$5F,$D9,$67,$2F,$F5,$6D,$17,$C7,$95,$87,$19,$CD,$71,$9C,$B7,$E1
	.dc.b	$F5,$7C,$5B,$19,$3B,$46,$5A,$73,$F8,$83,$FE,$27,$D9,$9F,$F8,$17
	.dc.b	$C0,$CC,$7F,$97,$EA,$F7,$E2,$B5,$7F,$E1,$59,$CB,$FE,$56,$8C,$D7
	.dc.b	$18,$ED,$18,$AF,$F1,$37,$CF,$F0,$AF,$0F,$E4,$CE,$FC,$70,$3C,$15
	.dc.b	$EF,$15,$EF,$2F,$F1,$B4,$E6,$31,$8F,$FA,$AE,$31,$8F,$7F,$9C,$7C
	.dc.b	$93,$EC,$F0,$DC,$7E,$24,$EB,$E1,$33,$52,$73,$5B,$C7,$FD,$5B,$7B
	.dc.b	$9C,$D6,$E7,$35,$BD,$BE,$6D,$EE,$73,$5F,$0F,$CE,$79,$4F,$53,$0B
	.dc.b	$EA,$2C,$61,$B9,$FF,$9C,$9B,$87,$F9,$C1,$F9,$1D,$0F,$6E,$9F,$3B
	.dc.b	$A5,$F6,$98,$C3,$EB,$D6,$DB,$5B,$9D,$61,$EC,$F7,$FE,$49,$F2,$BC
	.dc.b	$CB,$7F,$1A,$2C,$FC,$29,$BC,$1F,$0B,$85,$58,$BF,$F4,$FF,$0F,$D6
	.dc.b	$BC,$61,$FE,$4E,$F8,$FF,$C1,$79,$47,$D5,$F1,$F8,$66,$1F,$E6,$67
	.dc.b	$2D,$3F,$D8,$DE,$6F,$CF,$F6,$7F,$CC,$11,$83,$FD,$95,$E7,$66,$74
	.dc.b	$D8,$C9,$FA,$9C,$5F,$57,$F1,$DF,$F7,$7F,$94,$FB,$0F,$F2,$65,$6E
	.dc.b	$B5,$C3,$FA,$EA,$D7,$9D,$FA,$3C,$39,$F2,$5B,$DE,$1F,$EC,$2B,$7D
	.dc.b	$6B,$E8,$F5,$72,$BB,$3F,$AD,$AD,$CE,$70,$65,$FB,$BF,$05,$F1,$1F
	.dc.b	$A9,$3F,$63,$E2,$3E,$D7,$19,$FE,$EA,$B6,$DA,$DC,$EB,$0E,$B3,$FC
	.dc.b	$0F,$EC,$5F,$3D,$FB,$97,$FD,$69,$1F,$C1,$1C,$B7,$CA,$D4,$ED,$B1
	.dc.b	$93,$F6,$B5,$F6,$0F,$D8,$97,$3F,$D6,$BF,$E8,$B8,$DF,$1F,$F0,$18
	.dc.b	$D4,$7E,$39,$A3,$AB,$FD,$DF,$EB,$6E,$A5,$72,$3C,$31,$8A,$F7,$9F
	.dc.b	$CA,$BF,$03,$7D,$6F,$F1,$75,$F6,$96,$30,$FF,$1B,$F8,$E7,$9F,$4C
	.dc.b	$7C,$35,$CB,$F9,$B7,$C8,$3F,$08,$7C,$9F,$F2,$98,$C3,$FC,$A3,$97
	.dc.b	$F8,$45,$70,$AD,$BF,$CC,$BE,$53,$FE,$00,$3F,$CD,$30,$FF,$62,$39
	.dc.b	$F8,$4F,$E5,$D1,$8D,$FD,$BE,$A2,$F8,$C6,$1B,$C9,$FD,$17,$67,$D9
	.dc.b	$7F,$37,$9C,$E5,$FD,$16,$72,$FF,$44,$6D,$FC,$AD,$BD,$DE,$97,$C5
	.dc.b	$C3,$FA,$07,$CB,$BE,$AA,$9F,$D1,$BE,$63,$FB,$3E,$1E,$69,$D6,$5F
	.dc.b	$D4,$1B,$19,$AE,$1E,$D7,$2A,$D4,$59,$FB,$7E,$D3,$0E,$BB,$7F,$61
	.dc.b	$C7,$9B,$FD,$8F,$F0,$3F,$4B,$73,$FE,$B3,$F3,$9F,$80,$CF,$FB,$0D
	.dc.b	$69,$FE,$CC,$DC,$BF,$B8,$B7,$25,$71,$9F,$A3,$3B,$E1,$34,$E1,$3B
	.dc.b	$C7,$C5,$87,$F7,$0F,$C0,$FF,$BF,$B4,$FE,$DB,$F3,$DF,$D4,$DF,$EF
	.dc.b	$11,$AC,$7E,$FB,$6B,$FC,$00,$D9,$4B,$25,$F3,$8F,$F0,$5D,$6D,$B5
	.dc.b	$FD,$E6,$31,$5C,$38,$31,$3F,$DF,$7D,$BF,$DC,$5F,$E4,$1B,$8F,$F8
	.dc.b	$52,$DE,$FE,$2F,$FC,$21,$56,$FF,$C3,$6D,$D4,$E6,$F8,$F7,$F7,$25
	.dc.b	$FF,$85,$A7,$58,$71,$79,$7D,$73,$EC,$7F,$C2,$4D,$AF,$F1,$56,$0F
	.dc.b	$A1,$FE,$4B,$9F,$EA,$2B,$4D,$BE,$27,$F8,$7A,$77,$A7,$F8,$B6,$B6
	.dc.b	$D9,$DE,$1D,$AB,$6F,$F8,$BE,$B6,$FF,$8C,$4F,$DF,$DF,$EE,$1B,$7F
	.dc.b	$27,$E4,$BF,$A4,$47,$F8,$D0,$C9,$C7,$BF,$FC,$5C,$E2,$BE,$F4,$F0
	.dc.b	$7A,$4F,$F8,$CF,$EB,$5F,$1D,$7E,$4F,$0D,$C3,$FC,$83,$3A,$6C,$E9
	.dc.b	$A7,$F6,$BB,$FC,$6F,$C6,$7E,$3B,$E8,$9F,$2F,$F9,$09,$7D,$43,$C3
	.dc.b	$FF,$2B,$D6,$A7,$FD,$C5,$7C,$6F,$85,$EF,$0C,$56,$85,$F1,$7A,$E1
	.dc.b	$D2,$FF,$DC,$98,$70,$9C,$DD,$7C,$E7,$29,$EF,$5C,$63,$B5,$F5,$C3
	.dc.b	$C5,$1B,$33,$5C,$2F,$A4,$30,$F0,$57,$0B,$F3,$4F,$81,$E9,$87,$78
	.dc.b	$F8,$D5,$C2,$31,$1B,$C3,$B3,$E1,$36,$5B,$ED,$1F,$02,$31,$3C,$9F
	.dc.b	$21,$8C,$3C,$F7,$F5,$63,$81,$4F,$8A,$53,$89,$BF,$A7,$65,$FE,$74
	.dc.b	$8E,$11,$8F,$82,$E0,$D8,$CC,$ED,$E5,$F3,$1F,$E1,$C7,$8B,$5C,$F0
	.dc.b	$F2,$8C,$47,$E2,$E9,$1C,$75,$FC,$17,$D3,$BE,$A3,$F2,$11,$86,$ED
	.dc.b	$C6,$B8,$37,$1F,$C3,$5F,$1C,$6D,$F0,$7F,$94,$E3,$E0,$AF,$C6,$36
	.dc.b	$3E,$A5,$71,$6D,$71,$53,$BC,$7F,$30,$DC,$3F,$4D,$7C,$67,$D1,$5D
	.dc.b	$B1,$F3,$BE,$26,$A9,$E1,$5C,$B0,$F0,$61,$D6,$B8,$63,$EC,$61,$F8
	.dc.b	$B8,$D3,$E6,$C3,$AE,$1D,$23,$FA,$4B,$5D,$57,$ED,$3A,$DE,$33,$E8
	.dc.b	$3F,$80,$4F,$FB,$46,$1F,$51,$8E,$1F,$F6,$84,$63,$0F,$F6,$7B,$C5
	.dc.b	$B3,$B7,$FD,$BF,$1B,$7F,$C3,$AD,$AE,$D7,$BC,$1F,$51,$FE,$F6,$BF
	.dc.b	$F6,$DB,$5F,$F1,$2B,$6B,$F9,$77,$C6,$F1,$8C,$65,$FE,$E5,$6D,$7F
	.dc.b	$BC,$1B,$DA,$EA,$FF,$60,$4F,$FB,$A3,$97,$FC,$8A,$DF,$43,$C9,$18
	.dc.b	$8F,$F0,$BF,$6F,$FB,$A5,$D1,$0F,$F1,$CC,$6F,$3F,$ED,$53,$B9,$CD
	.dc.b	$74,$23,$FD,$3F,$E4,$3E,$17,$81,$FC,$96,$DA,$7F,$70,$E6,$C9,$F3
	.dc.b	$74,$FE,$EB,$EF,$E3,$1F,$ED,$93,$5F,$F1,$6F,$85,$FE,$D3,$72,$43
	.dc.b	$FC,$01,$1C,$E3,$51,$FE,$52,$F8,$67,$C3,$FD,$56,$C7,$F8,$53,$2F
	.dc.b	$2A,$E3,$FF,$63,$57,$97,$C8,$7C,$4F,$18,$95,$C7,$5F,$F0,$ED,$F0
	.dc.b	$8F,$EF,$B7,$C1,$D1,$D7,$1F,$0B,$E5,$BE,$6F,$F9,$2A,$78,$3F,$E6
	.dc.b	$AE,$D7,$CC,$FE,$35,$B1,$D3,$0F,$EA,$73,$F2,$FC,$B7,$D4,$6E,$25
	.dc.b	$6E,$F8,$BF,$E5,$8A,$DB,$99,$DB,$E1,$7A,$57,$E7,$7C,$FF,$E6,$33
	.dc.b	$FE,$67,$F0,$B0,$7D,$BF,$43,$F9,$E7,$07,$F9,$86,$33,$3B,$3F,$C4
	.dc.b	$33,$AD,$FE,$4D,$F4,$9F,$F3,$A5,$70,$7F,$CD,$B5,$B7,$EB,$78,$BE
	.dc.b	$AC,$3E,$CE,$BF,$DF,$73,$FD,$94,$E7,$2F,$F2,$8F,$7B,$EA,$3F,$93
	.dc.b	$61,$C3,$D5,$FF,$52,$E5,$FE,$9F,$F2,$FF,$75,$18,$8F,$F4,$DE,$3C
	.dc.b	$DF,$C3,$43,$E4,$F6,$48,$E4,$7F,$AC,$B0,$7D,$47,$93,$E6,$24,$E6
	.dc.b	$3F,$84,$7B,$1F,$84,$CF,$FD,$49,$CF,$83,$75,$FE,$AC,$FA,$8F,$C2
	.dc.b	$CF,$C0,$5F,$23,$FC,$45,$B3,$F8,$57,$7C,$1F,$E2,$3F,$A7,$7D,$47
	.dc.b	$F1,$DA,$46,$1F,$CB,$98,$8D,$47,$EB,$30,$FD,$3C,$E4,$DC,$71,$7F
	.dc.b	$DB,$A6,$4A,$F1,$AF,$15,$74,$24,$F2,$F0,$3F,$95,$7D,$3B,$EA,$2F
	.dc.b	$99,$2B,$F4,$13,$96,$CE,$5F,$F6,$8C,$7F,$3B,$6C,$7F,$35,$FA,$47
	.dc.b	$D0,$D2,$4F,$F9,$D3,$E3,$BE,$F7,$5A,$B7,$3F,$AF,$6B,$FE,$DE,$6E
	.dc.b	$DF,$E8,$FA,$FC,$34,$BE,$A3,$F9,$AB,$F1,$1B,$3F,$29,$BB,$7F,$95
	.dc.b	$FE,$A3,$EB,$D7,$FA,$A2,$72,$FF,$AA,$A3,$D9,$5F,$49,$B5,$FE,$5F
	.dc.b	$AD,$D7,$87,$D1,$FB,$D5,$A8,$E5,$B7,$E0,$6B,$AF,$96,$F9,$87,$C9
	.dc.b	$FF,$61,$8C,$77,$59,$FF,$35,$61,$FC,$72,$78,$56,$4D,$54,$9D,$57
	.dc.b	$47,$F8,$C7,$D1,$FF,$45,$D3,$F9,$4B,$63,$F9,$93,$7C,$E7,$CE,$E7
	.dc.b	$89,$34,$9E,$7F,$F4,$B9,$E5,$6F,$A8,$FF,$1B,$BF,$9C,$6B,$F9,$66
	.dc.b	$F1,$7E,$71,$FE,$24,$DF,$67,$F4,$D1,$AC,$FF,$D1,$8D,$AF,$F6,$EB
	.dc.b	$C2,$33,$3C,$09,$3E,$1F,$27,$E6,$7D,$46,$E3,$FD,$9E,$DE,$13,$9D
	.dc.b	$BF,$BF,$7C,$57,$0B,$97,$F4,$6B,$5E,$C2,$BE,$C7,$D0,$BE,$A2,$F5
	.dc.b	$CC,$3E,$C4,$E5,$B3,$95,$BF,$F6,$4D,$15,$FA,$39,$D9,$FC,$1B,$E5
	.dc.b	$1F,$1B,$FC,$2D,$B1,$98,$C7,$D4,$65,$91,$93,$94,$61,$FE,$13,$3C
	.dc.b	$9B,$3C,$9F,$DB,$7B,$5F,$BB,$F6,$64,$60,$87,$ED,$5C,$BE,$CE,$32
	.dc.b	$E2,$78,$1F,$E1,$DD,$7F,$C0,$71,$F9,$0F,$9F,$FF,$1E,$7D,$07,$FB
	.dc.b	$62,$FF,$B8,$35,$E8,$73,$E8,$4F,$E4,$CE,$3F,$11,$27,$FE,$00,$CE
	.dc.b	$9B,$F2,$DF,$EB,$2F,$A2,$FF,$05,$E1,$FD,$DB,$E0,$91,$8F,$63,$57
	.dc.b	$DE,$EE,$7F,$A6,$73,$E0,$31,$E5,$87,$1B,$FF,$73,$BF,$0A,$F8,$46
	.dc.b	$0F,$F1,$14,$F0,$7D,$9F,$1B,$DB,$33,$DA,$75,$87,$7C,$B5,$32,$FA
	.dc.b	$4E,$5E,$C3,$EA,$3B,$3F,$BC,$F2,$9F,$43,$AF,$13,$F9,$BB,$C2,$B6
	.dc.b	$70,$8C,$D7,$FB,$87,$5F,$F8,$01,$18,$DF,$E4,$CF,$E0,$2B,$86,$1D
	.dc.b	$A3,$81,$F8,$4A,$E1,$7C,$A7,$57,$CA,$BD,$19,$FD,$27,$D1,$8F,$9E
	.dc.b	$B7,$7C,$63,$D1,$5B,$C3,$AD,$F7,$BD,$B6,$B6,$F3,$63,$E6,$3B,$97
	.dc.b	$1F,$5C,$6B,$1F,$35,$F6,$AF,$39,$9C,$7C,$8F,$38,$D5,$F2,$8C,$61
	.dc.b	$E4,$7B,$DF,$C1,$9D,$3D,$A7,$55,$CB,$1F,$74,$60,$98,$FB,$30,$EA
	.dc.b	$F8,$DF,$80,$7A,$E7,$8C,$66,$3C,$38,$FE,$E1,$F5,$9C,$89,$87,$8E
	.dc.b	$7B,$36,$33,$3F,$02,$3D,$55,$DA,$72,$78,$A3,$2F,$87,$0F,$94,$7B
	.dc.b	$5C,$61,$EC,$7D,$06,$30,$F4,$63,$F1,$B1,$F0,$12,$7F,$25,$8F,$85
	.dc.b	$F6,$1F,$70,$B8,$7E,$71,$B8,$7E,$5D,$CF,$D3,$FF,$6B,$6F,$91,$C9
	.dc.b	$FC,$D4,$65,$FD,$1B,$5C,$C7,$EB,$63,$FB,$A6,$7F,$9D,$FA,$87,$B1
	.dc.b	$FE,$A4,$8F,$C5,$CE,$B3,$E7,$7C,$4F,$DD,$DF,$0E,$8C,$63,$F8,$DB
	.dc.b	$E0,$F8,$0F,$CD,$4E,$DD,$57,$0C,$3B,$BD,$31,$F1,$63,$E0,$AD,$92
	.dc.b	$BD,$13,$A8,$ED,$1C,$9D,$4F,$E2,$6F,$C1,$39,$24,$75,$79,$E3,$E3
	.dc.b	$8C,$12,$BF,$A4,$D7,$28,$C4,$7F,$85,$31,$F7,$B7,$1F,$55,$F3,$8F
	.dc.b	$2D,$70,$BE,$51,$82,$1C,$4F,$8F,$8F,$B5,$C6,$3E,$87,$73,$C9,$CD
	.dc.b	$79,$33,$65,$E4,$AD,$46,$32,$F2,$5F,$03,$DB,$8F,$79,$D4,$7C,$A3
	.dc.b	$BC,$66,$FD,$F3,$AC,$7C,$94,$C3,$BD,$68,$98,$7C,$49,$E8,$78,$49
	.dc.b	$3C,$23,$55,$D2,$31,$3F,$13,$2F,$4C,$62,$7F,$9F,$EF,$FD,$A3,$AF
	.dc.b	$F0,$9C,$9F,$E4,$AF,$A8,$FC,$34,$E7,$0F,$EC,$13,$97,$FA,$BE,$DF
	.dc.b	$E1,$C7,$3C,$AB,$E3,$39,$CF,$FA,$4D,$F1,$6D,$3E,$1B,$79,$78,$A3
	.dc.b	$18,$F6,$8C,$4E,$A3,$BB,$6B,$9F,$C3,$A7,$FE,$00,$62,$EF,$C5,$7F
	.dc.b	$84,$B2,$FF,$33,$DE,$3B,$3F,$DC,$B7,$BC,$3F,$C2,$33,$97,$FC,$47
	.dc.b	$AF,$F8,$3B,$1C,$63,$FE,$1B,$AB,$39,$AF,$F5,$37,$D4,$63,$E8,$9E
	.dc.b	$0F,$0D,$3F,$C5,$73,$98,$C7,$80,$8C,$5F,$23,$FC,$67,$AF,$FB,$EB
	.dc.b	$0E,$71,$83,$FC,$6F,$F3,$EF,$91,$7C,$B7,$9A,$30,$D8,$FF,$7A,$4F
	.dc.b	$1D,$95,$A8,$C3,$63,$FC,$72,$DC,$3E,$5B,$DF,$E0,$BA,$61,$2F,$8E
	.dc.b	$1E,$26,$CE,$70,$FF,$21,$F9,$1F,$E6,$8F,$5F,$E5,$BF,$E6,$AF,$A8
	.dc.b	$BE,$52,$6C,$61,$B8,$7F,$9D,$9B,$7F,$E4,$EE,$2F,$F2,$1F,$B7,$F5
	.dc.b	$39,$8F,$D1,$E9,$7D,$46,$62,$63,$F8,$87,$51,$8C,$3E,$0E,$CF,$80
	.dc.b	$F9,$47,$CB,$F3,$2E,$1F,$02,$2D,$70,$9B,$C7,$F2,$2F,$95,$EC,$F8
	.dc.b	$F3,$A7,$FC,$BB,$F1,$DF,$8C,$F2,$8F,$A7,$E0,$B0,$C9,$FF,$37,$B9
	.dc.b	$69,$FE,$CC,$F3,$BE,$6F,$B2,$F3,$AC,$61,$B3,$FE,$D2,$6E,$9F,$5A
	.dc.b	$B8,$4F,$E1,$E1,$F4,$FF,$32,$3F,$15,$E3,$BE,$A3,$FC,$AB,$1B,$9D
	.dc.b	$57,$FA,$0A,$B5,$E7,$FE,$AA,$39,$F1,$FE,$B2,$73,$7C,$63,$18,$74
	.dc.b	$AF,$F4,$59,$C2,$74,$78,$BE,$A1,$F5,$1D,$72,$F0,$F1,$FE,$A7,$E8
	.dc.b	$BF,$68,$7E,$C3,$EA,$2E,$1F,$A3,$EA,$78,$4A,$CE,$7F,$C2,$7C,$46
	.dc.b	$18,$C7,$F8,$6E,$0F,$A8,$7D,$47,$F0,$47,$F8,$66,$1F,$95,$73,$9F
	.dc.b	$77,$37,$C9,$CC,$EF,$CA,$E5,$7C,$DF,$DE,$47,$D4,$3F,$6A,$5A,$FD
	.dc.b	$AB,$FE,$92,$BF,$DD,$B9,$7F,$83,$B4,$FE,$02,$5F,$1B,$FD,$83,$EB
	.dc.b	$76,$E7,$95,$C9,$F0,$C6,$2B,$F6,$1F,$50,$FA,$8F,$E4,$6B,$EA,$2C
	.dc.b	$61,$FE,$4F,$F1,$CF,$3E,$98,$F8,$6B,$9F,$F1,$5F,$93,$7D,$43,$E5
	.dc.b	$35,$39,$8C,$3F,$67,$05,$FF,$34,$3F,$D8,$B1,$F7,$5B,$A7,$F0,$7F
	.dc.b	$94,$7C,$26,$FC,$CF,$E7,$71,$8B,$FC,$23,$6B,$F8,$76,$0F,$A7,$DE
	.dc.b	$1E,$C6,$5F,$97,$D3,$B7,$FC,$57,$E5,$DF,$51,$FD,$21,$B5,$FC,$CF
	.dc.b	$D8,$7C,$C7,$F6,$9E,$9F,$C6,$D6,$B0,$E5,$87,$13,$FA,$65,$6F,$0F
	.dc.b	$EA,$F5,$BB,$E3,$5B,$9E,$D5,$BB,$76,$B0,$E1,$5F,$D8,$3E,$A3,$DC
	.dc.b	$FE,$95,$B5,$FD,$AF,$E6,$FE,$F4,$E7,$5F,$EC,$75,$AC,$FF,$4A,$D9
	.dc.b	$FF,$76,$57,$0B,$FE,$71,$39,$7C,$6D,$C4,$E7,$45,$71,$C3,$AE,$3F
	.dc.b	$DC,$BE,$A3,$FC,$05,$97,$EE,$DB,$97,$F8,$07,$E7,$BF,$AA,$3F,$D1
	.dc.b	$E3,$37,$FB,$A6,$D7,$F8,$01,$B2,$C5,$7C,$AF,$85,$F8,$48,$7D,$B8
	.dc.b	$D3,$CC,$FF,$03,$FB,$5F,$B5,$BF,$EB,$0D,$AF,$F0,$BD,$FF,$89,$7B
	.dc.b	$76,$4B,$FE,$2D,$5B,$57,$E5,$F7,$B0,$FD,$C9,$BB,$1A,$6B,$0E,$D7
	.dc.b	$D3,$0F,$F0,$DF,$D3,$BE,$A2,$9F,$E1,$9F,$A1,$FE,$49,$5F,$C9,$E7
	.dc.b	$4D,$9D,$BF,$9B,$9D,$E5,$FD,$D6,$B6,$DA,$DB,$FE,$35,$9D,$E9,$FC
	.dc.b	$B2,$B6,$FF,$2E,$BF,$DF,$D7,$F7,$F6,$C7,$C8,$F9,$0F,$E8,$31,$FC
	.dc.b	$EC,$C9,$FE,$3A,$2C,$FE,$81,$B9,$27,$8C,$F5,$ED,$FF,$1E,$FD,$43
	.dc.b	$E4,$E9,$8F,$8F,$B8,$FF,$49,$94,$62,$FF,$C2,$91,$89,$D5,$7F,$61
	.dc.b	$9F,$B3,$F1,$1F,$6B,$E8,$9F,$2F,$FE,$FB,$FA,$8F,$E3,$78,$66,$3F
	.dc.b	$DB,$55,$6F,$F0,$35,$A5,$7C,$9F,$F7,$4C,$ED,$4E,$E7,$51,$F8,$A1
	.dc.b	$3B,$9D,$57,$82,$7A,$BE,$5F,$0F,$FC,$DB,$7C,$A7,$D6,$68,$F2,$C6
	.dc.b	$A7,$94,$F8,$2B,$AD,$79,$8E,$07,$9A,$B8,$39,$9E,$11,$88,$F4,$5F
	.dc.b	$9E,$31,$1E,$29,$DE,$1F,$06,$31,$87,$A6,$31,$1F,$00,$F0,$E1,$E4
	.dc.b	$7A,$98,$BE,$AE,$CF,$90,$68,$A7,$C2,$29,$B3,$E7,$7D,$3F,$E2,$FE
	.dc.b	$07,$2D,$8E,$2D,$7A,$B5,$E3,$F3,$0E,$2E,$2D,$8E,$2D,$78,$B5,$E3
	.dc.b	$1F,$83,$FA,$77,$D4,$7E,$07,$5F,$C1,$C6,$2B,$F1,$18,$30,$ED,$5F
	.dc.b	$88,$8C,$61,$CF,$0E,$5A,$FE,$2F,$1E,$58,$FD,$CB,$DB,$71,$E7,$5B
	.dc.b	$BE,$33,$FC,$72,$F8,$E3,$DE,$F7,$3F,$C8,$30,$E7,$3E,$8B,$E3,$2B
	.dc.b	$E3,$87,$C0,$8F,$06,$3E,$76,$18,$25,$F3,$72,$F8,$63,$E1,$E3,$E2
	.dc.b	$9E,$D8,$78,$30,$EF,$1E,$59,$F3,$D7,$F7,$CF,$33,$F5,$CF,$FB,$47
	.dc.b	$C3,$FF,$13,$57,$FA,$CE,$34,$FF,$AD,$7A,$DF,$9E,$7F,$B6,$A3,$F0
	.dc.b	$1C,$3F,$EC,$4B,$E3,$18,$3F,$08,$E2,$FB,$61,$E0,$AF,$E8,$F8,$74
	.dc.b	$71,$7F,$D1,$7F,$14,$FF,$2E,$BE,$17,$FC,$C4,$DF,$63,$74,$9D,$CF
	.dc.b	$FA,$F1,$B8,$7F,$BC,$27,$55,$C2,$79,$D6,$FD,$7E,$7F,$9B,$BE,$A6
	.dc.b	$FD,$9C,$AF,$85,$39,$6C,$E5,$FF,$65,$C2,$79,$CC,$9E,$31,$BC,$BC
	.dc.b	$95,$CE,$7F,$DA,$7F,$8A,$FF,$A9,$5F,$A0,$BC,$3F,$7B,$63,$3E,$BE
	.dc.b	$0C,$CC,$8C,$90,$EA,$E5,$EF,$1F,$2B,$EC,$1F,$09,$7A,$E5,$7E,$06
	.dc.b	$72,$D9,$CB,$FE,$EA,$8F,$F2,$FB,$92,$19,$9D,$E5,$F4,$E7,$9C,$FF
	.dc.b	$BC,$BF,$15,$7C,$9F,$DC,$5F,$79,$4C,$44,$9F,$CA,$4E,$9B,$39,$7A
	.dc.b	$4F,$17,$27,$07,$F1,$3F,$8E,$7C,$F8,$FC,$C7,$22,$3E,$74,$61,$B8
	.dc.b	$FE,$2D,$E0,$DD,$39,$EB,$2B,$F5,$15,$AE,$5F,$CF,$B6,$3F,$C4,$4F
	.dc.b	$48,$CC,$F9,$2B,$F3,$BF,$88,$3D,$9A,$4C,$3F,$D2,$9F,$61,$FE,$03
	.dc.b	$7F,$6D,$3B,$7F,$C6,$79,$31,$FE,$EB,$F4,$1F,$EF,$A9,$CC,$63,$E3
	.dc.b	$FA,$47,$EE,$7C,$55,$6E,$30,$DB,$E9,$1F,$DE,$A3,$17,$FB,$CC,$3C
	.dc.b	$53,$CF,$F0,$3F,$D2,$3E,$03,$FC,$02,$E1,$D7,$47,$EF,$3F,$1D,$E3
	.dc.b	$8F,$EE,$57,$E0,$5F,$C1,$BE,$05,$F8,$BF,$EF,$B3,$FF,$28,$C6,$3B
	.dc.b	$B9,$57,$F0,$6F,$94,$FC,$36,$7F,$79,$B9,$7F,$05,$6E,$1C,$B1,$E9
	.dc.b	$5F,$E5,$CF,$AE,$FF,$60,$3E,$90,$FF,$AE,$1C,$35,$FE,$44,$6B,$D1
	.dc.b	$FE,$3A,$D3,$7F,$03,$89,$E1,$8F,$F2,$D7,$BC,$F2,$6C,$66,$78,$1C
	.dc.b	$6B,$F9,$1F,$D3,$7F,$5E,$9F,$E4,$0D,$C7,$F9,$7B,$70,$FF,$39,$2F
	.dc.b	$C4,$38,$F9,$92,$7F,$A0,$4E,$FB,$0E,$E3,$FA,$27,$CB,$7E,$EB,$1F
	.dc.b	$E5,$CF,$F0,$28,$FD,$5B,$63,$3E,$1F,$F3,$88,$E1,$97,$F4,$27,$4D
	.dc.b	$F0,$F5,$3F,$23,$EB,$9F,$6F,$49,$8F,$D3,$FA,$F7,$CC,$E2,$3E,$8D
	.dc.b	$70,$7E,$C3,$5F,$AB,$3C,$1C,$D7,$D6,$8F,$C7,$C7,$5E,$5E,$9A,$7F
	.dc.b	$4B,$FC,$53,$D3,$3F,$E9,$B7,$D2,$DC,$FF,$AA,$C6,$B7,$34,$DB,$88
	.dc.b	$DE,$9F,$D7,$3F,$11,$FF,$31,$E1,$FD,$65,$7D,$3E,$4D,$7A,$D3,$2F
	.dc.b	$EE,$D3,$A6,$CE,$5F,$7F,$D6,$7B,$7E,$9E,$47,$C9,$8F,$80,$61,$F8
	.dc.b	$B1,$87,$47,$CF,$75,$87,$F8,$0F,$65,$7F,$74,$FA,$77,$D7,$DF,$B0
	.dc.b	$98,$FF,$82,$E7,$2D,$9C,$BF,$CC,$A1,$3C,$E7,$FB,$FF,$E2,$CF,$A2
	.dc.b	$BE,$04,$C3,$FC,$2D,$39,$6C,$E5,$FE,$7B,$09,$E7,$3F,$B4,$F9,$A7
	.dc.b	$DB,$58,$97,$FB,$F9,$CB,$67,$2F,$EF,$A3,$C0,$D8,$E8,$6B,$5F,$F1
	.dc.b	$44,$E9,$B1,$FE,$1A,$FC,$5B,$F9,$7D,$7F,$8B,$DF,$DF,$F5,$FF,$89
	.dc.b	$9B,$1F,$EE,$97,$6F,$F5,$86,$D7,$F8,$7D,$DF,$31,$A6,$AB,$FC,$63
	.dc.b	$FA,$43,$E7,$BF,$C5,$6F,$BB,$0F,$EC,$6D,$C3,$FC,$54,$EE,$3F,$C8
	.dc.b	$8D,$77,$8F,$23,$9D,$7F,$8D,$BF,$21,$FF,$38,$FC,$37,$F5,$5C,$D1
	.dc.b	$8F,$4B,$FC,$86,$DC,$13,$AF,$13,$FD,$5B,$5A,$F1,$2D,$6A,$F8,$5F
	.dc.b	$49,$CB,$C6,$7F,$DB,$B5,$C5,$A7,$6F,$59,$F6,$1C,$1B,$1A,$AE,$8D
	.dc.b	$C3,$BC,$EA,$3F,$DA,$F7,$FE,$5C,$8C,$65,$EF,$AF,$65,$3E,$D3,$F9
	.dc.b	$5B,$E2,$1F,$69,$FD,$C3,$C6,$C3,$63,$A8,$39,$79,$1F,$09,$A7,$0F
	.dc.b	$4F,$F8,$E6,$3F,$DC,$AB,$93,$FD,$D3,$0F,$0D,$74,$3C,$35,$C3,$3F
	.dc.b	$F7,$95,$F0,$6C,$66,$7C,$55,$CA,$FB,$4E,$67,$9E,$3E,$33,$DD,$5B
	.dc.b	$BE,$71,$99,$F1,$61,$DD,$EB,$8F,$B6,$31,$1D,$0D,$E1,$DF,$0E,$91
	.dc.b	$F1,$67,$64,$9D,$57,$37,$CD,$19,$BF,$2C,$E5,$F2,$63,$EC,$73,$78
	.dc.b	$C7,$D2,$F8,$A3,$2D,$24,$F1,$7E,$76,$3E,$67,$D5,$19,$25,$7D,$D8
	.dc.b	$FB,$67,$27,$BC,$6A,$3E,$29,$E3,$3E,$06,$3E,$C3,$DE,$E4,$EF,$3C
	.dc.b	$8E,$D5,$E3,$C7,$E1,$4F,$73,$A9,$4E,$27,$6C,$3D,$F8,$78,$E3,$24
	.dc.b	$AE,$98,$78,$EB,$C5,$18,$7C,$97,$F7,$23,$15,$F0,$4E,$C5,$3F,$00
	.dc.b	$66,$3F,$48,$DF,$B4,$FF,$45,$BF,$98,$6B,$F8,$36,$BF,$AE,$6C,$1F
	.dc.b	$F4,$EE,$1F,$A7,$6E,$3F,$4F,$ED,$3F,$AE,$F3,$7E,$D2,$73,$87,$5F
	.dc.b	$17,$07,$EE,$EB,$71,$F9,$C9,$DB,$E1,$9D,$BF,$B1,$9D,$E1,$E1,$9C
	.dc.b	$B6,$75,$E0,$7F,$9C,$23,$04,$C7,$E0,$57,$86,$75,$18,$AF,$14,$F4
	.dc.b	$8F,$64,$EA,$7D,$46,$89,$5F,$E7,$AB,$F2,$D7,$38,$F0,$4F,$07,$55
	.dc.b	$CE,$3D,$C7,$29,$E0,$78,$31,$F1,$47,$18,$CB,$FC,$76,$FB,$B9,$21
	.dc.b	$D2,$FB,$D7,$28,$C5,$72,$55,$C9,$C9,$CB,$2F,$C0,$63,$F1,$EF,$94
	.dc.b	$62,$7D,$07,$8F,$1F,$64,$67,$1F,$83,$3D,$0E,$F3,$AC,$BE,$FD,$6F
	.dc.b	$2F,$26,$5E,$38,$C5,$F8,$EB,$F0,$B3,$97,$89,$32,$F3,$4F,$90,$EF
	.dc.b	$97,$C1,$AE,$13,$AC,$BC,$B1,$8A,$FE,$93,$C3,$FE,$EE,$E6,$FC,$02
	.dc.b	$FC,$85,$F5,$F8,$E4,$FF,$11,$E7,$FD,$86,$72,$FF,$63,$8C,$E1,$D2
	.dc.b	$72,$F6,$8C,$E5,$FD,$FE,$72,$FF,$80,$63,$3B,$7F,$67,$CB,$C1,$DB
	.dc.b	$E1,$66,$3D,$DC,$B4,$FE,$E9,$FB,$27,$FB,$59,$FF,$66,$68,$87,$9F
	.dc.b	$FE,$D6,$ED,$FF,$6B,$CE,$79,$B1,$7F,$DA,$E3,$85,$63,$6F,$F0,$8E
	.dc.b	$1D,$56,$1C,$A3,$AD,$F8,$1D,$46,$6B,$95,$FF,$8C,$7E,$D6,$F7,$BF
	.dc.b	$02,$B1,$86,$E3,$FE,$41,$6E,$5F,$E3,$D7,$C9,$F5,$2F,$91,$7C,$1E
	.dc.b	$17,$DB,$58,$C3,$70,$FF,$78,$57,$18,$C6,$1E,$C9,$FF,$06,$4E,$5D
	.dc.b	$DF,$66,$D7,$0D,$7F,$E0,$0D,$6E,$F9,$1F,$E3,$8F,$B2,$FF,$3C,$DF
	.dc.b	$D4,$7F,$97,$37,$CA,$F8,$B1,$89,$E3,$B2,$38,$C6,$1D,$61,$F6,$9C
	.dc.b	$B7,$AF,$E7,$F9,$7F,$0B,$1F,$F4,$76,$F8,$D3,$2D,$8C,$9F,$E9,$38
	.dc.b	$CB,$F0,$3D,$4F,$99,$F7,$5F,$35,$B1,$93,$E2,$C6,$5A,$67,$3F,$87
	.dc.b	$B7,$D7,$9D,$46,$37,$47,$E1,$BD,$17,$D5,$5F,$B6,$C3,$71,$FF,$5E
	.dc.b	$B7,$0F,$8F,$5C,$27,$FD,$83,$5F,$90,$F5,$9F,$0F,$7C,$AC,$36,$BF
	.dc.b	$4F,$3A,$6C,$E9,$A7,$F9,$3F,$88,$FF,$85,$7E,$C5,$F4,$3F,$A7,$97
	.dc.b	$E1,$6E,$3F,$D5,$6A,$BD,$E1,$CC,$E9,$5B,$71,$3F,$E5,$D9,$DB,$F8
	.dc.b	$C3,$96,$1F,$B5,$F8,$77,$E1,$6F,$77,$BD,$F0,$CF,$CF,$87,$3C,$32
	.dc.b	$71,$BE,$18,$F8,$27,$AB,$FC,$2B,$DF,$FC,$93,$7E,$2F,$F7,$71,$89
	.dc.b	$FC,$BA,$8C,$F5,$E9,$B5,$C4,$FE,$2B,$F9,$5F,$F2,$8C,$BF,$89,$B6
	.dc.b	$3F,$89,$3F,$ED,$D7,$7E,$2F,$F0,$A8,$D6,$3F,$EA,$46,$F7,$1E,$11
	.dc.b	$FA,$A7,$7D,$5F,$C1,$EF,$F8,$37,$B7,$FC,$1D,$A7,$EF,$8B,$5F,$95
	.dc.b	$6E,$3F,$BE,$EB,$FC,$0C,$7F,$04,$AE,$4D,$3F,$98,$FE,$10,$F5,$BF
	.dc.b	$90,$B6,$3E,$EF,$C3,$BB,$F9,$E5,$FC,$02,$B4,$FF,$02,$F8,$87,$CA
	.dc.b	$FF,$33,$BE,$18,$7F,$2C,$9C,$BD,$30,$E7,$A7,$E0,$3A,$1F,$15,$FA
	.dc.b	$EA,$FC,$6C,$63,$C2,$E5,$F1,$5F,$8B,$AD,$69,$FC,$6A,$9F,$8D,$3D
	.dc.b	$1F,$C8,$37,$1F,$E9,$FF,$1A,$C7,$35,$FD,$02,$31,$3E,$0A,$E1,$87
	.dc.b	$F1,$E9,$CB,$DE,$33,$9F,$F5,$69,$CB,$70,$F8,$5F,$63,$E6,$8C,$61
	.dc.b	$FD,$83,$F3,$9F,$E4,$6F,$F5,$B7,$F9,$3B,$7C,$CF,$EA,$71,$8C,$BF
	.dc.b	$DD,$AF,$06,$F9,$9A,$9D,$B6,$BB,$36,$7F,$9B,$D7,$3E,$A7,$D9,$BE
	.dc.b	$8F,$FB,$0A,$F9,$BB,$18,$6E,$3F,$DE,$27,$4D,$9D,$34,$D4,$F0,$C3
	.dc.b	$FA,$3D,$70,$7F,$A4,$FC,$BB,$E1,$FF,$B8,$AF,$86,$B1,$86,$E1,$FE
	.dc.b	$08,$9D,$36,$74,$D3,$53,$C2,$BB,$BF,$DF,$BD,$9F,$CA,$C7,$F5,$A7
	.dc.b	$EB,$4F,$F8,$35,$CE,$FF,$E1,$27,$32,$D1,$1A,$AE,$1C,$FD,$7A,$3B
	.dc.b	$36,$3F,$C3,$F3,$CA,$B8,$D6,$A7,$FC,$47,$F3,$3F,$51,$BF,$0B,$FE
	.dc.b	$14,$8C,$63,$FD,$C1,$46,$76,$D3,$6B,$89,$D2,$32,$D3,$35,$FD,$EB
	.dc.b	$27,$A3,$FE,$22,$FC,$6F,$F6,$B5,$FE,$2F,$9C,$BF,$E3,$88,$CD,$7C
	.dc.b	$39,$CB,$FE,$04,$8C,$E7,$FD,$A2,$FA,$5F,$F5,$1F,$B7,$F9,$7B,$1F
	.dc.b	$89,$F3,$FF,$E5,$48,$FF,$21,$46,$89,$97,$F8,$DA,$75,$5F,$D7,$E7
	.dc.b	$EA,$E5,$FE,$F3,$F8,$02,$FA,$27,$C9,$7F,$BD,$FC,$07,$C9,$FF,$50
	.dc.b	$9C,$D7,$FB,$6E,$F8,$C6,$21,$5C,$31,$CE,$1F,$ED,$FB,$B5,$E0,$AD
	.dc.b	$3E,$19,$D3,$4D,$61,$E1,$8C,$E1,$DE,$B8,$BE,$1A,$F5,$DF,$67,$C7
	.dc.b	$7C,$5F,$41,$D4,$CE,$1B,$78,$99,$BE,$B8,$78,$0F,$89,$5C,$E7,$E5
	.dc.b	$1B,$70,$F2,$C3,$CD,$18,$86,$1E,$5B,$EB,$1E,$4A,$DE,$1E,$38,$D9
	.dc.b	$D5,$C5,$7D,$88,$87,$88,$A7,$A8,$A7,$23,$7F,$46,$FA,$4F,$F3,$84
	.dc.b	$70,$8C,$79,$7C,$1B,$19,$9D,$BC,$BE,$8F,$FD,$12,$F1,$6B,$9E,$0E
	.dc.b	$51,$88,$F4,$52,$38,$E5,$F8,$9F,$A4,$FE,$DD,$5F,$65,$AF,$D7,$AE
	.dc.b	$17,$F7,$5B,$1F,$6E,$77,$5C,$FC,$CF,$8C,$FE,$56,$B6,$DC,$3C,$3B
	.dc.b	$FC,$BB,$E2,$DB,$E2,$FA,$27,$78,$7E,$A3,$0E,$73,$B8,$F4,$57,$07
	.dc.b	$55,$C1,$57,$07,$F6,$37,$E0,$6D,$71,$77,$5C,$6B,$84,$2B,$83,$6B
	.dc.b	$84,$E8,$DC,$E8,$DB,$E1,$C5,$F4,$9F,$D6,$EF,$FC,$AF,$18,$EF,$FC
	.dc.b	$73,$FE,$CF,$9D,$4F,$FA,$73,$8F,$FC,$A6,$FF,$A9,$A7,$83,$FE,$D8
	.dc.b	$FC,$FF,$F9,$3B,$79,$B9,$F1,$7F,$9C,$70,$E7,$7F,$ED,$AF,$A5,$FF
	.dc.b	$08,$5F,$FB,$81,$FF,$13,$B7,$CE,$E2,$FF,$4D,$8D,$BF,$EA,$E6,$CF
	.dc.b	$39,$DF,$4A,$79,$D6,$FC,$8C,$F7,$3E,$96,$F9,$B2,$BF,$A7,$C6,$8C
	.dc.b	$3F,$D3,$63,$0E,$8F,$EA,$AE,$B2,$FF,$7A,$D7,$59,$FF,$7D,$FD,$2F
	.dc.b	$F9,$B5,$BF,$03,$E6,$8F,$F4,$EC,$6B,$D7,$C1,$99,$91,$92,$1F,$EC
	.dc.b	$E9,$E2,$D8,$E2,$FF,$80,$FE,$43,$E1,$F9,$3F,$CF,$23,$FC,$0D,$E0
	.dc.b	$62,$64,$60,$87,$F8,$49,$CB,$FE,$14,$8E,$66,$32,$FF,$4A,$EE,$FA
	.dc.b	$5F,$F2,$75,$7F,$A8,$A3,$1F,$11,$F5,$27,$FD,$FF,$39,$76,$FF,$BA
	.dc.b	$1B,$5D,$1D,$C6,$67,$97,$A7,$23,$94,$75,$30,$FF,$BC,$BE,$97,$E8
	.dc.b	$37,$4F,$C2,$C7,$C8,$C3,$D2,$E9,$BB,$6A,$65,$7A,$EB,$5C,$DF,$DA
	.dc.b	$B3,$46,$27,$FC,$44,$F5,$9D,$36,$33,$7F,$96,$F4,$DF,$1F,$49,$3F
	.dc.b	$E8,$EF,$A5,$FE,$70,$FF,$9D,$A3,$F4,$7E,$3B,$E4,$1F,$4B,$86,$D6
	.dc.b	$CF,$F4,$3B,$85,$B7,$D8,$FC,$0F,$FD,$EF,$1F,$5D,$AF,$FA,$26,$78
	.dc.b	$4F,$FA,$4F,$E9,$BF,$D2,$73,$F7,$7D,$BF,$F6,$C3,$7E,$1F,$CB,$3A
	.dc.b	$8F,$F5,$14,$7E,$EB,$DD,$7D,$37,$EE,$38,$BF,$C2,$71,$8E,$FF,$DC
	.dc.b	$E3,$D3,$0F,$3D,$7C,$FF,$9F,$FC,$1E,$7F,$90,$6C,$67,$9F,$B6,$5D
	.dc.b	$6F,$FC,$3F,$F4,$CF,$95,$FE,$17,$08,$7D,$36,$1B,$1F,$EE,$B7,$83
	.dc.b	$67,$53,$23,$04,$3F,$DF,$D3,$C9,$B1,$D6,$7F,$8A,$FB,$1F,$D7,$B3
	.dc.b	$FD,$4B,$72,$FC,$63,$6F,$FC,$D8,$BE,$83,$8F,$9F,$27,$F5,$D3,$BE
	.dc.b	$CF,$F3,$0D,$7F,$22,$F8,$A7,$85,$7C,$29,$87,$ED,$67,$2D,$9C,$BF
	.dc.b	$40,$FF,$3E,$47,$F0,$97,$16,$8E,$31,$B1,$F4,$DF,$EE,$57,$FC,$9F
	.dc.b	$A7,$F4,$88,$C4,$6D,$B1,$FD,$33,$E9,$1F,$43,$F5,$A7,$3B,$7F,$88
	.dc.b	$67,$F0,$95,$BE,$8D,$7B,$D3,$1E,$F7,$B9,$FB,$F7,$C2,$3F,$C5,$58
	.dc.b	$74,$AF,$F4,$C7,$CA,$7A,$B1,$FE,$36,$FB,$DB,$B7,$F4,$C8,$D5,$FF
	.dc.b	$01,$6D,$72,$9D,$1B,$DB,$FA,$9F,$CA,$FF,$99,$F0,$FE,$4E,$BE,$9B
	.dc.b	$26,$BD,$99,$87,$F2,$99,$D3,$67,$2F,$CF,$D1,$F4,$3E,$EF,$A6,$93
	.dc.b	$F0,$8F,$F3,$FC,$E9,$C0,$70,$F3,$8E,$2F,$59,$F9,$26,$31,$FA,$D3
	.dc.b	$89,$FE,$7D,$F3,$6F,$8F,$FE,$E2,$BE,$32,$98,$89,$3C,$23,$2F,$28
	.dc.b	$FE,$F3,$F4,$DF,$EC,$A8,$FE,$94,$BC,$0C,$9A,$F8,$89,$3A,$8C,$BC
	.dc.b	$23,$F7,$FF,$14,$F9,$3B,$12,$7E,$1C,$E5,$B3,$97,$F3,$91,$E3,$6C
	.dc.b	$74,$35,$A7,$F8,$92,$74,$D8,$FF,$09,$FC,$F7,$F3,$7A,$FE,$C8,$FF
	.dc.b	$0C,$7F,$56,$D9,$FE,$D6,$EE,$33,$E4,$F2,$F0,$64,$72,$8F,$F0,$FF
	.dc.b	$D3,$3E,$36,$E3,$2B,$FB,$94,$E5,$B3,$97,$F7,$71,$F0,$5B,$1F,$E2
	.dc.b	$AF,$A6,$FF,$39,$FC,$57,$F5,$69,$FF,$7F,$46,$3D,$1F,$EC,$4D,$8C
	.dc.b	$F8,$9F,$EC,$F9,$CF,$8D,$67,$8B,$7C,$55,$F1,$6D,$E6,$31,$EB,$BE
	.dc.b	$77,$83,$70,$FF,$60,$E3,$D9,$A7,$62,$E0,$9E,$11,$FE,$B4,$9F,$EF
	.dc.b	$51,$8C,$BD,$D5,$E4,$CB,$D1,$6F,$A6,$79,$38,$FA,$6F,$EE,$5E,$4F
	.dc.b	$82,$91,$87,$B1,$FE,$F0,$AE,$2E,$DA,$E8,$E3,$C5,$FC,$89,$C4,$F3
	.dc.b	$CB,$E9,$47,$8F,$0F,$15,$6E,$73,$5B,$D3,$FD,$90,$DB,$F1,$5F,$0C
	.dc.b	$7D,$13,$99,$F0,$C7,$96,$75,$1E,$2C,$3B,$5F,$46,$C7,$43,$9B,$87
	.dc.b	$99,$C6,$BA,$3A,$63,$93,$E1,$63,$F1,$31,$F5,$D7,$3C,$7C,$C7,$9E
	.dc.b	$75,$87,$8F,$0E,$77,$C1,$EF,$87,$58,$C3,$67,$FD,$D9,$5C,$A0,$7A
	.dc.b	$27,$64,$9D,$E3,$F1,$23,$13,$C3,$0F,$24,$E6,$BC,$93,$DC,$F8,$C5
	.dc.b	$3B,$3A,$AF,$BB,$7E,$39,$C9,$D7,$1F,$84,$7C,$D8,$D6,$1F,$12,$33
	.dc.b	$8F,$C0,$BF,$2D,$F2,$25,$F0,$AD,$C7,$8E,$31,$5F,$2B,$0F,$21,$F0
	.dc.b	$27,$07,$8E,$2B,$F9,$33,$ED,$9F,$28,$D1,$4A,$5C,$3F,$4C,$DF,$94
	.dc.b	$FF,$05,$BF,$98,$6B,$F3,$5A,$FE,$B9,$B0,$7F,$C3,$78,$FE,$89,$B7
	.dc.b	$E3,$FA,$6F,$EC,$FE,$47,$E8,$27,$38,$75,$F1,$B0,$7E,$EE,$B7,$1F
	.dc.b	$B2,$9D,$BE,$19,$DB,$FB,$99,$DE,$1E,$19,$CB,$67,$5E,$13,$0C,$E3
	.dc.b	$EF,$BB,$5F,$92,$C3,$9C,$E6,$3C,$53,$A5,$3A,$7D,$46,$89,$87,$F0
	.dc.b	$88,$D4,$7A,$4C,$D7,$59,$D6,$1E,$59,$C9,$D2,$3D,$46,$B0,$F1,$65
	.dc.b	$F4,$71,$F2,$63,$DA,$31,$7C,$8F,$05,$F1,$71,$1A,$8E,$B8,$FD,$7C
	.dc.b	$7C,$51,$8A,$F1,$4F,$9C,$D6,$3E,$2C,$BE,$AE,$1C,$E7,$35,$DA,$72
	.dc.b	$7D,$AC,$7C,$CE,$32,$F4,$E5,$E0,$BE,$D7,$E0,$21,$F4,$23,$4E,$30
	.dc.b	$F0,$3A,$C3,$AD,$FE,$2E,$B7,$3F,$0A,$FC,$B9,$7B,$63,$A6,$5E,$23
	.dc.b	$F0,$7C,$1F,$D6,$79,$FB,$AF,$A5,$7D,$36,$39,$3F,$C3,$D7,$FD,$82
	.dc.b	$72,$FF,$62,$8C,$E1,$FD,$52,$F9,$5F,$1C,$3A,$E1,$C2,$B5,$7C,$31
	.dc.b	$FE,$CA,$DC,$BF,$B4,$F7,$71,$CF,$B6,$37,$0E,$33,$70,$FE,$DB,$F4
	.dc.b	$DF,$ED,$9B,$FF,$2B,$68,$8F,$F5,$B4,$61,$D5,$FF,$A4,$1C,$B7,$93
	.dc.b	$F2,$DD,$DF,$97,$BE,$5D,$B8,$C3,$FD,$ED,$87,$2A,$D3,$C9,$FF,$17
	.dc.b	$FD,$1E,$F1,$60,$EF,$7D,$36,$63,$15,$67,$38,$FF,$8F,$23,$15,$96
	.dc.b	$C7,$F9,$CF,$E7,$9F,$2A,$F5,$FC,$11,$8F,$A6,$B3,$9A,$E0,$A3,$35
	.dc.b	$A6,$CE,$9B,$9F,$FA,$E6,$FA,$4F,$2A,$E0,$FD,$0F,$A1,$FF,$3B,$37
	.dc.b	$B3,$DD,$18,$BF,$F5,$FB,$FC,$B6,$78,$79,$5A,$AD,$B6,$3F,$D0,$78
	.dc.b	$FC,$7F,$2D,$A5,$8C,$36,$7F,$D1,$2D,$C7,$FD,$9B,$5C,$27,$E0,$FA
	.dc.b	$BF,$13,$E6,$7E,$2B,$63,$27,$C6,$8C,$B4,$CD,$7C,$3D,$93,$A8,$C3
	.dc.b	$F7,$7D,$17,$D3,$5F,$91,$C3,$70,$FF,$5A,$36,$FF,$DC,$75,$C2,$7F
	.dc.b	$D7,$15,$F8,$7F,$59,$F1,$17,$CA,$C3,$6B,$F4,$33,$A6,$CE,$9A,$7F
	.dc.b	$AE,$3E,$95,$F4,$BF,$97,$EA,$7D,$37,$C7,$D3,$FD,$E5,$5A,$BF,$F6
	.dc.b	$15,$6B,$8B,$F3,$D4,$AD,$3F,$EF,$F9,$D5,$77,$C3,$B6,$1D,$67,$FA
	.dc.b	$25,$71,$9E,$77,$CA,$7F,$A4,$7C,$3B,$E9,$9E,$F6,$AF,$F5,$5E,$FB
	.dc.b	$AA,$FC,$5E,$1B,$AF,$F0,$49,$D3,$67,$4D,$3F,$84,$FD,$37,$F1,$CC
	.dc.b	$FF,$81,$36,$3F,$C7,$7E,$0A,$1E,$37,$EE,$DD,$6F,$FC,$01,$B3,$BF
	.dc.b	$1F,$55,$B7,$F6,$B7,$FB,$6F,$6E,$BF,$B6,$6E,$1F,$8F,$6E,$3F,$DA
	.dc.b	$FC,$2F,$E0,$31,$8A,$FF,$27,$C6,$1F,$DF,$9D,$4F,$E4,$BF,$4C,$F5
	.dc.b	$7F,$89,$36,$3F,$DF,$DF,$10,$F0,$3F,$C1,$91,$C2,$BF,$8C,$7C,$3B
	.dc.b	$E5,$BF,$8F,$5E,$EB,$31,$8B,$E3,$A3,$4F,$B7,$F2,$9F,$A8,$BF,$E6
	.dc.b	$F8,$F3,$7F,$CC,$F3,$AF,$0B,$9C,$60,$FF,$35,$F4,$BE,$9B,$FD,$65
	.dc.b	$1F,$D0,$3F,$1D,$FE,$19,$97,$78,$DC,$EA,$30,$7F,$34,$9D,$E1,$FC
	.dc.b	$FF,$E4,$5F,$1B,$FC,$E7,$1E,$DE,$A7,$F3,$79,$CE,$5E,$3C,$55,$A8
	.dc.b	$FE,$A3,$3A,$9E,$FC,$E8,$C6,$3F,$E2,$3E,$17,$D3,$FC,$3B,$DD,$6B
	.dc.b	$0F,$C7,$FD,$3F,$F2,$9C,$3F,$AA,$3E,$88,$CF,$9B,$FD,$2E,$31,$7F
	.dc.b	$D4,$1E,$13,$CB,$CE,$D5,$6D,$B1,$FD,$A2,$78,$E6,$78,$DF,$3A,$FE
	.dc.b	$B1,$F4,$CF,$A7,$BF,$41,$86,$E3,$FD,$CA,$74,$D9,$D3,$4D,$4F,$0B
	.dc.b	$F8,$F5,$C1,$B5,$FD,$9F,$E6,$5F,$11,$7E,$1F,$0D,$C7,$FC,$03,$3A
	.dc.b	$6C,$E9,$A6,$A7,$85,$FD,$6A,$E0,$DA,$FE,$ED,$E0,$FE,$76,$3E,$DB
	.dc.b	$F5,$F7,$FF,$03,$39,$6F,$47,$0E,$BE,$2D,$CB,$FC,$2B,$5C,$5F,$E1
	.dc.b	$6F,$16,$C6,$EF,$BB,$FE,$07,$F9,$B6,$B7,$E1,$B0,$DB,$FF,$10,$CE
	.dc.b	$9B,$3A,$69,$A9,$E1,$7F,$96,$AE,$0D,$AF,$F0,$B5,$BD,$1F,$F0,$C7
	.dc.b	$D3,$FE,$F3,$4F,$F1,$34,$E7,$2F,$F1,$8C,$E5,$FE,$4F,$19,$C3,$F4
	.dc.b	$73,$97,$FC,$4B,$19,$AF,$F1,$84,$E5,$7D,$3F,$C7,$AF,$F1,$5F,$D1
	.dc.b	$7C,$77,$F8,$64,$67,$99,$39,$AE,$38,$7F,$5D,$AD,$4F,$F9,$4E,$7F
	.dc.b	$C7,$F1,$D6,$FF,$D9,$DF,$01,$FE,$F8,$FA,$67,$C9,$FF,$B4,$7B,$9F
	.dc.b	$29,$FE,$9B,$9C,$DF,$FB,$4E,$F8,$5E,$EB,$4F,$FB,$5E,$53,$A7,$F8
	.dc.b	$DC,$E9,$E4,$2B,$AB,$FC,$3A,$34,$F9,$8D,$57,$7D,$BA,$46,$6B,$8E
	.dc.b	$1E,$89,$CC,$8F,$19,$C8,$F1,$BA,$73,$3C,$AF,$C3,$1C,$70,$EF,$7D
	.dc.b	$63,$91,$E1,$C3,$C8,$E0,$E0,$6A,$38,$5F,$BA,$32,$F7,$C3,$C5,$7D
	.dc.b	$B0,$F0,$D7,$93,$0E,$B7,$C6,$3D,$67,$CF,$29,$F4,$6B,$0F,$88,$F8
	.dc.b	$47,$C3,$FA,$47,$D3,$FF,$9B,$23,$EC,$46,$3D,$DE,$8D,$8C,$CE,$DE
	.dc.b	$62,$1F,$4F,$FD,$86,$39,$B5,$DC,$6B,$6E,$B3,$99,$F2,$CE,$5F,$9B
	.dc.b	$F4,$FF,$DE,$5F,$BE,$D7,$E8,$C6,$F2,$FC,$4B,$73,$F7,$D6,$E7,$F1
	.dc.b	$AD,$8E,$F7,$E7,$9E,$2D,$8E,$3E,$67,$C8,$9E,$2D,$8E,$38,$FE,$4B
	.dc.b	$0F,$06,$3F,$8F,$C7,$B3,$6F,$83,$F4,$71,$EC,$DB,$F6,$CE,$A1,$5C
	.dc.b	$9F,$9D,$5C,$1B,$5C,$18,$7C,$48,$D5,$F9,$23,$9D,$F5,$BE,$93,$95
	.dc.b	$F1,$1F,$86,$C3,$FC,$AF,$18,$BF,$F0,$11,$E4,$9D,$6F,$F3,$BE,$92
	.dc.b	$CF,$C1,$9A,$EC,$E3,$5C,$DF,$F0,$A9,$EE,$9D,$FE,$32,$FB,$7A,$49
	.dc.b	$F0,$FE,$34,$F9,$CC,$47,$2A,$E0,$F9,$9A,$F6,$9E,$0E,$7C,$0E,$E4
	.dc.b	$7F,$9D,$FD,$F5,$7A,$45,$D7,$F9,$BC,$DD,$7F,$86,$7C,$93,$F0,$AF
	.dc.b	$F3,$ED,$27,$31,$87,$F6,$39,$7F,$93,$F1,$F9,$7F,$1E,$DB,$E3,$B6
	.dc.b	$33,$1F,$2B,$4F,$F4,$04,$60,$87,$CA,$9E,$0F,$BF,$DE,$E7,$5C,$7E
	.dc.b	$1E,$4F,$0C,$FF,$D2,$7F,$9B,$3E,$4F,$EB,$B6,$7E,$FB,$5F,$F1,$EF
	.dc.b	$E4,$8F,$9F,$DD,$49,$DB,$F8,$2F,$11,$5B,$9C,$9F,$E4,$B8,$7D,$53
	.dc.b	$D3,$E0,$D9,$D9,$F8,$6F,$73,$ED,$E7,$FE,$AA,$6F,$17,$F1,$C6,$CF
	.dc.b	$F3,$FF,$C6,$DF,$37,$F9,$29,$CC,$9A,$62,$77,$1F,$9E,$8C,$7A,$9F
	.dc.b	$9A,$9F,$F2,$14,$E7,$AF,$A6,$12,$38,$3F,$A2,$F8,$1F,$0F,$56,$67
	.dc.b	$D5,$37,$3F,$F4,$E6,$6A,$C4,$FF,$A7,$21,$ED,$F3,$6D,$F6,$3F,$67
	.dc.b	$F4,$FF,$CE,$9F,$DA,$FE,$32,$F8,$EF,$D8,$B6,$37,$E4,$62,$64,$70
	.dc.b	$7F,$6F,$5C,$E3,$F6,$B4,$FA,$9E,$64,$8E,$07,$FA,$17,$CD,$4B,$E5
	.dc.b	$FF,$81,$C6,$1B,$19,$9F,$F4,$66,$8F,$CE,$BF,$6B,$1E,$0F,$9B,$F4
	.dc.b	$91,$ED,$5C,$5F,$F4,$EE,$1F,$C9,$BE,$FF,$BF,$90,$69,$F9,$3F,$7B
	.dc.b	$F4,$BC,$7F,$EF,$2F,$A8,$FD,$96,$1F,$C4,$9B,$19,$F2,$7F,$8C,$FD
	.dc.b	$47,$E1,$2B,$51,$FC,$72,$B7,$B7,$F6,$C6,$CF,$F8,$3E,$B8,$5F,$38
	.dc.b	$FE,$53,$3A,$4D,$9D,$D7,$F7,$67,$CE,$DE,$BF,$19,$E4,$F1,$25,$FF
	.dc.b	$34,$AF,$CE,$36,$72,$FF,$2E,$8F,$EE,$AE,$FC,$FF,$69,$EE,$E6,$B1
	.dc.b	$92,$1C,$5F,$81,$F0,$32,$3D,$6F,$C6,$F7,$71,$1F,$2E,$2E,$9F,$C8
	.dc.b	$E3,$E2,$4E,$33,$F9,$15,$67,$27,$AF,$E8,$3E,$8F,$8D,$23,$04,$3D
	.dc.b	$8E,$5F,$74,$7D,$8E,$19,$1F,$15,$FE,$A7,$F0,$5F,$E1,$8C,$BC,$92
	.dc.b	$D7,$F1,$0E,$C8,$7F,$56,$9E,$19,$7F,$8A,$BF,$19,$7C,$C7,$F5,$E6
	.dc.b	$D7,$F6,$66,$BC,$7F,$12,$D4,$C8,$D3,$E4,$8C,$47,$F6,$49,$E3,$8F
	.dc.b	$99,$B7,$FE,$3A,$C1,$7E,$CF,$4E,$61,$FD,$DA,$7F,$57,$87,$B9,$7E
	.dc.b	$1B,$A9,$F6,$55,$AD,$4F,$C6,$F9,$2F,$8F,$F0,$B2,$FA,$46,$5E,$D1
	.dc.b	$97,$E3,$47,$C3,$23,$F0,$DF,$A3,$E8,$E2,$3D,$D1,$6F,$CD,$37,$3F
	.dc.b	$9D,$36,$BF,$CF,$59,$7F,$62,$7E,$A7,$CB,$7D,$CF,$9D,$91,$F0,$63
	.dc.b	$EF,$98,$7E,$D4,$61,$FF,$0D,$FE,$92,$F9,$5F,$F0,$BB,$FD,$46,$3F
	.dc.b	$C2,$AD,$CF,$F0,$EE,$E3,$8C,$F0,$F4,$24,$60,$FF,$13,$FD,$3B,$9B
	.dc.b	$FD,$5F,$EE,$7F,$9C,$A3,$13,$FE,$8D,$8C,$4E,$A3,$FD,$31,$3F,$DA
	.dc.b	$5F,$F5,$5C,$E9,$E0,$6A,$78,$29,$EA,$F0,$39,$F4,$6A,$33,$F0,$5F
	.dc.b	$E3,$37,$FC,$35,$3D,$0E,$5E,$E6,$EE,$46,$67,$6F,$53,$66,$63,$B7
	.dc.b	$C4,$F9,$3E,$9D,$E4,$FF,$6E,$C7,$FC,$85,$18,$F3,$FA,$57,$0C,$3C
	.dc.b	$53,$97,$B3,$6B,$D1,$3D,$6F,$A6,$9F,$E9,$EA,$D3,$63,$37,$F3,$2B
	.dc.b	$86,$1D,$A7,$32,$BE,$2E,$CD,$4C,$7C,$73,$C3,$25,$F4,$9C,$D6,$F1
	.dc.b	$F3,$C5,$C7,$C9,$36,$3B,$1C,$09,$3F,$1B,$0E,$D7,$D6,$31,$5C,$6B
	.dc.b	$F0,$13,$9B,$EB,$1C,$9F,$C8,$46,$63,$FD,$C9,$8F,$B9,$F8,$18,$76
	.dc.b	$7A,$46,$4D,$63,$EE,$8C,$CF,$C8,$9C,$D7,$59,$E7,$87,$96,$F8,$47
	.dc.b	$8C,$F0,$DF,$13,$C0,$E4,$E2,$58,$E3,$8F,$C0,$8D,$61,$E0,$7C,$77
	.dc.b	$DF,$0F,$14,$FA,$70,$EB,$7C,$49,$8F,$C6,$7E,$AE,$1F,$44,$FF,$6E
	.dc.b	$D7,$E4,$5B,$5F,$E0,$37,$E6,$9A,$EF,$7B,$FF,$EF,$3A,$46,$27,$FD
	.dc.b	$CE,$7E,$35,$C3,$C6,$78,$46,$EF,$F4,$AD,$C3,$FD,$F3,$F4,$FF,$C2
	.dc.b	$FA,$3F,$4B,$39,$C3,$ED,$B5,$ED,$87,$EE,$EB,$84,$7E,$AA,$75,$8F
	.dc.b	$9A,$B9,$9B,$DB,$FB,$C6,$5E,$28,$CD,$F1,$75,$8E,$27,$53,$93,$F8
	.dc.b	$26,$3E,$08,$C9,$32,$FD,$B6,$5E,$79,$D4,$63,$0E,$8D,$BE,$78,$F9
	.dc.b	$23,$18,$73,$9F,$83,$5E,$47,$C5,$19,$7E,$19,$CB,$1E,$AF,$03,$75
	.dc.b	$F3,$AF,$B4,$66,$B9,$65,$F1,$E3,$91,$C0,$D5,$79,$0E,$57,$DA,$3C
	.dc.b	$35,$E3,$C7,$DF,$1C,$08,$6A,$B8,$BD,$4C,$1A,$8E,$2E,$09,$7E,$5A
	.dc.b	$FF,$77,$4E,$A3,$15,$E5,$AE,$58,$7B,$9E,$39,$7E,$10,$FA,$91,$9A
	.dc.b	$E3,$21,$7D,$6B,$78,$FB,$E3,$19,$7D,$3C,$BE,$06,$9D,$E5,$F8,$75
	.dc.b	$8C,$36,$FF,$30,$DC,$7F,$16,$DB,$F6,$D7,$3F,$0F,$1A,$FE,$9A,$2D
	.dc.b	$F9,$F1,$56,$3D,$57,$C3,$7F,$AC,$73,$F6,$74,$7F,$95,$E4,$FF,$B2
	.dc.b	$A3,$27,$C1,$8E,$3E,$B7,$F9,$76,$78,$1B,$C7,$8D,$6E,$FB,$46,$05
	.dc.b	$F6,$E3,$E5,$5B,$C2,$D6,$B0,$E8,$79,$70,$FF,$5F,$57,$92,$79,$C7
	.dc.b	$FA,$C6,$31,$87,$BB,$D5,$E6,$FC,$0F,$4E,$D7,$2B,$F2,$4E,$6F,$8C
	.dc.b	$E7,$0F,$05,$F9,$2D,$F8,$46,$6F,$F0,$91,$8C,$7D,$3C,$4F,$C2,$DE
	.dc.b	$77,$1C,$38,$5F,$E2,$DA,$F1,$C3,$D9,$7C,$6B,$86,$12,$7E,$05,$6C
	.dc.b	$F9,$95,$B8,$CD,$F6,$D3,$EF,$FD,$3F,$C0,$FA,$7F,$82,$DA,$F7,$B7
	.dc.b	$0F,$87,$B2,$31,$B7,$D1,$F6,$FE,$49,$F2,$BD,$BB,$87,$C8,$8B,$87
	.dc.b	$29,$B5,$FE,$90,$CD,$F1,$DF,$0C,$FB,$FE,$C3,$E8,$9E,$AE,$73,$F5
	.dc.b	$D5,$CB,$FD,$C3,$5A,$9C,$92,$3F,$1C,$7D,$69,$EB,$3B,$8C,$E1,$E2
	.dc.b	$AF,$BF,$EE,$7E,$10,$FC,$2F,$D0,$D8,$C9,$2B,$E2,$C6,$5F,$8F,$EC
	.dc.b	$FE,$33,$1F,$C6,$B6,$BF,$16,$D9,$DC,$7E,$3A,$7F,$09,$E1,$C8,$FC
	.dc.b	$0F,$3B,$EC,$F1,$7E,$8A,$BB,$FF,$B3,$79,$BF,$1D,$58,$C3,$F2,$7A
	.dc.b	$FF,$94,$1F,$C1,$DF,$E5,$31,$E9,$18,$3F,$3F,$F8,$6F,$E9,$0F,$D2
	.dc.b	$FC,$35,$8C,$93,$3F,$C2,$C6,$5F,$C3,$FD,$97,$EB,$3F,$00,$FD,$6B
	.dc.b	$6B,$F5,$6D,$9D,$C7,$EB,$A7,$F5,$9F,$4E,$F4,$69,$F6,$3E,$9E,$D7
	.dc.b	$E8,$27,$4D,$9D,$34,$97,$F8,$29,$DB,$F8,$8A,$E6,$FF,$13,$E7,$7C
	.dc.b	$87,$F9,$6A,$FF,$86,$46,$3C,$6F,$E2,$7F,$14,$F5,$9B,$7E,$63,$0E
	.dc.b	$18,$FE,$42,$BF,$C4,$38,$FE,$D9,$B1,$DF,$AF,$F7,$F1,$8E,$65,$7F
	.dc.b	$4D,$CF,$F4,$3E,$17,$E9,$2F,$F8,$3C,$BF,$2B,$EF,$7C,$31,$61,$C2
	.dc.b	$BF,$82,$F4,$3E,$3B,$F8,$C3,$63,$3B,$62,$72,$C8,$FE,$1B,$F8,$03
	.dc.b	$E0,$B1,$8F,$EB,$23,$13,$C8,$DC,$EB,$2F,$25,$3C,$E7,$37,$F3,$2C
	.dc.b	$3D,$38,$7F,$28,$9C,$BF,$CE,$23,$35,$FD,$BE,$72,$FE,$F1,$B7,$FD
	.dc.b	$5E,$7F,$78,$F8,$23,$35,$D6,$FF,$9C,$46,$2F,$FA,$8D,$F0,$AD,$1F
	.dc.b	$CB,$3C,$77,$B7,$FD,$4A,$1F,$8B,$3D,$CF,$E8,$55,$A9,$FE,$A4,$6A
	.dc.b	$33,$52,$33,$4F,$7B,$18,$FF,$16,$C3,$93,$6F,$C1,$5B,$BF,$24,$7F
	.dc.b	$51,$FB,$F7,$BD,$F9,$4F,$8F,$7C,$96,$63,$0D,$BF,$E8,$F5,$C1,$FE
	.dc.b	$D1,$D2,$F9,$FF,$0F,$17,$D8,$AD,$69,$FC,$A9,$B1,$9C,$7F,$47,$5C
	.dc.b	$30,$FE,$C3,$5A,$6F,$E2,$FE,$A3,$D0,$57,$C6,$E1,$F8,$BF,$0B,$71
	.dc.b	$F0,$4D,$C7,$F3,$D5,$F4,$AB,$0F,$F7,$0E,$9F,$DE,$CF,$F2,$3D,$7F
	.dc.b	$C6,$D3,$FD,$9A,$72,$FF,$8E,$7A,$53,$AC,$7F,$A3,$D6,$9F,$D9,$78
	.dc.b	$5F,$C0,$3E,$7D,$F4,$BF,$E0,$1A,$D5,$7E,$C2,$B0,$6E,$6E,$3E,$1E
	.dc.b	$2E,$07,$F8,$13,$DF,$FE,$A1,$7E,$08,$CC,$EC,$FF,$00,$CE,$B4,$FD
	.dc.b	$5E,$3F,$3E,$FF,$83,$7E,$0F,$E3,$3C,$9B,$DC,$3E,$04,$59,$F6,$CD
	.dc.b	$BF,$DA,$C7,$EB,$47,$CE,$3E,$73,$E6,$7C,$EF,$F0,$D8,$C4,$FF,$87
	.dc.b	$8F,$DE,$1C,$27,$78,$FF,$99,$67,$73,$FB,$D3,$47,$F1,$DA,$E9,$5F
	.dc.b	$EC,$4E,$DE,$0F,$29,$D9,$C6,$B9,$E6,$BE,$33,$E3,$BE,$2F,$5C,$2B
	.dc.b	$C7,$0E,$98,$79,$23,$E0,$3D,$2F,$9C,$6C,$E4,$78,$8D,$CF,$23,$D0
	.dc.b	$F9,$B0,$F2,$5F,$82,$36,$72,$BF,$03,$F2,$1F,$04,$6E,$31,$5D,$5F
	.dc.b	$61,$FE,$E8,$3E,$B1,$70,$C3,$83,$FC,$A3,$E2,$BE,$3F,$FC,$FB,$1F
	.dc.b	$EF,$08,$C7,$B3,$CD,$B1,$99,$DB,$F7,$61,$F2,$1F,$E8,$28,$E4,$D7
	.dc.b	$4D,$7C,$2D,$7F,$D4,$D1,$CB,$5F,$A7,$F2,$6F,$9D,$FF,$33,$EB,$F8
	.dc.b	$68,$C5,$FD,$8B,$E0,$FE,$09,$B3,$BA,$E3,$A7,$F9,$C2,$F8,$DE,$F1
	.dc.b	$F4,$5F,$E4,$B0,$E7,$EA,$71,$AF,$F5,$D3,$5F,$F1,$76,$1E,$AC,$B2
	.dc.b	$D8,$FC,$9E,$3D,$B0,$E0,$75,$C7,$BD,$6E,$32,$F9,$72,$F3,$DF,$C4
	.dc.b	$AD,$B7,$0F,$8B,$3A,$86,$1D,$EF,$A1,$23,$FD,$A5,$19,$6C,$F1,$CB
	.dc.b	$E3,$D7,$C5,$F8,$4F,$F4,$AE,$1F,$DD,$30,$46,$37,$D6,$3F,$E8,$C9
	.dc.b	$FF,$4B,$B6,$7F,$1F,$1C,$23,$F2,$8D,$7F,$82,$1E,$88,$FC,$AF,$CE
	.dc.b	$BD,$DA,$49,$FC,$E4,$65,$B1,$E3,$7F,$DD,$DF,$3B,$FD,$BD,$FF,$4F
	.dc.b	$B5,$FF,$3E,$CF,$16,$FB,$5C,$54,$23,$0D,$8F,$EC,$3F,$40,$F8,$4C
	.dc.b	$D7,$A0,$93,$E6,$6B,$FE,$AE,$F8,$CF,$F0,$EE,$BF,$E7,$16,$EB,$FE
	.dc.b	$C2,$C7,$B5,$7F,$A4,$BC,$7F,$96,$DF,$84,$F3,$46,$30,$FF,$13,$E9
	.dc.b	$FE,$10,$AF,$F5,$EC,$F0,$7F,$98,$35,$FF,$67,$B5,$1F,$E0,$2A,$FE
	.dc.b	$0B,$1D,$70,$FF,$6A,$D7,$F9,$0E,$BE,$5F,$A3,$E2,$BF,$91,$F3,$DF
	.dc.b	$E5,$C9,$FF,$66,$B5,$FF,$55,$B7,$1F,$F2,$87,$94,$F9,$E7,$C7,$F2
	.dc.b	$F6,$3F,$8A,$DF,$1C,$3F,$A4,$35,$FF,$73,$B5,$F4,$BD,$74,$F5,$E1
	.dc.b	$FD,$DB,$E4,$3F,$08,$DA,$F2,$DF,$DF,$F6,$3F,$05,$9F,$F9,$41,$B8
	.dc.b	$7F,$BD,$1B,$5F,$1B,$E7,$5F,$17,$C7,$9F,$F9,$5D,$7E,$2E,$BF,$AF
	.dc.b	$35,$E5,$1A,$3F,$1D,$4F,$9E,$A7,$39,$FF,$7A,$D7,$E1,$1F,$AA,$D7
	.dc.b	$FD,$A4,$31,$FF,$3D,$7A,$9E,$4E,$8F,$66,$7E,$42,$57,$F4,$3B,$FE
	.dc.b	$F2,$FB,$63,$F5,$15,$EC,$F8,$9B,$A1,$E7,$5F,$86,$C7,$FC,$EB,$F3
	.dc.b	$8F,$7B,$F5,$0D,$9D,$63,$F8,$C5,$19,$7F,$DD,$BF,$3D,$FD,$3D,$D9
	.dc.b	$F9,$28,$D3,$FE,$07,$6B,$CD,$AE,$EB,$FA,$DC,$3E,$2A,$92,$7F,$B1
	.dc.b	$46,$5F,$E1,$FE,$57,$93,$E3,$BF,$B3,$36,$33,$29,$CB,$63,$9F,$C4
	.dc.b	$78,$08,$E0,$EB,$E1,$FF,$A8,$5B,$F0,$1D,$A3,$11,$FE,$23,$C3,$F9
	.dc.b	$BC,$FF,$74,$F9,$CF,$E0,$3B,$7F,$A8,$9B,$B7,$CB,$6D,$FE,$E3,$E7
	.dc.b	$3F,$26,$FC,$A1,$1F,$C2,$9F,$F1,$83,$5E,$5E,$27,$72,$3C,$65,$F0
	.dc.b	$DF,$C8,$DB,$1D,$6B,$C2,$A3,$F8,$33,$E3,$E5,$E0,$71,$F0,$24,$75
	.dc.b	$8C,$BD,$E3,$B7,$97,$23,$C4,$F8,$7C,$EC,$47,$E7,$F3,$FC,$66,$1F
	.dc.b	$E0,$E8,$D9,$E9,$9F,$E7,$58,$BE,$52,$92,$7D,$DF,$25,$5F,$EA,$33
	.dc.b	$B7,$FC,$50,$D7,$33,$F5,$4F,$89,$F2,$92,$3E,$24,$7C,$7F,$5F,$E4
	.dc.b	$78,$B2,$7E,$64,$65,$F9,$D1,$97,$D8,$7D,$B8,$FA,$1C,$52,$3E,$A3
	.dc.b	$FD,$57,$E1,$3F,$CD,$5C,$7F,$EE,$6D,$7F,$16,$D9,$E1,$5D,$0F,$EC
	.dc.b	$05,$F9,$B7,$CE,$7F,$5F,$6C,$7E,$25,$AF,$1C,$1F,$03,$BA,$FF,$24
	.dc.b	$D6,$AB,$FC,$F8,$D7,$F8,$03,$5F,$E9,$AF,$4C,$3F,$DF,$B6,$F6,$FD
	.dc.b	$87,$5A,$FC,$D4,$3E,$73,$E0,$12,$33,$F0,$F5,$F3,$45,$BE,$B3,$F1
	.dc.b	$FE,$2B,$E4,$7C,$E4,$BE,$B1,$97,$BC,$65,$F8,$11,$F1,$48,$FC,$57
	.dc.b	$E9,$7C,$06,$23,$FC,$33,$9A,$7E,$29,$23,$FC,$E5,$19,$7D,$BE,$8F
	.dc.b	$DB,$39,$63,$F1,$3E,$0B,$EF,$7C,$44,$AF,$87,$1F,$81,$30,$FD,$58
	.dc.b	$C3,$FC,$07,$E7,$1F,$39,$FE,$1B,$5E,$9D,$31,$12,$78,$46,$5E,$51
	.dc.b	$FE,$23,$F9,$C7,$3F,$F3,$BE,$EF,$F5,$D4,$62,$BF,$DC,$47,$F8,$29
	.dc.b	$C3,$FE,$C6,$6C,$70,$9F,$E8,$9F,$07,$FE,$41,$97,$B7,$8F,$77,$FA
	.dc.b	$44,$63,$0F,$F0,$AD,$6E,$F5,$1D,$DA,$A7,$82,$9E,$D7,$A9,$F2,$C7
	.dc.b	$39,$EB,$7E,$0A,$D1,$BB,$EB,$F3,$8F,$2B,$F9,$7E,$1F,$E5,$98,$C7
	.dc.b	$A3,$CB,$1F,$F1,$5C,$61,$F1,$63,$EC,$F9,$1E,$78,$72,$8F,$EC,$D5
	.dc.b	$BC,$3A,$4E,$6F,$E6,$5F,$0C,$3B,$42,$F8,$61,$CA,$F8,$4F,$C3,$BE
	.dc.b	$0F,$88,$F2,$C6,$AB,$8C,$76,$3A,$B8,$7B,$46,$CE,$B9,$AB,$AE,$3E
	.dc.b	$8C,$78,$D6,$A7,$38,$F9,$67,$27,$AA,$F9,$61,$83,$CC,$78,$23,$18
	.dc.b	$7E,$1E,$F9,$CF,$AA,$73,$7F,$36,$29,$93,$9E,$3E,$BC,$3B,$C6,$31
	.dc.b	$F6,$5F,$08,$F9,$05,$AF,$84,$71,$C3,$C3,$1E,$53,$94,$FA,$E3,$33
	.dc.b	$C2,$BE,$15,$73,$7B,$BE,$EB,$F1,$E3,$F2,$63,$99,$1F,$80,$F1,$C3
	.dc.b	$CD,$3C,$31,$F9,$93,$B6,$D7,$08,$C1,$F0,$DE,$C6,$4B,$1C,$09,$1F
	.dc.b	$83,$3C,$67,$F8,$EC,$F8,$05,$C3,$F3,$CD,$C3,$FC,$26,$E7,$C1,$7C
	.dc.b	$5F,$FA,$FA,$3F,$31,$18,$7F,$D8,$35,$FD,$DE,$74,$D8,$CC,$EF,$0F
	.dc.b	$D8,$37,$0F,$F2,$77,$CE,$7F,$16,$DF,$F3,$33,$9B,$FD,$9B,$5F,$0E
	.dc.b	$3F,$AA,$AE,$11,$F8,$69,$D6,$3E,$1C,$7B,$DE,$F6,$63,$DA,$26,$1C
	.dc.b	$5D,$46,$2B,$8C,$E4,$FC,$EE,$3E,$08,$C9,$32,$FE,$07,$97,$9E,$75
	.dc.b	$18,$C3,$A3,$6F,$9E,$3E,$47,$B6,$1C,$DF,$83,$5E,$47,$35,$C6,$75
	.dc.b	$0B,$EE,$F0,$36,$58,$FF,$7C,$46,$A3,$E5,$9A,$2C,$6A,$FC,$8F,$6C
	.dc.b	$BE,$65,$F8,$4F,$B3,$3A,$71,$5D,$DA,$F9,$63,$4E,$48,$78,$23,$9E
	.dc.b	$5F,$1A,$33,$5E,$07,$FD,$BF,$87,$A6,$31,$1C,$4E,$71,$9C,$BF,$06
	.dc.b	$4C,$BE,$56,$5E,$B8,$C5,$7D,$2C,$3C,$0F,$32,$5F,$3B,$CE,$1E,$08
	.dc.b	$C4,$FE,$13,$6F,$6B,$F4,$FC,$4B,$3F,$CB,$F0,$FC,$C4,$62,$B2,$D8
	.dc.b	$C9,$F9,$D8,$CB,$4C,$D7,$77,$FB,$CB,$F8,$18,$FC,$54,$7E,$FE,$BC
	.dc.b	$13,$F8,$D3,$E4,$CE,$90,$F5,$5F,$0E,$F5,$FF,$29,$D3,$FE,$D6,$3F
	.dc.b	$DD,$38,$FF,$B0,$2B,$73,$A8,$E7,$1F,$ED,$DE,$3C,$61,$D7,$0E,$95
	.dc.b	$C7,$0E,$73,$FE,$6A,$BE,$73,$9B,$E3,$1E,$4F,$56,$E1,$FE,$BD,$AF
	.dc.b	$13,$FE,$DB,$6F,$A9,$A9,$DB,$63,$27,$FA,$BF,$E5,$F8,$63,$FE,$7A
	.dc.b	$87,$E5,$3F,$E4,$37,$57,$EA,$73,$E0,$7B,$29,$7C,$9F,$E5,$CD,$7F
	.dc.b	$DD,$4D,$7E,$7B,$5E,$35,$CE,$3C,$7A,$7F,$03,$E2,$7C,$C3,$D1,$D5
	.dc.b	$FC,$2F,$9C,$F8,$7F,$3B,$74,$F8,$6D,$CF,$E6,$6E,$8C,$67,$FC,$3F
	.dc.b	$83,$EF,$C7,$CC,$F7,$E6,$1F,$33,$DD,$B8,$F2,$9B,$87,$1C,$55,$8B
	.dc.b	$F7,$FC,$DF,$E0,$37,$F4,$F6,$7B,$63,$1C,$08,$C5,$7F,$4D,$E1,$57
	.dc.b	$E5,$F4,$E7,$39,$8C,$47,$CA,$78,$BF,$EB,$2F,$69,$F3,$AF,$67,$3B
	.dc.b	$79,$FD,$EF,$C4,$9F,$8A,$F9,$DB,$97,$E2,$A2,$E3,$F2,$26,$D7,$BF
	.dc.b	$DD,$FC,$8E,$7F,$92,$6E,$1F,$92,$6D,$FD,$06,$D7,$EC,$74,$FF,$29
	.dc.b	$57,$EC,$23,$EF,$6D,$FE,$F8,$8F,$CA,$78,$9F,$1B,$DD,$FC,$23,$F0
	.dc.b	$31,$FC,$7B,$70,$E7,$1F,$DF,$4C,$D7,$CB,$9E,$2F,$73,$F7,$FC,$3F
	.dc.b	$8C,$F8,$FC,$D7,$E1,$BE,$77,$F5,$A7,$EB,$BE,$16,$DF,$EB,$A2,$CF
	.dc.b	$E3,$26,$E9,$F8,$3F,$9D,$FD,$AF,$CD,$FE,$D9,$B8,$7E,$D9,$B7,$F9
	.dc.b	$46,$D7,$F1,$5F,$9D,$7A,$5F,$C2,$21,$F5,$FF,$A4,$C3,$F8,$BE,$0B
	.dc.b	$DC,$7B,$61,$5C,$3C,$0F,$DB,$56,$A7,$E3,$35,$FF,$12,$35,$EA,$D7
	.dc.b	$F9,$2B,$C6,$7C,$D8,$7F,$1C,$E9,$FC,$7F,$17,$68,$C7,$C9,$7F,$30
	.dc.b	$C3,$CF,$48,$C4,$7F,$30,$8C,$79,$1F,$9C,$6E,$7F,$E2,$FF,$8B,$79
	.dc.b	$BF,$81,$6C,$66,$30,$71,$EC,$FE,$1D,$5F,$A1,$DB,$04,$AF,$F1,$D7
	.dc.b	$2B,$D6,$FC,$63,$63,$31,$83,$3E,$3C,$8E,$2D,$D3,$F2,$15,$FC,$9D
	.dc.b	$FA,$33,$F5,$A3,$1C,$1F,$D4,$7E,$75,$F0,$3F,$CA,$67,$35,$FE,$CC
	.dc.b	$E6,$FE,$7F,$5F,$D6,$2F,$FA,$C7,$A2,$9C,$F0,$FF,$31,$8C,$5F,$F9
	.dc.b	$55,$B9,$7E,$8D,$AF,$F9,$85,$AF,$78,$FC,$43,$FE,$CF,$AF,$0E,$FD
	.dc.b	$5F,$E7,$DE,$43,$DA,$FE,$85,$F3,$AF,$7F,$31,$86,$E3,$FE,$67,$AE
	.dc.b	$0D,$9F,$DD,$46,$31,$FD,$C3,$74,$FE,$EC,$DA,$FE,$EA,$30,$7B,$DD
	.dc.b	$AF,$E7,$9F,$DE,$BE,$71,$EE,$FF,$70,$E8,$7C,$E3,$E4,$FF,$AC,$56
	.dc.b	$B2,$FD,$45,$6A,$73,$7F,$C3,$EB,$58,$7F,$0D,$6B,$FE,$8B,$6B,$FB
	.dc.b	$47,$7A,$7F,$83,$DB,$87,$F8,$3B,$07,$CE,$FB,$6F,$D1,$87,$93,$E6
	.dc.b	$FE,$69,$6F,$F7,$87,$C1,$71,$3E,$37,$CC,$48,$FF,$4B,$60,$8D,$E3
	.dc.b	$FE,$24,$D5,$F0,$F8,$BF,$DA,$F7,$FF,$57,$DB,$E7,$E5,$F3,$07,$6F
	.dc.b	$F0,$1A,$FE,$53,$3A,$C7,$FC,$59,$3F,$C1,$EB,$77,$FC,$DE,$30,$47
	.dc.b	$FC,$10,$F3,$86,$CF,$9D,$7C,$EE,$AF,$7C,$9F,$DA,$F8,$5F,$59,$8D
	.dc.b	$7F,$90,$76,$7C,$C8,$C6,$7F,$D0,$F4,$FC,$73,$FE,$C1,$6B,$FC,$53
	.dc.b	$0F,$37,$C4,$CC,$3F,$D7,$D4,$8C,$57,$37,$D2,$7A,$B8,$AD,$7E,$AA
	.dc.b	$7F,$C9,$7D,$7F,$E4,$CF,$9D,$7C,$E7,$F2,$2F,$9E,$FF,$4F,$46,$2B
	.dc.b	$F6,$81,$C3,$F4,$23,$73,$FB,$B8,$D3,$CE,$FB,$39,$AE,$39,$2F,$8D
	.dc.b	$F6,$9C,$C6,$23,$FA,$5E,$1C,$9F,$E6,$91,$97,$D6,$78,$0C,$E1,$B7
	.dc.b	$89,$93,$D0,$E5,$E3,$E6,$F8,$23,$FC,$DE,$FB,$A3,$38,$FA,$D4,$65
	.dc.b	$5F,$7C,$7D,$97,$D6,$72,$78,$63,$2F,$8C,$F4,$E3,$E1,$70,$70,$70
	.dc.b	$77,$AF,$F7,$F4,$7C,$03,$F9,$61,$87,$FC,$1C,$FF,$5B,$F9,$2F,$F4
	.dc.b	$3B,$FE,$11,$F2,$11,$88,$FE,$EE,$6F,$D9,$E3,$3A,$6F,$C9,$FF,$78
	.dc.b	$9E,$4D,$78,$CA,$31,$1F,$E0,$28,$C7,$47,$23,$EA,$7C,$A3,$E7,$BF
	.dc.b	$C0,$F9,$FE,$32,$31,$7F,$66,$F8,$BF,$65,$B3,$BA,$E3,$BF,$F8,$42
	.dc.b	$F8,$DE,$F1,$F4,$5F,$E6,$30,$E3,$7B,$C3,$8D,$FF,$AB,$B0,$E3,$7B
	.dc.b	$C3,$8D,$FA,$DB,$5B,$BE,$2A,$F8,$D6,$F1,$F5,$5F,$0A,$D5,$F0,$BD
	.dc.b	$C7,$FA,$FB,$1F,$2B,$6B,$84,$EA,$38,$9D,$2E,$3F,$9D,$8D,$1B,$AE
	.dc.b	$71,$E5,$F6,$FF,$C6,$B5,$FE,$CF,$AF,$F3,$F4,$FF,$6E,$6B,$FC,$B5
	.dc.b	$FF,$4E,$39,$9E,$3B,$FC,$FF,$3A,$CF,$C9,$6B,$DC,$FE,$F3,$F3,$AF
	.dc.b	$7B,$3F,$3B,$FD,$D6,$7F,$BE,$B5,$FE,$79,$3B,$78,$CF,$0F,$22,$46
	.dc.b	$0E,$BE,$D7,$82,$8D,$3C,$BE,$0B,$E7,$BF,$DD,$15,$C0,$FE,$6B,$F1
	.dc.b	$BF,$EF,$1D,$FF,$D7,$8D,$D7,$FC,$B7,$8F,$4A,$E0,$7F,$7C,$EF,$56
	.dc.b	$E1,$18,$AF,$14,$7F,$86,$BC,$9F,$96,$FF,$59,$6E,$1F,$F0,$05,$B3
	.dc.b	$FE,$6A,$DB,$FC,$39,$5C,$7E,$13,$E4,$FA,$72,$30,$43,$FB,$24,$FF
	.dc.b	$92,$5F,$9B,$1F,$52,$FA,$C0,$7C,$EB,$DA,$FB,$2D,$8F,$F1,$8B,$5F
	.dc.b	$F0,$DE,$EF,$9E,$7C,$C7,$0B,$F9,$FE,$5B,$E4,$A9,$9C,$39,$CE,$1F
	.dc.b	$F0,$EC,$ED,$C9,$FE,$40,$F5,$3F,$04,$FF,$7F,$6F,$03,$83,$FD,$1F
	.dc.b	$18,$7F,$D2,$3F,$3C,$F8,$CF,$C9,$CE,$6F,$E0,$4B,$E7,$BF,$2E,$D8
	.dc.b	$F0,$75,$70,$33,$32,$32,$79,$70,$EA,$47,$FD,$3D,$EB,$78,$7E,$3B
	.dc.b	$11,$FE,$42,$8F,$27,$C5,$7E,$A1,$B1,$CC,$86,$F3,$FF,$7C,$BD,$1A
	.dc.b	$FC,$3B,$FD,$7F,$CE,$BD,$FA,$49,$FF,$34,$FC,$F7,$F4,$07,$FC,$D7
	.dc.b	$3B,$6F,$C5,$F2,$38,$3F,$E2,$D8,$47,$6F,$85,$93,$C0,$FF,$5E,$79
	.dc.b	$CA,$7C,$87,$93,$E2,$25,$6A,$72,$F8,$A3,$F6,$BF,$41,$F9,$4E,$B3
	.dc.b	$F9,$69,$E9,$3F,$C3,$B0,$FE,$5D,$F0,$9F,$BC,$FB,$97,$C7,$53,$5E
	.dc.b	$67,$FA,$6B,$0F,$F4,$4F,$CF,$FE,$E5,$FF,$24,$37,$AF,$F1,$51,$FA
	.dc.b	$38,$C7,$1F,$FB,$6B,$E7,$FF,$15,$3C,$BC,$BF,$E3,$B0,$F3,$69,$FC
	.dc.b	$82,$F8,$D7,$3A,$CB,$FE,$9A,$AE,$0F,$33,$1E,$47,$4E,$F9,$18,$21
	.dc.b	$FE,$5F,$94,$6A,$31,$3E,$4F,$42,$47,$77,$CF,$EA,$E2,$39,$45,$D3
	.dc.b	$E0,$C3,$0F,$C3,$3F,$1B,$E2,$BD,$84,$7D,$5F,$41,$5F,$27,$CB,$FC
	.dc.b	$43,$17,$D3,$5F,$F7,$2C,$FC,$19,$D3,$F5,$2F,$E5,$FC,$CF,$CC,$F2
	.dc.b	$64,$FC,$78,$CB,$F2,$63,$2F,$C3,$8F,$A7,$CB,$23,$E7,$3F,$D4,$BD
	.dc.b	$3F,$F0,$EE,$5E,$ED,$3F,$09,$5F,$2E,$B8,$3F,$E8,$BA,$DF,$1F,$F2
	.dc.b	$58,$7C,$EB,$E8,$3F,$AE,$B6,$3F,$DA,$AD,$79,$7C,$3E,$A7,$C5,$F0
	.dc.b	$32,$77,$19,$7F,$DB,$B8,$7F,$9E,$DB,$39,$7C,$9F,$41,$EB,$23,$C3
	.dc.b	$E3,$2B,$A3,$9C,$7C,$18,$F8,$5F,$21,$F0,$FE,$82,$57,$82,$32,$F8
	.dc.b	$A3,$2F,$C3,$8F,$79,$1F,$7B,$F3,$7E,$1B,$11,$F2,$7D,$7F,$A0,$63
	.dc.b	$87,$F1,$BB,$2F,$35,$F0,$2B,$5B,$FC,$FD,$5F,$13,$F6,$7E,$3A,$5F
	.dc.b	$C6,$8F,$B8,$61,$FB,$11,$87,$FA,$E6,$5F,$9C,$FA,$07,$D0,$7F,$85
	.dc.b	$1F,$E9,$D3,$FE,$C9,$6C,$EE,$74,$6F,$D8,$95,$D3,$2F,$C2,$5F,$48
	.dc.b	$C1,$FD,$AF,$E8,$1D,$1F,$E9,$8F,$82,$FC,$54,$7F,$81,$A7,$FD,$63
	.dc.b	$18,$F8,$0D,$47,$1A,$AA,$B9,$BC,$0E,$D3,$A8,$E9,$1F,$E9,$2A,$EF
	.dc.b	$59,$35,$52,$79,$1D,$63,$0F,$13,$A9,$AA,$EB,$F1,$1F,$E4,$47,$C7
	.dc.b	$1B,$1F,$40,$F3,$3F,$93,$D7,$F9,$2A,$31,$EA,$F8,$1B,$5C,$6D,$F2
	.dc.b	$9F,$02,$73,$7C,$70,$E7,$9F,$EC,$E7,$2F,$CC,$AE,$13,$AC,$3F,$C3
	.dc.b	$95,$A6,$C6,$5F,$99,$3C,$31,$F2,$4E,$6B,$85,$68,$CC,$C7,$E0,$46
	.dc.b	$B1,$F4,$C6,$71,$F4,$55,$C7,$CD,$34,$F1,$C5,$7B,$1D,$30,$F1,$12
	.dc.b	$3E,$96,$3E,$79,$D4,$71,$C3,$C3,$18,$26,$1E,$27,$E7,$57,$18,$C1
	.dc.b	$30,$F2,$CF,$07,$E7,$61,$E1,$8E,$27,$9F,$1F,$8A,$7C,$1B,$E9,$5C
	.dc.b	$31,$F2,$46,$2B,$DE,$75,$3F,$4C,$E5,$FC,$4D,$F7,$C3,$F1,$D3,$9C
	.dc.b	$3B,$CF,$D9,$C7,$DF,$5E,$9B,$EA,$D7,$E5,$B6,$B8,$E1,$F2,$A7,$D5
	.dc.b	$7D,$F1,$F8,$37,$F1,$E7,$FD,$C7,$7F,$96,$6D,$7F,$80,$9B,$7F,$EE
	.dc.b	$E9,$78,$3F,$E8,$09,$FF,$75,$9F,$A4,$70,$F4,$6B,$F9,$B6,$BF,$98
	.dc.b	$79,$5F,$EA,$5B,$87,$F5,$3F,$A0,$FE,$1B,$BF,$EA,$E7,$37,$C7,$E1
	.dc.b	$30,$7E,$FA,$B7,$1F,$98,$9D,$BE,$59,$DB,$FB,$B9,$DE,$1E,$19,$CB
	.dc.b	$67,$5B,$7F,$8A,$23,$04,$C7,$D9,$5E,$19,$D4,$62,$BC,$53,$D2,$3D
	.dc.b	$93,$A9,$F5,$1A,$25,$7F,$B5,$2F,$CB,$5C,$E3,$C1,$3C,$1D,$57,$38
	.dc.b	$F7,$1C,$A7,$81,$E0,$C7,$C5,$1C,$63,$2F,$FA,$96,$FB,$B9,$21,$D2
	.dc.b	$FB,$D7,$28,$C5,$72,$55,$C9,$C9,$CB,$2F,$C4,$E3,$F1,$EF,$94,$62
	.dc.b	$7D,$07,$8F,$1F,$64,$67,$1F,$83,$3D,$0E,$F3,$AC,$BE,$FD,$6F,$2F
	.dc.b	$26,$5E,$38,$C5,$F8,$EB,$F0,$B3,$97,$89,$32,$F3,$4F,$90,$F0,$57
	.dc.b	$27,$C3,$97,$AA,$B8,$47,$E2,$23,$E3,$5B,$C4,$FA,$CD,$8F,$CC,$6F
	.dc.b	$89,$CB,$63,$27,$E7,$63,$2D,$33,$5F,$9B,$8C,$B6,$33,$A7,$E2,$E7
	.dc.b	$F2,$1E,$CB,$E2,$DE,$E7,$FA,$13,$9F,$FD,$0B,$27,$FD,$99,$87,$FA
	.dc.b	$FA,$B6,$FF,$B0,$23,$38,$FF,$AF,$E7,$2F,$FB,$06,$32,$74,$8C,$BD
	.dc.b	$8C,$CF,$FB,$D6,$BC,$9E,$DF,$8E,$31,$5F,$EA,$E7,$C5,$19,$B5,$7A
	.dc.b	$FD,$8C,$3C,$AB,$D1,$18,$BF,$F6,$44,$60,$FF,$0D,$D3,$E6,$B2,$FC
	.dc.b	$18,$3F,$EF,$1F,$A0,$FF,$1E,$47,$09,$DE,$5F,$EA,$C8,$4E,$AB,$91
	.dc.b	$CA,$B7,$3D,$0D,$F8,$1E,$B9,$F1,$67,$ED,$9C,$9D,$DD,$57,$29,$71
	.dc.b	$BE,$71,$E8,$EA,$FE,$07,$D0,$7C,$1F,$A0,$B9,$7D,$B6,$E3,$F6,$5B
	.dc.b	$AF,$D7,$7E,$47,$C1,$CA,$FC,$17,$C1,$5B,$FC,$04,$DB,$FB,$D3,$6B
	.dc.b	$FD,$19,$AB,$E7,$FE,$C6,$DE,$7E,$BF,$5C,$63,$74,$62,$FF,$C9,$F5
	.dc.b	$B7,$E8,$56,$E7,$59,$FC,$E3,$ED,$47,$E1,$7D,$D7,$D0,$7F,$B2,$AF
	.dc.b	$89,$F9,$BF,$83,$FC,$21,$F8,$5F,$7A,$C6,$49,$87,$E7,$63,$2F,$E7
	.dc.b	$FD,$CF,$C6,$63,$F8,$D6,$D7,$E9,$DB,$3B,$8F,$D5,$4F,$DA,$F0,$FE
	.dc.b	$07,$BB,$67,$DF,$B7,$FB,$BD,$FC,$74,$E9,$A6,$A7,$85,$FC,$1D,$FF
	.dc.b	$CA,$93,$CB,$E4,$F3,$C5,$FB,$1F,$A0,$FD,$11,$FA,$3F,$8A,$B5,$FB
	.dc.b	$D8,$B8,$7E,$EE,$6D,$7E,$E7,$E1,$BF,$55,$F3,$BF,$AB,$6E,$1F,$C1
	.dc.b	$5B,$7F,$C0,$DB,$5F,$EF,$0F,$9F,$FF,$4F,$D7,$FA,$2A,$1F,$41,$F9
	.dc.b	$BB,$56,$9F,$D4,$CE,$BB,$3F,$4D,$39,$8F,$E3,$15,$C6,$31,$7F,$EA
	.dc.b	$48,$C7,$A5,$C0,$FC,$27,$37,$CD,$C7,$F1,$17,$C1,$73,$BE,$53,$FC
	.dc.b	$EF,$8F,$FA,$83,$E2,$5F,$5C,$F8,$6F,$C7,$56,$B0,$FD,$FD,$6A,$7F
	.dc.b	$78,$FF,$8A,$2F,$97,$86,$E2,$CF,$9B,$FC,$6E,$BF,$D9,$11,$F9,$9F
	.dc.b	$87,$F8,$91,$C7,$1F,$E1,$7A,$27,$37,$FC,$27,$C0,$7B,$3E,$0F,$5F
	.dc.b	$0D,$D3,$F8,$FC,$FF,$15,$E4,$47,$FC,$01,$8E,$2E,$B5,$E2,$D2,$67
	.dc.b	$FA,$FF,$A1,$7C,$1F,$86,$BF,$8D,$7A,$0E,$AF,$E6,$75,$F1,$72,$EF
	.dc.b	$87,$F2,$CA,$D6,$3F,$EC,$CF,$39,$7B,$8F,$E6,$34,$AD,$BF,$CB,$2B
	.dc.b	$7E,$8F,$83,$1E,$ED,$BF,$E7,$B7,$C3,$1F,$EB,$B8,$FF,$03,$C7,$F8
	.dc.b	$44,$E6,$BB,$E2,$F7,$7F,$99,$7D,$03,$DD,$FE,$6D,$39,$AF,$F7,$18
	.dc.b	$9D,$3F,$D3,$23,$58,$7F,$42,$8D,$57,$2B,$EA,$D8,$E1,$1E,$0F,$80
	.dc.b	$C5,$7F,$1D,$FA,$17,$BF,$F8,$7F,$92,$7C,$AF,$F4,$99,$CE,$3F,$97
	.dc.b	$9C,$BF,$D7,$A3,$37,$F9,$69,$CB,$D6,$FA,$5F,$36,$5F,$E8,$7E,$7F
	.dc.b	$D7,$C1,$F6,$AD,$8F,$48,$C2,$F9,$EF,$92,$79,$4B,$7E,$36,$D7,$B5
	.dc.b	$B5,$C9,$E1,$3F,$3B,$77,$C4,$E2,$FF,$A0,$76,$7E,$E7,$A1,$B7,$F6
	.dc.b	$D9,$CF,$89,$FD,$42,$31,$9F,$A6,$F8,$46,$32,$FC,$7C,$62,$FF,$BA
	.dc.b	$3F,$C1,$5F,$E0,$B7,$FB,$7F,$A1,$7D,$0E,$A3,$7C,$7F,$D7,$B8,$1F
	.dc.b	$43,$8C,$BF,$AD,$C6,$1F,$F0,$8E,$5F,$31,$FF,$07,$EA,$9D,$67,$EE
	.dc.b	$C3,$A6,$1E,$4F,$8B,$F2,$FC,$FD,$C7,$9B,$6F,$E2,$CF,$AB,$8B,$1C
	.dc.b	$1F,$66,$7F,$64,$3E,$81,$7F,$0B,$E8,$7F,$6F,$18,$9F,$D9,$9F,$B3
	.dc.b	$38,$4E,$F1,$FF,$13,$CE,$E7,$F7,$3D,$78,$AD,$47,$03,$9C,$65,$FF
	.dc.b	$17,$EE,$C3,$C3,$5F,$EC,$EA,$D5,$7F,$36,$9C,$BC,$8E,$86,$70,$DB
	.dc.b	$EE,$C3,$C9,$87,$59,$CC,$FA,$E7,$35,$D2,$3D,$27,$4A,$E0,$F6,$3A
	.dc.b	$61,$EC,$AF,$81,$18,$3D,$75,$CD,$C1,$E5,$39,$1E,$19,$E4,$7B,$DF
	.dc.b	$04,$71,$7E,$81,$FE,$F0,$29,$B3,$18,$67,$C8,$FE,$AD,$F0,$17,$C9
	.dc.b	$FF,$40,$57,$FB,$E2,$72,$D8,$CB,$F7,$A1,$F2,$9F,$E8,$38,$E0,$D7
	.dc.b	$4D,$7F,$9B,$C7,$0D,$7C,$53,$C7,$3F,$C2,$7C,$AB,$E8,$7F,$9E,$C6
	.dc.b	$1B,$AF,$2A,$E0,$DB,$FC,$0D,$F1,$C6,$DF,$07,$FC,$EF,$8F,$85,$FC
	.dc.b	$1B,$71,$ED,$1C,$0F,$C5,$C6,$70,$ED,$1F,$8A,$BE,$CD,$BE,$CF,$F8
	.dc.b	$17,$1F,$2B,$E0,$6A,$9E,$38,$78,$23,$C7,$3A,$C3,$C1,$1F,$E8,$EA
	.dc.b	$DE,$1F,$D8,$1B,$1D,$71,$F3,$DF,$69,$CC,$F7,$C7,$CB,$87,$4F,$AB
	.dc.b	$FE,$8B,$F0,$3F,$DA,$B1,$FE,$E3,$18,$BF,$F4,$EC,$63,$C0,$56,$F8
	.dc.b	$BF,$D4,$3E,$6B,$EC,$5F,$27,$9C,$59,$73,$7F,$D3,$ED,$CB,$A6,$3B
	.dc.b	$C3,$95,$DE,$EF,$F4,$BE,$1F,$E2,$D7,$FC,$38,$DF,$3F,$5F,$0B,$23
	.dc.b	$04,$3F,$C5,$F3,$C9,$FF,$57,$FE,$08,$FC,$11,$EC,$79,$FF,$06,$7E
	.dc.b	$2E,$F1,$30,$DC,$3E,$1F,$17,$FB,$03,$1E,$58,$7B,$DB,$3A,$18,$74
	.dc.b	$C3,$99,$BC,$7C,$33,$BA,$E2,$FF,$B1,$2B,$83,$CD,$AF,$AA,$78,$BF
	.dc.b	$E4,$CF,$0F,$FD,$7E,$FF,$73,$EE,$F9,$91,$FC,$9E,$3F,$BB,$78,$98
	.dc.b	$33,$32,$32,$43,$A3,$97,$B4,$7B,$F1,$7C,$EF,$C9,$E2,$FC,$0D,$79
	.dc.b	$DA,$FD,$09,$DB,$EF,$6B,$EF,$8D,$F2,$7F,$90,$B3,$46,$36,$FA,$17
	.dc.b	$C2,$BF,$C8,$FA,$FD,$7C,$7C,$5C,$73,$2E,$D1,$F8,$DF,$C0,$9F,$6D
	.dc.b	$7D,$69,$5F,$E7,$99,$CB,$67,$2F,$F9,$E2,$3F,$CD,$2E,$48,$66,$77
	.dc.b	$8F,$F9,$6B,$45,$74,$AF,$C8,$FE,$15,$F9,$A7,$EF,$78,$2F,$91,$FC
	.dc.b	$93,$63,$F0,$93,$BC,$3D,$ED,$DD,$3F,$8E,$AE,$0F,$4B,$FC,$A5,$6E
	.dc.b	$1B,$23,$1C,$1F,$DF,$6B,$C2,$70,$E9,$FD,$A7,$DB,$3E,$16,$93,$1F
	.dc.b	$CF,$79,$0F,$A4,$ED,$C0,$BE,$15,$FA,$3E,$F7,$E0,$5F,$E2,$48,$C3
	.dc.b	$5F,$F5,$2C,$F0,$7F,$83,$C6,$78,$7F,$8A,$B7,$87,$E0,$DF,$4F,$3F
	.dc.b	$C9,$7F,$C3,$BF,$03,$D3,$FC,$49,$B2,$73,$C5,$FE,$32,$F8,$DF,$E0
	.dc.b	$5D,$8F,$A2,$FE,$35,$6F,$C9,$BF,$C3,$BA,$FF,$26,$6E,$9F,$4B,$97
	.dc.b	$FC,$29,$7F,$C2,$FF,$0F,$7C,$AD,$26,$DF,$0F,$0E,$2F,$DA,$FA,$47
	.dc.b	$CE,$7F,$29,$84,$78,$DF,$E1,$86,$30,$D4,$70,$8C,$B5,$67,$FE,$73
	.dc.b	$8E,$51,$8C,$FF,$D5,$31,$89,$D5,$F4,$3A,$DC,$BE,$B1,$FC,$9F,$F1
	.dc.b	$6F,$6C,$7F,$15,$6B,$FC,$65,$B5,$FB,$6F,$B3,$FB,$8F,$EF,$63,$38
	.dc.b	$FC,$96,$D7,$F2,$86,$F0,$AF,$8D,$7F,$50,$FB,$E7,$C4,$7F,$49,$5F
	.dc.b	$49,$4C,$44,$C7,$FD,$ED,$3A,$6C,$E5,$F8,$87,$EB,$3E,$95,$F4,$FF
	.dc.b	$4D,$B1,$F4,$9A,$F0,$3F,$D8,$5F,$01,$B9,$91,$B7,$EE,$7C,$2F,$FA
	.dc.b	$5A,$30,$7D,$C5,$EF,$68,$E3,$1E,$4E,$E7,$E0,$AF,$A8,$FE,$D0,$FF
	.dc.b	$23,$DB,$FB,$9B,$6B,$F9,$4B,$B7,$FA,$A5,$72,$AE,$3E,$5C,$9C,$D6
	.dc.b	$DF,$EC,$FF,$8A,$FF,$31,$7F,$AD,$FD,$2F,$FB,$7D,$FE,$A8,$D7,$7E
	.dc.b	$77,$F6,$39,$E1,$DA,$AD,$73,$FF,$3D,$6E,$9F,$DA,$EB,$B6,$1F,$EE
	.dc.b	$BF,$8E,$F8,$FA,$48,$F8,$D1,$F2,$8C,$3F,$47,$7F,$46,$3F,$ED,$15
	.dc.b	$3B,$9F,$9D,$5C,$2B,$53,$FD,$E7,$9B,$F9,$4C,$7F,$51,$E4,$7C,$8F
	.dc.b	$F8,$15,$B8,$7F,$77,$AD,$E7,$FE,$03,$6C,$7F,$80,$27,$6F,$F7,$B6
	.dc.b	$DF,$E8,$FC,$BF,$F1,$07,$E0,$2F,$9B,$BE,$EC,$C3,$EB,$CE,$5B,$39
	.dc.b	$7F,$09,$1F,$EF,$E7,$27,$E2,$FD,$4F,$C7,$FC,$64,$8C,$10,$F9,$0E
	.dc.b	$5F,$95,$1F,$96,$31,$8F,$CB,$9C,$4F,$E2,$BF,$06,$7D,$DF,$F8,$E5
	.dc.b	$F9,$F1,$9E,$0C,$4C,$8C,$10,$FC,$E4,$F2,$7F,$C4,$5F,$8C,$B8,$3F
	.dc.b	$C6,$9F,$A4,$3E,$72,$9F,$E2,$F9,$D5,$70,$AC,$BF,$E2,$7A,$E0,$E4
	.dc.b	$FD,$0F,$E3,$DF,$E5,$A7,$FB,$4F,$C6,$3E,$97,$FC,$B7,$3F,$E2,$C8
	.dc.b	$C4,$FF,$84,$A3,$13,$AF,$85,$FF,$6C,$B6,$79,$BC,$BE,$0B,$31,$C5
	.dc.b	$BE,$0A,$FC,$52,$C3,$BC,$E6,$B7,$5E,$0A,$E4,$76,$3C,$98,$6A,$64
	.dc.b	$66,$B9,$36,$BB,$DF,$B2,$3F,$69,$C8,$FA,$CE,$91,$86,$F7,$7E,$55
	.dc.b	$CF,$D6,$BE,$67,$DB,$5B,$3D,$5A,$FF,$85,$1C,$63,$DF,$0E,$51,$A3
	.dc.b	$FC,$33,$7C,$DE,$F8,$F8,$DD,$1C,$AF,$A5,$72,$BF,$55,$F1,$AD,$CE
	.dc.b	$6B,$77,$CE,$FC,$4D,$8F,$35,$F1,$9C,$E3,$E2,$26,$1E,$29,$D5,$74
	.dc.b	$9C,$D7,$0A,$D6,$3F,$13,$1F,$14,$7F,$8D,$9A,$FA,$CF,$83,$1B,$C7
	.dc.b	$C9,$1B,$6B,$B1,$8F,$89,$D6,$1E,$2C,$7C,$91,$E2,$30,$68,$F8,$B5
	.dc.b	$D2,$B9,$47,$5C,$3C,$A7,$9A,$32,$E2,$7C,$6E,$70,$F4,$61,$DA,$FC
	.dc.b	$D7,$C2,$F6,$49,$FC,$4D,$70,$AF,$1C,$71,$AE,$0F,$3B,$ED,$36,$F9
	.dc.b	$3E,$D3,$89,$31,$FA,$47,$E1,$30,$EC,$E3,$1F,$8A,$E6,$7A,$61,$E1
	.dc.b	$AF,$24,$7E,$64,$FC,$E1,$70,$FC,$CB,$70,$FC,$B3,$9F,$9C,$FF,$58
	.dc.b	$B7,$E1,$39,$3F,$81,$8C,$BF,$9A,$6B,$98,$FD,$7C,$7F,$AE,$B1,$FD
	.dc.b	$B7,$D6,$FF,$91,$38,$B3,$52,$FF,$3F,$7B,$D3,$F6,$17,$B9,$CD,$6F
	.dc.b	$0F,$D9,$5E,$E7,$FD,$13,$5B,$9C,$D6,$EF,$CC,$FF,$04,$9D,$46,$2B
	.dc.b	$A9,$2B,$CF,$E8,$7F,$34,$79,$C6,$67,$84,$6B,$1E,$D3,$F0,$4D,$1E
	.dc.b	$4A,$E5,$5C,$09,$1B,$AE,$4F,$03,$FC,$1D,$8F,$B5,$B8,$FA,$6F,$9C
	.dc.b	$71,$BE,$6E,$8F,$E1,$F8,$FB,$9D,$10,$D6,$1E,$03,$DA,$49,$E7,$3C
	.dc.b	$67,$71,$DB,$36,$3E,$5A,$D4,$63,$1F,$87,$7C,$0F,$85,$96,$27,$51
	.dc.b	$EF,$AE,$37,$EF,$9C,$E3,$E3,$AE,$27,$C5,$8C,$E3,$E6,$CB,$E1,$CF
	.dc.b	$72,$4E,$9E,$35,$CD,$CD,$FA,$6B,$8D,$6E,$73,$97,$CF,$CB,$EE,$F0
	.dc.b	$7F,$B7,$7E,$8F,$E8,$B6,$3F,$AB,$77,$FF,$51,$6C,$EA,$0D,$31,$C1
	.dc.b	$67,$8B,$70,$FE,$99,$7F,$EE,$B8,$C6,$9F,$DB,$30,$ED,$5A,$7A,$DF
	.dc.b	$23,$C2,$7F,$6D,$F4,$1C,$FF,$E8,$E8,$FF,$77,$F6,$FF,$84,$3C,$77
	.dc.b	$9D,$CB,$1F,$F4,$4E,$1C,$3D,$0F,$F0,$AD,$EE,$B4,$FF,$86,$27,$58
	.dc.b	$74,$3F,$D9,$53,$A7,$FD,$FB,$1A,$BF,$F4,$AE,$1E,$0D,$BA,$D2,$5B
	.dc.b	$3F,$06,$FF,$26,$4F,$FA,$01,$78,$9C,$2B,$71,$A7,$FD,$1C,$EA,$78
	.dc.b	$73,$EE,$FA,$5D,$AE,$1A,$7F,$8C,$A7,$37,$EB,$C1,$F3,$8F,$95,$76
	.dc.b	$F8,$63,$1F,$5B,$67,$37,$C6,$72,$D8,$CD,$FF,$AD,$A7,$2D,$8C,$E3
	.dc.b	$FE,$44,$9C,$B6,$33,$5F,$1E,$72,$D8,$CD,$7A,$15,$FC,$1F,$1F,$39
	.dc.b	$1C,$DF,$F9,$DB,$2D,$56,$DB,$19,$3F,$CF,$71,$F4,$78,$3F,$DB,$9E
	.dc.b	$B7,$D3,$E1,$7E,$2C,$E6,$D7,$96,$AD,$4E,$71,$FF,$49,$46,$2B,$6D
	.dc.b	$8F,$91,$ED,$7C,$DF,$13,$E7,$36,$32,$7C,$D8,$CB,$4C,$F0,$7C,$D8
	.dc.b	$CB,$63,$EF,$66,$F4,$3F,$DA,$91,$F8,$1D,$3F,$C2,$F7,$EF,$9C,$DF
	.dc.b	$E6,$BC,$44,$E7,$CF,$F8,$51,$8C,$3A,$DF,$0E,$5F,$D8,$4E,$70,$FF
	.dc.b	$72,$7C,$D7,$FC,$00,$BF,$F6,$EB,$FC,$B9,$BE,$7F,$E6,$63,$13,$C2
	.dc.b	$31,$E2,$65,$B5,$F9,$99,$7D,$3B,$EA,$7F,$51,$E2,$3E,$A7,$FD,$11
	.dc.b	$7B,$AD,$3F,$A1,$9D,$79,$1F,$AF,$C3,$9E,$16,$F7,$0D,$7F,$D3,$7F
	.dc.b	$10,$FA,$77,$C5,$EF,$09,$9F,$77,$F7,$55,$B9,$CE,$1C,$EB,$F8,$27
	.dc.b	$73,$DE,$F9,$19,$7E,$DD,$FC,$D9,$89,$DB,$8E,$5F,$E0,$0D,$8F,$DE
	.dc.b	$57,$3E,$EF,$CC,$6A,$F9,$BA,$7E,$63,$F0,$AB,$7F,$83,$9D,$36,$74
	.dc.b	$D3,$F7,$DF,$57,$FC,$0F,$7F,$E0,$8D,$8C,$9F,$C0,$E3,$2D,$33,$AF
	.dc.b	$F0,$38,$CB,$63,$F2,$9F,$0E,$F8,$EF,$D8,$36,$3F,$8C,$7C,$43,$E1
	.dc.b	$DE,$EE,$AF,$F6,$DF,$8A,$BA,$3F,$95,$5E,$EB,$31,$8B,$E3,$87,$4F
	.dc.b	$C1,$7F,$2D,$9F,$D5,$18,$23,$3C,$BF,$D1,$7E,$89,$E2,$76,$9C,$E1
	.dc.b	$FC,$FE,$DC,$1F,$EE,$78,$F4,$B7,$1E,$DC,$AF,$90,$65,$E8,$AD,$67
	.dc.b	$FC,$E6,$73,$87,$23,$FA,$03,$8C,$3D,$18,$F8,$27,$38,$47,$FA,$07
	.dc.b	$57,$E6,$38,$F1,$7F,$D9,$3E,$49,$F7,$3F,$C7,$FF,$14,$7D,$27,$F2
	.dc.b	$39,$CE,$3F,$C2,$67,$2F,$F0,$E8,$CD,$FF,$62,$9C,$BD,$6F,$A5,$F3
	.dc.b	$6D,$F1,$6C,$FC,$4F,$C1,$7E,$5B,$FD,$A1,$FE,$0B,$B7,$F5,$B7,$3C
	.dc.b	$07,$37,$FA,$C7,$26,$2B,$97,$87,$FD,$9E,$31,$5C,$63,$1E,$26,$5B
	.dc.b	$7F,$CA,$B9,$DF,$35,$FD,$0E,$EC,$E6,$30,$7F,$49,$9C,$E3,$FE,$04
	.dc.b	$AB,$97,$F8,$36,$6D,$FF,$81,$7C,$AF,$F0,$2E,$1C,$4E,$71,$D4,$FE
	.dc.b	$71,$7C,$30,$FF,$05,$FC,$DB,$E2,$BF,$C0,$4B,$E1,$EC,$61,$B7,$FC
	.dc.b	$1A,$74,$D9,$D3,$4D,$4F,$0A,$FC,$BD,$70,$7F,$C5,$7F,$39,$FD,$B9
	.dc.b	$FF,$10,$D7,$F8,$40,$FF,$12,$C6,$74,$FF,$1A,$46,$65,$E7,$AB,$53
	.dc.b	$F0,$CD,$46,$7D,$4F,$EE,$51,$89,$DD,$35,$4E,$6B,$7E,$D7,$F8,$97
	.dc.b	$E7,$9F,$73,$7E,$1B,$0D,$BF,$E5,$33,$A6,$CE,$9A,$7F,$12,$FC,$01
	.dc.b	$E3,$FF,$0E,$6C,$7F,$32,$FA,$2F,$EC,$8F,$F0,$38,$CE,$9F,$B6,$6E
	.dc.b	$FE,$88,$FE,$1F,$18,$9F,$E8,$B5,$CA,$73,$7F,$D5,$DF,$F2,$67,$C2
	.dc.b	$7F,$93,$7E,$7B,$F9,$F4,$7F,$93,$FC,$0E,$F1,$88,$FF,$29,$09,$E1
	.dc.b	$E5,$E2,$B6,$FF,$B9,$23,$33,$CE,$FC,$D7,$CE,$3C,$21,$ED,$7C,$A3
	.dc.b	$FD,$ED,$5C,$94,$65,$F1,$C6,$63,$11,$D2,$79,$D7,$82,$FC,$93,$DD
	.dc.b	$F4,$46,$63,$B9,$93,$D1,$8F,$85,$E9,$3C,$23,$11,$F0,$CE,$07,$9A
	.dc.b	$7A,$3E,$63,$37,$E6,$78,$5F,$9A,$B9,$E1,$EB,$BE,$35,$F1,$8F,$A2
	.dc.b	$73,$29,$93,$DA,$6B,$E8,$DF,$4F,$FE,$75,$8D,$C6,$3E,$03,$8C,$23
	.dc.b	$0E,$A7,$83,$CF,$E9,$FF,$C3,$4F,$26,$B9,$DF,$9D,$27,$33,$F8,$88
	.dc.b	$C4,$FD,$BF,$A7,$7D,$3F,$E3,$35,$FC,$74,$62,$FF,$01,$7C,$1F,$9A
	.dc.b	$D9,$DD,$71,$DF,$F2,$B7,$C6,$F7,$8F,$A2,$FF,$35,$87,$1B,$DE,$1C
	.dc.b	$6F,$FD,$5D,$87,$1B,$DE,$1C,$6F,$D4,$DA,$DD,$F1,$57,$C6,$B7,$8F
	.dc.b	$AA,$F8,$56,$AF,$85,$EE,$3F,$D7,$D8,$F9,$5B,$5C,$27,$51,$C4,$E9
	.dc.b	$71,$FC,$8C,$68,$DD,$73,$F2,$7F,$D8,$BE,$B7,$EE,$AF,$F8,$44,$62
	.dc.b	$FF,$C4,$BF,$0D,$FE,$78,$7F,$89,$7C,$B6,$F8,$BF,$C2,$93,$99,$FF
	.dc.b	$6B,$D3,$EC,$F5,$1F,$CA,$E3,$0F,$4C,$73,$E1,$3A,$5F,$51,$87,$FC
	.dc.b	$C7,$5F,$CA,$BD,$BC,$4C,$8C,$10,$FF,$10,$CF,$27,$FD,$2F,$E0,$7F
	.dc.b	$B4,$9F,$F4,$B7,$D4,$7F,$8F,$9B,$E3,$F9,$BC,$3F,$E9,$1E,$DD,$BE
	.dc.b	$15,$FE,$F2,$BE,$0D,$BD,$54,$8E,$64,$3F,$AD,$BE,$EF,$17,$FC,$95
	.dc.b	$1F,$D9,$FB,$30,$67,$D7,$91,$92,$1F,$D9,$9C,$BF,$DA,$23,$8B,$FD
	.dc.b	$C6,$3D,$56,$F9,$BF,$A3,$E3,$B8,$1C,$D8,$6E,$1F,$E1,$3A,$E2,$66
	.dc.b	$64,$64,$87,$F8,$45,$CB,$FE,$5D,$DB,$FC,$D3,$3F,$73,$E9,$FF,$C3
	.dc.b	$B3,$FE,$8A,$5F,$51,$93,$5E,$64,$9D,$46,$5E,$11,$97,$FC,$BB,$3C
	.dc.b	$9D,$9A,$3F,$C9,$1C,$9F,$03,$C0,$FF,$76,$47,$FB,$55,$AF,$F9,$56
	.dc.b	$3F,$01,$E9,$FE,$3A,$3F,$CE,$DA,$FF,$51,$F8,$8E,$46,$AA,$4F,$38
	.dc.b	$CB,$E0,$3F,$D4,$FE,$BF,$FA,$3D,$BE,$EF,$E7,$A3,$11,$FE,$83,$F2
	.dc.b	$1F,$51,$E1,$F0,$7F,$CE,$97,$FA,$8E,$E7,$D4,$3E,$A3,$0D,$C7,$A9
	.dc.b	$FE,$9A,$70,$FF,$A0,$B8,$FF,$D1,$6F,$FA,$E9,$AF,$F9,$F2,$37,$87
	.dc.b	$F0,$0F,$A8,$BE,$27,$F9,$27,$AD,$F1,$9F,$8A,$6E,$1F,$EA,$8A,$DB
	.dc.b	$FE,$27,$8C,$BF,$E8,$8F,$81,$FE,$AB,$87,$F0,$C6,$E1,$D7,$0F,$CA
	.dc.b	$FD,$3F,$92,$BF,$C3,$1C,$88,$C5,$7F,$96,$DB,$87,$FA,$AB,$0E,$68
	.dc.b	$7D,$3B,$E6,$BF,$88,$C6,$32,$B3,$E6,$CB,$FC,$4D,$3C,$2B,$26,$AA
	.dc.b	$47,$59,$DD,$75,$8C,$63,$FE,$22,$8C,$47,$FA,$FC,$F0,$1A,$3C,$CF
	.dc.b	$86,$38,$5F,$ED,$BE,$9F,$DB,$99,$A7,$C0,$ED,$7C,$97,$E2,$B2,$F8
	.dc.b	$F3,$AF,$86,$E4,$7F,$1B,$9D,$63,$FE,$FE,$F9,$8F,$E5,$FF,$16,$F7
	.dc.b	$E9,$88,$9B,$7F,$C0,$09,$D3,$67,$2F,$B7,$E6,$1F,$4D,$F3,$9B,$1E
	.dc.b	$F0,$E1,$F8,$9F,$02,$FA,$7E,$15,$F5,$A1,$F4,$F4,$CD,$73,$BE,$5B
	.dc.b	$FD,$8A,$DC,$F7,$6B,$E5,$BF,$E9,$BF,$4E,$F9,$7B,$D9,$2B,$EA,$79
	.dc.b	$6A,$D4,$61,$AF,$FB,$B1,$FE,$58,$E6,$7A,$D7,$F4,$9F,$97,$FE,$57
	.dc.b	$3F,$C2,$BE,$7F,$FD,$54,$FE,$07,$9D,$18,$8F,$E5,$13,$C3,$4F,$F3
	.dc.b	$C7,$3A,$31,$C9,$FE,$BB,$BE,$15,$FA,$66,$C7,$E8,$6B,$87,$9B,$F2
	.dc.b	$FE,$97,$E6,$67,$32,$F8,$D1,$97,$E4,$46,$5F,$AB,$19,$7E,$BC,$67
	.dc.b	$1F,$A1,$3F,$8D,$F2,$3F,$A6,$F2,$BE,$9E,$99,$C3,$F4,$19,$7E,$62
	.dc.b	$B3,$1F,$0E,$B8,$3A,$E9,$FE,$6B,$8F,$FA,$67,$89,$F3,$8F,$88,$FF
	.dc.b	$03,$AF,$72,$98,$89,$87,$DC,$9D,$36,$72,$FE,$4B,$D3,$FC,$87,$D3
	.dc.b	$C8,$FC,$54,$7E,$4C,$C3,$F9,$88,$C3,$A3,$F2,$0E,$B1,$FC,$A4,$FF
	.dc.b	$87,$3E,$9F,$FD,$D1,$5F,$8E,$5C,$19,$35,$F1,$92,$75,$19,$78,$47
	.dc.b	$F8,$57,$E8,$BF,$8B,$E1,$FC,$9B,$E9,$9F,$4F,$FD,$F9,$B7,$F4,$2B
	.dc.b	$7C,$FF,$85,$6C,$7F,$B7,$67,$6F,$F6,$06,$DF,$E4,$A7,$7C,$0F,$A7
	.dc.b	$FF,$1F,$7C,$5B,$E7,$BF,$98,$4E,$7D,$FF,$F5,$8C,$62,$BF,$C5,$35
	.dc.b	$BC,$BF,$C2,$51,$CA,$75,$F0,$9C,$27,$37,$70,$E9,$1F,$ED,$4A,$DB
	.dc.b	$4D,$96,$36,$FF,$B4,$E3,$C5,$3C,$EF,$C1,$3C,$0D,$17,$0F,$35,$F2
	.dc.b	$6C,$F0,$9F,$DC,$F2,$BE,$9B,$C5,$18,$6F,$77,$E5,$DC,$FD,$13,$E4
	.dc.b	$7D,$D5,$B3,$C2,$5C,$7E,$33,$74,$FF,$76,$D7,$18,$EB,$8F,$85,$E3
	.dc.b	$8F,$76,$C6,$6F,$9D,$71,$9C,$E3,$C8,$8F,$8A,$35,$3E,$2B,$E8,$DB
	.dc.b	$E9,$87,$FC,$00,$9C,$BE,$6C,$7D,$B1,$8C,$3F,$BA,$35,$F6,$1F,$DD
	.dc.b	$F0,$EF,$5C,$EF,$AC,$F0,$6B,$C0,$4F,$07,$58,$FA,$B1,$F3,$3F,$43
	.dc.b	$1F,$1D,$F2,$C7,$C4,$78,$F0,$E5,$87,$58,$CC,$F8,$27,$31,$8B,$ED
	.dc.b	$8F,$BE,$E9,$E2,$3E,$69,$F2,$F0,$F0,$CE,$A3,$D5,$7D,$30,$F2,$12
	.dc.b	$7D,$13,$BC,$3C,$0F,$33,$13,$B7,$07,$83,$1F,$82,$DC,$7C,$CD,$3C
	.dc.b	$D1,$A3,$F0,$B8,$F9,$A3,$47,$99,$C4,$6E,$7E,$B6,$1E,$6A,$F7,$9F
	.dc.b	$E3,$13,$E1,$14,$FC,$F1,$70,$FC,$B3,$7E,$53,$FD,$46,$DF,$83,$E0
	.dc.b	$FE,$22,$32,$FE,$3D,$AE,$63,$F5,$F1,$FE,$AC,$C3,$F5,$3F,$4F,$FE
	.dc.b	$41,$E4,$CD,$4B,$FD,$3D,$EF,$2F,$45,$EE,$73,$5B,$C3,$F4,$57,$B9
	.dc.b	$FF,$27,$56,$E7,$35,$BB,$F3,$3F,$C1,$27,$51,$8A,$EA,$4A,$F3,$FA
	.dc.b	$2C,$79,$4F,$B4,$D5,$79,$A7,$51,$83,$55,$E2,$7F,$85,$DF,$81,$CD
	.dc.b	$78,$67,$47,$82,$FB,$CF,$F8,$53,$1F,$73,$71,$F5,$63,$E0,$BE,$38
	.dc.b	$73,$25,$7C,$49,$D4,$62,$39,$1E,$71,$8F,$95,$CD,$74,$7C,$E7,$3C
	.dc.b	$7C,$B9,$32,$F2,$B7,$2F,$2C,$EA,$72,$7A,$B0,$ED,$5B,$BF,$13,$70
	.dc.b	$F8,$51,$8A,$E3,$87,$48,$C3,$F1,$AF,$9C,$EA,$7D,$91,$89,$F1,$1C
	.dc.b	$6B,$D9,$3E,$EC,$7E,$25,$6E,$3F,$A8,$F0,$3C,$07,$D3,$FC,$A9,$FF
	.dc.b	$7C,$46,$34,$FE,$BC,$FF,$51,$6C,$EE,$B9,$36,$78,$57,$67,$83,$63
	.dc.b	$5B,$6D,$B7,$D0,$EF,$19,$69,$99,$FE,$D3,$E1,$39,$9E,$87,$E5,$BA
	.dc.b	$3F,$D7,$7E,$6F,$89,$FF,$06,$46,$70,$FF,$5C,$79,$3F,$E8,$6C,$F1
	.dc.b	$93,$2E,$97,$8A,$D5,$FF,$87,$63,$18,$F5,$8F,$F0,$9F,$D3,$FF,$91
	.dc.b	$DF,$F3,$EB,$FE,$47,$6F,$3F,$6D,$D3,$9C,$9A,$7F,$94,$63,$94,$62
	.dc.b	$FA,$F8,$7C,$67,$35,$B8,$F4,$FD,$3B,$E5,$1D,$B8,$5F,$31,$63,$0D
	.dc.b	$C7,$FD,$8D,$5C,$63,$19,$78,$44,$66,$77,$7D,$9B,$5C,$31,$FE,$0F
	.dc.b	$5B,$BE,$59,$FF,$9B,$31,$F8,$5E,$3F,$93,$1F,$F3,$44,$63,$1C,$B6
	.dc.b	$32,$7F,$9B,$63,$EB,$F0,$FF,$B5,$3D,$8F,$AD,$F4,$FF,$9B,$AF,$F3
	.dc.b	$E4,$63,$9F,$7E,$52,$31,$3A,$3F,$D4,$D5,$C7,$4F,$81,$AF,$CB,$F5
	.dc.b	$FE,$57,$89,$F2,$DB,$7F,$09,$B5,$C3,$85,$18,$E1,$F8,$71,$8A,$F9
	.dc.b	$37,$CD,$A7,$FA,$63,$D0,$FF,$64,$3F,$EB,$9F,$11,$3F,$02,$73,$8F
	.dc.b	$D5,$E9,$FF,$0A,$CE,$7C,$FF,$CC,$5F,$0C,$3A,$46,$27,$F2,$5F,$4F
	.dc.b	$FE,$F3,$9F,$F6,$9B,$FC,$7D,$BE,$7E,$BC,$4D,$B6,$32,$70,$8F,$D4
	.dc.b	$E4,$FA,$47,$D0,$7E,$B3,$C6,$7D,$3F,$F9,$E7,$0F,$F7,$B5,$69,$5F
	.dc.b	$E3,$EB,$7E,$37,$4A,$D4,$E6,$F9,$3F,$95,$3C,$18,$77,$8C,$F1,$7F
	.dc.b	$58,$9C,$E3,$E0,$AF,$DF,$FC,$3B,$E9,$DF,$15,$DE,$7F,$6D,$EF,$7C
	.dc.b	$9C,$3F,$32,$FE,$7D,$BC,$3A,$EF,$DB,$72,$FF,$7E,$D7,$1C,$FF,$D9
	.dc.b	$8F,$17,$FB,$5F,$2F,$C7,$F9,$AF,$E1,$DB,$9F,$CB,$FD,$3E,$19,$87
	.dc.b	$FB,$EE,$74,$FF,$B6,$67,$39,$FE,$47,$E9,$7F,$80,$6B,$FC,$05,$B5
	.dc.b	$FB,$86,$CE,$E3,$F7,$73,$B6,$9B,$2C,$7E,$3F,$E1,$9F,$17,$FC,$95
	.dc.b	$B3,$F6,$B4,$FF,$03,$FC,$5B,$E1,$DE,$E6,$F0,$FE,$59,$F4,$EE,$9F
	.dc.b	$E6,$37,$BA,$CC,$62,$F8,$E3,$D4,$FC,$3D,$3E,$9F,$F3,$5A,$7E,$8F
	.dc.b	$1E,$95,$B8,$56,$FB,$3A,$CE,$61,$BF,$F8,$8E,$38,$79,$6F,$B0,$7C
	.dc.b	$FF,$9A,$B5,$8F,$F1,$C9,$CD,$F0,$F9,$17,$D4,$6E,$3F,$A0,$46,$1F
	.dc.b	$05,$F2,$E2,$FF,$16,$E4,$AD,$63,$FC,$A1,$B3,$FE,$32,$EE,$FC,$BF
	.dc.b	$2F,$4C,$BF,$91,$FC,$9B,$E9,$BF,$AC,$7D,$43,$EA,$33,$18,$6E,$DF
	.dc.b	$17,$B5,$5B,$8F,$91,$E8,$B0,$E9,$1B,$9D,$6D,$FD,$16,$B5,$8F,$5A
	.dc.b	$E3,$7C,$2B,$F8,$B5,$F0,$D7,$DF,$F5,$1F,$61,$BE,$3F,$F6,$58,$C6
	.dc.b	$1F,$EF,$E7,$F8,$A5,$FF,$4A,$72,$B8,$51,$AB,$EB,$CD,$CF,$0E,$8D
	.dc.b	$C3,$EE,$D7,$2B,$E3,$5F,$D3,$3A,$DF,$35,$FE,$01,$9B,$7F,$12,$6E
	.dc.b	$5F,$CC,$66,$E5,$FE,$66,$DB,$F3,$F8,$71,$3A,$17,$0E,$4D,$BF,$95
	.dc.b	$87,$29,$FF,$0A,$E1,$67,$DD,$F3,$8F,$90,$FF,$05,$2F,$88,$B1,$86
	.dc.b	$DF,$E6,$A7,$4D,$9D,$34,$D4,$F0,$AF,$EE,$F5,$C1,$FB,$DF,$3D,$FD
	.dc.b	$DD,$FE,$A1,$5F,$E1,$23,$F9,$CE,$69,$D6,$9F,$C7,$A7,$55,$FE,$2D
	.dc.b	$9C,$C7,$F8,$C8,$CD,$71,$F6,$3F,$C6,$11,$89,$DD,$39,$3F,$D3,$1E
	.dc.b	$E7,$F8,$0A,$73,$18,$AF,$DC,$FC,$FB,$EA,$3F,$C6,$0B,$E2,$2C,$61
	.dc.b	$B3,$C1,$B8,$FF,$86,$2B,$84,$FF,$07,$FA,$87,$91,$FE,$45,$FA,$C7
	.dc.b	$D3,$FF,$1A,$9C,$D7,$EB,$67,$2F,$1A,$D1,$FE,$43,$FA,$9F,$EF,$7C
	.dc.b	$3F,$CA,$2F,$8E,$5F,$D7,$4F,$EC,$BF,$15,$FD,$9B,$E8,$7F,$A3,$47
	.dc.b	$F6,$8F,$0F,$71,$88,$FF,$72,$4E,$FD,$0F,$F7,$35,$69,$FF,$6F,$C6
	.dc.b	$67,$96,$5F,$EB,$EB,$E5,$1E,$00,$F5,$BE,$51,$FE,$EF,$AE,$4A,$32
	.dc.b	$F8,$8E,$37,$E2,$BE,$CF,$8A,$32,$78,$B3,$E6,$F8,$70,$F0,$57,$08
	.dc.b	$F0,$1C,$0F,$03,$97,$C1,$1C,$27,$9C,$30,$F3,$E1,$E6,$8E,$D8,$79
	.dc.b	$6B,$75,$AC,$3A,$56,$8F,$31,$E1,$3E,$31,$FE,$1A,$3C,$A5,$3E,$11
	.dc.b	$4D,$9B,$FA,$77,$D4,$FF,$9E,$63,$51,$8F,$7B,$84,$23,$0E,$A7,$83
	.dc.b	$DE,$3F,$27,$AF,$06,$C7,$4E,$0F,$F4,$7B,$67,$51,$8D,$7F,$15,$F5
	.dc.b	$3F,$D8,$2B,$EE,$B5,$FA,$17,$D2,$7F,$20,$DD,$7E,$15,$6E,$79,$F9
	.dc.b	$9F,$15,$DE,$1D,$EB,$E1,$37,$0F,$F2,$1D,$F1,$9E,$77,$C5,$DE,$3F
	.dc.b	$90,$C7,$C1,$7C,$2B,$4D,$35,$87,$57,$E5,$37,$0E,$D1,$B3,$DB,$1B
	.dc.b	$AE,$B0,$9D,$B6,$BB,$BD,$2F,$B9,$F2,$23,$33,$C2,$BC,$FE,$6B,$E1
	.dc.b	$FF,$D6,$B8,$7F,$84,$DA,$FF,$AC,$63,$5D,$7F,$8D,$C7,$FC,$0F,$F3
	.dc.b	$5F,$D5,$AB,$84,$E7,$8B,$FC,$4B,$87,$2B,$FF,$19,$B6,$3F,$CD,$FE
	.dc.b	$0B,$0E,$58,$7F,$3E,$F6,$79,$3F,$CB,$FE,$9F,$FC,$72,$FF,$81,$DA
	.dc.b	$EC,$FF,$03,$F7,$7F,$B4,$1D,$78,$9C,$BB,$D5,$C3,$4F,$EC,$D7,$BB
	.dc.b	$FF,$05,$7E,$24,$FB,$9E,$1D,$4F,$C1,$A9,$E4,$F1,$15,$BB,$FF,$0B
	.dc.b	$B5,$FF,$52,$57,$17,$FC,$3C,$D7,$BD,$F3,$9E,$18,$FF,$9D,$2B,$4D
	.dc.b	$C3,$55,$E1,$D1,$F3,$DE,$DE,$07,$CF,$53,$51,$33,$FF,$17,$4E,$9B
	.dc.b	$39,$7F,$C0,$33,$C5,$C9,$FE,$0F,$7E,$07,$E2,$7F,$E5,$68,$FF,$1C
	.dc.b	$7C,$9B,$E6,$7E,$73,$63,$FC,$73,$3B,$7F,$C6,$CD,$D3,$FD,$23,$3B
	.dc.b	$9F,$F6,$53,$63,$FC,$35,$3B,$E5,$FE,$FA,$DA,$FE,$E5,$5B,$9F,$F5
	.dc.b	$CF,$1A,$31,$2E,$15,$F1,$7F,$C9,$9F,$07,$F8,$5E,$F7,$C8,$53,$EF
	.dc.b	$7A,$9F,$EB,$4C,$3F,$C9,$75,$97,$FC,$B1,$5C,$1C,$90,$FF,$72,$39
	.dc.b	$6D,$74,$9F,$C5,$7E,$25,$FE,$91,$9F,$F3,$17,$E2,$BF,$E6,$96,$FA
	.dc.b	$FA,$F8,$19,$18,$21,$FE,$61,$AE,$4D,$8F,$D0,$57,$38,$FF,$7B,$37
	.dc.b	$83,$FC,$8D,$5C,$1F,$F6,$1F,$90,$F9,$1F,$D4,$36,$FF,$CE,$CD,$7F
	.dc.b	$82,$7D,$43,$EA,$B7,$5F,$B8,$F1,$3F,$D6,$15,$7E,$13,$FA,$96,$3D
	.dc.b	$AB,$13,$0F,$F5,$14,$EB,$9B,$ED,$E5,$DE,$FF,$D7,$7F,$97,$66,$FF
	.dc.b	$D5,$6D,$C4,$E7,$93,$FD,$43,$F7,$9E,$AD,$3F,$81,$7C,$2F,$E1,$3A
	.dc.b	$4D,$CF,$D5,$97,$FA,$4F,$F1,$C7,$D9,$7F,$09,$6C,$66,$39,$F6,$7F
	.dc.b	$16,$8F,$F3,$27,$D2,$5F,$75,$F2,$74,$C4,$49,$FF,$62,$1F,$E6,$B8
	.dc.b	$D3,$89,$4E,$E1,$C2,$E2,$FE,$FD,$5D,$A7,$85,$EE,$BF,$DF,$BF,$57
	.dc.b	$FC,$75,$BE,$5F,$F2,$38,$C6,$1F,$E6,$8F,$9A,$7C,$A6,$6A,$63,$FE
	.dc.b	$B5,$BD,$BF,$AB,$AD,$BF,$EB,$67,$F3,$D8,$79,$63,$FC,$CD,$F0,$AF
	.dc.b	$A8,$FE,$64,$D8,$E1,$E0,$FF,$2B,$BE,$39,$FE,$EA,$B5,$5F,$02,$31
	.dc.b	$7B,$BF,$A1,$F3,$DF,$1A,$BE,$5B,$5E,$00,$39,$C7,$CD,$1F,$E9,$BF
	.dc.b	$97,$E7,$C1,$23,$9B,$F6,$E3,$11,$D0,$FF,$51,$7D,$8F,$DF,$33,$5F
	.dc.b	$EE,$AF,$9E,$7C,$2F,$F5,$AF,$35,$1F,$DB,$EF,$FD,$FA,$D7,$CF,$1A
	.dc.b	$9E,$13,$CA,$77,$EB,$60,$CC,$FF,$53,$F9,$B7,$DB,$7F,$5F,$FA,$5F
	.dc.b	$77,$0F,$F2,$46,$E1,$FC,$8E,$B6,$FF,$BB,$9A,$FC,$8E,$E4,$F0,$6E
	.dc.b	$FD,$F0,$FC,$F7,$A5,$F9,$FF,$AA,$98,$7E,$66,$32,$FE,$6E,$32,$FC
	.dc.b	$9E,$BF,$EE,$59,$FE,$6A,$B3,$C1,$F5,$25,$7C,$AF,$FD,$FD,$A3,$8B
	.dc.b	$FB,$E7,$E3,$5F,$C6,$F2,$FE,$82,$DA,$F7,$4F,$C5,$37,$E6,$C8,$F8
	.dc.b	$11,$F5,$7E,$AF,$FD,$B5,$1F,$87,$CF,$F3,$5C,$89,$F9,$2D,$DF,$87
	.dc.b	$C6,$CB,$E1,$9F,$F8,$72,$F6,$49,$FE,$0D,$F8,$0B,$E9,$AF,$64,$C3
	.dc.b	$F8,$1C,$E5,$B3,$97,$F2,$10,$9E,$73,$27,$8C,$6F,$3F,$CE,$D7,$39
	.dc.b	$FD,$1F,$92,$FA,$7F,$D0,$D7,$ED,$F9,$9F,$56,$FA,$CF,$F1,$43,$70
	.dc.b	$F6,$9F,$DA,$EF,$8E,$DF,$EF,$06,$C7,$FB,$96,$78,$3F,$8B,$6E,$1F
	.dc.b	$C1,$71,$F1,$5F,$F8,$AF,$E0,$1F,$75,$FA,$7F,$C6,$FF,$C4,$73,$FE
	.dc.b	$46,$6E,$3F,$EF,$99,$E0,$6F,$DF,$91,$FE,$42,$8F,$D8,$7E,$56,$FA
	.dc.b	$FF,$F6,$9F,$95,$FE,$50,$8C,$57,$FB,$72,$31,$3A,$87,$3A,$BF,$B0
	.dc.b	$7C,$1F,$F2,$08,$C5,$6D,$B5,$CC,$E3,$5D,$9F,$F6,$05,$F8,$7D,$7E
	.dc.b	$13,$9C,$3A,$36,$7C,$75,$CE,$FC,$57,$C3,$09,$39,$BE,$0F,$82,$7D
	.dc.b	$4F,$EE,$BC,$37,$D7,$79,$63,$0D,$F1,$3F,$8B,$B9,$FA,$A7,$CB,$72
	.dc.b	$9C,$D7,$A8,$FE,$58,$E3,$5F,$E8,$71,$D0,$D0,$AE,$6F,$7C,$3C,$6E
	.dc.b	$AB,$96,$BF,$ED,$96,$C7,$F3,$9A,$E2,$EE,$B8,$DF,$82,$B7,$3E,$18
	.dc.b	$F1,$D7,$07,$55,$C2,$FB,$36,$BD,$6D,$7D,$18,$7C,$18,$C6,$1D,$8E
	.dc.b	$F8,$75,$BE,$78,$FB,$49,$87,$C2,$9D,$3D,$A3,$38,$F9,$1F,$A1,$8F
	.dc.b	$77,$37,$D2,$B9,$4E,$5F,$85,$3A,$7E,$34,$66,$BA,$46,$8E,$26,$CF
	.dc.b	$8D,$5D,$63,$E3,$CF,$18,$C1,$27,$A9,$C7,$0F,$85,$8F,$C7,$C7,$DB
	.dc.b	$3D,$0F,$71,$E1,$C3,$AB,$C6,$7E,$59,$C7,$0F,$04,$70,$9F,$49,$C1
	.dc.b	$E5,$87,$8D,$EC,$7D,$63,$A6,$3E,$DC,$3C,$78,$F9,$A3,$18,$7C,$9A
	.dc.b	$FF,$02,$1E,$62,$9F,$99,$2C,$FE,$81,$BF,$35,$FE,$6E,$7F,$2C,$D7
	.dc.b	$F4,$6D,$7F,$B4,$B6,$0F,$F9,$FF,$1F,$CC,$37,$1F,$1F,$D7,$FF,$17
	.dc.b	$E3,$FD,$A4,$E7,$0E,$BF,$01,$83,$F7,$55,$B8,$FD,$54,$ED,$F0,$CE
	.dc.b	$DF,$D9,$CE,$F0,$F0,$CE,$5B,$3A,$E0,$FF,$1D,$46,$09,$8F,$C0,$AF
	.dc.b	$0C,$EA,$31,$5E,$29,$E9,$1E,$C9,$D4,$FA,$8D,$12,$BF,$84,$DF,$96
	.dc.b	$B9,$C7,$82,$78,$3A,$AE,$71,$EE,$39,$4F,$03,$C1,$8F,$8A,$38,$C6
	.dc.b	$5F,$F2,$25,$F7,$72,$43,$A5,$F7,$AE,$51,$8A,$E4,$AB,$93,$93,$96
	.dc.b	$5F,$80,$C7,$E3,$DF,$28,$C4,$FA,$0F,$1E,$3E,$C8,$CE,$3F,$06,$7A
	.dc.b	$1D,$E7,$59,$7D,$FA,$DE,$5E,$4C,$BC,$71,$8B,$F1,$D7,$E1,$67,$2F
	.dc.b	$12,$65,$E6,$9F,$21,$E0,$AE,$4F,$87,$2F,$55,$70,$8F,$C4,$47,$CF
	.dc.b	$E1,$FF,$1B,$7C,$67,$E0,$1B,$1F,$D6,$FC,$37,$E0,$79,$55,$49,$3B
	.dc.b	$9B,$7F,$D4,$6F,$9B,$6F,$8D,$6C,$ED,$7C,$36,$FE,$F1,$3A,$AE,$18
	.dc.b	$67,$5F,$F0,$D7,$81,$FE,$99,$EB,$FF,$12,$79,$FF,$EE,$3C,$FF,$C4
	.dc.b	$3D,$9F,$E2,$9A,$D4,$7F,$A3,$7C,$0F,$F4,$6F,$46,$33,$FF,$13,$55
	.dc.b	$9C,$D7,$2E,$FF,$F1,$D3,$FE,$2C,$8C,$D7,$F8,$DA,$7F,$C8,$D3,$9C
	.dc.b	$5F,$0F,$D7,$2F,$F1,$D7,$87,$D6,$73,$5D,$63,$15,$BB,$E7,$87,$F9
	.dc.b	$F6,$F6,$FF,$A9,$BE,$B9,$F2,$BF,$53,$1F,$F1,$FC,$7F,$B4,$3A,$FD
	.dc.b	$8E,$B8,$6F,$07,$F8,$32,$B7,$7C,$70,$E9,$39,$7B,$46,$63,$FD,$8D
	.dc.b	$7D,$6B,$8D,$F4,$C3,$BD,$EE,$7F,$D7,$55,$BC,$3F,$D4,$FF,$61,$FE
	.dc.b	$AE,$9F,$F2,$AF,$32,$1E,$8F,$D3,$8C,$77,$D9,$CD,$70,$79,$5F,$4B
	.dc.b	$D5,$6D,$FA,$1E,$6F,$FA,$52,$FE,$C7,$33,$CF,$CF,$9F,$F7,$E3,$13
	.dc.b	$FE,$AD,$8C,$4E,$A3,$8B,$EE,$F6,$3D,$FD,$96,$7E,$5B,$73,$F9,$2D
	.dc.b	$E2,$F9,$3C,$1F,$9B,$6D,$F2,$BF,$A6,$DB,$FF,$54,$F9,$DF,$93,$F1
	.dc.b	$1F,$81,$7B,$23,$0D,$BF,$CA,$4E,$9F,$F6,$34,$E9,$FF,$6C,$DF,$CD
	.dc.b	$F5,$3C,$51,$FE,$EF,$E3,$FF,$6F,$CE,$74,$D3,$67,$89,$FE,$E2,$8F
	.dc.b	$D4,$64,$FA,$B5,$7E,$AE,$9F,$81,$B2,$DD,$6A,$7F,$D2,$A6,$A3,$35
	.dc.b	$23,$FD,$33,$E2,$BE,$C1,$F7,$79,$CB,$FD,$F9,$8F,$36,$E1,$FB,$BB
	.dc.b	$E1,$5B,$E0,$FF,$4D,$7B,$9F,$C1,$6F,$FD,$D1,$F6,$6F,$73,$F6,$D3
	.dc.b	$9A,$FF,$57,$4E,$74,$FE,$17,$39,$7F,$85,$46,$6F,$FD,$6F,$39,$5E
	.dc.b	$22,$75,$7F,$EB,$9D,$1F,$4F,$FC,$3D,$B8,$F6,$F9,$AC,$4C,$8F,$E2
	.dc.b	$13,$AC,$7B,$D6,$5F,$DB,$FD,$9F,$EC,$BE,$CB,$F6,$6D,$BF,$DF,$B6
	.dc.b	$77,$1F,$C0,$A7,$6D,$36,$58,$63,$FE,$D1,$FC,$6D,$F6,$5F,$C9,$DB
	.dc.b	$5F,$C9,$7E,$29,$F0,$39,$AF,$E5,$90,$C2,$CE,$1B,$5F,$ED,$8F,$8E
	.dc.b	$7C,$73,$E6,$7F,$9B,$46,$1F,$E5,$9F,$59,$F9,$FB,$F2,$DF,$5A,$D3
	.dc.b	$CF,$E4,$7C,$6D,$C7,$C7,$D1,$D2,$73,$7C,$8F,$B3,$9A,$75,$5C,$2F
	.dc.b	$91,$EF,$8C,$57,$1C,$1F,$63,$FD,$22,$3A,$73,$31,$E9,$87,$F5,$6F
	.dc.b	$94,$7D,$17,$F5,$1F,$96,$6D,$FD,$26,$73,$5F,$D4,$67,$2F,$1A,$E1
	.dc.b	$5B,$6C,$64,$E1,$19,$69,$FD,$9B,$E6,$5F,$15,$FD,$87,$EB,$7F,$1D
	.dc.b	$CD,$FF,$00,$67,$35,$FD,$86,$72,$FD,$BF,$61,$3A,$DB,$DF,$5A,$53
	.dc.b	$AC,$3B,$76,$7F,$65,$BE,$18,$B1,$FE,$ED,$8F,$1C,$38,$47,$86,$FF
	.dc.b	$B6,$7C,$DB,$EF,$7F,$BC,$7B,$3F,$87,$AF,$F0,$0C,$E5,$FE,$F1,$19
	.dc.b	$D3,$EF,$60,$AD,$47,$D6,$9D,$65,$E3,$8F,$4C,$EA,$31,$C5,$FD,$FB
	.dc.b	$A3,$C7,$97,$89,$B9,$78,$1F,$F0,$17,$3B,$EB,$7F,$C1,$DE,$55,$AF
	.dc.b	$F0,$BC,$DC,$FF,$0F,$37,$2F,$16,$BF,$AC,$C3,$89,$D0,$B8,$72,$6D
	.dc.b	$FB,$B0,$E5,$3F,$E2,$1C,$2C,$FD,$4F,$A0,$7C,$D7,$F8,$87,$F0,$07
	.dc.b	$C7,$7F,$86,$67,$38,$7F,$88,$67,$2F,$F8,$BA,$33,$7F,$9E,$9C,$BF
	.dc.b	$3A,$33,$A7,$C8,$9C,$BE,$C8,$CD,$7F,$8D,$67,$2F,$E3,$63,$35,$FC
	.dc.b	$B2,$72,$FE,$33,$E3,$7F,$AD,$BF,$CC,$2B,$FC,$0B,$7F,$22,$73,$C1
	.dc.b	$F8,$DF,$67,$FC,$69,$18,$AD,$9F,$9B,$71,$1F,$C4,$3E,$A7,$F8,$C4
	.dc.b	$FF,$15,$6E,$1F,$C1,$F5,$7D,$27,$E2,$2B,$F8,$95,$6B,$2F,$C5,$78
	.dc.b	$08,$C4,$EB,$0F,$F1,$05,$F1,$55,$B7,$FB,$79,$FC,$A7,$E1,$BF,$95
	.dc.b	$7D,$0F,$F0,$88,$FE,$59,$F5,$2F,$9F,$FC,$C4,$E6,$7F,$DD,$35,$C3
	.dc.b	$C0,$FF,$75,$DE,$F2,$C5,$7F,$BB,$2A,$A9,$D2,$BE,$8F,$8C,$C5,$F4
	.dc.b	$C3,$C2,$FF,$BD,$A1,$7D,$2F,$9D,$F2,$9C,$9C,$8F,$5D,$F2,$7C,$86
	.dc.b	$4F,$23,$97,$DA,$66,$7A,$DF,$86,$FC,$15,$BA,$F4,$5F,$C5,$BD,$C7
	.dc.b	$68,$C4,$EE,$3C,$B1,$8C,$3E,$2C,$72,$C3,$E0,$46,$23,$71,$87,$D0
	.dc.b	$7C,$73,$E2,$1F,$00,$B8,$61,$D1,$FD,$9C,$7D,$33,$EA,$7F,$D0,$31
	.dc.b	$F3,$63,$1E,$E7,$46,$C6,$67,$6F,$33,$9F,$D4,$FF,$7C,$79,$35,$D4
	.dc.b	$A3,$11,$E3,$8C,$70,$77,$9E,$59,$FE,$3F,$EA,$5E,$8F,$A6,$30,$DD
	.dc.b	$39,$57,$06,$DF,$E4,$2F,$8E,$36,$F8,$3F,$E0,$9C,$7C,$2F,$E0,$DB
	.dc.b	$8F,$68,$E0,$7E,$6A,$33,$87,$68,$FC,$55,$F6,$6D,$F6,$7E,$2E,$1C
	.dc.b	$71,$F3,$47,$F8,$66,$B8,$63,$DE,$3C,$53,$A7,$89,$D0,$F1,$9A,$AE
	.dc.b	$86,$27,$91,$FB,$7C,$7C,$F3,$FC,$0B,$E0,$FF,$DD,$EF,$F8,$95,$AF
	.dc.b	$F8,$3D,$B5,$FD,$5F,$93,$F8,$9F,$6F,$E2,$63,$F8,$8C,$EB,$83,$FD
	.dc.b	$C5,$C7,$FD,$A7,$0F,$F6,$E3,$5E,$9E,$E7,$19,$FE,$49,$EC,$D9,$FF
	.dc.b	$2E,$FD,$4F,$F8,$F9,$FF,$72,$35,$FE,$DB,$D5,$FE,$84,$7F,$DE,$6D
	.dc.b	$7F,$C4,$71,$BC,$BF,$B5,$B6,$FB,$CF,$F4,$DF,$96,$F0,$76,$7F,$0A
	.dc.b	$F0,$EF,$47,$F9,$16,$B4,$43,$FC,$5E,$E5,$B1,$E9,$F0,$DE,$03,$0C
	.dc.b	$36,$BF,$DF,$D7,$C7,$3F,$EF,$ED,$CB,$FA,$A5,$F3,$F0,$DF,$51,$E6
	.dc.b	$E4,$7D,$4D,$35,$13,$2F,$F3,$16,$5F,$E0,$AA,$DC,$63,$1F,$EE,$D7
	.dc.b	$B8,$FE,$F5,$7C,$DD,$1F,$E8,$EA,$E2,$E8,$FF,$10,$FD,$97,$C8,$F2
	.dc.b	$5F,$4B,$4F,$85,$E5,$26,$99,$AF,$F1,$3C,$E1,$E6,$60,$87,$F8,$AA
	.dc.b	$79,$3F,$E4,$BF,$AA,$FA,$FD,$9F,$EA,$D8,$F3,$F2,$7F,$49,$F7,$FE
	.dc.b	$97,$9E,$BA,$70,$89,$5C,$C8,$7F,$9A,$AF,$A3,$63,$8B,$FE,$6B,$F5
	.dc.b	$BF,$0F,$E9,$FF,$97,$23,$FC,$D1,$E9,$7E,$07,$D4,$FF,$3E,$46,$CC
	.dc.b	$90,$FF,$22,$BA,$6D,$74,$99,$1C,$4F,$F3,$EF,$92,$F9,$3F,$CB,$B6
	.dc.b	$3E,$0B,$5F,$F4,$37,$D5,$7F,$8C,$1C,$CF,$EB,$7C,$27,$B3,$FC,$DF
	.dc.b	$A7,$FC,$12,$DE,$9F,$F5,$57,$17,$F7,$FB,$E1,$39,$C3,$FD,$A3,$F7
	.dc.b	$6F,$7F,$F5,$2D,$8F,$F5,$21,$DF,$DC,$E1,$87,$F8,$27,$D6,$F4,$99
	.dc.b	$9A,$7F,$AF,$7A,$3F,$67,$ED,$FE,$32,$3D,$B3,$FB,$6F,$AA,$7D,$55
	.dc.b	$24,$F1,$3F,$32,$3E,$31,$F1,$1F,$C3,$63,$1F,$55,$64,$D9,$93,$CA
	.dc.b	$72,$D9,$CB,$FE,$BB,$3B,$4F,$38,$91,$C6,$3F,$86,$FD,$57,$F1,$5E
	.dc.b	$3F,$F3,$EB,$78,$FE,$7B,$6B,$FD,$37,$F1,$0F,$90,$E3,$E7,$C9,$FC
	.dc.b	$FD,$6F,$CC,$5E,$E7,$F7,$31,$8E,$FF,$E3,$BF,$32,$FA,$AA,$7F,$17
	.dc.b	$BF,$CF,$E9,$FE,$82,$6E,$9F,$82,$AE,$18,$FE,$B9,$AF,$FA,$7A,$76
	.dc.b	$FF,$BC,$7E,$63,$E7,$BF,$49,$AF,$C7,$6D,$FF,$A8,$63,$EB,$7C,$DF
	.dc.b	$D5,$9C,$E1,$FE,$15,$E1,$7D,$57,$D7,$35,$C3,$30,$FA,$D1,$97,$C1
	.dc.b	$1F,$D2,$FC,$B7,$C9,$3E,$0F,$0B,$D8,$A6,$22,$5F,$F2,$2A,$E3,$3A
	.dc.b	$8C,$57,$37,$E7,$FC,$B3,$EA,$BF,$A5,$FC,$FF,$AA,$5D,$E8,$C7,$32
	.dc.b	$F8,$57,$2B,$E3,$A4,$8C,$1F,$93,$F5,$3F,$31,$F3,$12,$30,$43,$F2
	.dc.b	$9C,$08,$D4,$62,$7F,$40,$63,$93,$E9,$56,$E7,$94,$7F,$B1,$31,$72
	.dc.b	$7F,$66,$FA,$AF,$E4,$14,$AF,$F6,$BC,$61,$E1,$3F,$D8,$EB,$8D,$7F
	.dc.b	$36,$AD,$72,$7F,$0E,$6C,$7F,$BE,$A7,$6F,$FA,$87,$EA,$9F,$55,$F6
	.dc.b	$DF,$99,$C8,$69,$A9,$FB,$0D,$79,$6D,$C4,$E5,$5F,$C0,$FE,$A9,$F4
	.dc.b	$57,$AE,$5F,$F0,$A9,$CB,$67,$2F,$F0,$18,$4F,$39,$93,$C6,$37,$B7
	.dc.b	$E7,$EB,$9C,$FC,$9D,$4B,$8D,$F4,$5F,$A5,$B5,$7F,$34,$FA,$AF,$F7
	.dc.b	$4D,$FF,$60,$6B,$CC,$E5,$5C,$67,$AD,$F4,$7F,$C3,$F5,$C2,$FA,$DF
	.dc.b	$E5,$A3,$13,$C2,$F9,$4F,$FB,$8F,$EA,$7F,$59,$F5,$5F,$E1,$B8,$FF
	.dc.b	$6B,$37,$0F,$F7,$24,$FF,$88,$FE,$10,$F0,$F8,$09,$7F,$E3,$88,$FD
	.dc.b	$B7,$D5,$3E,$A7,$FD,$97,$E4,$7C,$4E,$54,$E6,$FF,$DA,$D3,$98,$C4
	.dc.b	$7F,$B0,$A3,$1E,$9F,$28,$57,$2B,$C3,$63,$99,$AB,$E7,$F0,$7F,$CA
	.dc.b	$E7,$C3,$7C,$27,$38,$78,$70,$C1,$99,$B8,$74,$BE,$CF,$94,$ED,$3B
	.dc.b	$7C,$27,$F0,$5F,$0D,$F4,$D8,$F0,$7F,$8C,$36,$33,$3C,$C7,$D1,$BE
	.dc.b	$AB,$E0,$47,$3A,$E0,$47,$84,$ED,$C6,$BF,$D2,$5E,$07,$5C,$3C,$2F
	.dc.b	$0B,$EB,$18,$9E,$0D,$D3,$FD,$98,$DA,$FC,$ED,$F7,$7A,$63,$EA,$BE
	.dc.b	$2D,$BE,$2D,$9E,$38,$78,$23,$11,$E6,$AE,$38,$FC,$AC,$3B,$46,$27
	.dc.b	$BC,$FB,$B0,$FF,$70,$56,$9E,$F1,$99,$FD,$9E,$3E,$97,$E6,$E3,$E1
	.dc.b	$9D,$47,$86,$B9,$CE,$63,$B9,$30,$F9,$98,$77,$8E,$B8,$FA,$B1,$EC
	.dc.b	$77,$9D,$4F,$53,$78,$F8,$23,$33,$C2,$B9,$1F,$12,$32,$E2,$78,$61
	.dc.b	$E5,$C3,$E2,$13,$1F,$4E,$3E,$CC,$7A,$5F,$CB,$BE,$15,$F1,$AB,$D8
	.dc.b	$F0,$69,$E6,$29,$F2,$0E,$98,$78,$DF,$C5,$18,$AF,$2B,$F9,$53,$F9
	.dc.b	$19,$CC,$A7,$E5,$0B,$7F,$94,$6F,$D4,$FF,$9A,$DF,$C8,$B5,$FB,$6D
	.dc.b	$7F,$B9,$36,$0F,$F9,$E3,$1F,$CD,$B7,$0F,$1F,$D5,$7F,$15,$E6,$FD
	.dc.b	$A4,$E7,$0E,$BE,$F6,$0F,$DD,$D6,$E3,$F5,$53,$B7,$C3,$3B,$7F,$75
	.dc.b	$3B,$C3,$C3,$39,$6C,$EB,$6F,$E6,$F1,$82,$63,$F0,$2B,$C3,$3A,$8C
	.dc.b	$57,$8A,$7A,$47,$B2,$75,$3E,$A3,$44,$AF,$E0,$97,$E5,$AE,$71,$E0
	.dc.b	$9E,$0E,$AB,$9C,$7B,$8E,$53,$C0,$F0,$63,$E2,$8E,$31,$97,$FB,$1D
	.dc.b	$F7,$72,$43,$A5,$F7,$AE,$51,$8A,$E4,$AB,$93,$93,$96,$5F,$80,$C7
	.dc.b	$E3,$DF,$28,$C4,$FA,$0F,$1E,$3E,$C8,$CE,$3F,$06,$7A,$1D,$E7,$59
	.dc.b	$7D,$FA,$DE,$5E,$4C,$BC,$71,$8B,$F1,$D7,$E1,$67,$2F,$12,$65,$E6
	.dc.b	$9F,$21,$E0,$AE,$4F,$87,$2F,$55,$70,$8F,$C4,$47,$CF,$DF,$FC,$77
	.dc.b	$F1,$9F,$80,$6C,$7F,$57,$F0,$DF,$55,$95,$54,$93,$B9,$B7,$FD,$76
	.dc.b	$F9,$B6,$F8,$D6,$CE,$D7,$C3,$6F,$ED,$D3,$AA,$E1,$86,$75,$FF,$09
	.dc.b	$77,$7F,$85,$BA,$7F,$C2,$FE,$4B,$CD,$FF,$72,$69,$FE,$A0,$ED,$FF
	.dc.b	$19,$BF,$E2,$CA,$DC,$EB,$2F,$F1,$37,$7F,$F8,$BB,$1E,$97,$FE,$E0
	.dc.b	$AD,$5F,$F8,$92,$B5,$3C,$CD,$74,$F0,$C3,$FC,$5F,$E0,$DC,$FC,$9F
	.dc.b	$11,$9C,$3F,$C7,$73,$9C,$93,$9D,$3F,$CA,$39,$75,$C7,$FD,$0F,$5A
	.dc.b	$8F,$65,$F1,$BF,$07,$D5,$3E,$5B,$EB,$D7,$18,$FF,$3A,$75,$6A,$73
	.dc.b	$8F,$F9,$16,$73,$5B,$9C,$ED,$FE,$42,$9C,$D6,$E7,$35,$FE,$56,$9C
	.dc.b	$D6,$E7,$34,$9C,$D6,$E7,$31,$89,$FF,$19,$FD,$57,$FA,$CE,$BF,$D5
	.dc.b	$CF,$FA,$CD,$BD,$9F,$37,$9D,$39,$AF,$F2,$F4,$62,$75,$39,$C7,$FC
	.dc.b	$B7,$39,$BE,$BE,$87,$D3,$F2,$BF,$D4,$91,$F3,$F8,$BF,$DD,$DE,$66
	.dc.b	$7C,$EF,$B5,$CE,$8C,$56,$E3,$0F,$FB,$02,$77,$3F,$3F,$D6,$FA,$1D
	.dc.b	$96,$BE,$73,$77,$F9,$8D,$D7,$E5,$36,$7E,$43,$67,$E6,$70,$FE,$96
	.dc.b	$BF,$D6,$9E,$67,$E3,$FC,$47,$D5,$7B,$63,$0D,$C7,$DD,$3A,$7F,$DA
	.dc.b	$F3,$A7,$FD,$B7,$97,$CF,$F4,$BB,$47,$F9,$C5,$46,$73,$D3,$67,$89
	.dc.b	$FE,$E4,$8F,$D2,$66,$FA,$6E,$33,$F9,$D1,$F5,$4C,$77,$3A,$AF,$F4
	.dc.b	$D5,$6A,$73,$87,$EB,$E3,$07,$E9,$E7,$B4,$66,$FA,$4E,$6F,$A6,$9F
	.dc.b	$E9,$EB,$DC,$FE,$BF,$C5,$7D,$53,$EA,$BB,$CF,$EC,$FC,$07,$B1,$FC
	.dc.b	$12,$3F,$D6,$FC,$CF,$AA,$7B,$7F,$BF,$AD,$67,$FF,$00,$27,$37,$CE
	.dc.b	$73,$7C,$6F,$87,$89,$FC,$26,$F7,$87,$FA,$EE,$BE,$57,$D2,$7F,$0D
	.dc.b	$DC,$F7,$F9,$CC,$33,$3F,$CE,$4E,$9F,$04,$E6,$FF,$79,$F0,$FF,$C1
	.dc.b	$7E,$AB,$F8,$33,$6B,$F8,$23,$67,$71,$FC,$1E,$76,$D3,$65,$8F,$47
	.dc.b	$D5,$3E,$AB,$F9,$23,$67,$F6,$1F,$0E,$F7,$F3,$5F,$CA,$63,$FD,$C7
	.dc.b	$3A,$9C,$5E,$5F,$E6,$1F,$18,$F8,$A7,$CD,$6A,$FF,$95,$4E,$63,$17
	.dc.b	$FE,$E5,$87,$C6,$F8,$6B,$F8,$17,$C7,$BE,$D7,$AC,$E6,$FE,$94,$E6
	.dc.b	$BD,$3D,$7A,$CF,$D0,$D8,$C9,$2B,$C1,$F5,$7F,$D5,$1F,$B9,$AA,$BF
	.dc.b	$A3,$7C,$9B,$E8,$BF,$A5,$7C,$AB,$EA,$F3,$18,$6E,$5F,$C6,$67,$4F
	.dc.b	$E5,$EB,$93,$99,$DE,$3F,$8B,$AD,$BF,$D8,$AB,$6E,$67,$FA,$C7,$CC
	.dc.b	$BE,$5F,$FA,$E7,$D5,$FE,$2E,$BF,$B4,$CE,$71,$FE,$55,$39,$F1,$3F
	.dc.b	$B4,$4E,$6B,$F3,$53,$9A,$43,$C2,$56,$EF,$FB,$ED,$EE,$B5,$39,$BF
	.dc.b	$D8,$CE,$6F,$8E,$1E,$6E,$DF,$EE,$1F,$36,$FA,$BF,$EE,$9E,$CF,$E1
	.dc.b	$B1,$FD,$E4,$E5,$FE,$85,$19,$CF,$FC,$0B,$39,$7F,$65,$19,$D3,$FC
	.dc.b	$09,$39,$7F,$C2,$71,$9A,$F6,$E4,$AD,$47,$F0,$89,$D5,$FC,$19,$CA
	.dc.b	$F0,$7F,$C0,$BD,$5F,$E0,$9E,$67,$D1,$7F,$83,$7D,$6B,$97,$F1,$59
	.dc.b	$B8,$7F,$07,$9B,$A7,$F8,$66,$6D,$7F,$18,$9B,$5F,$46,$6E,$29,$CE
	.dc.b	$9F,$B5,$C7,$C0,$78,$CB,$87,$F2,$5F,$A1,$7C,$CF,$F8,$87,$EA,$9F
	.dc.b	$57,$FE,$21,$9C,$D7,$F6,$F9,$CB,$C6,$B8,$56,$DB,$19,$38,$46,$5A
	.dc.b	$7F,$35,$F8,$EF,$EA,$FA,$FC,$98,$FF,$06,$3F,$E3,$E8,$CE,$89,$CE
	.dc.b	$DF,$E3,$D9,$CC,$60,$FB,$2E,$3D,$DF,$E1,$FF,$57,$FC,$53,$0F,$E6
	.dc.b	$9F,$5D,$FC,$6D,$FE,$19,$19,$C3,$FC,$AB,$39,$7C,$12,$F1,$15,$BC
	.dc.b	$7F,$47,$18,$9F,$F1,$07,$D6,$7E,$1A,$31,$3F,$D7,$E7,$F5,$3F,$13
	.dc.b	$FA,$AF,$A2,$FE,$11,$1F,$AC,$FA,$C7,$8D,$E0,$9C,$CF,$FB,$B2,$B8
	.dc.b	$78,$7F,$EE,$EB,$DE,$18,$AF,$F7,$7D,$55,$3A,$57,$D1,$F1,$98,$BE
	.dc.b	$98,$78,$5F,$F7,$CC,$2F,$A5,$F3,$BE,$53,$93,$91,$EB,$BE,$4F,$90
	.dc.b	$C9,$E4,$72,$FB,$4C,$CF,$5B,$F0,$DF,$82,$B7,$5E,$8B,$F8,$B7,$B8
	.dc.b	$ED,$18,$9D,$C7,$96,$31,$87,$C5,$8E,$58,$7C,$08,$C4,$6E,$30,$FA
	.dc.b	$0F,$8E,$7C,$43,$E0,$17,$0C,$3A,$3F,$93,$8F,$A9,$7D,$67,$FA,$06
	.dc.b	$3E,$6C,$63,$DF,$E8,$D8,$CC,$ED,$E6,$73,$FA,$B7,$07,$06,$C7,$46
	.dc.b	$BC,$29,$18,$8F,$2C,$62,$7F,$11,$F5,$8F,$4F,$D3,$AF,$E4,$A3,$17
	.dc.b	$F8,$DB,$E0,$FD,$C6,$CE,$EB,$8E,$9F,$0A,$F8,$DE,$F1,$F3,$DF,$E7
	.dc.b	$B0,$E7,$E0,$F1,$AF,$E9,$ED,$7F,$D6,$38,$7A,$2F,$DF,$3C,$5E,$71
	.dc.b	$DF,$1F,$0D,$7F,$58,$C7,$C3,$8E,$F2,$F5,$61,$C2,$B5,$7C,$15,$F0
	.dc.b	$C3,$C1,$7C,$0F,$8F,$5B,$9D,$57,$F6,$6C,$BE,$1D,$F0,$8F,$D0,$CE
	.dc.b	$9E,$B5,$DE,$B7,$1F,$DB,$B0,$7C,$B7,$E2,$B1,$FF,$31,$46,$27,$FC
	.dc.b	$D4,$D7,$F0,$7B,$FE,$43,$F3,$6F,$C7,$C6,$AF,$F2,$B5,$B8,$FC,$5B
	.dc.b	$5F,$CB,$B5,$FE,$FF,$F3,$CF,$9D,$FC,$EB,$63,$F4,$4D,$7F,$83,$B5
	.dc.b	$FC,$BB,$9C,$7F,$95,$F0,$BE,$F3,$FD,$6D,$5F,$EE,$98,$C7,$C3,$FB
	.dc.b	$F4,$FD,$4E,$1C,$5B,$19,$AF,$2B,$54,$F4,$7E,$19,$FE,$EF,$E4,$7C
	.dc.b	$CE,$E7,$53,$E6,$AF,$69,$1F,$6B,$DF,$0F,$47,$15,$CB,$F9,$AE,$7F
	.dc.b	$E9,$5C,$3C,$0F,$F4,$8E,$87,$E8,$6F,$A6,$F4,$46,$1B,$19,$CF,$FC
	.dc.b	$09,$F5,$18,$9F,$E1,$D5,$B9,$FE,$66,$D7,$F8,$1B,$5F,$E1,$EE,$6F
	.dc.b	$FA,$CF,$83,$FE,$05,$FA,$3F,$90,$DA,$FF,$37,$E3,$E1,$7F,$CD,$DF
	.dc.b	$A1,$3E,$7B,$CE,$D8,$FF,$3A,$B5,$FF,$17,$FD,$BE,$A7,$CF,$39,$9F
	.dc.b	$1B,$5F,$EC,$ED,$7F,$B6,$3C,$F1,$FF,$19,$FC,$E7,$9B,$B8,$F5,$AF
	.dc.b	$F4,$2F,$E4,$1F,$69,$BC,$BF,$32,$31,$1F,$E5,$EE,$DC,$4F,$CA,$9C
	.dc.b	$CF,$C6,$6B,$FD,$E9,$AF,$F7,$F7,$9E,$1F,$2B,$BD,$F4,$5F,$29,$B5
	.dc.b	$E2,$3F,$D3,$3F,$50,$FA,$0F,$BE,$D8,$CC,$7E,$67,$D0,$7B,$7F,$E0
	.dc.b	$09,$C9,$F7,$AB,$83,$F7,$9A,$FD,$E6,$BF,$E1,$27,$78,$FE,$03,$DC
	.dc.b	$7E,$15,$F8,$06,$D7,$63,$FD,$59,$ED,$3E,$AF,$F2,$AD,$8F,$F5,$9B
	.dc.b	$5F,$F4,$87,$AF,$A9,$FC,$AC,$E6,$7F,$22,$D7,$FC,$4A,$D7,$FC,$55
	.dc.b	$F0,$4F,$C0,$3F,$3A,$D8,$FC,$EB,$5F,$CE,$B5,$FC,$EF,$C2,$D7,$FC
	.dc.b	$62,$D7,$FC,$6C,$DE,$DF,$F6,$75,$7E,$AB,$E4,$FE,$0D,$7E,$A1,$AF
	.dc.b	$EA,$1A,$FF,$8F,$1A,$FF,$90,$5F,$C2,$C9,$A4,$E5,$FE,$4B,$E1,$FF
	.dc.b	$0D,$25,$7F,$AA,$37,$FE,$0D,$F6,$78,$8F,$E4,$F3,$C1,$FE,$56,$D7
	.dc.b	$F8,$C9,$FC,$73,$F0,$0F,$E3,$07,$1A,$FF,$2A,$F2,$3E,$45,$F2,$5F
	.dc.b	$BE,$FA,$E7,$CB,$D3,$F2,$13,$AC,$7C,$9E,$6F,$FA,$82,$31,$1F,$CA
	.dc.b	$3C,$55,$F2,$8D,$9F,$E9,$F7,$0F,$F2,$EC,$7C,$58,$FF,$0E,$FA,$57
	.dc.b	$8B,$49,$5C,$AB,$D1,$DF,$88,$F0,$57,$07,$BE,$28,$C4,$77,$CB,$F6
	.dc.b	$2F,$F3,$EF,$5D,$F8,$E7,$FB,$86,$3C,$F1,$8E,$9F,$E9,$0D,$8F,$5F
	.dc.b	$C1,$BF,$1B,$E7,$C5,$FE,$80,$AD,$57,$F8,$C1,$AF,$F9,$C7,$0E,$CE
	.dc.b	$5A,$E7,$0F,$EA,$D8,$75,$F9,$7B,$89,$F1,$DF,$2C,$7E,$DF,$D0,$7F
	.dc.b	$03,$C7,$E9,$7D,$57,$FB,$8D,$F8,$6D,$7E,$84,$6E,$78,$FB,$BF,$CF
	.dc.b	$AB,$51,$FD,$09,$BC,$BF,$ED,$EA,$E0,$FF,$BF,$38,$1F,$54,$FB,$6F
	.dc.b	$E8,$ED,$CB,$E9,$B5,$F0,$FA,$FA,$99,$1A,$7F,$8F,$61,$FD,$93,$0F
	.dc.b	$EF,$7E,$07,$F7,$12,$57,$E3,$0F,$EF,$3F,$1B,$87,$FC,$17,$1F,$E1
	.dc.b	$02,$3F,$E1,$07,$FD,$49,$F4,$7F,$E6,$EF,$A1,$7C,$96,$28,$23,$2F
	.dc.b	$FA,$CA,$7C,$6D,$1D,$4C,$BF,$DF,$31,$9D,$BF,$C1,$BE,$17,$56,$C7
	.dc.b	$1A,$E1,$F1,$BE,$16,$C7,$1F,$6B,$FA,$AF,$BA,$CB,$E8,$46,$1F,$A9
	.dc.b	$7E,$9E,$9F,$CA,$7A,$77,$AB,$FC,$D3,$87,$1C,$7F,$E0,$04,$63,$0F
	.dc.b	$7D,$FE,$6F,$E5,$F8,$61,$2F,$DD,$3D,$1B,$6A,$F9,$11,$F1,$FD,$8B
	.dc.b	$C1,$FD,$FF,$2F,$04,$E4,$F9,$07,$CD,$F9,$DD,$D7,$FA,$83,$97,$FD
	.dc.b	$2F,$7F,$E0,$76,$C6,$6B,$EC,$35,$FA,$F3,$DF,$4F,$B9,$E1,$78,$8E
	.dc.b	$B3,$EA,$F7,$5F,$35,$F1,$D7,$D0,$53,$11,$27,$84,$65,$E5,$1F,$3F
	.dc.b	$EB,$3C,$10,$E0,$46,$1B,$9E,$F5,$E0,$CA,$F9,$B5,$B1,$3F,$33,$B5
	.dc.b	$F1,$AF,$A8,$F9,$31,$86,$C6,$6F,$FD,$E1,$B7,$E7,$7A,$79,$4E,$78
	.dc.b	$BE,$9E,$1C,$6F,$67,$E8,$1C,$47,$D8,$E1,$FC,$74,$E6,$B9,$61,$FE
	.dc.b	$D6,$BD,$B6,$33,$3F,$53,$0F,$0E,$3E,$62,$47,$80,$FD,$74,$69,$C5
	.dc.b	$F4,$7C,$13,$F8,$8C,$3C,$71,$FC,$42,$FB,$B6,$7A,$E3,$E9,$C3,$A3
	.dc.b	$F0,$2B,$78,$75,$BF,$09,$C3,$1F,$56,$16,$3D,$45,$3B,$1E,$6A,$F0
	.dc.b	$E3,$E4,$BE,$84,$8C,$E3,$E2,$C3,$AD,$F1,$C3,$D9,$39,$8F,$79,$93
	.dc.b	$E3,$61,$D6,$FD,$24,$AE,$33,$E5,$C3,$C1,$18,$8E,$58,$FC,$18,$F5
	.dc.b	$5F,$37,$85,$F9,$AA,$C6,$B0,$F2,$47,$C0,$BE,$4F,$92,$33,$87,$82
	.dc.b	$30,$72,$3C,$38,$FE,$4D,$C5,$72,$3E,$B3,$88,$F8,$B1,$88,$E2,$7C
	.dc.b	$47,$84,$73,$8C,$9E,$57,$89,$98,$F3,$C7,$F3,$C3,$E6,$94,$FB,$65
	.dc.b	$9F,$D2,$37,$DA,$7C,$CE,$E3,$F4,$D1,$87,$13,$FE,$79,$9D,$47,$F6
	.dc.b	$49,$D3,$CE,$3F,$73,$1C,$6B,$F9,$2F,$85,$FE,$A9,$E5,$CD,$4B,$FD
	.dc.b	$1D,$EF,$1F,$55,$EE,$73,$5B,$C3,$F7,$B7,$B9,$FE,$B7,$5B,$9C,$D6
	.dc.b	$EF,$F7,$57,$BB,$5E,$E7,$35,$B9,$F8,$2F,$F3,$A9,$D4,$60,$F1,$92
	.dc.b	$FD,$BC,$0C,$79,$4F,$C5,$BB,$7C,$A3,$07,$1C,$3C,$2F,$F0,$78,$D6
	.dc.b	$5F,$3E,$34,$78,$71,$EF,$87,$69,$D4,$7C,$09,$E5,$8F,$8D,$E2,$78
	.dc.b	$23,$47,$27,$33,$CB,$0F,$1C,$E4,$DB,$82,$4F,$D0,$AE,$D3,$A8,$F3
	.dc.b	$4E,$9F,$9F,$3A,$8C,$57,$79,$D9,$CE,$FD,$99,$7D,$79,$D4,$F2,$9C
	.dc.b	$E5,$F3,$30,$E9,$3C,$8F,$76,$1E,$1C,$BE,$5E,$3C,$C9,$7F,$76,$F8
	.dc.b	$DF,$03,$96,$3E,$0B,$E3,$5E,$09,$F9,$B8,$DC,$BE,$04,$79,$4F,$97
	.dc.b	$8F,$67,$66,$2B,$D1,$5F,$8F,$E1,$FF,$7D,$F9,$DE,$A6,$C7,$F8,$53
	.dc.b	$DE,$7D,$66,$5F,$F4,$CC,$66,$B0,$66,$6C,$7F,$87,$3C,$9F,$ED,$8D
	.dc.b	$F8,$0E,$D1,$8C,$BF,$B2,$E3,$FE,$71,$9C,$F1,$69,$B7,$C4,$FD,$27
	.dc.b	$17,$C8,$DF,$FC,$F1,$18,$3F,$51,$8E,$67,$F2,$51,$FC,$4E,$51,$8C
	.dc.b	$7F,$7E,$D7,$F7,$6D,$7F,$6E,$F3,$9F,$DA,$7D,$DD,$9F,$4F,$D3,$7F
	.dc.b	$B4,$67,$FC,$76,$D8,$CF,$B7,$F3,$63,$1C,$5F,$E4,$7D,$51,$8E,$5F
	.dc.b	$F2,$3B,$6F,$FC,$40,$AF,$97,$B1,$F9,$0F,$13,$FC,$C9,$EE,$7E,$4B
	.dc.b	$C6,$FF,$2D,$7B,$39,$F2,$13,$9C,$D3,$98,$C3,$76,$FF,$36,$4E,$9F
	.dc.b	$E2,$5F,$6D,$F3,$FD,$2F,$A0,$D8,$C9,$FC,$D3,$8D,$1F,$C0,$77,$FA
	.dc.b	$0D,$7F,$96,$B5,$FE,$52,$FF,$07,$F4,$38,$4F,$2A,$FF,$49,$3F,$EE
	.dc.b	$A6,$CE,$DF,$F5,$37,$80,$FB,$9C,$60,$DB,$FD,$1D,$18,$DF,$FD,$15
	.dc.b	$F1,$7A,$9C,$E1,$FE,$DC,$C3,$93,$6F,$87,$02,$F8,$56,$B1,$FF,$5F
	.dc.b	$F9,$FC,$CF,$F2,$6E,$1F,$EE,$BB,$46,$7C,$4E,$6D,$9D,$36,$BF,$DD
	.dc.b	$1C,$8F,$37,$D7,$CF,$F8,$6F,$43,$DB,$EA,$FE,$0E,$73,$D2,$9C,$DF
	.dc.b	$1C,$BF,$B6,$5E,$DB,$19,$C7,$FD,$D4,$D7,$FB,$D4,$ED,$F1,$E1,$DB
	.dc.b	$2E,$AF,$C9,$FA,$5F,$D5,$E1,$FE,$BC,$E5,$FE,$81,$C4,$9C,$E0,$9F
	.dc.b	$E8,$9C,$6F,$5F,$FB,$74,$70,$AD,$B7,$BB,$B1,$F6,$67,$78,$FF,$B9
	.dc.b	$6B,$6F,$FB,$32,$FA,$73,$7D,$DF,$A5,$7A,$FC,$63,$0D,$AF,$41,$FE
	.dc.b	$EF,$9D,$B8,$CF,$FD,$A3,$8F,$E4,$DF,$54,$6E,$BF,$1F,$1F,$DF,$3E
	.dc.b	$8E,$F5,$7F,$B2,$EB,$F3,$31,$8E,$CC,$BF,$A1,$AE,$13,$FA,$98,$C5
	.dc.b	$1A,$70,$8F,$D1,$F8,$76,$BC,$63,$E0,$1F,$9D,$FA,$A3,$FC,$65,$18
	.dc.b	$CB,$F6,$0D,$BF,$F1,$83,$67,$75,$F8,$B9,$CB,$63,$3C,$FC,$A3,$11
	.dc.b	$FE,$41,$A7,$C1,$7F,$10,$BE,$AD,$9F,$1B,$5E,$0D,$47,$AF,$E0,$DF
	.dc.b	$6B,$E8,$6D,$7A,$1B,$3E,$E6,$BE,$3F,$CD,$FF,$D7,$B3,$EE,$6C,$7C
	.dc.b	$46,$BB,$F4,$BE,$2C,$6B,$9F,$FC,$35,$92,$31,$7F,$8F,$F8,$97,$D6
	.dc.b	$7F,$10,$AD,$61,$7F,$37,$FE,$BD,$8F,$E9,$3E,$25,$F3,$F9,$F4,$3F
	.dc.b	$89,$46,$33,$FC,$86,$2C,$3F,$B3,$6B,$F3,$5A,$FC,$68,$D5,$FD,$C9
	.dc.b	$CC,$FE,$0D,$F0,$D6,$DB,$0E,$B6,$AF,$89,$F2,$5F,$8B,$F1,$FF,$E4
	.dc.b	$0F,$E8,$FC,$07,$D3,$7F,$1C,$8D,$79,$7F,$CA,$E3,$19,$7E,$53,$56
	.dc.b	$1C,$A3,$F2,$6D,$7F,$10,$D7,$D8,$D7,$F3,$8F,$1E,$0F,$A6,$DC,$3F
	.dc.b	$80,$7D,$3B,$E7,$7E,$94,$E6,$FF,$49,$FA,$13,$E2,$3F,$A0,$77,$7D
	.dc.b	$8C,$7E,$7C,$E5,$F7,$46,$6B,$D5,$39,$7F,$56,$D7,$F4,$0D,$7E,$A3
	.dc.b	$5F,$0D,$7E,$16,$39,$65,$FC,$8B,$E9,$DF,$31,$9A,$FC,$57,$E8,$6F
	.dc.b	$93,$FE,$AD,$E8,$3D,$EC,$C6,$1B,$AF,$EE,$AB,$83,$67,$F7,$F1,$8D
	.dc.b	$3F,$81,$B7,$3F,$CE,$B6,$BF,$95,$7B,$1D,$DF,$C9,$C6,$6B,$F6,$BF
	.dc.b	$97,$FF,$91,$27,$F9,$37,$CE,$3F,$11,$F6,$C7,$FB,$0E,$75,$5F,$C5
	.dc.b	$AB,$5A,$FF,$28,$9C,$CF,$FB,$1D,$AF,$1F,$80,$FE,$4D,$18,$8F,$0C
	.dc.b	$71,$3F,$BF,$D7,$F4,$DD,$71,$87,$F3,$9A,$5F,$F9,$61,$F1,$63,$F8
	.dc.b	$57,$3F,$03,$CA,$31,$3B,$9F,$24,$F8,$89,$5F,$CA,$3F,$2D,$7C,$87
	.dc.b	$F2,$C6,$C6,$63,$04,$7F,$7B,$1A,$AF,$E7,$23,$EF,$9F,$88,$FF,$0B
	.dc.b	$9D,$70,$FF,$89,$FE,$0F,$F9,$2F,$CD,$BE,$6B,$FB,$BF,$A3,$FE,$AF
	.dc.b	$8C,$65,$FE,$F1,$AD,$D6,$B3,$B3,$91,$5C,$17,$B9,$FD,$26,$31,$1F
	.dc.b	$CD,$2F,$8B,$AC,$3C,$18,$7F,$BE,$67,$37,$C6,$31,$5D,$9F,$13,$4F
	.dc.b	$15,$F8,$67,$84,$62,$77,$3E,$28,$57,$F4,$6E,$CE,$18,$7F,$B3,$DD
	.dc.b	$46,$6F,$BC,$62,$BC,$07,$A7,$1F,$5B,$ED,$33,$8F,$89,$FE,$A6,$78
	.dc.b	$0C,$E1,$F0,$1F,$31,$FD,$62,$38,$3C,$CE,$27,$2C,$2B,$F4,$4A,$75
	.dc.b	$37,$F4,$9F,$C6,$F9,$32,$D8,$E2,$D7,$7D,$E9,$DF,$D1,$7F,$8B,$E7
	.dc.b	$6D,$7A,$B5,$F6,$D2,$31,$1F,$DC,$A3,$1B,$7E,$03,$E9,$5F,$2D,$F1
	.dc.b	$38,$3F,$11,$18,$BF,$C2,$DF,$07,$E7,$B6,$77,$5C,$78,$3E,$6D,$F1
	.dc.b	$BD,$E3,$E9,$BF,$CA,$61,$C6,$F7,$87,$1B,$FF,$03,$E1,$C6,$F7,$87
	.dc.b	$1B,$F6,$B6,$B7,$7C,$55,$F1,$AD,$E3,$EA,$BE,$15,$AB,$E1,$7B,$8F
	.dc.b	$8D,$3A,$C7,$CD,$1D,$8F,$91,$5D,$E3,$30,$6D,$73,$8E,$C7,$86,$FF
	.dc.b	$5F,$6F,$A4,$FC,$9C,$7F,$AB,$63,$13,$FC,$7D,$AF,$F0,$CF,$A5,$FE
	.dc.b	$C8,$E6,$B8,$F0,$FF,$8B,$BC,$07,$D6,$FF,$BF,$AB,$FC,$C1,$18,$ED
	.dc.b	$DC,$7F,$8D,$31,$EB,$97,$FA,$F7,$E6,$BF,$D4,$93,$FD,$0B,$E9,$9F
	.dc.b	$39,$C3,$8D,$18,$9F,$E6,$95,$C1,$FE,$42,$DB,$69,$FE,$42,$BE,$38
	.dc.b	$73,$C3,$96,$1C,$3C,$CF,$2A,$C7,$38,$F9,$6F,$31,$8C,$7A,$73,$3E
	.dc.b	$13,$35,$D7,$96,$63,$E6,$8C,$BE,$D8,$CB,$E8,$8F,$8B,$E6,$3E,$97
	.dc.b	$FB,$DC,$FF,$8E,$9A,$FF,$88,$DB,$9F,$FA,$8B,$C7,$7C,$D7,$29,$CE
	.dc.b	$9F,$E6,$2F,$41,$C8,$FA,$6F,$A7,$18,$78,$79,$93,$3F,$F1,$B5,$6D
	.dc.b	$B3,$97,$FC,$69,$F4,$CF,$99,$F4,$B6,$3F,$CA,$2D,$7F,$D1,$BE,$B3
	.dc.b	$D0,$7D,$17,$0B,$F1,$F8,$DF,$E1,$5B,$FF,$14,$F9,$D9,$79,$CE,$A2
	.dc.b	$46,$0F,$F2,$FF,$D3,$7D,$36,$F3,$FC,$C8,$C4,$78,$BD,$5F,$E0,$DC
	.dc.b	$BF,$E7,$49,$CB,$FE,$D0,$F1,$9F,$3F,$F1,$9B,$7D,$08,$7F,$AA,$38
	.dc.b	$FF,$B0,$CE,$CF,$06,$3F,$73,$E9,$9F,$39,$49,$7F,$EC,$5E,$1F,$F3
	.dc.b	$4F,$99,$88,$E9,$EF,$7F,$8E,$5F,$F3,$43,$5F,$F3,$25,$F4,$7F,$CF
	.dc.b	$ED,$78,$FA,$FF,$7A,$78,$6B,$FE,$92,$AD,$3F,$E9,$4F,$97,$7C,$97
	.dc.b	$E4,$9B,$87,$FA,$29,$AF,$DA,$F7,$9D,$FB,$AF,$CC,$FC,$2B,$E4,$A9
	.dc.b	$9B,$FF,$49,$C6,$5E,$9E,$0F,$73,$6F,$FA,$63,$E2,$34,$FF,$A0,$7E
	.dc.b	$67,$D3,$E1,$7F,$A5,$DB,$1F,$ED,$79,$DE,$BF,$EA,$E9,$CF,$37,$FB
	.dc.b	$E3,$D9,$FC,$FF,$1F,$F8,$BB,$0F,$F5,$2D,$72,$F1,$4F,$E0,$B2,$FD
	.dc.b	$47,$B2,$F9,$8F,$D3,$B6,$F8,$F4,$FF,$04,$F2,$64,$7F,$14,$7F,$83
	.dc.b	$FC,$CE,$23,$F8,$AF,$C2,$7F,$0A,$31,$BF,$F2,$1A,$C4,$FF,$1C,$9C
	.dc.b	$E5,$FE,$DA,$EB,$7C,$8B,$E9,$7F,$73,$F4,$CF,$80,$A7,$F2,$EA,$DC
	.dc.b	$DA,$E1,$8F,$FA,$C6,$F8,$36,$F5,$5F,$C1,$3E,$9B,$F6,$B3,$FE,$B7
	.dc.b	$6B,$F9,$F6,$E9,$FC,$BA,$3F,$D8,$58,$7E,$99,$D7,$93,$CC,$8F,$E7
	.dc.b	$E3,$53,$C7,$DF,$7D,$33,$C0,$FE,$8F,$18,$6C,$66,$FF,$A6,$FD,$37
	.dc.b	$2D,$FF,$C5,$13,$99,$61,$E5,$F4,$16,$FA,$0F,$EA,$71,$87,$AE,$32
	.dc.b	$FA,$C6,$AB,$84,$0E,$B8,$F8,$2B,$E8,$7C,$F3,$AF,$F9,$6F,$CB,$BE
	.dc.b	$BA,$9A,$89,$87,$FB,$8E,$74,$D9,$CB,$FC,$37,$E8,$FF,$DA,$0F,$FB
	.dc.b	$A5,$AE,$7E,$2F,$C3,$EC,$FF,$3C,$9C,$EF,$F5,$7C,$24,$63,$5F,$E8
	.dc.b	$8D,$C3,$FD,$F5,$A7,$F5,$D7,$DD,$ED,$FE,$B2,$BF,$CB,$2D,$8C,$F8
	.dc.b	$FF,$D9,$3E,$63,$FB,$2B,$FE,$55,$F9,$AF,$F2,$FF,$D0,$FE,$BE,$BF
	.dc.b	$CD,$F1,$8F,$1F,$51,$C8,$FF,$42,$47,$6A,$CC,$2B,$97,$BC,$AE,$53
	.dc.b	$C0,$E4,$76,$3D,$7F,$3F,$DB,$6F,$EF,$9E,$2D,$D3,$FA,$1E,$1C,$EF
	.dc.b	$67,$3F,$09,$BB,$1E,$18,$4A,$D4,$E5,$F1,$75,$7E,$3B,$E2,$BD,$AD
	.dc.b	$8C,$F5,$7F,$95,$72,$F0,$4F,$E5,$3E,$9F,$73,$9D,$BF,$C5,$8D,$BF
	.dc.b	$EF,$2D,$AE,$11,$F0,$4F,$8D,$ED,$FC,$8E,$0F,$EE,$18,$FB,$67,$26
	.dc.b	$E1,$F4,$5B,$AE,$1E,$C4,$AD,$46,$5E,$11,$97,$E7,$1D,$0F,$A1,$87
	.dc.b	$F8,$4B,$E9,$BE,$0F,$A2,$FA,$7A,$62,$25,$7C,$89,$D3,$67,$2F,$4F
	.dc.b	$A7,$F5,$ED,$FE,$4F,$6C,$7F,$87,$1B,$81,$F6,$C7,$D4,$72,$4C,$BE
	.dc.b	$AB,$A8,$F5,$F5,$BE,$71,$F3,$7F,$36,$30,$D8,$CC,$FC,$2D,$FF,$17
	.dc.b	$97,$FA,$FA,$F7,$3D,$2B,$7C,$1F,$4E,$72,$FD,$5B,$E3,$5F,$C3,$67
	.dc.b	$37,$CE,$F8,$B7,$0F,$B5,$CB,$FA,$6A,$DD,$F1,$BD,$B6,$B6,$FA,$58
	.dc.b	$F9,$5C,$B4,$98,$F9,$E7,$9C,$7E,$CE,$FC,$11,$8B,$F0,$CF,$F0,$E8
	.dc.b	$DE,$3E,$C8,$C6,$1E,$0C,$3A,$3E,$06,$C7,$83,$1F,$73,$F3,$71,$EF
	.dc.b	$5C,$67,$35,$F2,$67,$35,$C4,$F8,$07,$13,$81,$88,$EE,$71,$72,$79
	.dc.b	$8E,$47,$BC,$CE,$3E,$E3,$DF,$87,$B5,$C6,$3F,$0C,$F8,$37,$E4,$BE
	.dc.b	$F8,$7A,$0F,$E4,$B1,$97,$18,$79,$5C,$BE,$0A,$E9,$8F,$C7,$9D,$CF
	.dc.b	$9E,$FA,$BF,$D5,$71,$F2,$5F,$08,$F5,$C6,$2F,$A1,$CE,$3C,$04,$3A
	.dc.b	$5F,$89,$F8,$78,$FB,$A7,$84,$68,$E1,$1F,$13,$1F,$A7,$1F,$D9,$CF
	.dc.b	$A6,$53,$E7,$16,$3F,$4E,$DF,$75,$F4,$FB,$8F,$D2,$46,$1C,$4F,$F9
	.dc.b	$B2,$75,$1F,$D7,$A7,$4F,$38,$FD,$D4,$71,$CF,$F9,$27,$89,$FE,$99
	.dc.b	$F0,$33,$52,$FF,$81,$DE,$F3,$F5,$5E,$E7,$35,$BC,$3F,$79,$7B,$9F
	.dc.b	$F0,$35,$6E,$73,$5B,$BF,$E0,$D7,$BB,$5E,$E7,$35,$B9,$F8,$2F,$F3
	.dc.b	$F9,$D4,$60,$F1,$92,$FD,$BB,$FF,$8D,$5F,$0C,$66,$FA,$BF,$CE,$AB
	.dc.b	$9B,$F2,$6F,$C0,$D3,$57,$CE,$B8,$DF,$42,$3F,$1A,$35,$3E,$59,$CB
	.dc.b	$67,$58,$F6,$CB,$ED,$4F,$94,$F9,$37,$C9,$F1,$5F,$26,$DF,$27,$81
	.dc.b	$83,$B1,$CA,$FB,$65,$F3,$F0,$EF,$7F,$1A,$73,$8F,$C0,$CB,$E1,$D6
	.dc.b	$B1,$F7,$93,$1E,$87,$82,$34,$F8,$0F,$B7,$97,$A6,$33,$1E,$52,$57
	.dc.b	$6C,$BC,$84,$C3,$EF,$5F,$18,$C5,$7B,$AB,$85,$7A,$23,$EA,$E3,$72
	.dc.b	$F8,$D5,$A9,$F3,$5F,$E3,$E3,$13,$F2,$F1,$F2,$C6,$32,$FF,$2F,$70
	.dc.b	$3E,$0B,$FD,$1D,$97,$F8,$79,$B9,$7F,$80,$17,$C6,$FE,$02,$B8,$7A
	.dc.b	$16,$77,$9F,$F8,$81,$B8,$F6,$78,$E3,$D6,$B9,$79,$5F,$DC,$A3,$FC
	.dc.b	$5A,$BE,$13,$3C,$5A,$6C,$E9,$BA,$7F,$85,$2B,$84,$FE,$A7,$8B,$F1
	.dc.b	$9B,$7F,$A0,$E3,$07,$EB,$67,$33,$A6,$C6,$4D,$D7,$EA,$67,$9F,$D3
	.dc.b	$EA,$3F,$C4,$7F,$4F,$FE,$CC,$AF,$F2,$2B,$63,$3E,$F7,$C7,$8C,$6F
	.dc.b	$FE,$44,$6E,$FF,$49,$B4,$E0,$47,$FB,$E3,$0E,$BE,$DF,$EC,$3C,$4F
	.dc.b	$F2,$CF,$BF,$FB,$2F,$1F,$FC,$D1,$ED,$67,$C6,$FF,$7C,$E1,$FE,$69
	.dc.b	$9C,$BF,$E5,$CA,$E1,$A2,$F8,$46,$1F,$93,$F4,$DF,$2B,$D2,$B7,$FC
	.dc.b	$55,$BB,$7F,$0D,$6E,$FF,$CC,$1C,$F9,$DF,$EC,$99,$FE,$43,$7F,$EC
	.dc.b	$B8,$C7,$2F,$FB,$7F,$0F,$F5,$84,$62,$75,$1F,$EB,$3D,$3F,$D0,$DE
	.dc.b	$EE,$30,$46,$27,$89,$FE,$EB,$D9,$F1,$BB,$9D,$57,$FA,$F0,$E3,$E0
	.dc.b	$FF,$C0,$1B,$E3,$18,$C3,$B7,$0F,$FA,$1B,$0E,$57,$BA,$E5,$3F,$E8
	.dc.b	$AF,$3B,$B1,$FE,$70,$C3,$9C,$65,$EB,$8A,$75,$5C,$B2,$C5,$6A,$73
	.dc.b	$A7,$FB,$C7,$0E,$57,$E9,$C3,$97,$93,$F0,$FD,$67,$47,$FA,$6F,$2F
	.dc.b	$5F,$47,$E2,$BD,$1D,$4E,$71,$F8,$53,$9A,$F8,$37,$CB,$CF,$F9,$D3
	.dc.b	$9A,$F9,$13,$9A,$EB,$D4,$C7,$AE,$12,$73,$B5,$C3,$D1,$F4,$EE,$DF
	.dc.b	$A8,$DA,$F8,$FE,$C7,$F7,$1E,$1F,$F6,$E4,$E6,$F8,$DF,$03,$EA,$D6
	.dc.b	$EF,$F0,$D5,$BB,$F9,$12,$E6,$FC,$17,$D3,$FF,$BE,$6F,$BC,$63,$D8
	.dc.b	$E6,$D8,$C9,$F6,$27,$86,$5F,$89,$3F,$BD,$FD,$3F,$E4,$DF,$EF,$BD
	.dc.b	$F7,$AD,$36,$30,$DB,$FC,$7D,$72,$6D,$7F,$B8,$5B,$3C,$1C,$F8,$76
	.dc.b	$BF,$35,$F0,$4F,$A7,$F2,$EC,$BF,$D4,$7D,$8B,$D7,$FD,$45,$EF,$7F
	.dc.b	$D2,$4E,$6B,$F4,$33,$CF,$97,$F5,$14,$F9,$DE,$39,$7D,$FB,$FC,$DC
	.dc.b	$E5,$E3,$F0,$DF,$C4,$34,$F2,$D6,$DF,$3C,$67,$4F,$34,$32,$F0,$5F
	.dc.b	$8B,$E1,$DF,$51,$E8,$FA,$87,$C2,$7A,$A7,$35,$E8,$9C,$B7,$D6,$D5
	.dc.b	$F1,$6C,$64,$98,$FE,$EB,$E2,$9F,$51,$B9,$F8,$7F,$4F,$FD,$56,$BF
	.dc.b	$61,$F1,$CF,$A8,$F9,$73,$9D,$FD,$7E,$3A,$B5,$87,$2F,$53,$F8,$94
	.dc.b	$63,$1F,$B3,$87,$1A,$FE,$3D,$39,$C3,$E2,$F0,$AB,$57,$F1,$DB,$E4
	.dc.b	$B5,$7C,$67,$93,$3F,$BF,$F4,$FF,$C8,$B4,$FE,$23,$F5,$1F,$C3,$B2
	.dc.b	$FC,$4C,$E5,$FE,$19,$19,$D3,$C3,$39,$7D,$B1,$9A,$57,$D9,$AD,$47
	.dc.b	$F3,$79,$D4,$E7,$CF,$F1,$9F,$CB,$3E,$55,$F4,$78,$BF,$C8,$FD,$43
	.dc.b	$E3,$3F,$9C,$78,$5F,$52,$73,$18,$E6,$FC,$DD,$6E,$7F,$96,$D6,$F2
	.dc.b	$FD,$2C,$E5,$FE,$51,$19,$D7,$E6,$CE,$5B,$87,$F4,$0F,$98,$7C,$DF
	.dc.b	$68,$FD,$4F,$D4,$3E,$63,$FA,$67,$A7,$FE,$29,$D3,$F5,$33,$97,$F9
	.dc.b	$FC,$67,$4F,$DB,$CE,$5F,$E7,$51,$9C,$BE,$FC,$E5,$FC,$5C,$66,$BE
	.dc.b	$7F,$42,$B5,$18,$6D,$B9,$DE,$E7,$18,$FD,$74,$E9,$7D,$3F,$F8,$D1
	.dc.b	$FE,$C9,$F4,$0F,$A8,$F8,$98,$FF,$AF,$6B,$5A,$7E,$E2,$B5,$B7,$F2
	.dc.b	$19,$CD,$7E,$12,$72,$FF,$1E,$8C,$E0,$9C,$BF,$CA,$E3,$2F,$93,$E1
	.dc.b	$7F,$BF,$4B,$0F,$54,$E6,$FA,$7C,$26,$B2,$FC,$8C,$FE,$53,$1F,$F2
	.dc.b	$CD,$F9,$E7,$33,$F0,$2B,$87,$47,$A5,$B8,$7F,$22,$FA,$87,$C9,$FF
	.dc.b	$62,$FD,$1F,$E1,$E1,$FC,$5E,$73,$5F,$DB,$A7,$3C,$49,$CD,$7F,$36
	.dc.b	$6E,$7F,$E2,$3F,$9C,$FE,$51,$39,$C7,$FA,$53,$FA,$CF,$84,$FD,$6F
	.dc.b	$CF,$3E,$77,$FD,$DD,$F0,$1F,$D4,$A3,$15,$FE,$EF,$AD,$D6,$BD,$1F
	.dc.b	$E5,$B5,$AA,$FF,$7B,$56,$9F,$DC,$DF,$3B,$FF,$79,$4E,$65,$39,$7B
	.dc.b	$0A,$E9,$19,$8E,$A6,$6B,$C5,$B7,$38,$F6,$19,$3C,$38,$79,$A7,$C3
	.dc.b	$19,$3C,$31,$97,$C3,$7E,$08,$C4,$78,$23,$15,$DE,$18,$7A,$F0,$F3
	.dc.b	$47,$8A,$B6,$F7,$31,$87,$A1,$C1,$C8,$F2,$4F,$27,$07,$F2,$23,$E2
	.dc.b	$94,$F8,$65,$3E,$09,$9F,$A9,$FE,$43,$C9,$98,$46,$1E,$0D,$78,$78
	.dc.b	$3F,$A9,$73,$F4,$BF,$A5,$6E,$FC,$E9,$39,$9F,$03,$5F,$44,$6A,$7F
	.dc.b	$01,$F5,$3F,$E0,$79,$BE,$C7,$B7,$4F,$C5,$5E,$E7,$EB,$36,$3F,$1D
	.dc.b	$5C,$67,$ED,$36,$38,$E1,$E7,$6E,$1D,$AA,$1D,$27,$5A,$79,$9B,$1F
	.dc.b	$D4,$2F,$A4,$FD,$16,$C7,$F5,$4C,$3C,$93,$C6,$FA,$E1,$E6,$C7,$D9
	.dc.b	$39,$BF,$05,$6A,$3A,$4E,$A7,$A6,$3E,$27,$EB,$B7,$1F,$0C,$1B,$8F
	.dc.b	$8A,$3F,$D2,$FC,$6F,$88,$FC,$46,$5F,$E0,$88,$C4,$FF,$A1,$9A,$FF
	.dc.b	$9D,$BE,$9F,$FD,$E2,$E6,$B8,$F0,$FF,$A9,$6B,$77,$CE,$73,$5F,$EC
	.dc.b	$BB,$E2,$FF,$92,$3F,$1A,$79,$7A,$C1,$18,$9F,$F2,$1B,$5F,$E8,$BF
	.dc.b	$50,$F9,$EE,$13,$9C,$BF,$C2,$ED,$CF,$FA,$BB,$6F,$FA,$05,$F1,$7F
	.dc.b	$C9,$FF,$5F,$FD,$B9,$BE,$3F,$8E,$31,$1E,$0E,$2C,$4C,$8D,$9F,$EB
	.dc.b	$8E,$55,$FF,$1F,$CB,$FC,$53,$87,$27,$FC,$6B,$F2,$DE,$3F,$50,$F9
	.dc.b	$76,$56,$F5,$FE,$DF,$F5,$2F,$B4,$F8,$CD,$D3,$FC,$CE,$D7,$FB,$DF
	.dc.b	$D5,$6A,$64,$69,$FA,$79,$7F,$BF,$67,$33,$FE,$BB,$87,$CE,$69,$B1
	.dc.b	$C4,$FF,$3A,$CB,$E6,$FE,$C3,$63,$31,$DB,$D1,$7C,$CE,$A2,$4E,$8F
	.dc.b	$F1,$3C,$61,$FF,$66,$7E,$4F,$F0,$5F,$F7,$6B,$76,$61,$DE,$7F,$D2
	.dc.b	$3F,$54,$FA,$1F,$C5,$36,$3F,$D1,$6D,$7F,$DB,$1E,$05,$E3,$FF,$2E
	.dc.b	$CD,$E1,$FF,$0E,$B6,$BF,$D0,$8D,$DB,$FA,$86,$1D,$8F,$97,$F7,$82
	.dc.b	$99,$AF,$F4,$DF,$D5,$BE,$EB,$F3,$CD,$8C,$C7,$D4,$F8,$0F,$F1,$E4
	.dc.b	$60,$87,$FA,$BA,$78,$3F,$E8,$2F,$5F,$9D,$7D,$CF,$CC,$3E,$2C,$FF
	.dc.b	$BF,$9A,$F1,$8F,$D7,$7C,$37,$FC,$01,$8F,$F5,$7F,$CE,$3E,$93,$F6
	.dc.b	$4D,$8F,$F5,$17,$CB,$E2,$64,$70,$79,$C7,$E3,$3F,$04,$7D,$35,$25
	.dc.b	$6C,$FF,$5E,$FD,$0F,$B9,$FF,$16,$71,$FF,$9E,$B8,$9E,$8F,$F8,$82
	.dc.b	$7F,$D7,$BE,$97,$F0,$5B,$E1,$CD,$FE,$94,$7F,$DF,$53,$D3,$E6,$B4
	.dc.b	$7E,$7A,$F9,$47,$61,$83,$D4,$7C,$E7,$E9,$1B,$3F,$E8,$CC,$7F,$D5
	.dc.b	$71,$88,$E5,$5C,$7C,$EF,$E3,$B0,$FC,$F1,$F4,$58,$6C,$7F,$BB,$9A
	.dc.b	$FF,$BE,$7E,$AB,$FD,$4E,$E6,$78,$11,$FF,$7A,$C6,$9F,$F8,$01,$3C
	.dc.b	$B1,$FF,$6C,$C7,$86,$B8,$3D,$7E,$2F,$F7,$FF,$5F,$FC,$C2,$BF,$D8
	.dc.b	$BF,$8B,$FF,$11,$3F,$89,$46,$78,$BE,$93,$6F,$E3,$CE,$FD,$3E,$7F
	.dc.b	$05,$FC,$2B,$E5,$3F,$6D,$BF,$FB,$93,$EB,$70,$67,$D3,$98,$FC,$BC
	.dc.b	$FE,$F5,$AB,$73,$93,$EA,$C7,$F0,$1F,$89,$7D,$63,$EA,$7F,$A5,$C7
	.dc.b	$D8,$F6,$AF,$39,$CD,$6F,$0E,$95,$CA,$F7,$52,$73,$5B,$7F,$0D,$5D
	.dc.b	$5E,$37,$D5,$B5,$DA,$FF,$95,$FC,$F7,$F6,$0C,$FF,$34,$DA,$FC,$7B
	.dc.b	$75,$3F,$06,$3E,$04,$63,$9F,$99,$23,$33,$B7,$F2,$3F,$55,$FB,$56
	.dc.b	$F8,$3F,$CF,$23,$1C,$AC,$7F,$59,$1F,$ED,$BA,$EE,$EA,$33,$5C,$4E
	.dc.b	$5F,$15,$F9,$CF,$A9,$FD,$4F,$C2,$BE,$87,$FC,$9D,$18,$6F,$85,$FE
	.dc.b	$7C,$7F,$CA,$5D,$DF,$C3,$73,$FF,$30,$C6,$9F,$F4,$63,$5C,$9F,$EF
	.dc.b	$AA,$E8,$FF,$BF,$27,$95,$F8,$63,$57,$DD,$A7,$42,$9D,$04,$F2,$BF
	.dc.b	$E1,$CF,$F0,$88,$F0,$FC,$7B,$EA,$5F,$37,$D6,$30,$F7,$E2,$E7,$E2
	.dc.b	$5C,$39,$67,$E5,$BD,$E3,$E3,$6B,$EE,$BE,$2F,$9F,$9F,$F3,$BF,$0F
	.dc.b	$EB,$7F,$89,$FE,$25,$7A,$4E,$7A,$BF,$CB,$99,$7C,$3A,$E3,$87,$3B
	.dc.b	$DB,$6F,$86,$1F,$1E,$1C,$3F,$DA,$70,$F9,$91,$8F,$4F,$53,$FD,$9E
	.dc.b	$72,$EF,$87,$F0,$3E,$D6,$AA,$5F,$D1,$3A,$C7,$23,$2F,$D3,$AE,$95
	.dc.b	$CA,$3E,$17,$2F,$F8,$39,$FC,$1E,$9F,$D4,$FD,$CF,$89,$39,$CB,$E0
	.dc.b	$3F,$D3,$AB,$8C,$E8,$8F,$4F,$2F,$E6,$7D,$6F,$A6,$18,$23,$0D,$CF
	.dc.b	$7F,$57,$C1,$95,$F5,$EB,$62,$7E,$67,$5B,$E7,$BF,$CB,$1A,$7C,$C8
	.dc.b	$C7,$8D,$F1,$5B,$87,$C3,$E8,$E3,$7C,$37,$FF,$5D,$56,$B0,$E5,$39
	.dc.b	$BF,$D3,$5F,$0C,$3A,$4E,$65,$7C,$70,$FD,$BD,$EE,$7F,$49,$39,$47
	.dc.b	$0A,$E5,$C1,$F9,$CB,$F1,$C7,$5A,$E1,$5B,$C3,$F0,$17,$B8,$E7,$3A
	.dc.b	$25,$74,$C3,$C5,$8F,$B3,$0F,$49,$CA,$33,$8F,$82,$33,$87,$C7,$2D
	.dc.b	$71,$7D,$77,$D5,$CD,$71,$73,$5D,$5C,$9E,$69,$F6,$1C,$5D,$92,$BB
	.dc.b	$1F,$A2,$BF,$2C,$71,$AE,$6F,$EB,$23,$2F,$C5,$39,$5F,$D2,$77,$87
	.dc.b	$C4,$AF,$24,$68,$E8,$53,$A1,$C3,$0F,$3C,$70,$C3,$CB,$7E,$F6,$D7
	.dc.b	$E1,$8E,$13,$E6,$C7,$EB,$47,$96,$76,$DA,$F3,$3E,$1C,$3C,$CF,$D1
	.dc.b	$31,$87,$89,$C5,$72,$9D,$BF,$1E,$3C,$27,$AE,$FC,$33,$FA,$28,$FC
	.dc.b	$89,$F0,$0A,$7C,$92,$BB,$7F,$4B,$F5,$BF,$EA,$87,$67,$E8,$1C,$3C
	.dc.b	$1A,$F0,$F8,$4D,$BB,$7F,$69,$1C,$EF,$F9,$1F,$83,$FE,$9D,$F0,$33
	.dc.b	$52,$FF,$53,$7B,$CF,$D3,$7B,$9C,$D6,$F0,$FE,$05,$7B,$9F,$D7,$D6
	.dc.b	$E7,$35,$BB,$F3,$57,$F5,$9A,$D4,$62,$B7,$3D,$8D,$70,$7F,$7F,$71
	.dc.b	$1A,$24,$F9,$A7,$51,$89,$FD,$BE,$1E,$57,$E2,$1D,$31,$F2,$3A,$C6
	.dc.b	$BF,$E1,$AB,$E7,$87,$94,$91,$B3,$91,$E4,$72,$F6,$C3,$E7,$CF,$28
	.dc.b	$E1,$1E,$57,$53,$E9,$BF,$0E,$4C,$7D,$0D,$C7,$FA,$96,$5F,$23,$0E
	.dc.b	$78,$78,$2B,$51,$E4,$C7,$B4,$61,$F0,$57,$19,$F0,$9C,$88,$7C,$1C
	.dc.b	$7C,$78,$F8,$8F,$95,$3C,$6F,$A6,$1E,$19,$CE,$5E,$FC,$3A,$92,$B6
	.dc.b	$FE,$1A,$35,$18,$8D,$4F,$2C,$3C,$39,$7D,$4C,$7C,$13,$E3,$8C,$61
	.dc.b	$CA,$31,$7F,$4C,$ED,$7C,$9C,$1E,$D2,$E1,$E1,$3C,$58,$76,$C3,$FC
	.dc.b	$35,$F0,$0F,$91,$FB,$5C,$7F,$F0,$01,$B0,$F2,$AC,$4C,$B8,$BA,$8C
	.dc.b	$65,$FE,$02,$F1,$9F,$05,$D1,$B7,$FE,$29,$5C,$79,$99,$39,$78,$46
	.dc.b	$2B,$F5,$9C,$9F,$47,$5F,$F4,$65,$FE,$22,$31,$86,$5B,$19,$3F,$1B
	.dc.b	$19,$69,$9A,$FD,$C4,$7C,$2E,$D7,$A7,$F1,$3F,$27,$7B,$99,$8C,$37
	.dc.b	$0F,$F2,$24,$E9,$FF,$1E,$4E,$9F,$F2,$F7,$85,$FE,$73,$F4,$BF,$51
	.dc.b	$E2,$7F,$96,$7C,$3F,$D5,$4E,$6B,$FC,$A5,$E3,$FF,$92,$9E,$3D,$5F
	.dc.b	$ED,$67,$35,$C2,$31,$B7,$F9,$92,$B8,$3F,$C4,$FD,$8F,$90,$7C,$9F
	.dc.b	$A9,$B8,$7E,$C2,$2C,$FF,$19,$9B,$DB,$FC,$B2,$6E,$1F,$CA,$26,$D7
	.dc.b	$FB,$D3,$B9,$E8,$FF,$1B,$F3,$7F,$8E,$6B,$FE,$58,$8C,$67,$FE,$18
	.dc.b	$8F,$F2,$DF,$9D,$8C,$39,$55,$9E,$15,$FE,$AE,$C3,$AB,$6F,$51,$8B
	.dc.b	$FF,$7E,$FB,$2F,$27,$93,$6F,$FD,$6E,$FF,$90,$5B,$1A,$23,$AF,$16
	.dc.b	$D7,$FB,$33,$95,$F0,$FE,$7F,$05,$18,$BF,$F6,$7F,$D3,$66,$31,$7F
	.dc.b	$ED,$4E,$A4,$E7,$0E,$51,$8B,$FF,$6F,$DF,$0E,$0F,$E8,$98,$AB,$58
	.dc.b	$F8,$F0,$E3,$8F,$3C,$7F,$DA,$7F,$56,$FA,$EF,$94,$BE,$06,$C6,$1B
	.dc.b	$3C,$1B,$83,$3F,$B3,$7C,$6B,$FD,$BF,$F5,$6F,$1F,$EB,$3F,$E5,$28
	.dc.b	$CC,$62,$F2,$D8,$C9,$CA,$3F,$00,$3F,$2B,$FF,$6B,$47,$F9,$73,$0F
	.dc.b	$EC,$93,$9E,$8D,$F8,$7A,$8F,$35,$7F,$84,$AF,$17,$C1,$B1,$9B,$E2
	.dc.b	$78,$23,$35,$FE,$6A,$6B,$DE,$7C,$47,$F8,$1A,$78,$1F,$95,$FA,$B7
	.dc.b	$55,$31,$E9,$5A,$E3,$6F,$8A,$7D,$53,$E0,$BF,$41,$18,$7F,$40,$3E
	.dc.b	$AF,$33,$BB,$E7,$5C,$B9,$37,$5A,$6C,$F8,$3B,$BB,$79,$7D,$0E,$FF
	.dc.b	$54,$F7,$B1,$5D,$23,$0D,$8F,$2F,$C3,$7F,$BB,$23,$07,$89,$C6,$BF
	.dc.b	$48,$F7,$57,$27,$19,$A7,$7C,$7F,$E1,$F8,$56,$E3,$1E,$5F,$F5,$3C
	.dc.b	$3D,$5F,$11,$C1,$FF,$4F,$DF,$F3,$2F,$47,$F8,$34,$E7,$2F,$9D,$7C
	.dc.b	$5F,$0B,$70,$FE,$27,$F1,$2F,$AA,$FA,$33,$9C,$BA,$B6,$F9,$E3,$FC
	.dc.b	$4E,$F8,$57,$D2,$AD,$64,$FA,$87,$C7,$FF,$19,$F8,$D7,$95,$F7,$A7
	.dc.b	$38,$7D,$59,$CB,$F7,$23,$39,$FC,$86,$EB,$FC,$AB,$EA,$1E,$57,$3C
	.dc.b	$A6,$1C,$5F,$E5,$13,$A8,$C1,$F8,$8F,$C0,$3F,$6D,$8F,$D7,$E6,$7C
	.dc.b	$8F,$F7,$2A,$D4,$FF,$31,$8C,$3F,$6A,$B8,$E5,$F6,$5B,$5F,$CE,$F9
	.dc.b	$DF,$51,$F9,$C9,$CD,$FE,$85,$BC,$07,$39,$7F,$81,$AB,$85,$7E,$93
	.dc.b	$E9,$5F,$3F,$FD,$0F,$D6,$FE,$FF,$3F,$8C,$9C,$E7,$F9,$F9,$CE,$9F
	.dc.b	$8D,$9C,$D6,$EF,$FA,$67,$9A,$F7,$7B,$44,$C7,$F5,$B7,$C3,$0F,$EB
	.dc.b	$37,$B9,$CD,$7E,$E3,$F2,$CD,$47,$F5,$BF,$26,$E5,$FB,$F9,$BA,$FF
	.dc.b	$27,$BE,$04,$BF,$E0,$BF,$34,$FA,$57,$C6,$E7,$1F,$D4,$E3,$D3,$88
	.dc.b	$F6,$BE,$15,$BD,$7F,$59,$F3,$BF,$E9,$B6,$FC,$2F,$F6,$DA,$ED,$F4
	.dc.b	$BF,$C4,$2B,$5A,$7E,$B1,$B9,$FF,$17,$6D,$70,$9F,$E2,$31,$AD,$9F
	.dc.b	$76,$F2,$FF,$BC,$36,$BF,$90,$F1,$3E,$9B,$DB,$1F,$C5,$67,$59,$7E
	.dc.b	$4D,$BA,$FF,$82,$1B,$29,$5F,$3D,$BB,$E3,$FD,$2A,$75,$EF,$6F,$AD
	.dc.b	$F4,$1F,$ED,$6F,$A5,$FE,$55,$5F,$A0,$8F,$F6,$8D,$6D,$D7,$9F,$FE
	.dc.b	$B3,$BE,$35,$FE,$DC,$AD,$3E,$3B,$E7,$3A,$8F,$F7,$1C,$E9,$4E,$AF
	.dc.b	$BC,$78,$4E,$57,$DE,$33,$7C,$EB,$85,$78,$4E,$07,$4A,$E0,$E6,$78
	.dc.b	$46,$23,$C6,$70,$9E,$6E,$27,$79,$F9,$E3,$E4,$9B,$AE,$4E,$27,$6F
	.dc.b	$88,$C5,$FC,$8A,$E5,$0C,$3E,$16,$1E,$08,$EA,$7F,$BE,$A3,$FD,$F6
	.dc.b	$E0,$F8,$A7,$DF,$3F,$01,$F4,$7F,$C0,$B8,$BF,$E0,$0B,$63,$46,$BA
	.dc.b	$13,$A6,$C6,$84,$BE,$97,$FC,$3D,$19,$8C,$70,$F6,$9C,$CF,$69,$CB
	.dc.b	$B8,$E3,$A7,$E0,$7E,$97,$F4,$53,$F8,$16,$BF,$62,$B8,$CF,$E1,$DB
	.dc.b	$AF,$BE,$B7,$3C,$F7,$F8,$4E,$F0,$EF,$5F,$16,$FC,$7E,$0E,$F1,$FD
	.dc.b	$33,$5F,$F5,$8C,$EE,$BD,$F3,$CF,$2F,$36,$3D,$AB,$C4,$DB,$FA,$37
	.dc.b	$E0,$6D,$70,$61,$ED,$8D,$3F,$39,$B9,$7B,$2F,$AC,$FE,$97,$1F,$82
	.dc.b	$F4,$3F,$3B,$3B,$7F,$80,$35,$E9,$7F,$B2,$E5,$FF,$16,$78,$FF,$87
	.dc.b	$8F,$F4,$7F,$5F,$F7,$0A,$FF,$11,$35,$FF,$0B,$72,$BE,$5F,$FD,$B9
	.dc.b	$1A,$BE,$5C,$9F,$E9,$5B,$E1,$87,$49,$CD,$7F,$12,$BE,$2F,$F8,$7B
	.dc.b	$A7,$FD,$B6,$FE,$3E,$D1,$88,$FC,$BC,$63,$E2,$FF,$29,$1F,$CB,$DB
	.dc.b	$19,$8F,$F1,$D7,$E0,$8F,$B4,$F0,$EA,$8C,$4F,$F3,$4B,$F2,$DF,$FA
	.dc.b	$5B,$E1,$AF,$C0,$FF,$AD,$2B,$FC,$92,$D7,$FC,$71,$D0,$FA,$6F,$E3
	.dc.b	$5B,$FF,$8D,$6B,$55,$FB,$3C,$3A,$3E,$D8,$CB,$FE,$34,$6B,$FE,$5F
	.dc.b	$7B,$4F,$FB,$0B,$C4,$F8,$4D,$CB,$CB,$97,$A3,$E9,$9F,$4D,$49,$8F
	.dc.b	$F9,$6F,$93,$FC,$5A,$FF,$98,$A7,$6F,$CF,$6B,$F3,$9A,$FA,$5A,$FF
	.dc.b	$92,$9D,$E1,$FE,$D6,$DD,$F2,$FF,$39,$B5,$DF,$0F,$9B,$E7,$3C,$6A
	.dc.b	$49,$FF,$38,$76,$FF,$83,$5F,$F3,$BC,$ED,$FC,$23,$5F,$C1,$B5,$F9
	.dc.b	$CD,$7F,$CB,$EE,$F1,$FF,$75,$FD,$37,$CC,$E9,$FF,$75,$36,$BF,$D1
	.dc.b	$D5,$B9,$FF,$33,$FD,$37,$FB,$8E,$7F,$D1,$2D,$7F,$CE,$AD,$BF,$F7
	.dc.b	$BF,$81,$6B,$8F,$C5,$FE,$52,$B8,$5F,$E0,$DA,$E8,$FD,$03,$C5,$6A
	.dc.b	$F5,$24,$7F,$98,$5F,$CB,$CE,$F9,$9E,$AB,$E9,$BF,$36,$DB,$FF,$38
	.dc.b	$47,$E8,$BD,$7F,$F6,$E4,$60,$FD,$84,$ED,$FD,$7B,$5F,$CD,$35,$FF
	.dc.b	$3F,$CF,$EB,$3D,$AF,$C6,$C7,$F9,$95,$BB,$7F,$A5,$27,$F5,$BE,$E7
	.dc.b	$F9,$CA,$BF,$D6,$7F,$1C,$F9,$BF,$D8,$B6,$3F,$D4,$DE,$D6,$27,$F7
	.dc.b	$B3,$99,$FE,$06,$D7,$F5,$CD,$7F,$C4,$1B,$BE,$99,$F3,$9F,$C0,$9B
	.dc.b	$3F,$EB,$F6,$BF,$EB,$0A,$ED,$9F,$FA,$BF,$E5,$7E,$3C,$FF,$A0,$7D
	.dc.b	$17,$D0,$7F,$0C,$6C,$7F,$B3,$A7,$75,$C7,$E6,$3F,$8A,$C1,$A4,$9F
	.dc.b	$ED,$D9,$BE,$99,$F2,$5F,$C6,$1B,$3F,$ED,$56,$BF,$E9,$37,$F9,$7F
	.dc.b	$4F,$F8,$07,$E9,$9F,$33,$FC,$C2,$91,$87,$FD,$DA,$62,$35,$F4,$2F
	.dc.b	$A6,$E1,$5B,$AF,$F0,$13,$5F,$F0,$4B,$5F,$05,$7F,$85,$23,$86,$AC
	.dc.b	$3F,$CA,$5E,$33,$E5,$F1,$87,$F9,$79,$AF,$FB,$8C,$F2,$1F,$EF,$3F
	.dc.b	$A6,$FE,$8F,$7C,$36,$7D,$37,$F2,$4C,$7E,$FB,$6F,$F5,$37,$C2,$7F
	.dc.b	$06,$D8,$FD,$6C,$EE,$B9,$F9,$7F,$D1,$AB,$51,$FE,$FF,$F2,$FF,$A4
	.dc.b	$FC,$6B,$E9,$A9,$88,$98,$FF,$9B,$F3,$F3,$56,$E3,$19,$FE,$52,$31
	.dc.b	$1F,$EC,$B7,$F8,$96,$AF,$A5,$7D,$07,$F6,$88,$FC,$1F,$AF,$79,$BF
	.dc.b	$B9,$61,$FE,$86,$BD,$BF,$E8,$88,$CB,$FD,$E1,$B9,$7D,$0B,$E7,$3C
	.dc.b	$70,$E9,$84,$8E,$4F,$87,$0E,$B1,$FE,$E1,$F9,$8F,$C7,$BF,$C5,$9B
	.dc.b	$B1,$A7,$BB,$1F,$B0,$E3,$D9,$FE,$E6,$EB,$7F,$EE,$BF,$45,$FB,$0D
	.dc.b	$FF,$72,$DC,$7F,$91,$B7,$13,$FE,$C0,$8F,$9A,$6B,$C1,$E6,$47,$33
	.dc.b	$C3,$E0,$FF,$49,$F3,$FF,$B8,$9F,$F3,$87,$CC,$FF,$9A,$63,$1F,$08
	.dc.b	$8C,$78,$5F,$E9,$67,$FC,$DD,$DD,$FC,$6E,$F8,$9F,$EA,$FB,$E9,$A7
	.dc.b	$F9,$B9,$B1,$C8,$FF,$5A,$46,$6F,$FC,$F0,$D8,$FF,$59,$DF,$8F,$E4
	.dc.b	$FC,$11,$D5,$AA,$FC,$F3,$FC,$83,$0F,$E1,$D3,$C8,$7C,$73,$E8,$1F
	.dc.b	$4D,$D2,$30,$F6,$D9,$7C,$AF,$DB,$18,$F1,$38,$61,$CA,$7B,$61,$C9
	.dc.b	$F0,$56,$DF,$17,$4F,$E7,$7E,$1F,$D6,$FF,$1E,$FA,$2B,$D8,$73,$D5
	.dc.b	$FE,$82,$E3,$FA,$58,$7F,$B6,$A3,$1B,$FF,$0C,$BE,$13,$D9,$AF,$D4
	.dc.b	$C3,$C0,$FC,$33,$E4,$7B,$5F,$2B,$87,$FB,$E6,$3D,$A7,$E6,$7A,$D9
	.dc.b	$A9,$1C,$6B,$78,$FE,$1E,$B3,$3F,$82,$38,$9B,$7B,$1A,$BF,$5E,$1F
	.dc.b	$D8,$7D,$CF,$87,$D1,$FE,$28,$BF,$43,$6B,$6F,$F5,$66,$CF,$02,$4F
	.dc.b	$CE,$9D,$61,$FC,$C3,$E9,$BF,$C5,$17,$E0,$6C,$67,$3F,$A7,$0F,$A3
	.dc.b	$B1,$30,$FB,$33,$C1,$64,$8E,$11,$F3,$7A,$DF,$4D,$FE,$7C,$CB,$E6
	.dc.b	$46,$3C,$6D,$4E,$73,$FD,$87,$3F,$E7,$2B,$58,$7C,$B6,$E1,$CA,$F8
	.dc.b	$4F,$EB,$EB,$84,$62,$3F,$82,$B6,$B8,$DF,$69,$F0,$E1,$1E,$D5,$C7
	.dc.b	$83,$F4,$18,$F7,$AD,$36,$BE,$F4,$EA,$FF,$DE,$18,$76,$91,$E2,$BE
	.dc.b	$33,$E1,$8E,$F1,$87,$73,$C5,$FD,$E6,$3E,$C9,$D4,$73,$3C,$31,$97
	.dc.b	$99,$30,$FC,$D6,$1E,$58,$EB,$8F,$C3,$25,$FE,$6E,$F9,$5F,$B4,$98
	.dc.b	$F9,$DB,$7C,$AB,$C7,$87,$58,$F2,$4E,$9B,$5C,$4F,$8B,$87,$82,$3C
	.dc.b	$D1,$97,$A1,$82,$CF,$AB,$0F,$2C,$F2,$C3,$C3,$18,$AD,$CE,$5F,$8A
	.dc.b	$7C,$8C,$7E,$61,$F3,$71,$F4,$38,$C7,$C9,$8F,$7A,$E1,$18,$BF,$34
	.dc.b	$7B,$0E,$E4,$3E,$B9,$C3,$1F,$2B,$4F,$61,$A2,$DF,$86,$3E,$94,$7E
	.dc.b	$4C,$FC,$49,$F1,$0A,$7E,$04,$B1,$FA,$66,$FD,$33,$E0,$BF,$DC,$B2
	.dc.b	$8C,$4F,$23,$FB,$4B,$87,$8C,$F0,$8D,$E3,$FB,$86,$E1,$E4,$F0,$7F
	.dc.b	$D7,$5D,$DF,$BF,$9C,$E1,$F9,$A6,$BD,$B0,$FE,$61,$5C,$23,$F7,$93
	.dc.b	$AC,$7C,$D5,$CC,$DF,$07,$F9,$1B,$2F,$14,$66,$F8,$BA,$C7,$13,$A9
	.dc.b	$C9,$FC,$CF,$1F,$04,$64,$99,$7E,$E3,$2F,$3C,$EA,$31,$87,$46,$DF
	.dc.b	$3C,$7C,$91,$8C,$39,$CF,$C1,$AF,$23,$E2,$8C,$BF,$0C,$E5,$8F,$57
	.dc.b	$81,$BA,$F9,$D7,$DA,$33,$5C,$B2,$F8,$F1,$C8,$E0,$6A,$BC,$87,$2B
	.dc.b	$ED,$1E,$1A,$F1,$E3,$EF,$8E,$04,$35,$5C,$5E,$A6,$0D,$47,$17,$04
	.dc.b	$BF,$2D,$7F,$6B,$9D,$46,$2B,$CB,$5C,$B0,$F7,$3C,$72,$FC,$21,$F5
	.dc.b	$23,$35,$C6,$7C,$CD,$7F,$CA,$11,$A9,$E3,$97,$C0,$C3,$C3,$18,$AF
	.dc.b	$5C,$62,$7E,$2D,$F3,$8C,$4F,$C0,$8C,$5F,$5A,$DD,$F9,$72,$FB,$5C
	.dc.b	$0F,$95,$7C,$D7,$D9,$8C,$36,$1F,$4D,$62,$57,$F8,$9A,$78,$B6,$73
	.dc.b	$9F,$F8,$6F,$C8,$7C,$1F,$56,$D7,$F8,$B9,$70,$E6,$64,$E5,$E1,$18
	.dc.b	$AF,$D3,$57,$D2,$DF,$FD,$23,$7F,$B8,$E4,$C4,$E5,$B1,$93,$F7,$B1
	.dc.b	$96,$99,$AE,$AF,$F8,$D7,$D3,$F8,$9E,$7B,$DC,$F8,$DD,$A9,$CC,$61
	.dc.b	$B5,$C4,$FF,$25,$D7,$17,$FC,$9F,$CB,$FE,$E0,$7F,$CA,$7E,$1F,$F9
	.dc.b	$2D,$CF,$A5,$FA,$BF,$13,$FC,$C5,$E1,$7E,$B7,$C4,$FF,$3F,$4E,$6F
	.dc.b	$86,$BF,$E6,$0C,$7A,$3F,$E7,$EC,$38,$E3,$FC,$73,$D6,$F9,$1F,$4B
	.dc.b	$F2,$5B,$19,$3F,$CF,$F5,$FC,$A6,$3F,$DC,$31,$FC,$F3,$89,$18,$9D
	.dc.b	$4F,$07,$FC,$17,$3A,$8E,$91,$FC,$77,$E8,$7C,$11,$FE,$70,$F0,$1E
	.dc.b	$97,$FB,$BB,$D0,$FF,$74,$46,$27,$FD,$46,$BD,$0D,$47,$F8,$63,$83
	.dc.b	$FD,$19,$78,$C5,$18,$9E,$33,$98,$FF,$49,$55,$BE,$D6,$C3,$FD,$F1
	.dc.b	$ED,$BC,$9E,$0D,$CB,$FD,$6A,$FF,$94,$5B,$1A,$23,$AF,$1E,$DF,$58
	.dc.b	$FF,$17,$F8,$0E,$67,$C4,$7A,$63,$0D,$EA,$7C,$FE,$6B,$77,$FE,$CA
	.dc.b	$BD,$B6,$33,$87,$FB,$AE,$B8,$61,$CD,$B5,$FE,$F1,$6C,$F9,$4F,$93
	.dc.b	$F4,$AF,$A6,$F9,$0B,$9A,$C6,$1B,$7F,$ED,$89,$D3,$67,$4F,$FB,$73
	.dc.b	$E9,$BF,$C0,$0D,$BF,$A9,$18,$7C,$5C,$68,$FC,$9F,$0A,$BF,$01,$1B
	.dc.b	$8C,$57,$E0,$23,$73,$AF,$1F,$AC,$FE,$14,$7D,$37,$FB,$9E,$7E,$47
	.dc.b	$C1,$E6,$31,$3F,$0F,$83,$EB,$4E,$7B,$3B,$CF,$F8,$7A,$B6,$D8,$CB
	.dc.b	$FE,$24,$8C,$B6,$39,$5F,$F9,$C3,$2E,$7D,$5E,$07,$AD,$D9,$FC,$F7
	.dc.b	$D3,$3A,$FF,$34,$D8,$E4,$6B,$D3,$CC,$FE,$8F,$E1,$DF,$4C,$F8,$3F
	.dc.b	$2E,$A9,$CC,$61,$B8,$7F,$8C,$5B,$3F,$E3,$C7,$9F,$37,$EA,$E9,$F4
	.dc.b	$DD,$F0,$FC,$BE,$1E,$4C,$3F,$67,$5A,$8C,$72,$78,$FC,$BE,$47,$9F
	.dc.b	$E9,$9F,$01,$8A,$E5,$A2,$73,$1E,$56,$E3,$E5,$6B,$FE,$54,$9E,$15
	.dc.b	$ED,$F8,$87,$B9,$DE,$3E,$34,$63,$0E,$F8,$74,$E3,$FC,$6E,$1F,$D0
	.dc.b	$7D,$2F,$E1,$5F,$13,$BC,$FF,$51,$39,$C3,$DB,$39,$BE,$3E,$77,$F6
	.dc.b	$6B,$F9,$4D,$7C,$CD,$7F,$D6,$77,$CD,$DE,$1F,$6E,$79,$57,$E8,$36
	.dc.b	$7C,$53,$DF,$D5,$FE,$07,$E9,$9F,$1F,$FC,$87,$E3,$BF,$95,$6D,$F3
	.dc.b	$27,$2F,$D4,$8C,$E9,$EC,$9C,$BF,$81,$6B,$F3,$5A,$FA,$1A,$FE,$35
	.dc.b	$E3,$9F,$CA,$6D,$7F,$39,$FA,$67,$97,$E7,$C7,$F3,$1F,$4D,$FB,$AC
	.dc.b	$7F,$9D,$FC,$8F,$F8,$1B,$0F,$C1,$CE,$77,$4E,$6B,$E9,$CF,$F5,$37
	.dc.b	$E5,$B5,$FD,$03,$78,$4D,$3C,$0E,$23,$FA,$47,$3B,$E9,$BC,$B1,$FA
	.dc.b	$DF,$A6,$7D,$07,$F4,$DF,$57,$FC,$27,$3F,$84,$9C,$E9,$FA,$09,$CF
	.dc.b	$0F,$E3,$27,$35,$FA,$D6,$BF,$99,$6B,$F8,$46,$BF,$B9,$78,$C7,$3B
	.dc.b	$FE,$BD,$E3,$FF,$8F,$70,$FD,$4D,$69,$FC,$6F,$D2,$DC,$3F,$B3,$F9
	.dc.b	$37,$54,$E7,$83,$FD,$6D,$97,$F0,$B8,$F0,$CE,$A3,$F8,$44,$60,$8E
	.dc.b	$CF,$EE,$9F,$34,$FA,$67,$C6,$E7,$3F,$DE,$FC,$F7,$FA,$BF,$7F,$CE
	.dc.b	$7C,$EF,$F0,$7D,$BF,$5B,$39,$AF,$F0,$04,$E6,$BF,$6B,$3F,$EB,$7A
	.dc.b	$FE,$00,$D7,$F5,$6D,$7F,$90,$37,$57,$D2,$3C,$AF,$F0,$33,$71,$FE
	.dc.b	$DD,$F4,$BE,$87,$F8,$CC,$67,$0F,$EA,$ED,$C7,$F6,$8D,$9F,$EF,$F5
	.dc.b	$C6,$B7,$7F,$D4,$D0,$F8,$AB,$D2,$FA,$3F,$F7,$6F,$D2,$FF,$39,$C7
	.dc.b	$31,$FE,$E5,$AB,$7C,$AB,$4F,$E9,$27,$4D,$35,$E5,$7F,$68,$C3,$B4
	.dc.b	$7F,$BE,$2F,$9A,$BE,$73,$A8,$FF,$7F,$CE,$A7,$37,$DA,$B5,$18,$AE
	.dc.b	$87,$84,$E9,$7E,$1A,$E0,$7B,$9C,$47,$0E,$BE,$71,$FD,$F9,$F1,$47
	.dc.b	$C8,$C3,$E4,$C6,$B1,$B7,$D2,$BD,$F1,$EC,$78,$1E,$3C,$3C,$CF,$B0
	.dc.b	$C6,$1E,$27,$18,$F5,$7D,$47,$23,$EC,$9E,$E2,$9C,$CA,$6C,$DF,$D2
	.dc.b	$7F,$08,$E4,$CC,$23,$0F,$06,$BD,$1A,$F2,$FA,$5F,$F2,$C3,$B6,$BB
	.dc.b	$6B,$BE,$0E,$F2,$9E,$37,$F7,$7E,$97,$F4,$57,$F8,$26,$BF,$66,$B8
	.dc.b	$5F,$E1,$9B,$1F,$76,$77,$5C,$F7,$F8,$AF,$E2,$6B,$6D,$C3,$D3,$B6
	.dc.b	$F0,$FB,$27,$E9,$30,$F0,$C6,$F1,$CC,$FC,$7A,$F4,$ED,$F4,$AF,$C2
	.dc.b	$DC,$BC,$2F,$C1,$AF,$55,$70,$7B,$4E,$9A,$7B,$63,$4F,$7B,$F8,$93
	.dc.b	$B8,$5F,$81,$FD,$DE,$5E,$EC,$3C,$19,$7B,$67,$E3,$5F,$1A,$D4,$33
	.dc.b	$7E,$91,$FD,$3E,$3F,$C5,$51,$8E,$FF,$C7,$E7,$FE,$60,$8C,$4F,$F8
	.dc.b	$BC,$7E,$13,$F7,$5E,$37,$CD,$FC,$AC,$F0,$6C,$65,$FC,$DC,$65,$FE
	.dc.b	$57,$F6,$7F,$96,$ED,$74,$23,$0D,$9F,$F2,$43,$5F,$F2,$43,$71,$FF
	.dc.b	$0C,$7E,$1D,$E3,$6E,$7E,$CE,$5F,$E6,$F8,$5C,$3F,$CA,$DE,$FF,$FB
	.dc.b	$05,$FF,$21,$B5,$FF,$18,$B6,$FF,$D0,$7E,$1B,$F0,$27,$C9,$F3,$D7
	.dc.b	$FD,$5F,$F0,$FE,$BC,$39,$63,$FE,$47,$6B,$FE,$32,$6B,$E2,$8F,$27
	.dc.b	$73,$F0,$2F,$F4,$1E,$7F,$E3,$36,$EB,$C6,$BC,$1F,$52,$FA,$6A,$4C
	.dc.b	$7F,$D7,$5F,$8D,$7F,$B0,$9F,$F2,$FC,$ED,$FF,$39,$35,$FF,$39,$35
	.dc.b	$FF,$2F,$35,$FF,$24,$BB,$C7,$FC,$8D,$E5,$3E,$BB,$D2,$DA,$EF,$87
	.dc.b	$C5,$FB,$3F,$E3,$37,$FE,$60,$6E,$7F,$E5,$46,$D2,$DF,$8D,$3E,$C3
	.dc.b	$94,$FE,$1E,$BF,$8E,$BF,$E8,$F6,$F0,$9A,$74,$E0,$78,$4F,$BA,$F9
	.dc.b	$ED,$BF,$F4,$1D,$78,$AD,$F5,$AF,$8F,$FB,$CD,$AF,$F4,$4B,$5F,$F6
	.dc.b	$F7,$43,$ED,$9F,$65,$C2,$FF,$3D,$7B,$BF,$F4,$53,$5F,$F3,$AB,$5F
	.dc.b	$F5,$53,$D7,$87,$FC,$E9,$E2,$7E,$42,$7F,$D3,$ED,$7F,$D0,$F3,$BE
	.dc.b	$2F,$F3,$FF,$BA,$FC,$03,$F2,$CD,$AF,$F5,$0B,$5F,$4F,$DC,$7F,$9A
	.dc.b	$1C,$CF,$ED,$67,$51,$FE,$8C,$8C,$11,$FF,$58,$FB,$FF,$EF,$D6,$FC
	.dc.b	$7F,$82,$7F,$61,$0F,$A8,$FF,$00,$69,$FE,$E0,$6E,$5F,$D8,$1B,$3F
	.dc.b	$E1,$A5,$F3,$5C,$9F,$EE,$F1,$BF,$33,$F0,$71,$88,$E1,$18,$FA,$F9
	.dc.b	$1F,$D9,$25,$18,$8F,$ED,$70,$F8,$BE,$64,$8E,$0F,$6F,$7E,$47,$36
	.dc.b	$BF,$DD,$3E,$61,$F5,$5F,$DE,$B3,$FE,$F3,$18,$9F,$F4,$53,$5F,$F7
	.dc.b	$EB,$63,$7F,$3F,$FE,$24,$F9,$3F,$F0,$1C,$A3,$15,$FE,$3E,$8C,$3F
	.dc.b	$E1,$99,$FF,$20,$F9,$EF,$6F,$FD,$C3,$39,$38,$5F,$95,$EA,$D7,$27
	.dc.b	$99,$C2,$BE,$EF,$F9,$36,$FC,$BF,$41,$E4,$8D,$36,$77,$ED,$2B,$BD
	.dc.b	$78,$CD,$96,$55,$C6,$E4,$E6,$3A,$18,$7C,$47,$A0,$A7,$42,$CF,$47
	.dc.b	$B4,$F1,$BF,$24,$7C,$BA,$EA,$E8,$87,$27,$2F,$49,$E5,$7F,$0A,$3C
	.dc.b	$C7,$0A,$EA,$F8,$DB,$7E,$38,$E8,$6E,$79,$38,$24,$F0,$AF,$0B,$CA
	.dc.b	$ED,$7C,$CA,$DB,$CA,$75,$1D,$4F,$71,$E6,$36,$7C,$97,$6E,$0F,$44
	.dc.b	$75,$8C,$E5,$F9,$CA,$FC,$2E,$1E,$B8,$C5,$FE,$7A,$7F,$09,$F4,$AF
	.dc.b	$A6,$7D,$1F,$E0,$63,$0F,$8F,$1F,$09,$F2,$23,$37,$F7,$1B,$1A,$CB
	.dc.b	$DC,$7F,$9A,$23,$2D,$BE,$F5,$EB,$CB,$E8,$5F,$59,$CB,$70,$F1,$5F
	.dc.b	$4B,$FF,$55,$E1,$D1,$B7,$C2,$75,$1F,$18,$CD,$7B,$67,$51,$F4,$0F
	.dc.b	$53,$85,$5D,$5B,$19,$C3,$CF,$7E,$BB,$E8,$4A,$FA,$D7,$C4,$97,$C7
	.dc.b	$1F,$A2,$F7,$8C,$D7,$BA,$BC,$E7,$91,$E9,$8F,$CD,$8C,$BF,$15,$D1
	.dc.b	$E0,$AE,$84,$9E,$91,$EA,$79,$E1,$E9,$8E,$11,$87,$AC,$7C,$AB,$F2
	.dc.b	$C7,$6B,$F1,$3C,$CC,$4F,$38,$CC,$F0,$7C,$C6,$2F,$C9,$5C,$0F,$A5
	.dc.b	$1A,$AE,$44,$8C,$B5,$F0,$46,$DC,$9E,$82,$1A,$8F,$54,$F8,$27,$84
	.dc.b	$FE,$14,$D1,$C2,$BF,$01,$8F,$DE,$8C,$57,$1B,$E7,$87,$C7,$9E,$27
	.dc.b	$3B,$F8,$13,$E7,$BE,$53,$A9,$CB,$E8,$31,$7D,$5C,$46,$9E,$B3,$CA
	.dc.b	$30,$79,$B1,$FC,$5E,$1E,$38,$C5,$5B,$E8,$F9,$5B,$1A,$8F,$A4,$78
	.dc.b	$A7,$F0,$F1,$DA,$7F,$31,$3F,$9D,$3F,$DC,$31,$FC,$09,$B1,$FB,$C6
	.dc.b	$CF,$FB,$82,$5F,$4C,$F8,$5C,$36,$BF,$DD,$33,$A9,$C9,$23,$F8,$74
	.dc.b	$EF,$3F,$E7,$3F,$31,$FE,$80,$D7,$F8,$C4,$E7,$5F,$E2,$0D,$BF,$E0
	.dc.b	$ED,$55,$C5,$EB,$87,$22,$3F,$D6,$23,$58,$7F,$21,$AE,$58,$F9,$EB
	.dc.b	$89,$BF,$4D,$97,$2B,$E1,$8D,$C9,$73,$2E,$B1,$E8,$9D,$4F,$A1,$FE
	.dc.b	$E9,$3A,$8C,$1E,$0A,$EA,$F8,$6F,$8B,$FC,$6B,$2F,$39,$F0,$6B,$94
	.dc.b	$66,$F7,$97,$AF,$0F,$9B,$97,$AA,$BB,$65,$E5,$69,$E9,$8D,$4F,$88
	.dc.b	$93,$B7,$9E,$5D,$B2,$F1,$E3,$EC,$BE,$37,$B2,$61,$EE,$C3,$8B,$E0
	.dc.b	$3D,$05,$CB,$D7,$97,$A6,$B8,$CE,$AB,$89,$E4,$7A,$46,$AB,$8E,$3E
	.dc.b	$49,$E9,$87,$91,$D9,$0E,$95,$F6,$9F,$C5,$C6,$DC,$9B,$C3,$C8,$FC
	.dc.b	$7C,$BE,$16,$5F,$5B,$1F,$0B,$89,$FB,$47,$BE,$B8,$BD,$0E,$F5,$BC
	.dc.b	$BD,$91,$83,$86,$5E,$B8,$F8,$39,$7C,$4B,$E6,$74,$C7,$C3,$7C,$31
	.dc.b	$FA,$98,$F5,$BF,$8D,$97,$B7,$0E,$39,$75,$9F,$85,$8F,$2C,$BC,$75
	.dc.b	$D6,$F8,$5E,$F2,$F9,$57,$BC,$38,$D6,$B2,$F0,$D7,$D2,$CB,$E3,$4E
	.dc.b	$63,$E2,$E5,$F0,$30,$E5,$1F,$34,$E0,$48,$D4,$F1,$7E,$1C,$78,$A3
	.dc.b	$18,$78,$AF,$F6,$A7,$DD,$9F,$98,$DB,$FE,$10,$DA,$FD,$33,$4F,$4D
	.dc.b	$FF,$16,$F9,$FF,$C2,$57,$EC,$09,$58,$8C,$B2,$B8,$E1,$F3,$27,$18
	.dc.b	$7C,$08,$F3,$3F,$C5,$63,$38,$7B,$23,$F0,$4F,$86,$B9,$C7,$C5,$8D
	.dc.b	$60,$AE,$37,$DE,$F7,$7F,$C9,$A7,$2F,$F1,$76,$B9,$8D,$D7,$36,$CE
	.dc.b	$A3,$13,$E9,$39,$4E,$88,$E4,$E8,$58,$E3,$7E,$8A,$E8,$4A,$FA,$F5
	.dc.b	$E9,$3F,$8C,$1D,$8A,$74,$7E,$D4,$7D,$17,$FD,$65,$C5,$FE,$62,$FB
	.dc.b	$FC,$57,$F4,$1F,$A2,$FB,$AF,$E7,$9A,$FF,$39,$6B,$FD,$31,$B8,$BE
	.dc.b	$67,$C3,$3F,$CF,$DA,$FF,$52,$6C,$6F,$0F,$C8,$4F,$F4,$5B,$C9,$23
	.dc.b	$31,$FD,$6A,$38,$E7,$27,$35,$FE,$BD,$6C,$76,$F6,$64,$69,$B7,$E0
	.dc.b	$A4,$7F,$41,$EE,$F9,$71,$8A,$FE,$B3,$18,$F3,$35,$1F,$C0,$AA,$DB
	.dc.b	$CB,$78,$8F,$4B,$3E,$7B,$C6,$79,$5C,$2F,$75,$A9,$CE,$5E,$03,$FB
	.dc.b	$4C,$E9,$F8,$ED,$C7,$F7,$0D,$7F,$9B,$35,$E3,$8F,$8E,$7A,$47,$9E
	.dc.b	$30,$B4,$FE,$C5,$E6,$BC,$97,$87,$E7,$BD,$D7,$F6,$7C,$5F,$1E,$F1
	.dc.b	$FD,$FF,$1E,$F8,$5F,$74,$E7,$5F,$0F,$0A,$B5,$1E,$4C,$39,$E1,$F8
	.dc.b	$A6,$C7,$C0,$BE,$5B,$7C,$06,$BF,$01,$CE,$5E,$56,$CF,$F8,$4B,$67
	.dc.b	$A0,$F8,$5F,$2E,$5F,$4F,$D2,$7C,$7F,$D8,$E7,$7D,$C6,$63,$0D,$CB
	.dc.b	$E9,$D7,$06,$CF,$A6,$31,$AF,$D6,$6E,$BF,$25,$B5,$FE,$26,$F6,$1F
	.dc.b	$8A,$7E,$2A,$3E,$24,$EA,$FF,$1D,$F2,$0F,$88,$FC,$6E,$0F,$AB,$FC
	.dc.b	$64,$E7,$0F,$9F,$39,$7E,$74,$66,$FF,$12,$D7,$F1,$2D,$7F,$22,$D7
	.dc.b	$C3,$7D,$8F,$D0,$D3,$D6,$7C,$E6,$70,$FB,$9F,$24,$F6,$FF,$47,$F0
	.dc.b	$6F,$9F,$FD,$14,$E6,$BF,$0B,$39,$78,$D7,$E0,$E7,$59,$FE,$81,$AF
	.dc.b	$E8,$1A,$FE,$A1,$DF,$07,$E5,$DB,$5F,$83,$F8,$6F,$D7,$D7,$E8,$63
	.dc.b	$85,$FE,$D7,$E2,$5E,$B7,$EE,$A7,$FC,$FA,$7E,$5E,$32,$FE,$70,$CD
	.dc.b	$73,$3F,$82,$46,$9F,$DB,$B5,$FD,$BB,$5F,$DF,$B5,$79,$8E,$3F,$9B
	.dc.b	$5F,$CA,$23,$1B,$7F,$30,$6C,$66,$77,$5F,$DF,$67,$2F,$95,$BA,$7F
	.dc.b	$4A,$FA,$1F,$E9,$AF,$F9,$8E,$33,$7F,$CD,$BC,$FF,$F3,$DF,$C0,$7E
	.dc.b	$FF,$0F,$F3,$EF,$C7,$FE,$E9,$46,$5F,$F2,$44,$F0,$EC,$FF,$25,$56
	.dc.b	$EB,$55,$FE,$86,$AD,$3D,$3C,$4F,$F3,$A4,$EA,$13,$A5,$5C,$EF,$CD
	.dc.b	$1E,$27,$BD,$F2,$73,$7E,$5A,$E1,$5F,$E2,$1B,$E9,$18,$8F,$49,$CC
	.dc.b	$F2,$61,$E1,$7D,$03,$B7,$A6,$3F,$E8,$B7,$D1,$7F,$EA,$27,$35,$71
	.dc.b	$F1,$9E,$67,$A5,$FB,$8F,$01,$EE,$BF,$53,$E7,$33,$87,$9D,$F1,$E3
	.dc.b	$D9,$E2,$79,$8F,$ED,$78,$FF,$6C,$70,$7C,$83,$FC,$85,$F1,$CF,$94
	.dc.b	$FC,$8C,$7F,$76,$8C,$69,$FD,$D1,$B1,$99,$DB,$CB,$C0,$FE,$98,$F1
	.dc.b	$6B,$9D,$B9,$46,$23,$FD,$1D,$5F,$CD,$67,$EF,$7C,$87,$F1,$F9,$BD
	.dc.b	$9F,$E9,$3C,$FF,$1D,$7B,$9F,$8E,$D8,$FC,$95,$71,$9F,$96,$D8,$E3
	.dc.b	$87,$99,$B8,$76,$A8,$74,$9D,$6F,$E6,$6C,$7F,$B7,$AF,$A4,$FF,$A9
	.dc.b	$DB,$1F,$EE,$2C,$3C,$93,$C6,$FA,$E1,$E6,$C7,$D9,$39,$BF,$05,$6A
	.dc.b	$3A,$4E,$A7,$A6,$3E,$25,$87,$67,$C9,$3B,$58,$F9,$1D,$D7,$AB,$37
	.dc.b	$C7,$BE,$3B,$FD,$81,$9F,$F0,$38,$C5,$FF,$82,$AB,$75,$AE,$1F,$F6
	.dc.b	$37,$C8,$7F,$6D,$AD,$5F,$07,$F9,$27,$95,$FE,$27,$FA,$4E,$10,$8C
	.dc.b	$3C,$30,$CF,$94,$EA,$FF,$1B,$CF,$F8,$2B,$E4,$7A,$6B,$FE,$F9,$C1
	.dc.b	$18,$D3,$F9,$5F,$A7,$CB,$0E,$38,$78,$AB,$4F,$85,$B8,$F7,$AF,$0F
	.dc.b	$C8,$BE,$46,$92,$FF,$C5,$7E,$6B,$C2,$7B,$99,$9F,$7F,$8B,$FE,$76
	.dc.b	$DF,$CC,$DD,$BF,$C7,$B9,$75,$AF,$F1,$13,$5E,$31,$F1,$B7,$57,$19
	.dc.b	$FF,$28,$56,$EF,$FC,$69,$F2,$1F,$E5,$19,$FF,$16,$35,$FE,$A9,$B2
	.dc.b	$31,$2C,$7F,$C2,$5F,$22,$C3,$E8,$D6,$A5,$87,$D7,$F2,$9F,$23,$E5
	.dc.b	$6C,$78,$3D,$8E,$26,$66,$46,$4F,$F0,$E9,$F7,$BC,$6F,$F6,$96,$7F
	.dc.b	$E2,$58,$FC,$2D,$BE,$45,$F2,$3B,$9F,$89,$E3,$BE,$46,$99,$8E,$A7
	.dc.b	$F9,$42,$34,$D8,$D3,$F1,$3E,$47,$FC,$EA,$FF,$94,$DA,$E7,$DD,$FC
	.dc.b	$BF,$3F,$F9,$2B,$CE,$FC,$AD,$6B,$4F,$C9,$78,$FF,$82,$C3,$FC,$BE
	.dc.b	$D7,$91,$D7,$0F,$B9,$19,$9E,$3D,$1D,$27,$F4,$3E,$D3,$E4,$69,$30
	.dc.b	$FC,$8F,$BA,$F8,$EC,$47,$EB,$BD,$FF,$F1,$A3,$FE,$70,$6B,$F8,$2B
	.dc.b	$E8,$FF,$9C,$5A,$F1,$F5,$BF,$21,$3C,$33,$FF,$40,$56,$9F,$F6,$8F
	.dc.b	$C7,$FF,$7F,$6F,$9D,$FB,$CC,$13,$9E,$AF,$C4,$C6,$3E,$47,$53,$98
	.dc.b	$FC,$6A,$F9,$1C,$9A,$BB,$3F,$E9,$A1,$F6,$4B,$F0,$36,$35,$7C,$DB
	.dc.b	$F0,$7E,$0D,$A4,$60,$DB,$FE,$18,$F8,$5F,$3F,$C8,$3E,$4B,$15,$FA
	.dc.b	$28,$C3,$FE,$19,$9F,$6E,$1F,$4E,$30,$6E,$B8,$AA,$7C,$D6,$1B,$1C
	.dc.b	$0D,$4E,$7D,$D9,$19,$9F,$F2,$AB,$FE,$FC,$F9,$37,$CF,$6A,$3F,$C4
	.dc.b	$F9,$7A,$BE,$53,$B1,$9F,$71,$1E,$EC,$73,$5D,$6B,$53,$98,$C1,$0E
	.dc.b	$A7,$C3,$9D,$BF,$16,$79,$BA,$3E,$04,$69,$F8,$51,$AA,$F1,$C7,$33
	.dc.b	$DD,$39,$AF,$54,$60,$87,$5C,$57,$B8,$E4,$65,$C3,$E0,$38,$DF,$81
	.dc.b	$C1,$B3,$DC,$E6,$7E,$7C,$64,$91,$99,$F6,$4F,$8A,$78,$D7,$A1,$F3
	.dc.b	$5F,$8A,$75,$1C,$A7,$E7,$34,$F9,$47,$23,$D7,$7D,$DC,$1E,$87,$CD
	.dc.b	$5E,$EF,$92,$7C,$A3,$E5,$3F,$19,$18,$78,$E3,$E2,$BF,$85,$39,$C3
	.dc.b	$EC,$36,$3C,$11,$E5,$11,$9C,$7F,$7A,$D8,$F4,$9B,$70,$F0,$C3,$C8
	.dc.b	$F3,$BF,$15,$72,$9C,$DF,$91,$B1,$F2,$E3,$C6,$67,$0F,$6C,$EA,$FC
	.dc.b	$33,$AA,$E9,$5A,$3E,$4E,$1E,$42,$5F,$1B,$F9,$95,$C1,$F8,$71,$99
	.dc.b	$F5,$E1,$E8,$9F,$EC,$B1,$CA,$FB,$47,$C1,$AE,$04,$8F,$0B,$63,$AC
	.dc.b	$60,$E3,$87,$B6,$39,$46,$30,$F2,$4F,$B6,$7C,$35,$CE,$38,$19,$8F
	.dc.b	$A1,$5C,$5D,$BD,$DF,$09,$23,$85,$FA,$E3,$17,$F4,$4F,$09,$1E,$67
	.dc.b	$98,$D9,$89,$E7,$7E,$59,$E1,$3F,$38,$D3,$8A,$F9,$B9,$7F,$28,$7C
	.dc.b	$57,$EC,$9F,$15,$70,$21,$EE,$3C,$51,$B7,$38,$FD,$77,$51,$96,$CF
	.dc.b	$46,$9D,$2B,$C6,$FC,$93,$24,$AE,$91,$E6,$9F,$6C,$63,$0F,$81,$5E
	.dc.b	$2C,$3D,$91,$88,$E5,$18,$C7,$B6,$1E,$09,$F9,$67,$81,$C6,$3E,$97
	.dc.b	$C1,$1F,$A0,$C7,$E5,$61,$D7,$1F,$16,$1D,$23,$F3,$D1,$8A,$EB,$7D
	.dc.b	$2F,$F4,$11,$FD,$E0,$F9,$85,$8F,$E0,$CD,$AF,$DD,$B9,$F9,$47,$CA
	.dc.b	$71,$8F,$D2,$C6,$1C,$52,$31,$1C,$67,$83,$FC,$3A,$38,$65,$FC,$59
	.dc.b	$B8,$7F,$83,$3E,$43,$FC,$E3,$D3,$FC,$7A,$73,$7E,$0C,$30,$7F,$6D
	.dc.b	$AD,$C7,$F1,$E9,$DB,$DE,$76,$FF,$59,$9D,$E1,$E1,$9C,$B6,$75,$E0
	.dc.b	$FF,$9A,$DE,$93,$AC,$7F,$92,$E1,$DB,$0E,$71,$89,$D5,$79,$CE,$78
	.dc.b	$E1,$F2,$1D,$0C,$97,$1F,$11,$2B,$E2,$63,$E3,$8C,$4F,$56,$AC,$3C
	.dc.b	$6E,$6B,$C8,$F5,$79,$DF,$8F,$2F,$B1,$8F,$96,$F8,$DF,$9B,$0E,$B5
	.dc.b	$A8,$F3,$C6,$32,$F9,$F3,$98,$FE,$E5,$5C,$A3,$18,$F8,$A7,$CF,$39
	.dc.b	$AE,$24,$C3,$EE,$65,$EF,$BE,$04,$C7,$D8,$7F,$A6,$AF,$95,$7A,$EF
	.dc.b	$EA,$5F,$1A,$D9,$EB,$BE,$4F,$C7,$8D,$39,$37,$8F,$A1,$ED,$1A,$8C
	.dc.b	$46,$C9,$87,$89,$F9,$C7,$81,$C6,$3F,$1F,$1F,$6E,$5E,$77,$E8,$61
	.dc.b	$D7,$2F,$85,$87,$23,$CD,$5B,$71,$87,$4C,$BD,$79,$79,$23,$13,$EA
	.dc.b	$BF,$87,$7C,$72,$FA,$38,$F7,$CB,$19,$7E,$13,$0F,$06,$5E,$E3,$A5
	.dc.b	$6E,$3F,$03,$3F,$2B,$2E,$F1,$E6,$3C,$B3,$BC,$3F,$06,$DA,$ED,$1E
	.dc.b	$2C,$7E,$09,$2B,$8B,$63,$51,$82,$38,$3A,$C7,$F0,$3F,$94,$FB,$78
	.dc.b	$7E,$68,$98,$62,$32,$C9,$FA,$77,$FC,$16,$70,$FB,$8F,$C0,$C7,$53
	.dc.b	$77,$E0,$8D,$1F,$B1,$9E,$31,$92,$61,$FC,$76,$32,$FC,$22,$4F,$E6
	.dc.b	$E3,$2D,$AE,$F3,$C5,$B1,$93,$C3,$82,$B8,$C7,$8E,$76,$D3,$71,$A9
	.dc.b	$E1,$3A,$6C,$EA,$31,$3B,$23,$82,$47,$C9,$7F,$D4,$FE,$F7,$F9,$5F
	.dc.b	$E4,$BE,$84,$7F,$37,$F9,$4F,$AB,$F4,$9F,$59,$B1,$FC,$EF,$E5,$70
	.dc.b	$D3,$81,$FC,$FE,$BF,$9A,$78,$3F,$CD,$71,$FE,$41,$58,$AF,$E9,$55
	.dc.b	$67,$F8,$34,$D7,$F9,$36,$97,$E8,$FF,$D5,$37,$FE,$D3,$D6,$D3,$1E
	.dc.b	$D5,$87,$C1,$2F,$77,$A3,$9B,$FE,$B5,$18,$C3,$2C,$8E,$DE,$7D,$C7
	.dc.b	$FD,$F5,$EA,$F9,$FD,$B7,$90,$F3,$3D,$5E,$D3,$CA,$5E,$EB,$59,$7F
	.dc.b	$C0,$18,$C5,$F0,$AD,$CB,$1F,$63,$71,$EF,$5B,$B9,$3B,$DB,$FB,$17
	.dc.b	$A2,$F3,$5E,$46,$B2,$E6,$35,$7C,$AB,$CC,$F6,$7C,$B3,$E3,$3D,$73
	.dc.b	$9D,$3B,$1E,$39,$D3,$8C,$3A,$4E,$DE,$FB,$AB,$71,$E2,$C7,$C3,$8F
	.dc.b	$83,$1E,$EC,$C7,$A6,$8F,$96,$7C,$46,$71,$E5,$ED,$BE,$57,$EA,$73
	.dc.b	$3E,$5B,$E9,$CE,$72,$F3,$6E,$AD,$47,$A7,$65,$6F,$1F,$4D,$69,$EF
	.dc.b	$8F,$6C,$7A,$B3,$0F,$85,$EE,$BE,$5B,$F0,$51,$D7,$1E,$9C,$BF,$E1
	.dc.b	$7C,$DF,$2A,$F8,$EF,$C6,$62,$F9,$6F,$C5,$CE,$71,$F4,$CE,$5F,$44
	.dc.b	$67,$5F,$9F,$39,$7E,$74,$67,$1F,$F1,$36,$EA,$D4,$78,$E7,$57,$F9
	.dc.b	$1C,$BC,$B9,$7E,$D6,$DE,$D3,$E5,$3C,$B3,$FA,$0F,$95,$7C,$07,$E9
	.dc.b	$7E,$21,$F2,$D9,$8C,$37,$3F,$A7,$3A,$7E,$95,$72,$73,$3B,$DF,$F1
	.dc.b	$13,$97,$F3,$7D,$0F,$8A,$FD,$6C,$6B,$1F,$D9,$FC,$6B,$D8,$FD,$C6
	.dc.b	$5F,$E6,$13,$F0,$11,$97,$F0,$A6,$6B,$F2,$D1,$97,$F3,$71,$9E,$2F
	.dc.b	$B5,$39,$7E,$6C,$64,$FD,$D7,$9D,$F8,$CD,$FF,$84,$7C,$C3,$CC,$FE
	.dc.b	$41,$39,$AF,$CA,$CE,$5F,$D6,$46,$71,$FD,$CB,$70,$FE,$59,$F2,$FF
	.dc.b	$C1,$EB,$7D,$3F,$55,$FD,$A5,$FF,$0B,$1E,$97,$F0,$DF,$85,$FE,$01
	.dc.b	$1F,$C3,$FE,$5F,$F6,$B1,$88,$E1,$3B,$F0,$F9,$56,$9F,$F4,$74,$66
	.dc.b	$79,$78,$9F,$E7,$3B,$E5,$1F,$C6,$E7,$4A,$F9,$3F,$E9,$F8,$CC,$62
	.dc.b	$3C,$22,$7C,$91,$98,$E8,$66,$FC,$B7,$E1,$7C,$B1,$93,$CB,$87,$89
	.dc.b	$E3,$7D,$2B,$84,$78,$CE,$07,$9D,$CB,$E4,$8E,$13,$CE,$3E,$69,$C2
	.dc.b	$FC,$CE,$27,$6F,$43,$C3,$7E,$8A,$E5,$87,$FA,$CA,$7D,$A7,$FA,$D1
	.dc.b	$C3,$D8,$FF,$63,$7C,$B3,$E5,$DB,$7F,$AE,$A3,$0D,$8C,$CF,$FB,$16
	.dc.b	$72,$DF,$17,$FB,$53,$77,$E5,$18,$8E,$ED,$7F,$19,$3F,$CD,$27,$EE
	.dc.b	$FC,$BF,$F1,$B9,$BD,$9F,$91,$D7,$F1,$97,$B9,$F9,$AD,$8F,$C8,$57
	.dc.b	$19,$FA,$2D,$8E,$38,$79,$9B,$87,$F6,$7F,$1B,$8E,$3F,$9F,$6E,$19
	.dc.b	$DB,$CF,$87,$BA,$78,$CF,$6A,$F3,$57,$E9,$F1,$F0,$3C,$30,$FC,$CE
	.dc.b	$5E,$78,$D5,$70,$BF,$EF,$17,$C1,$FE,$F5,$5B,$7F,$BE,$7E,$05,$F5
	.dc.b	$DF,$DA,$E4,$F8,$AF,$DB,$65,$FE,$27,$8C,$4F,$EE,$AB,$83,$FD,$E8
	.dc.b	$38,$9E,$15,$CB,$83,$F8,$05,$F2,$BE,$17,$B9,$FE,$09,$ED,$BD,$E7
	.dc.b	$C0,$7E,$DF,$D6,$CC,$63,$0F,$E1,$31,$89,$D4,$7F,$85,$FE,$01,$F0
	.dc.b	$3F,$C2,$2B,$53,$C0,$FF,$0E,$61,$DE,$F8,$C2,$30,$EC,$5F,$47,$71
	.dc.b	$AE,$65,$F6,$6C,$F6,$C7,$D1,$18,$8E,$78,$F9,$5D,$63,$D2,$B7,$7C
	.dc.b	$CF,$E5,$18,$F8,$6B,$C9,$8F,$86,$75,$1C,$8F,$44,$EF,$1F,$05,$6E
	.dc.b	$75,$18,$BE,$F1,$F2,$71,$F2,$3F,$0E,$32,$DC,$7C,$C7,$C8,$C3,$B6
	.dc.b	$3E,$29,$D1,$E6,$AE,$33,$A8,$F0,$CE,$F1,$F2,$47,$C4,$3A,$9C,$6E
	.dc.b	$E3,$F2,$6F,$94,$F6,$8F,$54,$EA,$FE,$3E,$3E,$AC,$3A,$93,$1F,$1C
	.dc.b	$65,$F6,$C6,$6B,$D9,$39,$7A,$61,$E8,$3E,$31,$83,$C5,$7F,$07,$1F
	.dc.b	$2D,$F3,$9F,$51,$F5,$2F,$A3,$8C,$7D,$2E,$4E,$93,$C6,$3E,$04,$F0
	.dc.b	$76,$4B,$F7,$D7,$0C,$7D,$67,$B9,$C6,$3E,$7C,$7C,$B3,$A8,$C6,$1D
	.dc.b	$6F,$EA,$4F,$1B,$F3,$46,$2B,$E2,$C7,$43,$27,$C4,$29,$F1,$0F,$90
	.dc.b	$7D,$D3,$38,$FB,$67,$D3,$18,$AF,$23,$F0,$CC,$57,$99,$F0,$63,$F0
	.dc.b	$5F,$03,$5F,$01,$31,$F7,$CF,$0B,$EF,$5B,$C3,$C5,$7C,$A7,$DC,$FA
	.dc.b	$AF,$8B,$E2,$3C,$95,$C7,$0F,$3B,$F5,$A3,$B1,$F0,$27,$8C,$6F,$0F
	.dc.b	$93,$3C,$E3,$D1,$3A,$BF,$0C,$73,$9D,$47,$E3,$71,$FA,$D1,$F3,$31
	.dc.b	$FD,$4D,$F1,$8C,$63,$F6,$70,$F0,$47,$AE,$F8,$BF,$1A,$F8,$C6,$1F
	.dc.b	$87,$1F,$5E,$78,$9E,$69,$D5,$7C,$0C,$3E,$59,$E8,$3C,$D5,$F0,$8F
	.dc.b	$A5,$3C,$31,$FB,$51,$EC,$3E,$34,$65,$F3,$1C,$70,$F2,$B6,$FD,$97
	.dc.b	$DE,$38,$99,$BF,$9B,$87,$A2,$7C,$73,$DC,$F1,$9B,$3E,$DB,$CD,$C9
	.dc.b	$BB,$F0,$B6,$FC,$2F,$CE,$BF,$5D,$F7,$8F,$19,$B8,$F1,$DF,$5B,$E7
	.dc.b	$5B,$AE,$ED,$8C,$CE,$EF,$BD,$7D,$77,$E8,$3F,$97,$3C,$47,$C8,$3C
	.dc.b	$07,$B0,$CD,$7D,$97,$E0,$3F,$8D,$3C,$11,$DA,$BF,$02,$78,$A7,$FA
	.dc.b	$61,$FE,$79,$3F,$9E,$95,$FE,$CE,$D8,$FE,$EC,$E7,$E0,$5F,$03,$C6
	.dc.b	$3F,$B8,$C6,$1C,$52,$31,$1C,$67,$83,$FD,$D6,$38,$5F,$F5,$C6,$E1
	.dc.b	$FE,$8D,$F6,$DF,$02,$DB,$FB,$BD,$6B,$0F,$0E,$38,$3F,$C0,$37,$C2
	.dc.b	$B1,$3F,$DB,$AB,$83,$E1,$AD,$DF,$58,$D4,$70,$C7,$C8,$EA,$F1,$8F
	.dc.b	$F8,$7B,$0E,$79,$7B,$EB,$58,$72,$26,$1C,$9F,$F0,$96,$5E,$57,$38
	.dc.b	$73,$AE,$B0,$13,$BA,$E4,$4A,$E1,$3D,$32,$F5,$4E,$63,$E2,$9D,$8E
	.dc.b	$95,$CE,$B8,$CE,$E7,$E2,$1C,$F2,$F1,$B8,$8D,$57,$18,$EB,$1C,$71
	.dc.b	$F8,$51,$C4,$F0,$1C,$6F,$B9,$F7,$70,$F3,$B8,$8E,$6D,$7E,$0C,$69
	.dc.b	$C9,$0E,$87,$B1,$D3,$82,$4F,$92,$BF,$C7,$17,$E9,$9C,$E5,$F6,$8F
	.dc.b	$C2,$C6,$6B,$86,$1E,$36,$D7,$7C,$7E,$2C,$62,$3C,$15,$C6,$FA,$65
	.dc.b	$F2,$A3,$0F,$08,$EF,$97,$A2,$F9,$12,$33,$86,$EB,$95,$6E,$7A,$E1
	.dc.b	$F3,$B0,$EA,$FD,$0C,$3C,$91,$8C,$BE,$B5,$74,$AD,$3C,$A3,$E2,$DF
	.dc.b	$5C,$3C,$15,$F6,$E7,$8C,$78,$B1,$CC,$7E,$12,$B8,$57,$4A,$FD,$1C
	.dc.b	$F0,$8F,$34,$EA,$30,$78,$2B,$9E,$5F,$81,$C7,$D0,$FD,$1C,$7D,$15
	.dc.b	$C2,$35,$7C,$DD,$63,$E0,$6E,$3E,$0C,$7A,$BF,$89,$C7,$AB,$F8,$BC
	.dc.b	$BE,$3E,$1C,$E3,$19,$78,$30,$E7,$97,$BF,$1E,$75,$F8,$0B,$E0,$FE
	.dc.b	$33,$2F,$13,$C9,$F9,$0F,$C5,$3C,$85,$AE,$8D,$BF,$6D,$F6,$C7,$CA
	.dc.b	$DC,$7C,$95,$D4,$E7,$7F,$15,$F7,$63,$E4,$BF,$C6,$63,$E4,$9F,$31
	.dc.b	$C3,$1F,$69,$E0,$7C,$47,$2C,$7C,$13,$C6,$B9,$CF,$13,$C5,$87,$A8
	.dc.b	$F1,$3F,$16,$B8,$BF,$8B,$BE,$EF,$53,$B9,$B2,$D7,$C0,$6E,$3F,$01
	.dc.b	$F4,$DF,$7A,$E7,$1C,$63,$C3,$5C,$5B,$5C,$63,$D3,$1C,$67,$D3,$5F
	.dc.b	$06,$37,$87,$BE,$FF,$4D,$39,$3B,$DF,$56,$C7,$53,$77,$ED,$BE,$6F
	.dc.b	$C1,$3F,$8D,$7D,$A3,$C2,$FE,$A5,$C9,$FD,$4F,$CD,$7D,$CB,$E1,$B3
	.dc.b	$B7,$F4,$5C,$58,$F1,$8F,$E9,$18,$71,$BE,$11,$83,$84,$EE,$95,$BB
	.dc.b	$EB,$1F,$D4,$67,$53,$9A,$E5,$3F,$E0,$9C,$BF,$AB,$76,$BB,$3C,$9E
	.dc.b	$03,$C1,$FE,$D9,$39,$BE,$01,$C2,$9E,$18,$74,$BF,$F0,$76,$16,$BF
	.dc.b	$B2,$56,$9E,$53,$A6,$9A,$C3,$C3,$19,$EF,$EF,$5C,$5F,$0D,$7F,$6F
	.dc.b	$BF,$EE,$11,$AB,$FF,$02,$3A,$94,$F8,$DF,$44,$F6,$7D,$B8,$F1,$C7
	.dc.b	$51,$D4,$F2,$9C,$1C,$BD,$0C,$E1,$ED,$9F,$05,$FF,$8A,$27,$91,$FE
	.dc.b	$2A,$70,$F1,$3F,$C7,$9D,$AF,$01,$E0,$FF,$8C,$23,$0D,$8C,$CF,$F8
	.dc.b	$EE,$72,$DE,$EF,$C2,$B7,$6E,$51,$88,$EE,$D7,$F2,$92,$8E,$38,$FD
	.dc.b	$EF,$01,$E0,$FE,$66,$30,$DE,$0E,$35,$C1,$B8,$FC,$FB,$E3,$8D,$BE
	.dc.b	$0F,$F8,$9F,$1F,$05,$7D,$56,$C7,$D0,$AE,$2D,$AE,$2A,$77,$8F,$E5
	.dc.b	$1B,$87,$EB,$EF,$8C,$FA,$2B,$B6,$3E,$77,$C4,$D5,$3C,$2B,$96,$1E
	.dc.b	$0C,$3A,$D7,$F8,$FF,$C0,$FB,$0F,$EB,$8F,$95,$E0,$FE,$FF,$3F,$F2
	.dc.b	$B4,$60,$D4,$F0,$7F,$77,$C1,$FC,$16,$31,$3C,$33,$5F,$3A,$FE,$15
	.dc.b	$18,$BE,$7A,$E5,$FF,$47,$F4,$FF,$AC,$38,$3F,$D1,$BD,$FF,$EA,$29
	.dc.b	$C9,$FE,$A6,$8C,$BF,$E9,$73,$FD,$3D,$1F,$EA,$C9,$FF,$51,$3C,$63
	.dc.b	$27,$FA,$7E,$78,$9F,$E9,$B9,$DC,$73,$EB,$FF,$4D,$F7,$FF,$A7,$4F
	.dc.b	$F5,$4C,$7F,$AC,$27,$75,$C6,$3C,$A6,$4B,$5F,$EA,$37,$C3,$1F,$EA
	.dc.b	$B7,$FD,$5E,$F1,$73,$3F,$04,$EA,$7F,$B9,$9E,$75,$E3,$BF,$F7,$37
	.dc.b	$83,$FE,$F9,$9E,$27,$FA,$F4,$A7,$8C,$B8,$78,$CF,$6C,$FC,$56,$C7
	.dc.b	$C1,$AF,$7C,$6D,$A7,$86,$3F,$DC,$95,$E1,$9F,$F8,$02,$D9,$F8,$ED
	.dc.b	$AF,$A9,$5E,$A7,$C1,$3D,$1A,$7B,$3C,$1F,$F7,$E6,$3F,$1C,$FA,$F8
	.dc.b	$FC,$3B,$F5,$DF,$8E,$7F,$DC,$98,$77,$9F,$BD,$5F,$29,$F8,$98,$FB
	.dc.b	$AB,$E8,$36,$7A,$36,$77,$7E,$57,$E2,$63,$E5,$7F,$DC,$75,$C2,$75
	.dc.b	$7D,$1B,$19,$AE,$15,$F8,$08,$C3,$F2,$B1,$F7,$C7,$E0,$63,$E3,$1E
	.dc.b	$03,$47,$80,$F2,$CF,$4B,$F8,$B1,$83,$67,$73,$0F,$C6,$AE,$F1,$FA
	.dc.b	$BA,$F8,$CB,$1F,$95,$5F,$46,$73,$18,$7C,$D3,$F4,$AF,$EC,$BE,$0A
	.dc.b	$F1,$BC,$EF,$BB,$F9,$3C,$7E,$73,$93,$D8,$53,$D0,$59,$E7,$87,$86
	.dc.b	$3D,$07,$C6,$8F,$A1,$7F,$0F,$0F,$3E,$1F,$E1,$99,$47,$F8,$6F,$06
	.dc.b	$2B,$FF,$0F,$61,$98,$C3,$81,$3C,$27,$11,$89,$DB,$CA,$7F,$DA,$AD
	.dc.b	$EE,$77,$F0,$9C,$CF,$FB,$B0,$D4,$7F,$9E,$31,$FF,$1C,$9F,$E9,$78
	.dc.b	$FF,$29,$C7,$F9,$14,$FF,$41,$F8,$3F,$E8,$C8,$FF,$2A,$E5,$FE,$DF
	.dc.b	$DF,$FC,$C3,$7C,$4E,$71,$99,$E6,$7F,$91,$4E,$27,$F9,$25,$FF,$26
	.dc.b	$DF,$78,$FF,$33,$DF,$7C,$3F,$CB,$75,$CE,$FA,$D7,$FB,$5C,$EA,$F8
	.dc.b	$63,$9C,$77,$9E,$2F,$73,$FC,$C6,$53,$C6,$7F,$9B,$63,$8C,$F3,$8D
	.dc.b	$99,$2C,$FF,$99,$5B,$8F,$F9,$97,$0F,$8F,$E0,$FF,$A5,$63,$FC,$CF
	.dc.b	$3F,$F0,$03,$2F,$8F,$8F,$AC,$EE,$58,$E7,$7F,$0F,$1F,$F4,$7E,$5F
	.dc.b	$0F,$0E,$6D,$9F,$B5,$7C,$AF,$E2,$5F,$C6,$7C,$07,$C6,$8E,$CF,$C0
	.dc.b	$CB,$D4,$DC,$3D,$4F,$D2,$CB,$D5,$87,$68,$F2,$9F,$36,$7D,$B5,$C8
	.dc.b	$F9,$F1,$F1,$CE,$93,$EB,$BF,$F4,$DE,$3E,$99,$FF,$57,$1A,$3A,$4F
	.dc.b	$CA,$CB,$E3,$C7,$B2,$BD,$CF,$B3,$2F,$75,$F3,$7F,$D4,$11,$CD,$FB
	.dc.b	$4D,$9F,$83,$87,$98,$F4,$63,$EA,$3E,$0D,$7D,$0C,$BE,$0D,$F9,$AB
	.dc.b	$85,$6A,$FD,$75,$AB,$F0,$D6,$B2,$F8,$18,$7E,$2B,$1F,$36,$1F,$EA
	.dc.b	$F8,$F9,$77,$E6,$BE,$11,$F0,$8F,$A5,$97,$C2,$7E,$05,$7C,$2B,$E7
	.dc.b	$1F,$9B,$AE,$37,$CF,$2F,$66,$3E,$2A,$FC,$FD,$F0,$7E,$B5,$F0,$C3
	.dc.b	$E2,$DF,$09,$F1,$61,$F0,$F1,$F1,$61,$C7,$0F,$93,$8F,$D7,$BF,$F4
	.dc.b	$5E,$3F,$C8,$71,$FE,$27,$18,$3F,$D6,$5E,$0F,$FA,$D3,$C1,$FF,$61
	.dc.b	$F8,$1F,$EB,$69,$FF,$60,$CE,$7C,$1F,$F5,$FC,$7F,$90,$8E,$3A,$7F
	.dc.b	$AE,$2F,$8B,$FE,$4D,$F0,$7F,$D7,$AA,$53,$A7,$91,$FC,$DC,$FF,$4A
	.dc.b	$F8,$1F,$E9,$F7,$1D,$DF,$E9,$F7,$FD,$3A,$7F,$A7,$DC,$1F,$E9,$A7
	.dc.b	$8D,$7F,$A8,$38,$3F,$D5,$5E,$07,$FA,$7F,$BB,$FD,$75,$7C,$1A,$7F
	.dc.b	$B3,$30,$F0,$61,$D6,$31,$87,$38,$EA,$78,$8F,$F5,$65,$73,$7F,$DA
	.dc.b	$27,$FA,$A4,$FF,$56,$61,$FE,$AA,$3F,$D4,$9E,$04,$8F,$F5,$27,$4F
	.dc.b	$FA,$99,$FF,$56,$B5,$FF,$4A,$9F,$EA,$52,$1F,$EA,$4E,$CF,$F5,$39
	.dc.b	$FE,$97,$61,$FE,$C2,$EE,$FF,$40,$BF,$E9,$1E,$CF,$F4,$47,$81,$FE
	.dc.b	$87,$8F,$A8,$D7,$55,$FE,$90,$39,$E5,$FE,$93,$8E,$95,$FE,$91,$BE
	.dc.b	$B7,$C4,$E4,$F8,$0C,$F8,$1F,$EA,$28,$CC,$FF,$A6,$7C,$0F,$F4,$4E
	.dc.b	$BF,$E8,$6F,$03,$FD,$09,$8F,$FA,$44,$FC,$2C,$E9,$FF,$4D,$69,$AB
	.dc.b	$FF,$46,$47,$E2,$9B,$A7,$FA,$6A,$B8,$4F,$F8,$5C,$FF,$16,$6D,$FE
	.dc.b	$23,$7F,$C2,$3E,$07,$F8,$35,$FF,$0B,$77,$7F,$8E,$9F,$F1,$4F,$57
	.dc.b	$F8,$17,$BF,$FC,$B7,$1C,$EF,$FC,$22,$FF,$88,$67,$FC,$22,$F4,$F0
	.dc.b	$3F,$C4,$3D,$DF,$E6,$BB,$FF,$0B,$4F,$F8,$66,$BF,$C8,$77,$E3,$8F
	.dc.b	$F1,$0E,$3E,$3B,$E1,$1D,$4E,$07,$90,$FF,$48,$C6,$F1,$FF,$32,$5F
	.dc.b	$9E,$7F,$CC,$67,$03,$FC,$58,$F9,$9E,$F8,$F9,$A7,$8C,$74,$8C,$63
	.dc.b	$FE,$3E,$C7,$CC,$7F,$8D,$23,$33,$F3,$2F,$C1,$3F,$E4,$98,$F0,$61
	.dc.b	$E5,$BE,$55,$A3,$FC,$97,$8F,$F9,$12,$75,$5C,$CF,$11,$F2,$C8,$7A
	.dc.b	$0B,$7E,$36,$D7,$B6,$31,$1B,$23,$82,$46,$C9,$1E,$8C,$3D,$B3,$F7
	.dc.b	$8F,$83,$3A,$9F,$C1,$3F,$68,$F6,$3F,$E5,$63,$27,$D7,$3F,$D6,$B5
	.dc.b	$EE,$9E,$93,$C2,$7F,$23,$87,$E1,$9E,$27,$D8,$3C,$87,$B8,$A7,$B2
	.dc.b	$3C,$98,$7C,$BC,$3D,$31,$E5,$9F,$96,$6C,$E3,$7F,$09,$B1,$CB,$1F
	.dc.b	$D1,$47,$84,$FF,$30,$61,$FE,$5B,$8F,$F3,$0C,$7E,$3C,$FF,$31,$38
	.dc.b	$7F,$A0,$61,$FE,$60,$6F,$6F,$F5,$27,$FC,$C2,$7F,$9B,$E3,$4F,$F6
	.dc.b	$4F,$03,$FB,$D6,$1F,$DD,$63,$FC,$10,$7F,$84,$4F,$EE,$EF,$F6,$6E
	.dc.b	$FF,$EC,$19,$7F,$84,$AB,$8D,$7F,$73,$C3,$FB,$54,$FF,$73,$AD,$DF
	.dc.b	$F7,$9B,$DD,$FF,$72,$BD,$CF,$58,$FF,$00,$9F,$E3,$93,$C4,$79,$0F
	.dc.b	$F0,$19,$E4,$33,$87,$F7,$B6,$CF,$46,$DF,$89,$E0,$72,$3A,$78,$1F
	.dc.b	$DF,$67,$FC,$3D,$1F,$E1,$9C,$3C,$F7,$FE,$00,$8F,$2E,$3F,$07,$8F
	.dc.b	$FC,$0B,$5A,$BE,$35,$F3,$67,$24,$7D,$87,$86,$FD,$0D,$9F,$43,$71
	.dc.b	$F4,$3E,$6B,$ED,$1F,$E1,$DA,$FF,$28,$E3,$E9,$8C,$5F,$F9,$A2,$B7
	.dc.b	$7F,$E1,$CC,$BF,$C7,$38,$F8,$F2,$F5,$4F,$93,$1E,$4E,$CF,$85,$3B
	.dc.b	$AF,$B4,$D8,$F9,$73,$C2,$BF,$01,$1F,$1A,$B6,$F5,$BF,$1C,$E8,$FF
	.dc.b	$17,$46,$6F,$A3,$EF,$6E,$5F,$1B,$0E,$D8,$7E,$1E,$F8,$9E,$A2,$3E
	.dc.b	$06,$DF,$9B,$0F,$0C,$F1,$3E,$1C,$78,$49,$3C,$1F,$11,$88,$E8,$7F
	.dc.b	$8F,$F2,$FF,$18,$4F,$FB,$8A,$7F,$C6,$E7,$F8,$E7,$B3,$FC,$65,$C2
	.dc.b	$CF,$FC,$6D,$3F,$E4,$07,$FC,$79,$DD,$FE,$40,$9D,$4F,$26,$D7,$2A
	.dc.b	$E3,$DD,$FE,$40,$AF,$F1,$FD,$F1,$9F,$F1,$7C,$7F,$90,$FB,$BF,$C6
	.dc.b	$8F,$FB,$CB,$1F,$F1,$BF,$77,$F8,$CA,$3F,$C5,$BD,$DF,$E3,$27,$FC
	.dc.b	$61,$3A,$BF,$F1,$A7,$6F,$F8,$C6,$7F,$C8,$F5,$B8,$FF,$22,$19,$8F
	.dc.b	$F1,$E1,$E1,$BE,$D7,$CE,$3A,$9F,$E3,$E9,$FF,$26,$3F,$E3,$B2,$1F
	.dc.b	$18,$F9,$A7,$F8,$EB,$BB,$FC,$78,$FF,$8E,$9A,$FF,$8E,$8F,$F1,$C1
	.dc.b	$FE,$3A,$30,$7F,$8F,$A3,$FC,$5C,$4B,$C1,$23,$FC,$6F,$DB,$FE,$2B
	.dc.b	$3E,$BC,$A7,$53,$FE,$3A,$11,$CD,$E2,$7F,$8F,$4F,$F1,$DE,$DF,$E3
	.dc.b	$E9,$FC,$0F,$6F,$F8,$DA,$35,$5C,$69,$39,$7A,$D7,$E2,$3A,$BF,$C5
	.dc.b	$FD,$7F,$E3,$2D,$7F,$C7,$95,$FE,$81,$C7,$FC,$69,$D9,$FE,$3E,$BD
	.dc.b	$F6,$FF,$8E,$CC,$47,$F9,$0B,$A3,$FA,$D5,$FF,$5F,$3F,$B1,$F6,$FF
	.dc.b	$69,$E7,$FE,$B9,$1F,$DB,$3B,$7F,$AF,$BF,$DB,$E9,$D8,$ED,$FF,$1C
	.dc.b	$C7,$49,$FE,$C4,$74,$AE,$2F,$F5,$F8,$D7,$27,$F6,$58,$D1,$E3,$8C
	.dc.b	$E1,$FD,$9E,$FF,$BC,$46,$27,$FB,$59,$0E,$27,$81,$D3,$8A,$ED,$18
	.dc.b	$26,$1F,$DC,$B1,$F8,$15,$ED,$8F,$63,$FE,$3E,$39,$61,$E3,$7F,$C9
	.dc.b	$06,$0F,$15,$74,$C3,$CF,$5D,$71,$F8,$38,$76,$BF,$85,$39,$7E,$03
	.dc.b	$5F,$53,$71,$F4,$B7,$1F,$14,$7F,$8A,$27,$C3,$39,$6E,$1E,$22,$3E
	.dc.b	$8C,$7C,$CD,$9E,$0F,$AC,$91,$AA,$ED,$18,$9D,$CE,$6F,$FC,$5C,$FF
	.dc.b	$9B,$2B,$6E,$6B,$FC,$4B,$1E,$AB,$F1,$BA,$BE,$64,$8C,$DF,$48,$FA
	.dc.b	$73,$FE,$84,$AF,$F1,$1D,$7F,$89,$C8,$FF,$87,$FB,$7F,$C4,$4F,$F3
	.dc.b	$73,$1D,$7F,$C3,$5F,$F1,$3B,$5F,$EF,$07,$F3,$F3,$FA,$47,$07,$F5
	.dc.b	$29,$FE,$EB,$DB,$FD,$4A,$7F,$C1,$A7,$F5,$F3,$67,$13,$FA,$4E,$1F
	.dc.b	$D2,$F3,$FE,$B5,$DB,$FD,$4A,$78,$CF,$F5,$AD,$3F,$A6,$C7,$28,$E1
	.dc.b	$87,$82,$B8,$4F,$F8,$14,$F0,$1D,$AB,$8B,$4E,$4E,$2F,$C3,$8F,$AF
	.dc.b	$2F,$F0,$6C,$62,$BD,$0D,$C7,$C3,$3A,$C7,$C5,$7C,$6F,$C3,$7B,$C3
	.dc.b	$B5,$6A,$79,$B7,$0E,$71,$C2,$31,$3C,$30,$EC,$49,$D3,$6B,$E3,$47
	.dc.b	$B0,$87,$BC,$97,$E3,$72,$D7,$37,$E0,$72,$D7,$35,$DE,$FF,$C0,$B7
	.dc.b	$E8,$3C,$33,$A3,$CA,$47,$37,$DC,$97,$D5,$B7,$D6,$BF,$C5,$A7,$C6
	.dc.b	$7F,$C1,$77,$FD,$FB,$BF,$FC,$07,$C7,$FE,$02,$EC,$FE,$FD,$7F,$E0
	.dc.b	$5C,$CF,$F8,$0B,$C0,$DC,$71,$3F,$C1,$AE,$23,$81,$B8,$FE,$90,$7F
	.dc.b	$82,$17,$6F,$F8,$21,$FE,$FE,$7F,$7B,$8F,$F0,$1F,$47,$F8,$04,$E1
	.dc.b	$3B,$9F,$F0,$09,$CB,$B7,$FC,$02,$FF,$7F,$9D,$C6,$3A,$FF,$C1,$13
	.dc.b	$FE,$12,$3B,$1F,$E0,$67,$3C,$1F,$B5,$8F,$F0,$19,$FE,$06,$3F,$C3
	.dc.b	$87,$F8,$14,$87,$F8,$14,$A7,$F8,$08,$F8,$FE,$03,$AD,$AF,$F0,$18
	.dc.b	$C3,$63,$33,$FC,$1E,$72,$FF,$74,$3F,$BF,$94,$FE,$F3,$DB,$FE,$08
	.dc.b	$CB,$FC,$0F,$3C,$31,$FE,$FD,$DB,$FD,$F5,$FF,$06,$5F,$93,$0E,$FE
	.dc.b	$0E,$23,$99,$F7,$9C,$53,$1F,$F2,$5C,$6F,$0F,$2B,$6F,$D3,$87,$82
	.dc.b	$BC,$98,$77,$CB,$FC,$23,$97,$F8,$52,$7F,$C4,$13,$83,$F1,$BC,$5F
	.dc.b	$E5,$47,$E3,$EF,$F9,$0E,$F7,$E1,$DF,$E6,$19,$C8,$8C,$D7,$F9,$83
	.dc.b	$1E,$4F,$36,$DE,$4E,$97,$CE,$3F,$CC,$75,$CA,$FF,$2E,$79,$2B,$FC
	.dc.b	$35,$D7,$FD,$1E,$3F,$AD,$75,$FF,$48,$7F,$A7,$E3,$FD,$98,$FE,$95
	.dc.b	$1F,$AE,$8F,$EF,$CF,$F8,$62,$79,$70,$FF,$53,$8E,$53,$FE,$21,$3F
	.dc.b	$BA,$B8,$E8,$FF,$10,$BC,$4C,$57,$F5,$6B,$FF,$3F,$CF,$F6,$37,$C3
	.dc.b	$4B,$F9,$1D,$9F,$E0,$32,$76,$FF,$81,$89,$3E,$03,$DA,$49,$F3,$36
	.dc.b	$BA,$C7,$49,$E3,$3F,$E0,$AC,$7C,$A7,$9D,$E0,$7C,$63,$FC,$0E,$60
	.dc.b	$F8,$07,$13,$18,$7C,$07,$FC,$48,$FF,$7D,$77,$5F,$DF,$5D,$E3,$F0
	.dc.b	$DF,$E2,$F3,$F0,$DF,$01,$FE,$0B,$AF,$2B,$D2,$7E,$9E,$1E,$C3,$FC
	.dc.b	$A7,$5E,$1B,$FC,$83,$CE,$BC,$33,$F7,$DE,$B1,$E5,$38,$47,$D4,$AF
	.dc.b	$24,$FF,$8C,$AF,$C3,$7D,$27,$38,$FD,$4A,$FA,$53,$F4,$09,$7E,$58
	.dc.b	$FC,$59,$E9,$BE,$71,$DD,$FC,$84,$77,$3E,$2B,$F4,$E7,$9C,$7A,$2F
	.dc.b	$E0,$B6,$79,$BF,$2A,$3C,$B8,$FD,$EA,$FA,$07,$E1,$27,$FC,$4A,$7F
	.dc.b	$87,$F1,$FB,$0E,$0F,$F5,$72,$EB,$FF,$13,$C7,$F8,$90,$FF,$13,$90
	.dc.b	$DA,$EB,$F5,$CF,$17,$DB,$DB,$FE,$25,$AF,$F0,$B5,$FF,$64,$BF,$EC
	.dc.b	$44,$AF,$EC,$53,$FE,$30,$9F,$ED,$D9,$FF,$70,$3F,$C0,$91,$FD,$B5
	.dc.b	$C3,$FD,$E2,$78,$E1,$E1,$8F,$D6,$4F,$F8,$FE,$BF,$8A,$61,$E3,$27
	.dc.b	$6F,$F8,$A9,$BB,$7F,$86,$E7,$CD,$8F,$4C,$BD,$D8,$F4,$3E,$24,$F4
	.dc.b	$C3,$91,$2F,$96,$3E,$6A,$FF,$03,$57,$47,$D4,$7F,$80,$1F,$7B,$63
	.dc.b	$DF,$1B,$23,$CF,$2F,$F0,$4C,$F2,$3E,$44,$70,$35,$87,$CD,$8D,$4F
	.dc.b	$C5,$AF,$D3,$CF,$B1,$F5,$5F,$A2,$BC,$47,$CD,$38,$CE,$EF,$E6,$1F
	.dc.b	$E1,$08,$FB,$AF,$83,$1F,$90,$7E,$12,$78,$4F,$F9,$7A,$74,$F4,$9D
	.dc.b	$1F,$88,$C7,$E5,$47,$F8,$8C,$D1,$F4,$46,$3E,$BA,$F1,$65,$F6,$A7
	.dc.b	$86,$5F,$21,$F5,$9F,$4A,$7C,$31,$98,$F9,$27,$BD,$C5,$75,$8F,$BF
	.dc.b	$3D,$AF,$CB,$5F,$62,$78,$9D,$4E,$F5,$F5,$A3,$FB,$27,$DD,$BC,$0F
	.dc.b	$F1,$97,$0B,$F4,$97,$CD,$FF,$8D,$71,$5E,$F7,$FE,$BF,$7B,$AD,$4E
	.dc.b	$6B,$FB,$7C,$E6,$BF,$B2,$C7,$2C,$3B,$E1,$DA,$73,$3F,$E3,$6E,$EF
	.dc.b	$F1,$C3,$FE,$31,$3F,$C7,$11,$FE,$3E,$7F,$C6,$7D,$3F,$E1,$37,$87
	.dc.b	$77,$F8,$F3,$B3,$FC,$2D,$68,$5F,$F8,$5F,$AF,$FD,$5D,$B5,$C3,$FC
	.dc.b	$7B,$87,$37,$C1,$87,$8A,$7F,$CB,$78,$FF,$90,$27,$FC,$90,$7F,$8F
	.dc.b	$9C,$3F,$E4,$13,$F2,$51,$FE,$45,$7F,$B6,$1F,$E4,$03,$FC,$7F,$E1
	.dc.b	$7F,$84,$DC,$0E,$E4,$66,$3F,$38,$67,$BB,$FC,$82,$E2,$7F,$C8,$2D
	.dc.b	$8C,$BF,$E4,$D3,$87,$67,$AB,$3F,$F1,$64,$53,$FC,$8F,$8F,$F8,$D6
	.dc.b	$7F,$C5,$D1,$FE,$3E,$3F,$C9,$5D,$FF,$E4,$5A,$FD,$76,$1F,$E4,$97
	.dc.b	$9C,$EB,$BF,$FC,$76,$F8,$0F,$F2,$94,$ED,$EF,$3C,$30,$F0,$46,$B1
	.dc.b	$FF,$22,$CE,$49,$1F,$E4,$53,$FC,$B9,$CF,$FE,$37,$DF,$FC,$6A,$FF
	.dc.b	$8D,$E3,$FC,$7D,$DF,$FE,$37,$9F,$E0,$93,$B9,$FF,$1B,$1F,$9A,$9D
	.dc.b	$3C,$23,$92,$78,$3F,$E0,$33,$FB,$29,$FD,$AC,$BB,$7F,$75,$EE,$FE
	.dc.b	$D7,$DF,$FD,$D6,$7F,$C2,$D3,$FD,$A9,$FF,$01,$9F,$E1,$C7,$1D,$3F
	.dc.b	$E0,$28,$FE,$D5,$5C,$BB,$BF,$B5,$61,$FE,$09,$9E,$27,$8E,$F8,$F7
	.dc.b	$FF,$82,$30,$FF,$08,$1C,$E7,$F8,$AC,$7F,$74,$7F,$C0,$F5,$FE,$05
	.dc.b	$7F,$C0,$15,$CB,$0F,$F0,$04,$E7,$1F,$54,$FF,$7A,$8C,$63,$E4,$BF
	.dc.b	$54,$61,$EA,$7F,$81,$4F,$8A,$FF,$83,$DF,$E5,$15,$D2,$3D,$98,$F9
	.dc.b	$CF,$4E,$3E,$B3,$D1,$5F,$2A,$7B,$34,$ED,$87,$A5,$FA,$8D,$3E,$A3
	.dc.b	$E0,$C3,$D0,$FC,$99,$E7,$3E,$D8,$E7,$18,$3C,$B5,$E3,$C7,$EA,$4F
	.dc.b	$B8,$FA,$78,$7F,$8E,$E7,$51,$DC,$D4,$7B,$63,$D7,$7D,$E7,$8C,$7F
	.dc.b	$8B,$1B,$5F,$E2,$C7,$EC,$BE,$07,$E0,$1A,$3E,$69,$F8,$83,$C0,$6A
	.dc.b	$BC,$11,$C6,$BC,$11,$89,$DC,$62,$BD,$2F,$F9,$DE,$3F,$CF,$C7,$F8
	.dc.b	$D4,$FB,$45,$3F,$C6,$87,$F8,$CB,$BB,$F8,$BB,$83,$FD,$1F,$DF,$FE
	.dc.b	$3A,$7F,$AA,$DF,$F5,$3C,$7F,$B0,$92,$3F,$AB,$CF,$F8,$1E,$FF,$B0
	.dc.b	$BF,$DE,$23,$FB,$7B,$7B,$FF,$B7,$4F,$F6,$1B,$FE,$CD,$5A,$C3,$FB
	.dc.b	$D6,$DF,$E0,$B8,$F0,$E1,$D7,$1F,$04,$62,$BF,$C0,$EF,$F7,$DB,$E7
	.dc.b	$8F,$6A,$FF,$0E,$4F,$9A,$B7,$8F,$F8,$69,$EF,$1F,$E0,$09,$EB,$7E
	.dc.b	$43,$9F,$7F,$F7,$88,$E7,$3E,$13,$FC,$1B,$7C,$DF,$F0,$76,$1E,$17
	.dc.b	$FC,$09,$3D,$31,$F8,$98,$74,$CB,$DC,$F7,$8F,$8D,$8F,$7B,$E2,$7A
	.dc.b	$4E,$F7,$FE,$2A,$7F,$C4,$86,$71,$FA,$39,$7F,$89,$23,$C7,$5C,$DF
	.dc.b	$F0,$71,$E9,$9D,$BF,$26,$FC,$38,$F9,$1F,$39,$88,$F8,$C7,$7C,$3D
	.dc.b	$98,$77,$BF,$F2,$1C,$E8,$ED,$3A,$7B,$4F,$D0,$9F,$C1,$57,$68,$C5
	.dc.b	$7C,$EC,$7E,$24,$62,$FE,$19,$D0,$F1,$10,$C1,$23,$8B,$F0,$B0,$F6
	.dc.b	$4F,$9E,$BF,$C6,$91,$89,$F2,$9B,$3D,$55,$CB,$0F,$69,$F7,$CF,$F1
	.dc.b	$AF,$67,$F8,$9B,$BF,$FC,$5A,$DB,$FF,$15,$F7,$FF,$8B,$FA,$7F,$C3
	.dc.b	$B8,$BB,$FF,$C6,$0A,$32,$FF,$8C,$6B,$FC,$73,$DF,$FE,$31,$8D,$47
	.dc.b	$F8,$91,$FF,$0E,$9F,$E1,$A2,$9F,$E2,$E7,$FC,$64,$FF,$8D,$07,$47
	.dc.b	$F8,$A7,$3F,$F0,$64,$66,$30,$F2,$3F,$C5,$BD,$FE,$88,$C9,$FE,$2D
	.dc.b	$27,$47,$F9,$EF,$BB,$FC,$65,$5F,$C3,$01,$FE,$2C,$3F,$C6,$06,$2F
	.dc.b	$5D,$7F,$E2,$AD,$2F,$77,$F8,$AE,$BF,$C3,$2D,$7F,$C6,$71,$C3,$B3
	.dc.b	$FC,$53,$1F,$E0,$F3,$FC,$59,$B6,$4F,$17,$87,$FE,$36,$8F,$F2,$EB
	.dc.b	$5F,$96,$79,$9C,$47,$D2,$97,$A1,$FE,$0E,$9C,$EF,$FE,$99,$9F,$F3
	.dc.b	$A5,$FF,$9A,$1A,$F4,$13,$C1,$F8,$7C,$EF,$80,$FF,$0C,$B6,$3F,$CE
	.dc.b	$2D,$79,$7E,$3A,$EA,$DD,$49,$DB,$FE,$22,$9D,$46,$0F,$F3,$DA,$F8
	.dc.b	$CD,$1C,$63,$CD,$D3,$FD,$8F,$BB,$FC,$17,$1F,$DB,$08,$7F,$59,$E2
	.dc.b	$FE,$ED,$D7,$FE,$05,$7F,$A8,$47,$F5,$9E,$EF,$EA,$73,$FE,$BD,$F0
	.dc.b	$3F,$C1,$13,$FE,$10,$39,$E1,$FD,$6E,$F8,$9F,$D6,$F0,$F0,$CF,$7B
	.dc.b	$FE,$F8,$F1,$30,$72,$C3,$BB,$F8,$EC,$3B,$BF,$E0,$73,$55,$D3,$0F
	.dc.b	$05,$F9,$7C,$0F,$F0,$6C,$7C,$1A,$FF,$09,$CF,$81,$E4,$78,$09,$1C
	.dc.b	$9F,$04,$F2,$C3,$FC,$09,$39,$9F,$F1,$BD,$73,$8E,$D8,$FF,$87,$8F
	.dc.b	$83,$1F,$E1,$8C,$3C,$D7,$EC,$BE,$27,$A8,$F2,$96,$B9,$57,$AE,$3F
	.dc.b	$C3,$33,$C7,$1F,$87,$87,$82,$BE,$C1,$BA,$FB,$18,$78,$08,$F6,$3D
	.dc.b	$F7,$CF,$1F,$74,$FF,$87,$EF,$FC,$82,$F0,$C7,$C7,$7F,$10,$F1,$C6
	.dc.b	$70,$F0,$C6,$4F,$09,$DE,$BE,$46,$3E,$8C,$7E,$59,$E1,$9E,$66,$DC
	.dc.b	$3C,$B0,$F0,$BF,$20,$F4,$E1,$FC,$3F,$0F,$0C,$2D,$8F,$FB,$0A,$7F
	.dc.b	$C4,$F3,$0F,$F1,$B1,$FE,$29,$3F,$C2,$C7,$F8,$A3,$16,$DF,$E2,$76
	.dc.b	$F7,$FF,$21,$6C,$EE,$7F,$AF,$3F,$D3,$72,$FE,$C3,$5F,$D4,$E3,$FB
	.dc.b	$D9,$FD,$66,$77,$3F,$D4,$2F,$FB,$24,$61,$FE,$B3,$3F,$D5,$A3,$0F
	.dc.b	$2B,$FE,$B1,$3A,$AE,$F3,$FD,$79,$E8,$FF,$6C,$9D,$BC,$6B,$9A,$F0
	.dc.b	$3F,$C3,$8F,$F8,$2A,$76,$E7,$83,$FC,$1D,$1F,$E4,$78,$FE,$E8,$EA
	.dc.b	$78,$B9,$8F,$76,$3F,$05,$E5,$C7,$FD,$CF,$1F,$ED,$D9,$7A,$4F,$83
	.dc.b	$5B,$C3,$9E,$5F,$E4,$5C,$3B,$BE,$88,$CC,$FF,$90,$2B,$FC,$51,$5D
	.dc.b	$8F,$91,$1C,$F2,$F4,$DF,$39,$F2,$9F,$1E,$BC,$4F,$C7,$8E,$D3,$C2
	.dc.b	$BB,$4F,$F6,$0C,$BF,$C7,$CF,$63,$91,$F3,$63,$C1,$97,$BF,$0F,$2E
	.dc.b	$3E,$8A,$FF,$1F,$9C,$F1,$FF,$19,$65,$F6,$8F,$8B,$1C,$CE,$A7,$B5
	.dc.b	$D5,$74,$CB,$CD,$3D,$0F,$CD,$CF,$E0,$9F,$B9,$5E,$CC,$3C,$91,$E1
	.dc.b	$C7,$FC,$64,$E9,$F4,$BE,$DA,$F1,$BE,$1C,$3E,$BB,$E4,$C7,$D9,$3F
	.dc.b	$E5,$03,$6E,$27,$FC,$FB,$F7,$0F,$5B,$FC,$4F,$CF,$FE,$28,$EB,$7E
	.dc.b	$5E,$F8,$9C,$EB,$FE,$0E,$D1,$8F,$18,$FF,$08,$E1,$C6,$F8,$46,$0E
	.dc.b	$13,$BA,$FE,$F0,$DA,$FF,$1B,$FC,$EB,$D8,$FF,$1D,$5F,$0C,$FC,$38
	.dc.b	$7F,$87,$B0,$E1,$39,$3B,$42,$F9,$78,$BC,$23,$CF,$77,$CE,$79,$DD
	.dc.b	$3C,$F7,$9E,$F3,$F5,$39,$8C,$67,$DB,$16,$1C,$A3,$0D,$8F,$F0,$FF
	.dc.b	$2B,$F2,$F7,$D3,$74,$BE,$55,$FE,$35,$3F,$C7,$47,$F8,$E4,$FF,$28
	.dc.b	$1F,$E3,$DE,$7F,$F1,$BF,$8B,$23,$FD,$97,$E3,$FF,$8E,$DB,$E3,$FF
	.dc.b	$8E,$54,$3C,$6F,$F1,$CC,$61,$B1,$9E,$6F,$F1,$D4,$FF,$B5,$BC,$6F
	.dc.b	$F1,$53,$C7,$93,$F0,$13,$FE,$37,$7F,$C6,$D8,$49,$FF,$1A,$46,$B3
	.dc.b	$FF,$1C,$39,$8F,$F1,$B7,$6F,$F8,$BD,$FC,$51,$FC,$F3,$C6,$F8,$31
	.dc.b	$98,$FC,$73,$5F,$F1,$DC,$7F,$8B,$1F,$F1,$7F,$8B,$FE,$2D,$C7,$FC
	.dc.b	$F7,$5F,$E3,$B7,$FC,$FB,$E3,$7F,$8C,$EB,$67,$F8,$DC,$FF,$19,$F8
	.dc.b	$DF,$D8,$3C,$6F,$ED,$11,$FC,$2E,$11,$FD,$6E,$3F,$B5,$70,$FF,$8A
	.dc.b	$5F,$EC,$91,$C1,$78,$9F,$DD,$9C,$92,$7F,$B3,$F8,$DF,$DB,$EB,$FB
	.dc.b	$E5,$F6,$A5,$7F,$85,$9F,$F1,$95,$FF,$76,$F1,$BF,$C0,$71,$FE,$F1
	.dc.b	$AF,$EE,$75,$E3,$3A,$CE,$9F,$F2,$4C,$FA,$EF,$D4,$FF,$77,$BE,$8F
	.dc.b	$F7,$E9,$E2,$DF,$13,$FB,$CE,$3E,$78,$F0,$5F,$BE,$FD,$D7,$C2,$7F
	.dc.b	$C0,$97,$C3,$0F,$13,$FE,$33,$C7,$E7,$4F,$C3,$8D,$3D,$CF,$31,$F0
	.dc.b	$9F,$F1,$45,$7B,$5D,$C7,$63,$71,$F1,$DF,$F2,$4D,$FB,$1D,$1F,$52
	.dc.b	$BE,$74,$EE,$7F,$C6,$D3,$CA,$BB,$9F,$E4,$57,$CC,$64,$FC,$0B,$99
	.dc.b	$F8,$07,$96,$34,$7C,$DA,$FA,$B3,$B3,$D3,$3B,$9F,$26,$1F,$53,$1F
	.dc.b	$9D,$3F,$62,$72,$7C,$33,$C2,$7E,$12,$FA,$B6,$BE,$A4,$7B,$AF,$D7
	.dc.b	$39,$3D,$67,$4A,$F8,$53,$F5,$1F,$5C,$F9,$E7,$D7,$7E,$48,$FB,$57
	.dc.b	$FE,$2F,$BF,$CC,$57,$F8,$DF,$04,$62,$3F,$A3,$47,$F8,$D1,$FD,$FB
	.dc.b	$5C,$E9,$FE,$3B,$27,$47,$F5,$EC,$BF,$B5,$CF,$F6,$33,$FC,$75,$8F
	.dc.b	$F5,$E8,$FE,$B7,$E3,$7F,$6E,$7F,$AF,$47,$F7,$1E,$2F,$ED,$75,$FD
	.dc.b	$C2,$3F,$C1,$6F,$F6,$B3,$FB,$35,$71,$8F,$F0,$BE,$3F,$E0,$AB,$F0
	.dc.b	$D7,$13,$FC,$00,$F8,$2F,$AC,$EE,$BC,$87,$86,$BF,$BB,$36,$B9,$CE
	.dc.b	$E7,$FC,$39,$87,$BE,$B8,$C7,$BE,$BF,$C0,$D5,$FE,$39,$3F,$C7,$98
	.dc.b	$79,$E7,$FC,$63,$7C,$5E,$67,$B3,$0F,$2E,$3F,$E1,$AC,$7A,$CF,$AC
	.dc.b	$FF,$04,$1C,$31,$F6,$3E,$83,$C5,$1F,$23,$1F,$15,$7F,$85,$E7,$47
	.dc.b	$3F,$1B,$FC,$18,$F9,$0F,$11,$A9,$F8,$0F,$D1,$BE,$E7,$C4,$AF,$34
	.dc.b	$78,$EF,$E4,$1E,$8A,$E2,$78,$B1,$F8,$13,$CE,$76,$F1,$3D,$27,$D8
	.dc.b	$72,$EC,$C9,$F0,$8C,$1E,$BC,$BE,$73,$FE,$54,$C3,$CF,$3E,$3C,$3C
	.dc.b	$F3,$C3,$0F,$2C,$30,$FE,$5F,$E1,$7F,$85,$A9,$E3,$7F,$8B,$3C,$6F
	.dc.b	$F0,$A4,$E6,$31,$AF,$F8,$36,$31,$E3,$7F,$8A,$AF,$FC,$4F,$E2,$7F
	.dc.b	$88,$AF,$71,$8B,$FE,$D6,$FF,$89,$65,$18,$8F,$F1,$0B,$6F,$FC,$65
	.dc.b	$1F,$DC,$BC,$0F,$F0,$C7,$37,$08,$78,$9F,$E2,$4F,$13,$FC,$31,$09
	.dc.b	$D4,$E7,$0F,$F1,$0F,$89,$FE,$27,$BD,$D6,$BC,$4F,$F1,$4E,$1C,$6F
	.dc.b	$FC,$3A,$C3,$FC,$07,$C7,$FE,$22,$7F,$C4,$40,$FF,$0D,$F7,$7F,$87
	.dc.b	$7C,$0F,$F1,$04,$7F,$82,$9A,$FF,$95,$5A,$F1,$7F,$C3,$CF,$F8,$A3
	.dc.b	$C1,$FF,$10,$93,$1F,$8D,$E0,$FF,$88,$21,$E0,$FF,$86,$EB,$C5,$18
	.dc.b	$7F,$C8,$53,$FE,$26,$38,$4E,$E3,$FC,$50,$6B,$0F,$F0,$B6,$5F,$E6
	.dc.b	$9E,$FF,$F0,$A4,$FC,$33,$DC,$E3,$1F,$F0,$DF,$83,$C2,$73,$E0,$FF
	.dc.b	$88,$5F,$F1,$5C,$65,$57,$1B,$ED,$59,$EA,$FE,$BF,$5F,$D0,$7A,$FF
	.dc.b	$A2,$BF,$DA,$3B,$BF,$A5,$B7,$BF,$FA,$59,$FD,$57,$C1,$FE,$F7,$1F
	.dc.b	$D7,$AF,$FC,$03,$7F,$D7,$0F,$E9,$11,$D4,$FE,$AE,$73,$3F,$A8,$C7
	.dc.b	$07,$FA,$6C,$F4,$78,$13,$C1,$FE,$A0,$F7,$C3,$FA,$E3,$D2,$BD,$67
	.dc.b	$37,$85,$73,$8F,$F0,$17,$1F,$F6,$F7,$44,$AE,$75,$FD,$DF,$1F,$7B
	.dc.b	$FD,$BC,$C1,$E8,$8D,$E1,$E4,$8F,$3D,$73,$AF,$8D,$1F,$E1,$B3,$C6
	.dc.b	$58,$D3,$D6,$79,$47,$9B,$1F,$92,$EB,$0F,$EE,$B7,$FE,$46,$3A,$DF
	.dc.b	$F8,$72,$75,$5F,$69,$F8,$78,$79,$B1,$F8,$4F,$DC,$8C,$DF,$2C,$3E
	.dc.b	$24,$EF,$0F,$33,$5E,$CD,$8D,$CF,$1C,$3C,$F3,$96,$C6,$6B,$FC,$21
	.dc.b	$5F,$E0,$C6,$CF,$E1,$DA,$7E,$08,$F0,$CF,$17,$F1,$13,$C5,$FF,$11
	.dc.b	$9F,$E0,$FA,$FF,$07,$3F,$E4,$83,$FC,$1B,$DF,$FE,$0A,$7F,$91,$E1
	.dc.b	$FE,$0F,$7F,$C1,$46,$BB,$FF,$A4,$C7,$F4,$B8,$FE,$AF,$7F,$D4,$E7
	.dc.b	$FD,$E3,$39,$8F,$E8,$33,$FD,$34,$FE,$8D,$AF,$F6,$EA,$DD,$FF,$41
	.dc.b	$9F,$F8,$01,$7F,$D1,$27,$FC,$E2,$FF,$7A,$AE,$13,$FE,$1C,$3A,$46
	.dc.b	$DE,$35,$FD,$3A,$3C,$59,$7F,$5A,$6C,$FF,$5A,$F0,$7F,$B8,$BF,$D5
	.dc.b	$E7,$55,$E3,$C7,$FC,$4B,$3B,$F0,$7F,$AE,$E1,$E3,$8F,$EE,$D5,$DC
	.dc.b	$E1,$7F,$E2,$47,$19,$7A,$9F,$F0,$16,$1F,$1E,$BD,$47,$F6,$38,$DB
	.dc.b	$EA,$9F,$8A,$71,$3E,$8C,$76,$8F,$F0,$44,$FC,$E7,$94,$F0,$79,$1D
	.dc.b	$AF,$E0,$BF,$48,$F4,$C6,$E7,$DF,$19,$7C,$C7,$F8,$B2,$BC,$11,$F1
	.dc.b	$CF,$89,$97,$D4,$9D,$57,$9E,$7F,$0B,$87,$F8,$06,$3B,$3F,$15,$B1
	.dc.b	$EA,$8F,$A3,$1C,$67,$D9,$87,$A0,$FB,$CF,$99,$B3,$CE,$BD,$D8,$79
	.dc.b	$A7,$EB,$57,$AA,$BF,$C6,$93,$93,$E9,$BE,$5A,$E3,$7E,$18,$F1,$C7
	.dc.b	$F1,$DE,$DF,$F0,$6F,$83,$FE,$12,$EF,$FF,$06,$78,$3F,$E0,$FE,$6F
	.dc.b	$F0,$15,$FF,$82,$0F,$EA,$DE,$0F,$F8,$42,$33,$87,$2D,$BF,$C3,$18
	.dc.b	$72,$C7,$A4,$7F,$79,$AF,$F0,$65,$F1,$8F,$EF,$98,$3C,$1F,$F0,$8C
	.dc.b	$7F,$7F,$F0,$7F,$C1,$93,$FD,$94,$D7,$83,$FE,$0D,$9F,$F0,$79,$D4
	.dc.b	$E1,$E0,$7F,$83,$B1,$F4,$E1,$D2,$FF,$C3,$73,$9A,$FF,$08,$47,$8B
	.dc.b	$87,$FC,$20,$47,$FC,$57,$E0,$7F,$82,$9F,$F0,$8F,$85,$FE,$2B,$8C
	.dc.b	$13,$C1,$FF,$05,$4F,$F8,$37,$2D,$47,$F8,$9E,$3F,$C6,$BC,$8E,$EF
	.dc.b	$EF,$0D,$8F,$51,$FE,$02,$E7,$FE,$F9,$DD,$FD,$EE,$3F,$C0,$26,$84
	.dc.b	$F1,$7E,$37,$E4,$0F,$C5,$BF,$C2,$4D,$8E,$0D,$7E,$67,$E0,$3F,$EB
	.dc.b	$AC,$FE,$FC,$E5,$B5,$C8,$FF,$0E,$4E,$F7,$FB,$35,$B7,$ED,$C3,$E3
	.dc.b	$B6,$71,$C3,$D1,$D9,$FD,$49,$FE,$C5,$E0,$7F,$56,$F0,$3F,$B8,$F1
	.dc.b	$FF,$60,$9F,$EE,$55,$FD,$95,$FE,$8E,$7F,$6C,$3F,$AA,$CF,$F4,$78
	.dc.b	$E9,$3C,$4F,$F7,$53,$FD,$9A,$7F,$C1,$1E,$07,$FB,$7F,$BB,$FA,$94
	.dc.b	$7F,$4E,$6C,$F2,$7F,$B6,$4E,$E7,$FA,$B1,$E0,$3D,$27,$F5,$82,$99
	.dc.b	$AF,$14,$FF,$BA,$23,$D7,$3F,$DC,$CE,$8F,$AE,$7F,$B6,$BC,$9A,$72
	.dc.b	$7F,$C4,$57,$EF,$9E,$37,$ED,$9F,$F0,$BD,$7F,$67,$8F,$54,$7C,$0A
	.dc.b	$FF,$05,$3D,$6F,$FB,$BB,$4F,$9B,$3C,$FC,$1F,$F0,$AE,$1E,$7B,$E3
	.dc.b	$87,$8E,$FF,$C4,$33,$FE,$01,$78,$B5,$F9,$87,$C8,$8F,$5E,$1F,$32
	.dc.b	$B9,$B6,$7E,$2C,$7E,$0A,$31,$3F,$6A,$BA,$61,$EC,$BF,$7D,$7D,$89
	.dc.b	$F8,$66,$8B,$7F,$1C,$EF,$3F,$0C,$EE,$7F,$8F,$5D,$3F,$E0,$88,$FF
	.dc.b	$06,$BF,$E1,$03,$F8,$BD,$23,$2F,$F8,$36,$0D,$27,$07,$F8,$4F,$0F
	.dc.b	$E8,$2F,$F3,$7F,$03,$FB,$09,$FC,$D6,$BF,$A8,$9F,$D6,$8F,$E7,$27
	.dc.b	$F6,$09,$FE,$A6,$DC,$7F,$A0,$CF,$18,$FE,$BD,$1F,$D0,$67,$67,$F4
	.dc.b	$DA,$FE,$E6,$DA,$FE,$AF,$5F,$D8,$4E,$27,$F8,$4A,$77,$5F,$D4,$EB
	.dc.b	$9B,$4F,$0C,$F8,$CF,$F7,$71,$EA,$8F,$EA,$CE,$0E,$A7,$F7,$C3,$DA
	.dc.b	$F8,$DB,$1D,$5E,$17,$FD,$91,$FE,$D8,$75,$8F,$71,$E2,$9E,$47,$9A
	.dc.b	$36,$78,$27,$FB,$85,$7B,$23,$31,$D5,$F8,$4D,$3F,$B7,$D7,$68,$F1
	.dc.b	$CF,$17,$6D,$9D,$C7,$C9,$BF,$9B,$18,$8F,$81,$7E,$57,$8C,$EA,$BF
	.dc.b	$BC,$C7,$13,$E1,$9E,$67,$B1,$A7,$9D,$FC,$13,$E4,$2F,$07,$FC,$87
	.dc.b	$87,$84,$F8,$37,$DC,$FC,$8C,$FD,$93,$87,$27,$FC,$01,$C5,$87,$4B
	.dc.b	$F8,$DF,$7C,$F8,$77,$17,$F5,$8E,$57,$D4,$BE,$EF,$3A,$7F,$79,$C3
	.dc.b	$85,$EE,$B4,$FF,$5B,$8C,$CE,$F0,$E3,$87,$F8,$23,$F1,$DF,$E9,$97
	.dc.b	$B9,$CB,$FE,$0A,$BE,$35,$FE,$18,$6C,$2D,$E1,$30,$E7,$1D,$E7,$FC
	.dc.b	$32,$7F,$72,$9F,$F0,$B4,$71,$F1,$7F,$C3,$CF,$F8,$4D,$78,$BF,$E1
	.dc.b	$E8,$FF,$7D,$79,$CF,$3D,$E5,$E7,$4F,$F1,$5D,$78,$F8,$BF,$C1,$CF
	.dc.b	$F8,$88,$FF,$09,$14,$A6,$7C,$2F,$F0,$7F,$96,$F1,$3F,$C2,$4D,$9F
	.dc.b	$F0,$BB,$5E,$1E,$27,$F8,$51,$B8,$7F,$85,$0E,$CF,$F8,$2B,$C4,$FF
	.dc.b	$19,$C7,$F9,$03,$C4,$FF,$08,$49,$BD,$9F,$E0,$F9,$FF,$22,$B7,$8B
	.dc.b	$FC,$77,$3C,$0D,$DF,$F8,$06,$FD,$9E,$1F,$F8,$45,$F8,$0D,$7F,$C1
	.dc.b	$7E,$27,$91,$47,$F8,$35,$FB,$5E,$27,$F8,$35,$FF,$11,$35,$FF,$10
	.dc.b	$CF,$7C,$BF,$CB,$33,$96,$F8,$99,$CB,$FB,$0F,$07,$F5,$BE,$CF,$F0
	.dc.b	$76,$3F,$D6,$0F,$EE,$9D,$BA,$3F,$A9,$9F,$DE,$E7,$FA,$A7,$85,$FD
	.dc.b	$21,$B3,$FD,$AC,$E5,$5F,$DA,$FC,$4F,$F0,$F4,$7F,$4F,$35,$1F,$D6
	.dc.b	$FC,$4F,$ED,$F7,$FD,$86,$7F,$AB,$5F,$43,$CD,$E2,$7F,$73,$8F,$EB
	.dc.b	$31,$FD,$FE,$3F,$AF,$C7,$46,$C7,$09,$E9,$3F,$DA,$0F,$EE,$98,$FB
	.dc.b	$5E,$EF,$B4,$8F,$F8,$40,$E9,$1F,$DA,$9E,$47,$C6,$3F,$C0,$2F,$C3
	.dc.b	$8E,$31,$FE,$06,$8C,$3C,$E7,$C8,$FC,$36,$9E,$B8,$ED,$7F,$12,$75
	.dc.b	$5F,$49,$E4,$D7,$B1,$E6,$3E,$2B,$FE,$05,$9E,$78,$FD,$2B,$ED,$5A
	.dc.b	$3E,$19,$EA,$39,$3F,$46,$BB,$63,$F8,$63,$EC,$57,$38,$ED,$5C,$CF
	.dc.b	$88,$7E,$0B,$1F,$F0,$A6,$1E,$28,$FF,$24,$61,$FE,$5F,$7F,$81,$61
	.dc.b	$FE,$0D,$9F,$DA,$B6,$3F,$C2,$91,$8E,$27,$87,$49,$3F,$C2,$6B,$FC
	.dc.b	$23,$1F,$CF,$AF,$FB,$2E,$3F,$D1,$EB,$FA,$64,$FF,$43,$7F,$DB,$A7
	.dc.b	$F5,$19,$FE,$AD,$5F,$D2,$6F,$FC,$31,$3A,$3F,$A8,$9F,$D9,$2B,$95
	.dc.b	$7F,$87,$9F,$F0,$A5,$7F,$6B,$AF,$EB,$0F,$F5,$2A,$FE,$BA,$7F,$62
	.dc.b	$2B,$DE,$78,$CF,$F6,$07,$D5,$7F,$D7,$BC,$4F,$EB,$B3,$D0,$F2,$E1
	.dc.b	$E1,$C3,$FA,$E4,$EA,$F8,$C7,$C0,$3B,$D7,$18,$EE,$7A,$AF,$FD,$CF
	.dc.b	$7F,$22,$BD,$C7,$09,$E6,$FA,$63,$E3,$D7,$38,$F0,$F8,$9F,$DA,$EF
	.dc.b	$FC,$15,$1F,$DB,$B6,$EE,$FC,$BC,$BE,$26,$1C,$B1,$EB,$1F,$0E,$79
	.dc.b	$D7,$0A,$F7,$1E,$58,$E8,$7F,$92,$5E,$4F,$5C,$7E,$2E,$1E,$C3,$E5
	.dc.b	$63,$E5,$9F,$C0,$47,$E1,$5F,$63,$70,$F1,$5F,$7C,$BE,$65,$75,$CB
	.dc.b	$C9,$7B,$C7,$E1,$5E,$F1,$ED,$7F,$88,$3C,$F1,$FE,$46,$6D,$7F,$9B
	.dc.b	$AB,$F3,$B0,$9F,$F0,$9B,$FD,$EF,$C1,$FE,$FA,$B8,$7F,$C2,$3E,$27
	.dc.b	$F8,$12,$35,$E1,$D9,$CE,$DF,$E0,$93,$FC,$25,$5B,$A5,$F4,$AD,$C7
	.dc.b	$F8,$14,$FE,$E9,$1F,$E1,$67,$FB,$C3,$5F,$F0,$95,$FF,$7E,$7F,$C1
	.dc.b	$1E,$07,$8F,$C2,$FF,$05,$C3,$C2,$FF,$05,$E1,$FE,$26,$8C,$56,$FC
	.dc.b	$A7,$95,$FE,$09,$D3,$FC,$35,$8F,$2B,$FF,$05,$1F,$D9,$F5,$FF,$05
	.dc.b	$BF,$E0,$97,$FC,$0B,$E0,$48,$FF,$02,$B5,$FF,$01,$9F,$E0,$FF,$05
	.dc.b	$E0,$7F,$81,$E3,$FC,$50,$23,$3A,$7F,$82,$38,$7F,$C0,$13,$FE,$2D
	.dc.b	$F0,$3F,$C0,$1E,$5F,$F8,$2B,$4E,$35,$AA,$FF,$07,$1F,$E3,$3B,$FF
	.dc.b	$07,$56,$E3,$93,$5D,$9F,$DF,$0F,$EF,$FD,$B9,$EE,$FE,$F9,$87,$F8
	.dc.b	$22,$7A,$35,$FC,$23,$EA,$8D,$C7,$F8,$2A,$7F,$C0,$4F,$F9,$17,$C0
	.dc.b	$FF,$07,$29,$DB,$67,$6F,$F3,$BE,$CF,$EB,$6F,$F5,$0D,$FF,$AC,$F4
	.dc.b	$7F,$42,$7F,$A4,$47,$F6,$C3,$FA,$B9,$4E,$11,$FD,$33,$B7,$FB,$E4
	.dc.b	$F0,$9F,$ED,$BD,$BF,$E0,$97,$FA,$3D,$6E,$FF,$BD,$F7,$FF,$65,$BF
	.dc.b	$EB,$67,$17,$13,$E3,$3F,$B9,$DF,$F5,$9C,$3F,$C0,$B1,$CF,$0F,$ED
	.dc.b	$A7,$96,$75,$87,$F5,$78,$FE,$B9,$18,$6D,$7F,$5C,$6C,$FC,$49,$D6
	.dc.b	$1E,$3E,$FF,$EC,$D5,$FD,$B8,$F9,$47,$F6,$EC,$3C,$B8,$7C,$26,$E1
	.dc.b	$EE,$AE,$35,$FD,$B3,$0F,$EE,$2F,$C4,$3D,$17,$C5,$F9,$E7,$CA,$7C
	.dc.b	$D8,$78,$E3,$81,$9B,$F4,$4F,$18,$FA,$A7,$48,$F1,$3C,$31,$F9,$AF
	.dc.b	$AE,$FC,$6F,$C2,$3E,$59,$D4,$F7,$1E,$48,$E2,$FA,$EB,$E9,$BF,$E9
	.dc.b	$78,$61,$FB,$2C,$3F,$90,$CA,$FF,$C0,$8F,$F8,$22,$D8,$7F,$82,$1B
	.dc.b	$1F,$E0,$83,$FC,$09,$C1,$FC,$B1,$FF,$07,$B5,$FE,$61,$DD,$FE,$0A
	.dc.b	$72,$FF,$52,$8F,$E7,$79,$FF,$48,$DF,$F9,$FD,$7F,$BC,$CF,$E9,$27
	.dc.b	$F6,$6C,$BF,$A8,$9F,$D5,$A7,$84,$7F,$55,$7F,$A3,$5F,$52,$47,$F4
	.dc.b	$B3,$FA,$F9,$B3,$A4,$70,$75,$3F,$D3,$6B,$C7,$1F,$EF,$FB,$FE,$BE
	.dc.b	$F9,$4F,$EA,$1D,$FF,$D6,$2B,$C9,$1F,$E0,$37,$CF,$7F,$D9,$63,$0F
	.dc.b	$78,$DC,$7F,$64,$3C,$64,$3F,$AC,$9F,$08,$A7,$83,$BF,$FB,$0E,$9F
	.dc.b	$DF,$AB,$A1,$DE,$7A,$1C,$23,$4F,$5B,$F4,$4E,$E7,$FB,$9B,$F3,$2F
	.dc.b	$FB,$54,$EA,$BE,$05,$72,$7A,$61,$F1,$1F,$8D,$97,$C8,$AE,$11,$83
	.dc.b	$E8,$DF,$4C,$7D,$37,$C7,$0F,$F1,$E6,$1F,$62,$BE,$2C,$79,$2F,$E4
	.dc.b	$9E,$9B,$E3,$5D,$DB,$1D,$B1,$FC,$8B,$FD,$FD,$F6,$18,$BF,$D0,$F7
	.dc.b	$7F,$75,$F1,$DE,$07,$F7,$98,$C3,$7C,$0F,$F0,$04,$EB,$6F,$F0,$06
	.dc.b	$3F,$D1,$0F,$F0,$44,$65,$FE,$FD,$CE,$9F,$EA,$D3,$98,$FE,$ED,$87
	.dc.b	$69,$FE,$9F,$5F,$DF,$9B,$D9,$FD,$FB,$C0,$FE,$FB,$1F,$DA,$9B,$3B
	.dc.b	$3C,$53,$FD,$ED,$FF,$0F,$1F,$DE,$C8,$7F,$BD,$B3,$FE,$F0,$FF,$74
	.dc.b	$EB,$78,$7F,$EF,$F8,$C3,$51,$FD,$E5,$C3,$FD,$EF,$B7,$FB,$F3,$A9
	.dc.b	$FE,$FA,$FF,$80,$E3,$FC,$58,$FF,$80,$8E,$7C,$7F,$DE,$FB,$7F,$BF
	.dc.b	$C7,$C2,$3F,$BB,$EB,$FE,$1B,$9D,$76,$FF,$7B,$A3,$FD,$F6,$3F,$C0
	.dc.b	$9D,$BF,$DC,$3B,$7F,$B7,$9F,$DC,$1F,$EE,$66,$7B,$7F,$C0,$07,$F7
	.dc.b	$27,$18,$FF,$81,$0F,$13,$FD,$EB,$A3,$F9,$9E,$DF,$CA,$1B,$CD,$FE
	.dc.b	$DF,$AF,$E6,$93,$FD,$E1,$FE,$75,$D5,$FC,$C9,$CC,$FF,$65,$9F,$ED
	.dc.b	$AF,$F5,$48,$FE,$7B,$DB,$FC,$EA,$FF,$BB,$47,$F4,$1A,$DB,$FC,$FF
	.dc.b	$B7,$FA,$CC,$EA,$7C,$98,$77,$BE,$4F,$F5,$6C,$3B,$4F,$2C,$3F,$A9
	.dc.b	$9E,$43,$FA,$A1,$E2,$C3,$AD,$71,$8F,$01,$FD,$5E,$33,$DB,$FD,$78
	.dc.b	$FE,$B7,$1E,$7B,$FF,$03,$3C,$0C,$1F,$12,$3F,$C0,$57,$E8,$6C,$FC
	.dc.b	$BC,$3D,$31,$D6,$BC,$87,$84,$87,$43,$FB,$03,$C9,$FF,$78,$57,$D3
	.dc.b	$72,$6E,$7A,$BF,$DA,$8F,$86,$E2,$38,$63,$F3,$6F,$C9,$87,$C0,$BE
	.dc.b	$13,$FE,$14,$9F,$F2,$EB,$FE,$1B,$8F,$ED,$58,$7F,$69,$23,$FD,$A3
	.dc.b	$B3,$FB,$23,$FC,$58,$FE,$CA,$64,$FE,$A7,$5F,$C6,$1F,$E8,$5D,$BF
	.dc.b	$CB,$2F,$F9,$3C,$68,$FE,$4A,$7F,$34,$AF,$E4,$9D,$1F,$D6,$A7,$79
	.dc.b	$7F,$2C,$E0,$FE,$E3,$7C,$30,$E9,$1F,$EE,$58,$E8,$7F,$4A,$8C,$E1
	.dc.b	$FE,$DB,$8F,$E8,$73,$FC,$E4,$E1,$5D,$67,$F9,$D6,$3F,$D3,$67,$BD
	.dc.b	$73,$ED,$FE,$6F,$7F,$DB,$E3,$FB,$A1,$E9,$3F,$9D,$1F,$02,$37,$5F
	.dc.b	$0E,$3F,$B1,$65,$ED,$BF,$0C,$F7,$C3,$C3,$5E,$67,$E1,$BD,$B1,$F5
	.dc.b	$BF,$EF,$13,$DF,$97,$F4,$F7,$CD,$87,$5A,$F8,$11,$FD,$6F,$2F,$41
	.dc.b	$FD,$6E,$75,$1C,$E3,$E1,$3D,$31,$F8,$37,$D4,$8E,$09,$7E,$89,$ED
	.dc.b	$8F,$BE,$BF,$B8,$63,$E8,$C3,$FC,$1B,$5E,$D3,$E3,$9F,$E3,$BC,$3E
	.dc.b	$A6,$1F,$CE,$27,$FC,$7F,$4C,$BE,$5D,$FF,$B7,$32,$FF,$6C,$9F,$D7
	.dc.b	$39,$3F,$AC,$47,$F5,$CE,$DF,$EC,$53,$A8,$FE,$BA,$FF,$5C,$8C,$C7
	.dc.b	$F6,$39,$FF,$6F,$9F,$D7,$BB,$7F,$B4,$CF,$FB,$46,$5D,$5F,$D7,$7C
	.dc.b	$1F,$F6,$B7,$81,$98,$C6,$3F,$D8,$BC,$1F,$F6,$CB,$FD,$16,$FF,$DA
	.dc.b	$E7,$F5,$B3,$FA,$E9,$FE,$D7,$3F,$DA,$A4,$3F,$A8,$76,$7F,$5E,$F0
	.dc.b	$7F,$B0,$C7,$F5,$7E,$DF,$EB,$2F,$F5,$BD,$7F,$AB,$39,$6A,$E9,$FE
	.dc.b	$97,$E1,$7F,$AE,$BB,$3F,$CC,$B3,$FE,$CD,$25,$7F,$BC,$EB,$FA,$A7
	.dc.b	$47,$FB,$4F,$F1,$37,$DC,$D2,$5F,$F6,$1F,$A4,$7C,$EE,$23,$FC,$1C
	.dc.b	$2B,$8B,$FD,$9D,$AE,$A7,$8B,$9A,$E5,$8F,$DC,$8E,$51,$FE,$E5,$9D
	.dc.b	$C7,$6E,$BF,$EB,$AF,$F2,$AE,$CF,$EB,$07,$F3,$C7,$F9,$3F,$6F,$F2
	.dc.b	$3E,$BF,$E4,$FD,$FF,$D1,$E7,$27,$F3,$68,$FE,$7C,$FF,$31,$9E,$15
	.dc.b	$D2,$3F,$A2,$19,$38,$3F,$CC,$1A,$78,$CF,$EA,$F3,$FD,$6E,$FF,$99
	.dc.b	$47,$F4,$F8,$EE,$7F,$AD,$5F,$E8,$D1,$FC,$ED,$E4,$7F,$61,$27,$7F
	.dc.b	$F4,$1A,$F2,$47,$F4,$8A,$E4,$74,$8F,$ED,$67,$C3,$AF,$0C,$FC,$03
	.dc.b	$BE,$1F,$0E,$32,$47,$FD,$88,$DD,$BF,$B4,$DF,$5A,$F5,$5F,$C1,$79
	.dc.b	$CF,$F6,$39,$F8,$17,$C8,$91,$CA,$B8,$4F,$3A,$F2,$36,$FC,$11,$FD
	.dc.b	$B0,$EA,$7F,$5C,$BE,$6E,$67,$84,$61,$F0,$DF,$79,$F9,$78,$7C,$97
	.dc.b	$FD,$A1,$8F,$FB,$3F,$1F,$EC,$B8,$7E,$9A,$6D,$FF,$64,$7F,$B3,$9F
	.dc.b	$D8,$4F,$EC,$64,$3F,$B3,$1F,$C8,$FB,$FF,$8E,$BF,$CB,$EB,$FA,$03
	.dc.b	$FC,$BF,$BF,$F9,$56,$7F,$D8,$27,$39,$FF,$34,$23,$82,$57,$F2,$B8
	.dc.b	$FE,$7B,$1F,$E8,$73,$C1,$5F,$D2,$A3,$FA,$04,$FF,$4B,$EE,$FE,$56
	.dc.b	$FF,$48,$31,$D1,$FD,$12,$7F,$9B,$C7,$F3,$C3,$FB,$84,$F0,$C3,$FA
	.dc.b	$1D,$FA,$31,$FE,$B7,$5E,$3C,$7C,$2F,$F3,$EA,$FE,$87,$7C,$6B,$FA
	.dc.b	$4C,$FC,$12,$61,$EE,$79,$1E,$0C,$7C,$98,$7F,$B2,$23,$27,$A6,$7D
	.dc.b	$A4,$8F,$85,$1C,$4F,$69,$23,$4F,$F7,$79,$EA,$FF,$AD,$DA,$FF,$7D
	.dc.b	$8F,$9A,$EC,$E5,$3D,$9F,$13,$70,$F8,$B1,$EC,$8F,$AE,$FC,$B3,$67
	.dc.b	$FA,$5A,$16,$C7,$FB,$15,$7F,$A6,$A9,$F6,$CF,$C2,$BF,$AF,$F1,$7F
	.dc.b	$25,$8C,$67,$FE,$B2,$F3,$5F,$09,$A8,$E5,$3A,$C3,$F9,$AB,$6F,$96
	.dc.b	$1B,$9F,$EB,$F8,$72,$BF,$F4,$F5,$ED,$E7,$5B,$AF,$2B,$C7,$B9,$7C
	.dc.b	$67,$BC,$FF,$70,$C3,$FB,$75,$F9,$FC,$4C,$3E,$FF,$25,$E2,$63,$0F
	.dc.b	$1D,$57,$E2,$70,$FF,$5E,$7F,$B6,$9F,$DA,$CF,$ED,$C7,$F6,$AF,$23
	.dc.b	$FD,$83,$E0,$FF,$6B,$8F,$ED,$C7,$FB,$23,$C1,$FE,$CF,$1F,$EC,$28
	.dc.b	$FE,$E3,$E0,$FF,$6E,$8C,$47,$D3,$EF,$FF,$58,$BF,$D7,$BC,$0F,$F5
	.dc.b	$C7,$83,$FD,$B7,$3F,$F5,$BC,$FF,$AD,$23,$FB,$23,$FE,$CB,$F0,$3F
	.dc.b	$B0,$B7,$C0,$F5,$46,$23,$F0,$0F,$F6,$6E,$1F,$EC,$5E,$03,$E8,$BF
	.dc.b	$B1,$1F,$D9,$B8,$F7,$86,$30,$FE,$E9,$39,$7F,$8F,$F8,$1F,$D6,$9F
	.dc.b	$EA,$7D,$5F,$D0,$FC,$0F,$E4,$CD,$F0,$3F,$93,$4F,$FA,$C3,$BF,$FA
	.dc.b	$93,$6B,$FA,$31,$C1,$FE,$C7,$CD,$FD,$42,$FF,$9D,$92,$3C,$27,$F6
	.dc.b	$43,$95,$FF,$45,$8F,$04,$7F,$39,$78,$3A,$3F,$9F,$47,$92,$FF,$BF
	.dc.b	$C6,$6B,$FB,$DC,$F3,$8F,$1D,$7F,$60,$69,$FD,$58,$F1,$9F,$D2,$A7
	.dc.b	$8B,$D8,$F1,$1A,$7F,$A5,$CF,$86,$FF,$DB,$18,$7C,$BA,$FE,$E6,$DF
	.dc.b	$07,$FA,$64,$60,$F8,$18,$FA,$09,$5C,$A3,$2D,$7C,$98,$FF,$7C,$9F
	.dc.b	$A2,$70,$C3,$F8,$2C,$FA,$C4,$2B,$FA,$C4,$7F,$0D,$6B,$A6,$C6,$E7
	.dc.b	$FD,$7B,$3F,$C4,$FC,$0F,$E2,$4F,$F5,$D3,$FA,$D9,$FC,$AF,$2F,$E5
	.dc.b	$2F,$FA,$70,$FE,$46,$7F,$25,$73,$5F,$D0,$35,$FE,$57,$E3,$FF,$A0
	.dc.b	$F2,$FE,$75,$7F,$C8,$EB,$F9,$3D,$FF,$41,$8C,$61,$FD,$1A,$31,$3F
	.dc.b	$D9,$67,$FA,$0B,$C2,$BF,$96,$B9,$8E,$3E,$0F,$F5,$53,$FD,$47,$19
	.dc.b	$25,$FF,$58,$9D,$9F,$CD,$CE,$AF,$86,$BA,$BB,$9E,$35,$D2,$30,$47
	.dc.b	$04,$AF,$F0,$01,$D1,$F2,$38,$3C,$37,$EE,$9E,$51,$9A,$EE,$D8,$CC
	.dc.b	$FF,$76,$9C,$B6,$33,$7F,$3A,$72,$D8,$CD,$7F,$5F,$3E,$C9,$F6,$EB
	.dc.b	$EC,$C7,$F8,$61,$E9,$7E,$B6,$CF,$8E,$31,$5F,$D5,$F0,$FF,$5C,$63
	.dc.b	$F9,$6A,$FF,$5F,$55,$F0,$3F,$A5,$47,$F5,$FF,$C4,$58,$7F,$4B,$9C
	.dc.b	$D2,$73,$7F,$D4,$BB,$FF,$AB,$4E,$6F,$84,$62,$B7,$87,$F5,$0B,$FF
	.dc.b	$51,$E1,$FD,$86,$3F,$B6,$1F,$D6,$23,$35,$FE,$8F,$8F,$EA,$5D,$DF
	.dc.b	$EC,$29,$FF,$67,$1F,$EC,$92,$C7,$FB,$3D,$BC,$BF,$EC,$A7,$FA,$94
	.dc.b	$43,$BE,$9F,$D5,$D3,$FE,$BC,$5D,$7F,$D5,$5F,$EB,$1D,$B9,$8F,$EA
	.dc.b	$31,$FE,$BD,$3F,$D7,$EE,$1F,$F6,$11,$FE,$C2,$EC,$FE,$AB,$87,$FA
	.dc.b	$A7,$B3,$FA,$AC,$FF,$53,$07,$F5,$9E,$CF,$EA,$EF,$FB,$0B,$2F,$E9
	.dc.b	$6F,$F4,$A2,$1F,$EB,$CE,$BF,$EA,$91,$A8,$FF,$5D,$74,$FF,$51,$75
	.dc.b	$D5,$FD,$45,$75,$FF,$59,$79,$E5,$FE,$BD,$C3,$9C,$7F,$2F,$E3,$FE
	.dc.b	$83,$CF,$FC,$7E,$3F,$D5,$C7,$F2,$43,$FA,$C9,$FC,$F0,$FF,$43,$C7
	.dc.b	$0E,$9F,$E5,$5C,$3F,$CE,$27,$F9,$D9,$FD,$1E,$7F,$96,$4F,$F5,$AE
	.dc.b	$9F,$E7,$87,$E5,$EF,$FD,$1F,$7F,$E8,$8B,$E3,$1E,$38,$FE,$BD,$1E
	.dc.b	$72,$1D,$4F,$29,$FE,$B3,$BF,$EA,$4D,$3F,$9C,$4F,$38,$CB,$FC,$FA
	.dc.b	$FD,$31,$EB,$BF,$24,$7F,$6F,$AE,$47,$C0,$3C,$35,$CD,$EB,$1C,$9C
	.dc.b	$B5,$F3,$1F,$D5,$4F,$98,$5B,$F9,$07,$C1,$25,$FF,$76,$6C,$7D,$1A
	.dc.b	$F8,$F3,$92,$46,$6B,$D1,$3E,$DB,$F5,$9F,$42,$7E,$ED,$FF,$BC,$1F
	.dc.b	$F5,$69,$0F,$F5,$89,$FD,$4C,$FE,$AD,$CF,$FD,$61,$FE,$1A,$7E,$F0
	.dc.b	$FF,$36,$9F,$C6,$F0,$FF,$3A,$74,$FF,$3B,$8F,$E6,$47,$F3,$F3,$F9
	.dc.b	$8C,$7F,$3E,$9F,$E3,$71,$FC,$96,$BF,$91,$C7,$FA,$AE,$79,$BF,$CE
	.dc.b	$A3,$F9,$19,$23,$F9,$74,$72,$9F,$F4,$54,$7F,$A9,$27,$FD,$17,$1E
	.dc.b	$53,$FB,$44,$74,$7C,$98,$7F,$38,$7C,$51,$99,$FE,$8D,$5D,$7A,$7F
	.dc.b	$B6,$56,$8E,$64,$3B,$12,$FF,$BA,$3F,$D0,$1B,$5E,$E8,$FF,$5C,$61
	.dc.b	$FD,$07,$0F,$5C,$7F,$3F,$24,$FB,$DF,$01,$D6,$78,$38,$AE,$B5,$E9
	.dc.b	$3E,$74,$72,$7A,$D7,$38,$F8,$E7,$F8,$3E,$38,$3F,$62,$3E,$0B,$C4
	.dc.b	$FE,$37,$7F,$D4,$1F,$E5,$B3,$FE,$B5,$E8,$FE,$97,$D3,$FD,$13,$6F
	.dc.b	$F4,$FF,$17,$F2,$24,$FF,$24,$D3,$FA,$63,$FD,$3C,$D1,$99,$E4,$FF
	.dc.b	$A8,$8C,$E1,$FD,$6F,$0F,$EA,$71,$8A,$E7,$3F,$D4,$C5,$F8,$1E,$13
	.dc.b	$FD,$2B,$CC,$79,$3F,$EA,$1A,$FE,$99,$5F,$CE,$3C,$1F,$F5,$2D,$F8
	.dc.b	$27,$32,$9C,$BD,$A7,$FA,$C3,$4F,$EA,$67,$FA,$A7,$5F,$EA,$84,$7F
	.dc.b	$D4,$A7,$F4,$6E,$9F,$EA,$8F,$FA,$9C,$C9,$FD,$40,$FF,$53,$DF,$FA
	.dc.b	$AB,$C0,$F6,$C7,$F5,$66,$BF,$DC,$3E,$85,$F7,$9F,$ED,$DA,$F8,$5D
	.dc.b	$AE,$C7,$E1,$59,$D3,$FA,$BC,$E6,$31,$8F,$F5,$AC,$79,$BF,$48,$38
	.dc.b	$7E,$04,$F1,$7E,$BB,$5F,$88,$70,$C7,$AB,$CB,$EA,$3F,$D9,$D4,$69
	.dc.b	$23,$F8,$81,$B3,$AE,$3F,$D8,$1B,$D7,$D2,$B9,$47,$9D,$AF,$9A,$78
	.dc.b	$36,$78,$57,$93,$9F,$F9,$F4,$FF,$41,$3F,$9D,$F8,$5F,$E9,$02,$78
	.dc.b	$7F,$E9,$32,$47,$FA,$3B,$AF,$F9,$8B,$FD,$73,$B3,$FA,$25,$F4,$3F
	.dc.b	$9D,$C7,$F3,$99,$FE,$6D,$C7,$FD,$82,$3B,$DF,$47,$FA,$4E,$1E,$07
	.dc.b	$B1,$FC,$F3,$A3,$FA,$D3,$4F,$EA,$77,$E2,$BF,$E9,$71,$FD,$73,$0F
	.dc.b	$0B,$71,$FE,$92,$7F,$77,$7D,$15,$C6,$FF,$A7,$C6,$E3,$CF,$1F,$DB
	.dc.b	$27,$85,$F6,$BF,$EA,$57,$E3,$3D,$67,$A2,$3D,$D1,$F0,$E7,$C3,$7F
	.dc.b	$EC,$DA,$EC,$78,$8E,$53,$B8,$F0,$C7,$F7,$27,$CC,$FC,$B3,$81,$FD
	.dc.b	$74,$FF,$5B,$3F,$8A,$8F,$F5,$A4,$7F,$5F,$3F,$AE,$9F,$E5,$85,$D9
	.dc.b	$FE,$B6,$8C,$35,$FE,$9D,$84,$7F,$95,$1F,$D0,$E7,$F9,$9B,$67,$F9
	.dc.b	$BB,$FC,$C0,$FE,$D3,$1A,$7F,$9E,$74,$FF,$61,$78,$35,$C9,$FD,$21
	.dc.b	$C4,$71,$24,$F6,$EC,$FE,$AF,$5D,$E7,$FD,$74,$FF,$3D,$BF,$E8,$B9
	.dc.b	$7F,$43,$AD,$93,$C9,$FE,$51,$1F,$D5,$CF,$ED,$77,$DB,$1F,$0E,$3D
	.dc.b	$EB,$78,$7F,$4A,$C7,$FB,$F9,$DC,$86,$09,$7F,$DC,$5F,$35,$7F,$82
	.dc.b	$5C,$BF,$EA,$A7,$37,$FE,$03,$AF,$EB,$64,$8F,$EC,$53,$E3,$E5,$FF
	.dc.b	$71,$E1,$E4,$9F,$96,$7C,$92,$63,$EC,$AE,$11,$FD,$DD,$FF,$7A,$C7
	.dc.b	$FA,$A2,$3F,$CF,$7F,$04,$FB,$4F,$E9,$9D,$1F,$CB,$3B,$7F,$8C,$65
	.dc.b	$FC,$96,$B4,$FF,$22,$8C,$DF,$F2,$29,$CB,$FC,$A2,$33,$5F,$D5,$E1
	.dc.b	$3A,$8E,$DC,$7E,$0B,$FF,$46,$D6,$B0,$FF,$56,$3E,$66,$B9,$8F,$E5
	.dc.b	$37,$E0,$8F,$F5,$AB,$FD,$8A,$3C,$DF,$4F,$D6,$7D,$FD,$D7,$A9,$9A
	.dc.b	$B5,$5E,$6A,$D4,$E6,$FE,$14,$62,$B7,$5F,$EA,$4D,$3F,$B1,$3F,$CF
	.dc.b	$1A,$FF,$61,$3F,$D6,$A7,$F6,$2E,$DF,$EC,$AF,$FA,$D4,$FE,$C6,$7F
	.dc.b	$61,$D7,$FB,$21,$FD,$1B,$AB,$F9,$FB,$FD,$83,$AB,$F9,$FB,$7A,$BF
	.dc.b	$9F,$D7,$F4,$5E,$0F,$EB,$FB,$7F,$65,$9B,$7C,$35,$F9,$F8,$F4,$6C
	.dc.b	$FF,$58,$26,$1E,$16,$BF,$D3,$FE,$07,$99,$FD,$56,$24,$7F,$5A,$8C
	.dc.b	$BF,$D5,$E3,$FA,$97,$4F,$F4,$82,$74,$FF,$1E,$9F,$E4,$BB,$7F,$2F
	.dc.b	$9F,$E6,$3C,$DF,$CD,$BA,$3F,$98,$CF,$F2,$E9,$E3,$5F,$CA,$9B,$19
	.dc.b	$BE,$95,$C2,$FF,$9E,$47,$F3,$38,$C5,$FF,$2F,$39,$47,$0B,$F0,$C7
	.dc.b	$F3,$1B,$F2,$4F,$19,$CC,$FA,$2F,$CD,$18,$BE,$B3,$FD,$BD,$FE,$7C
	.dc.b	$FF,$41,$BF,$33,$E0,$23,$82,$57,$A2,$3F,$CA,$A7,$FB,$1A,$7A,$9F
	.dc.b	$D8,$0B,$5F,$D4,$23,$67,$FA,$D6,$7F,$B5,$D7,$48,$F8,$0D,$7C,$D3
	.dc.b	$F2,$A7,$C1,$3F,$C1,$E7,$91,$FD,$C8,$F3,$C6,$8F,$F7,$2D,$FE,$8D
	.dc.b	$FD,$DF,$53,$AF,$F7,$B7,$FE,$B7,$BF,$F5,$A9,$FD,$6C,$63,$FD,$69
	.dc.b	$FE,$AB,$D5,$FE,$B7,$8F,$EB,$2D,$E9,$FE,$59,$1F,$CB,$FA,$FF,$9F
	.dc.b	$6F,$FC,$98,$97,$FC,$86,$FF,$9E,$91,$FE,$4F,$9F,$F5,$F7,$FA,$CE
	.dc.b	$1F,$CA,$31,$FE,$8D,$39,$C7,$FA,$7D,$7F,$3A,$BF,$F5,$57,$5F,$F3
	.dc.b	$38,$FE,$B7,$7F,$CC,$67,$B3,$63,$B9,$FD,$B0,$DD,$F5,$C3,$D6,$7F
	.dc.b	$45,$CB,$D7,$3E,$93,$89,$E6,$BE,$B5,$C5,$FE,$95,$1F,$D2,$1C,$BD
	.dc.b	$F1,$FE,$BC,$F2,$7D,$67,$C9,$C7,$D4,$EB,$1F,$41,$2F,$B5,$7F,$B3
	.dc.b	$B1,$F3,$5F,$07,$E4,$4F,$F6,$7B,$E9,$97,$D2,$C3,$F1,$51,$F6,$B0
	.dc.b	$FC,$AB,$63,$37,$FD,$5A,$FF,$A1,$E5,$FE,$B5,$9C,$E5,$F8,$DC,$3F
	.dc.b	$AC,$FD,$F7,$F3,$6F,$B4,$FE,$73,$D8,$8C,$69,$FD,$65,$B5,$FD,$5D
	.dc.b	$BD,$1F,$D5,$6B,$85,$7F,$4C,$9F,$E9,$F1,$8A,$FE,$AA,$78,$27,$FA
	.dc.b	$8F,$6F,$F4,$6E,$9F,$0B,$63,$FB,$43,$67,$FB,$33,$70,$FE,$BB,$9F
	.dc.b	$F4,$69,$FF,$56,$9F,$D2,$4F,$E8,$DE,$63,$AB,$FD,$0F,$D7,$FD,$22
	.dc.b	$4D,$13,$B8,$7D,$3B,$CD,$FF,$59,$36,$3F,$D6,$D3,$B8,$C4,$74,$8C
	.dc.b	$4F,$ED,$A2,$1F,$E5,$03,$E1,$9F,$B8,$E9,$74,$FF,$94,$DB,$1A,$E6
	.dc.b	$FD,$E3,$C8,$E1,$3B,$23,$B9,$E2,$FF,$05,$9E,$2F,$84,$C4,$69,$E9
	.dc.b	$1A,$7A,$9C,$E7,$AB,$C1,$AF,$09,$E4,$F5,$31,$1A,$70,$43,$A9,$CC
	.dc.b	$EC,$43,$04,$39,$99,$29,$93,$A4,$FA,$49,$3E,$48,$EE,$70,$21,$C1
	.dc.b	$D4,$70,$9F,$44,$F0,$34,$E2,$35,$1E,$33,$25,$38,$11,$DB,$E7,$31
	.dc.b	$3E,$02,$4F,$56,$C7,$39,$E2,$FA,$0E,$33,$B2,$3B,$9E,$2F,$79,$E2
	.dc.b	$F9,$8C,$46,$9E,$91,$A7,$A9,$CE,$7A,$BC,$1A,$F0,$9E,$4F,$53,$11
	.dc.b	$A7,$04,$3A,$9C,$CE,$C4,$30,$43,$99,$92,$99,$3A,$4F,$A4,$93,$E4
	.dc.b	$8E,$E7,$02,$1C,$1D,$47,$09,$F4,$4F,$03,$4E,$23,$51,$E3,$32,$53
	.dc.b	$81,$1D,$BE,$73,$13,$E0,$24,$F5,$6C,$73,$9E,$2F,$A0,$E3,$3B,$23
	.dc.b	$B9,$E2,$F7,$9E,$2F,$98,$C4,$69,$E9,$1A,$7A,$9C,$E7,$AB,$C1,$AF
	.dc.b	$09,$E4,$F5,$31,$1A,$70,$43,$A9,$CC,$EC,$43,$04,$39,$99,$3F,$2F
	.dc.b	$CD,$FE,$F1,$47,$FB,$E0,$FF,$51,$D7,$F6,$77,$FC,$46,$7F,$BF,$5C
	.dc.b	$3A,$8E,$4F,$13,$9C,$69,$B1,$A8,$C4,$6C,$8E,$09,$1B,$27,$27,$F8
	.dc.b	$C4,$93,$E3,$8C,$B6,$32,$FA,$08,$79,$DC,$B4,$E0,$43,$67,$A1,$DB
	.dc.b	$4D,$1C,$4D,$9E,$A8,$F1,$B8,$9F,$5C,$F2,$8D,$36,$33,$1D,$0E,$C7
	.dc.b	$19,$E6,$F7,$21,$A2,$9D,$0E,$45,$39,$4F,$78,$E7,$1C,$48,$6C,$E7
	.dc.b	$3E,$47,$89,$D1,$C3,$A8,$E6,$F1,$39,$C6,$9B,$1A,$8C,$46,$C8,$E0
	.dc.b	$91,$B2,$4F,$A4,$93,$E3,$8C,$B6,$32,$FA,$08,$79,$DC,$B4,$E0,$43
	.dc.b	$67,$A1,$DB,$4D,$1C,$4D,$9E,$A8,$F1,$B8,$9F,$5C,$F2,$8D,$36,$33
	.dc.b	$1D,$0E,$C7,$19,$E6,$F7,$21,$A2,$9D,$0E,$45,$39,$4F,$78,$E7,$1C
	.dc.b	$48,$6C,$E7,$3E,$47,$89,$D1,$C3,$A8,$E6,$F1,$39,$C6,$9B,$1A,$8C
	.dc.b	$46,$C8,$E0,$91,$B2,$4F,$A4,$93,$E3,$8C,$B6,$32,$FA,$08,$79,$DC
	.dc.b	$B4,$E0,$43,$F7,$79,$FF,$97,$34,$FE,$4B,$A7,$F9,$5A,$31,$3F,$E5
	.dc.b	$EC,$FF,$CB,$A4,$7F,$CB,$B5,$FE,$68,$D3,$FC,$C9,$1F,$E7,$AA,$FF
	.dc.b	$34,$46,$23,$81,$FE,$61,$9E,$F5,$8A,$EF,$7F,$E6,$6C,$BE,$2E,$3E
	.dc.b	$AC,$3E,$6B,$5F,$97,$97,$C4,$7E,$09,$A3,$DC,$7C,$13,$27,$08,$E5
	.dc.b	$3D,$A3,$67,$13,$6E,$5B,$3D,$DE,$84,$38,$14,$EC,$74,$29,$CA,$7C
	.dc.b	$31,$CE,$38,$90,$D9,$CE,$7C,$4D,$8E,$0E,$8E,$11,$A6,$C6,$A3,$11
	.dc.b	$B2,$38,$24,$6C,$93,$E6,$24,$FA,$E7,$71,$88,$F6,$19,$9E,$2D,$8D
	.dc.b	$3E,$96,$BE,$66,$BE,$48,$E2,$F4,$21,$C4,$E8,$43,$27,$33,$06,$CC
	.dc.b	$9D,$63,$BC,$FA,$A3,$67,$13,$6E,$5B,$3E,$47,$EA,$69,$FE,$1C,$7F
	.dc.b	$C2,$E7,$F8,$48,$FF,$0A,$1F,$E1,$B9,$FF,$14,$12,$38,$37,$4F,$F1
	.dc.b	$14,$F2,$38,$4E,$89,$1A,$9E,$13,$FE,$1F,$37,$19,$9D,$C6,$4E,$84
	.dc.b	$9E,$2F,$98,$C4,$70,$71,$1A,$7D,$84,$3B,$9D,$5D,$38,$38,$C7,$37
	.dc.b	$81,$DA,$7C,$C4,$79,$46,$DE,$47,$09,$EA,$F1,$37,$3C,$5E,$ED,$78
	.dc.b	$4F,$37,$81,$A3,$89,$0C,$10,$E3,$1C,$DF,$19,$0E,$66,$8E,$45,$39
	.dc.b	$1B,$9F,$49,$27,$C8,$D9,$ED,$3D,$0E,$13,$A2,$46,$A7,$84,$F8,$0D
	.dc.b	$C6,$67,$71,$93,$A1,$FE,$90,$C7,$F9,$7E,$5F,$CC,$74,$69,$FC,$D2
	.dc.b	$BF,$9E,$9F,$D0,$4A,$7F,$3D,$8F,$E7,$8E,$0F,$F3,$69,$FC,$F4,$A4
	.dc.b	$D6,$B4,$A7,$F9,$D1,$AD,$37,$C4,$E0,$70,$FF,$9E,$B8,$6D,$7A,$3C
	.dc.b	$07,$1E,$27,$32,$C3,$95,$F1,$3F,$D0,$79,$B0,$E3,$19,$AF,$F4,$3D
	.dc.b	$70,$57,$E0,$BE,$18,$77,$BF,$83,$7B,$AD,$61,$D3,$0E,$35,$AE,$4F
	.dc.b	$2C,$58,$F8,$76,$9C,$CF,$89,$AF,$AD,$F0,$3F,$07,$96,$B9,$AF,$56
	.dc.b	$5E,$D9,$CD,$F5,$9C,$E0,$9C,$E1,$D2,$F8,$57,$C3,$C1,$C4,$E3,$72
	.dc.b	$7D,$9E,$7F,$F4,$77,$01,$FC,$1F,$3B,$A3,$EB,$CE,$71,$FF,$47,$35
	.dc.b	$EB,$87,$45,$19,$C7,$C1,$8F,$79,$FF,$5D,$E1,$D7,$2C,$47,$29,$B5
	.dc.b	$C2,$39,$E5,$E6,$3C,$B9,$79,$1C,$47,$91,$AB,$0E,$CF,$23,$44,$3C
	.dc.b	$06,$4E,$13,$D5,$D5,$78,$F1,$F6,$65,$EF,$7E,$15,$F6,$73,$8F,$AA
	.dc.b	$75,$8F,$92,$B4,$F9,$AF,$A3,$9C,$7C,$18,$F9,$31,$E6,$F9,$6B,$6F
	.dc.b	$CA,$9D,$C6,$23,$85,$73,$CB,$C5,$94,$9E,$59,$79,$1E,$53,$9C,$7B
	.dc.b	$92,$FE,$7C,$65,$F4,$57,$1A,$F8,$F3,$97,$E3,$C6,$72,$F7,$C6,$30
	.dc.b	$D3,$C0,$F1,$B8,$C7,$D5,$3C,$27,$D6,$71,$3E,$B3,$98,$F8,$04,$AF
	.dc.b	$06,$5F,$11,$F3,$46,$6F,$F1,$F5,$C6,$72,$48,$CD,$71,$7C,$91,$D3
	.dc.b	$0B,$1E,$AA,$E4,$E3,$0F,$2D,$F1,$7D,$A4,$CB,$F0,$F1,$97,$E7,$47
	.dc.b	$E4,$EB,$9B,$BC,$BE,$D3,$4F,$85,$3C,$B0,$F0,$3C,$A3,$E6,$37,$2F
	.dc.b	$81,$1E,$49,$DD,$74,$AD,$65,$F0,$6B,$47,$73,$E8,$93,$0F,$2C,$F0
	.dc.b	$CB,$E3,$E1,$D2,$BD,$59,$7B,$71,$FC,$3D,$6A,$7A,$65,$E4,$C3,$D5
	.dc.b	$18,$CB,$C5,$18,$3B,$47,$8F,$2F,$0E,$1D,$23,$CF,$3A,$3E,$14,$EA
	.dc.b	$3E,$3C,$FE,$69,$ED,$87,$D9,$C7,$CD,$1A,$3F,$46,$68,$F0,$3A,$7E
	.dc.b	$1E,$3F,$96,$8C,$3E,$BA,$ED,$87,$8D,$F9,$C7,$9D,$F2,$CF,$0C,$7D
	.dc.b	$F3,$C0,$D1,$C2,$FC,$87,$D1,$9E,$18,$7A,$72,$FA,$13,$96,$E3,$E5
	.dc.b	$6E,$1F,$99,$9F,$66,$3E,$29,$CD,$FD,$58,$F4,$4F,$DA,$9C,$BF,$3B
	.dc.b	$1F,$2B,$93,$9B,$E7,$7E,$2D,$FB,$1F,$97,$5D,$1F,$31,$C0,$E4,$E7
	.dc.b	$0F,$44,$64,$DC,$66,$7D,$4D,$C3,$D1,$7E,$4A,$DB,$6F,$9D,$6E,$7D
	.dc.b	$47,$CC,$8C,$CE,$EF,$AB,$6F,$AD,$6E,$72,$D8,$CB,$F0,$1F,$A8,$FA
	.dc.b	$6B,$AB,$B3,$C2,$6C,$FF,$16,$BF,$D7,$DF,$F1,$8B,$5F,$F1,$AB,$5F
	.dc.b	$F2,$39,$FD,$0F,$99,$CF,$FD,$A1,$B1,$FD,$6F,$A9,$D5,$FE,$3E,$CF
	.dc.b	$FB,$A5,$69,$B7,$C4,$FE,$E7,$5C,$2F,$07,$78,$B3,$FE,$01,$9D,$4E
	.dc.b	$49,$1D,$1F,$15,$F8,$73,$F5,$E1,$E1,$6E,$1D,$1F,$F2,$9E,$1D,$30
	.dc.b	$E4,$48,$DE,$1D,$0E,$00,$FF,$0F,$DF,$57,$18,$A3,$10,$BE,$D5,$D2
	.dc.b	$FB,$4F,$0A,$DD,$6A,$31,$87,$81,$EE,$FC,$8C,$7D,$D1,$A3,$E1,$E1
	.dc.b	$D4,$F6,$C6,$A3,$89,$A3,$86,$1F,$36,$35,$8F,$A6,$FC,$0D,$AF,$84
	.dc.b	$71,$39,$D7,$A9,$F7,$57,$2C,$7E,$0D,$71,$AD,$63,$ED,$BE,$38,$F9
	.dc.b	$DF,$1B,$E8,$7C,$37,$C8,$F8,$B8,$75,$C7,$27,$C3,$9A,$79,$6F,$B6
	.dc.b	$3E,$1C,$3B,$4F,$80,$FB,$53,$AC,$3B,$5F,$CF,$8C,$3D,$31,$F8,$D5
	.dc.b	$6F,$9D,$6A,$3E,$26,$1D,$DE,$B8,$F9,$DC,$E3,$EA,$79,$1F,$00,$D9
	.dc.b	$83,$CC,$7C,$C3,$13,$D5,$C1,$D3,$1F,$45,$77,$C7,$EA,$56,$F0,$F3
	.dc.b	$46,$30,$F1,$61,$C4,$F9,$17,$C7,$1F,$8F,$87,$18,$C4,$F9,$CF,$2E
	.dc.b	$1D,$2B,$77,$D3,$0E,$D8,$F9,$2F,$D3,$39,$AD,$CE,$67,$C5,$5F,$86
	.dc.b	$C3,$BC,$7C,$2C,$7C,$4F,$2A,$FC,$03,$C2,$FE,$9D,$70,$AF,$8C,$70
	.dc.b	$9D,$1E,$9C,$7D,$0F,$AE,$35,$87,$A2,$3E,$19,$DC,$FD,$0B,$93,$CC
	.dc.b	$E5,$FC,$2B,$D8,$F8,$B5,$E3,$C7,$F1,$B1,$CB,$0F,$82,$F6,$38,$5F
	.dc.b	$8A,$33,$7D,$A7,$35,$F2,$5F,$43,$D2,$3C,$27,$96,$7E,$1C,$F9,$5F
	.dc.b	$D6,$3C,$B1,$F9,$33,$B8,$F8,$98,$79,$E7,$EA,$4E,$67,$D6,$73,$3F
	.dc.b	$2E,$73,$3D,$87,$92,$38,$D7,$A8,$F1,$C6,$E7,$CF,$19,$8F,$09,$D4
	.dc.b	$FC,$EC,$78,$70,$F8,$91,$D4,$F9,$11,$9C,$3E,$24,$7E,$1D,$F8,$95
	.dc.b	$F0,$27,$57,$F0,$23,$C1,$5E,$C3,$C8,$EC,$CD,$F8,$23,$A4,$FB,$1F
	.dc.b	$15,$7D,$53,$AB,$F0,$87,$6B,$B3,$FA,$44,$7F,$BF,$87,$0B,$B9,$DD
	.dc.b	$AE,$87,$17,$F4,$FA,$D4,$E6,$39,$E3,$D7,$0F,$F7,$93,$5E,$D5,$C1
	.dc.b	$FF,$7F,$E1,$D9,$DE,$3C,$DF,$EB,$95,$C5,$F3,$9E,$A8,$AF,$81,$AF
	.dc.b	$06,$BF,$03,$A7,$A5,$FB,$66,$DF,$97,$2F,$6D,$F8,$2F,$C4,$F1,$3C
	.dc.b	$F5,$C0,$CE,$09,$DF,$6B,$C3,$61,$E4,$8C,$36,$33,$87,$AA,$72,$DE
	.dc.b	$DF,$A6,$DC,$F9,$46,$23,$E1,$CA,$31,$1F,$13,$6F,$82,$9A,$47,$E4
	.dc.b	$F6,$3C,$1F,$83,$97,$C2,$C5,$DA,$F0,$73,$C3,$F8,$8C,$38,$3F,$E1
	.dc.b	$1B,$E1,$5B,$8C,$1B,$C3,$8E,$3F,$8E,$9D,$3F,$91,$17,$D1,$FC,$3C
	.dc.b	$6A,$FE,$AC,$6B,$BE,$BE,$36,$BF,$E2,$3A,$46,$23,$86,$1F,$86,$9C
	.dc.b	$BE,$68,$F2,$6F,$E6,$76,$78,$9C,$4F,$08,$C4,$F1,$77,$7F,$0A,$B9
	.dc.b	$72,$FF,$A9,$67,$FD,$07,$CD,$FE,$6D,$7F,$CC,$71,$97,$FC,$F2,$7F
	.dc.b	$99,$F6,$FF,$61,$4E,$DB,$19,$8F,$F5,$43,$4E,$A5,$3F,$DC,$6F,$13
	.dc.b	$6F,$F9,$C6,$3F,$D4,$4F,$F9,$DF,$AF,$FC,$FA,$E6,$7F,$D3,$95,$E5
	.dc.b	$3F,$CF,$71,$93,$FD,$33,$1D,$5F,$09,$93,$D6,$58,$E2,$73,$9E,$ED
	.dc.b	$EB,$FF,$45,$B6,$7E,$2C,$FF,$AA,$DE,$71,$9B,$F8,$71,$D6,$79,$9B
	.dc.b	$9E,$ED,$33,$5F,$EC,$8E,$DF,$F7,$35,$79,$DB,$5F,$4D,$B5,$E1,$6D
	.dc.b	$7F,$A7,$AF,$FD,$58,$78,$AB,$C9,$87,$CE,$6E,$1F,$11,$AF,$D4,$3E
	.dc.b	$B5,$F4,$7F,$DD,$F9,$FF,$B4,$2F,$A6,$3E,$9C,$7E,$2B,$6B,$E2,$C7
	.dc.b	$AF,$B7,$FD,$CB,$1C,$8F,$B9,$1C,$0D,$5F,$5C,$BC,$D7,$CF,$0E,$F5
	.dc.b	$AB,$E2,$76,$3F,$DA,$98,$F9,$E7,$F1,$27,$C2,$BE,$75,$C6,$31,$5F
	.dc.b	$06,$FB,$56,$A3,$9E,$3E,$8A,$E7,$3F,$8B,$9F,$8C,$7D,$DB,$F5,$3C
	.dc.b	$31,$F5,$3D,$63,$C1,$3C,$71,$F3,$3F,$9A,$C7,$C3,$5B,$39,$1E,$EC
	.dc.b	$3E,$0B,$87,$D2,$F8,$B2,$FB,$37,$D2,$3F,$42,$7D,$AC,$BD,$58,$77
	.dc.b	$C7,$C7,$39,$CB,$CE,$F3,$CB,$C6,$D9,$DB,$EF,$3E,$86,$4B,$FF,$7F
	.dc.b	$63,$E1,$9F,$D1,$DA,$FF,$D8,$57,$B9,$FF,$60,$CE,$5F,$F7,$7B,$5F
	.dc.b	$F7,$7B,$5F,$F7,$7B,$5F,$F7,$7C,$7F,$8B,$FB,$7F,$CA,$B1,$FE,$10
	.dc.b	$3F,$C1,$99,$FF,$97,$3B,$3F,$C3,$B1,$FE,$32,$AF,$F0,$89,$FE,$2C
	.dc.b	$7F,$C2,$B9,$7F,$8C,$23,$2F,$F8,$76,$32,$7F,$8C,$0F,$F1,$09,$CD
	.dc.b	$FF,$10,$CF,$F8,$B2,$BC,$1D,$9F,$E2,$1A,$F2,$5F,$F8,$86,$F7,$3F
	.dc.b	$E2,$D9,$EA,$7F,$8B,$67,$C2,$F5,$33,$5F,$E2,$D3,$AB,$FE,$3C,$AE
	.dc.b	$6F,$C0,$BF,$7B,$CE,$33,$5E,$F8,$F1,$C7,$BC,$C9,$EF,$72,$7B,$E3
	.dc.b	$C6,$FF,$93,$A3,$6F,$94,$D9,$DC,$F3,$CF,$C3,$3C,$75,$E9,$69,$9B
	.dc.b	$FF,$5D,$61,$F6,$A7,$DB,$87,$9D,$B8,$79,$F0,$EB,$7F,$25,$F0,$CF
	.dc.b	$58,$E7,$DB,$FE,$75,$9F,$8C,$70,$3E,$34,$69,$FF,$35,$61,$E1,$C7
	.dc.b	$E4,$B6,$7E,$D5,$F8,$AB,$E2,$61,$F0,$9F,$88,$75,$C7,$D1,$87,$BA
	.dc.b	$FC,$D7,$C6,$BE,$31,$E8,$9F,$84,$FF,$A8,$F0,$F0,$3E,$83,$55,$CA
	.dc.b	$7A,$47,$A8,$E1,$8F,$BF,$0F,$C7,$47,$8F,$1F,$B9,$8F,$84,$F8,$D1
	.dc.b	$C0,$F8,$35,$C6,$7E,$74,$78,$0E,$D5,$FA,$F9,$D4,$7C,$23,$B1,$F8
	.dc.b	$3A,$ED,$7E,$CA,$E1,$3D,$88,$7E,$A8,$F9,$27,$77,$E4,$E1,$F5,$63
	.dc.b	$58,$FC,$49,$F6,$C6,$F1,$F6,$DF,$C4,$9E,$33,$ED,$38,$E3,$E8,$7E
	.dc.b	$8E,$3E,$6C,$7E,$61,$DE,$BF,$3F,$8F,$7C,$7E,$1F,$57,$F9,$AF,$0F
	.dc.b	$F3,$79,$0F,$F2,$33,$96,$F0,$7F,$9C,$DD,$70,$FF,$91,$A7,$4D,$9F
	.dc.b	$F4,$1C,$71,$9F,$F4,$5B,$5F,$F4,$CC,$6B,$C2,$FF,$42,$F8,$72,$7F
	.dc.b	$D3,$71,$FE,$80,$30,$FF,$A7,$3B,$93,$98,$FF,$22,$76,$7F,$9F,$1F
	.dc.b	$F3,$BF,$6F,$FA,$01,$B9,$7F,$A0,$CF,$F3,$EF,$64,$8F,$F4,$01,$0F
	.dc.b	$F4,$EE,$DF,$E7,$96,$9F,$E9,$E7,$FD,$43,$E2,$3C,$CF,$F3,$7D,$FF
	.dc.b	$9E,$BB,$7F,$CB,$93,$9E,$9F,$F3,$04,$E5,$FF,$2D,$46,$70,$FF,$62
	.dc.b	$4E,$5F,$F3,$B4,$66,$3F,$D0,$FD,$BF,$E7,$77,$FC,$EB,$E3,$BC,$DF
	.dc.b	$F3,$FF,$5B,$C8,$FE,$B5,$97,$F9,$9E,$B4,$FF,$9C,$67,$5C,$DF,$E7
	.dc.b	$19,$CE,$7F,$ED,$59,$CD,$71,$C3,$AE,$1D,$2B,$FD,$15,$7C,$31,$FF
	.dc.b	$5D,$E1,$C2,$73,$7C,$31,$FF,$3A,$F6,$7F,$8F,$7B,$BF,$BC,$D7,$F7
	.dc.b	$C3,$FB,$97,$07,$F8,$10,$FE,$D8,$FF,$91,$E3,$6F,$F7,$2C,$3F,$BC
	.dc.b	$F5,$FF,$81,$A7,$2F,$F9,$4A,$33,$7F,$E0,$0E,$EF,$EF,$53,$9B,$FF
	.dc.b	$08,$63,$FE,$19,$C3,$8D,$FF,$96,$27,$27,$F8,$2C,$FF,$0C,$4F,$F8
	.dc.b	$AF,$1F,$2D,$FF,$96,$6F,$8E,$3F,$E2,$89,$CE,$1E,$FD,$BF,$C2,$79
	.dc.b	$78,$A3,$B4,$EA,$BF,$CB,$D8,$F8,$EF,$78,$74,$CB,$FC,$2B,$3F,$E1
	.dc.b	$48,$C5,$FC,$78,$C5,$70,$25,$7A,$0F,$F1,$94,$F7,$7F,$CB,$0F,$F8
	.dc.b	$C3,$1F,$26,$1E,$9B,$E7,$5C,$4F,$F1,$41,$F1,$0F,$AE,$FC,$33,$75
	.dc.b	$CF,$0F,$2C,$66,$7F,$C4,$B5,$EE,$7E,$65,$6F,$0F,$3B,$63,$C1,$87
	.dc.b	$91,$F8,$27,$23,$D3,$7F,$1D,$DE,$1E,$07,$DF,$1E,$43,$55,$D4,$8E
	.dc.b	$09,$1E,$46,$BE,$E3,$99,$E1,$8F,$F3,$4B,$B2,$1B,$3E,$3B,$F3,$CD
	.dc.b	$B9,$AF,$64,$7D,$98,$E9,$3F,$E5,$B7,$E7,$65,$F9,$DC,$3E,$2C,$F9
	.dc.b	$23,$13,$F0,$27,$2F,$8A,$33,$5E,$1A,$FF,$27,$E2,$AF,$F2,$AB,$5F
	.dc.b	$F2,$AB,$5D,$94,$53,$B9,$DB,$99,$CC,$61,$E1,$3B,$9D,$36,$33,$3F
	.dc.b	$D7,$6F,$FA,$FD,$7F,$57,$7F,$C2,$5D,$BF,$D5,$EB,$FB,$A1,$FD,$94
	.dc.b	$FE,$E9,$5C,$5B,$5C,$5F,$EF,$27,$F7,$2B,$FE,$C8,$FF,$80,$2F,$E0
	.dc.b	$47,$F8,$66,$FB,$4F,$F6,$AB,$E8,$F1,$6C,$6A,$7F,$C1,$F8,$FF,$80
	.dc.b	$AB,$FB,$99,$FE,$02,$7F,$B7,$F7,$7F,$75,$9D,$E1,$E0,$C3,$FC,$67
	.dc.b	$39,$7F,$C1,$91,$9C,$7F,$C6,$B1,$E0,$9D,$47,$F7,$CC,$3F,$C0,$58
	.dc.b	$76,$7C,$17,$F2,$2B,$71,$EF,$3F,$CE,$78,$FA,$70,$EF,$18,$9D,$DF
	.dc.b	$F8,$38,$F6,$1D,$63,$DA,$47,$04,$BF,$04,$78,$8F,$A3,$7D,$F0,$F4
	.dc.b	$57,$4A,$DD,$73,$6C,$73,$8E,$37,$E4,$9F,$C2,$C7,$C0,$76,$47,$67
	.dc.b	$84,$FC,$31,$F6,$6B,$F0,$8E,$2F,$C8,$FC,$53,$FC,$FE,$E2,$FE,$15
	.dc.b	$7D,$39,$FF,$2C,$63,$F8,$A9,$CC,$7C,$93,$C1,$19,$9F,$3C,$FB,$E7
	.dc.b	$CA,$7B,$CA,$FB,$DF,$26,$1E,$F7,$F3,$67,$BC,$DC,$FB,$DD,$9E,$0C
	.dc.b	$7F,$0C,$FF,$B5,$23,$F7,$9E,$4B,$CA,$FD,$F7,$13,$6F,$E0,$1E,$2B
	.dc.b	$CB,$CC,$61,$B8,$FF,$07,$C3,$93,$FC,$26,$F9,$39,$C3,$6B,$0E,$8E
	.dc.b	$F0,$E8,$D3,$81,$45,$F6,$57,$D9,$E9,$87,$6B,$EB,$1A,$BF,$7E,$1E
	.dc.b	$2F,$35,$E5,$F9,$3C,$17,$9B,$4D,$44,$C3,$C5,$3A,$6C,$E5,$EC,$7F
	.dc.b	$29,$F0,$7F,$C9,$CF,$F2,$96,$B9,$F0,$7F,$CA,$0E,$5A,$F1,$F2,$5E
	.dc.b	$17,$F9,$46,$3F,$98,$52,$73,$1F,$E5,$33,$F9,$97,$27,$F9,$4D,$A7
	.dc.b	$F9,$4C,$FE,$6D,$E0,$7F,$91,$F6,$FA,$DE,$07,$F9,$1D,$FE,$7D,$19
	.dc.b	$9F,$9B,$39,$7F,$A2,$46,$78,$BF,$CA,$47,$D1,$71,$3B,$97,$7F,$F9
	.dc.b	$21,$FE,$90,$D7,$FA,$63,$5F,$EA,$0D,$78,$F7,$FF,$91,$70,$FF,$27
	.dc.b	$1F,$E5,$5F,$03,$FC,$9D,$8F,$F9,$7A,$75,$5F,$E4,$CA,$D7,$3F,$F9
	.dc.b	$33,$B7,$FC,$19,$3F,$DB,$7B,$FF,$C0,$4F,$F7,$29,$FF,$01,$3F,$D9
	.dc.b	$7C,$0F,$F2,$24,$EB,$C0,$FE,$E7,$5F,$E1,$7A,$FF,$28,$61,$FD,$B1
	.dc.b	$E8,$DE,$EF,$ED,$B3,$FD,$CF,$0E,$73,$96,$D7,$F7,$52,$70,$7F,$84
	.dc.b	$1F,$16,$5F,$E5,$62,$3F,$E0,$09,$E3,$3B,$8F,$09,$CA,$FC,$15,$C2
	.dc.b	$7F,$BF,$4F,$F9,$AF,$0F,$2B,$D8,$E9,$3B,$AF,$87,$1D,$2B,$B5,$70
	.dc.b	$8C,$4F,$67,$89,$27,$BB,$FE,$73,$31,$1D,$27,$91,$2F,$D7,$1C,$EB
	.dc.b	$8C,$FF,$9C,$DB,$3F,$E6,$49,$E4,$F0,$AF,$54,$EA,$FF,$CA,$D1,$FE
	.dc.b	$28,$3E,$04,$ED,$F4,$9C,$2B,$9C,$FC,$18,$E0,$7A,$6B,$A3,$FE,$4E
	.dc.b	$3F,$02,$7C,$F3,$91,$C2,$FD,$15,$D2,$3E,$09,$9B,$F5,$36,$FC,$EF
	.dc.b	$4B,$F9,$F3,$E7,$BF,$8D,$5E,$5A,$D5,$FC,$73,$FC,$66,$FF,$8D,$8F
	.dc.b	$F1,$C9,$0F,$F1,$C4,$FF,$8B,$5B,$DB,$FE,$33,$6F,$81,$FE,$32,$8C
	.dc.b	$47,$F4,$B9,$FE,$B0,$FF,$50,$C7,$FA,$C3,$FD,$38,$FE,$B0,$57,$FA
	.dc.b	$C4,$70,$8F,$ED,$C7,$F5,$73,$57,$D6,$BF,$AA,$C7,$29,$E1,$E0,$7F
	.dc.b	$6F,$AF,$EC,$E7,$F6,$FB,$FF,$56,$DF,$F8,$5A,$FF,$BA,$F8,$1F,$D9
	.dc.b	$5E,$F3,$E6,$C3,$9C,$E7,$0E,$95,$EB,$C3,$A1,$1C,$1F,$DE,$4F,$26
	.dc.b	$1E,$6C,$3D,$AD,$9F,$56,$3F,$E0,$AC,$7B,$DF,$F7,$18,$C5,$6E,$3F
	.dc.b	$BB,$90,$F8,$65,$9E,$2F,$82,$78,$91,$F9,$51,$AB,$F1,$CE,$5F,$81
	.dc.b	$5D,$EB,$A9,$2B,$85,$FA,$23,$9C,$EE,$73,$1E,$38,$C1,$2F,$B5,$7C
	.dc.b	$79,$FF,$32,$46,$1F,$91,$7E,$47,$AC,$65,$FA,$E7,$C5,$32,$48,$F4
	.dc.b	$C6,$EF,$FC,$E5,$8F,$D9,$25,$71,$AD,$BC,$6F,$C9,$5F,$E6,$C8,$F1
	.dc.b	$C7,$F8,$09,$FF,$06,$77,$7F,$81,$B8,$7F,$C1,$4D,$DB,$53,$FD,$F0
	.dc.b	$FF,$08,$F7,$7F,$83,$7B,$E4,$7F,$85,$7C,$0F,$F0,$A6,$1F,$E1,$37
	.dc.b	$FC,$2D,$1F,$EA,$76,$BD,$0C,$57,$FA,$95,$78,$1F,$E1,$47,$FC,$1A
	.dc.b	$43,$C4,$E5,$BE,$07,$F8,$5D,$D5,$FF,$88,$E7,$4F,$F8,$97,$B3,$FC
	.dc.b	$23,$26,$93,$BB,$FC,$23,$19,$ED,$FF,$06,$BF,$E1,$06,$BF,$E1,$B6
	.dc.b	$BF,$E1,$7E,$EF,$F0,$83,$FE,$C5,$E1,$AF,$FB,$25,$FF,$61,$77,$7F
	.dc.b	$83,$60,$FF,$83,$FC,$F7,$9F,$FE,$32,$E1,$FF,$20,$4E,$C9,$E2,$7F
	.dc.b	$93,$35,$FF,$26,$DF,$0F,$0B,$17,$FE,$1A,$AB,$8F,$F8,$33,$BE,$BF
	.dc.b	$E0,$C9,$FF,$27,$5F,$83,$D0,$CF,$84,$BA,$4A,$DE,$8A,$C6,$1F,$E0
	.dc.b	$F2,$67,$E4,$9D,$57,$F8,$62,$7F,$15,$D2,$EF,$77,$FF,$85,$23,$F9
	.dc.b	$EC,$FF,$46,$3F,$9E,$9F,$D3,$0A,$7F,$58,$23,$C3,$83,$FA,$A8,$EB
	.dc.b	$FF,$0B,$47,$7A,$ED,$7F,$D2,$C8,$FF,$4B,$8D,$5A,$FF,$BC,$F8,$1F
	.dc.b	$E2,$08,$1F,$E2,$73,$FA,$C1,$FD,$8E,$78,$CF,$0F,$07,$FC,$50,$4C
	.dc.b	$BF,$C4,$18,$77,$C7,$E0,$E1,$DF,$2F,$F7,$7C,$7F,$64,$8F,$7E,$3E
	.dc.b	$D9,$FE,$D5,$8F,$B9,$F6,$D7,$28,$F0,$61,$DA,$FE,$55,$F9,$2F,$AE
	.dc.b	$1F,$DB,$A3,$13,$FE,$04,$8F,$21,$F1,$8A,$F9,$4D,$1D,$30,$F0,$57
	.dc.b	$48,$F8,$51,$87,$78,$79,$DF,$93,$97,$CB,$7E,$51,$EE,$9F,$A6,$F5
	.dc.b	$BF,$13,$4F,$11,$FE,$13,$BF,$86,$47,$E9,$DF,$E0,$5E,$F9,$7D,$17
	.dc.b	$FC,$2C,$7E,$22,$FB,$CF,$96,$3C,$78,$FC,$47,$B1,$D2,$FE,$7E,$1E
	.dc.b	$4A,$E1,$8F,$AC,$98,$F7,$8F,$F0,$CC,$FC,$A7,$EA,$DF,$8B,$0F,$B8
	.dc.b	$F5,$3C,$F9,$7D,$99,$EB,$7C,$F1,$FF,$0F,$56,$CF,$3E,$5F,$3D,$F3
	.dc.b	$9F,$2F,$1F,$04,$FD,$5B,$F8,$93,$C2,$7D,$53,$F2,$61,$8F,$E8,$9F
	.dc.b	$F1,$51,$0F,$F1,$69,$FE,$FF,$3F,$C4,$E7,$E9,$0C,$10,$FE,$D1,$DF
	.dc.b	$FD,$67,$C0,$3F,$EF,$18,$FD,$5C,$FF,$55,$7F,$B1,$9F,$DF,$CF,$EB
	.dc.b	$47,$F5,$83,$99,$83,$78,$7F,$6D,$3F,$B0,$92,$FF,$B0,$DF,$F8,$96
	.dc.b	$BF,$B4,$D7,$F6,$79,$CD,$F8,$E3,$FC,$4F,$18,$7F,$1F,$E0,$FF,$89
	.dc.b	$9C,$9E,$33,$FC,$15,$7F,$E0,$37,$C9,$7D,$9F,$56,$1E,$37,$84,$7F
	.dc.b	$83,$1E,$38,$7C,$53,$C5,$1F,$0C,$E5,$3E,$DC,$3C,$8F,$DA,$C3,$C8
	.dc.b	$4F,$07,$FC,$07,$87,$82,$B6,$7F,$92,$1F,$0E,$3E,$7B,$F2,$47,$CA
	.dc.b	$9F,$F1,$14,$63,$0F,$0C,$60,$E7,$3F,$E4,$43,$81,$F5,$27,$85,$7D
	.dc.b	$E8,$FD,$BD,$7A,$E3,$13,$FE,$58,$C7,$E8,$63,$E9,$BF,$A8,$79,$A3
	.dc.b	$E2,$3C,$AB,$ED,$1C,$B0,$FF,$18,$57,$19,$F8,$07,$13,$E9,$CE,$DF
	.dc.b	$C7,$BC,$AB,$E7,$B6,$3B,$47,$93,$1F,$97,$87,$D1,$9E,$8F,$BA,$78
	.dc.b	$9B,$8F,$49,$F1,$0F,$BA,$78,$9D,$BF,$8C,$9E,$B8,$7C,$D8,$FC,$67
	.dc.b	$7F,$F8,$AB,$6F,$F1,$8F,$0F,$F8,$AF,$BF,$F5,$CF,$F8,$DB,$D1,$B3
	.dc.b	$A2,$5F,$F5,$6B,$E4,$FF,$58,$F0,$95,$C2,$35,$85,$7F,$C7,$A7,$F8
	.dc.b	$ED,$C4,$F0,$8F,$EC,$38,$7F,$8B,$7C,$1E,$1E,$0F,$F8,$BB,$2F,$E0
	.dc.b	$4D,$AF,$EC,$FE,$2F,$FA,$B6,$F8,$47,$2A,$DD,$F2,$3C,$55,$B8,$FE
	.dc.b	$19,$E1,$7F,$8D,$6F,$F8,$75,$E3,$C2,$FF,$1B,$CF,$F8,$EB,$C2,$FF
	.dc.b	$19,$CF,$F8,$E7,$C2,$DB,$FE,$35,$2C,$A7,$71,$FD,$F3,$8F,$FC,$6E
	.dc.b	$D3,$FC,$A8,$F2,$5E,$BB,$D5,$FF,$16,$F8,$2F,$51,$EA,$FF,$8B,$AB
	.dc.b	$5B,$7F,$8B,$A7,$2F,$C1,$C3,$FC,$17,$3A,$5B,$7F,$8F,$2B,$71,$FE
	.dc.b	$71,$9D,$4E,$6F,$FC,$77,$E3,$3C,$5F,$F1,$FC,$BC,$97,$95,$A9,$CC
	.dc.b	$63,$3F,$F2,$6C,$62,$75,$1E,$5A,$E5,$18,$CB,$CB,$3F,$82,$BE,$51
	.dc.b	$FE,$85,$C3,$C5,$3C,$67,$36,$8F,$E9,$DD,$FF,$DB,$4F,$ED,$DE,$1F
	.dc.b	$F6,$88,$FE,$C1,$D9,$AF,$0B,$FA,$E5,$6B,$C3,$FE,$C5,$9F,$F9,$0E
	.dc.b	$39,$4F,$F8,$9D,$BE,$1F,$F8,$76,$38,$99,$2B,$93,$FB,$0C,$FF,$6B
	.dc.b	$C7,$C4,$FF,$75,$9F,$F0,$2E,$1E,$3A,$DC,$FF,$6D,$3F,$C2,$9C,$9F
	.dc.b	$DF,$0F,$EF,$EE,$27,$FB,$81,$FE,$08,$7F,$91,$D7,$8E,$B9,$E1,$E8
	.dc.b	$9F,$F1,$4C,$62,$7C,$F1,$83,$8C,$7A,$5A,$66,$FC,$57,$FE,$06,$8F
	.dc.b	$34,$7F,$82,$B2,$F9,$B1,$EF,$CB,$E5,$BF,$E2,$3B,$FF,$03,$BF,$E3
	.dc.b	$8B,$F8,$0F,$6B,$FF,$04,$E5,$F1,$9F,$79,$D6,$3E,$45,$72,$7F,$C2
	.dc.b	$79,$7B,$67,$8E,$1E,$08,$FF,$10,$CF,$9B,$0E,$F5,$EA,$8F,$BD,$97
	.dc.b	$C2,$8E,$39,$7A,$A7,$81,$EB,$71,$7D,$EF,$E8,$57,$3A,$F7,$1F,$54
	.dc.b	$F7,$1C,$4F,$49,$93,$D3,$7E,$19,$E6,$7D,$C8,$FF,$1B,$9C,$27,$E1
	.dc.b	$1F,$12,$3D,$66,$E3,$FC,$60,$7C,$B7,$F0,$8F,$F8,$B4,$FF,$11,$CF
	.dc.b	$FA,$11,$C1,$FE,$34,$F0,$BF,$B6,$3F,$E3,$02,$1F,$E2,$F3,$FA,$7D
	.dc.b	$7F,$55,$7F,$AD,$9F,$D3,$CF,$EA,$B7,$FD,$F1,$FE,$F1,$7F,$DF,$1F
	.dc.b	$F0,$E9,$FD,$66,$3F,$C2,$C4,$AF,$F0,$44,$EF,$C3,$FE,$E1,$7F,$D9
	.dc.b	$6F,$85,$69,$FE,$CF,$5B,$F0,$BF,$AF,$C7,$80,$C9,$0E,$3E,$1F,$F8
	.dc.b	$02,$FF,$C0,$D1,$E9,$3F,$BE,$9A,$3F,$B6,$1E,$E7,$FC,$49,$3D,$A3
	.dc.b	$E1,$D7,$85,$F8,$37,$FE,$03,$7E,$65,$FF,$79,$7A,$E3,$F1,$5E,$27
	.dc.b	$95,$C1,$F0,$E3,$CD,$5F,$E1,$2B,$F0,$4E,$6B,$CB,$87,$A8,$F2,$BE
	.dc.b	$2B,$FF,$01,$C6,$23,$C3,$87,$96,$BF,$C5,$15,$AC,$3F,$C7,$55,$F0
	.dc.b	$6F,$EF,$1D,$63,$35,$E6,$BF,$A5,$5B,$9F,$84,$F6,$BE,$8F,$F8,$EE
	.dc.b	$B8,$47,$B2,$3F,$C4,$EF,$C5,$BF,$BD,$3A,$9F,$6D,$7C,$17,$D8,$7D
	.dc.b	$37,$C7,$7E,$B3,$FC,$5D,$5E,$79,$F0,$46,$6F,$B5,$7F,$88,$E7,$D0
	.dc.b	$7D,$19,$D3,$F8,$3A,$F1,$C6,$1F,$94,$75,$8F,$F1,$97,$7F,$F8,$4D
	.dc.b	$77,$3C,$8F,$F0,$79,$FE,$11,$8F,$EE,$82,$78,$4E,$E3,$2D,$7F,$B1
	.dc.b	$83,$87,$81,$CD,$C9,$E0,$AE,$EF,$6A,$F2,$6C,$F6,$3F,$C6,$5E,$C3
	.dc.b	$D8,$A7,$F8,$D6,$75,$7E,$66,$BD,$A7,$83,$EA,$6B,$E2,$8D,$F3,$7F
	.dc.b	$73,$8F,$47,$83,$FE,$2D,$7F,$C4,$BE,$1F,$F8,$B0,$FF,$16,$F8,$5F
	.dc.b	$E2,$C9,$FF,$14,$78,$7F,$E2,$36,$D7,$F8,$B2,$7F,$C5,$6B,$83,$FC
	.dc.b	$56,$D3,$FC,$56,$7F,$8D,$3C,$0F,$F0,$DF,$47,$C9,$F0,$BF,$C5,$5E
	.dc.b	$0F,$F8,$A7,$C3,$FF,$16,$F8,$7F,$E2,$2D,$BD,$18,$FF,$8C,$A7,$35
	.dc.b	$BC,$3F,$C5,$CF,$F8,$5E,$BD,$C7,$F8,$6A,$3F,$C3,$E7,$F8,$57,$C2
	.dc.b	$D3,$FE,$1B,$6D,$FF,$8C,$BC,$1F,$F0,$CC,$FF,$B7,$7C,$2F,$F0,$EC
	.dc.b	$FE,$02,$FA,$3F,$E2,$1F,$0F,$FC,$55,$5F,$8A,$EE,$FE,$A5,$3F,$D1
	.dc.b	$B7,$FE,$88,$FF,$4F,$6B,$FD,$68,$91,$C1,$FE,$8F,$1F,$D6,$63,$FB
	.dc.b	$39,$9F,$0F,$FB,$97,$77,$F7,$F8,$C7,$85,$FD,$B7,$9F,$FC,$33,$87
	.dc.b	$7A,$D4,$7F,$5B,$8F,$24,$63,$C3,$FF,$08,$12,$3F,$AB,$1A,$3F,$B9
	.dc.b	$B9,$C7,$C9,$5D,$A7,$D2,$F5,$7D,$35,$FD,$87,$1F,$3C,$72,$3A,$9A
	.dc.b	$3F,$B9,$9C,$AF,$B4,$F1,$C3,$D9,$1F,$E0,$DC,$7F,$C8,$99,$7F,$80
	.dc.b	$31,$F0,$63,$DB,$1F,$EF,$F9,$7F,$74,$CB,$DD,$1E,$09,$D6,$1F,$32
	.dc.b	$F8,$E5,$DA,$BF,$C0,$53,$D9,$FE,$FD,$7F,$DF,$1F,$3C,$7D,$0C,$7C
	.dc.b	$59,$7A,$27,$38,$74,$9C,$C7,$87,$0E,$D9,$79,$63,$E2,$3F,$32,$FD
	.dc.b	$D7,$FE,$47,$9E,$17,$CA,$B5,$1E,$69,$D4,$60,$E9,$1F,$0B,$1F,$43
	.dc.b	$FE,$1C,$BF,$BC,$FF,$83,$B1,$5F,$F8,$46,$3F,$8E,$08,$CC,$FF,$A8
	.dc.b	$D7,$81,$FE,$14,$9F,$F0,$94,$63,$C1,$D4,$E6,$7F,$C0,$07,$F3,$C2
	.dc.b	$4F,$F4,$BC,$7F,$AD,$E3,$FD,$2E,$7F,$AB,$1F,$D2,$CF,$E8,$F7,$FD
	.dc.b	$61,$FF,$0B,$D7,$19,$FE,$BB,$5F,$D1,$2B,$5E,$1F,$F8,$5A,$7F,$A6
	.dc.b	$56,$EF,$FA,$65,$6B,$C3,$FE,$F3,$8F,$F6,$8A,$D5,$79,$A3,$FA,$75
	.dc.b	$FF,$5C,$BE,$13,$FD,$83,$1F,$EA,$18,$F7,$8F,$7E,$1C,$89,$87,$28
	.dc.b	$F5,$9E,$D3,$D6,$E5,$B7,$D9,$F8,$77,$DB,$1F,$3C,$FF,$77,$9F,$F0
	.dc.b	$D9,$FD,$97,$1F,$7D,$F3,$C7,$C9,$3F,$E1,$C3,$D9,$5D,$70,$F9,$11
	.dc.b	$E3,$9D,$E3,$FE,$1B,$9F,$AB,$3F,$E0,$43,$B1,$EE,$BE,$D8,$F9,$23
	.dc.b	$FC,$0A,$6E,$7F,$BA,$E3,$E4,$C7,$C0,$D9,$F0,$57,$1A,$EF,$1E,$C7
	.dc.b	$8B,$DE,$B8,$B4,$E2,$7D,$3A,$FC,$54,$FB,$0F,$91,$19,$3C,$91,$F6
	.dc.b	$DE,$27,$CE,$78,$9E,$58,$E3,$1F,$5F,$0F,$F2,$37,$81,$FE,$06,$F0
	.dc.b	$7F,$BE,$CF,$F8,$02,$3F,$C2,$06,$5F,$F0,$4C,$7F,$81,$78,$BF,$AD
	.dc.b	$CE,$67,$FC,$0F,$E1,$FF,$83,$67,$3E,$1F,$F8,$07,$C3,$FF,$01,$D7
	.dc.b	$F7,$DF,$0F,$0F,$F8,$0E,$32,$FF,$80,$E3,$2F,$F8,$4B,$C0,$FE,$FC
	.dc.b	$FE,$F0,$38,$7F,$7E,$FF,$7C,$74,$D7,$3E,$07,$F8,$0C,$FE,$FF,$AF
	.dc.b	$F7,$E3,$A4,$75,$6B,$FD,$DF,$97,$FB,$A9,$FE,$2E,$F0,$3F,$C0,$30
	.dc.b	$F0,$3F,$C0,$33,$9E,$FF,$EF,$AD,$38,$CF,$F6,$F7,$AF,$67,$F7,$48
	.dc.b	$F6,$35,$FE,$F1,$D9,$FE,$08,$6C,$FF,$80,$4F,$F0,$1B,$FD,$EF,$B7
	.dc.b	$FB,$FB,$FE,$3B,$AF,$EF,$FC,$1F,$E0,$A9,$CC,$A7,$35,$FE,$09,$3F
	.dc.b	$C0,$1D,$7F,$CF,$E7,$FA,$39,$FD,$C4,$FE,$7B,$CD,$FD,$6E,$7F,$9D
	.dc.b	$36,$36,$7F,$61,$3F,$9F,$98,$E6,$FE,$9C,$DA,$ED,$B7,$F8,$2A,$72
	.dc.b	$D8,$CD,$7F,$60,$8F,$E8,$8F,$F4,$29,$FE,$B9,$1B,$EB,$FE,$B1,$7F
	.dc.b	$D1,$4F,$0C,$65,$B5,$CC,$F4,$46,$4E,$B1,$EF,$3F,$A8,$16,$BF,$B7
	.dc.b	$57,$67,$C3,$5C,$E7,$FB,$3C,$E6,$FF,$AD,$CF,$63,$53,$E5,$BE,$EF
	.dc.b	$C7,$6D,$FC,$29,$DB,$DA,$7E,$89,$D8,$F5,$98,$3A,$47,$77,$11,$C5
	.dc.b	$F5,$9D,$E3,$4E,$23,$C0,$F0,$AF,$4B,$DE,$BC,$B1,$F5,$23,$57,$FE
	.dc.b	$89,$95,$FF,$70,$ED,$FF,$0A,$DF,$F8,$B7,$0E,$B8,$77,$8E,$4D,$C3
	.dc.b	$E4,$1C,$EF,$97,$6F,$F9,$FF,$1E,$B5,$FE,$97,$C3,$FB,$DB,$FE,$77
	.dc.b	$AF,$EF,$F1,$88,$FF,$02,$61,$FE,$03,$78,$35,$FE,$9E,$7F,$0B,$8F
	.dc.b	$F0,$03,$FC,$09,$AE,$7B,$3F,$C0,$4F,$F5,$18,$FE,$C7,$D7,$FC,$EA
	.dc.b	$31,$3F,$D2,$9F,$E6,$C7,$F4,$1B,$FE,$6D,$87,$F3,$23,$F9,$E3,$FC
	.dc.b	$DA,$3F,$A1,$46,$63,$F9,$C1,$98,$EB,$5D,$EF,$FA,$E4,$FF,$82,$0D
	.dc.b	$CF,$F4,$43,$FB,$AD,$73,$7A,$C7,$39,$E3,$1F,$E0,$A3,$FA,$4C,$F9
	.dc.b	$1E,$47,$F7,$A3,$33,$D9,$F6,$1D,$9E,$97,$FE,$11,$3E,$14,$7F,$53
	.dc.b	$39,$F6,$FF,$74,$8D,$61,$EB,$6C,$F6,$3C,$E7,$0E,$DF,$ED,$97,$EF
	.dc.b	$8C,$9E,$23,$CD,$87,$BD,$F8,$E7,$32,$B9,$C7,$E0,$C7,$F8,$28,$D3
	.dc.b	$83,$EE,$B8,$BF,$2C,$F1,$6D,$7C,$0C,$3F,$C3,$57,$F8,$AA,$FE,$FF
	.dc.b	$7E,$0A,$D3,$FE,$0C,$3D,$18,$FC,$3B,$E3,$81,$FD,$14,$FC,$D3,$E1
	.dc.b	$1F,$8C,$C3,$B5,$FE,$30,$EB,$87,$F8,$42,$33,$3C,$4E,$F8,$FF,$99
	.dc.b	$27,$F5,$71,$F2,$67,$FC,$07,$D5,$FE,$02,$E2,$FF,$6B,$F5,$FF,$7B
	.dc.b	$E1,$FF,$35,$F6,$7F,$81,$2B,$51,$FE,$6E,$9D,$60,$15,$B7,$FB,$5A
	.dc.b	$AF,$ED,$B1,$A6,$C6,$63,$FB,$8E,$FF,$F0,$02,$FF,$CF,$7E,$1B,$B3
	.dc.b	$FC,$0F,$97,$FB,$B2,$B5,$7E,$16,$AA,$E0,$FF,$84,$4F,$F0,$81,$E7
	.dc.b	$9F,$F7,$AF,$67,$F8,$2E,$3D,$4D,$7F,$C1,$FD,$9F,$E0,$9F,$0B,$FC
	.dc.b	$0B,$3F,$E0,$D9,$DC,$FF,$83,$1F,$6E,$FF,$E0,$D7,$FC,$1B,$DB,$FE
	.dc.b	$08,$7F,$C1,$83,$06,$9F,$27,$DB,$7B,$6F,$6F,$37,$BA,$FF,$4E,$E1
	.dc.b	$CC,$FF,$09,$F8,$1F,$E1,$29,$D1,$F4,$E3,$3E,$2B,$C5,$E7,$9F,$4A
	.dc.b	$FF,$02,$78,$38,$C7,$FC,$1F,$1C,$B1,$79,$6F,$17,$53,$FE,$0F,$BA
	.dc.b	$F2,$BF,$F0,$AE,$1E,$8C,$38,$DE,$BC,$2F,$F0,$8D,$F1,$BF,$F0,$E3
	.dc.b	$FE,$25,$3F,$C3,$9E,$37,$4E,$BF,$E9,$27,$F7,$28,$FE,$C2,$7F,$49
	.dc.b	$2B,$9E,$2F,$EA,$FE,$17,$F4,$C9,$FE,$8C,$7E,$57,$C3,$FE,$AB,$1F
	.dc.b	$ED,$63,$F5,$F1,$FE,$FB,$3F,$B5,$9F,$D8,$7B,$FF,$B3,$DF,$56,$F5
	.dc.b	$7F,$5D,$3F,$AD,$C6,$67,$81,$0E,$64,$7F,$BA,$1B,$71,$1F,$DC,$1B
	.dc.b	$5E,$4B,$F3,$C7,$FB,$9E,$31,$E1,$FF,$BC,$B0,$FF,$12,$5F,$1C,$3A
	.dc.b	$BE,$5A,$F8,$58,$FF,$00,$C7,$9B,$FE,$EE,$AF,$11,$C2,$7E,$26,$3F
	.dc.b	$20,$F8,$91,$97,$C4,$66,$B9,$E1,$E2,$AF,$2E,$7F,$E4,$6A,$F8,$57
	.dc.b	$C7,$0F,$06,$3E,$59,$FB,$2D,$8F,$B3,$18,$BF,$1C,$7A,$72,$F8,$13
	.dc.b	$FE,$35,$AE,$18,$FD,$4B,$DB,$FE,$0A,$C7,$C5,$5F,$30,$F4,$5F,$27
	.dc.b	$FC,$48,$79,$AF,$E5,$12,$FD,$71,$FE,$59,$7F,$D0,$B8,$FF,$83,$E3
	.dc.b	$FC,$23,$3F,$E0,$D8,$C4,$7F,$B3,$CF,$F0,$57,$83,$C3,$C2,$FF,$05
	.dc.b	$CF,$F8,$3A,$7F,$B6,$3F,$CE,$1F,$EB,$79,$7F,$3C,$3F,$B0,$3F,$D2
	.dc.b	$CF,$E8,$67,$F6,$D3,$F9,$E9,$FD,$16,$7F,$A5,$3F,$ED,$E3,$81,$A3
	.dc.b	$84,$F7,$F0,$FF,$A9,$47,$F5,$58,$C7,$87,$FD,$C2,$3F,$C0,$DA,$FF
	.dc.b	$61,$C7,$FA,$B9,$FA,$8B,$E3,$1F,$D4,$70,$F0,$BF,$D8,$B0,$FE,$D9
	.dc.b	$1F,$E0,$48,$FF,$07,$5F,$8D,$B3,$E0,$8F,$55,$F5,$C3,$C5,$1F,$DD
	.dc.b	$EB,$C4,$F9,$EF,$BE,$3F,$D8,$E3,$E2,$CF,$49,$CC,$7F,$63,$C3,$B3
	.dc.b	$F0,$A3,$2F,$D0,$38,$D7,$D4,$8C,$E3,$EE,$8F,$A7,$5E,$CC,$7F,$C3
	.dc.b	$47,$C4,$BF,$A1,$19,$7F,$C1,$51,$9C,$3E,$26,$1F,$E0,$A9,$EE,$7F
	.dc.b	$87,$2B,$CB,$3E,$9A,$F8,$53,$F1,$0E,$B5,$C5,$F9,$57,$E7,$9D,$C7
	.dc.b	$B5,$B5,$CE,$38,$C7,$8E,$7E,$44,$72,$3C,$55,$F0,$31,$FF,$04,$4F
	.dc.b	$F9,$BA,$3E,$3E,$1F,$E0,$8A,$FE,$D5,$AF,$F8,$0F,$87,$FC,$0B,$1F
	.dc.b	$E0,$3F,$07,$FC,$04,$FF,$54,$3F,$C0,$07,$03,$5E,$17,$F8,$25,$D7
	.dc.b	$87,$F9,$A7,$84,$F2,$E6,$FE,$B0,$FF,$81,$E3,$7A,$F1,$9F,$F0,$27
	.dc.b	$8D,$FE,$06,$CB,$FC,$13,$E2,$65,$ED,$C5,$FE,$06,$24,$E7,$0F,$F0
	.dc.b	$53,$5F,$F0,$57,$25,$F0,$BF,$C0,$84,$F0,$BF,$C0,$93,$E9,$3F,$C0
	.dc.b	$7E,$1F,$F8,$1A,$3F,$C0,$91,$93,$FD,$50,$7F,$81,$88,$71,$F7,$5E
	.dc.b	$F7,$F6,$EF,$7D,$EF,$BE,$07,$FB,$7D,$6B,$6F,$95,$5A,$9C,$DF,$1B
	.dc.b	$E1,$5F,$3F,$06,$1C,$5F,$F0,$3E,$1C,$67,$5E,$2F,$F8,$23,$C4,$FE
	.dc.b	$FA,$D8,$CF,$97,$FE,$02,$F0,$FF,$C2,$1C,$9F,$E0,$BA,$D6,$3F,$7D
	.dc.b	$B1,$9A,$E1,$E3,$7F,$84,$F0,$F2,$75,$FF,$7E,$E5,$FE,$77,$8F,$F3
	.dc.b	$3F,$07,$F9,$8F,$81,$FE,$F7,$EF,$FE,$7D,$3F,$D4,$A3,$FA,$26,$1F
	.dc.b	$CF,$B0,$FE,$9D,$5F,$D4,$EB,$F9,$FC,$E7,$0F,$E7,$91,$83,$F9,$E1
	.dc.b	$FD,$3C,$E9,$3F,$D2,$5C,$DF,$F5,$42,$47,$89,$FE,$B7,$87,$F5,$09
	.dc.b	$FF,$07,$73,$FF,$4D,$C7,$D9,$39,$9F,$ED,$78,$78,$AF,$D3,$1D,$B0
	.dc.b	$E8,$7A,$63,$27,$C1,$8F,$88,$7A,$5C,$3F,$D7,$5D,$C7,$F5,$4A,$F0
	.dc.b	$78,$5F,$DD,$AB,$FB,$44,$63,$1F,$36,$1F,$E1,$A3,$E8,$CF,$3B,$F8
	.dc.b	$58,$76,$C7,$E6,$CF,$23,$E3,$9F,$03,$2F,$9D,$8F,$D8,$BF,$14,$E8
	.dc.b	$F8,$31,$97,$FD,$EA,$F8,$E7,$8C,$7B,$67,$8D,$FC,$E8,$E5,$97,$AD
	.dc.b	$FE,$F2,$79,$CF,$11,$F2,$CF,$8F,$1F,$11,$EE,$72,$CB,$FC,$DB,$1F
	.dc.b	$EA,$4C,$59,$32,$FF,$22,$C7,$F7,$CB,$34,$93,$FE,$01,$F0,$3F,$BE
	.dc.b	$93,$B3,$FC,$00,$4A,$FE,$B0,$FF,$34,$CF,$F9,$C4,$7F,$34,$32,$5C
	.dc.b	$7F,$B6,$4E,$E3,$F9,$AB,$5F,$EF,$B8,$7F,$48,$AE,$75,$FD,$3A,$BF
	.dc.b	$9F,$57,$F6,$1B,$FE,$78,$FF,$4B,$9D,$3E,$06,$9E,$07,$C5,$5F,$D8
	.dc.b	$27,$FA,$85,$F7,$9F,$19,$A3,$FC,$0F,$3F,$E1,$96,$9F,$D7,$8F,$3C
	.dc.b	$F8,$9F,$EA,$C7,$13,$FB,$F4,$71,$69,$C6,$7F,$B6,$57,$F8,$08,$F3
	.dc.b	$1E,$89,$F3,$5F,$C0,$AF,$EF,$87,$F5,$D3,$C2,$FC,$9A,$E7,$3D,$DF
	.dc.b	$F8,$03,$5D,$63,$9D,$7C,$57,$FB,$3C,$FC,$18,$E6,$7A,$5E,$31,$CE
	.dc.b	$FF,$C4,$0F,$D4,$38,$BF,$06,$BF,$BA,$1E,$9A,$F8,$E7,$9E,$BC,$AF
	.dc.b	$C1,$9F,$A2,$71,$AE,$91,$F7,$6F,$D9,$3A,$7F,$0D,$87,$FB,$93,$CA
	.dc.b	$FF,$71,$D7,$F7,$36,$F0,$7F,$68,$8F,$F7,$29,$9E,$DF,$ED,$4F,$F6
	.dc.b	$43,$31,$FE,$C9,$9E,$3B,$FF,$6E,$9C,$BF,$DE,$FE,$05,$D7,$FD,$AF
	.dc.b	$CD,$FF,$6E,$57,$F5,$36,$BF,$E0,$59,$DB,$FD,$5D,$AF,$F8,$26,$37
	.dc.b	$87,$FB,$65,$B7,$DE,$FF,$BB,$47,$FB,$B6,$BF,$B8,$77,$7F,$78,$8C
	.dc.b	$47,$F7,$78,$FF,$7D,$3F,$EF,$EE,$EF,$EF,$04,$7F,$BA,$9F,$DE,$78
	.dc.b	$7F,$DF,$8F,$F7,$7E,$DF,$ED,$DD,$DF,$DC,$32,$FE,$E7,$DD,$FD,$C2
	.dc.b	$BF,$C4,$1D,$DF,$DB,$F5,$FF,$7D,$F7,$7F,$6F,$9C,$E3,$CB,$0F,$ED
	.dc.b	$A7,$F8,$1E,$1D,$8E,$CF,$F7,$C4,$7F,$89,$E3,$13,$B6,$C7,$F8,$3D
	.dc.b	$FE,$E3,$1B,$9F,$EE,$3D,$DF,$DF,$1B,$1E,$0A,$FF,$80,$1D,$BF,$DD
	.dc.b	$5F,$F1,$C7,$57,$F5,$A7,$F9,$44,$FF,$36,$DF,$F9,$64,$7F,$BC,$DF
	.dc.b	$EC,$AF,$F3,$0E,$AF,$E9,$27,$F3,$A3,$FA,$E1,$87,$F9,$9C,$7F,$3D
	.dc.b	$24,$6A,$3F,$9D,$C7,$F5,$1E,$8F,$ED,$13,$C6,$7C,$0F,$F6,$87,$FA
	.dc.b	$A4,$F2,$6C,$73,$7F,$A8,$9F,$D3,$4F,$0B,$9E,$DF,$E9,$98,$7B,$EB
	.dc.b	$95,$7F,$4E,$8F,$EC,$2D,$BF,$1B,$E6,$BF,$73,$E0,$21,$BE,$CF,$EC
	.dc.b	$0F,$A0,$F8,$C7,$13,$C6,$70,$AF,$14,$F9,$7B,$7F,$BB,$D7,$FB,$F2
	.dc.b	$3E,$14,$FF,$BE,$AF,$C7,$7E,$EC,$3E,$AC,$E6,$FC,$2D,$9F,$84,$78
	.dc.b	$CF,$ED,$47,$1B,$EB,$7F,$E0,$49,$F8,$27,$92,$7C,$D1,$FE,$09,$AF
	.dc.b	$EF,$51,$FD,$C4,$E7,$8F,$CC,$9D,$47,$E2,$8D,$57,$48,$CE,$1F,$DC
	.dc.b	$DF,$F7,$61,$2F,$FD,$E2,$FF,$6F,$EC,$FE,$D9,$7F,$EE,$E3,$FC,$DD
	.dc.b	$7F,$EE,$D7,$FA,$AD,$FF,$2D,$7F,$A5,$9F,$CB,$7B,$7F,$93,$3F,$CE
	.dc.b	$5F,$E6,$EF,$0C,$BF,$9B,$DF,$F2,$C3,$FA,$67,$8B,$FE,$C3,$9F,$E6
	.dc.b	$F8,$7F,$43,$3F,$9B,$BD,$67,$7E,$2F,$FB,$6A,$7F,$A2,$CF,$F3,$B9
	.dc.b	$CB,$FD,$CE,$FB,$BE,$38,$FF,$66,$BF,$D4,$4F,$E7,$C6,$4F,$F6,$69
	.dc.b	$93,$FA,$FC,$F8,$F0,$F6,$E3,$F0,$63,$C3,$3E,$53,$86,$3E,$B7,$E2
	.dc.b	$C7,$1B,$F2,$C7,$F7,$6A,$E5,$7F,$1A,$3F,$B0,$C6,$27,$CE,$FF,$74
	.dc.b	$C3,$D6,$E6,$7C,$EE,$5F,$5B,$93,$FB,$83,$F3,$2B,$FA,$F3,$F4,$CD
	.dc.b	$CF,$42,$38,$21,$F0,$A3,$8B,$E6,$8D,$B9,$AF,$B8,$FC,$2C,$7E,$B5
	.dc.b	$7B,$70,$3F,$EE,$03,$E0,$65,$FD,$AF,$BF,$FB,$2D,$FF,$92,$3C,$1F
	.dc.b	$F6,$44,$E6,$3F,$D6,$1E,$07,$F6,$67,$FB,$51,$FE,$E3,$84,$FF,$3F
	.dc.b	$8E,$47,$F4,$18,$CB,$63,$5E,$07,$F6,$4E,$9F,$EC,$5E,$57,$F6,$D3
	.dc.b	$FD,$B1,$19,$E4,$FE,$C9,$7F,$DB,$E7,$4F,$F6,$DF,$07,$FB,$3D,$FF
	.dc.b	$4F,$F0,$3F,$B3,$3F,$D8,$67,$FB,$34,$FF,$B7,$9B,$E0,$7F,$62,$7F
	.dc.b	$B2,$1F,$D9,$81,$FE,$CE,$EC,$FE,$C8,$A9,$8E,$B9,$9F,$06,$F8,$3D
	.dc.b	$D6,$AF,$9F,$81,$FD,$A7,$0E,$BE,$07,$F6,$9A,$D5,$77,$9F,$EE,$3E
	.dc.b	$17,$F5,$D7,$FA,$E4,$F3,$F3,$9E,$73,$87,$58,$7F,$B5,$FC,$1B,$7B
	.dc.b	$AD,$61,$C8,$FE,$D4,$7B,$A3,$CD,$E3,$7F,$B8,$A3,$37,$FD,$AB,$C3
	.dc.b	$F5,$57,$9B,$AF,$F9,$44,$FF,$43,$3F,$98,$BF,$CC,$FB,$FF,$9B,$F4
	.dc.b	$FF,$2F,$3F,$9D,$F8,$3F,$CA,$79,$3F,$A0,$5F,$F4,$73,$F9,$C5,$FF
	.dc.b	$29,$8C,$C7,$F3,$93,$51,$D7,$C0,$FE,$77,$3F,$CD,$B0,$FE,$D8,$FF
	.dc.b	$44,$C3,$C0,$D3,$53,$FD,$72,$FD,$15,$E6,$9C,$C7,$F5,$83,$33,$D9
	.dc.b	$FE,$7D,$3D,$9F,$7D,$76,$6D,$72,$7D,$71,$F0,$3C,$2F,$EB,$F1,$FE
	.dc.b	$F5,$AF,$EE,$78,$7C,$A2,$57,$F4,$8C,$3F,$B3,$69,$FE,$15,$7C,$55
	.dc.b	$C2,$75,$7F,$D6,$31,$F8,$F9,$7C,$D8,$ED,$7C,$9F,$EA,$B5,$E3,$C3
	.dc.b	$A4,$7C,$C8,$F5,$61,$F0,$E3,$DF,$5F,$1F,$1F,$2E,$3D,$B2,$F5,$DF
	.dc.b	$F8,$9A,$FF,$B2,$C7,$B9,$FD,$49,$FE,$D6,$29,$FE,$D0,$C3,$FB,$13
	.dc.b	$FD,$7B,$BF,$FD,$9C,$DF,$07,$FB,$1B,$FE,$84,$3F,$8F,$1F,$D1,$EF
	.dc.b	$F9,$6B,$FC,$FA,$7F,$92,$E3,$FD,$60,$FE,$81,$1F,$C8,$CF,$EA,$87
	.dc.b	$F2,$CF,$0B,$FA,$34,$7F,$4A,$37,$5F,$CE,$E7,$BB,$FC,$EF,$3F,$F5
	.dc.b	$16,$1F,$D1,$1F,$F5,$5C,$6E,$F9,$E1,$FC,$BC,$FE,$AF,$3F,$CD,$A3
	.dc.b	$35,$C6,$FF,$A2,$3F,$CE,$2F,$97,$85,$FC,$E2,$35,$3E,$D7,$57,$E0
	.dc.b	$24,$72,$3E,$23,$E6,$7F,$AA,$BA,$79,$1A,$3F,$BC,$1F,$14,$FE,$C8
	.dc.b	$F9,$E7,$C5,$1C,$8E,$D5,$F0,$DE,$2F,$A1,$F9,$A7,$C8,$9E,$4E,$2F
	.dc.b	$E0,$BC,$CF,$1C,$F3,$7F,$BD,$3C,$DF,$19,$F5,$EB,$C7,$0A,$FC,$6E
	.dc.b	$3F,$ED,$9C,$BF,$AB,$5F,$FA,$FC,$FF,$5C,$9F,$D5,$B7,$FF,$62,$3F
	.dc.b	$D5,$FC,$0F,$CA,$C7,$FB,$07,$5F,$EB,$F1,$27,$88,$F0,$53,$FD,$53
	.dc.b	$BF,$FA,$8C,$FF,$AE,$F9,$F0,$FF,$53,$21,$FE,$C3,$8F,$F6,$47,$C2
	.dc.b	$7F,$56,$8F,$F5,$DF,$8F,$FE,$BA,$7F,$4F,$E1,$23,$11,$FD,$4A,$31
	.dc.b	$9F,$F5,$B8,$DD,$FF,$5B,$6F,$85,$FD,$6D,$FD,$7F,$07,$F7,$34,$D3
	.dc.b	$FA,$EF,$C2,$BE,$0B,$F6,$FD,$6F,$86,$FE,$B1,$39,$BE,$61,$C2,$9E
	.dc.b	$1D,$5D,$2F,$FB,$5D,$6A,$73,$5C,$67,$27,$82,$FB,$B8,$8F,$EE,$33
	.dc.b	$C2,$BB,$C7,$F7,$4C,$3B,$3A,$BE,$2F,$9A,$7C,$11,$88,$FE,$27,$3C
	.dc.b	$1F,$44,$65,$E8,$79,$4E,$98,$6D,$CD,$74,$8E,$5F,$03,$FD,$B2,$7F
	.dc.b	$C0,$91,$FE,$08,$C9,$F0,$EF,$88,$FE,$D3,$3F,$C9,$2F,$86,$DF,$E0
	.dc.b	$5B,$DD,$6B,$0F,$F0,$8D,$6A,$FA,$61,$CD,$FE,$F0,$73,$E2,$E1,$87
	.dc.b	$28,$CD,$FF,$45,$3F,$A1,$97,$4F,$E7,$1C,$FF,$DB,$E3,$27,$F6,$03
	.dc.b	$FA,$27,$6F,$F5,$B8,$FE,$E1,$18,$3F,$A2,$9F,$D9,$DF,$E8,$AF,$26
	.dc.b	$CF,$6E,$8F,$EA,$13,$FD,$4A,$BF,$BA,$3C,$8F,$EC,$EE,$27,$FA,$F3
	.dc.b	$FD,$3E,$FF,$AB,$CF,$F8,$0C,$9D,$DF,$D7,$EB,$B5,$71,$EA,$FE,$C1
	.dc.b	$1F,$DD,$CF,$EB,$87,$4A,$F3,$61,$FD,$A9,$F8,$07,$4E,$8F,$EB,$B5
	.dc.b	$FE,$F3,$8C,$E1,$DE,$31,$97,$B9,$B7,$CE,$31,$5F,$EE,$43,$A6,$5E
	.dc.b	$AB,$E5,$19,$6B,$FD,$AA,$35,$87,$8A,$34,$D7,$47,$CB,$3F,$C2,$27
	.dc.b	$EF,$B0,$FE,$4B,$87,$F6,$E8,$FE,$E3,$82,$31,$1F,$CA,$BB,$BF,$B8
	.dc.b	$BF,$ED,$12,$1F,$DC,$7B,$7F,$9C,$E7,$FD,$C2,$FF,$B4,$57,$F7,$03
	.dc.b	$FA,$FD,$7F,$37,$3F,$A0,$76,$FF,$38,$C7,$FA,$A5,$F1,$AF,$EC,$DD
	.dc.b	$DF,$EB,$48,$FE,$79,$7F,$EA,$53,$FA,$51,$CC,$FE,$86,$60,$98,$F9
	.dc.b	$2F,$A9,$23,$FA,$AB,$FD,$87,$1F,$3D,$FF,$6A,$9F,$EB,$73,$FD,$59
	.dc.b	$EB,$5C,$DE,$26,$8F,$15,$F9,$9F,$F5,$91,$89,$F8,$D8,$FB,$E3,$89
	.dc.b	$FD,$7A,$3B,$4F,$1C,$3F,$C2,$D8,$FC,$9C,$7E,$15,$72,$BF,$92,$43
	.dc.b	$CB,$87,$57,$FB,$3C,$7B,$6B,$E5,$C7,$F7,$BC,$3B,$47,$D4,$8F,$8F
	.dc.b	$5C,$70,$FF,$0F,$DF,$1C,$3F,$96,$3F,$83,$6D,$7F,$A8,$B1,$5F,$F2
	.dc.b	$AA,$FF,$1E,$9F,$E9,$8E,$4F,$E3,$5F,$13,$F1,$5B,$1F,$DB,$B6,$D9
	.dc.b	$70,$FF,$56,$C7,$18,$FE,$51,$5C,$5B,$1B,$9F,$F0,$04,$7F,$69,$E6
	.dc.b	$FE,$A3,$F1,$6F,$8B,$FF,$27,$C6,$1B,$D7,$FD,$B3,$BB,$FD,$71,$87
	.dc.b	$F3,$D1,$DD,$FD,$B1,$C6,$1C,$DF,$E8,$CD,$78,$46,$F0,$FE,$C2,$DB
	.dc.b	$FF,$2F,$D6,$E7,$C5,$5F,$D3,$EB,$4F,$F7,$BB,$F1,$CE,$A3,$C3,$87
	.dc.b	$A1,$D6,$9C,$6B,$78,$F6,$AD,$53,$1F,$04,$FB,$EF,$81,$FE,$E3,$BF
	.dc.b	$F0,$5D,$7C,$06,$E3,$F0,$63,$FD,$C6,$7F,$9D,$8F,$8B,$DD,$FE,$0D
	.dc.b	$7F,$C2,$4B,$A3,$FC,$1D,$1F,$DB,$0F,$F7,$41,$73,$FE,$CC,$FC,$F3
	.dc.b	$FB,$4E,$3F,$DF,$F4,$FE,$D9,$1F,$E1,$43,$EA,$77,$7F,$7F,$9F,$F0
	.dc.b	$14,$7F,$76,$9D,$F2,$FF,$BD,$3E,$33,$A9,$FE,$04,$9E,$4E,$2B,$8D
	.dc.b	$F3,$9F,$F0,$24,$78,$5F,$21,$E1,$C3,$CE,$73,$BF,$F0,$35,$71,$97
	.dc.b	$6F,$F7,$37,$E1,$76,$FF,$73,$7F,$C0,$27,$F7,$8E,$DF,$EE,$71,$9E
	.dc.b	$0F,$F7,$ED,$71,$D7,$FB,$B5,$F0,$3A,$DF,$0E,$EF,$F0,$14,$7F,$50
	.dc.b	$EC,$FE,$81,$D5,$FD,$DF,$4F,$EB,$F3,$FD,$01,$FE,$B8,$60,$FE,$BF
	.dc.b	$CD,$FD,$88,$FF,$5D,$76,$7F,$55,$27,$77,$F6,$DA,$FE,$95,$5C,$1F
	.dc.b	$F0,$35,$7F,$5C,$9D,$4F,$F6,$B8,$FF,$69,$9D,$67,$FC,$05,$5D,$1F
	.dc.b	$ED,$B5,$C2,$7C,$2D,$8D,$B5,$FF,$00,$9E,$B8,$E8,$E2,$BC,$8F,$C3
	.dc.b	$E3,$FE,$CA,$47,$D3,$3E,$47,$D2,$7F,$68,$8D,$4F,$A6,$3F,$C3,$4F
	.dc.b	$C8,$9F,$74,$7A,$09,$1F,$2E,$3A,$47,$84,$EE,$72,$7C,$F8,$7F,$89
	.dc.b	$A3,$FB,$F9,$A2,$CF,$85,$B3,$F1,$1E,$EF,$89,$FE,$EF,$87,$F7,$67
	.dc.b	$FD,$D4,$7F,$35,$3F,$DC,$C4,$3F,$BB,$9F,$DA,$7A,$FF,$AD,$3F,$D1
	.dc.b	$8F,$EB,$47,$F4,$23,$FA,$65,$7F,$46,$EC,$FE,$A3,$1C,$7B,$3F,$A3
	.dc.b	$76,$FF,$55,$8F,$E9,$98,$FF,$55,$7F,$A9,$9C,$9C,$5F,$38,$D9,$CC
	.dc.b	$D1,$4E,$24,$78,$1F,$EC,$12,$C7,$F5,$58,$FF,$64,$65,$FD,$E9,$FF
	.dc.b	$6D,$1E,$D3,$FB,$71,$FD,$CC,$ED,$3C,$9F,$83,$8F,$C4,$9F,$EC,$31
	.dc.b	$8C,$3D,$F8,$FF,$81,$63,$99,$F1,$CF,$36,$1E,$E2,$3D,$9A,$FF,$81
	.dc.b	$0E,$18,$78,$5E,$78,$78,$71,$5F,$F6,$E9,$4F,$F6,$E7,$17,$FD,$95
	.dc.b	$FE,$61,$DB,$FD,$95,$B1,$FD,$94,$D3,$FD,$98,$FF,$40,$F0,$A7,$83
	.dc.b	$67,$86,$BF,$DA,$F9,$FF,$A6,$5F,$F6,$0E,$DF,$E9,$BD,$7F,$D6,$23
	.dc.b	$FA,$47,$36,$58,$7F,$64,$EC,$FE,$AA,$FF,$58,$1D,$7F,$D7,$49,$87
	.dc.b	$F5,$58,$CB,$7A,$FF,$B1,$11,$C1,$20,$7F,$5F,$C7,$FB,$0F,$6F,$F6
	.dc.b	$08,$F0,$57,$F7,$4C,$3F,$AE,$93,$1F,$EC,$47,$FB,$A6,$7F,$AD,$EB
	.dc.b	$5F,$EB,$EF,$F5,$C7,$FD,$86,$7F,$63,$3F,$AB,$17,$8B,$FA,$2F,$7F
	.dc.b	$FB,$26,$3F,$A5,$67,$FD,$37,$1F,$E9,$51,$FD,$72,$75,$5C,$36,$FE
	.dc.b	$B7,$39,$6D,$7F,$52,$BF,$EB,$37,$C2,$B4,$7F,$58,$BE,$4D,$33,$5F
	.dc.b	$EF,$4F,$25,$E6,$7F,$B2,$FC,$A7,$A2,$F2,$FF,$A6,$D6,$B4,$FF,$63
	.dc.b	$D6,$A7,$37,$C6,$F8,$57,$FB,$33,$16,$1C,$5F,$EC,$38,$71,$9D,$75
	.dc.b	$7F,$A8,$A7,$FD,$2F,$9F,$F2,$1F,$13,$FD,$3F,$3F,$D6,$F9,$BF,$98
	.dc.b	$CF,$F2,$FA,$FE,$58,$7F,$3A,$38,$C7,$F2,$78,$D4,$7F,$2A,$3F,$A5
	.dc.b	$9F,$CA,$8F,$E8,$07,$F3,$1A,$EF,$D9,$FE,$D2,$9F,$17,$77,$F6,$C8
	.dc.b	$C5,$F5,$3C,$64,$BE,$75,$C3,$1F,$E9,$73,$B9,$CB,$5F,$39,$E4,$9F
	.dc.b	$3E,$1E,$2E,$EF,$F6,$B4,$FB,$72,$F8,$4F,$23,$85,$F8,$5E,$67,$F6
	.dc.b	$37,$FA,$9C,$7C,$D3,$8C,$69,$FE,$A5,$5D,$E3,$FB,$61,$DF,$1F,$7B
	.dc.b	$D8,$F0,$1D,$4F,$A2,$7F,$5F,$C3,$E0,$3F,$2F,$2F,$91,$8F,$AA,$18
	.dc.b	$7F,$B3,$63,$F5,$0D,$8F,$F2,$63,$5F,$EB,$B1,$AE,$DF,$F6,$8C,$7F
	.dc.b	$5D,$3F,$AD,$9F,$CA,$FB,$3F,$94,$76,$FF,$35,$BF,$E8,$58,$FF,$AC
	.dc.b	$E7,$35,$FD,$12,$7F,$9E,$BF,$CC,$3B,$BF,$9E,$B9,$C3,$FA,$E5,$FF
	.dc.b	$4F,$D7,$F9,$CD,$FF,$AD,$31,$ED,$5F,$EA,$CA,$D5,$FF,$3F,$21,$E5
	.dc.b	$C3,$AC,$78,$DB,$87,$FB,$00,$EB,$5F,$EB,$3C,$7C,$F5,$BB,$FE,$CA
	.dc.b	$73,$2E,$3E,$37,$D1,$8F,$90,$FE,$91,$1E,$E3,$E1,$E1,$DE,$7C,$B1
	.dc.b	$C6,$BB,$47,$F5,$A3,$38,$78,$F1,$F3,$C7,$F6,$93,$C4,$6C,$B3,$FD
	.dc.b	$6E,$BB,$C6,$DE,$D7,$FC,$DA,$3F,$AD,$D7,$FB,$1D,$FE,$73,$E7,$3C
	.dc.b	$5F,$F5,$E7,$5B,$C2,$A7,$FA,$FE,$75,$5F,$C3,$DA,$FF,$99,$E7,$83
	.dc.b	$FC,$4D,$AF,$F9,$AE,$37,$D8,$E3,$FE,$95,$3F,$D2,$4F,$EB,$3D,$3F
	.dc.b	$D1,$E7,$FD,$4F,$1F,$EB,$23,$3D,$7C,$3C,$0F,$EC,$5D,$DF,$D3,$E7
	.dc.b	$91,$FD,$86,$B8,$DF,$F5,$98,$FE,$AF,$3B,$AF,$ED,$47,$2E,$FF,$EB
	.dc.b	$27,$F6,$EC,$FF,$B4,$D7,$F6,$E3,$FD,$65,$3F,$EB,$57,$FB,$41,$F3
	.dc.b	$4D,$1F,$EB,$82,$1F,$45,$F9,$DD,$FF,$D7,$63,$FA,$D9,$FE,$BD,$E2
	.dc.b	$FE,$BC,$FF,$58,$3F,$B6,$F8,$BF,$E9,$F7,$FD,$6E,$63,$C0,$FE,$A5
	.dc.b	$9F,$F5,$58,$FF,$55,$99,$DB,$FA,$CE,$DF,$D8,$4F,$EB,$2E,$23,$47
	.dc.b	$F5,$79,$E4,$FF,$B1,$E7,$83,$D0,$FF,$60,$1F,$EF,$7E,$DF,$E9,$DB
	.dc.b	$FD,$B8,$76,$FF,$4F,$7F,$D5,$53,$AC,$3F,$C0,$F3,$9A,$E1,$5F,$D6
	.dc.b	$3B,$FF,$AE,$78,$1F,$D7,$AF,$FB,$0F,$81,$FD,$7A,$BF,$D6,$9D,$3F
	.dc.b	$CB,$7B,$FF,$93,$C7,$F2,$53,$FA,$81,$7C,$0F,$E5,$DC,$3F,$E9,$6A
	.dc.b	$E1,$DF,$FC,$95,$B5,$FC,$C8,$FE,$59,$D9,$FD,$1E,$3F,$9C,$93,$0E
	.dc.b	$EF,$F3,$AF,$03,$F9,$7D,$71,$8E,$78,$77,$AF,$F5,$83,$FD,$3A,$B6
	.dc.b	$E7,$0F,$F5,$75,$73,$CF,$FA,$CE,$1F,$D0,$4F,$E9,$18,$2B,$F8,$6B
	.dc.b	$D3,$1F,$EA,$57,$E5,$9F,$64,$E6,$BF,$A4,$CF,$57,$FA,$35,$F6,$7E
	.dc.b	$36,$3E,$4C,$3F,$AC,$DF,$19,$F0,$E1,$E0,$3C,$F1,$E3,$86,$3F,$EB
	.dc.b	$77,$BC,$66,$31,$1C,$8D,$D7,$FA,$D2,$74,$DA,$ED,$3F,$E0,$73,$85
	.dc.b	$72,$C7,$DD,$1F,$3E,$FF,$D6,$4F,$F5,$E6,$C3,$BB,$FA,$CC,$FF,$95
	.dc.b	$A3,$1D,$FF,$D7,$49,$1F,$EA,$43,$F9,$5E,$3F,$D1,$23,$FD,$63,$18
	.dc.b	$3F,$9A,$B8,$9F,$E5,$B1,$FC,$BE,$BF,$9C,$3F,$D1,$9A,$FF,$37,$8D
	.dc.b	$BF,$E9,$2A,$EC,$F8,$63,$3E,$07,$F5,$29,$FE,$91,$3E,$12,$4E,$BC
	.dc.b	$0F,$E8,$B1,$CC,$F2,$9C,$EF,$C3,$97,$F4,$57,$FB,$45,$FF,$AE,$63
	.dc.b	$FD,$45,$87,$7A,$F5,$63,$ED,$BE,$37,$E6,$BF,$ED,$F5,$E2,$9C,$BE
	.dc.b	$08,$CE,$3E,$58,$FE,$C7,$3F,$EA,$C7,$F0,$27,$D3,$71,$3E,$2C,$7F
	.dc.b	$AA,$4B,$BB,$FA,$3E,$DF,$D6,$7B,$3F,$A5,$F2,$7F,$08,$6C,$7F,$5D
	.dc.b	$9D,$D7,$F5,$CF,$03,$FA,$CB,$FD,$4F,$C0,$FE,$B5,$1F,$EA,$D8,$FE
	.dc.b	$87,$DF,$FD,$01,$FE,$8F,$D5,$FD,$53,$C0,$FE,$98,$7F,$47,$CB,$FA
	.dc.b	$1F,$81,$FD,$1E,$BF,$A9,$9F,$E9,$EA,$E1,$7C,$E3,$FA,$91,$3B,$7F
	.dc.b	$A9,$3F,$D5,$9F,$EB,$CE,$8F,$E9,$55,$FD,$35,$FF,$5C,$18,$B6,$BF
	.dc.b	$EA,$59,$FE,$97,$3F,$D5,$32,$78,$1F,$D3,$63,$FD,$4A,$7F,$4D,$3F
	.dc.b	$D5,$06,$0F,$EB,$1C,$18,$7F,$A8,$1F,$CF,$B1,$FF,$53,$9F,$D1,$1F
	.dc.b	$EA,$26,$73,$FE,$96,$FF,$49,$7F,$A3,$BF,$D2,$4F,$E9,$B1,$CA,$33
	.dc.b	$1C,$BC,$0F,$EB,$8F,$81,$B5,$FE,$AA,$AF,$F5,$47,$6F,$F4,$4E,$1F
	.dc.b	$F4,$F7,$7F,$F4,$BC,$3F,$A7,$F6,$FF,$45,$DB,$FA,$5C,$E7,$BF,$FA
	.dc.b	$75,$FF,$88,$63,$15,$B8,$FE,$21,$7F,$94,$3F,$A0,$9F,$CC,$79,$FF
	.dc.b	$87,$E7,$FC,$DE,$3F,$8A,$46,$3A,$3F,$93,$F6,$7F,$24,$3F,$8C,$D7
	.dc.b	$F3,$FB,$FE,$5D,$5F,$CC,$63,$FC,$EE,$4A,$FE,$9D,$D9,$FD,$30,$FE
	.dc.b	$47,$87,$FA,$0F,$0F,$E7,$67,$57,$CD,$7F,$D3,$AB,$C3,$1E,$4A,$E5
	.dc.b	$5E,$29,$FE,$8B,$18,$9E,$2F,$F3,$DA,$F5,$47,$FA,$31,$E8,$D8,$D1
	.dc.b	$C6,$FF,$99,$36,$BE,$4C,$62,$FC,$F5,$E3,$6C,$7C,$38,$FE,$A2,$47
	.dc.b	$E1,$3F,$3E,$BA,$3F,$04,$FE,$7D,$3F,$06,$7B,$46,$0E,$C5,$7B,$1D
	.dc.b	$1F,$EB,$B8,$7C,$37,$FB,$67,$73,$BB,$FC,$31,$4C,$7F,$CC,$D9,$7F
	.dc.b	$A9,$63,$17,$FD,$27,$2F,$E9,$A7,$F4,$EE,$CF,$E9,$B1,$FE,$AE,$6D
	.dc.b	$7F,$21,$8F,$E8,$11,$FC,$99,$FE,$51,$DB,$FC,$B4,$8F,$F3,$76,$DF
	.dc.b	$F2,$A9,$E1,$5C,$E3,$F9,$D4,$FF,$2C,$3F,$D2,$15,$F8,$4A,$FF,$54
	.dc.b	$E1,$FC,$F4,$F5,$B8,$8F,$E6,$13,$C5,$FF,$4C,$92,$BF,$D2,$67,$67
	.dc.b	$A4,$76,$8F,$E6,$D1,$FD,$76,$3C,$F5,$DE,$3A,$9C,$DF,$EB,$37,$E9
	.dc.b	$7F,$A5,$1B,$8F,$91,$3D,$63,$CE,$D8,$E7,$5E,$49,$E3,$18,$76,$F7
	.dc.b	$3F,$AC,$9F,$2C,$91,$F2,$E7,$CF,$1C,$48,$7F,$69,$3E,$A9,$C2,$7C
	.dc.b	$D5,$FE,$BB,$BF,$ED,$33,$C2,$31,$5F,$2E,$FC,$D3,$FD,$8E,$F8,$DF
	.dc.b	$D1,$8F,$F7,$00,$3F,$88,$76,$7F,$58,$5D,$3F,$D7,$1D,$35,$FE,$C9
	.dc.b	$D1,$FD,$69,$FE,$B8,$43,$FD,$5C,$FF,$5D,$7F,$A7,$73,$FF,$B0,$7A
	.dc.b	$FF,$AD,$72,$FF,$5E,$9F,$EC,$27,$F5,$A3,$FD,$3D,$CF,$FD,$50,$FE
	.dc.b	$AF,$3C,$23,$FA,$D5,$73,$EB,$FE,$BD,$5B,$8F,$EC,$33,$CF,$AF,$FD
	.dc.b	$5B,$97,$F5,$88,$FF,$54,$B5,$FE,$B9,$8F,$FA,$A5,$FE,$BE,$7F,$A9
	.dc.b	$F2,$FE,$B6,$FF,$5D,$EB,$FE,$C1,$18,$8E,$07,$F5,$B3,$FD,$BD,$BF
	.dc.b	$FA,$A8,$FE,$95,$8F,$F5,$6C,$BF,$A9,$4F,$F5,$48,$C1,$FD,$97,$C1
	.dc.b	$3F,$EE,$7E,$C7,$83,$FE,$9D,$9C,$D7,$F6,$80,$E1,$FF,$52,$BD,$9F
	.dc.b	$EC,$C7,$FA,$A6,$7F,$AE,$AB,$FE,$D8,$FF,$61,$AF,$F5,$54,$7F,$60
	.dc.b	$E8,$FF,$50,$F8,$3F,$EA,$3A,$FF,$00,$C7,$FA,$AB,$AF,$FA,$D4,$7E
	.dc.b	$2F,$B3,$FA,$ED,$6F,$7F,$F5,$A7,$6F,$F6,$2A,$FE,$59,$B7,$F5,$B3
	.dc.b	$F9,$3F,$3F,$F5,$2E,$DF,$E6,$EF,$F4,$86,$CF,$F9,$F1,$FE,$54,$FF
	.dc.b	$9D,$23,$FA,$09,$1E,$AF,$F5,$49,$EB,$1F,$CF,$67,$9B,$C4,$D9,$FD
	.dc.b	$3D,$FE,$7D,$3C,$E3,$FB,$99,$FC,$E8,$FE,$90,$F9,$DF,$E8,$0F,$9D
	.dc.b	$A7,$9C,$FE,$BF,$1D,$1E,$E7,$FB,$1B,$AB,$FB,$A4,$F3,$AF,$88,$F3
	.dc.b	$6C,$F3,$7F,$A8,$98,$7D,$47,$9A,$79,$BE,$F3,$C6,$69,$FE,$DA,$F8
	.dc.b	$7A,$FF,$BF,$BC,$5B,$1A,$ED,$FF,$6C,$DF,$CB,$3C,$85,$AF,$23,$F1
	.dc.b	$E5,$5F,$D5,$B0,$FD,$2C,$61,$FF,$4F,$D7,$F1,$13,$F9,$B9,$4F,$E6
	.dc.b	$67,$EC,$E3,$FD,$4E,$DE,$BF,$EA,$64,$EC,$FE,$57,$1F,$C7,$0F,$EB
	.dc.b	$EE,$27,$8E,$5F,$D2,$27,$27,$F2,$D3,$F9,$55,$7F,$25,$9D,$E3,$FC
	.dc.b	$F2,$BF,$95,$D7,$F3,$47,$F9,$D3,$C4,$F0,$3F,$D5,$CE,$C5,$8E,$47
	.dc.b	$F5,$2A,$F4,$4F,$1A,$EE,$FF,$42,$F0,$7F,$D4,$4F,$F4,$38,$FE,$B0
	.dc.b	$FF,$3B,$3C,$F8,$7F,$55,$7E,$3E,$1E,$69,$DC,$79,$F0,$FE,$83,$5B
	.dc.b	$9F,$E8,$F3,$FD,$B0,$FE,$B0,$77,$C3,$CB,$7E,$E2,$5F,$29,$DE,$1E
	.dc.b	$62,$78,$3F,$EC,$26,$C7,$F7,$53,$DD,$7F,$D6,$1F,$3D,$7C,$47,$FA
	.dc.b	$94,$AB,$FD,$63,$DD,$FD,$4F,$AF,$FA,$81,$FD,$45,$FE,$AC,$67,$C1
	.dc.b	$FF,$56,$DF,$F5,$03,$FA,$E7,$0B,$97,$FA,$1C,$7F,$1F,$EA,$FE,$90
	.dc.b	$FF,$51,$3F,$A7,$B8,$74,$7F,$54,$3F,$A7,$9F,$D9,$FB,$7F,$B2,$D7
	.dc.b	$F6,$9E,$BF,$E6,$1D,$EE,$FF,$CE,$46,$1B,$87,$E7,$EF,$8E,$1F,$CE
	.dc.b	$23,$CB,$D7,$FE,$B4,$9F,$EA,$F8,$FF,$63,$C7,$FB,$04,$FF,$AD,$AF
	.dc.b	$FA,$B9,$FD,$6A,$7F,$A8,$D7,$FA,$DE,$0F,$FA,$D1,$FF,$5D,$F7,$7F
	.dc.b	$57,$7F,$AE,$77,$FF,$AD,$79,$3F,$AB,$47,$FA,$D1,$FE,$AE,$FF,$47
	.dc.b	$EE,$FE,$87,$9F,$F5,$68,$C6,$DF,$E0,$9C,$FF,$A7,$77,$FF,$4C,$7F
	.dc.b	$A5,$E1,$CA,$BF,$D6,$47,$F5,$22,$CE,$EF,$FD,$67,$1F,$D4,$6F,$A4
	.dc.b	$77,$8D,$E1,$E2,$7C,$27,$D6,$C5,$F1,$BF,$EB,$36,$F8,$5F,$81,$9F
	.dc.b	$EA,$E7,$F8,$7F,$E3,$BF,$C4,$18,$FE,$18,$FF,$11,$57,$07,$13,$FE
	.dc.b	$DA,$BF,$F5,$F7,$CC,$3E,$41,$F0,$7D,$78,$BB,$63,$C7,$0E,$F7,$BA
	.dc.b	$D6,$3F,$E2,$FF,$09,$F2,$2F,$92,$CE,$0C,$3F,$31,$E1,$E2,$F7,$5A
	.dc.b	$9C,$E1,$E6,$9C,$BE,$88,$CD,$FC,$09,$CB,$E4,$9F,$E6,$A7,$1F,$45
	.dc.b	$F2,$7D,$DB,$1F,$0B,$E5,$5F,$2A,$F4,$F5,$72,$BC,$1F,$2C,$F4,$F1
	.dc.b	$97,$C2,$8C,$4F,$7A,$F5,$46,$27,$CD,$97,$81,$E1,$7F,$07,$E6,$3F
	.dc.b	$59,$87,$BD,$B3,$F7,$3E,$61,$EC,$7F,$BB,$E7,$3C,$BF,$E1,$06,$DF
	.dc.b	$F9,$7F,$81,$6F,$99,$ED,$E1,$FC,$D8,$C5,$77,$8C,$4E,$A3,$E7,$4E
	.dc.b	$B2,$F9,$DD,$2F,$9B,$FB,$0D,$AF,$C5,$D3,$E6,$DF,$45,$AB,$95,$F6
	.dc.b	$7C,$77,$C4,$E3,$2F,$C5,$C6,$27,$F0,$19,$75,$9C,$D6,$F7,$FD,$06
	.dc.b	$DF,$E4,$48,$FC,$57,$CF,$FF,$93,$30,$FC,$43,$6F,$F4,$BD,$0F,$9E
	.dc.b	$FF,$40,$D6,$BC,$4E,$27,$E4,$7E,$7F,$FC,$A7,$19,$3F,$20,$E2,$FF
	.dc.b	$21,$87,$5C,$FC,$73,$97,$FC,$87,$19,$AF,$A9,$39,$7F,$4D,$BB,$E8
	.dc.b	$BF,$44,$D8,$FC,$A7,$D1,$3E,$8B,$35,$27,$F4,$5F,$47,$F9,$77,$F4
	.dc.b	$D7,$C7,$3F,$DA,$CE,$5F,$D3,$46,$73,$F9,$33,$97,$FC,$B3,$19,$AF
	.dc.b	$C5,$4E,$5F,$F0,$D7,$B4,$F8,$EF,$DD,$E0,$8C,$4F,$F0,$C8,$C7,$26
	.dc.b	$A7,$31,$FC,$58,$E1,$59,$F5,$F4,$DB,$E5,$49,$EF,$F4,$96,$73,$18
	.dc.b	$3B,$7C,$37,$7F,$26,$CF,$27,$0F,$FB,$0F,$D0,$E9,$97,$FB,$17,$17
	.dc.b	$D2,$FF,$2B,$9C,$CF,$B6,$F1,$8F,$C2,$AB,$3C,$5A,$FC,$3F,$A6,$B8
	.dc.b	$74,$C7,$C1,$7F,$2A,$73,$F4,$F2,$73,$6F,$39,$EF,$B5,$D4,$B5,$F2
	.dc.b	$E6,$C7,$BF,$C2,$6A,$72,$DD,$3E,$CB,$6F,$E0,$35,$FB,$4D,$B3,$49
	.dc.b	$9F,$9A,$7F,$01,$87,$82,$31,$5E,$36,$D7,$0A,$F4,$5F,$E2,$AF,$84
	.dc.b	$60,$FA,$4E,$30,$F6,$B6,$7D,$37,$D0,$EF,$19,$AF,$5C,$65,$E8,$79
	.dc.b	$9C,$57,$C4,$6D,$78,$DB,$3C,$8E,$38,$7C,$28,$EE,$6A,$3E,$3B,$63
	.dc.b	$E0,$47,$0C,$3C,$8E,$8E,$78,$62,$7A,$46,$EB,$AB,$49,$87,$E4,$BE
	.dc.b	$03,$FD,$19,$18,$C7,$FD,$19,$18,$9F,$64,$FF,$AC,$8D,$01,$1B,$9F
	.dc.b	$97,$1B,$9E,$27,$7F,$43,$FD,$4F,$87,$9A,$31,$F0,$3F,$07,$0E,$57
	.dc.b	$F1,$E7,$2F,$80,$EF,$7C,$B3,$FD,$2D,$70,$9F,$89,$5C,$0F,$0E,$3F
	.dc.b	$EB,$08,$FF,$6B,$1E,$57,$FD,$64,$FF,$AF,$6F,$FD,$72,$75,$CB,$FD
	.dc.b	$CF,$7F,$EB,$BC,$3B,$CE,$B1,$FF,$6A,$5F,$FB,$05,$DC,$71,$8F,$F6
	.dc.b	$F9,$FE,$CB,$C7,$FD,$A1,$7E,$7C,$BF,$D8,$B3,$E4,$9F,$F7,$7B,$FE
	.dc.b	$E7,$7D,$59,$7F,$B1,$6F,$C1,$3F,$ED,$4B,$E5,$8F,$93,$1F,$09,$FE
	.dc.b	$F6,$8C,$E5,$8A,$F8,$07,$7C,$3E,$FE,$1E,$7C,$7B,$CF,$DD,$BF,$C2
	.dc.b	$BF,$EC,$D6,$BC,$30,$FF,$6A,$D1,$FF,$6A,$C7,$E2,$BE,$07,$F8,$7B
	.dc.b	$FE,$D1,$DB,$2F,$FA,$CC,$FF,$60,$E3,$FE,$9B,$69,$9C,$3F,$D7,$AF
	.dc.b	$FA,$80,$FF,$4F,$FC,$07,$FA,$9C,$8F,$FA,$9F,$2F,$F5,$7B,$FE,$BE
	.dc.b	$3F,$D5,$F1,$A7,$FD,$B3,$1A,$7B,$9F,$EB,$F3,$83,$FE,$D8,$F8,$1F
	.dc.b	$F5,$64,$7F,$15,$AF,$19,$E4,$8C,$C7,$F0,$46,$CF,$FB,$66,$3C,$E7
	.dc.b	$FB,$3E,$33,$7F,$EF,$96,$C7,$82,$31,$7F,$EF,$DA,$FF,$6A,$63,$F2
	.dc.b	$9F,$F7,$2C,$EA,$BC,$93,$98,$F9,$57,$D2,$4D,$C3,$D3,$5B,$9F,$29
	.dc.b	$E9,$C3,$C1,$7D,$63,$07,$F3,$5B,$FF,$38,$36,$76,$F2,$C3,$E4,$7A
	.dc.b	$98,$8F,$0C,$DA,$F2,$12,$71,$12,$7A,$1D,$E3,$0F,$88,$95,$FD,$AA
	.dc.b	$BC,$71,$8B,$F4,$4F,$8A,$39,$99,$3C,$D1,$F4,$6B,$C8,$76,$9E,$13
	.dc.b	$F6,$E3,$13,$D0,$E1,$1E,$53,$AC,$F0,$3C,$93,$C2,$3B,$9F,$EF,$67
	.dc.b	$BB,$C2,$7A,$39,$3A,$1E,$73,$E4,$98,$9E,$AF,$13,$A9,$F7,$A7,$AB
	.dc.b	$E9,$38,$4F,$FA,$DB,$EA,$1F,$51,$F4,$FE,$13,$EA,$7C,$7C,$BF,$BD
	.dc.b	$1F,$67,$E9,$1F,$39,$B9,$C9,$FD,$DE,$32,$DF,$0F,$FA,$23,$FE,$33
	.dc.b	$8C,$B6,$73,$E1,$7F,$B3,$61,$3F,$E5,$37,$FC,$7C,$BE,$83,$7C,$DD
	.dc.b	$3E,$A5,$F4,$F8,$BF,$EA,$51,$87,$8D,$65,$EB,$E5,$5C,$3A,$46,$8E
	.dc.b	$F1,$92,$38,$3C,$77,$DF,$DD,$FF,$48,$FC,$47,$87,$D0,$79,$1F,$EA
	.dc.b	$AB,$FF,$34,$FB,$1B,$F2,$97,$4E,$15,$BA,$FF,$35,$D6,$5F,$F3,$7C
	.dc.b	$D7,$FC,$DA,$FB,$FD,$07,$9B,$F3,$38,$7F,$DE,$D9,$BE,$AB,$FD,$9D
	.dc.b	$87,$FB,$6F,$E1,$BF,$C5,$ED,$F1,$3D,$53,$A8,$FF,$6F,$36,$FF,$D4
	.dc.b	$35,$C2,$7F,$DE,$6D,$8E,$F3,$BE,$4F,$F7,$A3,$69,$87,$8F,$DF,$EE
	.dc.b	$7A,$7D,$3F,$F7,$3C,$FF,$AC,$8F,$95,$E8,$3E,$AE,$9F,$3F,$D6,$FC
	.dc.b	$5C,$FF,$B1,$8C,$9F,$EB,$F8,$D3,$0E,$84,$3F,$DB,$DF,$58,$FA,$9F
	.dc.b	$B5,$F5,$BF,$EA,$59,$FF,$69,$FD,$1F,$DB,$F7,$4E,$BD,$89,$3A,$20
	.dc.b	$E1,$FA,$37,$CE,$31,$C3,$FE,$25,$E0,$4E,$6B,$77,$D2,$FF,$DE,$5F
	.dc.b	$5E,$FA,$FF,$C7,$0F,$B0,$FC,$77,$C3,$7E,$9F,$0F,$F6,$AB,$73,$ED
	.dc.b	$5B,$C3,$FC,$69,$5F,$84,$8C,$47,$90,$77,$2F,$94,$61,$BA,$78,$F1
	.dc.b	$F1,$92,$7A,$E3,$E0,$8F,$CA,$7C,$47,$E1,$63,$E0,$7C,$4F,$FA,$EE
	.dc.b	$B6,$78,$3E,$6D,$F6,$3F,$A9,$6C,$6B,$15,$71,$8F,$44,$63,$E0,$38
	.dc.b	$5C,$9E,$15,$BD,$98,$FF,$95,$7D,$C7,$D9,$76,$BF,$85,$BF,$E5,$BE
	.dc.b	$6B,$F7,$5A,$1F,$E1,$97,$C3,$3B,$87,$1C,$BA,$E3,$FE,$AB,$8C,$65
	.dc.b	$DD,$B1,$FE,$65,$FB,$37,$CC,$7F,$13,$F9,$6B,$87,$F9,$A7,$C2,$79
	.dc.b	$7F,$CA,$3E,$5D,$F1,$BF,$CA,$A7,$39,$7F,$9B,$E7,$2F,$1A,$FE,$5F
	.dc.b	$87,$48,$57,$05,$19,$8C,$57,$29,$DD,$78,$6F,$C0,$F4,$C7,$F3,$D3
	.dc.b	$A3,$FD,$01,$87,$39,$E1,$1E,$33,$CD,$5C,$9D,$D7,$2C,$3C,$97,$C6
	.dc.b	$76,$71,$18,$79,$67,$75,$C7,$0F,$26,$3B,$AD,$47,$8A,$FC,$C7,$C0
	.dc.b	$AE,$0F,$8A,$B8,$63,$E5,$8E,$ED,$C7,$C9,$1C,$67,$6C,$3A,$C6,$8F
	.dc.b	$A0,$FF,$16,$6B,$E6,$3F,$8C,$9C,$CA,$BE,$5F,$FD,$E9,$18,$6B,$CD
	.dc.b	$AF,$DE,$6B,$FC,$93,$DA,$EA,$FF,$91,$CA,$7F,$BE,$CA,$E1,$C6,$7F
	.dc.b	$85,$F9,$77,$CB,$FE,$72,$3F,$11,$3F,$EB,$DA,$FA,$95,$AE,$0F,$AD
	.dc.b	$39,$3E,$7C,$65,$FA,$26,$6B,$B5,$F5,$D7,$E1,$E1,$FE,$C9,$F1,$B3
	.dc.b	$8D,$9C,$E0,$9C,$B6,$33,$86,$0F,$3C,$7B,$23,$FC,$23,$8F,$8E,$73
	.dc.b	$87,$AB,$1F,$15,$FC,$93,$D2,$E1,$F9,$53,$D6,$FB,$9D,$27,$53,$ED
	.dc.b	$8C,$63,$F1,$23,$C4,$7F,$B9,$70,$F1,$4F,$F1,$3F,$19,$F2,$FF,$C2
	.dc.b	$B6,$FE,$17,$CA,$9C,$E3,$FC,$A3,$17,$CB,$FF,$BB,$E3,$F9,$5D,$6F
	.dc.b	$0E,$51,$D5,$FF,$16,$55,$8F,$C8,$CE,$DB,$7D,$A7,$FB,$6F,$47,$F3
	.dc.b	$0F,$7B,$8C,$7F,$7A,$EB,$46,$27,$57,$F9,$C6,$D7,$2A,$DC,$FF,$40
	.dc.b	$EB,$FE,$67,$F0,$BE,$6A,$E9,$85,$9C,$D7,$F4,$1F,$95,$7C,$B6,$F2
	.dc.b	$46,$27,$FA,$5F,$A4,$F0,$9E,$7E,$EB,$59,$FF,$48,$6B,$FE,$54,$C3
	.dc.b	$9E,$3E,$B9,$CC,$FF,$77,$6B,$FE,$1A,$79,$F1,$7F,$70,$F9,$6E,$D5
	.dc.b	$E9,$F1,$91,$E7,$F9,$6F,$E7,$3A,$FF,$5B,$6E,$3F,$DC,$7C,$D7,$AD
	.dc.b	$F4,$72,$FF,$1C,$D6,$9B,$E0,$7C,$F8,$E3,$7E,$96,$BF,$E0,$D6,$BF
	.dc.b	$E3,$07,$96,$08,$FE,$FF,$83,$E5,$BC,$D7,$F4,$63,$15,$FD,$C7,$99
	.dc.b	$F2,$BF,$45,$B1,$FD,$EA,$5F,$2C,$F6,$F5,$72,$BF,$09,$7C,$32,$FC
	.dc.b	$6B,$5F,$A4,$D7,$FC,$4C,$EF,$4F,$ED,$BC,$DF,$C0,$7E,$5B,$A5,$7F
	.dc.b	$78,$F0,$D1,$F8,$5F,$96,$FE,$D3,$8F,$F8,$25,$B5,$FE,$1F,$E9,$7B
	.dc.b	$1F,$DC,$AB,$5E,$0F,$13,$F3,$97,$C2,$FF,$0C,$D7,$FC,$74,$D7,$FC
	.dc.b	$BA,$F6,$CF,$CF,$F2,$DF,$C8,$5F,$F3,$9C,$66,$FE,$96,$8A,$D4,$7F
	.dc.b	$91,$77,$7A,$9F,$9F,$6C,$7E,$7B,$DC,$7C,$26,$6A,$4F,$EC,$6B,$79
	.dc.b	$FE,$E9,$AF,$E7,$DA,$F7,$9F,$D8,$7B,$FF,$A8,$AF,$F1,$85,$72,$D3
	.dc.b	$E8,$CE,$6F,$FC,$90,$3E,$41,$EA,$7E,$EE,$30,$D8,$FE,$71,$87,$F1
	.dc.b	$97,$6D,$8F,$E1,$B3,$C2,$3F,$AC,$CE,$A3,$FB,$3B,$5F,$F4,$0A,$F8
	.dc.b	$AE,$2D,$9E,$31,$83,$3F,$25,$23,$AB,$6B,$AF,$CA,$F2,$24,$F8,$9E
	.dc.b	$BF,$2B,$FE,$3E,$8F,$0B,$67,$FD,$BB,$82,$31,$1F,$AF,$8C,$4F,$F3
	.dc.b	$78,$DF,$CA,$7F,$92,$3D,$1F,$54,$6E,$31,$97,$F9,$5F,$E4,$D9,$E2
	.dc.b	$B5,$87,$F7,$3A,$E6,$F5,$8C,$9F,$D0,$31,$61,$E4,$8F,$B0,$D8,$CC
	.dc.b	$F3,$9F,$47,$C2,$F7,$77,$19,$7F,$01,$19,$9E,$30,$CB,$D6,$EA,$B8
	.dc.b	$4E,$A7,$2A,$B9,$5F,$36,$D7,$06,$DF,$5C,$3E,$7C,$63,$0E,$6D,$8C
	.dc.b	$D7,$82,$FA,$34,$E8,$5B,$E8,$7C,$07,$17,$D1,$B5,$D2,$76,$7B,$4B
	.dc.b	$87,$AE,$77,$1C,$CC,$E1,$E2,$77,$85,$C3,$B4,$6A,$B8,$BD,$63,$24
	.dc.b	$C3,$EF,$C6,$C9,$19,$7F,$02,$D7,$94,$7A,$63,$E8,$D7,$86,$3D,$95
	.dc.b	$FA,$CB,$FC,$21,$1F,$C6,$5F,$23,$FA,$57,$C9,$BC,$D7,$CA,$7C,$C9
	.dc.b	$F8,$B3,$98,$C6,$5F,$84,$8C,$4F,$A3,$1F,$EA,$8A,$33,$3E,$96,$CF
	.dc.b	$A4,$DB,$89,$F9,$D3,$D6,$FC,$71,$D4,$F3,$B8,$8E,$8D,$AE,$93,$AB
	.dc.b	$F3,$C7,$69,$D4,$F6,$8C,$4F,$E8,$EF,$C5,$3E,$9A,$DB,$F1,$CC,$4E
	.dc.b	$E3,$33,$C3,$3F,$C4,$57,$C8,$9C,$92,$BC,$71,$E5,$34,$F3,$32,$F6
	.dc.b	$32,$4C,$3D,$F1,$B8,$C6,$1F,$01,$E1,$3F,$26,$74,$F0,$8D,$9E,$17
	.dc.b	$06,$DC,$9E,$FC,$3E,$19,$2F,$E1,$E1,$EE,$AE,$04,$C3,$E5,$DF,$D7
	.dc.b	$3C,$6E,$23,$51,$F4,$4F,$14,$64,$8E,$4E,$26,$88,$79,$6B,$E6,$3E
	.dc.b	$8A,$F2,$4F,$28,$E2,$48,$CB,$CA,$FD,$98,$7D,$06,$C7,$3C,$BF,$94
	.dc.b	$46,$31,$F4,$CF,$86,$BE,$6E,$1D,$F0,$EB,$3E,$0B,$E1,$3F,$55,$B1
	.dc.b	$EA,$9D,$C6,$23,$CD,$87,$83,$1F,$6E,$3E,$0B,$DC,$63,$0F,$05,$6F
	.dc.b	$1F,$1E,$1D,$63,$0F,$DC,$C7,$F2,$D1,$8C,$7D,$71,$8C,$B8,$D7,$A7
	.dc.b	$1F,$BF,$7B,$C7,$E6,$5E,$EB,$A5,$7E,$D9,$FE,$02,$7E,$64,$B1,$FC
	.dc.b	$19,$B1,$B7,$F8,$0F,$CA,$36,$FF,$6F,$D2,$31,$3F,$EE,$E8,$C4,$EA
	.dc.b	$73,$3D,$9D,$E5,$FC,$55,$B8,$76,$AF,$EC,$1F,$26,$F9,$4F,$E2,$F7
	.dc.b	$B9,$FE,$05,$7C,$30,$FE,$4A,$D5,$7C,$5D,$CC,$7F,$93,$CE,$E7,$51
	.dc.b	$89,$D5,$77,$8F,$EA,$33,$A8,$C1,$30,$F4,$4E,$B0,$F3,$C6,$27,$53
	.dc.b	$E6,$BF,$6D,$6A,$30,$78,$49,$3F,$DF,$2F,$C1,$7C,$C9,$3C,$27,$55
	.dc.b	$CE,$3F,$9E,$61,$F0,$F1,$FA,$71,$D2,$FF,$A0,$E1,$D2,$39,$47,$47
	.dc.b	$8E,$3E,$67,$E1,$C6,$6B,$9E,$3E,$09,$C9,$C8,$87,$D2,$C7,$C9,$3C
	.dc.b	$2B,$96,$3F,$46,$72,$F9,$C9,$87,$86,$FD,$07,$4B,$E4,$FF,$AB,$71
	.dc.b	$F5,$3E,$83,$65,$C7,$CC,$E2,$78,$B9,$C7,$E1,$E3,$E2,$C3,$D7,$39
	.dc.b	$3B,$46,$5E,$58,$FA,$5D,$DF,$43,$AE,$3E,$47,$13,$EC,$9F,$C3,$1C
	.dc.b	$1C,$63,$E6,$3F,$0F,$1E,$12,$57,$39,$DE,$3F,$5E,$31,$8F,$33,$E4
	.dc.b	$5F,$2C,$3C,$31,$8A,$FA,$2F,$BA,$B8,$E1,$EC,$8D,$3F,$3F,$0E,$EF
	.dc.b	$C2,$3B,$1F,$3A,$7E,$89,$CB,$0F,$4B,$8C,$7E,$44,$61,$FC,$B4,$6F
	.dc.b	$0F,$54,$60,$DB,$8A,$F1,$9F,$0F,$1F,$9B,$3B,$8F,$1D,$FC,$9A,$FB
	.dc.b	$71,$FE,$2A,$BF,$DE,$E9,$FC,$81,$B1,$A1,$19,$BE,$91,$F5,$8F,$E0
	.dc.b	$3E,$8E,$30,$ED,$1F,$0C,$C9,$E6,$BF,$80,$FF,$1A,$9E,$18,$77,$AE
	.dc.b	$33,$FC,$5A,$72,$F3,$8C,$D2,$73,$3F,$AB,$3B,$3F,$28,$CC,$7D,$27
	.dc.b	$11,$E3,$BA,$F9,$8F,$65,$76,$9C,$BE,$09,$B7,$FC,$1A,$7A,$C6,$2B
	.dc.b	$F6,$11,$87,$DC,$43,$AC,$FB,$CE,$4E,$70,$FA,$53,$FE,$52,$6C,$7D
	.dc.b	$78,$C3,$C0,$F4,$5F,$F2,$38,$FE,$51,$E9,$BE,$51,$F2,$9F,$8C,$F9
	.dc.b	$3E,$38,$7F,$05,$6B,$FD,$11,$BA,$1C,$FC,$83,$E4,$FF,$B2,$D7,$F9
	.dc.b	$9A,$30,$E0,$38,$76,$D7,$FB,$6A,$EF,$E1,$1C,$1B,$19,$8E,$8D,$79
	.dc.b	$FC,$96,$88,$E9,$E3,$E2,$C8,$E2,$D7,$C6,$FF,$9F,$BB,$FF,$3B,$03
	.dc.b	$D3,$19,$E0,$F7,$36,$35,$6F,$92,$C7,$AE,$F0,$9F,$25,$BF,$1D,$E2
	.dc.b	$3D,$4C,$37,$2F,$06,$BE,$2C,$79,$61,$D3,$1E,$75,$F4,$27,$2D,$8C
	.dc.b	$BF,$E1,$88,$CB,$63,$35,$D5,$BA,$FB,$EB,$73,$F0,$BC,$B7,$91,$F6
	.dc.b	$63,$C9,$F2,$4F,$6B,$D9,$F0,$3F,$EE,$7C,$FD,$39,$7C,$9D,$D3,$99
	.dc.b	$61,$CA,$FE,$3E,$1C,$E7,$D8,$D7,$D4,$D7,$FD,$67,$9A,$31,$1E,$97
	.dc.b	$85,$75,$8F,$18,$9E,$15,$F6,$73,$79,$CF,$6B,$EC,$FA,$5F,$E4,$0C
	.dc.b	$7F,$07,$A3,$D2,$7C,$97,$DB,$6E,$9D,$B4,$61,$C7,$7F,$98,$FE,$72
	.dc.b	$33,$3F,$A5,$6B,$F6,$E3,$47,$17,$D0,$D8,$DC,$7C,$A9,$FC,$B7,$AC
	.dc.b	$F9,$27,$C9,$7E,$4B,$E4,$5E,$A7,$E4,$69,$F2,$54,$D4,$4C,$7F,$19
	.dc.b	$5C,$1B,$1E,$78,$C4,$7E,$35,$BB,$FE,$5D,$B9,$7E,$2D,$B2,$C1,$F2
	.dc.b	$4F,$92,$FD,$38,$F9,$2F,$D3,$7C,$17,$EC,$AF,$DC,$38,$D1,$9C,$FF
	.dc.b	$4D,$7C,$6B,$E9,$35,$FD,$13,$5F,$C4,$B5,$FD,$03,$5F,$C3,$3C,$63
	.dc.b	$AD,$F9,$8F,$C4,$7C,$27,$E6,$E3,$F6,$DF,$0A,$F6,$FF,$7D,$F2,$4F
	.dc.b	$91,$A7,$F0,$49,$D6,$5F,$87,$6B,$FB,$76,$BF,$9E,$6B,$FB,$56,$AF
	.dc.b	$39,$F2,$3D,$E3,$F9,$9C,$62,$BF,$99,$36,$33,$3B,$7F,$D3,$B1,$97
	.dc.b	$E7,$B7,$1F,$E4,$6D,$E2,$FE,$49,$2F,$91,$7C,$87,$F2,$38,$7C,$86
	.dc.b	$30,$FE,$55,$E1,$BC,$BE,$3F,$22,$F8,$8E,$73,$9C,$3F,$98,$4E,$5E
	.dc.b	$35,$E0,$C3,$A4,$7F,$0C,$8C,$63,$FC,$32,$30,$72,$C3,$AC,$F8,$1B
	.dc.b	$1F,$CF,$6B,$9E,$1E,$18,$F4,$DF,$38,$C6,$1E,$39,$EA,$1C,$47,$1A
	.dc.b	$E7,$87,$93,$1F,$4D,$F8,$CA,$71,$C7,$C6,$F8,$23,$35,$EE,$F0,$3C
	.dc.b	$35,$F4,$1B,$5F,$D5,$5F,$06,$5F,$05,$B7,$DB,$2F,$2E,$3F,$06,$FA
	.dc.b	$DF,$B6,$3C,$D1,$EC,$9D,$BE,$73,$DF,$8F,$B1,$F9,$15,$F3,$DF,$F7
	.dc.b	$2C,$7D,$46,$C6,$5E,$A6,$BE,$47,$F3,$0F,$F8,$22,$51,$88,$D6,$DF
	.dc.b	$E0,$99,$CB,$FE,$04,$87,$AF,$48,$FF,$82,$1B,$AF,$DB,$F9,$07,$C8
	.dc.b	$FF,$7C,$8F,$BD,$5F,$DF,$6B,$E9,$D6,$B8,$7E,$AC,$E4,$FB,$F1,$97
	.dc.b	$F0,$26,$6B,$B5,$F5,$D7,$FC,$47,$87,$D5,$7F,$4F,$7C,$77,$F0,$E1
	.dc.b	$EA,$3F,$A6,$4E,$AF,$FC,$23,$3A,$BF,$2C,$75,$AF,$D8,$9C,$70,$F0
	.dc.b	$E3,$E8,$8C,$61,$CE,$31,$7C,$9B,$3E,$CB,$E8,$FE,$F2,$74,$FB,$8E
	.dc.b	$67,$F8,$8F,$1F,$85,$87,$96,$39,$5F,$E7,$BA,$DF,$2F,$FB,$D8,$FE
	.dc.b	$E3,$18,$E2,$FF,$16,$F1,$7F,$8B,$BB,$9F,$68,$FB,$4C,$E1,$FE,$09
	.dc.b	$C7,$96,$5F,$D0,$B2,$E5,$3D,$27,$38,$F3,$C3,$94,$E7,$2E,$78,$7F
	.dc.b	$80,$7A,$5F,$07,$FD,$36,$3F,$96,$35,$D3,$7C,$74,$F0,$BF,$F1,$83
	.dc.b	$6B,$C9,$EE,$3E,$4F,$BD,$F0,$C3,$C3,$AF,$F9,$C6,$73,$F2,$7F,$A0
	.dc.b	$CB,$FC,$10,$DC,$7F,$98,$FD,$5F,$A9,$FE,$C3,$8A,$75,$B7,$F9,$15
	.dc.b	$E1,$3F,$E2,$46,$BF,$E3,$B6,$BF,$E2,$16,$CF,$3E,$AF,$2F,$C3,$F4
	.dc.b	$C7,$84,$60,$F5,$FC,$9E,$70,$FF,$15,$37,$4F,$EF,$9F,$7F,$AA,$97
	.dc.b	$FE,$55,$AE,$E4,$8F,$98,$F1,$F4,$FF,$D7,$13,$FE,$29,$8C,$7C,$1F
	.dc.b	$CA,$FB,$6F,$F3,$94,$FF,$8B,$5B,$97,$E0,$FB,$BF,$B7,$4E,$63,$81
	.dc.b	$F6,$67,$6F,$F9,$21,$AF,$F9,$7D,$AF,$F9,$15,$EB,$BF,$F9,$03,$A9
	.dc.b	$F1,$3D,$2F,$F3,$71,$87,$FC,$81,$F2,$2F,$93,$FC,$0B,$63,$F4,$3F
	.dc.b	$25,$FE,$A7,$77,$3A,$3F,$1D,$3B,$CB,$FC,$F0,$D7,$FC,$BE,$D7,$FC
	.dc.b	$EC,$EF,$1F,$F2,$AF,$D2,$7F,$8A,$6F,$A9,$FB,$07,$15,$F8,$AF,$79
	.dc.b	$F6,$BF,$96,$6D,$7E,$DB,$E0,$5F,$39,$9A,$93,$FE,$8D,$7F,$49,$5C
	.dc.b	$2F,$FD,$19,$18,$3F,$3F,$F6,$EF,$BA,$E3,$3F,$B9,$8C,$5F,$F9,$9F
	.dc.b	$EE,$AC,$FE,$7B,$EE,$BF,$DA,$B3,$9B,$FD,$5C,$62,$75,$F1,$FF,$B5
	.dc.b	$8C,$47,$FA,$11,$AF,$FA,$6D,$AF,$F9,$27,$CB,$FF,$3D,$5F,$F1,$46
	.dc.b	$C7,$48,$C7,$37,$EE,$BE,$EF,$FD,$A9,$1B,$F0,$B1,$F1,$BF,$BC,$C3
	.dc.b	$F8,$04,$FF,$A0,$6B,$FA,$36,$1F,$E9,$66,$BF,$EA,$D6,$BF,$E9,$36
	.dc.b	$F7,$BE,$F1,$F3,$7F,$C2,$1B,$7F,$CA,$9F,$E3,$EF,$FA,$B5,$CD,$FF
	.dc.b	$99,$87,$D2,$3E,$53,$F9,$0C,$61,$BF,$33,$9B,$FF,$52,$CE,$5B,$59
	.dc.b	$8F,$F7,$EB,$5F,$EA,$2D,$7F,$C8,$6D,$7F,$DD,$0F,$1E,$9F,$E7,$DF
	.dc.b	$7A,$FA,$1F,$1C,$E6,$FF,$D3,$33,$97,$FA,$3F,$DC,$BC,$8F,$E5,$2F
	.dc.b	$ED,$23,$FD,$98,$E5,$DE,$3A,$9D,$B6,$32,$7F,$52,$FA,$3F,$F7,$83
	.dc.b	$7E,$9B,$F9,$AC,$62,$7F,$A7,$28,$CF,$CA,$69,$B5,$C4,$FF,$66,$FA
	.dc.b	$CF,$A5,$FC,$3E,$1F,$E9,$88,$67,$FE,$9F,$AD,$D6,$B6,$F8,$BF,$0D
	.dc.b	$FD,$5A,$F6,$DB,$E1,$87,$3B,$E1,$87,$C8,$AD,$11,$EB,$CF,$6B,$FB
	.dc.b	$47,$CC,$BE,$83,$FA,$9B,$6B,$FB,$AB,$FE,$FB,$6C,$68,$8E,$BE,$52
	.dc.b	$D7,$F7,$3F,$9C,$7D,$F7,$F5,$E6,$D7,$F8,$01,$FE,$23,$B7,$FB,$A9
	.dc.b	$CD,$71,$AE,$07,$1F,$8B,$B7,$F9,$4E,$77,$D1,$A3,$FC,$71,$D3,$FE
	.dc.b	$92,$9C,$C6,$30,$08,$FD,$2F,$C0,$FE,$8E,$33,$1C,$1F,$F2,$5D,$73
	.dc.b	$BE,$D1,$FE,$05,$9D,$5F,$F8,$C1,$B1,$B8,$FE,$97,$1C,$EF,$B4,$F8
	.dc.b	$79,$39,$C6,$27,$58,$7A,$67,$2F,$A3,$B7,$F7,$1F,$37,$8C,$1E,$AB
	.dc.b	$EF,$DF,$73,$9B,$DE,$9F,$06,$B5,$39,$C5,$8F,$39,$CD,$7C,$7B,$E3
	.dc.b	$8F,$C2,$3B,$E3,$DA,$7E,$67,$4F,$E7,$7E,$FB,$C8,$FE,$73,$CD,$BD
	.dc.b	$49,$CF,$5A,$73,$18,$6D,$F4,$9D,$3F,$67,$1F,$B1,$F4,$5F,$05,$FB
	.dc.b	$5E,$C5,$E2,$4D,$8C,$36,$B9,$4E,$9B,$39,$AF,$71,$D2,$37,$59,$E3
	.dc.b	$FC,$36,$8B,$FE,$8B,$18,$6F,$0F,$E3,$A7,$5E,$BF,$C4,$9C,$E2,$9E
	.dc.b	$5E,$56,$4D,$55,$CB,$E6,$FB,$7F,$DA,$F2,$F0,$B7,$2E,$CB,$EF,$FE
	.dc.b	$8B,$9C,$BF,$C6,$3E,$96,$0C,$CD,$C3,$E6,$7D,$23,$E9,$BF,$96,$5F
	.dc.b	$C2,$8C,$78,$FF,$23,$E0,$BE,$84,$62,$75,$5C,$34,$FC,$EC,$E5,$FC
	.dc.b	$E5,$70,$9D,$69,$FB,$2B,$5E,$E3,$0D,$8F,$D9,$D7,$1C,$3C,$F7,$CA
	.dc.b	$7F,$77,$5B,$7F,$79,$5B,$7C,$B5,$C0,$F5,$63,$63,$24,$C7,$CF,$19
	.dc.b	$7D,$78,$77,$9F,$3B,$78,$BF,$CF,$D3,$AB,$E1,$97,$AA,$B5,$8F,$8E
	.dc.b	$B5,$87,$7A,$FE,$2D,$39,$78,$5F,$29,$F8,$38,$D7,$C3,$17,$1F,$0D
	.dc.b	$70,$9E,$87,$C4,$C7,$C3,$7C,$72,$F8,$C7,$C7,$BE,$35,$BC,$35,$7C
	.dc.b	$23,$99,$E0,$2C,$F0,$71,$97,$B0,$98,$FB,$31,$E7,$87,$9F,$2F,$75
	.dc.b	$F7,$39,$B8,$79,$D7,$AE,$FE,$9C,$78,$B2,$F4,$E5,$E5,$9E,$F3,$CD
	.dc.b	$AF,$E0,$B1,$F0,$BD,$9B,$1A,$CB,$D5,$7D,$32,$F3,$57,$0B,$FF,$75
	.dc.b	$CF,$E7,$DB,$87,$E9,$9B,$5F,$EF,$07,$F2,$7F,$4D,$FE,$84,$71,$48
	.dc.b	$C4,$74,$6B,$FA,$06,$BC,$B7,$FD,$AC,$73,$CB,$F5,$AD,$CF,$F5,$5E
	.dc.b	$0F,$F2,$0B,$DD,$C7,$F7,$55,$B7,$FB,$0C,$ED,$CC,$EE,$FA,$BF,$CB
	.dc.b	$E7,$51,$89,$D9,$1F,$EB,$71,$AB,$F5,$57,$36,$D7,$F2,$C9,$D3,$ED
	.dc.b	$39,$46,$AB,$9F,$27,$F0,$57,$C0,$48,$D1,$F0,$23,$4E,$27,$A4,$ED
	.dc.b	$F7,$CE,$9C,$CE,$DF,$E2,$38,$7C,$55,$7E,$67,$38,$7A,$DE,$27,$43
	.dc.b	$06,$8F,$59,$92,$99,$C3,$C8,$F6,$C3,$C8,$E7,$0F,$26,$3F,$87,$BF
	.dc.b	$A5,$5A,$8C,$57,$46,$DF,$C8,$8F,$1C,$62,$FD,$D8,$7A,$E3,$13,$EA
	.dc.b	$3C,$07,$9E,$BC,$B1,$A8,$C9,$E7,$3A,$E3,$F1,$E3,$55,$D6,$30,$4C
	.dc.b	$7D,$35,$C1,$F8,$F3,$B7,$11,$F8,$5C,$7D,$D5,$C6,$3E,$4D,$6D,$F8
	.dc.b	$11,$F5,$E3,$18,$7A,$23,$15,$F7,$F8,$DF,$80,$38,$30,$DC,$7F,$C3
	.dc.b	$2D,$7F,$C3,$D3,$C3,$C4,$DD,$49,$DB,$CA,$3B,$7D,$5B,$EE,$A9,$2B
	.dc.b	$FC,$4B,$0F,$9C,$C4,$76,$8E,$0E,$A7,$9C,$FE,$13,$5F,$F3,$A4,$FE
	.dc.b	$CF,$C8,$C1,$9C,$64,$64,$87,$EB,$DC,$BF,$B1,$8E,$2F,$EF,$63,$E3
	.dc.b	$7E,$02,$F8,$BF,$91,$E0,$7F,$92,$AB,$FC,$7A,$D7,$E2,$CE,$DF,$F2
	.dc.b	$0B,$5F,$93,$1B,$E1,$FF,$13,$B7,$93,$FC,$49,$5B,$9F,$F3,$87,$84
	.dc.b	$FA,$8F,$E0,$BE,$27,$F9,$6F,$DB,$FD,$C3,$72,$FF,$2A,$4F,$F0,$AA
	.dc.b	$E1,$CB,$AF,$15,$85,$3F,$80,$4E,$AF,$F8,$E7,$D5,$7C,$DF,$7E,$46
	.dc.b	$08,$7F,$9F,$8F,$E4,$F1,$FE,$17,$EE,$FF,$0C,$47,$F0,$68,$C3,$FC
	.dc.b	$C9,$AF,$03,$9E,$9F,$C9,$A3,$F7,$F9,$BC,$FF,$2F,$6B,$A3,$FC,$49
	.dc.b	$E7,$FF,$A3,$B4,$FF,$12,$C6,$33,$FF,$41,$56,$F3,$FF,$4D,$DE,$E1
	.dc.b	$F2,$1B,$96,$1C,$9D,$DF,$2C,$35,$39,$96,$1C,$EF,$3E,$13,$CF,$A4
	.dc.b	$AE,$67,$68,$7E,$05,$98,$E7,$97,$FA,$66,$79,$47,$0C,$7F,$D6,$57
	.dc.b	$DA,$FD,$1D,$FF,$EE,$7F,$C0,$EF,$49,$A6,$BA,$3F,$CE,$58,$FF,$BA
	.dc.b	$23,$2D,$3F,$DE,$31,$A7,$FD,$E0,$DD,$7F,$D7,$77,$DE,$78,$1A,$F2
	.dc.b	$BE,$2F,$C0,$7F,$B6,$A3,$D0,$FB,$9B,$96,$BD,$39,$18,$21,$F0,$8F
	.dc.b	$A7,$3C,$DC,$6A,$FB,$AF,$F3,$D4,$7C,$5F,$BE,$FF,$03,$E7,$F8,$36
	.dc.b	$F0,$FF,$B0,$EB,$75,$C9,$B5,$BF,$23,$EE,$DF,$0A,$FC,$1D,$6A,$FF
	.dc.b	$CE,$5F,$65,$FE,$C2,$9F,$B7,$B3,$D0,$FD,$16,$5F,$19,$B7,$C3,$D3
	.dc.b	$FC,$75,$B0,$E4,$FF,$88,$DA,$FE,$A1,$AF,$13,$FC,$50,$71,$8E,$71
	.dc.b	$CA,$FF,$64,$DA,$FF,$44,$75,$3E,$BF,$F3,$3F,$68,$E6,$A4,$CB,$F1
	.dc.b	$D1,$F9,$07,$E9,$7C,$8F,$FB,$4F,$C3,$F0,$B6,$3F,$C7,$4D,$7F,$7E
	.dc.b	$D7,$FD,$30,$FF,$B6,$FD,$8F,$D4,$FC,$23,$C1,$A7,$EC,$67,$55,$F9
	.dc.b	$FF,$C1,$3F,$A1,$37,$CD,$FD,$8C,$62,$3C,$BE,$DF,$EF,$5B,$E1,$79
	.dc.b	$A3,$11,$E2,$6A,$16,$8C,$C6,$3E,$23,$F8,$83,$FB,$89,$D5,$79,$EB
	.dc.b	$C1,$C6,$E2,$E9,$4C,$38,$E1,$F2,$BD,$A7,$C1,$FF,$0A,$F8,$97,$C6
	.dc.b	$D3,$E1,$4E,$AB,$C0,$D7,$E6,$B5,$F9,$AD,$7C,$2D,$7D,$AF,$1C,$BC
	.dc.b	$D1,$8B,$EE,$7F,$1C,$F8,$B7,$17,$D3,$FA,$FF,$ED,$75,$FB,$0F,$8D
	.dc.b	$1F,$C8,$8A,$FC,$58,$DE,$1D,$B5,$47,$29,$FC,$1B,$5F,$8E,$D7,$E9
	.dc.b	$B6,$38,$FA,$BE,$32,$4F,$F2,$3F,$90,$EF,$C1,$F4,$36,$FF,$0A,$63
	.dc.b	$FD,$0F,$F0,$5F,$E1,$0D,$F1,$BF,$1D,$3A,$AE,$1E,$BF,$E7,$23,$FC
	.dc.b	$39,$8F,$E6,$1A,$E8,$FC,$FB,$C5,$FC,$4B,$73,$FB,$73,$FD,$1B,$E0
	.dc.b	$9F,$05,$F9,$DF,$A9,$7E,$0D,$FD,$2B,$CA,$FF,$0D,$E5,$FC,$71,$BA
	.dc.b	$FC,$9A,$DC,$FE,$45,$AF,$EB,$9A,$FE,$B9,$AF,$E4,$9A,$FE,$79,$E3
	.dc.b	$1D,$6B,$C8,$7F,$5C,$F1,$FF,$27,$5F,$9F,$F8,$8F,$93,$B7,$E9,$F1
	.dc.b	$7C,$17,$F2,$56,$F9,$9F,$B9,$AD,$DF,$EB,$1F,$E7,$0F,$F0,$26,$BF
	.dc.b	$A6,$6B,$BE,$1F,$E1,$B1,$CB,$80,$D3,$AF,$CF,$BF,$00,$7D,$07,$F0
	.dc.b	$9F,$9D,$F2,$7C,$EB,$EC,$7F,$BB,$B7,$3F,$DD,$4F,$F3,$DC,$8F,$FB
	.dc.b	$2A,$3F,$81,$B6,$33,$A7,$F2,$08,$D6,$BF,$C2,$3D,$8F,$F0,$1D,$7F
	.dc.b	$27,$E3,$FF,$09,$37,$87,$F8,$77,$17,$F7,$3F,$8A,$DD,$CC,$FF,$BE
	.dc.b	$5E,$D5,$6D,$FF,$03,$8F,$A2,$7D,$1F,$F8,$42,$1D,$BF,$09,$FE,$39
	.dc.b	$F4,$9F,$E4,$D8,$C4,$23,$1C,$5F,$C8,$63,$A5,$73,$AD,$DF,$6D,$7F
	.dc.b	$93,$47,$2C,$38,$DF,$6F,$87,$E5,$3C,$30,$E5,$8F,$5C,$79,$5F,$9A
	.dc.b	$3C,$18,$72,$8F,$26,$1C,$A7,$85,$F2,$9F,$E6,$D7,$CB,$0F,$15,$EE
	.dc.b	$7C,$47,$9A,$F8,$E3,$63,$8E,$1F,$1F,$B3,$8C,$7F,$41,$6A,$7D,$53
	.dc.b	$9A,$E7,$3E,$93,$B9,$4F,$49,$DC,$F9,$71,$71,$F4,$4F,$C0,$C3,$D1
	.dc.b	$5C,$0E,$53,$A7,$CF,$5F,$DF,$A3,$E4,$C7,$F9,$A9,$AF,$90,$C7,$CF
	.dc.b	$B9,$F8,$36,$3F,$CE,$27,$E2,$5C,$3C,$23,$FC,$16,$FF,$63,$F8,$3A
	.dc.b	$9A,$47,$6F,$BF,$E9,$1E,$8F,$F4,$CD,$BF,$1F,$5C,$CF,$8D,$3A,$E4
	.dc.b	$F8,$D1,$99,$FC,$AC,$E6,$3E,$61,$9D,$7F,$2D,$19,$9E,$B3,$E5,$3A
	.dc.b	$DF,$86,$BD,$F8,$79,$63,$13,$FE,$92,$8C,$61,$E2,$8E,$57,$E2,$AD
	.dc.b	$C7,$E9,$6B,$6F,$90,$DD,$F9,$27,$71,$E7,$3F,$C1,$98,$79,$E7,$C7
	.dc.b	$1E,$6C,$3C,$17,$D3,$0F,$5B,$6B,$F9,$77,$D2,$7F,$B3,$6B,$FD,$00
	.dc.b	$DB,$FF,$32,$71,$3E,$93,$FD,$33,$5A,$E0,$FE,$FB,$C1,$FE,$A3,$C3
	.dc.b	$83,$63,$33,$FC,$52,$B8,$E1,$FE,$6C,$BD,$CF,$F8,$6B,$A1,$E1,$F0
	.dc.b	$AF,$E4,$AD,$E0,$7D,$23,$EB,$B5,$D1,$FC,$B2,$AE,$1C,$B1,$FE,$63
	.dc.b	$1C,$6F,$84,$60,$FF,$17,$7D,$2B,$97,$9B,$63,$FD,$53,$97,$83,$E1
	.dc.b	$F3,$52,$78,$BD,$63,$1A,$7F,$96,$AF,$93,$FE,$8A,$F1,$2F,$0F,$F8
	.dc.b	$5E,$6E,$1F,$D7,$E6,$D7,$F9,$1B,$E9,$6E,$C7,$3D,$FF,$CE,$3C,$8F
	.dc.b	$83,$1A,$26,$9F,$26,$75,$8F,$F9,$2E,$72,$FF,$AF,$7E,$39,$F0,$DC
	.dc.b	$AF,$84,$62,$7F,$CA,$5F,$4B,$FE,$59,$6E,$7D,$7B,$DF,$51,$F4,$AB
	.dc.b	$53,$FE,$7A,$6F,$07,$F8,$D5,$B7,$FD,$F6,$F8,$BD,$09,$5F,$EB,$0E
	.dc.b	$D7,$8B,$D6,$3D,$91,$89,$FF,$33,$7D,$2F,$F9,$71,$BE,$77,$C1,$FA
	.dc.b	$6F,$F7,$D4,$60,$97,$FE,$8E,$AE,$0F,$FA,$FA,$76,$E6,$BF,$1B,$3A
	.dc.b	$CF,$FE,$00,$7A,$7F,$EE,$FC,$39,$C6,$3C,$8F,$C9,$FA,$EF,$9A,$A7
	.dc.b	$D3,$F5,$DF,$23,$8C,$3F,$CD,$F3,$74,$FF,$67,$D6,$A7,$24,$8F,$F6
	.dc.b	$9F,$B4,$F9,$2E,$19,$FF,$9B,$BA,$53,$9B,$FF,$45,$7B,$97,$0F,$F5
	.dc.b	$6F,$D3,$3E,$37,$F5,$33,$9C,$7F,$CD,$53,$9B,$FF,$49,$36,$F8,$DF
	.dc.b	$0C,$BF,$63,$39,$BF,$17,$9D,$FD,$77,$0F,$F5,$04,$63,$9F,$78,$7E
	.dc.b	$56,$B5,$8F,$FB,$03,$E9,$5F,$3F,$9F,$A6,$FD,$EC,$63,$2E,$67,$F9
	.dc.b	$16,$D3,$A6,$46,$AD,$BF,$F4,$8A,$CC,$7F,$BF,$3E,$95,$F4,$DF,$A7
	.dc.b	$6D,$F1,$B4,$7F,$16,$AE,$6E,$96,$BF,$E9,$E5,$F4,$CF,$9B,$FE,$35
	.dc.b	$18,$6F,$CC,$67,$1F,$F1,$7D,$5A,$E7,$1F,$E2,$78,$C5,$F5,$6C,$7F
	.dc.b	$8A,$EA,$D7,$FB,$36,$6B,$DE,$7A,$DF,$80,$F1,$CF,$0C,$5F,$4D,$FE
	.dc.b	$6E,$AF,$80,$FE,$DA,$FE,$4B,$97,$CA,$60,$FE,$1C,$E7,$1F,$C9,$3C
	.dc.b	$6B,$AF,$93,$AA,$FE,$29,$F4,$CF,$A6,$BF,$11,$86,$E5,$F4,$67,$4D
	.dc.b	$9D,$34,$FB,$DE,$BB,$E4,$7F,$0D,$87,$F5,$8F,$96,$FF,$3D,$D6,$A7
	.dc.b	$3A,$FF,$B5,$2D,$5A,$BF,$EB,$F5,$AB,$FF,$35,$CE,$6F,$84,$E6,$BF
	.dc.b	$CD,$33,$9A,$DF,$C2,$7F,$59,$C3,$AB,$6F,$84,$62,$7F,$1D,$F3,$AF
	.dc.b	$8A,$FE,$70,$D8,$CF,$D0,$E2,$64,$70,$77,$18,$CB,$EA,$F0,$3E,$81
	.dc.b	$F1,$3F,$D0,$DB,$87,$E7,$5F,$E3,$55,$C1,$CB,$67,$81,$C7,$C7,$B5
	.dc.b	$FE,$86,$E7,$7C,$7F,$FB,$FB,$E0,$BF,$41,$80,$4E,$AB,$FC,$81,$D3
	.dc.b	$F9,$9E,$BE,$0F,$E6,$F0,$B7,$C6,$F8,$3F,$E4,$59,$D5,$F2,$8D,$5F
	.dc.b	$07,$BD,$DB,$E3,$87,$92,$73,$7C,$5F,$2D,$EE,$FC,$3A,$7E,$E7,$E6
	.dc.b	$7C,$7E,$9B,$E8,$1F,$4D,$AD,$55,$A9,$CE,$7E,$5C,$39,$3E,$7C,$39
	.dc.b	$61,$C6,$F8,$3F,$03,$93,$F3,$5F,$4D,$7A,$1E,$76,$7A,$13,$98,$C3
	.dc.b	$72,$F8,$97,$C9,$B7,$C9,$F8,$1C,$4F,$98,$F8,$DE,$CF,$C7,$6C,$E9
	.dc.b	$BC,$7F,$44,$FA,$8E,$39,$3E,$65,$F3,$AE,$B7,$CD,$A7,$91,$FE,$7F
	.dc.b	$3F,$86,$C7,$FC,$23,$E4,$FC,$19,$FE,$F9,$8F,$79,$CD,$F1,$AD,$1B
	.dc.b	$F5,$AC,$7C,$0F,$6F,$FB,$83,$7D,$1F,$99,$0F,$8C,$B1,$27,$47,$18
	.dc.b	$C6,$3E,$F1,$F4,$4F,$A6,$78,$FF,$89,$8C,$3F,$33,$E1,$3F,$25,$18
	.dc.b	$AE,$4D,$DF,$F2,$4D,$EB,$FE,$75,$5C,$30,$EB,$7C,$1B,$7C,$1B,$5C
	.dc.b	$32,$F1,$46,$2B,$F6,$18,$75,$6C,$65,$FD,$9E,$1D,$F1,$E2,$48,$F8
	.dc.b	$1C,$3F,$E6,$97,$C5,$19,$CB,$C4,$FB,$F1,$C6,$5F,$EF,$DC,$BC,$17
	.dc.b	$F2,$32,$F0,$63,$CF,$1E,$33,$D0,$8F,$E2,$B2,$F2,$46,$30,$E9,$7C
	.dc.b	$A7,$35,$C8,$93,$F0,$EB,$81,$F0,$32,$F2,$BF,$C2,$EB,$8C,$F4,$9C
	.dc.b	$CF,$C4,$CA,$E5,$EA,$3D,$79,$7A,$6F,$85,$76,$3D,$D7,$F0,$DC,$57
	.dc.b	$1C,$BD,$18,$74,$7E,$94,$74,$C7,$C7,$5E,$73,$A6,$3D,$DC,$65,$EE
	.dc.b	$6E,$5E,$F8,$C4,$78,$8E,$35,$D1,$C4,$6A,$BA,$46,$2B,$D7,$39,$AE
	.dc.b	$17,$E9,$CB,$D9,$3D,$A3,$17,$F8,$6A,$DE,$3E,$C3,$DF,$1F,$09,$FC
	.dc.b	$DB,$FD,$0C,$F7,$14,$F4,$16,$7F,$50,$DF,$A6,$FF,$40,$BB,$3F,$66
	.dc.b	$E1,$D6,$F9,$72,$D9,$4F,$2C,$3F,$62,$DC,$BF,$30,$DB,$FE,$B1,$E0
	.dc.b	$FF,$25,$7F,$7D,$19,$7A,$9F,$BC,$71,$1B,$8C,$61,$C0,$8F,$F0,$B8
	.dc.b	$D5,$F8,$A7,$2D,$9D,$72,$7C,$78,$C1,$30,$F3,$D7,$F5,$9B,$E7,$18
	.dc.b	$BF,$09,$2B,$FB,$04,$EA,$3F,$82,$E1,$EB,$C7,$E2,$E1,$E2,$C3,$EB
	.dc.b	$5E,$EF,$DF,$87,$85,$E3,$5E,$1C,$7C,$D7,$E1,$AD,$57,$08,$E7,$3A
	.dc.b	$8E,$27,$96,$38,$1A,$7C,$46,$A7,$9D,$F8,$F0,$F3,$CF,$13,$55,$C5
	.dc.b	$C1,$E1,$32,$7D,$CB,$F1,$39,$24,$F4,$BF,$8F,$3A,$8C,$57,$9A,$B9
	.dc.b	$9F,$E2,$D9,$D6,$3F,$2C,$98,$FC,$59,$F5,$9D,$A3,$F0,$F3,$C2,$73
	.dc.b	$7D,$AB,$CC,$62,$B9,$BF,$84,$31,$3D,$31,$F8,$93,$AC,$7E,$3E,$1D
	.dc.b	$6F,$E1,$E1,$D5,$F4,$1C,$8F,$8F,$7C,$B0,$C5,$FC,$29,$FF,$05,$FD
	.dc.b	$33,$4F,$B5,$C8,$8C,$4E,$A7,$30,$F0,$EC,$49,$D1,$D2,$39,$FC,$CB
	.dc.b	$E9,$A9,$27,$FC,$37,$0F,$9C,$C4,$7F,$81,$63,$83,$A9,$E7,$3F,$7B
	.dc.b	$4F,$2C,$FE,$C7,$C7,$C1,$9C,$64,$7E,$AE,$32,$FE,$B6,$32,$F1,$34
	.dc.b	$64,$FF,$17,$FD,$2B,$E3,$FE,$2F,$81,$FE,$46,$9F,$F1,$9B,$5F,$F0
	.dc.b	$E4,$ED,$E3,$3C,$3C,$19,$18,$3F,$C9,$5F,$43,$FB,$9F,$13,$FC,$A5
	.dc.b	$ED,$7E,$AD,$B3,$AE,$6F,$D7,$78,$CB,$E1,$E3,$B0,$E0,$DB,$E0,$7F
	.dc.b	$0B,$F9,$FF,$90,$47,$FC,$CD,$4F,$37,$11,$FC,$6E,$6F,$47,$F9,$B6
	.dc.b	$B5,$3C,$C8,$6C,$E7,$5E,$6E,$A7,$37,$F4,$5F,$3D,$E0,$FF,$A2,$75
	.dc.b	$FF,$18,$CE,$63,$15,$C3,$0F,$F7,$B2,$F8,$BD,$47,$FC,$00,$8D,$3F
	.dc.b	$EA,$46,$E1,$DA,$35,$7C,$30,$FF,$1F,$DD,$F0,$9E,$75,$25,$FF,$A7
	.dc.b	$70,$E5,$7F,$E9,$EF,$A5,$CC,$FF,$AC,$AF,$8B,$BC,$FD,$91,$88,$D4
	.dc.b	$FA,$3B,$FF,$DB,$5F,$4D,$7A,$4D,$35,$D2,$F2,$A9,$9C,$3F,$DB,$D3
	.dc.b	$87,$FD,$CB,$3B,$72,$43,$FD,$F7,$68,$D4,$7F,$B2,$BD,$DF,$F6,$4B
	.dc.b	$E9,$1A,$21,$97,$C9,$6C,$EE,$75,$E9,$64,$D5,$49,$D5,$76,$7F,$DB
	.dc.b	$3F,$4B,$FE,$77,$79,$7D,$37,$F7,$FE,$0F,$9E,$DE,$0F,$C3,$56,$EB
	.dc.b	$8F,$97,$F6,$EB,$51,$E8,$6E,$6A,$E1,$87,$E0,$E9,$F4,$DF,$EB,$C7
	.dc.b	$E0,$18,$F4,$FF,$1A,$D3,$3D,$2F,$A6,$CD,$71,$95,$F5,$8C,$36,$F9
	.dc.b	$DF,$1A,$EF,$5A,$D3,$FC,$50,$D8,$FF,$76,$35,$FD,$F6,$1F,$99,$FA
	.dc.b	$67,$2D,$24,$FD,$38,$FC,$86,$5F,$A1,$F9,$1F,$F6,$5F,$83,$FA,$86
	.dc.b	$D7,$DF,$6B,$EA,$CB,$F6,$9F,$02,$F0,$38,$7A,$CF,$6F,$F5,$53,$9B
	.dc.b	$FC,$13,$5E,$43,$0F,$F6,$DC,$FE,$C7,$E9,$BF,$9F,$4F,$E2,$5A,$F0
	.dc.b	$E3,$46,$34,$F5,$7B,$5F,$BC,$D1,$7D,$63,$0D,$F0,$7D,$D1,$8C,$9F
	.dc.b	$0D,$FC,$62,$BC,$B1,$8F,$13,$77,$27,$E2,$56,$F1,$F9,$5E,$DB,$E9
	.dc.b	$BF,$85,$7C,$43,$E3,$69,$F3,$67,$58,$7C,$D6,$BE,$06,$DF,$4B,$FE
	.dc.b	$29,$F1,$5F,$36,$B8,$47,$32,$3F,$AF,$8D,$3E,$AF,$A6,$7A,$5F,$C7
	.dc.b	$7E,$34,$53,$F0,$11,$A9,$FC,$03,$5E,$53,$A3,$F9,$37,$C8,$67,$1F
	.dc.b	$C5,$55,$E2,$FA,$3C,$28,$C6,$1F,$7D,$BC,$8C,$33,$5F,$1F,$D4,$FE
	.dc.b	$65,$F4,$DF,$C0,$F2,$FE,$5A,$DC,$BA,$F0,$7F,$7C,$9F,$CD,$36,$33
	.dc.b	$5F,$9B,$8C,$67,$FC,$63,$E0,$9F,$05,$9A,$93,$F9,$FA,$DE,$8B,$DC
	.dc.b	$FE,$8F,$E9,$9F,$4D,$FD,$1E,$BF,$C2,$33,$F9,$56,$E7,$FA,$DA,$DC
	.dc.b	$F1,$AF,$D7,$46,$23,$F1,$ED,$9F,$E9,$7E,$6F,$E2,$67,$81,$0D,$D7
	.dc.b	$F5,$78,$D4,$7E,$63,$E1,$FE,$33,$FC,$DB,$E1,$3F,$8F,$9F,$C8,$B8
	.dc.b	$13,$A6,$F9,$DF,$C0,$63,$07,$2F,$99,$7D,$33,$E8,$35,$53,$1F,$E0
	.dc.b	$D7,$BA,$D4,$7F,$09,$F9,$DF,$01,$3E,$17,$FB,$AE,$3D,$3E,$BB,$F8
	.dc.b	$84,$E6,$B8,$56,$AD,$5F,$DD,$73,$FE,$33,$CD,$FE,$05,$6D,$FF,$54
	.dc.b	$6E,$DF,$DA,$3D,$2E,$1E,$CC,$AF,$EA,$95,$AF,$5F,$FA,$0E,$1F,$ED
	.dc.b	$5F,$A3,$7D,$27,$F3,$8B,$FC,$B7,$17,$FB,$67,$E9,$7F,$C7,$11,$83
	.dc.b	$F8,$7B,$8E,$EF,$83,$0B,$E7,$CF,$FC,$52,$73,$18,$9E,$42,$B9,$5F
	.dc.b	$7F,$84,$FF,$74,$5F,$0C,$7C,$8D,$C7,$A5,$6F,$0A,$79,$2F,$8C,$7B
	.dc.b	$6F,$8B,$6B,$75,$E1,$AD,$BC,$4C,$E3,$E7,$C3,$C1,$1E,$8C,$7C,$EE
	.dc.b	$CF,$5C,$EF,$0E,$D1,$BB,$EA,$F7,$3A,$E3,$E4,$7F,$9C,$D7,$17,$A6
	.dc.b	$1F,$33,$0E,$F1,$F3,$6B,$83,$71,$FE,$82,$C3,$05,$29,$9B,$FA,$4D
	.dc.b	$F9,$E7,$2E,$E3,$ED,$46,$1E,$52,$8C,$47,$19,$E0,$FF,$52,$F8,$1A
	.dc.b	$9A,$47,$6F,$D2,$FA,$57,$A1,$FE,$55,$D3,$F1,$75,$CC,$FC,$3C,$EB
	.dc.b	$7F,$B3,$19,$9F,$C8,$CE,$63,$F1,$A6,$77,$F8,$B1,$99,$EB,$3E,$53
	.dc.b	$AD,$F8,$6B,$E4,$61,$E5,$8C,$4F,$CD,$8C,$61,$E2,$8E,$57,$E2,$AD
	.dc.b	$C7,$D4,$AD,$BE,$43,$77,$E4,$9D,$C7,$9C,$FE,$F7,$87,$9E,$7C,$71
	.dc.b	$E6,$C3,$C1,$7D,$30,$F5,$B6,$BF,$A9,$FD,$CF,$F8,$5E,$BF,$C3,$8D
	.dc.b	$BF,$F0,$98,$F5,$5F,$31,$FE,$85,$DF,$FC,$4D,$5B,$6F,$AB,$9C,$3F
	.dc.b	$C1,$FD,$0F,$A6,$FF,$49,$47,$69,$D6,$4A,$D7,$C2,$62,$1F,$76,$F9
	.dc.b	$3D,$F1,$BE,$95,$F4,$DB,$AD,$67,$FE,$73,$AD,$4E,$6F,$8D,$F0,$AF
	.dc.b	$F2,$CD,$F0,$C3,$FC,$63,$87,$09,$CF,$77,$53,$C3,$B7,$F6,$68,$C9
	.dc.b	$2B,$8E,$1E,$E7,$E4,$E0,$BF,$0E,$7C,$EF,$CD,$7B,$C3,$FC,$DD,$7B
	.dc.b	$9C,$D7,$C6,$ED,$7D,$2F,$F9,$2E,$7F,$CA,$4D,$CB,$FC,$93,$F1,$AF
	.dc.b	$B8,$F3,$47,$F8,$E2,$75,$97,$F9,$75,$B7,$FE,$CB,$E0,$DC,$E7,$DD
	.dc.b	$FA,$BE,$67,$D5,$F4,$3C,$58,$70,$C7,$FD,$3F,$87,$0C,$7F,$C3,$B3
	.dc.b	$9B,$E1,$1F,$E6,$7F,$B3,$F7,$3F,$E2,$26,$DB,$47,$E0,$EF,$33,$E0
	.dc.b	$56,$B7,$3C,$31,$FF,$09,$7D,$46,$B8,$7C,$FE,$8A,$31,$A7,$FA,$BE
	.dc.b	$30,$F8,$E7,$73,$F3,$7E,$9B,$F0,$9A,$FF,$97,$9B,$C5,$FE,$E3,$D9
	.dc.b	$F6,$DF,$C7,$71,$FF,$41,$56,$9F,$F1,$EB,$6F,$FD,$87,$F5,$0F,$81
	.dc.b	$ED,$D9,$F3,$23,$17,$CA,$31,$3A,$8F,$F6,$34,$EA,$BF,$CE,$3F,$84
	.dc.b	$3D,$1F,$C2,$36,$BF,$DA,$59,$BE,$E3,$FD,$61,$1F,$E9,$79,$D5,$F2
	.dc.b	$BF,$BB,$EF,$3E,$AB,$9C,$61,$B7,$FE,$69,$6D,$7F,$A3,$EB,$84,$FF
	.dc.b	$A8,$FE,$A9,$F3,$FF,$97,$6C,$79,$32,$FA,$37,$FE,$8A,$9C,$B6,$BF
	.dc.b	$3D,$F0,$AF,$99,$E5,$18,$66,$1D,$FE,$4E,$DF,$37,$F4,$DF,$84,$FE
	.dc.b	$86,$7F,$D6,$DF,$54,$FA,$BC,$F8,$9F,$AB,$8C,$4E,$A7,$F3,$FE,$9F
	.dc.b	$F1,$E8,$C1,$FE,$B6,$F3,$B8,$47,$E9,$3E,$AD,$EC,$E2,$BF,$DF,$91
	.dc.b	$87,$FD,$CB,$F3,$6F,$B4,$FE,$38,$D8,$CF,$83,$89,$91,$C3,$E9,$6F
	.dc.b	$07,$F3,$2B,$46,$3C,$0F,$F7,$A4,$62,$BB,$DF,$6A,$FE,$5B,$D6,$FC
	.dc.b	$29,$F5,$3F,$BC,$FA,$CF,$EC,$59,$7E,$76,$73,$8F,$FB,$1A,$0F,$FB
	.dc.b	$23,$C6,$57,$08,$FF,$65,$E1,$D7,$0E,$9C,$1E,$0C,$79,$EB,$FE,$D5
	.dc.b	$C3,$82,$C5,$F5,$6F,$97,$FE,$1E,$DC,$FF,$37,$83,$E3,$33,$52,$7C
	.dc.b	$C7,$96,$B8,$61,$FC,$8A,$F8,$4E,$A3,$C1,$3A,$D7,$DB,$0F,$79,$F8
	.dc.b	$53,$EE,$FF,$AA,$C6,$1F,$EA,$FE,$93,$E6,$BF,$69,$87,$2A,$D6,$5D
	.dc.b	$DB,$8F,$FB,$F2,$F8,$75,$74,$CB,$B5,$F9,$67,$2F,$18,$CD,$71,$C7
	.dc.b	$F2,$EB,$1F,$04,$E9,$F5,$DF,$29,$FE,$87,$F4,$BE,$6C,$3E,$D7,$94
	.dc.b	$FA,$AF,$C2,$CE,$6B,$EA,$4E,$6B,$F8,$5C,$E7,$8B,$F8,$D5,$F0,$AF
	.dc.b	$E8,$A6,$EF,$C9,$7F,$87,$FA,$87,$C5,$7F,$4A,$6C,$74,$F3,$B1,$32
	.dc.b	$3E,$B6,$2A,$DD,$FA,$AF,$4F,$E8,$FA,$3F,$56,$7E,$B3,$EF,$6E,$1F
	.dc.b	$B0,$8F,$BB,$F0,$BF,$A8,$A0,$EE,$7C,$E6,$BB,$BF,$CC,$B1,$F9,$8A
	.dc.b	$E1,$5B,$AD,$3C,$A7,$55,$C2,$73,$3D,$6E,$D7,$6C,$3F,$CC,$17,$FA
	.dc.b	$8E,$0F,$E2,$D7,$FE,$61,$8F,$D7,$0F,$A2,$EF,$DE,$AD,$E7,$EB,$DF
	.dc.b	$F8,$1D,$6A,$73,$7C,$A3,$1E,$37,$A2,$B7,$97,$8B,$2F,$F1,$87,$3B
	.dc.b	$E7,$FD,$BC,$DF,$9D,$6C,$7E,$92,$B8,$4E,$5B,$1C,$5B,$3C,$3A,$38
	.dc.b	$BF,$0B,$BA,$4F,$87,$0F,$A5,$F5,$F8,$7E,$11,$F3,$67,$E4,$FD,$46
	.dc.b	$78,$3E,$9F,$12,$31,$C1,$F6,$1B,$7F,$55,$F5,$9F,$D8,$2B,$F0,$5A
	.dc.b	$F1,$F1,$73,$E9,$E9,$BA,$7B,$AD,$18,$C3,$CB,$7C,$6B,$E1,$FB,$FF
	.dc.b	$E4,$9C,$7A,$B7,$2F,$F3,$0B,$FE,$3B,$BE,$2E,$5F,$F1,$67,$C2,$FD
	.dc.b	$08,$C6,$07,$85,$F4,$9F,$C6,$FD,$4B,$EB,$9F,$2F,$F2,$E3,$0F,$CD
	.dc.b	$DF,$F9,$0E,$5F,$30,$4E,$9C,$37,$0F,$CF,$B7,$3F,$D1,$F9,$1E,$0C
	.dc.b	$7B,$DF,$06,$DF,$06,$D7,$0D,$7C,$91,$8A,$FD,$8E,$1D,$5B,$19,$7F
	.dc.b	$69,$87,$7C,$B8,$92,$3D,$59,$7B,$32,$F0,$4E,$5B,$3A,$AE,$4F,$9E
	.dc.b	$75,$97,$96,$30,$4A,$E2,$7E,$3A,$B8,$B8,$9F,$F7,$4E,$3E,$9A,$F3
	.dc.b	$E5,$D1,$F5,$E1,$5C,$9D,$6B,$8E,$5E,$FB,$E5,$97,$AE,$73,$7C,$67
	.dc.b	$39,$78,$A7,$38,$73,$3F,$89,$46,$5F,$09,$93,$C3,$97,$A5,$CB,$5C
	.dc.b	$CF,$48,$C5,$78,$32,$F1,$65,$E7,$3E,$B6,$5E,$7C,$39,$E1,$C4,$8F
	.dc.b	$92,$B8,$E5,$F0,$EB,$84,$FE,$4A,$FA,$4E,$72,$F5,$4E,$63,$17,$CA
	.dc.b	$3D,$97,$CB,$1F,$A4,$4C,$3C,$B8,$72,$8C,$61,$FA,$38,$C4,$FE,$6A
	.dc.b	$FF,$A1,$9F,$20,$A7,$84,$A5,$33,$5F,$E5,$EE,$0C,$C7,$EB,$3D,$3F
	.dc.b	$F3,$A3,$C2,$74,$D8,$CB,$C9,$DE,$9F,$B0,$6E,$9F,$8E,$6C,$FF,$58
	.dc.b	$F0,$3F,$C4,$CF,$EF,$E3,$2F,$87,$0C,$1F,$CB,$67,$71,$FB,$E9,$DE
	.dc.b	$1E,$4C,$3C,$17,$BE,$3F,$87,$18,$BE,$13,$D1,$FE,$15,$87,$49,$D1
	.dc.b	$31,$F4,$57,$3A,$E1,$39,$7F,$80,$E3,$E8,$BE,$57,$C2,$7B,$92,$BF
	.dc.b	$B8,$CE,$A3,$8D,$F4,$AE,$15,$AC,$3D,$31,$82,$57,$03,$D3,$1A,$71
	.dc.b	$3E,$57,$E5,$1A,$3B,$1A,$9E,$6F,$F8,$56,$79,$3C,$27,$96,$3F,$36
	.dc.b	$BF,$1B,$87,$A6,$B5,$1C,$A7,$51,$F7,$6F,$9C,$62,$36,$78,$4D,$1E
	.dc.b	$43,$24,$AE,$D1,$C4,$98,$FC,$FA,$EA,$FE,$08,$98,$FD,$7B,$FA,$18
	.dc.b	$79,$67,$84,$77,$BE,$75,$BC,$7D,$F5,$CA,$BE,$41,$C8,$F4,$1E,$5C
	.dc.b	$3C,$35,$CE,$7A,$CE,$CF,$5D,$FC,$1C,$7D,$73,$D7,$0F,$AD,$1F,$83
	.dc.b	$E1,$7D,$77,$FA,$17,$0F,$F0,$AB,$5F,$EE,$ED,$AF,$F0,$0B,$FE,$61
	.dc.b	$8D,$EB,$F8,$3A,$E6,$D9,$E1,$A7,$F9,$22,$7B,$F8,$9B,$FA,$37,$D6
	.dc.b	$DF,$AE,$99,$7F,$89,$A7,$2D,$9C,$BF,$E0,$A8,$4F,$39,$FC,$5E,$9E
	.dc.b	$0A,$FE,$01,$F4,$38,$33,$A4,$8F,$DF,$46,$5F,$E0,$11,$97,$F6,$ED
	.dc.b	$D7,$F5,$79,$7B,$FE,$B7,$FD,$B3,$5F,$E3,$C6,$BF,$E1,$E6,$CF,$F8
	.dc.b	$A1,$7D,$AF,$1F,$56,$57,$17,$A4,$7E,$B7,$C5,$FF,$2C,$79,$FF,$AF
	.dc.b	$F0,$DF,$57,$FB,$B6,$F4,$6F,$C7,$FF,$58,$65,$FE,$55,$C3,$8B,$FE
	.dc.b	$57,$AD,$B6,$FF,$82,$4E,$A3,$FC,$3D,$0F,$37,$E5,$FC,$8C,$9F,$F2
	.dc.b	$84,$E5,$B3,$97,$F8,$EC,$FE,$F2,$75,$1D,$27,$FA,$3F,$9F,$FE,$C8
	.dc.b	$8F,$5C,$63,$C0,$FE,$97,$5C,$2B,$4B,$D0,$D5,$FF,$B8,$7B,$5F,$0F
	.dc.b	$8D,$3F,$C6,$AD,$8C,$CF,$FA,$CB,$E7,$31,$32,$3F,$D4,$F3,$A7,$FD
	.dc.b	$7F,$3D,$B5,$FF,$44,$7A,$FD,$7C,$47,$D7,$53,$5C,$29,$A6,$70,$FF
	.dc.b	$59,$D7,$07,$FD,$87,$1E,$AF,$55,$F8,$5F,$FA,$AA,$7F,$DB,$5D,$1F
	.dc.b	$E7,$56,$F9,$7B,$9D,$47,$FB,$18,$FF,$78,$38,$E2,$FF,$6C,$D7,$19
	.dc.b	$FF,$74,$36,$3F,$DD,$D3,$BE,$3F,$F5,$DE,$1E,$2F,$4F,$B1,$30,$F1
	.dc.b	$F7,$7C,$DF,$7D,$F8,$55,$24,$FF,$B1,$32,$63,$FE,$C6,$E1,$7B,$4C
	.dc.b	$77,$7E,$9E,$66,$5D,$23,$FD,$B2,$D7,$DD,$7C,$AB,$EA,$5B,$D7,$FC
	.dc.b	$07,$D7,$BC,$FA,$6A,$26,$1F,$EE,$49,$D3,$67,$2F,$CB,$3F,$5B,$F3
	.dc.b	$5F,$8E,$7F,$5F,$F5,$D7,$B4,$D9,$93,$F0,$76,$FC,$CC,$EC,$CC,$75
	.dc.b	$A5,$63,$7F,$CE,$BD,$7D,$9F,$CE,$FC,$87,$FB,$2F,$C2,$FC,$E7,$CD
	.dc.b	$BD,$EA,$6B,$D9,$FF,$0A,$E3,$F4,$33,$61,$FE,$D3,$F9,$5F,$D2,$76
	.dc.b	$BC,$6A,$71,$8E,$FF,$0D,$FA,$FE,$1F,$F7,$ED,$F1,$C3,$D8,$74,$C9
	.dc.b	$7C,$3C,$7F,$3F,$C4,$71,$AF,$0F,$6B,$D2,$F5,$B6,$39,$E2,$AE,$0A
	.dc.b	$5E,$EF,$4C,$25,$F0,$9C,$B6,$7A,$91,$F3,$F6,$3E,$29,$F0,$1F,$33
	.dc.b	$EB,$3F,$D8,$15,$EC,$F0,$5F,$19,$EC,$8C,$35,$FD,$76,$8A,$E3,$18
	.dc.b	$6F,$2F,$D1,$BE,$7C,$1F,$C4,$F0,$E5,$3F,$57,$0E,$5E,$EF,$8C,$98
	.dc.b	$FC,$4C,$79,$F1,$FE,$A7,$4F,$E1,$93,$F2,$3E,$AB,$EF,$CA,$1A,$3E
	.dc.b	$7B,$F5,$D1,$ED,$6B,$EA,$AE,$2F,$E1,$9B,$3C,$7C,$4F,$E5,$55,$C3
	.dc.b	$4F,$45,$70,$D0,$D2,$76,$7F,$27,$F9,$3F,$EC,$35,$EF,$FC,$33,$FC
	.dc.b	$14,$FC,$5F,$84,$FF,$7A,$69,$F4,$5A,$FC,$E9,$DE,$BF,$92,$6C,$7B
	.dc.b	$A7,$75,$CF,$E1,$3F,$4F,$5A,$CC,$D2,$7B,$7F,$CE,$FE,$63,$F8,$65
	.dc.b	$FD,$3F,$A9,$7C,$CF,$F4,$9F,$82,$7D,$55,$3F,$67,$3A,$D7,$F3,$2D
	.dc.b	$7F,$8D,$4F,$07,$F1,$8D,$7F,$13,$1B,$BF,$05,$F4,$24,$FE,$63,$AB
	.dc.b	$F4,$CF,$F0,$7F,$63,$FA,$FF,$CF,$BE,$9F,$F8,$4C,$FE,$D2,$31,$E5
	.dc.b	$FF,$15,$A9,$39,$AD,$BF,$D1,$EB,$6F,$EF,$0F,$F0,$5C,$F0,$7C,$06
	.dc.b	$7C,$C4,$78,$BE,$07,$F1,$33,$FE,$12,$F8,$EF,$2E,$1A,$8F,$05,$7F
	.dc.b	$27,$99,$18,$21,$FC,$8D,$FE,$5D,$F4,$AF,$C0,$3F,$BE,$F0,$3E,$A1
	.dc.b	$F4,$78,$6E,$1E,$9B,$E0,$DB,$E6,$47,$DD,$F8,$13,$E9,$FF,$9E,$7D
	.dc.b	$42,$DC,$9F,$AB,$F0,$7E,$3F,$D4,$FF,$8E,$9B,$F1,$5F,$ED,$E7,$E5
	.dc.b	$DF,$1A,$E1,$CB,$FE,$E2,$8C,$67,$FB,$38,$FF,$72,$E1,$CD,$B8,$73
	.dc.b	$78,$63,$E2,$AE,$07,$EF,$6F,$9E,$3E,$28,$EE,$7F,$01,$C3,$A3,$BA
	.dc.b	$E5,$3E,$66,$DF,$82,$78,$8A,$E2,$AF,$BC,$F9,$31,$DD,$68,$DC,$EA
	.dc.b	$BC,$13,$AC,$7D,$91,$E5,$6B,$D7,$0F,$1B,$C0,$F7,$9E,$38,$DD,$F6
	.dc.b	$9D,$CF,$1A,$FE,$40,$C3,$05,$29,$9B,$FA,$6D,$C9,$CB,$B8,$F8,$11
	.dc.b	$87,$94,$FE,$1E,$74,$D8,$E0,$39,$2B,$4C,$B8,$23,$EE,$F7,$FF,$C8
	.dc.b	$DE,$0F,$D2,$8F,$E7,$53,$F5,$27,$31,$F1,$0C,$9C,$7D,$5F,$EF,$DB
	.dc.b	$FE,$F2,$32,$FE,$FA,$78,$B9,$8F,$27,$85,$9A,$FF,$03,$39,$BF,$E9
	.dc.b	$4E,$70,$BC,$1F,$2E,$7C,$95,$E0,$8C,$5F,$8E,$BD,$4F,$C1,$C3,$93
	.dc.b	$E9,$31,$3F,$63,$1F,$2B,$CA,$76,$78,$E7,$A1,$C7,$0F,$0D,$F6,$6C
	.dc.b	$7E,$9B,$C1,$7D,$4F,$F8,$76,$BF,$C1,$AD,$9F,$F1,$E7,$B0,$F9,$CF
	.dc.b	$F1,$37,$0F,$F7,$3A,$D3,$7D,$8F,$F4,$A5,$FF,$88,$A7,$93,$F9,$0C
	.dc.b	$7B,$CE,$A3,$FC,$5F,$3A,$7F,$2B,$F4,$FF,$E5,$3C,$BF,$1C,$D8,$FC
	.dc.b	$B3,$5F,$CD,$35,$E3,$EE,$F9,$A7,$FB,$F7,$1B,$EA,$1D,$D9,$BF,$F0
	.dc.b	$16,$1D,$74,$FF,$21,$5E,$EB,$54,$AF,$E5,$78,$7F,$41,$6C,$7F,$45
	.dc.b	$B7,$93,$C2,$73,$1F,$EB,$77,$73,$E8,$EE,$F8,$4F,$56,$C6,$A3,$E1
	.dc.b	$C7,$9F,$1F,$57,$C5,$F5,$8C,$67,$FE,$BA,$CB,$FC,$8B,$F5,$1F,$E4
	.dc.b	$E6,$FA,$BE,$FF,$5B,$FC,$77,$AF,$C1,$E2,$5E,$F4,$FF,$1B,$B5,$FB
	.dc.b	$0D,$7F,$CB,$AD,$7F,$CC,$ED,$8E,$77,$F3,$7D,$27,$89,$F1,$32,$F1
	.dc.b	$9F,$DF,$B1,$F1,$7D,$47,$A7,$2F,$F2,$3B,$78,$7F,$CE,$9A,$3C,$DF
	.dc.b	$F0,$7E,$1F,$E6,$2A,$D3,$FE,$70,$79,$5F,$E3,$DA,$E7,$83,$F2,$1A
	.dc.b	$A7,$33,$FE,$80,$79,$BF,$E2,$7F,$A8,$F2,$D7,$AF,$BD,$1F,$27,$EA
	.dc.b	$3F,$17,$8F,$F9,$D1,$B7,$FE,$B0,$F4,$BF,$B4,$4E,$79,$F8,$1E,$2A
	.dc.b	$DD,$7E,$95,$AF,$FA,$29,$AF,$FA,$51,$EB,$B7,$F9,$9F,$EA,$1E,$EF
	.dc.b	$4C,$3F,$CE,$91,$87,$FD,$0F,$F5,$0F,$9C,$FC,$03,$63,$FC,$FF,$ED
	.dc.b	$7F,$AD,$1D,$CE,$8F,$8F,$3B,$CB,$FA,$47,$0A,$31,$1F,$BD,$6B,$FE
	.dc.b	$A2,$78,$65,$FE,$7D,$F7,$5F,$51,$D7,$1F,$F4,$3B,$F9,$3F,$A8,$7C
	.dc.b	$F5,$3F,$31,$A7,$D0,$AF,$F5,$0C,$E9,$B3,$F7,$23,$19,$79,$9A,$FF
	.dc.b	$11,$6B,$FE,$B6,$F8,$37,$D4,$7E,$11,$B7,$FD,$8F,$5F,$F5,$74,$E4
	.dc.b	$FC,$4F,$C3,$FF,$A9,$27,$4D,$BF,$8A,$D7,$F9,$03,$5F,$F6,$2F,$D4
	.dc.b	$3E,$63,$F2,$6D,$AF,$03,$67,$E7,$B5,$FE,$54,$FF,$5E,$6F,$9F,$C8
	.dc.b	$93,$FD,$D7,$E5,$5F,$4F,$FC,$69,$B1,$FE,$59,$5D,$79,$99,$3C,$5F
	.dc.b	$E2,$B3,$C2,$FF,$C0,$CD,$7F,$C1,$4E,$78,$FF,$DD,$50,$87,$D0,$74
	.dc.b	$8F,$F2,$3D,$FF,$84,$A7,$2F,$FB,$43,$E4,$9F,$4B,$FC,$27,$E9,$FF
	.dc.b	$BD,$E3,$FC,$0A,$73,$5F,$ED,$19,$CB,$FE,$BB,$F0,$D3,$AC,$7B,$9F
	.dc.b	$EB,$EA,$DA,$9D,$E1,$DF,$5F,$E3,$1F,$4E,$F8,$8A,$7F,$29,$C5,$F0
	.dc.b	$F8,$CF,$FD,$C5,$5B,$6D,$6E,$75,$97,$8B,$D8,$FF,$7A,$FC,$F7,$F5
	.dc.b	$16,$DF,$FB,$4F,$63,$FE,$D5,$F2,$9F,$4F,$FB,$BC,$39,$7C,$7B,$1E
	.dc.b	$4D,$BF,$F1,$D6,$3D,$2B,$FC,$7C,$D7,$FC,$15,$3B,$72,$DC,$7C,$31
	.dc.b	$FC,$E3,$0F,$C9,$46,$2B,$A3,$AC,$7C,$BC,$FF,$D3,$7E,$77,$CD,$3F
	.dc.b	$EF,$BF,$15,$F4,$39,$7F,$A2,$18,$D3,$E0,$1F,$66,$B8,$BF,$EF,$5F
	.dc.b	$3F,$F8,$ED,$71,$8D,$E9,$FE,$F8,$BE,$4F,$DE,$A7,$D1,$3E,$5B,$FA
	.dc.b	$93,$6F,$8F,$3B,$2F,$E5,$71,$DE,$75,$87,$36,$E7,$F1,$31,$EB,$8F
	.dc.b	$BF,$CB,$EA,$FE,$B7,$A2,$63,$FA,$DF,$9A,$B9,$7E,$48,$FD,$24,$FE
	.dc.b	$B0,$93,$FB,$76,$C7,$F9,$97,$B3,$F4,$DF,$2B,$F5,$3D,$FF,$D5,$C8
	.dc.b	$E7,$7F,$E6,$2A,$DB,$63,$F3,$DB,$7F,$15,$AE,$3A,$7E,$D3,$E6,$3A
	.dc.b	$61,$F4,$7E,$3A,$E0,$AD,$63,$CF,$87,$E4,$56,$A7,$35,$E7,$3B,$C6
	.dc.b	$6D,$87,$4E,$5F,$CD,$FC,$C5,$E8,$37,$A7,$0C,$D3,$D5,$5C,$1B,$4A
	.dc.b	$D6,$5E,$13,$F8,$87,$67,$A9,$B5,$F0,$DF,$E2,$9F,$4F,$9F,$0B,$F7
	.dc.b	$11,$C5,$B3,$FC,$72,$BF,$46,$D3,$E8,$16,$7F,$CB,$F3,$CB,$83,$EA
	.dc.b	$C7,$F2,$1F,$A0,$ED,$7D,$DF,$84,$7F,$6B,$9F,$B7,$9A,$3F,$99,$E1
	.dc.b	$F0,$1D,$6D,$FC,$0D,$CE,$1F,$3E,$B8,$7A,$3D,$3C,$6E,$AD,$C3,$E4
	.dc.b	$FB,$3F,$E6,$9C,$BB,$B7,$2F,$E1,$CF,$F9,$6A,$33,$F0,$1F,$5E,$35
	.dc.b	$97,$F8,$E9,$B9,$7F,$97,$BD,$ED,$57,$CF,$FA,$27,$CD,$3E,$9F,$E8
	.dc.b	$46,$1F,$A5,$C3,$FC,$AB,$5C,$D6,$F3,$FB,$57,$B6,$C6,$74,$F9,$B7
	.dc.b	$C6,$73,$87,$22,$3F,$6F,$C7,$F2,$5F,$1D,$BF,$DC,$38,$F5,$C7,$EB
	.dc.b	$B6,$3F,$AA,$57,$0C,$39,$D6,$AF,$96,$1F,$B1,$C7,$C5,$39,$AD,$E1
	.dc.b	$DA,$F7,$7F,$D8,$F0,$E8,$FA,$09,$7E,$C8,$CB,$CE,$33,$5E,$4C,$BA
	.dc.b	$C7,$F0,$1C,$BD,$0D,$C3,$BE,$5E,$4B,$F1,$65,$C2,$97,$C1,$F2,$61
	.dc.b	$98,$C1,$9B,$E5,$3D,$6B,$58,$76,$CB,$CF,$8F,$6A,$F3,$9F,$16,$32
	.dc.b	$F1,$32,$78,$0B,$87,$82,$BA,$63,$DB,$0F,$46,$3E,$2B,$F4,$D7,$CC
	.dc.b	$BE,$15,$F8,$CC,$3B,$D6,$A3,$8C,$62,$77,$3F,$85,$C3,$B3,$E4,$C3
	.dc.b	$B4,$EB,$2F,$A4,$7D,$5A,$DB,$8C,$BE,$A5,$FA,$B2,$F2,$1E,$A7,$19
	.dc.b	$78,$AF,$BC,$7F,$39,$3E,$21,$4F,$41,$4A,$67,$0F,$F3,$2E,$F9,$8F
	.dc.b	$B7,$E9,$FF,$9F,$5E,$13,$A6,$C6,$5E,$4E,$F2,$FD,$0B,$72,$FD,$2B
	.dc.b	$67,$FA,$77,$87,$FE,$23,$7F,$77,$19,$7C,$38,$60,$FE,$5D,$3B,$8F
	.dc.b	$DD,$CE,$F0,$F2,$61,$E0,$BD,$F2,$7C,$38,$C5,$F0,$9E,$8F,$F3,$DC
	.dc.b	$3A,$4E,$89,$8F,$A2,$B9,$D7,$09,$CB,$FA,$7C,$7D,$17,$CA,$F8,$4F
	.dc.b	$72,$57,$F3,$A9,$D4,$71,$BE,$95,$C2,$B5,$87,$A6,$30,$4A,$E0,$7A
	.dc.b	$63,$4E,$27,$CA,$FC,$A3,$47,$63,$53,$CD,$FF,$0E,$CF,$27,$84,$F2
	.dc.b	$C7,$E6,$D7,$E3,$70,$F4,$D6,$A3,$94,$EA,$3E,$ED,$F3,$8C,$46,$CF
	.dc.b	$09,$A3,$C8,$64,$95,$DA,$38,$93,$1F,$9F,$5D,$5F,$C1,$13,$1F,$AF
	.dc.b	$7F,$43,$0F,$2C,$F0,$8E,$F7,$CE,$B7,$8F,$BE,$B9,$57,$C8,$39,$1E
	.dc.b	$83,$CB,$87,$86,$B9,$CF,$59,$D9,$EB,$BF,$83,$8F,$AE,$7A,$E1,$F5
	.dc.b	$A3,$F0,$7C,$2F,$A7,$FF,$41,$E1,$FE,$16,$6B,$FD,$D5,$B5,$FD,$F9
	.dc.b	$FF,$31,$C6,$F8,$3F,$07,$5C,$DB,$3C,$34,$FF,$41,$3C,$2B,$FC,$01
	.dc.b	$F3,$AF,$A6,$E2,$BE,$9E,$98,$89,$3F,$E2,$A9,$D3,$67,$2F,$EB,$70
	.dc.b	$FC,$66,$6F,$9E,$A6,$22,$4F,$F8,$88,$FC,$73,$FC,$6F,$87,$F8,$D5
	.dc.b	$7C,$26,$C6,$67,$67,$F0,$67,$9B,$FB,$DF,$2B,$E1,$FD,$2B,$E5,$29
	.dc.b	$27,$FC,$87,$0F,$A7,$C4,$7F,$88,$E3,$83,$A9,$E7,$3F,$B0,$F1,$7F
	.dc.b	$D0,$1E,$87,$EC,$BC,$67,$CD,$FE,$F9,$BD,$3B,$F2,$3F,$D5,$D9,$7F
	.dc.b	$98,$5A,$FF,$97,$E7,$6F,$F9,$BD,$AF,$F9,$8F,$0E,$EF,$F9,$46,$DE
	.dc.b	$7F,$CE,$F8,$E9,$39,$8F,$DE,$4E,$A3,$F9,$F3,$5F,$E8,$8D,$7F,$A4
	.dc.b	$35,$FE,$98,$EF,$D0,$FF,$79,$B7,$87,$35,$FD,$46,$73,$E8,$D9,$FF
	.dc.b	$58,$76,$BE,$1B,$18,$FF,$A0,$E3,$0F,$FA,$CB,$4F,$F0,$5C,$EB,$4F
	.dc.b	$F5,$67,$CD,$70,$9C,$D6,$EF,$8B,$FE,$B7,$6E,$5F,$ED,$0B,$E3,$3F
	.dc.b	$EB,$6D,$9E,$BF,$3F,$11,$F4,$F4,$F0,$4E,$AB,$85,$65,$F3,$57,$07
	.dc.b	$27,$FB,$2B,$D6,$7D,$3B,$D3,$F7,$F8,$6E,$AF,$F3,$A5,$FF,$B4,$23
	.dc.b	$1E,$6F,$C4,$C3,$8D,$DC,$39,$5F,$FB,$C9,$AF,$FB,$C9,$AF,$F8,$06
	.dc.b	$78,$BB,$9E,$44,$7C,$18,$7E,$CD,$FA,$1F,$4C,$FA,$7F,$9C,$D8,$CC
	.dc.b	$7D,$3F,$6F,$84,$7F,$B9,$67,$4D,$9E,$26,$FC,$2E,$1E,$C7,$D9,$FA
	.dc.b	$77,$9F,$4C,$DF,$FB,$AB,$15,$66,$3F,$09,$5C,$1D,$47,$E2,$9F,$D7
	.dc.b	$FC,$C5,$EE,$7D,$3D,$31,$12,$7F,$42,$FF,$87,$DE,$26,$08,$7F,$BB
	.dc.b	$65,$1C,$63,$1E,$CF,$E6,$BE,$13,$FD,$9F,$E1,$FF,$8A,$5B,$3F,$E2
	.dc.b	$F6,$BF,$E3,$36,$BF,$E3,$77,$FD,$AF,$F2,$3C,$7D,$B3,$49,$3C,$FE
	.dc.b	$13,$F6,$8A,$91,$86,$F8,$BE,$38,$C7,$C3,$48,$F3,$36,$BC,$DC,$DF
	.dc.b	$EE,$16,$FB,$9E,$B8,$C4,$7A,$5A,$FA,$5C,$F8,$2F,$19,$3E,$59,$CC
	.dc.b	$63,$2F,$D2,$5F,$EF,$7E,$21,$EF,$FC,$BF,$A7,$FF,$5F,$E3,$FC,$0F
	.dc.b	$E2,$FE,$1B,$77,$D7,$B9,$23,$E7,$57,$07,$E7,$35,$F9,$CD,$7C,$4E
	.dc.b	$F5,$F4,$47,$CB,$8C,$69,$AA,$F3,$78,$9F,$7F,$E9,$FF,$CA,$58,$70
	.dc.b	$F2,$9F,$37,$FA,$48,$C3,$70,$F8,$15,$C1,$BC,$DF,$87,$8F,$F2,$E6
	.dc.b	$7F,$83,$6B,$A3,$F1,$4F,$17,$F0,$0D,$E1,$39,$F2,$9F,$23,$FD,$62
	.dc.b	$99,$7B,$AD,$F4,$FF,$E0,$C8,$F5,$7C,$1B,$E0,$FF,$97,$36,$FE,$DC
	.dc.b	$F1,$AE,$1A,$7D,$B7,$FC,$2B,$3F,$9B,$6B,$9F,$63,$F4,$91,$88,$FC
	.dc.b	$C3,$C9,$B1,$FA,$B8,$C6,$6F,$95,$FE,$1B,$87,$A7,$E9,$9F,$4D,$FD
	.dc.b	$0F,$CF,$7D,$3D,$3F,$69,$3A,$CF,$D3,$A7,$E4,$63,$11,$FB,$36,$BF
	.dc.b	$B3,$6B,$BB,$FD,$E4,$70,$C3,$C5,$1D,$8F,$EA,$DD,$4F,$7F,$F8,$97
	.dc.b	$AD,$FD,$7B,$E6,$DF,$4F,$FC,$2B,$2F,$18,$8C,$F8,$FC,$23,$27,$F8
	.dc.b	$2A,$BF,$8B,$79,$06,$92,$3F,$BD,$BC,$5F,$E3,$0F,$87,$E1,$7F,$11
	.dc.b	$1F,$E1,$4F,$8E,$F2,$63,$CE,$3C,$78,$F9,$AF,$F9,$55,$49,$FF,$03
	.dc.b	$BE,$22,$3F,$EC,$7F,$A3,$7D,$37,$F2,$B6,$C6,$63,$04,$7D,$B1,$A7
	.dc.b	$F2,$BC,$3F,$ED,$98,$C7,$07,$ED,$25,$F4,$CF,$A5,$FE,$81,$F4,$AB
	.dc.b	$FC,$ED,$7E,$CB,$83,$FD,$C5,$F4,$DF,$A1,$F0,$BE,$84,$A3,$13,$FE
	.dc.b	$E7,$BE,$59,$7F,$18,$BD,$C6,$2F,$A1,$FC,$72,$F8,$B7,$E1,$BF,$DE
	.dc.b	$15,$C3,$1F,$1B,$70,$E9,$8F,$92,$B6,$DB,$C9,$FC,$A3,$1F,$05,$F1
	.dc.b	$8F,$09,$E9,$AE,$2F,$84,$E2,$27,$8D,$7B,$70,$F8,$BD,$DF,$0B,$2F
	.dc.b	$E0,$8D,$C3,$F9,$BE,$08,$F0,$1E,$1C,$75,$1D,$AB,$E0,$BD,$EF,$DF
	.dc.b	$87,$7C,$B5,$8F,$9A,$BD,$07,$9B,$2F,$2B,$EF,$BF,$B2,$F8,$8E,$95
	.dc.b	$5D,$35,$C9,$99,$FC,$27,$0E,$5B,$19,$D3,$F5,$CD,$9F,$F0,$02,$E3
	.dc.b	$A4,$7F,$C0,$8D,$D3,$F1,$3E,$DF,$F9,$23,$83,$E7,$D7,$F5,$EA,$FA
	.dc.b	$33,$98,$FA,$E6,$4F,$B2,$E6,$BF,$27,$19,$7A,$F8,$DF,$D4,$9C,$C7
	.dc.b	$94,$E3,$19,$DB,$E2,$5F,$82,$31,$3F,$26,$31,$87,$9A,$39,$5F,$68
	.dc.b	$C4,$EA,$7C,$E6,$AF,$FC,$95,$7D,$23,$F0,$B3,$BB,$F2,$B6,$BC,$73
	.dc.b	$E6,$BE,$AD,$3F,$53,$E1,$3F,$0E,$FF,$5F,$6D,$FD,$4D,$B5,$FE,$08
	.dc.b	$EE,$7E,$1C,$FA,$9F,$EE,$B7,$BC,$6D,$FF,$82,$F0,$E3,$0F,$25,$F7
	.dc.b	$7F,$EB,$BA,$FE,$5F,$9F,$86,$B8,$9B,$9D,$3F,$C4,$87,$88,$FC,$3F
	.dc.b	$FC,$AF,$7F,$E9,$51,$F2,$5C,$A7,$5E,$EF,$4C,$BF,$C4,$53,$9B,$E3
	.dc.b	$87,$59,$CE,$1F,$D0,$B0,$E2,$F5,$BE,$9F,$80,$BE,$87,$A7,$E0,$4F
	.dc.b	$B1,$CC,$61,$B7,$FE,$A9,$9D,$3D,$6B,$93,$99,$DE,$5E,$3A,$DB,$F0
	.dc.b	$71,$5F,$08,$D5,$7B,$FC,$77,$C7,$FB,$F9,$6E,$5E,$B9,$B7,$EF,$9B
	.dc.b	$C7,$FE,$B3,$9B,$5F,$E4,$6D,$95,$8B,$FA,$35,$67,$38,$78,$4F,$01
	.dc.b	$6F,$FD,$91,$7C,$73,$FF,$04,$7C,$5F,$D3,$F0,$BF,$15,$5F,$3F,$CD
	.dc.b	$FF,$5F,$BF,$EC,$F8,$CE,$7F,$06,$72,$FD,$78,$CE,$9F,$EC,$B9,$CB
	.dc.b	$FE,$28,$E7,$FC,$14,$62,$FD,$95,$C1,$FF,$31,$E3,$E4,$7C,$15,$F8
	.dc.b	$9F,$47,$FD,$8D,$1F,$EE,$1F,$A5,$7C,$7E,$7D,$CF,$C6,$46,$27,$F1
	.dc.b	$15,$C3,$4F,$C6,$4E,$5F,$B5,$19,$D3,$FC,$AD,$39,$7F,$DC,$91,$9A
	.dc.b	$FA,$53,$97,$CF,$19,$AF,$D6,$4E,$5F,$D2,$FB,$7F,$EA,$4A,$FF,$3B
	.dc.b	$7C,$87,$E8,$E3,$13,$F9,$FF,$71,$39,$F0,$3F,$4D,$18,$BF,$D1,$5F
	.dc.b	$0C,$BF,$CD,$93,$9C,$3F,$29,$39,$AF,$C4,$4E,$6B,$FD,$FD,$39,$AF
	.dc.b	$E0,$53,$9A,$FC,$77,$E2,$0F,$9A,$FD,$B3,$63,$F8,$2F,$E2,$2F,$8C
	.dc.b	$CD,$49,$FE,$13,$F0,$5F,$E9,$67,$FC,$73,$D2,$C3,$96,$5F,$C2,$7C
	.dc.b	$35,$EE,$30,$DD,$7F,$D1,$1E,$43,$E8,$FF,$89,$DF,$0E,$DF,$E1,$53
	.dc.b	$9C,$FF,$8C,$FC,$2F,$E2,$F2,$FE,$3D,$9F,$F2,$0C,$3F,$D8,$3D,$1E
	.dc.b	$C9,$97,$E6,$AF,$03,$F0,$A3,$E0,$E3,$FC,$5F,$EC,$7F,$90,$E9,$FE
	.dc.b	$9B,$6D,$7C,$89,$D7,$CB,$7F,$28,$FC,$07,$FC,$B1,$F6,$B7,$D0,$7C
	.dc.b	$E5,$8C,$37,$0F,$F2,$FD,$7F,$97,$E3,$19,$FF,$2F,$8C,$4E,$AD,$7C
	.dc.b	$A3,$0D,$AF,$E5,$5F,$4D,$F9,$3E,$43,$F9,$5F,$01,$F2,$2F,$5F,$F9
	.dc.b	$A4,$EC,$9E,$17,$F3,$99,$D6,$5F,$72,$79,$43,$E6,$FF,$9E,$7D,$33
	.dc.b	$E6,$38,$46,$1B,$7F,$84,$9D,$3F,$CF,$6B,$93,$99,$FE,$A3,$A7,$A3
	.dc.b	$E7,$AE,$27,$FA,$AF,$6B,$F1,$17,$CA,$EA,$66,$3F,$E8,$A1,$F8,$0B
	.dc.b	$E5,$33,$97,$F5,$B9,$CC,$63,$AB,$FD,$BF,$18,$3F,$D1,$F9,$F8,$5E
	.dc.b	$97,$CD,$FF,$4B,$F9,$CF,$C4,$B5,$8F,$F6,$78,$C7,$AF,$FD,$B1,$D6
	.dc.b	$76,$BF,$B0,$E5,$FE,$EE,$9C,$E1,$D2,$72,$FF,$80,$5E,$B3,$B7,$38
	.dc.b	$7F,$74,$F9,$EF,$23,$FD,$AF,$E8,$5F,$4F,$FD,$CE,$73,$7F,$EF,$59
	.dc.b	$CE,$1F,$DB,$E7,$2F,$F5,$18,$CD,$FF,$61,$9C,$BF,$E0,$B8,$CD,$29
	.dc.b	$2E,$CF,$F5,$D6,$7F,$A3,$D3,$F4,$8D,$AF,$CE,$36,$BF,$41,$D9,$EE
	.dc.b	$9E,$77,$FE,$C3,$BE,$31,$FA,$36,$D7,$07,$DD,$F5,$FF,$E0,$FF,$A5
	.dc.b	$FF,$62,$57,$B6,$31,$F1,$5A,$8F,$E2,$F3,$AB,$F7,$56,$A7,$EE,$FE
	.dc.b	$27,$FE,$1C,$6F,$DE,$7D,$B8,$C5,$7D,$F5,$19,$F7,$B4,$DA,$E2,$7E
	.dc.b	$EF,$EA,$7F,$84,$6E,$FA,$9F,$E1,$11,$86,$CF,$F0,$69,$D3,$C4,$C4
	.dc.b	$6B,$1F,$E1,$13,$A7,$E5,$D7,$3A,$FF,$16,$7E,$28,$FB,$DF,$F1,$A7
	.dc.b	$E2,$9F,$E9,$D9,$CC,$63,$DC,$FF,$1D,$56,$E7,$F9,$ED,$6F,$0F,$F1
	.dc.b	$24,$E5,$FF,$1D,$46,$71,$FC,$24,$7F,$92,$27,$FA,$55,$72,$F0,$FA
	.dc.b	$FA,$9E,$26,$C6,$4F,$3C,$65,$FD,$7F,$CD,$7F,$92,$3E,$25,$C7,$FE
	.dc.b	$65,$8C,$37,$6F,$0C,$EB,$2F,$F1,$5D,$69,$BC,$BA,$BE,$4D,$8D,$DD
	.dc.b	$ED,$5F,$16,$D6,$63,$02,$71,$13,$1F,$F5,$26,$1D,$9B,$87,$66,$E1
	.dc.b	$D1,$CC,$3E,$A9,$F3,$DF,$DD,$E3,$FD,$6D,$7F,$ED,$69,$CF,$8B,$C3
	.dc.b	$EA,$9F,$3B,$EB,$BE,$58,$71,$DB,$E2,$46,$2F,$BE,$1C,$A3,$A9,$FB
	.dc.b	$CC,$39,$E5,$AC,$39,$65,$E1,$6D,$7F,$02,$BE,$58,$F6,$C3,$8C,$F4
	.dc.b	$3C,$05,$9D,$DF,$3C,$7D,$59,$7A,$67,$35,$D7,$1E,$EF,$2C,$7B,$56
	.dc.b	$E7,$35,$C2,$B5,$87,$FB,$F3,$1F,$1E,$1F,$0A,$B4,$4B,$EB,$3A,$AF
	.dc.b	$25,$F9,$E7,$A9,$AC,$3C,$18,$FC,$52,$4F,$BE,$F9,$3E,$98,$CE,$3E
	.dc.b	$1B,$E5,$96,$31,$EF,$39,$AF,$87,$39,$BE,$47,$8E,$32,$D2,$4F,$17
	.dc.b	$C5,$8F,$99,$CD,$F6,$3E,$24,$66,$78,$12,$33,$7E,$1A,$F9,$67,$8A
	.dc.b	$7E,$8C,$67,$2F,$95,$87,$92,$79,$56,$EF,$C0,$FE,$37,$0F,$04,$62
	.dc.b	$3A,$8C,$BE,$4B,$8C,$3A,$BE,$CC,$7C,$19,$7E,$96,$3F,$A9,$9F,$18
	.dc.b	$A7,$E4,$0B,$3F,$9F,$6F,$47,$F3,$37,$67,$EC,$5C,$3A,$E7,$CB,$96
	.dc.b	$CF,$FB,$89,$FD,$5C,$F3,$CF,$F1,$CD,$CB,$CB,$DE,$F9,$FF,$3D,$6A
	.dc.b	$F8,$93,$1F,$D4,$56,$DC,$4E,$DF,$D5,$61,$D5,$FE,$E7,$3B,$73,$5C
	.dc.b	$30,$FE,$11,$5B,$95,$6E,$FA,$12,$B7,$7F,$C5,$27,$51,$8A,$EA,$4A
	.dc.b	$FE,$1B,$3A,$8F,$E0,$97,$DF,$1F,$8D,$1B,$BF,$87,$8F,$C2,$7A,$61
	.dc.b	$E1,$9E,$38,$F8,$6B,$51,$83,$91,$67,$C3,$1A,$8F,$AC,$68,$B1,$A7
	.dc.b	$E2,$1C,$6B,$C2,$79,$1C,$9C,$1C,$BF,$3A,$7B,$5F,$B2,$BD,$13,$B3
	.dc.b	$47,$28,$D5,$72,$8F,$A3,$5C,$B0,$F8,$C4,$C3,$E0,$4F,$18,$C3,$F8
	.dc.b	$49,$E1,$39,$37,$1E,$FB,$EE,$FB,$4F,$19,$9A,$E7,$87,$C3,$8C,$57
	.dc.b	$C0,$C7,$ED,$6E,$FA,$CF,$E9,$B5,$FD,$61,$AF,$F4,$88,$D6,$FF,$59
	.dc.b	$E1,$CB,$6B,$78,$FF,$75,$9D,$DF,$3B,$D5,$ED,$91,$93,$C1,$18,$7C
	.dc.b	$17,$D3,$5F,$F2,$9D,$F4,$9C,$D7,$E2,$3C,$F7,$C6,$3E,$5F,$F8,$3F
	.dc.b	$57,$F9,$CF,$F1,$57,$D4,$E3,$97,$9C,$63,$2F,$F2,$FD,$7F,$87,$B1
	.dc.b	$E7,$86,$E5,$18,$7F,$D0,$F5,$C5,$FF,$43,$36,$B9,$63,$E3,$9F,$F1
	.dc.b	$BE,$1C,$FD,$47,$E1,$0F,$8D,$C3,$6B,$64,$18,$7F,$8C,$AB,$8D,$70
	.dc.b	$3C,$D9,$3B,$FF,$C6,$F7,$C3,$1F,$E5,$3F,$87,$7F,$91,$E3,$DF,$18
	.dc.b	$F0,$3D,$D0,$94,$61,$E1,$1C,$CF,$F4,$2B,$8C,$BB,$BC,$23,$A9,$A9
	.dc.b	$E5,$E1,$C8,$C9,$DE,$3B,$19,$7F,$D6,$1F,$8B,$3E,$77,$E9,$72,$23
	.dc.b	$FD,$85,$3F,$EB,$86,$BC,$27,$71,$C6,$35,$E9,$60,$CC,$FC,$8F,$3B
	.dc.b	$FC,$87,$3E,$8E,$07,$CF,$BE,$3B,$EC,$36,$7F,$D5,$DE,$3F,$13,$33
	.dc.b	$23,$27,$BB,$CD,$7B,$FF,$39,$79,$D4,$C4,$49,$E1,$19,$79,$46,$5F
	.dc.b	$A4,$27,$9B,$C0,$D9,$FE,$C3,$FA,$3F,$C8,$78,$BF,$1B,$93,$F1,$4D
	.dc.b	$E5,$FF,$5F,$F3,$B7,$5F,$C5,$38,$F0,$CC,$7A,$1F,$ED,$DF,$17,$EA
	.dc.b	$FA,$9F,$EA,$88,$FF,$70,$79,$F8,$99,$18,$21,$FE,$ED,$94,$7E,$4B
	.dc.b	$CE,$F2,$4F,$FB,$4B,$67,$DA,$FC,$93,$58,$C9,$F9,$D1,$97,$F1,$B5
	.dc.b	$C9,$34,$93,$F4,$F8,$FF,$53,$ED,$BD,$BF,$DB,$75,$BE,$9E,$9A,$89
	.dc.b	$8F,$E1,$E7,$4D,$9C,$BF,$A7,$E9,$7D,$6B,$D8,$FE,$04,$D9,$FC,$83
	.dc.b	$5F,$D7,$72,$32,$FF,$3E,$F3,$FF,$01,$AD,$36,$F9,$9E,$AF,$27,$F1
	.dc.b	$DB,$7E,$3F,$EA,$A6,$5F,$93,$8C,$BF,$95,$8C,$BF,$8C,$E2,$F9,$04
	.dc.b	$8C,$E1,$FC,$32,$3E,$AE,$8F,$BF,$7D,$3F,$F1,$D6,$E1,$FA,$B6,$BF
	.dc.b	$0F,$E2,$1F,$4F,$FC,$8F,$E3,$5F,$39,$FC,$96,$73,$5F,$B3,$6B,$F2
	.dc.b	$E7,$6F,$F0,$EF,$8D,$7C,$77,$F9,$13,$5F,$E4,$D5,$AA,$E1,$87,$F3
	.dc.b	$BB,$DC,$5F,$90,$E1,$E0,$A1,$B3,$F0,$47,$C6,$7F,$39,$6C,$F8,$6F
	.dc.b	$8B,$C3,$27,$AB,$D6,$21,$FC,$36,$32,$D8,$F8,$44,$7D,$9F,$8B,$FA
	.dc.b	$75,$FE,$A9,$F2,$BF,$E6,$37,$FA,$17,$CA,$FF,$13,$AF,$C7,$36,$37
	.dc.b	$F0,$3F,$D1,$2B,$86,$DE,$DD,$97,$B8,$C3,$70,$FE,$BF,$5C,$1B,$7F
	.dc.b	$D6,$67,$57,$CB,$CC,$ED,$3F,$D4,$BE,$65,$F6,$DF,$D7,$FE,$8F,$F5
	.dc.b	$98,$FF,$2D,$6B,$F9,$E9,$DB,$FD,$23,$C1,$46,$23,$F5,$53,$C3,$5F
	.dc.b	$EC,$4D,$D7,$F8,$CD,$6E,$7F,$BD,$36,$3E,$04,$EE,$BE,$EB,$6F,$BE
	.dc.b	$1F,$DA,$BE,$71,$F8,$B7,$F7,$7F,$89,$FC,$8E,$7F,$DC,$5B,$97,$F2
	.dc.b	$8A,$DE,$5E,$96,$BF,$C0,$67,$6F,$E3,$1B,$8A,$77,$7F,$E1,$16,$DF
	.dc.b	$D7,$AE,$3B,$1E,$33,$AC,$3F,$C2,$BE,$8B,$ED,$BF,$C1,$DE,$BB,$EC
	.dc.b	$A9,$FE,$06,$9D,$57,$CF,$6B,$FC,$5E,$78,$3F,$D8,$DA,$FF,$14,$8D
	.dc.b	$E7,$FE,$27,$6E,$7F,$8D,$AD,$CF,$F8,$5D,$B1,$FD,$3A,$77,$5F,$E2
	.dc.b	$CF,$A2,$FF,$14,$4F,$E5,$7E,$15,$F3,$7F,$E2,$76,$D7,$E1,$6B,$6F
	.dc.b	$1F,$84,$D5,$48,$D4,$7F,$89,$27,$7B,$AF,$9D,$7F,$35,$6E,$7F,$E4
	.dc.b	$1A,$EB,$C1,$FE,$4C,$F2,$BF,$C7,$1F,$87,$7F,$7A,$C3,$FC,$0C,$D7
	.dc.b	$F9,$BF,$C1,$FF,$34,$D7,$FB,$5B,$63,$F8,$5C,$EF,$1F,$F1,$D3,$6B
	.dc.b	$77,$FC,$57,$D7,$7C,$6F,$F9,$1F,$C4,$7D,$15,$35,$13,$2F,$E4,$55
	.dc.b	$C1,$B1,$FE,$28,$FA,$77,$D2,$7F,$8F,$BE,$A9,$F5,$7F,$E4,$CB,$77
	.dc.b	$3E,$93,$FD,$DF,$F5,$8F,$A0,$FF,$78,$CE,$6B,$F9,$34,$E5,$FE,$51
	.dc.b	$19,$58,$76,$E2,$CE,$1F,$EF,$A8,$C4,$F8,$23,$BE,$1D,$9B,$1C,$0F
	.dc.b	$E6,$93,$C2,$FB,$36,$77,$3C,$CF,$2D,$73,$78,$E3,$EF,$BE,$2D,$AE
	.dc.b	$31,$E9,$C7,$D9,$5C,$CF,$5D,$6D,$E7,$8F,$95,$4F,$0A,$E5,$87,$86
	.dc.b	$78,$D6,$F1,$F4,$47,$98,$E5,$7C,$71,$F8,$D8,$F7,$9E,$27,$8E,$74
	.dc.b	$F1,$3C,$D5,$D1,$D9,$FD,$6C,$F3,$94,$A5,$7E,$9B,$7E,$B1,$F3,$5B
	.dc.b	$8F,$A3,$18,$71,$28,$C4,$71,$9E,$0F,$87,$A2,$BF,$DB,$5A,$FE,$CC
	.dc.b	$CB,$83,$E6,$7D,$63,$CC,$73,$7C,$FA,$E4,$7D,$59,$D4,$66,$3A,$6F
	.dc.b	$F9,$39,$CC,$7D,$B3,$8C,$66,$BA,$C6,$4E,$3A,$FC,$9A,$E3,$5F,$9D
	.dc.b	$9E,$33,$9A,$E3,$39,$9E,$27,$F8,$2E,$B8,$D7,$CB,$C3,$DB,$7C,$23
	.dc.b	$15,$EB,$8C,$3F,$5E,$77,$87,$B5,$B1,$E7,$6E,$1E,$16,$DF,$48,$F7
	.dc.b	$57,$EA,$F9,$5F,$59,$FE,$1F,$8F,$F1,$53,$70,$FF,$5F,$78,$0E,$4F
	.dc.b	$E6,$75,$A9,$F8,$FC,$5F,$23,$0E,$71,$89,$FF,$15,$46,$1F,$F6,$1D
	.dc.b	$F3,$EF,$D5,$2B,$4F,$F2,$AE,$6C,$F7,$7E,$03,$E0,$F7,$DD,$FE,$FE
	.dc.b	$9C,$E5,$FE,$88,$3F,$C1,$1D,$DF,$CA,$1F,$F5,$D1,$9C,$3A,$38,$BE
	.dc.b	$7D,$FF,$E5,$08,$FF,$60,$7D,$5F,$FB,$CD,$FE,$9D,$E9,$F8,$A3,$17
	.dc.b	$FE,$74,$AE,$0F,$F9,$F6,$33,$BF,$F4,$9A,$E1,$F2,$1D,$9B,$87,$43
	.dc.b	$BC,$7B,$3C,$4F,$E6,$B1,$FD,$BF,$BF,$07,$C2,$E2,$CE,$5F,$DC,$63
	.dc.b	$06,$A7,$81,$FD,$ED,$CB,$4C,$C7,$FA,$33,$C7,$FE,$97,$1F,$ED,$FF
	.dc.b	$23,$EE,$CE,$7C,$3F,$A3,$18,$D3,$FC,$29,$18,$E5,$FA,$78,$72,$BF
	.dc.b	$F0,$F4,$62,$75,$87,$F8,$8D,$B3,$FE,$3F,$F3,$7E,$EF,$D5,$BE,$97
	.dc.b	$EE,$F9,$DF,$EE,$FC,$3F,$DE,$53,$97,$FC,$89,$19,$BF,$F5,$64,$E5
	.dc.b	$E9,$3C,$2B,$9D,$F6,$6C,$64,$FF,$27,$7A,$3F,$D7,$DF,$F8,$01,$E8
	.dc.b	$BE,$AF,$F1,$F3,$9C,$3F,$CD,$B1,$8F,$03,$F2,$B5,$BB,$E7,$18,$9D
	.dc.b	$47,$46,$D7,$28,$FF,$4E,$7A,$5F,$9E,$F6,$4F,$E9,$3B,$DF,$56,$F8
	.dc.b	$5D,$4C,$BF,$F4,$87,$45,$C5,$CB,$FB,$3F,$61,$3C,$DB,$5C,$0F,$09
	.dc.b	$6B,$8B,$FB,$8F,$6F,$F0,$0D,$E6,$FD,$8E,$CF,$A2,$F3,$56,$AF,$FD
	.dc.b	$4F,$BF,$F0,$9C,$FF,$95,$F8,$88,$F0,$CF,$FB,$04,$F8,$1A,$7E,$FB
	.dc.b	$EA,$9F,$55,$4C,$EB,$FE,$B3,$C7,$9C,$EA,$30,$7F,$AD,$FE,$93,$FA
	.dc.b	$BF,$BF,$FC,$31,$FE,$B3,$F5,$57,$D0,$FF,$5C,$D7,$0C,$D5,$A8,$C3
	.dc.b	$C9,$B2,$AE,$57,$CD,$B3,$A6,$DF,$CA,$3A,$D7,$19,$F0,$6A,$FA,$7F
	.dc.b	$CC,$67,$F1,$9B,$3F,$ED,$1F,$2F,$FE,$00,$7A,$DF,$C8,$A7,$38,$FF
	.dc.b	$B5,$1B,$87,$3E,$DF,$E4,$BF,$15,$FE,$37,$C7,$BD,$70,$C5,$3F,$ED
	.dc.b	$49,$F1,$FD,$53,$E3,$78,$FC,$1F,$F3,$08,$C4,$FF,$AC,$EB,$86,$7E
	.dc.b	$DC,$D5,$A8,$F8,$18,$F6,$C3,$96,$3F,$CF,$A7,$E0,$7B,$97,$33,$FD
	.dc.b	$0B,$B1,$F5,$4F,$AA,$D4,$CD,$BE,$C6,$1C,$9F,$B5,$7C,$2B,$53,$F4
	.dc.b	$4D,$1F,$4E,$5F,$54,$F8,$DD,$F1,$FF,$80,$27,$31,$FD,$53,$CE,$7C
	.dc.b	$A7,$9A,$7F,$AC,$BF,$D3,$70,$FE,$CA,$E5,$BE,$B7,$18,$C6,$1F,$E0
	.dc.b	$CB,$E7,$39,$D3,$4F,$F5,$CF,$96,$F2,$63,$F8,$6F,$A4,$FD,$73,$FE
	.dc.b	$E3,$8C,$E5,$F4,$67,$2F,$D0,$8C,$E3,$FB,$09,$CB,$FE,$1C,$E3,$4E
	.dc.b	$AD,$6E,$56,$1C,$27,$F5,$9D,$3E,$A6,$E5,$FA,$BC,$FF,$58,$D8,$FD
	.dc.b	$3B,$70,$FD,$17,$67,$F9,$2A,$31,$A7,$F8,$BA,$31,$5F,$AE,$BE,$6F
	.dc.b	$BF,$E6,$7F,$C0,$3F,$24,$F9,$1F,$8D,$18,$6C,$FE,$B2,$74,$F1,$31
	.dc.b	$1F,$6F,$EA,$BF,$C0,$58,$FD,$35,$F0,$F9,$F7,$34,$DB,$FF,$0B,$CE
	.dc.b	$9B,$3A,$7F,$82,$FD,$57,$F0,$FB,$7D,$3F,$F0,$F8,$C3,$67,$F8,$6C
	.dc.b	$E9,$E2,$62,$35,$BF,$C8,$9D,$3F,$2B,$81,$7D,$2F,$F8,$D7,$23,$E8
	.dc.b	$9F,$43,$FE,$28,$F9,$B7,$B9,$FE,$28,$9C,$DF,$F1,$29,$CB,$CA,$33
	.dc.b	$97,$EF,$67,$2F,$EF,$23,$39,$FF,$3C,$9C,$BF,$8D,$8F,$EA,$D5,$D7
	.dc.b	$C2,$F0,$7A,$5E,$36,$C6,$4F,$04,$65,$FF,$22,$FC,$DF,$EB,$7E,$1D
	.dc.b	$CB,$F9,$6E,$D4,$E6,$30,$FC,$4A,$FF,$46,$46,$3A,$BF,$CF,$77,$CB
	.dc.b	$56,$1C,$A7,$35,$BB,$E3,$5A,$C3,$A2,$C3,$96,$1C,$6B,$6D,$9D,$5F
	.dc.b	$36,$9C,$F8,$7B,$BD,$71,$B5,$C7,$0E,$B8,$FB,$6B,$86,$3D,$A3,$10
	.dc.b	$F9,$E7,$D4,$FF,$82,$23,$B9,$97,$17,$F0,$61,$F5,$4F,$A9,$F1,$C6
	.dc.b	$FC,$6F,$82,$D3,$9E,$1D,$A3,$35,$C7,$2D,$CE,$6B,$78,$F8,$36,$FF
	.dc.b	$68,$4E,$72,$F1,$61,$C6,$F7,$8F,$2C,$38,$47,$F1,$4B,$E1,$87,$FB
	.dc.b	$66,$F8,$61,$EB,$9C,$D6,$EF,$8B,$E9,$9E,$F9,$7B,$AB,$73,$9C,$3C
	.dc.b	$B7,$C3,$0E,$B1,$8C,$7C,$84,$AF,$9B,$3A,$7C,$11,$9C,$7C,$97,$CF
	.dc.b	$1F,$03,$E4,$C7,$C2,$78,$B0,$E5,$87,$8C,$E6,$7C,$43,$BD,$72,$C3
	.dc.b	$B6,$3E,$59,$CE,$58,$78,$E5,$E2,$AF,$AF,$7C,$67,$A4,$E5,$E2,$62
	.dc.b	$FA,$46,$4F,$93,$7D,$67,$CC,$4C,$3D,$38,$FB,$32,$F7,$1F,$06,$B8
	.dc.b	$46,$72,$F1,$C6,$27,$71,$8B,$F1,$1E,$AC,$7C,$67,$E0,$72,$F5,$65
	.dc.b	$E5,$8C,$CF,$F5,$AC,$BC,$78,$74,$8F,$EB,$A7,$C7,$29,$F9,$D2,$E5
	.dc.b	$F9,$96,$F4,$FF,$3A,$76,$7E,$BD,$C3,$AE,$9C,$B9,$6C,$A7,$96,$3F
	.dc.b	$97,$6E,$3F,$A4,$6D,$FF,$72,$F0,$1F,$55,$E8,$BE,$18,$FE,$BA,$F7
	.dc.b	$7D,$CF,$DF,$DF,$07,$53,$1F,$DC,$CF,$09,$D4,$62,$75,$5D,$A3,$9D
	.dc.b	$F8,$70,$F3,$56,$9B,$3A,$30,$7E,$DF,$0F,$2E,$3F,$06,$BF,$85,$E3
	.dc.b	$E0,$C7,$B1,$30,$E6,$FF,$15,$8D,$46,$09,$87,$5A,$E7,$8F,$96,$30
	.dc.b	$7A,$C9,$7C,$E3,$51,$F1,$E3,$C0,$F0,$74,$FA,$8D,$57,$81,$FF,$11
	.dc.b	$CF,$2C,$3D,$91,$DA,$75,$87,$A2,$3A,$47,$19,$DD,$F6,$9E,$06,$88
	.dc.b	$76,$3C,$51,$C2,$7A,$DF,$87,$1F,$9A,$4C,$7E,$1D,$F0,$8C,$3F,$54
	.dc.b	$EE,$7B,$0E,$F8,$F9,$DC,$13,$0F,$03,$F9,$3B,$F0,$5F,$53,$BC,$7F
	.dc.b	$5C,$E0,$7D,$57,$F4,$B8,$FE,$C6,$D7,$4D,$CF,$FB,$4A,$E5,$FB,$55
	.dc.b	$CB,$7B,$3B,$D7,$FC,$00,$DC,$3F,$BC,$BC,$5E,$F3,$C0,$8F,$8A,$EB
	.dc.b	$E2,$8B,$87,$8A,$7F,$15,$D6,$F8,$A7,$D5,$7F,$08,$EA,$FF,$3C,$7D
	.dc.b	$53,$E5,$F1,$CF,$FE,$13,$8C,$69,$FE,$32,$C3,$8B,$70,$DB,$FE,$10
	.dc.b	$9D,$BF,$EC,$5D,$D1,$8C,$7F,$CE,$D7,$CA,$BF,$C6,$18,$72,$F1,$DF
	.dc.b	$54,$F8,$BC,$36,$BF,$C7,$55,$E0,$D7,$FC,$88,$EA,$A4,$E8,$FF,$55
	.dc.b	$FD,$57,$F9,$21,$BD,$FE,$85,$08,$C3,$B8,$FF,$57,$37,$35,$FF,$30
	.dc.b	$8E,$13,$D0,$EF,$E1,$F2,$39,$DC,$8C,$9F,$EF,$5D,$9F,$54,$FA,$AF
	.dc.b	$97,$18,$F3,$EC,$FF,$AE,$EB,$83,$B6,$D6,$A6,$46,$0F,$F5,$67,$83
	.dc.b	$F3,$7E,$A9,$F1,$54,$F9,$5E,$87,$F9,$42,$B6,$7F,$BE,$A7,$6D,$9D
	.dc.b	$D7,$C9,$F3,$5E,$FD,$F3,$E5,$FF,$B6,$A7,$2D,$9C,$BF,$ED,$98,$4F
	.dc.b	$39,$93,$C6,$37,$C5,$F4,$23,$FD,$77,$5D,$1F,$F6,$27,$D5,$7E,$27
	.dc.b	$C3,$F8,$D1,$FE,$F9,$6D,$7F,$B1,$74,$57,$C0,$38,$6B,$2F,$FD,$A5
	.dc.b	$1F,$4B,$D3,$FE,$BD,$1F,$EE,$4F,$47,$06,$7C,$79,$19,$21,$FE,$E8
	.dc.b	$7F,$1D,$E7,$78,$BA,$3F,$BB,$EA,$FA,$AF,$94,$6B,$39,$87,$FB,$63
	.dc.b	$5F,$80,$B0,$E4,$FF,$B7,$1A,$FF,$B2,$78,$78,$A9,$E3,$3A,$37,$87
	.dc.b	$DF,$9D,$63,$FE,$E6,$EB,$7B,$6F,$6B,$F6,$9D,$4F,$AA,$A7,$EC,$67
	.dc.b	$58,$7F,$AD,$9A,$F4,$3F,$DE,$13,$C5,$F6,$61,$FE,$F4,$F0,$3F,$61
	.dc.b	$7C,$F0,$FE,$05,$F5,$4F,$5E,$93,$0E,$DE,$5D,$7B,$60,$AE,$11,$D8
	.dc.b	$FF,$80,$0E,$1F,$26,$9E,$EA,$FC,$D7,$0F,$E6,$FE,$AA,$57,$E6,$63
	.dc.b	$2F,$E6,$E3,$2F,$E6,$63,$2F,$E6,$E3,$3C,$3F,$63,$83,$EB,$E4,$FA
	.dc.b	$97,$D5,$7F,$1F,$6E,$1E,$79,$D4,$FF,$23,$F8,$77,$D5,$7F,$25,$F9
	.dc.b	$E7,$CC,$E6,$26,$1C,$67,$2F,$C1,$F9,$17,$C6,$7F,$91,$35,$FD,$54
	.dc.b	$62,$7F,$9C,$B5,$FF,$77,$FC,$73,$EE,$73,$F2,$3C,$7A,$54,$B4,$E6
	.dc.b	$30,$DD,$D9,$FF,$5B,$9F,$AE,$DC,$BC,$75,$C6,$F9,$D7,$BF,$D4,$8F
	.dc.b	$B5,$AF,$E3,$61,$F5,$5F,$02,$3E,$EF,$CB,$7F,$B1,$DF,$C2,$FC,$B7
	.dc.b	$F1,$A6,$F9,$7F,$D2,$67,$59,$7F,$0D,$EA,$46,$36,$FC,$B6,$1C,$6B
	.dc.b	$F8,$96,$C8,$C4,$FE,$A6,$B8,$62,$C3,$9F,$C0,$7F,$5D,$BD,$CB,$0E
	.dc.b	$59,$78,$EB,$FA,$E7,$CC,$3E,$AB,$FB,$27,$D0,$FE,$B3,$5F,$0B,$75
	.dc.b	$FD,$C5,$6F,$2F,$D0,$68,$8C,$47,$E9,$AB,$83,$E1,$6C,$EF,$0F,$09
	.dc.b	$E0,$F0,$65,$74,$BE,$78,$FC,$7F,$9C,$7D,$57,$F7,$9F,$89,$FC,$BD
	.dc.b	$FE,$6D,$AF,$EC,$27,$6F,$99,$AF,$EE,$A3,$75,$F3,$9B,$A7,$F1,$5A
	.dc.b	$DC,$F8,$6F,$C1,$D5,$30,$E9,$1F,$E0,$5F,$49,$F5,$5F,$E0,$DF,$5D
	.dc.b	$F5,$54,$D4,$4B,$FD,$94,$E9,$B3,$97,$F8,$4D,$74,$72,$43,$35,$FE
	.dc.b	$15,$FA,$2F,$F1,$06,$9F,$C7,$FE,$09,$F3,$9F,$E1,$F6,$DF,$D8,$AD
	.dc.b	$BF,$89,$6F,$3A,$77,$3F,$C8,$1B,$B7,$E6,$6B,$8F,$21,$E3,$3A,$C7
	.dc.b	$C5,$8F,$97,$E0,$FF,$C5,$5F,$03,$FC,$97,$1F,$CE,$FD,$57,$F6,$D6
	.dc.b	$ED,$AF,$3A,$6B,$FC,$5E,$B5,$97,$F3,$B6,$C7,$E8,$70,$E9,$3F,$C2
	.dc.b	$3B,$DE,$B3,$E3,$3F,$88,$B6,$FF,$8C,$35,$FF,$25,$7D,$23,$E8,$F7
	.dc.b	$52,$76,$FF,$8A,$27,$51,$FD,$04,$7D,$43,$EA,$BF,$B2,$43,$C0,$7D
	.dc.b	$1F,$FB,$AB,$EA,$9F,$3D,$FE,$EC,$9C,$D7,$F2,$09,$CB,$FC,$8A,$32
	.dc.b	$F6,$BE,$B1,$8C,$3C,$1C,$7B,$C7,$B6,$3D,$5B,$86,$4F,$E5,$B5,$C1
	.dc.b	$F3,$47,$FB,$FA,$31,$7E,$0C,$7C,$75,$A3,$CD,$7D,$5E,$78,$F9,$30
	.dc.b	$E4,$D8,$D9,$E1,$15,$D2,$FD,$53,$B9,$F7,$63,$BA,$D5,$FA,$A7,$C4
	.dc.b	$68,$B8,$FA,$23,$58,$7A,$71,$F3,$56,$E3,$BE,$1D,$67,$38,$76,$AF
	.dc.b	$5C,$F9,$F0,$F8,$D1,$FD,$50,$C9,$4A,$57,$6F,$CB,$FA,$A7,$CB,$EE
	.dc.b	$3E,$8C,$61,$C4,$A3,$11,$CD,$AF,$1E,$8E,$4F,$8D,$AF,$F6,$86,$BF
	.dc.b	$BB,$32,$FD,$1F,$AA,$FE,$CF,$C7,$F4,$E7,$81,$F3,$23,$2F,$CE,$33
	.dc.b	$A7,$E4,$A3,$2F,$E5,$23,$3A,$FC,$59,$CB,$E7,$8E,$9E,$77,$F9,$B6
	.dc.b	$B8,$47,$95,$E9,$19,$9E,$11,$8B,$F3,$56,$E7,$B5,$6E,$3E,$96,$1E
	.dc.b	$D8,$C4,$7C,$53,$C3,$5C,$A7,$4D,$8C,$9E,$F3,$B1,$FE,$12,$C3,$D6
	.dc.b	$F6,$37,$87,$EE,$39,$1F,$55,$F2,$36,$FF,$10,$B7,$0F,$F0,$DF,$82
	.dc.b	$FC,$61,$EC,$FF,$A3,$EF,$7A,$DB,$FE,$35,$87,$1C,$FF,$BD,$7D,$77
	.dc.b	$F0,$D9,$FF,$18,$4E,$7C,$24,$E6,$BF,$DC,$9F,$8C,$BE,$EF,$A4,$35
	.dc.b	$4B,$E6,$33,$18,$AB,$39,$CB,$FD,$94,$F3,$6D,$6A,$76,$F6,$F1,$5D
	.dc.b	$FF,$CC,$6B,$FC,$A0,$3E,$1F,$51,$9C,$BF,$CD,$B3,$93,$FC,$1B,$19
	.dc.b	$DF,$FC,$47,$19,$9F,$F3,$6C,$E6,$3F,$CF,$C6,$67,$FE,$00,$F7,$7C
	.dc.b	$0D,$BF,$DD,$78,$78,$23,$15,$B3,$EA,$FE,$34,$F7,$FE,$1B,$FE,$E7
	.dc.b	$8F,$F7,$53,$96,$FC,$7E,$A7,$6D,$8C,$9C,$23,$2D,$33,$9F,$F7,$EF
	.dc.b	$0B,$CD,$C3,$FD,$46,$7F,$C5,$BE,$35,$AF,$F4,$8C,$DD,$7F,$C9,$53
	.dc.b	$6B,$FD,$3D,$36,$BF,$C8,$93,$75,$FC,$06,$1F,$56,$31,$5B,$DF,$FD
	.dc.b	$DB,$3E,$23,$F0,$7F,$8A,$3E,$1A,$FC,$E6,$1B,$9F,$FA,$87,$05,$6E
	.dc.b	$30,$AB,$84,$61,$B9,$FF,$AE,$6B,$8C,$63,$4F,$F7,$94,$62,$BF,$C8
	.dc.b	$9E,$BF,$DB,$FA,$E7,$D4,$7E,$3F,$D6,$4E,$63,$0D,$AE,$27,$FA,$86
	.dc.b	$B8,$B8,$CB,$FD,$7B,$18,$7F,$D7,$37,$D2,$78,$5F,$E9,$BF,$1B,$7C
	.dc.b	$0D,$3F,$5B,$E1,$FF,$9A,$FE,$F3,$17,$F1,$6B,$6F,$FA,$9F,$D6,$3A
	.dc.b	$CB,$FD,$1D,$18,$8D,$1F,$B4,$FC,$73,$FC,$0E,$E0,$FA,$0E,$27,$FD
	.dc.b	$2D,$F6,$3F,$45,$FF,$6B,$FC,$3B,$DF,$FE,$03,$E4,$2B,$5C,$7F,$E8
	.dc.b	$EA,$D6,$3F,$EC,$56,$E1,$FE,$A1,$F8,$47,$57,$EA,$E7,$37,$FE,$AC
	.dc.b	$8F,$D7,$78,$EF,$B5,$A7,$82,$F8,$E3,$FE,$C9,$8F,$E2,$7F,$51,$FD
	.dc.b	$93,$7B,$E8,$B8,$AC,$61,$B7,$FE,$EF,$AF,$ED,$71,$8D,$5A,$7C,$C5
	.dc.b	$3A,$AE,$18,$7F,$A2,$9B,$7C,$6F,$85,$6E,$FF,$DF,$FF,$59,$FE,$38
	.dc.b	$BF,$F7,$43,$63,$FD,$FB,$F2,$2F,$B1,$FE,$1D,$39,$C3,$FD,$72,$DB
	.dc.b	$FF,$6D,$F8,$48,$C6,$1F,$EB,$9C,$39,$57,$6F,$4B,$F9,$9F,$E3,$AF
	.dc.b	$B0,$BE,$EE,$1B,$5F,$B1,$9D,$36,$74,$D3,$FC,$AD,$8F,$CD,$6E,$DF
	.dc.b	$98,$FC,$77,$FD,$85,$18,$26,$7F,$31,$FF,$79,$F4,$BE,$C1,$F2,$9F
	.dc.b	$D3,$2B,$D9,$87,$F4,$EA,$D4,$E7,$1F,$05,$73,$3F,$29,$7C,$B0,$EF
	.dc.b	$C1,$F6,$70,$E3,$1F,$C3,$7E,$63,$F2,$EF,$F6,$A7,$F8,$63,$9F,$23
	.dc.b	$FA,$CC,$63,$0F,$E1,$8F,$07,$FB,$03,$9B,$FE,$1A,$E7,$C6,$39,$B5
	.dc.b	$7F,$BB,$7E,$73,$FB,$BC,$62,$FF,$DE,$BF,$55,$F4,$2B,$D9,$E6,$2F
	.dc.b	$71,$EF,$E6,$56,$EF,$F2,$D7,$BF,$63,$FA,$B4,$E7,$1F,$0C,$E5,$F0
	.dc.b	$1F,$8A,$8D,$61,$F9,$EE,$A7,$D9,$7E,$AB,$9F,$F5,$6D,$BF,$D0,$B6
	.dc.b	$BE,$97,$5F,$F9,$52,$31,$9F,$E8,$B0,$E7,$7F,$A8,$C3,$9B,$4F,$F8
	.dc.b	$03,$D4,$F9,$2F,$ED,$DF,$64,$F8,$FF,$F0,$14,$E6,$BD,$3A,$38,$17
	.dc.b	$B9,$E7,$87,$2C,$38,$B6,$3F,$A7,$7C,$F3,$B3,$FC,$2E,$FF,$80,$71
	.dc.b	$FF,$0E,$39,$9D,$FA,$DA,$AD,$B6,$32,$7F,$04,$8F,$E2,$31,$FC,$49
	.dc.b	$B1,$93,$F8,$4C,$65,$A6,$6B,$E5,$EB,$E3,$E0,$FF,$3F,$DF,$29,$FE
	.dc.b	$BD,$F6,$1F,$E0,$F7,$FC,$64,$FF,$B5,$AF,$FB,$53,$9E,$73,$99,$E3
	.dc.b	$EC,$7F,$8D,$A3,$1C,$5F,$D9,$9E,$06,$2F,$F1,$3F,$60,$FA,$6F,$F1
	.dc.b	$17,$D0,$3E,$53,$FC,$6D,$39,$AF,$AB,$C8,$AD,$47,$D8,$E9,$56,$F4
	.dc.b	$FC,$E5,$69,$FD,$0C,$67,$3F,$E6,$58,$73,$C7,$FC,$9B,$C8,$FA,$EF
	.dc.b	$E6,$DF,$0D,$FC,$2B,$1F,$F2,$A4,$F3,$8C,$70,$7E,$2E,$B7,$E1,$FF
	.dc.b	$9A,$AB,$4D,$8C,$DF,$2C,$2D,$F1,$DB,$FD,$49,$7C,$67,$38,$72,$BE
	.dc.b	$15,$A7,$C7,$86,$AA,$4E,$70,$FC,$F4,$72,$9C,$D3,$1F,$62,$FA,$4F
	.dc.b	$F1,$17,$85,$D6,$31,$ED,$7F,$75,$78,$69,$FE,$CD,$FA,$F7,$D7,$7B
	.dc.b	$27,$77,$C7,$0E,$71,$C0,$FE,$2E,$E3,$1F,$0D,$F2,$D7,$86,$5D,$E7
	.dc.b	$CB,$5B,$CB,$FD,$B8,$DC,$3F,$90,$E1,$C5,$EC,$78,$8B,$1B,$71,$5E
	.dc.b	$0C,$BD,$13,$92,$61,$E1,$C3,$DD,$3A,$8C,$5F,$A9,$AF,$49,$DE,$3E
	.dc.b	$AC,$A4,$F8,$AB,$79,$79,$F2,$F6,$C7,$2C,$BC,$37,$F4,$6B,$58,$FC
	.dc.b	$0A,$D1,$27,$59,$79,$5E,$F8,$74,$73,$96,$2F,$94,$5C,$BE,$4D,$F1
	.dc.b	$7C,$D1,$95,$8F,$89,$F8,$11,$9C,$BE,$64,$E6,$7E,$11,$A3,$99,$F5
	.dc.b	$1D,$47,$BA,$33,$8F,$D3,$C7,$A6,$5E,$D9,$E1,$8F,$96,$BD,$91,$C3
	.dc.b	$2F,$9E,$D3,$C1,$1A,$3A,$39,$3C,$A4,$C7,$C7,$7D,$E7,$85,$74,$6D
	.dc.b	$7A,$CF,$ED,$E5,$3E,$59,$83,$07,$E7,$B9,$BF,$85,$39,$3F,$5C,$E1
	.dc.b	$E1,$97,$FB,$3F,$4D,$46,$9B,$19,$8F,$CD,$CF,$4A,$FD,$2B,$6F,$CB
	.dc.b	$E0,$FF,$B9,$EB,$F6,$53,$97,$C9,$86,$0F,$DF,$4E,$E3,$AC,$65,$A7
	.dc.b	$EB,$23,$57,$E0,$9D,$36,$75,$F0,$7F,$1A,$30,$4C,$3F,$73,$87,$8E
	.dc.b	$BF,$C1,$F5,$B9,$CD,$F4,$6C,$F4,$30,$7A,$8F,$2D,$FF,$1A,$C3,$CB
	.dc.b	$3C,$23,$C1,$3B,$75,$7D,$E3,$13,$B2,$3F,$E2,$CA,$E6,$E8,$87,$D5
	.dc.b	$3C,$87,$8A,$38,$B9,$AE,$EF,$A0,$C4,$6A,$7A,$C6,$0E,$B8,$FD,$FA
	.dc.b	$E3,$1D,$4C,$E3,$F2,$6B,$8D,$6E,$FC,$2D,$AF,$56,$1E,$F7,$B1,$23
	.dc.b	$4F,$D2,$3C,$CE,$0C,$D7,$6B,$FB,$11,$97,$C5,$5D,$63,$18,$FB,$EB
	.dc.b	$83,$D2,$B7,$7F,$D5,$7E,$5F,$FC,$4C,$DE,$AF,$A9,$C4,$9C,$CF,$F4
	.dc.b	$98,$C7,$85,$A9,$CE,$BF,$DC,$9B,$AF,$F9,$92,$B9,$61,$BB,$98,$77
	.dc.b	$21,$D6,$FF,$D3,$4F,$96,$3B,$4C,$8C,$1F,$DC,$7A,$7F,$C1,$FC,$5F
	.dc.b	$C1,$39,$FF,$C1,$7F,$58,$FA,$CC,$4F,$FA,$B2,$31,$7F,$E1,$7E,$AF
	.dc.b	$F6,$13,$FE,$14,$9D,$BF,$E5,$19,$DB,$FE,$75,$6D,$FF,$9D,$2B,$8B
	.dc.b	$FE,$CB,$D1,$F5,$6F,$85,$E7,$18,$FA,$5B,$26,$CC,$C7,$FC,$C3,$5B
	.dc.b	$6D,$6D,$B3,$E6,$5F,$0D,$FE,$F0,$7F,$CD,$26,$3E,$AB,$50,$6F,$B5
	.dc.b	$27,$73,$96,$CE,$5F,$F5,$91,$DA,$79,$C4,$8E,$31,$EE,$F1,$DC,$D7
	.dc.b	$E5,$A5,$7F,$9D,$23,$46,$1F,$F5,$F6,$6A,$56,$A7,$FD,$13,$E0,$FD
	.dc.b	$4F,$09,$E8,$3E,$A3,$0D,$AD,$90,$FF,$56,$39,$6C,$7D,$EE,$67,$9B
	.dc.b	$FE,$93,$9F,$87,$C5,$F2,$3C,$F9,$18,$21,$F1,$38,$51,$A8,$C4,$FD
	.dc.b	$03,$1C,$9F,$8E,$AD,$CF,$28,$FF,$5E,$FA,$0F,$9D,$F7,$78,$5F,$72
	.dc.b	$BF,$CD,$ED,$7F,$D9,$33,$B7,$FC,$E0,$D7,$FD,$95,$1B,$AF,$F3,$6B
	.dc.b	$7A,$BF,$D8,$F5,$B9,$FC,$C7,$CC,$79,$2F,$E9,$AF,$9D,$FC,$CB,$63
	.dc.b	$F2,$AE,$E7,$8E,$9A,$A9,$5A,$8F,$CB,$79,$6F,$C4,$BF,$27,$F8,$04
	.dc.b	$AF,$CB,$CE,$5B,$39,$7E,$74,$F2,$89,$1A,$21,$F4,$FF,$00,$FD,$7F
	.dc.b	$B9,$FE,$9B,$8F,$D5,$BF,$88,$6F,$2E,$BD,$09,$18,$21,$F9,$49,$E4
	.dc.b	$FF,$BD,$7E,$AF,$FD,$D2,$FF,$A8,$9A,$E7,$D8,$FC,$67,$33,$E0,$7F
	.dc.b	$31,$5A,$E0,$FF,$4F,$B7,$8D,$87,$FB,$7E,$30,$BA,$3F,$2B,$7F,$BF
	.dc.b	$FB,$CF,$F3,$FC,$FE,$F9,$FF,$76,$37,$D9,$D7,$A7,$23,$04,$3F,$77
	.dc.b	$3C,$9B,$1D,$66,$47,$58,$FD,$87,$DD,$F9,$3B,$8D,$3C,$FF,$10,$F8
	.dc.b	$5F,$F1,$96,$DF,$EB,$CF,$86,$F2,$E2,$86,$FF,$02,$F8,$4F,$F3,$06
	.dc.b	$BF,$EC,$EA,$E3,$5C,$21,$2F,$A8,$7D,$5F,$F0,$5B,$DE,$1F,$33,$34
	.dc.b	$63,$0E,$37,$C1,$FA,$38,$F5,$C5,$F8,$F6,$67,$F8,$AF,$6F,$D4,$F2
	.dc.b	$3F,$49,$CD,$E7,$AD,$61,$FD,$07,$E3,$7F,$01,$87,$4D,$7D,$D5,$A6
	.dc.b	$ED,$FD,$12,$B7,$5F,$C8,$9B,$7C,$F9,$34,$7F,$54,$F8,$E7,$E3,$EF
	.dc.b	$A9,$FE,$35,$F2,$77,$1F,$F7,$AF,$CF,$FF,$1B,$7E,$86,$9F,$0E,$31
	.dc.b	$97,$CB,$EA,$5F,$1B,$D9,$F9,$17,$13,$FE,$F6,$AE,$37,$CF,$CC,$F0
	.dc.b	$E3,$30,$F0,$C7,$F3,$7A,$DC,$73,$9F,$EC,$7F,$8D,$3E,$66,$FD,$8C
	.dc.b	$AF,$EC,$33,$96,$CE,$5F,$E8,$F0,$9E,$73,$27,$8C,$6F,$2F,$EE,$95
	.dc.b	$CE,$7D,$BA,$3D,$3F,$E3,$91,$F0,$3A,$9C,$DF,$C9,$76,$FC,$53,$67
	.dc.b	$F6,$95,$C3,$1F,$94,$DA,$DF,$9B,$FD,$5E,$F8,$63,$DF,$A3,$FA,$85
	.dc.b	$EF,$16,$5E,$0C,$38,$1A,$AF,$F0,$4F,$CD,$FF,$C0,$1A,$FE,$CE,$FE
	.dc.b	$71,$FF,$03,$B5,$CF,$C2,$F0,$38,$F6,$C8,$DC,$65,$E3,$19,$78,$9F
	.dc.b	$99,$3F,$89,$7C,$E7,$F9,$0B,$6F,$D9,$64,$FA,$7F,$F0,$85,$7C,$96
	.dc.b	$CF,$E3,$6B,$86,$FF,$C7,$B6,$46,$2F,$E0,$DF,$0A,$FD,$C3,$63,$C3
	.dc.b	$EF,$7F,$77,$AE,$18,$F8,$77,$FF,$02,$5E,$F1,$EB,$7F,$E3,$0F,$9F
	.dc.b	$7D,$D5,$E6,$99,$FF,$8B,$A7,$2D,$9C,$BF,$E1,$48,$FD,$0E,$9F,$E5
	.dc.b	$B7,$71,$FA,$6F,$07,$FC,$2F,$F0,$3F,$E1,$55,$F6,$34,$FE,$39,$ED
	.dc.b	$7F,$97,$2B,$83,$78,$7F,$88,$37,$1F,$E4,$8D,$9F,$C9,$FD,$27,$E4
	.dc.b	$57,$D3,$3E,$A7,$F9,$64,$3B,$9F,$33,$FE,$E6,$FA,$87,$D4,$FF,$BA
	.dc.b	$27,$35,$FD,$42,$72,$FF,$52,$8C,$BD,$AF,$AC,$63,$0F,$06,$9B,$C7
	.dc.b	$B6,$3D,$5B,$86,$4F,$EB,$B5,$C1,$F3,$47,$FB,$EA,$31,$7E,$0C,$7C
	.dc.b	$75,$A3,$CD,$7D,$5E,$78,$F9,$30,$E4,$D8,$D9,$E1,$15,$D2,$FD,$53
	.dc.b	$B9,$F7,$63,$BA,$D5,$FA,$A7,$C4,$68,$B8,$FA,$23,$58,$7A,$71,$F3
	.dc.b	$56,$E3,$BE,$1D,$67,$38,$76,$AF,$5C,$F9,$F0,$F8,$D1,$FE,$04,$32
	.dc.b	$52,$95,$DB,$F2,$FE,$A5,$F2,$BB,$8F,$A3,$18,$71,$28,$C4,$73,$6B
	.dc.b	$C7,$E3,$79,$3E,$36,$BF,$E1,$86,$A6,$9F,$47,$C8,$FE,$E3,$C1,$F4
	.dc.b	$E7,$81,$F3,$23,$2F,$CE,$33,$A7,$E4,$A3,$2F,$E5,$23,$3A,$FC,$59
	.dc.b	$CB,$E7,$8E,$9D,$BF,$E0,$7A,$E1,$1E,$53,$37,$E3,$71,$1D,$E7,$85
	.dc.b	$73,$8C,$56,$F0,$F6,$56,$AF,$EA,$46,$27,$51,$D8,$F6,$1D,$AF,$D1
	.dc.b	$5C,$63,$E6,$1B,$BF,$3D,$F3,$8F,$D1,$71,$BE,$A7,$E3,$ED,$FE,$17
	.dc.b	$6E,$5F,$EE,$FE,$2F,$F4,$9B,$FC,$E6,$B8,$65,$FE,$29,$BE,$0D,$9F
	.dc.b	$F1,$95,$F2,$C3,$F8,$7F,$BD,$97,$FC,$5F,$19,$F1,$1D,$0F,$A8,$D4
	.dc.b	$63,$E4,$AC,$E7,$1F,$E5,$D1,$8A,$CB,$63,$27,$F4,$FE,$EF,$E8,$BF
	.dc.b	$31,$E1,$F0,$BF,$C8,$FF,$13,$E3,$9C,$D7,$F5,$59,$CB,$C7,$C4,$5F
	.dc.b	$18,$FF,$69,$99,$71,$7C,$E7,$73,$CC,$DE,$BF,$E0,$68,$CB,$FE,$05
	.dc.b	$8C,$F0,$7F,$BF,$FC,$6F,$F7,$8C,$7F,$72,$F8,$BF,$87,$18,$AF,$EE
	.dc.b	$F1,$8F,$87,$CB,$6B,$81,$FD,$EA,$32,$D3,$33,$FE,$18,$F1,$FF,$C5
	.dc.b	$11,$F0,$78,$6C,$E7,$CD,$F9,$11,$8C,$BF,$C2,$B1,$83,$E5,$5E,$6F
	.dc.b	$FC,$3B,$18,$34,$FD,$38,$D5,$FF,$85,$AE,$8F,$A8,$FF,$79,$D7,$F8
	.dc.b	$1E,$31,$F2,$BF,$84,$9C,$D7,$FB,$F9,$46,$7E,$0B,$4D,$BE,$53,$A6
	.dc.b	$CE,$9A,$6A,$7F,$09,$E9,$3E,$A3,$F1,$1F,$50,$FA,$8C,$C6,$1B,$7F
	.dc.b	$89,$9D,$3F,$91,$AE,$4E,$67,$79,$7F,$91,$6B,$6F,$F9,$2A,$B6,$E6
	.dc.b	$7F,$47,$EC,$3E,$A2,$9F,$AB,$F1,$3F,$CF,$1F,$45,$8C,$BE,$45,$6D
	.dc.b	$FD,$C7,$D2,$FF,$A9,$9C,$DF,$EC,$FC,$24,$E7,$2F,$F0,$93,$79,$3F
	.dc.b	$D3,$4D,$BF,$F6,$6F,$13,$1E,$D7,$C2,$7A,$7D,$3F,$D6,$C3,$FD,$44
	.dc.b	$D8,$FF,$1C,$64,$F7,$F7,$C7,$FB,$5F,$A7,$7A,$FE,$5F,$82,$FF,$36
	.dc.b	$61,$FB,$59,$CB,$FC,$4E,$33,$9F,$F9,$BF,$C7,$FF,$73,$4F,$FA,$FF
	.dc.b	$DF,$E5,$D5,$FC,$06,$73,$5F,$E9,$4F,$A1,$FE,$BB,$C5,$FC,$55,$FE
	.dc.b	$BF,$C9,$7D,$0F,$8B,$97,$F9,$66,$72,$D8,$CD,$F3,$55,$C6,$F9,$36
	.dc.b	$B8,$36,$B8,$3F,$C5,$7E,$97,$FC,$5F,$3F,$C4,$DB,$3F,$E8,$EF,$8A
	.dc.b	$7D,$3E,$FD,$19,$7B,$7F,$8E,$4E,$A3,$07,$EA,$FE,$31,$ED,$7F,$2E
	.dc.b	$5E,$ED,$8C,$36,$78,$37,$1F,$F5,$DF,$5A,$F8,$D7,$F9,$F6,$59,$FC
	.dc.b	$28,$C3,$63,$FD,$87,$DA,$FA,$7F,$F6,$85,$6B,$6F,$E8,$D5,$C9,$B7
	.dc.b	$C7,$B3,$FD,$F1,$87,$1F,$7B,$35,$30,$FF,$47,$74,$BE,$9D,$F4,$19
	.dc.b	$CF,$FD,$A7,$39,$8F,$E9,$FF,$4F,$F9,$59,$FF,$5B,$BF,$C1,$DF,$7B
	.dc.b	$9D,$CE,$74,$63,$FA,$48,$D1,$97,$1D,$BD,$BC,$7F,$EA,$1F,$2F,$FD
	.dc.b	$B3,$1F,$F6,$37,$D1,$FD,$4B,$FE,$B1,$39,$C7,$FD,$4B,$39,$AF,$F6
	.dc.b	$7C,$E6,$91,$FE,$C0,$9D,$57,$FB,$CE,$B5,$39,$E5,$FC,$E7,$83,$F9
	.dc.b	$D6,$E1,$F9,$B6,$CF,$E6,$DB,$7E,$E8,$76,$7F,$91,$27,$35,$E2,$8C
	.dc.b	$3F,$A5,$BE,$CF,$F7,$3E,$87,$C8,$7F,$73,$FA,$77,$C5,$FE,$62,$30
	.dc.b	$DA,$FE,$E9,$3A,$79,$4E,$9F,$E8,$DF,$3B,$F9,$2C,$FF,$2A,$FF,$6E
	.dc.b	$D3,$D1,$C4,$8D,$46,$1D,$57,$F1,$C7,$2D,$F5,$7F,$C2,$71,$8B,$FF
	.dc.b	$7D,$BC,$1F,$87,$E0,$E7,$6F,$F3,$3D,$7F,$09,$EE,$FE,$14,$DC,$3F
	.dc.b	$84,$36,$FF,$84,$37,$1F,$9D,$7C,$AB,$F8,$75,$F2,$7F,$B2,$7D,$3F
	.dc.b	$F8,$21,$FE,$66,$BE,$1B,$FC,$52,$E7,$16,$88,$D5,$72,$F5,$F7,$7C
	.dc.b	$1B,$1F,$E2,$9F,$A1,$7C,$BF,$F8,$A7,$E6,$DF,$4F,$FE,$29,$9C,$D7
	.dc.b	$F8,$A6,$72,$F1,$AE,$15,$B6,$C7,$F7,$DE,$47,$D3,$7E,$37,$E1,$7F
	.dc.b	$6F,$B7,$A2,$79,$BF,$E8,$3E,$FF,$EA,$31,$C2,$FF,$D1,$11,$8C,$79
	.dc.b	$3A,$C2,$DF,$39,$BB,$FE,$42,$6E,$1D,$B1,$F0,$CF,$E5,$2B,$26,$AA
	.dc.b	$E3,$E0,$C3,$A3,$6B,$83,$70,$F3,$4F,$A4,$D9,$47,$CF,$FF,$86,$BC
	.dc.b	$1E,$91,$8E,$FF,$EC,$6E,$FE,$9D,$F3,$5E,$C8,$D5,$70,$BE,$51,$D5
	.dc.b	$BA,$FC,$C8,$5F,$2C,$B8,$63,$DA,$B5,$7C,$36,$FF,$64,$37,$0F,$E1
	.dc.b	$F8,$71,$7B,$5F,$1A,$DB,$63,$F8,$AD,$71,$7C,$39,$7A,$63,$04,$9E
	.dc.b	$13,$AC,$7C,$B9,$48,$EE,$FC,$6C,$7C,$59,$7A,$2F,$8E,$1C,$EB,$58
	.dc.b	$F7,$9C,$92,$33,$8F,$5B,$E5,$17,$2F,$F7,$46,$3D,$9F,$6C,$67,$2F
	.dc.b	$6E,$1E,$8A,$D4,$62,$74,$FD,$63,$ED,$63,$E3,$8C,$E5,$F1,$6B,$8D
	.dc.b	$6F,$1F,$1B,$F0,$0E,$39,$70,$C7,$D5,$18,$35,$3C,$09,$97,$AB,$2F
	.dc.b	$04,$EA,$3D,$6F,$19,$DB,$E9,$34,$7E,$02,$78,$C6,$A3,$CD,$3D,$6F
	.dc.b	$C9,$97,$C0,$C7,$D7,$97,$D3,$BE,$67,$DD,$2B,$F9,$87,$FA,$D9,$E1
	.dc.b	$29,$F2,$4A,$ED,$FC,$AF,$17,$F0,$A7,$67,$EB,$1C,$3C,$1A,$F0,$DB
	.dc.b	$6E,$DF,$D5,$47,$3C,$BF,$46,$DC,$BC,$BE,$0B,$E0,$FC,$F5,$AB,$E2
	.dc.b	$4C,$3F,$7D,$5B,$71,$3B,$7F,$6D,$87,$57,$FB,$DC,$ED,$CD,$70,$C3
	.dc.b	$D1,$1D,$30,$F0,$C6,$30,$E8,$47,$CF,$1A,$9F,$D9,$E1,$DD,$F8,$07
	.dc.b	$63,$A1,$DC,$DD,$D5,$87,$97,$0E,$D1,$A3,$93,$FC,$3E,$35,$87,$94
	.dc.b	$93,$B9,$D6,$1E,$58,$F1,$C7,$46,$CF,$A9,$F1,$62,$C7,$D4,$DC,$7F
	.dc.b	$94,$E3,$E0,$AE,$D7,$E0,$79,$9F,$3B,$1F,$05,$F1,$8F,$4C,$EE,$B9
	.dc.b	$47,$02,$4E,$F1,$F4,$C7,$93,$1F,$19,$F3,$6B,$8D,$6E,$FC,$4D,$AF
	.dc.b	$11,$C7,$1F,$9D,$5C,$2F,$9C,$63,$1F,$3C,$62,$BC,$E7,$22,$9E,$67
	.dc.b	$C5,$1F,$81,$E0,$7D,$3F,$FA,$02,$BF,$AC,$35,$FE,$AD,$1A,$E8,$E1
	.dc.b	$3F,$E8,$89,$DC,$60,$8F,$2F,$0B,$66,$1E,$73,$D7,$5F,$ED,$AF,$38
	.dc.b	$CC,$EE,$73,$5C,$A7,$FB,$77,$C2,$3A,$3F,$CA,$50,$EC,$FF,$2A,$7C
	.dc.b	$23,$E9,$9F,$4F,$FC,$2A,$73,$18,$AF,$F0,$E7,$36,$4F,$F3,$4E,$3D
	.dc.b	$5C,$43,$0E,$98,$61,$B1,$C8,$FF,$38,$E1,$FE,$4E,$FA,$77,$C2,$F2
	.dc.b	$E0,$43,$E9,$70,$DB,$E3,$97,$FB,$B2,$38,$C6,$0F,$F1,$CB,$87,$FD
	.dc.b	$62,$F1,$6C,$F4,$BF,$4C,$3E,$19,$EC,$E1,$7D,$3D,$31,$12,$7F,$D7
	.dc.b	$53,$A6,$CE,$5F,$A1,$8F,$F3,$69,$EB,$1A,$21,$A3,$FD,$2A,$E9,$FF
	.dc.b	$3A,$F3,$7C,$5F,$23,$FC,$B1,$1F,$E7,$AF,$13,$13,$23,$04,$38,$9F
	.dc.b	$E8,$BF,$07,$EA,$79,$CF,$A7,$A7,$DA,$E4,$79,$BF,$E8,$EB,$E0,$7C
	.dc.b	$DE,$45,$F1,$8C,$35,$FF,$78,$61,$F2,$78,$7E,$57,$C7,$49,$F8,$F1
	.dc.b	$97,$E4,$C6,$5F,$84,$7D,$38,$D3,$8D,$93,$BE,$5F,$CE,$C7,$DD,$7F
	.dc.b	$D7,$5F,$35,$F9,$1F,$07,$EE,$5F,$A5,$AF,$FB,$12,$3E,$07,$A2,$78
	.dc.b	$5C,$9F,$F6,$6C,$7E,$57,$E9,$FC,$31,$F8,$FE,$94,$3E,$6F,$46,$F3
	.dc.b	$95,$B9,$D3,$67,$2F,$C0,$F3,$1F,$4D,$F9,$1F,$A7,$93,$F9,$D1,$3A
	.dc.b	$70,$D7,$94,$FE,$72,$31,$B7,$C6,$FA,$7C,$4C,$AE,$2F,$E7,$E3,$F6
	.dc.b	$5E,$F7,$FA,$6A,$7D,$CF,$E0,$1B,$C7,$AF,$42,$46,$08,$7E,$6F,$65
	.dc.b	$73,$8F,$CF,$72,$BE,$9D,$EC,$7F,$07,$6C,$FE,$3D,$AF,$E9,$7A,$7F
	.dc.b	$2B,$F0,$9B,$A9,$3B,$7F,$35,$D4,$9F,$DE,$FD,$3F,$FA,$0A,$BE,$53
	.dc.b	$FE,$EC,$6F,$B3,$AF,$4E,$46,$08,$7E,$52,$79,$36,$3A,$CC,$8E,$B1
	.dc.b	$FC,$87,$E9,$FD,$3C,$7F,$CC,$1B,$E1,$AB,$F9,$44,$61,$7C,$63,$E2
	.dc.b	$3F,$C7,$77,$FC,$E3,$E2,$3D,$15,$FC,$1E,$B4,$FF,$32,$6B,$FB,$78
	.dc.b	$DC,$B0,$F0,$53,$E9,$DF,$4D,$C5,$B8,$7F,$2B,$9D,$4F,$F0,$5F,$8C
	.dc.b	$7D,$EF,$7F,$A8,$DD,$F8,$DB,$97,$78,$F5,$C6,$32,$F8,$1D,$0F,$90
	.dc.b	$FA,$D5,$FA,$B6,$BC,$F2,$EC,$7F,$45,$8C,$E1,$D6,$B9,$71,$E4,$97
	.dc.b	$FD,$0F,$E2,$5F,$50,$F8,$AF,$E3,$DF,$27,$6B,$F8,$E7,$CF,$7F,$1F
	.dc.b	$7F,$2A,$DC,$BD,$5D,$6A,$E1,$5F,$93,$6C,$FF,$6C,$AE,$1B,$7D,$DE
	.dc.b	$D4,$62,$D8,$73,$C7,$C6,$7F,$67,$9D,$57,$0F,$37,$37,$FC,$3F,$E9
	.dc.b	$DF,$33,$FD,$9D,$7D,$45,$31,$12,$78,$46,$5E,$51,$97,$F6,$D3,$C9
	.dc.b	$D9,$A3,$FB,$66,$5F,$D9,$3C,$9F,$EF,$9C,$5F,$C9,$E7,$FA,$23,$5F
	.dc.b	$DB,$7A,$DF,$C7,$9B,$3C,$3C,$D9,$1D,$27,$5C,$1F,$39,$BC,$1F,$27
	.dc.b	$0E,$93,$FE,$18,$6C,$7F,$84,$A7,$74,$97,$CD,$BE,$1B,$FC,$18,$BB
	.dc.b	$A9,$88,$95,$FC,$EE,$74,$D9,$CB,$FC,$6E,$3A,$39,$21,$99,$FE,$D5
	.dc.b	$F3,$9F,$E4,$5C,$BF,$C5,$FF,$51,$FD,$ED,$B9,$6B,$E0,$26,$1F,$DB
	.dc.b	$6B,$5A,$7E,$25,$B1,$F8,$8C,$3A,$4F,$F8,$C9,$B1,$FE,$31,$9D,$F2
	.dc.b	$1E,$3C,$0A,$DE,$CF,$9F,$7D,$47,$F8,$C5,$FC,$DE,$5F,$B6,$6E,$5F
	.dc.b	$35,$DB,$FD,$85,$AF,$F3,$F7,$75,$D7,$A3,$A5,$FF,$15,$F6,$FF,$CB
	.dc.b	$5B,$FF,$05,$F0,$D1,$8C,$BF,$8A,$B7,$2F,$B9,$F4,$5F,$80,$7F,$92
	.dc.b	$C6,$6B,$8F,$C2,$F5,$3F,$C9,$55,$BA,$7D,$23,$E9,$FF,$CC,$51,$FB
	.dc.b	$4F,$9A,$FF,$74,$FD,$33,$E9,$FF,$DD,$7E,$0A,$73,$18,$7F,$DD,$F8
	.dc.b	$73,$DB,$F9,$D5,$6F,$0F,$E7,$93,$97,$B1,$FC,$FF,$0E,$CF,$0B,$EA
	.dc.b	$D3,$8E,$3E,$6B,$E9,$1E,$A8,$C4,$F0,$8E,$A2,$BA,$B8,$BF,$15,$F2
	.dc.b	$6D,$71,$8E,$B5,$C1,$D5,$F6,$BF,$1E,$1E,$57,$B6,$3F,$0D,$F6,$35
	.dc.b	$F3,$46,$F1,$EF,$1A,$39,$61,$DF,$1F,$97,$87,$79,$E2,$7B,$27,$4F
	.dc.b	$13,$D9,$1B,$AF,$9D,$1F,$DA,$CF,$39,$4A,$57,$6F,$DA,$FA,$77,$CB
	.dc.b	$EE,$3E,$CC,$61,$C4,$A3,$11,$CD,$AF,$1F,$8C,$E4,$F8,$DA,$FF,$7F
	.dc.b	$6A,$69,$F3,$7C,$7F,$E2,$DC,$5F,$3E,$78,$1F,$7A,$32,$FE,$00,$CE
	.dc.b	$DF,$86,$8C,$BF,$88,$8C,$E9,$F1,$67,$2F,$9E,$3A,$77,$FF,$77,$AE
	.dc.b	$11,$E5,$33,$7E,$37,$11,$DE,$78,$57,$38,$C5,$6F,$0F,$65,$6A,$FE
	.dc.b	$A4,$62,$75,$1D,$8F,$61,$DA,$FD,$15,$C6,$3E,$61,$BB,$F3,$DF,$38
	.dc.b	$F8,$DE,$EF,$F8,$47,$7F,$F7,$0E,$48,$C6,$5F,$0F,$E8,$AD,$FF,$8C
	.dc.b	$EB,$53,$FE,$6F,$6C,$7F,$A5,$1F,$E0,$91,$9A,$E3,$29,$CB,$BF,$9A
	.dc.b	$EB,$E0,$FF,$54,$F4,$37,$49,$CF,$86,$9C,$C6,$1B,$8F,$FB,$76,$F9
	.dc.b	$3E,$5F,$0F,$99,$EA,$ED,$FE,$79,$E7,$62,$74,$71,$FC,$6B,$FC,$A0
	.dc.b	$E3,$4F,$F6,$86,$09,$CC,$7B,$F0,$F4,$C6,$37,$F9,$71,$89,$F8,$3F
	.dc.b	$56,$E0,$A7,$CC,$C5,$F9,$03,$E0,$33,$1F,$E9,$BA,$E0,$FF,$BE,$4D
	.dc.b	$D6,$AF,$FD,$C9,$18,$38,$47,$5B,$E3,$3F,$EC,$8A,$B8,$FF,$94,$6F
	.dc.b	$84,$E7,$E3,$3C,$E7,$C7,$E1,$FA,$53,$E3,$8C,$BF,$EB,$DD,$3F,$AA
	.dc.b	$B9,$31,$E0,$EA,$FF,$D3,$D3,$A7,$B4,$E8,$FF,$5C,$7C,$DF,$68,$FF
	.dc.b	$6E,$B7,$1F,$36,$6F,$B6,$FC,$5C,$E6,$FF,$8E,$36,$FF,$CF,$54,$F8
	.dc.b	$8D,$F2,$FC,$36,$DF,$CB,$9D,$3E,$29,$D3,$FE,$CB,$E8,$7D,$D7,$E6
	.dc.b	$9B,$1F,$E9,$5C,$FF,$CF,$59,$3F,$00,$FF,$35,$4F,$1B,$E0,$7E,$AA
	.dc.b	$75,$F2,$F8,$9E,$47,$FA,$7E,$32,$F1,$33,$A2,$32,$B0,$FF,$45,$3F
	.dc.b	$C9,$EF,$B6,$0C,$3F,$63,$CC,$F8,$EA,$7E,$D2,$5F,$88,$E3,$3F,$F7
	.dc.b	$55,$6D,$BE,$B6,$72,$F0,$47,$F4,$B9,$FF,$36,$BA,$3F,$DE,$5D,$6F
	.dc.b	$88,$F0,$DF,$F0,$28,$C4,$FF,$A4,$7C,$D7,$A7,$9F,$9D,$FE,$1F,$E0
	.dc.b	$FF,$A4,$E3,$15,$D2,$7F,$75,$B3,$E4,$73,$AF,$FB,$AE,$31,$5F,$E9
	.dc.b	$DF,$23,$E1,$3F,$2B,$37,$D8,$7F,$57,$8D,$71,$7F,$05,$8C,$61,$EE
	.dc.b	$3F,$DA,$5F,$8B,$BC,$3F,$E3,$6D,$BF,$F5,$57,$D3,$FF,$BA,$DF,$AF
	.dc.b	$49,$D3,$CC,$C5,$7F,$2C,$E7,$7E,$21,$F0,$AF,$8D,$F0,$9F,$E6,$D3
	.dc.b	$9B,$E5,$39,$AF,$C9,$4E,$6B,$F0,$73,$9A,$FC,$34,$E6,$BF,$1D,$E6
	.dc.b	$BE,$6B,$F2,$8F,$E7,$2B,$E0,$75,$FF,$80,$27,$3D,$7A,$6D,$FE,$52
	.dc.b	$74,$DB,$E5,$97,$E9,$17,$E3,$BD,$8E,$FC,$76,$B8,$45,$CB,$C3,$5F
	.dc.b	$69,$B1,$FD,$2B,$EA,$5F,$33,$4F,$E9,$3F,$50,$FA,$5C,$63,$F4,$AB
	.dc.b	$6F,$F4,$59,$FA,$D8,$7D,$39,$CB,$FC,$FA,$33,$9F,$EB,$27,$2F,$FB
	.dc.b	$DF,$B9,$3A,$BF,$F1,$65,$69,$5F,$FA,$B6,$74,$FE,$E7,$37,$D5,$FD
	.dc.b	$0F,$AA,$FA,$2D,$BF,$94,$DA,$F8,$6D,$CB,$E6,$7D,$9F,$D8,$3E,$CF
	.dc.b	$C6,$DB,$FA,$51,$6F,$E5,$CD,$AF,$66,$1F,$2C,$FD,$7F,$3F,$F1,$8A
	.dc.b	$F8,$91,$9C,$3E,$47,$D3,$D9,$F8,$CD,$BF,$89,$A2,$31,$5F,$0B,$6F
	.dc.b	$8F,$1F,$8E,$FB,$A9,$3F,$87,$FC,$12,$E3,$F8,$09,$B8,$FE,$26,$6D
	.dc.b	$7E,$63,$E6,$3F,$2F,$F0,$1F,$98,$6C,$7E,$45,$B8,$7E,$29,$B7,$F6
	.dc.b	$FC,$3F,$C5,$1F,$88,$F8,$2B,$7F,$87,$8F,$EE,$9F,$81,$FF,$21,$C3
	.dc.b	$F0,$D1,$8F,$13,$71,$FD,$C2,$75,$87,$5B,$FE,$ED,$18,$7F,$5F,$7F
	.dc.b	$AE,$FA,$DF,$D7,$B6,$3F,$50,$DB,$FD,$0B,$6F,$F5,$5F,$81,$FF,$49
	.dc.b	$5F,$57,$FB,$88,$C3,$6F,$F6,$93,$A7,$F6,$93,$A7,$F8,$AF,$D6,$FF
	.dc.b	$24,$C3,$FC,$DD,$C3,$FE,$85,$87,$C8,$AB,$FC,$E9,$39,$F1,$F1,$7C
	.dc.b	$0F,$CF,$DF,$0A,$DC,$62,$77,$E8,$FE,$16,$37,$86,$30,$E9,$7C,$25
	.dc.b	$87,$37,$B1,$8C,$38,$3C,$A7,$84,$FA,$F0,$ED,$3C,$B0,$DB,$6B,$71
	.dc.b	$8A,$E1,$87,$C1,$8D,$FD,$33,$CC,$73,$F9,$63,$0F,$83,$E6,$F7,$3A
	.dc.b	$CB,$CF,$5A,$6C,$67,$3F,$2E,$4B,$E3,$87,$5A,$D6,$5F,$EB,$7C,$79
	.dc.b	$61,$C2,$F7,$8F,$4C,$38,$1E,$3C,$EC,$64,$95,$C6,$32,$FC,$6B,$EB
	.dc.b	$87,$7C,$7C,$37,$F7,$31,$EF,$7B,$9C,$CA,$F8,$BB,$C3,$AE,$1C,$F1
	.dc.b	$B3,$93,$C7,$19,$69,$9C,$3B,$4F,$08,$ED,$87,$83,$2F,$91,$86,$71
	.dc.b	$F0,$D6,$AF,$E1,$3D,$0C,$16,$B8,$B9,$9E,$27,$28,$CE,$1E,$29,$D6
	.dc.b	$1E,$1C,$BE,$1E,$5E,$9A,$D1,$F3,$6B,$78,$FA,$70,$E7,$5B,$24,$FF
	.dc.b	$46,$C7,$C1,$87,$48,$C6,$3F,$3A,$73,$97,$BE,$30,$78,$DC,$63,$DE
	.dc.b	$F9,$5F,$C0,$3F,$DF,$D7,$F9,$36,$DF,$E5,$DB,$5F,$EF,$49,$FE,$DB
	.dc.b	$BD,$F0,$BF,$87,$D7,$FC,$01,$9C,$B6,$32,$F6,$77,$A7,$E3,$5B,$9F
	.dc.b	$E6,$DB,$3F,$D9,$BE,$B7,$FC,$AC,$FE,$B6,$32,$FE,$35,$B7,$E0,$BF
	.dc.b	$DE,$D7,$0A,$FD,$7D,$68,$98,$7F,$0B,$9D,$57,$8E,$3B,$12,$BF,$86
	.dc.b	$CE,$B0,$F3,$4E,$6B,$85,$6B,$7F,$EF,$AF,$82,$75,$5C,$23,$46,$0F
	.dc.b	$BD,$87,$91,$C9,$C4,$97,$D8,$F6,$DF,$67,$15,$D5,$B3,$F2,$67,$53
	.dc.b	$92,$47,$82,$FE,$14,$EA,$30,$78,$27,$6B,$0F,$84,$EA,$FB,$BF,$4E
	.dc.b	$F9,$BE,$D3,$07,$33,$C8,$67,$1F,$B1,$7D,$AF,$AC,$F0,$BE,$B5,$A8
	.dc.b	$E2,$73,$35,$5C,$70,$F5,$46,$0F,$34,$65,$F0,$63,$F3,$EF,$E5,$C6
	.dc.b	$31,$F4,$63,$E2,$C7,$C1,$3C,$C9,$1E,$61,$F7,$1E,$06,$C7,$E3,$63
	.dc.b	$1A,$C8,$C1,$0F,$C6,$B9,$7F,$1F,$1C,$CC,$61,$F9,$C9,$C4,$F1,$8C
	.dc.b	$3F,$D9,$7F,$01,$FF,$56,$3F,$E8,$E5,$F4,$9F,$94,$E4,$FF,$61,$C6
	.dc.b	$27,$5E,$AE,$78,$BF,$33,$3C,$A7,$FD,$28,$ED,$B7,$B7,$FC,$B7,$1B
	.dc.b	$BF,$F6,$7F,$B5,$FE,$A3,$9F,$F4,$FA,$F7,$B8,$47,$FB,$C5,$B1,$B8
	.dc.b	$FF,$51,$36,$79,$CE,$8D,$FA,$3C,$BE,$9B,$FD,$B1,$7D,$BD,$8E,$66
	.dc.b	$7B,$A4,$64,$87,$FA,$BE,$78,$B6,$36,$D3,$D3,$ED,$F9,$A3,$CF,$F5
	.dc.b	$12,$7D,$B1,$97,$DF,$19,$7D,$71,$EE,$F9,$69,$1F,$11,$F8,$3E,$FE
	.dc.b	$23,$E0,$C5,$D3,$E5,$CF,$BF,$E9,$FE,$3F,$C4,$CB,$FF,$65,$E6,$9D
	.dc.b	$47,$D0,$9D,$3E,$F8,$CB,$4E,$2F,$D2,$F8,$7F,$AB,$F4,$9F,$F0,$02
	.dc.b	$7F,$DA,$ED,$7F,$DA,$CD,$C7,$EF,$C7,$0F,$81,$8F,$07,$EE,$BD,$3A
	.dc.b	$3A,$FD,$24,$8E,$B1,$97,$BC,$65,$EF,$1F,$88,$FC,$83,$FC,$61,$1E
	.dc.b	$4F,$17,$CE,$67,$C4,$91,$92,$1F,$EE,$D9,$E2,$D8,$DB,$4F,$7F,$DC
	.dc.b	$FC,$2F,$92,$91,$F0,$9F,$8B,$F2,$58,$8F,$87,$17,$4F,$F3,$0C,$35
	.dc.b	$57,$CC,$F3,$7E,$6F,$C3,$4A,$F9,$B1,$97,$E7,$C6,$5F,$9F,$1F,$07
	.dc.b	$D8,$FA,$BF,$43,$27,$FD,$97,$8F,$83,$5F,$F6,$4C,$63,$C1,$FE,$A7
	.dc.b	$18,$8F,$B3,$5C,$5F,$89,$7D,$5E,$07,$FB,$5A,$78,$5F,$59,$FC,$2F
	.dc.b	$8F,$FC,$07,$E1,$DF,$29,$F8,$3C,$BF,$DB,$BF,$0E,$F9,$D7,$B9,$FC
	.dc.b	$1E,$30,$F1,$AE,$1A,$7F,$B9,$3F,$21,$FF,$7C,$BE,$2A,$7E,$16,$E0
	.dc.b	$E4,$D9,$F3,$8A,$E2,$B6,$FF,$49,$E9,$9F,$5B,$F8,$8D,$6A,$7E,$BB
	.dc.b	$75,$FF,$7A,$D6,$EF,$AC,$BD,$FE,$3F,$0D,$27,$31,$D3,$D8,$EA,$C8
	.dc.b	$E3,$19,$69,$0F,$36,$1F,$E0,$8F,$A3,$FC,$3C,$7B,$39,$BF,$17,$9F
	.dc.b	$FA,$B3,$E1,$FF,$D7,$BE,$F7,$F3,$16,$D6,$F0,$FE,$5B,$7C,$14,$27
	.dc.b	$6F,$80,$C7,$2A,$F8,$D4,$9E,$84,$38,$57,$FB,$3B,$DB,$EF,$EF,$FF
	.dc.b	$3C,$FC,$89,$F7,$74,$CD,$FA,$E7,$0F,$D8,$9D,$B9,$21,$E5,$19,$A7
	.dc.b	$9C,$66,$7F,$C9,$BA,$7E,$C3,$F2,$47,$C5,$7E,$3F,$E4,$BF,$AC,$4F
	.dc.b	$FB,$53,$A5,$18,$DB,$FC,$6F,$7C,$31,$FC,$5B,$5F,$8D,$5C,$3D,$7E
	.dc.b	$87,$F5,$A9,$D6,$1E,$1C,$70,$4B,$FC,$D7,$E4,$8F,$27,$FB,$0F,$CA
	.dc.b	$7E,$66,$7D,$DF,$8D,$3D,$6F,$EC,$AD,$8F,$CE,$CE,$F1,$FE,$1A,$DD
	.dc.b	$7F,$C9,$F5,$BA,$E7,$7D,$BB,$E4,$72,$6B,$E0,$F5,$BF,$B9,$FE,$01
	.dc.b	$FB,$EC,$3F,$0F,$F0,$9F,$E1,$C7,$FB,$F3,$6B,$F5,$F3,$B7,$F1,$AD
	.dc.b	$8D,$47,$E3,$5B,$9F,$F2,$49,$E1,$26,$93,$C6,$FE,$0D,$F9,$2B,$A7
	.dc.b	$CB,$CF,$F1,$3D,$C9,$87,$8A,$32,$FF,$02,$DD,$3A,$8F,$C5,$63,$F8
	.dc.b	$97,$F8,$3F,$8C,$F8,$4F,$F0,$9A,$F2,$E9,$88,$93,$C1,$AF,$E5,$67
	.dc.b	$E9,$3F,$AE,$F9,$57,$9F,$FC,$46,$1F,$3A,$F9,$FF,$F4,$2D,$BE,$83
	.dc.b	$FD,$0F,$39,$AF,$DB,$CE,$5F,$DC,$C6,$56,$1D,$B8,$B3,$87,$FA,$42
	.dc.b	$31,$3E,$08,$EF,$87,$66,$C7,$03,$D1,$8F,$9F,$0E,$D7,$C9,$F0,$63
	.dc.b	$E3,$AE,$87,$9B,$0E,$8F,$2C,$7A,$3E,$6C,$3A,$63,$E0,$8E,$27,$F4
	.dc.b	$8A,$E2,$EE,$B8,$E3,$E1,$AE,$18,$F8,$AF,$73,$9B,$F0,$1D,$31,$F3
	.dc.b	$56,$E3,$9E,$3E,$27,$71,$97,$85,$F4,$9D,$47,$8C,$FF,$5D,$9D,$8B
	.dc.b	$7F,$39,$B1,$F7,$DC,$FA,$2F,$91,$E3,$1F,$85,$8C,$38,$AF,$F0,$2C
	.dc.b	$EA,$3E,$EC,$E9,$C7,$C2,$E5,$FE,$76,$D4,$D2,$3F,$3B,$CE,$7D,$03
	.dc.b	$A7,$EA,$CF,$23,$51,$97,$EA,$99,$D3,$F2,$15,$CA,$B8,$CF,$80,$E2
	.dc.b	$E2,$FF,$A7,$FC,$A7,$0C,$3A,$3F,$99,$AF,$03,$9A,$DE,$FF,$1A,$FF
	.dc.b	$13,$87,$7B,$F1,$9D,$F1,$E0,$E2,$77,$8F,$A6,$7B,$4E,$EF,$FD,$E7
	.dc.b	$7C,$E7,$37,$C3,$0F,$45,$FA,$8E,$85,$9D,$B6,$F9,$FB,$DF,$1F,$87
	.dc.b	$FC,$7F,$8A,$31,$87,$83,$E7,$AE,$1F,$E0,$C9,$AA,$B8,$57,$E0,$5B
	.dc.b	$1F,$88,$6B,$F9,$26,$BE,$CE,$DF,$F2,$7B,$67,$94,$62,$3F,$C3,$6D
	.dc.b	$7C,$0F,$93,$B5,$F3,$AF,$A0,$F2,$E0,$AD,$4E,$70,$FF,$17,$CB,$E1
	.dc.b	$F6,$7B,$BC,$97,$D1,$BE,$87,$37,$C2,$BC,$57,$C6,$3E,$04,$60,$8F
	.dc.b	$F9,$4B,$3F,$EE,$FA,$72,$3F,$A6,$79,$9F,$E1,$C9,$FF,$20,$FD,$0B
	.dc.b	$83,$E6,$B6,$3F,$CB,$63,$E8,$5F,$0B,$AA,$93,$FB,$1A,$E1,$87,$EE
	.dc.b	$1A,$FF,$02,$6B,$FE,$2C,$0F,$0E,$3F,$D5,$4F,$7A,$EB,$5F,$E2,$F8
	.dc.b	$C1,$E8,$F8,$CF,$39,$F2,$78,$7E,$84,$64,$93,$E3,$9F,$B5,$87,$FB
	.dc.b	$AE,$7D,$11,$83,$E9,$66,$F2,$FA,$CF,$F9,$AE,$30,$FF,$9A,$3D,$4E
	.dc.b	$F8,$7F,$8B,$DB,$9F,$AB,$57,$C4,$FE,$62,$73,$7F,$05,$E3,$7F,$C9
	.dc.b	$9A,$E7,$97,$F2,$11,$88,$FF,$29,$BC,$9F,$E2,$97,$FD,$E2,$BC,$95
	.dc.b	$FA,$87,$EC,$78,$DF,$A8,$EB,$7C,$ED,$3F,$5B,$8B,$E8,$7F,$C8,$F7
	.dc.b	$FE,$84,$AD,$B6,$BF,$45,$18,$CB,$F8,$84,$60,$93,$FE,$A2,$BE,$9F
	.dc.b	$2B,$DE,$31,$82,$3F,$77,$CA,$F8,$EA,$7E,$FA,$5F,$17,$8B,$FF,$4B
	.dc.b	$D6,$DF,$E8,$0D,$CB,$FA,$3B,$5F,$EA,$2D,$7F,$20,$FF,$91,$E3,$8F
	.dc.b	$A5,$FB,$D8,$C4,$F2,$8F,$E0,$5D,$5F,$E6,$AF,$7A,$FB,$46,$93,$0F
	.dc.b	$EA,$5D,$EF,$89,$FE,$10,$D8,$CC,$60,$EB,$E4,$FC,$38,$FE,$30,$DC
	.dc.b	$BF,$8C,$35,$FE,$66,$D7,$F5,$5F,$41,$FC,$0D,$FF,$40,$37,$A3,$FD
	.dc.b	$73,$F3,$BF,$E3,$5B,$FF,$5F,$CE,$5E,$3E,$0F,$F1,$D8,$E5,$7F,$DE
	.dc.b	$1A,$EA,$7F,$B7,$47,$18,$FF,$63,$7D,$07,$C3,$9E,$B5,$FC,$E3,$04
	.dc.b	$E6,$BF,$D4,$3E,$63,$E1,$3F,$22,$FE,$2E,$BF,$D6,$CE,$70,$FC,$53
	.dc.b	$97,$F4,$2E,$6B,$9F,$47,$2B,$E6,$F4,$FA,$09,$7C,$BD,$DB,$7E,$29
	.dc.b	$B8,$F8,$26,$D7,$F1,$5F,$A0,$7C,$D7,$F5,$06,$C7,$F2,$DF,$9F,$7A
	.dc.b	$39,$A9,$3F,$E1,$3A,$DE,$5F,$E1,$A6,$BF,$E2,$46,$BD,$E7,$FA,$65
	.dc.b	$FD,$78,$FF,$5D,$4F,$2C,$7F,$C1,$DD,$2A,$D4,$7C,$8F,$77,$E5,$FD
	.dc.b	$07,$CC,$6E,$5F,$29,$B8,$FD,$26,$D7,$D0,$F8,$9F,$A6,$7D,$4F,$8D
	.dc.b	$B7,$F4,$A2,$DF,$DA,$9B,$5F,$E2,$18,$66,$C7,$E4,$1F,$DE,$BB,$1C
	.dc.b	$FF,$C3,$AB,$DB,$3A,$CB,$E4,$3F,$DB,$EF,$94,$60,$93,$FE,$A5,$D3
	.dc.b	$FD,$25,$1F,$EF,$6C,$7C,$55,$C2,$B5,$ED,$7E,$3B,$DE,$97,$F8,$CF
	.dc.b	$88,$B5,$F9,$49,$B8,$7E,$4A,$6D,$7E,$43,$E7,$FF,$2F,$F4,$1F,$98
	.dc.b	$6C,$7E,$59,$B8,$7E,$79,$B7,$E4,$F0,$3F,$14,$7D,$EF,$84,$B7,$E9
	.dc.b	$3F,$21,$1F,$04,$8F,$E2,$DB,$7F,$D8,$FE,$83,$F9,$05,$F5,$AF,$EE
	.dc.b	$F1,$87,$F4,$F5,$FA,$EF,$8F,$FD,$7B,$63,$F5,$AD,$BF,$DA,$B6,$FF
	.dc.b	$67,$F0,$BF,$D1,$97,$C6,$FE,$E2,$30,$DB,$FE,$81,$3A,$7F,$73,$3A
	.dc.b	$7D,$D6,$1F,$41,$FC,$7E,$3F,$D0,$ED,$8F,$F0,$33,$5E,$0D,$78,$F9
	.dc.b	$9F,$E7,$1F,$3B,$FD,$A9,$1A,$F9,$6C,$71,$F2,$AB,$39,$C3,$FD,$07
	.dc.b	$39,$6C,$64,$FF,$42,$C6,$5A,$66,$FB,$BE,$29,$E2,$FB,$2F,$BB,$93
	.dc.b	$89,$98,$DF,$D0,$7F,$47,$ED,$BE,$49,$E7,$9B,$E6,$39,$D6,$F0,$E6
	.dc.b	$DC,$BC,$6D,$CF,$E4,$D7,$0C,$7A,$5F,$06,$DF,$06,$D7,$0C,$BC,$79
	.dc.b	$4B,$E0,$F9,$30,$E8,$E7,$0E,$4E,$6B,$17,$FE,$3F,$C7,$C9,$87,$E0
	.dc.b	$5A,$FE,$0B,$0E,$6F,$62,$61,$F3,$B2,$F8,$58,$74,$9E,$59,$7A,$1B
	.dc.b	$97,$85,$E3,$97,$86,$75,$0C,$BC,$38,$74,$8D,$9C,$8A,$79,$70,$F0
	.dc.b	$3C,$6B,$9D,$72,$CB,$D3,$97,$C2,$AE,$38,$75,$8C,$13,$1F,$26,$5E
	.dc.b	$78,$C5,$78,$5A,$F8,$4F,$1E,$1F,$20,$F0,$65,$E4,$71,$97,$BB,$1F
	.dc.b	$06,$3D,$AF,$8C,$E7,$2F,$3C,$E7,$1E,$F7,$C2,$FC,$53,$92,$3F,$B2
	.dc.b	$AE,$2E,$88,$79,$EB,$D3,$87,$E5,$E3,$FD,$00,$7C,$9C,$BF,$DF,$11
	.dc.b	$98,$F3,$1A,$7F,$B8,$EF,$AF,$03,$F8,$8C,$F2,$3E,$D3,$87,$83,$5F
	.dc.b	$C8,$3C,$2F,$F4,$AD,$C7,$35,$FB,$8F,$A0,$FF,$2A,$5F,$ED,$2B,$4F
	.dc.b	$E5,$DA,$F8,$27,$86,$FF,$E9,$BC,$3A,$C7,$12,$4F,$F0,$0B,$EA,$FE
	.dc.b	$E2,$F9,$E3,$EE,$8C,$57,$47,$8D,$F4,$72,$73,$C3,$C1,$8F,$AA,$33
	.dc.b	$3B,$3F,$8A,$E3,$E0,$71,$3B,$6C,$FF,$33,$9D,$4E,$49,$1F,$06,$FC
	.dc.b	$8F,$68,$CB,$F0,$CF,$21,$E8,$73,$87,$B6,$FC,$78,$7A,$2B,$A1,$CC
	.dc.b	$B1,$A7,$91,$8A,$EF,$87,$9A,$30,$43,$C5,$3C,$5E,$86,$27,$8D,$78
	.dc.b	$63,$07,$1B,$F3,$4F,$FB,$56,$FB,$4E,$89,$3A,$7E,$45,$73,$8C,$46
	.dc.b	$B1,$FC,$3D,$72,$9C,$92,$33,$5C,$5F,$C1,$93,$1F,$9F,$8F,$D6,$C3
	.dc.b	$A4,$63,$1F,$36,$7E,$BC,$9F,$41,$4C,$44,$BF,$CF,$46,$5F,$D0,$46
	.dc.b	$5F,$CB,$46,$5F,$CC,$46,$77,$FD,$CD,$7F,$53,$9F,$EF,$51,$89,$FF
	.dc.b	$7E,$36,$32,$7F,$BF,$FE,$7D,$F4,$1F,$F0,$07,$93,$FC,$91,$0F,$93
	.dc.b	$C5,$ED,$B7,$FE,$07,$9E,$2F,$F8,$28,$38,$55,$D2,$78,$C6,$9B,$3B
	.dc.b	$9F,$F2,$C7,$AE,$F3,$B6,$BD,$DA,$62,$26,$1E,$19,$D3,$67,$2F,$8B
	.dc.b	$1F,$F0,$F2,$F9,$EF,$F6,$F3,$C7,$D5,$EE,$79,$A3,$31,$C4,$FF,$13
	.dc.b	$38,$7F,$94,$B5,$EA,$D7,$7E,$C7,$F2,$56,$9F,$1E,$3A,$E5,$FE,$77
	.dc.b	$6D,$F9,$B1,$7B,$5F,$0B,$0F,$87,$F4,$13,$1F,$81,$19,$7E,$14,$65
	.dc.b	$F4,$0A,$EA,$FF,$8F,$D7,$D0,$7C,$E2,$47,$CB,$F7,$AB,$CA,$76,$FC
	.dc.b	$F9,$FA,$B3,$F1,$FE,$83,$EB,$FC,$4C,$AF,$74,$65,$F8,$5A,$23,$E2
	.dc.b	$47,$D3,$6C,$66,$37,$1C,$5F,$5C,$65,$F2,$7C,$37,$E3,$3E,$7D,$EF
	.dc.b	$D2,$4F,$E1,$4A,$7C,$48,$DB,$F9,$96,$BE,$F3,$A7,$3F,$5F,$96,$91
	.dc.b	$B8,$CB,$C6,$32,$F5,$8F,$CF,$FC,$FB,$C4,$F3,$78,$B2,$BF,$51,$39
	.dc.b	$6C,$FC,$48,$C4,$7F,$6D,$6B,$FD,$E1,$AF,$E7,$1D,$F2,$FF,$B0,$5B
	.dc.b	$CB,$9A,$F8,$DF,$41,$F1,$FE,$4A,$61,$F0,$5F,$95,$F2,$58,$8F,$8D
	.dc.b	$17,$1F,$F6,$66,$AA,$FA,$1E,$6F,$D1,$F8,$49,$87,$CB,$8C,$BF,$36
	.dc.b	$32,$FD,$18,$F8,$BE,$D7,$D9,$F9,$59,$3F,$12,$32,$FE,$A7,$64,$7C
	.dc.b	$52,$46,$5D,$BF,$A0,$8D,$BF,$77,$C6,$FE,$33,$EC,$BE,$57,$EF,$CF
	.dc.b	$F0,$48,$C7,$2F,$ED,$DB,$1F,$E1,$D6,$BF,$E1,$FF,$9E,$79,$7F,$9A
	.dc.b	$6C,$7E,$FD,$AE,$E3,$A4,$3D,$2A,$49,$ED,$19,$79,$7B,$BD,$48,$E0
	.dc.b	$ED,$E9,$F8,$48,$E0,$87,$F0,$29,$47,$F3,$26,$C7,$F3,$9F,$9E,$FB
	.dc.b	$8F,$F3,$BE,$17,$47,$F4,$0F,$8A,$FF,$67,$FB,$37,$80,$FF,$3C,$AD
	.dc.b	$DF,$1B,$FD,$EC,$62,$3F,$C8,$CD,$7F,$CA,$2D,$7F,$9A,$BB,$E7,$E0
	.dc.b	$D9,$C9,$E1,$F5,$BC,$1E,$D7,$F4,$CF,$9D,$7C,$F5,$33,$8F,$AE,$7C
	.dc.b	$64,$7F,$9A,$60,$8C,$47,$F2,$88,$C3,$C4,$FE,$BD,$5F,$E5,$DC,$7D
	.dc.b	$9F,$3C,$F8,$4F,$EC,$BF,$18,$F8,$FF,$ED,$58,$7F,$BB,$A3,$1E,$A7
	.dc.b	$F6,$2B,$E1,$3F,$DA,$5A,$FF,$9A,$DA,$FF,$9D,$9A,$FF,$5A,$6B,$FE
	.dc.b	$F9,$78,$E1,$E4,$C3,$B1,$E4,$F9,$E7,$8F,$FD,$CB,$E4,$5E,$0F,$F7
	.dc.b	$4F,$9D,$7A,$94,$D4,$4C,$FF,$AD,$57,$06,$C7,$F3,$A8,$C4,$7F,$A0
	.dc.b	$DB,$97,$E0,$5B,$C9,$FD,$C9,$B2,$18,$7F,$60,$7F,$C1,$9F,$3B,$FB
	.dc.b	$A7,$FB,$97,$C0,$7F,$8B,$2F,$FC,$26,$D8,$DF,$89,$FE,$0E,$AE,$11
	.dc.b	$FE,$9E,$6B,$FE,$02,$6B,$E1,$6A,$7C,$18,$FE,$65,$B7,$C6,$BE,$FF
	.dc.b	$CF,$3A,$3C,$1C,$FF,$2F,$D4,$99,$7A,$A3,$2F,$F8,$0F,$1F,$C6,$BF
	.dc.b	$EA,$96,$BE,$78,$FE,$CB,$E2,$BD,$EA,$49,$FF,$06,$46,$5B,$1F,$E2
	.dc.b	$E8,$C4,$7F,$80,$9B,$A7,$EC,$DF,$A3,$0F,$92,$79,$9F,$E3,$88,$7C
	.dc.b	$E3,$E7,$3F,$C5,$36,$F9,$EF,$F1,$5C,$E6,$BE,$D4,$E5,$FB,$91,$97
	.dc.b	$B5,$F5,$8C,$61,$E0,$C7,$78,$F6,$C7,$AB,$70,$C9,$F8,$7A,$E0,$F9
	.dc.b	$A3,$FC,$7B,$18,$BF,$06,$3E,$3A,$D1,$E6,$BE,$AF,$3C,$7C,$98,$72
	.dc.b	$6C,$6C,$F0,$9F,$D2,$67,$8B,$EC,$C7,$38,$F9,$3E,$3F,$C3,$5F,$E5
	.dc.b	$06,$D7,$E7,$DF,$83,$7F,$3D,$B7,$DB,$1F,$46,$5C,$2F,$D2,$77,$2E
	.dc.b	$3E,$FC,$BE,$34,$6A,$B8,$C7,$AE,$FC,$B8,$FD,$DA,$FF,$80,$37,$F2
	.dc.b	$9B,$19,$78,$1F,$EF,$D8,$FF,$00,$7C,$9F,$F9,$6D,$B1,$9F,$41,$39
	.dc.b	$6C,$7F,$8F,$BE,$17,$FC,$7E,$9A,$47,$47,$BB,$CE,$7C,$F7,$F4,$CD
	.dc.b	$7F,$15,$3F,$C1,$EB,$E9,$D6,$B9,$FE,$AC,$E4,$E5,$7C,$75,$FF,$40
	.dc.b	$C6,$70,$FC,$A5,$F1,$8F,$9A,$66,$7F,$C0,$95,$C6,$38,$E1,$E0,$8C
	.dc.b	$47,$FA,$32,$31,$87,$8A,$3D,$C7,$89,$C4,$6A,$3B,$9A,$C3,$C7,$5D
	.dc.b	$23,$C4,$7F,$25,$C3,$CD,$3F,$0A,$38,$61,$E6,$C3,$AC,$77,$3A,$E1
	.dc.b	$DE,$BD,$FE,$DF,$F7,$6D,$BF,$BA,$62,$8C,$61,$FE,$23,$F9,$BB,$8F
	.dc.b	$86,$6A,$AE,$15,$FE,$C1,$6C,$7F,$A4,$15,$70,$73,$7E,$0F,$81,$C1
	.dc.b	$DB,$AD,$ED,$3E,$43,$39,$75,$AB,$97,$F4,$2A,$D4,$E4,$F1,$7C,$27
	.dc.b	$73,$D3,$E4,$3E,$75,$F3,$B9,$AD,$93,$0F,$F5,$F5,$6E,$75,$1E,$BB
	.dc.b	$F3,$79,$7F,$EE,$AD,$7E,$4C,$62,$7E,$2F,$CE,$7C,$0C,$7F,$DC,$8D
	.dc.b	$8F,$F6,$54,$3E,$71,$F0,$3C,$38,$A4,$F0,$7E,$8B,$B8,$CD,$7F,$AF
	.dc.b	$2F,$A6,$3F,$E3,$0A,$E9,$7A,$8C,$61,$A9,$FF,$70,$D7,$FB,$AD,$F2
	.dc.b	$FC,$47,$C6,$F9,$DF,$F6,$7C,$60,$98,$F7,$9D,$70,$FB,$99,$3F,$E1
	.dc.b	$DD,$5E,$4E,$2F,$E5,$C6,$27,$FB,$F6,$7F,$E4,$3F,$5B,$BD,$7F,$B5
	.dc.b	$9B,$1E,$8F,$86,$FC,$7C,$62,$BF,$DF,$2D,$AF,$F0,$43,$63,$E6,$4F
	.dc.b	$0C,$A4,$78,$0F,$F0,$71,$F4,$7C,$7C,$DF,$D4,$DB,$FD,$CD,$3B,$8F
	.dc.b	$E1,$DF,$33,$FA,$0E,$C3,$FA,$4C,$9F,$3A,$F5,$B5,$33,$0E,$58,$7F
	.dc.b	$1F,$8E,$15,$B9,$CE,$1D,$1B,$3A,$F9,$0E,$56,$DB,$FC,$DB,$97,$82
	.dc.b	$7F,$67,$8F,$FB,$83,$0E,$FD,$8F,$8C,$CF,$C1,$7E,$E2,$31,$3A,$9C
	.dc.b	$1A,$70,$72,$F4,$F8,$1A,$70,$7C,$5F,$03,$E0,$7B,$FF,$C1,$9B,$1F
	.dc.b	$24,$7E,$4C,$CF,$5C,$3F,$E9,$29,$F8,$BD,$FF,$EE,$56,$FC,$46,$B1
	.dc.b	$FE,$1F,$39,$E1,$F8,$55,$FC,$8F,$CA,$E8,$F8,$DB,$1F,$BA,$D7,$FC
	.dc.b	$E9,$39,$CB,$F8,$CF,$CF,$3C,$1A,$7A,$7E,$69,$F1,$18,$BF,$B3,$5B
	.dc.b	$6D,$6E,$75,$8F,$90,$FF,$2D,$56,$DC,$07,$1C,$DF,$EA,$6E,$16,$3E
	.dc.b	$0C,$7A,$9F,$91,$E6,$FE,$D7,$F3,$BF,$8F,$F3,$BF,$0D,$97,$79,$CD
	.dc.b	$6E,$73,$8F,$E1,$95,$F1,$E9,$E4,$DA,$E1,$F3,$BD,$4E,$DE,$F5,$8C
	.dc.b	$93,$2F,$06,$1F,$31,$B5,$FD,$13,$E6,$FF,$DC,$75,$FA,$06,$CF,$F9
	.dc.b	$E3,$E7,$1E,$AE,$FE,$2E,$5E,$DF,$E8,$53,$A8,$C1,$2B,$E5,$DF,$4D
	.dc.b	$7F,$D9,$FD,$AB,$E1,$5A,$C3,$E4,$7B,$BF,$27,$D0,$B9,$7C,$76,$D7
	.dc.b	$B1,$B8,$FC,$97,$E8,$7C,$54,$AF,$9D,$F3,$B6,$FE,$1C,$DB,$FA,$13
	.dc.b	$6B,$FD,$39,$4D,$3F,$AB,$BE,$AC,$3F,$21,$DF,$F2,$59,$5F,$25,$CB
	.dc.b	$49,$3C,$5F,$D9,$72,$FF,$0B,$8C,$1F,$5B,$DC,$FC,$41,$F8,$9F,$88
	.dc.b	$B5,$F8,$78,$B3,$F7,$26,$E7,$F8,$9F,$9C,$FC,$87,$CE,$FE,$45,$B8
	.dc.b	$7E,$3D,$B7,$F8,$46,$D7,$FB,$FF,$4F,$DF,$4F,$D8,$F0,$BE,$EF,$C2
	.dc.b	$7E,$1D,$98,$7E,$1D,$CB,$49,$3C,$5F,$E6,$9C,$9F,$7E,$B8,$CF,$E5
	.dc.b	$2B,$F4,$DF,$1D,$FA,$76,$DF,$E9,$5B,$3B,$8F,$D4,$4F,$EB,$7E,$0F
	.dc.b	$F9,$E6,$4F,$8C,$C5,$7E,$CE,$30,$FE,$BE,$77,$3A,$BF,$DA,$56,$9F
	.dc.b	$E0,$3F,$39,$FC,$53,$CD,$FF,$59,$B7,$E1,$3F,$0A,$BE,$49,$97,$F9
	.dc.b	$76,$73,$E3,$E3,$0F,$F3,$44,$62,$75,$1F,$EE,$9B,$E8,$DB,$E1,$87
	.dc.b	$83,$0B,$87,$1A,$E6,$71,$F9,$3C,$BC,$E6,$AC,$13,$87,$BC,$66,$31
	.dc.b	$39,$75,$EF,$72,$F9,$C7,$9C,$E3,$F4,$E1,$D6,$B5,$39,$8F,$29,$9D
	.dc.b	$77,$7C,$AF,$8E,$5F,$E4,$89,$CD,$F1,$D3,$FD,$C3,$7B,$C3,$9B,$63
	.dc.b	$E9,$E3,$E0,$C7,$89,$2B,$C3,$8F,$57,$8E,$1D,$74,$FD,$EE,$3C,$DB
	.dc.b	$19,$9F,$A9,$87,$57,$8E,$1D,$71,$F0,$E5,$70,$E3,$36,$B8,$CD,$58
	.dc.b	$F7,$46,$31,$F0,$3F,$1C,$EE,$4C,$3E,$1C,$6B,$2F,$64,$63,$0E,$D7
	.dc.b	$D6,$F9,$4E,$5E,$58,$F9,$9F,$8B,$97,$B6,$74,$49,$FC,$EC,$65,$E1
	.dc.b	$19,$9F,$69,$D0,$F5,$D7,$84,$F1,$61,$E6,$AE,$24,$9D,$E3,$EB,$C3
	.dc.b	$B4,$63,$1F,$91,$7C,$63,$07,$73,$E4,$47,$E2,$A7,$8C,$60,$D4,$F1
	.dc.b	$71,$8F,$A5,$B8,$7B,$B2,$F9,$38,$FD,$2C,$BD,$38,$F4,$7F,$25,$7F
	.dc.b	$ED,$26,$BF,$2C,$C4,$FE,$85,$D7,$7F,$F1,$37,$FD,$B2,$7E,$09,$C3
	.dc.b	$AE,$0C,$B9,$6D,$27,$96,$5F,$A4,$6E,$5F,$9A,$6D,$FE,$33,$E6,$FF
	.dc.b	$C7,$0F,$EC,$23,$2F,$53,$F5,$6E,$23,$71,$8C,$38,$11,$FD,$DC,$6A
	.dc.b	$FC,$53,$96,$CE,$B7,$FF,$76,$61,$E5,$AD,$36,$74,$60,$FC,$4E,$1E
	.dc.b	$6C,$7E,$05,$7F,$0E,$C7,$C1,$7C,$71,$EE,$47,$FB,$BC,$6B,$0E,$75
	.dc.b	$D5,$F8,$14,$AD,$E3,$E6,$9D,$9F,$04,$93,$B7,$E6,$46,$DE,$73,$E3
	.dc.b	$23,$A3,$C1,$87,$A9,$E7,$1E,$07,$38,$FB,$CF,$A5,$5D,$6B,$85,$FB
	.dc.b	$6B,$CA,$F8,$D6,$3E,$F7,$C6,$6B,$0F,$1B,$AB,$E9,$87,$7C,$7E,$25
	.dc.b	$F1,$C7,$BE,$1C,$1B,$87,$AC,$F4,$9D,$48,$75,$3D,$75,$D3,$0F,$8D
	.dc.b	$7D,$70,$F0,$57,$E0,$09,$8F,$B5,$F4,$9C,$CF,$C7,$E7,$CB,$E7,$24
	.dc.b	$7E,$0E,$3A,$98,$7F,$2B,$18,$74,$43,$99,$F8,$BF,$8D,$D3,$82,$57
	.dc.b	$CD,$C1,$3F,$7F,$E6,$DF,$39,$FE,$44,$E5,$FE,$F9,$82,$5F,$25,$9E
	.dc.b	$2F,$F2,$4C,$62,$7F,$D9,$6F,$26,$DE,$DF,$E8,$71,$B7,$FD,$22,$DB
	.dc.b	$EE,$F1,$8F,$F4,$AF,$B5,$FE,$A3,$7F,$DA,$EB,$DF,$E9,$08,$C3,$AF
	.dc.b	$4F,$31,$C6,$39,$CE,$0C,$CF,$2F,$99,$FF,$6A,$4F,$FB,$91,$AF,$FA
	.dc.b	$99,$B8,$7C,$4B,$7B,$BC,$7D,$69,$7C,$5F,$0C,$62,$39,$45,$AE,$53
	.dc.b	$E6,$8F,$3F,$CE,$4A,$F7,$C6,$5F,$83,$19,$7E,$2C,$7B,$BE,$72,$47
	.dc.b	$C6,$7E,$0F,$C0,$62,$3E,$5C,$5C,$3D,$53,$F0,$7E,$72,$57,$C1,$7E
	.dc.b	$4F,$C3,$E2,$3E,$14,$5A,$F8,$53,$78,$7F,$D1,$7A,$AB,$17,$FE,$02
	.dc.b	$D5,$7F,$63,$E7,$1F,$03,$49,$5F,$78,$A7,$FB,$5A,$36,$FF,$83,$32
	.dc.b	$FC,$AC,$62,$3F,$0E,$75,$E9,$ED,$CD,$23,$86,$9F,$EA,$0A,$56,$E7
	.dc.b	$27,$F8,$62,$3F,$4A,$E0,$38,$7C,$91,$F9,$2F,$9C,$FF,$24,$C6,$DA
	.dc.b	$FF,$BB,$9B,$87,$A3,$C4,$E1,$E0,$CA,$E0,$FC,$58,$C4,$71,$8B,$3C
	.dc.b	$63,$C9,$83,$E4,$7E,$61,$23,$3F,$23,$5F,$95,$16,$BE,$4C,$FD,$3F
	.dc.b	$2B,$EA,$7A,$52,$FA,$C6,$5E,$F1,$97,$EA,$47,$C1,$F5,$64,$7C,$17
	.dc.b	$EF,$78,$98,$8F,$85,$17,$0F,$B3,$37,$7F,$F6,$9E,$CA,$C5,$FD,$3A
	.dc.b	$FE,$15,$EF,$3E,$4B,$F0,$18,$FF,$A4,$63,$1C,$7A,$9C,$E5,$FE,$52
	.dc.b	$F8,$37,$D0,$35,$FE,$13,$F3,$BF,$E1,$1D,$3E,$F3,$56,$7F,$C8,$A3
	.dc.b	$0F,$75,$E8,$79,$1B,$3F,$8A,$D1,$18,$8E,$79,$7F,$99,$27,$35,$FC
	.dc.b	$96,$1F,$93,$BB,$BC,$BE,$DC,$C3,$CA,$FA,$3D,$39,$1E,$86,$BD,$8F
	.dc.b	$F3,$8F,$CF,$39,$BF,$93,$B6,$38,$10,$E0,$1D,$38,$E6,$FE,$87,$B7
	.dc.b	$FA,$FF,$E1,$3F,$DA,$5E,$CF,$F3,$D6,$D7,$E9,$EB,$7A,$FF,$A0,$A1
	.dc.b	$18,$7F,$D0,$AD,$73,$1C,$F9,$9E,$B7,$F3,$EA,$D1,$27,$51,$9F,$6B
	.dc.b	$FA,$1F,$CE,$BE,$7A,$99,$C7,$C9,$35,$F6,$3C,$31,$C6,$9F,$D8,$32
	.dc.b	$4E,$32,$F5,$63,$FE,$5C,$AF,$EC,$BF,$3C,$F8,$5F,$EB,$DD,$4F,$8F
	.dc.b	$FE,$B9,$87,$E4,$5B,$A7,$FA,$92,$F8,$3F,$C3,$5A,$FD,$26,$E1,$DB
	.dc.b	$0E,$BF,$3D,$FC,$46,$31,$5C,$67,$FB,$1D,$70,$C3,$F4,$17,$B7,$E9
	.dc.b	$7C,$8F,$E8,$6F,$F8,$E7,$CF,$3E,$43,$FB,$5B,$63,$F9,$04,$ED,$FC
	.dc.b	$2B,$72,$FF,$02,$B6,$FA,$F7,$72,$F4,$7F,$BA,$CE,$6F,$FD,$86,$D4
	.dc.b	$DC,$3F,$B5,$C7,$F2,$FF,$9D,$FE,$04,$FF,$34,$F8,$07,$B7,$F8,$96
	.dc.b	$E5,$FE,$02,$AD,$D7,$E6,$9B,$9F,$F8,$69,$B3,$F9,$36,$CB,$85,$F3
	.dc.b	$DF,$90,$8F,$17,$8F,$D4,$8F,$F4,$69,$DF,$4F,$87,$9A,$4F,$A5,$F9
	.dc.b	$7E,$06,$23,$C9,$17,$3F,$F1,$6E,$DF,$BF,$9F,$E9,$DE,$37,$F9,$86
	.dc.b	$7F,$C6,$2D,$D7,$F5,$4D,$9F,$F7,$6B,$F2,$CC,$3F,$C9,$7E,$03,$FC
	.dc.b	$4B,$18,$3E,$87,$C8,$FF,$47,$BF,$F7,$AF,$CD,$BE,$73,$AF,$CF,$7E
	.dc.b	$DE,$31,$8F,$68,$C4,$FE,$EE,$F9,$4A,$B6,$A3,$31,$8C,$17,$CA,$BA
	.dc.b	$1E,$2B,$E9,$E7,$7F,$30,$8E,$91,$8C,$75,$1C,$0C,$97,$0F,$0E,$3E
	.dc.b	$0A,$E1,$3A,$3F,$A4,$61,$E1,$69,$D0,$F4,$61,$E1,$BE,$6D,$AF,$8B
	.dc.b	$3B,$C7,$D6,$F1,$C3,$C0,$8E,$78,$EF,$1F,$05,$70,$9F,$25,$70,$76
	.dc.b	$79,$E3,$58,$FC,$9A,$F4,$56,$CF,$A3,$5F,$CB,$4E,$25,$72,$FC,$23
	.dc.b	$1E,$2B,$E4,$3F,$DE,$73,$FD,$15,$B1,$99,$FF,$7D,$CE,$5F,$F7,$C8
	.dc.b	$F8,$2A,$D3,$2E,$08,$FD,$1F,$31,$F3,$CE,$0F,$A1,$3F,$1E,$BE,$95
	.dc.b	$6A,$73,$3D,$73,$FC,$6C,$E6,$3E,$E1,$C4,$ED,$7D,$5C,$9C,$76,$FE
	.dc.b	$A5,$3C,$6B,$F2,$F3,$C5,$EB,$5C,$5C,$C7,$13,$FB,$C4,$66,$BE,$84
	.dc.b	$64,$57,$58,$CE,$1E,$CA,$E1,$5F,$5E,$31,$87,$B9,$B1,$E8,$6E,$1E
	.dc.b	$26,$DF,$49,$F7,$D7,$E9,$36,$7D,$93,$E3,$BE,$14,$48,$C4,$FF,$19
	.dc.b	$FC,$A3,$FA,$C3,$FE,$60,$9E,$35,$C2,$3F,$CD,$D5,$C2,$FF,$C1,$9D
	.dc.b	$0C,$3F,$C6,$D5,$C6,$B5,$87,$D1,$AD,$74,$F3,$C3,$F9,$76,$2F,$C1
	.dc.b	$71,$5F,$CC,$31,$B3,$97,$F9,$77,$5F,$E1,$DF,$E6,$11,$9B,$F1,$CE
	.dc.b	$5F,$2C,$67,$1F,$14,$E5,$F2,$7D,$03,$D9,$F0,$AF,$8A,$B1,$86,$E3
	.dc.b	$FC,$FE,$74,$D9,$D3,$4D,$4F,$0C,$7E,$3D,$70,$7E,$4F,$C1,$FC,$86
	.dc.b	$E5,$F0,$A3,$18,$FC,$97,$E5,$E1,$FE,$59,$72,$D9,$E1,$AF,$16,$C7
	.dc.b	$F5,$59,$E3,$7F,$D7,$5E,$2D,$9F,$A3,$F3,$BF,$EE,$9A,$FE,$B8,$BE
	.dc.b	$EB,$3F,$03,$A6,$CE,$9B,$7F,$84,$AE,$13,$CA,$B8,$34,$E0,$7F,$69
	.dc.b	$D7,$FC,$6C,$DF,$6B,$F0,$71,$8B,$FC,$7B,$FE,$78,$9E,$1E,$96,$AB
	.dc.b	$6D,$8C,$9C,$CB,$5C,$A7,$F1,$FF,$82,$BE,$2E,$FC,$06,$1B,$7F,$DF
	.dc.b	$67,$4D,$9D,$34,$D4,$F0,$BF,$D2,$D7,$06,$D7,$E7,$BF,$29,$7D,$1D
	.dc.b	$3F,$49,$F5,$6F,$C1,$7E,$BF,$42,$B7,$9F,$F8,$3D,$B8,$FF,$9F,$7C
	.dc.b	$07,$E5,$6F,$73,$A5,$FF,$A4,$63,$0D,$8F,$D7,$FE,$56,$D7,$F6,$97
	.dc.b	$CA,$73,$18,$F6,$FC,$7F,$6C,$FA,$CC,$74,$FE,$E2,$31,$97,$F8,$8D
	.dc.b	$BA,$7F,$A4,$78,$3F,$82,$CE,$6F,$DF,$D7,$FB,$38,$FF,$5A,$7D,$77
	.dc.b	$F8,$11,$FF,$16,$36,$F9,$4E,$DF,$F1,$B4,$67,$B7,$FD,$25,$E5,$67
	.dc.b	$E0,$3F,$AE,$7C,$A7,$F0,$F7,$FA,$F6,$36,$53,$63,$0D,$7F,$E0,$0C
	.dc.b	$F2,$6C,$F5,$38,$B8,$38,$9F,$E4,$3F,$CB,$1F,$25,$FC,$71,$B1,$9F
	.dc.b	$0B,$13,$23,$83,$B8,$C6,$7F,$CB,$BE,$7D,$EE,$7F,$26,$5F,$6D,$63
	.dc.b	$0D,$C3,$FC,$A9,$3A,$6C,$E9,$FF,$0A,$6B,$FD,$96,$DF,$4C,$F9,$6F
	.dc.b	$E6,$D1,$87,$F0,$BF,$42,$FA,$7C,$D6,$F0,$E6,$7F,$AF,$5C,$5F,$FB
	.dc.b	$47,$07,$D7,$BE,$33,$8E,$5F,$D0,$E7,$31,$8C,$3F,$A9,$56,$DE,$B7
	.dc.b	$FE,$2B,$C3,$AD,$FF,$5C,$AD,$4F,$F5,$4F,$B1,$FF,$4D,$5F,$F9,$D5
	.dc.b	$7C,$C7,$F4,$E7,$33,$B8,$C7,$C3,$E5,$B5,$C5,$FE,$AB,$F9,$6B,$E7
	.dc.b	$2F,$A9,$86,$D7,$FA,$0E,$74,$D9,$D3,$4F,$C5,$F5,$FE,$4F,$BF,$F2
	.dc.b	$8D,$8C,$9F,$90,$8C,$B4,$CE,$5F,$EF,$DE,$AF,$F1,$D4,$EA,$31,$D5
	.dc.b	$FC,$2B,$EF,$9F,$4D,$FD,$B9,$F8,$FC,$3F,$DD,$5C,$E3,$FE,$9B,$72
	.dc.b	$FF,$96,$7C,$94,$6B,$0F,$F0,$0B,$A7,$FD,$50,$E6,$BA,$72,$7F,$7C
	.dc.b	$E8,$FE,$45,$1F,$E0,$37,$F8,$86,$1F,$EA,$97,$2F,$77,$37,$FE,$0E
	.dc.b	$72,$FF,$AD,$5C,$CF,$3F,$37,$B7,$8B,$FB,$DF,$A6,$FD,$F6,$97,$1F
	.dc.b	$DC,$B7,$1F,$8A,$DC,$3F,$84,$B6,$7F,$CF,$1B,$FF,$99,$63,$14,$8C
	.dc.b	$4F,$FB,$07,$F0,$1F,$85,$B1,$7C,$BF,$F8,$71,$C9,$89,$E1,$B6,$67
	.dc.b	$4D,$8F,$F6,$67,$E0,$CF,$A4,$FF,$12,$BF,$EE,$68,$CC,$63,$E3,$F2
	.dc.b	$D8,$C9,$CA,$3F,$C5,$BE,$67,$F1,$2D,$3F,$CF,$1F,$11,$F7,$63,$1F
	.dc.b	$02,$8C,$78,$DF,$E6,$98,$C5,$7F,$B9,$6B,$75,$C7,$C7,$D3,$71,$FF
	.dc.b	$3D,$3C,$9B,$5C,$67,$6D,$8C,$EE,$FC,$9F,$FC,$A7,$3E,$63,$45,$8D
	.dc.b	$7C,$A7,$27,$FB,$05,$DA,$E3,$97,$A2,$73,$87,$69,$CE,$3F,$EB,$5C
	.dc.b	$5F,$4E,$FA,$0F,$2C,$3E,$9F,$FB,$37,$C9,$FA,$A3,$73,$A9,$CC,$79
	.dc.b	$CC,$E3,$B8,$CE,$7F,$2A,$72,$6E,$33,$8F,$8D,$B1,$FE,$C9,$6B,$C2
	.dc.b	$79,$46,$09,$D4,$C7,$CD,$97,$AF,$1F,$14,$E7,$0F,$03,$70,$E6,$DC
	.dc.b	$7E,$25,$F1,$AD,$E1,$C8,$8F,$BE,$35,$7C,$A7,$2D,$9D,$65,$F1,$F0
	.dc.b	$ED,$1E,$D6,$BF,$3D,$B7,$D3,$0F,$89,$1A,$CB,$DF,$8F,$BB,$0E,$8F
	.dc.b	$82,$B6,$E7,$2C,$63,$DE,$2B,$F2,$E7,$6F,$B8,$99,$7D,$7C,$BC,$51
	.dc.b	$C4,$CC,$F0,$9D,$61,$E2,$9C,$9B,$8C,$E1,$E0,$C7,$D3,$3F,$12,$31
	.dc.b	$97,$B4,$F8,$37,$CE,$FE,$51,$30,$FA,$B1,$97,$C4,$4B,$1E,$D3,$C4
	.dc.b	$E5,$F9,$D9,$7C,$57,$CE,$60,$B9,$7C,$07,$17,$F6,$DF,$D4,$1B,$2E
	.dc.b	$5F,$A2,$6E,$3F,$8F,$73,$C4,$F9,$AF,$F7,$3C,$7E,$B6,$30,$FF,$BA
	.dc.b	$AB,$F9,$04,$E9,$B1,$99,$DD,$FE,$89,$B9,$7E,$A5,$B3,$F0,$BE,$9F
	.dc.b	$FD,$78,$FE,$CE,$32,$F8,$70,$C1,$FB,$D9,$DC,$7E,$D6,$77,$87,$8B
	.dc.b	$0F,$05,$EF,$E1,$7F,$85,$46,$2F,$84,$F4,$7F,$82,$61,$D2,$74,$4C
	.dc.b	$7D,$15,$CE,$B8,$4E,$5F,$E0,$58,$FA,$2F,$95,$F0,$9E,$E4,$AF,$E1
	.dc.b	$D3,$A8,$E3,$7D,$2B,$85,$6B,$0F,$4C,$60,$95,$C0,$F4,$C6,$9C,$4F
	.dc.b	$95,$F9,$46,$8E,$C6,$A7,$9A,$BF,$23,$C2,$79,$63,$F3,$6B,$F9,$A5
	.dc.b	$6A,$31,$5C,$C8,$73,$BF,$23,$CE,$BB,$CF,$08,$D1,$E4,$34,$72,$AE
	.dc.b	$D8,$FD,$E2,$63,$F4,$E7,$E6,$5F,$49,$FA,$58,$78,$30,$EC,$48,$CD
	.dc.b	$71,$C7,$DB,$18,$8F,$8C,$74,$3C,$55,$F3,$78,$1F,$8D,$3E,$3F,$1C
	.dc.b	$BF,$E5,$AB,$FE,$CE,$D7,$FB,$24,$ED,$FE,$D6,$D7,$FB,$44,$6F,$1F
	.dc.b	$EC,$8D,$CF,$FC,$7D,$5B,$9F,$ED,$0D,$8C,$DF,$96,$31,$6C,$3C,$FF
	.dc.b	$85,$66,$A4,$F8,$9E,$11,$ED,$FC,$19,$F1,$1F,$DC,$7C,$17,$37,$F9
	.dc.b	$3E,$7F,$D5,$38,$7F,$A0,$31,$46,$21,$5C,$1F,$14,$74,$78,$35,$DD
	.dc.b	$75,$E8,$F0,$36,$37,$86,$1F,$67,$7F,$F8,$C5,$FF,$1F,$76,$67,$2F
	.dc.b	$F1,$F3,$67,$53,$D8,$E3,$C5,$99,$91,$FE,$47,$8E,$CB,$E1,$FD,$9A
	.dc.b	$7F,$AC,$D7,$6E,$9B,$3F,$EA,$23,$86,$99,$35,$32,$35,$3D,$34,$FF
	.dc.b	$0F,$36,$3F,$A2,$BD,$1F,$41,$E3,$9E,$8F,$D1,$FC,$17,$58,$7C,$FD
	.dc.b	$10,$FB,$AF,$88,$DA,$E1,$78,$33,$32,$32,$7F,$A3,$FD,$AF,$99,$E4
	.dc.b	$BC,$8A,$62,$24,$FF,$A5,$63,$2F,$13,$E2,$7D,$3F,$F8,$E5,$BF,$51
	.dc.b	$F3,$A0,$DF,$0A,$4E,$88,$7C,$29,$E4,$D8,$E5,$32,$3B,$C7,$E0,$FD
	.dc.b	$07,$C8,$DF,$9A,$95,$FE,$C3,$8D,$18,$7F,$0D,$BA,$95,$A9,$FF,$6C
	.dc.b	$78,$BF,$8F,$F5,$9F,$29,$4C,$44,$9F,$F6,$F4,$65,$E2,$7E,$1F,$EA
	.dc.b	$1E,$A7,$C3,$F7,$24,$E6,$3F,$27,$3A,$8F,$97,$3A,$7F,$0B,$19,$6C
	.dc.b	$F0,$23,$F4,$BC,$7F,$D3,$7B,$DF,$E0,$1E,$8F,$C4,$3E,$A9,$FD,$E3
	.dc.b	$5F,$86,$EF,$AB,$FD,$A0,$D8,$FE,$1C,$EE,$79,$E5,$DB,$3F,$DA,$78
	.dc.b	$FF,$C7,$E7,$F2,$CD,$C3,$FD,$B4,$DC,$3F,$8C,$2F,$0B,$8F,$87,$27
	.dc.b	$F8,$84,$EF,$88,$EE,$3F,$65,$F8,$FB,$C1,$FD,$DA,$FB,$3A,$62,$24
	.dc.b	$F0,$8C,$BC,$A3,$2F,$EC,$A7,$93,$B3,$47,$F2,$3E,$D7,$AB,$FC,$4D
	.dc.b	$B1,$98,$FE,$35,$F0,$AF,$C0,$9F,$1D,$FC,$67,$E2,$1D,$1A,$9C,$E1
	.dc.b	$FE,$48,$E2,$FE,$59,$87,$1C,$38,$5F,$F2,$5F,$8A,$7D,$3B,$E2,$F3
	.dc.b	$8F,$F3,$7C,$FA,$BC,$F9,$7F,$CA,$77,$C2,$B9,$79,$4F,$B3,$FE,$59
	.dc.b	$F5,$2F,$83,$A6,$A3,$F9,$87,$C8,$7F,$AA,$78,$3F,$A0,$63,$FE,$E5
	.dc.b	$9D,$67,$FD,$1E,$B5,$9F,$F9,$8B,$A1,$18,$9F,$EB,$0D,$8C,$FA,$7F
	.dc.b	$23,$F0,$37,$CF,$7F,$3F,$FC,$09,$F1,$D7,$E2,$A6,$5F,$D5,$67,$2D
	.dc.b	$9C,$BF,$BE,$84,$F3,$99,$3C,$63,$7B,$7F,$2D,$AE,$73,$FC,$1B,$EA
	.dc.b	$9E,$DF,$F5,$17,$EF,$C6,$7A,$F1,$32,$3F,$B4,$4E,$AF,$F9,$C3,$77
	.dc.b	$F8,$8F,$49,$DD,$75,$B9,$18,$3F,$B9,$7C,$D7,$EC,$F7,$3F,$B3,$C3
	.dc.b	$E7,$36,$7F,$C0,$8E,$C9,$E2,$FF,$69,$75,$5C,$BC,$5C,$19,$99,$5E
	.dc.b	$08,$FE,$E1,$F8,$E7,$C4,$D3,$FA,$93,$F3,$27,$FC,$0C,$D7,$F9,$6B
	.dc.b	$BC,$7F,$04,$D8,$FF,$0C,$3B,$9E,$7B,$76,$DE,$57,$48,$F0,$98,$7F
	.dc.b	$9C,$FC,$E3,$EA,$7F,$C0,$8F,$E8,$E3,$FC,$30,$D7,$71,$A3,$7D,$72
	.dc.b	$37,$19,$78,$C6,$5E,$27,$F8,$8F,$F2,$EF,$DB,$BF,$D2,$97,$CB,$FF
	.dc.b	$85,$5B,$1B,$F0,$70,$66,$64,$64,$FD,$B1,$FE,$2C,$FC,$4D,$F3,$BF
	.dc.b	$E1,$FF,$CB,$DF,$27,$4F,$F1,$36,$EA,$DD,$F2,$8C,$5E,$A7,$33,$FE
	.dc.b	$25,$9C,$BF,$E3,$DF,$A3,$FF,$1D,$3F,$E1,$2F,$A5,$7D,$3F,$F8,$FA
	.dc.b	$1F,$4D,$FD,$B9,$FF,$6E,$FD,$3F,$F9,$76,$31,$08,$C7,$B7,$FE,$E2
	.dc.b	$85,$F4,$C7,$FD,$CB,$7C,$23,$13,$DA,$B8,$5F,$28,$EF,$E9,$F3,$BE
	.dc.b	$18,$F9,$5B,$8F,$4C,$38,$D6,$DB,$79,$3F,$DE,$B8,$F8,$2F,$8C,$78
	.dc.b	$4E,$E6,$CF,$65,$73,$78,$E1,$E1,$BF,$75,$F3,$7B,$CF,$33,$89,$CE
	.dc.b	$34,$F2,$C3,$B6,$3E,$F8,$D9,$C6,$76,$FC,$AB,$F0,$3E,$63,$B4,$F1
	.dc.b	$9D,$9F,$32,$3E,$91,$C6,$BF,$E0,$0C,$66,$3D,$35,$F7,$DD,$C7,$F3
	.dc.b	$2F,$97,$DF,$84,$69,$27,$A9,$FE,$F2,$28,$1F,$17,$87,$0D,$72,$65
	.dc.b	$FA,$1E,$47,$F8,$C3,$B3,$E9,$4F,$03,$E6,$46,$5F,$9C,$67,$6F,$C9
	.dc.b	$C6,$5F,$CA,$C6,$75,$FC,$A4,$E5,$F4,$DF,$83,$BB,$FD,$39,$7C,$2F
	.dc.b	$CA,$F4,$8C,$D7,$09,$F1,$C6,$1F,$9F,$5C,$F0,$F4,$C7,$2A,$F0,$D6
	.dc.b	$E3,$E4,$5F,$5A,$DD,$F1,$9F,$69,$C4,$FE,$F7,$5C,$5F,$61,$89,$E2
	.dc.b	$F9,$1F,$D0,$6A,$FA,$77,$CC,$FF,$AE,$62,$46,$27,$F8,$27,$C9,$E5
	.dc.b	$FF,$0A,$46,$6F,$F8,$53,$6B,$8D,$EE,$7F,$DA,$9C,$6C,$31,$5E,$1A
	.dc.b	$D7,$85,$E2,$E7,$BF,$4F,$FE,$D2,$CF,$FD,$A1,$5F,$C8,$1F,$F6,$94
	.dc.b	$67,$2F,$F3,$CC,$E5,$FF,$3D,$C6,$70,$FF,$53,$4E,$5F,$F5,$44,$66
	.dc.b	$BF,$C2,$9D,$EF,$A7,$F1,$BF,$E4,$E8,$FE,$93,$18,$F6,$B2,$D8,$C9
	.dc.b	$FD,$3A,$32,$D3,$38,$FF,$56,$8F,$7F,$C4,$FF,$A0,$A7,$FD,$D8,$FC
	.dc.b	$97,$FC,$4C,$E6,$7F,$D2,$AE,$5F,$F7,$F3,$9C,$7F,$D8,$4E,$74,$FF
	.dc.b	$6B,$FB,$FD,$B4,$EE,$DC,$3B,$B7,$0F,$F5,$0F,$91,$FE,$E1,$9F,$F0
	.dc.b	$6F,$C7,$7D,$48,$C4,$EA,$31,$F1,$99,$6D,$70,$3F,$C2,$B1,$96,$99
	.dc.b	$9F,$F5,$7E,$AF,$A7,$FC,$0B,$FD,$7A,$7F,$B0,$E7,$F4,$E7,$3E,$0E
	.dc.b	$9B,$3A,$AE,$4D,$9D,$34,$D4,$F0,$BF,$F5,$FF,$D3,$FF,$B7,$67,$BA
	.dc.b	$F9,$5F,$C9,$B9,$9D,$C6,$3E,$33,$2D,$AE,$2D,$9D,$C7,$49,$DB,$FE
	.dc.b	$63,$F1,$FF,$3F,$EA,$FF,$A0,$A3,$FB,$FF,$BA,$FA,$7F,$B5,$39,$C3
	.dc.b	$FD,$AA,$DE,$5F,$F3,$7F,$D3,$FF,$94,$B0,$E8,$7F,$82,$3C,$24,$66
	.dc.b	$D5,$C3,$D4,$FD,$A5,$69,$F1,$E7,$8B,$E1,$5F,$E7,$CF,$3B,$F2,$17
	.dc.b	$FE,$8D,$E8,$FD,$BF,$BB,$E3,$CF,$FD,$17,$5F,$91,$8F,$F1,$0C,$7B
	.dc.b	$3A,$7F,$5D,$A3,$99,$5A,$8F,$2D,$7F,$45,$E0,$FD,$23,$F8,$2F,$A6
	.dc.b	$79,$9F,$C1,$9B,$1D,$4C,$47,$F0,$FF,$9E,$FE,$A9,$17,$C9,$7C,$8D
	.dc.b	$8C,$35,$FF,$17,$CF,$F5,$B8,$C6,$3E,$41,$19,$9D,$D7,$B2,$31,$3B
	.dc.b	$7C,$33,$AA,$FE,$3B,$F4,$CF,$95,$A6,$3C,$1B,$5F,$A0,$9D,$36,$74
	.dc.b	$D3,$FD,$EF,$F4,$3F,$CA,$3E,$33,$FD,$47,$18,$3F,$DF,$7E,$CE,$67
	.dc.b	$4D,$8C,$9F,$E4,$EC,$DF,$48,$FA,$5F,$E7,$11,$87,$F0,$FF,$4D,$FE
	.dc.b	$D6,$AE,$15,$B6,$CF,$08,$C5,$7C,$9E,$85,$F1,$8F,$98,$75,$BE,$27
	.dc.b	$F9,$5F,$E9,$9F,$47,$CF,$1F,$E8,$F1,$89,$FE,$0B,$F2,$9F,$EB,$A6
	.dc.b	$FD,$37,$F4,$A8,$C6,$1F,$67,$A5,$09,$D7,$C0,$6D,$B7,$C8,$FE,$23
	.dc.b	$F2,$FF,$EE,$16,$FA,$3F,$D5,$A3,$18,$F5,$7F,$A6,$4F,$FB,$75,$CF
	.dc.b	$80,$8C,$57,$1F,$8A,$E4,$FE,$57,$BF,$F2,$DE,$0D,$AF,$C6,$B7,$1F
	.dc.b	$C6,$76,$7F,$41,$EC,$FF,$18,$5F,$0B,$FC,$9B,$6B,$6D,$CB,$C5,$1F
	.dc.b	$B1,$F9,$B7,$D3,$7F,$70,$7E,$64,$7C,$77,$37,$FC,$C9,$CD,$72,$E1
	.dc.b	$46,$AB,$87,$27,$3C,$3A,$3F,$DF,$3A,$7F,$95,$61,$C5,$FE,$5D,$8F
	.dc.b	$F3,$E7,$2F,$F9,$17,$14,$6B,$0F,$B4,$E9,$FE,$14,$E6,$BA,$79,$3D
	.dc.b	$FC,$AF,$E0,$1F,$0F,$FC,$07,$3B,$97,$EE,$1B,$8F,$CC,$6E,$1F,$1B
	.dc.b	$6F,$F2,$E4,$62,$BF,$CB,$11,$87,$F8,$0D,$F5,$7F,$AB,$FC,$F7,$C6
	.dc.b	$7D,$4B,$E9,$BF,$C3,$AE,$71,$FE,$44,$E6,$78,$ED,$AA,$DB,$63,$FB
	.dc.b	$1F,$D3,$3E,$47,$FC,$4E,$BE,$5E,$C6,$1B,$3C,$1B,$8F,$E8,$EB,$84
	.dc.b	$FF,$6E,$F3,$BF,$87,$CF,$F9,$D3,$E4,$3E,$94,$63,$E0,$11,$8F,$23
	.dc.b	$F1,$71,$8A,$FE,$79,$5B,$AE,$3E,$3E,$9B,$87,$F9,$F9,$E4,$DA,$E3
	.dc.b	$3B,$6C,$67,$99,$F4,$DF,$E4,$98,$C6,$1D,$EB,$FA,$64,$E5,$F5,$1C
	.dc.b	$8B,$5C,$BD,$EE,$CA,$B9,$39,$BF,$25,$3E,$99,$F3,$FE,$58,$7D,$2F
	.dc.b	$F8,$1F,$E4,$77,$5A,$BE,$33,$9A,$F1,$9C,$EF,$94,$67,$6F,$95,$39
	.dc.b	$AE,$3D,$4C,$69,$27,$9E,$3D,$DE,$B8,$73,$DF,$FD,$DC,$D8,$FF,$58
	.dc.b	$35,$EC,$70,$BE,$CD,$BE,$D8,$F9,$30,$EC,$F9,$AF,$91,$2F,$8C,$6F
	.dc.b	$2F,$6C,$63,$1F,$1B,$71,$F1,$B5,$F1,$D7,$58,$C5,$F5,$C7,$E6,$E1
	.dc.b	$D1,$F2,$E1,$D1,$CE,$58,$C7,$AC,$57,$AC,$ED,$F7,$13,$1F,$45,$6F
	.dc.b	$0F,$04,$FC,$EC,$3C,$0E,$EF,$95,$F0,$8C,$1E,$02,$9E,$03,$0F,$B7
	.dc.b	$0F,$26,$5F,$32,$B9,$92,$B6,$FF,$B9,$DE,$47,$E4,$32,$F8,$A4,$AF
	.dc.b	$83,$1E,$0C,$7D,$38,$79,$4F,$04,$66,$BE,$86,$1E,$18,$EE,$43,$B5
	.dc.b	$F4,$AE,$27,$28,$F2,$1A,$3F,$DF,$F7,$F9,$F7,$2D,$7F,$27,$C6,$F9
	.dc.b	$EE,$31,$F9,$48,$C3,$C9,$AF,$26,$BC,$A3,$F5,$B1,$C2,$BF,$28,$DC
	.dc.b	$BE,$47,$D3,$3E,$0B,$E5,$56,$AF,$89,$30,$FD,$AD,$6D,$C4,$ED,$FD
	.dc.b	$6E,$1D,$50,$9E,$0E,$AB,$8E,$1E,$78,$EB,$87,$8A,$31,$87,$52,$3E
	.dc.b	$88,$D4,$FC,$5A,$E6,$D3,$99,$83,$F8,$9E,$1E,$56,$C7,$F1,$AC,$3C
	.dc.b	$B8,$76,$9E,$4D,$7D,$57,$E0,$78,$9E,$B2,$47,$F1,$69,$E7,$87,$A8
	.dc.b	$93,$D1,$F5,$3E,$5C,$7F,$AC,$61,$E7,$AE,$33,$B2,$1E,$0B,$ED,$87
	.dc.b	$8E,$F8,$CF,$8B,$0E,$EF,$A0,$ED,$8F,$C1,$9D,$DF,$29,$CE,$3E,$1C
	.dc.b	$3A,$5E,$C9,$1F,$1E,$32,$F0,$9F,$9E,$78,$4F,$3D,$F6,$C3,$DF,$1A
	.dc.b	$C3,$E6,$4E,$63,$F0,$F8,$FC,$07,$D0,$73,$3E,$05,$72,$AF,$ED,$9B
	.dc.b	$BE,$9B,$FD,$43,$5F,$D8,$3C,$8F,$C2,$CF,$F5,$96,$CE,$A7,$FA,$FB
	.dc.b	$6B,$9F,$17,$F9,$86,$79,$1F,$DB,$67,$58,$7F,$8C,$5B,$1F,$EB,$76
	.dc.b	$E1,$E3,$F5,$3B,$FD,$07,$81,$95,$CD,$CB,$FD,$E7,$E9,$9F,$09,$FC
	.dc.b	$6F,$C3,$FF,$25,$47,$FA,$8E,$7F,$DF,$ED,$7F,$CF,$0F,$07,$33,$C7
	.dc.b	$9F,$FC,$8D,$3B,$AE,$7D,$59,$BF,$F6,$74,$F1,$7F,$D0,$5E,$07,$F8
	.dc.b	$A5,$FF,$1B,$F7,$6F,$44,$FF,$8A,$E7,$FC,$77,$CD,$98,$E1,$1C,$AB
	.dc.b	$06,$67,$C1,$0F,$84,$F2,$67,$FC,$E5,$77,$69,$B3,$FE,$B1,$76,$FF
	.dc.b	$0C,$6B,$CE,$38,$1C,$73,$91,$C6,$32,$F3,$8F,$67,$D3,$76,$8F,$F5
	.dc.b	$12,$FA,$0C,$9A,$A9,$3F,$C7,$36,$FF,$3F,$56,$E3,$18,$FF,$9D,$63
	.dc.b	$11,$D2,$BE,$57,$93,$FE,$EE,$8F,$F4,$47,$91,$83,$3E,$D4,$8C,$90
	.dc.b	$FF,$48,$3F,$2B,$E9,$BF,$C5,$F1,$FE,$D5,$5F,$4D,$93,$5E,$1C,$9D
	.dc.b	$46,$5E,$11,$97,$FD,$9F,$3C,$9D,$9A,$3F,$D7,$FF,$21,$F8,$3F,$4D
	.dc.b	$F4,$D4,$C4,$49,$FF,$62,$C6,$5E,$27,$FB,$53,$C4,$FC,$67,$AC,$F9
	.dc.b	$0A,$62,$24,$FF,$B6,$A3,$2F,$13,$FD,$E9,$F4,$CF,$4F,$E2,$7D,$1C
	.dc.b	$9C,$C7,$E4,$27,$51,$82,$1F,$03,$C8,$D4,$C8,$CF,$37,$FB,$11,$6D
	.dc.b	$FD,$C2,$3F,$D7,$DF,$01,$FE,$44,$9F,$F7,$53,$EB,$7F,$06,$3B,$91
	.dc.b	$98,$DE,$9F,$AD,$8E,$55,$D3,$4C,$19,$9F,$D9,$79,$5F,$C6,$67,$FD
	.dc.b	$9C,$DE,$9F,$F6,$23,$70,$F0,$2F,$0F,$8F,$87,$27,$CB,$9F,$F8,$06
	.dc.b	$3A,$CF,$ED,$BE,$97,$F7,$D4,$C7,$E1,$43,$E9,$B4,$6F,$C2,$95,$F0
	.dc.b	$04,$ED,$C3,$5E,$99,$FF,$8E,$63,$BD,$F6,$AE,$B5,$C2,$FA,$BC,$0F
	.dc.b	$99,$DE,$F5,$3F,$8B,$36,$3D,$AD,$78,$7C,$2B,$E9,$9F,$15,$FC,$77
	.dc.b	$E2,$1D,$3F,$C9,$6B,$53,$6B,$77,$F2,$2F,$6F,$DE,$F8,$47,$D3,$3E
	.dc.b	$2B,$F9,$74,$61,$B1,$99,$E5,$E7,$EB,$CB,$7D,$2D,$24,$FD,$6F,$8D
	.dc.b	$FF,$50,$F1,$B9,$7F,$C5,$BD,$09,$CC,$7F,$8C,$34,$FF,$71,$61,$CB
	.dc.b	$0F,$91,$7B,$CB,$F9,$E6,$1C,$27,$F2,$CD,$9D,$47,$E6,$B8,$1F,$4A
	.dc.b	$F9,$AF,$E8,$9F,$4A,$F8,$DB,$F1,$33,$0F,$C6,$CE,$5B,$39,$7C,$70
	.dc.b	$9E,$73,$27,$8C,$6F,$7F,$81,$1F,$AB,$AE,$8F,$B3,$E9,$7F,$DA,$CF
	.dc.b	$E7,$1F,$C0,$4F,$0C,$D1,$8E,$0F,$92,$F0,$AE,$5E,$C7,$F5,$F8,$D4
	.dc.b	$7E,$E5,$B5,$CA,$74,$6F,$AA,$57,$86,$3F,$B4,$7C,$BF,$EF,$30,$FA
	.dc.b	$EF,$F0,$1C,$7C,$8D,$8C,$F8,$5C,$0E,$3E,$1C,$C3,$F6,$F5,$CD,$FE
	.dc.b	$18,$DC,$8F,$38,$CC,$F3,$7F,$7F,$F4,$BF,$23,$1F,$88,$FD,$18,$CE
	.dc.b	$DF,$DF,$E3,$59,$7E,$39,$B3,$FC,$6D,$E0,$FE,$31,$B5,$D3,$7E,$47
	.dc.b	$3B,$95,$E2,$8F,$F0,$67,$CC,$BE,$96,$F5,$CA,$FE,$2B,$39,$6C,$E5
	.dc.b	$FC,$DC,$7E,$AD,$B1,$D0,$D6,$1F,$81,$9D,$36,$3F,$C3,$9F,$4B,$FB
	.dc.b	$CA,$FD,$2A,$FA,$5C,$9A,$F0,$24,$FF,$28,$9D,$36,$72,$FF,$15,$B7
	.dc.b	$D2,$3E,$73,$FC,$61,$18,$6C,$66,$7F,$99,$FC,$FF,$F4,$FA,$FC,$F9
	.dc.b	$B8,$FE,$96,$D7,$B4,$F1,$7F,$7B,$3C,$78,$3F,$DC,$9F,$44,$FA,$3F
	.dc.b	$E7,$54,$FA,$37,$AD,$FE,$E9,$FA,$4F,$F3,$24,$62,$7F,$D7,$31,$89
	.dc.b	$E7,$5C,$AB,$54,$BE,$D9,$FF,$B0,$27,$38,$71,$AE,$7E,$7F,$FB,$DE
	.dc.b	$78,$DF,$09,$F2,$63,$DB,$0E,$4F,$96,$F8,$36,$72,$7F,$B3,$AB,$95
	.dc.b	$F4,$C7,$CD,$5B,$AE,$A7,$88,$E2,$7A,$63,$6F,$36,$AC,$7D,$4D,$C3
	.dc.b	$C6,$EF,$1E,$78,$F8,$B0,$F0,$63,$F0,$AF,$77,$DE,$F7,$8F,$6C,$7A
	.dc.b	$9D,$B1,$E5,$5D,$4F,$F7,$59,$F1,$4A,$64,$AF,$CB,$6F,$7B,$E5,$37
	.dc.b	$1F,$1E,$30,$E2,$7F,$9E,$4E,$A3,$EE,$4E,$9C,$7C,$3E,$5F,$F7,$EB
	.dc.b	$53,$48,$FD,$01,$E3,$FF,$8F,$F6,$FA,$31,$C8,$D4,$65,$E0,$67,$5F
	.dc.b	$C7,$C6,$5F,$C8,$C6,$76,$F8,$B3,$97,$DF,$7F,$7A,$31,$3F,$22,$31
	.dc.b	$D7,$FD,$32,$F9,$CF,$A8,$F1,$C6,$6F,$9C,$63,$0F,$4B,$BA,$F2,$D6
	.dc.b	$E3,$B1,$B3,$D6,$6E,$BC,$F1,$B8,$FB,$B7,$E4,$BE,$AD,$8E,$2D,$BF
	.dc.b	$DB,$7E,$0C,$FC,$01,$F4,$7F,$C2,$30,$60,$AD,$69,$98,$C5,$F0,$AF
	.dc.b	$F2,$BD,$69,$B1,$FD,$FD,$B5,$C6,$7C,$17,$6B,$57,$FE,$35,$9F,$F0
	.dc.b	$B7,$A5,$BC,$0F,$87,$F0,$57,$E0,$AF,$AE,$D4,$CA,$F1,$9C,$0F,$F3
	.dc.b	$5D,$70,$9D,$61,$DF,$F0,$17,$E0,$1E,$56,$DF,$0F,$C4,$BF,$9E,$3E
	.dc.b	$E8,$CD,$71,$AE,$07,$96,$B7,$19,$AE,$13,$A7,$A7,$6D,$9F,$81,$E3
	.dc.b	$FF,$86,$E7,$E5,$B6,$3F,$CB,$19,$3E,$8B,$1D,$7F,$E1,$CD,$FF,$D0
	.dc.b	$DE,$37,$CD,$C3,$38,$FD,$A9,$CB,$F0,$AF,$38,$FD,$DA,$B5,$A8,$F1
	.dc.b	$DF,$CF,$AC,$D6,$BF,$27,$FF,$25,$61,$F8,$06,$DF,$9B,$F1,$47,$D9
	.dc.b	$FD,$F9,$CE,$7E,$76,$DF,$95,$B6,$D1,$87,$2A,$EC,$73,$21,$F7,$1D
	.dc.b	$4F,$E2,$6B,$9E,$8F,$C5,$1F,$84,$7E,$45,$B8,$7E,$13,$F0,$BF,$ED
	.dc.b	$CF,$E6,$A3,$38,$7E,$4D,$B9,$FC,$F6,$CF,$E9,$B3,$5F,$2B,$E1,$87
	.dc.b	$72,$57,$7F,$1A,$CF,$E7,$FF,$15,$7D,$FF,$E9,$5B,$5F,$98,$F2,$3F
	.dc.b	$DC,$EF,$EB,$A3,$35,$C6,$B8,$1F,$A5,$AD,$C6,$6B,$84,$E9,$E9,$F1
	.dc.b	$76,$7F,$61,$F8,$AF,$B1,$FE,$00,$DF,$65,$F6,$3F,$3B,$83,$F2,$13
	.dc.b	$BE,$17,$E1,$FF,$EB,$DC,$BF,$81,$B7,$1E,$BB,$FF,$01,$8C,$61,$F9
	.dc.b	$9F,$15,$EE,$FF,$05,$6E,$1C,$BE,$17,$13,$FC,$47,$5F,$1F,$83,$FC
	.dc.b	$2F,$5F,$E8,$18,$FF,$9F,$A7,$38,$7F,$14,$9C,$BF,$A3,$5F,$59,$C4
	.dc.b	$E5,$95,$9F,$E3,$7F,$8D,$BE,$D7,$F8,$C3,$F7,$71,$FD,$83,$99,$DF
	.dc.b	$37,$F2,$5F,$AA,$FE,$53,$EC,$DE,$2F,$F4,$F5,$7F,$BF,$27,$F9,$6F
	.dc.b	$83,$9A,$93,$9A,$D9,$FE,$D7,$AD,$C6,$67,$73,$93,$78,$FF,$A7,$AB
	.dc.b	$71,$99,$FE,$61,$F8,$43,$E8,$2F,$CF,$61,$B8,$FF,$3A,$9D,$36,$74
	.dc.b	$D3,$F9,$FF,$E2,$DF,$C7,$B2,$FE,$24,$D8,$FD,$27,$DC,$FF,$B9,$9D
	.dc.b	$F8,$FF,$D0,$67,$59,$FF,$4D,$6F,$82,$BE,$13,$FD,$0E,$76,$BE,$D1
	.dc.b	$F3,$FF,$D4,$E3,$0F,$F5,$5F,$83,$7E,$5D,$FD,$17,$3F,$EB,$7A,$7F
	.dc.b	$B0,$AF,$71,$83,$FC,$4B,$87,$47,$10,$B7,$E1,$6F,$99,$F0,$5E,$E9
	.dc.b	$87,$FB,$3A,$F7,$5A,$CB,$C9,$87,$1C,$7A,$E1,$E2,$38,$BF,$DC,$2F
	.dc.b	$8B,$E2,$F3,$1F,$2E,$FB,$2F,$EE,$51,$8E,$5B,$39,$AE,$0F,$E1,$1F
	.dc.b	$F7,$FB,$9C,$3F,$DF,$6E,$55,$76,$BE,$77,$FD,$DF,$F1,$A7,$C8,$53
	.dc.b	$FB,$97,$E2,$AF,$97,$C6,$7F,$E0,$2A,$DB,$6B,$73,$AC,$BF,$39,$F0
	.dc.b	$0F,$C3,$BF,$50,$DB,$FD,$32,$FB,$0C,$CC,$9C,$BC,$23,$15,$D2,$2F
	.dc.b	$1F,$E8,$EF,$95,$F0,$3F,$C2,$FE,$3F,$D9,$9F,$ED,$4D,$8F,$E7,$9A
	.dc.b	$3E,$DF,$FC,$31,$39,$C7,$FC,$48,$DF,$33,$F8,$CF,$2F,$C4,$AD,$BF
	.dc.b	$5B,$E9,$DF,$6D,$7E,$D7,$0D,$CF,$EF,$4E,$9B,$3A,$69,$F1,$3D,$9F
	.dc.b	$8E,$7C,$8F,$93,$B1,$92,$69,$F3,$A3,$2F,$D0,$9B,$B7,$C5,$9B,$5F
	.dc.b	$B8,$FA,$7F,$DB,$7E,$1D,$FE,$2F,$F0,$51,$89,$FF,$15,$C6,$1F,$E0
	.dc.b	$73,$B9,$D6,$3F,$E1,$8E,$1F,$E5,$F5,$F4,$3C,$4F,$C4,$F9,$BF,$31
	.dc.b	$B1,$93,$FC,$6F,$3C,$2B,$FC,$89,$C6,$AF,$C8,$FB,$AB,$FF,$45,$CF
	.dc.b	$F8,$7C,$FE,$C3,$F0,$FF,$D3,$2F,$FC,$19,$87,$F8,$A6,$B7,$1F,$6B
	.dc.b	$D7,$E4,$FF,$A4,$AF,$F0,$38,$63,$0E,$5C,$DF,$D5,$AF,$15,$AC,$39
	.dc.b	$CE,$5B,$1E,$23,$F1,$7F,$61,$C3,$1F,$04,$F8,$BC,$86,$6A,$D6,$1C
	.dc.b	$EF,$C8,$7A,$A3,$38,$75,$8C,$61,$EC,$5F,$66,$FA,$AF,$04,$3E,$CD
	.dc.b	$F6,$9F,$C7,$63,$A9,$C2,$77,$3A,$F1,$FA,$5F,$3C,$3B,$4E,$6F,$8C
	.dc.b	$E5,$E7,$19,$DF,$E9,$47,$87,$1F,$04,$62,$75,$97,$FB,$5A,$F9,$4E
	.dc.b	$6B,$73,$FE,$DA,$AD,$DF,$FA,$59,$AF,$A7,$0F,$2D,$70,$8D,$35,$F1
	.dc.b	$63,$EC,$CB,$DF,$39,$BE,$67,$4C,$BC,$78,$F8,$67,$44,$BE,$58,$F8
	.dc.b	$19,$8F,$C2,$9F,$05,$F2,$8F,$1C,$EB,$2F,$73,$CA,$38,$39,$C3,$C0
	.dc.b	$E6,$77,$3A,$23,$F3,$AB,$B4,$7B,$8C,$D7,$27,$EB,$E5,$F0,$2F,$B4
	.dc.b	$79,$0C,$E3,$87,$96,$1E,$03,$B5,$71,$8E,$A7,$18,$CC,$F1,$CB,$E5
	.dc.b	$E5,$F0,$AF,$85,$73,$C7,$C8,$F0,$C3,$99,$E0,$8C,$9F,$85,$9E,$2E
	.dc.b	$23,$C0,$E7,$0F,$05,$FD,$59,$F5,$9D,$30,$FC,$C3,$FE,$DE,$3E,$01
	.dc.b	$4F,$90,$57,$6F,$E4,$7E,$83,$FC,$04,$EC,$FD,$63,$87,$83,$5E,$1E
	.dc.b	$56,$DD,$BF,$AA,$8E,$79,$7E,$8D,$B8,$F9,$7E,$D1,$F5,$1F,$EF,$E8
	.dc.b	$E7,$87,$EB,$AB,$6F,$E8,$DA,$AB,$8B,$DF,$08,$F1,$7F,$84,$46,$A3
	.dc.b	$04,$C3,$C9,$3A,$C3,$C5,$3D,$E3,$9C,$EA,$3F,$85,$D7,$38,$C1,$30
	.dc.b	$F4,$BF,$C1,$23,$51,$82,$4F,$53,$9D,$F8,$1F,$92,$79,$63,$6F,$BE
	.dc.b	$35,$8B,$1F,$86,$DC,$7C,$F7,$F3,$6B,$75,$A2,$4F,$BE,$F6,$78,$F1
	.dc.b	$F5,$23,$C0,$68,$F2,$BA,$72,$79,$CE,$57,$E8,$C7,$E0,$C6,$08,$72
	.dc.b	$8F,$33,$D7,$1F,$53,$CC,$EB,$5D,$E3,$2F,$86,$BA,$CF,$33,$AC,$EA
	.dc.b	$BA,$C7,$38,$C3,$E2,$9E,$27,$28,$CC,$73,$3C,$33,$CA,$BC,$58,$7E
	.dc.b	$4A,$FD,$BC,$0F,$1F,$DF,$C7,$FD,$83,$D4,$FC,$5B,$72,$FE,$D7,$3C
	.dc.b	$0D,$F2,$2B,$6D,$9D,$ED,$8F,$27,$F2,$B8,$F7,$8B,$87,$29,$D5,$70
	.dc.b	$21,$FD,$BB,$1F,$EE,$71,$DB,$CA,$FE,$0D,$DF,$FE,$06,$C7,$D5,$F1
	.dc.b	$B8,$E4,$FC,$14,$63,$1F,$F0,$CD,$E5,$FF,$08,$B6,$35,$1F,$E8,$00
	.dc.b	$E2,$FF,$C2,$78,$FF,$81,$2B,$93,$FE,$A5,$EE,$E5,$F5,$FF,$E6,$5C
	.dc.b	$3F,$D2,$F8,$23,$11,$D6,$B8,$3F,$E3,$8A,$E3,$E5,$48,$C1,$FE,$94
	.dc.b	$EE,$76,$FA,$63,$FC,$60,$D8,$CC,$6F,$F0,$DC,$4C,$8E,$4E,$E3,$11
	.dc.b	$EF,$39,$9A,$39,$1F,$E6,$2F,$B0,$F9,$1E,$7B,$C8,$F6,$4F,$F9,$9B
	.dc.b	$4C,$C2,$A9,$88,$93,$CC,$FF,$55,$7E,$36,$F9,$2E,$AD,$8C,$C7,$CE
	.dc.b	$F9,$17,$C1,$EA,$24,$E8,$FF,$62,$F0,$2B,$71,$FE,$BC,$FA,$D7,$CC
	.dc.b	$FC,$96,$C6,$63,$17,$96,$46,$4E,$51,$87,$E4,$4F,$26,$CF,$27,$F0
	.dc.b	$3E,$0B,$E5,$BF,$1C,$D8,$CC,$7E,$2B,$B7,$11,$F9,$1E,$6F,$F2,$24
	.dc.b	$F0,$24,$7F,$9C,$9F,$A7,$EB,$FF,$91,$E3,$FD,$93,$F0,$78,$33,$E4
	.dc.b	$48,$FF,$68,$C6,$5F,$F6,$9C,$7E,$2F,$EB,$FF,$D4,$71,$FE,$E2,$FA
	.dc.b	$D7,$DD,$7E,$3D,$B1,$FE,$E6,$9D,$BF,$89,$6F,$77,$FB,$EE,$77,$3F
	.dc.b	$8D,$6C,$7F,$BF,$E7,$7D,$FF,$DF,$DB,$5F,$CE,$5A,$9B,$5F,$AA,$AF
	.dc.b	$E0,$1C,$DF,$D2,$EB,$F5,$ED,$7E,$53,$63,$F7,$3E,$8F,$F6,$B7,$8F
	.dc.b	$8F,$FB,$82,$57,$FA,$AB,$C5,$FE,$09,$F8,$B3,$F0,$4B,$AC,$AF,$DE
	.dc.b	$CE,$5B,$39,$7E,$D4,$7C,$0D,$7F,$89,$BB,$8F,$D8,$7E,$40,$FB,$7F
	.dc.b	$DE,$36,$3F,$AD,$67,$FE,$A2,$51,$97,$8F,$B5,$C2,$A4,$70,$C3,$FC
	.dc.b	$55,$5B,$BE,$4A,$F7,$5E,$2B,$B5,$A9,$FE,$49,$F8,$E7,$CF,$C3,$F9
	.dc.b	$5A,$F7,$B2,$6B,$C7,$FE,$55,$F1,$8F,$80,$A4,$CF,$F9,$1F,$3F,$FA
	.dc.b	$85,$F3,$0F,$8D,$7D,$73,$F0,$AF,$E6,$95,$BA,$FD,$23,$76,$FF,$62
	.dc.b	$DF,$0A,$FE,$75,$F2,$2F,$7E,$93,$0F,$E6,$70,$F9,$1D,$92,$32,$F8
	.dc.b	$39,$E4,$7A,$1A,$FF,$45,$C3,$CF,$F4,$9F,$D2,$7E,$5F,$F8,$C4,$FE
	.dc.b	$E5,$B9,$FF,$55,$6F,$1F,$F8,$57,$C9,$7C,$BF,$F9,$76,$B5,$E4,$F3
	.dc.b	$26,$1F,$C7,$27,$57,$FD,$73,$0E,$73,$FD,$59,$BE,$29,$CE,$3D,$A7
	.dc.b	$FA,$F7,$D0,$FE,$86,$FF,$B2,$B6,$3F,$AF,$B6,$5F,$32,$FA,$8E,$3B
	.dc.b	$CB,$FE,$BB,$7B,$AD,$46,$27,$4F,$F1,$B9,$D5,$7F,$6F,$F8,$CF,$C3
	.dc.b	$4F,$F6,$66,$BF,$C8,$5B,$B7,$F7,$0F,$9D,$FC,$94,$FF,$74,$9C,$E7
	.dc.b	$FD,$F9,$B1,$9E,$8F,$EE,$B5,$C3,$65,$69,$B3,$E3,$37,$D5,$FE,$03
	.dc.b	$F5,$3F,$99,$E5,$FE,$0D,$6E,$5F,$E0,$C6,$E4,$DD,$F0,$DF,$AD,$8F
	.dc.b	$F0,$0C,$EB,$23,$4E,$35,$C3,$BF,$C1,$E6,$4B,$E2,$FF,$83,$B0,$E9
	.dc.b	$1F,$D1,$7E,$01,$F0,$FF,$E1,$96,$C7,$F8,$41,$AF,$F8,$3F,$E7,$DE
	.dc.b	$E6,$EA,$4E,$DF,$F0,$DC,$EA,$30,$43,$C1,$1A,$9F,$F1,$47,$E2,$EF
	.dc.b	$89,$A4,$AF,$EE,$7E,$BF,$F3,$CD,$D5,$B9,$FF,$12,$B5,$FE,$C0,$DE
	.dc.b	$1F,$F1,$7D,$71,$EE,$7D,$BF,$F8,$DD,$B7,$FE,$2E,$6B,$FD,$A5,$FE
	.dc.b	$60,$FF,$8E,$9B,$87,$F0,$F9,$E4,$73,$F3,$7F,$C5,$9F,$4D,$FE,$06
	.dc.b	$C7,$FD,$15,$D7,$FD,$D3,$E9,$DE,$17,$FB,$56,$30,$DE,$4F,$F6,$C5
	.dc.b	$F1,$D1,$5F,$DF,$23,$0F,$08,$EB,$3F,$ED,$FB,$EB,$18,$3F,$D3,$55
	.dc.b	$C2,$7A,$FB,$2C,$75,$4B,$E2,$F8,$2B,$A6,$3E,$58,$E7,$7C,$5D,$18
	.dc.b	$BE,$77,$CB,$1F,$34,$E8,$63,$E3,$78,$19,$9E,$18,$F8,$23,$CB,$3B
	.dc.b	$78,$E1,$E1,$BE,$A4,$AF,$6E,$1C,$27,$51,$CE,$BB,$63,$EB,$79,$D6
	.dc.b	$CF,$F7,$D6,$1F,$46,$3F,$D8,$47,$98,$A7,$02,$9E,$CF,$0D,$F2,$5C
	.dc.b	$9B,$1A,$9F,$F2,$B4,$E4,$E3,$1D,$B9,$AB,$FE,$D1,$6A,$69,$1F,$99
	.dc.b	$F4,$EF,$29,$DB,$F7,$27,$91,$A8,$CB,$F4,$CC,$ED,$FE,$48,$AE,$55
	.dc.b	$C6,$7C,$07,$17,$17,$F8,$89,$DC,$F3,$37,$5C,$EF,$BB,$FE,$54,$8E
	.dc.b	$FA,$FC,$CA,$E3,$39,$9E,$33,$9A,$E2,$FC,$EB,$F0,$46,$27,$E9,$C6
	.dc.b	$30,$F2,$C7,$9C,$F2,$46,$5A,$66,$FC,$11,$97,$FC,$D9,$C4,$FA,$77
	.dc.b	$C3,$7C,$8D,$FF,$D8,$D1,$FE,$8F,$C7,$FD,$6A,$E9,$BB,$ED,$FF,$64
	.dc.b	$C6,$72,$FF,$3A,$3C,$5B,$7B,$7F,$D8,$F1,$9B,$FF,$6A,$57,$F8,$3C
	.dc.b	$FF,$61,$96,$7F,$D8,$5C,$CF,$A7,$E5,$97,$E0,$BC,$AB,$7B,$C3,$9C
	.dc.b	$77,$F0,$3F,$D3,$D7,$A8,$E3,$F4,$EF,$A7,$A6,$FC,$07,$CE,$63,$2F
	.dc.b	$F6,$AD,$6D,$B5,$B9,$D6,$BF,$E9,$3A,$D3,$67,$55,$C0,$FE,$BF,$F1
	.dc.b	$BE,$AF,$14,$F9,$FC,$4B,$3A,$25,$FF,$A7,$EB,$2F,$FB,$96,$6E,$DF
	.dc.b	$DC,$6F,$A1,$2F,$FD,$7F,$3A,$7F,$BA,$47,$CA,$FA,$7F,$D5,$C7,$FB
	.dc.b	$C9,$B7,$FE,$F1,$F2,$9F,$2D,$BA,$86,$CF,$9D,$1A,$70,$49,$FF,$5B
	.dc.b	$57,$4C,$7F,$C2,$D7,$CA,$7F,$D7,$5F,$4E,$FA,$5F,$BA,$DC,$3F,$DA
	.dc.b	$FE,$83,$E8,$F3,$52,$7F,$CC,$BC,$CB,$E1,$87,$F4,$19,$FF,$25,$61
	.dc.b	$C5,$DC,$EB,$2F,$E0,$F7,$C7,$0E,$98,$71,$3F,$C3,$18,$71,$7C,$F1
	.dc.b	$F9,$2F,$49,$F4,$F4,$FC,$87,$A8,$F8,$FC,$69,$E6,$AD,$B6,$B7,$3A
	.dc.b	$C3,$FC,$C9,$17,$3F,$E3,$7D,$5F,$E6,$2A,$C5,$AD,$A3,$DA,$F6,$4E
	.dc.b	$63,$1A,$7E,$C7,$CC,$7D,$3F,$D7,$C3,$FD,$D5,$87,$07,$9F,$B0,$F8
	.dc.b	$CF,$D7,$B7,$2F,$9F,$5B,$C9,$39,$BF,$F7,$A3,$75,$F9,$BE,$33,$E5
	.dc.b	$B9,$E5,$FB,$A8,$C6,$5F,$E9,$ED,$3C,$BE,$07,$9F,$5F,$DE,$4F,$FA
	.dc.b	$0F,$BD,$F3,$B9,$AF,$E0,$7B,$FF,$8F,$AB,$CB,$EA,$E3,$2C,$BF,$C2
	.dc.b	$3E,$99,$F4,$F7,$AF,$0D,$BF,$F5,$7F,$D3,$FC,$DF,$98,$FE,$30,$D8
	.dc.b	$CF,$27,$EA,$C4,$65,$FC,$43,$78,$FF,$A9,$B6,$7E,$C3,$F6,$7C,$1E
	.dc.b	$B4,$BA,$4A,$EF,$36,$7F,$29,$CC,$FA,$67,$D3,$FF,$24,$5F,$33,$63
	.dc.b	$0D,$C3,$F4,$F3,$A6,$CE,$9F,$D5,$FD,$3F,$F2,$26,$F2,$FF,$2B,$F8
	.dc.b	$E7,$D5,$7F,$32,$AD,$67,$E4,$6E,$FF,$AC,$6E,$1F,$EB,$BB,$E2,$F4
	.dc.b	$3F,$DB,$B4,$FA,$7F,$F7,$84,$61,$B1,$FB,$3F,$94,$7D,$47,$F4,$3A
	.dc.b	$D7,$C1,$FF,$4D,$AE,$4D,$BE,$39,$7F,$81,$30,$E3,$E8,$E6,$A6,$1F
	.dc.b	$EE,$E9,$7D,$43,$E8,$B3,$97,$86,$31,$5F,$EF,$4F,$41,$F3,$0F,$A0
	.dc.b	$FE,$BD,$18,$E8,$B3,$9A,$E0,$FD,$F6,$D6,$A7,$6D,$8F,$E3,$BF,$43
	.dc.b	$FE,$B4,$6F,$C5,$FF,$5D,$F9,$B7,$C6,$7F,$64,$AD,$69,$F8,$66,$F0
	.dc.b	$FF,$1E,$6E,$1F,$EE,$DB,$E2,$F4,$3F,$43,$F0,$4F,$92,$FD,$0B,$63
	.dc.b	$3F,$1B,$89,$91,$C1,$DC,$63,$1F,$D1,$D7,$16,$CE,$AB,$81,$FC,$EB
	.dc.b	$CC,$FA,$EF,$F1,$96,$CF,$FB,$27,$07,$D4,$7F,$80,$A7,$3E,$07,$F0
	.dc.b	$DE,$6C,$1F,$16,$6E,$1F,$5F,$E8,$7F,$84,$B7,$EA,$3E,$BC,$63,$2F
	.dc.b	$E8,$1C,$A8,$4E,$BC,$CD,$B6,$F9,$1F,$23,$D9,$F8,$E7,$C8,$F9,$3B
	.dc.b	$19,$26,$3F,$12,$32,$FC,$69,$BC,$7F,$C2,$38,$55,$8B,$FD,$57,$D4
	.dc.b	$7E,$C7,$E4,$BF,$68,$D8,$FE,$89,$7F,$D2,$3C,$14,$E7,$2F,$EB,$9E
	.dc.b	$BA,$73,$7F,$C0,$30,$FC,$67,$9D,$F3,$7C,$9B,$3F,$55,$B8,$7F,$12
	.dc.b	$E5,$FE,$14,$F4,$E6,$FE,$64,$E6,$7F,$2D,$EF,$FF,$9D,$3E,$23,$FA
	.dc.b	$14,$FF,$58,$BF,$EC,$D8,$7F,$A0,$DF,$CA,$56,$DF,$C1,$7B,$36,$BF
	.dc.b	$05,$37,$A7,$FA,$64,$E7,$66,$6A,$D6,$1D,$B0,$FF,$47,$3D,$31,$F0
	.dc.b	$47,$E3,$7E,$4B,$78,$F3,$C3,$FD,$F1,$5A,$CB,$BB,$C3,$2E,$F7,$C2
	.dc.b	$31,$5C,$30,$EE,$7F,$77,$FA,$87,$D4,$7A,$E1,$F4,$4F,$A8,$79,$3E
	.dc.b	$49,$5F,$0C,$71,$7B,$AD,$4E,$6B,$B9,$D2,$F9,$B8,$8D,$70,$7D,$28
	.dc.b	$D5,$74,$AD,$CE,$6F,$B5,$71,$C7,$D1,$18,$9D,$57,$0D,$FF,$D9,$73
	.dc.b	$97,$FD,$9B,$8F,$81,$CC,$EF,$0F,$F0,$D5,$6D,$B1,$97,$E0,$E1,$E0
	.dc.b	$8C,$4E,$E7,$D2,$79,$72,$F5,$65,$23,$E1,$CE,$F1,$F1,$47,$69,$D6
	.dc.b	$5E,$A7,$E5,$E3,$DD,$CE,$5E,$87,$38,$E2,$F9,$1D,$B2,$F4,$63,$DA
	.dc.b	$3C,$B8,$F5,$75,$7C,$9A,$48,$D1,$F5,$67,$96,$5E,$F2,$5F,$6C,$BC
	.dc.b	$D7,$C7,$2F,$33,$F0,$CC,$1E,$6C,$3A,$C7,$D4,$9E,$31,$89,$D4,$E7
	.dc.b	$2F,$74,$7A,$A3,$AC,$F9,$6B,$8C,$60,$93,$F9,$F8,$CB,$D0,$87,$DC
	.dc.b	$AE,$B8,$7B,$32,$F8,$98,$78,$EB,$EE,$9F,$E5,$F3,$E7,$94,$C9,$F0
	.dc.b	$0C,$4B,$E8,$3F,$BD,$46,$CF,$D7,$B8,$76,$D7,$F3,$2D,$7F,$2A,$F2
	.dc.b	$9F,$D2,$37,$1D,$F9,$5F,$B7,$FA,$87,$CD,$79,$EF,$73,$FB,$8B,$E1
	.dc.b	$87,$48,$E1,$5B,$C3,$9C,$FE,$62,$B6,$FA,$2B,$6F,$36,$3B,$72,$D2
	.dc.b	$5F,$89,$FE,$0B,$87,$99,$E3,$19,$25,$FC,$F9,$D4,$F2,$8F,$2C,$EA
	.dc.b	$3F,$D2,$53,$A8,$C4,$79,$CE,$06,$B0,$F7,$3F,$04,$C9,$B8,$FE,$29
	.dc.b	$3C,$EB,$B9,$27,$84,$EA,$B9,$C7,$19,$D3,$4F,$01,$EE,$73,$8F,$E1
	.dc.b	$70,$F8,$F7,$E8,$8C,$57,$93,$0F,$13,$E0,$BE,$38,$FD,$49,$CC,$F7
	.dc.b	$25,$72,$BF,$34,$78,$4E,$44,$13,$CA,$78,$5F,$6C,$3C,$AD,$AF,$26
	.dc.b	$3F,$02,$F8,$63,$F8,$6C,$38,$D7,$C0,$3E,$FD,$7C,$68,$CC,$79,$EB
	.dc.b	$DB,$C4,$F1,$BD,$FC,$7F,$DA,$BD,$5F,$AE,$DB,$FE,$C5,$3C,$0D,$F1
	.dc.b	$AB,$6D,$9D,$E5,$8F,$2F,$F0,$38,$F7,$8B,$87,$29,$D5,$70,$21,$FD
	.dc.b	$AB,$1F,$EF,$B1,$DB,$CB,$FE,$09,$DD,$FE,$5D,$F8,$EB,$C9,$FE,$E1
	.dc.b	$C5,$18,$C7,$FC,$95,$77,$1F,$F0,$3B,$9C,$FE,$15,$5B,$E3,$7F,$E2
	.dc.b	$CC,$3F,$CD,$17,$C7,$B9,$DD,$C7,$E8,$9E,$5D,$35,$13,$0E,$47,$FB
	.dc.b	$E2,$76,$E3,$34,$62,$3F,$C6,$F7,$D9,$FF,$53,$F7,$BB,$BD,$91,$FE
	.dc.b	$C0,$6C,$66,$37,$F5,$18,$99,$1C,$9D,$C6,$21,$C0,$D7,$F9,$85,$F3
	.dc.b	$9C,$C6,$2F,$8C,$7F,$9B,$7E,$A3,$E7,$79,$5F,$E6,$9C,$EF,$95,$EC
	.dc.b	$AD,$10,$FF,$37,$B9,$6C,$70,$AE,$7F,$50,$F9,$3A,$4A,$F9,$7C,$AF
	.dc.b	$92,$7C,$36,$67,$FC,$F7,$7C,$5F,$F5,$C3,$5F,$F3,$1C,$F0,$7F,$DA
	.dc.b	$7F,$50,$F9,$AF,$9A,$D8,$CC,$63,$1C,$B2,$32,$72,$8C,$3F,$20,$09
	.dc.b	$E8,$E4,$C4,$F5,$7F,$09,$DE,$F9,$7F,$CB,$36,$33,$1F,$9B,$EE,$C4
	.dc.b	$77,$E7,$FF,$26,$CF,$02,$47,$C8,$9D,$F2,$FD,$F8,$71,$7F,$B7,$3C
	.dc.b	$5F,$B7,$EB,$BE,$16,$98,$89,$5F,$ED,$C8,$CB,$FE,$DE,$8F,$CA,$FD
	.dc.b	$43,$D4,$FC,$BF,$D4,$3E,$7A,$9A,$89,$7F,$A6,$15,$C5,$C3,$5F,$91
	.dc.b	$5F,$A7,$CF,$D5,$1C,$2B,$9B,$F3,$6B,$84,$EA,$B8,$75,$FD,$4D,$FF
	.dc.b	$63,$E4,$7E,$EC,$98,$7E,$3B,$BF,$FB,$1E,$1F,$1E,$B7,$3F,$B9,$EB
	.dc.b	$FF,$08,$36,$FF,$83,$78,$DF,$C3,$7E,$9D,$F5,$17,$79,$7F,$C0,$67
	.dc.b	$2D,$9C,$BF,$6E,$3E,$43,$93,$E3,$7D,$43,$EA,$3F,$81,$B6,$33,$18
	.dc.b	$F7,$32,$C8,$C9,$CA,$30,$FE,$4E,$79,$36,$79,$3F,$C5,$FE,$A3,$E4
	.dc.b	$AD,$50,$F7,$BF,$96,$36,$78,$79,$3F,$CB,$FE,$39,$F0,$14,$93,$FC
	.dc.b	$8F,$A7,$FD,$4E,$FF,$28,$F9,$17,$D4,$3E,$A3,$F9,$E4,$61,$B1,$99
	.dc.b	$FE,$0B,$F1,$7F,$31,$FD,$27,$CA,$3E,$4F,$FA,$03,$FE,$EE,$8F,$E6
	.dc.b	$4D,$77,$1A,$23,$A7,$FA,$46,$3F,$84,$EB,$FC,$53,$67,$E2,$37,$83
	.dc.b	$F5,$5C,$FE,$6F,$A4,$FE,$A1,$F3,$1F,$C4,$FB,$3F,$18,$DC,$7F,$28
	.dc.b	$DC,$5E,$6B,$E6,$7F,$CC,$57,$C3,$CB,$E6,$4A,$FE,$4B,$3A,$C3,$9C
	.dc.b	$E6,$7F,$83,$61,$D7,$0E,$91,$FD,$8F,$E8,$BD,$04,$CF,$FB,$47,$CD
	.dc.b	$BE,$A3,$51,$2B,$FA,$D5,$7C,$1B,$5E,$E7,$35,$B7,$F8,$B5,$6E,$75
	.dc.b	$1F,$C6,$7E,$37,$F0,$AF,$F6,$56,$B9,$E7,$FE,$F5,$F3,$FF,$92,$E1
	.dc.b	$56,$B8,$3F,$6C,$DD,$FF,$7C,$DC,$3E,$5D,$F1,$8C,$5F,$2C,$3F,$C0
	.dc.b	$91,$BD,$15,$DF,$CB,$FE,$53,$8F,$F0,$96,$E3,$FA,$36,$D3,$E1,$DE
	.dc.b	$0F,$F8,$3E,$77,$9F,$F3,$0F,$3B,$07,$F8,$56,$B7,$87,$F8,$16,$B6
	.dc.b	$49,$D5,$FF,$86,$7E,$0D,$F1,$14,$97,$FD,$9B,$E8,$BF,$C1,$4F,$F4
	.dc.b	$59,$DB,$FE,$11,$6B,$99,$FF,$13,$4F,$0E,$0F,$E4,$15,$A6,$CF,$6A
	.dc.b	$FE,$BD,$A2,$F8,$6B,$FC,$B7,$EA,$1F,$15,$FE,$2F,$6D,$FF,$88,$5A
	.dc.b	$FF,$74,$F4,$35,$32,$34,$FF,$8E,$7D,$7F,$E7,$77,$FC,$E7,$C3,$FE
	.dc.b	$4B,$5F,$D8,$57,$D4,$7F,$8F,$5B,$1F,$E2,$F7,$7A,$FF,$3B,$6D,$73
	.dc.b	$AE,$07,$1F,$37,$FC,$5F,$F4,$FF,$E0,$5B,$FE,$F7,$D5,$FD,$C3,$EA
	.dc.b	$1E,$0F,$FB,$5E,$30,$DE,$5F,$F6,$D5,$F1,$D1,$5F,$DE,$23,$0F,$08
	.dc.b	$EB,$3F,$EE,$1B,$EB,$18,$3F,$C1,$75,$C2,$7A,$FB,$3F,$E3,$A9,$EB
	.dc.b	$1C,$70,$F1,$47,$F8,$4A,$F9,$BA,$39,$D7,$1C,$3C,$8D,$BE,$53,$E1
	.dc.b	$15,$CA,$B7,$8E,$71,$F2,$57,$7C,$7C,$2F,$06,$C6,$A0,$DB,$E9,$87
	.dc.b	$82,$7C,$C6,$B1,$F2,$3E,$8C,$3C,$95,$C2,$3C,$07,$8E,$76,$F1,$9F
	.dc.b	$F1,$9B,$0C,$14,$A6,$6F,$EC,$37,$C1,$7C,$86,$E3,$D3,$18,$79,$4F
	.dc.b	$F9,$66,$74,$D8,$E0,$39,$EB,$4C,$B8,$23,$F0,$BE,$A3,$FC,$69,$C1
	.dc.b	$F4,$E3,$FC,$9B,$3F,$56,$73,$1F,$20,$C9,$C7,$C5,$FF,$20,$6F,$FE
	.dc.b	$48,$8C,$BF,$E4,$A9,$E2,$E6,$3C,$8E,$67,$8E,$DF,$1E,$FC,$11,$8A
	.dc.b	$F9,$91,$8C,$3C,$D1,$CA,$BE,$8E,$1D,$9E,$51,$E5,$9D,$D7,$18,$F5
	.dc.b	$1C,$4F,$F3,$5D,$71,$7B,$9C,$4D,$C7,$1C,$3D,$4D,$BF,$0E,$1F,$E6
	.dc.b	$1D,$3F,$C9,$38,$7C,$87,$FC,$97,$87,$C4,$AD,$E3,$FE,$55,$AD,$3E
	.dc.b	$AF,$C4,$79,$3A,$AF,$F3,$E4,$A7,$31,$E8,$F9,$CB,$83,$F1,$A7,$C4
	.dc.b	$6A,$98,$2F,$75,$AC,$7C,$51,$8B,$E1,$5B,$9F,$F6,$7E,$1D,$67,$24
	.dc.b	$8F,$F6,$9F,$B8,$F9,$0F,$F3,$A4,$7B,$67,$59,$7F,$A6,$69,$F6,$0F
	.dc.b	$3F,$38,$FF,$A2,$67,$31,$8C,$3F,$D3,$F1,$87,$A1,$E0,$AE,$4F,$F9
	.dc.b	$FF,$E4,$5F,$81,$7C,$16,$C7,$FB,$A0,$7E,$2C,$BD,$54,$9D,$4F,$D4
	.dc.b	$CB,$FD,$5F,$7C,$9F,$F5,$94,$F2,$E2,$E1,$18,$AE,$0A,$B8,$67,$FE
	.dc.b	$A5,$BE,$07,$FA,$57,$EB,$5F,$8B,$FD,$76,$C7,$8B,$8F,$13,$23,$CA
	.dc.b	$DC,$3C,$0D,$BE,$D8,$F2,$9F,$F6,$26,$16,$73,$5F,$85,$BF,$F7,$21
	.dc.b	$FE,$A7,$2D,$7A,$FE,$33,$5D,$52,$7E,$57,$C7,$3E,$AB,$17,$F1,$23
	.dc.b	$13,$CE,$BF,$2E,$A3,$33,$FE,$D7,$8E,$F8,$FC,$59,$CB,$FE,$B2,$FA
	.dc.b	$5F,$CB,$57,$FB,$75,$B1,$EE,$9F,$CF,$46,$2B,$F9,$9B,$69,$1F,$80
	.dc.b	$F9,$0D,$F2,$F9,$36,$46,$27,$F0,$2D,$AF,$B8,$FE,$AF,$EB,$9F,$65
	.dc.b	$4F,$D7,$7E,$10,$FA,$3C,$61,$FE,$D1,$AD,$B7,$E7,$B3,$8F,$FB,$DE
	.dc.b	$51,$D4,$FD,$C7,$E5,$7F,$89,$AE,$07,$FB,$6D,$C5,$71,$E1,$FD,$F7
	.dc.b	$4F,$FA,$E6,$BC,$11,$8F,$86,$91,$FE,$E3,$97,$CD,$E2,$DC,$7F,$EE
	.dc.b	$09,$CD,$7F,$BA,$7D,$57,$CB,$7F,$0B,$F0,$1F,$45,$FC,$36,$73,$5F
	.dc.b	$EE,$F9,$CB,$FE,$F1,$F8,$B7,$D5,$7F,$13,$6C,$33,$F0,$B7,$D5,$7E
	.dc.b	$58,$FB,$2E,$18,$4A,$E1,$B7,$DC,$C3,$F8,$EF,$E5,$FF,$E8,$B9,$F0
	.dc.b	$57,$F2,$6F,$21,$39,$AE,$E7,$87,$3F,$1B,$89,$DF,$C2,$77,$8C,$57
	.dc.b	$1C,$7C,$1D,$DE,$9F,$C7,$BF,$99,$BE,$BE,$FB,$69,$B1,$86,$E9,$EA
	.dc.b	$3E,$15,$73,$71,$48,$E9,$18,$7E,$47,$D0,$F6,$3B,$F3,$59,$D5,$27
	.dc.b	$0F,$C0,$AF,$A1,$58,$C7,$E5,$4B,$EA,$9F,$65,$FD,$25,$B8,$FF,$3E
	.dc.b	$FA,$CF,$BB,$5F,$28,$E1,$5B,$3F,$AA,$F4,$A7,$F9,$0D,$7F,$89,$AF
	.dc.b	$9D,$FF,$43,$9C,$BF,$83,$8C,$E1,$FD,$0E,$73,$3D,$AB,$C5,$E6,$7D
	.dc.b	$0F,$57,$F7,$B1,$F1,$5F,$C9,$CF,$BD,$CB,$7A,$F8,$4F,$16,$BC,$5E
	.dc.b	$9F,$2F,$23,$A7,$D3,$DA,$E9,$37,$2F,$C1,$E0,$AC,$5F,$BF,$E1,$7E
	.dc.b	$51,$F9,$3F,$C2,$DF,$80,$E2,$7E,$FE,$5F,$46,$BF,$BD,$FE,$17,$F3
	.dc.b	$B2,$FA,$BB,$FF,$B0,$6F,$FB,$9C,$E5,$EC,$63,$C1,$E9,$87,$E7,$AD
	.dc.b	$5B,$8F,$C6,$68,$F9,$8F,$94,$7C,$BF,$97,$B8,$FC,$B8,$B5,$F2,$E6
	.dc.b	$E1,$F0,$3E,$0B,$E9,$7C,$17,$D3,$6E,$1F,$4D,$B7,$F4,$DB,$5F,$90
	.dc.b	$FB,$2F,$F0,$3C,$FE,$2D,$B1,$FB,$DE,$F7,$0F,$F2,$AA,$D6,$3F,$E1
	.dc.b	$DB,$DE,$3F,$E0,$6F,$B0,$7D,$66,$6B,$F2,$71,$89,$F9,$F5,$C2,$FF
	.dc.b	$C0,$53,$97,$AC,$FF,$87,$23,$19,$7D,$EF,$BC,$FC,$87,$C5,$7E,$45
	.dc.b	$B5,$F9,$16,$DF,$E4,$5B,$5F,$86,$E2,$FC,$C7,$4F,$E6,$5B,$7F,$99
	.dc.b	$6D,$FE,$65,$B5,$FB,$6F,$AE,$F8,$5F,$07,$EE,$AF,$E5,$50,$FA,$DF
	.dc.b	$E4,$93,$99,$4F,$0D,$FF,$D2,$EE,$F1,$FF,$3F,$C6,$3D,$5E,$2E,$B0
	.dc.b	$B7,$DA,$6B,$D2,$2D,$F3,$8F,$F4,$95,$71,$BE,$B1,$C5,$AF,$16,$CF
	.dc.b	$1C,$3D,$0B,$ED,$3A,$E1,$D9,$EF,$E7,$78,$BE,$A3,$F6,$1E,$17,$28
	.dc.b	$C7,$7F,$F8,$F7,$3F,$19,$F1,$A3,$33,$BC,$FC,$77,$CA,$B6,$1C,$67
	.dc.b	$F2,$DB,$A7,$CA,$C3,$9D,$70,$9C,$BF,$D7,$5A,$FD,$28,$DD,$73,$C7
	.dc.b	$CC,$DC,$7C,$95,$CE,$73,$87,$FB,$3B,$0E,$67,$4D,$3F,$CE,$19,$7A
	.dc.b	$AB,$77,$E1,$6B,$EB,$6C,$F0,$8E,$D8,$71,$23,$F0,$A3,$58,$F9,$63
	.dc.b	$64,$7E,$BE,$5E,$CC,$BC,$D8,$78,$32,$F3,$4E,$6F,$C9,$8F,$6C,$38
	.dc.b	$E5,$71,$E3,$36,$BE,$C4,$D7,$D5,$17,$2F,$2D,$70,$9F,$23,$6F,$AC
	.dc.b	$60,$F0,$94,$FA,$97,$D7,$1F,$9B,$97,$C6,$CB,$D1,$5E,$03,$86,$3E
	.dc.b	$07,$9E,$1F,$14,$98,$FE,$17,$1E,$98,$7C,$33,$A3,$8C,$BD,$58,$F6
	.dc.b	$CB,$F2,$58,$FF,$89,$4F,$39,$4F,$11,$5D,$BF,$85,$FA,$3F,$F4,$83
	.dc.b	$B3,$F3,$4E,$1E,$0D,$78,$7C,$0E,$DD,$BF,$A4,$8E,$79,$FE,$85,$B9
	.dc.b	$F9,$7E,$BD,$F3,$5E,$7A,$D5,$F1,$26,$1F,$AC,$AD,$B8,$9D,$BF,$A9
	.dc.b	$C3,$AB,$FE,$45,$9D,$B9,$AE,$18,$7A,$23,$A6,$1E,$18,$C6,$1D,$08
	.dc.b	$F9,$E3,$53,$FB,$5C,$3B,$BF,$00,$EC,$74,$3B,$9B,$BA,$B0,$F2,$E1
	.dc.b	$DA,$34,$72,$7F,$99,$46,$B0,$F2,$92,$77,$3A,$C3,$CB,$1E,$38,$E8
	.dc.b	$D9,$F5,$3E,$2C,$58,$FA,$9B,$8F,$F2,$EC,$7C,$15,$DA,$FC,$0F,$33
	.dc.b	$E7,$63,$E0,$BE,$31,$E9,$9D,$D7,$28,$E0,$49,$DE,$3E,$98,$F2,$63
	.dc.b	$E3,$3E,$6D,$71,$AD,$DF,$89,$B5,$E2,$38,$E3,$F3,$AB,$85,$F3,$8C
	.dc.b	$63,$F1,$A7,$31,$8A,$F8,$13,$EB,$AE,$11,$F6,$CF,$C9,$7C,$37,$1F
	.dc.b	$B4,$91,$F1,$E3,$A1,$1C,$1F,$D6,$3E,$C1,$F0,$DF,$59,$B1,$A3,$A9
	.dc.b	$B8,$D1,$CD,$C9,$FD,$3A,$38,$BF,$D4,$3E,$01,$E3,$FF,$01,$E8,$FF
	.dc.b	$3A,$C3,$E5,$B1,$C9,$FC,$16,$31,$7F,$E4,$A7,$8B,$6F,$6F,$FA,$BE
	.dc.b	$36,$FF,$9E,$5B,$7D,$A7,$73,$FE,$E4,$F6,$ED,$3E,$5E,$98,$89,$3E
	.dc.b	$18,$CB,$E3,$8E,$7F,$8C,$BE,$5B,$9F,$A3,$2B,$FC,$C3,$39,$6C,$E5
	.dc.b	$F3,$4F,$28,$91,$A2,$1E,$BD,$BD,$3F,$5A,$E9,$BF,$05,$2B,$FC,$DF
	.dc.b	$39,$6C,$E5,$FF,$4E,$47,$FB,$D1,$C9,$C3,$E4,$FA,$91,$E7,$F1,$35
	.dc.b	$E3,$17,$83,$FD,$5B,$4A,$F9,$BF,$56,$F3,$7C,$FE,$9C,$C7,$FD,$07
	.dc.b	$39,$6C,$E5,$F2,$CF,$28,$91,$A2,$1E,$AE,$BF,$B7,$F5,$6F,$AC,$A4
	.dc.b	$9F,$F4,$8C,$3E,$CF,$11,$FE,$BD,$8E,$0E,$A7,$9C,$F4,$FA,$5E,$BF
	.dc.b	$61,$2B,$9C,$65,$EB,$19,$7A,$47,$E3,$BC,$6F,$F2,$B6,$FF,$ED,$2F
	.dc.b	$AC,$7D,$A7,$E4,$9B,$7F,$EA,$6A,$DF,$6F,$F6,$56,$C0,$4F,$07,$FD
	.dc.b	$AC,$DB,$FF,$6F,$CF,$09,$FF,$5C,$78,$DE,$32,$3F,$EB,$4A,$FD,$5B
	.dc.b	$CB,$AF,$C8,$47,$5D,$95,$D1,$C6,$3D,$51,$EB,$E9,$F6,$7C,$9C,$AE
	.dc.b	$D1,$97,$C1,$19,$7D,$31,$F1,$FA,$BF,$6B,$EC,$3C,$4A,$66,$BE,$34
	.dc.b	$E1,$FF,$66,$CE,$DC,$90,$FF,$7D,$69,$F9,$F6,$CF,$6E,$CF,$E0,$7F
	.dc.b	$8C,$B8,$FF,$82,$FE,$60,$FB,$6A,$6A,$37,$E9,$70,$FB,$19,$97,$FB
	.dc.b	$7B,$04,$FF,$0E,$FC,$0F,$F4,$93,$98,$E2,$FF,$B9,$23,$2B,$95,$CD
	.dc.b	$FE,$25,$97,$C6,$7F,$18,$FC,$C5,$F0,$D4,$ED,$1F,$C7,$3F,$32,$FE
	.dc.b	$5F,$87,$2C,$7F,$11,$87,$27,$C1,$5B,$E5,$91,$FE,$F6,$BE,$39,$7F
	.dc.b	$8D,$5B,$1F,$91,$FC,$6D,$F2,$9F,$CA,$BD,$AF,$D6,$4F,$FB,$FA,$7F
	.dc.b	$D6,$CD,$F6,$B5,$E0,$C8,$C1,$0F,$81,$8F,$F9,$0E,$F9,$C6,$0F,$D0
	.dc.b	$38,$D7,$D9,$08,$C4,$F1,$AF,$E7,$B8,$7F,$90,$EF,$7F,$03,$E8,$C3
	.dc.b	$F5,$51,$CB,$E9,$5F,$05,$FD,$11,$B3,$AC,$7F,$CA,$D7,$C6,$3C,$87
	.dc.b	$FB,$E3,$4E,$17,$23,$81,$FB,$79,$E1,$87,$8A,$BF,$A8,$C6,$2F,$AC
	.dc.b	$78,$7E,$27,$FA,$AF,$8A,$F9,$0F,$EA,$FF,$8C,$7E,$03,$2E,$A1,$C3
	.dc.b	$E4,$9E,$0E,$9A,$FA,$23,$79,$7C,$A6,$E3,$FE,$73,$AD,$CF,$CE,$6C
	.dc.b	$66,$FC,$A7,$93,$E4,$3F,$B5,$FE,$5A,$FA,$2F,$3F,$D7,$CC,$FC,$F3
	.dc.b	$96,$CE,$5F,$24,$F2,$89,$1A,$21,$E9,$F2,$3F,$BC,$FE,$35,$F8,$08
	.dc.b	$FE,$49,$AB,$D9,$F8,$E6,$B9,$64,$FC,$89,$D3,$67,$2F,$DC,$AE,$89
	.dc.b	$A6,$25,$F4,$3E,$2F,$87,$FA,$BE,$D4,$BF,$AB,$69,$D4,$78,$67,$4F
	.dc.b	$78,$FF,$09,$FD,$4B,$A2,$F5,$4A,$FB,$F3,$96,$CF,$02,$3F,$84,$FA
	.dc.b	$97,$B7,$FC,$2E,$1F,$52,$FA,$9F,$F5,$85,$6B,$A7,$FA,$64,$E6,$BF
	.dc.b	$A7,$63,$DD,$E3,$7F,$EB,$78,$C4,$EE,$57,$CD,$D9,$FD,$5D,$C4,$6E
	.dc.b	$78,$D7,$3C,$7F,$D8,$53,$CE,$B7,$3E,$6C,$7D,$53,$C6,$B7,$3C,$70
	.dc.b	$F4,$BB,$C6,$C7,$86,$F8,$CE,$E3,$CA,$31,$F1,$BB,$AE,$2F,$B2,$B8
	.dc.b	$BF,$12,$FC,$2D,$37,$8F,$A1,$E4,$76,$BE,$AF,$79,$E2,$F5,$BE,$37
	.dc.b	$DA,$34,$78,$AB,$8D,$F5,$8E,$07,$FB,$85,$FA,$0F,$FB,$A5,$AF,$C3
	.dc.b	$35,$EE,$3E,$4B,$FD,$C5,$3F,$E3,$76,$C6,$67,$71,$FE,$E7,$7F,$C2
	.dc.b	$1F,$13,$53,$48,$ED,$F9,$3F,$22,$FA,$7F,$EA,$5B,$7E,$1E,$B9,$9F
	.dc.b	$4A,$75,$DB,$F4,$23,$33,$F8,$D9,$CC,$7D,$83,$3B,$FF,$88,$D4,$6A
	.dc.b	$7B,$47,$98,$7D,$87,$3C,$3F,$B2,$3A,$C7,$FC,$58,$E7,$1B,$8F,$5B
	.dc.b	$EE,$7F,$8C,$71,$ED,$19,$8E,$D8,$F1,$7C,$C6,$23,$CD,$3C,$23,$A6
	.dc.b	$1F,$1B,$1F,$8B,$5C,$E3,$18,$7C,$3A,$E4,$78,$6B,$6F,$23,$E5,$38
	.dc.b	$7E,$24,$F2,$73,$3F,$D2,$2F,$E5,$EF,$FD,$AA,$7D,$A7,$CA,$BF,$F0
	.dc.b	$44,$7F,$6E,$9F,$A3,$3F,$0D,$AF,$B2,$79,$3E,$5A,$F8,$8F,$D4,$EA
	.dc.b	$CB,$63,$35,$FD,$1D,$B3,$A8,$C4,$FF,$80,$6D,$F5,$4E,$EF,$2E,$1F
	.dc.b	$33,$0B,$7F,$E2,$9A,$D4,$FB,$EB,$33,$F8,$67,$7F,$1D,$FE,$A9,$F1
	.dc.b	$73,$97,$0A,$FE,$FA,$D7,$E0,$57,$94,$97,$EA,$79,$FD,$53,$E2,$F9
	.dc.b	$C6,$1F,$0F,$C1,$7F,$43,$AF,$EF,$2D,$BF,$F0,$F4,$3E,$A1,$9F,$0E
	.dc.b	$B9,$3F,$32,$F8,$5F,$E5,$5A,$FE,$8D,$AF,$FA,$1D,$ED,$BF,$F9,$0E
	.dc.b	$7B,$70,$74,$AF,$0E,$3F,$84,$9F,$D6,$7D,$53,$DE,$F9,$2D,$8F,$F0
	.dc.b	$E9,$E7,$F3,$F3,$32,$3F,$03,$3B,$BF,$D3,$35,$FD,$63,$5E,$F3,$F6
	.dc.b	$A7,$E1,$3C,$E3,$CA,$FD,$A8,$FF,$67,$1F,$E2,$22,$C7,$AB,$C1,$D7
	.dc.b	$54,$9F,$C9,$9A,$9F,$DE,$35,$FE,$14,$D7,$FD,$5C,$F4,$CF,$E4,$7D
	.dc.b	$57,$F9,$82,$3C,$15,$E7,$8C,$57,$F9,$13,$E7,$BF,$15,$7F,$E5,$66
	.dc.b	$DF,$4D,$17,$F9,$49,$CD,$FF,$AF,$5F,$C8,$3F,$C7,$5A,$E7,$8B,$F5
	.dc.b	$71,$88,$3C,$AF,$F1,$5E,$2F,$5A,$F9,$98,$7F,$89,$67,$FC,$93,$F5
	.dc.b	$1F,$E6,$96,$F8,$7F,$A6,$FA,$A7,$D5,$7E,$A2,$B5,$9F,$F9,$81,$AF
	.dc.b	$F2,$56,$BF,$CC,$DA,$FF,$B5,$DA,$97,$C0,$7F,$04,$6E,$1F,$C2,$1B
	.dc.b	$3F,$CB,$5A,$F7,$F4,$78,$12,$78,$3F,$C3,$BA,$1E,$65,$24,$FF,$14
	.dc.b	$8C,$BF,$C5,$BE,$A5,$E9,$7F,$16,$6C,$66,$30,$7F,$1A,$FA,$6F,$F5
	.dc.b	$9C,$7F,$9C,$9B,$8F,$FA,$5F,$EA,$5F,$52,$FA,$9F,$E2,$35,$AE,$7F
	.dc.b	$F4,$CB,$C2,$FF,$A7,$35,$FE,$B8,$D7,$FE,$00,$35,$3C,$F0,$FE,$A9
	.dc.b	$F5,$3F,$E8,$77,$C9,$EE,$7F,$24,$F1,$13,$9B,$FF,$56,$74,$97,$D4
	.dc.b	$FF,$2A,$8C,$74,$D9,$CE,$3F,$E9,$C9,$CB,$63,$27,$43,$F3,$71,$D6
	.dc.b	$B9,$57,$3F,$A9,$97,$C7,$E0,$2E,$1C,$66,$EB,$FE,$D9,$C7,$CD,$58
	.dc.b	$BF,$F0,$07,$D3,$BD,$5F,$E7,$AD,$8F,$F5,$8F,$2B,$EA,$7F,$DE,$71
	.dc.b	$FE,$A8,$9D,$67,$FE,$BA,$6B,$FD,$FD,$AF,$6B,$FE,$B3,$18,$8F,$90
	.dc.b	$F2,$CB,$FD,$DD,$87,$7F,$23,$C7,$7F,$D3,$B1,$FE,$63,$8F,$EF,$7D
	.dc.b	$2F,$F7,$EB,$F2,$DF,$C7,$E5,$F5,$1C,$B7,$9F,$EA,$46,$32,$F8,$8F
	.dc.b	$06,$D6,$A7,$7F,$53,$23,$B7,$C6,$5C,$3B,$4D,$BE,$73,$6B,$E6,$FC
	.dc.b	$DF,$CC,$3F,$15,$F0,$2F,$97,$C4,$FC,$B9,$F8,$A4,$7F,$C0,$3C,$08
	.dc.b	$C6,$5F,$DE,$FE,$A7,$E9,$BF,$46,$3F,$BC,$E4,$9C,$BE,$EF,$90,$F8
	.dc.b	$27,$C2,$FA,$9B,$8F,$C9,$8B,$87,$C9,$9B,$5F,$1B,$E6,$7E,$4F,$AB
	.dc.b	$F2,$9B,$87,$D1,$6D,$FD,$16,$D2,$7F,$C9,$DF,$52,$BF,$F0,$5B,$63
	.dc.b	$FD,$FF,$DA,$DF,$F9,$3D,$F1,$C7,$F5,$D7,$B5,$19,$51,$97,$FC,$A0
	.dc.b	$D7,$C3,$87,$F8,$AA,$31,$1D,$23,$FC,$37,$F5,$2F,$98,$F2,$C6,$1F
	.dc.b	$C4,$FD,$47,$E2,$BE,$A6,$E9,$F9,$16,$E1,$F9,$16,$E5,$F8,$E7,$F2
	.dc.b	$3F,$23,$F9,$2E,$BB,$7F,$99,$6D,$FE,$65,$B7,$F9,$67,$FC,$63,$F5
	.dc.b	$1F,$D0,$89,$E8,$7F,$91,$1B,$7F,$E8,$EC,$FF,$CF,$90,$F8,$FF,$95
	.dc.b	$F0,$1E,$DC,$BF,$8C,$4F,$F5,$48,$C4,$23,$85,$F3,$8D,$D7,$2B,$EA
	.dc.b	$D8,$E7,$85,$79,$45,$BE,$71,$FE,$67,$6D,$FF,$A1,$5B,$1D,$23,$FC
	.dc.b	$12,$FA,$BE,$63,$FC,$60,$D8,$CB,$EA,$8C,$B6,$32,$FA,$E3,$C9,$F5
	.dc.b	$0E,$CF,$F2,$2C,$FF,$AC,$B2,$E8,$D3,$F9,$E4,$EF,$C2,$F8,$50,$8C
	.dc.b	$3B,$31,$5D,$34,$F9,$79,$31,$F0,$CE,$71,$FF,$58,$61,$CA,$FE,$8B
	.dc.b	$63,$C7,$18,$9E,$18,$79,$1A,$F8,$31,$F2,$E3,$DF,$2F,$3E,$3C,$EB
	.dc.b	$85,$62,$F8,$DE,$E7,$D5,$5B,$7A,$18,$9D,$E5,$EA,$BE,$44,$9F,$85
	.dc.b	$7C,$F1,$F4,$4E,$5B,$3A,$AE,$4F,$59,$D6,$5F,$00,$F8,$B8,$7B,$B1
	.dc.b	$F2,$1E,$5C,$7B,$4E,$AB,$8D,$FB,$8E,$D9,$61,$F5,$E5,$E2,$3E,$AE
	.dc.b	$5E,$27,$17,$CE,$78,$9E,$52,$46,$9F,$C7,$18,$8E,$11,$98,$FC,$85
	.dc.b	$78,$71,$F4,$E5,$F2,$A3,$0F,$B4,$C4,$F3,$3D,$55,$CF,$2F,$7C,$63
	.dc.b	$0E,$E4,$7C,$73,$E1,$C3,$F2,$B1,$FE,$F1,$6B,$ED,$31,$7F,$A2,$75
	.dc.b	$F4,$CF,$7F,$FD,$A3,$48,$C4,$FF,$BD,$E3,$13,$A9,$CC,$F2,$8F,$C8
	.dc.b	$C7,$0B,$FC,$EB,$70,$FF,$14,$53,$EA,$1F,$51,$FE,$F0,$9F,$D2,$D6
	.dc.b	$EB,$F6,$6D,$7F,$64,$DA,$E3,$1E,$0C,$23,$F9,$C9,$D5,$F5,$8E,$73
	.dc.b	$AC,$3C,$AF,$EF,$AF,$9C,$F1,$8D,$61,$E4,$75,$7C,$E7,$C7,$1A,$7A
	.dc.b	$9F,$C1,$6B,$9D,$F6,$8E,$84,$7F,$98,$C6,$AF,$B4,$E5,$B3,$AB,$F0
	.dc.b	$63,$F4,$AB,$47,$82,$74,$F8,$CF,$29,$A3,$E8,$4F,$2C,$7E,$65,$73
	.dc.b	$AF,$31,$CA,$B8,$46,$6F,$C6,$7D,$AC,$3D,$13,$E1,$8C,$B4,$FF,$33
	.dc.b	$46,$9F,$1D,$76,$9D,$C7,$90,$F4,$57,$77,$91,$93,$C7,$7E,$83,$EE
	.dc.b	$BC,$1C,$63,$F5,$A3,$0F,$C3,$C3,$C4,$78,$EB,$B4,$F9,$2F,$D9,$8F
	.dc.b	$DD,$AF,$C1,$E9,$F9,$8A,$F1,$36,$33,$1C,$0F,$EA,$2B,$E0,$BF,$DC
	.dc.b	$B1,$F1,$DA,$FC,$46,$BF,$A3,$6C,$71,$8F,$CA,$47,$8C,$FC,$87,$D4
	.dc.b	$3E,$0A,$99,$9D,$9F,$D6,$23,$4D,$8F,$53,$FB,$28,$CB,$86,$BE,$F6
	.dc.b	$BF,$B8,$F7,$7F,$CA,$8F,$FA,$87,$9B,$FC,$65,$0F,$A8,$FD,$EC,$61
	.dc.b	$B9,$71,$F1,$2D,$FF,$AA,$6F,$9B,$FE,$3C,$9E,$D5,$A8,$4F,$57,$C1
	.dc.b	$3C,$DA,$73,$F6,$38,$4B,$E8,$29,$88,$93,$C0,$EB,$1A,$71,$3F,$EE
	.dc.b	$27,$5F,$50,$FA,$8E,$BE,$7C,$9C,$C7,$9E,$75,$1F,$38,$38,$7E,$8B
	.dc.b	$5F,$C6,$35,$FE,$4A,$EF,$4F,$F2,$F3,$72,$CD,$7B,$B6,$F8,$DF,$51
	.dc.b	$FE,$0F,$C3,$D8,$BD,$FC,$9A,$EB,$93,$A8,$CB,$C2,$39,$FC,$8C,$8E
	.dc.b	$6F,$6F,$85,$C4,$7D,$73,$FD,$C5,$69,$CE,$9F,$E7,$58,$57,$DB,$FA
	.dc.b	$87,$97,$E8,$F4,$26,$3F,$4E,$72,$D9,$F9,$4A,$32,$FE,$69,$AF,$ED
	.dc.b	$5A,$FF,$4E,$77,$87,$FA,$1D,$BB,$66,$BE,$4F,$67,$E3,$FE,$A1,$F5
	.dc.b	$14,$98,$FE,$1E,$1F,$31,$88,$F8,$42,$38,$BB,$9E,$93,$D7,$E7,$3B
	.dc.b	$7D,$1C,$AE,$91,$97,$B4,$65,$EB,$69,$FD,$27,$89,$FE,$78,$D7,$FD
	.dc.b	$45,$F4,$EF,$A8,$FD,$43,$70,$FC,$F5,$6F,$6F,$EE,$2F,$FA,$02,$3F
	.dc.b	$8E,$35,$CF,$87,$F4,$23,$10,$3F,$CD,$51,$E2,$23,$FA,$BA,$FD,$DE
	.dc.b	$1E,$4E,$DF,$19,$27,$5D,$B5,$F9,$35,$FA,$F6,$CF,$AA,$7D,$7D,$3E
	.dc.b	$CF,$8F,$95,$DE,$32,$F8,$63,$2F,$48,$F8,$FD,$5F,$C2,$FD,$67,$D4
	.dc.b	$53,$35,$F1,$A7,$E0,$11,$FD,$D6,$BF,$AB,$6C,$70,$7F,$03,$19,$7F
	.dc.b	$07,$5F,$B3,$7F,$0B,$F4,$AE,$3F,$CD,$B6,$3F,$83,$35,$FE,$6A,$D7
	.dc.b	$FA,$82,$F4,$38,$92,$38,$BF,$A6,$FA,$4F,$D7,$C7,$F1,$16,$BF,$D0
	.dc.b	$1A,$FF,$5B,$6B,$FE,$34,$5F,$0D,$CC,$91,$CD,$FD,$9F,$D2,$7F,$44
	.dc.b	$8F,$C2,$35,$34,$9A,$FF,$3B,$7F,$A7,$F1,$7F,$BE,$5B,$C9,$FE,$44
	.dc.b	$FA,$47,$C7,$7F,$41,$F5,$3F,$5F,$8F,$F3,$79,$FF,$6E,$37,$D4,$D7
	.dc.b	$64,$8F,$E6,$15,$C1,$FE,$BE,$D7,$FB,$D3,$5F,$F2,$9B,$BC,$98,$E2
	.dc.b	$BA,$57,$EF,$B0,$FF,$2F,$D6,$A3,$F4,$9E,$DF,$A1,$FE,$99,$F3,$EF
	.dc.b	$6B,$FA,$D3,$63,$FA,$84,$ED,$FE,$12,$D7,$FC,$1C,$D7,$FC,$CE,$D7
	.dc.b	$BE,$9F,$C9,$23,$95,$F8,$E5,$18,$CB,$8D,$78,$BE,$0B,$FB,$67,$7B
	.dc.b	$E9,$3F,$B7,$7D,$13,$E1,$E9,$A8,$99,$7F,$5B,$AE,$0D,$8F,$E6,$B1
	.dc.b	$88,$FF,$06,$B7,$3F,$23,$72,$FF,$40,$B6,$7F,$86,$60,$F8,$AF,$F0
	.dc.b	$07,$D2,$3E,$77,$CB,$F4,$92,$FF,$BC,$4E,$5B,$3F,$D1,$A3,$11,$FE
	.dc.b	$1E,$6B,$FE,$39,$6B,$FE,$92,$77,$97,$C9,$6E,$59,$AF,$8B,$E1,$FF
	.dc.b	$84,$FE,$93,$F0,$D3,$FE,$9A,$CF,$FC,$2B,$8F,$F8,$2D,$AF,$F4,$49
	.dc.b	$DB,$FC,$A5,$FC,$83,$79,$7F,$AE,$C6,$20,$D3,$8B,$88,$F1,$1F,$63
	.dc.b	$E6,$FC,$9F,$05,$F6,$BD,$69,$5F,$E1,$DB,$4E,$A3,$FC,$29,$18,$6C
	.dc.b	$78,$8F,$ED,$FF,$48,$E7,$A4,$9F,$F0,$E4,$65,$B1,$FE,$37,$8C,$47
	.dc.b	$F6,$46,$EF,$F9,$5F,$A4,$FF,$03,$3F,$D2,$FE,$8D,$F4,$7F,$EC,$49
	.dc.b	$CD,$A7,$2F,$F5,$1B,$EB,$18,$84,$62,$B9,$7A,$EC,$3B,$46,$1B,$5C
	.dc.b	$F1,$FF,$69,$E1,$CF,$0E,$53,$FD,$7F,$1F,$1C,$F7,$6E,$1F,$ED,$AA
	.dc.b	$DD,$78,$AF,$9C,$F7,$8F,$53,$6F,$B4,$EA,$7C,$47,$90,$5F,$67,$C5
	.dc.b	$8F,$AD,$7D,$27,$27,$FB,$A3,$5C,$E1,$E1,$8F,$44,$F1,$C3,$44,$9F
	.dc.b	$21,$2B,$C6,$7A,$0B,$1C,$B2,$F3,$61,$F5,$A7,$FD,$48,$D7,$E0,$18
	.dc.b	$BF,$43,$AF,$5D,$F2,$3F,$E6,$5A,$FF,$22,$B6,$33,$3F,$E9,$69,$CB
	.dc.b	$FE,$42,$F8,$6F,$F2,$1A,$69,$1D,$1E,$91,$F1,$AF,$A2,$FE,$71,$9F
	.dc.b	$E2,$67,$FC,$39,$5F,$46,$B5,$CD,$F4,$E7,$27,$38,$CB,$D4,$CD,$7E
	.dc.b	$4E,$32,$F5,$8F,$31,$C6,$FC,$3A,$FF,$AC,$B0,$F0,$46,$27,$FD,$9D
	.dc.b	$18,$C3,$C5,$1C,$AF,$B5,$6E,$3F,$C6,$86,$6E,$E1,$E2,$7B,$47,$86
	.dc.b	$F9,$E1,$D6,$3D,$47,$13,$FC,$7D,$5C,$5F,$19,$C4,$DC,$71,$C3,$C7
	.dc.b	$5F,$EA,$BA,$F8,$BA,$FF,$53,$BF,$8B,$87,$93,$1F,$ED,$13,$FD,$5E
	.dc.b	$B7,$79,$3F,$D6,$97,$C0,$F1,$D7,$08,$F6,$F4,$59,$F3,$B7,$E2,$9F
	.dc.b	$47,$FE,$A1,$AF,$24,$EE,$BF,$06,$DC,$3F,$A7,$4F,$FB,$2A,$F8,$FA
	.dc.b	$39,$AD,$77,$4A,$D7,$BF,$FD,$F5,$F8,$11,$9C,$7F,$D0,$78,$FF,$5B
	.dc.b	$9D,$F8,$D9,$D7,$E0,$5F,$9F,$47,$C3,$66,$FF,$CD,$DA,$FA,$A3,$D1
	.dc.b	$EE,$7F,$3A,$9F,$F6,$C3,$73,$FF,$59,$43,$E8,$57,$C3,$AE,$4F,$07
	.dc.b	$E4,$3B,$8C,$D7,$FA,$6A,$FA,$67,$FE,$1D,$AE,$9C,$3A,$8C,$63,$F6
	.dc.b	$2F,$FD,$09,$F4,$6F,$63,$E7,$37,$0F,$F7,$A9,$E5,$F4,$33,$32,$3F
	.dc.b	$DE,$53,$BA,$F5,$DA,$F5,$1F,$EA,$E9,$D6,$1E,$2C,$32,$DC,$3F,$DF
	.dc.b	$87,$FB,$40,$BD,$3F,$03,$17,$83,$F8,$39,$D6,$3F,$EB,$A6,$DF,$F8
	.dc.b	$AD,$A7,$FA,$84,$F9,$D3,$C6,$E4,$7B,$BE,$8F,$FC,$B2,$F8,$0C,$79
	.dc.b	$9F,$91,$F9,$8F,$C2,$63,$FE,$C8,$6E,$3F,$32,$F5,$C9,$30,$FF,$58
	.dc.b	$4F,$F8,$F6,$B7,$1E,$3C,$D7,$BC,$3A,$61,$EF,$F8,$7C,$63,$F6,$1F
	.dc.b	$D1,$7D,$1B,$E8,$FF,$49,$F4,$6F,$53,$31,$86,$69,$E2,$3F,$87,$CE
	.dc.b	$DC,$65,$FE,$C5,$BE,$91,$88,$4E,$DF,$F7,$3B,$FB,$56,$F7,$76,$6C
	.dc.b	$F1,$9F,$97,$1C,$37,$FD,$D7,$E6,$7D,$A3,$15,$DF,$A1,$E6,$FE,$FD
	.dc.b	$B1,$CE,$F1,$3F,$C0,$FE,$91,$EA,$53,$87,$A9,$6F,$D5,$F4,$3F,$EA
	.dc.b	$46,$FB,$5F,$C4,$7E,$91,$F4,$8F,$A4,$FE,$1D,$7B,$DB,$FC,$25,$87
	.dc.b	$F3,$68,$C7,$37,$F9,$DD,$B1,$FE,$ED,$C3,$9D,$F2,$C6,$4F,$80,$F9
	.dc.b	$F3,$A7,$18,$76,$1D,$5F,$EF,$7F,$A3,$FE,$4C,$FF,$BE,$7A,$6C,$A6
	.dc.b	$C6,$1B,$5F,$72,$B9,$36,$BF,$22,$76,$1F,$47,$D0,$EB,$B5,$9E,$11
	.dc.b	$67,$CD,$47,$FD,$21,$3A,$3F,$11,$F4,$2F,$5B,$F9,$C3,$63,$F9,$B7
	.dc.b	$D1,$FF,$B8,$DD,$CE,$88,$EB,$E0,$ED,$FC,$51,$7C,$DC,$37,$87,$FD
	.dc.b	$79,$CE,$C7,$B5,$6B,$1F,$97,$E9,$FF,$B1,$B2,$F8,$CF,$E2,$9A,$6B
	.dc.b	$97,$6D,$32,$70,$79,$FD,$1C,$8E,$7F,$11,$70,$F8,$F6,$AC,$67,$FE
	.dc.b	$D1,$DB,$FC,$57,$78,$C3,$FC,$07,$AB,$E5,$FE,$49,$FD,$9F,$D0,$FF
	.dc.b	$62,$FC,$9E,$2B,$E7,$CD,$D3,$F1,$55,$A9,$C9,$23,$E3,$CF,$3F,$83
	.dc.b	$5F,$1C,$FC,$9A,$BE,$2B,$E3,$9F,$23,$E8,$AE,$5F,$22,$2E,$5E,$E9
	.dc.b	$B7,$CB,$E4,$FE,$7F,$AB,$F4,$1B,$87,$D0,$6D,$FC,$66,$D7,$FB,$8F
	.dc.b	$E8,$BF,$BB,$4F,$E0,$5B,$1F,$EF,$9D,$BF,$DC,$FE,$47,$F8,$16,$73
	.dc.b	$87,$F8,$B1,$B7,$F9,$76,$D7,$09,$FE,$3D,$1A,$93,$4D,$ED,$FC,$87
	.dc.b	$0F,$26,$5F,$8A,$FA,$1F,$C5,$FD,$0D,$C7,$F1,$6D,$C7,$F0,$2D,$BF
	.dc.b	$9A,$FE,$4B,$AB,$F2,$7E,$B5,$BF,$C9,$B6,$FF,$12,$DB,$FA,$AF,$F8
	.dc.b	$8F,$E8,$7C,$95,$FE,$4A,$6D,$7E,$D5,$7C,$56,$F7,$95,$FC,$F6,$77
	.dc.b	$47,$71,$99,$ED,$F1,$88,$FF,$37,$FC,$A7,$78,$C5,$FF,$9E,$E3,$C5
	.dc.b	$58,$F8,$14,$6E,$F8,$D7,$96,$F5,$39,$8E,$47,$A6,$F2,$EE,$B8,$BE
	.dc.b	$7F,$7B,$2D,$8C,$E2,$FB,$7C,$AA,$A4,$9D,$CD,$AF,$08,$FA,$17,$6F
	.dc.b	$F8,$D2,$33,$18,$F0,$FB,$E1,$E2,$8C,$4E,$AB,$86,$BF,$0A,$3F,$94
	.dc.b	$4E,$A3,$11,$E9,$38,$5F,$BA,$75,$A7,$FA,$19,$B3,$F7,$70,$F0,$61
	.dc.b	$DA,$B4,$D9,$EE,$D7,$B6,$1E,$0B,$E7,$1E,$F9,$D4,$63,$3F,$91,$9F
	.dc.b	$8E,$72,$4C,$BE,$0C,$65,$97,$EA,$3E,$5D,$72,$71,$87,$86,$F9,$46
	.dc.b	$B2,$F5,$DF,$2A,$D3,$DA,$78,$E5,$EE,$8D,$3F,$A1,$AE,$33,$A3,$E0
	.dc.b	$CF,$1A,$E7,$19,$3E,$04,$66,$BA,$46,$72,$F8,$57,$D2,$F8,$92,$BC
	.dc.b	$19,$78,$AF,$B9,$D7,$2F,$81,$5B,$8C,$B5,$FA,$75,$E0,$8C,$61,$E0
	.dc.b	$BE,$95,$E6,$3A,$61,$E3,$BF,$3C,$7F,$B2,$8D,$16,$BF,$26,$DC,$BF
	.dc.b	$34,$E7,$E8,$9E,$FF,$18,$FC,$AC,$61,$C5,$23,$11,$C6,$78,$3F,$A7
	.dc.b	$8E,$19,$7E,$35,$B8,$7E,$65,$B3,$FE,$09,$FA,$17,$D1,$7F,$B8,$AF
	.dc.b	$86,$1F,$A1,$BD,$DF,$58,$FE,$03,$7C,$23,$51,$1F,$DA,$CF,$09,$D4
	.dc.b	$62,$75,$5D,$A3,$F7,$F3,$A8,$C1,$30,$F4,$57,$83,$0F,$34,$62,$FA
	.dc.b	$12,$BF,$77,$7C,$E7,$51,$CE,$7E,$6D,$F3,$7C,$73,$A2,$3F,$16,$FB
	.dc.b	$BC,$0E,$D1,$A5,$7D,$DD,$5F,$77,$E4,$1E,$EA,$E6,$E2,$35,$8F,$DD
	.dc.b	$C3,$C9,$3D,$8E,$78,$77,$C3,$B4,$71,$8F,$15,$71,$8E,$C7,$B0,$E2
	.dc.b	$43,$CF,$87,$8A,$38,$61,$E2,$C7,$EC,$4E,$5F,$C0,$CE,$AF,$C5,$39
	.dc.b	$C7,$B6,$1E,$33,$E4,$18,$AE,$0F,$87,$1F,$49,$2B,$E1,$B8,$FB,$EB
	.dc.b	$12,$7E,$1E,$5F,$72,$77,$F0,$72,$7F,$26,$FC,$CF,$A3,$FF,$70,$46
	.dc.b	$08,$7E,$3A,$35,$F0,$72,$3F,$D5,$D3,$C1,$FF,$2F,$37,$A3,$FB,$6C
	.dc.b	$23,$FD,$EC,$F0,$9F,$F3,$93,$C1,$B3,$C2,$79,$70,$E9,$E3,$5F,$E6
	.dc.b	$BF,$77,$FC,$9C,$FF,$AF,$D7,$D1,$F2,$AE,$B1,$89,$D7,$8F,$97,$FD
	.dc.b	$05,$1C,$9E,$95,$95,$1C,$5F,$04,$E6,$3F,$D7,$F1,$E1,$F9,$BC,$19
	.dc.b	$F6,$64,$64,$87,$A5,$ED,$F4,$6F,$A3,$A4,$9F,$66,$B5,$FF,$45,$4E
	.dc.b	$DF,$F6,$F3,$5F,$89,$1B,$D7,$E8,$CE,$5B,$5E,$99,$F7,$7D,$1F,$F8
	.dc.b	$36,$FF,$DD,$0B,$E0,$32,$6B,$AA,$4E,$A3,$2F,$08,$E7,$F1,$F2,$39
	.dc.b	$BD,$BE,$1B,$11,$FE,$A3,$B4,$E3,$2F,$C3,$C2,$BE,$97,$D1,$BC,$CA
	.dc.b	$4C,$3C,$FD,$75,$FF,$57,$4E,$DF,$F5,$F6,$1F,$92,$8C,$47,$86,$78
	.dc.b	$4E,$6A,$F2,$7E,$53,$1F,$E8,$F3,$F9,$AF,$A2,$7D,$1F,$E1,$5B,$8F
	.dc.b	$6B,$FC,$DC,$62,$3F,$D8,$0B,$E5,$F8,$54,$9E,$07,$FB,$4E,$78,$5F
	.dc.b	$7A,$ED,$F3,$7D,$FE,$82,$5F,$01,$69,$DC,$66,$13,$C2,$3C,$73,$F9
	.dc.b	$DF,$0F,$FC,$73,$87,$ED,$9B,$97,$ED,$1B,$5F,$EE,$7F,$A2,$FD,$35
	.dc.b	$6F,$C3,$E4,$49,$F2,$4E,$AF,$94,$E6,$7F,$DC,$57,$D1,$56,$DE,$DD
	.dc.b	$1F,$B1,$8F,$D6,$C7,$FB,$A5,$B8,$7F,$0C,$5D,$3B,$A9,$5E,$27,$94
	.dc.b	$73,$E7,$E9,$F1,$F2,$BC,$F1,$97,$D3,$19,$7C,$11,$F1,$FA,$24,$7D
	.dc.b	$1C,$BF,$4C,$FD,$3F,$57,$F5,$4F,$BA,$76,$D3,$5D,$38,$D3,$E8,$4D
	.dc.b	$DB,$F9,$1D,$F2,$9F,$E5,$3F,$43,$F6,$B0,$FC,$03,$54,$6B,$87,$F8
	.dc.b	$C4,$BE,$A6,$FA,$12,$F8,$BF,$C7,$7E,$8B,$F5,$D1,$F8,$B6,$BD,$32
	.dc.b	$FE,$6B,$18,$C7,$F9,$DB,$63,$F9,$CF,$D4,$FD,$9A,$4B,$E1,$FB,$12
	.dc.b	$3F,$31,$F4,$7F,$CF,$10,$70,$DE,$1F,$E6,$51,$8F,$A3,$7C,$77,$F3
	.dc.b	$56,$D7,$4A,$D9,$99,$FE,$75,$E9,$FE,$C6,$7F,$51,$4F,$53,$91,$AE
	.dc.b	$D9,$3F,$AD,$AE,$0F,$C3,$3C,$EE,$36,$FE,$BB,$7F,$C1,$B6,$FE,$B7
	.dc.b	$39,$6D,$F4,$BE,$42,$B8,$3F,$0B,$DA,$F5,$BA,$F9,$E7,$B7,$FD,$4D
	.dc.b	$B1,$FC,$06,$77,$97,$F6,$E6,$E3,$FD,$B1,$B3,$D7,$4E,$5F,$09,$FD
	.dc.b	$76,$73,$26,$98,$C1,$3F,$D8,$FC,$0F,$F7,$9D,$7F,$11,$FA,$17,$C4
	.dc.b	$FF,$67,$6C,$7F,$15,$9D,$BF,$01,$B9,$FF,$81,$1B,$9F,$F8,$09,$B3
	.dc.b	$F8,$4A,$7C,$5F,$87,$DC,$93,$D5,$F0,$47,$F7,$7F,$A2,$7C,$E7,$8F
	.dc.b	$E8,$E5,$FF,$27,$9C,$B6,$78,$11,$F6,$78,$B5,$FA,$F3,$C1,$B3,$EF
	.dc.b	$9F,$F0,$37,$D1,$FD,$F9,$F4,$7A,$7F,$0C,$CF,$34,$8F,$E7,$13,$B4
	.dc.b	$DD,$1F,$37,$B9,$FC,$3C,$EC,$D4,$78,$3E,$0A,$4F,$47,$E9,$75,$E2
	.dc.b	$38,$C5,$E2,$FF,$19,$E1,$FA,$EF,$A3,$73,$7F,$87,$DB,$7F,$D3,$5A
	.dc.b	$FE,$63,$E8,$BF,$BB,$DF,$EE,$3E,$89,$F4,$5F,$E3,$AB,$DC,$FF,$90
	.dc.b	$1B,$CF,$9C,$7F,$C5,$98,$72,$C5,$8F,$27,$F8,$1C,$66,$31,$0B,$F1
	.dc.b	$C6,$1B,$8F,$F9,$2A,$B8,$37,$D7,$FE,$4B,$7D,$A3,$C4,$FA,$5A,$E4
	.dc.b	$F0,$C7,$0B,$F0,$D7,$49,$FE,$2F,$7E,$06,$E1,$FE,$5A,$AD,$DF,$1C
	.dc.b	$3C,$59,$77,$BD,$94,$F6,$5F,$19,$DC,$78,$C6,$3E,$17,$67,$1C,$7C
	.dc.b	$4E,$B0,$F0,$3E,$AC,$35,$87,$58,$F0,$9D,$6F,$A1,$D6,$F9,$9D,$1C
	.dc.b	$DF,$CA,$9F,$F7,$B4,$7D,$96,$C6,$5F,$49,$AF,$5D,$F1,$9F,$EE,$EA
	.dc.b	$FF,$1E,$36,$33,$3F,$EF,$99,$CB,$FE,$F4,$1F,$01,$5A,$65,$C1,$1F
	.dc.b	$91,$F1,$6F,$A1,$6F,$F3,$E7,$FA,$7D,$7D,$1A,$D4,$E6,$7A,$EF,$F9
	.dc.b	$09,$CC,$7D,$33,$89,$DA,$FA,$B9,$38,$EF,$FE,$9C,$9E,$35,$F9,$A9
	.dc.b	$E2,$F5,$AE,$2E,$63,$89,$FD,$92,$78,$D7,$FA,$9A,$38,$BD,$6B,$8B
	.dc.b	$98,$E2,$27,$8B,$C7,$0F,$81,$5C,$23,$07,$AF,$0F,$85,$3A,$6C,$64
	.dc.b	$F7,$46,$5A,$67,$0F,$54,$65,$B1,$9C,$3C,$D3,$97,$FC,$03,$F6,$5F
	.dc.b	$5B,$CE,$F9,$7C,$3F,$32,$31,$8F,$F9,$9B,$EC,$9F,$64,$F7,$77,$3C
	.dc.b	$8F,$F2,$F9,$FE,$6C,$AE,$39,$FF,$79,$6E,$1F,$63,$0E,$77,$CA,$95
	.dc.b	$AB,$F2,$FC,$FB,$D3,$D4,$A3,$13,$E5,$CB,$FC,$2A,$F8,$63,$37,$E1
	.dc.b	$6E,$5E,$D6,$CF,$F8,$8B,$2F,$6F,$91,$DA,$73,$7E,$46,$E3,$E1,$CB
	.dc.b	$D1,$7C,$30,$F0,$47,$9B,$E7,$BF,$D9,$91,$F0,$DB,$87,$F8,$6E,$9F
	.dc.b	$07,$8C,$FE,$85,$5B,$FA,$AD,$BF,$8A,$DB,$E3,$F0,$5D,$C9,$5D,$DF
	.dc.b	$2E,$1C,$72,$5D,$FC,$0F,$FC,$1B,$7F,$5D,$B8,$F9,$7E,$D1,$F5,$FF
	.dc.b	$42,$73,$A7,$E1,$5B,$7F,$8A,$6D,$B4,$61,$CA,$BB,$79,$BC,$C8,$FE
	.dc.b	$36,$35,$5F,$87,$BE,$79,$BF,$03,$FE,$BF,$7F,$8D,$6D,$7C,$BF,$B6
	.dc.b	$FF,$30,$4E,$79,$BF,$0F,$5A,$D3,$F3,$8D,$9F,$C0,$57,$0B,$FC,$B4
	.dc.b	$E5,$FC,$9B,$70,$F1,$61,$F8,$0A,$E1,$87,$48,$FC,$AF,$E6,$1F,$E6
	.dc.b	$CC,$7F,$44,$DC,$7F,$01,$F7,$2F,$A9,$DC,$4A,$DB,$F8,$C9,$D4,$60
	.dc.b	$91,$FA,$BA,$E9,$87,$EA,$EB,$73,$93,$97,$D7,$FE,$A3,$EE,$9F,$4D
	.dc.b	$8C,$7F,$6D,$18,$6C,$6C,$D5,$7E,$E6,$74,$D8,$FD,$9F,$7B,$D3,$E1
	.dc.b	$7F,$C0,$63,$0F,$F0,$3F,$BC,$79,$7F,$BA,$6C,$67,$E1,$B1,$39,$CE
	.dc.b	$4F,$18,$B9,$7F,$0B,$FC,$0D,$F2,$BF,$C1,$1B,$5D,$EF,$13,$FC,$37
	.dc.b	$E8,$BF,$D3,$FC,$77,$0F,$BD,$1F,$EA,$2F,$43,$13,$30,$FF,$52,$F1
	.dc.b	$7F,$4E,$AE,$31,$83,$FD,$52,$F1,$71,$68,$DB,$FE,$99,$6C,$F3,$8F
	.dc.b	$09,$FE,$50,$9E,$33,$A7,$F9,$AF,$CF,$FF,$BB,$9F,$E5,$8D,$9F,$E1
	.dc.b	$5F,$7E,$F3,$BF,$92,$CE,$70,$FE,$74,$DB,$E7,$F1,$9F,$C6,$2F,$84
	.dc.b	$74,$8F,$E5,$BF,$97,$BE,$7A,$9B,$F9,$9B,$5F,$CF,$67,$4D,$9D,$34
	.dc.b	$F8,$79,$BE,$AD,$7F,$CF,$E3,$0F,$F4,$1A,$7D,$43,$E8,$F5,$7C,$30
	.dc.b	$E6,$DC,$3F,$D8,$97,$C3,$E8,$F7,$8F,$3C,$51,$8C,$3E,$AD,$6D,$E7
	.dc.b	$5B,$9F,$E7,$7F,$80,$BA,$3F,$A7,$FD,$0F,$EB,$AF,$FA,$34,$E5,$FE
	.dc.b	$BD,$19,$CB,$F9,$04,$E5,$FE,$C1,$19,$F2,$13,$9E,$2F,$ED,$D3,$98
	.dc.b	$7D,$5F,$F5,$EF,$A0,$7C,$CF,$92,$30,$DC,$BF,$B7,$4E,$9F,$EE,$D5
	.dc.b	$C9,$CC,$FE,$87,$DB,$7D,$4D,$31,$F3,$36,$BF,$23,$3A,$6C,$E9,$A4
	.dc.b	$D7,$F3,$93,$B6,$D7,$37,$FC,$01,$8B,$ED,$3D,$87,$CE,$F8,$9B,$87
	.dc.b	$AA,$2D,$7F,$65,$CD,$58,$DB,$FC,$7F,$E9,$3F,$11,$7D,$1F,$F8,$26
	.dc.b	$31,$E1,$D9,$CE,$7F,$B4,$9C,$B6,$33,$93,$0E,$B0,$BE,$58,$73,$7E
	.dc.b	$57,$CC,$CC,$3E,$3F,$CF,$DC,$7E,$1C,$DC,$7F,$84,$70,$2B,$17,$FC
	.dc.b	$22,$15,$8B,$FE,$11,$1B,$9F,$A7,$3F,$94,$EE,$FF,$10,$FD,$2F,$E7
	.dc.b	$27,$EF,$C6,$3A,$75,$1F,$E3,$19,$D6,$1F,$2E,$B5,$3F,$86,$F8,$CE
	.dc.b	$7E,$C7,$E8,$1B,$19,$39,$46,$5A,$75,$F8,$BF,$19,$FA,$8F,$8D,$B1
	.dc.b	$92,$61,$F6,$E3,$2F,$F9,$0F,$1F,$83,$3E,$7C,$F3,$B7,$F9,$46,$73
	.dc.b	$5B,$9F,$82,$6A,$BF,$CB,$33,$AA,$FF,$41,$D3,$E9,$DF,$3F,$FE,$86
	.dc.b	$F8,$5F,$D5,$C8,$D4,$67,$B3,$11,$C7,$EA,$30,$EE,$F9,$E5,$FE,$9A
	.dc.b	$AD,$C6,$27,$FD,$4F,$18,$BE,$31,$CE,$B8,$3C,$E3,$3D,$EE,$DF,$E6
	.dc.b	$13,$C8,$F3,$46,$7C,$3E,$31,$E8,$BB,$39,$8E,$01,$C6,$7F,$02,$31
	.dc.b	$7D,$A3,$19,$7A,$5D,$57,$0C,$BE,$1B,$6F,$C7,$7C,$67,$DD,$87,$8A
	.dc.b	$31,$3A,$C7,$07,$A3,$1E,$D7,$D2,$B7,$5E,$4A,$DB,$EA,$AD,$BC,$CF
	.dc.b	$6E,$1E,$37,$81,$BC,$BE,$5D,$F4,$9D,$1B,$AE,$75,$C3,$2F,$7D,$6C
	.dc.b	$F1,$CE,$AB,$8C,$67,$0F,$56,$1D,$9F,$3D,$F2,$73,$96,$32,$F7,$45
	.dc.b	$CB,$F0,$13,$5F,$14,$E9,$F5,$46,$71,$F0,$D7,$49,$F8,$38,$F8,$DB
	.dc.b	$19,$C3,$C5,$18,$23,$F7,$B1,$F4,$46,$31,$F0,$4E,$71,$F0,$61,$D6
	.dc.b	$31,$5B,$3E,$93,$D8,$C9,$6B,$D5,$1F,$EE,$7C,$7F,$28,$DA,$FC,$EB
	.dc.b	$71,$FF,$74,$3F,$97,$FA,$8F,$F6,$63,$FE,$EB,$A4,$62,$3A,$35,$FC
	.dc.b	$AB,$5F,$D3,$BC,$B0,$FC,$DB,$70,$DF,$C3,$7E,$AB,$7F,$F2,$4D,$FE
	.dc.b	$AA,$B4,$F2,$9C,$1B,$F4,$3F,$82,$E1,$D2,$33,$3F,$94,$AE,$2A,$FA
	.dc.b	$B9,$9D,$D7,$18,$FD,$BC,$EA,$30,$4C,$3C,$F3,$AC,$3C,$31,$89,$D4
	.dc.b	$F9,$CD,$1D,$A3,$44,$7F,$90,$57,$78,$C4,$7D,$1B,$E0,$EA,$FB,$BD
	.dc.b	$6F,$BB,$B3,$B9,$ED,$23,$AB,$ED,$1C,$6F,$BE,$3F,$73,$0F,$99,$3E
	.dc.b	$03,$A5,$F1,$C3,$C3,$18,$BE,$38,$FC,$EC,$7A,$CE,$4E,$C4,$3E,$95
	.dc.b	$72,$C7,$C1,$1E,$6B,$F8,$35,$CA,$73,$87,$B9,$B3,$FD,$12,$75,$87
	.dc.b	$DE,$9C,$DF,$79,$CE,$3E,$B9,$CC,$F8,$8F,$79,$F2,$30,$FE,$69,$5E
	.dc.b	$A1,$F5,$1F,$01,$C6,$54,$C4,$7D,$4F,$77,$F9,$F7,$8B,$FD,$63,$8F
	.dc.b	$FC,$96,$DE,$0F,$EB,$D3,$AC,$BF,$CA,$FE,$0E,$54,$65,$FF,$2B,$D6
	.dc.b	$63,$17,$FE,$56,$AC,$69,$FE,$34,$BE,$64,$3F,$C7,$1F,$25,$A9,$91
	.dc.b	$9A,$63,$FE,$6A,$AF,$17,$8F,$FE,$64,$F7,$7A,$F6,$4C,$3F,$CB,$71
	.dc.b	$AF,$17,$FD,$D1,$1F,$E7,$09,$D3,$FE,$47,$9D,$7A,$9E,$6C,$24,$FF
	.dc.b	$99,$23,$2F,$F9,$E2,$32,$FF,$93,$E3,$E2,$2F,$A5,$7C,$46,$1B,$1F
	.dc.b	$E9,$C6,$BF,$E4,$9C,$FF,$C1,$37,$FE,$0A,$8C,$1F,$E0,$F9,$DE,$9F
	.dc.b	$E1,$66,$D7,$F8,$41,$B3,$FE,$15,$AE,$55,$BB,$F1,$12,$7F,$D5,$B3
	.dc.b	$AB,$FF,$0E,$79,$1F,$E6,$5F,$5B,$EF,$3F,$E6,$7C,$AF,$92,$6C,$CC
	.dc.b	$3F,$CC,$91,$A3,$0F,$FA,$6A,$3F,$02,$47,$FD,$1E,$D7,$47,$FA,$91
	.dc.b	$FB,$07,$CC,$FA,$5F,$F3,$14,$7F,$B1,$1A,$FC,$60,$E3,$3F,$F1,$DA
	.dc.b	$F0,$F9,$79,$12,$B9,$3D,$63,$11,$CE,$79,$67,$FE,$3C,$6C,$7F,$90
	.dc.b	$5E,$8F,$FB,$E3,$89,$F7,$EF,$98,$F9,$3F,$81,$BF,$10,$A7,$E6,$27
	.dc.b	$57,$C5,$AF,$E0,$38,$15,$C6,$3F,$D2,$4D,$7F,$0B,$87,$83,$0F,$CF
	.dc.b	$78,$5F,$B1,$F1,$5F,$59,$4C,$DF,$FA,$FA,$70,$FF,$96,$F8,$3F,$5A
	.dc.b	$FE,$F2,$B8,$BB,$3F,$DE,$7E,$2B,$E8,$7F,$3F,$F5,$2E,$7A,$7E,$8A
	.dc.b	$75,$5C,$2B,$2F,$FB,$F2,$B8,$39,$21,$F6,$E7,$8B,$67,$BF,$37,$F0
	.dc.b	$DF,$53,$FC,$C1,$3E,$9E,$4F,$E2,$F1,$88,$FD,$4F,$81,$FE,$D8,$C3
	.dc.b	$F8,$3F,$B6,$FA,$4E,$51,$2B,$91,$FC,$72,$F9,$6D,$F1,$38,$98,$71
	.dc.b	$AD,$E2,$F6,$9F,$21,$FC,$85,$B3,$A6,$BF,$62,$BF,$DB,$8E,$67,$81
	.dc.b	$1F,$C6,$46,$9F,$CF,$47,$F8,$6F,$0F,$F4,$BB,$74,$FF,$48,$B6,$7F
	.dc.b	$D2,$CF,$19,$DF,$95,$FC,$5A,$35,$3E,$88,$C7,$AF,$96,$47,$C4,$34
	.dc.b	$59,$E9,$7F,$D1,$BF,$04,$7C,$A7,$F4,$16,$D7,$C6,$6B,$F5,$3E,$15
	.dc.b	$EB,$7F,$47,$6C,$7C,$C6,$BB,$8F,$E9,$9F,$0D,$FD,$7A,$3E,$8B,$5F
	.dc.b	$BE,$DD,$3F,$37,$BA,$31,$97,$F5,$C5,$EA,$7F,$57,$F8,$9F,$F2,$3E
	.dc.b	$1F,$71,$FD,$EB,$F8,$A6,$E1,$D1,$DC,$67,$D3,$E4,$47,$F8,$27,$C5
	.dc.b	$3E,$4B,$FA,$BB,$63,$B4,$ED,$FD,$DC,$65,$FC,$CF,$E0,$6F,$87,$FE
	.dc.b	$D8,$D8,$FC,$8B,$5F,$E0,$7F,$1D,$FD,$0B,$64,$E7,$3F,$86,$DD,$BF
	.dc.b	$BE,$B6,$DD,$0C,$39,$4E,$70,$E7,$8F,$88,$FE,$E3,$3D,$FE,$5B,$F4
	.dc.b	$98,$7E,$C1,$B9,$7D,$26,$E5,$FE,$E9,$CB,$B7,$B5,$FE,$07,$9C,$DF
	.dc.b	$F8,$31,$B9,$1C,$D7,$09,$FD,$8C,$68,$9D,$BE,$4B,$FF,$09,$74,$FE
	.dc.b	$2B,$2F,$C1,$B7,$2F,$F0,$EB,$70,$F8,$3F,$25,$FB,$6A,$DF,$2F,$22
	.dc.b	$4F,$F3,$F9,$D5,$F2,$9C,$CF,$F1,$0B,$E9,$9A,$F7,$3F,$E2,$8F,$21
	.dc.b	$F6,$3F,$E2,$76,$C7,$F0,$C6,$BF,$CF,$BE,$55,$EA,$EE,$A4,$ED,$FF
	.dc.b	$11,$CE,$A3,$04,$3F,$91,$D7,$4D,$3F,$8E,$4E,$E7,$F9,$3F,$BA,$8F
	.dc.b	$F1,$D3,$63,$F8,$FB,$5F,$EA,$3F,$2F,$FE,$96,$AF,$D7,$CE,$5B,$5C
	.dc.b	$8F,$F1,$BC,$EF,$6F,$F2,$6D,$6D,$B3,$BB,$FE,$D9,$5B,$7F,$C9,$FF
	.dc.b	$66,$FA,$3F,$F2,$73,$63,$F9,$9B,$5F,$EC,$BE,$83,$E3,$77,$52,$76
	.dc.b	$FF,$73,$8F,$F1,$CF,$CD,$FE,$BF,$45,$7E,$47,$D7,$AB,$E7,$5F,$3D
	.dc.b	$8E,$95,$6A,$73,$5C,$A7,$2F,$4C,$3B,$3D,$4C,$76,$31,$E1,$87,$3C
	.dc.b	$7B,$36,$B2,$70,$9A,$AF,$AE,$1E,$08,$F4,$5F,$27,$DB,$8F,$99,$F1
	.dc.b	$1C,$71,$F1,$38,$BE,$78,$F8,$63,$75,$C7,$0E,$D5,$A9,$F2,$1A,$C3
	.dc.b	$C0,$F8,$8F,$3C,$6D,$F0,$63,$EA,$BE,$98,$77,$C3,$AE,$3E,$46,$D6
	.dc.b	$F1,$EB,$7C,$30,$F0,$D6,$9F,$11,$B3,$C3,$7C,$F0,$FB,$B8,$F8,$EF
	.dc.b	$95,$79,$CF,$85,$5B,$7D,$A7,$F9,$DE,$7E,$DB,$63,$E4,$B7,$1F,$F3
	.dc.b	$E3,$F8,$1F,$9E,$7C,$5E,$23,$EC,$46,$1D,$DA,$31,$1D,$DA,$F4,$E3
	.dc.b	$E4,$FE,$61,$A9,$A4,$7E,$FF,$CF,$7F,$65,$E4,$FC,$4C,$71,$33,$E3
	.dc.b	$7F,$91,$36,$FC,$54,$65,$FC,$65,$71,$8F,$C4,$99,$DB,$F4,$F1,$99
	.dc.b	$E3,$3E,$33,$AD,$F8,$1C,$47,$4B,$F0,$5F,$B6,$76,$76,$36,$F6,$8D
	.dc.b	$C7,$FA,$A7,$0F,$D1,$E1,$E1,$AE,$57,$D7,$0F,$05,$6B,$0E,$38,$7E
	.dc.b	$FA,$31,$3A,$8F,$0D,$7A,$A7,$31,$E3,$6C,$66,$37,$7E,$03,$CD,$5C
	.dc.b	$E7,$C5,$3F,$C0,$7B,$FF,$DB,$BE,$4F,$F9,$CB,$6F,$F0,$CD,$FF,$8B
	.dc.b	$0F,$E4,$7E,$5E,$5F,$F7,$A9,$9B,$FF,$2F,$B6,$78,$D6,$E7,$FB,$EF
	.dc.b	$3A,$F1,$87,$83,$0F,$F2,$DF,$CF,$7F,$96,$9B,$E6,$71,$EF,$61,$FE
	.dc.b	$DC,$9C,$93,$1F,$F3,$7D,$F1,$7F,$A2,$FB,$FD,$A3,$1A,$7F,$8D,$1B
	.dc.b	$5F,$CB,$B2,$46,$2F,$CF,$7C,$72,$F3,$F6,$FF,$8A,$F0,$FE,$CD,$E4
	.dc.b	$F9,$7B,$B1,$87,$F5,$E8,$D9,$27,$FD,$6D,$3A,$CF,$FC,$8B,$76,$7C
	.dc.b	$4D,$BF,$EE,$D7,$D4,$E4,$FF,$A3,$3E,$75,$F3,$DF,$29,$B7,$FE,$05
	.dc.b	$F8,$EC,$4C,$8F,$F4,$3C,$EA,$BF,$D7,$3E,$22,$F5,$1D,$6E,$E5,$FE
	.dc.b	$91,$AD,$61,$CA,$7E,$CF,$CF,$7F,$50,$8F,$F4,$9B,$6F,$FC,$87,$E4
	.dc.b	$EB,$9E,$56,$9F,$B9,$E5,$7F,$A8,$E3,$04,$AF,$F5,$4D,$F3,$CB,$FD
	.dc.b	$45,$5B,$9C,$DF,$E2,$BE,$7B,$FB,$3E,$1F,$EA,$46,$CF,$F9,$73,$E1
	.dc.b	$B4,$C9,$D3,$F8,$DF,$39,$39,$8F,$F5,$CD,$72,$C7,$FC,$F6,$DB,$FF
	.dc.b	$25,$5F,$07,$99,$FE,$70,$F9,$2F,$CE,$FA,$2F,$9E,$ED,$18,$64,$FF
	.dc.b	$9F,$DC,$B4,$98,$FF,$A3,$3E,$7B,$6D,$8F,$F4,$B7,$CB,$FE,$AB,$D6
	.dc.b	$7C,$F5,$8C,$39,$C6,$57,$FA,$76,$2D,$FF,$B5,$BE,$79,$F2,$5F,$B4
	.dc.b	$6D,$7F,$A8,$EF,$13,$FB,$BF,$9E,$FF,$0A,$F1,$5C,$3E,$B5,$FF,$86
	.dc.b	$27,$FC,$33,$F0,$19,$A9,$3F,$ED,$E9,$E2,$DA,$56,$9F,$F0,$94,$73
	.dc.b	$7A,$89,$E2,$E8,$C1,$CA,$3A,$57,$F0,$EF,$9E,$FF,$5F,$C7,$FB,$A1
	.dc.b	$B3,$FE,$E3,$F3,$B5,$EC,$4A,$D3,$FC,$47,$DE,$FF,$51,$46,$0F,$F7
	.dc.b	$8F,$CF,$3E,$1F,$F8,$DB,$63,$FD,$D3,$F3,$B8,$99,$1C,$1E,$91,$8C
	.dc.b	$7D,$47,$FB,$ED,$FE,$50,$D9,$FF,$21,$F4,$9F,$F2,$2D,$3E,$7B,$FC
	.dc.b	$9D,$1F,$EE,$7F,$83,$61,$C9,$B1,$98,$FE,$69,$5C,$7D,$8D,$44,$9E
	.dc.b	$E7,$6B,$7C,$E6,$72,$FA,$18,$78,$63,$F9,$B7,$CA,$3A,$3F,$9E,$FC
	.dc.b	$B7,$E9,$EF,$FD,$7B,$39,$C7,$AC,$3C,$E5,$6F,$D0,$5E,$EB,$EF,$56
	.dc.b	$9E,$F8,$76,$7C,$3F,$2F,$FD,$33,$E7,$9F,$09,$8C,$FB,$E0,$9C,$C7
	.dc.b	$8B,$2E,$F7,$C6,$F8,$61,$E0,$BD,$BF,$8F,$E9,$FF,$14,$FC,$4F,$E8
	.dc.b	$FD,$D3,$7E,$2B,$0C,$CF,$F4,$13,$A6,$CE,$6F,$A4,$E5,$B5,$CE,$B8
	.dc.b	$BE,$DF,$86,$91,$E9,$DD,$F3,$18,$BF,$45,$5E,$8F,$C4,$47,$D2,$9C
	.dc.b	$5F,$FA,$3B,$D3,$7C,$BB,$E7,$BF,$BC,$46,$3C,$4B,$39,$AF,$B3,$39
	.dc.b	$6C,$67,$0F,$DA,$4E,$5B,$19,$C7,$C7,$39,$7C,$BB,$FF,$95,$3E,$77
	.dc.b	$E6,$1F,$37,$E6,$6E,$3F,$2A,$2E,$3F,$22,$6D,$F2,$E0,$56,$2F,$F8
	.dc.b	$A4,$2B,$17,$F9,$4F,$07,$FC,$27,$F3,$9F,$9E,$BF,$99,$18,$EB,$FC
	.dc.b	$EB,$63,$27,$E3,$27,$85,$7F,$13,$E8,$FC,$CF,$C6,$74,$F5,$FF,$4A
	.dc.b	$DB,$FB,$0D,$AE,$1C,$BF,$EE,$E9,$E9,$5E,$4F,$9A,$91,$F7,$FE,$7A
	.dc.b	$DF,$E9,$32,$56,$38,$3E,$FE,$7F,$AA,$BC,$61,$FE,$1F,$BF,$8A,$FA
	.dc.b	$78,$33,$8F,$EC,$E7,$2F,$58,$F8,$A6,$BB,$53,$AC,$7F,$1B,$4F,$96
	.dc.b	$7C,$F7,$E4,$3E,$1B,$F6,$73,$FE,$DC,$9C,$E5,$FE,$87,$EF,$FF,$6E
	.dc.b	$C3,$E4,$F8,$46,$30,$B7,$C5,$AF,$76,$D7,$1B,$E9,$3B,$6B,$FC,$3B
	.dc.b	$BF,$F9,$C3,$FE,$95,$BF,$F7,$6F,$86,$9C,$E5,$E8,$AE,$D7,$CA,$B5
	.dc.b	$E2,$5B,$DE,$1C,$9B,$7E,$13,$B1,$FE,$FC,$AE,$2E,$0F,$25,$73,$3C
	.dc.b	$F8,$FC,$68,$D4,$F0,$C7,$C8,$DC,$3C,$17,$C3,$2F,$8B,$39,$BF,$06
	.dc.b	$3E,$47,$B9,$E1,$C7,$B6,$3D,$27,$55,$C6,$B7,$3F,$0E,$B7,$8F,$82
	.dc.b	$7C,$B7,$C0,$ED,$3A,$8F,$0E,$3E,$D7,$6F,$9F,$2C,$61,$E1,$8B,$87
	.dc.b	$E0,$66,$D7,$23,$89,$E9,$CB,$DF,$8F,$86,$F7,$3F,$60,$DE,$1D,$5B
	.dc.b	$87,$56,$9D,$63,$47,$BE,$FA,$D7,$4C,$BE,$26,$5E,$DB,$F8,$B8,$7C
	.dc.b	$3C,$38,$BD,$5B,$1D,$63,$46,$4A,$E4,$87,$A6,$38,$BD,$8C,$CF,$86
	.dc.b	$BE,$B1,$F8,$C8,$1F,$54,$A7,$C9,$29,$F8,$22,$AF,$93,$FF,$5B,$46
	.dc.b	$8F,$EE,$AE,$1D,$B5,$FD,$4B,$5F,$CF,$3C,$AB,$F2,$ED,$BD,$FC,$37
	.dc.b	$ED,$37,$FF,$28,$DF,$EB,$6B,$4F,$29,$C1,$BF,$47,$FD,$CD,$7C,$A7
	.dc.b	$F4,$B5,$C1,$F3,$D6,$DE,$26,$9B,$87,$EF,$70,$F0,$CE,$49,$1E,$89
	.dc.b	$E5,$18,$BF,$04,$E6,$3F,$75,$5C,$A3,$04,$9D,$9F,$4A,$34,$E2,$7D
	.dc.b	$C7,$53,$EC,$C6,$9C,$CF,$36,$CF,$3C,$3E,$1C,$60,$87,$C2,$3C,$AE
	.dc.b	$9C,$4F,$36,$CF,$37,$E2,$BE,$57,$27,$92,$79,$63,$F8,$7C,$37,$3E
	.dc.b	$0C,$3D,$0F,$02,$4E,$DF,$19,$C0,$CE,$3F,$66,$78,$CE,$F1,$F3,$37
	.dc.b	$1F,$35,$FC,$73,$8E,$1F,$5A,$B7,$87,$E0,$EF,$85,$7A,$4D,$5F,$2A
	.dc.b	$FC,$16,$3F,$17,$17,$CF,$58,$97,$C0,$FA,$B1,$F4,$38,$BF,$CB,$3E
	.dc.b	$3F,$EB,$B8,$FF,$C5,$F1,$8A,$FF,$28,$E6,$AD,$71,$7E,$DA,$BF,$CC
	.dc.b	$3E,$26,$2F,$FC,$C9,$7B,$78,$DE,$6B,$A4,$7F,$99,$27,$FC,$82,$F6
	.dc.b	$AD,$46,$23,$51,$D7,$9F,$FC,$DB,$F3,$1A,$89,$3C,$4F,$F2,$64,$6F
	.dc.b	$AF,$FD,$CB,$1E,$AE,$CE,$A6,$7D,$C9,$19,$21,$FE,$1C,$79,$7A,$9E
	.dc.b	$4B,$93,$FE,$59,$8C,$BF,$1A,$32,$FF,$89,$23,$E2,$2F,$9E,$7C,$3E
	.dc.b	$1B,$1F,$E8,$C3,$51,$C2,$BF,$8F,$F0,$7F,$85,$AB,$75,$FE,$B0,$6C
	.dc.b	$66,$B8,$4F,$56,$3D,$67,$7C,$1F,$CB,$2F,$AB,$FE,$6B,$F1,$DE,$B7
	.dc.b	$DB,$F2,$3F,$CD,$B9,$D9,$FF,$6F,$E3,$FE,$68,$AC,$19,$99,$19,$3F
	.dc.b	$13,$1F,$70,$8F,$FA,$05,$AF,$E0,$DB,$9F,$F9,$0B,$6F,$1C,$FC,$5F
	.dc.b	$9E,$FF,$31,$37,$C7,$F9,$B1,$88,$FF,$69,$BF,$E4,$C7,$44,$33,$1C
	.dc.b	$3C,$4E,$2C,$8D,$5F,$F9,$12,$B9,$47,$F9,$97,$E5,$1F,$39,$F1,$7E
	.dc.b	$75,$F3,$D4,$FC,$B4,$EA,$BF,$D6,$CD,$7F,$D1,$B3,$C1,$FF,$49,$B5
	.dc.b	$ED,$5F,$99,$F0,$FF,$51,$E1,$BE,$7A,$99,$AF,$F5,$54,$E1,$FF,$63
	.dc.b	$CE,$DC,$9F,$8D,$F1,$9F,$3B,$F9,$4F,$9E,$73,$D3,$F2,$F3,$AA,$E1
	.dc.b	$58,$23,$F9,$7A,$FC,$1A,$DF,$FC,$D7,$5C,$63,$1C,$FF,$DE,$A3,$13
	.dc.b	$AB,$F1,$72,$FF,$05,$F5,$3F,$CB,$58,$7F,$B8,$1A,$FF,$BA,$9B,$5F
	.dc.b	$E8,$6F,$71,$F1,$3C,$71,$97,$C5,$FE,$11,$3A,$8F,$F7,$5F,$B8,$F8
	.dc.b	$FF,$E2,$8D,$8C,$C7,$0F,$63,$2C,$8C,$9F,$EF,$1A,$4F,$EF,$A7,$FC
	.dc.b	$1D,$8F,$FA,$31,$BA,$FF,$86,$9B,$7F,$E1,$E7,$8B,$FC,$05,$CB,$E5
	.dc.b	$F3,$7A,$1D,$6E,$C6,$E2,$BF,$38,$E9,$1F,$CB,$FE,$79,$F2,$74,$93
	.dc.b	$FE,$2E,$C7,$F9,$FF,$C2,$BD,$6F,$E6,$CD,$9D,$67,$FC,$FE,$31,$1D
	.dc.b	$FC,$A4,$62,$FF,$A4,$FC,$AF,$F5,$2A,$E2,$29,$1F,$EA,$78,$CC,$63
	.dc.b	$CD,$FE,$8D,$3C,$67,$B3,$B7,$C1,$F1,$3F,$E3,$F7,$8B,$FB,$6B,$FE
	.dc.b	$B2,$DC,$7F,$AB,$3B,$9E,$3E,$9F,$F5,$38,$C5,$70,$8F,$E9,$DF,$14
	.dc.b	$F9,$1A,$6E,$25,$FB,$3C,$F5,$6E,$3D,$7F,$3C,$F8,$7F,$EC,$0D,$8C
	.dc.b	$C7,$F5,$BF,$8E,$FE,$67,$1F,$16,$74,$FC,$D6,$E5,$FD,$D1,$BA,$7F
	.dc.b	$99,$B1,$5F,$2D,$D5,$AC,$3A,$63,$E2,$3C,$1E,$7F,$F7,$3F,$96,$FD
	.dc.b	$24,$FD,$B6,$E7,$FE,$02,$6E,$7D,$F2,$FF,$66,$4E,$7B,$39,$12,$FE
	.dc.b	$24,$EA,$BF,$C0,$D7,$CA,$7F,$04,$DC,$CE,$74,$F5,$FB,$3F,$E0,$7E
	.dc.b	$8F,$C6,$61,$FE,$0F,$6E,$1F,$8B,$6D,$3E,$4B,$F7,$53,$F9,$59,$DE
	.dc.b	$3F,$B6,$E7,$E8,$7F,$82,$EB,$7A,$FF,$88,$2B,$64,$9D,$75,$FF,$86
	.dc.b	$7D,$17,$CF,$52,$65,$F8,$3F,$95,$7A,$D8,$8F,$D3,$57,$07,$F3,$ED
	.dc.b	$7F,$71,$AA,$B8,$C6,$AF,$96,$7F,$E3,$4B,$E0,$B6,$FC,$15,$F0,$BF
	.dc.b	$DA,$7B,$9C,$DF,$D6,$B5,$CC,$7F,$8A,$FE,$5F,$FD,$13,$7F,$E4,$0A
	.dc.b	$D3,$67,$FC,$71,$18,$BF,$C8,$57,$07,$FC,$85,$19,$BF,$DF,$37,$80
	.dc.b	$E7,$35,$F1,$AF,$E0,$9F,$3C,$F9,$EF,$F2,$33,$63,$F8,$2B,$5F,$EE
	.dc.b	$5F,$1D,$A9,$91,$A7,$FB,$8C,$7F,$8C,$3E,$73,$F6,$5A,$FF,$B3,$39
	.dc.b	$BF,$C1,$76,$F9,$D7,$CF,$7F,$B3,$A7,$31,$89,$E1,$ED,$FF,$B4,$B0
	.dc.b	$E7,$1F,$98,$8C,$76,$FF,$61,$8C,$1F,$ED,$AA,$E4,$E0,$F0,$61,$DE
	.dc.b	$B8,$E3,$DA,$7F,$DC,$38,$76,$C7,$C7,$7B,$9E,$8D,$53,$BB,$E7,$1D
	.dc.b	$F0,$F0,$DF,$28,$EC,$70,$C7,$D1,$1F,$12,$F9,$4F,$63,$75,$CA,$FA
	.dc.b	$D6,$EF,$8C,$7F,$BD,$A7,$55,$C4,$F8,$17,$D1,$F0,$63,$D3,$1F,$55
	.dc.b	$70,$9C,$9C,$30,$EA,$F3,$C7,$D4,$E6,$FD,$78,$75,$BE,$47,$37,$C0
	.dc.b	$7D,$89,$1F,$18,$A7,$D0,$BF,$B4,$E0,$F1,$96,$1F,$3C,$F8,$BD,$4F
	.dc.b	$D8,$8C,$38,$9D,$CE,$A3,$EB,$CE,$9C,$70,$E5,$F2,$B5,$FD,$49,$97
	.dc.b	$07,$DA,$F9,$EF,$ED,$7C,$9F,$72,$78,$9F,$8F,$8C,$BF,$91,$33,$A7
	.dc.b	$E8,$A3,$2F,$E9,$23,$3B,$7C,$49,$CB,$EE,$BF,$07,$8D,$FC,$46,$BA
	.dc.b	$1E,$6B,$E9,$19,$7A,$46,$6B,$A4,$E5,$E9,$1E,$78,$C3,$F4,$6B,$BE
	.dc.b	$1E,$D8,$E4,$7D,$AC,$3C,$AE,$CE,$51,$B9,$E3,$1E,$66,$E1,$E4,$6E
	.dc.b	$1E,$28,$F4,$D7,$1A,$E6,$71,$2C,$71,$6C,$FF,$24,$EC,$F8,$9B,$FF
	.dc.b	$BA,$AD,$F3,$38,$74,$7F,$88,$0C,$9F,$E0,$BE,$5F,$F2,$5B,$98,$E1
	.dc.b	$83,$9B,$F9,$4E,$18,$E7,$FE,$85,$5A,$E9,$7D,$2B,$E4,$31,$F8,$DF
	.dc.b	$F4,$7C,$FF,$C6,$75,$A9,$CC,$71,$AE,$0E,$AB,$86,$7F,$E4,$DB,$E1
	.dc.b	$39,$38,$5F,$27,$AD,$F2,$AE,$17,$CB,$5F,$C4,$69,$FD,$7F,$B3,$D5
	.dc.b	$97,$F6,$1C,$7F,$B5,$46,$0F,$83,$3B,$8F,$ED,$31,$83,$86,$1F,$DB
	.dc.b	$9C,$1C,$63,$99,$C0,$F8,$3E,$03,$EB,$FE,$43,$F2,$23,$31,$8C,$F2
	.dc.b	$D8,$C9,$CA,$32,$D3,$37,$F1,$EB,$A3,$FD,$FB,$C4,$7C,$57,$D9,$F9
	.dc.b	$E7,$AF,$EE,$9C,$E1,$FE,$64,$9C,$BF,$E2,$CA,$E1,$5B,$6C,$64,$E1
	.dc.b	$19,$69,$F8,$0F,$23,$FD,$82,$DF,$C0,$BF,$11,$18,$C7,$EF,$28,$CF
	.dc.b	$B3,$A6,$D7,$13,$A4,$65,$A6,$6B,$FC,$3B,$F3,$3F,$81,$8F,$F2,$C3
	.dc.b	$63,$F3,$8B,$E8,$3E,$FB,$9E,$13,$BF,$C1,$3E,$C7,$C0,$EB,$0F,$D2
	.dc.b	$C6,$1F,$F4,$7B,$70,$FF,$44,$FD,$5E,$B8,$FF,$DD,$55,$FE,$90,$F0
	.dc.b	$53,$CB,$F0,$8F,$AF,$D1,$FC,$77,$A1,$F8,$7B,$F0,$0D,$74,$FE,$BE
	.dc.b	$31,$97,$EB,$1B,$A7,$FA,$CF,$C4,$7E,$34,$F9,$AE,$18,$FE,$F2,$73
	.dc.b	$1F,$B9,$FC,$75,$F0,$74,$D5,$F1,$CB,$F7,$33,$A7,$FD,$5B,$F3,$BF
	.dc.b	$D6,$9B,$93,$E2,$7F,$84,$CE,$63,$FD,$8C,$FF,$62,$8C,$C7,$F6,$49
	.dc.b	$E3,$18,$8C,$B6,$32,$76,$8C,$BF,$EB,$AF,$A3,$7D,$77,$F1,$56,$C7
	.dc.b	$FB,$9F,$CD,$C4,$C8,$E0,$FE,$FB,$E1,$BF,$B6,$C6,$0F,$F7,$77,$B9
	.dc.b	$FC,$7F,$E9,$3F,$AD,$F2,$FF,$C0,$18,$C7,$47,$16,$D7,$03,$FD,$9B
	.dc.b	$5C,$1F,$EF,$FF,$10,$F0,$BF,$96,$EA,$8C,$4F,$FB,$63,$D9,$7D,$E7
	.dc.b	$F3,$6A,$FF,$77,$1B,$BE,$37,$C2,$3F,$01,$F8,$2B,$EA,$73,$5F,$7E
	.dc.b	$31,$8F,$FB,$92,$B9,$3F,$EE,$6E,$84,$EA,$7B,$E1,$D7,$6F,$EE,$15
	.dc.b	$AC,$7F,$92,$FC,$BF,$FA,$53,$1F,$F7,$5F,$CE,$3E,$A7,$3F,$01,$FD
	.dc.b	$1E,$31,$3F,$ED,$9A,$E1,$9F,$FB,$EE,$72,$FF,$A5,$23,$39,$7F,$7F
	.dc.b	$F5,$15,$A8,$FE,$77,$2F,$A7,$ED,$7E,$46,$E1,$DD,$B1,$D5,$B5,$FE
	.dc.b	$0B,$EA,$FF,$16,$57,$E6,$3D,$BF,$E0,$8F,$E7,$9F,$E1,$11,$FE,$31
	.dc.b	$F2,$F5,$3B,$6B,$B6,$E1,$FA,$0A,$E6,$D9,$E0,$7F,$64,$F9,$9F,$CB
	.dc.b	$C7,$F5,$F5,$F7,$BF,$DB,$5C,$E3,$FC,$15,$CD,$71,$EB,$D5,$6D,$B1
	.dc.b	$FD,$C3,$F1,$4F,$F4,$F3,$7E,$33,$FB,$C4,$63,$1F,$ED,$CA,$33,$F0
	.dc.b	$BA,$6D,$71,$3E,$27,$99,$F1,$FC,$9F,$90,$D8,$C9,$F2,$63,$2D,$33
	.dc.b	$87,$C5,$8C,$B6,$3F,$C0,$F8,$BE,$4F,$FC,$1C,$FF,$11,$9F,$F0,$23
	.dc.b	$99,$DF,$8F,$AA,$DB,$63,$27,$F8,$4B,$F1,$67,$DC,$FF,$86,$57,$D9
	.dc.b	$58,$C3,$67,$83,$71,$FF,$0A,$57,$09,$FC,$E7,$57,$F8,$94,$7D,$DF
	.dc.b	$E9,$9C,$36,$3F,$D0,$7C,$7F,$9D,$9D,$C7,$23,$27,$FA,$23,$AB,$F5
	.dc.b	$95,$D7,$EA,$24,$EB,$15,$6A,$30,$FE,$CE,$77,$2C,$3C,$11,$E4,$AD
	.dc.b	$E0,$8F,$E2,$BF,$36,$FA,$8F,$F3,$87,$C4,$3E,$47,$FD,$33,$39,$C3
	.dc.b	$FD,$13,$18,$BD,$BA,$BB,$97,$FA,$8A,$F8,$3F,$EF,$69,$D5,$F2,$8D
	.dc.b	$5F,$05,$5C,$30,$F1,$47,$FA,$B3,$BF,$FA,$CF,$9D,$C6,$3F,$83,$F1
	.dc.b	$7F,$53,$7B,$1E,$1C,$FA,$57,$08,$C9,$C2,$B9,$46,$6F,$BC,$F8,$AB
	.dc.b	$8D,$F5,$9C,$D6,$F3,$FF,$5E,$B6,$FC,$19,$F4,$BE,$15,$A6,$CF,$AD
	.dc.b	$AF,$79,$E1,$8C,$8F,$3D,$F5,$74,$62,$B8,$C6,$4D,$65,$F2,$70,$F0
	.dc.b	$5F,$3C,$7D,$37,$E0,$AD,$61,$E3,$AD,$5F,$08,$F8,$F3,$A7,$E4,$1E
	.dc.b	$19,$DB,$F0,$0C,$12,$FB,$C7,$D7,$9E,$58,$79,$F0,$F4,$9E,$3C,$BE
	.dc.b	$7D,$74,$C3,$BE,$1F,$53,$0E,$37,$C2,$B5,$18,$78,$61,$F0,$B0,$E7
	.dc.b	$87,$8F,$1F,$66,$3C,$E7,$2F,$9A,$74,$7A,$DC,$12,$78,$9F,$9A,$8D
	.dc.b	$9E,$67,$2F,$16,$BF,$20,$FF,$80,$15,$F9,$36,$E3,$F9,$96,$E1,$FE
	.dc.b	$FB,$97,$CA,$3E,$6B,$0D,$A1,$D1,$C3,$63,$67,$E7,$A3,$86,$3F,$9A
	.dc.b	$6E,$3C,$A9,$F5,$1F,$E8,$A7,$85,$C7,$F2,$75,$C1,$F0,$CE,$DC,$CE
	.dc.b	$EF,$B3,$FB,$79,$D4,$62,$76,$47,$F6,$31,$AB,$FD,$6D,$F7,$AF,$F0
	.dc.b	$2B,$70,$F8,$B8,$75,$F8,$1F,$56,$3D,$49,$87,$29,$D6,$1E,$58,$C3
	.dc.b	$E1,$7E,$16,$3E,$18,$FE,$21,$8F,$72,$57,$28,$E0,$69,$E8,$FC,$4C
	.dc.b	$3C,$77,$DB,$0F,$1C,$EC,$FE,$27,$87,$77,$DB,$7C,$9A,$6A,$F9,$BC
	.dc.b	$6B,$9B,$A7,$E4,$C6,$A3,$17,$E1,$23,$E5,$3B,$63,$F2,$9F,$59,$F8
	.dc.b	$29,$E5,$19,$AE,$ED,$AF,$8F,$3A,$C7,$E2,$DF,$2C,$7D,$97,$B8,$F4
	.dc.b	$5F,$89,$EC,$78,$AB,$8E,$1E,$A8,$F9,$CD,$7E,$11,$F8,$BF,$98,$F7
	.dc.b	$61,$F2,$49,$8E,$23,$E3,$F1,$FF,$4C,$FA,$3F,$CF,$72,$7F,$93,$F8
	.dc.b	$51,$8E,$2F,$EE,$51,$8C,$BF,$B6,$AF,$07,$51,$FD,$CA,$34,$FF,$6F
	.dc.b	$9E,$35,$AC,$3C,$13,$CA,$F8,$FC,$03,$C7,$7A,$5B,$CB,$B5,$EC,$FF
	.dc.b	$35,$47,$09,$DC,$64,$F0,$1E,$3B,$D4,$F9,$BD,$B7,$5F,$F0,$AA,$FF
	.dc.b	$12,$7C,$FE,$60,$DE,$39,$3A,$21,$FE,$22,$7A,$7E,$2C,$F8,$8B,$F5
	.dc.b	$12,$BF,$C2,$D8,$AB,$51,$86,$BE,$88,$D4,$F1,$8C,$4E,$B8,$9C,$5F
	.dc.b	$C5,$7E,$9F,$FC,$2D,$5F,$E9,$D7,$FC,$62,$DF,$03,$EA,$A8,$DB,$FE
	.dc.b	$AC,$6D,$74,$8F,$E5,$31,$C9,$D4,$F2,$8D,$D7,$87,$BB,$13,$2B,$C5
	.dc.b	$1F,$3F,$C9,$FE,$75,$87,$FB,$F5,$FA,$F3,$FE,$70,$E0,$D7,$74,$8F
	.dc.b	$F3,$C4,$71,$30,$FF,$A1,$63,$ED,$71,$72,$BF,$F4,$25,$62,$65,$7F
	.dc.b	$A3,$63,$66,$1F,$F6,$5F,$D3,$7D,$E7,$FD,$88,$BE,$83,$E3,$B6,$33
	.dc.b	$C7,$83,$33,$23,$24,$3A,$B9,$7B,$C7,$E5,$EF,$F3,$9F,$0D,$27,$FD
	.dc.b	$89,$19,$7F,$5B,$19,$7F,$D8,$91,$FA,$6F,$A5,$FC,$C4,$7E,$1F,$E9
	.dc.b	$5F,$35,$F9,$D6,$C7,$F9,$97,$AD,$5C,$35,$FB,$D1,$8E,$8E,$18,$4B
	.dc.b	$E1,$7B,$E3,$FE,$C5,$E2,$FE,$0F,$1F,$C3,$53,$E7,$5F,$0D,$F8,$E6
	.dc.b	$C7,$FB,$BE,$B8,$5F,$2E,$6D,$5C,$AD,$47,$E5,$AB,$71,$FD,$E9,$A9
	.dc.b	$B7,$F9,$28,$68,$F8,$17,$C6,$FF,$0F,$F3,$BC,$FA,$7F,$BA,$DB,$A7
	.dc.b	$F8,$1A,$F8,$57,$F1,$2B,$E3,$87,$FA,$2A,$F6,$FF,$A3,$9B,$5C,$23
	.dc.b	$FD,$1F,$EE,$3E,$53,$F8,$F3,$63,$31,$D7,$CC,$CB,$23,$27,$D7,$C3
	.dc.b	$FC,$2F,$9F,$CE,$FC,$D3,$F2,$D3,$FC,$70,$3F,$95,$70,$D8,$E1,$CF
	.dc.b	$FC,$02,$38,$C7,$3E,$8C,$19,$99,$5F,$EA,$B3,$AC,$69,$C5,$72,$DB
	.dc.b	$FC,$5B,$E0,$BE,$AD,$F5,$DF,$CB,$5B,$7F,$C8,$1A,$FF,$34,$F8,$0F
	.dc.b	$F1,$CF,$CC,$7F,$33,$F8,$DF,$F4,$6E,$BF,$C4,$67,$35,$FE,$BD,$6B
	.dc.b	$FC,$96,$77,$97,$F8,$FB,$E4,$1F,$86,$BE,$2F,$A7,$A9,$30,$FF,$63
	.dc.b	$56,$A7,$FA,$BA,$FC,$5F,$84,$FF,$B2,$3E,$9B,$FA,$7B,$FD,$2B,$D1
	.dc.b	$46,$35,$FE,$4D,$7C,$70,$FE,$B3,$F4,$CF,$82,$EB,$97,$4C,$BF,$A9
	.dc.b	$DE,$DF,$DA,$56,$DB,$5C,$88,$EB,$E0,$BF,$AC,$7E,$2A,$F9,$3B,$F4
	.dc.b	$92,$BF,$A6,$CE,$5B,$39,$7F,$A6,$42,$79,$CC,$9E,$31,$BD,$3F,$A1
	.dc.b	$73,$AF,$A5,$7F,$6C,$EF,$7C,$DB,$F0,$3F,$EF,$9F,$45,$FA,$8A,$F9
	.dc.b	$0D,$C7,$F8,$7D,$6E,$7F,$8A,$35,$DF,$89,$C8,$FF,$01,$C6,$67,$8D
	.dc.b	$E6,$BF,$DF,$B7,$D7,$9F,$F7,$57,$B7,$F2,$BF,$8A,$FC,$DA,$FE,$D8
	.dc.b	$B7,$FF,$01,$36,$3F,$BA,$BB,$C7,$E0,$B6,$B9,$D7,$03,$8F,$04,$BE
	.dc.b	$31,$97,$9C,$7F,$87,$3E,$7D,$F6,$5F,$E2,$1E,$07,$C7,$D3,$51,$2F
	.dc.b	$FD,$33,$9F,$F7,$3E,$95,$F0,$9C,$E7,$FE,$9E,$9C,$E1,$C6,$7E,$B6
	.dc.b	$3E,$07,$62,$76,$E1,$AF,$0C,$7F,$C4,$DF,$8A,$7F,$3A,$8F,$F0,$A2
	.dc.b	$F9,$EC,$9A,$F3,$A4,$FF,$87,$A7,$4D,$9C,$BC,$A7,$8B,$93,$83,$FD
	.dc.b	$77,$9F,$FC,$8D,$F2,$FF,$46,$7D,$0D,$E6,$FE,$B7,$3B,$9F,$54,$EF
	.dc.b	$23,$BF,$43,$F3,$3C,$1F,$D6,$3E,$99,$F4,$E8,$FF,$63,$F3,$BE,$9F
	.dc.b	$FD,$81,$F5,$1F,$DA,$B3,$4E,$5F,$F6,$6C,$66,$31,$0C,$3A,$DF,$4B
	.dc.b	$FF,$65,$46,$27,$9C,$76,$F5,$7F,$D9,$D7,$CB,$0E,$93,$C3,$1F,$35
	.dc.b	$F0,$3F,$DB,$B8,$73,$BE,$18,$72,$BE,$95,$E1,$37,$7C,$AB,$CF,$7C
	.dc.b	$6B,$75,$D4,$61,$E2,$78,$E3,$CD,$F7,$61,$CE,$B8,$E1,$E2,$9D,$61
	.dc.b	$D5,$E2,$75,$C3,$A3,$DB,$1F,$43,$AA,$F2,$E1,$D6,$3E,$7C,$E9,$E0
	.dc.b	$76,$8D,$3E,$4C,$7C,$D7,$D8,$EF,$87,$8E,$74,$78,$EF,$A4,$EA,$F9
	.dc.b	$D7,$D7,$A1,$F4,$EE,$B0,$E6,$58,$E0,$FE,$03,$C2,$7C,$A6,$E3,$EC
	.dc.b	$C6,$1C,$4F,$F5,$99,$D4,$7E,$2E,$74,$E3,$E2,$F2,$F9,$1A,$9A,$47
	.dc.b	$EE,$79,$3F,$EA,$CE,$9F,$B5,$1C,$4C,$FB,$BF,$E8,$1D,$BF,$37,$19
	.dc.b	$7F,$3B,$5C,$63,$F0,$C6,$77,$F8,$71,$99,$E3,$3E,$33,$AD,$F8,$1C
	.dc.b	$47,$4B,$F0,$5F,$C5,$C3,$F1,$D8,$73,$8C,$5F,$23,$C2,$72,$7C,$35
	.dc.b	$CA,$3E,$A5,$72,$8C,$47,$D8,$8C,$61,$E3,$6C,$74,$6E,$1E,$A6,$DF
	.dc.b	$8A,$3D,$15,$C6,$BC,$91,$B6,$CE,$CF,$2C,$FE,$CF,$91,$F5,$0F,$33
	.dc.b	$FD,$61,$BF,$F3,$FA,$FF,$25,$1F,$E7,$99,$D3,$FC,$DC,$CE,$1F,$DC
	.dc.b	$E3,$2F,$F8,$5E,$33,$3F,$E7,$E9,$CB,$D7,$D3,$EF,$3F,$E3,$AF,$93
	.dc.b	$7D,$46,$30,$FE,$9F,$F5,$1F,$EC,$EA,$FF,$6C,$D6,$9F,$F6,$DC,$66
	.dc.b	$7F,$DB,$13,$97,$FD,$B5,$19,$C3,$FC,$17,$39,$7F,$C1,$B1,$9A,$F1
	.dc.b	$5F,$87,$C1,$4E,$70,$FE,$C3,$87,$1B,$FF,$28,$CE,$74,$78,$12,$7F
	.dc.b	$AE,$61,$70,$FE,$C7,$3E,$DE,$BF,$ED,$6F,$BE,$F8,$B9,$AE,$19,$FF
	.dc.b	$A9,$A2,$E1,$FD,$BE,$6E,$1F,$E9,$09,$B8,$74,$AF,$F5,$84,$62,$B6
	.dc.b	$3E,$21,$F2,$DE,$F5,$A5,$8C,$36,$FF,$BC,$D7,$18,$C6,$3F,$D6,$67
	.dc.b	$FC,$E7,$7C,$9D,$BD,$9B,$3C,$32,$FF,$1E,$CE,$EB,$FC,$79,$E3,$BE
	.dc.b	$3F,$E3,$7D,$43,$EA,$3E,$34,$E7,$2F,$F0,$44,$E5,$FF,$1D,$D7,$0A
	.dc.b	$DB,$63,$27,$08,$CB,$4F,$6F,$CF,$BE,$A3,$F0,$4F,$FA,$EA,$33,$18
	.dc.b	$F3,$B2,$D8,$C9,$CA,$32,$D3,$39,$7C,$73,$FC,$57,$5D,$5F,$F2,$E7
	.dc.b	$8F,$F8,$B2,$7C,$E7,$E1,$5C,$57,$E9,$7D,$3F,$B3,$F1,$19,$8C,$71
	.dc.b	$A2,$92,$3F,$DE,$55,$F9,$BA,$E3,$1F,$E9,$4F,$A7,$FA,$5C,$5F,$E8
	.dc.b	$C6,$C7,$FB,$56,$9F,$37,$E1,$9C,$E7,$FE,$44,$F6,$FF,$DE,$37,$FE
	.dc.b	$A7,$FA,$7F,$F4,$33,$F8,$88,$CC,$FF,$AC,$67,$2F,$C3,$E7,$7D,$3F
	.dc.b	$FB,$B6,$FF,$1E,$D9,$E1,$E7,$7E,$D6,$31,$A2,$BF,$B2,$FD,$07,$F6
	.dc.b	$47,$FC,$9E,$FF,$68,$F8,$DF,$E1,$8E,$70,$FA,$4E,$67,$71,$FD,$B9
	.dc.b	$B3,$C2,$30,$70,$0E,$5D,$36,$38,$9F,$E5,$DF,$A7,$7D,$07,$F0,$16
	.dc.b	$C7,$FB,$BF,$CB,$C4,$C8,$E0,$FF,$05,$F8,$7F,$F6,$64,$60,$FF,$7B
	.dc.b	$7D,$3F,$F0,$BF,$A7,$FE,$C9,$A7,$C3,$CD,$E2,$30,$5E,$FC,$DE,$8D
	.dc.b	$C7,$99,$FE,$73,$C7,$9B,$FE,$0E,$F8,$B7,$67,$F3,$5C,$91,$89,$FF
	.dc.b	$3E,$FA,$CF,$A7,$FE,$83,$1C,$6F,$85,$6E,$FF,$44,$72,$8F,$E8,$17
	.dc.b	$C7,$CB,$DC,$4B,$FF,$79,$7D,$3B,$E9,$F1,$87,$FB,$EE,$31,$3F,$E8
	.dc.b	$ED,$DF,$3A,$FA,$7F,$E6,$1F,$46,$FA,$5F,$E6,$13,$9B,$FF,$49,$4E
	.dc.b	$5E,$51,$9C,$7F,$DE,$53,$97,$FD,$A5,$19,$E4,$FF,$53,$4E,$5F,$A5
	.dc.b	$8F,$8B,$E0,$7C,$37,$E0,$6E,$1D,$9B,$1D,$1B,$5F,$6F,$AB,$FC,$73
	.dc.b	$5F,$9A,$F6,$5E,$3D,$8C,$35,$FD,$0C,$69,$A6,$8A,$6A,$38,$61,$FA
	.dc.b	$49,$E0,$FE,$D7,$E6,$BF,$2E,$DF,$83,$FE,$CD,$18,$C7,$FD,$EE,$A3
	.dc.b	$3D,$7A,$6E,$3F,$D2,$A7,$4F,$F4,$CF,$9B,$FF,$72,$37,$E9,$FF,$B9
	.dc.b	$46,$31,$FE,$0C,$A3,$3F,$0D,$A6,$D7,$13,$E0,$F9,$9F,$1F,$D4,$F9
	.dc.b	$0D,$8C,$9F,$1A,$32,$D3,$38,$7C,$38,$CB,$63,$FB,$1E,$6F,$96,$FF
	.dc.b	$04,$3F,$C5,$23,$FC,$EF,$E3,$FF,$00,$9C,$FC,$C6,$9B,$3C,$9B,$7F
	.dc.b	$C8,$FE,$87,$F9,$7C,$FF,$70,$5F,$4F,$FE,$15,$73,$3B,$8C,$79,$D9
	.dc.b	$6D,$71,$7F,$3F,$D5,$FE,$11,$1F,$37,$89,$FF,$79,$46,$1F,$D0,$BB
	.dc.b	$8D,$5F,$F5,$9F,$6F,$F5,$51,$CF,$E9,$E5,$F2,$8B,$5B,$B5,$63,$9B
	.dc.b	$F8,$B6,$1C,$B0,$F0,$CF,$F0,$DF,$A4,$7D,$0F,$F1,$BF,$85,$7C,$B7
	.dc.b	$FA,$82,$73,$7F,$E9,$18,$C5,$ED,$D5,$DC,$FF,$D5,$37,$C1,$FF,$4B
	.dc.b	$4E,$AF,$94,$6A,$F8,$2A,$E1,$87,$8A,$3F,$D6,$7E,$03,$8B,$FA,$1C
	.dc.b	$74,$32,$E2,$78,$47,$3C,$BA,$D7,$1F,$43,$C3,$28,$C3,$CC,$C5,$F3
	.dc.b	$CF,$D1,$19,$9D,$E1,$D9,$B3,$FE,$C6,$6C,$FC,$2C,$3C,$37,$CE,$3C
	.dc.b	$F8,$78,$AB,$71,$E7,$CF,$15,$B8,$F1,$D6,$0F,$4E,$5F,$1B,$1F,$05
	.dc.b	$6F,$0E,$67,$49,$D5,$F8,$0F,$54,$EC,$ED,$1A,$8E,$E6,$AF,$C0,$FC
	.dc.b	$58,$D5,$76,$AE,$55,$E4,$C7,$DE,$F0,$CB,$E3,$4F,$69,$CE,$1D,$49
	.dc.b	$3F,$63,$0E,$B5,$D2,$73,$8F,$D2,$AE,$78,$75,$8F,$82,$F8,$72,$C5
	.dc.b	$7B,$EB,$6F,$A7,$1F,$0C,$60,$97,$D4,$E7,$3E,$03,$93,$96,$90,$F8
	.dc.b	$F1,$E9,$9E,$B3,$CA,$3A,$CF,$E0,$EB,$CA,$FC,$E0,$78,$0A,$7D,$72
	.dc.b	$9B,$3F,$1D,$F3,$EF,$9E,$CC,$7E,$9A,$30,$E3,$44,$62,$39,$C6,$27
	.dc.b	$8C,$7E,$8E,$78,$DF,$E7,$5B,$87,$82,$BF,$53,$F4,$FF,$E8,$9B,$FC
	.dc.b	$D5,$69,$F1,$4E,$0D,$FB,$FF,$EF,$3B,$E5,$3F,$AB,$AE,$0F,$86,$B6
	.dc.b	$F1,$34,$DC,$3F,$7B,$87,$86,$72,$48,$F4,$4F,$28,$C5,$F8,$27,$31
	.dc.b	$FB,$FA,$E5,$18,$24,$EC,$FA,$71,$A7,$13,$EE,$3A,$9F,$6A,$34,$E6
	.dc.b	$79,$B6,$79,$E1,$F0,$E3,$04,$3E,$11,$E5,$74,$E2,$79,$B6,$79,$BF
	.dc.b	$15,$F2,$B9,$3C,$93,$CB,$1F,$C0,$47,$A0,$D0,$BF,$4B,$88,$DD,$F8
	.dc.b	$27,$31,$8B,$EE,$DA,$EE,$78,$CF,$63,$97,$B5,$F8,$F1,$FC,$0C,$E4
	.dc.b	$D5,$F6,$8C,$93,$1F,$B9,$8F,$E0,$AF,$E2,$D7,$A0,$9F,$4F,$88,$98
	.dc.b	$F0,$CB,$ED,$CF,$D1,$E1,$FE,$A1,$F4,$FF,$B0,$E3,$FE,$CF,$C4,$8C
	.dc.b	$6F,$FB,$6A,$DE,$3F,$B8,$BD,$F8,$3A,$84,$E9,$FE,$F5,$87,$37,$85
	.dc.b	$F3,$8C,$46,$AB,$97,$8F,$FE,$6C,$F4,$FA,$61,$FD,$96,$6D,$9D,$F9
	.dc.b	$2A,$A6,$A2,$61,$FE,$6E,$8F,$37,$73,$B7,$F8,$2C,$FF,$88,$7E,$87
	.dc.b	$30,$6F,$2C,$9D,$10,$FF,$15,$3D,$3E,$99,$F0,$D7,$E9,$E5,$7C,$01
	.dc.b	$8A,$B7,$19,$8C,$57,$08,$FF,$6C,$4A,$BB,$4F,$CF,$97,$D3,$FF,$86
	.dc.b	$9E,$C6,$3C,$0D,$47,$F8,$CA,$32,$ED,$FF,$52,$B7,$8F,$FD,$79,$1C
	.dc.b	$4D,$CF,$2E,$C9,$18,$21,$C9,$F9,$BE,$4F,$F8,$5A,$3F,$CD,$7B,$E0
	.dc.b	$CF,$74,$8F,$F3,$A4,$65,$FF,$3B,$47,$D6,$E1,$E5,$8F,$FA,$02,$B0
	.dc.b	$66,$65,$7F,$9F,$E3,$2F,$FA,$06,$3E,$37,$D3,$FD,$B8,$F8,$AB,$E9
	.dc.b	$72,$6B,$8A,$4E,$A3,$2F,$08,$CB,$FE,$CC,$9E,$4E,$CD,$1F,$EB,$CA
	.dc.b	$FC,$BF,$C2,$C8,$C1,$0F,$D2,$B9,$7F,$D7,$F1,$FA,$2F,$97,$FC,$AC
	.dc.b	$32,$FF,$78,$71,$BE,$9D,$F3,$1F,$9C,$6D,$7F,$B5,$6F,$8D,$FF,$B1
	.dc.b	$9B,$D9,$FE,$88,$AD,$D7,$58,$C5,$F8,$79,$FF,$1D,$E2,$FE,$26,$BF
	.dc.b	$DD,$7F,$4E,$F9,$AF,$C9,$36,$3F,$23,$3B,$CF,$F1,$6D,$C7,$FB,$EB
	.dc.b	$67,$FC,$02,$D5,$5F,$93,$9D,$F3,$78,$89,$3F,$9E,$9D,$2E,$F7,$C0
	.dc.b	$BE,$2F,$F8,$4F,$9D,$F9,$CC,$FF,$D0,$6D,$CB,$FC,$23,$7C,$2B,$F8
	.dc.b	$8D,$F1,$C3,$F7,$1E,$17,$FA,$C6,$F9,$9A,$AF,$E2,$BE,$FB,$DE,$A4
	.dc.b	$BF,$E0,$18,$F9,$3C,$BC,$47,$FA,$76,$B8,$3D,$9A,$FD,$33,$D9,$97
	.dc.b	$F8,$67,$4F,$DA,$FD,$3F,$E6,$E7,$41,$FC,$9B,$86,$C7,$0E,$AF,$E1
	.dc.b	$11,$C6,$39,$F5,$E0,$CC,$CA,$FE,$17,$5C,$B5,$FF,$1A,$B6,$3F,$C7
	.dc.b	$2F,$38,$FF,$69,$FD,$33,$E9,$FF,$97,$B6,$BF,$8F,$35,$F7,$7C,$D7
	.dc.b	$F8,$DB,$E6,$3F,$9A,$FC,$6F,$FA,$6B,$46,$3F,$E4,$6A,$D5,$FF,$2B
	.dc.b	$6B,$F1,$FB,$57,$C6,$BF,$C9,$BF,$20,$FA,$77,$C5,$FF,$57,$8C,$36
	.dc.b	$33,$87,$D8,$F9,$17,$B9,$BF,$A6,$CC,$FD,$5C,$11,$8D,$7C,$F2,$9C
	.dc.b	$E3,$FD,$73,$E9,$9F,$05,$4E,$91,$2B,$F0,$F3,$A6,$CE,$DA,$6B,$E0
	.dc.b	$BF,$AD,$FD,$2B,$E4,$AF,$D3,$4A,$FE,$99,$39,$6C,$E5,$F9,$10,$9E
	.dc.b	$73,$27,$8C,$6F,$3F,$E2,$35,$CE,$7F,$2D,$DA,$F9,$97,$D3,$7F,$78
	.dc.b	$FA,$0F,$D5,$57,$FB,$ED,$B8,$7C,$6B,$E1,$5F,$E0,$0B,$E3,$7F,$B0
	.dc.b	$BD,$F2,$7F,$84,$5B,$1F,$BF,$9D,$BE,$26,$E1,$FF,$00,$27,$7E,$19
	.dc.b	$E3,$8F,$82,$FF,$B8,$7D,$37,$CF,$7F,$14,$B6,$FF,$03,$B6,$37,$BE
	.dc.b	$0C,$CC,$8C,$90,$E6,$E5,$EB,$1F,$E1,$EF,$9D,$7D,$37,$F8,$7B,$A9
	.dc.b	$F1,$F4,$D4,$4B,$FE,$27,$3A,$6C,$E5,$FA,$75,$D1,$C9,$0C,$D7,$F8
	.dc.b	$8B,$E9,$BF,$9C,$47,$EA,$57,$56,$4D,$79,$B2,$7F,$C2,$D3,$A6,$CE
	.dc.b	$5E,$53,$C5,$C9,$C1,$FE,$B9,$CD,$FE,$3C,$F9,$5F,$A1,$C5,$FA,$8E
	.dc.b	$24,$63,$93,$F9,$AF,$A5,$F8,$FA,$FE,$B5,$7C,$78,$26,$5F,$D7,$E3
	.dc.b	$FC,$49,$F4,$8F,$A5,$FE,$D1,$3F,$81,$FA,$5F,$F6,$6F,$D3,$7F,$60
	.dc.b	$E8,$46,$3D,$5F,$F6,$8C,$62,$BF,$DA,$B8,$73,$CF,$FD,$AF,$39,$94
	.dc.b	$E5,$EC,$7F,$B7,$30,$ED,$87,$58,$E2,$FF,$B7,$A7,$85,$73,$6C,$6E
	.dc.b	$79,$8A,$E6,$F9,$31,$F6,$DF,$16,$D7,$18,$F3,$57,$0A,$F3,$D6,$9E
	.dc.b	$58,$79,$1F,$46,$3F,$0B,$0E,$47,$9E,$B8,$4F,$2A,$E0,$F9,$6B,$83
	.dc.b	$C8,$CE,$3E,$C7,$B9,$EC,$C3,$C8,$79,$63,$51,$D4,$D4,$F4,$7B,$1D
	.dc.b	$2F,$C7,$3C,$2B,$9D,$6E,$39,$9F,$56,$87,$AE,$FF,$E0,$0C,$66,$30
	.dc.b	$68,$FA,$DD,$EF,$93,$DB,$63,$47,$E1,$9C,$38,$9F,$E9,$F3,$A8,$DF
	.dc.b	$C5,$D7,$B3,$5F,$2B,$53,$4F,$9F,$E3,$FF,$A7,$F8,$7E,$8C,$F0,$3F
	.dc.b	$0F,$19,$7F,$12,$67,$6F,$CC,$46,$5F,$CD,$46,$76,$F8,$13,$97,$D1
	.dc.b	$7E,$0F,$73,$FA,$ED,$F0,$BF,$29,$E0,$8C,$D7,$08,$C6,$1E,$67,$75
	.dc.b	$E2,$AD,$C7,$C4,$36,$7A,$0D,$D7,$92,$37,$1F,$20,$DC,$F8,$EF,$CD
	.dc.b	$7D,$DB,$3C,$DB,$3B,$8E,$37,$E7,$8F,$25,$71,$9C,$9E,$58,$CB,$FB
	.dc.b	$4F,$8A,$7E,$13,$FE,$00,$DB,$FC,$EB,$E3,$FF,$B2,$70,$FF,$59,$D6
	.dc.b	$9F,$F5,$F4,$66,$FF,$C7,$33,$97,$FC,$35,$19,$AF,$F5,$9C,$E5,$EB
	.dc.b	$E7,$F7,$BF,$F7,$97,$E6,$B9,$6E,$39,$C7,$FD,$9F,$E5,$3B,$F1,$8B
	.dc.b	$0E,$15,$FE,$8D,$6F,$83,$FE,$20,$F3,$F7,$18,$C3,$FC,$79,$97,$FB
	.dc.b	$6F,$0E,$59,$77,$C3,$84,$E6,$F8,$61,$E3,$C5,$F9,$B7,$F8,$B2,$BF
	.dc.b	$DA,$ED,$8F,$F1,$D7,$E4,$EF,$43,$84,$E7,$3F,$F6,$AB,$6F,$F7,$79
	.dc.b	$2F,$FA,$04,$74,$D3,$C9,$18,$C7,$FC,$CF,$7C,$5F,$F7,$9C,$67,$0E
	.dc.b	$B5,$C3,$AF,$E3,$FE,$6E,$F7,$29,$EF,$F8,$77,$C1,$63,$4F,$89,$5B
	.dc.b	$7F,$DF,$9F,$91,$BE,$DB,$5A,$FF,$95,$E3,$13,$C3,$1F,$95,$18,$9F
	.dc.b	$F4,$16,$8C,$3A,$47,$1C,$3E,$67,$E5,$6F,$82,$A7,$D3,$F9,$AF,$F0
	.dc.b	$5E,$7F,$C8,$31,$FF,$7E,$D6,$E3,$07,$F1,$F7,$18,$7F,$9A,$E9,$F6
	.dc.b	$3D,$2B,$72,$E3,$56,$A7,$32,$BE,$78,$7F,$B6,$2B,$53,$FE,$98,$9C
	.dc.b	$D7,$DA,$9C,$D7,$F9,$AB,$C0,$7D,$63,$EE,$FF,$39,$18,$7F,$21,$F4
	.dc.b	$9F,$E7,$CC,$3F,$D0,$95,$B6,$CF,$E5,$7F,$1C,$7D,$1F,$2A,$D6,$9F
	.dc.b	$E8,$99,$CD,$F0,$3D,$FA,$A3,$33,$FE,$98,$AE,$39,$FF,$BD,$27,$2F
	.dc.b	$FA,$0A,$7F,$CE,$9C,$7F,$A1,$F2,$FF,$C5,$B8,$FC,$67,$ED,$D6,$DC
	.dc.b	$D7,$FA,$F1,$B1,$CF,$CC,$FF,$2A,$57,$1F,$47,$F5,$91,$8B,$FF,$61
	.dc.b	$B6,$B8,$7A,$BE,$1A,$DB,$FB,$7F,$7F,$C3,$87,$EA,$9B,$87,$DD,$FC
	.dc.b	$75,$E3,$6F,$4F,$A3,$F4,$DF,$C1,$BE,$09,$DD,$8F,$1B,$F8,$14,$63
	.dc.b	$3F,$93,$F8,$8B,$E7,$3F,$83,$36,$3F,$D8,$77,$89,$FE,$0D,$EF,$FF
	.dc.b	$21,$9F,$EE,$4D,$D5,$F2,$0F,$C0,$1F,$1F,$C3,$CE,$98,$7E,$5A,$B5
	.dc.b	$E3,$63,$0F,$F7,$64,$62,$7F,$2F,$8F,$58,$C4,$FF,$BB,$23,$18,$FE
	.dc.b	$FA,$B7,$AF,$F2,$DB,$F1,$79,$1E,$88,$C5,$7F,$0D,$C3,$FD,$47,$18
	.dc.b	$9D,$5F,$96,$B5,$83,$E8,$B8,$FC,$A7,$C2,$6D,$79,$9B,$8F,$95,$BD
	.dc.b	$3F,$8F,$F8,$D7,$DB,$78,$BB,$7F,$98,$46,$30,$F0,$46,$27,$51,$FB
	.dc.b	$89,$D6,$1F,$1B,$3F,$F6,$64,$E6,$B7,$2C,$3C,$11,$87,$C5,$E1,$FE
	.dc.b	$DB,$EA,$5D,$1E,$28,$C3,$70,$FF,$7D,$4E,$9F,$14,$E9,$F2,$77,$7C
	.dc.b	$CF,$13,$D6,$D8,$C9,$F2,$E3,$2D,$3F,$89,$FE,$21,$FB,$47,$FC,$78
	.dc.b	$DC,$FF,$17,$A3,$E9,$BF,$AB,$4E,$64,$EF,$97,$F2,$1C,$3A,$F1,$BE
	.dc.b	$FB,$71,$FB,$8D,$BF,$94,$F4,$EB,$FC,$47,$E2,$AF,$93,$C5,$FE,$5E
	.dc.b	$30,$FA,$27,$73,$AB,$FC,$CD,$69,$FA,$55,$EC,$F5,$FF,$BE,$FE,$2D
	.dc.b	$EB,$8F,$83,$18,$D7,$DA,$D8,$C9,$FD,$EE,$78,$57,$43,$FC,$CF,$F6
	.dc.b	$0F,$99,$FE,$F4,$DA,$F4,$7B,$18,$99,$1E,$BE,$CB,$7E,$86,$F4,$27
	.dc.b	$FC,$E7,$C2,$D5,$5A,$9C,$C6,$1F,$E9,$B9,$7F,$79,$AE,$1E,$17,$93
	.dc.b	$BB,$E6,$B6,$BE,$63,$72,$F9,$2F,$AB,$7F,$F1,$34,$BF,$29,$FE,$A2
	.dc.b	$FF,$A0,$71,$F8,$7A,$79,$F0,$F4,$56,$9F,$8B,$7C,$A3,$10,$AE,$37
	.dc.b	$FE,$90,$3B,$7A,$5F,$D6,$B8,$DE,$67,$F8,$A7,$C0,$FE,$EF,$87,$D1
	.dc.b	$F4,$3F,$96,$CE,$7E,$33,$83,$71,$FE,$E7,$3A,$7F,$AB,$DF,$2D,$7D
	.dc.b	$1E,$8B,$F2,$17,$CB,$FF,$8D,$DB,$7C,$32,$C4,$C8,$FE,$C3,$3A,$CF
	.dc.b	$F0,$96,$C7,$B5,$7F,$8C,$AB,$58,$FF,$61,$FA,$57,$D0,$7C,$9F,$84
	.dc.b	$FE,$85,$97,$FA,$CA,$73,$87,$FA,$83,$93,$FD,$61,$0F,$AA,$E1,$18
	.dc.b	$C2,$DF,$16,$AA,$E6,$FF,$AA,$27,$9D,$6A,$73,$87,$8A,$9F,$55,$FE
	.dc.b	$29,$E8,$F1,$47,$FB,$1A,$30,$F3,$3C,$19,$69,$E3,$ED,$FF,$54,$74
	.dc.b	$62,$B5,$5C,$AF,$C1,$87,$8B,$2F,$F6,$74,$E5,$B8,$73,$AF,$34,$E7
	.dc.b	$0F,$09,$D9,$C3,$DC,$C4,$F1,$AF,$91,$8F,$8A,$31,$3E,$D6,$BC,$6F
	.dc.b	$CB,$9F,$CC,$C3,$A9,$2B,$78,$FA,$6F,$96,$5D,$48,$FC,$28,$D3,$D7
	.dc.b	$2F,$6C,$EB,$1E,$F9,$59,$CD,$7A,$AB,$E7,$5F,$2C,$7C,$51,$B9,$F8
	.dc.b	$78,$7C,$6C,$7D,$75,$A9,$F1,$65,$EB,$9C,$92,$FA,$C6,$5E,$93,$E7
	.dc.b	$3A,$E3,$E4,$C3,$B4,$EB,$1F,$26,$1C,$EF,$D0,$78,$32,$E7,$97,$8F
	.dc.b	$0F,$7C,$E9,$E6,$DC,$7E,$BB,$4F,$9A,$53,$F3,$D9,$79,$DE,$E0,$F9
	.dc.b	$25,$38,$94,$FC,$31,$6B,$F9,$E7,$81,$98,$FD,$07,$6F,$F5,$57,$13
	.dc.b	$A6,$C6,$5E,$4E,$F0,$FC,$BB,$70,$FC,$EB,$67,$CD,$F5,$5F,$E4,$C7
	.dc.b	$F4,$B1,$97,$F6,$6D,$BF,$05,$FE,$DE,$B8,$57,$E8,$6B,$44,$C3,$F6
	.dc.b	$93,$AA,$F1,$C7,$62,$57,$ED,$A7,$58,$79,$A7,$35,$C2,$B5,$EA,$FC
	.dc.b	$B8,$C1,$2B,$84,$E8,$E2,$E2,$35,$1F,$02,$79,$4E,$49,$1F,$42,$BD
	.dc.b	$33,$A8,$C1,$EF,$8E,$AF,$32,$9F,$21,$B7,$EC,$8E,$37,$E2,$74,$79
	.dc.b	$CE,$84,$75,$1C,$63,$71,$C7,$1F,$A7,$87,$E3,$6B,$51,$8A,$EF,$3E
	.dc.b	$C5,$3A,$C7,$EB,$47,$79,$E5,$8F,$BE,$3B,$C7,$1B,$E9,$1C,$A7,$51
	.dc.b	$82,$57,$E2,$6F,$B6,$3F,$12,$78,$C7,$AC,$95,$DD,$C7,$D5,$58,$98
	.dc.b	$7D,$23,$F1,$71,$F4,$F9,$3F,$B0,$7D,$CF,$EB,$78,$FF,$B6,$F2,$A3
	.dc.b	$1C,$1F,$B3,$AD,$E1,$FE,$CA,$FC,$85,$8E,$A3,$FD,$D5,$87,$17,$FB
	.dc.b	$CE,$19,$BE,$31,$FD,$C2,$F8,$BF,$E0,$0F,$97,$7D,$C6,$E7,$33,$BC
	.dc.b	$7B,$75,$AF,$73,$C6,$F8,$54,$8C,$1F,$EF,$AF,$C5,$7F,$E0,$0B,$FE
	.dc.b	$FC,$FC,$0D,$EF,$FA,$61,$18,$78,$78,$39,$AE,$31,$FE,$4B,$37,$5C
	.dc.b	$6B,$84,$66,$77,$C1,$FE,$E9,$5F,$65,$BF,$36,$57,$96,$3B,$18,$7B
	.dc.b	$5F,$FB,$F9,$D1,$FC,$96,$FF,$DE,$9F,$8A,$BB,$3D,$5F,$59,$F2,$27
	.dc.b	$FD,$F0,$DE,$5F,$E5,$55,$BB,$E5,$18,$BE,$3B,$49,$C9,$FC,$CE,$B8
	.dc.b	$3F,$05,$B8,$78,$AB,$E5,$46,$2F,$F9,$C6,$0F,$53,$3F,$0B,$E6,$F3
	.dc.b	$66,$5D,$86,$09,$DC,$66,$FF,$CD,$6D,$AC,$BE,$3F,$53,$ED,$7D,$5B
	.dc.b	$CE,$A7,$92,$24,$F7,$9D,$36,$72,$FD,$DD,$3F,$C5,$9A,$FF,$99,$E1
	.dc.b	$F8,$27,$E7,$5F,$C1,$36,$35,$3C,$36,$FE,$B0,$DA,$FE,$B9,$5B,$9F
	.dc.b	$EB,$8D,$8F,$EC,$13,$B7,$EE,$78,$5F,$95,$8C,$19,$97,$A7,$9E,$8F
	.dc.b	$CD,$47,$E6,$AF,$E6,$35,$F6,$46,$DC,$F1,$F0,$30,$FC,$1F,$83,$FD
	.dc.b	$27,$C0,$BE,$76,$9E,$38,$93,$F6,$67,$4D,$9C,$BF,$A7,$FC,$E1,$F6
	.dc.b	$9D,$3E,$6E,$4F,$E0,$67,$2D,$9C,$BF,$AF,$B5,$7E,$DE,$B7,$3D,$6B
	.dc.b	$7C,$1F,$5E,$F7,$3C,$CE,$DF,$43,$FB,$FF,$A9,$FF,$40,$47,$E4,$5A
	.dc.b	$F9,$E3,$B1,$BF,$1A,$47,$23,$D1,$E0,$73,$E2,$91,$82,$1F,$C1,$EF
	.dc.b	$F4,$8F,$F8,$75,$AB,$3F,$E1,$39,$7E,$3F,$F3,$97,$D7,$7F,$15,$FC
	.dc.b	$AD,$F6,$54,$FE,$31,$3A,$C3,$F5,$4D,$7F,$8B,$4F,$07,$F5,$CD,$7F
	.dc.b	$8D,$46,$EF,$C1,$7D,$09,$3C,$3C,$5F,$E7,$58,$3E,$02,$92,$7F,$83
	.dc.b	$F9,$4F,$89,$7E,$01,$FC,$E3,$F3,$AF,$C2,$D7,$FA,$C2,$72,$F0,$6B
	.dc.b	$FE,$C7,$8D,$E7,$FE,$C0,$F9,$17,$C4,$FF,$4B,$6D,$78,$1A,$FF,$AF
	.dc.b	$3E,$45,$F3,$1F,$CF,$7E,$B5,$F4,$14,$95,$FC,$46,$1F,$85,$6C,$CC
	.dc.b	$7F,$3F,$8E,$2E,$E3,$FA,$47,$C6,$FF,$17,$8C,$10,$FE,$8C,$E5,$FF
	.dc.b	$71,$7E,$1E,$F0,$7F,$B2,$7C,$BF,$F9,$6F,$6F,$F6,$EB,$7C,$8F,$F3
	.dc.b	$17,$AC,$BE,$18,$7F,$B9,$9A,$FF,$B6,$AB,$82,$E4,$46,$2B,$96,$3F
	.dc.b	$DA,$AF,$8E,$2C,$7B,$C6,$1B,$3F,$DE,$7E,$B9,$E6,$FF,$67,$FB,$9F
	.dc.b	$AF,$87,$FB,$D3,$F2,$EF,$9F,$AF,$E0,$9A,$FF,$67,$9D,$E3,$FE,$7B
	.dc.b	$6C,$79,$E7,$75,$CF,$8B,$B7,$44,$AF,$E8,$D7,$C1,$3C,$B7,$61,$FE
	.dc.b	$04,$96,$1E,$D6,$FC,$D3,$C6,$FF,$0A,$7C,$FF,$F4,$C9,$DE,$9F,$E1
	.dc.b	$F8,$C5,$F1,$17,$C6,$BF,$B2,$C6,$2F,$9F,$A5,$27,$35,$B7,$A9,$E2
	.dc.b	$9F,$F1,$45,$71,$8D,$C7,$F8,$9F,$DB,$FF,$03,$C7,$F6,$5E,$E7,$CF
	.dc.b	$7F,$8A,$F4,$7C,$2D,$3F,$C3,$53,$AA,$E1,$59,$7E,$A4,$EA,$24,$68
	.dc.b	$87,$9B,$6F,$97,$D1,$E6,$FC,$EF,$F9,$7E,$9F,$2E,$5F,$5E,$FC,$0B
	.dc.b	$FC,$7C,$DC,$3E,$2D,$F0,$E2,$FF,$0C,$36,$3F,$2D,$3B,$7C,$0D,$C3
	.dc.b	$E8,$4E,$D3,$49,$E7,$FF,$93,$7D,$7F,$D0,$DB,$F3,$CF,$DD,$BF,$C5
	.dc.b	$3F,$06,$FF,$15,$3E,$D6,$E1,$FD,$F6,$77,$3F,$E3,$D6,$C7,$BE,$77
	.dc.b	$97,$F9,$8D,$B5,$FE,$D7,$AD,$CF,$FB,$5F,$EB,$7F,$88,$DF,$FB,$6F
	.dc.b	$EA,$FF,$84,$DB,$EB,$FF,$AA,$A8,$CC,$61,$DD,$72,$9D,$78,$BD,$7B
	.dc.b	$7F,$3B,$18,$3F,$DD,$17,$C6,$7F,$DD,$0D,$8F,$F7,$65,$73,$BE,$B7
	.dc.b	$C9,$F0,$FA,$DF,$EE,$F6,$D7,$A7,$0F,$06,$36,$39,$DF,$19,$EF,$7C
	.dc.b	$5B,$3B,$18,$78,$63,$9C,$7A,$4F,$89,$3C,$5F,$5B,$5E,$35,$DD,$F0
	.dc.b	$DF,$5C,$7D,$D1,$C8,$F0,$61,$DD,$D9,$DE,$36,$70,$8D,$47,$69,$E9
	.dc.b	$7E,$1A,$D5,$F8,$6B,$47,$39,$D1,$E7,$BE,$67,$D9,$81,$F0,$0A,$79
	.dc.b	$0A,$74,$2A,$FA,$F7,$D7,$EA,$7E,$DC,$61,$C4,$EE,$75,$1F,$42,$74
	.dc.b	$E3,$E3,$F2,$F9,$1A,$9A,$47,$EB,$78,$CF,$AF,$FE,$59,$B7,$E7,$27
	.dc.b	$91,$A8,$CE,$DE,$7B,$FC,$15,$F1,$DF,$F0,$97,$C3,$0F,$D2,$CE,$63
	.dc.b	$F0,$87,$4F,$9D,$CE,$3F,$CE,$DF,$04,$79,$C3,$AB,$F9,$EE,$B0,$F0
	.dc.b	$D7,$92,$BE,$7C,$62,$BC,$71,$CF,$0E,$4F,$88,$C6,$3F,$63,$1F,$34
	.dc.b	$F5,$9F,$81,$5C,$AF,$A4,$71,$6C,$EC,$EA,$5B,$E9,$87,$96,$BF,$0B
	.dc.b	$5C,$27,$E5,$47,$63,$A1,$6B,$F8,$6F,$C6,$FF,$56,$C3,$F0,$2D,$9F
	.dc.b	$C4,$35,$FC,$7B,$5F,$C9,$BB,$E1,$FE,$7D,$F5,$EF,$A7,$FE,$C5,$7A
	.dc.b	$F0,$EE,$1F,$E0,$FA,$D4,$FF,$64,$FA,$FE,$4F,$FB,$5D,$B3,$FE,$15
	.dc.b	$E8,$FF,$4C,$CE,$EF,$9E,$2B,$F2,$5F,$F7,$A6,$C6,$67,$FD,$69,$3D
	.dc.b	$5A,$ED,$F2,$F6,$FF,$B0,$FE,$BF,$AD,$F8,$A3,$18,$7F,$B4,$7E,$BF
	.dc.b	$FD,$90,$DC,$B8,$FD,$7F,$E3,$23,$04,$CB,$F6,$55,$C1,$B5,$E2,$51
	.dc.b	$97,$F7,$AE,$F3,$FE,$D5,$7F,$E5,$6B,$E9,$18,$E9,$F8,$7F,$5F,$FE
	.dc.b	$47,$9F,$F0,$7B,$70,$FE,$ED,$F4,$0F,$AC,$DF,$A3,$2F,$F8,$36,$2C
	.dc.b	$38,$61,$FC,$42,$31,$5F,$21,$B5,$C2,$BB,$47,$F7,$FF,$37,$FC,$23
	.dc.b	$3E,$2A,$F9,$B1,$8B,$FF,$0E,$8F,$AF,$7C,$1F,$E1,$A3,$0F,$D4,$F8
	.dc.b	$F7,$D0,$E6,$B7,$8F,$F2,$6C,$BF,$DE,$11,$8C,$3F,$90,$35,$FE,$4E
	.dc.b	$D7,$F9,$5B,$5F,$E5,$EF,$18,$FF,$0C,$5B,$E0,$3C,$18,$F8,$DF,$CD
	.dc.b	$F7,$3E,$BF,$FC,$BF,$97,$8C,$FF,$39,$B8,$CB,$FC,$4D,$F2,$7F,$CD
	.dc.b	$67,$50,$AE,$5E,$BF,$4B,$E1,$3F,$C2,$5A,$F4,$8F,$CC,$CF,$0A,$FE
	.dc.b	$7C,$D9,$FF,$3F,$7C,$B7,$3A,$FC,$EF,$B6,$8F,$9D,$F1,$5F,$63,$0F
	.dc.b	$E9,$AB,$D1,$CF,$1E,$99,$3F,$D1,$A7,$4D,$8F,$CD,$46,$31,$FE,$A1
	.dc.b	$F4,$6F,$8B,$FB,$8D,$BF,$83,$2A,$E1,$C5,$FE,$71,$85,$6D,$FF,$5B
	.dc.b	$78,$7F,$B6,$F8,$2F,$44,$7F,$A6,$7E,$09,$39,$D3,$EA,$FC,$1F,$F0
	.dc.b	$4E,$1F,$F6,$1F,$89,$F2,$23,$07,$F0,$8F,$09,$F5,$EF,$93,$FE,$06
	.dc.b	$DE,$0F,$F6,$87,$81,$89,$FC,$B7,$BF,$FC,$6B,$6F,$CD,$B7,$87,$FD
	.dc.b	$15,$F4,$2F,$AB,$FF,$45,$63,$F0,$2B,$4F,$F7,$C7,$95,$FF,$7C,$6B
	.dc.b	$9F,$2F,$F8,$34,$62,$3F,$C0,$AF,$25,$7D,$FC,$4F,$2D,$7F,$30,$8C
	.dc.b	$5F,$87,$E2,$7D,$26,$71,$78,$38,$9D,$1F,$96,$AF,$57,$93,$9C,$BF
	.dc.b	$D2,$B1,$89,$F6,$7B,$FF,$10,$E9,$8D,$8C,$9C,$F1,$7D,$7E,$70,$E7
	.dc.b	$18,$9F,$C2,$FD,$7B,$EB,$FF,$82,$36,$39,$7D,$FE,$26,$47,$F8,$6A
	.dc.b	$75,$7F,$E2,$16,$BF,$E2,$86,$BF,$E2,$C7,$7C,$3F,$63,$0F,$CA,$57
	.dc.b	$8A,$FE,$37,$77,$8B,$C5,$93,$FB,$AF,$B2,$7E,$06,$FB,$BC,$E9,$FC
	.dc.b	$7B,$46,$1F,$C2,$3E,$CD,$D9,$AD,$BF,$DA,$1D,$C9,$CE,$9F,$0B,$C5
	.dc.b	$F8,$27,$D7,$FB,$8B,$19,$3E,$67,$93,$8D,$3E,$5B,$FC,$AF,$ED,$3F
	.dc.b	$80,$4F,$EC,$1B,$3F,$43,$EC,$B5,$E5,$4A,$FF,$21,$D6,$AB,$FC,$9E
	.dc.b	$D7,$FC,$A6,$D7,$FC,$AE,$F4,$D3,$E5,$74,$F3,$9F,$BA,$F8,$FC,$19
	.dc.b	$7F,$83,$F1,$2D,$7F,$44,$C1,$5F,$99,$FB,$37,$D6,$6B,$2F,$D6,$C6
	.dc.b	$31,$F6,$7D,$17,$BC,$FE,$6B,$F8,$BB,$C5,$C4,$E8,$FF,$79,$FD,$AB
	.dc.b	$83,$3A,$FF,$BC,$32,$4E,$6B,$F8,$BF,$DA,$3E,$C3,$FA,$B3,$63,$F4
	.dc.b	$DA,$7F,$80,$70,$FE,$BB,$39,$6D,$7F,$7F,$8C,$47,$F9,$E9,$AF,$F9
	.dc.b	$F9,$AF,$FA,$09,$E5,$87,$C0,$CB,$C3,$5F,$15,$F1,$F8,$B3,$1F,$E9
	.dc.b	$BD,$0D,$7F,$3B,$1D,$67,$59,$7E,$EB,$3F,$E3,$CF,$A2,$3E,$87,$93
	.dc.b	$EE,$DA,$5F,$BB,$AA,$DF,$BA,$7F,$C7,$B2,$FB,$5F,$ED,$0F,$C8,$8C
	.dc.b	$C6,$25,$B7,$E8,$38,$FF,$9F,$DF,$F5,$A6,$CF,$FA,$7B,$5F,$C1,$FA
	.dc.b	$3F,$D8,$27,$33,$FE,$6D,$77,$6B,$FF,$54,$F9,$BF,$D3,$7D,$8F,$EE
	.dc.b	$F5,$A9,$CF,$89,$FE,$0C,$7F,$B9,$D6,$DB,$F6,$8F,$B5,$FF,$22,$D6
	.dc.b	$F2,$F4,$D7,$6A,$D3,$C2,$BF,$C9,$71,$89,$D8,$BE,$CE,$14,$6F,$2F
	.dc.b	$EC,$5F,$68,$FB,$5F,$65,$FF,$AB,$F8,$BF,$D8,$72,$F9,$6C,$FC,$47
	.dc.b	$F5,$9A,$FF,$64,$CF,$1D,$3F,$D7,$17,$C2,$F6,$D8,$CD,$70,$9C,$BC
	.dc.b	$B0,$AA,$FC,$15,$C2,$1F,$6B,$FE,$3A,$F4,$70,$D8,$F4,$1D,$67,$3E
	.dc.b	$F7,$F5,$88,$C9,$C9,$C6,$7E,$B8,$C5,$F7,$C3,$C3,$1C,$30,$ED,$5B
	.dc.b	$6E,$7F,$EE,$86,$CF,$84,$E1,$87,$69,$D3,$63,$FD,$E3,$3C,$1F,$81
	.dc.b	$9F,$C2,$8C,$12,$78,$4E,$B2,$F6,$E5,$23,$BB,$E9,$BE,$55,$C2,$F9
	.dc.b	$4E,$6F,$86,$3E,$BA,$D1,$DB,$2F,$83,$19,$AD,$E5,$EA,$79,$65,$E0
	.dc.b	$24,$ED,$F9,$F5,$CA,$B8,$46,$9B,$3F,$46,$75,$3F,$12,$FC,$77,$E0
	.dc.b	$8F,$99,$5D,$49,$7C,$8F,$C7,$47,$53,$94,$F1,$7E,$41,$93,$EF,$96
	.dc.b	$FC,$2E,$2B,$AB,$C8,$F5,$E5,$F3,$2B,$A6,$5F,$66,$F8,$BF,$3A,$BA
	.dc.b	$56,$DF,$8C,$0F,$86,$53,$B1,$4F,$29,$63,$FA,$27,$81,$98,$FC,$4F
	.dc.b	$6F,$F5,$87,$13,$A6,$C6,$5E,$4E,$F1,$FC,$D3,$72,$FC,$D3,$67,$CD
	.dc.b	$F6,$BF,$E6,$07,$F5,$F1,$97,$F6,$6D,$BF,$05,$FE,$AE,$B8,$57,$EA
	.dc.b	$6B,$44,$C3,$F7,$93,$AA,$F1,$C7,$62,$57,$EE,$27,$58,$79,$A7,$35
	.dc.b	$C2,$B5,$EC,$FC,$B8,$C1,$2B,$84,$E8,$E2,$E2,$35,$1F,$02,$79,$4E
	.dc.b	$49,$1F,$42,$BD,$33,$A8,$C1,$EF,$8E,$AF,$32,$9F,$21,$B7,$EC,$8E
	.dc.b	$37,$E2,$74,$79,$CE,$84,$75,$1C,$63,$71,$C7,$1F,$A7,$87,$E3,$6B
	.dc.b	$51,$8A,$EF,$3E,$C5,$3A,$C7,$EB,$47,$79,$E5,$8F,$BE,$3B,$C7,$1B
	.dc.b	$E9,$1C,$A7,$51,$82,$57,$E2,$6F,$B6,$3F,$12,$78,$C7,$AC,$95,$DD
	.dc.b	$C7,$D8,$D3,$11,$30,$FC,$5C,$7D,$3E,$3F,$F6,$07,$DA,$FE,$B7,$8F
	.dc.b	$FA,$EC,$62,$BF,$B9,$F3,$2B,$5C,$7F,$DF,$3E,$D5,$8E,$2F,$FB,$64
	.dc.b	$61,$E3,$59,$C3,$A4,$7F,$6F,$9D,$3F,$DD,$3A,$DF,$3A,$FA,$FE,$15
	.dc.b	$AD,$3C,$AD,$CF,$F8,$35,$F0,$AF,$13,$70,$FF,$79,$E1,$CD,$B3,$FC
	.dc.b	$15,$FF,$80,$11,$97,$F8,$47,$DA,$BE,$17,$D9,$0F,$B0,$FF,$75,$3F
	.dc.b	$EE,$86,$BF,$E0,$C1,$3C,$5F,$F7,$6B,$5D,$37,$3F,$F7,$73,$72,$FE
	.dc.b	$3D,$4F,$C0,$BC,$32,$FB,$6F,$1C,$BD,$0A,$6A,$26,$3E,$06,$BE,$59
	.dc.b	$7C,$EF,$B3,$ED,$9D,$FE,$EF,$B6,$F9,$97,$FC,$C5,$AA,$76,$FF,$35
	.dc.b	$D7,$FC,$46,$E6,$BE,$63,$4F,$9D,$3C,$1F,$E6,$8D,$7F,$9D,$39,$9E
	.dc.b	$6D,$D7,$FA,$03,$67,$C9,$EC,$7A,$7E,$0B,$D1,$E7,$C9,$F2,$5F,$09
	.dc.b	$D4,$60,$FE,$99,$7F,$EE,$E7,$07,$B6,$3E,$17,$6F,$C5,$F5,$BF,$03
	.dc.b	$F6,$CF,$3E,$99,$BE,$38,$2B,$11,$F8,$38,$C3,$DA,$9F,$0B,$F8,$CD
	.dc.b	$BF,$C5,$9B,$7C,$AF,$B6,$FD,$75,$F9,$1A,$FE,$36,$77,$AF,$F6,$37
	.dc.b	$FC,$69,$1F,$D8,$DA,$E7,$C4,$FA,$71,$88,$FE,$D8,$F2,$6C,$78,$CF
	.dc.b	$CF,$D3,$D6,$F3,$75,$E3,$96,$5F,$C8,$C3,$F5,$19,$7E,$09,$B9,$79
	.dc.b	$6F,$8D,$7A,$FE,$A7,$E0,$7C,$4F,$EB,$3E,$A1,$F6,$D4,$CD,$75,$9A
	.dc.b	$FC,$17,$87,$97,$FB,$4F,$B4,$FC,$DC,$FD,$AB,$7D,$17,$53,$5D,$92
	.dc.b	$7F,$80,$57,$07,$FC,$0C,$D7,$FC,$18,$D7,$FC,$20,$EF,$6F,$CF,$B7
	.dc.b	$5C,$D7,$AF,$D8,$F2,$7D,$B7,$F0,$8F,$B6,$79,$74,$CE,$5F,$90,$C5
	.dc.b	$58,$8F,$E1,$51,$87,$F2,$B4,$F8,$CF,$81,$F2,$5D,$79,$25,$73,$8C
	.dc.b	$B4,$87,$F1,$5A,$7E,$07,$FC,$73,$ED,$9F,$6D,$FC,$7B,$ED,$DF,$6F
	.dc.b	$4D,$44,$C3,$F9,$45,$70,$6C,$7F,$00,$8C,$47,$F8,$C5,$BC,$5F,$C1
	.dc.b	$9B,$C5,$FE,$3B,$6C,$FF,$8F,$F0,$FE,$44,$FF,$AD,$70,$7C,$47,$F4
	.dc.b	$96,$C7,$FB,$19,$AF,$FA,$E7,$E4,$DD,$5F,$D0,$BE,$DF,$F1,$B3,$FE
	.dc.b	$43,$F4,$51,$88,$FF,$6E,$35,$FF,$27,$46,$BB,$3F,$A3,$7A,$6F,$97
	.dc.b	$7C,$F7,$F5,$BF,$99,$75,$D3,$FA,$F4,$EB,$1F,$E1,$BF,$6E,$FA,$6F
	.dc.b	$E9,$6D,$8C,$C6,$3E,$DE,$47,$F5,$59,$47,$F2,$C6,$C6,$63,$F9,$7F
	.dc.b	$C9,$7F,$1B,$8F,$EA,$53,$A6,$CF,$F5,$A9,$D7,$37,$FB,$6B,$67,$DB
	.dc.b	$BC,$5F,$EE,$DF,$3D,$FE,$6D,$AF,$F0,$1B,$7A,$7F,$B7,$D6,$F2,$FE
	.dc.b	$59,$87,$2C,$3F,$B6,$36,$3E,$84,$ED,$FF,$3C,$46,$5B,$19,$C3,$C8
	.dc.b	$DB,$EF,$7F,$E0,$9F,$B7,$FE,$57,$7F,$DB,$FE,$DF,$EF,$BF,$CF,$78
	.dc.b	$DF,$6F,$F4,$63,$FD,$F8,$DC,$BF,$AB,$D7,$09,$E5,$CF,$FD,$62,$31
	.dc.b	$1F,$E8,$D6,$BF,$E9,$16,$BF,$E9,$57,$94,$F4,$9F,$F0,$3C,$63,$2F
	.dc.b	$F4,$D6,$1E,$4F,$2B,$FC,$4D,$6F,$96,$79,$5F,$E2,$5F,$A4,$7B,$14
	.dc.b	$DC,$EB,$4F,$EB,$57,$C6,$FF,$C3,$4D,$8F,$F5,$03,$5F,$B5,$3C,$5B
	.dc.b	$3D,$08,$FC,$9F,$80,$FF,$07,$3F,$D9,$3C,$07,$DB,$FF,$8E,$78,$1E
	.dc.b	$7D,$3F,$C4,$B3,$AA,$E1,$5F,$E0,$28,$C4,$7F,$AC,$DA,$FF,$AE,$5A
	.dc.b	$FF,$AF,$5D,$E5,$FE,$0C,$6E,$F9,$AF,$17,$3F,$CA,$FB,$5F,$E6,$58
	.dc.b	$38,$5F,$6C,$FB,$7F,$F2,$83,$67,$E4,$D7,$0C,$CF,$F3,$F8,$46,$3D
	.dc.b	$1F,$9B,$39,$8C,$4F,$CE,$8E,$44,$9F,$F2,$F7,$B7,$FA,$79,$FF,$6D
	.dc.b	$7D,$B7,$EC,$EF,$FC,$DB,$F6,$FF,$E3,$99,$FC,$C3,$5F,$F2,$54,$EF
	.dc.b	$6F,$F7,$03,$FB,$B8,$FF,$70,$35,$CF,$7F,$F8,$C2,$31,$1F,$EE,$9F
	.dc.b	$B6,$FE,$2F,$87,$FB,$AF,$EC,$BF,$88,$4B,$ED,$FF,$B3,$B9,$F1,$3F
	.dc.b	$DD,$D1,$88,$4F,$0B,$E7,$18,$AF,$F7,$A4,$62,$7A,$D7,$16,$FB,$1F
	.dc.b	$EF,$97,$97,$93,$F3,$B0,$E4,$79,$DC,$47,$08,$F2,$03,$D1,$3C,$F0
	.dc.b	$F0,$D7,$06,$E3,$C2,$B6,$5C,$3C,$93,$A8,$F0,$8C,$7D,$51,$F0,$B0
	.dc.b	$ED,$3E,$1A,$E2,$F9,$EF,$B5,$F0,$8D,$57,$07,$C2,$76,$8D,$3D,$F0
	.dc.b	$F8,$5F,$6B,$EF,$83,$4F,$8F,$7F,$05,$B1,$E8,$C3,$C6,$F6,$CB,$24
	.dc.b	$CB,$D5,$1F,$07,$2F,$26,$5E,$23,$ED,$C7,$CD,$3D,$85,$29,$5D,$BF
	.dc.b	$80,$FB,$67,$DB,$FF,$AA,$E5,$18,$9E,$47,$CD,$70,$F3,$6B,$AF,$94
	.dc.b	$FF,$44,$A6,$91,$FF,$46,$BF,$77,$C8,$7D,$BB,$8B,$F0,$15,$F8,$7A
	.dc.b	$D4,$E6,$3E,$F9,$93,$F0,$35,$CA,$B8,$C7,$84,$CC,$F1,$D3,$F5,$31
	.dc.b	$C5,$FD,$5D,$71,$8E,$A7,$18,$CC,$F1,$D3,$DD,$3D,$2F,$BE,$1E,$66
	.dc.b	$C6,$4E,$71,$96,$99,$BF,$1B,$DE,$BF,$C3,$37,$DE,$3E,$D5,$71,$AD
	.dc.b	$B6,$BC,$71,$E6,$BF,$0B,$6B,$BB,$4E,$2F,$86,$7F,$11,$F2,$5F,$D6
	.dc.b	$EB,$FD,$56,$D9,$FE,$57,$1F,$D8,$BE,$DD,$F2,$DF,$4F,$8B,$FD,$93
	.dc.b	$7B,$6F,$89,$9B,$EA,$DC,$3F,$9F,$61,$CA,$7B,$CE,$69,$F6,$DC,$E7
	.dc.b	$FD,$AC,$DB,$FE,$8D,$D3,$FE,$9D,$9F,$ED,$73,$9A,$E3,$7C,$E3,$87
	.dc.b	$77,$FB,$07,$E0,$F1,$32,$FF,$C2,$18,$3E,$DF,$19,$7C,$03,$BB,$E9
	.dc.b	$5F,$6F,$FE,$CB,$AF,$F6,$8B,$63,$86,$2F,$B7,$FE,$07,$5A,$C3,$95
	.dc.b	$EF,$4F,$F7,$8B,$67,$FC,$1F,$5C,$30,$F1,$1E,$0B,$E2,$7C,$BC,$3B
	.dc.b	$75,$E2,$FF,$C5,$57,$FE,$5E,$7F,$C1,$FF,$6F,$FE,$45,$6F,$A3,$C7
	.dc.b	$E8,$7F,$DB,$71,$82,$65,$F0,$74,$5F,$18,$F6,$D7,$07,$51,$99,$FC
	.dc.b	$23,$6B,$FC,$53,$E7,$7F,$84,$9E,$B3,$FB,$BC,$D3,$9B,$F7,$FD,$BB
	.dc.b	$E1,$3E,$83,$70,$FF,$23,$7C,$8B,$E8,$F3,$52,$7F,$A7,$65,$FC,$1F
	.dc.b	$06,$1C,$5F,$AB,$87,$1A,$D9,$FC,$A6,$75,$5C,$5B,$1F,$E4,$29,$7C
	.dc.b	$CF,$7B,$E6,$FE,$77,$B9,$F6,$EF,$93,$C3,$70,$FF,$2F,$7C,$AF,$F2
	.dc.b	$E7,$FC,$E3,$19,$C3,$F2,$17,$CA,$F8,$9F,$7E,$F8,$4E,$AB,$86,$7F
	.dc.b	$E8,$36,$E1,$FE,$58,$F7,$1F,$31,$CA,$3F,$CD,$D1,$87,$E0,$DF,$F7
	.dc.b	$5F,$8F,$FB,$0D,$F4,$BF,$41,$18,$8F,$F7,$1C,$60,$93,$CA,$3F,$DC
	.dc.b	$FC,$D8,$7F,$33,$F4,$AF,$8D,$FD,$53,$6B,$86,$0C,$33,$5F,$9D,$F8
	.dc.b	$27,$87,$F9,$FF,$84,$F2,$F4,$FF,$AE,$E3,$1A,$7D,$1F,$84,$FE,$0D
	.dc.b	$C9,$FE,$BC,$F0,$B3,$AF,$FA,$7A,$7F,$81,$FC,$3B,$BB,$F5,$5F,$6E
	.dc.b	$F9,$4F,$D6,$4E,$6F,$FD,$29,$39,$7D,$DF,$02,$FB,$7F,$E0,$8D,$8F
	.dc.b	$97,$F4,$8F,$B6,$FF,$42,$C7,$0A,$D9,$1D,$C6,$B9,$FF,$7F,$7C,$DB
	.dc.b	$5B,$7F,$8A,$5F,$5F,$84,$C7,$67,$E7,$ED,$9F,$90,$FE,$61,$5E,$3F
	.dc.b	$03,$19,$FA,$A6,$E3,$D3,$35,$63,$1E,$9F,$03,$EE,$39,$CF,$6C,$3E
	.dc.b	$1C,$65,$A7,$B3,$0C,$61,$FC,$6B,$EC,$9F,$6D,$FC,$AD,$B5,$F0,$BE
	.dc.b	$DF,$13,$23,$F1,$93,$AA,$F3,$56,$5E,$95,$E4,$ED,$C6,$7F,$E0,$7D
	.dc.b	$7E,$94,$77,$F0,$A6,$5F,$A0,$FB,$77,$DB,$BE,$DF,$39,$F9,$EE,$E5
	.dc.b	$FE,$11,$BD,$D6,$30,$FF,$18,$6B,$F9,$97,$CF,$E0,$FC,$6F,$B6,$F7
	.dc.b	$32,$32,$7E,$02,$31,$97,$F3,$EF,$B6,$FD,$A3,$7C,$6F,$9F,$F6,$5F
	.dc.b	$D4,$67,$3A,$7F,$1D,$6E,$1F,$E4,$16,$DB,$83,$F2,$57,$C7,$09,$7F
	.dc.b	$B7,$BD,$D6,$A3,$13,$DF,$B2,$47,$E5,$BB,$AE,$3E,$79,$B9,$7C,$FE
	.dc.b	$7F,$EE,$D2,$CB,$F5,$4F,$B7,$D5,$F5,$1F,$C9,$FE,$D3,$FD,$CF,$97
	.dc.b	$8E,$32,$D3,$F8,$AF,$82,$9D,$D7,$39,$FE,$21,$F6,$BF,$E0,$B8,$FE
	.dc.b	$CE,$DC,$BF,$05,$F6,$BF,$DA,$DF,$E8,$91,$9A,$FF,$34,$36,$BF,$D2
	.dc.b	$CD,$DD,$7C,$6B,$7E,$DF,$F8,$0E,$B8,$59,$E1,$5B,$8E,$73,$BF,$5E
	.dc.b	$5F,$B3,$2F,$CE,$65,$E4,$9C,$DF,$A9,$F8,$DD,$78,$9F,$4F,$1F,$84
	.dc.b	$FD,$EE,$4F,$17,$19,$7C,$49,$B8,$FF,$08,$AD,$4E,$4F,$9B,$C4,$F3
	.dc.b	$3F,$AD,$36,$3F,$A2,$F9,$2E,$E7,$C4,$EA,$E5,$72,$7F,$C2,$93,$A8
	.dc.b	$C1,$C9,$FE,$E1,$8F,$F8,$9B,$EC,$DF,$69,$C3,$D6,$FF,$0E,$46,$2F
	.dc.b	$84,$62,$75,$1E,$49,$D5,$7F,$89,$BE,$CD,$F6,$9F,$E8,$7F,$83,$FE
	.dc.b	$79,$3F,$ED,$19,$7C,$96,$7C,$FF,$F5,$44,$70,$C7,$FD,$21,$18,$9D
	.dc.b	$57,$1C,$2A,$BE,$77,$C2,$34,$D7,$55,$C3,$ED,$3F,$C3,$9D,$79,$6C
	.dc.b	$FF,$A9,$F1,$F0,$5E,$7D,$BF,$DD,$4E,$4F,$F7,$55,$70,$CF,$DB,$08
	.dc.b	$C5,$F2,$8E,$06,$6B,$9D,$70,$AD,$CF,$49,$CD,$71,$AD,$46,$27,$58
	.dc.b	$79,$4F,$76,$5F,$0A,$3B,$E1,$D1,$F0,$4E,$DF,$1C,$EE,$FA,$CE,$5B
	.dc.b	$3A,$CB,$E4,$61,$E1,$8E,$67,$67,$17,$CF,$1F,$F7,$16,$3E,$12,$4E
	.dc.b	$AF,$9E,$3D,$EB,$73,$9C,$B1,$8F,$B8,$ED,$5E,$EB,$E3,$18,$26,$5F
	.dc.b	$4B,$2F,$14,$71,$8F,$59,$31,$F9,$87,$82,$FB,$CF,$B2,$79,$5F,$67
	.dc.b	$84,$73,$AE,$2D,$21,$C5,$CB,$4C,$E5,$F4,$F0,$F2,$C6,$49,$19,$BF
	.dc.b	$05,$79,$67,$DA,$F4,$9F,$97,$82,$61,$B2,$9F,$58,$B1,$F9,$E6,$F5
	.dc.b	$BE,$B3,$71,$F9,$C8,$C3,$89,$FF,$67,$CE,$A3,$F3,$D3,$A7,$9C,$7E
	.dc.b	$AE,$38,$E5,$F9,$D6,$E1,$E6,$FB,$4F,$F2,$5D,$EE,$E3,$F9,$8A,$DB
	.dc.b	$ED,$9D,$B9,$9D,$DF,$57,$F7,$D3,$A8,$C4,$EC,$8F,$ED,$A3,$57,$F9
	.dc.b	$6B,$EF,$5E,$CB,$EB,$3E,$7F,$57,$E6,$E1,$D0,$91,$C6,$7C,$33,$A8
	.dc.b	$C5,$78,$1B,$5E,$19,$D4,$FC,$FA,$EE,$D3,$F8,$3E,$1E,$67,$15,$E0
	.dc.b	$6D,$F9,$5E,$27,$A4,$F0,$1F,$26,$FB,$B9,$9E,$4D,$9E,$4F,$03,$07
	.dc.b	$63,$91,$DB,$1F,$C1,$BB,$BF,$33,$C0,$91,$AA,$F8,$11,$C7,$1F,$9F
	.dc.b	$39,$BF,$0D,$79,$23,$15,$E1,$17,$DA,$33,$1D,$AF,$EC,$CE,$5E,$A4
	.dc.b	$C7,$EA,$CF,$E1,$BE,$D3,$E1,$D7,$A0,$8E,$0C,$E5,$27,$81,$FD,$37
	.dc.b	$9D,$F6,$6A,$FD,$8F,$6F,$FB,$AF,$9D,$F6,$98,$E3,$FD,$B6,$1F,$DE
	.dc.b	$63,$18,$A3,$9E,$18,$6C,$7F,$77,$9C,$5F,$48,$FF,$01,$E3,$E1,$7F
	.dc.b	$C6,$BF,$34,$F9,$BE,$BE,$2A,$31,$3A,$9E,$D5,$FD,$DA,$30,$43,$FD
	.dc.b	$C3,$3C,$16,$8D,$3D,$75,$C6,$73,$1F,$C4,$BE,$CF,$FD,$EB,$1F,$ED
	.dc.b	$36,$BA,$3F,$8C,$38,$AE,$91,$FC,$6F,$EB,$5F,$69,$F1,$A1,$5B,$BE
	.dc.b	$DF,$09,$98,$C4,$F1,$AC,$7D,$A4,$8F,$13,$CF,$0F,$F1,$4C,$60,$87
	.dc.b	$94,$EF,$1B,$F0,$BC,$5F,$66,$EE,$A4,$AF,$F7,$87,$DA,$3D,$7C,$47
	.dc.b	$3A,$E0,$FF,$BD,$5A,$F7,$9E,$0E,$6B,$8E,$DF,$12,$15,$C2,$31,$5C
	.dc.b	$27,$B7,$B9,$E6,$23,$E4,$EA,$AE,$8F,$14,$7C,$D9,$FE,$75,$5D,$ED
	.dc.b	$F0,$FF,$00,$97,$FD,$3E,$7C,$9F,$68,$FB,$4C,$CF,$BE,$AD,$6A,$7E
	.dc.b	$D6,$7F,$E2,$68,$F6,$37,$8F,$FA,$CB,$6F,$FA,$E7,$DA,$3D,$3E,$3E
	.dc.b	$1C,$9F,$15,$F0,$8F,$C3,$56,$E7,$27,$F6,$5B,$E8,$FD,$99,$E4,$FD
	.dc.b	$F9,$E0,$FE,$06,$7F,$41,$87,$9B,$A7,$D2,$76,$FA,$6B,$5D,$8F,$C2
	.dc.b	$46,$1E,$91,$F0,$3E,$1A,$46,$ED,$1F,$9D,$F8,$4F,$F3,$1C,$E6,$36
	.dc.b	$F8,$DD,$79,$1F,$A3,$FB,$3F,$CC,$CF,$E5,$E9,$F3,$9C,$CD,$74,$49
	.dc.b	$F3,$57,$0D,$3F,$18,$DD,$FF,$C1,$38,$7F,$A6,$6B,$77,$C5,$FD,$25
	.dc.b	$6D,$FF,$08,$4E,$DC,$FD,$A7,$94,$8F,$C5,$FB,$37,$DA,$62,$36,$79
	.dc.b	$27,$F7,$D3,$A8,$F3,$FC,$77,$BC,$8F,$3F,$B4,$AF,$92,$2E,$9F,$A7
	.dc.b	$A5,$7F,$09,$FB,$37,$DA,$52,$61,$F6,$7E,$D3,$F8,$4A,$9D,$BF,$75
	.dc.b	$AF,$F0,$56,$E1,$FB,$59,$E3,$18,$DF,$FC,$66,$DA,$FF,$1B,$35,$4F
	.dc.b	$F1,$8F,$2D,$F1,$8E,$6F,$E8,$5F,$68,$F8,$EA,$7F,$45,$9D,$61,$FE
	.dc.b	$F4,$FB,$47,$DA,$7F,$40,$6C,$66,$3F,$A6,$7C,$A7,$D7,$DB,$F9,$7E
	.dc.b	$6F,$95,$7C,$E6,$BE,$12,$63,$FE,$D0,$9C,$BF,$9F,$FB,$47,$D1,$7F
	.dc.b	$35,$6C,$66,$31,$F5,$72,$38,$35,$FE,$5D,$F2,$2F,$33,$F9,$F3,$63
	.dc.b	$64,$37,$1F,$D7,$3E,$D3,$F5,$31,$FE,$E4,$F9,$BF,$F2,$FE,$DE,$C6
	.dc.b	$F4,$7F,$99,$7D,$85,$F0,$BF,$EA,$51,$8F,$17,$A1,$23,$FD,$DF,$F6
	.dc.b	$9F,$C8,$9F,$E1,$73,$B6,$D7,$3C,$7F,$AC,$CE,$67,$FA,$F7,$12,$F9
	.dc.b	$47,$F4,$DF,$B4,$FB,$2E,$BE,$D3,$E6,$B7,$93,$FB,$04,$EB,$1F,$E8
	.dc.b	$4D,$D3,$FC,$FE,$DA,$FF,$41,$37,$16,$1F,$C6,$AB,$7D,$7D,$09,$7F
	.dc.b	$C9,$AB,$7E,$57,$F8,$53,$17,$CA,$3C,$6F,$F0,$A7,$D0,$BE,$76,$9B
	.dc.b	$9D,$63,$FC,$C2,$F8,$DE,$49,$1D,$FD,$E7,$B3,$88,$FF,$02,$57,$07
	.dc.b	$95,$F4,$7E,$5D,$70,$7F,$98,$78,$2F,$B4,$FF,$14,$70,$7F,$2A,$9F
	.dc.b	$99,$E7,$FF,$82,$6B,$51,$C2,$B0,$47,$E1,$53,$AB,$11,$FE,$03,$AE
	.dc.b	$2D,$AE,$F3,$F1,$FE,$CB,$F9,$27,$0B,$EC,$DF,$69,$FE,$3B,$6E,$1F
	.dc.b	$16,$BF,$95,$6C,$69,$3D,$1F,$F2,$17,$CB,$7E,$6A,$BF,$C6,$33,$97
	.dc.b	$94,$E9,$FF,$13,$7D,$9B,$C1,$A4,$8F,$BD,$F6,$9F,$B6,$7F,$C3,$D3
	.dc.b	$B7,$FA,$D3,$5C,$CF,$F9,$4E,$78,$63,$FE,$13,$FB,$1F,$E0,$EF,$FB
	.dc.b	$47,$EB,$FA,$7D,$9F,$F5,$2C,$31,$E3,$7F,$B5,$23,$10,$9F,$15,$F2
	.dc.b	$8C,$57,$FB,$6A,$31,$3D,$3D,$8F,$F6,$EB,$6B,$A9,$FE,$E1,$9D,$D7
	.dc.b	$1F,$07,$FC,$AF,$3D,$0F,$F7,$36,$1E,$2B,$E5,$19,$17,$C9,$E9,$5C
	.dc.b	$A7,$64,$BE,$D8,$F8,$AB,$8D,$6E,$7E,$06,$3E,$86,$CE,$F0,$EC,$F0
	.dc.b	$BE,$58,$F9,$6B,$73,$E5,$AD,$9C,$27,$51,$EC,$BE,$CF,$A2,$FB,$61
	.dc.b	$E3,$8D,$1D,$A7,$6F,$84,$ED,$1B,$55,$D9,$B3,$C2,$FC,$75,$F3,$A4
	.dc.b	$7C,$42,$94,$A7,$72,$C3,$EC,$9F,$67,$A9,$FB,$91,$87,$13,$B9,$D4
	.dc.b	$7D,$89,$D3,$8F,$92,$CB,$E4,$6A,$69,$1F,$C1,$78,$CF,$B3,$FF,$49
	.dc.b	$ED,$FA,$79,$E4,$6A,$33,$B7,$9E,$FF,$05,$7C,$7C,$5F,$C2,$5F,$0C
	.dc.b	$3F,$61,$39,$8F,$C8,$19,$BF,$8F,$19,$9F,$04,$F9,$CF,$05,$74,$71
	.dc.b	$1D,$30,$F1,$C7,$CE,$C3,$C7,$18,$8F,$A7,$18,$AE,$4D,$8F,$19,$D3
	.dc.b	$0F,$1D,$71,$8F,$31,$FD,$AD,$C3,$B3,$15,$D1,$F3,$9D,$2B,$B5,$FB
	.dc.b	$1B,$7F,$26,$77,$39,$6C,$7E,$33,$EE,$7F,$DA,$1E,$17,$FB,$1B,$87
	.dc.b	$FD,$8F,$18,$C7,$FC,$FF,$F7,$4F,$C2,$7F,$AB,$47,$FA,$0A,$75,$8F
	.dc.b	$FB,$45,$B7,$CA,$7F,$AF,$1C,$AB,$8D,$6E,$32,$DF,$AD,$F1,$47,$94
	.dc.b	$38,$86,$0F,$BB,$FE,$D7,$5E,$33,$65,$AE,$18,$FF,$BB,$5E,$55,$C7
	.dc.b	$C5,$CD,$70,$C3,$C4,$F1,$3E,$0E,$1D,$A3,$27,$BF,$BD,$F6,$BD,$9B
	.dc.b	$5F,$23,$E1,$AD,$6E,$B4,$49,$D3,$63,$51,$F3,$6B,$8E,$BF,$EF,$AC
	.dc.b	$BF,$D7,$17,$B8,$FA,$F2,$F8,$0F,$11,$EE,$C7,$18,$73,$8C,$B4,$F3
	.dc.b	$38,$C7,$D8,$DB,$F0,$57,$29,$CD,$7E,$07,$EC,$7F,$B2,$37,$B7,$E6
	.dc.b	$FD,$E3,$E5,$FE,$8D,$6A,$FF,$0C,$DC,$7F,$1E,$DC,$3F,$8F,$EA,$C3
	.dc.b	$94,$75,$25,$F2,$C7,$BE,$3D,$9E,$FF,$57,$2B,$F3,$36,$FC,$03,$18
	.dc.b	$78,$AA,$E7,$F3,$6B,$57,$ED,$9C,$BF,$A7,$CD,$1E,$EC,$33,$87,$E5
	.dc.b	$3C,$8F,$F5,$1D,$FE,$A1,$B0,$C7,$ED,$43,$E2,$7F,$26,$EA,$3F,$63
	.dc.b	$F8,$57,$E9,$1B,$D1,$AF,$85,$FC,$DC,$E7,$1F,$DD,$76,$FF,$B5,$9F
	.dc.b	$DA,$37,$2F,$DA,$4E,$D7,$4F,$F3,$3F,$0F,$F5,$9F,$7E,$F1,$9F,$69
	.dc.b	$FA,$8F,$C3,$7D,$F9,$FF,$C0,$0A,$D4,$E7,$8B,$C7,$F4,$D2,$BF,$82
	.dc.b	$5F,$0C,$FC,$33,$FB,$2D,$5F,$21,$FB,$4F,$C4,$BF,$DE,$F8,$7A,$23
	.dc.b	$1E,$FF,$EF,$E1,$F9,$EF,$D4,$C7,$E5,$47,$F6,$43,$3C,$FB,$97,$CC
	.dc.b	$D8,$93,$FD,$AB,$98,$F1,$AE,$91,$8B,$CC,$72,$9E,$06,$EB,$9C,$F8
	.dc.b	$8D,$BE,$1B,$CD,$FF,$2A,$FB,$0F,$F3,$9D,$FF,$27,$7F,$55,$19,$F6
	.dc.b	$FF,$8A,$C6,$27,$F9,$A3,$C3,$0F,$E5,$8E,$5F,$B2,$E6,$BB,$78,$3F
	.dc.b	$C7,$BF,$0A,$FA,$11,$FC,$E9,$B7,$FA,$CF,$81,$F9,$18,$7F,$37,$9C
	.dc.b	$BF,$84,$6F,$41,$CF,$84,$AE,$38,$75,$E5,$FE,$69,$87,$FA,$67,$DA
	.dc.b	$BE,$F7,$FB,$B7,$3F,$93,$2F,$A7,$7D,$9F,$FB,$D2,$73,$3C,$5B,$5C
	.dc.b	$63,$E7,$57,$0F,$95,$EB,$87,$F4,$38,$C5,$F0,$8C,$60,$AD,$CF,$8A
	.dc.b	$B7,$7F,$67,$47,$E0,$BF,$E8,$CA,$FE,$B9,$B7,$F8,$2F,$5F,$E6,$13
	.dc.b	$9B,$F9,$1D,$3F,$D2,$A3,$13,$F3,$E3,$1E,$F7,$D3,$7E,$A6,$53,$1F
	.dc.b	$ED,$77,$FE,$A3,$7B,$C6,$70,$FA,$8F,$5F,$13,$18,$FE,$57,$F0,$AF
	.dc.b	$F2,$1F,$D7,$FF,$63,$6E,$1F,$39,$FF,$23,$F2,$E7,$15,$52,$56,$E3
	.dc.b	$F1,$95,$C9,$FC,$E3,$6F,$C1,$1C,$A7,$85,$F2,$8F,$F0,$07,$E4,$DE
	.dc.b	$B6,$E3,$F1,$7E,$47,$FB,$8C,$E7,$0F,$F0,$6B,$70,$FF,$09,$36,$FF
	.dc.b	$BE,$E8,$BE,$38,$7F,$80,$7C,$FF,$B9,$F4,$77,$8F,$C9,$87,$EC,$E7
	.dc.b	$39,$7F,$BC,$67,$39,$FE,$F6,$73,$5B,$9C,$D7,$F9,$BA,$73,$5B,$9C
	.dc.b	$D3,$5F,$F3,$5F,$95,$F4,$7D,$BF,$E5,$91,$FE,$6E,$F9,$DC,$4C,$CB
	.dc.b	$EA,$56,$84,$E6,$FE,$7C,$66,$BA,$4E,$71,$F4,$7E,$4C,$F9,$FF,$84
	.dc.b	$BE,$D2,$C6,$1B,$97,$EC,$35,$56,$E3,$0F,$36,$CF,$E6,$FC,$5F,$AF
	.dc.b	$EA,$59,$FF,$1F,$72,$7F,$72,$DB,$F5,$1C,$9F,$CE,$1F,$C0,$D6,$DB
	.dc.b	$5C,$3E,$AF,$F0,$27,$E9,$FE,$2E,$C6,$49,$3D,$23,$2F,$E3,$3C,$6F
	.dc.b	$C4,$F8,$FF,$8A,$6D,$79,$1B,$3B,$8F,$3C,$FF,$14,$FA,$F7,$D8,$7F
	.dc.b	$3D,$F8,$1F,$EA,$37,$FE,$AC,$96,$39,$F3,$3F,$D5,$B1,$C3,$1F,$F4
	.dc.b	$F4,$62,$75,$5C,$70,$AA,$F9,$DF,$08,$D3,$5D,$57,$0F,$B0,$FF,$01
	.dc.b	$78,$19,$6C,$FF,$AA,$F1,$F0,$5E,$7D,$7F,$E5,$53,$93,$FD,$8B,$5C
	.dc.b	$33,$F6,$C2,$31,$7C,$A3,$81,$9A,$E7,$5C,$2B,$73,$D2,$73,$5C,$6B
	.dc.b	$51,$89,$D6,$1E,$53,$DD,$97,$C2,$8E,$F8,$74,$7C,$13,$B7,$C7,$3B
	.dc.b	$BE,$B3,$96,$CE,$B2,$F9,$18,$78,$63,$99,$D9,$C5,$F3,$C7,$FD,$CF
	.dc.b	$8F,$84,$93,$AB,$E7,$8F,$7A,$DC,$E7,$2C,$63,$EE,$3B,$57,$BA,$F8
	.dc.b	$C6,$09,$97,$D2,$CB,$C5,$1C,$63,$D6,$4C,$7E,$61,$E0,$BE,$F3,$EC
	.dc.b	$9E,$57,$D9,$E1,$1C,$EB,$8B,$48,$71,$72,$D3,$39,$7D,$3C,$3C,$B1
	.dc.b	$92,$46,$6F,$C1,$5E,$59,$F6,$BD,$27,$E5,$E0,$98,$6C,$A7,$D6,$2C
	.dc.b	$7E,$79,$BF,$60,$FA,$9D,$C7,$E7,$23,$0E,$25,$18,$8E,$33,$C1,$FD
	.dc.b	$04,$70,$CB,$F3,$8D,$C3,$F0,$6D,$9F,$47,$D8,$38,$7F,$DF,$75,$AC
	.dc.b	$3F,$69,$7B,$BE,$B1,$FB,$BB,$E1,$1A,$88,$FE,$82,$78,$4E,$A3,$13
	.dc.b	$AC,$3C,$AF,$3C,$3C,$33,$F0,$63,$58,$78,$E3,$F7,$37,$CE,$F8,$C7
	.dc.b	$42,$3F,$C0,$A3,$53,$C6,$36,$47,$EB,$C6,$A7,$C0,$72,$8E,$06,$AF
	.dc.b	$CB,$1F,$08,$D1,$C2,$BC,$11,$AC,$7E,$BE,$1E,$AA,$D5,$F3,$70,$73
	.dc.b	$BF,$23,$F5,$2F,$96,$3F,$47,$0E,$73,$E0,$18,$FA,$5E,$58,$7A,$1E
	.dc.b	$E7,$33,$BC,$F2,$C3,$BC,$72,$C3,$BE,$3A,$C7,$EC,$C6,$27,$C1,$7C
	.dc.b	$70,$F5,$46,$2B,$CC,$70,$79,$5F,$BB,$1F,$31,$FE,$E5,$8F,$C0,$10
	.dc.b	$FB,$BC,$5F,$3A,$35,$1F,$80,$C7,$06,$67,$F0,$FE,$73,$DA,$7C,$FF
	.dc.b	$EC,$BB,$7F,$D3,$93,$9F,$0B,$1C,$7F,$B7,$8C,$65,$FE,$76,$79,$5E
	.dc.b	$A3,$FD,$8F,$87,$47,$4F,$F9,$B5,$B7,$BA,$E3,$1A,$F8,$FC,$51,$A7
	.dc.b	$5F,$33,$FC,$E3,$CF,$8A,$E0,$43,$FC,$1E,$EF,$AB,$E1,$E9,$EC,$C3
	.dc.b	$FC,$57,$5E,$4F,$07,$5B,$48,$C1,$0F,$F0,$CB,$ED,$F0,$3F,$C4,$BF
	.dc.b	$47,$7C,$1F,$95,$5F,$E2,$6C,$31,$32,$30,$43,$FC,$A1,$28,$F9,$10
	.dc.b	$FA,$A7,$C6,$E1,$B1,$FE,$A0,$6B,$FE,$CB,$38,$7C,$66,$E6,$46,$DE
	.dc.b	$51,$88,$E1,$3C,$B4,$FF,$6D,$47,$86,$B6,$FF,$98,$FB,$3E,$07,$1B
	.dc.b	$CE,$A6,$22,$4F,$F9,$A6,$32,$F1,$3F,$03,$D5,$F7,$08,$FF,$A7,$3A
	.dc.b	$2B,$FE,$E5,$E6,$FC,$A4,$E3,$2F,$F5,$5D,$7C,$6F,$A8,$7D,$6D,$26
	.dc.b	$1F,$EC,$48,$7D,$9E,$23,$E3,$47,$07,$53,$CE,$75,$7F,$0A,$B5,$5C
	.dc.b	$2B,$4F,$E6,$BC,$3F,$CA,$91,$FF,$5A,$F8,$B5,$F5,$72,$27,$18,$7F
	.dc.b	$B1,$2B,$F2,$9E,$2F,$E7,$BB,$A6,$1F,$EB,$E8,$CB,$FA,$88,$CB,$FE
	.dc.b	$6A,$8F,$C3,$FD,$53,$EA,$BF,$13,$F5,$4F,$B0,$A6,$A2,$5F,$FB,$A2
	.dc.b	$74,$D9,$CB,$F8,$BA,$E8,$FF,$BB,$9A,$E6,$BB,$F2,$FE,$8F,$F0,$97
	.dc.b	$B7,$49,$5F,$EF,$AF,$09,$F1,$DF,$0E,$3F,$D1,$D5,$C1,$B5,$C1,$FF
	.dc.b	$77,$4E,$DF,$C2,$7D,$D7,$F7,$A7,$F8,$83,$9E,$0F,$E2,$ED,$DF,$F8
	.dc.b	$B3,$B9,$E3,$CF,$FC,$3E,$35,$0E,$84,$61,$B3,$FB,$4F,$C1,$FF,$CF
	.dc.b	$F5,$FE,$17,$C3,$F4,$F1,$88,$FD,$93,$63,$3E,$87,$1F,$3A,$57,$1C
	.dc.b	$BF,$C3,$F7,$B6,$D6,$A7,$9D,$5B,$E7,$1F,$CB,$FE,$2F,$EC,$BF,$A7
	.dc.b	$FC,$4B,$D0,$7A,$63,$33,$FC,$E9,$7C,$9F,$F2,$F6,$D7,$38,$FE,$65
	.dc.b	$DF,$FE,$0F,$A6,$5F,$E3,$5F,$8B,$7E,$14,$F8,$ED,$F8,$53,$2F,$60
	.dc.b	$AD,$CF,$7F,$3E,$4F,$A5,$F1,$77,$C8,$F6,$35,$65,$E2,$5D,$0F,$A9
	.dc.b	$7E,$29,$FD,$0D,$B3,$F8,$F6,$BC,$B5,$FF,$27,$57,$1A,$E1,$EF,$7F
	.dc.b	$51,$F9,$9F,$81,$5F,$D7,$5B,$7F,$85,$6D,$BE,$4D,$F5,$9C,$78,$25
	.dc.b	$FF,$63,$BD,$D6,$A3,$13,$A7,$F8,$F4,$EA,$BC,$5F,$4D,$27,$93,$FD
	.dc.b	$7E,$DF,$84,$E6,$78,$CD,$DB,$E3,$65,$E5,$AC,$67,$EB,$AB,$7F,$2B
	.dc.b	$0E,$6F,$F7,$9F,$49,$F3,$9F,$DC,$DB,$19,$8F,$EE,$1F,$33,$FC,$B2
	.dc.b	$30,$43,$F9,$C4,$F0,$7E,$81,$F6,$23,$83,$A9,$EC,$47,$F8,$34,$6B
	.dc.b	$3F,$F0,$8D,$3E,$87,$C8,$4A,$DF,$DA,$D7,$C3,$16,$BE,$1D,$AB,$19
	.dc.b	$FF,$86,$70,$E9,$87,$C1,$C3,$DB,$5F,$E1,$3F,$C0,$5F,$29,$49,$5F
	.dc.b	$C6,$FC,$57,$C3,$62,$3F,$AC,$57,$07,$FC,$48,$D7,$F8,$DC,$6F,$4F
	.dc.b	$F1,$6C,$E5,$B7,$FE,$1D,$BE,$B7,$CB,$97,$C8,$47,$C5,$F3,$35,$F0
	.dc.b	$C5,$BF,$C1,$D2,$BF,$C1,$BF,$43,$FE,$0C,$9F,$F5,$87,$37,$D7,$FA
	.dc.b	$2F,$DE,$63,$8E,$2F,$F5,$9C,$62,$13,$F7,$6F,$94,$62,$BF,$D7,$31
	.dc.b	$89,$E9,$DB,$FE,$BC,$6D,$75,$3F,$D8,$13,$BA,$E3,$EC,$7F,$00,$9E
	.dc.b	$87,$FB,$17,$0F,$15,$F2,$8C,$8B,$E4,$F4,$AE,$53,$B2,$5F,$6C,$7C
	.dc.b	$55,$C6,$B7,$3F,$03,$1F,$43,$67,$78,$76,$78,$5F,$2C,$7C,$B5,$B9
	.dc.b	$F2,$D6,$CE,$13,$A8,$F6,$5F,$67,$D1,$7D,$B0,$F1,$C6,$8E,$D3,$B7
	.dc.b	$C2,$76,$8D,$AA,$EC,$D9,$E1,$7E,$3A,$F9,$D3,$FE,$EE,$8F,$B6,$D8
	.dc.b	$CB,$F1,$CF,$F7,$6C,$3D,$A7,$C5,$FF,$BA,$63,$0D,$8C,$CF,$06,$BF
	.dc.b	$EE,$FF,$87,$FF,$78,$3F,$F0,$05,$A9,$A4,$7E,$FF,$D1,$7F,$94,$79
	.dc.b	$FE,$CC,$71,$33,$E2,$FF,$47,$DF,$F8,$94,$65,$FE,$29,$5C,$63,$F1
	.dc.b	$26,$77,$32,$35,$3C,$A3,$C8,$76,$BF,$0B,$88,$EB,$7E,$1B,$E7,$33
	.dc.b	$0F,$C7,$E1,$D2,$73,$3E,$13,$C1,$5C,$5C,$D7,$08,$C4,$73,$3C,$47
	.dc.b	$A8,$F1,$4F,$82,$BB,$47,$23,$77,$EB,$BE,$98,$7A,$9B,$3C,$DB,$3B
	.dc.b	$8E,$37,$E8,$7B,$9E,$8A,$E4,$E8,$FF,$16,$2F,$8C,$FE,$F5,$C1,$FE
	.dc.b	$CC,$8C,$1F,$E4,$5F,$A2,$FF,$24,$BF,$EC,$B3,$35,$FE,$83,$6C,$F1
	.dc.b	$AE,$71,$FD,$F2,$B8,$D6,$E7,$2F,$8B,$C8,$BE,$7F,$F8,$71,$CF,$83
	.dc.b	$FE,$54,$FA,$2C,$E3,$FE,$26,$3C,$53,$83,$7C,$E9,$C4,$E8,$98,$7F
	.dc.b	$96,$23,$2F,$F6,$4F,$A2,$F0,$F7,$9F,$0F,$81,$6F,$FB,$94,$68,$8F
	.dc.b	$F6,$46,$D7,$38,$D9,$92,$BF,$DC,$BE,$4F,$B9,$D8,$B1,$FD,$FD,$B1
	.dc.b	$FD,$DB,$35,$A7,$3B,$FF,$72,$6D,$FF,$8A,$A7,$4F,$33,$FC,$83,$AB
	.dc.b	$E8,$5F,$45,$F4,$1B,$3F,$E8,$3F,$11,$F0,$F9,$A9,$3C,$AB,$7A,$79
	.dc.b	$6F,$73,$9A,$DC,$FF,$B1,$EB,$79,$FF,$88,$AF,$73,$9A,$F1,$7C,$DC
	.dc.b	$C3,$C5,$EF,$DA,$E5,$37,$0E,$53,$6B,$F8,$2E,$7E,$DA,$C5,$FF,$8A
	.dc.b	$EA,$EC,$F9,$57,$C1,$7E,$39,$B8,$FF,$A8,$DF,$F6,$23,$63,$F2,$BE
	.dc.b	$A7,$E7,$A3,$BB,$70,$FF,$30,$D7,$27,$FD,$CD,$19,$C8,$DE,$DF,$F3
	.dc.b	$B5,$FF,$A1,$FA,$3F,$43,$18,$BF,$F1,$AF,$A6,$FA,$2F,$F0,$35,$FF
	.dc.b	$9E,$A3,$1E,$37,$A7,$E7,$3F,$57,$18,$9F,$F4,$97,$D0,$FA,$0F,$D6
	.dc.b	$FD,$13,$8B,$13,$F8,$4C,$DF,$29,$FB,$2F,$A2,$7B,$BE,$48,$C3,$FB
	.dc.b	$8A,$7D,$13,$9B,$0D,$8F,$F4,$79,$AF,$89,$CC,$C8,$CC,$EF,$87,$FC
	.dc.b	$A7,$1B,$33,$3B,$8C,$CE,$FA,$3F,$D2,$15,$B9,$CD,$7F,$0A,$FA,$2F
	.dc.b	$F2,$2C,$FE,$81,$FD,$1C,$FF,$98,$39,$91,$A8,$C3,$AB,$FF,$4E,$4E
	.dc.b	$BD,$1E,$9E,$07,$56,$E1,$FA,$CF,$A0,$F9,$8D,$E1,$FE,$33,$F4,$5F
	.dc.b	$D6,$23,$04,$CF,$C5,$5C,$1F,$DF,$4E,$DC,$D7,$F2,$67,$FD,$F3,$CD
	.dc.b	$7E,$18,$FF,$2A,$F9,$07,$D1,$3E,$0B,$53,$33,$FF,$60,$E1,$C9,$FF
	.dc.b	$61,$DF,$0A,$D4,$FF,$A1,$0D,$1F,$E8,$68,$EB,$F1,$5A,$E0,$FF,$60
	.dc.b	$46,$2F,$FD,$9B,$E1,$3E,$59,$F4,$5F,$CF,$38,$3F,$BE,$61,$FE,$A2
	.dc.b	$9C,$DF,$C9,$EB,$FE,$7F,$18,$CF,$E2,$C6,$30,$E4,$F3,$CA,$5F,$82
	.dc.b	$FF,$78,$FD,$38,$CD,$FC,$E7,$AF,$87,$8C,$7F,$DB,$1F,$41,$FE,$0C
	.dc.b	$FA,$2F,$EA,$6D,$C3,$FD,$CE,$FF,$84,$39,$B3,$8A,$A9,$2B,$71,$F8
	.dc.b	$97,$FC,$2F,$5C,$E3,$FD,$5E,$DB,$F0,$C7,$39,$E3,$7C,$E3,$F2,$DF
	.dc.b	$41,$EC,$6E,$3F,$1B,$E8,$BF,$B3,$CE,$70,$FC,$BB,$70,$FE,$46,$DB
	.dc.b	$FF,$7A,$E8,$BE,$38,$7F,$72,$F3,$DC,$4F,$7B,$0F,$97,$0F,$F6,$3C
	.dc.b	$EA,$73,$B6,$9B,$9F,$F8,$DB,$43,$FE,$38,$AE,$0E,$4C,$61,$D6,$7E
	.dc.b	$9F,$B5,$FC,$AF,$7F,$F2,$CF,$93,$F5,$21,$F3,$76,$26,$1F,$5A,$78
	.dc.b	$1A,$71,$1A,$D3,$E8,$CE,$AB,$18,$79,$A5,$93,$E6,$FF,$D9,$38,$F9
	.dc.b	$E7,$30,$FA,$2F,$90,$EA,$3E,$2B,$99,$E1,$5C,$FE,$0F,$77,$C1,$B3
	.dc.b	$F9,$9F,$1B,$ED,$FA,$7F,$A2,$6C,$64,$FE,$F3,$CB,$FD,$A3,$3F,$D5
	.dc.b	$4E,$B3,$FC,$3F,$C4,$FE,$10,$FD,$5F,$C7,$DC,$38,$C5,$9F,$F0,$67
	.dc.b	$3A,$B1,$87,$E2,$BC,$9F,$C6,$F9,$1F,$8E,$6E,$3E,$26,$E1,$F9,$66
	.dc.b	$D7,$CB,$87,$D1,$3E,$7B,$E7,$7C,$07,$F4,$E9,$FF,$3D,$CF,$87,$0F
	.dc.b	$F4,$77,$9B,$FE,$74,$8E,$11,$89,$DE,$7C,$AB,$58,$59,$CD,$6D,$F0
	.dc.b	$C6,$5F,$1D,$5F,$A2,$7C,$83,$C1,$E1,$18,$79,$65,$E5,$BF,$F5,$06
	.dc.b	$5F,$E8,$8A,$D4,$60,$FF,$4A,$B8,$CF,$DD,$08,$C4,$FB,$71,$F1,$63
	.dc.b	$E1,$AE,$46,$DC,$5F,$2B,$E3,$3E,$E3,$6E,$1E,$18,$FA,$AB,$8E,$1D
	.dc.b	$F3,$C4,$CA,$E3,$5A,$7A,$61,$DE,$31,$1D,$0D,$CF,$C4,$C7,$D2,$F3
	.dc.b	$AE,$0E,$6F,$18,$7C,$7C,$39,$E3,$E7,$CB,$E6,$61,$C3,$2E,$A4,$7E
	.dc.b	$4C,$6A,$F9,$63,$D2,$BE,$7E,$1E,$6A,$D4,$62,$74,$FA,$6F,$9E,$5E
	.dc.b	$88,$ED,$87,$47,$E8,$DF,$27,$84,$66,$BC,$C7,$2C,$BA,$E5,$F4,$4F
	.dc.b	$66,$1D,$67,$75,$C4,$FA,$57,$E0,$3C,$C5,$3F,$2B,$7D,$32,$F8,$E7
	.dc.b	$B6,$FA,$57,$3A,$E3,$7E,$37,$96,$1F,$94,$7F,$DB,$2D,$7D,$C6,$2F
	.dc.b	$F2,$0E,$BE,$51,$F4,$5F,$EC,$98,$FC,$C4,$61,$FF,$66,$52,$31,$1C
	.dc.b	$67,$84,$E5,$FD,$5C,$F1,$BF,$CD,$37,$0F,$F7,$8F,$D1,$38,$FF,$DE
	.dc.b	$51,$FA,$D9,$D6,$3F,$AC,$6B,$E4,$C3,$F7,$D7,$C6,$3B,$C4,$7F,$4B
	.dc.b	$3C,$27,$51,$89,$D6,$1E,$47,$9E,$1E,$29,$F0,$46,$B0,$F2,$3A,$8C
	.dc.b	$1A,$3A,$1F,$BA,$BE,$F7,$D6,$32,$D3,$F8,$2C,$6A,$BA,$CE,$9B,$3A
	.dc.b	$BF,$24,$FC,$13,$47,$86,$34,$F9,$0D,$3D,$E7,$E7,$CF,$2C,$7E,$B5
	.dc.b	$72,$37,$3A,$79,$46,$AB,$94,$05,$FA,$5F,$15,$7A,$08,$EE,$34,$F6
	.dc.b	$36,$43,$9D,$F9,$71,$FB,$F8,$78,$6F,$9D,$68,$F2,$CE,$8F,$BD,$5C
	.dc.b	$71,$C9,$F5,$E7,$DD,$7C,$DF,$06,$1E,$43,$FD,$BE,$EF,$1F,$96,$FD
	.dc.b	$BE,$6F,$A3,$3A,$8F,$BD,$86,$0C,$CF,$E2,$7D,$0F,$F6,$6F,$95,$F1
	.dc.b	$EB,$FD,$21,$ED,$7F,$8B,$9F,$F7,$9B,$5C,$C6,$F9,$3F,$D2,$33,$FE
	.dc.b	$7B,$9D,$F8,$59,$E2,$FF,$6F,$CF,$1A,$DD,$E6,$D8,$7F,$B7,$BE,$8B
	.dc.b	$0D,$8F,$0C,$6F,$CC,$FF,$3C,$73,$EB,$14,$D2,$57,$F9,$BE,$78,$75
	.dc.b	$7C,$3C,$BC,$FE,$0F,$13,$3B,$4B,$F7,$C6,$5F,$F3,$C4,$79,$FC,$0F
	.dc.b	$F1,$87,$D1,$5F,$0B,$FD,$EB,$5F,$E2,$A6,$BF,$E2,$E8,$D1,$BD,$24
	.dc.b	$6E,$3E,$12,$FA,$27,$C6,$61,$B1,$FE,$A4,$6B,$FE,$A1,$5F,$17,$B9
	.dc.b	$93,$B7,$94,$62,$38,$45,$D3,$FC,$45,$5C,$A7,$FC,$87,$C9,$FC,$C2
	.dc.b	$7F,$CA,$19,$FF,$9C,$3C,$EF,$85,$26,$F8,$12,$F8,$C6,$8C,$3F,$8C
	.dc.b	$EC,$FB,$04,$7F,$D4,$9D,$15,$D1,$F9,$08,$FB,$31,$F1,$BE,$85,$F4
	.dc.b	$34,$95,$FE,$C8,$87,$D1,$62,$3F,$D8,$D1,$C1,$D4,$F3,$9D,$69,$FE
	.dc.b	$43,$AD,$57,$0A,$D3,$FA,$2F,$07,$F2,$04,$7F,$D7,$9E,$25,$7F,$53
	.dc.b	$16,$FF,$D9,$53,$F9,$2F,$17,$F3,$1D,$D2,$FF,$D8,$91,$97,$F5,$F1
	.dc.b	$97,$FC,$C3,$1F,$86,$FA,$07,$D0,$7E,$23,$E8,$9F,$45,$4D,$44,$BF
	.dc.b	$F7,$5D,$70,$6C,$71,$E3,$D4,$7F,$BA,$E7,$4D,$9D,$3F,$E8,$6F,$A2
	.dc.b	$7B,$7F,$A3,$6C,$66,$3F,$6D,$F1,$DF,$E5,$88,$C1,$0F,$F5,$AC,$F0
	.dc.b	$5B,$7F,$A7,$7C,$35,$71,$9C,$ED,$FC,$EF,$C4,$7D,$17,$F7,$D9,$FF
	.dc.b	$50,$46,$38,$BF,$84,$CF,$1C,$7F,$D5,$13,$C3,$9F,$26,$AA,$5F,$2B
	.dc.b	$E8,$F8,$0F,$E9,$5F,$45,$FE,$91,$7F,$D5,$C1,$C3,$B3,$1E,$6F,$EF
	.dc.b	$21,$E7,$D8,$93,$CA,$78,$E5,$FE,$93,$77,$19,$9D,$BD,$63,$79,$FF
	.dc.b	$37,$77,$1E,$7C,$DF,$15,$F8,$48,$F8,$BF,$45,$F0,$EF,$E5,$36,$3F
	.dc.b	$98,$CE,$DE,$4B,$E4,$BF,$99,$B6,$3E,$7B,$5F,$07,$77,$F8,$63,$3F
	.dc.b	$F6,$B7,$C5,$3E,$89,$F3,$3F,$CF,$E3,$0D,$8C,$DF,$FB,$67,$91,$EF
	.dc.b	$3E,$BB,$C3,$C9,$E2,$F4,$65,$F8,$A3,$2D,$21,$F6,$6D,$F4,$6F,$A3
	.dc.b	$FE,$94,$D9,$F0,$35,$E0,$DD,$7F,$28,$BE,$03,$FA,$AF,$CD,$7B,$F9
	.dc.b	$3F,$28,$DB,$FE,$B6,$DC,$5F,$28,$FA,$3E,$3E,$54,$BF,$CC,$DE,$EB
	.dc.b	$51,$89,$D3,$F7,$67,$55,$E2,$FA,$29,$3C,$9F,$EC,$D6,$FA,$3C,$CF
	.dc.b	$19,$BA,$FE,$BE,$B5,$3C,$48,$6C,$FA,$8E,$9F,$C7,$79,$CF,$9D,$FE
	.dc.b	$EE,$D8,$FD,$9B,$5F,$A5,$F3,$5F,$CD,$1C,$CF,$02,$3F,$81,$8D,$3D
	.dc.b	$A3,$2D,$35,$E7,$F8,$3E,$8E,$47,$81,$FF,$03,$53,$E8,$F3,$3E,$89
	.dc.b	$B9,$7F,$0F,$AD,$4F,$12,$1B,$3C,$95,$FE,$14,$FA,$37,$CA,$52,$4F
	.dc.b	$E6,$7C,$87,$C3,$E2,$3F,$69,$5C,$1F,$E2,$ED,$7F,$2D,$3C,$1C,$D7
	.dc.b	$1C,$7F,$8C,$D6,$DF,$E4,$D5,$BA,$EB,$CD,$E5,$23,$E3,$F9,$AA,$E8
	.dc.b	$F0,$C7,$F3,$68,$FF,$08,$FD,$17,$F8,$42,$FF,$CD,$1C,$BF,$67,$E8
	.dc.b	$FF,$6B,$82,$31,$E0,$FF,$91,$DF,$BD,$7C,$67,$5C,$5F,$E4,$B8,$C4
	.dc.b	$F0,$9E,$B3,$FE,$4F,$BE,$B3,$CB,$DA,$FE,$87,$5C,$8F,$F3,$D5,$F2
	.dc.b	$BE,$18,$78,$5B,$3D,$0D,$D7,$2B,$E9,$8E,$65,$8F,$82,$F8,$C7,$71
	.dc.b	$8F,$77,$9D,$71,$AE,$11,$FE,$8D,$6B,$C3,$0F,$05,$FC,$16,$C6,$F1
	.dc.b	$F4,$34,$F2,$61,$DD,$E5,$87,$79,$DC,$7A,$1B,$3B,$38,$61,$E2,$3C
	.dc.b	$6E,$6F,$C3,$5C,$4E,$87,$81,$CC,$7D,$98,$FF,$41,$1E,$22,$94,$CB
	.dc.b	$B7,$EF,$7B,$6F,$8C,$FF,$7A,$4F,$F9,$51,$B1,$99,$E7,$39,$7F,$DF
	.dc.b	$43,$E1,$AB,$4C,$B8,$23,$F5,$FE,$8F,$FC,$A5,$DF,$F7,$63,$FD,$81
	.dc.b	$3F,$7E,$73,$1F,$74,$C9,$C7,$C8,$57,$18,$F0,$6F,$FC,$66,$B8,$BF
	.dc.b	$C6,$E3,$8B,$9A,$E3,$1D,$4E,$3B,$FF,$AA,$EB,$8B,$F2,$B0,$F2,$57
	.dc.b	$0B,$E9,$7C,$0F,$F6,$C5,$F0,$7E,$9D,$70,$C3,$CC,$D8,$F0,$36,$B9
	.dc.b	$B7,$0F,$04,$FB,$EB,$8B,$4E,$25,$AE,$2D,$9E,$31,$EF,$9E,$2D,$9F
	.dc.b	$DB,$78,$0F,$A3,$7C,$5F,$C4,$E2,$FF,$18,$E0,$F8,$5C,$3F,$EF,$5B
	.dc.b	$E1,$58,$C3,$FD,$87,$3B,$3F,$C0,$98,$74,$7F,$C7,$E6,$63,$FB,$F4
	.dc.b	$65,$FF,$11,$46,$70,$FE,$7D,$1D,$7C,$F7,$E2,$F9,$AF,$17,$6A,$B5
	.dc.b	$39,$59,$27,$55,$FE,$50,$AD,$4E,$6B,$8E,$7E,$B8,$C5,$6E,$31,$7C
	.dc.b	$AB,$FC,$71,$7C,$A1,$F7,$5E,$7F,$06,$E1,$FD,$96,$30,$72,$2B,$93
	.dc.b	$FB,$8B,$96,$99,$BF,$EF,$31,$E5,$FC,$4D,$F8,$17,$C5,$FB,$17,$C4
	.dc.b	$66,$30,$DC,$3F,$CD,$F3,$A7,$FC,$71,$5C,$9C,$CE,$F0,$FF,$2A,$56
	.dc.b	$DF,$F1,$A5,$6D,$CC,$FC,$7F,$19,$F1,$9F,$5F,$E9,$BF,$BD,$63,$FE
	.dc.b	$12,$8C,$78,$DB,$6C,$64,$FF,$41,$CF,$0C,$3F,$C8,$57,$D2,$73,$7C
	.dc.b	$5A,$7F,$90,$9F,$C3,$7E,$2E,$F8,$9F,$B5,$E7,$F3,$E6,$FF,$45,$CE
	.dc.b	$6F,$81,$FE,$92,$71,$8F,$F9,$2B,$0E,$78,$72,$6D,$70,$9F,$F3,$C5
	.dc.b	$70,$BF,$F5,$17,$A8,$F7,$BF,$14,$FF,$9C,$EF,$FC,$CC,$E5,$BE,$CE
	.dc.b	$8F,$CF,$7B,$3F,$94,$9E,$18,$FF,$9F,$3A,$32,$7E,$8F,$C1,$7D,$6E
	.dc.b	$31,$FF,$55,$C6,$1F,$F5,$6F,$D2,$FF,$8F,$DB,$E0,$F9,$3E,$AB,$F5
	.dc.b	$93,$9C,$7F,$D2,$4B,$EC,$7F,$61,$EF,$7F,$2E,$C3,$9C,$61,$FD,$A7
	.dc.b	$D3,$FF,$A0,$1F,$F5,$DA,$F9,$EC,$ED,$A7,$F8,$34,$BE,$9D,$F8,$0E
	.dc.b	$1F,$F4,$7C,$7E,$EF,$93,$F9,$AB,$B9,$D7,$A1,$B6,$C6,$CF,$DE,$D7
	.dc.b	$47,$1A,$FF,$46,$8E,$AF,$FB,$16,$3A,$1A,$3C,$95,$D5,$C4,$F0,$BF
	.dc.b	$F6,$F7,$CB,$3D,$AF,$E2,$9F,$51,$FC,$36,$FF,$D4,$53,$97,$FD,$BB
	.dc.b	$19,$C7,$FD,$39,$39,$7F,$D8,$11,$9B,$EB,$7D,$3C,$6E,$BE,$C7,$6F
	.dc.b	$A2,$B9,$74,$6E,$3C,$9B,$87,$FB,$A3,$14,$7F,$35,$9F,$F8,$03,$F0
	.dc.b	$57,$E2,$CF,$F3,$9F,$01,$F4,$DF,$EB,$9C,$BF,$DC,$F5,$A6,$D6,$E5
	.dc.b	$F3,$5C,$E3,$0D,$9F,$A3,$7D,$23,$18,$7F,$BD,$63,$13,$F1,$BD,$BF
	.dc.b	$96,$7F,$31,$E0,$7A,$F8,$9D,$12,$FE,$5C,$65,$F9,$7F,$01,$F5,$0E
	.dc.b	$5E,$D5,$AF,$A7,$16,$7E,$9C,$DE,$0F,$F0,$1F,$D1,$7F,$4B,$F8,$5F
	.dc.b	$EB,$8F,$F4,$DC,$ED,$FF,$B4,$F1,$FC,$6C,$E5,$B1,$91,$7D,$1F,$F6
	.dc.b	$94,$6A,$F9,$DF,$46,$C6,$4E,$71,$F8,$EF,$8F,$FC,$7F,$93,$67,$F1
	.dc.b	$AD,$BF,$C5,$B7,$0F,$C1,$3F,$DD,$7D,$37,$E5,$6F,$97,$C3,$70,$FD
	.dc.b	$3C,$EB,$6C,$D4,$9C,$D6,$F1,$F6,$C7,$F7,$1B,$E0,$EA,$BE,$4F,$BE
	.dc.b	$FA,$AB,$F8,$2E,$1B,$5F,$2A,$74,$D9,$D3,$4D,$4F,$0C,$3C,$23,$83
	.dc.b	$FC,$B5,$19,$D7,$F4,$5E,$47,$ED,$8F,$C6,$79,$57,$0F,$DA,$C5,$C3
	.dc.b	$F6,$B3,$6B,$F5,$5F,$81,$7E,$F8,$FC,$A7,$85,$70,$FD,$EC,$5B,$FD
	.dc.b	$EC,$DA,$FE,$0F,$E1,$FF,$06,$F2,$FF,$38,$DC,$3F,$83,$B6,$FF,$82
	.dc.b	$B6,$BF,$75,$D5,$FC,$3A,$7F,$87,$B6,$BF,$86,$B6,$3F,$85,$B6,$7F
	.dc.b	$CB,$B0,$FA,$3F,$ED,$4F,$F9,$DF,$DF,$FF,$14,$39,$F9,$2F,$F3,$1F
	.dc.b	$C1,$7F,$90,$A3,$FC,$59,$5C,$23,$11,$BC,$7F,$CE,$F7,$CA,$B4,$DB
	.dc.b	$E6,$E7,$D8,$B7,$D2,$6A,$FA,$27,$C7,$FF,$79,$8F,$31,$FE,$3D,$8C
	.dc.b	$E9,$E0,$8E,$59,$57,$95,$71,$CF,$CF,$7C,$27,$99,$AB,$E1,$9B,$1F
	.dc.b	$1D,$6E,$F8,$E1,$CF,$3F,$8F,$39,$7D,$D9,$78,$63,$19,$79,$EB,$83
	.dc.b	$72,$EF,$8F,$69,$CC,$F8,$CF,$2D,$72,$7E,$71,$88,$DC,$64,$D6,$5E
	.dc.b	$9C,$3C,$53,$A2,$61,$E0,$8C,$B2,$FA,$39,$69,$32,$F8,$31,$96,$C6
	.dc.b	$72,$F8,$77,$EA,$C3,$C5,$39,$3E,$01,$4F,$5D,$72,$25,$72,$BE,$CD
	.dc.b	$9E,$18,$FB,$B2,$F9,$15,$E9,$BE,$35,$F8,$23,$78,$FC,$07,$EF,$E3
	.dc.b	$E4,$9D,$C7,$CF,$6C,$67,$1F,$04,$6B,$1F,$06,$5D,$E7,$24,$BF,$05
	.dc.b	$6F,$0F,$CC,$47,$FB,$CA,$FF,$20,$D8,$CB,$ED,$35,$EB,$3E,$8F,$FD
	.dc.b	$EF,$1F,$A3,$8C,$3F,$EF,$AA,$FF,$26,$4E,$9B,$19,$9D,$DF,$E8,$1B
	.dc.b	$87,$E6,$1B,$3F,$EA,$7F,$A3,$7C,$8F,$F0,$2B,$E1,$8F,$E6,$EF,$77
	.dc.b	$E3,$8F,$D7,$5F,$08,$D4,$47,$F3,$33,$C2,$75,$18,$9D,$57,$68,$FD
	.dc.b	$9C,$EA,$30,$4C,$3D,$15,$E0,$C3,$CD,$18,$BE,$84,$AF,$E2,$B7,$CE
	.dc.b	$75,$1C,$E7,$F8,$56,$1E,$5B,$E7,$18,$C3,$CB,$7C,$0F,$E4,$B8,$73
	.dc.b	$9D,$1D,$E3,$4F,$CB,$AE,$6F,$C7,$34,$70,$C3,$CB,$1A,$C7,$EC,$C7
	.dc.b	$96,$78,$3A,$7C,$67,$1A,$EE,$AF,$C8,$F0,$8C,$D7,$47,$D3,$7D,$F1
	.dc.b	$FB,$73,$9A,$E8,$7A,$4E,$E7,$33,$27,$79,$E5,$8E,$A7,$94,$E4,$C5
	.dc.b	$76,$3E,$73,$89,$EB,$87,$A0,$7D,$1F,$C8,$71,$85,$89,$5F,$81,$21
	.dc.b	$F8,$87,$EC,$FC,$33,$E6,$1E,$CF,$F5,$CE,$47,$03,$05,$BF,$05,$FD
	.dc.b	$6C,$7F,$82,$A3,$1A,$71,$AE,$11,$D7,$0C,$57,$2C,$7B,$BF,$E8,$E9
	.dc.b	$E2,$DC,$3A,$E3,$E2,$8E,$17,$CB,$E2,$1E,$5B,$E5,$3A,$E3,$EB,$FA
	.dc.b	$BB,$99,$E1,$5B,$C7,$FD,$1F,$7A,$99,$18,$3C,$94,$FA,$7E,$64,$8F
	.dc.b	$F4,$BF,$25,$7E,$14,$ED,$E3,$3E,$09,$F1,$FC,$4C,$9E,$AF,$93,$D3
	.dc.b	$C4,$72,$8B,$B7,$FA,$A2,$7E,$4C,$BE,$93,$D4,$FF,$8E,$8C,$77,$6A
	.dc.b	$3F,$D8,$B1,$FE,$BA,$38,$69,$93,$53,$23,$53,$D2,$FF,$C4,$2D,$E2
	.dc.b	$FF,$0F,$BD,$27,$C0,$D9,$DD,$FD,$EF,$27,$F9,$2F,$83,$79,$CD,$31
	.dc.b	$E1,$DC,$7F,$CD,$33,$A6,$CE,$9F,$A7,$E9,$FD,$CF,$4E,$46,$08,$7D
	.dc.b	$27,$2F,$D4,$8F,$93,$F4,$8F,$97,$BE,$3C,$AF,$8D,$39,$6C,$E5,$F9
	.dc.b	$D0,$9E,$73,$27,$8C,$6F,$5F,$F2,$4D,$73,$9F,$C2,$FD,$47,$E5,$08
	.dc.b	$FF,$B8,$3E,$66,$BA,$3F,$23,$1F,$98,$8F,$CD,$F9,$B2,$7F,$2A,$FE
	.dc.b	$73,$BF,$11,$F9,$28,$B8,$7F,$BC,$27,$F1,$3F,$41,$FE,$DC,$AF,$C2
	.dc.b	$7D,$1F,$F9,$21,$BE,$3E,$BD,$B9,$18,$21,$F8,$FA,$E4,$D8,$DB,$4E
	.dc.b	$BF,$4D,$FB,$2F,$1F,$F6,$77,$27,$F5,$B1,$97,$F5,$F1,$97,$F5,$71
	.dc.b	$FB,$FF,$9A,$7C,$87,$E3,$7E,$8B,$F8,$2F,$67,$FA,$A5,$B6,$F1,$97
	.dc.b	$C2,$3F,$8D,$6F,$FC,$2E,$73,$87,$3D,$E5,$6B,$1F,$E4,$15,$A8,$FD
	.dc.b	$57,$D0,$F6,$4B,$8B,$F6,$AD,$8D,$DE,$0C,$CC,$8C,$90,$FD,$D4,$F1
	.dc.b	$6C,$71,$7F,$93,$7C,$57,$AD,$FE,$60,$B7,$FE,$42,$D8,$DE,$5F,$C9
	.dc.b	$3D,$37,$D3,$7F,$26,$F8,$CF,$F6,$D5,$7F,$31,$6B,$FC,$EE,$76,$FF
	.dc.b	$35,$5F,$4D,$CC,$91,$F1,$CF,$A1,$F1,$6F,$8E,$FE,$67,$F1,$FE,$7E
	.dc.b	$2F,$E7,$5E,$7A,$1E,$0E,$8D,$EB,$FC,$F7,$E4,$9F,$79,$8C,$FF,$A3
	.dc.b	$5F,$3E,$6F,$AB,$5A,$9F,$17,$32,$1F,$05,$CF,$09,$5A,$9C,$BC,$3E
	.dc.b	$51,$BF,$F4,$AF,$C5,$9E,$4D,$3F,$A8,$4E,$AB,$85,$65,$FE,$B5,$5C
	.dc.b	$1C,$90,$FE,$CC,$E5,$B5,$D2,$7A,$F5,$7F,$5A,$F9,$C7,$C5,$D3,$37
	.dc.b	$E7,$9C,$3F,$DA,$67,$6E,$48,$7F,$75,$72,$D9,$E6,$B8,$FF,$49,$F3
	.dc.b	$0F,$17,$FB,$5F,$6F,$E1,$27,$EC,$B7,$8B,$F4,$B5,$BB,$FA,$2D,$7F
	.dc.b	$C0,$93,$B7,$FC,$12,$DE,$24,$EE,$FC,$38,$78,$3E,$1A,$61,$D2,$3F
	.dc.b	$C0,$1F,$20,$FB,$1F,$EF,$FE,$83,$B6,$9A,$89,$7F,$E0,$D9,$D3,$67
	.dc.b	$2F,$C1,$D7,$F7,$F7,$D6,$34,$43,$47,$F0,$D7,$4E,$34,$7C,$EF,$EA
	.dc.b	$23,$FC,$49,$F0,$2F,$92,$FF,$0A,$B6,$3F,$2F,$3B,$DB,$F8,$A3,$71
	.dc.b	$FF,$14,$D6,$EB,$9D,$FF,$88,$6B,$51,$FC,$7D,$BD,$CA,$E1,$87,$EC
	.dc.b	$1B,$87,$F1,$CA,$E3,$8F,$F2,$6F,$A8,$FF,$14,$FD,$1B,$E4,$69,$FC
	.dc.b	$C6,$26,$7F,$E3,$79,$D3,$67,$2F,$F8,$4F,$E8,$5F,$27,$F8,$D8,$74
	.dc.b	$3E,$33,$FC,$99,$0F,$A2,$FF,$00,$3A,$8C,$C6,$27,$84,$FF,$94,$EF
	.dc.b	$AE,$FF,$E5,$59,$CC,$A7,$2F,$33,$FC,$B5,$5C,$EF,$AF,$B3,$E2,$BE
	.dc.b	$37,$C1,$B8,$FA,$5E,$B7,$C7,$0E,$77,$C7,$0F,$5E,$1C,$5B,$87,$17
	.dc.b	$BD,$79,$DA,$F9,$4F,$44,$6D,$F7,$35,$F8,$D5,$DD,$F0,$9C,$CD,$E3
	.dc.b	$5E,$6D,$C3,$D1,$1C,$8D,$61,$E7,$78,$9E,$38,$D9,$C2,$35,$1D,$0D
	.dc.b	$57,$8D,$E8,$7A,$E3,$67,$09,$E7,$3B,$8C,$DF,$E1,$6B,$F0,$67,$52
	.dc.b	$C7,$DA,$6C,$6D,$F9,$FD,$6F,$90,$FF,$80,$35,$FC,$A1,$B1,$99,$EA
	.dc.b	$7E,$15,$C3,$83,$F5,$05,$29,$9E,$2C,$38,$39,$BF,$3F,$E8,$BF,$89
	.dc.b	$71,$FD,$19,$FF,$54,$4E,$79,$FF,$D3,$5B,$FE,$5A,$72,$7E,$06,$F8
	.dc.b	$BF,$83,$33,$1F,$9C,$8C,$BC,$63,$CE,$E6,$7A,$EF,$E4,$9E,$91,$87
	.dc.b	$E6,$DF,$8F,$0F,$34,$72,$AF,$D9,$E1,$D9,$E5,$87,$5A,$DD,$F1,$BF
	.dc.b	$31,$C4,$FF,$63,$5F,$17,$B9,$E2,$9E,$2D,$38,$CF,$3A,$F6,$36,$FC
	.dc.b	$27,$93,$0F,$55,$7F,$08,$9D,$3D,$AB,$CC,$FE,$FF,$8F,$F9,$9F,$C1
	.dc.b	$7F,$B2,$78,$7F,$C2,$39,$7F,$85,$4F,$E6,$31,$97,$FD,$AE,$66,$BF
	.dc.b	$9E,$72,$A7,$51,$FE,$E2,$9D,$57,$F3,$89,$CF,$37,$FB,$A3,$CA,$DE
	.dc.b	$2E,$9F,$F7,$56,$1F,$ED,$FF,$A2,$E2,$FF,$50,$CB,$8C,$61,$FE,$9C
	.dc.b	$62,$F9,$D7,$28,$FF,$78,$4E,$EF,$B4,$6E,$FF,$C4,$75,$BC,$BF,$C6
	.dc.b	$37,$B8,$7D,$17,$A3,$B2,$DF,$08,$C1,$E0,$2B,$93,$FB,$63,$96,$99
	.dc.b	$C3,$FB,$3E,$6A,$EB,$1F,$D9,$FC,$07,$C2,$7C,$8F,$A2,$FF,$3E,$E1
	.dc.b	$FE,$0A,$8C,$79,$BF,$2A,$B7,$3F,$E4,$FC,$57,$C3,$0E,$98,$70,$AE
	.dc.b	$46,$E7,$5A,$FF,$87,$A7,$55,$FD,$E7,$C9,$7C,$97,$DA,$FA,$27,$CB
	.dc.b	$FD,$A9,$CE,$5E,$19,$CB,$FE,$49,$AE,$15,$B6,$C6,$4E,$11,$96,$9F
	.dc.b	$E5,$37,$F1,$3F,$44,$F9,$4F,$C4,$F9,$FD,$B2,$F9,$73,$9B,$E0,$7C
	.dc.b	$37,$11,$FE,$13,$9E,$1A,$7F,$9C,$27,$2F,$FA,$86,$33,$4A,$6C,$BD
	.dc.b	$CF,$F9,$BB,$1F,$2E,$3F,$9C,$F3,$5E,$E5,$FA,$2C,$36,$FF,$CE,$72
	.dc.b	$9F,$D2,$FA,$9F,$97,$8C,$4F,$F9,$F3,$AB,$27,$EA,$BB,$FF,$CE,$31
	.dc.b	$93,$FD,$25,$E8,$F8,$7E,$43,$F5,$FE,$C7,$F8,$2E,$7B,$C6,$3C,$3C
	.dc.b	$FD,$17,$ED,$3D,$B5,$60,$FF,$55,$9F,$B8,$FA,$2F,$F4,$0B,$F8,$D5
	.dc.b	$F4,$59,$D7,$4F,$F0,$29,$7D,$13,$E3,$F0,$BC,$FB,$18,$6C,$FF,$A9
	.dc.b	$CE,$2E,$3A,$3F,$C9,$B1,$83,$55,$D5,$B3,$BE,$AF,$5D,$7B,$2B,$C1
	.dc.b	$5F,$EA,$2F,$94,$7D,$17,$F1,$4F,$A2,$FE,$25,$39,$8C,$78,$FF,$C5
	.dc.b	$AB,$73,$FA,$DA,$DE,$1F,$E5,$D9,$CB,$FB,$C8,$CE,$3F,$EB,$C7,$FC
	.dc.b	$BD,$5C,$7E,$23,$C1,$EC,$78,$5B,$19,$3B,$46,$5F,$F6,$55,$BE,$4D
	.dc.b	$F2,$1F,$CC,$23,$0F,$F3,$2E,$E7,$D1,$3E,$97,$57,$C3,$0E,$6D,$C3
	.dc.b	$FC,$F2,$72,$8F,$F6,$C5,$F2,$AE,$38,$49,$CE,$1F,$ED,$67,$9C,$78
	.dc.b	$63,$17,$F0,$FE,$8F,$E7,$1F,$2F,$D6,$B7,$F2,$E2,$CF,$CB,$9B,$B7
	.dc.b	$C9,$F8,$5F,$AC,$7D,$3F,$6A,$E1,$F4,$E2,$DF,$D3,$9B,$5F,$82,$FA
	.dc.b	$1F,$DD,$7B,$BF,$D5,$9F,$DE,$65,$6F,$FC,$F9,$8F,$FB,$DE,$72,$D8
	.dc.b	$C9,$FE,$9E,$8C,$B4,$CD,$F7,$55,$D6,$FB,$36,$FA,$3F,$91,$FA,$3F
	.dc.b	$C9,$79,$16,$7F,$1C,$DB,$FC,$63,$70,$FC,$63,$FD,$C3,$D3,$7D,$1B
	.dc.b	$E1,$30,$DC,$3F,$8D,$5F,$E9,$2B,$7B,$6A,$E5,$68,$FE,$41,$3D,$23
	.dc.b	$9C,$EB,$1F,$85,$EF,$3E,$8F,$E6,$3F,$EE,$5C,$3E,$53,$96,$FC,$2E
	.dc.b	$A7,$6D,$8C,$9C,$23,$2D,$33,$5F,$AD,$F1,$BF,$70,$7E,$D7,$C9,$B3
	.dc.b	$FB,$58,$B3,$FB,$59,$BA,$7E,$CF,$E5,$BF,$80,$1F,$BD,$F0,$AE,$1F
	.dc.b	$BD,$8B,$7F,$BD,$9B,$5F,$C3,$3D,$C7,$81,$FC,$46,$BF,$80,$79,$9F
	.dc.b	$C1,$FC,$5F,$E1,$8D,$CB,$F8,$43,$67,$71,$FC,$2F,$BB,$F8,$94,$7F
	.dc.b	$13,$6C,$64,$FE,$21,$19,$69,$FC,$13,$E8,$BF,$B8,$C7,$FB,$77,$DE
	.dc.b	$FF,$64,$C3,$E8,$FF,$91,$FC,$17,$F8,$DE,$BF,$AE,$4F,$08,$C4,$23
	.dc.b	$85,$FF,$9D,$AF,$9F,$B1,$C9,$EB,$85,$BE,$73,$5E,$51,$7E,$8D,$F2
	.dc.b	$BF,$DE,$67,$CC,$7F,$6D,$8C,$E7,$DE,$3C,$B5,$C6,$B8,$67,$E6,$BD
	.dc.b	$D7,$20,$E2,$3D,$79,$FA,$A7,$53,$9C,$BF,$DD,$AD,$8F,$15,$F6,$9F
	.dc.b	$6E,$7E,$39,$EB,$39,$77,$96,$6F,$94,$7C,$0A,$C5,$6E,$38,$D6,$0E
	.dc.b	$98,$79,$71,$F6,$56,$F0,$E6,$74,$9D,$63,$E8,$9D,$12,$F9,$E5,$F1
	.dc.b	$DE,$D8,$FC,$09,$F1,$5F,$28,$F4,$99,$9D,$E3,$F2,$2F,$94,$E4,$91
	.dc.b	$9C,$B7,$8F,$8E,$ED,$7D,$89,$AF,$C0,$38,$CF,$CC,$69,$E2,$AF,$83
	.dc.b	$5E,$AA,$E3,$97,$C2,$CB,$D3,$8F,$92,$F7,$87,$23,$E8,$16,$B8,$61
	.dc.b	$CE,$31,$7F,$1D,$FC,$46,$5F,$0D,$F3,$5F,$42,$65,$E7,$6C,$7C,$83
	.dc.b	$FD,$E7,$87,$E4,$9B,$7F,$9D,$6E,$1F,$EE,$99,$FF,$21,$FC,$65,$F6
	.dc.b	$0D,$25,$7F,$BC,$A7,$6F,$E6,$27,$6D,$9F,$CA,$7D,$1B,$E7,$3F,$D8
	.dc.b	$19,$FE,$72,$F7,$9F,$E9,$DB,$19,$BE,$35,$C3,$0F,$D9,$56,$9B,$3F
	.dc.b	$9D,$8C,$43,$0F,$03,$99,$8F,$94,$FD,$A4,$69,$C6,$1E,$39,$DB,$FC
	.dc.b	$42,$74,$E6,$77,$87,$9D,$E7,$87,$9D,$FD,$FE,$1E,$4C,$7E,$3D,$F1
	.dc.b	$3A,$56,$E3,$37,$CE,$74,$FF,$9F,$6F,$D9,$5B,$BE,$84,$C7,$E3,$CF
	.dc.b	$A2,$B9,$11,$FF,$43,$63,$EA,$8C,$47,$9A,$78,$3A,$BE,$D1,$88,$D1
	.dc.b	$EE,$31,$8F,$A1,$FA,$66,$4F,$8A,$63,$1F,$5B,$D4,$E3,$8F,$A5,$E1
	.dc.b	$1A,$7B,$B7,$1F,$44,$73,$C7,$D3,$7E,$18,$F8,$A7,$4C,$3D,$15,$C2
	.dc.b	$3F,$09,$F4,$7F,$56,$30,$47,$06,$72,$93,$F9,$A8,$FC,$97,$BA,$FA
	.dc.b	$17,$B3,$F5,$7A,$3F,$CF,$72,$FA,$3F,$D1,$E1,$FE,$EA,$6C,$EA,$F8
	.dc.b	$CF,$FB,$8D,$AE,$F3,$E9,$C9,$FE,$BD,$C3,$A4,$62,$7F,$D7,$F5,$C2
	.dc.b	$FF,$C0,$1D,$0F,$A0,$D6,$1E,$7F,$A3,$B6,$79,$75,$FF,$81,$AA,$DF
	.dc.b	$1C,$7C,$95,$C0,$CC,$F2,$FA,$39,$3E,$97,$8F,$1E,$21,$9F,$BE,$70
	.dc.b	$F9,$A7,$C7,$F0,$D2,$BE,$13,$E2,$F4,$F1,$1C,$B7,$F8,$5A,$AB,$38
	.dc.b	$7C,$0B,$F8,$96,$FA,$3F,$53,$FE,$88,$7F,$D7,$CF,$FA,$59,$AE,$73
	.dc.b	$E3,$1C,$2B,$FD,$6B,$3C,$FB,$F2,$6A,$A4,$73,$8C,$BD,$63,$EB,$79
	.dc.b	$3F,$C8,$7C,$1B,$CF,$FE,$CD,$AF,$F2,$D0,$9D,$38,$7F,$94,$78,$7C
	.dc.b	$2A,$4E,$6B,$93,$F1,$4E,$4A,$7E,$FF,$A7,$F8,$0F,$98,$95,$F8,$F8
	.dc.b	$CB,$F9,$18,$CB,$CF,$9B,$F3,$D3,$F3,$3E,$89,$F4,$77,$C7,$97,$FE
	.dc.b	$C0,$9C,$B6,$72,$FF,$AE,$21,$3C,$E6,$4F,$18,$DE,$FF,$42,$B9,$CF
	.dc.b	$8F,$8D,$F4,$7F,$9A,$24,$7F,$B9,$FE,$66,$BB,$8B,$7F,$A4,$9F,$D0
	.dc.b	$F9,$92,$BF,$60,$FE,$83,$B7,$11,$DE,$2D,$7E,$B2,$7F,$1B,$F4,$4F
	.dc.b	$73,$F1,$FF,$47,$FE,$48,$C3,$FD,$C2,$D7,$FD,$AF,$3B,$7F,$DC,$AD
	.dc.b	$73,$3D,$0E,$7F,$41,$63,$71,$5F,$F7,$0D,$27,$F7,$7E,$47,$EF,$3C
	.dc.b	$A9,$5F,$C2,$E3,$2F,$F0,$D8,$CB,$F2,$E3,$F8,$0F,$D1,$3E,$4F,$F2
	.dc.b	$FF,$47,$FC,$5B,$C1,$FF,$56,$B6,$DE,$32,$F8,$47,$E6,$A3,$17,$CB
	.dc.b	$69,$39,$AD,$BF,$EF,$AF,$A3,$E8,$97,$17,$EE,$9B,$1F,$A6,$35,$59
	.dc.b	$35,$32,$35,$3D,$09,$19,$8F,$04,$FF,$24,$F8,$AF,$1F,$17,$F3,$87
	.dc.b	$E6,$4F,$ED,$3A,$91,$88,$FD,$23,$C3,$5F,$E6,$9E,$9B,$E8,$FF,$97
	.dc.b	$FC,$6F,$FB,$76,$F9,$35,$FD,$64,$ED,$F9,$1F,$16,$F8,$EF,$E6,$F2
	.dc.b	$FA,$3A,$7F,$36,$9D,$57,$97,$E4,$3C,$4B,$44,$3C,$2F,$E8,$2D,$9E
	.dc.b	$1C,$7F,$D0,$3E,$4D,$F4,$7A,$89,$3F,$6E,$72,$DF,$7F,$85,$48,$C1
	.dc.b	$F8,$1F,$96,$67,$FD,$43,$E8,$FF,$89,$4F,$CB,$6B,$FC,$5F,$A5,$5C
	.dc.b	$23,$95,$71,$F2,$24,$60,$87,$F1,$5B,$E7,$CD,$FD,$C1,$BC,$A9,$FC
	.dc.b	$BE,$1D,$5F,$03,$F9,$BE,$07,$CE,$BE,$2F,$FB,$3B,$6B,$D9,$D3,$FD
	.dc.b	$9A,$F8,$4E,$67,$F9,$33,$72,$FF,$78,$4F,$E9,$F5,$F6,$DF,$4C,$26
	.dc.b	$1D,$0F,$D4,$7C,$D3,$C5,$FE,$EF,$D9,$F8,$7C,$BF,$C1,$6D,$D7,$D9
	.dc.b	$92,$F8,$61,$CB,$0E,$3F,$0B,$23,$04,$3F,$9E,$61,$CD,$FC,$43,$6F
	.dc.b	$A5,$7F,$81,$3E,$45,$F4,$7F,$E0,$AF,$45,$D9,$4D,$44,$C7,$F5,$13
	.dc.b	$A6,$CE,$5F,$C0,$57,$47,$24,$33,$5F,$E1,$4F,$9E,$FD,$34,$7F,$53
	.dc.b	$F8,$17,$C9,$7F,$87,$1B,$1F,$C4,$67,$7C,$0E,$0F,$8F,$18,$7F,$C5
	.dc.b	$55,$C2,$BA,$5F,$7F,$0A,$47,$F0,$19,$D6,$1F,$E3,$C6,$DF,$4A,$FE
	.dc.b	$53,$F4,$7F,$E2,$EF,$A3,$7C,$85,$33,$8F,$F2,$19,$C3,$FC,$36,$76
	.dc.b	$E4,$FE,$3D,$F4,$4F,$93,$FF,$14,$C7,$E5,$BE,$33,$FD,$05,$F4,$7F
	.dc.b	$E0,$0E,$C4,$63,$DB,$FF,$43,$46,$2B,$F4,$58,$73,$DF,$F4,$93,$99
	.dc.b	$4E,$5E,$C7,$EA,$70,$ED,$87,$58,$E2,$FF,$A4,$E7,$85,$73,$6C,$6E
	.dc.b	$79,$8A,$E6,$F9,$31,$F6,$DF,$16,$D7,$18,$F3,$57,$0A,$F3,$D6,$9E
	.dc.b	$58,$79,$1F,$46,$3F,$0B,$0E,$47,$9E,$B8,$4F,$2A,$E0,$F9,$6B,$83
	.dc.b	$C8,$CE,$3E,$C7,$B9,$EC,$C3,$C8,$79,$63,$51,$D4,$D4,$F4,$7B,$1D
	.dc.b	$2F,$C7,$3C,$2B,$9D,$6E,$39,$9F,$56,$BF,$D8,$47,$52,$B9,$7D,$D7
	.dc.b	$A7,$F0,$3D,$6F,$8F,$C4,$7E,$02,$30,$EE,$BF,$91,$4E,$9B,$1C,$07
	.dc.b	$0D,$69,$97,$04,$7F,$03,$F4,$7F,$C2,$38,$BE,$7C,$7F,$35,$9F,$A3
	.dc.b	$39,$8F,$C3,$99,$38,$F3,$FE,$ED,$C9,$E0,$DF,$F3,$53,$C5,$FC,$E4
	.dc.b	$71,$73,$3C,$5C,$C7,$1D,$FF,$DD,$95,$C5,$F9,$17,$E4,$BE,$11,$8A
	.dc.b	$F1,$E1,$EA,$79,$1F,$D3,$EF,$93,$F5,$AF,$94,$ED,$B1,$EA,$6E,$1E
	.dc.b	$46,$DF,$78,$F4,$D7,$1A,$F0,$9C,$4B,$1C,$5B,$3C,$4F,$6E,$1E,$A8
	.dc.b	$FE,$C9,$F8,$47,$F9,$AA,$FF,$D2,$3A,$BF,$25,$78,$FF,$3E,$3F,$D4
	.dc.b	$35,$BB,$FF,$47,$46,$E7,$57,$FD,$BE,$55,$BA,$E5,$7D,$2B,$FC,$75
	.dc.b	$39,$7F,$BB,$4F,$37,$FC,$7D,$A3,$E5,$7A,$E1,$F3,$5B,$F1,$EF,$C9
	.dc.b	$5F,$57,$AA,$5F,$1C,$39,$E1,$CB,$1F,$77,$CC,$6E,$B5,$7E,$CD,$13
	.dc.b	$9B,$E1,$18,$C3,$99,$F2,$1C,$57,$FA,$C4,$78,$CF,$A2,$F2,$36,$BF
	.dc.b	$D7,$FF,$56,$FC,$DF,$FC,$7D,$1F,$E5,$D9,$D5,$F3,$6E,$9F,$A1,$F1
	.dc.b	$1F,$9B,$FF,$62,$BE,$19,$7C,$69,$CF,$85,$E0,$8C,$56,$F1,$E9,$18
	.dc.b	$9F,$2F,$C8,$4C,$3F,$DA,$63,$F2,$67,$26,$70,$F2,$57,$D4,$AD,$73
	.dc.b	$E7,$4F,$DB,$46,$0F,$64,$EF,$0F,$DF,$4F,$FB,$B3,$54,$E6,$FE,$D7
	.dc.b	$83,$FE,$FD,$C3,$FD,$00,$DC,$7F,$DB,$DB,$7C,$2F,$1F,$7F,$1F,$2F
	.dc.b	$FD,$D1,$C0,$FB,$8C,$61,$C2,$31,$3D,$2B,$E2,$D6,$B5,$F9,$D5,$AB
	.dc.b	$FB,$97,$FE,$45,$3F,$D2,$45,$7F,$DE,$0F,$C7,$FA,$7F,$C9,$39,$AF
	.dc.b	$E3,$13,$FE,$13,$F0,$BE,$47,$94,$FB,$4D,$C6,$1B,$7F,$4A,$74,$F8
	.dc.b	$27,$4F,$FA,$8F,$C2,$7C,$6F,$E8,$5B,$1F,$0B,$EB,$31,$32,$3F,$63
	.dc.b	$C7,$72,$FC,$C7,$83,$F4,$A7,$27,$ED,$0F,$F3,$57,$13,$D1,$79,$7F
	.dc.b	$AF,$AE,$17,$F5,$27,$37,$FB,$EF,$98,$FD,$A7,$D6,$BE,$4F,$84,$61
	.dc.b	$93,$FE,$BD,$9F,$DD,$B6,$38,$F8,$BF,$BD,$FC,$A7,$F5,$B3,$FE,$C3
	.dc.b	$FC,$FB,$FC,$D0,$FF,$B1,$63,$39,$7E,$CE,$72,$FE,$D7,$E1,$7E,$5B
	.dc.b	$79,$7F,$85,$FD,$37,$FB,$AF,$0F,$F6,$75,$6B,$9F,$EC,$B6,$3F,$D1
	.dc.b	$9E,$2B,$F2,$2F,$F4,$FC,$70,$C3,$F8,$BC,$E7,$DF,$FE,$29,$7C,$72
	.dc.b	$4E,$6F,$81,$FC,$13,$0E,$73,$F8,$CF,$AB,$F9,$13,$FB,$45,$E9,$FE
	.dc.b	$C5,$CC,$EE,$31,$C1,$96,$D7,$17,$87,$C3,$48,$E9,$EC,$5B,$F0,$4D
	.dc.b	$DB,$F7,$78,$FF,$B7,$6F,$C5,$3F,$CB,$FE,$6B,$F8,$AC,$7F,$41,$6D
	.dc.b	$FF,$08,$E0,$FE,$7C,$FF,$AC,$E3,$29,$DF,$92,$9F,$F7,$15,$7F,$BF
	.dc.b	$E7,$31,$8A,$FE,$5D,$18,$F8,$8F,$17,$99,$F6,$9F,$E8,$FF,$51,$FD
	.dc.b	$4F,$0F,$CD,$5A,$3F,$D8,$98,$EA,$3F,$D8,$D3,$A9,$CB,$63,$23,$4F
	.dc.b	$F0,$94,$EA,$31,$3F,$EC,$F8,$F0,$46,$0F,$01,$FC,$AA,$3A,$7A,$BE
	.dc.b	$13,$C5,$E2,$D8,$C9,$27,$95,$7F,$B6,$E7,$51,$FD,$3F,$E5,$7E,$A3
	.dc.b	$78,$3F,$B3,$F9,$6F,$C4,$3F,$43,$9F,$F1,$2B,$DB,$6B,$FB,$6F,$AA
	.dc.b	$FC,$61,$F3,$9F,$DA,$EF,$7A,$FF,$87,$67,$2F,$CC,$CD,$3A,$7F,$B8
	.dc.b	$CE,$9F,$24,$FD,$2B,$ED,$7F,$47,$EE,$64,$FC,$FF,$97,$B7,$F0,$04
	.dc.b	$E3,$3F,$A3,$09,$C5,$7D,$1F,$7B,$FD,$80,$FD,$B7,$F7,$93,$F2,$9C
	.dc.b	$B7,$E5,$BF,$0F,$18,$BF,$BA,$F0,$6D,$6A,$7F,$C1,$DF,$44,$F5,$29
	.dc.b	$F2,$BF,$11,$FC,$E5,$7C,$59,$D3,$67,$FC,$25,$F2,$8F,$A2,$E3,$3F
	.dc.b	$E3,$28,$C3,$B3,$11,$AC,$FF,$C6,$33,$A7,$D9,$F5,$BE,$AF,$95,$F5
	.dc.b	$B6,$FE,$13,$67,$71,$F1,$67,$E0,$FB,$3F,$0B,$ED,$AC,$FC,$96,$D7
	.dc.b	$C9,$6E,$DF,$19,$FF,$18,$7D,$4F,$C0,$9F,$F1,$12,$FA,$5C,$F8,$3A
	.dc.b	$6E,$1F,$89,$9D,$36,$74,$FF,$3B,$9F,$E8,$06,$2B,$FC,$5D,$E6,$F0
	.dc.b	$6C,$7F,$88,$7F,$10,$FE,$1B,$39,$CB,$FC,$51,$1F,$C4,$31,$7D,$67
	.dc.b	$FA,$2F,$DA,$F3,$62,$12,$FA,$CD,$65,$FE,$5D,$8E,$39,$FF,$A6,$63
	.dc.b	$13,$AA,$E5,$5A,$7F,$D4,$F3,$AA,$E1,$1A,$C3,$83,$E0,$BB,$F5,$8F
	.dc.b	$A1,$FF,$18,$CF,$23,$CD,$19,$F1,$B9,$C7,$1C,$2C,$E6,$38,$07,$19
	.dc.b	$FC,$08,$C5,$F6,$8C,$67,$F0,$6F,$9C,$E7,$3F,$F3,$F6,$1C,$9B,$19
	.dc.b	$C3,$E0,$E5,$EA,$8C,$47,$6C,$B9,$9E,$8C,$7B,$47,$9B,$0E,$8F,$BA
	.dc.b	$B8,$C6,$23,$DC,$6E,$32,$4A,$E9,$7E,$17,$BC,$ED,$CE,$38,$CB,$E1
	.dc.b	$61,$D4,$ED,$8F,$92,$33,$8F,$5C,$39,$1F,$32,$75,$97,$C2,$C7,$A3
	.dc.b	$67,$59,$78,$DE,$38,$74,$AF,$1C,$62,$B8,$4E,$63,$8E,$5D,$1F,$89
	.dc.b	$55,$F8,$73,$B2,$57,$23,$F2,$11,$D4,$E2,$53,$8B,$B7,$31,$C9,$FB
	.dc.b	$75,$DF,$2F,$8D,$97,$DC,$CB,$E0,$D7,$0C,$39,$3F,$7C,$F4,$9D,$EB
	.dc.b	$F2,$8F,$FB,$1C,$D1,$5C,$BF,$24,$C7,$D0,$FF,$AB,$1E,$07,$E6,$DC
	.dc.b	$3A,$F8,$4C,$B9,$6C,$A7,$96,$5F,$80,$6D,$FE,$71,$B7,$FC,$0F,$EB
	.dc.b	$1F,$53,$FE,$DD,$BE,$18,$FE,$BA,$F7,$7D,$CF,$D7,$DF,$07,$53,$1F
	.dc.b	$CF,$4F,$09,$D4,$62,$75,$5D,$A3,$9D,$F8,$70,$F3,$56,$9B,$3A,$30
	.dc.b	$7F,$06,$C3,$CB,$8F,$C1,$AF,$E0,$F8,$F8,$31,$EC,$4C,$39,$BF,$C1
	.dc.b	$63,$51,$82,$61,$D6,$B9,$E3,$E5,$8C,$1E,$B2,$5F,$38,$D4,$7C,$78
	.dc.b	$F0,$3C,$1D,$3E,$A3,$55,$E0,$57,$E4,$78,$46,$70,$F7,$47,$C3,$3C
	.dc.b	$31,$9C,$7E,$86,$1E,$2A,$DE,$1E,$2B,$EE,$7E,$06,$75,$5F,$28,$98
	.dc.b	$FC,$09,$F3,$5F,$2C,$3C,$D8,$F9,$67,$99,$27,$4F,$53,$8D,$F4,$7D
	.dc.b	$18,$FD,$1F,$AC,$F8,$F7,$CC,$8E,$0C,$E5,$2B,$EE,$C7,$CE,$FC,$F1
	.dc.b	$E5,$7F,$5D,$9F,$F7,$4F,$1F,$FB,$53,$93,$FB,$2F,$0E,$5F,$F6,$BD
	.dc.b	$F0,$AF,$F1,$F3,$5F,$F2,$E0,$71,$1C,$2F,$95,$77,$FC,$85,$F7,$F8
	.dc.b	$C3,$FB,$74,$3C,$EC,$42,$36,$FF,$93,$5A,$F8,$1E,$0E,$63,$8C,$6B
	.dc.b	$0F,$F7,$7B,$6F,$C5,$3D,$BE,$CB,$FD,$B5,$7F,$E5,$95,$F2,$79,$35
	.dc.b	$F0,$32,$75,$19,$78,$47,$A3,$EB,$FF,$C0,$D1,$FE,$63,$5F,$01,$93
	.dc.b	$5D,$72,$75,$19,$78,$47,$C2,$FC,$41,$E1,$F5,$F0,$25,$7F,$9B,$E7
	.dc.b	$2D,$9C,$BF,$53,$0F,$F5,$63,$92,$1F,$62,$78,$B6,$7D,$5D,$7F,$4B
	.dc.b	$F1,$07,$AF,$7E,$CE,$5F,$F9,$FE,$72,$D9,$CB,$F8,$48,$4F,$39,$FB
	.dc.b	$BF,$61,$F4,$2B,$F1,$2D,$E4,$71,$FE,$56,$30,$DE,$1E,$3C,$32,$FF
	.dc.b	$D2,$15,$A7,$FC,$70,$FD,$A7,$FD,$29,$E0,$FD,$27,$FD,$2D,$F6,$0F
	.dc.b	$B7,$FC,$7B,$63,$7C,$9F,$92,$AE,$11,$F9,$E9,$D3,$6B,$55,$23,$89
	.dc.b	$0E,$1C,$9D,$B8,$24,$78,$9F,$0F,$D1,$E2,$3B,$45,$DB,$E1,$C7,$F9
	.dc.b	$4F,$15,$7E,$9B,$EC,$3E,$8C,$FE,$C9,$F9,$EF,$FA,$CD,$AE,$78,$38
	.dc.b	$1C,$78,$64,$6E,$3F,$63,$F8,$83,$EB,$7F,$67,$F8,$83,$EB,$69,$A8
	.dc.b	$97,$FE,$C1,$9D,$36,$72,$FF,$06,$AE,$8F,$FB,$1D,$AE,$6B,$BF,$1F
	.dc.b	$F0,$2F,$43,$F5,$73,$FC,$39,$BB,$7C,$48,$46,$39,$BF,$8E,$3F,$62
	.dc.b	$74,$E6,$7F,$CF,$8D,$78,$F0,$F7,$E2,$F0,$71,$78,$78,$65,$7F,$B7
	.dc.b	$6F,$FC,$1B,$87,$4A,$CC,$F9,$6B,$8B,$FE,$E3,$FC,$D9,$F2,$DF,$C7
	.dc.b	$3E,$95,$F5,$14,$FE,$3F,$3A,$AE,$15,$97,$FD,$D9,$5C,$1C,$90,$FE
	.dc.b	$63,$3C,$5B,$3D,$F8,$FF,$95,$FE,$51,$FE,$E5,$9F,$F7,$B7,$A8,$FA
	.dc.b	$FF,$E5,$ED,$8F,$F7,$D4,$ED,$FE,$7B,$F9,$3B,$EE,$3F,$9A,$FE,$50
	.dc.b	$F8,$6A,$7F,$38,$9D,$57,$FA,$A7,$87,$EB,$7E,$50,$F9,$5D,$63,$2B
	.dc.b	$5C,$FF,$D4,$EB,$51,$FD,$3B,$E9,$FF,$A3,$7C,$E3,$E4,$3F,$8D,$CB
	.dc.b	$2F,$D2,$CE,$7C,$BF,$E9,$3E,$8B,$0F,$27,$C3,$61,$98,$78,$0F,$F5
	.dc.b	$F6,$DE,$FC,$7F,$18,$DD,$3F,$B3,$36,$77,$7E,$4F,$C7,$BF,$B4,$57
	.dc.b	$F6,$BF,$85,$F4,$D7,$F6,$4F,$93,$7C,$1F,$F5,$C6,$C7,$84,$4F,$0C
	.dc.b	$BF,$38,$DB,$FE,$ED,$5C,$2B,$A5,$FF,$66,$AD,$47,$FB,$5D,$AF,$4E
	.dc.b	$2D,$1E,$1F,$87,$FE,$DB,$F9,$FF,$ED,$A7,$E5,$63,$FE,$DF,$E1,$68
	.dc.b	$BF,$71,$C3,$C7,$97,$F2,$6B,$73,$A8,$F6,$61,$D7,$5F,$DB,$37,$1F
	.dc.b	$F0,$83,$67,$FD,$DC,$DB,$F0,$F8,$DF,$E0,$2F,$96,$FB,$55,$F1,$FF
	.dc.b	$28,$E0,$9B,$1A,$8F,$F0,$9B,$71,$FE,$0D,$3C,$0D,$F1,$F5,$25,$7D
	.dc.b	$3A,$E1,$97,$F8,$96,$BF,$A6,$4B,$F2,$9F,$E2,$F8,$7E,$43,$6F,$F7
	.dc.b	$A4,$FC,$36,$BF,$E2,$49,$DD,$FF,$8B,$DB,$0C,$FE,$7F,$A1,$F8,$B3
	.dc.b	$FC,$43,$3A,$9F,$06,$98,$23,$F9,$9F,$99,$7E,$80,$FA,$DE,$3F,$47
	.dc.b	$FE,$2A,$DB,$FC,$41,$63,$E4,$7B,$DA,$7F,$25,$92,$B8,$4C,$9C,$C6
	.dc.b	$DF,$63,$77,$F9,$BD,$FF,$B0,$F8,$EF,$F1,$E4,$3E,$D7,$FB,$A3,$7D
	.dc.b	$1D,$61,$FE,$4A,$AD,$D6,$8F,$9F,$DE,$8F,$F2,$1F,$D8,$BC,$CF,$E0
	.dc.b	$16,$FB,$1F,$F2,$0A,$FB,$27,$83,$FE,$5D,$9C,$C6,$27,$E1,$47,$0B
	.dc.b	$E9,$A6,$3E,$0B,$FC,$C9,$18,$AE,$D1,$FE,$67,$BE,$51,$83,$E6,$38
	.dc.b	$BE,$98,$F9,$9B,$87,$5A,$DE,$3E,$5C,$3A,$36,$3E,$B6,$3E,$1B,$E3
	.dc.b	$1D,$CF,$74,$ED,$A7,$51,$3C,$6B,$D7,$87,$8D,$EF,$3B,$C7,$D1,$1A
	.dc.b	$C7,$95,$F1,$AF,$39,$BC,$3C,$11,$98,$EF,$87,$57,$67,$9F,$0E,$AF
	.dc.b	$B2,$34,$F0,$BE,$8C,$3D,$D5,$D4,$ED,$5C,$A3,$DE,$6A,$B9,$CF,$D7
	.dc.b	$8F,$F4,$E1,$D4,$A5,$32,$ED,$FB,$FE,$73,$E5,$37,$1F,$2A,$30,$F2
	.dc.b	$9F,$EF,$73,$A6,$C7,$01,$F1,$55,$A6,$5C,$11,$FB,$9F,$2E,$FA,$DF
	.dc.b	$F1,$FE,$DF,$9C,$9F,$DC,$57,$DD,$AD,$77,$7D,$F9,$C9,$CA,$F8,$EF
	.dc.b	$F0,$2F,$84,$62,$7E,$2C,$7F,$03,$F5,$7F,$53,$87,$49,$FF,$26,$DF
	.dc.b	$16,$9F,$C2,$63,$53,$E5,$8F,$6E,$3C,$B0,$ED,$1F,$EC,$DC,$3C,$38
	.dc.b	$F9,$E3,$78,$FF,$B4,$71,$F0,$61,$CE,$F7,$5F,$9F,$3D,$B1,$D6,$F9
	.dc.b	$4E,$A3,$84,$FC,$19,$CB,$4E,$47,$80,$F5,$61,$E6,$7A,$E1,$E3,$8C
	.dc.b	$47,$D8,$C7,$C0,$FC,$23,$D8,$5C,$7E,$1D,$7F,$BA,$FE,$9D,$EF,$FF
	.dc.b	$95,$6B,$0D,$7F,$99,$46,$B1,$FC,$7B,$63,$F2,$ED,$E1,$FF,$40,$8F
	.dc.b	$B1,$FF,$6A,$D7,$F8,$26,$74,$FF,$8B,$E3,$35,$FE,$F4,$C7,$FC,$45
	.dc.b	$3F,$E0,$F6,$C7,$F8,$00,$7C,$9B,$EC,$BF,$DA,$F5,$BF,$1F,$C2,$FD
	.dc.b	$77,$D5,$4B,$E6,$FE,$C1,$AE,$E7,$F5,$71,$E0,$CB,$D7,$8F,$F5,$79
	.dc.b	$EC,$7F,$8C,$E3,$DD,$E0,$78,$2F,$FA,$C7,$B7,$FE,$D2,$AF,$F1,$43
	.dc.b	$63,$D5,$E1,$BE,$CB,$FC,$8D,$5A,$F6,$B9,$52,$F8,$3F,$C1,$9A,$FF
	.dc.b	$65,$35,$1E,$E8,$CC,$78,$A3,$C9,$E4,$3E,$BD,$F6,$5D,$AF,$C9,$A7
	.dc.b	$F9,$52,$73,$8F,$A3,$C4,$D1,$ED,$5F,$65,$F4,$9D,$13,$3F,$E3,$38
	.dc.b	$7F,$75,$8F,$E2,$AD,$8C,$D7,$9C,$E9,$5F,$BB,$C3,$FD,$4D,$5B,$7F
	.dc.b	$D9,$77,$E0,$E8,$F8,$BF,$46,$F7,$69,$EE,$CF,$FB,$95,$7F,$95,$E7
	.dc.b	$4D,$9F,$C1,$C6,$33,$FE,$5E,$D7,$F9,$9B,$5F,$9C,$EF,$0F,$8D,$F6
	.dc.b	$5E,$0B,$FF,$26,$60,$8F,$B3,$1F,$EA,$A3,$FC,$D0,$5B,$FC,$D3,$F2
	.dc.b	$BE,$4B,$7E,$0C,$8F,$E7,$66,$E7,$FA,$2B,$5F,$C0,$35,$FF,$13,$3D
	.dc.b	$70,$F8,$FF,$65,$FE,$99,$8F,$0D,$7D,$98,$C5,$7E,$AC,$FE,$83,$F6
	.dc.b	$3F,$90,$F0,$4F,$CB,$FB,$2C,$33,$0F,$E9,$F3,$A6,$C7,$EB,$A3,$15
	.dc.b	$FC,$13,$8B,$F6,$3D,$DF,$81,$8F,$EB,$ED,$7F,$89,$46,$BE,$C6,$61
	.dc.b	$FC,$39,$B3,$FD,$93,$EC,$5F,$5D,$FB,$C6,$D7,$F6,$26,$CE,$E7,$F8
	.dc.b	$0F,$D1,$7E,$AE,$BF,$B3,$B6,$7F,$B7,$B5,$FE,$E6,$D7,$F5,$DF,$5F
	.dc.b	$F2,$38,$3F,$DF,$4D,$AF,$F2,$B7,$86,$FA,$EF,$F4,$94,$7F,$A9,$E7
	.dc.b	$57,$FE,$5B,$7E,$8B,$FE,$00,$6B,$DB,$DE,$FE,$2F,$18,$8E,$11,$FC
	.dc.b	$37,$C0,$7B,$6F,$B0,$FE,$0F,$7D,$A7,$31,$FB,$7F,$B0,$F8,$7A,$7E
	.dc.b	$A9,$FF,$27,$E5,$FE,$CB,$0E,$9F,$C5,$BA,$CB,$F2,$EE,$54,$70,$F4
	.dc.b	$7B,$6F,$DD,$F0,$7C,$34,$9E,$D7,$FE,$F1,$6C,$66,$FA,$BE,$7F,$88
	.dc.b	$C6,$3D,$A7,$FD,$05,$F0,$FF,$C5,$AF,$F9,$03,$63,$FA,$1F,$07,$F4
	.dc.b	$C7,$7E,$37,$F2,$A9,$FE,$47,$06,$92,$3F,$A2,$BC,$F8,$BE,$D7,$D8
	.dc.b	$71,$9F,$E7,$91,$8C,$3F,$DE,$5F,$61,$FB,$2C,$3F,$AA,$3F,$82,$8C
	.dc.b	$C7,$ED,$0C,$9F,$C9,$CA,$7F,$47,$8E,$0E,$23,$8C,$E6,$1E,$85,$31
	.dc.b	$E0,$5A,$EF,$16,$7B,$CD,$D3,$E5,$43,$EB,$FE,$93,$FC,$05,$B1,$F3
	.dc.b	$FC,$47,$D8,$3E,$47,$FA,$9D,$6B,$2F,$EA,$0F,$2C,$3F,$CB,$2D,$73
	.dc.b	$C3,$FD,$C2,$31,$1F,$EB,$B7,$93,$FD,$CF,$EC,$1F,$61,$DA,$BF,$B5
	.dc.b	$7A,$49,$CE,$1F,$33,$EC,$3E,$21,$F1,$7E,$3A,$C6,$49,$8F,$D0,$8C
	.dc.b	$BF,$5B,$DB,$FC,$CC,$FD,$A7,$F6,$F1,$98,$FA,$91,$8D,$FF,$25,$C7
	.dc.b	$FE,$6F,$9C,$FC,$5E,$9A,$7F,$B5,$E7,$E7,$FD,$87,$F5,$2C,$3F,$97
	.dc.b	$36,$3F,$39,$A3,$E3,$BF,$29,$7C,$4F,$EF,$95,$B6,$FC,$0F,$F8,$6A
	.dc.b	$3A,$61,$FE,$88,$6B,$A9,$FF,$41,$C7,$18,$60,$F9,$67,$D8,$7F,$83
	.dc.b	$AB,$53,$98,$F8,$3F,$45,$F0,$BE,$4A,$E3,$F1,$DB,$8F,$CB,$6D,$FC
	.dc.b	$6D,$D1,$F2,$BD,$5F,$97,$F6,$17,$0F,$A0,$DC,$3E,$A3,$70,$E4,$FF
	.dc.b	$88,$BE,$BF,$E2,$CF,$FA,$91,$AF,$FA,$A1,$AF,$F8,$8D,$A8,$3C,$A7
	.dc.b	$F4,$6B,$EB,$FB,$78,$5D,$DF,$F1,$8F,$9B,$9C,$7E,$A6,$1C,$E3,$0F
	.dc.b	$F6,$9B,$B8,$7E,$BA,$6D,$7F,$8D,$5D,$D7,$F4,$CB,$FE,$A6,$49,$FF
	.dc.b	$5C,$DB,$EC,$3F,$C3,$F5,$EB,$6B,$EB,$8D,$4F,$F0,$D8,$C4,$E9,$F7
	.dc.b	$BF,$CA,$7D,$3F,$74,$71,$F9,$EC,$61,$D6,$FF,$D6,$EF,$C9,$8C,$CF
	.dc.b	$C9,$FB,$07,$CA,$FF,$9C,$23,$FD,$8F,$5F,$E5,$19,$CF,$8D,$CA,$7B
	.dc.b	$61,$54,$E9,$E2,$75,$BE,$98,$F9,$A3,$71,$8C,$3C,$97,$CE,$F9,$5E
	.dc.b	$DB,$3E,$3A,$E1,$7E,$23,$B9,$D3,$0E,$51,$A8,$C1,$AC,$D9,$7C,$28
	.dc.b	$CE,$3D,$EB,$86,$3E,$09,$ED,$39,$C3,$A9,$23,$A5,$F3,$AE,$11,$D3
	.dc.b	$1F,$45,$7C,$7C,$7C,$39,$71,$C7,$A5,$FA,$B0,$E3,$5A,$8C,$65,$E4
	.dc.b	$26,$1C,$F2,$F0,$46,$27,$73,$9C,$7B,$65,$E2,$9C,$9E,$88,$CB,$F5
	.dc.b	$E6,$BF,$AC,$BE,$B1,$8C,$7C,$B1,$89,$E7,$97,$A8,$FB,$D9,$7A,$5C
	.dc.b	$63,$E1,$6C,$FC,$C8,$C1,$1F,$11,$E5,$3A,$B9,$7E,$61,$C4,$DC,$74
	.dc.b	$72,$7A,$71,$F4,$65,$F9,$D8,$FF,$56,$1F,$84,$2E,$3F,$99,$6C,$6D
	.dc.b	$FD,$17,$D8,$7F,$AD,$DE,$27,$E8,$1C,$3C,$9A,$F0,$F7,$F6,$ED,$FC
	.dc.b	$F4,$73,$C3,$F3,$CD,$C7,$FD,$81,$F6,$0F,$B0,$FF,$61,$56,$AF,$89
	.dc.b	$30,$FD,$0D,$6D,$C4,$ED,$FC,$66,$1D,$5F,$E5,$13,$B7,$35,$C3,$0F
	.dc.b	$44,$74,$C3,$C3,$18,$C3,$A1,$1F,$3C,$6A,$7F,$86,$61,$DD,$F8,$07
	.dc.b	$63,$A1,$DC,$DD,$D5,$87,$97,$0E,$D1,$A3,$93,$FC,$2E,$35,$87,$94
	.dc.b	$93,$B9,$D6,$1E,$58,$F1,$C7,$46,$CF,$A9,$F1,$64,$C7,$D4,$DC,$7F
	.dc.b	$90,$63,$E0,$AE,$D7,$E0,$79,$9F,$3B,$1F,$05,$F1,$8F,$4C,$EE,$B9
	.dc.b	$47,$02,$4E,$F1,$F4,$C7,$93,$1F,$19,$F3,$6B,$8D,$6E,$FC,$4D,$AF
	.dc.b	$11,$C7,$1F,$9D,$5C,$2F,$9C,$63,$1F,$8D,$2F,$B0,$B1,$2F,$E0,$C6
	.dc.b	$8F,$F3,$33,$F8,$C6,$BF,$69,$B3,$CE,$3F,$AE,$0F,$91,$79,$7F,$A6
	.dc.b	$8C,$36,$33,$1F,$AA,$6B,$FB,$26,$F2,$7F,$98,$9B,$BF,$F8,$16,$79
	.dc.b	$9F,$DF,$87,$D7,$BE,$5B,$57,$FE,$6A,$8F,$F3,$9C,$ED,$D3,$67,$75
	.dc.b	$D6,$77,$3F,$EE,$F6,$BF,$C3,$9A,$EF,$CE,$F7,$46,$23,$FC,$35,$F5
	.dc.b	$EF,$96,$EB,$18,$FA,$CA,$A5,$18,$78,$47,$B5,$AF,$FC,$01,$76,$E6
	.dc.b	$38,$4F,$4F,$AF,$FE,$C7,$3F,$15,$7D,$3E,$4D,$74,$C9,$D4,$65,$E1
	.dc.b	$1E,$6F,$AF,$78,$7D,$7B,$E5,$7D,$19,$CB,$67,$F0,$11,$88,$FE,$5A
	.dc.b	$D7,$F9,$BB,$5F,$B8,$EF,$83,$FD,$42,$DD,$F3,$5E,$DE,$6F,$93,$F5
	.dc.b	$EF,$5A,$FD,$64,$CB,$F0,$93,$96,$CE,$5F,$F3,$6E,$1F,$8B,$95,$74
	.dc.b	$AF,$AB,$F5,$FF,$43,$3F,$F5,$9B,$6F,$F2,$4D,$C3,$FC,$B0,$FD,$47
	.dc.b	$A9,$8E,$0E,$64,$7F,$31,$1B,$AF,$F5,$FF,$7F,$D6,$9F,$CE,$7C,$BB
	.dc.b	$EB,$FF,$0E,$D8,$FD,$04,$ED,$EE,$7F,$90,$38,$BF,$61,$19,$8F,$EC
	.dc.b	$AD,$7F,$54,$D7,$FC,$E0,$F3,$BF,$1C,$FE,$5A,$31,$97,$EC,$6F,$F1
	.dc.b	$B7,$E0,$7F,$B3,$DF,$FB,$61,$B3,$BD,$BD,$3F,$33,$EA,$65,$FB,$9C
	.dc.b	$BF,$BB,$FA,$FF,$B0,$FB,$1F,$AA,$E7,$7F,$D4,$C6,$B1,$FF,$73,$B7
	.dc.b	$0F,$E0,$CF,$0A,$E5,$5C,$37,$FD,$8F,$D7,$3E,$97,$F6,$7F,$5E,$FA
	.dc.b	$9A,$6A,$26,$1F,$E0,$0A,$E0,$D8,$FE,$3D,$18,$8F,$F0,$5B,$73,$FF
	.dc.b	$7D,$B7,$3F,$E3,$4D,$9F,$F5,$06,$1F,$8B,$7F,$D8,$1D,$1F,$C1,$BD
	.dc.b	$2F,$9F,$83,$0F,$E4,$51,$86,$F1,$6E,$A5,$FF,$25,$AD,$3F,$EA,$D7
	.dc.b	$ED,$B7,$7E,$DC,$12,$38,$47,$84,$C3,$F8,$E9,$4E,$63,$F9,$97,$D7
	.dc.b	$BE,$53,$F8,$EF,$D1,$3E,$06,$9F,$C8,$27,$55,$C2,$BF,$A3,$46,$23
	.dc.b	$FC,$66,$D7,$FC,$7A,$D7,$FA,$13,$BD,$7E,$24,$7F,$B2,$63,$1A,$EA
	.dc.b	$BC,$7C,$5F,$CC,$3E,$B9,$EE,$FF,$55,$6E,$1F,$E3,$C6,$BF,$E4,$46
	.dc.b	$BF,$E4,$EF,$AC,$7D,$3F,$F5,$D6,$C7,$F9,$11,$AF,$F9,$31,$AF,$F9
	.dc.b	$5B,$EB,$5F,$25,$FD,$95,$B1,$FE,$4C,$6B,$FE,$54,$6B,$FE,$5E,$F2
	.dc.b	$9F,$47,$FD,$AC,$91,$FD,$AD,$F2,$65,$23,$C8,$D7,$C8,$FF,$52,$C3
	.dc.b	$DF,$BF,$CA,$94,$63,$6F,$F7,$A7,$D0,$3E,$B3,$57,$BF,$81,$F3,$BC
	.dc.b	$7E,$41,$EF,$7F,$5E,$6E,$5F,$DD,$EB,$78,$FF,$7B,$7D,$B1,$77,$FF
	.dc.b	$04,$46,$20,$1C,$36,$3F,$B5,$C7,$48,$DB,$55,$3E,$13,$FB,$8F,$D6
	.dc.b	$FE,$FE,$7A,$E5,$F8,$AD,$3D,$D5,$FD,$E1,$B3,$FE,$0F,$AE,$0F,$F3
	.dc.b	$06,$BF,$E9,$26,$BF,$E1,$76,$BE,$A6,$A7,$9D,$7D,$16,$E5,$C7,$C1
	.dc.b	$F1,$6B,$FC,$1E,$DF,$2D,$F8,$27,$FC,$45,$F5,$9C,$1F,$8E,$D7,$7C
	.dc.b	$5F,$E1,$09,$E1,$1F,$EA,$06,$BF,$EA,$A6,$BF,$E2,$F6,$BF,$49,$E3
	.dc.b	$97,$E1,$DB,$8E,$EB,$FC,$3B,$7F,$C6,$FE,$B3,$F8,$F6,$9F,$89,$F2
	.dc.b	$3D,$67,$F8,$8A,$3D,$87,$F2,$9C,$BF,$D7,$6D,$7F,$C8,$0D,$7F,$06
	.dc.b	$D4,$F2,$F1,$7A,$1F,$E3,$0F,$97,$7D,$6B,$E9,$FC,$1F,$3B,$E6,$F6
	.dc.b	$66,$3F,$E4,$BC,$53,$A8,$FF,$5F,$C6,$1B,$1F,$E5,$D8,$C4,$7F,$B2
	.dc.b	$9B,$C1,$FA,$B6,$EF,$EA,$AF,$8B,$DB,$FE,$42,$EC,$7C,$5F,$F9,$1B
	.dc.b	$07,$D6,$FF,$83,$5B,$E9,$6E,$66,$7F,$E6,$97,$FC,$D8,$47,$FD,$AB
	.dc.b	$F5,$8F,$1D,$7F,$AA,$F2,$3F,$CB,$BF,$5B,$FA,$4B,$C7,$BB,$FE,$6D
	.dc.b	$8C,$47,$F3,$E8,$FF,$38,$61,$CE,$F7,$3F,$B4,$8C,$2F,$AD,$EB,$E0
	.dc.b	$7F,$9D,$DB,$1F,$BA,$7F,$CD,$B3,$C1,$B9,$79,$A7,$96,$1E,$3B,$E4
	.dc.b	$D8,$E4,$F9,$27,$85,$72,$C3,$35,$FC,$23,$1F,$5C,$F8,$A3,$A9,$E9
	.dc.b	$AE,$37,$CE,$F8,$CA,$F8,$CF,$86,$38,$8C,$7D,$51,$C6,$33,$3C,$1F
	.dc.b	$16,$3E,$A7,$C1,$7D,$AF,$E4,$E5,$E5,$26,$1C,$DF,$65,$74,$BE,$4F
	.dc.b	$95,$B5,$E2,$8E,$A6,$CF,$5E,$1E,$37,$D1,$8F,$C0,$76,$7F,$AB,$4C
	.dc.b	$E3,$F8,$A8,$FA,$91,$9A,$FB,$ED,$F2,$5F,$1B,$F9,$06,$C7,$33,$E3
	.dc.b	$38,$78,$47,$FB,$15,$7C,$3D,$23,$FE,$C9,$6E,$BF,$EB,$4F,$90,$7D
	.dc.b	$47,$F8,$02,$3F,$D6,$73,$F6,$27,$3C,$DF,$6A,$FF,$A4,$CE,$6F,$87
	.dc.b	$07,$96,$31,$5B,$AF,$19,$F7,$AF,$9B,$F8,$03,$99,$9B,$E6,$FC,$9A
	.dc.b	$E6,$EC,$E7,$87,$A6,$77,$1F,$ED,$B9,$DC,$FA,$A3,$A5,$F8,$E7,$38
	.dc.b	$79,$E7,$27,$58,$CB,$E1,$33,$5C,$CF,$05,$73,$8C,$D7,$39,$CC,$73
	.dc.b	$32,$7F,$70,$FA,$37,$B3,$FE,$2E,$DF,$FC,$0D,$18,$C3,$FB,$49,$FE
	.dc.b	$06,$F2,$9F,$67,$B8,$FF,$7C,$4E,$F0,$FF,$4B,$B6,$F9,$E1,$B9,$78
	.dc.b	$5F,$E0,$AC,$31,$0B,$E1,$87,$F8,$12,$1B,$7F,$89,$23,$C7,$1F,$E0
	.dc.b	$88,$C3,$FE,$B4,$F8,$D7,$D7,$7F,$B9,$2B,$73,$CB,$C6,$EF,$7C,$34
	.dc.b	$FF,$12,$36,$BF,$C2,$78,$72,$F6,$37,$72,$7C,$36,$FA,$EF,$EA,$11
	.dc.b	$DB,$0E,$58,$FF,$34,$9C,$BF,$EB,$AF,$63,$FD,$B9,$5E,$46,$C7,$FB
	.dc.b	$33,$C6,$7D,$73,$E8,$7E,$2D,$6B,$2F,$F1,$D3,$70,$F8,$CD,$C1,$87
	.dc.b	$2B,$E3,$EB,$CB,$F9,$3C,$4F,$AC,$7D,$76,$71,$FF,$6E,$CE,$63,$C3
	.dc.b	$E3,$CB,$FA,$23,$EB,$9C,$79,$C3,$DD,$57,$6F,$F3,$CD,$EE,$B0,$FA
	.dc.b	$27,$F0,$8F,$3E,$CC,$DF,$FA,$8F,$0F,$DA,$E1,$FE,$59,$FA,$27,$CC
	.dc.b	$FE,$21,$B8,$7B,$35,$FE,$DE,$FE,$02,$33,$87,$FB,$A9,$B7,$CE,$FF
	.dc.b	$29,$AF,$C3,$F1,$BA,$44,$C3,$A7,$D7,$5C,$7E,$2E,$48,$FC,$CC,$FF
	.dc.b	$A9,$4F,$C7,$96,$7F,$BA,$BF,$E5,$E3,$E4,$7C,$87,$0F,$0E,$47,$07
	.dc.b	$E1,$EA,$9C,$D7,$23,$94,$FC,$8F,$AE,$FF,$47,$DF,$13,$EA,$B8,$9F
	.dc.b	$3F,$D7,$7E,$BB,$C4,$7D,$77,$C0,$D1,$39,$89,$87,$EA,$2B,$6F,$CD
	.dc.b	$9C,$DF,$EF,$B8,$FF,$73,$E0,$7F,$8B,$2B,$86,$18,$9C,$FD,$74,$9F
	.dc.b	$C8,$FE,$13,$FE,$C9,$7F,$6B,$69,$FE,$07,$F5,$EF,$AD,$A7,$1C,$BE
	.dc.b	$10,$8D,$5F,$EF,$BE,$81,$DF,$FC,$31,$B1,$FC,$0A,$7F,$82,$57,$29
	.dc.b	$CB,$D3,$EB,$FE,$47,$4F,$FA,$9D,$B7,$FC,$13,$EB,$7F,$D1,$CF,$F0
	.dc.b	$C8,$CD,$7F,$A6,$9A,$9C,$F6,$AB,$8C,$F7,$F8,$4F,$E4,$35,$C0,$DC
	.dc.b	$EB,$0F,$14,$FE,$9B,$EB,$BD,$F7,$F8,$3C,$51,$FE,$A2,$7F,$6A,$EA
	.dc.b	$BF,$62,$E6,$78,$39,$9F,$CA,$39,$9D,$EB,$DF,$D3,$F0,$78,$72,$FA
	.dc.b	$FD,$55,$C7,$9C,$DC,$3B,$CD,$AF,$E7,$5F,$31,$FC,$56,$CF,$F1,$7D
	.dc.b	$BE,$27,$9D,$FD,$0A,$34,$4F,$1B,$FA,$44,$EA,$FF,$9D,$CE,$5B,$3A
	.dc.b	$DF,$FA,$4E,$1F,$EC,$CA,$FE,$49,$F5,$1F,$D2,$EF,$FA,$4B,$FC,$F9
	.dc.b	$B3,$FB,$53,$F5,$D1,$96,$99,$CD,$3F,$ED,$89,$D4,$62,$51,$DD,$E3
	.dc.b	$8F,$F6,$18,$FF,$6F,$FA,$56,$4D,$F0,$30,$CC,$3C,$AE,$5A,$7F,$5C
	.dc.b	$FA,$C6,$DF,$DA,$9B,$1F,$D9,$BC,$77,$D6,$FE,$BF,$1F,$D4,$D6,$9B
	.dc.b	$7C,$4F,$EE,$D5,$C3,$E4,$31,$87,$79,$B5,$FC,$4E,$B5,$39,$3E,$5F
	.dc.b	$D5,$FB,$4F,$77,$C7,$58,$C9,$2B,$E6,$C6,$5F,$A3,$EC,$7C,$E9,$FA
	.dc.b	$EF,$E8,$DF,$A0,$E6,$37,$F1,$9F,$36,$31,$C3,$F4,$F8,$3F,$CB,$33
	.dc.b	$9B,$D3,$FE,$10,$FA,$BF,$E8,$F8,$7F,$0C,$6D,$FD,$BF,$51,$F1,$DF
	.dc.b	$98,$C3,$FC,$17,$5A,$7F,$C2,$CD,$B0,$71,$5C,$B9,$7C,$1E,$7C,$BE
	.dc.b	$6F,$F8,$96,$75,$18,$3E,$0F,$CE,$7B,$7D,$BF,$73,$63,$27,$C6,$8C
	.dc.b	$B4,$F9,$3E,$8F,$C5,$FA,$AF,$8C,$D8,$C9,$F3,$A3,$2D,$3F,$C5,$3F
	.dc.b	$53,$F0,$74,$FE,$62,$DD,$BF,$41,$08,$7C,$17,$F8,$F1,$D5,$7F,$8E
	.dc.b	$DB,$3F,$19,$AF,$19,$E1,$AF,$F3,$9F,$A9,$97,$E2,$6D,$7F,$39,$6E
	.dc.b	$1E,$2C,$11,$8A,$F0,$7D,$57,$ED,$7D,$1C,$E1,$E1,$8C,$57,$D8,$E5
	.dc.b	$FE,$93,$F5,$B8,$AD,$61,$FE,$8E,$AE,$6D,$8C,$9F,$E9,$E8,$E5,$E6
	.dc.b	$2B,$94,$FC,$5C,$2F,$D5,$BE,$43,$FC,$53,$3C,$4F,$C4,$C6,$7C,$3E
	.dc.b	$D1,$F3,$7E,$6F,$8D,$70,$D7,$D7,$2B,$E0,$F3,$31,$3B,$CF,$E9,$5F
	.dc.b	$2C,$3A,$DF,$0C,$39,$B6,$B7,$A6,$B0,$E5,$8F,$99,$AF,$6C,$3C,$FB
	.dc.b	$32,$ED,$3E,$3C,$7A,$3C,$6B,$73,$AC,$BC,$38,$74,$BD,$DF,$C1,$C7
	.dc.b	$C5,$84,$8E,$95,$E2,$C7,$C1,$59,$C3,$9D,$DC,$BD,$78,$70,$8C,$3E
	.dc.b	$33,$E4,$5F,$27,$44,$AE,$6E,$B2,$F4,$5F,$FA,$5F,$2F,$39,$E8,$BE
	.dc.b	$57,$C0,$93,$BC,$BD,$17,$CB,$0F,$15,$69,$FD,$34,$65,$F9,$C7,$8E
	.dc.b	$BD,$73,$B8,$CB,$6F,$C3,$3D,$F1,$F7,$D6,$A3,$89,$97,$A1,$E2,$72
	.dc.b	$D2,$1C,$4F,$2E,$5D,$32,$FC,$8C,$7F,$AE,$4F,$39,$5C,$BD,$CC,$7D
	.dc.b	$5F,$F8,$E1,$E0,$7E,$6D,$C3,$AD,$B2,$E5,$B2,$9E,$59,$7E,$69,$B8
	.dc.b	$7D,$56,$DF,$F3,$6F,$A9,$7D,$5F,$FB,$DE,$B5,$8F,$E9,$2F,$77,$DC
	.dc.b	$FD,$65,$F0,$75,$31,$FC,$F4,$F0,$9D,$46,$27,$58,$79,$5F,$DE,$CE
	.dc.b	$A3,$13,$B2,$3F,$BE,$8D,$5F,$9A,$72,$D9,$D6,$1E,$B9,$FE,$17,$87
	.dc.b	$AB,$1F,$8D,$8F,$8B,$0F,$F7,$76,$3D,$2F,$73,$96,$9F,$03,$0E,$8F
	.dc.b	$C3,$C7,$CE,$4B,$E7,$3A,$C3,$BD,$79,$8E,$75,$E6,$7B,$1C,$EB,$D4
	.dc.b	$72,$73,$3C,$A7,$B4,$7D,$9A,$ED,$8F,$B4,$98,$7C,$C8,$CB,$CA,$3E
	.dc.b	$A4,$62,$7C,$87,$84,$E8,$6E,$7C,$B8,$FC,$58,$D1,$30,$F2,$63,$E7
	.dc.b	$8E,$71,$87,$DF,$19,$C7,$ED,$7D,$5F,$CA,$B3,$4F,$B1,$C5,$FE,$4A
	.dc.b	$F9,$5C,$61,$F8,$69,$AF,$F4,$79,$D3,$FD,$1B,$E4,$3F,$D9,$0F,$F8
	.dc.b	$4B,$97,$FB,$AC,$62,$3F,$D5,$2D,$8F,$EF,$3F,$56,$F9,$4F,$F0,$CD
	.dc.b	$6B,$CC,$B5,$C9,$BC,$5A,$C3,$93,$C3,$0E,$57,$E1,$8C,$79,$9C,$23
	.dc.b	$FC,$4B,$3F,$E6,$07,$46,$23,$4A,$3A,$B9,$36,$66,$37,$1C,$E3,$FD
	.dc.b	$8A,$F0,$5F,$56,$F9,$2E,$AB,$EA,$E9,$88,$93,$DA,$74,$D9,$CB,$FE
	.dc.b	$D2,$FA,$BF,$EB,$CF,$91,$7C,$77,$A9,$B1,$9E,$BC,$19,$99,$19,$3D
	.dc.b	$BF,$56,$F1,$7E,$1B,$63,$31,$D7,$A7,$2C,$8C,$9E,$38,$C3,$FE,$E8
	.dc.b	$9E,$4D,$9E,$4F,$CC,$FA,$B7,$AB,$F4,$97,$D4,$53,$11,$27,$84,$65
	.dc.b	$E5,$1F,$6B,$EA,$FE,$84,$7D,$86,$BF,$EF,$A6,$F2,$FF,$92,$9F,$9F
	.dc.b	$19,$E0,$E4,$47,$96,$FF,$86,$F0,$BE,$6C,$2F,$FC,$A7,$18,$6E,$DF
	.dc.b	$8C,$F6,$3E,$BD,$6F,$97,$F3,$F7,$B6,$D7,$53,$87,$0D,$AF,$C8,$D6
	.dc.b	$DF,$F3,$FB,$55,$F6,$C3,$FC,$CA,$FF,$5A,$6F,$0F,$8C,$F2,$78,$F6
	.dc.b	$BF,$3B,$5B,$CB,$F6,$6D,$8F,$F4,$6B,$FE,$6E,$7F,$47,$F5,$5F,$61
	.dc.b	$BC,$1F,$AA,$8F,$B3,$1A,$E0,$C1,$99,$91,$93,$F5,$BF,$54,$EE,$FD
	.dc.b	$9B,$63,$31,$FB,$6F,$AA,$FB,$39,$7F,$08,$AD,$57,$CD,$3F,$D4,$6E
	.dc.b	$35,$FF,$43,$B6,$7F,$87,$D7,$2A,$DD,$F8,$08,$FF,$00,$8D,$2F,$AA
	.dc.b	$F9,$D5,$FC,$61,$B1,$FE,$B2,$6B,$BE,$1F,$E1,$B1,$82,$70,$E2,$33
	.dc.b	$C1,$AE,$09,$7F,$C4,$2F,$E9,$3E,$19,$CB,$EF,$3C,$93,$C9,$FE,$2F
	.dc.b	$F5,$4F,$8E,$FE,$3A,$D8,$CC,$7F,$20,$F9,$CF,$B5,$1C,$E7,$4D,$9E
	.dc.b	$26,$F8,$6C,$72,$21,$E3,$AE,$7C,$1F,$EB,$6F,$AA,$7B,$3F,$CB,$9B
	.dc.b	$5F,$CB,$1A,$FA,$FE,$C1,$EE,$FF,$34,$FA,$A7,$D1,$52,$57,$F3,$4F
	.dc.b	$AB,$7C,$4F,$F3,$A6,$C6,$63,$07,$F4,$3F,$97,$7C,$83,$E4,$7F,$A0
	.dc.b	$F9,$4F,$9D,$A4,$C3,$FA,$0E,$7E,$6E,$0F,$F6,$8B,$7A,$BF,$AC,$B6
	.dc.b	$7F,$DE,$6D,$C9,$F5,$3F,$ED,$8B,$5F,$FB,$E2,$72,$FB,$F8,$A6,$1F
	.dc.b	$D5,$A7,$2F,$F5,$2F,$82,$F6,$BF,$81,$F8,$C7,$C0,$7F,$5A,$6C,$6F
	.dc.b	$83,$FA,$ED,$70,$CB,$FD,$D2,$DC,$BF,$B8,$B6,$82,$DF,$09,$CB,$09
	.dc.b	$3D,$4F,$EC,$B3,$A7,$FB,$3F,$D5,$BE,$AF,$E8,$6D,$99,$FD,$B6,$BF
	.dc.b	$13,$1F,$EE,$37,$B9,$E3,$78,$23,$F0,$EF,$F7,$4F,$9A,$78,$3F,$A3
	.dc.b	$6E,$9F,$3A,$35,$5E,$2F,$0F,$FB,$FF,$D5,$FD,$F9,$FE,$FB,$F5,$7C
	.dc.b	$37,$FF,$09,$B6,$3E,$1C,$ED,$FD,$83,$70,$ED,$08,$C5,$7F,$87,$1B
	.dc.b	$1F,$32,$BF,$85,$C6,$23,$87,$1F,$90,$FF,$09,$CF,$F1,$0F,$A9,$F7
	.dc.b	$ED,$F8,$AF,$2F,$E5,$9F,$E1,$88,$F8,$76,$69,$33,$FF,$11,$7D,$0B
	.dc.b	$EA,$DF,$57,$C6,$65,$FF,$87,$AF,$E6,$DE,$EB,$B7,$57,$93,$DA,$97
	.dc.b	$B8,$CB,$48,$7F,$15,$D3,$F7,$2E,$38,$3E,$3E,$3F,$B1,$8C,$70,$7D
	.dc.b	$0E,$EF,$D9,$FC,$5C,$9F,$F1,$8F,$6A,$3F,$75,$D7,$FE,$40,$9F,$F2
	.dc.b	$6C,$6D,$A7,$EA,$63,$58,$FD,$6F,$A9,$78,$9F,$DE,$69,$F5,$3F,$E3
	.dc.b	$BF,$AA,$FD,$04,$7E,$8A,$31,$DB,$FE,$56,$7F,$C8,$17,$CB,$0D,$7B
	.dc.b	$BF,$E5,$A8,$C5,$F4,$8F,$F2,$EE,$1D,$23,$07,$EB,$9C,$5F,$4C,$7C
	.dc.b	$8D,$C3,$AD,$6F,$1F,$16,$1D,$1B,$1F,$B8,$C7,$C3,$7C,$63,$B8,$C7
	.dc.b	$BD,$F1,$8E,$A2,$B8,$C7,$AF,$1F,$1B,$B9,$E7,$5C,$9F,$04,$E9,$B8
	.dc.b	$F4,$9D,$47,$87,$1F,$3B,$DE,$B8,$36,$B8,$4E,$AB,$8E,$1E,$4A,$EA
	.dc.b	$78,$A7,$47,$48,$D4,$73,$34,$74,$3D,$EE,$8F,$AE,$FF,$A4,$8E,$05
	.dc.b	$72,$FC,$3A,$D3,$F3,$FC,$97,$BF,$88,$F9,$51,$87,$75,$FD,$CE,$74
	.dc.b	$D8,$E0,$3E,$0A,$B4,$CB,$82,$3F,$73,$E3,$9F,$4B,$FC,$7F,$6F,$F4
	.dc.b	$A4,$FF,$3A,$AF,$BB,$5A,$E3,$FB,$F3,$93,$95,$F1,$DF,$DB,$7C,$23
	.dc.b	$13,$F1,$63,$FA,$2F,$9F,$FE,$A1,$C3,$A4,$A3,$C2,$FA,$5D,$57,$F3
	.dc.b	$37,$35,$D7,$0F,$1D,$7B,$F1,$DC,$62,$75,$3F,$9A,$3D,$13,$A7,$07
	.dc.b	$1C,$7F,$D9,$78,$F9,$27,$BC,$FB,$AF,$A3,$4E,$37,$DA,$76,$75,$2D
	.dc.b	$F4,$C7,$E1,$D7,$FB,$62,$B8,$4F,$C8,$8E,$95,$FE,$CF,$EA,$FE,$D5
	.dc.b	$F0,$1F,$D1,$A7,$FB,$6C,$63,$1F,$C5,$B6,$3F,$24,$D7,$FD,$CA,$E7
	.dc.b	$7F,$EB,$9F,$71,$F8,$7F,$73,$FA,$F4,$7F,$5F,$9D,$5F,$F7,$AA,$D3
	.dc.b	$FE,$E0,$87,$CB,$BE,$C7,$8C,$F9,$23,$0F,$FB,$9A,$9F,$19,$FE,$CE
	.dc.b	$8E,$53,$AC,$BF,$CE,$18,$74,$C3,$DD,$08,$C5,$7B,$AB,$84,$FB,$1A
	.dc.b	$F8,$6B,$C6,$B0,$F0,$FB,$38,$C3,$9C,$FF,$9F,$E7,$38,$7B,$7E,$C9
	.dc.b	$F0,$FB,$6C,$7F,$C0,$0D,$1F,$56,$FC,$03,$57,$2B,$E9,$DF,$0C,$7E
	.dc.b	$9B,$5F,$A6,$D7,$ED,$BB,$C3,$EA,$F3,$3C,$4E,$95,$E7,$8C,$5F,$D3
	.dc.b	$FC,$8D,$F5,$3E,$D6,$C7,$9F,$4F,$F2,$A3,$B9,$D1,$F3,$B6,$57,$0C
	.dc.b	$3F,$12,$D7,$F1,$2D,$7F,$22,$EF,$2F,$C5,$F2,$B4,$EB,$8F,$E0,$63
	.dc.b	$0F,$E2,$7E,$2B,$FD,$6C,$DF,$27,$EA,$7D,$6F,$F9,$E2,$30,$4C,$BF
	.dc.b	$37,$5C,$1B,$5E,$98,$C4,$7E,$79,$D6,$3F,$9E,$FB,$B7,$D1,$7E,$06
	.dc.b	$73,$87,$E7,$67,$2F,$E7,$71,$7E,$0B,$E4,$3F,$17,$E8,$3E,$B1,$F5
	.dc.b	$F9,$BE,$15,$F2,$EF,$8C,$7E,$D1,$BA,$7E,$D1,$AF,$ED,$1C,$E9,$FB
	.dc.b	$4F,$BE,$FF,$33,$D7,$43,$F5,$4E,$2B,$F4,$FF,$21,$FE,$91,$9F,$DA
	.dc.b	$64,$FB,$5F,$CF,$4E,$6B,$F5,$B3,$9A,$DD,$7E,$06,$31,$3F,$C2,$1A
	.dc.b	$FF,$08,$6B,$FC,$0F,$C4,$FF,$10,$D7,$EF,$5B,$1D,$AB,$F6,$3D,$9F
	.dc.b	$D3,$3C,$BF,$D9,$CB,$E8,$DE,$B6,$BE,$FF,$F1,$59,$7F,$A2,$2B,$53
	.dc.b	$9E,$BF,$DC,$7E,$76,$FB,$5F,$CC,$36,$FF,$90,$B6,$FF,$90,$B5,$E9
	.dc.b	$9B,$F0,$7F,$F5,$FD,$7F,$15,$6C,$FF,$03,$FA,$2F,$F4,$CC,$E7,$E0
	.dc.b	$BD,$3F,$5C,$F8,$EA,$7B,$3E,$BF,$F8,$E6,$1F,$EE,$B9,$D5,$FF,$A8
	.dc.b	$B9,$FF,$CF,$B1,$89,$FF,$00,$B5,$FF,$04,$35,$F8,$4D,$7F,$D5,$6F
	.dc.b	$18,$FF,$43,$FD,$E3,$EA,$FC,$77,$DC,$F9,$51,$97,$DD,$E6,$7D,$4D
	.dc.b	$5F,$73,$63,$0D,$9F,$88,$DC,$BF,$93,$57,$09,$FC,$07,$D0,$F3,$F9
	.dc.b	$5F,$BF,$D4,$8C,$4E,$8F,$9D,$3A,$8C,$DF,$FA,$EB,$ED,$1F,$39,$FA
	.dc.b	$EF,$B4,$E5,$87,$FB,$66,$72,$FF,$B6,$E3,$35,$FE,$C2,$6E,$FF,$EB
	.dc.b	$7F,$11,$18,$AF,$F7,$35,$F2,$BE,$18,$F7,$24,$F7,$C6,$CF,$FB,$3B
	.dc.b	$F2,$8F,$CB,$CF,$FB,$AD,$B0,$CB,$FD,$8D,$ED,$3F,$29,$7E,$1D,$FC
	.dc.b	$22,$F7,$5E,$56,$E1,$FE,$F9,$6E,$0C,$39,$5F,$1E,$EF,$B1,$F5,$13
	.dc.b	$0F,$ED,$9E,$03,$EE,$31,$8F,$D6,$AB,$9F,$F7,$EA,$D5,$FE,$7E,$7F
	.dc.b	$67,$C3,$FC,$2E,$BF,$DA,$2D,$9F,$D9,$7B,$4F,$94,$DF,$C9,$C9,$FF
	.dc.b	$2B,$57,$F1,$3C,$BF,$CC,$0D,$7F,$74,$D4,$1E,$CE,$3D,$BF,$E2,$38
	.dc.b	$F8,$AF,$F8,$3C,$62,$FF,$C2,$5D,$1F,$D1,$6F,$F2,$7F,$6B,$F9,$BF
	.dc.b	$99,$FC,$E3,$6B,$F7,$8D,$9D,$C7,$E2,$7E,$1B,$F1,$C7,$F2,$66,$99
	.dc.b	$24,$7F,$2E,$72,$FD,$F7,$D7,$C7,$FE,$9F,$DF,$FC,$1E,$D8,$FD,$94
	.dc.b	$FB,$FE,$01,$F7,$DA,$F8,$56,$14,$FF,$0F,$E1,$C6,$FF,$D1,$4D,$7F
	.dc.b	$CA,$73,$C3,$0F,$F4,$4C,$EF,$7F,$D0,$76,$3E,$1F,$A5,$7E,$9D,$FD
	.dc.b	$AF,$91,$23,$FC,$75,$F9,$27,$FB,$8D,$FE,$75,$19,$CB,$FC,$49,$E0
	.dc.b	$A3,$FC,$51,$5B,$9C,$C7,$F0,$BF,$72,$5F,$E5,$3E,$2E,$D7,$F0,$89
	.dc.b	$B9,$FE,$4E,$6D,$7F,$96,$BC,$DF,$F4,$0D,$FF,$95,$9A,$E7,$2F,$E7
	.dc.b	$E4,$9F,$E7,$1C,$5F,$D1,$23,$F8,$E3,$FC,$77,$DB,$93,$A3,$F9,$F3
	.dc.b	$8A,$FE,$36,$F0,$8C,$9F,$62,$1F,$5D,$FD,$E2,$3F,$D7,$8D,$8F,$EA
	.dc.b	$CD,$7F,$3A,$D7,$8F,$B2,$9F,$F5,$57,$8B,$C3,$9F,$15,$FE,$B7,$8E
	.dc.b	$75,$8F,$89,$FE,$55,$3D,$2B,$15,$FD,$97,$EB,$7F,$B4,$F8,$1D,$2B
	.dc.b	$F5,$B1,$87,$A1,$E4,$ED,$E2,$EB,$C3,$FF,$39,$BA,$13,$C7,$1F,$0C
	.dc.b	$62,$77,$18,$D1,$7D,$2B,$58,$78,$DA,$75,$C3,$A6,$3F,$22,$B8,$46
	.dc.b	$27,$C9,$5C,$63,$11,$F4,$A7,$78,$F8,$AF,$E1,$36,$78,$9E,$BC,$3C
	.dc.b	$BA,$73,$9D,$4F,$93,$2F,$6E,$3D,$AF,$73,$E7,$AD,$CF,$88,$D6,$3E
	.dc.b	$2C,$BA,$12,$BC,$11,$97,$94,$67,$0F,$91,$39,$7C,$37,$D8,$F2,$E1
	.dc.b	$E2,$8C,$D7,$1C,$BD,$EF,$CD,$9D,$46,$30,$EE,$4A,$EF,$97,$18,$F8
	.dc.b	$B1,$D1,$F3,$C6,$5F,$89,$3C,$E7,$8C,$73,$8C,$11,$FC,$64,$F1,$CB
	.dc.b	$E3,$63,$EA,$CB,$DD,$7B,$BE,$A7,$03,$A1,$E4,$75,$8F,$B2,$30,$7D
	.dc.b	$C3,$FC,$88,$7E,$30,$B7,$F9,$F6,$DF,$E7,$1C,$FD,$77,$F8,$DD,$E2
	.dc.b	$7E,$79,$C3,$AF,$47,$2E,$5B,$49,$E5,$9F,$E2,$1B,$8F,$EC,$DB,$7F
	.dc.b	$D7,$BE,$BB,$F9,$EB,$F9,$B8,$CB,$D4,$FD,$23,$88,$DC,$63,$0E,$04
	.dc.b	$7F,$6D,$1A,$BF,$14,$E5,$B3,$AE,$DF,$F3,$54,$60,$98,$79,$EB,$F6
	.dc.b	$D7,$CE,$31,$7E,$12,$57,$F0,$29,$D4,$7F,$0E,$C3,$D7,$8F,$C5,$C3
	.dc.b	$C5,$87,$F9,$E6,$F7,$7E,$FC,$3C,$2F,$1A,$F0,$E3,$E6,$BF,$0D,$6A
	.dc.b	$B8,$47,$39,$D4,$71,$3C,$B1,$C0,$D3,$E2,$35,$3C,$EF,$C7,$87,$9E
	.dc.b	$78,$9A,$AE,$2E,$0F,$09,$93,$EE,$5F,$89,$C9,$27,$A5,$FC,$79,$D4
	.dc.b	$62,$BC,$D5,$CC,$5F,$93,$1F,$96,$4C,$7E,$2C,$FA,$CE,$D1,$F8,$7C
	.dc.b	$3A,$BB,$8F,$45,$7D,$01,$E7,$79,$DF,$F3,$FC,$7E,$03,$CE,$C4,$C8
	.dc.b	$FF,$4F,$9F,$65,$FF,$59,$6F,$FB,$0A,$E7,$3A,$3F,$3D,$3A,$7F,$C7
	.dc.b	$BE,$A3,$EE,$7F,$CC,$31,$FE,$22,$8C,$79,$7F,$EB,$DE,$BF,$EC,$31
	.dc.b	$89,$D5,$3F,$0F,$D6,$5D,$23,$9C,$FF,$8C,$45,$71,$70,$DB,$EB,$3C
	.dc.b	$E7,$FC,$67,$F0,$CF,$1B,$F5,$F5,$AC,$7B,$D7,$4B,$E0,$F7,$BE,$1B
	.dc.b	$E9,$B1,$FC,$31,$AF,$F1,$66,$BF,$ED,$B6,$BE,$BF,$47,$BE,$5F,$1D
	.dc.b	$B6,$73,$E6,$FF,$B7,$EB,$E2,$C6,$88,$FF,$A0,$BD,$8E,$7E,$FB,$ED
	.dc.b	$29,$27,$FC,$7E,$53,$E3,$C6,$DF,$AC,$D7,$E0,$78,$DF,$CC,$36,$FF
	.dc.b	$4E,$FC,$8B,$DB,$F8,$FE,$8B,$E7,$1F,$81,$61,$B8,$FF,$98,$BC,$E6
	.dc.b	$1C,$63,$0D,$7F,$CB,$ED,$7F,$D0,$E7,$FA,$8D,$C3,$FE,$70,$6B,$9C
	.dc.b	$3E,$CC,$F6,$6C,$F5,$C7,$FD,$89,$E2,$3E,$D3,$CD,$B4,$C7,$FC,$EF
	.dc.b	$A7,$AE,$3E,$F7,$D0,$7E,$A2,$7F,$21,$39,$77,$A7,$C7,$C7,$FD,$A0
	.dc.b	$DE,$9F,$F4,$FF,$33,$0E,$55,$F9,$4F,$4B,$EF,$FD,$74,$C7,$F4,$31
	.dc.b	$F8,$23,$0F,$E7,$39,$D3,$98,$FF,$2B,$7E,$58,$FA,$EF,$CC,$C3,$F2
	.dc.b	$FF,$D3,$E3,$FE,$4E,$8C,$77,$EA,$F8,$47,$2B,$FE,$C2,$A3,$2F,$F6
	.dc.b	$86,$BF,$DC,$9A,$FD,$67,$75,$DD,$BA,$E6,$BC,$1D,$5F,$9E,$65,$FF
	.dc.b	$00,$7F,$3F,$EF,$62,$3E,$04,$5C,$3F,$DB,$53,$F9,$0A,$7E,$01,$49
	.dc.b	$7F,$8F,$F1,$5F,$27,$88,$FB,$35,$C1,$FB,$ED,$7F,$86,$75,$7F,$7F
	.dc.b	$9F,$CD,$7D,$04,$9F,$BE,$F8,$7E,$5B,$11,$F8,$4E,$64,$E3,$4F,$F3
	.dc.b	$EF,$3F,$EE,$EF,$85,$F2,$21,$E4,$7F,$73,$C3,$FC,$0B,$CD,$93,$FB
	.dc.b	$48,$CB,$FC,$9A,$32,$FE,$EA,$3F,$84,$FC,$97,$F0,$AF,$7E,$4F,$F2
	.dc.b	$C8,$CB,$FC,$BA,$32,$FE,$36,$3F,$13,$F5,$7F,$EF,$F8,$FF,$51,$76
	.dc.b	$3D,$AF,$C6,$B6,$3F,$D4,$D3,$B7,$F9,$A0,$70,$FF,$8E,$1A,$FE,$F1
	.dc.b	$A8,$E1,$F0,$3F,$92,$8E,$9A,$FE,$65,$BC,$FC,$67,$F3,$FF,$56,$F8
	.dc.b	$1F,$E2,$7F,$06,$F8,$EA,$4C,$BF,$7B,$1F,$59,$FE,$87,$E7,$FF,$B7
	.dc.b	$5F,$E7,$4D,$7F,$9D,$35,$FE,$74,$D7,$F8,$AF,$E6,$DF,$E4,$57,$FA
	.dc.b	$F7,$BC,$F7,$3F,$9E,$B6,$3F,$D9,$93,$B7,$F9,$8F,$C8,$FF,$38,$23
	.dc.b	$FD,$5B,$F4,$1F,$E3,$1B,$7F,$0A,$7F,$8F,$F9,$06,$93,$CC,$FF,$2C
	.dc.b	$C7,$F0,$FE,$DF,$EA,$D3,$CA,$D3,$C6,$BF,$8E,$D7,$F5,$BF,$A6,$7D
	.dc.b	$EF,$F5,$4F,$C4,$5A,$D3,$51,$31,$FE,$C1,$5C,$1B,$1E,$28,$C4,$7F
	.dc.b	$63,$6F,$37,$FB,$83,$24,$63,$4F,$E4,$ED,$A7,$77,$F2,$9E,$DF,$EE
	.dc.b	$1F,$6D,$FE,$78,$C7,$FC,$13,$F8,$97,$F7,$BB,$FF,$07,$35,$FE,$F1
	.dc.b	$3B,$C7,$F9,$9B,$FE,$80,$8F,$EE,$8D,$73,$EE,$FC,$A8,$C4,$7C,$87
	.dc.b	$93,$63,$C6,$7C,$9F,$5B,$FC,$09,$F2,$8F,$96,$FE,$FB,$F8,$9B,$EC
	.dc.b	$E9,$A8,$98,$7F,$82,$6B,$83,$63,$ED,$C6,$23,$FC,$18,$DD,$3E,$53
	.dc.b	$73,$FE,$9E,$D9,$4B,$F0,$77,$D6,$FF,$86,$7F,$42,$7C,$77,$F8,$4B
	.dc.b	$E0,$3F,$9B,$E1,$FE,$2C,$6F,$2A,$B7,$3F,$E2,$A6,$BF,$E1,$F6,$BF
	.dc.b	$E1,$F6,$BF,$D8,$9A,$FE,$49,$E3,$8F,$E0,$DB,$9E,$F0,$FF,$19,$CF
	.dc.b	$A7,$D1,$FF,$18,$7A,$6F,$2B,$FC,$53,$F3,$5F,$D7,$B0,$FF,$28,$B7
	.dc.b	$2F,$CF,$D6,$DF,$F2,$0B,$5F,$F1,$B3,$5F,$F1,$B3,$5F,$EF,$2D,$79
	.dc.b	$D7,$88,$8F,$F9,$76,$7F,$C8,$F8,$7E,$77,$E4,$FF,$BB,$CF,$F9,$87
	.dc.b	$07,$E0,$1F,$E4,$06,$C7,$F9,$52,$76,$FF,$93,$DA,$FF,$92,$5A,$FF
	.dc.b	$83,$1A,$FE,$F1,$A9,$FC,$95,$7F,$2D,$FA,$27,$D4,$FF,$8B,$3E,$05
	.dc.b	$F2,$5F,$E2,$E9,$CE,$1F,$CC,$E7,$2F,$1A,$FF,$19,$E1,$D2,$15,$C1
	.dc.b	$F0,$73,$7E,$41,$CD,$77,$C3,$B3,$63,$A1,$FD,$06,$FA,$5F,$36,$CE
	.dc.b	$E7,$99,$E5,$AE,$6F,$1B,$F0,$5F,$16,$D7,$18,$F3,$61,$EC,$AE,$67
	.dc.b	$A3,$1F,$1B,$4C,$C6,$E7,$8E,$1E,$09,$D4,$77,$AE,$35,$BB,$F0,$1F
	.dc.b	$23,$0F,$1D,$F9,$EB,$FB,$07,$87,$D2,$2D,$7B,$E3,$A4,$7F,$86,$A3
	.dc.b	$94,$F6,$C7,$C5,$87,$7A,$E8,$7B,$32,$F9,$2F,$C7,$38,$1E,$E7,$2F
	.dc.b	$BC,$FF,$37,$94,$EE,$60,$C1,$F5,$FE,$55,$F5,$59,$8F,$C8,$46,$1E
	.dc.b	$12,$8C,$47,$68,$C4,$FF,$94,$21,$F2,$14,$8F,$F9,$51,$BA,$FE,$3F
	.dc.b	$DA,$FE,$AB,$BF,$D8,$AF,$F0,$55,$7D,$A9,$CC,$7D,$E3,$27,$E0,$1C
	.dc.b	$D7,$E6,$63,$2F,$5F,$19,$7D,$67,$CA,$7D,$33,$FC,$31,$5D,$B7,$E5
	.dc.b	$1A,$79,$4F,$DC,$8F,$9B,$87,$96,$FA,$C7,$94,$EB,$87,$07,$13,$BA
	.dc.b	$F1,$36,$32,$7A,$EB,$9D,$78,$A7,$4D,$9D,$57,$26,$9C,$A3,$C0,$6C
	.dc.b	$BB,$FE,$06,$BD,$52,$C7,$E0,$4E,$63,$E2,$19,$C3,$DB,$8F,$F0,$CE
	.dc.b	$8F,$E1,$FF,$61,$FE,$93,$AF,$F5,$2F,$02,$31,$1F,$E1,$68,$C6,$9F
	.dc.b	$EE,$D6,$C7,$FB,$B5,$BC,$3F,$E0,$BF,$D0,$DF,$55,$F3,$23,$FC,$1D
	.dc.b	$3A,$BF,$F5,$9D,$69,$FF,$28,$4F,$1B,$E1,$1A,$69,$FE,$28,$9F,$C5
	.dc.b	$CE,$E7,$89,$B3,$F4,$D8,$A7,$FC,$B9,$18,$6D,$7E,$BA,$74,$FC,$06
	.dc.b	$CF,$48,$76,$3E,$59,$F3,$DE,$4B,$DD,$7F,$7E,$BF,$F3,$55,$EE,$B5
	.dc.b	$ED,$3C,$2A,$63,$D3,$B7,$DF,$1F,$F3,$F4,$F8,$EB,$84,$FB,$1B,$8F
	.dc.b	$A5,$AF,$FB,$DA,$B8,$CF,$A7,$47,$D4,$FC,$FA,$F3,$1F,$E5,$2B,$E7
	.dc.b	$3F,$5E,$74,$FF,$91,$7E,$A5,$F5,$9E,$F6,$E1,$E7,$7E,$E6,$1F,$E5
	.dc.b	$C7,$2F,$6C,$51,$88,$F3,$35,$FF,$7A,$B5,$E5,$5F,$4E,$39,$DF,$97
	.dc.b	$8B,$C5,$3F,$84,$8F,$6C,$7E,$1A,$9E,$53,$E1,$30,$DC,$BE,$77,$D5
	.dc.b	$3C,$FC,$D4,$9F,$E5,$55,$BB,$FE,$60,$D7,$F9,$AB,$5F,$F5,$41,$E0
	.dc.b	$9F,$C2,$7D,$47,$FA,$8E,$BF,$15,$39,$7E,$F4,$66,$3F,$D6,$F5,$FE
	.dc.b	$55,$9D,$B6,$3E,$54,$BE,$3F,$D7,$E6,$CB,$F3,$BF,$B8,$AF,$82,$DB
	.dc.b	$E5,$1E,$88,$FD,$5F,$A4,$FA,$87,$D4,$F3,$AF,$C2,$CE,$63,$E6,$7D
	.dc.b	$4F,$F8,$FF,$1F,$5B,$6B,$EF,$5F,$FB,$C2,$73,$E5,$F0,$3C,$95,$BA
	.dc.b	$FE,$BA,$D7,$E4,$B5,$FF,$52,$FB,$7C,$DF,$DB,$35,$FC,$F3,$5F,$F6
	.dc.b	$C7,$2A,$31,$1F,$EA,$EE,$9F,$C1,$E1,$FE,$CC,$D7,$F7,$D1,$89,$FF
	.dc.b	$66,$C7,$9B,$86,$57,$F6,$BC,$7F,$0C,$D8,$FE,$0D,$F4,$2F,$56,$92
	.dc.b	$7F,$88,$C6,$5F,$EE,$9E,$73,$E8,$7F,$2C,$D8,$FC,$2F,$D1,$BF,$07
	.dc.b	$CD,$49,$E2,$FD,$DA,$E1,$A7,$F8,$0D,$AF,$F8,$21,$CF,$57,$F5,$C8
	.dc.b	$47,$F1,$8A,$FC,$9B,$A7,$F3,$2E,$63,$F8,$2F,$82,$F9,$A7,$C6,$7F
	.dc.b	$32,$87,$CA,$7F,$74,$C3,$F9,$34,$6E,$75,$39,$BE,$77,$CA,$F7,$39
	.dc.b	$C7,$FB,$8E,$1C,$67,$27,$7F,$9A,$7A,$D4,$F1,$60,$FA,$CC,$57,$8A
	.dc.b	$B6,$DA,$DC,$EB,$2F,$E4,$3E,$7B,$E6,$7E,$85,$FF,$3F,$8C,$70,$7E
	.dc.b	$9B,$F2,$4F,$89,$AF,$F8,$1A,$73,$5B,$BE,$38,$7F,$8E,$27,$2D,$8C
	.dc.b	$BF,$C1,$A3,$2D,$8E,$35,$FE,$12,$EE,$7C,$A3,$E5,$7F,$B0,$E1,$D2
	.dc.b	$B5,$39,$C7,$FA,$B5,$B2,$7D,$0F,$0F,$A9,$7C,$AF,$AE,$30,$DB,$F8
	.dc.b	$13,$A7,$9D,$72,$73,$3F,$43,$E3,$A4,$7E,$03,$8D,$F6,$38,$E0,$FB
	.dc.b	$F5,$75,$FB,$B5,$AB,$E1,$3F,$CD,$FD,$27,$83,$4F,$D2,$FC,$65,$AF
	.dc.b	$E5,$33,$A7,$FC,$BC,$DC,$7F,$CC,$8D,$7F,$CD,$2D,$7F,$9A,$BB,$C7
	.dc.b	$F4,$8B,$E4,$FB,$4F,$F6,$43,$FB,$F3,$F8,$EF,$83,$97,$EF,$F6,$AD
	.dc.b	$7E,$72,$6E,$1E,$E9,$B4,$A7,$13,$F2,$47,$D1,$FF,$82,$7C,$8F,$D1
	.dc.b	$5E,$EB,$53,$98,$FE,$6F,$5C,$1F,$F0,$35,$70,$BE,$B3,$97,$FB,$84
	.dc.b	$65,$F0,$F8,$DF,$D9,$70,$FE,$37,$1F,$A2,$F0,$DF,$27,$9A,$98,$FF
	.dc.b	$A2,$EF,$77,$FE,$94,$6B,$FD,$81,$AF,$A6,$7F,$A6,$F0,$BE,$7B,$9C
	.dc.b	$FF,$4C,$8C,$5F,$E2,$78,$7F,$C5,$5F,$42,$F8,$9F,$DE,$C6,$1B,$3F
	.dc.b	$E2,$59,$D3,$C4,$C4,$7F,$04,$BF,$E0,$AD,$8F,$CB,$36,$3F,$89,$B7
	.dc.b	$87,$F2,$AF,$F8,$EF,$C9,$FF,$13,$4F,$F8,$AD,$AF,$F4,$E6,$AF,$35
	.dc.b	$18,$6F,$BF,$FC,$9F,$91,$CF,$D6,$30,$F7,$89,$97,$F2,$A7,$6F,$F9
	.dc.b	$0E,$FF,$CB,$EF,$86,$3F,$A8,$91,$FE,$9E,$FF,$80,$7E,$9F,$FA,$1E
	.dc.b	$3F,$E1,$46,$BF,$93,$6A,$69,$FD,$33,$E0,$DF,$4D,$C7,$E5,$FF,$D4
	.dc.b	$B1,$8A,$E5,$CD,$8A,$D6,$3F,$EC,$07,$FC,$47,$F4,$FF,$E2,$DE,$2C
	.dc.b	$5F,$E8,$E3,$0F,$FB,$3A,$7C,$DF,$51,$CE,$35,$CD,$FE,$6F,$74,$62
	.dc.b	$F9,$AB,$E7,$97,$96,$3D,$97,$CD,$D5,$F3,$9D,$57,$0C,$3C,$F3,$97
	.dc.b	$C7,$8F,$69,$D5,$F8,$4F,$19,$D3,$1F,$1E,$8C,$BC,$15,$C2,$B4,$70
	.dc.b	$71,$3B,$BE,$D8,$F8,$6B,$B6,$5E,$68,$C4,$7B,$27,$78,$C8,$E9,$3A
	.dc.b	$BE,$91,$89,$D5,$F3,$7C,$F8,$F9,$AF,$AE,$3E,$73,$CB,$19,$9E,$39
	.dc.b	$7C,$67,$8E,$58,$C7,$CF,$39,$7E,$2E,$5E,$2C,$BC,$11,$E3,$23,$F5
	.dc.b	$E3,$53,$C6,$36,$47,$E9,$CF,$07,$44,$7F,$39,$87,$9D,$D1,$E4,$33
	.dc.b	$7E,$77,$E9,$D7,$57,$27,$03,$35,$D5,$FA,$06,$71,$F8,$D9,$7C,$FC
	.dc.b	$3C,$17,$EE,$AF,$F4,$19,$F8,$02,$D7,$E8,$1B,$5F,$9D,$73,$ED,$FF
	.dc.b	$A6,$1E,$27,$E7,$5C,$3A,$F5,$72,$E5,$B5,$FE,$9F,$7F,$31,$3C,$EF
	.dc.b	$F3,$ED,$C7,$FD,$2F,$F5,$0F,$A7,$FF,$4C,$D6,$AF,$89,$31,$FC,$FD
	.dc.b	$6D,$C4,$ED,$FC,$E6,$1D,$5F,$EC,$13,$B7,$35,$C3,$0F,$E0,$B5,$B9
	.dc.b	$56,$EF,$A1,$2B,$77,$FC,$02,$75,$18,$AE,$A4,$AF,$E0,$53,$A8,$FE
	.dc.b	$21,$7D,$F1,$F8,$D1,$BB,$F8,$78,$FC,$27,$A6,$1E,$19,$E3,$8F,$86
	.dc.b	$B5,$18,$39,$16,$7C,$31,$A8,$FA,$C6,$8B,$1A,$7E,$21,$C6,$BC,$27
	.dc.b	$91,$C9,$C1,$CB,$F3,$A7,$B5,$FB,$2B,$D1,$3B,$34,$72,$8D,$57,$28
	.dc.b	$FA,$35,$CB,$0F,$8C,$4C,$3E,$04,$F1,$8C,$3F,$84,$9E,$13,$93,$71
	.dc.b	$EF,$BE,$EF,$B4,$F1,$9F,$CA,$BF,$02,$FF,$24,$BF,$77,$C0,$F9,$19
	.dc.b	$FF,$89,$A3,$F9,$FC,$E9,$E1,$E2,$FE,$89,$46,$5F,$D8,$B5,$FF,$7A
	.dc.b	$B5,$FF,$62,$63,$FA,$AE,$8F,$F2,$8B,$7B,$3F,$31,$3F,$EF,$C9,$CB
	.dc.b	$FB,$46,$AE,$B7,$17,$FB,$4A,$BF,$41,$18,$9E,$E7,$F7,$DC,$FF,$C8
	.dc.b	$8F,$1A,$E1,$EE,$E2,$BF,$D4,$7D,$1F,$06,$BF,$D6,$6D,$8E,$98,$FF
	.dc.b	$9A,$2B,$85,$F2,$BF,$36,$29,$F6,$12,$3F,$86,$FA,$6C,$7F,$D0,$F1
	.dc.b	$FE,$7B,$F6,$39,$A9,$FF,$61,$46,$1E,$1E,$65,$AE,$35,$FE,$BE,$AD
	.dc.b	$79,$3F,$EE,$58,$FF,$0E,$56,$FE,$1B,$AE,$8F,$0B,$F9,$3E,$DE,$C6
	.dc.b	$ED,$FE,$69,$D3,$CD,$7C,$71,$FF,$21,$1C,$AB,$6F,$9F,$44,$FA,$A1
	.dc.b	$F4,$8E,$7F,$6F,$8B,$F1,$7D,$2B,$C3,$FE,$C3,$BF,$F3,$C6,$BF,$EE
	.dc.b	$9C,$38,$46,$35,$FB,$31,$88,$FE,$6E,$D7,$FA,$0B,$5D,$E3,$F9,$78
	.dc.b	$E1,$5D,$63,$CD,$18,$AF,$F4,$47,$A3,$E8,$FA,$3F,$3F,$96,$FA,$AF
	.dc.b	$93,$F5,$1F,$43,$2F,$C1,$B7,$0F,$F4,$55,$6F,$3F,$F6,$2B,$5F,$C8
	.dc.b	$1F,$E5,$99,$E0,$DE,$4E,$F8,$7D,$AF,$A6,$93,$FE,$BF,$7F,$01,$E1
	.dc.b	$E2,$3F,$DF,$B1,$72,$FC,$CC,$FB,$3C,$DF,$6F,$AB,$2F,$E5,$1F,$EA
	.dc.b	$5D,$FF,$D2,$2E,$0F,$AF,$3F,$1C,$8F,$F6,$08,$DB,$F4,$23,$B3,$C6
	.dc.b	$33,$7F,$B1,$87,$D1,$BE,$53,$F2,$BE,$C4,$9F,$F6,$D3,$F9,$BF,$C0
	.dc.b	$F1,$1F,$1B,$AD,$DB,$FE,$F3,$8D,$9F,$E7,$79,$FC,$5F,$B1,$FE,$51
	.dc.b	$CB,$F8,$07,$37,$E3,$FE,$47,$F5,$6D,$BF,$F7,$6F,$93,$C0,$CF,$4A
	.dc.b	$A9,$99,$F8,$F4,$FC,$97,$EB,$F6,$B8,$DF,$03,$F9,$C6,$DF,$0A,$FB
	.dc.b	$B7,$C7,$A3,$FC,$12,$D9,$FF,$07,$B5,$FF,$0C,$B5,$FE,$2E,$E7,$4F
	.dc.b	$6D,$74,$3C,$7A,$FE,$A6,$BE,$47,$C0,$3E,$23,$F8,$43,$63,$F0,$3C
	.dc.b	$48,$C4,$7F,$1E,$F2,$37,$52,$37,$3A,$C7,$F9,$2D,$7F,$11,$FA,$5F
	.dc.b	$E1,$FE,$44,$BF,$F5,$54,$65,$EF,$CA,$9D,$47,$F2,$C9,$FE,$0D,$F7
	.dc.b	$7F,$D3,$70,$FC,$47,$4A,$7F,$85,$36,$BF,$2D,$5B,$F0,$7F,$9D,$BF
	.dc.b	$D3,$60,$D2,$79,$9F,$D5,$DE,$10,$37,$1E,$23,$F8,$C7,$DE,$FF,$B2
	.dc.b	$23,$F4,$FE,$57,$A7,$C2,$FE,$AC,$48,$FE,$7E,$FF,$63,$FA,$7F,$E2
	.dc.b	$11,$D5,$A9,$A6,$A5,$F7,$BF,$91,$9F,$EB,$CD,$7F,$CB,$71,$AE,$BF
	.dc.b	$ED,$13,$99,$FE,$CD,$E2,$D7,$D9,$48,$C4,$7F,$0F,$AF,$EE,$11,$88
	.dc.b	$F2,$72,$7A,$48,$FF,$97,$B0,$FC,$2B,$FE,$6C,$7F,$05,$AF,$F6,$A6
	.dc.b	$C7,$BE,$7F,$A6,$FD,$47,$F6,$7A,$FE,$DF,$F7,$2F,$8D,$FE,$83,$8B
	.dc.b	$D7,$A7,$F4,$79,$D6,$3F,$C4,$DA,$FF,$9D,$1A,$FF,$9F,$9A,$F6,$AF
	.dc.b	$F0,$A4,$70,$E4,$FE,$24,$DC,$35,$5F,$73,$E9,$DF,$47,$E1,$C7,$FB
	.dc.b	$2C,$7F,$69,$9F,$E6,$FB,$A3,$18,$F4,$AE,$15,$FE,$84,$C7,$D8,$D8
	.dc.b	$F1,$91,$F6,$3C,$1E,$BF,$5C,$F8,$AF,$EE,$BF,$34,$F9,$AF,$F0,$1E
	.dc.b	$2F,$A5,$F3,$E1,$FC,$EA,$31,$AF,$0A,$93,$FE,$27,$BE,$4D,$8F,$F1
	.dc.b	$84,$62,$3F,$1A,$DA,$FF,$0E,$B7,$6F,$C5,$B6,$5B,$21,$F3,$BF,$E0
	.dc.b	$CF,$A5,$7D,$4F,$A3,$47,$C6,$53,$51,$31,$FF,$1E,$57,$06,$C7,$F9
	.dc.b	$16,$31,$1F,$A2,$6E,$7F,$E3,$26,$E7,$FA,$06,$CF,$99,$F8,$EF,$DA
	.dc.b	$F9,$2F,$EA,$95,$FE,$45,$F1,$5F,$28,$FB,$0F,$F1,$73,$6F,$F8,$0D
	.dc.b	$F0,$C7,$F5,$AF,$F5,$D8,$FF,$2B,$35,$CC,$7F,$97,$63,$11,$F2,$1E
	.dc.b	$4D,$8F,$F1,$9C,$63,$77,$D7,$FC,$8A,$FE,$23,$F4,$AF,$B4,$FF,$1C
	.dc.b	$CB,$E7,$29,$FE,$3C,$9D,$57,$0A,$FF,$37,$46,$23,$F8,$5B,$5F,$F3
	.dc.b	$73,$5F,$F3,$2A,$FA,$8F,$F3,$74,$7F,$65,$FA,$5F,$F3,$4F,$D4,$7E
	.dc.b	$02,$3F,$95,$78,$69,$CF,$37,$F9,$BA,$31,$3F,$CB,$F0,$E9,$C5,$C2
	.dc.b	$FB,$E3,$DB,$E9,$F7,$6C,$BB,$28,$CC,$61,$DE,$1D,$27,$57,$D3,$0F
	.dc.b	$E7,$D5,$C3,$2F,$F4,$0B,$6B,$35,$B7,$89,$E7,$C7,$C9,$1E,$C8,$D9
	.dc.b	$FD,$27,$1F,$14,$71,$AF,$1D,$70,$AD,$5F,$82,$B9,$47,$61,$8F,$A6
	.dc.b	$F8,$E3,$DA,$7E,$15,$F1,$79,$99,$AE,$17,$E9,$CB,$D9,$3F,$5C,$D9
	.dc.b	$4E,$F8,$FD,$C8,$EF,$1E,$3C,$BD,$6F,$0C,$B2,$4B,$F4,$BD,$8C,$47
	.dc.b	$C7,$CB,$FD,$5A,$FC,$B2,$E3,$F5,$5B,$8F,$E3,$1C,$FB,$6F,$A8,$E3
	.dc.b	$1F,$87,$8C,$38,$AF,$F2,$54,$EA,$3E,$4C,$E9,$C7,$C7,$7F,$B5,$53
	.dc.b	$48,$FF,$B6,$5F,$9D,$EA,$FF,$AC,$F9,$3E,$84,$FF,$81,$67,$3F,$0F
	.dc.b	$FE,$CB,$D3,$F2,$B3,$93,$ED,$5F,$17,$EE,$19,$8F,$CD,$C6,$5E,$31
	.dc.b	$E4,$3E,$A4,$71,$DF,$37,$FA,$59,$FF,$49,$DF,$FB,$6E,$73,$1E,$08
	.dc.b	$F2,$6D,$F2,$2B,$CF,$8F,$AA,$BB,$63,$E2,$BD,$C7,$D6,$AD,$BE,$83
	.dc.b	$77,$FE,$50,$9D,$C7,$FB,$A2,$77,$5E,$23,$A5,$F7,$8C,$DF,$8E,$BC
	.dc.b	$D3,$96,$9D,$E7,$6F,$8C,$CD,$7F,$08,$E4,$FE,$9B,$E8,$FF,$2D,$6D
	.dc.b	$FE,$19,$B5,$FE,$26,$E2,$FF,$27,$4E,$5B,$1F,$DE,$97,$D2,$27,$FC
	.dc.b	$7B,$1A,$E2,$FF,$7E,$BA,$8F,$F2,$0F,$D4,$7D,$D9,$FF,$80,$2D,$8F
	.dc.b	$F2,$57,$E2,$0F,$99,$7C,$A7,$5C,$BF,$C1,$37,$B7,$FB,$4B,$F8,$77
	.dc.b	$FC,$A4,$63,$83,$DD,$19,$8F,$F1,$DB,$5D,$61,$E4,$7B,$63,$E4,$AB
	.dc.b	$3C,$26,$9F,$DE,$3C,$4F,$74,$7F,$91,$9B,$0C,$BF,$C1,$D3,$C3,$5F
	.dc.b	$F0,$96,$D6,$FE,$5D,$FA,$27,$35,$F4,$E7,$33,$D6,$BE,$C4,$76,$AF
	.dc.b	$F0,$E7,$89,$CE,$7A,$3D,$FE,$A3,$E8,$5F,$F8,$C1,$B9,$FC,$BC,$BF
	.dc.b	$C4,$5F,$55,$BE,$19,$A7,$E0,$EF,$8E,$1E,$56,$BF,$E6,$46,$BF,$E1
	.dc.b	$A7,$AC,$67,$EA,$79,$CF,$85,$F8,$5F,$53,$FE,$57,$9F,$F4,$1B,$67
	.dc.b	$FD,$F7,$8B,$F1,$17,$C5,$F0,$F3,$A6,$1F,$97,$C3,$86,$1F,$25,$AF
	.dc.b	$F9,$E1,$AF,$F8,$B9,$ED,$AF,$FB,$AF,$E6,$1F,$0F,$D2,$93,$98,$FB
	.dc.b	$1F,$54,$F9,$1A,$7E,$0B,$57,$E2,$5F,$03,$0F,$F1,$DD,$6F,$3F,$F4
	.dc.b	$FB,$63,$FD,$22,$D7,$F0,$0D,$7F,$D1,$ED,$79,$DF,$E4,$16,$5F,$ED
	.dc.b	$3A,$EB,$39,$8C,$4F,$CA,$F9,$77,$95,$F9,$1F,$AB,$7C,$A7,$A6,$73
	.dc.b	$8F,$FA,$D2,$72,$D9,$FE,$0D,$18,$BF,$F5,$4B,$5F,$CC,$35,$FD,$8F
	.dc.b	$D5,$FF,$A5,$6B,$FD,$89,$BF,$EE,$65,$39,$7F,$D0,$75,$D6,$75,$39
	.dc.b	$7F,$C9,$9F,$0C,$E5,$FD,$8F,$A2,$8C,$63,$FE,$E4,$E8,$CC,$63,$2F
	.dc.b	$F4,$7F,$5F,$C8,$AF,$E3,$CD,$8F,$F6,$1B,$5F,$D9,$35,$F5,$F3,$74
	.dc.b	$EB,$FA,$5A,$7E,$2A,$FF,$DC,$19,$78,$E7,$EB,$7D,$5F,$FA,$46,$7E
	.dc.b	$09,$FC,$51,$C6,$1F,$8E,$8F,$F7,$47,$C7,$FF,$B6,$3C,$FE,$3E,$CC
	.dc.b	$9F,$EE,$97,$C2,$FF,$C0,$6D,$7A,$4F,$F9,$92,$3A,$F0,$7C,$FF,$4D
	.dc.b	$F3,$FC,$EF,$C0,$7F,$BD,$E3,$38,$79,$A3,$F8,$C7,$2F,$28,$F6,$F0
	.dc.b	$FF,$94,$DD,$75,$A7,$54,$DF,$EA,$79,$7F,$48,$97,$C2,$B7,$3A,$8F
	.dc.b	$07,$86,$FC,$77,$FD,$11,$18,$6E,$1F,$9B,$E9,$FE,$55,$18,$DB,$FA
	.dc.b	$25,$F1,$F5,$BF,$03,$F5,$7A,$6E,$3E,$DE,$47,$CC,$BE,$37,$F9,$DC
	.dc.b	$3E,$3D,$F3,$78,$AD,$5F,$19,$CB,$CE,$33,$B7,$CC,$BE,$57,$C7,$0E
	.dc.b	$B8,$70,$AD,$37,$4F,$EB,$97,$C3,$1E,$97,$C3,$0F,$7F,$DE,$49,$F9
	.dc.b	$5F,$35,$72,$FF,$75,$EE,$AE,$FF,$29,$9C,$FA,$E8,$8F,$D2,$FD,$37
	.dc.b	$E9,$BE,$42,$E7,$F8,$1E,$4F,$D6,$46,$63,$1D,$08,$DE,$9F,$8A,$AD
	.dc.b	$D7,$E2,$63,$FD,$83,$F4,$DF,$C5,$FA,$8F,$F1,$9F,$CC,$BF,$0F,$3A
	.dc.b	$AE,$15,$A7,$FC,$A4,$D8,$FF,$2C,$B5,$FF,$31,$B5,$FE,$FE,$EF,$5F
	.dc.b	$E1,$DE,$AF,$F7,$29,$EC,$7E,$80,$FF,$05,$F8,$9F,$8A,$F8,$E9,$87
	.dc.b	$E5,$7D,$3B,$5F,$32,$7F,$85,$FA,$B8,$C7,$F8,$9B,$FE,$02,$F7,$3F
	.dc.b	$5B,$C9,$7C,$2F,$A5,$97,$E1,$79,$31,$32,$3F,$11,$E0,$5C,$FF,$11
	.dc.b	$D9,$FA,$B9,$FD,$96,$3F,$B8,$E9,$E8,$7F,$84,$21,$F8,$2F,$E4,$35
	.dc.b	$FF,$12,$56,$E7,$8E,$8B,$FF,$75,$4E,$71,$FF,$45,$B5,$FF,$13,$B5
	.dc.b	$07,$85,$FF,$39,$ED,$FE,$41,$39,$BE,$D8,$7E,$06,$73,$E0,$7E,$EF
	.dc.b	$E3,$7F,$6E,$7F,$4F,$FB,$6B,$5F,$BD,$8F,$E6,$FC,$B8,$C7,$F1,$9D
	.dc.b	$7F,$E4,$0A,$FE,$0B,$F9,$13,$EB,$BF,$C5,$4D,$8F,$E1,$9F,$5F,$89
	.dc.b	$91,$F9,$8B,$A7,$43,$8F,$BD,$FE,$61,$BF,$F2,$74,$FF,$98,$5B,$3F
	.dc.b	$EB,$0D,$11,$88,$FF,$1B,$3A,$DF,$F8,$8F,$D6,$3C,$1E,$B8,$7F,$1C
	.dc.b	$8C,$36,$3F,$AC,$F4,$FE,$5E,$3E,$7E,$2A,$DE,$3F,$AD,$AD,$29,$D5
	.dc.b	$F2,$EF,$FF,$26,$5F,$0C,$7F,$CC,$DE,$03,$1E,$38,$BE,$AD,$F4,$DF
	.dc.b	$DC,$E3,$FC,$CF,$18,$D3,$FD,$84,$D8,$FF,$64,$B5,$FF,$69,$B9,$C3
	.dc.b	$FB,$E3,$FD,$2A,$3F,$5C,$E6,$7A,$F3,$7F,$86,$7E,$AD,$E9,$FF,$9F
	.dc.b	$2F,$FD,$97,$F4,$7F,$E7,$89,$CF,$37,$FB,$1E,$3C,$BE,$FF,$FA,$07
	.dc.b	$E1,$91,$BC,$39,$46,$30,$FC,$CE,$1C,$9E,$A6,$27,$75,$D3,$2E,$D3
	.dc.b	$9C,$6A,$BE,$58,$75,$8C,$61,$C9,$FD,$3E,$7F,$E5,$6C,$38,$DF,$0C
	.dc.b	$7C,$D8,$E3,$0E,$18,$7D,$3C,$BD,$95,$A8,$F3,$9D,$AB,$81,$E4,$C7
	.dc.b	$B4,$6B,$0E,$CF,$A4,$C6,$1D,$A7,$79,$7A,$6B,$75,$D4,$DD,$F5,$C7
	.dc.b	$C7,$1C,$EB,$86,$1E,$8B,$E3,$87,$4C,$7D,$15,$AC,$7C,$F8,$7A,$AB
	.dc.b	$4F,$3A,$D9,$1F,$93,$8F,$81,$F2,$5F,$37,$CA,$73,$8D,$61,$EA,$26
	.dc.b	$3E,$F8,$CB,$E7,$26,$3E,$DC,$BC,$B8,$75,$26,$3F,$83,$C7,$BC,$61
	.dc.b	$F4,$1E,$69,$E3,$3E,$98,$F3,$BC,$67,$84,$60,$E2,$7D,$F9,$E2,$E3
	.dc.b	$1F,$97,$8F,$A3,$0F,$93,$3F,$ED,$A3,$EB,$16,$7F,$38,$DC,$7F,$54
	.dc.b	$E7,$DF,$FF,$42,$BC,$4F,$CD,$38,$75,$EE,$65,$CB,$6B,$FD,$46,$FE
	.dc.b	$6A,$79,$E5,$F8,$96,$E7,$FE,$EE,$FA,$BF,$F1,$56,$1B,$B8,$FE,$56
	.dc.b	$B6,$FF,$4C,$9D,$B9,$9D,$DF,$57,$F6,$33,$A8,$C4,$EC,$8F,$EC,$E3
	.dc.b	$57,$FC,$26,$FB,$D7,$B0,$DC,$79,$FC,$FF,$EB,$B1,$83,$F8,$3C,$F0
	.dc.b	$C7,$E0,$C7,$EC,$AF,$AB,$AC,$7C,$F8,$77,$7C,$F8,$74,$C7,$CE,$49
	.dc.b	$D5,$73,$9D,$46,$27,$53,$D4,$D1,$D8,$C3,$A5,$7E,$47,$DD,$19,$C3
	.dc.b	$D2,$FC,$5B,$EF,$3C,$AF,$C0,$47,$EA,$46,$9E,$71,$96,$9A,$3C,$86
	.dc.b	$09,$7E,$3B,$F0,$CF,$16,$CF,$13,$E8,$1C,$5C,$BE,$CA,$F1,$46,$09
	.dc.b	$8F,$DC,$BE,$CF,$D7,$26,$3F,$4A,$FF,$9C,$FD,$5F,$F7,$37,$FC,$A8
	.dc.b	$FF,$96,$5B,$E6,$FE,$0B,$8B,$FD,$A9,$3B,$9F,$F2,$33,$5F,$D9,$35
	.dc.b	$FC,$4B,$5F,$F1,$3B,$7A,$5E,$5B,$DF,$ED,$C7,$FE,$C6,$E7,$FF,$54
	.dc.b	$61,$FE,$6F,$6B,$FE,$D6,$9D,$BF,$DF,$EB,$93,$BF,$23,$3D,$3F,$DF
	.dc.b	$5D,$E3,$E0,$BB,$DB,$EF,$BE,$87,$FA,$BB,$0E,$6F,$FB,$0A,$BF,$C3
	.dc.b	$BA,$F0,$3F,$82,$38,$8F,$F1,$24,$62,$BA,$47,$F8,$88,$F1,$E5,$FE
	.dc.b	$8D,$7D,$9E,$57,$89,$87,$F8,$E8,$A7,$FB,$4B,$D6,$FF,$65,$7A,$7E
	.dc.b	$AF,$9C,$E5,$3F,$E2,$1F,$53,$A9,$9B,$91,$FE,$C8,$9D,$F1,$7F,$B0
	.dc.b	$9B,$C3,$FE,$5A,$C7,$FD,$AD,$F7,$3F,$33,$CB,$7D,$4B,$D8,$F9,$F9
	.dc.b	$63,$A2,$C9,$E0,$EE,$3F,$CE,$AD,$7F,$CD,$6D,$7F,$CD,$6D,$7F,$CD
	.dc.b	$75,$E0,$AF,$E6,$0D,$8E,$71,$D8,$FC,$0E,$2F,$CE,$9F,$51,$E9,$FA
	.dc.b	$4F,$F7,$DD,$FF,$B4,$7E,$8B,$F0,$FE,$8F,$FB,$71,$D7,$8F,$27,$FD
	.dc.b	$17,$5C,$1F,$F8,$03,$9A,$31,$C3,$FE,$D2,$F5,$3F,$C3,$CF,$CB,$FC
	.dc.b	$33,$EE,$F9,$13,$0F,$A1,$1F,$5C,$8E,$0F,$F7,$F7,$D4,$FC,$CC,$1F
	.dc.b	$27,$F3,$3D,$69,$5F,$EC,$79,$CB,$67,$FA,$A3,$E5,$6C,$6A,$3F,$AE
	.dc.b	$35,$FE,$CC,$E7,$17,$0F,$F3,$E8,$C3,$6B,$EF,$FD,$D3,$D2,$FC,$CF
	.dc.b	$D5,$CC,$7F,$68,$FE,$73,$49,$18,$21,$FE,$75,$7F,$31,$7F,$4B,$DF
	.dc.b	$F9,$38,$FF,$AA,$B8,$D1,$8B,$FF,$3D,$75,$F9,$1B,$7C,$9F,$CE,$3F
	.dc.b	$C0,$BA,$DE,$2F,$FB,$F6,$7F,$D6,$BB,$FF,$BF,$3D,$1F,$01,$23,$39
	.dc.b	$F8,$79,$1F,$31,$F8,$66,$CE,$86,$BE,$6B,$E5,$1F,$ED,$4A,$E0,$63
	.dc.b	$0F,$04,$71,$A5,$F4,$8F,$E1,$9A,$3E,$53,$F8,$2F,$9D,$27,$F7,$11
	.dc.b	$FB,$F2,$38,$3C,$9A,$FC,$0F,$B2,$FE,$1B,$E5,$49,$FE,$50,$FC,$5D
	.dc.b	$DF,$65,$4D,$4F,$FA,$A6,$70,$FB,$C7,$D4,$BD,$8F,$E0,$3F,$55,$FC
	.dc.b	$9F,$83,$F1,$35,$FE,$E2,$8C,$71,$FF,$82,$2F,$84,$F3,$6B,$FE,$2F
	.dc.b	$6B,$FE,$39,$6B,$C6,$FF,$3B,$1C,$2F,$BC,$F9,$27,$F8,$9C,$E6,$BF
	.dc.b	$15,$BF,$F0,$FB,$DF,$D5,$7F,$15,$C3,$FA,$57,$D9,$7F,$32,$3F,$8D
	.dc.b	$C6,$63,$E9,$FC,$E7,$F3,$CB,$34,$91,$FE,$47,$E4,$FF,$77,$7C,$17
	.dc.b	$F0,$CF,$7C,$D3,$1F,$0D,$DB,$E1,$A5,$FC,$DC,$91,$FD,$53,$EA,$FF
	.dc.b	$AE,$EB,$F3,$3D,$54,$7F,$60,$BF,$F1,$9C,$63,$97,$FA,$8C,$E6,$7F
	.dc.b	$DC,$AD,$5C,$EF,$97,$FE,$DF,$1F,$9B,$BE,$4D,$AF,$0F,$17,$F4,$3A
	.dc.b	$DC,$FF,$61,$FA,$A7,$D4,$7F,$42,$FA,$A7,$C4,$D3,$FB,$6C,$EA,$BB
	.dc.b	$35,$FF,$3B,$64,$8C,$47,$B1,$AF,$FB,$F9,$AF,$FB,$ED,$E5,$AF,$C5
	.dc.b	$6E,$5B,$C7,$F8,$17,$CE,$7F,$5C,$B6,$0F,$AA,$7C,$A7,$A7,$E1,$A5
	.dc.b	$7C,$39,$CB,$67,$FB,$EC,$62,$3D,$2D,$7E,$9B,$51,$E0,$78,$67,$F6
	.dc.b	$DB,$96,$B0,$F2,$5F,$F1,$9F,$AA,$CE,$1E,$8F,$97,$FE,$08,$FD,$75
	.dc.b	$F4,$1F,$E1,$A6,$C6,$7C,$4C,$1F,$E1,$48,$E3,$1F,$41,$AF,$E2,$9A
	.dc.b	$F8,$DD,$E7,$F9,$16,$E5,$9A,$F3,$7C,$EF,$F8,$AB,$EA,$3E,$8E,$1F
	.dc.b	$39,$7D,$4E,$4D,$70,$C9,$FF,$57,$6C,$AF,$BA,$D9,$FF,$15,$C6,$0E
	.dc.b	$B1,$99,$FC,$C4,$78,$39,$FF,$AA,$7B,$3F,$E3,$CF,$A9,$7D,$57,$F8
	.dc.b	$FB,$3F,$E0,$FB,$FE,$AD,$BB,$7D,$1A,$DE,$5F,$A3,$6B,$F9,$66,$BF
	.dc.b	$B0,$6B,$C6,$FF,$C8,$31,$C2,$FF,$70,$DA,$D5,$7F,$36,$FC,$5B,$FC
	.dc.b	$97,$5F,$BD,$FA,$96,$3F,$E5,$45,$F3,$F4,$C4,$4B,$F9,$33,$C1,$B1
	.dc.b	$FE,$52,$8C,$47,$FB,$6D,$FE,$B7,$5F,$94,$FA,$87,$CF,$FF,$AF,$3E
	.dc.b	$AB,$F0,$B1,$8B,$FC,$CC,$62,$7F,$D8,$37,$CA,$55,$B5,$19,$8C,$57
	.dc.b	$19,$E7,$C9,$FE,$CA,$9E,$15,$BB,$EB,$E9,$B1,$F3,$47,$96,$F8,$37
	.dc.b	$0C,$9F,$AE,$BE,$4D,$BE,$4F,$9A,$78,$34,$F1,$9C,$8F,$4C,$70,$C3
	.dc.b	$CA,$D9,$E3,$1E,$CB,$F1,$D7,$7C,$39,$61,$E2,$86,$3F,$05,$B5,$C1
	.dc.b	$EF,$5C,$31,$F3,$F9,$3D,$67,$F8,$4B,$5F,$EA,$39,$7C,$6A,$CB,$63
	.dc.b	$D8,$76,$3D,$98,$78,$71,$F2,$1E,$57,$C8,$67,$0F,$74,$FA,$CF,$C4
	.dc.b	$3F,$29,$FF,$80,$27,$A0,$CD,$E1,$FA,$5E,$AB,$EA,$B1,$1F,$84,$8C
	.dc.b	$3F,$93,$6B,$D3,$0F,$F2,$FC,$FF,$96,$A1,$F2,$B4,$8F,$F9,$79,$BA
	.dc.b	$FE,$47,$E0,$5F,$19,$FE,$12,$8F,$CA,$57,$F3,$2A,$FB,$D5,$AE,$3F
	.dc.b	$C0,$4E,$4F,$93,$19,$7E,$59,$9A,$ED,$7D,$76,$F8,$18,$7D,$2A,$FF
	.dc.b	$4B,$CF,$F8,$8B,$7F,$0C,$7A,$8F,$B2,$6B,$1E,$0F,$F4,$63,$58,$76
	.dc.b	$8F,$2C,$63,$0D,$63,$E5,$AE,$B8,$FB,$31,$F2,$5F,$D6,$8C,$61,$C9
	.dc.b	$B1,$EC,$3A,$1E,$33,$0E,$AB,$AC,$7C,$4A,$EA,$F9,$E7,$58,$7C,$28
	.dc.b	$D5,$F7,$69,$D0,$B7,$FC,$37,$A7,$FC,$BD,$F6,$1F,$E3,$A9,$FE,$C5
	.dc.b	$C4,$8C,$47,$F7,$88,$C5,$FF,$48,$6C,$7F,$82,$9B,$C1,$FE,$52,$F9
	.dc.b	$1F,$C4,$DF,$F9,$52,$3F,$C5,$53,$AB,$FE,$EB,$5A,$7F,$C4,$5F,$56
	.dc.b	$F9,$BE,$11,$FD,$D6,$30,$FF,$89,$E9,$F4,$DA,$CB,$FC,$1B,$5A,$7F
	.dc.b	$8D,$57,$1B,$F5,$E6,$AD,$57,$F8,$E9,$AE,$EB,$FB,$2C,$73,$CB,$D3
	.dc.b	$F4,$FD,$EF,$96,$17,$24,$E7,$C3,$7D,$8D,$31,$F6,$77,$3F,$F2,$04
	.dc.b	$E9,$FE,$E6,$DD,$3F,$C6,$AD,$7F,$C9,$ED,$7F,$C3,$0E,$F3,$FA,$1C
	.dc.b	$BF,$D3,$2B,$AE,$1F,$17,$7F,$E3,$31,$FE,$A9,$A7,$DB,$FB,$63,$FD
	.dc.b	$6A,$DC,$7F,$C6,$F8,$BC,$6F,$9D,$7C,$36,$B7,$F2,$2F,$FA,$43,$67
	.dc.b	$58,$7F,$15,$6C,$FF,$1A,$69,$FE,$6D,$C1,$E1,$74,$C3,$E5,$3F,$3F
	.dc.b	$EB,$5F,$25,$4F,$C0,$60,$F9,$B7,$9B,$9C,$FF,$CD,$97,$C1,$FF,$0F
	.dc.b	$B6,$BF,$CC,$2D,$7F,$CF,$6D,$7F,$C8,$8E,$F3,$FC,$E7,$E0,$3F,$EC
	.dc.b	$5A,$EB,$8F,$CD,$8C,$5F,$E4,$0F,$F2,$FF,$D7,$3E,$CB,$F1,$0D,$BF
	.dc.b	$F5,$E6,$7D,$5F,$F2,$94,$66,$B9,$73,$A3,$11,$FE,$90,$6B,$C2,$FF
	.dc.b	$3B,$18,$8F,$0C,$7C,$9F,$4F,$C2,$DB,$FF,$66,$7D,$73,$EF,$7F,$39
	.dc.b	$F6,$1F,$E7,$49,$CC,$63,$D3,$91,$FE,$48,$9D,$63,$FE,$8A,$6B,$FE
	.dc.b	$A4,$6B,$FB,$EF,$0D,$E3,$FE,$4D,$B5,$FE,$A1,$AE,$7C,$3F,$C2,$67
	.dc.b	$37,$F2,$C6,$5F,$E9,$EB,$F4,$E9,$E8,$F9,$8F,$C1,$CE,$6B,$E4,$FC
	.dc.b	$4A,$7F,$80,$B6,$FF,$CC,$2D,$9F,$F5,$1B,$5F,$E8,$9D,$7F,$D7,$3C
	.dc.b	$AF,$41,$F1,$BE,$4E,$DF,$FB,$D7,$EC,$39,$6B,$FE,$E4,$6C,$7F,$37
	.dc.b	$F9,$5F,$CB,$C6,$31,$FF,$4A,$BD,$DF,$EA,$4D,$73,$9F,$F3,$25,$19
	.dc.b	$6C,$7F,$21,$F8,$57,$D0,$BE,$73,$BC,$E6,$31,$8F,$FA,$93,$2F,$E6
	.dc.b	$DF,$68,$F8,$FF,$A9,$E1,$7F,$07,$8C,$56,$CF,$B5,$19,$9D,$C6,$73
	.dc.b	$70,$2F,$8C,$7E,$8E,$3F,$C0,$B7,$D7,$DE,$FC,$17,$95,$70,$E5,$5F
	.dc.b	$E0,$A8,$CC,$63,$9B,$F5,$B9,$FA,$2F,$38,$FE,$17,$DD,$7D,$BF,$F0
	.dc.b	$46,$E1,$E5,$97,$D5,$BE,$1B,$57,$2B,$57,$C2,$FF,$21,$7B,$9C,$D6
	.dc.b	$E7,$C0,$FF,$12,$9D,$46,$27,$FD,$F9,$D1,$F0,$B2,$47,$FB,$3E,$30
	.dc.b	$FF,$B5,$BD,$EE,$CF,$0E,$1F,$E9,$D3,$AB,$F9,$BA,$BE,$73,$35,$FA
	.dc.b	$B7,$EC,$7B,$9F,$AD,$F9,$5B,$8F,$C9,$6E,$1F,$4F,$04,$63,$3F,$D8
	.dc.b	$3F,$CB,$3E,$BD,$F4,$54,$C6,$57,$1F,$99,$3A,$7F,$C9,$ED,$C7,$FC
	.dc.b	$A8,$D7,$FC,$B2,$D7,$FC,$60,$EF,$0F,$8B,$1F,$EF,$8F,$C0,$7F,$9E
	.dc.b	$57,$6C,$3F,$9F,$C6,$27,$F1,$9C,$FF,$EB,$8A,$FC,$37,$D5,$7E,$67
	.dc.b	$E0,$7F,$87,$32,$32,$7E,$3A,$78,$57,$C3,$9F,$F0,$57,$BC,$FB,$9F
	.dc.b	$C3,$DF,$E7,$63,$1C,$9F,$AF,$F9,$B7,$D2,$F0,$BD,$E3,$FD,$06,$F7
	.dc.b	$87,$E9,$7C,$17,$E0,$1F,$D5,$A3,$A5,$F0,$E0,$F8,$35,$AF,$53,$FA
	.dc.b	$EF,$D6,$FE,$4E,$7E,$2E,$48,$C6,$1F,$E5,$97,$F8,$0B,$F9,$F7,$30
	.dc.b	$38,$75,$FF,$0B,$8F,$CF,$47,$FA,$31,$AF,$38,$D1,$FD,$BB,$0F,$E8
	.dc.b	$D5,$D6,$31,$E7,$7E,$5A,$21,$EF,$F9,$EF,$D1,$E3,$F7,$3D,$27,$6F
	.dc.b	$E8,$AB,$F4,$A7,$F1,$9F,$23,$F7,$AF,$E9,$4C,$78,$7F,$E2,$FF,$BA
	.dc.b	$7D,$15,$3F,$C3,$9F,$63,$F0,$EB,$F8,$96,$2A,$DC,$61,$BB,$FE,$87
	.dc.b	$E9,$F9,$4F,$F0,$D7,$B6,$FF,$E4,$6A,$F8,$74,$BE,$93,$92,$5F,$F5
	.dc.b	$9F,$47,$8B,$4F,$E3,$1F,$01,$FD,$7C,$FE,$DB,$16,$32,$4B,$FE,$FF
	.dc.b	$E9,$FF,$59,$E0,$FE,$C5,$7F,$E1,$EF,$B0,$76,$FF,$43,$6E,$1F,$EB
	.dc.b	$E6,$E1,$FE,$C3,$6B,$FE,$C8,$73,$7F,$8A,$F0,$2D,$7F,$A8,$66,$D7
	.dc.b	$EC,$6F,$FC,$5B,$F6,$1F,$E2,$CE,$8F,$F5,$D4,$7F,$A9,$E3,$13,$FE
	.dc.b	$C6,$8E,$5F,$53,$FB,$AF,$8B,$F3,$CF,$E4,$6B,$71,$88,$46,$2B,$A7
	.dc.b	$5F,$E7,$67,$A1,$F9,$9C,$58,$75,$AD,$E3,$E4,$AE,$D7,$C7,$1E,$D8
	.dc.b	$F5,$9E,$58,$71,$77,$7C,$5F,$1D,$F1,$C3,$AC,$78,$86,$3E,$2C,$B8
	.dc.b	$63,$D6,$78,$E1,$8C,$39,$47,$82,$B1,$32,$3A,$57,$1B,$E9,$18,$BE
	.dc.b	$2F,$43,$15,$E7,$9D,$E5,$EE,$AD,$DF,$8C,$E9,$97,$BF,$0E,$55,$BC
	.dc.b	$3C,$4F,$D3,$C3,$9E,$5F,$02,$35,$5E,$03,$E1,$4E,$9F,$92,$66,$36
	.dc.b	$73,$BF,$06,$1E,$3C,$7C,$A7,$B2,$F9,$5F,$93,$0F,$01,$1F,$90,$75
	.dc.b	$70,$EC,$CE,$1E,$08,$CD,$FC,$0C,$BD,$B1,$8C,$3C,$18,$F9,$1F,$1E
	.dc.b	$5A,$C7,$9D,$6E,$31,$1F,$80,$AE,$11,$F6,$63,$F9,$61,$F0,$8A,$7D
	.dc.b	$72,$E1,$FA,$66,$FB,$DF,$E9,$77,$67,$E8,$9C,$3A,$F7,$72,$E5,$B3
	.dc.b	$FE,$A2,$7F,$23,$3C,$F3,$FC,$FB,$71,$F2,$FD,$83,$E9,$BC,$F5,$AB
	.dc.b	$E2,$4C,$7F,$3D,$5B,$71,$3B,$7F,$2F,$87,$57,$FA,$64,$ED,$CD,$70
	.dc.b	$C3,$F8,$35,$6E,$55,$BB,$E8,$4A,$DD,$FE,$CE,$75,$18,$AE,$A4,$AF
	.dc.b	$E1,$53,$A8,$FE,$1D,$7D,$F1,$F8,$D1,$BB,$F8,$78,$FC,$27,$A6,$1E
	.dc.b	$19,$E3,$8F,$86,$B5,$18,$39,$16,$7C,$31,$A8,$FA,$C6,$8B,$1A,$7E
	.dc.b	$21,$C6,$BC,$27,$91,$C9,$C1,$CB,$F3,$A7,$B5,$FB,$2B,$D1,$3B,$34
	.dc.b	$72,$8D,$57,$28,$FA,$35,$CB,$0F,$8C,$4C,$3E,$04,$F1,$8C,$3F,$84
	.dc.b	$9E,$13,$93,$71,$EF,$BE,$EF,$B4,$F1,$9F,$CA,$7E,$C1,$F6,$DF,$A8
	.dc.b	$E2,$FF,$0D,$FE,$07,$FF,$00,$2B,$FC,$EC,$D7,$FD,$81,$5C,$1F,$F3
	.dc.b	$D3,$5F,$D4,$B5,$FD,$8B,$5F,$DA,$B5,$FF,$4A,$F9,$2F,$A1,$F1,$71
	.dc.b	$FF,$89,$7D,$7F,$F4,$C4,$62,$B6,$DE,$B7,$D2,$BE,$33,$57,$C2,$FF
	.dc.b	$C9,$97,$8C,$7B,$47,$F8,$52,$F8,$3F,$E1,$3F,$B2,$E9,$18,$8F,$F7
	.dc.b	$FB,$75,$FF,$14,$77,$AB,$C4,$DA,$FE,$13,$9F,$F8,$7E,$7F,$83,$B6
	.dc.b	$33,$3B,$7F,$8A,$C7,$F8,$95,$F3,$7D,$24,$7F,$C4,$C7,$62,$18,$3F
	.dc.b	$D6,$5E,$8F,$C3,$FB,$AF,$F3,$0C,$7F,$97,$7E,$3D,$5F,$27,$CC,$46
	.dc.b	$27,$FC,$6B,$5C,$16,$7F,$E4,$58,$C3,$7D,$0E,$C7,$3F,$31,$E3,$FD
	.dc.b	$5B,$7D,$8B,$E0,$F0,$BD,$5A,$62,$24,$FF,$2C,$9E,$0D,$8F,$E5,$EF
	.dc.b	$46,$C6,$A3,$CF,$CF,$FD,$02,$31,$97,$F9,$D1,$B1,$9F,$84,$FC,$57
	.dc.b	$A8,$F0,$1E,$4F,$C5,$C9,$F3,$94,$DC,$EB,$3F,$F4,$83,$5F,$F4,$63
	.dc.b	$73,$FF,$53,$E1,$D7,$0F,$A1,$EB,$7F,$8B,$67,$FD,$FB,$EB,$FE,$4C
	.dc.b	$98,$7F,$A8,$7E,$97,$F2,$C6,$1F,$F5,$5C,$61,$FF,$74,$7D,$83,$E3
	.dc.b	$3E,$87,$D8,$7F,$BB,$A7,$FD,$D8,$D7,$E2,$9F,$E8,$F9,$E0,$F2,$9F
	.dc.b	$9B,$18,$8F,$EC,$2D,$7F,$B4,$B5,$FE,$DA,$EF,$0F,$F5,$0B,$78,$33
	.dc.b	$5F,$83,$FA,$DF,$CF,$57,$FB,$2B,$C2,$78,$3F,$47,$1F,$F6,$07,$D9
	.dc.b	$7E,$D4,$FD,$B7,$0C,$8C,$90,$FF,$66,$BF,$9F,$F8,$EF,$DE,$7C,$64
	.dc.b	$9F,$CD,$C6,$5A,$43,$E1,$FC,$33,$EA,$1F,$11,$F9,$95,$F3,$B4,$95
	.dc.b	$FE,$B7,$D7,$FD,$7F,$2F,$98,$CC,$FC,$E3,$A5,$62,$B7,$37,$87,$FC
	.dc.b	$2B,$7D,$AF,$A4,$62,$FC,$55,$FC,$5F,$AA,$57,$FB,$95,$FE,$35,$EB
	.dc.b	$C8,$C1,$0F,$F7,$8B,$FC,$7F,$DB,$FE,$41,$F6,$D2,$7F,$DD,$D1,$97
	.dc.b	$FD,$E1,$19,$7F,$DF,$31,$FB,$BF,$41,$F3,$9F,$C2,$7E,$C3,$F7,$F5
	.dc.b	$FE,$E9,$6B,$F0,$67,$6F,$F4,$89,$FD,$CC,$62,$3F,$C5,$AD,$7F,$C6
	.dc.b	$AD,$7F,$C7,$2E,$F6,$FE,$A7,$3E,$33,$F1,$9E,$FF,$F3,$BF,$84,$97
	.dc.b	$F2,$8F,$E9,$6F,$F4,$5F,$AB,$7D,$87,$F2,$5F,$0B,$F9,$9F,$97,$FB
	.dc.b	$9B,$FF,$23,$FA,$08,$C4,$7F,$BD,$63,$13,$FC,$4B,$EB,$5E,$27,$F1
	.dc.b	$F6,$D7,$84,$38,$7F,$63,$4C,$5F,$23,$FC,$CF,$EB,$9E,$35,$25,$FF
	.dc.b	$BA,$2B,$F9,$8F,$4F,$8D,$FE,$6D,$1E,$86,$B9,$F3,$FF,$A7,$C6,$20
	.dc.b	$79,$3C,$DF,$89,$32,$3D,$19,$FF,$35,$7F,$DE,$FF,$5B,$FD,$7F,$0F
	.dc.b	$85,$F7,$DF,$E6,$2E,$54,$63,$DE,$FE,$E5,$5B,$BF,$F7,$A3,$FD,$A5
	.dc.b	$F9,$AD,$7E,$63,$5D,$FB,$7F,$D9,$A3,$97,$01,$A7,$59,$7D,$7B,$B7
	.dc.b	$FC,$05,$F2,$CF,$99,$FF,$00,$FD,$FF,$E2,$74,$FF,$05,$37,$06,$4B
	.dc.b	$E1,$1F,$89,$8C,$5F,$F7,$89,$CC,$FF,$76,$6B,$F8,$86,$BF,$87,$77
	.dc.b	$AF,$C6,$8E,$F1,$8C,$35,$5F,$D9,$F2,$FE,$45,$5F,$CD,$AD,$F5,$FF
	.dc.b	$D4,$A3,$C9,$F5,$CF,$8F,$FF,$10,$36,$3F,$41,$5C,$3D,$6F,$F1,$33
	.dc.b	$63,$FD,$36,$D7,$FD,$46,$D7,$FC,$81,$7D,$DC,$B6,$3D,$47,$12,$B9
	.dc.b	$CB,$F4,$D7,$FC,$AA,$BC,$FE,$DF,$F8,$CB,$D1,$FD,$16,$1F,$D0,$5B
	.dc.b	$7C,$27,$D6,$7F,$89,$AF,$FC,$2B,$7F,$E2,$A6,$BF,$E2,$66,$BF,$B8
	.dc.b	$73,$7F,$3F,$64,$62,$FB,$57,$CA,$F0,$BF,$C9,$1F,$5C,$FB,$CF,$F2
	.dc.b	$57,$D6,$FE,$0E,$BF,$04,$D7,$F7,$33,$B7,$F0,$CD,$7F,$C7,$2D,$7F
	.dc.b	$85,$B5,$FE,$14,$D7,$F2,$2F,$1C,$BD,$0D,$C3,$77,$E4,$F9,$CF,$2F
	.dc.b	$83,$FE,$5C,$9F,$F1,$C6,$1F,$DF,$36,$FD,$9F,$C0,$EA,$3D,$35,$FE
	.dc.b	$50,$CF,$F8,$94,$7F,$1E,$6B,$FC,$75,$AF,$F0,$65,$F4,$1F,$E4,$28
	.dc.b	$7D,$67,$F5,$75,$F5,$CF,$97,$FF,$77,$CE,$63,$13,$FC,$42,$38,$5F
	.dc.b	$4D,$F1,$E9,$7F,$BD,$A3,$15,$DA,$3F,$DF,$37,$CA,$30,$7F,$1A,$71
	.dc.b	$7D,$31,$F3,$37,$0E,$B5,$BC,$7C,$B8,$74,$6C,$7B,$EF,$8C,$F8,$DB
	.dc.b	$1C,$4E,$67,$C0,$9E,$2B,$1F,$4E,$1E,$18,$F3,$9B,$38,$E3,$E8,$C3
	.dc.b	$B6,$3D,$2F,$8C,$79,$6B,$78,$76,$9C,$C7,$9E,$BF,$9D,$F9,$1F,$3E
	.dc.b	$BF,$9E,$CF,$F9,$6A,$F9,$34,$E0,$53,$D1,$97,$B9,$F3,$1E,$8C,$7C
	.dc.b	$CF,$0C,$BD,$87,$77,$CA,$7A,$67,$96,$5F,$6A,$3C,$67,$6C,$AB,$E2
	.dc.b	$33,$1B,$7F,$05,$EB,$3E,$BB,$71,$F3,$63,$0F,$29,$46,$23,$9B,$5E
	.dc.b	$FF,$23,$FE,$88,$4D,$23,$FE,$8C,$7E,$EF,$C1,$3E,$83,$F9,$C6,$7F
	.dc.b	$9F,$AF,$ED,$35,$A9,$CE,$BD,$A7,$EF,$DF,$1E,$3F,$C3,$C7,$03,$35
	.dc.b	$C1,$F8,$B7,$F6,$AF,$EC,$DF,$19,$FA,$C7,$1E,$0F,$D8,$46,$5F,$24
	.dc.b	$7B,$0E,$B8,$78,$1C,$63,$C3,$1F,$04,$7E,$1B,$1F,$55,$6D,$B1,$F9
	.dc.b	$83,$D1,$5C,$67,$51,$D8,$D1,$DB,$1F,$4D,$F6,$8E,$31,$89,$E3,$5C
	.dc.b	$E7,$8D,$F8,$23,$D6,$FC,$38,$FF,$0B,$E3,$EF,$BF,$9D,$8F,$13,$FC
	.dc.b	$D1,$D3,$FC,$47,$AB,$FB,$75,$FF,$8E,$B8,$BF,$DB,$B9,$FF,$BC,$9B
	.dc.b	$1F,$EE,$E6,$BB,$9F,$EB,$B3,$FE,$3F,$97,$D7,$62,$BF,$C0,$F1,$CB
	.dc.b	$87,$FA,$73,$AA,$FF,$04,$3F,$E6,$66,$F5,$71,$C7,$F9,$37,$D2,$5A
	.dc.b	$D6,$1F,$E4,$46,$BE,$9A,$F5,$BE,$36,$BE,$26,$D7,$F3,$17,$B7,$D7
	.dc.b	$3E,$B3,$B4,$61,$F1,$7D,$33,$C9,$A7,$B7,$DF,$FF,$07,$FD,$06,$35
	.dc.b	$FF,$11,$56,$DF,$DC,$E4,$8C,$67,$FB,$E8,$C6,$1E,$49,$CC,$FF,$9B
	.dc.b	$1D,$ED,$FD,$D5,$F9,$57,$FE,$6D,$72,$DE,$0F,$3F,$BD,$23,$C7,$C1
	.dc.b	$70,$FF,$35,$DF,$FA,$C6,$7C,$AD,$8F,$23,$5F,$15,$79,$4F,$B9,$83
	.dc.b	$EB,$7E,$45,$FC,$09,$CC,$62,$BF,$D7,$9E,$77,$F9,$3E,$BF,$CF,$CD
	.dc.b	$8F,$F6,$5F,$D8,$BE,$47,$F1,$10,$9D,$63,$F9,$67,$84,$7F,$26,$6B
	.dc.b	$FC,$E9,$AF,$F9,$E1,$AF,$3F,$3F,$F1,$9F,$5A,$F8,$6E,$F5,$DB,$2F
	.dc.b	$F4,$C4,$E5,$F3,$CE,$8E,$35,$63,$27,$F3,$0A,$E1,$7F,$CE,$1A,$F9
	.dc.b	$EB,$F2,$6F,$B5,$FC,$F7,$D6,$3E,$2F,$A4,$FF,$A8,$23,$17,$FE,$A5
	.dc.b	$FA,$B7,$2F,$C6,$6C,$7F,$AE,$79,$BF,$D4,$BE,$56,$6A,$4F,$F4,$4A
	.dc.b	$DE,$1F,$D3,$1A,$FF,$50,$6B,$FE,$B9,$FA,$37,$D6,$7E,$A1,$B5,$F9
	.dc.b	$C6,$CF,$E6,$DA,$FF,$B1,$1F,$F4,$64,$6F,$D4,$E4,$47,$C9,$F0,$DF
	.dc.b	$13,$03,$49,$1D,$FE,$4B,$C4,$47,$07,$F6,$0F,$15,$F1,$3F,$8A,$6C
	.dc.b	$7F,$B7,$7E,$59,$93,$E0,$B5,$72,$B9,$3F,$5A,$F8,$E9,$FD,$92,$30
	.dc.b	$7E,$5B,$E9,$BD,$B0,$E8,$FE,$57,$5C,$1F,$E5,$B3,$A7,$FD,$39,$C5
	.dc.b	$FE,$DA,$F4,$7F,$93,$7A,$6F,$90,$7D,$67,$F0,$DF,$86,$FF,$73,$EB
	.dc.b	$FE,$F5,$AD,$4E,$6F,$8D,$F0,$AF,$C1,$5F,$0D,$3F,$CB,$D3,$96,$FB
	.dc.b	$DC,$0E,$DC,$76,$FB,$57,$FA,$F5,$B3,$A6,$9F,$EF,$7F,$86,$FF,$79
	.dc.b	$B7,$0F,$DC,$C7,$EE,$E3,$04,$C3,$FC,$37,$1C,$1B,$3F,$A4,$8C,$47
	.dc.b	$F8,$89,$D6,$BF,$D5,$BE,$3D,$F1,$7D,$AF,$F8,$3C,$63,$0F,$82,$7F
	.dc.b	$49,$F8,$BF,$81,$7F,$42,$BF,$DE,$33,$FE,$D7,$9C,$BF,$EF,$EE,$0F
	.dc.b	$5B,$9D,$FE,$93,$CD,$F9,$FD,$92,$3F,$61,$F1,$FF,$8B,$7E,$4C,$67
	.dc.b	$5F,$A2,$FF,$47,$9D,$46,$24,$FE,$25,$FF,$29,$FD,$53,$EA,$FF,$91
	.dc.b	$B6,$37,$A7,$E3,$3B,$B3,$52,$7F,$C8,$35,$BB,$FF,$24,$B5,$FF,$26
	.dc.b	$B5,$EF,$3D,$BC,$6E,$51,$E3,$3C,$B3,$F6,$DF,$C5,$FC,$5F,$E4,$0F
	.dc.b	$CC,$FD,$4B,$E8,$71,$8F,$E2,$A6,$D7,$FA,$BE,$B5,$39,$3F,$69,$F3
	.dc.b	$2F,$8E,$FF,$04,$7B,$6F,$AA,$74,$7F,$82,$FC,$35,$EF,$0F,$86,$7F
	.dc.b	$88,$2B,$6F,$4C,$39,$C6,$32,$F6,$5F,$29,$CE,$1F,$D4,$3E,$9D,$E9
	.dc.b	$D3,$E8,$FC,$9D,$CB,$BE,$3F,$62,$B7,$1F,$E7,$06,$C7,$F9,$E1,$AF
	.dc.b	$F9,$F1,$AF,$FA,$41,$DF,$27,$F8,$C9,$FC,$D4,$7D,$10,$EA,$78,$76
	.dc.b	$7C,$3F,$9E,$FD,$2C,$7E,$2B,$D8,$76,$7E,$7B,$0F,$7D,$69,$4E,$AB
	.dc.b	$97,$6F,$E9,$DA,$7E,$94,$FB,$B8,$FF,$8C,$FE,$7D,$F1,$D8,$9F,$87
	.dc.b	$69,$CC,$7F,$17,$E5,$F4,$BA,$8C,$E3,$F8,$3C,$3A,$61,$FE,$3F,$F2
	.dc.b	$BF,$B3,$CF,$F5,$1F,$47,$FA,$81,$30,$FE,$A0,$F8,$F6,$E0,$D8,$F1
	.dc.b	$D7,$F4,$A6,$F8,$B9,$BF,$CE,$69,$FD,$0A,$3C,$07,$F5,$D2,$4F,$F5
	.dc.b	$D7,$FB,$BF,$D4,$BE,$82,$92,$7F,$D5,$F1,$96,$C7,$F8,$6A,$1F,$4F
	.dc.b	$63,$0F,$F8,$7B,$E9,$9E,$BF,$DE,$CB,$FD,$5F,$3F,$ED,$49,$CE,$FF
	.dc.b	$EA,$A8,$FF,$6B,$E0,$FA,$96,$59,$C3,$F8,$5E,$1C,$1E,$66,$31,$ED
	.dc.b	$E7,$F9,$23,$19,$71,$58,$F5,$C3,$18,$73,$BE,$14,$8C,$65,$E7,$C7
	.dc.b	$AB,$E6,$CB,$CB,$5E,$23,$84,$E9,$F0,$99,$AE,$6F,$F2,$4C,$7C,$4E
	.dc.b	$67,$83,$CB,$2F,$2C,$62,$77,$8F,$9A,$F9,$DE,$CE,$57,$EF,$AD,$E3
	.dc.b	$E9,$8E,$64,$9F,$44,$E9,$F3,$DF,$5C,$3C,$78,$76,$AD,$57,$7A,$D5
	.dc.b	$F3,$7E,$4C,$EA,$BC,$A4,$BF,$56,$3F,$22,$B8,$E1,$E8,$23,$CA,$35
	.dc.b	$8F,$A6,$72,$D9,$D6,$5E,$18,$C1,$27,$E7,$57,$27,$DB,$19,$7D,$4F
	.dc.b	$DB,$9E,$51,$A2,$1F,$88,$9E,$4F,$03,$38,$7C,$58,$CB,$EF,$CB,$9C
	.dc.b	$F1,$AB,$39,$8F,$A3,$5C,$31,$F9,$E7,$E9,$E3,$FB,$49,$F8,$A2,$9F
	.dc.b	$7C,$B3,$FA,$36,$EF,$FE,$8D,$76,$7E,$6D,$C3,$AF,$63,$2E,$5B,$3F
	.dc.b	$E8,$97,$F4,$B3,$CF,$3F,$C8,$37,$3F,$2F,$D4,$BE,$A7,$CF,$5A,$BE
	.dc.b	$24,$C7,$F4,$D5,$B7,$13,$B7,$F4,$D8,$75,$7F,$C1,$33,$B7,$35,$C3
	.dc.b	$0F,$E1,$55,$B9,$56,$EF,$A1,$2B,$77,$FC,$0E,$75,$18,$AE,$A4,$AF
	.dc.b	$E0,$73,$A8,$FE,$23,$7D,$F1,$F8,$D1,$BB,$F8,$78,$FC,$27,$A6,$1E
	.dc.b	$19,$E3,$8F,$86,$B5,$18,$39,$16,$7C,$31,$A8,$FA,$C6,$8B,$1A,$7E
	.dc.b	$21,$C6,$BC,$27,$91,$C9,$C1,$CB,$F3,$A7,$B5,$FB,$2B,$D1,$3B,$34
	.dc.b	$72,$8D,$57,$28,$FA,$35,$CB,$0F,$8C,$4C,$3E,$04,$F1,$8C,$3F,$84
	.dc.b	$9E,$13,$93,$71,$EF,$BE,$EF,$B4,$F1,$9F,$CC,$7E,$A5,$F1,$3F,$A8
	.dc.b	$EC,$FE,$6B,$29,$FF,$1D,$3F,$E1,$E3,$1E,$36,$F8,$3F,$5F,$3C,$A7
	.dc.b	$4D,$7F,$56,$D7,$F6,$6D,$78,$DF,$EB,$39,$5F,$50,$F8,$4C,$36,$BF
	.dc.b	$6B,$3A,$9C,$92,$3F,$C2,$33,$B7,$FB,$EB,$72,$FE,$A3,$5F,$E1,$5F
	.dc.b	$A8,$F5,$57,$F8,$8A,$9F,$15,$8A,$D4,$FF,$85,$EB,$15,$A8,$EF,$5E
	.dc.b	$62,$3E,$68,$D3,$FC,$4B,$D7,$7C,$6B,$CE,$ED,$E0,$7F,$16,$CF,$CA
	.dc.b	$DD,$FF,$D0,$77,$C2,$BC,$ED,$C3,$FC,$54,$72,$33,$87,$57,$FD,$BD
	.dc.b	$E1,$FF,$85,$DF,$F1,$AF,$CF,$FF,$9B,$5B,$E4,$FC,$2F,$29,$5B,$9F
	.dc.b	$F2,$3F,$81,$F0,$EF,$84,$7F,$8A,$1B,$AF,$F7,$CB,$E3,$5F,$32,$9F
	.dc.b	$50,$F7,$B0,$FF,$BD,$23,$F1,$8D,$7F,$CB,$CE,$DC,$C7,$F3,$07,$9B
	.dc.b	$63,$51,$FC,$E1,$AF,$F4,$07,$37,$F9,$C6,$ED,$FD,$19,$B3,$FE,$B0
	.dc.b	$FA,$77,$C8,$FC,$3F,$43,$FC,$3B,$87,$F8,$63,$D1,$77,$BE,$43,$F1
	.dc.b	$31,$87,$87,$83,$32,$FF,$42,$CE,$5B,$5C,$5F,$F4,$FC,$FD,$CF,$75
	.dc.b	$EF,$52,$4F,$FB,$FF,$97,$F9,$E6,$5F,$E2,$AE,$65,$F0,$AF,$F7,$BD
	.dc.b	$F0,$7F,$D7,$13,$F9,$0F,$0F,$EA,$7D,$3B,$B7,$EB,$70,$BE,$1E,$9A
	.dc.b	$89,$87,$F6,$0A,$E0,$D8,$FA,$91,$88,$FE,$D2,$DC,$FF,$85,$B7,$93
	.dc.b	$FB,$AB,$67,$FD,$BF,$F6,$6F,$53,$E1,$70,$C9,$FF,$56,$EB,$EE,$8F
	.dc.b	$DB,$FB,$EF,$75,$F3,$3B,$BF,$7F,$17,$D6,$78,$B5,$FF,$6D,$57,$19
	.dc.b	$E7,$4F,$81,$FC,$F7,$4C,$BF,$F7,$DC,$68,$C3,$FA,$18,$C3,$FC,$96
	.dc.b	$5F,$1B,$FA,$17,$F9,$17,$07,$F2,$C6,$F0,$7F,$85,$9D,$D7,$1A,$FC
	.dc.b	$F4,$62,$3F,$C2,$6D,$7F,$C3,$6D,$79,$47,$72,$3F,$CD,$FD,$8F,$F6
	.dc.b	$0B,$F0,$BD,$0F,$C7,$11,$FF,$6C,$E9,$F2,$FC,$9C,$47,$CA,$9B,$97
	.dc.b	$F9,$E2,$B5,$3C,$31,$7C,$7F,$E6,$7D,$59,$39,$8F,$9A,$61,$FA,$B1
	.dc.b	$87,$FA,$67,$2B,$D7,$FE,$11,$F4,$8F,$92,$A7,$E6,$22,$57,$E4,$DF
	.dc.b	$C8,$91,$C1,$D1,$DB,$FE,$FE,$F1,$FF,$AE,$E9,$FD,$0B,$E9,$3F,$21
	.dc.b	$3F,$D3,$1A,$FF,$92,$3C,$F4,$62,$3F,$C9,$B1,$8E,$AF,$E2,$9F,$49
	.dc.b	$FB,$19,$FF,$24,$B5,$FF,$28,$B5,$FF,$2A,$B5,$FF,$78,$D3,$8F,$FB
	.dc.b	$13,$63,$C9,$18,$33,$E6,$48,$F4,$35,$F0,$FC,$17,$90,$8F,$F9,$62
	.dc.b	$D1,$88,$65,$FE,$00,$9C,$C7,$F7,$F6,$BF,$A8,$FC,$5F,$E4,$F4,$7F
	.dc.b	$9B,$BE,$8B,$FD,$AF,$3F,$F0,$03,$E9,$3F,$B7,$4A,$1B,$27,$31,$9A
	.dc.b	$E2,$F8,$AB,$A7,$47,$F3,$F9,$CC,$FF,$9E,$9A,$E8,$FF,$39,$BC,$B2
	.dc.b	$FF,$0C,$B6,$B5,$3F,$EF,$EF,$8B,$7C,$4F,$F5,$5A,$FE,$09,$7F,$E0
	.dc.b	$49,$7C,$87,$F8,$05,$B7,$FE,$21,$AD,$E1,$F0,$7C,$2F,$31,$27,$C6
	.dc.b	$D7,$D8,$7B,$DF,$01,$FC,$6B,$E8,$BA,$C6,$B5,$7D,$17,$F1,$38,$F9
	.dc.b	$59,$FF,$BA,$05,$3E,$8F,$FC,$33,$18,$7F,$C3,$15,$C6,$BA,$35,$FC
	.dc.b	$CB,$5F,$AD,$7E,$0F,$0F,$FC,$65,$39,$9F,$B2,$F3,$BF,$F1,$4B,$72
	.dc.b	$E2,$D9,$DF,$D9,$F8,$BA,$FC,$7D,$F2,$7D,$77,$C2,$75,$1F,$E3,$CB
	.dc.b	$FE,$00,$D8,$F1,$91,$E2,$79,$1E,$2F,$F3,$CF,$87,$FE,$5A,$FE,$43
	.dc.b	$E7,$DF,$21,$FE,$3B,$FA,$4F,$E8,$F3,$CB,$64,$62,$3E,$95,$70,$74
	.dc.b	$D7,$F8,$23,$5F,$E0,$6D,$7E,$5B,$5F,$15,$7F,$92,$E3,$96,$BF,$E5
	.dc.b	$26,$E7,$C3,$AB,$5C,$5B,$F4,$66,$3E,$28,$CB,$FE,$62,$BF,$EB,$8F
	.dc.b	$8A,$3F,$CD,$51,$88,$FF,$71,$AF,$A2,$FD,$FC,$7F,$86,$7E,$07,$FD
	.dc.b	$CF,$F4,$6F,$A1,$FF,$74,$FB,$3F,$E7,$98,$C5,$7E,$EE,$3F,$CF,$98
	.dc.b	$73,$D7,$18,$76,$C7,$AF,$D1,$71,$CB,$FD,$E6,$DC,$7F,$82,$3C,$A7
	.dc.b	$75,$C6,$F2,$F9,$2F,$8B,$6F,$27,$F0,$DC,$3C,$4B,$0E,$97,$A8,$65
	.dc.b	$EA,$75,$7C,$DA,$73,$C7,$C6,$DA,$E5,$5C,$6F,$D5,$3C,$06,$1E,$28
	.dc.b	$F0,$C6,$67,$86,$1D,$E7,$26,$B2,$E6,$48,$F8,$55,$F5,$4F,$11,$CF
	.dc.b	$2F,$86,$FF,$2D,$BE,$AE,$AF,$AD,$77,$24,$65,$EA,$FF,$35,$3A,$94
	.dc.b	$F5,$E5,$F8,$67,$18,$FC,$E6,$F9,$EF,$A3,$E1,$1F,$4E,$30,$E2,$7F
	.dc.b	$CD,$73,$A8,$F8,$F3,$A7,$1F,$15,$FE,$7D,$4D,$23,$FE,$81,$7E,$E7
	.dc.b	$B9,$FE,$AA,$E2,$FB,$D3,$FD,$3E,$73,$ED,$A7,$33,$D0,$FB,$15,$C5
	.dc.b	$FB,$47,$13,$33,$C5,$C9,$C7,$4F,$F6,$E4,$F1,$7F,$DB,$57,$E0,$BE
	.dc.b	$ED,$AE,$2D,$9D,$C7,$69,$DB,$4D,$96,$36,$78,$2F,$C8,$E3,$4F,$B5
	.dc.b	$87,$8D,$FA,$78,$77,$6D,$F3,$6D,$73,$C3,$C2,$D9,$F0,$B4,$FF,$7E
	.dc.b	$73,$7F,$05,$EA,$FE,$A5,$5F,$D2,$38,$3F,$C8,$2F,$F9,$2B,$EC,$DF
	.dc.b	$61,$AC,$7F,$C0,$35,$A7,$FA,$F3,$79,$3F,$DE,$6D,$BF,$F0,$47,$A6
	.dc.b	$C3,$9C,$7F,$37,$9D,$5F,$4F,$94,$ED,$9F,$F1,$9E,$DB,$49,$B1,$2F
	.dc.b	$93,$A8,$C5,$F9,$22,$E5,$E4,$9B,$5E,$4F,$09,$F3,$DE,$0F,$A3,$7A
	.dc.b	$FE,$FD,$BF,$BA,$CE,$62,$5F,$F4,$FA,$DB,$FE,$70,$BE,$6E,$67,$79
	.dc.b	$FF,$80,$EB,$6F,$FA,$62,$B6,$E6,$7C,$58,$F8,$FD,$8B,$87,$77,$0D
	.dc.b	$3C,$51,$8A,$ED,$37,$3F,$D0,$6C,$CB,$E5,$61,$C6,$FF,$D5,$3F,$44
	.dc.b	$F8,$9F,$BD,$0F,$90,$79,$DF,$0E,$B5,$7F,$DE,$2B,$53,$C2,$B7,$39
	.dc.b	$6C,$F6,$BD,$54,$AD,$3D,$63,$13,$FE,$29,$FA,$2E,$39,$FE,$E6,$3C
	.dc.b	$9E,$17,$E1,$A7,$38,$7F,$21,$C7,$FD,$F6,$F8,$67,$C8,$6F,$1B,$19
	.dc.b	$25,$F7,$8C,$BF,$EC,$CF,$53,$FC,$1F,$3F,$EA,$C6,$C6,$7C,$EF,$C3
	.dc.b	$46,$32,$F3,$7D,$0F,$DC,$6F,$26,$BC,$8F,$9B,$39,$C7,$F9,$C3,$71
	.dc.b	$FF,$36,$FC,$BF,$F9,$72,$7E,$74,$E7,$89,$39,$AF,$37,$93,$FA,$88
	.dc.b	$FF,$2F,$DF,$9E,$31,$E8,$F8,$3D,$D9,$3F,$E2,$7A,$DD,$FF,$97,$06
	.dc.b	$2A,$FD,$9F,$CC,$EB,$5F,$C5,$DB,$E2,$F3,$97,$85,$FE,$0D,$F1,$BE
	.dc.b	$6F,$A1,$FC,$5F,$8C,$FA,$1C,$E0,$9C,$C4,$DB,$F1,$95,$B7,$FC,$8B
	.dc.b	$7C,$DC,$CE,$F6,$F4,$56,$DF,$F3,$AD,$6D,$CC,$FD,$6E,$6F,$AF,$E8
	.dc.b	$5C,$3E,$E3,$6B,$F0,$0D,$D7,$FD,$CD,$CA,$8F,$BF,$C1,$2F,$F0,$9C
	.dc.b	$F7,$0F,$C4,$4D,$C3,$8C,$DA,$FE,$ED,$F3,$9F,$AA,$BF,$A0,$FF,$B1
	.dc.b	$E3,$F5,$8E,$6B,$F4,$0E,$67,$7D,$7F,$B6,$8C,$4E,$A3,$19,$65,$B7
	.dc.b	$D1,$B7,$F9,$D8,$FC,$C9,$F9,$AE,$8A,$FE,$7C,$B1,$FA,$38,$B7,$FD
	.dc.b	$AF,$C2,$7C,$27,$DB,$BF,$F3,$DF,$C3,$BE,$1F,$F4,$91,$FD,$0A,$75
	.dc.b	$8F,$FA,$AA,$B5,$D9,$FC,$BE,$73,$8F,$F7,$79,$CD,$FF,$A1,$B1,$EF
	.dc.b	$CB,$FD,$4E,$73,$8F,$E6,$7E,$87,$F4,$98,$7F,$07,$F9,$F7,$D0,$7F
	.dc.b	$1C,$9C,$E5,$FE,$27,$9C,$BF,$C6,$DB,$7F,$A3,$6D,$FF,$11,$AE,$33
	.dc.b	$FE,$E3,$9C,$D7,$3B,$FE,$53,$18,$9F,$F7,$45,$70,$51,$9C,$1F,$3E
	.dc.b	$F9,$AF,$63,$71,$FF,$4E,$7D,$E3,$EB,$73,$52,$7E,$7D,$6F,$1F,$F6
	.dc.b	$85,$EE,$73,$5F,$B4,$F9,$57,$C4,$53,$FB,$7D,$3E,$B3,$17,$FB,$0A
	.dc.b	$DB,$6B,$73,$AC,$3E,$37,$C5,$FF,$48,$FA,$1F,$E9,$B1,$8B,$FE,$8D
	.dc.b	$18,$E1,$FE,$74,$D8,$FF,$2C,$36,$7E,$65,$73,$F3,$E5,$72,$7E,$6C
	.dc.b	$62,$7F,$7F,$F3,$5F,$A2,$A4,$3C,$24,$E6,$1E,$D7,$E1,$33,$FE,$2F
	.dc.b	$39,$AD,$F6,$E0,$FF,$10,$56,$F0,$FE,$27,$5C,$F8,$FE,$CE,$9F,$60
	.dc.b	$F9,$7E,$05,$AF,$AB,$17,$3F,$AB,$37,$6F,$E5,$7E,$17,$F3,$93,$F4
	.dc.b	$5F,$0F,$70,$FE,$6D,$16,$D5,$FB,$1E,$C5,$E6,$3F,$BF,$FC,$DB,$E4
	.dc.b	$3F,$A2,$65,$FE,$2C,$8C,$65,$FD,$06,$31,$3A,$8F,$F7,$54,$EB,$1E
	.dc.b	$F1,$F3,$3C,$EF,$EA,$37,$FE,$06,$FA,$27,$CE,$53,$1F,$0D,$6F,$FC
	.dc.b	$0D,$3A,$6C,$E9,$A7,$F8,$2F,$E8,$3F,$C1,$FC,$1F,$E8,$1F,$9C,$FF
	.dc.b	$4E,$46,$33,$FF,$45,$C7,$FA,$7F,$07,$D1,$7F,$9F,$7C,$E4,$F0,$F3
	.dc.b	$18,$F3,$C3,$F0,$B8,$F3,$BD,$C6,$2B,$6F,$63,$F1,$78,$F6,$CB,$94
	.dc.b	$2F,$96,$3D,$F1,$CC,$62,$78,$E5,$E5,$AD,$47,$53,$96,$3E,$0C,$BC
	.dc.b	$71,$8C,$3B,$E5,$E1,$8E,$18,$F5,$74,$62,$F9,$61,$F1,$71,$F0,$4F
	.dc.b	$B7,$0F,$2D,$6A,$30,$78,$F0,$ED,$8F,$8B,$1E,$C4,$AE,$14,$AD,$E3
	.dc.b	$EC,$BE,$15,$AB,$F3,$56,$AF,$BD,$6B,$0F,$19,$EE,$C7,$CF,$3E,$C9
	.dc.b	$D3,$ED,$9D,$65,$C2,$30,$49,$F9,$37,$DE,$39,$65,$E5,$23,$F5,$2B
	.dc.b	$95,$76,$24,$FC,$98,$CB,$C2,$79,$9C,$4F,$03,$9C,$7D,$F8,$FA,$9F
	.dc.b	$91,$8F,$2A,$E1,$87,$67,$E1,$DF,$5C,$7D,$17,$C2,$7A,$9C,$27,$B6
	.dc.b	$5F,$96,$8F,$F7,$F4,$FE,$95,$B1,$97,$F1,$E6,$B9,$3F,$CF,$8F,$FB
	.dc.b	$FA,$51,$88,$D7,$AD,$FE,$FD,$9C,$B6,$78,$E3,$F8,$56,$E1,$FB,$66
	.dc.b	$DF,$F3,$2F,$A2,$7C,$F7,$F3,$48,$FC,$CC,$E9,$FD,$53,$71,$EF,$1F
	.dc.b	$A5,$AE,$2F,$EB,$27,$44,$C7,$F6,$73,$A9,$EB,$1B,$23,$CE,$FC,$58
	.dc.b	$FA,$67,$2D,$9D,$18,$C7,$D9,$36,$3D,$95,$CA,$73,$1E,$02,$57,$EE
	.dc.b	$AF,$C8,$F5,$24,$F0,$75,$5C,$EB,$F8,$AE,$1E,$D7,$9D,$7C,$9B,$E6
	.dc.b	$F1,$24,$70,$C3,$E1,$DF,$38,$F9,$B1,$E9,$69,$EE,$3A,$9C,$4A,$2B
	.dc.b	$BB,$9C,$3D,$2D,$BF,$25,$F7,$C7,$EC,$D6,$AB,$BB,$EE,$7C,$51,$AA
	.dc.b	$F4,$C7,$7A,$E5,$8F,$C1,$BE,$58,$78,$AB,$9D,$F0,$C3,$94,$7D,$43
	.dc.b	$AD,$F1,$9F,$2C,$7F,$38,$FA,$27,$C3,$7E,$4F,$9F,$FD,$99,$18,$9F
	.dc.b	$F6,$E3,$5F,$F5,$22,$FB,$3E,$1C,$5F,$9B,$8D,$57,$2B,$CD,$FF,$84
	.dc.b	$A7,$35,$BB,$CF,$33,$9D,$F3,$D8,$FA,$27,$C1,$53,$73,$AC,$BF,$BA
	.dc.b	$D6,$5F,$F0,$FE,$9F,$EF,$EB,$E3,$1A,$33,$97,$F7,$98,$7D,$1F,$72
	.dc.b	$4F,$F9,$B7,$B9,$F2,$58,$8F,$F0,$08,$D1,$7C,$A3,$A9,$97,$0F,$0F
	.dc.b	$4B,$FD,$0B,$5F,$13,$C1,$7C,$63,$D7,$C3,$6F,$FD,$12,$D7,$C3,$8B
	.dc.b	$ED,$7F,$CB,$31,$AA,$E2,$47,$C7,$3B,$F3,$3E,$9F,$93,$FE,$42,$F0
	.dc.b	$AF,$8F,$FE,$18,$BF,$F1,$DC,$7D,$8F,$21,$F4,$7F,$E8,$89,$C9,$0F
	.dc.b	$F2,$7C,$EE,$35,$1F,$EE,$9F,$A4,$FF,$72,$3F,$10,$38,$75,$EF,$FE
	.dc.b	$19,$EF,$F2,$16,$75,$CE,$E7,$FE,$51,$18,$6D,$F1,$C1,$74,$E5,$13
	.dc.b	$0F,$14,$E9,$FF,$3E,$77,$7C,$4F,$47,$FC,$1F,$97,$79,$FF,$34,$46
	.dc.b	$31,$FF,$20,$71,$71,$9F,$F7,$6C,$E5,$DD,$F3,$3F,$3B,$3B,$BF,$F5
	.dc.b	$D7,$1F,$F3,$56,$F8,$9F,$80,$37,$F1,$97,$0F,$C0,$C5,$7F,$03,$69
	.dc.b	$F9,$1F,$43,$FA,$39,$FC,$A8,$71,$7E,$B8,$C7,$6F,$CF,$FA,$37,$A7
	.dc.b	$CA,$24,$FD,$7B,$E3,$B7,$F5,$7B,$DC,$FD,$B6,$BE,$09,$DB,$FE,$90
	.dc.b	$6A,$8D,$E1,$FE,$77,$FA,$3F,$D1,$7B,$FF,$DC,$2B,$FD,$CE,$D7,$F5
	.dc.b	$1E,$02,$7F,$85,$F7,$FF,$BB,$9E,$3E,$3C,$AF,$DC,$46,$A9,$F3,$74
	.dc.b	$93,$FB,$DF,$4E,$BD,$A2,$E5,$FE,$8B,$9F,$CA,$7A,$AF,$87,$FC,$B6
	.dc.b	$0F,$8E,$A6,$A2,$63,$F9,$F9,$D3,$67,$2F,$D6,$AE,$8E,$48,$66,$BD
	.dc.b	$5E,$4C,$8F,$E2,$8F,$B7,$E6,$F1,$1F,$1A,$2E,$5F,$EA,$39,$F8,$5E
	.dc.b	$4F,$C3,$F9,$49,$7F,$C7,$63,$2F,$F1,$F8,$CB,$F4,$23,$F1,$BA,$48
	.dc.b	$FC,$23,$FC,$2B,$E8,$BF,$19,$87,$E3,$79,$3F,$D5,$D5,$A2,$1F,$CA
	.dc.b	$3C,$55,$70,$E6,$FF,$6A,$C6,$1E,$D3,$F8,$BC,$3B,$47,$F4,$6A,$7B
	.dc.b	$1F,$D2,$3E,$8B,$F1,$73,$FE,$B3,$6D,$FF,$35,$AD,$BF,$C4,$7E,$89
	.dc.b	$F1,$5F,$C3,$DB,$19,$8E,$9C,$5F,$89,$F7,$EC,$FF,$2A,$C1,$39,$8E
	.dc.b	$9F,$1F,$E3,$F3,$24,$60,$87,$ED,$31,$4F,$84,$FC,$DF,$D1,$7C,$39
	.dc.b	$76,$23,$0D,$D7,$67,$0F,$8B,$FE,$51,$F4,$0F,$76,$93,$0F,$E8,$3F
	.dc.b	$24,$FA,$1C,$47,$F0,$6A,$E0,$FF,$35,$6B,$FD,$3A,$78,$3C,$A7,$8F
	.dc.b	$3F,$F0,$BA,$DB,$FC,$9F,$D3,$5F,$0B,$ED,$E1,$79,$88,$FF,$66,$D1
	.dc.b	$F2,$EF,$93,$CC,$F0,$3C,$95,$FC,$02,$75,$1C,$5A,$EC,$98,$FF,$39
	.dc.b	$BE,$B7,$FE,$0E,$FA,$1F,$E0,$75,$FD,$91,$AF,$F6,$C6,$D7,$AD,$77
	.dc.b	$F1,$F1,$25,$71,$7A,$44,$8C,$10,$E8,$F8,$7E,$66,$47,$85,$FE,$37
	.dc.b	$ED,$7F,$89,$23,$FC,$19,$69,$DC,$79,$5D,$FB,$F2,$39,$10,$FF,$07
	.dc.b	$57,$3D,$DF,$3C,$F9,$3F,$F1,$7B,$6B,$FC,$0A,$D7,$FC,$1A,$FE,$59
	.dc.b	$CC,$70,$23,$C3,$A3,$99,$87,$8C,$61,$EB,$C3,$E0,$23,$F8,$1E,$87
	.dc.b	$CA,$E2,$3C,$33,$78,$3E,$D5,$6A,$7F,$C1,$58,$BE,$7D,$49,$7C,$F7
	.dc.b	$F9,$5F,$E8,$3F,$B3,$46,$63,$1E,$9F,$FA,$C5,$FF,$2E,$61,$CE,$F5
	.dc.b	$C1,$FE,$B5,$8C,$5F,$48,$FF,$5C,$61,$D2,$30,$78,$5C,$5F,$4C,$7C
	.dc.b	$B3,$A3,$E9,$B8,$8F,$1C,$78,$0E,$58,$F9,$6F,$C5,$5B,$6E,$3B,$A6
	.dc.b	$1E,$09,$F2,$DF,$19,$F8,$98,$FA,$B1,$F2,$61,$C7,$0E,$15,$E0,$3A
	.dc.b	$56,$F0,$EF,$1E,$99,$DB,$E0,$C3,$B3,$EC,$BE,$38,$F9,$E3,$A1,$C6
	.dc.b	$77,$87,$58,$F2,$CE,$DE,$87,$96,$77,$19,$9E,$13,$A3,$A1,$E0,$A8
	.dc.b	$FF,$BB,$4E,$AD,$69,$63,$EE,$BB,$7F,$C9,$3F,$41,$B8,$FD,$4F,$B8
	.dc.b	$8C,$4E,$A7,$94,$E5,$E0,$3E,$26,$B4,$CB,$82,$3F,$85,$F8,$0F,$F4
	.dc.b	$1F,$83,$F7,$63,$F3,$93,$F7,$E7,$31,$F8,$B3,$27,$1F,$6B,$FC,$B1
	.dc.b	$BF,$FA,$DE,$32,$FF,$AE,$67,$8B,$98,$F2,$39,$9E,$3B,$FC,$7B,$F0
	.dc.b	$46,$2B,$E6,$46,$30,$F3,$47,$2A,$FD,$92,$C3,$BC,$73,$8F,$34,$F0
	.dc.b	$AE,$51,$EB,$39,$1F,$BB,$AE,$2F,$80,$E2,$6E,$38,$E1,$EB,$6D,$F7
	.dc.b	$8F,$65,$F7,$AE,$31,$B6,$CE,$CF,$4D,$F8,$39,$7F,$CD,$DD,$5F,$E8
	.dc.b	$3D,$BF,$D0,$9F,$07,$FE,$93,$F4,$B1,$9F,$F1,$1B,$DB,$FE,$C2,$9D
	.dc.b	$3F,$EB,$2E,$2F,$F2,$9B,$93,$FC,$98,$67,$0E,$79,$AB,$8B,$FE,$C0
	.dc.b	$FA,$A7,$95,$A9,$7C,$B7,$2A,$D4,$E6,$31,$A7,$F7,$5B,$E3,$8F,$96
	.dc.b	$72,$FA,$63,$37,$FE,$DE,$C3,$AE,$9E,$4D,$15,$AC,$7E,$1E,$3C,$B0
	.dc.b	$FF,$72,$4E,$71,$FE,$69,$8F,$27,$FD,$A7,$F3,$EF,$92,$F5,$4F,$FA
	.dc.b	$F6,$73,$18,$7C,$13,$B7,$E6,$4E,$EF,$FD,$C7,$39,$7D,$11,$9C,$7F
	.dc.b	$C2,$93,$97,$EC,$C6,$6B,$EC,$4E,$5F,$3C,$66,$BF,$DF,$DF,$7A,$F5
	.dc.b	$BE,$87,$E0,$1F,$E2,$FB,$FF,$76,$CE,$5F,$C0,$46,$6B,$FD,$EF,$39
	.dc.b	$7E,$64,$67,$1F,$F1,$44,$E5,$FC,$6C,$66,$BF,$19,$39,$7E,$5C,$66
	.dc.b	$BC,$B7,$E4,$F2,$FF,$13,$1F,$B1,$F1,$7F,$17,$F2,$4F,$C5,$7F,$CB
	.dc.b	$38,$FF,$BC,$F9,$BF,$C3,$AF,$F6,$9D,$3F,$DC,$15,$BF,$47,$9D,$6B
	.dc.b	$1F,$ED,$F7,$FD,$BE,$B4,$F3,$C7,$BE,$3D,$AB,$FD,$A7,$7C,$71,$EB
	.dc.b	$39,$7F,$DA,$91,$9B,$E9,$39,$7F,$67,$E3,$7A,$BF,$26,$FE,$FD,$1E
	.dc.b	$98,$C6,$1E,$66,$C6,$4F,$61,$E7,$9E,$33,$E4,$9D,$B6,$76,$B2,$57
	.dc.b	$0C,$7F,$DD,$5F,$26,$F1,$7E,$7B,$70,$E3,$CE,$FA,$2F,$AD,$92,$B4
	.dc.b	$FF,$BD,$9B,$87,$F9,$A7,$EB,$1F,$55,$D3,$2F,$D8,$46,$32,$EA,$DC
	.dc.b	$3E,$98,$CB,$C1,$F3,$3C,$30,$FD,$C6,$68,$E7,$F6,$5E,$8F,$6A,$E5
	.dc.b	$F0,$31,$F3,$46,$63,$E1,$FB,$CF,$A4,$FF,$79,$46,$F7,$FF,$55,$BA
	.dc.b	$8C,$DF,$93,$46,$AF,$94,$FE,$31,$F0,$EF,$6F,$DB,$18,$7F,$33,$8B
	.dc.b	$CB,$FE,$46,$FF,$BB,$FE,$2B,$31,$8F,$86,$46,$27,$EC,$9B,$3D,$87
	.dc.b	$33,$31,$C1,$CC,$F3,$9E,$59,$FC,$C9,$E4,$FC,$EF,$65,$F0,$EF,$B9
	.dc.b	$FE,$65,$F8,$1F,$E1,$EF,$FD,$39,$39,$7F,$A1,$46,$6F,$94,$E5,$FE
	.dc.b	$5D,$19,$E9,$FF,$4C,$4E,$5F,$E9,$F1,$9A,$FE,$57,$F6,$6F,$A1,$FE
	.dc.b	$84,$D8,$FF,$56,$7D,$4B,$A3,$35,$27,$FA,$A5,$6F,$2F,$65,$EE,$73
	.dc.b	$5F,$D4,$BC,$15,$BE,$DF,$FA,$AF,$4B,$F2,$7E,$5A,$B1,$FE,$F1,$7B
	.dc.b	$7F,$13,$3A,$73,$7C,$A3,$19,$FF,$81,$EB,$83,$F9,$CC,$3B,$65,$E3
	.dc.b	$9C,$E3,$DE,$95,$AC,$7F,$73,$97,$A3,$2E,$D8,$FA,$3B,$BF,$25,$5F
	.dc.b	$9C,$7E,$06,$1D,$DC,$B7,$5E,$1C,$DC,$5B,$5C,$5F,$C0,$CC,$8F,$C0
	.dc.b	$F2,$D8,$FB,$A5,$D3,$EE,$C5,$9F,$F1,$3F,$E2,$7F,$C8,$6B,$E5,$7E
	.dc.b	$2F,$F8,$19,$CF,$8D,$FA,$28,$C4,$FF,$A7,$EB,$85,$27,$32,$9C,$C6
	.dc.b	$1B,$CB,$FD,$19,$B5,$FE,$17,$FB,$DF,$D2,$5F,$F8,$05,$B9,$7E,$53
	.dc.b	$E8,$3F,$C2,$11,$8A,$FE,$FA,$DA,$FF,$0C,$37,$A3,$F4,$F5,$C3,$A7
	.dc.b	$F1,$3F,$9C,$FF,$13,$C3,$F8,$6B,$FD,$6F,$94,$E7,$1F,$C1,$39,$31
	.dc.b	$97,$E0,$67,$7E,$5F,$4F,$13,$AB,$FD,$8B,$DA,$FF,$70,$63,$FC,$F3
	.dc.b	$6F,$F1,$3F,$CC,$67,$E0,$FF,$C5,$F1,$89,$FE,$45,$5C,$34,$FE,$5B
	.dc.b	$39,$7F,$97,$9F,$E7,$DA,$E3,$EB,$30,$E9,$87,$78,$C3,$D2,$77,$3F
	.dc.b	$E3,$EE,$07,$77,$E4,$36,$FF,$26,$F6,$BE,$E3,$53,$9C,$7F,$80,$C6
	.dc.b	$3A,$BF,$80,$B6,$3F,$7A,$F5,$F0,$5E,$E7,$D8,$F4,$70,$DA,$FE,$2E
	.dc.b	$DD,$FF,$8A,$B7,$3F,$EA,$9E,$57,$E0,$3E,$3B,$FA,$9B,$6F,$FA,$9B
	.dc.b	$6F,$FA,$07,$93,$FD,$7E,$7F,$19,$5F,$E4,$9F,$B7,$7C,$57,$F2,$8F
	.dc.b	$93,$C3,$6B,$FB,$4C,$E9,$FE,$53,$3A,$69,$AF,$1B,$F2,$7F,$45,$D5
	.dc.b	$B1,$93,$F3,$91,$96,$9F,$D7,$FE,$95,$D7,$FD,$FF,$4F,$F5,$C7,$CE
	.dc.b	$FF,$A6,$A7,$3E,$B7,$FA,$CA,$3F,$D3,$BF,$4E,$FA,$8F,$E2,$5D,$E9
	.dc.b	$E1,$8F,$38,$C6,$21,$8F,$47,$B9,$97,$11,$C2,$18,$77,$CB,$C5,$1B
	.dc.b	$BE,$D8,$78,$32,$CC,$F6,$9C,$BE,$1B,$E4,$F8,$0D,$4F,$0C,$7C,$37
	.dc.b	$CE,$75,$5C,$31,$F2,$56,$A7,$C5,$7C,$70,$EF,$1E,$8C,$BE,$2E,$3C
	.dc.b	$F0,$E3,$5A,$C3,$95,$EF,$2E,$58,$C8,$F3,$4E,$B2,$F1,$63,$CC,$93
	.dc.b	$E2,$3C,$F9,$78,$1C,$65,$E0,$C3,$A3,$F3,$6F,$9C,$EA,$F9,$BF,$10
	.dc.b	$E7,$8F,$08,$CE,$3E,$3C,$7B,$56,$F0,$EC,$FA,$32,$F8,$78,$F6,$8D
	.dc.b	$61,$CA,$73,$7C,$B0,$F8,$64,$8C,$E3,$E1,$BF,$31,$0F,$90,$77,$CB
	.dc.b	$4F,$63,$F0,$17,$D6,$BE,$81,$27,$F0,$F8,$F9,$70,$F0,$47,$28,$C4
	.dc.b	$EE,$FF,$DF,$A7,$CF,$2D,$7E,$6D,$B1,$FA,$B7,$3E,$BF,$F9,$85,$E2
	.dc.b	$7E,$89,$C3,$AE,$3C,$B9,$6D,$7F,$9B,$DF,$CB,$4F,$3D,$3F,$26,$DD
	.dc.b	$3F,$5D,$6F,$A8,$FE,$68,$F0,$B8,$FE,$32,$B8,$3F,$B5,$9D,$B9,$9D
	.dc.b	$DF,$67,$F5,$B3,$A8,$C4,$EC,$8F,$EE,$63,$57,$FC,$2E,$FB,$D7,$B4
	.dc.b	$DC,$7A,$27,$71,$FC,$0E,$77,$87,$C1,$8D,$9E,$7E,$EF,$7E,$1E,$22
	.dc.b	$61,$DF,$1F,$3C,$EA,$31,$5C,$8F,$15,$79,$E7,$51,$E1,$AF,$83,$5C
	.dc.b	$2B,$53,$83,$47,$3C,$7D,$2F,$49,$E0,$47,$D0,$71,$34,$72,$9E,$AF
	.dc.b	$97,$0F,$81,$1E,$C8,$E2,$67,$1F,$8A,$7C,$CB,$F0,$D7,$0B,$F0,$57
	.dc.b	$63,$F0,$98,$FB,$9C,$4E,$DF,$14,$FD,$39,$D4,$E6,$77,$7D,$E7,$2F
	.dc.b	$8B,$1F,$66,$1F,$6B,$D5,$71,$BC,$0F,$CA,$F4,$7F,$96,$7D,$F7,$93
	.dc.b	$FE,$C0,$D3,$FC,$65,$E7,$EB,$97,$FA,$46,$5C,$B1,$C5,$7F,$AF,$F2
	.dc.b	$E8,$F0,$21,$FE,$B1,$72,$DF,$63,$8E,$4F,$51,$E8,$61,$B5,$FD,$CA
	.dc.b	$BB,$E3,$EA,$75,$52,$74,$7F,$78,$E5,$EB,$AF,$FA,$46,$3F,$C9,$9E
	.dc.b	$8F,$2C,$FF,$B7,$C6,$CC,$90,$F4,$57,$BB,$C7,$FF,$57,$F4,$F4,$F1
	.dc.b	$65,$7F,$AB,$A3,$5E,$27,$F0,$FE,$3F,$99,$9F,$FB,$3E,$BF,$CB,$97
	.dc.b	$AB,$E9,$E3,$3E,$BB,$E1,$FC,$43,$F0,$0A,$6A,$26,$1F,$DE,$72,$FB
	.dc.b	$15,$B8,$C7,$37,$F7,$DE,$54,$E6,$7F,$D5,$B7,$F6,$1F,$05,$7D,$AF
	.dc.b	$31,$D4,$FB,$1C,$36,$BF,$DE,$97,$FE,$9F,$7F,$CA,$2D,$CB,$FD,$45
	.dc.b	$7C,$AF,$89,$DE,$75,$87,$F2,$46,$DF,$FA,$17,$0E,$CF,$F9,$79,$B8
	.dc.b	$77,$9D,$F9,$FF,$EB,$FF,$89,$FC,$8B,$FE,$C1,$F1,$EF,$B2,$6C,$CC
	.dc.b	$7F,$D8,$31,$A3,$0F,$FA,$37,$DA,$FC,$39,$1F,$F7,$49,$4D,$1F,$E9
	.dc.b	$57,$EF,$BE,$DF,$AF,$F9,$9F,$45,$F7,$E3,$FC,$EA,$DC,$FD,$53,$B9
	.dc.b	$FA,$E7,$F8,$CE,$75,$AF,$F4,$86,$C7,$F8,$DF,$45,$71,$8E,$BC,$BE
	.dc.b	$03,$18,$7E,$43,$E4,$3F,$42,$49,$FF,$79,$FA,$F5,$FF,$56,$45,$9F
	.dc.b	$F6,$D7,$57,$FA,$0A,$BE,$6F,$C1,$7D,$2D,$3F,$C9,$1C,$28,$7D,$D6
	.dc.b	$8D,$F2,$4C,$FD,$B5,$B9,$D4,$62,$BA,$2F,$7D,$F0,$EF,$43,$ED,$CC
	.dc.b	$3F,$DD,$71,$96,$C6,$5F,$F7,$B7,$37,$F8,$13,$6F,$F7,$A5,$F4,$AF
	.dc.b	$E2,$ED,$AF,$2D,$6F,$4F,$6B,$63,$C9,$87,$75,$87,$EC,$3C,$BF,$DB
	.dc.b	$92,$3E,$DE,$FF,$EA,$AF,$B5,$C4,$70,$C7,$F6,$9D,$7F,$DF,$30,$E5
	.dc.b	$3C,$3B,$97,$FC,$13,$D7,$FE,$33,$F1,$52,$B5,$19,$7F,$DC,$D1,$97
	.dc.b	$B0,$CB,$FD,$77,$3F,$BF,$FC,$39,$CB,$FC,$07,$E4,$1F,$61,$4D,$44
	.dc.b	$C3,$F2,$A2,$76,$E1,$AF,$FB,$4A,$BA,$BC,$1A,$FF,$B5,$A3,$7A,$7F
	.dc.b	$09,$6E,$DF,$E2,$6A,$DC,$A3,$FA,$7F,$DC,$3C,$0F,$E8,$9F,$27,$31
	.dc.b	$FD,$4F,$9C,$AD,$47,$A3,$E7,$9F,$05,$FC,$1B,$E8,$BF,$9F,$CF,$C6
	.dc.b	$ED,$46,$34,$FE,$39,$7C,$2B,$AE,$0F,$88,$FE,$51,$39,$9F,$F7,$8B
	.dc.b	$6B,$D1,$0C,$BC,$95,$FD,$62,$DF,$4D,$FD,$6B,$EF,$3F,$C9,$F8,$70
	.dc.b	$B4,$63,$7F,$83,$5E,$0F,$43,$93,$32,$F7,$46,$5A,$43,$F0,$76,$FA
	.dc.b	$17,$D1,$7A,$BE,$63,$FB,$CF,$A2,$FC,$07,$F9,$9B,$73,$D6,$3F,$2E
	.dc.b	$B7,$5A,$3F,$9C,$C3,$81,$3C,$23,$F2,$77,$CD,$FE,$3E,$D7,$E4,$C7
	.dc.b	$CC,$F2,$BF,$AA,$7E,$18,$F8,$EA,$66,$BE,$E4,$E1,$FC,$AC,$ED,$C9
	.dc.b	$0F,$D1,$87,$4E,$1B,$C3,$F8,$09,$FF,$06,$FB,$3F,$ED,$F8,$FF,$12
	.dc.b	$7D,$3B,$E8,$FF,$C2,$ED,$C3,$F2,$35,$BD,$BF,$C4,$AD,$87,$AC,$AE
	.dc.b	$11,$FB,$68,$C5,$F8,$3E,$0E,$4F,$F4,$49,$CB,$6B,$8E,$3F,$87,$C3
	.dc.b	$B7,$1B,$E9,$1F,$1B,$FE,$2C,$EC,$FF,$09,$6B,$FE,$03,$6E,$5F,$95
	.dc.b	$AD,$BF,$AA,$6B,$E5,$8D,$E6,$78,$E1,$DF,$0F,$F1,$9D,$F0,$CB,$FC
	.dc.b	$95,$5C,$0D,$C6,$71,$FF,$24,$FD,$DB,$E7,$FF,$C8,$3F,$11,$FD,$DF
	.dc.b	$4F,$80,$D7,$F3,$33,$B7,$F7,$ED,$7E,$54,$6F,$1F,$F2,$D3,$6F,$BD
	.dc.b	$FF,$92,$2F,$85,$6B,$DA,$F0,$99,$9F,$F2,$6F,$D0,$3E,$5F,$FC,$AB
	.dc.b	$F7,$4F,$9E,$A6,$A2,$69,$F5,$E7,$4D,$9C,$BF,$92,$A7,$B5,$FD,$2A
	.dc.b	$3F,$C6,$BF,$27,$FE,$C6,$FA,$3F,$E5,$D0,$32,$57,$2E,$05,$74,$D3
	.dc.b	$87,$A8,$BF,$F6,$8C,$E6,$B8,$5F,$2A,$D4,$75,$F3,$7F,$DA,$F1,$88
	.dc.b	$EB,$8F,$9E,$31,$7C,$DD,$5F,$3C,$3D,$58,$72,$7C,$37,$C5,$B8,$F9
	.dc.b	$9D,$E3,$4F,$11,$4F,$0C,$5C,$7C,$13,$C2,$7D,$77,$DD,$F9,$13,$A6
	.dc.b	$DF,$85,$EA,$70,$3A,$39,$7B,$99,$26,$3F,$02,$35,$3D,$65,$8F,$C1
	.dc.b	$7A,$D6,$F1,$EA,$7A,$E7,$53,$C3,$0F,$43,$E0,$AE,$0F,$E0,$4F,$8E
	.dc.b	$53,$D7,$55,$A5,$B7,$C0,$3E,$8F,$53,$F8,$68,$C3,$B6,$BD,$67,$8C
	.dc.b	$F0,$8F,$C9,$71,$D4,$D2,$3B,$7E,$C7,$B4,$F3,$BF,$B1,$E5,$F8,$BA
	.dc.b	$E6,$7E,$66,$75,$CB,$F7,$E3,$33,$F9,$19,$CC,$7E,$7C,$CE,$FF,$16
	.dc.b	$33,$3D,$67,$CA,$7F,$2F,$9D,$5F,$F7,$4E,$CD,$63,$FA,$D7,$B4,$66
	.dc.b	$7E,$64,$FA,$A3,$D3,$18,$AF,$44,$61,$FA,$93,$BC,$7D,$51,$E2,$C3
	.dc.b	$AC,$62,$75,$3E,$B3,$57,$E6,$9D,$47,$66,$D7,$28,$E1,$5C,$A3,$D7
	.dc.b	$1B,$32,$7C,$13,$F8,$56,$3F,$2B,$1D,$9C,$CF,$E0,$3D,$5F,$E3,$DF
	.dc.b	$3F,$FD,$79,$5F,$EF,$EE,$0F,$F1,$86,$5F,$8F,$AD,$36,$33,$3F,$95
	.dc.b	$9C,$BF,$D7,$67,$FC,$06,$7F,$8D,$A7,$7C,$1F,$E2,$6E,$CF,$F1,$F7
	.dc.b	$87,$FE,$28,$8F,$F1,$17,$D1,$FF,$45,$BF,$F1,$EB,$70,$FF,$2E,$FA
	.dc.b	$AF,$A0,$FC,$2D,$F0,$EB,$E5,$87,$F2,$9D,$18,$F1,$C7,$F5,$91,$8A
	.dc.b	$F7,$C7,$F8,$DE,$74,$F1,$6B,$98,$EA,$FF,$7C,$F9,$9F,$6B,$FE,$2E
	.dc.b	$9F,$36,$1E,$D9,$CB,$FE,$3D,$F6,$1F,$47,$CB,$14,$63,$2F,$F4,$4F
	.dc.b	$D3,$FF,$96,$1F,$F2,$EC,$67,$2F,$F1,$DB,$5F,$5B,$5F,$80,$D7,$FC
	.dc.b	$A4,$DA,$EF,$7E,$AD,$FF,$C2,$B7,$E3,$AE,$D8,$FD,$59,$FF,$39,$4F
	.dc.b	$F9,$BB,$E9,$3F,$B1,$37,$BB,$E1,$F0,$34,$7D,$8F,$B6,$F7,$AF,$F9
	.dc.b	$49,$AF,$D2,$6B,$F5,$DA,$FF,$9A,$5B,$A9,$E7,$9F,$E2,$7E,$01,$E9
	.dc.b	$F7,$CF,$D9,$18,$D3,$FD,$33,$D1,$FE,$79,$6F,$AF,$F6,$BD,$07,$CC
	.dc.b	$7F,$A8,$A7,$24,$D7,$F9,$75,$F1,$6D,$FE,$21,$FF,$6C,$36,$35,$9F
	.dc.b	$FA,$0B,$07,$D2,$BD,$FF,$C4,$E1,$C3,$1F,$F3,$1C,$E5,$F3,$F4,$EB
	.dc.b	$1F,$CF,$47,$FB,$AB,$F3,$77,$CE,$E6,$A4,$FF,$3E,$AD,$E1,$E8,$9F
	.dc.b	$21,$27,$FC,$E3,$F0,$FF,$E0,$D7,$B9,$89,$F2,$FC,$6B,$E6,$69,$F9
	.dc.b	$7F,$A6,$F9,$93,$FE,$6C,$9D,$57,$FB,$19,$B1,$FD,$55,$AF,$F5,$96
	.dc.b	$BF,$D7,$5A,$FF,$A9,$FE,$7D,$F3,$5F,$B4,$E8,$46,$2F,$F8,$93,$6B
	.dc.b	$F8,$83,$5F,$DE,$AF,$97,$E4,$D9,$E3,$18,$3E,$D6,$0F,$98,$A4,$9F
	.dc.b	$ED,$31,$97,$C3,$F2,$3D,$08,$E0,$FE,$DE,$7F,$72,$3F,$DD,$25,$C3
	.dc.b	$FD,$EA,$BE,$8D,$F4,$7F,$84,$8D,$73,$FF,$B2,$AB,$FD,$C5,$E8,$FE
	.dc.b	$3E,$79,$47,$F8,$0D,$AF,$F8,$1D,$AF,$DD,$8F,$E2,$F0,$F6,$FF,$E0
	.dc.b	$0B,$E4,$8C,$C6,$27,$F9,$6F,$CC,$FF,$AC,$6B,$FD,$25,$F3,$6E,$BF
	.dc.b	$DF,$CE,$6B,$FD,$35,$39,$AF,$F2,$C6,$3F,$EE,$2F,$11,$7B,$8F,$E8
	.dc.b	$15,$BB,$EF,$F0,$5F,$C4,$E3,$13,$FD,$17,$E8,$9F,$49,$FC,$5D,$B9
	.dc.b	$76,$3F,$CC,$D9,$3E,$4B,$EF,$65,$FE,$AD,$AD,$B6,$F9,$1F,$C6,$27
	.dc.b	$78,$77,$8F,$E7,$BD,$8B,$7D,$16,$1B,$97,$87,$1F,$F5,$27,$4F,$D8
	.dc.b	$8F,$ED,$F3,$A9,$62,$9C,$DF,$F5,$0C,$3B,$36,$F8,$5F,$F8,$75,$AF
	.dc.b	$0A,$F0,$7C,$0B,$E5,$7D,$B8,$F5,$A6,$3D,$2F,$75,$A7,$FD,$19,$7F
	.dc.b	$AE,$E6,$FE,$B6,$FE,$C3,$A6,$E8,$9B,$18,$6D,$78,$A7,$4D,$9F,$B5
	.dc.b	$8C,$8F,$8D,$BF,$E0,$B9,$31,$3F,$88,$2E,$9F,$87,$E3,$FC,$B5,$FE
	.dc.b	$42,$7F,$93,$FC,$DF,$F2,$C8,$FF,$32,$36,$3E,$67,$47,$07,$FD,$FF
	.dc.b	$19,$4F,$F3,$97,$FC,$D4,$D7,$9F,$8B,$FC,$EE,$31,$1F,$EB,$17,$B5
	.dc.b	$7E,$DF,$9B,$C7,$5F,$AC,$BF,$CE,$4F,$FA,$DF,$CD,$7C,$EF,$EA,$F3
	.dc.b	$FF,$0F,$36,$7F,$2F,$9F,$F0,$07,$FD,$FD,$19,$CD,$7C,$BE,$53,$FB
	.dc.b	$B4,$E7,$6F,$D5,$F6,$62,$7E,$CF,$D2,$7F,$81,$F0,$FC,$7E,$3F,$7A
	.dc.b	$3F,$9B,$E7,$F8,$5E,$44,$6E,$30,$EA,$78,$7C,$A7,$4F,$3B,$AB,$63
	.dc.b	$F4,$18,$7F,$A0,$BE,$83,$FA,$DF,$BA,$7F,$C4,$BC,$8F,$59,$F3,$BA
	.dc.b	$99,$87,$FA,$52,$F8,$36,$3F,$B6,$46,$30,$FF,$39,$C7,$4B,$7C,$D7
	.dc.b	$29,$F7,$9F,$7A,$33,$7F,$90,$F0,$7F,$2B,$3F,$C0,$FE,$2F,$F8,$07
	.dc.b	$6F,$F0,$B6,$D7,$EE,$F9,$BF,$80,$46,$2B,$84,$FE,$6B,$9F,$FD,$D8
	.dc.b	$FD,$8F,$47,$F9,$1F,$C0,$FD,$66,$C6,$4E,$11,$97,$F8,$47,$97,$FD
	.dc.b	$73,$C9,$FE,$AA,$48,$FE,$BA,$DC,$3D,$B9,$7E,$33,$D0,$FF,$24,$79
	.dc.b	$8A,$FC,$2C,$7F,$92,$FE,$81,$F4,$94,$CE,$3E,$08,$CB,$63,$FB,$74
	.dc.b	$63,$1F,$F3,$D3,$5F,$F3,$3B,$5F,$F3,$63,$FC,$51,$64,$8C,$37,$C9
	.dc.b	$EA,$4B,$FC,$FF,$CE,$F9,$E3,$F3,$AF,$F8,$93,$E8,$FF,$C7,$7E,$77
	.dc.b	$FB,$43,$E4,$3F,$DC,$51,$8C,$BF,$DA,$31,$FE,$E5,$FA,$4F,$F1,$CE
	.dc.b	$3B,$90,$AE,$1D,$4B,$E7,$DD,$FD,$6E,$B9,$9C,$2B,$73,$D7,$D9,$FE
	.dc.b	$59,$7C,$6B,$57,$C6,$31,$1D,$E3,$15,$DB,$2F,$24,$B2,$F3,$BE,$03
	.dc.b	$84,$6B,$1F,$0C,$76,$8E,$33,$C3,$0E,$F5,$AA,$E2,$73,$71,$3B,$BF
	.dc.b	$14,$E7,$0F,$2B,$D4,$F1,$4E,$DC,$5F,$69,$DC,$76,$8F,$7C,$EE,$31
	.dc.b	$5D,$A7,$86,$1E,$8C,$7E,$4C,$E4,$95,$D9,$F9,$D9,$7C,$1C,$7B,$E5
	.dc.b	$ED,$3D,$77,$CA,$F8,$10,$F0,$9F,$80,$CB,$C9,$3C,$6B,$99,$EA,$8C
	.dc.b	$B8,$C7,$C0,$79,$F0,$B7,$F1,$E2,$E5,$F0,$31,$F1,$4F,$80,$98,$F9
	.dc.b	$AF,$79,$7B,$A7,$59,$78,$31,$CC,$7A,$63,$0F,$72,$63,$DE,$F7,$1F
	.dc.b	$78,$FC,$E9,$F6,$4A,$64,$FB,$E6,$30,$FF,$3D,$70,$EA,$3F,$3D,$EA
	.dc.b	$FF,$09,$71,$3A,$6C,$65,$E4,$EE,$FF,$22,$DC,$BF,$6A,$D9,$FD,$8F
	.dc.b	$D2,$7F,$36,$7F,$39,$19,$7F,$68,$DB,$F0,$5F,$EA,$AB,$85,$7E,$6E
	.dc.b	$B4,$4C,$3F,$57,$3A,$AF,$1C,$76,$25,$7E,$B6,$75,$87,$9A,$73,$5C
	.dc.b	$2B,$5D,$BF,$32,$30,$4A,$E1,$3A,$38,$B8,$8D,$47,$C0,$9E,$53,$92
	.dc.b	$47,$D0,$AF,$4C,$EA,$30,$7B,$E3,$AB,$CC,$A7,$C8,$6D,$FB,$23,$8D
	.dc.b	$F8,$9D,$1E,$73,$A1,$1D,$47,$18,$DC,$71,$C7,$E9,$E1,$F8,$EA,$D4
	.dc.b	$62,$BB,$CF,$B1,$4E,$B1,$FA,$D1,$DE,$79,$63,$EF,$8E,$F1,$C6,$FA
	.dc.b	$47,$29,$D4,$60,$95,$F8,$9B,$ED,$8F,$C4,$9E,$31,$F0,$F8,$7F,$A9
	.dc.b	$77,$7E,$9B,$C3,$FF,$18,$F9,$D7,$6F,$F1,$B7,$93,$FE,$CE,$D7,$FB
	.dc.b	$25,$E5,$FE,$B7,$CD,$FE,$1B,$7F,$51,$8F,$46,$C6,$5F,$EC,$F3,$FA
	.dc.b	$D7,$8F,$AB,$6B,$FB,$23,$6B,$FD,$9D,$3C,$A3,$07,$86,$77,$1F,$DD
	.dc.b	$E3,$C9,$D7,$FE,$4C,$F3,$F5,$6E,$7C,$E5,$FE,$DF,$C3,$FC,$97,$5A
	.dc.b	$99,$1E,$87,$C4,$47,$F8,$27,$97,$FE,$B6,$E7,$E7,$E1,$49,$FF,$10
	.dc.b	$47,$C5,$9D,$47,$F9,$62,$74,$FF,$B2,$27,$5E,$1F,$F9,$02,$3F,$C5
	.dc.b	$9F,$10,$FA,$4F,$5B,$63,$FC,$6D,$E3,$F0,$33,$32,$32,$7F,$94,$23
	.dc.b	$FD,$C9,$C9,$FE,$9F,$F2,$3F,$C4,$5F,$47,$F6,$5B,$5B,$E5,$FB,$8F
	.dc.b	$F9,$49,$FA,$CD,$3E,$C5,$FF,$BB,$DF,$E5,$6D,$7F,$97,$BE,$08,$A7
	.dc.b	$47,$11,$FE,$8E,$E9,$7C,$D7,$E2,$FC,$DF,$F5,$B7,$C5,$5F,$55,$F2
	.dc.b	$74,$C4,$4C,$3F,$D6,$F1,$97,$FD,$73,$1F,$7B,$DB,$91,$FE,$88,$7E
	.dc.b	$A7,$9B,$FD,$0B,$7F,$F4,$7C,$6A,$7F,$C9,$3D,$08,$DC,$7A,$7E,$5F
	.dc.b	$E2,$FD,$27,$FA,$FE,$3F,$CF,$AD,$7F,$CD,$A7,$F9,$EA,$38,$3F,$21
	.dc.b	$F9,$7B,$9A,$48,$FB,$0F,$07,$9C,$FF,$4D,$C3,$FD,$4B,$F2,$7F,$9A
	.dc.b	$24,$FF,$9E,$7D,$5A,$FA,$A2,$CF,$FB,$7F,$AB,$FE,$00,$57,$CF,$F8
	.dc.b	$07,$35,$FA,$49,$8F,$FB,$A2,$72,$D9,$CB,$FE,$8A,$E3,$4A,$BA,$57
	.dc.b	$F0,$5F,$75,$F0,$AF,$3F,$EE,$63,$0F,$F2,$8E,$9F,$EF,$3A,$DC,$F1
	.dc.b	$AE,$15,$FD,$CE,$72,$D8,$CB,$FD,$F2,$32,$D8,$CD,$75,$6F,$4F,$F9
	.dc.b	$82,$B7,$3F,$EB,$FC,$9E,$87,$EC,$3C,$C9,$8F,$ED,$1F,$DB,$FD,$26
	.dc.b	$23,$FD,$1F,$DF,$FD,$CB,$06,$5F,$ED,$D8,$C3,$FE,$6E,$E1,$78,$0C
	.dc.b	$3F,$80,$FA,$DF,$C5,$3E,$2E,$65,$FE,$B3,$8C,$BF,$EE,$08,$CB,$FE
	.dc.b	$BA,$BE,$73,$FB,$8F,$A3,$72,$FE,$EB,$E4,$5F,$49,$4D,$44,$C3,$F9
	.dc.b	$95,$70,$6C,$7F,$00,$8C,$47,$F8,$8D,$BB,$7B,$A3,$FD,$13,$18,$D7
	.dc.b	$FC,$5E,$D8,$FF,$46,$E0,$F9,$9F,$E7,$1F,$46,$F0,$7F,$A0,$36,$BF
	.dc.b	$A4,$B5,$FF,$1B,$F9,$E8,$C4,$7F,$8E,$BE,$7D,$F0,$1F,$C5,$9B,$1F
	.dc.b	$D3,$5A,$FF,$8E,$DA,$FF,$91,$2B,$A7,$C2,$78,$89,$1E,$27,$CB,$F4
	.dc.b	$52,$3C,$AD,$7A,$79,$FE,$12,$3E,$16,$BF,$C1,$2F,$FA,$93,$FA,$0D
	.dc.b	$BF,$CA,$B1,$88,$FF,$7A,$E0,$FA,$2F,$E4,$B1,$B3,$FD,$6F,$2F,$97
	.dc.b	$FE,$E5,$E8,$3E,$8B,$F9,$6B,$71,$FE,$77,$5B,$BE,$41,$C3,$F1,$6F
	.dc.b	$E5,$11,$F9,$4F,$22,$3E,$3F,$11,$F9,$A7,$9A,$3D,$DE,$47,$F4,$AF
	.dc.b	$A3,$7C,$5D,$3C,$71,$27,$EB,$3F,$64,$8E,$0F,$3E,$3F,$C9,$DB,$1D
	.dc.b	$7E,$2B,$FC,$05,$F1,$9F,$EC,$CE,$AF,$CA,$7B,$4F,$9B,$7D,$1F,$F8
	.dc.b	$11,$B8,$FC,$BE,$46,$1C,$78,$7F,$DC,$71,$8F,$83,$FF,$0A,$4F,$FB
	.dc.b	$62,$BF,$10,$D7,$6D,$7F,$DF,$AF,$36,$C7,$98,$FF,$02,$7C,$CF,$F8
	.dc.b	$72,$3E,$FF,$D1,$BE,$37,$FC,$33,$C9,$FE,$EF,$96,$DE,$C8,$C3,$FB
	.dc.b	$1C,$97,$C6,$BD,$8D,$7F,$52,$D7,$F4,$0D,$7D,$ED,$51,$D2,$FF,$C5
	.dc.b	$D3,$CE,$BC,$0D,$CF,$F8,$E7,$CA,$BE,$4F,$FC,$69,$F4,$6F,$9E,$FF
	.dc.b	$19,$FC,$5F,$F7,$AD,$3F,$35,$C6,$8C,$47,$E8,$2B,$83,$F5,$5A,$FF
	.dc.b	$02,$6B,$FB,$96,$BF,$61,$AF,$A9,$E3,$9F,$C7,$6E,$5B,$C3,$FC,$6D
	.dc.b	$F7,$BF,$DE,$6B,$E6,$7D,$17,$F1,$7C,$3C,$FF,$48,$F9,$DF,$F2,$6B
	.dc.b	$63,$F1,$53,$B7,$F7,$2D,$7F,$87,$35,$FC,$6B,$5F,$9E,$D4,$BD,$BF
	.dc.b	$C0,$C7,$E6,$FE,$57,$FD,$B7,$F4,$8F,$33,$FD,$BB,$D4,$9C,$C6,$1F
	.dc.b	$F7,$0E,$1C,$F3,$61,$CC,$B1,$95,$EA,$F3,$77,$8F,$77,$8C,$65,$46
	.dc.b	$5E,$37,$E0,$CB,$D3,$5B,$BE,$CF,$1B,$E0,$DC,$7F,$DE,$EE,$EB,$8D
	.dc.b	$F3,$C7,$C2,$FB,$B1,$A7,$9A,$F8,$C6,$E3,$C6,$7B,$AB,$9B,$B3,$89
	.dc.b	$EA,$9E,$2E,$EB,$38,$FA,$2F,$C1,$3D,$0E,$06,$AF,$AB,$F6,$9F,$B3
	.dc.b	$97,$B6,$3E,$8B,$5F,$A4,$6E,$F3,$5E,$9C,$74,$47,$04,$9F,$51,$E3
	.dc.b	$CB,$E5,$C6,$63,$A1,$F8,$03,$E2,$14,$C9,$71,$FC,$13,$7E,$01,$F4
	.dc.b	$9F,$EB,$79,$FF,$32,$36,$33,$3C,$CF,$CC,$B8,$7F,$D7,$DC,$7F,$E8
	.dc.b	$A7,$F8,$01,$FE,$8C,$70,$47,$E6,$7B,$6F,$A4,$FF,$07,$67,$F8,$F9
	.dc.b	$FC,$CD,$7D,$8A,$D7,$27,$DA,$9C,$9C,$AF,$8E,$7E,$5B,$E1,$18,$9F
	.dc.b	$85,$18,$AE,$51,$E6,$C3,$F3,$D8,$76,$8F,$C0,$1D,$A3,$37,$CE,$3E
	.dc.b	$75,$F3,$76,$73,$C3,$E9,$5F,$3B,$F6,$4E,$4F,$F2,$A5,$F3,$7B,$57
	.dc.b	$08,$C4,$6C,$57,$39,$F7,$47,$C1,$C3,$D7,$87,$8F,$0E,$ED,$AD,$C7
	.dc.b	$AA,$77,$5C,$4F,$4D,$6D,$FE,$8B,$C1,$FE,$39,$9F,$ED,$4D,$8F,$F2
	.dc.b	$A7,$1F,$FB,$1F,$7F,$EA,$D3,$96,$FA,$1F,$E4,$96,$DF,$F8,$77,$D6
	.dc.b	$5F,$18,$FE,$E7,$7C,$5F,$EE,$BF,$49,$FD,$35,$FE,$F4,$DC,$7F,$CC
	.dc.b	$9D,$7F,$5F,$1F,$EC,$D3,$96,$F4,$78,$6F,$9B,$6D,$93,$0E,$55,$A3
	.dc.b	$BD,$6F,$B9,$8F,$59,$FE,$C7,$C3,$98,$C6,$3F,$DB,$3D,$87,$D2,$71
	.dc.b	$6D,$FC,$1F,$0B,$FC,$B6,$FF,$85,$63,$35,$FE,$44,$6E,$BF,$E0,$16
	.dc.b	$CF,$FA,$36,$B8,$D6,$EF,$B1,$27,$B7,$A7,$67,$FC,$9F,$F4,$7F,$D7
	.dc.b	$67,$FC,$48,$D8,$FF,$47,$5B,$E9,$3E,$04,$E7,$2F,$F0,$93,$6F,$FC
	.dc.b	$52,$DB,$F9,$77,$C6,$7A,$F7,$F2,$23,$F0,$70,$B5,$FE,$EA,$E8,$7A
	.dc.b	$BF,$56,$3F,$3E,$FD,$7F,$45,$F3,$3F,$E9,$EC,$3F,$CB,$B7,$C1,$B5
	.dc.b	$C6,$31,$7F,$E2,$C9,$FE,$C7,$7C,$BD,$BD,$5C,$9E,$27,$F9,$C6,$74
	.dc.b	$E2,$BC,$DC,$BD,$4F,$C5,$79,$2F,$A3,$7C,$E6,$6F,$CB,$56,$FF,$DD
	.dc.b	$55,$A9,$C9,$23,$CC,$30,$FF,$77,$BC,$78,$7F,$D6,$51,$89,$FF,$2E
	.dc.b	$7D,$1B,$D9,$F9,$1F,$20,$F9,$BF,$CC,$4E,$6B,$FD,$07,$39,$6C,$F0
	.dc.b	$37,$56,$3F,$CD,$1F,$3C,$F9,$AF,$D1,$36,$38,$9A,$8C,$FC,$BC,$8F
	.dc.b	$DF,$7C,$8F,$F9,$D1,$E7,$1E,$6F,$99,$91,$83,$CA,$F6,$F9,$89,$1F
	.dc.b	$E9,$1F,$A8,$7D,$86,$38,$7E,$39,$FE,$F2,$F3,$3F,$11,$47,$F7,$CB
	.dc.b	$E8,$DE,$E6,$A6,$4F,$FA,$98,$E0,$7F,$AC,$4E,$11,$AA,$EF,$1A,$6C
	.dc.b	$F6,$AE,$B3,$CE,$3F,$87,$79,$0D,$7F,$D4,$B3,$FE,$64,$FA,$5F,$BC
	.dc.b	$FC,$6F,$80,$FE,$29,$EC,$27,$39,$7F,$C0,$1B,$E2,$FE,$1A,$75,$CF
	.dc.b	$DA,$73,$D1,$FE,$A4,$BE,$2F,$FA,$9B,$E9,$7F,$CF,$19,$7F,$9D,$1B
	.dc.b	$8F,$E4,$33,$7C,$73,$93,$C7,$5A,$C3,$97,$B7,$83,$DB,$5B,$C7,$A4
	.dc.b	$7F,$2D,$F0,$96,$F9,$0C,$36,$BF,$D9,$9D,$1F,$2F,$14,$7F,$96,$2F
	.dc.b	$8D,$F0,$24,$70,$E2,$BA,$FE,$A2,$B7,$3F,$EE,$96,$C7,$EB,$3C,$1F
	.dc.b	$CD,$57,$0C,$3B,$CA,$7F,$3D,$D1,$FB,$CC,$BD,$93,$C5,$46,$76,$D3
	.dc.b	$72,$F5,$B6,$BE,$B7,$07,$F9,$CA,$7F,$01,$1F,$5F,$E0,$3E,$FC,$E6
	.dc.b	$36,$60,$D4,$F0,$3F,$DF,$7F,$39,$FC,$73,$3F,$4C,$A3,$13,$DF,$9F
	.dc.b	$FD,$BE,$F0,$F1,$3F,$96,$4E,$AB,$FD,$0A,$D9,$FD,$36,$3A,$A9,$1C
	.dc.b	$4F,$D2,$5F,$4B,$FD,$8D,$F3,$E7,$7C,$CB,$E3,$FF,$36,$DC,$3F,$81
	.dc.b	$F5,$E2,$64,$7F,$B1,$A7,$58,$7E,$7E,$B2,$FD,$9F,$A5,$FE,$BF,$AF
	.dc.b	$D8,$E2,$47,$F1,$D9,$D7,$9D,$FC,$D2,$31,$5B,$78,$E3,$FE,$A4,$BE
	.dc.b	$3E,$16,$EF,$83,$FD,$47,$D8,$7D,$05,$3F,$C1,$1B,$BE,$97,$FC,$31
	.dc.b	$A7,$F9,$5A,$75,$97,$FA,$C3,$4F,$AB,$87,$28,$C1,$F4,$2F,$8D,$FF
	.dc.b	$B9,$6F,$8B,$FC,$72,$B8,$3F,$D1,$30,$F0,$7C,$E6,$B1,$FC,$CF,$7F
	.dc.b	$EE,$39,$FF,$6F,$F0,$DF,$B4,$6E,$9F,$E2,$FF,$03,$F3,$2F,$F0,$19
	.dc.b	$D3,$4F,$E6,$FE,$F7,$D8,$F9,$4F,$E1,$8D,$8C,$9F,$5E,$32,$D3,$FA
	.dc.b	$3F,$97,$FD,$18,$FE,$29,$E1,$D8,$C9,$36,$FE,$39,$19,$7F,$B0,$FB
	.dc.b	$4F,$A1,$7D,$2F,$F8,$C5,$B7,$F8,$7F,$4B,$13,$23,$E6,$DA,$F9,$4F
	.dc.b	$E4,$AB,$53,$DA,$FF,$89,$37,$C6,$FE,$55,$AF,$76,$63,$D3,$1F,$24
	.dc.b	$6A,$31,$7C,$B8,$7F,$67,$F4,$9F,$DC,$FC,$AF,$F4,$CF,$C5,$FF,$AC
	.dc.b	$E3,$19,$B9,$D3,$FE,$B7,$FA,$5F,$F0,$EE,$5C,$2B,$FD,$F3,$7C,$2F
	.dc.b	$71,$8E,$BF,$F0,$33,$CF,$0E,$D1,$89,$DC,$B0,$EA,$EC,$E5,$8F,$75
	.dc.b	$87,$7C,$39,$CE,$6B,$8D,$F4,$AD,$5F,$23,$A1,$CB,$0F,$16,$1E,$DF
	.dc.b	$5F,$F9,$7D,$F1,$9C,$D7,$18,$C6,$3D,$EB,$77,$E2,$AE,$B8,$78,$1E
	.dc.b	$D5,$BC,$BD,$53,$92,$63,$DA,$BE,$16,$1D,$1F,$06,$1D,$71,$F0,$C7
	.dc.b	$AB,$2F,$7C,$E4,$91,$C0,$1D,$31,$F0,$E1,$DF,$1F,$0E,$5E,$5A,$F4
	.dc.b	$CE,$49,$19,$C7,$C2,$4C,$3B,$D7,$53,$8C,$F6,$23,$E7,$CB,$D8,$FC
	.dc.b	$43,$64,$AF,$C1,$E5,$EC,$8F,$51,$B8,$F0,$3E,$C8,$D3,$9B,$F1,$E3
	.dc.b	$F4,$67,$2F,$BA,$7B,$4F,$DF,$CB,$CB,$7D,$27,$91,$D7,$0F,$3D,$F1
	.dc.b	$8F,$96,$7E,$2C,$A7,$AC,$AF,$EA,$9B,$E9,$B9,$37,$1F,$A8,$8C,$38
	.dc.b	$9F,$F3,$24,$EA,$3F,$3F,$3A,$79,$C7,$E2,$A3,$8D,$FE,$7D,$B8,$F9
	.dc.b	$BE,$95,$F3,$7E,$8A,$D5,$F1,$26,$1F,$A4,$AD,$B8,$9D,$BF,$9C,$C3
	.dc.b	$AB,$F8,$C9,$DB,$9A,$E1,$87,$F0,$BA,$DC,$AB,$77,$D0,$95,$BB,$FD
	.dc.b	$F4,$EA,$31,$5D,$49,$5F,$C2,$27,$51,$F2,$EB,$C2,$D3,$D9,$87,$C1
	.dc.b	$BE,$77,$E0,$9D,$3C,$E7,$51,$89,$D9,$1F,$81,$1A,$7E,$AD,$78,$1A
	.dc.b	$78,$0C,$63,$F7,$1F,$4D,$7D,$5A,$E5,$39,$BF,$25,$72,$7C,$8F,$2C
	.dc.b	$7E,$5C,$EE,$F9,$4E,$70,$F1,$63,$F1,$23,$94,$FA,$5A,$7A,$8E,$07
	.dc.b	$90,$C1,$4D,$92,$3A,$57,$86,$3E,$31,$0F,$BD,$8F,$CA,$E2,$FE,$CD
	.dc.b	$DD,$FA,$8E,$1F,$F2,$DC,$62,$BF,$AA,$CE,$76,$FE,$2F,$D9,$FE,$93
	.dc.b	$FB,$1C,$57,$F6,$7E,$4F,$F0,$FB,$C0,$87,$FA,$46,$37,$87,$F7,$5F
	.dc.b	$56,$D1,$A7,$4F,$23,$FC,$87,$E6,$E3,$0F,$EC,$F3,$5F,$F4,$D7,$91
	.dc.b	$FE,$8F,$7F,$B5,$F4,$74,$23,$FE,$CA,$F8,$2A,$FF,$79,$8B,$BF,$F8
	.dc.b	$8B,$D3,$57,$B3,$6F,$F5,$87,$57,$8F,$1F,$F3,$CE,$3F,$E4,$FF,$0F
	.dc.b	$58,$7F,$6A,$8F,$F5,$75,$6E,$64,$7F,$B6,$A3,$A7,$86,$F1,$3D,$DF
	.dc.b	$0E,$FA,$6A,$6A,$25,$FF,$85,$38,$55,$B8,$C3,$5F,$F5,$95,$F5,$7F
	.dc.b	$C3,$67,$7E,$5F,$AE,$D3,$37,$FE,$82,$DF,$FD,$37,$E3,$FF,$23,$FA
	.dc.b	$6D,$CE,$6F,$F9,$6B,$6F,$FA,$CB,$6B,$84,$74,$9E,$24,$7F,$C5,$B1
	.dc.b	$A7,$FD,$59,$F4,$CF,$8A,$FC,$5F,$9A,$87,$D3,$61,$B5,$FE,$36,$9D
	.dc.b	$56,$4D,$54,$9D,$57,$57,$FD,$0D,$EB,$FD,$B2,$3F,$E4,$0E,$5F,$F7
	.dc.b	$73,$83,$F1,$B1,$75,$FC,$34,$7C,$4F,$95,$8F,$CA,$7E,$6F,$D3,$7F
	.dc.b	$43,$9F,$F5,$6E,$C9,$DD,$27,$31,$C3,$C4,$CC,$62,$7A,$1F,$6A,$FA
	.dc.b	$6F,$FE,$08,$E3,$FF,$53,$FC,$97,$E5,$49,$5F,$E7,$7F,$56,$BF,$A6
	.dc.b	$8B,$87,$FA,$03,$93,$FD,$5D,$3F,$2F,$E8,$9F,$4B,$F4,$97,$D3,$53
	.dc.b	$11,$27,$84,$65,$E5,$1F,$C1,$3D,$E7,$D2,$BE,$9B,$0D,$AF,$F7,$73
	.dc.b	$5F,$F7,$76,$1F,$EF,$F7,$33,$C0,$8F,$9F,$D1,$AF,$FB,$E6,$B9,$67
	.dc.b	$FE,$F4,$C3,$C3,$7E,$BB,$FD,$7F,$95,$30,$FD,$4B,$FB,$4F,$A6,$C4
	.dc.b	$7D,$5E,$1F,$9D,$DD,$FD,$AB,$2F,$F7,$FF,$17,$EC,$B0,$E1,$7F,$EA
	.dc.b	$5F,$5F,$F8,$37,$C8,$48,$C1,$0F,$E2,$EE,$5F,$DB,$C7,$ED,$3E,$9B
	.dc.b	$E9,$C7,$84,$FF,$22,$B8,$C3,$FD,$EE,$DC,$BF,$C3,$FF,$28,$FA,$6E
	.dc.b	$5C,$F3,$0F,$E0,$37,$B7,$FC,$9F,$5B,$73,$3B,$7C,$02,$78,$29,$E0
	.dc.b	$FF,$AB,$BE,$9B,$F5,$8D,$F0,$7F,$9E,$C6,$32,$FF,$77,$7D,$4B,$CC
	.dc.b	$FE,$23,$F4,$2F,$81,$A4,$C7,$FD,$4D,$83,$E1,$79,$92,$32,$FF,$2A
	.dc.b	$F7,$BB,$FD,$3C,$9E,$2F,$6C,$64,$7F,$22,$C7,$C4,$D8,$FE,$A7,$F4
	.dc.b	$EF,$32,$9C,$27,$E1,$D7,$07,$FD,$9B,$3B,$7A,$7D,$37,$F6,$2F,$3D
	.dc.b	$F4,$FF,$C9,$5B,$CD,$F0,$26,$57,$02,$1F,$7F,$CE,$7D,$3F,$58,$F0
	.dc.b	$CF,$06,$9F,$C8,$63,$5C,$6B,$FF,$30,$E7,$FC,$3E,$BE,$07,$95,$F4
	.dc.b	$0F,$E8,$3F,$4E,$F8,$AC,$4E,$88,$7E,$0A,$34,$E3,$0E,$CF,$FA,$62
	.dc.b	$FB,$3F,$1A,$FA,$DF,$F7,$EF,$8C,$FF,$64,$DF,$F8,$31,$B8,$FF,$A8
	.dc.b	$3D,$C4,$63,$07,$A2,$F9,$AF,$F0,$24,$7F,$BA,$67,$55,$C3,$E0,$BB
	.dc.b	$1F,$DD,$EB,$75,$FA,$99,$D3,$C6,$76,$FF,$3F,$9D,$E3,$FE,$20,$FA
	.dc.b	$67,$C4,$7F,$85,$9B,$5F,$EF,$96,$BF,$B5,$C9,$F3,$0F,$8B,$E1,$72
	.dc.b	$B8,$3F,$E0,$B9,$D4,$60,$87,$9E,$FA,$E1,$F3,$B0,$EB,$87,$F8,$A3
	.dc.b	$E9,$9F,$41,$49,$87,$F5,$9F,$A1,$FF,$0A,$BF,$06,$76,$FF,$B3,$B0
	.dc.b	$FF,$1E,$C6,$23,$55,$FE,$1F,$AE,$35,$A8,$EB,$3A,$3B,$7B,$FF,$E3
	.dc.b	$AF,$A3,$FE,$07,$5F,$E4,$86,$F1,$7D,$D6,$D7,$DC,$FA,$B7,$CF,$71
	.dc.b	$F2,$E4,$F9,$6F,$84,$7F,$8C,$6B,$73,$FE,$1A,$B7,$B5,$FD,$A6,$3F
	.dc.b	$C4,$7F,$25,$FE,$C0,$FA,$5F,$E2,$D1,$FE,$EC,$F4,$D3,$9F,$37,$FD
	.dc.b	$8B,$18,$9F,$F7,$8E,$1D,$35,$E1,$7F,$EC,$A8,$C4,$F6,$3F,$DE,$F8
	.dc.b	$76,$70,$1C,$5F,$58,$4F,$28,$C3,$6B,$C9,$3E,$03,$AE,$1E,$68,$F1
	.dc.b	$63,$EF,$BE,$2D,$8E,$35,$E2,$C3,$C3,$1C,$31,$E9,$3E,$2A,$E0,$F6
	.dc.b	$33,$5C,$1F,$87,$1A,$78,$4E,$B0,$F2,$47,$13,$E0,$E3,$EB,$78,$CE
	.dc.b	$D8,$6A,$BA,$4E,$E3,$E4,$57,$46,$9E,$88,$D1,$BB,$F3,$46,$8E,$97
	.dc.b	$BB,$FC,$24,$7F,$C0,$0A,$FB,$2D,$8C,$BF,$14,$FF,$7F,$C3,$E0,$1F
	.dc.b	$4B,$FE,$FD,$8C,$36,$33,$3C,$1A,$88,$7F,$BE,$CA,$4F,$91,$AD,$29
	.dc.b	$C5,$AD,$3F,$1B,$ED,$7F,$A5,$38,$3E,$DC,$67,$6F,$C7,$4E,$63,$F2
	.dc.b	$66,$76,$F1,$C6,$67,$F2,$93,$93,$D1,$19,$9F,$2D,$FE,$22,$31,$3E
	.dc.b	$7A,$EF,$F1,$3F,$D0,$9E,$70,$AE,$13,$9B,$E7,$3E,$A8,$C3,$CE,$3C
	.dc.b	$B1,$88,$F9,$15,$DB,$0F,$7C,$75,$39,$38,$BE,$ED,$8C,$9F,$08,$E1
	.dc.b	$3C,$DC,$4F,$06,$9C,$23,$D7,$19,$6C,$66,$FF,$AC,$6F,$FD,$02,$FF
	.dc.b	$8C,$4F,$F9,$EF,$F2,$57,$AD,$FD,$9B,$83,$FD,$D7,$7B,$AF,$F1,$E4
	.dc.b	$2B,$77,$CB,$C0,$FF,$17,$76,$F1,$8F,$E7,$13,$AB,$FE,$B9,$87,$3C
	.dc.b	$3B,$4E,$5E,$FE,$A7,$1A,$7B,$3F,$E0,$5F,$27,$19,$FF,$8E,$EB,$53
	.dc.b	$9C,$BC,$59,$7F,$AC,$DF,$F6,$80,$BE,$31,$9C,$3F,$CA,$D5,$C5,$F2
	.dc.b	$DF,$4B,$E0,$74,$71,$7E,$2F,$D1,$1F,$27,$E2,$C3,$FC,$B5,$7F,$0E
	.dc.b	$72,$FF,$B8,$63,$38,$7F,$7D,$9C,$BE,$28,$CE,$7F,$E1,$6C,$D5,$A8
	.dc.b	$FF,$3C,$CE,$AF,$FC,$25,$39,$7F,$D5,$91,$9B,$5F,$2C,$3C,$DD,$5F
	.dc.b	$2B,$DA,$7D,$37,$A7,$1C,$35,$FB,$91,$A7,$D4,$68,$A6,$A3,$86,$1F
	.dc.b	$E1,$99,$E0,$FB,$67,$EE,$2E,$67,$D5,$7D,$FD,$5F,$81,$7F,$9A,$6F
	.dc.b	$FE,$00,$D6,$9F,$CB,$CE,$BD,$8E,$73,$9C,$3F,$2B,$7C,$5F,$F4,$1C
	.dc.b	$66,$BF,$39,$39,$7F,$C8,$73,$E3,$C3,$C5,$39,$C3,$F8,$8C,$63,$0E
	.dc.b	$E4,$8F,$D4,$FA,$4F,$97,$FF,$02,$C7,$A2,$75,$87,$FA,$8E,$9F,$59
	.dc.b	$F9,$FA,$FE,$3B,$5F,$9F,$9C,$F4,$70,$CF,$FC,$AF,$18,$AE,$87,$82
	.dc.b	$F9,$61,$FC,$E7,$E2,$BF,$D2,$D3,$FB,$C8,$46,$1F,$F2,$07,$C4,$FF
	.dc.b	$B6,$E7,$53,$9C,$3E,$1B,$63,$FC,$BD,$F4,$BF,$8D,$BE,$DE,$DF,$EC
	.dc.b	$A3,$1C,$A7,$5C,$CF,$92,$FF,$58,$61,$FC,$16,$73,$18,$C3,$FC,$BD
	.dc.b	$1D,$DF,$57,$3F,$5C,$7F,$0F,$97,$86,$7D,$73,$FC,$0B,$E9,$5F,$43
	.dc.b	$A8,$C3,$E5,$F8,$2F,$E0,$B9,$3E,$3F,$1A,$7F,$20,$D7,$CD,$1F,$94
	.dc.b	$FA,$C7,$C7,$FF,$11,$F7,$DF,$4D,$F9,$AE,$3F,$C3,$D6,$B1,$FE,$53
	.dc.b	$5A,$EE,$E5,$87,$1C,$38,$57,$F6,$CB,$E1,$C1,$FE,$A7,$9C,$BF,$DC
	.dc.b	$63,$38,$FF,$A5,$B1,$E9,$7F,$6E,$72,$FF,$77,$8C,$D7,$DC,$F8,$1F
	.dc.b	$E4,$7F,$48,$FA,$5F,$E5,$5E,$CB,$A7,$FD,$77,$7F,$C8,$2B,$58,$FF
	.dc.b	$42,$AD,$61,$CB,$0E,$38,$70,$CF,$FD,$53,$C0,$C7,$8E,$5C,$FD,$EF
	.dc.b	$09,$C3,$E5,$EE,$9F,$3E,$BF,$D4,$ED,$9D,$34,$FC,$C7,$E5,$FF,$83
	.dc.b	$DB,$F9,$C7,$B3,$FC,$22,$31,$E4,$FF,$A2,$AB,$7D,$9F,$5E,$31,$9B
	.dc.b	$5F,$DE,$DF,$0C,$39,$5E,$DB,$87,$DA,$C3,$8D,$FF,$A6,$76,$7C,$77
	.dc.b	$CC,$F8,$2F,$86,$DB,$F9,$13,$A7,$DD,$3A,$69,$F0,$7D,$0F,$AA,$7C
	.dc.b	$CE,$CB,$19,$26,$BF,$B6,$8C,$BE,$B8,$FF,$1A,$9F,$E3,$F3,$FB,$C9
	.dc.b	$FE,$D2,$F8,$EF,$E5,$AE,$63,$81,$FE,$ED,$E0,$4E,$BD,$5F,$ED,$D5
	.dc.b	$C4,$FE,$10,$71,$D3,$FD,$CE,$E6,$F9,$9F,$EF,$79,$E3,$7D,$27,$8C
	.dc.b	$62,$FF,$27,$EF,$C9,$F8,$FF,$37,$70,$FC,$94,$DC,$7E,$84,$DA,$FC
	.dc.b	$F6,$4F,$7B,$F6,$8A,$6C,$61,$B3,$FB,$49,$D3,$67,$4F,$F2,$5F,$99
	.dc.b	$FE,$7D,$8A,$1F,$4B,$F9,$CC,$7F,$A2,$D6,$A7,$3D,$5F,$D6,$A3,$15
	.dc.b	$B9,$FF,$34,$C6,$27,$5E,$67,$CE,$AD,$61,$FE,$6F,$AD,$5F,$F0,$AF
	.dc.b	$A5,$78,$BF,$C4,$5F,$D1,$18,$8D,$74,$66,$74,$D8,$C9,$F9,$4D,$3F
	.dc.b	$81,$FC,$2F,$EE,$9B,$19,$3F,$69,$19,$69,$FE,$3B,$FC,$63,$FB,$43
	.dc.b	$FE,$85,$EF,$FE,$01,$F4,$1F,$E3,$29,$CE,$9F,$E3,$6E,$F4,$E7,$D2
	.dc.b	$FE,$26,$DC,$3F,$81,$3C,$7A,$3F,$C7,$3E,$B3,$E2,$B1,$5F,$CD,$63
	.dc.b	$0F,$8E,$77,$3A,$C7,$FC,$73,$5A,$7F,$3F,$E8,$33,$FF,$74,$E1,$FC
	.dc.b	$47,$DF,$F4,$76,$FF,$26,$6C,$64,$F8,$11,$97,$FB,$57,$D1,$38,$9E
	.dc.b	$A7,$FA,$7F,$E2,$FF,$D4,$F3,$98,$C6,$7F,$EA,$BF,$A4,$FF,$02,$61
	.dc.b	$88,$56,$DF,$F4,$85,$F2,$C3,$AC,$60,$FF,$5A,$4E,$FE,$09,$5F,$EB
	.dc.b	$6C,$3A,$D6,$CE,$B7,$CA,$B7,$1C,$A3,$15,$DF,$1F,$25,$F2,$BD,$DF
	.dc.b	$29,$F4,$61,$E1,$BE,$13,$DE,$F8,$46,$2F,$8D,$6E,$BC,$17,$DB,$D9
	.dc.b	$FD,$74,$79,$E7,$4F,$73,$18,$F9,$9F,$95,$3C,$2F,$AC,$E6,$78,$46
	.dc.b	$2F,$B6,$5E,$8A,$D6,$1D,$8E,$F9,$7A,$30,$F2,$9E,$0B,$E0,$72,$70
	.dc.b	$4C,$3B,$3F,$07,$1F,$26,$5F,$04,$91,$D2,$7E,$5D,$F7,$8F,$3C,$E9
	.dc.b	$A7,$BB,$2F,$0B,$DE,$35,$3C,$71,$F5,$65,$EC,$3E,$14,$65,$C6,$5C
	.dc.b	$B1,$F0,$E5,$E6,$C7,$C9,$8F,$87,$1F,$87,$39,$9F,$19,$23,$39,$73
	.dc.b	$AE,$18,$F8,$2F,$85,$79,$31,$61,$0F,$C3,$94,$FB,$C5,$8F,$D1,$37
	.dc.b	$D4,$FF,$2B,$BB,$3F,$4E,$E1,$D7,$AD,$97,$2D,$9F,$F5,$9B,$F9,$89
	.dc.b	$E7,$9F,$E5,$1B,$97,$97,$E9,$3F,$BE,$E1,$BB,$8F,$E5,$6B,$6F,$B2
	.dc.b	$76,$E6,$77,$7D,$5F,$DF,$4E,$A3,$13,$B2,$3F,$AE,$8D,$5F,$F0,$AB
	.dc.b	$EF,$5E,$C3,$71,$E7,$F2,$7E,$65,$F4,$24,$71,$C3,$C3,$3A,$8C,$57
	.dc.b	$23,$F8,$5D,$F6,$C3,$C3,$18,$BF,$5D,$70,$C3,$CD,$87,$3A,$D1,$CC
	.dc.b	$F3,$46,$9E,$26,$9F,$2B,$F2,$63,$93,$C2,$7B,$11,$D1,$C8,$D1,$CD
	.dc.b	$F8,$4E,$71,$FB,$EF,$8F,$0F,$5B,$C0,$91,$AB,$F3,$47,$A6,$31,$5E
	.dc.b	$18,$57,$29,$CC,$78,$48,$6A,$FD,$71,$A8,$EF,$8F,$CD,$9C,$BF,$72
	.dc.b	$74,$6E,$BE,$8F,$13,$EB,$DF,$37,$FA,$9F,$13,$FC,$DF,$F0,$BF,$EC
	.dc.b	$DC,$FF,$D3,$0D,$7F,$AF,$D7,$07,$F6,$33,$FC,$BA,$FA,$5F,$3E,$2F
	.dc.b	$EF,$58,$71,$AE,$BC,$39,$C1,$87,$07,$9D,$FE,$F2,$3C,$DE,$96,$FB
	.dc.b	$9D,$32,$FF,$BC,$3C,$EB,$71,$C8,$EB,$A2,$F9,$57,$19,$FE,$F6,$F1
	.dc.b	$9D,$47,$1A,$F8,$DC,$1E,$1D,$33,$3F,$E5,$EC,$51,$8D,$3E,$16,$1F
	.dc.b	$B8,$6D,$F1,$CA,$61,$D6,$B5,$9F,$F1,$AF,$4F,$FD,$51,$EC,$77,$F5
	.dc.b	$25,$FF,$A0,$23,$E4,$4E,$A3,$EA,$CE,$9F,$F7,$7C,$FC,$DF,$51,$F4
	.dc.b	$0F,$83,$C2,$FA,$1A,$62,$25,$FF,$B3,$A7,$4D,$9C,$BF,$E7,$4B,$FF
	.dc.b	$15,$3F,$F0,$07,$E8,$7F,$CD,$0D,$F2,$3E,$F3,$FE,$E3,$F0,$ED,$70
	.dc.b	$8F,$F3,$ED,$70,$6D,$6A,$A4,$73,$21,$E1,$7D,$DF,$40,$FA,$1B,$EE
	.dc.b	$CA,$FF,$73,$CE,$5B,$39,$7F,$D2,$FA,$A7,$FD,$E7,$1A,$3F,$43,$C7
	.dc.b	$F8,$12,$3F,$EB,$2F,$9F,$AF,$E0,$F9,$BE,$DA,$E7,$FC,$2D,$7D,$8F
	.dc.b	$65,$7F,$13,$E7,$FF,$D4,$B7,$F9,$71,$F1,$4F,$92,$F9,$2D,$8D,$4F
	.dc.b	$07,$FD,$4A,$DE,$BF,$F0,$FC,$72,$AE,$27,$7F,$6A,$47,$7B,$E8,$FF
	.dc.b	$AB,$9B,$5C,$AF,$A7,$57,$EE,$BD,$8F,$D1,$12,$FF,$DA,$7E,$2D,$7E
	.dc.b	$14,$5B,$FF,$6E,$4F,$C1,$FB,$BF,$D5,$7B,$52,$FF,$DB,$71,$97,$F8
	.dc.b	$7C,$65,$FF,$65,$F7,$37,$57,$E3,$3E,$35,$EF,$7E,$7F,$F0,$A7,$C9
	.dc.b	$53,$51,$31,$FE,$09,$D0,$AD,$C6,$1A,$FF,$B4,$2F,$AB,$F5,$5A,$FE
	.dc.b	$22,$37,$AF,$F9,$3A,$BA,$61,$DE,$BF,$09,$D9,$FC,$24,$8F,$FB,$EB
	.dc.b	$EC,$6B,$A3,$F9,$7C,$7E,$1E,$3F,$11,$E7,$7F,$11,$F8,$19,$5F,$C7
	.dc.b	$77,$F8,$93,$A8,$FE,$6F,$3A,$7E,$4E,$FF,$E1,$BC,$FF,$0B,$F8,$A3
	.dc.b	$EC,$3F,$83,$7D,$27,$F2,$39,$F5,$37,$9B,$FC,$59,$5B,$C3,$F2,$CD
	.dc.b	$7F,$53,$3B,$5B,$A3,$10,$9E,$17,$E2,$C3,$C3,$E6,$4A,$EB,$1F,$CC
	.dc.b	$F4,$FE,$81,$EE,$3E,$1A,$99,$AF,$D4,$7C,$EF,$FA,$9D,$FE,$71,$F4
	.dc.b	$8F,$A3,$FE,$88,$D8,$DF,$B3,$FD,$1F,$E8,$9F,$47,$FC,$5B,$0E,$54
	.dc.b	$DB,$F5,$71,$87,$FA,$F5,$70,$9F,$E8,$CB,$EA,$FC,$FF,$07,$27,$31
	.dc.b	$D7,$DB,$F0,$32,$3D,$31,$96,$90,$FD,$0F,$D1,$75,$7D,$FE,$7F,$F2
	.dc.b	$CF,$A3,$7C,$65,$3D,$51,$27,$F8,$64,$E9,$B3,$97,$DD,$D3,$FE,$69
	.dc.b	$D7,$F9,$A6,$BF,$03,$0D,$3F,$C3,$30,$FD,$0B,$F1,$33,$F1,$7C,$6C
	.dc.b	$C7,$E9,$C7,$CB,$30,$F6,$3F,$CE,$F4,$9D,$47,$03,$CE,$F0,$AE,$D1
	.dc.b	$FD,$CF,$E8,$DF,$13,$FE,$03,$EC,$FF,$6C,$CF,$F7,$36,$D7,$F8,$4E
	.dc.b	$B7,$3C,$6B,$87,$C1,$C8,$C1,$0E,$4E,$5F,$F7,$EC,$F2,$7F,$DF,$FE
	.dc.b	$33,$E6,$3C,$51,$0F,$F0,$44,$65,$B1,$97,$CC,$7F,$85,$E3,$4F,$48
	.dc.b	$D6,$1F,$ED,$8F,$4B,$F5,$73,$B9,$FF,$0F,$7C,$FF,$EB,$E5,$F2,$7F
	.dc.b	$AC,$E2,$FF,$51,$C6,$3B,$F7,$E0,$4C,$7F,$C4,$93,$96,$DF,$17,$FC
	.dc.b	$67,$87,$EE,$FE,$04,$F7,$9D,$57,$1A,$FB,$95,$AB,$FF,$19,$7E,$8A
	.dc.b	$F9,$6F,$F1,$F6,$BF,$CE,$7A,$AF,$8C,$77,$33,$1F,$F1,$C5,$6D,$B5
	.dc.b	$B7,$FC,$69,$7D,$5F,$ED,$1A,$7F,$40,$AF,$F6,$1F,$B0,$F8,$AF,$F6
	.dc.b	$27,$D0,$FF,$6C,$BC,$C6,$3D,$6F,$F6,$45,$6E,$7F,$D9,$78,$73,$7F
	.dc.b	$C7,$B5,$FE,$CD,$8C,$4E,$E5,$7D,$5D,$9F,$ED,$47,$11,$B9,$E3,$5C
	.dc.b	$FC,$FF,$E1,$D7,$C0,$E7,$87,$92,$7C,$D1,$E8,$3C,$B3,$C2,$FA,$36
	.dc.b	$77,$3C,$F0,$F0,$C7,$7C,$79,$4F,$DC,$AE,$0F,$9F,$0F,$03,$4D,$E3
	.dc.b	$E8,$7C,$35,$CB,$0E,$B1,$B3,$D5,$3B,$7B,$4E,$EB,$AC,$7B,$8E,$B7
	.dc.b	$E2,$76,$7A,$63,$77,$DA,$33,$8F,$C0,$BE,$91,$0F,$0D,$F3,$C3,$ED
	.dc.b	$40,$F8,$47,$FB,$FD,$CB,$83,$47,$E0,$BC,$C7,$D0,$FF,$C0,$19,$46
	.dc.b	$27,$93,$5E,$27,$EA,$1C,$38,$1F,$17,$5F,$F7,$69,$4C,$95,$C3,$8C
	.dc.b	$7F,$21,$F3,$CF,$8E,$74,$7D,$CA,$F7,$5F,$DE,$AD,$4E,$67,$96,$FE
	.dc.b	$E9,$CC,$7E,$44,$E2,$7E,$4D,$CE,$1F,$9F,$8C,$FA,$5E,$2C,$7C,$30
	.dc.b	$8D,$3E,$A7,$58,$7D,$17,$38,$F9,$6B,$8D,$E2,$FC,$11,$8C,$7E,$45
	.dc.b	$6F,$0F,$53,$6B,$D0,$7B,$0F,$BD,$8F,$A2,$B8,$C7,$4A,$E2,$D3,$8D
	.dc.b	$73,$C3,$C4,$7A,$23,$26,$DF,$36,$39,$2D,$FC,$BC,$7E,$1C,$7F,$0B
	.dc.b	$E6,$FF,$3D,$FD,$0F,$F9,$D3,$0F,$F6,$E7,$12,$31,$1F,$EE,$28,$C6
	.dc.b	$7F,$87,$AE,$2D,$8C,$CF,$F6,$BE,$17,$AD,$FD,$4E,$BF,$DB,$15,$A9
	.dc.b	$FA,$A6,$BB,$2C,$F8,$BE,$1E,$E1,$E1,$6F,$AA,$FA,$37,$C3,$6B,$4F
	.dc.b	$F4,$AD,$5C,$FF,$D9,$98,$F9,$63,$D7,$3A,$8C,$13,$1F,$F7,$42,$F9
	.dc.b	$9E,$55,$E2,$EF,$FC,$DC,$77,$FA,$2F,$E8,$D5,$FE,$F3,$6D,$FF,$89
	.dc.b	$3D,$2D,$78,$12,$BE,$45,$6A,$BF,$77,$92,$31,$1F,$C0,$E3,$15,$E8
	.dc.b	$9E,$BC,$1F,$DC,$72,$78,$3D,$70,$FF,$80,$11,$8C,$7F,$86,$57,$E8
	.dc.b	$B1,$7D,$17,$F6,$19,$C9,$DD,$C5,$FF,$8F,$3E,$5F,$E6,$46,$2F,$93
	.dc.b	$A9,$F4,$B5,$E5,$E1,$FC,$E8,$C4,$30,$F0,$CF,$CB,$C7,$F8,$2E,$3E
	.dc.b	$2A,$D5,$F7,$8F,$F0,$3C,$EA,$7F,$C6,$C3,$B1,$E9,$7E,$1A,$30,$FE
	.dc.b	$23,$C9,$7D,$0B,$E4,$B5,$7C,$31,$FF,$6E,$37,$0F,$E4,$4D,$7F,$93
	.dc.b	$B5,$FE,$29,$87,$E9,$1E,$F3,$FE,$04,$FA,$27,$CE,$F6,$AF,$CC,$EE
	.dc.b	$9C,$D7,$82,$BC,$67,$9F,$E5,$3F,$DD,$91,$92,$61,$E4,$9F,$06,$3C
	.dc.b	$A3,$13,$FC,$DE,$31,$87,$F2,$0D,$3E,$0F,$67,$6A,$F9,$D1,$87,$FD
	.dc.b	$F9,$EF,$39,$FE,$53,$63,$FD,$75,$E5,$FC,$97,$73,$A3,$E1,$4E,$F2
	.dc.b	$FE,$94,$D7,$FA,$7B,$5F,$CE,$7D,$17,$E7,$9F,$C2,$37,$94,$D2,$57
	.dc.b	$EE,$1E,$53,$C2,$30,$78,$BA,$66,$1F,$D6,$70,$ED,$47,$A7,$3F,$32
	.dc.b	$4F,$F1,$BC,$11,$F7,$7C,$D7,$C7,$FE,$35,$B5,$FE,$F6,$FA,$0F,$E3
	.dc.b	$4F,$F7,$48,$CE,$3F,$96,$6B,$FD,$C1,$AE,$EB,$B1,$27,$A7,$57,$78
	.dc.b	$FE,$43,$EA,$30,$FD,$25,$BC,$37,$03,$E5,$BF,$8B,$7C,$17,$FA,$C3
	.dc.b	$3F,$B3,$18,$F7,$3F,$94,$D6,$E7,$FC,$DF,$5B,$C3,$FD,$E7,$7B,$9C
	.dc.b	$FC,$DF,$43,$8F,$CB,$58,$C9,$2B,$FD,$F7,$19,$7F,$6F,$F4,$1F,$85
	.dc.b	$E9,$FE,$E7,$08,$C5,$7F,$88,$76,$FE,$61,$ED,$7E,$EE,$73,$5F,$E2
	.dc.b	$57,$F0,$2F,$1E,$7F,$D9,$46,$08,$FC,$8F,$1B,$FD,$E9,$87,$F0,$DA
	.dc.b	$E0,$FF,$0F,$C3,$F9,$F7,$C2,$FC,$4F,$92,$F9,$3E,$DD,$AF,$8A,$DB
	.dc.b	$FF,$63,$78,$CE,$14,$E7,$6F,$F5,$24,$62,$FE,$6F,$A5,$F5,$0F,$A3
	.dc.b	$F1,$D6,$32,$4C,$7F,$DA,$5C,$DF,$72,$75,$1F,$C8,$27,$FC,$7A,$7A
	.dc.b	$CA,$FF,$B6,$57,$D0,$78,$63,$87,$AF,$FC,$96,$77,$1F,$E4,$26,$BF
	.dc.b	$E4,$C6,$BF,$E5,$36,$BF,$E3,$A8,$FE,$E7,$0E,$57,$C8,$F5,$9C,$C6
	.dc.b	$27,$F1,$BF,$01,$F9,$43,$F2,$3F,$1F,$63,$24,$C3,$E0,$F5,$27,$51
	.dc.b	$FE,$F5,$E4,$68,$F8,$DF,$D2,$BB,$C3,$EE,$39,$6D,$EA,$76,$D8,$C9
	.dc.b	$F8,$5F,$A0,$FE,$52,$FF,$9B,$9B,$87,$F9,$6B,$6F,$81,$3B,$F2,$BF
	.dc.b	$A8,$57,$F3,$3A,$FF,$36,$B5,$E5,$1F,$D1,$E3,$11,$FE,$7F,$7A,$CF
	.dc.b	$D8,$EE,$E7,$5F,$3A,$FF,$69,$C9,$FE,$5A,$D1,$F4,$1F,$C5,$B0,$FC
	.dc.b	$D3,$FA,$68,$CF,$89,$FC,$22,$30,$6F,$3F,$94,$E2,$35,$D9,$FC,$13
	.dc.b	$6F,$E1,$3D,$D7,$1F,$E0,$AD,$C3,$F4,$3D,$08,$C6,$7F,$A0,$7F,$C6
	.dc.b	$5F,$3F,$FD,$9E,$BF,$C0,$0D,$F2,$4D,$27,$9F,$FC,$63,$E8,$38,$B7
	.dc.b	$2F,$F0,$25,$70,$E8,$EA,$67,$0F,$F7,$EE,$1F,$51,$B1,$FE,$3E,$8C
	.dc.b	$47,$F8,$B9,$AF,$F8,$B5,$AF,$9B,$B3,$81,$2F,$FC,$BD,$3F,$EF,$76
	.dc.b	$C7,$F2,$EF,$77,$DE,$D8,$FE,$1D,$F3,$FF,$DA,$FC,$EF,$F4,$FF,$CA
	.dc.b	$FF,$AE,$23,$18,$FF,$AB,$A3,$FD,$77,$F4,$0F,$4F,$FD,$2F,$F0,$49
	.dc.b	$E1,$CA,$C7,$9E,$1F,$EC,$7C,$79,$DE,$E3,$15,$B7,$B1,$FE,$CD,$C7
	.dc.b	$B6,$5C,$A1,$7C,$B1,$EF,$8E,$63,$13,$C7,$2F,$2D,$6A,$3A,$9C,$B1
	.dc.b	$F0,$65,$E3,$8C,$61,$DF,$2F,$0C,$70,$C7,$AB,$A3,$17,$CB,$0F,$8B
	.dc.b	$8F,$82,$7D,$B8,$79,$6B,$51,$83,$C7,$87,$6C,$7C,$58,$F6,$25,$70
	.dc.b	$A5,$6F,$1F,$65,$F0,$AD,$5F,$9A,$B5,$7D,$EB,$58,$78,$CF,$76,$3E
	.dc.b	$79,$F6,$4E,$9F,$6C,$EB,$2E,$11,$82,$4F,$C9,$BE,$F1,$CB,$2F,$29
	.dc.b	$1F,$A9,$5C,$AB,$B1,$27,$E4,$C6,$5E,$13,$CC,$E2,$78,$1C,$E3,$EF
	.dc.b	$C7,$D4,$FC,$8C,$79,$57,$0C,$3B,$3F,$0E,$FA,$E3,$E8,$BE,$13,$D4
	.dc.b	$E1,$3D,$B2,$FC,$B4,$0F,$C6,$14,$FB,$85,$3F,$0C,$59,$7A,$5F,$E3
	.dc.b	$D8,$D1,$F9,$C7,$0E,$DA,$FE,$CD,$AF,$EA,$DE,$58,$7E,$21,$B7,$BF
	.dc.b	$4B,$F5,$1F,$40,$F2,$7C,$D7,$B9,$FD,$45,$F0,$C3,$A4,$70,$AD,$E1
	.dc.b	$CE,$7F,$45,$5B,$7D,$15,$B7,$9B,$1D,$B9,$69,$2F,$C4,$FE,$D3,$0F
	.dc.b	$33,$C6,$32,$4B,$F9,$D3,$A9,$E5,$1E,$59,$D4,$7D,$59,$D4,$62,$3C
	.dc.b	$E7,$03,$58,$7B,$9F,$82,$64,$DC,$7F,$05,$9E,$75,$DC,$93,$C2,$75
	.dc.b	$5C,$E3,$8C,$E9,$A7,$80,$F7,$39,$C7,$F0,$B8,$7C,$7B,$F4,$46,$2B
	.dc.b	$C9,$87,$89,$F0,$5F,$1C,$7E,$A4,$E6,$7B,$92,$B9,$5F,$9A,$3C,$27
	.dc.b	$22,$09,$E5,$3C,$2F,$B6,$1E,$56,$D7,$93,$1F,$95,$C6,$FA,$0F,$F6
	.dc.b	$D4,$FF,$B9,$FC,$A7,$C2,$7F,$49,$8C,$35,$75,$7F,$C0,$1A,$E3,$18
	.dc.b	$E3,$FD,$86,$1D,$7D,$9F,$D4,$CF,$FB,$DE,$B6,$FE,$CE,$74,$FE,$B9
	.dc.b	$AF,$ED,$5F,$37,$A3,$C9,$B1,$B9,$FF,$4B,$36,$75,$18,$8F,$21,$E1
	.dc.b	$8F,$E0,$58,$7A,$5B,$87,$F0,$7C,$BE,$0C,$62,$BD,$19,$D9,$CC,$7C
	.dc.b	$49,$D6,$3F,$2E,$AE,$3F,$A1,$AF,$73,$5F,$8A,$D7,$A9,$FC,$3B,$CC
	.dc.b	$FF,$58,$FC,$07,$7D,$5E,$95,$3C,$93,$AA,$FF,$75,$D6,$5F,$AF,$3A
	.dc.b	$F2,$D3,$EB,$F1,$5F,$40,$F8,$0C,$2F,$A0,$A6,$22,$63,$FE,$9D,$9D
	.dc.b	$36,$72,$FC,$E3,$BF,$C8,$7C,$95,$F4,$1F,$E6,$D7,$FD,$50,$D7,$86
	.dc.b	$29,$E0,$D8,$FC,$4B,$5F,$E6,$6D,$7F,$9D,$35,$FE,$7C,$EF,$93,$FD
	.dc.b	$8A,$DE,$4C,$D7,$96,$FF,$9D,$D7,$F9,$8F,$E7,$DF,$41,$F1,$DF,$F7
	.dc.b	$BE,$3F,$EB,$D6,$BF,$E8,$97,$6E,$63,$87,$B7,$23,$07,$EA,$B9,$3E
	.dc.b	$E1,$1F,$EA,$D6,$CB,$FA,$C7,$17,$FB,$CB,$E8,$35,$5F,$EC,$6E,$4F
	.dc.b	$E5,$97,$B8,$74,$7E,$E3,$1F,$C0,$5F,$D1,$F9,$FF,$F5,$5D,$7F,$B5
	.dc.b	$FE,$21,$F4,$1F,$55,$B1,$FE,$DB,$9D,$BF,$6D,$AF,$F6,$76,$BF,$DA
	.dc.b	$DA,$FF,$6F,$6A,$79,$E3,$FE,$F6,$6E,$9C,$7D,$AF,$17,$C4,$7E,$7F
	.dc.b	$D9,$98,$7F,$10,$7F,$35,$E0,$62,$3E,$14,$5C,$7E,$17,$03,$AF,$FC
	.dc.b	$8D,$18,$5F,$41,$FA,$BF,$6A,$61,$F4,$34,$F9,$F5,$A8,$C0,$C3,$A5
	.dc.b	$F0,$8F,$A5,$7F,$AD,$8F,$CF,$7C,$67,$FB,$E2,$78,$F7,$3E,$81,$F2
	.dc.b	$5F,$B1,$6C,$F6,$15,$C5,$FD,$53,$5F,$F0,$B3,$5F,$F0,$D3,$5F,$F0
	.dc.b	$F3,$5E,$7F,$01,$E4,$23,$E4,$9F,$D8,$6B,$F8,$9E,$7F,$E7,$FE,$C7
	.dc.b	$F0,$F2,$57,$C2,$D7,$E6,$7C,$86,$23,$F2,$53,$6F,$E7,$4F,$F0,$FF
	.dc.b	$27,$F8,$C7,$C0,$CB,$E5,$C3,$FA,$EE,$2F,$9F,$59,$9F,$CE,$57,$17
	.dc.b	$E2,$7C,$FB,$E5,$BF,$8C,$7D,$03,$A2,$9F,$C3,$67,$55,$C2,$BF,$A7
	.dc.b	$A8,$CB,$FE,$3A,$6B,$FE,$41,$6B,$FE,$45,$77,$A7,$EB,$63,$94,$F7
	.dc.b	$3C,$7E,$F7,$F3,$EC,$BF,$A7,$92,$7F,$AA,$3E,$2F,$9B,$7A,$94,$93
	.dc.b	$FD,$92,$32,$FF,$61,$F9,$D7,$CF,$7F,$5F,$6C,$66,$30,$67,$E7,$64
	.dc.b	$71,$6B,$FE,$58,$AF,$F2,$DB,$DF,$D7,$F2,$11,$C1,$F9,$38,$DB,$FC
	.dc.b	$ED,$BA,$7E,$BB,$E7,$97,$FC,$F6,$7F,$96,$57,$F9,$BF,$E7,$9F,$11
	.dc.b	$FD,$05,$B1,$D7,$CE,$E0,$7C,$52,$3F,$0D,$AF,$85,$DB,$FA,$78,$FD
	.dc.b	$36,$3F,$56,$FC,$73,$9F,$8A,$F0,$1E,$5C,$64,$7C,$F7,$EA,$11,$C1
	.dc.b	$F0,$5D,$BF,$D6,$ED,$F0,$9F,$DF,$BE,$79,$F0,$DF,$E0,$3F,$0F,$FD
	.dc.b	$1F,$9F,$EA,$DB,$D5,$FE,$22,$AD,$E3,$FC,$33,$D9,$46,$23,$FC,$3C
	.dc.b	$D7,$FC,$48,$D7,$96,$1F,$E2,$18,$E1,$87,$F4,$1A,$E7,$3E,$53,$FC
	.dc.b	$0B,$F3,$4D,$7F,$C2,$BE,$1B,$E5,$3D,$19,$43,$C9,$19,$6C,$7F,$8B
	.dc.b	$E3,$11,$FE,$33,$6B,$FE,$39,$6B,$FE,$40,$77,$8F,$95,$B9,$E6,$BC
	.dc.b	$B8,$FF,$8A,$3E,$6F,$F7,$98,$7F,$94,$7E,$53,$F6,$37,$FE,$33,$6C
	.dc.b	$34,$57,$08,$FF,$1E,$C6,$3B,$BF,$C8,$D3,$99,$FF,$27,$35,$FF,$2A
	.dc.b	$B5,$E8,$F3,$CF,$E6,$37,$0D,$D7,$F8,$C3,$E8,$FF,$B3,$DF,$83,$E7
	.dc.b	$DF,$2D,$FE,$47,$C7,$FA,$97,$55,$F1,$3F,$53,$97,$F9,$92,$F7,$87
	.dc.b	$E0,$DA,$FF,$97,$DA,$FF,$99,$5A,$F1,$C3,$F8,$EE,$7F,$D3,$31,$FF
	.dc.b	$6F,$78,$AF,$89,$FF,$6F,$7C,$FB,$D4,$FF,$70,$4E,$71,$FF,$72,$4E
	.dc.b	$5E,$35,$FE,$E5,$C3,$A4,$2B,$83,$F4,$3C,$EF,$F4,$0B,$FE,$86,$AE
	.dc.b	$F8,$7F,$BB,$7D,$8D,$E3,$D6,$B4,$7F,$BC,$F1,$EB,$3D,$1B,$7C,$A7
	.dc.b	$FD,$F1,$3C,$AF,$A4,$F9,$70,$F0,$CF,$16,$C7,$1A,$EE,$30,$F1,$56
	.dc.b	$F2,$EB,$5C,$DB,$5F,$13,$0E,$ED,$32,$7C,$18,$B8,$FA,$23,$47,$83
	.dc.b	$0E,$F7,$D6,$3C,$73,$F3,$32,$F6,$47,$D0,$38,$BE,$57,$E8,$4F,$1C
	.dc.b	$77,$1D,$71,$F2,$BE,$33,$AE,$5F,$09,$F1,$CF,$03,$A1,$D5,$1E,$52
	.dc.b	$9F,$56,$3E,$EB,$83,$B1,$70,$79,$4F,$9F,$FF,$5E,$C6,$1B,$19,$9E
	.dc.b	$47,$EE,$1C,$3F,$EC,$4F,$88,$FE,$EE,$9A,$47,$FB,$DB,$F4,$BE,$69
	.dc.b	$F1,$BF,$E6,$1C,$BF,$2B,$5E,$CA,$D4,$E7,$4F,$0C,$63,$8B,$ED,$D6
	.dc.b	$CF,$C3,$E1,$D2,$33,$1E,$33,$37,$CD,$F8,$95,$CD,$FC,$79,$CF,$0F
	.dc.b	$D3,$DF,$38,$C9,$CE,$33,$5C,$E3,$E8,$D7,$39,$F5,$C6,$F0,$FA,$B7
	.dc.b	$CE,$73,$5B,$9F,$6D,$79,$B0,$F4,$CE,$63,$CE,$64,$F1,$B9,$7C,$26
	.dc.b	$6B,$AB,$6B,$AC,$7A,$2B,$F4,$FC,$FF,$C3,$7B,$7F,$D2,$5C,$5F,$D4
	.dc.b	$F8,$11,$8A,$FF,$4D,$FC,$FF,$F1,$37,$FB,$14,$E9,$B3,$BC,$BF,$9F
	.dc.b	$4E,$5F,$F7,$CF,$A2,$BE,$51,$FD,$02,$75,$87,$87,$E1,$79,$65,$FC
	.dc.b	$85,$E9,$87,$1F,$83,$CD,$FF,$B8,$73,$F1,$D6,$0D,$E5,$FE,$E8,$8C
	.dc.b	$93,$2F,$04,$EA,$F8,$4F,$F7,$DF,$9F,$75,$F8,$9B,$1F,$E6,$BF,$2F
	.dc.b	$FC,$86,$EE,$74,$47,$5E,$75,$C7,$FA,$ED,$F2,$6D,$7B,$2F,$9C,$62
	.dc.b	$FF,$C1,$DF,$3F,$FD,$51,$BD,$9B,$F4,$7E,$2C,$E6,$BF,$B4,$B7,$87
	.dc.b	$FC,$18,$DB,$FF,$69,$E0,$BE,$38,$49,$EA,$7B,$63,$18,$30,$FF,$35
	.dc.b	$73,$3E,$4F,$EF,$4A,$31,$7F,$EB,$0F,$90,$7C,$FB,$E3,$B7,$5A,$BE
	.dc.b	$97,$FA,$4C,$32,$7E,$27,$0E,$37,$89,$FF,$57,$5F,$1C,$7F,$2F,$F3
	.dc.b	$EF,$99,$D6,$5D,$0F,$44,$66,$F8,$CF,$88,$ED,$F2,$9F,$EE,$A8,$C9
	.dc.b	$2B,$91,$ED,$9D,$FC,$37,$19,$D6,$1F,$EF,$9C,$FD,$FE,$66,$70,$F1
	.dc.b	$46,$2B,$FD,$53,$A7,$C8,$F6,$7E,$FE,$1F,$ED,$D8,$C7,$34,$AF,$F6
	.dc.b	$65,$69,$FF,$6E,$CE,$70,$FF,$13,$FC,$FF,$E2,$F0,$FF,$25,$36,$BF
	.dc.b	$A1,$DF,$1A,$FF,$18,$4E,$5B,$87,$EB,$A3,$5D,$9F,$EF,$1D,$B1,$3E
	.dc.b	$7E,$79,$8F,$F9,$86,$FA,$60,$C7,$A7,$6A,$7F,$1D,$5B,$E5,$FD,$DF
	.dc.b	$D1,$BE,$17,$C1,$1F,$8A,$F3,$DF,$17,$4F,$E1,$5E,$AB,$E8,$3F,$82
	.dc.b	$E3,$FD,$7A,$B6,$FF,$BC,$5B,$AF,$D8,$6F,$12,$7F,$B3,$DF,$3F,$13
	.dc.b	$B4,$71,$C6,$E1,$F8,$9A,$D4,$E4,$FF,$70,$FC,$A7,$F1,$DF,$84,$7C
	.dc.b	$D7,$FA,$9A,$72,$7F,$74,$CB,$FD,$E9,$19,$C3,$F9,$0F,$A6,$C3,$94
	.dc.b	$E6,$BC,$D4,$E7,$5E,$F0,$E8,$7F,$80,$31,$F0,$FB,$FC,$BE,$6A,$4F
	.dc.b	$FB,$DB,$E5,$6E,$3F,$A3,$BF,$3D,$63,$2E,$F9,$76,$BC,$61,$FE,$0A
	.dc.b	$FA,$0F,$BA,$7E,$27,$67,$D0,$7F,$AA,$27,$C7,$9F,$F8,$4A,$B7,$D7
	.dc.b	$FC,$11,$E1,$5B,$9F,$D5,$C6,$AB,$8C,$23,$7E,$DF,$1B,$FE,$7F,$BB
	.dc.b	$E4,$BE,$3F,$BB,$86,$E3,$E2,$D7,$E3,$C6,$32,$F9,$ED,$DB,$E9,$FA
	.dc.b	$5F,$40,$EB,$D5,$70,$F5,$F1,$7D,$C9,$C6,$1F,$62,$AD,$FD,$68,$FC
	.dc.b	$67,$AB,$FC,$8B,$0F,$F1,$59,$F1,$07,$CF,$34,$D4,$C9,$F8,$27,$0A
	.dc.b	$FE,$30,$F0,$9D,$47,$F1,$28,$D5,$F8,$5A,$7F,$90,$8E,$07,$4A,$FC
	.dc.b	$6F,$BF,$23,$E3,$6C,$F8,$FC,$61,$F8,$5E,$8F,$F6,$75,$7C,$26,$C7
	.dc.b	$E6,$27,$78,$FE,$5B,$97,$F7,$5F,$19,$FB,$6D,$BA,$63,$F8,$49,$CB
	.dc.b	$63,$38,$7E,$B5,$57,$1C,$39,$36,$BF,$0B,$F3,$FF,$CE,$6F,$F9,$9B
	.dc.b	$71,$FD,$3C,$FC,$EE,$9F,$EA,$73,$98,$FF,$31,$B6,$BF,$CC,$ED,$8E
	.dc.b	$07,$6F,$2F,$89,$1F,$B5,$C2,$BC,$5F,$F0,$5F,$9F,$FE,$3A,$DE,$7F
	.dc.b	$E2,$F1,$8C,$BE,$33,$FA,$93,$18,$7F,$12,$71,$E1,$70,$AF,$CF,$E9
	.dc.b	$FC,$13,$A6,$DF,$D3,$6E,$3F,$5D,$B8,$FF,$0E,$7F,$C4,$FF,$3F,$FD
	.dc.b	$96,$BF,$D0,$BE,$32,$31,$7F,$C3,$79,$3F,$12,$FF,$BC,$E3,$33,$BF
	.dc.b	$43,$99,$FB,$2E,$7B,$87,$F0,$09,$E9,$BF,$80,$FE,$37,$F0,$D6,$75
	.dc.b	$87,$C9,$6C,$E9,$A6,$BA,$BD,$27,$F0,$0F,$80,$B1,$92,$67,$FC,$AA
	.dc.b	$32,$FF,$76,$F9,$D7,$0B,$CF,$FF,$45,$7C,$A7,$FA,$92,$73,$18,$AF
	.dc.b	$F5,$4F,$CF,$BD,$0F,$E6,$3F,$00,$C3,$86,$3C,$EF,$FD,$39,$18,$C3
	.dc.b	$94,$31,$EA,$E8,$EB,$87,$4A,$E1,$19,$38,$5F,$5C,$3C,$11,$FE,$BC
	.dc.b	$BE,$71,$88,$63,$E7,$CF,$C3,$5A,$C3,$AD,$EE,$31,$7E,$2A,$DE,$1D
	.dc.b	$EF,$64,$9D,$1E,$5A,$E2,$F9,$B2,$E2,$7C,$0C,$7C,$51,$BB,$E8,$FA
	.dc.b	$5A,$F1,$3E,$04,$70,$C7,$CD,$3A,$BE,$E7,$82,$75,$87,$96,$F9,$5F
	.dc.b	$0C,$3C,$59,$7C,$D9,$F0,$47,$99,$F2,$3A,$CB,$15,$C6,$72,$79,$2B
	.dc.b	$93,$DA,$FC,$18,$F8,$E7,$2F,$9B,$2F,$41,$EE,$CB,$CF,$7E,$23,$AE
	.dc.b	$1D,$CF,$B5,$1A,$8E,$C6,$E3,$58,$7F,$BE,$A7,$57,$DF,$2F,$4C,$E7
	.dc.b	$1F,$1E,$3E,$29,$CE,$3F,$6B,$1E,$6F,$5B,$F9,$58,$F4,$BE,$13,$FE
	.dc.b	$FC,$C7,$F3,$28,$FC,$B9,$4F,$61,$4F,$D0,$7A,$3F,$E4,$17,$93,$5D
	.dc.b	$47,$FA,$AB,$D1,$E1,$1F,$9D,$9D,$3F,$B5,$79,$63,$F9,$96,$E3,$9A
	.dc.b	$FD,$7F,$CF,$FF,$80,$EF,$F4,$35,$A7,$A4,$E0,$DF,$8D,$FE,$F4,$BE
	.dc.b	$53,$F9,$DA,$E0,$FA,$AB,$6F,$13,$4D,$C3,$F6,$58,$78,$67,$24,$8F
	.dc.b	$44,$F2,$8C,$5F,$82,$73,$1F,$C0,$6B,$94,$60,$93,$B3,$E8,$46,$9C
	.dc.b	$4F,$B8,$EA,$7D,$78,$D3,$99,$E6,$D9,$E7,$87,$C3,$8C,$10,$F8,$47
	.dc.b	$95,$D3,$89,$E6,$D9,$E6,$FC,$57,$CA,$E4,$F2,$4F,$2C,$7F,$05,$1E
	.dc.b	$83,$42,$FD,$2E,$23,$77,$E0,$9C,$C6,$2F,$BB,$6B,$B9,$E3,$3D,$8E
	.dc.b	$5E,$D7,$E3,$C7,$F0,$33,$93,$57,$DA,$32,$4C,$7E,$E6,$3F,$CD,$B9
	.dc.b	$1F,$3F,$FE,$DD,$CB,$FB,$23,$5F,$F6,$94,$6B,$D4,$FD,$57,$5F,$F9
	.dc.b	$C3,$A9,$EE,$63,$93,$FB,$BE,$3C,$6B,$FD,$07,$87,$1C,$7A,$E3,$B7
	.dc.b	$FB,$D6,$16,$7C,$7E,$7D,$C3,$FD,$1D,$5E,$48,$E1,$E5,$3E,$0B,$33
	.dc.b	$E6,$AB,$97,$81,$B9,$F9,$30,$E3,$78,$B6,$3F,$0B,$8B,$C0,$48,$FF
	.dc.b	$27,$FC,$15,$7B,$45,$CB,$E1,$4F,$AF,$CD,$FF,$4F,$FC,$07,$6F,$46
	.dc.b	$61,$FE,$58,$8D,$70,$9C,$C7,$FA,$77,$35,$6A,$7E,$37,$98,$F9,$F7
	.dc.b	$C1,$61,$7C,$FD,$31,$13,$1F,$F2,$D4,$E9,$B3,$95,$AF,$FA,$0F,$07
	.dc.b	$C8,$FD,$59,$F3,$36,$33,$1F,$EA,$0F,$9F,$FE,$4D,$5C,$6B,$44,$9D
	.dc.b	$74,$D7,$BF,$27,$F3,$3C,$39,$C6,$0F,$E7,$0E,$2F,$FD,$2D,$87,$6A
	.dc.b	$FF,$70,$FC,$FB,$E7,$FE,$2B,$FE,$F0,$C3,$FC,$F0,$D7,$FD,$C4,$ED
	.dc.b	$CC,$70,$F6,$A4,$60,$FF,$54,$F0,$FF,$25,$C7,$FD,$B7,$D3,$F8,$FE
	.dc.b	$6F,$C4,$9F,$6B,$E7,$ED,$F0,$8A,$FF,$B9,$A3,$E5,$7C,$FB,$E7,$E9
	.dc.b	$27,$FD,$93,$D2,$F9,$3C,$47,$C3,$15,$C5,$FF,$67,$B5,$D4,$F1,$73
	.dc.b	$5C,$B2,$FF,$5F,$56,$DF,$F6,$05,$6E,$BF,$21,$EC,$7E,$90,$8F,$FA
	.dc.b	$BB,$C5,$AE,$8F,$8B,$1F,$9A,$8F,$CD,$FC,$FF,$EB,$3D,$99,$5F,$EB
	.dc.b	$68,$CB,$F3,$23,$2F,$DF,$EE,$77,$2B,$F2,$1F,$3F,$FE,$CF,$AF,$F7
	.dc.b	$6B,$5F,$88,$DC,$FF,$D7,$5F,$3F,$FE,$B2,$9D,$FB,$DC,$88,$FF,$AE
	.dc.b	$E3,$4F,$28,$CB,$4D,$7A,$FF,$C1,$88,$FE,$11,$FD,$CF,$C9,$E2,$3F
	.dc.b	$71,$16,$FF,$92,$4F,$EF,$BB,$7F,$87,$FA,$D2,$FF,$DB,$31,$97,$F1
	.dc.b	$F1,$97,$E2,$69,$F9,$F7,$FC,$35,$F3,$CF,$9E,$FD,$A3,$63,$FC,$3D
	.dc.b	$8F,$FB,$EE,$31,$1F,$AB,$F1,$5F,$3E,$F7,$38,$E1,$2B,$8F,$83,$FE
	.dc.b	$25,$C3,$8E,$3F,$EE,$5C,$38,$56,$8F,$1D,$F1,$7F,$2D,$5C,$2F,$FC
	.dc.b	$61,$8F,$83,$1F,$E7,$5A,$4B,$FE,$9A,$FF,$3F,$FA,$17,$D1,$7F,$0F
	.dc.b	$F9,$FF,$F5,$0D,$77,$0E,$1F,$D2,$37,$5F,$E9,$BF,$42,$FA,$1F,$E9
	.dc.b	$4D,$AF,$F5,$F3,$5F,$EB,$6D,$96,$9F,$E4,$7E,$87,$CB,$66,$3F,$DA
	.dc.b	$34,$FA,$1E,$A4,$C3,$CD,$93,$D6,$F5,$12,$32,$FF,$45,$C3,$94,$7C
	.dc.b	$BC,$7B,$DE,$E3,$DB,$5B,$8F,$89,$E6,$7F,$33,$FA,$17,$C6,$D3,$37
	.dc.b	$E2,$9A,$FC,$67,$87,$5E,$36,$F7,$CF,$F9,$97,$4F,$E5,$DB,$7F,$99
	.dc.b	$5B,$1F,$E6,$9C,$7F,$5D,$39,$CB,$C5,$F1,$72,$BE,$0C,$7D,$92,$38
	.dc.b	$3E,$D5,$3E,$13,$31,$F6,$A6,$E3,$E4,$85,$6D,$FF,$7B,$7D,$0B,$E1
	.dc.b	$FF,$C0,$CD,$8C,$C7,$F7,$1F,$8F,$65,$A8,$93,$A3,$E8,$4E,$9C,$1F
	.dc.b	$E1,$EA,$E8,$D9,$E9,$5F,$E1,$5D,$5F,$29,$49,$5D,$2A,$BF,$E2,$89
	.dc.b	$DB,$FA,$9F,$65,$18,$8F,$E4,$58,$F8,$33,$FF,$13,$CE,$5B,$87,$47
	.dc.b	$EC,$F1,$AB,$86,$FF,$E9,$FC,$5F,$3D,$FC,$F6,$FF,$C6,$2D,$9D,$37
	.dc.b	$26,$0F,$A1,$7C,$BF,$F8,$CE,$B5,$8F,$4F,$3F,$07,$F8,$7E,$B7,$87
	.dc.b	$B6,$B6,$49,$D5,$FF,$8F,$7E,$85,$F3,$94,$97,$FA,$3C,$1F,$2B,$88
	.dc.b	$FF,$28,$D7,$07,$E5,$35,$FE,$95,$3C,$1F,$E1,$18,$BE,$2F,$FC,$3B
	.dc.b	$1F,$E4,$7F,$8C,$FF,$61,$7D,$0B,$D3,$FF,$62,$7B,$29,$CC,$61,$FF
	.dc.b	$64,$61,$CF,$8B,$FD,$99,$5B,$C3,$FD,$9D,$39,$7B,$1F,$ED,$1C,$3B
	.dc.b	$3C,$2F,$AB,$4E,$38,$F9,$AF,$A4,$7D,$A8,$C4,$F0,$8E,$A2,$BA,$B8
	.dc.b	$BF,$15,$F2,$6D,$71,$8E,$B5,$C1,$D5,$F6,$BF,$1E,$1E,$57,$B6,$3F
	.dc.b	$0D,$F6,$35,$F3,$46,$F1,$EF,$1A,$39,$61,$DF,$1F,$97,$87,$79,$E2
	.dc.b	$7B,$27,$4F,$13,$D9,$1B,$AE,$B1,$E4,$3A,$E1,$E3,$8C,$9A,$AE,$73
	.dc.b	$B3,$9E,$1E,$7B,$FC,$2C,$0F,$2C,$7F,$BF,$9F,$29,$93,$E2,$94,$FC
	.dc.b	$0F,$98,$FA,$1E,$18,$23,$13,$B3,$F6,$AE,$1E,$51,$B9,$F0,$47,$8F
	.dc.b	$E2,$2A,$69,$1D,$BF,$5F,$E7,$5F,$43,$FC,$FB,$6F,$C5,$D7,$33,$F1
	.dc.b	$33,$AE,$7F,$BD,$19,$9F,$C8,$CE,$63,$F1,$E6,$77,$F8,$51,$99,$EB
	.dc.b	$3E,$53,$AD,$F8,$6B,$E0,$61,$E5,$8C,$4F,$C9,$8C,$61,$E2,$8E,$57
	.dc.b	$E2,$AD,$C7,$D1,$AD,$BE,$43,$77,$E4,$9D,$C7,$9C,$FB,$98,$79,$E7
	.dc.b	$C7,$1E,$6C,$3C,$17,$D3,$0F,$5B,$6B,$71,$E3,$9D,$D7,$13,$C5,$5B
	.dc.b	$7D,$86,$67,$F8,$37,$33,$E0,$7F,$C1,$F7,$FE,$2B,$F7,$BF,$C2,$3C
	.dc.b	$FF,$DD,$38,$3F,$DC,$F5,$AC,$7F,$D4,$95,$AB,$FE,$95,$39,$AF,$F4
	.dc.b	$AC,$7F,$38,$32,$73,$3F,$39,$3C,$6B,$FD,$F7,$FA,$37,$19,$CF,$D7
	.dc.b	$71,$9C,$C7,$15,$49,$D5,$FF,$BF,$2F,$95,$6A,$73,$7F,$D4,$E7,$37
	.dc.b	$C7,$1E,$99,$27,$37,$C3,$0F,$F7,$9C,$62,$B7,$F4,$5F,$D1,$31,$FF
	.dc.b	$29,$64,$F6,$5E,$BE,$BA,$FD,$13,$98,$C5,$70,$AD,$BF,$DC,$EF,$A6
	.dc.b	$5F,$ED,$1B,$E1,$39,$C3,$9D,$F0,$AF,$F6,$B6,$8C,$38,$E1,$F6,$6B
	.dc.b	$4D,$F8,$5E,$3F,$19,$C9,$B9,$78,$DB,$8F,$86,$D1,$8B,$E5,$5E,$9B
	.dc.b	$E4,$D3,$FC,$18,$FF,$6C,$6F,$77,$D9,$F0,$3F,$CB,$7E,$A7,$D0,$AD
	.dc.b	$65,$EA,$6E,$1F,$ED,$EE,$2F,$B7,$AA,$73,$97,$E2,$A7,$31,$8B,$FC
	.dc.b	$2C,$63,$2F,$C4,$3F,$88,$AF,$BF,$F6,$7F,$E9,$97,$FC,$44,$DA,$FF
	.dc.b	$61,$7A,$0F,$B3,$DF,$B3,$2F,$F4,$DF,$58,$F9,$3C,$61,$FA,$48,$C4
	.dc.b	$F4,$AF,$0C,$62,$7E,$BE,$3D,$F6,$FF,$7D,$CE,$5F,$F3,$1E,$DE,$FF
	.dc.b	$97,$F8,$C7,$C7,$FA,$7B,$97,$CE,$8F,$C6,$FD,$2F,$F9,$4E,$30,$7E
	.dc.b	$51,$C6,$5F,$8B,$6E,$1D,$B8,$7F,$C8,$35,$BF,$3B,$EA,$79,$DF,$1B
	.dc.b	$E6,$FE,$3F,$AF,$6B,$E3,$B8,$7B,$45,$CB,$B3,$F3,$B3,$FA,$3E,$0F
	.dc.b	$D9,$AF,$9F,$18,$CB,$F7,$96,$F7,$7F,$D3,$95,$FE,$BD,$1F,$3B,$FD
	.dc.b	$07,$5F,$F3,$A4,$EA,$91,$FE,$8A,$AD,$F8,$FD,$2B,$53,$9B,$E5,$5F
	.dc.b	$DA,$67,$2F,$F6,$F3,$89,$FE,$BF,$73,$7F,$EB,$49,$E2,$FF,$75,$8C
	.dc.b	$CF,$08,$FC,$4F,$C6,$FE,$03,$AF,$F4,$1D,$28,$C5,$7F,$AB,$E3,$0F
	.dc.b	$E3,$27,$73,$AA,$FE,$5F,$F0,$0F,$90,$FC,$DB,$6F,$FD,$13,$18,$23
	.dc.b	$FC,$EE,$35,$3F,$E7,$C9,$CB,$FC,$37,$F3,$27,$CB,$FF,$2D,$A7,$E5
	.dc.b	$FF,$5E,$B5,$39,$8C,$6D,$FD,$16,$F8,$E3,$FE,$F3,$C3,$8F,$07,$F2
	.dc.b	$09,$CE,$3F,$F0,$07,$0E,$37,$FE,$1D,$C1,$5A,$7F,$D8,$78,$72,$9F
	.dc.b	$E5,$B3,$97,$FD,$F7,$5C,$AD,$4F,$BD,$FE,$43,$5F,$F0,$01,$FE,$75
	.dc.b	$19,$F1,$BD,$D1,$83,$75,$C9,$C4,$6B,$C8,$FD,$CF,$AF,$30,$FB,$1F
	.dc.b	$27,$70,$FE,$01,$37,$4F,$F5,$1F,$12,$B1,$7F,$E3,$7F,$CE,$7F,$62
	.dc.b	$FE,$3F,$96,$E3,$FE,$3F,$9B,$3F,$A0,$EC,$FD,$2C,$62,$B7,$48,$C4
	.dc.b	$E8,$FE,$CD,$5C,$4F,$9C,$78,$5C,$D7,$83,$C5,$78,$3F,$D9,$0F,$ED
	.dc.b	$3E,$3B,$37,$C1,$7F,$1A,$C6,$1F,$C8,$8F,$CE,$B8,$CF,$F6,$71,$FE
	.dc.b	$59,$E2,$7D,$53,$C0,$FC,$72,$F0,$2C,$61,$B5,$F8,$B9,$D3,$67,$4F
	.dc.b	$B2,$FE,$47,$93,$F9,$9F,$96,$7D,$97,$C8,$8C,$36,$BF,$2D,$3A,$79
	.dc.b	$57,$CB,$AD,$C6,$38,$BF,$A6,$66,$BE,$17,$EE,$3C,$35,$D3,$0F,$EB
	.dc.b	$DE,$F3,$E6,$7F,$6E,$60,$B3,$FF,$00,$1B,$5F,$EF,$06,$F3,$7E,$2B
	.dc.b	$F0,$CF,$EB,$3E,$93,$F0,$AF,$E1,$71,$86,$E1,$FE,$39,$9D,$3F,$C3
	.dc.b	$67,$4F,$FB,$FF,$F4,$2F,$EA,$0F,$F8,$CA,$7F,$B4,$79,$A9,$CE,$5F
	.dc.b	$E2,$E9,$CC,$63,$A3,$FC,$3B,$C3,$FE,$28,$8C,$4F,$2F,$3B,$F9,$55
	.dc.b	$FF,$91,$BC,$BF,$E7,$99,$7F,$6E,$8F,$E8,$7E,$4F,$F7,$26,$C6,$4F
	.dc.b	$F2,$94,$F0,$8F,$F0,$EE,$0F,$2F,$E1,$F0,$F8,$5B,$7E,$D6,$D7,$06
	.dc.b	$F0,$27,$FD,$51,$F4,$3F,$DE,$7E,$8F,$FD,$39,$F3,$18,$F4,$D1,$8C
	.dc.b	$7F,$D6,$5F,$46,$FA,$4E,$51,$FE,$A1,$BE,$15,$B8,$61,$D5,$D1,$FE
	.dc.b	$BB,$71,$EE,$AB,$FD,$75,$87,$5C,$B9,$56,$A7,$AD,$F2,$3A,$E5,$E6
	.dc.b	$71,$7C,$5D,$57,$17,$C1,$3B,$C3,$C1,$18,$C7,$D1,$7C,$67,$33,$DA
	.dc.b	$7C,$75,$C7,$0F,$1D,$6A,$F8,$46,$09,$7D,$4F,$2E,$5F,$22,$BA,$E3
	.dc.b	$E5,$9E,$C7,$A3,$0E,$CF,$D0,$AE,$17,$D2,$37,$7C,$AB,$70,$C3,$C6
	.dc.b	$F8,$67,$58,$FA,$8F,$7E,$1D,$27,$58,$75,$C7,$C9,$87,$29,$EF,$39
	.dc.b	$C7,$BC,$E6,$BD,$44,$8C,$DF,$5B,$E2,$76,$21,$E7,$75,$3C,$70,$F3
	.dc.b	$4F,$C8,$CA,$E5,$EA,$8F,$8D,$3E,$A9,$CE,$5E,$58,$C4,$E8,$95,$C4
	.dc.b	$FB,$18,$78,$31,$F9,$38,$F8,$6B,$C0,$70,$C3,$F2,$28,$A4,$3E,$F1
	.dc.b	$93,$F3,$DF,$1F,$FE,$3B,$70,$63,$86,$B4,$B4,$9E,$11,$C2,$31,$3F
	.dc.b	$A4,$9E,$98,$FE,$29,$B7,$E1,$FA,$47,$D0,$7F,$BD,$63,$F7,$F3,$AC
	.dc.b	$7C,$17,$C7,$0F,$DC,$61,$C2,$7B,$44,$7F,$57,$5C,$54,$ED,$CC,$EE
	.dc.b	$BA,$47,$ED,$E7,$51,$82,$5F,$92,$BC,$38,$79,$EF,$CB,$39,$7A,$9C
	.dc.b	$E3,$4F,$F0,$EC,$3C,$98,$FC,$68,$DE,$1F,$BA,$C7,$C3,$87,$38,$F0
	.dc.b	$CF,$63,$EE,$4E,$9C,$1D,$71,$F6,$57,$CD,$C7,$C3,$87,$7C,$39,$DE
	.dc.b	$CF,$31,$C0,$E7,$1B,$AE,$51,$93,$84,$F2,$C3,$C5,$18,$9F,$41,$1F
	.dc.b	$3D,$F8,$B1,$F8,$24,$C7,$E5,$CF,$5C,$7C,$0F,$B3,$0E,$B1,$88,$EC
	.dc.b	$66,$FA,$57,$1B,$FC,$05,$F0,$7B,$57,$CE,$C3,$D6,$FF,$3E,$E8,$FE
	.dc.b	$D4,$FE,$8F,$83,$FA,$3F,$43,$9F,$FC,$09,$3F,$DD,$FA,$F5,$C5,$FA
	.dc.b	$DC,$38,$CF,$F7,$73,$27,$F7,$39,$B7,$CD,$CC,$F0,$9F,$F0,$A3,$FD
	.dc.b	$9F,$87,$7E,$33,$EF,$69,$C3,$9B,$FA,$E5,$FF,$8A,$EF,$83,$FE,$DE
	.dc.b	$9D,$B9,$9D,$BD,$BF,$42,$BA,$FB,$B5,$7D,$4F,$99,$B5,$D6,$B7,$B7
	.dc.b	$BD,$B1,$FE,$0F,$AE,$07,$B2,$33,$E4,$FF,$94,$FC,$FE,$9F,$2F,$2F
	.dc.b	$FD,$55,$1F,$1A,$75,$1F,$E2,$39,$D3,$FE,$FC,$9D,$76,$3D,$DF,$97
	.dc.b	$E3,$3D,$BA,$6A,$25,$FF,$AA,$E3,$2F,$F9,$47,$15,$F3,$8F,$F2,$5F
	.dc.b	$E0,$0F,$B9,$FB,$38,$FF,$92,$7D,$CB,$E5,$1F,$BB,$6B,$E0,$F2,$BD
	.dc.b	$CC,$8C,$10,$FF,$29,$5F,$57,$FC,$BB,$1D,$2B,$9C,$6B,$CD,$FF,$04
	.dc.b	$37,$DD,$F2,$C1,$BE,$24,$9F,$F3,$94,$68,$C3,$FE,$61,$EC,$FC,$E6
	.dc.b	$FF,$7C,$91,$FE,$DD,$CE,$BB,$E5,$F8,$CD,$9F,$B1,$E8,$F9,$B9,$FF
	.dc.b	$C5,$B3,$FE,$7A,$FA,$5F,$8B,$F2,$BF,$EE,$18,$FF,$3B,$B5,$FF,$39
	.dc.b	$1B,$3F,$CF,$7E,$7A,$78,$FB,$D2,$73,$1F,$42,$75,$1D,$27,$4F,$48
	.dc.b	$CB,$E9,$F6,$BF,$38,$47,$FD,$F5,$F3,$35,$D1,$FE,$73,$8F,$6C,$7E
	.dc.b	$0B,$F3,$2F,$E0,$33,$F6,$1F,$56,$5F,$EE,$06,$F7,$7F,$AA,$5D,$C6
	.dc.b	$7C,$4E,$59,$4B,$FF,$78,$56,$AF,$FD,$D9,$F3,$6C,$7F,$23,$EF,$3E
	.dc.b	$8A,$9F,$C3,$33,$56,$F1,$FF,$56,$46,$2F,$71,$FE,$FD,$BE,$4E,$AB
	.dc.b	$83,$FE,$A3,$AE,$0E,$EF,$8B,$99,$DC,$FF,$24,$EA,$7C,$6F,$EC,$89
	.dc.b	$5F,$AD,$ED,$FF,$7E,$3F,$EB,$FE,$7F,$F7,$FC,$EA,$3F,$25,$5C,$9E
	.dc.b	$1C,$1F,$8C,$FB,$BF,$E2,$5F,$0F,$27,$F7,$1C,$C9,$D4,$7F,$AA,$67
	.dc.b	$4F,$FA,$F7,$C0,$FF,$71,$4F,$ED,$BF,$1C,$FF,$6C,$4F,$E4,$BA,$7F
	.dc.b	$2A,$67,$DD,$91,$92,$1F,$96,$9E,$2D,$8E,$2E,$B7,$FD,$0C,$E9,$FF
	.dc.b	$47,$F9,$5F,$EA,$1F,$94,$FE,$77,$D5,$2F,$F9,$A4,$7F,$3B,$F8,$6F
	.dc.b	$B9,$B7,$F3,$09,$FA,$8D,$C3,$9F,$96,$AF,$E8,$5F,$2D,$C6,$FF,$9F
	.dc.b	$46,$1F,$C1,$DB,$DC,$FE,$9F,$F3,$7F,$C8,$27,$FA,$43,$5E,$F1,$AE
	.dc.b	$FF,$1B,$23,$71,$96,$90,$FE,$B3,$3E,$7E,$DF,$EC,$99,$7F,$A5,$33
	.dc.b	$FE,$9D,$C1,$FD,$34,$FE,$B5,$E9,$4C,$3F,$DC,$51,$97,$FA,$C4,$65
	.dc.b	$FE,$CB,$5C,$EF,$AC,$FF,$12,$DF,$DB,$F0,$9F,$D1,$3C,$79,$87,$BA
	.dc.b	$32,$FE,$4F,$A7,$F9,$5E,$7F,$9B,$AC,$CF,$AA,$AC,$FD,$AD,$11,$FD
	.dc.b	$4B,$E6,$FF,$65,$F4,$4F,$8E,$BF,$1B,$31,$FD,$9C,$E5,$B3,$97,$F6
	.dc.b	$10,$9E,$73,$27,$8C,$6F,$1F,$F6,$C8,$AE,$93,$FE,$08,$F4,$1F,$55
	.dc.b	$FC,$17,$EF,$3F,$C3,$13,$FC,$11,$AE,$E3,$FC,$23,$3C,$23,$5E,$67
	.dc.b	$13,$33,$23,$24,$39,$1F,$E1,$1F,$B9,$F9,$7C,$7F,$E8,$FF,$11,$F0
	.dc.b	$1F,$C7,$69,$18,$E8,$FF,$10,$DF,$09,$77,$23,$15,$B9,$FE,$15,$7C
	.dc.b	$EB,$5C,$9C,$8F,$F1,$6C,$77,$AE,$15,$FE,$2C,$BE,$D6,$F9,$EF,$E6
	.dc.b	$F3,$E4,$FB,$4F,$E2,$EF,$F6,$56,$F0,$FE,$EA,$77,$5C,$7C,$EF,$F1
	.dc.b	$DD,$6B,$5F,$80,$DC,$3D,$77,$D9,$7C,$E3,$C2,$65,$FE,$B5,$F8,$7F
	.dc.b	$E4,$54,$F9,$E7,$89,$FE,$B5,$9C,$C6,$27,$87,$97,$FE,$B9,$C7,$A6
	.dc.b	$18,$9F,$F6,$04,$61,$5F,$2C,$3A,$5F,$FB,$07,$0E,$95,$A8,$E9,$8F
	.dc.b	$9A,$30,$74,$71,$7C,$E3,$C0,$7A,$6F,$AB,$BC,$73,$8F,$8E,$B8,$4F
	.dc.b	$43,$81,$AB,$EB,$87,$8A,$FA,$E3,$E6,$BE,$27,$C8,$C3,$A5,$6F,$0E
	.dc.b	$95,$B8,$F7,$63,$E4,$9F,$A7,$8F,$67,$CC,$70,$8D,$57,$18,$CE,$1D
	.dc.b	$EB,$84,$7A,$0E,$07,$A7,$18,$E0,$93,$C6,$33,$1F,$EE,$EC,$7E,$B3
	.dc.b	$6B,$E8,$36,$7F,$DE,$2F,$E2,$BE,$79,$F3,$D8,$B4,$62,$7F,$DE,$F1
	.dc.b	$89,$D4,$E6,$79,$F1,$F4,$47,$EE,$0C,$39,$70,$70,$3E,$C7,$CF,$7F
	.dc.b	$94,$78,$3F,$03,$1C,$A7,$F0,$73,$98,$FB,$C6,$4E,$3E,$C2,$B8,$C7
	.dc.b	$87,$4F,$D1,$D7,$17,$F4,$B1,$C7,$A7,$37,$C6,$3F,$D6,$DC,$1E,$A6
	.dc.b	$63,$EC,$AD,$E3,$6B,$55,$E3,$C3,$2F,$98,$F1,$47,$3C,$7C,$4F,$08
	.dc.b	$C9,$F2,$23,$2F,$4C,$3E,$0E,$3E,$5B,$F4,$CF,$01,$5C,$1F,$39,$F1
	.dc.b	$30,$F8,$77,$C9,$B1,$93,$CD,$19,$69,$9A,$EF,$1E,$F8,$F3,$5F,$C7
	.dc.b	$C7,$31,$F2,$0C,$DE,$F9,$1F,$3D,$F5,$63,$E0,$46,$1F,$F3,$FC,$FF
	.dc.b	$B8,$67,$4F,$CC,$9D,$4F,$09,$CE,$FF,$EE,$C8,$7C,$B7,$F5,$C7,$91
	.dc.b	$6F,$FA,$3B,$A3,$FD,$BD,$F3,$BC,$9F,$F7,$83,$61,$E8,$7F,$AF,$3E
	.dc.b	$8B,$96,$4A,$D3,$6F,$D3,$7F,$D4,$27,$D2,$FE,$26,$75,$3F,$02,$73
	.dc.b	$3E,$08,$F3,$75,$BE,$7F,$CD,$3F,$EF,$AA,$F2,$4E,$72,$F1,$FC,$FB
	.dc.b	$E0,$E9,$EF,$F2,$9F,$9E,$3E,$63,$3A,$7F,$BF,$6F,$83,$F4,$9B,$9F
	.dc.b	$CF,$6B,$E9,$6B,$FB,$CD,$51,$C3,$83,$FC,$17,$F4,$0E,$3E,$F7,$FE
	.dc.b	$83,$8C,$61,$FE,$8C,$5F,$41,$F5,$6D,$18,$9F,$E1,$31,$FB,$3D,$3D
	.dc.b	$7F,$7B,$BA,$D6,$1D,$70,$FA,$37,$C9,$B3,$A9,$FD,$D4,$E7,$0F,$D8
	.dc.b	$36,$3C,$DE,$B7,$5B,$EC,$FE,$00,$FF,$30,$57,$90,$F8,$4E,$27,$FD
	.dc.b	$0D,$EB,$7F,$BE,$E7,$51,$FE,$12,$9D,$5F,$F9,$E9,$E1,$1F,$85,$6B
	.dc.b	$D3,$D6,$FC,$5E,$DF,$E7,$56,$CF,$2B,$FF,$12,$71,$FE,$4A,$31,$7F
	.dc.b	$E2,$48,$FC,$AF,$CB,$BD,$9F,$CC,$B6,$FF,$D4,$3E,$8E,$26,$47,$ED
	.dc.b	$27,$57,$FB,$16,$BF,$9B,$6B,$F8,$37,$79,$FF,$29,$C1,$F1,$CF,$95
	.dc.b	$EB,$7F,$E1,$6C,$3F,$C7,$93,$97,$E3,$EB,$FE,$93,$F6,$7E,$81,$F4
	.dc.b	$7D,$DB,$3A,$D9,$3F,$10,$91,$FC,$32,$32,$FD,$EF,$A1,$FF,$37,$B7
	.dc.b	$D8,$FC,$3C,$63,$2F,$E7,$3A,$A3,$11,$EF,$F7,$3E,$9E,$E6,$93,$0F
	.dc.b	$91,$F3,$9A,$23,$83,$E6,$67,$C8,$8E,$0F,$EE,$7F,$2A,$FA,$4F,$E3
	.dc.b	$18,$A3,$13,$F8,$C1,$F3,$CF,$5B,$73,$AA,$F9,$C7,$F7,$27,$0F,$F5
	.dc.b	$66,$BF,$D7,$5A,$EE,$FF,$8D,$BE,$17,$F9,$55,$F8,$3C,$8E,$F1,$8C
	.dc.b	$3F,$BA,$6D,$FD,$A7,$C7,$75,$7F,$A4,$F0,$FF,$7C,$72,$FE,$97,$2F
	.dc.b	$D3,$FB,$9F,$C8,$23,$18,$71,$BE,$15,$EA,$E8,$FC,$5F,$C7,$FF,$3B
	.dc.b	$CB,$AE,$3C,$63,$F9,$4F,$80,$F5,$E9,$BF,$85,$C6,$BF,$EB,$78,$CB
	.dc.b	$4F,$DB,$E0,$9D,$E5,$E1,$BB,$A7,$FA,$A3,$E8,$5F,$45,$FC,$65,$B8
	.dc.b	$FF,$00,$A7,$C3,$BE,$3F,$57,$2B,$F8,$8D,$F0,$C7,$F2,$2D,$7F,$8B
	.dc.b	$B5,$FE,$46,$EF,$1F,$E4,$9F,$17,$FE,$F9,$AF,$E1,$F0,$9D,$47,$F8
	.dc.b	$CB,$25,$6E,$FF,$0B,$1E,$6F,$33,$FA,$46,$7F,$D4,$31,$FE,$60,$FF
	.dc.b	$3D,$C3,$93,$96,$CF,$0F,$37,$8B,$63,$EF,$FD,$17,$DD,$3E,$C7,$23
	.dc.b	$E8,$B1,$3A,$26,$7F,$CA,$A3,$2F,$E6,$87,$D1,$7E,$DF,$3F,$F1,$B4
	.dc.b	$63,$C2,$FC,$97,$A7,$FA,$8A,$D6,$1F,$E3,$97,$F7,$75,$6B,$F3,$B1
	.dc.b	$88,$FF,$2B,$B5,$E9,$3F,$AF,$EA,$E7,$1A,$AF,$D0,$1F,$99,$98,$7E
	.dc.b	$2B,$2F,$EE,$0F,$E8,$0C,$CF,$E6,$FA,$BF,$C1,$DC,$DF,$A6,$F4,$33
	.dc.b	$A7,$F5,$FF,$93,$78,$5F,$DE,$DF,$E2,$18,$FE,$35,$CB,$7C,$3D,$4E
	.dc.b	$DB,$19,$3F,$A9,$E5,$F2,$BC,$CF,$8C,$75,$F9,$8B,$3A,$E0,$FE,$A7
	.dc.b	$5F,$12,$BE,$39,$27,$FD,$13,$3F,$D4,$39,$DF,$2C,$F8,$CE,$78,$7E
	.dc.b	$3A,$31,$7F,$AE,$EE,$7C,$37,$EE,$0F,$DB,$16,$34,$DA,$FE,$37,$C3
	.dc.b	$FB,$58,$C7,$07,$F0,$7F,$A1,$FF,$14,$FA,$AF,$A2,$FC,$7C,$61,$B8
	.dc.b	$FF,$74,$F1,$BF,$C2,$D1,$8C,$BF,$87,$5F,$1A,$FE,$A7,$F4,$5F,$D7
	.dc.b	$A5,$E8,$A3,$0D,$F4,$FF,$BB,$C6,$0F,$EC,$75,$FE,$A2,$C7,$FC,$49
	.dc.b	$C7,$FE,$4A,$9C,$E0,$9C,$D7,$F6,$2F,$3F,$FD,$52,$D9,$F6,$DF,$F7
	.dc.b	$A6,$47,$F7,$77,$FA,$A9,$1F,$F3,$6D,$FF,$62,$8F,$F3,$53,$5F,$F2
	.dc.b	$A3,$5F,$F1,$E3,$5F,$7B,$C7,$5F,$EF,$99,$FF,$25,$AF,$8B,$E7,$FD
	.dc.b	$0E,$1F,$EA,$FF,$43,$FD,$FF,$D0,$FF,$5C,$7C,$8F,$FA,$A6,$31,$9B
	.dc.b	$DF,$4F,$FA,$B7,$E8,$9F,$3F,$FE,$B1,$9C,$D7,$FB,$2A,$F8,$C6,$21
	.dc.b	$18,$BE,$DE,$C7,$F6,$39,$ED,$1D,$6F,$A3,$AC,$7C,$B3,$D6,$B8,$61
	.dc.b	$D7,$0E,$93,$FE,$DA,$C7,$CB,$A7,$08,$F0,$DF,$0C,$7A,$47,$80,$E9
	.dc.b	$8E,$EF,$8C,$64,$E3,$1C,$F0,$EF,$18,$35,$3C,$31,$F2,$61,$DE,$7A
	.dc.b	$9C,$EB,$8D,$6F,$0F,$1D,$F0,$C3,$B6,$1C,$70,$5E,$DE,$75,$B7,$A4
	.dc.b	$EE,$75,$1E,$59,$D1,$2F,$E7,$63,$DB,$09,$5C,$9F,$85,$97,$D2,$8E
	.dc.b	$F5,$E6,$23,$F3,$E3,$4F,$78,$CB,$4D,$57,$6B,$F1,$47,$43,$27,$84
	.dc.b	$8E,$9F,$83,$5D,$B2,$FA,$24,$CB,$E0,$46,$5F,$91,$19,$C7,$D7,$87
	.dc.b	$87,$1F,$2C,$FD,$2B,$E1,$59,$C3,$C8,$78,$2F,$0F,$DF,$07,$E2,$8A
	.dc.b	$7D,$E2,$9B,$3F,$59,$F4,$1F,$E4,$57,$27,$E9,$1C,$3C,$2D,$18,$8E
	.dc.b	$3E,$B7,$07,$83,$F9,$38,$E9,$9F,$E8,$DB,$97,$93,$E8,$5F,$45,$E6
	.dc.b	$AD,$5F,$12,$61,$F9,$FA,$DB,$89,$DB,$F9,$9C,$3A,$BF,$0A,$76,$E6
	.dc.b	$B8,$61,$E8,$8E,$98,$78,$63,$18,$74,$23,$E7,$8D,$4F,$F0,$BC,$3B
	.dc.b	$BF,$00,$EC,$74,$3B,$9B,$BA,$B0,$F2,$E1,$DA,$34,$72,$7F,$8B,$46
	.dc.b	$B0,$F2,$92,$77,$3A,$C3,$CB,$1E,$38,$E8,$D9,$F5,$3E,$2C,$7E,$DE
	.dc.b	$1E,$4C,$3B,$4F,$4A,$EC,$F1,$3E,$66,$1D,$AF,$8C,$79,$E7,$75,$CA
	.dc.b	$38,$12,$77,$87,$9E,$3C,$58,$FC,$A9,$CD,$F9,$A7,$A1,$E3,$21,$F2
	.dc.b	$70,$F1,$57,$12,$63,$FC,$B7,$1F,$16,$1D,$2F,$8B,$5E,$47,$F3,$AE
	.dc.b	$8F,$EE,$8F,$FB,$FB,$3F,$EE,$1C,$5F,$E0,$E8,$FF,$00,$35,$FF,$52
	.dc.b	$CF,$07,$FB,$F9,$FA,$37,$0F,$F6,$83,$1C,$9F,$A7,$7F,$C0,$7C,$7F
	.dc.b	$D6,$1F,$D8,$CF,$FB,$53,$83,$94,$E6,$7F,$73,$39,$77,$19,$E4,$63
	.dc.b	$FC,$0A,$B5,$5E,$6A,$D4,$FB,$3C,$1F,$74,$FF,$85,$3E,$87,$2F,$F8
	.dc.b	$2F,$34,$EB,$A5,$18,$9F,$F7,$7D,$70,$C4,$F9,$E3,$F8,$8C,$63,$3F
	.dc.b	$89,$39,$9F,$33,$57,$77,$FB,$03,$CE,$F0,$60,$FA,$1A,$7A,$27,$55
	.dc.b	$FE,$67,$AC,$BF,$5A,$75,$E5,$27,$87,$5B,$BB,$D9,$E2,$3E,$52,$9A
	.dc.b	$89,$8F,$F8,$FE,$74,$D9,$CB,$FE,$8D,$FA,$1F,$F5,$B3,$FE,$B9,$F1
	.dc.b	$5E,$CF,$CE,$BF,$F2,$EB,$F4,$25,$5C,$1E,$5E,$C7,$E1,$63,$11,$FC
	.dc.b	$E1,$AF,$F3,$C6,$BF,$D0,$1E,$5C,$FF,$E9,$C6,$F3,$6E,$BC,$5E,$B7
	.dc.b	$FB,$42,$FF,$D0,$5E,$0F,$97,$CD,$7B,$94,$C4,$49,$FF,$44,$46,$5E
	.dc.b	$27,$E9,$37,$FA,$E4,$7F,$D3,$58,$57,$FC,$CB,$17,$3F,$F7,$B4,$FC
	.dc.b	$3F,$2D,$F4,$1F,$17,$E8,$7F,$DE,$58,$7F,$A4,$1A,$FF,$AE,$E7,$6F
	.dc.b	$FA,$51,$AF,$F5,$A6,$BF,$D8,$1A,$FF,$64,$6B,$FD,$A1,$E3,$7F,$EC
	.dc.b	$46,$EB,$BB,$F2,$76,$FE,$53,$D9,$98,$7F,$04,$7F,$1F,$F4,$18,$8F
	.dc.b	$F4,$1C,$5C,$7F,$6B,$1F,$9E,$E8,$57,$E1,$7E,$87,$F8,$65,$6B,$95
	.dc.b	$1F,$02,$30,$DC,$FF,$DE,$8F,$06,$FA,$1C,$FD,$E9,$97,$F8,$12,$F8
	.dc.b	$CF,$F0,$05,$F4,$1E,$29,$D5,$FF,$BF,$F9,$9F,$43,$F8,$83,$85,$6E
	.dc.b	$97,$EC,$39,$E9,$FE,$03,$6C,$7F,$83,$1A,$FF,$84,$9A,$FF,$85,$9C
	.dc.b	$9C,$E7,$B3,$67,$E3,$7D,$FF,$EE,$3E,$4A,$5F,$ED,$5F,$DE,$F0,$62
	.dc.b	$3C,$B9,$FF,$AA,$77,$56,$71,$FA,$7E,$0B,$BD,$8F,$F0,$8F,$A2,$FE
	.dc.b	$23,$F1,$13,$3F,$F7,$A4,$65,$FE,$87,$19,$7F,$D9,$BC,$89,$FD,$F7
	.dc.b	$D1,$3D,$AF,$CE,$76,$4B,$FF,$75,$CE,$5B,$3F,$D1,$63,$11,$FE,$34
	.dc.b	$6B,$FE,$3A,$6B,$FE,$3E,$77,$A7,$E3,$A1,$18,$DF,$55,$FA,$DC,$BF
	.dc.b	$A0,$F9,$2F,$2D,$F3,$1F,$D4,$09,$8F,$F4,$E7,$FA,$9F,$C5,$48,$FE
	.dc.b	$A2,$D7,$FA,$AC,$BE,$6B,$FA,$C9,$23,$FA,$BB,$F8,$5D,$DE,$CD,$24
	.dc.b	$FE,$1E,$32,$FF,$64,$FA,$2F,$EE,$0F,$F4,$C6,$BF,$D7,$DA,$FF,$BB
	.dc.b	$9B,$91,$FD,$34,$67,$B3,$E0,$C6,$27,$5D,$FF,$13,$2F,$F6,$37,$0F
	.dc.b	$FB,$63,$7F,$EA,$66,$7D,$79,$97,$F7,$27,$FB,$B9,$1C,$1F,$AC,$3F
	.dc.b	$BF,$BC,$1F,$C3,$F8,$5E,$2F,$82,$FE,$A7,$E3,$C9,$E1,$9F,$A6,$75
	.dc.b	$1F,$0D,$FF,$07,$68,$69,$23,$8C,$65,$B1,$FD,$8B,$E8,$BF,$81,$7D
	.dc.b	$13,$E3,$EF,$C6,$CB,$F4,$6C,$AD,$46,$1A,$FE,$ED,$AF,$EE,$DA,$FF
	.dc.b	$11,$6B,$FC,$21,$E3,$1F,$C2,$5B,$D1,$BB,$FE,$D1,$F9,$FF,$F8,$06
	.dc.b	$FE,$1F,$A2,$FA,$2F,$E3,$5F,$47,$FE,$14,$9F,$3B,$5F,$F1,$AC,$ED
	.dc.b	$E3,$3F,$E3,$58,$C4,$7F,$8B,$1A,$FF,$86,$1A,$FF,$8A,$DD,$E5,$FC
	.dc.b	$EB,$7F,$F0,$45,$7E,$AA,$73,$97,$FA,$CB,$C4,$65,$FC,$DF,$16,$0F
	.dc.b	$A3,$71,$7F,$8B,$7E,$07,$F2,$BE,$D9,$FC,$A4,$EA,$FF,$06,$D7,$FC
	.dc.b	$86,$D7,$FC,$5E,$D7,$FC,$84,$D7,$C2,$F1,$D7,$D4,$DC,$F7,$87,$F8
	.dc.b	$B7,$EF,$1F,$3F,$FC,$1B,$E8,$FD,$59,$BE,$93,$F9,$3E,$7F,$9D,$8C
	.dc.b	$79,$1B,$F4,$64,$FF,$97,$2F,$FB,$CB,$FE,$44,$6B,$FE,$57,$6B,$FE
	.dc.b	$DA,$FA,$3F,$E2,$7B,$7F,$B7,$3D,$EF,$E7,$33,$FE,$DE,$FA,$4F,$E1
	.dc.b	$CE,$7C,$FF,$F6,$FC,$62,$3F,$CA,$F1,$FE,$E2,$BE,$71,$83,$B7,$93
	.dc.b	$FD,$69,$7D,$17,$65,$7C,$E3,$0D,$BE,$95,$C6,$78,$5F,$67,$7E,$4F
	.dc.b	$FB,$C9,$B8,$7F,$BD,$9D,$D7,$1B,$E7,$96,$5F,$5D,$F1,$6D,$E4,$F4
	.dc.b	$CD,$58,$78,$67,$71,$EC,$6E,$1E,$0A,$E1,$5B,$AF,$99,$7D,$DA,$6B
	.dc.b	$0F,$33,$E0,$3E,$51,$F1,$A0,$FD,$33,$89,$A9,$FA,$39,$78,$63,$C8
	.dc.b	$6B,$0F,$8A,$E0,$F5,$DF,$81,$B8,$7B,$F2,$FA,$51,$C7,$2F,$92,$F1
	.dc.b	$C7,$EF,$BE,$43,$D0,$53,$99,$4A,$65,$FD,$0F,$D2,$66,$3E,$6F,$0A
	.dc.b	$31,$3A,$9E,$4D,$78,$73,$7F,$AF,$DF,$DE,$1F,$EC,$17,$04,$7E,$07
	.dc.b	$D2,$7F,$9C,$B9,$BE,$E4,$7C,$C9,$FB,$D3,$98,$F8,$C6,$4E,$3E,$E7
	.dc.b	$F9,$3B,$6F,$CD,$46,$5F,$CE,$4F,$17,$31,$E4,$73,$3C,$73,$F8,$97
	.dc.b	$E0,$8C,$57,$C9,$8C,$61,$E6,$8E,$55,$F3,$F0,$EC,$F2,$8F,$2C,$EE
	.dc.b	$B8,$C7,$A8,$E2,$7E,$2A,$B8,$BD,$CE,$26,$E3,$8E,$1E,$A6,$DF,$78
	.dc.b	$F5,$DF,$7A,$E3,$1B,$6C,$EC,$F4,$61,$EE,$E9,$7D,$27,$FB,$B3,$0F
	.dc.b	$EE,$6D,$BF,$F2,$69,$FE,$D8,$E8,$79,$DF,$E6,$3A,$FF,$79,$56,$F8
	.dc.b	$7F,$A4,$36,$32,$6E,$F9,$55,$C3,$C3,$37,$E8,$F8,$57,$F4,$A6,$FA
	.dc.b	$BF,$E5,$9F,$97,$D5,$43,$FC,$13,$3A,$F5,$93,$9B,$F4,$E0,$AE,$AD
	.dc.b	$BF,$26,$1D,$5E,$67,$97,$E8,$BF,$DD,$8F,$F8,$59,$B7,$F1,$3C,$27
	.dc.b	$D1,$BE,$5F,$87,$C3,$CC,$38,$3F,$06,$B7,$39,$BF,$2E,$1D,$38,$3F
	.dc.b	$B8,$37,$0F,$ED,$D8,$F5,$79,$9F,$E1,$8F,$A2,$7D,$17,$CD,$B4,$62
	.dc.b	$7F,$C6,$99,$79,$BE,$8F,$53,$9C,$BE,$ED,$F0,$C3,$27,$9F,$15,$E3
	.dc.b	$1F,$04,$F2,$E2,$C6,$1C,$B1,$FC,$2D,$7F,$8F,$27,$FC,$C2,$DF,$67
	.dc.b	$E7,$FB,$2F,$89,$FC,$15,$6B,$0F,$AE,$DD,$3F,$C2,$4D,$C3,$F9,$8F
	.dc.b	$0B,$0E,$51,$D4,$97,$F9,$8F,$1C,$7E,$02,$F9,$46,$23,$F0,$7F,$2C
	.dc.b	$F7,$BF,$28,$DB,$FF,$2E,$63,$DF,$D1,$CD,$49,$FF,$2C,$D6,$F0,$FF
	.dc.b	$52,$D6,$5F,$C8,$C6,$73,$FF,$72,$E2,$F9,$B7,$CB,$6E,$73,$18,$C7
	.dc.b	$E5,$7B,$FF,$1B,$DD,$F8,$ED,$8F,$9D,$17,$2F,$93,$36,$FF,$DB,$1F
	.dc.b	$47,$FE,$69,$BF,$F2,$FB,$6B,$E7,$DF,$E2,$63,$15,$FE,$41,$CF,$FD
	.dc.b	$79,$7E,$EF,$7B,$EC,$F3,$7C,$67,$E3,$FC,$E4,$9D,$1F,$11,$F9,$78
	.dc.b	$C8,$F0,$CD,$CB,$F8,$68,$F9,$67,$D1,$FF,$07,$8C,$3F,$C2,$69,$F3
	.dc.b	$DF,$E4,$C7,$FD,$21,$3A,$6D,$72,$6D,$71,$AE,$15,$E0,$AD,$36,$7C
	.dc.b	$11,$8D,$17,$FE,$00,$3C,$3E,$B6,$AB,$F3,$7E,$EF,$F1,$DE,$3F,$B9
	.dc.b	$D3,$8B,$FF,$24,$71,$FE,$5B,$3F,$F7,$4D,$EE,$31,$97,$FB,$DB,$A5
	.dc.b	$39,$C3,$A7,$73,$D5,$A7,$F1,$8F,$1B,$19,$FF,$95,$66,$E7,$FA,$16
	.dc.b	$E5,$F9,$06,$CF,$EA,$AA,$D7,$E9,$A6,$D3,$E8,$9F,$47,$FC,$35,$B8
	.dc.b	$7F,$0C,$F6,$1F,$23,$9A,$93,$FE,$AC,$AD,$E3,$FC,$9E,$F7,$39,$AD
	.dc.b	$CF,$F1,$2A,$DE,$1F,$42,$18,$B0,$E3,$1B,$BF,$E5,$91,$FC,$9E,$36
	.dc.b	$BE,$8F,$3E,$46,$9B,$97,$A5,$B5,$E5,$7E,$A7,$D1,$C9,$FE,$03,$C2
	.dc.b	$FA,$3C,$5F,$F0,$DE,$35,$E3,$5F,$B5,$57,$0F,$F1,$A7,$D1,$FE,$B1
	.dc.b	$BF,$0D,$F9,$3F,$3F,$F3,$D3,$9C,$BF,$C9,$AD,$4E,$6B,$F4,$33,$C2
	.dc.b	$A4,$7E,$8A,$75,$D7,$FD,$8A,$31,$3F,$8F,$61,$EC,$F3,$6E,$1E,$B8
	.dc.b	$B9,$7E,$4E,$6C,$7F,$AF,$FE,$59,$E0,$7F,$5B,$7F,$82,$DF,$F6,$37
	.dc.b	$33,$FD,$41,$CD,$FD,$BE,$44,$6A,$B8,$77,$F3,$C3,$A3,$F1,$3C,$A9
	.dc.b	$8F,$3F,$8B,$B7,$FC,$23,$6F,$E7,$6F,$C6,$AF,$24,$EB,$3F,$C5,$4E
	.dc.b	$6F,$E3,$56,$9A,$7F,$3E,$EA,$7C,$67,$EC,$0F,$D7,$16,$34,$DA,$FF
	.dc.b	$12,$F9,$6A,$E3,$5F,$B3,$AF,$EB,$FF,$47,$FC,$17,$E0,$1F,$47,$8A
	.dc.b	$FF,$66,$46,$1F,$C7,$CE,$E7,$58,$FF,$5B,$AD,$3F,$E1,$CF,$9E,$FE
	.dc.b	$99,$7F,$E9,$4F,$55,$18,$96,$7F,$D9,$FE,$97,$FC,$61,$39,$AD,$FA
	.dc.b	$38,$3F,$BA,$CD,$C3,$FA,$EE,$1F,$D9,$0F,$56,$D6,$32,$48,$F3,$46
	.dc.b	$5F,$37,$97,$E0,$3F,$C4,$5F,$6D,$FB,$AE,$1F,$DF,$65,$E8,$9D,$61
	.dc.b	$FE,$AE,$9C,$E6,$9E,$3E,$2E,$B1,$FE,$93,$F4,$3F,$DA,$BC,$CF,$F4
	.dc.b	$57,$C9,$7F,$A8,$E3,$12,$F8,$04,$FF,$A9,$FE,$95,$F4,$5F,$C3,$E7
	.dc.b	$35,$FE,$BE,$BE,$31,$88,$46,$2F,$B7,$AC,$C7,$C3,$5F,$EC,$6C,$7A
	.dc.b	$ED,$88,$E3,$5B,$7C,$97,$CA,$F5,$87,$59,$E5,$97,$96,$B5,$1C,$0F
	.dc.b	$35,$F3,$CB,$CD,$7C,$65,$97,$89,$E0,$78,$B1,$EF,$7C,$70,$EB,$87
	.dc.b	$BE,$F8,$3D,$27,$57,$E0,$BE,$27,$B6,$B7,$7C,$89,$3B,$3C,$31,$A7
	.dc.b	$17,$D6,$B9,$C6,$B1,$F8,$0E,$AF,$9B,$E9,$35,$97,$0B,$E5,$1E,$EC
	.dc.b	$BE,$16,$3D,$E3,$84,$E8,$97,$F8,$3C,$3B,$CE,$A3,$BD,$F2,$9C,$BF
	.dc.b	$34,$99,$7C,$4A,$F9,$F3,$A8,$C4,$EB,$2F,$8D,$39,$7D,$17,$EC,$C7
	.dc.b	$CD,$5B,$C3,$D5,$87,$59,$CB,$D0,$97,$D3,$2E,$75,$F1,$C8,$79,$1C
	.dc.b	$BF,$24,$1F,$8C,$29,$F5,$CA,$6C,$FD,$67,$D2,$7F,$8D,$DC,$9F,$A4
	.dc.b	$70,$F0,$C1,$18,$8E,$3E,$9F,$07,$83,$FA,$08,$E9,$A7,$E6,$9B,$9F
	.dc.b	$93,$E9,$1F,$4B,$E6,$AD,$5F,$12,$61,$F9,$EA,$DB,$89,$DB,$F8,$5C
	.dc.b	$3A,$BF,$0A,$76,$E6,$B8,$61,$E8,$8E,$98,$78,$63,$18,$74,$23,$E7
	.dc.b	$8D,$4F,$F0,$BC,$3B,$BF,$00,$EC,$74,$3B,$9B,$BA,$B0,$F2,$E1,$DA
	.dc.b	$34,$72,$7F,$8B,$46,$B0,$F2,$92,$77,$3A,$C3,$CB,$1E,$38,$E8,$D9
	.dc.b	$F5,$3E,$2C,$7E,$DE,$1E,$4C,$3B,$4F,$4A,$EC,$F1,$3E,$66,$1D,$AF
	.dc.b	$8C,$79,$E7,$75,$CA,$38,$12,$77,$87,$9E,$3C,$58,$FC,$A9,$CD,$F9
	.dc.b	$A7,$A1,$E3,$21,$F2,$70,$F1,$57,$12,$63,$FC,$D7,$1F,$16,$1D,$2F
	.dc.b	$8B,$5E,$47,$F3,$EE,$3F,$ED,$EF,$EA,$38,$FF,$AB,$B7,$4F,$EE,$9C
	.dc.b	$7F,$AA,$9F,$F1,$17,$D9,$BC,$1D,$5B,$0E,$6F,$F6,$1C,$38,$BA,$8C
	.dc.b	$AB,$E7,$59,$E5,$FE,$B7,$3C,$4E,$8E,$1E,$17,$FE,$07,$8C,$57,$FB
	.dc.b	$53,$AD,$E0,$73,$E6,$98,$73,$CB,$FB,$94,$7A,$3D,$55,$F0,$AD,$7D
	.dc.b	$37,$4C,$FF,$C2,$8D,$8F,$73,$7B,$BF,$DF,$5E,$7F,$B6,$35,$91,$A6
	.dc.b	$FC,$FF,$F3,$E5,$FF,$9D,$3D,$0F,$16,$4F,$A6,$A7,$CC,$C9,$6A,$C3
	.dc.b	$6F,$FC,$63,$5C,$1E,$87,$F9,$0F,$C2,$F7,$79,$4F,$94,$A7,$5F,$21
	.dc.b	$34,$CD,$7F,$AC,$E7,$0F,$FB,$5B,$E9,$7F,$D4,$ED,$F0,$FF,$02,$A3
	.dc.b	$2F,$D6,$3F,$D4,$3F,$45,$FE,$B2,$78,$10,$FF,$74,$C7,$0E,$6F,$F3
	.dc.b	$74,$EE,$73,$5D,$AB,$5C,$7F,$CF,$9B,$3D,$BD,$6F,$C9,$E9,$FE,$76
	.dc.b	$F0,$EF,$9C,$F8,$9A,$62,$25,$FF,$9F,$E3,$2F,$FA,$06,$3E,$AE,$D2
	.dc.b	$3F,$CB,$EF,$D2,$CF,$11,$FE,$5D,$8B,$AF,$FB,$8E,$7E,$2F,$CE,$3E
	.dc.b	$92,$92,$FF,$E0,$0F,$D2,$FE,$45,$F9,$02,$78,$3F,$A3,$31,$3C,$5D
	.dc.b	$CF,$1B,$FF,$43,$D7,$09,$E6,$D8,$CD,$7D,$8A,$E1,$BF,$F6,$AE,$BF
	.dc.b	$D1,$FB,$5F,$94,$26,$3F,$EC,$AF,$A2,$AF,$F9,$EE,$2E,$3F,$F0,$06
	.dc.b	$7E,$8F,$D2,$FF,$A4,$AB,$F7,$5C,$28,$7B,$DA,$37,$EA,$4A,$DC,$E9
	.dc.b	$B3,$97,$F1,$98,$3E,$91,$8F,$F1,$28,$C3,$63,$33,$D3,$E9,$31,$32
	.dc.b	$3F,$87,$57,$07,$46,$2B,$8B,$E0,$6C,$7F,$BF,$4E,$11,$F9,$FC,$BD
	.dc.b	$D7,$FB,$0F,$91,$93,$FC,$6F,$55,$27,$F6,$5A,$EA,$BF,$51,$5A,$9C
	.dc.b	$60,$DB,$FC,$3F,$DF,$FE,$B7,$C3,$F8,$1F,$D2,$7F,$0A,$F8,$99,$97
	.dc.b	$FC,$00,$8C,$BF,$EB,$68,$CB,$FE,$C0,$F0,$13,$FB,$AF,$A4,$FE,$79
	.dc.b	$3F,$8F,$6B,$F9,$46,$E5,$FE,$D4,$E4,$61,$C7,$CA,$97,$C5,$FC,$74
	.dc.b	$62,$39,$45,$AE,$53,$FD,$07,$E6,$A5,$7F,$37,$F2,$D1,$F7,$73,$FC
	.dc.b	$BC,$FF,$44,$F8,$DF,$E9,$04,$78,$7C,$DF,$F4,$B2,$38,$3F,$A6,$6F
	.dc.b	$FC,$B3,$F0,$4E,$9E,$DC,$9F,$C1,$E0,$8F,$EB,$1F,$4B,$F9,$B9,$FE
	.dc.b	$39,$E6,$A3,$1B,$7F,$B0,$9B,$C9,$FE,$DB,$6D,$7F,$65,$7D,$35,$C7
	.dc.b	$D0,$EB,$DD,$2B,$FA,$B5,$6B,$83,$FD,$71,$8F,$F5,$DE,$2F,$E8,$E7
	.dc.b	$F6,$7F,$62,$5F,$C7,$6B,$F7,$7C,$1B,$5F,$ED,$6B,$E7,$A7,$F6,$9B
	.dc.b	$FE,$97,$5E,$EF,$83,$FE,$8F,$E2,$CC,$7C,$71,$96,$90,$FE,$7D,$C4
	.dc.b	$F3,$1F,$4B,$B9,$F8,$FB,$2F,$1A,$FE,$6E,$7F,$DF,$B7,$C1,$FD,$9F
	.dc.b	$D2,$BE,$4B,$F7,$8B,$E3,$A9,$88,$93,$C1,$AF,$EC,$FE,$8F,$FB,$63
	.dc.b	$92,$1F,$C2,$AB,$B3,$FC,$5A,$36,$FD,$EF,$45,$F4,$9F,$E1,$96,$C7
	.dc.b	$F8,$2D,$AF,$7F,$27,$53,$23,$4F,$F5,$4A,$FE,$39,$1D,$63,$4D,$3C
	.dc.b	$53,$FE,$27,$FA,$4F,$F7,$74,$7E,$41,$BA,$7C,$9E,$94,$63,$97,$F8
	.dc.b	$2F,$CE,$3E,$87,$97,$2C,$C7,$FC,$45,$7B,$5C,$AB,$E1,$1A,$AE,$0F
	.dc.b	$85,$FF,$0A,$46,$A3,$F4,$7F,$47,$EC,$CF,$C8,$DC,$BF,$D6,$B9,$A3
	.dc.b	$15,$E0,$FA,$9F,$EF,$33,$9D,$7F,$C8,$D5,$A8,$C5,$7E,$E2,$BF,$BF
	.dc.b	$AF,$A2,$FE,$0B,$8F,$FB,$0B,$E8,$7F,$96,$4F,$FB,$13,$E9,$1E,$27
	.dc.b	$FB,$12,$73,$29,$CB,$C7,$CC,$FF,$64,$E1,$D2,$15,$C2,$30,$F2,$AE
	.dc.b	$33,$AA,$FF,$68,$C6,$27,$87,$8F,$FE,$5F,$AD,$1F,$ED,$8B,$E7,$7C
	.dc.b	$30,$F3,$36,$7C,$86,$EB,$95,$F4,$C7,$33,$EB,$BE,$2F,$C1,$6B,$F1
	.dc.b	$EB,$8D,$FB,$9B,$1B,$39,$E3,$E9,$9D,$5F,$6C,$7C,$F5,$B3,$D6,$78
	.dc.b	$4D,$5F,$47,$C1,$3D,$30,$F1,$E1,$D6,$7D,$93,$C0,$D3,$0E,$B7,$E0
	.dc.b	$79,$9E,$09,$E4,$F8,$4D,$DF,$86,$7E,$D4,$0F,$59,$FF,$00,$1C,$B8
	.dc.b	$34,$7D,$5E,$37,$D2,$7F,$BF,$E7,$F5,$8D,$8C,$CF,$36,$BC,$8F,$F7
	.dc.b	$F1,$4A,$51,$F1,$35,$F3,$35,$C9,$F3,$BE,$93,$FC,$B5,$C5,$F8,$08
	.dc.b	$D9,$9F,$67,$F9,$66,$FF,$9E,$8C,$BF,$A0,$AE,$31,$F6,$0C,$EF,$F0
	.dc.b	$23,$33,$C6,$7C,$2F,$1A,$EB,$5E,$BC,$3F,$23,$7C,$23,$17,$E3,$C3
	.dc.b	$CC,$F2,$8E,$D7,$CA,$3E,$35,$F2,$8C,$47,$C8,$C3,$C1,$7C,$AB,$4D
	.dc.b	$9E,$C7,$AC,$EC,$7D,$FC,$3C,$8F,$63,$75,$ED,$AE,$71,$C5,$B1,$98
	.dc.b	$DE,$1E,$C8,$F1,$C7,$EE,$B8,$5F,$48,$F8,$CF,$F2,$26,$5F,$E6,$B1
	.dc.b	$F1,$F8,$BF,$F7,$35,$6E,$7F,$C8,$86,$8F,$F6,$B3,$98,$FF,$21,$C6
	.dc.b	$5F,$F0,$8C,$67,$8B,$FB,$57,$A9,$FE,$45,$BE,$D8,$F9,$30,$F0,$4F
	.dc.b	$F9,$27,$E1,$73,$87,$F9,$2E,$7C,$31,$87,$61,$C3,$A3,$CA,$69,$CE
	.dc.b	$3E,$67,$C3,$F4,$6F,$92,$F0,$E0,$F9,$CC,$C6,$1B,$87,$F9,$96,$74
	.dc.b	$FF,$BD,$EB,$93,$99,$DE,$1F,$E6,$6A,$DB,$FE,$34,$AD,$B9,$9F,$1F
	.dc.b	$C3,$4B,$FF,$49,$FD,$1B,$E5,$F1,$7C,$6B,$EB,$6E,$FA,$27,$C5,$6A
	.dc.b	$3F,$D0,$11,$8C,$7C,$D5,$C2,$31,$7C,$E7,$FD,$1B,$18,$AF,$EF,$19
	.dc.b	$BE,$4B,$E3,$9F,$85,$F1,$3F,$DF,$FC,$98,$C3,$E1,$4F,$CA,$F9,$CC
	.dc.b	$65,$F0,$5B,$8F,$E0,$B8,$11,$8B,$FF,$5E,$46,$1F,$E4,$53,$FE,$09
	.dc.b	$F6,$BF,$CB,$D1,$83,$FB,$BB,$8B,$FF,$43,$70,$FE,$32,$31,$5C,$7D
	.dc.b	$8F,$BB,$79,$6B,$CE,$DC,$CB,$E3,$3E,$43,$81,$F8,$13,$B7,$C6,$47
	.dc.b	$FC,$EF,$ED,$7A,$6F,$FC,$E3,$4A,$D4,$63,$E2,$2E,$5E,$D7,$D0,$7E
	.dc.b	$8B,$93,$FD,$61,$8F,$97,$2F,$24,$E7,$B3,$DD,$EE,$FB,$FD,$9B,$7B
	.dc.b	$D1,$19,$8F,$F6,$4F,$D1,$7F,$97,$5D,$47,$E2,$A3,$1A,$7A,$3D,$1F
	.dc.b	$17,$B1,$F2,$DB,$7F,$EF,$DF,$94,$FF,$80,$0F,$FB,$A3,$EE,$DF,$0D
	.dc.b	$FC,$27,$05,$69,$FF,$68,$4E,$AB,$97,$B9,$FC,$3F,$E1,$93,$9A,$71
	.dc.b	$FF,$9C,$5E,$75,$B9,$D4,$E7,$0E,$CF,$F9,$DE,$F9,$E3,$E3,$9C,$E3
	.dc.b	$FD,$B6,$72,$FF,$AB,$BD,$B7,$CC,$FF,$12,$9B,$39,$8C,$1F,$A1,$9C
	.dc.b	$E7,$FE,$E2,$AB,$8F,$FA,$1F,$BB,$FD,$1D,$7C,$30,$FF,$76,$57,$E3
	.dc.b	$2D,$D7,$FE,$EE,$EA,$E8,$78,$49,$5C,$1C,$B4,$FC,$C4,$BF,$38,$FE
	.dc.b	$45,$1F,$9B,$B7,$E1,$BF,$ED,$48,$DC,$EA,$BE,$6C,$7C,$39,$D4,$60
	.dc.b	$FF,$09,$4E,$F3,$F9,$93,$A7,$EA,$46,$6B,$97,$A3,$F8,$18,$C5,$B0
	.dc.b	$F1,$CE,$7C,$8F,$C0,$63,$CF,$0F,$F1,$06,$BF,$EB,$5D,$BE,$91,$F3
	.dc.b	$CB,$1F,$5D,$B8,$FF,$AF,$1F,$F5,$E4,$7C,$E7,$ED,$70,$48,$FB,$CD
	.dc.b	$9F,$B0,$DD,$3E,$C5,$7E,$32,$BF,$C6,$5B,$7A,$6B,$F1,$5F,$47,$FD
	.dc.b	$96,$5F,$3F,$FE,$16,$C3,$F2,$53,$98,$C4,$FF,$8D,$B8,$93,$9F,$43
	.dc.b	$9C,$FF,$11,$39,$B8,$CF,$E9,$46,$0F,$D8,$4F,$17,$F1,$7E,$C7,$E5
	.dc.b	$CF,$CA,$15,$C9,$27,$E3,$46,$5F,$BD,$C2,$F8,$B7,$95,$FB,$37,$EC
	.dc.b	$C6,$63,$1C,$59,$6E,$1F,$8F,$AE,$2D,$9F,$E0,$5F,$3D,$FC,$33,$6F
	.dc.b	$F7,$54,$FC,$8F,$83,$C9,$F1,$CA,$FC,$3D,$F8,$3F,$E5,$EA,$B9,$A3
	.dc.b	$91,$FA,$28,$CB,$88,$FC,$D0,$8C,$CF,$2C,$DF,$24,$F9,$FF,$E2,$4B
	.dc.b	$E2,$2C,$61,$B8,$FE,$B6,$74,$D9,$D3,$FC,$2F,$E2,$7F,$85,$1F,$C1
	.dc.b	$CA,$E4,$FE,$06,$E5,$A7,$F3,$FF,$D0,$DF,$6B,$FC,$4F,$E3,$3F,$C4
	.dc.b	$F1,$89,$FE,$17,$18,$9D,$47,$16,$EB,$FB,$BF,$29,$7F,$EE,$FA,$FE
	.dc.b	$C1,$D3,$FE,$EC,$F8,$6E,$BE,$EF,$F8,$CB,$B5,$18,$AD,$C6,$1F,$34
	.dc.b	$EE,$7C,$BF,$03,$FC,$A7,$CF,$B5,$FC,$81,$B9,$FF,$27,$6F,$37,$F9
	.dc.b	$13,$8F,$FB,$35,$7F,$1C,$F9,$DF,$F2,$B7,$C3,$7F,$A3,$BE,$37,$FD
	.dc.b	$3F,$18,$C7,$FD,$27,$1F,$EA,$2F,$A0,$7C,$E7,$E1,$67,$27,$FA,$AE
	.dc.b	$B8,$7C,$12,$F9,$FB,$4C,$7B,$5E,$E3,$0F,$21,$8F,$6C,$B7,$0C,$7A
	.dc.b	$E1,$D3,$0C,$C6,$27,$FD,$79,$7C,$5E,$91,$97,$96,$1E,$0C,$7C,$D1
	.dc.b	$E3,$BE,$F7,$C6,$73,$3C,$27,$BD,$71,$C3,$C5,$3E,$9B,$E1,$18,$BE
	.dc.b	$38,$78,$4F,$14,$EA,$BC,$07,$9A,$FA,$38,$AF,$0C,$EE,$3C,$46,$F0
	.dc.b	$F1,$9E,$8C,$BE,$A5,$78,$2F,$9D,$70,$8C,$56,$F2,$E3,$3C,$C9,$3E
	.dc.b	$0B,$E7,$97,$AC,$93,$E5,$7F,$09,$1A,$8C,$10,$F0,$46,$AB,$BD,$78
	.dc.b	$A7,$BC,$E5,$A7,$1C,$BE,$3D,$F5,$8E,$66,$71,$F6,$E5,$F1,$6B,$E7
	.dc.b	$47,$C0,$78,$E3,$E0,$C3,$D3,$7C,$8F,$C7,$E1,$D6,$B8,$E5,$B9,$FC
	.dc.b	$5C,$E7,$0E,$87,$E6,$23,$FD,$F9,$5F,$A7,$6C,$65,$FC,$29,$AF,$6B
	.dc.b	$FC,$6E,$FF,$C0,$19,$FF,$40,$78,$BF,$F0,$05,$B1,$99,$DB,$C6,$7F
	.dc.b	$3F,$1C,$B4,$FC,$6B,$72,$54,$F9,$F7,$D0,$62,$3F,$47,$5B,$7F,$70
	.dc.b	$D7,$F5,$CD,$9E,$31,$E0,$C2,$3C,$9F,$DA,$C6,$A3,$04,$C3,$CB,$3A
	.dc.b	$BF,$14,$62,$75,$87,$F1,$0C,$3C,$15,$A8,$C5,$F8,$2F,$81,$FB,$4C
	.dc.b	$3C,$0E,$27,$78,$74,$3E,$15,$F3,$7D,$F7,$CE,$77,$5C,$EB,$55,$F0
	.dc.b	$AB,$4F,$03,$F8,$B5,$F8,$2F,$89,$23,$57,$CD,$E2,$73,$3C,$82,$79
	.dc.b	$3C,$27,$93,$B9,$E4,$F0,$C7,$F0,$B1,$EF,$8F,$69,$DA,$33,$3C,$AB
	.dc.b	$AC,$F3,$8E,$B3,$A8,$E2,$75,$C7,$F0,$51,$98,$F3,$63,$F1,$23,$04
	.dc.b	$9F,$AF,$3E,$17,$89,$87,$FA,$2F,$43,$F4,$37,$D3,$FE,$4F,$8F,$FA
	.dc.b	$A7,$75,$E2,$FF,$21,$F2,$7F,$81,$F1,$FF,$5A,$61,$C1,$B8,$6D,$FF
	.dc.b	$3E,$CE,$DF,$F4,$35,$E7,$1F,$05,$5B,$7C,$97,$1E,$BF,$F4,$2F,$C9
	.dc.b	$54,$EE,$26,$3F,$DE,$A7,$2D,$9C,$BF,$E0,$1A,$7C,$3F,$5C,$FF,$D8
	.dc.b	$B3,$FE,$94,$F6,$7C,$53,$23,$04,$39,$3E,$DE,$E7,$9E,$F7,$7A,$7B
	.dc.b	$12,$FF,$D5,$91,$B9,$D4,$7F,$A8,$A7,$4F,$F9,$E6,$7E,$46,$3F,$E3
	.dc.b	$4E,$CB,$E3,$3D,$0A,$62,$25,$FF,$8C,$E3,$2F,$F8,$D6,$3E,$92,$F6
	.dc.b	$1F,$37,$8F,$96,$7A,$54,$DC,$EA,$BF,$D5,$AD,$7F,$C9,$73,$C1,$FF
	.dc.b	$58,$B5,$FF,$27,$C6,$FA,$7F,$DB,$0D,$E9,$FE,$3D,$5B,$9F,$F6,$E3
	.dc.b	$63,$F9,$14,$EE,$BF,$D5,$1F,$36,$F3,$3E,$2B,$B8,$CE,$38,$99,$1F
	.dc.b	$EC,$19,$D6,$C6,$9C,$E3,$FD,$A7,$F0,$1F,$83,$23,$FE,$E3,$F4,$EB
	.dc.b	$FE,$DD,$8B,$A7,$FA,$9A,$7E,$6F,$CC,$7F,$99,$EB,$E1,$F1,$7C,$D3
	.dc.b	$3E,$84,$8C,$90,$FF,$3A,$CF,$16,$C7,$17,$57,$FF,$00,$67,$4F,$F9
	.dc.b	$1F,$DE,$FC,$D1,$27,$FD,$F5,$F0,$35,$FF,$7A,$45,$9F,$F5,$EC,$FE
	.dc.b	$4E,$FF,$45,$EA,$CB,$FE,$AB,$5F,$EC,$29,$D4,$7F,$BE,$27,$4F,$FB
	.dc.b	$3E,$3F,$1B,$EC,$3E,$07,$F3,$3F,$6D,$FE,$52,$AF,$F7,$93,$5F,$F4
	.dc.b	$9C,$ED,$FF,$7C,$B5,$FF,$59,$C6,$F0,$FF,$2D,$70,$23,$1C,$9F,$D9
	.dc.b	$EF,$85,$7F,$97,$EF,$FD,$9B,$87,$8B,$CE,$91,$DA,$3F,$80,$7D,$2C
	.dc.b	$8F,$F4,$0E,$48,$FD,$B7,$8F,$99,$61,$FE,$75,$9C,$3F,$EE,$69,$FD
	.dc.b	$E7,$83,$FC,$2F,$B2,$5F,$CA,$D3,$FD,$F7,$DF,$FC,$3A,$3F,$DE,$D5
	.dc.b	$A7,$AC,$FF,$2C,$AE,$71,$C4,$FF,$03,$FD,$93,$E1,$FF,$82,$FD,$AB
	.dc.b	$D2,$A7,$EF,$A7,$55,$C4,$38,$7F,$DB,$F3,$C5,$ED,$B2,$31,$1F,$EC
	.dc.b	$89,$E1,$AF,$CD,$6E,$5F,$E1,$AB,$E1,$87,$8A,$FC,$15,$A5,$E5,$BE
	.dc.b	$6F,$F9,$3C,$7F,$37,$F9,$CF,$E7,$87,$F4,$38,$CC,$7F,$AD,$DF,$DB
	.dc.b	$3F,$18,$FA,$7F,$14,$F8,$6F,$E9,$DA,$23,$13,$F2,$70,$F0,$78,$3D
	.dc.b	$7D,$EE,$CC,$9D,$62,$9F,$17,$7F,$F4,$FF,$12,$56,$A3,$CD,$D9,$A8
	.dc.b	$C1,$C9,$CB,$48,$7A,$7C,$AF,$E8,$FE,$1F,$BF,$BF,$FA,$97,$91,$2B
	.dc.b	$C7,$19,$7F,$2B,$19,$7E,$24,$65,$A7,$03,$F1,$15,$FD,$53,$C2,$FE
	.dc.b	$DD,$E5,$3E,$CF,$FB,$05,$43,$E3,$46,$5B,$19,$7E,$87,$9E,$E8,$FF
	.dc.b	$2B,$47,$FB,$AE,$B8,$36,$BB,$91,$F1,$F7,$32,$F1,$FB,$9F,$63,$EC
	.dc.b	$1F,$19,$4C,$DF,$92,$70,$FE,$8F,$A1,$5C,$23,$44,$3C,$93,$C9,$FC
	.dc.b	$46,$6F,$9D,$FE,$91,$D0,$FC,$29,$F1,$5F,$DC,$3E,$33,$FC,$0B,$87
	.dc.b	$E9,$1A,$FE,$46,$B8,$4F,$F8,$66,$B8,$CE,$FD,$4E,$46,$66,$46,$48
	.dc.b	$73,$9E,$2F,$5F,$9C,$FE,$B9,$C6,$FB,$17,$A3,$FE,$23,$6D,$7E,$76
	.dc.b	$BD,$67,$08,$CC,$7D,$86,$F0,$3C,$5F,$F4,$8E,$8B,$E3,$E9,$F5,$C6
	.dc.b	$63,$D6,$3F,$C6,$D3,$FE,$18,$8E,$9E,$8F,$8C,$CF,$74,$8F,$E3,$1A
	.dc.b	$7A,$67,$51,$E7,$9F,$F0,$9F,$CD,$FD,$0D,$3F,$D4,$7F,$32,$EF,$FF
	.dc.b	$51,$7C,$E3,$E1,$FF,$D4,$93,$9C,$7F,$D5,$33,$97,$8D,$7F,$AA,$70
	.dc.b	$E9,$0A,$E0,$A3,$31,$8A,$E5,$3B,$AF,$0D,$F8,$1E,$9C,$7F,$D0,$A7
	.dc.b	$47,$FA,$EB,$0E,$73,$C2,$3C,$67,$9A,$B9,$3B,$AE,$58,$79,$2F,$8C
	.dc.b	$EC,$E2,$30,$F2,$CE,$EB,$8E,$1E,$4C,$77,$5A,$8F,$15,$F9,$8F,$81
	.dc.b	$5C,$1F,$15,$70,$C7,$CB,$1D,$DB,$8F,$92,$38,$CE,$D8,$75,$8D,$1B
	.dc.b	$C3,$C5,$1E,$B3,$57,$E1,$79,$9E,$58,$DB,$E1,$3E,$0C,$F4,$9F,$A1
	.dc.b	$1F,$C4,$23,$EE,$B6,$32,$FC,$33,$F8,$A9,$F7,$FE,$21,$F3,$98,$8C
	.dc.b	$36,$93,$A9,$E3,$18,$9E,$11,$FE,$F8,$F8,$6A,$9A,$47,$6F,$DA,$F9
	.dc.b	$CF,$F3,$C7,$2F,$DC,$9E,$47,$D5,$8C,$BF,$5C,$CE,$DF,$98,$8C,$BF
	.dc.b	$9A,$8C,$EC,$64,$EA,$3E,$0D,$E7,$BB,$97,$A5,$FC,$06,$FF,$97,$D6
	.dc.b	$9A,$72,$8D,$4F,$78,$F7,$1E,$78,$E5,$86,$6F,$DD,$18,$BE,$93,$E0
	.dc.b	$C7,$E0,$5F,$28,$C3,$EF,$31,$3E,$B5,$8F,$AD,$FE,$9F,$87,$6B,$E6
	.dc.b	$D8,$E4,$DB,$F2,$47,$9E,$76,$D3,$67,$C0,$3B,$E1,$F0,$30,$F8,$98
	.dc.b	$F9,$8F,$E2,$1C,$EF,$9C,$FF,$18,$61,$F8,$E8,$C7,$A3,$FE,$51,$C7
	.dc.b	$F9,$6D,$69,$FE,$CF,$3A,$AE,$13,$9E,$2F,$A9,$CD,$FC,$D7,$B3,$EB
	.dc.b	$47,$D7,$BE,$18,$7F,$72,$C3,$FD,$B2,$7F,$87,$70,$FE,$78,$D3,$89
	.dc.b	$F8,$5F,$1D,$F3,$DF,$E7,$58,$D1,$2B,$CB,$5C,$9B,$1E,$B7,$FC,$73
	.dc.b	$19,$7E,$B5,$FC,$39,$EF,$0C,$BF,$A7,$7C,$8E,$2B,$55,$FD,$10,$46
	.dc.b	$7E,$65,$F3,$B4,$E7,$9F,$F8,$EF,$E2,$31,$7F,$E3,$BA,$DB,$F4,$5B
	.dc.b	$97,$CE,$6B,$E6,$6B,$FD,$81,$DE,$BF,$43,$3F,$F3,$65,$75,$C7,$D7
	.dc.b	$1E,$BF,$63,$CD,$F2,$92,$BF,$CC,$FF,$3A,$F9,$DC,$67,$FE,$05,$CD
	.dc.b	$7F,$27,$74,$63,$0F,$44,$E6,$31,$5E,$98,$F9,$7A,$3E,$77,$FC,$4B
	.dc.b	$7F,$06,$73,$18,$AF,$06,$DF,$2C,$F9,$3E,$C7,$E3,$6F,$E0,$46,$5F
	.dc.b	$90,$F8,$48,$E0,$FE,$F3,$B3,$E7,$78,$5F,$F9,$0A,$31,$3F,$E6,$DF
	.dc.b	$81,$FF,$39,$37,$DB,$F0,$F0,$FE,$06,$73,$7F,$E2,$96,$BF,$8A,$6B
	.dc.b	$FE,$14,$6B,$FE,$88,$6E,$FF,$EC,$DA,$FA,$B8,$7E,$5F,$6F,$D3,$57
	.dc.b	$67,$EE,$63,$E1,$F9,$48,$FF,$99,$35,$F4,$E5,$E3,$9C,$BF,$B6,$B4
	.dc.b	$7D,$96,$D7,$EF,$DB,$3B,$F1,$B4,$49,$D3,$C7,$C7,$91,$C5,$B3,$DF
	.dc.b	$E0,$3B,$12,$7B,$3E,$1F,$5A,$47,$85,$B3,$FD,$7B,$E4,$DE,$DF,$EC
	.dc.b	$A1,$18,$AF,$F5,$17,$E0,$2F,$97,$D4,$E6,$BF,$D5,$87,$0B,$FE,$00
	.dc.b	$DB,$FE,$D1,$3B,$8F,$F4,$0B,$5F,$E7,$0F,$3D,$FF,$DE,$FF,$3A,$F9
	.dc.b	$BE,$F8,$7F,$06,$F5,$D3,$9C,$7F,$D0,$BE,$B7,$EC,$24,$FF,$02,$F6
	.dc.b	$EC,$FF,$BD,$A3,$06,$CA,$E4,$FF,$7D,$F3,$73,$39,$15,$C9,$34,$FF
	.dc.b	$5A,$57,$E2,$63,$FB,$67,$37,$FA,$7E,$BE,$67,$CE,$FF,$18,$9F,$F7
	.dc.b	$63,$6F,$FD,$1B,$F3,$BF,$B3,$DB,$FB,$74,$27,$51,$87,$84,$7F,$AA
	.dc.b	$9A,$FF,$61,$6B,$FD,$F5,$AF,$3F,$13,$E5,$57,$FA,$62,$FC,$71,$8F
	.dc.b	$4B,$E4,$67,$F2,$CB,$3F,$E7,$4D,$51,$8C,$3F,$D7,$B1,$89,$FF,$5B
	.dc.b	$CF,$17,$87,$07,$D5,$3E,$A1,$5C,$92,$BE,$5C,$65,$FC,$6F,$0B,$E2
	.dc.b	$FF,$22,$7F,$48,$A7,$CE,$3E,$0B,$31,$FE,$1D,$DF,$E5,$DF,$18,$FC
	.dc.b	$9D,$FE,$5A,$31,$87,$E8,$E3,$07,$5C,$3F,$C5,$B7,$CA,$31,$E7,$FE
	.dc.b	$3F,$D9,$98,$FF,$B5,$9B,$87,$C4,$6D,$FF,$8A,$7C,$AF,$93,$18,$AF
	.dc.b	$F1,$96,$DF,$EB,$6F,$1B,$F3,$DE,$D7,$E0,$27,$2F,$F3,$EC,$3F,$C6
	.dc.b	$55,$FE,$BC,$E2,$DD,$F0,$6D,$FC,$13,$8B,$FB,$DF,$9D,$FF,$75,$46
	.dc.b	$73,$FF,$21,$46,$16,$1F,$E3,$F9,$F9,$BF,$19,$C7,$E0,$A4,$7F,$98
	.dc.b	$EF,$85,$FF,$BF,$1A,$FF,$8C,$DA,$F6,$9F,$98,$72,$7B,$CF,$F9,$4C
	.dc.b	$FB,$71,$FB,$FF,$93,$7C,$6D,$F8,$7C,$36,$BD,$5A,$AB,$71,$85,$5C
	.dc.b	$23,$0F,$F0,$5F,$8A,$FE,$10,$58,$FC,$F3,$71,$FC,$EB,$74,$FF,$17
	.dc.b	$E0,$8F,$F0,$B7,$CE,$FF,$25,$BF,$F1,$E3,$6F,$FC,$72,$D7,$FC,$42
	.dc.b	$D7,$FD,$D6,$EE,$FA,$FC,$57,$4F,$0F,$F8,$E3,$DB,$E0,$26,$1F,$B3
	.dc.b	$8B,$87,$CA,$1E,$6B,$1F,$F4,$AD,$79,$2F,$FC,$19,$18,$AF,$2C,$63
	.dc.b	$E0,$A5,$7A,$9B,$3E,$A7,$F9,$77,$9B,$F0,$1B,$3F,$8B,$97,$CC,$BD
	.dc.b	$BF,$DA,$47,$FA,$7F,$E6,$FF,$CE,$D3,$9F,$03,$FD,$49,$1F,$E7,$9F
	.dc.b	$9D,$7C,$97,$FA,$3F,$E0,$93,$C3,$1E,$71,$8C,$7F,$D3,$58,$F3,$7B
	.dc.b	$18,$9D,$D7,$5B,$E8,$F3,$C7,$C3,$5B,$6B,$EA,$AE,$2F,$FA,$9F,$1F
	.dc.b	$2E,$5A,$C7,$C1,$97,$8E,$B8,$61,$9C,$38,$C3,$0E,$2F,$98,$D9,$6F
	.dc.b	$9E,$3E,$58,$D4,$F0,$C3,$B5,$6A,$7D,$07,$1B,$EF,$5B,$BE,$57,$C2
	.dc.b	$7C,$77,$C6,$FC,$99,$7C,$47,$9D,$6F,$1F,$51,$F1,$31,$E9,$3A,$C7
	.dc.b	$A9,$1F,$8D,$7C,$DF,$6D,$F3,$7C,$D7,$CE,$75,$87,$8C,$98,$FA,$A3
	.dc.b	$2F,$A0,$93,$C7,$2F,$2D,$F4,$26,$1F,$4F,$2F,$3C,$78,$B2,$F3,$34
	.dc.b	$F3,$D7,$2A,$F3,$12,$7F,$11,$19,$78,$E5,$F2,$6F,$AB,$E8,$C3,$58
	.dc.b	$74,$3C,$75,$C3,$1F,$66,$1A,$9F,$C0,$9F,$ED,$92,$9F,$81,$31,$7F
	.dc.b	$9D,$6F,$AF,$FE,$36,$78,$1F,$B0,$70,$EB,$C4,$CB,$96,$CA,$79,$67
	.dc.b	$F8,$C6,$E3,$FB,$66,$D8,$7C,$EB,$E7,$7F,$DE,$17,$C7,$1F,$D2,$5F
	.dc.b	$0B,$F0,$1F,$9F,$BE,$0E,$A6,$3F,$A1,$9E,$13,$A8,$C4,$EA,$BB,$47
	.dc.b	$3B,$F0,$E1,$E7,$AD,$36,$74,$60,$FE,$0F,$87,$97,$1F,$85,$5F,$C0
	.dc.b	$31,$F0,$63,$D8,$98,$73,$7F,$7D,$1A,$8C,$13,$0E,$B5,$CF,$1F,$2C
	.dc.b	$60,$F5,$92,$F9,$C6,$A3,$E3,$C7,$81,$E0,$E9,$F5,$1A,$AF,$03,$F5
	.dc.b	$27,$96,$1E,$C8,$ED,$3A,$C3,$D1,$1D,$23,$8C,$EE,$FB,$4F,$03,$44
	.dc.b	$3B,$1E,$28,$E1,$3D,$6F,$C3,$8F,$CD,$26,$3F,$0E,$F8,$46,$1F,$AA
	.dc.b	$77,$3D,$87,$7C,$7C,$FC,$6F,$9A,$7C,$EF,$E4,$EB,$FC,$A3,$DB,$7A
	.dc.b	$7F,$BA,$B7,$7F,$CC,$61,$C6,$F1,$7F,$D9,$B0,$E6,$FE,$AA,$74,$D8
	.dc.b	$CB,$FB,$58,$CB,$61,$F3,$BF,$AE,$9F,$33,$63,$33,$CE,$30,$78,$A7
	.dc.b	$71,$FE,$33,$6B,$FE,$6C,$F5,$B5,$4F,$84,$F4,$FA,$4E,$F5,$61,$B7
	.dc.b	$CE,$F8,$36,$FC,$84,$7D,$CF,$F0,$6F,$27,$FC,$FB,$EC,$F3,$F6,$25
	.dc.b	$7F,$91,$E7,$2D,$F4,$15,$A8,$FF,$41,$D7,$CE,$AD,$C7,$F8,$57,$1F
	.dc.b	$F1,$B7,$57,$D2,$7F,$C7,$1E,$6D,$F1,$8D,$99,$5F,$E3,$B8,$D1,$87
	.dc.b	$FD,$A3,$EB,$BE,$73,$EE,$C3,$E7,$1E,$86,$1B,$1F,$E1,$7A,$E0,$DA
	.dc.b	$FE,$4E,$FF,$B9,$1B,$1A,$8F,$E5,$ED,$7F,$9C,$39,$E8,$FF,$5C,$B7
	.dc.b	$9F,$F2,$CD,$9F,$E5,$56,$F1,$FE,$17,$CE,$7F,$92,$E7,$FC,$B6,$F9
	.dc.b	$9F,$90,$DC,$3F,$D0,$2E,$E3,$3E,$57,$2C,$E4,$72,$8F,$C2,$7C,$4C
	.dc.b	$8F,$F5,$D3,$F5,$79,$1E,$96,$67,$FD,$77,$37,$1F,$F4,$C5,$6A,$7F
	.dc.b	$C7,$DF,$39,$FE,$6C,$7F,$C8,$1A,$7E,$09,$BC,$DF,$4A,$75,$1F,$E5
	.dc.b	$50,$E1,$FE,$C0,$D7,$F7,$8D,$7F,$A8,$35,$E3,$E7,$75,$23,$E2,$F7
	.dc.b	$BF,$45,$F2,$F2,$3F,$DB,$0F,$BE,$31,$1F,$ED,$88,$B9,$7B,$A7,$E2
	.dc.b	$57,$EA,$FD,$19,$7F,$9F,$8C,$BF,$EE,$38,$CB,$FE,$71,$E9,$FF,$72
	.dc.b	$CF,$E4,$BD,$67,$C0,$7E,$7F,$E7,$3F,$CB,$B7,$FE,$98,$6B,$F1,$67
	.dc.b	$6F,$F9,$AD,$AF,$F7,$E6,$BF,$E0,$76,$BF,$C8,$DA,$FF,$77,$78,$E4
	.dc.b	$E7,$D5,$F9,$4F,$27,$CC,$7F,$09,$F5,$E6,$1F,$EF,$F7,$F0,$DE,$3E
	.dc.b	$23,$FD,$FF,$17,$0F,$C2,$CF,$E2,$FC,$3F,$E2,$5D,$B2,$FF,$83,$46
	.dc.b	$5F,$F6,$87,$89,$FC,$66,$75,$1F,$27,$9B,$F8,$CB,$F8,$7F,$9B,$7C
	.dc.b	$2F,$F0,$7F,$9C,$FF,$46,$4F,$FB,$01,$AE,$E3,$F8,$0C,$F0,$8F,$F1
	.dc.b	$0B,$5F,$F1,$3B,$5F,$F1,$83,$5F,$E9,$8E,$F3,$F9,$11,$FE,$21,$8C
	.dc.b	$6F,$AA,$FE,$75,$F6,$DF,$BB,$AF,$E7,$5F,$38,$CB,$FA,$23,$63,$FA
	.dc.b	$73,$5F,$E9,$8D,$7F,$AE,$79,$48,$FD,$F4,$61,$AF,$F5,$C6,$BF,$EC
	.dc.b	$06,$C7,$1F,$0B,$C2,$49,$F0,$BE,$4E,$C9,$1E,$46,$BE,$4E,$7F,$49
	.dc.b	$1F,$4B,$5F,$F6,$85,$79,$E3,$FA,$E5,$FF,$00,$D7,$F8,$4F,$76,$3A
	.dc.b	$BF,$AB,$78,$B3,$1F,$8F,$19,$7E,$2B,$F2,$08,$E0,$D9,$0F,$C4,$76
	.dc.b	$FF,$6B,$F2,$1D,$3F,$D8,$E6,$1F,$DB,$63,$2D,$8F,$EF,$11,$88,$FE
	.dc.b	$FC,$D7,$E9,$B5,$FF,$05,$BB,$C7,$F0,$BA,$23,$1C,$5A,$BF,$E5,$D8
	.dc.b	$F9,$7E,$6F,$FC,$23,$83,$E7,$1F,$27,$86,$E7,$E7,$AE,$0F,$F8,$57
	.dc.b	$24,$7A,$09,$19,$76,$FD,$48,$DB,$FE,$79,$BF,$CD,$F9,$6F,$9C,$FE
	.dc.b	$99,$A7,$F8,$9B,$E7,$1F,$0D,$FD,$D7,$E2,$BF,$C1,$38,$FF,$79,$6E
	.dc.b	$3F,$02,$B7,$3F,$E1,$96,$BF,$E2,$76,$BF,$9C,$6B,$DE,$BF,$C5,$11
	.dc.b	$C2,$7A,$47,$F8,$96,$31,$7F,$E9,$1F,$15,$FA,$3B,$1F,$F1,$57,$CE
	.dc.b	$7A,$B2,$78,$2F,$9D,$FE,$B1,$0D,$3F,$77,$39,$F3,$38,$F9,$D2,$BF
	.dc.b	$C8,$58,$7F,$71,$7F,$66,$D7,$FC,$96,$D5,$CE,$8F,$13,$67,$A4,$FF
	.dc.b	$45,$FA,$17,$D3,$FA,$3B,$FF,$C7,$7A,$3C,$DA,$66,$FF,$C8,$99,$7F
	.dc.b	$6D,$7F,$CA,$4D,$7C,$B5,$FE,$51,$8C,$47,$F9,$97,$E7,$BE,$DE,$DF
	.dc.b	$EB,$9E,$A7,$7F,$F9,$A7,$E7,$DF,$0B,$FE,$6B,$9C,$E1,$FE,$BF,$9C
	.dc.b	$BC,$6B,$A7,$47,$CB,$8F,$CC,$C6,$30,$FF,$63,$7C,$EF,$29,$5F,$25
	.dc.b	$19,$8C,$3B,$CB,$CF,$3A,$BE,$8D,$3C,$99,$7F,$A0,$6B,$93,$67,$71
	.dc.b	$CF,$1F,$33,$AA,$F2,$CF,$C1,$AE,$11,$FE,$DD,$BF,$0D,$71,$C7,$73
	.dc.b	$E0,$29,$EC,$AE,$2A,$B8,$DF,$88,$ED,$1C,$31,$F7,$C6,$63,$C1,$8F
	.dc.b	$CB,$CB,$CD,$1F,$30,$F0,$14,$ED,$1B,$9E,$51,$E3,$3B,$61,$E4,$C3
	.dc.b	$C7,$1D,$B1,$CB,$0F,$98,$74,$7D,$B1,$F4,$A3,$FD,$5C,$64,$A5,$32
	.dc.b	$ED,$FC,$87,$C4,$3E,$7F,$71,$F1,$63,$0F,$29,$46,$23,$9B,$5E,$FF
	.dc.b	$0B,$FE,$74,$4D,$23,$FE,$78,$7E,$AF,$CF,$FF,$A2,$BB,$3E,$BC,$8E
	.dc.b	$27,$C0,$8D,$3F,$08,$D4,$F2,$F3,$FF,$C0,$3C,$1F,$9B,$9C,$9C,$63
	.dc.b	$3A,$F9,$A3,$18,$78,$1E,$F5,$FA,$3C,$3B,$BC,$E3,$35,$FE,$DB,$9C
	.dc.b	$BC,$E7,$85,$F6,$AF,$A7,$87,$97,$0E,$CD,$8F,$51,$D0,$F8,$18,$74
	.dc.b	$78,$9E,$63,$71,$C6,$FC,$8D,$C3,$C4,$79,$70,$F1,$57,$FB,$C6,$74
	.dc.b	$F4,$AF,$3E,$1E,$48,$ED,$1F,$C3,$3A,$1F,$3C,$F8,$8F,$F1,$6E,$FF
	.dc.b	$DA,$A3,$18,$7F,$B3,$FE,$7F,$FA,$9B,$FE,$E5,$8C,$E3,$FE,$10,$6D
	.dc.b	$71,$BE,$75,$FD,$1E,$F8,$DE,$E7,$2F,$F3,$8B,$FF,$32,$57,$53,$3E
	.dc.b	$02,$FF,$CF,$31,$92,$57,$85,$DC,$6A,$31,$1A,$C3,$D8,$EB,$D6,$F0
	.dc.b	$E5,$F0,$A2,$FC,$E3,$E7,$E9,$E2,$F9,$B7,$C4,$62,$BF,$BD,$1F,$9B
	.dc.b	$B5,$F1,$8C,$E5,$FD,$52,$30,$FF,$A5,$6F,$9C,$EA,$B8,$E1,$D6,$BF
	.dc.b	$B3,$4F,$E9,$FC,$3F,$ED,$7E,$AE,$74,$FE,$D3,$1F,$E7,$49,$E5,$93
	.dc.b	$E5,$3D,$67,$CE,$F2,$9F,$3E,$F9,$FC,$DF,$0A,$B9,$F8,$05,$F0,$AC
	.dc.b	$E5,$FE,$2E,$C7,$B6,$1E,$E8,$7C,$FF,$F8,$AA,$3A,$19,$71,$87,$F9
	.dc.b	$5A,$BE,$0B,$78,$3E,$77,$2F,$D2,$65,$FB,$1C,$B4,$93,$E7,$D7,$FC
	.dc.b	$1F,$F3,$FC,$23,$10,$8F,$BF,$C6,$F9,$DA,$7D,$DE,$1B,$87,$D3,$9D
	.dc.b	$36,$74,$D2,$61,$D9,$E2,$D7,$F0,$3C,$CE,$5F,$14,$75,$F9,$5F,$C6
	.dc.b	$65,$E4,$64,$64,$E3,$3C,$2F,$FC,$FB,$8A,$BD,$7E,$34,$BE,$15,$A8
	.dc.b	$F7,$F8,$F2,$79,$96,$79,$FB,$FF,$10,$C7,$1F,$F9,$AB,$14,$7C,$8F
	.dc.b	$4A,$5F,$DF,$FA,$27,$AD,$8C,$7F,$10,$3E,$45,$F1,$5F,$BE,$8C,$3F
	.dc.b	$B0,$FA,$07,$CC,$E6,$B7,$8F,$EA,$2F,$6D,$8C,$C7,$FA,$7A,$B8,$CE
	.dc.b	$6F,$91,$1F,$D6,$46,$A7,$CB,$5C,$B5,$FF,$65,$F1,$FF,$B1,$7D,$5F
	.dc.b	$E2,$7C,$FF,$A0,$7F,$88,$FC,$0F,$F9,$77,$2F,$F7,$3F,$75,$CB,$FD
	.dc.b	$CD,$5C,$23,$18,$F4,$BB,$7F,$B9,$9B,$7F,$E6,$47,$F1,$AF,$E6,$5B
	.dc.b	$5C,$7A,$3A,$12,$BC,$2F,$90,$C5,$7D,$0F,$9F,$7C,$E7,$CF,$6D,$7F
	.dc.b	$12,$C1,$F3,$FF,$B2,$C7,$FD,$D7,$82,$F6,$F3,$C3,$91,$FC,$26,$B8
	.dc.b	$61,$FD,$E6,$78,$4E,$A3,$EC,$57,$2B,$FF,$4E,$DF,$0B,$FF,$00,$4E
	.dc.b	$6B,$FD,$15,$97,$FB,$D5,$F9,$AF,$D0,$3E,$61,$5C,$9F,$AC,$78,$4F
	.dc.b	$C9,$7E,$97,$CF,$C9,$D3,$67,$E7,$37,$0F,$AC,$D9,$F6,$6E,$F8,$BF
	.dc.b	$CA,$1F,$D5,$A1,$F3,$EF,$84,$CC,$7E,$56,$6E,$3F,$CA,$63,$73,$8C
	.dc.b	$7E,$7E,$1D,$67,$81,$FD,$18,$D9,$8A,$FC,$27,$B9,$F8,$D3,$F2,$65
	.dc.b	$8F,$C9,$B6,$3F,$9C,$71,$B4,$4E,$70,$FF,$67,$FB,$BF,$CE,$2F,$F9
	.dc.b	$F7,$0A,$3E,$A4,$7F,$40,$75,$3C,$38,$7F,$A5,$46,$32,$FF,$1D,$5F
	.dc.b	$19,$CF,$91,$A7,$F8,$07,$CF,$FF,$06,$AF,$DE,$37,$1F,$8D,$C0,$F8
	.dc.b	$DD,$F5,$CB,$DB,$F4,$67,$51,$82,$57,$F1,$0B,$E9,$A2,$7F,$C9,$C7
	.dc.b	$D8,$C3,$A5,$7C,$0F,$94,$7C,$77,$F6,$55,$F1,$56,$30,$DB,$FD,$54
	.dc.b	$E9,$B3,$A7,$F2,$9F,$25,$FC,$0C,$FE,$12,$57,$27,$F0,$A7,$2D,$3F
	.dc.b	$C0,$5F,$3F,$FC,$5F,$2F,$56,$3F,$ED,$C8,$C4,$FE,$A3,$E3,$7F,$C3
	.dc.b	$31,$89,$D7,$83,$83,$F8,$CC,$DA,$EB,$ED,$F0,$3F,$8C,$45,$8C,$92
	.dc.b	$BF,$8A,$46,$5F,$F0,$27,$A8,$C3,$FD,$AF,$5E,$8F,$7B,$FC,$1D,$F7
	.dc.b	$4F,$BC,$F2,$C4,$9D,$1F,$C4,$E7,$4E,$30,$FE,$93,$F3,$FF,$E5,$3E
	.dc.b	$EF,$F4,$0F,$CF,$FF,$0F,$8C,$4F,$FA,$42,$3F,$89,$7D,$03,$E8,$7E
	.dc.b	$3F,$BE,$9E,$1E,$BF,$ED,$EF,$81,$FE,$91,$C3,$9D,$6E,$31,$3B,$8C
	.dc.b	$43,$0E,$D9,$79,$23,$79,$55,$87,$4C,$7C,$33,$9B,$E3,$5E,$23,$78
	.dc.b	$F7,$AD,$C6,$6F,$93,$FE,$A7,$AE,$38,$F9,$27,$33,$CB,$0F,$14,$62
	.dc.b	$B7,$7D,$2F,$86,$1D,$F0,$F4,$9E,$3A,$DB,$89,$DE,$1D,$2B,$84,$FC
	.dc.b	$3C,$BE,$7E,$1D,$49,$1C,$5F,$AB,$1A,$8C,$10,$F4,$1E,$17,$4E,$2F
	.dc.b	$C2,$FC,$3A,$F0,$E1,$E6,$8D,$5F,$81,$A7,$C7,$BE,$78,$FC,$62,$5F
	.dc.b	$28,$CB,$EC,$8C,$E3,$F5,$E7,$2F,$C1,$25,$79,$F2,$B9,$79,$89,$8F
	.dc.b	$C0,$8C,$BE,$5C,$3E,$A6,$1D,$2F,$D9,$86,$9F,$45,$D7,$D6,$4C,$BE
	.dc.b	$05,$F9,$A3,$FE,$00,$BF,$A4,$7F,$BA,$1F,$30,$AE,$5F,$D5,$7A,$FF
	.dc.b	$E3,$57,$8B,$5D,$B5,$DF,$89,$D1,$FD,$44,$71,$CF,$F2,$4D,$CF,$73
	.dc.b	$F9,$DF,$A0,$7D,$0F,$FB,$56,$73,$7C,$30,$FC,$BE,$1C,$70,$0C,$38
	.dc.b	$BA,$98,$FE,$96,$78,$3B,$C3,$B3,$E6,$9D,$DF,$78,$F2,$4E,$9F,$DB
	.dc.b	$D7,$2C,$7F,$D7,$B7,$C3,$1F,$F0,$F6,$3E,$2B,$F0,$4E,$5E,$C7,$43
	.dc.b	$07,$46,$C7,$1C,$7C,$44,$AF,$86,$76,$C7,$C9,$87,$8B,$0E,$E4,$AF
	.dc.b	$E2,$77,$CE,$F8,$CF,$8A,$7A,$3F,$2A,$74,$E6,$76,$F9,$4F,$41,$D5
	.dc.b	$CD,$78,$6B,$BD,$F9,$31,$F8,$C4,$C7,$F9,$06,$3D,$B1,$E9,$7E,$26
	.dc.b	$E3,$ED,$7E,$2E,$3D,$63,$C7,$5C,$63,$13,$E7,$21,$F7,$4E,$38,$78
	.dc.b	$5F,$4D,$F7,$9E,$31,$E0,$3E,$16,$3F,$84,$E3,$7C,$FB,$E8,$7F,$CB
	.dc.b	$5D,$3F,$DB,$A7,$31,$8D,$FF,$DF,$95,$BC,$3F,$CD,$5F,$70,$FA,$4D
	.dc.b	$43,$0E,$2F,$FB,$EE,$B8,$3F,$DA,$1B,$8F,$F7,$CF,$19,$87,$3A,$E1
	.dc.b	$F4,$7D,$30,$F3,$B7,$03,$E0,$F4,$BF,$DE,$BE,$BE,$62,$5F,$23,$FD
	.dc.b	$2F,$1C,$7E,$8E,$47,$95,$F4,$F9,$DF,$E4,$7F,$73,$33,$E7,$9B,$AF
	.dc.b	$FA,$6E,$77,$E7,$BD,$CF,$16,$EF,$6A,$9A,$89,$8F,$FA,$A2,$32,$FF
	.dc.b	$9E,$63,$E4,$CB,$C8,$FE,$ED,$3F,$E3,$5F,$88,$CC,$1B,$DD,$27,$FC
	.dc.b	$67,$1A,$30,$FF,$B4,$BD,$A7,$D1,$FD,$68,$FF,$69,$B6,$33,$1F,$EC
	.dc.b	$FF,$A2,$79,$19,$A9,$3F,$ED,$3A,$DB,$FE,$DE,$AD,$B9,$9D,$BE,$17
	.dc.b	$C1,$1A,$8F,$F3,$7F,$D1,$BC,$CF,$82,$FF,$9D,$A3,$FD,$6C,$D7,$71
	.dc.b	$A3,$7E,$B4,$8D,$C7,$D2,$F8,$C9,$1F,$ED,$67,$EB,$FA,$F8,$8F,$F6
	.dc.b	$74,$5E,$4F,$F6,$B4,$FC,$5F,$A3,$FF,$30,$D7,$FB,$C5,$AF,$FB,$F1
	.dc.b	$B3,$FE,$C8,$C9,$F0,$5C,$7D,$39,$7C,$5F,$83,$18,$8E,$51,$6B,$94
	.dc.b	$FE,$2B,$E7,$A5,$7F,$BB,$9F,$C7,$C6,$23,$FD,$CD,$16,$7F,$DD,$D3
	.dc.b	$F9,$4A,$FC,$9F,$B1,$2F,$FD,$8D,$19,$7F,$DF,$31,$97,$FD,$4F,$D1
	.dc.b	$EC,$F0,$FF,$DB,$9E,$DB,$E0,$FF,$10,$DA,$FE,$AB,$97,$FB,$A2,$31
	.dc.b	$1F,$EA,$AF,$A3,$7C,$FF,$0B,$93,$C3,$0E,$B1,$FA,$0A,$DC,$E4,$FE
	.dc.b	$C3,$7D,$30,$ED,$5C,$A7,$B6,$1F,$B0,$F6,$A4,$71,$F1,$1D,$BF,$EA
	.dc.b	$B8,$FF,$73,$7B,$5A,$58,$FF,$A3,$2B,$77,$98,$FF,$47,$F8,$7F,$EC
	.dc.b	$9B,$FD,$FF,$89,$FB,$EE,$E9,$8F,$FB,$C6,$32,$F8,$72,$FD,$B3,$67
	.dc.b	$2F,$83,$7F,$EF,$AB,$E8,$9F,$13,$FB,$56,$C7,$C3,$35,$1F,$C8,$FE
	.dc.b	$8B,$F5,$5A,$7F,$A9,$27,$26,$EB,$8B,$DC,$FF,$05,$B8,$8F,$E7,$D3
	.dc.b	$CA,$FC,$73,$96,$DF,$8C,$F0,$78,$3F,$CC,$FE,$8B,$FD,$D9,$8F,$D7
	.dc.b	$F2,$D1,$88,$FF,$52,$B7,$2F,$F5,$C8,$F8,$A7,$C3,$F3,$F0,$65,$F3
	.dc.b	$BF,$E0,$8D,$C7,$C0,$47,$07,$E1,$F8,$BF,$DF,$DE,$4F,$97,$C9,$93
	.dc.b	$CD,$F4,$76,$C8,$EC,$D7,$F4,$3E,$1F,$F3,$C3,$1A,$FF,$63,$CB,$F4
	.dc.b	$92,$C5,$DA,$BF,$89,$DD,$FD,$0B,$C8,$97,$EB,$8C,$B4,$87,$EC,$3C
	.dc.b	$1C,$BF,$22,$2E,$5F,$92,$C1,$5F,$D8,$F5,$FD,$3D,$FD,$DE,$CF,$CF
	.dc.b	$BF,$D4,$8D,$76,$C7,$EB,$4F,$0C,$3F,$BF,$37,$9B,$FC,$9F,$AF,$B3
	.dc.b	$3E,$17,$FE,$DE,$AD,$3F,$DF,$5B,$3B,$8E,$BF,$43,$FE,$02,$C1,$F4
	.dc.b	$4F,$9A,$C3,$6B,$64,$3B,$57,$B4,$FE,$C7,$F4,$5F,$D0,$6F,$CF,$19
	.dc.b	$9E,$11,$FE,$09,$9D,$67,$FE,$68,$E9,$7D,$13,$E7,$3F,$B6,$FC,$73
	.dc.b	$D2,$A7,$F8,$36,$75,$97,$E6,$DA,$FF,$BF,$9B,$97,$F8,$95,$B8,$FC
	.dc.b	$8F,$A2,$61,$CF,$19,$87,$37,$FC,$1B,$3A,$8F,$D4,$7C,$FF,$A1,$FF
	.dc.b	$14,$37,$33,$9E,$07,$D1,$3E,$8B,$FC,$49,$3B,$27,$A3,$FE,$1A,$9D
	.dc.b	$61,$DA,$36,$47,$F5,$11,$A7,$F8,$36,$1E,$0E,$EE,$A4,$8D,$FD,$55
	.dc.b	$77,$87,$F8,$85,$F5,$CF,$86,$7F,$C1,$DF,$41,$F4,$78,$7F,$D3,$7F
	.dc.b	$3F,$FA,$7B,$FF,$4E,$FD,$13,$E2,$FF,$91,$CE,$65,$39,$78,$F8,$3F
	.dc.b	$CA,$30,$E9,$0A,$E1,$18,$79,$57,$19,$D5,$7F,$30,$8C,$4F,$0E,$2F
	.dc.b	$E8,$75,$A3,$FD,$63,$7C,$EF,$86,$1E,$66,$CF,$90,$DD,$72,$BE,$98
	.dc.b	$E6,$7D,$77,$C5,$F8,$2D,$58,$77,$79,$CF,$1C,$7C,$F8,$F8,$27,$CC
	.dc.b	$70,$C7,$C1,$1F,$15,$B1,$D8,$E5,$87,$7C,$3B,$47,$1C,$32,$C3,$AC
	.dc.b	$68,$DE,$1E,$28,$E8,$6A,$FB,$BC,$CF,$2C,$6D,$F8,$27,$C3,$9E,$93
	.dc.b	$F4,$23,$FD,$EF,$1F,$71,$B1,$97,$D9,$5F,$EF,$17,$E3,$FC,$73,$E8
	.dc.b	$B1,$68,$C4,$FF,$BD,$E3,$13,$A9,$CC,$F3,$F8,$5E,$8F,$FB,$C1,$A9
	.dc.b	$A4,$7E,$8F,$D1,$7F,$9C,$FA,$3E,$F4,$71,$33,$E7,$A3,$31,$CC,$CB
	.dc.b	$8A,$FC,$35,$73,$9F,$BA,$73,$DF,$F4,$91,$97,$F4,$D1,$9D,$FF,$C1
	.dc.b	$73,$97,$FC,$1B,$19,$02,$34,$EA,$BC,$38,$79,$E3,$C2,$79,$EF,$C3
	.dc.b	$19,$7C,$31,$9B,$F0,$CE,$5F,$0C,$64,$F0,$DF,$81,$FF,$10,$DF,$83
	.dc.b	$0F,$85,$1F,$10,$F0,$C6,$5A,$67,$0F,$0C,$65,$B1,$9B,$F0,$CE,$5F
	.dc.b	$8B,$E0,$FF,$84,$AF,$E3,$F8,$96,$FC,$B3,$5F,$8D,$17,$14,$D8,$C3
	.dc.b	$6B,$FC,$B1,$2F,$D2,$1E,$DF,$5F,$49,$F8,$9E,$2F,$75,$72,$FF,$1B
	.dc.b	$56,$AF,$C9,$87,$37,$96,$1B,$E2,$FA,$D3,$9C,$7F,$CD,$17,$C5,$EB
	.dc.b	$7C,$0E,$78,$F1,$8F,$C1,$CE,$F1,$FE,$83,$F4,$9F,$97,$9F,$F1,$E3
	.dc.b	$63,$FC,$95,$DC,$FA,$4F,$F4,$25,$68,$F9,$5D,$8B,$E1,$A7,$F9,$B6
	.dc.b	$5F,$31,$A9,$CC,$7E,$53,$CB,$FD,$34,$E6,$3E,$A4,$61,$EC,$7F,$9F
	.dc.b	$4E,$67,$F9,$47,$EE,$79,$E6,$C5,$18,$CF,$FC,$6F,$ED,$79,$A3,$73
	.dc.b	$AB,$EF,$F7,$7D,$30,$FE,$C7,$5D,$31,$4E,$74,$FD,$36,$1C,$70,$FF
	.dc.b	$48,$4E,$5F,$F0,$95,$F1,$8C,$61,$E5,$EE,$FF,$70,$57,$F9,$A5,$B1
	.dc.b	$F3,$B6,$7D,$D7,$CC,$9C,$E5,$E0,$6D,$FC,$91,$F3,$9B,$F5,$BE,$4C
	.dc.b	$60,$EC,$E2,$35,$1F,$2D,$FB,$3F,$85,$FF,$B6,$A3,$FC,$F8,$DC,$BF
	.dc.b	$D8,$BD,$BA,$F8,$29,$5F,$02,$ED,$7F,$94,$F5,$F0,$3F,$E0,$CC,$39
	.dc.b	$67,$FE,$AF,$9C,$BF,$2A,$33,$7F,$83,$9C,$BF,$E3,$BC,$DE,$BF,$C3
	.dc.b	$3E,$26,$56,$7E,$54,$7A,$BA,$F1,$97,$85,$C4,$FC,$FF,$7E,$CF,$F8
	.dc.b	$83,$04,$7C,$2D,$7E,$1C,$DA,$F8,$6E,$1F,$F6,$77,$AE,$FB,$1C,$4F
	.dc.b	$F9,$CE,$31,$3F,$EA,$5F,$C8,$FF,$DC,$6F,$0F,$98,$FD,$3C,$62,$B8
	.dc.b	$47,$F0,$FD,$3F,$8A,$70,$3E,$91,$EE,$E7,$1F,$F7,$34,$E6,$31,$87
	.dc.b	$FA,$FE,$31,$3F,$EC,$9F,$07,$E4,$BC,$2F,$9B,$9C,$3B,$C6,$2F,$FD
	.dc.b	$90,$7F,$47,$8C,$E1,$FC,$9F,$B3,$FD,$83,$7F,$A0,$F4,$71,$EE,$7F
	.dc.b	$4E,$8C,$1F,$EB,$57,$18,$7E,$16,$76,$FE,$C1,$FC,$BF,$C5,$CA,$FF
	.dc.b	$2A,$FD,$95,$AF,$F4,$F4,$27,$1A,$FF,$BD,$BD,$35,$E2,$BF,$AF,$CB
	.dc.b	$F3,$DC,$C7,$F2,$6F,$0F,$FD,$A2,$FC,$A8,$CE,$3F,$CC,$A7,$2F,$CB
	.dc.b	$8C,$D7,$F6,$5F,$17,$F8,$DF,$C0,$FD,$69,$CE,$1F,$CB,$30,$E3,$7F
	.dc.b	$EE,$09,$CE,$0C,$38,$DF,$87,$1F,$E8,$D8,$70,$8E,$5E,$C7,$F4,$9F
	.dc.b	$C2,$BF,$8E,$DF,$F2,$A8,$C4,$F0,$6B,$ED,$6E,$1F,$B3,$F5,$7F,$8E
	.dc.b	$9F,$67,$C4,$7C,$FE,$2F,$F8,$CC,$DA,$FE,$25,$37,$5F,$F0,$24,$7F
	.dc.b	$82,$79,$3F,$B4,$75,$3E,$43,$F3,$55,$AE,$9F,$C5,$3F,$E0,$B9,$D7
	.dc.b	$C1,$FF,$6B,$BE,$37,$C3,$3F,$F0,$CC,$62,$75,$1F,$68,$FF,$0C,$D7
	.dc.b	$27,$F3,$93,$B7,$F9,$8F,$DA,$7F,$3A,$3D,$D5,$63,$24,$CB,$F9,$74
	.dc.b	$65,$FE,$E3,$F8,$6B,$E8,$7F,$BC,$2F,$9A,$B1,$86,$DF,$F7,$09,$D3
	.dc.b	$67,$4F,$EC,$7A,$FE,$3F,$CC,$7E,$8B,$6F,$8D,$18,$C7,$FC,$1C,$DA
	.dc.b	$E0,$7C,$18,$6A,$BE,$53,$FE,$40,$AE,$15,$FB,$8D,$D1,$8C,$3F,$4D
	.dc.b	$F2,$1F,$D1,$3E,$73,$F7,$5D,$88,$C6,$3F,$EC,$46,$E3,$FE,$00,$9D
	.dc.b	$4F,$F7,$1A,$FC,$57,$E1,$9F,$DA,$3E,$6E,$CF,$F6,$26,$DF,$F4,$36
	.dc.b	$E3,$FD,$05,$FF,$17,$7E,$6F,$FA,$4F,$7D,$F9,$3B,$E1,$33,$E0,$7F
	.dc.b	$8D,$67,$31,$8C,$BE,$24,$61,$F8,$D9,$30,$E9,$87,$39,$FE,$65,$F0
	.dc.b	$3F,$E0,$CE,$3F,$F4,$34,$60,$FE,$77,$1F,$CD,$0D,$15,$C9,$FC,$EF
	.dc.b	$E5,$7F,$C0,$FE,$2F,$F3,$A6,$C6,$4F,$F0,$1C,$65,$A7,$9B,$E7,$DF
	.dc.b	$49,$F7,$33,$FF,$47,$7D,$17,$F9,$D9,$4E,$B2,$FF,$4A,$4E,$7E,$91
	.dc.b	$E7,$7F,$9D,$FE,$15,$3C,$6F,$86,$1D,$0F,$F4,$0B,$89,$DC,$63,$1F
	.dc.b	$14,$62,$18,$76,$CB,$C3,$1B,$CA,$BE,$6B,$E3,$8F,$85,$B1,$FE,$90
	.dc.b	$BE,$4F,$8C,$E1,$5C,$B1,$F2,$5F,$18,$EB,$5C,$31,$F2,$56,$AF,$C4
	.dc.b	$73,$70,$FA,$32,$F8,$98,$73,$8D,$63,$E4,$7C,$27,$8F,$0E,$D3,$BC
	.dc.b	$3B,$D6,$F0,$EB,$5A,$BF,$0D,$6B,$1F,$55,$F5,$AD,$57,$6A,$D3,$E2
	.dc.b	$AF,$6E,$3E,$68,$E7,$8F,$84,$8F,$C9,$8D,$3C,$E3,$2D,$35,$97,$8A
	.dc.b	$A4,$77,$3D,$D5,$CB,$2F,$8E,$4C,$BD,$11,$97,$E0,$12,$BC,$F7,$D3
	.dc.b	$1F,$59,$31,$FC,$A5,$F1,$AD,$47,$9B,$2F,$87,$8F,$79,$CE,$5E,$83
	.dc.b	$CD,$87,$37,$E1,$99,$70,$FD,$83,$FD,$BB,$3F,$9E,$6E,$5F,$AB,$6D
	.dc.b	$7F,$B4,$9F,$D5,$F3,$3E,$93,$16,$8C,$4F,$FB,$62,$31,$3A,$9C,$CF
	.dc.b	$81,$DE,$1F,$94,$6E,$5F,$B1,$6C,$FF,$BC,$3E,$93,$FA,$53,$F9,$F8
	.dc.b	$CB,$E4,$BC,$1F,$AC,$9D,$C7,$E7,$67,$78,$79,$23,$15,$BF,$03,$FD
	.dc.b	$D7,$87,$86,$39,$CE,$8F,$D8,$61,$D3,$1F,$44,$E9,$B3,$AB,$E6,$FF
	.dc.b	$08,$C7,$C2,$4A,$F0,$47,$1C,$3C,$0F,$7B,$FE,$15,$87,$79,$CC,$76
	.dc.b	$23,$FC,$7E,$35,$3D,$EB,$A3,$4F,$21,$83,$E4,$1D,$E7,$96,$1E,$88
	.dc.b	$CD,$F6,$79,$3E,$48,$D6,$3F,$2E,$35,$3C,$2B,$9B,$DE,$75,$1E,$A3
	.dc.b	$31,$E1,$3A,$B9,$7F,$01,$5D,$B1,$F9,$11,$B1,$8F,$C0,$7C,$07,$22
	.dc.b	$1A,$9E,$58,$F8,$EB,$F1,$35,$C7,$0F,$2D,$F0,$6B,$C4,$FE,$73,$EF
	.dc.b	$BA,$1F,$61,$FD,$3F,$AB,$FD,$69,$C5,$4F,$F5,$91,$6F,$FD,$D9,$5C
	.dc.b	$1F,$F0,$A5,$F3,$BD,$7C,$0F,$FB,$9A,$EF,$B9,$FE,$48,$C7,$FD,$CF
	.dc.b	$8F,$1C,$51,$8F,$91,$E3,$C5,$FA,$E8,$60,$A6,$3F,$E2,$2D,$FC,$9A
	.dc.b	$7A,$F1,$B9,$1E,$73,$2F,$CF,$39,$7F,$C0,$FF,$8B,$FF,$B0,$1F,$F0
	.dc.b	$5F,$E1,$CF,$8B,$F8,$1D,$C8,$C4,$F0,$F7,$7E,$22,$BE,$4F,$B3,$5F
	.dc.b	$45,$F1,$C3,$9E,$1C,$AB,$FD,$5B,$7C,$36,$F7,$C3,$D8,$7A,$2F,$C2
	.dc.b	$7E,$6F,$53,$D3,$A7,$D5,$AD,$CD,$AE,$17,$FE,$49,$E3,$61,$C6,$30
	.dc.b	$D7,$D0,$78,$67,$8B,$FE,$41,$FB,$9F,$BB,$F6,$FF,$E4,$C8,$E8,$D7
	.dc.b	$47,$F9,$5F,$17,$92,$AE,$3E,$5C,$AD,$4E,$5F,$F1,$CF,$CF,$BE,$1B
	.dc.b	$F2,$1D,$5F,$07,$96,$EF,$F8,$B7,$FC,$4D,$5B,$BE,$37,$A9,$91,$82
	.dc.b	$1E,$EB,$E8,$FF,$9B,$5A,$FF,$B8,$2B,$A3,$FE,$84,$FA,$6F,$D1,$7B
	.dc.b	$2F,$4D,$F1,$38,$6D,$7F,$A0,$EB,$06,$66,$46,$4F,$A3,$EF,$7D,$C2
	.dc.b	$3F,$27,$89,$F3,$58,$8F,$AB,$CB,$F6,$AB,$1C,$FF,$88,$E4,$FC,$3D
	.dc.b	$FE,$53,$5F,$D7,$4F,$FA,$E3,$E9,$3F,$1D,$C1,$FE,$4D,$6E,$5F,$AD
	.dc.b	$AD,$CF,$CC,$F0,$FF,$3B,$5A,$A3,$6B,$71,$C6,$77,$DD,$23,$ED,$C7
	.dc.b	$E6,$39,$A4,$7E,$41,$FC,$B7,$8F,$88,$F6,$F5,$7E,$2F,$3F,$D1,$EF
	.dc.b	$F9,$8B,$DD,$7F,$A8,$F9,$1F,$9D,$3D,$4F,$AA,$BD,$9A,$62,$25,$71
	.dc.b	$9D,$36,$72,$FE,$8B,$17,$E7,$4F,$8F,$EB,$EF,$CA,$FF,$5A,$56,$A7
	.dc.b	$F9,$33,$63,$B4,$ED,$E9,$5C,$DE,$2D,$7F,$63,$97,$B2,$78,$57,$F0
	.dc.b	$43,$A4,$FE,$0B,$C0,$7D,$8F,$ED,$09,$1F,$AC,$FA,$DA,$EE,$2D,$7E
	.dc.b	$1F,$C1,$57,$F0,$7F,$7B,$F8,$6D,$4B,$FC,$94,$65,$FE,$0F,$19,$7F
	.dc.b	$83,$F5,$7F,$16,$9F,$C8,$7B,$9F,$88,$F9,$69,$5F,$BC,$8C,$BF,$EB
	.dc.b	$6E,$A4,$EA,$3F,$61,$3A,$7F,$89,$9F,$90,$8D,$67,$FC,$9A,$1D,$5F
	.dc.b	$ED,$97,$F9,$EF,$D2,$FF,$50,$FC,$12,$46,$0F,$E8,$3D,$BF,$D5,$3B
	.dc.b	$24,$60,$FE,$B7,$F7,$AF,$03,$FA,$5D,$49,$CC,$7F,$50,$9F,$A3,$1F
	.dc.b	$D8,$BE,$6D,$F0,$FF,$D5,$1B,$1D,$38,$B1,$F0,$BC,$59,$1C,$63,$2D
	.dc.b	$21,$FC,$D3,$D4,$57,$F7,$98,$F3,$FC,$9F,$F4,$4F,$9F,$91,$82,$1F
	.dc.b	$AC,$72,$F3,$C9,$3E,$B2,$47,$0E,$1F,$E1,$9E,$3B,$F0,$AF,$67,$DC
	.dc.b	$7F,$42,$F8,$69,$7F,$06,$32,$FF,$44,$8C,$BE,$9C,$D3,$A8,$C3,$70
	.dc.b	$FC,$A7,$C6,$FC,$DF,$3F,$FA,$FF,$CC,$4B,$FA,$B1,$97,$FA,$CC,$65
	.dc.b	$F6,$46,$5A,$70,$3F,$96,$7C,$37,$F8,$23,$E8,$5C,$1F,$D7,$3E,$76
	.dc.b	$57,$F8,$72,$72,$D9,$CB,$FB,$19,$E5,$12,$34,$43,$F4,$BE,$13,$E4
	.dc.b	$BF,$C5,$5F,$1A,$F9,$DF,$F7,$F5,$7F,$00,$FA,$47,$CC,$FF,$4E,$6C
	.dc.b	$7F,$88,$78,$3F,$C0,$B7,$C6,$39,$DF,$29,$FF,$04,$F9,$4B,$E3,$5D
	.dc.b	$70,$DB,$0F,$E5,$51,$AA,$E2,$DE,$EF,$F1,$86,$3D,$EB,$C5,$1F,$E3
	.dc.b	$35,$F4,$4F,$63,$FC,$46,$BE,$26,$98,$89,$3C,$8F,$F1,$E4,$ED,$C1
	.dc.b	$D2,$76,$FF,$02,$E7,$7D,$17,$F9,$22,$1F,$45,$FE,$2B,$B7,$CE,$F9
	.dc.b	$63,$0D,$F8,$4F,$46,$1C,$B6,$FF,$26,$57,$F8,$CE,$31,$87,$4A,$F7
	.dc.b	$61,$D2,$37,$18,$C3,$A4,$7F,$95,$6B,$8D,$F3,$C7,$CC,$F8,$A7,$89
	.dc.b	$F0,$2B,$8B,$FB,$C9,$DD,$F9,$9B,$1B,$38,$9F,$42,$78,$DF,$82,$B8
	.dc.b	$63,$AC,$7D,$17,$EB,$39,$14,$F0,$47,$4A,$E5,$1A,$C7,$D0,$EE,$32
	.dc.b	$F0,$39,$57,$69,$D9,$F0,$6B,$A6,$3E,$AA,$F4,$1E,$5C,$7D,$31,$9C
	.dc.b	$3A,$E1,$CF,$1F,$9D,$7C,$B0,$8F,$EA,$0F,$11,$FA,$AA,$FB,$2D,$BF
	.dc.b	$C3,$39,$E7,$7C,$2F,$18,$FE,$57,$18,$7A,$52,$31,$1C,$67,$83,$FE
	.dc.b	$93,$F6,$AB,$FE,$6C,$32,$E0,$8F,$CF,$E0,$FD,$EF,$07,$DA,$8E,$46
	.dc.b	$78,$BF,$D0,$3B,$FF,$A1,$63,$2F,$FA,$1E,$B8,$C7,$D9,$33,$AF,$EC
	.dc.b	$E3,$33,$C6,$7C,$9F,$4D,$C6,$FF,$D1,$AE,$71,$FF,$56,$39,$C6,$E3
	.dc.b	$F3,$67,$F1,$F8,$F4,$8C,$61,$E7,$AF,$53,$F0,$71,$E4,$F9,$4C,$4F
	.dc.b	$97,$1F,$6B,$CA,$76,$78,$E7,$A1,$C7,$0F,$0D,$70,$8F,$E4,$F1,$8B
	.dc.b	$F0,$B6,$38,$1C,$AB,$A6,$3E,$42,$E1,$E3,$BF,$7D,$7F,$1E,$F9,$FF
	.dc.b	$F0,$63,$F2,$DA,$E6,$3F,$11,$92,$73,$28,$C4,$FF,$1A,$7F,$C3,$AD
	.dc.b	$9F,$F1,$D3,$F5,$FC,$EB,$5C,$79,$DD,$78,$C3,$FA,$33,$AE,$87,$E9
	.dc.b	$2F,$A0,$D5,$F0,$CB,$F9,$66,$1C,$1F,$D3,$37,$1F,$E9,$ED,$7F,$11
	.dc.b	$3C,$5C,$B7,$97,$FA,$5F,$D2,$5B,$FE,$87,$36,$99,$2B,$1E,$4B,$E9
	.dc.b	$3F,$DF,$38,$7F,$3F,$8C,$3F,$03,$D6,$FF,$6E,$FC,$BF,$C2,$9C,$CE
	.dc.b	$EB,$8E,$3F,$D9,$A7,$2D,$8C,$BF,$D4,$E3,$2D,$8E,$39,$FF,$A2,$A9
	.dc.b	$E3,$77,$9C,$C2,$9F,$02,$8D,$9E,$9F,$A4,$DC,$B1,$46,$33,$7C,$17
	.dc.b	$FA,$1E,$F8,$1E,$D9,$F3,$DF,$7B,$E3,$3F,$8E,$6B,$FC,$75,$B1,$F0
	.dc.b	$67,$C3,$83,$E4,$7C,$B3,$FE,$45,$9C,$C6,$34,$F5,$4F,$9F,$C1,$79
	.dc.b	$5F,$3D,$B9,$7F,$B3,$FE,$1F,$13,$23,$C5,$3A,$AF,$A4,$D7,$E8,$35
	.dc.b	$F6,$3B,$BF,$A1,$F1,$3C,$E7,$FD,$B3,$5F,$EF,$D3,$F0,$9F,$49,$FE
	.dc.b	$52,$6F,$A5,$F8,$5D,$FF,$8F,$E1,$EA,$DD,$7B,$C3,$FC,$C8,$D7,$F0
	.dc.b	$4D,$7E,$73,$5F,$B0,$D8,$E7,$7F,$B2,$DF,$E6,$65,$F9,$3A,$ED,$87
	.dc.b	$FB,$4B,$6F,$46,$3F,$E7,$3D,$1F,$45,$F2,$CF,$99,$95,$D7,$F4,$13
	.dc.b	$A7,$E5,$CF,$DF,$8C,$57,$CC,$8F,$C1,$63,$FC,$2A,$3F,$40,$D7,$DB
	.dc.b	$DB,$A3,$7E,$14,$C3,$F1,$15,$FD,$1A,$BE,$1C,$F4,$64,$E9,$A7,$EB
	.dc.b	$7E,$93,$FD,$D1,$1F,$91,$6C,$67,$E1,$BF,$2D,$18,$8F,$E1,$78,$FF
	.dc.b	$1A,$7F,$91,$3F,$C7,$3D,$17,$D0,$FF,$17,$C3,$F9,$24,$EF,$BB,$E0
	.dc.b	$DF,$29,$FE,$5B,$3F,$CB,$27,$33,$FD,$76,$3F,$93,$F0,$3E,$91,$F4
	.dc.b	$5C,$E7,$31,$8C,$3F,$C9,$DD,$5F,$B4,$C3,$EF,$71,$FF,$9C,$E7,$3C
	.dc.b	$FF,$C0,$5F,$F2,$A7,$37,$EB,$7D,$1F,$C9,$5F,$06,$CE,$AF,$94,$FE
	.dc.b	$73,$E7,$BF,$15,$27,$F8,$F7,$A6,$FA,$1C,$C4,$9E,$E6,$CF,$3C,$ED
	.dc.b	$CC,$7C,$0F,$A4,$C5,$B8,$3F,$7A,$23,$2F,$FA,$0B,$C1,$FE,$27,$7F
	.dc.b	$E6,$F8,$D6,$3C,$CF,$EB,$AE,$23,$FC,$58,$DB,$FF,$46,$5F,$81,$FF
	.dc.b	$40,$46,$4E,$DF,$01,$CE,$BF,$CF,$9E,$2F,$F2,$E8,$FE,$B1,$8B,$E5
	.dc.b	$7F,$AD,$7D,$27,$F2,$AA,$FE,$67,$18,$BD,$B5,$E7,$5C,$5E,$B5,$FC
	.dc.b	$7B,$E0,$BE,$D9,$FD,$2F,$DF,$7C,$66,$2B,$EB,$4D,$D3,$E4,$47,$F3
	.dc.b	$FE,$55,$F0,$8D,$9F,$DE,$7A,$3F,$2C,$7E,$6F,$E8,$7F,$96,$47,$F8
	.dc.b	$BE,$32,$FF,$81,$1F,$CC,$11,$C1,$FE,$9F,$EC,$FD,$1F,$77,$0C,$7C
	.dc.b	$62,$33,$3F,$8D,$F8,$3F,$CB,$9E,$DA,$B1,$92,$57,$2C,$BD,$73,$A8
	.dc.b	$EB,$C0,$F9,$57,$C6,$7F,$82,$D7,$C9,$D8,$C3,$71,$FB,$73,$A6,$CE
	.dc.b	$9F,$F0,$F7,$6F,$EE,$CE,$9F,$27,$63,$27,$F8,$A2,$3F,$78,$47,$07
	.dc.b	$C5,$FA,$4F,$D9,$65,$F1,$0F,$F1,$64,$FF,$0B,$8C,$F4,$7F,$5C,$D5
	.dc.b	$D8,$F8,$EF,$D8,$FD,$07,$EC,$1B,$97,$EB,$1B,$97,$FB,$D1,$B8,$73
	.dc.b	$E5,$69,$FC,$1B,$E7,$3F,$18,$DC,$3F,$85,$CE,$9F,$E0,$53,$A6,$9F
	.dc.b	$ED,$AF,$A4,$FD,$46,$3F,$EE,$76,$BF,$E0,$F6,$A6,$9F,$BA,$8F,$F0
	.dc.b	$A7,$BE,$FA,$47,$C5,$FF,$2F,$AD,$5F,$FA,$72,$73,$7F,$E2,$FF,$2F
	.dc.b	$FD,$34,$DF,$82,$FF,$27,$E3,$CF,$1F,$F1,$85,$6B,$0F,$C9,$FA,$68
	.dc.b	$C4,$7F,$8B,$5A,$FF,$86,$5A,$FF,$88,$DB,$1C,$7D,$8F,$E8,$E4,$C7
	.dc.b	$FA,$63,$6B,$F4,$3E,$57,$F8,$6F,$C8,$F1,$CF,$F3,$67,$FA,$67,$D0
	.dc.b	$BE,$8F,$F4,$D6,$F8,$47,$C4,$7F,$8E,$61,$5B,$F0,$BF,$D6,$D3,$9F
	.dc.b	$A4,$78,$BF,$E2,$4C,$38,$E1,$FE,$41,$C3,$83,$D6,$FA,$46,$27,$73
	.dc.b	$DB,$1F,$1C,$63,$0E,$BA,$F8,$2B,$A5,$6B,$0E,$6D,$8C,$E5,$B5,$7C
	.dc.b	$9F,$11,$A2,$9E,$18,$E1,$8F,$96,$FD,$D9,$7B,$5E,$F3,$AC,$3C,$17
	.dc.b	$C4,$E7,$87,$7C,$BD,$B8,$F3,$9F,$04,$62,$B8,$D6,$F0,$EB,$5E,$4B
	.dc.b	$E0,$F5,$BF,$89,$8F,$8E,$39,$D7,$09,$E9,$8F,$C1,$C3,$A5,$68,$87
	.dc.b	$3A,$F6,$E5,$EB,$A9,$3E,$1C,$BC,$B8,$F7,$8E,$07,$2C,$BC,$B8,$F6
	.dc.b	$8C,$92,$33,$3C,$5F,$59,$2B,$AE,$1E,$0C,$BD,$37,$E1,$C7,$D5,$87
	.dc.b	$1B,$F9,$F8,$71,$9C,$E1,$C7,$2F,$84,$4C,$3F,$D2,$59,$79,$6F,$84
	.dc.b	$75,$3E,$BE,$3D,$70,$E5,$1F,$1A,$B8,$63,$E0,$3F,$4F,$1F,$A5,$3F
	.dc.b	$00,$5B,$FC,$FB,$63,$6F,$E8,$3A,$9F,$45,$C6,$3F,$4D,$18,$71,$48
	.dc.b	$C4,$75,$6B,$CE,$7F,$2F,$1C,$74,$FC,$7B,$73,$FD,$D7,$D2,$7F,$54
	.dc.b	$BD,$DC,$7F,$1D,$5B,$7F,$D7,$13,$B7,$33,$BB,$EA,$FE,$B6,$75,$18
	.dc.b	$9D,$91,$FD,$EC,$6A,$FD,$13,$96,$CE,$BC,$3F,$E2,$0F,$32,$46,$8F
	.dc.b	$5C,$69,$C4,$FC,$AB,$F3,$46,$A2,$99,$3C,$07,$BA,$79,$39,$AE,$ED
	.dc.b	$9E,$4F,$03,$F9,$1E,$1F,$1D,$B7,$F0,$EF,$9C,$60,$E1,$8F,$DB,$9D
	.dc.b	$47,$90,$D6,$1E,$05,$8F,$C7,$7C,$44,$37,$8F,$92,$78,$5F,$59,$F3
	.dc.b	$57,$17,$67,$18,$D4,$FC,$F8,$F1,$46,$2B,$E5,$C7,$8B,$0F,$5C,$E6
	.dc.b	$39,$63,$F0,$70,$ED,$39,$9E,$E4,$8F,$57,$13,$A5,$F2,$9F,$E6,$0E
	.dc.b	$15,$B9,$3F,$DE,$51,$98,$C7,$A3,$FE,$F0,$7F,$B5,$DF,$27,$8F,$77
	.dc.b	$FB,$6E,$31,$1F,$B5,$C7,$C1,$8F,$EB,$E1,$18,$7F,$64,$D7,$84,$BE
	.dc.b	$93,$84,$FA,$5B,$19,$9E,$2D,$78,$C7,$A6,$78,$47,$9B,$1F,$7C,$62
	.dc.b	$3F,$CA,$FD,$4F,$A4,$7C,$4F,$C1,$97,$D2,$3E,$87,$E2,$5E,$EA,$DF
	.dc.b	$0B,$EC,$7C,$CB,$E0,$E0,$E7,$7F,$0D,$A7,$C5,$8D,$BF,$C5,$5A,$FF
	.dc.b	$80,$FC,$6F,$F3,$EF,$D1,$FD,$5E,$1F,$F0,$47,$A9,$7A,$3F,$D2,$97
	.dc.b	$FE,$3C,$AF,$B5,$7C,$23,$FD,$87,$3A,$7F,$C2,$ED,$E2,$FF,$24,$57
	.dc.b	$0F,$21,$F4,$9D,$FE,$83,$FD,$A1,$87,$F8,$79,$B7,$FE,$59,$AD,$CF
	.dc.b	$FB,$02,$B8,$79,$92,$30,$7C,$3F,$9E,$7C,$2F,$E3,$FC,$26,$7F,$8F
	.dc.b	$C9,$CD,$F9,$69,$CB,$F5,$6F,$8E,$1C,$F0,$F9,$91,$88,$FE,$7C,$D7
	.dc.b	$FA,$43,$5F,$F1,$73,$BB,$FC,$F3,$6B,$35,$FA,$3F,$45,$EB,$FF,$B8
	.dc.b	$70,$FF,$45,$FC,$AE,$7D,$37,$C4,$53,$11,$27,$FD,$23,$19,$78,$9F
	.dc.b	$53,$DB,$FB,$44,$7F,$CE,$BF,$49,$5F,$AF,$17,$5F,$F3,$F4,$FC,$3F
	.dc.b	$03,$F3,$1F,$41,$F8,$A9,$FD,$C3,$76,$F7,$CE,$F0,$FD,$A6,$FF,$B8
	.dc.b	$7E,$2C,$7C,$A2,$3F,$5D,$E0,$FF,$A6,$E9,$F5,$1F,$91,$E3,$93,$F8
	.dc.b	$E7,$F2,$3E,$A7,$F5,$BC,$BE,$DC,$FF,$0F,$E8,$FC,$4E,$7F,$C4,$BA
	.dc.b	$FF,$CC,$58,$72,$AF,$F6,$16,$1C,$6B,$F8,$BF,$1B,$E9,$7F,$D3,$B5
	.dc.b	$FE,$F6,$5E,$D7,$D7,$6C,$67,$D7,$C1,$99,$91,$93,$F7,$3A,$3E,$91
	.dc.b	$F4,$BC,$FE,$02,$56,$A7,$2D,$9F,$E3,$D5,$C5,$B3,$A9,$FF,$0C,$4E
	.dc.b	$5F,$F0,$93,$5F,$F7,$4B,$4F,$8E,$5C,$FE,$0F,$C0,$F2,$7E,$47,$80
	.dc.b	$FA,$2F,$DB,$12,$FF,$89,$F8,$5F,$EC,$CF,$A5,$C4,$7E,$E6,$6E,$3F
	.dc.b	$B9,$9F,$AF,$F0,$1F,$C2,$E6,$57,$FB,$1A,$32,$FF,$0A,$8C,$BF,$5A
	.dc.b	$3F,$69,$EF,$FE,$4B,$E6,$24,$FE,$FE,$32,$FE,$41,$FD,$B9,$1C,$1B
	.dc.b	$21,$F8,$FD,$BF,$9E,$79,$EF,$A5,$FE,$86,$49,$FE,$86,$FF,$46,$DA
	.dc.b	$47,$F4,$66,$BF,$D5,$7E,$91,$E0,$7F,$4E,$6C,$7F,$54,$6B,$B8,$D6
	.dc.b	$92,$34,$D7,$FB,$07,$CE,$BE,$17,$FA,$CB,$63,$FA,$FB,$5D,$C6,$8F
	.dc.b	$EB,$11,$9E,$0F,$D3,$FA,$48,$C6,$5F,$D6,$9F,$EF,$97,$FD,$42,$59
	.dc.b	$3E,$4F,$FA,$27,$D0,$CA,$D4,$E5,$B3,$F0,$48,$FC,$16,$BF,$EE,$A8
	.dc.b	$E0,$D8,$FD,$EF,$D2,$7A,$3E,$93,$FA,$37,$C2,$CA,$F2,$64,$9D,$47
	.dc.b	$C2,$7E,$29,$1C,$1C,$1D,$AD,$3F,$A7,$74,$FF,$7D,$CB,$FD,$07,$1F
	.dc.b	$C3,$3C,$2F,$F7,$57,$CB,$FF,$63,$35,$E7,$CC,$3E,$E3,$F7,$C8,$E0
	.dc.b	$F8,$AE,$DF,$E2,$3F,$0F,$FE,$0B,$FA,$47,$07,$F6,$3F,$9F,$95,$F4
	.dc.b	$67,$2D,$9F,$F0,$CC,$62,$3F,$C4,$8D,$7F,$DF,$DC,$48,$C4,$7F,$85
	.dc.b	$9E,$19,$FE,$09,$BC,$79,$BF,$EE,$DF,$23,$FE,$37,$F9,$47,$CE,$FF
	.dc.b	$05,$C5,$E4,$3E,$91,$F3,$1F,$D6,$DB,$3F,$E1,$8B,$E2,$F3,$BF,$F0
	.dc.b	$B4,$62,$3F,$D5,$0D,$7F,$D6,$4D,$7F,$C9,$61,$D6,$9F,$D0,$1B,$9F
	.dc.b	$F9,$41,$B3,$E6,$FA,$0F,$F0,$0B,$FE,$27,$FA,$47,$AD,$FE,$33,$5F
	.dc.b	$13,$4C,$44,$9F,$F2,$74,$F0,$6C,$7F,$8B,$E3,$11,$FE,$C1,$EA,$7D
	.dc.b	$27,$E3,$61,$F4,$7F,$E3,$B5,$4F,$9F,$F6,$CE,$63,$13,$FE,$6A,$8E
	.dc.b	$17,$D3,$5C,$7C,$2F,$C4,$8C,$53,$E9,$39,$63,$F2,$1B,$1F,$E7,$47
	.dc.b	$9E,$5E,$56,$E1,$9C,$3E,$7E,$5E,$2B,$E9,$18,$C3,$A3,$E1,$C3,$93
	.dc.b	$6F,$95,$7F,$A0,$AF,$96,$1E,$28,$E6,$7F,$A1,$B1,$F4,$C7,$8A,$32
	.dc.b	$7C,$6C,$7B,$BC,$DA,$FF,$19,$C7,$C2,$F3,$6B,$C4,$F1,$94,$EB,$5E
	.dc.b	$0C,$7D,$11,$F5,$9F,$73,$F3,$8F,$01,$4F,$06,$36,$79,$46,$8F,$4E
	.dc.b	$3F,$02,$BA,$E3,$ED,$3E,$11,$1C,$9E,$29,$FC,$F9,$D8,$FC,$66,$1F
	.dc.b	$8B,$73,$7F,$80,$6F,$4B,$E1,$FF,$D4,$92,$8C,$4F,$E5,$5A,$F0,$9F
	.dc.b	$F4,$42,$F5,$E9,$1F,$F4,$63,$71,$FC,$D7,$0F,$F0,$0E,$2F,$9B,$5F
	.dc.b	$EC,$0A,$F9,$F3,$98,$FB,$66,$4F,$BA,$E6,$BF,$49,$19,$7A,$F2,$FF
	.dc.b	$8F,$1C,$C7,$94,$E3,$19,$D3,$E3,$C6,$67,$8E,$1E,$48,$C4,$FF,$0A
	.dc.b	$BF,$03,$C3,$0F,$3E,$1C,$EB,$E8,$C6,$2F,$93,$63,$D6,$74,$3F,$88
	.dc.b	$4F,$17,$F8,$BC,$6F,$0F,$4B,$6F,$C1,$3E,$AB,$EA,$D3,$AC,$ED,$F1
	.dc.b	$99,$AE,$6D,$AF,$67,$D0,$7F,$8F,$2E,$9F,$EC,$8E,$77,$E4,$0F,$9F
	.dc.b	$D6,$1F,$C9,$EE,$BF,$CB,$5B,$A1,$DD,$F2,$AC,$C7,$29,$FE,$8B,$DD
	.dc.b	$7A,$4F,$0E,$97,$D2,$E6,$26,$9F,$D1,$6B,$6F,$9E,$B6,$E6,$77,$9F
	.dc.b	$F7,$78,$FF,$3B,$FD,$2F,$08,$C1,$FD,$16,$F1,$CF,$FE,$84,$D5,$58
	.dc.b	$BF,$F7,$BF,$D2,$FF,$B0,$9F,$F0,$F3,$70,$FF,$42,$FC,$CF,$C6,$F0
	.dc.b	$93,$9C,$7F,$C5,$2D,$BE,$67,$CC,$F5,$9E,$EB,$0E,$55,$BC,$58,$F4
	.dc.b	$C3,$E9,$F8,$FF,$ED,$4F,$27,$71,$80,$F8,$FE,$97,$95,$1E,$99,$BD
	.dc.b	$FF,$F4,$6D,$FF,$8E,$67,$64,$7C,$D1,$AC,$7E,$4D,$F3,$9E,$C7,$93
	.dc.b	$E4,$7C,$2F,$33,$18,$FC,$5F,$09,$E5,$D3,$E1,$6F,$FE,$D8,$AF,$F2
	.dc.b	$54,$E9,$B3,$B7,$FD,$17,$3B,$C7,$C7,$5B,$6D,$7C,$3F,$89,$E1,$1F
	.dc.b	$80,$DB,$FD,$F7,$1F,$5F,$E9,$7F,$CA,$D8,$FF,$96,$5B,$5F,$86,$DF
	.dc.b	$E3,$78,$1B,$F4,$A5,$ED,$F8,$B1,$8A,$F8,$B7,$CB,$93,$FC,$89,$BB
	.dc.b	$2D,$46,$1F,$89,$87,$C5,$ED,$B9,$7C,$27,$0D,$3E,$2C,$63,$1E,$FD
	.dc.b	$BE,$B3,$18,$7F,$9E,$32,$FD,$A5,$FB,$73,$F8,$07,$DC,$F3,$AC,$7C
	.dc.b	$1C,$3E,$21,$83,$D6,$7E,$77,$E7,$3F,$DB,$B1,$FE,$7F,$6C,$7C,$D9
	.dc.b	$FD,$9B,$F8,$68,$CE,$1F,$EF,$08,$FF,$1A,$67,$FC,$20,$FE,$1B,$4F
	.dc.b	$A5,$E3,$5B,$F0,$3F,$89,$62,$C3,$F7,$27,$27,$13,$BC,$39,$DF,$29
	.dc.b	$FE,$75,$1D,$DD,$98,$2D,$30,$F1,$78,$1B,$8F,$C1,$74,$78,$6B,$7C
	.dc.b	$6E,$27,$9D,$FA,$6B,$FD,$13,$6B,$73,$A7,$FD,$C7,$7C,$EB,$F2,$93
	.dc.b	$97,$F2,$D2,$F9,$DF,$DE,$1F,$C1,$FE,$41,$F4,$98,$9F,$19,$B2,$CE
	.dc.b	$E3,$55,$FE,$FC,$FA,$4E,$35,$F0,$38,$11,$89,$FE,$4B,$E1,$7F,$0A
	.dc.b	$CF,$FA,$0D,$6B,$0B,$7B,$86,$1F,$4A,$30,$FD,$7B,$E4,$F7,$8C,$E1
	.dc.b	$FE,$AB,$6D,$75,$9F,$EA,$5D,$BF,$C5,$5E,$18,$7F,$4C,$87,$C9,$FF
	.dc.b	$4E,$F9,$1F,$E4,$D9,$7F,$20,$8C,$5E,$DA,$F4,$6E,$5F,$43,$E9,$3F
	.dc.b	$8C,$1F,$C4,$BE,$22,$D7,$D8,$8B,$3F,$C0,$FC,$15,$63,$4F,$CB,$7D
	.dc.b	$27,$F6,$23,$F3,$7F,$43,$FC,$8A,$FF,$2F,$D4,$9D,$65,$F9,$09,$FF
	.dc.b	$4B,$47,$0D,$B3,$1C,$22,$E1,$F3,$A3,$F2,$D8,$75,$8F,$0D,$FE,$3B
	.dc.b	$DF,$FE,$64,$7F,$2C,$F8,$BB,$19,$25,$FA,$E3,$2F,$E0,$3E,$51,$F1
	.dc.b	$3F,$DE,$97,$CF,$58,$C3,$6F,$FC,$11,$3A,$6C,$E9,$F7,$6B,$F6,$EF
	.dc.b	$E1,$F3,$75,$F8,$EF,$F0,$4B,$2B,$6E,$5A,$49,$E2,$F8,$27,$84,$7F
	.dc.b	$37,$E9,$7C,$57,$F6,$3F,$9E,$FD,$43,$6F,$F7,$FE,$4A,$B7,$1F,$AA
	.dc.b	$AD,$B4,$FE,$9F,$87,$F5,$DF,$8A,$FE,$AE,$D8,$C9,$FB,$E8,$CB,$4F
	.dc.b	$BD,$F4,$7F,$A0,$C8,$FE,$8B,$DA,$71,$FE,$16,$B7,$DB,$FE,$2E,$AE
	.dc.b	$06,$E7,$57,$FC,$8A,$3F,$57,$E7,$BC,$8F,$F3,$D6,$1F,$CD,$7D,$BF
	.dc.b	$F1,$CE,$4F,$A3,$7C,$47,$F3,$B9,$95,$B3,$F9,$1D,$6E,$73,$97,$F3
	.dc.b	$BF,$57,$FA,$11,$FD,$E3,$C9,$B7,$FD,$0E,$2C,$FE,$9E,$6E,$BF,$D0
	.dc.b	$7E,$81,$F4,$5F,$CB,$EB,$FD,$15,$F0,$BF,$EF,$89,$CF,$9B,$FE,$85
	.dc.b	$8F,$F7,$DF,$D1,$BC,$1F,$D1,$7C,$0F,$FB,$CB,$0F,$E8,$98,$70,$79
	.dc.b	$98,$C7,$B4,$EE,$13,$B8,$C5,$F5,$D7,$C1,$5C,$72,$AF,$0C,$39,$E5
	.dc.b	$E7,$6C,$7F,$55,$BE,$4F,$88,$E1,$5C,$B1,$F0,$5F,$18,$EB,$5C,$31
	.dc.b	$F3,$56,$AF,$C4,$73,$70,$A7,$85,$F3,$8D,$61,$E4,$7C,$27,$8E,$76
	.dc.b	$F4,$BE,$F3,$BB,$EB,$5A,$AE,$35,$AC,$3D,$57,$D6,$B5,$7E,$42,$46
	.dc.b	$72,$FA,$F8,$F9,$70,$E5,$8F,$9B,$1F,$83,$5A,$8F,$25,$6C,$8F,$CD
	.dc.b	$8D,$3E,$48,$CB,$4D,$65,$E6,$A9,$1D,$CF,$85,$5C,$B1,$F3,$CE,$B0
	.dc.b	$FC,$1C,$E7,$1F,$88,$49,$F3,$E1,$BC,$7C,$38,$7C,$8B,$E1,$59,$C7
	.dc.b	$D6,$79,$0F,$16,$3F,$08,$97,$F9,$EA,$E3,$3E,$73,$FD,$FB,$7F,$9D
	.dc.b	$6E,$3F,$A9,$6E,$1F,$ED,$C7,$F3,$BD,$0F,$A3,$C5,$A3,$13,$FE,$E2
	.dc.b	$8C,$4E,$A7,$33,$E0,$77,$97,$E5,$1B,$97,$ED,$DB,$3F,$E3,$5F,$A3
	.dc.b	$FE,$8E,$FE,$7E,$32,$F9,$2F,$07,$EB,$27,$71,$F9,$D9,$DE,$1E,$48
	.dc.b	$C5,$6F,$C2,$FF,$73,$E1,$E1,$8E,$73,$A3,$F6,$18,$74,$C7,$D1,$3A
	.dc.b	$6C,$EA,$F9,$BF,$C2,$B1,$F0,$92,$BC,$11,$C7,$0F,$03,$DE,$FF,$86
	.dc.b	$61,$DE,$73,$1D,$88,$FF,$96,$65,$3B,$7C,$15,$D5,$A7,$94,$C1,$F2
	.dc.b	$4F,$04,$F3,$9E,$51,$9A,$EE,$F2,$7C,$B8,$7B,$A3,$55,$E6,$23,$F6
	.dc.b	$63,$58,$FD,$1A,$E7,$3C,$0D,$1E,$48,$D5,$72,$8F,$6D,$72,$C7,$E5
	.dc.b	$47,$A0,$E2,$53,$A1,$30,$F5,$DF,$48,$F5,$5F,$59,$CD,$74,$C3,$CB
	.dc.b	$1F,$2F,$9D,$D4,$F9,$AF,$A1,$C5,$FE,$29,$9C,$C6,$38,$7F,$D1,$91
	.dc.b	$8C,$FF,$D9,$BC,$39,$7F,$DE,$38,$73,$7F,$D0,$F3,$B7,$FD,$C7,$19
	.dc.b	$7F,$C8,$D3,$C6,$7C,$7F,$47,$68,$D3,$B7,$3B,$E8,$DF,$0B,$9C,$7B
	.dc.b	$DF,$06,$DF,$0A,$DC,$7F,$92,$FE,$89,$F4,$7E,$98,$46,$1D,$47,$83
	.dc.b	$83,$F6,$3B,$FF,$AD,$6B,$55,$FE,$AE,$6C,$66,$B8,$72,$7C,$5F,$0D
	.dc.b	$E7,$3E,$8B,$E4,$7B,$0C,$FF,$CE,$F9,$FF,$AA,$1A,$FF,$9A,$E7,$95
	.dc.b	$71,$8D,$7A,$5C,$AB,$79,$FF,$98,$71,$EF,$87,$87,$C7,$7D,$1F,$83
	.dc.b	$E7,$1E,$5D,$35,$13,$3F,$F5,$CC,$E9,$B3,$97,$FD,$9B,$F4,$3F,$EF
	.dc.b	$77,$FC,$EC,$D7,$3F,$05,$F8,$5F,$5D,$EC,$7F,$AD,$27,$73,$F4,$6B
	.dc.b	$86,$BF,$EE,$16,$F6,$7F,$A9,$6F,$85,$F4,$38,$90,$FF,$59,$BA,$AF
	.dc.b	$F4,$16,$0F,$49,$F2,$3E,$35,$F4,$74,$C4,$4C,$3F,$CE,$F1,$97,$FC
	.dc.b	$F3,$1F,$3F,$E6,$A4,$7D,$77,$EA,$FD,$1E,$23,$FC,$AF,$9F,$FB,$CF
	.dc.b	$9F,$EF,$56,$73,$FA,$DC,$EC,$3E,$17,$85,$31,$F9,$6F,$E5,$7E,$8D
	.dc.b	$F4,$79,$9F,$A3,$37,$0F,$F5,$04,$2B,$6F,$22,$1C,$0F,$F5,$63,$A7
	.dc.b	$CE,$32,$FC,$A7,$1F,$E3,$49,$1F,$E8,$4C,$BF,$47,$A7,$B7,$A7,$F0
	.dc.b	$D3,$CA,$F8,$BF,$9B,$BF,$CC,$F5,$FE,$0F,$E8,$DE,$B5,$F6,$A5,$8F
	.dc.b	$F7,$6C,$E9,$C6,$BF,$72,$31,$1A,$9E,$2F,$0A,$FE,$29,$4F,$A3,$7D
	.dc.b	$0E,$1B,$5F,$EF,$D6,$BF,$EF,$CF,$7F,$FA,$93,$F1,$2B,$95,$71,$24
	.dc.b	$7F,$BC,$8A,$6E,$B9,$CF,$FB,$4C,$E3,$1B,$3F,$5B,$F4,$1F,$B3,$23
	.dc.b	$FE,$B0,$EF,$7D,$1E,$23,$E6,$77,$AB,$17,$FB,$FB,$DD,$7D,$5F,$7F
	.dc.b	$F8,$25,$48,$C1,$0F,$AD,$C5,$CD,$FE,$21,$1F,$C3,$A7,$F6,$9E,$F4
	.dc.b	$9F,$DE,$BF,$93,$D3,$11,$FC,$0A,$2E,$3F,$92,$E6,$F8,$3C,$BF,$DE
	.dc.b	$FC,$0F,$F6,$7D,$F2,$21,$F9,$03,$F3,$18,$7F,$41,$FA,$29,$1F,$CF
	.dc.b	$1F,$E4,$FF,$0F,$FD,$27,$1F,$EA,$DE,$67,$E8,$1B,$8F,$F4,$93,$FA
	.dc.b	$67,$D0,$3B,$E9,$31,$FE,$AB,$19,$7F,$39,$F6,$BC,$38,$24,$69,$FE
	.dc.b	$AD,$F3,$6F,$85,$A4,$9F,$EB,$F1,$96,$E7,$C0,$D3,$FA,$EF,$4D,$3B
	.dc.b	$8C,$3F,$DD,$1F,$2F,$C8,$FF,$3D,$FA,$09,$18,$21,$CF,$34,$FC,$03
	.dc.b	$FA,$3F,$D1,$D9,$DE,$FF,$DF,$6B,$71,$FB,$B9,$F6,$FD,$1F,$F4,$5F
	.dc.b	$83,$93,$EC,$8C,$B4,$87,$F0,$8F,$45,$D6,$EF,$DC,$FB,$2A,$E5,$F9
	.dc.b	$79,$FA,$BE,$FF,$F5,$9F,$95,$97,$F5,$23,$2D,$21,$DF,$C8,$7C,$2E
	.dc.b	$A3,$E1,$4D,$C3,$F9,$6C,$FF,$83,$7E,$8D,$C3,$FD,$87,$E7,$26,$1F
	.dc.b	$E1,$69,$CB,$67,$81,$1F,$DA,$7C,$95,$7E,$D4,$F0,$6C,$FF,$6F,$9F
	.dc.b	$F1,$27,$C6,$BE,$73,$F8,$5E,$5F,$D5,$3E,$8D,$F2,$FF,$D9,$5B,$8F
	.dc.b	$F1,$EE,$FF,$E4,$58,$71,$8E,$78,$72,$9C,$B6,$13,$B7,$90,$F5,$DF
	.dc.b	$3D,$FE,$09,$8F,$E7,$F3,$B6,$D7,$38,$C6,$5F,$DA,$AB,$83,$FD,$3B
	.dc.b	$E8,$BF,$B5,$DF,$60,$E1,$FD,$E3,$63,$FB,$9F,$D0,$3E,$BB,$FC,$61
	.dc.b	$39,$CF,$FD,$00,$DB,$FE,$84,$DC,$D7,$C6,$7C,$1D,$CF,$A3,$65,$FD
	.dc.b	$27,$E8,$FF,$C7,$F5,$E4,$FA,$0F,$F4,$64,$63,$24,$63,$E2,$3F,$D2
	.dc.b	$18,$F3,$C3,$94,$63,$1E,$B8,$B1,$E5,$87,$18,$F3,$B5,$FD,$EC,$F0
	.dc.b	$9F,$1E,$1D,$67,$95,$FC,$0B,$E1,$8F,$92,$7C,$27,$F6,$3A,$E1,$7C
	.dc.b	$B0,$E9,$87,$1A,$D9,$EF,$2D,$70,$8F,$56,$3A,$C3,$B5,$78,$8F,$49
	.dc.b	$C2,$FA,$BE,$16,$BD,$4E,$05,$C3,$C9,$5C,$4E,$D8,$FB,$0F,$89,$8F
	.dc.b	$8E,$BA,$4E,$98,$74,$8D,$3F,$2A,$33,$5F,$42,$3B,$1B,$BF,$AA,$FF
	.dc.b	$78,$AF,$9C,$D8,$CB,$F0,$8F,$EF,$A7,$F3,$2E,$77,$C4,$E2,$30,$DA
	.dc.b	$FE,$1E,$D9,$DC,$EC,$FF,$3D,$17,$D1,$C3,$83,$A3,$83,$EB,$F1,$7F
	.dc.b	$02,$E0,$FB,$33,$FE,$21,$9C,$F2,$A7,$33,$DB,$3F,$CE,$DF,$18,$FA
	.dc.b	$67,$13,$EA,$B9,$8F,$D1,$C6,$5E,$35,$D6,$38,$9D,$77,$FF,$79,$57
	.dc.b	$17,$E3,$E1,$E4,$C3,$C6,$DA,$E2,$DC,$3B,$C7,$33,$F8,$DE,$1D,$E3
	.dc.b	$33,$BB,$EB,$5E,$8A,$D4,$62,$BC,$F7,$E0,$AD,$36,$33,$1E,$23,$25
	.dc.b	$72,$78,$1C,$B4,$E3,$AF,$F9,$AA,$3E,$35,$7C,$2B,$FF,$1D,$5F,$8E
	.dc.b	$30,$FF,$21,$9D,$9E,$E7,$F9,$26,$2F,$A0,$E3,$3F,$CF,$1B,$C5,$FC
	.dc.b	$73,$E7,$B3,$3F,$CE,$A6,$D7,$1B,$E7,$F4,$B8,$9C,$B6,$32,$78,$23
	.dc.b	$2F,$FB,$CB,$E8,$DF,$29,$E2,$F2,$91,$89,$FF,$3E,$72,$66,$31,$97
	.dc.b	$F8,$5D,$BB,$FE,$1F,$DF,$FC,$A7,$C7,$6F,$2F,$6C,$63,$1F,$04,$61
	.dc.b	$FF,$7D,$98,$3F,$AA,$C6,$F0,$ED,$1E,$13,$E3,$F9,$1F,$EE,$8F,$97
	.dc.b	$C4,$7A,$CE,$9E,$3E,$2B,$FD,$21,$1F,$E1,$13,$47,$63,$0F,$67,$11
	.dc.b	$F0,$FA,$BF,$DC,$71,$FD,$A3,$6F,$E8,$BF,$03,$E2,$F1,$FF,$A3,$CE
	.dc.b	$6B,$78,$7F,$8E,$DB,$B7,$F9,$FF,$EF,$1F,$01,$D7,$1F,$C0,$CA,$73
	.dc.b	$87,$39,$CC,$63,$0E,$91,$8B,$F3,$79,$52,$FF,$D6,$5F,$0F,$6F,$C5
	.dc.b	$3F,$3F,$E2,$B1,$97,$D7,$B4,$61,$FF,$6A,$18,$3D,$93,$C5,$F7,$79
	.dc.b	$FF,$14,$F8,$5F,$11,$63,$27,$C3,$F3,$71,$8F,$C3,$7E,$61,$F8,$4F
	.dc.b	$04,$E2,$7F,$03,$18,$7E,$67,$A7,$F4,$FB,$BE,$A3,$63,$E8,$47,$E6
	.dc.b	$7E,$8F,$F4,$98,$7D,$18,$C7,$6F,$DB,$F8,$FF,$06,$17,$0E,$EF,$EB
	.dc.b	$BF,$1D,$7D,$0D,$3F,$67,$D2,$FB,$4C,$5F,$F8,$76,$B6,$FF,$8B,$38
	.dc.b	$DF,$55,$E9,$D7,$FD,$35,$5A,$7F,$D9,$11,$99,$FF,$7A,$CE,$5F,$F6
	.dc.b	$24,$66,$BF,$DC,$14,$AD,$47,$FB,$82,$75,$2E,$D7,$5B,$E6,$7F,$5D
	.dc.b	$E9,$FE,$5D,$B8,$7F,$BD,$DB,$7F,$E6,$96,$E5,$F8,$5D,$7F,$6C,$7E
	.dc.b	$7F,$F0,$CB,$7F,$A9,$8B,$7F,$E7,$19,$B5,$FD,$26,$1F,$1A,$FA,$2F
	.dc.b	$E3,$6D,$C3,$FC,$A7,$E5,$3F,$01,$7C,$E6,$AE,$56,$A5,$F4,$F8,$C7
	.dc.b	$FC,$7B,$27,$FC,$B9,$3A,$8F,$F2,$96,$3D,$E7,$66,$AB,$96,$5F,$D7
	.dc.b	$F2,$F1,$47,$EF,$7E,$AF,$F8,$B7,$C1,$DC,$7F,$84,$37,$0F,$E1,$AD
	.dc.b	$C3,$F5,$8F,$F1,$DF,$83,$95,$FE,$70,$E3,$FA,$BF,$41,$8C,$3F,$8A
	.dc.b	$D5,$C3,$FB,$85,$6A,$FF,$8B,$4F,$F4,$0E,$C9,$1F,$E7,$AF,$80,$7D
	.dc.b	$4B,$E8,$F3,$97,$FB,$A7,$D4,$61,$FC,$3B,$A9,$F4,$FA,$D7,$F1,$ED
	.dc.b	$CF,$F1,$8D,$C3,$FC,$D3,$5C,$27,$F9,$47,$0F,$F3,$93,$F9,$9F,$C2
	.dc.b	$FC,$DC,$3F,$86,$46,$67,$67,$C1,$9D,$3F,$C3,$A3,$F5,$BF,$3D,$FD
	.dc.b	$83,$E7,$3F,$C3,$93,$F8,$4F,$05,$39,$8D,$9F,$88,$9D,$E5,$F7,$39
	.dc.b	$BF,$D9,$D5,$BA,$FE,$DF,$A7,$F6,$83,$FB,$7F,$0B,$E8,$7F,$A0,$EB
	.dc.b	$F9,$7E,$0F,$76,$5F,$26,$F7,$C5,$FC,$C6,$B5,$18,$F3,$BA,$65,$F4
	.dc.b	$EB,$4F,$F8,$9A,$74,$FE,$7B,$E7,$5F,$2B,$FD,$F9,$FD,$2C,$66,$31
	.dc.b	$D7,$96,$C6,$4E,$51,$F4,$3E,$57,$FB,$57,$7D,$9F,$EB,$8D,$D3,$FB
	.dc.b	$0B,$74,$FE,$80,$FF,$83,$3E,$9D,$7F,$E1,$2F,$82,$7D,$37,$E2,$31
	.dc.b	$FF,$5D,$56,$9F,$F6,$4C,$EB,$AB,$95,$F1,$BE,$19,$7E,$B8,$61,$C5
	.dc.b	$FD,$8D,$F1,$C3,$FB,$67,$CB,$FD,$36,$F1,$F3,$CB,$FC,$29,$39,$BF
	.dc.b	$DC,$FB,$28,$C6,$BF,$E1,$1F,$C4,$5F,$53,$C3,$1F,$F0,$8C,$63,$2E
	.dc.b	$FF,$43,$FE,$11,$F7,$FA,$B6,$32,$7F,$81,$63,$2D,$3F,$DB,$9F,$4F
	.dc.b	$FE,$28,$EA,$FF,$3A,$79,$9F,$CD,$A3,$1B,$3D,$5F,$E6,$4F,$F1,$5F
	.dc.b	$A8,$F5,$46,$21,$3C,$23,$0D,$AE,$98,$B0,$EF,$7C,$F4,$FF,$3B,$61
	.dc.b	$C2,$B5,$3E,$1B,$E3,$7C,$23,$BC,$EA,$BC,$58,$F9,$2F,$75,$C4,$FE
	.dc.b	$59,$8F,$8D,$B7,$C6,$7F,$AA,$D7,$19,$CE,$1D,$EF,$84,$78,$23,$C5
	.dc.b	$3B,$8C,$1D,$CB,$87,$AE,$77,$8F,$BB,$0F,$2C,$E6,$B8,$61,$DB,$0E
	.dc.b	$8F,$80,$98,$F9,$E7,$E4,$65,$E5,$8C,$47,$83,$0E,$A7,$B3,$1F,$04
	.dc.b	$7B,$0E,$F3,$C2,$FC,$13,$A9,$E4,$6F,$1F,$31,$F3,$A3,$53,$E4,$CB
	.dc.b	$C4,$7B,$B1,$B7,$DA,$2D,$F1,$9B,$8F,$9A,$BB,$3E,$78,$CC,$60,$83
	.dc.b	$2F,$8B,$97,$A4,$98,$78,$67,$C0,$4C,$3D,$38,$75,$8C,$3E,$17,$DE
	.dc.b	$75,$CB,$DB,$7D,$23,$FD,$D7,$97,$E9,$1F,$F0,$D9,$F8,$02,$B9,$7F
	.dc.b	$55,$CC,$F2,$3F,$DE,$13,$FE,$04,$6C,$66,$7A,$B5,$E6,$E7,$1F,$D5
	.dc.b	$B7,$1E,$B5,$FA,$3F,$A8,$FF,$0A,$DF,$E5,$EB,$4F,$E8,$1A,$FE,$D9
	.dc.b	$B1,$C2,$3F,$23,$1A,$C3,$CD,$18,$AE,$DD,$3F,$E3,$AC,$7C,$71,$DC
	.dc.b	$93,$FA,$1C,$3A,$4E,$A3,$91,$1F,$E1,$11,$AC,$7D,$73,$96,$CE,$AF
	.dc.b	$C1,$18,$24,$FB,$6F,$CA,$F8,$AB,$C0,$49,$FF,$28,$DF,$79,$E0,$77
	.dc.b	$2C,$F7,$79,$4F,$67,$E4,$57,$67,$81,$82,$D7,$57,$15,$E1,$7A,$D7
	.dc.b	$8A,$FD,$71,$82,$1C,$23,$4F,$88,$CE,$1F,$52,$BC,$D1,$E2,$8C,$4F
	.dc.b	$9E,$FC,$51,$89,$E3,$8F,$E0,$30,$F0,$C7,$73,$38,$FB,$71,$F4,$46
	.dc.b	$49,$87,$C9,$E4,$FE,$B1,$E7,$7F,$9F,$39,$3F,$DD,$1F,$3C,$F9,$AC
	.dc.b	$71,$FF,$96,$FE,$11,$39,$CB,$FB,$3E,$1A,$58,$F4,$7F,$DD,$33,$B7
	.dc.b	$FC,$BF,$5C,$B0,$E1,$5B,$9F,$ED,$7D,$1F,$A9,$CB,$FB,$57,$C2,$75
	.dc.b	$C1,$D0,$8E,$35,$BD,$3F,$CE,$18,$FF,$73,$DF,$24,$CB,$FC,$0F,$F5
	.dc.b	$0F,$6B,$D1,$0F,$A8,$FF,$2A,$BA,$8F,$67,$72,$B8,$4F,$F9,$F7,$6F
	.dc.b	$F4,$65,$F0,$8F,$F0,$73,$61,$97,$FA,$0E,$F9,$61,$D3,$0E,$78,$70
	.dc.b	$EE,$F9,$5E,$8B,$CE,$7D,$3F,$C1,$E4,$7C,$2D,$37,$3A,$C3,$FC,$B9
	.dc.b	$59,$7E,$8C,$EA,$3E,$B7,$9A,$F2,$1F,$03,$85,$F0,$B4,$C4,$4C,$3F
	.dc.b	$C9,$11,$97,$FC,$95,$1F,$2B,$E8,$5F,$4D,$F3,$39,$FD,$BC,$9F,$EB
	.dc.b	$16,$F2,$7F,$B2,$AB,$78,$7C,$B8,$C5,$F1,$E3,$93,$93,$FD,$59,$5C
	.dc.b	$1F,$F1,$E3,$70,$FF,$5F,$36,$F9,$CF,$13,$FC,$6D,$F4,$FF,$0B,$E9
	.dc.b	$DF,$4F,$7D,$29,$5F,$EC,$29,$CB,$67,$2F,$F9,$F6,$13,$CE,$7E,$B7
	.dc.b	$D0,$CA,$FD,$0B,$F5,$BE,$9B,$11,$FA,$2E,$64,$E3,$4F,$F5,$7F,$32
	.dc.b	$FE,$67,$A3,$F2,$FE,$4A,$63,$EB,$DF,$FC,$CE,$D9,$CB,$FE,$5F,$E2
	.dc.b	$57,$28,$D1,$0F,$93,$3F,$5F,$E9,$BF,$1D,$F3,$F2,$7F,$72,$FE,$3B
	.dc.b	$EE,$31,$1E,$98,$BA,$7A,$27,$F2,$7F,$5F,$F9,$7F,$90,$97,$FE,$AF
	.dc.b	$8C,$BF,$C1,$23,$2F,$FB,$8B,$BB,$F4,$13,$FB,$4F,$CE,$DF,$3F,$F9
	.dc.b	$8F,$AF,$F5,$DF,$FB,$D9,$AF,$F9,$F6,$76,$FD,$F6,$BF,$EB,$38,$DE
	.dc.b	$5F,$EE,$D6,$F4,$FF,$BD,$E5,$5C,$2B,$F7,$31,$FE,$FB,$8C,$4E,$DB
	.dc.b	$19,$BF,$37,$CD,$7E,$DF,$DD,$93,$FC,$89,$FD,$BF,$6E,$23,$F0,$51
	.dc.b	$71,$FC,$0C,$FE,$F3,$F0,$3F,$E0,$1F,$21,$2F,$FD,$B1,$19,$7F,$96
	.dc.b	$C6,$5F,$95,$D0,$9F,$E3,$BF,$8F,$FF,$08,$9F,$D5,$60,$FB,$77,$C9
	.dc.b	$FF,$39,$8C,$3C,$3C,$89,$5C,$2B,$6D,$9C,$BF,$EA,$DB,$EC,$DA,$CD
	.dc.b	$FC,$AF,$1D,$F0,$FF,$42,$BF,$A3,$FD,$0B,$E4,$7F,$9F,$72,$7E,$86
	.dc.b	$7F,$80,$7A,$9F,$CA,$BC,$1F,$E8,$F7,$B9,$FE,$09,$E9,$21,$F1,$DF
	.dc.b	$D3,$3E,$AB,$F5,$B5,$A6,$BF,$EB,$89,$DB,$C9,$7C,$7F,$6A,$93,$98
	.dc.b	$F0,$FC,$6F,$9B,$C0,$93,$C6,$3C,$BE,$F6,$A3,$07,$57,$2D,$21,$F6
	.dc.b	$3E,$77,$DB,$F5,$3F,$CF,$FE,$6E,$4F,$96,$32,$FF,$6C,$8C,$BE,$B8
	.dc.b	$CB,$4E,$07,$F2,$9F,$0F,$FA,$67,$1F,$F7,$0F,$A9,$FE,$E9,$CD,$F9
	.dc.b	$E1,$F8,$33,$3F,$C2,$37,$0F,$9F,$5C,$39,$7F,$B7,$B6,$17,$CA,$F8
	.dc.b	$9D,$FD,$F9,$3C,$9A,$FE,$63,$E7,$BC,$9E,$4F,$F5,$FF,$9E,$93,$E0
	.dc.b	$8C,$BF,$E0,$C8,$CB,$FE,$F2,$C5,$3A,$8C,$37,$1F,$E5,$BF,$35,$F0
	.dc.b	$78,$3F,$B6,$77,$CB,$F6,$C6,$5F,$F0,$CC,$65,$ED,$19,$69,$C0,$F9
	.dc.b	$DE,$17,$E5,$5F,$87,$B7,$F0,$DF,$A6,$FE,$DF,$1F,$D4,$DF,$F7,$86
	.dc.b	$3F,$C3,$A1,$18,$9E,$1E,$CF,$F8,$5A,$38,$CF,$40,$70,$F8,$9D,$1B
	.dc.b	$99,$1D,$DA,$F1,$F0,$5F,$4F,$FC,$6A,$3F,$55,$F4,$9F,$69,$7D,$33
	.dc.b	$CC,$F5,$7D,$27,$8C,$FF,$14,$F8,$2B,$E1,$5B,$BF,$2C,$E7,$6F,$3C
	.dc.b	$E6,$7D,$F7,$CD,$FE,$ED,$5B,$7C,$38,$7F,$35,$8C,$57,$1D,$57,$FC
	.dc.b	$E6,$30,$DA,$E1,$8F,$91,$B8,$F8,$9F,$5E,$34,$F9,$78,$74,$70,$7C
	.dc.b	$49,$B8,$F8,$AB,$8C,$F9,$8E,$38,$78,$5F,$83,$87,$85,$B7,$D3,$0E
	.dc.b	$D1,$C4,$F8,$67,$69,$E0,$F8,$4D,$47,$0C,$7D,$F1,$BA,$8E,$AF,$C4
	.dc.b	$F4,$3B,$63,$E9,$7C,$66,$CF,$24,$70,$8D,$47,$86,$BB,$BB,$39,$E1
	.dc.b	$E6,$9D,$E1,$E5,$C7,$E2,$DF,$DF,$9E,$67,$DD,$BF,$F7,$8B,$5F,$AA
	.dc.b	$7F,$BD,$0F,$80,$6B,$E9,$DF,$0B,$98,$FE,$81,$18,$78,$4A,$31,$1C
	.dc.b	$67,$83,$FA,$8F,$8C,$A9,$A4,$76,$FE,$0B,$E6,$5F,$4F,$FE,$04,$C7
	.dc.b	$F3,$55,$CC,$FB,$53,$AE,$3F,$C4,$C6,$67,$F4,$13,$98,$FC,$01,$9C
	.dc.b	$FF,$C2,$51,$99,$EB,$3E,$51,$F3,$FC,$B0,$FF,$0A,$BA,$C7,$F8,$1B
	.dc.b	$9C,$6E,$3F,$2E,$7D,$55,$E9,$8C,$5F,$A2,$30,$FF,$88,$CF,$06,$1C
	.dc.b	$DF,$39,$89,$F3,$CF,$08,$FF,$15,$57,$6B,$F8,$31,$B3,$F8,$BE,$3E
	.dc.b	$DB,$F0,$E1,$D4,$F5,$4F,$8E,$B8,$DF,$38,$F0,$B6,$77,$8F,$94,$B7
	.dc.b	$E4,$BF,$83,$7E,$BB,$AB,$0F,$A5,$7C,$23,$CB,$8F,$CC,$C3,$CA,$73
	.dc.b	$8F,$E6,$07,$B8,$F3,$18,$A7,$D2,$7F,$6B,$C7,$ED,$B6,$5C,$7F,$CC
	.dc.b	$3E,$27,$11,$E3,$9B,$7C,$AF,$F3,$CD,$9F,$D2,$35,$FA,$CD,$7B,$C7
	.dc.b	$FA,$7E,$3F,$D0,$2D,$9E,$BE,$EF,$8F,$CC,$7D,$3B,$E5,$30,$DB,$FF
	.dc.b	$04,$7D,$27,$F3,$D7,$FC,$5F,$19,$CB,$FA,$95,$F2,$BF,$EB,$51,$FE
	.dc.b	$F2,$9D,$3F,$E6,$D9,$D5,$7E,$F5,$B1,$F8,$F6,$F4,$7F,$9C,$3D,$1F
	.dc.b	$F7,$4F,$97,$E4,$C7,$E5,$46,$30,$F1,$9E,$83,$E7,$79,$4C,$DF,$21
	.dc.b	$99,$D9,$E1,$D3,$FD,$EA,$D3,$D3,$A3,$4F,$8B,$18,$6B,$FE,$54,$3F
	.dc.b	$CB,$67,$48,$FE,$E9,$8F,$FC,$00,$F2,$FE,$77,$6F,$FB,$E2,$BF,$CA
	.dc.b	$ED,$AF,$F5,$06,$5F,$D6,$A7,$3E,$A7,$03,$C1,$5B,$AF,$94,$D7,$FC
	.dc.b	$03,$8F,$F6,$29,$D9,$F8,$FD,$DF,$4F,$FE,$D6,$C3,$A9,$FE,$82,$71
	.dc.b	$7F,$6B,$07,$C0,$7A,$0F,$97,$93,$E3,$31,$87,$AA,$7C,$A4,$7E,$E5
	.dc.b	$72,$6C,$7D,$85,$E6,$72,$C3,$E3,$C6,$0F,$8B,$E3,$FC,$83,$C3,$E2
	.dc.b	$D8,$C9,$DA,$3E,$29,$1F,$0F,$99,$F2,$08,$E0,$F2,$F9,$3F,$30,$8E
	.dc.b	$0F,$4F,$5F,$D0,$23,$83,$DB,$F1,$9F,$50,$8E,$0F,$EA,$9F,$4D,$FE
	.dc.b	$9B,$7F,$C6,$BC,$3F,$E3,$D8,$C6,$7F,$E3,$1E,$47,$BB,$FC,$0A,$73
	.dc.b	$5B,$F6,$BF,$27,$69,$E2,$FF,$35,$6B,$FE,$D9,$6B,$FE,$44,$79,$E1
	.dc.b	$E2,$FA,$36,$BD,$EF,$F0,$D1,$89,$FD,$FF,$9B,$F5,$FC,$FF,$C0,$B6
	.dc.b	$32,$7E,$86,$32,$D3,$F4,$95,$F7,$CF,$C4,$FC,$AD,$8C,$93,$2F,$D0
	.dc.b	$C6,$5F,$E5,$BF,$2E,$FA,$6F,$E4,$AD,$AF,$96,$7F,$98,$3C,$67,$D3
	.dc.b	$7F,$A5,$E3,$37,$FE,$64,$AE,$0F,$4B,$F9,$71,$88,$FE,$BC,$D7,$FC
	.dc.b	$D6,$1C,$3C,$E3,$F9,$D5,$F8,$3D,$AF,$2C,$63,$04,$7E,$A3,$CB,$FD
	.dc.b	$07,$BB,$72,$FE,$10,$DC,$7F,$42,$DE,$2F,$E4,$13,$FE,$DA,$3F,$87
	.dc.b	$7B,$BF,$A9,$3F,$67,$F2,$F6,$32,$4B,$FE,$17,$19,$7F,$AA,$74,$7F
	.dc.b	$4C,$3F,$A9,$7C,$E3,$E8,$DF,$4D,$9C,$3F,$A4,$57,$E4,$3C,$F3,$49
	.dc.b	$1F,$D5,$1F,$C4,$FC,$47,$1C,$7F,$D2,$9A,$3A,$13,$FC,$07,$82,$61
	.dc.b	$FE,$A6,$E2,$B8,$7F,$02,$9B,$82,$BF,$D3,$BE,$63,$DB,$FB,$73,$F1
	.dc.b	$AB,$E9,$F7,$3E,$97,$EA,$FD,$37,$F3,$CC,$BF,$5D,$18,$CB,$6D,$7F
	.dc.b	$0F,$D0,$BE,$51,$8B,$E5,$8F,$FA,$D7,$83,$D7,$3F,$DB,$76,$F9,$7F
	.dc.b	$41,$FD,$0B,$3F,$F0,$14,$7F,$7A,$3E,$51,$27,$FB,$DB,$63,$FD,$75
	.dc.b	$C9,$F4,$FE,$5F,$85,$7E,$6A,$30,$F9,$BE,$69,$F4,$DF,$92,$7F,$2D
	.dc.b	$19,$8C,$76,$E5,$B1,$93,$94,$7F,$66,$E2,$7C,$A7,$F3,$5F,$86,$C3
	.dc.b	$6B,$FD,$B6,$DC,$FF,$DA,$1C,$BF,$4E,$31,$C1,$FE,$D9,$FA,$5E,$77
	.dc.b	$FE,$75,$6D,$FF,$9E,$DA,$FF,$93,$5A,$FE,$C1,$DC,$FE,$06,$3F,$CF
	.dc.b	$FF,$4D,$F9,$2A,$FD,$EC,$2B,$71,$FE,$E2,$F8,$05,$6E,$BF,$C2,$BF
	.dc.b	$2E,$E3,$FF,$10,$36,$FF,$83,$DF,$0E,$CC,$D4,$9F,$F0,$B5,$6E,$FF
	.dc.b	$C2,$CD,$7F,$C3,$8D,$7F,$B9,$AF,$A5,$FF,$04,$36,$79,$C6,$08,$FF
	.dc.b	$82,$FD,$BF,$2C,$7F,$67,$7F,$C8,$BF,$49,$FE,$35,$E5,$FF,$4C,$79
	.dc.b	$5F,$C5,$A3,$18,$FF,$A6,$63,$F8,$CF,$D2,$F9,$FE,$0D,$1B,$C1,$87
	.dc.b	$2B,$0C,$39,$DF,$08,$CC,$62,$FA,$E9,$F5,$6B,$55,$C7,$0E,$AF,$8C
	.dc.b	$E2,$E2,$37,$7C,$E7,$C3,$8F,$9F,$0F,$36,$3E,$59,$E8,$DA,$E1,$0A
	.dc.b	$E0,$F6,$9D,$61,$DE,$F9,$CF,$8C,$E1,$8F,$A6,$33,$8F,$92,$B8,$57
	.dc.b	$33,$86,$1D,$9F,$19,$CF,$1F,$19,$E4,$8D,$38,$9E,$73,$C2,$BE,$3D
	.dc.b	$6A,$73,$5B,$C7,$E0,$4E,$5F,$59,$2F,$C7,$8F,$A5,$F3,$61,$E9,$9D
	.dc.b	$3D,$A7,$53,$E1,$C7,$DE,$47,$E8,$C6,$B0,$F1,$47,$B7,$1F,$4E,$3E
	.dc.b	$38,$C6,$3D,$B0,$FC,$1E,$76,$7C,$71,$D5,$F9,$F1,$97,$C4,$68,$F0
	.dc.b	$B9,$C3,$C3,$97,$A6,$31,$8F,$C6,$3C,$79,$79,$DC,$63,$FE,$D7,$C3
	.dc.b	$EC,$1F,$ED,$9C,$3F,$4A,$D9,$FC,$AB,$9E,$47,$8B,$C6,$3F,$45,$18
	.dc.b	$7A,$52,$31,$1C,$67,$84,$E5,$FC,$B4,$F1,$CB,$F3,$ED,$CF,$FC,$1B
	.dc.b	$F4,$BF,$E0,$DC,$37,$71,$FC,$65,$6D,$FD,$AC,$ED,$CC,$EE,$FA,$BF
	.dc.b	$A7,$9D,$46,$27,$64,$7F,$6F,$1A,$BF,$E1,$57,$DE,$BE,$0D,$F5,$9F
	.dc.b	$3F,$47,$FB,$F1,$DC,$66,$7F,$82,$D7,$1B,$E9,$18,$9D,$3D,$0F,$D9
	.dc.b	$46,$9C,$61,$E7,$25,$F8,$23,$8E,$3F,$35,$F5,$46,$5A,$77,$28,$AE
	.dc.b	$D3,$C6,$BA,$CE,$DF,$87,$8F,$B6,$31,$7E,$E7,$9D,$F2,$C7,$E3,$61
	.dc.b	$CE,$B5,$87,$2B,$F3,$D6,$8F,$89,$87,$58,$F7,$D7,$02,$3F,$7A,$35
	.dc.b	$7D,$63,$C6,$43,$D9,$1A,$AE,$D8,$FC,$EC,$7C,$CF,$88,$FC,$24,$FB
	.dc.b	$7D,$C7,$0B,$CC,$FF,$2F,$F3,$FF,$BB,$FE,$79,$F1,$D8,$E3,$FE,$8B
	.dc.b	$EF,$A7,$3B,$7F,$9B,$B2,$E7,$8E,$E3,$FD,$D1,$97,$57,$FB,$4C,$7E
	.dc.b	$B6,$77,$18,$3F,$B6,$E9,$9A,$FD,$6C,$7F,$6C,$8C,$11,$F0,$47,$EC
	.dc.b	$B3,$FF,$03,$57,$B2,$BC,$5D,$0F,$A3,$7A,$DE,$5C,$5F,$49,$FE,$5C
	.dc.b	$7F,$C0,$33,$A6,$CE,$E7,$50,$C7,$FC,$28,$F2,$EB,$FE,$10,$D3,$F8
	.dc.b	$54,$F0,$7F,$D1,$66,$4B,$DF,$FE,$5B,$FA,$4F,$91,$C1,$F3,$7E,$0A
	.dc.b	$F0,$9F,$99,$5B,$BF,$F4,$EB,$5F,$27,$17,$D6,$AE,$11,$FE,$A2,$BE
	.dc.b	$8F,$0F,$09,$8F,$D7,$F2,$9E,$2B,$DC,$C3,$FE,$FF,$BF,$F2,$87,$C0
	.dc.b	$62,$64,$60,$87,$88,$57,$48,$FA,$1F,$40,$F9,$BF,$A5,$D1,$F3,$EF
	.dc.b	$81,$F9,$4E,$4D,$CC,$8F,$E6,$95,$C5,$B1,$F4,$23,$11,$FD,$05,$BC
	.dc.b	$9F,$59,$BC,$1F,$EB,$C6,$CF,$F8,$83,$07,$D2,$7C,$3F,$A4,$7D,$25
	.dc.b	$F3,$65,$FF,$A1,$A7,$2D,$9C,$BF,$07,$64,$AB,$A5,$7D,$CF,$A4,$97
	.dc.b	$F5,$79,$51,$F7,$B9,$1F,$39,$AA,$F9,$9C,$DF,$B1,$AC,$70,$7F,$99
	.dc.b	$79,$FF,$D4,$97,$F3,$BC,$EF,$9B,$F1,$B3,$1F,$F3,$3F,$17,$F9,$F9
	.dc.b	$B3,$F2,$48,$FC,$E6,$BF,$52,$38,$36,$3E,$CF,$D2,$7E,$5B,$E6,$24
	.dc.b	$FA,$DF,$57,$CE,$62,$3D,$3B,$A7,$1A,$7A,$77,$5F,$C1,$FA,$4F,$D0
	.dc.b	$FC,$5C,$C7,$F3,$11,$97,$E4,$46,$5F,$F4,$9D,$27,$F7,$DF,$46,$FA
	.dc.b	$3F,$CF,$7D,$23,$82,$9A,$89,$87,$F8,$06,$B8,$36,$3F,$7B,$18,$8F
	.dc.b	$F0,$73,$73,$FE,$06,$DE,$BF,$F3,$5C,$A3,$0F,$E9,$30,$7C,$B7,$EF
	.dc.b	$BD,$A9,$3F,$83,$7F,$05,$DD,$88,$F9,$31,$71,$F9,$13,$F8,$7F,$A4
	.dc.b	$FE,$0F,$F1,$B2,$FF,$7D,$19,$7F,$1D,$19,$7E,$3C,$7F,$23,$FA,$3F
	.dc.b	$E0,$F1,$FE,$F8,$FA,$47,$C8,$7F,$3B,$C5,$18,$9F,$F5,$8D,$70,$9F
	.dc.b	$E4,$6F,$28,$FE,$3E,$D7,$E1,$B5,$FF,$47,$B5,$3C,$F6,$DF,$93,$E5
	.dc.b	$3F,$8B,$7C,$17,$D9,$AF,$EA,$1F,$3A,$F8,$EF,$E9,$8D,$8F,$F2,$17
	.dc.b	$A4,$8C,$47,$F3,$28,$C4,$F0,$F8,$C9,$3C,$29,$18,$8E,$10,$F8,$CE
	.dc.b	$E4,$8E,$0F,$8B,$E2,$24,$73,$6B,$CB,$DA,$F3,$91,$F0,$35,$FE,$55
	.dc.b	$8F,$A6,$3F,$B6,$36,$3F,$9D,$7C,$CE,$30,$7D,$1F,$F4,$8F,$94,$95
	.dc.b	$F1,$A3,$47,$C6,$23,$EF,$6B,$E0,$9E,$6F,$13,$FA,$46,$BE,$2F,$A2
	.dc.b	$FF,$00,$C7,$F3,$FE,$27,$D1,$FF,$99,$63,$FD,$97,$1F,$EF,$93,$3A
	.dc.b	$71,$F7,$64,$FD,$6B,$E6,$FE,$A3,$54,$62,$3F,$DA,$B1,$89,$FF,$05
	.dc.b	$4F,$8A,$FF,$C0,$2D,$E3,$DB,$6D,$E5,$3C,$27,$CB,$FA,$FC,$6F,$ED
	.dc.b	$FF,$2D,$2F,$FC,$39,$87,$F1,$76,$CF,$80,$8F,$A5,$AF,$D8,$8E,$0D
	.dc.b	$8F,$E7,$9F,$2D,$F2,$78,$FF,$BE,$F6,$4A,$EF,$19,$7E,$23,$EE,$23
	.dc.b	$83,$64,$3F,$90,$F4,$FF,$87,$7E,$87,$FB,$D4,$7D,$2D,$BF,$89,$C7
	.dc.b	$FB,$DB,$7F,$F0,$CB,$63,$3E,$C7,$13,$FC,$37,$1C,$AB,$FD,$58,$D7
	.dc.b	$FC,$36,$D7,$89,$DC,$07,$95,$78,$27,$FC,$21,$F4,$7F,$C9,$2B,$F6
	.dc.b	$DF,$47,$FE,$2D,$5F,$45,$FC,$99,$E7,$83,$E7,$3D,$C7,$F8,$EF,$A9
	.dc.b	$7C,$6F,$85,$F4,$CB,$C7,$F3,$58,$C7,$94,$E3,$C8,$F8,$B9,$7F,$38
	.dc.b	$AE,$39,$74,$C3,$57,$EC,$8C,$29,$E9,$5C,$6F,$9C,$F0,$3C,$46,$DF
	.dc.b	$E8,$B5,$CB,$1E,$33,$F2,$1A,$FC,$29,$DE,$5E,$D6,$E3,$B7,$BC,$EC
	.dc.b	$F8,$F8,$7F,$52,$BF,$5D,$F1,$9F,$53,$5F,$AD,$3B,$AE,$75,$C2,$FB
	.dc.b	$57,$06,$BC,$0B,$97,$CC,$7C,$47,$C0,$9E,$4E,$23,$51,$EE,$32,$6B
	.dc.b	$2F,$82,$4B,$F7,$D7,$C9,$3C,$47,$3C,$3D,$0F,$C0,$3C,$B9,$7C,$D9
	.dc.b	$E3,$7D,$DF,$69,$83,$AC,$FF,$9D,$A7,$EE,$B7,$2F,$CA,$37,$2F,$F3
	.dc.b	$DB,$F8,$3F,$A3,$7C,$2E,$23,$FA,$84,$61,$DD,$A3,$11,$DD,$AF,$4F
	.dc.b	$8C,$FF,$45,$A6,$91,$FF,$47,$BF,$4B,$E5,$5F,$46,$E2,$FA,$B5,$FB
	.dc.b	$2A,$D4,$E6,$3E,$A9,$93,$EB,$D7,$2A,$E3,$1E,$13,$33,$C7,$5F,$D6
	.dc.b	$C7,$1D,$FF,$5F,$7C,$63,$EB,$1F,$E2,$2C,$3F,$45,$19,$9E,$33,$D6
	.dc.b	$B8,$DF,$5D,$BE,$14,$F8,$31,$F1,$E3,$EA,$6C,$64,$F5,$9C,$E7,$FC
	.dc.b	$63,$8F,$8F,$0E,$F7,$C6,$B6,$D8,$CC,$73,$69,$E4,$BE,$ED,$8E,$26
	.dc.b	$EF,$CF,$19,$39,$E3,$B2,$C7,$93,$1F,$A5,$1D,$A3,$17,$D9,$B7,$DA
	.dc.b	$FA,$47,$63,$2F,$48,$FE,$5F,$8F,$CD,$9F,$93,$7E,$17,$84,$EB,$0F
	.dc.b	$9C,$FF,$B5,$31,$F8,$53,$A8,$5F,$C9,$8C,$36,$38,$1F,$79,$F0,$38
	.dc.b	$BF,$03,$AC,$3D,$2D,$AF,$49,$D3,$0F,$1C,$7E,$1C,$F9,$87,$38,$F8
	.dc.b	$75,$EF,$8F,$2C,$FC,$A9,$E5,$8F,$C9,$9F,$73,$5D,$B7,$0F,$54,$77
	.dc.b	$BF,$3B,$6B,$B9,$D1,$FA,$4F,$A1,$F5,$1D,$8C,$97,$0F,$35,$7C,$48
	.dc.b	$E8,$6E,$BF,$00,$EE,$BB,$56,$23,$CE,$74,$2B,$D3,$0F,$81,$1F,$14
	.dc.b	$E9,$3C,$88,$F9,$8E,$2E,$1F,$19,$DA,$BA,$61,$FD,$5E,$3F,$82,$5F
	.dc.b	$D8,$7E,$A3,$5C,$D7,$C3,$72,$D7,$F8,$23,$F2,$9F,$31,$DD,$F1,$3C
	.dc.b	$8F,$84,$79,$F0,$FA,$AE,$4F,$49,$9C,$3F,$0E,$E6,$36,$FF,$01,$39
	.dc.b	$9E,$42,$9B,$8E,$2E,$CE,$A6,$CA,$6E,$FF,$11,$1C,$31,$FD,$3C,$F8
	.dc.b	$E7,$24,$7E,$C5,$79,$27,$71,$D6,$30,$FE,$78,$C5,$75,$70,$78,$4F
	.dc.b	$E3,$DA,$FF,$39,$7F,$0A,$D7,$4E,$67,$F9,$94,$7F,$B4,$8F,$C5,$95
	.dc.b	$CB,$FC,$38,$C4,$FF,$9D,$EF,$F9,$4F,$AD,$67,$F3,$D7,$D6,$31,$ED
	.dc.b	$BE,$FD,$F5,$98,$7F,$AF,$61,$FE,$43,$9D,$4E,$77,$93,$9A,$DE,$2C
	.dc.b	$7F,$AB,$61,$C2,$B5,$7E,$39,$7E,$6E,$F1,$B0,$DA,$FF,$29,$7C,$13
	.dc.b	$F0,$67,$76,$AE,$56,$A7,$FD,$06,$FF,$96,$70,$EB,$3F,$06,$F9,$78
	.dc.b	$5F,$0F,$4F,$E4,$BE,$C5,$E2,$37,$E9,$F0,$CC,$BF,$C0,$93,$A6,$C7
	.dc.b	$D9,$AF,$E7,$2D,$7A,$C7,$F9,$AD,$FE,$67,$CB,$CD,$F8,$D1,$FD,$7A
	.dc.b	$3E,$B7,$90,$F7,$FF,$B9,$6B,$FD,$FE,$31,$CD,$EB,$FC,$FD,$E3,$F0
	.dc.b	$BD,$E3,$FE,$AE,$6E,$4E,$5F,$BD,$DD,$BC,$7E,$F3,$C6,$65,$FD,$0F
	.dc.b	$D0,$1F,$3F,$89,$FF,$6D,$E3,$FE,$84,$AF,$C0,$7D,$4E,$76,$FB,$FD
	.dc.b	$48,$F5,$FE,$00,$F8,$8F,$8D,$C2,$8C,$61,$FE,$FE,$6E,$5F,$EF,$D6
	.dc.b	$BF,$EC,$B7,$38,$FF,$88,$79,$DF,$A3,$BC,$AE,$F9,$7D,$F8,$C6,$3F
	.dc.b	$95,$F1,$BF,$38,$66,$AC,$64,$99,$7D,$41,$87,$69,$DC,$67,$C6,$7E
	.dc.b	$1E,$FA,$FF,$C1,$46,$1B,$19,$C3,$FC,$9F,$39,$6C,$66,$BC,$1C,$DF
	.dc.b	$B3,$F9,$0F,$DC,$5E,$E7,$3C,$49,$FC,$D7,$CD,$3E,$71,$F3,$BF,$E4
	.dc.b	$7F,$9D,$FF,$5C,$E0,$AD,$4E,$63,$FD,$7B,$F3,$EF,$9F,$FF,$26,$79
	.dc.b	$8C,$38,$E1,$C2,$FF,$CB,$51,$8C,$7B,$C6,$0F,$F6,$3C,$EE,$B9,$47
	.dc.b	$6B,$EB,$7C,$71,$F1,$4F,$97,$0E,$77,$C2,$31,$3A,$C7,$CD,$BF,$0C
	.dc.b	$BB,$63,$CE,$FC,$47,$02,$D7,$07,$33,$C1,$F3,$E3,$E0,$AE,$58,$F8
	.dc.b	$6F,$83,$E3,$BE,$17,$D6,$31,$7C,$63,$DD,$87,$96,$31,$1B,$C3,$A9
	.dc.b	$F5,$30,$EA,$E2,$3C,$78,$F1,$3E,$06,$5F,$22,$32,$7A,$AF,$AE,$5E
	.dc.b	$4C,$39,$CE,$B2,$F2,$E1,$CE,$3E,$1E,$1C,$DF,$21,$BC,$7C,$4A,$F9
	.dc.b	$BA,$26,$1D,$EB,$93,$F1,$09,$8F,$C0,$AE,$31,$8C,$3C,$58,$FC,$DB
	.dc.b	$E3,$7E,$A2,$4F,$C0,$CB,$79,$79,$30,$F0,$46,$32,$F4,$10,$F4,$61
	.dc.b	$D2,$F5,$1E,$9C,$BC,$78,$F6,$8F,$F6,$0E,$1F,$9D,$7F,$B7,$35,$FC
	.dc.b	$61,$8F,$69,$F3,$9F,$EC,$59,$46,$27,$9B,$5E,$13,$B8,$D6,$9F,$8B
	.dc.b	$6E,$9F,$AE,$6D,$07,$CF,$BE,$7F,$FD,$65,$5B,$C3,$F4,$57,$C2,$FF
	.dc.b	$7C,$DB,$E1,$1A,$98,$FE,$8A,$75,$7D,$23,$F7,$31,$CF,$1F,$6D,$70
	.dc.b	$C7,$F6,$98,$74,$BE,$53,$9C,$7C,$91,$FB,$29,$D5,$7B,$EB,$53,$D9
	.dc.b	$FE,$03,$3A,$AE,$35,$BA,$EA,$73,$C7,$CC,$FB,$AF,$9B,$4D,$18,$C3
	.dc.b	$D2,$F2,$AF,$4D,$72,$9C,$C7,$72,$5F,$F2,$8C,$3C,$97,$E7,$39,$47
	.dc.b	$98,$D5,$F3,$C7,$E4,$13,$1F,$6C,$65,$F2,$47,$D0,$AB,$1E,$AB,$EF
	.dc.b	$1E,$BC,$3C,$6F,$85,$F4,$BC,$0F,$19,$F8,$78,$E3,$8F,$B3,$1F,$34
	.dc.b	$F1,$7C,$EF,$1A,$EA,$FE,$3B,$1F,$C1,$70,$BF,$48,$7B,$DF,$D2,$BB
	.dc.b	$3F,$C7,$7D,$CF,$8F,$C7,$27,$E5,$F9,$D3,$9C,$FF,$35,$6A,$FC,$E6
	.dc.b	$5D,$71,$E3,$48,$C3,$FD,$9B,$2F,$0B,$67,$FC,$C4,$F2,$8C,$BF,$E6
	.dc.b	$5C,$7C,$3F,$17,$D1,$B1,$FE,$59,$C9,$3E,$0E,$CF,$F0,$EF,$D3,$6E
	.dc.b	$73,$5B,$AC,$1B,$F1,$15,$E8,$97,$8C,$F8,$7D,$4C,$9F,$F2,$04,$E5
	.dc.b	$B3,$EA,$7F,$7D,$19,$7D,$AD,$7F,$D0,$0D,$7F,$82,$60,$F8,$67,$CA
	.dc.b	$7A,$7E,$32,$57,$D0,$E1,$DD,$6A,$30,$7D,$67,$0F,$F8,$7E,$FA,$5F
	.dc.b	$D4,$6C,$7F,$90,$FE,$4B,$14,$F9,$5F,$91,$F1,$52,$BA,$DF,$96,$76
	.dc.b	$48,$CD,$71,$7F,$CF,$BE,$83,$E6,$7E,$67,$C0,$FF,$80,$7A,$3F,$D4
	.dc.b	$CB,$E8,$72,$6B,$B6,$61,$F2,$E7,$FC,$10,$FF,$A9,$DA,$FF,$88,$9A
	.dc.b	$FF,$2D,$7B,$6F,$99,$E7,$3F,$E4,$11,$2F,$84,$FC,$BE,$0F,$C0,$1F
	.dc.b	$6D,$F8,$BF,$B6,$FF,$33,$E1,$FE,$98,$6B,$FE,$4B,$AE,$13,$CA,$B8
	.dc.b	$CF,$FA,$5D,$AE,$DB,$3F,$E5,$FF,$7B,$FD,$90,$FE,$CB,$DD,$FC,$C1
	.dc.b	$2B,$F3,$DE,$F7,$E6,$CC,$3F,$E4,$CE,$34,$E6,$3F,$D2,$9A,$3D,$2F
	.dc.b	$A3,$E7,$49,$CC,$7D,$43,$F3,$0F,$C4,$6C,$F1,$8F,$A5,$18,$38,$3C
	.dc.b	$8F,$F5,$17,$C3,$7E,$A3,$2F,$EB,$DF,$0A,$F7,$7F,$CA,$F3,$9A,$4E
	.dc.b	$5F,$B1,$7F,$E5,$C8,$C4,$78,$CE,$75,$CB,$B7,$FD,$A5,$1C,$A3,$18
	.dc.b	$6A,$39,$94,$E4,$6C,$FE,$E5,$87,$8E,$FB,$B6,$BF,$19,$87,$8D,$B5
	.dc.b	$CE,$74,$F2,$C3,$C7,$7E,$8C,$7E,$44,$F7,$3C,$31,$DA,$77,$7E,$89
	.dc.b	$D5,$70,$AE,$58,$78,$67,$B6,$3E,$1C,$38,$C7,$B8,$E3,$3B,$58,$F9
	.dc.b	$9B,$3C,$1F,$1E,$3E,$C7,$CD,$3B,$38,$63,$E3,$9D,$63,$E9,$C7,$C4
	.dc.b	$7C,$D9,$D5,$F8,$E7,$4F,$4C,$3C,$13,$A8,$F0,$9E,$C3,$55,$CF,$1F
	.dc.b	$54,$EF,$1F,$25,$F2,$9C,$9E,$F8,$CD,$7F,$8B,$7E,$1B,$E4,$E1,$72
	.dc.b	$F7,$47,$90,$F2,$E5,$A9,$F9,$27,$B0,$F1,$98,$23,$C2,$7E,$C4,$F1
	.dc.b	$23,$83,$EE,$47,$52,$38,$24,$FC,$77,$E6,$98,$7E,$BD,$7A,$5F,$BD
	.dc.b	$5D,$9F,$90,$F8,$DE,$44,$38,$9E,$AA,$F0,$B9,$BF,$8E,$A7,$A3,$FA
	.dc.b	$67,$E3,$5F,$8F,$2F,$B0,$7D,$3C,$BF,$15,$7F,$86,$CB,$E9,$E1,$F2
	.dc.b	$71,$E2,$FA,$30,$FA,$27,$C5,$7F,$01,$1E,$47,$F0,$67,$A1,$F6,$38
	.dc.b	$C3,$C1,$97,$D5,$CB,$CF,$1F,$AE,$7B,$CF,$1B,$F9,$6F,$BF,$2F,$39
	.dc.b	$F8,$97,$F1,$C7,$9E,$3C,$8F,$C1,$7C,$EF,$CE,$7E,$16,$3E,$E8,$F9
	.dc.b	$91,$87,$99,$83,$81,$F0,$48,$7D,$4A,$F6,$DF,$CC,$AF,$CD,$5F,$9B
	.dc.b	$0E,$78,$FC,$5C,$BE,$BE,$1F,$02,$FF,$E0,$0B,$F0,$23,$E9,$63,$C3
	.dc.b	$2F,$C5,$46,$23,$F3,$31,$8C,$7E,$14,$60,$FA,$71,$EB,$70,$7C,$B3
	.dc.b	$27,$E3,$72,$FA,$6F,$BC,$FA,$D8,$7A,$B2,$FA,$D5,$F2,$A7,$E6,$1F
	.dc.b	$10,$E8,$70,$3E,$91,$D8,$C1,$D7,$0F,$E9,$0F,$FA,$A4,$FE,$10,$53
	.dc.b	$F3,$E5,$76,$FF,$50,$F8,$4F,$F2,$93,$FE,$4E,$84,$62,$38,$35,$E1
	.dc.b	$E4,$7F,$64,$5E,$5D,$23,$FD,$A1,$BA,$FF,$58,$F8,$17,$BF,$FE,$65
	.dc.b	$AF,$E8,$47,$FB,$06,$BF,$AA,$4E,$BA,$BF,$AB,$4E,$4F,$E9,$11,$97
	.dc.b	$89,$E2,$33,$5D,$6F,$FB,$3C,$71,$7C,$51,$9C,$FF,$CD,$B8,$78,$23
	.dc.b	$15,$F2,$63,$18,$78,$A3,$B5,$F2,$8C,$4E,$A7,$C6,$D9,$F0,$C7,$0B
	.dc.b	$FA,$D8,$7A,$67,$35,$B9,$F1,$D7,$AA,$FA,$B4,$EA,$5A,$E2,$D9,$E3
	.dc.b	$1E,$58,$DB,$4D,$96,$36,$79,$30,$F8,$33,$F8,$BB,$F2,$3F,$0C,$EC
	.dc.b	$5A,$E4,$DA,$E5,$7E,$89,$CC,$F3,$38,$95,$CB,$F5,$67,$8B,$E8,$6C
	.dc.b	$7A,$1A,$7A,$20,$30,$F9,$91,$98,$C6,$1E,$C7,$B1,$C2,$76,$FA,$4C
	.dc.b	$46,$DF,$B0,$62,$FD,$95,$E2,$8E,$07,$68,$CB,$B3,$D4,$5B,$F5,$36
	.dc.b	$7C,$D7,$E2,$8F,$89,$1D,$5A,$6C,$EF,$1B,$7D,$67,$84,$EE,$59,$EE
	.dc.b	$D9,$EF,$7E,$DC,$3E,$EC,$F6,$34,$76,$8D,$3F,$60,$EF,$7E,$17,$E2
	.dc.b	$9C,$0E,$4E,$67,$AC,$70,$9E,$77,$F8,$08,$E3,$5B,$7B,$18,$9D,$DF
	.dc.b	$C3,$8E,$58,$7C,$E7,$C0,$60,$F8,$35,$CE,$FB,$61,$E2,$9E,$95,$C6
	.dc.b	$B5,$7D,$67,$27,$2C,$3C,$2F,$D1,$3D,$AE,$1F,$09,$C3,$0F,$9F,$5D
	.dc.b	$23,$81,$E5,$8C,$CF,$19,$D4,$F2,$C3,$CC,$47,$E1,$B6,$BD,$71,$82
	.dc.b	$38,$3F,$5F,$3F,$C9,$74,$FE,$5D,$3F,$85,$6B,$F1,$DC,$C6,$F1,$FD
	.dc.b	$8C,$6A,$7E,$1B,$AC,$3F,$9C,$36,$7F,$DD,$B3,$F1,$1F,$F7,$7F,$92
	.dc.b	$F5,$9F,$0B,$C2,$7F,$6F,$39,$DB,$16,$8E,$15,$87,$55,$D2,$F9,$B6
	.dc.b	$F5,$18,$B6,$1F,$D3,$6F,$8C,$7F,$B4,$FE,$15,$D1,$E0,$A4,$63,$1F
	.dc.b	$E1,$71,$FE,$D1,$F8,$5D,$4E,$6F,$95,$E0,$F4,$F9,$AE,$B5,$E7,$1F
	.dc.b	$8B,$BB,$E1,$3D,$F8,$FF,$17,$6C,$3E,$16,$C4,$AE,$2E,$A3,$15,$FE
	.dc.b	$F7,$AE,$95,$A3,$BC,$EB,$4F,$ED,$DD,$AF,$83,$FB,$D8,$23,$1C,$7E
	.dc.b	$AF,$19,$F0,$BF,$52,$73,$97,$C4,$3F,$97,$B8,$BF,$9F,$18,$C3,$AE
	.dc.b	$EC,$39,$CF,$1A,$F0,$57,$8A,$FC,$91,$DE,$3F,$9D,$CE,$ED,$F0,$AF
	.dc.b	$47,$57,$2B,$A6,$1C,$9E,$15,$B7,$38,$74,$AF,$39,$F0,$7E,$11,$F0
	.dc.b	$9E,$F6,$D7,$E7,$3F,$2D,$7C,$1E,$6A,$4F,$82,$B7,$7D,$05,$F0,$9D
	.dc.b	$57,$E5,$BC,$19,$1E,$DF,$85,$7C,$2E,$30,$F2,$F5,$FE,$A6,$F1,$C9
	.dc.b	$FE,$27,$E1,$61,$C3,$1F,$D2,$57,$EA,$BC,$27,$C2,$F8,$72,$F7,$C6
	.dc.b	$37,$FC,$8F,$E7,$FF,$0F,$3F,$F1,$94,$E7,$1F,$AC,$DB,$E3,$87,$3A
	.dc.b	$FE,$17,$7C,$6F,$FB,$8D,$EE,$7F,$87,$FB,$EF,$83,$7C,$2F,$F8,$F7
	.dc.b	$D2,$FF,$2B,$4E,$63,$19,$FF,$7A,$F8,$57,$C3,$7F,$91,$7C,$96,$1C
	.dc.b	$31,$E7,$7F,$E5,$08,$C6,$1C,$A1,$8F,$57,$47,$5C,$3A,$57,$08,$C9
	.dc.b	$C2,$FA,$E1,$E0,$8F,$F0,$6D,$F3,$8C,$43,$1F,$3E,$9E,$1A,$D6,$1D
	.dc.b	$6F,$71,$8B,$F1,$56,$F0,$EF,$7B,$24,$E8,$F2,$D7,$17,$CD,$97,$13
	.dc.b	$E0,$63,$E2,$8D,$DF,$47,$D2,$D7,$89,$F0,$23,$86,$3E,$69,$D5,$F7
	.dc.b	$3C,$13,$AC,$3C,$B7,$CA,$F8,$61,$E2,$CB,$E6,$CF,$82,$3C,$CF,$91
	.dc.b	$D6,$58,$AE,$33,$93,$C9,$5C,$9E,$D7,$E0,$C7,$C7,$39,$7C,$D9,$7A
	.dc.b	$0F,$76,$5E,$7B,$F1,$1D,$70,$EE,$7D,$A8,$D4,$76,$37,$1A,$C3,$FD
	.dc.b	$59,$3A,$BE,$F9,$7A,$67,$38,$F8,$F1,$F1,$4E,$71,$FB,$58,$F3,$7A
	.dc.b	$DF,$CA,$C7,$A5,$F0,$9F,$F5,$C1,$F5,$8F,$F5,$D6,$3F,$9B,$6C,$7E
	.dc.b	$89,$CF,$C3,$3E,$1B,$8C,$7E,$8A,$30,$F4,$A4,$62,$38,$CF,$09,$CB
	.dc.b	$F9,$B9,$E3,$8F,$E6,$9B,$8F,$FB,$3F,$E1,$9F,$0D,$FE,$87,$8F,$DE
	.dc.b	$CE,$B1,$FD,$4B,$5F,$26,$1F,$B7,$BE,$31,$DE,$23,$FA,$09,$E1,$3A
	.dc.b	$8C,$4E,$B0,$F2,$3C,$F0,$F1,$4F,$82,$35,$87,$91,$D4,$60,$D1,$D0
	.dc.b	$FE,$07,$7D,$EF,$AC,$65,$A7,$F0,$28,$D5,$75,$9D,$36,$75,$7E,$49
	.dc.b	$F8,$26,$8F,$0C,$69,$F2,$1A,$7B,$CF,$CF,$9E,$58,$FD,$6A,$E4,$6E
	.dc.b	$74,$F2,$8D,$57,$28,$FF,$53,$DF,$A1,$F0,$D7,$9C,$8E,$AF,$BE,$3F
	.dc.b	$67,$0E,$75,$AA,$5F,$37,$C4,$74,$21,$AB,$E7,$8F,$97,$0F,$BB,$18
	.dc.b	$9F,$5E,$1D,$9F,$11,$BA,$F8,$B7,$E9,$E0,$7C,$23,$E1,$7F,$D8,$77
	.dc.b	$FD,$82,$30,$FF,$5C,$6B,$FD,$DF,$F4,$87,$A9,$C3,$93,$FB,$1D,$69
	.dc.b	$B7,$9E,$4F,$EB,$F3,$97,$BD,$F2,$C3,$FC,$C3,$E1,$3E,$1B,$84,$7F
	.dc.b	$97,$63,$0E,$A3,$3D,$9F,$E1,$DF,$86,$DF,$77,$F8,$D6,$6D,$F0,$BF
	.dc.b	$F0,$2B,$74,$FF,$02,$78,$AC,$3B,$D7,$B3,$E1,$AE,$E6,$9A,$FC,$C9
	.dc.b	$E9,$E2,$BF,$CF,$75,$B6,$D6,$E7,$D7,$93,$E0,$9F,$0D,$86,$DF,$F8
	.dc.b	$D1,$AF,$F8,$C3,$7F,$F1,$DF,$C3,$6E,$A4,$ED,$F8,$7E,$5A,$7A,$7C
	.dc.b	$2C,$AE,$8F,$C8,$F3,$5F,$0B,$99,$FF,$1E,$E7,$E5,$AC,$71,$7F,$8A
	.dc.b	$6F,$75,$FE,$98,$F7,$7F,$BD,$CF,$FA,$99,$BC,$FF,$8A,$6C,$82,$F8
	.dc.b	$4F,$9E,$EA,$33,$E0,$E0,$CC,$D8,$F0,$45,$7F,$C9,$13,$CD,$FF,$33
	.dc.b	$5B,$E0,$9F,$07,$F8,$A8,$C3,$63,$33,$FE,$67,$F6,$7F,$D7,$F5,$FE
	.dc.b	$B6,$9C,$CF,$F9,$82,$38,$BD,$A3,$31,$FE,$4D,$9E,$39,$7F,$AD,$B5
	.dc.b	$FF,$53,$DF,$E7,$3E,$0F,$F2,$67,$E5,$3D,$B9,$19,$21,$F9,$67,$E2
	.dc.b	$F9,$F2,$3E,$83,$F9,$6F,$81,$FD,$8E,$5F,$37,$E0,$1F,$03,$F4,$6B
	.dc.b	$85,$FC,$67,$87,$67,$FA,$FA,$15,$C1,$FA,$DC,$0A,$E3,$87,$DD,$78
	.dc.b	$BD,$F0,$FE,$0F,$8F,$8A,$F9,$C7,$85,$D1,$D2,$BF,$86,$E1,$DE,$15
	.dc.b	$C6,$30,$DA,$E5,$3D,$0E,$47,$43,$8E,$3E,$BA,$E2,$D9,$E3,$1D,$F1
	.dc.b	$F6,$57,$1B,$EC,$FE,$55,$AF,$E5,$B0,$F3,$DF,$23,$DF,$87,$59,$E0
	.dc.b	$D7,$81,$E5,$BE,$EF,$A4,$F3,$E3,$E1,$9E,$13,$AC,$7C,$D8,$78,$2F
	.dc.b	$AD,$EC,$E9,$19,$8E,$E7,$5C,$7D,$27,$AA,$FA,$63,$E9,$9E,$66,$F1
	.dc.b	$F0,$DF,$3A,$DC,$EB,$1F,$97,$8F,$82,$21,$C5,$F2,$3E,$D9,$E2,$F9
	.dc.b	$27,$8B,$B3,$55,$DC,$F5,$63,$E8,$C3,$C7,$3D,$AB,$89,$31,$F5,$47
	.dc.b	$57,$E9,$E3,$E2,$9F,$93,$3C,$67,$2F,$86,$32,$F8,$8E,$11,$A8,$C9
	.dc.b	$1C,$12,$3F,$22,$71,$3A,$39,$7C,$86,$A3,$D0,$FC,$B8,$F4,$E3,$F1
	.dc.b	$B0,$F8,$17,$E7,$AE,$15,$CE,$7E,$D6,$3F,$1E,$7C,$E7,$33,$64,$3F
	.dc.b	$1A,$79,$0C,$5F,$7C,$7F,$2D,$5D,$27,$C0,$70,$C7,$CE,$FA,$89,$1E
	.dc.b	$67,$E0,$47,$C5,$7C,$47,$1C,$3D,$B8,$7D,$CB,$EE,$7D,$C9,$E3,$3E
	.dc.b	$D8,$C5,$73,$9E,$47,$4C,$3F,$16,$7A,$71,$F8,$31,$E5,$8C,$3E,$F3
	.dc.b	$18,$78,$67,$73,$CA,$31,$3B,$8C,$3D,$31,$F8,$8F,$98,$F6,$B8,$BF
	.dc.b	$33,$F0,$DF,$44,$F8,$9F,$92,$F4,$9F,$24,$F1,$C7,$F3,$2F,$17,$E3
	.dc.b	$C7,$87,$1F,$87,$1A,$7E,$4E,$3F,$5A,$FB,$4F,$E3,$DE,$D8,$7D,$8C
	.dc.b	$3B,$61,$F8,$37,$B3,$F8,$2C,$7E,$0C,$FE,$D2,$3F,$C7,$E7,$E2,$0B
	.dc.b	$3F,$CE,$DB,$8F,$F4,$77,$3D,$6F,$17,$FD,$4F,$5F,$E2,$26,$C6,$67
	.dc.b	$FD,$67,$39,$7F,$D5,$A3,$CB,$AD,$32,$E0,$8F,$F2,$9F,$71,$ED,$B8
	.dc.b	$BF,$A1,$4F,$F8,$36,$BF,$A2,$56,$A7,$33,$D7,$3F,$EA,$13,$98,$FE
	.dc.b	$68,$71,$3B,$5F,$57,$27,$1D,$BF,$CC,$B3,$C6,$BF,$AD,$4F,$17,$AD
	.dc.b	$71,$73,$1C,$4F,$F1,$14,$F1,$AF,$F6,$BE,$1E,$8B,$E1,$18,$AF,$1E
	.dc.b	$1E,$AA,$D3,$63,$27,$9A,$32,$D3,$38,$78,$A3,$2D,$8C,$D7,$78,$F4
	.dc.b	$D7,$56,$CF,$36,$9C,$4B,$3C,$4F,$8B,$7E,$A9,$FC,$6D,$F8,$27,$51
	.dc.b	$F0,$A7,$4F,$FA,$42,$74,$F0,$3C,$A5,$AE,$ED,$AE,$F7,$E7,$6C,$79
	.dc.b	$DB,$3E,$76,$9E,$28,$61,$F4,$DE,$87,$96,$BA,$34,$E8,$5A,$E8,$DA
	.dc.b	$E9,$5C,$E3,$D8,$7A,$8B,$3E,$A6,$CF,$91,$F0,$9E,$4C,$3E,$2B,$F8
	.dc.b	$D8,$D3,$C0,$F0,$1C,$2B,$C0,$EC,$C9,$B3,$A5,$FB,$1A,$7B,$0B,$7E
	.dc.b	$C6,$D7,$B2,$FC,$71,$F9,$B3,$27,$85,$CE,$1F,$6A,$3C,$18,$79,$1E
	.dc.b	$0D,$8D,$47,$9C,$E6,$70,$9E,$6F,$C9,$35,$3C,$5C,$1D,$23,$77,$F3
	.dc.b	$63,$71,$8B,$F9,$31,$8C,$3E,$15,$F3,$BE,$13,$CE,$73,$7C,$63,$E5
	.dc.b	$9C,$4F,$91,$5C,$6F,$AC,$71,$33,$3C,$5F,$69,$C6,$FE,$75,$71,$8E
	.dc.b	$A7,$18,$CC,$F1,$9D,$4F,$23,$55,$F5,$6F,$CE,$48,$F0,$B6,$F9,$B6
	.dc.b	$78,$4E,$89,$3A,$6C,$6A,$30,$47,$04,$8F,$C1,$4F,$56,$9C,$8B,$3C
	.dc.b	$5B,$3C,$5F,$C0,$4F,$16,$9C,$4B,$3C,$5B,$3C,$5F,$98,$D7,$E2,$9D
	.dc.b	$88,$70,$29,$C0,$A7,$03,$44,$34,$53,$46,$3D,$87,$AB,$FD,$27,$87
	.dc.b	$FA,$57,$23,$D9,$7B,$39,$CF,$FD,$C9,$8B,$1E,$31,$FE,$E7,$C3,$8D
	.dc.b	$F0,$8C,$1C,$27,$75,$FE,$F8,$9D,$2A,$FE,$B4,$F5,$C7,$C0,$B1,$F2
	.dc.b	$63,$88,$D6,$3D,$9D,$61,$D3,$0E,$78,$F8,$1F,$2D,$F1,$8C,$5F,$FB
	.dc.b	$FF,$0E,$AF,$3C,$3A,$9E,$7A,$F0,$7D,$5B,$C5,$F2,$F1,$BF,$49,$7C
	.dc.b	$56,$A7,$31,$8C,$7A,$5F,$17,$D1,$7C,$6F,$84,$62,$B8,$63,$F1,$63
	.dc.b	$E8,$FA,$4F,$55,$EB,$FF,$82,$FC,$1C,$C6,$2F,$FC,$31,$1F,$63,$D7
	.dc.b	$7B,$1F,$E0,$7E,$94,$F0,$C3,$95,$86,$3D,$E3,$31,$8C,$7B,$5F,$28
	.dc.b	$E1,$8F,$6C,$B7,$87,$19,$F1,$9D,$5C,$57,$18,$E9,$87,$6B,$E5,$1C
	.dc.b	$0C,$96,$B8,$E3,$E9,$8D,$4F,$0B,$E9,$5A,$9F,$06,$3E,$8B,$E1,$5E
	.dc.b	$1B,$E1,$18,$25,$70,$C3,$C9,$97,$C9,$9C,$CF,$92,$FB,$E1,$D6,$3E
	.dc.b	$1D,$6D,$E4,$7A,$63,$4F,$1A,$FA,$D3,$B8,$C1,$C2,$BA,$DF,$8B,$2F
	.dc.b	$9B,$87,$2C,$BD,$D7,$C0,$65,$EC,$79,$65,$E8,$9D,$CE,$67,$84,$72
	.dc.b	$33,$97,$8D,$C5,$FE,$36,$7F,$CE,$71,$F7,$5B,$19,$7E,$A1,$AF,$61
	.dc.b	$EB,$FF,$9E,$A3,$F0,$51,$87,$14,$8C,$47,$19,$E0,$FE,$52,$38,$69
	.dc.b	$F7,$9B,$97,$E9,$9B,$3F,$9B,$F5,$DE,$C7,$F8,$E6,$B5,$87,$E5,$AF
	.dc.b	$77,$D6,$3F,$3F,$7C,$23,$51,$1F,$C9,$4F,$09,$D4,$62,$75,$87,$95
	.dc.b	$E7,$87,$86,$7E,$0C,$6B,$0F,$1C,$7E,$7E,$F9,$DF,$18,$E8,$47,$F5
	.dc.b	$11,$A9,$E3,$1B,$23,$FC,$E2,$35,$3E,$03,$94,$70,$35,$7E,$58,$F8
	.dc.b	$46,$8E,$15,$E0,$8D,$63,$F5,$F0,$F5,$56,$AF,$9B,$83,$9D,$F9,$1F
	.dc.b	$A9,$7C,$B1,$FA,$38,$73,$9F,$00,$C7,$D2,$FF,$18,$D7,$FA,$9F,$3F
	.dc.b	$FA,$87,$89,$FA,$4F,$F0,$B8,$C3,$67,$FA,$CE,$9F,$D3,$A7,$FA,$7B
	.dc.b	$63,$F9,$05,$72,$8E,$B1,$8A,$E5,$7F,$C9,$67,$83,$FC,$9F,$D2,$78
	.dc.b	$3C,$23,$F9,$57,$85,$FD,$1D,$FE,$98,$D7,$F9,$E1,$C6,$38,$29,$E6
	.dc.b	$F0,$E8,$CD,$71,$CF,$FB,$3D,$6A,$B9,$9C,$6F,$77,$23,$FA,$E4,$EE
	.dc.b	$B8,$D6,$67,$8D,$F5,$CB,$FB,$FD,$ED,$FF,$7D,$78,$7F,$E0,$2F,$01
	.dc.b	$E0,$7F,$C0,$1C,$BF,$A4,$46,$31,$FE,$95,$E1,$FF,$BF,$97,$88,$C3
	.dc.b	$FD,$FB,$87,$49,$E2,$F3,$57,$CD,$5F,$3B,$E5,$1C,$0E,$F9,$FF,$82
	.dc.b	$DE,$37,$D5,$A7,$F5,$D9,$E2,$E0,$EB,$3C,$6F,$C1,$7D,$F0,$F3,$D7
	.dc.b	$6C,$7C,$0F,$B2,$3E,$26,$3E,$9B,$E3,$87,$38,$FF,$14,$E3,$E0,$9F
	.dc.b	$63,$63,$FC,$5B,$8A,$B8,$61,$DC,$F3,$56,$DF,$56,$1E,$47,$D3,$8F
	.dc.b	$92,$75,$1E,$53,$CF,$7D,$66,$DF,$69,$D1,$DA,$FA,$57,$C1,$8D,$D4
	.dc.b	$74,$77,$9D,$D7,$3B,$F0,$9E,$2B,$F0,$57,$CF,$7A,$36,$7C,$51,$8C
	.dc.b	$7E,$4E,$1D,$DF,$8F,$87,$CE,$BF,$82,$67,$1F,$BD,$3A,$8C,$1F,$25
	.dc.b	$C3,$E8,$3C,$15,$A7,$E0,$C6,$8F,$0B,$A3,$83,$99,$FA,$F8,$7A,$0F
	.dc.b	$81,$7E,$5A,$E7,$3A,$BE,$D8,$FC,$4A,$F8,$F1,$F2,$6A,$1D,$F0,$F0
	.dc.b	$9D,$DF,$79,$9B,$F7,$9F,$8F,$2B,$A8,$E3,$18,$8F,$19,$F0,$DC,$5F
	.dc.b	$D5,$C3,$D5,$7E,$2B,$F0,$5F,$E3,$0F,$01,$D6,$FC,$0F,$33,$04,$38
	.dc.b	$47,$89,$C5,$F7,$AF,$87,$87,$C7,$8C,$4F,$63,$CE,$4C,$7F,$13,$7C
	.dc.b	$E3,$33,$C2,$30,$FC,$D8,$FB,$93,$F8,$BC,$3E,$94,$75,$9D,$3D,$E7
	.dc.b	$4F,$AC,$F4,$CF,$EE,$2F,$CC,$7D,$88,$FC,$1C,$7C,$27,$67,$BA,$35
	.dc.b	$7F,$2A,$FC,$D5,$F4,$27,$2F,$5A,$FA,$0F,$9A,$BD,$73,$E0,$9F,$5E
	.dc.b	$3F,$8E,$76,$79,$F0,$F4,$BE,$22,$61,$EE,$BF,$91,$3A,$3E,$5E,$1E
	.dc.b	$3A,$EB,$8F,$47,$E8,$61,$F8,$47,$DE,$7D,$DA,$E2,$FC,$7C,$3B,$5F
	.dc.b	$A4,$E9,$1F,$36,$BD,$0F,$23,$CD,$5E,$38,$ED,$87,$C7,$C7,$E2,$D7
	.dc.b	$CB,$3C,$A7,$1C,$3E,$5B,$E1,$8F,$8C,$F7,$C7,$D5,$3A,$8F,$87,$3A
	.dc.b	$AF,$87,$7E,$47,$F3,$F8,$7B,$2B,$67,$C3,$70,$7F,$1A,$7F,$DC,$67
	.dc.b	$E9,$4A,$E5,$FC,$79,$8D,$1C,$1F,$E5,$A9,$FF,$03,$B6,$33,$3F,$E6
	.dc.b	$59,$CB,$FE,$60,$1C,$B5,$A6,$5C,$11,$FE,$B9,$E1,$BC,$46,$FF,$D6
	.dc.b	$27,$FD,$7D,$5F,$D6,$AB,$53,$99,$EB,$97,$F7,$19,$CC,$7F,$67,$38
	.dc.b	$9D,$AF,$AB,$93,$8E,$FF,$EC,$B5,$3C,$AB,$FB,$E4,$72,$7B,$57,$27
	.dc.b	$31,$C8,$FF,$6A,$4F,$1A,$FF,$45,$61,$E9,$BE,$11,$8A,$F2,$61,$EB
	.dc.b	$AD,$36,$32,$79,$E3,$2D,$33,$87,$8A,$32,$D8,$CD,$77,$8F,$4D,$75
	.dc.b	$6C,$F3,$69,$C4,$B3,$C4,$F8,$D8,$7B,$67,$F1,$B7,$E0,$9D,$47,$C2
	.dc.b	$C3,$DE,$E8,$F1,$96,$FA,$B6,$7A,$CF,$08,$FA,$95,$D5,$D1,$D7,$0F
	.dc.b	$55,$70,$6C,$64,$F6,$46,$5A,$66,$FD,$D1,$96,$C6,$63,$D6,$FE,$12
	.dc.b	$FC,$8F,$E0,$E3,$81,$AC,$3E,$24,$7A,$0F,$79,$6B,$DE,$D9,$E2,$FB
	.dc.b	$0E,$D7,$EA,$69,$E8,$2C,$F9,$9B,$3E,$1B,$F2,$C7,$E4,$CC,$9D,$6F
	.dc.b	$CA,$FA,$4F,$71,$98,$E3,$87,$CF,$8F,$31,$A2,$9D,$0B,$3D,$1F,$03
	.dc.b	$63,$94,$77,$38,$16,$34,$FC,$63,$13,$E2,$70,$74,$71,$7F,$2A,$37
	.dc.b	$87,$C7,$AE,$73,$99,$E1,$39,$AE,$38,$7D,$09,$F4,$C6,$1F,$71,$D0
	.dc.b	$FA,$17,$D2,$78,$46,$6F,$E9,$4E,$63,$89,$E0,$38,$D7,$57,$E0,$99
	.dc.b	$9E,$31,$99,$E5,$3A,$C3,$E4,$BB,$BF,$BD,$87,$AC,$91,$D9,$B8,$74
	.dc.b	$6C,$F0,$9D,$12,$74,$D8,$D4,$60,$8E,$09,$1F,$87,$9E,$AD,$39,$16
	.dc.b	$78,$B6,$78,$BF,$42,$78,$B4,$E2,$59,$E2,$D9,$E2,$FD,$16,$BF,$14
	.dc.b	$EC,$43,$81,$4E,$05,$38,$1A,$21,$A2,$9A,$F0,$9E,$27,$F4,$AE,$1F
	.dc.b	$E9,$7E,$1B,$C4,$78,$99,$CF,$FA,$66,$1C,$16,$3F,$EE,$CB,$E1,$5B
	.dc.b	$8C,$D6,$DD,$5F,$4B,$E3,$87,$F5,$A9,$56,$E3,$C7,$5B,$9F,$F7,$7C
	.dc.b	$E6,$BC,$38,$78,$1E,$67,$9B,$1F,$14,$66,$B8,$C6,$2F,$A6,$1E,$1C
	.dc.b	$3A,$52,$F8,$4F,$97,$C4,$73,$FB,$78,$DE,$1B,$C4,$F8,$18,$AF,$7A
	.dc.b	$F9,$23,$FC,$27,$5B,$8C,$46,$F0,$E9,$39,$86,$3E,$19,$CE,$1D,$BB
	.dc.b	$5E,$0B,$C4,$FF,$06,$F6,$FF,$80,$67,$31,$8C,$7E,$BF,$88,$F1,$3F
	.dc.b	$C1,$3C,$EC,$38,$63,$CE,$FF,$C4,$31,$8C,$39,$43,$1E,$AE,$8E,$B8
	.dc.b	$74,$AE,$11,$93,$85,$F5,$C3,$C1,$1F,$8A,$BE,$71,$88,$63,$E7,$DB
	.dc.b	$C3,$5A,$C3,$AD,$EE,$31,$7E,$2A,$DE,$1D,$EF,$64,$9D,$1E,$5A,$E2
	.dc.b	$F9,$B2,$E2,$7C,$0C,$7C,$51,$BB,$E8,$FA,$5A,$F1,$3E,$04,$70,$C7
	.dc.b	$CD,$3A,$BE,$E7,$82,$75,$87,$96,$F9,$5F,$0C,$3C,$59,$7C,$D9,$F0
	.dc.b	$47,$99,$F2,$3A,$CB,$15,$C6,$72,$79,$2B,$93,$DA,$FC,$18,$F8,$E7
	.dc.b	$2F,$9B,$2F,$C2,$57,$F9,$CA,$3F,$02,$D8,$CB,$F8,$03,$5E,$13,$C4
	.dc.b	$FF,$3C,$C7,$DF,$8C,$38,$A4,$62,$38,$CF,$07,$F2,$11,$C3,$1F,$C0
	.dc.b	$37,$1F,$D0,$36,$7F,$8A,$F8,$6F,$13,$FC,$63,$5A,$C3,$F1,$77,$BB
	.dc.b	$EB,$1F,$9E,$BE,$11,$A8,$8F,$E1,$E7,$84,$EA,$31,$3A,$C3,$CA,$F3
	.dc.b	$C3,$C3,$3F,$06,$35,$87,$8E,$3F,$4D,$7C,$EF,$8C,$74,$23,$FA,$58
	.dc.b	$D4,$F1,$8D,$91,$FF,$41,$46,$A7,$C0,$72,$8E,$06,$AF,$CB,$1F,$08
	.dc.b	$D1,$C2,$BC,$11,$AC,$7E,$BE,$1E,$AA,$D5,$F3,$70,$73,$BF,$23,$F5
	.dc.b	$2F,$96,$3F,$47,$0E,$73,$E0,$18,$FA,$5F,$E2,$DA,$7F,$A7,$78,$7F
	.dc.b	$D4,$3C,$8F,$13,$F8,$44,$61,$B3,$FE,$A3,$D3,$FA,$74,$FF,$4F,$6C
	.dc.b	$7F,$5A,$AE,$51,$D6,$31,$5C,$AF,$FA,$FC,$F0,$7F,$B0,$F8,$8F,$07
	.dc.b	$84,$7F,$64,$F0,$BF,$A3,$BF,$CF,$9A,$FF,$3C,$C3,$FD,$6D,$5C,$63
	.dc.b	$FA,$1B,$5D,$1C,$F9,$BA,$AA,$E7,$9F,$F6,$9A,$E0,$DC,$37,$3A,$C3
	.dc.b	$FD,$88,$DB,$F4,$CF,$4B,$E3,$5A,$C3,$89,$99,$F0,$F8,$5F,$EF,$7F
	.dc.b	$11,$E2,$7F,$BF,$7C,$1F,$F6,$84,$FF,$BE,$33,$FF,$7F,$47,$FB,$4E
	.dc.b	$B7,$9F,$F8,$23,$C2,$4F,$0A,$FE,$BF,$29,$CB,$FE,$0E,$BF,$06,$1D
	.dc.b	$E3,$15,$C2,$B7,$0A,$DB,$DE,$FC,$57,$FE,$E5,$AD,$D7,$9F,$5D,$5F
	.dc.b	$F8,$73,$1F,$3B,$E3,$C3,$8B,$A3,$B6,$3E,$18,$F5,$56,$6B,$8E,$1E
	.dc.b	$46,$CE,$AF,$A1,$EA,$AE,$18,$FB,$63,$AC,$77,$C7,$D7,$1C,$05,$F4
	.dc.b	$79,$19,$8D,$E1,$E1,$BF,$81,$87,$86,$3E,$0E,$3F,$5A,$FE,$05,$69
	.dc.b	$E7,$3A,$C7,$CB,$10,$ED,$5F,$05,$DD,$FC,$67,$CD,$8F,$95,$F5,$9C
	.dc.b	$2B,$95,$F9,$63,$DD,$7C,$DD,$C6,$5E,$78,$79,$5C,$E1,$EB,$3C,$13
	.dc.b	$B3,$D1,$1D,$CE,$27,$82,$38,$3E,$83,$E0,$99,$30,$76,$32,$72,$3A
	.dc.b	$47,$63,$C1,$1E,$23,$AC,$71,$3E,$BC,$FD,$DC,$7F,$21,$1C,$EB,$8B
	.dc.b	$E0,$8F,$B3,$5C,$E3,$17,$F5,$DE,$67,$C8,$C2,$3C,$E7,$83,$E1,$AE
	.dc.b	$6F,$AA,$FD,$13,$E6,$C7,$EB,$1F,$5E,$BC,$B3,$E4,$AE,$E7,$C0,$AE
	.dc.b	$18,$FA,$F1,$FB,$17,$DA,$B7,$7D,$9F,$97,$87,$6B,$FB,$55,$DB,$0F
	.dc.b	$3C,$F8,$70,$E6,$7B,$DC,$63,$EE,$C7,$C9,$18,$C7,$C8,$7B,$6F,$E0
	.dc.b	$E1,$D6,$30,$43,$E2,$61,$D7,$1F,$1B,$E0,$C7,$F0,$53,$EA,$BF,$B7
	.dc.b	$7E,$DC,$3C,$18,$FB,$6B,$47,$59,$F0,$38,$23,$89,$F8,$51,$EB,$AF
	.dc.b	$96,$F8,$27,$58,$FC,$DC,$35,$3E,$9B,$F0,$DF,$08,$FC,$01,$F8,$69
	.dc.b	$F4,$CE,$A3,$E8,$D7,$37,$E3,$9E,$58,$D3,$F9,$77,$CA,$EA,$FE,$04
	.dc.b	$FC,$6A,$F1,$CF,$18,$C3,$D2,$B8,$E1,$F1,$6F,$8C,$F0,$6B,$D7,$1F
	.dc.b	$A4,$E5,$AE,$70,$F5,$E1,$DA,$FE,$86,$1D,$B1,$FA,$B8,$70,$9F,$03
	.dc.b	$63,$37,$CB,$0E,$CD,$7B,$1F,$28,$FA,$F3,$F8,$37,$17,$F1,$B1,$F4
	.dc.b	$E1,$DA,$F9,$36,$BF,$CB,$F3,$FD,$DE,$7F,$CB,$4D,$8C,$DF,$F6,$07
	.dc.b	$FD,$D6,$7F,$5F,$C9,$C1,$C1,$B1,$D1,$AF,$03,$FB,$03,$87,$FC,$DF
	.dc.b	$CB,$FE,$69,$7F,$DE,$ED,$4D,$23,$FD,$37,$C3,$78,$8E,$0F,$EC,$13
	.dc.b	$C8,$D4,$65,$FE,$A8,$67,$4F,$EE,$75,$CA,$B8,$CF,$80,$E2,$E2,$FF
	.dc.b	$B6,$4E,$E7,$99,$BA,$E7,$7D,$DF,$EF,$F1,$DF,$6F,$93,$5C,$67,$33
	.dc.b	$C6,$73,$5C,$5F,$8B,$1E,$23,$BD,$F8,$A7,$51,$F3,$63,$18,$79,$E3
	.dc.b	$85,$F6,$9C,$C7,$23,$37,$DA,$7D,$35,$C5,$B3,$B6,$9C,$DE,$27,$84
	.dc.b	$A7,$E1,$AB,$BC,$6C,$E3,$5D,$CF,$7D,$77,$6D,$77,$6C,$F1,$69,$C6
	.dc.b	$03,$0F,$99,$19,$8C,$61,$F1,$9F,$21,$E6,$3E,$A1,$CA,$78,$BE,$BA
	.dc.b	$F3,$B6,$BC,$2D,$38,$57,$57,$99,$E0,$BF,$63,$A6,$B9,$3D,$C6,$4A
	.dc.b	$66,$BE,$03,$96,$B9,$AF,$75,$FA,$23,$DB,$3E,$86,$CF,$A1,$A7,$37
	.dc.b	$8D,$FA,$63,$F2,$46,$4F,$71,$DC,$B5,$DD,$F7,$B5,$E2,$6C,$EE,$53
	.dc.b	$71,$DF,$0F,$C2,$D7,$8B,$0F,$8B,$7C,$E7,$37,$E0,$BE,$0F,$D9,$3B
	.dc.b	$CE,$DC,$4F,$07,$C3,$19,$C3,$E2,$47,$8A,$FB,$DF,$49,$CC,$F4,$8C
	.dc.b	$61,$F2,$63,$81,$F5,$AF,$A4,$66,$38,$5F,$42,$3B,$AE,$6F,$CD,$3C
	.dc.b	$AE,$1F,$8C,$72,$38,$47,$2B,$FC,$0D,$F8,$27,$91,$E4,$8C,$D7,$39
	.dc.b	$D4,$F4,$C3,$E2,$3B,$C3,$E8,$3D,$89,$7C,$B0,$F0,$4E,$A3,$13,$A8
	.dc.b	$C1,$1F,$AD,$3C,$1A,$6C,$B3,$C8,$93,$C6,$3E,$1C,$F1,$69,$C4,$B3
	.dc.b	$C1,$E7,$3C,$5F,$61,$82,$1C,$1D,$3F,$36,$39,$39,$36,$72,$30,$43
	.dc.b	$04,$39,$47,$4A,$F9,$B1,$83,$99,$4E,$45,$39,$3C,$08,$FE,$68,$93
	.dc.b	$F9,$48,$CB,$FB,$DF,$CE,$5F,$5F,$FD,$77,$7F,$EB,$D9,$3F,$3C,$79
	.dc.b	$B9,$D3,$FB,$CE,$1C,$1F,$E5,$37,$C2,$B7,$18,$37,$87,$1C,$FF,$97
	.dc.b	$4E,$9F,$F8,$02,$2F,$A2,$BE,$98,$71,$8D,$5F,$3C,$7C,$D3,$9B,$E3
	.dc.b	$7C,$31,$F1,$61,$FC,$FB,$F3,$A7,$CE,$78,$E9,$F6,$2F,$AD,$ED,$5A
	.dc.b	$C3,$8D,$6A,$73,$7C,$6F,$85,$76,$C1,$87,$17,$D5,$1F,$1A,$77,$1F
	.dc.b	$02,$75,$87,$D1,$9C,$BF,$E0,$F9,$E7,$7D,$F0,$F0,$E1,$E4,$BD,$BF
	.dc.b	$E2,$CC,$39,$BD,$E7,$E4,$7E,$82,$F3,$A9,$F2,$7F,$20,$7C,$86,$2F
	.dc.b	$D5,$5B,$7E,$4F,$8E,$FD,$15,$E0,$6F,$6F,$68,$9D,$46,$77,$FF,$14
	.dc.b	$46,$2F,$AE,$7F,$83,$8C,$5F,$E0,$BC,$C7,$B7,$F8,$7E,$CF,$AB,$F6
	.dc.b	$5F,$81,$9C,$E3,$F7,$27,$2F,$D1,$87,$BF,$C7,$8D,$F1,$3C,$63,$0F
	.dc.b	$E2,$FC,$87,$99,$E9,$D3,$FC,$9C,$23,$3A,$E2,$31,$3B,$7F,$CA,$93
	.dc.b	$A7,$8A,$F2,$9E,$67,$18,$C3,$63,$33,$C1,$AF,$F9,$7F,$EC,$9F,$63
	.dc.b	$FE,$68,$E3,$FF,$29,$70,$3E,$CD,$E2,$67,$8B,$FC,$37,$39,$8C,$63
	.dc.b	$FE,$1E,$C7,$9B,$F9,$75,$C7,$AA,$E9,$95,$C7,$77,$FB,$6B,$DB,$F9
	.dc.b	$FF,$D1,$1F,$61,$DA,$3C,$FF,$25,$70,$39,$AD,$CF,$1A,$D4,$C8,$C1
	.dc.b	$0F,$E0,$B3,$C1,$B3,$D7,$E7,$7C,$1F,$9E,$BD,$EF,$16,$6F,$9B,$A6
	.dc.b	$A2,$61,$FB,$79,$D3,$67,$2F,$F1,$4A,$E8,$E4,$86,$6B,$7C,$3F,$B7
	.dc.b	$BE,$73,$FC,$71,$B1,$FB,$BA,$E7,$3F,$C8,$39,$5F,$73,$F2,$7E,$85
	.dc.b	$F6,$74,$F9,$93,$AA,$FE,$1C,$D7,$F9,$4C,$F0,$7F,$89,$35,$ED,$5C
	.dc.b	$8E,$8E,$AB,$0F,$F1,$9F,$4B,$FD,$7F,$CB,$FE,$CA,$8F,$F6,$1F,$A9
	.dc.b	$FE,$D3,$8F,$F7,$97,$9F,$F9,$09,$FF,$34,$FA,$9F,$EC,$7F,$53,$FD
	.dc.b	$A5,$1F,$EF,$53,$89,$FE,$D2,$29,$FE,$D2,$9E,$4F,$FB,$B4,$D3,$89
	.dc.b	$FF,$70,$47,$38,$C5,$FF,$BF,$5F,$F7,$2B,$6F,$CA,$7F,$B3,$AB,$C9
	.dc.b	$B7,$FB,$62,$3D,$95,$CA,$FF,$D0,$2D,$7C,$91,$C6,$FA,$47,$FA,$1B
	.dc.b	$1F,$8A,$F8,$E7,$4F,$26,$BC,$8F,$8B,$3C,$94,$F2,$6E,$3F,$EF,$1B
	.dc.b	$F1,$12,$FC,$E7,$47,$13,$E1,$3A,$9E,$12,$9E,$13,$E6,$17,$1F,$96
	.dc.b	$F8,$31,$E0,$D3,$EA,$CF,$FB,$F4,$EE,$A7,$E5,$BA,$AF,$2B,$E0,$9D
	.dc.b	$E1,$FF,$00,$5F,$AB,$1D,$AB,$8C,$74,$6D,$7B,$49,$3E,$D3,$D3,$3E
	.dc.b	$38,$CC,$78,$F1,$FC,$A3,$F5,$71,$FC,$8B,$EB,$AF,$0A,$7C,$B1,$F3
	.dc.b	$EB,$9D,$F8,$AB,$D5,$1D,$B0,$F5,$3F,$8C,$8C,$9E,$59,$F9,$73,$A8
	.dc.b	$F1,$63,$F8,$F7,$58,$FE,$06,$7F,$00,$6E,$3C,$D7,$E2,$9F,$5C,$60
	.dc.b	$EE,$7B,$CF,$D2,$C6,$DA,$6A,$FD,$4F,$C8,$8F,$AA,$D3,$C6,$7D,$39
	.dc.b	$EF,$5E,$4C,$3E,$24,$F3,$9C,$C7,$CC,$9F,$41,$BC,$3F,$12,$DA,$F9
	.dc.b	$D7,$F1,$DF,$41,$F9,$1B,$F8,$57,$CC,$EF,$1E,$53,$E0,$13,$0F,$9B
	.dc.b	$1A,$6E,$1E,$08,$E4,$D3,$E8,$4F,$2C,$3C,$98,$FE,$08,$FC,$0C,$EE
	.dc.b	$7F,$34,$7C,$43,$D2,$7D,$09,$F7,$D7,$68,$C9,$E5,$C3,$D7,$5E,$0A
	.dc.b	$DB,$6B,$D1,$3A,$AE,$E7,$96,$3A,$5F,$F2,$2B,$47,$F9,$86,$7F,$C2
	.dc.b	$71,$8B,$FF,$09,$36,$33,$3B,$44,$3F,$87,$FA,$8E,$9E,$2D,$8E,$AD
	.dc.b	$78,$1C,$23,$FC,$40,$F4,$F4,$EA,$69,$1D,$BF,$E8,$ED,$3F,$DB,$8D
	.dc.b	$8F,$F6,$79,$FE,$99,$23,$97,$FD,$9C,$7F,$A6,$23,$8B,$4E,$31,$BF
	.dc.b	$47,$FD,$AA,$4F,$53,$FD,$B0,$4E,$EF,$F6,$15,$7F,$A9,$8E,$47,$6D
	.dc.b	$7F,$D4,$8E,$2B,$94,$F0,$8F,$29,$9A,$E8,$FF,$B2,$CE,$95,$C8,$98
	.dc.b	$7A,$1B,$1E,$83,$95,$75,$C3,$D5,$7D,$2B,$FD,$97,$5D,$2F,$91,$1E
	.dc.b	$51,$D1,$F5,$46,$5A,$69,$C6,$1E,$C8,$F4,$1E,$C9,$E5,$19,$7C,$31
	.dc.b	$99,$E3,$5E,$57,$FD,$AD,$1E,$A3,$9D,$FA,$B0,$F6,$46,$6F,$9C,$E6
	.dc.b	$36,$66,$B9,$CF,$8A,$FB,$4F,$FB,$82,$B8,$4F,$28,$E0,$64,$A7,$FB
	.dc.b	$92,$37,$5D,$0F,$4D,$F9,$2B,$B3,$63,$35,$E4,$9C,$B6,$33,$1E,$13
	.dc.b	$25,$72,$77,$8F,$A0,$43,$E9,$93,$0F,$9D,$1B,$8C,$61,$F0,$1F,$51
	.dc.b	$EC,$3F,$04,$73,$9E,$2F,$C1,$AF,$1D,$73,$9C,$CF,$F0,$1C,$3F,$BE
	.dc.b	$C2,$D8,$7F,$80,$3D,$1F,$F7,$3B,$FE,$F6,$3F,$DC,$FF,$66,$FC,$25
	.dc.b	$EB,$FF,$BB,$7D,$69,$5F,$EE,$CC,$7F,$DE,$2F,$FB,$87,$0E,$79,$FF
	.dc.b	$BF,$BD,$6D,$BF,$EC,$6A,$DB,$FF,$00,$27,$71,$8E,$3F,$F7,$07,$AB
	.dc.b	$FE,$DA,$AF,$F7,$C7,$AD,$FE,$D7,$EF,$FF,$70,$C7,$FB,$A6,$3F,$DC
	.dc.b	$99,$7F,$BC,$6B,$85,$FF,$BD,$E3,$13,$AA,$E1,$1F,$EE,$9F,$5B,$9D
	.dc.b	$7F,$BC,$5B,$1F,$EE,$93,$FD,$AE,$49,$F5,$7A,$DF,$EE,$88,$C5,$7F
	.dc.b	$BD,$FD,$5F,$F7,$45,$7F,$B9,$3C,$DF,$F6,$EC,$E7,$D3,$FF,$70,$7A
	.dc.b	$BF,$EE,$AC,$38,$9F,$CE,$2B,$6F,$F3,$9C,$39,$DF,$FB,$AA,$B8,$61
	.dc.b	$FE,$E7,$F5,$3F,$DB,$DE,$6F,$FB,$95,$B8,$7F,$B8,$FD,$2F,$F6,$FB
	.dc.b	$4F,$F7,$47,$A7,$FE,$E4,$7F,$A3,$E3,$F8,$4C,$3F,$A4,$FA,$5F,$ED
	.dc.b	$B7,$FD,$BB,$E3,$FF,$B7,$DF,$E9,$91,$96,$C6,$5E,$3E,$8F,$FB,$69
	.dc.b	$DB,$5C,$FA,$5F,$EE,$13,$FD,$D9,$E9,$7F,$B3,$B0,$FF,$6E,$F9,$9F
	.dc.b	$E8,$46,$F8,$3F,$EC,$D9,$D6,$7F,$D7,$5F,$F6,$9E,$0A,$E3,$E9,$7F
	.dc.b	$B6,$4E,$1C,$1F,$ED,$1F,$43,$FD,$9D,$5B,$C3,$FB,$35,$7F,$B5,$A3
	.dc.b	$FD,$A1,$7F,$0D,$7A,$3F,$ED,$18,$C4,$EE,$79,$1F,$DC,$1C,$47,$19
	.dc.b	$E1,$5F,$DC,$DB,$5F,$DD,$9A,$F6,$BF,$F6,$E3,$F5,$27,$FB,$D7,$9F
	.dc.b	$FE,$C8,$BF,$F6,$77,$67,$FB,$72,$B4,$7F,$80,$27,$55,$C3,$D0,$D4
	.dc.b	$FF,$B4,$EB,$FD,$AF,$39,$8F,$F6,$DD,$F4,$79,$8B,$E6,$F5,$BE,$2D
	.dc.b	$3F,$DD,$5B,$FF,$B2,$9F,$F6,$9F,$9D,$FE,$CE,$7F,$DA,$27,$FB,$63
	.dc.b	$CF,$C6,$1F,$37,$D0,$FF,$6B,$3F,$3F,$93,$FD,$B8,$D9,$E8,$DA,$FF
	.dc.b	$6E,$FB,$9F,$E1,$D8,$FA,$DE,$7F,$F8,$FB,$93,$FC,$A1,$E8,$FF,$98
	.dc.b	$8F,$F2,$3D,$FF,$8E,$27,$4F,$F9,$D8,$FF,$19,$BF,$E3,$86,$FA,$1F
	.dc.b	$E3,$87,$FC,$CC,$7F,$9B,$8F,$F2,$B4,$6F,$CE,$FF,$4D,$7A,$3F,$E6
	.dc.b	$6C,$7F,$C8,$07,$F9,$36,$33,$5C,$70,$F1,$C7,$F9,$4F,$1F,$F4,$76
	.dc.b	$3D,$EB,$FC,$9B,$5E,$9C,$3F,$C9,$17,$CA,$7F,$CD,$47,$90,$A7,$08
	.dc.b	$F1,$3E,$AC,$3C,$58,$7F,$A1,$E3,$75,$FE,$62,$7E,$26,$1F,$E5,$29
	.dc.b	$F9,$37,$CA,$BC,$07,$FA,$C6,$BF,$D0,$87,$AE,$3C,$B3,$D6,$75,$3F
	.dc.b	$1A,$BD,$91,$DC,$F8,$11,$FE,$6D,$8C,$CF,$C2,$BF,$3C,$F6,$3C,$F1
	.dc.b	$97,$CF,$1F,$39,$A7,$BC,$B5,$E8,$9F,$F5,$0E,$1F,$41,$F6,$CF,$A4
	.dc.b	$F9,$53,$E8,$9D,$47,$E1,$63,$93,$FE,$85,$7D,$53,$CA,$3C,$A7,$AE
	.dc.b	$BD,$2F,$C1,$6D,$74,$8F,$F4,$41,$E0,$9F,$2E,$1F,$2A,$3E,$05,$F8
	.dc.b	$E3,$11,$CC,$F6,$CF,$CF,$38,$BC,$CF,$33,$F1,$E7,$E7,$C7,$FA,$C9
	.dc.b	$E8,$EC,$FF,$49,$5F,$CB,$8D,$DF,$E4,$67,$58,$FE,$7E,$FC,$11,$D8
	.dc.b	$F8,$07,$90,$F4,$BE,$47,$F2,$87,$CF,$29,$F2,$0F,$23,$BA,$F9,$B1
	.dc.b	$F4,$63,$BB,$E6,$3E,$F4,$79,$B0,$FC,$B4,$EA,$BC,$D5,$E2,$3C,$C7
	.dc.b	$63,$A1,$FC,$04,$FF,$4F,$BF,$EA,$B6,$E3,$FE,$AB,$7F,$D5,$51,$96
	.dc.b	$FA,$3F,$EA,$76,$FA,$3F,$EA,$C2,$3F,$EA,$22,$1F,$EA,$C3,$FD,$54
	.dc.b	$60,$FF,$1C,$7A,$1F,$E4,$17,$FD,$2D,$F3,$EC,$7F,$D7,$5E,$5F,$F8
	.dc.b	$E7,$9F,$FD,$7B,$39,$AF,$F5,$BF,$A5,$8B,$DE,$1F,$DE,$EB,$FC,$AD
	.dc.b	$5F,$E4,$98,$EC,$FF,$9F,$4C,$5F,$F9,$8D,$A7,$FB,$78,$F3,$CF,$75
	.dc.b	$87,$8D,$46,$5D,$DF,$8A,$BC,$91,$E6,$37,$19,$69,$99,$ED,$5D,$5F
	.dc.b	$4C,$7F,$99,$AB,$FD,$1B,$1A,$BF,$75,$FC,$DC,$3B,$DF,$9A,$72,$FF
	.dc.b	$99,$E7,$FC,$EE,$FF,$A2,$6B,$95,$F0,$8F,$49,$FE,$EA,$C3,$C5,$5F
	.dc.b	$5A,$7E,$34,$FF,$B5,$AB,$8B,$EA,$3F,$CE,$05,$AE,$2F,$84,$F9,$BB
	.dc.b	$7F,$A6,$E7,$89,$F2,$63,$FD,$D8,$71,$C3,$D2,$FF,$A1,$27,$FC,$FE
	.dc.b	$FA,$70,$F9,$55,$C7,$1F,$83,$87,$82,$B8,$57,$C0,$7F,$D1,$B5,$E0
	.dc.b	$8F,$91,$1D,$F1,$F8,$4F,$8A,$74,$FD,$59,$EE,$70,$8C,$C7,$C0,$6B
	.dc.b	$F2,$8F,$A5,$3D,$2F,$CB,$8F,$B2,$75,$87,$9D,$B1,$97,$F2,$C7,$CF
	.dc.b	$C3,$C2,$F8,$CF,$AD,$1B,$AF,$F6,$0C,$7C,$F3,$F1,$07,$FA,$EC,$FA
	.dc.b	$33,$F6,$DC,$1D,$4E,$77,$E7,$7E,$21,$F7,$4F,$F5,$99,$D3,$1F,$CA
	.dc.b	$E1,$FE,$B0,$8C,$B5,$F4,$B6,$FE,$0D,$75,$C7,$E3,$D6,$B1,$FD,$AD
	.dc.b	$7A,$A3,$EC,$9E,$83,$D9,$7C,$D6,$1F,$92,$3F,$04,$FA,$2E,$9D,$AB
	.dc.b	$A4,$79,$AF,$BB,$FE,$CF,$3C,$58,$7E,$BB,$1F,$9F,$1F,$75,$AF,$18
	.dc.b	$FD,$1E,$3F,$C0,$F1,$EF,$5A,$AF,$C5,$37,$0F,$83,$1F,$2C,$C9,$F4
	.dc.b	$0C,$3E,$67,$27,$E2,$8C,$9E,$D8,$FB,$53,$F4,$B0,$F5,$BF,$0E,$3F
	.dc.b	$2C,$78,$67,$81,$F2,$A7,$FD,$41,$3F,$C7,$7C,$DF,$F6,$CC,$9F,$F6
	.dc.b	$DF,$A5,$FE,$E3,$F4,$7F,$DB,$7C,$7F,$E8,$59,$7A,$3F,$E8,$7B,$FF
	.dc.b	$75,$1F,$CF,$E3,$3A,$FF,$41,$8E,$93,$B4,$62,$3F,$DD,$9E,$CB,$D2
	.dc.b	$EF,$1F,$C4,$E3,$0F,$FB,$70,$93,$FE,$D4,$EF,$FF,$6C,$BF,$EE,$88
	.dc.b	$F0,$7A,$4F,$4B,$C1,$28,$C4,$EA,$73,$1F,$EE,$AF,$29,$E9,$74,$D7
	.dc.b	$FD,$D9,$87,$FB,$A8,$FF,$50,$B8,$3F,$DF,$37,$FE,$EF,$7F,$DC,$18
	.dc.b	$7F,$B7,$A7,$FA,$4F,$A5,$FE,$E9,$7F,$D5,$23,$D0,$FF,$6F,$CF,$FB
	.dc.b	$87,$D2,$F3,$C7,$FB,$B2,$7E,$39,$F2,$1C,$3B,$9F,$9D,$1C,$A1,$E9
	.dc.b	$7F,$BB,$1E,$15,$FE,$F2,$F6,$3F,$DD,$17,$F5,$7D,$87,$A7,$F1,$9B
	.dc.b	$19,$8C,$7A,$72,$38,$37,$0F,$EF,$DE,$9F,$FB,$95,$FF,$75,$F8,$FF
	.dc.b	$EE,$55,$19,$6C,$66,$1E,$9F,$FB,$9E,$76,$D8,$CF,$97,$FE,$E9,$9F
	.dc.b	$F0,$67,$A4,$F4,$7F,$1D,$08,$C3,$FE,$EE,$8C,$C6,$23,$8F,$A5,$82
	.dc.b	$3F,$EF,$47,$9F,$A5,$FE,$D6,$AF,$F6,$EE,$9F,$EE,$8F,$5F,$FD,$D9
	.dc.b	$7F,$EC,$8A,$7A,$7C,$09,$1F,$A1,$8C,$B6,$38,$BF,$82,$8E,$DE,$BF
	.dc.b	$FB,$3A,$1E,$AB,$B7,$FD,$B7,$D3,$FE,$F5,$F5,$7F,$DA,$B5,$FE,$D7
	.dc.b	$F5,$73,$A7,$E1,$B0,$E1,$3F,$ED,$CA,$D8,$AF,$8E,$D8,$FD,$B3,$5F
	.dc.b	$F1,$C3,$5F,$C8,$9D,$CE,$91,$E7,$97,$07,$FB,$62,$FF,$2F,$EA,$7F
	.dc.b	$B7,$9B,$EA,$E7,$0F,$E0,$3C,$AF,$57,$EC,$B6,$3F,$84,$35,$FF,$25
	.dc.b	$35,$F1,$BF,$EE,$8A,$FD,$2B,$63,$AC,$FF,$B9,$E7,$C7,$EA,$D9,$FF
	.dc.b	$31,$9F,$E8,$DF,$2B,$FC,$7D,$E9,$3D,$37,$07,$F8,$B3,$D5,$7A,$BF
	.dc.b	$B2,$F4,$FF,$C8,$91,$8C,$7F,$C6,$38,$71,$BB,$87,$FB,$36,$7B,$D7
	.dc.b	$FA,$8E,$BF,$DE,$70,$78,$3F,$E2,$B8,$E8,$DC,$3C,$1E,$8F,$F9,$D2
	.dc.b	$FC,$B3,$FE,$31,$BE,$67,$2C,$7C,$B7,$C2,$7F,$97,$E1,$D2,$7D,$F7
	.dc.b	$C9,$B8,$FB,$EF,$85,$AF,$86,$3E,$1A,$F6,$9F,$EA,$97,$FD,$89,$97
	.dc.b	$2A,$FF,$6A,$E1,$E2,$7A,$1C,$0F,$46,$3F,$15,$FF,$62,$9F,$EA,$69
	.dc.b	$FF,$5B,$B6,$BF,$D4,$C7,$FA,$F4,$EE,$78,$DE,$EF,$F2,$6C,$3C,$F1
	.dc.b	$F0,$31,$F5,$E5,$F1,$6B,$8C,$FF,$A9,$AB,$FD,$80,$F7,$31,$5F,$EA
	.dc.b	$E7,$FD,$58,$7F,$AC,$EF,$FD,$CA,$FF,$AD,$71,$F7,$3F,$5C,$F7,$3F
	.dc.b	$EB,$77,$FD,$85,$1F,$EB,$E7,$DE,$DC,$3E,$CE,$3F,$0D,$B9,$7C,$8A
	.dc.b	$E8,$7C,$87,$E5,$57,$8A,$7F,$D8,$33,$F9,$BC,$BD,$AF,$9D,$FF,$63
	.dc.b	$5F,$E2,$2B,$C9,$3F,$ED,$3C,$BF,$1F,$1F,$64,$FB,$58,$FB,$09,$7F
	.dc.b	$3D,$E3,$1B,$8C,$1F,$47,$2F,$C4,$D7,$D2,$BF,$CF,$61,$F8,$FA,$DC
	.dc.b	$7D,$53,$DD,$87,$82,$33,$1F,$91,$BE,$67,$DA,$AE,$51,$FE,$0A,$C7
	.dc.b	$EE,$3C,$CF,$8F,$7C,$DD,$4F,$07,$A1,$E0,$BE,$F5,$F4,$23,$FD,$DF
	.dc.b	$1E,$2C,$3D,$47,$AE,$7B,$CF,$DA,$9F,$CA,$3F,$BE,$C3,$FB,$BE,$3F
	.dc.b	$77,$0F,$06,$1C,$A7,$D5,$97,$DA,$C3,$EA,$BF,$E2,$4B,$F9,$B1,$F9
	.dc.b	$D7,$D5,$87,$A0,$F6,$C7,$A3,$0F,$C4,$12,$BD,$93,$F0,$F2,$E3,$97
	.dc.b	$B7,$0F,$C4,$5F,$E0,$F0,$F7,$63,$F3,$B0,$FC,$CC,$7E,$2A,$FF,$1B
	.dc.b	$97,$F1,$9B,$F5,$DE,$EF,$F2,$84,$C7,$F0,$EF,$87,$0F,$36,$5F,$C1
	.dc.b	$E3,$E3,$1E,$F8,$F9,$2F,$CB,$3E,$D6,$5F,$C0,$8F,$95,$3C,$63,$F1
	.dc.b	$6F,$B9,$F8,$95,$EE,$C3,$E1,$47,$DA,$3F,$03,$3F,$E5,$3C,$BF,$97
	.dc.b	$C7,$F5,$C6,$BF,$C2,$CC,$7A,$AF,$57,$83,$63,$FB,$19,$FC,$E5,$C3
	.dc.b	$C2,$3F,$C2,$6F,$FB,$70,$FF,$31,$7A,$98,$23,$FE,$18,$7F,$D0,$04
	.dc.b	$7F,$D2,$8D,$CB,$FD,$C3,$1F,$EA,$43,$FD,$65,$9F,$FA,$08,$8F,$F9
	.dc.b	$F7,$7F,$F4,$AF,$A9,$FE,$84,$9F,$F7,$14,$FF,$A6,$4F,$F6,$0F,$AB
	.dc.b	$FE,$AB,$BE,$B1,$97,$FD,$19,$7C,$89,$1C,$2F,$A3,$FE,$B5,$3C,$95
	.dc.b	$CE,$FB,$C7,$FA,$B0,$E1,$5C,$AF,$AC,$60,$8E,$09,$3E,$C8,$EB,$7E
	.dc.b	$07,$E0,$9E,$F3,$8C,$6D,$F5,$35,$F1,$4F,$47,$83,$5E,$B5,$E5,$9E
	.dc.b	$18,$7D,$17,$D6,$D9,$F8,$07,$63,$9C,$70,$7E,$C9,$2B,$E6,$BE,$28
	.dc.b	$D4,$62,$7A,$4F,$B0,$F1,$9E,$C3,$A9,$0C,$C7,$A5,$F1,$18,$9F,$91
	.dc.b	$F0,$4F,$84,$61,$FE,$B0,$DA,$DB,$E0,$DF,$08,$F8,$5C,$D6,$A9,$8F
	.dc.b	$3C,$39,$1F,$EB,$AC,$38,$DF,$08,$FF,$4B,$CE,$EB,$FC,$57,$1B,$C3
	.dc.b	$B3,$CA,$3F,$D8,$73,$BC,$3C,$B8,$74,$AD,$57,$8B,$1E,$CF,$A7,$0E
	.dc.b	$6F,$FA,$B3,$0E,$55,$DF,$1F,$F6,$5C,$B6,$7A,$BF,$EC,$C8,$FF,$59
	.dc.b	$FB,$1F,$ED,$87,$CE,$6E,$35,$A7,$FB,$87,$D8,$F0,$DF,$2F,$63,$AB
	.dc.b	$72,$F3,$E3,$CE,$FE,$7C,$E7,$1E,$F8,$72,$C7,$AE,$1F,$EE,$AB,$DC
	.dc.b	$62,$BF,$DC,$F8,$73,$AF,$F7,$26,$3F,$EE,$09,$FF,$6E,$7A,$BE,$4F
	.dc.b	$57,$FD,$11,$5F,$EE,$6C,$7F,$DA,$AF,$CF,$7F,$DB,$0F,$FB,$33,$D8
	.dc.b	$FF,$70,$7A,$FF,$ED,$FF,$63,$FD,$C3,$3C,$4F,$C4,$38,$C7,$FD,$B3
	.dc.b	$EA,$5C,$3F,$17,$D2,$A7,$C5,$FD,$6B,$5E,$F0,$FF,$80,$37,$BA,$FF
	.dc.b	$69,$D7,$D0,$8E,$5E,$86,$31,$EB,$8A,$3A,$7A,$5F,$EC,$B2,$78,$BF
	.dc.b	$EC,$B3,$FD,$99,$E9,$FF,$B2,$49,$E4,$FF,$B2,$72,$FF,$4D,$FA,$7C
	.dc.b	$DF,$C9,$67,$FE,$B0,$F4,$FF,$D0,$CD,$8F,$C8,$CE,$DF,$F7,$CA,$FA
	.dc.b	$5D,$F9,$AF,$9B,$A7,$FA,$3E,$BF,$DA,$93,$FE,$A7,$F4,$7F,$D6,$F0
	.dc.b	$3F,$D5,$87,$E4,$0B,$C1,$FB,$1B,$FF,$5D,$C6,$5B,$1F,$EC,$3F,$43
	.dc.b	$FD,$91,$E8,$FF,$AF,$9F,$04,$FF,$AD,$2B,$F5,$3B,$FF,$A9,$57,$A3
	.dc.b	$FE,$B5,$7C,$4D,$51,$B3,$84,$64,$D3,$9F,$43,$DD,$19,$F4,$7F,$D7
	.dc.b	$D5,$CA,$73,$12,$3C,$71,$F1,$7E,$09,$E6,$7C,$7F,$39,$EE,$D3,$51
	.dc.b	$2F,$F8,$1C,$E9,$B3,$97,$E5,$57,$47,$24,$33,$5F,$57,$CF,$FF,$12
	.dc.b	$79,$9F,$E3,$29,$79,$BF,$E3,$27,$FD,$06,$7F,$9E,$A3,$FC,$2F,$3C
	.dc.b	$23,$FD,$0A,$FF,$89,$BD,$1F,$F2,$0C,$EB,$2F,$F0,$9D,$FF,$8C,$38
	.dc.b	$BF,$C4,$18,$7F,$88,$8F,$F4,$81,$FE,$77,$39,$CF,$F8,$D5,$D4,$FF
	.dc.b	$88,$9F,$F0,$E1,$FE,$27,$3F,$C8,$8F,$F8,$9E,$BC,$15,$FE,$27,$C3
	.dc.b	$CD,$5F,$E8,$53,$9C,$E9,$E7,$1A,$F4,$BF,$C7,$D5,$C7,$0F,$4D,$FC
	.dc.b	$5C,$3E,$26,$1E,$73,$A1,$E3,$BF,$F5,$5C,$79,$CF,$81,$1F,$32,$7F
	.dc.b	$C9,$95,$ED,$8E,$18,$FF,$AB,$DE,$78,$F8,$E3,$E5,$C7,$F9,$12,$BF
	.dc.b	$D8,$6D,$3E,$A9,$FE,$9B,$8E,$58,$7A,$A3,$DB,$7E,$A7,$E6,$E3,$F4
	.dc.b	$9F,$09,$F7,$4E,$87,$B6,$79,$63,$EA,$3E,$0E,$3F,$7E,$79,$4F,$C0
	.dc.b	$C7,$E6,$56,$DF,$7D,$F7,$BF,$65,$F0,$C7,$CE,$7B,$9F,$B7,$5E,$E6
	.dc.b	$9E,$EB,$FB,$CF,$9A,$BC,$4F,$E2,$9B,$1F,$46,$7E,$C1,$E2,$9F,$A4
	.dc.b	$7E,$08,$E5,$8F,$D3,$AF,$2C,$73,$8C,$1F,$25,$FA,$AD,$C7,$E2,$D7
	.dc.b	$69,$E8,$7A,$0F,$95,$87,$B6,$7F,$D9,$F3,$FE,$9D,$3F,$CE,$13,$0F
	.dc.b	$F3,$B9,$FE,$6E,$BF,$F3,$AB,$FE,$BF,$CB,$FC,$DE,$FF,$29,$3F,$CE
	.dc.b	$5D,$1F,$E7,$62,$3F,$E6,$D2,$1F,$E7,$22,$9F,$E3,$2A,$FF,$1E,$BF
	.dc.b	$E3,$76,$E5,$FE,$35,$8C,$1F,$E4,$03,$FC,$74,$7F,$8C,$CF,$F1,$89
	.dc.b	$0C,$9F,$E3,$33,$46,$0D,$7A,$1F,$E4,$0F,$2F,$FC,$D1,$1D,$63,$1E
	.dc.b	$8F,$F9,$7E,$73,$3F,$E4,$53,$89,$E4,$30,$75,$24,$F2,$F0,$FF,$CB
	.dc.b	$91,$E2,$8C,$3F,$E4,$D3,$CC,$43,$AD,$FC,$12,$61,$F1,$09,$1E,$73
	.dc.b	$C5,$7E,$62,$61,$E9,$26,$1E,$29,$EC,$79,$8E,$31,$A7,$C8,$78,$AB
	.dc.b	$BB,$EE,$C3,$D3,$87,$F9,$86,$39,$3E,$67,$C8,$D7,$81,$DA,$FC,$4F
	.dc.b	$53,$75,$DB,$0F,$4C,$FF,$05,$9F,$EE,$D7,$FC,$1E,$58,$7F,$9A,$B0
	.dc.b	$61,$FE,$60,$F4,$3F,$C4,$1E,$7F,$F9,$4B,$17,$B8,$FC,$F9,$F5,$9A
	.dc.b	$B9,$AF,$F9,$8F,$D4,$FF,$30,$61,$FE,$5B,$9F,$F2,$FC,$63,$97,$FC
	.dc.b	$E9,$97,$58,$C5,$F8,$4E,$5E,$A7,$F9,$92,$BF,$CB,$11,$FE,$54,$F4
	.dc.b	$FF,$CD,$3F,$00,$F8,$2F,$ED,$F3,$92,$67,$FE,$62,$E2,$FF,$2C,$61
	.dc.b	$CA,$31,$EA,$FF,$95,$57,$AB,$FE,$55,$7F,$CD,$1E,$AF,$F9,$92,$75
	.dc.b	$39,$BE,$78,$7F,$97,$CF,$F2,$D9,$FE,$4F,$BF,$F1,$27,$A6,$F5,$7F
	.dc.b	$CB,$9E,$47,$F9,$7E,$3F,$CA,$F5,$BC,$FF,$CB,$77,$C2,$FF,$D0,$B1
	.dc.b	$88,$FF,$32,$CF,$F9,$87,$D1,$FF,$29,$47,$F8,$C7,$D3,$BD,$9F,$E5
	.dc.b	$3A,$FF,$47,$C6,$2B,$FC,$B7,$18,$F7,$BF,$D2,$35,$FE,$56,$8C,$1F
	.dc.b	$E4,$A9,$FF,$2B,$3F,$E9,$8F,$4B,$FC,$AB,$D5,$FC,$5F,$1F,$E3,$11
	.dc.b	$87,$01,$7A,$5F,$E5,$88,$CC,$63,$D2,$FF,$2D,$47,$F2,$1F,$47,$FC
	.dc.b	$A9,$AF,$F9,$CB,$CD,$7A,$5F,$E5,$43,$FC,$A5,$3A,$CF,$87,$A3,$FE
	.dc.b	$56,$C3,$FC,$C3,$7F,$E5,$97,$FC,$BF,$1F,$E5,$77,$FD,$63,$1F,$E5
	.dc.b	$09,$FE,$61,$E8,$FF,$93,$E7,$3C,$9F,$E5,$C7,$FC,$B3,$3A,$F4,$7F
	.dc.b	$CA,$A7,$F9,$53,$4F,$F2,$2F,$A1,$FE,$47,$AF,$F2,$19,$0F,$F3,$13
	.dc.b	$97,$FC,$C9,$E8,$FF,$93,$DF,$F6,$1C,$67,$6F,$F6,$B4,$E6,$77,$E8
	.dc.b	$76,$F6,$7F,$C8,$95,$FE,$DB,$F3,$FF,$C9,$15,$FE,$6F,$6C,$7F,$B7
	.dc.b	$FD,$0F,$F2,$44,$FF,$91,$8F,$F2,$44,$71,$F4,$79,$EB,$FD,$46,$B7
	.dc.b	$3C,$7C,$EC,$DF,$F9,$F7,$07,$C9,$FF,$93,$DF,$ED,$1E,$97,$F6,$33
	.dc.b	$FC,$03,$3F,$E0,$33,$FC,$24,$7F,$69,$F4,$7F,$BD,$F9,$DF,$E0,$C3
	.dc.b	$FB,$74,$7F,$86,$0F,$F1,$01,$C3,$D2,$FF,$09,$3F,$DF,$32,$FF,$08
	.dc.b	$61,$E0,$8E,$C7,$02,$7A,$1F,$E0,$C7,$C1,$1F,$E0,$09,$FF,$21,$12
	.dc.b	$3F,$C5,$6D,$9F,$F0,$F4,$7F,$80,$2B,$C2,$F2,$9F,$09,$BB,$F4,$E1
	.dc.b	$FE,$0E,$BF,$F0,$B5,$7F,$8D,$0F,$F0,$A9,$ED,$7C,$97,$D0,$98,$FC
	.dc.b	$89,$FF,$0B,$61,$D9,$F2,$92,$BC,$AF,$3C,$3F,$C5,$97,$CA,$3E,$45
	.dc.b	$F3,$AF,$1C,$7F,$87,$AF,$D0,$F9,$63,$FC,$83,$8F,$C5,$8E,$D8,$7B
	.dc.b	$23,$E7,$47,$90,$F5,$E3,$EF,$9F,$B5,$1C,$DF,$F1,$A5,$74,$BF,$53
	.dc.b	$E8,$AE,$15,$F6,$CE,$77,$DA,$7C,$F8,$FF,$8C,$A7,$A1,$F4,$F1,$F6
	.dc.b	$63,$E0,$8F,$61,$E8,$36,$7C,$53,$FC,$9C,$FC,$68,$E3,$1E,$1A,$F3
	.dc.b	$3E,$2C,$7E,$2B,$D0,$FE,$3D,$8F,$F8,$E1,$34,$99,$FF,$8E,$A3,$FC
	.dc.b	$79,$DF,$FE,$38,$78,$C6,$5F,$F1,$F7,$A5,$FC,$99,$A7,$F8,$E8,$FE
	.dc.b	$58,$E5,$FE,$EE,$64,$D9,$FE,$41,$21,$FC,$C1,$E0,$9A,$47,$91,$D0
	.dc.b	$E2,$78,$0C,$9C,$4A,$F1,$76,$78,$8F,$01,$A2,$1A,$3B,$1A,$30,$75
	.dc.b	$2B,$D5,$D1,$0F,$21,$A3,$A9,$93,$C4,$68,$C1,$E0,$30,$43,$64,$34
	.dc.b	$53,$C8,$72,$29,$C4,$D9,$0C,$9C,$5F,$53,$B3,$26,$DC,$BB,$3D,$66
	.dc.b	$DC,$BB,$32,$6C,$F8,$44,$3C,$CF,$07,$B9,$1C,$11,$E6,$75,$39,$1E
	.dc.b	$13,$27,$12,$BC,$5D,$9E,$33,$C2,$68,$86,$8E,$C6,$8C,$1D,$4A,$F5
	.dc.b	$74,$43,$C8,$68,$EA,$64,$F1,$1A,$30,$78,$0C,$10,$D9,$0D,$14,$F2
	.dc.b	$1C,$8A,$71,$36,$43,$27,$17,$D4,$EC,$C9,$B7,$2E,$CF,$59,$B7,$2E
	.dc.b	$CC,$9B,$3E,$11,$0F,$33,$C1,$EE,$47,$04,$79,$9D,$4E,$47,$84,$C9
	.dc.b	$C4,$AF,$17,$67,$8C,$F0,$9A,$21,$A3,$B1,$A3,$07,$52,$BD,$5D,$10
	.dc.b	$F2,$1A,$3A,$99,$3C,$46,$8C,$1F,$72,$92,$A4,$BF,$08,$73,$FF,$BC
	.dc.b	$6F,$75,$A9,$CD,$7F,$3C,$38,$56,$EF,$A4,$60,$D4,$F0,$7F,$DF,$66
	.dc.b	$4A,$E6,$FB,$C6,$5F,$E9,$9B,$32,$FE,$8D,$18,$6C,$66,$7F,$CD,$F3
	.dc.b	$96,$C6,$5F,$EB,$2D,$7D,$CD,$7E,$1C,$71,$77,$3D,$88,$F4,$6B,$A3
	.dc.b	$04,$30,$43,$BB,$C1,$A6,$CA,$6C,$F1,$CF,$95,$FE,$DE,$D9,$F3,$47
	.dc.b	$02,$1C,$27,$B3,$4E,$C5,$3B,$1D,$1D,$B4,$D9,$4D,$C7,$81,$E4,$78
	.dc.b	$0D,$15,$D3,$83,$64,$30,$47,$64,$9F,$71,$27,$CB,$19,$6C,$66,$7C
	.dc.b	$F3,$96,$C6,$5F,$4B,$5F,$33,$5F,$24,$71,$77,$3D,$88,$F4,$6B,$A3
	.dc.b	$04,$30,$43,$BB,$C1,$A6,$CA,$6C,$F1,$CF,$95,$F8,$CD,$9F,$34,$70
	.dc.b	$21,$C2,$7B,$34,$EC,$53,$B1,$D1,$DB,$4D,$94,$DC,$78,$1E,$47,$80
	.dc.b	$D1,$5D,$38,$36,$43,$04,$76,$49,$F7,$12,$7C,$B1,$96,$C6,$67,$CF
	.dc.b	$39,$6C,$65,$F4,$B5,$F3,$35,$F2,$47,$17,$73,$D8,$8F,$46,$BA,$30
	.dc.b	$43,$04,$3B,$BC,$1A,$6C,$A6,$CF,$1C,$F9,$5F,$8C,$D9,$F3,$47,$02
	.dc.b	$1C,$27,$B3,$4E,$C5,$3B,$1D,$1D,$B4,$D9,$4D,$C7,$81,$FE,$84,$7F
	.dc.b	$9F,$31,$FF,$3F,$3F,$D0,$DA,$E8,$C1,$FC,$DB,$0F,$F4,$14,$FF,$40
	.dc.b	$8E,$46,$67,$8B,$FD,$20,$C9,$0E,$65,$3B,$10,$E0,$68,$86,$63,$C0
	.dc.b	$E0,$D9,$99,$F2,$92,$7C,$73,$C2,$75,$18,$9D,$47,$A4,$87,$94,$EA
	.dc.b	$49,$E4,$D8,$E2,$EE,$32,$48,$CC,$79,$E3,$13,$E1,$24,$78,$0D,$4F
	.dc.b	$07,$CE,$64,$A7,$77,$4F,$8E,$39,$B9,$6B,$D1,$AF,$23,$07,$53,$91
	.dc.b	$67,$CE,$D9,$EF,$18,$8E,$C6,$67,$AB,$C8,$C9,$0E,$E5,$3C,$44,$38
	.dc.b	$1A,$21,$98,$F0,$38,$36,$66,$7C,$A4,$9F,$1C,$F0,$9D,$46,$27,$51
	.dc.b	$E9,$21,$E5,$3A,$92,$79,$36,$38,$BB,$8C,$92,$33,$1E,$78,$C4,$F8
	.dc.b	$49,$1E,$03,$53,$C1,$F3,$99,$29,$DD,$D3,$E3,$8E,$6E,$5A,$F4,$6B
	.dc.b	$C8,$C1,$D4,$E4,$59,$F3,$B6,$7B,$C6,$23,$B1,$99,$EA,$F2,$32,$43
	.dc.b	$B9,$4F,$11,$0E,$06,$88,$66,$3C,$0E,$0D,$99,$9F,$29,$27,$C7,$3C
	.dc.b	$27,$51,$89,$D4,$7A,$48,$79,$4E,$A4,$9E,$4D,$8E,$2E,$E3,$24,$8C
	.dc.b	$C7,$9E,$31,$3E,$12,$47,$80,$D4,$F0,$7C,$E6,$4A,$77,$74,$F8,$E3
	.dc.b	$9B,$96,$BD,$1A,$F2,$30,$75,$39,$16,$7C,$ED,$9E,$F1,$88,$EC,$66
	.dc.b	$7A,$BC,$8C,$90,$EE,$53,$C4,$43,$81,$A2,$19,$8F,$03,$83,$66,$67
	.dc.b	$CA,$49,$F1,$CF,$09,$D4,$62,$75,$1E,$92,$1E,$53,$A9,$27,$93,$63
	.dc.b	$8B,$B8,$C9,$23,$31,$E7,$8C,$4F,$84,$91,$E0,$35,$3C,$1F,$39,$92
	.dc.b	$9D,$DD,$3E,$38,$E6,$E5,$AF,$EA,$FF,$3F,$10

;--------------------------------------------------------------------------------
;	FSNEG.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsneg_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FSNEG.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSNEG.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsneg_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsneg.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSNEG.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1589+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsneg.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSNEG.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsneg_data_compressed::
	.dc.b	$00,$02,$6C,$BC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$4F,$E6,$3A,$29,$F7,$3F,$80,$9C,$C6,$1C,$4E,$E7,$51
	.dc.b	$E6,$9D,$3F,$88,$9D,$3C,$23,$2F,$96,$78,$CE,$A3,$84,$E9,$F0,$CE
	.dc.b	$A3,$11,$DE,$30,$7F,$4A,$FA,$EF,$F3,$4F,$C1,$7E,$A1,$B9,$FF,$3A
	.dc.b	$7F,$CD,$B8,$79,$1B,$87,$B5,$DE,$1C,$FD,$2F,$DD,$5F,$17,$D1,$09
	.dc.b	$E3,$8E,$E3,$F7,$95,$B6,$99,$3D,$07,$1A,$E7,$5E,$08,$C1,$C0,$A7
	.dc.b	$07,$BB,$8C,$7D,$D3,$E6,$8E,$D8,$FB,$63,$A9,$98,$E1,$7E,$57,$A4
	.dc.b	$7B,$3E,$1B,$C4,$47,$04,$8F,$24,$65,$FA,$A7,$F6,$38,$D3,$FC,$7E
	.dc.b	$7F,$B4,$CE,$E3,$E0,$BE,$AC,$3E,$01,$2B,$CF,$3D,$23,$0F,$C3,$78
	.dc.b	$46,$A3,$B1,$F3,$70,$F8,$6F,$18,$D4,$62,$36,$7C,$CC,$3E,$33,$B3
	.dc.b	$B4,$FC,$7A,$EF,$19,$38,$BE,$47,$E4,$99,$AE,$C7,$7A,$EB,$1A,$38
	.dc.b	$D7,$53,$D5,$1A,$9E,$38,$7D,$58,$E2,$65,$FB,$11,$D7,$2F,$CC,$57
	.dc.b	$BA,$72,$71,$8C,$DF,$96,$32,$FE,$1A,$32,$F9,$6B,$C2,$E7,$28,$FE
	.dc.b	$20,$F9,$D5,$D6,$33,$8F,$C3,$AF,$24,$63,$0F,$75,$FC,$C3,$E5,$63
	.dc.b	$EB,$79,$9B,$3C,$C6,$0F,$AD,$8F,$9A,$38,$47,$8C,$CD,$71,$7E,$69
	.dc.b	$DC,$FA,$11,$C3,$0F,$B0,$FC,$CA,$F8,$38,$7C,$0A,$E6,$4A,$F6,$63
	.dc.b	$EF,$8C,$63,$EC,$7B,$9D,$AF,$9C,$FC,$E8,$D9,$EC,$3D,$15,$F8,$C7
	.dc.b	$E8,$46,$E3,$18,$79,$EF,$B3,$E7,$C7,$E2,$4F,$AB,$0F,$88,$78,$30
	.dc.b	$CD,$FE,$32,$31,$1E,$08,$C5,$6A,$3E,$35,$F5,$74,$62,$BA,$CE,$E3
	.dc.b	$DF,$3B,$BF,$8B,$1F,$46,$3D,$0F,$19,$FA,$98,$FC,$DA,$F6,$DF,$91
	.dc.b	$F5,$4E,$B1,$F8,$58,$F9,$B0,$F1,$9E,$0B,$F1,$B8,$C3,$F1,$EF,$CE
	.dc.b	$3E,$9B,$8C,$3E,$73,$E1,$AE,$91,$E8,$9E,$18,$FA,$1E,$D8,$F8,$DF
	.dc.b	$81,$1E,$FA,$FC,$05,$F5,$7F,$61,$3A,$C3,$CB,$87,$53,$DF,$87,$C9
	.dc.b	$BE,$15,$F8,$39,$F8,$2F,$A6,$3C,$CF,$E9,$AB,$E3,$5F,$E0,$9E,$AF
	.dc.b	$8B,$1F,$CA,$5F,$B9,$F9,$35,$C7,$0F,$04,$6A,$BC,$F3,$F3,$31,$F6
	.dc.b	$3F,$A7,$C7,$CA,$FE,$50,$F6,$5F,$C3,$78,$63,$E2,$78,$61,$D5,$F6
	.dc.b	$61,$E7,$AE,$18,$FB,$8F,$34,$F4,$3E,$05,$7A,$1F,$D5,$4F,$A1,$FC
	.dc.b	$65,$FC,$27,$F1,$87,$03,$95,$74,$7C,$B3,$E0,$BF,$14,$6C,$C9,$67
	.dc.b	$C9,$1F,$11,$B1,$EA,$3B,$BD,$4C,$BF,$14,$E6,$E6,$3B,$B6,$38,$C7
	.dc.b	$53,$25,$72,$77,$3F,$22,$79,$5D,$B9,$C3,$EA,$E3,$F9,$76,$E1,$F1
	.dc.b	$67,$8D,$F8,$5B,$7D,$DA,$73,$C7,$EC,$3E,$B9,$F6,$E1,$D6,$72,$D9
	.dc.b	$FA,$78,$FD,$57,$F5,$38,$78,$71,$F1,$CE,$70,$FD,$0D,$7E,$DF,$0E
	.dc.b	$77,$B6,$D6,$F1,$FD,$A9,$F9,$DA,$EE,$71,$2D,$FE,$B6,$37,$1F,$92
	.dc.b	$32,$F1,$7C,$B3,$D5,$FD,$21,$CC,$C9,$67,$CE,$75,$76,$D7,$F2,$A6
	.dc.b	$E7,$C2,$FF,$10,$AF,$5B,$93,$8C,$78,$E7,$AB,$4E,$38,$7F,$00,$9E
	.dc.b	$B3,$DF,$1F,$DA,$C7,$53,$DA,$7B,$CF,$4C,$6F,$1F,$DA,$46,$CC,$DF
	.dc.b	$77,$C6,$71,$9F,$1B,$63,$C7,$87,$EC,$EF,$C3,$87,$95,$B1,$9C,$3B
	.dc.b	$CE,$5B,$19,$C7,$F6,$58,$7E,$78,$FC,$76,$1F,$9C,$69,$9C,$3F,$1C
	.dc.b	$FA,$0F,$21,$E8,$3D,$98,$FA,$1B,$3E,$87,$F8,$1B,$F9,$FC,$7F,$27
	.dc.b	$1F,$4C,$F9,$B1,$F0,$DA,$76,$29,$D0,$F0,$47,$B6,$FD,$71,$FA,$A3
	.dc.b	$FC,$B7,$3F,$E6,$DC,$BF,$CB,$D8,$7F,$A8,$1F,$F3,$B9,$FE,$61,$BF
	.dc.b	$F7,$DB,$89,$FF,$33,$C6,$1E,$47,$F9,$AA,$7E,$1E,$1E,$F9,$F1,$5F
	.dc.b	$F0,$53,$E4,$65,$FA,$B7,$F0,$A7,$A4,$F5,$C7,$09,$FE,$08,$FC,$47
	.dc.b	$F1,$18,$FB,$23,$18,$7C,$5C,$3E,$FD,$FC,$C3,$F8,$41,$DA,$7A,$4F
	.dc.b	$CF,$CB,$F4,$38,$74,$AF,$31,$F1,$EF,$D7,$5F,$5E,$3F,$94,$BC,$5E
	.dc.b	$95,$F8,$97,$8D,$F9,$1F,$AD,$5E,$47,$89,$E4,$3F,$10,$7C,$17,$57
	.dc.b	$FC,$73,$1F,$8D,$3A,$AF,$4B,$C6,$7E,$75,$F8,$5F,$8E,$FB,$30,$F8
	.dc.b	$97,$F9,$EA,$F8,$87,$E6,$8F,$73,$DA,$B9,$C6,$AD,$68,$98,$BE,$3B
	.dc.b	$FD,$6C,$D8,$FF,$61,$35,$FE,$A0,$DA,$E7,$1C,$A7,$89,$FD,$FA,$75
	.dc.b	$5D,$E7,$FD,$FB,$3A,$7F,$CA,$93,$A7,$91,$E0,$8E,$46,$DF,$F5,$4C
	.dc.b	$78,$F2,$FF,$39,$BC,$63,$A9,$C4,$FF,$30,$1E,$13,$FD,$62,$5B,$F4
	.dc.b	$B9,$32,$7F,$B5,$CD,$5E,$5F,$F7,$21,$83,$11,$C9,$C1,$CD,$C3,$CC
	.dc.b	$F8,$4B,$7E,$90,$8C,$3D,$23,$31,$88,$E9,$18,$23,$B2,$1A,$29,$EF
	.dc.b	$79,$F0,$48,$F4,$B5,$F4,$35,$F8,$CD,$7D,$2D,$7F,$15,$28,$C4,$7A
	.dc.b	$F8,$7D,$24,$7E,$3B,$5F,$8C,$D7,$EA,$35,$F8,$ED,$7B,$BF,$13,$8A
	.dc.b	$47,$D2,$6B,$F4,$5A,$FD,$F6,$BF,$49,$AF,$A9,$AB,$27,$B5,$F9,$9C
	.dc.b	$FF,$45,$F7,$7F,$8B,$6D,$67,$5F,$CF,$5F,$08,$FC,$D9,$CE,$F4,$FE
	.dc.b	$C6,$F9,$57,$E9,$5B,$1F,$B5,$AE,$4F,$E7,$0E,$47,$48,$DA,$6C,$70
	.dc.b	$BF,$D3,$5F,$82,$B4,$5A,$DC,$66,$B8,$5F,$8E,$B4,$66,$63,$E1,$32
	.dc.b	$7B,$A7,$8B,$FC,$37,$E1,$DC,$FE,$16,$C7,$C4,$6B,$ED,$C1,$ED,$6E
	.dc.b	$A4,$ED,$F4,$46,$23,$E2,$57,$2C,$BF,$8B,$37,$2F,$99,$5B,$AF,$E0
	.dc.b	$53,$E0,$6D,$FE,$CA,$31,$1F,$C5,$DA,$FA,$E3,$44,$7F,$82,$46,$93
	.dc.b	$49,$1B,$8F,$77,$B3,$F4,$4F,$56,$1E,$89,$C4,$FD,$D9,$CE,$1E,$39
	.dc.b	$CC,$F2,$3E,$21,$0C,$12,$BA,$39,$7F,$33,$5C,$9C,$C7,$CA,$9C,$57
	.dc.b	$E4,$EB,$C3,$3F,$6F,$BB,$ED,$11,$FC,$97,$B3,$5F,$A7,$17,$0F,$D3
	.dc.b	$CD,$BF,$99,$36,$BF,$45,$35,$F9,$31,$73,$4E,$72,$4E,$63,$0D,$AF
	.dc.b	$C6,$3F,$6C,$8F,$D6,$6B,$FC,$FE,$79,$B9,$21,$98,$DB,$F8,$86,$BE
	.dc.b	$A3,$F6,$5E,$C7,$EA,$CF,$1B,$82,$1F,$10,$F2,$98,$C7,$F8,$5C,$70
	.dc.b	$8F,$D1,$3C,$DD,$9C,$4C,$9C,$71,$F1,$39,$9E,$6E,$4F,$B0,$72,$72
	.dc.b	$F2,$38,$47,$28,$C9,$1C,$12,$3F,$8C,$4F,$95,$E2,$7E,$FF,$C4,$FD
	.dc.b	$51,$1F,$D2,$B5,$FA,$0D,$7F,$66,$D7,$F6,$0D,$7A,$1F,$AE,$F1,$7F
	.dc.b	$6C,$47,$04,$3E,$E3,$96,$99,$3F,$80,$B9,$69,$FB,$AF,$1B,$F8,$11
	.dc.b	$1C,$10,$FC,$43,$96,$99,$3F,$86,$B9,$69,$FC,$1B,$C7,$FE,$1C,$47
	.dc.b	$04,$3F,$28,$E5,$A6,$4F,$E3,$2E,$5A,$7F,$12,$F2,$3F,$8D,$11,$C1
	.dc.b	$0F,$D0,$39,$69,$93,$F9,$2B,$96,$9F,$C7,$BC,$9F,$E4,$C4,$70,$43
	.dc.b	$F5,$8E,$5A,$64,$FE,$62,$E5,$A7,$F2,$AF,$2B,$F9,$91,$1C,$10,$FD
	.dc.b	$C3,$96,$99,$3F,$9E,$B9,$69,$FC,$DB,$CB,$FE,$7C,$47,$04,$3F,$82
	.dc.b	$39,$69,$93,$FA,$4B,$96,$9F,$D0,$BE,$1F,$FA,$51,$1C,$10,$FE,$20
	.dc.b	$E5,$A6,$4F,$EA,$AE,$5A,$7F,$94,$7E,$21,$ED,$7F,$7B,$6C,$7F,$83
	.dc.b	$9A,$FF,$6B,$F6,$B5,$32,$34,$FF,$80,$63,$11,$FE,$0D,$9E,$3C,$3F
	.dc.b	$E7,$F6,$F0,$FF,$85,$6B,$73,$9B,$B3,$9B,$EA,$43,$A4,$77,$71,$3C
	.dc.b	$5B,$3C,$5D,$98,$3F,$D2,$D7,$E2,$7F,$C5,$17,$E0,$73,$75,$FF,$11
	.dc.b	$45,$9E,$10,$6F,$2A,$75,$5C,$63,$15,$E1,$38,$46,$67,$71,$99,$DB
	.dc.b	$E8,$26,$1E,$D7,$E9,$1C,$EF,$C6,$78,$AB,$9C,$70,$C3,$E1,$57,$12
	.dc.b	$3F,$EA,$9A,$E1,$3C,$F0,$F2,$34,$F2,$5F,$4B,$E3,$1F,$20,$D1,$DC
	.dc.b	$C1,$4F,$79,$C8,$DB,$C9,$F9,$A6,$2F,$C9,$5F,$76,$39,$5F,$BA,$31
	.dc.b	$5C,$63,$13,$FE,$59,$BF,$35,$F2,$8C,$1D,$4F,$BA,$60,$EB,$7F,$01
	.dc.b	$A7,$9C,$A7,$AF,$0F,$AB,$87,$86,$78,$36,$78,$35,$E1,$7E,$47,$C2
	.dc.b	$7D,$A3,$AC,$7C,$3A,$EB,$7E,$99,$CB,$F6,$8C,$4F,$53,$FD,$13,$5D
	.dc.b	$67,$E7,$9D,$63,$53,$C0,$E5,$1E,$23,$FD,$B5,$87,$EF,$9B,$87,$FB
	.dc.b	$2A,$31,$1E,$33,$27,$F0,$98,$D1,$1C,$10,$FF,$71,$47,$27,$E0,$D7
	.dc.b	$FA,$83,$D9,$FF,$44,$1E,$9C,$4E,$27,$EC,$1F,$E9,$37,$0F,$C4,$3F
	.dc.b	$D4,$0E,$27,$DF,$18,$3F,$D3,$91,$D4,$D6,$1F,$0E,$3C,$6F,$FA,$9E
	.dc.b	$33,$18,$21,$E1,$39,$9E,$4A,$F3,$B8,$3C,$86,$CF,$F4,$94,$78,$5D
	.dc.b	$57,$99,$D3,$F0,$DE,$91,$E2,$7E,$DC,$7E,$16,$31,$7F,$A0,$9E,$A7
	.dc.b	$4B,$F6,$3C,$1F,$63,$93,$D7,$1F,$C6,$27,$9E,$1F,$C5,$DB,$3F,$C4
	.dc.b	$DA,$FE,$85,$B7,$E3,$8E,$87,$73,$E3,$E1,$F1,$1F,$29,$F1,$1C,$47
	.dc.b	$BE,$31,$3D,$30,$F4,$D6,$9F,$4C,$E8,$EF,$87,$C1,$BE,$44,$9F,$83
	.dc.b	$1E,$D9,$DE,$1E,$D6,$9E,$D2,$9E,$D2,$D7,$B5,$F4,$BF,$5E,$FC,$91
	.dc.b	$F8,$D3,$C8,$6A,$FC,$35,$F6,$67,$73,$F1,$48,$E0,$CD,$FA,$EB,$DF
	.dc.b	$7E,$BA,$FA,$F7,$CF,$0F,$C5,$47,$73,$F2,$BA,$BB,$BF,$DE,$8D,$81
	.dc.b	$FE,$FF,$70,$FF,$B9,$7D,$8D,$CC,$8D,$E0,$C7,$8C,$2F,$8C,$E6,$3A
	.dc.b	$E6,$C3,$94,$77,$C2,$D6,$E1,$5B,$7B,$E1,$DB,$0E,$AD,$BE,$2F,$8A
	.dc.b	$B8,$57,$53,$89,$EE,$AE,$2D,$AF,$4D,$F8,$6B,$8B,$E3,$AD,$D7,$B6
	.dc.b	$BB,$BE,$CC,$2B,$93,$B9,$4F,$94,$62,$79,$92,$7C,$38,$7A,$6B,$9C
	.dc.b	$E6,$B9,$E1,$E4,$AF,$81,$1C,$EF,$C3,$39,$8F,$61,$92,$1E,$23,$E2
	.dc.b	$13,$0F,$74,$F8,$A3,$D8,$70,$8C,$B6,$BC,$CD,$3B,$9C,$DC,$B4,$CC
	.dc.b	$72,$7C,$4D,$24,$7B,$23,$2D,$8C,$C7,$28,$EB,$18,$8F,$9C,$DA,$F8
	.dc.b	$51,$89,$E4,$47,$DA,$79,$4D,$1D,$4C,$CF,$5B,$FA,$51,$E4,$34,$70
	.dc.b	$25,$78,$23,$73,$E3,$C3,$ED,$E1,$F5,$9B,$7F,$24,$FC,$FB,$87,$C9
	.dc.b	$7C,$E7,$15,$F9,$5A,$B8,$7A,$EF,$89,$CF,$0F,$0C,$7C,$63,$4E,$23
	.dc.b	$DB,$18,$BF,$24,$71,$6B,$F1,$67,$83,$D8,$C4,$6A,$FE,$33,$6B,$A4
	.dc.b	$FB,$70,$F5,$3D,$4F,$CD,$E1,$EA,$7E,$5C,$69,$EA,$68,$FC,$C9,$FC
	.dc.b	$38,$FD,$61,$70,$FE,$2E,$DB,$D3,$F8,$B3,$06,$0E,$26,$48,$7E,$68
	.dc.b	$D9,$CA,$3A,$BF,$3E,$5E,$BF,$E9,$E3,$F8,$BB,$EB,$69,$FB,$87,$11
	.dc.b	$C2,$30,$7C,$31,$8F,$F1,$19,$E9,$3B,$9F,$86,$6B,$1F,$E3,$D3,$A3
	.dc.b	$A9,$63,$A5,$73,$7B,$9C,$9C,$3C,$CF,$E6,$B1,$C9,$DC,$7D,$C6,$CF
	.dc.b	$81,$EA,$D7,$91,$F0,$0E,$CE,$63,$B9,$CE,$37,$5E,$D8,$CC,$FE,$22
	.dc.b	$73,$5E,$B7,$99,$E6,$3C,$67,$EC,$AF,$F9,$AD,$F3,$AF,$01,$CE,$FA
	.dc.b	$D7,$13,$99,$E7,$72,$FE,$EE,$B9,$DF,$86,$7D,$06,$AF,$B5,$78,$70
	.dc.b	$D6,$1E,$08,$C1,$E6,$3B,$3E,$33,$4E,$0D,$9F,$00,$E1,$87,$EC,$2B
	.dc.b	$BC,$6C,$F3,$57,$81,$F6,$5F,$9A,$FA,$D7,$0B,$E5,$86,$EF,$89,$E5
	.dc.b	$9D,$61,$FD,$2A,$75,$5C,$2B,$C7,$39,$69,$FB,$D9,$DC,$66,$3D,$47
	.dc.b	$CE,$AF,$49,$D9,$E0,$72,$3F,$8F,$CF,$27,$27,$2C,$3D,$8F,$BB,$0F
	.dc.b	$93,$87,$E0,$3D,$6F,$D3,$92,$7F,$B3,$FA,$D5,$FC,$7C,$5B,$FE,$FD
	.dc.b	$36,$FF,$C1,$13,$6B,$FC,$2F,$35,$FF,$09,$C5,$C5,$36,$30,$D9,$FE
	.dc.b	$EF,$1F,$BA,$8F,$E0,$DF,$14,$E7,$FE,$25,$48,$C3,$FC,$55,$AF,$F1
	.dc.b	$A6,$BF,$C8,$5A,$FF,$8C,$1A,$FF,$8E,$9A,$F5,$8E,$F1,$82,$38,$33
	.dc.b	$CF,$27,$C4,$D7,$C4,$D7,$C4,$D7,$C4,$D7,$C4,$D7,$DD,$CF,$E2,$23
	.dc.b	$EF,$6B,$EF,$6B,$EF,$6B,$EF,$6B,$EF,$7D,$BC,$F2,$3E,$53,$5F,$94
	.dc.b	$D7,$E5,$35,$F9,$4D,$7E,$53,$5F,$AD,$AF,$CA,$23,$F5,$DA,$FD,$76
	.dc.b	$BF,$5D,$AF,$D7,$6B,$F5,$DF,$AB,$CF,$23,$F0,$4D,$7F,$04,$D7,$F0
	.dc.b	$4D,$7F,$04,$D7,$F0,$4D,$7F,$1D,$CF,$F8,$22,$3F,$8F,$6B,$F8,$F6
	.dc.b	$BF,$8F,$6B,$F8,$F6,$BF,$8F,$7F,$1B,$CF,$23,$F3,$4D,$7F,$34,$D7
	.dc.b	$F3,$4D,$7F,$34,$D7,$F3,$4D,$7F,$4D,$F1,$BF,$9A,$23,$FA,$76,$BF
	.dc.b	$A7,$6B,$FA,$76,$BF,$A7,$6B,$FA,$77,$F4,$BC,$12,$3F,$64,$D7,$F6
	.dc.b	$4D,$7F,$64,$D7,$F6,$4D,$7F,$64,$D5,$F1,$8F,$67,$F8,$0B,$63,$FC
	.dc.b	$28,$D7,$FC,$32,$D7,$FC,$40,$D7,$FD,$5E,$D7,$FD,$72,$D7,$BC,$76
	.dc.b	$23,$D9,$AF,$6E,$0F,$19,$1F,$1B,$5D,$18,$21,$82,$1E,$67,$83,$E8
	.dc.b	$E1,$F7,$91,$F7,$B5,$F7,$B5,$E0,$68,$86,$8A,$7C,$4F,$6B,$E5,$91
	.dc.b	$F9,$6D,$7E,$5B,$5E,$06,$88,$68,$A7,$D0,$E5,$FA,$E4,$7E,$BB,$5F
	.dc.b	$AE,$D7,$81,$A2,$1A,$29,$F7,$3D,$BF,$C1,$91,$FC,$1B,$5F,$C1,$B5
	.dc.b	$E0,$68,$86,$8A,$7E,$23,$A3,$F1,$E4,$7F,$1E,$D7,$F1,$ED,$78,$1A
	.dc.b	$21,$A2,$9F,$94,$F7,$3F,$36,$47,$F3,$6D,$7F,$36,$D7,$81,$A2,$1A
	.dc.b	$29,$FA,$0E,$AF,$D3,$91,$FD,$3B,$5F,$D3,$B5,$E0,$68,$86,$8A,$7F
	.dc.b	$A2,$FE,$35,$E4,$7F,$9E,$DB,$1F,$EF,$A6,$BF,$E5,$EF,$35,$C3,$BA
	.dc.b	$93,$B7,$FD,$09,$3A,$8C,$10,$FF,$33,$D7,$4E,$C7,$12,$65,$71,$77
	.dc.b	$7D,$9A,$6E,$B8,$BC,$30,$D4,$62,$18,$77,$C3,$B4,$62,$FC,$F7,$C2
	.dc.b	$73,$5B,$AF,$23,$56,$1E,$5C,$3B,$4E,$49,$87,$4A,$F8,$13,$A8,$ED
	.dc.b	$1E,$73,$37,$E0,$8C,$C7,$6C,$2D,$F4,$8F,$81,$5E,$AC,$3C,$4F,$18
	.dc.b	$E6,$7A,$2F,$C4,$FC,$82,$61,$E3,$79,$9E,$33,$33,$CD,$A6,$63,$83
	.dc.b	$83,$E5,$1C,$8F,$41,$E2,$32,$6E,$BC,$EF,$94,$D5,$FC,$D8,$CC,$70
	.dc.b	$9E,$E4,$AF,$81,$87,$C4,$C3,$E3,$DF,$0C,$3E,$36,$1D,$EB,$4F,$8E
	.dc.b	$75,$58,$8F,$CE,$4D,$C3,$B4,$7B,$F0,$F8,$51,$88,$F2,$47,$89,$F4
	.dc.b	$D7,$8A,$B9,$4F,$03,$53,$F1,$F0,$F2,$DF,$E5,$70,$F1,$CF,$23,$E5
	.dc.b	$38,$78,$61,$F6,$A7,$C5,$5C,$AF,$AC,$62,$7C,$55,$CC,$F8,$45,$8F
	.dc.b	$55,$F5,$78,$47,$83,$0F,$4B,$F5,$4F,$8A,$71,$BF,$63,$F1,$4D,$4F
	.dc.b	$27,$67,$A0,$FA,$13,$D1,$F9,$27,$12,$93,$0F,$C1,$47,$EF,$E3,$F8
	.dc.b	$B6,$1F,$E8,$89,$FE,$27,$18,$3F,$76,$7E,$5C,$C1,$83,$F2,$26,$08
	.dc.b	$6C,$87,$BC,$E3,$5F,$57,$DC,$FE,$36,$7A,$71,$38,$9F,$88,$7F,$18
	.dc.b	$70,$FC,$13,$F9,$43,$89,$F8,$C7,$58,$FB,$F1,$D4,$D6,$3F,$3A,$3E
	.dc.b	$2D,$78,$1D,$9E,$03,$F8,$9C,$F0,$7E,$9C,$F1,$9E,$11,$E2,$38,$38
	.dc.b	$AF,$14,$79,$DA,$49,$F1,$5F,$ED,$6F,$B5,$7D,$7C,$2E,$1C,$EF,$71
	.dc.b	$8A,$F6,$57,$A0,$ED,$3F,$C6,$63,$9D,$FE,$6E,$7E,$51,$BC,$3E,$3D
	.dc.b	$6E,$30,$F6,$69,$C0,$EA,$58,$F5,$3F,$10,$87,$C8,$38,$9E,$B2,$E1
	.dc.b	$EB,$77,$5F,$05,$DB,$E7,$79,$1F,$34,$F0,$98,$9F,$0B,$D4,$D3,$88
	.dc.b	$E5,$3D,$E7,$85,$7B,$5D,$BF,$05,$EB,$1B,$38,$4F,$2B,$F9,$78,$7F
	.dc.b	$24,$6D,$7F,$21,$6B,$FC,$CD,$B8,$7F,$04,$6E,$1F,$89,$6C,$FC,$8C
	.dc.b	$3C,$11,$F7,$23,$CA,$7C,$6B,$FC,$03,$F8,$E8,$E2,$BC,$E7,$8B,$FD
	.dc.b	$CB,$1F,$EE,$D1,$89,$FE,$94,$D5,$1F,$DE,$FE,$F1,$F1,$DC,$73,$FE
	.dc.b	$F7,$7B,$AE,$78,$67,$0F,$F0,$64,$E6,$BF,$C0,$38,$72,$9E,$F5,$B7
	.dc.b	$CC,$DA,$E1,$1F,$E1,$59,$E0,$F9,$2B,$C3,$5B,$24,$7F,$63,$9D,$DF
	.dc.b	$92,$3B,$90,$E4,$75,$C3,$8B,$83,$C6,$6E,$3E,$14,$69,$E9,$1B,$C3
	.dc.b	$CE,$AB,$C3,$87,$9A,$39,$35,$ED,$1A,$7E,$3E,$1D,$67,$57,$E0,$9E
	.dc.b	$26,$8A,$77,$C3,$83,$93,$C9,$5F,$36,$FB,$47,$B0,$C9,$C9,$CE,$1E
	.dc.b	$A2,$4F,$1A,$EB,$5C,$6F,$E0,$9E,$57,$99,$1C,$E1,$F8,$09,$E3,$7D
	.dc.b	$2B,$EB,$46,$2F,$AE,$1E,$27,$13,$A3,$DF,$5C,$5F,$93,$5A,$7C,$17
	.dc.b	$DD,$DD,$74,$9D,$61,$E4,$9F,$C0,$1A,$3E,$14,$6A,$BE,$89,$CE,$3A
	.dc.b	$9F,$36,$79,$3F,$44,$E4,$68,$C1,$6F,$D1,$1F,$1C,$E2,$E2,$7B,$46
	.dc.b	$2F,$D1,$5D,$23,$BE,$1F,$16,$31,$5C,$0F,$3B,$E5,$6B,$E2,$9E,$8F
	.dc.b	$B8,$E2,$6C,$F8,$67,$9E,$BC,$2D,$38,$14,$F7,$57,$8D,$C4,$FB,$6B
	.dc.b	$C3,$7E,$D9,$EB,$3F,$A8,$7F,$C1,$31,$FC,$3D,$B7,$FE,$09,$7F,$34
	.dc.b	$60,$C4,$6D,$FD,$51,$93,$FC,$20,$E5,$E2,$68,$87,$BC,$FD,$B7,$C7
	.dc.b	$7E,$94,$FC,$03,$FC,$3D,$C3,$F1,$8D,$9F,$C3,$5D,$B8,$3F,$76,$E3
	.dc.b	$03,$CA,$3E,$2E,$1F,$C8,$A3,$18,$7D,$49,$E2,$75,$3C,$15,$C9,$C1
	.dc.b	$C8,$F3,$18,$8E,$47,$F3,$79,$EB,$5F,$4B,$0B,$5C,$0E,$47,$A6,$B8
	.dc.b	$3E,$0A,$DB,$FB,$FA,$FC,$5D,$F8,$E3,$CE,$72,$71,$5C,$A3,$15,$CB
	.dc.b	$0F,$C1,$47,$7B,$F8,$4F,$73,$D4,$F9,$DA,$79,$EB,$DB,$1C,$23,$C3
	.dc.b	$18,$75,$19,$8C,$47,$73,$27,$E0,$67,$F9,$BD,$FA,$2B,$81,$E1,$9E
	.dc.b	$AE,$1E,$18,$7C,$49,$E1,$18,$3C,$E7,$73,$E4,$1F,$8F,$76,$E6,$BE
	.dc.b	$1C,$60,$CB,$8C,$3E,$45,$7F,$24,$6D,$7F,$37,$6B,$FB,$26,$E1,$FC
	.dc.b	$19,$B8,$7D,$86,$CF,$8D,$AF,$98,$F4,$61,$E2,$BF,$04,$7B,$30,$F2
	.dc.b	$36,$32,$FC,$63,$E6,$5F,$EE,$5F,$F0,$83,$64,$D2,$47,$EF,$1A,$FE
	.dc.b	$E9,$AF,$F0,$66,$BF,$E1,$07,$B7,$C7,$BA,$29,$27,$F8,$54,$65,$B1
	.dc.b	$97,$FD,$2D,$19,$6C,$65,$FF,$4E,$C6,$5B,$19,$44,$87,$47,$90,$91
	.dc.b	$E2,$6B,$E0,$6B,$D9,$AF,$41,$1D,$1E,$44,$3D,$3D,$1F,$08,$8F,$C0
	.dc.b	$6B,$A3,$04,$30,$43,$DE,$F0,$7E,$0F,$47,$CE,$23,$F3,$1A,$FC,$A6
	.dc.b	$BC,$0D,$10,$D1,$4F,$9B,$D1,$F6,$88,$FD,$86,$BF,$59,$AF,$03,$44
	.dc.b	$34,$53,$FD,$AB,$A3,$87,$FC,$9D,$83,$CE,$A7,$F8,$96,$25,$7D,$CE
	.dc.b	$0F,$CB,$BF,$76,$76,$FF,$93,$67,$6F,$F9,$1E,$BA,$3C,$75,$FB,$51
	.dc.b	$88,$FF,$1F,$CF,$27,$85,$5B,$FF,$37,$4D,$BE,$33,$6B,$FC,$C5,$35
	.dc.b	$FC,$2C,$5C,$3D,$D1,$E1,$3D,$05,$36,$5C,$3C,$98,$78,$23,$15,$D6
	.dc.b	$73,$87,$F9,$E3,$0C,$4F,$76,$D7,$03,$E2,$14,$F6,$96,$FB,$38,$23
	.dc.b	$F3,$9F,$29,$F1,$8E,$15,$F0,$9A,$7C,$22,$9C,$23,$95,$FA,$23,$47
	.dc.b	$69,$E4,$D3,$33,$C2,$FD,$71,$E1,$34,$75,$9E,$33,$EA,$38,$47,$8A
	.dc.b	$BC,$0F,$FA,$06,$F9,$CE,$A3,$A3,$49,$5D,$0F,$9D,$1C,$EF,$E7,$34
	.dc.b	$F7,$E1,$F7,$9F,$2E,$1E,$03,$C3,$19,$BE,$B3,$F8,$23,$D7,$19,$BF
	.dc.b	$04,$66,$76,$FA,$4F,$01,$C3,$0F,$5C,$F0,$8D,$1E,$2C,$3C,$ED,$3C
	.dc.b	$F3,$C1,$ED,$3E,$E7,$47,$53,$BD,$78,$AF,$CD,$1C,$08,$7D,$33,$EC
	.dc.b	$CF,$8F,$0F,$85,$87,$ED,$2F,$D7,$18,$AF,$CA,$9E,$08,$ED,$87,$96
	.dc.b	$3A,$1E,$C8,$E3,$5E,$C6,$C7,$B0,$FC,$F1,$FD,$1A,$3F,$A9,$91,$FE
	.dc.b	$A2,$D7,$FA,$7B,$5E,$06,$88,$68,$B1,$FA,$5F,$90,$7C,$7F,$FB,$9D
	.dc.b	$B3,$F2,$31,$FE,$91,$18,$8F,$F6,$D3,$6B,$D3,$18,$AE,$47,$C1,$9D
	.dc.b	$57,$28,$C1,$A9,$EC,$7F,$60,$9D,$B6,$36,$FF,$0F,$8C,$BF,$33,$1F
	.dc.b	$D2,$C6,$0C,$BF,$56,$3C,$87,$19,$E6,$E0,$FB,$15,$E5,$74,$79,$05
	.dc.b	$75,$9E,$37,$F2,$5B,$5C,$6B,$99,$E9,$39,$98,$39,$C7,$C1,$71,$5D
	.dc.b	$5F,$03,$5F,$E0,$ED,$AF,$04,$7F,$16,$8C,$57,$86,$7F,$5F,$18,$9F
	.dc.b	$11,$CA,$BC,$13,$CA,$30,$79,$9C,$1C,$A3,$75,$D4,$FD,$B5,$75,$AF
	.dc.b	$59,$E3,$9D,$BE,$0A,$E4,$F5,$AE,$4F,$EC,$2B,$91,$F6,$30,$E3,$1D
	.dc.b	$0F,$AD,$5C,$9B,$5C,$9F,$D6,$D7,$2B,$FE,$B4,$D8,$F2,$1E,$13,$C9
	.dc.b	$3E,$A9,$FF,$26,$4F,$E7,$67,$FC,$4C,$78,$4A,$E5,$FB,$C6,$23,$89
	.dc.b	$1F,$F6,$74,$6D,$FA,$26,$CC,$C6,$D5,$7F,$57,$FB,$D6,$9F,$E3,$78
	.dc.b	$C3,$F1,$2F,$F4,$33,$AA,$FE,$E4,$D7,$E2,$9E,$E3,$FC,$96,$E6,$B9
	.dc.b	$C7,$C3,$3F,$3B,$ED,$FF,$01,$23,$D3,$7A,$FF,$00,$8B,$3E,$79,$B9
	.dc.b	$7F,$06,$9B,$5E,$09,$E2,$FC,$A8,$F9,$E4,$7E,$3B,$5F,$A2,$79,$0C
	.dc.b	$90,$CE,$08,$E1,$1D,$4D,$5F,$F8,$07,$E4,$9E,$FF,$CB,$C7,$FC,$C5
	.dc.b	$18,$AF,$E1,$97,$E8,$51,$97,$EA,$CF,$0A,$D6,$1E,$19,$FF,$11,$56
	.dc.b	$F0,$F0,$56,$B0,$F0,$57,$F9,$F2,$B5,$87,$38,$F0,$90,$EE,$6E,$B8
	.dc.b	$BE,$47,$A3,$E0,$73,$3D,$23,$F9,$55,$7A,$E3,$11,$C4,$ED,$7F,$D0
	.dc.b	$67,$8D,$49,$CD,$FD,$B7,$B1,$0F,$C1,$1A,$3C,$C6,$6B,$A3,$93,$C0
	.dc.b	$7B,$0A,$E9,$E6,$66,$FC,$E7,$A3,$0F,$9D,$1A,$9F,$9A,$F7,$AF,$33
	.dc.b	$97,$E6,$3C,$EF,$E3,$B4,$E0,$7C,$62,$9E,$48,$E3,$3E,$53,$BC,$71
	.dc.b	$23,$98,$D9,$23,$E1,$9E,$32,$C7,$22,$47,$A1,$B1,$A8,$E4,$FE,$84
	.dc.b	$EE,$E2,$FE,$26,$1F,$5E,$72,$71,$3D,$86,$0F,$61,$93,$A5,$78,$1C
	.dc.b	$93,$0F,$9F,$19,$AF,$A7,$87,$E2,$CF,$04,$71,$9D,$E1,$E5,$9C,$BE
	.dc.b	$D3,$17,$F3,$A7,$6F,$8E,$FA,$BF,$87,$6B,$E9,$8D,$BE,$68,$CD,$FD
	.dc.b	$18,$E5,$7E,$07,$E0,$C7,$A6,$31,$1F,$02,$FF,$0A,$FF,$3E,$AF,$DF
	.dc.b	$36,$7F,$88,$BF,$68,$C1,$88,$DB,$C8,$86,$CA,$7E,$80,$E2,$7F,$1C
	.dc.b	$7F,$3B,$F0,$9F,$B8,$3C,$AE,$23,$F8,$5C,$62,$75,$18,$3F,$2F,$3B
	.dc.b	$7F,$51,$3B,$74,$60,$DC,$F2,$70,$7E,$16,$79,$5F,$F2,$7A,$E5,$3A
	.dc.b	$8C,$3C,$8E,$AE,$1E,$51,$D6,$30,$7E,$25,$C5,$FF,$30,$7D,$E6,$27
	.dc.b	$F9,$23,$87,$AC,$6D,$F7,$C6,$DF,$E6,$AE,$DF,$34,$6D,$F1,$9F,$5E
	.dc.b	$3E,$DB,$C0,$C1,$0F,$A8,$EA,$8D,$24,$6C,$FC,$43,$87,$89,$FD,$0B
	.dc.b	$C6,$FE,$66,$79,$E7,$9C,$64,$87,$A0,$F8,$26,$2F,$FA,$C4,$F1,$AE
	.dc.b	$73,$9B,$F7,$57,$CE,$7F,$AD,$D6,$A7,$A1,$EF,$3F,$96,$39,$9E,$11
	.dc.b	$FD,$4E,$31,$3C,$A3,$E8,$99,$3C,$05,$3A,$DA,$78,$C6,$6B,$AB,$FD
	.dc.b	$AA,$32,$D3,$A9,$4F,$29,$D8,$F4,$47,$C2,$71,$3D,$DE,$2D,$78,$92
	.dc.b	$3C,$31,$97,$94,$66,$BE,$49,$EF,$70,$7A,$5C,$BE,$F3,$F8,$35,$7E
	.dc.b	$14,$8F,$F0,$36,$BF,$BD,$6B,$C0,$D1,$0D,$16,$52,$89,$F7,$CF,$84
	.dc.b	$FF,$00,$B6,$3E,$1D,$E2,$64,$60,$E5,$7D,$1E,$71,$A7,$DE,$7F,$87
	.dc.b	$BC,$7F,$D0,$11,$FE,$3C,$D7,$F8,$C3,$5F,$E5,$0D,$7F,$C5,$6D,$7F
	.dc.b	$C3,$CF,$EA,$FC,$79,$1F,$CA,$5A,$FF,$23,$6B,$FC,$CD,$AF,$F8,$D9
	.dc.b	$AF,$F8,$A5,$AF,$F8,$DB,$C7,$FD,$C1,$1F,$E6,$CD,$7F,$98,$35,$FE
	.dc.b	$80,$D7,$FC,$86,$D7,$FC,$6C,$FF,$03,$F0,$A4,$7F,$41,$6B,$FC,$ED
	.dc.b	$AF,$F4,$B6,$BF,$E4,$E6,$BF,$E4,$16,$AC,$9F,$07,$F7,$B7,$93,$F6
	.dc.b	$B5,$FC,$23,$B9,$FC,$19,$A7,$FC,$59,$5F,$E3,$E3,$9B,$C4,$FC,$5C
	.dc.b	$6F,$4F,$C2,$47,$86,$F9,$BB,$3B,$16,$BA,$46,$A3,$BD,$73,$BF,$2D
	.dc.b	$78,$63,$57,$DA,$38,$D7,$39,$FF,$2D,$FA,$EE,$CF,$E6,$AD,$9F,$EE
	.dc.b	$2D,$7F,$D0,$6D,$7F,$D6,$AD,$7E,$2B,$5F,$D5,$35,$34,$91,$E0,$24
	.dc.b	$78,$39,$3C,$64,$70,$43,$C6,$E5,$A6,$4F,$2B,$96,$9E,$CE,$6F,$82
	.dc.b	$47,$E8,$70,$BC,$1C,$47,$F9,$9F,$0F,$F5,$C6,$9E,$EB,$FF,$41,$DE
	.dc.b	$EB,$EC,$56,$9F,$83,$3C,$B4,$FB,$F5,$B9,$E9,$3F,$40,$8F,$94,$87
	.dc.b	$D0,$72,$D3,$27,$D2,$72,$D3,$8C,$F8,$3D,$AF,$39,$E3,$C3,$D5,$3F
	.dc.b	$02,$32,$FA,$AB,$4F,$F2,$3B,$4E,$A3,$D2,$7C,$0F,$1F,$F1,$64,$7E
	.dc.b	$5F,$A3,$5F,$95,$17,$1F,$87,$37,$1F,$C4,$CD,$AF,$04,$F1,$7F,$1F
	.dc.b	$1F,$96,$23,$F9,$66,$BF,$96,$6B,$C0,$D1,$0D,$16,$BE,$AF,$8D,$F7
	.dc.b	$0F,$14,$7E,$62,$70,$FE,$1A,$33,$9A,$73,$3F,$EA,$DC,$95,$AF,$53
	.dc.b	$F5,$E4,$9F,$C7,$FB,$D5,$FC,$74,$5C,$7F,$0F,$37,$3F,$D6,$CD,$AF
	.dc.b	$A9,$3C,$5F,$D9,$C7,$EE,$88,$FE,$E9,$AF,$EE,$9A,$F0,$34,$43,$45
	.dc.b	$AF,$CC,$F8,$FF,$A0,$3C,$51,$FB,$C9,$C3,$FA,$A8,$CE,$69,$CC,$FC
	.dc.b	$59,$CF,$C4,$7F,$12,$24,$7E,$CB,$C0,$AF,$EC,$62,$E1,$FA,$A9,$B9
	.dc.b	$7F,$10,$9B,$5F,$97,$9E,$2F,$F1,$68,$FE,$3A,$47,$F8,$EB,$5F,$E3
	.dc.b	$AD,$78,$1A,$21,$A2,$D7,$EF,$3E,$47,$F8,$19,$E2,$8F,$E3,$F3,$87
	.dc.b	$F8,$6C,$67,$24,$E6,$7F,$11,$39,$9F,$E6,$A4,$8F,$E2,$BF,$21,$5F
	.dc.b	$E2,$91,$6F,$F8,$6C,$DC,$7F,$99,$CD,$AF,$DD,$4F,$17,$F9,$CC,$7F
	.dc.b	$40,$23,$FD,$01,$AF,$F4,$06,$BC,$0D,$10,$D1,$6B,$F8,$FF,$C7,$FF
	.dc.b	$26,$3C,$51,$FD,$0A,$70,$FF,$30,$8C,$C7,$F4,$68,$C1,$0F,$79,$FC
	.dc.b	$BB,$E3,$7F,$AC,$91,$FE,$73,$E3,$D7,$F9,$C4,$5B,$FE,$63,$37,$1F
	.dc.b	$EA,$F3,$6B,$F8,$F4,$F1,$7F,$AE,$C7,$F6,$42,$3F,$D9,$1A,$FF,$64
	.dc.b	$6B,$C0,$D1,$0D,$16,$BF,$C0,$3F,$7E,$F8,$FF,$F0,$6C,$61,$F2,$4F
	.dc.b	$F6,$8A,$CE,$1F,$E1,$2B,$E1,$3F,$E0,$76,$BF,$00,$F8,$C6,$6B,$8B
	.dc.b	$FE,$5D,$8D,$DF,$F8,$4B,$16,$1A,$A7,$CA,$B2,$C3,$63,$FC,$2C,$D5
	.dc.b	$7E,$CF,$27,$75,$27,$6F,$5C,$53,$99,$E3,$9F,$F8,$BB,$0E,$98,$7F
	.dc.b	$89,$9B,$1F,$06,$B8,$57,$F8,$D0,$F3,$E1,$FE,$CE,$AF,$24,$65,$AF
	.dc.b	$58,$D1,$1E,$47,$FA,$06,$36,$D8,$D3,$A8,$CB,$E0,$38,$94,$F5,$9F
	.dc.b	$3F,$2F,$80,$7B,$F0,$F6,$56,$27,$ED,$4E,$70,$F3,$46,$08,$74,$3B
	.dc.b	$57,$C5,$F9,$2F,$AA,$48,$FC,$27,$C1,$57,$86,$29,$C5,$7E,$4A,$AD
	.dc.b	$FD,$19,$B7,$F8,$79,$E4,$FC,$B8,$F9,$C4,$7E,$0D,$F5,$6C,$65,$F9
	.dc.b	$0D,$79,$1A,$21,$A3,$F3,$9F,$03,$F8,$73,$C6,$FE,$52,$30,$E8,$FC
	.dc.b	$9B,$AD,$13,$A9,$F8,$33,$F9,$AE,$19,$1F,$9A,$7F,$29,$CB,$88,$F8
	.dc.b	$F1,$73,$FD,$BC,$DC,$FF,$2D,$36,$BF,$59,$3C,$5F,$C8,$47,$E4,$C8
	.dc.b	$FC,$66,$BF,$8B,$6B,$C0,$D1,$0D,$16,$BF,$84,$FA,$BF,$AB,$3C,$51
	.dc.b	$FB,$89,$C3,$FC,$16,$33,$9A,$73,$3F,$8B,$CD,$5A,$F9,$5F,$DF,$92
	.dc.b	$7F,$87,$7A,$F5,$FC,$5C,$5C,$7F,$8D,$CD,$CF,$F7,$53,$6B,$F8,$84
	.dc.b	$F1,$7F,$69,$1F,$B7,$23,$F8,$A6,$BF,$AF,$6B,$C0,$D1,$0D,$16,$BF
	.dc.b	$96,$F7,$FF,$0F,$3C,$51,$FC,$72,$70,$FF,$29,$8C,$E6,$9C,$CF,$C5
	.dc.b	$9C,$F5,$7F,$22,$24,$7F,$31,$F0,$6B,$FA,$E8,$B8,$7F,$3E,$9B,$97
	.dc.b	$F1,$D9,$B5,$FC,$CE,$78,$BF,$C5,$E3,$F8,$D1,$1F,$D6,$B5,$FE,$24
	.dc.b	$D7,$81,$A2,$1A,$2D,$7F,$50,$F4,$FF,$99,$1E,$28,$FE,$7F,$38,$7F
	.dc.b	$A6,$46,$72,$4E,$67,$F1,$13,$9F,$53,$FA,$29,$23,$FD,$B5,$F8,$03
	.dc.b	$C9,$C4,$7F,$12,$9B,$8F,$FB,$8A,$B5,$3F,$D3,$27,$39,$FF,$43,$7F
	.dc.b	$AA,$4E,$A7,$91,$FC,$42,$3F,$A0,$11,$C1,$0F,$E7,$0E,$5A,$64,$FE
	.dc.b	$8C,$E5,$A7,$AB,$CB,$F0,$9E,$47,$FA,$34,$61,$D1,$E9,$21,$FC,$DE
	.dc.b	$53,$B8,$F4,$9F,$CD,$3C,$5F,$ED,$A4,$7E,$4F,$B7,$5D,$1F,$DC,$23
	.dc.b	$E1,$C6,$71,$FE,$CD,$17,$15,$4C,$17,$CE,$3F,$B2,$57,$F6,$72,$3F
	.dc.b	$CE,$9A,$FF,$5F,$6B,$C0,$D1,$0D,$16,$FE,$E7,$B9,$F6,$0F,$24,$7F
	.dc.b	$6D,$9C,$3F,$86,$8C,$BF,$E1,$08,$CB,$F2,$63,$3F,$2E,$F8,$FF,$F0
	.dc.b	$B3,$67,$DF,$6C,$BE,$0C,$27,$4F,$D3,$F8,$0E,$37,$23,$8D,$70,$AD
	.dc.b	$DF,$E5,$6F,$55,$FD,$C6,$73,$5B,$BF,$CA,$D6,$AB,$FC,$29,$5E,$06
	.dc.b	$CF,$F8,$19,$F2,$36,$35,$1C,$CF,$F0,$5B,$87,$ED,$B5,$E2,$6A,$39
	.dc.b	$11,$C1,$FB,$2F,$73,$F5,$27,$AB,$0F,$8D,$1D,$27,$2E,$CF,$86,$64
	.dc.b	$A7,$80,$FC,$E4,$F0,$7E,$24,$7D,$4F,$3A,$47,$0B,$47,$EE,$BE,$47
	.dc.b	$33,$F2,$E2,$E1,$FC,$0E,$6E,$5F,$4E,$6D,$7E,$6E,$78,$BF,$3A,$3E
	.dc.b	$51,$1F,$92,$D7,$FC,$7B,$7E,$27,$24,$33,$1B,$7A,$19,$BF,$E2,$DE
	.dc.b	$A7,$F1,$23,$C7,$1F,$8D,$9C,$3F,$C8,$23,$35,$DE,$3F,$81,$46,$9F
	.dc.b	$80,$7C,$5F,$77,$F3,$64,$7F,$94,$D5,$7F,$21,$16,$FF,$94,$4D,$C7
	.dc.b	$F3,$13,$6B,$F8,$04,$F1,$7F,$27,$1F,$8F,$23,$F8,$E6,$BF,$19,$AF
	.dc.b	$03,$44,$34,$5A,$FE,$75,$F1,$1F,$CD,$8F,$14,$7E,$C6,$70,$FF,$44
	.dc.b	$8C,$E2,$9C,$CF,$F2,$79,$CE,$9F,$BE,$24,$7E,$4B,$F0,$17,$A5,$88
	.dc.b	$FD,$A4,$DC,$7F,$3B,$0A,$DB,$F9,$39,$D6,$7F,$BF,$7F,$D4,$13,$A9
	.dc.b	$E6,$7E,$CA,$3F,$6C,$47,$04,$3F,$1A,$E5,$A6,$4F,$E0,$6E,$5A,$7A
	.dc.b	$F8,$7C,$27,$91,$FD,$FC,$61,$D1,$E9,$21,$FE,$83,$94,$EE,$3D,$27
	.dc.b	$F1,$7F,$83,$FE,$4E,$47,$E5,$7A,$75,$D1,$FB,$78,$F8,$71,$9C,$7E
	.dc.b	$5E,$A9,$C6,$5F,$CB,$B1,$5F,$3C,$3F,$90,$D7,$F1,$D2,$3F,$C7,$1A
	.dc.b	$FE,$D1,$AF,$03,$44,$34,$5C,$3E,$ED,$3D,$1F,$B2,$79,$27,$F9,$3D
	.dc.b	$61,$FC,$44,$65,$FE,$89,$19,$7E,$54,$67,$17,$A5,$FC,$D4,$93,$FC
	.dc.b	$D1,$AF,$F0,$C6,$BF,$D0,$1A,$FF,$4B,$6B,$FC,$E9,$FE,$81,$CD,$23
	.dc.b	$F9,$FB,$5F,$E2,$ED,$7F,$A5,$B5,$FE,$AC,$D7,$FA,$2B,$5F,$EC,$1E
	.dc.b	$C7,$F4,$D2,$3F,$D3,$1A,$FF,$24,$6B,$FD,$61,$AF,$F6,$36,$BF,$D4
	.dc.b	$9F,$EB,$1D,$32,$3F,$AB,$B5,$FE,$5E,$D7,$FB,$1B,$5F,$ED,$CD,$7F
	.dc.b	$AE,$B5,$FE,$F1,$EC,$7F,$65,$23,$FD,$91,$AF,$F3,$C6,$BF,$DC,$1A
	.dc.b	$FF,$7B,$6B,$FD,$A9,$FE,$E1,$D9,$23,$FB,$7B,$5F,$E8,$ED,$7F,$BD
	.dc.b	$B5,$FF,$03,$35,$FE,$EA,$D5,$F3,$2E,$EF,$F0,$9E,$FF,$C5,$23,$13
	.dc.b	$FD,$B1,$AF,$EE,$FB,$F7,$C5,$FE,$1A,$9C,$D7,$1B,$CD,$FF,$8F,$A7
	.dc.b	$35,$FB,$3B,$CC,$F8,$AA,$D6,$A3,$84,$7F,$20,$9C,$BF,$E1,$59,$F1
	.dc.b	$CF,$F8,$A8,$E4,$53,$89,$4C,$99,$78,$BE,$98,$E8,$47,$84,$69,$E6
	.dc.b	$71,$8D,$3F,$C7,$7D,$5F,$E6,$A7,$99,$F9,$11,$87,$FA,$24,$65,$61
	.dc.b	$F1,$DC,$90,$CE,$1E,$57,$E9,$7C,$0F,$C6,$23,$FD,$3F,$D6,$AF,$03
	.dc.b	$E8,$46,$2B,$F8,$9C,$DA,$FB,$13,$6B,$FC,$E7,$C8,$AE,$71,$F3,$72
	.dc.b	$57,$D1,$24,$7C,$66,$BF,$35,$AF,$16,$BE,$03,$81,$0F,$E8,$DE,$57
	.dc.b	$F5,$E3,$C8,$FE,$5A,$30,$FF,$6D,$8C,$CF,$92,$30,$43,$E1,$1C,$4F
	.dc.b	$CD,$79,$1F,$8E,$23,$FD,$DF,$D5,$AF,$09,$4E,$33,$FE,$77,$57,$3F
	.dc.b	$D0,$4D,$BF,$93,$3C,$9F,$91,$1F,$97,$23,$F8,$C6,$BF,$93,$6B,$C0
	.dc.b	$D1,$0D,$16,$BF,$A6,$FA,$BF,$E0,$43,$C5,$1F,$BC,$9C,$3F,$E1,$38
	.dc.b	$CE,$1E,$C8,$C1,$0F,$81,$3F,$B2,$F9,$59,$1F,$C0,$5F,$F0,$8F,$A7
	.dc.b	$88,$F8,$D1,$70,$FE,$BF,$37,$2F,$E0,$93,$6B,$F1,$B3,$C5,$FC,$64
	.dc.b	$7E,$EC,$8F,$EC,$1A,$FE,$DD,$AF,$03,$44,$34,$5A,$FF,$16,$FA,$4F
	.dc.b	$5B,$F8,$7B,$67,$D8,$43,$F0,$EE,$5B,$1E,$97,$F9,$3D,$72,$89,$1C
	.dc.b	$6B,$F9,$47,$A7,$FC,$68,$91,$FE,$94,$B7,$AB,$88,$FC,$74,$DA,$F3
	.dc.b	$CD,$CB,$F9,$64,$DA,$FF,$73,$DF,$27,$F9,$A4,$7F,$23,$23,$FC,$61
	.dc.b	$AF,$F1,$F6,$BF,$BB,$6B,$D8,$E0,$43,$81,$E9,$F5,$7E,$49,$E3,$7F
	.dc.b	$A4,$C6,$1D,$1E,$42,$1F,$32,$34,$FC,$03,$FA,$67,$C8,$7F,$40,$23
	.dc.b	$F1,$EA,$BF,$B3,$8B,$8F,$DC,$9B,$87,$F5,$19,$B5,$F1,$A7,$8B,$FC
	.dc.b	$66,$3F,$A4,$11,$FE,$7A,$D7,$FA,$23,$5E,$06,$88,$68,$B5,$F8,$EC
	.dc.b	$3F,$1A,$78,$A3,$FB,$6C,$E1,$FC,$14,$67,$64,$E6,$7F,$23,$39,$CB
	.dc.b	$FB,$19,$23,$F1,$5E,$AD,$7F,$8B,$45,$C3,$F3,$F3,$72,$FE,$E9,$36
	.dc.b	$BF,$13,$3C,$5F,$E7,$91,$FD,$AC,$8F,$F6,$06,$BF,$D9,$DA,$F0,$34
	.dc.b	$43,$45,$AF,$D8,$FA,$9F,$B0,$3C,$51,$FE,$10,$9C,$3F,$A5,$8C,$E5
	.dc.b	$F9,$D8,$FD,$9C,$69,$F7,$9F,$E1,$33,$FC,$0E,$43,$F5,$DC,$15,$FE
	.dc.b	$74,$5B,$FE,$07,$17,$1F,$F0,$C4,$D3,$F5,$B1,$C5,$FE,$C0,$7F,$83
	.dc.b	$C8,$7F,$81,$0A,$7F,$82,$CA,$70,$34,$43,$45,$9F,$E2,$BE,$9F,$F1
	.dc.b	$43,$C5,$1F,$E3,$43,$07,$F0,$97,$38,$7F,$87,$E7,$8C,$E4,$F8,$80
	.dc.b	$FE,$C3,$DC,$E3,$FF,$17,$B6,$35,$3F,$ED,$28,$C4,$7F,$22,$F4,$F7
	.dc.b	$52,$37,$97,$55,$5C,$2F,$F9,$7D,$EE,$7C,$37,$67,$31,$C8,$F3,$57
	.dc.b	$81,$FF,$1D,$35,$EC,$D7,$73,$FE,$3B,$7C,$2D,$8D,$46,$0F,$F2,$03
	.dc.b	$82,$46,$C9,$1A,$7F,$A2,$F5,$C8,$F3,$1C,$CF,$59,$FC,$ED,$D3,$82
	.dc.b	$57,$F5,$08,$CB,$F3,$67,$E8,$FC,$C4,$AE,$B6,$8F,$F3,$1F,$E0,$4E
	.dc.b	$FD,$57,$DA,$9B,$83,$3F,$A5,$7B,$AE,$75,$A9,$F9,$2F,$4B,$E7,$3F
	.dc.b	$3A,$3E,$41,$1F,$8E,$D7,$FC,$D5,$87,$95,$C9,$0C,$C6,$DE,$86,$70
	.dc.b	$F3,$3F,$93,$C1,$E0,$FB,$8F,$34,$7E,$0A,$B0,$FD,$09,$DB,$EC,$21
	.dc.b	$FE,$3C,$72,$D9,$E6,$FE,$53,$B7,$F3,$E4,$7A,$78,$55,$FD,$2C,$5C
	.dc.b	$3C,$F0,$B5,$67,$3F,$C9,$DD,$AF,$64,$F3,$7F,$33,$1F,$93,$23,$F9
	.dc.b	$26,$BF,$31,$AF,$03,$44,$34,$5B,$F2,$DB,$D2,$FC,$39,$E3,$9F,$80
	.dc.b	$7C,$97,$2F,$E3,$AB,$53,$C6,$34,$64,$FC,$9F,$C1,$7F,$06,$23,$F4
	.dc.b	$FD,$3A,$FF,$0C,$8B,$87,$92,$6E,$5F,$B8,$9B,$5F,$81,$9E,$2F,$EF
	.dc.b	$A3,$F7,$04,$7F,$6E,$D7,$F2,$2D,$78,$1A,$21,$A2,$D7,$C6,$F3,$3F
	.dc.b	$56,$78,$A3,$F8,$BC,$E1,$FC,$74,$67,$64,$E6,$7F,$1D,$39,$F2,$7F
	.dc.b	$94,$92,$3F,$2D,$E9,$57,$F9,$7C,$5C,$3E,$BC,$DC,$BF,$8D,$CD,$AF
	.dc.b	$D2,$4F,$17,$F9,$0C,$7F,$1B,$23,$FC,$69,$AF,$ED,$1A,$F0,$34,$43
	.dc.b	$45,$AF,$C4,$FA,$5F,$C3,$8F,$14,$7F,$3A,$9C,$3F,$B1,$8C,$E4,$9C
	.dc.b	$CF,$EC,$67,$3F,$21,$FD,$30,$91,$FB,$9F,$4E,$BF,$D4,$A2,$E1,$F9
	.dc.b	$A9,$B9,$7F,$3E,$9B,$5F,$C1,$E7,$8B,$FD,$12,$3F,$9F,$11,$FE,$7A
	.dc.b	$D7,$F8,$BB,$5E,$06,$88,$68,$B5,$FA,$CF,$4B,$F9,$89,$E2,$8F,$EB
	.dc.b	$B3,$87,$F8,$A4,$67,$24,$E6,$7F,$8A,$4E,$7D,$2F,$ED,$E4,$8F,$E9
	.dc.b	$7F,$81,$A3,$11,$FD,$D9,$B8,$7F,$12,$9D,$4F,$F4,$69,$CE,$5F,$D9
	.dc.b	$1F,$EE,$55,$B7,$9C,$FF,$6A,$7F,$B2,$11,$FE,$C6,$D7,$F9,$EB,$5E
	.dc.b	$06,$88,$68,$B1,$E5,$F4,$FC,$35,$FE,$03,$B4,$79,$AB,$FB,$ED,$E9
	.dc.b	$FF,$0A,$4E,$9F,$87,$3B,$7F,$C1,$73,$B5,$3B,$7F,$DF,$6D,$BF,$F0
	.dc.b	$C4,$EF,$E7,$1F,$05,$49,$7F,$43,$D5,$AA,$FF,$9F,$D7,$08,$FE,$FE
	.dc.b	$D7,$BC,$F0,$9D,$D5,$CB,$E0,$4F,$4A,$FF,$12,$0B,$E4,$E1,$AF,$F8
	.dc.b	$A5,$AF,$F8,$DD,$AF,$1A,$FF,$0F,$46,$08,$E0,$91,$E0,$39,$38,$7E
	.dc.b	$3F,$07,$CB,$CB,$F7,$51,$EE,$9F,$53,$F5,$27,$46,$1F,$11,$F9,$87
	.dc.b	$2F,$13,$E3,$F0,$7D,$12,$3F,$9E,$F4,$6B,$F3,$F2,$F8,$13,$8C,$FF
	.dc.b	$3D,$56,$73,$5D,$0E,$45,$3D,$D9,$2F,$E2,$92,$3E,$11,$F4,$5C,$3C
	.dc.b	$1A,$FD,$36,$BE,$03,$47,$28,$EC,$FE,$E3,$D1,$F9,$47,$9D,$FB,$13
	.dc.b	$87,$47,$9F,$0F,$C6,$C6,$63,$F3,$33,$C5,$FC,$37,$C2,$7E,$64,$8F
	.dc.b	$F0,$6E,$9A,$FC,$93,$F0,$D1,$8B,$FE,$0D,$37,$1F,$C6,$4D,$AF,$D7
	.dc.b	$CF,$27,$F2,$51,$F8,$C2,$3E,$F6,$BF,$39,$AF,$03,$44,$34,$5A,$FE
	.dc.b	$31,$E8,$7E,$3C,$F1,$47,$B8,$FE,$42,$E5,$FD,$7C,$EB,$64,$E6,$7E
	.dc.b	$01,$C4,$B9,$BE,$43,$F4,$A4,$8F,$B0,$D7,$F0,$ED,$7F,$64,$D7,$F6
	.dc.b	$4D,$7F,$64,$FE,$C3,$AA,$47,$E0,$DA,$FE,$49,$AF,$EF,$5A,$FE,$F5
	.dc.b	$AF,$EF,$5A,$FF,$05,$F5,$FF,$7A,$47,$F2,$0D,$7F,$3E,$D7,$F8,$53
	.dc.b	$5F,$E1,$4D,$7F,$85,$3F,$C2,$3D,$6F,$E4,$84,$8F,$EA,$3E,$9D,$7E
	.dc.b	$5C,$5D,$3F,$A8,$4D,$DB,$F8,$B4,$DA,$FE,$73,$3C,$5F,$E3,$91,$DD
	.dc.b	$FD,$43,$5F,$DB,$B5,$D9,$A2,$1A,$29,$A3,$F9,$FF,$89,$FC,$64,$F1
	.dc.b	$3F,$CB,$63,$0F,$F6,$C8,$CE,$69,$CC,$F2,$3F,$9E,$BF,$C8,$FD,$3F
	.dc.b	$E9,$DF,$82,$7F,$BA,$F1,$FF,$6E,$36,$BF,$AB,$CE,$DF,$F6,$1C,$65
	.dc.b	$37,$74,$EE,$7F,$33,$3B,$35,$19,$CF,$FD,$1F,$19,$9C,$5F,$F6,$BA
	.dc.b	$F2,$C4,$3F,$C1,$2E,$9F,$F7,$F4,$F1,$72,$79,$BC,$DF,$69,$1F,$4F
	.dc.b	$CD,$D7,$CB,$17,$0F,$1C,$DC,$3E,$1C,$DA,$EF,$35,$F8,$31,$6D,$36
	.dc.b	$30,$D9,$F2,$1E,$23,$FD,$DB,$F3,$CF,$2B,$FC,$00,$D8,$CC,$7D,$6F
	.dc.b	$4F,$2C,$8C,$9F,$52,$30,$FF,$08,$9E,$4D,$9E,$4F,$03,$FB,$44,$69
	.dc.b	$FE,$F2,$D4,$47,$0D,$8F,$E9,$EA,$32,$D8,$CC,$60,$8E,$09,$1F,$D6
	.dc.b	$23,$CF,$97,$E1,$3D,$2F,$CA,$9E,$C7,$FD,$AB,$1A,$CF,$F2,$F3,$AA
	.dc.b	$F6,$56,$AF,$11,$F9,$29,$E5,$5F,$9E,$F4,$BF,$43,$EB,$4A,$FC,$EC
	.dc.b	$7E,$90,$C3,$F9,$A8,$C3,$A3,$F4,$6E,$B0,$FC,$DC,$EA,$78,$1A,$38
	.dc.b	$1F,$A4,$78,$57,$EC,$2B,$85,$72,$AE,$07,$23,$81,$FB,$2F,$4B,$F7
	.dc.b	$24,$7F,$6B,$F1,$D5,$FD,$8C,$59,$FD,$7C,$DB,$FD,$FC,$DA,$FD,$5C
	.dc.b	$D7,$F7,$B1,$77,$4D,$8C,$36,$7C,$87,$88,$FC,$EF,$C3,$7E,$70,$8F
	.dc.b	$F0,$AF,$4A,$BF,$97,$8B,$8F,$F0,$49,$B8,$FE,$4A,$BF,$1D,$58,$7F
	.dc.b	$80,$CF,$27,$F3,$31,$F9,$02,$3F,$8D,$6B,$FA,$06,$BC,$0D,$10,$D1
	.dc.b	$6B,$F8,$BE,$BF,$CA,$0F,$14,$79,$0F,$E4,$AE,$5F,$82,$43,$F9,$1B
	.dc.b	$97,$E0,$1F,$AE,$85,$FB,$FE,$69,$E9,$FF,$A0,$D7,$B9,$4C,$44,$BE
	.dc.b	$55,$C6,$75,$18,$3F,$77,$5D,$1C,$07,$0F,$F8,$41,$EE,$E9,$AE,$6B
	.dc.b	$C0,$EE,$7B,$B9,$38,$46,$2F,$DB,$1E,$29,$E3,$18,$8F,$14,$EE,$FD
	.dc.b	$51,$C8,$F1,$7A,$5F,$0C,$8F,$97,$A2,$BE,$98,$B7,$D7,$1F,$EC,$93
	.dc.b	$C8,$86,$CE,$46,$23,$84,$F2,$BF,$A9,$76,$BE,$5C,$D7,$E5,$45,$B4
	.dc.b	$D8,$FE,$7B,$C9,$2B,$FA,$0B,$F7,$7B,$71,$1F,$CE,$62,$E3,$E9,$9B
	.dc.b	$8F,$E8,$F9,$55,$8B,$FB,$B5,$C9,$FE,$77,$1F,$CC,$88,$FF,$2F,$6B
	.dc.b	$FD,$09,$AF,$03,$44,$34,$5B,$F9,$9E,$7F,$E5,$4F,$14,$7F,$43,$9C
	.dc.b	$3F,$89,$8C,$DF,$A2,$30,$43,$E0,$1F,$8F,$F3,$FF,$41,$EF,$48,$C1
	.dc.b	$0F,$CB,$39,$7F,$31,$1F,$A7,$31,$87,$E0,$67,$13,$C6,$30,$FE,$76
	.dc.b	$32,$F2,$30,$43,$04,$8E,$2F,$88,$E9,$97,$EE,$7D,$0F,$DC,$12,$3F
	.dc.b	$55,$E5,$57,$F5,$D1,$6B,$F2,$D3,$6B,$95,$7E,$C6,$BF,$81,$4E,$63
	.dc.b	$F7,$71,$5F,$DD,$45,$DD,$36,$30,$D9,$FF,$08,$9F,$E1,$02,$1F,$C1
	.dc.b	$3F,$05,$79,$B8,$8F,$F0,$61,$71,$FD,$C4,$EA,$39,$90,$D9,$CE,$79
	.dc.b	$3F,$E0,$E3,$FC,$0A,$43,$FC,$04,$53,$FC,$2A,$53,$81,$A2,$1A,$2C
	.dc.b	$F9,$3C,$EF,$E9,$0D,$8F,$23,$FE,$0A,$73,$97,$F4,$DB,$56,$27,$FC
	.dc.b	$43,$18,$78,$CF,$EC,$5D,$08,$D3,$FE,$B8,$6D,$77,$8F,$99,$E7,$70
	.dc.b	$97,$C1,$53,$11,$27,$E9,$46,$5E,$24,$3B,$39,$7C,$11,$C5,$C9,$C8
	.dc.b	$E1,$5F,$62,$38,$3F,$6A,$78,$39,$9E,$07,$23,$A9,$F8,$5F,$3B,$D4
	.dc.b	$47,$F1,$1E,$9D,$7E,$14,$59,$FC,$3C,$DC,$7E,$84,$DA,$F9,$F3,$5F
	.dc.b	$99,$17,$14,$D8,$C3,$67,$C8,$78,$8F,$EB,$BF,$41,$FE,$EC,$6F,$99
	.dc.b	$F6,$23,$11,$F5,$FC,$EC,$4C,$8D,$9F,$A3,$8C,$3F,$EF,$49,$E3,$97
	.dc.b	$F5,$96,$E5,$F8,$BA,$DD,$7F,$86,$44,$F8,$1C,$35,$FF,$0F,$35,$FF
	.dc.b	$38,$B5,$E3,$7F,$E7,$08,$C1,$1C,$12,$3C,$07,$27,$0F,$E5,$5F,$C9
	.dc.b	$1E,$CB,$F5,$38,$8F,$E0,$4E,$6F,$CA,$E6,$39,$1F,$18,$FD,$07,$97
	.dc.b	$F9,$DF,$87,$91,$FC,$2E,$3F,$40,$61,$FD,$34,$61,$D1,$FC,$45,$D5
	.dc.b	$FF,$0E,$9D,$4F,$03,$47,$03,$C2,$60,$D1,$82,$47,$F1,$89,$EB,$7F
	.dc.b	$C0,$6B,$F6,$9E,$4C,$BF,$E4,$4F,$EF,$BF,$06,$FF,$70,$6D,$FE,$DE
	.dc.b	$9C,$90,$63,$FE,$1B,$9D,$D7,$39,$FE,$08,$F3,$71,$3F,$CA,$A2,$E1
	.dc.b	$FE,$5F,$AE,$35,$D6,$B4,$FF,$8E,$23,$2F,$43,$53,$C5,$FC,$F7,$2C
	.dc.b	$7F,$46,$FA,$FC,$AC,$47,$E6,$E2,$DF,$A6,$6D,$7E,$52,$6D,$7F,$B1
	.dc.b	$2D,$5C,$A3,$F3,$73,$F9,$52,$3F,$92,$3F,$42,$E1,$FC,$AE,$BF,$E5
	.dc.b	$17,$4D,$73,$1B,$7A,$9A,$79,$61,$F3,$FD,$FF,$AC,$79,$A7,$B9,$F1
	.dc.b	$DC,$BF,$08,$87,$D8,$C7,$E8,$46,$A3,$E2,$1F,$3F,$B3,$F0,$5F,$15
	.dc.b	$23,$04,$39,$1F,$15,$D3,$81,$9F,$F8,$85,$C3,$70,$FC,$44,$F0,$7F
	.dc.b	$15,$5F,$91,$31,$87,$E1,$67,$17,$E1,$AE,$2F,$E1,$63,$2F,$6A,$E2
	.dc.b	$F9,$4E,$15,$F8,$FF,$2B,$F3,$C4,$8F,$C7,$79,$55,$F3,$C5,$C3,$F3
	.dc.b	$D3,$6B,$95,$7E,$56,$BF,$51,$39,$8F,$C4,$45,$7E,$1C,$5C,$53,$63
	.dc.b	$0D,$9F,$E7,$7D,$3F,$CE,$48,$FE,$8F,$C9,$AF,$F3,$58,$B8,$FE,$CE
	.dc.b	$6E,$3F,$CB,$A6,$D7,$E6,$E7,$8B,$FC,$D2,$3F,$97,$11,$FC,$AB,$5F
	.dc.b	$CC,$B5,$E0,$68,$86,$8B,$5F,$C1,$7D,$EF,$E1,$87,$8A,$3F,$A2,$CE
	.dc.b	$1F,$DF,$46,$77,$FF,$17,$47,$E2,$23,$4F,$BC,$FE,$09,$E4,$FF,$17
	.dc.b	$F8,$B9,$18,$21,$F9,$07,$2F,$E4,$A3,$F8,$F1,$8C,$3F,$8D,$4E,$27
	.dc.b	$8C,$61,$FE,$25,$19,$79,$18,$21,$82,$47,$17,$C4,$74,$CB,$FD,$FD
	.dc.b	$F8,$3F,$FB,$4A,$7F,$D8,$8D,$8C,$F9,$1F,$E3,$08,$FF,$69,$CE,$A3
	.dc.b	$04,$AF,$F6,$EC,$EA,$B8,$C6,$27,$4F,$ED,$67,$48,$3E,$17,$0D,$73
	.dc.b	$1E,$3F,$27,$9D,$F8,$E7,$B1,$E2,$8C,$DF,$8A,$3C,$15,$C5,$FF,$79
	.dc.b	$F7,$BC,$3F,$F2,$6B,$63,$31,$F0,$7C,$9C,$B2,$32,$7C,$08,$C3,$FC
	.dc.b	$46,$79,$36,$79,$3C,$0F,$F0,$3C,$69,$FF,$0C,$35,$F8,$81,$C3,$F1
	.dc.b	$DA,$E6,$3F,$A0,$34,$FE,$85,$1A,$24,$69,$B1,$A8,$FB,$DE,$EC,$8F
	.dc.b	$36,$4C,$7C,$77,$EC,$8F,$C3,$D6,$EB,$DC,$70,$AF,$C6,$C6,$67,$94
	.dc.b	$FA,$6F,$F1,$9E,$4C,$8F,$C4,$BF,$8F,$F5,$71,$1F,$87,$8B,$5F,$96
	.dc.b	$9B,$87,$E6,$A6,$D7,$E2,$26,$BF,$8C,$8B,$AA,$6C,$61,$B3,$E2,$3F
	.dc.b	$36,$FE,$87,$C8,$FD,$16,$32,$7F,$3F,$19,$7F,$43,$19,$7F,$3D,$19
	.dc.b	$7F,$41,$19,$C7,$F5,$F3,$97,$F6,$31,$9C,$7F,$6B,$39,$7F,$6F,$19
	.dc.b	$AF,$D2,$CE,$5F,$D3,$C6,$5F,$D5,$C7,$E7,$32,$91,$FA,$07,$F7,$9E
	.dc.b	$3E,$23,$F3,$11,$70,$FE,$07,$37,$0F,$C0,$CD,$AF,$DA,$CF,$17,$F3
	.dc.b	$31,$F8,$D2,$3F,$92,$6B,$FA,$06,$BC,$0D,$10,$D1,$6B,$F8,$AF,$B5
	.dc.b	$FC,$4C,$F1,$47,$E9,$67,$0F,$F1,$28,$CE,$A9,$CC,$FF,$11,$9E,$31
	.dc.b	$FC,$FA,$BF,$2B,$3B,$E5,$FF,$67,$CE,$71,$FD,$9A,$F9,$DC,$CC,$AF
	.dc.b	$E3,$57,$CA,$B8,$47,$F1,$D7,$83,$AB,$E7,$5C,$63,$FC,$31,$3C,$1F
	.dc.b	$45,$E6,$FA,$55,$BF,$1C,$DA,$ED,$35,$E2,$77,$32,$53,$DC,$5B,$F0
	.dc.b	$1E,$67,$C7,$8F,$C2,$F2,$24,$6E,$3E,$29,$87,$D3,$18,$74,$75,$75
	.dc.b	$8F,$C5,$9D,$4F,$03,$47,$03,$C2,$60,$D1,$82,$47,$CF,$9E,$B8,$FD
	.dc.b	$6A,$FE,$76,$CA,$FE,$7E,$FD,$CF,$23,$11,$FC,$DE,$2D,$FD,$99,$B1
	.dc.b	$FC,$80,$B3,$F6,$E3,$8B,$FC,$E2,$3F,$96,$11,$FE,$62,$D7,$F9,$FB
	.dc.b	$5E,$06,$88,$68,$B3,$F8,$BF,$67,$F2,$67,$8A,$3F,$A2,$4E,$1F,$B5
	.dc.b	$19,$BF,$CD,$D7,$19,$C9,$F0,$8F,$C0,$79,$1F,$9E,$23,$F8,$3F,$A2
	.dc.b	$AF,$E5,$22,$CF,$E7,$66,$E1,$FA,$49,$B5,$F9,$F9,$AF,$EC,$62,$E6
	.dc.b	$9B,$18,$6C,$F9,$0F,$11,$F9,$5F,$1F,$F6,$7E,$44,$8C,$10,$FD,$4B
	.dc.b	$97,$F5,$71,$FB,$A3,$18,$FE,$D6,$71,$3C,$63,$0F,$E8,$E3,$2F,$23
	.dc.b	$04,$30,$48,$E2,$F8,$8E,$98,$FF,$7A,$FA,$37,$7F,$F7,$CA,$FE,$1F
	.dc.b	$18,$AF,$DA,$35,$FE,$07,$E4,$6E,$F9,$5D,$9F,$F1,$45,$6E,$F9,$CF
	.dc.b	$F1,$8B,$D5,$2B,$57,$CA,$3B,$D6,$DA,$7F,$B2,$67,$C4,$FF,$84,$8E
	.dc.b	$51,$FE,$0E,$8C,$C7,$33,$46,$8E,$51,$B2,$38,$21,$C2,$39,$BE,$C8
	.dc.b	$F4,$FE,$10,$F8,$7A,$7B,$2B,$D7,$5B,$3F,$8E,$4E,$DB,$3B,$A9,$18
	.dc.b	$21,$FE,$C7,$AE,$6A,$BF,$DD,$EF,$4F,$2B,$C6,$79,$3D,$8B,$19,$21
	.dc.b	$FE,$D1,$94,$6A,$31,$37,$3F,$E9,$F5,$A9,$E3,$18,$9D,$65,$FE,$42
	.dc.b	$AD,$57,$23,$71,$99,$F2,$46,$0F,$F3,$64,$FA,$9E,$07,$96,$FE,$6F
	.dc.b	$91,$F3,$BC,$99,$3F,$2A,$32,$FC,$C8,$CB,$ED,$8C,$BE,$F8,$CE,$3E
	.dc.b	$C9,$CB,$EE,$8C,$E3,$EB,$9C,$BE,$D8,$CD,$7A,$27,$2F,$AA,$32,$FD
	.dc.b	$58,$FC,$14,$48,$FD,$4B,$F8,$8F,$1F,$11,$F7,$62,$DF,$D0,$9B,$7F
	.dc.b	$B5,$9B,$5F,$22,$78,$BF,$A1,$CD,$3F,$A1,$24,$7E,$9D,$AF,$E6,$DA
	.dc.b	$F1,$6B,$DC,$E0,$43,$F1,$1F,$0D,$FA,$03,$C6,$FE,$E6,$30,$FE,$7A
	.dc.b	$33,$3E,$48,$C1,$0F,$80,$7E,$93,$5F,$D5,$91,$FD,$3F,$AF,$5F,$CC
	.dc.b	$45,$CB,$F2,$B3,$72,$FC,$94,$DA,$FC,$6C,$D7,$F5,$11,$75,$4D,$8C
	.dc.b	$36,$7C,$87,$88,$FD,$CF,$8F,$FC,$03,$C7,$91,$82,$1F,$AD,$72,$FE
	.dc.b	$BE,$3F,$83,$98,$C7,$F5,$B3,$89,$E3,$18,$7F,$83,$C6,$5E,$46,$08
	.dc.b	$60,$91,$C5,$F1,$1D,$31,$FE,$7D,$C3,$FC,$54,$91,$FE,$2F,$FC,$27
	.dc.b	$FD,$B5,$A7,$F1,$58,$FE,$21,$3A,$57,$FE,$52,$AD,$CE,$63,$FA,$3B
	.dc.b	$FC,$F2,$33,$3F,$CD,$E3,$F9,$A1,$1C,$10,$FE,$5C,$E5,$A6,$4F,$CF
	.dc.b	$39,$69,$E9,$F8,$5F,$01,$E3,$C3,$DF,$19,$21,$EC,$3F,$AD,$CF,$07
	.dc.b	$37,$EA,$7E,$27,$8F,$EB,$23,$F1,$FD,$7A,$FC,$4B,$FF,$71,$E2,$AC
	.dc.b	$DF,$C7,$BB,$5A,$9B,$61,$6A,$CC,$62,$6C,$E6,$3C,$A4,$3C,$47,$37
	.dc.b	$C4,$F9,$CD,$5F,$D9,$9E,$75,$F5,$FC,$7F,$B1,$E3,$CB,$FC,$24,$65
	.dc.b	$FC,$34,$65,$EE,$7A,$23,$4E,$30,$FF,$6D,$3C,$0E,$64,$35,$3C,$F0
	.dc.b	$FC,$65,$72,$7F,$1D,$19,$C7,$E6,$CE,$5F,$9F,$19,$AF,$81,$3F,$E0
	.dc.b	$CF,$A6,$7C,$3D,$24,$FE,$7F,$C8,$AF,$AA,$76,$FF,$A4,$5A,$FA,$E3
	.dc.b	$78,$FF,$AC,$A7,$2D,$AF,$D0,$4F,$4A,$FF,$0A,$57,$26,$CF,$F8,$49
	.dc.b	$FC,$04,$65,$FC,$10,$70,$EA,$38,$39,$6C,$6A,$30,$75,$70,$48,$D9
	.dc.b	$FB,$5F,$83,$93,$EF,$8F,$6E,$5F,$9A,$9C,$CF,$C5,$9C,$BF,$8D,$8C
	.dc.b	$E1,$F0,$63,$F8,$57,$93,$FC,$04,$8F,$F0,$DF,$62,$BF,$85,$8B,$3F
	.dc.b	$C2,$E6,$E1,$FA,$A9,$B5,$F9,$B9,$AF,$E9,$A2,$EA,$9B,$18,$6C,$F9
	.dc.b	$0F,$11,$FE,$CD,$FC,$2B,$FD,$B4,$7F,$B2,$F0,$FF,$76,$B7,$0F,$A6
	.dc.b	$EB,$C9,$FF,$2B,$B9,$AF,$DC,$35,$3B,$7B,$79,$5C,$0E,$33,$23,$24
	.dc.b	$3F,$57,$3C,$5B,$3C,$5F,$44,$F1,$72,$71,$BF,$4C,$7E,$87,$C9,$93
	.dc.b	$F9,$B7,$D3,$E4,$62,$3F,$41,$16,$BE,$0C,$DC,$7F,$27,$BA,$B1,$7F
	.dc.b	$02,$B9,$3F,$9E,$8F,$C9,$91,$FC,$93,$5F,$F3,$8C,$F3,$72,$43,$31
	.dc.b	$B7,$A1,$9B,$FA,$BF,$0B,$F4,$8F,$1C,$7E,$9E,$70,$FD,$D8,$CD,$77
	.dc.b	$8F,$9B,$1A,$7E,$09,$F8,$39,$FB,$A4,$7F,$0D,$EC,$D7,$EA,$C5,$BF
	.dc.b	$C3,$CD,$C7,$F0,$F3,$6B,$F0,$B3,$5F,$C2,$C5,$D1,$36,$30,$D9,$F2
	.dc.b	$1E,$23,$F3,$1E,$4F,$E4,$BC,$99,$18,$21,$F8,$A7,$2F,$E3,$23,$F3
	.dc.b	$26,$31,$FC,$FC,$E2,$78,$C6,$1F,$D3,$46,$5E,$46,$08,$60,$91,$C5
	.dc.b	$F1,$1D,$31,$FC,$5E,$DF,$CD,$49,$1F,$B3,$F2,$6B,$FC,$CE,$2D,$7E
	.dc.b	$D2,$6D,$72,$AF,$E6,$D5,$FB,$49,$E3,$1F,$CC,$23,$F9,$29,$1F,$E4
	.dc.b	$8D,$7F,$26,$D7,$81,$A2,$1A,$2D,$7F,$06,$F8,$3F,$E0,$67,$92,$3F
	.dc.b	$9D,$CE,$1F,$E1,$D1,$97,$F9,$E4,$65,$FD,$FC,$67,$47,$93,$FC,$38
	.dc.b	$93,$FE,$05,$B6,$3F,$D1,$BC,$5F,$F7,$04,$7F,$18,$AD,$D7,$F0,$A7
	.dc.b	$FD,$DB,$3A,$AF,$E3,$71,$AA,$E1,$5D,$67,$A3,$97,$6F,$F2,$79,$E6
	.dc.b	$E4,$FE,$42,$D7,$FC,$FC,$6C,$87,$8C,$EC,$7F,$87,$EB,$C8,$DF,$27
	.dc.b	$CF,$E4,$C8,$D9,$FE,$9C,$8F,$59,$97,$FC,$23,$3F,$ED,$0C,$3E,$2C
	.dc.b	$73,$38,$C7,$FB,$36,$33,$1F,$EA,$6A,$E5,$18,$9D,$61,$F3,$6F,$2F
	.dc.b	$CF,$AF,$0D,$F4,$9C,$BE,$B8,$CB,$E4,$8C,$BD,$CE,$07,$B3,$EA,$1D
	.dc.b	$3F,$E0,$0C,$7C,$B1,$89,$D4,$FD,$AF,$2B,$58,$FF,$81,$AB,$55,$FE
	.dc.b	$32,$6D,$F2,$BE,$13,$C8,$86,$DC,$B4,$FF,$0A,$4F,$47,$FC,$26,$1C
	.dc.b	$3E,$18,$FF,$09,$C6,$23,$FC,$B4,$D7,$84,$68,$8F,$66,$BD,$8D,$99
	.dc.b	$3E,$CF,$C1,$7E,$68,$F8,$2F,$68,$AF,$03,$F0,$71,$B7,$DF,$5F,$0D
	.dc.b	$E6,$43,$E0,$1F,$8A,$F2,$FF,$4E,$47,$F1,$3E,$BD,$74,$7E,$32,$3F
	.dc.b	$5B,$19,$9F,$C5,$45,$C7,$F6,$D3,$6B,$F4,$13,$5F,$F7,$BC,$5D,$53
	.dc.b	$63,$0D,$9F,$CD,$F9,$5F,$BC,$F2,$A4,$60,$87,$E9,$1C,$BF,$A6,$8F
	.dc.b	$E0,$66,$31,$FD,$04,$E2,$78,$C6,$1F,$D4,$C6,$5E,$46,$08,$60,$91
	.dc.b	$C5,$F1,$1D,$31,$FC,$C7,$C2,$FD,$8D,$4D,$3F,$8B,$F9,$76,$BF,$4D
	.dc.b	$15,$FD,$FC,$5B,$FD,$3C,$DB,$FE,$49,$3B,$AF,$CC,$DF,$67,$F3,$B5
	.dc.b	$F9,$82,$38,$21,$F9,$38,$E0,$D8,$DB,$4D,$9D,$9F,$E5,$FF,$03,$23
	.dc.b	$D2,$43,$F8,$BB,$97,$F7,$11,$E9,$7F,$92,$46,$70,$F6,$C7,$F2,$2F
	.dc.b	$C2,$DE,$BF,$E8,$1B,$19,$8C,$7B,$19,$64,$64,$E5,$18,$7F,$D7,$33
	.dc.b	$C9,$B3,$C9,$E0,$60,$B5,$FC,$AA,$55,$CE,$31,$57,$3F,$F4,$DD,$6A
	.dc.b	$F8,$C6,$27,$51,$FE,$E1,$9D,$3E,$4F,$2F,$DD,$E5,$C8,$F8,$51,$F0
	.dc.b	$4C,$3E,$A8,$C3,$A3,$E4,$3A,$C3,$D1,$3A,$9E,$06,$8E,$07,$84,$C1
	.dc.b	$A3,$04,$8F,$85,$3D,$70,$F7,$D7,$E7,$70,$97,$FC,$FD,$FB,$1E,$56
	.dc.b	$23,$F9,$C4,$5C,$3E,$64,$DB,$FE,$7D,$36,$BE,$34,$F1,$7F,$3B,$1F
	.dc.b	$CC,$88,$FF,$3A,$6B,$F9,$B6,$BC,$0D,$10,$D1,$6B,$EF,$7C,$17,$E3
	.dc.b	$CF,$14,$7F,$43,$9C,$3F,$95,$8C,$ED,$F8,$BA,$E3,$39,$3E,$11,$F9
	.dc.b	$BF,$2F,$F3,$44,$7F,$3D,$EC,$57,$F2,$B1,$67,$F2,$13,$72,$FC,$6C
	.dc.b	$DA,$FC,$54,$D5,$AA,$70,$A6,$C6,$1B,$3E,$53,$C6,$7E,$C7,$CA,$FD
	.dc.b	$87,$95,$23,$04,$3F,$56,$E5,$FD,$6C,$7E,$DC,$C6,$3F,$A9,$9C,$4F
	.dc.b	$18,$C3,$FB,$F8,$CB,$C8,$C1,$0C,$12,$38,$BE,$23,$A6,$26,$76,$B9
	.dc.b	$7F,$B4,$B6,$3F,$C5,$ED,$55,$FE,$28,$FC,$31,$E5,$F0,$B9,$5C,$1F
	.dc.b	$EE,$73,$A8,$C1,$0E,$78,$78,$D6,$1D,$F0,$ED,$87,$49,$FF,$02,$4F
	.dc.b	$8D,$FF,$09,$B5,$E4,$D7,$DB,$87,$F8,$2D,$A5,$8D,$3E,$08,$D3,$4D
	.dc.b	$38,$8D,$91,$F6,$FC,$14,$9F,$65,$FF,$93,$2B,$D5,$87,$B7,$15,$EE
	.dc.b	$FE,$09,$E4,$29,$E1,$2C,$70,$7F,$CC,$2D,$77,$5D,$9F,$B9,$0F,$33
	.dc.b	$E8,$92,$38,$FB,$35,$EB,$17,$1F,$B9,$37,$1F,$44,$DA,$EB,$35,$5F
	.dc.b	$77,$CB,$3F,$80,$9D,$4E,$67,$B9,$E2,$36,$43,$27,$C0,$8F,$C4,$F9
	.dc.b	$7F,$8A,$F2,$E4,$FC,$28,$CB,$F1,$23,$2F,$C8,$8C,$BF,$2A,$33,$8F
	.dc.b	$E6,$67,$2F,$E6,$E3,$38,$FD,$79,$CB,$F6,$63,$35,$F3,$A7,$2F,$D0
	.dc.b	$8C,$BF,$12,$3F,$57,$E0,$C8,$FD,$5B,$FA,$4F,$2F,$11,$F6,$A2,$E1
	.dc.b	$FA,$D9,$B8,$7E,$72,$6D,$7D,$E9,$E2,$FE,$8A,$3F,$2C,$47,$F3,$CD
	.dc.b	$7F,$52,$D7,$81,$A2,$1A,$2D,$7E,$FF,$E0,$BF,$76,$78,$A3,$F6,$F3
	.dc.b	$87,$F2,$B1,$9D,$53,$99,$FC,$DC,$F1,$D1,$E5,$FF,$06,$24,$FE,$7B
	.dc.b	$D8,$AF,$E8,$A2,$E3,$FC,$46,$6E,$7F,$AD,$9B,$5F,$A2,$9A,$FE,$5A
	.dc.b	$2D,$A6,$C6,$1B,$3F,$AE,$8F,$E3,$31,$FC,$6B,$C9,$FF,$46,$FE,$1B
	.dc.b	$FE,$D6,$A3,$6F,$FD,$EF,$1F,$B6,$9D,$B0,$C1,$0F,$DF,$CF,$33,$94
	.dc.b	$77,$8F,$F6,$FC,$65,$C1,$FC,$12,$3A,$3C,$8C,$E3,$FC,$D6,$32,$FF
	.dc.b	$B7,$A7,$AC,$FF,$3C,$F9,$7F,$E7,$E4,$7B,$F9,$75,$FE,$1B,$16,$B9
	.dc.b	$4D,$CF,$F9,$7C,$DA,$FE,$99,$7F,$EE,$87,$FA,$6C,$7F,$29,$23,$FC
	.dc.b	$C5,$AF,$F3,$C6,$BF,$CF,$1A,$F6,$38,$10,$E0,$7C,$6F,$82,$FA,$A7
	.dc.b	$8D,$FE,$97,$18,$74,$7C,$12,$1F,$4A,$34,$FC,$03,$E8,$F9,$9F,$80
	.dc.b	$23,$F5,$3C,$BA,$FD,$A8,$B9,$7C,$F9,$B8,$FF,$5D,$C5,$58,$BF,$8F
	.dc.b	$55,$FC,$5C,$59,$4D,$8C,$36,$7C,$A7,$8C,$FC,$3F,$95,$F9,$6F,$2E
	.dc.b	$46,$08,$7E,$41,$CB,$F9,$28,$FC,$E9,$8C,$7F,$11,$38,$9E,$31,$87
	.dc.b	$F2,$B1,$97,$91,$82,$18,$24,$71,$7C,$47,$4C,$7F,$C0,$5F,$59,$FC
	.dc.b	$12,$3F,$C2,$D2,$9C,$F0,$FF,$BF,$AB,$55,$FA,$AF,$33,$13,$23,$8D
	.dc.b	$70,$7F,$57,$39,$7F,$B7,$D7,$2B,$ED,$1F,$E2,$58,$C5,$7D,$9A,$E1
	.dc.b	$3F,$E1,$71,$3E,$27,$0D,$7C,$06,$CD,$DF,$F8,$3A,$35,$1D,$E3,$04
	.dc.b	$7F,$C3,$11,$A7,$89,$88,$D3,$FC,$2B,$E0,$7F,$8A,$9E,$DC,$3D,$71
	.dc.b	$C0,$D1,$FC,$32,$35,$87,$9A,$7F,$83,$47,$29,$F9,$27,$94,$FE,$27
	.dc.b	$E6,$FF,$22,$23,$FC,$5B,$DB,$AF,$F1,$C8,$B5,$FC,$46,$6D,$FE,$8E
	.dc.b	$6D,$7F,$08,$9A,$FF,$2C,$8B,$BA,$6C,$61,$B3,$E4,$3C,$47,$A0,$75
	.dc.b	$7F,$A4,$6F,$FC,$9C,$FF,$B3,$30,$F8,$AD,$8C,$F9,$1F,$E5,$39,$CB
	.dc.b	$63,$8F,$99,$6F,$FC,$D5,$7C,$5B,$59,$7C,$2D,$8D,$47,$92,$71,$32
	.dc.b	$3D,$13,$A2,$46,$63,$FC,$8B,$19,$9C,$5F,$9A,$78,$C7,$73,$37,$FA
	.dc.b	$CE,$DF,$D4,$12,$7B,$F9,$75,$FD,$1C,$5B,$F8,$B3,$6B,$95,$7C,$DD
	.dc.b	$D7,$ED,$AE,$53,$FA,$28,$FC,$F1,$1F,$CD,$B5,$FD,$24,$F2,$72,$43
	.dc.b	$31,$B7,$A1,$9B,$FB,$3E,$F7,$CD,$3C,$D1,$FB,$99,$C3,$F3,$23,$2F
	.dc.b	$EE,$23,$2F,$CC,$8C,$E4,$F3,$3E,$C7,$B3,$2B,$EA,$C7,$DB,$30,$FD
	.dc.b	$48,$C3,$A3,$ED,$BA,$C3,$F1,$F3,$A9,$E0,$68,$E0,$7D,$E7,$86,$7F
	.dc.b	$98,$AE,$15,$CA,$B8,$1C,$8E,$07,$E2,$BC,$AF,$C8,$11,$FC,$6F,$95
	.dc.b	$5F,$C5,$C5,$9F,$D1,$CD,$BF,$D3,$CD,$AF,$CF,$CD,$7F,$5B,$17,$1F
	.dc.b	$F7,$54,$78,$48,$76,$3C,$26,$2F,$F6,$33,$C6,$BF,$9E,$78,$D2,$BF
	.dc.b	$1C,$FE,$A7,$CA,$C4,$7F,$38,$8B,$7F,$BA,$9B,$8F,$F4,$A9,$B5,$FB
	.dc.b	$69,$E2,$FF,$36,$8F,$E6,$44,$7F,$97,$B5,$FC,$AB,$5E,$06,$88,$68
	.dc.b	$B5,$FC,$33,$E0,$3F,$80,$1E,$28,$FE,$8F,$38,$7F,$7F,$19,$DF,$FD
	.dc.b	$79,$5C,$67,$27,$C2,$3F,$8E,$7E,$1D,$FE,$EB,$7F,$D8,$EB,$D6,$FE
	.dc.b	$20,$D8,$DF,$99,$83,$33,$23,$24,$39,$B9,$7A,$C7,$17,$C3,$19,$AF
	.dc.b	$E9,$B5,$E3,$6D,$67,$3F,$F7,$E4,$F8,$EF,$99,$DE,$B8,$DE,$E0,$27
	.dc.b	$A5,$72,$F2,$FD,$A4,$9F,$87,$E5,$D7,$E1,$C5,$C3,$CB,$36,$B9,$57
	.dc.b	$CA,$AF,$8F,$39,$8F,$0D,$FF,$B5,$EF,$DB,$5A,$9E,$04,$3B,$9E,$23
	.dc.b	$11,$C1,$B7,$F0,$AF,$FA,$27,$D7,$3E,$B2,$92,$BE,$EF,$15,$58,$7F
	.dc.b	$7A,$AE,$11,$FD,$75,$AF,$E0,$A7,$84,$EE,$AE,$1F,$4A,$7A,$57,$F8
	.dc.b	$66,$F8,$B6,$BF,$C2,$CF,$D2,$8C,$BF,$50,$38,$75,$1C,$1C,$B6,$35
	.dc.b	$18,$3A,$B8,$24,$6C,$FC,$CF,$C1,$C9,$F7,$47,$A9,$FA,$B1,$97,$E2
	.dc.b	$DF,$96,$BE,$1E,$1F,$06,$7F,$4F,$5B,$8C,$47,$EA,$3C,$EF,$D3,$FB
	.dc.b	$52,$3F,$5B,$1F,$AA,$30,$FE,$BA,$30,$E8,$FD,$9B,$A9,$FD,$1C,$EA
	.dc.b	$78,$1A,$38,$1E,$13,$06,$8C,$12,$3F,$5B,$3D,$70,$FD,$5D,$7E,$FB
	.dc.b	$C1,$97,$FC,$11,$FE,$03,$E6,$62,$3F,$82,$C5,$C3,$F6,$F3,$6F,$F6
	.dc.b	$B3,$6B,$F4,$33,$5F,$D9,$45,$DF,$FD,$D1,$1D,$C8,$76,$3F,$8A,$CE
	.dc.b	$E7,$83,$FA,$0E,$99,$3F,$99,$7F,$8B,$F9,$78,$8F,$B6,$7E,$46,$31
	.dc.b	$8F,$F0,$E9,$B8,$FD,$F9,$B5,$FB,$89,$E4,$FE,$76,$3F,$24,$47,$F2
	.dc.b	$2D,$7F,$CE,$9A,$FF,$7F,$72,$43,$31,$B7,$A1,$9B,$FF,$36,$FE,$1E
	.dc.b	$F8,$0F,$F2,$03,$67,$DC,$43,$DA,$1D,$38,$6C,$FE,$FE,$BD,$B3,$BA
	.dc.b	$EA,$62,$BF,$DD,$D1,$CF,$D1,$F0,$9E,$2F,$3A,$47,$F9,$8E,$53,$A8
	.dc.b	$C4,$E9,$F1,$46,$5F,$24,$67,$0F,$7C,$E5,$F8,$31,$9C,$FF,$83,$5F
	.dc.b	$6B,$F0,$5E,$5F,$E7,$B1,$E8,$34,$F8,$0F,$19,$E7,$F3,$3E,$59,$1F
	.dc.b	$57,$99,$5F,$91,$17,$1F,$A7,$37,$0F,$1C,$DA,$F0,$CD,$7E,$94,$5B
	.dc.b	$FF,$77,$C7,$84,$87,$63,$C2,$62,$FF,$03,$3C,$6B,$F9,$EF,$89,$2B
	.dc.b	$F9,$B3,$F8,$4F,$33,$11,$F8,$2D,$BF,$D0,$67,$FA,$13,$14,$EB,$1F
	.dc.b	$E8,$17,$72,$F9,$B3,$CB,$0F,$E7,$53,$FC,$BC,$8F,$F2,$E6,$BF,$9E
	.dc.b	$6B,$C0,$D1,$0D,$16,$BF,$3B,$F0,$9F,$9C,$3C,$51,$FD,$16,$70,$FE
	.dc.b	$36,$32,$FF,$46,$8C,$BF,$9D,$8C,$E0,$F3,$3F,$4F,$E9,$CA,$FC,$B4
	.dc.b	$7E,$A8,$C3,$FA,$58,$C3,$A3,$F3,$8E,$B0,$FD,$9C,$EA,$78,$1A,$38
	.dc.b	$1F,$AF,$78,$65,$F9,$8A,$E1,$5C,$AB,$81,$C8,$E0,$7E,$AF,$CB,$FD
	.dc.b	$F9,$1F,$D7,$79,$75,$FD,$E4,$59,$FE,$0F,$36,$FF,$4F,$36,$BF,$4B
	.dc.b	$35,$FE,$0D,$17,$1F,$F7,$14,$78,$48,$76,$3C,$26,$2F,$F8,$AC,$F1
	.dc.b	$AF,$F6,$A7,$D8,$3C,$EA,$4B,$FF,$74,$7E,$21,$FE,$E7,$7F,$C6,$73
	.dc.b	$B7,$FD,$D8,$D7,$F8,$DB,$70,$65,$FE,$07,$AE,$54,$AD,$4F,$F8,$2D
	.dc.b	$B1,$DA,$FF,$DD,$15,$C2,$7B,$4F,$F8,$44,$E5,$3F,$E0,$F8,$CC,$73
	.dc.b	$8C,$11,$FF,$0D,$C6,$9E,$C6,$23,$4F,$53,$A9,$92,$1F,$CE,$25,$F0
	.dc.b	$9E,$63,$E4,$47,$C2,$9C,$3F,$2A,$33,$87,$C4,$8F,$09,$CE,$36,$F1
	.dc.b	$3E,$8F,$9D,$F0,$BD,$B9,$18,$21,$E3,$72,$F9,$63,$E4,$98,$AF,$04
	.dc.b	$30,$56,$A3,$9C,$EA,$73,$1F,$82,$9D,$3D,$4C,$10,$C1,$27,$94,$78
	.dc.b	$72,$FC,$15,$FD,$CE,$99,$7F,$8E,$7E,$FF,$9B,$88,$F9,$D1,$70,$F2
	.dc.b	$4D,$BF,$C6,$4D,$AF,$F4,$3E,$2A,$C4,$7B,$E6,$CE,$63,$C0,$43,$B9
	.dc.b	$FA,$09,$DC,$F0,$7F,$4F,$F5,$32,$7F,$42,$FE,$5B,$CE,$C4,$7E,$8A
	.dc.b	$2E,$3F,$2E,$6E,$3F,$9F,$9B,$5E,$89,$E2,$FE,$77,$64,$FE,$74,$91
	.dc.b	$F9,$A6,$BF,$97,$8D,$B9,$21,$98,$DB,$D5,$E3,$7F,$7F,$DF,$FD,$79
	.dc.b	$E4,$9F,$DB,$4E,$1F,$E0,$71,$9A,$EF,$1F,$A3,$8D,$3F,$04,$FE,$13
	.dc.b	$E7,$7E,$EB,$DA,$91,$82,$1F,$AB,$72,$FE,$B6,$3F,$80,$98,$C3,$F2
	.dc.b	$93,$89,$E3,$18,$7F,$8B,$C6,$5E,$46,$08,$60,$91,$C5,$F1,$1D,$32
	.dc.b	$FF,$4F,$7B,$FF,$ED,$08,$FF,$69,$B5,$34,$FF,$6E,$4F,$EB,$9C,$79
	.dc.b	$BF,$EA,$78,$C3,$6B,$AF,$9B,$6F,$AC,$D7,$AB,$6F,$B4,$D8,$93,$FE
	.dc.b	$E6,$BE,$AD,$BE,$AD,$3A,$94,$EA,$5B,$EA,$DA,$EA,$FF,$95,$A3,$2F
	.dc.b	$F3,$EF,$07,$FA,$49,$1E,$5E,$6D,$74,$7F,$46,$8F,$54,$67,$0F,$E2
	.dc.b	$7A,$FF,$96,$0F,$F2,$CD,$FF,$BB,$F1,$FE,$89,$7F,$CF,$49,$87,$F3
	.dc.b	$A6,$BF,$A1,$6B,$FA,$96,$BC,$8E,$04,$38,$1F,$07,$E0,$BE,$71,$E5
	.dc.b	$9E,$11,$F0,$63,$2F,$AC,$86,$4E,$46,$30,$FC,$17,$9B,$F6,$7D,$49
	.dc.b	$3F,$3E,$3E,$E1,$87,$E7,$46,$1D,$1F,$55,$D4,$F4,$8C,$9C,$0D,$18
	.dc.b	$74,$E4,$8E,$09,$7F,$9E,$9E,$AF,$E8,$2B,$F2,$1E,$5C,$BF,$C3,$3F
	.dc.b	$93,$F3,$31,$1F,$85,$8B,$87,$E8,$66,$DF,$EA,$A6,$D7,$EA,$A6,$BF
	.dc.b	$AA,$8B,$9A,$6C,$61,$B3,$E2,$3F,$30,$FE,$9F,$AB,$FD,$A7,$1F,$EA
	.dc.b	$C6,$E1,$FD,$81,$B8,$7E,$F3,$CB,$E1,$E5,$CB,$FD,$25,$69,$FD,$FC
	.dc.b	$E5,$E9,$36,$FF,$7D,$39,$9F,$F0,$A5,$78,$9B,$5F,$E0,$B4,$7F,$82
	.dc.b	$1C,$34,$D4,$71,$76,$64,$FF,$0A,$39,$79,$B5,$D1,$E0,$39,$C7,$68
	.dc.b	$F6,$47,$F1,$8F,$82,$FE,$18,$7A,$F0,$EF,$3A,$CD,$3A,$AF,$8B,$7B
	.dc.b	$7F,$91,$C6,$5F,$8A,$7F,$0B,$F3,$3F,$8C,$11,$FE,$21,$F8,$8F,$F8
	.dc.b	$E7,$3F,$E4,$D3,$99,$FE,$1C,$FF,$21,$9D,$65,$FC,$CA,$35,$5C,$2B
	.dc.b	$99,$89,$D3,$E1,$32,$53,$A9,$63,$FC,$47,$5D,$A3,$F9,$ED,$73,$69
	.dc.b	$C8,$FF,$7B,$C7,$FB,$B8,$FF,$6B,$CF,$2F,$33,$D1,$C5,$2B,$DF,$1E
	.dc.b	$B3,$0F,$28,$C3,$A3,$E2,$BA,$9E,$91,$93,$81,$A3,$0E,$9C,$91,$C1
	.dc.b	$30,$FC,$5D,$AB,$F5,$1E,$7C,$BF,$D0,$BF,$47,$CD,$C4,$7E,$AE,$2E
	.dc.b	$1F,$22,$6E,$1F,$88,$9B,$5F,$9E,$C1,$5C,$A3,$F4,$B3,$F9,$32,$3F
	.dc.b	$A0,$6B,$F9,$E6,$BC,$0D,$10,$D1,$6F,$ED,$FB,$DF,$89,$3C,$71,$FE
	.dc.b	$76,$9C,$3F,$92,$8C,$EE,$9C,$CF,$DE,$9E,$38,$7F,$BC,$A3,$F3,$3E
	.dc.b	$67,$E5,$4F,$CB,$7B,$56,$32,$43,$F0,$EE,$5F,$C4,$C5,$C7,$F2,$33
	.dc.b	$73,$FA,$B3,$6B,$E5,$4D,$57,$E3,$79,$9F,$AC,$8D,$46,$6F,$F3,$B5
	.dc.b	$E6,$AF,$D8,$F9,$7F,$AD,$F2,$E4,$60,$87,$E7,$9C,$BF,$A0,$8F,$DA
	.dc.b	$18,$AF,$D4,$CE,$27,$8C,$61,$FD,$FC,$65,$E4,$60,$86,$09,$1C,$5F
	.dc.b	$11,$D3,$1F,$E2,$BF,$29,$FC,$FC,$91,$FC,$33,$CB,$AF,$F4,$18,$B5
	.dc.b	$FB,$E9,$B5,$CA,$BF,$9F,$D7,$F0,$F9,$E3,$1F,$CE,$A3,$F9,$41,$1F
	.dc.b	$E6,$AD,$7F,$99,$B5,$E0,$68,$86,$8B,$5F,$C4,$3E,$0F,$F8,$F1,$E4
	.dc.b	$8F,$E7,$73,$87,$F9,$3C,$65,$FF,$67,$C6,$5F,$E2,$71,$9F,$C4,$BF
	.dc.b	$DD,$A4,$F7,$3F,$C9,$4B,$D1,$B1,$2B,$FC,$85,$3C,$27,$10,$C3,$FD
	.dc.b	$8F,$3C,$E3,$73,$A9,$CD,$5C,$BF,$AB,$DF,$7B,$EB,$39,$7F,$B0,$C7
	.dc.b	$77,$24,$3C,$13,$CA,$7D,$37,$EA,$F3,$3D,$7E,$6C,$BF,$4C,$65,$F5
	.dc.b	$C6,$5E,$67,$82,$34,$E0,$61,$FE,$DE,$70,$DC,$7E,$0C,$F0,$7E,$1D
	.dc.b	$75,$73,$5E,$23,$C2,$43,$9C,$F1,$C7,$E2,$D7,$17,$E3,$C7,$F9,$FF
	.dc.b	$ED,$1E,$45,$25,$FD,$6F,$36,$BE,$99,$DB,$FE,$7B,$6B,$EA,$8D,$E3
	.dc.b	$FF,$00,$27,$2D,$AF,$C2,$CF,$4A,$FF,$07,$8A,$E6,$E1,$AF,$F8,$51
	.dc.b	$AF,$F8,$71,$AF,$19,$FF,$0E,$C6,$08,$E0,$91,$FE,$25,$8C,$BF,$94
	.dc.b	$F8,$5F,$CC,$9E,$AC,$7C,$15,$89,$D1,$F0,$5D,$3F,$8E,$9D,$3F,$1E
	.dc.b	$3F,$47,$E6,$C8,$FD,$0B,$FA,$6F,$77,$11,$F8,$28,$B9,$7E,$8A,$6E
	.dc.b	$1F,$9E,$9B,$5F,$9C,$9A,$FE,$32,$2E,$A9,$B1,$86,$CF,$88,$FD,$9B
	.dc.b	$FB,$8C,$3F,$73,$E6,$49,$FD,$BC,$65,$FD,$CC,$65,$FC,$CC,$65,$FC
	.dc.b	$DC,$67,$1F,$DD,$CE,$5F,$DE,$C6,$71,$FD,$CC,$E5,$FD,$DC,$66,$BF
	.dc.b	$73,$39,$7F,$77,$19,$7F,$57,$1F,$9B,$F3,$E4,$7E,$55,$FF,$6E,$7E
	.dc.b	$27,$FE,$E9,$D5,$1A,$9F,$CE,$3A,$F3,$71,$97,$E6,$62,$E5,$FE,$64
	.dc.b	$C3,$94,$7E,$38,$97,$F8,$B6,$BF,$9C,$3F,$38,$E1,$FC,$E3,$5F,$CE
	.dc.b	$35,$FC,$E3,$CA,$95,$E1,$F7,$BD,$67,$92,$B6,$7F,$BB,$6D,$1A,$8F
	.dc.b	$84,$43,$F8,$94,$69,$E1,$1C,$D4,$7F,$B8,$DF,$F4,$4A,$F7,$3E,$59
	.dc.b	$F3,$3C,$F9,$3E,$18,$CB,$E3,$8C,$BE,$08,$CB,$E2,$8C,$E1,$EE,$8E
	.dc.b	$F3,$A8,$C4,$79,$27,$78,$7D,$3A,$D3,$F5,$63,$33,$C8,$F2,$1E,$33
	.dc.b	$E3,$79,$9F,$84,$23,$F2,$7C,$DA,$FC,$68,$B8,$78,$B1,$56,$30,$FC
	.dc.b	$25,$5B,$F8,$F3,$5F,$97,$16,$B8,$C7,$88,$87,$73,$C4,$62,$FC,$13
	.dc.b	$C6,$BF,$9C,$78,$72,$BF,$99,$3F,$9E,$F3,$71,$1F,$CD,$E2,$DF,$A6
	.dc.b	$6E,$3F,$CA,$E6,$D7,$DE,$9E,$2F,$F3,$18,$FE,$56,$47,$F9,$43,$5F
	.dc.b	$CF,$6E,$8C,$47,$13,$64,$36,$F1,$BF,$99,$F0,$9F,$B6,$3C,$73,$FD
	.dc.b	$0E,$70,$FE,$76,$33,$5D,$E3,$04,$3E,$01,$FA,$3F,$37,$F8,$0F,$99
	.dc.b	$23,$04,$3F,$46,$E5,$FD,$2C,$7F,$08,$31,$87,$E0,$67,$13,$C6,$30
	.dc.b	$FE,$C6,$32,$F2,$30,$43,$04,$8E,$2F,$88,$E9,$97,$F0,$FF,$2F,$F8
	.dc.b	$C9,$23,$FD,$F7,$3F,$A4,$F1,$FF,$D9,$51,$FB,$89,$D3,$FB,$E7,$FD
	.dc.b	$F7,$3B,$8C,$BF,$9B,$8B,$5F,$F0,$02,$B9,$EE,$9C,$DF,$F0,$27,$9D
	.dc.b	$6E,$32,$78,$4F,$E2,$51,$A4,$2B,$AC,$70,$AE,$D3,$EC,$AE,$8D,$38
	.dc.b	$FD,$B3,$E7,$7A,$42,$30,$EC,$C4,$6B,$0F,$F6,$84,$66,$76,$FF,$8C
	.dc.b	$DB,$8F,$F0,$E6,$DF,$C3,$F3,$3B,$C6,$2B,$C7,$86,$0E,$D3,$63,$FA
	.dc.b	$01,$E2,$2C,$F8,$1F,$F0,$FB,$5F,$8A,$6C,$FE,$90,$E5,$EE,$72,$9E
	.dc.b	$AE,$CE,$6E,$23,$84,$64,$EA,$FA,$5F,$A5,$F0,$9F,$14,$F4,$3F,$0E
	.dc.b	$70,$FC,$BC,$53,$AC,$7E,$35,$6A,$BF,$CB,$95,$A7,$E7,$47,$D1,$F3
	.dc.b	$64,$73,$7E,$A7,$BB,$88,$E3,$16,$FF,$39,$37,$3F,$CA,$4D,$AF,$C5
	.dc.b	$E0,$AC,$47,$E3,$66,$CE,$63,$C0,$43,$B9,$ED,$9D,$CF,$07,$F1,$1E
	.dc.b	$5F,$E2,$7C,$D9,$5F,$0E,$32,$FC,$58,$CB,$F0,$A3,$2F,$C4,$8C,$E3
	.dc.b	$FB,$79,$CB,$FB,$98,$CE,$3F,$B1,$9C,$BF,$B3,$8C,$D7,$C0,$9C,$BF
	.dc.b	$0A,$32,$FE,$B6,$3F,$2E,$47,$F4,$0F,$E8,$FC,$CC,$47,$E8,$62,$DF
	.dc.b	$F0,$B9,$B7,$F8,$89,$B5,$F4,$27,$8B,$F9,$A8,$FC,$81,$1F,$CF,$B5
	.dc.b	$FC,$DB,$5E,$06,$88,$68,$B5,$FC,$77,$D9,$FD,$D1,$E2,$8F,$E0,$13
	.dc.b	$87,$F2,$D1,$9D,$93,$99,$FC,$AC,$F1,$D1,$E6,$7F,$05,$F6,$A5,$7E
	.dc.b	$7E,$3F,$84,$18,$7F,$3B,$18,$74,$7E,$95,$D6,$3F,$CC,$E7,$53,$C0
	.dc.b	$D1,$C0,$FD,$3B,$C3,$3F,$E6,$15,$C2,$B9,$57,$03,$91,$C0,$FE,$35
	.dc.b	$B7,$EF,$C8,$FF,$0E,$6B,$FC,$29,$AF,$F1,$06,$BF,$C1,$1A,$FF,$2E
	.dc.b	$7F,$86,$FE,$28,$F3,$3F,$97,$BF,$EE,$D9,$FF,$6D,$36,$BF,$C6,$AE
	.dc.b	$E3,$F9,$93,$5D,$F9,$9C,$8E,$75,$23,$71,$97,$8C,$65,$E2,$68,$C9
	.dc.b	$0C,$92,$BC,$71,$C5,$F2,$CF,$59,$F4,$CF,$F3,$CF,$4B,$F9,$41,$23
	.dc.b	$F9,$BB,$5F,$E6,$6D,$7F,$9C,$B5,$FE,$5A,$D7,$FA,$BB,$FC,$CF,$96
	.dc.b	$47,$F4,$26,$BF,$CF,$9A,$FF,$44,$6B,$FC,$E1,$AF,$F6,$26,$BF,$D8
	.dc.b	$7E,$1B,$FA,$01,$1F,$E9,$ED,$7F,$A5,$B5,$FE,$A2,$D7,$FA,$2B,$5F
	.dc.b	$ED,$EF,$F4,$BF,$80,$91,$FD,$69,$AF,$F5,$66,$BF,$D7,$1A,$FF,$50
	.dc.b	$6B,$FD,$E9,$AF,$F7,$93,$FA,$C1,$0F,$EC,$E5,$3F,$B1,$94,$FE,$D2
	.dc.b	$53,$FA,$E9,$4F,$F0,$39,$FD,$9B,$EE,$3F,$C1,$AF,$FA,$CD,$AF,$F7
	.dc.b	$00,$E1,$FE,$E6,$D7,$FC,$32,$D7,$7D,$5D,$48,$FF,$80,$23,$44,$8C
	.dc.b	$C7,$87,$EE,$3C,$84,$78,$35,$D1,$82,$18,$21,$E9,$94,$71,$8C,$7A
	.dc.b	$FF,$04,$8F,$95,$AF,$C3,$6B,$C0,$D1,$0D,$14,$F0,$76,$FC,$D2,$3F
	.dc.b	$05,$AF,$CF,$6B,$C0,$D1,$0D,$14,$F6,$77,$7D,$92,$3F,$35,$AF,$DB
	.dc.b	$6B,$C0,$D1,$0D,$14,$F9,$1F,$1F,$F8,$52,$3F,$65,$AF,$E1,$DA,$F0
	.dc.b	$34,$43,$45,$3E,$A7,$83,$F9,$12,$3F,$85,$6B,$F9,$36,$BC,$0D,$10
	.dc.b	$D1,$4F,$C0,$7B,$5F,$9D,$23,$F9,$16,$BF,$9F,$6B,$C0,$D1,$0D,$14
	.dc.b	$FC,$67,$89,$FA,$92,$3F,$9D,$6B,$FA,$B6,$BC,$0D,$10,$D1,$4F,$CC
	.dc.b	$7D,$37,$ED,$48,$FE,$A5,$AF,$ED,$DA,$F0,$34,$43,$45,$3F,$49,$ED
	.dc.b	$7F,$01,$23,$FB,$56,$BF,$C0,$DA,$F0,$34,$43,$45,$3F,$61,$E3,$FF
	.dc.b	$0D,$23,$FC,$05,$AF,$F0,$F6,$BC,$0D,$10,$D1,$4F,$DE,$79,$1F,$C6
	.dc.b	$48,$FF,$0D,$6B,$FC,$6D,$AF,$03,$44,$34,$53,$F8,$47,$CB,$7F,$25
	.dc.b	$23,$FC,$65,$AF,$F2,$76,$BC,$0D,$10,$D1,$4F,$E2,$9E,$EF,$F3,$12
	.dc.b	$3F,$C9,$5A,$FF,$33,$6B,$C0,$D1,$0D,$14,$FE,$41,$E6,$7F,$3D,$23
	.dc.b	$FC,$C5,$AF,$F3,$F6,$BC,$0D,$10,$D1,$4F,$E5,$9E,$67,$F4,$92,$3F
	.dc.b	$CF,$5A,$FF,$4B,$6B,$C0,$D1,$0D,$14,$FE,$71,$ED,$7F,$55,$23,$FD
	.dc.b	$25,$AF,$F5,$76,$BC,$0D,$10,$D1,$4F,$E8,$9E,$77,$F6,$12,$3F,$D5
	.dc.b	$5A,$FF,$63,$6B,$C0,$D1,$0D,$14,$FE,$A1,$EF,$FF,$6D,$23,$FD,$85
	.dc.b	$AF,$F6,$F6,$BC,$0D,$10,$D1,$4F,$EB,$9E,$8F,$F7,$92,$3F,$DB,$5A
	.dc.b	$FF,$7B,$6B,$C0,$D1,$0D,$14,$FF,$00,$53,$E9,$3F,$DD,$6D,$8F,$F7
	.dc.b	$CB,$5F,$F0,$30,$70,$F2,$8E,$24,$78,$B5,$E3,$1C,$1F,$F7,$FB,$51
	.dc.b	$D6,$3C,$7F,$15,$23,$C6,$D7,$9B,$5F,$5B,$5E,$AD,$7A,$CA,$31,$1E
	.dc.b	$BF,$57,$E0,$91,$F8,$2D,$7D,$2D,$7E,$4B,$5F,$5B,$5E,$EF,$CC,$F5
	.dc.b	$64,$7C,$C6,$BF,$11,$AF,$D4,$6B,$F1,$9A,$FA,$9A,$FD,$3F,$89,$FB
	.dc.b	$24,$7E,$CB,$5F,$A2,$D7,$F0,$2D,$7E,$9B,$5F,$8E,$FE,$13,$E0,$E4
	.dc.b	$7E,$11,$AF,$DC,$6B,$F8,$C6,$BF,$79,$AF,$D2,$6B,$F8,$BF,$63,$F2
	.dc.b	$24,$7F,$22,$D7,$F1,$2D,$7F,$32,$D7,$F1,$6D,$7E,$FB,$F9,$CF,$66
	.dc.b	$47,$E7,$1A,$FE,$51,$AF,$E9,$1A,$FE,$59,$AF,$E2,$9A,$FE,$8F,$E7
	.dc.b	$BF,$52,$47,$F5,$2D,$7F,$42,$D7,$F6,$2D,$7F,$46,$D7,$F2,$EF,$ED
	.dc.b	$3D,$C9,$1F,$B4,$6B,$FA,$C6,$BF,$BC,$6B,$FA,$E6,$BF,$A2,$6B,$FB
	.dc.b	$BF,$6F,$F8,$09,$1F,$E0,$2D,$7F,$72,$D7,$F8,$4B,$5F,$DD,$B5,$FD
	.dc.b	$7B,$FC,$33,$DD,$91,$FC,$31,$AF,$F0,$46,$BF,$C5,$1A,$FF,$06,$6B
	.dc.b	$FB,$A6,$BF,$C4,$FE,$0F,$F8,$C9,$1F,$E3,$2D,$7F,$88,$B5,$FE,$42
	.dc.b	$D7,$F8,$9B,$5F,$E0,$EF,$F2,$4F,$8C,$91,$FC,$91,$AF,$F1,$C6,$BF
	.dc.b	$CB,$1A,$FF,$1E,$6B,$FC,$49,$AF,$F2,$BF,$81,$FE,$62,$47,$F9,$8B
	.dc.b	$5F,$E5,$2D,$7F,$9C,$B5,$FE,$56,$D7,$F8,$FB,$FC,$F3,$E0,$64,$7F
	.dc.b	$3C,$6B,$FC,$D1,$AF,$F4,$46,$BF,$CD,$9A,$FF,$2A,$6B,$FD,$0F,$E7
	.dc.b	$FF,$A4,$91,$FE,$92,$D7,$FA,$0B,$5F,$EA,$2D,$7F,$A1,$B5,$FE,$6E
	.dc.b	$FF,$54,$F8,$19,$1F,$D5,$1A,$FF,$4C,$6B,$FD,$71,$AF,$F4,$E6,$BF
	.dc.b	$D0,$9A,$FF,$5B,$F4,$3F,$B0,$91,$FE,$C2,$D7,$FA,$CB,$5F,$ED,$2D
	.dc.b	$7F,$AD,$B5,$FE,$9E,$FF,$6C,$F4,$24,$7F,$6C,$6B,$FD,$91,$AF,$F7
	.dc.b	$46,$BF,$D9,$9A,$FF,$5A,$6B,$FD,$CF,$D5,$FE,$F2,$47,$FB,$CB,$5F
	.dc.b	$EE,$2D,$7F,$C0,$2D,$7F,$B9,$B5,$FE,$CE,$FF,$80,$1F,$EE,$BF,$72
	.dc.b	$F8,$4C,$42,$30,$ED,$AF,$F8,$21,$AF,$F7,$E3,$9C,$72,$8C,$C6,$5A
	.dc.b	$E8,$E9,$1B,$23,$83,$C1,$EB,$F3,$23,$82,$1E,$37,$2D,$32,$79,$4F
	.dc.b	$63,$A7,$12,$8F,$47,$B1,$27,$E0,$35,$F8,$0D,$78,$B5,$F7,$35,$F9
	.dc.b	$AD,$7E,$67,$C0,$7C,$52,$3F,$35,$AF,$CD,$6B,$E6,$6B,$F2,$DA,$FD
	.dc.b	$A7,$E8,$7B,$72,$3E,$C3,$5F,$B0,$D7,$E0,$B5,$FA,$CD,$7F,$0A,$D7
	.dc.b	$F0,$9F,$0B,$F7,$48,$FE,$15,$AF,$E1,$5A,$FC,$E6,$BF,$83,$6B,$F9
	.dc.b	$27,$F1,$1E,$FC,$8F,$C8,$35,$FC,$83,$5F,$B2,$D7,$F1,$CD,$7F,$3A
	.dc.b	$D7,$F3,$9F,$3B,$F9,$52,$3F,$9D,$6B,$F9,$D6,$BF,$86,$6B,$F9,$B6
	.dc.b	$BF,$AA,$7F,$41,$F1,$92,$3F,$50,$D7,$F5,$0D,$7F,$22,$D7,$F4,$CD
	.dc.b	$7F,$6A,$D7,$F6,$9F,$05,$FA,$D2,$3F,$B5,$6B,$FB,$56,$BF,$9E,$6B
	.dc.b	$FB,$36,$BF,$C0,$9F,$DC,$7C,$0C,$8F,$E0,$0D,$7F,$80,$35,$FD,$4B
	.dc.b	$5F,$DF,$35,$FE,$1A,$D7,$F8,$67,$C5,$7F,$05,$23,$FC,$35,$AF,$F0
	.dc.b	$D6,$BF,$B6,$6B,$FC,$2D,$AF,$F1,$A7,$F8,$87,$C1,$48,$FE,$30,$D7
	.dc.b	$F8,$C3,$5F,$E0,$2D,$7F,$8B,$35,$FE,$4A,$D7,$F9,$27,$C2,$7F,$1D
	.dc.b	$23,$FC,$95,$AF,$F2,$56,$BF,$C3,$9A,$FF,$23,$6B,$FC,$C9,$FE,$51
	.dc.b	$F0,$72,$3F,$98,$35,$FE,$60,$D7,$F8,$CB,$5F,$E5,$CD,$7F,$9E,$B5
	.dc.b	$FE,$79,$F4,$7F,$CD,$48,$FF,$3D,$6B,$FC,$F5,$AF,$F2,$66,$BF,$CE
	.dc.b	$DA,$FF,$4A,$7F,$A0,$7C,$34,$8F,$E9,$0D,$7F,$A4,$35,$FE,$62,$D7
	.dc.b	$FA,$33,$5F,$EA,$AD,$7F,$AA,$7C,$1F,$F4,$D2,$3F,$D5,$5A,$FF,$55
	.dc.b	$6B,$FC,$F9,$AF,$F5,$36,$BF,$D8,$9F,$EB,$1F,$07,$23,$FB,$03,$5F
	.dc.b	$EC,$0D,$7F,$A4,$B5,$FE,$BC,$D7,$FB,$6B,$5F,$ED,$8F,$F6,$7F,$C3
	.dc.b	$BF,$DE,$11,$FE,$E6,$FB,$B7,$C0,$E7,$CE,$FE,$D9,$18,$9F,$F7,$5D
	.dc.b	$70,$79,$98,$9F,$0B,$4E,$95,$CD,$A6,$4E,$22,$7A,$74,$B0,$E1,$2C
	.dc.b	$3C,$4E,$F0,$F2,$E1,$AB,$E1,$3D,$5A,$EC,$F0,$94,$E9,$1C,$6E,$47
	.dc.b	$96,$FB,$E1,$EF,$6E,$1E,$E9,$E5,$87,$6A,$D6,$1E,$8B,$DD,$7D,$23
	.dc.b	$A4,$64,$8F,$9F,$0A,$F9,$67,$8B,$E6,$34,$72,$21,$B9,$F4,$B4,$CD
	.dc.b	$79,$1E,$26,$48,$E5,$F5,$D7,$8A,$30,$4C,$3E,$03,$BE,$BF,$F0,$E6
	.dc.b	$1F,$51,$B8,$7F,$8C,$DB,$80,$38,$69,$F6,$8D,$39,$23,$93,$67,$DA
	.dc.b	$72,$D3,$F2,$FF,$03,$E1,$69,$D0,$A7,$23,$F1,$8E,$DF,$D1,$35,$C9
	.dc.b	$C4,$C9,$4C,$9C,$E9,$1C,$5B,$1B,$6B,$B8,$D7,$C0,$C8,$D3,$5E,$8D
	.dc.b	$AF,$9E,$D8,$FD,$AB,$5F,$DD,$35,$FD,$47,$C0,$F8,$C8,$F8,$DA,$FA
	.dc.b	$5B,$3F,$71,$B1,$FC,$09,$AF,$F0,$77,$DD,$F0,$32,$3D,$CD,$7E,$23
	.dc.b	$67,$F0,$ED,$8F,$E1,$AD,$7F,$89,$35,$FE,$01,$F0,$3F,$2C,$8F,$CB
	.dc.b	$6B,$F4,$5B,$3F,$94,$6C,$7F,$1A,$6B,$FC,$7D,$FA,$DF,$03,$23,$EB
	.dc.b	$35,$FB,$8D,$9F,$CF,$B6,$3F,$92,$B5,$FE,$54,$D7,$F8,$C3,$FC,$6B
	.dc.b	$EF,$1E,$33,$D7,$FE,$45,$18,$7F,$88,$4F,$F0,$93,$84,$EE,$BF,$8C
	.dc.b	$56,$DF,$E0,$D3,$B7,$9C,$EF,$7F,$E6,$D5,$B7,$F8,$D5,$6E,$B9,$D6
	.dc.b	$F8,$3F,$D5,$57,$B7,$DB,$3E,$43,$AE,$11,$F0,$99,$39,$B9,$C3,$BC
	.dc.b	$59,$EB,$7D,$E3,$F8,$25,$F7,$7E,$49,$B3,$31,$B5,$1B,$7E,$69,$B3
	.dc.b	$31,$B8,$D6,$EC,$7E,$5E,$3B,$9C,$C6,$2F,$D2,$4C,$7D,$2D,$C7,$D3
	.dc.b	$7D,$F0,$E9,$3C,$F1,$F0,$1E,$78,$CC,$71,$3D,$73,$C7,$0F,$06,$1F
	.dc.b	$5A,$F8,$CF,$CB,$8C,$3D,$4C,$61,$E2,$BE,$87,$C0,$8C,$BD,$8C,$E3
	.dc.b	$EE,$79,$1C,$31,$F2,$DF,$68,$DE,$1E,$9C,$7C,$B8,$F7,$AD,$1F,$10
	.dc.b	$AE,$49,$5F,$68,$E7,$8F,$BD,$F0,$CF,$07,$4D,$7D,$47,$96,$BC,$2F
	.dc.b	$AE,$35,$87,$A2,$FB,$CF,$09,$CD,$70,$8C,$4E,$AB,$84,$73,$8E,$38
	.dc.b	$FC,$58,$F3,$9E,$13,$CF,$1C,$6B,$A4,$FE,$44,$E1,$7E,$97,$18,$FC
	.dc.b	$9B,$E2,$77,$3D,$F7,$E0,$8F,$BF,$1F,$AA,$3C,$35,$EA,$78,$9E,$A3
	.dc.b	$89,$CC,$C1,$30,$F6,$B8,$BF,$1E,$3F,$84,$8C,$5F,$82,$FE,$D6,$1E
	.dc.b	$3A,$FA,$D5,$E2,$8F,$80,$F8,$AF,$DF,$1B,$8F,$8C,$6C,$F2,$61,$E2
	.dc.b	$9F,$0C,$6A,$FE,$36,$1E,$1B,$FE,$1B,$3C,$4F,$56,$3F,$3A,$75,$1F
	.dc.b	$00,$F4,$E3,$E6,$BE,$6D,$BF,$24,$7C,$99,$E9,$87,$A1,$F9,$0F,$92
	.dc.b	$3E,$0C,$62,$FB,$BF,$C3,$0E,$2F,$DB,$6D,$FC,$BA,$E5,$7F,$1A,$72
	.dc.b	$7D,$BB,$F2,$BF,$06,$3C,$A6,$E3,$27,$9C,$AE,$4D,$C7,$CC,$C3,$E2
	.dc.b	$5F,$BE,$3D,$0D,$7E,$29,$DC,$C9,$4C,$D7,$7A,$EB,$1F,$40,$FB,$D8
	.dc.b	$7B,$E7,$C7,$1E,$13,$35,$F0,$67,$E3,$E1,$F3,$67,$AD,$FA,$5F,$0D
	.dc.b	$F7,$6C,$65,$D9,$D4,$F9,$31,$F5,$1C,$9E,$53,$31,$F8,$88,$F2,$C6
	.dc.b	$5B,$3E,$33,$EA,$C7,$58,$CD,$7B,$6B,$E0,$D6,$E3,$D6,$66,$BF,$95
	.dc.b	$1F,$E5,$92,$9F,$CC,$4C,$1F,$C6,$0B,$1F,$DD,$5B,$1F,$DE,$5B,$7C
	.dc.b	$1E,$64,$39,$1D,$48,$7F,$66,$F1,$DF,$0B,$C6,$3F,$B1,$C6,$1E,$96
	.dc.b	$8C,$47,$91,$AF,$F7,$E6,$BD,$67,$A4,$F1,$6C,$EA,$7B,$C6,$08,$F8
	.dc.b	$23,$53,$C6,$3F,$C2,$F1,$E5,$C7,$FD,$91,$F7,$AE,$0F,$F4,$0C,$7F
	.dc.b	$85,$A7,$59,$F9,$71,$FF,$12,$C6,$30,$FF,$45,$61,$C7,$6C,$D4,$9E
	.dc.b	$EF,$FB,$36,$75,$18,$26,$1D,$E7,$58,$F8,$B1,$F0,$56,$9B,$5E,$89
	.dc.b	$D4,$FC,$73,$45,$8D,$61,$E3,$7B,$46,$86,$1E,$48,$C9,$2F,$B9,$F8
	.dc.b	$F8,$D3,$89,$EC,$D7,$9D,$F9,$1F,$AA,$77,$CB,$E9,$5F,$92,$31,$97
	.dc.b	$CB,$77,$19,$3C,$4F,$62,$1D,$B1,$F6,$61,$E1,$BE,$AF,$9F,$0F,$14
	.dc.b	$62,$B8,$B5,$E4,$78,$B2,$E6,$FD,$C9,$E3,$8F,$96,$B8,$1C,$72,$F8
	.dc.b	$75,$D2,$74,$7A,$E3,$31,$EC,$C3,$CD,$39,$BE,$B8,$78,$2B,$47,$87
	.dc.b	$1C,$63,$E0,$BE,$11,$87,$BD,$7E,$22,$30,$F3,$C7,$C3,$5C,$63,$18
	.dc.b	$78,$30,$F8,$03,$2F,$A1,$97,$9A,$B7,$8F,$4A,$FC,$46,$3D,$32,$F0
	.dc.b	$C6,$31,$E9,$8F,$D8,$8F,$59,$F3,$9C,$65,$E3,$8F,$06,$5F,$5F,$1E
	.dc.b	$D5,$C7,$1E,$D9,$62,$B5,$7E,$8A,$D6,$3E,$09,$F4,$4F,$43,$E5,$1B
	.dc.b	$7F,$2D,$7D,$23,$06,$AB,$CB,$97,$B1,$F8,$71,$F7,$70,$F8,$6F,$03
	.dc.b	$DB,$5C,$E7,$8C,$FD,$39,$D6,$5F,$87,$C7,$CD,$1C,$CF,$31,$F1,$71
	.dc.b	$F4,$DF,$B3,$2F,$81,$3F,$A0,$C7,$B5,$68,$F8,$75,$E0,$9F,$A5,$5C
	.dc.b	$1F,$2D,$F8,$1F,$A7,$87,$8E,$74,$D8,$F3,$E5,$F8,$E6,$E5,$EB,$C3
	.dc.b	$B1,$E5,$8C,$9E,$AC,$BC,$AF,$B6,$3E,$4B,$E0,$AF,$99,$97,$A0,$F5
	.dc.b	$47,$B7,$2F,$36,$3F,$85,$BE,$2D,$AF,$B3,$3A,$BE,$77,$F6,$67,$9B
	.dc.b	$F0,$32,$F6,$CF,$AF,$0F,$4D,$72,$9C,$E5,$ED,$8E,$38,$FC,$3A,$DE
	.dc.b	$1E,$5B,$DB,$EF,$C7,$C5,$87,$E1,$2B,$73,$E1,$3E,$2B,$EA,$C7,$CF
	.dc.b	$19,$8D,$E1,$DF,$1F,$0D,$79,$DB,$7E,$5B,$F9,$91,$E9,$7A,$BF,$2D
	.dc.b	$C9,$E7,$32,$7C,$53,$27,$85,$EA,$FA,$F1,$F6,$E1,$E7,$AF,$C5,$65
	.dc.b	$F7,$2B,$ED,$47,$8D,$E4,$77,$9F,$C2,$12,$FC,$CD,$8D,$46,$08,$E0
	.dc.b	$87,$43,$E5,$5F,$F8,$72,$FF,$B4,$B7,$22,$9A,$7F,$06,$74,$64,$AE
	.dc.b	$8F,$E0,$AE,$08,$F1,$21,$A3,$FB,$17,$07,$68,$F2,$36,$33,$1D,$DA
	.dc.b	$77,$3F,$B7,$C6,$DF,$F0,$4B,$5C,$C7,$17,$2D,$72,$73,$A4,$F1,$6C
	.dc.b	$ED,$B3,$BE,$0D,$12,$74,$D8,$D4,$60,$8E,$08,$FD,$38,$E0,$FD,$5D
	.dc.b	$FC,$44,$7C,$4D,$9F,$43,$63,$81,$A2,$46,$9A,$7E,$07,$E1,$3D,$C4
	.dc.b	$7D,$CD,$8F,$88,$D8,$E0,$68,$91,$A6,$9F,$8D,$F8,$4F,$94,$47,$E5
	.dc.b	$36,$3E,$83,$63,$81,$A2,$46,$9A,$7E,$67,$E1,$3E,$B1,$1F,$AC,$D8
	.dc.b	$FB,$8D,$8E,$06,$89,$1A,$69,$A7,$F9,$94,$7F,$8C,$71,$7C,$05,$89
	.dc.b	$5F,$E5,$37,$51,$8A,$E3,$39,$3F,$CB,$11,$9D,$BF,$CC,$F5,$76,$FF
	.dc.b	$22,$4D,$BF,$F3,$1C,$D7,$FC,$9B,$17,$04,$D8,$98,$77,$AF,$E9,$51
	.dc.b	$87,$C6,$62,$35,$87,$F9,$E2,$74,$F9,$23,$CE,$6B,$0F,$2B,$6B,$FC
	.dc.b	$EF,$5C,$27,$FC,$B9,$7F,$0A,$B5,$5E,$86,$E1,$F2,$9F,$F4,$F3,$63
	.dc.b	$FD,$50,$D7,$F8,$B5,$FE,$FD,$EF,$3C,$23,$11,$D4,$E2,$E2,$79,$47
	.dc.b	$A0,$CC,$F1,$3C,$3F,$01,$B2,$38,$24,$79,$23,$2D,$8C,$BF,$CA,$E3
	.dc.b	$2D,$8F,$6E,$1E,$32,$38,$24,$7C,$08,$CB,$63,$2F,$F3,$78,$CB,$63
	.dc.b	$E4,$FC,$07,$BC,$8E,$09,$1F,$32,$32,$D8,$CB,$FD,$0E,$32,$D8,$FA
	.dc.b	$BF,$01,$F2,$C8,$E0,$91,$F6,$23,$2D,$8C,$BF,$D3,$E3,$2D,$8F,$C0
	.dc.b	$F2,$FD,$72,$38,$24,$7E,$12,$32,$D8,$CB,$FD,$6E,$32,$D8,$FF,$75
	.dc.b	$7D,$F3,$E0,$7F,$0E,$D8,$CC,$60,$8E,$09,$3F,$EF,$B9,$E0,$D9,$DB
	.dc.b	$67,$6F,$F6,$93,$FB,$54,$BE,$07,$C2,$49,$F2,$B6,$7B,$35,$E4,$27
	.dc.b	$94,$F0,$24,$70,$6B,$C2,$1F,$03,$F0,$09,$3F,$11,$B3,$ED,$6B,$C5
	.dc.b	$AF,$AE,$79,$CF,$12,$47,$C5,$F8,$1F,$9A,$48,$FA,$2D,$8D,$46,$08
	.dc.b	$E0,$87,$A6,$38,$3E,$BF,$81,$FB,$24,$9F,$BA,$D8,$FA,$CD,$78,$46
	.dc.b	$88,$E9,$AF,$C7,$F8,$1F,$C2,$92,$3F,$12,$D8,$FC,$13,$5E,$11,$A2
	.dc.b	$3A,$6B,$F4,$FE,$07,$F2,$24,$8F,$CA,$B6,$3F,$1C,$D7,$84,$68,$8E
	.dc.b	$9A,$FD,$FF,$81,$FC,$E9,$23,$F4,$2D,$8F,$CD,$35,$E1,$1A,$23,$A6
	.dc.b	$BF,$8B,$F8,$1F,$D4,$92,$3F,$5A,$D8,$FD,$33,$5E,$11,$A2,$3A,$6B
	.dc.b	$F9,$7F,$81,$FD,$A9,$23,$F7,$2D,$8F,$D9,$35,$E1,$1A,$23,$A6,$BF
	.dc.b	$A3,$F8,$1F,$E0,$24,$8F,$E0,$AD,$8F,$DF,$35,$E1,$1A,$23,$A6,$BF
	.dc.b	$AF,$F8,$1F,$E1,$A4,$8F,$E2,$2D,$8F,$E1,$4D,$78,$46,$88,$E9,$AF
	.dc.b	$EE,$FE,$07,$F8,$C9,$23,$F8,$EB,$63,$F8,$B3,$5E,$11,$A2,$3A,$6B
	.dc.b	$FC,$1F,$E0,$7F,$92,$92,$3F,$94,$B6,$3F,$91,$35,$E1,$1A,$23,$A6
	.dc.b	$BF,$C4,$FE,$07,$F9,$89,$23,$F9,$AB,$63,$F9,$73,$5E,$11,$A2,$3A
	.dc.b	$6B,$FC,$7F,$E0,$7F,$9E,$92,$3F,$A0,$B6,$3F,$9D,$35,$E1,$1A,$23
	.dc.b	$A6,$BF,$CA,$FC,$7F,$E9,$24,$8F,$E9,$AD,$8F,$E8,$CD,$78,$46,$88
	.dc.b	$E9,$AF,$F3,$7F,$1B,$FA,$A9,$23,$FA,$CB,$63,$FA,$93,$5E,$11,$A2
	.dc.b	$3A,$6B,$FD,$0F,$E0,$7F,$B0,$92,$3F,$B2,$B6,$3F,$AF,$35,$E1,$1A
	.dc.b	$23,$A6,$BF,$D3,$FE,$07,$FB,$69,$23,$FB,$8B,$63,$FB,$53,$5E,$11
	.dc.b	$A2,$3A,$6B,$FD,$6D,$FE,$F2,$43,$FB,$E9,$5F,$EE,$CD,$78,$3A,$23
	.dc.b	$A6,$BF,$DB,$BE,$FD,$F0,$5F,$E4,$C6,$CF,$F7,$29,$FF,$72,$46,$1C
	.dc.b	$35,$E5,$3C,$49,$3C,$61,$3B,$7F,$C0,$CD,$7A,$CF,$07,$AB,$E3,$F8
	.dc.b	$29,$5C,$E9,$18,$AF,$1B,$5F,$63,$5E,$ED,$7D,$6D,$7C,$BF,$05,$F0
	.dc.b	$49,$5D,$DB,$1F,$01,$AF,$C9,$6B,$EC,$6B,$F2,$1F,$99,$F0,$52,$7D
	.dc.b	$4D,$9F,$96,$D7,$EA,$35,$F8,$ED,$7E,$9B,$5F,$9B,$F0,$5F,$64,$93
	.dc.b	$F1,$DB,$1F,$61,$AF,$E0,$5A,$FD,$46,$BF,$80,$7F,$09,$F0,$52,$7E
	.dc.b	$93,$67,$F0,$6D,$7F,$18,$D7,$EF,$B5,$FC,$5B,$5F,$C2,$FC,$17,$E4
	.dc.b	$49,$3F,$7D,$B1,$F9,$06,$BF,$99,$6B,$F8,$C6,$BF,$98,$7F,$39,$F0
	.dc.b	$52,$7F,$14,$D9,$FC,$DB,$5F,$D2,$35,$FC,$BB,$5F,$D1,$B5,$FC,$EF
	.dc.b	$C1,$7E,$A4,$93,$F9,$76,$C7,$EA,$1A,$FE,$C5,$AF,$E9,$1A,$FE,$C1
	.dc.b	$FD,$A7,$C1,$49,$FD,$13,$67,$F6,$6D,$7F,$78,$D7,$F5,$ED,$7F,$76
	.dc.b	$D7,$F6,$BF,$05,$FC,$04,$93,$FA,$F6,$C7,$F0,$06,$BF,$C2,$5A,$FE
	.dc.b	$F1,$AF,$F0,$87,$F8,$67,$C1,$49,$FD,$D3,$67,$F8,$5B,$5F,$E2,$8D
	.dc.b	$7F,$83,$B5,$FE,$26,$D7,$F8,$6F,$C1,$7F,$19,$24,$FF,$07,$6C,$7F
	.dc.b	$18,$6B,$FC,$85,$AF,$F1,$46,$BF,$C8,$1F,$E4,$9F,$05,$27,$F8,$93
	.dc.b	$67,$F9,$1B,$5F,$E5,$8D,$7F,$8F,$B5,$FE,$56,$D7,$F9,$2F,$C1,$7F
	.dc.b	$31,$24,$FF,$1F,$6C,$7F,$30,$6B,$FC,$E5,$AF,$F2,$C6,$BF,$CE,$1F
	.dc.b	$E7,$9F,$05,$27,$F9,$53,$67,$F9,$DB,$5F,$E8,$8D,$7F,$9B,$B5,$FE
	.dc.b	$86,$D7,$F9,$EF,$C1,$7F,$49,$24,$FF,$37,$6C,$7F,$48,$6B,$FD,$45
	.dc.b	$AF,$F4,$46,$BF,$D4,$1F,$EA,$9F,$05,$27,$FA,$13,$67,$FA,$9B,$5F
	.dc.b	$EB,$8D,$7F,$A7,$B5,$FE,$B6,$D7,$FA,$AF,$C0,$FF,$61,$24,$FF,$4F
	.dc.b	$6C,$7F,$60,$6B,$FD,$A5,$AF,$F5,$C6,$BF,$DA,$1F,$ED,$9F,$13,$27
	.dc.b	$FA,$D3,$67,$FB,$5B,$5F,$EE,$8D,$7F,$B3,$B5,$FE,$E6,$D7,$FB,$6F
	.dc.b	$BB,$FD,$E4,$93,$FD,$9D,$B1,$FD,$E1,$AF,$F8,$05,$AF,$F7,$46,$BF
	.dc.b	$E0,$07,$FC,$08,$C8,$FE,$E4,$D3,$FC,$06,$53,$FC,$18,$53,$FB,$F9
	.dc.b	$4F,$F0,$59,$47,$5B,$E2,$53,$FE,$09,$8C,$36,$33,$3F,$E1,$59,$CB
	.dc.b	$63,$28,$87,$67,$2E,$23,$8B,$A3,$53,$DC,$8F,$56,$BB,$87,$C1,$F4
	.dc.b	$8F,$04,$62,$7D,$53,$99,$F7,$B6,$75,$1E,$B8,$C1,$27,$89,$23,$4D
	.dc.b	$7C,$5F,$07,$CC,$91,$F2,$5B,$1F,$1D,$AF,$08,$D1,$1D,$35,$F7,$7C
	.dc.b	$1F,$9C,$91,$F5,$5B,$1F,$4D,$AF,$08,$D1,$1D,$35,$F9,$5F,$03,$F0
	.dc.b	$C9,$1F,$81,$6C,$7D,$F6,$BC,$23,$44,$74,$D7,$EB,$7C,$0F,$CF,$24
	.dc.b	$7E,$35,$B1,$F8,$B6,$BC,$23,$44,$74,$D7,$F0,$5F,$03,$F6,$C9,$1F
	.dc.b	$99,$6C,$7E,$5D,$AF,$08,$D1,$1D,$35,$FC,$77,$C0,$FE,$1C,$91,$FA
	.dc.b	$56,$C7,$E8,$DA,$F0,$8D,$11,$D3,$5F,$CD,$7C,$0F,$E4,$C9,$1F,$B1
	.dc.b	$6C,$7E,$BD,$AF,$08,$D1,$1D,$35,$FD,$37,$C0,$FE,$7C,$91,$FB,$D6
	.dc.b	$C7,$EE,$DA,$F0,$8D,$11,$D3,$5F,$D9,$7C,$07,$EA,$C9,$1F,$C2,$5B
	.dc.b	$1F,$C1,$DA,$F0,$8D,$11,$D3,$5F,$DF,$7C,$47,$ED,$C9,$1F,$C5,$5B
	.dc.b	$1F,$C4,$DA,$F0,$8D,$11,$D3,$5F,$E1,$5F,$11,$FC,$0C,$91,$FC,$85
	.dc.b	$B1,$FC,$7D,$AF,$08,$D1,$1D,$35,$FE,$2D,$F1,$1F,$C3,$C9,$1F,$CB
	.dc.b	$5B,$1F,$CA,$DA,$F0,$8D,$11,$D3,$5F,$E4,$5F,$11,$FC,$6C,$91,$FC
	.dc.b	$E5,$B1,$FC,$DD,$AF,$08,$D1,$1D,$35,$FE,$5D,$F0,$9F,$C9,$C9,$1F
	.dc.b	$D1,$5B,$1F,$D0,$DA,$F0,$8D,$11,$D3,$5F,$E7,$5F,$07,$FC,$CC,$91
	.dc.b	$FD,$45,$B1,$FD,$3D,$AF,$08,$D1,$1D,$35,$FE,$A3,$F8,$7B,$F1,$47
	.dc.b	$C0,$FF,$5A,$F8,$CF,$EB,$72,$5F,$07,$9C,$3F,$AD,$4F,$1A,$C6,$08
	.dc.b	$E1,$58,$75,$5F,$D8,$6F,$A9,$D1,$C1,$D3,$0F,$1B,$8B,$E9,$3B,$7C
	.dc.b	$91,$BC,$3B,$4E,$4F,$4D,$72,$AD,$9C,$B0,$F2,$46,$67,$C7,$18,$AE
	.dc.b	$6D,$BF,$16,$1E,$C6,$DF,$17,$E1,$DF,$85,$F2,$1C,$1C,$61,$F1,$67
	.dc.b	$83,$E3,$9F,$73,$FE,$10,$F8,$4F,$EE,$44,$8F,$F0,$83,$63,$FC,$1C
	.dc.b	$D7,$FC,$00,$D7,$FC,$14,$D7,$FC,$02,$FF,$80,$EF,$F0,$1F,$07,$F0
	.dc.b	$BE,$13,$E1,$B6,$FD,$AD,$BF,$93,$5C,$27,$E0,$D7,$06,$9C,$0B,$3C
	.dc.b	$30,$5F,$3C,$7E,$AB,$6B,$F0,$0D,$AF,$8C,$DC,$3C,$11,$CF,$0F,$03
	.dc.b	$6B,$E4,$FE,$05,$D9,$FC,$40,$F9,$C7,$95,$F9,$6A,$56,$DF,$C2,$D6
	.dc.b	$E7,$53,$9C,$BF,$C7,$B3,$93,$BE,$5F,$A0,$71,$7F,$8E,$3A,$C6,$70
	.dc.b	$F0,$CF,$03,$CA,$E2,$78,$5F,$8D,$FF,$20,$35,$FF,$23,$B5,$FF,$25
	.dc.b	$35,$EF,$3F,$E4,$77,$35,$FE,$3D,$9D,$11,$C1,$27,$84,$F2,$39,$4F
	.dc.b	$C1,$8F,$0C,$FE,$FB,$C1,$7E,$B6,$E1,$F8,$F3,$2B,$DD,$87,$EC,$2F
	.dc.b	$75,$98,$F7,$D5,$CB,$FC,$D3,$5A,$BE,$31,$FB,$AF,$85,$FD,$BF,$C3
	.dc.b	$7E,$E1,$B3,$FB,$26,$E1,$F8,$F6,$D7,$19,$FD,$BD,$71,$69,$C4,$B5
	.dc.b	$C5,$B3,$C6,$3F,$35,$3C,$5B,$3C,$5A,$71,$2C,$FF,$12,$F8,$5F,$E1
	.dc.b	$E7,$F1,$0F,$86,$B5,$FC,$2A,$2D,$7E,$9E,$6D,$71,$AD,$1F,$C4,$23
	.dc.b	$2F,$F2,$7A,$B5,$FC,$36,$6F,$0F,$F3,$D9,$BC,$49,$B1,$87,$8D,$7E
	.dc.b	$93,$E1,$3E,$F6,$EF,$89,$FD,$1C,$61,$B9,$7F,$35,$DB,$F0,$71,$8B
	.dc.b	$FC,$05,$F1,$AF,$DA,$DF,$16,$9C,$63,$53,$FA,$29,$FC,$C1,$27,$F2
	.dc.b	$BC,$28,$C4,$FE,$59,$AF,$EA,$1A,$FE,$A1,$AF,$EA,$1E,$3A,$FF,$AD
	.dc.b	$27,$A5,$7F,$4F,$FC,$55,$F7,$4E,$AF,$65,$EF,$0B,$7F,$EB,$BB,$DD
	.dc.b	$63,$1E,$D3,$EE,$7F,$A8,$CE,$A3,$17,$FD,$5B,$E2,$3F,$A9,$9F,$D5
	.dc.b	$3E,$26,$CF,$F4,$E8,$B7,$FC,$AE,$6D,$71,$AD,$1F,$D5,$23,$2F,$F6
	.dc.b	$BA,$B5,$FD,$46,$6F,$17,$F8,$06,$6F,$12,$6C,$61,$CF,$C4,$FF,$68
	.dc.b	$D3,$FB,$4B,$71,$FE,$C6,$DC,$7F,$A0,$37,$2F,$EE,$6D,$9F,$EF,$CD
	.dc.b	$9F,$EC,$8D,$CB,$FC,$28,$DC,$8D,$25,$7F,$72,$C3,$C6,$D3,$FA,$1F
	.dc.b	$C2,$7E,$6B,$E2,$FF,$9D,$B6,$32,$7E,$46,$32,$D3,$38,$FE,$6A,$32
	.dc.b	$D8,$CB,$FC,$DA,$7F,$97,$12,$3F,$2E,$D8,$FE,$56,$D7,$84,$68,$8E
	.dc.b	$9B,$97,$E5,$67,$F6,$7D,$BE,$28,$C4,$F9,$9F,$E5,$3C,$49,$D4,$FF
	.dc.b	$48,$AB,$6E,$4F,$E7,$35,$8D,$7F,$C8,$5F,$13,$FE,$24,$3F,$C4,$BF
	.dc.b	$15,$71,$FF,$0F,$C5,$C3,$FB,$E4,$DC,$FF,$C5,$F3,$6B,$FC,$75,$36
	.dc.b	$BF,$C4,$13,$6F,$FC,$A1,$37,$74,$D8,$C3,$6F,$C2,$7F,$8C,$3E,$27
	.dc.b	$FC,$73,$F1,$5F,$E3,$A6,$C6,$4F,$F1,$AC,$65,$A6,$71,$FF,$0A,$46
	.dc.b	$5B,$19,$CB,$FC,$93,$39,$6C,$64,$FF,$2B,$46,$5A,$66,$7F,$C7,$D1
	.dc.b	$96,$C7,$F8,$FB,$F0,$57,$C2,$FF,$9A,$2D,$F3,$BF,$DB,$EC,$2B,$77
	.dc.b	$FE,$63,$AD,$F8,$1F,$E1,$B9,$CD,$7F,$97,$A7,$35,$DF,$0E,$D8,$75
	.dc.b	$AE,$91,$8C,$39,$36,$B8,$1F,$E0,$4C,$3A,$D7,$90,$F2,$B8,$70,$D7
	.dc.b	$32,$9E,$4D,$8F,$F0,$B4,$ED,$F1,$CE,$88,$EA,$78,$3E,$13,$66,$63
	.dc.b	$CE,$FF,$2E,$EA,$72,$7B,$27,$39,$72,$C3,$E3,$BE,$3E,$9F,$E7,$F5
	.dc.b	$BC,$3F,$C7,$55,$AB,$F6,$E3,$FE,$9E,$BD,$C7,$43,$FD,$31,$C9,$FE
	.dc.b	$8C,$F8,$CF,$F4,$6B,$63,$FD,$0E,$DC,$3F,$CC,$8D,$CB,$FD,$2E,$D9
	.dc.b	$E6,$7F,$A3,$AB,$9B,$4E,$65,$AE,$6D,$AE,$6D,$39,$C7,$F9,$CE,$79
	.dc.b	$B6,$3F,$D3,$7E,$B7,$FA,$90,$FF,$52,$FC,$65,$8C,$92,$7F,$CF,$91
	.dc.b	$97,$FC,$FD,$36,$FF,$D6,$33,$6B,$FD,$79,$36,$BF,$D4,$53,$79,$7F
	.dc.b	$DA,$53,$79,$53,$63,$0F,$1A,$FD,$27,$C1,$7F,$85,$BC,$9F,$D1,$37
	.dc.b	$1F,$D7,$36,$7E,$DB,$72,$FC,$13,$67,$F3,$2D,$9F,$CE,$BF,$90,$73
	.dc.b	$1F,$A3,$8C,$46,$A3,$04,$7F,$3B,$1A,$7C,$F1,$97,$F3,$5B,$A7,$4D
	.dc.b	$9D,$4F,$80,$F4,$BF,$E2,$FF,$D6,$FF,$D9,$31,$D0,$C9,$2F,$FC,$D7
	.dc.b	$1A,$AE,$93,$9A,$E9,$5A,$F2,$FE,$4C,$7F,$B7,$0F,$F7,$2F,$C6,$FF
	.dc.b	$B9,$9A,$64,$FF,$70,$B9,$69,$9A,$FF,$60,$46,$5B,$19,$AF,$F7,$AC
	.dc.b	$E5,$B1,$93,$F9,$0E,$CD,$95,$B9,$CC,$61,$E1,$7F,$EE,$D9,$D3,$63
	.dc.b	$FD,$F2,$BE,$2F,$F8,$96,$1F,$C4,$F0,$FE,$5F,$18,$AF,$0D,$3E,$37
	.dc.b	$0A,$B7,$3D,$2B,$31,$89,$FF,$73,$9C,$67,$44,$73,$3C,$1D,$18,$8D
	.dc.b	$5F,$96,$3D,$11,$EA,$BE,$15,$23,$67,$A2,$31,$3C,$C9,$5C,$E3,$2F
	.dc.b	$F2,$3F,$81,$FF,$5E,$6A,$F3,$FF,$A6,$46,$1B,$87,$E4,$5B,$87,$FA
	.dc.b	$F6,$B8,$4F,$F5,$DA,$E0,$D3,$81,$67,$FA,$0C,$7E,$80,$91,$FD,$3D
	.dc.b	$B1,$A8,$C1,$1C,$10,$FD,$5C,$70,$BF,$64,$FE,$9F,$A7,$CF,$18,$AF
	.dc.b	$E7,$B5,$F0,$27,$35,$FD,$6A,$2C,$FF,$89,$EF,$FB,$46,$E9,$FB,$1F
	.dc.b	$19,$F6,$7A,$2E,$1F,$8E,$6E,$1F,$82,$6E,$5F,$6D,$B3,$F8,$E6,$CF
	.dc.b	$E3,$4F,$EA,$05,$AE,$D3,$F9,$A6,$C7,$E3,$9B,$5F,$87,$6C,$FE,$69
	.dc.b	$B3,$CE,$3C,$C7,$E9,$3E,$2F,$F2,$47,$E4,$FE,$32,$C6,$49,$3F,$9C
	.dc.b	$8C,$BF,$9E,$9B,$7F,$97,$9B,$5F,$AA,$9B,$5E,$D9,$BC,$1F,$E1,$E9
	.dc.b	$B2,$9B,$18,$78,$C3,$F0,$77,$C0,$FE,$63,$D0,$FD,$AC,$E6,$31,$8F
	.dc.b	$F8,$AB,$2F,$F0,$4D,$F0,$8F,$E5,$57,$C2,$31,$7F,$9B,$8C,$5F,$28
	.dc.b	$5F,$28,$C3,$6F,$FC,$29,$3F,$E3,$8B,$E6,$FF,$85,$DA,$FF,$8A,$5A
	.dc.b	$EE,$BF,$A3,$46,$23,$D5,$3D,$1A,$76,$8D,$12,$78,$92,$7E,$01,$DD
	.dc.b	$FD,$EF,$A3,$EC,$8C,$D7,$85,$AF,$6B,$FE,$93,$3B,$8F,$83,$3C,$CD
	.dc.b	$DF,$C8,$AD,$C6,$2F,$F8,$E7,$C6,$7F,$1D,$F8,$DB,$3F,$CC,$9B,$87
	.dc.b	$F2,$66,$DF,$F1,$F6,$CF,$F3,$26,$CF,$E9,$5B,$AF,$F9,$99,$B0,$69
	.dc.b	$27,$FA,$0D,$F8,$DA,$78,$CF,$E8,$BD,$5F,$CC,$CF,$E6,$9F,$1B,$63
	.dc.b	$24,$BF,$E7,$B1,$97,$F9,$FC,$DC,$7F,$9C,$4D,$AF,$E9,$73,$6B,$F7
	.dc.b	$D3,$71,$FF,$3F,$CD,$94,$D8,$C3,$C6,$3F,$45,$EF,$7F,$94,$FE,$D9
	.dc.b	$EB,$FE,$6E,$30,$DC,$7E,$FF,$88,$AD,$C7,$DA,$AD,$B4,$DC,$6A,$FF
	.dc.b	$02,$7E,$76,$76,$E0,$F1,$CF,$E5,$F7,$4F,$E7,$09,$19,$6C,$FE,$6D
	.dc.b	$AF,$E9,$DA,$FE,$9D,$AF,$86,$3C,$CF,$93,$F1,$47,$DC,$3C,$AF,$81
	.dc.b	$3E,$5C,$3C,$77,$C3,$7F,$F4,$96,$1C,$2F,$71,$FC,$1B,$1B,$5F,$EB
	.dc.b	$2A,$D3,$C4,$FF,$02,$FC,$87,$F7,$93,$FB,$D7,$B9,$63,$FC,$1F,$17
	.dc.b	$2F,$F0,$1C,$DC,$3F,$BE,$4D,$AF,$F0,$7C,$DA,$FE,$85,$37,$87,$FD
	.dc.b	$87,$36,$53,$63,$0D,$AF,$09,$FE,$20,$F8,$FF,$F0,$77,$C8,$7F,$83
	.dc.b	$DB,$19,$3F,$C4,$B1,$96,$99,$BF,$F0,$DC,$65,$B1,$9C,$7F,$C2,$D3
	.dc.b	$96,$C6,$4F,$F1,$5C,$65,$A6,$67,$FB,$04,$65,$B1,$FC,$3B,$D9,$FF
	.dc.b	$0A,$FB,$9F,$CD,$DB,$19,$3F,$CC,$91,$96,$99,$9F,$F0,$B4,$65,$B1
	.dc.b	$97,$F9,$B4,$FF,$32,$24,$7E,$6D,$B1,$FC,$C5,$AF,$08,$D1,$1D,$36
	.dc.b	$FF,$A8,$CF,$F8,$BB,$DD,$F1,$46,$27,$CC,$FF,$9E,$23,$31,$FD,$1E
	.dc.b	$6D,$7F,$84,$E6,$E8,$87,$C7,$FF,$92,$CF,$F2,$67,$C8,$5C,$3F,$CC
	.dc.b	$31,$67,$FC,$AB,$37,$1F,$F2,$74,$DA,$FF,$30,$CD,$AF,$F0,$A4,$DB
	.dc.b	$F0,$4D,$9F,$0D,$76,$3F,$CD,$F1,$9A,$D3,$BF,$90,$FF,$30,$78,$7F
	.dc.b	$E6,$16,$D7,$F9,$D9,$B7,$FE,$6A,$6E,$3F,$E6,$36,$CF,$F9,$D9,$B3
	.dc.b	$FE,$29,$6E,$3F,$15,$B0,$69,$27,$FD,$0B,$7E,$36,$9F,$D9,$3D,$2F
	.dc.b	$F2,$ED,$7F,$97,$FD,$DF,$F0,$7C,$63,$1F,$F5,$7C,$63,$C9,$FF,$35
	.dc.b	$B6,$FF,$D7,$0D,$BE,$51,$D6,$76,$D3,$FC,$01,$7E,$09,$FB,$C7,$DF
	.dc.b	$70,$E1,$AE,$67,$FD,$A1,$18,$83,$49,$1A,$24,$F8,$1B,$3F,$CF,$67
	.dc.b	$DD,$3F,$8F,$F6,$9E,$4B,$DE,$F6,$64,$AF,$26,$38,$78,$56,$DF,$76
	.dc.b	$3D,$67,$4F,$F9,$BE,$EE,$BF,$9D,$F8,$FF,$F5,$31,$FE,$A7,$F9,$0B
	.dc.b	$87,$FA,$F2,$2D,$FF,$AC,$26,$E5,$FE,$A9,$9B,$5F,$EB,$C9,$B5,$FE
	.dc.b	$6F,$9B,$7F,$AB,$9B,$29,$B1,$86,$D7,$84,$FF,$66,$CF,$FA,$EF,$3F
	.dc.b	$F5,$E3,$63,$27,$FB,$4A,$32,$D3,$31,$FE,$C8,$8C,$B6,$33,$87,$FB
	.dc.b	$0E,$72,$D8,$C9,$FE,$D8,$8C,$B4,$CC,$FF,$A3,$63,$2D,$8F,$CF,$63
	.dc.b	$FE,$C2,$F7,$FF,$3A,$D8,$C9,$FE,$48,$AF,$C0,$CE,$A3,$1A,$23,$17
	.dc.b	$FE,$C5,$6E,$1F,$85,$8F,$F5,$64,$69,$FC,$E9,$F9,$6C,$3F,$32,$4B
	.dc.b	$FD,$1B,$63,$FC,$DF,$AF,$E2,$DB,$19,$79,$C6,$5B,$19,$7F,$50,$F9
	.dc.b	$27,$F2,$BF,$01,$67,$F5,$D3,$72,$FC,$FC,$E6,$7C,$D5,$F1,$1F,$F7
	.dc.b	$14,$71,$87,$E0,$FF,$E6,$57,$F9,$B1,$8F,$23,$E1,$67,$F0,$E3,$7F
	.dc.b	$23,$87,$75,$C7,$0F,$D6,$36,$39,$1F,$E7,$A3,$89,$5D,$12,$FA,$4F
	.dc.b	$2A,$B5,$CA,$3F,$C7,$AD,$BF,$25,$F2,$AD,$BC,$E7,$B5,$E2,$72,$C8
	.dc.b	$C9,$E3,$8C,$5F,$6A,$7C,$87,$4F,$91,$EA,$DB,$F5,$36,$B8,$37,$1F
	.dc.b	$2B,$67,$E3,$B6,$7E,$3B,$63,$E9,$B6,$FC,$31,$F1,$EF,$C2,$D3,$C2
	.dc.b	$5A,$F0,$B6,$FF,$35,$1F,$CB,$3D,$CF,$9B,$A7,$F2,$4F,$82,$FE,$8B
	.dc.b	$18,$E1,$46,$30,$FD,$3B,$6F,$FC,$DB,$7C,$67,$FA,$85,$F1,$AD,$61
	.dc.b	$FA,$9A,$D5,$FF,$35,$AF,$E7,$E4,$BF,$E9,$2D,$8F,$D3,$B5,$FD,$2B
	.dc.b	$5F,$E9,$4D,$7F,$A7,$3C,$6B,$DF,$C0,$8F,$67,$C1,$79,$2F,$89,$E8
	.dc.b	$3F,$81,$B9,$BE,$71,$97,$DF,$92,$75,$3F,$A8,$F9,$0F,$C7,$9F,$90
	.dc.b	$F9,$1B,$1F,$9D,$8B,$7F,$8B,$9B,$97,$E6,$26,$D7,$EA,$A6,$D7,$EA
	.dc.b	$A6,$CF,$ED,$A6,$EC,$9B,$18,$6C,$F8,$4F,$D7,$7C,$87,$E9,$3E,$47
	.dc.b	$F4,$AD,$8C,$9F,$B3,$8C,$B4,$CE,$1F,$A3,$8C,$B6,$33,$8F,$EC,$A7
	.dc.b	$2D,$8C,$9F,$C1,$A3,$2D,$33,$3F,$C2,$23,$2D,$8F,$F7,$FF,$E1,$2F
	.dc.b	$77,$F2,$DE,$5F,$F0,$6A,$4E,$6B,$F9,$EC,$E6,$31,$3F,$6A,$31,$5C
	.dc.b	$AB,$AB,$67,$FD,$8D,$5C,$30,$FC,$D6,$DF,$EB,$7B,$D9,$FD,$0A,$BF
	.dc.b	$CB,$95,$D9,$FE,$8A,$D7,$FC,$46,$D7,$77,$FB,$47,$2D,$73,$3D,$5F
	.dc.b	$6C,$65,$B1,$99,$E9,$18,$3C,$8E,$0F,$54,$76,$3F,$B0,$FC,$3F,$C2
	.dc.b	$7D,$73,$F0,$2B,$F8,$14,$E6,$31,$3A,$C3,$E1,$CE,$5F,$E6,$3F,$25
	.dc.b	$FC,$C0,$FE,$63,$E6,$DA,$FE,$7D,$16,$FF,$96,$4D,$AE,$35,$A3,$FA
	.dc.b	$24,$65,$FE,$9B,$56,$BF,$A6,$CD,$9F,$EB,$33,$78,$3F,$DD,$35,$FC
	.dc.b	$E6,$30,$79,$27,$B4,$7F,$3D,$F9,$1F,$E8,$FF,$25,$FD,$21,$B3,$FD
	.dc.b	$55,$B3,$B8,$FE,$B1,$3B,$69,$B2,$C6,$CF,$E9,$51,$B6,$C6,$DA,$6F
	.dc.b	$1F,$EB,$73,$B6,$CE,$DB,$3F,$96,$F5,$7F,$9C,$72,$FE,$7D,$B5,$FA
	.dc.b	$86,$CE,$E3,$FC,$43,$3B,$69,$B2,$C6,$CF,$C5,$CF,$E4,$76,$34,$93
	.dc.b	$F9,$36,$CF,$08,$C1,$1C,$10,$FD,$3C,$78,$DF,$E5,$FF,$11,$70,$FE
	.dc.b	$B3,$7E,$0B,$F3,$56,$9B,$3A,$BE,$11,$FE,$7A,$BB,$5F,$E1,$4F,$91
	.dc.b	$FF,$02,$9F,$E0,$6F,$92,$B5,$FE,$13,$8B,$3F,$E0,$19,$B8,$7F,$83
	.dc.b	$A6,$D7,$F8,$82,$6D,$75,$8E,$47,$F8,$76,$32,$FF,$8B,$A6,$EC,$9B
	.dc.b	$18,$73,$E6,$FF,$86,$BE,$4B,$FC,$36,$DB,$FF,$14,$37,$0F,$F0,$B3
	.dc.b	$71,$FF,$11,$B6,$7F,$C6,$AD,$9F,$F1,$AB,$63,$FC,$82,$DC,$4D,$24
	.dc.b	$7F,$8D,$EF,$C6,$D3,$F9,$AF,$9B,$FE,$09,$F8,$AF,$E6,$0D,$8C,$9F
	.dc.b	$E7,$08,$CB,$4C,$E1,$FD,$BA,$32,$D8,$CB,$FC,$B2,$7F,$24,$48,$FE
	.dc.b	$44,$D8,$FE,$42,$D7,$84,$68,$8E,$9B,$8F,$E3,$A7,$FC,$45,$F1,$7E
	.dc.b	$28,$C4,$F9,$9F,$E8,$71,$98,$FE,$7B,$36,$9D,$FF,$ED,$0A,$C5,$7F
	.dc.b	$9B,$BE,$47,$FC,$AA,$7F,$95,$BD,$CB,$1F,$E6,$38,$B7,$FE,$51,$9B
	.dc.b	$8F,$F9,$7A,$6D,$7F,$9C,$26,$D7,$F9,$C2,$6C,$FF,$9F,$26,$EA,$9B
	.dc.b	$18,$6C,$F8,$4F,$F3,$A7,$C7,$FF,$9A,$BC,$2F,$F3,$5B,$63,$27,$F9
	.dc.b	$E6,$32,$D3,$38,$7F,$9A,$63,$2D,$8C,$E3,$FE,$78,$9C,$B6,$32,$7F
	.dc.b	$A3,$23,$2D,$33,$3F,$E8,$E8,$CB,$60,$A7,$FC,$C3,$F8,$53,$C7,$FF
	.dc.b	$46,$FE,$2C,$FC,$1B,$FD,$31,$6A,$D6,$1C,$B0,$FF,$61,$5E,$FC,$AF
	.dc.b	$F0,$85,$6A,$FB,$5F,$82,$B4,$F1,$3C,$71,$D7,$1F,$0E,$3E,$0B,$E5
	.dc.b	$8F,$68,$C4,$EB,$1E,$EF,$F8,$06,$BC,$17,$FE,$D2,$6B,$D9,$AF,$C2
	.dc.b	$6B,$E1,$C3,$FC,$19,$18,$24,$F5,$24,$69,$B3,$EE,$3A,$3F,$E8,$7F
	.dc.b	$8D,$F2,$CF,$BE,$72,$D7,$C0,$7F,$9A,$A3,$6F,$33,$38,$FD,$6A,$E9
	.dc.b	$5C,$6F,$E2,$BF,$ED,$4F,$E0,$17,$C8,$7F,$AF,$BE,$47,$FD,$8B,$18
	.dc.b	$9F,$F6,$64,$62,$75,$1F,$ED,$19,$D6,$3F,$EB,$DA,$D4,$F0,$35,$5C
	.dc.b	$27,$FD,$9F,$5C,$31,$FF,$67,$DF,$09,$E5,$7C,$0E,$4F,$0B,$E5,$5F
	.dc.b	$EE,$9B,$E5,$5F,$EE,$8B,$E5,$3D,$23,$CD,$F1,$FE,$C3,$DB,$EB,$58
	.dc.b	$C9,$31,$F2,$C6,$5F,$3C,$DA,$F0,$4D,$AF,$95,$36,$90,$C1,$59,$8C
	.dc.b	$55,$9C,$C7,$86,$FD,$93,$9A,$E9,$7F,$AB,$F2,$BF,$D9,$97,$FA,$86
	.dc.b	$BF,$75,$B1,$F8,$3C,$D1,$8C,$3F,$72,$DA,$FC,$F3,$6B,$F3,$CF,$E9
	.dc.b	$49,$1F,$9E,$6C,$7F,$9D,$EB,$AC,$E4,$91,$99,$DB,$FA,$D6,$BE,$08
	.dc.b	$F4,$1F,$9F,$E7,$7E,$B8,$F9,$1F,$74,$EA,$BD,$97,$FE,$F0,$AD,$4E
	.dc.b	$71,$F6,$B4,$FE,$33,$87,$4C,$39,$E1,$F9,$2F,$8B,$FC,$FF,$C6,$7E
	.dc.b	$81,$B7,$F9,$D6,$CE,$E3,$F4,$13,$B6,$9B,$2C,$6C,$FC,$FC,$6D,$B1
	.dc.b	$B6,$9B,$C3,$F3,$93,$B6,$CE,$DB,$3F,$AA,$F7,$FF,$6E,$7E,$E3,$E0
	.dc.b	$2D,$7E,$D6,$2C,$FE,$C2,$6D,$7E,$A6,$6D,$7F,$03,$9B,$5F,$9A,$E0
	.dc.b	$56,$2B,$F3,$95,$67,$31,$DC,$FE,$09,$19,$9D,$3D,$2B,$FA,$27,$81
	.dc.b	$F6,$7C,$DF,$E8,$2D,$BF,$E2,$CD,$9F,$CD,$37,$1F,$EA,$2D,$9F,$E6
	.dc.b	$CD,$9F,$E6,$CF,$F3,$D2,$47,$F3,$66,$C7,$E6,$5A,$FF,$41,$6B,$D2
	.dc.b	$38,$11,$E0,$F9,$8F,$CC,$6C,$F9,$0F,$5C,$66,$7D,$2E,$A3,$31,$82
	.dc.b	$5F,$F0,$87,$FC,$3F,$5C,$EB,$94,$FF,$1F,$F6,$7F,$9A,$FB,$DF,$CD
	.dc.b	$9B,$3F,$CC,$DB,$7F,$CB,$9B,$5C,$67,$F9,$C5,$71,$69,$C4,$B5,$C5
	.dc.b	$B3,$C6,$3F,$9C,$CF,$16,$CF,$16,$9C,$4B,$3F,$CD,$FD,$AF,$E9,$A7
	.dc.b	$F4,$EF,$6E,$D7,$F4,$B8,$B5,$FD,$1A,$6D,$71,$AD,$1F,$D4,$63,$2F
	.dc.b	$F5,$BA,$B5,$F9,$99,$B3,$FC,$CE,$6E,$E9,$B1,$87,$8D,$3F,$0E,$FE
	.dc.b	$9A,$FD,$4F,$91,$FA,$D8,$FD,$78,$D7,$5F,$F1,$39,$CE,$5F,$E1,$6B
	.dc.b	$E3,$CD,$FC,$72,$73,$7F,$E2,$C9,$CE,$C9,$CC,$F4,$BF,$C6,$C6,$2B
	.dc.b	$6D,$C3,$C5,$5D,$DA,$F5,$6B,$CC,$FF,$14,$99,$AA,$D2,$D7,$F7,$C7
	.dc.b	$6D,$7B,$47,$07,$A1,$A9,$EF,$3B,$8E,$98,$FC,$78,$F7,$FC,$87,$A6
	.dc.b	$FF,$C3,$91,$F3,$AB,$CD,$5F,$A3,$CB,$FA,$1D,$EE,$7F,$C3,$D5,$B3
	.dc.b	$9D,$F2,$C3,$FC,$33,$ED,$FF,$86,$BD,$CB,$3F,$E1,$76,$DF,$F8,$49
	.dc.b	$B9,$7F,$82,$DB,$3F,$E2,$56,$CF,$F2,$36,$C7,$F7,$D6,$EE,$69,$23
	.dc.b	$FC,$4D,$7E,$36,$9E,$33,$FC,$55,$EB,$FF,$8C,$0F,$F1,$8F,$B3,$63
	.dc.b	$24,$C3,$FC,$53,$19,$7F,$C5,$73,$71,$FF,$12,$4D,$AF,$F1,$F4,$DA
	.dc.b	$FE,$79,$36,$7F,$C2,$D3,$71,$4D,$8C,$3C,$6B,$F4,$9C,$5F,$DC,$7E
	.dc.b	$47,$F2,$8D,$C3,$ED,$B6,$7E,$BB,$71,$FC,$13,$67,$F4,$CD,$9F,$D2
	.dc.b	$3F,$8A,$24,$7E,$3F,$54,$62,$7F,$2E,$D7,$F2,$ED,$7A,$CF,$18,$C4
	.dc.b	$76,$8F,$39,$FC,$EB,$E4,$7D,$6E,$23,$D2,$E4,$C5,$FA,$9A,$7F,$83
	.dc.b	$B8,$95,$CA,$BF,$CB,$FE,$9F,$F9,$B3,$D6,$FF,$36,$B6,$FF,$CD,$4D
	.dc.b	$9D,$C7,$F9,$B6,$76,$D3,$65,$8D,$9F,$E6,$C8,$DB,$63,$6D,$37,$87
	.dc.b	$F9,$A6,$76,$D9,$DB,$67,$FC,$EF,$E7,$FF,$A1,$0F,$F4,$2F,$A3,$6B
	.dc.b	$FD,$05,$16,$7F,$CF,$B3,$6B,$FC,$ED,$36,$BF,$D2,$13,$6B,$FC,$1D
	.dc.b	$36,$7F,$CD,$13,$78,$13,$63,$0D,$9F,$09,$FC,$2B,$83,$FC,$89,$F2
	.dc.b	$3F,$C9,$DB,$19,$3F,$BB,$C6,$5A,$67,$0F,$F5,$A7,$37,$F9,$4A,$75
	.dc.b	$18,$D9,$18,$8F,$E4,$B7,$FC,$80,$97,$F9,$76,$C7,$F2,$C6,$BC,$23
	.dc.b	$44,$74,$DC,$FF,$94,$4F,$F9,$7F,$D6,$F1,$C6,$27,$D0,$FF,$27,$C1
	.dc.b	$3A,$9F,$EA,$55,$6F,$FC,$1B,$36,$7F,$D2,$C7,$FB,$03,$CA,$FF,$60
	.dc.b	$B4,$FF,$5F,$16,$7F,$D7,$2D,$C3,$FD,$64,$D3,$FD,$98,$58,$FF,$25
	.dc.b	$34,$EE,$7F,$B2,$67,$BB,$4E,$E5,$8E,$ED,$9E,$ED,$11,$FE,$B0,$FE
	.dc.b	$01,$78,$1F,$DB,$A5,$18,$7F,$DD,$86,$3C,$5D,$46,$6B,$FD,$E9,$3C
	.dc.b	$2B,$26,$AA,$C7,$2A,$E1,$7F,$E3,$98,$E0,$66,$77,$7D,$2B,$73,$E8
	.dc.b	$37,$19,$9D,$D7,$3A,$DC,$FB,$6F,$C5,$7E,$17,$6F,$39,$DB,$ED,$34
	.dc.b	$7C,$6A,$FF,$0C,$7C,$1F,$FA,$A2,$FF,$39,$F2,$5F,$E1,$08,$C4,$FE
	.dc.b	$2E,$31,$8F,$FA,$BD,$B7,$FE,$60,$6D,$F2,$8E,$C7,$D4,$57,$D2,$3F
	.dc.b	$93,$4F,$09,$FF,$09,$CF,$8D,$B5,$FE,$18,$8C,$47,$36,$BF,$80,$6B
	.dc.b	$DA,$38,$12,$BF,$C2,$D3,$A2,$46,$65,$7F,$12,$3F,$D9,$9F,$1B,$EC
	.dc.b	$8C,$5F,$B2,$BC,$15,$A8,$C4,$78,$5B,$7E,$F8,$DC,$7F,$A1,$AF,$A3
	.dc.b	$F5,$FB,$FF,$23,$E1,$FE,$49,$B1,$F8,$D6,$D7,$E2,$5B,$8F,$E1,$5B
	.dc.b	$3C,$CF,$CA,$57,$36,$9C,$CB,$5C,$DB,$5C,$DA,$73,$8F,$CC,$4F,$36
	.dc.b	$C7,$E8,$7A,$FF,$52,$7E,$AB,$C0,$B1,$92,$4F,$E8,$63,$2F,$E8,$E6
	.dc.b	$DF,$E7,$E6,$D7,$E6,$E6,$D7,$EF,$66,$CF,$EF,$26,$EA,$9B,$18,$78
	.dc.b	$D7,$E4,$B8,$FF,$CE,$3C,$8E,$47,$C6,$FE,$86,$73,$18,$CB,$ED,$C6
	.dc.b	$27,$51,$FE,$C0,$6D,$FF,$45,$6E,$7F,$CC,$DC,$C7,$E7,$AB,$F3,$84
	.dc.b	$AF,$CA,$B6,$3F,$CE,$7B,$FF,$4E,$9C,$92,$33,$3B,$7A,$46,$6F,$F7
	.dc.b	$B3,$EA,$F8,$CF,$1C,$FE,$A6,$2D,$7F,$B3,$A6,$BF,$16,$7D,$E7,$FB
	.dc.b	$72,$36,$FF,$2A,$9F,$E4,$FD,$9F,$CA,$1B,$1F,$C8,$DB,$5F,$C7,$DB
	.dc.b	$7F,$C6,$DB,$3F,$C5,$DB,$3F,$CE,$5B,$1D,$CF,$E5,$D5,$DD,$A7,$72
	.dc.b	$CF,$76,$D7,$76,$9F,$C9,$78,$FF,$9F,$9F,$D0,$39,$AC,$7F,$3B,$8B
	.dc.b	$5F,$CD,$E6,$D7,$F3,$39,$B5,$FC,$BE,$6D,$7F,$51,$9B,$3F,$D4,$26
	.dc.b	$F0,$26,$C6,$1B,$3E,$13,$F8,$4F,$C6,$7E,$4B,$E3,$7F,$9A,$36,$32
	.dc.b	$7F,$13,$E0,$FC,$D4,$EA,$31,$C0,$8C,$65,$FE,$19,$6E,$7F,$E0,$77
	.dc.b	$31,$FC,$CB,$0F,$E6,$04,$BF,$E5,$2D,$8F,$CA,$B5,$E1,$1A,$23,$A6
	.dc.b	$D7,$F5,$09,$FD,$69,$E2,$8F,$E7,$8D,$96,$1F,$B5,$8C,$47,$F4,$A7
	.dc.b	$51,$99,$F8,$73,$FD,$B7,$83,$FB,$DF,$2F,$F7,$C6,$CF,$F7,$76,$CE
	.dc.b	$E3,$FB,$D4,$ED,$A6,$CB,$1B,$3F,$BD,$46,$DB,$1B,$69,$BC,$3F,$BD
	.dc.b	$4E,$DB,$3B,$6C,$FF,$81,$71,$FF,$09,$9F,$E1,$4E,$0B,$5F,$E0,$F8
	.dc.b	$B3,$FE,$0B,$9B,$5F,$E0,$79,$B5,$FE,$03,$9B,$5F,$E2,$59,$B3,$FE
	.dc.b	$24,$9B,$B7,$E1,$E3,$B1,$FE,$1D,$8C,$CE,$9E,$75,$FD,$F0,$FE,$4F
	.dc.b	$ED,$FF,$82,$5B,$5F,$D8,$9A,$7F,$6E,$7F,$BD,$38,$C3,$FC,$26,$D8
	.dc.b	$C9,$F9,$A9,$E8,$FF,$7A,$8F,$F0,$21,$1C,$11,$FE,$52,$64,$AE,$5F
	.dc.b	$F0,$79,$92,$BE,$73,$F5,$3F,$19,$E0,$8F,$41,$FE,$0F,$31,$5F,$E6
	.dc.b	$F8,$B3,$E8,$9D,$9F,$15,$E6,$75,$7F,$C8,$55,$FE,$50,$D7,$FC,$A2
	.dc.b	$D9,$FF,$26,$36,$77,$1F,$E4,$F9,$DB,$4D,$96,$36,$7F,$93,$E3,$6D
	.dc.b	$8D,$B4,$DD,$FF,$93,$E7,$6D,$9D,$B6,$7F,$CB,$5F,$1B,$FE,$68,$3F
	.dc.b	$CD,$38,$5A,$FF,$32,$45,$9F,$F3,$04,$DA,$FF,$2E,$4D,$AF,$F2,$C4
	.dc.b	$DA,$FF,$3D,$4D,$9F,$F3,$CC,$DC,$D3,$63,$0D,$9F,$09,$FE,$34,$FC
	.dc.b	$41,$F1,$1F,$DF,$25,$39,$8C,$34,$FE,$53,$1A,$C3,$ED,$D7,$29,$C9
	.dc.b	$B8,$FF,$62,$4E,$F1,$E1,$0A,$E0,$EA,$B8,$3F,$E6,$53,$57,$C1,$FF
	.dc.b	$34,$4F,$07,$4D,$73,$1B,$7F,$D4,$31,$96,$C6,$5E,$B3,$DD,$CB,$5C
	.dc.b	$99,$9E,$4E,$0F,$0C,$FB,$67,$C2,$7F,$4E,$F8,$DC,$4F,$43,$E3,$CE
	.dc.b	$1E,$B7,$FE,$D5,$9C,$5F,$C4,$8F,$8E,$DB,$EB,$5F,$EE,$FA,$DD,$FF
	.dc.b	$AD,$B9,$FF,$D6,$F1,$6B,$FD,$64,$DA,$FF,$57,$37,$0F,$F5,$53,$67
	.dc.b	$FD,$4C,$D9,$FF,$64,$36,$3F,$D8,$ED,$CC,$D2,$47,$FB,$06,$FC,$6D
	.dc.b	$3C,$67,$FB,$16,$BF,$D9,$A7,$FB,$3B,$8A,$C6,$49,$1F,$EC,$58,$CB
	.dc.b	$FE,$C7,$2E,$1F,$EC,$38,$B3,$FE,$BF,$9A,$7F,$B8,$62,$C7,$FB,$82
	.dc.b	$2E,$1F,$DB,$23,$BD,$7F,$B3,$67,$35,$C9,$FC,$C7,$C1,$FF,$98,$B0
	.dc.b	$FA,$ED,$BF,$D6,$36,$7F,$4A,$DB,$FB,$6D,$9F,$CD,$B6,$7F,$2E,$FE
	.dc.b	$57,$33,$49,$3F,$95,$8C,$B6,$33,$3F,$97,$6B,$FA,$66,$BE,$08,$E0
	.dc.b	$7A,$0F,$F1,$2E,$1E,$C7,$11,$EA,$72,$62,$FD,$6D,$11,$FE,$16,$E4
	.dc.b	$F9,$8F,$13,$FA,$3C,$3B,$5E,$D7,$E9,$A3,$1E,$D5,$9C,$C7,$FB,$34
	.dc.b	$C9,$5C,$D7,$45,$3C,$6B,$93,$6B,$83,$6B,$83,$70,$FE,$39,$3D,$A7
	.dc.b	$A5,$72,$8F,$03,$63,$31,$D0,$C9,$5C,$9D,$9F,$47,$C6,$FB,$4F,$77
	.dc.b	$C6,$D8,$C9,$27,$D9,$19,$7D,$D3,$6F,$CF,$8A,$B1,$7F,$48,$F9,$D5
	.dc.b	$8B,$FE,$7B,$5B,$9E,$D7,$CA,$B5,$1F,$56,$ED,$26,$C7,$3B,$FE,$75
	.dc.b	$EE,$7F,$B1,$1F,$E3,$2D,$BF,$E9,$CD,$8F,$E8,$ED,$C3,$F8,$EB,$67
	.dc.b	$F9,$EB,$67,$F9,$CB,$FA,$32,$47,$E8,$5B,$1F,$CD,$5A,$FF,$44,$6B
	.dc.b	$D2,$38,$11,$E0,$F9,$8F,$F4,$47,$B1,$F8,$A7,$11,$E8,$6D,$7F,$61
	.dc.b	$D5,$5B,$AD,$12,$7D,$B5,$D3,$4F,$8A,$FE,$67,$E3,$3F,$3D,$F1,$BF
	.dc.b	$9F,$6C,$64,$FC,$F4,$65,$A6,$6F,$F4,$11,$96,$C6,$6F,$F0,$B3,$96
	.dc.b	$C6,$4F,$C3,$6A,$9D,$46,$1B,$5F,$88,$6C,$7E,$E5,$FD,$B7,$B7,$23
	.dc.b	$F6,$9C,$D6,$BF,$69,$37,$0F,$CB,$CD,$AF,$0C,$DA,$FC,$AC,$D9,$FE
	.dc.b	$0D,$37,$0F,$EC,$55,$F9,$A8,$C1,$FC,$1E,$AD,$7F,$05,$8C,$34,$FE
	.dc.b	$2B,$7F,$E1,$2F,$C4,$5E,$DF,$F7,$6D,$BE,$DF,$17,$F7,$59,$CD,$FE
	.dc.b	$12,$31,$5B,$DF,$F8,$8E,$1C,$E3,$13,$AA,$E0,$78,$0F,$E4,$B9,$B0
	.dc.b	$EA,$FF,$9A,$B1,$F0,$CE,$E3,$F9,$5C,$E9,$FF,$12,$35,$E6,$D7,$77
	.dc.b	$FE,$1D,$8C,$47,$78,$C1,$27,$FC,$3D,$3A,$24,$66,$7F,$C4,$91,$E6
	.dc.b	$C3,$DD,$5F,$A6,$E1,$F7,$57,$F9,$F2,$B5,$3E,$4B,$F8,$D5,$AB,$E3
	.dc.b	$7B,$9C,$D7,$BE,$3C,$CF,$F3,$DF,$8C,$FE,$85,$F1,$BF,$D0,$DB,$19
	.dc.b	$3F,$A1,$46,$5A,$66,$7F,$A2,$46,$5B,$19,$BF,$E5,$33,$96,$C6,$4F
	.dc.b	$D2,$C6,$5A,$66,$7F,$95,$C6,$5B,$1F,$D5,$7E,$37,$FA,$D1,$FD,$6F
	.dc.b	$E3,$6C,$64,$93,$FD,$66,$32,$FF,$5B,$9B,$7F,$CF,$A6,$D7,$EE,$26
	.dc.b	$D7,$F3,$C9,$B3,$FD,$B6,$6F,$0A,$6C,$61,$E3,$5F,$93,$F5,$FF,$39
	.dc.b	$CF,$FA,$16,$E1,$F5,$76,$46,$2B,$F0,$0D,$CB,$F5,$8D,$AF,$CA,$B6
	.dc.b	$7F,$36,$FE,$54,$91,$F9,$26,$C7,$E4,$1A,$FE,$81,$AF,$48,$E0,$47
	.dc.b	$83,$E7,$8F,$DB,$72,$7A,$E3,$04,$9F,$D8,$71,$7F,$21,$8C,$5F,$A9
	.dc.b	$A7,$F8,$73,$87,$F9,$FD,$F3,$BF,$F0,$87,$4F,$F8,$5F,$E3,$3F,$C3
	.dc.b	$0D,$BF,$F0,$B3,$67,$71,$FE,$19,$9D,$B4,$D9,$63,$67,$F8,$76,$36
	.dc.b	$D8,$DB,$4D,$E1,$FE,$04,$9D,$B6,$76,$D9,$FF,$14,$7C,$67,$F8,$BC
	.dc.b	$FF,$18,$7C,$05,$AF,$F1,$74,$59,$FF,$17,$4D,$AF,$F0,$8C,$DA,$FE
	.dc.b	$97,$36,$BF,$C1,$F3,$67,$FC,$81,$37,$85,$36,$30,$D9,$F0,$9F,$91
	.dc.b	$3F,$98,$74,$FF,$35,$69,$93,$FB,$CB,$96,$99,$BF,$BF,$19,$6C,$65
	.dc.b	$FE,$5A,$FF,$29,$21,$FC,$98,$AF,$F2,$46,$BC,$1D,$11,$D3,$70,$FE
	.dc.b	$A0,$FF,$49,$E9,$F1,$46,$23,$CC,$FF,$83,$76,$FE,$BF,$5F,$D1,$E7
	.dc.b	$67,$F4,$03,$91,$C0,$FE,$F7,$F1,$5F,$E6,$9F,$8B,$FF,$35,$36,$3F
	.dc.b	$CD,$0D,$9F,$F3,$43,$6F,$FC,$96,$D9,$E6,$7F,$9B,$EB,$9B,$4E,$65
	.dc.b	$AE,$6D,$AE,$6D,$39,$C7,$F9,$DE,$79,$B6,$3F,$CF,$9F,$17,$FE,$81
	.dc.b	$3F,$D0,$5F,$17,$63,$24,$9F,$F4,$04,$65,$FF,$41,$4D,$BF,$F3,$2C
	.dc.b	$DA,$FF,$0F,$CD,$AF,$F3,$1C,$D9,$FF,$4A,$4D,$D9,$36,$30,$F1,$A7
	.dc.b	$E2,$4F,$43,$FC,$09,$DF,$FE,$FD,$8C,$63,$FE,$4C,$8C,$4E,$A3,$FB
	.dc.b	$04,$EA,$BF,$89,$D6,$A7,$81,$AA,$E1,$2C,$93,$FE,$0C,$BE,$53,$FE
	.dc.b	$0E,$3C,$B5,$A7,$06,$26,$B4,$CC,$F2,$7A,$4F,$17,$A4,$65,$FF,$5F
	.dc.b	$B5,$E4,$6A,$38,$3C,$A7,$BC,$EC,$F8,$15,$D5,$B1,$FC,$BF,$57,$5F
	.dc.b	$C3,$BF,$81,$3C,$DA,$78,$B0,$FF,$1A,$4E,$F0,$E9,$16,$BA,$4D,$AF
	.dc.b	$93,$FC,$02,$F8,$3F,$F6,$93,$6B,$FD,$9D,$E1,$E2,$64,$7F,$B4,$E7
	.dc.b	$55,$FE,$D1,$AC,$BF,$ED,$E8,$CE,$3F,$EC,$C6,$D7,$F9,$C2,$B7,$39
	.dc.b	$BB,$39,$3B,$38,$BF,$3B,$FE,$E5,$AC,$4A,$FA,$D6,$A3,$D2,$72,$3D
	.dc.b	$1F,$17,$EE,$F8,$BF,$7B,$63,$27,$B6,$32,$D3,$33,$ED,$8C,$B6,$33
	.dc.b	$8F,$BE,$72,$D8,$C9,$FE,$D5,$B4,$EA,$30,$DA,$FA,$AD,$8F,$73,$FA
	.dc.b	$DF,$43,$FC,$E5,$E1,$59,$FF,$5C,$B6,$BF,$BE,$B7,$0F,$E5,$8D,$9F
	.dc.b	$C7,$36,$5D,$BF,$EB,$C8,$E5,$1F,$A0,$BF,$CD,$BF,$A0,$6B,$FC,$09
	.dc.b	$BC,$06,$92,$FF,$42,$D9,$FD,$7B,$5E,$D1,$A3,$F8,$2C,$7F,$A3,$FC
	.dc.b	$1F,$34,$F7,$9C,$9F,$ED,$88,$CD,$78,$63,$33,$FE,$1E,$9F,$74,$7E
	.dc.b	$1F,$87,$F3,$E7,$E8,$3E,$2A,$C7,$E7,$62,$D7,$E6,$E6,$E7,$F9,$B9
	.dc.b	$B5,$F6,$E6,$D7,$EB,$26,$CF,$E5,$66,$EB,$FE,$E8,$8E,$C7,$EA,$E3
	.dc.b	$33,$A7,$7E,$27,$ED,$3E,$2B,$F6,$AD,$AF,$D9,$37,$0F,$D8,$37,$1F
	.dc.b	$D8,$36,$7F,$20,$D9,$FE,$02,$D8,$FD,$53,$71,$FF,$4D,$CF,$F0,$6B
	.dc.b	$F0,$D7,$F0,$16,$C7,$EF,$DB,$7F,$BF,$7F,$A1,$6D,$FE,$C7,$F1,$6D
	.dc.b	$FF,$AF,$61,$9A,$73,$87,$F9,$51,$B8,$7F,$80,$DB,$7F,$C9,$5B,$3F
	.dc.b	$98,$AE,$53,$FD,$32,$FF,$26,$47,$F9,$8B,$67,$F9,$FB,$5F,$E7,$AD
	.dc.b	$7F,$A6,$35,$FE,$7E,$F1,$BF,$E7,$0F,$FB,$8F,$C0,$F1,$CF,$F6,$48
	.dc.b	$F2,$9E,$E8,$CE,$DF,$AE,$BF,$31,$F0,$F0,$4F,$B2,$3F,$8E,$7C,$57
	.dc.b	$F3,$43,$F9,$AF,$C2,$D8,$FE,$65,$17,$1F,$E6,$13,$72,$FE,$61,$36
	.dc.b	$BF,$88,$4D,$AF,$E8,$33,$67,$F9,$54,$DD,$53,$63,$0D,$9F,$09,$FD
	.dc.b	$0F,$E2,$BF,$A4,$FC,$57,$F4,$A6,$C6,$4F,$E9,$11,$96,$99,$C3,$FA
	.dc.b	$44,$65,$B1,$9C,$7F,$A5,$4E,$5B,$19,$3F,$96,$46,$5A,$66,$7F,$AF
	.dc.b	$C6,$5B,$1F,$EF,$1F,$81,$78,$7F,$E0,$7B,$FF,$29,$78,$D6,$97,$2F
	.dc.b	$71,$F6,$67,$53,$9B,$BC,$1F,$9E,$9C,$CA,$F8,$BF,$EB,$28,$CC,$63
	.dc.b	$0F,$0E,$1E,$5C,$39,$EE,$BD,$C7,$F3,$2B,$E7,$87,$F8,$85,$AF,$A0
	.dc.b	$F0,$46,$C9,$7F,$E2,$29,$DC,$FA,$63,$13,$A8,$FF,$14,$CE,$9E,$73
	.dc.b	$A3,$CC,$7E,$6F,$C6,$F2,$B8,$8F,$F0,$B9,$C0,$FF,$68,$C6,$9C,$CF
	.dc.b	$C1,$9F,$43,$63,$F7,$D3,$BC,$7F,$C1,$9F,$13,$FE,$18,$3F,$C3,$3F
	.dc.b	$15,$70,$FF,$0B,$45,$BF,$F0,$A4,$DA,$E3,$5A,$3F,$C3,$91,$97,$FB
	.dc.b	$B5,$5A,$FF,$13,$CD,$9F,$F0,$84,$DE,$14,$D8,$C3,$9F,$84,$FF,$16
	.dc.b	$F9,$7F,$E2,$E6,$DF,$F8,$AD,$B8,$7F,$8A,$5B,$8F,$F8,$A5,$B3,$FE
	.dc.b	$09,$6C,$FF,$8E,$9B,$1F,$E2,$36,$E2,$69,$23,$FC,$75,$7E,$36,$9F
	.dc.b	$A3,$E4,$FC,$7F,$95,$F9,$46,$C6,$4F,$E3,$11,$96,$99,$C3,$FC,$AB
	.dc.b	$C3,$FC,$FA,$75,$18,$E1,$FF,$70,$D6,$E3,$F3,$17,$F9,$B2,$47,$E5
	.dc.b	$5B,$1F,$92,$6B,$C2,$34,$47,$4D,$CF,$FC,$F5,$BF,$F2,$E8,$FE,$61
	.dc.b	$E4,$79,$27,$37,$67,$FC,$E5,$36,$FF,$80,$4D,$CB,$FA,$CC,$E6,$BF
	.dc.b	$CA,$5F,$13,$FE,$6A,$3F,$CD,$7F,$13,$63,$FC,$D1,$17,$0F,$F3,$34
	.dc.b	$DC,$7F,$CC,$D3,$6B,$FC,$8D,$36,$BF,$CE,$F3,$67,$FC,$C1,$37,$74
	.dc.b	$D8,$C3,$67,$C2,$7F,$9E,$7E,$1F,$FC,$FF,$F1,$1F,$E8,$06,$C6,$4F
	.dc.b	$F3,$F4,$65,$A6,$70,$FF,$3F,$46,$5B,$19,$C7,$FD,$01,$39,$6C,$64
	.dc.b	$FF,$31,$46,$5A,$66,$7F,$D2,$B1,$96,$C7,$F3,$7F,$2F,$F9,$1F,$9B
	.dc.b	$FC,$A1,$B1,$93,$FB,$64,$65,$A6,$67,$F2,$51,$96,$C6,$5F,$E5,$53
	.dc.b	$FC,$B4,$91,$FC,$9D,$B1,$FC,$91,$AF,$08,$D1,$1D,$36,$FF,$27,$3F
	.dc.b	$CF,$3C,$CF,$14,$62,$7C,$CF,$FB,$67,$6F,$F1,$3C,$EA,$7F,$92,$D5
	.dc.b	$BF,$F1,$14,$D9,$FF,$52,$1F,$EB,$C3,$FD,$7B,$C1,$5F,$F5,$D1,$67
	.dc.b	$FD,$71,$17,$0F,$F5,$C4,$D3,$FD,$31,$16,$7F,$D9,$31,$4F,$F5,$94
	.dc.b	$5D,$BF,$DE,$B1,$D8,$FF,$64,$46,$5D,$3C,$7F,$80,$5F,$13,$FD,$AA
	.dc.b	$50,$F3,$F0,$EA,$3F,$B9,$E3,$FE,$FD,$77,$3A,$AD,$B6,$36,$7F,$7A
	.dc.b	$36,$63,$13,$E0,$9E,$2E,$4C,$16,$FC,$8D,$8D,$CF,$FB,$D2,$36,$D9
	.dc.b	$DB,$67,$77,$E9,$9F,$64,$EE,$FD,$0D,$BE,$C7,$86,$3F,$C1,$7F,$8A
	.dc.b	$BC,$4F,$F0,$2E,$B8,$6D,$7B,$6D,$5B,$8F,$F2,$85,$6D,$A6,$E3,$53
	.dc.b	$FC,$DE,$74,$D9,$D5,$7F,$B0,$AB,$73,$FE,$0F,$3B,$C6,$49,$39,$8F
	.dc.b	$F1,$5C,$F0,$7F,$C2,$CD,$73,$3C,$1E,$11,$97,$AC,$F3,$69,$A9,$F3
	.dc.b	$E3,$FA,$9E,$67,$99,$8B,$EF,$56,$FF,$D1,$53,$5E,$77,$F8,$47,$C7
	.dc.b	$5F,$E7,$79,$E0,$FE,$5B,$E2,$BF,$2A,$7E,$5B,$E2,$AC,$7E,$1E,$2D
	.dc.b	$7E,$0E,$6D,$FE,$06,$6D,$7E,$57,$55,$62,$FF,$2B,$0A,$C5,$7E,$9A
	.dc.b	$AC,$E6,$3C,$07,$E5,$E3,$33,$A7,$7F,$15,$FA,$CF,$8A,$FD,$6B,$6B
	.dc.b	$F3,$CD,$C3,$F3,$4D,$C7,$F3,$0D,$9F,$43,$67,$CE,$D8,$FD,$DB,$71
	.dc.b	$34,$91,$FA,$DB,$F1,$B4,$FC,$A7,$81,$FE,$6C,$FD,$7D,$E9,$7E,$8A
	.dc.b	$30,$D9,$FF,$67,$4E,$9E,$26,$23,$58,$FF,$85,$A7,$4F,$F9,$5B,$81
	.dc.b	$7D,$2F,$8C,$FE,$4A,$3F,$40,$48,$FC,$94,$F0,$6C,$66,$78,$46,$08
	.dc.b	$E0,$87,$EA,$67,$93,$E3,$9F,$81,$E8,$FA,$34,$FC,$A4,$FE,$F2,$71
	.dc.b	$7F,$E9,$6E,$86,$2C,$33,$7F,$B0,$BB,$5A,$9C,$DF,$F3,$0F,$8B,$FE
	.dc.b	$5E,$7F,$30,$F5,$6C,$7F,$21,$8B,$87,$F1,$D9,$B9,$7F,$1B,$9B,$5F
	.dc.b	$9B,$9B,$5F,$9A,$9B,$3F,$CF,$A6,$EE,$9B,$18,$6C,$F8,$4F,$E6,$5F
	.dc.b	$15,$FD,$13,$E2,$FF,$A2,$B6,$32,$7F,$37,$8C,$B4,$CE,$1F,$CD,$63
	.dc.b	$2D,$8C,$E3,$FC,$D6,$72,$D8,$C9,$FB,$78,$CB,$4C,$CF,$EE,$23,$2D
	.dc.b	$8F,$E5,$DD,$5F,$EC,$6F,$CE,$DF,$07,$FC,$F2,$30,$D9,$FE,$AD,$B2
	.dc.b	$B7,$1C,$8C,$B8,$BF,$CD,$5F,$29,$D4,$7F,$9D,$27,$58,$FE,$2B,$87
	.dc.b	$FD,$A9,$5F,$9A,$BF,$E7,$84,$70,$4B,$FE,$67,$19,$6C,$65,$FC,$D4
	.dc.b	$65,$B1,$E2,$AF,$F0,$E7,$0B,$D8,$F7,$57,$A2,$B4,$7E,$6E,$74,$E2
	.dc.b	$3F,$A4,$DF,$0B,$E7,$76,$BF,$DC,$1F,$8B,$7F,$70,$BF,$EA,$8D,$73
	.dc.b	$DB,$FA,$A6,$CF,$F4,$B9,$D4,$7F,$C0,$06,$A6,$92,$79,$46,$DA,$ED
	.dc.b	$B1,$C3,$BA,$92,$BC,$0D,$AF,$EC,$D3,$FE,$04,$7D,$31,$96,$CF,$02
	.dc.b	$3C,$23,$BF,$4C,$9F,$63,$67,$AB,$5F,$8A,$D7,$E4,$B5,$F8,$4D,$7E
	.dc.b	$17,$47,$F6,$5C,$BF,$B3,$7C,$BB,$F0,$EF,$F2,$0D,$7F,$C0,$0C,$97
	.dc.b	$BC,$52,$BE,$11,$98,$FA,$13,$67,$33,$FE,$1B,$AE,$6F,$F8,$E6,$33
	.dc.b	$5D,$EB,$85,$F8,$B0,$FE,$4F,$87,$0A,$47,$A2,$75,$87,$E2,$2F,$EB
	.dc.b	$91,$FC,$23,$6B,$EA,$B5,$FC,$6B,$5E,$51,$C0,$8F,$07,$D5,$1E,$DF
	.dc.b	$C4,$3C,$97,$E6,$8F,$85,$5F,$3A,$B6,$7B,$EB,$6F,$68,$F8,$0E,$9F
	.dc.b	$6B,$F3,$B1,$4F,$09,$F9,$B3,$F2,$F9,$BF,$B6,$7B,$3E,$0C,$FF,$99
	.dc.b	$4F,$F6,$A8,$C4,$EA,$38,$19,$2D,$7F,$6B,$9D,$C6,$63,$F0,$53,$FA
	.dc.b	$22,$47,$EB,$DB,$1F,$A0,$6B,$FB,$86,$BF,$BC,$6B,$FB,$57,$8E,$5F
	.dc.b	$79,$FF,$03,$78,$3E,$29,$F0,$D7,$9A,$BF,$88,$CE,$5E,$11,$93,$F8
	.dc.b	$84,$65,$FE,$17,$F1,$FF,$E1,$1F,$0B,$E3,$36,$3F,$28,$DB,$FD,$03
	.dc.b	$6F,$F2,$2D,$9E,$67,$F8,$5A,$B9,$BF,$C4,$63,$F8,$39,$1C,$12,$BF
	.dc.b	$82,$C6,$5B,$19,$7F,$8F,$C6,$5B,$1E,$63,$FC,$4B,$ED,$78,$23,$CF
	.dc.b	$19,$7F,$C4,$91,$9B,$F4,$CE,$5F,$E7,$31,$98,$FE,$5F,$EE,$7F,$8D
	.dc.b	$7D,$9F,$C5,$36,$3F,$70,$DB,$FE,$08,$DB,$FD,$AB,$67,$99,$FE,$3A
	.dc.b	$AE,$6F,$F3,$58,$FE,$56,$47,$04,$AF,$E5,$31,$96,$C6,$5F,$E8,$71
	.dc.b	$96,$C7,$98,$FF,$24,$FB,$3E,$08,$F3,$C6,$5F,$F2,$44,$66,$FD,$33
	.dc.b	$97,$FA,$EC,$66,$3F,$A9,$FC,$67,$F9,$67,$D9,$FD,$73,$63,$F8,$E3
	.dc.b	$6F,$F9,$43,$6F,$F8,$CB,$67,$99,$FE,$5E,$AE,$6F,$F5,$98,$FE,$9E
	.dc.b	$47,$04,$AF,$E9,$B1,$96,$C6,$5F,$EC,$F1,$96,$C7,$98,$FF,$36,$FB
	.dc.b	$5E,$08,$F3,$C6,$5F,$F3,$64,$66,$FD,$33,$97,$FC,$03,$19,$8F,$EE
	.dc.b	$FD,$FF,$E7,$DF,$5F,$F8,$93,$63,$FA,$03,$6F,$FA,$63,$6F,$F9,$EB
	.dc.b	$67,$99,$FE,$82,$AE,$6F,$F7,$D8,$FE,$E6,$47,$04,$AF,$EE,$31,$96
	.dc.b	$C6,$5F,$F0,$5C,$65,$B1,$E6,$3F,$D2,$3E,$37,$82,$3C,$F1,$97,$FD
	.dc.b	$21,$19,$BF,$4C,$E5,$FF,$12,$C6,$63,$FB,$3C,$BF,$18,$7C,$47,$FA
	.dc.b	$8F,$D8,$FE,$8F,$18,$AF,$F5,$0C,$62,$75,$1F,$DC,$30,$FF,$44,$D6
	.dc.b	$EF,$FD,$45,$7B,$AE,$26,$CF,$0C,$EE,$BF,$C7,$33,$BA,$FF,$7E,$4F
	.dc.b	$37,$CE,$79,$EF,$FC,$47,$19,$8E,$71,$82,$5F,$FB,$EE,$75,$3F,$E0
	.dc.b	$66,$BC,$E3,$53,$C2,$3C,$2F,$8E,$FF,$C9,$BE,$CD,$AE,$E7,$AE,$BA
	.dc.b	$57,$FB,$2E,$75,$18,$9D,$3D,$CF,$F6,$34,$69,$FA,$5F,$0D,$FE,$D3
	.dc.b	$F5,$FE,$8B,$6B,$8E,$28,$C5,$7F,$74,$3D,$8E,$31,$FA,$6D,$AF,$9E
	.dc.b	$D9,$DC,$7C,$58,$F9,$64,$70,$48,$FA,$71,$96,$C6,$5F,$AB,$5E,$58
	.dc.b	$C4,$78,$1F,$4C,$FF,$BB,$FD,$7F,$6C,$60,$93,$FE,$E9,$8C,$BE,$D3
	.dc.b	$E2,$C6,$AF,$FC,$FF,$3A,$7F,$23,$F0,$9F,$EF,$BF,$5F,$F3,$2D,$AF
	.dc.b	$92,$DB,$E9,$8A,$31,$7C,$AB,$FD,$FF,$7C,$9A,$72,$3F,$19,$3F,$92
	.dc.b	$23,$82,$5F,$E6,$A3,$2D,$8C,$BF,$22,$32,$D8,$EC,$FD,$DF,$2B,$FD
	.dc.b	$E3,$EC,$78,$E3,$D3,$57,$0F,$F5,$0C,$DB,$FD,$6C,$DA,$5F,$29,$EB
	.dc.b	$7F,$B8,$F8,$2F,$F3,$34,$FF,$BF,$BC,$9F,$E0,$31,$8F,$2B,$FD,$77
	.dc.b	$87,$C8,$8C,$4E,$B0,$FF,$65,$D6,$B0,$E9,$18,$9D,$5F,$AE,$B5,$3F
	.dc.b	$C0,$AF,$F7,$64,$7F,$70,$D9,$FE,$06,$D7,$F1,$ED,$79,$47,$02,$3C
	.dc.b	$1F,$D4,$C7,$AD,$F1,$DF,$9A,$7D,$91,$EA,$32,$74,$72,$7A,$5C,$BF
	.dc.b	$C2,$FD,$BF,$93,$FA,$FB,$CB,$FE,$53,$18,$6D,$7F,$BC,$27,$4F,$EF
	.dc.b	$27,$4D,$35,$3C,$2B,$F0,$55,$C1,$FC,$CD,$7F,$22,$25,$7F,$1F,$6C
	.dc.b	$7F,$2A,$6B,$FB,$66,$BF,$CC,$9A,$FA,$E3,$91,$E3,$7C,$3E,$5F,$8F
	.dc.b	$0F,$E7,$F5,$6B,$FD,$63,$A3,$05,$E7,$0E,$87,$A2,$BE,$14,$66,$3F
	.dc.b	$A9,$7B,$BF,$2B,$CC,$FE,$A2,$DA,$FE,$42,$D9,$FD,$CB,$72,$FE,$A2
	.dc.b	$D9,$FE,$9A,$D9,$FE,$80,$FF,$48,$24,$7F,$45,$6C,$7F,$50,$6B,$FC
	.dc.b	$71,$AF,$48,$E0,$47,$83,$E6,$3E,$F7,$AF,$EA,$71,$1E,$87,$26,$2B
	.dc.b	$D2,$D3,$E5,$E2,$AE,$55,$FC,$63,$DB,$FC,$7F,$9B,$FD,$D9,$B7,$FD
	.dc.b	$19,$B3,$B8,$FC,$84,$ED,$A6,$CB,$1B,$3F,$B8,$CF,$F6,$E2,$47,$F6
	.dc.b	$B6,$C6,$A3,$04,$70,$43,$FA,$2C,$70,$C3,$FC,$0E,$FE,$A3,$D8,$F2
	.dc.b	$4F,$86,$73,$5F,$A7,$9C,$C6,$27,$53,$F1,$A7,$2F,$F8,$73,$DA,$7B
	.dc.b	$3F,$B4,$F5,$FF,$DC,$31,$89,$FD,$94,$62,$75,$1F,$DD,$A7,$57,$FB
	.dc.b	$0A,$D4,$F0,$35,$5C,$27,$FA,$B5,$70,$7F,$C2,$ED,$7C,$90,$9F,$F0
	.dc.b	$DC,$04,$ED,$C3,$5C,$CF,$17,$8C,$64,$91,$98,$F0,$B5,$F0,$CF,$81
	.dc.b	$E6,$D9,$F7,$E1,$C5,$B1,$F9,$2F,$37,$E0,$D7,$BA,$ED,$F8,$F8,$7F
	.dc.b	$80,$56,$2F,$BD,$ED,$FE,$41,$19,$8F,$6F,$9D,$FC,$A3,$D6,$FA,$AD
	.dc.b	$AF,$A4,$D9,$F9,$AD,$C3,$CC,$D9,$FC,$03,$67,$AC,$7B,$63,$E1,$46
	.dc.b	$0F,$92,$E1,$FA,$2D,$C8,$D2,$4F,$32,$4F,$26,$BF,$80,$8F,$5C,$7F
	.dc.b	$3F,$F5,$EC,$FA,$A3,$DA,$64,$AE,$49,$7F,$37,$85,$5C,$A3,$EB,$79
	.dc.b	$BF,$D4,$3C,$EF,$CE,$B7,$0F,$CD,$37,$0F,$CA,$B6,$B8,$CF,$F5,$3A
	.dc.b	$E2,$D3,$89,$F8,$59,$F8,$84,$70,$4A,$FC,$9C,$65,$B1,$97,$F4,$71
	.dc.b	$96,$C7,$67,$FA,$EF,$C9,$78,$9F,$3C,$66,$FE,$14,$E5,$F4,$CE,$8E
	.dc.b	$33,$B3,$F8,$17,$99,$FD,$D3,$E4,$BF,$80,$B6,$3F,$7C,$D9,$FD,$CB
	.dc.b	$6B,$F4,$4D,$9E,$67,$F7,$9A,$E6,$FE,$A6,$3F,$10,$47,$04,$AF,$DB
	.dc.b	$C6,$5B,$19,$7F,$83,$C6,$5B,$1E,$63,$FC,$11,$E7,$F8,$23,$CF,$19
	.dc.b	$7F,$C0,$B1,$9B,$F4,$CE,$5F,$C5,$46,$63,$F9,$37,$95,$FE,$18,$F9
	.dc.b	$1F,$E4,$AD,$8F,$E4,$4D,$BF,$E3,$AD,$BF,$E0,$CD,$9E,$67,$F8,$76
	.dc.b	$B9,$BF,$C3,$63,$F5,$84,$70,$4A,$FE,$37,$19,$6C,$65,$FE,$57,$19
	.dc.b	$6C,$79,$8F,$F1,$67,$C9,$78,$23,$CF,$19,$7F,$C5,$31,$9B,$F4,$CE
	.dc.b	$5F,$D7,$46,$63,$FA,$57,$95,$FC,$0C,$8F,$F2,$76,$D7,$F3,$26,$BF
	.dc.b	$D0,$1A,$F2,$8E,$04,$7F,$9C,$F9,$5F,$C3,$C9,$1F,$CC,$DB,$1A,$8C
	.dc.b	$11,$C1,$0F,$E9,$F1,$C1,$FE,$A3,$E4,$FF,$1B,$24,$FF,$3F,$6C,$7F
	.dc.b	$4A,$6B,$C2,$34,$47,$4D,$7F,$AE,$F9,$1F,$C9,$C9,$1F,$D2,$DB,$1F
	.dc.b	$D5,$9A,$F0,$8D,$11,$D3,$5F,$ED,$3E,$47,$F3,$32,$47,$F5,$76,$C7
	.dc.b	$F6,$26,$BC,$23,$44,$74,$D7,$FB,$AF,$8F,$FC,$FC,$91,$FD,$8D,$B1
	.dc.b	$FD,$B9,$AF,$08,$D1,$1D,$35,$FE,$F8,$7F,$7D,$FC,$6B,$FC,$F1,$A7
	.dc.b	$F3,$BF,$47,$FB,$D3,$FD,$DE,$33,$5F,$E4,$09,$CB,$CA,$33,$5F,$EB
	.dc.b	$AE,$75,$6A,$3B,$78,$FF,$E9,$28,$C4,$2B,$8F,$6F,$F9,$D6,$73,$18
	.dc.b	$BE,$78,$7F,$84,$F1,$F2,$63,$FD,$96,$BF,$D2,$18,$FB,$0E,$91,$92
	.dc.b	$57,$F8,$36,$77,$3E,$99,$CB,$FE,$20,$8C,$BE,$89,$E1,$3A,$8E,$D3
	.dc.b	$F0,$1E,$F3,$FE,$A9,$F9,$3F,$4D,$48,$F4,$1D,$CD,$18,$9E,$C2,$FD
	.dc.b	$B1,$98,$EA,$7F,$AB,$AB,$A3,$B8,$FA,$5E,$27,$FA,$FB,$E4,$FE,$4B
	.dc.b	$63,$96,$3F,$E7,$77,$FD,$7D,$3A,$69,$A9,$E1,$5D,$4F,$C3,$38,$C3
	.dc.b	$F0,$8F,$D1,$24,$FC,$D6,$CF,$CB,$6B,$F2,$1A,$FA,$23,$81,$1E,$0F
	.dc.b	$8C,$FF,$6A,$7A,$3E,$78,$F2,$1F,$89,$9C,$57,$F8,$4E,$6D,$75,$AD
	.dc.b	$E1,$F0,$5F,$F6,$BC,$E9,$FD,$17,$87,$FE,$E6,$DB,$F2,$2D,$BF,$92
	.dc.b	$D9,$E6,$DC,$3F,$4C,$D9,$FD,$23,$67,$AC,$7E,$46,$3F,$30,$47,$04
	.dc.b	$8F,$C9,$46,$5B,$19,$7F,$23,$19,$6C,$79,$8F,$F7,$D3,$EA,$70,$FA
	.dc.b	$1C,$98,$9F,$4B,$4F,$F5,$EC,$71,$3E,$4F,$77,$EE,$ED,$FA,$FB,$D1
	.dc.b	$FD,$BC,$E6,$31,$7F,$90,$8C,$4E,$A3,$F4,$15,$CA,$31,$7F,$8F,$8C
	.dc.b	$4F,$13,$77,$D2,$7C,$71,$FB,$BB,$FE,$04,$47,$04,$8F,$DE,$46,$5B
	.dc.b	$19,$7F,$77,$19,$6C,$78,$5F,$24,$BE,$47,$CF,$8F,$F2,$5B,$B7,$FE
	.dc.b	$A7,$C3,$E2,$D6,$30,$FE,$5D,$56,$73,$5D,$6F,$F9,$77,$77,$C9,$F9
	.dc.b	$1F,$E4,$2D,$9F,$DC,$B7,$0F,$CD,$B6,$B8,$CF,$C6,$AE,$2D,$38,$9F
	.dc.b	$C8,$67,$F9,$41,$1C,$12,$BF,$91,$46,$5B,$19,$7F,$90,$C6,$5B,$1D
	.dc.b	$9F,$8F,$E9,$F8,$9F,$3C,$66,$FF,$D6,$99,$AB,$51,$EA,$AD,$9C,$AB
	.dc.b	$81,$FD,$3F,$B3,$F1,$DF,$21,$FD,$19,$B1,$FC,$79,$B3,$FC,$01,$B5
	.dc.b	$FD,$6D,$B3,$CC,$FC,$6D,$73,$7F,$A3,$47,$F4,$D2,$38,$25,$7F,$47
	.dc.b	$8C,$B6,$32,$FF,$46,$8C,$B6,$3C,$C7,$E7,$3E,$2B,$C1,$1E,$78,$CB
	.dc.b	$F9,$08,$CD,$FA,$67,$2F,$CA,$8C,$C7,$F7,$3E,$4F,$D9,$7A,$9F,$DA
	.dc.b	$9B,$1F,$D0,$9B,$7F,$C9,$1B,$7F,$DF,$DB,$3C,$CF,$D8,$D7,$37,$FB
	.dc.b	$54,$7F,$71,$23,$82,$57,$F6,$B8,$CB,$63,$2F,$F6,$A8,$CB,$63,$CC
	.dc.b	$7F,$00,$F9,$0F,$04,$79,$E3,$2F,$ED,$23,$37,$E9,$9C,$BF,$8E,$8C
	.dc.b	$C7,$F8,$5E,$7F,$8B,$7C,$87,$F8,$39,$B1,$FD,$99,$A7,$F4,$82,$D7
	.dc.b	$F8,$8D,$B1,$CC,$FE,$2B,$3C,$DF,$F0,$74,$7F,$85,$48,$E0,$8F,$F8
	.dc.b	$3C,$C9,$5C,$BF,$E0,$E3,$25,$7C,$C7,$F2,$4F,$57,$C1,$1E,$78,$CB
	.dc.b	$FC,$62,$33,$5E,$99,$CB,$FB,$28,$CC,$7F,$91,$3F,$1C,$70,$FF,$3B
	.dc.b	$F9,$1F,$F1,$BC,$62,$7F,$99,$C6,$27,$51,$FE,$17,$9D,$61,$FC,$A2
	.dc.b	$B5,$3C,$0D,$57,$09,$FE,$F7,$5C,$1F,$F1,$D3,$5F,$24,$27,$FC,$7D
	.dc.b	$01,$3B,$70,$D7,$33,$C5,$E3,$19,$24,$66,$3C,$2D,$7C,$33,$E0,$79
	.dc.b	$B6,$7D,$F8,$71,$6C,$7F,$43,$E1,$7C,$8F,$C2,$BF,$7E,$17,$0F,$E4
	.dc.b	$73,$6F,$BD,$69,$FE,$C3,$19,$8F,$6E,$FF,$DB,$FD,$8F,$AA,$DA,$FA
	.dc.b	$4D,$9F,$9A,$DB,$F3,$36,$7F,$00,$D9,$EB,$1E,$D8,$F8,$51,$83,$E4
	.dc.b	$B8,$7E,$8B,$72,$34,$93,$CC,$93,$C9,$AF,$E0,$23,$D7,$1F,$E0,$0F
	.dc.b	$90,$B3,$EA,$8F,$69,$92,$B9,$25,$FF,$4E,$9E,$2F,$D5,$DB,$FC,$2D
	.dc.b	$F2,$1F,$9C,$6D,$FE,$65,$B7,$F9,$46,$D7,$19,$FF,$09,$D7,$16,$9C
	.dc.b	$4F,$C2,$4F,$C3,$23,$82,$57,$E4,$A3,$2D,$8C,$BF,$A2,$8C,$B6,$3B
	.dc.b	$3F,$E1,$4F,$63,$C4,$F9,$E3,$37,$FD,$BA,$72,$FA,$67,$47,$19,$D9
	.dc.b	$FC,$07,$4F,$F1,$D7,$C8,$7F,$00,$6C,$7E,$F5,$B3,$FB,$86,$D7,$E8
	.dc.b	$5B,$3C,$CF,$F1,$CD,$73,$7F,$51,$1F,$87,$23,$82,$57,$ED,$A3,$2D
	.dc.b	$8C,$BF,$C1,$A3,$2D,$8F,$31,$FE,$48,$F9,$0F,$04,$79,$E3,$2F,$F8
	.dc.b	$F2,$33,$7E,$99,$CB,$FE,$16,$8C,$C7,$F2,$5C,$BF,$CB,$DE,$CF,$F2
	.dc.b	$46,$C7,$F2,$16,$DF,$F1,$C6,$DF,$F0,$56,$CF,$33,$FC,$BB,$5C,$DF
	.dc.b	$E1,$91,$FA,$B2,$38,$25,$7F,$1A,$8C,$B6,$32,$FF,$2A,$8C,$B6,$3C
	.dc.b	$C7,$F9,$B3,$D5,$F0,$47,$9E,$32,$FF,$98,$23,$37,$E9,$9C,$BF,$E3
	.dc.b	$A8,$CC,$7F,$49,$C3,$FD,$05,$F1,$FF,$D2,$1B,$1F,$D1,$5B,$7F,$D0
	.dc.b	$1B,$7F,$CA,$5B,$3C,$CF,$F4,$0D,$73,$7F,$98,$47,$F0,$F2,$38,$25
	.dc.b	$7F,$3E,$8C,$B6,$32,$FF,$4E,$8C,$B6,$3C,$C7,$FA,$5B,$A9,$EC,$F8
	.dc.b	$67,$D1,$3A,$7F,$D3,$13,$A8,$C6,$1E,$B6,$9F,$E8,$58,$D4,$F2,$3F
	.dc.b	$B0,$E1,$FE,$AD,$87,$C7,$FF,$70,$8C,$36,$BC,$8D,$AF,$0B,$6F,$FA
	.dc.b	$F7,$7B,$0E,$77,$FD,$D7,$0E,$6D,$3F,$AD,$3F,$CE,$48,$E0,$98,$7F
	.dc.b	$68,$8C,$B6,$32,$FF,$78,$8C,$B6,$3B,$3F,$1B,$E3,$3C,$D1,$FE,$0F
	.dc.b	$9B,$2B,$F4,$56,$23,$FB,$EC,$EA,$73,$5D,$30,$FE,$F1,$F8,$EB,$0F
	.dc.b	$95,$ED,$7F,$7F,$9C,$C6,$2B,$FC,$11,$18,$7E,$F4,$ED,$FF,$06,$62
	.dc.b	$BE,$91,$A3,$89,$A7,$31,$F8,$29,$E4,$FF,$85,$63,$35,$F5,$32,$56
	.dc.b	$AB,$B6,$1F,$DA,$EF,$9C,$78,$8C,$E1,$FE,$11,$8C,$C7,$86,$78,$12
	.dc.b	$74,$D8,$FF,$13,$CF,$07,$C9,$19,$7B,$47,$9A,$7A,$1F,$A1,$F8,$9F
	.dc.b	$34,$66,$78,$CF,$C2,$69,$FA,$29,$DD,$49,$D4,$FA,$8D,$38,$AE,$91
	.dc.b	$F3,$FE,$27,$F4,$47,$80,$B3,$B6,$C3,$2F,$C0,$C6,$1B,$3F,$B9,$8E
	.dc.b	$4D,$39,$1F,$2D,$FA,$04,$70,$49,$F9,$31,$96,$C6,$5F,$93,$19,$6C
	.dc.b	$76,$7F,$82,$FB,$7E,$27,$D1,$1F,$89,$9C,$57,$E1,$6A,$CE,$4E,$51
	.dc.b	$96,$9F,$91,$F8,$8F,$E1,$1F,$13,$F5,$9B,$1F,$35,$B8,$78,$0F,$CA
	.dc.b	$38,$AF,$D4,$B6,$7A,$1F,$C8,$6B,$A3,$F9,$18,$FC,$B1,$1C,$12,$BF
	.dc.b	$1F,$19,$6C,$65,$FC,$7C,$65,$B1,$E7,$3F,$98,$47,$82,$3D,$07,$EC
	.dc.b	$DC,$57,$EA,$DB,$3E,$89,$DE,$29,$FE,$58,$6A,$3F,$69,$F1,$3F,$AF
	.dc.b	$25,$7E,$B5,$B1,$FA,$86,$BF,$A6,$6B,$FC,$05,$AF,$48,$FD,$B7,$C4
	.dc.b	$FE,$EC,$8E,$09,$1F,$B4,$8C,$B6,$32,$FE,$D2,$32,$D8,$FF,$50,$8F
	.dc.b	$90,$7C,$07,$F1,$08,$C7,$C0,$D9,$CC,$7E,$DC,$C9,$5C,$D7,$E1,$63
	.dc.b	$2D,$8C,$E7,$FE,$A9,$9C,$B6,$33,$1F,$EB,$33,$25,$73,$3F,$EB,$95
	.dc.b	$5E,$5B,$F3,$36,$BC,$8D,$AF,$13,$6F,$C0,$D9,$EC,$D9,$EC,$FF,$18
	.dc.b	$F7,$7F,$96,$12,$7F,$95,$36,$7F,$92,$B5,$FE,$46,$D7,$F9,$F3,$5F
	.dc.b	$E7,$8F,$F3,$7F,$76,$4F,$F3,$56,$CF,$F3,$06,$BF,$CB,$9A,$FF,$49
	.dc.b	$6B,$FD,$1D,$AF,$F5,$1F,$77,$FA,$21,$27,$FA,$13,$63,$F9,$EB,$5F
	.dc.b	$E7,$6D,$7F,$AB,$35,$FE,$A8,$FF,$4F,$F7,$64,$FF,$4D,$6C,$FF,$48
	.dc.b	$6B,$FD,$19,$AF,$F6,$16,$BF,$D7,$DA,$FF,$69,$F7,$3F,$AF,$73,$FF
	.dc.b	$3E,$E1,$46,$39,$7F,$A6,$B7,$0F,$E8,$8D,$BF,$DD,$B7,$5F,$EF,$ED
	.dc.b	$9F,$EE,$4F,$F6,$92,$3F,$D9,$DB,$3F,$D8,$9A,$FF,$60,$6B,$CA,$38
	.dc.b	$11,$E0,$F9,$0F,$EB,$7D,$0F,$85,$F4,$C6,$67,$CE,$EA,$33,$18,$25
	.dc.b	$A7,$FC,$1F,$5C,$DE,$51,$EE,$69,$AA,$E7,$19,$AE,$73,$93,$6F,$F8
	.dc.b	$46,$F7,$19,$8D,$99,$76,$FF,$8C,$6B,$6E,$4D,$99,$8D,$E3,$F6,$9E
	.dc.b	$38,$7A,$67,$6D,$C3,$CC,$7B,$E0,$73,$37,$7E,$07,$11,$C2,$BD,$0D
	.dc.b	$8E,$35,$D6,$7E,$C4,$7F,$46,$8C,$3E,$99,$F0,$3D,$27,$71,$89,$DC
	.dc.b	$FE,$5E,$7E,$C1,$EC,$AF,$24,$EA,$FE,$05,$77,$9E,$AF,$CD,$3F,$C7
	.dc.b	$13,$F0,$DB,$1A,$7D,$2D,$7E,$11,$AA,$FF,$20,$3A,$6B,$A3,$B1,$0D
	.dc.b	$1C,$27,$D6,$4D,$FF,$C9,$CE,$2A,$98,$8B,$5F,$02,$BD,$8F,$E7,$A3
	.dc.b	$31,$82,$61,$F2,$63,$2F,$CB,$AB,$5F,$9C,$9B,$5E,$6A,$D4,$F5,$3C
	.dc.b	$0F,$36,$9C,$4B,$1F,$80,$9E,$6F,$FA,$1D,$B7,$FD,$32,$77,$39,$6C
	.dc.b	$66,$3D,$31,$D5,$A7,$57,$73,$D0,$8E,$09,$1C,$0E,$C5,$9E,$D5,$EE
	.dc.b	$8F,$CA,$57,$27,$4F,$E0,$67,$C9,$1F,$88,$8C,$12,$BF,$03,$3A,$9F
	.dc.b	$41,$B8,$CC,$EE,$3F,$04,$D7,$47,$8A,$3F,$CE,$7F,$8F,$3D,$6F,$C1
	.dc.b	$47,$D3,$9D,$1F,$E3,$09,$D4,$67,$0F,$AD,$3A,$7F,$5B,$3A,$8C,$4F
	.dc.b	$CD,$9D,$9A,$AE,$F3,$A3,$E8,$1D,$DC,$C7,$09,$E4,$7F,$AD,$2B,$B4
	.dc.b	$66,$B8,$B5,$46,$A7,$8C,$69,$AE,$AB,$BC,$63,$0F,$B2,$F7,$8C,$CF
	.dc.b	$12,$1C,$63,$E2,$1C,$70,$FF,$5E,$5F,$86,$77,$7D,$1E,$31,$97,$F1
	.dc.b	$38,$7A,$67,$51,$D4,$F8,$C6,$AF,$D3,$5C,$67,$64,$BF,$0E,$1E,$8B
	.dc.b	$E0,$78,$1C,$4E,$EF,$F0,$71,$8A,$E1,$7D,$5E,$73,$EE,$BF,$E7,$33
	.dc.b	$98,$E2,$66,$3D,$18,$7D,$39,$D4,$7D,$D9,$D4,$62,$7F,$09,$18,$BF
	.dc.b	$09,$1C,$12,$BE,$1D,$71,$8F,$AC,$66,$BC,$15,$C6,$B6,$D8,$CC,$EC
	.dc.b	$F9,$71,$96,$99,$9F,$90,$D7,$E5,$9E,$E7,$6F,$67,$F9,$BD,$7E,$8E
	.dc.b	$3F,$DF,$4D,$7C,$E7,$F3,$E7,$0F,$CC,$9F,$E8,$91,$88,$FC,$11,$CA
	.dc.b	$32,$49,$EE,$D8,$F9,$C7,$FA,$A3,$F5,$B7,$7F,$DA,$8C,$E1,$F4,$1D
	.dc.b	$46,$63,$04,$BF,$EE,$F2,$AE,$71,$F8,$08,$FE,$0D,$EA,$7F,$7D,$AF
	.dc.b	$ED,$FF,$A2,$3C,$0F,$D1,$CE,$6F,$9C,$E5,$FE,$27,$19,$BF,$F0,$2C
	.dc.b	$E5,$EB,$3F,$E0,$8B,$EB,$7D,$23,$07,$47,$13,$FE,$C8,$AE,$55,$FC
	.dc.b	$5A,$FE,$69,$2F,$EB,$B6,$BE,$A3,$5F,$B2,$D7,$F8,$93,$5F,$D1,$BC
	.dc.b	$67,$B4,$FF,$2E,$F2,$3C,$71,$8A,$F5,$9F,$CB,$63,$33,$D0,$F8,$67
	.dc.b	$12,$C7,$09,$E4,$7F,$30,$9E,$31,$E2,$9D,$CE,$5A,$75,$8E,$4F,$F2
	.dc.b	$B9,$E3,$1F,$CA,$2A,$E8,$9B,$18,$6D,$74,$AF,$E0,$71,$87,$A9,$88
	.dc.b	$D3,$D5,$B5,$D5,$B5,$D5,$A6,$AB,$A5,$FC,$08,$F4,$1A,$BE,$E7,$F4
	.dc.b	$09,$E4,$E2,$FB,$36,$33,$5C,$E3,$CC,$66,$BC,$04,$70,$48,$E0,$F7
	.dc.b	$8D,$46,$2B,$BB,$6B,$BC,$FD,$18,$FA,$95,$CE,$30,$7B,$1C,$1F,$6C
	.dc.b	$FC,$33,$D8,$DB,$E8,$8E,$D1,$98,$E1,$1D,$8C,$BD,$0F,$A4,$74,$31
	.dc.b	$4B,$FE,$B3,$7E,$89,$E7,$7E,$5B,$E6,$47,$C7,$1F,$0C,$E7,$5C,$A3
	.dc.b	$F9,$1D,$7F,$1C,$23,$82,$4F,$F2,$38,$CB,$63,$2F,$F2,$A8,$CB,$63
	.dc.b	$F4,$CA,$BF,$CF,$FF,$A2,$7F,$16,$8F,$74,$7C,$39,$F5,$1F,$C7,$27
	.dc.b	$51,$99,$F9,$93,$A7,$FD,$2F,$3A,$8C,$47,$BE,$78,$29,$D3,$C8,$86
	.dc.b	$E7,$C1,$A7,$C5,$8F,$2C,$74,$9E,$31,$B3,$8E,$3F,$8C,$9E,$27,$C0
	.dc.b	$BF,$0C,$66,$7C,$F7,$D6,$73,$87,$F2,$D9,$F3,$4E,$6F,$D7,$7C,$67
	.dc.b	$31,$D4,$CD,$71,$C3,$CB,$5B,$24,$66,$FA,$5F,$A2,$31,$7C,$E3,$0F
	.dc.b	$F3,$87,$CE,$E5,$F8,$04,$BE,$F3,$D2,$31,$56,$FC,$37,$CE,$7D,$27
	.dc.b	$CC,$3B,$9F,$85,$7E,$4B,$93,$BC,$7C,$1B,$EF,$3E,$38,$C5,$77,$BF
	.dc.b	$15,$6A,$F9,$56,$AF,$85,$6A,$F8,$47,$B6,$F8,$61,$F5,$23,$D1,$18
	.dc.b	$7E,$79,$89,$E3,$87,$9E,$74,$FB,$2B,$F3,$F3,$A7,$F2,$73,$A9,$F6
	.dc.b	$DF,$D2,$7C,$EF,$96,$FC,$E7,$96,$F9,$3F,$93,$6C,$FE,$6A,$BA,$61
	.dc.b	$F4,$23,$E6,$1F,$68,$EB,$5C,$C8,$FA,$6B,$C0,$D8,$CE,$1E,$89,$FB
	.dc.b	$DF,$64,$EE,$FF,$73,$36,$BF,$DC,$6D,$BF,$F7,$3B,$5F,$F0,$B8,$70
	.dc.b	$FF,$86,$9A,$8D,$4F,$62,$4F,$66,$C7,$67,$C7,$F0,$F2,$7C,$6D,$9F
	.dc.b	$1B,$5F,$1D,$23,$11,$E3,$6B,$E3,$8D,$4F,$43,$DF,$DF,$F0,$88,$E0
	.dc.b	$93,$F0,$A3,$2D,$8C,$BE,$28,$CB,$63,$E5,$F7,$FC,$E2,$38,$24,$7C
	.dc.b	$E8,$CB,$63,$2F,$BA,$32,$D8,$FA,$FF,$0B,$F6,$88,$E0,$91,$F6,$A3
	.dc.b	$2D,$8C,$BF,$2A,$32,$D8,$FC,$1F,$C7,$FE,$18,$8E,$09,$1F,$86,$8C
	.dc.b	$B6,$32,$FD,$68,$CB,$63,$F1,$6F,$E1,$FE,$91,$EF,$FF,$95,$E5,$E1
	.dc.b	$BD,$DF,$E5,$D5,$A9,$FE,$3C,$6A,$33,$5C,$6B,$84,$FF,$19,$AE,$17
	.dc.b	$FA,$C8,$E0,$66,$B8,$5F,$28,$EB,$7C,$AB,$F6,$62,$F9,$F0,$2B,$F4
	.dc.b	$07,$49,$DC,$EA,$33,$5C,$EB,$85,$7C,$1C,$3C,$35,$E3,$3C,$38,$78
	.dc.b	$E7,$D1,$7C,$67,$FD,$13,$87,$59,$F1,$4E,$5E,$93,$C8,$E8,$F2,$AE
	.dc.b	$98,$78,$A7,$D0,$F9,$27,$DD,$87,$91,$F5,$B6,$35,$87,$86,$B8,$9E
	.dc.b	$49,$D3,$4F,$15,$72,$6D,$72,$9D,$CF,$A0,$F3,$4F,$A2,$7C,$D8,$7B
	.dc.b	$67,$CB,$3C,$8E,$CF,$03,$B1,$67,$B4,$6E,$3B,$47,$4C,$3E,$1C,$FD
	.dc.b	$66,$CF,$CC,$6C,$7B,$A3,$8B,$B3,$CC,$57,$C3,$3D,$63,$31,$CD,$A7
	.dc.b	$37,$D6,$79,$DF,$5C,$71,$8E,$71,$D6,$3D,$71,$B3,$C6,$75,$36,$57
	.dc.b	$67,$32,$C7,$33,$B9,$63,$9B,$E5,$C3,$EF,$5F,$D1,$6D,$FC,$48,$C6
	.dc.b	$1F,$6A,$FD,$78,$77,$C3,$AC,$78,$DA,$75,$9E,$18,$74,$8F,$8B,$7D
	.dc.b	$1F,$B1,$5D,$1F,$33,$4E,$87,$A2,$38,$3F,$14,$D4,$77,$39,$CF,$78
	.dc.b	$F8,$F1,$97,$84,$66,$FE,$3C,$76,$39,$D7,$68,$F7,$47,$87,$0F,$88
	.dc.b	$4B,$F5,$B6,$3B,$D7,$19,$F5,$DF,$F1,$69,$FE,$DE,$D7,$F3,$A6,$30
	.dc.b	$FD,$B3,$82,$1F,$A9,$35,$3F,$CB,$9F,$C6,$98,$3F,$98,$CF,$57,$81
	.dc.b	$F5,$8F,$6F,$C4,$FE,$89,$F9,$C6,$27,$F3,$6E,$27,$F0,$B3,$B7,$F8
	.dc.b	$A4,$6C,$FC,$99,$BA,$FE,$6D,$3B,$7F,$14,$DA,$FC,$0B,$67,$E3,$B7
	.dc.b	$1F,$A7,$3F,$D1,$A3,$0F,$E3,$30,$46,$2B,$F0,$8D,$7C,$95,$F7,$1F
	.dc.b	$25,$7F,$40,$9D,$61,$EC,$8F,$7B,$EC,$31,$3E,$C7,$18,$7B,$27,$6F
	.dc.b	$B1,$A7,$B0,$B3,$EC,$6D,$FB,$30,$F4,$C6,$0F,$66,$1E,$4B,$ED,$7E
	.dc.b	$17,$B3,$67,$B3,$4E,$C5,$BE,$D5,$CE,$7E,$01,$DE,$7E,$3B,$5F,$8E
	.dc.b	$FE,$17,$0F,$7B,$C2,$3A,$3E,$63,$84,$7C,$6A,$F9,$CF,$09,$FE,$B7
	.dc.b	$5E,$16,$D7,$0B,$F6,$DF,$07,$DA,$D8,$F6,$B6,$7D,$AD,$3D,$B1,$F9
	.dc.b	$61,$E8,$7F,$83,$6D,$D5,$FE,$B9,$9C,$9F,$E0,$C8,$CB,$FE,$A1,$33
	.dc.b	$7F,$E0,$98,$CB,$FE,$04,$8C,$DF,$F6,$B9,$CB,$FD,$AE,$32,$7F,$81
	.dc.b	$A3,$2F,$F9,$9E,$7F,$C1,$57,$E5,$BF,$23,$67,$C4,$D9,$F0,$36,$FB
	.dc.b	$36,$7A,$1E,$09,$FF,$0B,$5F,$81,$FF,$0B,$CE,$DB,$3F,$06,$FC,$2D
	.dc.b	$3C,$25,$8F,$0B,$6F,$C2,$D3,$C3,$5D,$67,$DB,$19,$7B,$9F,$2E,$BB
	.dc.b	$B6,$3B,$B4,$EE,$5A,$EE,$D8,$EE,$75,$7B,$99,$3B,$C7,$D5,$9E,$ED
	.dc.b	$9E,$ED,$9E,$ED,$3B,$96,$3B,$C7,$53,$BC,$65,$EF,$1F,$7E,$7B,$B4
	.dc.b	$EE,$58,$EE,$D9,$EE,$D3,$BB,$D6,$3B,$C6,$5E,$E7,$DE,$9E,$ED,$8E
	.dc.b	$ED,$3B,$96,$7B,$B6,$3B,$9D,$5E,$E6,$4E,$EF,$F2,$E9,$EE,$D3,$B9
	.dc.b	$63,$BB,$67,$BB,$4E,$F1,$D4,$EE,$E5,$EF,$7F,$99,$72,$D8,$FC,$F1
	.dc.b	$B8,$CD,$79,$27,$57,$F9,$3A,$DC,$61,$FC,$41,$89,$E0,$47,$D7,$1B
	.dc.b	$AF,$C0,$46,$CC,$CE,$E3,$DF,$3B,$BF,$B1,$5B,$9C,$D6,$E7,$E1,$D6
	.dc.b	$EB,$E8,$1B,$8F,$CF,$D7,$FB,$CA,$3F,$89,$B6,$FF,$C0,$6F,$EB,$0C
	.dc.b	$47,$F8,$04,$DB,$F9,$83,$27,$FA,$C9,$CB,$C4,$D1,$0F,$C6,$9F,$53
	.dc.b	$27,$C4,$FE,$3A,$3F,$0A,$65,$C4,$FF,$11,$8C,$3F,$92,$76,$FD,$89
	.dc.b	$E6,$E4,$E0,$66,$79,$D7,$E3,$A7,$75,$F9,$A3,$9C,$7F,$B0,$67,$75
	.dc.b	$F9,$2A,$DD,$FE,$7A,$F7,$3F,$EB,$EA,$DC,$7E,$5A,$B7,$3F,$94,$8D
	.dc.b	$9F,$EB,$29,$DD,$7E,$6A,$B7,$3F,$1A,$BD,$37,$E3,$7D,$8F,$A6,$F8
	.dc.b	$C7,$E8,$0F,$F6,$FB,$89,$FD,$6D,$74,$7C,$37,$DE,$75,$1F,$30,$F8
	.dc.b	$55,$E0,$3D,$87,$80,$F8,$51,$C1,$F9,$86,$63,$E0,$46,$6F,$D1,$1D
	.dc.b	$1F,$44,$EE,$BC,$4D,$3D,$57,$E7,$7C,$55,$D1,$D3,$EC,$BE,$D1,$E6
	.dc.b	$8E,$CF,$98,$DB,$ED,$AF,$B1,$1D,$1F,$F6,$1C,$F2,$9F,$A9,$3C,$0F
	.dc.b	$6B,$D0,$F1,$16,$7D,$2D,$8F,$4C,$F3,$9E,$B1,$CE,$FE,$1C,$7E,$27
	.dc.b	$E4,$DE,$67,$EC,$30,$46,$2F,$FC,$E0,$D8,$FE,$FC,$0D,$CF,$18,$D4
	.dc.b	$69,$AF,$F8,$15,$AF,$16,$BC,$A7,$BC,$F3,$23,$E2,$F3,$3B,$11,$C1
	.dc.b	$27,$CA,$7A,$A3,$4E,$29,$18,$8F,$33,$5F,$33,$EB,$F3,$25,$7C,$26
	.dc.b	$CF,$36,$BE,$F6,$BE,$F6,$BE,$F6,$BF,$3B,$CC,$F9,$04,$9F,$9E,$D8
	.dc.b	$F4,$35,$F9,$8D,$7E,$63,$5F,$98,$FD,$3E,$89,$3F,$69,$B3,$F0,$DA
	.dc.b	$FD,$76,$BF,$5D,$AF,$D7,$6B,$F8,$6E,$8F,$C0,$12,$7F,$0E,$D8,$FA
	.dc.b	$0D,$7F,$08,$D7,$F0,$8D,$7F,$08,$FE,$2F,$9E,$4F,$E4,$9B,$3F,$6D
	.dc.b	$AF,$E3,$DA,$FE,$3D,$AF,$E3,$DA,$FE,$7B,$9F,$F3,$04,$9F,$CF,$B6
	.dc.b	$3F,$10,$D7,$F3,$8D,$7F,$38,$D7,$F3,$8F,$E8,$F9,$E4,$FE,$A9,$B3
	.dc.b	$F9,$36,$BF,$A7,$6B,$FA,$76,$BF,$A7,$6B,$FB,$6E,$7F,$D8,$12,$7F
	.dc.b	$6E,$D8,$FD,$03,$5F,$DA,$35,$FD,$A3,$5F,$DA,$3F,$BB,$F2,$E4,$FF
	.dc.b	$02,$6C,$FE,$AD,$AF,$EF,$DA,$FE,$FD,$AF,$EF,$DA,$FF,$0E,$F2,$FF
	.dc.b	$84,$12,$7F,$87,$B6,$3F,$70,$D7,$F8,$63,$5F,$E1,$8D,$7F,$86,$3F
	.dc.b	$C4,$FC,$B9,$3F,$C6,$9B,$3F,$C0,$DA,$FF,$17,$6B,$FC,$5D,$AF,$F1
	.dc.b	$76,$BF,$C9,$BC,$BF,$E4,$04,$9F,$E4,$ED,$8F,$E2,$0D,$7F,$92,$35
	.dc.b	$FE,$48,$D7,$F9,$23,$FC,$AF,$E2,$A4,$FF,$32,$6C,$FF,$1B,$6B,$FC
	.dc.b	$BD,$AF,$F2,$F6,$BF,$CB,$DA,$FF,$3E,$F2,$BF,$9C,$12,$7F,$9F,$B6
	.dc.b	$3F,$94,$35,$FE,$78,$D7,$F9,$E3,$5F,$E7,$8F,$F4,$38,$87,$F4,$A2
	.dc.b	$C7,$F3,$36,$BF,$D1,$DA,$FF,$47,$6B,$FD,$1D,$AB,$F2,$27,$C4,$FF
	.dc.b	$59,$6C,$FF,$42,$9D,$4E,$49,$19,$9D,$CF,$F4,$86,$BF,$D7,$9A,$FF
	.dc.b	$B7,$5A,$FF,$61,$6B,$DA,$3A,$4F,$8F,$E4,$24,$77,$6C,$F7,$6B,$DD
	.dc.b	$AF,$76,$BD,$DB,$0E,$B7,$C6,$23,$FB,$A7,$55,$BC,$56,$A7,$27,$F7
	.dc.b	$93,$97,$C2,$EE,$B8,$47,$5B,$C3,$BA,$E5,$6A,$DC,$61,$B7,$C2,$7C
	.dc.b	$37,$6F,$A1,$E2,$8C,$B4,$F1,$96,$FA,$3E,$19,$E5,$18,$9F,$6D,$62
	.dc.b	$65,$71,$77,$18,$9F,$44,$F3,$34,$77,$8D,$5F,$C9,$9F,$C7,$7C,$5F
	.dc.b	$80,$F6,$38,$AF,$34,$63,$33,$B8,$F7,$4E,$CF,$09,$F5,$23,$81,$B7
	.dc.b	$26,$EB,$E6,$C6,$A7,$B9,$D2,$BB,$B8,$AE,$31,$87,$BC,$7E,$1E,$79
	.dc.b	$46,$31,$D5,$FB,$EE,$DF,$DF,$BD,$CF,$8A,$3C,$11,$E7,$8E,$0D,$8F
	.dc.b	$D3,$36,$FF,$4C,$D7,$D9,$1D,$4E,$67,$A6,$79,$B4,$E6,$58,$E6,$D7
	.dc.b	$99,$F2,$DE,$6F,$B6,$7D,$31,$DE,$3E,$2C,$78,$EB,$DF,$1C,$4C,$CF
	.dc.b	$58,$CB,$F2,$4E,$B5,$F0,$E7,$67,$78,$C9,$DE,$3D,$71,$93,$6F,$13
	.dc.b	$25,$8F,$7C,$6E,$33,$1C,$CC,$BB,$8F,$86,$6E,$32,$6D,$CC,$6D,$F8
	.dc.b	$F1,$EB,$7C,$6E,$DC,$DF,$C4,$6D,$FC,$A7,$E3,$57,$91,$B5,$E0,$69
	.dc.b	$E0,$7A,$5F,$CA,$BE,$8D,$BE,$93,$E3,$BE,$8D,$3A,$16,$FA,$1D,$8F
	.dc.b	$2B,$F4,$1C,$57,$58,$C1,$C6,$36,$7C,$03,$75,$F8,$32,$47,$E0,$5B
	.dc.b	$1F,$34,$F0,$C7,$39,$ED,$1C,$E3,$13,$C2,$BE,$85,$69,$F6,$DF,$C2
	.dc.b	$9D,$57,$C6,$6D,$F9,$23,$C4,$47,$FA,$B4,$7F,$5F,$24,$7F,$5B,$6C
	.dc.b	$6A,$30,$47,$04,$3F,$AF,$47,$0A,$F2,$FC,$2F,$DF,$9F,$D6,$9F,$84
	.dc.b	$71,$87,$A6,$FF,$9B,$46,$2F,$ED,$E0,$8C,$4F,$D2,$8C,$3F,$5E,$FA
	.dc.b	$DF,$E1,$6B,$71,$F2,$EB,$6F,$D7,$37,$7F,$D0,$A7,$78,$7A,$30,$E9
	.dc.b	$87,$96,$FF,$8F,$56,$A7,$F1,$78,$74,$BF,$CD,$57,$49,$F5,$9C,$B0
	.dc.b	$F5,$DF,$19,$FC,$0D,$F1,$AF,$D0,$47,$13,$BB,$C4,$F8,$13,$C6,$3E
	.dc.b	$54,$65,$F7,$46,$70,$F8,$53,$9C,$3D,$95,$F0,$67,$D9,$1F,$DC,$2B
	.dc.b	$86,$1F,$0E,$F8,$1E,$E8,$CB,$EC,$BE,$53,$EC,$8D,$C7,$B2,$76,$FA
	.dc.b	$8D,$C7,$A8,$DB,$EA,$32,$7A,$9C,$DF,$AA,$33,$7E,$A3,$E7,$C7,$AA
	.dc.b	$7F,$C1,$95,$E0,$9F,$64,$72,$BF,$65,$F2,$8F,$61,$93,$D9,$3D,$27
	.dc.b	$D9,$3B,$3D,$93,$F8,$FF,$31,$EB,$7F,$82,$BE,$27,$FB,$DC,$62,$7F
	.dc.b	$C2,$11,$89,$D4,$7F,$7A,$9D,$57,$F6,$9A,$D4,$F0,$35,$5C,$27,$FB
	.dc.b	$DD,$70,$C3,$FC,$4F,$7C,$27,$95,$F0,$39,$3C,$2F,$95,$7F,$84,$EF
	.dc.b	$95,$7F,$8C,$EF,$94,$F4,$8E,$57,$D0,$E4,$F4,$8E,$57,$D2,$13,$2F
	.dc.b	$AB,$FE,$48,$BE,$B3,$DC,$EB,$7D,$E7,$A9,$DD,$EB,$1D,$CE,$B0,$AE
	.dc.b	$B1,$FE,$59,$AE,$B3,$DE,$3A,$9D,$DE,$B5,$DC,$EA,$F7,$8E,$B5,$D3
	.dc.b	$35,$F4,$9E,$C7,$4A,$EC,$74,$7B,$5F,$43,$B3,$D3,$0F,$8D,$3F,$15
	.dc.b	$AF,$AD,$B8,$78,$6F,$C1,$87,$8A,$FC,$07,$8A,$FC,$07,$8B,$0F,$04
	.dc.b	$78,$8F,$01,$F2,$63,$C0,$D7,$C0,$DB,$F0,$3D,$AB,$C1,$3D,$A3,$C1
	.dc.b	$5D,$A3,$C1,$1D,$9F,$94,$76,$29,$D8,$B5,$D9,$E9,$5D,$AB,$A5,$76
	.dc.b	$3A,$3F,$2A,$7A,$35,$E8,$DA,$E8,$F2,$AE,$95,$C8,$F8,$95,$C9,$AF
	.dc.b	$26,$D7,$27,$84,$7B,$8E,$05,$38,$16,$38,$3A,$8F,$53,$67,$C6,$DB
	.dc.b	$FB,$A4,$9F,$A0,$D8,$F8,$AD,$BF,$63,$67,$CA,$76,$9F,$1B,$6F,$C7
	.dc.b	$1F,$44,$FD,$49,$4F,$B0,$61,$FE,$26,$DC,$3F,$D3,$0E,$88,$7D,$D3
	.dc.b	$53,$FE,$E5,$9E,$31,$A3,$B1,$F2,$0F,$2E,$0F,$47,$F6,$F1,$F2,$8C
	.dc.b	$B8,$9F,$E1,$31,$8A,$FD,$83,$B8,$FD,$B4,$EE,$FF,$4E,$DA,$FC,$D3
	.dc.b	$6B,$F1,$ED,$C7,$F0,$CD,$9F,$97,$87,$95,$EF,$87,$88,$EE,$E3,$0E
	.dc.b	$ED,$3B,$97,$0E,$ED,$C3,$BB,$4E,$F8,$78,$67,$F0,$CD,$7E,$01,$FB
	.dc.b	$E3,$E0,$57,$FB,$62,$35,$7E,$49,$D3,$E7,$9D,$57,$79,$EA,$D7,$AD
	.dc.b	$FE,$ED,$EB,$1F,$EE,$7B,$EB,$5C,$A7,$AB,$F0,$1B,$3F,$01,$A7,$C0
	.dc.b	$2D,$7C,$06,$C7,$C0,$3C,$57,$E9,$9F,$13,$E9,$8F,$11,$AA,$F1,$4E
	.dc.b	$9F,$13,$63,$C4,$DA,$F1,$34,$F1,$16,$3C,$51,$E4,$3E,$29,$4F,$8A
	.dc.b	$FF,$02,$7F,$A8,$72,$FE,$91,$FE,$BC,$D7,$31,$FA,$38,$C5,$9A,$48
	.dc.b	$D1,$27,$F6,$A7,$22,$CF,$18,$D4,$7E,$DF,$97,$A9,$23,$F6,$4D,$8E
	.dc.b	$2D,$78,$46,$88,$E9,$AF,$F0,$3E,$5F,$49,$23,$F7,$CD,$8F,$2B,$5E
	.dc.b	$11,$A2,$3A,$6B,$FC,$3F,$97,$E2,$92,$3F,$85,$36,$3E,$0B,$5E,$11
	.dc.b	$A2,$3A,$6B,$FC,$6F,$97,$E8,$92,$3F,$8B,$36,$3E,$6B,$5E,$11,$A2
	.dc.b	$3A,$6B,$FC,$9D,$FB,$A4,$3F,$91,$15,$FB,$2D,$78,$3A,$23,$A6,$BF
	.dc.b	$CD,$BE,$A5,$CD,$F8,$D6,$CF,$E1,$A5,$18,$7F,$12,$D7,$94,$F1,$24
	.dc.b	$F1,$9D,$4F,$F8,$B1,$AF,$F8,$E9,$AF,$69,$E2,$F6,$E6,$F2,$12,$BA
	.dc.b	$36,$35,$18,$23,$82,$1E,$F8,$E0,$FC,$1F,$91,$F8,$04,$9F,$43,$63
	.dc.b	$DE,$D7,$84,$68,$8E,$9A,$FC,$DF,$21,$1F,$E5,$8D,$EC,$FF,$55,$FC
	.dc.b	$F9,$F1,$98,$9C,$CF,$D0,$BC,$5F,$F9,$83,$8D,$87,$08,$E8,$7C,$D9
	.dc.b	$DD,$75,$C2,$BC,$E7,$6E,$5A,$E4,$61,$E6,$C3,$BD,$72,$D1,$87,$69
	.dc.b	$CE,$1C,$A7,$2F,$68,$CE,$1D,$23,$CC,$67,$1F,$54,$64,$FC,$25,$F5
	.dc.b	$8C,$E3,$F1,$30,$EB,$5B,$C3,$AC,$E6,$F8,$63,$E7,$6E,$3E,$56,$DF
	.dc.b	$73,$E0,$57,$BF,$1E,$F5,$E6,$9E,$D1,$87,$CB,$1E,$9B,$EC,$D3,$B1
	.dc.b	$63,$B6,$3E,$87,$F0,$91,$FA,$5C,$3F,$60,$47,$04,$8F,$D7,$C6,$5B
	.dc.b	$19,$79,$98,$3B,$C7,$D5,$F6,$5E,$0F,$9A,$73,$5E,$67,$E5,$4E,$A3
	.dc.b	$19,$7C,$B6,$9F,$91,$8D,$4F,$22,$4E,$DE,$75,$DE,$30,$4A,$E5,$19
	.dc.b	$7E,$74,$66,$7F,$C7,$15,$DA,$31,$1F,$C4,$BD,$AF,$CE,$C3,$F7,$7F
	.dc.b	$C1,$BF,$85,$8C,$C6,$2F,$F8,$84,$63,$1F,$CD,$36,$3F,$30,$DB,$FA
	.dc.b	$ED,$BE,$93,$FA,$87,$C1,$5C,$A3,$17,$DE,$FF,$8E,$5F,$19,$F0,$9A
	.dc.b	$AE,$13,$F9,$FB,$E7,$3E,$23,$45,$8F,$E5,$F5,$FC,$D4,$91,$FC,$B5
	.dc.b	$B1,$A8,$C1,$1C,$10,$FE,$81,$1C,$27,$DB,$1F,$9E,$F5,$7C,$B1,$82
	.dc.b	$57,$F1,$1D,$53,$A8,$F6,$D7,$C8,$7F,$79,$19,$8F,$E4,$3F,$C0,$2F
	.dc.b	$3F,$F8,$AD,$3A,$FF,$DF,$6F,$F1,$41,$3B,$71,$1A,$AF,$E0,$D1,$AA
	.dc.b	$E7,$39,$C7,$F9,$6C,$E4,$E3,$5E,$0B,$E8,$F2,$AE,$13,$E2,$9C,$D7
	.dc.b	$29,$CD,$7A,$AF,$BB,$CD,$B7,$E9,$BE,$2F,$B2,$B8,$BE,$3A,$F0,$1C
	.dc.b	$5F,$29,$C6,$B9,$C7,$F0,$8F,$03,$FB,$A1,$27,$FB,$63,$63,$FB,$9B
	.dc.b	$5F,$EF,$8D,$7F,$B8,$35,$FE,$E0,$FF,$7F,$F0,$64,$FF,$77,$6C,$FF
	.dc.b	$7E,$6B,$FE,$07,$6B,$FD,$ED,$AF,$F7,$B6,$BF,$E0,$8E,$FF,$F0,$61
	.dc.b	$27,$FC,$08,$D8,$FF,$05,$B5,$FF,$0A,$35,$FF,$04,$35,$FF,$04,$3F
	.dc.b	$E1,$7F,$02,$4F,$F8,$3D,$B3,$FE,$16,$6B,$FE,$1F,$6B,$FE,$13,$6B
	.dc.b	$FE,$13,$6B,$FE,$20,$F0,$3F,$C4,$84,$9F,$F0,$E3,$63,$FC,$46,$D7
	.dc.b	$FC,$58,$D7,$FC,$40,$D7,$FC,$40,$F1,$D3,$FA,$E3,$FD,$DA,$03,$D2
	.dc.b	$BF,$2E,$CF,$9E,$1F,$DB,$EF,$FB,$A1,$C6,$F8,$0A,$E0,$F8,$27,$81
	.dc.b	$1F,$EC,$47,$82,$FB,$DF,$68,$FE,$B3,$7D,$AB,$FC,$0F,$1B,$3B,$4E
	.dc.b	$F0,$F5,$CF,$57,$FD,$35,$7D,$2B,$C4,$74,$9D,$1F,$E9,$48,$D4,$7D
	.dc.b	$18,$EE,$FB,$30,$F5,$B9,$3C,$27,$C8,$3C,$8E,$B9,$98,$FF,$8C,$23
	.dc.b	$8C,$EB,$1F,$74,$62,$3B,$46,$0F,$23,$D8,$F0,$46,$F0,$F3,$DF,$6C
	.dc.b	$3D,$35,$D4,$F7,$D6,$DE,$53,$B7,$A1,$B3,$AD,$F8,$27,$D5,$87,$A6
	.dc.b	$30,$79,$67,$E0,$1E,$83,$BB,$A6,$9A,$C3,$CF,$3C,$E3,$07,$91,$C3
	.dc.b	$F2,$E3,$78,$7B,$23,$B1,$E6,$38,$DF,$9A,$35,$3C,$27,$DA,$70,$C3
	.dc.b	$E0,$57,$C3,$38,$46,$A7,$84,$67,$0F,$74,$F8,$CF,$35,$F6,$69,$C4
	.dc.b	$B3,$C5,$F8,$27,$17,$B9,$97,$D4,$66,$7B,$BF,$4E,$FC,$F1,$F2,$E3
	.dc.b	$66,$70,$F8,$ED,$BF,$25,$F6,$AE,$17,$CA,$BE,$3C,$76,$7E,$25,$F2
	.dc.b	$6D,$F2,$9D,$61,$F1,$1E,$75,$B8,$C5,$74,$3D,$E7,$8C,$F8,$98,$78
	.dc.b	$DD,$1E,$A3,$B9,$AA,$EE,$FC,$A7,$BD,$F8,$E3,$A9,$C2,$7E,$56,$1E
	.dc.b	$D8,$F9,$46,$AF,$CA,$E0,$F0,$38,$7C,$D3,$CB,$0F,$64,$78,$4F,$3D
	.dc.b	$75,$77,$87,$91,$F4,$1D,$CE,$07,$78,$E0,$F9,$E3,$85,$77,$C3,$D9
	.dc.b	$3E,$43,$89,$FC,$77,$E0,$3F,$2A,$7E,$E4,$B5,$FC,$11,$B1,$FA,$A6
	.dc.b	$BF,$97,$6B,$CE,$3F,$96,$47,$86,$3D,$27,$86,$BE,$B3,$E9,$C3,$E1
	.dc.b	$C6,$23,$E0,$9F,$12,$33,$5F,$62,$75,$3E,$F3,$57,$EE,$C3,$E1,$E3
	.dc.b	$F4,$67,$47,$79,$D6,$1D,$23,$E6,$5F,$4A,$E0,$48,$F8,$E7,$23,$CA
	.dc.b	$73,$3D,$B7,$E0,$77,$1C,$67,$A4,$60,$8E,$09,$87,$B4,$91,$E4,$6B
	.dc.b	$F2,$5B,$5D,$E7,$8C,$EC,$86,$CB,$1B,$9E,$37,$F0,$67,$44,$9D,$3C
	.dc.b	$23,$51,$88,$D4,$60,$98,$7C,$C2,$47,$C4,$8C,$B6,$33,$7F,$85,$25
	.dc.b	$FE,$60,$93,$F7,$0E,$04,$8D,$B9,$6B,$F4,$DA,$FD,$62,$1F,$64,$98
	.dc.b	$7D,$59,$F8,$75,$E4,$7C,$33,$CD,$AF,$29,$E0,$FC,$59,$E0,$F8,$E7
	.dc.b	$6F,$63,$EF,$BD,$9F,$53,$A9,$E7,$1F,$32,$79,$BE,$13,$A1,$0C,$10
	.dc.b	$F2,$1D,$8F,$21,$D8,$E0,$6A,$BF,$0E,$F7,$34,$73,$3D,$0E,$B0,$FC
	.dc.b	$01,$30,$FB,$71,$F3,$A7,$78,$7B,$63,$18,$7B,$2F,$86,$1E,$CB,$F1
	.dc.b	$5E,$CF,$84,$FA,$4F,$25,$F2,$71,$5C,$9F,$39,$C8,$F7,$C6,$CF,$46
	.dc.b	$1E,$6A,$EB,$18,$BF,$15,$7A,$0E,$33,$A3,$C8,$FC,$83,$47,$91,$D3
	.dc.b	$C0,$D6,$1E,$FB,$F0,$46,$0F,$4B,$87,$E7,$46,$EB,$C6,$D8,$F6,$9E
	.dc.b	$0C,$3E,$06,$1E,$DC,$3C,$B3,$C8,$EB,$7C,$9A,$72,$BE,$95,$C6,$3D
	.dc.b	$07,$17,$E2,$3B,$38,$BB,$7E,$01,$EB,$76,$66,$FC,$B7,$F1,$27,$67
	.dc.b	$57,$CE,$FD,$73,$25,$8E,$2F,$58,$DB,$F0,$70,$F9,$0E,$DA,$E7,$0F
	.dc.b	$81,$7E,$4A,$F1,$5F,$4A,$F1,$1F,$00,$E5,$5C,$27,$BE,$1F,$19,$E3
	.dc.b	$87,$68,$C5,$79,$4E,$47,$84,$F9,$18,$78,$5D,$1E,$E3,$4F,$C7,$9E
	.dc.b	$91,$F3,$AB,$F4,$38,$FF,$6A,$8F,$F4,$EB,$6F,$FD,$4E,$D8,$34,$93
	.dc.b	$FD,$C0,$F4,$E1,$ED,$9F,$3D,$7A,$9F,$86,$79,$5C,$3F,$E9,$F3,$FC
	.dc.b	$DE,$53,$FD,$0A,$60,$FF,$58,$95,$C3,$B3,$FC,$DE,$64,$CB,$A7,$89
	.dc.b	$D0,$87,$FB,$13,$F2,$37,$A2,$FA,$FF,$04,$61,$B1,$99,$F3,$CE,$5B
	.dc.b	$19,$7F,$D1,$31,$96,$C6,$5F,$F5,$64,$65,$B1,$97,$C8,$D5,$1B,$73
	.dc.b	$3F,$E9,$6F,$A9,$10

;--------------------------------------------------------------------------------
;	FSQRT.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsqrt_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#63,-(sp)
	peamsg	'FSQRT.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSQRT.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsqrt_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsqrt.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSQRT.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsqrt.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSQRT.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsqrt_data_compressed::
	.dc.b	$00,$02,$A1,$14,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$FD
	.dc.b	$4F,$E5,$1F,$97,$BE,$4B,$CF,$E9,$7A,$3D,$2D,$4B,$E1,$2F,$C9,$E3
	.dc.b	$E4,$73,$F1,$7A,$F8,$8D,$FB,$9C,$3D,$3E,$3E,$9F,$2F,$4F,$9F,$A7
	.dc.b	$D3,$D3,$EB,$E9,$F6,$F4,$FB,$FA,$7E,$0F,$4F,$C3,$E9,$F8,$BD,$3F
	.dc.b	$97,$EC,$7C,$CF,$63,$E6,$FB,$1F,$D8,$FF,$14,$FE,$40,$7E,$63,$C5
	.dc.b	$3F,$98,$7F,$B9,$3F,$D9,$9B,$BF,$1D,$F9,$6F,$CF,$7E,$9B,$F5,$DF
	.dc.b	$B6,$FD,$F7,$F0,$6F,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$FE,$ED,$F8,$AB
	.dc.b	$E6,$2F,$C7,$FE,$13,$DC,$FE,$F5,$F3,$9F,$47,$8F,$FC,$0D,$7F,$E7
	.dc.b	$B8,$FA,$11,$FE,$09,$7F,$DB,$87,$FA,$50,$FF,$03,$1F,$CF,$CF,$80
	.dc.b	$7B,$D3,$EF,$6F,$1F,$83,$8F,$C3,$C7,$E2,$E3,$F1,$F1,$F9,$38,$FC
	.dc.b	$BC,$7E,$6E,$3F,$3F,$1F,$A3,$8F,$D3,$C7,$EA,$E3,$FF,$0A,$77,$7B
	.dc.b	$F7,$FF,$10,$3F,$ED,$CE,$1F,$F2,$6B,$FE,$97,$2F,$C6,$F0,$F8,$AE
	.dc.b	$3F,$05,$CB,$DE,$E7,$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83
	.dc.b	$DE,$F0,$FB,$DE,$2F,$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$F5,$38,$3C
	.dc.b	$FE,$F7,$A3,$DE,$FF,$16,$7C,$EF,$F8,$14,$7C,$5E,$3E,$2F,$3F,$17
	.dc.b	$AF,$8B,$DF,$C5,$F0,$F8,$BE,$3F,$11,$CB,$E1,$79,$FC,$2F,$4F,$85
	.dc.b	$EB,$F0,$BD,$BE,$17,$BF,$C2,$F8,$3E,$17,$C3,$F0,$BE,$2F,$83,$F1
	.dc.b	$FC,$1F,$93,$DB,$F2,$FB,$7E,$6F,$6F,$CF,$EC,$7F,$90,$FE,$7B,$F1
	.dc.b	$5E,$61,$FC,$4B,$7C,$BD,$7C,$16,$FE,$0B,$87,$C1,$71,$F8,$2E,$5F
	.dc.b	$05,$CF,$E0,$BF,$33,$E8,$F5,$F7,$BB,$7B,$DD,$FD,$EF,$07,$BD,$E1
	.dc.b	$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD,$EF,$79,$FD,$EF
	.dc.b	$47,$BD,$FE,$5B,$F9,$FF,$C4,$73,$BD,$FF,$E7,$2A,$1F,$03,$BF,$8A
	.dc.b	$E1,$F2,$BC,$7D,$5E,$5C,$BC,$FE,$0B,$A7,$C1,$75,$F8,$2E,$DF,$05
	.dc.b	$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE,$0B,$C9,$F0,$5E
	.dc.b	$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$7B,$AF,$EF,$B5,$FD,$FE,$BF,$C0
	.dc.b	$35,$FE,$03,$AF,$B7,$E0,$BD,$DF,$05,$FE,$79,$F7,$FF,$CA,$47,$E3
	.dc.b	$BA,$3F,$D0,$CB,$E0,$34,$EF,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1C
	.dc.b	$FE,$03,$A7,$C0,$7F,$10,$E8,$ED,$F0,$1D,$FE,$03,$C1,$F0,$1F,$C5
	.dc.b	$3A,$FF,$8A,$F5,$FF,$16,$EB,$FE,$2F,$D7,$FC,$63,$AF,$F8,$CF,$5F
	.dc.b	$F1,$AE,$6F,$E3,$7C,$DF,$E9,$1F,$C5,$BF,$CD,$47,$A7,$9C,$FA,$5F
	.dc.b	$F5,$03,$FE,$FC,$3F,$D6,$C7,$F8,$98,$FE,$BC,$7F,$5D,$78,$F0,$F5
	.dc.b	$E1,$ED,$C1,$DF,$83,$E4,$7C,$0F,$C9,$F8,$1F,$95,$F0,$3F,$2F,$E0
	.dc.b	$7E,$67,$C0,$FC,$DF,$81,$F9,$DF,$03,$F3,$FE,$07,$E8,$7C,$0F,$FA
	.dc.b	$A7,$E8,$7F,$4E,$7E,$98,$7C,$0E,$7E,$07,$5F,$03,$BF,$81,$E1,$F0
	.dc.b	$3C,$7E,$07,$97,$C0,$F3,$F8,$1E,$9F,$03,$D7,$E0,$7B,$7C,$0F,$7F
	.dc.b	$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03,$F2,$FC,$9F,$99,$D7,$F3,$7A
	.dc.b	$FE,$77,$5F,$CF,$EB,$FA,$1D,$7F,$47,$AF,$E9,$75,$FD,$3E,$BF,$A9
	.dc.b	$D7,$FE,$C5,$F4,$FF,$81,$9F,$C0,$8F,$6F,$83,$FE,$D3,$5D,$7B,$EB
	.dc.b	$E1,$D7,$C7,$AF,$97,$5F,$3E,$BE,$9D,$7F,$27,$E0,$7E,$57,$C0,$FC
	.dc.b	$BF,$81,$F0,$77,$7C,$DF,$81,$FF,$6B,$7D,$17,$EF,$0F,$DC,$F0,$7F
	.dc.b	$B8,$D7,$BF,$AF,$7F,$7E,$FF,$0F,$7F,$8F,$BF,$CB,$DF,$E7,$EF,$F4
	.dc.b	$E7,$EB,$EF,$FB,$FC,$8F,$81,$E4,$7C,$1F,$23,$E1,$79,$1F,$0F,$C8
	.dc.b	$F1,$FB,$FE,$4F,$67,$CB,$EC,$F9,$BD,$9F,$3F,$B3,$E8,$F6,$7D,$3E
	.dc.b	$CF,$AB,$D9,$F5,$FB,$3E,$CF,$67,$FD,$E5,$F4,$7F,$EC,$B3,$F4,$47
	.dc.b	$E8,$7B,$0B,$DF,$DF,$BF,$C3,$DF,$E3,$EF,$F2,$F7,$FE,$C7,$B1,$D3
	.dc.b	$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F
	.dc.b	$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$F0,$DE,$A7
	.dc.b	$AB,$DC,$74,$4F,$3F,$FA,$59,$FD,$28,$FC,$4A,$7F,$12,$EF,$DB,$E2
	.dc.b	$E7,$DB,$B3,$FE,$33,$3F,$C2,$87,$F8,$4C,$7B,$1A,$F5,$B7,$E6,$70
	.dc.b	$F3,$38,$F9,$9C,$BC,$CE,$7E,$67,$4E,$CE,$BB,$F6,$F6,$3B,$FB,$1E
	.dc.b	$0F,$63,$C7,$F4,$9F,$D3,$0F,$C8,$F0,$9F,$C8,$BF,$01,$F6,$37,$D7
	.dc.b	$E3,$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$F6,$1A,$F8
	.dc.b	$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F,$5F
	.dc.b	$CF,$EB,$FA,$3D,$7F,$4F,$AF,$F3,$3E,$97,$E6,$F2,$9F,$E6,$07,$E3
	.dc.b	$CF,$C7,$3F,$69,$7A,$DC,$7D,$6E,$5E,$B7,$F0,$5E,$BE,$9E,$B7,$5F
	.dc.b	$5B,$B7,$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$78,$FD,$6F,$27
	.dc.b	$AD,$E5,$F5,$BC,$DE,$B7,$9F,$D6,$F4,$7A,$DE,$9F,$5B,$D5,$EB,$7E
	.dc.b	$1B,$E9,$BF,$C2,$E7,$F8,$58,$FC,$69,$F8,$CE,$72,$F5,$F8,$7A,$FC
	.dc.b	$7D,$EF,$E3,$DE,$77,$3F,$7B,$A7,$BD,$F9,$2F,$C5,$FF,$C1,$67,$E0
	.dc.b	$0F,$BF,$91,$FB,$ED,$F4,$38,$7A,$1C,$7D,$0E,$5E,$87,$3F,$43,$A7
	.dc.b	$A1,$D7,$D0,$ED,$E7,$F7,$F3,$FC,$1E,$7F,$87,$CF,$F1,$79,$FE,$3F
	.dc.b	$3F,$C9,$E7,$FC,$1F,$6B,$E1,$7B,$5F,$0F,$DA,$FF,$57,$FA,$7F,$EB
	.dc.b	$0F,$33,$FD,$63,$D3,$FE,$B7,$FC,$C3,$FC,$D9,$F2,$47,$F9,$5B,$7E
	.dc.b	$03,$5E,$F6,$FA,$78,$74,$F1,$E9,$E5,$D3,$CF,$A7,$A7,$4F,$5E,$9E
	.dc.b	$DD,$3D,$FA,$7C,$1D,$3F,$5F,$DE,$FB,$1E,$F7,$8F,$AF,$ED,$7B,$DF
	.dc.b	$6F,$DE,$FB,$9E,$F7,$DD,$F7,$BE,$F7,$4F,$DF,$E9,$FC,$07,$4F,$F0
	.dc.b	$6F,$82,$FF,$1F,$9F,$B4,$3E,$A9,$F5,$07,$36,$F9,$B8,$73,$71,$E6
	.dc.b	$E5,$CD,$CF,$9B,$A7,$37,$5E,$6E,$DC,$DD,$F9,$BC,$1C,$9E,$1F,$03
	.dc.b	$C5,$E0,$78,$FC,$0F,$27,$BD,$E5,$F7,$BC,$DE,$F7,$F1,$CF,$A9,$FD
	.dc.b	$89,$FE,$1F,$3F,$9F,$1F,$B0,$F1,$4F,$EC,$1F,$E5,$6B,$DD,$E5,$EE
	.dc.b	$F3,$F7,$7A,$7B,$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF
	.dc.b	$17,$BB,$E3,$F7,$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE
	.dc.b	$FA,$7D,$DF,$E6,$BF,$55,$FE,$86,$3F,$D0,$A7,$F3,$C3,$F6,$07,$EB
	.dc.b	$F9,$7F,$A1,$AF,$73,$8F,$B9,$CB,$DC,$E7,$EE,$74,$F6,$FA,$FB,$7D
	.dc.b	$BD,$BE,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F
	.dc.b	$2F,$AF,$E6,$F5,$FC,$FE,$BF,$F4,$CF,$AB,$FF,$42,$9F,$CE,$8F,$DD
	.dc.b	$9F,$BA,$3F,$0F,$C9,$FD,$69,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7
	.dc.b	$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F
	.dc.b	$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5,$BF,$B1,$FC,$07,$F3,$83
	.dc.b	$FC,$CC,$7F,$81,$4F,$C4,$1F,$87,$E4,$FE,$6F,$EB,$70,$F5,$B8,$FA
	.dc.b	$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD,$DB,$D6,$EF,$EB,$78,$3D,$6F
	.dc.b	$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6,$F2,$FA,$DE,$6F,$5B,$CF,$EB
	.dc.b	$7F,$7A,$FA,$CF,$F6,$59,$FE,$CA,$3F,$C0,$27,$F3,$53,$F0,$DC,$87
	.dc.b	$F0,$CF,$F8,$29,$7A,$BC,$FD,$5E,$9E,$AF,$5F,$57,$B7,$AB,$DF,$D5
	.dc.b	$F0,$7A,$BE,$1F,$57,$C5,$EA,$F8,$FD,$5F,$27,$AB,$E5,$F5,$7C,$DE
	.dc.b	$AF,$9F,$D5,$F4,$7A,$BF,$E1,$2F,$AD,$FF,$65,$9F,$E0,$03,$FA,$09
	.dc.b	$FD,$00,$FC,$3F,$19,$FC,$3B,$FE,$22,$7F,$DA,$23,$D3,$E9,$E9,$F5
	.dc.b	$F4,$FB,$7A,$7D,$FD,$3F,$07,$A7,$E1,$F4,$FC,$5E,$9F,$8F,$D3,$F2
	.dc.b	$7A,$7E,$5F,$4F,$CD,$E9,$F9,$FD,$3F,$47,$A7,$FE,$2A,$FA,$EF,$EF
	.dc.b	$E7,$FA,$B4,$FF,$26,$1F,$C3,$CF,$C4,$1F,$C3,$78,$BF,$C7,$2B,$D2
	.dc.b	$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9,$78
	.dc.b	$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5E,$7F,$4B
	.dc.b	$D1,$E9,$7F,$90,$BD,$B7,$CC,$5F,$97,$C7,$C5,$E7,$DF,$D7,$C2,$EF
	.dc.b	$DE,$E1,$EC,$71,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1
	.dc.b	$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6
	.dc.b	$3C,$DE,$C7,$F9,$63,$EB,$DF,$2F,$7E,$47,$1F,$0B,$9F,$85,$D7,$66
	.dc.b	$FE,$0B,$87,$C1,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7
	.dc.b	$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$E5
	.dc.b	$F7,$BC,$DE,$F7,$F9,$BB,$EC,$3F,$C8,$67,$FB,$C4,$FF,$48,$1F,$D4
	.dc.b	$CF,$E1,$87,$E1,$7D,$F3,$F8,$57,$FC,$FE,$FF,$BC,$47,$B9,$D7,$DC
	.dc.b	$ED,$EE,$77,$F7,$3C,$1E,$E7,$87,$DC,$F1,$7B,$9E,$3F,$73,$C9,$EE
	.dc.b	$79,$7D,$CF,$37,$B9,$E7,$F7,$3D,$1E,$E7,$A7,$DC,$FF,$43,$FD,$8F
	.dc.b	$E1,$B8,$BF,$D2,$2B,$DC,$CF,$B9,$AF,$73,$7E,$E7,$0F,$73,$8F,$B9
	.dc.b	$CB,$DC,$E7,$EE,$74,$F7,$3A,$FB,$9D,$BD,$CE,$FE,$E7,$83,$DC,$F0
	.dc.b	$FB,$9E,$2F,$73,$C7,$EE,$79,$3D,$CF,$2F,$B9,$E6,$F7,$3C,$FE,$E7
	.dc.b	$FA,$7B,$8B,$F0,$FC,$FF,$EA,$87,$FD,$1A,$7F,$93,$8F,$F2,$69,$FC
	.dc.b	$38,$FE,$1A,$FF,$96,$FD,$5E,$5E,$AF,$3F,$57,$A7,$AB,$D7,$D5,$ED
	.dc.b	$EA,$F7,$F5,$7C,$1E,$AF,$87,$D5,$F1,$7A,$BE,$3F,$57,$C9,$EA,$F9
	.dc.b	$7D,$5F,$37,$AB,$E7,$F5,$7F,$D6,$BF,$65,$FE,$8A,$1F,$31,$8F,$95
	.dc.b	$CF,$C8,$EB,$E2,$37,$F1,$1C,$3E,$17,$8F,$C2,$F2,$F7,$39,$FB,$9D
	.dc.b	$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC,$F1
	.dc.b	$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$FE,$CC,$FB,$3F,$C3
	.dc.b	$79,$07,$F0,$CF,$FB,$61,$FF,$6A,$9F,$E8,$93,$FA,$A9,$FD,$51,$E1
	.dc.b	$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D
	.dc.b	$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1
	.dc.b	$EC,$7F,$B9,$3E,$D3,$F0,$FC,$47,$FD,$A6,$7F,$A5,$0F,$F4,$99,$FD
	.dc.b	$50,$FC,$38,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1,$DF
	.dc.b	$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6,$3C
	.dc.b	$DE,$C7,$9F,$D8,$FF,$7D,$FA,$7F,$ED,$03,$F0,$FC,$63,$FC,$B9,$EB
	.dc.b	$6B,$D6,$DF,$AD,$C2,$B8,$FA,$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD
	.dc.b	$DB,$D6,$EF,$EB,$78,$3D,$6F,$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6
	.dc.b	$F2,$FA,$DE,$6F,$5B,$CF,$EB,$7A,$3D,$6F,$37,$33,$ED,$7F,$12,$7E
	.dc.b	$21,$C7,$BB,$A8,$FF,$68,$1F,$E1,$C3,$F1,$07,$F8,$6A,$3E,$02,$F5
	.dc.b	$B8,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77
	.dc.b	$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B,$C9,$EB,$7A,$BE,$DB
	.dc.b	$F1,$23,$E1,$B1,$F0,$D9,$F8,$6D,$7C,$36,$FE,$1B,$87,$B7,$C7,$DB
	.dc.b	$E5,$ED,$F3,$F6,$FA,$7B,$7D,$7D,$BE,$DE,$DF,$7F,$6F,$C1,$ED,$F8
	.dc.b	$7D,$BF,$17,$B7,$E3,$F6,$FC,$9E,$DF,$97,$DB,$F3,$7B,$7F,$43,$ED
	.dc.b	$FE,$8F,$8E,$7F,$12,$DF,$73,$5E,$E6,$FD,$CE,$1E,$E7,$1F,$73,$97
	.dc.b	$B9,$CF,$DC,$E9,$EE,$75,$F7,$3B,$7B,$9D,$FD,$CF,$07,$B9,$E1,$F7
	.dc.b	$3C,$5E,$E7,$8F,$DC,$F2,$7B,$9E,$5F,$73,$CD,$EE,$79,$FD,$CF,$47
	.dc.b	$B9,$F8,$8F,$1F,$F8,$79,$F8,$71,$EE,$67,$E1,$35,$F0,$7B,$F8,$4E
	.dc.b	$1F,$03,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D,$FE,$DE,$FF,$7F
	.dc.b	$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E,$FF,$97,$DF,$F3
	.dc.b	$7B,$FF,$9E,$FB,$8F,$E1,$A7,$E1,$7C,$BF,$DB,$2F,$7B,$5E,$F6,$FD
	.dc.b	$EE,$1D,$5C,$7D,$EE,$5E,$F7,$3F,$7B,$A7,$BD,$D7,$DE,$ED,$EF,$77
	.dc.b	$F7,$BC,$1E,$F7,$87,$DE,$F1,$7B,$DE,$3F,$7B,$C9,$EF,$79,$7D,$EF
	.dc.b	$37,$BD,$E7,$F7,$BF,$6B,$F7,$3F,$86,$3F,$0A,$3D,$EC,$FB,$DA,$F7
	.dc.b	$B7,$EF,$70,$F7,$B8,$FB,$DC,$BD,$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB
	.dc.b	$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD,$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2
	.dc.b	$FB,$DE,$6F,$7B,$CF,$EF,$7F,$0C,$FB,$AF,$EA,$67,$F0,$B3,$F0,$BE
	.dc.b	$67,$F2,$55,$EF,$6F,$DE,$E1,$EF,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E
	.dc.b	$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD
	.dc.b	$EF,$27,$BD,$E5,$F7,$BC,$DE,$F7,$9F,$DE,$FE,$47,$C3,$FD,$48,$FC
	.dc.b	$37,$2F,$FB,$F7,$DD,$CF,$BB,$AF,$77,$7E,$EF,$0F,$77,$8F,$BB,$CB
	.dc.b	$DD,$E7,$EE,$F4,$F7,$7A,$FB,$BD,$BD,$DE,$FE,$EF,$83,$DD,$F0,$FB
	.dc.b	$BE,$2F,$77,$C7,$EE,$F9,$3D,$DF,$2F,$BB,$E6,$F7,$7F,$9C,$FD,$DF
	.dc.b	$F0,$A3,$F8,$49,$F8,$5E,$6F,$EA,$4B,$DC,$DF,$B9,$C3,$DC,$E3,$EE
	.dc.b	$72,$F7,$39,$FB,$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C
	.dc.b	$3E,$E7,$8B,$DC,$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9
	.dc.b	$E8,$F7,$3F,$A8,$FD,$E7,$F8,$60,$FE,$A0,$7E,$1B,$9C,$FE,$19,$FE
	.dc.b	$D8,$FF,$BB,$0F,$F0,$C1,$FD,$45,$DF,$B3,$D3,$D9,$EB,$EC,$F6,$F6
	.dc.b	$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE
	.dc.b	$CF,$9B,$D9,$F3,$FB,$3E,$8F,$67,$FB,$4F,$DE,$FF,$86,$0F,$C3,$9F
	.dc.b	$C2,$F8,$7F,$C0,$4B,$D8,$DF,$B1,$C3,$D8,$E3,$EC,$72,$F6,$39,$FB
	.dc.b	$1D,$3D,$8E,$BE,$C7,$6F,$63,$BF,$B1,$E0,$F6,$3C,$3E,$C7,$8B,$D8
	.dc.b	$F1,$FB,$1E,$4F,$63,$CB,$EC,$79,$BD,$8F,$3F,$B1,$FE,$00,$F5,$3F
	.dc.b	$A8,$9F,$D4,$0F,$C4,$73,$7F,$B9,$7D,$7D,$7A,$FB,$F5,$F8,$57,$1F
	.dc.b	$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F,$07,$AF
	.dc.b	$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB,$F9,$FD
	.dc.b	$7F,$F0,$BF,$DF,$7F,$97,$8F,$F0,$B1,$F8,$73,$F0,$DC,$C5,$EE,$F0
	.dc.b	$F7,$38,$FB,$BC,$BD,$AE,$7E,$D7,$4F,$6B,$AF,$B5,$DB,$DA,$EF,$ED
	.dc.b	$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7,$93,$DA,$F2,$FB,$5E,$6F
	.dc.b	$6B,$CF,$ED,$7F,$8B,$BE,$FF,$FC,$BA,$7F,$0C,$3F,$0D,$CA,$7F,$0C
	.dc.b	$FF,$92,$57,$B3,$C7,$D9,$E5,$EC,$F3,$F6,$7A,$7B,$3D,$7D,$97,$F0
	.dc.b	$0C,$2E,$5C,$E1,$F8,$07,$F9,$57,$E1,$9C,$FC,$63,$0D,$AF,$EF,$6D
	.dc.b	$9D,$C7,$39,$DB,$4D,$96,$36,$FB,$6B,$C2,$E8,$C4,$F2,$69,$9D,$53
	.dc.b	$C2,$B9,$C6,$2B,$83,$6B,$85,$6E,$FD,$04,$8F,$23,$5E,$F7,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF
	.dc.b	$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63
	.dc.b	$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF
	.dc.b	$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88
	.dc.b	$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A
	.dc.b	$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB
	.dc.b	$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B
	.dc.b	$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D
	.dc.b	$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B
	.dc.b	$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60
	.dc.b	$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC
	.dc.b	$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE
	.dc.b	$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78
	.dc.b	$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12
	.dc.b	$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71
	.dc.b	$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C
	.dc.b	$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C
	.dc.b	$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6
	.dc.b	$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09
	.dc.b	$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC
	.dc.b	$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE
	.dc.b	$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A
	.dc.b	$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12
	.dc.b	$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D
	.dc.b	$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12
	.dc.b	$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2
	.dc.b	$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C
	.dc.b	$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09
	.dc.b	$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA
	.dc.b	$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB
	.dc.b	$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E
	.dc.b	$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD
	.dc.b	$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9
	.dc.b	$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A
	.dc.b	$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46
	.dc.b	$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31
	.dc.b	$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9
	.dc.b	$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26
	.dc.b	$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12
	.dc.b	$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3
	.dc.b	$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C
	.dc.b	$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F
	.dc.b	$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18
	.dc.b	$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A
	.dc.b	$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C
	.dc.b	$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67
	.dc.b	$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB
	.dc.b	$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D
	.dc.b	$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38
	.dc.b	$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E
	.dc.b	$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97
	.dc.b	$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D
	.dc.b	$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57
	.dc.b	$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA
	.dc.b	$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3
	.dc.b	$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF
	.dc.b	$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17
	.dc.b	$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF
	.dc.b	$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E
	.dc.b	$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D
	.dc.b	$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35
	.dc.b	$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1
	.dc.b	$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D
	.dc.b	$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4
	.dc.b	$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F
	.dc.b	$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D2,$12,$95,$2F,$10,$BE,$15,$BA,$D4,$F1,$9C,$BC,$AB
	.dc.b	$8C,$62,$77,$3C,$EF,$C2,$EC,$E4,$E2,$37,$F4,$6F,$83,$7C,$77,$92
	.dc.b	$FC,$B3,$98,$C4,$7A,$E3,$0E,$1A,$F1,$6B,$CA,$7A,$BB,$6B,$B6,$BB
	.dc.b	$9E,$EA,$79,$BE,$19,$DB,$67,$72,$F0,$1F,$27,$86,$D7,$9D,$AF,$9D
	.dc.b	$AF,$C0,$6B,$EF,$6B,$E3,$6C,$F7,$9E,$C4,$9E,$CD,$7E,$FF,$DB,$BF
	.dc.b	$73,$FA,$11,$F3,$23,$18,$FC,$06,$C6,$67,$6F,$41,$F7,$8F,$84,$F4
	.dc.b	$61,$F8,$F8,$C4,$FA,$63,$13,$F9,$0F,$A6,$7C,$17,$E4,$AF,$77,$ED
	.dc.b	$BD,$BC,$4E,$F7,$D9,$A7,$E0,$FA,$F9,$47,$E7,$F2,$F7,$6B,$F9,$EF
	.dc.b	$C0,$5F,$35,$9D,$7E,$B6,$3F,$A5,$9C,$AB,$E2,$E6,$B8,$79,$3C,$DF
	.dc.b	$16,$9E,$8E,$0F,$A5,$5E,$43,$D4,$5D,$FD,$EF,$B0,$CC,$F1,$CF,$F0
	.dc.b	$AD,$DB,$56,$DB,$DB,$7E,$5D,$5E,$4F,$8E,$31,$E4,$DD,$BE,$F3,$6B
	.dc.b	$73,$EB,$3E,$66,$5D,$DF,$46,$3E,$A8,$FA,$7A,$7A,$27,$3A,$79,$B8
	.dc.b	$3E,$09,$23,$E9,$1F,$63,$2F,$C7,$79,$3F,$8B,$C8,$D2,$4F,$E5,$F1
	.dc.b	$F1,$12,$FF,$0D,$C5,$F0,$27,$E3,$C7,$76,$9E,$03,$F0,$78,$7E,$77
	.dc.b	$C9,$F8,$18,$49,$F8,$13,$97,$E1,$1C,$89,$A7,$C1,$BF,$0C,$7D,$2C
	.dc.b	$FE,$03,$F2,$23,$C5,$3F,$56,$3E,$06,$7F,$C8,$B8,$FF,$17,$87,$EB
	.dc.b	$FC,$9F,$A4,$7D,$3E,$AF,$B1,$7E,$0C,$BA,$BE,$3D,$FE,$86,$FF,$0E
	.dc.b	$3C,$04,$3F,$8A,$71,$7E,$13,$1F,$DE,$79,$3D,$67,$DF,$A7,$8F,$97
	.dc.b	$F4,$13,$E0,$DB,$DD,$5C,$DA,$B2,$F0,$EF,$EE,$D3,$CB,$5A,$3E,$05
	.dc.b	$7D,$0C,$58,$7E,$3C,$91,$E2,$F2,$7C,$17,$C8,$9C,$DF,$A3,$CB,$BD
	.dc.b	$78,$E7,$2F,$E5,$35,$F7,$62,$C3,$F3,$84,$9F,$7F,$93,$DB,$1F,$B0
	.dc.b	$48,$FD,$97,$2F,$E2,$F4,$F5,$65,$DB,$83,$F7,$59,$F6,$F2,$7F,$50
	.dc.b	$4C,$3F,$51,$D6,$FD,$79,$F1,$9D,$E7,$F4,$BA,$77,$C7,$F5,$3A,$77
	.dc.b	$CF,$B7,$17,$E2,$34,$FD,$DE,$7C,$F2,$FE,$9E,$DC,$BF,$AF,$B5,$FE
	.dc.b	$C8,$D7,$FA,$46,$5D,$EF,$F4,$9C,$3F,$2A,$7F,$19,$39,$7F,$4D,$9F
	.dc.b	$F3,$4B,$E9,$CF,$FC,$81,$BC,$9F,$E0,$76,$D7,$F0,$FF,$2F,$DD,$7F
	.dc.b	$93,$8F,$CB,$6D,$ED,$C7,$B4,$7E,$AB,$5F,$AF,$E5,$FF,$11,$24,$7E
	.dc.b	$52,$39,$EB,$F8,$DA,$FC,$15,$F3,$AF,$96,$4C,$3E,$AB,$FA,$EF,$2F
	.dc.b	$F8,$E7,$97,$23,$F3,$F1,$D1,$FC,$55,$F4,$7C,$BC,$9F,$80,$8F,$E5
	.dc.b	$B9,$7E,$A6,$BF,$B9,$EF,$FB,$86,$EF,$FD,$2D,$B8,$7F,$45,$5F,$25
	.dc.b	$DE,$3E,$79,$F8,$E3,$BF,$1F,$E0,$1F,$A3,$B7,$3E,$3F,$E6,$D7,$FD
	.dc.b	$83,$1F,$04,$E2,$7A,$69,$E7,$BE,$87,$C3,$9F,$90,$FE,$94,$FE,$61
	.dc.b	$6F,$33,$A9,$23,$F9,$0C,$3C,$CE,$B9,$79,$67,$F8,$07,$0F,$D1,$E1
	.dc.b	$FB,$D9,$7E,$FF,$7F,$8D,$E6,$7A,$09,$7D,$3C,$CE,$7C,$5F,$82,$E7
	.dc.b	$FC,$96,$1E,$9C,$79,$E3,$F3,$FC,$CF,$82,$4C,$7C,$BE,$4F,$27,$DC
	.dc.b	$7F,$96,$70,$E7,$AF,$9F,$1F,$24,$FC,$0F,$8A,$F9,$7F,$15,$23,$DF
	.dc.b	$1C,$F3,$FC,$C6,$1C,$B4,$FA,$A7,$E0,$AB,$E4,$FC,$57,$D4,$E0,$91
	.dc.b	$F2,$23,$97,$27,$EB,$A7,$E3,$3F,$4B,$0F,$CA,$D7,$F5,$2E,$37,$C5
	.dc.b	$7D,$EF,$8A,$93,$F4,$A3,$9B,$FC,$EF,$3F,$23,$F0,$0F,$D5,$1F,$73
	.dc.b	$E2,$BF,$0F,$F1,$52,$3E,$DC,$72,$7F,$66,$7B,$0E,$5C,$1F,$CD,$B2
	.dc.b	$FF,$19,$F8,$EC,$DF,$15,$F9,$22,$5F,$E1,$BE,$2B,$A6,$1F,$D1,$33
	.dc.b	$F2,$E3,$F8,$4E,$4F,$E8,$7C,$FF,$A9,$F8,$AF,$F0,$87,$EE,$9F,$E7
	.dc.b	$9E,$FF,$EC,$2D,$F0,$7F,$55,$3D,$2F,$FA,$FE,$3D,$39,$3E,$51,$FE
	.dc.b	$80,$C7,$FA,$E6,$1F,$E7,$6C,$79,$FC,$87,$63,$BF,$8D,$CF,$A7,$F7
	.dc.b	$DC,$1F,$03,$3F,$C3,$63,$CB,$E3,$3C,$BF,$19,$2B,$FC,$07,$9A,$79
	.dc.b	$47,$EB,$B8,$3D,$D8,$FF,$08,$C3,$C7,$F1,$9E,$EF,$8C,$97,$CA,$38
	.dc.b	$DF,$F4,$08,$FA,$4F,$E8,$1F,$F1,$26,$2F,$8C,$F8,$FF,$19,$27,$C7
	.dc.b	$1C,$B8,$7F,$C2,$31,$EB,$CF,$C4,$F2,$F8,$CF,$9E,$47,$D7,$F1,$9C
	.dc.b	$5F,$D4,$78,$1F,$E4,$8F,$A3,$7D,$23,$C2,$FF,$24,$46,$1B,$19,$AF
	.dc.b	$F2,$34,$F0,$F0,$3E,$EF,$C7,$49,$F9,$2F,$E0,$3E,$3B,$C1,$1F,$C2
	.dc.b	$8F,$F3,$CD,$78,$3C,$3E,$A4,$8E,$B8,$7E,$49,$FA,$7F,$1D,$F8,$AF
	.dc.b	$8E,$91,$83,$FC,$6F,$E0,$FF,$0B,$E1,$FD,$94,$76,$B7,$85,$EA,$F0
	.dc.b	$64,$F6,$8F,$CA,$FC,$74,$8F,$C0,$BF,$0B,$C8,$FC,$B4,$7F,$48,$3F
	.dc.b	$65,$E1,$FF,$85,$36,$4E,$9B,$E0,$F1,$39,$54,$3F,$DA,$F1,$C1,$FF
	.dc.b	$65,$B5,$CC,$FF,$B3,$9F,$07,$5F,$F1,$6E,$5F,$B1,$39,$AE,$15,$FE
	.dc.b	$E5,$6B,$FE,$FA,$6B,$FE,$A1,$EB,$7C,$77,$ED,$09,$1F,$A0,$F8,$EF
	.dc.b	$AF,$9F,$B7,$07,$89,$F0,$FC,$29,$7F,$82,$9C,$BF,$A8,$F8,$EF,$DF
	.dc.b	$FC,$74,$8C,$1F,$5B,$C6,$FE,$6C,$7E,$FB,$6F,$5F,$8D,$F3,$C9,$3D
	.dc.b	$7C,$4F,$E1,$3F,$1D,$23,$F6,$AF,$F0,$CF,$8E,$F3,$DF,$C4,$F1,$FE
	.dc.b	$C1,$2B,$CD,$E3,$7F,$A8,$7F,$75,$FF,$5D,$C3,$CA,$FF,$39,$C6,$1B
	.dc.b	$E5,$EF,$E7,$3A,$D7,$E6,$38,$5E,$5F,$E1,$71,$93,$F0,$E3,$B7,$C6
	.dc.b	$77,$F8,$C9,$3F,$E8,$EC,$D3,$D2,$3F,$DA,$FD,$7F,$8A,$F3,$3F,$21
	.dc.b	$84,$9F,$9D,$1F,$93,$F3,$3D,$5C,$B2,$7E,$94,$73,$EE,$FF,$76,$E1
	.dc.b	$F9,$48,$FC,$D7,$95,$2B,$67,$D9,$F2,$BE,$27,$C6,$48,$C1,$F5,$FC
	.dc.b	$EF,$E2,$FE,$47,$F1,$0C,$BA,$F1,$30,$F2,$5F,$43,$E6,$72,$16,$9F
	.dc.b	$A6,$9F,$D6,$12,$3B,$9C,$0F,$F7,$77,$0F,$FB,$51,$C1,$FE,$2A,$C3
	.dc.b	$F4,$4F,$D3,$9E,$4F,$1A,$EC,$FE,$57,$CB,$FC,$07,$C5,$C8,$FA,$4F
	.dc.b	$EF,$63,$E4,$67,$E7,$AF,$DD,$C7,$4E,$7F,$F8,$03,$1D,$2F,$B6,$8A
	.dc.b	$E5,$1F,$B8,$7F,$77,$19,$8C,$1F,$ED,$BB,$FE,$B6,$FF,$32,$3F,$19
	.dc.b	$3C,$08,$F0,$3E,$51,$F2,$FD,$36,$9F,$EE,$88,$FC,$67,$C5,$FE,$74
	.dc.b	$FC,$F7,$C5,$FB,$23,$E8,$E3,$E3,$7D,$38,$7A,$1F,$5F,$9F,$FA,$5F
	.dc.b	$8A,$91,$F9,$77,$F5,$1F,$17,$CF,$5F,$E3,$FE,$6F,$F2,$52,$57,$F0
	.dc.b	$3F,$2F,$F5,$DF,$17,$23,$F4,$0F,$EC,$7E,$2B,$96,$9F,$C4,$A7,$F8
	.dc.b	$55,$F2,$7E,$6D,$FE,$DF,$E2,$BF,$71,$F1,$52,$7F,$55,$1C,$A7,$F9
	.dc.b	$B1,$23,$F8,$A7,$9F,$FB,$FF,$89,$FE,$01,$F1,$32,$7F,$65,$1E,$2F
	.dc.b	$3B,$F9,$F7,$9D,$27,$F8,$E4,$7F,$07,$F6,$FF,$84,$7B,$B2,$7F,$75
	.dc.b	$1E,$AF,$3B,$FA,$37,$9D,$27,$F9,$24,$7F,$0F,$F8,$9F,$B3,$DD,$FF
	.dc.b	$00,$3F,$76,$FF,$43,$F7,$1B,$DC,$AF,$9D,$F9,$BF,$F1,$3F,$13,$F2
	.dc.b	$8B,$CF,$FE,$EA,$DC,$FF,$B9,$BB,$58,$F5,$F8,$BF,$0D,$FE,$27,$2F
	.dc.b	$F6,$AD,$7E,$D7,$3F,$27,$C5,$EA,$38,$CF,$FA,$69,$D5,$7E,$EF,$F0
	.dc.b	$67,$E7,$AE,$9F,$CA,$F9,$2F,$1F,$F8,$07,$E0,$6F,$D5,$5E,$26,$A7
	.dc.b	$31,$8C,$FF,$83,$5F,$15,$97,$6C,$7F,$C2,$DC,$3F,$CA,$27,$DF,$EA
	.dc.b	$7C,$23,$E1,$FA,$9E,$9E,$3F,$99,$9D,$E0,$FF,$55,$70,$7C,$18,$C7
	.dc.b	$C7,$E7,$E3,$F5,$F1,$F3,$3F,$79,$F0,$A3,$E7,$FA,$B2,$3E,$7B,$F4
	.dc.b	$BD,$3E,$DF,$1F,$27,$8E,$9F,$76,$3C,$5F,$1F,$E3,$F8,$F9,$5A,$D9
	.dc.b	$1E,$7F,$8E,$FB,$BF,$25,$2B,$EE,$BF,$80,$F5,$BD,$9F,$21,$ED,$F3
	.dc.b	$E5,$73,$8E,$AF,$F2,$BD,$3F,$1D,$8E,$3B,$FF,$25,$97,$3A,$EA,$43
	.dc.b	$F9,$57,$3B,$2E,$F1,$D4,$87,$5D,$1F,$01,$D7,$9B,$F1,$D9,$5A,$FC
	.dc.b	$5E,$3D,$DE,$BB,$7B,$5E,$F8,$FF,$94,$3F,$16,$7C,$77,$F8,$91,$B3
	.dc.b	$FE,$28,$6B,$FE,$1F,$6B,$FE,$1F,$6B,$FC,$ED,$AF,$F5,$26,$FA,$A6
	.dc.b	$92,$7C,$04,$9F,$E6,$27,$EA,$7E,$47,$F1,$9F,$21,$23,$ED,$47,$E3
	.dc.b	$E7,$F2,$1F,$21,$27,$EF,$47,$EE,$3D,$C9,$1F,$B8,$7F,$77,$EE,$FE
	.dc.b	$5F,$E4,$64,$FE,$15,$FC,$D7,$C7,$7E,$6F,$E3,$E4,$FE,$25,$FD,$07
	.dc.b	$A1,$FE,$51,$E6,$7B,$DF,$C1,$BC,$A9,$7F,$C1,$A3,$7F,$1F,$23,$F2
	.dc.b	$F7,$FE,$AB,$E7,$FE,$9C,$FF,$8F,$23,$F1,$4F,$FA,$8D,$D9,$E0,$3C
	.dc.b	$3F,$19,$FC,$53,$E0,$24,$7F,$14,$7F,$8B,$FC,$07,$F1,$8F,$2F,$CF
	.dc.b	$F1,$92,$BF,$8C,$C7,$FA,$43,$A9,$7F,$C4,$FF,$05,$68,$F8,$6C,$37
	.dc.b	$0F,$F4,$9D,$F0,$79,$9F,$C9,$3E,$0B,$F9,$39,$E1,$7F,$93,$FB,$FF
	.dc.b	$23,$E3,$64,$60,$FE,$03,$F2,$CF,$C8,$5F,$1F,$E5,$D6,$56,$A7,$2F
	.dc.b	$13,$F5,$3F,$09,$FC,$D8,$FE,$6F,$8C,$8D,$79,$1F,$E0,$CE,$84,$FD
	.dc.b	$AC,$7F,$99,$65,$E1,$7B,$DF,$F8,$77,$5F,$73,$FE,$27,$9D,$3D,$39
	.dc.b	$BF,$A3,$DF,$78,$FD,$DF,$C0,$FE,$1B,$E3,$FF,$0F,$F1,$F2,$7F,$03
	.dc.b	$1E,$0D,$BD,$71,$DF,$7F,$0E,$CC,$BF,$C7,$99,$F8,$2B,$F8,$3E,$7D
	.dc.b	$EB,$D8,$FF,$56,$CF,$F2,$BF,$1F,$F9,$6F,$8F,$97,$F9,$08,$FE,$A0
	.dc.b	$BE,$CD,$1E,$3E,$09,$3F,$CD,$0D,$47,$9F,$69,$1F,$4D,$E7,$5C,$A3
	.dc.b	$EB,$71,$FB,$48,$F3,$8D,$3F,$E4,$E6,$E7,$EF,$86,$9E,$F8,$C3,$75
	.dc.b	$FF,$1A,$D7,$06,$E3,$E4,$BF,$AD,$BF,$FB,$F5,$F0,$47,$98,$FF,$5F
	.dc.b	$57,$8A,$3F,$87,$B5,$E6,$7F,$12,$3F,$CD,$8E,$5B,$1C,$88,$EC,$F8
	.dc.b	$E5,$8E,$0E,$0E,$8F,$3F,$25,$C6,$F8,$7F,$F0,$79,$2B,$F8,$3F,$C8
	.dc.b	$7A,$5F,$59,$0F,$CA,$7B,$FF,$D1,$1F,$F0,$AC,$F1,$7F,$A6,$BF,$E1
	.dc.b	$AF,$86,$FE,$33,$F2,$1F,$C6,$BE,$42,$57,$F1,$48,$FE,$3B,$F2,$1F
	.dc.b	$C7,$BE,$42,$4F,$F1,$88,$FF,$0C,$68,$FC,$EB,$F8,$24,$FF,$A0,$DB
	.dc.b	$C7,$FE,$7A,$6F,$17,$F9,$FD,$FF,$0F,$4F,$F1,$EF,$91,$FE,$56,$47
	.dc.b	$F9,$2F,$C8,$F8,$9F,$01,$0F,$E4,$BF,$21,$E2,$AF,$F1,$6C,$78,$7E
	.dc.b	$17,$06,$7E,$33,$F9,$BF,$C8,$FF,$38,$F3,$25,$7F,$32,$8F,$E7,$78
	.dc.b	$77,$8F,$E6,$3B,$61,$FF,$0A,$ED,$EF,$F8,$7F,$07,$C3,$49,$CE,$89
	.dc.b	$FE,$8B,$F2,$32,$7F,$9F,$BF,$E2,$7F,$DD,$FF,$C9,$1F,$04,$7F,$C9
	.dc.b	$93,$FE,$09,$7F,$C6,$0D,$F8,$AE,$1F,$23,$EF,$F8,$B9,$3E,$07,$E1
	.dc.b	$7C,$4F,$EF,$38,$DE,$47,$E4,$EB,$FC,$8C,$D8,$CC,$7E,$5D,$AF,$E7
	.dc.b	$DA,$FF,$72,$6B,$FE,$A4,$6B,$FE,$CF,$6B,$D8,$E9,$1D,$C8,$FE,$97
	.dc.b	$AD,$C9,$FC,$49,$B1,$FC,$3D,$AF,$FB,$49,$6B,$FC,$63,$E4,$3F,$CE
	.dc.b	$13,$FE,$64,$E9,$C1,$9D,$35,$AE,$FE,$36,$57,$F1,$BB,$F9,$BD,$7F
	.dc.b	$C3,$E3,$FC,$D5,$DF,$F1,$48,$FC,$56,$BF,$15,$AF,$C5,$6B,$F1,$5A
	.dc.b	$FC,$57,$F8,$27,$17,$DF,$BF,$C1,$12,$7F,$CB,$3C,$4F,$8D,$FC,$37
	.dc.b	$C6,$C9,$CB,$F8,$9F,$43,$F4,$1F,$1B,$FA,$1F,$89,$95,$F9,$28,$FC
	.dc.b	$17,$89,$F7,$C9,$1F,$7D,$AF,$DF,$6B,$F7,$DA,$FD,$F6,$BF,$7D,$F5
	.dc.b	$E3,$FE,$57,$8F,$4F,$C4,$7E,$74,$FC,$F7,$C1,$FE,$7F,$E0,$E4,$FB
	.dc.b	$9F,$D1,$7B,$FF,$B9,$F8,$9F,$DD,$7C,$64,$AF,$D5,$C7,$F5,$6F,$B5
	.dc.b	$FE,$1B,$1F,$E4,$46,$BF,$EF,$66,$A0,$E1,$BE,$EF,$E8,$E3,$04,$70
	.dc.b	$48,$FE,$75,$19,$7F,$A1,$71,$FE,$DF,$E0,$FF,$74,$48,$FB,$3E,$D7
	.dc.b	$EF,$7E,$0A,$47,$DC,$7F,$80,$7C,$0F,$F1,$7F,$8E,$FE,$31,$F1,$B2
	.dc.b	$BF,$84,$C7,$BF,$E4,$BC,$C4,$8F,$1B,$5F,$EA,$99,$FF,$2A,$7E,$24
	.dc.b	$65,$A7,$02,$1C,$1F,$B3,$F0,$5F,$C4,$C9,$1F,$92,$F8,$0F,$E2,$DF
	.dc.b	$05,$23,$F2,$AF,$F1,$9F,$1F,$F9,$87,$C6,$7F,$31,$F8,$D9,$5F,$C8
	.dc.b	$A3,$EC,$7A,$9F,$3C,$91,$F3,$1A,$FA,$9A,$FD,$D6,$BF,$75,$AF,$DD
	.dc.b	$CF,$F9,$FB,$F7,$CF,$F7,$8F,$57,$F6,$97,$F5,$7E,$17,$F2,$F3,$F9
	.dc.b	$87,$67,$FB,$5F,$F7,$8F,$EA,$73,$FE,$F8,$6F,$31,$A7,$F5,$6F,$13
	.dc.b	$FA,$D7,$C6,$FF,$4F,$F8,$E9,$7F,$D0,$23,$F2,$DE,$77,$E3,$49,$1F
	.dc.b	$8A,$6B,$F6,$1A,$FE,$6D,$AF,$E6,$DA,$FB,$DF,$E1,$37,$FD,$7A,$FF
	.dc.b	$C4,$D8,$79,$E7,$D8,$7B,$75,$F7,$7A,$3E,$FF,$4E,$63,$FE,$E6,$C7
	.dc.b	$FA,$B3,$FD,$63,$E3,$BF,$B5,$11,$FE,$B7,$F1,$DF,$1F,$D4,$91,$C5
	.dc.b	$F9,$5E,$A7,$CB,$F2,$24,$F5,$7E,$77,$93,$FD,$D7,$E3,$BF,$C0,$FF
	.dc.b	$9E,$FF,$AF,$38,$7F,$C7,$DE,$C2,$31,$EA,$7F,$81,$6B,$F8,$07,$07
	.dc.b	$ED,$C9,$97,$ED,$1A,$FE,$89,$AF,$F0,$56,$BF,$C1,$5A,$FE,$25,$F9
	.dc.b	$7E,$47,$E0,$C9,$1F,$17,$C9,$FC,$37,$87,$23,$E4,$3F,$89,$F1,$3D
	.dc.b	$7F,$19,$EC,$F9,$69,$5F,$E0,$6D,$3F,$D7,$8F,$E5,$F9,$BF,$88,$11
	.dc.b	$FE,$1A,$D7,$F7,$6D,$7F,$8F,$35,$FE,$3C,$D7,$A1,$FC,$53,$99,$E0
	.dc.b	$7E,$74,$91,$F7,$FC,$1F,$D0,$76,$48,$FC,$13,$FA,$3E,$DF,$AB,$F2
	.dc.b	$1F,$5B,$E4,$25,$7A,$E3,$F9,$97,$57,$F2,$92,$47,$F2,$66,$BF,$C5
	.dc.b	$1A,$FF,$37,$6B,$FC,$DD,$AF,$EE,$DF,$CE,$F6,$7E,$D4,$91,$F9,$7E
	.dc.b	$CF,$DC,$73,$48,$FC,$D3,$FC,$07,$A7,$F4,$7F,$BC,$FE,$2A,$5D,$5F
	.dc.b	$8D,$F8,$CF,$C8,$FC,$74,$9F,$AD,$1F,$D3,$7B,$FF,$A2,$12,$3F,$A0
	.dc.b	$B5,$FE,$5E,$D7,$FA,$93,$5F,$EA,$4D,$7F,$8B,$3F,$E0,$0B,$FC,$1F
	.dc.b	$84,$69,$39,$FD,$5D,$DF,$D6,$67,$5C,$18,$C1,$18,$7F,$AF,$69,$F0
	.dc.b	$27,$53,$D7,$87,$F5,$FF,$1D,$FB,$0F,$90,$98,$7E,$56,$3F,$B6,$78
	.dc.b	$5F,$D8,$C9,$1F,$D8,$1A,$FF,$4E,$6B,$FD,$C5,$AF,$F7,$16,$BF,$CD
	.dc.b	$5F,$81,$B7,$93,$53,$4C,$6F,$9B,$97,$EC,$D7,$E1,$5F,$FA,$E6,$1A
	.dc.b	$FD,$EC,$BF,$85,$FC,$6F,$F0,$CF,$90,$95,$FB,$18,$FF,$04,$78,$9F
	.dc.b	$DF,$C9,$1F,$DF,$1A,$FF,$6A,$6B,$FE,$0D,$6B,$FE,$0D,$6B,$FD,$45
	.dc.b	$FB,$91,$ED,$24,$7A,$4F,$AF,$82,$BE,$17,$C6,$49,$D7,$07,$E0,$E3
	.dc.b	$E3,$7C,$5F,$F2,$67,$F9,$3B,$27,$F8,$64,$7F,$88,$FC,$5F,$F0,$D1
	.dc.b	$23,$FC,$2E,$D7,$FC,$0A,$D7,$FC,$50,$D7,$FC,$50,$D7,$FB,$63,$F9
	.dc.b	$7F,$83,$FB,$04,$8F,$3E,$2F,$86,$FB,$9F,$17,$27,$2F,$DF,$F8,$BF
	.dc.b	$E7,$F7,$FD,$02,$A5,$7F,$25,$8F,$F1,$E7,$93,$FE,$33,$24,$7F,$8C
	.dc.b	$1A,$FF,$87,$9A,$FF,$90,$5A,$FF,$90,$5A,$FF,$81,$5F,$D8,$7C,$5F
	.dc.b	$E4,$49,$1E,$AF,$8A,$FC,$A7,$C5,$C8,$F6,$BF,$97,$F8,$AF,$EA,$F9
	.dc.b	$FF,$58,$CE,$57,$F3,$D8,$FF,$2A,$79,$BF,$E4,$B2,$47,$F9,$21,$AF
	.dc.b	$F8,$D9,$AF,$F9,$65,$AF,$F9,$65,$AF,$F8,$75,$FE,$11,$F1,$5F,$A9
	.dc.b	$24,$7E,$23,$C2,$FF,$64,$9F,$EC,$8E,$8F,$F0,$1B,$7A,$7F,$85,$36
	.dc.b	$7F,$84,$BF,$EA,$08,$FF,$33,$78,$5F,$DC,$76,$FE,$EB,$B4,$BF,$EB
	.dc.b	$51,$FE,$71,$F3,$BF,$CD,$04,$8F,$F3,$2B,$5F,$F2,$9B,$5F,$F3,$B3
	.dc.b	$5F,$F3,$B3,$5F,$F1,$D3,$FE,$A4,$C1,$F0,$FF,$C6,$9B,$3E,$F8,$C7
	.dc.b	$C3,$E7,$E1,$E4,$FC,$37,$87,$C3,$FF,$83,$B8,$FF,$C1,$FC,$72,$BF
	.dc.b	$BA,$47,$FA,$2F,$CE,$FF,$41,$12,$3F,$CF,$ED,$7F,$CD,$AD,$7F,$D2
	.dc.b	$0D,$7F,$D2,$0D,$7F,$CA,$8F,$AF,$E1,$FD,$44,$8F,$B1,$F0,$FE,$DF
	.dc.b	$87,$91,$F6,$DF,$81,$F0,$FF,$E2,$6E,$5F,$E8,$1C,$6F,$85,$FF,$6B
	.dc.b	$73,$7F,$8A,$1B,$CD,$F6,$BD,$6F,$F7,$2E,$5F,$C4,$4F,$F5,$0F,$A9
	.dc.b	$FE,$9F,$23,$FE,$9C,$6B,$FE,$8A,$6B,$FE,$AD,$6B,$FE,$AD,$6B,$FE
	.dc.b	$71,$7E,$C7,$C3,$7D,$72,$47,$E5,$6F,$ED,$7C,$34,$8F,$CC,$3F,$77
	.dc.b	$E1,$BD,$B3,$F9,$5C,$FD,$DF,$13,$83,$3F,$15,$FE,$C1,$F6,$7F,$D7
	.dc.b	$04,$C7,$FD,$6A,$D7,$FD,$46,$D7,$FD,$8C,$D7,$FD,$8C,$D7,$FD,$14
	.dc.b	$FE,$47,$E1,$FF,$20,$48,$FD,$77,$C3,$FE,$4F,$E1,$E4,$7E,$C5,$FC
	.dc.b	$B7,$C3,$FA,$BB,$BD,$7D,$D2,$BF,$A1,$69,$F6,$67,$FD,$B5,$EE,$FF
	.dc.b	$B4,$49,$3F,$EC,$E6,$BF,$EB,$C6,$BF,$ED,$F6,$BF,$ED,$F6,$BF,$EA
	.dc.b	$37,$F5,$5F,$0B,$FA,$92,$47,$F0,$7F,$85,$FF,$02,$FE,$F5,$FE,$9E
	.dc.b	$27,$C5,$63,$E2,$BE,$9F,$89,$F5,$3C,$49,$87,$AB,$95,$3F,$EF,$3F
	.dc.b	$81,$FF,$75,$12,$7F,$DC,$ED,$7F,$DA,$AD,$7F,$DF,$0D,$7F,$DF,$0D
	.dc.b	$7F,$D7,$8F,$FA,$ED,$FF,$5E,$7A,$8C,$BF,$D4,$1D,$BF,$E2,$56,$CF
	.dc.b	$F8,$91,$B3,$BF,$88,$F7,$FC,$3C,$AF,$ED,$F5,$FE,$A0,$7F,$C1,$77
	.dc.b	$CF,$E1,$BF,$27,$E3,$7E,$53,$C6,$95,$F0,$63,$FC,$64,$BE,$29,$F0
	.dc.b	$BE,$78,$FF,$69,$46,$1C,$1B,$8A,$D3,$31,$B7,$FD,$F7,$1A,$8C,$47
	.dc.b	$C3,$6B,$DA,$34,$47,$9B,$E0,$D3,$E4,$7C,$37,$E0,$09,$3F,$0B,$E1
	.dc.b	$BF,$07,$F0,$72,$3E,$2B,$F8,$6F,$7F,$F6,$BE,$67,$ED,$BC,$C9,$5F
	.dc.b	$83,$8F,$C5,$FB,$7F,$8C,$F6,$24,$FC,$E8,$FC,$7F,$AD,$F9,$0F,$52
	.dc.b	$4F,$D2,$8F,$E0,$3E,$67,$F0,$2E,$A9,$3F,$8F,$8F,$9F,$C2,$F5,$3E
	.dc.b	$21,$27,$E4,$B5,$FA,$51,$B7,$24,$33,$1B,$7E,$63,$5F,$C6,$7A,$5F
	.dc.b	$A6,$23,$F8,$8F,$3F,$F5,$3E,$5C,$8F,$C5,$BF,$AC,$F1,$3F,$8D,$F9
	.dc.b	$FF,$C7,$3C,$F9,$5F,$A9,$8F,$C4,$FD,$CF,$E0,$A0,$D2,$3F,$86,$6B
	.dc.b	$F8,$A6,$BF,$8F,$6B,$FA,$76,$BF,$A5,$F0,$7F,$7E,$47,$F4,$3D,$5F
	.dc.b	$C0,$B4,$91,$FA,$47,$FD,$13,$FB,$DF,$FA,$A7,$CA,$FE,$A1,$E7,$FF
	.dc.b	$36,$F3,$E5,$FF,$03,$8F,$5F,$C5,$79,$C9,$1E,$76,$BE,$76,$BE,$76
	.dc.b	$A2,$66,$8C,$C6,$1F,$CE,$61,$FE,$AA,$23,$E8,$F8,$9F,$4F,$C4,$CA
	.dc.b	$FF,$36,$D2,$3D,$BF,$13,$FD,$43,$D3,$FE,$A3,$E9,$CB,$FE,$3D,$1F
	.dc.b	$5B,$0F,$A4,$48,$FA,$4D,$7E,$93,$5F,$A4,$D7,$D8,$7D,$E7,$0F,$B9
	.dc.b	$F5,$FC,$4F,$D0,$24,$7A,$3E,$27,$E9,$FC,$4C,$8F,$5B,$F5,$BE,$27
	.dc.b	$FB,$4F,$AB,$FD,$AB,$D5,$95,$FC,$DE,$3F,$1F,$AF,$E2,$C9,$1F,$8B
	.dc.b	$6B,$F8,$B6,$BF,$8B,$6B,$EA,$E0,$F9,$4F,$12,$1F,$CD,$FC,$A7,$03
	.dc.b	$E2,$7F,$16,$49,$FA,$7F,$13,$F8,$EF,$88,$91,$F5,$9F,$C8,$FC,$4F
	.dc.b	$F8,$83,$F4,$17,$B5,$7D,$A9,$7F,$DB,$27,$2F,$F8,$7F,$47,$3F,$EA
	.dc.b	$49,$1F,$A9,$6B,$FA,$96,$BF,$A9,$6B,$EF,$6B,$F9,$47,$F2,$3F,$11
	.dc.b	$FA,$62,$47,$E3,$BE,$27,$F5,$3F,$11,$23,$F2,$2F,$EB,$3E,$27,$D9
	.dc.b	$F1,$9E,$DF,$76,$57,$AE,$3F,$82,$F4,$7F,$01,$24,$7F,$01,$6B,$FC
	.dc.b	$05,$AF,$F0,$16,$BF,$5D,$AF,$EB,$1F,$D4,$FC,$4F,$EF,$89,$1F,$A6
	.dc.b	$F8,$AF,$E0,$3F,$13,$23,$F5,$2F,$F0,$4F,$8A,$FA,$9F,$03,$F5,$7D
	.dc.b	$F9,$5F,$4E,$3F,$8E,$F3,$7F,$19,$24,$7F,$19,$6B,$FC,$65,$AF,$F1
	.dc.b	$96,$BF,$8F,$6B,$FC,$11,$FE,$03,$F1,$5F,$AC,$F8,$A7,$C1,$7F,$09
	.dc.b	$ED,$7E,$F9,$FE,$64,$C1,$18,$9F,$E3,$B7,$C5,$E1,$5F,$EA,$59,$E3
	.dc.b	$F1,$39,$BE,$33,$93,$FD,$50,$BC,$C7,$53,$E0,$BF,$30,$49,$FC,$B7
	.dc.b	$C1,$78,$1F,$E2,$79,$D7,$FC,$22,$76,$F8,$8F,$06,$7E,$28,$FE,$49
	.dc.b	$AF,$B9,$F1,$E9,$AD,$3C,$9F,$15,$DE,$79,$EB,$E4,$D3,$18,$71,$8F
	.dc.b	$3E,$9C,$7E,$2B,$97,$C4,$CC,$3B,$F0,$7C,$49,$ED,$F1,$52,$72,$7C
	.dc.b	$47,$C5,$F1,$1F,$B7,$F8,$8F,$DC,$7C,$44,$AF,$DB,$47,$9F,$E2,$64
	.dc.b	$72,$E2,$FA,$F1,$EB,$F8,$8F,$67,$C4,$CA,$D7,$1A,$3E,$07,$C2,$7F
	.dc.b	$05,$F8,$8F,$E0,$DF,$13,$2F,$F8,$24,$7C,$5F,$87,$F8,$DF,$03,$27
	.dc.b	$9C,$7C,$9F,$82,$F9,$5E,$F4,$9E,$F1,$FC,$47,$E2,$3F,$89,$7C,$4C
	.dc.b	$9F,$E2,$11,$D5,$E6,$43,$ED,$F0,$7E,$4B,$E0,$7B,$47,$AB,$0E,$CE
	.dc.b	$B7,$63,$DE,$3F,$8E,$63,$DD,$FB,$7E,$0B,$2F,$F0,$4F,$91,$27,$F9
	.dc.b	$03,$FE,$0C,$F2,$24,$7F,$B0,$7D,$67,$BB,$FE,$58,$6C,$7F,$27,$9D
	.dc.b	$BF,$92,$3F,$9C,$7B,$BB,$33,$E0,$FE,$46,$3F,$C2,$F3,$FE,$15,$23
	.dc.b	$FE,$15,$6B,$FE,$15,$6B,$FE,$15,$6B,$FD,$2D,$AF,$F8,$01,$E3,$1D
	.dc.b	$FD,$44,$78,$BB,$FF,$A9,$79,$96,$F8,$1E,$B3,$E0,$7B,$59,$7C,$E6
	.dc.b	$4F,$76,$5F,$D1,$9F,$C6,$FB,$FF,$AB,$F8,$09,$18,$3F,$23,$EE,$FE
	.dc.b	$BC,$8F,$E5,$3D,$EF,$EA,$DF,$11,$23,$FA,$A3,$FD,$67,$E2,$3F,$6D
	.dc.b	$84,$9E,$87,$E1,$AB,$FA,$CB,$F8,$3B,$EA,$D3,$B1,$D0,$87,$E1,$2F
	.dc.b	$B3,$FA,$B9,$7B,$9F,$C2,$9F,$F5,$19,$1C,$7A,$F2,$78,$3B,$F4,$FF
	.dc.b	$B8,$7C,$44,$9F,$ED,$CF,$F7,$3F,$88,$E9,$72,$7C,$0F,$7F,$5B,$A4
	.dc.b	$7B,$70,$BC,$BF,$C6,$7C,$06,$7F,$EB,$18,$FE,$29,$D6,$CF,$C1,$87
	.dc.b	$E8,$1F,$03,$ED,$F4,$7D,$9E,$84,$8C,$1F,$07,$CB,$F8,$04,$7E,$27
	.dc.b	$99,$FE,$0B,$F8,$8F,$F0,$67,$C4,$49,$FF,$05,$47,$C7,$C2,$47,$43
	.dc.b	$EB,$79,$3C,$DF,$B1,$85,$D3,$D9,$AB,$2E,$D5,$FE,$15,$CB,$B3,$FB
	.dc.b	$AC,$BE,$A6,$32,$3E,$C3,$F5,$FC,$AE,$9D,$BF,$D6,$F1,$67,$8B,$FE
	.dc.b	$B7,$1D,$DF,$49,$DC,$87,$73,$F0,$9E,$1F,$E0,$C8,$FE,$1F,$C4,$FC
	.dc.b	$37,$44,$8F,$C5,$3F,$89,$EB,$FF,$19,$7C,$47,$F6,$3F,$51,$FA,$1F
	.dc.b	$FA,$14,$FF,$43,$73,$7F,$AB,$F0,$FC,$E6,$3F,$91,$23,$F9,$5E,$6E
	.dc.b	$CF,$E8,$72,$BC,$9F,$2B,$3E,$F3,$FD,$3B,$2E,$EF,$87,$2E,$F9,$FE
	.dc.b	$77,$AF,$F3,$DC,$72,$BF,$45,$1F,$A1,$E4,$FD,$15,$49,$FD,$34,$73
	.dc.b	$F8,$9E,$9F,$13,$27,$C1,$8F,$FA,$45,$E6,$43,$F7,$18,$7F,$76,$FD
	.dc.b	$F7,$FC,$FB,$1F,$BB,$C3,$B3,$FB,$DC,$7C,$4F,$4C,$7B,$3F,$DD,$B0
	.dc.b	$63,$E3,$8F,$EC,$98,$78,$5F,$E9,$6E,$9C,$C7,$86,$BC,$5E,$24,$9E
	.dc.b	$2F,$97,$E1,$7E,$43,$23,$DA,$FC,$B8,$F5,$79,$5C,$E7,$C3,$58,$79
	.dc.b	$90,$F1,$D7,$57,$99,$0F,$63,$D5,$F5,$CF,$37,$D6,$FC,$9F,$2B,$D1
	.dc.b	$1D,$0E,$DC,$3F,$C1,$F2,$FE,$6D,$1D,$0E,$D1,$FE,$1B,$CF,$FD,$27
	.dc.b	$1F,$2F,$2F,$A9,$F0,$1F,$57,$E2,$A4,$FC,$C8,$FB,$1E,$4F,$D9,$F8
	.dc.b	$99,$3F,$42,$3F,$1D,$A7,$E3,$F4,$93,$F8,$48,$FC,$07,$C0,$F6,$8E
	.dc.b	$87,$D8,$7A,$BF,$6A,$FA,$BF,$6A,$FA,$BF,$8E,$BE,$AF,$E0,$EF,$F1
	.dc.b	$9F,$13,$F8,$DF,$88,$93,$F8,$48,$FC,$87,$C4,$48,$FC,$3B,$F9,$3F
	.dc.b	$88,$FD,$36,$92,$7F,$36,$FE,$A7,$5F,$CC,$7C,$47,$E6,$7D,$49,$5F
	.dc.b	$90,$8F,$CE,$7A,$72,$3F,$26,$FE,$7F,$D3,$FD,$96,$F2,$7F,$4E,$FE
	.dc.b	$D7,$83,$AC,$73,$21,$F9,$AF,$87,$E8,$FE,$72,$EF,$2F,$DF,$C3,$A4
	.dc.b	$FE,$97,$0B,$3F,$7F,$0F,$D8,$7C,$3C,$AF,$D3,$BF,$E6,$8F,$DF,$BF
	.dc.b	$D0,$1E,$79,$FF,$3E,$1F,$E7,$87,$1F,$0D,$BF,$87,$FE,$19,$D1,$30
	.dc.b	$FE,$02,$FF,$0F,$E5,$E9,$F1,$12,$7F,$CB,$18,$A3,$C1,$F0,$FE,$1F
	.dc.b	$88,$95,$A7,$C9,$F1,$1F,$C6,$7A,$A4,$FF,$0E,$7F,$8E,$75,$7A,$7E
	.dc.b	$23,$D5,$F1,$12,$BB,$C7,$FA,$4F,$8D,$F5,$0E,$8C,$36,$BF,$CE,$B3
	.dc.b	$A9,$CF,$27,$F2,$9E,$DF,$E5,$5D,$52,$FF,$8F,$C7,$C9,$F8,$9F,$95
	.dc.b	$F1,$32,$7D,$F1,$F3,$7E,$27,$E7,$7C,$44,$9F,$87,$1F,$CD,$7B,$E4
	.dc.b	$7F,$2A,$7F,$9C,$77,$FD,$5F,$89,$FA,$DF,$0D,$2B,$E5,$C7,$B7,$E9
	.dc.b	$DD,$17,$A2,$57,$B2,$72,$FF,$36,$F0,$BF,$A2,$11,$FE,$73,$C5,$F8
	.dc.b	$3F,$89,$91,$F6,$1F,$C3,$7C,$57,$FA,$F6,$1F,$32,$F8,$CC,$36,$3F
	.dc.b	$D3,$E6,$A3,$87,$DF,$FF,$A4,$E3,$FA,$3F,$3F,$F5,$52,$3F,$D6,$7F
	.dc.b	$46,$FC,$9F,$82,$F8,$BF,$CB,$7C,$64,$AF,$C5,$C7,$4E,$9F,$0E,$52
	.dc.b	$7C,$39,$77,$7F,$D8,$FD,$8E,$1C,$3F,$EC,$53,$17,$F2,$CF,$99,$D1
	.dc.b	$FA,$5F,$8C,$FD,$37,$C0,$CC,$3F,$37,$1E,$17,$DF,$C1,$F0,$5F,$F4
	.dc.b	$23,$70,$FF,$41,$65,$E0,$F8,$35,$7F,$B8,$F8,$1E,$4F,$F8,$1D,$B1
	.dc.b	$F0,$E3,$1E,$57,$DF,$EE,$93,$F7,$5F,$C1,$75,$F8,$E3,$C3,$C3,$F0
	.dc.b	$DF,$F4,$B6,$5E,$15,$9F,$C3,$EC,$FC,$57,$44,$9F,$C4,$47,$FB,$B3
	.dc.b	$F7,$FF,$F3,$7D,$7F,$94,$DF,$F3,$3C,$FF,$B4,$DA,$FF,$AE,$9D,$ED
	.dc.b	$EE,$86,$1E,$17,$F0,$87,$C8,$C7,$D1,$7F,$EE,$FD,$BF,$83,$4F,$E0
	.dc.b	$FC,$9F,$88,$7C,$5E,$EF,$D0,$75,$F9,$70,$95,$F9,$F8,$E7,$97,$87
	.dc.b	$8B,$E9,$36,$B3,$5F,$0B,$9D,$D3,$FA,$BE,$59,$3F,$A9,$8F,$85,$F1
	.dc.b	$52,$3E,$03,$F1,$7E,$37,$9C,$7B,$0F,$A2,$7E,$4B,$AB,$EE,$78,$1F
	.dc.b	$77,$CB,$93,$F9,$F8,$FE,$2D,$F8,$CF,$C8,$8F,$AA,$DE,$33,$74,$FF
	.dc.b	$41,$CF,$53,$E5,$10,$FB,$F8,$74,$F8,$EE,$BB,$49,$FE,$3F,$7F,$63
	.dc.b	$0F,$B9,$F1,$DF,$7F,$E3,$65,$7D,$D8,$F6,$F1,$7C,$A8,$FC,$CB,$5F
	.dc.b	$BB,$97,$E6,$7C,$CF,$CD,$79,$B2,$7F,$79,$1E,$EE,$1F,$7F,$0C,$9F
	.dc.b	$1C,$76,$D7,$F3,$1A,$7E,$A3,$1F,$E1,$B3,$2B,$F5,$EF,$DA,$F4,$9D
	.dc.b	$6F,$8E,$FC,$C7,$C5,$CB,$FC,$B4,$75,$7E,$36,$3D,$5F,$94,$7E,$E3
	.dc.b	$0E,$CF,$CA,$3E,$3F,$07,$D6,$23,$F2,$B8,$3A,$3F,$80,$38,$12,$9D
	.dc.b	$1F,$93,$BE,$B1,$E1,$9F,$DA,$F9,$BF,$CB,$7F,$02,$7C,$25,$3F,$94
	.dc.b	$C7,$6A,$E7,$5F,$4B,$83,$F2,$3E,$07,$EB,$79,$59,$76,$CF,$F6,$53
	.dc.b	$FB,$11,$8F,$4C,$3F,$93,$BC,$B8,$BE,$8D,$7E,$9E,$7E,$2E,$5E,$73
	.dc.b	$C2,$FF,$42,$C3,$B3,$FC,$73,$0E,$D7,$FE,$7E,$8F,$0B,$5D,$35,$FC
	.dc.b	$49,$98,$FE,$70,$FF,$A1,$9A,$F2,$35,$1C,$63,$31,$D0,$C9,$0F,$13
	.dc.b	$E4,$7D,$77,$F1,$3C,$8F,$81,$18,$3F,$3B,$F2,$58,$7E,$D4,$FF,$0B
	.dc.b	$E9,$FF,$06,$7E,$BE,$F4,$A9,$2B,$F9,$27,$C8,$F1,$D7,$E4,$61,$FC
	.dc.b	$4F,$1C,$3E,$BA,$FE,$8F,$8F,$F2,$FF,$91,$91,$FC,$B5,$FE,$65,$F2
	.dc.b	$3C,$23,$DC,$7F,$4E,$BC,$3F,$28,$E6,$7F,$35,$F9,$1F,$E7,$44,$7F
	.dc.b	$9C,$7C,$87,$0D,$FF,$11,$57,$83,$FB,$DE,$5F,$D0,$7E,$3E,$61,$FC
	.dc.b	$FD,$FE,$89,$F1,$FB,$8F,$EC,$37,$F1,$3E,$13,$E2,$FC,$4C,$9F,$F1
	.dc.b	$6F,$1A,$7F,$A6,$7C,$87,$F4,$DE,$A9,$5F,$D2,$E3,$A3,$ED,$3F,$C0
	.dc.b	$9D,$1F,$D9,$75,$91,$83,$FA,$97,$C7,$FF,$57,$23,$FD,$57,$E3,$F9
	.dc.b	$F3,$7E,$63,$AF,$F9,$77,$C8,$3D,$17,$67,$F7,$1F,$56,$63,$FE,$13
	.dc.b	$8F,$EC,$5F,$21,$FD,$6F,$1F,$D8,$F8,$EF,$73,$FD,$B7,$18,$6F,$77
	.dc.b	$F7,$E9,$FE,$2B,$5E,$28,$FC,$16,$1F,$E1,$FF,$5F,$FC,$00,$47,$FC
	.dc.b	$45,$EB,$77,$9E,$58,$7F,$B7,$71,$ED,$F2,$19,$B6,$5E,$38,$C3,$E5
	.dc.b	$F9,$0F,$04,$FE,$70,$FF,$17,$7B,$3F,$E0,$F2,$3F,$E3,$1F,$5F,$8E
	.dc.b	$3F,$EC,$6C,$7D,$9F,$1D,$83,$3D,$9D,$2F,$FC,$87,$DB,$FE,$19,$F7
	.dc.b	$A4,$7F,$8E,$DF,$F0,$E7,$B1,$D3,$BA,$4F,$F8,$8B,$1F,$91,$1E,$0F
	.dc.b	$8E,$E9,$D3,$F5,$BA,$1E,$D7,$F8,$94,$98,$7F,$91,$FD,$FF,$3F,$C7
	.dc.b	$7A,$3E,$3A,$4F,$5E,$7F,$93,$3D,$37,$F9,$1D,$AF,$8A,$FC,$19,$FD
	.dc.b	$8E,$BF,$C6,$B1,$9F,$23,$73,$FD,$D7,$A7,$F0,$1F,$1D,$F1,$BE,$3A
	.dc.b	$67,$DC,$FA,$7D,$09,$F0,$46,$71,$F0,$7A,$7D,$3D,$39,$39,$CB,$FB
	.dc.b	$76,$5F,$8D,$F8,$EF,$A6,$49,$ED,$1F,$91,$F8,$EF,$F4,$F7,$EB,$FE
	.dc.b	$B9,$FF,$4D,$A8,$C9,$AB,$F5,$FB,$B2,$BC,$0F,$BB,$D9,$FC,$0C,$FE
	.dc.b	$0B,$E3,$E5,$78,$37,$EC,$FF,$92,$B8,$BF,$2B,$EC,$FE,$23,$E3,$A4
	.dc.b	$EE,$3E,$4F,$B9,$F2,$BD,$D9,$3E,$F8,$FC,$7F,$C7,$7E,$43,$E3,$A4
	.dc.b	$F7,$8F,$C9,$FC,$77,$E5,$3E,$3A,$4F,$8E,$3E,$AF,$C1,$7D,$6F,$81
	.dc.b	$93,$F2,$E3,$F3,$7B,$7E,$93,$F8,$03,$FD,$BE,$7F,$B8,$1F,$DD,$4F
	.dc.b	$CC,$8F,$F3,$8B,$ED,$3F,$5A,$E3,$C7,$FE,$02,$DE,$0F,$DF,$2F,$8D
	.dc.b	$FC,$3F,$C1,$FE,$23,$E0,$E6,$1F,$76,$3C,$5F,$1B,$23,$C0,$F9,$7E
	.dc.b	$37,$CD,$F1,$92,$7C,$8F,$A7,$E3,$7F,$27,$F0,$FF,$94,$F8,$79,$5F
	.dc.b	$89,$8F,$77,$C5,$7B,$FE,$2E,$4F,$B6,$3E,$17,$C1,$C8,$F8,$0F,$C5
	.dc.b	$F8,$3F,$CF,$FC,$44,$9F,$CA,$3F,$A2,$F8,$8E,$B1,$CC,$87,$C7,$F1
	.dc.b	$9C,$7F,$CB,$23,$A9,$0D,$3F,$E1,$27,$E8,$5F,$77,$E8,$79,$5F,$B5
	.dc.b	$C3,$1C,$DF,$E2,$08,$FD,$C6,$3F,$6F,$E0,$FE,$E7,$C0,$CC,$3E,$D4
	.dc.b	$7D,$FF,$81,$91,$F7,$5F,$C1,$7B,$FF,$BC,$F8,$69,$3F,$B1,$7F,$7F
	.dc.b	$F0,$FF,$87,$F8,$19,$3F,$85,$7F,$15,$EF,$FE,$2F,$DF,$93,$F8,$97
	.dc.b	$F1,$DE,$DF,$F0,$8F,$87,$FE,$13,$F0,$F2,$BF,$7D,$1C,$9E,$84,$3E
	.dc.b	$2D,$F4,$7E,$3D,$DE,$1F,$C8,$E1,$D6,$7F,$84,$61,$D5,$F9,$3E,$7B
	.dc.b	$1F,$D0,$7B,$BF,$D1,$3F,$80,$7F,$8D,$1F,$C6,$F9,$FF,$D1,$77,$F6
	.dc.b	$63,$F9,$13,$FD,$31,$FE,$26,$4F,$6F,$F9,$1F,$C3,$FF,$2A,$F9,$EF
	.dc.b	$EB,$F4,$E8,$46,$1F,$E5,$D3,$E2,$9F,$E3,$37,$E2,$F7,$B1,$1E,$1C
	.dc.b	$3C,$DE,$E4,$8D,$C7,$81,$FE,$55,$8F,$53,$B6,$AC,$79,$F0,$7E,$DB
	.dc.b	$5B,$E5,$7F,$73,$D3,$FC,$CB,$8F,$43,$C7,$9F,$77,$C4,$7F,$33,$C3
	.dc.b	$DC,$C3,$07,$AB,$D8,$F9,$9E,$C4,$8C,$1E,$EF,$53,$E8,$11,$F8,$3E
	.dc.b	$AF,$C9,$F8,$79,$1E,$C7,$E6,$7C,$47,$D7,$C2,$4F,$43,$EB,$7A,$5C
	.dc.b	$DF,$B1,$85,$DF,$FC,$2B,$AF,$FA,$8E,$7A,$10,$FD,$6F,$0F,$F8,$2F
	.dc.b	$7F,$EB,$EB,$2F,$C3,$7A,$72,$BE,$BB,$F8,$9F,$43,$F1,$5E,$8F,$E2
	.dc.b	$FC,$C9,$5F,$72,$3F,$09,$F0,$F2,$3E,$BB,$F8,$7F,$88,$E5,$1D,$08
	.dc.b	$7E,$46,$FA,$BF,$86,$BE,$AF,$E5,$6F,$A3,$EA,$BB,$AF,$F0,$5C,$D8
	.dc.b	$FE,$7F,$CF,$FD,$07,$95,$30,$FC,$8C,$7E,$8F,$CA,$FD,$27,$8D,$27
	.dc.b	$F2,$B1,$F9,$DF,$88,$91,$F9,$07,$F4,$1F,$11,$D6,$39,$90,$FD,$57
	.dc.b	$91,$D5,$FC,$FD,$F5,$7F,$5D,$7D,$1F,$A7,$77,$4F,$56,$1F,$B7,$C3
	.dc.b	$B5,$7E,$D7,$1C,$3E,$63,$A9,$0E,$A7,$9C,$87,$63,$F5,$33,$FC,$0F
	.dc.b	$C3,$91,$FB,$47,$F8,$37,$89,$FC,$1F,$B7,$FC,$97,$F6,$3F,$BE,$F3
	.dc.b	$CD,$E9,$FD,$86,$1F,$B4,$F8,$6F,$E0,$84,$7F,$6F,$F0,$FD,$9F,$E1
	.dc.b	$D9,$5D,$BF,$2F,$9F,$79,$FF,$0E,$E3,$DC,$ED,$9F,$38,$FC,$CE,$36
	.dc.b	$FF,$01,$8F,$9B,$BF,$CF,$D7,$2F,$9C,$7A,$B9,$E4,$76,$7D,$BC,$DF
	.dc.b	$C6,$FE,$1A,$4F,$F0,$E7,$F8,$F7,$C3,$F5,$8E,$64,$3F,$B0,$6B,$FE
	.dc.b	$9D,$7A,$10,$CD,$F6,$7A,$10,$FD,$9D,$76,$7E,$0F,$17,$D0,$23,$F1
	.dc.b	$35,$FA,$7A,$48,$F8,$EF,$D6,$AF,$E6,$9F,$0D,$FC,$D7,$E1,$65,$7F
	.dc.b	$28,$8E,$6F,$A3,$0B,$BF,$FA,$73,$36,$5D,$6B,$DB,$95,$AD,$65,$DA
	.dc.b	$7F,$99,$E5,$D9,$FE,$1B,$97,$E2,$30,$EB,$E0,$7F,$A8,$F2,$67,$9B
	.dc.b	$F4,$90,$EE,$7E,$0B,$1E,$CF,$F1,$8C,$7B,$65,$FB,$FC,$BF,$2B,$AC
	.dc.b	$AF,$C4,$BF,$98,$8F,$CC,$D7,$E6,$BC,$89,$5F,$8E,$8F,$EB,$51,$23
	.dc.b	$FA,$6B,$FD,$77,$E1,$79,$47,$42,$1F,$80,$AE,$AF,$E6,$27,$AB,$F8
	.dc.b	$4A,$E8,$FF,$30,$8B,$87,$F1,$7B,$FD,$69,$FA,$EF,$84,$95,$FA,$28
	.dc.b	$FD,$8F,$C2,$FE,$CB,$E1,$24,$FE,$9A,$3F,$B9,$57,$F7,$36,$47,$F6
	.dc.b	$68,$EA,$F3,$21,$F9,$8F,$85,$EA,$FE,$C2,$BA,$BF,$9C,$AE,$AF,$F7
	.dc.b	$07,$AB,$FD,$2A,$BF,$82,$FC,$27,$F0,$6F,$85,$93,$FB,$A8,$FE,$13
	.dc.b	$F0,$B2,$3F,$7A,$FF,$B6,$7D,$14,$39,$3F,$C1,$99,$7F,$83,$73,$93
	.dc.b	$FE,$00,$8E,$EF,$52,$1F,$E7,$7A,$E7,$F2,$3D,$E3,$F6,$B8,$5D,$BB
	.dc.b	$42,$BC,$38,$F8,$23,$FC,$1B,$8F,$81,$FE,$CF,$8F,$9D,$FF,$40,$1C
	.dc.b	$4B,$5E,$07,$FD,$84,$FB,$3E,$1B,$C1,$7F,$A1,$C6,$D7,$18,$F0,$3F
	.dc.b	$DD,$30,$F0,$63,$FD,$6F,$1F,$07,$C3,$4A,$EA,$1F,$27,$C3,$79,$7E
	.dc.b	$1B,$CD,$F0,$D2,$B7,$1F,$E3,$7D,$E4,$7F,$8A,$9F,$F1,$D7,$07,$28
	.dc.b	$EB,$87,$77,$E6,$0C,$3B,$3E,$2C,$3B,$3A,$C3,$AB,$FE,$0E,$C2,$E5
	.dc.b	$FD,$D7,$1F,$95,$F0,$B2,$BD,$6F,$CD,$F8,$6F,$9D,$F0,$DF,$3F,$E1
	.dc.b	$A5,$7C,$08,$FF,$28,$70,$C8,$FF,$21,$BF,$EF,$CF,$D3,$3E,$EC,$7A
	.dc.b	$2F,$B3,$EA,$C3,$B3,$CC,$F9,$78,$76,$7D,$F8,$75,$7F,$C4,$F8,$5C
	.dc.b	$BF,$D2,$58,$F6,$7B,$EE,$F6,$FF,$37,$7F,$EC,$93,$EB,$38,$9D,$B7
	.dc.b	$8F,$75,$FE,$54,$AE,$0D,$AF,$5F,$E6,$7F,$ED,$69,$E8,$D8,$F0,$FC
	.dc.b	$57,$39,$FD,$5D,$71,$BF,$1F,$E6,$9E,$76,$9B,$9F,$C5,$55,$AF,$14
	.dc.b	$7A,$8C,$94,$EB,$15,$A5,$8E,$2F,$B2,$34,$F2,$39,$46,$9E,$26,$23
	.dc.b	$71,$98,$E0,$78,$D7,$B9,$E7,$99,$3F,$C2,$B0,$FA,$73,$88,$F9,$73
	.dc.b	$67,$E5,$F1,$5C,$3C,$D5,$C5,$AF,$CE,$E2,$FE,$01,$F1,$D2,$3F,$62
	.dc.b	$FF,$03,$F8,$AF,$E0,$9F,$17,$27,$F6,$AF,$F0,$7F,$8B,$FC,$A7,$77
	.dc.b	$E5,$7B,$65,$77,$3F,$D5,$1C,$7F,$C2,$96,$1D,$5F,$DF,$BF,$C4,$7E
	.dc.b	$2F,$B4,$7E,$D3,$1B,$B7,$F1,$CE,$FF,$E5,$79,$F8,$2B,$B7,$C3,$32
	.dc.b	$67,$E2,$9F,$F7,$7C,$78,$7E,$2F,$F8,$F7,$C5,$C9,$FE,$1F,$1F,$C8
	.dc.b	$7E,$2E,$47,$F1,$27,$F9,$27,$C5,$FE,$C7,$9F,$F6,$5E,$14,$AF,$0E
	.dc.b	$3F,$C5,$DE,$64,$3F,$86,$E1,$FC,$B7,$E2,$7A,$C7,$F1,$0C,$6E,$FE
	.dc.b	$D8,$65,$DD,$EB,$8F,$77,$D7,$97,$F4,$DF,$A7,$7C,$5D,$E1,$95,$FC
	.dc.b	$DA,$72,$DE,$EE,$1D,$72,$30,$7A,$BC,$4F,$E1,$DE,$2C,$8C,$1C,$F1
	.dc.b	$F0,$11,$ED,$DD,$E3,$C3,$A4,$7F,$57,$C7,$F6,$CF,$F5,$37,$F8,$53
	.dc.b	$D0,$FA,$B7,$DD,$FA,$F8,$7B,$6F,$B4,$73,$C3,$0F,$A0,$ED,$8F,$83
	.dc.b	$6F,$C3,$63,$DA,$79,$9F,$5B,$1E,$D3,$FC,$7B,$1F,$95,$F1,$7F,$2F
	.dc.b	$E2,$65,$FC,$98,$F9,$DF,$17,$F3,$FE,$2E,$4F,$CD,$8F,$E6,$9E,$47
	.dc.b	$F3,$5F,$26,$4F,$E1,$23,$AB,$CC,$87,$C5,$F8,$BE,$AF,$D7,$BE,$AF
	.dc.b	$C9,$BE,$AF,$E3,$AF,$AB,$F9,$CB,$FC,$17,$C5,$FE,$0F,$E2,$A4,$FE
	.dc.b	$06,$3F,$0D,$F1,$7F,$87,$F8,$59,$3F,$85,$8F,$E9,$9E,$6F,$F4,$DF
	.dc.b	$3A,$4F,$E7,$23,$AB,$CC,$87,$DD,$F8,$AE,$AF,$E3,$EF,$AB,$F8,$1B
	.dc.b	$EA,$FE,$9A,$FA,$BF,$B4,$BF,$CD,$7C,$4F,$E6,$FE,$2A,$4F,$E6,$63
	.dc.b	$F3,$DF,$13,$F9,$FF,$87,$93,$F9,$D8,$FE,$C9,$E6,$C8,$FD,$9B,$FD
	.dc.b	$9F,$CC,$EB,$1C,$C8,$7E,$57,$E1,$FA,$BF,$A7,$BE,$AF,$E6,$6F,$A3
	.dc.b	$FC,$2A,$EF,$4F,$F6,$6C,$3F,$65,$84,$AF,$D8,$3F,$B5,$F8,$7E,$B1
	.dc.b	$FA,$3C,$7A,$DF,$E5,$B1,$E7,$E8,$FF,$91,$BC,$CF,$E3,$59,$E3,$0F
	.dc.b	$E5,$39,$FF,$92,$F0,$FE,$B0,$7F,$23,$1F,$0D,$93,$53,$BF,$87,$8F
	.dc.b	$F8,$E2,$51,$CF,$E1,$BF,$C1,$9E,$8C,$AF,$E3,$6F,$F8,$3F,$D1,$E7
	.dc.b	$1C,$4A,$7F,$94,$ED,$F0,$FD,$A3,$18,$59,$F3,$61,$D6,$7F,$99,$E1
	.dc.b	$67,$FC,$1F,$87,$B7,$E2,$BD,$DF,$0B,$2F,$D1,$1F,$07,$E1,$FE,$17
	.dc.b	$C4,$49,$F6,$47,$F8,$97,$D7,$91,$FC,$D1,$FF,$14,$7A,$9F,$27,$E1
	.dc.b	$FB,$4F,$43,$E2,$3D,$5F,$8D,$7D,$5F,$75,$F4,$7F,$A5,$DD,$CF,$FC
	.dc.b	$4F,$87,$D5,$F8,$7F,$AD,$F0,$F2,$FE,$84,$7D,$9F,$87,$91,$F4,$DF
	.dc.b	$B9,$F0,$FF,$E3,$DF,$6E,$4F,$F4,$C7,$FC,$81,$EC,$FE,$07,$E1,$FB
	.dc.b	$4F,$43,$EE,$3D,$1F,$A9,$77,$3F,$C3,$61,$D2,$7F,$B1,$E1,$67,$FC
	.dc.b	$7F,$87,$59,$E6,$43,$EF,$7C,$37,$3E,$CF,$F4,$DF,$B0,$B6,$59,$BF
	.dc.b	$F4,$84,$32,$ED,$7F,$9D,$F7,$59,$E2,$FF,$C0,$B9,$FE,$77,$E1,$E4
	.dc.b	$7E,$5D,$FD,$07,$C3,$7E,$87,$E1,$FF,$45,$F0,$92,$BF,$39,$1F,$E6
	.dc.b	$5F,$80,$91,$FD,$D1,$FF,$34,$7C,$0F,$28,$FD,$15,$D7,$F3,$5F,$09
	.dc.b	$D5,$FD,$36,$1D,$5F,$CF,$61,$D1,$FF,$05,$E1,$77,$FF,$33,$E3,$FB
	.dc.b	$5F,$87,$FF,$60,$7D,$5F,$F9,$4E,$FF,$D6,$4D,$E3,$F9,$13,$AF,$86
	.dc.b	$95,$FE,$AC,$AF,$F5,$79,$FE,$1A,$7F,$C3,$BF,$A6,$FF,$7A,$7E,$87
	.dc.b	$1F,$F0,$78,$FF,$5B,$46,$07,$92,$8C,$E7,$FE,$C4,$8C,$7C,$17,$9B
	.dc.b	$E1,$7D,$17,$FC,$17,$3C,$3E,$3C,$FC,$AF,$FB,$47,$3F,$2F,$C3,$FF
	.dc.b	$9C,$F1,$E7,$3E,$6D,$71,$A7,$83,$5F,$36,$9F,$13,$E1,$E5,$FB,$DF
	.dc.b	$8F,$F0,$5F,$23,$E0,$E4,$FC,$37,$E5,$FB,$BD,$3E,$1E,$4F,$C4,$7B
	.dc.b	$FC,$37,$28,$E8,$43,$D1,$7D,$1F,$55,$DE,$3F,$F2,$67,$03,$1E,$B5
	.dc.b	$F2,$F7,$F9,$59,$62,$FE,$56,$DF,$E1,$CD,$D8,$BE,$03,$C5,$8F,$80
	.dc.b	$FB,$8F,$7E,$0F,$D4,$CB,$4C,$57,$E1,$B4,$ED,$5F,$77,$3B,$3F,$73
	.dc.b	$2F,$06,$0F,$6B,$F1,$FE,$DC,$AF,$C5,$47,$E4,$BD,$5F,$C9,$FA,$B2
	.dc.b	$7F,$1D,$1F,$5B,$E1,$FE,$BF,$C2,$C9,$FC,$7C,$7E,$6B,$D3,$FC,$DE
	.dc.b	$12,$7F,$2D,$1D,$5F,$CC,$61,$D5,$FC,$36,$1D,$5F,$CC,$E1,$D5,$F9
	.dc.b	$99,$FF,$A7,$9F,$D0,$7A,$7F,$A8,$F3,$A4,$60,$FD,$27,$9D,$FA,$CF
	.dc.b	$1E,$46,$0F,$D2,$FC,$37,$E4,$3E,$12,$46,$0E,$47,$E6,$70,$FD,$A9
	.dc.b	$FB,$6F,$27,$F6,$F8,$74,$9F,$CF,$61,$87,$91,$E8,$BE,$AF,$AB,$0F
	.dc.b	$DF,$F8,$92,$3F,$74,$FF,$02,$F1,$3F,$DB,$1D,$5F,$E7,$6F,$02,$9F
	.dc.b	$A5,$A4,$7F,$9E,$8C,$C7,$EA,$BE,$16,$4F,$F0,$27,$F5,$BF,$0D,$D6
	.dc.b	$3B,$90,$FD,$B5,$F2,$7F,$D6,$57,$D5,$B7,$E1,$8F,$24,$75,$21,$F5
	.dc.b	$AF,$BB,$FC,$56,$BC,$0F,$63,$9F,$4F,$77,$F8,$95,$77,$6D,$F7,$3F
	.dc.b	$8D,$77,$7E,$AB,$0F,$6C,$F7,$3F,$8F,$E3,$F1,$BC,$19,$1E,$87,$E4
	.dc.b	$F4,$FC,$AE,$5F,$97,$B4,$AF,$6C,$7F,$12,$F8,$59,$1F,$CA,$9F,$E2
	.dc.b	$BF,$0D,$CA,$3A,$10,$F3,$DF,$57,$E3,$DF,$57,$D7,$7D,$1F,$CD,$DD
	.dc.b	$CF,$F8,$2E,$1F,$7B,$87,$EF,$E9,$2F,$E8,$C7,$E0,$B4,$91,$F5,$1F
	.dc.b	$C2,$E1,$FC,$AB,$E1,$24,$FF,$42,$7F,$96,$FC,$2F,$58,$E6,$43,$E7
	.dc.b	$E3,$D1,$FA,$37,$72,$FF,$75,$6E,$C7,$AD,$7E,$D3,$1B,$5F,$C7,$B1
	.dc.b	$FC,$BC,$49,$FC,$4B,$F9,$A8,$FC,$DF,$C3,$7E,$73,$E1,$65,$7E,$3A
	.dc.b	$3F,$A1,$9F,$D1,$22,$47,$F5,$C8,$E4,$F4,$21,$F8,$89,$EA,$FE,$62
	.dc.b	$7A,$BF,$8C,$9E,$AF,$F6,$68,$EA,$FF,$0E,$8E,$AF,$33,$F4,$33,$FB
	.dc.b	$2F,$86,$EB,$1F,$96,$AE,$B8,$7C,$6B,$EB,$3F,$C6,$27,$AC,$FF,$42
	.dc.b	$9F,$DD,$FC,$1C,$AF,$D6,$BF,$BE,$F8,$4F,$DF,$FC,$37,$F0,$0F,$86
	.dc.b	$95,$FB,$28,$FE,$CF,$94,$8F,$EF,$EF,$F6,$AB,$E5,$1D,$08,$7E,$B2
	.dc.b	$FA,$BF,$BC,$BE,$AF,$EC,$2F,$A3,$FC,$C2,$EE,$3F,$D6,$B0,$FE,$2B
	.dc.b	$E5,$FF,$A3,$3C,$8F,$F1,$EB,$FE,$8C,$BF,$E1,$2D,$7F,$7A,$ED,$FE
	.dc.b	$32,$FF,$A3,$1D,$FC,$1F,$43,$AF,$C2,$7F,$81,$34,$FF,$02,$E7,$2B
	.dc.b	$FC,$3B,$1E,$26,$1E,$03,$CB,$F0,$BC,$9E,$A4,$3C,$75,$D5,$FF,$13
	.dc.b	$D7,$57,$FA,$7D,$19,$9B,$2C,$43,$D4,$6B,$8A,$D4,$F1,$9C,$BC,$AB
	.dc.b	$8C,$62,$3F,$C6,$47,$3D,$79,$4E,$9B,$1E,$3F,$3D,$8E,$23,$FC,$14
	.dc.b	$DB,$55,$35,$4C,$A9,$9B,$F2,$1E,$79,$F6,$B2,$B9,$B8,$34,$53,$D0
	.dc.b	$57,$6E,$DA,$F3,$8E,$0F,$03,$4E,$23,$71,$F0,$4E,$91,$93,$D0,$FC
	.dc.b	$88,$EF,$8F,$D0,$8F,$A9,$86,$A3,$13,$AC,$7A,$57,$D0,$C7,$E9,$56
	.dc.b	$B0,$E1,$5A,$8E,$93,$C6,$BF,$33,$F0,$9F,$A2,$F8,$59,$1F,$A0,$8F
	.dc.b	$D2,$FC,$1F,$E9,$BE,$12,$4F,$E9,$23,$FC,$BF,$C7,$FE,$63,$E6,$FD
	.dc.b	$27,$21,$BE,$5A,$7F,$5F,$8C,$C3,$B1,$FA,$EF,$85,$EE,$FE,$C3,$2E
	.dc.b	$EF,$EC,$32,$EE,$FD,$EC,$BB,$3E,$13,$27,$ED,$BE,$13,$F7,$BF,$0B
	.dc.b	$23,$F7,$71,$FC,$03,$E0,$E4,$7E,$F9,$FE,$07,$F0,$7E,$BF,$86,$F6
	.dc.b	$74,$CA,$F0,$70,$7F,$03,$E3,$FD,$2F,$93,$FE,$90,$CB,$F8,$59,$2B
	.dc.b	$F8,$4F,$C3,$76,$7F,$84,$E7,$6F,$8E,$BF,$EC,$79,$EB,$9F,$77,$FC
	.dc.b	$E5,$97,$F0,$D6,$C7,$F1,$8F,$86,$FE,$33,$F0,$B2,$7F,$8B,$C7,$F1
	.dc.b	$CF,$85,$91,$FC,$69,$FE,$3F,$F0,$7F,$5F,$C1,$FB,$1E,$0C,$AF,$6C
	.dc.b	$75,$79,$90,$FE,$43,$F0,$9D,$1F,$E4,$77,$75,$F5,$E1,$D6,$7E,$3E
	.dc.b	$1D,$5F,$5E,$1F,$CC,$30,$ED,$1F,$CB,$31,$C3,$E7,$21,$D8,$FE,$67
	.dc.b	$7D,$5F,$8B,$7D,$70,$FF,$5F,$68,$C7,$F9,$EF,$C2,$FF,$3E,$F8,$59
	.dc.b	$87,$F3,$C8,$FF,$2A,$FE,$43,$FE,$D5,$27,$81,$8F,$86,$FC,$E7,$6C
	.dc.b	$BF,$C0,$BF,$9F,$F0,$3A,$C7,$32,$1F,$E4,$2B,$FF,$6D,$36,$FB,$91
	.dc.b	$FF,$2A,$C2,$FB,$47,$E0,$2E,$E9,$EC,$C3,$CF,$E1,$CA,$E2,$FA,$BE
	.dc.b	$17,$D7,$F0,$BE,$CF,$85,$95,$DA,$3F,$6D,$E0,$7E,$DF,$C1,$93,$F9
	.dc.b	$C8,$E4,$F4,$21,$EB,$BE,$AF,$A2,$FA,$BE,$FB,$EA,$FE,$9A,$FA,$BF
	.dc.b	$9A,$BF,$9F,$F0,$BF,$43,$C6,$93,$F0,$A3,$E9,$FC,$1C,$8F,$8A,$FD
	.dc.b	$6F,$84,$FE,$1D,$E3,$49,$FD,$9B,$FC,$47,$C2,$EB,$1C,$C8,$7D,$3F
	.dc.b	$82,$E8,$FD,$5B,$BC,$3F,$BA,$E4,$63,$D6,$BF,$63,$8D,$AF,$C6,$E3
	.dc.b	$F8,$8C,$7B,$D7,$DF,$CB,$0F,$A0,$87,$73,$EF,$E1,$CF,$97,$FB,$A7
	.dc.b	$83,$FC,$9F,$2C,$52,$3F,$95,$63,$F9,$6F,$83,$FC,$BF,$C2,$4A,$FC
	.dc.b	$4C,$7E,$6B,$E0,$E4,$7E,$31,$FC,$EF,$C2,$7F,$35,$F2,$E4,$FF,$1A
	.dc.b	$7F,$9C,$78,$FD,$63,$99,$0F,$CD,$7C,$1F,$47,$F3,$97,$78,$7E,$46
	.dc.b	$1D,$27,$F8,$BE,$16,$7F,$67,$87,$EB,$BE,$0E,$57,$E8,$1F,$D8,$FC
	.dc.b	$27,$EC,$BE,$12,$4F,$E9,$1F,$DA,$FC,$27,$F4,$DF,$43,$FA,$77,$9D
	.dc.b	$2B,$F9,$9C,$72,$7A,$10,$FD,$95,$F4,$7F,$69,$77,$2F,$C0,$61,$D6
	.dc.b	$7F,$9E,$E1,$D5,$FE,$63,$87,$FB,$6B,$F2,$1F,$F5,$CF,$2F,$F8,$AF
	.dc.b	$D0,$47,$FB,$2C,$9E,$E6,$FE,$17,$87,$C2,$CC,$7F,$DA,$74,$9F,$ED
	.dc.b	$1E,$7F,$F6,$9F,$4A,$57,$F4,$E8,$F0,$61,$23,$8B,$E3,$F8,$5E,$D1
	.dc.b	$D3,$1E,$CF,$FB,$4B,$1E,$CF,$33,$FA,$CE,$1D,$9F,$E9,$B8,$73,$7A
	.dc.b	$90,$E5,$7D,$9E,$97,$DB,$7F,$EC,$38,$76,$9E,$67,$62,$18,$3C,$9F
	.dc.b	$09,$F2,$C8,$FC,$4F,$85,$F9,$DF,$09,$F3,$FE,$16,$4F,$C8,$8F,$F0
	.dc.b	$77,$AF,$FE,$0F,$F5,$A4,$FF,$74,$8E,$6F,$63,$E7,$E1,$F6,$70,$EB
	.dc.b	$1F,$4B,$0E,$AF,$C7,$C3,$AB,$FD,$FB,$0E,$AF,$F7,$2C,$3F,$05,$F0
	.dc.b	$92,$3E,$DB,$F8,$5F,$84,$FC,$37,$C2,$C9,$FB,$EF,$E2,$7E,$17,$FC
	.dc.b	$4D,$ED,$FF,$89,$FD,$79,$5F,$E0,$C8,$E4,$F4,$21,$F6,$EF,$A3,$F7
	.dc.b	$6E,$F0,$7D,$2C,$3A,$CF,$F8,$5B,$0E,$AF,$F8,$2B,$0F,$CD,$7B,$5F
	.dc.b	$9B,$F8,$49,$3F,$94,$8F,$CF,$7C,$2F,$E7,$FE,$16,$4F,$E6,$23,$FC
	.dc.b	$7D,$EE,$7F,$9E,$BC,$C7,$E9,$DF,$E5,$7D,$7F,$D0,$F1,$8E,$FD,$E1
	.dc.b	$D4,$FD,$0E,$7E,$07,$F2,$D9,$F8,$1F,$D2,$E7,$E0,$7F,$35,$9F,$25
	.dc.b	$8F,$6C,$BC,$18,$FF,$2A,$6C,$FF,$29,$C1,$9F,$C0,$8F,$11,$E0,$3F
	.dc.b	$67,$F0,$9E,$17,$F5,$19,$78,$71,$FC,$EE,$5E,$17,$C1,$97,$C1,$CB
	.dc.b	$BE,$9F,$E5,$FC,$BF,$83,$7C,$2F,$F0,$7F,$86,$99,$FF,$00,$8F,$E1
	.dc.b	$5F,$0B,$23,$F8,$1B,$FE,$A4,$F0,$FF,$D0,$11,$EE,$CB,$D9,$F0,$3D
	.dc.b	$FE,$22,$4F,$F9,$87,$4F,$F0,$43,$D4,$87,$F0,$1C,$7A,$5F,$F9,$5C
	.dc.b	$FE,$31,$DF,$FE,$73,$9F,$09,$0F,$00,$C7,$C7,$B7,$CF,$CB,$C3,$3A
	.dc.b	$DB,$F0,$39,$FC,$39,$EF,$F0,$D2,$BF,$D6,$78,$7F,$B3,$E3,$C7,$F0
	.dc.b	$9E,$4F,$85,$95,$AD,$11,$E8,$F8,$5F,$9D,$F0,$92,$BE,$03,$F4,$7E
	.dc.b	$13,$9C,$71,$C7,$C0,$FA,$B2,$FF,$6F,$65,$DA,$31,$97,$67,$C1,$97
	.dc.b	$77,$A9,$0F,$77,$0F,$FB,$95,$FA,$18,$FC,$CF,$85,$91,$E9,$7E,$7F
	.dc.b	$C1,$7D,$0F,$83,$FA,$3E,$EC,$AF,$74,$7E,$27,$E1,$24,$7D,$B7,$F1
	.dc.b	$9F,$0B,$CA,$3A,$10,$F4,$DF,$57,$E4,$5F,$57,$DB,$7D,$1F,$4D,$DE
	.dc.b	$5F,$D4,$61,$D6,$7E,$1E,$3E,$A7,$A9,$E6,$3A,$90,$EA,$79,$C8,$76
	.dc.b	$3F,$29,$1F,$90,$F7,$FF,$23,$ED,$49,$FB,$D1,$F9,$4F,$6A,$47,$E0
	.dc.b	$5F,$CB,$FA,$BF,$AD,$F8,$19,$3F,$9F,$7F,$61,$F0,$BD,$63,$99,$0F
	.dc.b	$BB,$EC,$74,$7E,$FD,$DD,$3F,$D6,$FC,$9F,$EA,$F9,$FB,$B8,$76,$69
	.dc.b	$FA,$DC,$3F,$53,$EA,$FE,$AB,$D6,$93,$F9,$78,$FD,$6F,$9F,$23,$F3
	.dc.b	$4F,$EC,$3D,$0F,$E0,$DF,$0B,$FC,$1F,$E0,$E5,$7E,$E6,$3A,$BC,$C8
	.dc.b	$7E,$5B,$CD,$E8,$FE,$62,$EE,$BE,$AC,$3A,$CF,$F0,$3C,$3A,$BF,$98
	.dc.b	$C3,$F8,$0F,$9D,$FC,$0B,$C9,$93,$FA,$F8,$FE,$21,$F9,$17,$F2,$8F
	.dc.b	$00,$FF,$0C,$9F,$D5,$37,$2F,$F6,$FB,$FC,$83,$B9,$F0,$BF,$C8,$FE
	.dc.b	$12,$5F,$F1,$58,$F0,$BD,$C8,$74,$F1,$FC,$3C,$BE,$2F,$1E,$4F,$F0
	.dc.b	$8C,$7B,$5F,$48,$FD,$9E,$1F,$CC,$63,$AC,$7E,$F3,$4B,$9F,$9B,$2E
	.dc.b	$B3,$CF,$1E,$AF,$B3,$2E,$4F,$F2,$BC,$7B,$5F,$F4,$DC,$7E,$2F,$8D
	.dc.b	$F1,$BC,$39,$7E,$58,$F9,$3E,$07,$CA,$EF,$93,$E9,$8F,$E8,$DF,$0B
	.dc.b	$FD,$1F,$E1,$24,$FF,$39,$8E,$AF,$32,$1F,$43,$B3,$AB,$F0,$EF,$AB
	.dc.b	$F5,$2F,$AB,$FD,$0E,$FA,$BF,$C6,$2F,$EE,$F6,$48,$F9,$8F,$E0,$39
	.dc.b	$BF,$03,$CF,$F8,$2D,$65,$7D,$18,$FE,$BD,$F0,$92,$3F,$A8,$3F,$D8
	.dc.b	$7E,$17,$94,$74,$21,$F8,$8B,$EA,$FD,$BB,$EA,$FE,$32,$FA,$3F,$CB
	.dc.b	$AE,$F0,$7F,$7D,$C3,$F2,$BC,$3F,$96,$CE,$5F,$E1,$63,$F3,$3A,$7E
	.dc.b	$6B,$09,$3F,$89,$8F,$EE,$DF,$0B,$23,$FB,$43,$FD,$E7,$E1,$7A,$C7
	.dc.b	$32,$1F,$A0,$C7,$AB,$F9,$3B,$EA,$FE,$92,$FA,$3F,$D1,$AE,$E5,$FE
	.dc.b	$15,$C3,$F5,$B8,$76,$AF,$D5,$E3,$87,$CC,$75,$21,$D4,$F3,$90,$EC
	.dc.b	$7F,$80,$67,$F7,$37,$23,$F5,$0F,$EF,$23,$F7,$BF,$0B,$FB,$EF,$85
	.dc.b	$95,$FA,$D8,$FF,$0D,$1F,$E1,$B8,$91,$FE,$0D,$8E,$4F,$42,$1F,$C1
	.dc.b	$27,$AB,$FB,$79,$EA,$FF,$08,$9E,$AF,$F8,$5E,$3A,$BF,$DB,$23,$FC
	.dc.b	$41,$F5,$7F,$EA,$C8,$FF,$08,$3F,$E1,$17,$FC,$12,$D7,$FB,$E3,$BE
	.dc.b	$D3,$FE,$91,$CD,$D1,$FE,$B5,$87,$C1,$F6,$3B,$FC,$37,$F9,$67,$F4
	.dc.b	$FF,$07,$A8,$7B,$61,$FE,$5A,$6E,$1F,$EF,$67,$C1,$84,$C7,$D2,$7A
	.dc.b	$BE,$1F,$AE,$7E,$8C,$FC,$0F,$AB,$3F,$04,$F8,$6A,$48,$FF,$16,$61
	.dc.b	$DB,$E1,$FE,$2F,$C3,$C9,$F8,$71,$F2,$3E,$1F,$E4,$FC,$34,$9F,$8F
	.dc.b	$1E,$DC,$A4,$75,$DB,$D9,$1F,$07,$4F,$A3,$84,$AE,$A7,$D0,$F8,$7E
	.dc.b	$CF,$D2,$C7,$B3,$F4,$B1,$EA,$FA,$30,$B8,$FC,$9C,$7E,$EE,$1D,$6B
	.dc.b	$EC,$E3,$87,$CE,$75,$21,$D4,$F4,$6F,$F4,$70,$EA,$D8,$C8,$E0,$FB
	.dc.b	$58,$7E,$33,$E1,$FF,$1B,$F0,$F2,$BF,$17,$1F,$90,$F8,$79,$1F,$8E
	.dc.b	$7F,$27,$F0,$FF,$85,$CE,$4F,$6C,$7F,$0D,$1F,$89,$E0,$ED,$3D,$08
	.dc.b	$7E,$5B,$E1,$BA,$BF,$98,$C2,$EF,$F9,$68,$FC,$FE,$3D,$9F,$C0,$E1
	.dc.b	$67,$F2,$58,$7E,$9B,$E1,$FF,$4F,$F0,$B2,$FF,$4B,$1F,$AA,$F8,$69
	.dc.b	$1F,$A8,$7F,$5B,$F0,$FF,$9F,$E4,$93,$EB,$E6,$47,$E8,$F9,$BB,$4F
	.dc.b	$42,$1F,$B0,$F8,$5E,$AF,$EC,$B0,$BA,$7A,$FA,$3F,$67,$97,$6A,$FC
	.dc.b	$DE,$36,$BF,$59,$8F,$E7,$E7,$FD,$1E,$3F,$0F,$7B,$99,$AF,$8D,$39
	.dc.b	$F8,$0C,$74,$F1,$B4,$62,$7F,$55,$1C,$9F,$C2,$9D,$6F,$4B,$F3,$77
	.dc.b	$8D,$86,$C7,$0A,$E4,$F9,$9B,$8F,$92,$5E,$2F,$13,$AC,$EB,$2F,$36
	.dc.b	$3E,$FC,$24,$F4,$6C,$65,$D9,$4F,$41,$63,$83,$A3,$83,$87,$E0,$C7
	.dc.b	$17,$89,$A7,$11,$B8,$CE,$1F,$3A,$3D,$11,$C8,$F8,$5E,$FF,$B2,$A4
	.dc.b	$7C,$57,$E4,$79,$18,$8F,$8D,$16,$FE,$37,$91,$FC,$FB,$E2,$65,$7F
	.dc.b	$3C,$7F,$A0,$FC,$4F,$FB,$87,$F2,$4F,$8D,$7F,$8B,$7F,$DB,$15,$FE
	.dc.b	$D3,$6F,$21,$DE,$BF,$CD,$7C,$4F,$E7,$84,$AF,$CF,$78,$3E,$07,$FA
	.dc.b	$4E,$57,$1F,$CC,$78,$DF,$EE,$39,$FF,$64,$63,$D1,$F0,$7C,$67,$94
	.dc.b	$F3,$3D,$4F,$E6,$F8,$DB,$FE,$91,$8F,$AF,$E2,$65,$7F,$B8,$B0,$47
	.dc.b	$BF,$E2,$7E,$07,$C4,$CA,$D3,$F0,$FE,$2B,$FA,$B7,$91,$FD,$5F,$C6
	.dc.b	$95,$FB,$98,$E4,$F5,$C3,$BB,$EA,$C3,$A3,$EC,$C2,$E7,$F3,$71,$EB
	.dc.b	$3F,$C0,$F1,$EA,$FF,$52,$C7,$EA,$7C,$47,$D5,$F8,$A9,$3E,$B8,$FB
	.dc.b	$1F,$15,$F6,$7B,$A4,$FB,$E3,$FB,$77,$93,$23,$F8,$83,$FD,$C7,$CA
	.dc.b	$EB,$1C,$C8,$7D,$2E,$AE,$AF,$C9,$BE,$AF,$D6,$BE,$8F,$F6,$8B,$BA
	.dc.b	$FF,$79,$C3,$F1,$98,$76,$AF,$BB,$8E,$1F,$31,$D4,$87,$53,$CE,$43
	.dc.b	$B1,$FC,$A6,$7F,$31,$D9,$23,$EF,$3F,$9B,$F8,$8F,$CE,$7C,$54,$9F
	.dc.b	$C1,$3F,$9F,$F8,$AF,$F0,$97,$95,$27,$F9,$A3,$FE,$15,$F2,$FA,$47
	.dc.b	$E5,$30,$BA,$7E,$97,$1E,$93,$F9,$8C,$6C,$FE,$A7,$1E,$93,$FE,$0C
	.dc.b	$C6,$CF,$F8,$77,$1F,$D8,$7C,$4F,$EC,$7E,$2A,$5F,$E5,$E3,$F6,$9F
	.dc.b	$13,$23,$F3,$4F,$ED,$FE,$27,$FC,$55,$E6,$C9,$FE,$98,$FF,$8B,$7C
	.dc.b	$DE,$B1,$CC,$87,$EB,$BE,$2B,$A3,$FB,$0B,$B9,$FF,$01,$C3,$A4,$FF
	.dc.b	$89,$30,$B3,$FE,$35,$C3,$F8,$47,$C3,$FF,$09,$F8,$A9,$7F,$AF,$8F
	.dc.b	$E8,$9E,$3B,$C4,$FF,$6F,$13,$C7,$CB,$FE,$DD,$7F,$C7,$DE,$8F,$F9
	.dc.b	$17,$C8,$98,$7F,$67,$8E,$EF,$52,$0C,$FF,$B1,$5B,$E2,$3A,$E1,$2B
	.dc.b	$FC,$2F,$8F,$8E,$3C,$3F,$11,$FE,$53,$C7,$BD,$7F,$90,$B2,$C3,$C3
	.dc.b	$1E,$8F,$83,$2F,$E4,$D9,$FA,$B3,$ED,$3E,$3C,$BF,$C9,$99,$F7,$8E
	.dc.b	$51,$E0,$7D,$87,$F8,$0F,$1F,$0E,$1F,$E6,$1C,$BE,$2F,$C3,$CA,$F1
	.dc.b	$3F,$23,$E2,$3E,$4F,$C3,$49,$F3,$3F,$33,$E1,$FF,$CD,$DE,$7F,$F7
	.dc.b	$5F,$6D,$FA,$87,$F2,$D6,$DE,$8F,$76,$1D,$DF,$50,$DB,$ED,$65,$E1
	.dc.b	$8E,$E4,$37,$1F,$7F,$2F,$14,$7F,$87,$F1,$F1,$63,$E1,$3C,$5E,$D7
	.dc.b	$F4,$1D,$7E,$CF,$C4,$FE,$10,$97,$F4,$7E,$27,$F0,$FF,$13,$23,$EA
	.dc.b	$3F,$8A,$F8,$8E,$BC,$DD,$BD,$29,$5F,$E4,$9C,$FF,$AC,$CB,$0E,$B3
	.dc.b	$E8,$86,$3F,$94,$9E,$D3,$9C,$6E,$5F,$8E,$CB,$BC,$F5,$CB,$BB,$F2
	.dc.b	$38,$7E,$7E,$5F,$9E,$F8,$99,$3F,$8B,$7F,$43,$F1,$3F,$A2,$F6,$A4
	.dc.b	$FE,$3D,$FD,$2F,$B5,$F2,$3D,$6F,$93,$EB,$4A,$F6,$47,$47,$D5,$85
	.dc.b	$D7,$F4,$78,$F4,$9F,$76,$36,$7F,$4F,$8F,$59,$F8,$D8,$F5,$7D,$5C
	.dc.b	$2C,$BC,$F1,$E9,$21,$F2,$71,$E9,$D7,$FA,$0D,$19,$E2,$BF,$15,$0C
	.dc.b	$BC,$11,$DB,$2F,$06,$7F,$E7,$5D,$7F,$2F,$9F,$F0,$6F,$7B,$F8,$3F
	.dc.b	$BB,$30,$FD,$A4,$7F,$0A,$F8,$9F,$E1,$7F,$13,$27,$F7,$11,$F9,$0F
	.dc.b	$6B,$F2,$3E,$9C,$9F,$C2,$47,$57,$99,$0F,$C2,$7C,$47,$57,$F8,$15
	.dc.b	$F5,$7F,$11,$7D,$5F,$C7,$5F,$57,$E4,$DF,$F1,$EF,$89,$FE,$47,$F6
	.dc.b	$FF,$EE,$93,$FD,$D5,$C2,$8C,$F7,$CB,$FE,$2F,$5F,$EE,$27,$F6,$4F
	.dc.b	$FB,$BF,$93,$FD,$B1,$C1,$FB,$CF,$6F,$F5,$A4,$AF,$D0,$7B,$5E,$17
	.dc.b	$F8,$D6,$35,$FD,$BE,$1F,$EF,$67,$BE,$3E,$17,$2F,$A3,$E2,$7A,$47
	.dc.b	$E1,$30,$BC,$9F,$5F,$A5,$97,$BB,$E2,$3D,$FF,$11,$30,$F5,$C7,$C2
	.dc.b	$F8,$89,$1E,$E7,$E2,$FC,$47,$F0,$6F,$80,$93,$FB,$97,$F8,$4F,$BF
	.dc.b	$D6,$3E,$2E,$1D,$5F,$5E,$1D,$1F,$6E,$17,$3F,$F5,$36,$EC,$BA,$D7
	.dc.b	$E7,$B2,$B5,$EB,$CB,$B4,$F4,$3D,$27,$D4,$F8,$5E,$9C,$7F,$CF,$B1
	.dc.b	$FF,$5D,$CF,$AC,$FB,$38,$F4,$9F,$AF,$B7,$FA,$A2,$7D,$87,$F1,$5C
	.dc.b	$3F,$0B,$F0,$B2,$3F,$02,$FE,$23,$E0,$FF,$13,$F0,$9F,$8A,$F8,$49
	.dc.b	$5F,$86,$8F,$E5,$BF,$07,$23,$F9,$03,$FC,$C3,$E0,$F9,$47,$42,$1F
	.dc.b	$82,$BE,$AF,$E4,$2F,$AB,$F8,$6B,$E8,$FF,$04,$BB,$BF,$E0,$70,$FC
	.dc.b	$EF,$C2,$7E,$7B,$E0,$A5,$FE,$6A,$3F,$43,$F0,$52,$3F,$3A,$FE,$93
	.dc.b	$DD,$FE,$8B,$F0,$9F,$D1,$BE,$12,$57,$F3,$98,$EA,$FE,$93,$0E,$AF
	.dc.b	$E6,$B0,$E8,$FE,$73,$0B,$97,$E0,$31,$EB,$3F,$D0,$F1,$EA,$FF,$1E
	.dc.b	$C7,$F6,$BE,$FF,$ED,$BD,$A9,$3F,$B2,$8F,$DC,$FB,$7F,$BA,$F5,$E4
	.dc.b	$FE,$DA,$3F,$AE,$FC,$1F,$F5,$EF,$84,$93,$FD,$46,$3A,$BC,$C8,$7F
	.dc.b	$68,$FB,$4F,$F5,$07,$88,$EC,$61,$DA,$BF,$6F,$8F,$89,$E8,$7F,$5D
	.dc.b	$C7,$BB,$FC,$E3,$1E,$EF,$EF,$B2,$BB,$BA,$FF,$C2,$BA,$78,$33,$EF
	.dc.b	$7F,$C4,$32,$F0,$5F,$EA,$33,$F6,$CF,$72,$1D,$CF,$F1,$6D,$FF,$B1
	.dc.b	$89,$E3,$FF,$90,$9F,$F5,$8C,$6B,$CF,$C7,$A1,$23,$07,$F7,$7F,$84
	.dc.b	$FF,$03,$FC,$24,$8C,$1D,$B0,$E8,$47,$C3,$EB,$F7,$C3,$A4,$7C,$DC
	.dc.b	$30,$F2,$3F,$AE,$DF,$57,$FA,$FE,$1E,$7F,$37,$D1,$E7,$49,$F6,$C7
	.dc.b	$AF,$CB,$F6,$79,$72,$7E,$0C,$7F,$87,$BE,$17,$FC,$3F,$F0,$72,$7F
	.dc.b	$C2,$51,$D5,$E6,$43,$F0,$5E,$57,$57,$E5,$DF,$57,$F0,$D7,$D5,$FF
	.dc.b	$0E,$5F,$57,$FB,$BD,$FC,$FF,$1F,$E8,$78,$72,$7E,$AC,$7D,$3F,$0E
	.dc.b	$47,$D8,$7E,$B7,$77,$F8,$DB,$E1,$64,$FF,$8A,$5F,$F2,$57,$B1,$F8
	.dc.b	$98,$E6,$7E,$03,$0E,$CF,$E6,$F0,$EA,$FE,$77,$0B,$D1,$C7,$BD,$D0
	.dc.b	$CF,$BD,$FF,$84,$33,$B7,$FE,$03,$CF,$C1,$3D,$4F,$09,$4F,$D5,$78
	.dc.b	$3D,$BB,$7F,$3B,$9D,$9F,$F2,$D6,$5D,$DF,$61,$0F,$00,$3F,$C7,$57
	.dc.b	$E0,$C3,$D1,$2E,$FF,$CE,$12,$BF,$43,$E0,$7E,$7F,$9F,$F4,$1D,$12
	.dc.b	$7F,$4B,$1F,$2B,$E1,$64,$77,$C7,$DD,$1F,$3B,$E1,$7F,$51,$87,$5A
	.dc.b	$FE,$0D,$8E,$1E,$67,$EC,$31,$EC,$FF,$0B,$C7,$AB,$EA,$C2,$E3,$F2
	.dc.b	$F1,$FD,$A7,$27,$ED,$79,$65,$FE,$EA,$3F,$71,$C3,$FE,$BA,$F9,$8F
	.dc.b	$85,$C3,$FE,$9C,$6D,$7C,$C9,$FC,$2F,$C2,$48,$F6,$57,$F8,$5B,$85
	.dc.b	$3F,$8A,$F8,$5F,$0D,$7F,$08,$CB,$C2,$FF,$23,$CB,$C2,$FF,$AE,$32
	.dc.b	$EF,$CC,$6F,$1B,$4F,$05,$7D,$8C,$AE,$DF,$89,$CF,$D5,$BF,$AF,$79
	.dc.b	$7D,$63,$DD,$72,$3C,$0F,$C1,$AF,$D0,$FC,$2F,$E8,$BE,$1A,$57,$B8
	.dc.b	$7B,$3F,$E6,$09,$EF,$1D,$48,$7F,$A9,$B5,$47,$78,$C6,$17,$3F,$D3
	.dc.b	$53,$1F,$0D,$77,$C3,$FC,$DB,$8F,$8A,$3F,$2F,$8F,$D7,$23,$F2,$5F
	.dc.b	$B5,$1D,$78,$BE,$D6,$56,$BF,$D7,$D8,$F7,$7B,$71,$79,$9B,$7D,$F3
	.dc.b	$FD,$A6,$3F,$85,$F8,$5F,$C3,$7C,$2C,$BF,$B1,$1F,$89,$F8,$59,$1F
	.dc.b	$6D,$FC,$67,$C2,$FF,$0C,$99,$3E,$07,$F8,$7D,$7E,$47,$E1,$7B,$4F
	.dc.b	$43,$F0,$8F,$47,$E6,$5D,$C7,$E4,$E1,$D2,$7F,$77,$85,$9F,$E1,$98
	.dc.b	$7E,$77,$E1,$7F,$3D,$A4,$BF,$C8,$47,$E8,$7E,$16,$47,$E4,$DF,$D2
	.dc.b	$7C,$27,$F2,$4C,$E4,$FC,$87,$F9,$3E,$7D,$63,$99,$0F,$C1,$7C,$2F
	.dc.b	$47,$F0,$97,$73,$FC,$0E,$1D,$27,$F8,$9E,$16,$7F,$92,$61,$FB,$5F
	.dc.b	$83,$FD,$B7,$C2,$4B,$FD,$44,$7E,$E7,$E1,$64,$7E,$AD,$FD,$E7,$C2
	.dc.b	$FF,$3C,$E1,$93,$F8,$07,$F9,$FE,$DD,$63,$99,$0F,$F7,$FF,$6B,$E3
	.dc.b	$3A,$C7,$E7,$70,$BA,$7E,$6B,$1E,$B3,$FC,$B3,$1B,$3F,$CF,$71,$FE
	.dc.b	$61,$3D,$30,$B3,$FD,$1F,$1E,$D3,$FA,$F4,$61,$F1,$BE,$0E,$8F,$E2
	.dc.b	$D9,$59,$FC,$EE,$5E,$09,$FE,$29,$95,$9F,$DC,$F1,$33,$F0,$D7,$F4
	.dc.b	$3C,$7B,$FC,$36,$2F,$C3,$1F,$EB,$36,$C6,$63,$CB,$F1,$95,$7C,$35
	.dc.b	$F8,$79,$7E,$B9,$CB,$FB,$0C,$BF,$B7,$91,$FD,$9F,$2F,$3C,$3A,$63
	.dc.b	$27,$94,$77,$C7,$AC,$7D,$6C,$70,$F5,$21,$FD,$7B,$0F,$EF,$B8,$79
	.dc.b	$FE,$1F,$D1,$F0,$D2,$BC,$D1,$EB,$F8,$7F,$67,$C3,$49,$F5,$47,$F8
	.dc.b	$1B,$A6,$47,$F0,$A7,$FC,$13,$D5,$F0,$FE,$1B,$B4,$F4,$3E,$13,$D5
	.dc.b	$F8,$97,$D5,$FC,$7D,$F4,$7F,$BA,$DD,$E6,$FF,$03,$61,$F3,$FE,$17
	.dc.b	$E8,$7C,$34,$BF,$9D,$1F,$4F,$E1,$64,$7D,$17,$EB,$7C,$37,$F8,$7B
	.dc.b	$B3,$FC,$3F,$DB,$2B,$F9,$1C,$7D,$BF,$86,$ED,$1D,$0F,$B4,$F4,$7F
	.dc.b	$4B,$77,$3F,$CD,$61,$D6,$7F,$96,$E1,$D5,$FF,$07,$E1,$D5,$E6,$43
	.dc.b	$F5,$FF,$0D,$CF,$AB,$F2,$9D,$4C,$B1,$5F,$94,$C7,$B3,$FE,$1A,$C7
	.dc.b	$B6,$5F,$E2,$BC,$BF,$29,$F0,$BF,$95,$F8,$69,$7F,$93,$8F,$CC,$7C
	.dc.b	$24,$8F,$CB,$3F,$9B,$F8,$5F,$F2,$1F,$85,$FE,$44,$F0,$A5,$7F,$48
	.dc.b	$8E,$AF,$32,$0C,$BF,$C8,$5F,$92,$FF,$A8,$37,$FF,$48,$CF,$F0,$CC
	.dc.b	$7C,$0F,$F0,$EC,$7B,$BF,$D5,$31,$EE,$FF,$8D,$B1,$FD,$86,$0F,$86
	.dc.b	$A4,$9F,$E1,$11,$FE,$92,$7C,$5C,$3F,$BD,$87,$C2,$F2,$F8,$59,$5C
	.dc.b	$23,$F8,$07,$A0,$FD,$47,$F8,$86,$FF,$E6,$38,$C7,$93,$A9,$CE,$5F
	.dc.b	$E4,$B9,$E1,$C1,$E6,$F8,$7F,$26,$5E,$23,$F8,$DE,$9E,$97,$CF,$1E
	.dc.b	$CC,$FC,$71,$BA,$53,$4D,$4F,$F8,$17,$2F,$0B,$FE,$14,$6E,$3E,$68
	.dc.b	$D5,$FC,$8F,$86,$F9,$3F,$0F,$27,$E3,$C7,$CC,$F8,$5F,$9B,$F0,$D2
	.dc.b	$7E,$5C,$7B,$FC,$CF,$81,$E6,$49,$F1,$47,$D4,$F8,$7E,$D1,$D0,$FA
	.dc.b	$6F,$57,$EA,$DF,$57,$DD,$B3,$0E,$D1,$ED,$C3,$B3,$F2,$74,$FA,$98
	.dc.b	$F7,$8E,$A4,$39,$FC,$37,$77,$AE,$3D,$F9,$3F,$27,$8F,$69,$E9,$BB
	.dc.b	$2E,$F5,$F7,$E3,$FA,$06,$FF,$4B,$E1,$FF,$22,$4B,$FC,$7F,$C2,$FE
	.dc.b	$53,$E1,$E4,$7E,$49,$FC,$BF,$C3,$FE,$1F,$D2,$FC,$47,$A5,$2B,$ED
	.dc.b	$47,$E7,$2E,$47,$43,$F3,$7F,$0D,$D1,$F8,$F8,$5D,$3F,$5B,$8F,$59
	.dc.b	$FC,$26,$3D,$5F,$6E,$3F,$A8,$F8,$7F,$D4,$FC,$34,$9F,$D3,$C7,$EB
	.dc.b	$3E,$1E,$47,$EA,$9F,$D7,$FC,$37,$E7,$FD,$49,$3F,$91,$7F,$45,$EA
	.dc.b	$FE,$D3,$E1,$FB,$4F,$43,$F6,$6F,$47,$EF,$DD,$D7,$F8,$2E,$1D,$27
	.dc.b	$E9,$E1,$67,$E9,$61,$FD,$CE,$5F,$8A,$3E,$17,$0D,$8F,$F3,$B1,$A8
	.dc.b	$CF,$C3,$48,$CF,$E7,$3F,$EC,$57,$F6,$81,$C6,$1F,$ED,$2B,$57,$1A
	.dc.b	$DD,$38,$7C,$8D,$8F,$0D,$70,$6D,$7F,$88,$19,$19,$70,$7B,$22,$BF
	.dc.b	$ED,$58,$D3,$C2,$32,$EC,$DC,$69,$FF,$16,$B5,$E5,$1C,$23,$11,$C4
	.dc.b	$F0,$B8,$3F,$D7,$2F,$A0,$87,$C6,$72,$D8,$F8,$97,$F1,$DF,$87,$CB
	.dc.b	$8F,$17,$C7,$16,$BF,$DE,$18,$39,$BF,$9D,$FC,$47,$F8,$AF,$F2,$7F
	.dc.b	$F6,$09,$FE,$C2,$F0,$7F,$D8,$17,$FE,$43,$8F,$BC,$FF,$25,$7E,$56
	.dc.b	$6C,$BF,$C7,$51,$86,$ED,$FC,$DF,$DA,$FE,$71,$ED,$CC,$3F,$95,$47
	.dc.b	$89,$FE,$93,$95,$7B,$7C,$47,$96,$FC,$DB,$49,$F2,$47,$57,$F9,$A6
	.dc.b	$1D,$5F,$E1,$B8,$7B,$7E,$27,$DD,$F0,$F2,$7D,$91,$F0,$7E,$23,$E1
	.dc.b	$7C,$44,$9F,$81,$1F,$D3,$FE,$07,$FA,$87,$BD,$27,$FA,$14,$75,$79
	.dc.b	$90,$F8,$DF,$11,$D5,$F9,$57,$D5,$F9,$57,$D5,$FE,$99,$7D,$5F,$E4
	.dc.b	$B7,$F5,$7E,$23,$EB,$7C,$44,$9F,$A9,$1F,$67,$E2,$3E,$D7,$C4,$49
	.dc.b	$FB,$11,$FD,$9F,$E1,$24,$7F,$59,$7F,$B5,$7C,$17,$58,$E6,$43,$EF
	.dc.b	$7C,$47,$57,$F0,$57,$D5,$FC,$15,$F4,$7F,$9D,$DD,$E9,$FE,$75,$87
	.dc.b	$E3,$70,$ED,$5F,$89,$C7,$0F,$98,$EA,$43,$A9,$E7,$21,$D8,$FE,$E5
	.dc.b	$3F,$99,$F8,$89,$1F,$97,$7F,$39,$F1,$1F,$9D,$F8,$89,$3F,$9B,$7F
	.dc.b	$41,$F0,$9F,$E0,$BF,$86,$FF,$06,$7C,$44,$AF,$EF,$D1,$C9,$FD,$15
	.dc.b	$D7,$F4,$5F,$0F,$D1,$FD,$26,$17,$5F,$D2,$47,$FA,$AB,$C0,$65,$DE
	.dc.b	$3F,$C1,$59,$77,$7F,$B1,$E5,$FB,$3F,$82,$FD,$A7,$C1,$C9,$FD,$94
	.dc.b	$7F,$A8,$7F,$27,$FF,$08,$FA,$A7,$FD,$6F,$3F,$EB,$96,$BF,$E1,$B7
	.dc.b	$FC,$2B,$EB,$3E,$27,$FC,$55,$F1,$12,$BF,$C3,$91,$E1,$7B,$90,$E5
	.dc.b	$EE,$F8,$7E,$17,$C5,$8C,$9F,$04,$7F,$8D,$6B,$A4,$7F,$7E,$BC,$3E
	.dc.b	$5F,$6F,$D4,$47,$D1,$EC,$7B,$7D,$99,$1E,$B7,$E0,$7A,$9F,$E4,$1F
	.dc.b	$89,$93,$FE,$33,$7F,$C8,$9F,$13,$F1,$6F,$A4,$FC,$1C,$30,$F3,$21
	.dc.b	$F1,$70,$EA,$F3,$21,$FE,$16,$BE,$AF,$98,$87,$CD,$AF,$A9,$EB,$7A
	.dc.b	$23,$A1,$DB,$9F,$F0,$7B,$7D,$7C,$FD,$33,$D4,$EF,$3F,$E2,$1C,$FE
	.dc.b	$F7,$95,$27,$EE,$3F,$81,$F2,$FF,$05,$E2,$FE,$0F,$C6,$95,$F8,$18
	.dc.b	$FF,$33,$7C,$4F,$F9,$B3,$F5,$2F,$EC,$59,$7F,$9A,$DB,$E9,$7F,$9A
	.dc.b	$79,$D5,$D7,$0E,$E4,$3F,$15,$97,$81,$FC,$76,$5E,$07,$F1,$D9,$78
	.dc.b	$11,$FE,$49,$C7,$BA,$F8,$34,$61,$FF,$40,$61,$F9,$DF,$1F,$F3,$DE
	.dc.b	$2C,$BF,$CE,$47,$E8,$7B,$A4,$7E,$7D,$FD,$27,$7F,$BF,$E2,$BE,$07
	.dc.b	$C5,$4A,$F1,$F0,$FC,$19,$ED,$1D,$08,$7E,$9F,$B3,$AB,$FA,$9C,$2D
	.dc.b	$7D,$AE,$F6,$3D,$EB,$AF,$1F,$FA,$4F,$7F,$F5,$15,$FC,$FC,$F5,$5F
	.dc.b	$B3,$ED,$FD,$D7,$5C,$8F,$DC,$47,$EF,$79,$E4,$7E,$ED,$FE,$01,$C3
	.dc.b	$F6,$FE,$2B,$EE,$7C,$54,$AF,$09,$F3,$73,$FF,$57,$BD,$08,$78,$4D
	.dc.b	$F5,$FF,$7C,$7F,$83,$61,$E0,$69,$FC,$2F,$0F,$03,$DB,$97,$EC,$E1
	.dc.b	$E1,$8F,$BD,$8F,$E0,$1F,$F5,$57,$A9,$F8,$0C,$7F,$8B,$67,$FC,$64
	.dc.b	$D6,$DF,$A4,$9D,$70,$FF,$B5,$A3,$E7,$69,$E3,$7E,$86,$9E,$2C,$7F
	.dc.b	$19,$9F,$86,$7F,$39,$9F,$4E,$0F,$C8,$E9,$FE,$57,$E4,$C6,$08,$FF
	.dc.b	$2D,$70,$E1,$B1,$F1,$35,$FF,$63,$3E,$AC,$35,$1F,$EC,$38,$D6,$58
	.dc.b	$9F,$D7,$7C,$4F,$EB,$FE,$2A,$57,$C7,$E2,$4F,$2C,$E4,$F7,$7A,$C7
	.dc.b	$38,$EA,$43,$C4,$73,$7A,$BF,$AD,$AF,$0B,$E9,$7C,$FF,$0F,$23,$07
	.dc.b	$B7,$E2,$3D,$64,$7E,$07,$C5,$7F,$07,$9F,$E1,$13,$27,$D5,$1F,$0A
	.dc.b	$A4,$74,$3E,$FF,$C5,$73,$7F,$03,$57,$87,$EF,$61,$D6,$7E,$24,$F5
	.dc.b	$7F,$83,$61,$F3,$BE,$07,$E7,$FC,$04,$9F,$A9,$1F,$4B,$E2,$7E,$9F
	.dc.b	$C2,$49,$FB,$11,$FC,$7F,$1F,$E4,$19,$C9,$FA,$51,$D5,$E6,$43,$F2
	.dc.b	$FF,$15,$D5,$FC,$15,$F5,$7F,$37,$7D,$5F,$B9,$7D,$5F,$E3,$D7,$F8
	.dc.b	$6B,$ED,$1F,$A2,$C3,$0F,$98,$EA,$43,$A9,$E6,$3A,$90,$EA,$7F,$AB
	.dc.b	$BF,$18,$78,$BF,$E9,$46,$C7,$FA,$69,$AF,$FA,$79,$AF,$FA,$8D,$AF
	.dc.b	$ED,$DA,$FF,$38,$6A,$69,$23,$C0,$48,$FF,$55,$FC,$BB,$91,$EA,$FE
	.dc.b	$C2,$30,$D8,$CC,$FF,$18,$9E,$1E,$8E,$7D,$29,$39,$A4,$73,$FC,$11
	.dc.b	$E9,$7F,$5D,$D3,$FA,$F7,$14,$C3,$F6,$11,$FE,$D9,$F4,$1E,$C7,$FA
	.dc.b	$E9,$B3,$FE,$C0,$6B,$FE,$C4,$6B,$FE,$C9,$6B,$FC,$75,$AF,$F5,$E6
	.dc.b	$A6,$92,$3C,$04,$8F,$F6,$8F,$77,$FB,$AF,$E3,$3F,$86,$FC,$6C,$9F
	.dc.b	$E2,$51,$FC,$43,$E3,$24,$7F,$15,$7F,$89,$FC,$67,$F7,$DE,$89,$3F
	.dc.b	$C3,$DF,$F0,$07,$4F,$F1,$8B,$93,$D0,$FE,$3D,$F1,$BD,$1F,$E6,$F8
	.dc.b	$5E,$DF,$E6,$98,$F4,$9F,$EF,$58,$D9,$FF,$05,$63,$FC,$93,$0E,$B5
	.dc.b	$FD,$03,$1C,$3E,$72,$1D,$8F,$E5,$D8,$73,$F1,$7F,$B4,$63,$67,$FA
	.dc.b	$D6,$1F,$CC,$3E,$2B,$F9,$B7,$E5,$1F,$E6,$C9,$67,$FE,$2C,$8C,$3F
	.dc.b	$D2,$E7,$7E,$34,$AF,$EB,$F5,$FE,$DA,$AE,$1F,$15,$FE,$25,$EE,$FF
	.dc.b	$13,$6B,$2F,$F9,$C4,$78,$5F,$E8,$98,$D7,$C3,$A7,$F4,$EC,$7F,$DE
	.dc.b	$B8,$F8,$F2,$F0,$56,$8F,$EA,$13,$E1,$AE,$19,$78,$67,$FA,$36,$5E
	.dc.b	$17,$FC,$51,$97,$C0,$F8,$BF,$06,$5E,$D9,$F5,$C7,$68,$F0,$47,$BB
	.dc.b	$2F,$8F,$87,$58,$FF,$19,$61,$D6,$BF,$C8,$58,$7C,$CF,$8A,$95,$F1
	.dc.b	$9F,$9F,$F1,$7F,$43,$E2,$64,$FC,$A7,$E9,$FC,$57,$F9,$37,$96,$4F
	.dc.b	$F5,$F7,$FC,$A1,$CB,$CA,$3A,$10,$F8,$9C,$CE,$16,$3D,$A7,$30,$C7
	.dc.b	$1A,$7D,$FC,$BB,$D7,$F9,$37,$2B,$3F,$E5,$DC,$BC,$13,$CB,$1B,$3F
	.dc.b	$87,$F8,$BE,$7C,$5F,$67,$2B,$7F,$5B,$2E,$73,$FE,$3B,$CA,$CF,$F8
	.dc.b	$B7,$2F,$C9,$FC,$3C,$8F,$C6,$BF,$96,$F8,$BF,$CB,$FC,$54,$9F,$C8
	.dc.b	$BF,$9A,$F8,$BF,$F3,$A7,$54,$9F,$F0,$33,$FE,$77,$EB,$E9,$1F,$0B
	.dc.b	$0B,$BF,$E7,$F1,$E9,$3F,$1B,$1B,$3F,$A3,$C7,$A4,$FF,$9C,$71,$B3
	.dc.b	$FE,$81,$C7,$CD,$3E,$82,$1F,$43,$0E,$7A,$7E,$77,$1B,$3F,$9A,$C7
	.dc.b	$9C,$FF,$97,$F1,$B3,$FE,$51,$C3,$F6,$FF,$17,$FB,$8E,$D9,$3F,$B2
	.dc.b	$8F,$DD,$FC,$5C,$8F,$DA,$BF,$BE,$F8,$BF,$F4,$C7,$6F,$F8,$E3,$DA
	.dc.b	$7E,$A7,$FB,$ED,$3F,$D1,$B1,$87,$78,$75,$21,$F8,$AC,$7F,$84,$FC
	.dc.b	$5F,$78,$FC,$6E,$57,$87,$F8,$56,$7C,$A7,$F7,$18,$F5,$7F,$CC,$19
	.dc.b	$78,$5F,$F3,$76,$7D,$1F,$E1,$B8,$D7,$F3,$3E,$67,$5D,$3F,$83,$67
	.dc.b	$67,$F8,$1E,$5D,$F3,$57,$B3,$2E,$D9,$F8,$6F,$FD,$55,$9F,$F2,$2F
	.dc.b	$2E,$47,$F1,$C7,$F9,$2F,$95,$FE,$5B,$FC,$A7,$FB,$2C,$FF,$9F,$9F
	.dc.b	$F3,$03,$7B,$B5,$E5,$7C,$DC,$BF,$8A,$4F,$FA,$D5,$78,$79,$35,$E2
	.dc.b	$76,$9F,$01,$0F,$DB,$E3,$D1,$EC,$43,$F7,$F8,$F9,$1F,$9B,$7E,$0D
	.dc.b	$3E,$EE,$1D,$E7,$E7,$E1,$DD,$EA,$7F,$3A,$C3,$DD,$F1,$9D,$E3,$AE
	.dc.b	$7F,$04,$C6,$FF,$32,$3D,$97,$D5,$F0,$E1,$EC,$7E,$37,$C6,$7C,$9F
	.dc.b	$2A,$46,$0F,$95,$F1,$9F,$34,$8F,$CD,$F8,$BF,$95,$E5,$7C,$BF,$22
	.dc.b	$4F,$E2,$23,$EA,$7C,$67,$D5,$C6,$4F,$D5,$8E,$6F,$C6,$C2,$E9,$E9
	.dc.b	$D5,$C4,$CF,$BD,$FE,$53,$3E,$EF,$CC,$CF,$BB,$D4,$87,$D2,$F8,$CE
	.dc.b	$BC,$FF,$EA,$6C,$19,$E2,$BF,$D4,$D1,$DD,$F5,$1D,$C8,$77,$3F,$1D
	.dc.b	$EA,$7E,$47,$E2,$A4,$60,$FC,$97,$C5,$FE,$54,$8F,$E5,$7E,$2B,$F2
	.dc.b	$5E,$6C,$8F,$D7,$BF,$95,$F3,$3F,$39,$84,$9E,$87,$E3,$3E,$2F,$9B
	.dc.b	$F8,$EC,$2F,$17,$C9,$CD,$97,$5A,$FC,$8E,$56,$BF,$0D,$D5,$FC,$3B
	.dc.b	$3F,$D5,$65,$2F,$F5,$4F,$EB,$7E,$2B,$AE,$FF,$27,$4B,$5F,$23,$3E
	.dc.b	$93,$FB,$BC,$EC,$FE,$B7,$1F,$DB,$7C,$57,$ED,$FE,$2E,$4F,$ED,$E3
	.dc.b	$F7,$5F,$15,$FB,$BF,$8B,$93,$FB,$B8,$FD,$C7,$A5,$23,$F8,$AB,$FB
	.dc.b	$BF,$4F,$AC,$73,$21,$FB,$2F,$8A,$EA,$FF,$05,$BE,$AF,$ED,$AF,$A3
	.dc.b	$FB,$6B,$BC,$5F,$1B,$C3,$63,$FC,$3F,$19,$7F,$C3,$DF,$E2,$5F,$15
	.dc.b	$DA,$3F,$80,$E5,$DB,$0F,$A9,$97,$4D,$FF,$90,$67,$6F,$F8,$96,$5F
	.dc.b	$EF,$5F,$CA,$51,$48,$63,$0F,$D3,$47,$FB,$E1,$BF,$17,$C7,$E0,$25
	.dc.b	$6F,$1E,$7E,$03,$05,$FF,$23,$F5,$A5,$FF,$40,$7F,$93,$7A,$DE,$3F
	.dc.b	$89,$F0,$CF,$F8,$6A,$FC,$AF,$51,$87,$59,$F5,$61,$E2,$9E,$19,$61
	.dc.b	$EC,$43,$F9,$2E,$3E,$07,$A5,$F3,$6E,$1E,$FC,$BD,$6D,$8F,$0A,$F8
	.dc.b	$BE,$D1,$E0,$38,$E1,$8D,$3E,$E6,$3F,$02,$7C,$27,$F6,$0D,$3E,$97
	.dc.b	$C5,$7D,$3F,$8A,$93,$E2,$8F,$AD,$F1,$3F,$5F,$E2,$A4,$F9,$A3,$FA
	.dc.b	$77,$B9,$FD,$3F,$DC,$93,$FD,$B2,$3A,$BC,$C8,$7C,$DF,$89,$EA,$FC
	.dc.b	$0B,$EA,$FD,$1B,$EA,$FF,$76,$BE,$AF,$F4,$EB,$FC,$55,$F6,$8F,$B1
	.dc.b	$86,$1F,$31,$D4,$87,$53,$CC,$75,$21,$D4,$FA,$7F,$15,$F9,$72,$3F
	.dc.b	$5B,$E1,$BF,$35,$F0,$FF,$9B,$F8,$79,$3F,$6A,$3F,$B9,$7B,$DF,$DC
	.dc.b	$FD,$F9,$3F,$E1,$08,$FD,$15,$F4,$8F,$C8,$61,$87,$91,$F8,$4C,$3A
	.dc.b	$BF,$95,$C3,$AB,$FE,$1A,$C3,$AB,$FD,$CB,$0E,$4F,$E6,$F0,$BC,$5F
	.dc.b	$AC,$F8,$7F,$4C,$F5,$21,$D4,$F3,$1D,$48,$75,$3F,$2F,$F0,$DF,$BB
	.dc.b	$23,$F9,$AF,$87,$FD,$F7,$C3,$7E,$FF,$E1,$A4,$FE,$7A,$3F,$C2,$DF
	.dc.b	$05,$FE,$17,$F8,$29,$3F,$E3,$48,$FE,$0D,$7D,$23,$F6,$98,$61,$E4
	.dc.b	$7E,$A3,$0E,$AF,$EE,$70,$EA,$FF,$8F,$B0,$EA,$FF,$85,$B0,$E4,$FE
	.dc.b	$FF,$0B,$97,$F1,$0F,$86,$F4,$CF,$52,$1D,$4F,$31,$D4,$87,$53,$F8
	.dc.b	$47,$C4,$BF,$4E,$FF,$27,$37,$C0,$FE,$1B,$7F,$E1,$B8,$C5,$7F,$0F
	.dc.b	$1F,$8C,$7F,$34,$70,$D7,$FD,$A5,$1D,$7E,$07,$9D,$FF,$2C,$8C,$4F
	.dc.b	$0F,$1B,$2F,$F3,$0B,$E5,$87,$F4,$1F,$88,$FF,$24,$1A,$9F,$0D,$5A
	.dc.b	$F0,$47,$F2,$43,$25,$3A,$45,$69,$63,$8B,$FC,$B1,$AE,$8E,$4E,$1E
	.dc.b	$A6,$23,$A3,$83,$C5,$1B,$3D,$78,$75,$9F,$41,$F0,$30,$FA,$5F,$05
	.dc.b	$8B,$F8,$84,$3E,$5E,$7F,$F0,$02,$F9,$46,$72,$FF,$28,$3E,$47,$FB
	.dc.b	$47,$C4,$7F,$5B,$F8,$99,$3F,$CE,$9F,$EB,$DF,$13,$F4,$3C,$9F,$EA
	.dc.b	$78,$5F,$2B,$FD,$21,$18,$BF,$E8,$71,$FE,$69,$F4,$5F,$13,$FE,$7B
	.dc.b	$24,$FF,$A6,$3E,$23,$AB,$DC,$87,$F5,$9C,$7B,$63,$23,$AB,$E1,$E3
	.dc.b	$E7,$1F,$E7,$8B,$BA,$FF,$80,$F0,$EB,$3C,$CF,$EA,$B8,$7A,$77,$E9
	.dc.b	$1F,$F0,$03,$3F,$59,$31,$E7,$97,$FA,$87,$2B,$5F,$E9,$3C,$7E,$17
	.dc.b	$C1,$48,$F8,$2F,$C5,$F8,$3F,$8D,$EC,$C9,$F8,$AF,$C9,$F6,$7F,$D9
	.dc.b	$5D,$FF,$E9,$1D,$FF,$C7,$ED,$CF,$FD,$96,$DB,$7B,$C8,$FF,$54,$C7
	.dc.b	$A7,$1F,$13,$EB,$CB,$C4,$F8,$08,$7B,$F2,$F1,$3F,$E0,$5B,$E9,$F1
	.dc.b	$3E,$42,$46,$70,$EF,$97,$FA,$07,$76,$5D,$AB,$C4,$7E,$0B,$D5,$F8
	.dc.b	$0F,$73,$C3,$5C,$75,$EB,$1F,$ED,$CC,$3D,$D9,$78,$A3,$C1,$C9,$F8
	.dc.b	$FE,$99,$1F,$8E,$7F,$25,$E8,$FE,$4F,$D1,$93,$F9,$27,$F2,$DE,$6F
	.dc.b	$CE,$F8,$9F,$9F,$F1,$12,$BD,$91,$C9,$E8,$43,$F0,$97,$D1,$FC,$35
	.dc.b	$DD,$7E,$16,$1D,$67,$E3,$61,$D5,$F6,$61,$D5,$E6,$7E,$9B,$0F,$D5
	.dc.b	$7A,$1C,$F9,$7F,$C8,$3C,$4C,$F1,$7F,$E3,$88,$65,$DE,$3E,$46,$5D
	.dc.b	$F3,$FC,$4E,$7F,$B5,$F2,$BF,$6D,$E5,$CB,$FD,$AC,$7E,$E7,$C6,$91
	.dc.b	$FB,$87,$F7,$9E,$3F,$E5,$3E,$27,$F2,$BF,$0D,$2B,$F0,$51,$D5,$E6
	.dc.b	$43,$F5,$5E,$3F,$47,$F5,$97,$73,$FC,$46,$1D,$67,$F1,$98,$75,$7F
	.dc.b	$05,$87,$F0,$DC,$3B,$47,$ED,$71,$C3,$E7,$21,$D8,$FE,$23,$7D,$5F
	.dc.b	$C6,$5F,$5C,$3F,$5B,$87,$F1,$9F,$12,$57,$F1,$87,$F8,$E7,$7F,$F1
	.dc.b	$DE,$F9,$3F,$C7,$1F,$E4,$5F,$95,$7E,$5B,$AB,$FD,$A5,$3F,$99,$F8
	.dc.b	$AF,$DD,$9F,$BC,$F8,$AF,$EA,$1D,$3F,$D0,$3C,$A3,$49,$5F,$D3,$5A
	.dc.b	$FF,$4A,$6B,$FD,$19,$AF,$F5,$16,$BF,$E7,$CC,$3F,$6B,$D3,$FE,$8A
	.dc.b	$6C,$67,$B3,$D3,$CD,$EA,$E7,$93,$FE,$7A,$A4,$7B,$B8,$3F,$8A,$FC
	.dc.b	$57,$F1,$6F,$89,$97,$FC,$0E,$3F,$B6,$76,$7B,$49,$1F,$D9,$DA,$FF
	.dc.b	$64,$6B,$FD,$81,$AF,$F6,$B6,$BF,$DF,$5F,$EE,$59,$31,$72,$FD,$22
	.dc.b	$57,$A7,$7F,$AB,$C1,$23,$D8,$FD,$8C,$7F,$97,$FC,$57,$F3,$0F,$87
	.dc.b	$95,$FC,$76,$3F,$C1,$1E,$27,$C2,$8F,$F0,$13,$5F,$F0,$33,$5F,$F0
	.dc.b	$53,$5F,$F5,$7B,$5F,$F6,$2B,$5F,$E4,$FB,$FF,$80,$A3,$04,$70,$48
	.dc.b	$FF,$0A,$69,$F0,$76,$FC,$8D,$C9,$FB,$11,$F9,$4C,$24,$7D,$B7,$F2
	.dc.b	$FF,$0F,$FD,$38,$FE,$9F,$F1,$32,$7F,$9E,$47,$E7,$2A,$47,$43,$F0
	.dc.b	$93,$D1,$75,$39,$3F,$1D,$86,$63,$13,$F9,$3C,$3C,$0F,$F4,$9B,$F0
	.dc.b	$3F,$CF,$E3,$BB,$D2,$FF,$31,$86,$2F,$9E,$5F,$81,$C3,$B5,$78,$2F
	.dc.b	$DC,$7E,$67,$0E,$EF,$F4,$B9,$EF,$7F,$DB,$27,$F7,$3F,$13,$2B,$F4
	.dc.b	$0F,$EF,$3E,$23,$F7,$BF,$13,$27,$F4,$8F,$F0,$0F,$86,$FE,$EF,$5F
	.dc.b	$DE,$2A,$57,$F6,$58,$E4,$F4,$21,$ED,$3F,$6D,$1D,$5C,$67,$F9,$DB
	.dc.b	$C6,$3F,$03,$95,$8F,$82,$FF,$B9,$61,$E0,$7F,$B4,$63,$E0,$79,$D2
	.dc.b	$71,$7C,$FE,$1F,$AE,$BF,$C1,$72,$B8,$7F,$09,$CB,$B3,$FD,$CF,$0E
	.dc.b	$D8,$7F,$85,$32,$FE,$43,$F1,$3F,$C8,$BE,$26,$5F,$F0,$98,$FF,$23
	.dc.b	$7E,$55,$FE,$6D,$BF,$B2,$FF,$C0,$1D,$8D,$35,$A7,$F8,$7B,$4F,$F1
	.dc.b	$1E,$93,$1F,$F0,$84,$78,$1E,$C4,$3E,$4D,$3C,$9F,$0C,$78,$BE,$2A
	.dc.b	$23,$67,$F8,$C2,$BB,$3F,$E1,$1B,$C3,$BC,$BD,$37,$E1,$8F,$9B,$9F
	.dc.b	$BA,$3B,$C6,$F2,$ED,$9F,$BB,$2F,$0C,$F8,$0F,$59,$FE,$2B,$C7,$C3
	.dc.b	$8F,$F9,$03,$2F,$93,$F1,$1F,$2B,$E2,$65,$F5,$18,$FC,$A9,$F9,$FF
	.dc.b	$0B,$F4,$3E,$26,$56,$EB,$E7,$CF,$F9,$3B,$8F,$FC,$9F,$BC,$AF,$F1
	.dc.b	$E4,$78,$1E,$1C,$0C,$71,$1F,$3B,$E2,$3B,$BA,$CB,$BB,$F4,$B1,$EC
	.dc.b	$FF,$91,$F1,$EC,$FF,$8F,$71,$E7,$AF,$E1,$31,$BC,$1F,$23,$36,$7D
	.dc.b	$B0,$FB,$B9,$F6,$C3,$F1,$79,$F6,$9F,$F2,$46,$7D,$A7,$FC,$D3,$9F
	.dc.b	$FA,$9E,$9E,$B7,$E5,$3E,$26,$61,$FE,$DC,$9C,$B7,$D6,$E0,$47,$5E
	.dc.b	$A7,$F9,$DB,$9B,$FD,$A3,$EA,$BF,$54,$FC,$CD,$7F,$9D,$1B,$1F,$E7
	.dc.b	$59,$F1,$65,$2B,$C0,$F9,$7D,$7E,$D1,$FE,$E2,$C9,$96,$35,$FF,$76
	.dc.b	$47,$FB,$12,$7F,$D0,$3E,$6B,$1F,$1E,$7E,$03,$C3,$19,$9F,$76,$3D
	.dc.b	$DA,$7D,$EF,$8A,$ED,$C9,$FE,$EF,$C7,$C3,$3F,$0C,$F8,$27,$FA,$07
	.dc.b	$1E,$EC,$30,$7E,$23,$E2,$BF,$80,$FC,$44,$8C,$1F,$8C,$F8,$9F,$E0
	.dc.b	$A4,$7F,$1F,$F1,$5E,$AE,$EF,$5F,$7C,$9F,$F4,$A6,$FF,$C1,$F1,$FE
	.dc.b	$1A,$48,$FC,$BE,$FC,$DF,$E0,$78,$DE,$0F,$E2,$39,$75,$9E,$78,$F5
	.dc.b	$7F,$9A,$71,$FD,$BC,$FF,$8C,$E3,$27,$A9,$FA,$4F,$8A,$E9,$C1,$F9
	.dc.b	$7C,$EC,$FE,$D3,$AD,$9F,$78,$EB,$B7,$F1,$57,$C2,$7B,$32,$FE,$4F
	.dc.b	$F1,$1F,$CA,$3E,$26,$4F,$EC,$A3,$F9,$5F,$C5,$7F,$2C,$F8,$A9,$3F
	.dc.b	$B6,$8F,$C0,$F8,$1F,$82,$F0,$A4,$FC,$88,$EA,$F3,$21,$FC,$A3,$E2
	.dc.b	$3A,$BF,$C0,$2F,$AB,$FC,$B2,$FA,$BF,$4A,$FA,$BE,$CB,$FE,$7F,$72
	.dc.b	$3F,$84,$BF,$D0,$BE,$1F,$AC,$7F,$36,$C3,$AF,$2F,$F4,$4C,$7A,$CF
	.dc.b	$C6,$C7,$AC,$FF,$82,$B9,$3F,$5D,$97,$F3,$9A,$FE,$CF,$F9,$5F,$F8
	.dc.b	$39,$BF,$0F,$9B,$FF,$03,$CA,$31,$C9,$FC,$D2,$38,$4E,$BE,$23,$37
	.dc.b	$FC,$B2,$5F,$11,$8F,$21,$18,$9F,$E5,$8A,$BF,$63,$DD,$27,$F2,$EF
	.dc.b	$ED,$3C,$7E,$55,$E0,$CB,$C4,$60,$CF,$A9,$AC,$37,$94,$C7,$BC,$75
	.dc.b	$7F,$17,$8D,$CF,$E3,$73,$B3,$E7,$5E,$6C,$6B,$F2,$3C,$4E,$CF,$AB
	.dc.b	$2E,$CF,$CC,$CB,$A6,$7F,$D9,$F8,$3F,$8F,$69,$8A,$FE,$3D,$97,$B3
	.dc.b	$C3,$91,$F1,$1F,$7F,$89,$F0,$3C,$59,$3F,$21,$F8,$7E,$E7,$F1,$8F
	.dc.b	$2B,$F8,$CF,$91,$2B,$F7,$F1,$C9,$E8,$43,$F0,$37,$D1,$FC,$1D,$DE
	.dc.b	$2F,$C0,$E1,$D6,$7F,$84,$E1,$D5,$FD,$9E,$1D,$5E,$67,$DC,$C3,$EB
	.dc.b	$FB,$5C,$F6,$FC,$46,$56,$BF,$0D,$8F,$57,$F7,$B8,$F5,$CB,$E9,$E5
	.dc.b	$F8,$0F,$86,$95,$F8,$87,$F9,$BF,$95,$F8,$4F,$87,$FC,$2F,$C2,$4A
	.dc.b	$FC,$6C,$69,$FE,$43,$E5,$7F,$3E,$30,$FF,$0C,$AE,$AF,$E9,$2F,$38
	.dc.b	$FE,$8A,$FA,$4F,$EA,$30,$F0,$3F,$CB,$2F,$A3,$D8,$C1,$0E,$C5,$AF
	.dc.b	$D6,$FC,$3F,$28,$E8,$7E,$92,$B1,$3F,$94,$AF,$1E,$DF,$8B,$E9,$47
	.dc.b	$F5,$8F,$33,$F4,$BF,$0F,$31,$FD,$6B,$FA,$8F,$86,$FD,$4F,$C3,$FE
	.dc.b	$AB,$E1,$E5,$7E,$CA,$3F,$B0,$F9,$B2,$3F,$A3,$BF,$D9,$3C,$CE,$51
	.dc.b	$D0,$87,$F0,$AB,$EA,$FE,$F2,$FA,$BF,$C3,$AF,$A3,$FC,$EA,$EE,$BF
	.dc.b	$B0,$C3,$AC,$F3,$3F,$82,$E1,$FC,$03,$E1,$FA,$C7,$F1,$8C,$7A,$DF
	.dc.b	$F1,$6C,$79,$ED,$F1,$72,$B7,$FE,$0B,$CB,$F8,$4F,$C3,$C8,$FE,$22
	.dc.b	$FF,$15,$F8,$67,$C4,$5F,$1B,$11,$2B,$FC,$84,$FF,$92,$1F,$ED,$7E
	.dc.b	$9F,$F8,$23,$D3,$91,$83,$77,$DD,$FE,$5F,$87,$78,$ED,$1D,$2F,$B3
	.dc.b	$FE,$9A,$7B,$3F,$D8,$2F,$AB,$FD,$8A,$FF,$C2,$F1,$CE,$3B,$90,$E5
	.dc.b	$F0,$5D,$9E,$95,$DB,$6F,$F6,$1B,$D9,$FE,$43,$5D,$9F,$E4,$77,$DA
	.dc.b	$BE,$0F,$C3,$FC,$72,$3F,$13,$E1,$FE,$57,$C3,$FC,$BF,$87,$93,$F2
	.dc.b	$23,$FC,$5B,$EA,$FF,$8B,$BD,$69,$3F,$E1,$48,$FA,$57,$D2,$3E,$16
	.dc.b	$18,$79,$1F,$4B,$0E,$AF,$C7,$C3,$AB,$FE,$1E,$C3,$AB,$FE,$08,$C3
	.dc.b	$93,$F3,$70,$B9,$7C,$2F,$87,$F4,$CF,$52,$1D,$4F,$31,$D4,$87,$53
	.dc.b	$FD,$18,$3F,$1C,$78,$9F,$D9,$23,$0D,$8C,$C7,$07,$F7,$6D,$ED,$E1
	.dc.b	$E2,$3C,$BA,$71,$89,$3E,$09,$FE,$B0,$DF,$2F,$2F,$FB,$9C,$F2,$4F
	.dc.b	$F8,$05,$97,$E0,$70,$68,$A5,$32,$ED,$E3,$1A,$74,$60,$E3,$1B,$7F
	.dc.b	$C2,$0D,$7F,$D9,$4D,$78,$9A,$8F,$4E,$5F,$E1,$59,$F3,$37,$C6,$CF
	.dc.b	$5E,$EF,$E4,$1F,$0F,$D3,$CB,$DB,$0A,$FF,$A9,$3D,$4F,$E0,$1F,$0D
	.dc.b	$23,$F7,$AF,$F0,$3F,$88,$FE,$09,$F1,$1F,$C1,$7E,$1E,$57,$F0,$28
	.dc.b	$FF,$43,$FB,$32,$3F,$CD,$AF,$FA,$2B,$DA,$E5,$1D,$08,$7E,$F6,$FA
	.dc.b	$BF,$C3,$2F,$AB,$FC,$06,$FA,$3F,$E6,$0B,$BB,$7F,$85,$B0,$EB,$3C
	.dc.b	$CF,$E2,$B8,$7F,$1B,$F8,$8E,$B1,$FC,$2F,$1E,$B7,$FC,$03,$1E,$7B
	.dc.b	$7F,$4B,$CA,$DF,$FA,$C7,$2F,$E4,$9F,$0D,$23,$F9,$0B,$FC,$9F,$E1
	.dc.b	$FF,$94,$7C,$37,$F2,$9F,$87,$95,$FC,$9A,$3F,$D5,$7F,$03,$23,$FD
	.dc.b	$22,$FF,$AB,$BE,$07,$94,$74,$21,$FC,$86,$FA,$BF,$CC,$2F,$AB,$FC
	.dc.b	$96,$FA,$3F,$E8,$4B,$BA,$FF,$8E,$B0,$EB,$3C,$CF,$E7,$38,$7F,$3F
	.dc.b	$F8,$6E,$B1,$FC,$BF,$1E,$B7,$FC,$93,$1E,$7A,$FF,$6F,$CA,$DF,$FB
	.dc.b	$57,$2F,$E9,$1F,$0F,$23,$FA,$2B,$FD,$2F,$E1,$FF,$CC,$1F,$95,$FF
	.dc.b	$B7,$3C,$C2,$F8,$6C,$9A,$BF,$F7,$1F,$EA,$BF,$E5,$D5,$F0,$28,$C3
	.dc.b	$74,$DF,$C3,$F7,$AF,$09,$0F,$E9,$59,$75,$43,$75,$63,$3F,$2E,$5E
	.dc.b	$4A,$DD,$69,$4F,$A9,$F3,$D4,$76,$FF,$97,$F2,$F3,$9F,$E8,$98,$FE
	.dc.b	$D3,$87,$9D,$F2,$9C,$0E,$B9,$FF,$AF,$5F,$35,$79,$63,$E5,$1F,$26
	.dc.b	$38,$5F,$A1,$B1,$FC,$77,$0F,$F5,$06,$1E,$2B,$F3,$57,$FB,$5A,$3C
	.dc.b	$8F,$63,$C3,$DB,$F6,$08,$F9,$74,$FA,$5F,$07,$F7,$3B,$7E,$EF,$87
	.dc.b	$2B,$C3,$1F,$6A,$7E,$DF,$C2,$49,$E5,$8F,$FA,$7D,$B3,$D9,$F1,$6B
	.dc.b	$F8,$6C,$B9,$47,$43,$DB,$8F,$77,$7A,$B2,$DC,$7B,$72,$F0,$3E,$FC
	.dc.b	$73,$AF,$C4,$6E,$BF,$37,$C3,$F6,$57,$62,$1D,$B0,$DB,$F9,$4D,$7D
	.dc.b	$4F,$53,$DB,$F0,$DF,$A0,$F1,$64,$7D,$47,$F4,$7E,$37,$E9,$3C,$49
	.dc.b	$3F,$61,$FD,$3F,$8B,$FA,$2F,$86,$FD,$1F,$C3,$4A,$F9,$31,$FD,$AF
	.dc.b	$F1,$EF,$B3,$1F,$D6,$DA,$FF,$B7,$5A,$FE,$81,$A9,$A4,$E9,$FE,$B4
	.dc.b	$D8,$FE,$C2,$D7,$FB,$2B,$D3,$9B,$F7,$7E,$F7,$EF,$BD,$99,$5F,$92
	.dc.b	$8F,$E0,$3E,$CF,$F0,$2F,$56,$4F,$E5,$A3,$F7,$FF,$0F,$FC,$03,$E1
	.dc.b	$A4,$FE,$1A,$3F,$84,$DC,$8E,$87,$E7,$BD,$7E,$AF,$E8,$30,$EA,$FE
	.dc.b	$13,$0E,$AF,$E4,$30,$EA,$FE,$87,$0F,$E2,$BE,$9C,$8F,$D4,$BF,$C6
	.dc.b	$3D,$4F,$E3,$3E,$64,$9F,$D6,$BF,$C7,$3C,$CF,$E2,$FF,$0D,$FC,$63
	.dc.b	$E1,$E5,$7E,$7A,$3E,$B7,$77,$CC,$24,$7C,$C6,$BF,$31,$AF,$DB,$6B
	.dc.b	$F6,$DA,$FD,$B7,$FC,$2B,$C3,$FC,$B3,$C9,$FE,$5E,$49,$FE,$07,$E4
	.dc.b	$FF,$20,$7F,$5B,$F9,$67,$F2,$DA,$34,$9C,$DF,$CC,$BE,$23,$F9,$9F
	.dc.b	$C4,$4B,$FD,$CC,$7E,$47,$C5,$FC,$39,$23,$F0,$ED,$7F,$0E,$D7,$F2
	.dc.b	$CD,$7F,$2C,$D7,$F2,$CF,$B2,$FF,$5B,$E2,$FF,$31,$C7,$CF,$3E,$B8
	.dc.b	$C4,$A3,$1E,$2F,$F3,$B8,$E2,$DF,$0B,$38,$7F,$3A,$94,$63,$0F,$86
	.dc.b	$D8,$76,$FF,$54,$F8,$7F,$EA,$BF,$11,$30,$FE,$37,$1B,$E1,$91,$FC
	.dc.b	$D2,$FC,$2F,$7F,$03,$A6,$32,$34,$F7,$C7,$B4,$7F,$26,$C7,$B3,$FC
	.dc.b	$DF,$1E,$AF,$DA,$C6,$EB,$F0,$B1,$E9,$3F,$7B,$1B,$3F,$1B,$1E,$B3
	.dc.b	$FD,$07,$1E,$B3,$FD,$1B,$1F,$81,$D3,$2B,$D4,$FC,$3E,$AF,$34,$7A
	.dc.b	$08,$7E,$33,$0F,$EF,$3F,$0F,$23,$F9,$FB,$FD,$F3,$E1,$F9,$47,$42
	.dc.b	$1F,$94,$AE,$AF,$A4,$EB,$8F,$A5,$E6,$43,$FA,$D5,$75,$7E,$84,$FA
	.dc.b	$5F,$CF,$E1,$E0,$7C,$C7,$4E,$0F,$C2,$3F,$5F,$1F,$43,$D0,$ED,$87
	.dc.b	$F6,$AC,$7F,$09,$CF,$27,$EF,$3F,$87,$E8,$F4,$D7,$E5,$36,$FA,$38
	.dc.b	$FE,$4F,$1F,$F0,$EF,$C2,$CC,$7F,$B1,$BF,$E2,$0F,$86,$E5,$1D,$08
	.dc.b	$7E,$E3,$1E,$B7,$E8,$8F,$8C,$7B,$0F,$EF,$95,$D5,$A7,$F7,$0A,$EA
	.dc.b	$F3,$3F,$2D,$5F,$9F,$E6,$EB,$1F,$A5,$C7,$D2,$FE,$9F,$1E,$99,$7F
	.dc.b	$83,$31,$B9,$7F,$87,$F2,$FD,$47,$3C,$C3,$F4,$4F,$EA,$F8,$7A,$F3
	.dc.b	$7F,$10,$CA,$F3,$39,$3F,$C8,$3F,$0F,$31,$FF,$03,$BF,$E4,$4F,$87
	.dc.b	$E5,$1D,$08,$7F,$1C,$C7,$9E,$1D,$2F,$94,$FF,$87,$EE,$ED,$FE,$24
	.dc.b	$C3,$9C,$F1,$3F,$71,$87,$F0,$1E,$6E,$71,$F8,$53,$F7,$98,$72,$DB
	.dc.b	$FC,$7D,$8D,$9F,$F0,$26,$1F,$C2,$79,$3F,$85,$70,$49,$FE,$09,$1D
	.dc.b	$1F,$B1,$9F,$F1,$AC,$7F,$8A,$FA,$2F,$D5,$9F,$11,$87,$FC,$A3,$7F
	.dc.b	$E3,$36,$BF,$E3,$99,$E1,$3E,$18,$EE,$43,$FA,$0F,$1F,$87,$0F,$D6
	.dc.b	$47,$F4,$4C,$3B,$57,$7E,$29,$3D,$E3,$F9,$25,$C8,$FE,$3C,$FF,$27
	.dc.b	$D7,$A6,$DF,$9F,$9F,$8E,$71,$3D,$17,$D1,$F4,$67,$EE,$7F,$94,$6D
	.dc.b	$FC,$CF,$F2,$BC,$47,$F9,$11,$BA,$7F,$40,$9E,$F3,$F9,$53,$B9,$D3
	.dc.b	$E2,$3E,$41,$1F,$91,$F1,$3D,$71,$FE,$7D,$52,$7F,$CD,$78,$7C,$CC
	.dc.b	$3E,$29,$32,$F6,$4F,$C7,$3E,$96,$3D,$5F,$01,$D8,$E7,$87,$76,$9D
	.dc.b	$A3,$D3,$1F,$A3,$8F,$4D,$7F,$A3,$F7,$FC,$15,$75,$8C,$66,$8F,$C2
	.dc.b	$67,$F1,$CF,$91,$F1,$33,$0F,$5C,$7F,$5B,$E0,$EF,$CD,$FA,$F9,$FC
	.dc.b	$55,$7E,$37,$E1,$E4,$60,$F8,$6F,$57,$99,$83,$CF,$AF,$C4,$CF,$E3
	.dc.b	$47,$F6,$DC,$3F,$31,$7D,$23,$EF,$DE,$1E,$47,$DD,$7F,$00,$7E,$07
	.dc.b	$E2,$3D,$11,$FB,$13,$F4,$F8,$7F,$7D,$BE,$51,$D0,$FD,$25,$63,$0F
	.dc.b	$F0,$46,$EF,$89,$FC,$67,$C4,$CB,$FB,$F1,$D9,$EA,$7E,$2C,$FB,$C7
	.dc.b	$E5,$B2,$FD,$91,$1F,$D9,$47,$E5,$C8,$F4,$3F,$17,$F1,$1D,$70,$F8
	.dc.b	$BC,$5F,$7E,$7A,$C7,$DC,$BE,$91,$DC,$EA,$72,$21,$FA,$3A,$FD,$37
	.dc.b	$C4,$75,$D7,$E9,$12,$7E,$91,$FC,$1E,$BA,$DF,$78,$EA,$7E,$B7,$E1
	.dc.b	$BF,$60,$47,$F4,$7F,$0F,$D5,$F8,$93,$EA,$E1,$FC,$54,$7E,$EB,$5F
	.dc.b	$E2,$D9,$C8,$FE,$2D,$1F,$BB,$BE,$91,$DC,$EA,$7E,$28,$EB,$97,$A7
	.dc.b	$4F,$F2,$AF,$07,$F2,$1B,$E9,$5F,$C2,$30,$C3,$C8,$FE,$0D,$87,$77
	.dc.b	$F7,$7F,$11,$E9,$C3,$E8,$67,$E9,$C7,$8D,$73,$3F,$96,$61,$89,$FC
	.dc.b	$4F,$C4,$7F,$13,$24,$FF,$31,$DF,$FC,$97,$F4,$2F,$D4,$9C,$38,$6D
	.dc.b	$7F,$BF,$67,$53,$FE,$B9,$F1,$3F,$9E,$5F,$F3,$DE,$29,$7D,$48,$7F
	.dc.b	$21,$C7,$C2,$FF,$28,$9E,$D7,$23,$B6,$3E,$07,$FD,$F8,$7F,$AF,$F0
	.dc.b	$EE,$FF,$49,$C3,$BB,$CC,$EF,$97,$F0,$EC,$37,$1F,$CC,$7E,$2B,$B2
	.dc.b	$97,$12,$1A,$7B,$9F,$1E,$7E,$F7,$AC,$EF,$AF,$FA,$F6,$3D,$B0,$FE
	.dc.b	$3B,$C5,$FD,$07,$E2,$26,$1F,$CD,$1F,$E8,$9F,$13,$CE,$3C,$58,$F8
	.dc.b	$73,$F0,$46,$0F,$1E,$FF,$0F,$1F,$3C,$7F,$B8,$F7,$FE,$D9,$C9,$FD
	.dc.b	$B7,$86,$4F,$F6,$D8,$EE,$F8,$48,$7F,$53,$CB,$C5,$7D,$31,$F0,$B7
	.dc.b	$1F,$06,$5F,$EF,$3C,$BB,$E1,$FD,$E7,$2E,$EF,$F6,$2C,$BB,$BD,$4F
	.dc.b	$EA,$79,$7F,$63,$F8,$AE,$F8,$7B,$F1,$EB,$5E,$EE,$0C,$AE,$0F,$F0
	.dc.b	$56,$3D,$B4,$FE,$93,$9F,$8A,$BF,$B6,$7C,$57,$F6,$F2,$3F,$E0,$DE
	.dc.b	$2E,$7B,$FB,$34,$F0,$67,$DB,$3F,$F0,$B5,$7F,$85,$78,$26,$1C,$48
	.dc.b	$76,$3F,$B6,$E3,$FE,$1B,$9E,$BA,$FA,$B2,$ED,$5F,$54,$E3,$A7,$F6
	.dc.b	$0C,$3A,$C7,$F8,$22,$F9,$3A,$CF,$F1,$9D,$E9,$FF,$05,$7C,$4F,$6D
	.dc.b	$7F,$33,$8F,$E0,$B0,$FD,$21,$C4,$F9,$17,$C5,$C6,$BF,$E3,$3E,$1F
	.dc.b	$F6,$37,$A3,$FE,$EB,$9F,$C7,$74,$A3,$18,$7F,$B5,$DB,$A7,$E9,$67
	.dc.b	$5F,$15,$E3,$BF,$D2,$9F,$A9,$BF,$27,$C0,$7F,$B5,$23,$FC,$7D,$CB
	.dc.b	$FE,$41,$E3,$91,$83,$86,$3D,$72,$63,$E6,$98,$68,$65,$E2,$70,$7C
	.dc.b	$23,$F5,$1C,$5F,$48,$FA,$6E,$73,$FF,$08,$E5,$E2,$8F,$F1,$CE,$5F
	.dc.b	$E5,$58,$F8,$B8,$49,$F0,$61,$DD,$F8,$F8,$BE,$3F,$E1,$47,$83,$93
	.dc.b	$F3,$6F,$7D,$3E,$DE,$3D,$F3,$FF,$10,$E7,$DE,$BD,$9F,$5A,$FD,$BD
	.dc.b	$F0,$7F,$EF,$E8,$C3,$63,$33,$F0,$38,$3E,$73,$FC,$99,$B8,$7E,$E7
	.dc.b	$93,$F0,$F9,$FE,$E6,$B8,$75,$7F,$1F,$7F,$87,$53,$F1,$67,$8F,$98
	.dc.b	$FE,$1E,$65,$FD,$AC,$FC,$F3,$F1,$58,$F8,$9B,$7F,$8D,$7C,$05,$8D
	.dc.b	$38,$36,$6C,$E8,$64,$87,$EA,$CE,$51,$B7,$E7,$19,$AF,$C8,$C7,$E7
	.dc.b	$BE,$3B,$CF,$5F,$13,$1F,$1D,$7F,$1D,$F1,$73,$7F,$C4,$23,$E1,$79
	.dc.b	$1F,$A6,$F8,$BF,$D3,$FC,$74,$CF,$F0,$7C,$C9,$E3,$1F,$30,$FE,$B3
	.dc.b	$8F,$FA,$7F,$B7,$FD,$47,$E0,$7E,$AA,$7F,$8E,$37,$B7,$FD,$4D,$3D
	.dc.b	$E7,$AD,$D7,$F1,$FF,$1B,$D9,$FB,$DA,$FF,$75,$C7,$B4,$7E,$97,$06
	.dc.b	$3D,$4E,$DD,$EC,$BF,$80,$E3,$D7,$6F,$DF,$67,$8C,$7D,$47,$62,$C7
	.dc.b	$C7,$7F,$D7,$15,$EC,$61,$83,$E2,$FC,$67,$F0,$EF,$8C,$91,$83,$9F
	.dc.b	$0F,$D2,$23,$FD,$7B,$5F,$8F,$DF,$23,$D9,$AF,$F0,$46,$F7,$7F,$1A
	.dc.b	$C6,$47,$53,$F6,$BF,$19,$F1,$F4,$FC,$1C,$7D,$43,$89,$D3,$0E,$CF
	.dc.b	$FB,$47,$0F,$DE,$3F,$7A,$BA,$BF,$C0,$2F,$F9,$4C,$75,$CB,$D3,$E0
	.dc.b	$FF,$8A,$73,$FB,$F9,$F4,$CF,$FD,$45,$97,$A7,$2E,$D3,$FC,$CB,$E3
	.dc.b	$3F,$99,$FC,$64,$9F,$C6,$C7,$57,$F3,$99,$7F,$67,$C7,$F1,$7D,$3F
	.dc.b	$8C,$E1,$95,$F1,$A3,$A3,$C8,$87,$F1,$8E,$EE,$8F,$EA,$35,$F4,$3C
	.dc.b	$4F,$9F,$5D,$1F,$89,$5F,$D2,$2B,$AC,$7F,$22,$C7,$0F,$DC,$3C,$E4
	.dc.b	$BE,$AF,$23,$A9,$0C,$1F,$47,$A3,$FA,$AF,$C5,$48,$FC,$0C,$7A,$3A
	.dc.b	$FF,$75,$1F,$87,$32,$7E,$1B,$9B,$F5,$E7,$EC,$26,$46,$8E,$C7,$EE
	.dc.b	$BE,$2F,$A7,$57,$EA,$E3,$5C,$7F,$62,$B8,$C7,$E1,$79,$7F,$B4,$DE
	.dc.b	$2B,$53,$CC,$F8,$47,$F1,$43,$E7,$78,$9F,$EB,$97,$AF,$17,$F9,$27
	.dc.b	$95,$8F,$38,$C6,$3F,$DD,$BE,$32,$5F,$F1,$07,$FB,$CF,$C6,$7F,$7A
	.dc.b	$F8,$C9,$3F,$C5,$1F,$EF,$BF,$19,$FC,$43,$B3,$F8,$8F,$5C,$AF,$D5
	.dc.b	$C7,$F3,$DC,$BF,$A1,$71,$3E,$03,$F9,$63,$6B,$F9,$A3,$5F,$E7,$51
	.dc.b	$A2,$3A,$6B,$A3,$04,$30,$7F,$84,$7C,$AF,$F0,$87,$C6,$7F,$85,$09
	.dc.b	$1F,$CB,$BE,$33,$FC,$2F,$F1,$92,$3F,$98,$BE,$97,$FC,$07,$F9,$6F
	.dc.b	$F8,$8F,$97,$F5,$FD,$7F,$CC,$3C,$09,$5F,$C2,$A3,$82,$D1,$F1,$7F
	.dc.b	$00,$91,$F0,$1A,$FF,$59,$6B,$FD,$81,$AF,$CA,$6B,$FE,$8B,$8E,$AF
	.dc.b	$F6,$3D,$18,$BE,$37,$D5,$18,$6C,$66,$7F,$C2,$2F,$F8,$D5,$BF,$1B
	.dc.b	$A2,$3A,$EC,$FE,$9F,$E0,$7F,$50,$EF,$93,$FC,$AA,$3E,$0E,$4F,$80
	.dc.b	$FB,$E4,$9F,$BE,$D7,$FB,$C3,$5F,$EF,$ED,$7F,$12,$D7,$E0,$3F,$DE
	.dc.b	$F3,$F7,$F1,$FB,$35,$95,$EC,$8F,$7F,$C6,$48,$F7,$BF,$0B,$E3,$7F
	.dc.b	$B4,$F8,$32,$7F,$A1,$BF,$DB,$3B,$7E,$3D,$C9,$E8,$7C,$8F,$8D,$E8
	.dc.b	$FF,$97,$B8,$58,$E3,$97,$FD,$57,$A7,$D4,$CF,$B5,$FF,$43,$CE,$D7
	.dc.b	$F3,$AC,$FB,$4F,$3C,$EC,$F2,$C9,$9F,$5E,$9F,$F3,$44,$78,$2F,$DA
	.dc.b	$7D,$CC,$BA,$CF,$F0,$3C,$AC,$FF,$81,$B2,$FC,$1F,$C6,$FE,$13,$E3
	.dc.b	$64,$FE,$12,$3B,$3F,$A2,$DA,$ED,$EB,$8F,$ED,$3E,$07,$F8,$4B,$C0
	.dc.b	$91,$83,$91,$0F,$8D,$7F,$92,$F8,$DE,$7B,$7A,$67,$E5,$63,$F5,$5E
	.dc.b	$47,$F6,$9B,$EA,$FF,$6B,$C3,$93,$D7,$83,$E3,$63,$88,$E6,$7C,$42
	.dc.b	$9F,$5F,$7F,$F1,$4D,$74,$AF,$ED,$77,$8A,$FD,$2F,$C6,$FE,$A3,$E0
	.dc.b	$A4,$60,$E9,$5F,$57,$2F,$5E,$3F,$E3,$4F,$0E,$57,$F8,$39,$FF,$1B
	.dc.b	$F8,$3F,$B2,$C6,$4F,$33,$F1,$5F,$1B,$F1,$30,$E7,$1C,$48,$7F,$82
	.dc.b	$EF,$9B,$FB,$A9,$E6,$FE,$4E,$BF,$7F,$1C,$E3,$CF,$C9,$F8,$8C,$7C
	.dc.b	$B1,$CC,$EB,$97,$F8,$33,$2F,$E0,$FF,$19,$FC,$23,$E3,$65,$7F,$08
	.dc.b	$8E,$8F,$CC,$AF,$CB,$5F,$F9,$53,$C4,$FF,$2A,$F8,$92,$BF,$C5,$F1
	.dc.b	$D1,$E4,$43,$F5,$5F,$19,$D2,$FE,$64,$7E,$6C,$E2,$7F,$8E,$2B,$A3
	.dc.b	$FE,$2D,$AF,$E3,$B5,$D6,$3F,$69,$86,$1F,$82,$7E,$6C,$BC,$DF,$9B
	.dc.b	$79,$1D,$48,$60,$FD,$F7,$C6,$7F,$2A,$EE,$91,$FC,$AA,$3D,$1B,$7E
	.dc.b	$D3,$8B,$F3,$52,$D1,$28,$7A,$CF,$D6,$3F,$A7,$4E,$63,$1E,$37,$F5
	.dc.b	$36,$D7,$F9,$CE,$B6,$FF,$56,$C7,$F9,$E9,$1F,$F6,$6F,$E5,$CF,$8F
	.dc.b	$F2,$C7,$C1,$3F,$44,$73,$C7,$B5,$FF,$45,$8F,$44,$78,$E3,$19,$7A
	.dc.b	$A7,$D3,$1E,$2D,$58,$77,$9F,$0E,$1F,$D4,$A3,$11,$D0,$FF,$77,$BF
	.dc.b	$EF,$07,$E9,$F6,$7F,$03,$C3,$F0,$31,$E6,$38,$67,$F1,$F2,$F3,$E9
	.dc.b	$27,$D4,$7C,$0F,$8F,$F0,$7B,$B2,$30,$76,$DF,$F5,$AF,$4D,$BF,$B1
	.dc.b	$37,$DC,$F8,$67,$C4,$F2,$FD,$58,$49,$E6,$7A,$F8,$3E,$0F,$BB,$F3
	.dc.b	$9F,$EE,$78,$FE,$0B,$6F,$91,$B7,$DF,$CB,$AE,$1F,$6D,$EC,$FD,$8C
	.dc.b	$3E,$41,$1F,$B7,$F1,$FD,$78,$3E,$37,$07,$EE,$E7,$91,$0F,$51,$F8
	.dc.b	$7D,$7E,$97,$C0,$7D,$3F,$6E,$4F,$E1,$A3,$A6,$BE,$7B,$FC,$07,$9F
	.dc.b	$F8,$4F,$2E,$5F,$E4,$63,$9B,$C4,$87,$C3,$E5,$7C,$6F,$48,$FD,$D1
	.dc.b	$F3,$E7,$9B,$F4,$6B,$F2,$31,$F8,$79,$EB,$3C,$F1,$C3,$F6,$4F,$C3
	.dc.b	$F7,$70,$7F,$49,$7D,$38,$BE,$7F,$2F,$F9,$F3,$93,$F2,$BF,$1B,$F9
	.dc.b	$32,$5F,$E9,$BE,$33,$CC,$FE,$3E,$F8,$F9,$FF,$A4,$24,$7E,$C7,$83
	.dc.b	$F3,$95,$23,$91,$FB,$0F,$8D,$E7,$7F,$C9,$23,$F7,$17,$C9,$FC,$83
	.dc.b	$FB,$0B,$FD,$35,$73,$9F,$95,$78,$7E,$01,$FC,$BF,$87,$F0,$0F,$23
	.dc.b	$A9,$0F,$CE,$57,$EC,$7E,$36,$47,$F0,$47,$F6,$9F,$1B,$E6,$AE,$79
	.dc.b	$7C,$8D,$FF,$81,$FA,$53,$0F,$E2,$11,$C1,$E4,$43,$F0,$17,$CD,$FF
	.dc.b	$34,$C7,$F0,$C3,$93,$FA,$F9,$C3,$FC,$5B,$DC,$E8,$F2,$30,$7C,$0B
	.dc.b	$FE,$06,$7C,$B7,$F7,$15,$D2,$3E,$FD,$74,$9F,$C9,$7C,$67,$F1,$12
	.dc.b	$3F,$CA,$7D,$0F,$F3,$BF,$BB,$FE,$A6,$8F,$E6,$97,$FC,$75,$FF,$3C
	.dc.b	$FC,$5F,$F9,$F3,$84,$F0,$AF,$E4,$F7,$FC,$A3,$D4,$97,$D4,$87,$88
	.dc.b	$FE,$67,$8F,$07,$BE,$DF,$D0,$DF,$F5,$BD,$F7,$C2,$46,$8F,$E9,$11
	.dc.b	$E7,$7F,$8C,$61,$E0,$7F,$8D,$61,$D5,$C9,$E2,$CB,$F4,$73,$FC,$CF
	.dc.b	$E3,$7D,$53,$D4,$F7,$97,$3F,$7B,$9C,$3C,$17,$CA,$3A,$65,$FD,$17
	.dc.b	$0F,$E9,$3E,$AF,$F3,$FF,$8D,$95,$FD,$49,$FE,$9B,$E9,$78,$23,$FA
	.dc.b	$11,$E0,$2E,$5E,$07,$39,$78,$31,$FE,$A5,$5F,$D5,$BD,$49,$5F,$D8
	.dc.b	$E3,$AB,$DC,$87,$F3,$D7,$C1,$9F,$DD,$C3,$B3,$76,$CE,$1D,$AF,$FB
	.dc.b	$76,$1D,$9E,$47,$72,$18,$3F,$90,$4F,$27,$FB,$F5,$F6,$AF,$BC,$D8
	.dc.b	$FB,$78,$FA,$1E,$58,$7F,$18,$C7,$FB,$9F,$A7,$FD,$A7,$E3,$65,$7F
	.dc.b	$7C,$7F,$BB,$F9,$FD,$E3,$E5,$5F,$AD,$FB,$18,$69,$FF,$04,$7A,$7F
	.dc.b	$DF,$8C,$3F,$D8,$6B,$AB,$FC,$E6,$F3,$7F,$A3,$BF,$A5,$3E,$CD,$3F
	.dc.b	$2A,$FD,$68,$DE,$9F,$E0,$7C,$7B,$4F,$F8,$5F,$1C,$BF,$E0,$86,$F6
	.dc.b	$FF,$82,$3E,$27,$D5,$B7,$C5,$AF,$5E,$1F,$A5,$C7,$C3,$E5,$FF,$89
	.dc.b	$F1,$E6,$DA,$FF,$12,$7A,$5F,$EE,$0F,$CB,$7F,$59,$1F,$DD,$5B,$B7
	.dc.b	$FA,$D1,$B8,$FF,$87,$7D,$2F,$0E,$FE,$CA,$F8,$19,$FF,$89,$27,$FC
	.dc.b	$4B,$E9,$CC,$7F,$C6,$51,$97,$F6,$DB,$B2,$EB,$08,$C3,$B3,$E4,$1F
	.dc.b	$AF,$C7,$C1,$97,$BF,$1D,$65,$FD,$6B,$2F,$05,$FF,$85,$F2,$CD,$FA
	.dc.b	$63,$C6,$78,$B2,$7C,$57,$B2,$3B,$1F,$84,$36,$73,$2E,$1F,$E1,$6F
	.dc.b	$D6,$7F,$C6,$EF,$E0,$7E,$2B,$F9,$39,$27,$91,$EA,$3F,$73,$A7,$A9
	.dc.b	$FC,$E9,$A9,$FD,$87,$27,$F2,$B3,$F3,$1A,$FF,$1B,$9F,$AE,$D7,$E1
	.dc.b	$1F,$68,$E6,$57,$4E,$0E,$2E,$DC,$1C,$8E,$0E,$5F,$E5,$E6,$DE,$6E
	.dc.b	$0F,$64,$FE,$09,$FE,$45,$F1,$1E,$89,$FC,$BF,$65,$9F,$76,$BF,$BE
	.dc.b	$EA,$FC,$97,$C5,$C9,$FA,$8F,$F9,$EF,$5F,$F3,$04,$76,$AF,$36,$5D
	.dc.b	$E3,$F9,$1E,$1B,$E2,$E1,$EA,$C9,$FE,$89,$CA,$C7,$9E,$1F,$66,$FC
	.dc.b	$15,$FA,$1C,$B9,$3E,$63,$E5,$CF,$73,$F9,$26,$3A,$8F,$F3,$7D,$B1
	.dc.b	$F0,$46,$30,$D5,$F7,$8F,$1D,$78,$B2,$FF,$4F,$57,$E8,$FE,$2F,$E3
	.dc.b	$57,$BC,$FE,$4B,$8F,$2A,$FF,$50,$5F,$19,$FF,$52,$B7,$3F,$77,$AF
	.dc.b	$FB,$DE,$29,$5F,$9F,$7E,$67,$AD,$DB,$4F,$66,$3F,$CC,$32,$CD,$F9
	.dc.b	$BD,$6F,$A4,$61,$F7,$DF,$47,$F8,$46,$19,$C3,$9C,$7C,$8C,$3E,$B9
	.dc.b	$D5,$F8,$35,$D1,$EC,$60,$87,$62,$CF,$F1,$3E,$1F,$45,$74,$9F,$8A
	.dc.b	$69,$F8,$B5,$E3,$7E,$3F,$A1,$FC,$7F,$8E,$57,$F0,$17,$F1,$DE,$A7
	.dc.b	$48,$F4,$1D,$0F,$BA,$FE,$4F,$D2,$91,$F6,$A3,$8B,$CC,$87,$03,$F1
	.dc.b	$47,$E9,$F2,$F8,$AB,$B9,$B7,$28,$FC,$5D,$F6,$79,$12,$BF,$88,$5E
	.dc.b	$DF,$E6,$DF,$13,$E9,$9E,$D9,$FC,$F7,$83,$F8,$8A,$E0,$FE,$2B,$D1
	.dc.b	$FE,$85,$F1,$12,$BF,$92,$3F,$B1,$F4,$7A,$70,$7F,$5C,$9F,$46,$99
	.dc.b	$7F,$37,$E9,$7E,$DC,$C3,$FA,$4A,$E8,$FF,$4D,$BC,$E9,$F0,$B0,$E8
	.dc.b	$75,$DB,$E1,$D7,$E9,$AF,$A3,$D8,$C1,$0E,$C5,$AF,$EB,$1F,$13,$E8
	.dc.b	$BF,$F2,$A6,$5E,$0D,$3F,$53,$85,$C3,$F5,$3E,$97,$F6,$3F,$88,$91
	.dc.b	$FD,$15,$FE,$CD,$F1,$3F,$D9,$FE,$22,$4F,$F4,$97,$FB,$57,$C4,$FF
	.dc.b	$17,$F5,$BF,$8C,$7A,$D2,$BF,$7F,$1F,$E6,$AF,$83,$FE,$59,$B1,$A4
	.dc.b	$9F,$E5,$CD,$7F,$9D,$35,$FE,$82,$D7,$FC,$00,$D7,$FB,$57,$90,$F8
	.dc.b	$8F,$EF,$C4,$8F,$EC,$1F,$11,$FE,$01,$F8,$89,$1F,$D8,$DF,$F0,$27
	.dc.b	$C4,$7F,$31,$F6,$7F,$99,$72,$4A,$FE,$33,$1F,$E8,$4F,$86,$F5,$92
	.dc.b	$3F,$A3,$35,$FE,$A4,$D7,$FA,$CB,$5F,$F0,$C3,$5F,$F3,$CB,$FD,$57
	.dc.b	$C8,$FF,$03,$7C,$47,$F8,$5C,$93,$FD,$E7,$5F,$F0,$D5,$48,$FE,$F8
	.dc.b	$FF,$87,$7E,$23,$FA,$57,$9B,$FD,$2F,$CF,$95,$FC,$CA,$3F,$D1,$9D
	.dc.b	$9F,$E9,$BF,$C9,$DF,$11,$FD,$85,$B5,$FD,$9D,$AF,$F6,$F6,$BF,$E2
	.dc.b	$A6,$BF,$E9,$76,$A0,$E1,$B1,$DC,$8F,$76,$BF,$E3,$6D,$1E,$77,$FA
	.dc.b	$C6,$DD,$DF,$D2,$A3,$0D,$3D,$A6,$6E,$DF,$F8,$BE,$5F,$11,$8C,$3A
	.dc.b	$B6,$33,$3F,$E3,$85,$F0,$FF,$DD,$7D,$3F,$EE,$DE,$CC,$AF,$EB,$B1
	.dc.b	$D9,$F7,$DF,$CB,$23,$F2,$DA,$FC,$B6,$BF,$2D,$AF,$CB,$6B,$F2,$F4
	.dc.b	$F2,$BF,$17,$07,$C3,$7A,$49,$1E,$9F,$86,$F6,$78,$D2,$3D,$8F,$BF
	.dc.b	$E1,$7E,$2F,$A8,$FD,$6B,$FC,$AA,$DF,$67,$31,$FE,$48,$BF,$F0,$6B
	.dc.b	$FE,$5A,$7E,$EE,$FF,$89,$23,$F8,$96,$BF,$89,$6B,$F8,$96,$BF,$89
	.dc.b	$6B,$EF,$7F,$49,$9B,$E1,$BE,$B1,$27,$EB,$7C,$37,$D9,$F8,$59,$1F
	.dc.b	$65,$FB,$9F,$0D,$EA,$F7,$7D,$78,$4A,$F5,$C7,$D9,$E1,$F9,$71,$FA
	.dc.b	$1E,$1F,$D2,$11,$C1,$0F,$D2,$39,$69,$93,$F4,$CE,$5F,$F7,$E7,$2F
	.dc.b	$E7,$3E,$1B,$F2,$04,$8F,$C8,$7C,$2F,$E4,$FE,$1A,$47,$E4,$DF,$CB
	.dc.b	$7C,$37,$D4,$F7,$7E,$AF,$97,$2B,$EA,$C7,$AF,$E1,$3F,$78,$48,$FD
	.dc.b	$E3,$5F,$DE,$35,$FD,$E3,$5F,$DE,$35,$FC,$93,$F5,$3E,$17,$F5,$04
	.dc.b	$8F,$D4,$7C,$37,$EA,$E6,$47,$EA,$DF,$D7,$7C,$2F,$E3,$3D,$FF,$C6
	.dc.b	$FB,$D2,$BF,$1B,$1F,$AB,$FC,$9F,$F8,$0C,$7F,$9E,$5A,$FF,$A0,$5A
	.dc.b	$FF,$AF,$1A,$FF,$11,$0E,$1F,$E2,$8D,$73,$C9,$FE,$81,$8C,$11,$C1
	.dc.b	$23,$FD,$21,$C3,$FC,$53,$E1,$FF,$84,$12,$3F,$D7,$FF,$47,$FE,$A6
	.dc.b	$F1,$BF,$CB,$EF,$F9,$66,$8D,$27,$C4,$7E,$BB,$DF,$FD,$7F,$BF,$30
	.dc.b	$FD,$7C,$7C,$3F,$4F,$CC,$48,$F8,$8D,$7E,$23,$5F,$88,$D7,$E2,$68
	.dc.b	$8C,$47,$C4,$7F,$A9,$60,$F8,$7F,$E4,$CD,$9F,$6C,$63,$E1,$B3,$F0
	.dc.b	$F2,$7E,$0B,$C3,$E1,$FF,$84,$FC,$0F,$F0,$AF,$81,$95,$FC,$2A,$3A
	.dc.b	$FC,$37,$6F,$87,$93,$F2,$A3,$C3,$F0,$FE,$2F,$86,$93,$F3,$A3,$F8
	.dc.b	$AF,$C2,$FF,$16,$F8,$59,$3F,$C5,$A3,$AB,$CC,$83,$D7,$FE,$96,$FF
	.dc.b	$BC,$F8,$1F,$0F,$E0,$8F,$B7,$8F,$81,$E1,$D1,$F7,$32,$F0,$C7,$F2
	.dc.b	$4C,$BC,$2E,$87,$A3,$FC,$B7,$3F,$97,$F0,$F2,$7F,$0E,$FF,$2E,$F8
	.dc.b	$4F,$D0,$7C,$07,$F5,$8F,$8F,$F1,$F2,$DD,$8E,$B0,$FE,$67,$7F,$CD
	.dc.b	$BE,$16,$46,$08,$78,$FB,$BF,$23,$97,$86,$3C,$DA,$7F,$41,$BE,$91
	.dc.b	$E9,$CB,$C2,$FF,$7B,$E2,$F8,$F1,$88,$F7,$1F,$AE,$D3,$D4,$FF,$46
	.dc.b	$CB,$C6,$F6,$F5,$7F,$C6,$CF,$92,$FE,$8E,$3E,$18,$C7,$0F,$EA,$3E
	.dc.b	$1F,$C7,$5E,$0D,$FE,$0D,$78,$EB,$FC,$1D,$9F,$8E,$BE,$DC,$F8,$CE
	.dc.b	$67,$D2,$C7,$C2,$F8,$F4,$EB,$7F,$B6,$F8,$8F,$CF,$92,$7F,$77,$F0
	.dc.b	$BA,$58,$F7,$AF,$0E,$5E,$49,$F8,$7D,$ED,$FE,$1C,$77,$3C,$67,$94
	.dc.b	$FA,$9C,$EC,$7C,$11,$8C,$BF,$63,$1E,$1C,$BC,$13,$E2,$C7,$C1,$AF
	.dc.b	$C0,$CB,$C1,$3F,$03,$7F,$98,$F8,$4F,$2E,$7D,$1F,$69,$0F,$1F,$C3
	.dc.b	$77,$C3,$AE,$BE,$EE,$4F,$85,$87,$4C,$BE,$CF,$07,$CC,$6F,$07,$C7
	.dc.b	$E4,$FE,$3F,$F0,$5F,$C2,$89,$A7,$F2,$6F,$83,$ED,$97,$E7,$4E,$B8
	.dc.b	$F3,$CF,$FC,$95,$9F,$76,$C6,$6F,$F9,$DE,$DF,$C5,$B0,$EB,$1F,$33
	.dc.b	$1C,$3F,$87,$36,$5E,$0F,$96,$7C,$C7,$EA,$6B,$EA,$7E,$56,$1D,$DC
	.dc.b	$E3,$FC,$F3,$0E,$B1,$F6,$9F,$E4,$FF,$03,$EC,$8E,$DB,$FC,$A7,$C2
	.dc.b	$7D,$23,$AE,$FF,$1F,$91,$F0,$5F,$CC,$09,$3E,$9E,$17,$E3,$DF,$E3
	.dc.b	$1E,$4F,$BF,$3F,$B7,$C7,$B6,$FE,$B9,$CE,$FF,$85,$E1,$FC,$9B,$ED
	.dc.b	$BF,$11,$FC,$17,$0F,$ED,$0F,$E5,$32,$F0,$B9,$3D,$87,$7E,$3F,$D2
	.dc.b	$3F,$10,$FD,$76,$BD,$B8,$3F,$A7,$E3,$FE,$6F,$C3,$AE,$3F,$CC,$35
	.dc.b	$F5,$E1,$93,$5E,$FF,$69,$FE,$01,$97,$6D,$3D,$79,$7E,$0B,$3E,$D8
	.dc.b	$7C,$CB,$E9,$93,$DA,$F0,$12,$7F,$04,$79,$CF,$46,$DF,$82,$DF,$EB
	.dc.b	$46,$72,$EE,$FB,$0F,$DC,$B8,$7F,$74,$7F,$84,$76,$FF,$05,$11,$FF
	.dc.b	$0C,$35,$FF,$0C,$35,$E0,$68,$86,$8B,$8F,$D9,$7D,$33,$E4,$DB,$F0
	.dc.b	$7E,$67,$C5,$BF,$D8,$66,$F4,$F1,$87,$09,$FF,$3D,$FC,$47,$D3,$F5
	.dc.b	$64,$FE,$36,$D1,$F5,$FD,$0E,$33,$F8,$6D,$3E,$6E,$38,$7C,$91,$FC
	.dc.b	$F7,$0E,$0F,$3B,$EA,$FE,$36,$F8,$70,$FE,$12,$3C,$18,$61,$FB,$51
	.dc.b	$FB,$EB,$E7,$1C,$4F,$35,$7E,$2F,$D1,$E1,$1F,$91,$D7,$F4,$B8,$63
	.dc.b	$6F,$B3,$D7,$FE,$9D,$C7,$F2,$5E,$74,$C3,$E0,$3F,$95,$F3,$F8,$70
	.dc.b	$7E,$5F,$1B,$5F,$EA,$0B,$D3,$C0,$87,$EA,$6B,$6F,$EA,$E6,$EF,$FC
	.dc.b	$9A,$B8,$CE,$CF,$A1,$5F,$A5,$F2,$B7,$BF,$FB,$A7,$5F,$5D,$7D,$92
	.dc.b	$1F,$4E,$3E,$33,$F6,$7C,$BF,$D7,$11,$FB,$9E,$37,$45,$5D,$1E,$47
	.dc.b	$D0,$26,$DF,$1D,$E2,$75,$21,$FC,$1E,$7A,$4F,$3C,$FF,$06,$4C,$3E
	.dc.b	$F5,$7E,$14,$E6,$71,$2C,$7E,$E3,$C7,$E7,$C9,$F9,$99,$F7,$CF,$37
	.dc.b	$CB,$E0,$36,$FE,$19,$D5,$27,$F3,$4F,$F0,$FF,$1B,$9F,$0F,$C1,$DB
	.dc.b	$D9,$5F,$C7,$E7,$83,$FC,$86,$B8,$3E,$0D,$3D,$39,$78,$75,$F7,$46
	.dc.b	$09,$3E,$4C,$B8,$BD,$4F,$D6,$61,$FC,$86,$3C,$71,$F8,$B3,$B1,$C4
	.dc.b	$FD,$97,$9F,$FC,$9F,$8E,$46,$0F,$DB,$7C,$47,$F2,$B2,$3F,$B9,$F8
	.dc.b	$8F,$F7,$BD,$48,$FF,$7B,$BF,$EF,$AA,$FE,$64,$C8,$E8,$7F,$41,$F8
	.dc.b	$8E,$6F,$F4,$3B,$BA,$7F,$2C,$C3,$A4,$FF,$8E,$EE,$CF,$F8,$0B,$0F
	.dc.b	$E7,$D8,$75,$AF,$E9,$95,$87,$D2,$43,$B1,$E9,$3C,$E4,$3B,$1F,$DE
	.dc.b	$76,$70,$7F,$21,$FC,$6B,$FC,$7F,$8F,$F4,$B5,$F1,$3F,$EC,$18,$C3
	.dc.b	$77,$FD,$64,$F0,$7F,$5D,$C6,$DB,$DF,$18,$7F,$AC,$6D,$EE,$8F,$F0
	.dc.b	$8E,$97,$4F,$F2,$66,$DD,$5F,$D6,$E3,$D5,$B5,$9E,$14,$E5,$BF,$13
	.dc.b	$E0,$AF,$DA,$E7,$E3,$7E,$36,$BF,$45,$F8,$8F,$F8,$C3,$2F,$16,$FF
	.dc.b	$13,$3E,$D9,$31,$B9,$CC,$7F,$C1,$53,$97,$AC,$7B,$DA,$78,$B4,$EB
	.dc.b	$BF,$FB,$1E,$3C,$33,$FE,$32,$C3,$BF,$C3,$78,$3E,$06,$57,$68,$F6
	.dc.b	$6F,$FE,$18,$E0,$BC,$0C,$FA,$EF,$DB,$2E,$78,$F1,$21,$FE,$99,$D5
	.dc.b	$F1,$3C,$B1,$E7,$87,$2A,$F2,$9F,$50,$F7,$CF,$47,$96,$7F,$1A,$39
	.dc.b	$47,$C0,$CF,$FC,$5B,$5C,$E3,$F2,$27,$FA,$67,$1F,$9B,$F1,$3F,$3B
	.dc.b	$E2,$64,$FC,$C8,$E6,$FE,$70,$FF,$16,$57,$3E,$3F,$B1,$1F,$9B,$E1
	.dc.b	$FC,$D5,$74,$8E,$47,$D6,$F8,$9E,$75,$FE,$1C,$D7,$F2,$73,$EE,$E4
	.dc.b	$FD,$1F,$07,$C9,$8E,$07,$23,$E4,$61,$F8,$4C,$39,$F1,$FD,$F9,$F2
	.dc.b	$9C,$F2,$FB,$CF,$C0,$CB,$F7,$8F,$E2,$3E,$27,$F1,$BF,$13,$23,$07
	.dc.b	$C9,$E4,$FC,$C3,$F0,$6B,$F3,$0F,$C9,$3B,$67,$FC,$4D,$E2,$7D,$DB
	.dc.b	$FC,$C9,$F9,$AF,$89,$F3,$C7,$E9,$36,$F8,$2F,$33,$CD,$9F,$ED,$1F
	.dc.b	$99,$C3,$C5,$FC,$FD,$57,$F1,$5F,$11,$D3,$8F,$F2,$91,$FC,$44,$E9
	.dc.b	$9F,$D6,$7A,$9F,$8B,$3F,$25,$F1,$1F,$AE,$F8,$89,$1F,$AC,$8E,$9B
	.dc.b	$7C,$2A,$F8,$27,$CA,$9E,$7C,$7F,$61,$F3,$9C,$48,$76,$3F,$6F,$F1
	.dc.b	$1D,$36,$F8,$35,$CC,$F3,$13,$6F,$DA,$57,$02,$1F,$A9,$AF,$E0,$BF
	.dc.b	$11,$CE,$3E,$05,$73,$AF,$95,$1E,$6E,$3F,$E5,$DF,$11,$FC,$33,$E2
	.dc.b	$24,$FF,$0A,$8E,$4F,$E2,$8F,$93,$5C,$AF,$C9,$1F,$25,$FE,$6D,$3F
	.dc.b	$C5,$89,$1F,$C5,$3E,$23,$9C,$FC,$8A,$F5,$CF,$8F,$97,$F1,$0F,$F0
	.dc.b	$3A,$FE,$3E,$47,$F8,$EF,$C4,$72,$AF,$E7,$BD,$5C,$30,$F1,$EF,$F4
	.dc.b	$6F,$CE,$7F,$89,$3E,$1D,$F1,$FF,$CD,$DF,$F1,$3C,$7F,$AC,$5A,$AB
	.dc.b	$FD,$35,$1C,$23,$F4,$5C,$3F,$C5,$1F,$BA,$7A,$4B,$3C,$2B,$B1,$23
	.dc.b	$B7,$C4,$FA,$5F,$E6,$47,$0C,$BE,$61,$1F,$98,$FF,$91,$F0,$4F,$A4
	.dc.b	$93,$E9,$8E,$9C,$DF,$C7,$A3,$A0,$CB,$E2,$63,$F1,$B8,$7F,$D1,$D6
	.dc.b	$F8,$9F,$89,$EC,$CC,$3E,$24,$74,$E6,$FE,$79,$1D,$31,$E7,$A7,$9D
	.dc.b	$FE,$AD,$1C,$08,$7C,$B8,$F5,$57,$D1,$F8,$9E,$9C,$1F,$66,$79,$67
	.dc.b	$FC,$5B,$4F,$CC,$61,$F6,$6F,$83,$F5,$DF,$B5,$F1,$1C,$A3,$C9,$5C
	.dc.b	$6F,$9F,$6F,$F7,$99,$7A,$9F,$C6,$9F,$F7,$1B,$51,$1C,$37,$8F,$FD
	.dc.b	$D2,$E8,$E6,$E1,$D1,$63,$8B,$A3,$11,$B7,$07,$48,$D9,$E0,$E1,$70
	.dc.b	$BE,$27,$F3,$04,$AF,$CC,$7C,$47,$E6,$FE,$26,$47,$E6,$DF,$CF,$7B
	.dc.b	$9F,$E1,$9E,$3F,$F1,$6E,$77,$6F,$B7,$7F,$54,$FC,$33,$8E,$16,$FF
	.dc.b	$12,$F8,$D7,$FB,$5B,$8B,$D5,$BF,$28,$F9,$71,$82,$3F,$0D,$AF,$BD
	.dc.b	$AF,$03,$44,$3E,$71,$FB,$9F,$89,$FD,$C1,$1F,$DC,$7C,$4F,$EE,$FE
	.dc.b	$26,$47,$EE,$DF,$DF,$7C,$4F,$BF,$D2,$F8,$5E,$2F,$F8,$E7,$33,$76
	.dc.b	$4F,$E1,$F4,$B8,$FD,$B8,$F7,$91,$C1,$0F,$B4,$E5,$A6,$4F,$C7,$B9
	.dc.b	$7F,$15,$CB,$F7,$BE,$27,$F8,$A1,$23,$F8,$A7,$C4,$FF,$17,$F8,$99
	.dc.b	$1F,$C5,$DF,$E3,$5F,$13,$EC,$F5,$BD,$BE,$B4,$AF,$F3,$66,$BF,$E6
	.dc.b	$87,$F3,$BC,$BF,$71,$FC,$FF,$83,$5F,$CE,$B5,$FA,$2D,$7F,$40,$D7
	.dc.b	$8F,$37,$E6,$E3,$10,$69,$23,$44,$8F,$D4,$47,$3D,$7F,$3F,$F1,$1F
	.dc.b	$CD,$49,$3E,$8A,$FF,$74,$7E,$27,$F8,$4C,$3D,$56,$87,$C4,$FF,$9C
	.dc.b	$DB,$7C,$4F,$F7,$2B,$87,$F4,$13,$C7,$E2,$7F,$0B,$EC,$7E,$1B,$DA
	.dc.b	$95,$F2,$A3,$F8,$56,$9F,$88,$8F,$4E,$BF,$04,$8E,$08,$7C,$C7,$2D
	.dc.b	$32,$7C,$E7,$2F,$E1,$31,$7C,$47,$AC,$91,$EB,$CB,$E0,$7B,$FE,$FF
	.dc.b	$88,$93,$9B,$47,$C3,$F8,$8F,$D0,$FB,$BF,$A2,$F7,$25,$FE,$1A,$3E
	.dc.b	$46,$1D,$A3,$E2,$E1,$E2,$7A,$47,$67,$38,$5E,$1F,$A7,$4C,$7B,$D7
	.dc.b	$E4,$B1,$EE,$FF,$0F,$ED,$FD,$8E,$5E,$08,$F4,$69,$E1,$C7,$9E,$7D
	.dc.b	$78,$FF,$03,$A5,$BF,$B7,$9F,$67,$A6,$CD,$3B,$E3,$F9,$FD,$3B,$D7
	.dc.b	$A8,$87,$BB,$2F,$C4,$7C,$47,$5D,$FF,$19,$9D,$DF,$D9,$7F,$1C,$F6
	.dc.b	$10,$F0,$EF,$C9,$FA,$57,$5F,$97,$F1,$1C,$EF,$A6,$DE,$9D,$3C,$DB
	.dc.b	$7A,$6F,$A4,$63,$1F,$CE,$61,$CE,$3E,$7E,$38,$AF,$C0,$63,$D5,$FC
	.dc.b	$16,$3D,$77,$FE,$75,$97,$59,$F8,$FB,$E5,$FE,$C3,$F9,$3B,$AB,$FD
	.dc.b	$C1,$F2,$EF,$DE,$7F,$DC,$0D,$C7,$FB,$65,$6E,$FF,$5C,$17,$57,$F7
	.dc.b	$1F,$41,$EF,$53,$8C,$4A,$E3,$2F,$7B,$81,$98,$FE,$AC,$75,$CF,$C2
	.dc.b	$47,$C3,$87,$81,$B7,$DD,$EB,$87,$3E,$1F,$E4,$15,$B9,$EB,$8F,$85
	.dc.b	$F1,$E3,$BC,$BF,$C0,$15,$B9,$FE,$AA,$D7,$38,$FF,$7D,$8C,$47,$CC
	.dc.b	$8C,$11,$FF,$04,$46,$9F,$91,$19,$7F,$C1,$F1,$92,$47,$8E,$3A,$1F
	.dc.b	$43,$F2,$8F,$E4,$95,$E2,$8F,$82,$43,$DB,$E0,$71,$7B,$99,$CB,$FA
	.dc.b	$16,$4F,$81,$FE,$CF,$F0,$7F,$CA,$FE,$32,$5F,$F4,$C8,$FF,$11,$5F
	.dc.b	$F4,$BF,$09,$FB,$6B,$C6,$CE,$9F,$E2,$C9,$CF,$8D,$8C,$7F,$9B,$DA
	.dc.b	$31,$3C,$AF,$C3,$5F,$D6,$BE,$17,$FB,$C9,$87,$F7,$F9,$5C,$BF,$52
	.dc.b	$FF,$41,$F8,$DE,$79,$78,$EB,$FA,$16,$3E,$26,$C7,$85,$B8,$F3,$CF
	.dc.b	$F4,$F8,$59,$FF,$35,$74,$BE,$1B,$FA,$77,$C6,$C9,$FE,$60,$FF,$51
	.dc.b	$F8,$DF,$EA,$5F,$1B,$27,$F9,$A3,$FD,$57,$E3,$7F,$C2,$7C,$7F,$E1
	.dc.b	$4E,$A9,$5F,$DE,$23,$FA,$D7,$C6,$FF,$5B,$F8,$D9,$3F,$CF,$63,$93
	.dc.b	$D0,$87,$F3,$8B,$FF,$0F,$F6,$48,$FF,$01,$BF,$E2,$2F,$8C,$FE,$CD
	.dc.b	$F1,$12,$7F,$A4,$3F,$DA,$7E,$07,$AC,$73,$3F,$A6,$5F,$F6,$DF,$73
	.dc.b	$FC,$57,$C3,$FE,$2C,$F8,$C9,$5F,$E1,$08,$FF,$17,$FC,$67,$F7,$3F
	.dc.b	$7A,$4F,$F8,$52,$3F,$CE,$9F,$87,$7F,$56,$8F,$EE,$DF,$1F,$FE,$32
	.dc.b	$D7,$F0,$CD,$8F,$E1,$B3,$B7,$F9,$FD,$7F,$8F,$3E,$36,$47,$53,$FC
	.dc.b	$43,$F1,$DF,$E0,$3F,$72,$46,$0E,$7E,$77,$88,$8F,$F1,$AF,$C9,$DF
	.dc.b	$1F,$FE,$48,$BF,$F2,$4B,$27,$91,$FD,$42,$34,$E3,$C3,$FF,$0A,$FE
	.dc.b	$2D,$FE,$2D,$7F,$C5,$DA,$7A,$5B,$8F,$82,$BF,$8E,$E3,$E0,$79,$78
	.dc.b	$1F,$CE,$31,$EB,$1F,$D7,$BE,$43,$F5,$5E,$BB,$3F,$F1,$F4,$FF,$87
	.dc.b	$17,$BF,$A6,$C7,$F8,$25,$DE,$7E,$28,$F8,$64,$3C,$7A,$F8,$23,$FA
	.dc.b	$AE,$37,$2F,$E9,$79,$75,$F7,$7B,$7B,$D3,$1E,$D1,$E1,$F6,$7C,$5E
	.dc.b	$B4,$9F,$14,$7F,$A5,$7F,$5D,$FF,$77,$47,$F8,$15,$FF,$79,$E7,$FD
	.dc.b	$E5,$B9,$7F,$A4,$DD,$C7,$CE,$C3,$C5,$5E,$03,$DD,$E4,$78,$5F,$81
	.dc.b	$8F,$85,$F8,$D8,$F8,$7E,$47,$C5,$C3,$27,$FC,$91,$EF,$33,$EF,$A7
	.dc.b	$FB,$1A,$BE,$87,$9F,$F3,$49,$7F,$37,$CB,$FF,$68,$7E,$25,$FE,$B9
	.dc.b	$DF,$FD,$9B,$F6,$9F,$E0,$AA,$7E,$50,$F5,$7F,$D9,$93,$98,$C4,$F0
	.dc.b	$F7,$7F,$DA,$18,$73,$85,$7F,$85,$36,$F3,$7C,$A7,$CB,$24,$FA,$B7
	.dc.b	$FF,$65,$BE,$53,$F0,$5A,$FA,$1F,$C4,$6B,$E8,$F9,$06,$DE,$19,$FF
	.dc.b	$72,$EB,$E4,$7C,$3A,$FF,$B8,$E3,$1A,$7B,$75,$F3,$3E,$5C,$DA,$79
	.dc.b	$30,$FB,$CF,$E0,$7E,$47,$F2,$DE,$7F,$E5,$FD,$09,$5F,$97,$8F,$26
	.dc.b	$9E,$38,$F9,$3C,$3F,$23,$5F,$13,$63,$FD,$F7,$A7,$8A,$7F,$11,$F2
	.dc.b	$7F,$8F,$24,$FC,$5F,$87,$F0,$BD,$C8,$7E,$AF,$D0,$F0,$BF,$00,$F8
	.dc.b	$78,$77,$7C,$67,$97,$0F,$02,$E4,$FC,$56,$1E,$08,$FC,$F3,$FA,$1C
	.dc.b	$3F,$6D,$E4,$FE,$DF,$CB,$95,$FB,$78,$F0,$F1,$7E,$17,$2F,$07,$1F
	.dc.b	$D0,$D7,$F0,$BC,$9F,$63,$0F,$36,$9F,$99,$BF,$D7,$12,$3F,$09,$F2
	.dc.b	$7F,$C1,$7C,$DF,$0C,$77,$3F,$83,$3E,$0C,$3E,$6C,$7D,$1C,$7C,$0F
	.dc.b	$CE,$C7,$C1,$97,$DF,$C7,$B4,$FE,$ED,$FD,$F6,$3F,$C4,$FC,$BF,$E2
	.dc.b	$9E,$5C,$AF,$E2,$91,$DD,$F9,$A7,$7A,$F9,$DC,$3F,$33,$5F,$DA,$65
	.dc.b	$FC,$27,$E4,$E5,$FE,$7E,$3B,$3D,$08,$7F,$25,$F1,$7B,$3F,$60,$FB
	.dc.b	$85,$D3,$F3,$53,$EC,$DB,$F8,$83,$DA,$7F,$8B,$5F,$F1,$98,$FE,$5D
	.dc.b	$E3,$49,$FE,$5C,$FF,$31,$F1,$BB,$F0,$7C,$89,$ED,$87,$D8,$E7,$FD
	.dc.b	$76,$9F,$C3,$FE,$4F,$F9,$27,$C9,$CB,$FD,$D4,$7F,$40,$FC,$47,$F8
	.dc.b	$9C,$3C,$24,$61,$BE,$26,$DF,$0F,$1F,$E2,$0F,$97,$BF,$B6,$FF,$2D
	.dc.b	$E0,$B1,$FE,$68,$48,$FE,$15,$F2,$9E,$1F,$0A,$47,$FB,$4F,$3F,$DE
	.dc.b	$B7,$C2,$EE,$FD,$93,$F4,$98,$F7,$DB,$F9,$66,$1D,$67,$FA,$06,$3F
	.dc.b	$D0,$E3,$FA,$27,$CA,$7B,$FC,$09,$5F,$C6,$E3,$AB,$F6,$38,$FE,$67
	.dc.b	$4F,$E1,$76,$57,$F4,$DF,$94,$97,$C4,$F4,$78,$3F,$2F,$86,$46,$0F
	.dc.b	$F4,$DE,$1F,$BA,$7E,$B9,$F9,$43,$F0,$EF,$7B,$F3,$C7,$53,$FA,$D7
	.dc.b	$CA,$7D,$7B,$91,$D4,$FE,$5B,$E4,$7F,$4B,$E2,$78,$1E,$CE,$DF,$DF
	.dc.b	$57,$C8,$E9,$FC,$2E,$CC,$BB,$CF,$F6,$7C,$FF,$B5,$47,$E1,$79,$24
	.dc.b	$FD,$57,$F1,$1D,$BD,$7A,$FF,$77,$3E,$A9,$FA,$98,$3E,$4B,$FB,$A1
	.dc.b	$27,$FA,$57,$C0,$72,$7A,$10,$F5,$9F,$93,$D9,$8F,$78,$F9,$1A,$FE
	.dc.b	$B3,$5F,$5E,$1D,$67,$FC,$00,$FF,$80,$F0,$FC,$F7,$3F,$E7,$FB,$65
	.dc.b	$7E,$3E,$3A,$3F,$50,$71,$FE,$07,$E1,$3F,$C1,$9F,$15,$27,$FB,$14
	.dc.b	$74,$79,$10,$F4,$F7,$7E,$AB,$AF,$AF,$57,$E3,$38,$BE,$2D,$F3,$BF
	.dc.b	$F0,$B6,$1F,$E1,$98,$FD,$A7,$2C,$9F,$D2,$3F,$B7,$EA,$E7,$D5,$FB
	.dc.b	$0E,$1F,$B7,$F1,$3F,$EA,$3E,$D7,$EB,$DF,$EB,$AC,$7F,$A4,$36,$3F
	.dc.b	$C2,$53,$CE,$7F,$61,$85,$7D,$3C,$FF,$85,$7F,$33,$AF,$EF,$5F,$8D
	.dc.b	$B7,$AF,$2F,$0C,$7F,$87,$B2,$EC,$77,$E8,$FF,$72,$4B,$B3,$F2,$DC
	.dc.b	$FF,$C4,$79,$E6,$1F,$C0,$63,$DA,$FE,$5C,$FE,$A3,$83,$E1,$BD,$CB
	.dc.b	$E4,$E9,$88,$FE,$39,$52,$BA,$1F,$C3,$78,$FA,$BF,$97,$CB,$F9,$3E
	.dc.b	$1D,$63,$D5,$87,$47,$D9,$87,$F2,$7E,$4F,$E5,$1C,$32,$7F,$8C,$47
	.dc.b	$4E,$AF,$CA,$47,$53,$FA,$1F,$67,$9D,$FF,$08,$35,$D4,$23,$F7,$53
	.dc.b	$DE,$7A,$83,$F1,$BE,$A7,$DD,$7F,$2F,$C5,$F5,$78,$BF,$A2,$1F,$56
	.dc.b	$3A,$DF,$17,$B6,$7F,$E0,$F8,$F1,$63,$E8,$3C,$53,$FD,$0F,$0F,$88
	.dc.b	$F8,$2B,$FD,$0F,$1E,$9D,$7E,$E4,$EF,$4F,$85,$87,$F5,$38,$F0,$9E
	.dc.b	$97,$B9,$E3,$2B,$FE,$01,$9F,$F2,$37,$4A,$FC,$31,$DA,$31,$7E,$23
	.dc.b	$C7,$F2,$7E,$0B,$E4,$78,$0F,$79,$F4,$31,$FD,$43,$FE,$31,$97,$C1
	.dc.b	$3E,$13,$E3,$63,$EC,$9C,$CE,$39,$7E,$94,$62,$FE,$A5,$6E,$B9,$47
	.dc.b	$EA,$23,$FA,$C3,$FD,$37,$AB,$C3,$B7,$CF,$DB,$E1,$67,$F4,$CC,$BD
	.dc.b	$B0,$E7,$7F,$C9,$31,$F3,$E7,$E3,$8F,$D2,$4E,$B2,$F8,$DB,$3E,$4F
	.dc.b	$E2,$57,$E1,$A7,$F5,$99,$FF,$A2,$72,$FA,$33,$F1,$8F,$AF,$BF,$D1
	.dc.b	$F7,$7E,$FC,$C9,$F7,$EF,$FE,$99,$CB,$FD,$A4,$FD,$FC,$3F,$A7,$CF
	.dc.b	$BC,$EF,$59,$3D,$EF,$D3,$32,$7E,$3E,$FF,$1A,$47,$B9,$F9,$3F,$91
	.dc.b	$F5,$F8,$1F,$DA,$CF,$ED,$95,$EA,$3D,$6F,$F9,$A3,$2F,$C5,$7C,$97
	.dc.b	$E6,$3E,$36,$61,$D2,$72,$FD,$73,$D0,$4C,$BE,$B3,$F8,$C7,$E4,$71
	.dc.b	$7D,$78,$F4,$CF,$27,$F6,$B5,$D9,$FD,$BD,$FA,$39,$BF,$95,$63,$E8
	.dc.b	$C3,$95,$7D,$0D,$9F,$1B,$FA,$EF,$8C,$93,$E7,$9C,$BE,$7D,$3E,$79
	.dc.b	$35,$F9,$D1,$E7,$7E,$7B,$FC,$4B,$0F,$C6,$BF,$BD,$C3,$F7,$64,$7B
	.dc.b	$7C,$67,$9F,$8B,$EA,$C7,$E1,$70,$F8,$2F,$C0,$C5,$F1,$9F,$C1,$3E
	.dc.b	$2A,$4F,$9E,$39,$3F,$B3,$CF,$EB,$67,$E2,$E1,$FD,$F5,$7F,$0C,$25
	.dc.b	$7B,$FE,$33,$8E,$5F,$4A,$3C,$58,$7B,$23,$EA,$DF,$F0,$7E,$8F,$E2
	.dc.b	$A4,$AF,$91,$D1,$CA,$FE,$2C,$72,$3F,$8A,$65,$E3,$D3,$F8,$1C,$70
	.dc.b	$21,$FC,$C3,$E3,$39,$3F,$34,$E2,$7E,$6C,$FB,$5F,$19,$FC,$98,$8F
	.dc.b	$D9,$F8,$CE,$2F,$EC,$CF,$A8,$71,$D3,$D7,$69,$FE,$5E,$49,$FC,$17
	.dc.b	$C5,$F2,$E5,$FC,$54,$71,$CF,$F0,$3B,$3F,$01,$FE,$6B,$3F,$E1,$45
	.dc.b	$F1,$FF,$C2,$67,$4D,$8D,$F2,$70,$C3,$FC,$B1,$E1,$7F,$4C,$F0,$7F
	.dc.b	$13,$8F,$A7,$6F,$C1,$BE,$A3,$F8,$4E,$1C,$BB,$FF,$AC,$67,$E3,$7B
	.dc.b	$1D,$FE,$2F,$D5,$D9,$FE,$0D,$DB,$F0,$B8,$7E,$B3,$0F,$E8,$4F,$ED
	.dc.b	$D7,$C5,$79,$CF,$47,$17,$2D,$FE,$15,$7D,$43,$C7,$1A,$79,$3E,$FF
	.dc.b	$88,$E1,$1E,$9D,$7D,$67,$E3,$30,$7C,$4F,$C5,$F8,$A9,$5E,$38,$E3
	.dc.b	$5C,$33,$FC,$A1,$F4,$AD,$F1,$5F,$36,$A5,$FA,$E3,$8D,$70,$8F,$45
	.dc.b	$F8,$7E,$2B,$EA,$7C,$4C,$9F,$85,$1C,$1D,$D7,$17,$EA,$3F,$43,$14
	.dc.b	$FD,$D2,$4F,$CB,$F8,$AE,$4F,$E8,$C9,$CD,$FC,$A6,$9F,$15,$F8,$52
	.dc.b	$4F,$D3,$F8,$9E,$38,$FD,$88,$FC,$0C,$69,$FC,$84,$FE,$36,$37,$DD
	.dc.b	$FC,$83,$4F,$9D,$F1,$5F,$91,$25,$FD,$FF,$8A,$DB,$F7,$0F,$69,$FE
	.dc.b	$D7,$C9,$1C,$23,$F3,$D3,$F9,$B8,$E3,$1F,$77,$2F,$A9,$F1,$3F,$9F
	.dc.b	$24,$7E,$37,$E2,$B7,$8F,$DB,$8F,$DF,$75,$FE,$AE,$B7,$1F,$AB,$7F
	.dc.b	$51,$5C,$2F,$E9,$6D,$F2,$7E,$2B,$F5,$9F,$15,$30,$FC,$D4,$6D,$FB
	.dc.b	$27,$6D,$51,$FB,$32,$47,$E8,$7E,$2B,$83,$F5,$BB,$3F,$DE,$1A,$3E
	.dc.b	$EF,$FD,$F0,$7F,$BE,$E7,$FA,$F3,$5F,$F4,$AB,$FF,$00,$1F,$96,$7F
	.dc.b	$9F,$3C,$0F,$DE,$E7,$FC,$32,$3E,$41,$F0,$A7,$AD,$7F,$09,$7A,$76
	.dc.b	$FF,$8D,$5E,$95,$FC,$22,$FF,$CA,$1C,$1F,$D5,$70,$FE,$50,$78,$36
	.dc.b	$FC,$44,$7F,$3A,$BF,$4F,$C5,$F5,$3B,$7C,$5F,$C0,$AF,$EF,$BA,$7C
	.dc.b	$0A,$7C,$5F,$8F,$E2,$E6,$5F,$02,$38,$73,$FF,$4F,$AD,$CE,$48,$7C
	.dc.b	$4F,$8B,$DB,$FD,$83,$66,$2D,$FF,$CD,$53,$FE,$5D,$2C,$EA,$33,$1F
	.dc.b	$E5,$23,$33,$C5,$CC,$ED,$B1,$97,$73,$C9,$FE,$8C,$60,$DB,$99,$E5
	.dc.b	$1E,$2F,$89,$7E,$11,$96,$C7,$81,$AE,$E3,$44,$74,$D7,$46,$08,$60
	.dc.b	$FF,$4A,$7E,$16,$20

;--------------------------------------------------------------------------------
;	FSSQRT.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fssqrt_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FSSQRT.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSSQRT.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fssqrt_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fssqrt.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSSQRT.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1723+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fssqrt.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSSQRT.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fssqrt_data_compressed::
	.dc.b	$00,$02,$A1,$14,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$FD
	.dc.b	$4F,$E5,$1F,$97,$BE,$4B,$CF,$E9,$7A,$3D,$2D,$4B,$E1,$2F,$C9,$E3
	.dc.b	$E4,$73,$F1,$7A,$F8,$8D,$FB,$9C,$3D,$3E,$3E,$9F,$2F,$4F,$9F,$A7
	.dc.b	$D3,$D3,$EB,$E9,$F6,$F4,$FB,$FA,$7E,$0F,$4F,$C3,$E9,$F8,$BD,$3F
	.dc.b	$97,$EC,$7C,$CF,$63,$E6,$FB,$1F,$D8,$FF,$14,$FE,$40,$7E,$63,$C5
	.dc.b	$3F,$98,$7F,$B9,$3F,$D9,$9B,$BF,$1D,$F9,$6F,$CF,$7E,$9B,$F5,$DF
	.dc.b	$B6,$FD,$F7,$F0,$6F,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$FE,$ED,$F8,$AB
	.dc.b	$E6,$2F,$C7,$FE,$13,$DC,$FE,$F5,$F3,$9F,$47,$8F,$FC,$0D,$7F,$E7
	.dc.b	$B8,$FA,$11,$FE,$09,$7F,$DB,$87,$FA,$50,$FF,$03,$1F,$CF,$CF,$80
	.dc.b	$7B,$D3,$EF,$6F,$1F,$83,$8F,$C3,$C7,$E2,$E3,$F1,$F1,$F9,$38,$FC
	.dc.b	$BC,$7E,$6E,$3F,$3F,$1F,$A3,$8F,$D3,$C7,$EA,$E3,$FF,$0A,$77,$7B
	.dc.b	$F7,$FF,$10,$3F,$ED,$CE,$1F,$F2,$6B,$FE,$97,$2F,$C6,$F0,$F8,$AE
	.dc.b	$3F,$05,$CB,$DE,$E7,$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83
	.dc.b	$DE,$F0,$FB,$DE,$2F,$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$F5,$38,$3C
	.dc.b	$FE,$F7,$A3,$DE,$FF,$16,$7C,$EF,$F8,$14,$7C,$5E,$3E,$2F,$3F,$17
	.dc.b	$AF,$8B,$DF,$C5,$F0,$F8,$BE,$3F,$11,$CB,$E1,$79,$FC,$2F,$4F,$85
	.dc.b	$EB,$F0,$BD,$BE,$17,$BF,$C2,$F8,$3E,$17,$C3,$F0,$BE,$2F,$83,$F1
	.dc.b	$FC,$1F,$93,$DB,$F2,$FB,$7E,$6F,$6F,$CF,$EC,$7F,$90,$FE,$7B,$F1
	.dc.b	$5E,$61,$FC,$4B,$7C,$BD,$7C,$16,$FE,$0B,$87,$C1,$71,$F8,$2E,$5F
	.dc.b	$05,$CF,$E0,$BF,$33,$E8,$F5,$F7,$BB,$7B,$DD,$FD,$EF,$07,$BD,$E1
	.dc.b	$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD,$EF,$79,$FD,$EF
	.dc.b	$47,$BD,$FE,$5B,$F9,$FF,$C4,$73,$BD,$FF,$E7,$2A,$1F,$03,$BF,$8A
	.dc.b	$E1,$F2,$BC,$7D,$5E,$5C,$BC,$FE,$0B,$A7,$C1,$75,$F8,$2E,$DF,$05
	.dc.b	$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE,$0B,$C9,$F0,$5E
	.dc.b	$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$7B,$AF,$EF,$B5,$FD,$FE,$BF,$C0
	.dc.b	$35,$FE,$03,$AF,$B7,$E0,$BD,$DF,$05,$FE,$79,$F7,$FF,$CA,$47,$E3
	.dc.b	$BA,$3F,$D0,$CB,$E0,$34,$EF,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1C
	.dc.b	$FE,$03,$A7,$C0,$7F,$10,$E8,$ED,$F0,$1D,$FE,$03,$C1,$F0,$1F,$C5
	.dc.b	$3A,$FF,$8A,$F5,$FF,$16,$EB,$FE,$2F,$D7,$FC,$63,$AF,$F8,$CF,$5F
	.dc.b	$F1,$AE,$6F,$E3,$7C,$DF,$E9,$1F,$C5,$BF,$CD,$47,$A7,$9C,$FA,$5F
	.dc.b	$F5,$03,$FE,$FC,$3F,$D6,$C7,$F8,$98,$FE,$BC,$7F,$5D,$78,$F0,$F5
	.dc.b	$E1,$ED,$C1,$DF,$83,$E4,$7C,$0F,$C9,$F8,$1F,$95,$F0,$3F,$2F,$E0
	.dc.b	$7E,$67,$C0,$FC,$DF,$81,$F9,$DF,$03,$F3,$FE,$07,$E8,$7C,$0F,$FA
	.dc.b	$A7,$E8,$7F,$4E,$7E,$98,$7C,$0E,$7E,$07,$5F,$03,$BF,$81,$E1,$F0
	.dc.b	$3C,$7E,$07,$97,$C0,$F3,$F8,$1E,$9F,$03,$D7,$E0,$7B,$7C,$0F,$7F
	.dc.b	$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03,$F2,$FC,$9F,$99,$D7,$F3,$7A
	.dc.b	$FE,$77,$5F,$CF,$EB,$FA,$1D,$7F,$47,$AF,$E9,$75,$FD,$3E,$BF,$A9
	.dc.b	$D7,$FE,$C5,$F4,$FF,$81,$9F,$C0,$8F,$6F,$83,$FE,$D3,$5D,$7B,$EB
	.dc.b	$E1,$D7,$C7,$AF,$97,$5F,$3E,$BE,$9D,$7F,$27,$E0,$7E,$57,$C0,$FC
	.dc.b	$BF,$81,$F0,$77,$7C,$DF,$81,$FF,$6B,$7D,$17,$EF,$0F,$DC,$F0,$7F
	.dc.b	$B8,$D7,$BF,$AF,$7F,$7E,$FF,$0F,$7F,$8F,$BF,$CB,$DF,$E7,$EF,$F4
	.dc.b	$E7,$EB,$EF,$FB,$FC,$8F,$81,$E4,$7C,$1F,$23,$E1,$79,$1F,$0F,$C8
	.dc.b	$F1,$FB,$FE,$4F,$67,$CB,$EC,$F9,$BD,$9F,$3F,$B3,$E8,$F6,$7D,$3E
	.dc.b	$CF,$AB,$D9,$F5,$FB,$3E,$CF,$67,$FD,$E5,$F4,$7F,$EC,$B3,$F4,$47
	.dc.b	$E8,$7B,$0B,$DF,$DF,$BF,$C3,$DF,$E3,$EF,$F2,$F7,$FE,$C7,$B1,$D3
	.dc.b	$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F
	.dc.b	$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$F0,$DE,$A7
	.dc.b	$AB,$DC,$74,$4F,$3F,$FA,$59,$FD,$28,$FC,$4A,$7F,$12,$EF,$DB,$E2
	.dc.b	$E7,$DB,$B3,$FE,$33,$3F,$C2,$87,$F8,$4C,$7B,$1A,$F5,$B7,$E6,$70
	.dc.b	$F3,$38,$F9,$9C,$BC,$CE,$7E,$67,$4E,$CE,$BB,$F6,$F6,$3B,$FB,$1E
	.dc.b	$0F,$63,$C7,$F4,$9F,$D3,$0F,$C8,$F0,$9F,$C8,$BF,$01,$F6,$37,$D7
	.dc.b	$E3,$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$F6,$1A,$F8
	.dc.b	$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F,$5F
	.dc.b	$CF,$EB,$FA,$3D,$7F,$4F,$AF,$F3,$3E,$97,$E6,$F2,$9F,$E6,$07,$E3
	.dc.b	$CF,$C7,$3F,$69,$7A,$DC,$7D,$6E,$5E,$B7,$F0,$5E,$BE,$9E,$B7,$5F
	.dc.b	$5B,$B7,$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$78,$FD,$6F,$27
	.dc.b	$AD,$E5,$F5,$BC,$DE,$B7,$9F,$D6,$F4,$7A,$DE,$9F,$5B,$D5,$EB,$7E
	.dc.b	$1B,$E9,$BF,$C2,$E7,$F8,$58,$FC,$69,$F8,$CE,$72,$F5,$F8,$7A,$FC
	.dc.b	$7D,$EF,$E3,$DE,$77,$3F,$7B,$A7,$BD,$F9,$2F,$C5,$FF,$C1,$67,$E0
	.dc.b	$0F,$BF,$91,$FB,$ED,$F4,$38,$7A,$1C,$7D,$0E,$5E,$87,$3F,$43,$A7
	.dc.b	$A1,$D7,$D0,$ED,$E7,$F7,$F3,$FC,$1E,$7F,$87,$CF,$F1,$79,$FE,$3F
	.dc.b	$3F,$C9,$E7,$FC,$1F,$6B,$E1,$7B,$5F,$0F,$DA,$FF,$57,$FA,$7F,$EB
	.dc.b	$0F,$33,$FD,$63,$D3,$FE,$B7,$FC,$C3,$FC,$D9,$F2,$47,$F9,$5B,$7E
	.dc.b	$03,$5E,$F6,$FA,$78,$74,$F1,$E9,$E5,$D3,$CF,$A7,$A7,$4F,$5E,$9E
	.dc.b	$DD,$3D,$FA,$7C,$1D,$3F,$5F,$DE,$FB,$1E,$F7,$8F,$AF,$ED,$7B,$DF
	.dc.b	$6F,$DE,$FB,$9E,$F7,$DD,$F7,$BE,$F7,$4F,$DF,$E9,$FC,$07,$4F,$F0
	.dc.b	$6F,$82,$FF,$1F,$9F,$B4,$3E,$A9,$F5,$07,$36,$F9,$B8,$73,$71,$E6
	.dc.b	$E5,$CD,$CF,$9B,$A7,$37,$5E,$6E,$DC,$DD,$F9,$BC,$1C,$9E,$1F,$03
	.dc.b	$C5,$E0,$78,$FC,$0F,$27,$BD,$E5,$F7,$BC,$DE,$F7,$F1,$CF,$A9,$FD
	.dc.b	$89,$FE,$1F,$3F,$9F,$1F,$B0,$F1,$4F,$EC,$1F,$E5,$6B,$DD,$E5,$EE
	.dc.b	$F3,$F7,$7A,$7B,$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF
	.dc.b	$17,$BB,$E3,$F7,$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE
	.dc.b	$FA,$7D,$DF,$E6,$BF,$55,$FE,$86,$3F,$D0,$A7,$F3,$C3,$F6,$07,$EB
	.dc.b	$F9,$7F,$A1,$AF,$73,$8F,$B9,$CB,$DC,$E7,$EE,$74,$F6,$FA,$FB,$7D
	.dc.b	$BD,$BE,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F
	.dc.b	$2F,$AF,$E6,$F5,$FC,$FE,$BF,$F4,$CF,$AB,$FF,$42,$9F,$CE,$8F,$DD
	.dc.b	$9F,$BA,$3F,$0F,$C9,$FD,$69,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7
	.dc.b	$AD,$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F
	.dc.b	$5B,$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5,$BF,$B1,$FC,$07,$F3,$83
	.dc.b	$FC,$CC,$7F,$81,$4F,$C4,$1F,$87,$E4,$FE,$6F,$EB,$70,$F5,$B8,$FA
	.dc.b	$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD,$DB,$D6,$EF,$EB,$78,$3D,$6F
	.dc.b	$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6,$F2,$FA,$DE,$6F,$5B,$CF,$EB
	.dc.b	$7F,$7A,$FA,$CF,$F6,$59,$FE,$CA,$3F,$C0,$27,$F3,$53,$F0,$DC,$87
	.dc.b	$F0,$CF,$F8,$29,$7A,$BC,$FD,$5E,$9E,$AF,$5F,$57,$B7,$AB,$DF,$D5
	.dc.b	$F0,$7A,$BE,$1F,$57,$C5,$EA,$F8,$FD,$5F,$27,$AB,$E5,$F5,$7C,$DE
	.dc.b	$AF,$9F,$D5,$F4,$7A,$BF,$E1,$2F,$AD,$FF,$65,$9F,$E0,$03,$FA,$09
	.dc.b	$FD,$00,$FC,$3F,$19,$FC,$3B,$FE,$22,$7F,$DA,$23,$D3,$E9,$E9,$F5
	.dc.b	$F4,$FB,$7A,$7D,$FD,$3F,$07,$A7,$E1,$F4,$FC,$5E,$9F,$8F,$D3,$F2
	.dc.b	$7A,$7E,$5F,$4F,$CD,$E9,$F9,$FD,$3F,$47,$A7,$FE,$2A,$FA,$EF,$EF
	.dc.b	$E7,$FA,$B4,$FF,$26,$1F,$C3,$CF,$C4,$1F,$C3,$78,$BF,$C7,$2B,$D2
	.dc.b	$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9,$78
	.dc.b	$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5E,$7F,$4B
	.dc.b	$D1,$E9,$7F,$90,$BD,$B7,$CC,$5F,$97,$C7,$C5,$E7,$DF,$D7,$C2,$EF
	.dc.b	$DE,$E1,$EC,$71,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1
	.dc.b	$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6
	.dc.b	$3C,$DE,$C7,$F9,$63,$EB,$DF,$2F,$7E,$47,$1F,$0B,$9F,$85,$D7,$66
	.dc.b	$FE,$0B,$87,$C1,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7
	.dc.b	$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$E5
	.dc.b	$F7,$BC,$DE,$F7,$F9,$BB,$EC,$3F,$C8,$67,$FB,$C4,$FF,$48,$1F,$D4
	.dc.b	$CF,$E1,$87,$E1,$7D,$F3,$F8,$57,$FC,$FE,$FF,$BC,$47,$B9,$D7,$DC
	.dc.b	$ED,$EE,$77,$F7,$3C,$1E,$E7,$87,$DC,$F1,$7B,$9E,$3F,$73,$C9,$EE
	.dc.b	$79,$7D,$CF,$37,$B9,$E7,$F7,$3D,$1E,$E7,$A7,$DC,$FF,$43,$FD,$8F
	.dc.b	$E1,$B8,$BF,$D2,$2B,$DC,$CF,$B9,$AF,$73,$7E,$E7,$0F,$73,$8F,$B9
	.dc.b	$CB,$DC,$E7,$EE,$74,$F7,$3A,$FB,$9D,$BD,$CE,$FE,$E7,$83,$DC,$F0
	.dc.b	$FB,$9E,$2F,$73,$C7,$EE,$79,$3D,$CF,$2F,$B9,$E6,$F7,$3C,$FE,$E7
	.dc.b	$FA,$7B,$8B,$F0,$FC,$FF,$EA,$87,$FD,$1A,$7F,$93,$8F,$F2,$69,$FC
	.dc.b	$38,$FE,$1A,$FF,$96,$FD,$5E,$5E,$AF,$3F,$57,$A7,$AB,$D7,$D5,$ED
	.dc.b	$EA,$F7,$F5,$7C,$1E,$AF,$87,$D5,$F1,$7A,$BE,$3F,$57,$C9,$EA,$F9
	.dc.b	$7D,$5F,$37,$AB,$E7,$F5,$7F,$D6,$BF,$65,$FE,$8A,$1F,$31,$8F,$95
	.dc.b	$CF,$C8,$EB,$E2,$37,$F1,$1C,$3E,$17,$8F,$C2,$F2,$F7,$39,$FB,$9D
	.dc.b	$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC,$F1
	.dc.b	$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$FE,$CC,$FB,$3F,$C3
	.dc.b	$79,$07,$F0,$CF,$FB,$61,$FF,$6A,$9F,$E8,$93,$FA,$A9,$FD,$51,$E1
	.dc.b	$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D
	.dc.b	$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1
	.dc.b	$EC,$7F,$B9,$3E,$D3,$F0,$FC,$47,$FD,$A6,$7F,$A5,$0F,$F4,$99,$FD
	.dc.b	$50,$FC,$38,$F6,$39,$7B,$1C,$FD,$8E,$9E,$C7,$5F,$63,$B7,$B1,$DF
	.dc.b	$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F,$27,$B1,$E5,$F6,$3C
	.dc.b	$DE,$C7,$9F,$D8,$FF,$7D,$FA,$7F,$ED,$03,$F0,$FC,$63,$FC,$B9,$EB
	.dc.b	$6B,$D6,$DF,$AD,$C2,$B8,$FA,$DC,$BD,$6E,$7E,$B7,$4F,$5B,$AF,$AD
	.dc.b	$DB,$D6,$EF,$EB,$78,$3D,$6F,$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6
	.dc.b	$F2,$FA,$DE,$6F,$5B,$CF,$EB,$7A,$3D,$6F,$37,$33,$ED,$7F,$12,$7E
	.dc.b	$21,$C7,$BB,$A8,$FF,$68,$1F,$E1,$C3,$F1,$07,$F8,$6A,$3E,$02,$F5
	.dc.b	$B8,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7,$AD,$D7,$D6,$ED,$EB,$77
	.dc.b	$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B,$C9,$EB,$7A,$BE,$DB
	.dc.b	$F1,$23,$E1,$B1,$F0,$D9,$F8,$6D,$7C,$36,$FE,$1B,$87,$B7,$C7,$DB
	.dc.b	$E5,$ED,$F3,$F6,$FA,$7B,$7D,$7D,$BE,$DE,$DF,$7F,$6F,$C1,$ED,$F8
	.dc.b	$7D,$BF,$17,$B7,$E3,$F6,$FC,$9E,$DF,$97,$DB,$F3,$7B,$7F,$43,$ED
	.dc.b	$FE,$8F,$8E,$7F,$12,$DF,$73,$5E,$E6,$FD,$CE,$1E,$E7,$1F,$73,$97
	.dc.b	$B9,$CF,$DC,$E9,$EE,$75,$F7,$3B,$7B,$9D,$FD,$CF,$07,$B9,$E1,$F7
	.dc.b	$3C,$5E,$E7,$8F,$DC,$F2,$7B,$9E,$5F,$73,$CD,$EE,$79,$FD,$CF,$47
	.dc.b	$B9,$F8,$8F,$1F,$F8,$79,$F8,$71,$EE,$67,$E1,$35,$F0,$7B,$F8,$4E
	.dc.b	$1F,$03,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D,$FE,$DE,$FF,$7F
	.dc.b	$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E,$FF,$97,$DF,$F3
	.dc.b	$7B,$FF,$9E,$FB,$8F,$E1,$A7,$E1,$7C,$BF,$DB,$2F,$7B,$5E,$F6,$FD
	.dc.b	$EE,$1D,$5C,$7D,$EE,$5E,$F7,$3F,$7B,$A7,$BD,$D7,$DE,$ED,$EF,$77
	.dc.b	$F7,$BC,$1E,$F7,$87,$DE,$F1,$7B,$DE,$3F,$7B,$C9,$EF,$79,$7D,$EF
	.dc.b	$37,$BD,$E7,$F7,$BF,$6B,$F7,$3F,$86,$3F,$0A,$3D,$EC,$FB,$DA,$F7
	.dc.b	$B7,$EF,$70,$F7,$B8,$FB,$DC,$BD,$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB
	.dc.b	$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD,$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2
	.dc.b	$FB,$DE,$6F,$7B,$CF,$EF,$7F,$0C,$FB,$AF,$EA,$67,$F0,$B3,$F0,$BE
	.dc.b	$67,$F2,$55,$EF,$6F,$DE,$E1,$EF,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E
	.dc.b	$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD
	.dc.b	$EF,$27,$BD,$E5,$F7,$BC,$DE,$F7,$9F,$DE,$FE,$47,$C3,$FD,$48,$FC
	.dc.b	$37,$2F,$FB,$F7,$DD,$CF,$BB,$AF,$77,$7E,$EF,$0F,$77,$8F,$BB,$CB
	.dc.b	$DD,$E7,$EE,$F4,$F7,$7A,$FB,$BD,$BD,$DE,$FE,$EF,$83,$DD,$F0,$FB
	.dc.b	$BE,$2F,$77,$C7,$EE,$F9,$3D,$DF,$2F,$BB,$E6,$F7,$7F,$9C,$FD,$DF
	.dc.b	$F0,$A3,$F8,$49,$F8,$5E,$6F,$EA,$4B,$DC,$DF,$B9,$C3,$DC,$E3,$EE
	.dc.b	$72,$F7,$39,$FB,$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C
	.dc.b	$3E,$E7,$8B,$DC,$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9
	.dc.b	$E8,$F7,$3F,$A8,$FD,$E7,$F8,$60,$FE,$A0,$7E,$1B,$9C,$FE,$19,$FE
	.dc.b	$D8,$FF,$BB,$0F,$F0,$C1,$FD,$45,$DF,$B3,$D3,$D9,$EB,$EC,$F6,$F6
	.dc.b	$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE
	.dc.b	$CF,$9B,$D9,$F3,$FB,$3E,$8F,$67,$FB,$4F,$DE,$FF,$86,$0F,$C3,$9F
	.dc.b	$C2,$F8,$7F,$C0,$4B,$D8,$DF,$B1,$C3,$D8,$E3,$EC,$72,$F6,$39,$FB
	.dc.b	$1D,$3D,$8E,$BE,$C7,$6F,$63,$BF,$B1,$E0,$F6,$3C,$3E,$C7,$8B,$D8
	.dc.b	$F1,$FB,$1E,$4F,$63,$CB,$EC,$79,$BD,$8F,$3F,$B1,$FE,$00,$F5,$3F
	.dc.b	$A8,$9F,$D4,$0F,$C4,$73,$7F,$B9,$7D,$7D,$7A,$FB,$F5,$F8,$57,$1F
	.dc.b	$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F,$07,$AF
	.dc.b	$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB,$F9,$FD
	.dc.b	$7F,$F0,$BF,$DF,$7F,$97,$8F,$F0,$B1,$F8,$73,$F0,$DC,$C5,$EE,$F0
	.dc.b	$F7,$38,$FB,$BC,$BD,$AE,$7E,$D7,$4F,$6B,$AF,$B5,$DB,$DA,$EF,$ED
	.dc.b	$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7,$93,$DA,$F2,$FB,$5E,$6F
	.dc.b	$6B,$CF,$ED,$7F,$8B,$BE,$FF,$FC,$BA,$7F,$0C,$3F,$0D,$CA,$7F,$0C
	.dc.b	$FF,$92,$57,$B3,$C7,$D9,$E5,$EC,$F3,$F6,$7A,$7B,$3D,$7D,$97,$F0
	.dc.b	$0C,$2E,$5C,$E1,$F8,$07,$F9,$57,$E1,$9C,$FC,$63,$0D,$AF,$EF,$6D
	.dc.b	$9D,$C7,$39,$DB,$4D,$96,$36,$FB,$6B,$C2,$E8,$C4,$F2,$69,$9D,$53
	.dc.b	$C2,$B9,$C6,$2B,$83,$6B,$85,$6E,$FD,$04,$8F,$23,$5E,$F7,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF
	.dc.b	$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63
	.dc.b	$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF
	.dc.b	$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88
	.dc.b	$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A
	.dc.b	$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB
	.dc.b	$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B
	.dc.b	$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D
	.dc.b	$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B
	.dc.b	$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60
	.dc.b	$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC
	.dc.b	$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE
	.dc.b	$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78
	.dc.b	$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12
	.dc.b	$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71
	.dc.b	$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C
	.dc.b	$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C
	.dc.b	$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6
	.dc.b	$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09
	.dc.b	$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC
	.dc.b	$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE
	.dc.b	$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A
	.dc.b	$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12
	.dc.b	$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D
	.dc.b	$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12
	.dc.b	$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2
	.dc.b	$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C
	.dc.b	$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09
	.dc.b	$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA
	.dc.b	$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB
	.dc.b	$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E
	.dc.b	$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD
	.dc.b	$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9
	.dc.b	$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A
	.dc.b	$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46
	.dc.b	$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31
	.dc.b	$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9
	.dc.b	$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26
	.dc.b	$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12
	.dc.b	$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3
	.dc.b	$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C
	.dc.b	$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F
	.dc.b	$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18
	.dc.b	$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A
	.dc.b	$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C
	.dc.b	$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67
	.dc.b	$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB
	.dc.b	$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D
	.dc.b	$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38
	.dc.b	$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E
	.dc.b	$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97
	.dc.b	$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D
	.dc.b	$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57
	.dc.b	$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA
	.dc.b	$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3
	.dc.b	$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF
	.dc.b	$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17
	.dc.b	$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF
	.dc.b	$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E
	.dc.b	$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D
	.dc.b	$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35
	.dc.b	$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1
	.dc.b	$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA,$31,$3A,$8C,$12,$BA,$12,$3D
	.dc.b	$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0,$D9,$DB,$67,$71,$E3,$9D,$B4
	.dc.b	$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE,$26,$2B,$AB,$4C,$CF,$12,$4F
	.dc.b	$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6,$12,$3D,$0D,$7C,$17,$D2,$78
	.dc.b	$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B,$E3,$7B,$38,$D6,$EF,$8C,$EA
	.dc.b	$31,$3A,$8C,$12,$BA,$12,$3D,$AD,$BF,$0C,$60,$8E,$09,$7E,$09,$E0
	.dc.b	$D9,$DB,$67,$71,$E3,$9D,$B4,$D9,$63,$6F,$BC,$97,$EC,$6D,$FA,$DE
	.dc.b	$26,$2B,$AB,$4C,$CF,$12,$4F,$1A,$EF,$18,$AE,$2D,$AE,$35,$BB,$F6
	.dc.b	$12,$3D,$0D,$7C,$17,$D2,$78,$46,$88,$6A,$78,$57,$7A,$E1,$7E,$2B
	.dc.b	$E3,$7B,$38,$D2,$12,$95,$2F,$10,$BE,$15,$BA,$D4,$F1,$9C,$BC,$AB
	.dc.b	$8C,$62,$77,$3C,$EF,$C2,$EC,$E4,$E2,$37,$F4,$6F,$83,$7C,$77,$92
	.dc.b	$FC,$B3,$98,$C4,$7A,$E3,$0E,$1A,$F1,$6B,$CA,$7A,$BB,$6B,$B6,$BB
	.dc.b	$9E,$EA,$79,$BE,$19,$DB,$67,$72,$F0,$1F,$27,$86,$D7,$9D,$AF,$9D
	.dc.b	$AF,$C0,$6B,$EF,$6B,$E3,$6C,$F7,$9E,$C4,$9E,$CD,$7B,$3F,$83,$FB
	.dc.b	$B7,$C1,$77,$8C,$36,$53,$A7,$F1,$5F,$48,$F8,$0F,$C4,$CE,$67,$C9
	.dc.b	$7C,$9F,$43,$5C,$9D,$2F,$F2,$37,$CE,$35,$87,$AD,$D1,$C8,$A7,$2B
	.dc.b	$F0,$4F,$18,$CC,$F1,$9F,$23,$6B,$D9,$1C,$30,$F4,$C6,$09,$5D,$E7
	.dc.b	$4F,$93,$1F,$B3,$1D,$1B,$3C,$25,$87,$8E,$B5,$8F,$12,$4F,$AA,$75
	.dc.b	$5C,$71,$F3,$4F,$86,$3E,$2E,$3E,$68,$C1,$0F,$85,$1A,$7A,$E1,$E0
	.dc.b	$7E,$1E,$3E,$67,$EC,$D7,$37,$89,$CE,$FC,$2F,$AA,$35,$18,$24,$F3
	.dc.b	$9D,$1E,$0B,$F5,$46,$70,$F6,$CF,$58,$E1,$5D,$A7,$33,$E2,$24,$66
	.dc.b	$FC,$35,$C0,$98,$FE,$2E,$F8,$D7,$92,$3C,$57,$EE,$29,$D2,$BC,$97
	.dc.b	$D2,$73,$5D,$EA,$D7,$67,$EB,$56,$DF,$97,$5D,$2F,$C3,$5B,$C7,$E8
	.dc.b	$1F,$88,$AD,$91,$CE,$3F,$5E,$30,$FC,$8B,$F6,$61,$DE,$B7,$5F,$7F
	.dc.b	$0E,$F5,$E6,$37,$14,$E6,$7C,$12,$9F,$05,$DB,$8C,$7E,$43,$EE,$39
	.dc.b	$57,$81,$FB,$A6,$F1,$F7,$63,$E9,$9D,$57,$27,$E0,$18,$9D,$D7,$8E
	.dc.b	$31,$7F,$27,$1F,$76,$1F,$06,$BC,$47,$79,$D3,$E4,$8F,$D7,$63,$F0
	.dc.b	$A3,$47,$0B,$EA,$FC,$08,$D4,$60,$F2,$CF,$77,$DD,$1A,$8C,$1E,$D2
	.dc.b	$CF,$D1,$BF,$3B,$F9,$BB,$F9,$77,$D6,$BC,$B5,$AA,$FA,$07,$96,$39
	.dc.b	$BF,$9F,$BF,$03,$CC,$D1,$6B,$CB,$5C,$F1,$F9,$D7,$E5,$AD,$3E,$07
	.dc.b	$E3,$1E,$72,$9D,$E7,$96,$3F,$52,$3E,$ED,$FA,$5E,$07,$CC,$7C,$2F
	.dc.b	$2B,$FD,$33,$9B,$F0,$34,$E3,$3C,$DF,$01,$C6,$BF,$62,$DC,$7E,$4D
	.dc.b	$7C,$4B,$EB,$7C,$5B,$19,$BE,$37,$DC,$EA,$59,$DB,$F1,$0F,$1C,$F1
	.dc.b	$BE,$F1,$93,$8D,$F7,$BE,$AD,$9D,$CF,$DA,$8F,$86,$F7,$9D,$D7,$58
	.dc.b	$FB,$26,$DF,$8B,$3C,$E3,$37,$F6,$9F,$29,$E3,$7D,$2F,$13,$25,$3C
	.dc.b	$F3,$DE,$7F,$37,$1C,$6B,$C6,$DB,$F4,$DF,$78,$E2,$64,$F1,$CF,$E0
	.dc.b	$E3,$C2,$6C,$AE,$4E,$B1,$E3,$8C,$CF,$E3,$B1,$FC,$DC,$F9,$DB,$3E
	.dc.b	$16,$D7,$E4,$5B,$5F,$BE,$AF,$98,$FA,$27,$C0,$D3,$B6,$3F,$9A,$78
	.dc.b	$61,$F3,$4F,$87,$5E,$5A,$F3,$E1,$EB,$69,$D2,$3E,$0D,$74,$7F,$28
	.dc.b	$60,$FC,$76,$1E,$67,$F8,$66,$1E,$16,$9C,$30,$F0,$E1,$DE,$FD,$13
	.dc.b	$CA,$B8,$57,$E2,$8F,$46,$3F,$BB,$9F,$45,$70,$9F,$D5,$61,$E1,$78
	.dc.b	$57,$E0,$4E,$18,$77,$BF,$44,$F0,$C3,$BD,$7E,$48,$E1,$3F,$CC,$DA
	.dc.b	$FF,$02,$6D,$F7,$78,$47,$E1,$EB,$DB,$8F,$C9,$6E,$3F,$26,$7E,$C9
	.dc.b	$E2,$AF,$E3,$F7,$CE,$3F,$37,$8F,$9E,$F8,$C7,$C5,$3E,$55,$7B,$30
	.dc.b	$EC,$DC,$3B,$57,$C8,$AF,$51,$E5,$9D,$3C,$CF,$1C,$76,$3E,$23,$D8
	.dc.b	$E2,$7D,$03,$D0,$7F,$05,$C3,$D0,$FE,$A0,$D4,$73,$9F,$6E,$3F,$3A
	.dc.b	$7D,$B8,$7A,$1E,$B8,$FB,$EB,$CB,$7E,$29,$F2,$C6,$0E,$6F,$94,$FA
	.dc.b	$83,$EF,$DF,$9C,$3B,$FF,$C8,$9F,$78,$FD,$41,$E2,$61,$B8,$7F,$9C
	.dc.b	$F0,$E3,$7F,$E4,$B6,$C7,$F9,$DE,$F9,$5F,$13,$58,$77,$9E,$4D,$7F
	.dc.b	$CE,$B3,$C1,$E5,$87,$81,$FF,$3C,$DF,$67,$FC,$F7,$F7,$8F,$DD,$9D
	.dc.b	$BF,$ED,$2F,$C0,$DF,$3D,$FE,$D7,$AD,$77,$E6,$31,$68,$C6,$0A,$E9
	.dc.b	$86,$1B,$1F,$EE,$1C,$3A,$BB,$BE,$4E,$70,$EB,$87,$19,$F2,$CE,$4E
	.dc.b	$18,$66,$64,$67,$0E,$EB,$0E,$F1,$EF,$E9,$E8,$49,$E1,$83,$E2,$F1
	.dc.b	$1C,$72,$F7,$D6,$30,$E7,$7B,$AE,$33,$9A,$DC,$74,$BC,$3E,$F9,$B8
	.dc.b	$7D,$19,$B6,$9B,$18,$6D,$76,$21,$E5,$3A,$1D,$CC,$E1,$E6,$8E,$31
	.dc.b	$BB,$F8,$4E,$B0,$F3,$B9,$C3,$D8,$F5,$FA,$57,$E2,$2F,$89,$FF,$79
	.dc.b	$C6,$1B,$19,$91,$0B,$3B,$70,$62,$38,$38,$8D,$36,$32,$F3,$8C,$B6
	.dc.b	$32,$F3,$9F,$19,$1E,$E7,$E3,$DF,$C6,$7C,$4C,$8F,$C0,$C7,$E3,$CC
	.dc.b	$3F,$0E,$30,$E8,$FC,$3B,$AC,$FF,$07,$3F,$97,$F8,$79,$5F,$8C,$7F
	.dc.b	$45,$EB,$7C,$0A,$FC,$4C,$6A,$7E,$83,$A7,$11,$A7,$91,$F9,$D9,$E5
	.dc.b	$9F,$ED,$57,$C3,$F1,$9D,$1E,$06,$38,$21,$F8,$38,$E2,$F5,$8E,$4E
	.dc.b	$4E,$27,$E0,$FD,$47,$A3,$FC,$03,$F0,$07,$84,$F0,$B0,$DB,$FD,$C7
	.dc.b	$03,$0E,$31,$86,$BF,$9B,$BE,$77,$C8,$FE,$1B,$5C,$70,$B7,$C3,$4F
	.dc.b	$E1,$77,$C3,$0F,$E0,$CD,$7D,$15,$FC,$5B,$C7,$7E,$38,$F8,$7E,$73
	.dc.b	$2B,$F8,$75,$6A,$78,$56,$E7,$D1,$5B,$7F,$65,$5B,$78,$CE,$DF,$09
	.dc.b	$C0,$CD,$74,$39,$61,$FC,$93,$2F,$74,$E8,$87,$22,$47,$F2,$AB,$F0
	.dc.b	$47,$8A,$FC,$13,$CF,$2F,$8D,$BF,$EC,$9B,$4B,$E3,$5C,$EF,$8D,$EF
	.dc.b	$2F,$2B,$FB,$CA,$DB,$67,$76,$C7,$D2,$F9,$A7,$59,$7B,$E3,$E1,$19
	.dc.b	$3C,$55,$C9,$7D,$73,$E1,$B8,$47,$B6,$BF,$A7,$BA,$8B,$3B,$8F,$8F
	.dc.b	$9F,$AE,$B3,$52,$73,$87,$91,$FE,$98,$73,$2D,$F3,$7C,$07,$03,$F8
	.dc.b	$EB,$9B,$FE,$B7,$1C,$4D,$91,$CC,$6E,$7F,$92,$36,$7E,$0D,$FB,$AB
	.dc.b	$77,$E7,$9F,$E3,$35,$E3,$7F,$94,$B5,$E7,$1F,$C7,$9F,$2B,$5F,$21
	.dc.b	$C1,$C4,$71,$8D,$9C,$5C,$3C,$0E,$A6,$63,$C9,$3E,$68,$FD,$3F,$86
	.dc.b	$F1,$B3,$8F,$09,$FE,$AC,$7C,$5A,$FC,$C5,$6E,$3E,$7D,$F0,$7C,$67
	.dc.b	$EB,$23,$4F,$F1,$7F,$82,$FE,$57,$F0,$52,$3F,$92,$47,$F2,$E3,$0F
	.dc.b	$F1,$D8,$C3,$A3,$F9,$53,$AB,$FF,$02,$F2,$AB,$75,$DC,$E0,$7F,$27
	.dc.b	$F8,$0C,$D7,$09,$B9,$7F,$33,$9B,$5F,$E2,$3B,$E8,$F2,$BE,$F3,$FD
	.dc.b	$10,$DF,$62,$B7,$5D,$CF,$E8,$13,$A8,$CD,$72,$9C,$9D,$6F,$FA,$26
	.dc.b	$1D,$63,$FC,$49,$FB,$BB,$E8,$7F,$C6,$8B,$CE,$A6,$22,$4F,$08,$CB
	.dc.b	$CA,$32,$FF,$82,$E7,$93,$B3,$47,$EB,$3C,$A7,$AF,$F9,$36,$C7,$E5
	.dc.b	$DA,$FF,$76,$6B,$FE,$06,$6B,$FE,$0B,$6B,$FE,$0E,$6A,$69,$23,$C0
	.dc.b	$48,$FD,$37,$93,$FB,$8C,$FF,$72,$DC,$FF,$C3,$0D,$9F,$F0,$DB,$F2
	.dc.b	$5F,$E0,$EE,$5F,$F4,$1C,$7F,$14,$8C,$1B,$8F,$D0,$B8,$8D,$1B,$C7
	.dc.b	$F8,$DE,$0F,$5F,$E3,$12,$FE,$33,$5F,$8C,$D7,$E3,$35,$F8,$CD,$7E
	.dc.b	$33,$FC,$17,$76,$6F,$7F,$F0,$64,$AF,$C3,$F5,$D7,$EF,$45,$D3,$FC
	.dc.b	$D7,$5F,$DE,$E7,$F1,$B3,$9C,$FF,$D1,$13,$D6,$79,$9F,$83,$F6,$3F
	.dc.b	$04,$47,$F0,$4D,$7F,$04,$D7,$F0,$4D,$7F,$04,$D7,$F0,$4F,$B6,$1E
	.dc.b	$EF,$E7,$89,$3F,$A2,$F5,$6B,$F9,$98,$B8,$FB,$26,$E1,$EA,$9B,$5F
	.dc.b	$E9,$DB,$57,$F5,$2F,$AF,$FF,$0F,$4F,$F9,$39,$AF,$F9,$79,$AF,$F9
	.dc.b	$8D,$AF,$F9,$99,$A9,$A4,$F0,$7F,$9D,$36,$3F,$9F,$35,$FF,$35,$EB
	.dc.b	$F7,$3D,$EF,$DE,$92,$3F,$80,$FA,$55,$FD,$C4,$5D,$3E,$DC,$DD,$3E
	.dc.b	$CC,$DA,$F8,$13,$EF,$F9,$DF,$19,$23,$C6,$D7,$E1,$35,$F8,$4D,$7E
	.dc.b	$13,$5F,$84,$FD,$9F,$77,$F8,$A9,$23,$F8,$CF,$35,$7F,$88,$45,$C7
	.dc.b	$F2,$73,$71,$FC,$8C,$DA,$FB,$13,$F5,$FE,$63,$E5,$92,$3E,$5B,$5F
	.dc.b	$B4,$D7,$ED,$35,$FB,$4D,$7E,$D3,$FA,$DF,$DE,$3F,$07,$87,$FB,$41
	.dc.b	$B8,$7F,$BA,$DB,$5F,$EE,$75,$8F,$FB,$9D,$CE,$86,$92,$35,$B4,$AF
	.dc.b	$F7,$04,$E5,$FF,$0D,$ED,$FC,$8D,$FC,$3F,$37,$E2,$88,$E0,$87,$E6
	.dc.b	$5C,$B4,$C9,$F9,$B7,$2F,$FA,$F7,$0F,$37,$BF,$E9,$F4,$E4,$F9,$23
	.dc.b	$D8,$61,$FF,$69,$61,$FE,$FC,$6C,$69,$F5,$CE,$F4,$F9,$53,$F1,$7D
	.dc.b	$E9,$5F,$01,$F9,$1E,$96,$23,$AC,$5A,$FA,$53,$76,$FA,$53,$6B,$E9
	.dc.b	$4F,$EE,$EF,$F5,$E4,$8F,$D7,$B5,$FD,$F3,$5F,$DF,$35,$FD,$F3,$5F
	.dc.b	$DF,$3F,$85,$F9,$5F,$BC,$48,$FB,$3E,$2D,$7E,$2C,$5C,$7F,$15,$37
	.dc.b	$1F,$C5,$4D,$AF,$C5,$4F,$F1,$3C,$BF,$83,$92,$3F,$83,$B5,$FE,$2C
	.dc.b	$D7,$F8,$B3,$5F,$E2,$CD,$7F,$8B,$3F,$9D,$F4,$3F,$2C,$48,$FC,$8F
	.dc.b	$A3,$5F,$BB,$17,$1F,$D1,$4D,$C7,$F4,$53,$6B,$F4,$53,$FC,$AF,$D5
	.dc.b	$FE,$3E,$48,$FE,$3E,$D7,$F9,$73,$5F,$E5,$CD,$7F,$97,$35,$FE,$5C
	.dc.b	$FE,$D7,$D2,$FD,$71,$20,$30,$61,$8E,$07,$EF,$3F,$EF,$0A,$FD,$C5
	.dc.b	$F1,$7F,$36,$F2,$79,$57,$19,$E1,$5F,$BB,$AE,$17,$FC,$15,$E0,$F8
	.dc.b	$2D,$3E,$07,$FA,$77,$C0,$FF,$42,$23,$FD,$09,$AF,$F5,$16,$BF,$D4
	.dc.b	$5A,$FF,$51,$6B,$FD,$45,$FF,$02,$5B,$D4,$F1,$E3,$E9,$6D,$79,$27
	.dc.b	$4F,$C8,$C3,$D1,$87,$C6,$F9,$2D,$BF,$22,$ED,$F0,$BF,$8A,$73,$8C
	.dc.b	$D6,$99,$1D,$70,$F1,$34,$F1,$4F,$F7,$1E,$3F,$EC,$A4,$8F,$EC,$AD
	.dc.b	$7F,$BA,$35,$FE,$E8,$D7,$FB,$A3,$5F,$EE,$8F,$C4,$F9,$1F,$49,$23
	.dc.b	$C5,$8F,$E1,$3E,$47,$11,$EC,$8F,$C3,$CE,$33,$F8,$B9,$AF,$19,$7E
	.dc.b	$16,$AE,$1F,$36,$7F,$C1,$9C,$BF,$E0,$02,$47,$F8,$01,$AF,$F8,$3D
	.dc.b	$AF,$F8,$3D,$AF,$F8,$3D,$AF,$F8,$3D,$FB,$FF,$21,$F5,$09,$1E,$CC
	.dc.b	$1F,$21,$88,$F6,$42,$71,$97,$B2,$AE,$5F,$9F,$9B,$7F,$87,$9F,$F1
	.dc.b	$37,$B9,$FE,$1A,$24,$7F,$86,$9A,$FF,$8A,$5A,$FF,$8A,$5A,$FF,$8A
	.dc.b	$5A,$FF,$8A,$5F,$CC,$FA,$1F,$8E,$24,$7A,$FE,$3E,$BE,$A8,$B8,$FD
	.dc.b	$49,$B8,$FE,$DE,$6D,$7E,$7E,$7F,$C7,$DF,$2F,$FE,$32,$24,$7F,$8C
	.dc.b	$9A,$FF,$90,$5A,$FF,$90,$5A,$FF,$90,$5A,$FF,$90,$5F,$D8,$FC,$7F
	.dc.b	$E9,$89,$1F,$4F,$D2,$AF,$D2,$8B,$8F,$E3,$26,$E3,$FC,$3E,$6D,$7E
	.dc.b	$DE,$7F,$CA,$DD,$9F,$E4,$A2,$47,$F9,$29,$AF,$F9,$65,$AF,$F9,$65
	.dc.b	$AF,$F9,$65,$AF,$F9,$65,$FF,$46,$73,$FF,$A7,$F0,$FE,$06,$DC,$3F
	.dc.b	$95,$36,$BF,$8E,$BF,$EC,$87,$F5,$0E,$78,$CD,$27,$A3,$FC,$31,$C7
	.dc.b	$A1,$4C,$47,$F9,$CF,$BB,$FC,$CE,$4B,$FF,$33,$B5,$FF,$3B,$35,$FF
	.dc.b	$3B,$35,$FF,$3B,$35,$FF,$3B,$3E,$EF,$3B,$C4,$48,$FE,$07,$7F,$EA
	.dc.b	$37,$D7,$19,$69,$AF,$3F,$19,$FC,$69,$B9,$FC,$69,$FF,$46,$7C,$BF
	.dc.b	$FA,$00,$97,$FE,$80,$6B,$FE,$8F,$6B,$FE,$8F,$6B,$FE,$8F,$6B,$FE
	.dc.b	$8F,$7E,$BF,$9F,$F2,$C9,$1E,$AF,$3A,$BF,$52,$2E,$3F,$72,$6E,$1F
	.dc.b	$72,$6D,$7D,$C9,$FF,$52,$7C,$BF,$FA,$60,$91,$FE,$98,$6B,$FE,$A7
	.dc.b	$6B,$FE,$A7,$6B,$FE,$A7,$6B,$FE,$A7,$7F,$1F,$E7,$FE,$0C,$91,$F4
	.dc.b	$BC,$FA,$FE,$32,$2E,$3F,$94,$9B,$8F,$E5,$26,$D7,$E5,$27,$FD,$79
	.dc.b	$E2,$FF,$AC,$09,$1F,$EB,$06,$BF,$EB,$F6,$BF,$EB,$F6,$BF,$EB,$F6
	.dc.b	$BF,$EB,$F7,$F4,$FE,$7F,$E6,$C9,$1F,$8A,$F3,$EB,$FA,$48,$B8,$FE
	.dc.b	$B2,$6E,$3F,$AC,$9B,$5F,$AC,$9F,$F6,$A7,$8F,$FE,$C8,$24,$7F,$B2
	.dc.b	$1A,$FF,$B5,$DA,$FF,$B5,$DA,$FF,$B5,$DA,$FF,$B5,$DF,$F0,$F3,$FE
	.dc.b	$81,$FD,$EB,$FD,$2E,$4F,$3B,$FC,$02,$FF,$A7,$23,$FB,$3B,$96,$C6
	.dc.b	$BC,$FD,$B2,$32,$70,$8F,$F7,$77,$CC,$7F,$B8,$89,$1F,$EE,$26,$BF
	.dc.b	$EF,$16,$BF,$EF,$16,$BF,$EF,$16,$BF,$EF,$17,$D6,$FF,$8F,$E9,$E6
	.dc.b	$FA,$89,$1E,$AF,$36,$BF,$DF,$E7,$6F,$BA,$7E,$0C,$EB,$5F,$F1,$2D
	.dc.b	$6A,$B8,$56,$9F,$F5,$BF,$CC,$7F,$BF,$88,$FF,$BF,$9A,$F9,$BD,$14
	.dc.b	$62,$3E,$9C,$61,$E0,$43,$84,$6D,$F7,$79,$7F,$58,$8F,$D6,$F2,$6B
	.dc.b	$ED,$8B,$8F,$AA,$6F,$0F,$2A,$FB,$B5,$F8,$09,$FC,$0F,$81,$F8,$2E
	.dc.b	$D9,$5F,$82,$8C,$BF,$84,$8C,$BF,$2E,$32,$FC,$D8,$CE,$1F,$2A,$7F
	.dc.b	$CA,$72,$CB,$FA,$3B,$F6,$8C,$76,$D6,$99,$8D,$BF,$2E,$35,$18,$8F
	.dc.b	$97,$18,$23,$C5,$AF,$13,$F1,$3E,$03,$67,$37,$E8,$89,$3F,$A2,$E4
	.dc.b	$AF,$E2,$62,$E5,$F8,$69,$BB,$7E,$4E,$6D,$7F,$A1,$2D,$5F,$0B,$EC
	.dc.b	$BD,$D2,$69,$1F,$84,$D7,$E3,$35,$FD,$0B,$5F,$DE,$B5,$FD,$26,$5F
	.dc.b	$BC,$23,$FE,$62,$FD,$EF,$FD,$3B,$7F,$33,$2F,$CF,$D6,$AB,$FC,$C9
	.dc.b	$5A,$8C,$4E,$A7,$91,$EB,$F8,$4F,$39,$1F,$3B,$5F,$3B,$5F,$3B,$5F
	.dc.b	$F5,$96,$08,$C4,$7F,$AF,$A3,$0F,$D8,$BF,$F4,$A5,$7F,$A6,$9B,$9F
	.dc.b	$FB,$05,$B3,$FE,$B4,$5E,$7F,$FA,$79,$CE,$A6,$92,$35,$E8,$4A,$F9
	.dc.b	$0F,$18,$FB,$5F,$3F,$F5,$89,$3F,$59,$AF,$D6,$6B,$F5,$9A,$FC,$06
	.dc.b	$BE,$F7,$F4,$97,$E5,$F4,$7D,$04,$9F,$B5,$E8,$D7,$EB,$C5,$CF,$E2
	.dc.b	$4D,$CF,$E2,$4D,$AF,$89,$3F,$93,$CB,$F1,$E4,$8F,$C7,$B5,$FC,$7B
	.dc.b	$5F,$C7,$B5,$FB,$2D,$7E,$C3,$EA,$CF,$FD,$25,$1F,$91,$F4,$3E,$91
	.dc.b	$F4,$FD,$0B,$19,$21,$F9,$27,$2F,$E5,$22,$E5,$F8,$19,$FD,$66,$BF
	.dc.b	$A8,$25,$7E,$A1,$AF,$EA,$1A,$FE,$A1,$AF,$E4,$9A,$FE,$45,$F5,$79
	.dc.b	$FF,$89,$24,$7E,$AF,$CF,$AF,$EA,$22,$DF,$E5,$66,$EB,$F9,$59,$B5
	.dc.b	$F9,$59,$FE,$09,$AF,$F0,$02,$47,$F0,$06,$BF,$C0,$1A,$FF,$00,$6B
	.dc.b	$FA,$A6,$BF,$A9,$7E,$97,$9D,$FA,$12,$47,$F0,$3F,$3A,$BF,$C0,$22
	.dc.b	$E3,$FA,$D9,$B8,$FE,$B6,$6D,$7E,$B6,$7F,$8E,$69,$FC,$60,$91,$FC
	.dc.b	$61,$AF,$F1,$86,$BF,$C6,$1A,$FF,$02,$6B,$FC,$05,$FD,$1F,$EF,$9F
	.dc.b	$E8,$F2,$79,$E9,$FF,$7E,$C7,$FA,$57,$16,$39,$E2,$4F,$0A,$DD,$F1
	.dc.b	$57,$C1,$57,$0F,$3F,$53,$98,$E7,$58,$C3,$B6,$1D,$70,$E7,$18,$C3
	.dc.b	$51,$D2,$75,$87,$37,$D9,$3A,$C3,$AD,$79,$88,$74,$39,$46,$E7,$9D
	.dc.b	$F8,$B0,$C4,$E7,$CE,$97,$F1,$9E,$1F,$2D,$88,$E9,$8F,$C6,$9C,$65
	.dc.b	$E5,$AB,$8F,$BA,$6D,$FA,$E7,$C5,$E6,$C8,$FA,$6F,$97,$E3,$71,$1D
	.dc.b	$B4,$4E,$31,$F8,$55,$71,$F9,$73,$6F,$E4,$4F,$BF,$CC,$F8,$1F,$2B
	.dc.b	$27,$F0,$31,$97,$F0,$71,$97,$C1,$19,$7C,$51,$9C,$3E,$94,$E5,$FA
	.dc.b	$91,$9C,$3E,$EC,$E5,$FB,$F1,$9A,$FB,$93,$97,$EF,$46,$74,$FC,$8C
	.dc.b	$FF,$80,$7E,$03,$FC,$47,$F0,$4F,$9C,$A6,$70,$FC,$E6,$5F,$93,$7F
	.dc.b	$C0,$D5,$A7,$39,$7E,$66,$F9,$46,$0F,$ED,$13,$CD,$FC,$40,$8D,$B8
	.dc.b	$6D,$77,$EE,$57,$1A,$FF,$06,$BF,$E1,$42,$3F,$E1,$46,$BF,$E1,$46
	.dc.b	$BC,$0D,$10,$D1,$67,$C1,$1E,$77,$E0,$57,$F7,$A9,$F8,$27,$B4,$A7
	.dc.b	$B2,$BC,$EE,$63,$84,$62,$35,$5D,$6F,$E4,$76,$7E,$A7,$E0,$25,$7F
	.dc.b	$0A,$7F,$59,$F0,$18,$8F,$CC,$45,$CB,$F7,$13,$6F,$F8,$0C,$DA,$FD
	.dc.b	$F4,$DD,$7F,$88,$CD,$E1,$4D,$8C,$36,$BC,$47,$F1,$C7,$F8,$20,$F7
	.dc.b	$A9,$23,$F0,$17,$D3,$CD,$C4,$7E,$F2,$B8,$3A,$6B,$98,$DE,$5F,$C7
	.dc.b	$67,$2D,$BF,$0C,$EB,$2F,$E5,$95,$A6,$CE,$AF,$B3,$B7,$F9,$5C,$ED
	.dc.b	$B5,$9C,$3E,$1D,$5A,$FF,$58,$61,$E3,$F3,$3E,$09,$30,$F5,$79,$95
	.dc.b	$F7,$45,$9F,$54,$DC,$BC,$F3,$6B,$C7,$37,$1F,$A7,$37,$04,$D8,$C3
	.dc.b	$6B,$E5,$47,$D4,$8C,$47,$88,$F2,$9A,$21,$93,$EA,$3C,$30,$F7,$47
	.dc.b	$03,$EC,$7A,$FF,$80,$23,$F6,$7D,$6A,$E8,$FB,$91,$F8,$68,$CC,$FD
	.dc.b	$B8,$B9,$7D,$89,$B5,$F5,$26,$E5,$F9,$49,$B9,$26,$C6,$1B,$5F,$8B
	.dc.b	$F3,$FF,$2B,$E4,$C8,$C1,$0F,$C9,$39,$7F,$29,$1F,$9C,$31,$97,$E4
	.dc.b	$67,$13,$C6,$30,$FE,$66,$32,$F2,$30,$43,$04,$8E,$2F,$88,$E9,$97
	.dc.b	$F8,$17,$F7,$DF,$F4,$19,$38,$BF,$B9,$3F,$E8,$89,$FD,$6B,$96,$C6
	.dc.b	$BC,$5D,$B2,$32,$70,$8C,$3F,$DB,$A7,$93,$63,$37,$FE,$0D,$9C,$B6
	.dc.b	$33,$7E,$7B,$B7,$FE,$16,$BF,$33,$D6,$33,$AA,$7D,$B7,$87,$AD,$FC
	.dc.b	$18,$C5,$65,$F8,$3C,$DF,$0B,$E7,$66,$3F,$0A,$32,$FC,$48,$CB,$ED
	.dc.b	$8C,$BE,$F8,$CE,$5E,$D9,$CB,$EF,$8C,$E1,$F4,$E7,$2F,$D5,$8C,$D7
	.dc.b	$BE,$72,$FC,$18,$CE,$5E,$F9,$FB,$3F,$09,$2B,$EC,$BF,$73,$C7,$C4
	.dc.b	$7D,$38,$B8,$FD,$19,$B8,$FE,$16,$6D,$7C,$F9,$FC,$2F,$44,$8F,$C2
	.dc.b	$BF,$88,$F1,$71,$1F,$7E,$2E,$1F,$76,$6E,$1F,$91,$9B,$5F,$6E,$6E
	.dc.b	$5F,$66,$6E,$C9,$B1,$86,$D7,$88,$FC,$BB,$FE,$4B,$FD,$F9,$E2,$5F
	.dc.b	$12,$57,$F4,$89,$CB,$67,$2F,$F9,$12,$3F,$CD,$99,$7F,$9A,$DD,$C6
	.dc.b	$67,$79,$FF,$99,$67,$73,$FE,$6A,$77,$19,$F7,$3C,$84,$9F,$17,$8B
	.dc.b	$5F,$04,$59,$E9,$37,$0F,$14,$DA,$ED,$3F,$E6,$EA,$7C,$DB,$C8,$C3
	.dc.b	$6B,$FC,$4D,$3A,$9C,$FC,$E6,$99,$3A,$AE,$0F,$F9,$C6,$BA,$D6,$A3
	.dc.b	$13,$A9,$FF,$45,$1D,$63,$94,$FD,$3F,$1B,$EA,$7C,$74,$9F,$A5,$19
	.dc.b	$7E,$A4,$65,$FA,$51,$97,$EA,$46,$6B,$E8,$4F,$C1,$F9,$97,$CB,$DD
	.dc.b	$26,$1F,$02,$72,$D9,$CB,$ED,$84,$F3,$99,$3C,$63,$77,$EE,$AE,$73
	.dc.b	$FE,$4E,$F8,$E7,$C4,$FF,$A8,$21,$18,$7F,$C7,$4D,$7F,$CA,$BF,$6D
	.dc.b	$FE,$84,$73,$3C,$67,$57,$FE,$63,$6F,$07,$FA,$E6,$F8,$5F,$86,$F8
	.dc.b	$11,$FF,$24,$46,$AF,$FC,$D8,$DA,$FF,$61,$5F,$4A,$FF,$66,$35,$4E
	.dc.b	$DF,$F6,$83,$5F,$6D,$F7,$7D,$7E,$6F,$C1,$9D,$57,$33,$A3,$C4,$CD
	.dc.b	$71,$C3,$CB,$5C,$4C,$61,$E6,$9C,$61,$EC,$9C,$BE,$88,$CB,$E2,$3E
	.dc.b	$59,$0C,$12,$BE,$64,$72,$8F,$7C,$F2,$9C,$C7,$17,$CA,$F8,$4C,$57
	.dc.b	$95,$C1,$D5,$C3,$F5,$4C,$C7,$27,$04,$70,$48,$E2,$F8,$A3,$A1,$F4
	.dc.b	$A7,$D1,$87,$CC,$7D,$15,$C0,$93,$B3,$C2,$E1,$FB,$26,$0D,$1E,$37
	.dc.b	$58,$7D,$79,$EF,$87,$6A,$DE,$1E,$52,$5F,$E5,$6B,$53,$D8,$E8,$70
	.dc.b	$8D,$3F,$28,$FB,$47,$8C,$FA,$66,$4E,$53,$F1,$5E,$47,$03,$A9,$C0
	.dc.b	$F8,$E7,$08,$F5,$CF,$93,$0F,$CA,$5F,$9B,$0F,$C9,$DF,$77,$E0,$3E
	.dc.b	$19,$F9,$CF,$95,$E2,$F9,$E3,$8B,$E2,$78,$BE,$48,$E2,$F9,$1E,$2F
	.dc.b	$9A,$38,$E1,$F2,$A3,$DA,$7A,$DC,$3E,$B3,$31,$C3,$0F,$6C,$79,$6F
	.dc.b	$F3,$93,$E2,$AE,$37,$E2,$7E,$55,$73,$24,$EE,$FF,$0D,$5B,$7E,$BC
	.dc.b	$FA,$EF,$E4,$60,$88,$43,$E9,$16,$FF,$8B,$B6,$34,$F4,$30,$60,$E2
	.dc.b	$64,$87,$E5,$8D,$9C,$A3,$AB,$E6,$C3,$DB,$1F,$C4,$BE,$03,$CF,$3F
	.dc.b	$65,$AF,$D3,$6C,$FC,$4C,$51,$8C,$7A,$1F,$CA,$1C,$57,$D6,$9D,$BF
	.dc.b	$83,$C3,$F1,$D3,$A8,$FB,$55,$C5,$D1,$E3,$29,$E3,$2D,$F8,$E7,$96
	.dc.b	$1E,$B9,$CD,$79,$1F,$E7,$33,$AB,$FE,$94,$D8,$FE,$A2,$D7,$FA,$C3
	.dc.b	$6F,$FA,$73,$6B,$CD,$3F,$95,$AF,$5B,$E0,$3E,$04,$ED,$F0,$99,$3D
	.dc.b	$E6,$0E,$47,$BC,$A7,$BC,$B5,$EF,$79,$5F,$BD,$CD,$FB,$DF,$ED,$B5
	.dc.b	$E1,$69,$C2,$FE,$0B,$4F,$82,$53,$E0,$96,$FE,$0B,$6B,$E0,$C7,$E2
	.dc.b	$A7,$E0,$CF,$88,$F8,$26,$CF,$74,$6D,$F7,$35,$F7,$34,$F7,$3F,$E2
	.dc.b	$DF,$D6,$5F,$41,$FE,$3B,$5E,$AD,$31,$12,$7F,$C7,$73,$A6,$CE,$5E
	.dc.b	$93,$C5,$C9,$C1,$FD,$BD,$30,$FD,$D7,$63,$8D,$DB,$A9,$CC,$3D,$5C
	.dc.b	$62,$BE,$11,$87,$95,$DA,$E7,$5F,$C1,$F0,$F0,$3C,$63,$A3,$9C,$3B
	.dc.b	$E1,$C6,$E4,$E4,$E1,$1D,$4C,$AB,$EB,$87,$AF,$B3,$C5,$D1,$2B,$94
	.dc.b	$E5,$B9,$7F,$A3,$DB,$3B,$7A,$65,$F0,$AF,$D8,$67,$0E,$F7,$FE,$63
	.dc.b	$9D,$57,$6F,$E0,$0F,$F4,$1C,$7C,$E6,$C7,$C9,$6E,$27,$FD,$16,$DE
	.dc.b	$CF,$99,$1A,$27,$6E,$22,$57,$77,$71,$C3,$9E,$4F,$8D,$E7,$D9,$88
	.dc.b	$F6,$9E,$38,$C6,$7F,$2B,$05,$63,$2F,$55,$5B,$F5,$4F,$9B,$E1,$E4
	.dc.b	$7C,$17,$D3,$C9,$88,$EF,$A7,$BE,$71,$97,$7A,$B8,$FC,$79,$B7,$F1
	.dc.b	$E6,$DF,$18,$EC,$7C,$F8,$CE,$4A,$B1,$86,$CF,$D5,$F4,$FE,$69,$1F
	.dc.b	$B1,$EA,$57,$CF,$69,$C6,$5F,$16,$AE,$5F,$76,$6D,$FD,$D9,$FB,$3F
	.dc.b	$0B,$23,$F0,$8F,$DC,$F4,$B1,$1D,$A2,$E1,$F4,$66,$E1,$F8,$99,$B5
	.dc.b	$F8,$99,$BA,$7E,$26,$6E,$49,$B1,$86,$D7,$88,$FC,$B3,$FB,$9F,$E0
	.dc.b	$17,$C3,$5E,$E9,$5F,$A8,$9C,$B6,$72,$FD,$C8,$FF,$46,$62,$77,$19
	.dc.b	$9D,$E7,$FB,$69,$D9,$A7,$11,$AC,$FF,$8A,$CE,$A3,$CC,$66,$BF,$89
	.dc.b	$46,$5F,$21,$E7,$C3,$58,$78,$EB,$8C,$E4,$87,$94,$E2,$7A,$E7,$8D
	.dc.b	$79,$6F,$C7,$58,$9E,$75,$F0,$7E,$37,$E5,$11,$F9,$1F,$1B,$5D,$1F
	.dc.b	$2E,$3E,$84,$67,$2F,$89,$17,$0F,$83,$36,$BD,$F3,$72,$F9,$D3,$72
	.dc.b	$4D,$8C,$36,$BE,$D7,$C5,$FD,$FF,$86,$91,$82,$1F,$6D,$CB,$F7,$63
	.dc.b	$F0,$C6,$32,$FB,$53,$89,$E3,$18,$7F,$11,$19,$79,$18,$21,$82,$47
	.dc.b	$17,$C4,$74,$CB,$F0,$7E,$9F,$E5,$FE,$16,$4F,$E5,$63,$F3,$46,$1F
	.dc.b	$CA,$46,$1D,$1F,$9B,$75,$5F,$93,$9D,$4F,$03,$47,$03,$F3,$CF,$0A
	.dc.b	$FC,$AD,$70,$AE,$55,$C0,$E4,$70,$39,$3F,$A1,$8E,$51,$F9,$AA,$E5
	.dc.b	$5D,$23,$95,$74,$79,$1D,$1E,$4F,$48,$FD,$3F,$A5,$FD,$CF,$E7,$FF
	.dc.b	$CE,$95,$FD,$91,$B3,$AA,$FD,$E5,$6D,$86,$08,$7F,$38,$9E,$2F,$F8
	.dc.b	$05,$BB,$7E,$FE,$7F,$85,$1C,$A7,$81,$83,$9D,$F8,$DD,$3F,$C0,$E3
	.dc.b	$51,$C0,$FE,$EF,$5D,$B6,$4F,$B7,$D3,$95,$A7,$E0,$7A,$78,$8F,$44
	.dc.b	$5C,$3E,$14,$DC,$BF,$C4,$38,$AB,$17,$ED,$AB,$92,$B5,$3D,$CF,$2C
	.dc.b	$EA,$33,$1E,$03,$26,$CC,$13,$0F,$1B,$EE,$F8,$2F,$B0,$47,$E0,$FA
	.dc.b	$75,$FA,$51,$67,$EE,$4D,$CB,$D7,$36,$BE,$B4,$DC,$7D,$53,$70,$4D
	.dc.b	$8C,$36,$BC,$87,$88,$A7,$88,$B8,$78,$9B,$87,$89,$A7,$88,$B8,$78
	.dc.b	$9B,$87,$89,$FB,$FE,$37,$E6,$7C,$69,$18,$21,$F9,$37,$2F,$E5,$63
	.dc.b	$F4,$06,$30,$FD,$04,$E2,$78,$C6,$1F,$C6,$C6,$5E,$46,$08,$60,$91
	.dc.b	$C5,$F1,$1D,$30,$FC,$5F,$8B,$FB,$1F,$A2,$93,$F9,$A8,$FD,$A1,$87
	.dc.b	$F5,$91,$87,$47,$E8,$1D,$57,$EE,$67,$53,$C0,$D1,$C0,$FD,$B3,$C2
	.dc.b	$BF,$3B,$5C,$2B,$95,$70,$39,$1C,$0F,$D7,$F8,$9F,$C2,$48,$FF,$9C
	.dc.b	$7A,$7F,$E0,$05,$E2,$3F,$82,$4D,$AF,$F3,$ED,$6A,$79,$90,$D9,$CC
	.dc.b	$FF,$41,$98,$24,$7F,$17,$9F,$F4,$7C,$F3,$9C,$D7,$EE,$27,$C0,$A3
	.dc.b	$FC,$2A,$E1,$FE,$2D,$16,$79,$57,$59,$F2,$35,$FF,$31,$C6,$AB,$B4
	.dc.b	$69,$AE,$8F,$4C,$F1,$24,$69,$F8,$71,$A8,$F3,$4F,$9F,$C2,$F6,$CF
	.dc.b	$F9,$B8,$F8,$9E,$16,$22,$46,$08,$78,$5C,$BE,$38,$E6,$62,$FC,$33
	.dc.b	$89,$E3,$18,$7E,$94,$65,$E4,$60,$86,$09,$1C,$5F,$85,$E0,$FA,$BC
	.dc.b	$29,$1F,$76,$3D,$A6,$1F,$91,$18,$74,$7E,$0D,$D4,$FC,$89,$D4,$F0
	.dc.b	$34,$70,$3C,$26,$0D,$18,$24,$7D,$29,$EB,$8F,$D2,$9F,$F4,$31,$FE
	.dc.b	$CD,$1F,$40,$F7,$3F,$DB,$F0,$FC,$9B,$FD,$CC,$ED,$AF,$FA,$A8,$E3
	.dc.b	$1C,$27,$94,$6F,$5E,$75,$C3,$1F,$D0,$D6,$AB,$93,$63,$F4,$B3,$BC
	.dc.b	$3F,$3D,$5E,$3B,$ED,$84,$8C,$9A,$C3,$A3,$86,$B9,$8D,$B4,$A7,$A8
	.dc.b	$B1,$C1,$FD,$73,$5D,$1C,$9C,$12,$39,$C6,$B0,$F7,$C7,$9E,$3E,$46
	.dc.b	$1F,$EF,$17,$E0,$C6,$67,$71,$F3,$FC,$8C,$63,$E7,$8D,$99,$3F,$86
	.dc.b	$7C,$7F,$EB,$08,$FF,$0F,$F1,$EB,$FB,$98,$B7,$FB,$69,$B5,$F8,$59
	.dc.b	$B5,$FE,$A0,$9B,$C1,$FC,$3E,$6C,$FF,$B1,$4F,$09,$0E,$C7,$84,$C5
	.dc.b	$FF,$04,$8E,$35,$FC,$0B,$C5,$95,$FC,$95,$FE,$0B,$E2,$E2,$3F,$88
	.dc.b	$45,$BF,$E1,$B3,$71,$FC,$E4,$DA,$FF,$5B,$CD,$C7,$F9,$34,$D9,$FF
	.dc.b	$63,$1D,$C8,$76,$3F,$9B,$4E,$E3,$83,$FC,$D3,$E6,$5E,$35,$D6,$5F
	.dc.b	$F4,$79,$CB,$67,$2F,$F2,$E8,$FC,$F6,$3F,$C9,$1D,$C6,$67,$79,$7F
	.dc.b	$4C,$9D,$CF,$FB,$59,$DC,$66,$77,$5C,$EF,$CD,$12,$FC,$91,$A7,$94
	.dc.b	$65,$F2,$46,$72,$61,$DA,$FC,$D7,$98,$F1,$CE,$A3,$96,$1E,$7B,$D4
	.dc.b	$62,$FC,$98,$78,$8F,$88,$7C,$7F,$1B,$E6,$7B,$F2,$30,$43,$E4,$39
	.dc.b	$7E,$54,$7D,$33,$18,$7C,$79,$C4,$F1,$8C,$3F,$56,$32,$F2,$30,$43
	.dc.b	$04,$8E,$2F,$88,$E9,$5F,$67,$C5,$FC,$27,$C6,$C9,$FC,$14,$7E,$1C
	.dc.b	$C3,$F8,$18,$C3,$A3,$F1,$0E,$AB,$F0,$13,$A9,$E0,$68,$E0,$7E,$2D
	.dc.b	$E1,$5F,$8B,$AE,$15,$CA,$B8,$1C,$8E,$07,$E5,$FE,$33,$F3,$9F,$1F
	.dc.b	$23,$04,$3F,$30,$E5,$FC,$D4,$7E,$8C,$C4,$FE,$5E,$71,$3C,$63,$0F
	.dc.b	$E9,$63,$2F,$23,$04,$30,$48,$E2,$F8,$8E,$97,$FA,$9A,$F0,$BF,$AB
	.dc.b	$AE,$B5,$FA,$99,$CD,$7E,$A6,$73,$5C,$63,$C1,$5D,$8E,$0F,$F9,$05
	.dc.b	$7C,$5F,$F8,$D1,$B1,$98,$C7,$8B,$96,$46,$4E,$51,$87,$FC,$5F,$3C
	.dc.b	$9B,$3C,$9E,$06,$0B,$1F,$E4,$E3,$FD,$5E,$58,$E8,$D8,$EE,$70,$AF
	.dc.b	$F1,$8C,$77,$35,$1E,$2A,$F6,$7C,$57,$B7,$C5,$93,$F0,$63,$2F,$C3
	.dc.b	$8C,$BE,$A8,$CB,$EC,$8C,$D7,$9E,$72,$FA,$63,$37,$E4,$9C,$BE,$68
	.dc.b	$CD,$38,$56,$BD,$CE,$61,$5C,$2B,$75,$F5,$FC,$29,$7F,$6D,$FB,$5E
	.dc.b	$16,$23,$E9,$C5,$C3,$E7,$4D,$C7,$E4,$CD,$AF,$04,$DC,$3C,$B3,$6D
	.dc.b	$36,$30,$DA,$F1,$1F,$8D,$70,$F8,$8C,$1A,$30,$49,$F2,$B9,$7F,$09
	.dc.b	$3D,$A3,$F3,$1E,$EF,$E5,$FA,$A4,$60,$87,$E4,$9C,$BF,$94,$8F,$CF
	.dc.b	$18,$AF,$C8,$4E,$27,$8C,$61,$FD,$34,$65,$E4,$60,$86,$09,$1C,$5F
	.dc.b	$11,$D3,$1F,$F4,$3F,$CE,$BD,$2F,$F5,$1B,$67,$FD,$68,$D7,$FD,$32
	.dc.b	$BD,$2D,$CC,$9D,$BC,$A3,$11,$C2,$2D,$FF,$AE,$AB,$94,$F0,$21,$B3
	.dc.b	$C3,$5C,$2F,$CF,$76,$7F,$31,$5C,$9E,$04,$BE,$D3,$E3,$8F,$5F,$89
	.dc.b	$EE,$23,$EE,$F8,$AA,$FB,$A2,$D7,$9E,$6E,$1F,$22,$6D,$78,$A6,$EC
	.dc.b	$C7,$FD,$61,$3C,$CE,$C7,$88,$86,$4F,$9B,$3C,$A3,$8E,$1F,$1E,$FE
	.dc.b	$AC,$E5,$D3,$6B,$9C,$F9,$DB,$87,$9D,$A7,$9C,$FB,$BE,$E7,$E0,$3D
	.dc.b	$D9,$18,$21,$F8,$07,$2F,$E0,$A3,$F0,$E6,$2B,$EE,$4E,$27,$8C,$61
	.dc.b	$FC,$5C,$65,$E4,$60,$86,$09,$1C,$5F,$11,$D2,$BF,$33,$F4,$5F,$98
	.dc.b	$F9,$29,$3F,$98,$8F,$CD,$98,$7F,$33,$18,$74,$7E,$79,$D5,$7E,$56
	.dc.b	$75,$3C,$0D,$1C,$0F,$D1,$BC,$2B,$F5,$75,$C2,$B9,$57,$03,$91,$C0
	.dc.b	$1F,$CC,$27,$FD,$F3,$ED,$8F,$F5,$9C,$A3,$3A,$E1,$B5,$CA,$74,$FF
	.dc.b	$3B,$AE,$4E,$67,$AC,$71,$8D,$55,$3C,$05,$8E,$31,$E5,$3A,$39,$89
	.dc.b	$1D,$4C,$95,$F5,$46,$A3,$94,$76,$75,$1D,$AB,$D3,$3C,$27,$2F,$18
	.dc.b	$CC,$EE,$BD,$75,$A2,$1D,$A3,$4F,$90,$C5,$73,$9E,$31,$B9,$C3,$F1
	.dc.b	$0F,$2B,$F0,$CC,$CF,$A6,$38,$1E,$AF,$92,$F9,$1E,$BC,$8C,$10,$F2
	.dc.b	$39,$7C,$D1,$F3,$CC,$61,$EE,$9C,$4F,$18,$C3,$F2,$23,$2F,$23,$04
	.dc.b	$30,$48,$E2,$F8,$8E,$98,$7E,$43,$E2,$7F,$00,$48,$FA,$BF,$27,$5F
	.dc.b	$99,$16,$BE,$9C,$DA,$E5,$5F,$72,$BF,$15,$39,$8F,$CA,$45,$C7,$F3
	.dc.b	$D3,$72,$4D,$8C,$36,$BF,$0F,$F1,$3F,$95,$F6,$24,$60,$87,$E0,$5C
	.dc.b	$BF,$83,$8F,$CE,$18,$CB,$F1,$73,$89,$E3,$18,$7F,$2B,$19,$79,$18
	.dc.b	$21,$82,$47,$17,$C4,$74,$CB,$F7,$B5,$E1,$7F,$7F,$5D,$6B,$F3,$B3
	.dc.b	$9A,$FD,$D4,$E6,$B8,$C7,$82,$BB,$1C,$1F,$D5,$FC,$4F,$EE,$7D,$89
	.dc.b	$18,$21,$FA,$57,$2F,$E9,$E3,$F8,$01,$89,$FD,$7C,$E2,$78,$C6,$1F
	.dc.b	$DC,$C6,$5E,$46,$08,$60,$91,$C5,$F1,$1D,$2B,$F9,$1E,$8F,$9B,$FF
	.dc.b	$35,$46,$1F,$F3,$9B,$7B,$FF,$8C,$46,$AB,$FC,$B4,$DA,$E5,$3A,$37
	.dc.b	$E0,$4A,$FE,$2D,$5D,$1E,$33,$97,$F8,$BC,$73,$35,$5D,$23,$8D,$6F
	.dc.b	$0F,$E3,$35,$A2,$46,$70,$F4,$57,$83,$0C,$3F,$D0,$BC,$F7,$87,$FC
	.dc.b	$D6,$1F,$46,$F1,$30,$D8,$FE,$60,$2B,$8B,$86,$BF,$CE,$4E,$73,$C8
	.dc.b	$ED,$1A,$CF,$FA,$4C,$F3,$BD,$3D,$AF,$C0,$F3,$AE,$18,$FF,$5A,$8E
	.dc.b	$05,$31,$1E,$27,$0D,$73,$5B,$69,$4E,$66,$63,$83,$F1,$E3,$4E,$8E
	.dc.b	$AE,$23,$71,$93,$A9,$E9,$8F,$82,$7C,$CF,$72,$C7,$C9,$8A,$FC,$9B
	.dc.b	$4E,$30,$F8,$F7,$F0,$AB,$C8,$7C,$79,$FC,$AF,$89,$F9,$6F,$16,$4F
	.dc.b	$E4,$23,$2F,$E4,$A3,$2F,$E4,$23,$2F,$E4,$A3,$37,$F8,$E9,$CB,$F9
	.dc.b	$08,$CE,$5F,$A7,$9C,$BF,$A9,$8C,$D7,$E9,$A7,$2F,$EA,$23,$3B,$FE
	.dc.b	$C6,$7F,$5B,$E2,$4A,$FD,$3B,$FB,$0F,$17,$11,$FA,$68,$B8,$FE,$8A
	.dc.b	$6E,$3F,$B9,$9B,$5F,$B6,$9B,$87,$EF,$E6,$F0,$FF,$AF,$2B,$B9,$0E
	.dc.b	$C7,$EE,$2B,$73,$C1,$FE,$07,$EB,$49,$FD,$EB,$FC,$1B,$C6,$C4,$7E
	.dc.b	$F2,$2E,$5F,$B8,$9B,$97,$F0,$F9,$B5,$FC,$32,$6E,$5F,$C5,$66,$E4
	.dc.b	$9B,$18,$6D,$78,$8F,$E1,$EE,$1F,$11,$83,$46,$09,$87,$95,$CB,$FC
	.dc.b	$3A,$7B,$47,$F8,$BF,$F1,$1F,$DC,$CF,$F8,$A9,$B8,$FF,$2C,$6C,$FF
	.dc.b	$BF,$DF,$F5,$E4,$EE,$7F,$C8,$51,$AC,$4D,$37,$1C,$CD,$92,$BC,$5E
	.dc.b	$2D,$7F,$C7,$F3,$C6,$7A,$CE,$6F,$C1,$5E,$99,$CB,$48,$72,$72,$F4
	.dc.b	$F5,$FC,$FE,$2C,$8F,$85,$1E,$A3,$0F,$8A,$30,$E8,$F9,$0E,$AB,$E4
	.dc.b	$4E,$A7,$81,$A3,$81,$E1,$30,$68,$C1,$23,$E9,$4F,$5A,$FA,$55,$F3
	.dc.b	$BD,$79,$7F,$65,$FA,$3E,$BE,$23,$E4,$C5,$C3,$EE,$4D,$BF,$B9,$36
	.dc.b	$BE,$E4,$DC,$FE,$CC,$DC,$D3,$63,$0D,$AF,$11,$F8,$B7,$0F,$88,$C1
	.dc.b	$A3,$04,$C3,$CA,$E5,$FC,$94,$F6,$8F,$CB,$7A,$9F,$8E,$23,$F9,$9F
	.dc.b	$80,$AF,$E2,$22,$D7,$E6,$E6,$E5,$F9,$B9,$B5,$F9,$B9,$B9,$7E,$62
	.dc.b	$6E,$49,$B1,$86,$D7,$90,$F1,$1F,$AB,$F5,$3F,$46,$47,$F5,$DF,$1B
	.dc.b	$5F,$CE,$C5,$CB,$F6,$13,$72,$FD,$84,$DA,$FD,$84,$DC,$BF,$5B,$37
	.dc.b	$24,$D8,$C3,$6B,$C8,$78,$8F,$F5,$77,$E2,$2F,$53,$FD,$74,$BD,$5A
	.dc.b	$62,$24,$F0,$8C,$BC,$A3,$2F,$FA,$DA,$79,$3B,$34,$43,$53,$C9,$E1
	.dc.b	$3C,$B3,$FF,$5D,$D7,$29,$D9,$C6,$33,$3C,$5D,$D7,$1C,$FF,$C8,$77
	.dc.b	$ED,$BF,$11,$C8,$EE,$73,$8D,$CF,$37,$8D,$F3,$96,$6B,$F2,$61,$E0
	.dc.b	$9E,$46,$A1,$EA,$FC,$CF,$56,$4F,$D0,$8F,$9E,$61,$F9,$11,$87,$47
	.dc.b	$D6,$75,$5F,$12,$75,$3C,$0D,$1C,$0F,$A4,$F0,$C7,$E1,$D7,$0A,$E5
	.dc.b	$5C,$0E,$47,$03,$F0,$BF,$03,$F8,$42,$3F,$88,$F5,$AB,$F7,$E2,$CF
	.dc.b	$D8,$9B,$7F,$4A,$6D,$7C,$59,$B9,$7C,$B9,$B9,$26,$C6,$1B,$5E,$43
	.dc.b	$C4,$7E,$6B,$C7,$FC,$CF,$7C,$8C,$10,$FC,$A3,$97,$F2,$D1,$FA,$03
	.dc.b	$19,$7E,$42,$71,$3C,$63,$0F,$EA,$23,$2F,$23,$04,$30,$48,$E2,$F8
	.dc.b	$8E,$99,$7E,$5A,$BC,$2F,$E6,$2B,$AD,$7E,$8A,$73,$5F,$8D,$9C,$D7
	.dc.b	$18,$F0,$57,$63,$83,$FB,$EF,$1B,$F7,$A4,$42,$13,$FE,$AA,$3F,$CE
	.dc.b	$4F,$EC,$A7,$84,$FE,$F5,$FF,$3C,$47,$08,$FD,$2C,$F0,$3F,$63,$3C
	.dc.b	$27,$94,$F8,$1E,$19,$FF,$A2,$23,$A1,$E7,$21,$E5,$AF,$23,$CD,$EA
	.dc.b	$EA,$79,$C7,$86,$3F,$65,$5B,$69,$E1,$8E,$6D,$3D,$3F,$01,$EC,$8F
	.dc.b	$84,$D8,$C9,$BB,$F8,$FF,$25,$97,$E4,$CD,$A3,$4C,$4D,$AE,$53,$A6
	.dc.b	$CE,$9A,$43,$E1,$1E,$22,$D7,$89,$D6,$1F,$0A,$7B,$46,$A9,$78,$AD
	.dc.b	$4F,$09,$FC,$04,$FA,$7C,$2F,$57,$C0,$CA,$F3,$C6,$5F,$4C,$65,$ED
	.dc.b	$19,$7C,$11,$9B,$FA,$B3,$97,$EB,$C6,$6F,$F2,$13,$97,$F2,$51,$9A
	.dc.b	$FC,$6C,$E5,$FC,$7C,$67,$1F,$CC,$CE,$5F,$CD,$C6,$69,$39,$51,$98
	.dc.b	$C3,$6B,$EA,$F8,$5F,$6C,$8F,$D8,$F0,$AB,$F3,$62,$E3,$F9,$29,$B8
	.dc.b	$7E,$AE,$6D,$7E,$9A,$6E,$3F,$B0,$9B,$5F,$E9,$9A,$E8,$78,$88,$7D
	.dc.b	$BA,$DC,$F0,$7F,$21,$E0,$49,$FC,$6B,$F9,$3F,$07,$11,$F8,$38,$B8
	.dc.b	$FE,$A2,$6E,$1F,$C0,$66,$D7,$EF,$26,$E3,$FC,$1A,$6D,$7F,$A6,$6B
	.dc.b	$B9,$0E,$C7,$E7,$AB,$73,$C1,$FD,$47,$E2,$3F,$E8,$73,$FD,$0B,$3F
	.dc.b	$DB,$9B,$67,$57,$C3,$C2,$95,$FB,$39,$CB,$63,$2F,$E9,$A3,$2D,$8C
	.dc.b	$E3,$FC,$4A,$72,$D8,$CD,$7F,$26,$BF,$13,$E8,$33,$7E,$79,$C3,$E3
	.dc.b	$8C,$CF,$A6,$73,$3E,$39,$CD,$FB,$8F,$0B,$DD,$C4,$F0,$BF,$37,$83
	.dc.b	$F2,$3C,$29,$87,$C1,$8F,$96,$61,$F8,$11,$87,$47,$C7,$75,$3D,$23
	.dc.b	$27,$03,$46,$1D,$39,$23,$82,$61,$E9,$9E,$AF,$AE,$BE,$FF,$85,$F8
	.dc.b	$0F,$56,$61,$F6,$63,$2F,$DB,$8C,$BF,$66,$32,$FD,$B8,$CE,$1F,$66
	.dc.b	$72,$FD,$B8,$CD,$FC,$B9,$CB,$F3,$63,$35,$F8,$F9,$CB,$F9,$18,$CE
	.dc.b	$7F,$91,$9F,$CB,$F8,$5F,$98,$F0,$A5,$FE,$46,$32,$FE,$4E,$32,$FE
	.dc.b	$46,$32,$FE,$4E,$33,$8F,$E4,$67,$2F,$E4,$E3,$38,$FE,$0E,$72,$FE
	.dc.b	$16,$33,$5F,$A7,$9C,$BF,$A9,$8C,$E1,$FA,$99,$CB,$FA,$B8,$CE,$A9
	.dc.b	$CB,$FE,$9F,$AF,$51,$ED,$21,$E9,$32,$7B,$4F,$D9,$78,$3F,$BC,$F0
	.dc.b	$64,$60,$87,$EC,$DC,$BF,$B5,$8F,$E0,$66,$32,$FD,$A4,$E2,$78,$C6
	.dc.b	$1F,$E0,$11,$97,$91,$82,$18,$24,$71,$7C,$47,$4C,$BF,$7D,$F3,$BF
	.dc.b	$E8,$78,$FE,$DA,$FF,$9F,$A3,$3E,$17,$F7,$F8,$D5,$7F,$86,$9B,$3A
	.dc.b	$AF,$F0,$34,$71,$7F,$D1,$2E,$9B,$E1,$73,$B9,$1C,$23,$B1,$87,$C7
	.dc.b	$18,$74,$76,$75,$7E,$39,$D4,$F0,$34,$70,$3D,$BE,$97,$AF,$C3,$91
	.dc.b	$E8,$8F,$71,$87,$E1,$C6,$1D,$1E,$E7,$55,$F0,$E7,$53,$C0,$D1,$C0
	.dc.b	$F0,$98,$34,$60,$91,$F3,$67,$AE,$3F,$5A,$BA,$CE,$6B,$93,$E1,$7B
	.dc.b	$39,$7A,$B9,$7A,$CF,$E0,$2B,$AC,$7D,$6A,$FB,$DE,$0C,$8F,$B6,$FE
	.dc.b	$07,$C1,$C4,$7E,$1A,$2E,$1F,$83,$9B,$87,$E3,$26,$D7,$E0,$A6,$F0
	.dc.b	$B4,$56,$1F,$F5,$15,$78,$08,$77,$3F,$1B,$5B,$9E,$0F,$E5,$7C,$09
	.dc.b	$3F,$92,$7F,$31,$E5,$62,$3F,$3B,$17,$2F,$CD,$4D,$CB,$F4,$73,$6B
	.dc.b	$F3,$33,$72,$F4,$4D,$C9,$36,$30,$DA,$F1,$1F,$A4,$7F,$CD,$8B,$EE
	.dc.b	$DF,$25,$8F,$C9,$BF,$89,$4F,$F9,$80,$38,$CF,$FD,$21,$1C,$67,$9D
	.dc.b	$72,$8F,$F3,$E1,$C3,$9B,$A1,$A9,$91,$AB,$E9,$A7,$FA,$92,$38,$1C
	.dc.b	$AF,$A3,$E3,$6B,$B3,$C4,$53,$FD,$5A,$E9,$E8,$D7,$35,$B6,$96,$34
	.dc.b	$F4,$3A,$38,$24,$71,$8D,$1E,$38,$F3,$E1,$F1,$9B,$5E,$2A,$F9,$DF
	.dc.b	$03,$F0,$AF,$D9,$16,$FE,$6E,$0A,$C5,$F1,$BF,$89,$7C,$9F,$F5,$27
	.dc.b	$E2,$5F,$EB,$1D,$8B,$E6,$BF,$D4,$CD,$9D,$98,$33,$11,$C9,$0F,$F4
	.dc.b	$FC,$71,$6C,$71,$75,$97,$FB,$32,$74,$D9,$D4,$F2,$3B,$1C,$8E,$38
	.dc.b	$7F,$8F,$EB,$D4,$72,$71,$3C,$AB,$B7,$8D,$EC,$25,$FC,$2F,$1A,$BE
	.dc.b	$88,$B3,$E5,$9B,$7D,$E6,$D3,$1F,$93,$58,$7E,$5C,$DC,$7F,$D7,$75
	.dc.b	$F4,$23,$07,$CE,$AB,$19,$3C,$87,$D6,$F1,$BE,$9F,$8D,$23,$04,$3E
	.dc.b	$7B,$97,$E8,$C7,$DA,$31,$97,$CF,$9C,$4F,$18,$C3,$F8,$48,$CB,$C8
	.dc.b	$C1,$0C,$12,$38,$BE,$23,$A6,$5F,$56,$BC,$2F,$D7,$AE,$B5,$F8,$19
	.dc.b	$CD,$7C,$BC,$7F,$D7,$EF,$23,$11,$AC,$BF,$2F,$7C,$B0,$FC,$17,$8B
	.dc.b	$F9,$82,$5F,$E7,$BE,$56,$BF,$94,$8B,$97,$E4,$66,$D7,$2A,$FD,$2D
	.dc.b	$7E,$8A,$73,$1E,$58,$B8,$7E,$A2,$6E,$5F,$EC,$3C,$3F,$39,$5C,$6F
	.dc.b	$99,$0F,$11,$E5,$3F,$67,$E3,$7E,$C0,$8F,$ED,$BC,$6A,$FE,$B2,$2D
	.dc.b	$7E,$A6,$6E,$5F,$A3,$9B,$5F,$2E,$6E,$5F,$BE,$9B,$8F,$FB,$0E,$BA
	.dc.b	$1E,$22,$1F,$BA,$AD,$CF,$07,$F9,$86,$0C,$3F,$D6,$1B,$FF,$3E,$8C
	.dc.b	$5F,$E9,$63,$13,$A3,$89,$C3,$C6,$B3,$B8,$B9,$7F,$13,$6D,$72,$9E
	.dc.b	$95,$89,$90,$17,$E2,$BE,$51,$98,$C5,$FF,$B6,$0E,$93,$A9,$E3,$5E
	.dc.b	$77,$83,$E7,$69,$E7,$2C,$71,$AE,$B3,$E8,$BF,$2D,$74,$38,$16,$B8
	.dc.b	$4F,$58,$C5,$F9,$09,$1C,$E3,$BC,$61,$B1,$97,$8B,$E0,$BF,$1C,$70
	.dc.b	$BF,$97,$E2,$C8,$E6,$FC,$EF,$1B,$11,$E2,$8B,$5F,$22,$6D,$FE,$02
	.dc.b	$6D,$7D,$F9,$B8,$FF,$4B,$9B,$8A,$6C,$61,$B5,$E2,$3D,$AF,$E0,$3C
	.dc.b	$D9,$1F,$01,$FC,$1F,$8D,$88,$F8,$F1,$72,$FB,$B3,$72,$FC,$AC,$DA
	.dc.b	$FC,$A4,$DC,$BF,$B0,$4D,$C9,$36,$30,$DA,$F1,$1F,$49,$FC,$AF,$34
	.dc.b	$8F,$AA,$FE,$63,$C6,$C4,$7D,$C8,$B9,$7E,$4A,$6E,$5F,$AA,$9B,$5F
	.dc.b	$A9,$9B,$97,$F7,$59,$B9,$7F,$B0,$EB,$B9,$0E,$C7,$E1,$AB,$73,$C1
	.dc.b	$B1,$BC,$3C,$4D,$C3,$C4,$D3,$C4,$5C,$3C,$4D,$C3,$C4,$FE,$63,$DB
	.dc.b	$FD,$BF,$C2,$48,$FC,$BC,$7E,$E8,$C3,$F9,$E8,$C3,$A3,$F3,$AE,$B0
	.dc.b	$FD,$BC,$EA,$78,$1A,$38,$1E,$13,$06,$8C,$12,$3F,$80,$CF,$5C,$3F
	.dc.b	$8C,$D7,$F1,$0F,$A7,$FE,$F9,$7F,$BC,$6B,$FB,$76,$DF,$F0,$D7,$FD
	.dc.b	$8D,$89,$CD,$FF,$13,$8C,$1D,$4E,$DE,$2D,$BE,$D1,$5E,$D1,$6F,$B4
	.dc.b	$DA,$E9,$7E,$03,$BD,$FF,$84,$DB,$3E,$08,$F5,$5F,$17,$51,$4F,$44
	.dc.b	$7C,$1F,$12,$4F,$C1,$7E,$27,$8B,$88,$F8,$51,$71,$F8,$53,$71,$F2
	.dc.b	$CD,$AF,$24,$DC,$3E,$34,$DC,$53,$63,$0D,$AF,$11,$F4,$DC,$3E,$23
	.dc.b	$06,$8C,$13,$0F,$2B,$97,$EC,$4F,$68,$FB,$DE,$2F,$E0,$7C,$59,$18
	.dc.b	$21,$F8,$17,$2F,$E0,$E3,$F1,$06,$2B,$F0,$93,$89,$E3,$18,$7F,$19
	.dc.b	$19,$79,$18,$21,$82,$47,$17,$C4,$74,$CB,$F0,$7E,$6F,$E6,$49,$1F
	.dc.b	$99,$F1,$6B,$F9,$98,$B5,$F9,$99,$B5,$CA,$BF,$41,$5F,$A3,$9C,$C7
	.dc.b	$E3,$62,$E9,$F9,$F9,$BA,$26,$C6,$1B,$5F,$AB,$F6,$3F,$5D,$E7,$48
	.dc.b	$C1,$0F,$D7,$39,$7F,$61,$1F,$B5,$31,$97,$EC,$67,$13,$C6,$30,$FE
	.dc.b	$E6,$32,$F2,$30,$43,$04,$8E,$2F,$88,$E9,$97,$EC,$2B,$C2,$FE,$CA
	.dc.b	$BA,$D7,$F0,$39,$CD,$7E,$CE,$73,$5C,$63,$C1,$5D,$8E,$0F,$F8,$2B
	.dc.b	$E9,$1E,$C7,$F8,$89,$78,$94,$C4,$49,$E1,$19,$79,$46,$5F,$F0,$74
	.dc.b	$F2,$76,$68,$86,$A7,$93,$C2,$79,$57,$F8,$AE,$B9,$4E,$CE,$31,$99
	.dc.b	$E2,$EE,$B8,$DF,$F2,$3B,$7A,$BE,$DF,$12,$63,$F0,$63,$E0,$18,$7D
	.dc.b	$31,$87,$47,$C7,$75,$87,$96,$75,$3C,$0D,$1C,$0F,$86,$F0,$BF,$25
	.dc.b	$70,$AE,$55,$C0,$E4,$70,$3E,$B7,$A3,$F5,$48,$FD,$9F,$0E,$BF,$42
	.dc.b	$2C,$FC,$99,$B7,$F0,$E6,$D7,$FB,$2F,$1F,$F4,$54,$E6,$3B,$90,$FB
	.dc.b	$75,$C3,$85,$4B,$C3,$CA,$BA,$46,$0F,$C6,$55,$8E,$98,$79,$5B,$87
	.dc.b	$85,$F5,$1E,$D3,$27,$42,$E1,$E2,$AF,$CC,$78,$5F,$99,$F0,$A5,$7E
	.dc.b	$76,$32,$FE,$7E,$32,$FE,$5A,$32,$FE,$62,$33,$87,$E4,$67,$2F,$E4
	.dc.b	$E3,$38,$7E,$36,$72,$FE,$3E,$33,$5F,$36,$72,$FC,$F8,$CE,$1F,$2A
	.dc.b	$7F,$61,$F1,$1F,$E9,$2F,$8B,$FF,$42,$9F,$E8,$6D,$11,$AC,$3F,$D0
	.dc.b	$33,$AC,$71,$1F,$B0,$9E,$51,$23,$F7,$D1,$AC,$BF,$D2,$D5,$C6,$72
	.dc.b	$43,$FD,$0C,$E5,$E8,$66,$38,$D7,$58,$C9,$C3,$D8,$F5,$11,$F3,$FC
	.dc.b	$A5,$74,$7A,$A3,$E0,$46,$6B,$CB,$16,$BF,$C3,$96,$AC,$5B,$34,$AF
	.dc.b	$31,$8A,$B3,$9A,$F0,$1E,$42,$9E,$42,$61,$F1,$A7,$8D,$F2,$6C,$F0
	.dc.b	$9D,$37,$0F,$3D,$6E,$3E,$BF,$87,$F7,$48,$FD,$AF,$0E,$BF,$42,$2D
	.dc.b	$FD,$69,$B7,$F0,$A6,$D7,$C1,$C7,$FD,$12,$F6,$21,$CC,$EC,$62,$FF
	.dc.b	$03,$3C,$B2,$F9,$38,$5B,$4D,$8F,$CA,$78,$53,$0F,$C8,$BF,$97,$F3
	.dc.b	$B1,$1F,$88,$8B,$8F,$E3,$A6,$E5,$F6,$A6,$D7,$A6,$6E,$5F,$6A,$6E
	.dc.b	$49,$B1,$86,$D7,$88,$FD,$1B,$FA,$9F,$4E,$47,$E9,$9F,$D6,$7A,$78
	.dc.b	$8F,$CE,$C5,$CB,$F4,$73,$72,$FC,$7C,$DA,$F9,$D3,$72,$FC,$7C,$DC
	.dc.b	$93,$63,$0D,$AF,$11,$FB,$87,$0F,$88,$C1,$A3,$04,$C3,$CA,$E5,$FD
	.dc.b	$CC,$F6,$81,$FE,$FD,$B7,$CC,$62,$31,$F0,$56,$D9,$7F,$A1,$ED,$39
	.dc.b	$8C,$5D,$BF,$D7,$57,$47,$BD,$F2,$73,$7B,$9F,$DA,$35,$E1,$1E,$5A
	.dc.b	$C4,$C8,$E7,$7E,$06,$9E,$09,$D3,$EA,$33,$5F,$C0,$63,$B1,$AB,$F2
	.dc.b	$47,$8A,$76,$F8,$A3,$2D,$AF,$85,$7C,$5F,$15,$7C,$BF,$67,$D5,$F0
	.dc.b	$12,$39,$C7,$B4,$C3,$F4,$63,$0E,$8F,$0B,$AC,$3C,$33,$A9,$E0,$68
	.dc.b	$E0,$78,$4C,$1A,$30,$48,$F5,$CF,$5A,$FB,$95,$F4,$7E,$26,$5F,$C2
	.dc.b	$7E,$A7,$C0,$E2,$3F,$11,$17,$0F,$8B,$36,$FF,$0D,$36,$BF,$07,$37
	.dc.b	$2F,$E6,$53,$54,$58,$C3,$67,$C4,$7D,$47,$0F,$88,$C1,$A3,$04,$BF
	.dc.b	$2B,$97,$ED,$CF,$68,$FB,$FE,$07,$E4,$48,$FE,$0B,$C0,$AF,$E9,$E2
	.dc.b	$D7,$E1,$26,$E3,$FA,$59,$B5,$FA,$29,$B8,$FF,$51,$9A,$A2,$C6,$1B
	.dc.b	$3E,$43,$C4,$7E,$6B,$F1,$2F,$F3,$8B,$7C,$0F,$D5,$47,$F9,$83,$1F
	.dc.b	$CD,$BA,$94,$63,$C1,$DD,$49,$E9,$1C,$8C,$3C,$32,$FE,$0F,$93,$0E
	.dc.b	$D1,$87,$B6,$1D,$6B,$C8,$43,$91,$FD,$C6,$37,$3C,$6F,$BD,$7A,$30
	.dc.b	$F1,$5F,$90,$E5,$19,$70,$74,$C3,$C9,$E0,$FC,$2F,$06,$4F,$C7,$8F
	.dc.b	$8A,$61,$E1,$92,$73,$1B,$3E,$73,$BB,$F9,$75,$A9,$E3,$39,$CB,$E1
	.dc.b	$4E,$6B,$E8,$CE,$6B,$E0,$4E,$6B,$E4,$4E,$70,$F4,$D7,$84,$EA,$78
	.dc.b	$8C,$5F,$89,$C1,$A3,$D7,$7D,$5F,$9D,$5D,$70,$FF,$4D,$C7,$E2,$7C
	.dc.b	$0F,$C3,$91,$FC,$67,$A5,$5F,$8B,$17,$0F,$C5,$CD,$CB,$F0,$13,$6B
	.dc.b	$EE,$4D,$CB,$F0,$D3,$70,$4D,$8C,$36,$BC,$87,$88,$FC,$F7,$83,$F9
	.dc.b	$C2,$3F,$A1,$F8,$AA,$FD,$A8,$B9,$7E,$82,$6E,$5F,$95,$9B,$5F,$91
	.dc.b	$9B,$97,$E6,$E6,$E4,$9B,$18,$6D,$79,$0F,$11,$FD,$40,$7D,$F3,$E2
	.dc.b	$FF,$B1,$43,$F2,$6F,$F4,$F3,$75,$CD,$6E,$39,$35,$FD,$39,$B3,$B4
	.dc.b	$F2,$8D,$69,$FD,$AE,$7C,$37,$E0,$AE,$E6,$DE,$E7,$F7,$D8,$95,$DD
	.dc.b	$B1,$97,$FC,$05,$19,$6C,$65,$FF,$04,$C6,$5B,$19,$7C,$66,$E3,$4E
	.dc.b	$0E,$CE,$09,$1E,$68,$F8,$6F,$AF,$E7,$B1,$1F,$12,$7E,$0E,$BE,$48
	.dc.b	$F8,$B3,$AC,$3E,$25,$6A,$7C,$A6,$8F,$F0,$1F,$E2,$6F,$5F,$FC,$24
	.dc.b	$EF,$73,$4D,$4F,$09,$DE,$52,$30,$43,$F8,$A4,$F1,$8E,$B3,$27,$C3
	.dc.b	$1C,$5D,$C6,$5E,$31,$9C,$3F,$C2,$B5,$E2,$9D,$9C,$88,$60,$F0,$FA
	.dc.b	$DE,$CF,$8D,$91,$F0,$23,$DE,$61,$F3,$C6,$1D,$1F,$19,$D5,$78,$A7
	.dc.b	$53,$C0,$D1,$C0,$F0,$98,$34,$60,$91,$EC,$9E,$BA,$7F,$4E,$B5,$FD
	.dc.b	$4F,$83,$FA,$BF,$1B,$31,$FB,$31,$97,$ED,$C6,$5F,$A3,$19,$7E,$9C
	.dc.b	$67,$1F,$97,$39,$7E,$6C,$67,$0F,$14,$E5,$F2,$46,$6B,$FD,$95,$82
	.dc.b	$B5,$18,$9D,$69,$F8,$A9,$CB,$F8,$C8,$CD,$71,$AF,$59,$EA,$21,$EA
	.dc.b	$32,$7A,$8C,$E1,$EA,$7F,$37,$F1,$BF,$9A,$23,$F9,$EF,$5E,$BF,$94
	.dc.b	$8B,$3F,$8C,$9B,$97,$D1,$9B,$5E,$E9,$B8,$FE,$92,$6E,$09,$B1,$86
	.dc.b	$D7,$90,$F1,$1F,$E7,$2F,$C4,$DF,$1D,$FE,$81,$7F,$DE,$91,$9C,$31
	.dc.b	$32,$30,$43,$FC,$01,$3C,$9E,$91,$CE,$64,$60,$87,$07,$2F,$28,$EC
	.dc.b	$E4,$ED,$97,$FA,$12,$BB,$46,$CC,$E5,$FE,$24,$8F,$4C,$64,$F2,$F8
	.dc.b	$7E,$D2,$3F,$07,$D7,$AE,$8F,$89,$1F,$0E,$33,$7E,$18,$B7,$DE,$6D
	.dc.b	$31,$61,$FE,$B4,$8E,$67,$72,$9D,$C9,$87,$B6,$B9,$D7,$2C,$2D,$A6
	.dc.b	$C6,$1B,$7D,$2F,$93,$70,$F8,$95,$B8,$FC,$07,$C7,$7D,$FF,$62,$46
	.dc.b	$08,$7D,$C7,$2F,$DE,$8F,$C3,$18,$BF,$A9,$38,$9E,$31,$87,$F1,$D1
	.dc.b	$97,$91,$82,$18,$24,$71,$7C,$47,$4B,$FB,$DE,$27,$E5,$C9,$1F,$9B
	.dc.b	$F5,$6B,$F9,$48,$B5,$F8,$99,$B5,$CA,$BF,$45,$5F,$A1,$9C,$C7,$DB
	.dc.b	$CF,$FD,$46,$E8,$EC,$53,$B1,$6F,$B3,$6F,$B3,$A3,$F4,$F5,$FA,$FF
	.dc.b	$12,$47,$EC,$DF,$D9,$78,$98,$8F,$D6,$C5,$BF,$D1,$4D,$BF,$D0,$4D
	.dc.b	$AF,$44,$DD,$BF,$27,$37,$1F,$F4,$F5,$77,$21,$D8,$FE,$09,$5B,$9E
	.dc.b	$0D,$8D,$E1,$E2,$6E,$1E,$26,$9E,$22,$E1,$E2,$6E,$1E,$27,$FD,$11
	.dc.b	$D1,$FE,$92,$7F,$54,$FF,$A7,$1B,$E1,$99,$83,$0F,$F5,$1D,$71,$C3
	.dc.b	$FC,$15,$E2,$6E,$FA,$E1,$70,$E1,$1D,$0E,$78,$72,$AD,$36,$32,$6C
	.dc.b	$E6,$62,$FB,$B7,$0F,$11,$E9,$8F,$F0,$E6,$1D,$27,$DB,$76,$78,$C7
	.dc.b	$42,$61,$E6,$8C,$B6,$FA,$DF,$17,$B1,$E0,$F0,$F0,$4B,$F2,$7C,$05
	.dc.b	$7E,$7C,$5B,$E9,$37,$2F,$97,$36,$BE,$9C,$DC,$7F,$C6,$D3,$70,$4D
	.dc.b	$8C,$36,$BE,$24,$79,$E3,$D1,$E0,$7A,$7C,$49,$5F,$26,$32,$FC,$B8
	.dc.b	$CB,$F8,$68,$CB,$F8,$88,$CE,$5F,$16,$72,$FC,$78,$CE,$5F,$86,$9C
	.dc.b	$BF,$88,$8C,$D7,$E4,$27,$2F,$E4,$A3,$39,$7F,$95,$67,$2F,$F9,$5E
	.dc.b	$33,$92,$72,$FF,$AB,$AB,$D4,$7B,$48,$7A,$4C,$9E,$D3,$FD,$35,$F8
	.dc.b	$0B,$C9,$FF,$3D,$36,$3F,$D0,$0D,$7F,$D0,$CD,$7F,$D1,$8D,$7F,$44
	.dc.b	$D7,$F5,$4D,$4D,$24,$78,$09,$1F,$E9,$3F,$8A,$6E,$F3,$7F,$39,$18
	.dc.b	$6C,$66,$7F,$53,$2F,$AE,$FE,$F3,$B7,$EC,$A3,$84,$EA,$B8,$29,$D4
	.dc.b	$61,$E1,$5C,$B6,$FE,$09,$7C,$AF,$84,$62,$F8,$37,$0F,$3D,$F1,$AF
	.dc.b	$E1,$9E,$6B,$D1,$FF,$54,$36,$3F,$D5,$CD,$7F,$D6,$8D,$7F,$D7,$4D
	.dc.b	$7F,$88,$35,$EB,$1D,$08,$F4,$6B,$D1,$FF,$5E,$70,$7F,$A8,$FC,$AF
	.dc.b	$DF,$12,$7F,$87,$79,$55,$FE,$53,$17,$5F,$E0,$D3,$73,$FE,$4B,$36
	.dc.b	$BF,$96,$CD,$BE,$D1,$FC,$66,$7F,$86,$4E,$78,$11,$63,$0D,$9E,$04
	.dc.b	$3C,$67,$98,$C6,$1F,$C6,$63,$84,$68,$91,$97,$81,$F6,$8F,$F1,$67
	.dc.b	$E2,$7F,$EA,$F8,$FE,$48,$DC,$7F,$A2,$62,$8C,$57,$0F,$23,$F9,$74
	.dc.b	$62,$75,$96,$0C,$CD,$AF,$E7,$95,$B2,$46,$63,$C9,$5C,$9D,$D7,$26
	.dc.b	$90,$DD,$72,$71,$87,$AA,$7A,$3E,$CB,$B8,$7B,$2F,$C5,$5D,$6F,$9B
	.dc.b	$A7,$E2,$B9,$BE,$91,$DA,$B9,$4E,$5E,$9E,$3F,$C4,$24,$FC,$0F,$22
	.dc.b	$BE,$88,$B7,$F4,$26,$D7,$2A,$F8,$F5,$F3,$A7,$31,$F6,$22,$E5,$F5
	.dc.b	$66,$E3,$FE,$CC,$AE,$27,$84,$A7,$84,$B8,$78,$5B,$87,$85,$D1,$F3
	.dc.b	$AB,$AE,$1D,$9B,$1C,$0F,$C5,$BF,$84,$F2,$24,$7E,$01,$FC,$3F,$91
	.dc.b	$88,$FB,$31,$70,$FC,$8C,$DC,$3E,$B4,$DA,$FC,$BC,$DC,$BF,$29,$37
	.dc.b	$24,$D8,$C3,$6B,$C4,$7E,$49,$C3,$E2,$30,$68,$C1,$30,$F2,$B9,$7F
	.dc.b	$4F,$3D,$A3,$F3,$BE,$97,$E9,$08,$FE,$83,$C8,$AF,$E6,$22,$D7,$EC
	.dc.b	$26,$E5,$F9,$59,$B5,$FB,$69,$B9,$7E,$CE,$6E,$49,$B1,$86,$D7,$90
	.dc.b	$F1,$14,$F1,$17,$0F,$13,$70,$F1,$34,$F1,$17,$0F,$13,$70,$F1,$3F
	.dc.b	$E3,$9F,$C4,$FF,$A1,$49,$FF,$64,$DF,$F8,$C5,$B7,$FC,$39,$D9,$27
	.dc.b	$FC,$74,$EA,$BF,$D8,$4E,$BC,$8C,$CC,$9F,$04,$70,$30,$FF,$0C,$9E
	.dc.b	$53,$E3,$9D,$4F,$9E,$37,$7E,$59,$CC,$F1,$9C,$DF,$9E,$73,$5E,$B9
	.dc.b	$CD,$79,$A7,$35,$ED,$9C,$9E,$19,$F3,$9E,$12,$1E,$79,$F8,$3E,$3C
	.dc.b	$8F,$98,$FC,$4F,$23,$11,$ED,$8B,$87,$CD,$9B,$87,$CD,$9B,$5F,$2A
	.dc.b	$6E,$1F,$56,$6E,$3F,$EE,$3A,$EE,$43,$B1,$F7,$6B,$73,$C1,$B1,$BC
	.dc.b	$3C,$4D,$C3,$C4,$D3,$C4,$5C,$3C,$4D,$C3,$C4,$FD,$9F,$1F,$F0,$64
	.dc.b	$7F,$1D,$EF,$57,$47,$E4,$63,$F1,$31,$9C,$3F,$21,$17,$0F,$C8,$4D
	.dc.b	$AF,$C6,$CD,$C3,$F3,$13,$70,$4D,$8C,$36,$BC,$04,$3C,$67,$E4,$63
	.dc.b	$6E,$70,$FD,$14,$62,$37,$18,$7F,$55,$F1,$5F,$A2,$F9,$19,$18,$21
	.dc.b	$F9,$D7,$2F,$E7,$E3,$F5,$26,$27,$F6,$13,$89,$E3,$18,$7F,$71,$19
	.dc.b	$79,$18,$21,$82,$47,$17,$C4,$74,$CB,$F7,$95,$E1,$7F,$7D,$5D,$6B
	.dc.b	$F8,$14,$E6,$BF,$7B,$39,$AE,$31,$E0,$AE,$C7,$07,$FD,$8D,$F8,$9B
	.dc.b	$E4,$FF,$D8,$0F,$FB,$5C,$8E,$0C,$F7,$C9,$E3,$19,$79,$C6,$5F,$F0
	.dc.b	$9C,$F5,$76,$68,$86,$A7,$93,$C2,$79,$57,$FB,$3E,$B9,$4E,$CE,$31
	.dc.b	$99,$E2,$EE,$B8,$DF,$F8,$CA,$FD,$77,$E2,$3B,$1C,$A3,$53,$C9,$E1
	.dc.b	$7C,$A5,$AA,$FC,$78,$79,$7D,$4F,$93,$E3,$CB,$F9,$B1,$F3,$0C,$3F
	.dc.b	$12,$30,$E8,$FA,$6E,$A7,$A4,$64,$E0,$68,$C3,$A7,$24,$70,$4C,$7E
	.dc.b	$14,$F5,$7E,$25,$78,$AB,$84,$69,$EA,$E9,$F0,$1A,$3E,$1D,$A7,$95
	.dc.b	$FD,$7B,$ED,$5A,$9F,$C6,$FC,$07,$E3,$BC,$69,$3F,$92,$8C,$BF,$94
	.dc.b	$8C,$BF,$8A,$8C,$BF,$8C,$8C,$DF,$E1,$67,$2F,$E1,$E3,$38,$FE,$06
	.dc.b	$72,$FE,$0E,$33,$5E,$A9,$CB,$EC,$8C,$E5,$F7,$E7,$2F,$E0,$63,$39
	.dc.b	$27,$2F,$FB,$BE,$BD,$47,$B4,$87,$A4,$C9,$ED,$3F,$67,$F0,$1F,$B2
	.dc.b	$EF,$91,$82,$1F,$AC,$72,$FE,$BA,$3F,$72,$63,$2F,$D4,$4E,$27,$8C
	.dc.b	$61,$FE,$03,$19,$79,$18,$21,$82,$47,$17,$C4,$74,$CB,$F5,$D5,$E1
	.dc.b	$7F,$61,$5D,$6B,$F7,$53,$9A,$FC,$4C,$E6,$B8,$C7,$82,$BB,$1C,$11
	.dc.b	$FE,$B1,$1F,$7E,$F9,$4F,$F7,$FF,$E7,$5F,$EA,$EF,$20,$FF,$AC,$2B
	.dc.b	$73,$C9,$E5,$5C,$5E,$B7,$FF,$00,$6B,$91,$C6,$75,$3C,$2B,$BF,$85
	.dc.b	$23,$8E,$3D,$6B,$78,$23,$17,$E1,$A9,$5D,$1B,$19,$76,$53,$D2,$58
	.dc.b	$E0,$E8,$E0,$E1,$F8,$ED,$78,$9A,$8E,$8E,$0F,$09,$EA,$38,$BF,$1E
	.dc.b	$BD,$51,$83,$58,$FC,$EF,$80,$D6,$1E,$3A,$DD,$78,$1B,$87,$C4,$9F
	.dc.b	$B7,$7C,$6B,$C0,$6E,$39,$11,$46,$DF,$F6,$FF,$27,$6B,$F2,$F0,$FF
	.dc.b	$5B,$BF,$1D,$7F,$38,$AD,$47,$FB,$D1,$FF,$78,$18,$8D,$47,$09,$E5
	.dc.b	$1F,$0D,$FE,$CD,$9E,$6B,$84,$6A,$7C,$75,$9A,$D3,$23,$95,$70,$3C
	.dc.b	$31,$98,$E2,$6A,$71,$7E,$CA,$E1,$5D,$6B,$9B,$A3,$D8,$65,$E6,$49
	.dc.b	$E5,$19,$6C,$72,$21,$F3,$2B,$DD,$F1,$72,$3D,$2F,$C1,$F1,$31,$1F
	.dc.b	$4A,$2D,$79,$E6,$DF,$CF,$9B,$5F,$4E,$6E,$3F,$E1,$19,$B8,$FF,$BC
	.dc.b	$6B,$A1,$E0,$21,$F8,$0A,$DC,$F0,$6C,$6B,$0F,$13,$70,$F1,$34,$F1
	.dc.b	$17,$0F,$13,$70,$F1,$57,$D1,$F8,$0F,$C0,$12,$7E,$DF,$87,$5F,$C9
	.dc.b	$C5,$9F,$B1,$37,$0F,$C8,$4D,$AF,$CA,$4D,$C3,$FC,$65,$37,$04,$D8
	.dc.b	$C3,$6B,$F1,$71,$F9,$38,$C4,$78,$8F,$29,$A2,$19,$3F,$52,$F0,$C3
	.dc.b	$DD,$1C,$0F,$D6,$F8,$7F,$9F,$23,$F9,$AF,$0E,$BA,$3F,$3B,$1F,$A5
	.dc.b	$8C,$CF,$E6,$62,$E5,$FB,$19,$B5,$FB,$59,$B9,$7F,$94,$E6,$E4,$9B
	.dc.b	$18,$6D,$78,$08,$78,$CF,$E0,$B1,$B7,$38,$7E,$AE,$31,$1B,$8C,$3F
	.dc.b	$B6,$F2,$FF,$7C,$47,$F8,$17,$E2,$9F,$EF,$4C,$13,$99,$FE,$26,$FF
	.dc.b	$09,$9D,$39,$AF,$DD,$39,$F1,$3F,$8A,$D7,$F5,$AF,$82,$FE,$67,$C6
	.dc.b	$69,$23,$FA,$03,$5F,$E7,$ED,$7F,$A5,$B5,$FE,$A4,$D7,$FC,$B3,$83
	.dc.b	$E4,$FC,$C4,$8F,$6F,$9B,$56,$5F,$CB,$2B,$84,$7C,$2A,$F7,$56,$EB
	.dc.b	$51,$77,$FE,$EF,$BB,$05,$FC,$0C,$9C,$3F,$00,$99,$7F,$5D,$6B,$FD
	.dc.b	$71,$AF,$F6,$86,$BF,$DB,$DA,$FF,$60,$7F,$B4,$5F,$87,$E2,$7E,$A1
	.dc.b	$27,$ED,$FC,$15,$7D,$F9,$7A,$32,$56,$74,$F8,$57,$6B,$53,$6F,$E6
	.dc.b	$CF,$DC,$F8,$5F,$8F,$1F,$DE,$DA,$FF,$80,$5A,$FF,$84,$DA,$FF,$87
	.dc.b	$9A,$9A,$4D,$7F,$BC,$B6,$3F,$C0,$CD,$7F,$C1,$D9,$7D,$1F,$1F,$F2
	.dc.b	$84,$8F,$CD,$F8,$F5,$F8,$27,$E1,$A3,$19,$FE,$06,$6E,$7F,$16,$6D
	.dc.b	$7E,$32,$6D,$F7,$8F,$D3,$CF,$E8,$E7,$3A,$22,$C6,$1B,$3C,$08,$79
	.dc.b	$2B,$A3,$69,$5C,$63,$78,$79,$DB,$87,$9D,$A7,$9C,$FD,$DF,$C4,$7E
	.dc.b	$DC,$8F,$EF,$BC,$8A,$F7,$8B,$5F,$A3,$9B,$5F,$87,$9B,$5F,$AE,$9B
	.dc.b	$8F,$E8,$66,$E2,$9B,$18,$6D,$79,$0F,$11,$4F,$11,$70,$F1,$37,$0F
	.dc.b	$13,$4F,$11,$70,$F1,$37,$0F,$13,$FE,$1D,$FC,$53,$FC,$A9,$7F,$E2
	.dc.b	$A5,$F1,$BF,$E2,$36,$CE,$CC,$19,$88,$E4,$87,$D9,$8E,$2D,$8E,$2E
	.dc.b	$A7,$C1,$1C,$A3,$81,$A3,$81,$E7,$71,$1E,$78,$E0,$F2,$7A,$61,$FC
	.dc.b	$4A,$BD,$AD,$AD,$B9,$9E,$6E,$5E,$C7,$23,$FE,$00,$CF,$26,$CF,$27
	.dc.b	$27,$C7,$F9,$0F,$91,$F6,$52,$3E,$74,$65,$FA,$11,$97,$E3,$46,$5F
	.dc.b	$91,$19,$C3,$EA,$4E,$5F,$AD,$19,$C3,$DD,$39,$7E,$04,$66,$BD,$33
	.dc.b	$97,$D7,$19,$C9,$8A,$B5,$18,$9D,$4E,$6B,$D6,$7A,$88,$7A,$8C,$9E
	.dc.b	$A3,$38,$7A,$9F,$F4,$4D,$BD,$2F,$F3,$BF,$83,$E1,$6C,$66,$77,$E1
	.dc.b	$E2,$64,$72,$77,$18,$8F,$E3,$75,$CD,$D1,$C8,$A6,$AB,$B3,$6B,$9B
	.dc.b	$AB,$FF,$50,$D7,$38,$D1,$CC,$A4,$8D,$D7,$46,$DF,$95,$CD,$DC,$FF
	.dc.b	$D4,$75,$FE,$A5,$BF,$F4,$FB,$C6,$33,$3B,$9C,$DD,$C3,$C9,$87,$49
	.dc.b	$F2,$3F,$BB,$F2,$3F,$74,$48,$FE,$01,$E4,$D7,$F6,$D1,$67,$F8,$14
	.dc.b	$DC,$7F,$55,$36,$BF,$47,$37,$5F,$C4,$CD,$D1,$36,$30,$DA,$FE,$21
	.dc.b	$1F,$C3,$63,$11,$E2,$3C,$A6,$88,$64,$FE,$18,$F0,$C3,$DD,$1C,$0F
	.dc.b	$E2,$1E,$3F,$F1,$9F,$52,$47,$F1,$F8,$FE,$38,$61,$FE,$2D,$18,$74
	.dc.b	$7F,$25,$75,$3F,$C9,$67,$53,$C0,$D1,$C0,$F0,$98,$34,$60,$91,$FC
	.dc.b	$BE,$7A,$E5,$FC,$5E,$BA,$CE,$6B,$93,$E1,$7B,$39,$7A,$B9,$7A,$CF
	.dc.b	$F2,$1A,$EB,$1F,$C6,$AB,$FD,$BD,$F8,$A3,$D5,$BE,$4C,$AF,$F7,$04
	.dc.b	$E5,$B3,$97,$FB,$A4,$27,$9C,$C9,$E3,$1B,$C7,$FD,$EB,$5C,$E7,$8D
	.dc.b	$67,$1F,$F2,$3C,$F3,$9E,$E4,$3B,$DE,$D6,$A9,$E9,$87,$6C,$35,$39
	.dc.b	$BF,$25,$F8,$0F,$29,$0C,$12,$BC,$F1,$CA,$FA,$E1,$E6,$C3,$13,$D2
	.dc.b	$BD,$BB,$FC,$C2,$61,$F2,$BC,$0A,$FC,$88,$B3,$F1,$26,$E1,$F0,$A6
	.dc.b	$D7,$BB,$1F,$F4,$23,$F5,$23,$39,$7C,$17,$ED,$DE,$2B,$F8,$54,$E6
	.dc.b	$BC,$47,$12,$1E,$33,$25,$26,$1E,$C7,$2D,$73,$87,$9D,$CB,$5C,$9E
	.dc.b	$9C,$3C,$AF,$E3,$3D,$5F,$C8,$FC,$8C,$8C,$10,$FC,$6B,$97,$F1,$F1
	.dc.b	$F9,$83,$17,$F8,$C9,$C4,$F1,$8C,$3F,$9A,$8C,$BC,$8C,$10,$C1,$23
	.dc.b	$8B,$E2,$3A,$5F,$E5,$AB,$C2,$FE,$62,$BA,$D7,$E7,$A7,$35,$F6,$67
	.dc.b	$35,$C6,$3C,$15,$D8,$E0,$FE,$C3,$E1,$3F,$6B,$F2,$32,$30,$43,$F6
	.dc.b	$2E,$5F,$D9,$C7,$EF,$0C,$4F,$EC,$27,$13,$C6,$30,$FE,$FA,$32,$F2
	.dc.b	$30,$43,$04,$8E,$2F,$88,$E9,$5F,$BA,$AF,$0B,$FB,$CA,$EB,$5F,$C0
	.dc.b	$A7,$35,$F9,$D9,$CD,$71,$8F,$05,$76,$38,$23,$FD,$09,$F2,$3F,$C0
	.dc.b	$9C,$42,$7F,$DA,$32,$38,$4F,$06,$CE,$E3,$F8,$5C,$73,$7A,$E7,$AA
	.dc.b	$E7,$36,$76,$D9,$EC,$57,$95,$75,$8F,$14,$64,$D3,$27,$D3,$5D,$9A
	.dc.b	$76,$2C,$76,$6C,$F6,$9E,$11,$ED,$33,$3D,$5C,$12,$BC,$95,$C6,$3E
	.dc.b	$0C,$F9,$5E,$B1,$C6,$77,$39,$9E,$55,$F3,$A3,$4F,$8A,$7A,$D6,$2B
	.dc.b	$E6,$FA,$FF,$17,$C3,$91,$EA,$8F,$90,$61,$FB,$B1,$87,$47,$C0,$75
	.dc.b	$5E,$A9,$D4,$F0,$34,$70,$3F,$0E,$F0,$BF,$C2,$D7,$0A,$E5,$5C,$0E
	.dc.b	$47,$03,$93,$F2,$23,$94,$7B,$0C,$CF,$47,$94,$F4,$79,$1D,$1E,$4F
	.dc.b	$03,$F3,$9F,$81,$BE,$03,$F1,$D0,$F5,$7F,$CE,$0F,$FB,$D8,$C7,$66
	.dc.b	$AB,$84,$7E,$20,$38,$78,$4F,$2B,$FD,$04,$72,$38,$4E,$E3,$2D,$C3
	.dc.b	$F5,$35,$C5,$B7,$E1,$C2,$61,$FA,$E3,$F5,$F7,$E1,$70,$D7,$31,$B6
	.dc.b	$94,$F4,$B5,$A6,$63,$93,$F2,$23,$51,$88,$D9,$D5,$C1,$EB,$78,$1F
	.dc.b	$02,$3E,$7D,$BD,$7F,$5C,$FC,$FA,$E0,$F4,$C7,$E4,$D6,$E3,$91,$0E
	.dc.b	$17,$D2,$FF,$82,$78,$9F,$C1,$7C,$49,$5F,$BA,$8C,$BF,$BC,$8C,$BF
	.dc.b	$C7,$A3,$2F,$F2,$08,$CE,$1F,$BA,$9C,$BF,$BC,$8C,$E7,$FC,$8A,$72
	.dc.b	$FF,$24,$8C,$D7,$F2,$09,$CB,$FC,$8A,$33,$BF,$F2,$99,$CB,$FC,$AE
	.dc.b	$33,$BA,$72,$A3,$31,$86,$D7,$F1,$8F,$13,$F9,$37,$89,$23,$04,$3F
	.dc.b	$9E,$B9,$7F,$9F,$C7,$F2,$D3,$19,$7F,$1A,$9C,$4F,$18,$C3,$FC,$9A
	.dc.b	$32,$F2,$30,$43,$04,$8E,$2F,$88,$E9,$97,$F4,$CA,$F0,$BF,$D3,$AB
	.dc.b	$AD,$7F,$2C,$9C,$D7,$F4,$C9,$CD,$71,$8F,$05,$76,$38,$3F,$CF,$BF
	.dc.b	$14,$FD,$54,$7F,$49,$5E,$27,$F4,$F6,$CE,$CC,$19,$88,$E4,$82,$7A
	.dc.b	$46,$A3,$15,$FD,$B7,$45,$F8,$A3,$0D,$9F,$F3,$8B,$E2,$78,$98,$AF
	.dc.b	$1B,$E0,$BE,$EF,$B6,$32,$F0,$BE,$EF,$8E,$76,$E6,$BC,$13,$A8,$E4
	.dc.b	$43,$47,$49,$E2,$4B,$F4,$C6,$5F,$5D,$F7,$73,$1F,$47,$E0,$BE,$77
	.dc.b	$14,$8C,$10,$F2,$39,$7C,$D1,$F5,$4C,$5F,$D9,$9C,$4F,$18,$C3,$F7
	.dc.b	$E3,$2F,$23,$04,$30,$48,$E2,$F8,$8E,$95,$F5,$AB,$C2,$FD,$8A,$EB
	.dc.b	$5F,$8B,$9C,$D7,$DA,$9C,$D7,$18,$F0,$57,$63,$83,$F9,$5E,$0F,$C9
	.dc.b	$11,$FC,$C7,$CA,$D7,$EC,$45,$9F,$CD,$4D,$AF,$C5,$CD,$AF,$C5,$CF
	.dc.b	$F6,$0F,$C1,$1E,$BF,$F4,$16,$CF,$F4,$56,$BF,$D2,$5A,$FF,$51,$6B
	.dc.b	$FE,$40,$6B,$FE,$49,$6A,$69,$23,$C0,$48,$FE,$AD,$A7,$E9,$3C,$3F
	.dc.b	$D9,$7C,$24,$9F,$DC,$47,$ED,$4C,$3F,$99,$8C,$3A,$3F,$7C,$EB,$3F
	.dc.b	$DF,$4E,$A7,$81,$A3,$81,$FA,$37,$86,$7F,$B4,$AE,$15,$CA,$B8,$1C
	.dc.b	$8E,$07,$F0,$DF,$0F,$F8,$51,$1F,$E2,$1E,$25,$7F,$61,$16,$7F,$89
	.dc.b	$4D,$BF,$E0,$73,$6B,$F8,$1C,$FD,$9F,$13,$EC,$12,$3E,$C3,$5F,$B0
	.dc.b	$D7,$EC,$35,$FB,$0D,$7E,$C3,$FE,$05,$C9,$F8,$A7,$F8,$6B,$1F,$F3
	.dc.b	$73,$71,$FE,$34,$DB,$57,$0D,$3F,$CF,$13,$98,$E7,$58,$33,$36,$BF
	.dc.b	$CE,$35,$B2,$47,$E5,$FC,$5F,$CB,$12,$7F,$2C,$D7,$F2,$CD,$7F,$2C
	.dc.b	$D7,$F2,$CD,$7F,$2C,$FB,$E5,$E6,$7A,$09,$3E,$4F,$66,$BF,$CA,$E7
	.dc.b	$6F,$A2,$7D,$F3,$AC,$BE,$3C,$5C,$7E,$3C,$DA,$F8,$F7,$C9,$E9,$5C
	.dc.b	$9E,$84,$32,$78,$CC,$5F,$66,$DF,$66,$9D,$A7,$C1,$97,$DC,$C2,$E1
	.dc.b	$E1,$C3,$95,$FB,$67,$37,$A7,$C9,$1E,$22,$4F,$E1,$23,$2F,$22,$61
	.dc.b	$E3,$9E,$87,$93,$0F,$C6,$DF,$09,$D5,$70,$AF,$41,$EF,$38,$9E,$08
	.dc.b	$FC,$8C,$F1,$8F,$97,$7E,$87,$85,$79,$5F,$06,$1E,$FB,$EE,$48,$EE
	.dc.b	$75,$39,$9C,$9E,$EF,$2A,$F0,$61,$EA,$8C,$1B,$39,$DF,$AE,$3E,$84
	.dc.b	$65,$E0,$F7,$72,$FB,$A3,$85,$F9,$E3,$C6,$66,$FC,$6F,$18,$5F,$91
	.dc.b	$E1,$7D,$AF,$E1,$57,$13,$CB,$7C,$A3,$24,$70,$48,$F0,$C6,$5E,$24
	.dc.b	$9F,$4C,$65,$B3,$F3,$5F,$E1,$7F,$05,$E5,$C3,$CC,$DB,$F8,$98,$23
	.dc.b	$11,$E7,$AE,$2E,$AB,$8B,$4E,$33,$C3,$0F,$15,$F3,$9D,$1C,$DC,$36
	.dc.b	$B9,$B4,$E6,$70,$9E,$75,$D7,$0F,$65,$F2,$9E,$F7,$C9,$B7,$C8,$E8
	.dc.b	$78,$6F,$93,$F3,$2F,$F9,$5C,$63,$0F,$93,$3F,$C1,$A3,$18,$7D,$E8
	.dc.b	$FE,$4F,$18,$C3,$F9,$7B,$63,$27,$A6,$FC,$F1,$98,$FE,$34,$DC,$3F
	.dc.b	$9E,$B6,$7C,$73,$C4,$F3,$1E,$33,$18,$79,$8F,$AA,$FB,$AB,$E3,$BE
	.dc.b	$23,$13,$D1,$F0,$9C,$B0,$F9,$51,$99,$F9,$33,$A2,$38,$3C,$EF,$57
	.dc.b	$11,$C4,$91,$E8,$AF,$35,$76,$8F,$51,$E4,$2D,$74,$71,$3E,$C8,$C1
	.dc.b	$DA,$37,$5D,$A7,$AE,$1F,$4A,$7C,$93,$F8,$68,$C4,$7A,$A3,$0E,$8E
	.dc.b	$67,$9E,$FE,$F3,$B3,$35,$D9,$E0,$F9,$67,$9C,$6E,$30,$F3,$AF,$0D
	.dc.b	$FA,$AB,$6D,$9E,$B7,$E0,$8F,$B3,$1C,$E3,$13,$F8,$A8,$F0,$C6,$1F
	.dc.b	$46,$1D,$A7,$A9,$F1,$67,$E8,$61,$E9,$AE,$95,$F9,$C3,$83,$E7,$3A
	.dc.b	$B9,$7B,$E1,$FA,$DB,$E8,$D8,$7C,$B3,$F3,$B7,$86,$99,$9B,$0C,$5E
	.dc.b	$EB,$55,$FE,$AE,$8C,$3C,$27,$FC,$BD,$A6,$AF,$94,$F8,$30,$EE,$EC
	.dc.b	$C6,$1B,$BE,$F3,$A9,$FF,$5D,$9F,$E6,$5B,$D3,$C0,$E5,$87,$89,$F2
	.dc.b	$4F,$1C,$3C,$31,$C4,$DC,$66,$FB,$E1,$D2,$F9,$5F,$0E,$4F,$E9,$91
	.dc.b	$8A,$DE,$1C,$AF,$FD,$0D,$8F,$1C,$7A,$4F,$94,$F1,$CE,$9E,$04,$32
	.dc.b	$7C,$33,$CC,$60,$98,$FC,$AC,$3C,$51,$FE,$89,$BF,$05,$48,$E9,$5E
	.dc.b	$99,$F0,$47,$2C,$3D,$2F,$CC,$C7,$FB,$43,$E0,$AE,$D8,$FA,$B1,$EF
	.dc.b	$39,$BF,$BA,$F6,$79,$D7,$78,$F5,$5F,$C8,$BE,$47,$C4,$8E,$47,$A4
	.dc.b	$E4,$7C,$79,$E4,$F9,$70,$F6,$3E,$93,$81,$93,$8E,$1E,$6B,$F5,$9E
	.dc.b	$1C,$78,$DF,$42,$63,$E8,$25,$F4,$8C,$BE,$38,$CB,$E8,$3D,$37,$F4
	.dc.b	$8F,$06,$1F,$3A,$BE,$4D,$F8,$2F,$E3,$DF,$9C,$F5,$5F,$07,$E3,$DE
	.dc.b	$DF,$01,$E3,$9E,$37,$D2,$3E,$39,$E3,$C3,$DC,$E1,$F8,$35,$CE,$FB
	.dc.b	$4F,$86,$B9,$3F,$2C,$E5,$3B,$C7,$EB,$5F,$29,$F9,$B1,$89,$F7,$61
	.dc.b	$F2,$30,$EB,$87,$C8,$7E,$25,$6E,$FD,$A7,$2C,$78,$BE,$D3,$DF,$3C
	.dc.b	$AF,$DE,$75,$AE,$53,$E9,$8C,$57,$28,$FA,$73,$F8,$E8,$C6,$3F,$12
	.dc.b	$7D,$CF,$AD,$FA,$33,$D3,$0F,$43,$C6,$3C,$E7,$D7,$3F,$DF,$58,$FF
	.dc.b	$C0,$12,$3F,$F0,$05,$AF,$FC,$01,$6B,$C0,$D1,$0D,$16,$7F,$1E,$FA
	.dc.b	$BC,$3F,$4C,$FC,$6C,$BF,$9B,$46,$27,$CB,$5F,$B6,$8C,$47,$C5,$3F
	.dc.b	$97,$38,$8F,$6B,$72,$F9,$B3,$A3,$9B,$87,$C4,$60,$FA,$C6,$0E,$C7
	.dc.b	$E2,$0F,$E5,$71,$D5,$FB,$95,$E8,$79,$1B,$3F,$9A,$C6,$DE,$95,$F0
	.dc.b	$B1,$FA,$CF,$B6,$38,$C6,$2F,$FA,$2D,$71,$8F,$E1,$CD,$C3,$F9,$93
	.dc.b	$6B,$CA,$72,$8F,$99,$8A,$75,$3E,$43,$47,$C1,$C7,$D4,$FC,$0B,$F1
	.dc.b	$3F,$24,$C5,$F4,$AE,$78,$FA,$67,$27,$3C,$3C,$0F,$63,$EC,$61,$DA
	.dc.b	$7E,$15,$F3,$9D,$47,$DA,$6E,$3F,$6A,$B8,$3F,$67,$1F,$2C,$62,$7E
	.dc.b	$CC,$63,$0F,$0C,$79,$0F,$0B,$8B,$F0,$C7,$9F,$1F,$5E,$1F,$42,$3C
	.dc.b	$27,$DE,$AE,$11,$E1,$C7,$D1,$3D,$6F,$D0,$D9,$F4,$57,$08,$F2,$D7
	.dc.b	$0A,$F0,$C7,$92,$FF,$BF,$EA,$E1,$7B,$D8,$6D,$7F,$82,$E7,$53,$9E
	.dc.b	$4D,$32,$75,$5C,$16,$1F,$E1,$6B,$ED,$7B,$9C,$E7,$FE,$25,$9E,$B1
	.dc.b	$CB,$0E,$CF,$F6,$0D,$FF,$C4,$AC,$D7,$FC,$AA,$D8,$CB,$FE,$2A,$8C
	.dc.b	$B6,$32,$FF,$8C,$E3,$2D,$8C,$A8,$CB,$63,$31,$82,$3D,$2B,$E5,$7C
	.dc.b	$B7,$C9,$25,$7C,$9F,$C4,$BF,$C4,$F9,$FC,$3D,$15,$AA,$F8,$62,$F9
	.dc.b	$BC,$73,$D5,$F1,$AB,$1F,$27,$0E,$AF,$C1,$F7,$FC,$E4,$BF,$83,$9A
	.dc.b	$31,$1F,$22,$30,$EA,$32,$FC,$F6,$BD,$4D,$1F,$43,$15,$B9,$7F,$1A
	.dc.b	$D9,$F8,$51,$8E,$0C,$B2,$32,$7C,$78,$C3,$F1,$2B,$AC,$E3,$3F,$C6
	.dc.b	$5F,$F8,$53,$DB,$FA,$A4,$BF,$7B,$5F,$73,$5F,$C5,$35,$FC,$43,$5F
	.dc.b	$C3,$3F,$76,$1E,$9F,$B0,$93,$F7,$7C,$EA,$F9,$E7,$6F,$E0,$67,$E1
	.dc.b	$CE,$AF,$D3,$17,$5F,$B5,$36,$BF,$05,$7C,$9E,$95,$C9,$E8,$43,$27
	.dc.b	$8C,$C5,$F6,$6D,$F6,$69,$DA,$7C,$19,$7D,$AC,$2E,$1E,$1C,$39,$5F
	.dc.b	$B6,$73,$7A,$6C,$74,$BE,$4D,$BE,$58,$7E,$5E,$FD,$53,$E8,$26,$1D
	.dc.b	$8F,$BB,$87,$48,$CD,$7A,$5A,$7A,$4E,$D1,$B9,$E5,$3B,$6C,$EC,$E3
	.dc.b	$87,$99,$B8,$79,$9D,$4F,$89,$F5,$61,$EF,$69,$E0,$29,$CB,$0F,$16
	.dc.b	$1E,$F8,$FA,$87,$69,$E4,$D3,$91,$4E,$47,$4A,$F4,$E1,$E7,$6E,$1E
	.dc.b	$78,$FD,$8D,$7C,$C9,$CB,$D9,$B5,$D9,$AF,$63,$C0,$F4,$C3,$D3,$1F
	.dc.b	$0B,$0F,$94,$7A,$DD,$B8,$74,$4B,$F1,$BF,$02,$BB,$9D,$5F,$64,$F3
	.dc.b	$8F,$9D,$1D,$0E,$58,$2F,$C7,$7D,$27,$F2,$47,$23,$71,$A9,$F8,$87
	.dc.b	$18,$D4,$F1,$AE,$E7,$92,$FC,$47,$AE,$34,$E2,$7A,$9C,$AB,$AD,$F7
	.dc.b	$C3,$E8,$56,$EF,$8D,$F0,$3D,$8E,$0F,$B5,$1B,$BF,$3D,$72,$AF,$82
	.dc.b	$FD,$99,$D3,$63,$27,$DB,$8C,$B4,$C9,$F7,$70,$F7,$3E,$36,$DF,$8E
	.dc.b	$FE,$34,$F9,$9F,$6B,$EF,$BF,$24,$ED,$F1,$36,$3C,$0D,$7C,$17,$E4
	.dc.b	$69,$E4,$BF,$C1,$3D,$EB,$AD,$FC,$28,$E6,$D7,$8D,$F5,$AE,$27,$AA
	.dc.b	$7C,$F3,$CC,$FC,$85,$F8,$67,$6D,$3E,$86,$1F,$66,$76,$79,$63,$2F
	.dc.b	$C1,$33,$7E,$18,$CC,$77,$3F,$26,$5A,$FC,$6C,$77,$8F,$A0,$D9,$FA
	.dc.b	$15,$E1,$9E,$35,$EB,$C3,$E1,$9F,$9B,$9D,$47,$49,$0E,$F8,$E2,$18
	.dc.b	$6A,$3F,$C0,$1F,$07,$67,$31,$CB,$64,$E6,$B8,$5F,$4A,$FF,$0A,$9D
	.dc.b	$DD,$3C,$66,$BD,$8F,$14,$F2,$9C,$9C,$EF,$C3,$3B,$3A,$DF,$89,$E2
	.dc.b	$7B,$A7,$85,$73,$8C,$9C,$6B,$9D,$71,$8F,$3D,$70,$79,$3E,$68,$D4
	.dc.b	$F0,$BF,$2C,$CA,$EB,$5A,$BF,$94,$D8,$E6,$78,$A7,$B3,$E3,$9D,$D7
	.dc.b	$38,$F1,$CF,$86,$24,$71,$8D,$5F,$8E,$3C,$33,$8A,$F8,$11,$F1,$A3
	.dc.b	$8B,$D0,$E4,$70,$21,$92,$4F,$57,$91,$C0,$CC,$FA,$1E,$31,$F5,$AB
	.dc.b	$DA,$7B,$AF,$E1,$4F,$BC,$F8,$67,$23,$F0,$87,$2A,$F4,$D7,$3A,$F4
	.dc.b	$BE,$33,$13,$E3,$71,$3E,$38,$E8,$7E,$44,$E8,$76,$33,$3E,$5B,$EB
	.dc.b	$3E,$02,$4E,$EB,$AB,$E8,$7C,$2F,$C4,$23,$82,$47,$8D,$E3,$1E,$33
	.dc.b	$CB,$7F,$10,$F8,$B7,$F2,$0F,$1C,$78,$6B,$C6,$F1,$AE,$75,$C2,$BF
	.dc.b	$01,$3B,$21,$BA,$E3,$3D,$EF,$73,$93,$71,$CD,$F9,$93,$E9,$23,$F0
	.dc.b	$5E,$73,$E6,$8E,$44,$9E,$CF,$23,$81,$DA,$3C,$35,$F4,$63,$EA,$9C
	.dc.b	$9C,$3F,$32,$BE,$0C,$62,$FE,$5B,$F1,$E7,$8D,$FA,$5F,$9B,$7E,$59
	.dc.b	$D4,$73,$3E,$2B,$8A,$F9,$73,$E4,$8F,$F5,$E9,$FC,$6E,$7F,$D9,$6D
	.dc.b	$7F,$5A,$61,$D9,$1F,$F6,$4B,$B7,$F3,$E6,$CC,$BB,$7F,$DA,$6F,$8E
	.dc.b	$FE,$BF,$8B,$E3,$9F,$94,$7F,$AC,$DC,$3F,$8A,$6D,$FE,$42,$3F,$9B
	.dc.b	$46,$27,$9C,$72,$AE,$CE,$8C,$3F,$84,$C1,$5C,$67,$EC,$9E,$13,$67
	.dc.b	$92,$36,$E0,$F2,$16,$FC,$8F,$13,$F6,$35,$DE,$BB,$47,$E2,$5B,$1F
	.dc.b	$C2,$5A,$FF,$33,$6C,$7F,$30,$6B,$FD,$11,$B3,$FC,$F9,$B7,$FC,$D1
	.dc.b	$B3,$F0,$4F,$19,$EB,$72,$F9,$E7,$D3,$5F,$22,$31,$5F,$94,$8F,$54
	.dc.b	$7B,$E3,$E7,$BD,$0F,$86,$59,$F8,$77,$F2,$23,$13,$C2,$7B,$D7,$49
	.dc.b	$C9,$F4,$AB,$84,$EA,$3D,$07,$84,$D5,$FA,$A7,$77,$E7,$9D,$9D,$0F
	.dc.b	$A0,$71,$29,$9B,$FA,$4E,$5A,$E5,$FA,$8F,$F7,$2C,$DF,$88,$3D,$3F
	.dc.b	$F0,$0C,$61,$B1,$99,$E0,$BC,$1C,$CC,$9C,$BC,$23,$11,$FE,$FF,$9E
	.dc.b	$4D,$9D,$3F,$DB,$FC,$0F,$DD,$91,$FD,$AB,$5F,$D8,$B5,$FD,$FB,$5F
	.dc.b	$F0,$EB,$5F,$F0,$73,$FE,$27,$7F,$C5,$17,$F0,$3E,$93,$DA,$48,$F6
	.dc.b	$F4,$D7,$D1,$17,$6F,$F1,$6D,$7C,$8A,$C6,$9F,$E3,$88,$E9,$7F,$1A
	.dc.b	$FE,$4D,$7F,$89,$BE,$A3,$F8,$B9,$27,$F8,$8B,$5F,$E1,$AD,$7F,$8D
	.dc.b	$B5,$FF,$1C,$B5,$FF,$15,$3E,$D7,$C5,$9A,$3E,$C9,$23,$EC,$F4,$57
	.dc.b	$EA,$45,$CB,$DB,$4A,$C6,$5D,$2A,$DB,$2F,$C4,$D7,$F9,$1B,$F0,$77
	.dc.b	$87,$E3,$AF,$F2,$DC,$61,$C3,$5F,$43,$5F,$AC,$D7,$FC,$BC,$D7,$9C
	.dc.b	$75,$51,$96,$C6,$63,$04,$7C,$94,$F3,$9D,$9F,$A0,$6D,$FE,$95,$AB
	.dc.b	$4F,$C0,$C3,$AB,$85,$4A,$E0,$F3,$8C,$47,$1D,$3F,$0D,$5D,$2F,$9D
	.dc.b	$EE,$BF,$CC,$39,$78,$1F,$87,$1E,$B2,$3F,$29,$AF,$CA,$6B,$C0,$D1
	.dc.b	$0D,$1F,$87,$C7,$E1,$72,$7C,$12,$47,$C1,$E8,$AF,$A4,$FC,$06,$29
	.dc.b	$CE,$3F,$12,$AE,$5E,$E9,$B7,$C2,$7F,$0D,$9F,$D5,$24,$7E,$11,$AF
	.dc.b	$E1,$1A,$FE,$11,$AF,$E1,$1A,$FE,$39,$FB,$B8,$3D,$DF,$B8,$49,$FB
	.dc.b	$9F,$09,$5F,$66,$7F,$83,$9C,$67,$F7,$AA,$E7,$F6,$26,$DF,$C2,$9F
	.dc.b	$D0,$76,$FE,$3C,$91,$F9,$E6,$BF,$9E,$6B,$F9,$E6,$BF,$9E,$6B,$FA
	.dc.b	$87,$D7,$E2,$FE,$54,$91,$F9,$5F,$84,$AF,$AB,$34,$E3,$2F,$CB,$D5
	.dc.b	$CB,$F2,$33,$6F,$ED,$CF,$EE,$7D,$4F,$D4,$12,$3F,$6E,$D7,$F6,$ED
	.dc.b	$7F,$6E,$D7,$F6,$ED,$7F,$80,$BF,$50,$FD,$71,$1F,$D7,$7C,$D5,$7D
	.dc.b	$51,$70,$FD,$7C,$DC,$3F,$53,$36,$BF,$27,$3F,$ED,$6F,$C1,$BF,$79
	.dc.b	$1F,$C4,$5A,$FF,$25,$6B,$FD,$0D,$AF,$FB,$AD,$AF,$FB,$C9,$A9,$A4
	.dc.b	$CB,$F8,$93,$63,$C1,$1F,$C3,$BE,$37,$F9,$34,$7F,$6F,$6B,$F9,$06
	.dc.b	$E3,$FD,$31,$7A,$BF,$DC,$9C,$E5,$FD,$35,$B5,$FC,$C5,$BC,$07,$8C
	.dc.b	$77,$8E,$44,$7C,$55,$F1,$FA,$BD,$24,$8F,$90,$D7,$E4,$35,$F9,$0D
	.dc.b	$7E,$43,$5F,$90,$FC,$1F,$8C,$F5,$12,$3E,$17,$99,$5F,$85,$17,$2F
	.dc.b	$85,$37,$2F,$81,$36,$BE,$04,$FC,$5F,$8E,$F8,$DF,$2B,$27,$E7,$46
	.dc.b	$5F,$A1,$19,$7E,$84,$65,$FA,$51,$9C,$3E,$94,$E5,$FA,$91,$9C,$3E
	.dc.b	$94,$E5,$FA,$91,$9A,$FA,$93,$97,$EB,$46,$73,$FA,$D3,$F9,$7F,$2D
	.dc.b	$53,$06,$18,$F3,$9E,$C3,$4D,$DF,$F3,$39,$CC,$73,$8E,$55,$F9,$BC
	.dc.b	$37,$87,$49,$5F,$07,$66,$30,$EC,$E0,$FD,$05,$75,$7C,$11,$A6,$E1
	.dc.b	$9B,$ED,$1C,$0E,$B7,$DD,$D9,$D7,$85,$87,$1C,$7E,$06,$3D,$70,$F4
	.dc.b	$61,$C5,$EB,$7C,$30,$E5,$87,$BB,$0E,$38,$D9,$F6,$3D,$5F,$DE,$4E
	.dc.b	$70,$E9,$5C,$27,$99,$0F,$39,$30,$F0,$D7,$3A,$E3,$3D,$8E,$13,$A7
	.dc.b	$E0,$E3,$EB,$BE,$D1,$FD,$AC,$F8,$C5,$C7,$D8,$F7,$8D,$D7,$38,$D6
	.dc.b	$1E,$9B,$E9,$1E,$13,$27,$8A,$B8,$91,$F8,$26,$4E,$15,$F3,$A7,$9C
	.dc.b	$79,$23,$C0,$67,$1F,$01,$F2,$E3,$55,$D7,$1F,$87,$18,$3B,$9E,$B3
	.dc.b	$BC,$7E,$55,$F1,$47,$CA,$38,$E3,$F0,$AF,$C3,$5D,$70,$F2,$CE,$63
	.dc.b	$95,$F5,$BE,$38,$FD,$5B,$F8,$B1,$89,$EF,$18,$39,$CE,$F1,$F6,$CF
	.dc.b	$C5,$3D,$55,$E5,$AD,$C7,$29,$DB,$F4,$09,$8F,$C7,$7F,$03,$1F,$1F
	.dc.b	$1F,$44,$7C,$22,$1F,$1C,$EF,$5E,$59,$E1,$87,$6B,$E9,$5D,$B1,$F8
	.dc.b	$31,$82,$38,$24,$77,$C7,$E1,$3C,$4C,$61,$EC,$71,$1E,$FC,$3C,$31
	.dc.b	$EA,$21,$C4,$F3,$BE,$5A,$EF,$18,$37,$7F,$69,$FC,$1B,$E6,$7E,$0E
	.dc.b	$3F,$11,$E0,$7E,$1C,$F8,$8F,$C8,$BF,$24,$7D,$58,$FD,$3B,$FD,$EF
	.dc.b	$1F,$EE,$64,$7F,$C0,$4D,$7F,$C0,$4D,$78,$1A,$21,$A2,$D7,$A5,$FC
	.dc.b	$E6,$6F,$C0,$1E,$AF,$C6,$9F,$99,$39,$8C,$57,$C1,$8C,$4F,$A6,$FF
	.dc.b	$8A,$46,$23,$E6,$61,$F0,$CF,$C3,$C6,$6B,$C5,$19,$AF,$F7,$BE,$1E
	.dc.b	$2B,$E8,$7E,$F6,$F8,$65,$F5,$2F,$84,$FF,$C0,$1A,$E0,$F8,$2F,$C7
	.dc.b	$18,$C3,$BC,$63,$1F,$5E,$1F,$8C,$C7,$AE,$1F,$72,$3B,$57,$92,$B7
	.dc.b	$39,$3F,$D9,$18,$75,$AD,$C7,$9E,$30,$EC,$FC,$0D,$73,$7D,$A7,$CE
	.dc.b	$8D,$BA,$39,$57,$81,$FA,$F1,$AC,$7E,$0C,$71,$9D,$47,$CE,$9D,$57
	.dc.b	$A3,$0F,$6D,$F5,$38,$1E,$2C,$7E,$E5,$70,$8C,$63,$FD,$F2,$B8,$E1
	.dc.b	$E0,$BD,$E3,$E0,$C7,$D9,$87,$D1,$3D,$27,$23,$E4,$5F,$28,$F2,$63
	.dc.b	$DE,$78,$1F,$E9,$A7,$17,$F6,$23,$EF,$5F,$76,$9D,$4F,$8A,$E5,$E4
	.dc.b	$D8,$FF,$04,$5F,$86,$7C,$8D,$7C,$31,$AA,$F0,$E3,$F7,$27,$99,$DF
	.dc.b	$1F,$9D,$87,$E6,$3C,$59,$3F,$93,$7F,$37,$E1,$E2,$3D,$91,$71,$ED
	.dc.b	$37,$0F,$9D,$8A,$B1,$7F,$8E,$AB,$97,$ED,$E6,$D2,$6C,$61,$B5,$E3
	.dc.b	$3F,$4C,$E1,$F1,$98,$34,$60,$98,$79,$9C,$BF,$16,$7B,$46,$30,$F4
	.dc.b	$C6,$23,$8C,$61,$D1,$EB,$2E,$1E,$B6,$E1,$EB,$6D,$73,$BF,$04,$60
	.dc.b	$F4,$1D,$8F,$D9,$C6,$EF,$C1,$19,$BE,$CF,$E3,$70,$F8,$57,$D2,$30
	.dc.b	$75,$38,$61,$E3,$6E,$1D,$5E,$E6,$E7,$E3,$3A,$3D,$75,$DB,$0F,$1C
	.dc.b	$E8,$F0,$11,$C9,$F8,$28,$E1,$3D,$8F,$29,$D4,$F2,$BA,$C3,$DB,$5D
	.dc.b	$6B,$9D,$71,$8E,$CF,$D1,$39,$38,$9F,$E8,$55,$CF,$0F,$C9,$57,$CE
	.dc.b	$AD,$46,$23,$C2,$7C,$1B,$EF,$7C,$A3,$07,$86,$3D,$8E,$23,$CF,$18
	.dc.b	$8F,$1B,$CA,$7E,$85,$78,$30,$FA,$17,$C2,$FA,$11,$F0,$3C,$8C,$57
	.dc.b	$2B,$E9,$5C,$A3,$D5,$7D,$6B,$DD,$5A,$7E,$C9,$E7,$9D,$C7,$9E,$78
	.dc.b	$4F,$C2,$3C,$17,$E5,$3C,$71,$A3,$E4,$E1,$F3,$70,$FE,$0B,$5D,$5A
	.dc.b	$F2,$C3,$BB,$F5,$EB,$96,$1F,$16,$F9,$4F,$A0,$F7,$DF,$B4,$FB,$51
	.dc.b	$A8,$FC,$4D,$72,$7E,$2C,$66,$3F,$11,$1C,$8F,$82,$78,$E7,$CB,$1C
	.dc.b	$70,$F7,$61,$F3,$6B,$81,$F0,$E7,$51,$EA,$9F,$87,$87,$CD,$AD,$36
	.dc.b	$BE,$2C,$7C,$33,$D9,$3D,$AF,$D9,$1F,$4F,$0F,$45,$FD,$33,$91,$F1
	.dc.b	$63,$91,$F0,$1C,$BF,$00,$CD,$FC,$0B,$ED,$1F,$02,$3D,$37,$F0,$2B
	.dc.b	$8D,$7C,$0B,$F0,$DF,$D5,$9C,$E1,$EF,$3E,$25,$74,$9F,$79,$F1,$B0
	.dc.b	$F7,$BF,$34,$F7,$CF,$1C,$3D,$F1,$F8,$3C,$3B,$DF,$CA,$8E,$58,$7B
	.dc.b	$DF,$F5,$85,$7F,$BB,$19,$97,$ED,$DB,$19,$7F,$DE,$31,$96,$C6,$5F
	.dc.b	$F7,$DC,$65,$B1,$95,$19,$6C,$66,$30,$47,$A5,$3E,$15,$C3,$FE,$DD
	.dc.b	$8C,$36,$33,$7F,$EF,$C9,$E1,$C1,$9A,$93,$9A,$DA,$D3,$FB,$FC,$2B
	.dc.b	$73,$9C,$BF,$C3,$93,$C2,$B7,$7E,$1B,$FF,$78,$5F,$BB,$3F,$61,$1F
	.dc.b	$8D,$9F,$F4,$57,$E7,$C6,$5F,$9D,$19,$7E,$7C,$65,$D9,$C9,$C3,$E4
	.dc.b	$F9,$6F,$84,$47,$DD,$F2,$D5,$F5,$0D,$3E,$64,$E7,$2F,$83,$57,$4E
	.dc.b	$97,$F1,$EB,$E7,$4F,$BB,$17,$8D,$F6,$49,$5E,$E6,$BE,$C6,$BF,$8E
	.dc.b	$6B,$F8,$C6,$BF,$8A,$7F,$03,$AB,$CD,$FB,$C4,$9F,$B5,$F2,$D5,$F6
	.dc.b	$E9,$F8,$79,$C6,$9F,$72,$AE,$9F,$6A,$6D,$FC,$3C,$15,$F6,$B9,$57
	.dc.b	$F9,$96,$D7,$E7,$DA,$FE,$95,$AF,$EA,$DA,$88,$E1,$B1,$D4,$8F,$56
	.dc.b	$BD,$5F,$D1,$FD,$C3,$DB,$FD,$13,$67,$F4,$0D,$7E,$1A,$F8,$ED,$CC
	.dc.b	$9D,$BC,$A3,$11,$C2,$2E,$7F,$E6,$2A,$E5,$3E,$B8,$F8,$38,$3C,$2F
	.dc.b	$8C,$49,$F8,$CD,$7E,$33,$5F,$8C,$D7,$E3,$08,$E8,$F2,$21,$FC,$5F
	.dc.b	$55,$BD,$0F,$86,$49,$F8,$7F,$13,$5F,$87,$17,$1F,$64,$DD,$3F,$D3
	.dc.b	$17,$CE,$FE,$5D,$FC,$EA,$D4,$7E,$A6,$74,$F4,$3B,$9C,$08,$64,$FA
	.dc.b	$71,$CB,$0F,$4D,$5A,$4D,$8C,$36,$B9,$E1,$E8,$6E,$1E,$8A,$DC,$60
	.dc.b	$F3,$38,$7C,$53,$F6,$E3,$0D,$8E,$18,$79,$1F,$59,$C2,$BB,$3A,$C3
	.dc.b	$CE,$DB,$E7,$3A,$8E,$07,$88,$E7,$7D,$DC,$DF,$77,$E2,$9C,$CE,$D8
	.dc.b	$7C,$17,$78,$78,$1F,$E3,$63,$F0,$67,$CC,$7F,$32,$EF,$79,$14,$FE
	.dc.b	$6B,$3A,$AF,$CD,$67,$F9,$08,$C4,$74,$AE,$2E,$8F,$DD,$38,$7C,$31
	.dc.b	$C3,$3F,$E0,$6D,$AF,$CD,$D6,$E7,$F8,$3B,$63,$C1,$87,$90,$F3,$5F
	.dc.b	$71,$3B,$7F,$85,$5F,$56,$CF,$07,$D1,$87,$37,$94,$ED,$E7,$1B,$C3
	.dc.b	$CF,$5B,$9F,$EF,$D5,$F0,$9F,$F0,$23,$5F,$39,$FD,$FD,$C3,$E2,$31
	.dc.b	$1F,$DF,$E3,$31,$F5,$63,$04,$76,$43,$E5,$BE,$39,$F8,$FE,$7D,$C7
	.dc.b	$C3,$5A,$3E,$36,$29,$D4,$78,$CD,$1F,$6D,$FE,$1B,$5F,$CE,$B8,$1F
	.dc.b	$B3,$3E,$0B,$38,$FF,$37,$9C,$F3,$63,$1F,$E6,$17,$FD,$4D,$FC,$24
	.dc.b	$F2,$BD,$36,$33,$3B,$7F,$B8,$C7,$47,$33,$C2,$7C,$51,$C2,$B0,$66
	.dc.b	$64,$64,$87,$49,$E2,$D8,$E2,$EB,$0F,$81,$3A,$6C,$EA,$79,$57,$67
	.dc.b	$FA,$4F,$AB,$FD,$38,$8F,$F4,$EF,$7E,$BA,$3F,$A8,$C7,$F5,$48,$CD
	.dc.b	$FF,$49,$8B,$97,$F2,$E9,$D5,$70,$21,$FD,$61,$CB,$D4,$95,$FD,$4E
	.dc.b	$32,$F5,$8E,$2E,$5F,$31,$0E,$87,$F6,$6F,$03,$8D,$78,$E3,$53,$FD
	.dc.b	$7E,$73,$1F,$CF,$EA,$F1,$FF,$50,$9F,$F1,$FF,$E2,$FF,$D0,$AF,$FC
	.dc.b	$0B,$E2,$FF,$89,$3E,$23,$F1,$0D,$C7,$F7,$95,$BC,$BF,$AF,$B6,$13
	.dc.b	$B7,$FC,$6E,$DB,$FF,$07,$CE,$E4,$F1,$9D,$57,$F8,$3D,$B3,$FE,$11
	.dc.b	$9E,$04,$E7,$F4,$9F,$E5,$8F,$C1,$5E,$6E,$2B,$CF,$35,$FE,$5F,$3A
	.dc.b	$72,$43,$F9,$9D,$79,$63,$51,$89,$E5,$8F,$A9,$F0,$5E,$31,$F8,$71
	.dc.b	$6B,$FC,$51,$7C,$AF,$84,$77,$6B,$E4,$BE,$0F,$C4,$AB,$5C,$6F,$AC
	.dc.b	$F1,$6B,$C4,$C1,$F1,$E7,$8D,$7F,$A3,$FE,$E9,$F0,$17,$D6,$97,$FE
	.dc.b	$8F,$9C,$B6,$72,$FF,$95,$21,$3C,$E6,$4F,$18,$DE,$5F,$E4,$ED,$D7
	.dc.b	$D2,$B9,$5E,$B2,$FF,$26,$56,$B0,$E7,$5A,$3A,$10,$C1,$30,$ED,$87
	.dc.b	$92,$35,$18,$C3,$FD,$37,$F6,$CF,$92,$FF,$53,$53,$E7,$BF,$B9,$7E
	.dc.b	$0B,$FD,$5A,$3F,$CB,$D1,$8E,$BF,$F1,$DB,$FD,$DE,$F9,$63,$AF,$03
	.dc.b	$FC,$7F,$18,$BE,$91,$FE,$42,$C3,$A4,$60,$FF,$35,$38,$BE,$98,$F9
	.dc.b	$1B,$87,$5A,$DE,$3E,$2C,$3A,$36,$3F,$CE,$98,$F8,$6F,$8C,$77,$18
	.dc.b	$F7,$BE,$31,$D4,$57,$18,$F5,$E3,$E3,$77,$3C,$EB,$93,$E0,$9D,$37
	.dc.b	$1E,$93,$A8,$F0,$E3,$E7,$7B,$D7,$06,$D7,$09,$D5,$71,$C3,$C9,$5D
	.dc.b	$4F,$14,$E8,$E9,$1A,$8E,$66,$8E,$87,$BD,$D1,$B9,$F3,$1D,$EF,$C9
	.dc.b	$19,$AE,$D5,$C6,$72,$7A,$6B,$8C,$F8,$A3,$37,$E0,$AF,$11,$F4,$EF
	.dc.b	$C3,$3C,$6B,$AC,$E6,$38,$BC,$31,$FC,$93,$E1,$B8,$E7,$1F,$AB,$3F
	.dc.b	$80,$23,$82,$47,$C9,$8C,$BF,$8C,$3C,$CE,$1F,$92,$F8,$1E,$C7,$18
	.dc.b	$FC,$2C,$F0,$7A,$C7,$07,$2F,$03,$F0,$27,$A2,$7B,$63,$F2,$23,$13
	.dc.b	$C2,$7C,$04,$70,$48,$F4,$1D,$71,$F9,$77,$E3,$8C,$D7,$AE,$7E,$D4
	.dc.b	$64,$F3,$1E,$B3,$75,$EA,$8E,$71,$E9,$BF,$C2,$5F,$9A,$36,$43,$99
	.dc.b	$23,$D8,$FD,$B8,$E0,$FC,$23,$D1,$87,$DE,$C3,$DD,$1C,$EF,$C3,$18
	.dc.b	$8E,$0F,$AE,$7F,$11,$1E,$B9,$DE,$1F,$4B,$0F,$3C,$F8,$4F,$56,$3F
	.dc.b	$95,$C7,$C7,$3E,$6C,$7A,$47,$C5,$8E,$31,$F5,$A7,$F2,$AE,$AB,$99
	.dc.b	$EC,$C3,$E6,$BE,$CB,$EB,$5C,$30,$FB,$13,$F1,$63,$0F,$D1,$30,$7A
	.dc.b	$F1,$F8,$93,$EB,$9F,$44,$FA,$F0,$F8,$98,$F9,$AB,$AB,$EB,$7B,$E3
	.dc.b	$E4,$9F,$94,$FC,$6A,$FB,$15,$C5,$F1,$3C,$5F,$24,$71,$AF,$C2,$61
	.dc.b	$E4,$9E,$38,$79,$30,$F8,$D1,$E9,$C7,$E0,$B4,$F4,$5F,$AD,$F9,$17
	.dc.b	$F6,$9F,$90,$68,$EA,$6A,$7E,$A5,$F7,$AF,$7B,$F2,$A7,$F2,$38,$79
	.dc.b	$DF,$C5,$5F,$9D,$FC,$4D,$7A,$AF,$CB,$1E,$0C,$3D,$2D,$9E,$AF,$A6
	.dc.b	$32,$7C,$AA,$E9,$8F,$C7,$9F,$0C,$7A,$E7,$57,$D9,$B3,$ED,$3C,$67
	.dc.b	$9A,$35,$7E,$58,$D5,$F5,$C3,$EC,$47,$02,$38,$8D,$61,$E4,$C7,$E2
	.dc.b	$4E,$5B,$19,$AF,$65,$FB,$9F,$80,$7E,$0A,$77,$87,$92,$3C,$C7,$92
	.dc.b	$32,$D3,$99,$C4,$A7,$E7,$0E,$A5,$C7,$E1,$9F,$32,$BE,$03,$63,$BB
	.dc.b	$5F,$7D,$FA,$5E,$78,$7D,$F7,$38,$79,$DA,$79,$4D,$CF,$D5,$BE,$E7
	.dc.b	$C1,$BF,$1B,$4C,$CF,$AC,$E3,$5F,$92,$7F,$28,$7F,$02,$3E,$E5,$F8
	.dc.b	$F1,$FD,$64,$7B,$EB,$9B,$4F,$6B,$DD,$A7,$7B,$F1,$B6,$38,$D7,$5C
	.dc.b	$7F,$45,$39,$38,$9D,$4F,$83,$5F,$26,$FC,$75,$C6,$7D,$33,$FC,$2A
	.dc.b	$B9,$C7,$1A,$DE,$3F,$1E,$3A,$DF,$17,$DF,$19,$7B,$E1,$EB,$69,$E5
	.dc.b	$29,$D6,$FB,$D7,$3C,$3F,$96,$1E,$39,$F2,$D7,$F0,$77,$BE,$3F,$45
	.dc.b	$A7,$D1,$9E,$33,$B6,$C7,$F0,$A3,$6F,$56,$C6,$CF,$E4,$F8,$7A,$F1
	.dc.b	$F8,$37,$DE,$F9,$DF,$EC,$67,$2F,$86,$F8,$D6,$DB,$19,$3F,$95,$DF
	.dc.b	$5C,$7E,$0B,$63,$DE,$79,$F0,$EF,$87,$EF,$30,$FC,$E3,$E0,$8F,$51
	.dc.b	$AB,$FD,$9B,$F7,$DE,$B3,$E3,$AF,$0B,$5F,$E2,$ED,$C7,$F8,$34,$7E
	.dc.b	$E1,$FA,$98,$FA,$AF,$BB,$4F,$91,$3D,$0F,$71,$D0,$B3,$D1,$FD,$49
	.dc.b	$8C,$7F,$75,$5F,$9E,$8F,$C2,$1F,$08,$F9,$D8,$7A,$9E,$18,$7C,$28
	.dc.b	$C4,$F9,$27,$C4,$D7,$C4,$DB,$F1,$3D,$DF,$6C,$70,$C3,$D6,$F0,$6C
	.dc.b	$70,$7F,$63,$87,$D7,$9F,$C4,$9F,$5E,$3E,$23,$4E,$85,$3A,$16,$7A
	.dc.b	$3F,$AC,$9E,$95,$F6,$E3,$E9,$46,$2B,$AC,$7C,$43,$E3,$63,$F8,$57
	.dc.b	$F8,$F5,$7B,$6B,$84,$78,$8E,$07,$A0,$B1,$DD,$FA,$AF,$C7,$8F,$AB
	.dc.b	$5F,$03,$1F,$C3,$46,$2B,$D1,$06,$1F,$E6,$02,$9F,$E9,$22,$E1,$FE
	.dc.b	$C5,$6C,$FF,$B4,$9B,$3C,$1F,$F4,$F1,$83,$13,$C5,$EC,$43,$67,$FB
	.dc.b	$4B,$D0,$7D,$7F,$38,$FF,$5E,$C6,$1E,$6D,$7C,$AD,$79,$47,$19,$4E
	.dc.b	$67,$FD,$9E,$D7,$AC,$EE,$31,$3C,$23,$04,$9F,$8F,$F6,$22,$00

;--------------------------------------------------------------------------------
;	FSSUB.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68040
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fssub_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#60,-(sp)
	peamsg	'FSSUB.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSSUB.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fssub_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fssub.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSSUB.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fssub.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSSUB.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fssub_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3D,$61,$F3,$7F,$DA,$BD,$E7,$99,$C6,$30,$DA,$FE
	.dc.b	$50,$D9,$DC,$7F,$72,$F8,$17,$BF,$FC,$8E,$3F,$B9,$88,$CF,$C6,$E2
	.dc.b	$31,$E6,$FF,$42,$7F,$BC,$CE,$9E,$2D,$73,$5D,$5D,$CF,$17,$F8,$A5
	.dc.b	$75,$AE,$11,$83,$9C,$ED,$F4,$57,$37,$47,$6A,$E5,$3B,$8F,$54,$ED
	.dc.b	$E3,$3D,$E3,$11,$EC,$F9,$3C,$4C,$8F,$5C,$EA,$D7,$D5,$F3,$CF,$43
	.dc.b	$75,$E2,$BE,$51,$E5,$32,$7B,$1C,$BE,$A2,$1F,$56,$BB,$5F,$84,$95
	.dc.b	$D6,$79,$9D,$48,$7F,$40,$AE,$AF,$94,$C9,$D5,$CB,$F1,$A2,$DF,$BA
	.dc.b	$6D,$78,$2B,$D0,$78,$0F,$39,$C4,$C9,$D8,$C1,$67,$A3,$C4,$F2,$1C
	.dc.b	$9C,$BB,$8F,$03,$F8,$89,$F1,$D7,$9A,$FE,$2B,$89,$F0,$9C,$EB,$8B
	.dc.b	$F0,$CF,$01,$E4,$3C,$13,$C9,$CD,$FD,$C7,$F0,$86,$27,$9B,$F2,$4C
	.dc.b	$57,$8E,$FC,$51,$EF,$BE,$8F,$AE,$7E,$C3,$4E,$F1,$A7,$D1,$1A,$69
	.dc.b	$A9,$EA,$FA,$1A,$FC,$A3,$C8,$75,$3C,$4E,$9F,$80,$68,$EA,$68,$F5
	.dc.b	$9D,$CA,$64,$E8,$78,$4C,$94,$C9,$E3,$3E,$49,$83,$C2,$60,$F0,$9C
	.dc.b	$4A,$71,$BF,$C1,$B8,$38,$B8,$75,$7F,$51,$E8,$62,$BE,$A3,$E6,$3A
	.dc.b	$5F,$CF,$74,$7E,$20,$A7,$BC,$F0,$57,$E0,$CF,$03,$E2,$AF,$04,$60
	.dc.b	$D3,$89,$ED,$3B,$AF,$B8,$F6,$3E,$01,$4F,$2D,$FD,$7B,$F0,$57,$27
	.dc.b	$AD,$72,$AE,$87,$D7,$3E,$74,$70,$AE,$57,$FC,$66,$F9,$3E,$58,$CD
	.dc.b	$F3,$AE,$B7,$CC,$F8,$35,$C6,$B7,$1F,$1C,$F0,$9F,$1C,$F5,$1F,$5A
	.dc.b	$FB,$DF,$D2,$8F,$2D,$73,$7E,$A1,$BA,$EA,$FE,$FC,$CD,$FA,$DB,$3C
	.dc.b	$5E,$75,$E9,$8F,$A6,$66,$3D,$31,$93,$E6,$5F,$92,$3D,$C7,$A2,$BB
	.dc.b	$3D,$0F,$73,$F1,$1F,$71,$F0,$0F,$B2,$53,$EC,$9E,$69,$F0,$3E,$A9
	.dc.b	$F0,$3C,$A7,$A3,$C9,$AF,$23,$C4,$7C,$32,$9F,$C6,$CA,$7E,$EC,$F5
	.dc.b	$57,$C1,$6D,$78,$9F,$C2,$D7,$89,$F0,$4F,$C8,$9D,$47,$E8,$27,$4F
	.dc.b	$8C,$F9,$67,$C4,$9E,$0D,$9F,$9B,$7F,$6A,$3C,$C7,$03,$D4,$53,$27
	.dc.b	$98,$E4,$7C,$77,$93,$ED,$3E,$71,$F8,$73,$E7,$1C,$0F,$6B,$C1,$FE
	.dc.b	$06,$70,$BF,$BA,$F8,$4E,$8F,$D0,$7A,$1F,$30,$E2,$E4,$E2,$F9,$CF
	.dc.b	$59,$4F,$59,$DE,$BE,$93,$6B,$E6,$BF,$8F,$BF,$9A,$F8,$0F,$21,$E0
	.dc.b	$9F,$9A,$F6,$39,$1E,$F3,$EB,$5F,$BE,$FF,$89,$4F,$09,$F1,$C7,$A6
	.dc.b	$3C,$6D,$BF,$1C,$FC,$5B,$F1,$CF,$BA,$3B,$9E,$17,$83,$E1,$3D,$57
	.dc.b	$E1,$9E,$37,$D8,$F1,$4F,$CE,$6B,$ED,$BF,$C3,$BF,$68,$ED,$1A,$BF
	.dc.b	$5C,$EA,$7E,$A4,$61,$FC,$31,$C4,$F7,$CF,$1A,$F1,$D7,$BA,$BE,$D4
	.dc.b	$62,$BD,$47,$C1,$BF,$51,$1E,$87,$A6,$BC,$4F,$D6,$AF,$13,$F9,$03
	.dc.b	$67,$B4,$F2,$9E,$8A,$F0,$3E,$E9,$E6,$FE,$24,$EC,$70,$3B,$47,$07
	.dc.b	$DB,$1C,$1F,$AA,$60,$E0,$73,$3C,$40,$FF,$41,$63,$1F,$F4,$66,$5F
	.dc.b	$E8,$86,$9F,$E8,$82,$9C,$4A,$7F,$A3,$4C,$1F,$E9,$CC,$FF,$D2,$AF
	.dc.b	$FA,$2B,$64,$63,$4F,$F4,$8C,$EE,$BC,$0D,$3C,$05,$8D,$BE,$9F,$95
	.dc.b	$FF,$49,$34,$AE,$13,$48,$E8,$87,$FA,$6D,$CB,$4C,$9F,$EA,$27,$2D
	.dc.b	$32,$7C,$07,$2D,$32,$7C,$77,$2D,$32,$7A,$4F,$01,$0E,$C5,$3A,$14
	.dc.b	$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08
	.dc.b	$7D,$03,$DA,$53,$CC,$53,$C2,$53,$B1,$4E,$A5,$3A,$BF,$29,$A7,$67
	.dc.b	$0E,$C8,$E0,$8E,$C8,$79,$1C,$B4,$C9,$E6,$72,$D3,$27,$A1,$CB,$4C
	.dc.b	$9E,$A7,$2D,$32,$7A,$8F,$01,$0E,$C5,$3A,$14,$E4,$53,$91,$C0,$87
	.dc.b	$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$7D,$03,$DA,$53,$CC
	.dc.b	$53,$C2,$53,$B1,$4E,$A5,$3A,$BF,$29,$A7,$67,$0E,$C8,$E0,$8E,$C8
	.dc.b	$79,$1C,$B4,$C9,$E6,$72,$D3,$27,$A1,$CB,$4C,$9E,$A7,$2D,$32,$7A
	.dc.b	$8F,$01,$0E,$C5,$3A,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34
	.dc.b	$43,$45,$34,$60,$86,$08,$7D,$03,$DA,$53,$CC,$53,$C2,$53,$B1,$4E
	.dc.b	$A5,$3A,$BF,$29,$A7,$67,$0E,$C8,$E0,$8E,$C8,$79,$1C,$B4,$C9,$E6
	.dc.b	$72,$D3,$27,$A1,$CB,$4C,$9E,$A7,$2D,$32,$7A,$8F,$01,$0E,$C5,$3A
	.dc.b	$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86
	.dc.b	$08,$7D,$03,$DA,$53,$CC,$53,$C2,$53,$B1,$4E,$A5,$3A,$BF,$29,$A7
	.dc.b	$67,$0E,$C8,$E0,$8E,$C8,$7C,$BB,$FF,$0A,$D7,$F8,$6A,$3F,$C3,$A4
	.dc.b	$3F,$C2,$87,$F8,$7A,$1B,$BE,$6F,$84,$A7,$31,$86,$C7,$F8,$8E,$78
	.dc.b	$3F,$E3,$61,$AB,$76,$BF,$E1,$8F,$9E,$71,$FF,$8C,$7E,$71,$F2,$0F
	.dc.b	$88,$FF,$1E,$56,$B2,$FF,$09,$35,$57,$09,$DD,$FF,$86,$FE,$83,$53
	.dc.b	$9A,$47,$17,$FC,$5C,$EB,$0F,$04,$6A,$B9,$1E,$58,$CC,$F1,$9F,$F0
	.dc.b	$23,$C6,$FF,$CA,$4E,$78,$3C,$73,$FE,$55,$C3,$C5,$8E,$60,$F1,$6E
	.dc.b	$3E,$07,$86,$3E,$08,$C4,$7C,$08,$C3,$FE,$35,$C3,$91,$D4,$B3,$C2
	.dc.b	$FE,$14,$66,$3D,$98,$FA,$1D,$5F,$4B,$FF,$1E,$57,$27,$B4,$F0,$7E
	.dc.b	$31,$83,$CF,$5C,$DC,$1E,$77,$0F,$48,$E5,$3E,$38,$E2,$74,$34,$61
	.dc.b	$D3,$CC,$D1,$F6,$A3,$99,$EA,$9E,$F1,$93,$94,$65,$E2,$7D,$A7,$8B
	.dc.b	$E8,$BF,$8F,$1B,$AF,$0C,$62,$75,$7F,$1D,$EC,$79,$67,$71,$98,$EC
	.dc.b	$78,$8E,$37,$E6,$72,$71,$BF,$01,$1D,$9E,$8C,$7E,$5E,$1E,$38,$F8
	.dc.b	$E7,$13,$78,$78,$AF,$B5,$70,$C3,$C8,$D8,$CD,$7D,$1B,$E8,$FB,$1A
	.dc.b	$F9,$9B,$3C,$63,$C8,$71,$AE,$AD,$AE,$B7,$F8,$B8,$F3,$DF,$27,$47
	.dc.b	$28,$D3,$E1,$AE,$71,$88,$F1,$46,$30,$FB,$31,$D2,$30,$FC,$F3,$17
	.dc.b	$E2,$9D,$C7,$58,$C3,$B3,$07,$8E,$36,$F8,$23,$6F,$CF,$3A,$9B,$39
	.dc.b	$1B,$30,$6C,$EB,$5F,$55,$CD,$FC,$38,$F7,$61,$F3,$5D,$12,$FF,$09
	.dc.b	$5E,$08,$C3,$F8,$08,$CB,$D9,$AF,$8E,$36,$F1,$8C,$CF,$76,$C7,$D4
	.dc.b	$9D,$BC,$CE,$C4,$39,$14,$D1,$82,$18,$21,$E3,$8E,$F1,$F0,$23,$C1
	.dc.b	$18,$3E,$8C,$F9,$63,$B1,$C4,$AE,$5E,$E6,$4A,$E5,$F0,$B5,$D1,$83
	.dc.b	$C2,$74,$29,$C8,$A6,$E7,$D0,$49,$F3,$C6,$8E,$A4,$9F,$01,$1D,$36
	.dc.b	$79,$47,$E2,$A3,$13,$D2,$73,$3F,$8B,$3A,$1C,$88,$EA,$BF,$19,$3C
	.dc.b	$49,$3D,$4D,$B8,$7A,$57,$81,$D5,$73,$C3,$F5,$B1,$A3,$89,$AA,$E5
	.dc.b	$7E,$08,$C1,$0E,$15,$CF,$0F,$D5,$CE,$E7,$35,$CA,$FB,$CE,$48,$E0
	.dc.b	$95,$C9,$B5,$FA,$C6,$DF,$E6,$A3,$E3,$BE,$17,$53,$C4,$8E,$09,$1C
	.dc.b	$70,$F8,$91,$E6,$3D,$97,$E4,$AE,$35,$B9,$F8,$A6,$C9,$3B,$7A,$E1
	.dc.b	$F9,$F7,$9E,$1E,$88,$C4,$70,$8C,$3F,$95,$AF,$51,$DC,$B3,$C2,$BB
	.dc.b	$46,$63,$E6,$5F,$99,$D5,$F4,$BF,$CF,$57,$27,$B4,$F0,$7C,$E6,$0F
	.dc.b	$3D,$73,$70,$79,$DC,$3D,$23,$94,$F8,$E3,$89,$D0,$D1,$87,$4F,$33
	.dc.b	$47,$EC,$63,$99,$EA,$9E,$F1,$93,$94,$65,$E2,$7E,$09,$E2,$FA,$2F
	.dc.b	$E3,$C6,$EB,$C3,$18,$9D,$5F,$C7,$7B,$1E,$59,$DC,$66,$3B,$1E,$23
	.dc.b	$8D,$F9,$9C,$9C,$6F,$C0,$47,$67,$A3,$0F,$C8,$E1,$E3,$8F,$8E,$71
	.dc.b	$37,$87,$8A,$FB,$57,$0C,$3C,$8D,$8C,$D7,$E5,$AF,$A3,$EC,$6B,$E6
	.dc.b	$6C,$F1,$8F,$21,$FE,$88,$3F,$DC,$1B,$7F,$A4,$9F,$F4,$93,$5F,$F4
	.dc.b	$87,$13,$87,$FD,$18,$D8,$FF,$46,$70,$FF,$A3,$F8,$7F,$D1,$B1,$FE
	.dc.b	$97,$6A,$25,$FF,$A5,$1D,$4F,$F9,$C9,$74,$BE,$0F,$84,$7F,$9C,$0E
	.dc.b	$51,$97,$85,$7F,$A7,$B8,$3F,$D2,$1C,$7F,$E9,$BE,$2F,$F5,$2B,$6B
	.dc.b	$FD,$5C,$B9,$9C,$FA,$E5,$FF,$50,$D7,$C5,$E0,$F9,$1C,$9F,$56,$3E
	.dc.b	$4B,$FA,$F7,$ED,$F1,$7C,$E7,$EA,$72,$7D,$86,$F2,$7D,$8A,$E1,$1F
	.dc.b	$5E,$75,$18,$8F,$B1,$18,$E5,$FC,$FC,$76,$8C,$3F,$95,$31,$5D,$67
	.dc.b	$71,$D6,$30,$EC,$C1,$EA,$8D,$BE,$08,$DB,$F9,$53,$A9,$B3,$91,$B3
	.dc.b	$06,$CE,$B5,$F1,$C9,$3F,$A5,$8E,$71,$87,$F4,$31,$97,$8B,$5E,$D1
	.dc.b	$B7,$8C,$67,$0F,$AA,$D8,$F9,$E7,$93,$0F,$A9,$5F,$36,$B7,$3E,$02
	.dc.b	$1D,$0A,$78,$8A,$78,$AB,$B4,$F3,$8E,$EF,$95,$E0,$73,$21,$82,$1C
	.dc.b	$CE,$B3,$E4,$76,$64,$AE,$5E,$2D,$74,$60,$F1,$CF,$98,$FB,$B1,$B2
	.dc.b	$46,$8F,$09,$4E,$E5,$3A,$4F,$BC,$93,$D2,$34,$78,$89,$3D,$A7,$71
	.dc.b	$F8,$88,$C4,$F1,$9C,$CF,$E2,$4E,$27,$49,$DB,$C4,$C1,$23,$8D,$7D
	.dc.b	$EA,$E8,$49,$E1,$3A,$C3,$F5,$31,$DE,$FA,$B4,$E9,$5C,$5C,$57,$1C
	.dc.b	$3F,$4B,$87,$79,$EB,$5B,$8C,$47,$08,$C6,$1E,$09,$C9,$1C,$12,$BA
	.dc.b	$B6,$BF,$34,$DB,$FC,$74,$7C,$67,$C8,$EA,$78,$91,$C1,$23,$8E,$1F
	.dc.b	$02,$FC,$55,$B9,$CD,$7C,$23,$DA,$43,$D8,$59,$E2,$F5,$C3,$F2,$0E
	.dc.b	$70,$F4,$46,$23,$89,$27,$FD,$98,$FF,$3E,$3F,$9F,$E1,$FD,$55,$34
	.dc.b	$FE,$A3,$CB,$23,$FA,$BF,$2F,$F5,$37,$FA,$BC,$7F,$53,$E4,$91,$FD
	.dc.b	$58,$87,$F5,$01,$93,$1F,$E9,$FC,$7F,$D3,$67,$31,$8E,$4F,$EA,$75
	.dc.b	$B9,$FE,$B7,$CA,$C3,$FA,$F4,$E6,$5C,$AA,$D4,$74,$3F,$AD,$DF,$4C
	.dc.b	$3B,$47,$27,$FB,$0D,$71,$C3,$13,$FD,$AF,$8F,$FA,$E3,$F2,$5A,$FF
	.dc.b	$68,$7F,$AF,$71,$7F,$5D,$E4,$F9,$53,$FD,$AC,$70,$38,$5C,$5F,$DC
	.dc.b	$E7,$FB,$C7,$17,$F7,$5A,$D4,$E6,$7E,$A9,$CE,$32,$F5,$38,$99,$AE
	.dc.b	$AE,$4F,$CD,$69,$FA,$6A,$FC,$93,$F8,$7A,$FC,$6E,$FF,$94,$3F,$3E
	.dc.b	$58,$E1,$5F,$91,$3F,$8D,$C6,$A3,$95,$FE,$4A,$73,$1C,$27,$94,$7E
	.dc.b	$5E,$FC,$2E,$8C,$1D,$4F,$E4,$B1,$D5,$D1,$F9,$BA,$EC,$EA,$BB,$1E
	.dc.b	$2A,$EB,$19,$3A,$D7,$89,$E5,$1A,$7F,$98,$CF,$27,$BD,$FC,$48,$E9
	.dc.b	$1E,$53,$67,$47,$35,$D6,$78,$4F,$2B,$F3,$BC,$0F,$09,$CC,$FA,$A7
	.dc.b	$33,$85,$F9,$5C,$9C,$2B,$B4,$F4,$8C,$D7,$81,$F8,$06,$EF,$F8,$03
	.dc.b	$FE,$E4,$E5,$FD,$44,$FE,$29,$FD,$54,$7E,$29,$CD,$7E,$29,$CC,$7E
	.dc.b	$49,$CE,$1F,$95,$C7,$F3,$15,$BC,$3E,$19,$2F,$D6,$7B,$E7,$F9,$24
	.dc.b	$FA,$5B,$5E,$F6,$CE,$CF,$E4,$C7,$AC,$E7,$8F,$C2,$9E,$11,$AC,$7E
	.dc.b	$13,$F3,$DA,$FC,$6B,$F0,$3B,$6D,$7E,$3E,$3E,$4C,$F0,$BF,$1C,$F7
	.dc.b	$9F,$23,$B7,$B5,$72,$C3,$F1,$51,$CA,$30,$FD,$D3,$18,$7A,$2B,$94
	.dc.b	$79,$63,$15,$CA,$3A,$C6,$1E,$46,$0F,$1C,$6D,$E4,$60,$D9,$89,$F5
	.dc.b	$12,$3E,$B4,$79,$0E,$07,$CF,$31,$1B,$7F,$C3,$78,$FE,$09,$FE,$1D
	.dc.b	$87,$E0,$AF,$F0,$91,$F3,$5F,$F5,$5B,$F9,$23,$FD,$47,$C7,$FD,$2D
	.dc.b	$FE,$62,$D7,$FD,$32,$D7,$F9,$EB,$77,$FF,$4E,$47,$F3,$8E,$1F,$F5
	.dc.b	$0B,$FD,$1B,$83,$FA,$9B,$FD,$47,$2F,$E8,$DC,$5F,$D3,$DB,$C5,$FD
	.dc.b	$0E,$78,$1F,$D1,$CF,$E9,$77,$FD,$2A,$77,$1F,$D1,$0F,$F5,$4C,$F1
	.dc.b	$77,$79,$C3,$FD,$4D,$AF,$F4,$E9,$FE,$85,$8F,$F5,$FD,$7F,$AC,$56
	.dc.b	$A7,$FA,$77,$27,$FA,$91,$FE,$9C,$D7,$FA,$77,$2F,$FA,$A9,$FF,$56
	.dc.b	$C6,$5F,$EF,$9B,$FD,$17,$E9,$9F,$D0,$F7,$FA,$7A,$FE,$3A,$7F,$37
	.dc.b	$9F,$E0,$2B,$F1,$53,$F7,$5F,$BF,$19,$7F,$10,$D9,$E3,$1F,$B2,$8C
	.dc.b	$3F,$7C,$FC,$4F,$0F,$EA,$5C,$57,$57,$BC,$F0,$8F,$24,$6A,$30,$7E
	.dc.b	$6E,$BB,$C7,$90,$F0,$10,$C1,$0E,$A6,$4A,$64,$F5,$99,$29,$97,$BB
	.dc.b	$C0,$86,$8A,$E9,$C1,$1C,$11,$F6,$1D,$DC,$4F,$C0,$24,$FB,$E7,$67
	.dc.b	$8C,$A7,$80,$A7,$72,$CF,$48,$E5,$3D,$27,$F8,$0C,$E5,$A7,$02,$46
	.dc.b	$DF,$E0,$07,$77,$15,$F6,$DB,$C3,$FC,$72,$76,$78,$E3,$38,$7F,$17
	.dc.b	$8C,$DF,$85,$A7,$83,$0F,$39,$1F,$2D,$72,$7F,$17,$7D,$F0,$F1,$56
	.dc.b	$E7,$9D,$6C,$FC,$73,$E3,$8D,$4F,$08,$EB,$18,$AE,$84,$C3,$D1,$5B
	.dc.b	$AE,$B5,$B9,$CD,$6E,$7C,$B1,$83,$FA,$9B,$FC,$7F,$0F,$F2,$1B,$4F
	.dc.b	$F3,$5F,$2F,$F9,$B5,$FF,$35,$98,$3F,$CD,$27,$F1,$DD,$FF,$8D,$B4
	.dc.b	$FF,$37,$6D,$FC,$51,$FE,$1B,$C0,$64,$7F,$0E,$E1,$FF,$31,$B6,$BF
	.dc.b	$8D,$70,$FF,$23,$D3,$F8,$B6,$D9,$C3,$F9,$17,$0F,$F1,$57,$F8,$F9
	.dc.b	$D2,$7F,$8E,$61,$B2,$38,$24,$7F,$2C,$8F,$F3,$B7,$2F,$F9,$A2,$3F
	.dc.b	$8F,$F2,$7F,$43,$8F,$F3,$4C,$7F,$1C,$8F,$F3,$4E,$1F,$6E,$FF,$81
	.dc.b	$CF,$D8,$21,$F5,$0B,$C3,$F5,$C9,$C3,$F5,$6B,$F0,$7C,$1F,$91,$9C
	.dc.b	$F0,$FD,$7B,$FB,$24,$AD,$BF,$95,$3A,$DF,$3C,$3C,$71,$E1,$25,$F4
	.dc.b	$79,$46,$A3,$15,$D0,$F6,$4F,$F0,$33,$B1,$CA,$BB,$39,$8D,$BF,$C0
	.dc.b	$DA,$FE,$D6,$BF,$28,$F8,$0E,$0E,$23,$C0,$D3,$C9,$3D,$A7,$A4,$65
	.dc.b	$E2,$7C,$0B,$F8,$D5,$CA,$30,$77,$38,$98,$74,$FA,$E3,$8C,$F8,$1B
	.dc.b	$1D,$8D,$1C,$CF,$61,$CE,$38,$39,$3F,$78,$64,$E5,$7F,$3A,$37,$5E
	.dc.b	$28,$C4,$EA,$FE,$03,$D8,$F0,$CE,$E3,$31,$D8,$CD,$74,$23,$CE,$FD
	.dc.b	$6F,$C2,$36,$64,$F4,$61,$F9,$1A,$EA,$D9,$E3,$3E,$3C,$3C,$EF,$F4
	.dc.b	$13,$F8,$41,$0F,$F3,$DC,$FF,$09,$7F,$C5,$3C,$35,$FF,$14,$BF,$C2
	.dc.b	$1F,$F0,$E9,$FE,$F3,$21,$FC,$2F,$1F,$E1,$44,$C7,$04,$8F,$F7,$49
	.dc.b	$FC,$23,$87,$FC,$41,$B7,$FB,$9E,$3F,$83,$F0,$FF,$89,$EB,$F8,$59
	.dc.b	$23,$FC,$5B,$7F,$C3,$4E,$CF,$43,$8F,$0F,$F8,$A6,$B9,$46,$67,$6F
	.dc.b	$F8,$B8,$FF,$1B,$6F,$FC,$2D,$34,$FE,$16,$FF,$8A,$B7,$FF,$79,$F0
	.dc.b	$7F,$95,$CF,$F4,$DE,$BF,$E2,$77,$F0,$1B,$FD,$2A,$FC,$06,$FF,$49
	.dc.b	$FA,$9C,$1F,$61,$FD,$BC,$65,$DC,$67,$2F,$DA,$C6,$2B,$A6,$1F,$7E
	.dc.b	$3F,$06,$72,$AE,$1A,$FE,$C6,$3F,$15,$87,$5A,$E1,$87,$9A,$78,$BD
	.dc.b	$A3,$35,$F9,$49,$E3,$7D,$1F,$D3,$18,$C3,$C6,$F8,$5B,$7C,$E3,$07
	.dc.b	$6A,$F2,$3F,$0A,$77,$18,$8F,$71,$B3,$B9,$B3,$07,$E9,$0F,$11,$8C
	.dc.b	$3D,$CF,$82,$BD,$C4,$9F,$64,$73,$8C,$3F,$AD,$6B,$D1,$AE,$E3,$8B
	.dc.b	$E6,$8C,$B6,$33,$87,$CA,$8E,$91,$87,$C8,$D9,$F8,$C4,$7E,$14,$F1
	.dc.b	$9F,$14,$6C,$CB,$C8,$EA,$43,$65,$3D,$05,$3C,$C5,$8F,$F0,$5B,$FC
	.dc.b	$08,$FF,$56,$07,$F8,$13,$FE,$A8,$E0,$BB,$B9,$3F,$D5,$5B,$FF,$AA
	.dc.b	$2B,$F5,$87,$F0,$43,$F8,$39,$FC,$04,$FF,$00,$9F,$EA,$5E,$2F,$F0
	.dc.b	$43,$FB,$CC,$75,$18,$8D,$E5,$FB,$F9,$CF,$17,$FA,$9E,$B5,$8F,$F0
	.dc.b	$9A,$D5,$FE,$FE,$FF,$84,$3F,$BF,$8C,$BD,$23,$33,$C5,$FF,$12,$6F
	.dc.b	$FE,$23,$23,$83,$FC,$49,$3F,$C1,$B8,$FF,$D5,$0D,$E1,$FF,$09,$72
	.dc.b	$7F,$AB,$63,$F8,$41,$F3,$37,$FB,$8F,$D4,$3E,$B9,$0C,$10,$FA,$E6
	.dc.b	$4A,$65,$E6,$E0,$8E,$08,$FE,$1C,$E4,$E3,$2F,$C8,$47,$EB,$8E,$26
	.dc.b	$CE,$C5,$32,$F0,$7F,$64,$69,$C7,$07,$EC,$9B,$C5,$FC,$26,$76,$72
	.dc.b	$CB,$F3,$73,$AE,$2F,$E0,$73,$E3,$C3,$CD,$5A,$24,$64,$F2,$46,$9C
	.dc.b	$57,$37,$F8,$25,$F8,$70,$F2,$5F,$29,$F1,$DF,$16,$9D,$09,$7C,$CF
	.dc.b	$DF,$C6,$EB,$9C,$F8,$2B,$A4,$62,$39,$D7,$19,$FD,$D4,$6C,$86,$9E
	.dc.b	$37,$E7,$26,$1F,$12,$74,$70,$BE,$C4,$9D,$C7,$C0,$9E,$13,$97,$F9
	.dc.b	$BC,$7F,$80,$4F,$F7,$F9,$F9,$53,$FD,$10,$7F,$A2,$88,$7F,$81,$0F
	.dc.b	$F4,$59,$FB,$C2,$1F,$C0,$0F,$DA,$9F,$DF,$4F,$EE,$5C,$91,$FE,$01
	.dc.b	$C1,$FB,$93,$FC,$03,$C3,$FD,$DA,$3F,$6D,$8F,$EE,$38,$7F,$7B,$39
	.dc.b	$E1,$FD,$B5,$7E,$F4,$87,$03,$F8,$51,$89,$E3,$87,$87,$4F,$F0,$0C
	.dc.b	$62,$7F,$C0,$E7,$90,$F4,$18,$8F,$F4,$7F,$07,$EF,$B9,$3F,$C0,$9C
	.dc.b	$1F,$BD,$AF,$E0,$71,$FE,$09,$D9,$E0,$B7,$FE,$0F,$5F,$C2,$27,$31
	.dc.b	$88,$FF,$57,$46,$1C,$35,$E2,$B8,$3A,$46,$9A,$E9,$AF,$76,$BD,$9E
	.dc.b	$3A,$70,$9E,$95,$FE,$BF,$8D,$F0,$FF,$B0,$74,$70,$73,$AD,$17,$15
	.dc.b	$F3,$C7,$AC,$FF,$4B,$BE,$35,$E1,$8E,$2D,$38,$BC,$EB,$C3,$09,$E2
	.dc.b	$B1,$F3,$BC,$0E,$53,$E1,$AE,$B1,$FC,$DB,$1F,$56,$3D,$66,$4F,$07
	.dc.b	$CD,$18,$8E,$D8,$FA,$9D,$1F,$A6,$C7,$C7,$8F,$8B,$1F,$04,$F8,$CE
	.dc.b	$38,$78,$9F,$F7,$56,$3E,$27,$99,$A3,$85,$73,$9F,$1C,$6A,$30,$F0
	.dc.b	$39,$9F,$17,$1F,$44,$6A,$3A,$5F,$CA,$C3,$CF,$8F,$AE,$73,$87,$BD
	.dc.b	$E7,$3A,$8F,$93,$7C,$CE,$93,$AC,$3B,$DF,$39,$CC,$FF,$81,$CF,$21
	.dc.b	$E9,$8D,$1E,$17,$91,$9A,$EF,$19,$8E,$57,$E3,$7E,$2C,$72,$9F,$99
	.dc.b	$1A,$BF,$1C,$72,$33,$8F,$CF,$9F,$79,$DC,$FA,$C7,$88,$C9,$30,$F5
	.dc.b	$C7,$17,$E1,$5F,$86,$78,$9E,$13,$A6,$3F,$82,$3E,$86,$1E,$48,$F4
	.dc.b	$E1,$E4,$7E,$29,$DE,$BA,$57,$08,$E7,$3F,$55,$EC,$6D,$F7,$3E,$2C
	.dc.b	$3C,$D5,$CA,$78,$46,$1D,$D7,$95,$FC,$6D,$73,$3C,$A7,$33,$CB,$87
	.dc.b	$C0,$8E,$27,$07,$11,$C2,$7D,$0F,$37,$B3,$FA,$7C,$3E,$14,$62,$3F
	.dc.b	$68,$76,$C3,$D0,$FE,$74,$F7,$D7,$69,$E1,$1F,$5E,$31,$7E,$E8,$E4
	.dc.b	$F9,$9E,$58,$7C,$17,$CD,$1B,$C7,$F1,$97,$E3,$BF,$BF,$5C,$8D,$9C
	.dc.b	$8C,$46,$F0,$FC,$B6,$1D,$63,$18,$7D,$DC,$3A,$BF,$00,$F9,$15,$EE
	.dc.b	$BF,$C7,$9E,$B7,$18,$78,$AB,$8C,$FE,$08,$F9,$45,$38,$9F,$1A,$78
	.dc.b	$9E,$47,$CF,$5E,$17,$89,$F5,$5F,$85,$8E,$63,$9E,$3E,$B9,$C4,$79
	.dc.b	$5F,$98,$E8,$EA,$78,$CD,$1F,$20,$D5,$FA,$9F,$01,$E2,$C3,$EC,$BF
	.dc.b	$36,$35,$18,$8F,$C2,$C6,$30,$F9,$11,$DB,$1F,$9E,$FE,$5E,$75,$7E
	.dc.b	$68,$D3,$EA,$7E,$33,$EA,$BF,$9D,$8F,$C1,$6C,$FC,$D6,$BE,$83,$D7
	.dc.b	$1B,$C3,$E7,$5F,$82,$7E,$01,$E0,$3E,$21,$E0,$3E,$25,$74,$3D,$C5
	.dc.b	$9F,$B3,$7F,$89,$BE,$CD,$AF,$24,$F8,$27,$31,$F0,$5B,$87,$EC,$AF
	.dc.b	$AD,$6E,$B8,$DF,$37,$D9,$3E,$18,$F8,$93,$B9,$CB,$67,$9C,$F1,$6C
	.dc.b	$F8,$4E,$A5,$38,$C6,$E7,$DB,$19,$9E,$35,$E7,$9C,$B6,$BC,$67,$12
	.dc.b	$D7,$19,$DC,$7E,$B0,$CC,$F1,$9F,$9F,$3E,$93,$6F,$E4,$5A,$FE,$3A
	.dc.b	$FE,$4D,$73,$8E,$2F,$85,$A7,$85,$EE,$6E,$BF,$26,$78,$4A,$E4,$FC
	.dc.b	$96,$3F,$AD,$9F,$C2,$B5,$FC,$71,$F8,$92,$9F,$89,$9E,$6F,$A6,$7C
	.dc.b	$2D,$3C,$B1,$E1,$8E,$F1,$FB,$63,$9C,$F7,$C3,$F5,$D1,$DD,$AF,$79
	.dc.b	$E3,$8F,$E7,$6F,$F9,$8C,$E4,$F4,$9D,$6B,$8D,$F5,$C7,$F6,$F3,$B3
	.dc.b	$C2,$6D,$CC,$6F,$0F,$DC,$4E,$EB,$9B,$B8,$CC,$EE,$3F,$5E,$6E,$B9
	.dc.b	$BB,$33,$1B,$BF,$4B,$6B,$F9,$0C,$FE,$62,$7E,$43,$EC,$C3,$E0,$4E
	.dc.b	$A3,$F3,$67,$E0,$4F,$9E,$FD,$D7,$86,$3F,$8A,$7E,$A5,$7D,$87,$F1
	.dc.b	$C7,$79,$E0,$FC,$4A,$EE,$FF,$23,$AE,$4F,$03,$C3,$5E,$67,$84,$69
	.dc.b	$FB,$58,$7C,$2A,$E7,$1F,$C6,$0F,$01,$E4,$BF,$BB,$5C,$63,$E5,$47
	.dc.b	$CE,$7E,$55,$FC,$86,$C7,$C3,$C7,$F7,$2F,$A9,$FA,$F7,$FC,$FE,$7F
	.dc.b	$A1,$BF,$EE,$67,$FD,$D3,$7F,$ED,$CA,$FF,$69,$E1,$FE,$9A,$AF,$F3
	.dc.b	$ED,$7F,$97,$0F,$F6,$A1,$FE,$06,$2E,$EE,$47,$2F,$FA,$A6,$53,$98
	.dc.b	$C3,$63,$FD,$B6,$0F,$F6,$BA,$E2,$FF,$4C,$C7,$FB,$70,$FF,$6D,$1F
	.dc.b	$EC,$A1,$C4,$E3,$FF,$6C,$47,$FB,$2F,$4F,$F6,$FB,$FE,$C9,$3F,$D9
	.dc.b	$0E,$32,$FF,$4C,$D7,$FB,$3D,$FF,$4D,$E1,$FE,$DA,$AE,$B7,$D1,$E2
	.dc.b	$78,$6B,$C0,$FF,$B5,$2B,$8E,$1F,$EE,$33,$FD,$D1,$3A,$9F,$14,$EB
	.dc.b	$6F,$34,$7F,$B8,$EF,$FD,$DF,$1F,$E6,$5F,$05,$DB,$F2,$E9,$18,$7E
	.dc.b	$73,$5F,$F6,$C3,$5F,$F6,$EF,$1B,$97,$FD,$CC,$7F,$B9,$F9,$7F,$DC
	.dc.b	$D3,$F4,$67,$FD,$CD,$C3,$FE,$89,$F9,$1F,$F1,$09,$38,$FF,$BB,$36
	.dc.b	$BF,$B4,$B7,$3F,$F1,$C3,$67,$FD,$42,$D9,$FF,$54,$35,$FF,$3E,$37
	.dc.b	$93,$FD,$38,$DB,$F1,$57,$80,$95,$E0,$6D,$F8,$1A,$78,$0B,$1E,$06
	.dc.b	$BE,$06,$DF,$81,$B5,$E0,$9E,$C4,$AE,$CD,$9E,$CD,$3B,$14,$EC,$5A
	.dc.b	$EC,$DA,$ED,$5D,$09,$5D,$1B,$3D,$1A,$F4,$6D,$74,$6D,$74,$AE,$44
	.dc.b	$AE,$4D,$79,$36,$B9,$36,$79,$4F,$02,$1C,$0B,$3C,$1B,$1C,$1D,$12
	.dc.b	$34,$D8,$D4,$60,$8E,$09,$3F,$EE,$0B,$FC,$23,$4F,$AE,$5A,$F9,$8D
	.dc.b	$BF,$86,$D3,$D8,$58,$F3,$B5,$F2,$36,$FC,$6D,$BF,$1C,$A9,$3F,$C6
	.dc.b	$61,$4B,$F3,$C6,$E7,$53,$97,$D3,$19,$3F,$91,$46,$5F,$59,$93,$AC
	.dc.b	$F4,$AF,$0C,$70,$8C,$57,$66,$C6,$63,$BC,$F0,$8E,$71,$83,$81,$5E
	.dc.b	$07,$8A,$38,$57,$95,$F9,$B1,$E5,$8C,$4F,$CE,$8C,$57,$89,$EE,$73
	.dc.b	$8E,$ED,$3B,$D7,$43,$BC,$F1,$AE,$B3,$F6,$0E,$B5,$E7,$7A,$D7,$18
	.dc.b	$EA,$D8,$EB,$1E,$58,$EA,$F1,$9E,$B3,$F8,$23,$A9,$F4,$8F,$BC,$7D
	.dc.b	$EB,$F9,$4F,$DC,$6B,$F6,$9A,$EE,$7F,$B2,$4A,$72,$D8,$CC,$60,$FF
	.dc.b	$04,$71,$7F,$80,$F9,$7F,$C4,$D7,$FE,$13,$3F,$C1,$27,$F8,$4E,$FF
	.dc.b	$C0,$2F,$F8,$2C,$FF,$10,$9A,$3F,$C2,$77,$FE,$0B,$E0,$FF,$09,$CF
	.dc.b	$FA,$3E,$7F,$C5,$27,$13,$FC,$36,$71,$3F,$C2,$67,$F8,$E8,$FF,$08
	.dc.b	$9E,$33,$FC,$47,$3F,$E1,$1C,$3D,$31,$C6,$FD,$2F,$F8,$84,$D9,$DE
	.dc.b	$BF,$C2,$F8,$7B,$E3,$A9,$B8,$F6,$BE,$FA,$FF,$43,$BE,$C3,$C9,$87
	.dc.b	$B2,$3F,$D1,$D8,$7F,$9D,$27,$83,$5E,$18,$FF,$8A,$EB,$84,$FF,$A3
	.dc.b	$63,$FC,$8D,$3A,$8F,$F1,$75,$78,$23,$E1,$CF,$1C,$3D,$17,$CA,$7F
	.dc.b	$CA,$93,$C9,$DD,$F2,$3C,$4E,$0F,$95,$87,$B9,$F0,$18,$C3,$C1,$8F
	.dc.b	$CA,$9F,$B1,$7D,$63,$13,$F0,$2B,$ED,$9F,$00,$A7,$43,$E6,$3D,$1F
	.dc.b	$9F,$5D,$2F,$FC,$A7,$1F,$64,$CC,$F5,$BF,$04,$62,$7E,$41,$C1,$FB
	.dc.b	$A7,$86,$37,$87,$C5,$7F,$CD,$17,$F8,$09,$ED,$39,$AF,$C5,$1F,$97
	.dc.b	$C7,$E6,$92,$76,$D7,$B9,$ED,$AF,$53,$FA,$23,$27,$CF,$3E,$0C,$FD
	.dc.b	$2C,$7E,$4B,$EA,$8F,$DF,$C7,$F9,$57,$A1,$7F,$BF,$AC,$3F,$C1,$29
	.dc.b	$58,$7F,$81,$4D,$7E,$96,$1F,$E7,$13,$9B,$A0,$E6,$BF,$CE,$91,$87
	.dc.b	$FD,$C5,$1F,$E7,$48,$EE,$7F,$91,$6B,$FC,$8C,$E0,$FF,$3A,$F3,$7F
	.dc.b	$9D,$1F,$F3,$79,$FE,$6F,$E5,$FF,$39,$3F,$C7,$1A,$FF,$5B,$7F,$CE
	.dc.b	$59,$FF,$9E,$39,$7F,$CE,$73,$B6,$BB,$6B,$FE,$65,$E4,$FF,$34,$63
	.dc.b	$FE,$5F,$8F,$F3,$49,$FE,$6C,$3F,$CD,$19,$71,$BF,$F3,$1C,$64,$FF
	.dc.b	$33,$C6,$5E,$3C,$BF,$E6,$BA,$FF,$3C,$DF,$F9,$A2,$72,$F8,$2B,$84
	.dc.b	$F8,$4D,$CF,$80,$FF,$45,$63,$FE,$7F,$BE,$95,$FD,$ED,$FF,$42,$36
	.dc.b	$FF,$CE,$51,$FE,$7B,$AF,$E8,$6D,$74,$7F,$A8,$5F,$F4,$06,$1F,$DB
	.dc.b	$D7,$6F,$F8,$1E,$30,$DD,$78,$D7,$F7,$FE,$2F,$F3,$C7,$4F,$FA,$AD
	.dc.b	$B1,$F8,$F5,$C9,$FE,$7A,$84,$69,$FE,$AF,$F2,$AF,$77,$FA,$9B,$70
	.dc.b	$FF,$5B,$36,$7F,$34,$D7,$FA,$FF,$23,$CF,$FF,$07,$E9,$FE,$10,$ED
	.dc.b	$7E,$5C,$F2,$33,$97,$F8,$7A,$EE,$2B,$DD,$62,$3F,$C4,$F3,$FD,$88
	.dc.b	$95,$FD,$99,$B1,$FE,$0D,$79,$D3,$C9,$7B,$9C,$F3,$F3,$61,$FE,$CB
	.dc.b	$CF,$FC,$3D,$F9,$D3,$E0,$7C,$7F,$34,$FC,$85,$E8,$F8,$EB,$58,$71
	.dc.b	$3F,$DA,$95,$B9,$D5,$FF,$7C,$AD,$ED,$FC,$05,$AF,$F8,$19,$AF,$F9
	.dc.b	$15,$DE,$3E,$49,$F1,$E2,$F3,$BA,$E5,$E1,$C3,$FD,$C9,$83,$E1,$37
	.dc.b	$AF,$9D,$B5,$BA,$56,$E3,$FD,$E4,$D8,$F3,$4F,$A4,$93,$C6,$75,$97
	.dc.b	$F9,$33,$4F,$8D,$E5,$77,$AE,$DB,$3C,$9F,$BB,$6F,$DC,$FE,$8E,$7F
	.dc.b	$E4,$39,$CC,$62,$7F,$C9,$31,$8E,$3F,$B0,$DC,$3A,$B6,$F9,$47,$C3
	.dc.b	$AE,$D5,$C2,$31,$7D,$5F,$F3,$6D,$F1,$AE,$F1,$AA,$E1,$7F,$E6,$F9
	.dc.b	$E7,$3E,$03,$45,$8D,$5F,$57,$FC,$EB,$7D,$67,$B9,$C2,$3C,$E7,$0B
	.dc.b	$EF,$7F,$A8,$BE,$53,$FE,$86,$BE,$4F,$4B,$E5,$5E,$88,$E5,$1E,$38
	.dc.b	$E5,$7D,$23,$F3,$D7,$D3,$0F,$A9,$1C,$8F,$2D,$F2,$3C,$B7,$CA,$3A
	.dc.b	$1C,$AF,$A4,$FF,$A7,$6F,$A6,$1E,$68,$F4,$57,$57,$D0,$75,$BF,$41
	.dc.b	$D5,$EF,$1D,$4E,$EF,$5B,$EF,$3F,$4A,$36,$66,$77,$19,$3A,$BC,$E3
	.dc.b	$71,$98,$F7,$BE,$13,$71,$CC,$DD,$FB,$27,$D5,$1E,$C3,$D4,$F2,$3B
	.dc.b	$4F,$23,$B4,$7C,$07,$B4,$6E,$3B,$46,$E7,$B4,$74,$38,$4F,$43,$84
	.dc.b	$F4,$8F,$13,$D1,$D9,$D2,$37,$3D,$27,$53,$C2,$7A,$3C,$23,$A6,$1F
	.dc.b	$84,$8F,$11,$B3,$A1,$BA,$E9,$5C,$A3,$55,$CB,$0F,$C1,$5F,$38,$F0
	.dc.b	$B6,$B8,$36,$BA,$9C,$AF,$AB,$88,$D5,$75,$AF,$A7,$18,$AD,$CE,$6B
	.dc.b	$D3,$39,$C3,$F1,$35,$A2,$38,$25,$FE,$2A,$32,$D8,$CD,$7E,$11,$B7
	.dc.b	$F6,$5B,$3F,$41,$B3,$F1,$DA,$FB,$5B,$7E,$59,$EC,$49,$EC,$DA,$EC
	.dc.b	$D3,$B1,$63,$B3,$5E,$CD,$9E,$D1,$D0,$93,$D1,$B3,$D1,$A7,$42,$9D
	.dc.b	$0B,$3D,$27,$91,$27,$93,$67,$93,$5E,$4D,$9E,$53,$C0,$93,$FC,$20
	.dc.b	$FF,$06,$E1,$FE,$E2,$24,$7F,$B9,$9B,$1F,$E6,$D9,$FC,$D3,$C2,$74
	.dc.b	$7F,$0C,$E3,$FF,$09,$FB,$DF,$C6,$A7,$F3,$C7,$E7,$E3,$33,$FC,$4A
	.dc.b	$33,$5F,$97,$9C,$C7,$F1,$23,$37,$FE,$10,$E6,$61,$D6,$FF,$3D,$5D
	.dc.b	$67,$38,$75,$9C,$9D,$63,$33,$D6,$33,$7D,$6F,$A3,$FC,$CB,$0F,$1E
	.dc.b	$3F,$C6,$AF,$C7,$8F,$F8,$16,$72,$FF,$11,$8C,$89,$FF,$1B,$63,$E0
	.dc.b	$79,$63,$E0,$AE,$53,$97,$94,$66,$B9,$41,$FF,$1D,$4E,$DF,$E2,$A7
	.dc.b	$82,$BB,$E3,$EA,$AF,$A9,$5B,$AD,$47,$F2,$7C,$3D,$31,$CB,$0F,$06
	.dc.b	$1C,$AB,$E5,$63,$F0,$F0,$E5,$5A,$9E,$51,$89,$F0,$5F,$A0,$F3,$E3
	.dc.b	$8C,$3F,$A1,$E3,$E0,$9F,$19,$FC,$B9,$C6,$3D,$63,$18,$75,$C7,$C6
	.dc.b	$7C,$8C,$3A,$3E,$26,$CF,$06,$BF,$85,$6D,$7E,$0D,$B8,$FE,$06,$BA
	.dc.b	$D7,$F6,$2C,$3C,$73,$D1,$B3,$AC,$7E,$35,$7E,$2E,$BE,$35,$F4,$AF
	.dc.b	$31,$B3,$CC,$59,$F3,$35,$F3,$37,$0F,$33,$6B,$CD,$3D,$E3,$EE,$D7
	.dc.b	$78,$FA,$18,$79,$9D,$18,$C7,$E5,$38,$C7,$D2,$7E,$53,$0F,$4E,$1D
	.dc.b	$CF,$4C,$ED,$F4,$B4,$F4,$94,$F4,$97,$1F,$4B,$67,$D3,$5D,$EB,$D2
	.dc.b	$77,$9F,$36,$3E,$B9,$F2,$CF,$E9,$2B,$CB,$5C,$E3,$CB,$3C,$E3,$EC
	.dc.b	$61,$EE,$8C,$1C,$A3,$F3,$D5,$CA,$33,$5E,$C6,$CF,$B1,$AF,$B1,$B8
	.dc.b	$7B,$1B,$5E,$CA,$F3,$1E,$CA,$F3,$1E,$C7,$F9,$07,$E7,$7F,$ED,$27
	.dc.b	$FD,$1D,$B7,$FA,$8F,$E1,$3F,$D6,$93,$9A,$FD,$54,$E6,$BF,$D1,$D3
	.dc.b	$99,$FE,$08,$7F,$05,$70,$E1,$AE,$7B,$93,$C2,$35,$1E,$03,$FD,$47
	.dc.b	$F1,$F6,$7B,$FC,$27,$FA,$9B,$E4,$3C,$AD,$8C,$9C,$EF,$C5,$1D,$9C
	.dc.b	$60,$8C,$47,$86,$31,$3C,$61,$18,$7F,$85,$18,$9F,$F5,$BF,$C0,$F8
	.dc.b	$AB,$FD,$61,$F0,$8F,$86,$FF,$5D,$FB,$9F,$ED,$D8,$C4,$FC,$EE,$C4
	.dc.b	$E6,$BC,$78,$FC,$AB,$DB,$EB,$6B,$F1,$DA,$EE,$39,$5F,$5F,$46,$CF
	.dc.b	$A3,$F3,$96,$DF,$EB,$2F,$96,$FF,$78,$46,$2F,$FD,$53,$18,$9F,$1D
	.dc.b	$7F,$3F,$AD,$3F,$35,$AF,$D9,$6B,$F8,$2E,$57,$A5,$FE,$F3,$3F,$DF
	.dc.b	$D1,$FD,$27,$D1,$FE,$95,$E7,$BC,$EE,$15,$BC,$FF,$D9,$AF,$0B,$5F
	.dc.b	$0A,$DD,$7F,$B7,$E7,$A0,$BE,$8F,$F5,$7C,$3C,$73,$C6,$3F,$21,$75
	.dc.b	$4E,$9E,$57,$F8,$B9,$D4,$E6,$FA,$D7,$E8,$27,$38,$7E,$3E,$73,$3E
	.dc.b	$47,$75,$C6,$FA,$4F,$63,$CD,$5C,$63,$35,$C2,$31,$3E,$83,$81,$E8
	.dc.b	$72,$F8,$CC,$D7,$77,$84,$F3,$C3,$E0,$47,$96,$FE,$5B,$D0,$C1,$E4
	.dc.b	$8D,$BF,$48,$EB,$7E,$07,$67,$8E,$37,$87,$EA,$E3,$EA,$9F,$AD,$70
	.dc.b	$F3,$3F,$81,$79,$EF,$45,$EB,$7E,$CE,$30,$D8,$CC,$FE,$FE,$72,$DF
	.dc.b	$4B,$FC,$40,$DE,$DE,$51,$88,$EE,$D7,$F8,$CC,$A3,$8E,$9F,$7B,$D1
	.dc.b	$7A,$5F,$C7,$A3,$0D,$DB,$8D,$70,$6E,$3F,$8D,$BE,$38,$DB,$E0,$FF
	.dc.b	$8A,$F1,$F0,$57,$E4,$DB,$1F,$7A,$B8,$B6,$B8,$A9,$DE,$3F,$94,$6E
	.dc.b	$1F,$CD,$2F,$8C,$FA,$2B,$B6,$3E,$77,$C4,$D5,$3C,$2B,$96,$1E,$0C
	.dc.b	$3A,$D7,$0C,$7D,$8C,$3F,$2B,$1A,$7C,$D8,$75,$C3,$A4,$7F,$92,$DA
	.dc.b	$EA,$B9,$1F,$CD,$A3,$E0,$B6,$FF,$A7,$BF,$CE,$AF,$B4,$EB,$5F,$81
	.dc.b	$D9,$A6,$E1,$FE,$59,$BE,$4D,$8F,$EB,$3F,$12,$F1,$BF,$30,$D9,$FE
	.dc.b	$88,$D7,$FA,$73,$5F,$EB,$5E,$8B,$9D,$E9,$FF,$4B,$8C,$37,$D3,$78
	.dc.b	$F9,$AD,$DF,$FA,$77,$C8,$FF,$47,$7A,$3F,$D6,$75,$FE,$AF,$D2,$9C
	.dc.b	$ED,$E9,$BF,$E3,$D1,$87,$FA,$BC,$EE,$7F,$CF,$B7,$CD,$FE,$6D,$5F
	.dc.b	$CF,$AB,$F9,$0B,$6B,$84,$78,$E7,$FB,$27,$96,$F3,$F9,$4B,$C7,$78
	.dc.b	$6F,$43,$32,$BE,$18,$7F,$61,$9C,$C6,$27,$FC,$1B,$5C,$6B,$D9,$1C
	.dc.b	$70,$EF,$87,$6C,$38,$C7,$A7,$0E,$55,$FE,$0B,$6B,$FE,$14,$6B,$FE
	.dc.b	$99,$5E,$77,$4C,$5E,$7F,$92,$30,$DC,$BF,$BC,$36,$FC,$ED,$7C,$13
	.dc.b	$E3,$24,$FF,$B5,$23,$FD,$25,$B3,$CF,$F9,$AF,$FB,$97,$CD,$F9,$6E
	.dc.b	$77,$F9,$3B,$FF,$BA,$2F,$FD,$DB,$87,$FB,$A1,$FF,$76,$B6,$35,$18
	.dc.b	$36,$43,$04,$76,$4B,$F0,$E1,$E4,$AE,$91,$8B,$F1,$B6,$3E,$AD,$71
	.dc.b	$7F,$25,$87,$A5,$CC,$EF,$0F,$C9,$B6,$7E,$04,$F0,$C3,$C1,$5E,$7B
	.dc.b	$E8,$FC,$88,$CB,$E9,$BE,$55,$E6,$39,$61,$E6,$69,$E2,$2C,$F8,$9B
	.dc.b	$87,$8A,$3B,$4F,$16,$CF,$16,$9C,$4B,$3C,$4F,$DB,$5F,$9B,$0F,$4C
	.dc.b	$62,$FA,$36,$FA,$36,$FA,$34,$E9,$1F,$C5,$EF,$A6,$1F,$02,$BE,$33
	.dc.b	$5F,$89,$3E,$C3,$E2,$3E,$66,$D7,$56,$9D,$6F,$CA,$D3,$CA,$59,$F2
	.dc.b	$B6,$7C,$B0,$7F,$00,$E8,$C5,$FA,$27,$C9,$1E,$86,$BE,$83,$CC,$64
	.dc.b	$A6,$67,$D0,$E5,$AE,$63,$C8,$7B,$1F,$0D,$7B,$1B,$1E,$C6,$9E,$79
	.dc.b	$F4,$9E,$C2,$C7,$36,$C7,$86,$3B,$B4,$E7,$1C,$6B,$E0,$47,$AB,$0F
	.dc.b	$C1,$D7,$08,$F4,$1C,$1E,$D3,$C2,$BD,$B5,$C2,$B4,$77,$8F,$1B,$C9
	.dc.b	$B1,$B6,$E1,$F8,$58,$C5,$75,$BF,$14,$63,$0F,$8D,$7C,$27,$EA,$CE
	.dc.b	$5F,$91,$19,$C3,$D9,$87,$52,$4F,$06,$C6,$A3,$04,$70,$49,$F7,$CF
	.dc.b	$06,$CE,$DB,$3B,$8F,$4C,$ED,$A6,$CB,$1B,$3E,$49,$F1,$F0,$F8,$4E
	.dc.b	$2F,$DC,$FE,$0E,$33,$7E,$F3,$C2,$4A,$EA,$D9,$E5,$7E,$36,$CF,$8D
	.dc.b	$A7,$8E,$7A,$47,$22,$46,$A3,$A1,$A7,$11,$A8,$C1,$2B,$CE,$4C,$3E
	.dc.b	$DB,$67,$EC,$B6,$BD,$D3,$A2,$4E,$9B,$1A,$8C,$11,$C1,$2B,$F4,$13
	.dc.b	$C1,$B3,$B6,$CE,$E3,$F3,$F3,$B6,$9B,$3F,$D9,$57,$FE,$13,$23,$FD
	.dc.b	$AE,$46,$9F,$DE,$71,$3C,$E7,$9F,$FC,$06,$BF,$D7,$B7,$FC,$0E,$B5
	.dc.b	$39,$CD,$39,$BE,$38,$7F,$07,$C3,$8B,$DA,$33,$1F,$C7,$CC,$9F,$C8
	.dc.b	$70,$EB,$3C,$6F,$AB,$C7,$0F,$06,$1D,$DB,$19,$38,$46,$5A,$66,$FC
	.dc.b	$15,$D1,$F0,$61,$E4,$CB,$FC,$8F,$7D,$27,$F9,$B5,$FC,$8C,$7C,$B1
	.dc.b	$CA,$7B,$36,$7B,$57,$27,$C3,$1C,$8F,$F2,$CD,$F7,$8F,$A5,$1C,$4D
	.dc.b	$5F,$6A,$E5,$19,$53,$E3,$AE,$53,$FE,$F7,$BF,$45,$6A,$3B,$E3,$8B
	.dc.b	$EF,$39,$C3,$B4,$60,$95,$CD,$FF,$33,$46,$A3,$07,$8E,$3C,$55,$D6
	.dc.b	$33,$5D,$70,$F2,$C7,$DE,$8F,$59,$2B,$89,$F1,$9C,$47,$A2,$30,$79
	.dc.b	$2F,$C5,$1C,$E3,$CE,$D3,$81,$E5,$3C,$6F,$67,$F2,$1E,$6B,$AF,$FC
	.dc.b	$B5,$E7,$FF,$9C,$23,$15,$FE,$85,$E5,$FF,$3A,$CE,$6B,$73,$FE,$6E
	.dc.b	$AD,$DF,$F9,$CA,$B5,$3F,$AE,$88,$27,$6E,$1A,$ED,$AF,$ED,$97,$9D
	.dc.b	$FE,$6F,$F3,$FC,$CD,$AF,$13,$6B,$F1,$C7,$77,$19,$FE,$0E,$31,$3E
	.dc.b	$47,$B4,$FF,$A6,$EF,$BC,$E6,$76,$6F,$CD,$79,$BF,$E8,$1F,$3B,$FD
	.dc.b	$2F,$18,$AF,$91,$CE,$9C,$D6,$E7,$FD,$2B,$5B,$BF,$F4,$CD,$6A,$7C
	.dc.b	$13,$99,$F9,$2D,$7E,$63,$5F,$E3,$7E,$63,$CD,$FF,$39,$F9,$DF,$EA
	.dc.b	$A8,$C5,$FF,$99,$E3,$13,$A8,$FF,$53,$CE,$AB,$FC,$D1,$5F,$E8,$AF
	.dc.b	$3A,$57,$FA,$B2,$7F,$94,$F9,$DF,$CA,$BC,$C7,$99,$F7,$FA,$98,$70
	.dc.b	$BF,$F6,$5B,$FE,$BC,$78,$8A,$E2,$FF,$34,$9E,$38,$77,$86,$1D,$F0
	.dc.b	$ED,$87,$EF,$70,$E3,$5F,$8F,$AD,$3C,$E7,$55,$F7,$E7,$32,$C3,$BB
	.dc.b	$E3,$9D,$34,$E5,$87,$07,$D3,$87,$82,$B8,$BE,$28,$CD,$79,$AB,$84
	.dc.b	$7A,$8E,$87,$87,$0F,$2B,$FA,$6C,$3C,$B8,$74,$BF,$1E,$1D,$EF,$77
	.dc.b	$EF,$39,$D6,$DE,$46,$23,$6F,$90,$F5,$9E,$23,$07,$8F,$0F,$3E,$1E
	.dc.b	$29,$E3,$3E,$93,$F4,$91,$FA,$57,$ED,$98,$3F,$59,$E5,$3C,$D7,$9D
	.dc.b	$FA,$B8,$C3,$63,$33,$C1,$AF,$F0,$89,$79,$DF,$E8,$78,$EA,$D7,$4D
	.dc.b	$7F,$88,$47,$0E,$6F,$04,$F1,$D3,$F0,$FE,$6B,$CE,$FE,$2D,$18,$6E
	.dc.b	$FC,$AB,$83,$6F,$F1,$37,$C7,$1B,$7C,$1F,$F0,$6E,$3E,$17,$EF,$B7
	.dc.b	$1E,$D1,$C0,$FC,$9C,$67,$0E,$D1,$F8,$6B,$EC,$DB,$EC,$AB,$83,$E0
	.dc.b	$C7,$CE,$FF,$87,$AB,$8E,$1E,$08,$F1,$CE,$9E,$4D,$9D,$E1,$FC,$E5
	.dc.b	$B3,$D7,$1F,$3D,$6F,$1F,$1D,$6A,$B9,$D7,$B5,$FE,$67,$1E,$A8,$FE
	.dc.b	$64,$DD,$BD,$F8,$F9,$67,$3D,$7F,$D2,$5B,$7F,$E3,$87,$8E,$19,$BF
	.dc.b	$E9,$DE,$B3,$DD,$C5,$FF,$8F,$BE,$F9,$F1,$B8,$84,$61,$E9,$7C,$AB
	.dc.b	$57,$CE,$72,$73,$BE,$57,$C6,$3F,$A7,$CE,$DF,$EA,$3E,$83,$CE,$7A
	.dc.b	$1F,$D2,$E3,$0D,$F4,$1E,$2E,$6B,$77,$FD,$57,$D6,$7A,$FF,$D6,$63
	.dc.b	$FB,$64,$63,$6F,$EA,$CD,$8E,$13,$B7,$FC,$BF,$F3,$EF,$88,$C4,$23
	.dc.b	$17,$FD,$7D,$B1,$FD,$95,$AF,$F7,$AF,$CF,$7C,$F3,$FD,$95,$B1,$FD
	.dc.b	$AD,$AE,$F2,$EF,$1A,$E3,$FE,$E4,$DA,$FF,$39,$F9,$4F,$19,$F2,$9F
	.dc.b	$E3,$9B,$DE,$16,$FF,$C1,$98,$71,$C7,$FC,$F6,$DC,$79,$61,$B9,$CD
	.dc.b	$D5,$EA,$BD,$6F,$86,$DC,$7F,$C0,$0D,$9F,$F0,$33,$5F,$F0,$DB,$9C
	.dc.b	$BF,$C1,$8D,$9F,$5E,$6F,$95,$F0,$4E,$6F,$FD,$22,$D7,$C7,$87,$9F
	.dc.b	$3F,$85,$5C,$23,$D1,$E2,$3D,$2F,$66,$9F,$E9,$79,$CC,$63,$D0,$E9
	.dc.b	$5B,$D3,$FC,$B3,$18,$9D,$57,$09,$CB,$72,$FF,$38,$D6,$E7,$27,$E8
	.dc.b	$A3,$FB,$45,$72,$8C,$5F,$79,$E3,$7C,$F5,$FF,$31,$36,$BF,$CF,$0D
	.dc.b	$9F,$F4,$1B,$71,$F6,$4F,$4C,$3B,$57,$F8,$F3,$0E,$D8,$74,$AD,$1E
	.dc.b	$69,$D5,$70,$9F,$F2,$05,$70,$AF,$09,$C2,$BC,$ED,$9E,$91,$CA,$3A
	.dc.b	$47,$2A,$E9,$1F,$E1,$F6,$CF,$F8,$65,$AF,$C6,$8E,$26,$CA,$ED,$AE
	.dc.b	$CF,$19,$E1,$2C,$F8,$63,$75,$EB,$9F,$93,$1E,$93,$DA,$F1,$9F,$59
	.dc.b	$C5,$E7,$1E,$36,$9E,$32,$9E,$38,$E3,$1B,$6C,$6D,$AE,$E3,$C3,$1D
	.dc.b	$CC,$CF,$9E,$FE,$ED,$79,$6B,$67,$52,$BD,$5A,$F5,$3D,$E7,$12,$D7
	.dc.b	$78,$DD,$77,$C7,$F3,$55,$EC,$8C,$63,$F8,$F9,$CD,$F9,$B1,$FB,$35
	.dc.b	$F9,$BB,$FF,$4C,$B5,$F9,$06,$23,$F9,$5C,$7D,$38,$F3,$62,$F7,$1E
	.dc.b	$B7,$E7,$E7,$D4,$6A,$33,$9F,$D9,$8F,$80,$4A,$E5,$5C,$6F,$B1,$F2
	.dc.b	$EF,$8D,$7C,$33,$AC,$66,$78,$61,$FA,$DA,$DE,$3F,$1E,$FB,$9E,$3A
	.dc.b	$DE,$3E,$3C,$7C,$2F,$96,$7A,$E1,$FC,$4B,$0E,$73,$F6,$67,$9E,$3E
	.dc.b	$18,$E0,$FA,$2B,$86,$1D,$AB,$ED,$61,$D9,$F5,$56,$EF,$FD,$E3,$87
	.dc.b	$69,$EB,$87,$AF,$0E,$58,$FA,$F1,$E4,$FB,$6B,$CF,$87,$2A,$DB,$E7
	.dc.b	$AD,$BE,$99,$D8,$A4,$23,$FB,$BE,$1E,$37,$73,$E1,$8C,$63,$E9,$9F
	.dc.b	$6C,$63,$1F,$0C,$72,$C6,$BF,$47,$0F,$8F,$3E,$D6,$9F,$4E,$B9,$38
	.dc.b	$9F,$33,$E5,$C7,$D7,$1F,$1A,$38,$5F,$6C,$3E,$A4,$FC,$EC,$3B,$36
	.dc.b	$B6,$75,$C7,$D9,$7C,$6B,$EC,$CE,$4F,$95,$8F,$92,$F9,$1E,$43,$DD
	.dc.b	$1B,$7F,$39,$1B,$BF,$AD,$87,$91,$B7,$E4,$6D,$F9,$2F,$AC,$79,$23
	.dc.b	$A4,$7C,$93,$2F,$93,$0F,$43,$F8,$13,$77,$DD,$FC,$03,$F5,$EB,$EF
	.dc.b	$D7,$A0,$E6,$7A,$2B,$9B,$ED,$C7,$E7,$BF,$82,$BF,$01,$DB,$0F,$8F
	.dc.b	$1C,$8F,$35,$FB,$1F,$C2,$DF,$48,$F0,$DF,$4C,$3C,$98,$74,$BE,$51
	.dc.b	$E9,$7E,$F1,$CA,$3C,$07,$BD,$DE,$1E,$2B,$F0,$1F,$26,$33,$3B,$8C
	.dc.b	$BD,$DB,$1F,$06,$3A,$BF,$07,$1F,$B5,$1F,$2A,$BB,$34,$FB,$15,$CA
	.dc.b	$BE,$0C,$65,$F4,$1E,$23,$A6,$3F,$0E,$3F,$39,$7D,$A7,$DF,$5F,$1A
	.dc.b	$31,$1F,$44,$FB,$11,$DF,$0F,$8D,$19,$3C,$27,$2A,$F7,$1E,$39,$F8
	.dc.b	$87,$1C,$3D,$73,$B3,$E1,$1E,$78,$F1,$61,$E7,$8F,$0D,$73,$3A,$BC
	.dc.b	$4F,$09,$E3,$BF,$29,$F7,$9E,$B5,$E1,$9F,$A8,$73,$8C,$B5,$E7,$3C
	.dc.b	$5B,$7F,$C3,$B1,$FC,$53,$71,$F8,$F5,$DE,$73,$7F,$B5,$6D,$FE,$81
	.dc.b	$B8,$7E,$09,$B3,$E3,$6C,$FC,$56,$BF,$4D,$B5,$DE,$FC,$38,$7E,$F6
	.dc.b	$FD,$53,$F0,$A3,$CD,$8F,$CB,$BE,$31,$BC,$3E,$55,$6E,$FD,$18,$74
	.dc.b	$7F,$1E,$7F,$3B,$FC,$F7,$FB,$2D,$7F,$39,$3F,$9D,$38,$70,$D7,$38
	.dc.b	$FF,$A9,$A3,$F9,$F1,$FD,$47,$6F,$F4,$B7,$C2,$FF,$A2,$E3,$33,$4F
	.dc.b	$F3,$D7,$0A,$71,$87,$F9,$A6,$9E,$6F,$52,$47,$97,$1F,$1C,$61,$FE
	.dc.b	$54,$27,$8C,$FF,$B5,$31,$FF,$54,$78,$1E,$3B,$93,$DD,$F1,$C6,$2F
	.dc.b	$FC,$D5,$EB,$7F,$86,$09,$5E,$18,$FE,$BD,$AA,$73,$1C,$27,$FD,$17
	.dc.b	$EB,$7F,$A5,$26,$47,$FA,$09,$FF,$49,$46,$27,$D3,$E8,$3D,$2F,$F6
	.dc.b	$8B,$6F,$C1,$28,$C3,$E3,$6B,$F1,$5C,$F1,$FF,$36,$8E,$15,$F4,$3B
	.dc.b	$BF,$D7,$D5,$AB,$FC,$06,$4F,$3F,$FD,$56,$7F,$B0,$F0,$FE,$F9,$E7
	.dc.b	$FF,$7D,$F2,$5E,$67,$0A,$DE,$1F,$E9,$E7,$85,$AF,$85,$6E,$BF,$D4
	.dc.b	$D3,$D0,$5F,$47,$E6,$DF,$F4,$E8,$C4,$7E,$56,$F9,$BA,$AA,$E4,$E7
	.dc.b	$7D,$DA,$7E,$26,$77,$87,$BE,$B8,$CF,$8E,$FB,$CE,$63,$89,$D6,$33
	.dc.b	$87,$97,$0F,$C2,$61,$D2,$31,$5D,$CE,$37,$CE,$FD,$71,$DE,$FC,$D1
	.dc.b	$88,$F9,$07,$0A,$E6,$F8,$67,$67,$81,$D6,$1E,$A7,$8D,$79,$A3,$18
	.dc.b	$79,$5B,$3F,$8C,$3B,$9F,$8E,$8C,$38,$3F,$82,$79,$AF,$39,$E7,$FE
	.dc.b	$52,$30,$D8,$CC,$FE,$9E,$72,$DF,$3F,$FC,$90,$DE,$0E,$51,$88,$FD
	.dc.b	$F4,$A3,$11,$E2,$78,$E5,$F8,$4F,$3D,$E7,$7F,$19,$8C,$37,$5E,$18
	.dc.b	$FE,$1B,$1E,$8F,$E2,$2F,$8E,$36,$F8,$3F,$E5,$BA,$DB,$F8,$26,$E1
	.dc.b	$DA,$38,$1D,$71,$F2,$E3,$E0,$8F,$C4,$E1,$D9,$B8,$76,$58,$74,$AE
	.dc.b	$18,$75,$9D,$9D,$71,$8E,$09,$1F,$A6,$AE,$51,$E9,$C7,$C6,$FF,$9D
	.dc.b	$71,$F1,$57,$53,$84,$EA,$53,$E7,$BF,$E1,$4D,$D7,$FA,$33,$67,$B4
	.dc.b	$FC,$27,$F9,$9C,$3C,$FF,$E9,$93,$C9,$B5,$CB,$3B,$7F,$E8,$9F,$35
	.dc.b	$CF,$F9,$56,$C7,$F2,$26,$BF,$CC,$DA,$FF,$45,$73,$87,$F2,$1F,$39
	.dc.b	$E7,$F6,$A4,$61,$FF,$4C,$4F,$F2,$7A,$E3,$8F,$F6,$3F,$6D,$EE,$7E
	.dc.b	$9D,$B7,$FC,$D1,$AF,$F4,$56,$BF,$D5,$FD,$17,$4B,$D1,$FE,$77,$18
	.dc.b	$6F,$A2,$F1,$F3,$5B,$BF,$EA,$1E,$8B,$D1,$FE,$79,$1F,$DD,$E3,$1A
	.dc.b	$7F,$46,$6C,$70,$9D,$BF,$EB,$5F,$3F,$B5,$7F,$65,$FC,$9F,$C8,$FC
	.dc.b	$23,$FC,$05,$1A,$9F,$F4,$B7,$8A,$AD,$C7,$F7,$09,$D3,$FD,$D5,$AF
	.dc.b	$2C,$BC,$2A,$32,$F8,$E3,$FD,$17,$C3,$FD,$5E,$7C,$0F,$FA,$3A,$DE
	.dc.b	$43,$D0,$DE,$5D,$4F,$46,$3F,$ED,$1F,$41,$E7,$FF,$83,$BC,$6D,$CE
	.dc.b	$63,$15,$C2,$15,$C2,$3F,$B7,$B6,$3F,$C0,$0D,$7F,$C2,$4D,$7F,$C5
	.dc.b	$0E,$F2,$FE,$E0,$D9,$FF,$72,$65,$EA,$F4,$3C,$77,$E0,$C2,$F8,$CF
	.dc.b	$3F,$CF,$BB,$D0,$E5,$7F,$E1,$09,$CC,$62,$15,$C7,$2F,$87,$7C,$70
	.dc.b	$EB,$87,$06,$CE,$B0,$E5,$7C,$70,$FF,$5C,$46,$2F,$85,$7C,$7A,$D6
	.dc.b	$9F,$EB,$2A,$D4,$F3,$C3,$AE,$1C,$9B,$5B,$6C,$FF,$BC,$EB,$8D,$6E
	.dc.b	$3B,$61,$E0,$8F,$F1,$EE,$1E,$09,$E1,$7C,$A7,$F5,$17,$CB,$0E,$D5
	.dc.b	$C8,$F2,$16,$7B,$1D,$30,$ED,$3F,$E2,$7C,$3B,$4F,$4C,$3B,$63,$F2
	.dc.b	$F0,$E5,$7D,$AB,$94,$76,$69,$DA,$3A,$61,$DA,$7F,$C5,$ED,$9F,$B9
	.dc.b	$1C,$67,$6D,$36,$58,$D9,$D5,$CB,$4C,$C7,$94,$E6,$E4,$FC,$3C,$F8
	.dc.b	$5B,$3D,$5A,$6F,$0F,$87,$7F,$06,$3D,$B7,$D6,$BC,$2D,$8F,$0C,$7D
	.dc.b	$63,$E3,$95,$E2,$D8,$F2,$9C,$63,$C2,$DA,$F0,$E3,$F7,$6B,$81,$E0
	.dc.b	$2D,$78,$0F,$B0,$F8,$1A,$78,$0B,$5E,$08,$E3,$1E,$06,$D7,$83,$1F
	.dc.b	$C0,$DF,$86,$B8,$D7,$85,$A7,$87,$0F,$8D,$87,$58,$C9,$D4,$B8,$75
	.dc.b	$AF,$35,$70,$7A,$36,$7A,$5F,$86,$31,$1C,$DB,$5C,$E3,$DD,$18,$C7
	.dc.b	$EF,$4E,$70,$F2,$4E,$70,$FC,$04,$E5,$F0,$E1,$DC,$98,$FC,$22,$47
	.dc.b	$A2,$32,$D8,$CE,$3E,$29,$CB,$63,$38,$78,$A7,$2D,$8C,$D7,$89,$B5
	.dc.b	$F0,$9A,$FC,$1A,$E3,$3B,$3C,$44,$8F,$03,$6B,$B3,$4E,$11,$A2,$46
	.dc.b	$9B,$1A,$8C,$11,$C1,$23,$CC,$F5,$69,$C8,$B1,$C5,$B1,$C5,$F6,$57
	.dc.b	$BA,$FE,$A4,$68,$86,$A7,$B3,$FB,$83,$FD,$CD,$1F,$CE,$DB,$8F,$F7
	.dc.b	$28,$C4,$7F,$08,$33,$3F,$EC,$E8,$CC,$7F,$0E,$32,$7E,$33,$D0,$7A
	.dc.b	$1F,$CC,$63,$F3,$87,$E4,$E3,$35,$FB,$18,$CD,$7F,$0A,$9C,$C7,$EC
	.dc.b	$CC,$DF,$F6,$08,$CE,$3F,$A9,$9C,$DF,$F5,$C9,$CB,$FB,$18,$C8,$94
	.dc.b	$23,$FC,$59,$3C,$1F,$E0,$C7,$09,$DB,$AC,$3D,$98,$7F,$42,$8F,$64
	.dc.b	$63,$0F,$E8,$77,$CE,$79,$61,$FB,$88,$C5,$F1,$25,$7B,$6F,$95,$7F
	.dc.b	$00,$AD,$CE,$2F,$D5,$3E,$9C,$D5,$A8,$F9,$98,$7C,$BC,$38,$63,$FB
	.dc.b	$CA,$D6,$1F,$36,$B5,$7C,$31,$F4,$E1,$C2,$3D,$0D,$AF,$E3,$57,$CB
	.dc.b	$0F,$A5,$87,$29,$E9,$87,$2C,$3A,$1E,$BC,$39,$4F,$C8,$8C,$BF,$89
	.dc.b	$8C,$D7,$3B,$EB,$7C,$EB,$AD,$F3,$38,$CF,$B6,$F8,$D7,$CB,$9C,$9E
	.dc.b	$A2,$D7,$AA,$77,$1F,$89,$6C,$7C,$E3,$F0,$11,$C1,$F4,$1E,$23,$D0
	.dc.b	$E6,$FD,$0F,$E4,$5F,$BA,$F5,$35,$5E,$59,$E5,$7E,$99,$C9,$B8,$CB
	.dc.b	$E9,$69,$E9,$76,$F9,$F1,$FC,$1E,$1E,$AA,$FC,$CC,$62,$B7,$18,$C7
	.dc.b	$ED,$47,$9F,$0F,$24,$FD,$08,$F9,$D8,$78,$0F,$83,$19,$7E,$09,$9C
	.dc.b	$7E,$0B,$FA,$A7,$E0,$BE,$5B,$EE,$E6,$FB,$DF,$84,$EF,$7D,$4F,$89
	.dc.b	$3D,$67,$6F,$19,$D9,$C5,$F1,$CF,$18,$F9,$0D,$8F,$90,$73,$8F,$9D
	.dc.b	$19,$8F,$6C,$73,$8F,$09,$9B,$F6,$CF,$EF,$0E,$71,$FC,$4F,$CF,$FE
	.dc.b	$15,$B7,$FA,$53,$E0,$BF,$CF,$71,$8E,$34,$63,$5F,$F5,$0C,$FF,$20
	.dc.b	$C3,$F8,$A1,$FC,$55,$C3,$86,$B9,$8E,$59,$7F,$A7,$B0,$FF,$51,$FA
	.dc.b	$38,$AF,$CD,$0B,$7A,$3F,$EA,$0A,$FF,$54,$FC,$37,$9A,$31,$9A,$31
	.dc.b	$8F,$93,$0F,$15,$7A,$09,$19,$8F,$44,$70,$7F,$D8,$B5,$FE,$F8,$9F
	.dc.b	$F5,$66,$1E,$27,$FD,$11,$E9,$7F,$AC,$B5,$7B,$9F,$C9,$63,$0D,$9F
	.dc.b	$EA,$51,$FD,$5C,$8E,$0E,$67,$FA,$67,$D3,$FF,$54,$C7,$FA,$B3,$D3
	.dc.b	$74,$FB,$BD,$8C,$36,$FF,$D8,$13,$A7,$D9,$A2,$30,$FB,$DA,$FC,$56
	.dc.b	$BF,$6B,$8D,$EA,$7F,$AE,$8F,$F5,$EC,$7F,$51,$F5,$3F,$A9,$7A,$0F
	.dc.b	$4B,$85,$6F,$1F,$F5,$63,$C2,$D7,$C2,$B7,$5F,$ED,$79,$E8,$2F,$A3
	.dc.b	$FD,$7F,$0F,$1C,$F1,$8F,$B9,$75,$FF,$63,$E1,$FD,$82,$F8,$C7,$E5
	.dc.b	$27,$57,$CA,$3B,$61,$F8,$19,$CD,$F1,$8C,$4E,$8F,$14,$65,$F0,$99
	.dc.b	$3C,$38,$78,$DF,$5E,$1E,$2C,$3C,$0F,$92,$33,$7F,$09,$E4,$60,$EF
	.dc.b	$5C,$DF,$92,$77,$AE,$73,$E6,$AE,$D7,$E5,$AF,$80,$6D,$C3,$D4,$F3
	.dc.b	$9E,$03,$AC,$72,$BF,$06,$1F,$8A,$9F,$AA,$7E,$31,$C3,$E4,$3F,$5D
	.dc.b	$E9,$3D,$37,$A9,$F9,$28,$C3,$63,$33,$FB,$19,$CB,$7D,$4F,$F1,$03
	.dc.b	$78,$F9,$46,$23,$BB,$5F,$E2,$32,$8E,$39,$7E,$0B,$D4,$7A,$7F,$C5
	.dc.b	$A3,$0D,$DB,$8D,$70,$6E,$3F,$8D,$BE,$38,$DB,$E0,$FF,$8C,$F1,$F0
	.dc.b	$57,$E4,$DB,$1F,$85,$AE,$2D,$AE,$2A,$77,$8F,$E6,$1B,$87,$F2,$CB
	.dc.b	$E3,$3E,$8A,$ED,$8F,$9D,$F1,$35,$4F,$0A,$E5,$87,$83,$0E,$B5,$C3
	.dc.b	$1F,$63,$0F,$CE,$46,$9F,$36,$1D,$70,$E9,$1F,$E5,$36,$BA,$AE,$47
	.dc.b	$F9,$49,$F1,$DF,$F4,$68,$C4,$FC,$47,$57,$DE,$31,$AF,$C1,$F5,$38
	.dc.b	$37,$0F,$F3,$15,$F2,$C3,$31,$FE,$65,$F4,$5E,$A6,$21,$18,$78,$DF
	.dc.b	$F5,$88,$C4,$7F,$50,$F4,$DF,$84,$71,$9C,$CF,$F4,$B6,$BC,$6F,$F4
	.dc.b	$D1,$88,$FF,$3B,$7A,$0E,$AF,$D5,$36,$3F,$9A,$B5,$FE,$A4,$D7,$FA
	.dc.b	$C7,$A6,$F5,$3F,$64,$D8,$FE,$7A,$D7,$FA,$D3,$5F,$EC,$1E,$A3,$D3
	.dc.b	$7A,$9F,$D1,$23,$0D,$C1,$E5,$E6,$B7,$7F,$D9,$7D,$1C,$F0,$7F,$77
	.dc.b	$F4,$B9,$BF,$2A,$7F,$D2,$93,$A7,$FC,$A5,$E4,$2B,$71,$FD,$F2,$75
	.dc.b	$C9,$FE,$96,$AD,$67,$FE,$98,$8F,$C4,$4E,$B0,$FF,$50,$56,$9F,$05
	.dc.b	$A7,$C0,$FF,$7F,$6B,$FE,$14,$6B,$E6,$CB,$F2,$8F,$F8,$63,$D1,$79
	.dc.b	$FF,$E0,$AF,$0F,$C5,$86,$63,$0F,$82,$76,$FF,$86,$7C,$8F,$ED,$B9
	.dc.b	$FE,$7E,$73,$87,$F8,$2B,$0E,$6F,$FA,$E6,$32,$FF,$85,$5A,$F7,$BF
	.dc.b	$EF,$31,$88,$FF,$72,$F8,$7F,$EE,$87,$FD,$CC,$7F,$BC,$4B,$7F,$EF
	.dc.b	$16,$D7,$FB,$C6,$78,$13,$CE,$F1,$E6,$F4,$3C,$F8,$71,$AD,$5F,$DD
	.dc.b	$AD,$69,$FE,$F9,$C3,$8E,$1C,$2B,$B4,$62,$75,$1C,$67,$58,$FA,$AB
	.dc.b	$53,$FE,$FE,$C3,$BC,$FF,$A0,$EB,$6F,$A8,$E1,$3C,$EF,$AC,$6D,$B8
	.dc.b	$FF,$A2,$2B,$8D,$F4,$8C,$4E,$AF,$FD,$A5,$5A,$C3,$C7,$87,$CB,$C3
	.dc.b	$85,$6A,$F8,$4F,$E9,$AF,$84,$F8,$CF,$55,$F2,$C3,$E5,$36,$3E,$29
	.dc.b	$BA,$ED,$19,$7D,$A7,$AE,$39,$9B,$9E,$6D,$39,$C6,$CE,$6E,$5E,$71
	.dc.b	$F0,$23,$9C,$6E,$39,$B6,$39,$C6,$E3,$99,$93,$9C,$7C,$33,$9C,$6C
	.dc.b	$E6,$57,$99,$B8,$E7,$19,$79,$9F,$2E,$FE,$6E,$3F,$8E,$AE,$A7,$1A
	.dc.b	$EA,$D3,$AD,$71,$76,$D3,$BB,$B7,$BC,$F5,$72,$6E,$7A,$B6,$7A,$CE
	.dc.b	$E7,$2D,$8C,$E1,$F8,$2A,$F8,$C7,$E5,$C9,$8F,$E2,$9E,$87,$8A,$38
	.dc.b	$5F,$C3,$8D,$E3,$F8,$36,$DF,$DB,$69,$DA,$FA,$61,$E3,$8E,$98,$FC
	.dc.b	$AC,$7A,$DF,$C1,$9E,$38,$F7,$C3,$8D,$6A,$7C,$A6,$AB,$84,$EA,$FA
	.dc.b	$12,$3E,$7E,$1E,$1A,$E0,$72,$AE,$13,$AB,$ED,$5B,$9C,$D7,$D2,$8C
	.dc.b	$63,$F0,$67,$38,$FD,$0B,$E2,$4C,$7C,$58,$71,$8C,$63,$DE,$73,$5C
	.dc.b	$5B,$1F,$61,$B5,$F2,$23,$81,$1E,$91,$A7,$DB,$8F,$9A,$7E,$A6,$1E
	.dc.b	$67,$A1,$C6,$FC,$0D,$33,$5F,$AC,$3F,$DA,$92,$C7,$FA,$9C,$7F,$A7
	.dc.b	$DA,$A3,$53,$FE,$BC,$6C,$7F,$B9,$9F,$F7,$E9,$8A,$FF,$18,$36,$7F
	.dc.b	$D7,$95,$FD,$03,$D0,$79,$FF,$D5,$30,$ED,$36,$F8,$47,$8E,$FF,$A1
	.dc.b	$5F,$07,$FB,$8C,$EA,$78,$C6,$2B,$B5,$F3,$64,$64,$E7,$18,$9F,$54
	.dc.b	$DC,$14,$9F,$F0,$AE,$1E,$17,$71,$CD,$A2,$78,$BC,$2F,$A4,$FC,$23
	.dc.b	$84,$EB,$15,$F5,$8F,$93,$87,$0C,$3E,$11,$D6,$FC,$97,$C3,$0D,$4F
	.dc.b	$C2,$3C,$77,$D7,$0E,$97,$E1,$9C,$B6,$75,$5C,$B1,$F9,$87,$DC,$C7
	.dc.b	$07,$CC,$AF,$49,$C4,$97,$EF,$72,$FA,$2F,$D7,$1F,$40,$CC,$79,$E3
	.dc.b	$2F,$B4,$E4,$79,$8E,$51,$DB,$1F,$80,$F2,$BF,$3D,$F9,$6B,$81,$D6
	.dc.b	$78,$63,$F0,$A3,$99,$0D,$1E,$0C,$3C,$64,$9E,$38,$FC,$27,$DB,$8F
	.dc.b	$8A,$BB,$E3,$E2,$9F,$74,$63,$1F,$B5,$5B,$C3,$A4,$F8,$6F,$E7,$5F
	.dc.b	$AE,$FD,$55,$AA,$F2,$9E,$6B,$E5,$5F,$42,$1E,$67,$B7,$85,$C0,$F4
	.dc.b	$3F,$5D,$39,$8C,$5F,$F9,$6A,$31,$3A,$D3,$FC,$8F,$7C,$A3,$15,$FA
	.dc.b	$3C,$21,$E0,$AF,$E0,$F7,$FA,$58,$CB,$5E,$0B,$CE,$EB,$8B,$CF,$F2
	.dc.b	$C6,$1B,$87,$F9,$75,$B7,$FE,$7C,$AE,$13,$E9,$AF,$09,$1F,$2C,$FE
	.dc.b	$E6,$FC,$11,$FC,$67,$CE,$FF,$31,$EA,$F3,$FF,$3F,$18,$6C,$FF,$A3
	.dc.b	$67,$5B,$FF,$15,$AD,$36,$7A,$13,$0F,$0F,$02,$BF,$95,$F9,$DF,$E7
	.dc.b	$6A,$FF,$45,$E8,$F3,$FF,$91,$4E,$70,$70,$FD,$A9,$CF,$3F,$C8,$6E
	.dc.b	$1F,$E8,$17,$F9,$47,$9F,$23,$FD,$2B,$7F,$C6,$7C,$FF,$E3,$5E,$6B
	.dc.b	$CD,$FB,$5B,$30,$E1,$7F,$CE,$DF,$E6,$8F,$11,$5C,$5F,$E4,$93,$C7
	.dc.b	$0E,$F0,$C3,$BE,$1D,$B0,$FE,$47,$87,$1A,$FC,$55,$69,$E7,$3A,$AF
	.dc.b	$B3,$39,$9F,$EA,$53,$97,$F2,$31,$9C,$3C,$51,$E6,$AE,$2D,$3A,$E1
	.dc.b	$CD,$F6,$61,$E3,$AE,$2F,$96,$33,$5E,$9A,$E1,$1E,$D3,$A1,$E4,$73
	.dc.b	$7E,$3C,$3C,$D3,$EE,$37,$87,$6B,$F2,$E1,$E1,$AD,$D7,$80,$F3,$9D
	.dc.b	$8C,$1D,$CC,$1E,$38,$E5,$87,$8A,$3F,$42,$7D,$92,$E1,$87,$07,$EE
	.dc.b	$FC,$C7,$9C,$F3,$F8,$C6,$1B,$19,$9F,$D6,$CE,$5B,$E7,$FF,$61,$6F
	.dc.b	$27,$28,$C4,$7F,$08,$94,$62,$3C,$4F,$1C,$BF,$0D,$E7,$BC,$EF,$E2
	.dc.b	$11,$86,$ED,$C3,$1F,$C3,$63,$D1,$FC,$45,$F1,$C6,$DF,$07,$FB,$E5
	.dc.b	$6D,$FB,$6D,$C3,$B4,$70,$3A,$E3,$E5,$C7,$C1,$1F,$82,$C3,$B3,$70
	.dc.b	$EC,$B0,$E9,$5C,$30,$EB,$3B,$3A,$E3,$1C,$12,$3F,$51,$5C,$A3,$D3
	.dc.b	$8F,$8D,$FF,$0B,$63,$E2,$AE,$A7,$09,$D4,$A7,$CF,$7F,$E1,$68,$C6
	.dc.b	$3F,$CF,$63,$13,$AD,$7D,$19,$7F,$41,$8C,$5F,$83,$CF,$E1,$59,$6C
	.dc.b	$79,$30,$AB,$F4,$27,$C2,$EB,$95,$E3,$BE,$23,$0D,$BF,$F1,$96,$BE
	.dc.b	$7C,$38,$C6,$33,$FE,$A9,$18,$8F,$F0,$56,$1D,$30,$E6,$78,$A7,$57
	.dc.b	$F0,$7F,$41,$FE,$A5,$7F,$A8,$6B,$FD,$29,$AF,$F6,$06,$BF,$D8,$DD
	.dc.b	$C7,$F9,$02,$FD,$79,$FA,$8F,$83,$3A,$C7,$87,$81,$9C,$3D,$47,$F4
	.dc.b	$0C,$FE,$6A,$AE,$0F,$F9,$4F,$CC,$79,$6F,$2F,$F7,$F1,$86,$F9,$8F
	.dc.b	$1F,$35,$BB,$FE,$CB,$EC,$3D,$8F,$F3,$FB,$75,$FE,$12,$D9,$DC,$FF
	.dc.b	$A1,$3D,$A7,$B7,$49,$7F,$C4,$23,$2F,$F1,$2F,$A9,$7D,$17,$FA,$31
	.dc.b	$B1,$98,$C1,$1F,$87,$82,$76,$FF,$91,$61,$FB,$F7,$E2,$9D,$CE,$B9
	.dc.b	$7F,$38,$D8,$FF,$0C,$36,$BF,$CF,$B5,$C5,$FF,$28,$1D,$EF,$FD,$29
	.dc.b	$C2,$F6,$3F,$D4,$6D,$C3,$F0,$D1,$82,$3E,$88,$D5,$7E,$4B,$8D,$D5
	.dc.b	$D2,$E5,$ED,$FF,$17,$CE,$A3,$07,$3F,$3B,$C2,$47,$F2,$CD,$BF,$4E
	.dc.b	$1F,$E4,$1A,$7C,$77,$AA,$A5,$F3,$3F,$1F,$D3,$FB,$09,$CE,$1F,$0B
	.dc.b	$9B,$FC,$8A,$FC,$FD,$7F,$B6,$EF,$FC,$1E,$72,$FF,$A7,$B0,$FA,$7E
	.dc.b	$2F,$FA,$EE,$73,$1F,$EA,$93,$27,$DC,$73,$7F,$4A,$78,$7A,$1F,$EA
	.dc.b	$8B,$E9,$87,$83,$0F,$F2,$86,$1C,$1E,$78,$79,$EF,$84,$77,$AE,$13
	.dc.b	$FE,$B9,$AE,$11,$88,$FD,$45,$70,$7C,$55,$E5,$8C,$3D,$AF,$E0,$46
	.dc.b	$2B,$95,$F4,$C3,$D3,$7F,$BA,$38,$1E,$8A,$E1,$1A,$C3,$E6,$37,$0F
	.dc.b	$7B,$6B,$D4,$D9,$F2,$9E,$F8,$ED,$87,$8A,$FA,$C7,$A4,$CD,$71,$7C
	.dc.b	$75,$F0,$E7,$6F,$33,$C7,$1C,$5A,$66,$FC,$B1,$96,$C6,$63,$CE,$64
	.dc.b	$AE,$4F,$4C,$7C,$A8,$F2,$9B,$8C,$BF,$1C,$CB,$C5,$F4,$BF,$61,$CC
	.dc.b	$73,$8F,$3B,$4F,$29,$63,$C6,$D8,$EE,$7C,$38,$E6,$78,$5E,$6D,$8E
	.dc.b	$6E,$63,$FD,$43,$87,$E9,$9B,$5F,$DA,$DD,$1C,$66,$3F,$C4,$39,$FE
	.dc.b	$E3,$EB,$3D,$D8,$FF,$37,$9F,$89,$19,$C3,$D1,$1F,$C8,$2B,$8E,$1F
	.dc.b	$AA,$AD,$4F,$E8,$67,$38,$7E,$07,$0E,$95,$F4,$CD,$9F,$04,$FD,$F4
	.dc.b	$F0,$AF,$A8,$77,$3C,$77,$DC,$FD,$5C,$6E,$7D,$2D,$C3,$CF,$7E,$07
	.dc.b	$E8,$4F,$80,$F0,$DF,$57,$C9,$7D,$70,$FD,$76,$1C,$EB,$E4,$9B,$BE
	.dc.b	$75,$B8,$FD,$1C,$EE,$FD,$95,$FC,$2A,$31,$1E,$0C,$3C,$95,$F5,$23
	.dc.b	$C9,$7E,$C8,$F7,$E1,$F2,$6F,$8C,$E6,$BC,$F3,$9A,$D9,$F2,$2F,$AC
	.dc.b	$F8,$A3,$B5,$71,$C3,$D3,$7C,$6B,$9D,$71,$C3,$BC,$FC,$28,$EB,$3F
	.dc.b	$39,$B5,$F3,$AB,$71,$F0,$8C,$9F,$23,$0F,$1D,$7B,$AF,$C1,$7E,$2A
	.dc.b	$F0,$1C,$27,$D0,$7E,$5D,$C3,$F2,$6F,$BB,$F8,$88,$CE,$1E,$D7,$DF
	.dc.b	$7D,$9A,$76,$9F,$79,$EB,$2D,$FA,$E3,$9C,$FC,$19,$CC,$7A,$EF,$C3
	.dc.b	$5F,$76,$36,$7C,$19,$D9,$EA,$7F,$10,$78,$9F,$04,$7C,$16,$C6,$63
	.dc.b	$C5,$1C,$8F,$81,$1D,$9F,$80,$66,$FE,$03,$F2,$0F,$9D,$1E,$39,$EC
	.dc.b	$74,$2B,$EB,$3A,$3E,$B8,$E9,$7F,$06,$37,$1E,$BC,$3E,$DD,$FE,$86
	.dc.b	$3D,$27,$85,$C4,$F3,$38,$C7,$A8,$EB,$7E,$53,$E0,$1E,$37,$27,$BA
	.dc.b	$38,$CF,$BA,$3D,$A7,$80,$B1,$ED,$9E,$71,$E7,$C3,$EB,$C7,$F1,$53
	.dc.b	$DE,$58,$EF,$3E,$78,$E3,$1E,$78,$F0,$E1,$F0,$E7,$37,$E3,$3D,$EF
	.dc.b	$A6,$76,$FB,$6B,$E2,$57,$E4,$DB,$1F,$93,$9D,$C7,$BE,$3A,$CF,$78
	.dc.b	$D9,$EF,$AE,$EF,$B6,$7D,$6F,$E9,$CF,$5B,$ED,$AE,$A7,$B6,$39,$BF
	.dc.b	$1D,$B1,$E9,$3C,$27,$A5,$F9,$AE,$5D,$9E,$13,$B9,$E1,$8E,$F1,$E1
	.dc.b	$39,$C7,$B4,$E2,$7C,$D7,$31,$E5,$8E,$E7,$57,$9B,$D6,$36,$75,$7F
	.dc.b	$41,$FA,$0B,$DF,$FE,$C5,$01,$19,$9C,$44,$33,$3C,$1B,$19,$79,$4F
	.dc.b	$17,$FC,$EE,$DA,$FF,$58,$6E,$F9,$8F,$F1,$E5,$72,$37,$86,$2F,$86
	.dc.b	$1E,$5D,$FF,$DA,$37,$9A,$D7,$17,$8E,$FF,$CF,$78,$79,$DA,$FA,$9A
	.dc.b	$F9,$A3,$8E,$1F,$E6,$EC,$5F,$0B,$8A,$F4,$D5,$CF,$D3,$36,$FA,$71
	.dc.b	$78,$48,$FC,$3B,$F9,$71,$87,$E1,$BF,$E8,$AF,$8D,$F7,$32,$7F,$D5
	.dc.b	$79,$7F,$A1,$67,$15,$FE,$7F,$F3,$7E,$86,$E6,$92,$FC,$11,$97,$FC
	.dc.b	$F5,$1F,$E6,$9F,$AC,$7E,$4A,$FA,$AF,$F4,$14,$CA,$D9,$FE,$DE,$AD
	.dc.b	$CE,$70,$FF,$58,$FA,$EF,$63,$F0,$4D,$BF,$80,$D9,$FC,$23,$5F,$C5
	.dc.b	$39,$DD,$3F,$8E,$8C,$3F,$7F,$D5,$FF,$5A,$CE,$AB,$F2,$D5,$A6,$DF
	.dc.b	$E2,$A3,$1C,$AB,$E3,$87,$5A,$FC,$7F,$C4,$FD,$E8,$FA,$CD,$C3,$F3
	.dc.b	$4D,$7F,$40,$D7,$F4,$6E,$F1,$F2,$42,$31,$3E,$9F,$8F,$EF,$C7,$F9
	.dc.b	$EF,$57,$FD,$38,$7F,$A7,$6F,$F6,$1E,$AF,$EC,$7D,$27,$A3,$FB,$0A
	.dc.b	$DF,$7F,$13,$F4,$F7,$C0,$57,$07,$F7,$53,$D5,$5F,$5B,$E9,$7F,$2E
	.dc.b	$F8,$57,$EF,$AB,$4F,$29,$D5,$7E,$F6,$73,$2B,$56,$A3,$F8,$34,$EA
	.dc.b	$73,$97,$09,$F4,$61,$E0,$9C,$C7,$AB,$0E,$CD,$3C,$55,$C5,$F5,$99
	.dc.b	$3C,$58,$78,$5F,$51,$9A,$EB,$7E,$E8,$F0,$DF,$C2,$BE,$98,$79,$F0
	.dc.b	$E5,$5A,$3C,$27,$73,$07,$82,$78,$BD,$E3,$8B,$E6,$BF,$74,$62,$BF
	.dc.b	$82,$4F,$D9,$3F,$83,$38,$7D,$A7,$F2,$FF,$4D,$EA,$3D,$5F,$E1,$91
	.dc.b	$86,$C6,$67,$F9,$94,$E5,$BE,$AB,$83,$8B,$63,$31,$82,$3F,$37,$D5
	.dc.b	$7A,$5F,$D0,$B4,$FC,$3C,$62,$7F,$07,$83,$0E,$95,$F8,$C8,$C6,$1F
	.dc.b	$88,$BE,$17,$D7,$0E,$9B,$7F,$4E,$BD,$CF,$DD,$6E,$1E,$0A,$DC,$FF
	.dc.b	$57,$BE,$73,$DF,$1F,$0E,$1C,$DB,$5B,$BE,$2A,$F8,$D6,$F1,$F5,$DF
	.dc.b	$0A,$D5,$F0,$BD,$C7,$F6,$7C,$7C,$B8,$71,$8F,$61,$EF,$AE,$F3,$A8
	.dc.b	$E1,$7D,$6F,$9C,$76,$3C,$38,$71,$7F,$BA,$3F,$C3,$A3,$E4,$B6,$FF
	.dc.b	$BA,$BF,$C4,$AF,$B5,$73,$3F,$01,$B7,$F7,$AB,$79,$1C,$23,$0D,$9F
	.dc.b	$8C,$78,$23,$FC,$03,$E9,$BD,$5F,$CC,$B6,$7F,$AF,$35,$FF,$01,$35
	.dc.b	$FF,$04,$39,$C7,$FC,$03,$EA,$F7,$E9,$78,$88,$F1,$5A,$F8,$63,$FE
	.dc.b	$12,$F2,$DE,$AE,$9B,$5E,$73,$FC,$34,$E1,$F8,$23,$1E,$FD,$DC,$6B
	.dc.b	$F6,$75,$BA,$FF,$0F,$35,$F0,$46,$8F,$46,$5F,$4B,$0F,$F0,$A3,$67
	.dc.b	$C3,$3E,$AF,$80,$7C,$0F,$F0,$06,$CF,$F8,$15,$AF,$F8,$A5,$AF,$F8
	.dc.b	$BF,$D7,$7A,$CF,$1B,$FC,$19,$18,$6F,$AE,$F3,$F3,$5B,$BF,$F1,$A7
	.dc.b	$AE,$F5,$FF,$C7,$0D,$DB,$FC,$2C,$D9,$DC,$FF,$8F,$7D,$57,$AB,$FD
	.dc.b	$C7,$9B,$FB,$96,$0F,$53,$2F,$D1,$C1,$78,$7E,$8C,$6E,$FF,$C7,$F3
	.dc.b	$99,$FF,$26,$35,$FF,$27,$B5,$FF,$2A,$35,$EB,$1C,$8F,$24,$FF,$96
	.dc.b	$4F,$27,$46,$5F,$27,$27,$F8,$0A,$BF,$05,$C8,$FA,$8F,$F2,$8C,$E7
	.dc.b	$9F,$F2,$71,$8F,$4B,$FC,$BD,$3F,$E4,$8B,$34,$91,$BC,$3F,$C9,$5D
	.dc.b	$76,$BF,$22,$70,$F4,$BC,$EF,$8B,$C3,$F0,$C6,$32,$FD,$1C,$62,$FF
	.dc.b	$C9,$58,$AB,$C6,$48,$CB,$BB,$FC,$FD,$FE,$B3,$D6,$C7,$0F,$F8,$5B
	.dc.b	$9D,$EB,$78,$B1,$FB,$5E,$47,$F9,$CA,$73,$1F,$AB,$32,$7D,$D7,$38
	.dc.b	$FD,$CE,$4F,$EF,$38,$75,$C3,$A4,$62,$FF,$D1,$37,$C2,$BF,$13,$39
	.dc.b	$BF,$F7,$36,$1E,$1C,$38,$B7,$0F,$15,$6E,$BF,$45,$5A,$9F,$23,$63
	.dc.b	$E7,$E1,$DE,$7F,$09,$5C,$63,$13,$FE,$50,$8C,$57,$1C,$3D,$57,$E0
	.dc.b	$C3,$C5,$7E,$F6,$CF,$A9,$B3,$D9,$B3,$D5,$B1,$F2,$A3,$CB,$5D,$E3
	.dc.b	$E2,$B6,$3E,$1C,$7B,$0F,$11,$5F,$13,$6B,$C4,$D3,$C4,$58,$F1,$34
	.dc.b	$F1,$47,$BA,$39,$9E,$48,$EE,$D3,$31,$B3,$BB,$96,$99,$8F,$0C,$65
	.dc.b	$B1,$98,$F1,$99,$2B,$98,$F2,$9F,$10,$F1,$B9,$BF,$AF,$18,$35,$3E
	.dc.b	$66,$CF,$91,$A7,$88,$B1,$E0,$69,$E0,$7E,$89,$E5,$70,$FE,$41,$F0
	.dc.b	$3A,$8F,$25,$FD,$EA,$F2,$CF,$8C,$D5,$76,$8C,$D7,$96,$72,$43,$B1
	.dc.b	$C4,$AE,$DB,$3E,$B7,$86,$1F,$9C,$8F,$3B,$6B,$BC,$6E,$78,$9F,$04
	.dc.b	$E3,$87,$E4,$5C,$C7,$28,$C5,$73,$C3,$DC,$49,$FC,$65,$F6,$BE,$87
	.dc.b	$27,$11,$CC,$F9,$31,$99,$E4,$FC,$F9,$D9,$23,$38,$7C,$EA,$EF,$18
	.dc.b	$7E,$63,$6B,$D7,$1D,$AF,$D0,$F0,$24,$77,$9E,$0F,$8E,$33,$5D,$E3
	.dc.b	$51,$83,$B9,$5E,$E7,$D7,$AE,$EF,$02,$4F,$32,$4F,$28,$CB,$63,$31
	.dc.b	$D1,$FC,$04,$FF,$30,$81,$FA,$02,$98,$74,$E4,$86,$70,$FF,$6B,$B8
	.dc.b	$3F,$1E,$72,$3F,$0A,$7D,$9E,$8F,$E3,$6E,$27,$F1,$91,$87,$F9,$5C
	.dc.b	$6E,$FF,$88,$3B,$AF,$E8,$75,$B7,$F4,$CD,$9F,$C4,$B6,$7F,$04,$DB
	.dc.b	$FE,$00,$D9,$FD,$53,$67,$D9,$1F,$D1,$A3,$11,$E3,$3F,$87,$1E,$38
	.dc.b	$D1,$E6,$BF,$49,$EC,$71,$7E,$66,$9E,$62,$CF,$99,$B7,$E6,$69,$E6
	.dc.b	$2C,$79,$8F,$ED,$17,$E6,$BF,$04,$79,$9F,$1D,$FC,$8A,$E3,$3E,$BB
	.dc.b	$FE,$54,$FA,$CE,$71,$AA,$E7,$3A,$79,$C7,$C0,$9E,$D5,$ED,$6C,$7B
	.dc.b	$5A,$7B,$4B,$5E,$D6,$C7,$B5,$A7,$B6,$3F,$BE,$DF,$B6,$BD,$53,$A3
	.dc.b	$D5,$5F,$CC,$9F,$55,$70,$8F,$54,$F0,$8F,$53,$E7,$6C,$79,$DB,$3E
	.dc.b	$76,$9E,$72,$C7,$9D,$B1,$E7,$8F,$F0,$2D,$79,$DF,$8D,$3A,$9F,$39
	.dc.b	$B2,$18,$3D,$31,$D4,$F6,$5F,$D0,$8F,$AA,$7C,$17,$EA,$C6,$AF,$D8
	.dc.b	$D3,$D8,$59,$F6,$36,$BD,$8D,$3C,$11,$E4,$AF,$04,$FC,$B8,$F0,$57
	.dc.b	$CB,$AE,$11,$EB,$7F,$CF,$57,$F0,$63,$CF,$5F,$06,$FC,$F1,$D6,$FC
	.dc.b	$EF,$C2,$F5,$1E,$AF,$58,$FE,$2B,$18,$C7,$E5,$9F,$C5,$9C,$47,$F0
	.dc.b	$F0,$E1,$FE,$24,$D7,$84,$FF,$18,$7A,$BF,$E6,$A7,$8B,$FE,$69,$E4
	.dc.b	$7A,$D9,$AE,$F9,$30,$F1,$72,$6E,$B4,$D8,$FF,$4D,$57,$F9,$FC,$91
	.dc.b	$E1,$C7,$CE,$D8,$FF,$46,$46,$27,$FD,$A5,$C1,$FE,$DC,$F3,$7A,$C7
	.dc.b	$4E,$2F,$37,$91,$E5,$6C,$64,$F8,$97,$F1,$23,$B1,$23,$38,$7F,$A0
	.dc.b	$DE,$13,$FE,$80,$D1,$EB,$75,$9F,$F3,$C7,$AB,$FE,$8D,$FC,$AD,$F3
	.dc.b	$DF,$E7,$F8,$C3,$67,$C3,$C6,$9F,$80,$D8,$CC,$60,$E1,$8F,$FB,$BB
	.dc.b	$D7,$7B,$0F,$27,$3C,$9F,$46,$BF,$AC,$FB,$0F,$5F,$FD,$41,$EC,$7F
	.dc.b	$BF,$33,$FF,$65,$4E,$5F,$97,$19,$AF,$F5,$5C,$E5,$FC,$53,$5F,$EB
	.dc.b	$4D,$7F,$B2,$B5,$F0,$D7,$E2,$63,$95,$F9,$79,$57,$F7,$27,$38,$7F
	.dc.b	$6E,$F6,$1C,$DE,$8F,$8E,$FC,$94,$62,$7C,$F1,$89,$FC,$44,$E7,$3F
	.dc.b	$D0,$35,$FE,$E0,$D7,$FB,$DF,$B1,$FE,$9F,$3F,$D4,$0F,$F8,$03,$D8
	.dc.b	$FF,$00,$FA,$EF,$5F,$7E,$0F,$E8,$EF,$83,$FA,$5B,$E0,$2B,$83,$FE
	.dc.b	$0A,$9E,$AA,$FA,$DF,$4C,$EC,$FF,$81,$66,$AA,$61,$F7,$70,$EB,$3F
	.dc.b	$E1,$4A,$DD,$6A,$F8,$CF,$92,$B7,$39,$8F,$36,$1F,$C1,$6F,$8E,$1D
	.dc.b	$63,$13,$A3,$D5,$19,$7C,$C6,$4F,$33,$97,$C8,$66,$BC,$17,$E8,$8F
	.dc.b	$25,$7C,$17,$91,$83,$BE,$1E,$A7,$E4,$1C,$EF,$B3,$B3,$C3,$1B,$C3
	.dc.b	$D6,$D8,$E0,$D7,$F8,$0C,$7F,$02,$7E,$C9,$83,$F9,$47,$AC,$F1,$9E
	.dc.b	$C7,$F0,$88,$C3,$63,$33,$C1,$AF,$F8,$FE,$5D,$EE,$3E,$4D,$9E,$ED
	.dc.b	$78,$35,$FF,$28,$CF,$29,$FC,$2F,$AE,$F6,$3F,$CA,$BB,$7E,$12,$31
	.dc.b	$7F,$8A,$BE,$0F,$D7,$6C,$EE,$B8,$ED,$FE,$5E,$BE,$37,$BC,$7C,$D7
	.dc.b	$F9,$0C,$38,$DE,$F0,$E3,$7F,$E6,$8C,$38,$DE,$F0,$E3,$7E,$A6,$D6
	.dc.b	$EF,$8A,$BE,$35,$BC,$7D,$57,$C2,$B5,$7C,$2F,$71,$FE,$78,$C7,$CA
	.dc.b	$DA,$E1,$3A,$8E,$27,$4B,$8F,$E6,$E3,$46,$EB,$9C,$70,$3E,$D6,$BF
	.dc.b	$05,$FF,$43,$D7,$F3,$17,$4D,$9E,$35,$F3,$23,$4B,$CA,$FF,$46,$C6
	.dc.b	$1B,$7D,$6F,$B9,$FE,$94,$F5,$DE,$6E,$A3,$0D,$51,$A7,$FB,$3F,$84
	.dc.b	$F6,$F5,$39,$9F,$F0,$33,$5F,$F4,$B3,$5F,$F5,$0B,$5F,$F5,$2F,$9A
	.dc.b	$F6,$3C,$0D,$8F,$87,$8F,$FA,$A2,$31,$1F,$DA,$7B,$B7,$52,$36,$74
	.dc.b	$7D,$9D,$AC,$BF,$AF,$4E,$85,$F4,$7F,$D5,$8D,$7C,$15,$FB,$D8,$C4
	.dc.b	$7F,$AF,$3D,$57,$C3,$7F,$01,$6C,$7F,$81,$1A,$FF,$8A,$5A,$FF,$AE
	.dc.b	$7D,$87,$B3,$FC,$25,$B1,$FE,$0C,$6B,$FE,$31,$6B,$FE,$C1,$F6,$5E
	.dc.b	$CB,$C8,$FF,$09,$C6,$1B,$83,$D2,$CD,$6E,$FF,$C7,$9E,$BB,$D7,$FF
	.dc.b	$00,$73,$FF,$80,$78,$DE,$AE,$5F,$91,$56,$FF,$17,$1A,$BF,$F1,$F4
	.dc.b	$E6,$F9,$5D,$AF,$F7,$0C,$DB,$FF,$1F,$4D,$51,$D5,$FF,$6F,$35,$E3
	.dc.b	$7F,$E3,$D8,$C4,$7E,$2B,$D2,$FF,$26,$7A,$FF,$E4,$FE,$24,$63,$1F
	.dc.b	$C0,$46,$1F,$CD,$CE,$E7,$55,$FA,$63,$F0,$AF,$F9,$2E,$B8,$47,$F9
	.dc.b	$8D,$AF,$E2,$23,$5E,$9F,$F9,$77,$C2,$F1,$37,$0F,$0B,$6B,$B5,$FF
	.dc.b	$91,$5F,$F2,$25,$70,$AF,$F2,$6C,$E6,$7F,$2B,$5D,$A7,$FC,$83,$EB
	.dc.b	$7D,$85,$3F,$33,$C6,$FF,$38,$71,$2B,$53,$9A,$FD,$14,$E5,$FD,$49
	.dc.b	$F8,$0A,$E3,$5C,$23,$07,$07,$1B,$7F,$80,$F7,$FE,$31,$7C,$F0,$FC
	.dc.b	$2C,$E6,$FC,$B1,$F9,$7B,$E3,$87,$69,$F0,$D6,$EB,$FD,$40,$D7,$84
	.dc.b	$6A,$78,$47,$66,$E1,$F9,$06,$CF,$C4,$3D,$33,$F7,$70,$F3,$56,$E7
	.dc.b	$D8,$75,$2C,$F5,$6C,$6E,$39,$61,$EA,$9C,$9D,$F0,$F4,$5F,$68,$C6
	.dc.b	$1E,$4B,$FA,$47,$28,$F0,$34,$F0,$17,$0F,$01,$F3,$B0,$F0,$46,$67
	.dc.b	$A9,$E2,$7A,$CE,$E3,$AB,$63,$AB,$4E,$B1,$F5,$AF,$D1,$3C,$A3,$15
	.dc.b	$E0,$AE,$11,$E0,$9F,$11,$B3,$EC,$DF,$BE,$3E,$8C,$F0,$9D,$1F,$70
	.dc.b	$F1,$C7,$63,$CE,$57,$27,$47,$DC,$74,$35,$1E,$A8,$D1,$E4,$8D,$38
	.dc.b	$8F,$14,$62,$3C,$51,$C5,$B1,$ED,$8D,$E1,$F7,$2F,$E1,$4F,$80,$D1
	.dc.b	$6B,$DB,$87,$CA,$8E,$CD,$AE,$51,$EE,$9F,$23,$C4,$DD,$76,$8C,$90
	.dc.b	$F2,$9D,$9D,$38,$8E,$27,$57,$13,$E0,$7E,$59,$8C,$3E,$B4,$67,$0F
	.dc.b	$85,$87,$9E,$72,$4B,$FA,$35,$B7,$E0,$DF,$96,$B8,$46,$AB,$84,$FD
	.dc.b	$3A,$E0,$FA,$67,$CD,$5C,$23,$55,$C2,$FE,$8B,$6B,$E3,$5F,$27,$DF
	.dc.b	$5E,$CB,$E5,$7C,$27,$C4,$49,$EA,$D8,$D4,$60,$8E,$09,$1E,$0B,$F1
	.dc.b	$4F,$BF,$0F,$A3,$7E,$0A,$D1,$D4,$B3,$D1,$B7,$D2,$39,$47,$C2,$AE
	.dc.b	$64,$70,$48,$F7,$46,$5B,$19,$9F,$B3,$39,$6C,$64,$FA,$F3,$F8,$17
	.dc.b	$CC,$F8,$8E,$64,$8E,$2D,$8E,$35,$EF,$8E,$13,$A2,$4E,$AB,$C4,$D3
	.dc.b	$C5,$3F,$81,$3A,$16,$78,$B6,$38,$3A,$7E,$FC,$69,$A6,$9C,$61,$F8
	.dc.b	$52,$61,$F6,$2B,$E0,$1F,$4C,$F4,$D7,$12,$3E,$12,$3D,$9B,$3D,$23
	.dc.b	$E4,$C7,$4A,$FB,$8D,$8F,$BE,$47,$B9,$CA,$BC,$98,$7C,$B8,$E0,$6A
	.dc.b	$7C,$04,$BF,$04,$7A,$4D,$92,$79,$13,$0F,$F4,$64,$7F,$2B,$3F,$B9
	.dc.b	$9F,$CC,$4B,$1F,$DE,$5B,$1F,$C1,$38,$7F,$AC,$47,$F1,$03,$F8,$5B
	.dc.b	$89,$FE,$8F,$18,$BF,$EA,$4E,$E3,$F9,$04,$EE,$BA,$56,$A7,$89,$A7
	.dc.b	$11,$AA,$FE,$9A,$73,$7F,$C2,$35,$CE,$7F,$89,$B6,$7F,$87,$37,$1F
	.dc.b	$E7,$2D,$9F,$E1,$2D,$9F,$F0,$7B,$5F,$51,$E0,$BF,$2D,$F8,$2B,$CA
	.dc.b	$78,$0B,$3E,$06,$DF,$81,$A7,$80,$B1,$E0,$6B,$E0,$8F,$EC,$CB,$D4
	.dc.b	$7A,$DE,$18,$FF,$2B,$46,$2B,$FC,$53,$8F,$F8,$FA,$78,$56,$EF,$FD
	.dc.b	$99,$7B,$9F,$F1,$AD,$6E,$31,$1F,$84,$6B,$FE,$38,$6B,$FE,$4E,$6B
	.dc.b	$E5,$DF,$FC,$9B,$D7,$FE,$AF,$F5,$BA,$46,$3B,$11,$8C,$FF,$D9,$6D
	.dc.b	$BF,$F6,$AD,$F1,$9F,$F5,$B6,$1F,$E6,$E9,$F1,$61,$F9,$A7,$C4,$D7
	.dc.b	$C5,$3F,$9C,$7C,$DC,$1E,$1F,$13,$DC,$D9,$EA,$D9,$F9,$6D,$D3,$E3
	.dc.b	$BD,$70,$FF,$39,$4F,$C3,$8C,$1E,$1D,$DE,$9F,$C2,$F5,$5E,$AF,$F9
	.dc.b	$CA,$30,$DC,$BF,$DA,$FA,$AB,$71,$FE,$74,$AD,$B4,$F6,$63,$FE,$BF
	.dc.b	$8F,$01,$23,$3E,$8B,$D3,$FA,$7E,$9F,$FB,$B6,$7F,$DC,$13,$97,$C3
	.dc.b	$19,$BF,$F5,$04,$E5,$FA,$71,$9D,$95,$FE,$7A,$AD,$46,$29,$92,$F8
	.dc.b	$5E,$F0,$FA,$D8,$75,$BF,$E2,$2D,$75,$3E,$EF,$45,$E9,$7A,$BD,$3F
	.dc.b	$CE,$46,$27,$D3,$18,$9D,$47,$FB,$F6,$75,$5E,$AA,$FD,$27,$A7,$FE
	.dc.b	$8B,$3F,$D1,$91,$FE,$A8,$F4,$FF,$D5,$3E,$93,$D2,$FF,$4B,$D6,$FB
	.dc.b	$B8,$9F,$AD,$BE,$02,$B8,$3F,$EB,$29,$EA,$AF,$AD,$F4,$BF,$91,$7C
	.dc.b	$2B,$FD,$4B,$5A,$79,$4E,$AB,$F3,$93,$99,$5A,$B5,$1F,$CF,$27,$53
	.dc.b	$9C,$B8,$4F,$A3,$0F,$04,$E6,$3D,$58,$76,$69,$E2,$AE,$2F,$AC,$C9
	.dc.b	$E2,$C3,$C2,$FA,$8C,$D7,$5B,$F7,$47,$86,$FE,$15,$F4,$C3,$CF,$87
	.dc.b	$2A,$D1,$E1,$3B,$98,$3C,$13,$C5,$EF,$1C,$5F,$35,$FB,$A3,$15,$FA
	.dc.b	$A9,$FB,$27,$EB,$1C,$3E,$D3,$FB,$2F,$A2,$F0,$5E,$9F,$EC,$A3,$0D
	.dc.b	$8C,$CF,$F1,$19,$CB,$7A,$9C,$9C,$5B,$19,$8C,$11,$F9,$BE,$93,$D3
	.dc.b	$FF,$74,$E9,$F8,$78,$C4,$FE,$0F,$06,$1D,$2B,$F1,$91,$8C,$3F,$11
	.dc.b	$7C,$2F,$AE,$1D,$36,$FF,$7A,$DE,$E7,$EE,$B7,$0F,$05,$6E,$42,$FA
	.dc.b	$4F,$83,$1F,$16,$1D,$1B,$5C,$2B,$73,$CA,$39,$5F,$55,$87,$86,$79
	.dc.b	$63,$EA,$AD,$5F,$2B,$F1,$DE,$DB,$87,$81,$EA,$71,$3E,$43,$98,$E1
	.dc.b	$1A,$3E,$29,$AB,$E0,$FD,$AD,$BD,$17,$F2,$E7,$F9,$14,$F1,$6D,$74
	.dc.b	$9D,$57,$0F,$4F,$E0,$57,$09,$46,$27,$AE,$37,$0F,$9F,$DC,$F9,$EC
	.dc.b	$74,$BC,$1C,$57,$F4,$C6,$BE,$73,$8C,$F3,$55,$AC,$FF,$9D,$5F,$3A
	.dc.b	$EB,$79,$C7,$EB,$5D,$C3,$F0,$3F,$A3,$F2,$7F,$AA,$B5,$FE,$E4,$D7
	.dc.b	$FC,$9E,$D7,$F0,$0E,$F1,$F0,$DB,$D0,$F0,$9E,$79,$D6,$1C,$30,$CE
	.dc.b	$1E,$FA,$D4,$7B,$7B,$5E,$CF,$EF,$DB,$5F,$DA,$9A,$FF,$81,$DA,$FF
	.dc.b	$99,$1C,$E5,$F7,$3E,$2D,$F2,$9F,$25,$B5,$FA,$29,$FC,$0D,$70,$CB
	.dc.b	$F3,$8D,$9F,$D3,$D7,$07,$EA,$07,$15,$CB,$0F,$AD,$E8,$BD,$07,$8B
	.dc.b	$FE,$02,$8C,$37,$D1,$79,$59,$AD,$E5,$FE,$23,$F7,$5E,$EF,$F8,$9D
	.dc.b	$BB,$7F,$83,$1B,$3B,$9F,$F1,$67,$C0,$3D,$AA,$4B,$FF,$09,$C6,$5F
	.dc.b	$C0,$FA,$FF,$E3,$3E,$53,$49,$1A,$3E,$41,$F8,$27,$FC,$8D,$F0,$DF
	.dc.b	$E5,$99,$E3,$3A,$C7,$F1,$F5,$93,$71,$98,$4F,$07,$E6,$AF,$1B,$C8
	.dc.b	$49,$FC,$BB,$67,$6F,$F9,$5F,$F4,$EF,$13,$27,$B5,$73,$7F,$5D,$39
	.dc.b	$AF,$B7,$EB,$BD,$8F,$F1,$27,$2F,$F8,$96,$DE,$B6,$5F,$45,$5B,$FD
	.dc.b	$B4,$6A,$FF,$CF,$53,$9B,$E5,$76,$BF,$94,$4D,$BF,$F1,$AC,$D5,$19
	.dc.b	$9E,$87,$82,$FC,$AF,$F3,$A3,$33,$EA,$8C,$BD,$1F,$19,$88,$DD,$FF
	.dc.b	$20,$AE,$D7,$E1,$6D,$70,$AF,$B4,$70,$3D,$2F,$39,$E0,$D3,$85,$FC
	.dc.b	$48,$CC,$F2,$6C,$66,$77,$7F,$19,$B5,$ED,$AD,$CE,$63,$E1,$1C,$E7
	.dc.b	$75,$EF,$8C,$BD,$5B,$3C,$63,$E3,$47,$13,$27,$79,$E6,$FB,$E3,$FB
	.dc.b	$3D,$FC,$48,$DC,$77,$39,$9D,$CB,$5D,$CF,$9D,$5F,$16,$78,$9D,$67
	.dc.b	$EF,$C6,$5E,$31,$E3,$37,$7F,$09,$C4,$68,$F1,$57,$68,$F6,$46,$5E
	.dc.b	$D7,$E9,$8C,$57,$37,$8C,$FD,$C3,$FD,$2F,$3F,$9D,$8F,$EF,$C7,$9A
	.dc.b	$33,$7F,$EB,$99,$E5,$1F,$A6,$F7,$7E,$9D,$6C,$EE,$E2,$77,$68,$C5
	.dc.b	$FD,$08,$C5,$FC,$EB,$E5,$87,$B5,$FC,$65,$F0,$9F,$8A,$D8,$EF,$1C
	.dc.b	$70,$EF,$39,$AE,$47,$ED,$1C,$4F,$C1,$AE,$AF,$1C,$3C,$93,$A9,$FC
	.dc.b	$1C,$7E,$86,$31,$1E,$69,$FC,$C4,$62,$FF,$5A,$DA,$E1,$3D,$4F,$49
	.dc.b	$6B,$D2,$7D,$D3,$A1,$E0,$7F,$55,$3C,$DA,$71,$BF,$6B,$E2,$30,$7B
	.dc.b	$5F,$80,$E2,$7A,$1F,$00,$F0,$5F,$C8,$8D,$47,$89,$FB,$C6,$23,$DA
	.dc.b	$DB,$F6,$CF,$27,$E4,$C6,$4F,$3C,$65,$F8,$2D,$3E,$0D,$7A,$9F,$2B
	.dc.b	$63,$C3,$1B,$8C,$61,$F8,$88,$C4,$FB,$CF,$84,$E0,$E6,$E2,$7B,$47
	.dc.b	$CE,$3D,$D3,$CE,$33,$7F,$0E,$79,$1A,$C3,$E9,$3F,$26,$77,$87,$C0
	.dc.b	$8D,$D7,$C6,$6C,$7C,$69,$DE,$1F,$09,$B3,$F0,$A3,$D4,$FA,$08,$FC
	.dc.b	$28,$D9,$F0,$AF,$C6,$FD,$E9,$EF,$3F,$42,$36,$F4,$39,$1D,$A3,$E1
	.dc.b	$5F,$9E,$77,$1E,$0C,$3E,$25,$6E,$BE,$25,$6A,$BE,$24,$FD,$F7,$E2
	.dc.b	$3B,$C3,$E2,$4E,$EB,$B4,$E4,$F8,$45,$AF,$84,$7C,$13,$CB,$5F,$26
	.dc.b	$35,$5F,$05,$F0,$3F,$04,$ED,$1B,$9E,$55,$E8,$AE,$27,$C3,$BF,$6C
	.dc.b	$EA,$BC,$F3,$A8,$F8,$58,$7C,$E8,$FC,$51,$98,$F8,$27,$D0,$3E,$09
	.dc.b	$CD,$F6,$B4,$F6,$9E,$C7,$2D,$33,$7F,$CD,$DE,$B8,$7B,$27,$EE,$1E
	.dc.b	$19,$F9,$73,$D6,$7E,$23,$E3,$7E,$24,$EE,$FE,$24,$F7,$9D,$9D,$5F
	.dc.b	$49,$99,$F8,$D7,$F0,$63,$FC,$32,$D8,$FC,$5C,$78,$5D,$9F,$50,$F8
	.dc.b	$EE,$4F,$96,$E5,$E6,$D3,$9B,$E1,$36,$7A,$DE,$6E,$4F,$8A,$FB,$4F
	.dc.b	$8A,$F7,$32,$5B,$F8,$EF,$16,$9F,$3C,$B7,$FE,$2C,$87,$E8,$CF,$8B
	.dc.b	$FF,$64,$C7,$FB,$8E,$31,$7F,$DA,$23,$FD,$DF,$18,$70,$D7,$FB,$6B
	.dc.b	$9C,$7F,$DA,$6D,$9F,$F7,$BF,$83,$FE,$C7,$8F,$11,$9A,$EB,$78,$3F
	.dc.b	$C8,$33,$BA,$FF,$37,$56,$EB,$D3,$BF,$A1,$F2,$B5,$CC,$FA,$5A,$7A
	.dc.b	$23,$C2,$76,$F0,$1F,$0D,$8A,$EF,$36,$BF,$DB,$58,$7F,$B7,$EB,$19
	.dc.b	$7F,$BC,$3A,$9E,$D7,$85,$B8,$F8,$5B,$87,$CB,$6B,$FE,$14,$73,$5C
	.dc.b	$EB,$FC,$A9,$D5,$F5,$E7,$36,$B9,$18,$6D,$70,$9F,$F4,$A4,$E6,$5E
	.dc.b	$C3,$D8,$F8,$CD,$AF,$8C,$D9,$FC,$03,$5F,$F1,$4B,$9C,$53,$E3,$9E
	.dc.b	$B5,$F2,$FE,$0F,$F1,$D7,$BA,$D3,$6B,$D4,$DD,$1B,$2F,$8E,$1F,$97
	.dc.b	$F5,$7C,$58,$77,$6B,$F0,$DA,$9A,$77,$71,$1F,$8C,$F5,$DF,$1F,$D6
	.dc.b	$26,$7F,$EB,$BB,$F2,$56,$A3,$13,$FA,$9F,$59,$E6,$7E,$3D,$B3,$F8
	.dc.b	$F6,$E3,$FA,$96,$BF,$E5,$27,$3B,$7C,$A8,$FB,$91,$89,$FD,$3E,$0F
	.dc.b	$63,$C1,$87,$EE,$7D,$77,$AF,$4F,$DE,$63,$F8,$5F,$5F,$F0,$DE,$AB
	.dc.b	$D5,$FA,$D7,$C3,$3F,$F6,$C3,$6D,$AF,$8D,$46,$63,$F5,$D8,$74,$BE
	.dc.b	$78,$F8,$2D,$39,$7F,$2D,$7D,$71,$F1,$47,$F9,$CA,$74,$F6,$AF,$85
	.dc.b	$3A,$8F,$C6,$4E,$AB,$95,$FF,$13,$8C,$5F,$08,$C5,$2F,$95,$F0,$8F
	.dc.b	$01,$99,$F4,$E3,$E2,$8F,$90,$64,$F4,$DF,$57,$8D,$F4,$8C,$47,$98
	.dc.b	$E0,$7A,$E7,$A3,$F1,$0C,$DF,$8D,$E1,$5C,$E7,$78,$7B,$1B,$3C,$2F
	.dc.b	$BB,$E6,$AE,$0F,$80,$F4,$9F,$C7,$EB,$F9,$03,$83,$E2,$9F,$EA,$2F
	.dc.b	$51,$EB,$3D,$7F,$E4,$D1,$86,$C6,$67,$FA,$14,$E5,$BE,$BF,$E2,$1B
	.dc.b	$AF,$28,$C4,$77,$6B,$FC,$1E,$51,$C7,$1F,$C0,$7A,$EF,$5F,$F8,$64
	.dc.b	$61,$BB,$F1,$AE,$0D,$C7,$F1,$77,$C7,$1B,$7C,$1F,$E5,$18,$F8,$2B
	.dc.b	$F2,$2D,$8F,$B3,$5C,$5B,$5C,$54,$EF,$1F,$CB,$37,$0F,$E4,$57,$C6
	.dc.b	$7D,$15,$DB,$1F,$3B,$E2,$6A,$9E,$15,$CB,$0F,$06,$1D,$6B,$86,$3E
	.dc.b	$C6,$1F,$90,$8D,$3E,$6C,$3A,$E1,$D2,$3F,$A4,$35,$D5,$72,$3F,$80
	.dc.b	$47,$C1,$6D,$FF,$3D,$7F,$82,$5F,$69,$D6,$DF,$03,$83,$4D,$C3,$FA
	.dc.b	$B5,$F2,$6C,$7F,$47,$F5,$5E,$BF,$E5,$9B,$3F,$DD,$DA,$FF,$84,$5A
	.dc.b	$FF,$BE,$9C,$D7,$F2,$9E,$BF,$E9,$35,$E2,$A4,$61,$F0,$9F,$DA,$6B
	.dc.b	$8D,$FF,$59,$F2,$51,$C6,$51,$87,$89,$88,$ED,$4F,$0B,$F9,$9C,$EA
	.dc.b	$B8,$3F,$E2,$96,$BF,$D9,$CE,$47,$F8,$55,$CB,$E3,$AE,$BD,$AF,$5E
	.dc.b	$9E,$5A,$EB,$9F,$F3,$7F,$59,$EA,$FA,$5B,$8F,$A1,$AF,$F3,$ED,$3F
	.dc.b	$BC,$7A,$CF,$89,$E1,$7F,$C2,$2F,$77,$E7,$6B,$FD,$FD,$AF,$F8,$4D
	.dc.b	$EB,$87,$F4,$9B,$7B,$1F,$2D,$B5,$E3,$BF,$9D,$F0,$1F,$26,$3F,$C6
	.dc.b	$CD,$7F,$CA,$0D,$7E,$43,$5F,$F0,$87,$B2,$F6,$1E,$47,$F8,$FE,$30
	.dc.b	$DF,$65,$E5,$E6,$B7,$7F,$E5,$DF,$65,$EC,$FF,$92,$5B,$C3,$FE,$49
	.dc.b	$6C,$EE,$7F,$C9,$DC,$5F,$81,$AF,$C6,$F3,$FE,$0F,$1F,$F3,$74,$E6
	.dc.b	$FF,$C9,$3C,$FE,$2B,$FC,$33,$C8,$8F,$E0,$DB,$1F,$5C,$FF,$20,$4F
	.dc.b	$3E,$2F,$D0,$73,$3D,$7E,$F7,$E0,$F5,$FF,$CD,$2F,$F9,$1F,$1F,$CA
	.dc.b	$46,$31,$D4,$7E,$46,$3D,$64,$70,$70,$3D,$B6,$F5,$F8,$D7,$94,$7A
	.dc.b	$AF,$57,$FC,$39,$9F,$F8,$77,$07,$AF,$97,$F5,$75,$6F,$DD,$1A,$BF
	.dc.b	$F4,$B4,$E6,$F9,$5D,$AF,$F6,$F5,$68,$EE,$FB,$E3,$57,$FE,$FF,$9D
	.dc.b	$4F,$F0,$23,$B1,$E2,$78,$C7,$F9,$72,$78,$BE,$C8,$CC,$F0,$8F,$81
	.dc.b	$7D,$A3,$17,$FE,$29,$8C,$5F,$26,$DF,$8A,$FD,$F3,$98,$EA,$DB,$F3
	.dc.b	$36,$77,$3C,$6F,$E1,$B6,$BE,$03,$67,$D4,$D8,$F2,$9F,$20,$F8,$07
	.dc.b	$67,$55,$DE,$BA,$CE,$AB,$C2,$D8,$CC,$7C,$2A,$F4,$9C,$89,$3B,$3C
	.dc.b	$A5,$8F,$2B,$4F,$2B,$F5,$DC,$57,$8D,$E3,$1D,$DF,$41,$A9,$EF,$19
	.dc.b	$3C,$E5,$AF,$39,$F5,$63,$CF,$3C,$49,$3B,$7B,$CE,$DE,$F1,$93,$D2
	.dc.b	$57,$D2,$FE,$1D,$DC,$71,$8F,$1B,$4F,$1C,$7D,$38,$E5,$5E,$F7,$AC
	.dc.b	$EE,$7C,$66,$DC,$46,$A3,$04,$7E,$51,$D2,$BE,$2C,$F0,$9F,$3D,$FE
	.dc.b	$51,$F8,$31,$9B,$F1,$46,$AB,$E3,$DF,$38,$D9,$DD,$F8,$07,$B9,$C5
	.dc.b	$F5,$8E,$04,$BE,$D5,$E4,$AD,$BD,$A3,$31,$DA,$7D,$71,$87,$94,$70
	.dc.b	$3E,$79,$5F,$11,$C1,$C4,$7D,$08,$C1,$0F,$4B,$96,$C7,$D0,$75,$5E
	.dc.b	$B9,$F1,$12,$7A,$9B,$9E,$27,$D5,$38,$D7,$B6,$37,$19,$36,$FC,$F3
	.dc.b	$D6,$F0,$6B,$AB,$FA,$A4,$9F,$95,$1B,$3E,$09,$F5,$0F,$F0,$85,$FF
	.dc.b	$2F,$6C,$7F,$AF,$DF,$F7,$D4,$65,$B1,$94,$73,$33,$5F,$E1,$B7,$33
	.dc.b	$C5,$B3,$FC,$33,$D7,$FE,$29,$1F,$10,$F5,$38,$9F,$5B,$67,$85,$7F
	.dc.b	$07,$AE,$15,$FC,$B2,$B5,$3F,$C1,$9B,$3F,$AA,$6E,$1F,$9B,$6C,$FA
	.dc.b	$27,$F9,$8C,$61,$F3,$C7,$F3,$08,$C4,$79,$A3,$C5,$5E,$66,$DF,$9A
	.dc.b	$7F,$82,$1E,$67,$F9,$DD,$72,$AF,$45,$70,$3D,$05,$AF,$43,$6F,$D0
	.dc.b	$D3,$D0,$FF,$62,$AE,$97,$DA,$75,$5E,$A3,$DA,$E1,$F5,$1E,$6B,$F5
	.dc.b	$36,$7D,$57,$FC,$32,$3D,$55,$F0,$CE,$55,$EA,$8E,$11,$EA,$69,$EA
	.dc.b	$2C,$FA,$9B,$1E,$A3,$FB,$B5,$FA,$9E,$D1,$A3,$D5,$3E,$F8,$F0,$DF
	.dc.b	$BE,$74,$7B,$27,$D1,$7E,$C6,$CF,$B2,$BC,$15,$D1,$A7,$F8,$52,$FC
	.dc.b	$51,$F2,$0E,$B7,$EF,$BE,$75,$EF,$6C,$FB,$DA,$7B,$CB,$1D,$B0,$5F
	.dc.b	$69,$F6,$57,$6C,$3D,$91,$C2,$3C,$C7,$F8,$4E,$78,$5F,$C8,$3E,$19
	.dc.b	$F2,$27,$84,$FC,$38,$F7,$BF,$68,$F6,$46,$B0,$F6,$4E,$9F,$90,$DB
	.dc.b	$F9,$15,$E2,$39,$57,$A0,$E5,$1E,$89,$E0,$7C,$52,$CF,$C5,$6C,$FC
	.dc.b	$57,$CB,$EB,$3D,$7F,$F1,$8F,$13,$57,$CE,$FF,$8D,$31,$5E,$F0,$FF
	.dc.b	$29,$CE,$63,$13,$FC,$4D,$AF,$FA,$BD,$AF,$F1,$A3,$9E,$1F,$C7,$5B
	.dc.b	$5F,$EB,$E9,$E7,$8F,$F9,$B9,$ED,$5D,$F8,$3C,$F8,$BE,$33,$18,$7F
	.dc.b	$99,$9B,$8F,$8E,$BB,$61,$FE,$83,$8F,$24,$60,$DE,$3F,$E7,$CF,$66
	.dc.b	$E1,$E7,$EA,$7B,$3F,$E4,$6F,$43,$E2,$46,$27,$FC,$8F,$18,$9F,$64
	.dc.b	$E7,$1F,$E6,$CD,$7F,$DA,$4D,$7B,$47,$1A,$FF,$28,$F4,$59,$FF,$3F
	.dc.b	$7B,$2F,$5B,$E0,$F9,$BF,$56,$31,$7F,$E8,$68,$C4,$FC,$D9,$CD,$7F
	.dc.b	$47,$6B,$FE,$E3,$6B,$F5,$5D,$E8,$BC,$57,$D0,$F6,$7A,$C7,$FB,$D3
	.dc.b	$D8,$DB,$FD,$5B,$95,$EB,$FC,$1C,$39,$65,$FD,$6A,$F6,$AB,$7A,$FE
	.dc.b	$3E,$7C,$73,$BC,$BF,$0D,$18,$83,$4E,$6E,$23,$F3,$78,$3D,$9E,$91
	.dc.b	$F9,$BF,$65,$E7,$EB,$E5,$70,$DD,$7F,$B7,$EB,$E8,$C9,$7C,$27,$FB
	.dc.b	$84,$62,$BF,$35,$39,$8F,$EE,$8D,$7F,$40,$3D,$97,$27,$EB,$5B,$7E
	.dc.b	$03,$FB,$E7,$B3,$FE,$00,$F5,$DE,$BF,$EA,$6B,$78,$7F,$B0,$5B,$9F
	.dc.b	$9D,$B8,$72,$95,$F2,$7F,$C1,$73,$A7,$CB,$87,$7C,$3B,$63,$F6,$2F
	.dc.b	$85,$7E,$E6,$B4,$F2,$9D,$57,$7A,$DD,$F9,$9E,$35,$FE,$1A,$BE,$2F
	.dc.b	$C3,$BE,$2F,$88,$F3,$DF,$8B,$0D,$C7,$13,$27,$85,$CA,$AE,$B8,$78
	.dc.b	$6B,$C8,$73,$9E,$27,$AC,$B1,$B7,$C5,$87,$AB,$0F,$2C,$6E,$31,$87
	.dc.b	$83,$0E,$F5,$AA,$F3,$56,$9E,$E6,$30,$F1,$47,$C4,$8F,$41,$F8,$C2
	.dc.b	$9F,$8E,$AC,$38,$3F,$95,$7A,$CF,$61,$EC,$FE,$52,$30,$D8,$CC,$FF
	.dc.b	$91,$E7,$2D,$E8,$6B,$C5,$B1,$98,$C1,$1F,$93,$EC,$3D,$9F,$E7,$99
	.dc.b	$FE,$1A,$31,$3F,$7B,$06,$1D,$2B,$F1,$71,$8C,$3F,$07,$7C,$2F,$AE
	.dc.b	$1D,$34,$F7,$E3,$C5,$B7,$C2,$75,$5F,$E6,$6B,$E5,$8F,$6C,$38,$4F
	.dc.b	$F9,$AB,$1F,$03,$D8,$E4,$79,$E3,$6A,$37,$3E,$DB,$EF,$2B,$EF,$8F
	.dc.b	$92,$3F,$1B,$7C,$23,$FB,$04,$EA,$7C,$D5,$CA,$B5,$1E,$5C,$3C,$35
	.dc.b	$B8,$ED,$5B,$9C,$BF,$DA,$F1,$FE,$2D,$5F,$25,$B1,$9D,$BC,$99,$7A
	.dc.b	$70,$FE,$1C,$7C,$7E,$1D,$34,$C9,$E0,$2F,$B0,$F6,$77,$08,$C3,$A8
	.dc.b	$E7,$D4,$F0,$3F,$C6,$33,$93,$F9,$4D,$70,$7F,$D3,$4D,$7F,$93,$B5
	.dc.b	$FF,$04,$79,$BD,$DF,$2B,$5F,$CC,$E5,$A9,$F0,$D7,$F2,$E9,$DC,$7A
	.dc.b	$2B,$86,$9F,$CF,$5B,$7F,$D6,$6F,$85,$F4,$3F,$D6,$33,$DD,$FE,$76
	.dc.b	$D7,$C0,$7F,$44,$70,$FF,$AE,$3D,$77,$AB,$F2,$1B,$1F,$39,$AF,$71
	.dc.b	$EB,$3D,$8E,$15,$27,$83,$F4,$A7,$51,$83,$FA,$AD,$74,$7F,$A8,$B5
	.dc.b	$FF,$19,$35,$FF,$66,$FA,$AF,$67,$F8,$93,$63,$FA,$FB,$5F,$F6,$83
	.dc.b	$5F,$EB,$BE,$C7,$DF,$7F,$B2,$B5,$FF,$6B,$35,$DE,$7F,$C7,$FD,$87
	.dc.b	$B0,$F2,$7F,$B6,$C6,$1B,$83,$C9,$CD,$6F,$0F,$F7,$16,$FF,$8F,$E0
	.dc.b	$FC,$EF,$1F,$E4,$6F,$FC,$B5,$39,$AF,$B9,$39,$AF,$C5,$E7,$FE,$62
	.dc.b	$3B,$38,$39,$B8,$9F,$AD,$5F,$E4,$B7,$F2,$B1,$FE,$5B,$8C,$47,$E2
	.dc.b	$23,$1E,$AF,$58,$F1,$B6,$BF,$36,$D8,$EA,$DE,$3F,$CB,$3C,$A7,$F1
	.dc.b	$CD,$9F,$F0,$68,$F4,$9E,$97,$F8,$C2,$FF,$C6,$36,$F5,$32,$FE,$FA
	.dc.b	$AD,$FE,$7E,$35,$7F,$E8,$19,$CD,$F2,$BB,$5F,$EA,$DA,$D1,$DD,$5F
	.dc.b	$F9,$26,$76,$FF,$AB,$23,$A9,$E1,$9E,$11,$B7,$9C,$7F,$16,$37,$1F
	.dc.b	$A5,$BE,$71,$8B,$FF,$1B,$46,$2F,$BD,$7C,$3B,$EB,$7B,$3D,$07,$84
	.dc.b	$ED,$3C,$2B,$94,$F8,$AB,$E2,$4E,$63,$9B,$6B,$8C,$6E,$FE,$33,$63
	.dc.b	$E1,$9E,$1A,$F8,$55,$C5,$EB,$3B,$7E,$14,$F9,$5C,$B6,$BA,$B4,$EB
	.dc.b	$5C,$E3,$33,$F1,$EB,$CE,$75,$AF,$14,$7C,$57,$D2,$73,$7C,$27,$17
	.dc.b	$2D,$3C,$65,$9E,$E7,$D2,$9E,$F3,$CC,$EF,$7F,$31,$C5,$F9,$DF,$4C
	.dc.b	$F9,$63,$17,$E2,$BE,$95,$E2,$6C,$78,$9A,$78,$A3,$EE,$C7,$95,$EA
	.dc.b	$78,$2B,$B1,$1C,$12,$38,$1E,$17,$9C,$F6,$79,$C6,$9F,$6C,$FC,$08
	.dc.b	$C5,$7A,$A7,$27,$A4,$E6,$48,$FC,$1C,$F8,$DE,$2D,$AF,$04,$78,$CD
	.dc.b	$CF,$9E,$32,$F6,$8F,$88,$F5,$7E,$F1,$83,$89,$D9,$E4,$6E,$FE,$B4
	.dc.b	$64,$93,$DA,$FD,$84,$BF,$35,$F1,$7E,$14,$ED,$CD,$78,$23,$13,$E0
	.dc.b	$8C,$3C,$4F,$8C,$E0,$87,$AA,$3E,$F3,$8B,$F3,$C7,$48,$F3,$CF,$37
	.dc.b	$E0,$1C,$5F,$B3,$1B,$23,$B7,$AC,$6C,$FA,$05,$8F,$9D,$1A,$9F,$0C
	.dc.b	$71,$3A,$96,$78,$C6,$9F,$45,$7C,$69,$D1,$D0,$E3,$1A,$7E,$1B,$E1
	.dc.b	$BF,$A1,$1F,$26,$31,$3F,$39,$B1,$9A,$F0,$46,$0F,$8A,$75,$3E,$2B
	.dc.b	$A2,$4F,$96,$35,$7E,$F8,$C4,$73,$8F,$92,$EA,$BD,$2E,$5F,$6D,$FA
	.dc.b	$23,$C2,$64,$E0,$7C,$6A,$EB,$7E,$2A,$F8,$2D,$9F,$3C,$E8,$93,$A6
	.dc.b	$C6,$A3,$04,$70,$4B,$F6,$46,$5B,$19,$8F,$5D,$78,$67,$24,$8C,$D7
	.dc.b	$82,$31,$5E,$B9,$E8,$68,$AE,$4E,$C5,$8E,$31,$F8,$38,$C3,$C5,$DB
	.dc.b	$92,$5F,$F4,$B8,$FE,$E2,$7F,$72,$BF,$EF,$AE,$0F,$E0,$65,$72,$F0
	.dc.b	$7F,$7B,$EA,$7F,$53,$8F,$D1,$1F,$91,$71,$1A,$AF,$E9,$13,$A8,$FE
	.dc.b	$E5,$3A,$C3,$F9,$E3,$6B,$FA,$0B,$6B,$F9,$3B,$70,$FE,$1C,$D9,$F2
	.dc.b	$57,$F7,$08,$C5,$7F,$6A,$6B,$E5,$9F,$1D,$F9,$63,$F9,$D9,$E5,$AE
	.dc.b	$53,$A7,$CA,$D3,$CA,$58,$F2,$B6,$FC,$AD,$3C,$A7,$F8,$2A,$FC,$77
	.dc.b	$F0,$A3,$55,$E7,$6B,$E7,$3C,$BE,$83,$D3,$FF,$11,$EC,$F4,$DE,$A7
	.dc.b	$F8,$9A,$D7,$BC,$3F,$CB,$F5,$A7,$FD,$73,$19,$9F,$F2,$FC,$E5,$FF
	.dc.b	$47,$46,$73,$57,$F8,$E3,$0F,$06,$1D,$EB,$95,$FE,$25,$AF,$E3,$9A
	.dc.b	$FF,$90,$1A,$F9,$B7,$F5,$E2,$F5,$3A,$46,$1B,$8F,$A9,$B5,$E8,$6B
	.dc.b	$E1,$9F,$3B,$FE,$60,$BF,$F3,$7B,$DE,$BF,$CC,$6F,$FA,$2F,$D1,$FF
	.dc.b	$1B,$F9,$5F,$27,$9D,$18,$AF,$F6,$FC,$61,$FF,$21,$CE,$E7,$55,$FE
	.dc.b	$B5,$AD,$3F,$E7,$06,$BF,$AD,$6B,$FB,$26,$BF,$E6,$4F,$4D,$E7,$FC
	.dc.b	$4F,$1F,$EE,$46,$31,$FB,$71,$89,$D4,$7F,$BD,$27,$55,$F7,$2B,$53
	.dc.b	$F0,$27,$33,$FB,$F6,$BF,$C1,$1A,$FF,$9D,$BD,$2E,$7C,$3F,$33,$85
	.dc.b	$E9,$7B,$A3,$E7,$D2,$B7,$87,$E1,$EB,$4F,$F1,$08,$CC,$AF,$9E,$9F
	.dc.b	$99,$9C,$C6,$2B,$FD,$0F,$87,$69,$F2,$DF,$7C,$FC,$2F,$56,$BE,$96
	.dc.b	$AF,$41,$E3,$FE,$AB,$37,$A3,$FA,$99,$CE,$5F,$EC,$69,$CB,$E2,$3F
	.dc.b	$94,$C6,$AF,$FD,$8F,$3A,$7F,$DE,$D1,$9C,$BE,$44,$E5,$F8,$FE,$8B
	.dc.b	$87,$F7,$0D,$C3,$B6,$1F,$CC,$FD,$1F,$F5,$3F,$96,$F3,$FF,$9C,$DE
	.dc.b	$E9,$8F,$4C,$79,$9F,$C0,$E3,$FD,$5F,$87,$2B,$E3,$39,$94,$E5,$FF
	.dc.b	$5A,$DF,$5C,$3C,$11,$E3,$0E,$14,$6E,$FF,$D9,$F5,$CE,$73,$7F,$B4
	.dc.b	$BE,$11,$E4,$39,$D7,$93,$1E,$11,$99,$ED,$87,$58,$F1,$46,$2F,$A5
	.dc.b	$F8,$27,$99,$EA,$8C,$BC,$23,$B3,$A3,$FB,$1E,$1E,$7B,$ED,$18,$34
	.dc.b	$E1,$F2,$4F,$2B,$F5,$C7,$CA,$37,$7E,$07,$94,$79,$F0,$F6,$C6,$E3
	.dc.b	$0F,$8C,$F8,$87,$F1,$6C,$3F,$8B,$B8,$3D,$C7,$F7,$7F,$3D,$E8,$3D
	.dc.b	$1F,$E3,$B1,$86,$C6,$67,$FB,$F0,$9D,$38,$5C,$DF,$8C,$8C,$35,$F0
	.dc.b	$35,$FF,$05,$C7,$0D,$7C,$33,$C7,$1F,$C3,$7A,$2F,$47,$FC,$23,$18
	.dc.b	$6E,$DC,$AB,$83,$6F,$F1,$57,$C7,$1B,$7C,$11,$EB,$BE,$2E,$31,$F2
	.dc.b	$3F,$7E,$B9,$36,$B9,$3C,$6F,$F3,$15,$C3,$1F,$34,$F8,$0C,$E1,$D9
	.dc.b	$E0,$73,$34,$7E,$36,$79,$63,$EB,$8D,$12,$BF,$55,$3A,$8F,$45,$75
	.dc.b	$C7,$CF,$5E,$29,$E9,$39,$C7,$FC,$77,$8F,$F3,$2B,$F6,$B6,$33,$AF
	.dc.b	$87,$0E,$53,$BE,$3C,$E3,$F2,$B0,$FE,$4A,$FF,$12,$0F,$4B,$FF,$24
	.dc.b	$FC,$87,$F6,$69,$D7,$9C,$FA,$CE,$70,$8C,$3C,$E7,$83,$FC,$62,$B9
	.dc.b	$F5,$70,$AD,$CE,$4F,$BA,$E3,$2F,$E9,$F7,$C9,$BE,$97,$1F,$94,$97
	.dc.b	$DE,$3C,$74,$73,$1D,$4F,$34,$EA,$24,$7B,$23,$75,$DA,$3E,$27,$E9
	.dc.b	$8F,$A9,$FD,$AB,$63,$FC,$1A,$D7,$F5,$ED,$7F,$CD,$CE,$67,$FB,$67
	.dc.b	$84,$F9,$1F,$A7,$E0,$23,$17,$E0,$C3,$E6,$E1,$C7,$6F,$F1,$A5,$6A
	.dc.b	$BE,$45,$F1,$C3,$78,$7D,$4F,$59,$F5,$7F,$C3,$1B,$7F,$E2,$46,$BF
	.dc.b	$C1,$9A,$FF,$A1,$5C,$E7,$E8,$D3,$CF,$91,$ED,$72,$FF,$AD,$E5,$EB
	.dc.b	$9D,$99,$8F,$A5,$E9,$BD,$27,$8D,$FE,$36,$8C,$37,$D3,$7A,$79,$AD
	.dc.b	$DF,$F1,$CF,$7D,$EF,$FF,$80,$1B,$C5,$FE,$40,$6C,$EE,$7F,$C0,$BF
	.dc.b	$04,$F8,$3A,$4B,$FF,$23,$C6,$5F,$F2,$4F,$D8,$B8,$FF,$C1,$AD,$8C
	.dc.b	$C6,$08,$FE,$AE,$35,$87,$E7,$61,$EB,$3E,$A7,$FC,$F5,$5A,$E5,$FF
	.dc.b	$16,$B6,$BF,$D7,$2D,$BF,$F3,$AD,$F1,$9E,$D8,$7F,$A0,$C7,$D6,$FF
	.dc.b	$88,$9B,$1F,$7E,$30,$47,$F8,$04,$6A,$BF,$17,$CD,$D5,$FD,$44,$67
	.dc.b	$1F,$F2,$0B,$6B,$FD,$9A,$D9,$FF,$1F,$57,$1A,$DD,$F9,$7E,$A7,$C8
	.dc.b	$4C,$3F,$24,$DB,$EE,$FF,$A8,$2D,$F3,$5C,$6B,$CF,$36,$BC,$F3,$70
	.dc.b	$FF,$2A,$D7,$F6,$63,$FC,$5A,$7C,$FF,$5B,$FD,$50,$E6,$BE,$A6,$BF
	.dc.b	$EA,$3A,$E3,$E4,$FF,$A3,$1F,$F4,$6D,$71,$3F,$93,$1C,$5C,$CF,$18
	.dc.b	$FF,$77,$99,$3F,$AB,$9C,$6B,$F8,$93,$97,$FD,$5F,$1C,$5D,$D7,$86
	.dc.b	$78,$CE,$67,$FD,$6B,$39,$78,$DF,$B2,$31,$5D,$EF,$D6,$FC,$2B,$E4
	.dc.b	$F8,$4D,$1E,$89,$E3,$1A,$6D,$75,$9E,$4D,$BF,$92,$DB,$F8,$6D,$9F
	.dc.b	$4B,$67,$CF,$1F,$2A,$BA,$B4,$EA,$58,$EA,$DA,$EA,$D3,$A9,$63,$A9
	.dc.b	$F3,$A7,$AB,$63,$AB,$4E,$A5,$9E,$AD,$8E,$AD,$3A,$C7,$CC,$7F,$C5
	.dc.b	$B1,$F8,$B7,$FC,$2A,$7D,$A7,$FD,$A2,$65,$F9,$CF,$9C,$FE,$07,$87
	.dc.b	$C5,$8F,$BB,$19,$D7,$F4,$2F,$B6,$32,$78,$DE,$26,$4F,$C6,$3A,$AF
	.dc.b	$CC,$CF,$28,$F5,$19,$9F,$1B,$FC,$56,$35,$3E,$D2,$4E,$9E,$75,$E7
	.dc.b	$8E,$E7,$E2,$E7,$8B,$B2,$57,$C3,$3B,$4F,$1B,$FC,$24,$FA,$A7,$CA
	.dc.b	$F4,$8F,$CD,$D7,$49,$E5,$7E,$46,$D7,$EB,$A7,$51,$ED,$32,$48,$ED
	.dc.b	$3C,$63,$46,$E3,$27,$A1,$CB,$ED,$BF,$43,$F5,$5B,$1F,$4E,$BC,$33
	.dc.b	$E9,$34,$7D,$5B,$F7,$C6,$9E,$37,$F3,$27,$47,$0A,$F8,$73,$C2,$F7
	.dc.b	$3F,$30,$F3,$CF,$B8,$F3,$CF,$C8,$9D,$C7,$C1,$BF,$05,$76,$8F,$A6
	.dc.b	$DB,$F9,$8F,$48,$E6,$F8,$E3,$E4,$3E,$D8,$E2,$79,$9D,$1F,$24,$D4
	.dc.b	$7B,$63,$47,$C7,$AF,$2B,$F4,$6B,$CA,$EA,$78,$1E,$0B,$F8,$57,$E7
	.dc.b	$3E,$15,$70,$9D,$47,$F4,$08,$C3,$E3,$9F,$88,$F1,$AF,$89,$7E,$28
	.dc.b	$FB,$C6,$63,$C9,$19,$9E,$31,$E5,$9F,$0C,$E5,$F0,$36,$3E,$2D,$F9
	.dc.b	$1F,$81,$3D,$4E,$EF,$33,$C6,$E6,$BB,$CF,$13,$EA,$DF,$B2,$BF,$01
	.dc.b	$5C,$A7,$4F,$90,$E4,$79,$0B,$1E,$47,$E2,$1F,$61,$E7,$5F,$39,$F2
	.dc.b	$1F,$0A,$38,$CF,$D4,$78,$C7,$B0,$DB,$E3,$33,$1E,$8B,$F8,$8F,$19
	.dc.b	$EF,$3E,$17,$F1,$E7,$38,$F5,$BF,$10,$EB,$5F,$12,$7A,$9E,$F8,$CB
	.dc.b	$EF,$69,$EF,$3B,$9D,$5E,$2F,$59,$EE,$FB,$CE,$2F,$9D,$F2,$BE,$B8
	.dc.b	$F8,$67,$F1,$1A,$F6,$B4,$CC,$FF,$8E,$23,$2D,$8C,$9F,$AF,$3F,$4E
	.dc.b	$7A,$E7,$C6,$FC,$B8,$F4,$C7,$85,$F2,$C7,$76,$9D,$4A,$F5,$AF,$E0
	.dc.b	$6D,$3F,$44,$58,$FC,$93,$67,$E8,$B4,$F8,$E5,$8F,$1B,$E1,$8F,$E5
	.dc.b	$9F,$A5,$BD,$14,$FF,$64,$AF,$EC,$B3,$98,$C4,$7F,$BF,$23,$0E,$1A
	.dc.b	$F1,$73,$5D,$23,$85,$7F,$91,$FC,$BF,$F1,$24,$2B,$77,$E0,$AC,$C6
	.dc.b	$27,$FC,$AF,$58,$9C,$FC,$BF,$59,$F0,$35,$F4,$B5,$F4,$35,$F0,$3C
	.dc.b	$AF,$FD,$47,$BF,$FB,$9D,$FF,$77,$60,$9D,$57,$FB,$56,$31,$3C,$3C
	.dc.b	$6C,$E5,$E3,$97,$57,$76,$D7,$CA,$BF,$81,$39,$24,$78,$1F,$F3,$67
	.dc.b	$B9,$F0,$99,$19,$3C,$F1,$89,$FF,$74,$7D,$9B,$E7,$7D,$CD,$AF,$03
	.dc.b	$67,$ED,$B5,$FC,$03,$9C,$7E,$63,$65,$2F,$05,$F1,$DF,$EB,$8B,$DE
	.dc.b	$1F,$73,$0E,$17,$C1,$B7,$C3,$B3,$55,$F8,$2F,$6B,$FD,$D9,$97,$C5
	.dc.b	$6E,$1F,$90,$6B,$F9,$56,$BF,$E3,$A7,$78,$FE,$0D,$B3,$E8,$C1,$EE
	.dc.b	$3E,$27,$C1,$C5,$2B,$DB,$87,$18,$FF,$3C,$5F,$0A,$D6,$2A,$FC,$87
	.dc.b	$B8,$F2,$FF,$18,$DC,$7E,$F3,$6B,$F5,$2D,$7F,$5C,$E6,$B9,$D7,$83
	.dc.b	$27,$C7,$7E,$9A,$B5,$87,$F9,$61,$B8,$7D,$46,$E5,$EE,$AE,$18,$7E
	.dc.b	$BB,$D9,$FF,$6A,$59,$FF,$6B,$78,$CF,$8C,$7B,$BB,$F0,$E6,$9F,$10
	.dc.b	$FF,$33,$DF,$0A,$DC,$67,$0F,$C2,$CF,$62,$57,$FA,$87,$0E,$D8,$74
	.dc.b	$AF,$95,$7D,$3C,$EF,$31,$BF,$9C,$7C,$3E,$27,$B5,$7A,$6B,$0F,$21
	.dc.b	$82,$F8,$D7,$0C,$97,$F0,$FD,$D9,$3D,$30,$FC,$EF,$BB,$F9,$EF,$65
	.dc.b	$E1,$7C,$0B,$DD,$77,$3E,$53,$FC,$69,$E2,$27,$8B,$F1,$23,$4A,$BB
	.dc.b	$57,$5B,$B3,$FC,$C6,$6B,$DA,$2B,$DA,$FA,$B9,$3C,$15,$FC,$DA,$B8
	.dc.b	$E2,$C3,$F9,$95,$F0,$BE,$57,$E1,$78,$9E,$12,$9D,$63,$83,$E9,$33
	.dc.b	$5D,$70,$F2,$C6,$67,$CB,$39,$8E,$67,$80,$F8,$93,$C3,$0F,$3E,$1E
	.dc.b	$6C,$3D,$55,$DA,$30,$F9,$70,$EB,$18,$8F,$F5,$DD,$72,$BE,$91,$C5
	.dc.b	$AF,$F3,$18,$FE,$64,$FD,$83,$07,$F5,$EF,$71,$E2,$3D,$DF,$E7,$11
	.dc.b	$86,$C6,$67,$83,$5F,$ED,$52,$F7,$7F,$84,$47,$56,$BA,$6B,$FE,$D4
	.dc.b	$8E,$1D,$3E,$09,$E3,$9F,$E0,$3D,$B7,$BB,$FE,$DA,$8C,$37,$6E,$55
	.dc.b	$C1,$B7,$F8,$CB,$E3,$8D,$BE,$0F,$F3,$EC,$7C,$2F,$E0,$9B,$8F,$68
	.dc.b	$E0,$7E,$5A,$33,$87,$68,$FC,$4D,$F6,$6D,$F6,$55,$C1,$F0,$63,$E7
	.dc.b	$7F,$A7,$D7,$1C,$3C,$11,$E3,$9D,$3C,$9B,$3B,$C3,$FD,$FA,$D9,$EB
	.dc.b	$8F,$9E,$B7,$8F,$8E,$B5,$5C,$EB,$DA,$FE,$FE,$3D,$51,$FE,$0F,$6E
	.dc.b	$BE,$FC,$7C,$B3,$9F,$77,$FB,$3D,$F1,$6C,$71,$C2,$DE,$BD,$B7,$BB
	.dc.b	$8C,$3F,$B7,$FB,$AF,$2B,$10,$8C,$3F,$C9,$2B,$8B,$6B,$73,$A8,$FE
	.dc.b	$30,$DC,$7F,$C5,$15,$C2,$7F,$8E,$3F,$D6,$DF,$E5,$F2,$F6,$B9,$9E
	.dc.b	$29,$D4,$F0,$ED,$CB,$C2,$75,$12,$30,$43,$81,$FC,$F3,$DD,$7B,$BE
	.dc.b	$E6,$C6,$63,$E2,$78,$2F,$15,$F1,$DB,$9F,$DD,$DF,$0B,$FF,$2B,$B5
	.dc.b	$FF,$2E,$35,$FA,$AE,$F4,$FF,$2B,$61,$D7,$B3,$BE,$3F,$D2,$2F,$C4
	.dc.b	$7F,$52,$F0,$5E,$F7,$D4,$6C,$66,$38,$FA,$F9,$7F,$87,$CE,$A7,$FC
	.dc.b	$D6,$D7,$FC,$E0,$D7,$F1,$2E,$F1,$F2,$DF,$D3,$9F,$F2,$BE,$DD,$5F
	.dc.b	$EC,$BF,$11,$F1,$5F,$F2,$73,$5F,$F3,$CB,$5F,$F3,$F3,$5F,$CB,$FB
	.dc.b	$EF,$79,$EF,$FF,$95,$E3,$0D,$F4,$FF,$B7,$D7,$FA,$12,$72,$FF,$A1
	.dc.b	$BD,$F7,$A3,$FE,$6F,$6F,$57,$F9,$85,$B3,$B9,$FF,$3A,$7B,$AD,$FF
	.dc.b	$D1,$1B,$FF,$A2,$70,$7B,$5F,$93,$7F,$C0,$75,$B7,$FC,$24,$6B,$7F
	.dc.b	$F4,$6C,$67,$1F,$F4,$CB,$5F,$F4,$F3,$5F,$DB,$B5,$FF,$2F,$BF,$E2
	.dc.b	$83,$95,$F8,$ED,$EE,$76,$C7,$FD,$36,$38,$FF,$C9,$3C,$5F,$E3,$08
	.dc.b	$C6,$7F,$E8,$E7,$73,$FE,$93,$73,$97,$E1,$23,$04,$9F,$C8,$3D,$2F
	.dc.b	$8C,$7E,$4F,$DD,$EF,$F4,$1E,$26,$C6,$4F,$F4,$86,$08,$F0,$12,$32
	.dc.b	$F1,$7F,$D4,$7E,$ED,$C7,$FC,$41,$1F,$D9,$0F,$9F,$83,$DE,$FF,$5D
	.dc.b	$C6,$67,$FC,$68,$FC,$68,$CF,$93,$F7,$23,$FD,$16,$64,$FF,$1E,$9C
	.dc.b	$76,$FC,$05,$F5,$8C,$CF,$1B,$FF,$6B,$CE,$5F,$2C,$66,$7A,$46,$1F
	.dc.b	$3D,$F8,$9A,$74,$2C,$6C,$F4,$D7,$5B,$FF,$67,$4F,$94,$F4,$C6,$5F
	.dc.b	$E7,$E6,$67,$CB,$5D,$DE,$6D,$AF,$6B,$67,$9B,$67,$8C,$79,$5E,$2D
	.dc.b	$9E,$2D,$38,$96,$38,$9E,$B8,$DB,$63,$6D,$37,$1E,$EB,$F9,$57,$ED
	.dc.b	$9F,$AF,$18,$BF,$35,$FC,$EB,$E0,$FC,$A6,$D7,$A0,$ED,$1B,$3E,$3D
	.dc.b	$73,$7D,$06,$63,$BB,$CC,$D9,$5C,$9D,$67,$B9,$EE,$2C,$7B,$9B,$1E
	.dc.b	$E8,$E6,$F1,$6C,$71,$69,$C4,$B1,$C4,$F0,$C6,$DB,$1B,$69,$B0,$FE
	.dc.b	$14,$F7,$38,$9F,$84,$FC,$6B,$FA,$0E,$67,$E0,$34,$F4,$1D,$9E,$67
	.dc.b	$58,$E2,$DB,$FA,$F1,$8A,$EF,$7F,$12,$B8,$C6,$1F,$79,$83,$C2,$73
	.dc.b	$3E,$BC,$F3,$73,$1C,$6F,$C9,$39,$7C,$E7,$A6,$BE,$43,$6B,$E4,$34
	.dc.b	$F9,$0F,$77,$E2,$5F,$A8,$91,$E2,$6C,$6A,$30,$47,$04,$9F,$81,$5E
	.dc.b	$57,$C7,$3D,$DC,$D7,$46,$C7,$36,$D7,$38,$F8,$F5,$CD,$F5,$9C,$CC
	.dc.b	$D7,$36,$9C,$CB,$3C,$CF,$93,$1B,$6C,$6D,$A6,$CF,$90,$FD,$77,$E7
	.dc.b	$BF,$CC,$DF,$F2,$99,$F8,$92,$B9,$7F,$46,$62,$FF,$2E,$F0,$38,$1A
	.dc.b	$21,$F9,$E3,$C3,$B3,$1F,$E3,$51,$F6,$CC,$B8,$AF,$E4,$D1,$87,$F8
	.dc.b	$C3,$B7,$F4,$D3,$CD,$C9,$C0,$CC,$F3,$AF,$E2,$73,$BA,$FE,$38,$73
	.dc.b	$8F,$E8,$91,$9C,$7F,$81,$CE,$6F,$FC,$13,$39,$7F,$7F,$19,$3F,$AD
	.dc.b	$C6,$5F,$81,$1E,$3A,$EB,$7E,$38,$FE,$37,$5B,$AE,$57,$FA,$F8,$C4
	.dc.b	$F1,$37,$5C,$6B,$F9,$3C,$EA,$F1,$7F,$E1,$1A,$B7,$FB,$6A,$D3,$F1
	.dc.b	$AB,$C6,$D3,$A9,$6F,$AD,$72,$7F,$85,$57,$27,$E4,$CF,$CE,$38,$5F
	.dc.b	$A6,$B8,$4F,$D9,$37,$7F,$4A,$33,$5C,$67,$2F,$89,$A7,$89,$EA,$7A
	.dc.b	$E3,$8C,$FA,$A7,$75,$E1,$AD,$C7,$A2,$B7,$1F,$69,$B1,$F6,$1B,$3F
	.dc.b	$26,$78,$CF,$A6,$3F,$07,$1F,$4E,$38,$3A,$3D,$51,$AA,$F7,$46,$5F
	.dc.b	$34,$66,$3D,$CD,$9F,$71,$EA,$8F,$71,$C4,$F4,$15,$F0,$9B,$3D,$CE
	.dc.b	$63,$D5,$19,$3D,$51,$F9,$83,$D5,$1A,$3C,$47,$C9,$8E,$11,$EC,$38
	.dc.b	$4F,$B2,$32,$FB,$23,$F4,$7E,$E3,$DE,$FF,$2E,$E9,$FE,$33,$B4,$E6
	.dc.b	$90,$C3,$FD,$33,$5B,$AD,$4F,$F0,$D6,$BF,$C5,$9A,$FF,$BA,$9A,$F4
	.dc.b	$78,$CF,$F1,$68,$F0,$DF,$F9,$C3,$CA,$EC,$F7,$F7,$BF,$CE,$7F,$3F
	.dc.b	$72,$F2,$E1,$FE,$AC,$8C,$61,$E3,$AE,$8F,$85,$AF,$85,$AF,$C0,$78
	.dc.b	$CF,$FA,$DB,$C7,$B7,$E9,$EA,$FF,$41,$FD,$87,$C1,$6C,$78,$8D,$1F
	.dc.b	$CB,$F0,$F9,$71,$DB,$43,$49,$3C,$67,$E2,$47,$FA,$43,$E0,$3B,$4E
	.dc.b	$65,$EF,$BD,$FF,$F3,$57,$88,$F8,$0F,$F6,$DD,$EE,$B4,$F5,$9D,$78
	.dc.b	$9F,$7E,$73,$87,$F4,$56,$BF,$D3,$9A,$FE,$08,$F0,$E1,$F2,$23,$8E
	.dc.b	$DF,$65,$B8,$7E,$23,$0F,$27,$7F,$7B,$FC,$77,$C0,$7F,$C0,$1A,$F2
	.dc.b	$7B,$EF,$86,$CF,$81,$EC,$8C,$4F,$C1,$AE,$1A,$7F,$5B,$6B,$FD,$91
	.dc.b	$AF,$9D,$AF,$E6,$9E,$3A,$7E,$3D,$B3,$E7,$A7,$C0,$F8,$A3,$E8,$7C
	.dc.b	$0B,$CE,$F2,$7C,$A7,$E9,$E3,$13,$EA,$8C,$4F,$E3,$27,$39,$7F,$71
	.dc.b	$6B,$FD,$E9,$AF,$CD,$77,$97,$E8,$DB,$7E,$5F,$82,$EA,$FE,$0F,$DD
	.dc.b	$FF,$6D,$BF,$55,$B6,$F7,$DF,$04,$F8,$2F,$DF,$56,$A8,$D3,$8D,$70
	.dc.b	$9D,$F8,$F3,$0F,$03,$FE,$B9,$9D,$46,$0F,$27,$A3,$D8,$D7,$C0,$BE
	.dc.b	$0B,$13,$A2,$65,$FC,$35,$FE,$1F,$6A,$E1,$1A,$DB,$F8,$36,$6F,$82
	.dc.b	$F5,$9F,$C6,$70,$FE,$2F,$F0,$5F,$C6,$3D,$E7,$8F,$F1,$6B,$78,$FC
	.dc.b	$97,$85,$AF,$85,$6E,$BF,$61,$3D,$05,$F4,$7F,$92,$E1,$E3,$9E,$31
	.dc.b	$FE,$36,$BA,$FC,$EC,$3F,$C6,$77,$C6,$3F,$59,$3A,$BE,$51,$DB,$0F
	.dc.b	$F1,$C4,$E6,$F8,$C6,$27,$47,$8A,$32,$F8,$4C,$9E,$1C,$3C,$6F,$AF
	.dc.b	$0F,$16,$1E,$07,$C9,$19,$BF,$84,$F2,$30,$77,$AE,$6F,$C9,$3B,$D7
	.dc.b	$39,$F3,$57,$6B,$F2,$D7,$C0,$36,$E1,$EA,$79,$CF,$01,$D6,$39,$5F
	.dc.b	$83,$0F,$E5,$D3,$F5,$4F,$E6,$0E,$1F,$21,$FE,$68,$F8,$17,$94,$F8
	.dc.b	$2F,$E6,$D1,$86,$C6,$67,$FC,$D1,$39,$6F,$C1,$7F,$63,$6E,$BC,$A3
	.dc.b	$11,$DD,$AF,$F2,$99,$47,$1C,$BF,$05,$F0,$4F,$81,$FE,$5D,$18,$6E
	.dc.b	$DC,$6B,$83,$71,$FC,$6D,$F1,$C6,$DF,$07,$FB,$CE,$3E,$0A,$FC,$9B
	.dc.b	$63,$F0,$B5,$C5,$B5,$C5,$4E,$F1,$FC,$C3,$70,$FE,$89,$7C,$67,$D1
	.dc.b	$5D,$B1,$F3,$BE,$26,$A9,$E1,$5C,$B0,$F0,$61,$D6,$B8,$63,$EC,$61
	.dc.b	$F9,$C8,$D3,$E6,$C3,$AE,$1D,$23,$FC,$32,$D7,$55,$C8,$FF,$08,$47
	.dc.b	$C1,$6D,$FF,$5F,$7F,$C2,$97,$DA,$75,$AF,$C0,$F8,$2D,$37,$0F,$EE
	.dc.b	$D3,$A7,$FC,$53,$F0,$0F,$82,$DC,$7F,$8B,$3E,$09,$E7,$62,$11,$87
	.dc.b	$FC,$01,$5C,$5B,$5B,$9D,$47,$F2,$93,$A5,$73,$AE,$11,$90,$FF,$5B
	.dc.b	$7F,$C1,$92,$F6,$BA,$1C,$7B,$E4,$EA,$32,$F9,$23,$2F,$59,$E4,$EC
	.dc.b	$D1,$FE,$19,$F8,$07,$C0,$FB,$DB,$19,$8F,$8B,$E5,$3C,$77,$C1,$6E
	.dc.b	$65,$6C,$FF,$0E,$4E,$9C,$1F,$ED,$FB,$EA,$D9,$FE,$0F,$18,$8F,$F7
	.dc.b	$5F,$80,$F3,$7E,$93,$63,$FC,$60,$D7,$E7,$63,$E5,$F7,$37,$52,$76
	.dc.b	$FA,$B1,$4E,$67,$FD,$E5,$7C,$DF,$0D,$7F,$1A,$8C,$47,$F8,$BF,$D1
	.dc.b	$FE,$D0,$FF,$92,$DA,$FF,$BE,$DA,$FF,$8D,$4F,$F2,$6F,$85,$FC,$93
	.dc.b	$E0,$BE,$54,$7F,$94,$DA,$EC,$38,$79,$35,$FF,$20,$7C,$03,$E0,$9F
	.dc.b	$05,$FE,$5D,$8C,$37,$0F,$F0,$85,$75,$9C,$BD,$FE,$07,$FC,$2B,$D2
	.dc.b	$F7,$BF,$20,$FF,$97,$63,$31,$89,$FC,$C4,$63,$9B,$FD,$2C,$DA,$E1
	.dc.b	$AF,$F9,$EA,$0D,$24,$7F,$9C,$DF,$A6,$FE,$21,$FF,$50,$18,$AF,$BB
	.dc.b	$C1,$FE,$25,$8C,$76,$AB,$FC,$B9,$39,$E1,$FF,$38,$B6,$3B,$47,$FA
	.dc.b	$32,$BA,$5F,$E3,$A3,$1E,$EF,$0F,$81,$F0,$B6,$FF,$1E,$DA,$FD,$C5
	.dc.b	$A3,$18,$FF,$A9,$E3,$15,$ED,$9E,$AD,$3F,$D2,$47,$F6,$03,$FD,$1C
	.dc.b	$3E,$05,$F0,$3F,$EC,$18,$D6,$1F,$C2,$DD,$46,$7C,$6F,$85,$39,$AE
	.dc.b	$35,$FE,$95,$38,$E3,$A7,$89,$9A,$E3,$19,$9E,$35,$FC,$56,$B8,$9F
	.dc.b	$ED,$03,$8B,$9A,$E3,$19,$9E,$35,$FE,$C9,$8E,$35,$F5,$E7,$31,$FE
	.dc.b	$99,$32,$74,$9E,$15,$D2,$B8,$5F,$BE,$BE,$3D,$F2,$9F,$25,$6E,$79
	.dc.b	$1D,$5C,$4E,$E7,$F1,$11,$8A,$E9,$1F,$25,$B7,$F0,$DB,$7E,$D6,$CF
	.dc.b	$92,$3E,$51,$E4,$8C,$BE,$28,$F2,$B9,$36,$F7,$6C,$F7,$69,$DC,$B1
	.dc.b	$DC,$FA,$0F,$36,$9C,$CB,$1C,$DB,$1C,$DA,$73,$8F,$9F,$1C,$DB,$1C
	.dc.b	$DA,$73,$2C,$73,$6C,$73,$8F,$A0,$73,$2B,$CD,$B1,$CD,$A7,$32,$C7
	.dc.b	$33,$E4,$9C,$4B,$1C,$5B,$1C,$5A,$71,$3E,$FB,$E3,$72,$FD,$57,$27
	.dc.b	$52,$C7,$56,$C7,$56,$9D,$5F,$CC,$DF,$E3,$5B,$5F,$82,$8C,$5F,$E3
	.dc.b	$E7,$ED,$9E,$13,$66,$23,$6F,$9E,$BB,$BB,$9E,$51,$F6,$A3,$06,$CE
	.dc.b	$6E,$67,$9B,$ED,$8C,$DF,$D8,$78,$C6,$6B,$AC,$E6,$B8,$BF,$0E,$B8
	.dc.b	$9F,$58,$E2,$E6,$78,$C6,$67,$8B,$F2,$4E,$35,$F4,$A3,$2F,$CB,$9E
	.dc.b	$2E,$6F,$E5,$12,$BE,$33,$63,$DC,$DB,$F4,$36,$7C,$4D,$9E,$91,$C8
	.dc.b	$8F,$26,$CF,$26,$9C,$8B,$1C,$8E,$04,$8E,$0D,$8E,$0D,$38,$46,$89
	.dc.b	$1A,$6C,$6A,$30,$47,$04,$3F,$0B,$7F,$2A,$3E,$99,$F3,$EF,$E5,$CE
	.dc.b	$AF,$E1,$36,$7D,$6D,$3C,$C5,$BF,$1B,$63,$C4,$D3,$C4,$FD,$C7,$04
	.dc.b	$70,$4A,$F1,$B4,$FC,$C3,$82,$38,$24,$7E,$55,$B3,$F8,$F6,$9E,$F2
	.dc.b	$C7,$E2,$DA,$7C,$47,$07,$32,$47,$36,$C7,$36,$9C,$23,$44,$8D,$36
	.dc.b	$35,$18,$23,$82,$3E,$32,$5F,$D9,$25,$FD,$68,$F8,$D3,$E7,$3F,$C9
	.dc.b	$8F,$F8,$11,$B7,$FE,$02,$6C,$7F,$83,$A3,$FA,$9E,$EF,$81,$FE,$B3
	.dc.b	$5F,$C6,$8F,$E1,$B1,$9A,$FE,$9F,$19,$7F,$BA,$C6,$63,$FA,$49,$9A
	.dc.b	$FE,$E1,$19,$C7,$FA,$1C,$E6,$BF,$C3,$B3,$97,$F9,$EC,$64,$FF,$0E
	.dc.b	$46,$5F,$E7,$26,$67,$FC,$31,$19,$8C,$4F,$F1,$18,$C3,$FE,$0E,$8C
	.dc.b	$9F,$E4,$58,$CD,$7C,$09,$F5,$DD,$8F,$8A,$D8,$FE,$6B,$3C,$1F,$86
	.dc.b	$70,$2C,$F0,$69,$C2,$7E,$6F,$BA,$F8,$1F,$F2,$9E,$9F,$E1,$BC,$53
	.dc.b	$9B,$FF,$53,$CE,$63,$13,$FE,$60,$8C,$57,$2A,$4F,$FB,$26,$B6,$FF
	.dc.b	$A3,$DA,$E8,$EF,$3F,$85,$71,$1A,$79,$AF,$80,$FF,$34,$FC,$0F,$8D
	.dc.b	$B7,$E0,$D3,$F1,$D5,$B8,$FF,$5F,$D6,$DA,$6C,$EF,$5E,$1B,$EE,$D7
	.dc.b	$E1,$B6,$BA,$47,$FA,$8D,$73,$7F,$9F,$3E,$07,$E3,$36,$FB,$57,$AA
	.dc.b	$30,$FC,$56,$E3,$E8,$7B,$12,$7F,$D6,$13,$F5,$23,$C9,$EF,$BD,$EF
	.dc.b	$F2,$97,$7B,$E0,$3F,$DA,$98,$78,$30,$5E,$E3,$C7,$5B,$F0,$3F,$01
	.dc.b	$39,$AF,$F6,$BD,$F2,$59,$7E,$E6,$74,$FF,$9F,$2F,$96,$1D,$E7,$95
	.dc.b	$7F,$B3,$5A,$FD,$F6,$BF,$86,$6A,$F7,$BF,$DF,$53,$E6,$F7,$9E,$FF
	.dc.b	$E5,$67,$35,$FE,$EC,$9C,$D7,$BE,$73,$8F,$96,$73,$5E,$AA,$FC,$85
	.dc.b	$69,$FC,$BE,$2A,$DD,$FF,$17,$AE,$F7,$FE,$E7,$6B,$CB,$B3,$F0,$D1
	.dc.b	$88,$FD,$0F,$BA,$F2,$7E,$AE,$0F,$7B,$F5,$D3,$9C,$FD,$D1,$8A,$DB
	.dc.b	$63,$27,$DB,$9E,$17,$FE,$F3,$6B,$FE,$FB,$6B,$FA,$D6,$BF,$B4,$F6
	.dc.b	$3F,$DB,$F3,$F8,$56,$C7,$C5,$1E,$D3,$DD,$7B,$BF,$C1,$6B,$7B,$7F
	.dc.b	$3B,$6E,$1F,$BB,$6E,$1C,$AF,$8C,$F7,$F0,$79,$11,$FD,$2E,$3D,$F1
	.dc.b	$C5,$F8,$C6,$0F,$77,$C4,$7F,$AC,$69,$EE,$BC,$CC,$E3,$ED,$AB,$87
	.dc.b	$B6,$F7,$58,$D3,$D5,$EE,$FB,$4F,$E4,$55,$FC,$6B,$DD,$FE,$37,$EE
	.dc.b	$3D,$CF,$75,$6F,$3F,$94,$F0,$B5,$F0,$AD,$D7,$EC,$E7,$A0,$BE,$8F
	.dc.b	$D5,$BF,$F7,$3C,$62,$3F,$93,$5F,$37,$55,$5C,$9C,$EF,$BB,$4F,$E5
	.dc.b	$C7,$90,$C1,$F3,$2F,$BC,$6A,$35,$7E,$1D,$55,$A9,$E7,$87,$86,$73
	.dc.b	$87,$8E,$7D,$B7,$C6,$F8,$47,$9A,$31,$3B,$8E,$F8,$7A,$27,$E0,$47
	.dc.b	$28,$C1,$E5,$C3,$C1,$7D,$27,$47,$C4,$3D,$A6,$0F,$75,$F7,$9E,$2D
	.dc.b	$8F,$E5,$A7,$C5,$3F,$97,$C6,$1C,$1F,$D5,$FD,$A7,$8C,$F7,$7F,$9A
	.dc.b	$C6,$1B,$19,$9F,$99,$39,$6F,$BB,$FC,$B1,$BA,$F2,$8C,$47,$D8,$94
	.dc.b	$62,$3C,$4F,$1D,$3F,$0D,$ED,$BD,$DF,$ED,$91,$86,$ED,$C3,$1F,$C4
	.dc.b	$63,$D1,$FC,$55,$F1,$C6,$DF,$07,$FB,$DD,$6D,$FC,$4B,$70,$ED,$1C
	.dc.b	$0E,$B8,$F9,$71,$F0,$47,$E4,$30,$EC,$DC,$3B,$2C,$3A,$57,$0C,$3A
	.dc.b	$CE,$CE,$B8,$C7,$04,$8F,$D4,$57,$28,$F4,$E3,$E3,$7F,$C2,$B8,$F8
	.dc.b	$AB,$A9,$C2,$75,$29,$F3,$DF,$F9,$65,$BA,$FF,$86,$5B,$3D,$A7,$E1
	.dc.b	$3F,$C6,$61,$EE,$FF,$87,$E7,$93,$6B,$96,$76,$FF,$C5,$3E,$DB,$DD
	.dc.b	$C4,$3D,$C7,$BB,$86,$C7,$F7,$FA,$E0,$DA,$D5,$48,$C1,$BE,$EE,$5E
	.dc.b	$B4,$8C,$10,$FE,$65,$82,$35,$1F,$CD,$70,$FE,$BC,$EF,$4F,$F3,$F5
	.dc.b	$6E,$3F,$C2,$4D,$78,$5F,$B3,$F5,$0C,$47,$F1,$66,$BF,$E4,$56,$BF
	.dc.b	$E4,$C6,$A7,$A5,$7F,$84,$FF,$32,$FF,$47,$37,$5E,$D9,$4B,$FF,$10
	.dc.b	$D6,$A7,$FC,$02,$D8,$ED,$3E,$5E,$3F,$E0,$B8,$7F,$14,$9F,$24,$60
	.dc.b	$8F,$F8,$B3,$C4,$7D,$97,$DB,$F1,$91,$89,$FF,$1A,$5F,$F3,$DC,$FF
	.dc.b	$D4,$8D,$CF,$FC,$27,$5B,$AF,$EE,$97,$CB,$0F,$24,$7F,$8F,$78,$BE
	.dc.b	$8B,$FC,$E5,$AF,$F4,$06,$BF,$E6,$F6,$BF,$E7,$67,$8E,$3F,$EB,$5F
	.dc.b	$C4,$1F,$37,$F9,$06,$D7,$7D,$FF,$27,$7C,$33,$FF,$3D,$56,$AB,$FB
	.dc.b	$C3,$70,$FF,$60,$56,$E7,$FD,$0B,$EE,$3D,$E7,$BD,$FE,$5E,$8C,$37
	.dc.b	$D1,$FF,$09,$D7,$F5,$59,$CB,$FD,$5F,$E1,$DF,$0D,$FE,$77,$6E,$9F
	.dc.b	$E6,$86,$CE,$E7,$FC,$F9,$F1,$0F,$89,$A4,$BF,$F3,$7C,$65,$FF,$38
	.dc.b	$F2,$7F,$5E,$BF,$EC,$CD,$7F,$31,$E2,$7F,$A1,$E3,$11,$F9,$98,$FF
	.dc.b	$56,$47,$1D,$D5,$FE,$A6,$F8,$20,$F5,$F7,$72,$7F,$0E,$F0,$C3,$FC
	.dc.b	$6D,$5C,$2B,$51,$FE,$27,$FD,$41,$F0,$1F,$E9,$96,$E1,$FE,$8A,$6C
	.dc.b	$FF,$7F,$6B,$FE,$06,$73,$8F,$FB,$DD,$B3,$F9,$4F,$63,$FD,$8D,$18
	.dc.b	$AF,$F5,$FB,$53,$9E,$9F,$D0,$CF,$0C,$3F,$D9,$D9,$7F,$91,$B0,$7C
	.dc.b	$07,$F0,$D7,$FD,$31,$19,$8C,$63,$FE,$43,$8C,$76,$F2,$F4,$7F,$DF
	.dc.b	$AD,$BF,$F7,$DB,$6B,$FD,$F2,$DD,$3F,$DE,$ED,$9F,$F6,$B9,$E2,$71
	.dc.b	$3F,$EF,$56,$C6,$4F,$2D,$78,$1A,$78,$0B,$3E,$06,$D7,$81,$A7,$83
	.dc.b	$0F,$6C,$7D,$D9,$FC,$1D,$75,$7E,$03,$63,$58,$7C,$2B,$E7,$5E,$76
	.dc.b	$9E,$77,$8B,$E5,$69,$E5,$2C,$F9,$5B,$1E,$5C,$3E,$0E,$1E,$68,$FF
	.dc.b	$7C,$5F,$86,$B6,$78,$4B,$1E,$1B,$F3,$E1,$E9,$8E,$2D,$9F,$D1,$3F
	.dc.b	$EE,$FA,$FF,$49,$C7,$F9,$61,$AF,$09,$FF,$71,$C7,$FA,$3A,$3E,$3C
	.dc.b	$7F,$BB,$8F,$C5,$FC,$0F,$DA,$8C,$9E,$4C,$11,$9A,$E4,$7E,$3D,$C5
	.dc.b	$FC,$06,$DF,$CC,$AE,$2F,$D0,$BE,$B8,$7B,$A3,$C5,$3A,$9F,$8A,$6A
	.dc.b	$3D,$4F,$FA,$BD,$F3,$9C,$F0,$F1,$E1,$FB,$FB,$FE,$01,$3A,$AE,$75
	.dc.b	$A8,$EF,$87,$9A,$7D,$CD,$AF,$24,$77,$C3,$C9,$18,$25,$79,$1E,$F1
	.dc.b	$A8,$C1,$27,$EB,$BE,$03,$D1,$87,$C7,$69,$F1,$EF,$C4,$F1,$3D,$57
	.dc.b	$E0,$AE,$67,$A2,$38,$B4,$F7,$D7,$56,$C7,$53,$73,$CD,$EA,$79,$E3
	.dc.b	$E8,$1D,$A3,$DB,$5E,$48,$C4,$FE,$CE,$31,$1C,$23,$9B,$A2,$19,$BF
	.dc.b	$8C,$E3,$0F,$A8,$FC,$FA,$E7,$1F,$8D,$AF,$39,$F0,$EB,$8D,$79,$89
	.dc.b	$3B,$7D,$E7,$C3,$BF,$90,$EB,$0F,$C3,$5F,$A4,$EB,$1E,$77,$D2,$ED
	.dc.b	$F9,$0F,$49,$F9,$17,$E2,$BF,$4C,$63,$0F,$4E,$1D,$AF,$FA,$CD,$EE
	.dc.b	$3E,$56,$1D,$B0,$E2,$FB,$6F,$8B,$F8,$7B,$F0,$39,$3C,$EF,$CB,$8D
	.dc.b	$BF,$26,$33,$3F,$0F,$0F,$82,$F6,$9F,$1D,$70,$9F,$BB,$5C,$2F,$D0
	.dc.b	$F9,$5F,$3E,$1E,$98,$F8,$E6,$4A,$E6,$FC,$E7,$68,$E8,$7A,$EB,$C7
	.dc.b	$5F,$95,$6C,$7C,$E6,$CF,$D9,$8F,$CA,$61,$E9,$7C,$11,$D9,$B3,$C0
	.dc.b	$F5,$C7,$88,$F8,$D1,$E2,$7E,$31,$99,$F8,$D1,$DA,$BC,$8D,$3C,$85
	.dc.b	$9F,$21,$F9,$98,$DB,$63,$6D,$36,$7E,$2C,$B3,$F4,$DF,$29,$E3,$9F
	.dc.b	$A3,$7F,$2E,$FD,$97,$F9,$7A,$D3,$F2,$CF,$24,$7E,$45,$F7,$3F,$08
	.dc.b	$E3,$1F,$9D,$9E,$93,$C0,$E9,$3A,$9F,$1B,$6F,$E2,$C7,$AE,$3B,$3E
	.dc.b	$23,$E5,$BB,$BF,$4E,$1F,$8D,$BE,$A7,$7A,$F1,$46,$5F,$C3,$1C,$EB
	.dc.b	$D1,$3B,$8F,$1B,$6F,$C7,$1F,$2E,$7C,$67,$86,$7E,$24,$79,$DF,$D4
	.dc.b	$3D,$5F,$1B,$FC,$53,$F5,$07,$C4,$7F,$82,$E0,$48,$CC,$E2,$20,$8C
	.dc.b	$B8,$31,$3C,$DE,$4F,$F0,$37,$33,$FE,$A1,$E5,$C5,$72,$BF,$F0,$E3
	.dc.b	$67,$C0,$DA,$6F,$FE,$A6,$BE,$38,$72,$F8,$4F,$05,$79,$5B,$66,$93
	.dc.b	$0F,$7E,$3E,$08,$FE,$31,$8F,$FA,$CB,$E6,$FA,$56,$8F,$1C,$E3,$0F
	.dc.b	$F5,$7F,$84,$AC,$61,$FE,$35,$B7,$9D,$E4,$69,$F1,$EF,$FC,$FB,$19
	.dc.b	$24,$77,$8F,$F2,$7C,$7E,$93,$E1,$7E,$54,$F2,$95,$77,$AC,$9B,$9E
	.dc.b	$53,$D9,$BF,$02,$EE,$F8,$ED,$8F,$23,$67,$EE,$35,$FC,$0B,$9B,$FF
	.dc.b	$29,$36,$7F,$DA,$7F,$19,$F7,$E3,$15,$F8,$C6,$A7,$3A,$FF,$B8,$27
	.dc.b	$86,$1F,$87,$F7,$9F,$21,$F7,$1B,$8F,$C9,$6E,$3F,$91,$6B,$F9,$67
	.dc.b	$39,$7F,$99,$DB,$3F,$9E,$C5,$EF,$A7,$C1,$84,$AF,$93,$1C,$30,$FC
	.dc.b	$8D,$6E,$73,$5F,$9D,$F7,$9F,$39,$F9,$26,$CF,$DF,$6D,$FE,$9D,$AF
	.dc.b	$EB,$1C,$E5,$FE,$7D,$6C,$F9,$F0,$7C,$07,$7C,$24,$FD,$EA,$DE,$1E
	.dc.b	$6A,$CB,$F8,$69,$FD,$D7,$B5,$FE,$63,$1D,$5F,$DD,$FD,$C7,$C0,$3E
	.dc.b	$03,$F7,$15,$BA,$A6,$E7,$57,$F8,$E9,$D3,$67,$55,$C2,$35,$EF,$F4
	.dc.b	$3C,$1F,$00,$F8,$0C,$4E,$89,$9F,$A4,$66,$AE,$13,$AE,$0F,$F5,$26
	.dc.b	$7F,$EE,$98,$F7,$F8,$FF,$C4,$70,$FE,$3B,$F0,$1F,$C7,$BD,$E7,$AD
	.dc.b	$AB,$F5,$5F,$C0,$38,$5A,$F8,$56,$E3,$C8,$7F,$25,$8C,$BF,$CA,$B0
	.dc.b	$57,$FB,$CF,$0E,$F3,$FC,$AA,$B6,$AB,$78,$62,$3C,$E2,$B8,$BB,$BE
	.dc.b	$98,$7F,$31,$C3,$9E,$1C,$AB,$84,$E5,$A7,$96,$B8,$BE,$33,$27,$5C
	.dc.b	$3C,$4F,$C1,$33,$5D,$70,$F1,$C7,$99,$F2,$C6,$6F,$C1,$18,$BE,$F5
	.dc.b	$F1,$CE,$6E,$1E,$53,$F3,$67,$AC,$78,$5E,$57,$E4,$C3,$DF,$7C,$6B
	.dc.b	$57,$C4,$F8,$0F,$F2,$F9,$FE,$60,$FC,$C3,$07,$F5,$9F,$7F,$F9,$BF
	.dc.b	$C0,$7F,$35,$6C,$67,$B5,$39,$6C,$71,$F8,$0F,$E6,$6D,$D7,$94,$62
	.dc.b	$3F,$B4,$DA,$31,$1F,$DB,$63,$13,$F8,$4F,$80,$7C,$07,$F7,$1D,$BF
	.dc.b	$03,$18,$BF,$C4,$DF,$07,$EC,$B6,$77,$5C,$74,$FE,$F5,$7C,$6F,$78
	.dc.b	$FA,$2F,$F1,$B8,$71,$BD,$E1,$C6,$FF,$C0,$B8,$71,$BD,$E1,$C6,$FD
	.dc.b	$4D,$AD,$DF,$15,$7C,$6B,$78,$FA,$AF,$85,$6A,$F8,$5E,$E3,$FC,$2B
	.dc.b	$8F,$95,$B5,$C2,$75,$1C,$4E,$97,$1F,$CE,$C6,$8D,$D7,$38,$E0,$7E
	.dc.b	$F2,$3E,$34,$B6,$F8,$73,$FD,$B2,$B8,$56,$A7,$E6,$FA,$FF,$E2,$7B
	.dc.b	$E1,$7F,$22,$31,$3C,$A3,$F8,$4F,$BC,$F8,$0C,$C7,$F8,$CB,$DF,$7C
	.dc.b	$06,$21,$18,$7F,$BE,$D7,$16,$D6,$E7,$51,$FC,$59,$B8,$F9,$21,$5C
	.dc.b	$5F,$1F,$BB,$C7,$D1,$91,$DA,$3C,$52,$73,$1D,$67,$31,$23,$FC,$23
	.dc.b	$69,$DD,$F8,$6F,$95,$7F,$85,$BD,$6D,$36,$BE,$04,$62,$3F,$A1,$78
	.dc.b	$AF,$80,$7C,$2E,$AF,$F7,$B7,$BB,$FF,$2B,$B5,$FF,$2D,$35,$FF,$2F
	.dc.b	$BD,$71,$FA,$F8,$75,$DB,$E9,$B6,$BC,$87,$9F,$93,$CA,$47,$FC,$71
	.dc.b	$1E,$5F,$96,$C4,$7F,$99,$70,$FE,$36,$F8,$2B,$CE,$47,$8F,$9C,$F8
	.dc.b	$2F,$0B,$63,$C0,$7F,$8F,$38,$FE,$4B,$FD,$49,$B9,$FD,$D6,$E7,$FE
	.dc.b	$79,$F8,$17,$C0,$7E,$2E,$78,$DF,$F9,$F1,$AF,$F9,$FD,$AF,$FA,$11
	.dc.b	$A9,$EA,$FF,$A2,$25,$F0,$5F,$97,$6C,$79,$F9,$3F,$35,$F1,$6C,$FF
	.dc.b	$99,$B6,$7F,$CC,$8D,$7F,$D1,$8D,$7F,$D2,$3F,$04,$F8,$3F,$5C,$7F
	.dc.b	$A4,$63,$19,$FF,$9B,$5F,$F2,$F7,$8B,$6B,$75,$FE,$9A,$F8,$37,$B5
	.dc.b	$FE,$88,$6E,$FF,$E7,$56,$CF,$FA,$75,$AF,$E3,$3C,$5F,$F4,$3F,$4F
	.dc.b	$8E,$BF,$35,$18,$C3,$FD,$41,$CB,$FD,$CE,$F8,$57,$FA,$06,$31,$1F
	.dc.b	$94,$AF,$F4,$CC,$7F,$86,$63,$F2,$B1,$FC,$2B,$8F,$F8,$5F,$B6,$1F
	.dc.b	$03,$DA,$7F,$D5,$DF,$02,$F8,$0F,$2F,$AF,$86,$E7,$FE,$84,$9D,$3F
	.dc.b	$E9,$36,$E5,$FE,$BB,$6B,$FE,$C0,$6B,$FE,$C3,$77,$97,$F0,$B6,$CF
	.dc.b	$F9,$B7,$91,$F0,$3E,$18,$FF,$49,$E9,$FE,$50,$7F,$D3,$C3,$E0,$7F
	.dc.b	$83,$3A,$F4,$BE,$03,$FE,$D4,$35,$E0,$7F,$B2,$EB,$96,$7F,$E4,$59
	.dc.b	$C9,$C4,$EB,$7D,$23,$FC,$6F,$3C,$33,$FE,$6B,$5C,$27,$38,$78,$1F
	.dc.b	$E8,$17,$C6,$BF,$94,$5F,$1B,$F2,$4E,$5F,$F7,$34,$66,$B8,$E1,$E7
	.dc.b	$C3,$A5,$77,$C3,$9C,$63,$0E,$57,$FE,$F2,$C3,$95,$EC,$EB,$5B,$C3
	.dc.b	$A4,$7B,$9B,$5F,$0E,$F9,$BF,$87,$AE,$15,$F1,$63,$55,$C3,$0F,$87
	.dc.b	$7C,$30,$F0,$B7,$0E,$37,$F1,$B0,$F0,$1C,$EF,$8E,$1C,$E3,$E2,$B6
	.dc.b	$7E,$1C,$6C,$C9,$5C,$9D,$4F,$2B,$93,$E1,$3F,$16,$BC,$75,$F2,$A7
	.dc.b	$C7,$3C,$63,$C7,$1C,$6F,$D4,$EA,$38,$4F,$63,$84,$79,$1A,$79,$23
	.dc.b	$67,$47,$31,$E4,$7E,$73,$D4,$FA,$E5,$7D,$0D,$76,$78,$63,$AC,$78
	.dc.b	$5B,$1E,$18,$E6,$ED,$B1,$E3,$36,$F8,$F0,$FB,$38,$7D,$18,$C6,$1F
	.dc.b	$21,$E2,$77,$2D,$F7,$3C,$E7,$72,$DF,$78,$D9,$D0,$F6,$5F,$81,$FA
	.dc.b	$17,$E1,$C3,$E8,$DF,$33,$E0,$DF,$37,$AD,$78,$DA,$78,$CD,$B9,$69
	.dc.b	$9B,$F0,$3F,$52,$BB,$D7,$58,$DD,$79,$DF,$15,$70,$9F,$93,$5C,$2B
	.dc.b	$B4,$E5,$F7,$C6,$6B,$C1,$87,$CB,$7C,$F3,$D6,$FC,$17,$C6,$73,$7C
	.dc.b	$6B,$47,$6B,$E5,$3E,$69,$D1,$30,$F9,$44,$AF,$9A,$49,$F8,$11,$96
	.dc.b	$C6,$70,$F6,$CE,$5B,$19,$3D,$B1,$96,$99,$9F,$4C,$65,$FC,$28,$3F
	.dc.b	$2A,$53,$26,$08,$60,$87,$E8,$8E,$07,$DC,$3E,$17,$C0,$FF,$09,$71
	.dc.b	$8F,$F1,$48,$C3,$FC,$BE,$37,$5F,$C2,$DD,$E3,$FC,$B6,$B6,$FE,$4D
	.dc.b	$B3,$FC,$15,$B3,$F8,$96,$E1,$F8,$46,$CF,$E0,$9B,$3F,$71,$AF,$88
	.dc.b	$FE,$12,$78,$A3,$FA,$9C,$6A,$FD,$51,$E7,$7D,$46,$27,$8B,$8B,$F5
	.dc.b	$4E,$DF,$53,$4F,$51,$67,$D4,$DB,$F5,$34,$F5,$16,$3D,$4D,$7D,$51
	.dc.b	$FC,$49,$F5,$1F,$D3,$6F,$D5,$5E,$48,$F3,$BE,$46,$D7,$91,$A7,$90
	.dc.b	$B5,$E4,$6C,$79,$1A,$79,$0A,$79,$1F,$E2,$EF,$92,$3F,$AC,$D7,$92
	.dc.b	$7E,$79,$EB,$BF,$99,$3E,$B7,$E6,$46,$AB,$AC,$6A,$3A,$BA,$8F,$81
	.dc.b	$1C,$EF,$E0,$57,$37,$E0,$36,$3E,$03,$67,$E0,$34,$F8,$05,$8F,$80
	.dc.b	$D8,$F8,$0D,$7E,$01,$FC,$90,$F8,$11,$FE,$12,$BF,$81,$1E,$63,$D7
	.dc.b	$5E,$69,$F5,$BE,$66,$9E,$62,$C7,$99,$B5,$E6,$7E,$7F,$C0,$3E,$07
	.dc.b	$FC,$C3,$F5,$5F,$E2,$7B,$4E,$6B,$FD,$47,$CC,$AD,$4E,$6B,$F7,$E7
	.dc.b	$0B,$FF,$6A,$3F,$C2,$63,$4F,$F0,$46,$B9,$8E,$67,$F8,$DF,$C4,$FF
	.dc.b	$46,$74,$E3,$1F,$E2,$99,$7F,$B7,$F0,$7C,$17,$F9,$E3,$E3,$BC,$6D
	.dc.b	$AD,$9E,$17,$C7,$97,$F1,$DA,$E3,$3E,$88,$F7,$CE,$9B,$19,$9D,$C2
	.dc.b	$FC,$55,$E7,$E3,$7C,$1F,$8B,$0F,$0D,$FF,$97,$FD,$2F,$F2,$3F,$0F
	.dc.b	$F9,$B2,$31,$3F,$E6,$88,$C4,$21,$3E,$36,$9E,$72,$C7,$D0,$6B,$FC
	.dc.b	$C1,$CE,$3C,$E3,$B9,$FE,$4B,$F8,$3F,$BD,$3A,$3F,$9F,$7C,$1B,$AB
	.dc.b	$D9,$D9,$F8,$38,$C4,$FF,$B7,$23,$13,$F1,$E7,$37,$F8,$26,$BF,$D0
	.dc.b	$9A,$F6,$8E,$35,$FE,$E6,$F8,$3B,$3F,$36,$3F,$A8,$7C,$1B,$E0,$FF
	.dc.b	$09,$F0,$7F,$1A,$31,$3F,$EF,$68,$C4,$FD,$99,$CD,$7E,$51,$AF,$F5
	.dc.b	$46,$BD,$A3,$F4,$11,$C3,$4F,$91,$CA,$B4,$EA,$BF,$4D,$F0,$4F,$83
	.dc.b	$FF,$40,$78,$D8,$6E,$7F,$82,$CD,$5B,$8F,$C8,$36,$3F,$52,$D7,$FB
	.dc.b	$4B,$5F,$EE,$4E,$F3,$FB,$91,$D2,$BE,$2F,$C1,$F8,$6F,$C1,$EF,$7F
	.dc.b	$6B,$3F,$D3,$58,$BE,$0D,$F0,$7F,$BF,$9D,$78,$14,$FF,$54,$56,$E7
	.dc.b	$17,$D2,$6E,$5F,$5C,$56,$E7,$47,$8B,$E0,$FB,$1D,$3E,$09,$F0,$78
	.dc.b	$9D,$12,$FA,$5A,$B7,$39,$DB,$E0,$E7,$FE,$ED,$B3,$5F,$F0,$D7,$C1
	.dc.b	$FF,$86,$FE,$09,$EC,$7F,$86,$EB,$81,$3C,$DF,$E4,$17,$C0,$57,$07
	.dc.b	$FC,$49,$3D,$55,$F5,$BE,$97,$FA,$FB,$E1,$5F,$E2,$9A,$D3,$CA,$75
	.dc.b	$5F,$CA,$27,$33,$EA,$9C,$BF,$E2,$A8,$CD,$A7,$33,$D4,$F3,$D7,$8B
	.dc.b	$1E,$71,$EB,$C3,$C7,$7C,$67,$33,$EA,$38,$9E,$57,$38,$78,$30,$EF
	.dc.b	$3C,$F0,$EC,$F0,$3C,$58,$76,$9D,$C7,$58,$C5,$F4,$8E,$18,$78,$DE
	.dc.b	$47,$53,$C4,$60,$E5,$7E,$0A,$F9,$B1,$F5,$CF,$E7,$58,$7F,$3B,$70
	.dc.b	$7C,$D3,$FA,$AF,$C0,$7F,$2C,$F8,$3F,$E8,$0D,$8C,$F7,$A7,$2D,$8E
	.dc.b	$3F,$05,$FD,$05,$BA,$F2,$8C,$47,$F9,$82,$D1,$88,$FF,$35,$46,$27
	.dc.b	$EF,$FC,$13,$E0,$FF,$CD,$D9,$FE,$22,$31,$7F,$84,$BE,$0F,$CF,$6C
	.dc.b	$EE,$B8,$E9,$FE,$77,$BE,$37,$BC,$7D,$17,$F9,$4C,$38,$DE,$F0,$E3
	.dc.b	$7F,$E8,$1C,$38,$DE,$F0,$E3,$7E,$A6,$D6,$EF,$8A,$BE,$35,$BC,$7D
	.dc.b	$57,$C2,$B5,$7C,$2F,$71,$FE,$91,$C7,$CA,$DA,$E1,$3A,$8E,$27,$4B
	.dc.b	$8F,$E4,$E3,$46,$EB,$9C,$70,$3F,$8C,$16,$7F,$95,$CF,$C8,$7F,$96
	.dc.b	$CF,$08,$C6,$DF,$37,$DC,$FF,$52,$57,$06,$C7,$CB,$BF,$E8,$1F,$00
	.dc.b	$F8,$2C,$C7,$FA,$B7,$E0,$9F,$07,$88,$46,$1E,$33,$9A,$E5,$1F,$D2
	.dc.b	$E7,$4F,$A9,$B8,$FF,$44,$9D,$CF,$B9,$7B,$DD,$E9,$EA,$53,$B4,$4A
	.dc.b	$ED,$19,$7D,$71,$97,$C1,$19,$7B,$47,$AF,$D9,$EB,$1F,$D8,$5A,$FD
	.dc.b	$06,$E1,$FD,$BB,$D4,$7C,$1F,$1C,$25,$F1,$7E,$4F,$8E,$AD,$CE,$4F
	.dc.b	$F1,$D6,$1D,$9F,$F6,$8B,$5F,$F6,$D3,$5F,$F6,$F3,$FC,$DB,$0F,$ED
	.dc.b	$6D,$DB,$7C,$9E,$0B,$7C,$1F,$B6,$26,$1E,$D7,$51,$88,$FF,$31,$DF
	.dc.b	$F1,$67,$99,$E6,$9F,$E2,$DC,$3F,$52,$3F,$C0,$CD,$7F,$2A,$DC,$7F
	.dc.b	$BF,$FC,$0B,$E0,$3F,$11,$39,$CF,$FD,$EC,$DB,$FF,$7D,$35,$57,$17
	.dc.b	$BE,$FE,$22,$3F,$E6,$39,$F1,$3F,$EF,$AF,$33,$FB,$BB,$FE,$67,$6B
	.dc.b	$FE,$80,$6B,$BF,$07,$F9,$97,$C0,$BD,$3A,$49,$FF,$38,$46,$5F,$F3
	.dc.b	$97,$C0,$BE,$03,$CD,$02,$08,$D6,$BC,$31,$FF,$08,$CF,$FA,$4A,$72
	.dc.b	$FD,$2F,$0F,$FD,$19,$C7,$D7,$7F,$CA,$46,$31,$FF,$50,$71,$7F,$53
	.dc.b	$E1,$FC,$A4,$62,$75,$5C,$0F,$C7,$BF,$E8,$7A,$FE,$B5,$87,$E4,$23
	.dc.b	$FD,$15,$EF,$FF,$9F,$F9,$7F,$31,$A7,$FA,$B6,$31,$9F,$F9,$A3,$3F
	.dc.b	$F1,$DC,$62,$75,$F0,$1E,$86,$E7,$FE,$97,$94,$F6,$24,$7F,$A6,$FD
	.dc.b	$E7,$95,$FE,$81,$C3,$FD,$05,$2F,$7F,$F8,$0B,$FE,$55,$AD,$BF,$E9
	.dc.b	$83,$5D,$5F,$EC,$38,$CE,$DF,$E3,$F9,$C9,$C7,$0E,$CF,$D7,$33,$1F
	.dc.b	$CB,$A7,$8E,$1E,$2C,$3C,$33,$9B,$E2,$FF,$A8,$63,$33,$FE,$5E,$AE
	.dc.b	$37,$D6,$72,$FE,$12,$33,$5C,$70,$F4,$61,$D2,$BC,$78,$73,$8C,$61
	.dc.b	$CA,$FF,$1D,$87,$2B,$D9,$D6,$B7,$87,$48,$F8,$D7,$C6,$D7,$C6,$30
	.dc.b	$DA,$E1,$3F,$26,$73,$7D,$2F,$E5,$1D,$2B,$78,$7C,$76,$C7,$C1,$6D
	.dc.b	$FB,$E3,$8B,$D6,$33,$1E,$A8,$F6,$D7,$77,$E8,$47,$73,$8D,$77,$69
	.dc.b	$DC,$B1,$DC,$E2,$77,$72,$77,$7E,$1C,$F7,$9F,$A1,$3D,$E7,$8C,$77
	.dc.b	$6C,$77,$69,$DE,$38,$BD,$CC,$C7,$7C,$3E,$73,$E2,$C3,$E6,$BD,$EB
	.dc.b	$89,$DC,$AF,$76,$D7,$78,$FC,$01,$C4,$AF,$86,$37,$1E,$1B,$F7,$3A
	.dc.b	$8E,$CD,$3B,$16,$7B,$1F,$51,$EC,$D3,$B3,$B3,$B1,$A7,$E6,$B6,$FE
	.dc.b	$0C,$7C,$4B,$F8,$2F,$23,$A5,$79,$1B,$5E,$46,$9E,$43,$A9,$C4,$B1
	.dc.b	$C5,$B1,$C5,$A7,$1A,$F1,$CF,$C6,$AF,$1B,$C6,$33,$5E,$D9,$CB,$CE
	.dc.b	$33,$5E,$48,$F8,$2F,$8E,$32,$F8,$23,$35,$D6,$FE,$9C,$F1,$BE,$AF
	.dc.b	$96,$32,$F1,$8C,$D7,$23,$C5,$3A,$BE,$E4,$70,$48,$FA,$4D,$AF,$33
	.dc.b	$6F,$C9,$3A,$24,$E9,$B1,$A8,$C1,$1C,$12,$3C,$A7,$02,$C6,$DB,$1B
	.dc.b	$8F,$81,$1B,$69,$B2,$C6,$E3,$E1,$46,$1E,$64,$70,$47,$64,$BF,$96
	.dc.b	$4C,$3F,$89,$B6,$3F,$6D,$18,$23,$82,$4F,$E6,$E3,$2D,$8C,$D7,$E3
	.dc.b	$67,$2D,$8C,$9F,$8D,$8C,$B4,$CC,$FE,$16,$32,$D8,$CB,$F8,$58,$CB
	.dc.b	$63,$38,$7E,$0D,$F9,$87,$C1,$9D,$C6,$67,$E7,$4E,$4E,$38,$7E,$15
	.dc.b	$F9,$86,$63,$DC,$4B,$F6,$36,$3D,$4D,$3D,$05,$3C,$C5,$AF,$14,$FD
	.dc.b	$DC,$3F,$82,$CF,$F8,$B0,$FD,$F9,$67,$FB,$FB,$6D,$7F,$E0,$D7,$ED
	.dc.b	$FB,$1F,$D4,$A7,$F8,$09,$FA,$B7,$15,$FD,$D2,$31,$5F,$D2,$9D,$C7
	.dc.b	$F6,$79,$DD,$74,$AD,$4F,$13,$4E,$23,$55,$FD,$18,$E6,$FF,$7C,$AE
	.dc.b	$73,$FD,$0D,$B3,$FC,$95,$B8,$7F,$15,$6C,$FF,$13,$6C,$FF,$09,$6B
	.dc.b	$EA,$3C,$17,$E5,$BF,$05,$7B,$7D,$B7,$BD,$FE,$50,$F7,$FF,$C3,$B8
	.dc.b	$27,$35,$FE,$5D,$CB,$FD,$45,$5A,$9C,$D2,$73,$5B,$3F,$CC,$73,$A8
	.dc.b	$CC,$FF,$B7,$23,$8B,$F7,$5A,$F0,$BF,$B6,$FE,$04,$F2,$38,$78,$7B
	.dc.b	$BF,$E6,$8F,$7B,$C8,$DC,$3C,$53,$FE,$75,$B4,$E6,$3C,$4D,$C3,$A4
	.dc.b	$EA,$BC,$15,$F0,$EB,$C0,$FB,$1A,$F5,$6B,$F9,$2F,$2F,$FC,$79,$AF
	.dc.b	$F9,$8A,$BC,$73,$97,$FC,$B5,$19,$8F,$F7,$FC,$27,$8F,$07,$D4,$8C
	.dc.b	$1C,$1C,$47,$82,$33,$3F,$1D,$AF,$EA,$5A,$FF,$A5,$BD,$BF,$F3,$E4
	.dc.b	$7F,$A6,$79,$7F,$D4,$5E,$E7,$E0,$A7,$34,$9C,$D7,$AA,$73,$18,$9F
	.dc.b	$C1,$4F,$5E,$5F,$F3,$3C,$E7,$5F,$B1,$69,$F8,$ED,$AF,$9E,$DB,$FC
	.dc.b	$03,$5F,$E0,$5E,$D3,$DA,$EB,$ED,$FC,$9C,$7F,$09,$39,$C3,$F0,$39
	.dc.b	$2B,$51,$FE,$B6,$9D,$61,$E8,$3D,$D3,$A7,$FD,$79,$1A,$BF,$C9,$DF
	.dc.b	$3D,$7F,$4F,$3C,$AB,$CB,$8F,$E0,$A3,$11,$DE,$3F,$0B,$ED,$BD,$BF
	.dc.b	$F3,$EF,$AF,$FB,$08,$C6,$9F,$E7,$88,$C4,$EA,$3F,$13,$3A,$BF,$F3
	.dc.b	$B5,$6A,$7F,$2F,$39,$9F,$E5,$0D,$7F,$DA,$AD,$7F,$DB,$1E,$B7,$F2
	.dc.b	$13,$FD,$2B,$4F,$69,$ED,$7F,$04,$9D,$75,$53,$FD,$47,$5B,$9C,$63
	.dc.b	$F9,$EA,$DB,$FB,$29,$D3,$9A,$E9,$ED,$48,$D7,$B4,$F2,$71,$9F,$D5
	.dc.b	$18,$AF,$3C,$5F,$C1,$B2,$61,$C7,$1F,$15,$7A,$7C,$39,$1F,$B2,$C7
	.dc.b	$FD,$F1,$ED,$7F,$BE,$7D,$97,$9F,$FE,$FA,$D3,$FD,$DD,$77,$0E,$38
	.dc.b	$78,$CE,$22,$F8,$BF,$32,$3F,$95,$BA,$8C,$BE,$A8,$FE,$3B,$18,$8D
	.dc.b	$5F,$7A,$F2,$3F,$C8,$70,$B3,$D6,$6A,$38,$99,$3C,$93,$C5,$A6,$6B
	.dc.b	$C0,$AB,$9E,$0C,$3C,$B7,$C6,$FD,$33,$9C,$3C,$38,$78,$0F,$63,$87
	.dc.b	$A1,$E0,$32,$53,$84,$73,$70,$70,$9E,$F7,$E2,$8F,$F0,$16,$1E,$68
	.dc.b	$C4,$7C,$13,$86,$1F,$25,$E7,$3B,$C3,$C8,$FF,$3C,$8F,$B2,$7F,$3E
	.dc.b	$70,$FA,$0F,$F0,$87,$AF,$FC,$CF,$DA,$FE,$86,$D8,$FE,$A8,$D7,$7D
	.dc.b	$A9,$DF,$B3,$FD,$96,$76,$D7,$AB,$5F,$55,$23,$11,$F7,$63,$1C,$1F
	.dc.b	$80,$F6,$5E,$D7,$C1,$DB,$F1,$B1,$8B,$FC,$2D,$F0,$7E,$93,$67,$75
	.dc.b	$C7,$5F,$95,$7C,$6F,$78,$FA,$6F,$F3,$38,$71,$BD,$E1,$C6,$FF,$25
	.dc.b	$87,$1B,$DE,$1C,$6F,$DA,$DA,$DD,$F1,$57,$C6,$B7,$8F,$AA,$F8,$56
	.dc.b	$AF,$85,$EE,$3E,$34,$EB,$1F,$34,$76,$3E,$45,$77,$8C,$C1,$B5,$CE
	.dc.b	$3B,$1E,$1B,$E2,$FF,$7A,$8F,$8D,$1F,$E1,$36,$EB,$E8,$9F,$D0,$46
	.dc.b	$1F,$8B,$E9,$7E,$5E,$B8,$36,$BE,$5C,$FF,$81,$BD,$87,$B3,$98,$FD
	.dc.b	$F7,$B2,$F6,$B1,$08,$C3,$C6,$73,$5C,$A3,$FC,$23,$3A,$7F,$96,$37
	.dc.b	$1F,$F1,$64,$EE,$7F,$98,$2F,$5F,$BD,$3C,$FA,$76,$89,$5D,$A3,$2F
	.dc.b	$AE,$32,$F8,$23,$2F,$68,$F5,$FD,$67,$F3,$D8,$F3,$FC,$6F,$80,$CF
	.dc.b	$4C,$8C,$90,$FE,$8D,$3C,$5B,$1C,$5D,$65,$FE,$49,$9D,$3F,$E4,$BF
	.dc.b	$6D,$F3,$3E,$EF,$42,$5F,$F8,$E2,$72,$D9,$CB,$F2,$A7,$E9,$CE,$A3
	.dc.b	$FA,$C3,$6B,$FC,$B1,$A7,$F9,$C2,$7E,$9D,$70,$BF,$B7,$FA,$A3,$8B
	.dc.b	$F8,$B3,$6B,$FC,$A6,$D7,$FC,$C8,$D7,$FD,$4C,$E7,$0F,$F0,$BF,$AC
	.dc.b	$F8,$FF,$C5,$79,$68,$C5,$FF,$95,$2F,$87,$2F,$F2,$B6,$C7,$F9,$06
	.dc.b	$76,$FF,$89,$EB,$95,$FA,$31,$F4,$FC,$97,$A0,$95,$E8,$6B,$C6,$DF
	.dc.b	$3F,$E4,$6C,$78,$E7,$9B,$FE,$3A,$8C,$BC,$67,$2F,$F8,$17,$DC,$7B
	.dc.b	$8F,$77,$FC,$F3,$18,$6F,$A6,$F2,$33,$5B,$BF,$F4,$77,$C4,$3E,$1F
	.dc.b	$FC,$F6,$DE,$1F,$F4,$1B,$67,$73,$FE,$80,$F8,$B7,$C6,$52,$5F,$FA
	.dc.b	$2A,$32,$FF,$A3,$31,$7B,$1F,$E8,$86,$C6,$63,$07,$FA,$33,$EE,$7F
	.dc.b	$D0,$D8,$FF,$B8,$DA,$FF,$61,$6B,$FE,$03,$6A,$FD,$51,$F5,$3F,$E9
	.dc.b	$66,$DF,$FA,$79,$B3,$FE,$AF,$6B,$DF,$D1,$FF,$62,$FE,$95,$F3,$6B
	.dc.b	$FD,$37,$87,$F8,$BE,$BF,$C6,$D3,$97,$FC,$71,$19,$79,$4F,$FA,$B2
	.dc.b	$31,$1E,$9A,$F7,$1F,$ED,$28,$E9,$D5,$FE,$BF,$7F,$49,$5E,$09,$D1
	.dc.b	$1F,$1F,$3D,$BF,$0C,$DB,$E5,$36,$BC,$07,$33,$89,$6B,$8B,$67,$8B
	.dc.b	$4E,$31,$A3,$DF,$3D,$5F,$A4,$D5,$7E,$A7,$85,$74,$6D,$7D,$BA,$E1
	.dc.b	$5B,$AF,$85,$7F,$16,$72,$4A,$F2,$5F,$81,$AF,$82,$34,$F6,$34,$53
	.dc.b	$51,$C2,$FC,$0F,$BA,$FC,$D3,$F9,$1A,$FC,$A1,$27,$EE,$3E,$57,$FD
	.dc.b	$BF,$F5,$BF,$8C,$AF,$71,$23,$DC,$6B,$3E,$17,$F2,$A3,$E9,$BF,$E8
	.dc.b	$1A,$E5,$1C,$23,$44,$A4,$7F,$90,$E7,$73,$EE,$9F,$03,$67,$55,$E0
	.dc.b	$9C,$CF,$C2,$3C,$11,$99,$F0,$4E,$AB,$AD,$6E,$38,$E1,$F2,$23,$FC
	.dc.b	$E9,$3A,$8C,$46,$EB,$99,$E2,$3D,$F5,$CB,$0F,$5C,$F8,$A7,$27,$07
	.dc.b	$D4,$76,$C3,$CC,$E0,$EC,$66,$38,$B9,$BF,$1E,$1E,$5B,$E3,$3C,$04
	.dc.b	$66,$B8,$C6,$63,$0F,$D1,$8F,$23,$C2,$F9,$DF,$C0,$3B,$B8,$BE,$F7
	.dc.b	$CB,$0F,$81,$87,$E0,$CE,$F7,$C6,$BE,$1C,$FC,$4C,$3B,$56,$A7,$9E
	.dc.b	$1D,$63,$CF,$87,$57,$CB,$87,$89,$E0,$6A,$78,$5F,$AC,$F9,$05,$3E
	.dc.b	$24,$F1,$70,$F9,$67,$D1,$18,$BF,$1C,$76,$8E,$57,$E3,$7E,$94,$FB
	.dc.b	$27,$F0,$51,$C5,$F2,$1C,$70,$FC,$86,$1E,$66,$C6,$4F,$A3,$7D,$6B
	.dc.b	$CF,$7D,$5F,$5D,$7C,$E7,$C7,$7C,$EF,$8B,$63,$F1,$57,$CC,$91,$EC
	.dc.b	$8E,$33,$E6,$69,$2B,$CA,$D8,$EA,$6E,$32,$47,$31,$B3,$C3,$3D,$E7
	.dc.b	$51,$89,$E4,$70,$8C,$C7,$E5,$67,$84,$E4,$91,$C4,$EC,$6D,$E6,$E6
	.dc.b	$79,$E1,$F7,$DB,$87,$CB,$9F,$C8,$9C,$A7,$55,$DA,$3A,$19,$9E,$18
	.dc.b	$7C,$A9,$EF,$3A,$BE,$B1,$F8,$A6,$C7,$F0,$13,$94,$64,$ED,$7E,$08
	.dc.b	$F4,$46,$2F,$BE,$1F,$0A,$72,$43,$C1,$3B,$3A,$39,$AE,$8E,$8C,$57
	.dc.b	$49,$F4,$CE,$89,$19,$9F,$39,$E9,$BF,$C1,$BF,$EF,$73,$E4,$97,$0F
	.dc.b	$F0,$7B,$FA,$23,$18,$73,$9F,$F2,$DF,$CF,$DC,$7F,$4F,$1F,$C0,$EF
	.dc.b	$F9,$FF,$5F,$F9,$76,$B5,$87,$E8,$2F,$61,$FF,$2E,$46,$9F,$B7,$7C
	.dc.b	$F0,$FD,$45,$F3,$BF,$8B,$5A,$79,$BC,$9C,$E1,$FD,$10,$F1,$F4,$B1
	.dc.b	$ED,$19,$C3,$AC,$E6,$B8,$C7,$A3,$0F,$06,$1F,$26,$7B,$BE,$13,$C0
	.dc.b	$64,$F7,$12,$38,$E3,$EB,$9E,$A7,$F9,$AF,$1F,$1D,$F7,$C3,$C5,$87
	.dc.b	$33,$B1,$DF,$0E,$77,$F0,$1B,$1D,$E7,$8E,$1D,$E3,$FC,$FD,$87,$53
	.dc.b	$E7,$7D,$13,$E5,$3F,$9D,$4F,$D3,$11,$9C,$7F,$0D,$1F,$5E,$30,$E1
	.dc.b	$AF,$E2,$DC,$DF,$F4,$56,$CF,$57,$FC,$96,$DA,$FD,$D4,$7F,$94,$AB
	.dc.b	$B6,$DE,$46,$D7,$89,$F2,$C6,$57,$A6,$FA,$FC,$36,$B8,$36,$7D,$6D
	.dc.b	$7F,$44,$E7,$1F,$EB,$8D,$9F,$CB,$68,$F9,$77,$CB,$7F,$11,$AD,$61
	.dc.b	$F9,$B6,$BE,$4C,$7F,$33,$5C,$30,$FD,$06,$3D,$A7,$F8,$D7,$C3,$3E
	.dc.b	$1B,$D8,$DA,$F8,$0D,$C3,$E9,$B5,$FD,$D3,$9B,$FE,$E8,$D9,$F4,$FC
	.dc.b	$73,$E5,$3B,$E1,$27,$C6,$7E,$06,$B8,$60,$D5,$87,$1A,$DC,$FE,$E6
	.dc.b	$B6,$BE,$1D,$F2,$3F,$55,$B5,$F6,$5B,$3F,$8C,$6B,$FC,$4D,$CE,$7F
	.dc.b	$E0,$D6,$CF,$AB,$D4,$FE,$81,$18,$AF,$E1,$4D,$7C,$57,$FC,$23,$15
	.dc.b	$70,$8F,$E9,$1F,$40,$F9,$4A,$7F,$75,$C5,$F1,$EF,$A1,$CC,$7C,$0C
	.dc.b	$97,$C6,$30,$D8,$69,$FE,$48,$9E,$75,$CA,$73,$1F,$AF,$AE,$15,$FE
	.dc.b	$2B,$6C,$7F,$25,$38,$9D,$B0,$FE,$C1,$E4,$3D,$8A,$7A,$72,$FF,$76
	.dc.b	$57,$9E,$74,$D9,$DB,$F6,$8F,$7F,$87,$8C,$BC,$A7,$F3,$5A,$E2,$E0
	.dc.b	$FF,$7E,$57,$79,$E1,$97,$F9,$0F,$E0,$9F,$09,$FD,$D5,$B8,$FE,$22
	.dc.b	$9F,$0C,$F8,$6D,$5C,$AF,$9D,$7C,$32,$F1,$E1,$C2,$F5,$1E,$39,$D6
	.dc.b	$3E,$1C,$3F,$C5,$D7,$BC,$39,$C7,$8F,$D8,$91,$E3,$F8,$6F,$E2,$78
	.dc.b	$FF,$53,$C5,$5A,$8E,$CF,$49,$DC,$27,$75,$E8,$F8,$6F,$80,$7F,$85
	.dc.b	$67,$F8,$20,$F8,$47,$C2,$7F,$2F,$B5,$70,$BE,$47,$E3,$EF,$8D,$F0
	.dc.b	$7C,$D3,$D5,$FE,$42,$75,$BE,$91,$FE,$98,$9E,$13,$FC,$F7,$BF,$FB
	.dc.b	$84,$63,$0F,$0E,$18,$C3,$F9,$2D,$F3,$3B,$C6,$5F,$E4,$E6,$4F,$85
	.dc.b	$87,$79,$E3,$7D,$63,$13,$C2,$31,$5C,$E3,$C8,$73,$BF,$1B,$D4,$F8
	.dc.b	$37,$EC,$9E,$11,$83,$CA,$6C,$F8,$75,$D9,$C6,$1D,$EF,$AC,$FC,$49
	.dc.b	$E4,$7A,$AB,$AB,$CA,$FC,$0F,$23,$CD,$87,$B1,$F4,$9F,$D3,$8F,$EA
	.dc.b	$C7,$D4,$30,$7F,$8B,$FE,$09,$F0,$CF,$86,$FE,$B9,$18,$6C,$66,$7F
	.dc.b	$B8,$CE,$5B,$F0,$DF,$C3,$9B,$C3,$CA,$31,$1D,$DA,$FF,$7D,$94,$71
	.dc.b	$CF,$F0,$BF,$0A,$F8,$6F,$F0,$14,$61,$BA,$71,$AE,$0D,$C7,$ED,$5F
	.dc.b	$1C,$6D,$F0,$7F,$C7,$D8,$F8,$2B,$F0,$0D,$8F,$C5,$57,$16,$D7,$15
	.dc.b	$3B,$C7,$F3,$6D,$C3,$FC,$31,$7C,$67,$D1,$5D,$B1,$F3,$BE,$26,$A9
	.dc.b	$E1,$5C,$B0,$F0,$61,$D6,$B8,$63,$EC,$61,$FA,$18,$D3,$E6,$C3,$AE
	.dc.b	$1D,$23,$FC,$BC,$D7,$55,$C8,$FF,$75,$47,$C1,$6D,$FF,$8C,$DF,$F7
	.dc.b	$75,$F6,$9D,$6B,$F0,$3C,$ED,$37,$0F,$EA,$D3,$A7,$FC,$DD,$F0,$DF
	.dc.b	$D1,$DF,$EB,$5F,$03,$FD,$91,$BE,$07,$27,$7D,$55,$57,$17,$FA,$15
	.dc.b	$71,$6D,$6E,$75,$9B,$1F,$F2,$64,$3E,$17,$A6,$0F,$86,$A7,$18,$98
	.dc.b	$F9,$A3,$2F,$F7,$1A,$E4,$D9,$CB,$E4,$9E,$2F,$F5,$7F,$85,$FE,$76
	.dc.b	$FC,$0F,$86,$F0,$9E,$D8,$CC,$6F,$AB,$2E,$CF,$F0,$04,$6B,$3F,$64
	.dc.b	$7F,$A6,$63,$19,$FF,$9A,$73,$61,$CA,$BA,$1F,$E0,$9F,$86,$7C,$2F
	.dc.b	$C2,$F8,$69,$39,$8F,$A1,$3A,$8F,$AF,$3A,$7F,$B9,$46,$5B,$39,$CB
	.dc.b	$CB,$0C,$17,$D2,$BE,$E7,$C3,$38,$E9,$2F,$FC,$7B,$E6,$3D,$6F,$F2
	.dc.b	$0C,$7F,$88,$6B,$83,$FE,$73,$6B,$FE,$86,$6B,$FE,$CD,$6B,$FE,$8D
	.dc.b	$77,$97,$CA,$AE,$DF,$0D,$F9,$16,$D7,$7C,$3C,$FF,$0D,$23,$C4,$FE
	.dc.b	$5F,$E4,$3F,$C6,$7F,$39,$99,$F1,$4F,$F3,$2C,$8D,$24,$7B,$B3,$7C
	.dc.b	$47,$85,$B1,$D8,$FF,$20,$7C,$A7,$C2,$7F,$D4,$6D,$7F,$DD,$4D,$7F
	.dc.b	$D5,$4D,$7F,$DB,$DF,$12,$F8,$87,$9B,$FE,$9B,$8C,$37,$E2,$5E,$66
	.dc.b	$6B,$77,$FE,$B2,$F8,$97,$C4,$FF,$A8,$5B,$D5,$FE,$A5,$6C,$EE,$7F
	.dc.b	$D4,$DC,$6F,$91,$A4,$BF,$F5,$64,$65,$FF,$57,$62,$F8,$8F,$F5,$73
	.dc.b	$63,$31,$83,$FD,$65,$F0,$CF,$8F,$A4,$9F,$F5,$C4,$65,$FF,$5D,$7C
	.dc.b	$47,$C2,$7F,$CD,$9F,$AA,$FF,$B6,$1F,$F6,$94,$67,$4F,$F2,$84,$E5
	.dc.b	$08,$D3,$FE,$D9,$31,$C1,$FE,$C5,$8C,$C7,$48,$C1,$FB,$AE,$BF,$D3
	.dc.b	$CF,$77,$F5,$38,$7F,$8F,$A3,$18,$71,$3F,$D0,$4E,$2F,$C7,$3B,$BE
	.dc.b	$95,$B9,$FF,$34,$56,$D5,$6E,$79,$D7,$09,$F6,$D6,$EB,$51,$CA,$74
	.dc.b	$FB,$AB,$FD,$29,$3A,$BF,$5D,$6A,$F9,$7C,$25,$89,$7E,$03,$8C,$63
	.dc.b	$0F,$F3,$F6,$1C,$E7,$E7,$C7,$A2,$75,$87,$C6,$6C,$EA,$B8,$57,$C4
	.dc.b	$9C,$B6,$7C,$0F,$68,$CB,$D6,$33,$5D,$5A,$75,$9D,$D7,$6A,$E0,$7C
	.dc.b	$59,$F5,$9F,$90,$9F,$CA,$91,$C1,$A8,$F7,$F5,$FC,$CA,$FF,$3E,$92
	.dc.b	$B1,$19,$65,$71,$BF,$C3,$CE,$2F,$D9,$39,$3E,$24,$66,$7D,$93,$A3
	.dc.b	$C3,$3B,$7E,$31,$98,$EB,$19,$8F,$C4,$CF,$38,$EA,$66,$FF,$9E,$36
	.dc.b	$35,$3F,$96,$6B,$CA,$34,$47,$C0,$F4,$69,$C5,$D5,$7A,$8F,$2C,$68
	.dc.b	$E8,$5A,$F1,$B6,$BC,$31,$89,$E5,$5D,$DB,$5D,$CE,$11,$E0,$3F,$04
	.dc.b	$7E,$82,$B9,$91,$D4,$7C,$A9,$F2,$9D,$1F,$2B,$6B,$C5,$3D,$27,$E9
	.dc.b	$9E,$02,$4E,$EB,$B4,$62,$78,$4F,$66,$9E,$B2,$C7,$AE,$39,$47,$9A
	.dc.b	$3F,$33,$1C,$5F,$44,$71,$69,$C5,$E6,$FA,$9B,$5E,$A8,$D9,$92,$B9
	.dc.b	$7B,$C7,$56,$9D,$4D,$D7,$BA,$BC,$F1,$EF,$33,$3E,$9A,$FD,$3D,$6E
	.dc.b	$3C,$6D,$9D,$9C,$4E,$E7,$94,$B3,$F8,$67,$F9,$D3,$5F,$C9,$18,$8F
	.dc.b	$E9,$E4,$9F,$E9,$AD,$8F,$E8,$47,$59,$F8,$1F,$12,$F8,$9F,$DD,$DF
	.dc.b	$E2,$49,$4C,$3F,$C5,$35,$B9,$CC,$7D,$59,$F4,$4E,$5F,$C3,$C6,$6F
	.dc.b	$BD,$F6,$9C,$DF,$CF,$BE,$17,$BC,$34,$F8,$2E,$D7,$D5,$C3,$9B,$E2
	.dc.b	$31,$5F,$D0,$AF,$9C,$60,$D6,$1F,$D9,$A3,$2F,$DD,$3E,$24,$F1,$BE
	.dc.b	$F1,$9B,$EE,$FA,$0C,$4E,$EF,$AC,$72,$7F,$B9,$E1,$E9,$73,$5F,$8F
	.dc.b	$39,$C7,$BE,$38,$E1,$E4,$AE,$47,$70,$E6,$B9,$5F,$9A,$73,$85,$9C
	.dc.b	$DF,$88,$FE,$AF,$7C,$6F,$47,$A6,$FB,$57,$07,$F0,$15,$C2,$F6,$F9
	.dc.b	$9B,$1E,$5B,$E9,$3E,$EB,$E9,$87,$B6,$B5,$7C,$23,$F1,$17,$C2,$7C
	.dc.b	$35,$C2,$FE,$C4,$70,$7A,$1E,$12,$B9,$38,$B9,$69,$9A,$F1,$1E,$0B
	.dc.b	$F1,$4F,$E4,$5B,$3F,$83,$BF,$25,$70,$AF,$84,$78,$1E,$4D,$39,$1E
	.dc.b	$17,$33,$E1,$8C,$B4,$ED,$5F,$66,$BB,$CF,$69,$D1,$FA,$C9,$D1,$23
	.dc.b	$B1,$D2,$BB,$91,$C1,$27,$BB,$C1,$A7,$6C,$3F,$27,$1B,$8F,$03,$B6
	.dc.b	$9F,$C2,$AF,$F7,$CF,$F9,$35,$AF,$A4,$C5,$FD,$B5,$C1,$FE,$A9,$F8
	.dc.b	$9C,$43,$E2,$1F,$13,$FE,$3A,$EF,$FF,$5A,$E5,$F9,$9A,$FF,$2E,$D6
	.dc.b	$A3,$10,$9D,$BF,$98,$37,$5F,$EC,$07,$FC,$6F,$5C,$5F,$E4,$6D,$7F
	.dc.b	$CC,$CD,$77,$1F,$E4,$C8,$E1,$A7,$FB,$29,$6D,$99,$FF,$28,$FB,$0F
	.dc.b	$87,$FE,$25,$E7,$E1,$B9,$FA,$27,$4F,$F3,$4C,$91,$87,$F9,$BB,$5F
	.dc.b	$F3,$D3,$5F,$F2,$8B,$BC,$BF,$DB,$AD,$9F,$F4,$1E,$AF,$88,$7C,$4F
	.dc.b	$8E,$7E,$4F,$C3,$BE,$23,$FC,$DD,$7F,$ED,$DE,$CF,$F3,$5C,$E6,$BF
	.dc.b	$D8,$91,$8A,$F9,$99,$FF,$A4,$2B,$4F,$F4,$E6,$BF,$E9,$16,$BF,$E6
	.dc.b	$87,$79,$FF,$BD,$1B,$7E,$8F,$87,$7C,$4F,$78,$FB,$FF,$12,$F8,$9F
	.dc.b	$3D,$3E,$77,$FD,$03,$39,$AF,$F7,$BC,$E5,$E3,$5E,$8C,$95,$B7,$FB
	.dc.b	$2B,$5F,$F5,$43,$5F,$F3,$FB,$BC,$95,$8A,$F1,$FB,$1E,$27,$FD,$27
	.dc.b	$F1,$2F,$89,$A7,$FA,$6E,$9F,$12,$F8,$AC,$C7,$E7,$2B,$83,$6B,$85
	.dc.b	$6E,$30,$4C,$FF,$40,$2B,$8B,$FA,$38,$EF,$7D,$A3,$7E,$0B,$A3,$FD
	.dc.b	$52,$DA,$FD,$81,$E6,$F1,$B3,$32,$3F,$63,$3B,$AE,$D8,$FF,$BD,$A7
	.dc.b	$35,$B9,$EE,$FA,$AF,$1A,$FC,$A9,$FF,$6C,$FC,$33,$E2,$3F,$D7,$6D
	.dc.b	$C3,$F8,$15,$BE,$29,$F1,$5A,$B9,$5F,$C1,$2F,$86,$1F,$C0,$F0,$E1
	.dc.b	$7A,$8F,$8F,$3A,$C7,$C5,$08,$C5,$FF,$BB,$2F,$8B,$D7,$0F,$2F,$BB
	.dc.b	$23,$CB,$F1,$5F,$C1,$1F,$14,$67,$0E,$EF,$FB,$6E,$D5,$B9,$CD,$7A
	.dc.b	$7E,$2B,$E0,$9F,$ED,$C8,$FF,$7E,$F7,$FF,$BF,$BE,$21,$F1,$1F,$ED
	.dc.b	$2D,$17,$C3,$0E,$47,$F2,$6C,$38,$8B,$E2,$FD,$09,$EC,$FF,$12,$35
	.dc.b	$7D,$63,$EB,$4F,$09,$FE,$22,$68,$E5,$5E,$5A,$FE,$33,$19,$3E,$65
	.dc.b	$72,$8C,$D7,$3B,$F4,$DF,$1B,$FE,$3E,$71,$8C,$BC,$CF,$5C,$F1,$BE
	.dc.b	$91,$C2,$31,$5C,$E3,$CF,$18,$79,$1E,$F7,$07,$B5,$EB,$7F,$06,$79
	.dc.b	$46,$0F,$79,$CB,$0E,$D8,$79,$AF,$94,$F5,$BE,$4E,$CF,$66,$1E,$87
	.dc.b	$E5,$9F,$D0,$EF,$FA,$23,$F6,$0C,$1F,$E7,$AF,$86,$7C,$4B,$E2,$BF
	.dc.b	$A5,$46,$1B,$19,$9E,$0D,$44,$12,$F8,$AF,$F4,$04,$77,$6B,$B3,$6E
	.dc.b	$1E,$51,$C7,$87,$C7,$3C,$B3,$FC,$47,$C4,$BE,$2B,$C1,$18,$6E,$BC
	.dc.b	$AB,$83,$6F,$ED,$5F,$1C,$6D,$F0,$7D,$78,$F8,$5F,$C8,$37,$1E,$D1
	.dc.b	$C0,$FC,$24,$67,$0E,$D1,$F9,$7B,$EC,$DB,$EC,$FC,$0C,$38,$E3,$E6
	.dc.b	$8F,$97,$5C,$31,$EF,$1E,$29,$D3,$C4,$E8,$78,$CD,$57,$43,$13,$C8
	.dc.b	$FD,$86,$3E,$79,$EF,$87,$D9,$AF,$E3,$38,$7A,$1B,$19,$DF,$DB,$87
	.dc.b	$92,$77,$E8,$E7,$0F,$4B,$56,$1C,$EF,$F0,$DF,$15,$FD,$1E,$7F,$C6
	.dc.b	$3F,$0A,$F0,$38,$36,$3F,$C6,$B3,$B7,$FA,$AB,$71,$FE,$5F,$3B,$9F
	.dc.b	$EB,$2D,$8F,$E6,$73,$BB,$F0,$FC,$57,$6F,$8A,$98,$76,$8F,$45,$1C
	.dc.b	$C7,$49,$CC,$48,$F4,$C6,$AF,$BC,$FC,$1F,$8A,$74,$FB,$3E,$2A,$61
	.dc.b	$FE,$50,$9C,$B6,$72,$FA,$A7,$94,$48,$F8,$F1,$AB,$FE,$9F,$5C,$A7
	.dc.b	$FB,$CF,$C5,$3E,$27,$E5,$FC,$54,$9F,$F2,$FC,$E5,$B3,$97,$E4,$CF
	.dc.b	$D9,$9D,$47,$49,$D5,$FD,$0A,$D4,$7D,$FF,$88,$71,$52,$5F,$F9,$C7
	.dc.b	$D0,$7C,$33,$CC,$CC,$FF,$84,$2F,$8B,$FE,$74,$6B,$FE,$73,$AE,$33
	.dc.b	$C2,$3F,$85,$4E,$9B,$7F,$C8,$23,$11,$FC,$55,$7C,$47,$9C,$91,$FE
	.dc.b	$85,$F8,$9F,$F1,$07,$B5,$88,$F2,$4D,$D7,$FC,$1D,$5A,$9F,$34,$F7
	.dc.b	$23,$FE,$73,$BE,$EF,$F9,$D7,$E1,$BF,$CF,$ED,$F6,$BF,$9A,$C6,$23
	.dc.b	$F9,$20,$70,$FF,$27,$F8,$8F,$90,$FF,$9F,$DA,$F0,$6B,$BD,$3F,$A0
	.dc.b	$FC,$43,$E2,$1E,$7F,$FA,$1E,$30,$DC,$1E,$56,$6B,$78,$79,$3E,$13
	.dc.b	$EC,$76,$FF,$A3,$DB,$19,$CB,$FC,$C9,$18,$8F,$E7,$DC,$6F,$6B,$FC
	.dc.b	$D0,$DB,$FF,$4C,$36,$7E,$0B,$5F,$E8,$D8,$BE,$1B,$FC,$E0,$DA,$FF
	.dc.b	$50,$36,$7E,$4B,$5F,$E9,$D8,$3E,$07,$FD,$2B,$CF,$FE,$40,$BF,$F6
	.dc.b	$0C,$E5,$FC,$B4,$66,$BF,$CC,$73,$97,$E1,$57,$B8,$95,$F0,$DF,$F2
	.dc.b	$97,$2F,$F9,$0A,$B8,$3F,$E4,$4F,$85,$FA,$35,$E0,$AD,$BF,$DA,$6B
	.dc.b	$6D,$8C,$DF,$F6,$EA,$E2,$F9,$23,$2A,$B8,$BF,$DD,$1A,$A7,$8D,$F9
	.dc.b	$6B,$8D,$68,$DC,$EA,$FC,$47,$9C,$97,$FE,$86,$8D,$3D,$EB,$99,$27
	.dc.b	$FD,$DB,$3A,$9F,$34,$FA,$23,$15,$E0,$E0,$E1,$5D,$5B,$87,$C9,$6E
	.dc.b	$1F,$E0,$FB,$E3,$87,$39,$CD,$FC,$76,$BF,$05,$B3,$C2,$BC,$B1,$82
	.dc.b	$3E,$EB,$F0,$34,$F0,$46,$AB,$F6,$F5,$F5,$1F,$F0,$E3,$5D,$CF,$FB
	.dc.b	$D6,$3E,$11,$F0,$FE,$17,$FC,$F1,$7D,$AB,$E0,$57,$90,$FF,$16,$C7
	.dc.b	$FA,$32,$77,$1D,$27,$64,$9D,$72,$7C,$39,$E5,$39,$7F,$C6,$D5,$CA
	.dc.b	$3E,$34,$6A,$BC,$47,$27,$07,$77,$2F,$2A,$ED,$3D,$4E,$67,$FB,$2A
	.dc.b	$BB,$3F,$46,$7D,$67,$13,$AB,$FE,$4A,$8D,$DF,$C5,$AE,$B7,$E3,$AE
	.dc.b	$C2,$B7,$87,$E8,$9B,$87,$69,$E9,$7C,$A3,$C1,$1C,$A7,$6F,$9E,$74
	.dc.b	$F0,$9F,$2B,$DE,$32,$D8,$CC,$F2,$BF,$31,$F1,$A7,$91,$DE,$B8,$CE
	.dc.b	$EB,$C7,$5B,$AE,$86,$E3,$AD,$70,$23,$82,$46,$C9,$3C,$EB,$B4,$78
	.dc.b	$A3,$15,$CA,$37,$5F,$74,$F1,$95,$E0,$D7,$CB,$3C,$A3,$C2,$D8,$E6
	.dc.b	$7E,$09,$F8,$71,$C6,$BC,$73,$CA,$31,$5E,$3A,$E4,$48,$C9,$D8,$B3
	.dc.b	$D8,$FC,$61,$F6,$5D,$3E,$17,$D6,$D3,$D6,$53,$C4,$5A,$F2,$92,$BF
	.dc.b	$DF,$71,$F8,$83,$FD,$31,$1F,$EE,$F6,$C6,$9F,$BE,$62,$7F,$BD,$BF
	.dc.b	$70,$C1,$FD,$FA,$7F,$4A,$4E,$9F,$CD,$54,$AF,$CD,$57,$98,$FD,$04
	.dc.b	$E5,$F8,$04,$BF,$E0,$D1,$97,$E8,$C7,$C0,$38,$CF,$4B,$F0,$DE,$E3
	.dc.b	$F8,$2D,$DA,$E1,$5E,$59,$E1,$59,$9F,$24,$FF,$90,$27,$24,$8F,$23
	.dc.b	$D0,$C5,$F9,$23,$8D,$F3,$AF,$1D,$79,$63,$17,$96,$C7,$23,$A5,$76
	.dc.b	$3C,$F3,$FC,$0E,$77,$19,$3E,$09,$6B,$BB,$FC,$5E,$37,$7E,$59,$F6
	.dc.b	$4E,$EB,$C5,$7E,$18,$EE,$F9,$CC,$3F,$12,$3C,$51,$89,$EF,$5F,$5A
	.dc.b	$31,$3D,$2B,$B4,$6E,$FC,$6F,$E6,$A3,$D3,$5E,$5A,$E0,$D9,$E0,$74
	.dc.b	$2C,$F4,$3E,$05,$74,$7E,$EB,$5F,$A7,$1C,$27,$E1,$4F,$68,$F1,$5F
	.dc.b	$C4,$9C,$C6,$CC,$CE,$EB,$AC,$65,$A7,$89,$C3,$E0,$30,$7E,$96,$3F
	.dc.b	$58,$FD,$83,$F6,$2E,$2B,$C5,$3F,$82,$A2,$A9,$F0,$AF,$85,$FA,$71
	.dc.b	$F7,$E7,$EC,$2A,$E2,$F0,$9D,$C7,$D7,$8E,$06,$E7,$B2,$AE,$D1,$D2
	.dc.b	$78,$11,$DC,$6A,$79,$CE,$5F,$EC,$58,$7F,$8F,$27,$53,$99,$FC,$0D
	.dc.b	$7B,$CE,$77,$E6,$71,$1C,$9F,$E1,$35,$C9,$CC,$72,$34,$7A,$CF,$80
	.dc.b	$FA,$4D,$E1,$F3,$A3,$27,$A2,$33,$87,$9E,$FC,$53,$D3,$0F,$46,$1E
	.dc.b	$18,$C5,$F0,$3C,$B5,$C2,$3C,$84,$9E,$0E,$30,$F1,$39,$9E,$13,$A9
	.dc.b	$CB,$70,$FF,$32,$F5,$BB,$7F,$16,$DC,$7F,$9A,$B7,$1F,$F5,$8B,$5F
	.dc.b	$BA,$3E,$15,$F0,$BD,$2F,$FC,$E3,$18,$BF,$E6,$F7,$CF,$4F,$C5,$E1
	.dc.b	$C7,$0F,$CD,$DE,$E3,$FD,$7E,$D7,$BB,$5F,$C3,$FB,$1F,$A2,$CC,$D2
	.dc.b	$63,$FD,$3B,$D2,$7C,$2F,$E2,$67,$FD,$0B,$18,$9F,$C1,$C6,$27,$F5
	.dc.b	$B3,$9B,$FF,$68,$B5,$F8,$2D,$7F,$3F,$F0,$4F,$83,$FF,$55,$F5,$FF
	.dc.b	$A4,$A3,$18,$FF,$AA,$A3,$13,$A8,$FE,$B7,$3A,$AF,$F7,$E7,$47,$FB
	.dc.b	$07,$6F,$8F,$39,$9F,$A2,$D7,$C3,$7F,$C1,$23,$11,$FB,$FF,$81,$7C
	.dc.b	$0F,$FA,$A3,$B5,$DC,$F8,$2F,$F5,$3D,$6B,$4F,$F2,$35,$6A,$73,$7C
	.dc.b	$6F,$85,$7F,$72,$BE,$18,$7E,$9F,$0E,$13,$9B,$FE,$2F,$5A,$AF,$C4
	.dc.b	$B5,$FE,$0E,$D7,$F8,$B7,$C0,$BE,$0B,$FD,$76,$DB,$FF,$2D,$0F,$81
	.dc.b	$7C,$16,$AA,$4F,$F8,$1A,$B8,$5F,$EE,$E3,$85,$6A,$3F,$7C,$27,$7B
	.dc.b	$FE,$59,$B5,$FC,$A4,$E2,$75,$BF,$F6,$7F,$77,$E4,$A3,$FC,$2A,$DB
	.dc.b	$F0,$62,$F0,$B7,$CD,$2F,$6F,$FB,$71,$F6,$46,$6B,$CB,$7C,$F2,$FD
	.dc.b	$4B,$6B,$C9,$87,$57,$99,$FE,$26,$F7,$9F,$01,$FE,$E4,$6C,$7F,$A0
	.dc.b	$F1,$7C,$17,$FB,$F7,$1F,$DA,$D6,$B2,$F8,$AD,$8C,$E1,$FE,$F2,$BE
	.dc.b	$3F,$05,$D6,$26,$3E,$17,$58,$F5,$8F,$E8,$B8,$73,$6F,$A7,$E5,$3A
	.dc.b	$7C,$17,$FA,$9A,$51,$99,$F6,$3A,$F1,$71,$A7,$9B,$E0,$BD,$C7,$F5
	.dc.b	$8D,$7F,$CA,$5D,$BF,$D8,$7D,$C7,$C0,$78,$AB,$79,$7C,$77,$83,$FC
	.dc.b	$03,$05,$70,$7F,$B5,$66,$AD,$C7,$F0,$43,$77,$D6,$FC,$0F,$3B,$E5
	.dc.b	$2B,$E5,$5A,$8E,$53,$A9,$CC,$74,$33,$5D,$5C,$1F,$DC,$30,$E4,$F5
	.dc.b	$C3,$FB,$CD,$F8,$23,$FC,$05,$3A,$BE,$B3,$99,$4E,$6B,$8D,$FA,$4F
	.dc.b	$05,$6E,$75,$87,$82,$FA,$9E,$4C,$3A,$DF,$18,$F9,$38,$74,$7D,$66
	.dc.b	$2B,$93,$E2,$39,$46,$DF,$71,$88,$DD,$FF,$A3,$23,$C2,$7F,$41,$30
	.dc.b	$FF,$A4,$3D,$F7,$C0,$BE,$0B,$0D,$AF,$EB,$A2,$77,$3A,$39,$7C,$17
	.dc.b	$F8,$96,$77,$28,$C4,$7F,$66,$8C,$6B,$C0,$91,$F6,$3E,$05,$F0,$5F
	.dc.b	$E2,$8D,$FF,$0B,$18,$9F,$B7,$7C,$9F,$BE,$D9,$DD,$71,$DB,$FC,$63
	.dc.b	$7C,$6F,$78,$F9,$EF,$F2,$38,$71,$BD,$E1,$C6,$FF,$C8,$D8,$71,$BD
	.dc.b	$E1,$C6,$FF,$C7,$B8,$71,$BD,$E1,$C5,$61,$C6,$B7,$7C,$67,$FC,$A5
	.dc.b	$7C,$5D,$DF,$1B,$E1,$1A,$AE,$0F,$D3,$6C,$6A,$38,$9F,$1A,$37,$5D
	.dc.b	$67,$46,$DC,$57,$F9,$5B,$1F,$F0,$9C,$FC,$76,$C6,$75,$F4,$C7,$EB
	.dc.b	$27,$51,$F3,$BC,$CD,$63,$F2,$1A,$F9,$0F,$E9,$BF,$01,$FE,$16,$7F
	.dc.b	$DB,$3F,$02,$F8,$2D,$B6,$3F,$DB,$B3,$B7,$8F,$6E,$AA,$46,$A3,$BC
	.dc.b	$BE,$07,$9F,$C1,$4A,$E7,$1A,$30,$FF,$60,$C1,$39,$8D,$9E,$67,$7A
	.dc.b	$7F,$8E,$2F,$B4,$FF,$BC,$BE,$07,$F9,$AB,$EC,$F8,$2F,$59,$F0,$A3
	.dc.b	$31,$FD,$B9,$B5,$FE,$42,$C1,$5C,$2B,$A7,$57,$13,$0F,$9E,$30,$FF
	.dc.b	$A6,$EB,$FC,$97,$F0,$4F,$82,$F9,$7E,$AC,$AD,$70,$FF,$A5,$2B,$51
	.dc.b	$83,$FD,$48,$E1,$F8,$93,$C5,$FF,$02,$C6,$6F,$BD,$E5,$F5,$57,$E0
	.dc.b	$27,$57,$FE,$A2,$EF,$7C,$A7,$F1,$56,$E3,$FC,$38,$FE,$E3,$19,$23
	.dc.b	$FE,$32,$F4,$3E,$7B,$FE,$44,$E1,$FC,$6D,$72,$8F,$F0,$CB,$72,$FF
	.dc.b	$5B,$F9,$8B,$E3,$18,$6E,$3E,$5C,$7F,$2F,$FA,$BB,$E1,$BF,$95,$37
	.dc.b	$2F,$33,$5F,$E4,$AD,$7F,$D0,$6E,$70,$ED,$F2,$5F,$A4,$CB,$C4,$F6
	.dc.b	$B9,$7E,$D8,$D6,$47,$31,$FE,$34,$FB,$87,$D5,$7F,$3D,$6C,$7C,$96
	.dc.b	$BF,$CE,$1A,$FF,$A5,$9F,$F3,$0B,$BF,$47,$91,$1F,$6F,$C4,$FC,$02
	.dc.b	$38,$3E,$B7,$C2,$BE,$11,$E8,$61,$B3,$FD,$22,$5F,$0B,$FE,$3F,$8C
	.dc.b	$4E,$9F,$F5,$4F,$C6,$BE,$37,$FD,$F6,$DE,$4F,$C2,$36,$7F,$A9,$B5
	.dc.b	$FF,$58,$FC,$8B,$E4,$BF,$D5,$71,$FE,$F0,$8C,$57,$E3,$5B,$1F,$D7
	.dc.b	$1A,$FF,$AE,$B8,$FF,$D1,$D3,$F9,$7E,$0F,$F0,$DC,$A7,$28,$FF,$23
	.dc.b	$C6,$A1,$EF,$78,$A7,$51,$8A,$FC,$F3,$63,$FB,$6B,$5F,$14,$F3,$3C
	.dc.b	$3E,$AF,$FB,$1E,$BC,$13,$E3,$9D,$47,$EE,$63,$0D,$9E,$11,$E2,$9D
	.dc.b	$BF,$E7,$EA,$F5,$BF,$E7,$3E,$CC,$F3,$FF,$B5,$2F,$8C,$F4,$3F,$D0
	.dc.b	$AE,$2B,$8B,$6F,$CF,$7C,$B0,$F3,$CE,$5F,$24,$66,$FE,$54,$FF,$BA
	.dc.b	$DD,$C7,$C9,$7C,$93,$FC,$16,$75,$1C,$27,$BC,$FB,$FE,$26,$CF,$9A
	.dc.b	$6E,$3E,$B9,$B5,$D8,$F1,$9C,$4B,$5E,$E8,$DB,$E6,$6B,$FE,$99,$6D
	.dc.b	$9A,$49,$E7,$5E,$17,$D5,$1E,$D9,$E1,$18,$25,$78,$5F,$B9,$F0,$AF
	.dc.b	$8F,$FC,$1D,$C9,$F4,$57,$FA,$E2,$B5,$18,$C3,$C1,$3C,$CD,$5F,$17
	.dc.b	$11,$A1,$3F,$8E,$AE,$73,$A8,$FB,$77,$C6,$B7,$3F,$A3,$AD,$C7,$3A
	.dc.b	$D9,$E3,$3A,$5F,$17,$35,$D4,$E5,$19,$6C,$71,$69,$B1,$5E,$08,$E7
	.dc.b	$5E,$49,$CD,$73,$6C,$F3,$69,$CE,$B9,$3F,$7E,$3E,$8C,$F1,$9C,$CF
	.dc.b	$E0,$67,$27,$38,$CE,$1F,$72,$7E,$7C,$E6,$B9,$11,$C1,$23,$C5,$19
	.dc.b	$6C,$66,$FE,$64,$7C,$26,$BE,$A6,$DF,$8B,$0F,$E3,$2D,$AF,$3D,$74
	.dc.b	$25,$70,$BF,$03,$67,$C0,$7C,$C3,$EC,$98,$9F,$8C,$7B,$8E,$6E,$DF
	.dc.b	$33,$5F,$33,$6B,$C9,$1C,$67,$C0,$F2,$3A,$BF,$6E,$36,$DA,$F3,$D7
	.dc.b	$36,$D7,$08,$F3,$BC,$CF,$B0,$77,$78,$1E,$B7,$33,$D6,$31,$87,$DB
	.dc.b	$8C,$4F,$38,$C4,$F6,$9F,$49,$EA,$9F,$E6,$E7,$E3,$8B,$7F,$AC,$6D
	.dc.b	$FF,$1F,$8C,$47,$A0,$CD,$7F,$89,$63,$31,$F3,$08,$3E,$49,$F1,$1F
	.dc.b	$80,$B9,$7F,$80,$C3,$F4,$95,$A8,$C1,$E9,$25,$FF,$5D,$8D,$3F,$1C
	.dc.b	$EF,$87,$E8,$A3,$35,$F1,$09,$5C,$AB,$C9,$87,$89,$E4,$7E,$08,$EC
	.dc.b	$E6,$78,$C7,$B2,$BA,$D7,$3B,$FD,$89,$E1,$BE,$11,$9C,$3F,$BB,$D7
	.dc.b	$0B,$F3,$C6,$9E,$F3,$F0,$1C,$BF,$DA,$67,$84,$79,$49,$7D,$E7,$84
	.dc.b	$62,$BB,$DF,$58,$C4,$F3,$8C,$57,$A2,$B7,$3C,$A3,$0E,$CC,$47,$3B
	.dc.b	$F1,$CE,$AF,$E4,$C6,$27,$A9,$D2,$3E,$6C,$EF,$0F,$BF,$86,$B0,$F4
	.dc.b	$E1,$C5,$EE,$78,$6F,$8E,$1E,$1A,$E1,$1E,$06,$DF,$D6,$8F,$BB,$5C
	.dc.b	$9D,$36,$35,$7D,$E7,$AC,$F7,$9E,$58,$79,$E3,$FC,$3B,$5B,$9F,$4C
	.dc.b	$7A,$2F,$C8,$D2,$5F,$80,$93,$E0,$9D,$C6,$27,$F2,$95,$D5,$EF,$3A
	.dc.b	$AE,$11,$D2,$BF,$73,$5F,$A2,$6D,$7F,$85,$63,$11,$CA,$7D,$87,$F8
	.dc.b	$DB,$E2,$DF,$29,$F5,$70,$F5,$D1,$A7,$F5,$79,$DD,$72,$9F,$CB,$9B
	.dc.b	$7B,$34,$FF,$29,$47,$07,$07,$D4,$2D,$FE,$70,$FC,$EC,$6E,$BC,$35
	.dc.b	$D1,$B5,$D2,$FD,$13,$92,$46,$5F,$33,$DE,$78,$AA,$ED,$3C,$23,$15
	.dc.b	$D1,$E4,$79,$B0,$FF,$37,$D7,$08,$F8,$37,$F1,$6B,$84,$72,$BF,$0D
	.dc.b	$6E,$3C,$51,$F4,$67,$73,$9C,$3E,$0C,$60,$93,$C9,$B5,$FC,$E0,$EC
	.dc.b	$5A,$F9,$CF,$F3,$C6,$BD,$E7,$FC,$E5,$1E,$9D,$FF,$DA,$1F,$0D,$EC
	.dc.b	$BF,$EE,$7E,$77,$F3,$F2,$5F,$F3,$D6,$E1,$FE,$3B,$7F,$D2,$3F,$0F
	.dc.b	$FC,$09,$BA,$7F,$99,$63,$F2,$33,$FE,$93,$6A,$6F,$C4,$3E,$27,$FC
	.dc.b	$D8,$DC,$3F,$DE,$0D,$AF,$80,$D7,$FA,$CB,$9D,$BF,$C5,$8D,$9F,$F4
	.dc.b	$9A,$F7,$7F,$D4,$EE,$6B,$FD,$4C,$E6,$7B,$CF,$88,$8F,$FC,$01,$F8
	.dc.b	$8F,$F3,$ED,$1A,$4E,$5F,$F4,$23,$FE,$94,$3F,$D3,$BB,$BE,$1F,$FD
	.dc.b	$8D,$97,$F9,$06,$75,$5F,$EB,$A7,$73,$AC,$3F,$D5,$71,$FE,$9E,$F8
	.dc.b	$77,$C9,$FF,$A4,$5B,$3F,$83,$C9,$18,$7F,$0C,$D7,$FB,$FB,$9C,$7F
	.dc.b	$CA,$4D,$9F,$F6,$1E,$0F,$6F,$FD,$9B,$39,$BF,$F6,$63,$5E,$F8,$7F
	.dc.b	$9B,$B0,$FF,$37,$E7,$EF,$BE,$33,$FE,$D6,$F8,$77,$C3,$D3,$86,$7F
	.dc.b	$ED,$71,$F0,$EF,$88,$D4,$7F,$B2,$6B,$8B,$6B,$8E,$7F,$E8,$4B,$E1
	.dc.b	$39,$24,$7F,$10,$AE,$2F,$F1,$58,$EF,$7D,$A3,$F8,$AF,$8E,$F6,$BF
	.dc.b	$DD,$AD,$AF,$F7,$66,$BF,$E8,$C7,$CF,$DD,$AA,$91,$FE,$DF,$AE,$17
	.dc.b	$DF,$25,$EA,$3B,$DD,$C3,$A5,$6B,$0E,$53,$FE,$FE,$F8,$57,$C3,$D3
	.dc.b	$E0,$E2,$F4,$DF,$11,$9C,$3F,$DE,$17,$C1,$B7,$C2,$B7,$18,$26,$7F
	.dc.b	$CD,$AB,$83,$FC,$EF,$0E,$D8,$75,$8D,$7B,$3E,$53,$C7,$F1,$1F,$AB
	.dc.b	$8C,$E1,$E2,$9D,$3F,$28,$FF,$76,$CE,$F5,$FF,$52,$62,$AF,$81,$F1
	.dc.b	$12,$F8,$61,$FE,$58,$F8,$82,$F8,$6F,$CC,$38,$C1,$7C,$63,$D8,$FE
	.dc.b	$21,$E2,$2B,$9D,$72,$85,$72,$BF,$0C,$70,$AD,$F3,$72,$AD,$4F,$8E
	.dc.b	$3F,$9B,$4E,$A3,$C9,$87,$FB,$53,$0E,$D1,$EA,$9D,$3E,$09,$D5,$72
	.dc.b	$8E,$87,$2C,$36,$F1,$32,$71,$72,$FB,$4C,$E1,$E6,$C3,$C3,$19,$7E
	.dc.b	$04,$66,$BC,$11,$8B,$F1,$D7,$D1,$39,$B8,$7E,$21,$8B,$F0,$E1,$EC
	.dc.b	$C3,$A4,$F1,$BE,$95,$A8,$E0,$7A,$8F,$64,$69,$C1,$EB,$3F,$90,$57
	.dc.b	$F2,$17,$E1,$18,$3F,$B3,$7C,$2B,$E2,$1F,$11,$FC,$9E,$30,$D8,$CC
	.dc.b	$F0,$6B,$FE,$A6,$97,$C4,$7F,$89,$63,$AB,$5D,$35,$FC,$6C,$70,$DB
	.dc.b	$C1,$3C,$71,$FC,$77,$C3,$BE,$23,$F2,$71,$86,$EF,$CA,$B8,$36,$FF
	.dc.b	$0F,$7C,$71,$B7,$C1,$FD,$2E,$3E,$17,$F0,$AD,$C7,$B4,$70,$3F,$23
	.dc.b	$19,$C3,$B4,$7E,$32,$FB,$36,$FB,$2A,$E0,$F8,$31,$F3,$BF,$B6,$AE
	.dc.b	$38,$78,$23,$C7,$3A,$79,$36,$77,$87,$ED,$9B,$3D,$71,$F3,$D6,$F1
	.dc.b	$F1,$D6,$AB,$9D,$7B,$5F,$F2,$8C,$7A,$A1,$AF,$5A,$FE,$07,$18,$AE
	.dc.b	$C7,$BC,$FE,$03,$4F,$1F,$77,$E5,$6C,$FB,$EE,$AF,$86,$FE,$98,$FF
	.dc.b	$0A,$F8,$57,$C4,$69,$B1,$FC,$3A,$76,$FF,$9A,$1B,$97,$96,$13,$C1
	.dc.b	$FF,$36,$B6,$35,$3C,$31,$FE,$3E,$BE,$0F,$C1,$F1,$13,$0F,$04,$7A
	.dc.b	$27,$51,$E7,$9D,$3E,$39,$D1,$8B,$F2,$60,$AC,$D7,$F2,$8F,$88,$FE
	.dc.b	$78,$FC,$1F,$88,$F8,$07,$C6,$8C,$C7,$FA,$19,$B7,$D2,$77,$3C,$FA
	.dc.b	$38,$18,$7E,$0C,$61,$FE,$6F,$F0,$EF,$88,$F9,$DE,$5C,$9C,$C7,$D6
	.dc.b	$9D,$47,$D5,$9D,$3F,$E9,$88,$CB,$67,$38,$7C,$0A,$E5,$1F,$80,$EF
	.dc.b	$7C,$45,$25,$FF,$5A,$F2,$5F,$0C,$FA,$0C,$CF,$F9,$F6,$F8,$BF,$E3
	.dc.b	$D6,$B9,$BF,$E4,$11,$8C,$BF,$C6,$37,$F6,$2B,$FD,$65,$87,$6E,$1E
	.dc.b	$B5,$F8,$9F,$89,$7C,$4F,$E5,$9B,$7F,$D7,$9A,$F2,$C1,$E9,$6E,$BF
	.dc.b	$98,$56,$AB,$FB,$58,$70,$FF,$6E,$6B,$FD,$E1,$EB,$7E,$8F,$89,$FD
	.dc.b	$4E,$5F,$11,$ED,$5F,$DC,$3E,$E3,$F9,$D3,$FE,$5F,$6B,$FE,$A3,$6B
	.dc.b	$E2,$6B,$FD,$FB,$E2,$9F,$03,$FD,$1D,$B1,$FE,$69,$6B,$FE,$AD,$6B
	.dc.b	$EC,$5F,$15,$C4,$91,$C5,$FE,$A5,$F2,$1F,$E7,$F8,$FF,$3A,$35,$FF
	.dc.b	$5C,$35,$F8,$ED,$7F,$C2,$7F,$14,$F8,$97,$A5,$FE,$7D,$8C,$37,$E2
	.dc.b	$BF,$C9,$75,$FE,$C3,$9C,$BF,$EC,$5F,$89,$7C,$57,$F0,$78,$FF,$0F
	.dc.b	$C6,$3C,$2F,$F4,$43,$63,$84,$ED,$FF,$7C,$FB,$7F,$E2,$7F,$D6,$3F
	.dc.b	$C9,$D7,$FE,$CA,$8C,$74,$FF,$B9,$1E,$15,$FE,$ED,$6F,$C2,$7F,$A8
	.dc.b	$6B,$FD,$2C,$D9,$FF,$6C,$4F,$19,$FF,$16,$D7,$D6,$7B,$12,$5C,$BF
	.dc.b	$C8,$6F,$93,$FC,$8E,$B8,$4E,$AB,$FC,$7B,$5A,$6D,$F7,$9D,$1F,$E4
	.dc.b	$AA,$E3,$7D,$67,$FC,$8D,$18,$AE,$27,$9B,$8B,$95,$78,$5B,$1F,$E4
	.dc.b	$C9,$E1,$7F,$2A,$38,$1E,$BA,$E0,$D3,$85,$7C,$AA,$F5,$D7,$EB,$1A
	.dc.b	$EA,$7F,$82,$4F,$98,$FF,$2E,$72,$BE,$27,$13,$E0,$AF,$79,$FE,$87
	.dc.b	$AD,$CE,$A3,$37,$E8,$75,$5F,$E6,$0B,$E8,$F3,$C3,$FA,$94,$62,$78
	.dc.b	$E1,$D2,$B4,$7F,$9B,$A7,$55,$F0,$2B,$C7,$18,$9F,$64,$63,$0F,$25
	.dc.b	$76,$6C,$F3,$7F,$1F,$5E,$1A,$FE,$22,$D7,$53,$FE,$E5,$8E,$D1,$FE
	.dc.b	$88,$6B,$FE,$B4,$EF,$F4,$3E,$19,$F8,$15,$FE,$87,$F5,$95,$AC,$3C
	.dc.b	$95,$A3,$FD,$0B,$3A,$71,$7F,$42,$30,$F5,$31,$5C,$E7,$FD,$27,$3A
	.dc.b	$AE,$B5,$A9,$E7,$1F,$E0,$19,$D4,$7B,$A7,$47,$59,$D5,$7E,$D2,$BC
	.dc.b	$33,$FC,$D9,$AF,$82,$34,$7B,$CF,$81,$F1,$36,$FD,$B1,$67,$D9,$36
	.dc.b	$FB,$D6,$8F,$87,$19,$7D,$AD,$9F,$F5,$CD,$A3,$11,$FE,$E6,$6C,$66
	.dc.b	$7A,$4F,$81,$B5,$F0,$EB,$64,$8C,$9E,$02,$CF,$82,$34,$FC,$89,$E5
	.dc.b	$3A,$8F,$13,$6B,$C5,$1B,$8E,$8D,$3A,$3B,$AE,$B5,$F1,$27,$D0,$6C
	.dc.b	$EA,$5A,$EB,$1B,$3C,$71,$99,$E7,$5F,$5C,$F9,$A7,$72,$46,$DB,$3C
	.dc.b	$DF,$3B,$63,$CE,$6C,$8E,$5F,$2B,$6B,$CB,$3E,$19,$F5,$CE,$6B,$CB
	.dc.b	$1F,$72,$FF,$13,$5E,$23,$85,$7B,$0E,$0E,$67,$83,$EC,$7C,$6F,$03
	.dc.b	$D2,$57,$D3,$1C,$9D,$9C,$8F,$2C,$6D,$F5,$B6,$FD,$71,$B9,$F9,$EF
	.dc.b	$E9,$6F,$F5,$04,$AF,$F0,$EC,$7C,$83,$4E,$1F,$B6,$60,$DD,$76,$7C
	.dc.b	$20,$7C,$43,$E2,$7E,$3C,$7A,$0F,$6D,$6D,$DC,$6B,$0F,$95,$3F,$12
	.dc.b	$7A,$57,$38,$F9,$53,$A9,$E6,$6A,$B9,$4F,$18,$F2,$9D,$9E,$26,$6F
	.dc.b	$D5,$7D,$67,$CC,$70,$8C,$D7,$48,$EE,$74,$9E,$17,$E3,$8E,$18,$7F
	.dc.b	$AB,$E3,$FD,$39,$3A,$C3,$C1,$3C,$EB,$C8,$78,$30,$F9,$87,$82,$7C
	.dc.b	$F1,$E0,$8C,$3B,$9E,$55,$D6,$35,$18,$C3,$B5,$74,$8F,$25,$FD,$1B
	.dc.b	$E1,$7E,$88,$EE,$F9,$30,$F6,$5F,$19,$FB,$D1,$EC,$C3,$AE,$1D,$E7
	.dc.b	$8D,$6E,$FF,$BA,$E1,$DA,$31,$87,$C7,$8C,$5F,$2B,$EA,$79,$70,$C6
	.dc.b	$1F,$52,$B7,$3D,$AB,$D7,$3F,$81,$8C,$1E,$A9,$DC,$7C,$78,$EA,$EC
	.dc.b	$C1,$D6,$BB,$DF,$95,$B3,$F8,$BA,$E7,$3F,$12,$FC,$37,$F1,$EB,$EE
	.dc.b	$9E,$1B,$F8,$91,$D6,$30,$47,$04,$AF,$04,$65,$B1,$98,$F1,$4F,$49
	.dc.b	$CC,$FF,$A5,$AF,$F3,$4D,$9F,$F2,$CC,$62,$3E,$65,$7C,$27,$D0,$7E
	.dc.b	$E7,$E2,$5F,$0F,$EC,$AF,$C2,$92,$FD,$F3,$AB,$E5,$39,$6C,$EA,$B9
	.dc.b	$32,$32,$78,$6B,$95,$FE,$1A,$74,$D9,$D5,$63,$0F,$C8,$47,$29,$FF
	.dc.b	$20,$DF,$29,$CC,$62,$B8,$57,$F0,$FA,$D3,$F0,$A7,$4F,$3A,$3B,$AE
	.dc.b	$33,$F1,$5A,$AB,$9B,$F8,$BA,$E2,$EE,$BA,$1C,$CA,$E4,$F7,$1E,$22
	.dc.b	$C7,$27,$B5,$78,$2B,$BC,$7D,$83,$35,$D1,$A7,$27,$C0,$FF,$3E,$8F
	.dc.b	$9E,$FF,$B7,$1A,$E8,$C4,$FD,$99,$FC,$37,$0B,$E2,$31,$5F,$BC,$F8
	.dc.b	$6F,$C6,$92,$BF,$A4,$66,$8F,$C8,$FA,$32,$72,$DA,$FE,$A3,$48,$FC
	.dc.b	$81,$FA,$CF,$88,$F7,$C6,$6B,$F3,$1E,$23,$E2,$3F,$63,$F1,$FF,$E4
	.dc.b	$8C,$7F,$A9,$4E,$6B,$F3,$D3,$99,$FF,$7F,$D7,$FA,$F2,$B5,$1F,$D8
	.dc.b	$9A,$FF,$C0,$06,$BF,$D5,$DD,$EF,$F5,$5B,$79,$CF,$F8,$1C,$62,$BF
	.dc.b	$B2,$7C,$4F,$97,$99,$E4,$3C,$AF,$E1,$33,$98,$C6,$3F,$B5,$8F,$F2
	.dc.b	$DD,$F2,$E7,$FF,$69,$C6,$20,$D3,$DC,$77,$EB,$FF,$71,$69,$F8,$B8
	.dc.b	$F8,$5F,$12,$F8,$9F,$E5,$3A,$2D,$E8,$7F,$1C,$AD,$67,$FC,$A2,$B5
	.dc.b	$39,$BF,$F7,$2D,$6B,$2F,$F0,$5B,$5F,$EF,$C1,$C3,$FE,$01,$78,$65
	.dc.b	$F9,$F6,$E1,$FE,$9B,$C5,$ED,$77,$8F,$E3,$1F,$10,$F8,$9F,$E6,$2D
	.dc.b	$8F,$E7,$A3,$E2,$1F,$13,$AA,$93,$C9,$FE,$92,$EE,$33,$9F,$FB,$32
	.dc.b	$FA,$36,$B7,$7D,$2F,$91,$FC,$AF,$C9,$7C,$4F,$F4,$46,$C7,$92,$5E
	.dc.b	$16,$6A,$4F,$17,$AC,$63,$0F,$F6,$ED,$F2,$6C,$EA,$B8,$1F,$CE,$FE
	.dc.b	$15,$F1,$3F,$D5,$1B,$1F,$D8,$30,$78,$EF,$89,$D5,$CA,$E4,$FF,$6C
	.dc.b	$AD,$CE,$71,$FF,$77,$E1,$D1,$B5,$BC,$3A,$61,$C8,$F2,$7B,$5E,$33
	.dc.b	$A7,$C4,$FF,$05,$8C,$92,$FA,$CE,$BB,$78,$4F,$B7,$E2,$7D,$87,$F7
	.dc.b	$0A,$FF,$28,$7C,$4F,$F9,$47,$E1,$BF,$0F,$1F,$E5,$3D,$15,$C2,$F9
	.dc.b	$1F,$D8,$67,$88,$AE,$2A,$B8,$CE,$E1,$7D,$F9,$7F,$CB,$55,$A9,$E5
	.dc.b	$1F,$E6,$C9,$D4,$78,$F0,$FE,$8B,$87,$68,$F0,$BE,$0B,$E4,$E6,$3B
	.dc.b	$9D,$0E,$58,$74,$7F,$CD,$57,$6B,$8E,$1E,$19,$E0,$7A,$F0,$F0,$C6
	.dc.b	$70,$F0,$46,$27,$F6,$F7,$CE,$31,$1D,$8E,$77,$D9,$D9,$88,$DB,$EC
	.dc.b	$BF,$8B,$18,$8F,$5D,$F8,$30,$F5,$46,$CF,$69,$E8,$30,$7A,$5D,$BF
	.dc.b	$10,$FF,$4B,$61,$FE,$97,$7E,$D9,$83,$FD,$3B,$F0,$EF,$88,$7C,$4F
	.dc.b	$FA,$76,$30,$D8,$CC,$F0,$6B,$FC,$6A,$5F,$12,$EC,$E4,$D9,$EE,$D7
	.dc.b	$83,$5F,$F5,$04,$F2,$9F,$B5,$F1,$2F,$89,$FF,$52,$69,$F9,$28,$C5
	.dc.b	$FD,$FB,$E0,$FD,$C6,$CE,$EB,$8E,$DF,$EA,$DB,$E3,$7B,$C7,$CD,$7F
	.dc.b	$9E,$C3,$8D,$EF,$0E,$37,$FE,$BB,$C3,$8D,$EF,$0E,$37,$EA,$6D,$6E
	.dc.b	$F8,$AB,$E3,$5B,$C7,$D5,$7C,$2B,$57,$C2,$F7,$1F,$EC,$BC,$7C,$AD
	.dc.b	$AE,$13,$A8,$E2,$74,$B8,$FE,$82,$34,$6E,$B9,$C7,$03,$ED,$6B,$F0
	.dc.b	$5F,$F3,$24,$FC,$8A,$FF,$19,$57,$08,$C5,$7C,$DF,$2F,$FD,$B5,$5F
	.dc.b	$E3,$79,$FF,$0D,$5F,$2C,$73,$7F,$ED,$FF,$88,$FE,$8B,$3F,$E7,$1F
	.dc.b	$88,$7C,$4E,$A3,$FD,$D5,$18,$7F,$CE,$D3,$C1,$F3,$9F,$EE,$F7,$18
	.dc.b	$FF,$91,$E7,$84,$E8,$38,$8F,$F2,$64,$F0,$BF,$27,$C2,$78,$6F,$7F
	.dc.b	$13,$70,$39,$B5,$53,$51,$2B,$CF,$19,$7D,$D1,$97,$C9,$3C,$9F,$F4
	.dc.b	$97,$C4,$BA,$3C,$1E,$3C,$9C,$C7,$96,$75,$1F,$26,$74,$FB,$E3,$2D
	.dc.b	$9C,$E7,$F1,$6B,$94,$7B,$FE,$25,$F1,$3E,$FF,$32,$61,$FE,$A5,$9C
	.dc.b	$B6,$72,$FC,$DD,$55,$CE,$64,$FD,$D8,$DE,$3F,$E7,$BB,$E7,$3F,$57
	.dc.b	$BB,$FC,$84,$FF,$AC,$DA,$E7,$CE,$FC,$27,$99,$FE,$48,$F8,$9F,$AD
	.dc.b	$5A,$C3,$FD,$0C,$DE,$26,$12,$30,$AF,$93,$DF,$0F,$E5,$13,$E0,$8F
	.dc.b	$F1,$BB,$5E,$67,$F8,$22,$3F,$97,$7C,$43,$E2,$7F,$10,$D9,$FF,$66
	.dc.b	$35,$FC,$BD,$FF,$98,$3D,$2D,$D4,$9D,$BE,$EF,$49,$39,$9F,$F2,$48
	.dc.b	$BE,$8F,$F9,$31,$AF,$99,$AF,$C2,$FC,$65,$F0,$3F,$D1,$5B,$1D,$03
	.dc.b	$87,$BB,$5F,$F2,$CF,$C4,$FF,$49,$78,$35,$F2,$B5,$DF,$85,$FD,$4F
	.dc.b	$E2,$5F,$05,$49,$3E,$18,$CB,$E3,$F8,$6F,$F5,$43,$7C,$AF,$EB,$91
	.dc.b	$88,$F8,$6D,$7F,$CD,$BF,$11,$FE,$79,$7D,$8D,$7E,$4B,$5D,$FA,$DF
	.dc.b	$D9,$FE,$1D,$F1,$0F,$53,$E2,$46,$1B,$87,$F9,$62,$FE,$9C,$E5,$FA
	.dc.b	$BF,$AC,$FF,$88,$7A,$7F,$DC,$3C,$DF,$E9,$EF,$81,$FF,$71,$C7,$FB
	.dc.b	$6E,$75,$D5,$F9,$6A,$D3,$F9,$89,$D7,$A1,$FE,$E1,$C3,$F2,$37,$FE
	.dc.b	$2D,$7F,$D1,$6D,$73,$1B,$7F,$43,$1D,$9B,$1E,$2A,$FC,$DF,$C4,$7F
	.dc.b	$98,$9F,$01,$C8,$B3,$C1,$AA,$BA,$29,$E4,$FF,$2F,$8C,$92,$BC,$11
	.dc.b	$FB,$49,$DD,$FB,$AF,$AC,$61,$F1,$4F,$F3,$98,$C3,$E3,$31,$D9,$D1
	.dc.b	$F2,$5D,$9D,$C6,$2B,$AD,$6F,$0F,$4D,$69,$B1,$F2,$6B,$D8,$D9,$F6
	.dc.b	$C7,$FA,$E9,$AF,$CA,$8F,$01,$E3,$F8,$3F,$F6,$55,$F8,$09,$3E,$87
	.dc.b	$E5,$C6,$A3,$04,$C5,$3D,$4D,$F0,$76,$69,$D2,$37,$3F,$C8,$2B,$A4
	.dc.b	$62,$3D,$73,$C1,$EC,$FF,$1C,$9D,$BD,$E3,$FB,$44,$61,$FB,$17,$ED
	.dc.b	$BF,$14,$69,$F9,$13,$EB,$8F,$F6,$FB,$5E,$D1,$A3,$95,$7F,$3F,$E4
	.dc.b	$B3,$CE,$3C,$D8,$7F,$3C,$9C,$BE,$BB,$F2,$51,$B8,$7F,$0E,$BE,$27
	.dc.b	$F3,$87,$13,$D4,$FE,$A1,$7D,$27,$4D,$8F,$E9,$0D,$7F,$D8,$ED,$9E
	.dc.b	$77,$E3,$C3,$F2,$51,$DA,$BF,$A4,$D7,$E2,$E3,$11,$DA,$FF,$98,$47
	.dc.b	$27,$11,$FE,$17,$D1,$F1,$1D,$2E,$57,$C7,$3F,$67,$3A,$71,$5E,$C6
	.dc.b	$E3,$F8,$19,$D4,$E4,$ED,$7D,$1B,$5C,$0F,$EB,$F3,$C2,$BE,$94,$78
	.dc.b	$2B,$84,$75,$6C,$70,$6D,$7F,$1D,$6B,$FD,$C5,$B5,$D6,$7E,$0D,$73
	.dc.b	$7C,$55,$C5,$AF,$16,$CF,$18,$FA,$F1,$E5,$33,$5D,$88,$E0,$91,$F0
	.dc.b	$EB,$C0,$D8,$F5,$1A,$8F,$83,$1F,$18,$D1,$E5,$29,$E5,$2D,$79,$5F
	.dc.b	$C4,$CF,$95,$CD,$79,$1B,$1D,$CF,$86,$78,$9C,$92,$BC,$67,$A0,$B1
	.dc.b	$B6,$C7,$95,$AF,$95,$B5,$E5,$C3,$F2,$53,$BA,$F0,$B4,$F9,$A7,$C1
	.dc.b	$AE,$37,$E9,$6C,$FA,$67,$71,$E5,$8E,$71,$E1,$6D,$F8,$70,$FE,$74
	.dc.b	$D8,$F8,$D5,$D2,$BE,$09,$FE,$32,$3E,$71,$67,$F3,$AD,$BF,$E7,$51
	.dc.b	$88,$F9,$46,$6B,$FC,$41,$19,$8F,$71,$E5,$3F,$C9,$3E,$C7,$F9,$72
	.dc.b	$7C,$67,$CD,$24,$6B,$C2,$F4,$DC,$9D,$CE,$5A,$F5,$25,$75,$7E,$64
	.dc.b	$74,$73,$5C,$DC,$9E,$69,$E1,$87,$CD,$8F,$4E,$1F,$8F,$9C,$DF,$38
	.dc.b	$FC,$91,$9B,$E9,$7C,$6F,$BD,$79,$AB,$78,$7F,$6D,$C3,$9D,$76,$8E
	.dc.b	$35,$FE,$64,$9C,$CF,$B6,$B8,$CF,$F8,$22,$73,$18,$9F,$58,$8C,$D7
	.dc.b	$48,$F7,$61,$5E,$91,$E3,$8C,$57,$3C,$3F,$AF,$5F,$2C,$3D,$57,$B6
	.dc.b	$9F,$4A,$35,$87,$8A,$FE,$24,$EA,$3C,$B3,$AB,$F8,$D5,$A9,$E1,$7E
	.dc.b	$5A,$D5,$FC,$F8,$F0,$4E,$A3,$C1,$87,$A2,$78,$36,$32,$49,$F0,$C6
	.dc.b	$A3,$18,$7D,$7B,$EF,$5F,$8D,$37,$3F,$E7,$B8,$FC,$64,$7E,$9D,$AF
	.dc.b	$D7,$AF,$F0,$F4,$7D,$82,$7C,$47,$A2,$FE,$5C,$DB,$FB,$D3,$6F,$C1
	.dc.b	$5A,$9F,$BD,$39,$9F,$B8,$DB,$34,$93,$F8,$5B,$FC,$74,$62,$BF,$DF
	.dc.b	$15,$E4,$AF,$64,$FA,$08,$E0,$F0,$9F,$E5,$6F,$84,$7C,$47,$48,$CD
	.dc.b	$F4,$AF,$F2,$45,$6A,$73,$8F,$E2,$A7,$85,$23,$FC,$55,$5B,$8E,$55
	.dc.b	$C0,$F8,$A7,$86,$F8,$D7,$38,$D4,$FF,$B5,$2B,$94,$E6,$39,$91,$EF
	.dc.b	$5C,$AB,$AC,$E5,$FF,$69,$D7,$27,$35,$C1,$EF,$29,$D3,$FC,$72,$FE
	.dc.b	$0D,$F5,$8F,$23,$5E,$0D,$9E,$31,$F7,$2B,$DC,$76,$8C,$E1,$F7,$A3
	.dc.b	$33,$E7,$3A,$BB,$AE,$84,$7F,$DA,$55,$F3,$67,$F9,$F3,$5F,$7C,$FF
	.dc.b	$9D,$63,$C6,$79,$3E,$21,$F1,$1E,$CB,$FF,$4B,$FC,$97,$FA,$ED,$FF
	.dc.b	$6F,$37,$0F,$F7,$5B,$5F,$E9,$8D,$7E,$6F,$67,$72,$4F,$99,$F1,$7A
	.dc.b	$92,$3D,$2D,$9F,$F7,$BF,$C4,$7E,$C1,$07,$0D,$DF,$FD,$21,$18,$F1
	.dc.b	$5F,$11,$FE,$9D,$8F,$E5,$B1,$8C,$3F,$D6,$71,$89,$FF,$4C,$4E,$6F
	.dc.b	$FB,$30,$70,$FF,$69,$6B,$E4,$1F,$11,$FC,$FB,$91,$E4,$3C,$AF,$F5
	.dc.b	$74,$E6,$31,$97,$FB,$12,$31,$3A,$CF,$FD,$7F,$7C,$A3,$15,$E4,$C3
	.dc.b	$BD,$7F,$AB,$EF,$E1,$9F,$13,$E1,$DF,$11,$CB,$AD,$E9,$7F,$AF,$E7
	.dc.b	$38,$FF,$B5,$B3,$56,$A3,$FD,$AF,$3A,$CB,$F7,$51,$FE,$EF,$9D,$46
	.dc.b	$0F,$AB,$3B,$E9,$FF,$7C,$CE,$9F,$F5,$54,$73,$9F,$F0,$93,$5F,$94
	.dc.b	$D7,$9E,$3F,$EC,$BF,$87,$7C,$3F,$FB,$5D,$B8,$FF,$BB,$87,$C3,$BE
	.dc.b	$23,$55,$27,$FD,$E3,$5C,$30,$F4,$47,$0A,$D4,$7A,$C4,$EF,$2F,$F6
	.dc.b	$13,$6B,$F2,$67,$13,$AD,$FF,$BA,$FC,$87,$C4,$53,$FD,$F3,$3E,$6E
	.dc.b	$EC,$3F,$9C,$C3,$FD,$9D,$5C,$23,$07,$E3,$5C,$63,$E8,$CF,$FC,$01
	.dc.b	$3A,$8F,$D2,$CE,$AB,$9E,$1D,$A7,$86,$16,$7F,$DB,$9F,$0A,$F8,$7F
	.dc.b	$F7,$0C,$61,$B8,$FF,$8E,$BC,$47,$C4,$63,$1F,$D9,$E1,$C6,$BC,$50
	.dc.b	$8C,$3F,$2D,$B8,$7F,$BB,$B1,$61,$D2,$F8,$D6,$9F,$D6,$36,$FB,$E3
	.dc.b	$E1,$BE,$38,$F8,$E7,$C1,$36,$9E,$D5,$38,$7C,$43,$BF,$19,$78,$26
	.dc.b	$E1,$F9,$BC,$97,$BA,$D1,$E6,$F2,$FB,$1E,$1A,$FE,$2B,$F1,$1F,$C5
	.dc.b	$BE,$1D,$F0,$FE,$AA,$DD,$F1,$3E,$3D,$F0,$15,$C1,$F9,$B3,$D5,$5F
	.dc.b	$5B,$E9,$76,$7F,$CF,$93,$5F,$A5,$49,$C4,$77,$BE,$CE,$A3,$2F,$63
	.dc.b	$C7,$7D,$9A,$76,$29,$C4,$A7,$57,$84,$F6,$8C,$1E,$C9,$EF,$7E,$7D
	.dc.b	$18,$79,$EB,$99,$E3,$BF,$3C,$EA,$F8,$CF,$8A,$30,$F0,$BE,$97,$CE
	.dc.b	$FC,$53,$CF,$0F,$29,$FC,$FA,$B7,$87,$4A,$EF,$7C,$6B,$57,$FC,$C4
	.dc.b	$F8,$C7,$F3,$3B,$C3,$83,$FA,$27,$C3,$3D,$67,$C4,$7F,$39,$8C,$36
	.dc.b	$33,$3F,$D4,$27,$2D,$F8,$8F,$F1,$53,$79,$39,$46,$23,$FD,$79,$28
	.dc.b	$C4,$78,$9E,$39,$FE,$03,$E2,$1F,$11,$FD,$66,$30,$DD,$38,$63,$F8
	.dc.b	$BC,$7A,$3F,$8D,$BE,$38,$DB,$E0,$FF,$76,$AD,$BF,$86,$6E,$1D,$A3
	.dc.b	$81,$D7,$1F,$2E,$3E,$08,$FC,$6E,$1D,$9B,$87,$65,$87,$4A,$E1,$87
	.dc.b	$59,$D9,$D7,$18,$E0,$91,$FA,$2A,$E5,$1E,$9C,$7C,$6F,$F8,$43,$1F
	.dc.b	$15,$75,$38,$4E,$A5,$3E,$7B,$FC,$D4,$63,$1F,$F0,$34,$62,$75,$B7
	.dc.b	$A3,$2F,$F0,$54,$62,$FC,$1E,$4F,$0A,$CB,$63,$C9,$85,$5F,$0B,$FD
	.dc.b	$41,$FE,$67,$F0,$8F,$88,$D3,$63,$F9,$BC,$ED,$FE,$C0,$DC,$FD,$33
	.dc.b	$B9,$FE,$C8,$D8,$F6,$CE,$F1,$FF,$11,$2F,$88,$EF,$F1,$13,$0E,$F1
	.dc.b	$E7,$9D,$47,$9A,$74,$F8,$A7,$46,$2F,$C7,$38,$9F,$E9,$FF,$11,$FC
	.dc.b	$F9,$F7,$F9,$1E,$E3,$E2,$46,$63,$FB,$E3,$6B,$DB,$3B,$9E,$7D,$5C
	.dc.b	$0C,$3F,$02,$30,$FF,$5E,$F8,$77,$C4,$7C,$CF,$32,$4E,$63,$EA,$4E
	.dc.b	$A3,$E9,$CE,$9F,$F0,$74,$65,$B3,$9C,$3E,$8D,$72,$8F,$BD,$F7,$5F
	.dc.b	$EB,$4A,$FE,$DF,$C3,$FD,$D1,$BE,$0E,$BD,$79,$18,$21,$FE,$1D,$AE
	.dc.b	$4D,$8F,$C5,$57,$38,$FF,$11,$37,$1F,$CD,$D7,$07,$F3,$BC,$9F,$6A
	.dc.b	$BF,$C5,$1F,$40,$FA,$4F,$CA,$36,$3F,$C0,$B3,$BA,$E3,$C5,$AB,$95
	.dc.b	$A8,$FC,$E5,$6F,$93,$FC,$DD,$87,$3A,$EF,$7A,$BF,$D1,$7C,$33,$E4
	.dc.b	$BF,$49,$F1,$0E,$1A,$6A,$2E,$1F,$E1,$8F,$8C,$7D,$A7,$EA,$DB,$1F
	.dc.b	$E1,$CB,$E4,$EE,$7F,$65,$F3,$5F,$ED,$DF,$AD,$BC,$07,$94,$FE,$D7
	.dc.b	$E3,$1F,$25,$FB,$7F,$C6,$FF,$D1,$98,$FF,$8A,$1A,$FF,$94,$67,$6F
	.dc.b	$F8,$B3,$E4,$1F,$2F,$FB,$FF,$D8,$3F,$B3,$CF,$F8,$C1,$AF,$F9,$66
	.dc.b	$76,$FB,$7A,$FF,$84,$7B,$9F,$E9,$4F,$89,$BF,$00,$7E,$04,$FF,$0B
	.dc.b	$F8,$E7,$C8,$3D,$9F,$F4,$84,$61,$BF,$20,$F5,$B3,$5B,$CF,$FD,$53
	.dc.b	$F8,$77,$F9,$61,$FF,$53,$07,$0F,$FA,$A5,$AF,$FA,$F9,$AB,$A1,$F4
	.dc.b	$FF,$CB,$23,$FD,$61,$18,$F0,$3F,$D4,$8D,$8F,$F5,$CB,$5F,$F5,$B7
	.dc.b	$4F,$F7,$2D,$BF,$CB,$4F,$F7,$4F,$72,$FA,$BF,$0F,$0F,$F3,$64,$E5
	.dc.b	$B1,$9B,$FF,$50,$4E,$5B,$19,$EB,$FF,$0D,$4E,$5B,$1E,$BA,$E7,$F2
	.dc.b	$5E,$0C,$BC,$2D,$8C,$9E,$08,$CB,$FE,$6E,$FC,$0F,$FC,$D1,$8F,$FA
	.dc.b	$5D,$B8,$7F,$8A,$7D,$A7,$8E,$F1,$FF,$CC,$D5,$AC,$3F,$DB,$0D,$C3
	.dc.b	$FD,$DA,$DB,$E3,$5C,$3C,$A9,$7E,$17,$FD,$05,$3A,$8C,$1F,$1B,$E1
	.dc.b	$7E,$5C,$FC,$C6,$C6,$4F,$97,$19,$69,$F3,$BE,$8B,$E9,$9F,$53,$E1
	.dc.b	$6C,$64,$99,$7D,$08,$CB,$FE,$B9,$F9,$97,$CB,$FF,$A1,$5B,$5F,$EA
	.dc.b	$7C,$BE,$2F,$D0,$E6,$A4,$FF,$A6,$AB,$77,$FE,$DD,$BD,$CE,$6B,$73
	.dc.b	$DD,$F8,$7B,$30,$CE,$FF,$80,$F8,$BF,$C6,$9F,$8E,$EB,$B8,$FE,$32
	.dc.b	$2E,$3F,$8A,$9B,$5F,$88,$AF,$CA,$FC,$A7,$E5,$9A,$7E,$50,$B5,$F9
	.dc.b	$26,$CF,$F8,$4B,$4F,$F0,$14,$7B,$7D,$AF,$F5,$57,$EB,$BF,$97,$85
	.dc.b	$CA,$DB,$F2,$EB,$E5,$1C,$3A,$AD,$73,$24,$FF,$AA,$6F,$A3,$CE,$BF
	.dc.b	$5E,$BE,$42,$98,$F3,$ED,$FE,$C2,$74,$D9,$D3,$4D,$7D,$07,$4F,$88
	.dc.b	$EA,$D8,$C9,$CE,$32,$D3,$FD,$AF,$F3,$FF,$D1,$DB,$E9,$7F,$B7,$7E
	.dc.b	$B1,$F8,$F3,$E8,$FF,$D9,$D7,$BC,$DC,$3F,$E3,$E8,$C3,$70,$E5,$7C
	.dc.b	$71,$EA,$4B,$E2,$FD,$05,$7C,$A3,$E9,$E3,$E2,$C7,$C3,$84,$9E,$58
	.dc.b	$FF,$A9,$EB,$51,$83,$A5,$F3,$C3,$EE,$B6,$77,$1E,$23,$37,$D5,$C1
	.dc.b	$E4,$7D,$78,$67,$0F,$76,$1D,$E9,$7D,$27,$31,$CB,$0F,$13,$CE,$BF
	.dc.b	$1F,$3E,$9C,$7E,$34,$EF,$1E,$D5,$EA,$7C,$2F,$78,$E4,$6D,$E6,$7C
	.dc.b	$89,$E7,$5E,$22,$47,$33,$73,$E8,$9D,$C6,$63,$A1,$27,$CD,$8F,$92
	.dc.b	$3E,$3C,$FC,$1C,$7E,$03,$C6,$BB,$CE,$8F,$9D,$3A,$71,$1A,$7D,$6F
	.dc.b	$69,$F7,$63,$F2,$0F,$B1,$87,$9A,$B8,$CF,$72,$3E,$BC,$3C,$CF,$48
	.dc.b	$CE,$1E,$18,$C1,$C8,$8E,$6F,$C1,$8F,$BE,$30,$F4,$8F,$1D,$D8,$D6
	.dc.b	$1E,$5B,$EA,$4A,$DE,$1E,$CA,$DD,$F1,$8F,$95,$7C,$6B,$C3,$18,$9D
	.dc.b	$5F,$D3,$8F,$BD,$18,$C7,$E7,$E3,$E6,$9E,$C7,$DE,$71,$5C,$4F,$7E
	.dc.b	$1D,$F1,$E5,$5D,$0E,$33,$BA,$E3,$7F,$5F,$1F,$5C,$F3,$3D,$D7,$F9
	.dc.b	$67,$11,$EA,$7F,$03,$87,$7C,$3A,$DF,$DF,$9D,$DF,$C4,$9C,$9F,$62
	.dc.b	$B8,$4E,$A3,$D9,$18,$AF,$25,$78,$5B,$3F,$24,$E7,$5F,$03,$1F,$7B
	.dc.b	$63,$2F,$E1,$31,$F7,$C7,$E3,$23,$18,$77,$8C,$63,$F0,$B0,$EF,$5F
	.dc.b	$1E,$B4,$FE,$34,$E3,$87,$81,$B7,$E0,$BE,$4F,$E6,$BC,$2F,$EA,$7F
	.dc.b	$53,$FD,$23,$E0,$7F,$A4,$B6,$7F,$A3,$36,$FF,$C6,$61,$C6,$5F,$E3
	.dc.b	$28,$7C,$2F,$F5,$23,$FA,$9F,$D6,$5B,$FE,$A1,$16,$79,$CD,$AF,$F7
	.dc.b	$DF,$53,$C7,$61,$FC,$2A,$FF,$C0,$11,$FD,$B3,$E9,$33,$1F,$CA,$AB
	.dc.b	$85,$6D,$B1,$9A,$E1,$39,$78,$DF,$81,$B1,$FD,$1E,$B8,$CF,$6F,$31
	.dc.b	$D0,$C7,$78,$FF,$7C,$BD,$E1,$FD,$FB,$6F,$EA,$0F,$F6,$FC,$3F,$BA
	.dc.b	$D7,$F7,$66,$C6,$4F,$EE,$91,$97,$F9,$48,$FB,$E7,$CC,$F8,$23,$0D
	.dc.b	$87,$C1,$58,$93,$CA,$78,$B6,$73,$5F,$27,$E0,$FD,$6F,$F8,$B5,$FF
	.dc.b	$0C,$F8,$D8,$CD,$14,$97,$FC,$FB,$2F,$8B,$58,$BE,$93,$F4,$EB,$9B
	.dc.b	$4F,$27,$D8,$7A,$7E,$AF,$D4,$D8,$C9,$E9,$8C,$B4,$F6,$7D,$C7,$C1
	.dc.b	$3E,$17,$5D,$8C,$93,$0F,$EE,$4F,$CB,$8D,$47,$F7,$2F,$C3,$DF,$11
	.dc.b	$F0,$9B,$5E,$DE,$3F,$F5,$BB,$F9,$38,$CD,$71,$AE,$07,$F9,$03,$D5
	.dc.b	$B5,$B8,$B3,$FE,$39,$9B,$CB,$F9,$AA,$FB,$DE,$7F,$DB,$F8,$8F,$B8
	.dc.b	$D8,$FB,$4D,$BF,$5F,$27,$FA,$46,$7E,$F5,$7E,$0F,$D5,$91,$F8,$1F
	.dc.b	$66,$DF,$09,$B8,$F8,$70,$56,$2F,$FC,$51,$F8,$7F,$E3,$1F,$D9,$37
	.dc.b	$0F,$F2,$97,$97,$F8,$88,$C6,$1F,$B4,$94,$62,$B9,$F4,$7F,$99,$2F
	.dc.b	$83,$23,$AB,$96,$9F,$C0,$7E,$8B,$95,$FF,$9A,$9A,$BE,$89,$F4,$5F
	.dc.b	$9A,$9C,$C6,$2B,$F8,$53,$6B,$FC,$63,$F4,$8F,$6B,$F4,$33,$9D,$3D
	.dc.b	$53,$9B,$FE,$24,$DB,$FF,$1E,$7C,$03,$E0,$3F,$82,$56,$B1,$FD,$5F
	.dc.b	$D4,$F3,$8F,$F2,$96,$7F,$5F,$0F,$F2,$0C,$E5,$F6,$72,$70,$D6,$5F
	.dc.b	$F8,$AB,$2E,$DE,$56,$2F,$C3,$9F,$F8,$0F,$1E,$55,$AB,$FE,$51,$1D
	.dc.b	$69,$E1,$BC,$DF,$DF,$B6,$BF,$87,$FC,$EF,$F3,$D7,$F9,$84,$66,$BF
	.dc.b	$CA,$ED,$E0,$FB,$CD,$9F,$F4,$E8,$AE,$55,$C2,$FB,$92,$7B,$F9,$56
	.dc.b	$7F,$CC,$70,$F9,$8F,$E2,$6D,$3C,$BE,$56,$22,$47,$F4,$53,$53,$FB
	.dc.b	$EC,$15,$A8,$EB,$57,$2F,$D8,$56,$AF,$94,$FF,$22,$F9,$F7,$CF,$D3
	.dc.b	$F9,$85,$3E,$83,$17,$FD,$56,$B6,$DA,$DC,$EB,$1F,$F3,$FD,$69,$B3
	.dc.b	$AA,$E0,$7F,$5F,$F8,$5F,$E9,$73,$FE,$C8,$6C,$7C,$DF,$2B,$F9,$BC
	.dc.b	$6A,$4D,$37,$8A,$77,$E7,$4B,$EC,$FB,$23,$17,$FE,$94,$F7,$BE,$E3
	.dc.b	$FE,$D8,$6C,$3E,$81,$F0,$3F,$D2,$27,$74,$69,$C2,$77,$87,$F7,$B8
	.dc.b	$CF,$17,$13,$FB,$5E,$1D,$70,$C5,$6F,$E5,$7B,$1A,$F9,$F7,$D0,$62
	.dc.b	$74,$4C,$FF,$D6,$18,$2B,$73,$9E,$3F,$E2,$D9,$BE,$83,$D4,$7F,$62
	.dc.b	$C3,$FD,$EF,$F4,$1F,$EF,$8F,$9F,$FF,$0E,$C7,$C3,$9D,$E1,$EB,$78
	.dc.b	$42,$B8,$7B,$1F,$ED,$1C,$3B,$5F,$F8,$6E,$BA,$8C,$3A,$BF,$27,$0F
	.dc.b	$F0,$EC,$62,$36,$EA,$B9,$C7,$1B,$FF,$6F,$3A,$7C,$15,$C9,$CC,$79
	.dc.b	$4E,$07,$AE,$FC,$EF,$2C,$2B,$E3,$8A,$F5,$38,$1C,$8F,$F7,$74,$70
	.dc.b	$BF,$2E,$1E,$99,$F0,$3E,$9C,$3C,$91,$DC,$E5,$3C,$30,$F3,$DF,$3A
	.dc.b	$DD,$76,$36,$E1,$F2,$1E,$73,$C8,$62,$BA,$3D,$EF,$C5,$87,$B2,$3E
	.dc.b	$21,$F1,$8A,$78,$8C,$18,$03,$E8,$3F,$8C,$7C,$67,$16,$C7,$D0,$6B
	.dc.b	$F5,$9A,$F2,$39,$7D,$03,$4E,$4D,$8E,$4D,$79,$35,$E5,$28,$E1,$87
	.dc.b	$E3,$3E,$81,$E7,$F7,$8C,$37,$6E,$35,$C1,$B8,$FE,$0A,$F8,$E3,$6F
	.dc.b	$83,$F0,$B1,$F0,$57,$E2,$1B,$1F,$85,$AE,$2D,$AE,$2F,$C0,$AE,$18
	.dc.b	$78,$63,$E6,$D7,$0C,$7C,$F1,$AA,$8F,$33,$C8,$62,$3F,$57,$87,$89
	.dc.b	$E3,$1B,$3F,$31,$19,$C7,$DF,$19,$9D,$9C,$B0,$F2,$43,$0F,$5E,$1F
	.dc.b	$E1,$F6,$E7,$F6,$DB,$3F,$16,$B8,$C6,$23,$97,$8B,$F8,$1B,$E4,$DB
	.dc.b	$E5,$9D,$AF,$C7,$FC,$FF,$FA,$95,$FF,$52,$7D,$03,$E1,$B5,$08,$C3
	.dc.b	$FE,$A9,$9E,$0F,$E3,$6F,$F2,$11,$8C,$7F,$C7,$15,$C2,$B4,$7E,$79
	.dc.b	$C4,$7F,$8F,$67,$85,$72,$F9,$DF,$0D,$3E,$06,$9E,$4A,$4D,$33,$87
	.dc.b	$AA,$70,$F9,$E3,$2F,$88,$F6,$46,$9C,$7D,$03,$A7,$C3,$EB,$C9,$FF
	.dc.b	$64,$4E,$5B,$39,$7A,$63,$FE,$80,$72,$43,$35,$BC,$FF,$CB,$97,$CE
	.dc.b	$7C,$9E,$F7,$F9,$A1,$FA,$7F,$11,$F0,$8F,$A7,$19,$8D,$FA,$D9,$76
	.dc.b	$7F,$B7,$63,$57,$FE,$69,$6E,$DF,$C1,$72,$61,$C6,$F9,$9F,$EE,$5F
	.dc.b	$8D,$FF,$60,$BF,$EE,$7E,$1F,$EE,$AD,$EF,$FC,$2C,$EA,$3E,$63,$72
	.dc.b	$FF,$3C,$57,$09,$FA,$0D,$8F,$F3,$E4,$EF,$1F,$E3,$0D,$AF,$E4,$75
	.dc.b	$B9,$FE,$38,$D8,$FE,$4F,$3B,$5C,$7F,$84,$7F,$DF,$DF,$40,$FA,$0F
	.dc.b	$CD,$B6,$3F,$97,$E3,$FE,$8D,$AE,$11,$F8,$56,$C7,$FA,$46,$78,$4F
	.dc.b	$E2,$5B,$1C,$78,$FF,$4F,$5C,$32,$FE,$65,$5A,$6E,$1A,$AF,$D6,$FD
	.dc.b	$03,$E0,$3F,$A1,$B7,$2F,$E9,$0D,$7F,$B7,$B5,$FE,$EB,$F4,$1F,$69
	.dc.b	$FE,$9C,$D7,$FB,$AB,$5D,$EF,$FD,$43,$E5,$9F,$3F,$49,$3F,$D5,$E3
	.dc.b	$2F,$F5,$9F,$9D,$7C,$FF,$F5,$86,$C6,$63,$07,$F5,$CF,$9C,$7C,$F5
	.dc.b	$24,$FF,$63,$8C,$BD,$FA,$3C,$84,$70,$7F,$69,$FC,$17,$F4,$CF,$F8
	.dc.b	$51,$AF,$F5,$F6,$BF,$DA,$9A,$BE,$79,$F0,$FF,$E5,$C8,$FF,$0C,$C6
	.dc.b	$3D,$2F,$EE,$CB,$E6,$3F,$CB,$D3,$AA,$E0,$FF,$89,$3E,$7B,$FD,$80
	.dc.b	$DF,$80,$FF,$6B,$3F,$DC,$63,$2D,$8C,$BF,$EA,$8F,$2D,$F3,$0E,$9F
	.dc.b	$C9,$9F,$E0,$2F,$75,$3F,$E7,$86,$CE,$EB,$F9,$5C,$6B,$D6,$C7,$76
	.dc.b	$63,$78,$74,$AF,$05,$6A,$3B,$61,$E2,$75,$5C,$DB,$0E,$5E,$17,$E9
	.dc.b	$F8,$3B,$7E,$69,$B5,$FE,$DB,$C7,$C4,$F1,$C3,$9D,$EA,$B6,$F4,$C1
	.dc.b	$F3,$DF,$E7,$C7,$FC,$DF,$FA,$F3,$E3,$35,$39,$CB,$E1,$7A,$EA,$D6
	.dc.b	$5E,$FC,$39,$F8,$DF,$41,$B1,$FE,$33,$C3,$B4,$7F,$97,$9A,$FF,$8A
	.dc.b	$5E,$D1,$67,$E0,$FA,$8F,$1B,$FD,$E1,$97,$D3,$C3,$A9,$F5,$36,$7C
	.dc.b	$FF,$C4,$C3,$E2,$B7,$1F,$9A,$DA,$E0,$DB,$F9,$3F,$27,$F3,$0F,$9B
	.dc.b	$F3,$F6,$FE,$AC,$5B,$F0,$4D,$AF,$27,$CE,$BE,$7F,$EC,$C3,$F2,$EF
	.dc.b	$C0,$5F,$F6,$3A,$D6,$3F,$87,$AD,$73,$FD,$59,$CD,$7F,$A0,$1A,$FF
	.dc.b	$A1,$9A,$FF,$94,$5A,$FF,$97,$1E,$33,$E4,$C3,$E5,$EC,$F9,$8E,$D1
	.dc.b	$F8,$AF,$A0,$99,$7E,$4B,$B6,$E5,$F3,$66,$E3,$F8,$E9,$B5,$F9,$C3
	.dc.b	$F2,$7F,$25,$F9,$46,$9F,$9C,$2D,$FD,$76,$D7,$C8,$F6,$3F,$D9,$3F
	.dc.b	$2D,$FE,$E3,$D7,$FC,$49,$5F,$AB,$E8,$4E,$6B,$F1,$D5,$AC,$3E,$34
	.dc.b	$60,$97,$F9,$59,$D6,$1F,$9A,$BF,$A6,$F8,$67,$F6,$76,$FA,$0F,$D7
	.dc.b	$B6,$33,$E5,$E2,$64,$70,$77,$18,$9F,$E0,$E3,$E8,$39,$FC,$2F,$46
	.dc.b	$DF,$81,$B3,$C1,$B8,$7E,$7F,$E7,$BF,$AC,$57,$F0,$DE,$B7,$E3,$EF
	.dc.b	$2B,$F7,$95,$AC,$7F,$C8,$35,$A9,$CF,$A5,$FA,$CB,$E1,$87,$EE,$70
	.dc.b	$E1,$8F,$F1,$FC,$38,$4E,$6F,$86,$1C,$A3,$C5,$87,$2C,$7F,$C7,$D8
	.dc.b	$F2,$9F,$F5,$6E,$1C,$A7,$38,$72,$9E,$87,$2C,$3A,$4F,$A3,$0E,$98
	.dc.b	$FC,$1B,$E1,$3C,$AF,$81,$C9,$E1,$7C,$AB,$D5,$7C,$AB,$FD,$6D,$8F
	.dc.b	$96,$D8,$F9,$6F,$B5,$74,$9E,$C7,$47,$B4,$74,$C3,$B4,$7B,$B0,$ED
	.dc.b	$5F,$EC,$CC,$3B,$61,$D6,$BC,$38,$FA,$B1,$EF,$7D,$4E,$EF,$59,$EE
	.dc.b	$75,$C7,$BC,$FA,$31,$EE,$B1,$EF,$18,$6C,$F0,$C7,$C5,$5E,$5C,$7C
	.dc.b	$58,$79,$0F,$81,$8F,$90,$F1,$4F,$92,$3C,$47,$91,$F1,$5F,$5C,$7C
	.dc.b	$17,$D2,$7B,$57,$43,$B3,$EF,$BE,$D1,$D2,$3B,$1D,$1E,$D8,$7A,$23
	.dc.b	$DD,$5D,$5B,$3D,$5E,$E7,$57,$BD,$7A,$8E,$F1,$D4,$EF,$5D,$71,$F8
	.dc.b	$F5,$E0,$C3,$D6,$FF,$BF,$6B,$D2,$59,$FD,$73,$70,$FE,$C1,$18,$8D
	.dc.b	$90,$F2,$9B,$9F,$95,$F4,$2F,$A1,$7C,$AF,$DE,$AF,$69,$F1,$27,$51
	.dc.b	$9B,$FC,$D4,$EB,$3F,$F0,$B5,$6B,$0F,$AD,$39,$7E,$18,$51,$B7,$E8
	.dc.b	$9B,$3E,$5C,$F4,$C3,$CB,$5D,$2F,$C2,$7E,$AE,$F9,$57,$CE,$BE,$47
	.dc.b	$38,$CD,$78,$A3,$AB,$63,$31,$82,$1C,$6F,$C2,$F9,$AF,$C3,$8F,$F8
	.dc.b	$9A,$B8,$CF,$82,$73,$7F,$E3,$39,$CC,$62,$7E,$65,$F4,$C3,$D1,$39
	.dc.b	$7E,$06,$3F,$3F,$1F,$D8,$E1,$C9,$C6,$3D,$70,$E4,$FB,$71,$F1,$4F
	.dc.b	$96,$73,$1E,$CC,$7B,$57,$D3,$BE,$8F,$CD,$BE,$18,$76,$BF,$3B,$71
	.dc.b	$FC,$DD,$F0,$9F,$89,$5C,$2F,$91,$F8,$EC,$7C,$33,$E4,$8C,$61,$CD
	.dc.b	$FA,$53,$AC,$7F,$93,$78,$FF,$87,$F9,$2F,$EC,$7F,$03,$FD,$91,$B7
	.dc.b	$FD,$A1,$B8,$FF,$60,$6E,$9F,$E6,$58,$7C,$37,$F6,$C3,$FB,$6F,$C9
	.dc.b	$5B,$FE,$E3,$16,$7F,$B4,$CD,$AF,$F2,$07,$73,$CC,$FF,$03,$FD,$77
	.dc.b	$F9,$6E,$31,$8F,$F8,$1F,$A5,$39,$AF,$F2,$A5,$6B,$1F,$DE,$1F,$E2
	.dc.b	$38,$C9,$2F,$FC,$B9,$3B,$C3,$FC,$11,$F4,$5C,$E7,$FC,$0F,$7F,$E0
	.dc.b	$6E,$DF,$F0,$3B,$63,$27,$F8,$32,$32,$D3,$FD,$0C,$3E,$75,$F4,$5F
	.dc.b	$E8,$08,$C3,$61,$F3,$B6,$24,$F2,$9E,$2D,$9C,$DB,$1F,$93,$F0,$BF
	.dc.b	$E2,$EF,$27,$E4,$36,$FD,$EF,$C6,$23,$F1,$AB,$FC,$91,$18,$7A,$47
	.dc.b	$F0,$B7,$2F,$D1,$F2,$BA,$BE,$5F,$A2,$F5,$7D,$17,$AD,$B1,$93,$D5
	.dc.b	$19,$69,$ED,$FA,$2F,$84,$7C,$3E,$CB,$19,$26,$1E,$43,$FD,$7D,$AF
	.dc.b	$F7,$F8,$F9,$D5,$FE,$A1,$FA,$27,$C5,$FF,$94,$34,$C3,$6B,$FD,$39
	.dc.b	$3A,$7E,$13,$70,$FC,$4B,$5F,$E8,$CD,$7E,$23,$BD,$BF,$D3,$0D,$9F
	.dc.b	$E6,$5C,$3F,$E5,$CF,$6B,$C3,$1F,$6F,$E2,$3F,$01,$F1,$3F,$81,$6C
	.dc.b	$64,$FC,$04,$65,$A7,$E0,$FD,$8F,$C4,$1F,$89,$F1,$AC,$64,$99,$FC
	.dc.b	$77,$F2,$51,$A8,$F9,$43,$E8,$BF,$23,$3F,$ED,$DF,$C0,$5F,$1B,$F9
	.dc.b	$0A,$DD,$69,$FA,$53,$AE,$9F,$F6,$84,$E6,$7F,$B6,$35,$FC,$53,$5F
	.dc.b	$EA,$CD,$7C,$5D,$3C,$8F,$F4,$3F,$A1,$71,$FF,$69,$7D,$0E,$C9,$3F
	.dc.b	$ED,$37,$97,$D1,$BE,$03,$FD,$AC,$DA,$FC,$F3,$70,$FE,$0A,$D7,$FC
	.dc.b	$05,$F4,$6F,$80,$FF,$6F,$36,$BF,$4C,$D9,$FE,$1A,$D7,$FC,$15,$87
	.dc.b	$FB,$6A,$7F,$50,$D8,$FE,$E9,$F4,$3E,$1E,$7F,$F6,$F4,$E6,$53,$E3
	.dc.b	$7F,$60,$D7,$8E,$5F,$EF,$18,$C7,$D1,$3E,$8F,$FA,$0C,$75,$C7,$FC
	.dc.b	$05,$39,$F3,$D1,$FE,$FB,$9D,$61,$FE,$07,$F3,$5E,$A6,$A3,$F0,$B1
	.dc.b	$8A,$FE,$37,$F3,$EF,$3B,$5B,$7F,$BD,$EB,$57,$C3,$0C,$E1,$FC,$26
	.dc.b	$73,$5D,$61,$87,$37,$FC,$31,$1C,$EB,$8D,$F9,$5A,$AF,$9D,$7F,$87
	.dc.b	$CE,$78,$6A,$33,$39,$FA,$3E,$34,$7F,$99,$62,$F3,$B3,$12,$7C,$67
	.dc.b	$F2,$13,$81,$F0,$A3,$55,$FA,$A8,$CB,$67,$AC,$EB,$2F,$99,$2F,$A2
	.dc.b	$7D,$17,$B1,$B8,$7C,$EF,$A3,$C4,$C8,$F1,$D7,$F8,$F6,$B6,$FE,$F7
	.dc.b	$25,$EE,$3F,$7B,$3A,$C3,$C2,$DB,$F9,$17,$C1,$EB,$5E,$5C,$3E,$7F
	.dc.b	$C6,$7F,$65,$9F,$E8,$4D,$9F,$C5,$D3,$C3,$DF,$A1,$2F,$9D,$ED,$FB
	.dc.b	$F3,$9B,$FE,$23,$3C,$D3,$BB,$E9,$5C,$AF,$F0,$3E,$CF,$DC,$BF,$F5
	.dc.b	$EF,$0F,$E1,$B1,$7D,$1B,$E1,$F7,$33,$1F,$6E,$1F,$84,$BD,$D6,$70
	.dc.b	$FE,$47,$3D,$4D,$61,$CE,$F8,$E1,$E2,$F9,$A9,$1F,$53,$17,$D1,$3E
	.dc.b	$8F,$39,$74,$BB,$C3,$FA,$DB,$DD,$63,$0F,$ED,$D9,$BE,$8B,$DE,$7E
	.dc.b	$83,$1F,$F3,$E7,$D1,$FF,$9F,$7E,$8B,$FC,$5F,$1F,$EF,$29,$DD,$FB
	.dc.b	$9E,$10,$AE,$1E,$E7,$69,$CC,$62,$79,$55,$7F,$C0,$11,$70,$F2,$4E
	.dc.b	$67,$C6,$30,$F1,$BF,$EF,$0B,$FE,$15,$18,$8D,$BA,$AE,$71,$C7,$0F
	.dc.b	$F1,$BE,$1D,$EF,$8C,$62,$74,$79,$8A,$79,$8B,$5D,$2F,$C1,$1F,$04
	.dc.b	$CD,$74,$7F,$D8,$17,$DF,$0F,$64,$6E,$BC,$D5,$B8,$ED,$18,$9D,$D7
	.dc.b	$19,$F1,$9E,$EC,$3E,$74,$EA,$3C,$2D,$7F,$C6,$13,$FE,$31,$7E,$59
	.dc.b	$83,$FC,$B3,$F4,$7F,$C5,$FE,$3F,$FC,$6C,$D8,$CF,$52,$72,$D8,$E3
	.dc.b	$F4,$7F,$E4,$56,$EF,$CA,$31,$1F,$EE,$1B,$46,$23,$FC,$81,$18,$9F
	.dc.b	$C2,$7D,$1B,$D0,$FF,$22,$6B,$F8,$08,$C5,$FE,$26,$F8,$3F,$59,$B3
	.dc.b	$BA,$E3,$AF,$F9,$36,$F8,$DE,$F1,$F4,$5F,$E3,$30,$E3,$7B,$C3,$8D
	.dc.b	$FF,$B7,$F0,$E3,$7B,$C3,$8D,$FA,$9B,$5B,$BE,$2A,$F8,$D6,$F1,$F5
	.dc.b	$5F,$0A,$D5,$F0,$BD,$C7,$FB,$BF,$1F,$2B,$6B,$84,$EA,$38,$9D,$2E
	.dc.b	$3F,$9A,$8D,$1B,$AE,$71,$C0,$FF,$1B,$C7,$C6,$9F,$E9,$6D,$D7,$D1
	.dc.b	$3C,$F0,$FF,$3C,$56,$BD,$1B,$5F,$3D,$B5,$F0,$CF,$E8,$3F,$47,$FE
	.dc.b	$B1,$7F,$A1,$FD,$1B,$E3,$37,$08,$C3,$C3,$AF,$35,$C6,$3F,$C9,$B3
	.dc.b	$A6,$D6,$AB,$7F,$43,$D2,$9F,$0B,$4C,$DF,$92,$70,$F9,$23,$2F,$F9
	.dc.b	$6B,$04,$EA,$3C,$33,$AC,$FF,$B8,$FA,$8E,$AF,$07,$B3,$2F,$FB,$0C
	.dc.b	$E5,$B3,$97,$D7,$3F,$0A,$75,$1F,$E6,$A6,$D7,$F7,$EC,$BF,$D2,$33
	.dc.b	$F2,$AB,$85,$FC,$AF,$69,$EB,$FC,$1F,$8B,$97,$FD,$CA,$72,$D9,$CB
	.dc.b	$F3,$27,$94,$48,$FA,$F1,$AC,$7F,$C1,$F5,$CA,$7F,$CF,$FF,$23,$FE
	.dc.b	$CE,$7F,$BF,$F1,$FF,$67,$6F,$87,$F7,$A7,$51,$C3,$D5,$FC,$0D,$6E
	.dc.b	$3F,$D1,$2D,$C3,$FC,$3B,$5C,$1B,$5C,$2E,$5F,$F8,$3A,$B5,$CB,$FE
	.dc.b	$CD,$E5,$FB,$33,$FE,$12,$FA,$37,$D1,$FE,$41,$B1,$FE,$16,$9D,$BF
	.dc.b	$E9,$A6,$E5,$FE,$2E,$9D,$CF,$FA,$79,$B1,$FE,$32,$9D,$DF,$87,$93
	.dc.b	$B6,$13,$0E,$D1,$FA,$0F,$A3,$7C,$15,$24,$FF,$8A,$3F,$07,$FF,$18
	.dc.b	$BF,$E2,$AF,$A4,$71,$7E,$A1,$B1,$98,$E2,$FF,$AC,$BE,$71,$F4,$9F
	.dc.b	$AE,$6C,$7F,$8D,$1A,$FF,$AE,$2B,$9B,$FE,$BA,$FA,$47,$D2,$7E,$D9
	.dc.b	$B1,$FE,$3C,$6B,$BF,$8D,$FD,$DF,$D1,$BE,$92,$92,$7F,$C8,$7F,$4C
	.dc.b	$FA,$AF,$27,$64,$9F,$23,$FC,$13,$E9,$FF,$C3,$11,$FE,$4A,$F9,$57
	.dc.b	$BF,$FD,$75,$B0,$21,$1C,$B8,$34,$E1,$C7,$D3,$FF,$06,$74,$73,$70
	.dc.b	$E1,$AF,$0F,$17,$FB,$77,$D3,$BE,$3B,$FC,$9F,$3C,$A3,$1E,$DF,$48
	.dc.b	$43,$E7,$FF,$CB,$D1,$BA,$E2,$FF,$A7,$7D,$37,$D1,$7F,$A6,$3E,$0B
	.dc.b	$FB,$A7,$9D,$67,$3E,$0F,$F9,$EA,$72,$D8,$CD,$BC,$3F,$F5,$95,$6A
	.dc.b	$31,$5F,$EB,$7A,$DD,$69,$FF,$3A,$57,$F8,$0F,$AF,$C1,$F1,$5E,$16
	.dc.b	$DF,$76,$E1,$D0,$FF,$56,$5A,$33,$87,$FA,$63,$E9,$7F,$CC,$55,$BF
	.dc.b	$71,$F4,$CF,$92,$FF,$5A,$D6,$AF,$FC,$BD,$39,$CB,$FD,$1B,$39,$7F
	.dc.b	$05,$7C,$AF,$8F,$97,$FE,$7B,$BE,$18,$F9,$46,$1E,$27,$F0,$CD,$7F
	.dc.b	$16,$D7,$F1,$CD,$D7,$FC,$4D,$F4,$DF,$4E,$FE,$A3,$72,$FA,$4D,$BF
	.dc.b	$89,$5C,$27,$EB,$57,$DB,$F9,$99,$1F,$67,$E6,$AD,$F0,$DD,$58,$DB
	.dc.b	$EB,$55,$BF,$D0,$FD,$1B,$E9,$7F,$DB,$F4,$FA,$6F,$BB,$5A,$9C,$C6
	.dc.b	$32,$FD,$55,$F1,$AF,$F6,$2D,$F1,$E5,$FF,$73,$CE,$6B,$EF,$4E,$6B
	.dc.b	$FD,$7F,$39,$8F,$F7,$DD,$F5,$A4,$E6,$BF,$DE,$B3,$E1,$D7,$FC,$A1
	.dc.b	$F4,$DF,$9D,$3F,$3D,$D1,$6B,$F3,$91,$70,$F8,$93,$71,$FC,$C5,$7E
	.dc.b	$97,$E5,$BF,$4C,$D3,$F4,$85,$AF,$98,$D9,$FF,$75,$E5,$FE,$15,$8F
	.dc.b	$F5,$97,$77,$FB,$2E,$31,$5F,$97,$8C,$7B,$9F,$EC,$76,$D7,$0E,$AF
	.dc.b	$8F,$39,$6C,$64,$FD,$C5,$3E,$9B,$F7,$CD,$8C,$F9,$B8,$99,$1C,$1D
	.dc.b	$C6,$27,$F1,$A3,$E9,$B9,$FC,$57,$46,$DF,$81,$B3,$C1,$BA,$27,$F9
	.dc.b	$5F,$D1,$FF,$56,$7F,$97,$46,$69,$F1,$2F,$A6,$FF,$08,$D7,$FC,$00
	.dc.b	$BE,$15,$FC,$0E,$B5,$E4,$F1,$C3,$94,$AB,$71,$D6,$B7,$7C,$FD,$4E
	.dc.b	$0F,$F3,$F8,$CD,$FE,$EE,$F8,$CE,$71,$F1,$CE,$6F,$8B,$E8,$8C,$E1
	.dc.b	$E2,$BE,$33,$9C,$76,$F9,$6F,$8E,$3E,$0B,$E1,$18,$C3,$9E,$3F,$22
	.dc.b	$31,$87,$2C,$7A,$E1,$C6,$B4,$7C,$C3,$D1,$3B,$AE,$31,$F0,$4D,$63
	.dc.b	$E5,$70,$FC,$DC,$3C,$2F,$F6,$3C,$3B,$C6,$23,$D5,$3C,$B0,$F1,$46
	.dc.b	$EF,$BB,$F1,$E7,$55,$C6,$F9,$E1,$E5,$9C,$B6,$BC,$4D,$C7,$ED,$5F
	.dc.b	$1C,$3A,$C7,$C8,$BE,$27,$A9,$DD,$71,$8F,$86,$64,$B7,$E5,$3B,$B9
	.dc.b	$7E,$21,$99,$E7,$19,$C7,$DF,$3F,$08,$E2,$6E,$3A,$CF,$96,$32,$D3
	.dc.b	$9B,$B3,$BB,$98,$E2,$7D,$C8,$E2,$D8,$E2,$6E,$38,$C6,$4E,$2F,$D3
	.dc.b	$7E,$2B,$DE,$7D,$38,$FD,$29,$F6,$63,$F3,$5D,$DF,$5A,$DE,$3F,$42
	.dc.b	$7B,$61,$E0,$3B,$1E,$62,$D7,$9A,$37,$5C,$9E,$A6,$2F,$C1,$5C,$4F
	.dc.b	$99,$5C,$5E,$73,$EC,$72,$DC,$3D,$51,$ED,$AE,$75,$C6,$36,$79,$CB
	.dc.b	$7E,$37,$C1,$5C,$A7,$E0,$57,$2B,$ED,$87,$C1,$BF,$19,$CE,$F8,$4F
	.dc.b	$AF,$1F,$C8,$4E,$6B,$83,$E0,$BE,$8E,$6B,$E1,$BD,$31,$F2,$63,$E9
	.dc.b	$C7,$AC,$E6,$F8,$D6,$AF,$CA,$72,$9E,$D8,$F8,$A7,$51,$DC,$F7,$E3
	.dc.b	$E4,$C3,$AC,$70,$BF,$C1,$DF,$1C,$3B,$91,$C1,$23,$64,$9F,$A5,$1F
	.dc.b	$D4,$BE,$4F,$FA,$9F,$91,$6B,$FA,$BB,$6B,$FA,$7B,$73,$FF,$26,$C7
	.dc.b	$FB,$C2,$7F,$AE,$FC,$3F,$F5,$C3,$FA,$EF,$C9,$D8,$C9,$27,$FA,$CC
	.dc.b	$65,$FF,$1B,$7D,$1B,$8F,$FB,$77,$D2,$7F,$34,$8C,$5F,$F7,$68,$C4
	.dc.b	$EA,$3F,$93,$D2,$B6,$FF,$79,$AD,$CF,$13,$73,$AC,$7F,$AF,$78,$EE
	.dc.b	$75,$FF,$76,$C3,$FB,$BC,$DC,$3F,$BD,$B6,$3F,$B9,$B6,$FE,$2B,$FD
	.dc.b	$4A,$5F,$44,$F9,$FC,$36,$3F,$9F,$9A,$F8,$2C,$CC,$8C,$CE,$F0,$FF
	.dc.b	$38,$46,$CC,$CF,$C7,$F8,$4F,$F0,$B7,$89,$67,$FC,$3A,$DA,$FF,$0E
	.dc.b	$37,$2F,$83,$A2,$35,$3F,$EA,$BD,$58,$7F,$7A,$8F,$F5,$8C,$63,$0F
	.dc.b	$F5,$03,$63,$FD,$65,$1F,$03,$E9,$3D,$FF,$49,$F0,$1B,$1E,$C6,$CF
	.dc.b	$B3,$44,$63,$3F,$F0,$95,$7C,$AF,$A4,$F9,$27,$CA,$E9,$B1,$F1,$A2
	.dc.b	$E1,$CE,$6D,$7F,$84,$BE,$91,$F1,$5F,$E3,$BD,$7F,$C0,$31,$8C,$7F
	.dc.b	$C8,$71,$89,$D4,$7E,$3A,$75,$87,$F9,$16,$B5,$3F,$EC,$19,$CC,$FF
	.dc.b	$B4,$DA,$FF,$B6,$1A,$FF,$8C,$73,$7C,$37,$E1,$3E,$27,$F0,$AD,$BF
	.dc.b	$C0,$B6,$BE,$33,$72,$F8,$B7,$FE,$A0,$6F,$A5,$F8,$EE,$CF,$C7,$B3
	.dc.b	$0F,$C5,$B9,$69,$FA,$AF,$A3,$73,$7F,$8A,$BE,$91,$F1,$DF,$91,$C7
	.dc.b	$FC,$77,$5A,$BF,$F0,$BD,$6B,$A3,$FC,$57,$39,$D7,$FD,$8D,$39,$AE
	.dc.b	$D8,$F5,$C7,$A4,$7F,$BF,$2F,$86,$3D,$AB,$FC,$E1,$E4,$F4,$7F,$C6
	.dc.b	$FF,$45,$AB,$65,$FE,$C4,$FA,$CC,$D7,$F8,$F3,$B1,$1F,$E3,$EF,$A5
	.dc.b	$7D,$26,$B0,$FE,$13,$18,$7F,$85,$E3,$D7,$1E,$BF,$4A,$F7,$39,$61
	.dc.b	$FE,$48,$8C,$5F,$28,$F2,$F2,$C8,$FE,$2B,$F4,$7D,$70,$FF,$6A,$63
	.dc.b	$F8,$09,$CD,$EE,$7F,$DB,$15,$BD,$FF,$CD,$37,$B9,$CD,$7F,$96,$2B
	.dc.b	$57,$F0,$BE,$91,$F4,$AF,$84,$D5,$FF,$97,$BD,$67,$A1,$4F,$E2,$DF
	.dc.b	$44,$F3,$71,$97,$F2,$CA,$DB,$6B,$73,$AD,$3F,$DD,$D5,$A6,$CE,$AB
	.dc.b	$81,$FC,$D8,$7C,$EF,$F9,$C1,$A7,$F2,$2C,$1E,$66,$66,$4F,$17,$CF
	.dc.b	$18,$BF,$F7,$F4,$F2,$6C,$EA,$38,$1F,$D1,$BE,$91,$F4,$9F,$E8,$06
	.dc.b	$C7,$87,$E9,$71,$32,$38,$3F,$1A,$31,$6E,$3F,$F4,$ED,$F2,$8C,$5F
	.dc.b	$E5,$14,$66,$7F,$8D,$E4,$BE,$71,$FE,$A6,$F8,$8F,$EA,$98,$71,$6E
	.dc.b	$1F,$EB,$7B,$75,$7F,$A6,$27,$32,$69,$C2,$95,$C1,$FE,$CF,$1A,$F4
	.dc.b	$39,$1F,$82,$BB,$3F,$DB,$3D,$5F,$B9,$8F,$FA,$E9,$B1,$FB,$09,$7D
	.dc.b	$2B,$E0,$FF,$D5,$75,$AC,$FF,$D7,$5C,$38,$3F,$D6,$B7,$C2,$B1,$87
	.dc.b	$79,$B5,$F2,$8F,$7E,$1D,$67,$47,$97,$E6,$FB,$9D,$3E,$95,$F4,$B8
	.dc.b	$9D,$12,$FA,$64,$AD,$CE,$74,$F7,$7D,$2F,$A8,$FF,$62,$DF,$F0,$EF
	.dc.b	$A5,$FE,$1F,$F4,$AF,$6B,$EF,$70,$2F,$86,$3F,$25,$E3,$83,$0E,$37
	.dc.b	$C2,$73,$18,$9F,$F7,$04,$7F,$BC,$70,$ED,$1F,$8B,$39,$07,$2F,$FB
	.dc.b	$DA,$38,$56,$E3,$73,$AC,$3C,$D8,$77,$9F,$CB,$61,$67,$31,$D6,$F9
	.dc.b	$4E,$70,$EB,$3E,$38,$C5,$F3,$8F,$8C,$71,$3D,$6E,$5F,$1D,$7C,$18
	.dc.b	$C4,$F2,$9E,$07,$90,$B7,$E4,$C3,$DF,$1D,$9F,$66,$1E,$48,$E1,$18
	.dc.b	$BE,$B1,$CA,$74,$F1,$C3,$C4,$F8,$0F,$49,$F2,$4F,$9E,$5C,$30,$E8
	.dc.b	$FD,$F7,$D2,$7F,$1C,$F8,$AE,$0D,$8F,$A4,$D7,$ED,$35,$E3,$F4,$BF
	.dc.b	$C2,$DD,$B5,$DB,$5D,$E9,$DA,$38,$67,$F8,$2F,$A5,$7A,$7F,$C4,$63
	.dc.b	$0D,$CF,$8E,$1F,$3B,$16,$1C,$A7,$F8,$C6,$1C,$AB,$F0,$CD,$78,$4E
	.dc.b	$DF,$EE,$35,$CB,$1D,$E1,$DE,$7F,$BA,$DF,$1C,$7F,$2F,$7B,$9E,$77
	.dc.b	$C1,$86,$09,$3E,$A7,$2F,$2C,$3C,$50,$AE,$93,$B8,$FC,$5D,$F8,$71
	.dc.b	$F8,$33,$C7,$0E,$71,$F0,$31,$F4,$5F,$3C,$7A,$E1,$F0,$6D,$AE,$EB
	.dc.b	$FD,$23,$39,$8E,$B8,$66,$7F,$8D,$09,$FF,$1F,$BC,$3C,$6F,$E8,$D7
	.dc.b	$D2,$3E,$3E,$38,$C3,$FA,$57,$D2,$FF,$A9,$67,$FD,$3F,$F4,$AF,$87
	.dc.b	$D4,$7F,$88,$A3,$0F,$FA,$92,$78,$3E,$26,$E1,$DE,$13,$C1,$F3,$B6
	.dc.b	$35,$3C,$2F,$C7,$F4,$5E,$0B,$E3,$F0,$77,$03,$9B,$55,$35,$12,$BC
	.dc.b	$D1,$97,$DB,$19,$7D,$11,$DF,$D3,$74,$F7,$F6,$E5,$7F,$B0,$A7,$2D
	.dc.b	$9C,$BE,$C9,$E5,$12,$3E,$5C,$6B,$3F,$89,$5C,$A7,$EC,$FB,$6F,$67
	.dc.b	$DD,$F0,$F2,$7F,$DA,$33,$96,$CE,$5F,$97,$3F,$1E,$75,$1D,$27,$57
	.dc.b	$F4,$F5,$5E,$E7,$E7,$7D,$2F,$F9,$D2,$FF,$DB,$FC,$3F,$DB,$DB,$E0
	.dc.b	$EB,$D6,$91,$82,$1F,$92,$AE,$4D,$8F,$B5,$5C,$E3,$F2,$ED,$CB,$F0
	.dc.b	$D5,$C1,$FC,$47,$27,$DC,$AF,$F7,$8F,$D2,$BE,$97,$F0,$8D,$8F,$F7
	.dc.b	$AC,$ED,$FD,$1B,$6B,$F2,$53,$B9,$FD,$3B,$63,$BF,$17,$E3,$EB,$87
	.dc.b	$2F,$F8,$B6,$B4,$DC,$35,$5F,$95,$FA,$57,$C0,$D2,$65,$FE,$88,$CB
	.dc.b	$F4,$5F,$4B,$FE,$33,$8C,$63,$FE,$39,$FA,$4F,$A7,$18,$3F,$D1,$EE
	.dc.b	$1F,$DE,$B6,$BF,$59,$87,$82,$78,$0F,$99,$7D,$2F,$EA,$5B,$1A,$6B
	.dc.b	$D3,$93,$F5,$DF,$48,$FA,$5A,$49,$EB,$0F,$A5,$FD,$A7,$D2,$BE,$96
	.dc.b	$92,$BC,$7F,$4A,$FA,$5F,$17,$5C,$9F,$13,$FB,$DF,$A5,$FF,$0E,$47
	.dc.b	$AB,$C2,$7C,$F7,$F0,$15,$F7,$14,$C4,$49,$E1,$19,$79,$46,$5F,$E4
	.dc.b	$B3,$C9,$D9,$A3,$E4,$FD,$2B,$E9,$7F,$CB,$D1,$FE,$1D,$8C,$7B,$1F
	.dc.b	$DD,$D7,$CF,$7F,$99,$A7,$55,$C1,$FF,$15,$7D,$4F,$FA,$CD,$BF,$07
	.dc.b	$FC,$8D,$FF,$3C,$46,$5B,$19,$5F,$04,$F4,$DF,$45,$FC,$E7,$B3,$F9
	.dc.b	$D4,$E7,$DD,$C7,$7F,$FB,$5A,$38,$56,$30,$E7,$87,$F9,$FA,$73,$5F
	.dc.b	$8F,$79,$36,$3B,$B6,$F9,$4F,$8F,$B2,$D7,$FB,$52,$FF,$CD,$53,$9B
	.dc.b	$E5,$F1,$7A,$BD,$B6,$32,$77,$8C,$B4,$CD,$7F,$5C,$FA,$5F,$F3,$A3
	.dc.b	$FE,$4F,$6B,$FE,$57,$6B,$FE,$A3,$6B,$FE,$F8,$78,$E7,$FD,$4D,$B3
	.dc.b	$FE,$FC,$F6,$9E,$4F,$FB,$86,$B5,$E6,$7B,$BE,$97,$DE,$DB,$F8,$CD
	.dc.b	$C7,$E0,$EB,$D1,$F9,$F5,$C2,$BE,$67,$D2,$FC,$83,$E4,$FC,$E5,$8C
	.dc.b	$92,$B9,$C6,$5E,$B3,$71,$FC,$0E,$0A,$C5,$FF,$B0,$FE,$8F,$F5,$2F
	.dc.b	$F9,$35,$B8,$7F,$9D,$DA,$FF,$9F,$DA,$F1,$EA,$FC,$42,$8E,$38,$7F
	.dc.b	$82,$75,$F8,$FF,$3B,$A9,$CE,$1F,$91,$C9,$F4,$9F,$86,$3F,$0F,$D3
	.dc.b	$67,$F1,$D1,$72,$F9,$33,$72,$F6,$D7,$E3,$BE,$57,$F1,$ED,$3F,$2C
	.dc.b	$5A,$FA,$2D,$9E,$2F,$E6,$E7,$8B,$4F,$E4,$F8,$FF,$87,$5B,$ED,$7F
	.dc.b	$B2,$3E,$7F,$F4,$38,$7F,$8B,$2B,$4D,$9F,$80,$6F,$A6,$DF,$EA,$E5
	.dc.b	$F4,$9F,$A8,$6C,$67,$CB,$C4,$C8,$E0,$EE,$31,$3F,$AC,$1F,$49,$CF
	.dc.b	$E2,$7A,$36,$FC,$0D,$9E,$0D,$D7,$F8,$3C,$72,$9F,$35,$72,$69,$C8
	.dc.b	$B1,$C9,$B3,$CA,$BA,$47,$F0,$F9,$E8,$79,$EB,$A3,$63,$A3,$4E,$85
	.dc.b	$AE,$91,$D8,$E9,$5D,$A7,$F9,$19,$DA,$7D,$75,$D9,$B5,$D9,$B3,$D9
	.dc.b	$A7,$67,$C1,$1D,$8F,$03,$DA,$BC,$1D,$88,$F0,$4F,$BE,$FC,$0D,$3C
	.dc.b	$05,$8F,$03,$6F,$C0,$78,$9F,$04,$78,$8F,$03,$E2,$AF,$05,$F8,$A7
	.dc.b	$C2,$7C,$2B,$F0,$B6,$BC,$2D,$3C,$25,$BF,$0C,$78,$CF,$0B,$E3,$8F
	.dc.b	$09,$E3,$BF,$0C,$F8,$E3,$C3,$1F,$0E,$BC,$2D,$9F,$0B,$67,$C2,$DA
	.dc.b	$F0,$C7,$8E,$7C,$27,$8D,$F0,$C7,$8E,$7C,$31,$E3,$3C,$22,$BC,$2D
	.dc.b	$AF,$0B,$4F,$09,$6B,$C3,$5E,$39,$F0,$CF,$8C,$F0,$BE,$38,$F0,$CF
	.dc.b	$8E,$7C,$35,$E0,$25,$78,$1B,$1E,$06,$CF,$82,$3C,$53,$E0,$AF,$14
	.dc.b	$78,$0F,$13,$E0,$8F,$14,$78,$1E,$C4,$7B,$36,$7B,$57,$82,$3B,$4F
	.dc.b	$80,$ED,$5E,$03,$B3,$E0,$3B,$1D,$09,$5D,$27,$B1,$D2,$BB,$4F,$43
	.dc.b	$B5,$74,$3B,$3D,$2B,$93,$D2,$B9,$1D,$1E,$55,$D2,$B9,$47,$43,$94
	.dc.b	$F0,$79,$1C,$1E,$51,$C2,$79,$4F,$03,$F7,$5F,$49,$FE,$76,$AF,$F3
	.dc.b	$3B,$61,$7F,$EE,$5F,$A9,$7D,$4B,$E5,$3F,$B1,$DE,$EB,$FD,$B8,$DC
	.dc.b	$3F,$83,$60,$8C,$2C,$39,$DF,$2F,$5B,$FA,$27,$CA,$7F,$45,$F2,$6E
	.dc.b	$3F,$D2,$5B,$87,$F4,$16,$E5,$FC,$E9,$FE,$99,$F0,$B2,$BF,$A8,$7C
	.dc.b	$A5,$BF,$E9,$13,$6F,$FA,$0C,$DA,$FE,$71,$36,$BF,$96,$CD,$5D,$6F
	.dc.b	$1B,$FC,$A9,$18,$6C,$66,$FE,$0C,$FC,$28,$C3,$83,$84,$7F,$A5,$A3
	.dc.b	$38,$23,$F9,$77,$C6,$BA,$7F,$D1,$D1,$BC,$3F,$B5,$61,$FD,$AE,$AC
	.dc.b	$FF,$6F,$6C,$7F,$67,$6D,$FF,$60,$7F,$CC,$52,$FA,$07,$CF,$E1,$B1
	.dc.b	$FE,$B1,$35,$F0,$59,$99,$19,$9D,$E1,$F8,$88,$FE,$8C,$70,$35,$3C
	.dc.b	$23,$47,$BE,$78,$E1,$F0,$63,$17,$B8,$C4,$7D,$2F,$84,$F9,$E4,$C3
	.dc.b	$E6,$46,$5B,$A2,$7F,$C7,$2D,$8F,$F6,$66,$5F,$37,$C5,$CD,$7C,$1F
	.dc.b	$A3,$F8,$5F,$49,$6F,$DC,$DC,$7D,$31,$F2,$63,$19,$7A,$63,$07,$C5
	.dc.b	$FA,$4F,$98,$7C,$DE,$5B,$19,$25,$F4,$DB,$F4,$F1,$F5,$6B,$73,$97
	.dc.b	$94,$70,$BF,$8B,$F4,$8F,$8A,$FC,$AB,$67,$E2,$B6,$FF,$24,$D7,$F2
	.dc.b	$EE,$6B,$9D,$7F,$90,$7D,$3F,$F3,$0C,$E6,$FE,$EE,$3F,$E8,$56,$FC
	.dc.b	$47,$E2,$7E,$23,$F1,$4D,$8F,$C3,$37,$2F,$73,$F8,$FF,$A4,$93,$F8
	.dc.b	$BF,$6E,$D7,$C3,$9B,$A7,$CD,$C5,$58,$BF,$1D,$5B,$FC,$64,$D5,$F3
	.dc.b	$EE,$6F,$F3,$DF,$2B,$E8,$BF,$CF,$B3,$9C,$7F,$C3,$D9,$FE,$42,$B5
	.dc.b	$18,$3F,$8E,$FD,$13,$D9,$E3,$8F,$FA,$1A,$31,$3B,$CB,$FC,$51,$8F
	.dc.b	$57,$F4,$23,$E8,$BF,$D9,$19,$7E,$7A,$F8,$CE,$A7,$37,$E0,$C7,$CF
	.dc.b	$C1,$23,$B5,$3E,$8B,$8E,$3F,$A8,$CB,$EB,$4F,$F8,$E6,$73,$5F,$CD
	.dc.b	$A7,$33,$C7,$0E,$8F,$9A,$B6,$F1,$8F,$F7,$57,$D1,$3E,$8B,$FA,$26
	.dc.b	$1E,$2B,$DD,$FF,$BE,$3C,$9F,$F5,$4F,$95,$FD,$C6,$9F,$3C,$F1,$F3
	.dc.b	$52,$7D,$75,$BA,$D4,$62,$75,$7F,$2A,$B5,$78,$D7,$F8,$3D,$BE,$8B
	.dc.b	$FD,$72,$D3,$F8,$E6,$2F,$1B,$33,$27,$E5,$9B,$C7,$F7,$F5,$B8,$CC
	.dc.b	$EE,$7E,$DC,$6E,$FF,$81,$56,$E7,$35,$FE,$CA,$FA,$17,$D1,$53,$F9
	.dc.b	$9D,$3E,$8B,$18,$FD,$CA,$DB,$6B,$73,$AC,$3F,$17,$5A,$6C,$EA,$B8
	.dc.b	$1F,$86,$F8,$5F,$E9,$EF,$F7,$66,$CF,$F5,$1C,$1D,$1F,$EE,$09,$CC
	.dc.b	$9A,$70,$AF,$B7,$3B,$F2,$64,$77,$7D,$B1,$8A,$FD,$0F,$B7,$FE,$33
	.dc.b	$4F,$F8,$D6,$1F,$44,$F8,$0F,$F7,$8C,$EC,$9C,$7F,$D1,$F8,$3F,$A4
	.dc.b	$5F,$0A,$DC,$CC,$BF,$4B,$7D,$70,$E9,$18,$9E,$BF,$33,$23,$FA,$2F
	.dc.b	$D0,$BE,$8B,$1A,$7A,$2A,$E9,$E0,$AD,$5F,$69,$F3,$FD,$14,$8F,$DA
	.dc.b	$DF,$FA,$A7,$E8,$BF,$D5,$5F,$42,$F5,$3F,$C4,$FB,$30,$E1,$7F,$19
	.dc.b	$F7,$61,$FE,$09,$C3,$8D,$F0,$9C,$D7,$F8,$42,$72,$FF,$AD,$EF,$A5
	.dc.b	$E2,$3F,$C6,$B3,$67,$30,$C3,$C8,$F3,$31,$87,$7B,$E5,$1E,$B3,$27
	.dc.b	$7C,$3F,$BD,$57,$28,$EE,$68,$FB,$F1,$AA,$F6,$46,$6F,$C7,$5C,$63
	.dc.b	$B3,$CE,$33,$3C,$23,$15,$CE,$BD,$11,$8B,$F3,$C3,$0F,$81,$5B,$8F
	.dc.b	$2D,$6D,$F3,$1E,$0B,$F8,$0E,$0E,$47,$9E,$79,$3B,$3E,$15,$F8,$2F
	.dc.b	$FC,$2B,$1F,$1C,$FF,$0B,$B8,$79,$9F,$E4,$8F,$A2,$FE,$31,$F0,$9F
	.dc.b	$E1,$C6,$C7,$F9,$25,$AE,$FA,$53,$BF,$A2,$FF,$3C,$CE,$DA,$F5,$6B
	.dc.b	$FE,$76,$A4,$62,$3F,$DF,$D1,$8D,$FF,$11,$F4,$4F,$33,$FC,$FB,$B7
	.dc.b	$DD,$8C,$5F,$E3,$6F,$83,$F7,$DB,$3B,$AE,$3B,$7F,$A1,$AF,$8D,$EF
	.dc.b	$1F,$4D,$FE,$27,$0E,$37,$BC,$38,$D8,$61,$CA,$F8,$61,$CA,$FD,$CD
	.dc.b	$AE,$15,$B9,$E9,$1E,$99,$D2,$C3,$C2,$F0,$C3,$C2,$F8,$1B,$1A,$8E
	.dc.b	$13,$CA,$FB,$47,$13,$B6,$31,$FD,$24,$69,$E8,$67,$1F,$90,$FF,$7E
	.dc.b	$8F,$8F,$3F,$E6,$46,$EB,$EE,$8F,$AF,$8F,$FA,$C9,$D7,$8B,$67,$E8
	.dc.b	$36,$FC,$87,$F8,$53,$E8,$BF,$D3,$CF,$F8,$33,$E8,$5F,$09,$B8,$46
	.dc.b	$1F,$F0,$8C,$F0,$79,$75,$EE,$75,$1F,$E2,$06,$E3,$FE,$7C,$AE,$0D
	.dc.b	$F9,$FE,$1E,$EC,$BE,$B1,$E1,$A3,$98,$E9,$39,$89,$1E,$98,$D5,$75
	.dc.b	$8F,$F1,$3F,$D1,$3A,$3C,$DE,$AC,$9C,$C7,$BE,$75,$1F,$0E,$74,$FF
	.dc.b	$8E,$A3,$2D,$9C,$D7,$FA,$63,$2F,$F3,$A4,$F4,$3F,$C6,$FE,$D7,$F9
	.dc.b	$69,$F9,$DF,$09,$F1,$CF,$9D,$19,$8F,$F2,$3B,$6B,$A4,$EE,$79,$FA
	.dc.b	$7C,$0C,$3F,$3E,$30,$FF,$91,$BE,$3B,$FC,$C4,$FF,$92,$38,$FF,$B7
	.dc.b	$37,$C2,$FB,$D3,$A8,$FF,$2B,$B7,$0F,$F5,$9D,$70,$9F,$F2,$DB,$63
	.dc.b	$FD,$6F,$3B,$E4,$FD,$B5,$23,$0F,$EE,$6B,$84,$E9,$B1,$FB,$D9,$DA
	.dc.b	$E4,$FC,$0B,$FE,$60,$FA,$27,$D1,$7E,$55,$B1,$FE,$64,$9D,$BF,$E7
	.dc.b	$06,$E3,$FE,$CC,$9D,$CF,$F9,$D1,$B1,$FE,$D0,$9D,$E3,$EA,$6D,$7F
	.dc.b	$11,$AD,$CF,$B9,$B1,$FC,$56,$76,$BE,$81,$EE,$FE,$AD,$B1,$FE,$7A
	.dc.b	$6B,$FE,$86,$FA,$2F,$F2,$23,$FE,$7E,$FA,$26,$FF,$B3,$6C,$66,$3A
	.dc.b	$5F,$ED,$FE,$55,$F4,$54,$95,$B3,$FD,$13,$F4,$2F,$A2,$FD,$F3,$63
	.dc.b	$31,$8F,$A2,$FE,$05,$F4,$4F,$A2,$A4,$AF,$F4,$9F,$D1,$3E,$8B,$C5
	.dc.b	$D3,$27,$FD,$2F,$A7,$E6,$F9,$3F,$0B,$F4,$5F,$E2,$19,$FF,$4D,$F8
	.dc.b	$0F,$9B,$FE,$20,$BE,$8A,$98,$89,$3C,$23,$2F,$28,$CB,$FE,$B0,$9E
	.dc.b	$4E,$CD,$1F,$EA,$AF,$A9,$47,$F7,$86,$C7,$E6,$8F,$CD,$B8,$70,$1C
	.dc.b	$3B,$FA,$37,$C6,$7F,$9A,$63,$71,$8F,$67,$8C,$21,$F3,$BF,$E7,$28
	.dc.b	$DD,$71,$7F,$57,$F0,$3F,$E0,$0F,$9B,$FF,$20,$BF,$E0,$2F,$32,$FA
	.dc.b	$69,$B1,$86,$F0,$7F,$B5,$6B,$93,$6B,$FD,$07,$1F,$E0,$CE,$1E,$7F
	.dc.b	$45,$D1,$B5,$B6,$E2,$F0,$D3,$FD,$D6,$F9,$3D,$6F,$93,$4E,$45,$AF
	.dc.b	$F7,$2F,$D1,$7F,$9D,$9E,$91,$FD,$FF,$DE,$7D,$4B,$E3,$FF,$CC,$34
	.dc.b	$BE,$19,$7F,$A1,$2B,$58,$7F,$B9,$A7,$A5,$7D,$B6,$BF,$C5,$9A,$FF
	.dc.b	$83,$5B,$97,$F8,$BB,$E8,$FE,$6F,$99,$F3,$9B,$7F,$25,$B7,$E8,$15
	.dc.b	$C6,$7E,$75,$7D,$6F,$96,$91,$F4,$7E,$8E,$DF,$1B,$FF,$1E,$D6,$33
	.dc.b	$F9,$B3,$D3,$81,$79,$C3,$E8,$42,$F1,$5B,$AA,$BE,$87,$F5,$D7,$FE
	.dc.b	$2B,$C3,$FD,$59,$F5,$3F,$89,$8C,$1F,$CE,$A7,$75,$ED,$A5,$6E,$3E
	.dc.b	$01,$D7,$AF,$FD,$2B,$3A,$E4,$FA,$F7,$CA,$3B,$C7,$E6,$3E,$92,$63
	.dc.b	$F9,$3E,$8B,$8F,$C3,$D5,$58,$C3,$E1,$ED,$FE,$62,$BC,$61,$F0,$8F
	.dc.b	$D1,$FD,$27,$E9,$1A,$7E,$7C,$B8,$F5,$6D,$F4,$C7,$F3,$71,$FA,$5A
	.dc.b	$E4,$D3,$E2,$D7,$F9,$4F,$3F,$F1,$3C,$63,$95,$F3,$DF,$EE,$A9,$CD
	.dc.b	$7E,$16,$73,$D3,$F1,$3D,$BF,$F5,$DC,$63,$0E,$B5,$F5,$E7,$38,$7E
	.dc.b	$DA,$1F,$49,$FC,$09,$B5,$FB,$D7,$FA,$3B,$63,$44,$75,$E6,$D9,$FC
	.dc.b	$80,$FA,$4E,$9F,$13,$D5,$B5,$D5,$B3,$C1,$BB,$FD,$48,$E5,$3E,$3A
	.dc.b	$E4,$D3,$91,$63,$93,$67,$95,$74,$8F,$E1,$93,$D0,$F3,$D7,$46,$C7
	.dc.b	$46,$9D,$0B,$5D,$23,$B1,$D2,$BB,$4F,$EE,$0E,$D3,$EB,$AE,$CD,$AE
	.dc.b	$CD,$9E,$CD,$3B,$3E,$08,$EC,$78,$1E,$D5,$E0,$E1,$FF,$4E,$36,$3C
	.dc.b	$33,$F0,$2F,$C2,$D3,$C2,$58,$F0,$B6,$FC,$27,$8D,$F0,$C7,$8C,$F0
	.dc.b	$BE,$3B,$F0,$B6,$3C,$27,$C3,$AF,$0B,$63,$C2,$D3,$C2,$5A,$F0,$C7
	.dc.b	$8C,$F0,$BE,$38,$F0,$9E,$3A,$F0,$B4,$F0,$BF,$EB,$FA,$F0,$B6,$BC
	.dc.b	$2D,$9F,$0B,$6B,$C3,$1E,$39,$F0,$9E,$37,$C3,$1E,$39,$F0,$B6,$3C
	.dc.b	$22,$BC,$2D,$AF,$0B,$4F,$09,$67,$C3,$5E,$39,$F0,$CF,$8C,$F0,$BE
	.dc.b	$38,$F0,$B6,$3C,$35,$E0,$24,$78,$1B,$1E,$06,$D7,$82,$3C,$53,$E0
	.dc.b	$AF,$14,$78,$0F,$13,$E0,$69,$E0,$8E,$C4,$7B,$36,$7B,$57,$82,$3B
	.dc.b	$4F,$80,$ED,$5E,$03,$B1,$5E,$C7,$42,$57,$49,$EC,$74,$AE,$D3,$D0
	.dc.b	$ED,$5D,$1A,$74,$AE,$4F,$4A,$E4,$74,$79,$57,$4A,$E4,$D8,$E5,$3C
	.dc.b	$1E,$47,$07,$94,$F0,$7F,$9B,$FD,$1F,$F9,$F9,$FE,$8B,$F1,$0F,$A9
	.dc.b	$79,$3F,$E1,$8A,$D6,$1F,$E0,$7A,$D4,$E7,$D2,$FE,$FF,$9F,$F8,$02
	.dc.b	$73,$8F,$69,$CA,$B4,$E9,$FE,$77,$F2,$5F,$D1,$7C,$9F,$E8,$CD,$AF
	.dc.b	$E8,$CD,$9D,$C7,$F4,$99,$FE,$99,$F0,$F2,$3F,$A6,$7C,$95,$AF,$E8
	.dc.b	$F3,$72,$FE,$81,$36,$BF,$9B,$CD,$AF,$E5,$93,$57,$D0,$B9,$FF,$AE
	.dc.b	$FD,$4F,$F8,$F2,$31,$87,$F6,$BE,$64,$E6,$B7,$3F,$D9,$AB,$77,$FD
	.dc.b	$B6,$F7,$3F,$D8,$E3,$FB,$37,$47,$F6,$76,$C6,$4F,$ED,$11,$96,$9F
	.dc.b	$DF,$87,$CF,$3E,$8F,$E8,$D7,$F9,$82,$31,$3B,$5F,$3D,$A8,$E3,$56
	.dc.b	$77,$5F,$E4,$79,$C9,$C6,$B5,$52,$3A,$3C,$63,$13,$FE,$30,$C5,$F0
	.dc.b	$7F,$E0,$6F,$17,$E5,$37,$0F,$8C,$FF,$92,$AB,$FD,$39,$5C,$27,$FD
	.dc.b	$26,$D8,$FF,$4E,$B5,$FF,$7D,$C7,$C2,$FA,$3F,$57,$D1,$FA,$DB,$1F
	.dc.b	$05,$B3,$ED,$CD,$18,$E2,$FF,$2D,$D7,$CC,$FA,$3F,$88,$7C,$5E,$7B
	.dc.b	$1F,$2E,$2E,$1C,$E6,$D7,$2C,$15,$8B,$FD,$21,$F6,$2B,$17,$FA,$7F
	.dc.b	$A3,$7C,$5F,$F8,$B3,$5F,$F4,$5C,$62,$FF,$21,$18,$9F,$91,$5F,$AF
	.dc.b	$15,$B7,$FD,$86,$D7,$FD,$90,$D7,$C7,$8B,$E1,$3F,$05,$F1,$1F,$83
	.dc.b	$6E,$3F,$89,$6D,$7C,$96,$E9,$EF,$F4,$BF,$14,$7E,$2F,$B6,$DF,$E4
	.dc.b	$62,$DF,$D1,$9B,$5C,$6B,$47,$E5,$E3,$2F,$B7,$E8,$3F,$32,$FF,$94
	.dc.b	$B9,$9F,$47,$FE,$B0,$9C,$E1,$FE,$80,$9C,$D7,$F9,$63,$E8,$DF,$45
	.dc.b	$BE,$8F,$F5,$BC,$62,$7F,$CB,$B8,$72,$CF,$F2,$9F,$46,$F6,$37,$8F
	.dc.b	$FA,$FA,$31,$7D,$63,$C5,$BC,$8F,$F3,$3F,$D0,$F5,$D7,$FD,$27,$97
	.dc.b	$CF,$9C,$DE,$E7,$FD,$2F,$5B,$CB,$F4,$97,$B9,$CD,$7F,$B3,$AB,$57
	.dc.b	$FC,$BF,$E8,$DF,$46,$FA,$4D,$5F,$FB,$53,$CA,$FE,$C5,$8F,$FB,$0B
	.dc.b	$CB,$FE,$05,$F4,$2F,$27,$51,$33,$FF,$3F,$DF,$F6,$95,$5B,$73,$3B
	.dc.b	$BF,$F5,$7D,$6E,$B5,$18,$9F,$F7,$2D,$3E,$72,$9E,$2F,$22,$C7,$FA
	.dc.b	$38,$D1,$63,$4D,$25,$7F,$AF,$9D,$BF,$C6,$6D,$5D,$23,$FD,$87,$F4
	.dc.b	$6F,$A3,$FF,$7B,$B6,$FF,$96,$5B,$E9,$33,$52,$7F,$D4,$35,$BB,$ED
	.dc.b	$59,$7A,$D5,$E3,$FE,$17,$87,$3C,$38,$CF,$F3,$6A,$7C,$2F,$F4,$6B
	.dc.b	$FE,$44,$DA,$FE,$0F,$6E,$AE,$13,$9B,$FE,$DC,$D7,$9F,$97,$F7,$27
	.dc.b	$6C,$BF,$F5,$CD,$71,$7E,$DC,$E6,$BF,$DC,$FE,$7F,$E1,$8F,$F1,$44
	.dc.b	$3E,$91,$F0,$1E,$99,$D7,$0D,$3F,$D8,$F3,$AC,$3F,$D8,$F3,$A6,$CE
	.dc.b	$AB,$84,$6B,$E6,$7A,$1F,$ED,$0F,$A3,$7D,$26,$27,$44,$CF,$FD,$B1
	.dc.b	$3A,$8C,$6F,$FD,$2F,$27,$D2,$79,$CF,$A1,$7F,$E1,$5F,$A4,$FF,$0B
	.dc.b	$7D,$13,$D6,$FE,$55,$7C,$31,$F4,$E1,$C5,$FF,$70,$E1,$C4,$5F,$17
	.dc.b	$FC,$41,$5C,$56,$3F,$CD,$96,$3E,$0C,$7B,$E2,$C7,$15,$FC,$EC,$D9
	.dc.b	$DA,$77,$1A,$8F,$E8,$15,$70,$F0,$63,$E3,$9C,$9C,$4F,$2E,$1D,$DC
	.dc.b	$1A,$8E,$13,$CB,$0F,$14,$F8,$A7,$35,$C5,$ED,$1F,$0D,$CE,$1E,$58
	.dc.b	$C5,$7A,$1E,$E6,$0F,$1C,$6D,$ED,$1B,$9E,$B1,$83,$67,$23,$07,$F9
	.dc.b	$3A,$BC,$93,$E8,$C3,$E2,$46,$27,$E6,$46,$2F,$C0,$FB,$A3,$E3,$95
	.dc.b	$C3,$F7,$CF,$A9,$F4,$9F,$C6,$7E,$07,$83,$63,$E3,$B5,$DF,$52,$77
	.dc.b	$F4,$9F,$84,$9D,$B5,$EA,$D7,$FB,$E5,$23,$11,$FE,$6D,$8C,$6B,$F8
	.dc.b	$3F,$A4,$79,$BF,$E0,$3D,$BF,$05,$18,$BF,$C4,$5F,$07,$F0,$2D,$9D
	.dc.b	$D7,$1D,$BF,$C1,$97,$C6,$F7,$8F,$A6,$FF,$1D,$87,$1B,$DE,$1C,$6F
	.dc.b	$FD,$05,$87,$1B,$DE,$1C,$6F,$DA,$DA,$DD,$F1,$57,$C6,$B7,$8F,$AA
	.dc.b	$F8,$56,$AF,$85,$EE,$3E,$34,$EB,$1F,$34,$76,$3E,$45,$77,$8C,$C1
	.dc.b	$B5,$CE,$3B,$1E,$1B,$E2,$FF,$A7,$DF,$F2,$D4,$7C,$76,$D7,$F8,$D5
	.dc.b	$FF,$2F,$57,$48,$C6,$BF,$3B,$C6,$B8,$7C,$86,$BE,$03,$F8,$5C,$7F
	.dc.b	$8F,$FE,$93,$FD,$2D,$1E,$18,$FF,$21,$7D,$13,$E0,$F8,$B6,$74,$7F
	.dc.b	$AD,$6B,$8B,$D0,$C7,$67,$0A,$91,$C0,$4F,$0C,$FF,$CF,$0B,$E9,$3B
	.dc.b	$FB,$F2,$B8,$4E,$5B,$39,$7C,$03,$15,$7C,$0A,$E1,$3A,$6D,$FF,$A1
	.dc.b	$EB,$73,$E5,$6B,$BB,$F9,$3F,$48,$EA,$F8,$7E,$B4,$AF,$64,$E5,$B3
	.dc.b	$97,$CB,$7F,$E7,$47,$24,$33,$5B,$C3,$FD,$35,$7C,$E7,$DB,$EB,$FF
	.dc.b	$BB,$9F,$37,$C2,$7D,$43,$EE,$C6,$63,$7E,$AE,$5D,$9F,$46,$35,$7F
	.dc.b	$EA,$76,$F0,$7E,$02,$F8,$57,$23,$EC,$7D,$27,$FB,$01,$FB,$5C,$7F
	.dc.b	$DE,$9B,$E0,$7E,$3A,$75,$1F,$51,$B8,$FF,$AD,$6B,$84,$FD,$86,$C7
	.dc.b	$FA,$E6,$77,$87,$FA,$4D,$B4,$3F,$8A,$DF,$E7,$2F,$8D,$6F,$D5,$FC
	.dc.b	$FF,$2F,$DE,$8F,$F4,$E7,$D1,$BE,$93,$F4,$6D,$C7,$F2,$95,$B7,$B3
	.dc.b	$54,$6D,$FC,$7B,$73,$FF,$68,$4E,$CE,$11,$9B,$F0,$DF,$93,$8F,$F6
	.dc.b	$7F,$48,$F7,$69,$32,$FD,$37,$D2,$7F,$92,$5F,$D4,$7D,$23,$87,$F7
	.dc.b	$6D,$8C,$C7,$17,$F4,$7F,$2E,$FA,$4F,$E0,$2D,$8F,$D9,$35,$FD,$4D
	.dc.b	$73,$7F,$57,$F4,$8F,$A3,$FE,$12,$BE,$92,$98,$8F,$E1,$BF,$44,$FA
	.dc.b	$2F,$E1,$FF,$48,$FA,$4A,$6A,$3F,$89,$FD,$27,$F8,$B3,$2F,$1F,$6F
	.dc.b	$98,$CF,$9D,$FC,$63,$C4,$FF,$5A,$CF,$F0,$95,$F6,$59,$35,$F0,$F2
	.dc.b	$75,$19,$78,$46,$5F,$E0,$F3,$C9,$D9,$A3,$B7,$CE,$78,$7E,$96,$46
	.dc.b	$08,$72,$72,$F4,$8F,$E5,$FF,$48,$FA,$4F,$F3,$AC,$FF,$88,$A3,$1E
	.dc.b	$DF,$F8,$1D,$7C,$EF,$F9,$D6,$75,$5C,$1F,$F1,$D7,$D4,$7F,$B0,$5B
	.dc.b	$F0,$BF,$E4,$35,$19,$7F,$DF,$4D,$5F,$09,$FE,$09,$FA,$0F,$F1,$C7
	.dc.b	$8D,$FE,$39,$8F,$F0,$67,$A5,$8F,$17,$F5,$D1,$B7,$F9,$FB,$A6,$FB
	.dc.b	$0B,$E1,$87,$68,$FF,$46,$57,$18,$FE,$88,$F1,$6C,$71,$AE,$77,$CA
	.dc.b	$B8,$61,$E5,$AE,$63,$BF,$4F,$FB,$A3,$0E,$71,$8F,$8D,$CB,$71,$F3
	.dc.b	$B6,$7F,$A3,$FD,$27,$FA,$35,$FF,$2A,$B5,$FF,$34,$35,$FE,$20,$1C
	.dc.b	$3F,$C4,$DE,$58,$7F,$98,$E3,$5E,$6F,$C4,$F7,$5E,$7F,$FB,$C6,$B5
	.dc.b	$7C,$31,$F9,$F9,$BE,$93,$DB,$8F,$B9,$B3,$F4,$1B,$97,$CA,$6E,$1F
	.dc.b	$1D,$B3,$D1,$B3,$F5,$7E,$93,$E4,$9F,$2B,$E6,$6D,$7D,$C8,$B3,$F5
	.dc.b	$A6,$DF,$B7,$E8,$5D,$3F,$ED,$16,$E1,$FE,$04,$DB,$DF,$F3,$59,$A9
	.dc.b	$3F,$E0,$1A,$DE,$3F,$E0,$7B,$DC,$E6,$BE,$EF,$D1,$C8,$FC,$7F,$4D
	.dc.b	$BF,$C4,$CD,$CB,$F0,$D3,$6B,$E5,$CD,$AF,$CA,$4D,$E4,$FF,$35,$57
	.dc.b	$E2,$BE,$8F,$F1,$6D,$3F,$3A,$5A,$FC,$BB,$70,$F8,$5E,$EF,$C1,$3F
	.dc.b	$43,$F1,$2F,$9E,$FC,$E5,$FF,$85,$36,$FF,$15,$61,$F0,$6F,$83,$FE
	.dc.b	$2A,$BE,$1D,$58,$BF,$D3,$8F,$A3,$FD,$23,$6B,$F6,$6F,$F4,$36,$C6
	.dc.b	$88,$EB,$D1,$B3,$D2,$6E,$7F,$C0,$63,$91,$9A,$E6,$47,$C7,$5C,$23
	.dc.b	$F8,$2C,$78,$CE,$06,$9E,$13,$D6,$B8,$4F,$F1,$23,$BC,$66,$76,$74
	.dc.b	$9F,$7D,$70,$D7,$FC,$01,$16,$F8,$D6,$CE,$2F,$58,$E3,$18,$9E,$2D
	.dc.b	$9E,$26,$DC,$4F,$08,$F5,$DF,$95,$B3,$F0,$2F,$C5,$5E,$36,$C7,$56
	.dc.b	$CF,$03,$E0,$38,$AE,$87,$D1,$71,$3E,$B9,$F9,$47,$1B,$F4,$B4,$F4
	.dc.b	$B8,$AE,$73,$C5,$E8,$FB,$0F,$21,$67,$95,$78,$1F,$84,$78,$27,$9B
	.dc.b	$86,$D7,$F8,$C6,$78,$3E,$1A,$F8,$87,$59,$DD,$79,$8F,$C3,$38,$BF
	.dc.b	$9C,$D9,$F2,$CF,$B6,$7E,$64,$73,$3D,$87,$68,$D3,$C2,$79,$47,$B2
	.dc.b	$38,$3E,$97,$84,$68,$F2,$DF,$AA,$7C,$B3,$A3,$BB,$C4,$97,$F0,$63
	.dc.b	$AB,$F0,$63,$EE,$C6,$1E,$66,$0F,$39,$E5,$2D,$78,$A3,$85,$FC,$47
	.dc.b	$C3,$5D,$49,$5F,$2E,$7D,$17,$E9,$78,$47,$98,$F0,$CF,$09,$D5,$7F
	.dc.b	$03,$3B,$96,$7C,$6F,$81,$EC,$FA,$E3,$8C,$7E,$16,$35,$5F,$5A,$3F
	.dc.b	$29,$F4,$7F,$C5,$BE,$2F,$F8,$BB,$6B,$F8,$BB,$67,$F8,$EB,$71,$FE
	.dc.b	$D1,$F4,$5F,$E6,$3A,$FF,$63,$36,$BF,$D9,$9F,$4E,$FA,$87,$CB,$7F
	.dc.b	$84,$25,$5B,$C3,$9F,$B3,$98,$C6,$3F,$E8,$9C,$79,$61,$9A,$FE,$5B
	.dc.b	$F2,$DF,$CB,$BC,$BB,$7F,$CB,$9B,$97,$F3,$46,$E5,$FC,$A5,$B3,$FC
	.dc.b	$91,$B3,$FC,$7D,$FE,$73,$F1,$72,$7F,$9C,$FC,$B5,$AF,$E7,$F3,$6F
	.dc.b	$F9,$A4,$DA,$FF,$08,$F6,$BC,$7F,$80,$D8,$FE,$95,$D3,$FE,$77,$7F
	.dc.b	$DD,$D1,$9C,$7F,$DC,$AD,$C3,$FC,$2C,$D9,$FF,$56,$73,$FF,$59,$F5
	.dc.b	$FA,$E3,$FD,$43,$0F,$EA,$33,$6B,$FA,$8B,$63,$FA,$BB,$70,$FE,$98
	.dc.b	$D9,$FE,$8E,$D9,$FE,$84,$FF,$92,$21,$F3,$CF,$A3,$C3,$63,$FD,$81
	.dc.b	$67,$FD,$84,$70,$F8,$4D,$D4,$8D,$9A,$AE,$58,$2F,$93,$CE,$17,$C6
	.dc.b	$3E,$3F,$C3,$7F,$9B,$74,$34,$98,$7F,$91,$27,$EA,$C7,$D4,$3F,$C0
	.dc.b	$7A,$BC,$5C,$57,$92,$BC,$1F,$45,$EE,$FA,$3F,$7B,$67,$47,$C4,$71
	.dc.b	$3E,$98,$F6,$46,$34,$F5,$C6,$CE,$3A,$2F,$AC,$61,$F9,$BF,$47,$23
	.dc.b	$7C,$D7,$1E,$C7,$C5,$9C,$61,$DB,$0F,$D5,$CE,$6F,$D8,$BE,$8D,$F1
	.dc.b	$9F,$E9,$56,$CF,$C7,$9F,$F7,$D4,$61,$C3,$5F,$98,$E6,$FF,$CD,$8D
	.dc.b	$9F,$CB,$EB,$FE,$AD,$7C,$31,$9A,$FB,$9F,$15,$F8,$AF,$8B,$FC,$5B
	.dc.b	$6B,$E2,$B6,$7D,$F1,$F9,$28,$C6,$1E,$F6,$C7,$E2,$DB,$21,$F4,$1F
	.dc.b	$8A,$9F,$87,$CA,$F6,$FF,$8B,$4E,$6B,$E6,$89,$D5,$7C,$AF,$A2,$7B
	.dc.b	$1C,$39,$FF,$8F,$46,$27,$E7,$E1,$CF,$3F,$F4,$87,$D1,$7F,$A8,$DF
	.dc.b	$F4,$95,$6F,$2E,$F3,$9B,$FF,$4D,$CE,$6B,$FD,$31,$7F,$E9,$59,$FF
	.dc.b	$4F,$B6,$FD,$DF,$45,$B9,$CE,$BF,$CC,$63,$13,$F0,$EB,$82,$8C,$BF
	.dc.b	$1A,$78,$67,$2F,$F4,$9F,$44,$FA,$2F,$E7,$B1,$F8,$CC,$38,$F9,$6F
	.dc.b	$27,$FA,$03,$67,$F4,$FF,$3E,$F1,$73,$52,$7F,$2B,$5B,$D3,$EF,$DE
	.dc.b	$E7,$35,$B9,$FA,$B5,$BD,$BF,$5B,$7B,$9C,$D7,$F5,$1B,$7C,$DD,$3F
	.dc.b	$8E,$53,$C4,$C4,$FE,$98,$D9,$67,$71,$AB,$FC,$2C,$E9,$B3,$A8,$E0
	.dc.b	$7E,$C3,$E8,$9F,$43,$FD,$8D,$B1,$F1,$3E,$8B,$13,$23,$83,$E6,$8C
	.dc.b	$5F,$E5,$2B,$8B,$67,$55,$C0,$FD,$FF,$C2,$7F,$49,$9F,$D1,$36,$19
	.dc.b	$7E,$75,$DF,$8F,$FD,$DE,$78,$51,$E1,$5B,$8F,$D2,$C6,$89,$3F,$33
	.dc.b	$A7,$93,$FE,$01,$F6,$7E,$B6,$1F,$EF,$66,$DF,$EB,$7E,$8B,$FD,$DC
	.dc.b	$FE,$DE,$32,$9A,$70,$B4,$F0,$9D,$6F,$DC,$97,$FE,$B9,$9D,$5F,$2F
	.dc.b	$9A,$F2,$1F,$EC,$4F,$A1,$7D,$16,$2F,$F9,$3D,$2B,$18,$FF,$30,$C9
	.dc.b	$87,$0B,$DD,$7B,$FE,$2E,$47,$F3,$0B,$FE,$25,$F4,$5F,$C4,$FE,$89
	.dc.b	$E8,$FC,$3B,$DD,$7F,$10,$CD,$1F,$E2,$98,$E2,$2B,$8B,$F1,$E3,$4A
	.dc.b	$FB,$5F,$5C,$2C,$FF,$63,$9A,$FC,$DA,$FE,$DF,$7D,$23,$F9,$41,$CC
	.dc.b	$F0,$5F,$36,$9C,$CA,$73,$BE,$AF,$A2,$7A,$61,$EC,$9C,$CF,$59,$CD
	.dc.b	$F8,$AF,$FA,$96,$1E,$6C,$38,$5F,$63,$A4,$F8,$EF,$D1,$18,$3E,$26
	.dc.b	$1D,$AF,$96,$1D,$67,$D1,$5E,$0A,$D4,$7A,$B0,$EF,$3D,$23,$CD,$87
	.dc.b	$73,$CA,$EE,$FA,$3F,$23,$0F,$ED,$4D,$29,$F2,$CC,$BA,$FA,$0F,$E2
	.dc.b	$BF,$0B,$79,$CD,$25,$7F,$75,$97,$D1,$7F,$89,$E3,$0D,$7F,$C4,$11
	.dc.b	$AE,$AE,$D3,$99,$FB,$B3,$97,$EC,$FD,$13,$C8,$FC,$0E,$BF,$83,$8C
	.dc.b	$4F,$E2,$2B,$84,$FE,$1D,$BA,$FE,$22,$B7,$5D,$2F,$9C,$F8,$30,$F4
	.dc.b	$61,$C5,$B1,$C6,$77,$5F,$DE,$6F,$9D,$EF,$1F,$16,$1C,$2B,$57,$C1
	.dc.b	$5F,$0C,$3C,$17,$C0,$FD,$0C,$66,$78,$99,$3F,$37,$8F,$B1,$C9,$EB
	.dc.b	$35,$8F,$8D,$C1,$1C,$1F,$27,$0F,$D5,$DF,$F9,$72,$7C,$B1,$8D,$FE
	.dc.b	$04,$7F,$8D,$AF,$C0,$E8,$FF,$1C,$E0,$F0,$F8,$47,$C5,$6B,$B3,$FC
	.dc.b	$33,$F4,$5F,$E9,$37,$FC,$83,$F4,$4F,$81,$DB,$63,$FC,$8B,$3B,$78
	.dc.b	$F5,$6A,$A4,$6A,$3B,$CB,$E7,$F9,$FB,$52,$B9,$C6,$8C,$3F,$E2,$CC
	.dc.b	$13,$98,$D9,$E6,$77,$B7,$C7,$BE,$D3,$FE,$5A,$FA,$2F,$E6,$8F,$B3
	.dc.b	$D1,$F5,$9F,$0A,$33,$1F,$E3,$A6,$D7,$D1,$C3,$FC,$C3,$3C,$4E,$9D
	.dc.b	$3C,$4E,$6E,$4E,$4F,$CE,$F6,$1F,$2B,$F2,$BE,$0E,$57,$F9,$BA,$72
	.dc.b	$D9,$CB,$F0,$67,$94,$48,$FA,$B1,$AC,$3F,$0D,$5C,$A7,$FC,$A7,$F1
	.dc.b	$DF,$EA,$27,$FC,$F7,$CA,$F8,$DF,$C0,$B6,$3F,$CF,$F3,$B7,$8F,$A7
	.dc.b	$AA,$91,$A8,$FC,$4C,$EF,$0F,$CB,$5F,$3A,$FF,$31,$B6,$3F,$35,$3B
	.dc.b	$E7,$FE,$8F,$CB,$F5,$AF,$FC,$D3,$F4,$4F,$A2,$FC,$AB,$63,$FD,$27
	.dc.b	$3B,$C3,$F4,$8D,$BF,$EA,$35,$BA,$E7,$7F,$9E,$AD,$47,$F9,$F5,$AA
	.dc.b	$78,$5F,$EB,$DB,$7E,$6E,$5F,$D3,$FD,$13,$DC,$A4,$C3,$FD,$4F,$F4
	.dc.b	$4F,$3B,$F5,$BF,$42,$E3,$A4,$AF,$F5,$6E,$1C,$DF,$F5,$7F,$CA,$3E
	.dc.b	$8B,$F6,$AD,$8C,$C7,$17,$FD,$1F,$F4,$4F,$87,$FD,$DA,$FA,$2A,$62
	.dc.b	$3F,$7F,$F4,$4F,$A0,$FE,$03,$F4,$4F,$A2,$A6,$A3,$F8,$2F,$D1,$7F
	.dc.b	$83,$32,$F2,$F7,$79,$0C,$F9,$FF,$C2,$BC,$5F,$F5,$24,$FF,$B2,$17
	.dc.b	$D1,$64,$D7,$C2,$C9,$D4,$65,$E1,$19,$7F,$D5,$73,$C9,$D9,$A3,$B7
	.dc.b	$CD,$78,$7E,$8A,$46,$08,$72,$72,$F4,$8F,$E4,$3F,$51,$FC,$32,$3F
	.dc.b	$B8,$B5,$FF,$6E,$35,$FF,$0E,$87,$0F,$F8,$83,$E8,$DF,$47,$FE,$61
	.dc.b	$8F,$F7,$24,$63,$DD,$FE,$FA,$BE,$77,$FD,$8D,$3A,$AE,$0F,$F8,$AB
	.dc.b	$E0,$FF,$BD,$FC,$DB,$C2,$FF,$6A,$78,$96,$95,$FE,$00,$F4,$71,$39
	.dc.b	$8E,$35,$C0,$FF,$6E,$3B,$71,$5C,$BD,$86,$1C,$A7,$71,$FE,$83,$AE
	.dc.b	$74,$BE,$D3,$93,$8D,$74,$AE,$D5,$B8,$E7,$7D,$63,$C1,$85,$AF,$F0
	.dc.b	$CF,$75,$C3,$FD,$B9,$5E,$58,$C7,$D1,$65,$B8,$F7,$A4,$62,$BF,$DE
	.dc.b	$7F,$45,$FE,$86,$F7,$DF,$50,$FA,$5F,$15,$2B,$53,$9C,$7E,$2C,$E7
	.dc.b	$D1,$FE,$E4,$FF,$97,$AB,$F9,$54,$FF,$8E,$1B,$1F,$86,$38,$45,$D3
	.dc.b	$FC,$6B,$F4,$9E,$DB,$F7,$36,$7D,$4D,$AF,$90,$7D,$57,$1C,$1F,$41
	.dc.b	$B3,$F5,$9B,$3F,$6F,$E6,$3E,$51,$F2,$FE,$92,$D7,$C7,$8B,$3D,$B1
	.dc.b	$FF,$24,$D6,$30,$FF,$08,$7D,$13,$A7,$EC,$43,$EA,$3E,$FE,$1E,$61
	.dc.b	$5B,$CB,$F1,$B5,$BE,$5F,$9F,$39,$AF,$8C,$D7,$F5,$AD,$7F,$D1,$6D
	.dc.b	$7F,$2D,$F4,$BF,$85,$3F,$0D,$D1,$70,$FC,$14,$59,$F6,$72,$FE,$BE
	.dc.b	$72,$48,$CC,$F2,$8C,$63,$F9,$C7,$F2,$5F,$4B,$F9,$36,$9F,$8F,$2D
	.dc.b	$76,$6C,$FC,$8F,$7F,$F3,$DC,$CF,$99,$FF,$72,$E9,$FE,$2B,$BF,$D8
	.dc.b	$CE,$7A,$BE,$2D,$F0,$BD,$B6,$3F,$4D,$8B,$E9,$69,$8F,$42,$E3,$FB
	.dc.b	$19,$D3,$67,$4D,$24,$7E,$04,$D9,$67,$9B,$99,$B9,$FF,$0D,$8F,$EB
	.dc.b	$95,$CA,$31,$3A,$8F,$E1,$F3,$A3,$93,$B3,$31,$E4,$8C,$92,$79,$CF
	.dc.b	$AE,$BA,$4F,$F2,$03,$CF,$3A,$7D,$13,$D5,$D5,$70,$E4,$47,$8E,$B8
	.dc.b	$12,$BC,$67,$73,$94,$F0,$AE,$55,$C6,$76,$F9,$DB,$5F,$16,$F7,$5C
	.dc.b	$1B,$3D,$2B,$91,$E6,$71,$3F,$0E,$BC,$93,$93,$B1,$F0,$6B,$9B,$8B
	.dc.b	$F6,$C6,$27,$8C,$62,$FA,$9E,$19,$D3,$E8,$3E,$BB,$8A,$F0,$57,$AE
	.dc.b	$78,$C7,$73,$D3,$5F,$16,$33,$3B,$9D,$5F,$C1,$9C,$9D,$E3,$D4,$E1
	.dc.b	$EE,$F8,$8C,$57,$58,$FB,$A7,$48,$F5,$1A,$7F,$06,$F1,$74,$79,$EF
	.dc.b	$D9,$3C,$EF,$CC,$7C,$69,$E2,$78,$8F,$29,$C9,$D4,$F2,$8D,$CF,$CD
	.dc.b	$75,$5E,$D8,$C5,$75,$BF,$85,$5A,$3B,$4E,$AB,$84,$74,$25,$F7,$AF
	.dc.b	$C2,$4E,$9E,$77,$D6,$3E,$1D,$75,$BF,$74,$7C,$39,$D4,$FC,$38,$F9
	.dc.b	$15,$D2,$3C,$67,$85,$C5,$72,$8C,$C6,$1D,$4F,$82,$BA,$9F,$19,$EA
	.dc.b	$7E,$AE,$3C,$87,$B1,$FD,$17,$D2,$FF,$1C,$F8,$EF,$E3,$AD,$8F,$E3
	.dc.b	$4D,$9F,$E2,$CD,$AF,$F7,$27,$D2,$7F,$9C,$EB,$FC,$37,$F1,$8F,$A8
	.dc.b	$FF,$57,$C7,$F7,$A9,$D6,$1F,$DB,$2B,$5E,$5F,$F8,$0A,$73,$87,$F8
	.dc.b	$B6,$73,$7C,$7D,$9F,$F0,$C4,$63,$0F,$E5,$1F,$2B,$FC,$D3,$CA,$FE
	.dc.b	$6A,$DB,$FE,$64,$D9,$DC,$7F,$35,$9D,$B4,$D9,$63,$F9,$A7,$C5,$FF
	.dc.b	$40,$3F,$A0,$FC,$AD,$9F,$E7,$D1,$67,$F9,$D4,$DC,$BF,$C8,$D0,$FA
	.dc.b	$27,$27,$F4,$DF,$A8,$FF,$10,$C6,$31,$FE,$A7,$CE,$9C,$D7,$F8,$82
	.dc.b	$FF,$94,$DE,$DF,$F0,$F3,$5F,$F3,$2B,$5F,$E8,$27,$F5,$8E,$8F,$EB
	.dc.b	$2D,$C3,$FA,$B3,$4F,$EA,$45,$C7,$FA,$73,$63,$99,$FD,$76,$7F,$C9
	.dc.b	$4B,$E8,$1F,$49,$86,$C7,$F9,$78,$D7,$C3,$E6,$64,$66,$77,$7F,$E3
	.dc.b	$08,$D9,$99,$F8,$FF,$11,$FD,$D7,$C5,$B3,$F2,$1B,$5F,$12,$7E,$43
	.dc.b	$F4,$5B,$99,$F9,$95,$E9,$FA,$4F,$47,$D2,$7A,$5B,$1E,$56,$DF,$97
	.dc.b	$24,$63,$5F,$F4,$25,$73,$3D,$97,$CD,$A7,$BF,$E9,$3E,$31,$F1,$F9
	.dc.b	$EC,$64,$98,$F7,$8C,$BF,$E3,$FF,$A4,$7C,$87,$F8,$5F,$4F,$F3,$5C
	.dc.b	$62,$FF,$17,$18,$9F,$87,$39,$AF,$F3,$33,$5F,$F5,$6B,$5F,$EE,$F5
	.dc.b	$F7,$7E,$2B,$EE,$7C,$6F,$DD,$6C,$7D,$96,$CF,$C5,$6E,$9F,$13,$64
	.dc.b	$62,$BA,$C6,$1F,$C1,$5F,$FA,$13,$E8,$9C,$7F,$E8,$6E,$47,$B7,$98
	.dc.b	$C3,$FE,$8A,$FA,$37,$D1,$E7,$1F,$F4,$4C,$61,$FF,$45,$E1,$D1,$FF
	.dc.b	$46,$FD,$27,$FA,$7B,$65,$6D,$FF,$36,$46,$63,$1B,$7E,$F2,$FD,$15
	.dc.b	$FE,$71,$6C,$7F,$00,$FA,$2D,$E1,$FE,$20,$A5,$6A,$31,$3A,$9F,$F1
	.dc.b	$2C,$E5,$FF,$13,$C6,$5E,$3B,$FF,$AA,$23,$07,$1F,$A4,$7D,$23,$E9
	.dc.b	$75,$AF,$FA,$BB,$CB,$79,$54,$FD,$CF,$D1,$BC,$6C,$6D,$FE,$B0,$AD
	.dc.b	$B6,$B7,$3A,$D7,$FD,$21,$5A,$6C,$EA,$B8,$1F,$EB,$A1,$F3,$FF,$EC
	.dc.b	$06,$9F,$C1,$70,$78,$B9,$99,$3C,$5F,$3C,$62,$FF,$D4,$13,$C9,$B3
	.dc.b	$A8,$E0,$7F,$B2,$BE,$95,$F4,$BF,$ED,$06,$C7,$87,$E5,$31,$32,$38
	.dc.b	$3F,$1A,$31,$7F,$EA,$FA,$E2,$D9,$D5,$70,$3F,$DB,$3F,$0B,$FC,$DF
	.dc.b	$97,$FA,$1B,$63,$F5,$B9,$BA,$BF,$DC,$73,$9C,$3F,$CC,$ED,$4E,$69
	.dc.b	$5C,$5E,$FE,$3F,$22,$3F,$36,$ED,$FF,$B1,$FC,$CF,$B1,$8F,$F5,$66
	.dc.b	$D7,$F9,$DA,$9F,$4B,$FE,$EC,$AD,$70,$7F,$BE,$2F,$78,$7F,$9F,$1B
	.dc.b	$7C,$3D,$FD,$54,$C7,$FD,$F7,$5A,$C7,$A4,$7E,$96,$39,$E0,$5F,$37
	.dc.b	$DC,$FF,$5D,$FD,$23,$E9,$71,$7D,$26,$DF,$FB,$E3,$35,$EE,$B4,$78
	.dc.b	$BE,$37,$D4,$1F,$F4,$AE,$BF,$DB,$FE,$97,$FD,$2F,$F4,$8F,$4F,$CD
	.dc.b	$7C,$2F,$91,$C0,$FD,$F6,$5F,$E9,$EB,$E3,$5C,$27,$53,$A9,$CB,$F2
	.dc.b	$6B,$C4,$AF,$C0,$EC,$C5,$F3,$69,$98,$DB,$F2,$67,$9C,$62,$3C,$87
	.dc.b	$03,$57,$E2,$7F,$8B,$5D,$D5,$7C,$B0,$F6,$D2,$B5,$39,$8F,$14,$F9
	.dc.b	$70,$F0,$46,$CF,$05,$F4,$C3,$C1,$1E,$49,$F4,$CE,$63,$CC,$66,$BA
	.dc.b	$1E,$2C,$3C,$37,$CA,$7B,$5F,$2C,$3C,$91,$EC,$3C,$2E,$1E,$87,$D1
	.dc.b	$9E,$55,$EE,$8E,$05,$70,$FC,$A3,$EF,$FD,$27,$F1,$CF,$80,$E0,$D8
	.dc.b	$F8,$ED,$77,$D2,$9D,$FD,$2B,$5A,$4A,$E3,$19,$6C,$7D,$7F,$A5,$79
	.dc.b	$5F,$D3,$75,$FC,$0C,$62,$BE,$E6,$4C,$39,$57,$E1,$A3,$18,$73,$C3
	.dc.b	$96,$DF,$D5,$F1,$E5,$8F,$E1,$EF,$6D,$C7,$9D,$6E,$F8,$CF,$FB,$FA
	.dc.b	$F8,$E3,$DE,$F7,$21,$87,$49,$F4,$DF,$29,$5F,$2C,$3F,$B5,$5E,$DB
	.dc.b	$87,$4A,$DD,$F1,$BE,$11,$DE,$F8,$36,$B8,$4E,$A3,$89,$F2,$23,$A4
	.dc.b	$F0,$8C,$9F,$A9,$9E,$11,$97,$E3,$E3,$FE,$80,$6E,$BF,$16,$BF,$80
	.dc.b	$9F,$02,$FC,$38,$7F,$81,$6F,$87,$8B,$6F,$E9,$36,$FE,$65,$55,$F4
	.dc.b	$BF,$E8,$C7,$AF,$D2,$3E,$07,$4D,$8F,$0C,$ED,$FC,$86,$1F,$E1,$58
	.dc.b	$C6,$3F,$E9,$8A,$E1,$5A,$3E,$C3,$88,$FF,$4E,$CF,$0A,$E5,$F4,$3E
	.dc.b	$0F,$72,$61,$E0,$8F,$45,$1C,$C7,$51,$82,$B7,$32,$7D,$B1,$C3,$0F
	.dc.b	$0C,$FC,$3F,$A5,$FE,$8F,$5F,$0B,$D3,$F8,$26,$7A,$A4,$64,$87,$86
	.dc.b	$78,$B6,$38,$BA,$BF,$F5,$DC,$E9,$FF,$5E,$FA,$8F,$7F,$E7,$FC,$1C
	.dc.b	$BF,$BF,$39,$6C,$E5,$F8,$1A,$2B,$EE,$56,$E7,$E2,$36,$FE,$A4,$F1
	.dc.b	$7C,$37,$F8,$3F,$A5,$FF,$33,$DF,$E3,$B8,$5E,$17,$E1,$DB,$1F,$92
	.dc.b	$9D,$BF,$45,$B7,$FE,$D8,$9D,$CF,$D5,$6C,$7F,$B7,$27,$7C,$3F,$E5
	.dc.b	$D6,$D7,$EF,$6B,$73,$FE,$61,$6C,$7F,$01,$9D,$BF,$E8,$4D,$FF,$39
	.dc.b	$F4,$8F,$A5,$A7,$E7,$F9,$7F,$11,$5F,$AA,$6B,$F8,$99,$E0,$FE,$B9
	.dc.b	$AF,$E3,$63,$79,$7F,$BD,$1B,$97,$F9,$D6,$B7,$3F,$EF,$86,$C7,$F9
	.dc.b	$E6,$77,$5F,$C6,$7E,$8D,$ED,$FE,$CD,$B1,$FB,$E6,$BF,$9D,$FA,$57
	.dc.b	$9B,$FB,$9F,$A5,$FC,$04,$FF,$04,$6B,$FA,$36,$E7,$FE,$04,$85,$F4
	.dc.b	$8C,$7C,$B3,$E9,$7F,$82,$36,$BF,$87,$B5,$E9,$C9,$FC,$27,$E9,$5F
	.dc.b	$49,$7E,$96,$57,$F1,$69,$CB,$FC,$63,$E9,$5F,$47,$FC,$47,$E9,$5F
	.dc.b	$4B,$4F,$E2,$93,$AA,$FE,$3F,$F4,$BF,$E0,$A7,$C7,$D9,$E6,$3F,$8D
	.dc.b	$46,$63,$F8,$17,$E1,$1D,$1F,$E4,$ED,$73,$7F,$C8,$7C,$3F,$97,$1A
	.dc.b	$8C,$78,$B2,$30,$43,$F8,$74,$FF,$19,$E9,$FF,$5A,$36,$3F,$AA,$5F
	.dc.b	$7E,$3F,$A1,$3F,$ED,$09,$FE,$72,$BE,$9B,$C4,$78,$FE,$1A,$4E,$A3
	.dc.b	$2F,$08,$F4,$7C,$E7,$A7,$E9,$A4,$F4,$8C,$BD,$A3,$2F,$A2,$3F,$A0
	.dc.b	$7D,$2B,$E9,$7F,$D0,$73,$FE,$34,$8C,$7B,$BF,$E1,$95,$F3,$BF,$E8
	.dc.b	$D9,$D5,$70,$7F,$CA,$FF,$53,$FE,$DF,$6F,$C2,$FE,$3D,$FF,$6C,$46
	.dc.b	$5B,$19,$7F,$BD,$43,$DF,$C4,$23,$1C,$3F,$E2,$B6,$C7,$F9,$29,$AF
	.dc.b	$FB,$97,$E1,$3F,$C4,$9F,$3F,$FE,$62,$E2,$FF,$31,$C7,$F8,$9F,$0C
	.dc.b	$61,$FE,$09,$8D,$BF,$E6,$C7,$4D,$F6,$17,$C3,$05,$69,$FE,$B8,$F2
	.dc.b	$6C,$FF,$AC,$63,$FC,$6D,$E0,$71,$7F,$C7,$1F,$4D,$65,$36,$30,$DC
	.dc.b	$3C,$4D,$BF,$EB,$DF,$49,$FC,$D2,$7F,$C9,$CD,$8F,$F3,$6B,$5F,$F4
	.dc.b	$5B,$5F,$F1,$40,$7F,$D5,$2F,$F8,$AD,$BE,$F3,$CE,$79,$7F,$06,$B8
	.dc.b	$7A,$16,$F8,$C6,$30,$FF,$4A,$E1,$C6,$F5,$7E,$FF,$A4,$F8,$0D,$AF
	.dc.b	$9C,$DB,$F9,$2D,$E0,$FB,$3A,$F9,$B6,$FF,$30,$7D,$27,$C9,$3E,$57
	.dc.b	$CC,$DC,$7E,$C4,$5C,$7E,$9C,$DC,$3F,$91,$7D,$0B,$AB,$D8,$DC,$3D
	.dc.b	$3C,$0F,$9A,$CD,$49,$FF,$18,$D6,$F1,$FF,$0E,$70,$7C,$7B,$D4,$7F
	.dc.b	$84,$E7,$58,$F8,$30,$FC,$55,$EF,$0E,$71,$F8,$3F,$A3,$91,$F9,$1E
	.dc.b	$AB,$8F,$E3,$26,$E1,$F1,$26,$D7,$E6,$F7,$FF,$69,$3F,$99,$FA,$3F
	.dc.b	$C7,$3F,$8F,$6B,$93,$F3,$C6,$4F,$6F,$87,$F1,$0F,$D0,$7C,$3B,$E7
	.dc.b	$FF,$47,$3A,$26,$7E,$FA,$DC,$EA,$31,$3A,$CB,$E8,$D6,$BA,$BB,$4E
	.dc.b	$4F,$CA,$EF,$FE,$51,$7F,$DB,$F8,$F8,$67,$78,$74,$26,$1F,$EE,$09
	.dc.b	$D5,$73,$9F,$F4,$35,$F3,$77,$1F,$E8,$99,$DC,$60,$DE,$1D,$AF,$A6
	.dc.b	$3E,$B7,$D3,$7D,$69,$87,$59,$DE,$3C,$6B,$D7,$85,$B6,$3E,$0B,$E7
	.dc.b	$1E,$13,$9C,$FC,$5A,$ED,$87,$57,$99,$F1,$31,$F2,$38,$AE,$53,$B7
	.dc.b	$BC,$E9,$F6,$CF,$4A,$ED,$1C,$1F,$36,$1E,$59,$E9,$87,$98,$8F,$C9
	.dc.b	$BE,$97,$E8,$8E,$87,$97,$1F,$81,$19,$3E,$75,$F5,$7C,$67,$86,$78
	.dc.b	$1C,$9D,$D7,$6C,$7D,$91,$CF,$0F,$1B,$EE,$C3,$BC,$E8,$F8,$58,$77
	.dc.b	$9E,$18,$7B,$08,$F1,$7C,$17,$D2,$78,$57,$3C,$7E,$2E,$3E,$37,$B9
	.dc.b	$F2,$AB,$C7,$87,$63,$DA,$FE,$13,$0E,$A7,$83,$0E,$B5,$CA,$3E,$0D
	.dc.b	$7C,$33,$E0,$E3,$96,$C7,$27,$AB,$F4,$E3,$57,$E9,$C3,$CD,$3B,$C3
	.dc.b	$BB,$63,$37,$DE,$BC,$CD,$9F,$34,$EA,$7D,$B8,$FC,$C8,$C4,$FC,$AA
	.dc.b	$F8,$17,$D9,$E4,$7C,$D2,$9A,$9E,$2F,$A8,$91,$C6,$FB,$E3,$F3,$6F
	.dc.b	$B6,$1F,$4D,$F8,$F3,$A7,$F8,$05,$71,$77,$5C,$6F,$B1,$F5,$4F,$16
	.dc.b	$3F,$22,$7D,$ED,$8F,$1E,$3E,$67,$86,$3F,$06,$75,$1C,$AB,$F8,$87
	.dc.b	$22,$1F,$47,$FE,$01,$6C,$FF,$83,$DF,$F5,$C3,$63,$44,$75,$E8,$59
	.dc.b	$FF,$07,$8F,$A3,$E9,$F1,$BD,$5B,$5D,$5B,$3C,$1B,$B7,$FA,$0F,$E8
	.dc.b	$BF,$D2,$35,$E7,$8C,$7B,$3F,$E1,$FF,$A8,$7D,$4B,$E5,$BF,$C0,$58
	.dc.b	$70,$CB,$FD,$E0,$DC,$38,$E3,$CC,$F8,$58,$72,$BE,$38,$C9,$D6,$1F
	.dc.b	$1F,$0E,$51,$C3,$1F,$03,$A9,$4E,$9F,$F1,$98,$F9,$6F,$9D,$E7,$7C
	.dc.b	$F6,$D7,$CF,$6C,$FC,$A6,$E7,$FE,$46,$87,$C5,$FD,$73,$EC,$7C,$B5
	.dc.b	$BF,$B1,$16,$7E,$9C,$DA,$FF,$57,$77,$7F,$95,$9B,$E4,$7F,$64,$F6
	.dc.b	$BF,$C5,$53,$9C,$7F,$C6,$6D,$C3,$F1,$0D,$EA,$FE,$E1,$5C,$2E,$5F
	.dc.b	$C3,$AD,$74,$FF,$AC,$A3,$13,$F8,$FC,$3F,$21,$85,$9F,$C8,$36,$3F
	.dc.b	$16,$DC,$7E,$13,$FE,$5F,$97,$CF,$3E,$8F,$0D,$8F,$B2,$6B,$E7,$F3
	.dc.b	$32,$33,$3B,$C1,$82,$7F,$4C,$71,$37,$5F,$1B,$E1,$FF,$D0,$7A,$1A
	.dc.b	$49,$FF,$2D,$DF,$CA,$8E,$11,$FD,$97,$C7,$B3,$E0,$8D,$1F,$26,$7F
	.dc.b	$C2,$F3,$AC,$3C,$95,$E5,$C5,$F4,$7F,$03,$E8,$FE,$0B,$71,$F6,$B6
	.dc.b	$BB,$1C,$5C,$5F,$49,$F9,$7F,$47,$F2,$8F,$97,$D1,$63,$E3,$C5,$BE
	.dc.b	$71,$F4,$A7,$15,$FE,$FB,$FA,$37,$C7,$FF,$AA,$1B,$5F,$12,$FF,$89
	.dc.b	$46,$1C,$35,$F9,$2E,$6B,$A5,$7F,$8E,$BD,$5F,$F4,$04,$E6,$FF,$D1
	.dc.b	$2D,$BE,$B5,$B9,$FD,$B4,$E6,$5F,$42,$E2,$FF,$45,$71,$BD,$AF,$F4
	.dc.b	$64,$E7,$0F,$DF,$4E,$5F,$E0,$1F,$44,$F6,$37,$87,$FA,$4E,$31,$3F
	.dc.b	$C1,$6F,$96,$3F,$48,$7D,$1B,$D8,$E1,$8F,$FA,$6E,$31,$7D,$A3,$C7
	.dc.b	$EB,$48,$FE,$21,$F4,$7D,$B0,$E9,$8F,$FA,$92,$73,$7B,$9F,$05,$6F
	.dc.b	$87,$F1,$71,$F0,$B0,$FF,$55,$4E,$6F,$FD,$66,$D9,$7D,$1B,$E8,$FF
	.dc.b	$90,$C7,$4C,$7F,$D6,$5E,$63,$CA,$A7,$ED,$FE,$87,$F9,$16,$5F,$C9
	.dc.b	$E7,$4D,$9D,$BF,$B8,$9D,$DF,$C7,$AD,$B6,$B6,$FE,$4A,$77,$A7,$B2
	.dc.b	$B6,$FB,$AD,$F3,$9F,$EC,$76,$9D,$FC,$4C,$44,$8F,$E7,$86,$B0,$ED
	.dc.b	$39,$7E,$D4,$66,$BB,$57,$BE,$B5,$1F,$3B,$E8,$DF,$47,$FE,$D7,$6D
	.dc.b	$FB,$FE,$47,$13,$23,$FA,$64,$EB,$1F,$75,$65,$FC,$54,$66,$FB,$5F
	.dc.b	$CF,$AD,$47,$DE,$F8,$3F,$E7,$56,$7F,$9E,$60,$D3,$FD,$02,$FF,$5B
	.dc.b	$8D,$93,$C6,$FB,$5D,$3F,$EE,$F8,$C4,$EE,$72,$D3,$FB,$27,$B7,$F4
	.dc.b	$9F,$F7,$EB,$70,$FF,$47,$FD,$1F,$FB,$DA,$31,$06,$9B,$C1,$3B,$E1
	.dc.b	$FF,$80,$13,$A7,$FD,$87,$19,$BF,$0F,$CD,$68,$FF,$63,$FD,$13,$E8
	.dc.b	$F1,$3A,$26,$5F,$C9,$B2,$56,$E7,$3B,$7E,$A3,$37,$D1,$F8,$8F,$E6
	.dc.b	$B1,$FA,$CC,$3F,$C5,$DF,$47,$FE,$2F,$FA,$3F,$EF,$31,$FE,$31,$BE
	.dc.b	$38,$7C,$6C,$39,$3F,$E0,$B9,$E2,$2B,$8B,$F1,$67,$BC,$FF,$2B,$BE
	.dc.b	$DE,$8F,$CE,$9F,$E8,$F7,$C2,$F9,$BC,$B0,$F0,$C7,$5A,$E4,$D3,$CD
	.dc.b	$5C,$A3,$35,$C6,$31,$3F,$D7,$AF,$C2,$F0,$C2,$BF,$E4,$F1,$18,$C3
	.dc.b	$C3,$7E,$59,$C9,$E1,$3D,$D5,$C9,$F4,$E1,$ED,$8C,$47,$09,$F2,$46
	.dc.b	$2F,$B6,$1E,$6A,$F1,$46,$27,$51,$DF,$0F,$1E,$1E,$E3,$B4,$78,$CF
	.dc.b	$09,$ED,$30,$7A,$0F,$EB,$85,$3E,$59,$8A,$CF,$D1,$FF,$12,$F8,$0B
	.dc.b	$D2,$69,$2B,$FB,$34,$BE,$8B,$FC,$2B,$18,$6B,$FE,$0E,$8D,$6B,$DA
	.dc.b	$73,$3F,$46,$72,$FC,$DF,$A3,$7D,$07,$D5,$DF,$F1,$B1,$89,$FB,$35
	.dc.b	$C2,$7F,$22,$DD,$3E,$E5,$6E,$BA,$5F,$39,$F0,$61,$E8,$C3,$8B,$63
	.dc.b	$8C,$EE,$BF,$D2,$97,$CE,$F7,$8F,$8B,$0E,$15,$AB,$E0,$AF,$86,$1E
	.dc.b	$0B,$E0,$7E,$4A,$33,$3C,$4C,$9F,$8F,$C7,$D8,$E4,$F5,$9A,$C7,$C6
	.dc.b	$E0,$8E,$0F,$93,$87,$E7,$EF,$FB,$6D,$FC,$56,$C6,$76,$F3,$5F,$39
	.dc.b	$DF,$85,$98,$F8,$6D,$7D,$D5,$7E,$8D,$D1,$9F,$A2,$7C,$25,$35,$12
	.dc.b	$FF,$D2,$D3,$A6,$CE,$5F,$F0,$A5,$3E,$8B,$A7,$B9,$27,$FD,$3D,$39
	.dc.b	$6C,$E5,$FF,$0E,$5A,$BC,$D3,$27,$C3,$1B,$CF,$FC,$23,$7C,$E7,$FC
	.dc.b	$49,$F4,$6E,$BF,$67,$A1,$27,$FD,$5D,$39,$6C,$E5,$F0,$CF,$C7,$9D
	.dc.b	$47,$49,$D5,$FF,$88,$32,$FF,$2A,$CF,$23,$FD,$75,$EC,$3E,$63,$E6
	.dc.b	$7C,$24,$9C,$C7,$D3,$9D,$47,$CC,$9D,$3F,$E3,$D8,$CB,$67,$37,$E1
	.dc.b	$AE,$51,$F7,$7E,$33,$FC,$A5,$5F,$EC,$BE,$17,$C7,$7E,$05,$B1,$FE
	.dc.b	$CF,$9D,$BF,$E4,$B6,$E1,$FE,$3C,$9D,$CF,$F9,$3D,$B1,$FE,$40,$9D
	.dc.b	$F2,$7F,$35,$6D,$0F,$E0,$B7,$E1,$3C,$DE,$9F,$E4,$F9,$7E,$C6,$1F
	.dc.b	$E4,$AF,$A2,$7D,$1F,$E5,$DB,$87,$FB,$96,$B7,$97,$F4,$36,$C7,$79
	.dc.b	$DB,$FE,$65,$6D,$FF,$95,$A7,$72,$6C,$EA,$BF,$A7,$36,$7D,$53,$C3
	.dc.b	$E8,$DE,$57,$EA,$FE,$89,$ED,$D3,$F5,$D3,$AB,$FF,$7E,$7D,$1B,$8B
	.dc.b	$F6,$4D,$8C,$C7,$16,$9F,$27,$4F,$F3,$6F,$CB,$3E,$8F,$F7,$4D,$9D
	.dc.b	$9F,$E8,$3F,$A2,$7C,$47,$F4,$A8,$46,$1F,$E9,$AD,$7F,$D5,$4D,$7F
	.dc.b	$D9,$FF,$40,$FA,$2E,$4D,$8F,$EA,$AD,$7F,$D6,$4D,$7F,$DA,$FF,$43
	.dc.b	$FE,$1C,$7F,$AD,$B5,$FF,$5C,$35,$DF,$9B,$E7,$FA,$07,$D0,$FF,$17
	.dc.b	$6C,$FC,$3E,$6F,$F0,$14,$62,$3F,$D4,$5E,$1F,$CF,$7F,$D4,$91,$97
	.dc.b	$FD,$D2,$D7,$8D,$77,$23,$D4,$FC,$C4,$FF,$1D,$7A,$F8,$BE,$39,$FE
	.dc.b	$55,$3F,$EB,$D9,$E0,$7F,$17,$FA,$1C,$9A,$F8,$49,$1A,$94,$EA,$31
	.dc.b	$3D,$3E,$6F,$AF,$D0,$C9,$E1,$19,$79,$46,$5F,$83,$1F,$CF,$7E,$A7
	.dc.b	$F9,$44,$7F,$82,$9A,$FF,$B9,$DA,$FF,$8A,$C3,$87,$FC,$5D,$F4,$4F
	.dc.b	$8A,$FF,$3F,$47,$FB,$BE,$31,$EE,$7F,$85,$97,$CE,$BD,$AC,$5F,$0B
	.dc.b	$ED,$F4,$5F,$EC,$D6,$FC,$27,$F9,$75,$FE,$0D,$19,$7F,$C7,$CD,$5F
	.dc.b	$09,$FE,$1D,$F9,$CF,$F2,$BC,$63,$DA,$FE,$B3,$CD,$F6,$6B,$4D,$8C
	.dc.b	$CB,$BD,$3F,$C5,$6B,$73,$98,$C5,$70,$AD,$CF,$7B,$F0,$DF,$F8,$BF
	.dc.b	$D5,$B6,$EF,$BF,$17,$86,$E5,$E3,$6E,$3D,$CF,$53,$8C,$BF,$DD,$7F
	.dc.b	$45,$FC,$E2,$BF,$29,$BB,$EA,$1F,$53,$F0,$B1,$5E,$DF,$F4,$84,$EA
	.dc.b	$FF,$2B,$39,$9F,$C0,$35,$FF,$36,$35,$FE,$70,$D7,$AF,$9F,$DE,$3B
	.dc.b	$63,$F1,$3E,$05,$E8,$FF,$2B,$C3,$BD,$FC,$4F,$A4,$F8,$AD,$8F,$A8
	.dc.b	$DC,$7E,$17,$0F,$F9,$66,$31,$8F,$D0,$8C,$57,$DB,$F9,$6F,$9E,$7D
	.dc.b	$0F,$A4,$B1,$92,$61,$D3,$89,$3A,$8F,$A0,$3E,$89,$DB,$F8,$CC,$9F
	.dc.b	$52,$E7,$F9,$D5,$AC,$BF,$83,$56,$A7,$37,$F9,$BC,$3A,$36,$F8,$47
	.dc.b	$EB,$CC,$DF,$F9,$FA,$33,$7F,$8E,$9C,$9F,$85,$E0,$F9,$BF,$35,$DA
	.dc.b	$FF,$39,$F4,$BF,$90,$3F,$23,$DB,$63,$24,$BF,$85,$19,$7E,$24,$7E
	.dc.b	$63,$E9,$7F,$32,$D3,$F5,$05,$8D,$BF,$AB,$8F,$87,$E3,$FE,$CB,$A5
	.dc.b	$F3,$DF,$A7,$AF,$F5,$B6,$BF,$E4,$A7,$F3,$31,$9C,$3F,$C0,$93,$97
	.dc.b	$FD,$CD,$19,$7E,$A5,$F5,$8C,$4E,$F5,$FD,$9E,$1E,$0E,$DF,$1E,$3D
	.dc.b	$9B,$8F,$46,$D7,$0C,$39,$E3,$D6,$73,$7F,$E1,$7C,$3A,$3F,$CF,$31
	.dc.b	$F0,$39,$BE,$2F,$73,$57,$CD,$A6,$71,$FC,$7D,$72,$7F,$93,$57,$27
	.dc.b	$CD,$3C,$1A,$70,$AE,$8D,$8E,$98,$FE,$1E,$FA,$61,$E8,$AD,$1C,$A7
	.dc.b	$57,$C9,$B7,$C8,$FA,$F1,$96,$99,$9F,$AF,$19,$6C,$65,$FA,$ED,$BF
	.dc.b	$1E,$1E,$DB,$FA,$18,$FC,$EB,$F8,$2D,$C7,$D1,$3E,$49,$F4,$61,$E2
	.dc.b	$69,$E0,$29,$E0,$2E,$1E,$0B,$E9,$87,$86,$BE,$C4,$60,$8F,$CF,$BF
	.dc.b	$2B,$4D,$96,$FC,$ED,$C3,$CE,$F9,$67,$8C,$FD,$06,$9E,$B2,$C7,$AE
	.dc.b	$3F,$B8,$61,$FA,$16,$E3,$FC,$1C,$93,$FC,$15,$B1,$FC,$26,$78,$3E
	.dc.b	$19,$8F,$E3,$38,$3E,$73,$F1,$6F,$C5,$3F,$99,$9C,$D7,$E4,$EB,$53
	.dc.b	$93,$E0,$38,$3E,$85,$7E,$A6,$B8,$DF,$CE,$6D,$7C,$5C,$BF,$94,$3C
	.dc.b	$B0,$EC,$7E,$12,$77,$87,$C7,$9F,$03,$4E,$EF,$B3,$1F,$E1,$B3,$C3
	.dc.b	$1F,$5D,$7B,$EB,$B6,$3E,$36,$9E,$3A,$E0,$F3,$C7,$CB,$5D,$CF,$71
	.dc.b	$EF,$C3,$BC,$F2,$8F,$24,$61,$E1,$8F,$B3,$0F,$E2,$D8,$76,$9F,$49
	.dc.b	$E5,$37,$1E,$56,$DF,$94,$ED,$8F,$B3,$0E,$CF,$BA,$7C,$B8,$7C,$9A
	.dc.b	$D9,$EB,$C3,$C1,$3E,$9C,$7D,$B7,$C6,$FC,$D7,$FC,$CA,$FF,$65,$CB
	.dc.b	$FE,$B5,$FA,$5F,$F0,$DE,$26,$FA,$58,$65,$FF,$8B,$67,$4D,$9C,$C7
	.dc.b	$F7,$A1,$F4,$BD,$3E,$43,$AB,$6B,$93,$67,$83,$73,$FE,$C5,$F4,$5F
	.dc.b	$EA,$CA,$FF,$3A,$72,$BE,$A7,$FD,$D9,$1F,$E7,$09,$D6,$1F,$D9,$E7
	.dc.b	$2F,$F7,$E8,$CB,$FE,$5F,$6B,$FE,$D5,$0E,$1F,$F6,$C3,$5C,$FB,$3F
	.dc.b	$E2,$F8,$EF,$87,$F9,$13,$E9,$9F,$4A,$FA,$6D,$5F,$CC,$FA,$6F,$9B
	.dc.b	$E9,$5D,$3E,$53,$73,$F9,$0D,$CB,$FC,$41,$08,$FA,$BF,$21,$2B,$E9
	.dc.b	$7C,$D5,$AF,$A1,$36,$F8,$CD,$AF,$77,$77,$F7,$1E,$9F,$EE,$FF,$53
	.dc.b	$FE,$49,$8C,$75,$23,$1D,$9F,$E1,$B9,$FF,$34,$E7,$FE,$87,$6B,$D2
	.dc.b	$34,$70,$C3,$FB,$FF,$BB,$67,$F0,$87,$E3,$B1,$FC,$7B,$67,$F1,$8D
	.dc.b	$36,$FE,$3E,$3F,$CF,$EB,$E9,$9F,$4F,$86,$C7,$FA,$E4,$D7,$C4,$66
	.dc.b	$64,$66,$77,$7F,$E7,$18,$D9,$99,$F8,$9F,$17,$FE,$11,$F1,$EC,$FF
	.dc.b	$86,$9B,$5F,$E3,$BD,$15,$F1,$C9,$5F,$E3,$F9,$4E,$A3,$FD,$8E,$D8
	.dc.b	$FF,$10,$F9,$3D,$38,$7F,$D1,$F5,$EF,$FA,$6F,$77,$D3,$FB,$DB,$1E
	.dc.b	$B6,$DF,$AF,$44,$62,$F9,$C3,$E9,$FE,$41,$F2,$7A,$2E,$1F,$16,$2E
	.dc.b	$1C,$E6,$DF,$F9,$57,$E9,$DF,$27,$FE,$35,$DB,$FD,$25,$18,$C7,$FC
	.dc.b	$7F,$18,$9F,$F3,$24,$E7,$0F,$F3,$13,$5F,$F6,$93,$5F,$DA,$3B,$BF
	.dc.b	$04,$FF,$90,$FD,$9E,$98,$7F,$BC,$38,$DE,$F7,$FA,$8D,$B1,$FE,$F4
	.dc.b	$FA,$57,$2E,$6B,$FD,$51,$F4,$EF,$73,$FD,$55,$5A,$BF,$F7,$DC,$E5
	.dc.b	$E5,$7C,$2A,$47,$FC,$00,$FA,$77,$B9,$8D,$FF,$D1,$B8,$FF,$A8,$DB
	.dc.b	$F1,$FF,$EB,$7F,$A7,$E9,$18,$C0,$FF,$A5,$23,$3C,$1F,$E8,$3B,$E4
	.dc.b	$FF,$A1,$63,$35,$FC,$66,$72,$FF,$1B,$8C,$D7,$46,$FD,$3B,$E9,$FF
	.dc.b	$8B,$57,$83,$0E,$9A,$7F,$B2,$FD,$07,$9D,$4F,$D8,$FD,$2B,$C8,$C6
	.dc.b	$1C,$4F,$E5,$55,$C1,$C6,$9F,$EE,$DB,$E3,$5C,$23,$FD,$63,$5B,$6C
	.dc.b	$EA,$B8,$C7,$FC,$00,$8F,$F6,$F4,$3E,$86,$9E,$4F,$1E,$C7,$03,$45
	.dc.b	$8D,$34,$95,$FE,$C2,$76,$FF,$03,$C5,$5D,$23,$FD,$8F,$F4,$EF,$A7
	.dc.b	$FF,$78,$36,$FF,$88,$5B,$E5,$B3,$52,$7D,$75,$BB,$ED,$59,$7A,$D5
	.dc.b	$DB,$F8,$0E,$1C,$F0,$E3,$3F,$EF,$FF,$88,$FF,$53,$1F,$EA,$B8,$4B
	.dc.b	$AB,$5E,$4C,$9F,$97,$5B,$AD,$46,$27,$58,$FF,$B9,$2B,$57,$8C,$7F
	.dc.b	$DA,$9E,$9F,$D7,$6E,$FF,$EC,$1F,$A7,$7C,$0B,$3F,$B3,$7B,$7F,$B7
	.dc.b	$37,$2F,$CF,$B7,$15,$F2,$AE,$38,$73,$BF,$F7,$BD,$F1,$4F,$1A,$E1
	.dc.b	$1A,$9E,$18,$FA,$3E,$76,$E1,$E2,$7F,$DB,$1F,$4C,$FA,$7C,$E1,$E4
	.dc.b	$9A,$F6,$6E,$47,$8C,$E1,$CF,$C9,$4C,$3F,$33,$7F,$BC,$FA,$7F,$DE
	.dc.b	$FD,$2B,$D5,$F8,$77,$BA,$F6,$BE,$07,$80,$9E,$0F,$C8,$8D,$2A,$EB
	.dc.b	$5D,$2E,$CF,$F8,$74,$D1,$CE,$34,$D3,$53,$C1,$F9,$B5,$F4,$6F,$C3
	.dc.b	$1F,$E2,$79,$D6,$4A,$D5,$F7,$AE,$F7,$C2,$B5,$87,$59,$F6,$CE,$6F
	.dc.b	$8C,$78,$8F,$21,$E2,$9E,$0F,$4B,$F4,$D7,$92,$39,$C6,$0F,$16,$1E
	.dc.b	$6B,$EB,$3E,$93,$AE,$1E,$47,$B9,$C8,$F9,$D8,$78,$2B,$93,$63,$8D
	.dc.b	$74,$7E,$B9,$FD,$FE,$7F,$C0,$0E,$0F,$19,$F1,$7E,$9B,$F8,$D7,$C4
	.dc.b	$7F,$81,$5B,$19,$EA,$4E,$5B,$1C,$7E,$9D,$C1,$C5,$B1,$98,$C1,$1F
	.dc.b	$97,$F4,$EF,$A2,$FF,$30,$6F,$F7,$23,$13,$F8,$1C,$98,$74,$AF,$C0
	.dc.b	$C6,$30,$FC,$35,$F0,$BE,$B8,$74,$DF,$FC,$D9,$7B,$9F,$B6,$DC,$3C
	.dc.b	$15,$B9,$FE,$C3,$7C,$E7,$BE,$3E,$1C,$39,$B6,$B7,$7C,$55,$F1,$AD
	.dc.b	$E3,$EB,$BE,$15,$AB,$E1,$7B,$8F,$EE,$38,$F9,$70,$E3,$1E,$C3,$DF
	.dc.b	$5D,$E7,$51,$C2,$FA,$DF,$38,$EC,$78,$70,$E2,$FF,$5C,$8F,$91,$1F
	.dc.b	$C7,$63,$67,$DC,$DB,$C3,$97,$C6,$8E,$1E,$2F,$FA,$5E,$B8,$5A,$31
	.dc.b	$7F,$E9,$B8,$EA,$FF,$A7,$7E,$99,$D5,$AC,$3F,$C1,$FF,$4A,$F8,$8C
	.dc.b	$42,$30,$FF,$72,$AE,$2D,$AD,$CE,$A3,$D0,$74,$AE,$75,$C2,$32,$17
	.dc.b	$D3,$F2,$F7,$E4,$F7,$8E,$26,$1E,$37,$FE,$A3,$74,$43,$33,$BD,$3E
	.dc.b	$65,$F6,$9F,$F0,$5F,$D3,$FF,$39,$7F,$C1,$B7,$FE,$5A,$6F,$A5,$F0
	.dc.b	$A7,$51,$E8,$6D,$FD,$7A,$E1,$3E,$4A,$E1,$D3,$23,$99,$0E,$4F,$CE
	.dc.b	$F5,$1F,$4F,$F2,$BE,$22,$57,$F8,$82,$72,$D9,$CB,$F0,$A7,$94,$48
	.dc.b	$F5,$C6,$B1,$FC,$65,$72,$9F,$A3,$F2,$5F,$E6,$27,$FC,$5B,$CF,$FD
	.dc.b	$E1,$BE,$1F,$E0,$A7,$51,$C3,$D3,$FC,$2D,$6E,$3E,$DB,$70,$FC,$D5
	.dc.b	$70,$6D,$70,$B9,$7F,$E3,$CA,$D7,$27,$FB,$C7,$9B,$EB,$4F,$E1,$BE
	.dc.b	$99,$F4,$FF,$94,$6C,$7F,$91,$27,$79,$7E,$A5,$B7,$FE,$FB,$AD,$D7
	.dc.b	$3B,$FC,$ED,$6A,$3F,$CD,$ED,$53,$C2,$FF,$68,$DB,$F3,$72,$FE,$9B
	.dc.b	$E9,$BF,$C7,$D7,$FE,$56,$FA,$77,$B9,$FA,$A6,$C7,$F9,$6A,$76,$FE
	.dc.b	$8A,$32,$FE,$93,$E9,$DC,$5F,$B2,$6C,$75,$BC,$4F,$ED,$BE,$6D,$F4
	.dc.b	$F4,$CD,$7F,$9A,$3E,$9D,$F1,$3F,$BC,$6C,$66,$31,$F2,$7F,$C0,$3E
	.dc.b	$99,$F4,$F4,$95,$CF,$EA,$7F,$82,$FD,$47,$F7,$F8,$FF,$3A,$B5,$FD
	.dc.b	$EB,$79,$3F,$8E,$7D,$4B,$EA,$3F,$86,$FD,$4D,$BF,$F3,$EB,$5F,$E0
	.dc.b	$AD,$AF,$E4,$9E,$2F,$C8,$7F,$93,$46,$79,$FF,$D4,$F4,$8C,$3F,$EA
	.dc.b	$A6,$AA,$E5,$3C,$27,$C1,$3F,$E2,$56,$B9,$BF,$EF,$31,$88,$FF,$5A
	.dc.b	$D7,$FB,$39,$FF,$4A,$AF,$A9,$FE,$52,$D8,$CF,$C4,$60,$CC,$C8,$C9
	.dc.b	$CB,$E8,$38,$FD,$4C,$8C,$10,$FE,$53,$68,$D4,$62,$7F,$9C,$FD,$4F
	.dc.b	$F1,$88,$FF,$20,$35,$F8,$ED,$7F,$D9,$ED,$4F,$F2,$D7,$FC,$74,$D7
	.dc.b	$9D,$7F,$87,$23,$11,$F4,$BE,$A5,$F1,$AF,$7B,$FC,$85,$29,$CC,$61
	.dc.b	$7D,$17,$FA,$5E,$7F,$C9,$11,$A7,$FC,$B1,$F0,$BF,$E2,$2E,$7F,$F3
	.dc.b	$AB,$FE,$24,$F4,$6F,$B0,$9B,$18,$6F,$1F,$FB,$4A,$B9,$36,$BF,$B1
	.dc.b	$B7,$B7,$9F,$C6,$74,$6C,$64,$E1,$19,$69,$FE,$D6,$FA,$6F,$F4,$8F
	.dc.b	$3F,$FA,$5F,$E0,$9F,$4E,$FA,$8F,$F4,$5D,$6A,$FF,$DF,$F5,$A9,$CF
	.dc.b	$9B,$FE,$92,$8C,$4F,$58,$C5,$F3,$8C,$1F,$EE,$E9,$DE,$FF,$DB,$B0
	.dc.b	$EC,$FF,$B0,$EB,$83,$FD,$DB,$5F,$F2,$6F,$D4,$7C,$AC,$7E,$5B,$71
	.dc.b	$F8,$CD,$AE,$1B,$23,$1A,$7F,$AD,$70,$FF,$2B,$FC,$C7,$D4,$3E,$AF
	.dc.b	$CD,$5C,$7E,$84,$5C,$79,$CD,$BF,$F1,$17,$D2,$BA,$BF,$D6,$70,$FA
	.dc.b	$8F,$BD,$8B,$0F,$F0,$5D,$EE,$31,$5F,$EC,$0A,$DD,$6A,$F9,$C6,$27
	.dc.b	$BC,$63,$93,$FD,$7F,$87,$3B,$FF,$0D,$D7,$86,$73,$5B,$C3,$F1,$5F
	.dc.b	$51,$F9,$23,$F2,$7D,$16,$32,$4C,$3E,$4C,$65,$F9,$71,$F9,$AF,$A8
	.dc.b	$FC,$DB,$4F,$CB,$16,$36,$FE,$66,$3C,$FE,$F7,$E1,$FE,$87,$F3,$75
	.dc.b	$F1,$23,$1D,$5A,$BF,$C3,$56,$B4,$F9,$F5,$A9,$F9,$27,$F2,$CD,$7F
	.dc.b	$C8,$8F,$0B,$E5,$87,$F8,$FB,$9D,$8F,$3A,$E1,$87,$3A,$F1,$47,$13
	.dc.b	$AB,$F6,$70,$EA,$F3,$AE,$4E,$6B,$96,$3E,$6B,$E5,$3D,$0E,$18,$F9
	.dc.b	$27,$F6,$17,$CA,$58,$76,$9C,$D6,$EF,$F6,$37,$BA,$D4,$62,$75,$6C
	.dc.b	$7D,$F1,$86,$CF,$B2,$7B,$E3,$E8,$8F,$86,$7A,$31,$F5,$4F,$A2,$38
	.dc.b	$61,$E0,$8E,$D8,$F8,$EF,$94,$7C,$3A,$E4,$74,$79,$1F,$0E,$FD,$B1
	.dc.b	$D2,$78,$CF,$92,$38,$BC,$CE,$33,$E4,$38,$BC,$E3,$E1,$1C,$CE,$2F
	.dc.b	$33,$8B,$E3,$8E,$27,$37,$8C,$73,$3D,$B8,$FC,$9B,$F7,$4F,$4C,$3E
	.dc.b	$16,$3F,$32,$F8,$37,$0E,$91,$CA,$BC,$8F,$88,$F6,$47,$8A,$3C,$87
	.dc.b	$89,$E3,$3C,$E3,$33,$CC,$EF,$7E,$47,$A1,$EF,$AE,$8F,$69,$E8,$72
	.dc.b	$AE,$91,$93,$A4,$F8,$2B,$A3,$CA,$3A,$57,$27,$85,$72,$72,$72,$8E
	.dc.b	$D5,$C9,$E1,$1A,$9E,$17,$F2,$AB,$53,$C4,$EA,$E2,$77,$8F,$E1,$AB
	.dc.b	$86,$3F,$54,$98,$7C,$EC,$7D,$13,$A8,$C1,$E2,$71,$8F,$CA,$AE,$2D
	.dc.b	$9D,$BE,$33,$D1,$8F,$8C,$93,$C2,$FE,$43,$6B,$E0,$B6,$BD,$6D,$79
	.dc.b	$B6,$FC,$75,$CA,$BC,$38,$FC,$07,$F2,$18,$FC,$12,$38,$26,$1F,$82
	.dc.b	$BE,$71,$E2,$39,$E3,$FC,$63,$8B,$FA,$A7,$D4,$7F,$85,$AC,$DF,$3B
	.dc.b	$0C,$BF,$F0,$A4,$E9,$B3,$98,$FE,$EC,$3E,$A3,$A7,$C6,$F5,$6D,$72
	.dc.b	$6C,$F0,$6E,$A9,$FF,$33,$FD,$3F,$FA,$59,$FE,$67,$F1,$8F,$A8,$FF
	.dc.b	$69,$C7,$F1,$B9,$D6,$1F,$EF,$CA,$D7,$C0,$73,$8F,$F7,$DC,$EB,$0C
	.dc.b	$56,$AF,$97,$AB,$B9,$CE,$3D,$B0,$E3,$39,$AF,$EF,$97,$C6,$73,$1E
	.dc.b	$6A,$E1,$7F,$E0,$F9,$7C,$B7,$CF,$F2,$7E,$83,$6F,$E6,$B6,$7E,$5B
	.dc.b	$6B,$97,$C5,$FD,$73,$EC,$7C,$B5,$BF,$AB,$16,$FE,$9C,$DA,$F8,$50
	.dc.b	$FA,$3F,$EB,$3C,$FF,$D9,$3E,$97,$D9,$3F,$E1,$79,$CF,$32,$73,$3C
	.dc.b	$7A,$7F,$BD,$46,$2B,$6D,$8E,$F3,$FE,$A0,$9F,$EF,$EF,$FA,$88,$E5
	.dc.b	$9F,$F9,$C2,$33,$1F,$E2,$7E,$7F,$F5,$3C,$FE,$43,$1F,$CB,$4F,$E5
	.dc.b	$DB,$19,$3F,$2B,$19,$7F,$BC,$8F,$A2,$7D,$2F,$BA,$30,$D8,$7C,$2D
	.dc.b	$89,$3C,$A7,$8B,$67,$34,$D7,$E6,$E4,$F8,$7F,$F0,$F7,$89,$F1,$5B
	.dc.b	$87,$F7,$19,$D3,$FE,$1E,$9D,$34,$D4,$FC,$AB,$FA,$8D,$7E,$D3,$6B
	.dc.b	$9C,$7C,$D9,$F8,$1F,$49,$ED,$FA,$5F,$73,$63,$D8,$D9,$F5,$47,$F9
	.dc.b	$5C,$E2,$7C,$19,$F8,$DF,$4B,$23,$E2,$74,$DA,$E2,$7E,$E7,$55,$67
	.dc.b	$4F,$D9,$5F,$D3,$BC,$3F,$EE,$4F,$A5,$7C,$6F,$F9,$17,$5F,$C7,$C6
	.dc.b	$27,$FC,$87,$18,$9D,$47,$F8,$6E,$75,$97,$F8,$D6,$B5,$39,$AF,$F2
	.dc.b	$87,$AB,$8A,$F3,$9F,$E2,$7F,$67,$F3,$BC,$67,$F3,$FF,$47,$F8,$4C
	.dc.b	$3F,$C5,$13,$97,$FC,$55,$19,$AE,$6D,$D3,$F5,$FF,$4A,$FA,$1E,$7C
	.dc.b	$BF,$AA,$8C,$61,$FE,$33,$6E,$1E,$4F,$A5,$7A,$FB,$C3,$C7,$F1,$B3
	.dc.b	$0F,$F1,$D7,$D2,$F0,$CF,$F7,$BA,$2B,$F6,$F3,$9A,$FF,$49,$64,$AD
	.dc.b	$47,$2C,$3A,$4F,$C4,$AE,$0F,$FA,$65,$BF,$4A,$FA,$57,$D3,$76,$BF
	.dc.b	$E0,$BD,$CF,$2E,$9F,$AF,$FA,$3F,$E4,$39,$FF,$94,$67,$4D,$9D,$BF
	.dc.b	$B8,$9D,$E7,$F5,$AB,$6D,$AD,$BF,$90,$A5,$70,$DB,$F8,$FD,$F0,$7F
	.dc.b	$90,$CB,$E7,$BF,$8D,$B4,$F0,$78,$B8,$89,$1F,$E6,$73,$58,$77,$C1
	.dc.b	$5A,$8F,$C5,$CE,$AF,$BD,$FC,$1A,$D4,$7F,$32,$FA,$67,$D2,$FF,$2C
	.dc.b	$6D,$FC,$1F,$A6,$C4,$C8,$FF,$3C,$4E,$B1,$F1,$56,$5F,$CD,$C6,$6F
	.dc.b	$B5,$FD,$1A,$D4,$7F,$42,$F8,$4F,$F4,$81,$FE,$9A,$C3,$E4,$F4,$FF
	.dc.b	$3E,$8C,$F8,$F4,$FB,$51,$67,$FD,$15,$3A,$7E,$C4,$67,$2F,$D6,$F9
	.dc.b	$DF,$46,$7F,$22,$D8,$FF,$4B,$7D,$37,$FC,$00,$51,$96,$F0,$FF,$50
	.dc.b	$8E,$16,$AE,$11,$B2,$4F,$F5,$A9,$D7,$C0,$71,$8E,$5F,$37,$31,$FF
	.dc.b	$5E,$D3,$E9,$9F,$4D,$9C,$FD,$37,$76,$FF,$05,$5E,$EB,$18,$7B,$3E
	.dc.b	$9B,$CC,$7F,$66,$C3,$FE,$00,$7D,$37,$FC,$01,$FA,$47,$A7,$BB,$FF
	.dc.b	$59,$6E,$C3,$8C,$7F,$AD,$70,$E2,$2F,$8B,$FE,$10,$9E,$CB,$0E,$D8
	.dc.b	$75,$8E,$0F,$F6,$F7,$46,$0E,$57,$C1,$FE,$E9,$35,$FA,$75,$FE,$29
	.dc.b	$BF,$0C,$7F,$0A,$9D,$61,$E9,$BE,$13,$E8,$AD,$CE,$63,$C0,$71,$2B
	.dc.b	$98,$EB,$1C,$70,$F5,$46,$23,$D2,$F4,$C3,$C9,$3E,$DC,$3C,$71,$8A
	.dc.b	$F5,$9D,$9C,$3E,$1A,$E5,$87,$A2,$3A,$DF,$80,$E8,$F1,$8D,$5F,$66
	.dc.b	$C7,$CD,$3E,$B9,$63,$0E,$0F,$E6,$3F,$4D,$FC,$47,$E1,$F8,$36,$3E
	.dc.b	$B3,$5F,$F1,$33,$5E,$3F,$4D,$FE,$0F,$76,$D7,$6D,$77,$B7,$68,$E1
	.dc.b	$97,$DA,$FA,$67,$D1,$FF,$44,$8C,$37,$4E,$38,$7D,$8C,$58,$72,$9F
	.dc.b	$E9,$78,$72,$AF,$C4,$35,$E1,$3B,$42,$B9,$E3,$C3,$0F,$04,$EA,$F9
	.dc.b	$63,$F9,$0B,$DC,$F4,$BE,$0C,$30,$49,$F5,$B9,$79,$61,$E3,$8F,$13
	.dc.b	$6F,$BE,$3E,$CC,$7C,$75,$AC,$3F,$49,$8F,$8B,$1E,$B3,$F9,$CC,$38
	.dc.b	$BE,$E9,$D6,$1C,$E3,$E0,$DF,$F9,$CA,$BE,$33,$70,$FE,$EA,$FF,$9D
	.dc.b	$AF,$8D,$7C,$39,$CE,$9F,$43,$C5,$B3,$F1,$DA,$B0,$EB,$3F,$E0,$0F
	.dc.b	$A6,$74,$EB,$E9,$5F,$09,$4E,$33,$AC,$3C,$F7,$C6,$F5,$18,$3E,$DD
	.dc.b	$F2,$70,$2B,$93,$FE,$2E,$5F,$45,$E1,$F7,$A4,$F0,$9C,$B6,$72,$FF
	.dc.b	$8D,$A9,$5E,$69,$93,$E0,$9C,$B6,$72,$FF,$8F,$BE,$99,$D3,$F0,$7D
	.dc.b	$29,$3E,$89,$CB,$67,$2F,$86,$7E,$6C,$EA,$3A,$4E,$B8,$7F,$D5,$B8
	.dc.b	$7F,$96,$A7,$91,$F1,$3D,$37,$C1,$7D,$1F,$84,$93,$98,$FB,$33,$A8
	.dc.b	$F6,$4E,$9F,$F2,$BC,$65,$B3,$9B,$F0,$D7,$28,$FC,$1F,$D3,$7F,$A7
	.dc.b	$2B,$EA,$F2,$3C,$1F,$C3,$B6,$3E,$CC,$ED,$FF,$32,$B7,$0F,$F6,$4C
	.dc.b	$EE,$7F,$CD,$2D,$8F,$F6,$6C,$EF,$8B,$FC,$9E,$DA,$FE,$03,$5B,$9F
	.dc.b	$F2,$9B,$63,$F8,$2C,$ED,$FF,$2D,$70,$7E,$73,$E9,$5F,$4D,$4F,$CF
	.dc.b	$F3,$FE,$22,$BF,$CF,$6D,$CB,$FD,$BF,$5C,$27,$F2,$AD,$77,$7C,$8E
	.dc.b	$8E,$AF,$0F,$E7,$3E,$99,$E3,$FE,$BB,$E9,$9E,$E5,$3F,$63,$3A,$AF
	.dc.b	$D1,$CE,$9F,$D2,$FD,$33,$8B,$F6,$ED,$8D,$DE,$27,$F7,$7F,$30,$FA
	.dc.b	$6A,$66,$BF,$5F,$F4,$CF,$A6,$FE,$02,$D8,$CC,$63,$E5,$7F,$82,$FD
	.dc.b	$33,$E9,$A9,$2B,$9F,$D3,$7F,$0A,$FA,$6F,$F0,$A4,$7E,$F1,$AF,$FA
	.dc.b	$A1,$BC,$AE,$9F,$F4,$6F,$D3,$3E,$9B,$F8,$9F,$E6,$7F,$E3,$7C,$3F
	.dc.b	$82,$BF,$E2,$E6,$FA,$3A,$F5,$64,$60,$87,$FA,$DA,$79,$36,$3A,$CC
	.dc.b	$8E,$B1,$FC,$8B,$E9,$FC,$51,$FD,$D5,$AF,$F8,$8D,$AF,$FA,$90,$38
	.dc.b	$7F,$D4,$EF,$0C,$BD,$D1,$FD,$6B,$C7,$FE,$40,$FF,$B2,$9B,$1A,$8E
	.dc.b	$78,$FF,$34,$9F,$F6,$64,$7F,$26,$5F,$4F,$93,$5F,$13,$27,$51,$97
	.dc.b	$84,$73,$F9,$EE,$9F,$4F,$27,$A4,$65,$ED,$19,$7C,$D1,$97,$D1,$19
	.dc.b	$C7,$D1,$82,$BF,$A5,$FD,$33,$E3,$5E,$EF,$F8,$7E,$30,$DF,$9F,$7B
	.dc.b	$99,$AD,$E7,$FE,$51,$FA,$AF,$F6,$EB,$F3,$DA,$FF,$B5,$1A,$FF,$81
	.dc.b	$9B,$F0,$AF,$6D,$D9,$F4,$B6,$FA,$73,$9F,$57,$1D,$3F,$E0,$E8,$E1
	.dc.b	$58,$C5,$1B,$9F,$EB,$4E,$9B,$87,$F6,$66,$D7,$19,$E1,$4C,$3C,$36
	.dc.b	$F0,$74,$FF,$BE,$30,$E7,$18,$F8,$CC,$B6,$32,$78,$23,$2D,$33,$7D
	.dc.b	$DF,$65,$75,$8C,$1F,$E2,$CF,$A6,$FE,$79,$3F,$E5,$36,$C7,$FA,$05
	.dc.b	$AF,$E7,$DA,$FF,$8D,$03,$FE,$B3,$38,$E1,$FA,$0F,$80,$79,$EF,$3F
	.dc.b	$E6,$5F,$0F,$47,$E0,$FD,$2F,$C2,$6E,$5F,$31,$B3,$E8,$DB,$EE,$46
	.dc.b	$32,$F9,$6F,$CA,$FA,$59,$7F,$53,$E6,$6E,$1C,$76,$56,$30,$FA,$75
	.dc.b	$6F,$E3,$CD,$BF,$C0,$DA,$B1,$0F,$A2,$FD,$05,$FF,$86,$9B,$1F,$3B
	.dc.b	$6F,$83,$F3,$5F,$02,$73,$97,$EF,$DB,$7F,$2D,$B7,$FE,$DC,$DD,$7C
	.dc.b	$7A,$FF,$0F,$F4,$93,$0F,$C8,$F5,$5C,$3D,$D3,$72,$FC,$7C,$DA,$FC
	.dc.b	$19,$F9,$0F,$A4,$FC,$8B,$4F,$CD,$96,$FE,$43,$6B,$F3,$2D,$39,$9F
	.dc.b	$A2,$9E,$6D,$39,$96,$79,$B6,$79,$B4,$E7,$1F,$5E,$79,$B6,$39,$B4
	.dc.b	$E6,$59,$E6,$D8,$E7,$1F,$B0,$9E,$6D,$8E,$6D,$9E,$6D,$39,$96,$39
	.dc.b	$9F,$AC,$9E,$6D,$39,$96,$79,$B6,$79,$B4,$E7,$1F,$8E,$9E,$6D,$8E
	.dc.b	$6D,$39,$96,$79,$B6,$39,$F0,$FF,$21,$AE,$6D,$8E,$6D,$9E,$6D,$AE
	.dc.b	$6D,$9E,$7B,$AF,$9B,$63,$9B,$6F,$9B,$6F,$9B,$67,$9D,$72,$23,$C9
	.dc.b	$B5,$C9,$B5,$C9,$B3,$C8,$E0,$49,$E0,$D9,$E0,$D3,$84,$68,$91,$A6
	.dc.b	$C6,$A3,$04,$70,$48,$FE,$0A,$FE,$8D,$AF,$E6,$DA,$7E,$4C,$B1,$F8
	.dc.b	$B6,$C7,$E0,$DA,$7D,$B2,$DF,$D3,$6D,$FC,$B6,$DF,$C4,$6D,$7B,$DA
	.dc.b	$7B,$0B,$7E,$B6,$C7,$AE,$FF,$B4,$3E,$B6,$BE,$B6,$C7,$AD,$B3,$EB
	.dc.b	$6C,$FA,$DB,$3E,$B6,$9E,$B2,$DF,$AD,$B7,$EB,$6D,$7A,$DB,$5E,$B6
	.dc.b	$DF,$AE,$3F,$9A,$7C,$07,$F2,$F3,$FB,$D7,$63,$E7,$FF,$C0,$78,$7F
	.dc.b	$37,$9D,$34,$FF,$6B,$4E,$FA,$B3,$7F,$D4,$47,$D2,$7F,$7D,$6D,$7F
	.dc.b	$81,$1F,$F3,$B3,$63,$44,$75,$E9,$59,$E9,$F4,$9C,$FE,$33,$A3,$6B
	.dc.b	$A3,$67,$71,$DE,$76,$D3,$65,$8F,$EF,$E3,$E9,$3F,$D2,$31,$FE,$59
	.dc.b	$81,$9F,$5B,$FC,$A4,$FF,$7D,$9D,$BB,$6B,$FE,$61,$73,$B7,$F9,$65
	.dc.b	$B5,$D3,$EA,$9F,$54,$F9,$6F,$F0,$0D,$6B,$0F,$F1,$68,$FA,$5F,$99
	.dc.b	$E5,$7C,$D6,$E3,$F3,$5B,$3F,$4D,$B9,$7F,$8F,$21,$F1,$9F,$54,$FA
	.dc.b	$DF,$29,$6F,$EB,$45,$9F,$B9,$36,$B9,$4D,$AF,$F2,$25,$2B,$0F,$C8
	.dc.b	$EC,$FF,$20,$3F,$2D,$AF,$FA,$15,$AF,$F8,$9D,$AF,$1F,$23,$FC,$47
	.dc.b	$1C,$B0,$FF,$43,$F2,$FE,$5F,$DA,$CD,$7E,$3E,$FF,$21,$37,$1F,$C8
	.dc.b	$36,$3F,$2C,$DB,$F8,$8F,$F9,$6E,$5F,$42,$FA,$4C,$36,$3F,$07,$5F
	.dc.b	$E9,$87,$7F,$41,$AA,$91,$A3,$F1,$27,$47,$9C,$6B,$0F,$F2,$7C,$E4
	.dc.b	$E0,$FF,$B0,$6F,$C0,$E9,$4F,$1B,$FF,$52,$66,$F8,$9F,$CA,$DF,$D1
	.dc.b	$6B,$FE,$26,$6C,$6F,$0F,$CF,$4E,$5F,$0B,$4F,$F3,$1E,$5F,$C1,$7C
	.dc.b	$7F,$A7,$7F,$13,$E9,$3E,$2F,$D2,$5C,$3D,$27,$CA,$71,$87,$A5,$B9
	.dc.b	$7C,$9D,$11,$F3,$BE,$92,$5F,$29,$FD,$13,$F5,$23,$37,$F0,$DF,$AB
	.dc.b	$D3,$8C,$3A,$45,$AF,$A3,$7C,$84,$7F,$9B,$B4,$7A,$D4,$FF,$38,$FD
	.dc.b	$0F,$CE,$BF,$87,$3A,$7E,$2C,$67,$2F,$E1,$2D,$C7,$FD,$AD,$F4,$6F
	.dc.b	$9F,$E7,$C5,$FE,$7C,$8C,$61,$F3,$9B,$87,$F1,$2F,$A3,$FF,$49,$61
	.dc.b	$CB,$0F,$1F,$A5,$30,$FA,$BF,$47,$C2,$BF,$DC,$99,$FF,$A2,$67,$37
	.dc.b	$BC,$7F,$8E,$DE,$E7,$FD,$29,$5B,$9C,$D7,$FA,$42,$B5,$0F,$A3,$7D
	.dc.b	$1F,$F1,$5C,$3A,$9F,$86,$F2,$9E,$27,$FA,$61,$B1,$F8,$2F,$A3,$FE
	.dc.b	$30,$EE,$74,$47,$5E,$05,$C7,$F9,$9D,$F2,$6D,$7E,$D6,$F9,$C6,$2F
	.dc.b	$F2,$E3,$E8,$FF,$D5,$2D,$3C,$5E,$06,$22,$47,$07,$A4,$62,$7F,$A1
	.dc.b	$CF,$16,$CE,$A3,$81,$FA,$6F,$A3,$7D,$1F,$FA,$E1,$B1,$F0,$7E,$4B
	.dc.b	$13,$23,$83,$EB,$8C,$5F,$F5,$2A,$E2,$D9,$D5,$70,$3F,$6B,$F0,$1F
	.dc.b	$C8,$DF,$F6,$73,$65,$0C,$9D,$1F,$EC,$CA,$D5,$1A,$71,$AE,$13,$BF
	.dc.b	$0E,$47,$81,$FA,$51,$8A,$FE,$D3,$EC,$7D,$7E,$6F,$0C,$7F,$68,$8C
	.dc.b	$5F,$F6,$AA,$7C,$37,$FB,$8A,$3F,$DC,$D3,$AA,$E1,$C1,$FE,$E2,$8C
	.dc.b	$5F,$0F,$6F,$13,$2F,$F8,$BD,$6B,$0E,$98,$72,$9C,$FC,$CF,$33,$FC
	.dc.b	$3F,$F4,$6F,$89,$C4,$E8,$95,$FC,$82,$75,$18,$D3,$DB,$2C,$5F,$47
	.dc.b	$E7,$3F,$DF,$31,$FE,$08,$FA,$3F,$F0,$4F,$D1,$7F,$70,$8F,$89,$3B
	.dc.b	$C3,$D2,$F0,$85,$70,$F3,$BE,$2C,$E6,$31,$3C,$AA,$BF,$E1,$63,$88
	.dc.b	$BF,$03,$E7,$AF,$AB,$18,$8F,$31,$BB,$EF,$5D,$23,$33,$CB,$0F,$7C
	.dc.b	$E6,$7B,$CE,$6F,$AD,$FF,$8A,$70,$F2,$E1,$C2,$FB,$1E,$73,$8C,$F8
	.dc.b	$1C,$9E,$B9,$E3,$5D,$23,$84,$63,$0F,$13,$63,$89,$E8,$AE,$98,$7B
	.dc.b	$E3,$B9,$F2,$4D,$98,$3E,$35,$74,$7C,$78,$7A,$67,$95,$FF,$68,$9F
	.dc.b	$92,$7F,$6A,$70,$FD,$23,$DD,$F4,$5F,$C5,$7D,$CF,$ED,$ED,$8F,$F2
	.dc.b	$73,$5D,$F4,$A7,$7F,$46,$D6,$92,$B8,$C6,$5B,$1F,$84,$FA,$37,$CF
	.dc.b	$7E,$17,$3F,$C1,$C6,$2B,$F1,$39,$30,$E5,$5F,$88,$8C,$61,$CF,$0E
	.dc.b	$5B,$7E,$3F,$1E,$58,$FD,$8B,$DB,$71,$E7,$5B,$BE,$33,$FE,$6C,$BE
	.dc.b	$38,$F7,$BD,$CF,$F9,$C3,$0E,$73,$E8,$BE,$32,$BE,$38,$7E,$86,$F6
	.dc.b	$DC,$39,$D6,$EF,$8D,$F0,$8E,$D3,$AB,$E3,$1E,$5A,$D9,$E0,$8C,$C1
	.dc.b	$B5,$CE,$3B,$1F,$32,$BC,$31,$F1,$B1,$FE,$12,$DD,$BE,$25,$7F,$8C
	.dc.b	$2F,$9E,$1F,$BB,$BD,$F8,$B6,$FE,$83,$6F,$CC,$7F,$8E,$3E,$8D,$D3
	.dc.b	$98,$7D,$0B,$DF,$C3,$63,$84,$EE,$B8,$C7,$F9,$0A,$74,$FF,$A9,$9B
	.dc.b	$8F,$F2,$29,$DC,$FF,$AA,$97,$CF,$F6,$F6,$24,$F9,$63,$B1,$87,$A4
	.dc.b	$61,$D1,$E1,$75,$5D,$E7,$D9,$F4,$4E,$AF,$5F,$D1,$CC,$3F,$CB,$33
	.dc.b	$96,$CE,$5F,$54,$F2,$89,$1E,$F8,$D5,$FF,$3E,$AE,$53,$FE,$C6,$F5
	.dc.b	$9F,$47,$F2,$BE,$06,$4F,$F9,$A2,$72,$D9,$CB,$F2,$67,$EA,$4E,$A3
	.dc.b	$A4,$EA,$FF,$A6,$EA,$BD,$CF,$DB,$F8,$9F,$F4,$0D,$FF,$9D,$B8,$DE
	.dc.b	$17,$DF,$6C,$7F,$9E,$A7,$6F,$FB,$79,$B8,$FF,$5A,$9D,$CF,$FB,$89
	.dc.b	$B1,$FD,$7A,$77,$C9,$FD,$35,$B5,$FE,$E4,$AD,$CF,$F5,$16,$C7,$FB
	.dc.b	$A2,$76,$B9,$3F,$0A,$FF,$A3,$BE,$8D,$F4,$7F,$97,$6C,$7F,$A4,$A7
	.dc.b	$6F,$FB,$D9,$B8,$FF,$72,$9D,$CF,$FB,$E9,$B1,$FD,$DA,$77,$8F,$F6
	.dc.b	$56,$D7,$FB,$F2,$B7,$3F,$DA,$5B,$1F,$F0,$02,$76,$BE,$8D,$E5,$FE
	.dc.b	$B7,$E8,$DE,$DD,$3F,$61,$3A,$AF,$F5,$5C,$E9,$FF,$56,$FD,$1F,$DD
	.dc.b	$6F,$0F,$16,$9F,$EB,$2F,$94,$7D,$1F,$EE,$DB,$1A,$3F,$A2,$6B,$FE
	.dc.b	$01,$FA,$37,$C2,$FF,$01,$6C,$FF,$AF,$1A,$F0,$01,$7C,$67,$F0,$9F
	.dc.b	$A3,$7D,$1D,$25,$73,$FA,$3F,$E1,$DF,$47,$FE,$22,$8F,$F6,$4B,$5E
	.dc.b	$C1,$C7,$2F,$F8,$AB,$E8,$DF,$43,$FC,$5F,$E8,$FF,$C8,$B7,$FE,$D2
	.dc.b	$7F,$C6,$CD,$F4,$F5,$EA,$48,$C1,$0E,$F3,$C9,$B1,$D6,$64,$75,$8F
	.dc.b	$E4,$DF,$47,$E2,$8F,$F7,$0B,$5F,$4B,$71,$FF,$20,$F8,$FF,$76,$73
	.dc.b	$8F,$F8,$0E,$B8,$47,$F8,$75,$AF,$F8,$D5,$AF,$D0,$9F,$F6,$8B,$FE
	.dc.b	$F0,$5F,$47,$FC,$E5,$B1,$9F,$88,$C1,$99,$91,$93,$97,$CD,$71,$FA
	.dc.b	$39,$18,$21,$F5,$2D,$1A,$8C,$4F,$84,$C6,$5F,$02,$71,$3F,$88,$8F
	.dc.b	$F3,$1F,$D5,$2F,$FC,$48,$D9,$FF,$1F,$1F,$8E,$70,$E1,$AF,$F9,$0F
	.dc.b	$E9,$1F,$49,$FE,$95,$8E,$31,$8F,$7F,$9C,$21,$F3,$BF,$E9,$E8,$DD
	.dc.b	$71,$7F,$CC,$9F,$0B,$FE,$2B,$E9,$FF,$4A,$E7,$FE,$96,$8F,$F1,$77
	.dc.b	$A7,$8E,$DF,$D1,$C6,$DF,$71,$8F,$6B,$FD,$5D,$19,$C3,$F5,$4F,$29
	.dc.b	$DC,$23,$57,$DE,$B8,$8E,$FF,$05,$6E,$3F,$E0,$0D,$F7,$9C,$DF,$F9
	.dc.b	$0F,$E8,$ED,$71,$AE,$F3,$96,$C6,$6F,$CA,$F4,$7B,$5F,$2B,$E6,$DA
	.dc.b	$F4,$CB,$E8,$FF,$D5,$7F,$06,$FA,$97,$D5,$7A,$2B,$5E,$8F,$AA,$B5
	.dc.b	$39,$BE,$37,$C3,$0E,$B1,$89,$F5,$FD,$2F,$B3,$0B,$5E,$A6,$E3,$EA
	.dc.b	$6E,$DF,$2D,$F8,$7F,$31,$2B,$E0,$FD,$2D,$BF,$83,$36,$FE,$94,$DA
	.dc.b	$FB,$93,$6B,$E9,$4D,$5F,$4A,$EB,$FA,$F8,$3E,$AB,$F9,$3E,$3F,$22
	.dc.b	$B4,$FE,$0A,$75,$CB,$F1,$E7,$38,$7F,$31,$9C,$D7,$1C,$3A,$E1,$D2
	.dc.b	$BE,$CF,$D3,$7E,$00,$FC,$0F,$45,$BF,$BF,$16,$7E,$FC,$DC,$3F,$21
	.dc.b	$5F,$88,$FA,$6F,$C4,$B4,$FC,$39,$6B,$F0,$ED,$9E,$2F,$E3,$27,$8B
	.dc.b	$4E,$25,$9E,$2D,$9E,$31,$F9,$29,$E2,$D9,$E2,$D3,$89,$67,$89,$FA
	.dc.b	$99,$E2,$D9,$E2,$D9,$E2,$D3,$8C,$7E,$E6,$78,$B4,$E2,$59,$E2,$D9
	.dc.b	$E3,$1F,$B5,$9E,$2D,$9E,$2D,$38,$96,$78,$F3,$FF,$AE,$EB,$8B,$6B
	.dc.b	$8B,$6B,$8B,$6B,$8F,$3A,$F8,$B6,$F8,$B6,$F8,$B6,$B8,$D7,$02,$57
	.dc.b	$06,$D7,$06,$CF,$09,$D1,$27,$4D,$8D,$46,$08,$E0,$91,$FC,$A4,$FC
	.dc.b	$61,$4F,$C3,$16,$3F,$00,$D8,$FB,$4D,$3E,$A1,$63,$E7,$36,$FE,$43
	.dc.b	$6F,$E1,$36,$FD,$AD,$3D,$45,$AF,$4B,$6F,$D3,$1F,$CD,$63,$D2,$D7
	.dc.b	$D2,$D3,$D2,$58,$F4,$B6,$3D,$2D,$3D,$25,$BF,$4B,$6F,$D2,$DB,$F4
	.dc.b	$B6,$BD,$2D,$3D,$25,$BF,$4D,$AD,$5E,$A8,$C3,$63,$2F,$B2,$32,$D8
	.dc.b	$CD,$7B,$A7,$2D,$8C,$D7,$C0,$9C,$B6,$33,$5F,$0A,$72,$D8,$CD,$7C
	.dc.b	$49,$CB,$63,$31,$F1,$8C,$95,$C9,$F2,$1C,$B4,$CD,$FC,$A8,$CB,$FC
	.dc.b	$D7,$E0,$7F,$AD,$7D,$37,$FB,$AB,$0F,$E7,$91,$8E,$AD,$47,$F7,$EE
	.dc.b	$85,$6F,$2F,$F2,$FF,$07,$FB,$BE,$B5,$3F,$E0,$89,$7D,$37,$F7,$C6
	.dc.b	$CF,$F7,$A7,$FD,$80,$D8,$D1,$1D,$7A,$36,$7F,$C2,$23,$E9,$BA,$7C
	.dc.b	$7F,$56,$D7,$56,$CF,$06,$F0,$7F,$88,$E3,$94,$F8,$EB,$93,$4F,$F0
	.dc.b	$2F,$D1,$FF,$A8,$1F,$F0,$FF,$C7,$3E,$AB,$FD,$CB,$1A,$F6,$BF,$C3
	.dc.b	$B3,$9F,$81,$FF,$04,$5F,$1C,$7F,$D2,$53,$9B,$E0,$74,$C3,$9C,$FB
	.dc.b	$BE,$5B,$E4,$79,$3F,$25,$B5,$F2,$5B,$3B,$8F,$99,$3F,$3F,$E3,$64
	.dc.b	$7C,$FF,$96,$B5,$F2,$A6,$ED,$F0,$E6,$D7,$F9,$12,$95,$87,$FC,$7C
	.dc.b	$27,$0B,$E8,$FF,$09,$7F,$B0,$87,$2F,$E2,$3E,$AB,$FC,$55,$5A,$C7
	.dc.b	$FD,$67,$39,$7F,$C6,$71,$F8,$DA,$FF,$5B,$9F,$EB,$87,$0E,$1A,$E7
	.dc.b	$9F,$C0,$FF,$98,$8F,$F5,$FC,$FE,$3E,$BF,$25,$1F,$93,$6C,$64,$FC
	.dc.b	$A4,$65,$FF,$2E,$8F,$A2,$7D,$2F,$F9,$6A,$30,$D8,$7D,$15,$89,$3C
	.dc.b	$A7,$8B,$67,$35,$FE,$D5,$8E,$66,$EB,$A4,$EB,$3F,$6E,$4F,$8B,$FD
	.dc.b	$CF,$8B,$F3,$1B,$87,$CA,$7E,$1E,$3F,$1D,$AF,$CB,$6B,$F6,$5C,$C7
	.dc.b	$BF,$E9,$7D,$9F,$4B,$ED,$6C,$7A,$DB,$5E,$9D,$11,$8C,$BF,$05,$5F
	.dc.b	$27,$E9,$7E,$31,$F1,$FA,$6C,$7C,$58,$B8,$73,$9B,$5C,$B0,$56,$2F
	.dc.b	$EE,$C7,$EB,$67,$53,$FE,$77,$DD,$EB,$D3,$FC,$F5,$F4,$7F,$5B,$0F
	.dc.b	$F4,$0C,$E9,$FF,$41,$C6,$72,$FF,$7D,$B7,$0F,$C6,$FD,$2B,$E9,$39
	.dc.b	$F1,$7F,$A1,$63,$18,$7F,$A3,$1B,$82,$FF,$2B,$F4,$BF,$E9,$58,$4E
	.dc.b	$A7,$F3,$38,$BE,$37,$CE,$7F,$A4,$7E,$97,$58,$FE,$92,$74,$FF,$A6
	.dc.b	$A3,$38,$04,$E9,$F1,$4F,$EB,$27,$71,$AA,$ED,$9C,$AF,$F5,$EF,$D2
	.dc.b	$BE,$95,$F4,$D9,$C3,$99,$FE,$CA,$F1,$3F,$D4,$DE,$61,$FD,$37,$D2
	.dc.b	$3C,$2C,$C4,$CB,$FD,$63,$5B,$7F,$5F,$5B,$73,$3B,$BE,$E7,$FB,$3D
	.dc.b	$EB,$79,$DF,$F8,$0D,$3E,$87,$FD,$70,$D3,$AF,$83,$88,$91,$FE,$C2
	.dc.b	$35,$87,$EC,$F1,$FE,$3D,$19,$9D,$CF,$8A,$37,$7F,$9E,$AD,$CE,$6B
	.dc.b	$FD,$91,$F4,$CF,$A6,$A7,$C4,$F9,$5B,$8F,$FB,$4A,$74,$D9,$D3,$49
	.dc.b	$7F,$0A,$76,$F9,$EB,$9B,$F1,$BE,$0B,$F9,$73,$F8,$36,$D3,$1F,$E6
	.dc.b	$AF,$FB,$8A,$37,$26,$9C,$27,$71,$D7,$C4,$FF,$72,$4E,$9F,$BF,$D5
	.dc.b	$87,$FD,$D1,$E4,$FD,$DA,$FB,$2D,$D7,$FA,$35,$3E,$21,$ED,$F0,$E2
	.dc.b	$9A,$7F,$BD,$B0,$E0,$FF,$48,$AD,$B9,$9D,$E3,$D9,$FF,$7B,$4E,$A3
	.dc.b	$13,$CF,$E7,$7C,$A7,$FB,$EF,$E9,$9F,$17,$89,$D1,$2B,$F3,$F9,$AB
	.dc.b	$73,$9D,$3F,$A8,$E6,$FA,$6F,$82,$7F,$63,$8F,$E6,$3F,$4D,$FB,$5F
	.dc.b	$A6,$FE,$DB,$97,$ED,$EF,$86,$1E,$FC,$38,$BF,$BB,$C3,$8D,$2F,$8A
	.dc.b	$BE,$2F,$E9,$8F,$E0,$58,$77,$F3,$FF,$B5,$CE,$AB,$8D,$68,$E9,$3A
	.dc.b	$C7,$FB,$D4,$EA,$57,$CD,$F4,$4E,$AB,$FB,$BC,$73,$AF,$2D,$F7,$9C
	.dc.b	$CF,$23,$B9,$CB,$1E,$8F,$F8,$03,$0B,$5C,$63,$AC,$63,$0F,$04,$60
	.dc.b	$E6,$7C,$37,$2F,$33,$33,$CD,$C6,$1E,$69,$DC,$7A,$70,$F8,$58,$79
	.dc.b	$23,$07,$9C,$EE,$60,$F0,$3B,$7D,$D5,$D6,$31,$87,$C2,$9F,$21,$5C
	.dc.b	$3F,$4C,$F9,$7F,$49,$FC,$6F,$DC,$E0,$D8,$F9,$6D,$77,$D3,$FE,$A1
	.dc.b	$73,$F4,$DF,$D4,$5B,$AF,$28,$C4,$76,$6B,$FE,$30,$A4,$71,$CB,$F1
	.dc.b	$FF,$4C,$FA,$1F,$F1,$AC,$61,$BA,$F1,$AE,$0D,$C7,$EE,$DF,$1C,$6D
	.dc.b	$F0,$7F,$AC,$63,$E0,$AF,$C1,$B6,$3E,$ED,$71,$6D,$71,$53,$BC,$7F
	.dc.b	$4A,$DC,$3F,$C9,$77,$C6,$7D,$15,$DB,$1F,$3B,$E2,$6A,$9E,$15,$CB
	.dc.b	$0F,$06,$1D,$6B,$86,$3E,$C6,$1F,$94,$8D,$3E,$6C,$3A,$E1,$D2,$3F
	.dc.b	$BC,$35,$D5,$72,$3F,$CB,$B1,$F0,$5B,$7F,$E6,$67,$FC,$C3,$7D,$A7
	.dc.b	$5A,$FC,$0F,$13,$4D,$C3,$F8,$63,$EB,$69,$FE,$09,$FA,$67,$4F,$08
	.dc.b	$FF,$06,$7D,$1B,$E0,$31,$08,$C3,$C8,$57,$27,$0D,$7F,$BD,$C7,$06
	.dc.b	$CF,$49,$D6,$7F,$E7,$C5,$F4,$DE,$0B,$7B,$34,$E1,$13,$0F,$14,$E9
	.dc.b	$B3,$97,$FC,$17,$5D,$1C,$90,$CD,$7B,$BE,$95,$D7,$E1,$FA,$69,$5E
	.dc.b	$F9,$CB,$67,$2F,$BA,$79,$44,$8F,$2C,$6B,$2F,$F4,$CD,$72,$9F,$F0
	.dc.b	$FF,$AE,$F7,$BE,$07,$C0,$49,$F9,$D3,$96,$CE,$5F,$9B,$3F,$5A,$75
	.dc.b	$1D,$27,$57,$FE,$A8,$D5,$7B,$9F,$BB,$F4,$DF,$E9,$CB,$FB,$9C,$BF
	.dc.b	$DD,$9B,$E1,$EB,$D4,$91,$82,$1F,$E3,$8A,$E4,$D8,$FC,$4D,$73,$8F
	.dc.b	$F1,$EB,$72,$FF,$5D,$D7,$07,$FD,$7B,$CB,$F5,$EB,$FC,$87,$F4,$AF
	.dc.b	$A6,$FC,$9B,$63,$F2,$53,$BA,$E3,$C5,$AB,$95,$A8,$FC,$DD,$6F,$9F
	.dc.b	$FD,$A3,$87,$3A,$EF,$7A,$BF,$D0,$FD,$33,$CA,$FD,$1F,$D3,$3D,$CA
	.dc.b	$6A,$2E,$1F,$A5,$FA,$6F,$9E,$DE,$3C,$BF,$E5,$E8,$CB,$FE,$60,$F9
	.dc.b	$57,$D3,$7E,$C1,$B1,$C4,$FD,$87,$D2,$BE,$4B,$F6,$AB,$E9,$A9,$88
	.dc.b	$CF,$D2,$7E,$EB,$E9,$9F,$4D,$4C,$E1,$FB,$BF,$A5,$7D,$27,$F0,$0F
	.dc.b	$A6,$FF,$0F,$4F,$EF,$DA,$FF,$9D,$A7,$6F,$F0,$3F,$A6,$FE,$96,$DF
	.dc.b	$2F,$F8,$4C,$62,$3F,$CF,$AF,$F9,$E5,$D1,$0C,$C7,$0F,$3B,$8B,$23
	.dc.b	$5D,$8E,$BF,$F5,$14,$2F,$A3,$DB,$D6,$F0,$FA,$D2,$3C,$D1,$E4,$30
	.dc.b	$F9,$23,$0F,$F1,$DF,$A6,$F6,$BF,$E1,$C6,$BF,$E2,$26,$BE,$26,$BF
	.dc.b	$EB,$57,$8E,$1E,$78,$FF,$5C,$79,$7F,$C8,$E3,$11,$FC,$7B,$CC,$FE
	.dc.b	$5F,$5C,$32,$F6,$DF,$FB,$7A,$BF,$98,$2F,$A6,$FE,$6A,$D8,$CF,$C4
	.dc.b	$E0,$CC,$C8,$C9,$0F,$F5,$7C,$F1,$6C,$71,$7B,$FC,$E7,$6F,$A6,$91
	.dc.b	$82,$1C,$9C,$BD,$23,$FA,$4F,$D2,$BE,$97,$FD,$21,$3F,$E4,$68,$C7
	.dc.b	$BF,$FE,$27,$5F,$3A,$F7,$F1,$7C,$2F,$FC,$A3,$F5,$3F,$EF,$06,$FC
	.dc.b	$5F,$CA,$7F,$07,$19,$6C,$65,$7C,$43,$DE,$FD,$57,$0F,$F6,$48,$C7
	.dc.b	$B5,$FE,$7B,$8F,$F7,$74,$6F,$C1,$C5,$F2,$BF,$F5,$8C,$E6,$B9,$46
	.dc.b	$0E,$B3,$FE,$EA,$9E,$B3,$AA,$E7,$18,$9D,$57,$0E,$6F,$C4,$4E,$5B
	.dc.b	$0F,$13,$8A,$FA,$6B,$18,$6E,$3E,$46,$E1,$D9,$F5,$AF,$A5,$FF,$55
	.dc.b	$61,$F9,$46,$D7,$FA,$67,$D2,$7A,$5C,$27,$36,$39,$7A,$B9,$3E,$3F
	.dc.b	$C1,$AF,$8E,$1F,$E0,$9B,$E3,$AF,$D3,$FA,$4F,$5E,$1E,$C6,$D7,$CC
	.dc.b	$6E,$3F,$19,$B9,$FC,$46,$CF,$D8,$D9,$18,$AF,$B1,$F4,$9F,$20,$F9
	.dc.b	$3F,$31,$6F,$EC,$C5,$9F,$A7,$37,$0F,$DB,$FD,$0B,$AF,$E4,$B7,$0F
	.dc.b	$F1,$6E,$FE,$FF,$9A,$CD,$49,$FF,$14,$56,$F1,$61,$FC,$17,$0E,$15
	.dc.b	$AB,$FB,$BF,$47,$23,$F1,$BD,$77,$0F,$C3,$CD,$C3,$F0,$B3,$6B,$E1
	.dc.b	$4D,$AF,$CC,$4D,$E0,$FD,$14,$DD,$FF,$63,$7F,$9A,$8C,$1F,$8E,$AB
	.dc.b	$19,$3C,$65,$BF,$1B,$70,$F1,$B4,$F1,$96,$3C,$6D,$C3,$C6,$DC,$3C
	.dc.b	$73,$D0,$95,$FA,$2C,$3C,$95,$B7,$D5,$87,$9A,$B7,$3F,$83,$8C,$57
	.dc.b	$18,$F4,$CE,$8F,$DE,$57,$28,$E8,$78,$EB,$8B,$EA,$9F,$DB,$C6,$1E
	.dc.b	$31,$D2,$BB,$47,$5C,$3D,$AF,$09,$EA,$E7,$0F,$7D,$70,$24,$F7,$7E
	.dc.b	$5C,$EF,$0F,$33,$67,$CC,$74,$9E,$31,$F3,$8D,$1D,$F0,$F4,$61,$E4
	.dc.b	$9E,$A7,$07,$15,$E1,$3E,$51,$D4,$C5,$75,$9E,$18,$30,$F2,$C6,$1B
	.dc.b	$5C,$2F,$AD,$FC,$83,$E2,$1E,$6A,$E7,$87,$9B,$0E,$97,$F3,$4E,$4E
	.dc.b	$2B,$D9,$1E,$57,$A5,$7B,$23,$78,$78,$A7,$31,$83,$A3,$82,$61,$E6
	.dc.b	$3F,$A4,$61,$E3,$AF,$C8,$CF,$9E,$BC,$73,$97,$9D,$71,$C3,$CD,$87
	.dc.b	$B2,$7C,$D3,$9A,$E1,$7F,$0A,$BE,$45,$7D,$C8,$EC,$79,$2F,$F0,$F1
	.dc.b	$E9,$8F,$82,$6A,$FA,$47,$69,$F0,$BF,$6A,$75,$18,$7E,$D4,$73,$AF
	.dc.b	$34,$7E,$56,$3F,$A7,$FD,$1F,$F5,$06,$9F,$D8,$0B,$1F,$D6,$9B,$3F
	.dc.b	$CA,$3C,$4F,$E4,$E7,$F6,$AE,$D7,$D1,$FF,$5F,$AF,$F1,$DE,$FF,$ED
	.dc.b	$17,$F9,$3D,$70,$7F,$DB,$35,$C3,$AB,$18,$FF,$6D,$1F,$47,$FD,$B1
	.dc.b	$B5,$FD,$F5,$FF,$50,$B6,$34,$47,$5E,$B5,$9E,$93,$73,$FE,$E7,$1C
	.dc.b	$8C,$D7,$5F,$A3,$F0,$7C,$7F,$85,$B1,$93,$C5,$19,$69,$FE,$EB,$FA
	.dc.b	$3F,$F4,$63,$7D,$BF,$F0,$7F,$D4,$BE,$A5,$F3,$1F,$DE,$2F,$79,$7F
	.dc.b	$4A,$C3,$FC,$C7,$18,$CF,$FC,$9A,$D8,$FE,$ED,$87,$3B,$E5,$8F,$C7
	.dc.b	$F9,$8F,$91,$E6,$DB,$F9,$0D,$CB,$E7,$37,$2F,$86,$D9,$FF,$13,$CF
	.dc.b	$FB,$E1,$FA,$15,$D1,$FA,$5F,$21,$F5,$4F,$AD,$F3,$16,$32,$4A,$FB
	.dc.b	$91,$97,$DC,$7F,$83,$FB,$9E,$47,$F8,$F6,$11,$8A,$FF,$42,$C7,$FA
	.dc.b	$1A,$30,$E0,$DC,$F1,$8C,$E5,$FE,$74,$6D,$7F,$87,$79,$7F,$2F,$EE
	.dc.b	$FF,$8A,$27,$FC,$83,$87,$F8,$86,$BF,$1D,$5F,$91,$8F,$C9,$36,$32
	.dc.b	$7E,$4E,$32,$D3,$2F,$E6,$CC,$9F,$01,$7D,$03,$E8,$FF,$CA,$71,$86
	.dc.b	$C3,$E2,$AC,$49,$FF,$1B,$65,$FE,$93,$78,$18,$3D,$55,$C5,$C5,$FF
	.dc.b	$AB,$E3,$C1,$18,$8F,$9F,$F1,$BF,$24,$91,$F3,$63,$2D,$C9,$3F,$24
	.dc.b	$FE,$05,$DA,$F1,$B1,$5E,$03,$DF,$F4,$5E,$CF,$A3,$F6,$B6,$35,$92
	.dc.b	$31,$97,$C7,$8C,$4E,$AB,$8B,$4E,$35,$C3,$3F,$F5,$9E,$CF,$5F,$FC
	.dc.b	$CA,$DA,$F8,$1F,$41,$F1,$72,$FC,$99,$FE,$97,$AF,$F3,$5C,$66,$FF
	.dc.b	$D7,$AD,$AF,$F6,$BF,$D0,$BD,$3D,$F1,$7F,$9C,$A3,$13,$F3,$5B,$8F
	.dc.b	$6F,$A2,$FF,$3F,$4E,$6F,$BF,$A5,$23,$EA,$7D,$17,$0C,$FC,$33,$FE
	.dc.b	$7E,$8C,$4F,$FB,$A3,$3F,$F3,$9E,$3E,$A9,$CB,$FE,$7B,$8C,$DF,$FB
	.dc.b	$5A,$73,$27,$7F,$44,$FA,$2F,$E2,$B8,$78,$B2,$FF,$46,$F9,$8F,$12
	.dc.b	$9F,$98,$FA,$07,$7E,$2F,$F1,$95,$B6,$D6,$E7,$59,$FD,$0A,$D3,$67
	.dc.b	$55,$C0,$FC,$B0,$F8,$4F,$F4,$FB,$4F,$C7,$60,$EF,$CC,$C9,$E2,$F9
	.dc.b	$E3,$17,$F8,$09,$E4,$D9,$D4,$70,$3F,$4D,$F4,$2F,$A2,$FF,$57,$B6
	.dc.b	$3C,$3F,$21,$89,$91,$C1,$F8,$D1,$8B,$FC,$7D,$71,$6C,$EA,$B8,$1F
	.dc.b	$B5,$F8,$0F,$E3,$33,$FE,$B5,$6C,$32,$FF,$77,$3B,$F0,$FF,$D8,$73
	.dc.b	$C2,$8F,$0A,$DC,$7E,$82,$34,$49,$FB,$9D,$5C,$9F,$F6,$6F,$AD,$F5
	.dc.b	$B9,$58,$FF,$B1,$F0,$4E,$63,$F9,$77,$C2,$FF,$B3,$27,$3C,$1C,$89
	.dc.b	$3F,$E0,$29,$D5,$71,$AF,$F6,$93,$5E,$33,$BF,$6B,$8C,$7E,$E2,$5F
	.dc.b	$37,$E6,$3F,$C2,$9F,$44,$F8,$7C,$61,$FC,$7E,$B6,$DA,$DC,$EA,$FA
	.dc.b	$7D,$17,$C0,$3F,$DD,$AF,$FB,$97,$E8,$BF,$DC,$DF,$44,$F3,$7E,$16
	.dc.b	$4B,$E1,$97,$77,$8E,$0C,$38,$DF,$09,$CC,$62,$7E,$0C,$7C,$E5,$87
	.dc.b	$77,$FB,$1C,$ED,$FA,$33,$B9,$D4,$7D,$59,$D6,$1E,$5C,$38,$4F,$AC
	.dc.b	$EB,$87,$8E,$33,$1E,$73,$27,$91,$CD,$74,$C3,$C1,$5F,$DC,$E3,$13
	.dc.b	$AA,$E2,$79,$9C,$61,$E6,$C3,$C9,$3A,$3C,$91,$9C,$3A,$C7,$C0,$39
	.dc.b	$CF,$09,$F7,$C7,$0C,$3D,$4F,$33,$D5,$1C,$EF,$C3,$3A,$8E,$33,$EF
	.dc.b	$3F,$B0,$61,$FD,$85,$F8,$C6,$0F,$F0,$87,$D1,$7F,$13,$F7,$3F,$B3
	.dc.b	$36,$33,$D2,$9C,$B6,$38,$FC,$6B,$8F,$8B,$63,$31,$82,$3F,$2B,$E8
	.dc.b	$9F,$3B,$FB,$1C,$FE,$FC,$62,$7F,$09,$93,$0E,$95,$F8,$48,$C6,$1F
	.dc.b	$89,$BE,$17,$D7,$0E,$9A,$7E,$F6,$F7,$3F,$69,$B8,$78,$2B,$72,$17
	.dc.b	$D2,$7C,$18,$F8,$B0,$E8,$DA,$E1,$5B,$9F,$E1,$8D,$8E,$98,$78,$27
	.dc.b	$8E,$3E,$9C,$7C,$35,$E0,$6D,$6E,$38,$E1,$DD,$D1,$C4,$F8,$EE,$63
	.dc.b	$84,$68,$F8,$86,$AB,$83,$FE,$08,$C3,$E4,$C7,$ED,$E3,$83,$67,$9C
	.dc.b	$EB,$5F,$6F,$8B,$FC,$A6,$B8,$4A,$31,$3F,$0B,$1B,$87,$CD,$FA,$2F
	.dc.b	$F3,$3B,$FE,$39,$FA,$07,$BF,$A8,$46,$1F,$F1,$F4,$F0,$7F,$C3,$ED
	.dc.b	$CB,$D1,$3B,$9F,$F1,$1B,$63,$D9,$3B,$C3,$FA,$02,$F9,$DF,$05,$3D
	.dc.b	$7A,$78,$E7,$58,$78,$AB,$2F,$A6,$75,$12,$30,$43,$CE,$FB,$7E,$87
	.dc.b	$F9,$FC,$7C,$1F,$A2,$F0,$99,$E9,$91,$92,$1F,$E3,$E9,$E2,$D8,$E2
	.dc.b	$EB,$3F,$B5,$3A,$7E,$E7,$AC,$FA,$2F,$7F,$C1,$4B,$FF,$35,$4E,$5B
	.dc.b	$39,$7E,$5C,$FD,$59,$D4,$7F,$94,$1B,$5F,$4A,$78,$3E,$0A,$FB,$9F
	.dc.b	$19,$FE,$83,$BF,$F3,$C7,$0B,$BF,$F0,$0D,$8F,$F3,$E4,$ED,$FF,$2E
	.dc.b	$B6,$FF,$2D,$3B,$9F,$F3,$0B,$63,$F3,$53,$BE,$1F,$EF,$CD,$AF,$DF
	.dc.b	$56,$E7,$FC,$04,$D8,$FE,$05,$3B,$7F,$39,$B7,$E5,$7E,$89,$F4,$54
	.dc.b	$FC,$C7,$2F,$E1,$AB,$FC,$E6,$DC,$BF,$59,$5C,$27,$FD,$30,$D7,$77
	.dc.b	$C8,$E8,$EA,$F0,$FF,$A7,$FE,$89,$E5,$FE,$9F,$E8,$9F,$03,$4F,$D5
	.dc.b	$4E,$AB,$FD,$4D,$3A,$7F,$D5,$1F,$44,$E2,$FD,$83,$63,$77,$89,$FD
	.dc.b	$A7,$C9,$BE,$8A,$99,$AF,$F5,$97,$D1,$3E,$33,$F7,$4B,$E1,$A9,$88
	.dc.b	$CF,$D1,$7E,$FF,$E8,$9F,$45,$4C,$E1,$FE,$BD,$FA,$07,$D1,$7F,$05
	.dc.b	$FA,$2F,$F0,$B4,$FF,$B0,$5A,$FF,$A8,$A7,$6F,$FB,$17,$E8,$BF,$A5
	.dc.b	$37,$C9,$FE,$1D,$18,$8F,$F5,$4B,$FE,$75,$74,$43,$31,$C3,$CC,$E2
	.dc.b	$C8,$D7,$57,$F2,$9A,$E5,$1D,$3D,$3E,$FE,$94,$8F,$24,$78,$8C,$3E
	.dc.b	$28,$C3,$FE,$DD,$FA,$2F,$5B,$FE,$BE,$6D,$FF,$37,$6F,$83,$FE,$5B
	.dc.b	$F2,$3F,$01,$5B,$DB,$FC,$05,$5C,$A3,$FC,$C0,$1C,$3F,$E6,$36,$BF
	.dc.b	$D8,$E7,$FD,$A4,$FF,$BB,$D7,$D1,$7F,$37,$6C,$67,$E2,$30,$66,$64
	.dc.b	$64,$87,$FB,$62,$78,$B6,$38,$BD,$FE,$6B,$B7,$D1,$48,$C1,$0E,$4E
	.dc.b	$5E,$91,$FD,$2F,$EA,$7F,$9E,$47,$C0,$3F,$BF,$38,$70,$D7,$E2,$B5
	.dc.b	$FF,$43,$7D,$1B,$E8,$FF,$D2,$71,$CE,$31,$EF,$F5,$84,$3E,$75,$F0
	.dc.b	$19,$8D,$CF,$D2,$F8,$5F,$F1,$77,$6F,$F9,$9F,$8F,$FC,$D1,$1F,$E3
	.dc.b	$1F,$81,$C7,$27,$F8,$56,$36,$FF,$85,$E7,$7E,$E7,$FA,$BA,$73,$87
	.dc.b	$F8,$D6,$B8,$4E,$4F,$DB,$3C,$4C,$DF,$69,$57,$2B,$EB,$39,$8C,$57
	.dc.b	$03,$95,$70,$B7,$87,$B5,$F4,$56,$30,$DA,$E3,$7F,$E4,$F8,$C5,$FA
	.dc.b	$4E,$8F,$95,$7D,$17,$F4,$9A,$F0,$EA,$FA,$97,$D4,$EE,$B5,$7C,$EB
	.dc.b	$53,$9F,$53,$FC,$41,$3F,$EC,$3A,$DB,$FA,$66,$B9,$8D,$BE,$FF,$A4
	.dc.b	$F7,$47,$BD,$B8,$FB,$35,$46,$2B,$D9,$A7,$F9,$B6,$31,$BF,$DA,$6D
	.dc.b	$FD,$86,$D7,$CD,$F9,$7F,$98,$7C,$DF,$A4,$B5,$DE,$2C,$F6,$C5,$58
	.dc.b	$C3,$FC,$7D,$F4,$4E,$9F,$C0,$43,$EA,$7F,$0B,$87,$D1,$15,$BC,$BE
	.dc.b	$FD,$6F,$97,$EA,$4E,$6B,$EC,$4E,$6B,$F0,$B1,$CB,$0E,$F8,$76,$AF
	.dc.b	$CC,$FD,$2F,$E1,$CF,$C4,$74,$5C,$3E,$54,$5C,$3D,$93,$6B,$8D,$68
	.dc.b	$FA,$11,$97,$F4,$35,$6B,$F2,$D3,$77,$FF,$4F,$4D,$D7,$F7,$17,$CE
	.dc.b	$FE,$54,$E6,$F4,$D9,$D6,$1E,$2A,$DC,$78,$8C,$9E,$22,$BE,$26,$E1
	.dc.b	$E2,$6E,$1E,$29,$E6,$4A,$EC,$7F,$00,$C3,$C7,$5C,$1B,$3B,$8F,$E0
	.dc.b	$31,$89,$E3,$18,$3B,$E1,$EB,$AE,$07,$EE,$DC,$4F,$08,$C4,$F7,$38
	.dc.b	$95,$E3,$87,$AD,$B8,$7A,$EB,$94,$E5,$B1,$9A,$F8,$D5,$C2,$FF,$07
	.dc.b	$1D,$A3,$9E,$1E,$B8,$DD,$FC,$37,$4F,$48,$D4,$FA,$0F,$C4,$CE,$8E
	.dc.b	$27,$C3,$7C,$0F,$AE,$7C,$B5,$F0,$A3,$73,$CE,$7B,$47,$38,$C3,$CE
	.dc.b	$38,$BC,$CE,$38,$7C,$16,$E1,$F0,$6B,$89,$27,$4F,$E1,$CF,$29,$F0
	.dc.b	$A7,$C1,$1A,$3F,$DF,$18,$7C,$07,$15,$D5,$B8,$7A,$EF,$B4,$A3,$13
	.dc.b	$AB,$F4,$56,$A7,$35,$B9,$F0,$BD,$4C,$4F,$9B,$0F,$A8,$D9,$F7,$E1
	.dc.b	$E5,$AD,$D6,$8F,$8F,$5E,$D3,$81,$E6,$9E,$17,$F8,$48,$CE,$1E,$E9
	.dc.b	$CE,$1F,$B1,$9C,$E1,$DA,$3E,$A1,$F4,$0F,$EA,$DF,$4B,$FD,$5D,$A7
	.dc.b	$F3,$92,$CF,$F2,$B6,$E3,$FC,$AB,$E0,$FF,$B4,$72,$BE,$75,$D5,$FC
	.dc.b	$92,$73,$18,$BF,$E4,$7C,$1F,$EE,$07,$FA,$CC,$67,$1F,$E6,$13,$97
	.dc.b	$FA,$C4,$7F,$7D,$5F,$4B,$FD,$CD,$B5,$FD,$A9,$FF,$3E,$36,$34,$47
	.dc.b	$5E,$9D,$9F,$EA,$67,$22,$CF,$58,$D6,$7F,$E1,$49,$7D,$2F,$87,$E3
	.dc.b	$FC,$4D,$C3,$C4,$D9,$EA,$DA,$4A,$67,$D2,$7F,$A6,$A3,$FC,$17,$F1
	.dc.b	$CF,$A9,$FF,$6C,$47,$F4,$A9,$D5,$7F,$81,$6B,$5F,$09,$FE,$05,$9C
	.dc.b	$CF,$F9,$79,$AF,$70,$E1,$F0,$FB,$5F,$DF,$39,$BC,$9F,$2D,$F3,$7C
	.dc.b	$BF,$9C,$DC,$BE,$73,$6F,$E4,$B7,$4F,$8C,$D9,$FF,$1C,$4A,$30,$FF
	.dc.b	$8B,$87,$C7,$7D,$B3,$EE,$7C,$B5,$AF,$B9,$16,$7E,$BC,$DB,$F9,$71
	.dc.b	$F2,$FE,$8B,$F1,$51,$8E,$8F,$CA,$7D,$4F,$F8,$A6,$73,$8F,$D1,$9C
	.dc.b	$BF,$DD,$A3,$F2,$57,$FE,$86,$6B,$F2,$5A,$F7,$E9,$F9,$0F,$DB,$9F
	.dc.b	$C9,$57,$E5,$63,$F2,$CD,$8C,$9F,$97,$8C,$B4,$CB,$F9,$43,$27,$F9
	.dc.b	$0D,$7D,$1B,$E9,$3F,$CB,$11,$86,$C3,$E8,$AC,$49,$FF,$2B,$9B,$2C
	.dc.b	$ED,$B1,$F1,$FE,$37,$F7,$3E,$2D,$9F,$DB,$B7,$0F,$F4,$66,$6A,$FF
	.dc.b	$22,$C2,$31,$5F,$E8,$56,$C7,$FA,$CA,$3D,$7F,$47,$E8,$FA,$4F,$4B
	.dc.b	$63,$CA,$D9,$F8,$59,$23,$1B,$FE,$0A,$B9,$9E,$CB,$E6,$D3,$FC,$67
	.dc.b	$AB,$D6,$FF,$36,$36,$3F,$C6,$FF,$45,$F0,$F2,$FF,$21,$4E,$5F,$F2
	.dc.b	$24,$67,$0F,$C8,$C2,$31,$5F,$AF,$FA,$47,$D0,$7F,$9E,$67,$35,$FC
	.dc.b	$06,$73,$5F,$E4,$66,$C7,$F9,$47,$E9,$3F,$D1,$33,$9B,$E5,$F2,$32
	.dc.b	$3F,$C9,$BF,$49,$C3,$2F,$1E,$4A,$FF,$44,$4E,$6B,$F8,$14,$E5,$E3
	.dc.b	$7C,$F1,$FF,$30,$56,$DF,$E2,$CD,$FA,$47,$D2,$3E,$97,$AD,$FF,$A5
	.dc.b	$7B,$9E,$2D,$3F,$2B,$F4,$7F,$C0,$73,$FF,$31,$CE,$9B,$3B,$7F,$2B
	.dc.b	$3B,$CF,$FC,$E5,$5B,$6D,$6D,$FD,$24,$EF,$6F,$A7,$5B,$7E,$AD,$BE
	.dc.b	$1F,$FD,$54,$D3,$BF,$76,$22,$47,$F9,$E0,$D6,$1D,$A7,$2F,$FA,$16
	.dc.b	$33,$5D,$AB,$DF,$5A,$8F,$C2,$FD,$2B,$E9,$7F,$D7,$4D,$BF,$7F,$CA
	.dc.b	$E2,$64,$7F,$A2,$27,$58,$FB,$AB,$2F,$FA,$56,$33,$7D,$AF,$E7,$D6
	.dc.b	$A3,$F2,$BF,$01,$FC,$82,$FF,$A8,$37,$06,$9F,$EF,$67,$FD,$2F,$1B
	.dc.b	$93,$4E,$13,$B8,$FF,$53,$46,$7C,$1F,$F6,$B3,$AE,$8B,$1F,$ED,$9F
	.dc.b	$27,$EA,$61,$FD,$4D,$B6,$A7,$C4,$7F,$B6,$32,$FD,$6D,$6A,$8D,$38
	.dc.b	$D7,$0F,$6F,$B7,$14,$C7,$FD,$63,$5A,$C7,$A4,$78,$67,$C9,$F3,$92
	.dc.b	$3F,$DD,$7F,$4A,$F8,$AC,$67,$FE,$C3,$C5,$78,$D3,$FA,$75,$EF,$0E
	.dc.b	$F3,$EF,$FA,$59,$1F,$EC,$2C,$3F,$C0,$1F,$4B,$FE,$01,$FA,$5F,$EC
	.dc.b	$31,$FE,$03,$C9,$7C,$23,$C0,$FF,$C0,$17,$88,$9E,$2F,$C4,$8D,$2A
	.dc.b	$ED,$E6,$7C,$F9,$FE,$CF,$5C,$2B,$9B,$CA,$FC,$38,$70,$9F,$F0,$41
	.dc.b	$CC,$F0,$5F,$38,$D4,$7F,$83,$70,$B5,$CA,$36,$72,$8C,$E1,$E2,$BE
	.dc.b	$B3,$D7,$0F,$14,$E6,$38,$1C,$8F,$76,$1E,$07,$D3,$5E,$37,$46,$0F
	.dc.b	$99,$3C,$DF,$31,$CF,$0F,$35,$FC,$38,$F1,$57,$38,$C6,$1E,$8C,$3C
	.dc.b	$15,$A3,$AB,$88,$D4,$78,$8F,$90,$7B,$8A,$7C,$07,$0F,$DC,$3E,$7F
	.dc.b	$D2,$FF,$17,$F6,$F8,$36,$3E,$3B,$5D,$F5,$27,$7F,$20,$DE,$92,$B8
	.dc.b	$C6,$5B,$1F,$6B,$E9,$5F,$43,$FE,$47,$D3,$F1,$B1,$8A,$FB,$F9,$30
	.dc.b	$E5,$5F,$92,$8C,$61,$CF,$0E,$5A,$FF,$94,$71,$E5,$8F,$DB,$BD,$B7
	.dc.b	$1E,$75,$BB,$E3,$3F,$E6,$7B,$E3,$8F,$7B,$DC,$FF,$9A,$F0,$E7,$3E
	.dc.b	$8B,$E3,$2B,$E3,$87,$F9,$8E,$F6,$DC,$39,$D6,$EF,$8D,$F0,$8E,$D3
	.dc.b	$AB,$E3,$1E,$5A,$D9,$E0,$8C,$C1,$B5,$CE,$3B,$1F,$32,$BC,$31,$F1
	.dc.b	$B1,$FF,$06,$B7,$5F,$89,$5F,$E8,$99,$FF,$08,$DF,$4C,$3F,$CF,$D1
	.dc.b	$C3,$C5,$B7,$F4,$5B,$7E,$73,$94,$7F,$A1,$7E,$93,$FD,$73,$18,$3F
	.dc.b	$D2,$DF,$4A,$F7,$F8,$C2,$30,$EC,$4F,$27,$F2,$55,$D7,$AF,$8D,$70
	.dc.b	$9C,$9D,$AF,$AD,$F1,$C7,$FC,$5A,$BE,$87,$8F,$AF,$2F,$C1,$39,$6C
	.dc.b	$E5,$EA,$79,$E7,$AC,$48,$F4,$60,$9E,$17,$E0,$AF,$7F,$D2,$7F,$40
	.dc.b	$BD,$FD,$2F,$A8,$CF,$4C,$8C,$90,$F0,$CF,$16,$C7,$17,$57,$FE,$4A
	.dc.b	$9D,$3F,$E4,$CF,$51,$EE,$7C,$8F,$7E,$5F,$CF,$9C,$B6,$72,$FC,$0D
	.dc.b	$15,$F6,$2B,$73,$F1,$1B,$7F,$E5,$AA,$DC,$FC,$4A,$E1,$7F,$80,$FA
	.dc.b	$5F,$F4,$64,$FD,$FE,$37,$85,$F8,$46,$C7,$E0,$E7,$6F,$D2,$6E,$3F
	.dc.b	$E6,$89,$DC,$FD,$66,$C7,$F9,$B2,$77,$C7,$FE,$D9,$6D,$7F,$BA,$6B
	.dc.b	$73,$FE,$DD,$6C,$7F,$BB,$67,$6F,$F9,$8F,$7F,$CC,$FD,$2B,$E9,$69
	.dc.b	$F9,$CE,$6F,$C5,$57,$E2,$1B,$97,$F9,$FA,$B8,$4F,$E7,$DA,$EE,$F9
	.dc.b	$1D,$1D,$5E,$1F,$D3,$7D,$2B,$C9,$FD,$5F,$D2,$BE,$06,$9F,$AE,$9D
	.dc.b	$57,$EB,$67,$4F,$EB,$FE,$97,$EB,$BF,$9E,$6E,$7F,$E9,$49,$DC,$E7
	.dc.b	$8B,$F7,$3F,$2A,$FA,$5A,$7E,$F3,$0E,$3C,$4B,$9F,$4A,$F9,$1F,$E0
	.dc.b	$2B,$E9,$69,$88,$CF,$D2,$FF,$06,$FA,$57,$D2,$D3,$38,$7F,$07,$FA
	.dc.b	$47,$D1,$FF,$0C,$FA,$5F,$F1,$0C,$FF,$0B,$6B,$FB,$A9,$DB,$FC,$3F
	.dc.b	$EC,$DF,$37,$FC,$51,$B1,$98,$FE,$2F,$F4,$DF,$E5,$A8,$C1,$FC,$5D
	.dc.b	$FF,$6B,$37,$E2,$F9,$4C,$8C,$10,$FE,$17,$5D,$9F,$E1,$B5,$FD,$8D
	.dc.b	$E9,$E6,$F8,$0F,$0F,$A3,$63,$24,$38,$BE,$6F,$63,$CF,$EB,$C9,$F3
	.dc.b	$46,$5F,$44,$65,$F0,$C7,$F3,$4F,$A6,$F8,$B1,$FE,$2C,$6B,$FE,$3A
	.dc.b	$6B,$F7,$43,$87,$F0,$0F,$2E,$BF,$F6,$EF,$99,$FC,$DD,$BE,$77,$7E
	.dc.b	$19,$7F,$D0,$6B,$5D,$1D,$5B,$1D,$27,$FA,$4C,$7F,$BC,$63,$87,$D3
	.dc.b	$60,$CF,$C5,$48,$C9,$0F,$E7,$73,$C5,$BF,$3B,$D3,$E9,$A4,$78,$E3
	.dc.b	$B9,$87,$8C,$61,$FE,$B1,$F4,$8F,$A5,$FC,$34,$7F,$98,$23,$1F,$03
	.dc.b	$FE,$39,$5F,$3D,$FE,$97,$9F,$F2,$CC,$69,$FF,$2D,$FD,$4F,$FB,$E9
	.dc.b	$FF,$34,$B5,$FA,$0D,$7F,$5C,$D5,$F4,$CF,$AE,$70,$7F,$92,$63,$0D
	.dc.b	$8C,$CF,$F9,$AA,$72,$FF,$9E,$3E,$2B,$FC,$89,$E1,$6B,$93,$71,$FE
	.dc.b	$47,$F6,$F1,$8F,$EF,$E3,$6F,$F8,$05,$D3,$7E,$03,$F8,$13,$67,$8E
	.dc.b	$0A,$E2,$FF,$7C,$78,$B6,$9E,$4F,$09,$FF,$2C,$65,$67,$38,$77,$8C
	.dc.b	$56,$5B,$19,$3F,$C2,$1F,$4B,$FE,$BC,$C3,$F8,$C3,$7E,$0D,$EA,$BD
	.dc.b	$5F,$3D,$2B,$83,$7D,$7C,$C7,$1A,$E1,$87,$69,$C9,$1F,$F1,$36,$1D
	.dc.b	$B0,$C6,$1E,$FF,$13,$D9,$7E,$D6,$DF,$CB,$6C,$EE,$3E,$74,$FC,$3F
	.dc.b	$A5,$91,$F4,$7E,$62,$D7,$CB,$9B,$C3,$F7,$37,$56,$2F,$F7,$DF,$44
	.dc.b	$EB,$F8,$2D,$8F,$4E,$CF,$23,$35,$27,$FC,$79,$5B,$CF,$FC,$6F,$BF
	.dc.b	$CC,$BD,$47,$F8,$BA,$75,$9F,$E5,$21,$18,$BF,$F6,$F6,$3D,$DE,$B8
	.dc.b	$F2,$BE,$15,$BB,$FF,$7F,$DE,$DE,$D5,$BB,$F6,$BF,$92,$C7,$C9,$1D
	.dc.b	$09,$3C,$1F,$7E,$3C,$63,$FA,$F6,$1D,$DF,$3D,$F3,$C3,$C1,$8F,$92
	.dc.b	$38,$4E,$B0,$E6,$76,$9D,$34,$F7,$E1,$C3,$0E,$C7,$B7,$1F,$3D,$F9
	.dc.b	$49,$7C,$9F,$9F,$8F,$8B,$0F,$79,$D6,$3D,$38,$F9,$F1,$EC,$FB,$70
	.dc.b	$EB,$87,$32,$5F,$2B,$EB,$7C,$23,$A5,$70,$75,$5C,$2B,$BC,$7C,$07
	.dc.b	$CE,$E8,$8F,$D8,$AE,$8E,$8C,$47,$A5,$F9,$46,$6B,$C0,$F6,$38,$1A
	.dc.b	$AF,$E3,$F1,$A8,$C4,$F8,$1A,$76,$38,$E3,$F4,$30,$F8,$D3,$E1,$8C
	.dc.b	$3E,$BB,$F7,$57,$23,$9D,$F7,$AD,$C7,$9F,$1F,$A8,$F9,$EB,$8C,$FD
	.dc.b	$D9,$CC,$F3,$8E,$33,$EA,$9C,$C6,$2F,$E4,$DF,$7C,$3E,$4D,$EF,$1B
	.dc.b	$5A,$C3,$8D,$63,$1F,$81,$7E,$28,$FC,$14,$EA,$BE,$04,$7C,$2A,$F8
	.dc.b	$AE,$AF,$94,$7C,$0A,$E0,$EA,$BE,$31,$F3,$F1,$F4,$47,$82,$31,$7C
	.dc.b	$DE,$E7,$E3,$2B,$9E,$1E,$58,$E7,$5F,$25,$E2,$61,$F9,$15,$CD,$F1
	.dc.b	$1F,$47,$0F,$3E,$1F,$3D,$B3,$ED,$BE,$D5,$BB,$F1,$B4,$E7,$1F,$9D
	.dc.b	$C7,$E8,$4E,$A3,$E8,$34,$FA,$17,$FC,$27,$8D,$A3,$E9,$3F,$AD,$1F
	.dc.b	$D6,$FA,$AD,$FF,$69,$8B,$97,$F6,$39,$B8,$FF,$39,$AF,$EC,$5F,$49
	.dc.b	$FD,$8D,$A7,$F7,$12,$D7,$F6,$B6,$CF,$F2,$8F,$0F,$F9,$E9,$FD,$D3
	.dc.b	$E3,$1F,$49,$FD,$E2,$FF,$9B,$4E,$AF,$F9,$EB,$B9,$D5,$F3,$9C,$92
	.dc.b	$BF,$DE,$F7,$D3,$AF,$94,$7F,$7F,$1F,$49,$FD,$F5,$B5,$FE,$0B,$7F
	.dc.b	$DA,$6D,$8D,$11,$D7,$AF,$67,$A7,$D2,$73,$F8,$FE,$8D,$AE,$8D,$9D
	.dc.b	$C7,$79,$FE,$FF,$F4,$7F,$EA,$D8,$FF,$4E,$67,$FD,$56,$31,$48,$FF
	.dc.b	$15,$FD,$4B,$EA,$9F,$31,$CB,$E0,$26,$3C,$8F,$F3,$56,$3C,$B0,$E3
	.dc.b	$7C,$27,$58,$FF,$AA,$EB,$51,$C2,$F9,$CF,$86,$3F,$D6,$18,$F8,$71
	.dc.b	$CE,$3F,$E3,$8A,$7D,$2F,$CD,$F3,$3E,$73,$70,$F9,$CD,$9F,$A8,$DA
	.dc.b	$E5,$F2,$1F,$54,$FA,$DF,$31,$6F,$EB,$45,$BF,$B9,$36,$BF,$BE,$75
	.dc.b	$BC,$8F,$EF,$D5,$FE,$8B,$8C,$61,$FE,$8D,$8C,$4E,$9F,$F6,$64,$E9
	.dc.b	$D1,$FE,$11,$8F,$F1,$AC,$66,$BF,$CB,$BC,$BF,$9A,$F7,$B3,$53,$0F
	.dc.b	$9B,$87,$17,$E7,$D6,$DC,$E1,$CD,$FF,$2F,$4F,$FB,$75,$FC,$A4,$FE
	.dc.b	$5E,$3F,$30,$D8,$C9,$F9,$98,$CB,$F7,$07,$D1,$3E,$8F,$FC,$C7,$18
	.dc.b	$6C,$3E,$92,$C4,$9E,$53,$C5,$B3,$9A,$FC,$21,$F2,$3E,$3B,$E5,$3F
	.dc.b	$E3,$76,$E2,$69,$2B,$E0,$BF,$BD,$F1,$FE,$6B,$23,$27,$C8,$EC,$4E
	.dc.b	$6F,$C2,$33,$7A,$9F,$E6,$A6,$DF,$D2,$FA,$0F,$73,$F5,$23,$35,$CA
	.dc.b	$72,$DC,$FF,$CE,$7F,$44,$F9,$EF,$F3,$AD,$6B,$8B,$FC,$95,$39,$C7
	.dc.b	$EF,$B7,$0F,$2F,$D1,$7F,$AA,$27,$38,$72,$F3,$E4,$7E,$1B,$E8,$B8
	.dc.b	$63,$FE,$55,$CB,$FC,$FB,$8F,$BA,$73,$5F,$EE,$59,$CD,$7F,$96,$E7
	.dc.b	$34,$13,$AA,$FF,$76,$4E,$A4,$F0,$FA,$27,$D1,$7F,$05,$9F,$06,$5F
	.dc.b	$E9,$4F,$2D,$E1,$53,$F1,$7F,$41,$FC,$03,$2F,$CF,$CE,$9B,$3B,$7F
	.dc.b	$29,$3B,$BF,$B3,$5B,$6D,$6D,$FC,$6D,$2B,$86,$9F,$E7,$9B,$E0,$FF
	.dc.b	$9E,$E5,$F0,$5F,$EA,$A6,$9E,$0E,$CC,$44,$8F,$DA,$9A,$C3,$BE,$0A
	.dc.b	$D4,$7E,$3A,$75,$7D,$EF,$E0,$D6,$A3,$FD,$19,$F4,$4F,$A2,$FF,$5D
	.dc.b	$B6,$FE,$0F,$C9,$62,$64,$7F,$06,$9D,$63,$E2,$AC,$BF,$9E,$8C,$DF
	.dc.b	$6B,$FA,$35,$A8,$FF,$4E,$7B,$DF,$C2,$EF,$F0,$4D,$C3,$E5,$73,$FF
	.dc.b	$B3,$A3,$52,$69,$BD,$3F,$59,$5F,$C6,$A3,$3D,$FC,$4F,$C0,$DD,$9F
	.dc.b	$E3,$FE,$97,$D4,$7F,$25,$82,$30,$FE,$DF,$E1,$3F,$DB,$B1,$88,$D7
	.dc.b	$0F,$FB,$8A,$37,$6A,$DC,$E4,$97,$FC,$B2,$32,$F8,$BE,$6B,$91,$FE
	.dc.b	$0F,$FA,$27,$C3,$E2,$74,$4C,$BF,$C1,$33,$A8,$C6,$DF,$EC,$7C,$DF
	.dc.b	$45,$E5,$3F,$DE,$F7,$FA,$FF,$A2,$FD,$87,$D1,$3C,$9F,$83,$7C,$31
	.dc.b	$F8,$B8,$71,$7F,$A2,$E1,$C4,$5F,$17,$F7,$55,$C5,$FE,$3E,$6E,$34
	.dc.b	$FC,$DA,$57,$8A,$FF,$AC,$E1,$6B,$9D,$69,$EF,$3A,$69,$AA,$E0,$FD
	.dc.b	$AA,$B8,$78,$F0,$F0,$CF,$8E,$FD,$F8,$75,$9C,$CA,$F8,$CE,$6B,$C1
	.dc.b	$5E,$AB,$E9,$18,$39,$E1,$E2,$C3,$B1,$E5,$3C,$E7,$94,$B3,$C9,$FE
	.dc.b	$FD,$87,$82,$79,$47,$8C,$DE,$1E,$37,$66,$23,$73,$C6,$3B,$9F,$5C
	.dc.b	$FF,$02,$9C,$CA,$E1,$D1,$FA,$2F,$A1,$FE,$2B,$ED,$70,$6C,$7F,$83
	.dc.b	$9A,$FF,$85,$5A,$F1,$F8,$CF,$F2,$13,$B6,$BB,$6B,$BD,$3B,$47,$0D
	.dc.b	$3F,$0F,$F4,$4F,$9D,$FD,$AC,$61,$BB,$71,$C3,$E6,$E2,$C3,$94,$FE
	.dc.b	$F3,$0E,$55,$F8,$66,$BC,$27,$6F,$F5,$3A,$E5,$8E,$F0,$EF,$3F,$D5
	.dc.b	$EF,$8E,$3F,$9C,$BD,$CF,$3B,$E0,$C3,$04,$9F,$53,$97,$96,$1E,$28
	.dc.b	$57,$49,$DC,$7E,$2A,$FC,$38,$FC,$19,$E3,$87,$38,$F8,$18,$FA,$2F
	.dc.b	$9E,$3D,$70,$F8,$37,$FA,$D9,$46,$34,$E3,$1F,$E4,$4C,$BC,$58,$E2
	.dc.b	$3F,$C3,$74,$F0,$FF,$94,$E1,$CE,$31,$3F,$E4,$C7,$A3,$FD,$FF,$E8
	.dc.b	$BF,$CC,$0F,$F9,$43,$E8,$1E,$E7,$2C,$FF,$99,$C6,$27,$FC,$AB,$5C
	.dc.b	$1F,$F1,$5B,$70,$F5,$4E,$E7,$FC,$5E,$D8,$F7,$4E,$EB,$F9,$FA,$F9
	.dc.b	$EF,$D1,$36,$FF,$4E,$D7,$FC,$DA,$D7,$FC,$E8,$E6,$BC,$3E,$A7,$A6
	.dc.b	$7F,$C3,$98,$23,$10,$AE,$18,$78,$B0,$CD,$EE,$BE,$25,$F0,$BD,$C7
	.dc.b	$F8,$9F,$E8,$5D,$1F,$B9,$6C,$7E,$F9,$AF,$FA,$11,$AF,$FA,$2D,$CE
	.dc.b	$5F,$23,$E8,$BC,$27,$9A,$33,$9F,$F6,$46,$D7,$B2,$76,$FF,$A2,$23
	.dc.b	$2F,$F9,$A7,$E8,$BF,$86,$7A,$46,$92,$3F,$89,$C6,$5F,$C1,$7B,$DF
	.dc.b	$08,$F8,$F1,$98,$FF,$34,$B7,$1F,$BF,$3B,$9F,$F4,$CC,$E7,$1F,$EF
	.dc.b	$7F,$0D,$FE,$87,$9F,$E4,$2D,$7F,$D4,$ED,$7F,$D5,$CD,$78,$FA,$3F
	.dc.b	$63,$9D,$F1,$14,$FB,$75,$CF,$2F,$D5,$37,$0F,$F0,$5D,$70,$C3,$FD
	.dc.b	$5A,$D7,$FD,$07,$3B,$7F,$CC,$5C,$BF,$8D,$7F,$C2,$6D,$73,$C7,$FC
	.dc.b	$DE,$31,$1F,$B1,$8F,$C6,$FD,$13,$E8,$A9,$F9,$0A,$FC,$4D,$6C,$FF
	.dc.b	$5F,$CE,$DB,$3B,$AF,$CB,$FD,$13,$C9,$FE,$8C,$DA,$FE,$96,$D7,$FD
	.dc.b	$A2,$D7,$FD,$B1,$F4,$4F,$81,$FE,$9C,$D8,$FE,$A6,$D7,$FD,$B2,$D7
	.dc.b	$FD,$C1,$F4,$2E,$3F,$EA,$CD,$8F,$EB,$6D,$7F,$DC,$2D,$7F,$DD,$1F
	.dc.b	$42,$F8,$8F,$EB,$CD,$8F,$EC,$6D,$7F,$DD,$2D,$78,$FD,$0F,$02,$3C
	.dc.b	$1F,$ED,$1F,$3C,$F9,$BA,$49,$FE,$DF,$19,$7F,$B8,$FC,$F3,$E4,$FF
	.dc.b	$81,$B6,$3F,$DE,$2D,$7F,$D8,$FF,$3D,$FE,$5F,$73,$3F,$DD,$E7,$51
	.dc.b	$FE,$58,$0E,$1F,$F2,$DB,$5F,$F5,$73,$D7,$A3,$FA,$05,$3E,$13,$F8
	.dc.b	$AB,$6B,$B9,$FE,$B5,$9D,$F9,$1A,$24,$67,$AB,$FC,$D3,$5F,$E9,$C8
	.dc.b	$C1,$FE,$BD,$7F,$BA,$38,$0E,$1F,$7C,$F8,$BD,$9F,$1F,$A3,$E4,$F4
	.dc.b	$65,$71,$BF,$F3,$73,$67,$2F,$C8,$9E,$2F,$7F,$9E,$FA,$4F,$FB,$C9
	.dc.b	$B9,$7F,$61,$6E,$5F,$E8,$2F,$1F,$F3,$93,$9E,$0F,$F1,$35,$72,$BF
	.dc.b	$A0,$D7,$FD,$12,$D7,$FD,$B8,$F4,$7F,$DB,$D5,$FD,$4E,$FB,$79,$3F
	.dc.b	$D1,$2B,$55,$FD,$1F,$E0,$FF,$A3,$DB,$E7,$A9,$88,$93,$F2,$E7,$4D
	.dc.b	$9C,$AC,$BF,$BF,$4B,$E5,$3A,$FC,$F4,$AE,$51,$E0,$30,$F1,$CF,$E2
	.dc.b	$4E,$63,$ED,$7D,$57,$F5,$77,$F2,$41,$C3,$F9,$56,$BF,$EA,$E6,$BF
	.dc.b	$F0,$07,$E7,$9F,$3F,$F9,$F8,$FC,$E4,$63,$DD,$FF,$2C,$AF,$97,$7B
	.dc.b	$B8,$BE,$17,$FE,$BA,$FC,$2F,$FD,$8D,$9F,$F9,$81,$B1,$E6,$6B,$FA
	.dc.b	$96,$A1,$F0,$8F,$6B,$F8,$55,$7E,$72,$31,$F0,$1F,$E9,$5E,$1F,$F4
	.dc.b	$B4,$6F,$C2,$C5,$73,$8D,$BC,$6F,$9E,$1F,$34,$EE,$FF,$B8,$E7,$94
	.dc.b	$EF,$0E,$75,$AF,$9E,$B5,$E5,$F9,$EB,$82,$6C,$61,$B8,$78,$5B,$7E
	.dc.b	$11,$F3,$DF,$D7,$A7,$F7,$3B,$3E,$AD,$F5,$7E,$2B,$DF,$A9,$E4,$AD
	.dc.b	$4E,$70,$E5,$5F,$84,$6B,$FC,$5D,$AF,$FB,$C1,$AF,$D1,$78,$EF,$FE
	.dc.b	$55,$6E,$3F,$EC,$EF,$4F,$FC,$A0,$FF,$0B,$8C,$D7,$CA,$F1,$BE,$1E
	.dc.b	$5F,$11,$B5,$F4,$5B,$3F,$47,$7F,$BD,$18,$C7,$ED,$EE,$F9,$1F,$9A
	.dc.b	$7C,$EF,$A0,$B8,$FD,$B8,$B8,$F3,$A5,$62,$FF,$CD,$DF,$41,$FB,$1C
	.dc.b	$3F,$92,$C3,$EA,$FF,$06,$EB,$B7,$EF,$B9,$E6,$FA,$17,$C7,$2F,$C4
	.dc.b	$4E,$70,$FE,$59,$39,$8F,$CA,$61,$D2,$7F,$39,$4A,$D4,$7F,$3D,$9D
	.dc.b	$4F,$E8,$67,$2F,$F0,$98,$CB,$C7,$0F,$2C,$62,$77,$7E,$3A,$FC,$64
	.dc.b	$62,$75,$5C,$30,$F5,$B7,$0F,$4B,$6F,$8B,$70,$F8,$93,$F2,$2F,$9B
	.dc.b	$6F,$9B,$4E,$65,$BE,$6D,$AE,$71,$F2,$AB,$9B,$6B,$9B,$6B,$9B,$4E
	.dc.b	$65,$AE,$67,$D0,$9E,$2D,$9E,$2D,$9E,$2D,$38,$C7,$D3,$9E,$2D,$38
	.dc.b	$96,$78,$B6,$78,$C7,$D7,$9E,$2D,$9E,$2D,$38,$96,$78,$9F,$64,$B5
	.dc.b	$F1,$1F,$34,$FE,$2B,$0F,$B9,$7D,$A7,$84,$E7,$0F,$B0,$D3,$E8,$17
	.dc.b	$0F,$90,$D9,$EB,$1F,$91,$9E,$2D,$9E,$2D,$38,$96,$78,$BF,$95,$8F
	.dc.b	$EA,$6D,$3F,$AA,$DF,$EA,$5F,$F1,$43,$5F,$21,$8A,$FF,$72,$38,$21
	.dc.b	$DE,$7D,$07,$F1,$3F,$27,$E7,$C7,$7B,$FA,$7F,$0E,$E2,$FF,$17,$5F
	.dc.b	$D2,$AD,$3C,$94,$EE,$F9,$E3,$FB,$29,$D9,$AB,$E6,$FD,$A3,$26,$AB
	.dc.b	$95,$F7,$AE,$47,$EF,$23,$33,$C6,$33,$8F,$F4,$E9,$CE,$3F,$8C,$AE
	.dc.b	$B5,$C7,$0F,$0C,$E6,$38,$E1,$FE,$40,$BE,$2F,$B6,$F8,$E1,$D7,$0E
	.dc.b	$33,$9B,$E3,$5F,$D8,$EF,$8B,$F3,$C8,$7B,$6B,$8B,$89,$E3,$82,$FE
	.dc.b	$A1,$F0,$70,$FD,$1E,$1D,$2F,$C7,$3F,$A4,$8C,$5F,$28,$FA,$91,$8A
	.dc.b	$FA,$67,$A7,$1B,$3C,$71,$F1,$E1,$D0,$FD,$6D,$F1,$9F,$11,$F1,$67
	.dc.b	$78,$77,$AF,$75,$6D,$FA,$98,$77,$BE,$31,$D2,$F8,$CF,$8A,$3E,$44
	.dc.b	$62,$7F,$C1,$DF,$43,$FE,$04,$3F,$C0,$BD,$96,$32,$4A,$FE,$B1,$19
	.dc.b	$7F,$AD,$47,$F8,$33,$E8,$7F,$C1,$AD,$3F,$C3,$45,$8D,$BF,$E1,$D8
	.dc.b	$FE,$AF,$E3,$7F,$81,$7E,$5D,$DB,$8A,$FF,$06,$C6,$1F,$EB,$33,$FB
	.dc.b	$0B,$FE,$A4,$D7,$FA,$C3,$75,$34,$E9,$1F,$E2,$BE,$A5,$FF,$55,$9F
	.dc.b	$F1,$3C,$BE,$86,$98,$F5,$2E,$3F,$E3,$D9,$D3,$67,$4D,$35,$F4,$3D
	.dc.b	$3E,$3B,$AB,$63,$27,$38,$CB,$4F,$F5,$17,$D0,$7F,$B9,$5F,$D4,$7D
	.dc.b	$5B,$E1,$3F,$C5,$D3,$9C,$7F,$49,$39,$AE,$3E,$DF,$C5,$AD,$57,$FB
	.dc.b	$0D,$AF,$F8,$01,$AF,$F9,$75,$E9,$D1,$FA,$78,$C4,$FF,$93,$BE,$85
	.dc.b	$F4,$4F,$A2,$F1,$CF,$CA,$FA,$2F,$97,$E8,$5C,$FE,$43,$71,$FF,$16
	.dc.b	$07,$19,$7F,$99,$A3,$13,$F3,$FE,$3F,$EA,$9F,$5B,$E6,$2C,$64,$93
	.dc.b	$D2,$32,$FF,$96,$7B,$FF,$14,$FE,$07,$99,$F5,$7F,$E5,$79,$CE,$3E
	.dc.b	$FE,$8F,$C2,$CE,$7A,$BF,$1D,$5C,$B0,$FF,$76,$35,$FF,$14,$B5,$FF
	.dc.b	$43,$3D,$32,$CC,$FE,$7F,$E8,$FA,$47,$E2,$EB,$F2,$38,$7E,$49,$B1
	.dc.b	$93,$F2,$11,$96,$9F,$EE,$A1,$F4,$4F,$A3,$FF,$41,$C6,$1B,$0F,$8D
	.dc.b	$B1,$27,$94,$F1,$6C,$E6,$FE,$57,$C7,$7E,$DF,$C8,$F9,$0D,$AF,$F7
	.dc.b	$C6,$88,$F8,$E4,$8F,$ED,$F1,$96,$C7,$27,$F7,$FE,$55,$BF,$F4,$86
	.dc.b	$6F,$6B,$F7,$CD,$AF,$EF,$2D,$9F,$F6,$6B,$5F,$F2,$D8,$FA,$1F,$81
	.dc.b	$3F,$E0,$06,$C7,$FB,$51,$AF,$06,$BF,$E6,$4F,$A3,$7C,$FF,$F0,$F6
	.dc.b	$D7,$F8,$29,$B3,$FE,$DE,$6B,$CF,$E4,$3F,$8B,$5F,$FA,$52,$3C,$11
	.dc.b	$82,$57,$FB,$37,$93,$F6,$13,$E9,$9F,$F5,$67,$D1,$76,$9C,$DF,$37
	.dc.b	$71,$E0,$AF,$F3,$DB,$5F,$F6,$C3,$53,$7D,$27,$87,$FC,$AD,$B8,$FF
	.dc.b	$89,$9B,$5F,$EF,$A6,$BF,$39,$CD,$FA,$1B,$3F,$EE,$71,$F3,$CE,$DF
	.dc.b	$F6,$34,$EA,$FF,$DC,$58,$60,$FC,$CD,$6E,$FE,$21,$FE,$96,$AE,$15
	.dc.b	$FD,$EF,$E1,$BF,$A1,$49,$A4,$7F,$C7,$E6,$4E,$71,$E5,$8E,$FD,$9F
	.dc.b	$ED,$99,$D4,$FF,$B8,$5B,$97,$D4,$69,$ED,$C1,$5C,$6F,$FD,$C5,$F4
	.dc.b	$0F,$A0,$FE,$AC,$DB,$FF,$28,$B7,$1F,$CB,$87,$0F,$E6,$9D,$63,$F8
	.dc.b	$46,$CF,$87,$E4,$7B,$DC,$8F,$6D,$F3,$8F,$97,$39,$8F,$F7,$E7,$C0
	.dc.b	$FF,$17,$4D,$27,$77,$F6,$D7,$F8,$D4,$0D,$C7,$D6,$E5,$FF,$80,$13
	.dc.b	$A9,$CE,$1E,$28,$E2,$FE,$0A,$77,$5F,$D3,$72,$FC,$FF,$A9,$F8,$CC
	.dc.b	$3F,$DA,$6D,$7C,$0D,$7F,$CC,$7C,$1F,$E0,$18,$FC,$19,$F8,$BB,$7C
	.dc.b	$47,$FC,$00,$AD,$E3,$F1,$2B,$77,$E1,$8F,$65,$F0,$F6,$B1,$7E,$0A
	.dc.b	$AF,$F9,$CB,$E6,$3C,$04,$BF,$F4,$5E,$08,$C4,$B2,$F6,$C7,$4C,$5F
	.dc.b	$40,$F8,$AE,$B8,$71,$C7,$CD,$F4,$18,$3C,$39,$7D,$1F,$A0,$FA,$5F
	.dc.b	$40,$F2,$BE,$3D,$6E,$F8,$9F,$82,$BE,$02,$B8,$3F,$32,$7A,$AB,$EB
	.dc.b	$7D,$2E,$CF,$F2,$C9,$AF,$FA,$52,$BF,$C0,$D7,$D2,$3F,$0C,$73,$3C
	.dc.b	$17,$CD,$A7,$32,$9C,$EF,$AB,$E8,$9E,$98,$2F,$BD,$FE,$A3,$0F,$1C
	.dc.b	$E7,$0E,$38,$78,$70,$E3,$87,$08,$F0,$46,$27,$77,$CA,$B5,$5E,$A8
	.dc.b	$C4,$EA,$3C,$F7,$DD,$F4,$9E,$7C,$3D,$D8,$7C,$38,$C1,$F3,$EB,$94
	.dc.b	$78,$63,$89,$F1,$CF,$F6,$AD,$FF,$B5,$9F,$B2,$60,$FE,$0F,$F4,$1F
	.dc.b	$C7,$FE,$03,$FD,$B6,$D8,$CF,$42,$72,$D8,$E3,$F2,$1F,$95,$6F,$0F
	.dc.b	$28,$C4,$7F,$6A,$B4,$62,$3E,$EC,$62,$7F,$07,$F4,$0F,$9C,$FC,$0E
	.dc.b	$7F,$83,$8C,$5F,$E2,$2F,$83,$F8,$06,$CE,$EB,$8E,$BF,$8A,$BE,$37
	.dc.b	$BC,$7D,$17,$F8,$FC,$38,$DE,$F0,$E3,$7F,$C5,$F0,$E3,$7B,$C3,$8D
	.dc.b	$FA,$9B,$5B,$BE,$2A,$F8,$D6,$F1,$F5,$5F,$0A,$D5,$F0,$BD,$C7,$F2
	.dc.b	$7C,$7C,$AD,$AE,$13,$A8,$E2,$74,$B8,$FE,$92,$34,$6E,$B9,$C7,$03
	.dc.b	$F8,$76,$BF,$05,$FE,$83,$3F,$22,$BF,$C1,$75,$C2,$31,$5F,$37,$C2
	.dc.b	$FD,$BD,$7F,$84,$25,$18,$9E,$97,$D8,$FE,$85,$F3,$FF,$EB,$87,$FA
	.dc.b	$7F,$D0,$7F,$AC,$1B,$EE,$F1,$77,$D7,$55,$71,$7A,$0A,$E4,$E1,$B9
	.dc.b	$7F,$89,$27,$83,$7E,$73,$A4,$1A,$77,$9F,$EE,$B5,$FD,$92,$74,$FA
	.dc.b	$3D,$4D,$1C,$1D,$56,$1D,$18,$3F,$B6,$FC,$FF,$F3,$F6,$F5,$FB,$63
	.dc.b	$11,$E9,$F2,$F1,$3E,$3F,$A0,$93,$C0,$87,$BD,$CB,$F1,$2B,$9E,$7F
	.dc.b	$E4,$4F,$3B,$FD,$85,$3F,$DF,$9A,$FC,$A6,$C7,$F9,$2B,$4F,$F1,$EC
	.dc.b	$EF,$DD,$F8,$9F,$01,$23,$04,$3E,$83,$97,$A4,$7D,$9F,$85,$FF,$45
	.dc.b	$47,$F8,$39,$AF,$DA,$6D,$FF,$96,$B9,$DF,$07,$C7,$D5,$FC,$14,$3C
	.dc.b	$5A,$67,$0E,$93,$87,$FC,$37,$3B,$72,$7F,$87,$B9,$FE,$DB,$F8,$A6
	.dc.b	$E1,$FE,$6A,$6F,$3F,$F5,$4F,$A0,$FF,$0F,$D7,$1E,$4F,$CA,$43,$E8
	.dc.b	$29,$F9,$7B,$E3,$5F,$E2,$16,$A9,$E0,$FF,$8C,$1A,$F8,$EB,$F4,$1F
	.dc.b	$40,$F2,$E9,$27,$86,$9F,$A5,$FA,$07,$BF,$49,$3F,$E3,$CB,$E4,$FF
	.dc.b	$8F,$BE,$81,$C9,$FA,$D6,$C6,$63,$8B,$FA,$EF,$A0,$7C,$6F,$EC,$DB
	.dc.b	$1F,$E4,$86,$BF,$B4,$C7,$FC,$D8,$E7,$E8,$38,$11,$E1,$F0,$3F,$BD
	.dc.b	$FA,$07,$CF,$52,$4F,$F9,$53,$EA,$1F,$51,$FC,$0F,$E8,$9F,$2B,$49
	.dc.b	$5F,$E5,$AF,$A2,$FA,$EF,$F0,$69,$DB,$FE,$5E,$6B,$99,$FE,$1D,$3C
	.dc.b	$3A,$BF,$D5,$75,$A7,$FD,$5B,$0F,$85,$D9,$E0,$BF,$E2,$71,$FD,$81
	.dc.b	$F1,$79,$BD,$0E,$BE,$85,$8C,$90,$E2,$F8,$BD,$9F,$1F,$AD,$27,$D9
	.dc.b	$19,$7D,$D1,$97,$C3,$1F,$C9,$FE,$8B,$E5,$C7,$F9,$F1,$AF,$F2,$86
	.dc.b	$E5,$FE,$C8,$F3,$3F,$0B,$39,$E0,$E4,$47,$FD,$0B,$1A,$9F,$F6,$7D
	.dc.b	$72,$F0,$7F,$C7,$33,$FC,$C1,$BE,$7F,$1B,$FE,$6F,$3F,$EF,$57,$9F
	.dc.b	$D1,$7F,$3D,$83,$7C,$E9,$3A,$21,$C9,$E1,$F2,$FC,$7E,$8A,$4F,$9A
	.dc.b	$32,$FA,$23,$2F,$8A,$3F,$A6,$7E,$C8,$F7,$7F,$C5,$0D,$9F,$F5,$68
	.dc.b	$70,$FF,$AC,$1A,$FF,$96,$07,$D1,$FF,$B5,$DE,$2D,$7F,$D6,$ED,$78
	.dc.b	$35,$FF,$30,$FD,$1B,$E8,$FF,$33,$18,$6B,$FE,$AB,$87,$CD,$F8,$A3
	.dc.b	$13,$AF,$88,$FF,$1D,$78,$7F,$E9,$47,$FC,$79,$EC,$DF,$7B,$F7,$7C
	.dc.b	$7F,$ED,$C9,$CB,$63,$38,$7F,$79,$9C,$B6,$33,$CA,$AF,$F5,$F5,$6A
	.dc.b	$31,$48,$FF,$26,$7D,$0F,$87,$E8,$BC,$4D,$BF,$03,$6B,$B3,$6F,$9C
	.dc.b	$3E,$83,$FA,$15,$7F,$BD,$38,$5F,$44,$FA,$2F,$F5,$8D,$6A,$FF,$C2
	.dc.b	$55,$A9,$CF,$AD,$FE,$AD,$BE,$18,$7F,$B8,$B0,$E1,$8F,$AB,$0E,$13
	.dc.b	$9B,$E1,$87,$28,$F1,$61,$CB,$1F,$8D,$E2,$7C,$FB,$FA,$0D,$AF,$9C
	.dc.b	$D9,$DC,$7D,$19,$FA,$DF,$27,$23,$EA,$7C,$A5,$AF,$A5,$37,$3F,$9F
	.dc.b	$BF,$F9,$E6,$B1,$7F,$EA,$CF,$A0,$74,$FF,$B5,$AD,$F4,$5F,$6A,$B5
	.dc.b	$39,$8C,$65,$FE,$F9,$BE,$35,$FE,$FB,$BF,$F7,$95,$6A,$95,$A7,$FC
	.dc.b	$81,$3D,$B9,$7F,$DC,$73,$98,$E3,$18,$C3,$C0,$DC,$3B,$1D,$F0,$ED
	.dc.b	$7C,$63,$13,$BC,$3B,$D6,$AF,$9D,$72,$9F,$54,$61,$FF,$24,$DF,$37
	.dc.b	$FD,$E1,$5C,$6B,$86,$1E,$97,$94,$FB,$AF,$94,$FE,$52,$F9,$3B,$9E
	.dc.b	$B3,$AA,$F9,$A7,$4B,$ED,$3E,$8C,$3D,$57,$F1,$2B,$BE,$1D,$2F,$73
	.dc.b	$9B,$EB,$5C,$EF,$8B,$F1,$0E,$38,$78,$67,$E0,$C6,$CC,$95,$E6,$78
	.dc.b	$DC,$B4,$CD,$78,$4F,$81,$3D,$63,$E2,$4F,$39,$CB,$4F,$83,$1C,$E3
	.dc.b	$2D,$8C,$C7,$56,$9E,$B8,$F0,$B6,$38,$9B,$8F,$69,$CC,$AE,$5F,$29
	.dc.b	$92,$B9,$C3,$E7,$5F,$CA,$9F,$C1,$B6,$7E,$E1,$E0,$AF,$11,$E6,$9D
	.dc.b	$BF,$22,$3A,$5F,$C5,$BE,$B3,$93,$8D,$79,$9A,$F9,$1B,$1C,$23,$E0
	.dc.b	$9D,$8B,$5E,$49,$F1,$36,$3A,$1B,$7D,$0D,$8D,$E1,$F8,$C7,$BD,$F9
	.dc.b	$C8,$FD,$99,$E7,$5D,$5F,$31,$CC,$CD,$73,$70,$6A,$38,$5F,$AD,$F1
	.dc.b	$98,$BE,$F3,$F5,$67,$57,$E6,$C3,$E9,$61,$D7,$0E,$13,$E5,$BE,$D3
	.dc.b	$92,$57,$02,$4F,$D1,$9F,$2D,$F6,$7D,$33,$A7,$E6,$D7,$2A,$E1,$7E
	.dc.b	$09,$F3,$61,$E8,$8F,$BF,$3E,$3C,$3A,$57,$93,$0E,$95,$F0,$30,$E9
	.dc.b	$3D,$A7,$FB,$A7,$D1,$4C,$3F,$B9,$74,$5A,$FE,$E1,$37,$0F,$E7,$FC
	.dc.b	$DF,$E0,$2B,$FE,$F1,$3F,$DF,$5F,$EF,$DE,$E5,$7F,$BE,$36,$FF,$BD
	.dc.b	$36,$F9,$BF,$CD,$BC,$1F,$EF,$31,$FD,$87,$E6,$DD,$59,$C3,$FB,$84
	.dc.b	$61,$FE,$2F,$5C,$16,$1F,$E2,$D9,$D3,$FC,$C2,$74,$E6,$BA,$3F,$E2
	.dc.b	$B3,$FA,$5C,$7F,$80,$1F,$EB,$4D,$7F,$AA,$37,$87,$BF,$D1,$7F,$87
	.dc.b	$E4,$DF,$53,$0C,$C7,$FC,$41,$3A,$6C,$E6,$3F,$C4,$03,$E8,$BA,$7C
	.dc.b	$87,$56,$D7,$26,$CF,$06,$E9,$FE,$18,$FA,$1F,$F5,$AD,$7F,$53,$F8
	.dc.b	$F7,$D1,$3E,$17,$E1,$E2,$BD,$AC,$3F,$C5,$77,$C2,$B7,$5A,$9C,$DF
	.dc.b	$3C,$3F,$AB,$CE,$5F,$A7,$7C,$5C,$FB,$BC,$8F,$A7,$F4,$4F,$A2,$7D
	.dc.b	$16,$6B,$E3,$FC,$DF,$CD,$F3,$FE,$73,$63,$27,$CC,$8C,$B4,$F9,$FF
	.dc.b	$25,$F5,$4F,$AD,$F3,$56,$32,$4E,$0F,$A3,$19,$7E,$14,$7E,$31,$FF
	.dc.b	$59,$73,$FE,$2B,$C7,$74,$BE,$8B,$E5,$FB,$5F,$90,$E4,$FC,$D4,$62
	.dc.b	$F8,$56,$F1,$FE,$FB,$7B,$7F,$33,$87,$4B,$E1,$1A,$BE,$3E,$FE,$32
	.dc.b	$FC,$39,$F9,$1B,$FC,$93,$6B,$F1,$ED,$36,$FE,$4A,$3F,$46,$BE,$85
	.dc.b	$F4,$58,$6C,$7F,$86,$CD,$7C,$3E,$66,$46,$67,$77,$F2,$0F,$F3,$B2
	.dc.b	$AD,$C6,$4F,$F6,$E7,$C8,$7C,$4F,$27,$E4,$B6,$FE,$2B,$67,$86,$A8
	.dc.b	$C6,$FF,$E7,$68,$C5,$70,$3F,$83,$F9,$5D,$1F,$F1,$5E,$8F,$5F,$F4
	.dc.b	$2B,$C0,$B1,$87,$F4,$BF,$41,$E9,$AE,$33,$8D,$77,$1F,$E3,$6F,$A1
	.dc.b	$7D,$16,$63,$0D,$DB,$E0,$60,$F8,$FE,$BF,$01,$D9,$B7,$FE,$40,$9D
	.dc.b	$36,$3F,$6F,$F4,$5D,$A3,$07,$F9,$19,$C5,$F6,$9D,$E1,$D6,$BF,$C9
	.dc.b	$7E,$43,$C7,$FC,$0F,$D1,$3B,$BF,$81,$CE,$6B,$F0,$13,$97,$8D,$68
	.dc.b	$E3,$DF,$63,$24,$CB,$F0,$4A,$93,$B9,$FC,$14,$BE,$0F,$F2,$73,$F3
	.dc.b	$E7,$31,$8E,$0F,$E2,$F5,$B3,$FC,$CA,$59,$EF,$6E,$FD,$DC,$AD,$BF
	.dc.b	$02,$31,$5F,$93,$FA,$27,$D1,$7C,$43,$F9,$2B,$8B,$F8,$71,$82,$3F
	.dc.b	$1B,$E4,$6D,$73,$9B,$D5,$FE,$D4,$AF,$85,$58,$7F,$86,$7D,$17,$F0
	.dc.b	$1B,$FF,$6E,$66,$FA,$1F,$E6,$B3,$9A,$F9,$51,$8F,$0A,$47,$D6,$EB
	.dc.b	$B8,$7E,$B6,$FA,$27,$96,$09,$DE,$5F,$ED,$EF,$2B,$F0,$17,$FE,$D8
	.dc.b	$FA,$2F,$DD,$CE,$78,$FF,$90,$46,$32,$FE,$49,$7C,$2B,$51,$FE,$E5
	.dc.b	$F8,$9F,$F6,$DB,$AB,$FE,$9F,$5B,$8C,$1B,$F6,$F3,$3A,$F9,$DF,$F6
	.dc.b	$47,$E0,$FB,$CF,$FA,$5C,$62,$75,$7F,$EA,$6C,$13,$E6,$EE,$CD,$FF
	.dc.b	$0F,$9F,$99,$5B,$E1,$FE,$C3,$3F,$E0,$8F,$A3,$FF,$04,$FD,$1B,$CD
	.dc.b	$F6,$D7,$A0,$97,$FE,$B4,$BE,$02,$B8,$3F,$D8,$27,$AA,$BE,$B7,$D2
	.dc.b	$EC,$FF,$4D,$34,$73,$8D,$34,$D4,$F0,$58,$7B,$1D,$1F,$D5,$E3,$51
	.dc.b	$9C,$3C,$B3,$93,$71,$9A,$FF,$11,$57,$0C,$18,$77,$BF,$7E,$1C,$B0
	.dc.b	$E1,$1E,$43,$9D,$70,$7C,$67,$0C,$3D,$75,$E2,$8C,$1F,$1E,$FB,$3F
	.dc.b	$0A,$3B,$57,$37,$84,$66,$B9,$9E,$87,$C3,$5C,$B0,$F5,$36,$3F,$C7
	.dc.b	$C7,$D8,$2C,$61,$C1,$FD,$FF,$E8,$FF,$8A,$7B,$7C,$1B,$1F,$E4,$76
	.dc.b	$BF,$65,$AF,$1F,$8F,$FF,$15,$3B,$6B,$B6,$BB,$E2,$ED,$1C,$32,$FA
	.dc.b	$FF,$46,$F9,$EF,$F0,$94,$61,$BA,$71,$C3,$F0,$38,$B0,$E5,$3F,$E1
	.dc.b	$9C,$39,$57,$E2,$5A,$F0,$9D,$BF,$E6,$AA,$E5,$8E,$F0,$EF,$3F,$E6
	.dc.b	$EB,$E3,$8F,$E3,$2F,$73,$CE,$F8,$30,$C1,$27,$D4,$E5,$E5,$87,$8A
	.dc.b	$15,$D2,$77,$1F,$9E,$BF,$0E,$3F,$06,$78,$E1,$CE,$3E,$06,$3E,$8B
	.dc.b	$E7,$8F,$5C,$3E,$0D,$FE,$5E,$51,$8D,$38,$C7,$FA,$32,$75,$97,$8F
	.dc.b	$1C,$CF,$F8,$6E,$FF,$BE,$3B,$BF,$A1,$E2,$7F,$92,$70,$ED,$5D,$6B
	.dc.b	$FD,$2F,$F4,$7F,$E9,$E7,$81,$FE,$9A,$FA,$17,$BD,$C7,$1F,$F2,$94
	.dc.b	$62,$76,$2B,$93,$FE,$2D,$6E,$1E,$D9,$E1,$3F,$E3,$16,$C7,$C1,$9D
	.dc.b	$D7,$F9,$7D,$7C,$F7,$86,$FC,$ED,$7C,$0D,$BF,$EE,$11,$E4,$9E,$2F
	.dc.b	$9A,$3F,$C1,$CD,$F1,$F9,$4E,$A3,$D1,$5C,$E6,$47,$38,$F0,$7D,$17
	.dc.b	$F3,$E8,$F8,$2D,$7F,$C9,$8D,$C7,$E0,$79,$BC,$3A,$3C,$DF,$47,$2F
	.dc.b	$8C,$65,$F3,$46,$5E,$71,$EF,$F4,$BF,$D5,$11,$F4,$1A,$FF,$96,$DB
	.dc.b	$7F,$3F,$AF,$87,$BF,$F1,$7D,$19,$7C,$63,$2F,$C5,$8C,$BC,$E3,$E7
	.dc.b	$7D,$1F,$F9,$A6,$3E,$FB,$5F,$F3,$53,$6F,$F3,$5D,$0F,$83,$E3,$E6
	.dc.b	$FD,$78,$78,$54,$CE,$1D,$27,$0F,$E3,$27,$6E,$4F,$C7,$F4,$7D,$7C
	.dc.b	$3F,$50,$DE,$6F,$F5,$33,$67,$FD,$E7,$F4,$5F,$E1,$17,$8F,$27,$E2
	.dc.b	$61,$F4,$74,$FC,$65,$71,$AF,$F3,$35,$65,$FC,$FD,$70,$79,$9F,$A2
	.dc.b	$FA,$37,$99,$F9,$56,$C6,$63,$C1,$A7,$E6,$7E,$85,$F0,$74,$95,$B3
	.dc.b	$F5,$BF,$47,$FE,$85,$6F,$27,$E8,$63,$11,$FE,$9B,$AE,$6F,$ED,$3E
	.dc.b	$8D,$F1,$9F,$A7,$6C,$66,$3C,$5E,$AE,$7E,$8E,$4F,$17,$87,$C1,$7E
	.dc.b	$BF,$E8,$DF,$3F,$49,$5F,$C0,$BE,$8D,$F4,$7F,$B5,$FC,$D3,$FD,$A9
	.dc.b	$3F,$C1,$9A,$FF,$AC,$5B,$D3,$FC,$85,$FA,$06,$3A,$7F,$DC,$0D,$90
	.dc.b	$70,$DE,$7F,$07,$C4,$48,$F0,$3E,$38,$FE,$11,$F4,$8F,$97,$FE,$E0
	.dc.b	$DA,$FE,$EA,$D7,$FC,$48,$1C,$3F,$E2,$77,$58,$FF,$39,$FA,$47,$C4
	.dc.b	$7F,$15,$6D,$7F,$26,$AD,$E7,$FE,$FB,$9F,$EC,$D3,$FC,$A9,$AF,$FB
	.dc.b	$55,$B7,$D7,$8B,$A7,$A3,$2B,$83,$DE,$3C,$1E,$CF,$87,$D8,$95,$D6
	.dc.b	$32,$F7,$8C,$BD,$E3,$2F,$86,$33,$7E,$D9,$FE,$61,$F4,$9F,$82,$AF
	.dc.b	$F0,$DB,$5F,$8E,$D7,$FC,$98,$D7,$EC,$87,$84,$FF,$61,$F3,$BF,$A1
	.dc.b	$E7,$A6,$D7,$E0,$2B,$73,$9F,$37,$FA,$15,$3E,$92,$98,$89,$7F,$D5
	.dc.b	$A7,$4D,$9C,$AC,$3C,$5F,$35,$D3,$E9,$24,$F1,$8E,$E6,$1E,$18,$F9
	.dc.b	$E7,$31,$B3,$C0,$EE,$BA,$C7,$F6,$8F,$A2,$7D,$17,$F1,$58,$FF,$39
	.dc.b	$46,$3E,$0F,$FC,$8E,$BE,$6F,$FD,$51,$3F,$94,$8D,$3F,$96,$FA,$5F
	.dc.b	$F5,$76,$5F,$E4,$E6,$C6,$6F,$FC,$1C,$FD,$28,$CB,$F5,$3E,$2D,$EF
	.dc.b	$FF,$08,$BF,$F0,$1C,$63,$E1,$FF,$CD,$9A,$7F,$9B,$63,$7E,$26,$2B
	.dc.b	$F6,$B1,$B7,$8D,$F3,$C3,$E3,$89,$D3,$87,$9C,$F0,$C3,$A4,$EE,$9F
	.dc.b	$45,$8B,$F3,$FD,$1D,$9C,$E1,$E0,$8C,$56,$5B,$19,$3C,$B1,$96,$99
	.dc.b	$BF,$38,$FA,$3F,$F6,$5B,$FC,$49,$BF,$0C,$F5,$DE,$9F,$A2,$B8,$7A
	.dc.b	$D6,$F8,$C6,$30,$FF,$17,$E1,$C6,$F5,$7E,$BF,$0B,$D8,$DA,$F9,$6D
	.dc.b	$BF,$91,$A7,$93,$5F,$B3,$7C,$2F,$E8,$FC,$9F,$C5,$3E,$37,$CA,$58
	.dc.b	$C9,$30,$E7,$19,$7A,$CD,$E2,$FC,$06,$2A,$C5,$FE,$02,$15,$8B,$FB
	.dc.b	$15,$71,$FC,$6C,$DA,$E5,$7E,$1C,$3A,$CE,$6F,$4D,$9D,$1D,$DC,$61
	.dc.b	$E6,$8C,$BF,$87,$8F,$7C,$63,$0F,$0B,$70,$F0,$D7,$12,$61,$E2,$9F
	.dc.b	$CC,$E1,$E2,$78,$36,$77,$1F,$06,$31,$3C,$67,$E5,$CF,$7C,$3C,$F3
	.dc.b	$92,$47,$AC,$EE,$6C,$EE,$59,$EE,$7C,$C7,$11,$CA,$30,$6B,$0F,$7B
	.dc.b	$70,$F7,$CE,$E7,$2D,$8C,$D7,$8E,$B8,$5F,$CA,$6C,$79,$1A,$75,$C3
	.dc.b	$E0,$C7,$1B,$FC,$3C,$62,$7E,$A4,$FB,$9E,$33,$E1,$9D,$61,$E4,$AF
	.dc.b	$98,$7B,$8E,$86,$2B,$8B,$87,$E0,$D7,$AE,$FE,$7C,$E6,$BA,$57,$88
	.dc.b	$FA,$A7,$23,$15,$E8,$9F,$04,$61,$F0,$47,$57,$E1,$36,$3E,$14,$70
	.dc.b	$9E,$E7,$F2,$CB,$F6,$E1,$FA,$69,$F8,$57,$C4,$F8,$4F,$18,$D5,$F9
	.dc.b	$A7,$53,$EB,$9F,$AF,$3E,$52,$5F,$82,$32,$DC,$3E,$1C,$EA,$7A,$DF
	.dc.b	$82,$3E,$15,$71,$C3,$CF,$7F,$81,$AF,$0D,$71,$9C,$E1,$EA,$3C,$95
	.dc.b	$DA,$FC,$11,$F1,$8F,$AD,$4F,$A1,$75,$7F,$27,$6C,$FF,$24,$E0,$FF
	.dc.b	$15,$BF,$EC,$E8,$CD,$7F,$34,$6E,$5F,$EE,$96,$CF,$F9,$1B,$7F,$E6
	.dc.b	$9E,$2F,$5C,$7F,$A3,$7D,$04,$AF,$EA,$BD,$57,$2F,$E5,$73,$72,$FE
	.dc.b	$3F,$36,$BF,$8F,$1F,$D3,$FD,$BF,$EA,$0D,$3F,$AF,$96,$FF,$9A,$B6
	.dc.b	$BF,$94,$B4,$E6,$7F,$66,$9F,$85,$F0,$9F,$DD,$AB,$FC,$BF,$A7,$FB
	.dc.b	$CA,$BF,$CA,$71,$F2,$67,$4D,$8C,$CE,$E3,$E3,$4F,$F6,$AF,$8D,$7C
	.dc.b	$CF,$F7,$3C,$7E,$44,$FF,$6B,$A7,$D0,$53,$1E,$C5,$CB,$FC,$07,$3A
	.dc.b	$6C,$E9,$A6,$BE,$83,$A7,$C1,$75,$6C,$64,$E7,$19,$69,$9A,$FF,$02
	.dc.b	$08,$D3,$87,$FB,$77,$CF,$FF,$AA,$A7,$E2,$B6,$7F,$CF,$2D,$7F,$CA
	.dc.b	$21,$C3,$FE,$54,$78,$64,$9C,$7B,$9F,$E1,$2F,$9F,$7C,$FB,$E7,$FC
	.dc.b	$33,$F2,$7E,$67,$E5,$79,$97,$0F,$8E,$DC,$FD,$6D,$CB,$FC,$55,$08
	.dc.b	$FA,$5F,$1F,$2B,$E7,$FC,$CD,$AF,$87,$36,$F8,$CD,$AF,$F1,$8D,$2B
	.dc.b	$0F,$F8,$BC,$4E,$1F,$F4,$BF,$8F,$F8,$AE,$B3,$49,$7F,$90,$8C,$BF
	.dc.b	$EB,$2C,$3E,$35,$7F,$AD,$27,$2F,$F8,$FB,$9B,$F1,$FE,$F7,$F8,$8E
	.dc.b	$73,$5F,$02,$BF,$21,$5F,$91,$6D,$FE,$39,$A7,$E0,$8B,$3E,$08,$7C
	.dc.b	$F3,$E7,$FF,$C8,$F1,$86,$C3,$E1,$6C,$49,$F5,$1F,$EC,$C3,$81,$8B
	.dc.b	$FC,$15,$71,$8C,$1A,$8F,$F2,$DB,$6F,$C9,$3E,$38,$DB,$C6,$76,$E5
	.dc.b	$FF,$26,$E0,$F4,$FF,$C9,$0B,$A6,$C6,$1F,$F2,$67,$CF,$73,$C7,$E8
	.dc.b	$DA,$F9,$63,$E9,$85,$6E,$3F,$D6,$33,$A9,$FF,$2E,$FC,$FB,$E7,$BA
	.dc.b	$EB,$FE,$59,$8C,$5F,$3C,$39,$5F,$FB,$C9,$FF,$2F,$C6,$7D,$9F,$2F
	.dc.b	$9B,$E6,$6C,$71,$8F,$F3,$27,$CF,$F4,$BF,$F3,$6D,$F2,$7F,$CD,$F1
	.dc.b	$9D,$3D,$B3,$97,$DF,$19,$BF,$F6,$5C,$E5,$FF,$4D,$76,$FE,$0B,$E7
	.dc.b	$7F,$59,$1F,$0F,$CA,$FE,$14,$FD,$A8,$CE,$3F,$E7,$E9,$CB,$6B,$97
	.dc.b	$57,$F9,$EE,$73,$97,$D0,$6E,$1E,$03,$A4,$EE,$BF,$D0,$50,$F7,$3F
	.dc.b	$2B,$3F,$2B,$0F,$E1,$D1,$8D,$F6,$CA,$F3,$74,$DA,$FF,$49,$C7,$33
	.dc.b	$EF,$15,$FC,$9E,$7F,$BE,$8C,$D7,$FA,$4F,$E7,$DF,$3F,$F1,$4C,$16
	.dc.b	$7F,$11,$1A,$6C,$7C,$BF,$8D,$DC,$60,$98,$78,$23,$2F,$FA,$A3,$D9
	.dc.b	$FD,$FC,$FE,$B7,$17,$CE,$FF,$A9,$67,$35,$FC,$BE,$73,$5F,$99,$9C
	.dc.b	$C9,$DC,$27,$F4,$0E,$FB,$BA,$11,$FB,$DD,$16,$FF,$D7,$7E,$7F,$D8
	.dc.b	$7F,$51,$F3,$FF,$B3,$E4,$FD,$5C,$E6,$93,$9B,$5F,$F8,$37,$8F,$FD
	.dc.b	$7D,$7B,$F8,$1F,$F5,$EB,$D6,$4D,$35,$EB,$F0,$AE,$3F,$2D,$FE,$DE
	.dc.b	$CD,$F3,$FB,$8C,$37,$2F,$DE,$60,$9F,$EF,$99,$BB,$39,$D6,$B8,$7C
	.dc.b	$A7,$FB,$4F,$4F,$4F,$CF,$FA,$BE,$7F,$FA,$34,$7F,$89,$E7,$75,$C4
	.dc.b	$FE,$15,$1C,$04,$F0,$7D,$B3,$D5,$57,$5F,$1F,$E3,$C6,$AF,$0F,$F7
	.dc.b	$29,$AE,$4E,$B3,$FD,$D9,$E6,$60,$F3,$4F,$07,$C9,$7E,$3C,$FF,$13
	.dc.b	$1B,$3C,$37,$DA,$33,$3E,$99,$E3,$7E,$69,$CB,$E1,$AE,$18,$7F,$5E
	.dc.b	$BE,$17,$E4,$7C,$11,$9A,$E1,$3E,$1B,$E6,$FB,$CC,$4E,$DC,$57,$2C
	.dc.b	$3D,$91,$D6,$FC,$95,$B8,$F5,$9C,$A7,$78,$79,$9B,$1F,$E4,$63,$E4
	.dc.b	$9F,$E4,$88,$C3,$83,$EA,$FC,$FF,$F1,$5F,$6F,$FC,$9A,$D8,$FA,$4D
	.dc.b	$7F,$C5,$2D,$78,$FC,$3B,$83,$8B,$63,$8B,$5E,$2D,$78,$C7,$E0,$BE
	.dc.b	$7D,$F3,$3F,$88,$D7,$F1,$D1,$8A,$FC,$3E,$0C,$3B,$57,$E4,$E3,$18
	.dc.b	$73,$C3,$96,$BF,$92,$C7,$96,$3F,$85,$BD,$B7,$1E,$75,$BB,$E3,$3F
	.dc.b	$C0,$2F,$8E,$3D,$EF,$73,$FC,$13,$0E,$73,$E8,$BE,$32,$BE,$38,$7C
	.dc.b	$08,$F0,$63,$E7,$61,$82,$5F,$37,$2F,$86,$3E,$1E,$3E,$29,$ED,$87
	.dc.b	$83,$0E,$F1,$E5,$9D,$47,$C2,$AF,$F0,$4B,$74,$F8,$15,$FE,$09,$CB
	.dc.b	$D7,$96,$23,$EF,$4F,$CB,$F1,$3F,$7B,$87,$26,$C6,$E3,$FC,$2B,$F3
	.dc.b	$DF,$E4,$CA,$FF,$0A,$7C,$FB,$D7,$DC,$7F,$2C,$8C,$3F,$E1,$99,$E0
	.dc.b	$F8,$0F,$E6,$2E,$2B,$FC,$43,$3C,$27,$41,$C4,$78,$BA,$FB,$CA,$9F
	.dc.b	$1D,$86,$C7,$86,$37,$7F,$E6,$88,$C1,$EA,$F4,$72,$71,$99,$19,$21
	.dc.b	$E4,$AE,$AF,$9B,$E7,$FF,$9D,$3F,$E3,$76,$B9,$E9,$F0,$F9,$DE,$89
	.dc.b	$CE,$7F,$03,$E7,$FC,$A6,$6A,$5F,$CE,$8C,$BD,$63,$E0,$FC,$FB,$E7
	.dc.b	$E9,$27,$EA,$F7,$D7,$E5,$ED,$FE,$53,$7E,$CF,$A3,$9C,$BF,$CA,$38
	.dc.b	$73,$9C,$90,$FA,$6E,$5F,$AB,$F0,$7F,$E7,$37,$FC,$AC,$D7,$3E,$67
	.dc.b	$D6,$EA,$7C,$07,$D8,$AD,$61,$FE,$60,$97,$89,$B3,$57,$30,$FF,$31
	.dc.b	$4E,$9B,$39,$7F,$0F,$D1,$F5,$2F,$FC,$D6,$DE,$8F,$EB,$2D,$9F,$F0
	.dc.b	$0F,$CF,$7F,$84,$DF,$F3,$74,$65,$43,$E7,$FF,$1E,$D8,$DF,$3F,$E4
	.dc.b	$AB,$84,$E6,$7F,$32,$DC,$3C,$75,$F9,$8F,$9F,$79,$94,$97,$FE,$7F
	.dc.b	$BF,$8B,$97,$F7,$0F,$9E,$7B,$9F,$A0,$6D,$7F,$A1,$5A,$FE,$A3,$0E
	.dc.b	$AF,$FA,$27,$E7,$FF,$00,$FE,$B1,$B9,$FF,$A4,$1B,$87,$7C,$3F,$5B
	.dc.b	$F3,$EF,$90,$A4,$C3,$C1,$EA,$6B,$E7,$E4,$7F,$A5,$E9,$1C,$BE,$03
	.dc.b	$F6,$FF,$3E,$F9,$FA,$4B,$FF,$4F,$7C,$FB,$E7,$FF,$7B,$F3,$EF,$9F
	.dc.b	$A4,$AF,$F5,$24,$7D,$73,$4E,$23,$4F,$F0,$A8,$D4,$63,$87,$FD,$5F
	.dc.b	$1C,$27,$5C,$DF,$C3,$3E,$7D,$F2,$B4,$93,$FE,$B3,$F9,$FF,$C4,$3F
	.dc.b	$C5,$A7,$6E,$7E,$27,$F8,$AC,$9E,$31,$FC,$72,$3F,$B0,$3C,$FD,$2D
	.dc.b	$9C,$38,$EC,$64,$87,$17,$AF,$B7,$DB,$D9,$93,$E5,$8C,$BE,$78,$CB
	.dc.b	$E1,$8C,$BE,$38,$CE,$BF,$ED,$2C,$15,$FC,$BB,$E7,$FE,$C5,$FF,$B4
	.dc.b	$DA,$FF,$36,$6D,$7F,$B6,$BD,$0F,$C8,$CE,$78,$FF,$9B,$C7,$F3,$AE
	.dc.b	$AF,$F2,$5B,$67,$FC,$E7,$7C,$A7,$3E,$7F,$0F,$9F,$95,$FE,$E3,$BF
	.dc.b	$F7,$73,$67,$2F,$F4,$88,$4F,$82,$7C,$3F,$2D,$E2,$F9,$F9,$7C,$E3
	.dc.b	$2F,$58,$CB,$CE,$32,$F5,$8C,$E7,$EF,$9F,$EA,$7F,$48,$F8,$7F,$F1
	.dc.b	$C3,$6F,$FC,$C4,$1C,$3F,$E6,$56,$BF,$E6,$E1,$F3,$EF,$A0,$FC,$7C
	.dc.b	$7F,$9C,$A3,$1F,$0D,$D1,$7C,$CB,$E1,$B1,$7C,$2D,$F1,$6F,$82,$FD
	.dc.b	$E7,$17,$FA,$8B,$63,$FE,$A4,$9D,$7B,$B8,$BF,$EF,$11,$C7,$C7,$C5
	.dc.b	$E6,$37,$7C,$CE,$58,$76,$BE,$2F,$FB,$1A,$79,$4E,$E7,$CB,$39,$1F
	.dc.b	$3D,$DA,$79,$FC,$FD,$AD,$CD,$BF,$64,$DB,$E3,$7D,$67,$2D,$8C,$DF
	.dc.b	$92,$B8,$5F,$16,$D6,$DF,$04,$BE,$7F,$FA,$BC,$FA,$75,$7D,$1B,$E9
	.dc.b	$37,$5A,$BF,$F1,$05,$6A,$73,$EA,$FC,$D9,$FC,$05,$6D,$B1,$97,$FD
	.dc.b	$63,$19,$7E,$6F,$89,$EF,$9F,$80,$DC,$7D,$CD,$9F,$A2,$DD,$BE,$C6
	.dc.b	$AF,$A0,$F9,$07,$C9,$FA,$1B,$87,$C7,$8B,$87,$DA,$9B,$7C,$A6,$D7
	.dc.b	$E0,$26,$D7,$DC,$9B,$C1,$F9,$19,$B8,$78,$6F,$B1,$F5,$63,$35,$A6
	.dc.b	$C6,$AF,$C4,$DC,$3C,$4D,$3C,$45,$8F,$13,$70,$F1,$37,$0F,$14,$FE
	.dc.b	$5B,$0F,$85,$39,$7C,$4D,$38,$61,$E4,$9E,$11,$E9,$34,$E0,$F3,$97
	.dc.b	$0F,$3B,$63,$CE,$D3,$CE,$5C,$3C,$ED,$C3,$CF,$5C,$8F,$B7,$7E,$37
	.dc.b	$17,$D1,$B5,$D1,$B7,$D1,$B7,$D2,$BF,$73,$87,$C2,$BF,$03,$F0,$9B
	.dc.b	$1F,$08,$F7,$D7,$2C,$3C,$51,$C8,$F8,$93,$C0,$F5,$46,$9F,$24,$79
	.dc.b	$CE,$0E,$E7,$DB,$1D,$8C,$B8,$8D,$46,$0E,$4F,$42,$61,$FC,$86,$FD
	.dc.b	$95,$F0,$1F,$63,$6B,$D9,$5D,$0E,$38,$30,$F4,$61,$CE,$BD,$B7,$C6
	.dc.b	$FD,$B1,$E7,$78,$D7,$0B,$F9,$D1,$FC,$82,$31,$87,$92,$BD,$53,$E0
	.dc.b	$8C,$36,$3C,$B7,$DC,$FA,$0F,$7C,$3C,$93,$F9,$09,$D4,$F4,$9F,$C0
	.dc.b	$37,$0F,$1C,$70,$9F,$98,$70,$AF,$4D,$75,$AD,$9F,$0D,$C5,$F9,$AF
	.dc.b	$94,$75,$8C,$4F,$07,$E7,$C6,$4F,$59,$F0,$29,$F4,$0E,$AF,$E8,$18
	.dc.b	$BE,$93,$FA,$63,$FC,$A6,$75,$7F,$D2,$E3,$5C,$9F,$CE,$27,$38,$7F
	.dc.b	$B4,$67,$35,$FD,$46,$3B,$E1,$DB,$0E,$B3,$FD,$6B,$E8,$BF,$A6,$1F
	.dc.b	$D3,$7A,$2C,$7F,$4B,$8B,$7F,$D7,$A6,$E1,$FC,$BA,$BF,$AA,$7B,$DF
	.dc.b	$D5,$5A,$7F,$53,$2D,$7F,$66,$6C,$F1,$7F,$AC,$4F,$16,$9E,$9C,$BF
	.dc.b	$DE,$4F,$F5,$3F,$A2,$74,$FC,$1F,$89,$FE,$D1,$CE,$9C,$D7,$19,$FE
	.dc.b	$9D,$90,$FA,$2E,$1B,$1F,$E1,$AC,$3B,$C7,$F8,$5F,$8B,$87,$D1,$7F
	.dc.b	$73,$83,$7E,$0F,$0C,$BF,$EE,$93,$A6,$CE,$63,$FC,$1E,$3E,$8B,$A7
	.dc.b	$C3,$75,$6D,$72,$6C,$F0,$6E,$FF,$E1,$58,$E5,$3E,$0A,$E4,$D3,$F0
	.dc.b	$5F,$45,$FE,$AE,$7F,$C2,$9F,$26,$FA,$4F,$F7,$74,$6B,$DD,$FF,$09
	.dc.b	$4E,$7E,$2B,$FC,$0B,$7C,$71,$65,$FD,$F6,$30,$FF,$8A,$FE,$8D,$F4
	.dc.b	$6F,$A3,$E3,$97,$C9,$F2,$FE,$57,$9F,$75,$F8,$ED,$CF,$E2,$B7,$2F
	.dc.b	$73,$F4,$7E,$4A,$57,$CE,$F9,$DB,$7F,$32,$6D,$FC,$59,$B5,$FE,$41
	.dc.b	$A5,$61,$FF,$24,$89,$C3,$FE,$9D,$EA,$FC,$6F,$4F,$FA,$DE,$FF,$59
	.dc.b	$EC,$FE,$3A,$BF,$C4,$33,$97,$E2,$CF,$E3,$DE,$4F,$0A,$E7,$F0,$3F
	.dc.b	$E3,$0B,$FF,$24,$57,$E5,$6B,$F1,$3D,$9F,$8F,$8F,$C8,$36,$32,$7E
	.dc.b	$3A,$32,$FF,$93,$C7,$D1,$3E,$8F,$FC,$99,$18,$6C,$3E,$3A,$C4,$9E
	.dc.b	$53,$C5,$B3,$9A,$F6,$9F,$E6,$C9,$D3,$89,$EC,$70,$39,$D7,$F9,$3A
	.dc.b	$DE,$AD,$EC,$C3,$6B,$F8,$C7,$D0,$79,$5F,$E3,$55,$C1,$CE,$5C,$1F
	.dc.b	$E3,$DF,$44,$FA,$3F,$F2,$DD,$F1,$CF,$CE,$27,$55,$C8,$EF,$F1,$FC
	.dc.b	$3D,$FE,$2D,$8C,$92,$B9,$FD,$1F,$27,$AE,$0A,$FF,$35,$CE,$79,$BF
	.dc.b	$CC,$53,$9B,$FE,$69,$39,$7F,$0D,$19,$BF,$3F,$88,$F0,$BE,$CF,$D0
	.dc.b	$BA,$BF,$CE,$F3,$9C,$3F,$01,$39,$7F,$A0,$46,$6F,$9D,$F2,$EB,$E2
	.dc.b	$CA,$DB,$96,$9F,$D2,$47,$C1,$7E,$2A,$7E,$66,$9F,$E7,$D9,$C9,$F9
	.dc.b	$18,$CB,$FA,$13,$37,$F4,$9B,$3F,$EB,$15,$8F,$EC,$5F,$05,$FF,$A5
	.dc.b	$1D,$9C,$DF,$87,$D7,$97,$FD,$33,$F4,$6F,$A3,$F8,$E6,$0B,$5F,$D8
	.dc.b	$E3,$4D,$8F,$9B,$F2,$7B,$8C,$13,$2F,$CE,$62,$9D,$47,$E3,$3D,$EF
	.dc.b	$E0,$59,$FA,$E9,$F4,$3F,$EA,$C9,$CE,$1F,$BD,$9C,$BF,$DD,$63,$37
	.dc.b	$FD,$55,$B1,$F1,$3A,$39,$77,$C8,$E0,$FE,$12,$31,$5F,$9E,$F2,$3F
	.dc.b	$01,$97,$FB,$76,$5F,$45,$FB,$68,$FF,$5A,$4E,$B0,$FE,$C9,$5A,$EE
	.dc.b	$FF,$61,$CE,$70,$FE,$D0,$D9,$7C,$37,$FB,$07,$0E,$7C,$9C,$09,$8F
	.dc.b	$FB,$26,$74,$FF,$23,$BE,$98,$17,$CD,$FF,$88,$BE,$8F,$31,$8A,$FF
	.dc.b	$10,$46,$27,$CA,$7F,$80,$34,$76,$62,$BF,$C4,$17,$F3,$B0,$F8,$9C
	.dc.b	$3F,$ED,$E7,$FC,$05,$F4,$7F,$E0,$3F,$A3,$FE,$87,$1E,$E9,$F8,$24
	.dc.b	$BF,$F1,$54,$70,$13,$C1,$F5,$4F,$55,$5D,$7C,$8F,$81,$2B,$C5,$7F
	.dc.b	$45,$36,$74,$9D,$C6,$A3,$FA,$55,$5C,$3D,$98,$78,$67,$C8,$F8,$E3
	.dc.b	$37,$D6,$7A,$CE,$70,$F1,$47,$73,$C5,$7D,$63,$33,$D6,$78,$D7,$91
	.dc.b	$E0,$79,$0B,$7E,$07,$99,$8A,$E7,$3E,$3B,$ED,$87,$B6,$BD,$33,$D7
	.dc.b	$0F,$05,$6A,$7B,$9A,$3E,$B5,$71,$7C,$A6,$0D,$9F,$30,$F8,$45,$3C
	.dc.b	$46,$0C,$1F,$E4,$5F,$A2,$FE,$3F,$EE,$70,$6C,$7C,$B6,$BF,$65,$AF
	.dc.b	$F7,$F5,$F1,$BF,$E2,$68,$E0,$D7,$83,$5E,$0D,$7C,$4F,$1C,$3F,$1D
	.dc.b	$F4,$6F,$9D,$FF,$06,$46,$1B,$9F,$2E,$4F,$C4,$63,$E2,$7F,$15,$7C
	.dc.b	$71,$B7,$C1,$EF,$3C,$31,$F0,$61,$FE,$19,$BE,$15,$F8,$06,$E3,$FA
	.dc.b	$1C,$7B,$4F,$E1,$1B,$1E,$58,$EE,$75,$C2,$3C,$F1,$F1,$2B,$EB,$87
	.dc.b	$AE,$3A,$63,$E4,$7F,$CF,$53,$A6,$D7,$2C,$7C,$37,$C7,$1F,$15,$77
	.dc.b	$3C,$2E,$31,$FD,$1C,$78,$6F,$FC,$71,$18,$C3,$E0,$BA,$C3,$AD,$78
	.dc.b	$0F,$F5,$74,$FF,$80,$5B,$E2,$72,$D3,$E5,$61,$67,$FD,$25,$F4,$7F
	.dc.b	$E9,$A7,$FD,$6D,$F4,$2F,$6B,$50,$8C,$3F,$EB,$B9,$E0,$FF,$87,$1B
	.dc.b	$9F,$AA,$77,$3F,$E2,$06,$C7,$BA,$77,$87,$F9,$55,$7C,$9F,$81,$B7
	.dc.b	$DE,$3C,$77,$FE,$65,$8C,$1E,$8F,$1B,$27,$19,$91,$92,$1E,$77,$DB
	.dc.b	$F4,$7F,$D0,$23,$FD,$AC,$D7,$FC,$6E,$DC,$7D,$9E,$6F,$0E,$9F,$1F
	.dc.b	$D1,$CB,$E3,$19,$7E,$5C,$65,$E7,$1F,$33,$E8,$DE,$D5,$24,$FD,$2E
	.dc.b	$FA,$FC,$9F,$47,$E0,$98,$CB,$F0,$F7,$CA,$7E,$CC,$E9,$FF,$78,$46
	.dc.b	$5F,$F2,$8F,$D1,$FF,$9B,$DF,$F7,$A3,$5C,$FA,$5F,$80,$E4,$7C,$07
	.dc.b	$E0,$AB,$58,$FE,$5A,$5F,$13,$B3,$57,$30,$FF,$80,$13,$A6,$CE,$5F
	.dc.b	$F3,$1F,$3F,$DE,$6F,$17,$E4,$23,$18,$7E,$93,$E8,$9E,$4E,$6B,$F2
	.dc.b	$B0,$FA,$3A,$67,$0F,$F7,$75,$FE,$B3,$0E,$7D,$3F,$EA,$3B,$E1,$5A
	.dc.b	$8C,$5F,$18,$FF,$3B,$FD,$17,$F8,$BD,$D1,$FE,$F9,$70,$EB,$CD,$F1
	.dc.b	$ED,$FA,$8F,$A3,$7B,$B4,$95,$C8,$5F,$2D,$3F,$80,$FD,$1B,$97,$F6
	.dc.b	$2D,$AE,$0D,$7F,$D1,$78,$76,$7A,$7D,$1B,$E3,$3F,$72,$D8,$CC,$79
	.dc.b	$7D,$5C,$FB,$FA,$FA,$39,$5C,$A3,$F8,$0F,$D1,$3E,$8B,$F8,$1F,$D1
	.dc.b	$BE,$7E,$9A,$8F,$E0,$FF,$44,$EF,$B9,$4C,$BE,$04,$E5,$B3,$97,$FD
	.dc.b	$4B,$09,$E7,$32,$78,$C6,$F8,$7F,$93,$D7,$39,$E3,$F4,$7E,$22,$3E
	.dc.b	$0E,$BA,$E8,$F1,$C7,$A2,$3F,$8F,$FD,$1B,$E5,$E9,$2B,$EC,$7D,$1F
	.dc.b	$E4,$9F,$B5,$3B,$73,$F1,$3F,$CA,$24,$F1,$8F,$F6,$0C,$7F,$74,$79
	.dc.b	$FA,$7B,$38,$71,$D8,$C9,$0E,$2E,$5E,$71,$74,$FE,$93,$5F,$EA,$69
	.dc.b	$F3,$FB,$7E,$3F,$66,$46,$08,$79,$1C,$BE,$68,$FE,$83,$F4,$7F,$7E
	.dc.b	$3F,$30,$D7,$FD,$B6,$DB,$FE,$B3,$E7,$FE,$6F,$C2,$54,$AD,$46,$23
	.dc.b	$FD,$95,$E9,$7F,$4E,$AF,$E8,$B3,$9A,$DD,$72,$9F,$F7,$BB,$FA,$95
	.dc.b	$F4,$7C,$4E,$5E,$84,$9D,$46,$5E,$11,$97,$FD,$E7,$3C,$9D,$9A,$3B
	.dc.b	$7C,$C7,$97,$E8,$E4,$60,$87,$27,$2F,$48,$FE,$D3,$F4,$4F,$A2,$FC
	.dc.b	$44,$FF,$9B,$E3,$1E,$67,$F9,$11,$7C,$CB,$E0,$F1,$7C,$2F,$E4,$7D
	.dc.b	$27,$F2,$2D,$FF,$C9,$6D,$8F,$F3,$73,$5F,$F3,$E3,$5F,$A7,$F1,$8F
	.dc.b	$80,$FE,$0D,$5F,$C7,$23,$1F,$0F,$FE,$74,$E4,$FF,$3A,$C6,$FC,$7C
	.dc.b	$57,$2B,$FF,$63,$CE,$63,$94,$60,$EB,$82,$F9,$C6,$6F,$9C,$AF,$C3
	.dc.b	$7C,$A7,$31,$89,$E5,$3E,$1F,$A3,$CD,$BE,$8E,$C6,$1B,$3E,$46,$D7
	.dc.b	$23,$C1,$5C,$67,$F9,$A7,$D0,$FF,$B1,$A7,$F3,$2D,$F8,$77,$B2,$F5
	.dc.b	$FD,$34,$AE,$0D,$F6,$33,$1C,$6B,$86,$1D,$A7,$D3,$E2,$7A,$9B,$1F
	.dc.b	$29,$B7,$E7,$D7,$EC,$46,$37,$F9,$6D,$AF,$A8,$DA,$F6,$B6,$B8,$CF
	.dc.b	$D6,$BF,$03,$4F,$01,$6B,$C1,$A7,$C3,$AF,$C4,$57,$09,$CB,$70,$FB
	.dc.b	$AD,$BF,$C3,$36,$7E,$7B,$6B,$9C,$F8,$27,$9B,$6B,$9B,$4E,$65,$8E
	.dc.b	$6D,$9E,$71,$E8,$C3,$E0,$61,$E9,$3F,$9E,$56,$DB,$3B,$8F,$81,$7F
	.dc.b	$31,$A7,$CC,$2C,$7C,$C6,$DF,$CC,$8F,$49,$CC,$B3,$CD,$B3,$CD,$A7
	.dc.b	$32,$CF,$39,$F8,$31,$CD,$A7,$32,$CF,$36,$C7,$36,$CF,$39,$F9,$33
	.dc.b	$CD,$B3,$CD,$A7,$32,$C7,$36,$CF,$38,$FA,$27,$32,$CF,$36,$CF,$36
	.dc.b	$9C,$CB,$3C,$CF,$B3,$3C,$5B,$3C,$5B,$3C,$5A,$71,$AF,$B7,$5C,$5B
	.dc.b	$5C,$5B,$5C,$5B,$5C,$70,$FC,$A4,$60,$FC,$43,$82,$57,$E2,$5B,$5F
	.dc.b	$06,$BC,$75,$C2,$B4,$49,$D4,$75,$C3,$F2,$A4,$9F,$C5,$C7,$19,$DD
	.dc.b	$F6,$C3,$D8,$DA,$F6,$36,$7D,$8D,$C3,$D9,$3E,$1A,$E2,$D3,$89,$6B
	.dc.b	$8B,$67,$8C,$78,$A3,$17,$E5,$AE,$73,$92,$38,$24,$F6,$8C,$B6,$3E
	.dc.b	$B7,$D0,$7F,$3A,$3F,$9D,$FD,$0D,$9F,$E9,$71,$6B,$F8,$9C,$DA,$FF
	.dc.b	$73,$FC,$FB,$AB,$FD,$D3,$5F,$E0,$D8,$C6,$1F,$E0,$F8,$C4,$E9,$FF
	.dc.b	$0A,$CE,$9D,$07,$15,$DA,$7F,$AB,$F0,$7F,$42,$F1,$BF,$A0,$D6,$B0
	.dc.b	$FE,$C9,$F4,$1F,$D5,$0F,$EA,$BD,$56,$3F,$B2,$C5,$CB,$F9,$6C,$DC
	.dc.b	$BF,$B2,$CD,$AF,$ED,$D3,$6B,$FA,$BC,$7F,$60,$F7,$BF,$B0,$B4,$FE
	.dc.b	$E2,$58,$FE,$72,$DB,$F8,$FE,$1F,$C1,$9F,$73,$71,$FF,$80,$15,$F0
	.dc.b	$A7,$2D,$8C,$C7,$F6,$93,$27,$F6,$BE,$C7,$CC,$FF,$80,$30,$EF,$7F
	.dc.b	$E0,$DA,$7D,$07,$F7,$C6,$C6,$7D,$9C,$4C,$8E,$0E,$E3,$13,$FD,$8C
	.dc.b	$E2,$58,$D4,$F0,$3C,$1F,$41,$DF,$E0,$BC,$0D,$8C,$9E,$58,$CB,$4F
	.dc.b	$F5,$B7,$CF,$FF,$AE,$9F,$F6,$0F,$13,$E8,$3F,$DF,$31,$AF,$7F,$F0
	.dc.b	$B1,$8C,$BE,$35,$F1,$6C,$FF,$88,$3E,$81,$F4,$0F,$A0,$FF,$05,$DE
	.dc.b	$F7,$FF,$18,$E0,$F9,$BF,$91,$E6,$7C,$96,$E9,$F3,$5B,$3F,$15,$B5
	.dc.b	$C9,$B3,$FE,$42,$9E,$8F,$D5,$AF,$A9,$F2,$52,$3E,$C7,$CD,$DA,$FA
	.dc.b	$13,$70,$F2,$CD,$AF,$F3,$E7,$5B,$C9,$FF,$1E,$F3,$FE,$13,$D9,$FF
	.dc.b	$13,$CE,$6B,$FD,$D7,$39,$7F,$DC,$91,$9C,$BF,$1C,$DC,$53,$FE,$8C
	.dc.b	$AF,$C7,$FC,$0F,$4C,$7F,$19,$87,$E3,$66,$D7,$E4,$5B,$1F,$89,$6D
	.dc.b	$FC,$E6,$CF,$C5,$6C,$FF,$95,$4F,$F3,$A4,$7F,$BE,$63,$FD,$EC,$DE
	.dc.b	$8F,$F2,$1C,$3D,$1B,$1F,$E4,$6F,$9C,$7C,$3E,$A7,$0D,$BF,$F5,$3E
	.dc.b	$3E,$EC,$7F,$D5,$3F,$3C,$F9,$FD,$D6,$8F,$F5,$54,$65,$C4,$FF,$AC
	.dc.b	$31,$EF,$18,$84,$7F,$94,$9F,$DD,$2F,$67,$CD,$E7,$79,$DB,$5C,$67
	.dc.b	$86,$3E,$5B,$7C,$FF,$85,$E7,$87,$F9,$92,$31,$3B,$AE,$37,$FE,$C1
	.dc.b	$9C,$BF,$EC,$38,$CD,$FF,$A5,$BB,$BF,$D0,$B2,$F9,$D7,$3F,$F9,$C3
	.dc.b	$CC,$56,$A7,$35,$EF,$9C,$B6,$75,$D3,$C5,$91,$93,$E4,$9D,$E3,$38
	.dc.b	$7F,$1F,$87,$BD,$E4,$9F,$9B,$39,$8C,$70,$FF,$A1,$2B,$67,$D1,$2C
	.dc.b	$FF,$A4,$5F,$67,$4E,$EA,$46,$DF,$64,$62,$7F,$98,$7C,$FB,$E7,$FE
	.dc.b	$01,$FE,$92,$71,$7E,$AD,$2D,$70,$AD,$1F,$1F,$E3,$AD,$7E,$16,$F7
	.dc.b	$C3,$F7,$5B,$1F,$EF,$A3,$E9,$4F,$0B,$FE,$85,$ED,$FE,$D6,$75,$8B
	.dc.b	$E7,$3F,$D4,$B3,$9A,$FD,$CC,$E6,$BF,$29,$39,$93,$B8,$FC,$07,$57
	.dc.b	$2E,$B9,$1C,$1F,$C0,$46,$2B,$FA,$A7,$8D,$F8,$07,$FA,$A7,$CE,$BE
	.dc.b	$7F,$FD,$6F,$39,$DB,$F5,$D6,$AD,$63,$FA,$7A,$D2,$C2,$46,$1F,$F7
	.dc.b	$B7,$C1,$7F,$B0,$23,$10,$69,$BB,$4E,$F9,$7C,$7F,$2D,$FE,$E2,$C9
	.dc.b	$F3,$FC,$23,$0D,$CB,$C9,$1F,$CB,$33,$FD,$C4,$60,$F4,$70,$7B,$8F
	.dc.b	$F6,$AE,$9E,$9F,$9F,$F5,$7C,$FB,$C7,$F6,$4F,$9E,$B8,$5F,$23,$F8
	.dc.b	$44,$71,$15,$C5,$F8,$73,$D9,$5F,$6B,$EB,$BF,$DF,$9E,$31,$FE,$2E
	.dc.b	$9D,$47,$2C,$2B,$E4,$8B,$5C,$70,$F0,$4E,$4E,$67,$12,$99,$77,$1C
	.dc.b	$6F,$C1,$3A,$8C,$3F,$E4,$5A,$EF,$87,$F3,$1B,$EF,$3F,$05,$E7,$19
	.dc.b	$BE,$71,$8C,$3D,$35,$B9,$F8,$35,$B8,$EC,$78,$5C,$3E,$78,$DC,$F2
	.dc.b	$8F,$96,$6D,$C5,$F9,$9B,$1F,$E3,$73,$67,$F8,$E6,$30,$E0,$FB,$FF
	.dc.b	$3F,$FC,$6B,$DA,$FF,$1F,$36,$3F,$CC,$2D,$7F,$CD,$0D,$78,$FC,$4B
	.dc.b	$7E,$2D,$8E,$2D,$78,$B5,$E3,$1F,$7B,$E7,$DF,$35,$F9,$2D,$7E,$EC
	.dc.b	$62,$BF,$07,$83,$0E,$D5,$F8,$28,$C6,$1C,$F0,$E5,$A7,$E7,$31,$E5
	.dc.b	$8F,$E1,$2F,$6D,$C7,$9D,$6E,$F8,$CF,$EF,$EF,$8E,$3D,$EF,$73,$FC
	.dc.b	$0F,$0E,$73,$E8,$BE,$32,$BE,$38,$7C,$08,$F0,$63,$E7,$61,$82,$5F
	.dc.b	$37,$2F,$86,$3E,$1E,$3E,$29,$ED,$87,$83,$0E,$F1,$E5,$9D,$47,$C2
	.dc.b	$AF,$EF,$0D,$D3,$E0,$57,$F8,$17,$2F,$5E,$58,$8F,$F3,$F5,$3C,$4F
	.dc.b	$E0,$F8,$72,$6E,$1B,$8F,$F0,$27,$CF,$FF,$90,$6F,$FC,$21,$F3,$EF
	.dc.b	$67,$71,$FC,$AE,$30,$FF,$85,$67,$83,$FE,$8D,$6D,$7F,$83,$E7,$73
	.dc.b	$FE,$91,$6C,$78,$3A,$BA,$CA,$9F,$1F,$86,$C7,$78,$DD,$BC,$1D,$46
	.dc.b	$3C,$69,$5B,$21,$E1,$72,$F9,$E3,$C3,$F3,$EE,$9A,$49,$F8,$7E,$6D
	.dc.b	$7D,$5F,$3F,$E2,$31,$AF,$F8,$B6,$F9,$4F,$CA,$9D,$3F,$E3,$B8,$CB
	.dc.b	$FE,$B3,$F9,$47,$CF,$FC,$26,$C7,$F9,$05,$AF,$C9,$EF,$D4,$FC,$2F
	.dc.b	$46,$4E,$E3,$2F,$DA,$8C,$BD,$E3,$E7,$FC,$3F,$F9,$72,$3F,$C9,$CD
	.dc.b	$7F,$D8,$CD,$CB,$FC,$91,$D4,$F8,$0E,$3E,$67,$D8,$87,$C3,$D3,$38
	.dc.b	$74,$9C,$3F,$E5,$B9,$DB,$93,$FC,$BD,$D3,$F7,$9F,$F6,$B3,$70,$FF
	.dc.b	$2C,$37,$A7,$FB,$0F,$CF,$FF,$84,$2B,$8F,$27,$E2,$A1,$F3,$F4,$FC
	.dc.b	$6D,$F1,$AF,$F0,$23,$54,$F0,$7F,$CE,$0D,$7C,$75,$F9,$6F,$9F,$79
	.dc.b	$B4,$93,$C3,$4F,$CD,$FC,$F3,$E0,$29,$27,$FC,$F9,$7C,$9F,$F3,$F7
	.dc.b	$CF,$B9,$3F,$46,$D8,$CC,$71,$7F,$DF,$5F,$3E,$F9,$1F,$D4,$B6,$3F
	.dc.b	$D1,$0D,$7F,$E0,$06,$3F,$E6,$27,$3E,$EF,$0F,$9B,$91,$83,$FD,$21
	.dc.b	$F3,$EF,$9D,$FD,$AF,$CF,$BE,$7E,$9F,$B8,$9D,$57,$FA,$5F,$E7,$DD
	.dc.b	$FF,$BC,$58,$D3,$11,$27,$84,$65,$E5,$19,$7F,$AC,$74,$7F,$88,$2B
	.dc.b	$9C,$70,$36,$78,$7E,$17,$C0,$47,$C9,$CD,$5E,$67,$F9,$16,$53,$9E
	.dc.b	$AF,$F5,$34,$7E,$17,$E7,$DF,$2B,$FC,$59,$B7,$FE,$B5,$6B,$D0,$7C
	.dc.b	$FF,$E5,$5C,$46,$9F,$F1,$C8,$70,$FF,$8F,$1A,$FF,$96,$5A,$F4,$F4
	.dc.b	$7F,$93,$CF,$F6,$D8,$FF,$63,$35,$F1,$3F,$7D,$B1,$C0,$8F,$0F,$80
	.dc.b	$AE,$8E,$71,$DA,$3B,$FB,$5E,$0F,$85,$95,$E0,$8C,$BE,$28,$CB,$E2
	.dc.b	$8F,$E7,$BF,$3F,$F8,$98,$FF,$6E,$35,$F9,$0D,$D3,$FD,$AD,$E6,$FE
	.dc.b	$96,$73,$C9,$FE,$27,$AE,$11,$F2,$9A,$FF,$99,$1A,$FF,$A0,$67,$FD
	.dc.b	$F8,$FF,$BA,$D7,$CF,$FF,$50,$6C,$67,$CD,$C1,$99,$91,$92,$1F,$7E
	.dc.b	$78,$B6,$38,$BD,$FE,$5B,$B7,$CF,$C8,$C1,$0E,$4E,$5E,$91,$FD,$93
	.dc.b	$E9,$1F,$0F,$FE,$47,$6C,$FE,$28,$FF,$4F,$38,$70,$D7,$F1,$FF,$40
	.dc.b	$FA,$0F,$CA,$47,$FA,$2A,$31,$F0,$DC,$E1,$0F,$99,$7C,$36,$63,$73
	.dc.b	$F9,$AF,$8B,$7C,$17,$F1,$3E,$EF,$F5,$87,$21,$FF,$59,$4E,$BD,$DC
	.dc.b	$5F,$EC,$A3,$8F,$91,$8B,$CC,$6E,$F9,$9C,$B0,$ED,$7C,$5F,$F6,$B4
	.dc.b	$F2,$9D,$CA,$39,$3B,$BE,$D3,$CF,$0B,$3B,$BE,$B8,$78,$9E,$A3,$E7
	.dc.b	$F7,$8F,$C0,$F9,$FB,$39,$C7,$C3,$39,$6C,$67,$0F,$45,$70,$BE,$2F
	.dc.b	$8A,$1F,$3F,$FD,$B2,$FD,$7A,$BE,$91,$F4,$9B,$AD,$5F,$F1,$CA,$D4
	.dc.b	$E7,$D8,$F6,$C6,$30,$F8,$B5,$DA,$73,$87,$49,$F3,$F8,$7E,$C9,$F6
	.dc.b	$B6,$32,$7B,$23,$2D,$33,$5F,$56,$32,$D8,$CE,$DF,$36,$72,$D8,$C9
	.dc.b	$F7,$E3,$2D,$33,$3F,$5E,$32,$D8,$CE,$DF,$94,$9C,$B6,$33,$B7,$C8
	.dc.b	$BF,$BD,$87,$B7,$0F,$C6,$36,$33,$5E,$D9,$CB,$63,$31,$ED,$32,$57
	.dc.b	$27,$B5,$CB,$4C,$C7,$B6,$32,$D8,$CC,$7B,$4C,$95,$CD,$FB,$63,$2D
	.dc.b	$8C,$DF,$BE,$BF,$55,$7E,$D6,$9E,$B2,$DF,$A5,$B5,$E7,$69,$E5,$2C
	.dc.b	$78,$DB,$7E,$16,$DF,$86,$7F,$5D,$7E,$16,$CF,$85,$A7,$84,$B7,$E1
	.dc.b	$6C,$78,$5A,$78,$4B,$7E,$16,$DF,$86,$3F,$8B,$DF,$85,$B5,$E1,$6D
	.dc.b	$78,$5B,$7E,$16,$CF,$85,$B3,$E1,$69,$E1,$2D,$F8,$4F,$F0,$5C,$F7
	.dc.b	$6C,$F7,$6D,$77,$6C,$F7,$6C,$F7,$6D,$77,$69,$DE,$FD,$CE,$30,$FD
	.dc.b	$1B,$9B,$F1,$4E,$88,$E0,$97,$E5,$8C,$B6,$33,$87,$9E,$72,$D8,$C9
	.dc.b	$E9,$8C,$B4,$CC,$FA,$E3,$2D,$8C,$E1,$ED,$9C,$B6,$33,$7E,$BA,$F2
	.dc.b	$12,$FC,$4D,$9F,$03,$4E,$C5,$AE,$8F,$D3,$FA,$1F,$E7,$A7,$F3,$EF
	.dc.b	$A1,$B5,$FC,$F2,$2C,$FF,$53,$9B,$7F,$D7,$A1,$F4,$0E,$AF,$E9,$F2
	.dc.b	$FA,$4F,$E9,$F8,$AE,$5E,$E3,$FA,$DD,$6F,$97,$FA,$34,$E6,$BF,$CB
	.dc.b	$73,$9A,$FE,$7B,$39,$8F,$ED,$97,$DA,$BF,$AF,$7D,$17,$F5,$83,$FA
	.dc.b	$CF,$45,$C3,$FA,$BC,$59,$FE,$DD,$36,$FF,$B2,$4D,$AF,$EE,$93,$6B
	.dc.b	$FB,$54,$7F,$64,$F7,$BF,$B2,$B4,$FE,$C6,$58,$FE,$F4,$DB,$F9,$18
	.dc.b	$BE,$23,$FA,$CF,$D1,$3A,$BE,$0C,$61,$B7,$FE,$19,$AE,$0D,$9F,$A5
	.dc.b	$87,$CD,$3E,$9B,$87,$FC,$0C,$3E,$8B,$FB,$EB,$67,$FB,$DB,$FE,$B5
	.dc.b	$6C,$68,$8E,$BD,$8B,$3F,$E1,$D3,$91,$67,$AC,$6B,$2F,$F0,$FC,$BE
	.dc.b	$8B,$C3,$F0,$BE,$26,$E1,$E2,$6C,$F5,$6D,$7E,$3F,$E7,$BF,$DA,$15
	.dc.b	$FE,$22,$F9,$37,$D1,$3E,$23,$FC,$1D,$5A,$C7,$FD,$47,$5A,$9C,$FC
	.dc.b	$0F,$C5,$95,$6D,$FF,$4C,$B5,$F9,$8D,$7E,$5F,$91,$F2,$BC,$BF,$96
	.dc.b	$DC,$BE,$5B,$67,$E3,$37,$3F,$86,$D9,$FF,$24,$CF,$F9,$89,$FA,$75
	.dc.b	$F5,$BE,$42,$47,$D6,$F8,$DB,$5F,$46,$6E,$1F,$3A,$6D,$7E,$0F,$E7
	.dc.b	$FF,$18,$FF,$8B,$B9,$BF,$D8,$6F,$F8,$9A,$33,$87,$F8,$B6,$72,$FF
	.dc.b	$AF,$E3,$2F,$FA,$F1,$AF,$5E,$6F,$95,$18,$8F,$C7,$9F,$8E,$F1,$3F
	.dc.b	$1E,$DB,$FC,$7B,$4F,$C5,$17,$1F,$C3,$B6,$3E,$2B,$4F,$E1,$53,$F9
	.dc.b	$77,$FD,$A8,$D9,$FE,$28,$BC,$FF,$F2,$83,$9A,$FD,$37,$CE,$3E,$1B
	.dc.b	$35,$FE,$55,$8C,$4F,$F9,$2A,$B9,$3F,$E4,$CF,$9E,$7C,$FE,$F0,$FF
	.dc.b	$2E,$46,$27,$FC,$A3,$7C,$B2,$FF,$74,$2F,$8B,$ED,$EC,$77,$6D,$70
	.dc.b	$70,$FF,$96,$BE,$7F,$A5,$F2,$9F,$F3,$54,$63,$5F,$E6,$58,$75,$BE
	.dc.b	$19,$F2,$C3,$FD,$F1,$E4,$3B,$3F,$D0,$74,$F9,$D7,$37,$F9,$D6,$B5
	.dc.b	$3F,$E8,$E3,$51,$9A,$91,$9A,$FC,$C5,$F6,$9C,$B6,$75,$CF,$8C,$FF
	.dc.b	$A4,$43,$DB,$F1,$CF,$C7,$9C,$C6,$38,$7F,$D1,$15,$B3,$FC,$FA,$5B
	.dc.b	$57,$FA,$C2,$3D,$9C,$FC,$2E,$47,$07,$E0,$E4,$9C,$DF,$2A,$FE,$55
	.dc.b	$F3,$CF,$9F,$D9,$FE,$99,$71,$1F,$06,$30,$47,$E2,$7C,$75,$AE,$87
	.dc.b	$F6,$1A,$E2,$E0,$38,$CB,$D5,$8A,$B9,$47,$F6,$FF,$5F,$F6,$EE,$BE
	.dc.b	$73,$FD,$59,$18,$C7,$FB,$71,$F7,$63,$3D,$72,$3E,$C7,$4D,$CF,$C5
	.dc.b	$7C,$5F,$EE,$CD,$9F,$C3,$4E,$F0,$FF,$00,$78,$3F,$86,$BF,$F5,$A7
	.dc.b	$CE,$BE,$7F,$FD,$7D,$39,$AF,$F7,$1D,$AB,$58,$7F,$AD,$AB,$4F,$F5
	.dc.b	$96,$E1,$FE,$F4,$F8,$1F,$F3,$62,$8C,$B7,$97,$B9,$FE,$C2,$9D,$CE
	.dc.b	$87,$CC,$7F,$85,$3E,$7F,$0D,$AF,$3B,$30,$FF,$69,$E2,$EE,$C6,$5F
	.dc.b	$5F,$17,$07,$B8,$FF,$6E,$47,$F8,$77,$E7,$FF,$C3,$DF,$3E,$F1,$7D
	.dc.b	$95,$BC,$BE,$03,$C2,$D7,$C2,$B7,$5F,$ED,$A9,$E8,$2F,$A3,$FE,$2B
	.dc.b	$C3,$C7,$3C,$63,$FC,$04,$74,$38,$57,$46,$9A,$BE,$0F,$CD,$C1,$58
	.dc.b	$8F,$0E,$1E,$07,$51,$97,$B4,$67,$0F,$1C,$76,$3F,$C2,$97,$DA,$33
	.dc.b	$3C,$E7,$AE,$1E,$58,$C4,$74,$8C,$1E,$C7,$07,$C1,$C3,$CE,$F7,$39
	.dc.b	$99,$C3,$CD,$87,$86,$77,$7D,$A3,$13,$AA,$F4,$61,$E6,$C3,$C1,$5A
	.dc.b	$9E,$C7,$CD,$3C,$85,$36,$60,$C1,$FE,$61,$F9,$FF,$E4,$1E,$C7,$06
	.dc.b	$C7,$B5,$AF,$DA,$6B,$FE,$3D,$5F,$0F,$FC,$82,$38,$35,$E0,$D7,$83
	.dc.b	$5F,$13,$C7,$2F,$B9,$F3,$EF,$9A,$FF,$25,$46,$1B,$97,$2E,$5F,$C3
	.dc.b	$E3,$E2,$7F,$13,$7C,$71,$B7,$C1,$EF,$3C,$31,$F0,$61,$FE,$57,$BE
	.dc.b	$15,$F8,$76,$E3,$F8,$FC,$7B,$4F,$E2,$DB,$1E,$58,$EE,$75,$C2,$3C
	.dc.b	$F1,$F1,$2B,$EB,$87,$AE,$3A,$63,$E4,$7F,$D2,$D3,$A6,$D7,$2C,$7C
	.dc.b	$37,$C7,$1F,$15,$77,$3C,$2E,$31,$FF,$53,$61,$EE,$6E,$1F,$E7,$77
	.dc.b	$FD,$55,$87,$4B,$EE,$7F,$A9,$E3,$8F,$8D,$BD,$BE,$43,$67,$67,$F4
	.dc.b	$BF,$9F,$FF,$45,$3C,$67,$FD,$09,$F3,$FF,$E8,$F8,$C1,$FE,$B3,$70
	.dc.b	$F1,$EB,$E8,$AB,$9B,$FD,$5E,$77,$87,$FB,$05,$B7,$FE,$91,$AD,$CF
	.dc.b	$FB,$15,$7C,$87,$46,$DF,$88,$ED,$1E,$0B,$FE,$E9,$19,$9D,$BE,$1F
	.dc.b	$17,$47,$3A,$91,$CE,$BF,$C3,$6D,$9C,$BF,$DB,$FE,$7B,$FA,$1B,$E1
	.dc.b	$6B,$9E,$8F,$17,$99,$F0,$A7,$39,$7B,$FE,$7F,$CC,$66,$A5,$F8,$63
	.dc.b	$2F,$58,$F9,$7F,$27,$FE,$95,$8F,$80,$D7,$FC,$04,$DA,$F9,$9D,$9C
	.dc.b	$3D,$3F,$89,$EB,$4B,$E3,$19,$7E,$04,$65,$E7,$1F,$6B,$E7,$FF,$CB
	.dc.b	$D1,$F3,$DA,$FF,$84,$5B,$7F,$EF,$BE,$57,$B7,$C7,$CC,$FC,$1C,$3E
	.dc.b	$12,$99,$C3,$A4,$E1,$FB,$93,$B7,$27,$DF,$E7,$FB,$78,$7F,$38,$8F
	.dc.b	$F3,$27,$27,$E4,$27,$3C,$DF,$ED,$BF,$9E,$79,$19,$AF,$CA,$C3,$E7
	.dc.b	$E9,$9C,$3B,$4E,$1F,$C7,$4E,$DC,$9F,$91,$F9,$FF,$F1,$73,$7C,$BF
	.dc.b	$D0,$C6,$23,$FC,$6D,$3F,$1B,$1F,$F7,$77,$CE,$BD,$BF,$D3,$B6,$BF
	.dc.b	$38,$D7,$FC,$7D,$7C,$F5,$F8,$1F,$3E,$E5,$FD,$7B,$6B,$F4,$8D,$7F
	.dc.b	$C8,$D7,$CE,$BE,$47,$CF,$BE,$1B,$F6,$ED,$AF,$D6,$35,$E3,$EA,$EB
	.dc.b	$E7,$F7,$F3,$92,$77,$1F,$C0,$3E,$75,$F3,$FF,$C0,$BE,$7D,$F3,$74
	.dc.b	$D4,$7F,$06,$F8,$A7,$C9,$52,$63,$FB,$DF,$CD,$1F,$1F,$88,$FF,$30
	.dc.b	$43,$EE,$7F,$87,$B6,$3F,$82,$5F,$46,$DE,$AA,$46,$0F,$E1,$5F,$40
	.dc.b	$F1,$BF,$8C,$AD,$E9,$88,$93,$C2,$32,$F2,$8E,$3F,$43,$23,$AB,$D3
	.dc.b	$B7,$11,$FE,$79,$B4,$E3,$97,$F3,$91,$F9,$A9,$0F,$A1,$7C,$B7,$F2
	.dc.b	$E6,$C7,$F2,$36,$BF,$E8,$5F,$A1,$FD,$0B,$99,$FE,$6B,$0F,$8C,$A7
	.dc.b	$38,$93,$FC,$B6,$74,$D9,$CB,$D7,$D8,$E6,$47,$87,$45,$74,$73,$8F
	.dc.b	$0C,$78,$BE,$03,$C7,$EF,$4A,$F0,$46,$5F,$14,$65,$F6,$47,$F4,$EF
	.dc.b	$A1,$FC,$6C,$7F,$43,$6B,$FE,$A5,$6E,$BF,$C1,$FD,$4F,$D5,$78,$9F
	.dc.b	$32,$95,$A8,$C4,$FF,$54,$9F,$A2,$D7,$7D,$5E,$13,$FA,$FC,$64,$D5
	.dc.b	$73,$9E,$3E,$A6,$BE,$86,$47,$F5,$9A,$4F,$23,$31,$FE,$BD,$84,$E6
	.dc.b	$38,$75,$F8,$3E,$86,$46,$08,$71,$72,$F3,$8F,$EE,$BF,$3E,$FA,$0F
	.dc.b	$C2,$CF,$FA,$02,$31,$E6,$FF,$95,$17,$73,$E1,$B1,$7C,$2F,$E5,$7D
	.dc.b	$2F,$F2,$EE,$0F,$F2,$DB,$63,$FC,$E0,$D7,$FD,$1A,$D7,$E9,$7E,$CD
	.dc.b	$FC,$0A,$7F,$CC,$6D,$8F,$F3,$C3,$5F,$F4,$AB,$5F,$D2,$FC,$7B,$E1
	.dc.b	$3F,$8E,$57,$F8,$5A,$31,$F0,$1F,$E7,$EE,$5F,$F3,$FC,$6F,$CB,$C5
	.dc.b	$72,$BF,$F6,$D4,$E6,$39,$46,$0E,$B8,$7F,$BD,$27,$AC,$EB,$E8,$B9
	.dc.b	$CA,$7C,$9F,$45,$87,$CD,$36,$D3,$63,$0D,$BF,$0B,$D2,$DF,$45,$FE
	.dc.b	$DC,$3F,$B6,$4F,$F8,$D3,$E2,$5E,$EB,$D9,$E1,$5B,$F7,$29,$E6,$AD
	.dc.b	$CE,$2F,$FC,$59,$5B,$7A,$5F,$37,$37,$BC,$1B,$7F,$73,$BF,$1C,$62
	.dc.b	$BF,$D9,$D7,$C6,$30,$7F,$1C,$9E,$CE,$0F,$6D,$F9,$67,$85,$F5,$8F
	.dc.b	$03,$6F,$CB,$86,$1F,$DE,$61,$D2,$7F,$D7,$B8,$FB,$A7,$D2,$4C,$7C
	.dc.b	$15,$E8,$AD,$E3,$E0,$AE,$78,$F5,$BF,$43,$E5,$C7,$11,$E8,$9E,$97
	.dc.b	$CA,$73,$87,$8E,$7E,$1C,$62,$FA,$63,$E5,$C3,$C7,$19,$AF,$69,$D0
	.dc.b	$F1,$96,$13,$C7,$0F,$89,$1C,$A7,$53,$97,$8C,$64,$93,$E7,$8E,$91
	.dc.b	$EB,$32,$FA,$0C,$E1,$EB,$8F,$0E,$3F,$1D,$F8,$B3,$AC,$3E,$33,$4F
	.dc.b	$1C,$EE,$7B,$5F,$A0,$93,$DE,$3C,$27,$42,$E1,$EB,$71,$3C,$1B,$3C
	.dc.b	$63,$11,$DA,$3E,$0C,$7B,$E3,$A9,$93,$89,$EC,$C7,$E9,$3E,$0B,$F4
	.dc.b	$63,$F0,$63,$86,$1E,$37,$91,$23,$F1,$98,$78,$CF,$09,$23,$BC,$F1
	.dc.b	$9E,$87,$08,$F1,$4F,$C9,$7C,$57,$E4,$C3,$E0,$5F,$1C,$71,$7D,$EF
	.dc.b	$D6,$78,$E3,$E9,$BF,$03,$0F,$B5,$5C,$6B,$B5,$FC,$53,$E4,$CE,$B1
	.dc.b	$F8,$38,$76,$C3,$CA,$73,$7F,$37,$1A,$BE,$67,$8E,$7C,$4F,$E0,$B1
	.dc.b	$E5,$1F,$2A,$78,$3C,$E7,$F4,$2D,$3E,$0B,$D1,$F9,$31,$C3,$0F,$3D
	.dc.b	$FB,$9B,$5D,$CE,$15,$DE,$FC,$38,$7C,$FC,$7E,$14,$F7,$BF,$85,$7F
	.dc.b	$5F,$5F,$E9,$DE,$27,$F3,$9C,$BF,$9D,$36,$7F,$A3,$B6,$FF,$A3,$6D
	.dc.b	$FE,$38,$9F,$E9,$75,$FD,$07,$E8,$64,$7F,$4E,$FA,$1B,$7C,$26,$E5
	.dc.b	$FD,$63,$05,$62,$FF,$C3,$5F,$3F,$FE,$66,$7F,$DC,$8D,$C3,$F9,$7E
	.dc.b	$9F,$D1,$3C,$6F,$E9,$53,$9C,$7F,$D5,$4D,$BE,$7D,$9F,$CC,$6B,$6C
	.dc.b	$8E,$AE,$5A,$7F,$66,$FA,$0F,$EB,$87,$F5,$DE,$BB,$19,$26,$5F,$12
	.dc.b	$32,$FC,$68,$FE,$C9,$F0,$3F,$D9,$5A,$7F,$72,$2C,$6D,$FE,$EB,$1E
	.dc.b	$CF,$A0,$FF,$62,$4F,$C4,$6B,$FE,$6A,$C9,$18,$84,$62,$BF,$C8,$91
	.dc.b	$C5,$F9,$73,$FD,$F3,$B5,$F3,$9F,$E0,$2D,$7E,$4C,$F8,$6B,$67,$F9
	.dc.b	$E2,$0F,$F9,$EA,$B8,$F5,$6F,$0F,$F0,$50,$FA,$0F,$F0,$33,$6B,$FC
	.dc.b	$34,$FF,$AF,$DB,$1A,$23,$AF,$76,$CF,$4F,$A0,$E7,$F0,$FD,$1B,$5D
	.dc.b	$1B,$3B,$8E,$F3,$FE,$0F,$F9,$FF,$F6,$B4,$7F,$83,$7E,$81,$F1,$39
	.dc.b	$F8,$4F,$F0,$54,$62,$7E,$01,$FE,$6D,$AE,$38,$7F,$7D,$6E,$DF,$12
	.dc.b	$1F,$4C,$FA,$67,$D3,$75,$C3,$9D,$7F,$A8,$B0,$E3,$1D,$72,$FF,$17
	.dc.b	$E1,$C6,$F8,$5F,$CC,$F9,$FF,$97,$E7,$FC,$C6,$E1,$F3,$1B,$3B,$8F
	.dc.b	$9F,$3F,$4B,$E5,$A4,$7D,$2F,$9F,$B5,$F6,$26,$EB,$FE,$4C,$84,$E2
	.dc.b	$BF,$D1,$DD,$AF,$2B,$F1,$5D,$3F,$EC,$09,$F6,$DF,$FA,$CE,$73,$EE
	.dc.b	$7F,$8F,$EF,$85,$69,$B5,$FE,$2A,$E9,$FC,$9F,$C3,$6F,$09,$7B,$7B
	.dc.b	$C6,$2B,$FC,$5D,$5F,$8B,$6F,$47,$F9,$36,$3F,$D2,$77,$E4,$75,$A7
	.dc.b	$D7,$A7,$D0,$BE,$43,$87,$AD,$FE,$55,$8C,$61,$C6,$3F,$CB,$1F,$46
	.dc.b	$FA,$2D,$5E,$1B,$5F,$0A,$FF,$C0,$83,$DC,$E9,$E7,$F5,$6C,$EE,$3F
	.dc.b	$CC,$9F,$47,$D6,$30,$7C,$C7,$19,$FF,$93,$6F,$FC,$21,$18,$D3,$D9
	.dc.b	$1F,$E1,$3F,$3D,$E0,$7F,$9E,$B0,$7D,$0B,$A7,$FC,$E5,$5A,$9F,$CA
	.dc.b	$9A,$8C,$D4,$8C,$D3,$AB,$19,$FF,$97,$F0,$E4,$DB,$F0,$56,$EF,$F0
	.dc.b	$B5,$AC,$BE,$94,$3D,$FF,$36,$3F,$43,$1F,$F3,$54,$E4,$F2,$C6,$6B
	.dc.b	$8D,$7F,$A2,$E7,$55,$FE,$74,$7D,$FD,$59,$99,$1F,$96,$37,$5F,$B9
	.dc.b	$76,$66,$3F,$D2,$DF,$44,$FA,$3F,$59,$82,$CF,$E8,$23,$4D,$8F,$83
	.dc.b	$F2,$9B,$8C,$13,$1F,$36,$89,$D4,$7F,$A2,$7D,$DF,$D2,$E5,$E5,$A7
	.dc.b	$D0,$7F,$AA,$67,$38,$2F,$F9,$AD,$6A,$3F,$67,$3A,$9D,$36,$3F,$07
	.dc.b	$D3,$CF,$B2,$47,$07,$EC,$C6,$2B,$FD,$39,$E8,$FD,$9A,$FE,$85,$F4
	.dc.b	$2F,$A3,$FF,$5E,$4E,$6B,$FD,$DD,$39,$AF,$E0,$B3,$99,$FF,$3F,$F2
	.dc.b	$7F,$53,$F8,$5F,$F2,$34,$62,$B7,$3D,$0D,$7B,$16,$39,$FC,$D7,$FB
	.dc.b	$5B,$81,$F4,$7D,$72,$4E,$63,$0F,$98,$FE,$C5,$9F,$ED,$6F,$99,$E6
	.dc.b	$E1,$F5,$9F,$ED,$7C,$FE,$77,$D1,$FC,$FF,$A3,$79,$3F,$46,$7F,$D3
	.dc.b	$35,$C2,$F9,$1F,$C9,$23,$88,$AE,$2F,$D9,$CF,$EC,$3F,$EC,$93,$71
	.dc.b	$A7,$AC,$F8,$6F,$FB,$EC,$30,$C3,$CF,$0F,$EA,$97,$C6,$39,$D7,$9F
	.dc.b	$0E,$F3,$99,$E4,$79,$67,$55,$C9,$EF,$7D,$A7,$C6,$6C,$FF,$07,$B9
	.dc.b	$BE,$B3,$C2,$74,$70,$BF,$05,$6E,$FB,$4F,$23,$77,$C9,$C5,$7C,$38
	.dc.b	$C4,$7D,$1A,$E5,$87,$96,$3D,$E7,$C3,$70,$FB,$AB,$95,$F8,$23,$91
	.dc.b	$E0,$70,$FD,$83,$E7,$9F,$18,$B5,$87,$47,$E0,$FE,$8B,$F8,$E7,$B9
	.dc.b	$C1,$B1,$F6,$1A,$FF,$8D,$9A,$F1,$F8,$B6,$FC,$5B,$1C,$5A,$F1,$6B
	.dc.b	$C6,$3E,$B7,$D1,$BE,$7F,$F2,$D9,$FE,$2E,$31,$5F,$6F,$06,$1D,$AB
	.dc.b	$F2,$11,$8C,$39,$E1,$CB,$6F,$D0,$63,$CB,$1F,$BB,$7B,$6E,$3C,$EB
	.dc.b	$77,$C6,$7F,$8C,$5F,$1C,$7B,$DE,$E7,$F8,$E6,$1C,$E7,$D1,$7C,$65
	.dc.b	$7C,$70,$F8,$11,$E0,$C7,$CE,$C3,$04,$BE,$6E,$5F,$0C,$7C,$3C,$7C
	.dc.b	$53,$DB,$0F,$06,$1D,$E3,$CB,$3A,$8F,$85,$5F,$E0,$D6,$EB,$F0,$2B
	.dc.b	$FC,$4B,$97,$AF,$2C,$47,$DE,$9F,$97,$E2,$7F,$0B,$C3,$93,$63,$71
	.dc.b	$FE,$19,$FA,$3F,$F1,$95,$7F,$8C,$3E,$8F,$FD,$06,$DF,$6F,$83,$FD
	.dc.b	$16,$76,$F7,$31,$D7,$A5,$5C,$9D,$18,$AD,$57,$2B,$F1,$7C,$9E,$09
	.dc.b	$5E,$3F,$0E,$BC,$B0,$FF,$08,$3E,$7F,$1B,$37,$CA,$30,$43,$C3,$83
	.dc.b	$0F,$37,$D1,$FF,$3C,$9F,$F2,$7B,$5F,$3B,$6B,$FC,$7D,$83,$CE,$E3
	.dc.b	$D5,$EF,$87,$D1,$D3,$38,$74,$9C,$3F,$36,$32,$FC,$4F,$96,$FF,$44
	.dc.b	$BF,$E6,$06,$B9,$F7,$3B,$78,$1F,$36,$73,$97,$93,$D2,$F1,$19,$A9
	.dc.b	$7F,$6E,$32,$F5,$8F,$AD,$F1,$3F,$E5,$78,$FF,$37,$B5,$FA,$ED,$AF
	.dc.b	$F2,$F7,$53,$E0,$B8,$FA,$7F,$7E,$1E,$25,$33,$87,$49,$C3,$FE,$7C
	.dc.b	$9D,$B9,$3F,$CF,$FD,$3F,$55,$BC,$7F,$8B,$8C,$47,$E2,$BE,$89,$E4
	.dc.b	$E6,$BF,$21,$0F,$A3,$A6,$6B,$99,$FE,$8C,$8D,$36,$34,$FE,$5B,$E8
	.dc.b	$FF,$C5,$4F,$FA,$49,$AE,$7C,$DE,$F9,$FE,$7E,$73,$B7,$FA,$0F,$E8
	.dc.b	$9E,$FF,$E8,$DB,$87,$3C,$3F,$4D,$F4,$6E,$4A,$49,$FF,$50,$D7,$07
	.dc.b	$F5,$3F,$46,$F8,$7F,$D7,$36,$3F,$D4,$ED,$7F,$5F,$3F,$E6,$37,$3E
	.dc.b	$EF,$F4,$D9,$34,$91,$FD,$53,$E8,$DF,$45,$FD,$4D,$B1,$98,$C1,$FD
	.dc.b	$5F,$E2,$DF,$43,$FC,$05,$B3,$FE,$BA,$6B,$FC,$0B,$E8,$9F,$1F,$BA
	.dc.b	$E4,$49,$FE,$CB,$5C,$1F,$F1,$20,$79,$F0,$FF,$AC,$6D,$F4,$5F,$C4
	.dc.b	$5B,$5E,$1B,$FE,$29,$F4,$2F,$02,$FD,$14,$AF,$F6,$7C,$E5,$B3,$97
	.dc.b	$F8,$CC,$27,$9C,$F4,$F8,$89,$5C,$9E,$FD,$78,$8E,$31,$73,$FF,$64
	.dc.b	$61,$F9,$C7,$FD,$95,$F4,$4F,$96,$FE,$54,$DB,$FF,$71,$B5,$FE,$57
	.dc.b	$F4,$5F,$A2,$73,$3F,$E0,$99,$D4,$7C,$86,$BF,$E4,$D6,$BF,$E7,$F7
	.dc.b	$A5,$7F,$B6,$E9,$F1,$7F,$CE,$DB,$5D,$8F,$F7,$AB,$FD,$F9,$BE,$B6
	.dc.b	$CE,$1F,$09,$63,$24,$3D,$2F,$5F,$7F,$B7,$C4,$C9,$EB,$19,$7B,$C6
	.dc.b	$5F,$83,$1F,$D3,$FE,$85,$F4,$54,$93,$FE,$9B,$CF,$FD,$ED,$EA,$7E
	.dc.b	$DB,$D8,$FF,$37,$5F,$0A,$FF,$19,$B5,$FF,$31,$B5,$FC,$8B,$5E,$37
	.dc.b	$FD,$7E,$BF,$AF,$5F,$6F,$5F,$C3,$5F,$D8,$27,$FA,$43,$E3,$C3,$FC
	.dc.b	$FD,$0F,$A2,$E4,$73,$F5,$65,$6C,$FF,$5A,$4E,$DC,$1F,$EF,$19,$DB
	.dc.b	$DF,$B3,$C9,$F4,$52,$37,$5F,$B0,$6C,$E5,$F3,$4F,$AE,$3F,$BD,$7D
	.dc.b	$37,$FB,$D6,$7F,$CA,$ED,$7F,$D0,$ED,$7F,$6A,$D7,$FD,$35,$F4,$6F
	.dc.b	$A3,$FC,$B4,$7E,$EA,$31,$E7,$FF,$98,$9F,$DE,$7C,$4D,$AD,$D7,$FA
	.dc.b	$8F,$E8,$9F,$5E,$DF,$FC,$D3,$18,$6C,$66,$7F,$D3,$73,$97,$F8,$4F
	.dc.b	$C8,$3E,$1B,$F9,$55,$FC,$D8,$C7,$C1,$71,$E4,$E5,$1B,$F3,$71,$5F
	.dc.b	$EE,$58,$DB,$C6,$F9,$E1,$FE,$23,$9C,$B6,$33,$7D,$1F,$F7,$54,$F8
	.dc.b	$FE,$8F,$AD,$23,$CD,$F4,$78,$75,$87,$78,$C5,$AE,$C6,$1B,$3F,$86
	.dc.b	$FA,$3F,$EE,$AF,$E2,$F7,$7D,$3F,$F3,$88,$D7,$B9,$E2,$9C,$F9,$5E
	.dc.b	$4D,$18,$72,$7F,$C8,$13,$AB,$E2,$75,$C3,$A4,$FF,$90,$2B,$86,$BF
	.dc.b	$56,$F8,$4E,$71,$F0,$5F,$0A,$FB,$53,$E0,$9C,$E3,$E3,$C3,$9E,$2A
	.dc.b	$63,$D3,$0E,$35,$A9,$CE,$3D,$6F,$FD,$DD,$39,$BE,$11,$8C,$7B,$BE
	.dc.b	$CB,$F7,$DF,$17,$C1,$7C,$63,$11,$E7,$9F,$85,$8F,$9E,$F9,$CF,$BE
	.dc.b	$72,$75,$C7,$E2,$5F,$2C,$3A,$61,$C6,$B7,$7C,$4F,$1C,$65,$B1,$CC
	.dc.b	$F8,$98,$FB,$23,$CF,$8F,$8B,$1F,$8A,$DA,$F8,$07,$68,$CB,$4C,$CF
	.dc.b	$63,$C5,$1E,$8B,$F1,$36,$7C,$53,$F4,$27,$31,$E4,$C7,$DD,$7E,$1A
	.dc.b	$F2,$CF,$76,$9D,$EF,$C6,$D8,$F1,$C7,$78,$F1,$C7,$5B,$EF,$1E,$18
	.dc.b	$F1,$E1,$F0,$E3,$EB,$1C,$CE,$F5,$CC,$EE,$59,$EE,$75,$39,$E1,$EA
	.dc.b	$7D,$13,$D2,$78,$3D,$23,$83,$E6,$3C,$87,$DB,$3F,$C3,$B8,$FE,$B5
	.dc.b	$B8,$FE,$89,$FA,$26,$27,$FC,$46,$FC,$63,$06,$27,$93,$E8,$3C,$43
	.dc.b	$E3,$5E,$FB,$DD,$F8,$F3,$E9,$BE,$15,$A9,$CD,$FE,$AE,$B4,$73,$BE
	.dc.b	$57,$F8,$08,$C9,$AB,$E5,$19,$AE,$47,$ED,$63,$33,$F1,$4E,$31,$9A
	.dc.b	$EB,$19,$8E,$38,$7F,$28,$AE,$38,$FD,$8C,$3E,$05,$F8,$70,$E3,$39
	.dc.b	$8F,$86,$67,$1F,$6E,$1F,$88,$C3,$9E,$1C,$B1,$EE,$FF,$73,$8C,$DF
	.dc.b	$29,$CB,$FE,$E3,$8C,$BE,$AA,$E9,$8F,$8A,$FA,$46,$30,$F0,$C7,$EA
	.dc.b	$2F,$8E,$1D,$31,$F4,$46,$2F,$78,$F7,$C7,$A3,$E3,$8E,$98,$7C,$F6
	.dc.b	$C7,$C9,$6D,$FC,$38,$F3,$4F,$B6,$B9,$C6,$0F,$CC,$5F,$86,$7A,$1B
	.dc.b	$C7,$C7,$7E,$6B,$E3,$8F,$F3,$1E,$0F,$EE,$1E,$47,$F5,$8C,$FF,$AC
	.dc.b	$B7,$1F,$ED,$4D,$CF,$FB,$43,$71,$FE,$D1,$F4,$9F,$D8,$0F,$EC,$3F
	.dc.b	$49,$6F,$FB,$94,$5B,$FE,$E1,$36,$BF,$A5,$FD,$1F,$FA,$42,$FF,$C5
	.dc.b	$90,$FA,$7F,$EE,$B1,$FD,$2E,$D5,$B5,$5F,$E0,$5A,$DC,$7F,$51,$9D
	.dc.b	$3F,$E9,$66,$BC,$DA,$EF,$BB,$F7,$8D,$BF,$F1,$DC,$F7,$BF,$F0,$D7
	.dc.b	$07,$ED,$BC,$BF,$EE,$53,$9A,$FF,$09,$73,$BE,$97,$FC,$06,$7F,$81
	.dc.b	$3D,$5B,$3F,$E1,$58,$B9,$7F,$84,$E6,$E5,$FE,$13,$AF,$F0,$5F,$C2
	.dc.b	$7F,$83,$1A,$7F,$87,$4B,$5F,$E1,$B6,$CF,$CC,$F1,$FF,$C1,$1F,$3C
	.dc.b	$EE,$F3,$C6,$1B,$5F,$C6,$2F,$E6,$CF,$C2,$6C,$67,$44,$E5,$B1,$F1
	.dc.b	$A3,$BF,$77,$4A,$FF,$18,$0F,$A5,$FF,$14,$36,$BF,$C7,$CF,$FB,$BD
	.dc.b	$B1,$A2,$3A,$F7,$AC,$F4,$FA,$5E,$7F,$13,$D1,$B5,$D1,$B3,$B8,$EF
	.dc.b	$3F,$E2,$9F,$A3,$7C,$27,$F1,$7F,$A5,$7C,$5E,$63,$0D,$BF,$8D,$5C
	.dc.b	$1B,$3F,$1E,$31,$87,$F4,$26,$F0,$FF,$8A,$1B,$5F,$E2,$5F,$A4,$7D
	.dc.b	$2B,$E9,$7B,$61,$F2,$3E,$87,$E3,$FA,$5F,$21,$B8,$7C,$86,$CE,$E3
	.dc.b	$E5,$CF,$CE,$F9,$69,$1F,$3B,$E5,$AD,$7D,$49,$BA,$7F,$9B,$21,$38
	.dc.b	$AF,$D8,$76,$FE,$25,$FF,$2B,$74,$BE,$03,$FC,$A5,$39,$BF,$F2,$8C
	.dc.b	$E6,$BF,$D4,$73,$99,$FD,$D1,$FE,$DA,$EA,$FC,$74,$66,$3B,$C6,$0E
	.dc.b	$A7,$E6,$BE,$22,$D7,$FA,$BB,$CC,$FC,$45,$7F,$70,$6B,$D9,$AF,$F0
	.dc.b	$96,$BF,$E3,$2F,$63,$FC,$D5,$07,$FC,$D9,$F4,$4F,$8F,$CC,$7F,$9B
	.dc.b	$FE,$95,$F4,$BA,$9C,$37,$1F,$F6,$0D,$FF,$36,$1F,$01,$D3,$D1,$EA
	.dc.b	$D9,$DC,$7F,$9E,$FE,$97,$AC,$60,$FF,$65,$B8,$D3,$F8,$FC,$7F,$23
	.dc.b	$BF,$E8,$73,$9E,$3F,$F3,$AC,$7F,$46,$F4,$1D,$FF,$E9,$AB,$7D,$1B
	.dc.b	$A3,$FD,$15,$5A,$9F,$F6,$E1,$A8,$CD,$48,$CD,$3A,$71,$9F,$90,$FF
	.dc.b	$59,$56,$32,$FF,$60,$5F,$82,$B0,$FF,$3B,$1F,$07,$E5,$8F,$A1,$1F
	.dc.b	$EE,$A3,$27,$91,$CD,$F7,$3F,$DA,$13,$FE,$A3,$9D,$C7,$8D,$F7,$F4
	.dc.b	$E6,$64,$75,$7E,$26,$3F,$A0,$8C,$1F,$EF,$AF,$A5,$7D,$2E,$8F,$F5
	.dc.b	$83,$88,$F8,$71,$82,$3F,$1B,$E5,$2E,$3E,$7A,$E0,$DA,$F0,$CE,$B3
	.dc.b	$FF,$56,$FC,$07,$EA,$DF,$93,$8B,$E8,$BF,$D8,$13,$9C,$BF,$D5,$C7
	.dc.b	$DC,$8C,$F5,$C8,$FA,$FD,$57,$3F,$91,$7C,$5F,$F5,$B3,$67,$F0,$B3
	.dc.b	$BC,$7F,$D7,$FE,$8F,$E1,$6F,$FD,$9B,$F4,$6F,$A5,$FF,$69,$4E,$6B
	.dc.b	$FD,$95,$39,$C3,$FD,$85,$1C,$B8,$F6,$7F,$B6,$FE,$27,$FC,$B1,$7C
	.dc.b	$6B,$FB,$F3,$71,$FF,$65,$37,$25,$70,$BF,$0F,$CE,$FF,$BD,$B1,$7D
	.dc.b	$2F,$1C,$7F,$CD,$93,$98,$C6,$28,$E3,$97,$FB,$C7,$3F,$DB,$BE,$83
	.dc.b	$C7,$1E,$ED,$E4,$67,$0F,$E3,$7F,$4B,$FE,$DF,$FA,$57,$8F,$EA,$BD
	.dc.b	$D7,$E5,$F2,$FB,$58,$71,$8F,$F0,$E6,$1C,$6F,$84,$E6,$7F,$DD,$53
	.dc.b	$97,$E1,$DA,$71,$1C,$86,$1E,$27,$71,$A7,$99,$AB,$EE,$D3,$D6,$53
	.dc.b	$FD,$29,$5D,$DC,$1C,$AB,$B3,$AA,$EC,$AF,$CB,$87,$C0,$8E,$26,$4E
	.dc.b	$CE,$6B,$9C,$75,$C3,$CB,$18,$8E,$0F,$82,$78,$4F,$FA,$C2,$BB,$E1
	.dc.b	$E7,$AE,$11,$8C,$3C,$AD,$C3,$C6,$D9,$DC,$79,$2B,$95,$F5,$8F,$F1
	.dc.b	$31,$E9,$3F,$C5,$15,$87,$07,$F9,$5F,$E9,$3F,$8E,$FB,$3F,$E2,$D6
	.dc.b	$C7,$FB,$19,$AF,$F6,$46,$BC,$7E,$3F,$FC,$BA,$ED,$AE,$DA,$EF,$7E
	.dc.b	$D1,$C3,$2F,$AB,$F4,$AF,$A1,$FF,$35,$46,$1B,$9F,$1C,$3F,$01,$8B
	.dc.b	$0E,$53,$FE,$71,$C3,$95,$7E,$41,$AF,$09,$DA,$15,$CF,$1E,$18,$78
	.dc.b	$27,$57,$CB,$1F,$C5,$DE,$E7,$A5,$F0,$61,$82,$4F,$AD,$CB,$CB,$0F
	.dc.b	$1C,$78,$9B,$7D,$F1,$F6,$63,$E3,$AD,$61,$FA,$CC,$7C,$58,$F5,$9F
	.dc.b	$D2,$61,$C5,$F7,$4E,$B0,$E7,$1F,$BD,$9F,$8B,$7F,$A9,$84,$63,$5F
	.dc.b	$A1,$87,$5A,$D4,$FF,$C0,$1A,$E5,$E1,$FF,$A8,$67,$39,$5C,$3F,$D4
	.dc.b	$9F,$4B,$FE,$7D,$BF,$E1,$9F,$49,$FE,$76,$6F,$B5,$B7,$EA,$4E,$DF
	.dc.b	$ED,$4D,$C3,$CD,$5C,$67,$FB,$73,$63,$D5,$3B,$BF,$0F,$CB,$F5,$B4
	.dc.b	$A3,$0D,$8F,$2F,$85,$89,$E1,$87,$F8,$4E,$7D,$3E,$26,$5F,$E4,$D7
	.dc.b	$C5,$F1,$C7,$03,$4F,$F8,$0F,$E9,$7F,$9E,$3F,$CB,$5A,$E7,$9F,$BF
	.dc.b	$9B,$E7,$9C,$E9,$F0,$3E,$97,$C6,$66,$A5,$FC,$C8,$CB,$D6,$3E,$07
	.dc.b	$CD,$7F,$A2,$23,$F9,$E3,$5F,$F0,$B3,$6B,$E6,$F6,$F0,$F6,$7E,$0F
	.dc.b	$B9,$2F,$8C,$65,$FB,$71,$97,$9C,$7C,$FF,$8A,$FF,$2B,$47,$F4,$B6
	.dc.b	$BF,$E2,$56,$DF,$E4,$79,$1F,$03,$C7,$D2,$FB,$10,$F8,$9A,$67,$0E
	.dc.b	$93,$87,$FA,$C4,$ED,$C9,$FD,$6F,$A3,$EB,$61,$F9,$F6,$F3,$FF,$BE
	.dc.b	$5B,$3F,$C1,$7E,$95,$E4,$7E,$26,$73,$29,$7D,$2F,$E3,$5B,$1D,$FA
	.dc.b	$78,$19,$99,$19,$3F,$B7,$7D,$2B,$CD,$FC,$B3,$63,$31,$D7,$5C,$BF
	.dc.b	$9B,$FA,$4F,$F6,$2C,$FF,$77,$6B,$FE,$51,$AE,$4F,$F9,$4E,$33,$AF
	.dc.b	$EF,$FE,$95,$C9,$FA,$56,$D7,$3F,$82,$ED,$E9,$FE,$A7,$E9,$5F,$0F
	.dc.b	$4C,$D7,$F8,$1E,$1E,$EE,$C9,$19,$9F,$D9,$7D,$2B,$E9,$69,$27,$97
	.dc.b	$D1,$FE,$DF,$E3,$1F,$2B,$49,$3F,$E1,$5F,$A4,$7C,$86,$23,$FC,$E1
	.dc.b	$5C,$1F,$F0,$D3,$5F,$F3,$A4,$F0,$73,$5F,$C1,$61,$F4,$B4,$E9,$C5
	.dc.b	$FC,$97,$E9,$5E,$27,$F0,$B5,$9D,$31,$13,$1F,$F1,$54,$E9,$B3,$97
	.dc.b	$FD,$0D,$1F,$ED,$77,$7F,$13,$D4,$ED,$D7,$63,$24,$38,$BF,$C7,$7E
	.dc.b	$95,$F2,$F4,$93,$FE,$3B,$FA,$5F,$CD,$BF,$E3,$D9,$DB,$67,$6F,$FA
	.dc.b	$66,$76,$FF,$A6,$E7,$6E,$7E,$2F,$F9,$64,$9E,$3A,$7F,$4B,$8F,$EE
	.dc.b	$D3,$E1,$F5,$F6,$70,$E8,$B1,$92,$1E,$17,$AF,$BD,$DB,$E2,$64,$FA
	.dc.b	$E3,$2F,$B6,$32,$FA,$63,$FA,$07,$D2,$BE,$96,$92,$7F,$CB,$DE,$9F
	.dc.b	$E9,$B3,$FE,$C5,$5B,$9F,$F3,$13,$5F,$F5,$E3,$63,$FB,$44,$F1,$F2
	.dc.b	$7F,$C1,$94,$AD,$46,$27,$FA,$B3,$7D,$7F,$15,$7F,$B1,$E7,$FD,$FC
	.dc.b	$F9,$3E,$97,$FA,$C4,$1B,$EA,$C9,$D1,$0E,$4F,$0E,$CE,$3F,$4B,$27
	.dc.b	$DF,$19,$7E,$0C,$65,$F1,$47,$F6,$7F,$A1,$FF,$70,$C7,$B0,$F6,$B8
	.dc.b	$70,$1C,$3B,$6B,$C3,$87,$FC,$AB,$1F,$84,$8E,$2D,$78,$47,$F9,$6B
	.dc.b	$E9,$1F,$49,$F9,$29,$F9,$B1,$8F,$43,$C3,$08,$78,$2F,$87,$CC,$6E
	.dc.b	$7D,$5F,$1C,$F8,$3F,$E3,$7C,$5F,$EA,$78,$C7,$C5,$7F,$A2,$F9,$7F
	.dc.b	$D1,$91,$BF,$23,$15,$FE,$49,$8D,$BC,$6F,$9E,$0B,$E3,$1F,$ED,$BA
	.dc.b	$E5,$5B,$C3,$9D,$A7,$C9,$F4,$58,$7C,$DF,$47,$70,$4D,$8C,$37,$0F
	.dc.b	$13,$6F,$C5,$1F,$94,$FA,$2F,$ED,$4F,$FB,$17,$57,$D1,$FF,$2E,$8F
	.dc.b	$F6,$04,$EA,$B8,$9F,$2E,$74,$E0,$46,$F4,$FF,$59,$5F,$4F,$2B,$DB
	.dc.b	$5A,$F6,$BD,$F3,$9C,$7B,$C6,$31,$EB,$87,$F9,$4F,$1E,$95,$FE,$77
	.dc.b	$BE,$18,$78,$67,$2F,$97,$1E,$F1,$89,$DD,$F3,$C7,$B5,$69,$B8,$78
	.dc.b	$8F,$89,$8F,$89,$F9,$11,$AC,$7C,$0D,$BE,$93,$C2,$FA,$34,$E9,$5F
	.dc.b	$2B,$0F,$1C,$74,$35,$1D,$1B,$5D,$23,$86,$1E,$37,$FD,$31,$8F,$B6
	.dc.b	$31,$1F,$36,$B8,$9D,$5E,$38,$79,$1B,$1E,$4B,$FA,$F5,$CA,$FA,$5E
	.dc.b	$E7,$2D,$AF,$A5,$5B,$C7,$E1,$47,$7A,$DB,$D4,$DD,$71,$C3,$EB,$46
	.dc.b	$5A,$7C,$13,$CA,$59,$E6,$D8,$E6,$71,$C7,$E5,$C7,$C4,$BF,$1B,$67
	.dc.b	$E1,$B4,$EB,$5F,$34,$EA,$58,$EB,$1D,$CF,$2C,$65,$B1,$EF,$37,$19
	.dc.b	$AF,$1C,$FD,$33,$BC,$ED,$E2,$78,$E7,$C3,$19,$9E,$38,$FD,$28,$F7
	.dc.b	$63,$F3,$DF,$5D,$F5,$BE,$2F,$38,$F1,$9C,$E7,$84,$78,$8D,$1E,$23
	.dc.b	$94,$70,$73,$7E,$2C,$3E,$5C,$68,$E1,$87,$95,$F8,$D3,$BA,$F2,$CE
	.dc.b	$70,$F8,$98,$78,$9F,$3D,$F3,$BE,$0F,$94,$C4,$EE,$FC,$91,$8C,$7F
	.dc.b	$15,$5C,$2F,$9E,$1D,$EF,$E0,$56,$B0,$E5,$87,$A3,$1F,$5D,$EE,$3D
	.dc.b	$F3,$AC,$7A,$61,$E0,$AD,$57,$86,$3E,$9E,$1D,$48,$E0,$91,$F1,$2F
	.dc.b	$A1,$27,$71,$96,$E3,$F1,$9B,$5E,$A3,$9C,$7F,$4A,$F1,$3F,$A1,$65
	.dc.b	$FD,$0D,$B3,$FD,$45,$B7,$FC,$E3,$6F,$EB,$D1,$8B,$FE,$A7,$F4,$7F
	.dc.b	$D2,$CF,$E9,$9F,$47,$70,$FE,$B5,$17,$0E,$5C,$0A,$C5,$AF,$FC,$3F
	.dc.b	$F4,$3F,$E7,$25,$7F,$D2,$25,$F4,$7F,$D9,$1E,$11,$AA,$E0,$EE,$75
	.dc.b	$5D,$3A,$FF,$B2,$DF,$2E,$5F,$E9,$78,$70,$8F,$EE,$D7,$C3,$0E,$97
	.dc.b	$A3,$FB,$6F,$C7,$BE,$07,$FC,$AD,$97,$9B,$0F,$1D,$B1,$E9,$3F,$DF
	.dc.b	$74,$7D,$1F,$F7,$13,$FB,$97,$55,$9F,$F0,$04,$5B,$F8,$F3,$70,$FF
	.dc.b	$00,$D7,$F7,$9F,$82,$FE,$F4,$D3,$FC,$10,$5A,$F9,$ED,$9F,$2E,$2F
	.dc.b	$13,$FB,$4F,$CE,$3B,$3E,$1C,$61,$B7,$FD,$F6,$74,$FB,$CF,$AF,$3B
	.dc.b	$71,$87,$D3,$8E,$2F,$F5,$79,$F9,$8F,$C5,$8F,$F0,$A2,$FA,$3A,$63
	.dc.b	$DC,$B7,$FE,$2B,$9D,$36,$74,$D3,$5F,$47,$D3,$E1,$FA,$B6,$32,$73
	.dc.b	$8C,$B4,$F4,$D3,$E8,$9E,$F7,$E0,$3E,$8D,$F1,$3F,$E1,$29,$CD,$FC
	.dc.b	$0E,$65,$6A,$39,$CE,$B0,$FF,$10,$D6,$AF,$96,$1D,$5B,$7C,$23,$FC
	.dc.b	$77,$F4,$6F,$A3,$7D,$1E,$B1,$FD,$A0,$F9,$FF,$97,$E8,$FC,$C6,$E3
	.dc.b	$F2,$5B,$3C,$1B,$CD,$FE,$54,$87,$CB,$7D,$43,$EA,$FC,$A5,$BF,$A5
	.dc.b	$16,$7C,$33,$6B,$F4,$3E,$D7,$E1,$A7,$F0,$1E,$EF,$F8,$E2,$31,$3F
	.dc.b	$E3,$CE,$84,$E7,$A7,$F2,$6D,$CB,$F2,$0D,$C3,$FC,$8D,$87,$29,$FF
	.dc.b	$1D,$5B,$E1,$75,$8F,$EC,$BC,$C7,$57,$F9,$23,$EA,$DD,$DF,$E4,$A9
	.dc.b	$CE,$3F,$80,$9C,$BF,$82,$87,$AD,$B3,$FC,$A3,$F4,$2F,$8D,$C5,$7E
	.dc.b	$22,$30,$D8,$FF,$2B,$FD,$23,$E9,$38,$C7,$2F,$7E,$57,$1F,$92,$BC
	.dc.b	$5F,$E6,$3C,$51,$E0,$3F,$CC,$3F,$49,$CE,$B8,$1F,$98,$71,$5C,$67
	.dc.b	$78,$F1,$AF,$CE,$FA,$2F,$0F,$FD,$0F,$0F,$A2,$76,$7F,$9B,$AB,$53
	.dc.b	$FB,$C3,$51,$9A,$91,$9A,$FC,$93,$DA,$33,$8F,$F9,$F6,$F9,$BF,$E8
	.dc.b	$29,$F0,$76,$6B,$97,$F9,$4C,$3E,$0F,$CB,$3F,$32,$BF,$D0,$D3,$93
	.dc.b	$C9,$19,$AE,$35,$FE,$86,$9D,$57,$7C,$5D,$99,$A9,$5F,$C0,$0D,$DF
	.dc.b	$F4,$C7,$E2,$4E,$AF,$FD,$2D,$F4,$8F,$A4,$F9,$38,$DB,$FE,$13,$36
	.dc.b	$BC,$15,$FE,$98,$8C,$61,$FD,$19,$B5,$FD,$78,$FA,$5F,$29,$BA,$91
	.dc.b	$D1,$EF,$18,$BF,$F5,$47,$BD,$FA,$37,$FB,$56,$2F,$A1,$FF,$53,$CE
	.dc.b	$70,$FB,$D1,$8F,$06,$47,$D9,$EA,$B8,$74,$9B,$A7,$C3,$C7,$FD,$C5
	.dc.b	$7F,$C9,$B1,$F0,$5B,$C8,$FB,$B7,$FE,$B8,$FA,$27,$D2,$7F,$AE,$27
	.dc.b	$38,$7F,$AC,$67,$2F,$F3,$08,$CD,$FF,$80,$1B,$9B,$E2,$7F,$D6,$B3
	.dc.b	$DF,$97,$A1,$2F,$FD,$A5,$3A,$BE,$75,$E5,$F9,$BF,$F0,$0F,$D2,$62
	.dc.b	$51,$8C,$3C,$CC,$C3,$FD,$B9,$93,$B7,$18,$FF,$B5,$F0,$F9,$D5,$E9
	.dc.b	$E1,$FF,$6A,$CF,$F8,$63,$E9,$3F,$C3,$3F,$47,$FC,$B2,$3D,$93,$F0
	.dc.b	$C9,$5F,$D4,$23,$80,$9E,$0F,$CD,$9E,$AA,$BA,$F9,$5F,$E2,$68,$D5
	.dc.b	$E1,$C1,$B3,$A4,$6F,$0F,$4C,$60,$DD,$71,$79,$99,$3B,$39,$8E,$38
	.dc.b	$79,$A7,$D2,$FA,$23,$35,$E2,$9F,$14,$E7,$0F,$F1,$CB,$C6,$33,$3E
	.dc.b	$59,$CB,$C6,$FC,$11,$89,$F3,$9E,$5B,$EE,$E2,$FB,$61,$EB,$8F,$2E
	.dc.b	$1D,$63,$11,$E0,$3E,$AB,$8C,$3B,$B6,$3B,$61,$E1,$7E,$19,$ED,$3E
	.dc.b	$C9,$67,$0E,$0F,$AD,$F4,$9F,$C7,$7D,$9E,$0D,$8F,$8C,$D7,$EC,$B5
	.dc.b	$E3,$F1,$2E,$4E,$2D,$8E,$2D,$78,$B5,$E3,$1F,$77,$E9,$1F,$41,$F8
	.dc.b	$9D,$7F,$1F,$18,$AF,$C1,$60,$C3,$B5,$7E,$52,$31,$87,$3C,$39,$6B
	.dc.b	$F9,$3C,$79,$63,$F8,$9B,$DB,$71,$E7,$5B,$BE,$33,$FE,$80,$BE,$38
	.dc.b	$F7,$BD,$CF,$FA,$13,$0E,$73,$E8,$BE,$32,$BE,$38,$7C,$08,$F0,$63
	.dc.b	$E7,$61,$82,$5F,$37,$2F,$86,$3E,$1E,$3E,$29,$ED,$87,$83,$0E,$F1
	.dc.b	$E5,$9D,$47,$C2,$AF,$F4,$2B,$75,$F8,$15,$FE,$50,$CB,$D7,$96,$23
	.dc.b	$EF,$4F,$CB,$F0,$FF,$7D,$87,$26,$C6,$E3,$FD,$23,$F4,$7F,$E7,$BA
	.dc.b	$FF,$2D,$FD,$27,$F9,$FD,$BE,$DF,$07,$FD,$5B,$3B,$7B,$98,$ED,$D2
	.dc.b	$AE,$4E,$8C,$56,$AB,$95,$F8,$BE,$57,$04,$AF,$1F,$85,$5E,$58,$7F
	.dc.b	$84,$1F,$3F,$8D,$9B,$E5,$18,$21,$E1,$C1,$87,$9B,$E9,$3F,$9D,$CF
	.dc.b	$F9,$E9,$AF,$9D,$B5,$FE,$B6,$C3,$FC,$7B,$3B,$EA,$F7,$C3,$E9,$29
	.dc.b	$D6,$24,$FC,$C8,$CB,$C4,$FF,$44,$FC,$D7,$FA,$19,$BE,$DF,$CA,$8C
	.dc.b	$47,$C0,$EE,$C4,$F8,$3D,$39,$3C,$08,$7C,$67,$2F,$D1,$AE,$79,$7F
	.dc.b	$B4,$3E,$0F,$FC,$A9,$3F,$E9,$A6,$BF,$5D,$B1,$FE,$D6,$E9,$FF,$31
	.dc.b	$CE,$FD,$0F,$BF,$0F,$8A,$A7,$E0,$AF,$8F,$37,$F4,$F6,$DF,$FB,$DA
	.dc.b	$B8,$3F,$EA,$A6,$BE,$1B,$FC,$5F,$3F,$D8,$8F,$F5,$7B,$5F,$C6,$37
	.dc.b	$1F,$F7,$5F,$D2,$3C,$8E,$3C,$9F,$94,$87,$D2,$53,$38,$74,$9C,$3F
	.dc.b	$EB,$C9,$DB,$93,$FD,$7F,$F4,$8F,$33,$F3,$ED,$8C,$C7,$7C,$F2,$FE
	.dc.b	$93,$E9,$3F,$D8,$B3,$FE,$C8,$6B,$FA,$5A,$E4,$FE,$9E,$33,$9B,$7F
	.dc.b	$EF,$1F,$49,$FE,$96,$AF,$F6,$8B,$5E,$1F,$05,$DF,$D2,$FD,$9F,$D2
	.dc.b	$3E,$36,$99,$AF,$F6,$CC,$7F,$A8,$DB,$EF,$70,$36,$FF,$B7,$BE,$91
	.dc.b	$F4,$9F,$BE,$6C,$66,$3B,$FD,$17,$F0,$2F,$A4,$FD,$3C,$FF,$01,$6E
	.dc.b	$DE,$B8,$08,$CE,$DF,$E1,$55,$F3,$DC,$FE,$06,$57,$FB,$C2,$76,$F2
	.dc.b	$76,$E6,$36,$F8,$5F,$24,$FF,$16,$F8,$F7,$CB,$52,$57,$5F,$92,$AF
	.dc.b	$A7,$E9,$3D,$86,$33,$F5,$43,$E9,$1E,$17,$F2,$25,$B5,$31,$12,$7E
	.dc.b	$0D,$72,$AD,$C6,$36,$FF,$1B,$46,$23,$F9,$53,$D9,$D1,$F2,$9F,$F7
	.dc.b	$2B,$83,$C9,$F1,$7E,$33,$C9,$DB,$63,$53,$C1,$EB,$3A,$72,$ED,$C9
	.dc.b	$C2,$7F,$9F,$FD,$23,$E6,$69,$27,$C5,$F4,$9F,$B3,$7C,$93,$B7,$3F
	.dc.b	$1D,$FD,$22,$4F,$18,$FE,$91,$1F,$E0,$B7,$9F,$AF,$B3,$87,$5D,$8C
	.dc.b	$90,$E2,$F5,$F8,$0E,$DE,$F4,$9F,$2C,$65,$F3,$C6,$5F,$0C,$65,$F1
	.dc.b	$C6,$77,$FC,$0E,$1F,$E8,$F9,$FA,$1F,$46,$FA,$4F,$EC,$6D,$8C,$C7
	.dc.b	$F6,$6F,$57,$F7,$F7,$F8,$AA,$D5,$7F,$67,$9F,$ED,$33,$F8,$E6,$F1
	.dc.b	$7F,$9E,$6B,$73,$9E,$AE,$9D,$53,$0F,$C1,$61,$FA,$67,$FD,$EF,$39
	.dc.b	$7B,$4F,$8E,$3C,$9D,$1E,$5F,$A4,$95,$CA,$32,$F4,$8C,$BE,$78,$CB
	.dc.b	$E9,$8C,$E1,$E9,$9F,$F0,$2F,$D1,$3E,$8F,$F0,$97,$FE,$90,$8C,$79
	.dc.b	$9F,$E6,$55,$DC,$F8,$6C,$5F,$0B,$FF,$4D,$7D,$23,$E9,$7F,$67,$18
	.dc.b	$6B,$FC,$3A,$35,$C3,$FE,$71,$6C,$7F,$A8,$3E,$39,$F0,$9F,$C9,$EB
	.dc.b	$F5,$B1,$8F,$80,$FE,$31,$BF,$F1,$98,$DF,$95,$8A,$FE,$37,$1B,$78
	.dc.b	$DF,$3C,$3F,$C4,$F3,$96,$C6,$6F,$A3,$FE,$F5,$9F,$1F,$D2,$75,$A4
	.dc.b	$79,$BE,$97,$0E,$B0,$EF,$18,$B5,$D8,$C3,$67,$CA,$DB,$F1,$B7,$0F
	.dc.b	$1B,$67,$DB,$7D,$1E,$E7,$22,$D7,$26,$D7,$26,$9C,$AB,$A4,$FB,$2B
	.dc.b	$A3,$F1,$2F,$CF,$5C,$23,$55,$C3,$0F,$81,$1B,$9F,$13,$4F,$11,$70
	.dc.b	$F1,$36,$3C,$47,$59,$E7,$3F,$4A,$FC,$35,$B3,$C3,$5B,$C3,$CF,$5B
	.dc.b	$8F,$0B,$67,$C2,$D3,$C2,$58,$F0,$C7,$5B,$F0,$C7,$CE,$BF,$0D,$6E
	.dc.b	$BC,$26,$EF,$C3,$5B,$3C,$25,$8F,$0B,$67,$C2,$D3,$C2,$F5,$BF,$0C
	.dc.b	$FD,$3B,$F0,$D6,$E3,$C3,$5B,$BF,$0C,$6E,$7C,$2D,$3C,$25,$9F,$0B
	.dc.b	$63,$C2,$75,$BF,$09,$F9,$D7,$9B,$63,$9B,$67,$9B,$67,$9C,$F5,$3C
	.dc.b	$EF,$5A,$F3,$CE,$E7,$26,$DF,$0C,$2F,$E0,$CE,$63,$89,$9A,$F2,$B6
	.dc.b	$3C,$AD,$BF,$2B,$67,$CB,$7E,$13,$BD,$F6,$71,$3D,$A3,$0F,$1B,$F3
	.dc.b	$CE,$B0,$FD,$14,$EE,$FC,$77,$C0,$8E,$09,$7E,$68,$CB,$63,$38,$7A
	.dc.b	$27,$2D,$8C,$9E,$AA,$F2,$61,$EA,$8C,$DF,$9A,$72,$79,$5E,$C4,$AE
	.dc.b	$8D,$9E,$97,$E0,$69,$D2,$FB,$4E,$CE,$D3,$B6,$CE,$EF,$C1,$7D,$8F
	.dc.b	$F1,$BF,$D2,$FF,$A5,$E7,$FC,$A3,$50,$46,$5C,$18,$F6,$7F,$A5,$B6
	.dc.b	$72,$D3,$FD,$BD,$3C,$6B,$FC,$BD,$F1,$4F,$69,$EC,$FF,$3B,$BE,$15
	.dc.b	$FD,$03,$C5,$FE,$71,$97,$F3,$96,$E5,$FD,$29,$B5,$FD,$21,$B5,$FD
	.dc.b	$23,$E9,$3F,$A0,$1F,$D0,$7E,$92,$DF,$F5,$28,$B7,$FD,$42,$6D,$71
	.dc.b	$AD,$1F,$D6,$23,$2F,$C6,$FA,$27,$5F,$F5,$3A,$7D,$27,$F5,$9A,$D4
	.dc.b	$E6,$31,$87,$F9,$FA,$F8,$D7,$F9,$55,$B0,$BF,$8F,$AF,$F9,$0A,$B5
	.dc.b	$87,$F5,$BF,$A4,$91,$FD,$B3,$AE,$E1,$FD,$AA,$6E,$1F,$DA,$A6,$D7
	.dc.b	$F5,$83,$FB,$27,$CE,$7F,$65,$69,$FD,$D8,$B7,$FD,$D1,$B5,$FD,$D1
	.dc.b	$A7,$33,$FB,$EC,$FD,$BC,$9F,$15,$FD,$DF,$B5,$F3,$AE,$BD,$4E,$63
	.dc.b	$18,$7C,$38,$CB,$FE,$7C,$AE,$3D,$78,$CF,$FC,$12,$3E,$93,$FC,$02
	.dc.b	$DA,$FF,$0B,$AF,$84,$CC,$C9,$CB,$C2,$31,$3D,$3E,$93,$9F,$C3,$F4
	.dc.b	$6D,$F4,$6C,$EE,$3B,$CE,$DA,$6C,$B1,$F7,$5F,$D7,$FD,$1B,$DF,$FF
	.dc.b	$0E,$76,$3E,$91,$F1,$3F,$A5,$AD,$5F,$FA,$7A,$B5,$39,$B9,$3F,$EA
	.dc.b	$2F,$A4,$7D,$23,$E9,$33,$9F,$C8,$F9,$FF,$8F,$E8,$7C,$86,$E1,$F2
	.dc.b	$1B,$3B,$8F,$97,$3F,$3B,$E5,$24,$7C,$EF,$94,$B5,$F5,$26,$F5,$FF
	.dc.b	$91,$61,$38,$AF,$F2,$2C,$2B,$0F,$F8,$F0,$DC,$7D,$A8,$FE,$0D,$CD
	.dc.b	$F6,$E3,$FD,$79,$F4,$8E,$CC,$F3,$7F,$0B,$8C,$4F,$FB,$06,$9E,$97
	.dc.b	$23,$F8,$87,$D1,$BE,$3B,$15,$FE,$C7,$8C,$36,$3F,$8B,$7D,$23,$E8
	.dc.b	$F8,$C7,$2F,$6E,$57,$1F,$32,$EF,$F7,$CF,$F7,$EE,$3F,$C8,$27,$47
	.dc.b	$FB,$57,$E9,$3A,$38,$9E,$02,$32,$E2,$79,$3A,$8C,$F1,$7F,$BC,$A7
	.dc.b	$2F,$FB,$D3,$D1,$77,$7F,$9A,$70,$7D,$13,$9B,$F9,$95,$6A,$7F,$DD
	.dc.b	$66,$A3,$35,$23,$34,$E7,$C6,$7E,$8C,$39,$36,$FC,$15,$BB,$FF,$7A
	.dc.b	$56,$B1,$47,$F5,$21,$EF,$F9,$F1,$FA,$78,$FC,$49,$C9,$E6,$8C,$D7
	.dc.b	$29,$1D,$2B,$FA,$75,$70,$8F,$A8,$FC,$2E,$8C,$CC,$8E,$35,$CE,$B9
	.dc.b	$47,$C0,$AD,$47,$D6,$FA,$47,$D2,$7B,$8F,$EB,$8E,$2B,$E3,$C6,$08
	.dc.b	$FC,$AF,$94,$B5,$CE,$6E,$1E,$BA,$FA,$F5,$87,$FB,$BF,$B9,$FC,$A2
	.dc.b	$FD,$7A,$BE,$87,$FB,$64,$E7,$0E,$53,$9B,$F6,$9C,$BC,$BD,$9F,$5B
	.dc.b	$AA,$C6,$B1,$F9,$78,$73,$8E,$07,$E0,$A7,$73,$F9,$8F,$13,$EE,$4F
	.dc.b	$D2,$97,$D1,$3E,$93,$FC,$01,$5A,$BF,$F6,$E4,$E6,$FC,$10,$9D,$37
	.dc.b	$93,$FA,$67,$C2,$FF,$83,$6B,$FC,$33,$5A,$24,$EB,$D8,$EE,$F8,$3E
	.dc.b	$73,$FB,$C7,$D2,$6D,$B1,$9B,$F0,$DA,$7F,$AD,$64,$ED,$D6,$7E,$FE
	.dc.b	$2F,$59,$FE,$1B,$7F,$23,$F4,$9F,$92,$FA,$4F,$E3,$D1,$F9,$4C,$7E
	.dc.b	$93,$C7,$2E,$F7,$C6,$17,$C7,$C7,$F6,$CE,$63,$13,$CE,$3F,$41,$87
	.dc.b	$68,$FF,$1D,$CE,$9F,$95,$3A,$AF,$F1,$DC,$E7,$0F,$1D,$F1,$BF,$48
	.dc.b	$7A,$46,$AB,$8C,$E5,$F2,$E1,$DE,$38,$D7,$0C,$3C,$11,$FE,$4D,$C3
	.dc.b	$C1,$18,$8F,$39,$CC,$F6,$57,$37,$E1,$99,$3D,$6E,$67,$9B,$63,$58
	.dc.b	$79,$DF,$01,$8A,$F0,$3E,$6C,$3D,$2F,$53,$07,$88,$C1,$E2,$3F,$CB
	.dc.b	$F7,$FE,$60,$7E,$39,$83,$FC,$CF,$F4,$7F,$C6,$BD,$FF,$F3,$2B,$63
	.dc.b	$3D,$89,$CB,$63,$8F,$C5,$FF,$90,$DB,$D3,$CA,$31,$1F,$E7,$7B,$46
	.dc.b	$23,$FC,$F7,$18,$9F,$C1,$7D,$23,$E8,$3F,$CF,$FB,$FE,$1E,$31,$7F
	.dc.b	$87,$BE,$0F,$CF,$6C,$EE,$B8,$E9,$FE,$88,$BE,$37,$BC,$7D,$17,$F9
	.dc.b	$3C,$38,$DE,$F0,$E3,$7F,$CD,$B0,$E3,$7B,$C3,$8D,$FA,$9B,$5B,$BE
	.dc.b	$2A,$F8,$D6,$F1,$F5,$5F,$0A,$D5,$F0,$BD,$C7,$F4,$6C,$7C,$AD,$AE
	.dc.b	$13,$A8,$E2,$74,$B8,$FE,$4E,$34,$6E,$B9,$C7,$03,$FB,$DC,$7C,$69
	.dc.b	$FE,$98,$DD,$FD,$13,$CF,$0F,$F5,$9D,$6B,$C5,$B5,$F3,$DB,$5F,$0C
	.dc.b	$FE,$4F,$F4,$9F,$E6,$D7,$F9,$4F,$D1,$FF,$9C,$9B,$EC,$F0,$53,$B7
	.dc.b	$FC,$1E,$DC,$3F,$B0,$57,$19,$FF,$09,$B6,$3F,$B2,$4E,$F0,$FF,$64
	.dc.b	$AF,$92,$EC,$DB,$F2,$C7,$86,$FF,$C8,$71,$83,$CF,$E3,$E4,$E3,$32
	.dc.b	$32,$43,$CE,$FA,$FE,$93,$F9,$B4,$7F,$48,$6B,$FE,$2A,$6E,$3E,$CF
	.dc.b	$3B,$87,$57,$8F,$E9,$25,$F1,$8C,$BF,$2A,$32,$F3,$8F,$95,$F2,$FF
	.dc.b	$E7,$68,$FE,$AE,$D7,$FC,$72,$DB,$F9,$7D,$BC,$3D,$8F,$85,$E9,$CB
	.dc.b	$E3,$19,$7E,$CC,$65,$E7,$1F,$67,$E0,$FF,$CB,$11,$FD,$99,$AF,$F9
	.dc.b	$21,$B7,$FE,$19,$E4,$FF,$33,$4E,$FD,$1F,$C1,$43,$C3,$A7,$58,$93
	.dc.b	$FD,$CA,$74,$D9,$CB,$FE,$56,$E6,$FB,$CD,$E1,$FC,$74,$63,$0F,$F1
	.dc.b	$57,$D1,$BC,$8C,$D7,$E4,$E1,$F4,$94,$CE,$1F,$E0,$2C,$39,$74,$7F
	.dc.b	$8F,$21,$5C,$1F,$F0,$3B,$5F,$11,$FE,$0A,$FA,$47,$9B,$FA,$16,$C6
	.dc.b	$63,$C5,$B6,$5F,$D3,$7D,$27,$FA,$D6,$7F,$C2,$6D,$7F,$CE,$B5,$C9
	.dc.b	$FF,$3B,$C6,$75,$FF,$24,$7D,$27,$FB,$46,$7F,$C3,$6D,$77,$EF,$76
	.dc.b	$F1,$3F,$69,$F4,$8F,$84,$A6,$6B,$FC,$47,$0F,$A4,$D9,$23,$33,$FB
	.dc.b	$DF,$A4,$7D,$25,$24,$F2,$FA,$1F,$E0,$5F,$49,$FE,$A4,$8F,$F1,$7B
	.dc.b	$5F,$F4,$8B,$75,$FF,$31,$2F,$9B,$E3,$E1,$4A,$E2,$F4,$8C,$47,$28
	.dc.b	$B5,$CA,$7F,$89,$7C,$6B,$E9,$29,$2F,$C7,$F2,$15,$F1,$FD,$27,$A0
	.dc.b	$C6,$1F,$E7,$5F,$A4,$78,$BF,$C7,$56,$B4,$C4,$4B,$FF,$26,$CE,$9B
	.dc.b	$39,$7F,$D6,$30,$FA,$4E,$84,$8E,$3D,$F5,$DC,$5C,$3C,$93,$7B,$7F
	.dc.b	$31,$7C,$EF,$C3,$7E,$7A,$FE,$69,$F4,$8F,$98,$A4,$AF,$F3,$1F,$D1
	.dc.b	$FE,$A5,$FF,$32,$CE,$DC,$FD,$27,$F3,$D9,$3C,$63,$FD,$97,$1F,$E0
	.dc.b	$67,$9F,$A9,$B3,$87,$5D,$8C,$90,$E2,$F5,$F7,$BB,$7B,$92,$7C,$B1
	.dc.b	$97,$CF,$19,$7C,$31,$97,$C7,$19,$DF,$FD,$5D,$82,$BF,$AB,$7D,$1B
	.dc.b	$E9,$29,$30,$FF,$42,$FA,$7F,$B1,$C3,$FD,$6D,$5B,$9F,$E0,$C7,$F0
	.dc.b	$77,$0E,$03,$87,$6D,$7F,$87,$4F,$E8,$78,$AF,$4F,$F6,$98,$FD,$14
	.dc.b	$F8,$BA,$B0,$66,$64,$64,$F3,$F4,$71,$FA,$49,$18,$21,$FF,$00,$2D
	.dc.b	$1A,$8C,$4F,$84,$C6,$5F,$ED,$5A,$DC,$F2,$8F,$F6,$C7,$D0,$FD,$F8
	.dc.b	$F8,$0D,$7E,$2B,$5F,$E5,$8D,$7F,$DD,$3F,$48,$FA,$4F,$C8,$47,$CC
	.dc.b	$8C,$79,$DF,$2E,$10,$EF,$7C,$36,$63,$73,$F5,$3E,$37,$FC,$C1,$E2
	.dc.b	$FF,$BD,$F3,$FF,$7C,$47,$F9,$8F,$E2,$31,$C5,$FE,$FA,$8D,$BF,$ED
	.dc.b	$37,$4D,$F8,$2F,$E5,$0D,$9E,$38,$2B,$8A,$BE,$F7,$FE,$DB,$8D,$2F
	.dc.b	$A3,$E3,$5F,$E7,$5F,$A3,$B3,$9C,$7C,$11,$8A,$CB,$63,$27,$96,$32
	.dc.b	$D3,$37,$E3,$7D,$D5,$D6,$30,$79,$2B,$A3,$8B,$F2,$57,$38,$C5,$2B
	.dc.b	$78,$7B,$9B,$5D,$9B,$5D,$63,$94,$ED,$A6,$CB,$1B,$3B,$E1,$EC,$6C
	.dc.b	$7B,$27,$E6,$CE,$5B,$19,$3D,$F1,$96,$99,$9F,$A7,$19,$6C,$67,$0F
	.dc.b	$F5,$44,$E5,$B1,$9C,$3E,$55,$FA,$EF,$CA,$7C,$6C,$3E,$0D,$FC,$A8
	.dc.b	$CC,$EC,$F3,$BB,$BF,$83,$3B,$9C,$9B,$8F,$4C,$EC,$E6,$FC,$96,$9F
	.dc.b	$26,$38,$C6,$5B,$19,$8E,$71,$D6,$3E,$09,$D5,$DC,$F8,$E7,$67,$32
	.dc.b	$CF,$33,$8C,$F3,$9F,$8F,$1C,$E7,$71,$CE,$37,$1C,$DA,$73,$8E,$33
	.dc.b	$CE,$FE,$7D,$6E,$72,$D8,$CD,$71,$6C,$F7,$37,$19,$C3,$EE,$4E,$8F
	.dc.b	$CE,$57,$D2,$6C,$FC,$CA,$E5,$1B,$C3,$EE,$BA,$C3,$E3,$B6,$FD,$ED
	.dc.b	$BF,$6D,$F2,$C3,$C2,$4A,$FC,$63,$67,$E4,$1C,$2B,$95,$7A,$8E,$55
	.dc.b	$C0,$93,$C2,$35,$3C,$0F,$44,$FB,$70,$F4,$91,$C1,$2B,$E0,$4F,$1B
	.dc.b	$F2,$61,$F4,$DB,$5F,$45,$B5,$F3,$63,$C7,$3B,$69,$B2,$C6,$CF,$45
	.dc.b	$79,$9F,$0D,$F9,$5B,$3E,$5A,$F9,$F3,$95,$1F,$E7,$9F,$A3,$FF,$48
	.dc.b	$2A,$FE,$3F,$AB,$E9,$7F,$C8,$93,$9B,$E3,$5B,$BE,$2F,$0A,$DB,$FC
	.dc.b	$A0,$46,$A7,$8D,$F5,$F2,$3F,$A2,$D6,$BD,$7F,$E8,$F3,$9C,$7F,$A6
	.dc.b	$78,$7F,$D0,$27,$FA,$0B,$63,$27,$F5,$18,$CB,$4F,$EA,$9F,$49,$FD
	.dc.b	$20,$FE,$93,$F4,$96,$32,$4C,$3F,$9C,$C6,$5F,$E7,$73,$72,$FE,$AF
	.dc.b	$36,$BF,$D4,$5F,$44,$EA,$FE,$BF,$4F,$A4,$FE,$C7,$8F,$C3,$AD,$5F
	.dc.b	$CB,$3E,$64,$EA,$BE,$6D,$2B,$6F,$F8,$9D,$AF,$F8,$B1,$AF,$8A,$F8
	.dc.b	$BF,$D6,$BE,$93,$FB,$39,$FD,$A3,$A2,$E3,$FD,$DE,$2C,$FF,$58,$9B
	.dc.b	$87,$F7,$4A,$FE,$DF,$F3,$9F,$DC,$1A,$7F,$7F,$2D,$7F,$60,$6C,$F1
	.dc.b	$7F,$C0,$93,$C5,$A7,$C1,$9F,$C8,$62,$F8,$8F,$C0,$FD,$27,$F8,$0E
	.dc.b	$7E,$14,$63,$A7,$F0,$F5,$BB,$E5,$5A,$7A,$71,$30,$F9,$91,$F5,$B0
	.dc.b	$FF,$05,$D3,$E9,$29,$8F,$84,$B5,$FE,$1D,$9D,$36,$74,$D3,$5F,$49
	.dc.b	$D3,$E1,$BA,$B6,$32,$73,$8C,$B4,$CD,$7F,$82,$C4,$69,$C3,$FA,$6F
	.dc.b	$A3,$7B,$9F,$E2,$CF,$93,$7D,$23,$E2,$75,$39,$8C,$63,$F1,$AF,$8B
	.dc.b	$67,$FC,$6B,$F4,$8F,$A4,$7D,$27,$2A,$F8,$5F,$3F,$F2,$3C,$CF,$92
	.dc.b	$D8,$C9,$F1,$E3,$2D,$3E,$57,$CA,$FD,$03,$E8,$FC,$FD,$8C,$93,$83
	.dc.b	$E1,$46,$5F,$89,$37,$AB,$FC,$AB,$09,$C5,$7F,$19,$8F,$F2,$FF,$27
	.dc.b	$F0,$CF,$A4,$75,$FF,$0E,$E9,$4E,$63,$1E,$94,$C3,$F1,$7F,$42,$F8
	.dc.b	$CC,$65,$FC,$6E,$31,$3A,$9F,$E3,$3F,$48,$FA,$4F,$E3,$75,$AB,$E1
	.dc.b	$ED,$E0,$CF,$C8,$5C,$7B,$B8,$7F,$37,$F4,$9C,$DC,$98,$C3,$57,$FC
	.dc.b	$B6,$74,$F9,$6B,$F9,$5F,$A5,$F8,$AB,$FD,$70,$FA,$27,$2E,$BB,$FF
	.dc.b	$CE,$B3,$98,$C3,$FA,$CA,$E5,$87,$F9,$8A,$39,$73,$EA,$3F,$9F,$4E
	.dc.b	$B9,$BF,$8C,$B6,$BF,$A6,$1C,$0E,$98,$7F,$40,$5E,$FF,$8B,$0F,$91
	.dc.b	$7F,$C8,$A7,$27,$AA,$32,$FD,$F3,$38,$7F,$4F,$6C,$FF,$56,$7E,$07
	.dc.b	$3F,$7B,$91,$FC,$20,$D9,$E6,$CF,$F1,$8E,$63,$FA,$AF,$D2,$3E,$93
	.dc.b	$E3,$D5,$AF,$E2,$33,$71,$F8,$D3,$71,$FE,$77,$36,$BF,$2C,$DA,$FE
	.dc.b	$E7,$4F,$93,$EF,$84,$BE,$AF,$82,$31,$5F,$DB,$7D,$AF,$F5,$45,$7D
	.dc.b	$DF,$A1,$FE,$D7,$18,$C1,$18,$F2,$E5,$7E,$12,$7F,$D8,$AF,$F2,$AA
	.dc.b	$FE,$60,$49,$F7,$F6,$70,$C3,$FB,$E7,$8D,$F7,$5F,$F6,$6F,$D1,$3E
	.dc.b	$93,$FB,$DC,$E6,$93,$9C,$7F,$9B,$CF,$1B,$3B,$F8,$4F,$F0,$34,$FE
	.dc.b	$EE,$34,$4E,$4F,$F0,$2C,$EA,$BF,$C3,$13,$AF,$9B,$FE,$E3,$F4,$97
	.dc.b	$0F,$E9,$DC,$68,$C5,$FA,$AB,$FC,$1F,$93,$B3,$D3,$1E,$AE,$19,$8F
	.dc.b	$F4,$EB,$FF,$13,$7D,$27,$F8,$9F,$E9,$1E,$37,$9E,$F7,$5F,$19,$FF
	.dc.b	$0E,$BC,$04,$F0,$7D,$31,$A5,$5D,$6B,$A5,$D9,$FE,$3C,$68,$E7,$1A
	.dc.b	$69,$A9,$E0,$FC,$EA,$FF,$1C,$DF,$86,$3F,$94,$4E,$B5,$FC,$3E,$1C
	.dc.b	$27,$D1,$5B,$9C,$C7,$63,$89,$5C,$C7,$58,$E3,$87,$96,$31,$1E,$97
	.dc.b	$A6,$1E,$49,$F6,$CE,$5E,$06,$30,$F3,$4E,$E7,$DC,$78,$5C,$3E,$4A
	.dc.b	$EB,$87,$AE,$3A,$1F,$41,$C5,$74,$6C,$78,$8F,$B0,$58,$C3,$83,$FA
	.dc.b	$87,$D2,$7F,$1A,$F8,$0E,$0D,$8F,$5B,$5F,$B2,$D7,$8F,$C5,$B8,$38
	.dc.b	$B6,$38,$B5,$E2,$D7,$8C,$7D,$EF,$A4,$7C,$FF,$F6,$3C,$FE,$F4,$62
	.dc.b	$BF,$07,$83,$0E,$D5,$F8,$38,$C6,$1C,$F0,$E5,$AF,$F6,$DC,$79,$63
	.dc.b	$F8,$1B,$DB,$71,$E7,$5B,$BE,$33,$FE,$8D,$BE,$38,$F7,$BD,$CF,$FA
	.dc.b	$47,$0E,$73,$E8,$BE,$32,$BE,$38,$7C,$08,$F0,$63,$E7,$61,$82,$5F
	.dc.b	$37,$2F,$86,$3E,$1E,$3E,$29,$ED,$87,$83,$0E,$F1,$E5,$9D,$47,$C2
	.dc.b	$AF,$F3,$BB,$74,$F8,$15,$FE,$54,$CB,$D7,$96,$23,$FD,$49,$4F,$0B
	.dc.b	$FC,$31,$87,$26,$E1,$B8,$FF,$41,$FD,$23,$A7,$38,$7F,$AD,$FE,$93
	.dc.b	$FC,$D0,$FF,$98,$9A,$FF,$AC,$67,$83,$FE,$66,$6B,$FE,$B5,$8D,$D7
	.dc.b	$FA,$41,$B7,$FE,$2B,$AD,$CF,$8B,$E4,$BC,$36,$94,$61,$B1,$E3,$F0
	.dc.b	$B1,$3C,$30,$FF,$08,$CF,$9F,$C5,$CB,$E5,$9D,$44,$8C,$1F,$E7,$EF
	.dc.b	$A4,$73,$F8,$9B,$19,$8E,$DE,$66,$5F,$07,$D2,$49,$D1,$0F,$84,$E5
	.dc.b	$FF,$6C,$57,$47,$FD,$B5,$F3,$1F,$E6,$D7,$FD,$1E,$D7,$3E,$C7,$B7
	.dc.b	$C0,$F8,$B3,$9D,$FE,$67,$A3,$F0,$4C,$D4,$BF,$B1,$19,$7A,$C7,$CE
	.dc.b	$FA,$4F,$F2,$94,$7F,$A7,$9A,$FF,$BB,$5B,$5F,$EC,$8E,$9F,$F2,$FC
	.dc.b	$EF,$D1,$FA,$F0,$F8,$5A,$75,$89,$3F,$EA,$A9,$D3,$67,$2F,$FB,$EF
	.dc.b	$9F,$EE,$B7,$8B,$F0,$71,$8C,$3F,$DB,$3F,$46,$F2,$33,$5F,$88,$87
	.dc.b	$D2,$53,$38,$7F,$AE,$B0,$E5,$D3,$FE,$7B,$AD,$CF,$FA,$F9,$AF,$84
	.dc.b	$FF,$60,$FD,$27,$F8,$A5,$BE,$67,$E5,$A3,$11,$FC,$B6,$BA,$6D,$9A
	.dc.b	$FC,$DF,$D2,$7F,$AB,$2B,$FD,$96,$D7,$80,$17,$D3,$5F,$F7,$BC,$ED
	.dc.b	$FF,$7C,$FD,$27,$FA,$3A,$BF,$DA,$AD,$79,$89,$E1,$8F,$FA,$65,$FF
	.dc.b	$31,$C7,$FB,$73,$E9,$1F,$25,$FB,$06,$C7,$9B,$D3,$C4,$E7,$E9,$24
	.dc.b	$F1,$8E,$1F,$41,$FB,$8F,$A4,$7D,$25,$33,$5F,$EE,$AF,$A4,$F6,$69
	.dc.b	$FE,$AB,$7F,$CC,$11,$9F,$5F,$F8,$0C,$6A,$3E,$23,$53,$C1,$F7,$B7
	.dc.b	$75,$75,$E1,$E6,$74,$B9,$7E,$38,$EF,$E1,$48,$F3,$3E,$2F,$9C,$C4
	.dc.b	$76,$8B,$87,$68,$F2,$CF,$F1,$8F,$8D,$7C,$A5,$24,$FC,$1F,$90,$AF
	.dc.b	$A3,$E9,$3D,$E6,$30,$E9,$0F,$A4,$78,$9F,$C9,$17,$15,$31,$12,$7F
	.dc.b	$15,$5C,$AB,$71,$8D,$FF,$DB,$51,$88,$FC,$3B,$D9,$D1,$F9,$37,$0E
	.dc.b	$AF,$C0,$E8,$F3,$5F,$97,$E9,$24,$66,$91,$E9,$EF,$CC,$F8,$22,$E3
	.dc.b	$E7,$9F,$E8,$3F,$48,$F9,$8A,$4B,$F0,$7D,$1F,$F0,$17,$C5,$3B,$73
	.dc.b	$F4,$9F,$D2,$A4,$F1,$8F,$D2,$C7,$F8,$55,$E7,$EB,$6C,$E1,$D9,$63
	.dc.b	$24,$38,$B9,$79,$C5,$CB,$EC,$D7,$EB,$A7,$BF,$BF,$E3,$F7,$64,$7A
	.dc.b	$A3,$CC,$61,$F2,$C6,$1F,$9D,$F4,$6F,$A4,$FE,$CA,$D8,$CC,$7F,$68
	.dc.b	$F4,$FF,$4D,$39,$9F,$E0,$71,$FD,$C1,$FE,$1C,$D5,$1B,$C7,$F2,$32
	.dc.b	$EA,$E9,$D5,$2F,$F0,$13,$96,$CE,$5E,$F3,$E3,$89,$1C,$23,$55,$DA
	.dc.b	$7D,$1D,$1E,$9F,$A4,$97,$DA,$32,$F8,$23,$2F,$B2,$3F,$C0,$FF,$44
	.dc.b	$FA,$3F,$C2,$4F,$FA,$6A,$31,$E6,$FF,$9B,$D7,$73,$E1,$71,$7C,$2F
	.dc.b	$F9,$91,$F3,$F5,$FD,$9B,$8D,$D1,$88,$FE,$71,$18,$ED,$FF,$3E,$36
	.dc.b	$3F,$D5,$7F,$14,$F8,$1F,$E5,$F5,$FA,$E8,$C7,$BB,$FC,$2F,$87,$F8
	.dc.b	$64,$6F,$C8,$C5,$7F,$45,$8D,$BC,$6F,$9E,$1F,$D2,$A7,$2D,$8C,$DF
	.dc.b	$47,$FC,$7D,$5C,$6F,$AD,$2F,$8D,$EE,$73,$18,$AF,$15,$F4,$6E,$1A
	.dc.b	$8C,$61,$CD,$EE,$F8,$EF,$53,$B6,$D7,$8C,$F6,$46,$5A,$66,$7D,$91
	.dc.b	$96,$C6,$70,$F7,$4E,$5B,$19,$3C,$E5,$9F,$34,$EE,$33,$1E,$77,$C2
	.dc.b	$D3,$B9,$67,$AB,$63,$8C,$7C,$97,$8B,$63,$8B,$4E,$25,$8E,$27,$CE
	.dc.b	$AF,$8B,$3D,$4F,$3C,$7A,$70,$FA,$91,$F3,$E7,$75,$CE,$36,$71,$9D
	.dc.b	$9F,$8D,$AF,$29,$DC,$B1,$CD,$B1,$B8,$F4,$C6,$DA,$6C,$B1,$B9,$F6
	.dc.b	$CE,$DB,$3B,$6C,$EE,$7E,$0C,$E5,$B1,$98,$F5,$CF,$B6,$3D,$78,$7E
	.dc.b	$41,$B7,$F0,$AF,$C0,$72,$9D,$1F,$94,$AE,$4D,$8D,$DF,$66,$DF,$47
	.dc.b	$E1,$CF,$2C,$3F,$11,$5B,$3D,$05,$AF,$41,$C5,$F8,$B3,$C6,$BE,$D3
	.dc.b	$E1,$8E,$33,$BB,$F8,$B3,$B9,$FB,$4E,$E7,$AB,$4D,$4F,$B2,$FC,$B3
	.dc.b	$CE,$FE,$16,$1F,$17,$0E,$F8,$71,$8C,$57,$4B,$E3,$39,$BF,$83,$5A
	.dc.b	$8E,$F1,$F1,$9D,$5F,$BA,$74,$49,$E2,$48,$D3,$70,$F7,$CE,$E7,$24
	.dc.b	$8C,$CE,$CF,$59,$63,$E8,$57,$91,$F6,$DF,$A8,$93,$F4,$1F,$E3,$1F
	.dc.b	$47,$FC,$67,$E8,$EC,$FF,$2D,$6D,$7F,$2B,$6E,$1F,$CA,$DF,$E5,$30
	.dc.b	$FA,$2F,$F5,$15,$7F,$8A,$B7,$7D,$1F,$F9,$12,$3F,$C5,$13,$AA,$FF
	.dc.b	$40,$D6,$BC,$7E,$73,$9C,$7F,$C8,$6D,$BF,$F4,$37,$C4,$3D,$9F,$F7
	.dc.b	$F4,$74,$F6,$3F,$99,$78,$9F,$CC,$DB,$1F,$D1,$DB,$9F,$F4,$56,$E3
	.dc.b	$FD,$15,$B3,$FC,$F9,$B3,$FC,$ED,$FE,$7D,$F4,$72,$7F,$A9,$7D,$1D
	.dc.b	$AF,$EA,$13,$6F,$FA,$84,$DA,$FE,$DB,$F4,$2E,$CF,$EA,$34,$FA,$3F
	.dc.b	$E9,$38,$AB,$57,$FE,$54,$AD,$73,$7D,$09,$CD,$7F,$A8,$1B,$3F,$E7
	.dc.b	$5D,$BE,$37,$8F,$BC,$BF,$AB,$7D,$1C,$CB,$FB,$57,$6D,$CB,$FB,$44
	.dc.b	$DC,$7F,$B4,$4D,$AF,$EC,$33,$6B,$FA,$F4,$DE,$AF,$EF,$D5,$FD,$8F
	.dc.b	$E6,$3F,$B2,$34,$FE,$F0,$5A,$FE,$EC,$DC,$3F,$D0,$3A,$3E,$8B,$FB
	.dc.b	$AF,$3B,$E8,$FF,$83,$61,$F1,$67,$31,$8A,$E1,$87,$F9,$D5,$B5,$F8
	.dc.b	$E9,$F9,$B1,$8C,$3F,$C1,$F2,$FA,$3F,$EF,$ED,$BF,$F0,$B2,$F8,$4C
	.dc.b	$CC,$9C,$BC,$23,$13,$FE,$13,$39,$16,$35,$3C,$0E,$BF,$47,$E0,$F8
	.dc.b	$6F,$0B,$63,$27,$8A,$32,$D3,$FD,$25,$F4,$4F,$A2,$FF,$0D,$F2,$BE
	.dc.b	$8D,$F1,$3F,$B6,$AD,$5F,$FA,$7A,$B5,$39,$B9,$3F,$EA,$2F,$A3,$7D
	.dc.b	$1B,$E8,$F3,$B7,$B7,$E7,$BE,$3F,$97,$F2,$1B,$87,$C8,$6C,$EE,$3E
	.dc.b	$5C,$ED,$A6,$CB,$1F,$EF,$AF,$A3,$74,$7F,$0E,$6C,$FF,$AC,$7A,$3E
	.dc.b	$A3,$FE,$B3,$8C,$D3,$97,$FD,$97,$5F,$E7,$BA,$E1,$E8,$FF,$18,$FA
	.dc.b	$27,$C5,$FF,$1A,$AD,$61,$C6,$FF,$CC,$75,$A7,$FD,$80,$D8,$D0,$1F
	.dc.b	$46,$FA,$3E,$D8,$77,$F6,$26,$1C,$BE,$3E,$E1,$C8,$63,$FC,$71,$C3
	.dc.b	$74,$FF,$72,$FD,$1F,$5B,$E4,$7F,$B4,$DC,$57,$0B,$FE,$6B,$18,$AF
	.dc.b	$F7,$6D,$F3,$BF,$F3,$FF,$82,$EB,$FF,$31,$C3,$E8,$BF,$CF,$57,$FE
	.dc.b	$F4,$AD,$4F,$FB,$98,$D7,$1F,$F4,$28,$CE,$7E,$B6,$D2,$BE,$31,$D7
	.dc.b	$93,$86,$32,$3F,$DD,$95,$C1,$DE,$2A,$D6,$1F,$1E,$5E,$FF,$A2,$7E
	.dc.b	$A6,$3D,$EB,$BC,$E7,$7F,$EA,$F7,$C0,$FF,$7D,$16,$7E,$DB,$F0,$7E
	.dc.b	$63,$75,$23,$6F,$C6,$C5,$39,$AF,$B5,$F4,$6F,$8A,$D1,$FD,$99,$C6
	.dc.b	$1F,$26,$ED,$70,$AD,$1D,$7E,$2A,$D7,$E3,$73,$F6,$53,$0E,$31,$86
	.dc.b	$CF,$C8,$AE,$35,$F8,$DF,$77,$FC,$5D,$3F,$77,$E8,$3F,$BC,$C6,$32
	.dc.b	$46,$3D,$A9,$5E,$7E,$9B,$87,$39,$B9,$FC,$FA,$FC,$45,$61,$FF,$65
	.dc.b	$F8,$5F,$6E,$FF,$DA,$1F,$42,$FA,$3F,$F0,$4C,$E6,$D3,$9B,$F8,$07
	.dc.b	$2E,$5D,$9F,$D1,$3E,$13,$FC,$09,$8F,$F8,$12,$33,$3C,$23,$0F,$F8
	.dc.b	$4E,$77,$0C,$3C,$8D,$F9,$BF,$EE,$7F,$47,$89,$46,$30,$F3,$B3,$0F
	.dc.b	$F0,$F6,$2E,$DC,$63,$FB,$BC,$3E,$5D,$78,$E3,$C1,$BF,$F8,$A1,$FE
	.dc.b	$11,$F4,$7F,$C2,$7E,$8D,$F4,$5F,$22,$BE,$29,$2F,$95,$F0,$15,$C1
	.dc.b	$FF,$12,$CF,$55,$7D,$6F,$A5,$D9,$FF,$7F,$9A,$39,$C6,$9A,$6A,$78
	.dc.b	$3E,$A9,$AF,$8E,$FC,$4E,$4E,$27,$F2,$53,$B3,$AC,$3E,$2C,$68,$E5
	.dc.b	$7E,$19,$E7,$18,$3D,$B5,$E0,$79,$CE,$CF,$5B,$83,$67,$53,$07,$96
	.dc.b	$7A,$B8,$AF,$1C,$62,$FD,$F1,$8A,$EB,$18,$C3,$F8,$75,$6E,$FD,$F7
	.dc.b	$B7,$E4,$1E,$43,$EC,$16,$B0,$E0,$F8,$DF,$47,$FC,$73,$E1,$38,$36
	.dc.b	$3D,$AD,$7E,$CB,$5E,$3F,$19,$FD,$85,$DB,$5D,$B5,$DF,$27,$68,$E1
	.dc.b	$8F,$DA,$FA,$37,$CF,$FD,$98,$C3,$75,$E3,$87,$DE,$C5,$87,$29,$FC
	.dc.b	$06,$1C,$AB,$F1,$2D,$78,$4E,$DF,$ED,$55,$CB,$1D,$E1,$DE,$7F,$B7
	.dc.b	$5F,$1C,$7F,$1F,$7B,$9E,$77,$C1,$86,$09,$3E,$A7,$2F,$2C,$3C,$50
	.dc.b	$AE,$93,$B8,$FC,$DD,$F8,$71,$F8,$33,$C7,$0E,$71,$F0,$31,$F4,$5F
	.dc.b	$3C,$7A,$E1,$F0,$6F,$F9,$DB,$76,$E1,$5F,$C1,$F2,$F0,$E3,$88,$FF
	.dc.b	$03,$53,$C5,$FD,$66,$1C,$9B,$86,$E3,$FA,$3F,$D1,$BA,$73,$87,$F8
	.dc.b	$5F,$E8,$FF,$CD,$CF,$F1,$56,$BF,$E1,$39,$E0,$FF,$19,$6B,$FE,$17
	.dc.b	$8D,$D7,$F5,$46,$DF,$EF,$EB,$73,$E2,$F9,$2F,$0D,$A5,$18,$6C,$78
	.dc.b	$FC,$1C,$4F,$0C,$1E,$35,$33,$5E,$59,$C3,$EB,$8C,$BF,$E2,$DF,$A2
	.dc.b	$FE,$76,$FF,$33,$6B,$9E,$9E,$DE,$67,$9A,$73,$9F,$B7,$E8,$FC,$46
	.dc.b	$6A,$5F,$C9,$8C,$BD,$63,$DF,$F2,$FF,$E7,$28,$FE,$84,$D7,$FC,$80
	.dc.b	$DA,$F9,$5D,$BC,$3D,$5F,$81,$ED,$4B,$E3,$19,$7E,$C4,$65,$E7,$1F
	.dc.b	$3B,$E2,$3F,$CB,$B1,$FD,$45,$AF,$F9,$3D,$B7,$FE,$0C,$E9,$FF,$37
	.dc.b	$CE,$FC,$BF,$AF,$0F,$87,$A7,$58,$93,$FD,$76,$74,$D9,$CB,$FE,$5D
	.dc.b	$E7,$FB,$AD,$E2,$FC,$1C,$63,$0F,$F0,$EF,$D1,$3C,$9C,$D7,$E2,$21
	.dc.b	$F4,$74,$CE,$1F,$DB,$70,$E5,$D3,$FD,$3E,$B7,$3F,$DC,$5A,$F8,$4F
	.dc.b	$EE,$9F,$46,$F3,$3F,$2A,$D8,$CC,$78,$76,$CB,$F9,$AF,$A2,$FF,$58
	.dc.b	$4F,$F7,$D6,$BF,$E7,$BA,$E4,$FF,$9F,$63,$3A,$FF,$8E,$BE,$8F,$FD
	.dc.b	$99,$3F,$E0,$56,$BB,$F7,$7B,$7A,$5F,$A8,$FA,$37,$C5,$53,$35,$FE
	.dc.b	$0D,$8C,$B7,$E8,$F8,$1B,$75,$F3,$FF,$B2,$FA,$37,$D1,$D3,$B4,$7E
	.dc.b	$DB,$E8,$FD,$3A,$FF,$94,$1F,$F2,$E5,$FF,$86,$DA,$FF,$A6,$5D,$F1
	.dc.b	$FF,$7F,$6C,$7F,$A9,$5D,$CF,$3F,$67,$B7,$14,$AE,$91,$E1,$30,$F5
	.dc.b	$F0,$FC,$84,$7C,$5F,$3B,$5D,$1E,$58,$F5,$47,$F1,$0F,$8D,$7C,$B5
	.dc.b	$25,$7A,$FE,$42,$BE,$BF,$A3,$F8,$46,$33,$FF,$38,$FD,$1B,$C6,$FE
	.dc.b	$36,$B8,$29,$88,$97,$FE,$41,$9D,$36,$72,$FF,$B0,$63,$A3,$92,$19
	.dc.b	$9E,$FF,$47,$23,$A3,$E2,$F0,$31,$1C,$E2,$E1,$E8,$9F,$E6,$3F,$46
	.dc.b	$F9,$9A,$4B,$FF,$2B,$7D,$1F,$EC,$5F,$F2,$C4,$ED,$CF,$CB,$FF,$3A
	.dc.b	$93,$C6,$3F,$DB,$51,$FE,$02,$79,$FA,$FB,$38,$76,$D8,$C9,$0E,$2E
	.dc.b	$5E,$71,$72,$FF,$50,$D7,$FB,$02,$7C,$FF,$01,$E3,$F7,$A4,$60,$87
	.dc.b	$91,$CB,$E6,$8F,$EA,$7F,$46,$FA,$0A,$49,$FF,$3C,$7A,$9F,$9A,$53
	.dc.b	$B7,$3F,$31,$FD,$72,$7F,$B0,$C0,$DD,$67,$AB,$5D,$52,$7A,$C6,$5E
	.dc.b	$F1,$95,$A7,$FB,$02,$D5,$B9,$CF,$77,$F9,$B6,$57,$C6,$F8,$5F,$97
	.dc.b	$A7,$CD,$F4,$73,$1F,$6C,$65,$F7,$C6,$5E,$A7,$AF,$04,$FF,$7A,$FD
	.dc.b	$9D,$F4,$7F,$E5,$A6,$D7,$F9,$7D,$AF,$F9,$D1,$AF,$FA,$43,$E9,$38
	.dc.b	$FA,$1F,$E6,$56,$C7,$F9,$E9,$78,$0F,$88,$D6,$1F,$95,$AF,$F4,$67
	.dc.b	$C6,$BE,$13,$F9,$2F,$0F,$FB,$06,$31,$F0,$08,$C4,$EA,$BD,$B5,$A9
	.dc.b	$CF,$99,$B9,$CD,$7F,$8A,$67,$27,$2B,$D7,$16,$F0,$F1,$61,$8A,$FC
	.dc.b	$14,$75,$C3,$9C,$E5,$46,$6B,$96,$1E,$56,$E1,$B9,$EB,$87,$84,$F1
	.dc.b	$DF,$9A,$75,$7C,$67,$C2,$FA,$67,$C3,$1C,$5A,$71,$79,$DF,$81,$A7
	.dc.b	$2A,$EB,$87,$66,$C7,$39,$F4,$CE,$63,$8D,$7B,$AF,$C5,$5A,$9E,$07
	.dc.b	$9A,$B8,$57,$92,$7C,$4D,$9F,$11,$D3,$0F,$90,$F1,$39,$C7,$7A,$E7
	.dc.b	$5D,$E3,$DC,$73,$3C,$98,$7A,$9C,$C6,$CF,$09,$6B,$C3,$1C,$6F,$C2
	.dc.b	$F7,$8F,$0C,$7D,$68,$E7,$1E,$56,$9B,$2C,$6E,$7B,$CE,$DB,$3B,$6C
	.dc.b	$EE,$7C,$71,$E5,$9F,$5C,$7C,$5C,$3E,$7B,$CC,$D9,$63,$67,$AE,$36
	.dc.b	$F7,$8D,$B4,$DC,$F7,$AF,$82,$78,$E3,$33,$D6,$33,$5F,$93,$9C,$9C
	.dc.b	$E3,$38,$7D,$6A,$E1,$5E,$29,$E8,$6A,$3E,$6C,$73,$8C,$57,$CA,$8C
	.dc.b	$1C,$4B,$87,$DD,$C3,$E4,$C6,$2F,$E3,$4E,$5F,$7D,$F8,$27,$96,$1E
	.dc.b	$5C,$3D,$11,$8A,$F1,$CE,$6B,$85,$F3,$BF,$7C,$62,$BB,$92,$3E,$84
	.dc.b	$EE,$7D,$78,$79,$AF,$A6,$1E,$7B,$DE,$1C,$6B,$E9,$4F,$8E,$7E,$15
	.dc.b	$70,$25,$79,$89,$1A,$6D,$77,$69,$F0,$AB,$C1,$1D,$B0,$F3,$47,$78
	.dc.b	$E1,$5C,$C8,$E0,$95,$CF,$E9,$3F,$8A,$FD,$27,$F1,$66,$C7,$F3,$46
	.dc.b	$D7,$F3,$06,$D7,$F3,$38,$7D,$1F,$FA,$4E,$BF,$B7,$4F,$F3,$7D,$5F
	.dc.b	$49,$FE,$3E,$9C,$9F,$DC,$EF,$8F,$B3,$FC,$C6,$B7,$E5,$7F,$33,$9C
	.dc.b	$D6,$DE,$18,$F7,$70,$68,$AE,$5F,$EA,$7A,$7F,$8B,$67,$FA,$7E,$1F
	.dc.b	$CD,$3C,$4F,$E6,$AD,$8C,$9F,$D5,$23,$2D,$33,$5F,$D4,$A3,$2F,$F5
	.dc.b	$4F,$A4,$FE,$7E,$7F,$40,$FA,$4B,$5F,$D7,$A2,$CF,$F5,$A9,$BA,$FC
	.dc.b	$BF,$A1,$FE,$F7,$7F,$1A,$BF,$C1,$FF,$49,$FB,$75,$3A,$7F,$C0,$74
	.dc.b	$9D,$4E,$63,$A7,$57,$F5,$DF,$8E,$7C,$07,$F9,$1B,$1E,$D8,$F4,$AF
	.dc.b	$EA,$FF,$49,$27,$FB,$AF,$55,$AF,$EE,$33,$71,$FE,$CB,$36,$BF,$AE
	.dc.b	$CD,$AF,$E9,$D3,$7A,$7F,$C0,$95,$FD,$93,$E7,$3F,$B2,$B4,$FF,$01
	.dc.b	$96,$BF,$BF,$B7,$0F,$8D,$E2,$7E,$1F,$E9,$3F,$92,$D7,$C4,$8C,$75
	.dc.b	$FE,$27,$74,$53,$87,$65,$C7,$FC,$29,$6F,$A4,$FE,$F8,$D8,$CF,$C3
	.dc.b	$62,$64,$70,$77,$18,$9F,$F0,$C9,$C4,$B1,$A9,$E0,$78,$3E,$93,$BF
	.dc.b	$C4,$78,$1B,$19,$3C,$B1,$96,$9F,$63,$E8,$DF,$01,$F7,$7E,$91,$F1
	.dc.b	$5F,$A9,$9C,$D7,$E2,$FA,$15,$A8,$E5,$7B,$3F,$C4,$5F,$48,$FA,$47
	.dc.b	$D2,$66,$BE,$3F,$CF,$FC,$6F,$37,$E3,$B6,$32,$7C,$C8,$CB,$4C,$F0
	.dc.b	$7C,$68,$CB,$63,$F2,$7F,$48,$E9,$FE,$04,$D8,$FC,$BF,$8E,$E4,$CD
	.dc.b	$7F,$07,$8F,$F3,$9C,$EB,$D0,$FE,$15,$F4,$2F,$8C,$D4,$61,$93,$FA
	.dc.b	$2F,$A4,$7D,$26,$34,$FF,$37,$8C,$1E,$CF,$43,$AF,$C8,$5C,$3F,$57
	.dc.b	$28,$CC,$7E,$BB,$E9,$3A,$BA,$8F,$E3,$D1,$8B,$EF,$5C,$6B,$86,$1C
	.dc.b	$A7,$C7,$E1,$BA,$FF,$CB,$14,$FA,$37,$1F,$F2,$9A,$D4,$FF,$0B,$35
	.dc.b	$19,$A9,$19,$A7,$26,$30,$EF,$57,$C4,$FF,$39,$5F,$7A,$C3,$FC,$EC
	.dc.b	$7B,$FE,$48,$FA,$11,$FC,$38,$C9,$E3,$73,$7F,$3A,$32,$D9,$F6,$E0
	.dc.b	$F9,$9F,$E8,$95,$AC,$3F,$A9,$34,$F1,$57,$09,$DB,$FD,$2F,$E9,$1F
	.dc.b	$15,$D0,$C1,$6B,$F9,$1C,$69,$B1,$E2,$F8,$BD,$C6,$09,$97,$FA,$D1
	.dc.b	$FA,$51,$A8,$FC,$4F,$B7,$FE,$3A,$9F,$AD,$F4,$3F,$D7,$63,$1A,$23
	.dc.b	$1E,$54,$AF,$A9,$D5,$70,$E7,$37,$6F,$A7,$87,$FA,$CE,$FF,$9D,$E3
	.dc.b	$FC,$93,$C3,$FC,$0E,$1F,$EC,$8C,$5F,$44,$FA,$4F,$ED,$D5,$AC,$BF
	.dc.b	$BC,$4E,$71,$FE,$8B,$39,$7F,$BC,$37,$0F,$F6,$17,$C2,$7F,$73,$BF
	.dc.b	$F0,$24,$E6,$4D,$38,$62,$AE,$17,$D3,$93,$C7,$F3,$5F,$EB,$3C,$5F
	.dc.b	$49,$C6,$30,$DA,$F2,$C7,$F8,$27,$3F,$DA,$C6,$0F,$4F,$07,$B8,$FF
	.dc.b	$06,$68,$1F,$49,$7E,$91,$F4,$7E,$EA,$E1,$7C,$8F,$EC,$F7,$C6,$F8
	.dc.b	$56,$E7,$33,$D0,$5F,$47,$C1,$BF,$DF,$9E,$31,$FE,$1D,$BA,$FC,$8C
	.dc.b	$3F,$DC,$17,$C6,$3F,$C5,$13,$AB,$E5,$1D,$B0,$FF,$11,$CE,$6F,$8C
	.dc.b	$62,$74,$78,$A3,$2F,$84,$C9,$E1,$C3,$C6,$FB,$CC,$DF,$83,$0F,$24
	.dc.b	$65,$F3,$46,$6F,$C1,$18,$BE,$F5,$EA,$39,$B8,$79,$4F,$CB,$7C,$66
	.dc.b	$0F,$45,$F9,$1F,$9D,$5D,$63,$18,$7C,$16,$CF,$F9,$18,$F5,$16,$30
	.dc.b	$E0,$FF,$2A,$7D,$27,$F1,$AF,$7B,$83,$63,$FC,$9E,$D7,$FC,$A8,$D7
	.dc.b	$8F,$C6,$37,$E2,$D8,$E2,$D7,$8B,$5E,$31,$F8,$2F,$A4,$7D,$07,$F9
	.dc.b	$9F,$3F,$C1,$46,$2B,$F0,$F8,$30,$ED,$5F,$87,$8C,$61,$CF,$0E,$5A
	.dc.b	$7F,$9C,$71,$E5,$8F,$E0,$2F,$6D,$C7,$9D,$6E,$F8,$CF,$EA,$6F,$8E
	.dc.b	$3D,$EF,$73,$FA,$DC,$39,$CF,$A2,$F8,$CA,$F8,$E1,$F0,$23,$C1,$8F
	.dc.b	$9D,$86,$09,$7C,$DC,$BE,$18,$F8,$78,$F8,$A7,$B6,$1E,$0C,$3B,$C7
	.dc.b	$96,$75,$1F,$0A,$BF,$D3,$8D,$D3,$E0,$57,$F9,$9B,$2F,$5E,$58,$8F
	.dc.b	$F0,$2D,$7F,$7F,$7F,$CD,$6D,$73,$8F,$FA,$7F,$0F,$F5,$47,$D2,$3A
	.dc.b	$71,$97,$FA,$8F,$E9,$3F,$CA,$F8,$7F,$9D,$1A,$FF,$84,$AB,$83,$FE
	.dc.b	$78,$6B,$FE,$16,$8D,$DF,$FA,$D5,$B5,$FE,$AF,$AD,$CF,$8B,$E4,$FC
	.dc.b	$37,$FC,$8E,$51,$88,$46,$1F,$27,$81,$98,$C4,$78,$7C,$2E,$18,$FF
	.dc.b	$84,$A3,$CD,$7B,$33,$32,$32,$7F,$A4,$BE,$91,$CF,$E1,$6C,$66,$3A
	.dc.b	$F9,$99,$7B,$FD,$24,$9D,$10,$F8,$8E,$5F,$F1,$DD,$74,$7F,$C7,$BF
	.dc.b	$2F,$FE,$68,$7F,$D4,$6D,$73,$EC,$7B,$3B,$7E,$24,$E7,$7F,$9D,$E8
	.dc.b	$FC,$03,$35,$2F,$ED,$46,$5E,$B1,$F3,$7E,$0B,$FC,$A9,$1F,$EB,$26
	.dc.b	$BF,$E5,$06,$D7,$FB,$BF,$93,$FC,$C9,$3B,$F4,$7E,$B4,$3E,$1E,$9D
	.dc.b	$62,$4F,$FA,$FA,$74,$D9,$CB,$FE,$5E,$F4,$3E,$DD,$FF,$BF,$5B,$CD
	.dc.b	$FE,$F7,$6C,$FF,$69,$FA,$37,$8F,$F8,$79,$CC,$A5,$F4,$9F,$8B,$6C
	.dc.b	$7F,$B4,$27,$67,$08,$CC,$7F,$9C,$1B,$87,$8F,$9B,$F2,$9F,$48,$F2
	.dc.b	$E9,$2F,$FD,$B7,$7F,$23,$3F,$EF,$D0,$AE,$0F,$FB,$83,$E9,$3F,$D5
	.dc.b	$2D,$F7,$39,$3F,$E8,$08,$CB,$FE,$82,$FA,$4F,$F4,$83,$FE,$E8,$6B
	.dc.b	$B9,$FD,$44,$FF,$98,$A3,$FD,$DB,$F4,$8F,$83,$FD,$63,$63,$C5,$E9
	.dc.b	$E2,$73,$F4,$92,$78,$C7,$0F,$A0,$FD,$A7,$D2,$3E,$92,$99,$AF,$F7
	.dc.b	$DF,$C5,$FF,$A5,$9B,$EF,$7E,$EE,$31,$1F,$E9,$CF,$90,$FF,$19,$3A
	.dc.b	$3F,$C4,$9C,$AE,$2F,$93,$2F,$AE,$FE,$08,$DA,$E0,$7F,$A9,$AF,$AD
	.dc.b	$F1,$8F,$F1,$74,$ED,$F0,$57,$8F,$0F,$E1,$DF,$4B,$F5,$27,$FD,$5E
	.dc.b	$FF,$9D,$9F,$1E,$3F,$6A,$31,$1A,$E6,$E2,$72,$F6,$E4,$70,$3F,$C7
	.dc.b	$D3,$D7,$2F,$24,$21,$E3,$F8,$89,$1C,$4D,$7C,$EE,$1F,$2C,$5A,$E7
	.dc.b	$4A,$F9,$DF,$2D,$2F,$E9,$BE,$EF,$92,$C4,$7B,$BE,$97,$E0,$98,$C3
	.dc.b	$CB,$C0,$BC,$D7,$8F,$15,$EF,$5F,$CE,$FD,$2B,$C8,$FE,$70,$B9,$69
	.dc.b	$88,$98,$7B,$67,$4D,$9C,$BF,$EE,$08,$7D,$2F,$42,$47,$1F,$02,$BB
	.dc.b	$8B,$AF,$CE,$AF,$E0,$7F,$4A,$F9,$9F,$E9,$8D,$BF,$9C,$D7,$FD,$DF
	.dc.b	$F4,$9F,$C3,$1C,$CF,$F5,$38,$7D,$2D,$39,$C4,$9F,$B1,$3A,$6C,$E5
	.dc.b	$EB,$ED,$73,$23,$C3,$BE,$BA,$39,$C7,$86,$3C,$5F,$05,$E3,$EA,$95
	.dc.b	$E0,$8C,$BE,$28,$CB,$EC,$8F,$ED,$9F,$4A,$FA,$5A,$49,$FC,$87,$AD
	.dc.b	$F9,$7E,$4F,$E3,$35,$B9,$CF,$CC,$7F,$75,$AF,$EF,$B1,$FE,$9B,$9D
	.dc.b	$3F,$E5,$1A,$E3,$7B,$B9,$3F,$9D,$AD,$F0,$FE,$93,$0F,$D1,$CF,$8B
	.dc.b	$AF,$B1,$DF,$B2,$46,$48,$77,$3F,$D5,$93,$C9,$C0,$5D,$9E,$9F,$A5
	.dc.b	$93,$F0,$23,$D8,$61,$E3,$29,$CC,$7E,$D7,$E9,$1F,$49,$F8,$78,$FE
	.dc.b	$7E,$23,$3E,$87,$FA,$05,$78,$0F,$89,$C5,$F1,$BF,$E8,$F5,$A9,$FF
	.dc.b	$5F,$78,$38,$BE,$15,$FE,$C8,$8F,$E9,$BF,$53,$FE,$DB,$E3,$DC,$FF
	.dc.b	$B3,$70,$E4,$EC,$FF,$68,$4E,$DC,$1E,$7B,$EB,$F0,$DF,$C0,$63,$7C
	.dc.b	$BF,$CA,$E3,$11,$BC,$BB,$AB,$E9,$8F,$C2,$6C,$79,$AF,$F9,$A4,$E5
	.dc.b	$B1,$EA,$F4,$EB,$E6,$6B,$D5,$B5,$FD,$F2,$BF,$DD,$47,$78,$F3,$67
	.dc.b	$CE,$3C,$B2,$F4,$F3,$7D,$67,$31,$88,$ED,$18,$38,$DE,$6B,$94,$60
	.dc.b	$D4,$FC,$A8,$F2,$36,$B8,$57,$16,$D7,$0C,$BE,$5D,$F0,$6C,$EA,$F9
	.dc.b	$57,$76,$9D,$EF,$A4,$66,$B8,$B7,$0F,$C2,$DF,$39,$CD,$F1,$AF,$15
	.dc.b	$7C,$78,$EB,$3E,$4B,$EB,$1B,$33,$3B,$9E,$66,$EB,$9C,$FC,$D3,$65
	.dc.b	$8D,$CF,$13,$CB,$3E,$79,$FC,$AD,$F9,$E7,$31,$E1,$AE,$2F,$38,$CC
	.dc.b	$78,$63,$B9,$F6,$8B,$87,$CD,$9D,$9E,$18,$EF,$7E,$17,$F3,$57,$E1
	.dc.b	$72,$78,$67,$8B,$E1,$33,$7E,$13,$CA,$F8,$5F,$9E,$FE,$8E,$7F,$C0
	.dc.b	$2F,$B0,$C5,$7F,$08,$7B,$98,$31,$7F,$90,$6C,$F2,$7C,$BC,$9F,$E0
	.dc.b	$C9,$D3,$7D,$8F,$F1,$13,$9A,$B7,$F3,$23,$0F,$F8,$3E,$95,$AB,$EB
	.dc.b	$3B,$23,$FC,$2A,$34,$FC,$88,$E5,$78,$AF,$1C,$7F,$16,$AC,$4C,$8D
	.dc.b	$9E,$18,$C5,$7F,$38,$AE,$2F,$88,$F0,$4F,$07,$55,$C2,$FE,$04,$E5
	.dc.b	$F6,$CE,$A3,$B1,$93,$D6,$FA,$4F,$05,$F3,$8C,$DF,$8A,$7F,$8F,$4E
	.dc.b	$5F,$21,$C6,$BC,$2F,$72,$3E,$23,$25,$AE,$B7,$F1,$EB,$A5,$73,$9F
	.dc.b	$0C,$62,$FC,$44,$71,$59,$6C,$FB,$6B,$E3,$9E,$27,$B9,$F0,$9C,$1A
	.dc.b	$BF,$23,$63,$C2,$7E,$55,$F8,$2D,$BF,$C3,$B6,$BF,$29,$3F,$D0,$F1
	.dc.b	$B5,$FE,$4B,$AE,$11,$82,$3F,$91,$7F,$11,$D9,$F8,$96,$C6,$4F,$CB
	.dc.b	$46,$5A,$78,$DE,$E4,$7F,$59,$F1,$EF,$87,$FF,$51,$47,$F9,$D6,$31
	.dc.b	$F0,$7F,$9F,$75,$E8,$5A,$4D,$8C,$36,$B9,$4E,$9B,$F4,$9C,$7E,$97
	.dc.b	$93,$63,$27,$08,$CB,$4F,$F4,$07,$D2,$7F,$BA,$DF,$F4,$57,$23,$E9
	.dc.b	$7E,$56,$FF,$E1,$CA,$D3,$FE,$21,$9D,$6D,$FB,$49,$CF,$0F,$FA,$41
	.dc.b	$B8,$2D,$F1,$AF,$80,$FC,$F5,$6A,$74,$D4,$E6,$78,$72,$77,$F7,$BE
	.dc.b	$47,$93,$F2,$5B,$8F,$C6,$6E,$1F,$E3,$7E,$3F,$A9,$18,$CF,$FD,$71
	.dc.b	$18,$AF,$9B,$EA,$FD,$33,$EA,$7D,$2D,$8C,$93,$0E,$9C,$69,$D4,$7F
	.dc.b	$90,$BE,$8B,$FC,$83,$5F,$EB,$5C,$5F,$4B,$F8,$4A,$D7,$89,$FC,$86
	.dc.b	$73,$D1,$FC,$7B,$0E,$35,$AC,$3F,$D7,$DC,$1F,$27,$A7,$58,$C9,$FF
	.dc.b	$5D,$5F,$0C,$7F,$DA,$97,$B9,$CD,$7E,$3F,$E9,$64,$7E,$33,$C0,$B9
	.dc.b	$7B,$26,$E1,$F1,$B6,$FF,$25,$DF,$E4,$A7,$F3,$0F,$E6,$7E,$72,$BF
	.dc.b	$96,$6D,$FC,$66,$DF,$37,$DD,$E5,$7F,$41,$E9,$7D,$2F,$E7,$6F,$E5
	.dc.b	$4E,$63,$17,$FE,$DC,$D9,$3C,$B1,$FC,$CD,$7C,$DA,$D4,$E6,$F9,$D6
	.dc.b	$B8,$31,$5F,$82,$97,$D2,$FE,$D5,$B5,$FB,$95,$F1,$19,$99,$39,$78
	.dc.b	$46,$27,$F2,$03,$E9,$7A,$7C,$57,$56,$DF,$56,$CF,$06,$E1,$FE,$FA
	.dc.b	$FA,$57,$C3,$7F,$67,$E5,$7D,$2F,$F2,$4C,$3F,$D2,$35,$AC,$7F,$0F
	.dc.b	$39,$C3,$96,$1F,$DA,$AB,$57,$FE,$9B,$FA,$57,$D2,$BE,$97,$57,$32
	.dc.b	$FD,$D6,$3C,$9A,$7E,$F6,$F8,$D6,$F0,$FF,$51,$FD,$2B,$9B,$FB,$CB
	.dc.b	$6F,$8C,$FF,$05,$E6,$F9,$B1,$88,$4E,$DF,$F5,$6C,$BD,$4F,$F0,$1B
	.dc.b	$6B,$87,$D1,$BE,$47,$35,$FE,$08,$8C,$4E,$A7,$FC,$17,$F4,$AF,$A4
	.dc.b	$D4,$6F,$DC,$95,$D3,$C8,$B9,$FF,$AF,$B1,$FE,$EC,$7F,$77,$FA,$5E
	.dc.b	$55,$B3,$D4,$E2,$78,$57,$F8,$6A,$31,$C2,$AF,$E4,$D4,$BF,$F0,$17
	.dc.b	$8E,$E9,$FF,$13,$53,$E8,$FF,$D9,$CF,$C7,$9D,$3F,$30,$D7,$B7,$FE
	.dc.b	$28,$8C,$E1,$FE,$D6,$6E,$3F,$EA,$6B,$7D,$2E,$F1,$98,$F5,$7C,$33
	.dc.b	$A8,$C1,$F5,$C7,$C0,$79,$AB,$E7,$E3,$FE,$E4,$9C,$9E,$58,$CE,$1F
	.dc.b	$06,$7F,$D5,$B5,$FE,$3E,$AD,$C7,$FB,$BD,$F8,$1F,$39,$99,$91,$E1
	.dc.b	$7C,$B1,$89,$FF,$79,$FD,$2B,$E3,$7D,$27,$F9,$29,$C5,$FC,$7C,$AD
	.dc.b	$70,$AD,$1F,$03,$E3,$AD,$7E,$56,$F7,$97,$A6,$D1,$89,$FF,$63,$E0
	.dc.b	$BE,$34,$BE,$1F,$01,$FC,$72,$F7,$F4,$5F,$E5,$E8,$C5,$7E,$17,$04
	.dc.b	$E7,$DA,$97,$E9,$EA,$B7,$D6,$3F,$17,$87,$28,$3D,$23,$F1,$53,$C3
	.dc.b	$1F,$57,$87,$F8,$49,$FD,$A7,$D1,$BE,$97,$FC,$E3,$39,$C5,$82,$B5
	.dc.b	$7F,$B9,$AD,$57,$F8,$03,$8F,$F9,$F7,$C0,$FF,$7D,$9C,$92,$33,$EB
	.dc.b	$76,$7B,$FC,$F7,$F6,$5F,$A5,$DB,$63,$38,$F7,$8F,$E9,$59,$3B,$73
	.dc.b	$9F,$C3,$3F,$C2,$59,$7C,$AB,$F3,$70,$FF,$A3,$27,$FC,$2D,$F0,$5F
	.dc.b	$E1,$7F,$A5,$7D,$27,$ED,$30,$E1,$5C,$CF,$F0,$DD,$FC,$43,$92,$BE
	.dc.b	$55,$C6,$73,$18,$9F,$1C,$7E,$86,$FB,$C7,$F0,$23,$90,$72,$F8,$30
	.dc.b	$E4,$FF,$08,$AB,$87,$9A,$B7,$3E,$C3,$96,$1E,$3B,$E3,$19,$9E,$B5
	.dc.b	$C6,$BF,$D8,$D7,$D3,$0F,$34,$62,$F9,$9E,$AC,$39,$BE,$8C,$39,$D7
	.dc.b	$08,$F6,$5F,$3A,$DB,$63,$99,$EB,$2E,$1E,$37,$89,$8A,$E8,$EE,$FB
	.dc.b	$DF,$48,$D3,$5F,$F4,$65,$FF,$A3,$5F,$88,$60,$FF,$72,$7D,$2F,$F1
	.dc.b	$AF,$5B,$FD,$24,$D8,$CF,$67,$F9,$F5,$B3,$AF,$8D,$FD,$E4,$72,$6B
	.dc.b	$F6,$E9,$18,$8F,$E7,$F1,$8E,$6E,$47,$DC,$FA,$57,$D0,$FF,$44,$D3
	.dc.b	$F0,$31,$8B,$FC,$0D,$F1,$7E,$4B,$67,$75,$C7,$7F,$E9,$B7,$C6,$F7
	.dc.b	$8F,$A2,$FF,$1B,$87,$1B,$DE,$1C,$6F,$FC,$F5,$87,$1B,$DE,$1C,$6F
	.dc.b	$D6,$DA,$DD,$F1,$57,$C6,$B7,$8F,$AA,$F8,$56,$AF,$85,$EE,$3F,$D1
	.dc.b	$58,$F9,$5B,$5C,$27,$51,$C4,$E9,$71,$FC,$6C,$68,$DD,$73,$8E,$07
	.dc.b	$EF,$63,$E3,$4F,$F3,$D6,$EB,$E8,$9E,$78,$7F,$78,$AD,$78,$B6,$BE
	.dc.b	$7B,$6B,$E1,$9F,$E1,$DF,$A5,$FF,$5F,$BF,$E1,$FF,$A5,$FF,$2D,$B7
	.dc.b	$D8,$E0,$A7,$6F,$F1,$16,$E1,$FD,$46,$B8,$CF,$F1,$56,$C7,$F5,$59
	.dc.b	$DE,$1F,$E0,$E5,$F2,$DD,$9B,$7E,$58,$F0,$DB,$C4,$D4,$63,$C7,$95
	.dc.b	$B2,$1E,$17,$2F,$9E,$3D,$7F,$48,$EA,$A4,$9F,$85,$E7,$57,$D5,$F4
	.dc.b	$BE,$23,$1A,$FF,$6F,$BE,$53,$F2,$67,$4F,$F9,$26,$32,$FF,$37,$F9
	.dc.b	$A7,$C7,$7C,$C6,$C7,$F9,$3D,$AF,$C9,$EF,$D4,$FC,$2F,$6E,$4E,$E3
	.dc.b	$2F,$D9,$8C,$BD,$E3,$EC,$FC,$4F,$F9,$4E,$3F,$CB,$8D,$7F,$A5,$37
	.dc.b	$2F,$F0,$57,$2F,$F9,$9E,$77,$E8,$FE,$0A,$1E,$25,$3A,$C4,$9F,$F3
	.dc.b	$44,$E9,$B3,$97,$FA,$C7,$3F,$DE,$6F,$0F,$E3,$A3,$18,$7F,$86,$FE
	.dc.b	$97,$FC,$0F,$5F,$E7,$39,$CF,$3F,$F9,$5A,$5F,$21,$F9,$66,$D6,$F0
	.dc.b	$E8,$66,$64,$64,$FF,$3F,$7D,$2F,$F8,$95,$BE,$77,$E8,$23,$11,$FD
	.dc.b	$BE,$FE,$36,$1F,$E6,$69,$CE,$5F,$E3,$1F,$A4,$FF,$52,$DF,$FA,$25
	.dc.b	$AF,$F7,$79,$DD,$7F,$9B,$A7,$35,$FE,$39,$FA,$5F,$F5,$ED,$FF,$A4
	.dc.b	$5A,$FF,$7F,$9D,$D7,$1F,$13,$F6,$BF,$4A,$F8,$4A,$7E,$E2,$BF,$CC
	.dc.b	$77,$FE,$9B,$9D,$37,$E9,$78,$9C,$1D,$7D,$07,$F0,$0F,$A5,$7D,$2D
	.dc.b	$3B,$C7,$F0,$4F,$8C,$FF,$0F,$57,$FA,$9D,$AF,$F8,$55,$BB,$7F,$95
	.dc.b	$7E,$45,$E9,$F1,$F4,$7F,$86,$C3,$E9,$69,$9C,$3A,$4E,$1F,$F5,$A4
	.dc.b	$ED,$C9,$FE,$B7,$FA,$5F,$9A,$83,$FE,$71,$75,$EE,$7F,$1C,$8D,$C6
	.dc.b	$B9,$F1,$32,$3A,$46,$FC,$59,$1D,$DE,$5F,$3D,$88,$FF,$1A,$5F,$FB
	.dc.b	$15,$D1,$E5,$8F,$91,$F2,$F2,$3E,$43,$E3,$F9,$2C,$47,$2F,$A5,$F3
	.dc.b	$18,$DF,$FC,$FD,$C0,$BC,$D7,$C1,$AD,$74,$FF,$AE,$BE,$91,$E4,$7F
	.dc.b	$3A,$5C,$94,$C4,$4B,$FF,$70,$4E,$9B,$39,$7F,$CA,$10,$FA,$5E,$84
	.dc.b	$8E,$3E,$0D,$77,$17,$4F,$F4,$B5,$7E,$EA,$7E,$B7,$D2,$BE,$6B,$FA
	.dc.b	$7B,$63,$FD,$E4,$D7,$FC,$C1,$F4,$BF,$C1,$1F,$F3,$14,$66,$FF,$D4
	.dc.b	$52,$F9,$9F,$EB,$2D,$AE,$77,$D8,$CC,$C8,$C9,$FF,$00,$67,$FC,$4E
	.dc.b	$DF,$63,$99,$D3,$C2,$B1,$92,$1E,$67,$C3,$D9,$E2,$EB,$93,$D6,$32
	.dc.b	$F7,$8C,$BE,$E8,$FE,$E3,$F4,$8F,$A5,$A4,$9F,$CF,$69,$FE,$C1,$F5
	.dc.b	$3F,$37,$5B,$9C,$F2,$7F,$BF,$27,$FB,$9B,$7D,$8E,$53,$FE,$8B,$AE
	.dc.b	$0F,$FA,$36,$B6,$FE,$AD,$AA,$37,$7E,$1E,$CF,$F0,$3D,$3A,$E9,$88
	.dc.b	$97,$E4,$3A,$46,$9C,$09,$FD,$53,$E3,$EB,$EB,$F4,$B2,$37,$29,$F0
	.dc.b	$99,$8F,$54,$E1,$EB,$F4,$8F,$A5,$FE,$9B,$F0,$3F,$AB,$9F,$E1,$13
	.dc.b	$9F,$88,$FF,$59,$F7,$FF,$0B,$94,$E5,$FF,$57,$46,$7F,$10,$7D,$9F
	.dc.b	$FA,$8E,$38,$56,$EB,$FD,$58,$D7,$3E,$17,$FB,$69,$FF,$58,$1D,$B2
	.dc.b	$FC,$F3,$E1,$FA,$7F,$E5,$3C,$9E,$48,$F3,$D7,$1D,$3D,$50,$AD,$BE
	.dc.b	$63,$7F,$15,$C5,$D7,$3B,$0C,$64,$C7,$A6,$7E,$1A,$D5,$F3,$3E,$1C
	.dc.b	$E9,$FE,$61,$8F,$F7,$4C,$3B,$2F,$B4,$E3,$9F,$9A,$35,$5C,$DD,$57
	.dc.b	$0F,$53,$79,$7C,$1C,$2D,$F6,$9A,$FB,$E2,$DB,$D5,$B5,$86,$E1,$E4
	.dc.b	$6D,$F8,$5B,$8F,$95,$B8,$FD,$16,$CF,$B9,$AF,$CF,$6E,$1E,$1B,$F5
	.dc.b	$1D,$CB,$3D,$DB,$7D,$DB,$7D,$DB,$3D,$DA,$F7,$6D,$77,$8F,$95,$87
	.dc.b	$BE,$73,$87,$90,$8E,$09,$3C,$1F,$33,$4F,$31,$70,$F3,$37,$0F,$33
	.dc.b	$4F,$31,$4F,$31,$70,$F3,$4F,$E0,$70,$F3,$4E,$70,$F2,$B7,$0F,$2D
	.dc.b	$6D,$F2,$B6,$3C,$AD,$3C,$A5,$C3,$CA,$D8,$F2,$B5,$F2,$B7,$0F,$2C
	.dc.b	$FE,$BF,$0F,$2C,$E7,$0F,$2B,$70,$F2,$D6,$DF,$2B,$70,$F2,$B6,$7C
	.dc.b	$AD,$3C,$A5,$8F,$2B,$5F,$2B,$70,$F2,$C7,$E8,$6B,$BB,$4E,$E5,$AE
	.dc.b	$ED,$AE,$ED,$3B,$94,$EE,$5A,$EE,$FF,$0B,$C3,$E1,$CF,$8F,$0F,$86
	.dc.b	$DC,$3E,$1D,$71,$9C,$B6,$32,$F3,$7D,$2D,$3D,$25,$C3,$D2,$DA,$F4
	.dc.b	$B4,$F4,$94,$F4,$97,$0F,$49,$FC,$3A,$7F,$88,$7B,$5F,$C6,$BA,$3F
	.dc.b	$8D,$D7,$E5,$9B,$5F,$EB,$6C,$3F,$A9,$BF,$AD,$8C,$D7,$19,$FD,$0C
	.dc.b	$61,$DE,$39,$BF,$E4,$78,$D5,$F7,$5F,$E3,$9A,$EB,$39,$8C,$4A,$FC
	.dc.b	$73,$E6,$8F,$E8,$4D,$7A,$D7,$F4,$87,$FA,$04,$65,$FE,$85,$E0,$F8
	.dc.b	$72,$EB,$7E,$E8,$C5,$65,$B7,$F0,$1B,$7E,$38,$FF,$36,$7C,$93,$E2
	.dc.b	$7F,$D2,$11,$FD,$EE,$31,$F1,$FF,$05,$D7,$AF,$69,$36,$30,$DA,$E5
	.dc.b	$3A,$6F,$D4,$F1,$FA,$9E,$4D,$8C,$9C,$23,$2D,$3F,$C1,$7F,$51,$FE
	.dc.b	$D8,$7F,$C2,$1D,$0F,$A9,$F8,$71,$FD,$F6,$75,$A7,$BC,$56,$F6,$FB
	.dc.b	$53,$AD,$FF,$12,$DB,$FF,$48,$7C,$83,$E0,$FF,$11,$1D,$7E,$07,$89
	.dc.b	$1F,$F1,$14,$EF,$97,$19,$F8,$78,$BE,$77,$9B,$F2,$70,$F9,$4D,$C3
	.dc.b	$E3,$B6,$B8,$70,$FF,$9A,$E7,$E5,$D7,$D1,$F6,$24,$7C,$EF,$A9,$B7
	.dc.b	$C2,$6E,$DF,$DF,$B2,$FB,$D5,$8B,$FF,$11,$7D,$3F,$F8,$BD,$FF,$1D
	.dc.b	$53,$EA,$7F,$07,$8F,$C9,$15,$B7,$E6,$4E,$F9,$FF,$49,$39,$8F,$D9
	.dc.b	$B6,$FE,$5F,$07,$F9,$ED,$FF,$22,$C6,$6F,$FD,$7C,$D7,$BF,$8F,$FB
	.dc.b	$1E,$1F,$A5,$D3,$DA,$3F,$25,$F5,$33,$0F,$C7,$F8,$37,$0F,$93,$37
	.dc.b	$3F,$85,$C5,$FE,$23,$BF,$CA,$CF,$E6,$DF,$CE,$7D,$05,$7F,$32,$DB
	.dc.b	$FA,$6D,$BE,$6F,$C0,$F3,$BF,$83,$74,$BE,$A7,$F4,$17,$F2,$77,$56
	.dc.b	$A7,$38,$7C,$68,$E5,$8F,$E7,$27,$E8,$CE,$63,$E3,$CE,$AF,$F3,$98
	.dc.b	$77,$E1,$C5,$FD,$F9,$7D,$4F,$EE,$5B,$87,$EF,$57,$C5,$E6,$64,$E5
	.dc.b	$E1,$18,$9F,$CD,$0F,$A9,$E9,$F1,$BD,$5B,$7D,$5B,$3C,$1B,$87,$F3
	.dc.b	$6F,$A8,$7D,$3F,$DD,$FA,$9F,$E6,$D7,$F8,$F9,$CE,$1F,$AC,$9C,$DF
	.dc.b	$1B,$FE,$65,$5A,$BF,$DF,$7D,$43,$EA,$5F,$53,$AB,$99,$7F,$A5,$30
	.dc.b	$E1,$8F,$F9,$8B,$0E,$17,$AC,$7F,$41,$F5,$2E,$5F,$E8,$0D,$C7,$FD
	.dc.b	$3B,$C5,$F2,$9F,$F4,$FC,$66,$BF,$51,$2F,$67,$FA,$43,$6B,$FD,$4B
	.dc.b	$F4,$CF,$95,$CD,$7F,$4D,$8C,$4E,$A7,$FA,$87,$D4,$BE,$A7,$51,$BF
	.dc.b	$80,$95,$D3,$C9,$B9,$FE,$E7,$1F,$E3,$28,$38,$5F,$53,$D6,$B9,$1F
	.dc.b	$EB,$A7,$11,$CE,$BF,$B1,$46,$38,$7F,$B8,$DF,$3F,$1D,$D1,$FD,$A7
	.dc.b	$17,$D3,$FF,$B0,$DF,$F6,$2C,$E9,$FF,$65,$1A,$F7,$FF,$B7,$46,$73
	.dc.b	$FE,$24,$DC,$7F,$D3,$35,$D3,$EA,$77,$8C,$8E,$AF,$87,$14,$E7,$0F
	.dc.b	$E3,$94,$F8,$4F,$34,$FD,$09,$CC,$63,$7F,$F0,$05,$6C,$FF,$6E,$96
	.dc.b	$7F,$A7,$3E,$EF,$9E,$DD,$48,$DB,$E3,$8C,$4F,$F2,$DF,$A9,$7C,$97
	.dc.b	$9C,$FF,$07,$38,$BF,$8B,$95,$AE,$15,$A3,$DD,$F2,$56,$BF,$DE,$37
	.dc.b	$BD,$7F,$C4,$AD,$8F,$C1,$1F,$52,$78,$5F,$F3,$FF,$84,$FE,$2B,$3A
	.dc.b	$C5,$F4,$DF,$E2,$19,$CD,$7F,$B3,$E7,$35,$FF,$00,$27,$32,$77,$1F
	.dc.b	$82,$E9,$E5,$ED,$48,$E0,$FD,$78,$C5,$7F,$51,$F3,$FF,$01,$5F,$99
	.dc.b	$97,$D3,$BE,$A7,$FC,$69,$5A,$BF,$EF,$F3,$9B,$FD,$06,$DF,$E4,$6A
	.dc.b	$D5,$1E,$18,$3E,$0F,$FC,$05,$3F,$E0,$2E,$5E,$A4,$C3,$FC,$23,$3A
	.dc.b	$9E,$53,$E2,$FA,$2F,$F2,$46,$2F,$A9,$EF,$92,$73,$18,$78,$63,$FE
	.dc.b	$1C,$CD,$DB,$8C,$BE,$17,$0F,$B8,$FF,$29,$3F,$BB,$F8,$3F,$DE,$7D
	.dc.b	$4B,$EA,$7F,$7D,$3E,$DA,$E1,$8F,$C2,$33,$A7,$D8,$9F,$EF,$C7,$31
	.dc.b	$7C,$DF,$F1,$1D,$78,$16,$1E,$0C,$39,$CB,$0C,$57,$88,$D9,$D2,$77
	.dc.b	$1A,$8F,$3D,$5C,$3C,$98,$78,$67,$27,$13,$C7,$5C,$5F,$44,$6E,$78
	.dc.b	$DF,$86,$7C,$33,$9C,$3C,$D1,$CC,$EB,$19,$BE,$91,$C2,$BC,$EE,$CE
	.dc.b	$EE,$27,$CD,$18,$3C,$A6,$CE,$46,$1D,$B8,$3E,$B4,$F8,$30,$F8,$57
	.dc.b	$E6,$AD,$C6,$27,$55,$FE,$87,$3E,$69,$FE,$89,$8C,$38,$3F,$07,$F5
	.dc.b	$3F,$C7,$FD,$5F,$F4,$73,$63,$E3,$35,$FB,$0D,$78,$FC,$8F,$FA,$6D
	.dc.b	$DB,$5D,$B5,$DF,$07,$68,$E1,$8F,$DB,$FA,$97,$D2,$FE,$5A,$30,$DD
	.dc.b	$78,$E1,$F5,$71,$61,$CA,$7F,$3B,$87,$2A,$FC,$53,$5E,$13,$B7,$FB
	.dc.b	$05,$72,$C7,$78,$77,$9F,$EC,$97,$C7,$1F,$C8,$5E,$E7,$9D,$F0,$61
	.dc.b	$82,$4F,$A9,$CB,$CB,$0F,$14,$2B,$A4,$EE,$3F,$25,$7E,$1C,$7E,$0C
	.dc.b	$F1,$C3,$9C,$7C,$0C,$7D,$17,$CF,$1E,$B8,$7C,$1B,$FF,$71,$B7,$4E
	.dc.b	$15,$FC,$04,$F8,$75,$C2,$FF,$83,$DF,$0F,$16,$DF,$CE,$6D,$FC,$6A
	.dc.b	$AB,$EA,$7F,$D8,$6F,$51,$F5,$3F,$E6,$07,$0D,$7F,$D1,$91,$C1,$BD
	.dc.b	$5D,$27,$58,$FF,$BF,$1B,$8F,$F1,$CB,$E1,$3F,$F0,$01,$7C,$D7,$91
	.dc.b	$B8,$78,$DA,$FF,$A6,$ED,$E0,$EE,$BC,$1E,$34,$AE,$11,$97,$D3,$19
	.dc.b	$7D,$31,$F0,$BE,$A3,$F9,$E4,$7C,$16,$BF,$EA,$96,$E3,$EB,$F3,$38
	.dc.b	$75,$7B,$FE,$A6,$5F,$18,$CB,$F2,$23,$2F,$38,$FA,$3F,$3C,$F8,$FA
	.dc.b	$49,$F9,$FD,$F5,$F8,$1E,$CF,$CD,$31,$97,$F2,$28,$5F,$37,$EB,$CE
	.dc.b	$DF,$B9,$19,$7F,$D8,$9F,$07,$FE,$59,$7F,$00,$D7,$3E,$5F,$E1,$B9
	.dc.b	$7F,$CD,$51,$A8,$C7,$C3,$C8,$C1,$0F,$F6,$84,$A3,$51,$D0,$FC,$77
	.dc.b	$3F,$DC,$CB,$D4,$DE,$6F,$EC,$2D,$9F,$F3,$07,$D4,$FF,$84,$DE,$3C
	.dc.b	$7F,$9A,$87,$C8,$D3,$F3,$B5,$C6,$8D,$9D,$C7,$E7,$DA,$FF,$B9,$E3
	.dc.b	$87,$D4,$BC,$CF,$D3,$B6,$7F,$4C,$D7,$FD,$DD,$5F,$27,$2F,$A9,$39
	.dc.b	$7E,$B7,$D4,$7F,$AA,$2B,$F5,$CD,$7F,$DE,$D3,$BC,$FF,$C0,$33,$9A
	.dc.b	$FB,$DF,$53,$FE,$C0,$BF,$DB,$35,$FF,$7F,$4E,$EB,$8F,$A3,$FB,$EF
	.dc.b	$A9,$7C,$1D,$3F,$80,$D7,$F9,$A6,$FF,$80,$4E,$9B,$F5,$3C,$4E,$0E
	.dc.b	$BE,$93,$F8,$4F,$D4,$BE,$A6,$9D,$E3,$F8,$6F,$C6,$7F,$88,$AB,$F8
	.dc.b	$63,$56,$DF,$99,$8C,$3F,$9B,$F9,$1F,$F1,$D4,$F0,$F4,$3F,$8B,$C3
	.dc.b	$EA,$69,$DA,$24,$FF,$19,$9D,$36,$72,$F8,$38,$3F,$C5,$BF,$53,$F3
	.dc.b	$E3,$FD,$58,$FF,$9E,$9F,$E4,$4D,$78,$1A,$E5,$FE,$53,$1C,$60,$EF
	.dc.b	$1F,$DA,$36,$BB,$D7,$0F,$0E,$5F,$67,$9F,$D1,$62,$3B,$9D,$23,$18
	.dc.b	$74,$9F,$9B,$F2,$F2,$BE,$6B,$E4,$F9,$2C,$47,$2C,$9F,$53,$E8,$33
	.dc.b	$87,$D1,$BC,$57,$C0,$9C,$E7,$FE,$52,$FA,$8F,$E1,$B3,$FD,$15,$70
	.dc.b	$FF,$4A,$6C,$67,$C3,$C1,$99,$91,$93,$97,$D4,$F1,$23,$D7,$C0,$AF
	.dc.b	$C2,$B4,$E3,$3F,$E4,$11,$FB,$DF,$A9,$7C,$D7,$F5,$C6,$DF,$F5,$B6
	.dc.b	$BF,$33,$EA,$7F,$85,$3F,$3A,$33,$7F,$CA,$E5,$F3,$3F,$D9,$9B,$5C
	.dc.b	$EF,$B1,$99,$91,$93,$FB,$54,$FF,$88,$1B,$EA,$73,$3A,$78,$16,$32
	.dc.b	$43,$CC,$F8,$7D,$FF,$17,$54,$9E,$B1,$97,$BC,$65,$F7,$47,$F7,$EF
	.dc.b	$A8,$7D,$4D,$24,$FF,$7F,$F4,$FF,$35,$3A,$8C,$7C,$C7,$F8,$1A,$7F
	.dc.b	$BC,$C0,$4E,$DF,$CB,$B7,$6F,$EA,$93,$C2,$4D,$9D,$72,$FF,$A7,$A1
	.dc.b	$D7,$DF,$AE,$5F,$96,$3C,$46,$1F,$D2,$CA,$73,$1E,$BE,$BF,$47,$D4
	.dc.b	$C8,$C1,$0E,$2E,$5E,$71,$FE,$24,$FA,$97,$D4,$FF,$A7,$3B,$7F,$D6
	.dc.b	$7F,$50,$FA,$9F,$EA,$5F,$06,$9C,$C6,$1F,$F5,$CD,$F2,$C7,$FD,$7D
	.dc.b	$18,$C3,$A4,$7C,$BC,$3A,$7C,$56,$A3,$18,$73,$7E,$86,$1C,$DF,$F5
	.dc.b	$CE,$1E,$0A,$FF,$53,$42,$F9,$BF,$EA,$95,$E3,$3E,$27,$71,$A9,$FF
	.dc.b	$5F,$3A,$8F,$14,$73,$9E,$0F,$80,$FC,$17,$D3,$FF,$BB,$F8,$B7,$7C
	.dc.b	$67,$42,$5F,$53,$FE,$B8,$DB,$E4,$4F,$28,$C6,$7F,$EE,$0A,$FE,$F7
	.dc.b	$87,$89,$E2,$D9,$DD,$71,$8D,$7A,$98,$8F,$04,$DA,$E7,$1E,$46,$E1
	.dc.b	$E2,$8F,$37,$A9,$BB,$98,$74,$3C,$51,$89,$E5,$36,$B9,$5F,$F8,$3A
	.dc.b	$F8,$56,$9B,$3A,$BE,$18,$FB,$F0,$B7,$F3,$26,$BF,$26,$2E,$1D,$63
	.dc.b	$A1,$2F,$C3,$3E,$FA,$DB,$C1,$B8,$78,$B0,$F9,$D1,$8B,$E5,$18,$9D
	.dc.b	$47,$DE,$AE,$51,$88,$F2,$37,$0F,$26,$1E,$28,$C5,$78,$DA,$6F,$0F
	.dc.b	$04,$EE,$FE,$A4,$63,$0F,$89,$5F,$40,$E3,$3E,$63,$E3,$38,$AE,$47
	.dc.b	$CD,$71,$3C,$A7,$AC,$61,$F6,$3D,$5E,$87,$B0,$B8,$7B,$27,$95,$76
	.dc.b	$69,$F1,$E7,$78,$7C,$66,$DF,$BC,$F0,$3D,$6B,$CC,$E8,$EF,$87,$C1
	.dc.b	$7C,$A7,$A0,$D4,$78,$5F,$82,$62,$FC,$6E,$23,$4F,$78,$CB,$F9,$EA
	.dc.b	$F1,$E1,$FA,$29,$F5,$57,$08,$C5,$78,$A3,$CE,$67,$0F,$73,$6F,$DD
	.dc.b	$3E,$33,$85,$F4,$3C,$37,$D2,$3C,$87,$9E,$33,$7D,$1B,$5D,$2F,$C9
	.dc.b	$82,$3F,$DA,$D8,$7E,$8F,$9F,$FD,$44,$FF,$AA,$E3,$33,$BC,$7F,$C5
	.dc.b	$73,$B9,$FB,$15,$C3,$DB,$D3,$CA,$EC,$FF,$5F,$B7,$D4,$FF,$3F,$8F
	.dc.b	$F7,$33,$63,$AE,$5C,$1F,$EC,$D1,$9B,$FF,$5B,$36,$BA,$36,$7F,$95
	.dc.b	$D7,$17,$BB,$71,$ED,$1E,$33,$F9,$64,$7F,$B0,$5F,$E5,$0D,$79,$D7
	.dc.b	$F4,$49,$ED,$1C,$23,$E4,$F8,$3D,$6B,$D1,$F2,$4F,$89,$FF,$43,$D7
	.dc.b	$F7,$FC,$BE,$54,$E7,$E1,$77,$5A,$C7,$FC,$CB,$39,$BE,$07,$F8,$1D
	.dc.b	$E9,$EB,$70,$BE,$2D,$9D,$D2,$BC,$5F,$53,$DB,$EA,$7B,$B6,$B7,$3A
	.dc.b	$6E,$2A,$DC,$61,$F9,$9F,$51,$FE,$D3,$8E,$FC,$EF,$A9,$F8,$B3,$9C
	.dc.b	$BF,$CF,$F3,$9D,$B8,$FB,$D3,$0F,$3F,$C7,$BE,$0B,$F1,$98,$7F,$89
	.dc.b	$71,$E5,$C3,$FE,$F5,$6D,$FE,$2D,$B2,$BE,$5C,$BD,$63,$E5,$ED,$FE
	.dc.b	$5C,$BF,$6E,$5F,$43,$CD,$FA,$2D,$8C,$9F,$0A,$32,$FC,$4F,$63,$EB
	.dc.b	$9F,$63,$EA,$6C,$64,$98,$78,$23,$2F,$FA,$4B,$E9,$DE,$27,$E7,$76
	.dc.b	$7D,$4F,$E0,$5F,$D7,$52,$B7,$87,$DC,$9D,$F3,$FE,$76,$73,$5D,$70
	.dc.b	$FD,$46,$FF,$2F,$A7,$51,$31,$FC,$2D,$75,$8D,$F1,$7F,$9D,$A7,$38
	.dc.b	$74,$CB,$F2,$FC,$0F,$A9,$FC,$B9,$F9,$8F,$0A,$DF,$E2,$22,$E5,$F4
	.dc.b	$26,$DF,$2A,$FC,$FF,$D0,$7E,$81,$A7,$E4,$0B,$5F,$61,$B3,$F1,$31
	.dc.b	$79,$BF,$C1,$7A,$1F,$52,$EE,$F8,$93,$98,$C6,$1E,$C8,$CB,$F9,$88
	.dc.b	$FA,$0F,$F1,$5B,$EA,$EC,$98,$FE,$6A,$76,$FF,$9E,$67,$6F,$EE,$07
	.dc.b	$D4,$FF,$02,$6D,$7E,$B9,$7C,$56,$66,$4E,$5E,$11,$89,$E9,$F5,$3C
	.dc.b	$FE,$33,$A3,$6F,$A3,$67,$71,$DE,$7F,$91,$7D,$33,$E9,$FF,$0B,$F5
	.dc.b	$3F,$CF,$6F,$FD,$CB,$39,$BF,$D9,$CE,$78,$3F,$27,$39,$AD,$DF,$F3
	.dc.b	$5F,$A8,$7D,$4B,$EA,$7B,$CC,$CB,$F1,$F8,$71,$C7,$FA,$56,$1C,$2F
	.dc.b	$58,$FF,$BD,$BE,$A5,$CB,$FC,$E5,$B8,$FE,$63,$97,$E4,$BF,$9A,$8C
	.dc.b	$D7,$FB,$FA,$7F,$D4,$33,$BF,$63,$FA,$17,$D3,$3E,$53,$FA,$25,$6A
	.dc.b	$DA,$7E,$C2,$B5,$87,$E9,$9B,$3A,$8F,$EB,$5F,$52,$FA,$9E,$B8,$F6
	.dc.b	$F8,$09,$5C,$7C,$9B,$87,$13,$FD,$81,$8F,$F1,$E4,$DF,$A9,$E9,$5C
	.dc.b	$4F,$DB,$B8,$9E,$47,$0A,$FE,$BD,$19,$CF,$D5,$1F,$D8,$7C,$F7,$3F
	.dc.b	$F6,$BC,$1F,$4D,$FE,$CA,$7F,$81,$CE,$9F,$E9,$26,$BD,$FE,$55,$23
	.dc.b	$34,$FA,$9C,$67,$E4,$C3,$93,$6F,$C1,$5B,$BE,$F1,$FD,$D5,$7C,$27
	.dc.b	$97,$0F,$9F,$9F,$F2,$29,$C9,$E4,$8C,$BF,$E0,$B3,$38,$7C,$46,$CF
	.dc.b	$F8,$5D,$F7,$FC,$EF,$7B,$91,$FC,$88,$D9,$FE,$F3,$36,$65,$FF,$05
	.dc.b	$7D,$43,$EA,$7D,$66,$0B,$3F,$CB,$23,$4D,$8F,$83,$F2,$5B,$8C,$13
	.dc.b	$1F,$C2,$3F,$52,$35,$1F,$7B,$E0,$3F,$8C,$BA,$D5,$F4,$DF,$E1,$D9
	.dc.b	$CE,$9E,$F9,$CE,$1F,$CF,$A7,$2F,$F8,$C5,$B7,$FE,$41,$9F,$F6,$D4
	.dc.b	$EF,$C8,$E8,$47,$EF,$F4,$DA,$FC,$8F,$9F,$F7,$B1,$FF,$22,$4B,$E9
	.dc.b	$DF,$53,$FE,$31,$AD,$60,$C3,$FD,$D1,$5A,$8F,$EA,$73,$A9,$D3,$6D
	.dc.b	$F1,$7F,$DF,$A7,$FB,$F7,$27,$52,$5F,$F8,$4A,$75,$7C,$EB,$FC,$09
	.dc.b	$F4,$1F,$DB,$FE,$A7,$12,$8C,$61,$E7,$66,$1F,$E1,$CC,$9D,$98,$C7
	.dc.b	$E0,$F0,$FB,$4F,$F2,$73,$ED,$F8,$3F,$77,$D4,$3E,$A3,$FC,$C1,$A7
	.dc.b	$D4,$AE,$39,$7C,$0C,$38,$C3,$0E,$37,$C2,$73,$18,$9F,$8B,$1F,$E7
	.dc.b	$3C,$3B,$47,$F9,$6E,$74,$FF,$88,$E7,$55,$FE,$5F,$8E,$C2,$B8,$BB
	.dc.b	$BE,$2F,$AC,$D4,$F8,$E3,$31,$CA,$78,$E1,$ED,$79,$C7,$17,$37,$E2
	.dc.b	$8F,$31,$D0,$F5,$D7,$4C,$3D,$37,$CF,$0F,$44,$62,$F9,$5F,$FA,$2E
	.dc.b	$31,$7C,$63,$A1,$EB,$29,$E3,$AF,$53,$83,$D4,$E3,$0F,$4C,$60,$F2
	.dc.b	$9B,$AF,$94,$F8,$4A,$7C,$C3,$2E,$BE,$A7,$F8,$EF,$AB,$7A,$BF,$D0
	.dc.b	$95,$F4,$E7,$2D,$8C,$BF,$E9,$58,$7C,$8F,$F5,$68,$E0,$D7,$4D,$7F
	.dc.b	$D5,$D1,$C3,$97,$C5,$3C,$71,$FB,$FF,$52,$FA,$5F,$F5,$A4,$61,$BA
	.dc.b	$72,$AE,$0D,$BF,$C6,$DF,$1C,$6D,$F0,$7F,$7D,$8F,$85,$FA,$6D,$C7
	.dc.b	$B4,$70,$3F,$2F,$19,$C3,$B4,$7D,$AB,$EC,$DB,$EC,$AB,$83,$E0,$C7
	.dc.b	$CE,$FF,$0F,$AE,$38,$78,$23,$C7,$3A,$79,$36,$77,$87,$FB,$5D,$B3
	.dc.b	$D7,$1F,$3D,$6F,$1F,$1D,$6A,$B9,$D7,$B5,$FE,$EF,$1E,$A8,$FF,$3E
	.dc.b	$B7,$5F,$7E,$3E,$59,$CF,$85,$FE,$E7,$6D,$FF,$9E,$2F,$83,$FC,$AF
	.dc.b	$EA,$7F,$CC,$0F,$F9,$F3,$EA,$7F,$CC,$AD,$F6,$39,$65,$FE,$F1,$AD
	.dc.b	$CF,$F8,$29,$B8,$7F,$A4,$2F,$8C,$FF,$83,$9B,$1F,$E9,$49,$DD,$73
	.dc.b	$F9,$AE,$B7,$E3,$6A,$94,$62,$DE,$1E,$A7,$35,$E9,$F1,$B4,$66,$E4
	.dc.b	$64,$87,$A5,$EF,$F5,$3F,$CE,$E3,$FD,$3A,$D7,$FC,$4E,$DC,$BD,$DE
	.dc.b	$6F,$0E,$BF,$27,$D4,$CB,$E3,$19,$7E,$64,$65,$E7,$1E,$EF,$9E,$7C
	.dc.b	$7D,$24,$FD,$3E,$FA,$FC,$BF,$6B,$E1,$18,$CB,$FD,$7F,$7C,$A7,$ED
	.dc.b	$4E,$9F,$F5,$BC,$65,$FF,$20,$FC,$07,$F9,$6D,$FF,$5D,$B5,$CF,$A1
	.dc.b	$F4,$BA,$3F,$CD,$71,$A8,$C7,$C4,$48,$C1,$0F,$F2,$6C,$A3,$51,$D0
	.dc.b	$FF,$64,$F4,$7D,$87,$FC,$A6,$DC,$BF,$DB,$2D,$E6,$FF,$00,$4B,$EA
	.dc.b	$5E,$5F,$E1,$AB,$53,$FE,$05,$97,$C8,$FE,$29,$B1,$E0,$E5,$E0,$66
	.dc.b	$64,$64,$FF,$6E,$FD,$4B,$CE,$FC,$9B,$63,$31,$D7,$6C,$BF,$98,$FA
	.dc.b	$8F,$F5,$34,$FF,$B9,$9A,$FF,$9C,$2B,$93,$FE,$72,$8C,$EB,$FE,$F8
	.dc.b	$FA,$9F,$F5,$EC,$FF,$BB,$9A,$EF,$DE,$ED,$E2,$FE,$9B,$EA,$5F,$15
	.dc.b	$4C,$D7,$FB,$DA,$32,$DF,$A9,$E0,$6D,$D7,$D2,$7E,$C3,$EA,$5F,$53
	.dc.b	$4E,$D1,$FB,$4F,$8B,$FF,$0D,$D7,$F8,$CF,$57,$3F,$D6,$9C,$CF,$FA
	.dc.b	$39,$B3,$A8,$FF,$1D,$7C,$8F,$F8,$D1,$E5,$5C,$7D,$2F,$E0,$32,$FA
	.dc.b	$9A,$78,$62,$57,$03,$9C,$69,$C0,$C3,$C3,$7F,$8A,$FA,$9F,$A3,$8F
	.dc.b	$8E,$91,$FE,$75,$78,$B5,$FF,$53,$BC,$27,$97,$2F,$F1,$68,$D5,$73
	.dc.b	$6C,$EA,$B9,$F8,$BB,$23,$DB,$E8,$AB,$DA,$2D,$F3,$9F,$A1,$F2,$F2
	.dc.b	$BD,$B9,$23,$C5,$F2,$59,$9F,$3F,$D4,$F9,$4C,$61,$EF,$BE,$13,$EE
	.dc.b	$9D,$3D,$2F,$F9,$8F,$D4,$3C,$5B,$C5,$2B,$E6,$CE,$5B,$39,$7F,$D9
	.dc.b	$B0,$9E,$73,$D3,$EA,$65,$72,$7B,$F8,$18,$8E,$31,$73,$FA,$D8,$7E
	.dc.b	$FA,$5F,$52,$F9,$AF,$E8,$ED,$8F,$C0,$B5,$FF,$6F,$FD,$4F,$F0,$87
	.dc.b	$33,$FD,$3A,$1F,$33,$4E,$71,$27,$F1,$73,$A6,$CE,$5E,$BE,$C7,$32
	.dc.b	$3C,$3B,$EB,$A3,$9C,$78,$63,$C5,$F0,$3E,$3E,$A9,$5E,$08,$CB,$E2
	.dc.b	$8C,$BE,$C8,$FE,$CF,$F5,$0F,$A9,$A4,$9F,$D1,$FA,$BF,$99,$7F,$D5
	.dc.b	$53,$B7,$3E,$CF,$F7,$09,$FE,$EF,$D7,$FC,$C2,$B5,$4E,$1F,$D4,$46
	.dc.b	$1F,$D5,$57,$09,$FF,$5B,$35,$4E,$EB,$F9,$DC,$BA,$FC,$1D,$72,$FC
	.dc.b	$F1,$E3,$30,$F8,$6B,$8C,$E6,$3D,$BD,$7E,$9F,$A9,$91,$82,$1C,$4F
	.dc.b	$2B,$A7,$03,$04,$FF,$84,$BE,$9D,$F5,$3F,$D4,$B1,$FE,$A7,$69,$CB
	.dc.b	$FF,$00,$4F,$EA,$EE,$1F,$E1,$C6,$E7,$93,$A7,$FD,$25,$D9,$FD,$7B
	.dc.b	$EA,$1F,$53,$F9,$7F,$84,$FD,$BD,$A7,$3E,$9F,$FB,$1E,$F9,$4D,$3F
	.dc.b	$DC,$D3,$A7,$FD,$89,$1D,$6F,$A3,$6B,$84,$F9,$4B,$5C,$5F,$7B,$5F
	.dc.b	$37,$84,$F8,$9F,$F4,$DC,$E6,$7F,$DE,$DF,$4E,$FA,$9F,$E8,$1B,$7C
	.dc.b	$88,$C3,$F1,$E7,$73,$A7,$FB,$FC,$E9,$F9,$2D,$D3,$E1,$B7,$3D,$FA
	.dc.b	$9F,$D3,$67,$37,$67,$38,$7B,$AF,$86,$1E,$6C,$34,$F8,$A7,$51,$88
	.dc.b	$4F,$06,$C6,$63,$07,$F8,$42,$7B,$5E,$67,$4D,$8C,$9D,$8B,$5F,$58
	.dc.b	$F2,$38,$C3,$DF,$5C,$A3,$84,$EC,$F2,$D7,$57,$17,$E4,$AE,$11,$86
	.dc.b	$DF,$4B,$ED,$87,$C6,$AE,$38,$73,$AF,$97,$5E,$06,$CF,$38,$E2,$FA
	.dc.b	$0E,$2E,$EB,$ED,$D7,$36,$C6,$67,$E8,$1C,$CE,$25,$AE,$E7,$9D,$CB
	.dc.b	$4C,$CF,$A5,$B1,$E3,$6B,$F6,$27,$CE,$D3,$9C,$F1,$3C,$AF,$85,$B3
	.dc.b	$DD,$A6,$E3,$D3,$19,$6C,$7C,$B9,$FD,$05,$FE,$76,$7E,$34,$60,$FD
	.dc.b	$13,$E3,$70,$FE,$82,$38,$39,$3C,$07,$A6,$BF,$47,$EB,$7C,$D8,$F8
	.dc.b	$17,$AC,$53,$99,$E1,$39,$8C,$4F,$E6,$67,$E0,$1F,$BA,$8C,$DE,$9B
	.dc.b	$87,$D9,$9F,$D2,$C6,$1E,$26,$23,$55,$D7,$0F,$8D,$39,$37,$19,$C3
	.dc.b	$C9,$5C,$30,$F0,$34,$E7,$87,$7C,$39,$DF,$BA,$75,$87,$4A,$D6,$1D
	.dc.b	$23,$BE,$1D,$1B,$5D,$0E,$38,$74,$BE,$33,$F3,$0F,$89,$3B,$7B,$35
	.dc.b	$FE,$3D,$1F,$D0,$1D,$9F,$D2,$0F,$80,$7F,$4A,$FA,$9F,$F1,$D4,$E6
	.dc.b	$11,$87,$83,$5E,$18,$75,$76,$FF,$9B,$7A,$3F,$CB,$93,$C5,$FF,$2F
	.dc.b	$65,$6F,$FD,$57,$36,$FF,$9B,$DF,$19,$CB,$63,$31,$FE,$6C,$9E,$11
	.dc.b	$FC,$92,$7C,$55,$E5,$8F,$E7,$31,$83,$E1,$7C,$9B,$E2,$7F,$D2,$F8
	.dc.b	$7F,$9E,$63,$1F,$1F,$DD,$D7,$A9,$69,$36,$30,$DA,$E5,$3A,$6F,$D4
	.dc.b	$F1,$FA,$9E,$4D,$8C,$9C,$23,$2D,$3F,$D1,$3F,$51,$FE,$D9,$7F,$D1
	.dc.b	$DC,$8F,$A9,$F8,$33,$9D,$7F,$C1,$93,$9D,$7E,$44,$27,$5C,$1F,$E8
	.dc.b	$F6,$E0,$97,$C8,$3E,$0B,$F3,$51,$FE,$14,$AD,$BB,$6B,$D7,$E0,$7E
	.dc.b	$B5,$71,$E4,$F0,$D7,$0E,$2F,$A1,$E6,$7C,$BB,$F9,$8D,$CB,$E4,$B7
	.dc.b	$1F,$F1,$35,$70,$9F,$9B,$5F,$4F,$D7,$91,$F4,$3E,$A6,$DF,$0E,$35
	.dc.b	$63,$6F,$F6,$06,$3F,$81,$BC,$61,$FE,$FC,$FA,$6F,$F1,$7B,$FE,$BB
	.dc.b	$A7,$D4,$FE,$1B,$2F,$F1,$AD,$69,$FF,$1D,$4E,$B9,$FE,$7C,$E6,$3F
	.dc.b	$D7,$8D,$BF,$9D,$C1,$F3,$7C,$FD,$F8,$92,$7F,$D8,$F5,$BC,$7E,$9F
	.dc.b	$1F,$FA,$12,$73,$5F,$93,$FA,$99,$1F,$90,$F0,$6D,$FC,$19,$B9,$FC
	.dc.b	$0E,$1F,$F6,$95,$FE,$5A,$7F,$38,$FE,$77,$E8,$6B,$F9,$A6,$DF,$CB
	.dc.b	$6D,$F3,$7D,$FE,$67,$ED,$7A,$5F,$53,$FA,$1B,$F8,$F3,$98,$C5,$FC
	.dc.b	$48,$E3,$87,$E7,$27,$E7,$4E,$63,$E1,$CE,$AB,$F3,$B5,$A9,$FB,$13
	.dc.b	$FB,$79,$7D,$4D,$31,$F1,$56,$BF,$7D,$3A,$6C,$E9,$A6,$BE,$A7,$A7
	.dc.b	$C6,$75,$6C,$64,$E7,$19,$69,$FF,$00,$3E,$9D,$F0,$FF,$A4,$FA,$97
	.dc.b	$C7,$FE,$D2,$73,$5F,$B1,$D8,$FE,$1A,$B6,$F3,$BE,$07,$F2,$3F,$A8
	.dc.b	$7D,$4B,$EA,$73,$5B,$26,$DF,$E8,$FA,$DC,$EA,$3F,$D1,$FF,$52,$E4
	.dc.b	$FE,$0E,$DB,$F2,$1F,$C1,$67,$FC,$DD,$C3,$AA,$FE,$1B,$C9,$FE,$A9
	.dc.b	$AD,$D7,$F0,$B8,$C3,$FE,$9C,$FA,$77,$C9,$70,$F6,$BF,$8A,$C6,$2F
	.dc.b	$8D,$23,$0F,$FA,$93,$EA,$5F,$53,$C3,$0E,$3F,$01,$30,$E3,$87,$F1
	.dc.b	$8E,$3F,$97,$8F,$F1,$D8,$57,$83,$CB,$97,$E0,$FA,$9E,$AF,$28,$FE
	.dc.b	$53,$18,$BF,$54,$63,$2E,$31,$FE,$B8,$F2,$1D,$3F,$D5,$F0,$7D,$37
	.dc.b	$FA,$ED,$F6,$CE,$9F,$8E,$6B,$DD,$FE,$6F,$19,$BE,$97,$CE,$DF,$53
	.dc.b	$A8,$98,$74,$75,$18,$AF,$F6,$5C,$BE,$13,$C9,$3F,$33,$7F,$8B,$5F
	.dc.b	$90,$9C,$F0,$7F,$48,$BE,$07,$D4,$2C,$FF,$B6,$5F,$6F,$CF,$6E,$A4
	.dc.b	$6D,$F4,$46,$27,$FD,$BB,$F5,$2F,$A9,$F6,$1F,$D5,$9C,$5F,$C4,$D2
	.dc.b	$D7,$0A,$D1,$F0,$FE,$4A,$D7,$E2,$2F,$79,$FB,$1B,$1F,$D7,$4F,$A7
	.dc.b	$3C,$2F,$FD,$DF,$F0,$1F,$C5,$27,$58,$BE,$9B,$FB,$4C,$E6,$BF,$6B
	.dc.b	$39,$AF,$CB,$CE,$64,$EE,$3F,$03,$D3,$CB,$DB,$91,$C1,$FB,$91,$8A
	.dc.b	$FF,$80,$3E,$87,$E0,$74,$FC,$CC,$BE,$9D,$F5,$3F,$DF,$2B,$58,$FE
	.dc.b	$BE,$72,$FE,$AA,$33,$87,$F5,$0E,$4F,$E6,$DF,$17,$FD,$E2,$73,$3F
	.dc.b	$E0,$09,$E8,$6B,$D8,$B7,$CF,$E8,$3F,$C6,$5A,$3E,$A7,$8E,$29,$CC
	.dc.b	$61,$F3,$1F,$D1,$73,$FD,$8D,$F3,$3C,$DC,$3E,$B3,$FC,$2F,$9F,$F8
	.dc.b	$4F,$E0,$FF,$C2,$9F,$4E,$FA,$8F,$E2,$59,$FD,$1A,$F8,$F7,$C7,$0F
	.dc.b	$E1,$31,$C4,$5F,$17,$FB,$8C,$F6,$58,$76,$C3,$AC,$FF,$08,$AE,$15
	.dc.b	$FC,$86,$B4,$F2,$BA,$E4,$EF,$7C,$A3,$31,$E3,$AE,$78,$7A,$5E,$B1
	.dc.b	$9A,$E0,$F2,$3F,$98,$1E,$63,$07,$F8,$DB,$0F,$34,$7F,$9E,$F0,$F1
	.dc.b	$DF,$58,$C5,$70,$8C,$3D,$27,$75,$CA,$30,$6D,$C3,$ED,$BF,$4C,$62
	.dc.b	$3D,$07,$C3,$3C,$A6,$0F,$36,$1F,$09,$F5,$4F,$28,$C5,$7F,$A3,$E7
	.dc.b	$EC,$1F,$E9,$17,$0F,$C1,$3E,$1F,$D4,$FF,$1B,$F5,$7F,$D2,$CD,$8F
	.dc.b	$F4,$E3,$5D,$F5,$FF,$9F,$DC,$FC,$8B,$83,$8B,$63,$31,$82,$3F,$43
	.dc.b	$EA,$5F,$4B,$FD,$8F,$4F,$C2,$C6,$27,$F0,$78,$B0,$E9,$5F,$8A,$8C
	.dc.b	$61,$F8,$8B,$E1,$7D,$70,$E9,$BF,$F6,$FB,$DC,$FE,$01,$B8,$78,$2B
	.dc.b	$73,$FE,$79,$BE,$73,$DF,$1F,$0E,$1C,$DB,$5B,$BE,$2A,$F8,$D6,$F1
	.dc.b	$F5,$DF,$0A,$D5,$F0,$BD,$C7,$FA,$27,$1F,$2E,$1C,$63,$D8,$7B,$EB
	.dc.b	$BC,$EA,$38,$5F,$5B,$E7,$1D,$8F,$0E,$1C,$5F,$F0,$2C,$7C,$88,$FE
	.dc.b	$B2,$DD,$7C,$19,$7C,$5A,$DF,$89,$FE,$19,$BD,$B6,$BF,$C3,$97,$C6
	.dc.b	$BF,$D4,$5F,$51,$FE,$BE,$7F,$C2,$7F,$52,$F6,$38,$47,$FA,$9E,$30
	.dc.b	$FF,$86,$27,$83,$F8,$DB,$FF,$14,$46,$31,$FE,$DF,$5C,$2B,$47,$3B
	.dc.b	$EA,$AB,$85,$F2,$F9,$AC,$C1,$A7,$0B,$78,$78,$BF,$4F,$8F,$93,$0F
	.dc.b	$80,$ED,$59,$76,$7A,$DF,$1F,$D4,$FF,$3D,$8F,$F1,$AB,$5E,$4D,$CF
	.dc.b	$DB,$E7,$70,$EA,$F2,$7D,$4C,$BE,$31,$97,$E6,$C6,$5E,$71,$F0,$BE
	.dc.b	$77,$FC,$CF,$1F,$E4,$46,$BE,$96,$DF,$CC,$EF,$E1,$ED,$7C,$3F,$4E
	.dc.b	$5F,$18,$CB,$F7,$23,$2F,$38,$FA,$3F,$03,$FE,$56,$8F,$F2,$9B,$5F
	.dc.b	$90,$DB,$FF,$14,$F4,$FF,$9A,$67,$7E,$67,$DA,$87,$87,$4E,$B1,$27
	.dc.b	$FC,$BF,$3A,$6C,$E5,$FA,$DD,$1F,$7D,$FF,$32,$B5,$CF,$0F,$E2,$3E
	.dc.b	$A7,$FC,$23,$1A,$8C,$7C,$8C,$8C,$10,$FC,$14,$A3,$51,$D0,$FF,$39
	.dc.b	$FD,$4B,$CD,$FC,$AB,$63,$31,$E1,$CF,$2F,$E6,$BE,$A7,$FD,$43,$3F
	.dc.b	$E7,$C6,$BF,$91,$AE,$4F,$E4,$E3,$3A,$FF,$95,$3E,$A5,$C3,$FA,$46
	.dc.b	$D7,$3F,$7B,$B7,$89,$FA,$8F,$A9,$7C,$55,$33,$5F,$E8,$C8,$CB,$7E
	.dc.b	$A7,$81,$B7,$5F,$49,$FB,$2F,$A9,$7D,$4D,$3B,$47,$ED,$BE,$2F,$FD
	.dc.b	$17,$5F,$E9,$86,$BF,$AD,$6E,$DF,$E6,$EF,$91,$7A,$7C,$7D,$FF,$DF
	.dc.b	$C3,$EA,$69,$9C,$3A,$4E,$1F,$F5,$2C,$ED,$C9,$FE,$A8,$FA,$9F,$97
	.dc.b	$87,$F9,$F1,$FF,$37,$46,$7D,$DF,$E1,$B1,$AF,$2F,$FD,$78,$D9,$07
	.dc.b	$8B,$FC,$1D,$B5,$D7,$A3,$9F,$8D,$2B,$C4,$F6,$FA,$2C,$47,$18,$B5
	.dc.b	$E5,$9F,$95,$F2,$F2,$BE,$53,$E6,$F9,$1C,$47,$87,$EA,$7D,$26,$2F
	.dc.b	$FD,$33,$C2,$BC,$D7,$C3,$AD,$69,$FE,$D9,$FA,$87,$93,$FC,$B9,$72
	.dc.b	$53,$11,$2F,$FD,$A7,$3A,$6C,$E5,$FE,$55,$0F,$A9,$E8,$48,$E3,$E2
	.dc.b	$57,$71,$74,$FF,$58,$D7,$EE,$DF,$F5,$9F,$D4,$BE,$6B,$FA,$1B,$6F
	.dc.b	$FD,$CA,$D7,$F9,$D7,$D4,$FF,$00,$7F,$9E,$46,$6B,$FD,$7B,$2F,$99
	.dc.b	$FE,$9A,$DA,$E7,$7D,$8C,$CC,$8C,$9F,$EF,$49,$FF,$0E,$37,$D9,$E6
	.dc.b	$74,$F0,$6C,$64,$87,$99,$F0,$F6,$78,$BA,$E4,$F5,$8C,$BD,$E3,$2F
	.dc.b	$BA,$3F,$B2,$FD,$43,$EA,$69,$27,$F3,$3A,$7F,$B7,$3D,$5F,$CE,$42
	.dc.b	$B8,$3A,$E4,$FC,$FA,$9F,$EE,$95,$C7,$D9,$E9,$3F,$D9,$EF,$8B,$FD
	.dc.b	$A6,$B7,$7C,$9B,$1F,$AA,$9D,$DF,$8F,$AF,$FB,$ED,$3A,$E9,$88,$97
	.dc.b	$E5,$3F,$67,$3B,$70,$79,$E7,$6F,$8F,$AF,$AF,$D4,$C8,$DD,$27,$C2
	.dc.b	$66,$3D,$53,$87,$AF,$D4,$FF,$B3,$9F,$F7,$41,$F1,$9C,$2C,$FF,$D2
	.dc.b	$B1,$86,$FC,$37,$CC,$6C,$FF,$AE,$FB,$79,$78,$7F,$EA,$B6,$C6,$4F
	.dc.b	$E2,$37,$DB,$E0,$BF,$79,$18,$AE,$0D,$38,$09,$FE,$48,$FF,$AF,$2B
	.dc.b	$D1,$87,$08,$D5,$79,$5F,$09,$EB,$2E,$1E,$C8,$D4,$7F,$24,$FA,$77
	.dc.b	$D4,$7E,$BE,$3F,$13,$18,$F8,$BF,$F5,$BC,$6F,$C3,$7C,$56,$21,$5B
	.dc.b	$76,$A7,$71,$E9,$39,$4F,$BB,$E9,$BF,$A4,$E5,$F0,$2F,$84,$E7,$A3
	.dc.b	$D5,$BF,$C8,$C3,$BE,$3D,$23,$18,$F2,$9E,$C6,$AF,$96,$3D,$6F,$86
	.dc.b	$DF,$A6,$BE,$18,$75,$CB,$C3,$8F,$29,$CD,$7C,$4B,$E3,$3F,$D3,$27
	.dc.b	$2F,$B0,$E1,$97,$A2,$B8,$65,$E2,$9F,$2E,$1D,$B1,$F4,$DE,$E7,$94
	.dc.b	$63,$2F,$45,$F1,$8F,$36,$5E,$2A,$65,$E2,$7E,$01,$E2,$BE,$2D,$3C
	.dc.b	$33,$C6,$78,$63,$E3,$CB,$E2,$65,$E3,$BE,$78,$F8,$DE,$39,$78,$EB
	.dc.b	$DB,$1D,$67,$CB,$39,$8F,$5C,$F4,$CB,$C7,$5C,$CD,$CF,$19,$E7,$19
	.dc.b	$69,$C6,$3C,$26,$CE,$A7,$D3,$BF,$46,$1E,$B8,$C5,$F8,$2B,$95,$F8
	.dc.b	$0E,$D3,$C8,$F9,$77,$CB,$2F,$8C,$DC,$7D,$93,$F1,$2F,$99,$E6,$7B
	.dc.b	$9E,$6B,$EA,$EC,$CE,$1F,$CD,$9E,$EF,$F1,$47,$27,$BD,$D6,$5F,$9D
	.dc.b	$6E,$1F,$BD,$69,$C6,$79,$9F,$C0,$4F,$1D,$78,$A9,$F8,$4F,$DD,$7F
	.dc.b	$10,$4B,$C4,$65,$97,$F0,$27,$8B,$63,$37,$FA,$7A,$B7,$DE,$79,$D6
	.dc.b	$CE,$53,$AC,$3E,$4D,$F9,$9F,$DA,$46,$63,$04,$3F,$3E,$E5,$F7,$DF
	.dc.b	$B6,$31,$87,$AE,$FF,$87,$46,$30,$CB,$63,$C6,$79,$6F,$F1,$87,$67
	.dc.b	$17,$D5,$B8,$7A,$27,$47,$0C,$3C,$0F,$C3,$9D,$BE,$98,$F9,$2E,$F0
	.dc.b	$F1,$CF,$8A,$39,$61,$87,$ED,$1D,$A3,$F0,$71,$F8,$18,$F8,$F1,$F8
	.dc.b	$1A,$7B,$1C,$27,$F8,$B3,$5F,$E0,$ED,$8D,$CF,$F3,$B6,$CF,$FB,$AD
	.dc.b	$BC,$FF,$8B,$5E,$E7,$F4,$D8,$C5,$DB,$FE,$59,$36,$FC,$31,$FD,$FA
	.dc.b	$F8,$B6,$B5,$3F,$C6,$AB,$C8,$49,$F2,$3E,$BF,$91,$7C,$4F,$FA,$92
	.dc.b	$7F,$C1,$91,$8F,$8E,$F0,$3A,$F5,$2D,$26,$C6,$1B,$5C,$A7,$4D,$FA
	.dc.b	$7E,$3F,$4F,$C9,$B1,$93,$84,$65,$A7,$F8,$73,$E9,$BF,$DC,$EF,$F8
	.dc.b	$87,$8D,$F4,$FF,$E7,$D8,$D7,$C0,$FC,$79,$CF,$27,$C4,$BE,$3B,$B8
	.dc.b	$BF,$79,$18,$7F,$BF,$7C,$83,$E1,$3E,$BE,$5F,$E4,$2A,$D3,$72,$E8
	.dc.b	$7D,$5A,$E1,$C7,$89,$F8,$5E,$67,$C8,$AF,$92,$D8,$C9,$F1,$E3,$2D
	.dc.b	$3E,$57,$AF,$F4,$0F,$A3,$F4,$F6,$32,$4D,$3D,$8F,$FA,$DB,$97,$EE
	.dc.b	$CE,$E7,$27,$FA,$EB,$E9,$BF,$C8,$15,$FE,$4B,$D9,$F4,$FF,$83,$AD
	.dc.b	$8A,$DB,$CA,$77,$CF,$F9,$09,$CC,$7F,$27,$6D,$FF,$B2,$36,$F9,$9E
	.dc.b	$86,$FC,$69,$3F,$E5,$8A,$DE,$2B,$DC,$E6,$BF,$23,$F4,$F2,$3F,$1D
	.dc.b	$E0,$5B,$F6,$ED,$F9,$CA,$C6,$7F,$E3,$98,$FC,$FD,$62,$FE,$37,$57
	.dc.b	$F9,$5D,$FC,$C7,$CF,$7E,$79,$FC,$FB,$5C,$9F,$9D,$32,$7A,$7C,$DF
	.dc.b	$C8,$7D,$3F,$E8,$33,$F9,$3B,$27,$3D,$DF,$1D,$97,$F9,$BB,$FA,$1A
	.dc.b	$7C,$99,$CE,$1C,$9E,$BC,$38,$9F,$64,$BE,$9F,$F6,$ED,$C3,$F7,$6B
	.dc.b	$E1,$73,$32,$72,$F0,$8C,$4F,$CA,$1F,$4F,$D3,$E3,$7A,$B6,$FA,$B6
	.dc.b	$78,$37,$0F,$ED,$1F,$4E,$F8,$7F,$D3,$7D,$3F,$F1,$8B,$FC,$6C,$E7
	.dc.b	$0F,$F4,$E7,$3A,$B5,$87,$2C,$3F,$8E,$DE,$F0,$FC,$DF,$D3,$3E,$9D
	.dc.b	$F4,$FA,$B9,$9F,$FA,$93,$0E,$18,$FF,$9E,$F0,$E1,$7A,$C7,$F3,$FF
	.dc.b	$4E,$E6,$FE,$56,$DC,$7F,$D5,$BE,$37,$CB,$7F,$D5,$F1,$9A,$FD,$3C
	.dc.b	$BD,$8F,$E6,$6D,$AF,$F5,$AF,$D3,$3E,$47,$35,$FC,$E2,$31,$3A,$9F
	.dc.b	$E7,$7F,$4E,$FA,$6D,$46,$FD,$F9,$5D,$3C,$9B,$9F,$EE,$31,$FE,$32
	.dc.b	$83,$85,$F4,$FD,$6B,$91,$FE,$CA,$71,$1C,$EB,$FA,$6C,$63,$87,$FC
	.dc.b	$49,$7C,$FC,$27,$4F,$F5,$7C,$5F,$4B,$FE,$E9,$7F,$DA,$B3,$A7,$FD
	.dc.b	$B4,$6B,$DE,$FE,$B3,$19,$CF,$F8,$8B,$71,$FF,$1B,$5B,$E9,$F7,$8C
	.dc.b	$C7,$AB,$E1,$9D,$46,$0F,$F7,$28,$F8,$3F,$35,$7D,$0C,$7F,$8F,$CE
	.dc.b	$4F,$2C,$67,$0F,$F2,$14,$66,$77,$5D,$FE,$77,$FB,$8C,$E6,$B7,$38
	.dc.b	$35,$1E,$27,$FB,$BF,$D3,$BE,$47,$D0,$60,$B3,$FE,$FC,$8D,$36,$3D
	.dc.b	$DF,$21,$B8,$C1,$31,$FE,$7E,$FD,$28,$D4,$7F,$3C,$F7,$FF,$8B,$63
	.dc.b	$F3,$3E,$97,$FC,$13,$18,$BF,$C4,$1F,$6A,$33,$E4,$C8,$FB,$3B,$7F
	.dc.b	$B6,$9F,$C7,$5F,$F9,$9E,$F9,$61,$F9,$16,$A7,$33,$F8,$6E,$AE,$93
	.dc.b	$FD,$43,$C3,$FC,$04,$FE,$66,$5F,$4C,$F8,$4F,$F0,$FD,$6A,$FF,$DD
	.dc.b	$73,$9B,$F1,$9F,$A1,$8F,$E9,$9C,$7C,$0F,$F0,$37,$C0,$7F,$7C,$8F
	.dc.b	$EA,$D1,$99,$E3,$18,$7F,$C2,$13,$B8,$E7,$1A,$7F,$AC,$7C,$F7,$F8
	.dc.b	$DF,$47,$D3,$F9,$67,$35,$FD,$BE,$31,$92,$FD,$67,$F8,$67,$27,$66
	.dc.b	$2F,$B5,$38,$7E,$09,$FE,$40,$C7,$F7,$DF,$07,$FB,$FF,$A5,$7D,$3F
	.dc.b	$C1,$BE,$18,$72,$C3,$8B,$E9,$C3,$88,$BE,$2F,$F7,$99,$EC,$B0,$ED
	.dc.b	$87,$59,$5E,$2B,$FD,$69,$55,$FE,$F9,$3A,$7B,$57,$FA,$E5,$E8,$60
	.dc.b	$F0,$57,$17,$CB,$7E,$47,$9C,$65,$E3,$19,$C3,$DF,$1C,$CF,$C1,$61
	.dc.b	$E3,$8C,$CF,$39,$EB,$7E,$BB,$E9,$7B,$C3,$D1,$7B,$BE,$F7,$B7,$B1
	.dc.b	$C8,$F4,$98,$36,$62,$39,$57,$5B,$EF,$3E,$7C,$3C,$B7,$F1,$6B,$53
	.dc.b	$CD,$F9,$0F,$B0,$A7,$CA,$32,$EB,$E9,$FF,$90,$7A,$97,$AB,$FD,$03
	.dc.b	$5F,$46,$72,$D8,$CB,$FE,$ED,$87,$C7,$FF,$9A,$63,$83,$5D,$35,$FD
	.dc.b	$84,$70,$E1,$F1,$4F,$1D,$3E,$D7,$D3,$BE,$8F,$F6,$D1,$86,$E9,$CA
	.dc.b	$B8,$36,$FF,$19,$7C,$71,$B7,$C1,$FE,$C1,$8F,$85,$FC,$13,$71,$ED
	.dc.b	$1C,$0F,$CB,$46,$70,$ED,$1F,$89,$BE,$CD,$BE,$CA,$B8,$3E,$0C,$7C
	.dc.b	$EF,$F7,$1A,$E3,$87,$82,$3C,$73,$A7,$93,$67,$78,$7F,$19,$6C,$F5
	.dc.b	$C7,$CF,$5B,$C7,$C7,$5A,$AE,$75,$ED,$7F,$CF,$71,$EA,$8F,$D9,$42
	.dc.b	$31,$B7,$C0,$C7,$CD,$5A,$9F,$F7,$1D,$7F,$1F,$6B,$D2,$FF,$97,$4F
	.dc.b	$4F,$A7,$FF,$63,$4F,$F2,$5F,$A7,$FF,$31,$37,$D7,$DE,$5F,$CD,$6B
	.dc.b	$73,$FE,$90,$6D,$F9,$AF,$8C,$FF,$A5,$1B,$1D,$FB,$79,$CF,$F8,$7E
	.dc.b	$9F,$31,$88,$46,$1F,$E7,$0D,$7A,$78,$5B,$8C,$5F,$87,$C5,$E5,$89
	.dc.b	$D6,$13,$0F,$74,$68,$C3,$FE,$AA,$FA,$7F,$E7,$CF,$F4,$76,$B9,$EA
	.dc.b	$EB,$E6,$F9,$E7,$38,$FC,$1F,$A7,$F0,$99,$A9,$7F,$36,$32,$F5,$8F
	.dc.b	$81,$F3,$8F,$8D,$A4,$9F,$A9,$E0,$D7,$E6,$7B,$5E,$E3,$18,$FE,$1A
	.dc.b	$F9,$4F,$DB,$9D,$3F,$D8,$63,$2F,$FB,$23,$E2,$7F,$CB,$CF,$F6,$56
	.dc.b	$B9,$F4,$3E,$A7,$37,$F9,$BE,$35,$18,$F8,$79,$18,$21,$FE,$D4,$94
	.dc.b	$6A,$3A,$1F,$DC,$FD,$1F,$B3,$CF,$FE,$01,$6C,$7F,$98,$DA,$BE,$9F
	.dc.b	$FC,$28,$F1,$E4,$FC,$44,$3E,$9E,$9F,$8B,$AE,$27,$28,$D5,$7F,$80
	.dc.b	$DA,$FF,$BA,$A3,$BF,$D3,$BC,$EF,$CA,$36,$7F,$C1,$4D,$7F,$DE,$35
	.dc.b	$F2,$38,$3F,$6F,$39,$7F,$73,$F4,$FF,$EA,$DA,$FF,$09,$35,$FF,$7C
	.dc.b	$CE,$F3,$FF,$11,$4E,$6B,$F8,$0F,$D3,$FF,$B1,$6F,$FC,$34,$D7,$FE
	.dc.b	$00,$CE,$EB,$8F,$89,$FA,$BF,$A7,$7C,$25,$3F,$5D,$5F,$E6,$DB,$FF
	.dc.b	$12,$4E,$9B,$F4,$FC,$4E,$0E,$BE,$8F,$F6,$DF,$4E,$FA,$7A,$77,$8F
	.dc.b	$DD,$7C,$5F,$F8,$92,$BF,$C6,$0D,$5B,$7F,$1C,$8C,$3F,$C7,$BE,$45
	.dc.b	$EA,$72,$F4,$BF,$82,$43,$E7,$29,$9C,$3A,$CE,$1F,$F1,$F4,$ED,$C9
	.dc.b	$FE,$41,$FA,$7F,$9C,$FF,$AB,$DF,$F3,$E3,$9F,$6F,$F8,$94,$6A,$0D
	.dc.b	$36,$FA,$B8,$FF,$CB,$11,$8C,$BF,$9B,$BB,$F1,$B9,$92,$BC,$5F,$41
	.dc.b	$5E,$47,$8E,$31,$5C,$A7,$E5,$FC,$BC,$AF,$96,$F9,$BE,$4B,$11,$C8
	.dc.b	$F3,$FD,$3F,$A8,$CD,$FC,$FA,$C4,$FC,$39,$CE,$5F,$E7,$1F,$A6,$78
	.dc.b	$BF,$CC,$D7,$1D,$31,$12,$FF,$CD,$93,$A6,$CE,$5F,$24,$3E,$9F,$A1
	.dc.b	$23,$8F,$85,$5D,$C5,$CF,$FB,$0D,$7F,$02,$FA,$77,$CD,$7F,$47,$6D
	.dc.b	$FF,$9F,$5A,$FC,$3F,$A7,$FE,$0E,$FC,$58,$CD,$FF,$6C,$97,$CC,$FF
	.dc.b	$51,$6D,$73,$BE,$C6,$66,$46,$4F,$F4,$64,$FF,$8C,$1B,$EC,$F3,$3A
	.dc.b	$78,$36,$32,$43,$CC,$F8,$7A,$BC,$5D,$32,$7A,$C6,$5E,$F1,$97,$DD
	.dc.b	$1F,$DA,$7E,$99,$F4,$F4,$93,$FE,$A1,$F5,$3F,$39,$3A,$8C,$7B,$3F
	.dc.b	$DC,$67,$FB,$BF,$2F,$FA,$CE,$B5,$5F,$8D,$6E,$FF,$E1,$4A,$DC,$8F
	.dc.b	$EE,$17,$E0,$7F,$29,$0E,$AE,$DD,$52,$7C,$91,$E1,$30,$F0,$8C,$3E
	.dc.b	$9E,$AF,$37,$D3,$C8,$C1,$0E,$EE,$5F,$0C,$7F,$83,$7E,$9F,$FD,$A9
	.dc.b	$1F,$EB,$2E,$CF,$F6,$84,$62,$3F,$D5,$11,$8F,$86,$FF,$49,$BF,$ED
	.dc.b	$5F,$AE,$FE,$81,$F0,$68,$D5,$73,$3E,$15,$FF,$11,$79,$98,$F0,$F9
	.dc.b	$BA,$C2,$A9,$E1,$7D,$63,$81,$D4,$D5,$70,$9F,$44,$EE,$72,$6E,$39
	.dc.b	$DF,$BA,$B8,$5F,$83,$E9,$DF,$51,$FD,$B3,$1E,$51,$8F,$8B,$F0,$37
	.dc.b	$2F,$59,$E7,$FA,$BE,$57,$FA,$E8,$FF,$7D,$4E,$DF,$F7,$E0,$5E,$43
	.dc.b	$E3,$3F,$D4,$B1,$B9,$FC,$50,$FA,$8F,$E8,$7D,$1E,$2B,$E1,$3E,$29
	.dc.b	$CB,$E4,$AE,$B7,$F1,$E7,$39,$FB,$2F,$8B,$F1,$E3,$3B,$7A,$27,$2F
	.dc.b	$8A,$33,$87,$E7,$70,$EA,$F8,$7A,$BF,$DC,$D8,$75,$C7,$C1,$18,$C3
	.dc.b	$A4,$63,$2F,$53,$6F,$D1,$97,$81,$FD,$56,$5E,$07,$E4,$46,$5F,$89
	.dc.b	$87,$57,$C9,$19,$9E,$19,$79,$B2,$F7,$5E,$EF,$9E,$1E,$19,$F4,$47
	.dc.b	$AF,$1F,$2B,$4F,$2C,$EF,$2F,$5C,$66,$B8,$9D,$AB,$8B,$63,$89,$B9
	.dc.b	$E3,$19,$38,$BF,$3E,$FC,$D8,$7A,$63,$17,$D1,$B7,$D2,$F9,$47,$43
	.dc.b	$2F,$48,$F9,$77,$D3,$2F,$8C,$DC,$7D,$93,$F2,$9A,$FC,$63,$D0,$64
	.dc.b	$A6,$6B,$C2,$E5,$AE,$6F,$C6,$79,$5F,$1E,$1E,$B6,$9E,$B8,$DB,$CD
	.dc.b	$AF,$33,$73,$ED,$7F,$07,$3C,$DA,$73,$37,$3E,$17,$33,$C7,$2F,$C3
	.dc.b	$E3,$F1,$2F,$94,$63,$0F,$87,$97,$D3,$C3,$99,$E5,$29,$E5,$3A,$E3
	.dc.b	$DE,$B9,$47,$C8,$9E,$4E,$8E,$45,$39,$47,$0C,$7E,$5C,$F0,$8C,$57
	.dc.b	$C4,$8C,$65,$F6,$EF,$B6,$5E,$B7,$E1,$1D,$67,$79,$7B,$AB,$86,$3D
	.dc.b	$AF,$83,$F1,$63,$38,$77,$9C,$BE,$D8,$CE,$1F,$0B,$0E,$AF,$83,$1F
	.dc.b	$3B,$9B,$FD,$B7,$D4,$FE,$CE,$7F,$7C,$1C,$5E,$5B,$97,$F1,$B6,$DF
	.dc.b	$F2,$1F,$5B,$97,$AF,$CD,$B5,$FC,$85,$3F,$DD,$7A,$ED,$7F,$0E,$9B
	.dc.b	$18,$8F,$E6,$1E,$F6,$2B,$F9,$15,$59,$FE,$2B,$35,$FE,$6D,$17,$2F
	.dc.b	$F6,$D5,$77,$38,$96,$78,$B5,$E2,$D8,$E3,$1F,$E5,$B5,$3E,$1F,$92
	.dc.b	$7C,$5F,$FA,$72,$FA,$46,$3E,$1F,$53,$9C,$3F,$CD,$B3,$9A,$DC,$E7
	.dc.b	$2F,$87,$39,$AD,$CE,$6B,$FC,$39,$39,$AD,$CB,$EA,$7B,$FD,$4F,$81
	.dc.b	$B7,$D9,$B8,$74,$6D,$FC,$FF,$A8,$FF,$72,$D7,$C5,$E5,$7D,$4F,$F8
	.dc.b	$36,$3E,$34,$EB,$0F,$F4,$3D,$6B,$97,$E9,$4E,$70,$E9,$F0,$1F,$5F
	.dc.b	$E4,$1F,$09,$F2,$B2,$E7,$3C,$08,$FD,$59,$DF,$0E,$33,$F8,$7E,$77
	.dc.b	$CC,$BF,$9A,$DB,$F9,$6D,$9D,$C7,$CE,$9F,$A7,$EC,$48,$FA,$3F,$53
	.dc.b	$6B,$E7,$CD,$CF,$E6,$CD,$AF,$F0,$1F,$D3,$BC,$5F,$C9,$53,$EA,$7E
	.dc.b	$B6,$3F,$E5,$1A,$D5,$FE,$3E,$B5,$F0,$3F,$97,$9C,$D7,$FA,$9D,$B3
	.dc.b	$FA,$4D,$BE,$3F,$A7,$BC,$A6,$3F,$97,$BD,$BF,$E5,$FA,$DB,$99,$FC
	.dc.b	$8F,$D4,$C8,$FC,$7F,$81,$72,$FC,$6C,$DC,$BF,$17,$36,$BF,$1F,$D3
	.dc.b	$FB,$87,$F2,$FF,$41,$F9,$C7,$F3,$AD,$72,$7E,$70,$C9,$E7,$F3,$FF
	.dc.b	$71,$D0,$FA,$97,$77,$C3,$9C,$C6,$2B,$81,$34,$FC,$7C,$7C,$E5,$BA
	.dc.b	$B9,$56,$F4,$FE,$1B,$8F,$7E,$1C,$E1,$FB,$21,$F5,$3F,$BB,$6D,$FE
	.dc.b	$E9,$7C,$66,$66,$4E,$5E,$11,$89,$E9,$F5,$3C,$FE,$3B,$A3,$6F,$A3
	.dc.b	$67,$71,$DE,$7F,$8D,$7D,$33,$E9,$FF,$51,$F5,$3F,$C4,$B5,$FE,$49
	.dc.b	$39,$BF,$D3,$4E,$5F,$F2,$BC,$67,$0F,$F7,$93,$6B,$F9,$7F,$D4,$3E
	.dc.b	$A5,$F5,$3E,$0C,$A6,$3F,$C8,$F0,$E0,$FF,$AA,$EF,$85,$6B,$1F,$F7
	.dc.b	$F7,$D4,$B9,$BF,$99,$36,$FF,$96,$71,$7C,$A7,$F9,$74,$66,$9A,$7F
	.dc.b	$3F,$AF,$F5,$A5,$70,$F6,$BF,$9E,$7D,$33,$E5,$BF,$9F,$56,$B0,$E3
	.dc.b	$7F,$E7,$6A,$D3,$FC,$ED,$B1,$A3,$FD,$89,$F5,$2F,$A9,$EB,$87,$6F
	.dc.b	$82,$98,$71,$F2,$2E,$1C,$46,$3F,$C6,$5C,$37,$4F,$EB,$1F,$53,$D6
	.dc.b	$F9,$1F,$D3,$5C,$57,$0B,$FE,$B1,$18,$AF,$EC,$17,$CE,$FF,$D2,$3E
	.dc.b	$83,$9F,$FB,$04,$3E,$9D,$F0,$5F,$D8,$6B,$53,$FD,$74,$D4,$66,$A4
	.dc.b	$66,$9F,$53,$8D,$3D,$D8,$72,$6D,$F8,$2B,$77,$FD,$9E,$B5,$97,$F7
	.dc.b	$78,$7C,$37,$9B,$1F,$A3,$39,$8C,$6F,$FD,$DE,$B6,$7F,$71,$2E,$5F
	.dc.b	$5D,$F7,$7D,$0E,$EA,$46,$DF,$24,$62,$7E,$B7,$D4,$BE,$4F,$D0,$7F
	.dc.b	$81,$5C,$5F,$C1,$8C,$11,$F8,$9F,$39,$6B,$9C,$DC,$BF,$C2,$F5,$F4
	.dc.b	$6B,$0F,$FB,$1B,$E0,$BF,$8C,$5F,$A7,$67,$D3,$FF,$85,$67,$38,$72
	.dc.b	$9C,$DF,$F8,$46,$72,$F3,$F2,$7E,$87,$4F,$16,$46,$DC,$B4,$FF,$0D
	.dc.b	$79,$DF,$51,$65,$FE,$DD,$FA,$87,$C2,$FF,$89,$EB,$53,$98,$C7,$2C
	.dc.b	$BF,$F0,$17,$C0,$FF,$7B,$7F,$DC,$77,$C2,$CD,$39,$E4,$AE,$13,$FD
	.dc.b	$56,$FC,$1F,$41,$FD,$A3,$EA,$78,$36,$33,$8F,$AA,$3F,$A3,$E6,$ED
	.dc.b	$CE,$1F,$B3,$CF,$E4,$B7,$87,$FC,$84,$FC,$0F,$85,$F8,$3F,$4E,$FA
	.dc.b	$8F,$4D,$7C,$12,$67,$FE,$41,$BE,$02,$B8,$3F,$DB,$67,$AA,$BE,$B7
	.dc.b	$D2,$EC,$FD,$03,$9D,$F2,$7E,$3D,$FF,$98,$23,$11,$AA,$E4,$F8,$5E
	.dc.b	$4E,$4F,$15,$76,$7B,$5D,$C3,$E1,$DF,$49,$F3,$1C,$CE,$25,$38,$96
	.dc.b	$38,$DF,$82,$38,$CF,$A6,$73,$1E,$4A,$F0,$61,$EB,$BE,$11,$8C,$3C
	.dc.b	$D5,$E9,$3E,$0B,$87,$AC,$FC,$0C,$3D,$15,$C6,$7C,$C7,$1C,$3C,$2F
	.dc.b	$A0,$F3,$1E,$E2,$C6,$1F,$B6,$7C,$AF,$A9,$FE,$2B,$EA,$70,$6C,$7B
	.dc.b	$DA,$EF,$AF,$FC,$FA,$E7,$E4,$BF,$AB,$B7,$A3,$94,$62,$3B,$35,$FC
	.dc.b	$15,$23,$8E,$5F,$83,$FA,$97,$D2,$FE,$1E,$30,$DD,$78,$D7,$06,$E3
	.dc.b	$F8,$1B,$E3,$8D,$BE,$0F,$F0,$0C,$7C,$15,$F8,$76,$C7,$E0,$6B,$8B
	.dc.b	$6B,$8A,$9D,$E3,$F9,$96,$E1,$F9,$DB,$E3,$3E,$8A,$ED,$8F,$9D,$F1
	.dc.b	$35,$4F,$0A,$E5,$87,$83,$0E,$B5,$C3,$1F,$63,$0F,$CC,$46,$9F,$36
	.dc.b	$1D,$70,$E9,$1F,$C8,$1A,$EA,$B9,$1F,$EC,$D8,$F8,$2D,$BF,$DC,$3F
	.dc.b	$ED,$1B,$ED,$3A,$DB,$E0,$78,$BA,$6E,$1F,$EC,$99,$D3,$FC,$C3,$EA
	.dc.b	$7F,$CB,$AF,$FB,$37,$EA,$7F,$CA,$CD,$F5,$F9,$3B,$EC,$AA,$B8,$BF
	.dc.b	$E1,$1A,$E2,$DA,$DC,$EB,$3F,$F7,$42,$F9,$BC,$92,$BC,$5E,$1D,$7A
	.dc.b	$78,$FE,$0C,$7F,$C1,$71,$FE,$E3,$BE,$4D,$BD,$54,$8C,$1F,$EE,$9F
	.dc.b	$A9,$75,$FA,$9B,$19,$8E,$DE,$6E,$5F,$07,$D4,$C9,$D1,$0F,$80,$E5
	.dc.b	$FF,$19,$57,$47,$FC,$69,$F3,$EF,$8E,$F9,$2D,$8F,$F7,$E3,$5F,$95
	.dc.b	$E0,$6A,$7E,$07,$B5,$27,$71,$97,$EC,$46,$5E,$F1,$F5,$BE,$23,$FC
	.dc.b	$A9,$1F,$DB,$B8,$5C,$BF,$D0,$67,$33,$FE,$49,$6C,$EA,$3F,$BA,$F4
	.dc.b	$7F,$9A,$5E,$55,$C7,$CC,$FC,$1C,$BE,$1E,$9E,$18,$95,$C0,$E7,$1A
	.dc.b	$70,$7F,$BC,$23,$F1,$BD,$3F,$72,$3F,$CB,$8D,$C3,$C9,$08,$C6,$1F
	.dc.b	$D6,$FE,$A5,$E5,$72,$E5,$FC,$BC,$3E,$42,$99,$C3,$C3,$87,$27,$FC
	.dc.b	$27,$3B,$7C,$AD,$7C,$75,$FA,$2F,$A9,$79,$D4,$93,$C3,$4A,$FF,$9D
	.dc.b	$7E,$A3,$FD,$50,$FC,$06,$BF,$E7,$9A,$E6,$E7,$DD,$FD,$7F,$D4,$FF
	.dc.b	$AF,$A3,$E3,$B5,$E9,$1F,$B4,$AF,$F2,$EC,$7C,$CF,$A9,$7C,$27,$EE
	.dc.b	$1B,$1E,$1F,$23,$13,$9F,$A9,$93,$C6,$38,$7D,$2F,$F0,$0F,$A9,$7D
	.dc.b	$4D,$33,$5F,$6B,$E2,$FF,$C4,$4D,$F4,$BF,$83,$C6,$23,$FD,$2D,$F2
	.dc.b	$3F,$E3,$17,$47,$E0,$97,$CE,$FF,$0E,$6C,$68,$FF,$4F,$4F,$3E,$AF
	.dc.b	$C1,$CE,$A7,$AD,$78,$2F,$F8,$BF,$D4,$FC,$C9,$FF,$53,$3F,$E7,$17
	.dc.b	$F2,$2D,$73,$CB,$C0,$E3,$EE,$48,$DC,$74,$F1,$E4,$72,$7B,$FD,$16
	.dc.b	$23,$BC,$5C,$FF,$33,$82,$BE,$2F,$CB,$CB,$F9,$4F,$A3,$E4,$B1,$1E
	.dc.b	$8F,$A9,$F5,$98,$C3,$F4,$7B,$AF,$35,$EC,$AD,$78,$3F,$E3,$FF,$A8
	.dc.b	$79,$3F,$CE,$97,$2D,$31,$12,$FF,$6B,$3A,$6C,$E5,$FF,$68,$43,$EA
	.dc.b	$7A,$12,$38,$F8,$15,$DC,$5D,$3F,$77,$5F,$B1,$97,$D4,$BE,$6B,$FA
	.dc.b	$7B,$63,$F8,$3B,$5F,$F7,$07,$D4,$FF,$00,$73,$3F,$D5,$A1,$F3,$34
	.dc.b	$E7,$12,$7F,$88,$CE,$9B,$39,$7A,$FB,$5C,$C8,$F0,$EF,$AE,$8E,$71
	.dc.b	$E1,$8F,$17,$C1,$F8,$FA,$A5,$78,$23,$2F,$8A,$32,$FB,$23,$FB,$7F
	.dc.b	$D4,$3E,$A6,$92,$7F,$94,$FA,$DF,$9D,$DF,$F9,$2D,$6E,$73,$ED,$7F
	.dc.b	$78,$AF,$EE,$B5,$AB,$E5,$48,$FE,$5B,$3A,$E5,$FF,$43,$36,$7F,$D6
	.dc.b	$F8,$75,$78,$4E,$BB,$3B,$75,$C9,$F2,$C7,$84,$C3,$C0,$FF,$60,$DF
	.dc.b	$E8,$DD,$9E,$CE,$BF,$47,$D4,$C8,$C1,$0E,$42,$51,$B8,$CC,$27,$FC
	.dc.b	$33,$F5,$3F,$C0,$FE,$0F,$F5,$15,$FE,$D2,$FA,$9F,$F6,$8B,$FE,$B5
	.dc.b	$EE,$FF,$53,$46,$23,$FD,$59,$1F,$EA,$9A,$E4,$DF,$07,$FD,$3E,$F0
	.dc.b	$F8,$AB,$5D,$5A,$AB,$8B,$93,$C0,$53,$B9,$70,$F3,$BC,$6F,$9E,$1E
	.dc.b	$59,$CB,$FC,$8D,$BE,$26,$B0,$DF,$D4,$F8,$BE,$A1,$F5,$3F,$A9,$BE
	.dc.b	$B1,$8F,$5B,$C7,$9B,$26,$1C,$A3,$35,$B7,$CE,$1C,$5F,$92,$7C,$18
	.dc.b	$7E,$DE,$BF,$D9,$35,$C5,$BE,$33,$E3,$7F,$D5,$33,$99,$FF,$6C,$29
	.dc.b	$D4,$78,$23,$13,$E8,$FA,$8F,$F7,$F7,$37,$9A,$F7,$97,$B9,$BC,$3D
	.dc.b	$1B,$88,$5F,$27,$B4,$69,$F7,$37,$1F,$0C,$7C,$29,$E1,$7D,$F1,$F1
	.dc.b	$12,$75,$D7,$FE,$2F,$9D,$47,$69,$D4,$7E,$02,$B9,$E3,$E7,$9C,$B6
	.dc.b	$B9,$E5,$87,$D5,$8F,$93,$0E,$A7,$3C,$7C,$17,$C6,$32,$D3,$E1,$57
	.dc.b	$C2,$9D,$BD,$32,$F8,$04,$C7,$E8,$4E,$A7,$E2,$C7,$8D,$A7,$98,$F7
	.dc.b	$12,$FC,$2F,$D1,$CB,$DE,$F7,$9E,$58,$FA,$F1,$F2,$D6,$AB,$AD,$F2
	.dc.b	$6C,$EB,$0F,$25,$F1,$9E,$87,$86,$75,$97,$6C,$BD,$93,$E7,$C3,$A5
	.dc.b	$F1,$8E,$A7,$9F,$0E,$B5,$C2,$BC,$58,$F9,$6F,$81,$27,$F1,$51,$97
	.dc.b	$86,$3F,$47,$0E,$98,$F8,$F2,$F9,$57,$B7,$13,$E5,$AE,$0F,$90,$93
	.dc.b	$B7,$DC,$7C,$0C,$7F,$05,$7C,$9F,$3E,$5E,$AC,$31,$87,$C4,$BE,$27
	.dc.b	$5C,$BE,$16,$3D,$70,$F8,$B9,$7B,$31,$E5,$7C,$2F,$71,$8A,$DE,$3F
	.dc.b	$88,$CB,$BD,$F9,$72,$F0,$56,$A3,$C3,$97,$79,$F7,$E3,$D2,$F8,$D7
	.dc.b	$C0,$C7,$B5,$6B,$2F,$25,$6A,$F8,$5F,$96,$7E,$8B,$F4,$0C,$65,$F4
	.dc.b	$30,$E7,$19,$CB,$E3,$E1,$93,$D5,$1F,$C0,$7E,$A7,$F4,$59,$FE,$54
	.dc.b	$FF,$49,$4F,$F1,$5F,$67,$3E,$CE,$99,$19,$3F,$8D,$46,$2F,$F8,$63
	.dc.b	$70,$FE,$08,$DD,$7F,$46,$FF,$AE,$7B,$3F,$D4,$F8,$FF,$A9,$F0,$FF
	.dc.b	$57,$BF,$ED,$36,$CE,$DF,$E4,$1F,$26,$F8,$CF,$F4,$4C,$7F,$AE,$63
	.dc.b	$1F,$11,$BC,$D1,$BF,$53,$0E,$AB,$85,$F2,$AF,$F0,$04,$E5,$B1,$9C
	.dc.b	$BF,$DA,$13,$96,$C3,$EA,$3A,$AF,$A9,$B1,$86,$E1,$E0,$6E,$1D,$9F
	.dc.b	$42,$FA,$8F,$F6,$C4,$FF,$B5,$78,$DF,$53,$FE,$6B,$8F,$F6,$AC,$EA
	.dc.b	$BF,$C3,$15,$AE,$0E,$73,$9C,$FF,$DA,$AD,$BF,$E9,$5F,$22,$F8,$4F
	.dc.b	$A5,$8F,$F8,$62,$B4,$FF,$4F,$6B,$E5,$F8,$3F,$9D,$5C,$78,$BB,$C7
	.dc.b	$C2,$F3,$FE,$1B,$71,$F9,$ED,$BF,$9A,$DC,$BE,$5B,$F2,$BD,$99,$5F
	.dc.b	$5B,$EA,$6D,$FD,$49,$B7,$F4,$A6,$D7,$F4,$4F,$A6,$FF,$16,$3F,$EF
	.dc.b	$D9,$7D,$4F,$DB,$8D,$78,$FF,$87,$9C,$F3,$FD,$2B,$E3,$97,$F8,$E9
	.dc.b	$B2,$BF,$93,$C1,$FE,$7A,$7F,$BA,$46,$6B,$FB,$B5,$70,$9E,$E6,$8B
	.dc.b	$1F,$E4,$FD,$FE,$97,$A5,$E1,$CB,$F1,$7F,$53,$2F,$F2,$FE,$05,$C3
	.dc.b	$F2,$B3,$72,$FC,$9C,$DA,$FD,$3F,$4F,$9D,$FC,$FF,$D0,$FE,$59,$FC
	.dc.b	$BB,$5C,$9F,$A4,$32,$79,$F0,$F5,$FA,$3F,$AD,$E8,$7D,$4B,$BB,$E4
	.dc.b	$4E,$63,$17,$C6,$DC,$0A,$DC,$FE,$76,$BE,$A5,$76,$8E,$15,$B6,$D7
	.dc.b	$5E,$2E,$35,$FB,$81,$F5,$3F,$BE,$6D,$7F,$05,$5F,$1B,$99,$93,$97
	.dc.b	$84,$62,$7A,$7D,$4F,$3F,$8F,$E8,$DB,$E8,$D9,$DC,$77,$9F,$DD,$FD
	.dc.b	$3B,$E2,$BF,$4F,$F5,$3F,$C7,$6F,$FD,$0B,$39,$BF,$9D,$D0,$AD,$72
	.dc.b	$FE,$2E,$D7,$B7,$8E,$1F,$C7,$FE,$A1,$F5,$2F,$A9,$F0,$CC,$CF,$E9
	.dc.b	$61,$C7,$1F,$CB,$61,$C2,$F5,$8F,$FA,$63,$EA,$5C,$FF,$C5,$1B,$8F
	.dc.b	$DB,$E7,$F9,$6F,$DD,$8C,$D7,$FA,$82,$5E,$D7,$F1,$C6,$D7,$E0,$FE
	.dc.b	$99,$F2,$79,$AF,$E4,$31,$89,$D4,$FF,$24,$FA,$97,$D4,$EA,$37,$F0
	.dc.b	$52,$BA,$79,$37,$3F,$F5,$8E,$3F,$C7,$38,$3E,$53,$6B,$FC,$93,$F5
	.dc.b	$3D,$1E,$27,$E5,$DC,$57,$2B,$FE,$6B,$18,$BE,$71,$FC,$E3,$D0,$74
	.dc.b	$7F,$51,$C1,$F4,$FF,$EC,$17,$F4,$53,$A7,$F5,$86,$BE,$03,$95,$48
	.dc.b	$CD,$3E,$A7,$19,$FF,$B3,$30,$E4,$DB,$F0,$56,$EF,$BC,$7F,$4C,$5F
	.dc.b	$0D,$E5,$C3,$E7,$5F,$FB,$5A,$72,$79,$23,$39,$FE,$46,$33,$3B,$AE
	.dc.b	$FF,$43,$D5,$93,$FB,$F3,$75,$FC,$8D,$F8,$73,$AA,$FF,$72,$FD,$4B
	.dc.b	$E4,$FD,$27,$F6,$27,$15,$F0,$A3,$04,$7E,$2F,$C2,$5A,$E7,$37,$0F
	.dc.b	$06,$8A,$C5,$FF,$A2,$3E,$0B,$F8,$AE,$1F,$EF,$F8,$FA,$FF,$4F,$FD
	.dc.b	$CA,$73,$87,$28,$C7,$95,$2B,$EA,$F5,$5C,$3F,$DF,$F8,$70,$9E,$D5
	.dc.b	$F5,$2B,$58,$7E,$EF,$C3,$FB,$2F,$F4,$78,$7D,$43,$E1,$7F,$C0,$75
	.dc.b	$AC,$3F,$D8,$15,$A9,$CF,$1C,$9F,$F0,$37,$C0,$FF,$76,$BE,$1A,$7F
	.dc.b	$7E,$AD,$3F,$E0,$69,$D4,$76,$FA,$2F,$EC,$5F,$53,$C2,$96,$9C,$D6
	.dc.b	$F0,$FE,$79,$39,$7F,$C2,$D9,$BB,$35,$83,$5F,$81,$C1,$EC,$3F,$C3
	.dc.b	$F1,$FE,$DB,$F8,$1F,$F6,$E7,$D4,$BE,$A7,$FD,$7B,$9A,$F8,$67,$E0
	.dc.b	$78,$E0,$C3,$8D,$F0,$9C,$C6,$27,$FC,$3F,$1F,$DD,$70,$ED,$1F,$EC
	.dc.b	$A3,$90,$72,$FC,$28,$E1,$5B,$8D,$CE,$B0,$F3,$61,$DE,$7F,$DA,$B8
	.dc.b	$59,$CC,$75,$BE,$53,$9C,$3A,$CF,$8E,$31,$7C,$E3,$E3,$1C,$4F,$5B
	.dc.b	$97,$C7,$5F,$06,$31,$3C,$A7,$81,$E4,$2D,$F9,$30,$F7,$C7,$67,$D9
	.dc.b	$87,$92,$38,$46,$2F,$AC,$72,$9D,$3C,$70,$F1,$3E,$03,$D2,$7C,$93
	.dc.b	$E7,$97,$0C,$3A,$3F,$DF,$BF,$53,$FC,$63,$D4,$E0,$D8,$FA,$4D,$7E
	.dc.b	$D3,$5E,$3F,$23,$F9,$17,$6D,$76,$D7,$7A,$76,$8E,$19,$FE,$0B,$EA
	.dc.b	$5F,$48,$BF,$BD,$18,$7F,$01,$82,$F8,$E7,$D2,$7E,$8D,$6D,$B7,$CD
	.dc.b	$FA,$B7,$C7,$1E,$78,$72,$9F,$05,$F1,$C7,$C9,$87,$17,$F2,$F8,$71
	.dc.b	$AE,$77,$C5,$5F,$1C,$7C,$0D,$8F,$65,$71,$BE,$77,$C6,$B5,$5C,$B0
	.dc.b	$F0,$61,$ED,$BD,$E3,$F1,$27,$35,$E8,$6A,$C7,$B5,$F1,$3C,$73,$AC
	.dc.b	$3B,$CE,$AB,$91,$FE,$4B,$8F,$95,$7F,$E2,$56,$EB,$CE,$76,$7F,$11
	.dc.b	$7D,$FE,$17,$CF,$AE,$0D,$8F,$0C,$FF,$95,$FE,$A3,$FC,$B8,$FF,$96
	.dc.b	$7E,$A7,$FC,$B0,$DF,$5F,$83,$F8,$39,$DB,$DC,$C7,$5E,$95,$72,$74
	.dc.b	$62,$B5,$5C,$B1,$FF,$20,$2F,$9A,$C9,$2B,$C9,$E1,$57,$9E,$0F,$1B
	.dc.b	$C0,$D8,$F6,$56,$F0,$FF,$26,$4E,$67,$8D,$67,$3F,$CA,$7D,$47,$F4
	.dc.b	$19,$FF,$3D,$35,$F4,$B6,$F9,$F9,$BE,$F8,$C5,$7E,$77,$07,$D4,$F9
	.dc.b	$0D,$76,$CC,$7E,$74,$65,$EF,$1F,$27,$E7,$9F,$53,$49,$3F,$57,$BE
	.dc.b	$BF,$0B,$DA,$F8,$26,$35,$FF,$35,$5F,$29,$FB,$93,$A7,$FD,$2D,$19
	.dc.b	$7E,$AF,$C1,$FF,$95,$9F,$F4,$D3,$5C,$FA,$1F,$77,$A7,$FC,$CF,$1A
	.dc.b	$8C,$7C,$3C,$8C,$10,$FC,$0C,$A3,$51,$D0,$FF,$54,$74,$7E,$05,$BC
	.dc.b	$9F,$8B,$8C,$47,$E2,$BE,$A5,$E4,$E6,$BF,$21,$0F,$90,$A6,$6B,$99
	.dc.b	$FE,$B5,$8D,$36,$34,$FE,$5B,$EA,$7F,$C5,$4F,$FA,$ED,$AE,$7C,$CE
	.dc.b	$FA,$7E,$7E,$73,$C1,$FE,$99,$FA,$8F,$F5,$15,$FF,$B0,$DA,$F1,$31
	.dc.b	$EE,$FE,$9F,$EA,$7F,$D6,$F1,$FE,$C9,$6B,$FA,$89,$E3,$5D,$7D,$3F
	.dc.b	$D7,$7D,$4B,$E1,$29,$FB,$1A,$FF,$30,$DF,$FB,$4E,$74,$DF,$A9,$E2
	.dc.b	$70,$75,$F4,$BF,$B9,$FA,$97,$D4,$D3,$BC,$7E,$F7,$E2,$FF,$C4,$15
	.dc.b	$FE,$DF,$6B,$FB,$E6,$E9,$FE,$BB,$F9,$17,$AD,$C7,$D4,$FE,$0D,$0F
	.dc.b	$9E,$A6,$70,$E9,$38,$7F,$DD,$93,$B7,$27,$FB,$BF,$EA,$7E,$8F,$8D
	.dc.b	$FD,$35,$FF,$37,$46,$7D,$CF,$E2,$91,$A8,$FE,$B8,$D4,$F0,$7F,$8B
	.dc.b	$B7,$2F,$F6,$A3,$B3,$FD,$7F,$1E,$1F,$27,$A9,$DB,$E8,$AC,$64,$87
	.dc.b	$47,$E5,$7C,$BC,$8F,$94,$F9,$BE,$4B,$11,$E2,$FA,$9F,$49,$8C,$7F
	.dc.b	$DC,$DC,$0B,$CD,$7C,$3A,$D6,$9F,$DB,$FE,$A1,$E3,$FF,$35,$5C,$74
	.dc.b	$C4,$4B,$FF,$00,$65,$FE,$F3,$AD,$C6,$2F,$A5,$EE,$3F,$9E,$CB,$EA
	.dc.b	$7B,$12,$38,$1F,$E0,$B8,$FD,$DD,$F5,$9D,$4F,$27,$CD,$E1,$FF,$4B
	.dc.b	$FA,$97,$CD,$52,$57,$41,$F5,$3F,$C1,$1C,$4F,$07,$50,$F9,$9F,$EA
	.dc.b	$CD,$8D,$D7,$47,$FA,$BC,$65,$B3,$9C,$5A,$7F,$86,$EB,$FC,$45,$5D
	.dc.b	$BD,$DE,$A7,$6F,$06,$C6,$48,$76,$7C,$7D,$7E,$4E,$A9,$3E,$88,$CB
	.dc.b	$EA,$8C,$BE,$68,$FE,$E1,$F5,$0F,$A9,$A4,$9F,$8D,$EB,$FE,$7E,$75
	.dc.b	$18,$F6,$BF,$BC,$4F,$F7,$7D,$7D,$B3,$FE,$41,$AD,$BF,$DE,$9B,$1A
	.dc.b	$8F,$EE,$55,$C8,$F0,$3F,$E0,$0A,$FD,$3B,$7A,$BB,$9E,$0E,$B9,$1E
	.dc.b	$68,$CB,$E8,$8F,$2F,$5F,$9B,$EA,$64,$FB,$23,$2F,$BA,$32,$F8,$63
	.dc.b	$FC,$29,$F5,$3F,$CE,$BE,$0F,$F5,$13,$FE,$5D,$FA,$9F,$F6,$93,$FE
	.dc.b	$B5,$EE,$FF,$2B,$46,$23,$FD,$61,$1F,$E5,$9A,$E4,$DF,$3B,$FD,$42
	.dc.b	$F0,$F8,$9B,$5D,$5A,$AB,$8B,$93,$C0,$53,$B9,$70,$F3,$BC,$6F,$9E
	.dc.b	$1E,$59,$CB,$FC,$96,$78,$87,$11,$AB,$E3,$F5,$3E,$4F,$A8,$7D,$4F
	.dc.b	$EA,$6B,$BC,$63,$D5,$F2,$E3,$CE,$F8,$B8,$9D,$3E,$06,$D7,$92,$7B
	.dc.b	$5F,$5D,$3F,$D0,$55,$A6,$C6,$6F,$EA,$57,$0C,$3F,$D5,$57,$DA,$7F
	.dc.b	$DB,$4B,$C5,$7C,$6E,$B0,$FE,$05,$5A,$9C,$BE,$56,$E5,$F3,$3E,$A3
	.dc.b	$FA,$AF,$77,$AE,$73,$A7,$99,$B7,$FE,$A3,$CB,$C1,$97,$C4,$C3,$8C
	.dc.b	$F5,$8E,$44,$7E,$9C,$6B,$1F,$9B,$7D,$25,$7D,$1E,$47,$0E,$AF,$D9
	.dc.b	$D7,$0C,$3A,$36,$75,$96,$32,$F4,$55,$9F,$F5,$D6,$5E,$19,$CC,$76
	.dc.b	$BE,$47,$C3,$C3,$BC,$EA,$32,$D3,$FD,$75,$5C,$EB,$E2,$65,$EB,$CB
	.dc.b	$D5,$7C,$27,$C3,$87,$47,$78,$F9,$5F,$83,$7F,$0F,$0E,$91,$89,$D6
	.dc.b	$1D,$31,$F0,$E1,$CE,$B4,$7A,$E3,$CC,$F8,$32,$E3,$8F,$94,$ED,$5C
	.dc.b	$6F,$A5,$79,$0E,$47,$12,$9D,$B2,$F6,$2C,$BC,$F3,$CA,$35,$8F,$89
	.dc.b	$D5,$F7,$C3,$C5,$97,$C1,$C3,$8C,$62,$BB,$C6,$27,$D2,$7E,$3E,$B9
	.dc.b	$38,$9F,$C0,$9E,$6C,$3C,$47,$8E,$B9,$46,$6F,$E0,$E5,$C7,$2F,$5D
	.dc.b	$69,$F6,$DF,$1C,$7B,$E1,$C7,$1F,$16,$1C,$72,$F7,$63,$C5,$F8,$18
	.dc.b	$71,$9D,$5F,$CD,$C7,$F1,$D1,$88,$F8,$E7,$CF,$8F,$0E,$5F,$0D,$DC
	.dc.b	$78,$4F,$8B,$97,$92,$37,$97,$92,$75,$1E,$C9,$D6,$3E,$4B,$F8,$35
	.dc.b	$A8,$F4,$CE,$A3,$CD,$18,$C3,$EB,$4E,$67,$ED,$C6,$0F,$05,$FF,$0B
	.dc.b	$FA,$9F,$CA,$67,$F9,$A3,$FD,$29,$3F,$C1,$FD,$7C,$FB,$3A,$64,$64
	.dc.b	$FE,$19,$18,$C3,$F8,$4B,$70,$FE,$30,$DC,$BF,$66,$FF,$78,$EB,$FE
	.dc.b	$E9,$8A,$31,$87,$EE,$3D,$FB,$8F,$F0,$49,$BF,$26,$F8,$CF,$F4,$65
	.dc.b	$FF,$8E,$23,$1F,$11,$A9,$CD,$FF,$C0,$19,$CD,$6E,$73,$97,$F8,$2A
	.dc.b	$73,$5B,$9C,$D2,$13,$AA,$E1,$4F,$A9,$F0,$7D,$4F,$85,$B8,$77,$6E
	.dc.b	$1D,$5B,$7F,$E1,$5F,$A8,$FF,$68,$57,$F9,$4B,$91,$F5,$3F,$E6,$98
	.dc.b	$FF,$25,$4E,$B0,$5F,$F8,$8A,$F7,$C1,$F2,$25,$5B,$AE,$DF,$05,$F3
	.dc.b	$3E,$49,$F0,$BF,$47,$2E,$B3,$C6,$F8,$47,$C1,$C3,$9C,$F1,$AE,$11
	.dc.b	$E2,$9D,$57,$08,$F8,$FB,$7F,$99,$27,$E8,$5F,$D5,$F3,$BE,$B3,$63
	.dc.b	$27,$D5,$8C,$BF,$5F,$D8,$FB,$87,$DD,$FA,$9B,$19,$26,$1F,$66,$32
	.dc.b	$F9,$4F,$B5,$F4,$CF,$17,$F1,$BB,$3E,$A7,$EF,$D6,$A7,$31,$88,$57
	.dc.b	$1C,$3F,$CF,$8D,$8E,$DB,$FC,$CF,$4B,$E9,$4E,$71,$FF,$29,$36,$FE
	.dc.b	$63,$6D,$7C,$67,$AE,$3F,$9B,$FA,$99,$1F,$99,$F0,$2D,$FE,$5E,$6E
	.dc.b	$3F,$94,$9B,$5F,$99,$E8,$FD,$43,$FA,$5F,$A1,$FD,$33,$4F,$D2,$17
	.dc.b	$0F,$D1,$36,$FD,$3E,$77,$ED,$39,$DF,$52,$EC,$F7,$CE,$63,$17,$F1
	.dc.b	$A3,$2F,$E7,$63,$E6,$6E,$9C,$E3,$F5,$CF,$9E,$F3,$F8,$5C,$E5,$FB
	.dc.b	$79,$7D,$4F,$F0,$46,$E3,$FC,$0D,$7C,$6E,$66,$4E,$5E,$11,$89,$FE
	.dc.b	$0C,$3E,$A7,$A7,$C7,$F5,$6D,$F5,$6C,$F0,$6E,$1F,$E8,$FF,$A7,$7C
	.dc.b	$67,$EB,$FE,$A5,$E3,$7E,$D6,$73,$8F,$FB,$43,$A1,$5A,$8F,$F6,$44
	.dc.b	$EB,$2F,$88,$DC,$7F,$D4,$3F,$50,$FA,$97,$D4,$F5,$D2,$63,$C9,$FE
	.dc.b	$57,$5B,$9C,$E3,$F3,$FE,$A5,$D1,$FC,$25,$B8,$7F,$B6,$7A,$3E,$5B
	.dc.b	$FE,$DB,$8C,$D7,$D7,$97,$B3,$FC,$45,$B5,$FE,$E1,$FA,$77,$CA,$66
	.dc.b	$BF,$8B,$46,$27,$53,$FC,$67,$EA,$5F,$53,$A8,$DF,$C1,$4A,$E9,$E8
	.dc.b	$5C,$FF,$11,$8F,$F1,$4E,$2F,$F6,$03,$6B,$F8,$D7,$D4,$F4,$78,$9F
	.dc.b	$EF,$57,$15,$CA,$FF,$95,$46,$2F,$9C,$7F,$2D,$F0,$5D,$5F,$D5,$B0
	.dc.b	$7D,$3B,$E0,$B5,$39,$8C,$4F,$F4,$BC,$3F,$DA,$97,$CA,$30,$7F,$2F
	.dc.b	$71,$97,$F1,$98,$EB,$F5,$3D,$1D,$C7,$EF,$70,$C3,$BC,$15,$8C,$3F
	.dc.b	$DC,$56,$F8,$6F,$44,$FD,$1B,$FD,$84,$E4,$EA,$72,$DF,$FA,$63,$98
	.dc.b	$DD,$7B,$7E,$83,$2C,$9F,$2C,$9E,$18,$7F,$BE,$1F,$8F,$5B,$9F,$E0
	.dc.b	$1F,$52,$F9,$3F,$90,$7F,$5B,$71,$5E,$4C,$ED,$70,$AD,$1F,$3F,$E1
	.dc.b	$2D,$78,$AF,$78,$F8,$EF,$FB,$E4,$62,$7F,$BF,$1F,$6A,$B8,$CF,$F1
	.dc.b	$3F,$81,$FE,$36,$FC,$DF,$A7,$FE,$DD,$18,$C3,$D9,$92,$73,$E5,$CC
	.dc.b	$7D,$1D,$57,$1E,$93,$70,$E3,$36,$BF,$B8,$78,$7F,$82,$9F,$EF,$DF
	.dc.b	$50,$F8,$5F,$EF,$D3,$9A,$FF,$0B,$5A,$B5,$1F,$4A,$75,$8F,$F8,$6D
	.dc.b	$B4,$F8,$2F,$EF,$13,$FD,$4B,$97,$A9,$2F,$FC,$29,$3A,$BE,$75,$FD
	.dc.b	$63,$E8,$3F,$B5,$FD,$4E,$25,$18,$C3,$CE,$CC,$3F,$C3,$D9,$3B,$71
	.dc.b	$8F,$F4,$1C,$3E,$55,$7A,$77,$FF,$0E,$4F,$F8,$CF,$E0,$BF,$C6,$9F
	.dc.b	$52,$FA,$9F,$8F,$5F,$10,$97,$F9,$2B,$E0,$2B,$83,$F0,$E7,$AA,$BE
	.dc.b	$B7,$D2,$EC,$FF,$88,$CD,$1C,$E3,$4D,$35,$3C,$1F,$EF,$B5,$F9,$1B
	.dc.b	$F0,$C7,$F8,$B6,$75,$87,$C0,$8E,$46,$4A,$E6,$78,$C7,$B1,$E7,$19
	.dc.b	$BF,$34,$F9,$A7,$37,$E1,$9C,$DF,$82,$31,$5E,$83,$BB,$87,$C5,$5C
	.dc.b	$B0,$F6,$47,$86,$FE,$0C,$F9,$23,$C1,$18,$3B,$BD,$27,$8B,$A3,$E9
	.dc.b	$38,$BF,$4C,$7C,$12,$B8,$7E,$D9,$F4,$FE,$A7,$F8,$E7,$BD,$C1,$B1
	.dc.b	$F0,$9A,$EF,$AF,$FD,$14,$E7,$E4,$7F,$CD,$6D,$E5,$E5,$18,$8E,$CD
	.dc.b	$7F,$CC,$94,$8E,$39,$FD,$CF,$A9,$7D,$2F,$F9,$A6,$30,$DD,$78,$D7
	.dc.b	$06,$E3,$F8,$7B,$E3,$8D,$BE,$0F,$FA,$4F,$1F,$05,$7E,$35,$B1,$F8
	.dc.b	$8A,$E2,$DA,$E2,$A7,$78,$FE,$49,$B8,$7F,$9F,$EF,$8C,$FA,$2B,$B6
	.dc.b	$3E,$77,$C4,$D5,$3C,$2B,$96,$1E,$0C,$3A,$D7,$0C,$7D,$8C,$3F,$3F
	.dc.b	$1A,$7C,$D8,$75,$C3,$A4,$7F,$AC,$DA,$EA,$B9,$1F,$CF,$5F,$1D,$FF
	.dc.b	$A6,$63,$13,$F1,$1D,$5F,$78,$C6,$BF,$07,$C3,$E0,$DC,$3F,$AD,$CE
	.dc.b	$9F,$F6,$27,$D4,$FF,$98,$5F,$EB,$FF,$51,$FE,$5B,$6F,$B1,$C9,$DF
	.dc.b	$65,$55,$C5,$FF,$30,$57,$16,$D6,$E7,$59,$FB,$57,$CC,$64,$95,$E2
	.dc.b	$F0,$EB,$D3,$C7,$F0,$63,$FE,$15,$8F,$EE,$B7,$C9,$B7,$AA,$91,$83
	.dc.b	$FB,$DF,$D4,$FF,$3C,$6F,$5F,$AE,$31,$1E,$AF,$37,$13,$E0,$FA,$99
	.dc.b	$3C,$08,$7C,$07,$2F,$FA,$06,$B9,$E9,$F6,$FE,$7B,$FC,$D3,$3F,$E0
	.dc.b	$C6,$BF,$E8,$76,$C7,$E0,$74,$FF,$1A,$CE,$FD,$8F,$83,$ED,$48,$C1
	.dc.b	$0F,$A2,$E5,$E9,$1F,$5F,$E2,$7F,$CB,$31,$FE,$1C,$6B,$FE,$97,$6D
	.dc.b	$FE,$47,$9F,$FC,$D7,$3B,$F3,$3F,$01,$0F,$88,$A7,$58,$93,$FE,$28
	.dc.b	$9D,$36,$72,$FF,$A9,$7A,$3E,$C3,$78,$BF,$19,$18,$C3,$F4,$5F,$52
	.dc.b	$F2,$F3,$5F,$91,$87,$C8,$53,$38,$7F,$8E,$30,$E5,$D7,$FF,$00,$6B
	.dc.b	$73,$FE,$3C,$6B,$E1,$3F,$C7,$FF,$52,$F3,$BF,$3C,$D8,$CC,$78,$76
	.dc.b	$CB,$FA,$3F,$A8,$FF,$53,$CF,$F9,$21,$AF,$FB,$0E,$B9,$3F,$EC,$58
	.dc.b	$CE,$BF,$BF,$FA,$9F,$F5,$FC,$FF,$94,$1A,$EF,$DD,$ED,$E2,$FE,$C7
	.dc.b	$EA,$5F,$09,$4C,$D7,$F9,$62,$32,$DF,$A9,$E0,$6D,$D7,$D2,$7E,$EF
	.dc.b	$EA,$5F,$53,$4E,$D1,$FB,$FF,$8B,$FF,$0F,$57,$F9,$91,$AF,$FB,$7D
	.dc.b	$BA,$FF,$19,$F9,$1F,$F1,$A4,$EF,$D2,$FE,$11,$0F,$A9,$A7,$58,$93
	.dc.b	$FE,$70,$9D,$36,$72,$FF,$BB,$7E,$A7,$E5,$DF,$F2,$77,$FC,$E1,$1F
	.dc.b	$E7,$86,$BB,$E4,$FE,$31,$1C,$39,$D3,$23,$75,$CF,$DD,$DD,$F0,$F1
	.dc.b	$E5,$F5,$79,$FD,$16,$23,$C9,$16,$FC,$93,$F2,$7E,$5E,$57,$C9,$7C
	.dc.b	$7F,$25,$88,$F6,$7D,$4F,$98,$C5,$FF,$43,$E0,$5E,$6B,$DD,$5A,$D7
	.dc.b	$D7,$0F,$A9,$79,$3F,$CD,$D7,$1D,$31,$12,$7F,$D3,$F3,$A6,$CE,$56
	.dc.b	$5F,$D5,$65,$F5,$3D,$49,$3C,$BC,$3A,$F0,$C7,$CB,$5C,$23,$BD,$79
	.dc.b	$6B,$FA,$67,$D4,$BE,$6A,$92,$BF,$D6,$7F,$53,$FC,$0D,$FF,$5A,$CE
	.dc.b	$DC,$FC,$CF,$F5,$59,$3C,$63,$C4,$7E,$02,$78,$38,$13,$C1,$E6,$76
	.dc.b	$AE,$9E,$CC,$8F,$1B,$DF,$C0,$C4,$73,$AF,$F1,$63,$A3,$D3,$1E,$5F
	.dc.b	$83,$F3,$76,$49,$F6,$C6,$5F,$7C,$65,$E9,$1F,$DC,$7E,$A1,$F5,$34
	.dc.b	$93,$FE,$D8,$F5,$BF,$3C,$FE,$7A,$76,$E7,$D9,$FE,$F5,$3F,$DD,$F8
	.dc.b	$FF,$C2,$15,$AA,$F8,$8D,$8F,$F0,$A4,$EF,$8F,$F1,$ED,$AE,$D8,$74
	.dc.b	$EB,$EB,$D7,$27,$C9,$19,$7C,$D1,$97,$E9,$52,$7C,$BD,$7E,$6F,$A9
	.dc.b	$95,$ED,$8C,$BE,$F8,$CB,$CA,$3F,$C2,$DF,$53,$FC,$C3,$E0,$BF,$53
	.dc.b	$3F,$C0,$3E,$A7,$FD,$A4,$FF,$AC,$FB,$BF,$53,$18,$8F,$F5,$54,$7E
	.dc.b	$B2,$B9,$37,$C0,$FF,$50,$3C,$3E,$2A,$D7,$56,$AA,$E2,$E4,$F0,$14
	.dc.b	$EE,$5C,$3C,$EF,$1B,$E7,$87,$96,$72,$FF,$26,$9E,$21,$C4,$6A,$F8
	.dc.b	$FD,$4F,$93,$EA,$1F,$53,$FA,$BA,$EF,$18,$F5,$BC,$B8,$F3,$BE,$2E
	.dc.b	$27,$4F,$81,$B5,$E4,$9E,$D7,$D7,$4F,$E2,$B5,$A6,$C6,$6F,$EA,$57
	.dc.b	$0C,$3B,$4E,$65,$7C,$5D,$9A,$9F,$F5,$7C,$79,$23,$FD,$BC,$BC,$57
	.dc.b	$C6,$EA,$FF,$85,$56,$A7,$2F,$AD,$B9,$7D,$2F,$A8,$75,$39,$FE,$15
	.dc.b	$6B,$5F,$6B,$70,$FE,$9B,$9F,$87,$3F,$97,$8F,$2C,$38,$C4,$AF,$03
	.dc.b	$AC,$7B,$46,$09,$97,$D6,$C7,$B4,$30,$E8,$EB,$2F,$1E,$1D,$6F,$D5
	.dc.b	$87,$47,$94,$66,$77,$5F,$8A,$BE,$75,$A8,$E8,$7B,$30,$F0,$CE,$AB
	.dc.b	$8D,$6C,$8F,$F5,$48,$D5,$71,$BF,$03,$4F,$06,$1E,$5A,$E6,$48,$D5
	.dc.b	$FA,$B1,$E3,$1F,$84,$C7,$E1,$E5,$F3,$E3,$9C,$EA,$31,$5E,$49,$FF
	.dc.b	$00,$E5,$E4,$79,$65,$BC,$3D,$B9,$78,$E7,$C3,$97,$8D,$D9,$D8,$F8
	.dc.b	$A6,$30,$EC,$FD,$59,$E4,$E6,$79,$63,$E4,$7E,$34,$EA,$31,$3E,$02
	.dc.b	$1C,$F2,$F3,$63,$F0,$B2,$F5,$57,$0A,$EC,$70,$8C,$E5,$F1,$89,$5F
	.dc.b	$8F,$BE,$75,$EA,$AD,$CE,$32,$FA,$13,$EE,$C3,$B1,$DE,$B7,$19,$C3
	.dc.b	$E4,$E1,$C7,$2F,$91,$5A,$C3,$E2,$E1,$C7,$2F,$81,$87,$93,$1F,$14
	.dc.b	$63,$2E,$51,$F6,$23,$0F,$78,$F5,$34,$E3,$97,$07,$E1,$9F,$24,$D6
	.dc.b	$1F,$5B,$1F,$26,$5F,$53,$1F,$4D,$6A,$3D,$38,$F8,$2F,$D3,$5A,$79
	.dc.b	$46,$67,$ED,$E1,$F5,$63,$F0,$38,$7E,$3C,$7D,$4E,$23,$3E,$CC,$AF
	.dc.b	$E0,$FE,$CD,$CF,$F6,$93,$72,$FE,$29,$37,$2F,$D7,$36,$BF,$3A,$D7
	.dc.b	$F8,$33,$FC,$33,$AF,$F8,$6C,$63,$0F,$F3,$1C,$63,$37,$BB,$B9,$EA
	.dc.b	$68,$FF,$34,$FC,$9B,$E2,$FF,$D2,$71,$9C,$13,$9F,$87,$E5,$18,$AD
	.dc.b	$BF,$E3,$49,$D5,$71,$7A,$7A,$FC,$1C,$CE,$EB,$8E,$5F,$C9,$A7,$2D
	.dc.b	$87,$D4,$71,$5F,$53,$63,$0D,$C7,$C1,$3A,$6C,$E9,$F6,$60,$FA,$8F
	.dc.b	$F7,$14,$7B,$F9,$5F,$53,$FE,$76,$8D,$CE,$AB,$FC,$A3,$5A,$E1,$F8
	.dc.b	$71,$CF,$E0,$AD,$6F,$4F,$E8,$3F,$22,$F8,$57,$26,$AE,$67,$D9,$F9
	.dc.b	$B5,$B9,$CE,$3F,$1F,$CE,$F8,$55,$F0,$DB,$7F,$01,$B3,$B8,$F8,$73
	.dc.b	$F2,$7D,$A9,$1F,$1B,$EA,$6D,$7D,$39,$B9,$FD,$19,$B5,$FE,$94,$FA
	.dc.b	$67,$8D,$F8,$2A,$7D,$4F,$DA,$AD,$4E,$63,$18,$FF,$AD,$AF,$8D,$7D
	.dc.b	$26,$C7,$F9,$F3,$5F,$8D,$EA,$7C,$C9,$CE,$3F,$9E,$6D,$F6,$3F,$2D
	.dc.b	$5C,$1E,$98,$FE,$1F,$EA,$64,$7E,$13,$C1,$B9,$7E,$46,$6E,$3F,$8F
	.dc.b	$9B,$5F,$A4,$E7,$FF,$03,$3F,$8C,$FA,$1F,$C9,$3F,$93,$6B,$93,$F2
	.dc.b	$26,$4F,$77,$A3,$FA,$8E,$77,$D4,$BB,$BD,$D3,$98,$C5,$70,$26,$9F
	.dc.b	$90,$8F,$96,$B7,$57,$2A,$DE,$9F,$E1,$0C,$7B,$F7,$67,$0F,$DB,$0F
	.dc.b	$A9,$FD,$A3,$6F,$F7,$0B,$E3,$73,$32,$72,$F0,$8C,$4F,$4F,$A9,$E7
	.dc.b	$F3,$7D,$1B,$7D,$1B,$3B,$8E,$F3,$FA,$AF,$A8,$7C,$57,$E8,$BE,$A7
	.dc.b	$F8,$CD,$7E,$E2,$72,$FE,$02,$33,$7F,$EC,$19,$CB,$6B,$F8,$BF,$D4
	.dc.b	$3E,$A5,$F5,$3D,$E6,$63,$D5,$DC,$EA,$30,$7F,$08,$FA,$97,$37,$F0
	.dc.b	$06,$C7,$F0,$CE,$1F,$90,$EE,$7F,$82,$C7,$FA,$8E,$35,$EC,$FF,$08
	.dc.b	$FA,$77,$CA,$6A,$30,$C9,$FE,$2F,$F5,$2F,$A9,$C6,$FF,$ED,$FC,$1F
	.dc.b	$05,$CC,$E9,$E4,$5C,$3F,$8F,$CF,$F1,$06,$C7,$F1,$8F,$A9,$E9,$18
	.dc.b	$3F,$92,$B8,$BE,$D3,$BC,$3A,$D7,$F2,$AF,$41,$CD,$FD,$12,$1F,$4E
	.dc.b	$F8,$1F,$E4,$B5,$A9,$FE,$98,$6A,$33,$52,$33,$4F,$A9,$C6,$5F,$EF
	.dc.b	$DC,$39,$36,$FC,$15,$BB,$F2,$47,$F3,$45,$F0,$DE,$5C,$3E,$7C,$7F
	.dc.b	$42,$32,$79,$1C,$DB,$BF,$FB,$6C,$EA,$30,$98,$E6,$3F,$CF,$EF,$C2
	.dc.b	$FA,$1F,$15,$C8,$F0,$3E,$88,$C5,$74,$3A,$7D,$4B,$E4,$BD,$A7,$F4
	.dc.b	$E7,$17,$F1,$A3,$04,$7E,$4F,$C2,$5A,$E8,$2B,$8B,$86,$E3,$E8,$D1
	.dc.b	$5C,$63,$FC,$1B,$F0,$3F,$C5,$2F,$CD,$87,$DE,$FA,$7F,$EC,$33,$9A
	.dc.b	$E5,$18,$F2,$65,$7D,$CE,$AB,$87,$93,$0E,$13,$F2,$E1,$18,$7F,$D6
	.dc.b	$D9,$7F,$B0,$27,$FB,$CC,$E6,$5E,$77,$E1,$32,$FE,$41,$F5,$0F,$85
	.dc.b	$FE,$EB,$39,$C3,$EC,$64,$AD,$47,$F8,$12,$75,$87,$DD,$6D,$7F,$B2
	.dc.b	$3D,$5E,$7C,$92,$F8,$3F,$CF,$7E,$07,$FC,$01,$1E,$BB,$E7,$87,$9B
	.dc.b	$E8,$FF,$B4,$7D,$4E,$25,$18,$C3,$CE,$CB,$FF,$0A,$67,$FB,$E7,$CD
	.dc.b	$1E,$CE,$19,$1F,$E1,$CC,$7F,$29,$F0,$5F,$95,$FA,$97,$D4,$F8,$EF
	.dc.b	$75,$F8,$4C,$D1,$FE,$1A,$8E,$22,$B8,$BF,$0E,$34,$AF,$B5,$F5,$C2
	.dc.b	$CF,$DB,$34,$73,$8D,$34,$D4,$F0,$7E,$DD,$7F,$BA,$2F,$C3,$1F,$85
	.dc.b	$9D,$61,$E8,$C3,$84,$FA,$2B,$73,$98,$EC,$71,$2B,$98,$EB,$1C,$70
	.dc.b	$F2,$C6,$23,$D2,$F4,$C3,$C9,$3E,$D9,$CB,$C0,$C6,$1E,$69,$DC,$FB
	.dc.b	$8F,$0B,$87,$C9,$5D,$70,$F5,$C7,$43,$E8,$38,$AE,$8D,$8F,$11,$F6
	.dc.b	$0B,$18,$70,$7E,$B7,$EA,$7F,$91,$FA,$BC,$1B,$1E,$B6,$BF,$65,$AF
	.dc.b	$1F,$92,$FE,$48,$ED,$AE,$DA,$EF,$93,$B4,$70,$C7,$ED,$FD,$4B,$E9
	.dc.b	$7F,$80,$46,$1B,$AF,$1C,$3E,$CE,$2C,$39,$4F,$F0,$7C,$39,$57,$E2
	.dc.b	$5A,$F0,$9D,$BF,$CD,$EB,$96,$3B,$C3,$BC,$FF,$3B,$BE,$38,$FE,$42
	.dc.b	$F7,$3C,$EF,$83,$0C,$12,$7D,$4E,$5E,$58,$78,$A1,$5D,$27,$71,$F9
	.dc.b	$6B,$F0,$E3,$F0,$67,$8E,$1C,$E3,$E0,$63,$E8,$BE,$78,$F5,$C3,$E0
	.dc.b	$DF,$FB,$11,$BA,$70,$AF,$F5,$0E,$5E,$1C,$71,$1F,$D1,$AB,$FD,$2A
	.dc.b	$FF,$A9,$9A,$E7,$1F,$E6,$D8,$7F,$B4,$3E,$A7,$FC,$B5,$3F,$EA,$CF
	.dc.b	$A7,$FF,$29,$B7,$D8,$DE,$7F,$CF,$6B,$73,$FD,$51,$B7,$FE,$DA,$BE
	.dc.b	$33,$FD,$61,$B1,$DF,$B3,$9C,$FF,$71,$A7,$CC,$62,$11,$87,$FD,$7E
	.dc.b	$D7,$A7,$83,$B8,$C5,$F8,$7C,$5E,$58,$FF,$86,$6F,$8E,$1B,$33,$32
	.dc.b	$32,$7F,$B3,$3E,$A3,$F9,$DB,$79,$FC,$91,$88,$F6,$79,$D8,$9E,$FF
	.dc.b	$53,$27,$81,$0F,$84,$E5,$FE,$F7,$5C,$F2,$FF,$80,$3F,$3D,$FE,$FE
	.dc.b	$9F,$F6,$EB,$5F,$F0,$0B,$63,$E7,$76,$7C,$78,$C7,$07,$F7,$1D,$7F
	.dc.b	$D0,$17,$F6,$FD,$AD,$7C,$6C,$8C,$10,$EA,$FD,$1F,$88,$FF,$2D,$C7
	.dc.b	$FB,$B9,$AF,$F8,$49,$B9,$30,$FF,$00,$72,$FF,$9B,$2B,$87,$A7,$F6
	.dc.b	$E1,$E1,$D3,$B4,$49,$FF,$7D,$CE,$9B,$39,$7F,$C4,$3E,$7F,$DE,$BF
	.dc.b	$F0,$73,$63,$FC,$74,$D5,$F5,$3F,$E1,$27,$FC,$7D,$8F,$9D,$DF,$2F
	.dc.b	$E3,$E5,$F2,$14,$FC,$95,$F3,$3A,$46,$AB,$9E,$1F,$E1,$F8,$C4,$79
	.dc.b	$AF,$F3,$5F,$53,$FE,$32,$8D,$9F,$E4,$F7,$0F,$F8,$F1,$B8,$F9,$63
	.dc.b	$D3,$A7,$2F,$33,$F4,$9F,$53,$FE,$AC,$9E,$61,$C3,$CE,$37,$ED,$FE
	.dc.b	$AF,$EA,$7F,$07,$3F,$E4,$A6,$E1,$DE,$F8,$B7,$D2,$FD,$97,$D4,$BE
	.dc.b	$12,$9E,$28,$B8,$78,$A5,$F5,$3C,$0C,$C7,$F9,$62,$1F,$49,$FB,$DF
	.dc.b	$A9,$7D,$4D,$3B,$C7,$F0,$1F,$8B,$FF,$51,$DF,$B9,$AF,$F9,$89,$B9
	.dc.b	$FC,$7F,$91,$7A,$9C,$7E,$03,$F8,$54,$3E,$A6,$99,$C3,$A4,$E1,$F9
	.dc.b	$73,$B7,$27,$CE,$FA,$9F,$AB,$D3,$FE,$5D,$7F,$CF,$91,$9F,$77,$F8
	.dc.b	$C4,$6A,$3F,$D1,$8D,$4F,$07,$FC,$F6,$DC,$BF,$00,$EC,$79,$1D,$09
	.dc.b	$3E,$3F,$A2,$AF,$28,$B5,$CA,$7E,$57,$CB,$CA,$F9,$4F,$9B,$E4,$B1
	.dc.b	$1E,$2F,$A9,$F4,$98,$BF,$C9,$70,$2F,$35,$F0,$EB,$5A,$7F,$9F,$FE
	.dc.b	$A5,$E4,$FF,$39,$7E,$FD,$7E,$59,$AF,$FA,$69,$DB,$98,$E1,$C7,$23
	.dc.b	$07,$1F,$A9,$E8,$47,$A7,$83,$5E,$91,$74,$FD,$56,$1F,$C0,$FE,$A5
	.dc.b	$F3,$5F,$D3,$5B,$7F,$AA,$6B,$FE,$AE,$FA,$9F,$E1,$0F,$FA,$C2,$33
	.dc.b	$7F,$B5,$97,$CC,$FF,$57,$6D,$73,$BE,$C6,$66,$46,$4F,$DC,$CF,$F8
	.dc.b	$8D,$BE,$DF,$33,$A7,$7D,$8C,$90,$F3,$3E,$1E,$BF,$17,$54,$9E,$B1
	.dc.b	$97,$BC,$65,$F7,$47,$F6,$FF,$A9,$7D,$4D,$24,$FF,$0E,$F5,$BF,$35
	.dc.b	$3A,$8C,$7B,$5F,$DD,$E7,$FB,$EF,$2F,$FB,$06,$B5,$5F,$ED,$86,$EF
	.dc.b	$FC,$76,$B7,$3F,$EF,$16,$A9,$DD,$7F,$B2,$E7,$F4,$73,$E3,$EA,$F0
	.dc.b	$1E,$1E,$B9,$19,$21,$FE,$E8,$9E,$2D,$EB,$F4,$7D,$4C,$8F,$74,$7A
	.dc.b	$CC,$3C,$63,$0F,$F3,$0F,$A9,$FE,$73,$F0,$5F,$A5,$80,$FA,$9F,$F6
	.dc.b	$C3,$FE,$BA,$EE,$FF,$72,$46,$23,$FD,$61,$1F,$EE,$7A,$E6,$DF,$33
	.dc.b	$FD,$4C,$F0,$F8,$9B,$5D,$AF,$96,$1E,$18,$61,$C9,$D5,$70,$7C,$8D
	.dc.b	$AE,$37,$DE,$7C,$C7,$02,$9E,$BF,$17,$81,$C6,$B9,$7D,$4F,$5F,$A8
	.dc.b	$7D,$4F,$F6,$FC,$FC,$91,$8F,$8D,$E7,$9A,$F8,$E5,$DA,$B4,$FC,$43
	.dc.b	$E5,$B8,$BF,$75,$71,$9E,$18,$77,$CF,$E8,$C2,$F8,$46,$1B,$1F,$27
	.dc.b	$0F,$45,$71,$8E,$ED,$79,$36,$FC,$11,$E5,$33,$8F,$88,$8E,$63,$78
	.dc.b	$FC,$2A,$FE,$17,$3E,$36,$BE,$EF,$19,$F1,$BF,$EC,$2C,$BE,$1C,$FF
	.dc.b	$BC,$54,$EA,$3C,$51,$89,$FA,$FF,$51,$FD,$8B,$9B,$E6,$5E,$F1,$FA
	.dc.b	$6D,$D7,$E2,$37,$1F,$CC,$DF,$17,$AC,$65,$F9,$8D,$C7,$C1,$1F,$0E
	.dc.b	$78,$5F,$6C,$7C,$24,$9D,$75,$FF,$8C,$67,$51,$DA,$75,$1F,$9B,$AE
	.dc.b	$78,$F9,$A7,$2D,$AE,$79,$61,$F9,$53,$55,$72,$72,$77,$25,$F7,$C3
	.dc.b	$C9,$87,$58,$F2,$91,$FC,$6D,$73,$AF,$87,$7D,$F1,$F8,$77,$CF,$1C
	.dc.b	$65,$F1,$CF,$95,$87,$1A,$DE,$3D,$C9,$8F,$D2,$CB,$C6,$FA,$31,$F1
	.dc.b	$36,$7C,$73,$A8,$F1,$9E,$C8,$E6,$FC,$28,$D5,$78,$23,$91,$0D,$63
	.dc.b	$E8,$CB,$E5,$63,$E7,$CB,$D5,$18,$C3,$9C,$62,$FC,$07,$A6,$F9,$38
	.dc.b	$3A,$97,$0F,$13,$E8,$CB,$E4,$39,$AE,$91,$83,$E0,$C6,$F2,$F6,$E5
	.dc.b	$E5,$CB,$BD,$6E,$7E,$0E,$1D,$8E,$F5,$BB,$ED,$7F,$A8,$CB,$D6,$F2
	.dc.b	$BF,$9B,$95,$AF,$8D,$7C,$71,$F1,$E3,$D3,$1E,$35,$D0,$FC,$A9,$67
	.dc.b	$D3,$5C,$70,$F2,$E5,$F8,$4B,$E3,$5D,$23,$18,$F8,$70,$F8,$07,$CD
	.dc.b	$3A,$5F,$C2,$BE,$2F,$A2,$3E,$86,$1D,$DE,$18,$7C,$57,$F7,$FF,$53
	.dc.b	$F8,$99,$FC,$E1,$FE,$A2,$C3,$F8,$5F,$B3,$9F,$67,$4C,$8C,$9F,$C4
	.dc.b	$23,$17,$FC,$0D,$B8,$7E,$ED,$B8,$FE,$A9,$FF,$59,$76,$7F,$A7,$F1
	.dc.b	$46,$30,$FD,$7F,$BB,$71,$FE,$2B,$37,$E4,$DF,$19,$FE,$95,$BF,$F5
	.dc.b	$6C,$63,$E2,$35,$39,$B6,$5F,$EC,$1A,$D5,$F0,$AD,$4E,$55,$F2,$F5
	.dc.b	$F9,$BD,$3E,$A3,$AF,$D4,$DC,$39,$37,$2E,$F1,$FE,$04,$CB,$CF,$39
	.dc.b	$CB,$B3,$ED,$A7,$D4,$7F,$B9,$AF,$FD,$8F,$CC,$FA,$9F,$F3,$94,$7F
	.dc.b	$B0,$27,$55,$D8,$FE,$93,$3A,$70,$74,$F8,$2E,$D3,$C3,$1F,$F7,$17
	.dc.b	$C9,$3E,$17,$E8,$E3,$D6,$F6,$47,$FD,$B3,$3B,$E1,$C6,$3E,$1D,$7E
	.dc.b	$9F,$A1,$F4,$30,$FA,$2D,$C3,$E7,$36,$B8,$71,$7F,$9D,$27,$E9,$57
	.dc.b	$D8,$F6,$A4,$7D,$5F,$A9,$B7,$C2,$6E,$BF,$5E,$6D,$7F,$80,$BE,$99
	.dc.b	$E4,$7E,$26,$9F,$53,$F7,$AB,$53,$98,$C6,$3F,$E3,$AB,$E3,$5F,$EE
	.dc.b	$E6,$C7,$CF,$DF,$E4,$7A,$9F,$66,$73,$8F,$FB,$E5,$B7,$D8,$F9,$3C
	.dc.b	$3F,$E8,$87,$A6,$5F,$96,$FA,$99,$1F,$93,$F0,$6E,$7F,$1A,$6E,$3F
	.dc.b	$95,$9B,$5F,$8D,$E0,$3F,$A0,$FA,$0F,$D0,$B4,$FC,$F1,$70,$FA,$2D
	.dc.b	$BF,$47,$A3,$FA,$FE,$87,$D4,$FE,$8A,$BE,$14,$E6,$31,$7F,$8E,$D9
	.dc.b	$3F,$0C,$D7,$6D,$BE,$24,$AF,$CC,$61,$DA,$75,$1F,$8B,$1F,$53,$FB
	.dc.b	$F6,$D7,$EF,$97,$C6,$E6,$64,$E5,$E1,$18,$9E,$9F,$53,$CF,$E4,$3A
	.dc.b	$36,$FA,$36,$77,$1D,$E7,$F6,$DF,$50,$F8,$CF,$E2,$9C,$EF,$A9,$78
	.dc.b	$BA,$9C,$C6,$30,$FF,$42,$F1,$7D,$6C,$BF,$D0,$F8,$F3,$9C,$9F,$E5
	.dc.b	$A7,$11,$FB,$DF,$A9,$7D,$43,$EA,$7B,$5F,$0A,$E7,$96,$1E,$15,$B9
	.dc.b	$D4,$30,$EB,$8F,$8E,$1F,$52,$E7,$FE,$28,$DB,$E3,$C9,$F4,$1E,$71
	.dc.b	$9A,$FE,$1F,$2F,$6F,$F8,$E3,$6B,$C3,$F5,$0F,$93,$CD,$7F,$21,$8C
	.dc.b	$4E,$A7,$F9,$27,$D4,$BE,$A3,$51,$BF,$83,$95,$D3,$CA,$B9,$FF,$20
	.dc.b	$C7,$F8,$DB,$63,$F9,$77,$D4,$F3,$AE,$07,$C4,$71,$3B,$8C,$63,$E5
	.dc.b	$AF,$E6,$DE,$5B,$A7,$FA,$A5,$3E,$9D,$F0,$5A,$9C,$C6,$27,$E8,$D7
	.dc.b	$16,$CE,$BE,$A7,$84,$60,$99,$7F,$3D,$7B,$46,$A3,$F9,$EC,$3E,$1F
	.dc.b	$C7,$3F,$2A,$73,$18,$DF,$FA,$75,$6C,$FC,$19,$67,$B6,$2F,$A1,$DD
	.dc.b	$CA,$DB,$EC,$8C,$57,$F4,$EF,$A9,$7C,$A7,$C0,$3F,$AE,$38,$BF,$CB
	.dc.b	$1F,$1E,$32,$47,$38,$3E,$1B,$17,$D6,$AF,$4F,$EA,$6F,$E8,$56,$1F
	.dc.b	$F6,$0F,$C1,$7F,$14,$9D,$60,$FA,$7F,$ED,$D3,$9A,$FF,$7B,$9F,$66
	.dc.b	$33,$ED,$48,$F1,$F5,$5C,$FF,$B5,$DF,$17,$FD,$FE,$D9,$FC,$14,$EF
	.dc.b	$1F,$D6,$78,$9F,$7E,$FF,$8F,$FD,$43,$E1,$BF,$C0,$13,$9A,$FE,$09
	.dc.b	$6A,$D6,$1F,$B8,$AD,$3F,$ED,$E6,$E1,$FE,$C8,$F8,$2F,$EF,$93,$9E
	.dc.b	$5E,$A4,$7F,$C1,$71,$A9,$0A,$E6,$DF,$A3,$FE,$D1,$F5,$38,$94,$63
	.dc.b	$0F,$4B,$30,$FF,$0B,$E2,$EE,$C6,$3F,$E1,$2C,$3E,$95,$7A,$F8,$3F
	.dc.b	$C4,$13,$E2,$F8,$2F,$1F,$D4,$BE,$A3,$DD,$5F,$18,$97,$FC,$6A,$F8
	.dc.b	$0A,$E0,$FB,$30,$EE,$FF,$88,$8D,$C6,$9F,$81,$5F,$CC,$E3,$11,$AB
	.dc.b	$F0,$DF,$91,$F8,$F1,$30,$C3,$DE,$F9,$4E,$A3,$91,$9A,$DB,$C4,$F4
	.dc.b	$47,$66,$9C,$6B,$78,$7A,$B0,$F2,$CA,$73,$87,$82,$38,$1C,$8F,$6E
	.dc.b	$1E,$07,$C6,$60,$D1,$8B,$ED,$87,$A6,$3B,$1C,$E3,$2F,$6C,$3E,$54
	.dc.b	$62,$3D,$91,$8B,$F2,$46,$2B,$DD,$3E,$AC,$3B,$4F,$C3,$2B,$87,$EE
	.dc.b	$9F,$2F,$EA,$7F,$8F,$7A,$BC,$1B,$1E,$76,$BB,$EB,$FF,$42,$B9,$F9
	.dc.b	$2F,$F4,$0B,$79,$39,$46,$23,$B3,$5F,$BF,$48,$E3,$8F,$E2,$7E,$A5
	.dc.b	$F4,$BF,$84,$8C,$37,$4E,$35,$C1,$B8,$FE,$0A,$F8,$E3,$6F,$83,$FB
	.dc.b	$4C,$7C,$15,$F8,$86,$C7,$DD,$AE,$2D,$AE,$2A,$77,$8F,$E7,$DB,$87
	.dc.b	$E6,$AF,$8C,$FA,$2B,$B6,$3E,$77,$C4,$D5,$3C,$2B,$96,$1E,$0C,$3A
	.dc.b	$D7,$0C,$7D,$8C,$3F,$29,$1A,$7C,$D8,$75,$C3,$A4,$7F,$14,$6B,$AA
	.dc.b	$E4,$7F,$85,$E3,$E0,$B6,$FF,$6A,$FF,$86,$EF,$B4,$EB,$7F,$81,$E1
	.dc.b	$E9,$B8,$7F,$89,$27,$4F,$F2,$4F,$A9,$FF,$60,$BF,$E2,$8F,$A7,$FF
	.dc.b	$2E,$B7,$D8,$E4,$EF,$B6,$AA,$E2,$FF,$AC,$AB,$8B,$6B,$73,$AC,$FF
	.dc.b	$C6,$4B,$E6,$F2,$4A,$F1,$78,$75,$E9,$E3,$F8,$31,$FF,$08,$C7,$F8
	.dc.b	$FA,$F9,$36,$F5,$52,$30,$7F,$91,$3E,$A3,$F9,$EB,$7A,$FD,$71,$88
	.dc.b	$F5,$79,$B8,$9F,$07,$D4,$C9,$E0,$43,$E0,$39,$7F,$DA,$75,$CF,$4F
	.dc.b	$F2,$87,$CF,$FF,$9A,$27,$FC,$AA,$D7,$FD,$B2,$D8,$FF,$2B,$69,$FE
	.dc.b	$31,$9D,$FB,$1F,$07,$DA,$91,$82,$1F,$45,$CB,$D2,$3E,$BF,$C1,$7F
	.dc.b	$94,$E3,$FC,$CA,$D7,$FD,$D2,$DB,$FF,$33,$73,$FF,$99,$67,$7E,$67
	.dc.b	$E0,$21,$F0,$F4,$EB,$12,$7F,$CE,$33,$A6,$CE,$5F,$F7,$BF,$47,$D8
	.dc.b	$6F,$17,$E3,$23,$18,$7F,$9D,$FE,$A5,$E5,$66,$BF,$23,$0F,$90,$A6
	.dc.b	$70,$FF,$40,$E1,$CB,$AF,$FA,$DD,$6E,$7F,$D0,$AD,$7C,$27,$FA,$23
	.dc.b	$EA,$7F,$C6,$8D,$F3,$BF,$3F,$18,$86,$7E,$5D,$B3,$5F,$A4,$FA,$8F
	.dc.b	$F5,$45,$7F,$A4,$9A,$F0,$C7,$FB,$6E,$1C,$F0,$FF,$4A,$56,$9F,$F4
	.dc.b	$B7,$D4,$FF,$B0,$2F,$FD,$36,$D7,$91,$FE,$1C,$9E,$17,$FD,$E1,$FF
	.dc.b	$33,$47,$FA,$8B,$EA,$1F,$09,$FB,$86,$C7,$97,$C6,$C4,$E7,$EA,$64
	.dc.b	$F1,$8E,$1F,$4B,$FC,$03,$EA,$5F,$53,$4C,$D7,$FA,$C3,$E2,$FF,$C4
	.dc.b	$8D,$F4,$3F,$83,$C6,$23,$D2,$3E,$45,$EA,$EA,$BF,$86,$C3,$E7,$A9
	.dc.b	$9A,$E8,$7F,$AF,$A3,$4D,$8D,$3D,$BE,$A7,$E7,$70,$FF,$87,$1F,$F3
	.dc.b	$CC,$7F,$B1,$9A,$EF,$97,$F8,$EC,$70,$8F,$F2,$4B,$AE,$4F,$F6,$53
	.dc.b	$6B,$A5,$70,$F1,$A5,$F5,$79,$FD,$16,$23,$C9,$17,$0E,$73,$F1,$BE
	.dc.b	$6E,$57,$CB,$7C,$7F,$25,$88,$F6,$7D,$4F,$98,$C5,$FF,$B7,$78,$17
	.dc.b	$9A,$F7,$56,$B3,$FF,$1E,$FD,$43,$C8,$FE,$76,$B9,$69,$88,$97,$FE
	.dc.b	$EC,$9D,$36,$72,$FD,$C8,$7D,$4F,$42,$47,$1F,$0A,$BB,$8B,$A7,$FB
	.dc.b	$D6,$BF,$71,$3F,$E7,$8F,$A9,$7C,$D7,$F5,$06,$C7,$FB,$F9,$AF,$E3
	.dc.b	$3E,$A7,$F8,$3B,$99,$FE,$AF,$0F,$99,$A7,$38,$93,$FE,$86,$C1,$AF
	.dc.b	$F9,$8A,$F8,$4E,$63,$17,$C2,$3C,$1E,$C7,$62,$3B,$3F,$D2,$35,$FB
	.dc.b	$47,$66,$43,$FE,$31,$70,$D7,$D9,$DF,$EA,$EB,$F5,$F5,$C9,$DC,$65
	.dc.b	$E3,$19,$7E,$24,$7F,$78,$FA,$87,$D4,$D2,$4F,$8F,$D4,$FD,$02,$9D
	.dc.b	$B9,$F5,$FF,$C0,$33,$FD,$FB,$6F,$8F,$3F,$EA,$EA,$DB,$FB,$E6,$C6
	.dc.b	$A1,$3C,$76,$FF,$47,$B6,$39,$61,$DB,$AF,$BF,$5C,$9F,$34,$65,$F4
	.dc.b	$46,$5F,$E1,$B4,$9F,$3F,$5F,$A3,$EA,$65,$7B,$E3,$2F,$C1,$8C,$BC
	.dc.b	$A3,$FC,$3F,$F5,$3F,$D0,$7E,$03,$F5,$B3,$FE,$CC,$FA,$8F,$F6,$F3
	.dc.b	$FE,$BD,$EE,$FF,$54,$C6,$23,$FD,$35,$1F,$EA,$CA,$E4,$DF,$03,$FD
	.dc.b	$5C,$F0,$F8,$8B,$5D,$5A,$AB,$8B,$93,$C0,$53,$B9,$70,$F3,$BC,$6F
	.dc.b	$9E,$1E,$59,$CB,$FC,$AA,$78,$87,$11,$AB,$E3,$F5,$3E,$4F,$A8,$7D
	.dc.b	$4F,$EC,$2B,$BC,$63,$D4,$F2,$E3,$CE,$F8,$B8,$9D,$3E,$06,$D7,$92
	.dc.b	$7B,$5F,$5D,$3F,$DA,$B5,$A6,$C6,$6F,$EA,$57,$0C,$3B,$4E,$65,$7C
	.dc.b	$5D,$9A,$98,$F8,$E7,$86,$4B,$E9,$39,$AD,$E3,$E7,$8E,$27,$FB,$97
	.dc.b	$0F,$6B,$7C,$57,$C5,$FF,$B1,$27,$33,$EC,$9C,$BF,$0D,$B5,$F6,$BE
	.dc.b	$A3,$FB,$3F,$37,$C9,$9C,$EB,$C7,$1C,$1F,$42,$B7,$87,$38,$CB,$4F
	.dc.b	$BF,$1A,$C3,$EC,$DF,$49,$1F,$21,$C3,$6B,$B7,$57,$F0,$2A,$E3,$1D
	.dc.b	$9B,$3D,$A7,$58,$79,$E3,$C5,$3A,$3D,$D3,$A8,$F0,$E4,$C3,$C2,$DC
	.dc.b	$3B,$61,$E6,$A6,$3E,$B8,$ED,$8F,$56,$9B,$8D,$4F,$6C,$3A,$92,$FD
	.dc.b	$B9,$76,$9D,$45,$7B,$E5,$E7,$7E,$21,$EC,$C3,$BB,$F2,$B0,$EE,$EB
	.dc.b	$1F,$23,$D0,$F8,$D8,$75,$CB,$DD,$1C,$CE,$58,$77,$8C,$D7,$29,$E9
	.dc.b	$87,$B6,$B5,$97,$AE,$79,$1E,$8C,$3B,$46,$6F,$E8,$C6,$32,$F7,$65
	.dc.b	$E8,$CB,$BE,$5E,$CC,$7B,$9E,$4B,$E0,$F6,$AE,$18,$F5,$AE,$11,$8C
	.dc.b	$37,$8F,$82,$FC,$B8,$F6,$7E,$24,$64,$F7,$46,$63,$C2,$7A,$CE,$38
	.dc.b	$77,$7C,$86,$2B,$D7,$97,$C4,$7C,$38,$FC,$7C,$AD,$78,$23,$0F,$13
	.dc.b	$D9,$5C,$5C,$63,$E5,$9F,$89,$8F,$94,$F3,$65,$EF,$AD,$CF,$4A,$F5
	.dc.b	$E5,$E7,$9F,$54,$63,$2F,$3E,$52,$7C,$2D,$9F,$04,$7E,$37,$EA,$7F
	.dc.b	$25,$8F,$E3,$8F,$F5,$35,$7F,$03,$F5,$F3,$EB,$E9,$91,$93,$F8,$44
	.dc.b	$62,$FF,$82,$B7,$0F,$E2,$4D,$CF,$F5,$4F,$F8,$0B,$AF,$FB,$F6,$28
	.dc.b	$C6,$1F,$AF,$F6,$EE,$3F,$A3,$9B,$F2,$4F,$89,$FF,$4B,$DF,$F8,$A6
	.dc.b	$31,$F0,$BA,$9C,$DF,$FA,$8A,$73,$5B,$9C,$E5,$FE,$14,$9C,$D6,$E7
	.dc.b	$35,$FC,$72,$73,$5B,$97,$D4,$F7,$FA,$9F,$03,$6F,$B3,$70,$E8,$DB
	.dc.b	$FF,$0F,$7D,$47,$FB,$9A,$BF,$C8,$5C,$4F,$A9,$FF,$3A,$C7,$F9,$06
	.dc.b	$75,$87,$FA,$F2,$B5,$C1,$F1,$A1,$3A,$C3,$AF,$C0,$7C,$AF,$8F,$7C
	.dc.b	$1F,$CF,$CB,$A4,$F1,$BE,$11,$FC,$95,$B8,$2F,$8F,$0E,$EF,$E7,$79
	.dc.b	$DF,$3D,$B8,$7C,$D6,$DF,$CB,$6E,$5F,$25,$FA,$DE,$C4,$AF,$A9,$F5
	.dc.b	$36,$FE,$94,$DB,$FA,$13,$6B,$F9,$D7,$D3,$3C,$5F,$C3,$D3,$EA,$7E
	.dc.b	$CE,$3F,$D5,$2B,$57,$FE,$6F,$AD,$7B,$DF,$26,$73,$5F,$ED,$06,$CF
	.dc.b	$F9,$63,$6F,$93,$E9,$EF,$29,$8F,$F9,$D6,$F6,$FF,$62,$AD,$B9,$9F
	.dc.b	$CA,$7D,$4C,$8F,$C9,$78,$17,$2F,$C8,$4D,$CB,$F1,$D3,$6B,$F2,$5D
	.dc.b	$3F,$E9,$07,$F3,$5F,$41,$F9,$F7,$F4,$0D,$72,$7E,$7C,$C9,$E7,$F4
	.dc.b	$7F,$5B,$D0,$FA,$97,$77,$C3,$9C,$C6,$2B,$81,$34,$FC,$94,$7C,$E5
	.dc.b	$BA,$B9,$56,$F4,$FF,$40,$E3,$DF,$87,$38,$7E,$D4,$7D,$4F,$EF,$9B
	.dc.b	$7F,$BD,$5F,$17,$99,$93,$97,$84,$62,$7A,$7D,$4F,$3F,$99,$E8,$DB
	.dc.b	$E8,$D9,$DC,$77,$9F,$D6,$7D,$23,$E9,$BF,$55,$F5,$2F,$1B,$31,$86
	.dc.b	$DF,$FA,$DE,$B8,$3F,$EB,$DF,$A9,$7D,$43,$EA,$77,$9B,$87,$FD,$41
	.dc.b	$87,$08,$C6,$1F,$E1,$78,$C4,$FF,$AA,$B0,$E7,$7C,$A3,$FD,$51,$F5
	.dc.b	$2E,$7F,$E0,$8D,$C3,$FD,$95,$C7,$F2,$9F,$F6,$64,$67,$2E,$E7,$F8
	.dc.b	$EA,$7F,$D4,$93,$C3,$D9,$FE,$1D,$F5,$0F,$93,$FE,$21,$5A,$BE,$22
	.dc.b	$75,$7F,$ED,$96,$C6,$BE,$A5,$F5,$1D,$30,$EB,$F0,$53,$0E,$1E,$4D
	.dc.b	$C3,$AE,$3F,$C6,$5B,$1F,$C8,$7E,$A7,$9D,$70,$3F,$DD,$0E,$27,$71
	.dc.b	$8C,$7D,$75,$FC,$A7,$D1,$74,$7F,$50,$A7,$D3,$7F,$B4,$9F,$F7,$94
	.dc.b	$E9,$FF,$7C,$1A,$F8,$1E,$55,$23,$34,$FA,$9C,$65,$F1,$B0,$E4,$DB
	.dc.b	$F0,$56,$EF,$BC,$7F,$3D,$5F,$0D,$E5,$C3,$E5,$F2,$FF,$B9,$E7,$27
	.dc.b	$92,$32,$FF,$97,$CC,$E1,$F5,$9B,$3F,$EF,$47,$DF,$F4,$1D,$EE,$47
	.dc.b	$F9,$F2,$7B,$47,$9A,$73,$5F,$80,$FA,$97,$D4,$FA,$8F,$EA,$8E,$30
	.dc.b	$F8,$71,$82,$3F,$1B,$E1,$2D,$73,$9B,$8F,$97,$45,$62,$FF,$CF,$5F
	.dc.b	$01,$FC,$57,$0F,$CC,$C7,$D8,$FA,$7F,$EC,$93,$9C,$39,$46,$3C,$A9
	.dc.b	$5F,$5B,$A6,$E1,$F9,$AC,$38,$4F,$6A,$FA,$B5,$AC,$13,$FE,$8B,$F3
	.dc.b	$FE,$DC,$7F,$1E,$FA,$87,$C3,$7F,$75,$9C,$D7,$EB,$67,$2F,$CD,$39
	.dc.b	$71,$EC,$FE,$6B,$F0,$7F,$DC,$B4,$FE,$E5,$19,$9E,$11,$FD,$3E,$37
	.dc.b	$EB,$E6,$3F,$67,$83,$E8,$FF,$B2,$FD,$4F,$8A,$7F,$AB,$46,$2F,$CC
	.dc.b	$CC,$3F,$C2,$79,$3B,$B1,$7D,$38,$7D,$87,$F8,$45,$FF,$60,$FC,$17
	.dc.b	$FB,$0B,$EA,$5F,$53,$F1,$6B,$78,$FA,$2F,$FC,$19,$1C,$55,$F1,$AE
	.dc.b	$15,$B9,$EA,$2F,$AB,$FD,$BE,$FE,$9B,$9A,$FF,$5F,$3C,$27,$6F,$82
	.dc.b	$BB,$3A,$BA,$E4,$E4,$64,$A6,$67,$9B,$D6,$BC,$EE,$8C,$1E,$5A,$EA
	.dc.b	$E0,$E5,$1E,$67,$E0,$9D,$B5,$61,$EC,$BE,$F3,$A7,$D7,$3A,$BE,$B3
	.dc.b	$9B,$F1,$D7,$0C,$3D,$B7,$BA,$F3,$1B,$3C,$C6,$23,$78,$78,$6F,$CD
	.dc.b	$18,$C3,$BB,$E9,$AF,$F7,$63,$5F,$AE,$65,$D7,$D4,$FF,$1A,$F5,$AF
	.dc.b	$67,$F9,$EE,$BE,$14,$E5,$B1,$97,$ED,$C3,$E4,$BF,$B4,$47,$06,$BA
	.dc.b	$6B,$FE,$FF,$8E,$1C,$BE,$29,$E3,$AF,$E1,$3E,$A5,$F4,$AC,$3F,$07
	.dc.b	$18,$7F,$0B,$87,$27,$F0,$6D,$AE,$55,$C6,$7F,$0F,$5C,$5D,$DF,$83
	.dc.b	$5F,$23,$63,$C1,$3B,$BE,$71,$E0,$9D,$2C,$3C,$17,$DE,$F8,$37,$0F
	.dc.b	$16,$1F,$A0,$AD,$DF,$B5,$AA,$F8,$BE,$2A,$E3,$7E,$6C,$7C,$47,$C4
	.dc.b	$8C,$CE,$CC,$9E,$E8,$D3,$0C,$1F,$12,$3E,$34,$7E,$E2,$3C,$51,$8D
	.dc.b	$7D,$51,$FC,$1F,$0F,$3E,$1F,$2E,$F8,$3F,$EC,$99,$FE,$18,$D7,$FC
	.dc.b	$26,$EE,$BE,$A7,$D4,$FF,$97,$A7,$F8,$97,$D3,$FF,$95,$DB,$EC,$F8
	.dc.b	$5F,$C1,$4E,$DF,$F0,$EB,$70,$F4,$57,$19,$FF,$10,$B6,$3D,$93,$BB
	.dc.b	$F0,$FC,$D7,$5B,$4A,$30,$D8,$F2,$F8,$78,$9E,$18,$3C,$6A,$66,$BF
	.dc.b	$96,$5F,$17,$C7,$1A,$73,$5E,$1F,$A8,$FE,$89,$1F,$CC,$DA,$FF,$8E
	.dc.b	$9B,$97,$BB,$CE,$E1,$D5,$E2,$FA,$99,$7C,$63,$2F,$CC,$8C,$BC,$E3
	.dc.b	$E0,$7C,$F7,$F9,$A6,$3F,$A1,$35,$FF,$24,$B6,$FE,$6F,$6F,$0F,$5F
	.dc.b	$E0,$FB,$72,$F8,$C6,$5F,$B7,$19,$79,$C7,$CF,$F8,$9F,$F2,$C4,$7F
	.dc.b	$51,$6B,$FE,$58,$6D,$FE,$4F,$9B,$FC,$D5,$3B,$F4,$BE,$C4,$3C,$3A
	.dc.b	$75,$89,$3F,$D7,$67,$4D,$9C,$BF,$E6,$6F,$3F,$EB,$5F,$E8,$5B,$CD
	.dc.b	$FC,$25,$B3,$FC,$4F,$EA,$5E,$57,$E2,$67,$32,$97,$C8,$7E,$35,$B1
	.dc.b	$FD,$BE,$76,$70,$8C,$C7,$F9,$E1,$B8,$78,$F9,$7F,$2D,$F5,$3F,$E3
	.dc.b	$0A,$FE,$F0,$D7,$FC,$FE,$D9,$FD,$D5,$7C,$AC,$BF,$91,$4E,$5F,$E4
	.dc.b	$9F,$51,$FE,$AB,$BF,$F0,$13,$5E,$7E,$77,$E9,$63,$18,$7F,$05,$FA
	.dc.b	$9F,$F6,$0D,$7F,$82,$5A,$FF,$A4,$27,$77,$DB,$C4,$FD,$77,$D4,$BE
	.dc.b	$12,$9D,$62,$DF,$F8,$4E,$5F,$53,$C0,$CC,$7F,$A7,$21,$F4,$9F,$B9
	.dc.b	$FA,$97,$D4,$D3,$BC,$7E,$F7,$E2,$FF,$C5,$17,$FE,$1F,$6B,$FE,$A6
	.dc.b	$6E,$DF,$C8,$7E,$45,$EA,$71,$F4,$BF,$83,$43,$EA,$69,$9C,$3A,$4E
	.dc.b	$1F,$F1,$64,$ED,$C9,$FE,$2F,$FA,$9F,$A4,$FF,$5D,$7F,$CE,$EE,$7D
	.dc.b	$BF,$E2,$91,$A8,$34,$DB,$FE,$BE,$6E,$3F,$CE,$5D,$78,$DC,$89,$3E
	.dc.b	$1F,$A2,$AF,$18,$B5,$C6,$7E,$47,$CB,$CA,$F9,$0F,$93,$E4,$B1,$1E
	.dc.b	$1F,$A9,$F3,$98,$BF,$E9,$5B,$AF,$35,$F0,$6B,$5C,$FF,$DD,$7E,$A1
	.dc.b	$E4,$FF,$33,$7E,$ED,$7F,$95,$5A,$FF,$B7,$1D,$B9,$8E,$1C,$72,$30
	.dc.b	$71,$FA,$9E,$84,$7A,$78,$15,$E9,$17,$4F,$EC,$58,$7E,$F3,$EA,$5F
	.dc.b	$35,$FD,$1D,$B7,$FE,$69,$6B,$FE,$F0,$FA,$9F,$E0,$AE,$67,$FA,$74
	.dc.b	$3E,$66,$9C,$E2,$4F,$F9,$CE,$74,$D9,$CB,$D7,$DB,$E6,$47,$87,$7D
	.dc.b	$74,$73,$8F,$0C,$78,$BA,$FC,$7D,$52,$BC,$11,$97,$C5,$19,$7D,$91
	.dc.b	$FD,$9F,$EA,$5F,$53,$49,$3F,$E8,$BF,$57,$F3,$13,$A8,$C7,$B3,$FD
	.dc.b	$BE,$7F,$BA,$72,$FF,$8D,$6B,$54,$E1,$FF,$09,$46,$1F,$F0,$A5,$70
	.dc.b	$9F,$F2,$23,$54,$EE,$BF,$C8,$12,$EB,$F0,$75,$CB,$F3,$47,$8C,$C3
	.dc.b	$E1,$AE,$33,$98,$F6,$75,$FA,$7E,$A6,$46,$08,$71,$3C,$AE,$9C,$1F
	.dc.b	$E0,$78,$FF,$08,$7D,$4F,$F3,$9F,$81,$FD,$34,$FF,$97,$FE,$A3,$FD
	.dc.b	$AE,$FF,$AE,$BB,$BF,$CB,$51,$88,$FF,$5A,$C7,$F9,$76,$B9,$37,$CB
	.dc.b	$FF,$40,$BC,$3E,$1E,$D7,$56,$AA,$E2,$E4,$F0,$14,$EE,$5C,$3C,$EF
	.dc.b	$1B,$E7,$87,$96,$72,$FF,$21,$9E,$21,$C4,$6A,$F8,$FD,$47,$93,$EA
	.dc.b	$1F,$53,$FA,$AA,$EF,$18,$F5,$3C,$B8,$F3,$BE,$2E,$27,$4F,$81,$B5
	.dc.b	$E4,$9E,$D7,$D7,$4F,$F4,$35,$69,$B1,$9B,$FA,$95,$C3,$0E,$D3,$99
	.dc.b	$5F,$17,$66,$A6,$3E,$39,$E1,$92,$FA,$4E,$6B,$78,$F9,$E2,$E3,$E4
	.dc.b	$9B,$1D,$8E,$07,$FB,$67,$1F,$82,$FF,$BC,$17,$8A,$F8,$BF,$F6,$34
	.dc.b	$6A,$7E,$0C,$E5,$F9,$0D,$C3,$EF,$FD,$47,$F6,$4E,$3F,$9D,$39,$CB
	.dc.b	$8E,$38,$3E,$AD,$6F,$0E,$71,$96,$9F,$84,$8D,$61,$F7,$AF,$A4,$FF
	.dc.b	$B0,$71,$F4,$DF,$1A,$F1,$F5,$32,$C6,$5E,$2C,$3A,$D6,$A7,$31,$CF
	.dc.b	$1E,$AF,$D8,$6E,$5E,$2B,$E5,$7E,$3C,$78,$DA,$B4,$F3,$8C,$E5,$E6
	.dc.b	$8F,$66,$5E,$57,$C7,$87,$42,$3F,$ED,$DB,$E6,$F8,$EF,$9E,$3E,$4B
	.dc.b	$E5,$8F,$8E,$B9,$9E,$13,$BE,$5C,$2F,$CF,$87,$4C,$BC,$97,$C4,$99
	.dc.b	$7F,$B9,$67,$59,$7C,$4A,$DC,$F9,$B1,$F8,$19,$78,$A3,$13,$A9,$F1
	.dc.b	$61,$F3,$DF,$29,$EB,$2E,$1D,$B1,$F3,$47,$A0,$DD,$73,$78,$57,$33
	.dc.b	$CE,$F6,$BF,$25,$F3,$8E,$38,$FC,$7C,$BA,$DF,$09,$F4,$C6,$2F,$8E
	.dc.b	$5E,$B9,$E8,$43,$C3,$1F,$16,$35,$18,$35,$8F,$9F,$1F,$16,$1D,$E3
	.dc.b	$15,$C1,$F6,$13,$1F,$35,$FA,$32,$FA,$F5,$B9,$FA,$44,$8F,$D5,$C6
	.dc.b	$5F,$54,$FE,$26,$30,$F3,$3E,$65,$F8,$EB,$94,$63,$1F,$B3,$18,$9F
	.dc.b	$9D,$1F,$13,$2F,$19,$2B,$81,$F9,$7F,$A9,$F8,$79,$FE,$60,$FF,$4F
	.dc.b	$CF,$F0,$0F,$5F,$3E,$BE,$99,$19,$3F,$82,$C6,$2F,$F8,$13,$71,$FE
	.dc.b	$1E,$DC,$7F,$2C,$FE,$AF,$AF,$F5,$95,$FB,$38,$C5,$FE,$6F,$DC,$B3
	.dc.b	$FA,$D9,$BF,$24,$F8,$AF,$F4,$95,$AB,$FB,$C4,$E7,$E3,$F7,$5A,$9C
	.dc.b	$A8,$CF,$AF,$A6,$DF,$3C,$BF,$80,$61,$CE,$30,$77,$7C,$5F,$51,$DB
	.dc.b	$EA,$7B,$B6,$77,$92,$B7,$18,$7D,$55,$C2,$31,$92,$3E,$0D,$3E,$A3
	.dc.b	$FD,$CD,$5D,$39,$5F,$53,$FE,$75,$8E,$F5,$BF,$80,$F9,$35,$AE,$1E
	.dc.b	$78,$72,$AF,$F0,$DB,$75,$FE,$3D,$F2,$2F,$84,$71,$70,$CA,$67,$DB
	.dc.b	$0E,$0F,$D1,$BE,$15,$AC,$7E,$7F,$9F,$F4,$B0,$FA,$6D,$BF,$7B,$67
	.dc.b	$71,$F0,$A7,$EC,$FB,$32,$3E,$2F,$D4,$DA,$FB,$73,$74,$FA,$FC,$2A
	.dc.b	$C5,$FE,$16,$DF,$4C,$F1,$BF,$1B,$83,$EA,$7F,$0C,$FE,$52,$75,$7F
	.dc.b	$6A,$35,$D1,$F5,$27,$38,$7F,$91,$1B,$3F,$2B,$87,$FD,$0D,$7F,$52
	.dc.b	$72,$F8,$DB,$5F,$99,$6F,$07,$D1,$F4,$BB,$67,$F9,$AF,$A9,$98,$7E
	.dc.b	$1B,$C0,$B8,$FE,$72,$6E,$7F,$0A,$6D,$7E,$5F,$7F,$ED,$CF,$E3,$BE
	.dc.b	$83,$F4,$AF,$E9,$9A,$E4,$FC,$99,$93,$D5,$E8,$7E,$33,$EA,$7F,$55
	.dc.b	$18,$3F,$28,$E3,$2F,$89,$3F,$1C,$E1,$C1,$6B,$84,$E9,$E3,$19,$8F
	.dc.b	$DD,$8F,$A9,$FE,$02,$D9,$FD,$52,$F8,$CC,$CC,$9C,$BC,$23,$13,$D3
	.dc.b	$EA,$79,$FC,$77,$46,$DF,$46,$CE,$E3,$BC,$FE,$D3,$E9,$9F,$51,$FC
	.dc.b	$5B,$91,$F5,$2F,$1B,$53,$98,$C6,$1F,$B6,$BE,$2D,$9F,$E3,$FF,$50
	.dc.b	$FA,$97,$D4,$F2,$AD,$13,$4F,$F4,$B5,$6E,$75,$1F,$E9,$6F,$A9,$72
	.dc.b	$7F,$0A,$6D,$FF,$06,$F1,$7E,$43,$FC,$22,$33,$5F,$E9,$E9,$FF,$58
	.dc.b	$4E,$FD,$AF,$E2,$7F,$50,$F9,$6F,$E2,$B5,$AB,$FF,$52,$CE,$6F,$F8
	.dc.b	$9B,$63,$FD,$53,$F5,$2F,$A8,$E7,$87,$4F,$82,$98,$70,$F2,$6E,$1D
	.dc.b	$1F,$09,$FC,$93,$EA,$79,$4E,$CF,$E4,$2E,$27,$84,$63,$0E,$2F,$F2
	.dc.b	$6F,$3D,$D1,$FD,$3A,$1F,$4E,$F8,$2F,$E6,$35,$A9,$FE,$76,$6A,$33
	.dc.b	$52,$33,$4F,$A9,$C6,$3F,$EC,$8C,$39,$36,$FC,$15,$BB,$F2,$47,$F4
	.dc.b	$05,$F0,$DE,$5C,$3E,$57,$8B,$FD,$A2,$72,$79,$23,$2F,$F1,$F3,$38
	.dc.b	$7F,$B6,$1B,$3F,$E5,$B7,$DF,$F4,$3D,$EE,$47,$F4,$A3,$67,$F6,$D3
	.dc.b	$66,$5F,$EA,$BF,$52,$F9,$3F,$59,$82,$CF,$F5,$48,$D3,$63,$E0,$FC
	.dc.b	$26,$E3,$04,$C7,$FD,$E4,$FD,$48,$D4,$7F,$BC,$BE,$03,$F8,$76,$5E
	.dc.b	$BD,$5F,$4F,$FD,$A2,$73,$87,$A6,$72,$FF,$66,$8C,$DF,$F8,$25,$B1
	.dc.b	$F8,$2E,$8E,$5E,$DC,$8E,$0F,$D8,$8C,$53,$1F,$F0,$DF,$89,$F7,$9F
	.dc.b	$F4,$44,$BE,$A1,$F0,$DF,$DF,$2B,$58,$FF,$86,$6B,$53,$9E,$39,$3F
	.dc.b	$DF,$FE,$0B,$FB,$7E,$5F,$DB,$E7,$2F,$86,$7F,$AF,$61,$D2,$3F,$D8
	.dc.b	$D5,$C1,$46,$5B,$F4,$1F,$D8,$BE,$A7,$12,$8C,$61,$E7,$66,$1F,$E1
	.dc.b	$1C,$DD,$98,$C7,$C4,$7F,$80,$F8,$7D,$C7,$F8,$69,$FF,$1B,$7C,$0F
	.dc.b	$F8,$DF,$EA,$5F,$53,$E8,$AD,$E3,$FE,$1A,$D3,$EB,$E1,$D0,$FE,$F5
	.dc.b	$87,$1B,$E1,$5F,$E2,$09,$EA,$30,$EA,$F8,$E3,$83,$FE,$03,$15,$87
	.dc.b	$E2,$E1,$FE,$BF,$AE,$31,$E0,$3A,$5F,$81,$A7,$3A,$E5,$87,$A2,$B8
	.dc.b	$CF,$91,$E7,$19,$AF,$14,$F8,$8E,$67,$8B,$0F,$23,$C6,$3C,$15,$D2
	.dc.b	$38,$3D,$30,$F4,$4F,$4C,$3C,$71,$8A,$F3,$1D,$B0,$F1,$38,$BE,$58
	.dc.b	$78,$23,$AE,$1E,$03,$A3,$C6,$FA,$3F,$16,$B8,$BE,$A3,$E9,$9E,$E3
	.dc.b	$DC,$5C,$30,$E8,$FF,$1B,$7D,$4F,$F2,$0F,$53,$83,$63,$E3,$B5,$FB
	.dc.b	$4D,$78,$FC,$8F,$EC,$DD,$B5,$DB,$5D,$F2,$76,$8E,$18,$7E,$03,$EA
	.dc.b	$5F,$4B,$FE,$49,$8C,$37,$4E,$38,$7D,$8C,$58,$72,$9F,$F2,$86,$1C
	.dc.b	$AB,$F1,$CD,$78,$4E,$DF,$F4,$D5,$72,$C7,$78,$77,$9F,$F4,$F5,$F1
	.dc.b	$C7,$F2,$B7,$B9,$E7,$7C,$18,$60,$93,$EA,$72,$F2,$C3,$C5,$0A,$E9
	.dc.b	$3B,$8F,$CA,$DF,$87,$1F,$83,$3C,$70,$E7,$1F,$03,$1F,$45,$F3,$C7
	.dc.b	$AE,$1F,$06,$FF,$D8,$2D,$D7,$85,$7F,$A6,$B2,$F0,$E3,$88,$FE,$05
	.dc.b	$5F,$E9,$47,$FD,$3E,$D7,$38,$FF,$9F,$F0,$FF,$66,$FD,$47,$FA,$FE
	.dc.b	$7F,$D4,$BF,$52,$F6,$35,$9F,$FA,$16,$31,$3F,$EA,$AA,$E0,$FF,$11
	.dc.b	$6D,$FF,$B6,$27,$73,$FC,$55,$B1,$E0,$EA,$EB,$3F,$ED,$5A,$7C,$D6
	.dc.b	$21,$18,$7C,$11,$9B,$78,$5B,$AD,$57,$7F,$17,$46,$6E,$47,$BA,$32
	.dc.b	$F4,$3F,$D8,$BF,$53,$FC,$F5,$BD,$3E,$38,$C4,$7A,$FC,$DC,$4F,$8F
	.dc.b	$EA,$64,$F0,$21,$F0,$5C,$BF,$CD,$EB,$9E,$BF,$EF,$EF,$9E,$FF,$32
	.dc.b	$4F,$FB,$59,$AF,$F3,$D6,$C7,$CD,$ED,$F8,$D1,$8E,$0F,$AD,$ED,$7C
	.dc.b	$53,$3E,$CC,$C3,$EE,$C6,$5E,$D1,$F3,$FE,$0F,$FC,$B1,$1F,$EE,$66
	.dc.b	$BF,$D3,$5B,$8A,$FF,$D9,$1C,$BF,$E6,$AA,$E1,$E8,$7D,$98,$7C,$3D
	.dc.b	$3B,$44,$9F,$F7,$9C,$E9,$B3,$97,$FA,$E7,$A3,$F7,$30,$F0,$40,$FF
	.dc.b	$78,$F1,$FE,$1E,$75,$8F,$FB,$6F,$EA,$5E,$4E,$6B,$F1,$90,$F8,$FA
	.dc.b	$67,$0E,$98,$F6,$7A,$98,$C3,$A4,$F0,$8C,$1E,$3C,$7C,$2F,$97,$EA
	.dc.b	$5E,$6F,$E6,$DB,$1C,$03,$87,$FB,$D6,$3F,$3B,$0F,$F7,$A4,$E7,$1F
	.dc.b	$27,$D4,$7F,$AB,$2F,$93,$5F,$F0,$14,$EE,$BF,$DF,$93,$9A,$F6,$7D
	.dc.b	$4F,$FB,$0E,$FC,$4D,$7F,$C1,$53,$BA,$E3,$E9,$7E,$CF,$EA,$1F,$09
	.dc.b	$4F,$DB,$57,$F9,$CA,$FD,$93,$A6,$FD,$4F,$13,$83,$AF,$A4,$FD,$F7
	.dc.b	$D4,$BE,$A6,$9D,$E3,$F8,$17,$C5,$7F,$8A,$AB,$E4,$35,$FF,$10,$37
	.dc.b	$6F,$B1,$F2,$2F,$57,$8F,$A7,$FC,$2E,$1F,$53,$4C,$E1,$D2,$70,$FD
	.dc.b	$59,$DB,$93,$EC,$7D,$4F,$CD,$7F,$24,$FF,$9F,$5C,$FB,$7F,$C6,$63
	.dc.b	$50,$69,$B7,$FC,$70,$DC,$7F,$20,$EB,$C7,$E4,$49,$F0,$FD,$15,$78
	.dc.b	$C5,$AE,$33,$F2,$3E,$5E,$57,$C8,$7C,$9F,$23,$88,$F0,$FD,$4F,$9C
	.dc.b	$C5,$FE,$87,$75,$E6,$BE,$0D,$6B,$A3,$F6,$50,$FA,$87,$91,$FC,$E5
	.dc.b	$71,$D3,$11,$27,$F4,$53,$A6,$CE,$5F,$F2,$E4,$3E,$A7,$A1,$23,$8F
	.dc.b	$85,$5D,$C5,$D3,$F7,$55,$FB,$49,$7D,$4B,$E6,$BF,$A7,$36,$3F,$68
	.dc.b	$D7,$FC,$D9,$F5,$3F,$C2,$9C,$CF,$F5,$58,$7C,$C5,$39,$C4,$9F,$DF
	.dc.b	$4E,$9B,$39,$7A,$FB,$7C,$C8,$F0,$F0,$6B,$A3,$9C,$78,$63,$C5,$D7
	.dc.b	$E3,$EA,$95,$E0,$8C,$BE,$28,$CB,$EC,$8F,$ED,$DF,$50,$FA,$9A,$49
	.dc.b	$FE,$2D,$EA,$FE,$6A,$75,$18,$F6,$BF,$BB,$4F,$F7,$CE,$5F,$C1,$D6
	.dc.b	$AB,$FD,$28,$DE,$1F,$E5,$95,$B9,$FE,$74,$D5,$3B,$AF,$C6,$CB,$AF
	.dc.b	$BF,$5C,$BF,$2C,$78,$8C,$3F,$EA,$39,$4E,$63,$D7,$D7,$E8,$FA,$99
	.dc.b	$18,$21,$C5,$CB,$CE,$3F,$C2,$9F,$53,$FC,$DF,$E0,$BF,$4B,$3F,$D4
	.dc.b	$FE,$A3,$FD,$B2,$FF,$AE,$BB,$BF,$4B,$18,$8F,$F5,$5C,$7E,$A2,$B9
	.dc.b	$37,$CD,$FF,$5C,$BC,$3E,$22,$D7,$56,$AA,$E2,$E4,$F0,$14,$EE,$5C
	.dc.b	$3C,$EF,$1B,$E7,$87,$96,$72,$FF,$20,$9E,$21,$C4,$6A,$F8,$FD,$47
	.dc.b	$93,$EA,$1F,$53,$FA,$7A,$EF,$18,$F5,$3C,$B8,$F3,$BE,$2E,$27,$4F
	.dc.b	$81,$B5,$E4,$9E,$D7,$D7,$4F,$E2,$35,$A6,$C6,$6F,$EA,$57,$0C,$3B
	.dc.b	$4E,$65,$7C,$5D,$9A,$98,$F8,$E7,$86,$4B,$E9,$39,$AD,$E3,$E7,$8B
	.dc.b	$8F,$92,$6C,$76,$38,$12,$7E,$3E,$1D,$AF,$AC,$FF,$BC,$F1,$F8,$AD
	.dc.b	$F1,$5F,$19,$FE,$C9,$9C,$CF,$C4,$9C,$BF,$2D,$B8,$7E,$0B,$EA,$3F
	.dc.b	$B3,$72,$FD,$19,$CE,$9C,$71,$C1,$F6,$2B,$78,$73,$8C,$B4,$FC,$3C
	.dc.b	$6B,$0F,$C0,$DF,$49,$FF,$1C,$62,$8C,$47,$D2,$6D,$71,$C3,$C5,$3D
	.dc.b	$6F,$9E,$1E,$3E,$AF,$F2,$1C,$62,$75,$5C,$23,$DD,$3A,$8C,$12,$7C
	.dc.b	$78,$78,$27,$51,$8B,$F6,$DF,$53,$9E,$58,$C3,$95,$7A,$6B,$C9,$39
	.dc.b	$8E,$47,$90,$FF,$2C,$63,$ED,$73,$7D,$6F,$9C,$E6,$7A,$1C,$4B,$3E
	.dc.b	$27,$19,$7C,$BC,$3C,$10,$AE,$53,$98,$DE,$5E,$FC,$77,$39,$C3,$C5
	.dc.b	$5E,$6C,$BE,$1B,$E0,$CB,$DF,$5B,$7F,$C0,$38,$F6,$C7,$ED,$CE,$6B
	.dc.b	$DB,$5A,$BE,$73,$99,$E8,$68,$F0,$63,$E4,$BE,$67,$9E,$3D,$39,$7B
	.dc.b	$B1,$F0,$3B,$3E,$8C,$68,$F9,$11,$A7,$19,$7C,$93,$DE,$F8,$89,$1B
	.dc.b	$8C,$E1,$E7,$8C,$65,$F0,$B0,$F2,$56,$9F,$3E,$1E,$28,$C3,$E7,$CB
	.dc.b	$E0,$CE,$B1,$F2,$C6,$27,$D5,$1F,$AA,$8C,$3E,$8B,$F9,$59,$73,$24
	.dc.b	$72,$7C,$11,$F5,$3E,$A7,$EC,$E5,$F8,$B3,$FD,$41,$3F,$BB,$F5,$F3
	.dc.b	$EB,$E9,$91,$93,$F8,$04,$63,$0F,$DE,$B7,$4F,$E1,$4D,$C7,$F0,$6F
	.dc.b	$FA,$DB,$AF,$FD,$55,$8A,$31,$87,$E2,$3D,$CB,$8F,$E9,$26,$FC,$93
	.dc.b	$E2,$FF,$D1,$F7,$FE,$AE,$8C,$7C,$3E,$A7,$36,$CB,$FD,$8B,$5A,$BE
	.dc.b	$15,$A9,$CA,$BE,$5E,$C7,$37,$A7,$D4,$F5,$FA,$9B,$87,$26,$E5,$DE
	.dc.b	$3F,$DA,$59,$79,$E7,$39,$76,$7D,$B4,$FA,$8F,$F7,$1D,$FF,$B2,$39
	.dc.b	$1F,$53,$FE,$6E,$8F,$F6,$24,$EA,$BB,$1F,$CF,$A7,$4E,$0E,$9F,$01
	.dc.b	$DA,$78,$63,$FE,$E6,$F9,$17,$C2,$7D,$1C,$7A,$DE,$C8,$FF,$B6,$A7
	.dc.b	$7C,$38,$C7,$C3,$AF,$D3,$F3,$FE,$86,$1F,$45,$B8,$7C,$E6,$D7,$0E
	.dc.b	$2F,$F3,$74,$FD,$2A,$FB,$1E,$CC,$8F,$AB,$F5,$36,$F8,$4D,$D7,$EB
	.dc.b	$CD,$AF,$F4,$7F,$D4,$3C,$6F,$C4,$D3,$EA,$7E,$F5,$6A,$73,$18,$C4
	.dc.b	$2F,$95,$7F,$BC,$DB,$1F,$43,$7F,$93,$E9,$7D,$A9,$CE,$2B,$EC,$71
	.dc.b	$C7,$F3,$77,$C2,$30,$7E,$4B,$EA,$7F,$30,$7E,$67,$C1,$B1,$92,$65
	.dc.b	$F2,$63,$2F,$CB,$8F,$D0,$7D,$07,$E8,$5A,$7E,$78,$B1,$B7,$F4,$31
	.dc.b	$E9,$F4,$3F,$5F,$D0,$FA,$9F,$D1,$57,$C1,$9C,$C6,$2F,$F1,$DA,$A7
	.dc.b	$95,$7E,$66,$BE,$65,$7C,$B9,$CB,$C1,$E7,$DB,$8D,$BF,$6B,$2F,$A9
	.dc.b	$FD,$FB,$70,$FD,$F2,$F8,$CC,$CC,$9C,$BC,$23,$13,$F9,$61,$F5,$3D
	.dc.b	$3E,$3B,$AB,$6F,$AB,$67,$83,$70,$FB,$91,$F8,$6F,$A8,$7D,$47,$DC
	.dc.b	$FA,$97,$8B,$FB,$C9,$CE,$3F,$A9,$9C,$BF,$8E,$C3,$9E,$0A,$DB,$FB
	.dc.b	$96,$C7,$83,$EA,$1F,$52,$FA,$9E,$D9,$CC,$7A,$1F,$EF,$0C,$38,$BF
	.dc.b	$9C,$BE,$35,$BC,$7F,$83,$7D,$4B,$9B,$F8,$9B,$6F,$9C,$FE,$93,$9B
	.dc.b	$E8,$46,$21,$3B,$7F,$88,$4F,$FA,$CE,$74,$D8,$FE,$29,$F5,$2F,$96
	.dc.b	$FE,$43,$39,$DB,$E1,$CE,$6F,$B3,$6F,$F8,$E7,$D4,$BE,$A3,$F9,$4C
	.dc.b	$E6,$F8,$FC,$16,$0C,$F9,$17,$1E,$6E,$1F,$57,$D4,$F3,$72,$63,$0D
	.dc.b	$65,$FC,$AE,$74,$FA,$AB,$F9,$CF,$92,$E8,$FE,$AF,$83,$E9,$DF,$03
	.dc.b	$A9,$CC,$62,$7E,$55,$71,$6C,$EB,$EA,$78,$46,$09,$8F,$F4,$07,$B4
	.dc.b	$6A,$3F,$A0,$43,$E1,$BC,$73,$F1,$63,$F8,$EC,$FF,$15,$9C,$F1,$7F
	.dc.b	$52,$9E,$07,$DD,$2C,$F7,$C5,$F4,$1B,$B9,$5B,$7D,$B1,$8A,$FE,$A5
	.dc.b	$F5,$2F,$93,$F8,$27,$F6,$07,$17,$F8,$E3,$E4,$46,$48,$E7,$07,$C1
	.dc.b	$E2,$FA,$D5,$E3,$FD,$AD,$FD,$1A,$C3,$F0,$3E,$07,$F8,$C4,$EB,$07
	.dc.b	$D3,$FF,$72,$9C,$D7,$F4,$23,$ED,$46,$7D,$99,$1E,$3E,$9B,$9F,$F6
	.dc.b	$FB,$E2,$FF,$4A,$6C,$FE,$0E,$77,$8F,$F0,$0F,$33,$F0,$17,$FD,$EA
	.dc.b	$5F,$50,$F8,$5F,$F0,$2D,$6A,$FF,$88,$D6,$A7,$3C,$93,$1F,$F0,$47
	.dc.b	$C0,$FF,$7E,$7E,$FE,$1C,$5B,$3B,$CB,$C1,$5B,$7F,$B0,$D6,$DB,$F4
	.dc.b	$5F,$DA,$BE,$A7,$10,$94,$E6,$FC,$EC,$C3,$FC,$35,$A3,$B7,$18,$7F
	.dc.b	$85,$70,$FA,$35,$FC,$2D,$B1,$FE,$19,$3F,$B9,$FC,$0F,$F7,$4F,$A9
	.dc.b	$7D,$47,$B3,$8F,$E2,$93,$0F,$E2,$97,$C0,$57,$07,$DB,$3F,$E3,$07
	.dc.b	$51,$97,$D9,$1F,$CF,$A3,$11,$AA,$F0,$57,$8D,$FD,$5C,$64,$EB,$19
	.dc.b	$73,$75,$EB,$15,$E0,$60,$F3,$CF,$45,$3C,$70,$F8,$B7,$E4,$9C,$9C
	.dc.b	$0F,$09,$4F,$DF,$CF,$1C,$3D,$97,$D6,$31,$87,$8E,$B9,$DF,$C3,$C3
	.dc.b	$BD,$EE,$7A,$46,$1F,$69,$8C,$3A,$B8,$AE,$51,$88,$F1,$9B,$BF,$06
	.dc.b	$1E,$F8,$C9,$5C,$3F,$6C,$F9,$FF,$53,$FC,$77,$DC,$E0,$D8,$F9,$0D
	.dc.b	$77,$D7,$FE,$87,$73,$F2,$3F,$E8,$36,$F2,$F2,$8C,$47,$66,$BF,$CA
	.dc.b	$29,$1C,$73,$FC,$47,$D4,$BE,$97,$F9,$6C,$61,$BB,$F1,$AE,$0D,$C7
	.dc.b	$F0,$97,$C7,$1B,$7C,$1F,$F2,$9E,$3E,$0A,$FC,$53,$63,$EF,$57,$16
	.dc.b	$D7,$15,$3B,$C7,$F3,$CD,$C3,$FA,$1D,$F1,$9F,$45,$76,$C7,$CE,$F8
	.dc.b	$9A,$A7,$85,$72,$C3,$C1,$87,$5A,$E1,$8F,$B1,$87,$E5,$63,$4F,$9B
	.dc.b	$0E,$B8,$74,$8F,$F3,$9B,$5D,$57,$23,$FC,$53,$1F,$05,$B7,$FD,$79
	.dc.b	$FF,$16,$DF,$69,$D6,$BF,$03,$C3,$D3,$70,$FF,$0E,$4E,$9F,$F4,$1F
	.dc.b	$D4,$7F,$B0,$DF,$F1,$07,$D4,$FF,$97,$DB,$EB,$F2,$77,$DB,$55,$71
	.dc.b	$7F,$D6,$95,$C5,$B5,$B9,$D6,$7F,$E4,$45,$F3,$59,$25,$78,$BC,$3A
	.dc.b	$F4,$F1,$FC,$18,$FF,$84,$23,$FC,$6D,$7C,$9B,$7A,$A9,$18,$3F,$C7
	.dc.b	$9F,$51,$FC,$F1,$BD,$7E,$B8,$C4,$7A,$BE,$57,$13,$E0,$FA,$99,$3C
	.dc.b	$08,$7C,$07,$2F,$FB,$56,$B9,$E9,$FE,$62,$F9,$EF,$F3,$4C,$FF,$93
	.dc.b	$5A,$FF,$B6,$DB,$1F,$E6,$8D,$3F,$C6,$13,$BF,$63,$E0,$FB,$52,$30
	.dc.b	$43,$E8,$B9,$7A,$47,$D7,$F8,$8F,$F2,$A4,$7F,$97,$5A,$FF,$BA,$DB
	.dc.b	$7F,$E7,$8E,$7F,$F3,$44,$EF,$CC,$FC,$04,$3E,$1E,$9D,$62,$4F,$F9
	.dc.b	$A6,$74,$D9,$CB,$FE,$F9,$E8,$FB,$0D,$E2,$FC,$64,$63,$0F,$F4,$4F
	.dc.b	$D4,$BC,$AC,$D7,$E4,$61,$F2,$14,$CE,$1F,$E7,$9C,$39,$75,$FF,$91
	.dc.b	$2B,$73,$FE,$7D,$6B,$E1,$3F,$D0,$1F,$53,$FE,$32,$6F,$9D,$F9,$F8
	.dc.b	$C4,$33,$F2,$ED,$9A,$FD,$27,$D4,$7F,$AA,$2B,$FD,$14,$D7,$86,$3F
	.dc.b	$E5,$6C,$39,$E1,$FE,$A6,$AD,$3F,$EA,$8F,$A9,$FF,$61,$5F,$FA,$4D
	.dc.b	$AF,$23,$FD,$F7,$3C,$2F,$FC,$C2,$FF,$99,$63,$FD,$35,$F5,$2F,$84
	.dc.b	$FD,$C3,$63,$CB,$E3,$62,$73,$F5,$32,$78,$C7,$0F,$A4,$FE,$01,$F5
	.dc.b	$2F,$A9,$A6,$6B,$FD,$51,$F1,$7F,$E2,$36,$FA,$1F,$C1,$E3,$11,$E9
	.dc.b	$1F,$22,$F5,$75,$5F,$C3,$61,$F5,$34,$CD,$74,$3F,$D6,$D1,$A6,$C6
	.dc.b	$9E,$DF,$53,$F3,$B8,$7F,$D0,$AF,$F9,$DE,$3F,$D7,$CD,$77,$CB,$FC
	.dc.b	$76,$38,$40,$3B,$E4,$FF,$6E,$B6,$BA,$D7,$1F,$1A,$5F,$67,$A7,$D1
	.dc.b	$62,$3C,$B1,$70,$E7,$3F,$1F,$E6,$E5,$7C,$C7,$C9,$F2,$58,$8F,$6F
	.dc.b	$D4,$F9,$CC,$5F,$FB,$C3,$81,$79,$AF,$7D,$6B,$3F,$F5,$3F,$D4,$3C
	.dc.b	$8F,$E7,$8B,$96,$98,$89,$7F,$EE,$59,$D3,$67,$2F,$DD,$87,$D4,$F4
	.dc.b	$24,$71,$F0,$AB,$B8,$BA,$31,$F2,$D7,$F4,$FF,$A9,$7C,$D5,$26,$1F
	.dc.b	$EF,$7F,$A9,$FE,$0C,$FF,$BE,$67,$6E,$7E,$67,$FA,$C4,$9E,$31,$F9
	.dc.b	$39,$DB,$F9,$59,$D3,$C4,$E9,$5C,$BD,$89,$1E,$17,$AF,$7E,$23,$94
	.dc.b	$5C,$FD,$91,$F8,$BB,$FD,$B3,$DF,$AF,$CB,$D7,$23,$D9,$1E,$83,$0F
	.dc.b	$96,$30,$FE,$56,$BA,$AF,$A8,$7D,$4F,$F7,$76,$C6,$F7,$FF,$73,$C6
	.dc.b	$23,$F6,$9E,$A7,$E8,$1E,$55,$FD,$FE,$3F,$BE,$5F,$66,$DF,$29,$EC
	.dc.b	$6F,$D7,$97,$DC,$FD,$56,$1D,$F0,$E5,$D7,$D7,$AE,$56,$C6,$49,$F1
	.dc.b	$1A,$8F,$3C,$E1,$E1,$2E,$BF,$47,$D4,$C9,$CC,$7A,$CC,$3F,$02,$30
	.dc.b	$F8,$7E,$A7,$F9,$FF,$C0,$7E,$AA,$3F,$81,$7D,$4F,$FB,$65,$FF,$5C
	.dc.b	$F7,$7F,$05,$51,$97,$FD,$3B,$19,$AE,$6D,$F0,$3F,$D3,$2F,$1A,$E1
	.dc.b	$7D,$EF,$9F,$C3,$E9,$FE,$25,$87,$27,$C7,$19,$76,$79,$0C,$94,$47
	.dc.b	$17,$67,$98,$F6,$F8,$7C,$EF,$F9,$64,$64,$DD,$F9,$61,$F5,$1E,$1F
	.dc.b	$A8,$7D,$4F,$EC,$AB,$D5,$18,$F5,$3C,$57,$C3,$0B,$87,$1C,$38,$69
	.dc.b	$F2,$F0,$E1,$5E,$1B,$E0,$F8,$E3,$38,$F6,$BE,$33,$FC,$DE,$73,$7F
	.dc.b	$37,$0E,$D8,$FC,$39,$CD,$6E,$F8,$BF,$6A,$B6,$F7,$8C,$CF,$08,$D1
	.dc.b	$CE,$32,$4A,$F6,$C6,$58,$F3,$72,$D2,$61,$EB,$8C,$B6,$33,$87,$C5
	.dc.b	$AF,$3E,$3F,$12,$B5,$5F,$EF,$8C,$3E,$4B,$7C,$67,$C5,$FF,$B4,$E7
	.dc.b	$33,$F3,$E7,$2F,$CD,$6D,$7E,$4F,$EA,$3F,$BB,$75,$7C,$99,$CE,$7C
	.dc.b	$71,$C1,$F8,$FA,$DE,$1C,$E3,$2D,$3F,$19,$1A,$C3,$F0,$F7,$D2,$7E
	.dc.b	$D3,$56,$28,$C5,$71,$C3,$C5,$5D,$6F,$9E,$1E,$3E,$CF,$E6,$31,$89
	.dc.b	$D5,$70,$8F,$74,$EA,$30,$49,$F1,$D7,$3C,$BD,$77,$CC,$95,$FC,$0F
	.dc.b	$0F,$25,$F0,$3B,$96,$7B,$D7,$2A,$ED,$3E,$EC,$7D,$B8,$F8,$E3,$84
	.dc.b	$EB,$2F,$85,$97,$8E,$7A,$9C,$5E,$8F,$F8,$67,$2F,$1B,$93,$73,$C9
	.dc.b	$FA,$77,$DD,$D6,$5E,$C7,$C2,$73,$8D,$4F,$19,$D4,$F0,$C3,$E1,$56
	.dc.b	$B1,$FB,$57,$B9,$F7,$E1,$E4,$8C,$5F,$63,$91,$DF,$2F,$5B,$82,$4F
	.dc.b	$1B,$EA,$4C,$BE,$26,$5E,$4C,$3A,$DF,$0C,$BC,$95,$D2,$73,$87,$38
	.dc.b	$C6,$1F,$4B,$1B,$1E,$29,$D6,$1F,$2A,$B4,$D9,$D6,$1D,$A3,$07,$CE
	.dc.b	$BE,$6F,$D8,$BE,$78,$F9,$F0,$E6,$DC,$39,$DF,$08,$F9,$F5,$C3,$2F
	.dc.b	$9F,$3A,$C3,$A6,$3F,$6E,$BF,$02,$3E,$A7,$11,$9F,$5E,$5F,$EE,$7D
	.dc.b	$7B,$87,$ED,$A6,$E5,$FC,$0E,$6E,$5F,$9F,$6D,$7E,$3D,$AB,$24,$7F
	.dc.b	$88,$BA,$FF,$C2,$F1,$8C,$7F,$C0,$0B,$DB,$D4,$75,$3F,$C1,$5F,$24
	.dc.b	$F8,$9F,$F4,$E4,$61,$7C,$2F,$0C,$95,$BF,$53,$0F,$19,$B5,$CA,$BF
	.dc.b	$D5,$93,$9B,$E9,$7A,$BD,$FD,$4F,$0F,$A9,$B5,$E1,$6E,$1C,$B1,$46
	.dc.b	$34,$FF,$20,$E5,$E9,$8F,$35,$3E,$A3,$FD,$D9,$87,$F8,$9F,$91,$F5
	.dc.b	$3F,$E7,$A8,$FF,$12,$4E,$AB,$FD,$85,$5A,$E1,$E7,$39,$CB,$FC,$86
	.dc.b	$DB,$FF,$4D,$7C,$7B,$E0,$FE,$86,$3F,$EB,$3A,$D3,$FE,$9F,$6B,$E5
	.dc.b	$F8,$2F,$99,$5C,$79,$3B,$C7,$CA,$F3,$FE,$5B,$71,$FA,$8D,$BF,$88
	.dc.b	$DC,$BE,$1F,$07,$83,$83,$EE,$7B,$3F,$50,$FA,$BF,$53,$71,$FB,$D1
	.dc.b	$71,$F9,$D3,$70,$FF,$30,$FD,$3F,$F9,$06,$FF,$CA,$10,$FA,$9F,$BC
	.dc.b	$EB,$C9,$FC,$33,$9E,$8F,$A7,$7C,$71,$FF,$6B,$36,$7F,$CF,$3C,$1F
	.dc.b	$27,$D5,$DE,$13,$2F,$F2,$C5,$ED,$F9,$FC,$7F,$E8,$F7,$33,$F9,$0F
	.dc.b	$A9,$91,$F9,$9F,$06,$E7,$F8,$99,$B9,$FC,$69,$B5,$FE,$D3,$E0,$FD
	.dc.b	$13,$F9,$AF,$A0,$FC,$DB,$4F,$D2,$17,$0F,$CA,$36,$FE,$3F,$A3,$FA
	.dc.b	$7E,$97,$D4,$FE,$AA,$BD,$73,$98,$C5,$FC,$48,$F8,$E6,$78,$6D,$70
	.dc.b	$24,$FE,$5A,$BE,$8C,$7F,$99,$F7,$5F,$ED,$29,$F5,$34,$C7,$C5,$DB
	.dc.b	$FE,$03,$3A,$6C,$E9,$A6,$BE,$A7,$A7,$CC,$F5,$6C,$64,$E7,$19,$69
	.dc.b	$FE,$91,$FA,$67,$D3,$FE,$83,$EA,$7F,$8A,$4E,$63,$1E,$37,$EE,$AB
	.dc.b	$73,$93,$FD,$41,$F5,$0F,$A9,$7D,$4E,$F0,$95,$BD,$5C,$9F,$EC,$6C
	.dc.b	$38,$5E,$A7,$35,$FE,$8F,$FA,$97,$37,$F0,$A6,$C7,$FA,$A7,$99,$C5
	.dc.b	$9A,$FE,$1F,$1F,$EA,$B9,$D7,$B9,$FC,$4B,$EA,$1F,$27,$A8,$C3,$27
	.dc.b	$FD,$69,$F5,$2F,$A9,$C6,$9E,$40,$30,$7C,$1F,$53,$B7,$95,$70,$FF
	.dc.b	$5F,$4F,$F1,$2E,$1F,$F6,$CB,$63,$FD,$59,$F5,$3D,$DD,$D7,$F2,$88
	.dc.b	$C6,$1E,$13,$C1,$19,$C7,$BD,$7F,$B2,$FB,$DD,$3F,$D1,$61,$F4,$DF
	.dc.b	$EC,$E7,$51,$98,$C6,$1D,$6B,$8D,$68,$FF,$6E,$1C,$BE,$A7,$F9,$DC
	.dc.b	$6A,$78,$5E,$0E,$93,$6B,$FD,$BC,$3E,$27,$C9,$1F,$4A,$33,$18,$E0
	.dc.b	$FE,$8F,$1B,$32,$78,$B1,$7D,$06,$AA,$4E,$B1,$FA,$CF,$C0,$9C,$C7
	.dc.b	$C9,$FA,$97,$D4,$FC,$13,$FA,$AB,$8A,$ED,$18,$23,$E1,$F8,$5B,$5C
	.dc.b	$E6,$F1,$2B,$F0,$A7,$D0,$AC,$BF,$EE,$9F,$81,$FE,$1B,$3B,$BE,$19
	.dc.b	$BE,$9F,$FB,$3D,$6A,$F9,$CE,$5F,$C8,$DF,$1B,$56,$DF,$CA,$B6,$7F
	.dc.b	$03,$D1,$D3,$DA,$91,$C8,$F2,$3F,$83,$8C,$D7,$E3,$FC,$FF,$BD,$87
	.dc.b	$9E,$5F,$50,$F8,$7F,$EF,$B5,$AC,$7F,$4B,$5A,$9C,$F2,$49,$FF,$00
	.dc.b	$7C,$07,$F7,$0B,$E1,$97,$F7,$8A,$D3,$FE,$00,$9D,$47,$6F,$A3,$FE
	.dc.b	$B7,$F5,$3C,$29,$69,$CD,$6F,$0F,$75,$7F,$81,$27,$FA,$16,$7F,$B8
	.dc.b	$9C,$CF,$29,$F5,$F0,$49,$CC,$FF,$00,$F8,$1F,$E0,$3F,$52,$FA,$9F
	.dc.b	$E0,$73,$E2,$BE,$0F,$C9,$7F,$C4,$2F,$11,$3C,$5F,$F0,$BC,$69,$57
	.dc.b	$6A,$EB,$9F,$E0,$67,$8C,$78,$1E,$97,$CD,$F1,$61,$C5,$D5,$D7,$27
	.dc.b	$23,$25,$33,$3C,$DE,$B5,$E2,$C3,$F3,$13,$99,$F3,$9C,$70,$F3,$46
	.dc.b	$6F,$C1,$18,$BF,$45,$70,$AF,$44,$E6,$3C,$71,$8B,$EF,$5C,$2F,$9C
	.dc.b	$62,$B7,$3E,$5C,$3D,$D1,$E8,$3E,$6B,$87,$C1,$1B,$7E,$01,$B3,$E0
	.dc.b	$BD,$DF,$39,$F6,$4B,$58,$70,$7E,$0B,$EA,$7F,$8F,$7B,$9C,$1B,$1F
	.dc.b	$01,$AF,$D9,$6B,$C7,$E4,$FF,$B2,$BB,$6B,$B6,$BB,$E3,$ED,$1C,$33
	.dc.b	$FC,$17,$D4,$BE,$97,$F2,$B1,$86,$E7,$C7,$0F,$C0,$62,$C3,$94,$FE
	.dc.b	$73,$0E,$55,$F8,$96,$BC,$27,$6F,$F6,$4A,$E5,$8E,$F0,$EF,$3F,$DA
	.dc.b	$2F,$8E,$3F,$97,$BD,$CF,$3B,$E0,$C3,$04,$9F,$53,$97,$96,$1E,$28
	.dc.b	$57,$49,$DC,$7E,$76,$FC,$38,$FC,$19,$E3,$87,$38,$F8,$18,$FA,$2F
	.dc.b	$9E,$3D,$70,$F8,$37,$FC,$C5,$BA,$70,$AF,$E1,$79,$78,$71,$C4,$7F
	.dc.b	$79,$AF,$F6,$2B,$FC,$41,$AE,$71,$FE,$17,$87,$F3,$DF,$A9,$FF,$5F
	.dc.b	$CF,$F1,$6F,$A7,$FF,$25,$B7,$D6,$DE,$7F,$C4,$AB,$73,$FE,$07,$6D
	.dc.b	$FF,$47,$BE,$33,$FE,$0B,$6C,$77,$E9,$E7,$3F,$E1,$EA,$7C,$CE,$21
	.dc.b	$18,$7F,$94,$B5,$E9,$E0,$6E,$31,$7E,$1F,$94,$E5,$89,$D6,$13,$0F
	.dc.b	$74,$68,$C3,$FE,$21,$FA,$8F,$E7,$CF,$F3,$86,$B9,$E7,$EB,$E6,$F9
	.dc.b	$E7,$38,$FC,$1F,$A9,$F0,$99,$A9,$7F,$36,$32,$F5,$8F,$81,$F3,$FF
	.dc.b	$EF,$78,$FE,$8E,$D7,$FC,$68,$DA,$F9,$DD,$9C,$3D,$8F,$7F,$A5,$2F
	.dc.b	$8C,$65,$FB,$91,$97,$9C,$7C,$FF,$82,$FF,$2B,$C7,$F5,$66,$BF,$E4
	.dc.b	$36,$DF,$F8,$0F,$97,$FC,$CF,$3B,$F2,$BE,$C4,$3E,$1E,$9D,$62,$4F
	.dc.b	$F6,$29,$D3,$67,$2F,$F9,$47,$CF,$FB,$7C,$FF,$D8,$1B,$1F,$E6,$46
	.dc.b	$AF,$A9,$FF,$0A,$BF,$DB,$A3,$30,$F9,$0F,$C6,$36,$75,$CB,$F8,$FA
	.dc.b	$DD,$68,$C4,$7E,$4F,$EA,$5E,$65,$25,$FF,$7C,$C3,$E2,$E3,$FD,$DA
	.dc.b	$B7,$3F,$E0,$0F,$A8,$FF,$50,$B7,$DC,$E2,$FF,$9B,$E3,$2F,$F9,$C7
	.dc.b	$EA,$7F,$D1,$EF,$F8,$21,$AE,$E7,$F4,$D3,$FE,$5F,$8F,$F0,$6F,$D4
	.dc.b	$BE,$1B,$F5,$4D,$8F,$17,$A9,$89,$CF,$D4,$C9,$E3,$1C,$3E,$97,$F6
	.dc.b	$5F,$52,$FA,$9A,$66,$BF,$C3,$7F,$17,$FE,$94,$6F,$BF,$FB,$98,$C4
	.dc.b	$7F,$A2,$FE,$45,$EA,$E6,$BF,$7D,$0F,$9E,$A6,$6B,$99,$FE,$28,$8D
	.dc.b	$36,$34,$FF,$A5,$FE,$A7,$E7,$B7,$DC,$FE,$15,$1F,$E7,$28,$FF,$18
	.dc.b	$BA,$EE,$FF,$2E,$37,$9B,$FC,$5A,$EE,$79,$1C,$3A,$39,$F8,$F2,$39
	.dc.b	$BD,$BE,$8B,$11,$E3,$8B,$87,$69,$F8,$9F,$2F,$2B,$E4,$BE,$6F,$92
	.dc.b	$C4,$7A,$FE,$A7,$D2,$62,$FF,$CE,$5C,$2B,$CD,$7B,$2B,$5A,$7F,$91
	.dc.b	$BE,$A5,$E4,$FF,$2B,$5C,$D4,$C4,$4B,$FF,$2A,$CE,$9B,$39,$7F,$D8
	.dc.b	$90,$FA,$9E,$84,$8E,$3E,$0D,$77,$17,$4F,$F4,$55,$7E,$E3,$EA,$5F
	.dc.b	$35,$FC,$FD,$B7,$FE,$66,$6B,$FE,$D3,$FA,$9F,$DE,$BF,$ED,$58,$CD
	.dc.b	$FF,$A5,$65,$F3,$3F,$D2,$5B,$5C,$EF,$B1,$99,$91,$93,$FC,$EB,$3F
	.dc.b	$E1,$56,$FB,$7C,$CE,$9E,$05,$8C,$90,$F3,$3E,$1E,$CF,$17,$64,$9E
	.dc.b	$B1,$97,$BC,$65,$F7,$47,$F6,$1F,$A8,$7D,$4D,$24,$FF,$A2,$BD,$5F
	.dc.b	$CC,$21,$3C,$1D,$7B,$5F,$DA,$E7,$FB,$8F,$2F,$FA,$2A,$B5,$5F,$EF
	.dc.b	$F6,$EF,$FE,$BF,$AD,$CF,$56,$BD,$B0,$E9,$D7,$D7,$AE,$4F,$96,$32
	.dc.b	$F9,$E3,$2A,$FB,$4F,$97,$AF,$CD,$F5,$32,$BD,$D1,$97,$E0,$46,$5E
	.dc.b	$59,$FF,$A9,$32,$57,$F8,$27,$EA,$7F,$81,$FC,$17,$E9,$EF,$E3,$7D
	.dc.b	$3F,$FB,$2D,$FF,$57,$77,$7F,$AB,$23,$11,$FE,$A4,$8F,$F5,$7D,$72
	.dc.b	$6F,$81,$FE,$B2,$BE,$94,$BE,$31,$86,$D7,$09,$EE,$78,$BE,$23,$94
	.dc.b	$78,$5B,$19,$36,$30,$F1,$39,$3C,$C7,$B3,$C4,$E2,$F8,$DB,$7E,$28
	.dc.b	$F8,$3F,$53,$AF,$A8,$7D,$4F,$F6,$8C,$FD,$71,$8F,$8D,$F0,$E0,$BE
	.dc.b	$67,$1B,$E1,$A7,$FB,$52,$F8,$63,$E4,$3E,$25,$6F,$0F,$5C,$EB,$1E
	.dc.b	$F8,$72,$BF,$29,$A7,$0A,$78,$61,$DE,$73,$8F,$96,$B8,$61,$1F,$15
	.dc.b	$F0,$7A,$98,$AE,$B1,$98,$ED,$1D,$31,$F4,$46,$EB,$A6,$3F,$76,$72
	.dc.b	$FA,$A3,$38,$FE,$2E,$72,$F9,$E3,$35,$E2,$AE,$0F,$52,$3D,$5F,$51
	.dc.b	$FE,$F1,$C3,$E6,$B7,$C6,$7C,$6F,$FB,$22,$73,$3F,$4E,$72,$FC,$26
	.dc.b	$D7,$E6,$3E,$A3,$FB,$9F,$1F,$D4,$9C,$E5,$C7,$1C,$1F,$92,$AD,$E1
	.dc.b	$CE,$32,$D3,$F0,$11,$AC,$3F,$17,$7D,$27,$F8,$B6,$28,$C4,$7D,$D6
	.dc.b	$D7,$1C,$3C,$53,$D6,$F9,$E1,$E3,$EB,$65,$E2,$9C,$D6,$EF,$F9,$22
	.dc.b	$C3,$A3,$E2,$C7,$C3,$97,$8F,$2E,$57,$C2,$F7,$7E,$59,$CB,$C5,$F1
	.dc.b	$63,$88,$D1,$E1,$3F,$9A,$4F,$2B,$ED,$19,$37,$8F,$D3,$CB,$D7,$19
	.dc.b	$9D,$CF,$B0,$D5,$F7,$8C,$E3,$F1,$A7,$35,$C7,$2F,$69,$F3,$30,$EB
	.dc.b	$3C,$EB,$81,$2B,$CF,$3A,$8C,$47,$99,$B8,$FA,$EF,$AE,$5F,$06,$30
	.dc.b	$F9,$63,$C1,$95,$8D,$61,$D7,$0F,$94,$79,$E3,$E5,$B4,$F8,$D7,$D1
	.dc.b	$C6,$3E,$E3,$D4,$F2,$6B,$F7,$EB,$B3,$E9,$35,$8F,$AA,$33,$97,$CA
	.dc.b	$9C,$90,$E5,$8F,$A5,$F1,$CF,$0C,$BE,$0C,$62,$75,$87,$8B,$1F,$43
	.dc.b	$E4,$C3,$B5,$74,$3D,$78,$F8,$EF,$E2,$92,$3A,$1F,$AC,$C7,$D7,$83
	.dc.b	$EA,$6C,$63,$D7,$97,$FB,$7F,$62,$E1,$FB,$49,$B9,$FF,$01,$9B,$97
	.dc.b	$E7,$DB,$5F,$98,$6B,$FA,$97,$F2,$BD,$5F,$96,$8C,$61,$FE,$11,$8C
	.dc.b	$66,$F7,$77,$3D,$4D,$1F,$E1,$7F,$93,$7C,$67,$FA,$5E,$33,$82,$73
	.dc.b	$F1,$1C,$A3,$15,$B7,$FC,$02,$62,$35,$3D,$3D,$5B,$49,$B1,$87,$3F
	.dc.b	$53,$AF,$A9,$B8,$F8,$B0,$F2,$C6,$31,$EF,$18,$9D,$62,$BE,$51,$EA
	.dc.b	$A7,$D4,$7F,$B8,$E3,$DB,$CC,$FA,$9F,$F3,$D4,$6E,$75,$5E,$CA,$D7
	.dc.b	$0F,$C1,$8E,$7F,$03,$6B,$7C,$3F,$C2,$3E,$49,$F0,$AE,$4D,$5C,$CF
	.dc.b	$B3,$F2,$EB,$73,$9C,$7D,$BB,$7F,$9D,$2F,$DB,$7F,$2B,$D0,$F9,$6D
	.dc.b	$8C,$9F,$0A,$32,$FC,$4F,$6B,$E8,$9F,$4B,$EA,$6C,$64,$98,$78,$23
	.dc.b	$2F,$DC,$1F,$4D,$FE,$3C,$9F,$C2,$EA,$FA,$9F,$C1,$3F,$93,$9D,$57
	.dc.b	$E1,$23,$5C,$FF,$26,$73,$1B,$F2,$3E,$17,$07,$FA,$23,$1F,$B1,$5A
	.dc.b	$7F,$A4,$B7,$0F,$F3,$73,$78,$3E,$7F,$A9,$D7,$0F,$C7,$FD,$4C,$C3
	.dc.b	$F0,$BE,$15,$C7,$E6,$CD,$CF,$F9,$FF,$4F,$E8,$AB,$17,$FA,$5E,$3F
	.dc.b	$EB,$AF,$E3,$BE,$83,$F3,$6F,$E7,$1A,$E4,$FC,$99,$93,$D7,$E9,$7E
	.dc.b	$A7,$A5,$F5,$3F,$A3,$9C,$9F,$94,$71,$A7,$D1,$E0,$5F,$C9,$39,$4F
	.dc.b	$D0,$AF,$8B,$19,$79,$D7,$E6,$67,$4F,$C8,$9F,$DA,$AF,$A9,$A6,$3E
	.dc.b	$36,$DF,$EC,$67,$4D,$9D,$34,$D7,$D4,$F4,$F8,$FE,$AD,$8C,$9C,$E3
	.dc.b	$2D,$3F,$6D,$F5,$0F,$8B,$FD,$27,$D4,$FF,$19,$9C,$C6,$3C,$6F,$D9
	.dc.b	$56,$E7,$27,$F1,$8F,$A8,$7D,$4B,$EA,$77,$84,$AD,$EB,$FC,$62,$F7
	.dc.b	$59,$8F,$D0,$7D,$4B,$93,$F8,$03,$6F,$F8,$67,$23,$87,$35,$FC,$16
	.dc.b	$1E,$DD,$3F,$84,$7D,$43,$E5,$B1,$87,$F1,$58,$C3,$63,$F8,$77,$D4
	.dc.b	$BE,$A3,$8C,$72,$F8,$39,$5C,$7C,$AB,$B7,$ED,$F1,$FE,$21,$39,$3F
	.dc.b	$91,$7D,$4F,$37,$13,$FB,$D8,$C3,$C6,$BF,$8F,$46,$3C,$0F,$E9,$D7
	.dc.b	$D2,$BF,$82,$78,$0E,$9F,$E8,$98,$BE,$9D,$F0,$3F,$C9,$EB,$53,$FD
	.dc.b	$30,$D4,$66,$A4,$66,$9F,$53,$8C,$3B,$D5,$CB,$FC,$6B,$7D,$EB,$0F
	.dc.b	$F9,$38,$7C,$3F,$92,$3E,$6D,$FF,$1B,$9C,$9E,$38,$CD,$7F,$6B,$8C
	.dc.b	$CE,$EB,$DB,$83,$E8,$7F,$A1,$56,$AF,$85,$60,$DC,$D9,$FE,$A5,$F5
	.dc.b	$2F,$94,$E6,$7F,$4B,$71,$1F,$06,$30,$47,$E2,$7C,$35,$CB,$F9,$A5
	.dc.b	$70,$6D,$79,$A7,$58,$FF,$81,$7E,$03,$F8,$5B,$F2,$33,$7D,$3F,$F5
	.dc.b	$B9,$CE,$4C,$7F,$CF,$75,$A8,$FE,$D5,$3A,$9D,$36,$3F,$01,$D3,$CF
	.dc.b	$DB,$91,$C1,$F9,$31,$8A,$FE,$99,$E7,$FE,$11,$FE,$DD,$F5,$0F,$88
	.dc.b	$FE,$DD,$39,$D3,$FC,$E7,$6A,$D6,$3F,$DF,$6B,$4F,$FA,$39,$B8,$7F
	.dc.b	$AD,$3E,$07,$FB,$AC,$E7,$93,$A9,$1F,$F0,$1C,$6A,$7E,$15,$72,$4B
	.dc.b	$E8,$3F,$C4,$59,$3E,$A7,$53,$9C,$3F,$B2,$C6,$2B,$D6,$7F,$41,$C9
	.dc.b	$D9,$8A,$F5,$70,$FC,$03,$FC,$1D,$97,$D3,$F8,$2F,$A9,$F5,$2F,$A8
	.dc.b	$F7,$56,$EF,$89,$FE,$1F,$BE,$02,$B8,$3E,$C9,$EA,$AF,$AD,$F4,$BB
	.dc.b	$3F,$EB,$83,$47,$38,$D3,$4D,$4F,$07,$E3,$D7,$FA,$AA,$FC,$31,$FE
	.dc.b	$C2,$9D,$69,$F8,$D8,$E4,$64,$AE,$67,$8C,$7B,$1E,$71,$9B,$F3,$4F
	.dc.b	$9A,$73,$7E,$19,$CD,$F8,$23,$15,$E8,$3B,$B8,$7C,$55,$CB,$0F,$64
	.dc.b	$78,$6F,$E0,$CF,$92,$3C,$11,$83,$BB,$D2,$78,$BA,$3E,$93,$8B,$F4
	.dc.b	$C7,$C1,$2B,$87,$ED,$9F,$4F,$EA,$7F,$8E,$FA,$9C,$1B,$1F,$09,$AE
	.dc.b	$FA,$FF,$CF,$AE,$7E,$49,$C3,$C5,$B1,$98,$C1,$1F,$AB,$F5,$2F,$A5
	.dc.b	$FF,$7D,$EB,$F7,$23,$13,$F8,$1C,$58,$74,$AF,$C0,$C6,$30,$FC,$35
	.dc.b	$F0,$BE,$B8,$74,$D7,$F8,$94,$2F,$83,$F8,$66,$E1,$E1,$AE,$13,$FC
	.dc.b	$66,$FA,$4F,$83,$1F,$16,$1D,$30,$E2,$F6,$AD,$B7,$0E,$78,$78,$6F
	.dc.b	$8B,$FC,$96,$F8,$BD,$E7,$77,$CA,$35,$5C,$1E,$58,$FC,$0A,$D9,$DA
	.dc.b	$F9,$CE,$8F,$86,$6C,$F9,$95,$CE,$FE,$45,$FF,$B0,$5B,$AF,$A6,$BF
	.dc.b	$D4,$17,$D2,$73,$1F,$17,$C4,$F9,$B7,$C5,$B7,$C6,$7F,$D0,$1F,$53
	.dc.b	$FE,$B7,$7F,$D5,$5F,$53,$FE,$52,$6F,$AF,$C1,$FE,$91,$3B,$78,$F6
	.dc.b	$65,$57,$17,$FD,$18,$DC,$7F,$DB,$55,$C5,$BF,$37,$C6,$0D,$3A,$4F
	.dc.b	$F7,$3A,$FF,$5F,$4E,$9F,$3F,$8B,$A3,$83,$AA,$C3,$A3,$07,$FB,$1F
	.dc.b	$EA,$7F,$9F,$37,$AB,$D5,$18,$8F,$4F,$9B,$89,$F1,$FD,$4C,$9E,$04
	.dc.b	$3D,$EE,$5F,$F5,$65,$73,$CF,$FD,$F5,$F3,$FF,$EF,$79,$FF,$6C,$35
	.dc.b	$FF,$59,$36,$3F,$E0,$06,$9F,$E3,$19,$DF,$B1,$F1,$3D,$29,$18,$21
	.dc.b	$F4,$1C,$BD,$23,$EB,$7C,$17,$F9,$56,$3F,$DD,$0D,$7F,$D8,$4D,$B7
	.dc.b	$17,$ED,$B9,$FF,$CC,$D5,$C3,$D0,$FC,$04,$3E,$22,$9D,$A2,$4F,$FB
	.dc.b	$D6,$74,$D9,$CB,$FE,$D1,$E6,$FB,$8D,$E2,$FC,$64,$63,$1F,$1C,$FF
	.dc.b	$86,$7E,$A5,$E4,$EA,$FF,$25,$0F,$90,$A6,$6B,$FC,$41,$7C,$9F,$F7
	.dc.b	$C5,$69,$CE,$9F,$C5,$2F,$94,$60,$FF,$14,$63,$E3,$9F,$E3,$3F,$52
	.dc.b	$F3,$7F,$44,$DA,$E0,$1C,$3F,$EE,$FB,$F9,$D5,$CA,$BF,$55,$F5,$1F
	.dc.b	$EA,$5A,$E2,$D7,$FD,$ED,$7C,$B2,$F5,$4E,$5F,$67,$D4,$FF,$A5,$EB
	.dc.b	$C2,$D7,$FD,$FD,$3B,$C3,$F9,$8B,$FE,$6A,$8F,$3F,$D4,$BE,$2B,$F7
	.dc.b	$8D,$8F,$1F,$8B,$89,$CF,$D4,$C9,$E3,$1C,$3E,$97,$F8,$27,$D4,$BE
	.dc.b	$A6,$99,$AF,$8B,$F1,$7F,$E2,$66,$F9,$FF,$C2,$E3,$11,$FD,$6B,$5F
	.dc.b	$E8,$FF,$22,$F4,$F5,$7F,$C4,$61,$F3,$D4,$CD,$F4,$3E,$9C,$69,$B1
	.dc.b	$A7,$B1,$FE,$65,$FA,$9F,$9D,$87,$F5,$67,$FC,$F9,$1F,$71,$AF,$00
	.dc.b	$F0,$79,$72,$FF,$25,$8D,$F2,$7E,$45,$B5,$D6,$75,$E3,$4B,$EC,$F1
	.dc.b	$FA,$2C,$47,$7B,$FF,$68,$BA,$3C,$91,$F2,$7E,$5E,$47,$CE,$7C,$5F
	.dc.b	$25,$88,$E5,$F5,$3E,$53,$19,$FE,$97,$81,$79,$AF,$81,$5A,$D3,$FB
	.dc.b	$CF,$D4,$3C,$8F,$E8,$4B,$96,$98,$89,$7F,$A4,$9D,$36,$72,$FC,$B8
	.dc.b	$7D,$4F,$42,$47,$1F,$02,$BB,$8B,$A7,$EF,$6B,$F6,$D3,$FE,$9F,$FA
	.dc.b	$97,$CD,$7F,$57,$6C,$7E,$D9,$AF,$DB,$FA,$9F,$E1,$AE,$67,$FA,$F4
	.dc.b	$3E,$66,$9C,$E2,$4F,$F0,$09,$D3,$67,$2F,$5F,$5F,$99,$1E,$1D,$F5
	.dc.b	$D1,$CE,$3C,$31,$E2,$EB,$F1,$F5,$CA,$F0,$46,$5F,$14,$65,$F6,$47
	.dc.b	$F7,$7F,$A9,$7D,$4D,$24,$FF,$18,$F5,$3F,$34,$A7,$6E,$7D,$7F,$F0
	.dc.b	$04,$FF,$7D,$E5,$FF,$1D,$D6,$AB,$F4,$8D,$E1,$FE,$61,$5B,$9F,$F6
	.dc.b	$93,$5E,$D8,$74,$EC,$EB,$D7,$27,$C9,$19,$7C,$D1,$97,$F6,$55,$FA
	.dc.b	$46,$F5,$F9,$BE,$A6,$47,$B6,$3D,$26,$1E,$31,$87,$F9,$CF,$D4,$FF
	.dc.b	$38,$F7,$FF,$47,$1F,$EE,$9F,$A9,$FF,$6D,$3F,$EB,$9E,$EF,$F2,$DC
	.dc.b	$62,$3F,$D6,$11,$FE,$5E,$AE,$4D,$F0,$7F,$D4,$CF,$0F,$87,$B5,$D5
	.dc.b	$AA,$B8,$B9,$3C,$05,$3B,$97,$0F,$3B,$C6,$F9,$E1,$E5,$9C,$BF,$C8
	.dc.b	$E7,$88,$71,$1A,$BE,$3F,$51,$E4,$FA,$87,$D4,$FE,$9A,$BB,$C6,$3D
	.dc.b	$3F,$2E,$3C,$EF,$8B,$89,$D3,$E0,$6D,$79,$27,$B5,$F5,$D3,$FD,$0F
	.dc.b	$5A,$6C,$66,$FE,$A5,$70,$C3,$B4,$E6,$57,$C5,$D9,$A9,$8F,$8E,$78
	.dc.b	$64,$BE,$93,$9A,$DE,$3E,$78,$B8,$F9,$26,$C7,$63,$81,$27,$E3,$E1
	.dc.b	$DA,$FA,$C6,$2B,$8D,$7C,$92,$E3,$E9,$7D,$77,$C5,$CB,$EF,$8D,$63
	.dc.b	$E9,$9F,$69,$FE,$FF,$8F,$A2,$DF,$15,$F1,$5F,$ED,$69,$CC,$FD,$09
	.dc.b	$CB,$F5,$DB,$8F,$E3,$BE,$A3,$FB,$A7,$47,$DD,$9C,$EB,$C7,$1C,$1F
	.dc.b	$84,$AD,$E1,$CE,$32,$D3,$F2,$71,$AC,$3F,$1B,$7D,$24,$7D,$C7,$0D
	.dc.b	$AE,$DD,$5F,$E4,$8A,$E3,$1D,$9B,$3D,$A7,$58,$79,$E3,$C5,$3A,$3D
	.dc.b	$D3,$A8,$F0,$E4,$C3,$C2,$DC,$3B,$61,$FE,$DE,$C7,$D4,$F5,$C7,$A3
	.dc.b	$4D,$63,$D7,$1F,$51,$E2,$8D,$3F,$02,$B9,$BE,$2C,$7C,$24,$C3,$E0
	.dc.b	$E5,$E2,$9D,$45,$55,$C9,$EF,$8F,$91,$D9,$F1,$6B,$86,$5E,$D8,$F2
	.dc.b	$1E,$3C,$BD,$11,$9A,$E5,$3D,$32,$F6,$DF,$2C,$BC,$D3,$D4,$E6,$6C
	.dc.b	$F0,$D7,$D0,$9E,$19,$7A,$72,$F4,$63,$DF,$1F,$8B,$87,$13,$DB,$97
	.dc.b	$93,$2E,$F9,$79,$30,$E3,$18,$BE,$31,$8C,$7B,$E1,$E5,$BF,$5E,$3D
	.dc.b	$E3,$CB,$7E,$C7,$D3,$19,$7C,$57,$CD,$CE,$1F,$96,$8C,$5F,$48,$C6
	.dc.b	$3E,$7C,$7C,$33,$F0,$6F,$8C,$7C,$19,$D4,$60,$F3,$63,$EA,$8C,$D7
	.dc.b	$13,$E0,$3F,$8D,$C3,$F0,$03,$EA,$71,$19,$F5,$A6,$9F,$B7,$F5,$AE
	.dc.b	$5F,$B4,$9B,$8F,$F0,$19,$B9,$7E,$7D,$B5,$F9,$76,$BF,$90,$7F,$D7
	.dc.b	$7D,$7F,$EB,$3C,$D1,$8C,$7F,$51,$ED,$DB,$E5,$1F,$EB,$3F,$92,$7C
	.dc.b	$47,$FA,$5E,$30,$BE,$13,$84,$EB,$D4,$B3,$9A,$DC,$62,$B2,$D8,$E7
	.dc.b	$F5,$1C,$3E,$A7,$8B,$63,$27,$82,$32,$D3,$FD,$83,$F5,$1F,$EE,$57
	.dc.b	$FD,$95,$C8,$FA,$9F,$F3,$6C,$6B,$DF,$F7,$4E,$78,$BE,$06,$CC,$39
	.dc.b	$3F,$E8,$F6,$F1,$7F,$B9,$3E,$39,$F0,$7F,$2B,$2E,$84,$BE,$19,$7C
	.dc.b	$BA,$DF,$1F,$6C,$79,$DF,$F2,$CE,$2F,$9B,$E7,$FC,$8C,$3E,$4B,$67
	.dc.b	$E7,$B7,$1F,$F5,$4D,$70,$9F,$A7,$5F,$43,$D9,$91,$F5,$BE,$A6,$DF
	.dc.b	$0E,$55,$63,$6F,$F7,$FE,$3F,$7A,$F1,$87,$F4,$0C,$1F,$4E,$F1,$FF
	.dc.b	$0D,$83,$EA,$7F,$09,$5F,$EB,$5A,$D5,$FF,$BE,$EB,$5C,$FF,$3A,$73
	.dc.b	$5B,$C3,$E3,$71,$7F,$9F,$F0,$FF,$7D,$D6,$B0,$47,$8E,$30,$DA,$E3
	.dc.b	$E9,$75,$7F,$23,$F5,$32,$7F,$2F,$E0,$DC,$BE,$14,$DC,$7E,$0F,$07
	.dc.b	$F4,$DB,$FC,$EC,$FE,$69,$FC,$DF,$CF,$D7,$F4,$2D,$BF,$98,$DB,$E6
	.dc.b	$FB,$FD,$1F,$C7,$FD,$4F,$E8,$EB,$E3,$46,$3B,$B5,$87,$C0,$8D,$F0
	.dc.b	$62,$B7,$39,$3F,$3B,$7F,$43,$1F,$8B,$3F,$DD,$B7,$61,$FB,$48,$7D
	.dc.b	$4D,$31,$F1,$77,$1F,$DF,$4E,$9B,$3A,$69,$AF,$A9,$E9,$F3,$3D,$5B
	.dc.b	$19,$39,$C6,$5A,$7F,$82,$B2,$F3,$DB,$EA,$1F,$4F,$FA,$7F,$A9,$78
	.dc.b	$DF,$B4,$9C,$E1,$C4,$FF,$09,$F8,$2A,$DC,$67,$0F,$F0,$BD,$70,$58
	.dc.b	$F8,$23,$0F,$C8,$FA,$87,$D4,$BE,$A7,$C1,$A4,$C7,$A6,$3E,$78,$FE
	.dc.b	$4D,$7C,$2B,$53,$DA,$BF,$51,$F5,$2E,$7F,$E1,$AD,$8E,$7E,$3B,$8F
	.dc.b	$35,$FC,$4A,$1E,$DD,$3F,$8A,$FD,$33,$E5,$71,$87,$92,$30,$D8,$FE
	.dc.b	$37,$F5,$2F,$A9,$E3,$1C,$BE,$0A,$57,$1F,$32,$ED,$E9,$3F,$C7,$F8
	.dc.b	$FF,$1D,$70,$1C,$71,$FD,$AF,$A9,$ED,$7C,$CF,$8E,$E2,$B8,$D7,$09
	.dc.b	$FE,$19,$39,$AF,$C1,$CE,$6B,$F9,$27,$4F,$F5,$3B,$7D,$2B,$E0,$7F
	.dc.b	$9B,$F9,$6A,$D4,$E6,$BE,$BC,$E5,$B3,$AF,$A9,$E2,$C8,$C9,$F8,$93
	.dc.b	$BC,$67,$0F,$83,$0F,$86,$F2,$4F,$D0,$9C,$C6,$37,$FE,$97,$5B,$3F
	.dc.b	$12,$59,$F9,$CF,$B3,$EA,$77,$52,$36,$FB,$23,$13,$F4,$BE,$A5,$F5
	.dc.b	$3F,$00,$FE,$B2,$E2,$FD,$5A,$5A,$E1,$5A,$3E,$3F,$C2,$5A,$FC,$ED
	.dc.b	$EF,$83,$F6,$AD,$8F,$F4,$09,$F4,$A7,$85,$FE,$13,$E0,$BF,$8E,$5F
	.dc.b	$F1,$2A,$FA,$FF,$4F,$FD,$B2,$73,$1C,$B2,$4E,$7C,$B9,$3E,$8E,$AB
	.dc.b	$8F,$59,$B8,$F5,$C3,$FD,$53,$7F,$B4,$C7,$AD,$FF,$04,$E1,$FE,$FB
	.dc.b	$F5,$0F,$86,$FE,$FB,$E3,$7E,$0E,$BF,$3D,$A2,$B5,$8F,$EF,$AB,$59
	.dc.b	$FF,$16,$6C,$7F,$A8,$A9,$F0,$3F,$DE,$B0,$F0,$4F,$1C,$BB,$9F,$E0
	.dc.b	$E9,$D4,$AF,$9B,$FC,$6B,$E8,$FF,$B5,$FD,$4E,$23,$51,$8B,$F4,$B3
	.dc.b	$0F,$F0,$EE,$6E,$DC,$5F,$ED,$B1,$F9,$B5,$EB,$E1,$FF,$0E,$CF,$FA
	.dc.b	$EB,$E0,$BF,$D7,$7F,$52,$FA,$9F,$EF,$D8,$70,$AE,$57,$F2,$0E,$2A
	.dc.b	$F8,$D7,$09,$CC,$62,$7F,$9A,$C7,$F5,$4B,$ED,$1F,$E0,$B3,$90,$72
	.dc.b	$FC,$58,$E1,$5B,$8D,$CE,$AF,$CC,$F0,$8F,$61,$CE,$7C,$98,$78,$E3
	.dc.b	$FC,$35,$85,$BE,$35,$D6,$31,$87,$82,$BC,$D1,$8B,$E5,$1D,$B0,$F1
	.dc.b	$56,$E3,$A9,$E1,$3E,$0C,$F0,$7D,$37,$E4,$74,$70,$AE,$73,$BB,$F0
	.dc.b	$4E,$8F,$05,$71,$AF,$81,$1D,$0E,$E6,$CF,$9A,$7C,$92,$9E,$F3,$06
	.dc.b	$0F,$F7,$EF,$D4,$FF,$1D,$F5,$B8,$36,$3E,$93,$5F,$B4,$D7,$FC,$8C
	.dc.b	$BE,$47,$FB,$7C,$70,$6B,$C1,$AF,$06,$BE,$27,$8D,$7E,$13,$EA,$5F
	.dc.b	$4B,$FE,$53,$8C,$37,$1E,$5C,$BF,$7F,$1F,$13,$F8,$1B,$E3,$8D,$BE
	.dc.b	$0F,$79,$E1,$8F,$83,$0F,$F3,$05,$F0,$AF,$AE,$DC,$7F,$2F,$8F,$69
	.dc.b	$FB,$6D,$8F,$2C,$77,$3A,$E1,$1E,$78,$F8,$95,$F5,$C3,$D7,$1D,$31
	.dc.b	$F2,$23,$75,$C6,$B8,$47,$AC,$E1,$7E,$17,$B9,$E1,$C3,$CA,$FF,$7C
	.dc.b	$9F,$74,$FF,$8D,$1B,$A7,$8A,$B7,$39,$F1,$7F,$D0,$8D,$BF,$F1,$2D
	.dc.b	$F0,$7C,$DF,$53,$FE,$C5,$7F,$C5,$3F,$4F,$FE,$54,$6F,$AF,$CB,$2F
	.dc.b	$F4,$75,$6E,$7F,$C1,$ED,$C3,$FC,$8F,$7C,$67,$FC,$26,$D8,$FF,$25
	.dc.b	$CE,$EB,$9F,$CD,$75,$BF,$1B,$54,$A3,$16,$F0,$B5,$39,$AF,$4F,$8F
	.dc.b	$A3,$37,$23,$24,$3D,$2F,$7F,$A8,$74,$D2,$4F,$C4,$F3,$6B,$EC,$FA
	.dc.b	$9F,$19,$8D,$7F,$CC,$F7,$CA,$7E,$5C,$E9,$FF,$2A,$C6,$5F,$F1,$BF
	.dc.b	$CF,$7F,$9A,$9F,$F2,$CB,$5C,$FB,$3F,$03,$BB,$E2,$4E,$70,$F9,$FE
	.dc.b	$97,$C4,$33,$52,$FE,$DC,$65,$EB,$1F,$37,$E0,$FF,$CA,$D1,$FE,$68
	.dc.b	$6B,$FE,$4A,$6D,$7F,$A0,$B9,$FF,$CC,$D3,$BF,$33,$EB,$43,$E1,$E9
	.dc.b	$D6,$24,$FF,$9D,$27,$4D,$9C,$BF,$E5,$8E,$8F,$B4,$DE,$1F,$C1,$46
	.dc.b	$30,$FF,$49,$FD,$4B,$C8,$CD,$7E,$1E,$1F,$21,$4C,$E1,$FE,$84,$C3
	.dc.b	$97,$4F,$FB,$9E,$B7,$3F,$E8,$86,$BE,$13,$FD,$17,$F5,$2F,$37,$F2
	.dc.b	$8D,$8C,$C7,$87,$6C,$BF,$99,$FA,$8F,$F5,$3C,$FF,$A5,$1A,$FF,$9D
	.dc.b	$AB,$93,$FE,$78,$8C,$EB,$FE,$B0,$FA,$9F,$F4,$84,$FF,$A7,$1A,$EF
	.dc.b	$DD,$ED,$E9,$FE,$9F,$EA,$5F,$0D,$4C,$D7,$FA,$92,$32,$DF,$A9,$E0
	.dc.b	$6D,$D7,$D2,$7E,$C7,$EA,$5F,$53,$4E,$D1,$FB,$5F,$8C,$FF,$0E,$57
	.dc.b	$FA,$C1,$AF,$FA,$49,$BA,$FF,$B2,$BE,$47,$FC,$67,$3B,$F4,$BF,$7D
	.dc.b	$0F,$A9,$A7,$58,$93,$FE,$BC,$9D,$36,$72,$FF,$A8,$3E,$A7,$E5,$DF
	.dc.b	$FB,$61,$FF,$38,$C7,$FB,$11,$AE,$F9,$3F,$87,$47,$0E,$7F,$E7,$12
	.dc.b	$8D,$D7,$3F,$77,$77,$C3,$C6,$97,$D5,$E7,$F4,$58,$8F,$24,$5B,$F2
	.dc.b	$4F,$C9,$F9,$79,$5F,$25,$F1,$FC,$96,$23,$D9,$F5,$3E,$63,$17,$FE
	.dc.b	$F4,$E0,$5E,$6B,$DD,$5A,$D7,$D7,$F5,$0F,$23,$F9,$62,$E3,$A6,$22
	.dc.b	$5F,$FB,$AA,$74,$D9,$CB,$FE,$CE,$87,$D4,$F4,$24,$71,$F0,$EB,$B8
	.dc.b	$BA,$31,$F2,$D7,$F3,$FF,$A9,$7C,$D5,$26,$1F,$EF,$CF,$A9,$FE,$00
	.dc.b	$FF,$BF,$A7,$6E,$7E,$67,$FA,$44,$9E,$31,$FE,$E4,$9D,$BF,$EE,$69
	.dc.b	$D3,$C4,$E9,$5C,$BD,$99,$1E,$17,$AF,$81,$88,$E5,$17,$3F,$64,$7F
	.dc.b	$80,$6F,$FD,$A4,$FA,$7E,$0B,$CB,$D9,$23,$04,$3C,$8E,$5F,$34,$7F
	.dc.b	$63,$FA,$97,$D4,$D2,$4F,$F8,$3F,$0F,$0B,$FE,$F1,$EA,$CA,$8C,$FA
	.dc.b	$DF,$DC,$F8,$7F,$3E,$D9,$FF,$0C,$61,$D9,$E9,$3F,$A4,$8C,$47,$38
	.dc.b	$E3,$ED,$7F,$77,$AE,$52,$AD,$DF,$83,$B3,$5D,$92,$BA,$9F,$E2,$7B
	.dc.b	$4F,$33,$51,$DC,$D4,$E5,$E2,$25,$D9,$E3,$FA,$99,$3A,$9C,$B6,$72
	.dc.b	$F5,$39,$4F,$AE,$3F,$C2,$5F,$53,$FC,$DB,$E0,$BF,$5F,$5F,$E3,$DF
	.dc.b	$A7,$FF,$6C,$3F,$EB,$7E,$FF,$E1,$31,$88,$FF,$59,$47,$F0,$CA,$E4
	.dc.b	$DE,$0F,$F4,$FB,$C3,$E2,$6D,$75,$6A,$AE,$2E,$4F,$01,$4E,$E5,$C3
	.dc.b	$CE,$F1,$BE,$78,$79,$67,$2F,$F2,$C9,$E2,$1C,$46,$AF,$8F,$D4,$F9
	.dc.b	$3E,$A1,$F5,$3F,$B4,$AE,$F1,$8F,$5B,$CB,$8F,$3B,$E2,$E2,$74,$F8
	.dc.b	$1B,$5E,$49,$ED,$7D,$74,$FE,$6B,$5A,$6C,$66,$FE,$A5,$70,$C3,$B4
	.dc.b	$E6,$57,$C5,$D9,$A9,$8F,$8E,$78,$64,$BE,$93,$9A,$DE,$3E,$78,$B8
	.dc.b	$F9,$26,$C7,$63,$81,$27,$E3,$E1,$DA,$FA,$C6,$2B,$8D,$7C,$92,$E3
	.dc.b	$E9,$7D,$77,$C5,$CB,$EF,$8D,$63,$E9,$9C,$D7,$48,$C4,$79,$0F,$F8
	.dc.b	$01,$3F,$51,$BE,$2B,$E3,$3F,$DA,$D3,$99,$FA,$73,$97,$ED,$37,$1F
	.dc.b	$C8,$FD,$43,$AB,$F9,$2C,$7E,$1A,$75,$AF,$2C,$7F,$09,$18,$C3,$FD
	.dc.b	$55,$87,$1E,$9C,$D4,$9E,$EF,$E3,$A7,$51,$82,$61,$DE,$75,$8F,$8B
	.dc.b	$1F,$05,$69,$B5,$E8,$9D,$4F,$FA,$EF,$26,$5E,$58,$FC,$5B,$63,$38
	.dc.b	$76,$CB,$CD,$87,$86,$FB,$1E,$CC,$3A,$BC,$AB,$78,$F8,$8F,$F0,$9E
	.dc.b	$5E,$CB,$E4,$48,$D5,$73,$C3,$8B,$E7,$CB,$C8,$79,$EF,$95,$70,$9F
	.dc.b	$01,$E4,$11,$A7,$96,$1D,$1D,$E1,$F1,$70,$C6,$5F,$02,$79,$5F,$17
	.dc.b	$78,$77,$9D,$5F,$19,$F0,$46,$27,$C9,$5C,$63,$13,$E2,$3A,$E5,$F2
	.dc.b	$23,$31,$D8,$D5,$F5,$3E,$33,$8B,$EB,$5C,$72,$F8,$B9,$7B,$71,$F0
	.dc.b	$46,$2F,$BE,$5E,$E8,$C6,$3D,$71,$F6,$46,$1F,$AF,$8F,$5C,$78,$1F
	.dc.b	$83,$C3,$DF,$5B,$7E,$C1,$88,$F1,$4F,$18,$C1,$FC,$D3,$1F,$2C,$F4
	.dc.b	$8C,$4E,$F0,$F2,$47,$BC,$EA,$FC,$27,$EE,$C7,$E4,$2B,$EB,$0F,$A9
	.dc.b	$C4,$67,$D9,$99,$FE,$B7,$D9,$B8,$78,$E6,$E3,$FB,$E9,$BA,$7E,$61
	.dc.b	$B5,$FB,$A6,$BF,$8E,$7F,$C4,$FD,$5F,$E1,$8C,$D1,$8C,$7F,$43,$EE
	.dc.b	$DB,$E5,$1F,$E0,$8F,$93,$7C,$5F,$FA,$82,$30,$BE,$47,$84,$EB,$D7
	.dc.b	$B3,$9A,$DC,$62,$B2,$D8,$E7,$F5,$3C,$3E,$A7,$8B,$63,$27,$82,$32
	.dc.b	$D3,$FC,$2B,$F5,$1F,$EE,$57,$FC,$3F,$C8,$FA,$9F,$F3,$CC,$6B,$E0
	.dc.b	$BD,$D3,$9E,$2F,$81,$B3,$0E,$4F,$F8,$5D,$BC,$5F,$E5,$0F,$92,$7C
	.dc.b	$2F,$CA,$CB,$A1,$2F,$86,$5F,$2E,$B7,$C7,$DB,$1E,$76,$B7,$17,$CE
	.dc.b	$F3,$BE,$4E,$3F,$29,$B5,$F4,$1B,$5F,$E3,$7A,$E1,$3F,$52,$BE,$8F
	.dc.b	$B5,$23,$EB,$FD,$4D,$BE,$1C,$AA,$C6,$DF,$E6,$CC,$7E,$FD,$E3,$0F
	.dc.b	$F5,$BF,$D3,$7F,$90,$1F,$F2,$5D,$3E,$A7,$F0,$99,$7F,$8F,$6B,$4F
	.dc.b	$F9,$2E,$75,$CD,$F3,$A7,$31,$FE,$6A,$6D,$FC,$3E,$0F,$9B,$EA,$6F
	.dc.b	$C8,$93,$FE,$52,$AD,$E2,$BD,$CE,$6B,$F2,$3F,$53,$23,$F2,$FE,$0D
	.dc.b	$BF,$83,$37,$2F,$81,$C1,$FD,$42,$FF,$3B,$3F,$9A,$7F,$37,$F4,$15
	.dc.b	$FD,$0B,$6F,$E5,$B6,$F9,$BE,$FF,$43,$F1,$FF,$53,$FA,$3A,$F8,$D1
	.dc.b	$8E,$ED,$61,$ED,$8D,$F0,$62,$B7,$39,$3F,$3B,$7F,$43,$1F,$95,$7F
	.dc.b	$B2,$A7,$D4,$D3,$1F,$1B,$70,$FD,$E4,$E9,$B3,$A6,$9A,$FA,$9E,$9F
	.dc.b	$1F,$D5,$B1,$93,$9C,$65,$A7,$FA,$23,$EA,$1F,$15,$FA,$3F,$A9,$78
	.dc.b	$DF,$B0,$9C,$D7,$FA,$5B,$99,$5A,$8E,$57,$B3,$F8,$B7,$D4,$3E,$A5
	.dc.b	$F5,$39,$AD,$93,$6F,$F4,$E5,$6E,$75,$1F,$E9,$3F,$A9,$73,$7F,$02
	.dc.b	$6D,$FF,$A8,$3C,$77,$16,$6B,$F8,$3C,$7F,$AA,$A7,$5E,$DF,$F0,$AF
	.dc.b	$A6,$7C,$A6,$A3,$0C,$9F,$F5,$5F,$D4,$BE,$A7,$1A,$7F,$A9,$70,$7C
	.dc.b	$17,$33,$A7,$97,$70,$FF,$59,$CF,$F1,$1A,$E1,$3F,$C6,$BE,$A7,$A5
	.dc.b	$6A,$31,$87,$28,$FF,$5D,$CE,$AB,$8C,$7F,$AF,$BC,$87,$4F,$F4,$38
	.dc.b	$7D,$3F,$FB,$05,$D4,$66,$31,$5F,$EB,$DA,$E3,$5A,$3F,$DA,$27,$2F
	.dc.b	$A9,$FE,$5F,$1A,$9E,$17,$83,$A4,$DA,$FF,$68,$8F,$88,$F2,$47,$CE
	.dc.b	$8C,$C6,$37,$FE,$77,$1B,$32,$78,$B1,$7D,$06,$AA,$4E,$B1,$FF,$71
	.dc.b	$3F,$02,$73,$1F,$ED,$9F,$A9,$7C,$97,$C1,$3F,$A4,$B8,$AE,$D1,$82
	.dc.b	$3E,$1F,$83,$B5,$CE,$6F,$88,$AF,$E8,$87,$D0,$AC,$BF,$D1,$E9,$F0
	.dc.b	$1F,$C2,$EB,$84,$71,$CD,$F4,$FF,$D7,$2B,$57,$D2,$72,$FF,$BF,$23
	.dc.b	$35,$FD,$4D,$B1,$F8,$0E,$8E,$5E,$4C,$8E,$0F,$E0,$63,$15,$FE,$FD
	.dc.b	$F0,$BE,$E6,$5E,$49,$7D,$43,$E1,$7F,$B7,$D6,$B1,$FE,$C5,$5A,$9C
	.dc.b	$F2,$49,$FE,$F7,$F0,$1F,$DA,$32,$FE,$8D,$39,$7C,$0F,$F7,$C8,$D4
	.dc.b	$7F,$45,$8E,$FF,$45,$FE,$1D,$CD,$F5,$3E,$1A,$FE,$A1,$39,$8C,$47
	.dc.b	$F4,$D8,$E3,$87,$F8,$33,$4F,$DA,$BF,$DE,$6F,$E8,$D6,$E3,$07,$D7
	.dc.b	$D3,$E7,$C7,$C5,$E0,$FF,$07,$BF,$E0,$1F,$81,$FF,$01,$7D,$3B,$EA
	.dc.b	$38,$57,$C6,$26,$1D,$2F,$F0,$42,$F8,$D7,$09,$D4,$FE,$1E,$72,$FF
	.dc.b	$6D,$AF,$C3,$DF,$78,$E7,$3A,$7F,$B6,$4E,$AF,$BC,$77,$15,$C5,$DD
	.dc.b	$F1,$7D,$86,$A7,$C8,$D3,$B9,$4E,$25,$38,$CF,$5C,$3E,$4B,$DE,$33
	.dc.b	$3D,$23,$11,$E5,$78,$5F,$A6,$3B,$C6,$2F,$CD,$1B,$AE,$47,$BB,$0F
	.dc.b	$85,$87,$82,$31,$3E,$78,$C3,$D2,$38,$B4,$F9,$D5,$DA,$BE,$4C,$7D
	.dc.b	$32,$9F,$2C,$CB,$AF,$A9,$FE,$3D,$EA,$5E,$AF,$F3,$8D,$7D,$29,$CB
	.dc.b	$63,$2F,$E1,$21,$F2,$4D,$7A,$36,$79,$35,$E0,$D7,$F8,$74,$F2,$9F
	.dc.b	$C2,$7D,$4B,$E9,$7F,$89,$6D,$F8,$18,$C5,$FE,$26,$F8,$3F,$75,$B3
	.dc.b	$BA,$E3,$AF,$F1,$BB,$E3,$7B,$C7,$CD,$7F,$8D,$C3,$8D,$EF,$0E,$37
	.dc.b	$FE,$69,$C3,$8D,$EF,$0E,$37,$EA,$6D,$6E,$F8,$AB,$E3,$5B,$C7,$D5
	.dc.b	$7C,$2B,$57,$C2,$F7,$1F,$E7,$9C,$7C,$AD,$AE,$13,$A8,$E2,$74,$B8
	.dc.b	$FE,$86,$34,$6E,$B9,$C7,$03,$ED,$6B,$F0,$5F,$DE,$CF,$C8,$AF,$E2
	.dc.b	$35,$C2,$31,$5F,$37,$C3,$FE,$91,$5F,$C5,$A7,$FD,$83,$7C,$B1,$CD
	.dc.b	$FF,$4D,$FA,$8F,$F5,$E4,$FF,$0A,$FA,$97,$AD,$A8,$FF,$4C,$46,$1F
	.dc.b	$E1,$F3,$C1,$F3,$9F,$E9,$D7,$18,$FF,$26,$9E,$13,$A0,$E2,$3F,$95
	.dc.b	$CF,$0B,$F2,$7C,$C7,$86,$D2,$8C,$36,$3D,$1E,$16,$27,$86,$0F,$93
	.dc.b	$A6,$6B,$D3,$38,$7E,$04,$65,$F3,$7D,$43,$A7,$C4,$D8,$FE,$5E,$D7
	.dc.b	$D9,$6F,$37,$75,$E2,$FA,$99,$5C,$23,$2F,$CF,$8C,$BE,$08,$F8,$1F
	.dc.b	$3D,$FE,$68,$8F,$E7,$AD,$7E,$53,$76,$F9,$DD,$DC,$3D,$9F,$83,$E9
	.dc.b	$4B,$E3,$19,$7E,$F4,$65,$E7,$1F,$3F,$E2,$7F,$CB,$71,$FD,$31,$AF
	.dc.b	$D9,$6D,$FF,$62,$E4,$FF,$35,$CE,$FC,$BF,$B1,$0F,$87,$A7,$58,$93
	.dc.b	$FD,$62,$74,$D9,$CB,$F8,$7E,$8F,$BC,$DE,$3F,$C2,$46,$30,$FE,$E7
	.dc.b	$F5,$3F,$E2,$6A,$FE,$C7,$39,$E7,$FF,$15,$4B,$E4,$3F,$18,$DA,$DE
	.dc.b	$1D,$0C,$CC,$8C,$9F,$DB,$BE,$A5,$E6,$FE,$55,$B1,$98,$EB,$AE,$5F
	.dc.b	$CD,$7D,$4F,$FA,$92,$7F,$BB,$B5,$FD,$25,$72,$7F,$4D,$19,$D3,$FC
	.dc.b	$1F,$F5,$3F,$E9,$49,$FE,$FE,$D7,$7E,$F7,$6F,$13,$F5,$1F,$52,$F8
	.dc.b	$5A,$66,$BF,$C0,$F1,$96,$FD,$4F,$03,$6E,$BE,$93,$F6,$5F,$52,$FA
	.dc.b	$9A,$76,$8F,$DB,$7C,$5F,$FA,$6A,$BF,$C2,$6D,$7F,$80,$37,$5F,$F1
	.dc.b	$4F,$C8,$BD,$3E,$3F,$01,$FB,$F8,$7D,$4D,$33,$87,$49,$C3,$FE,$20
	.dc.b	$9D,$B9,$3F,$C4,$7F,$53,$F2,$F0,$FF,$1B,$BF,$E7,$A8,$CF,$B9,$FC
	.dc.b	$36,$35,$D7,$FE,$6F,$6C,$83,$C5,$FE,$34,$DA,$EB,$CF,$CF,$C6,$95
	.dc.b	$E2,$7B,$7D,$16,$23,$8C,$5A,$F2,$CF,$CA,$F9,$79,$5F,$29,$F3,$7C
	.dc.b	$8E,$23,$C3,$F5,$3E,$93,$17,$FE,$54,$E1,$5E,$6B,$E1,$D6,$B4,$FF
	.dc.b	$46,$7D,$4B,$C9,$FE,$5C,$B9,$29,$88,$97,$FE,$4E,$9D,$36,$72,$FF
	.dc.b	$3D,$87,$D4,$F4,$24,$71,$F0,$AB,$B8,$BA,$7F,$9B,$2B,$F7,$13,$F4
	.dc.b	$7E,$A5,$F3,$5F,$D0,$DB,$1F,$E6,$06,$BF,$D3,$7E,$A7,$F8,$13,$FD
	.dc.b	$3E,$33,$7F,$E7,$89,$7C,$CF,$F4,$D6,$D7,$3B,$EC,$66,$64,$64,$FF
	.dc.b	$37,$CF,$F8,$C1,$BE,$CF,$33,$A7,$83,$63,$24,$3C,$CF,$87,$B3,$C5
	.dc.b	$D7,$27,$AC,$65,$EF,$19,$7D,$D1,$FD,$97,$EA,$1F,$53,$49,$3F,$E8
	.dc.b	$4F,$5B,$F3,$6A,$76,$E7,$D9,$FE,$DD,$3F,$DC,$F9,$7F,$DA,$15,$AA
	.dc.b	$FE,$F0,$DD,$FF,$D4,$35,$B9,$FD,$33,$5E,$D8,$74,$EB,$EB,$D7,$27
	.dc.b	$C9,$19,$7C,$D1,$97,$FC,$0D,$5F,$DA,$DF,$57,$5F,$9B,$EA,$64,$60
	.dc.b	$87,$07,$2F,$28,$FF,$05,$7D,$4F,$F2,$DF,$81,$FD,$2C,$FE,$F7,$E9
	.dc.b	$FF,$DA,$CF,$FA,$D7,$BB,$FD,$E1,$18,$8F,$F5,$84,$7F,$BC,$EB,$93
	.dc.b	$7C,$0F,$F5,$83,$C3,$E1,$ED,$75,$6A,$AE,$2E,$4F,$01,$4E,$E5,$C3
	.dc.b	$CE,$F1,$BE,$78,$79,$67,$2F,$F2,$29,$E2,$1C,$46,$AF,$8F,$D4,$F9
	.dc.b	$3E,$A1,$F5,$3F,$DA,$6B,$BC,$63,$E3,$7C,$B8,$F3,$BE,$2E,$27,$4F
	.dc.b	$81,$B5,$E4,$9E,$D7,$D7,$4F,$E5,$96,$BD,$C6,$1B,$1F,$56,$B8,$E1
	.dc.b	$DE,$F9,$4F,$F3,$5A,$DB,$E5,$35,$31,$F2,$4F,$07,$46,$23,$58,$7C
	.dc.b	$09,$D6,$4C,$3C,$37,$B2,$57,$B5,$E3,$8F,$A6,$FB,$63,$E4,$C3,$95
	.dc.b	$68,$F4,$4E,$8E,$F8,$76,$C3,$C5,$8E,$EB,$CF,$18,$7C,$38,$F7,$9F
	.dc.b	$AB,$87,$69,$C9,$DA,$39,$46,$6B,$91,$2B,$95,$7E,$22,$75,$18,$8F
	.dc.b	$AE,$7D,$7C,$3F,$96,$BF,$5F,$C5,$7C,$67,$FB,$83,$3F,$C1,$CF,$05
	.dc.b	$3A,$8F,$C3,$37,$1F,$CE,$FD,$47,$F8,$0F,$A3,$F1,$93,$9C,$BB,$63
	.dc.b	$83,$F2,$75,$BC,$39,$C6,$5A,$7E,$8A,$35,$87,$E7,$2F,$A4,$FE,$25
	.dc.b	$AF,$E4,$DB,$3C,$2F,$C3,$3D,$2B,$95,$F8,$BA,$98,$F8,$67,$35,$BB
	.dc.b	$FC,$FE,$1C,$E3,$C1,$86,$E7,$AD,$7C,$2C,$39,$D7,$13,$65,$17,$CD
	.dc.b	$F4,$5F,$37,$75,$CF,$0C,$46,$8E,$55,$C7,$2F,$AB,$7A,$6C,$FC,$9C
	.dc.b	$3C,$78,$7B,$09,$8F,$A0,$F6,$C6,$9C,$4F,$89,$AF,$A2,$34,$E4,$87
	.dc.b	$96,$7D,$0F,$BE,$BC,$71,$E0,$8F,$F1,$A6,$1E,$9C,$BE,$CD,$6B,$0E
	.dc.b	$F3,$9C,$7D,$95,$F7,$31,$F0,$E1,$DE,$FC,$B8,$71,$8E,$B3,$A8,$C4
	.dc.b	$6C,$F3,$E1,$D5,$F3,$5F,$5C,$BB,$E3,$F4,$2B,$59,$7A,$6F,$91,$F0
	.dc.b	$EB,$6F,$D4,$C7,$C1,$87,$B0,$98,$7E,$3E,$B7,$7C,$F0,$EB,$97,$AA
	.dc.b	$FC,$58,$F3,$BF,$16,$3D,$31,$E5,$3F,$76,$7F,$57,$F5,$3F,$13,$1F
	.dc.b	$C6,$1F,$EA,$78,$FD,$7F,$AF,$9F,$67,$4C,$8C,$9F,$B4,$8C,$5F,$EC
	.dc.b	$5B,$A7,$EF,$9B,$A7,$E9,$1F,$CF,$F5,$FE,$82,$BF,$3B,$18,$BF,$D4
	.dc.b	$FB,$B6,$7F,$03,$37,$E4,$DF,$17,$FE,$9F,$BF,$F3,$C6,$1F,$AA,$9C
	.dc.b	$FC,$86,$EB,$58,$2B,$50,$9D,$7A,$FB,$6E,$1D,$0F,$F4,$26,$1D,$1C
	.dc.b	$60,$8F,$1F,$D4,$F7,$FA,$9F,$03,$6B,$73,$A6,$E2,$AD,$C6,$1F,$81
	.dc.b	$6F,$A8,$FF,$76,$4F,$83,$8D,$F5,$3F,$E8,$08,$F1,$D6,$FE,$0B,$E4
	.dc.b	$56,$B8,$79,$E1,$CA,$BE,$15,$FF,$74,$8C,$6F,$FC,$4B,$E4,$5F,$0A
	.dc.b	$E2,$E3,$9C,$C3,$BE,$1C,$1F,$F5,$45,$F0,$AD,$63,$D8,$F8,$DB,$7F
	.dc.b	$9F,$2B,$E0,$E1,$F5,$7C,$FF,$AC,$D8,$C9,$F1,$A3,$2F,$C8,$F6,$7E
	.dc.b	$E1,$F7,$7E,$A6,$C6,$49,$87,$87,$8B,$F0,$F3,$A8,$F2,$0F,$A8,$78
	.dc.b	$DF,$8F,$E0,$7D,$4F,$E2,$1F,$CB,$4E,$B1,$FC,$0C,$6B,$A3,$EB,$CE
	.dc.b	$70,$EA,$D9,$F9,$9C,$5F,$E9,$2B,$FA,$F3,$97,$FC,$4A,$DA,$FF,$69
	.dc.b	$37,$87,$E9,$FA,$7D,$B3,$FC,$E7,$D4,$CC,$3F,$17,$E0,$DC,$7E,$07
	.dc.b	$1A,$B1,$A7,$E8,$2A,$DF,$E7,$2B,$FA,$DB,$F9,$4F,$A0,$FD,$43,$FA
	.dc.b	$96,$B9,$3F,$32,$64,$F5,$FA,$3F,$90,$FA,$9F,$D6,$46,$0F,$CC,$38
	.dc.b	$CB,$E2,$CF,$C8,$38,$70,$5A,$E1,$3A,$78,$C6,$63,$EE,$0F,$A9,$FE
	.dc.b	$08,$D9,$FD,$7A,$F8,$DC,$CC,$9C,$BC,$23,$13,$D3,$EA,$79,$FC,$7F
	.dc.b	$46,$DF,$46,$CE,$E3,$BC,$FE,$E3,$EA,$5F,$15,$FC,$67,$91,$F5,$2F
	.dc.b	$1B,$53,$98,$C6,$1F,$BA,$BE,$2D,$9F,$E4,$5F,$50,$FA,$97,$D4,$F2
	.dc.b	$AD,$13,$4F,$D8,$56,$E7,$51,$FA,$0F,$A9,$72,$7F,$0E,$6D,$FF,$0A
	.dc.b	$F1,$BE,$43,$FC,$32,$33,$5F,$A8,$97,$B7,$FC,$59,$B5,$FC,$4B,$EA
	.dc.b	$1F,$29,$9A,$FE,$37,$18,$9D,$4F,$F1,$EF,$A9,$7D,$4E,$A3,$7F,$05
	.dc.b	$2B,$A7,$95,$73,$FD,$CE,$3F,$C5,$67,$27,$F2,$2F,$A9,$E6,$E2,$7F
	.dc.b	$7F,$18,$78,$D7,$F2,$E8,$C7,$81,$FD,$4E,$FA,$57,$F0,$6F,$21,$D3
	.dc.b	$FD,$4F,$17,$D3,$BE,$07,$F9,$BD,$6A,$7F,$A2,$9A,$8C,$D4,$8C,$D3
	.dc.b	$EA,$71,$87,$7A,$B9,$7F,$91,$2F,$BD,$61,$FF,$55,$8F,$86,$F2,$47
	.dc.b	$CD,$BF,$E3,$B3,$93,$C7,$19,$AF,$EE,$11,$99,$DD,$7B,$70,$7D,$07
	.dc.b	$F5,$2A,$D5,$F0,$AC,$1B,$9B,$3F,$D4,$BE,$A5,$F2,$7C,$CF,$EB,$6E
	.dc.b	$23,$E0,$C6,$08,$FC,$4F,$84,$B9,$7F,$36,$AE,$0D,$AF,$34,$EB,$1F
	.dc.b	$F0,$57,$C1,$7F,$0D,$7E,$46,$6F,$A7,$FE,$D7,$39,$C9,$8F,$F9,$EA
	.dc.b	$B5,$1F,$DA,$A7,$53,$A6,$C7,$E0,$3A,$79,$FB,$72,$38,$3F,$26,$31
	.dc.b	$5F,$D3,$BC,$4F,$C2,$3F,$D8,$E1,$F5,$0F,$86,$FE,$FF,$5A,$D3,$FD
	.dc.b	$07,$5A,$9C,$F1,$CC,$BF,$C0,$5F,$05,$FD,$C7,$2F,$EE,$33,$97,$C0
	.dc.b	$FF,$80,$A3,$50,$C7,$B5,$7F,$BD,$17,$D0,$7F,$61,$FA,$9C,$D7,$F7
	.dc.b	$C8,$C6,$1E,$66,$5F,$F8,$4F,$47,$66,$2F,$E0,$70,$FB,$0F,$F0,$DB
	.dc.b	$F4,$3E,$07,$E8,$FD,$4B,$EA,$7C,$55,$BC,$7F,$C3,$7A,$FD,$4C,$3F
	.dc.b	$BB,$9C,$45,$F1,$7F,$C2,$35,$D9,$61,$DB,$0E,$B8,$59,$FF,$6E,$CD
	.dc.b	$7E,$25,$7F,$AC,$AF,$A4,$7E,$18,$E6,$78,$2F,$9B,$4E,$65,$39,$DF
	.dc.b	$57,$D1,$3D,$30,$F6,$3D,$A3,$35,$D2,$31,$1E,$07,$85,$F9,$67,$FD
	.dc.b	$EB,$39,$78,$18,$C3,$D9,$7D,$A3,$13,$E5,$8C,$3E,$8B,$EB,$87,$9E
	.dc.b	$3A,$1F,$31,$C5,$F4,$C3,$C5,$3C,$70,$F0,$BF,$08,$F2,$9F,$20,$B3
	.dc.b	$87,$ED,$1F,$34,$7D,$4F,$F1,$BF,$53,$8B,$63,$DE,$D7,$83,$5F,$AC
	.dc.b	$BE,$47,$FC,$FD,$1C,$1A,$F0,$6B,$C1,$AF,$F0,$48,$E5,$C9,$F8,$5F
	.dc.b	$A9,$7D,$2F,$E0,$F5,$FC,$2C,$62,$FF,$15,$7C,$1F,$C1,$36,$77,$5C
	.dc.b	$73,$FC,$65,$F1,$BD,$E3,$E8,$BF,$C8,$E1,$C6,$F7,$87,$1B,$FE,$33
	.dc.b	$87,$1B,$DE,$1C,$6F,$D9,$8F,$A7,$0E,$18,$30,$E1,$5A,$8F,$0D,$6D
	.dc.b	$86,$09,$87,$A9,$CB,$E0,$8F,$83,$87,$86,$7C,$B8,$78,$30,$EF,$1F
	.dc.b	$36,$74,$F8,$E7,$F5,$D5,$FE,$C3,$AF,$88,$D8,$CE,$BE,$7C,$3C,$57
	.dc.b	$FE,$7A,$BF,$F4,$B3,$FE,$A2,$6B,$9A,$E2,$FF,$40,$FA,$9F,$F6,$03
	.dc.b	$FE,$A8,$FA,$87,$B1,$A8,$46,$1F,$F5,$6C,$F0,$7F,$D1,$2D,$CB,$FD
	.dc.b	$B5,$3B,$9F,$F4,$6B,$63,$FD,$BD,$3B,$CB,$F8,$62,$F9,$AF,$03,$6F
	.dc.b	$BC,$78,$EF,$FB,$9C,$60,$F4,$78,$99,$38,$CC,$8C,$90,$F3,$BE,$DF
	.dc.b	$A9,$FE,$7F,$1F,$EC,$86,$BF,$EA,$46,$E3,$EC,$F3,$78,$74,$78,$FE
	.dc.b	$A6,$5F,$18,$CB,$F2,$E3,$2F,$38,$F9,$9F,$3D,$FE,$61,$8F,$F6,$BB
	.dc.b	$5F,$F5,$9B,$6F,$E5,$F5,$F0,$F6,$3E,$17,$B5,$2F,$8C,$65,$FB,$51
	.dc.b	$97,$9C,$7D,$BF,$83,$FF,$29,$C7,$FB,$99,$AF,$FB,$09,$B6,$E1,$FE
	.dc.b	$81,$C7,$FE,$66,$AE,$1E,$87,$E1,$61,$F1,$14,$ED,$12,$7F,$DE,$73
	.dc.b	$A6,$CE,$5F,$F6,$8F,$A3,$F7,$DB,$C5,$F9,$18,$C6,$3E,$39,$FF,$0B
	.dc.b	$FD,$4B,$C8,$D5,$FE,$5E,$1F,$1F,$4C,$D7,$F8,$7A,$F9,$3F,$EF,$BA
	.dc.b	$D3,$9D,$3F,$AF,$5F,$28,$C1,$FE,$26,$C7,$C7,$3F,$D8,$FE,$A5,$E6
	.dc.b	$7E,$9D,$B5,$C0,$38,$7F,$DD,$F7,$F3,$AB,$95,$7E,$BF,$EA,$7F,$D4
	.dc.b	$F5,$C5,$AF,$FB,$DA,$F9,$65,$EA,$9C,$BE,$CF,$A9,$FF,$5F,$57,$85
	.dc.b	$AF,$FB,$FA,$77,$87,$F7,$B7,$FC,$D1,$1E,$7F,$A9,$7C,$27,$F0,$16
	.dc.b	$C7,$8F,$D4,$C4,$E7,$EA,$64,$F1,$8E,$1F,$49,$FC,$27,$EA,$5F,$53
	.dc.b	$4C,$D7,$C5,$F8,$AF,$F1,$3B,$7C,$FF,$E2,$11,$88,$6F,$F6,$3E,$45
	.dc.b	$E9,$EA,$FF,$8B,$43,$EA,$69,$9B,$E8,$7D,$38,$D3,$63,$4F,$6E,$1F
	.dc.b	$F0,$F7,$D4,$FC,$E6,$FB,$5F,$C8,$E3,$FC,$F7,$3F,$75,$D3,$FE,$6A
	.dc.b	$6A,$76,$F5,$35,$1C,$F9,$BA,$78,$D2,$7A,$3D,$FE,$8B,$11,$C4,$E9
	.dc.b	$18,$D7,$F3,$78,$2B,$E4,$FC,$BC,$BF,$9C,$FA,$7E,$47,$11,$CE,$9F
	.dc.b	$53,$ED,$33,$87,$E9,$B8,$58,$66,$FE,$05,$6B,$5F,$F2,$27,$D4,$3C
	.dc.b	$9F,$E9,$0B,$9A,$98,$89,$7F,$A5,$9D,$36,$72,$FB,$E1,$F5,$3D,$09
	.dc.b	$1C,$7C,$1A,$EE,$2E,$9F,$BE,$AF,$DB,$CF,$FA,$7B,$EA,$5F,$35,$FD
	.dc.b	$75,$B1,$FB,$76,$BF,$3F,$EA,$7F,$87,$39,$9F,$EC,$90,$F9,$8A,$73
	.dc.b	$89,$3F,$C0,$67,$4D,$9C,$BD,$7D,$8E,$64,$78,$78,$15,$D1,$CE,$3C
	.dc.b	$31,$E2,$EC,$F1,$F6,$4A,$F0,$46,$5F,$14,$65,$F6,$47,$F7,$DF,$A9
	.dc.b	$7D,$4D,$24,$FF,$19,$F5,$7F,$37,$3A,$8C,$7B,$1F,$E0,$59,$FE,$F9
	.dc.b	$CB,$FE,$85,$AD,$57,$E5,$DB,$C3,$FC,$C2,$B7,$3F,$ED,$06,$A9,$DD
	.dc.b	$7F,$A3,$E5,$D9,$DF,$B2,$5F,$96,$3C,$46,$1F,$D3,$4F,$E8,$DD,$1E
	.dc.b	$BE,$CF,$47,$D4,$C8,$C1,$0E,$2E,$5E,$71,$FE,$23,$FA,$9F,$E7,$5F
	.dc.b	$01,$FA,$49,$FF,$74,$FD,$47,$FB,$65,$FF,$5E,$77,$7F,$A9,$E3,$11
	.dc.b	$FE,$B2,$8F,$F5,$55,$72,$6F,$83,$FE,$AD,$BE,$94,$BE,$31,$86,$D7
	.dc.b	$09,$EE,$78,$BE,$1B,$94,$78,$5B,$19,$36,$30,$F1,$39,$3C,$C7,$B3
	.dc.b	$C4,$E2,$F8,$DB,$7E,$28,$F8,$3F,$53,$AF,$A8,$7D,$4F,$EA,$33,$F5
	.dc.b	$C6,$3D,$4F,$0E,$0B,$E6,$71,$BE,$1A,$7F,$B3,$EF,$86,$3E,$43,$E2
	.dc.b	$56,$F0,$F5,$CE,$B1,$EF,$87,$2B,$F2,$9A,$70,$A7,$86,$1D,$E7,$38
	.dc.b	$F9,$6B,$86,$11,$F1,$5F,$07,$A9,$8A,$EB,$19,$8E,$D1,$D3,$1F,$44
	.dc.b	$6E,$BA,$63,$F7,$67,$2F,$AA,$33,$8F,$E2,$E7,$2F,$9E,$33,$5E,$2A
	.dc.b	$E0,$F5,$23,$D5,$D9,$E4,$8B,$8F,$BE,$37,$5D,$5A,$75,$71,$5D,$88
	.dc.b	$F9,$63,$AE,$3F,$04,$F8,$55,$D6,$BC,$B3,$FE,$DF,$9F,$BA,$DF,$19
	.dc.b	$F1,$5F,$EE,$89,$CC,$FE,$0E,$72,$FD,$26,$D7,$EA,$3E,$A3,$FC,$0B
	.dc.b	$CD,$F8,$49,$CE,$5C,$71,$C1,$FA,$2A,$DE,$1C,$E3,$2D,$3F,$21,$1A
	.dc.b	$C3,$F3,$77,$D2,$47,$E2,$DC,$36,$BB,$75,$FF,$96,$6B,$8C,$76,$6C
	.dc.b	$F6,$9D,$61,$E7,$8F,$14,$E8,$F7,$4E,$A3,$C3,$93,$0F,$0B,$70,$ED
	.dc.b	$87,$9A,$98,$FA,$E3,$B6,$3D,$5A,$6E,$35,$3D,$B0,$EA,$4B,$F6,$E5
	.dc.b	$DA,$75,$15,$EF,$97,$9D,$F8,$87,$B3,$0E,$EF,$CA,$C3,$BB,$AC,$7C
	.dc.b	$8F,$43,$E3,$61,$D7,$2F,$74,$73,$39,$61,$DE,$33,$5C,$A7,$A6,$1E
	.dc.b	$DA,$D6,$5E,$B9,$E4,$7A,$30,$ED,$19,$BF,$A3,$18,$CB,$DD,$97,$A3
	.dc.b	$2E,$F9,$7B,$31,$EE,$79,$2F,$83,$DA,$B8,$63,$D6,$B8,$46,$30,$DE
	.dc.b	$3E,$0B,$F2,$E3,$D9,$F8,$91,$93,$DD,$19,$8F,$09,$EB,$38,$E1,$DD
	.dc.b	$F2,$18,$AF,$5E,$5F,$11,$FB,$B1,$E7,$8C,$3F,$48,$97,$FB,$1F,$A9
	.dc.b	$FB,$D8,$FD,$A3,$FD,$67,$9F,$EB,$FD,$7C,$FA,$FA,$64,$64,$FD,$A4
	.dc.b	$63,$0F,$D8,$B7,$0F,$DF,$37,$5F,$CB,$3F,$EC,$5E,$CF,$F5,$EE,$28
	.dc.b	$C6,$1F,$9B,$F7,$2E,$3F,$8F,$9B,$F2,$4F,$89,$FF,$50,$DF,$FB,$02
	.dc.b	$31,$F1,$DA,$9C,$DF,$F9,$EE,$73,$5B,$9C,$E5,$FE,$D7,$9C,$D6,$E7
	.dc.b	$35,$FC,$22,$73,$5B,$97,$D4,$F7,$FA,$9F,$03,$6F,$B3,$70,$E8,$DB
	.dc.b	$FF,$6E,$FD,$47,$FB,$D2,$BF,$DB,$3C,$6F,$A9,$FF,$3D,$C7,$FB,$56
	.dc.b	$75,$87,$FA,$56,$B5,$C1,$F1,$A1,$3A,$C3,$AF,$C0,$7C,$AF,$90,$7C
	.dc.b	$27,$CF,$CB,$A4,$F1,$BE,$11,$FC,$45,$B8,$2F,$8F,$0E,$EF,$E7,$79
	.dc.b	$FF,$3D,$B8,$7C,$D6,$DF,$CB,$6E,$5F,$25,$FA,$DE,$CC,$AF,$A9,$F5
	.dc.b	$36,$FE,$94,$DB,$FA,$13,$6B,$F9,$37,$D4,$3C,$6F,$C3,$D3,$EA,$7E
	.dc.b	$CE,$3F,$D0,$2B,$57,$FD,$0A,$BF,$25,$5A,$A5,$69,$FF,$5D,$36,$55
	.dc.b	$FD,$2B,$5F,$97,$E9,$7D,$1A,$D4,$F9,$31,$C1,$F4,$78,$3F,$D2,$51
	.dc.b	$E1,$34,$7E,$4F,$EA,$7F,$30,$7E,$67,$C0,$B1,$92,$67,$F9,$58,$CB
	.dc.b	$F9,$78,$FD,$07,$D0,$7E,$85,$A7,$E7,$CB,$1B,$7F,$45,$1E,$9F,$43
	.dc.b	$F5,$FC,$CF,$A9,$76,$E6,$FF,$2B,$18,$7D,$58,$7C,$DB,$D3,$F9,$59
	.dc.b	$F9,$9A,$A7,$3C,$1F,$2A,$B8,$5F,$4F,$82,$CB,$FA,$F8,$7D,$4F,$F0
	.dc.b	$06,$C7,$EF,$D7,$C6,$66,$64,$E5,$E1,$18,$9F,$E0,$43,$EA,$7A,$7C
	.dc.b	$77,$56,$DF,$56,$CF,$06,$E3,$E9,$87,$D4,$3E,$A3,$F5,$DF,$52,$F1
	.dc.b	$BF,$69,$39,$C7,$E3,$F4,$AB,$51,$F2,$B0,$E7,$87,$03,$F9,$0F,$D4
	.dc.b	$3E,$A5,$F5,$39,$BE,$19,$75,$C1,$A7,$96,$30,$FC,$FC,$39,$5F,$1A
	.dc.b	$DD,$FF,$83,$FE,$A5,$D1,$FC,$2D,$B1,$E5,$E8,$F9,$EF,$9E,$33,$5C
	.dc.b	$8F,$F1,$6C,$FF,$B0,$A7,$87,$B7,$FC,$57,$E9,$9F,$2D,$FC,$5E,$B5
	.dc.b	$7C,$44,$EA,$FE,$23,$63,$5F,$52,$FA,$9E,$98,$75,$F8,$39,$87,$0F
	.dc.b	$26,$E1,$D7,$1F,$E4,$0D,$8F,$E5,$1F,$53,$CE,$B8,$1F,$4D,$C4,$EE
	.dc.b	$31,$8F,$AE,$BF,$98,$7A,$2E,$AF,$EA,$F4,$FA,$67,$C1,$6A,$73,$18
	.dc.b	$9F,$C0,$57,$16,$CE,$BE,$A7,$84,$60,$99,$7C,$F7,$B4,$6A,$3E,$54
	.dc.b	$3E,$1F,$C7,$3F,$27,$9F,$F3,$13,$93,$F7,$22,$34,$FF,$94,$4D,$57
	.dc.b	$4A,$F4,$FD,$4F,$F4,$F9,$CD,$F4,$9C,$1A,$8F,$CE,$1A,$3F,$CB,$9F
	.dc.b	$52,$F9,$4F,$86,$7F,$59,$71,$1E,$4D,$2E,$3F,$85,$9B,$87,$EC,$5B
	.dc.b	$5F,$E7,$C3,$E8,$7C,$37,$3C,$24,$7E,$B6,$77,$7F,$CE,$B0,$5E,$A3
	.dc.b	$F2,$5F,$03,$FC,$75,$F9,$D2,$FA,$7F,$ED,$D3,$9C,$BF,$8D,$63,$F7
	.dc.b	$E7,$3E,$5C,$8F,$57,$4D,$C7,$AC,$DB,$E3,$8F,$FA,$AA,$FF,$81,$63
	.dc.b	$D5,$FF,$41,$78,$7F,$85,$C3,$FC,$01,$F5,$0F,$88,$FF,$00,$4E,$71
	.dc.b	$FE,$59,$9A,$B5,$87,$F0,$DA,$D3,$D3,$8F,$F9,$FF,$C1,$7F,$7E,$CB
	.dc.b	$FB,$F4,$74,$24,$FF,$85,$27,$55,$CA,$BF,$B1,$7D,$07,$F7,$0F,$A9
	.dc.b	$C4,$7F,$81,$E3,$18,$79,$99,$87,$F8,$7B,$17,$6E,$2F,$F8,$16,$3F
	.dc.b	$3E,$BD,$1C,$3F,$E1,$D9,$FF,$5C,$FC,$17,$FA,$EB,$E9,$DF,$51,$F1
	.dc.b	$AB,$E1,$92,$FF,$96,$5F,$01,$5C,$1F,$EF,$53,$D5,$5F,$5B,$E9,$76
	.dc.b	$7F,$D6,$86,$8E,$71,$A6,$9A,$9E,$0F,$F8,$A6,$BF,$B0,$5F,$86,$3F
	.dc.b	$D7,$D3,$AC,$3E,$04,$72,$32,$57,$33,$C6,$3D,$8F,$38,$CD,$F9,$A7
	.dc.b	$CD,$39,$BF,$0C,$E6,$FC,$11,$8A,$F4,$1D,$DC,$3E,$2A,$E5,$87,$B2
	.dc.b	$3C,$37,$F0,$67,$C9,$1E,$08,$C1,$DD,$E9,$3C,$5D,$1F,$49,$C5,$FA
	.dc.b	$63,$E0,$95,$C3,$F6,$CF,$A7,$F5,$3F,$C7,$FE,$13,$83,$63,$E1,$35
	.dc.b	$DF,$57,$FA,$25,$CF,$C9,$7F,$68,$6F,$27,$28,$C4,$76,$6B,$FE,$F5
	.dc.b	$A4,$71,$CF,$EE,$7D,$4B,$E9,$7F,$DF,$51,$86,$EB,$C6,$B8,$37,$1F
	.dc.b	$C3,$DF,$1C,$6D,$F0,$43,$1F,$0D,$7E,$39,$B1,$F8,$9A,$E4,$DA,$E4
	.dc.b	$F1,$C3,$C2,$F4,$AE,$38,$F9,$E3,$C2,$75,$9D,$BC,$70,$CB,$D4,$D3
	.dc.b	$1C,$12,$7F,$41,$5D,$B1,$F8,$13,$97,$DA,$B1,$F4,$D7,$0C,$3C,$31
	.dc.b	$EC,$C3,$FC,$64,$DD,$78,$B6,$FD,$95,$C3,$2E,$B7,$C3,$C2,$B8,$70
	.dc.b	$6D,$FC,$EC,$39,$2F,$A9,$FF,$31,$BF,$E3,$7F,$A9,$FF,$2E,$37,$D8
	.dc.b	$E2,$EF,$BA,$D7,$08,$FF,$0F,$57,$06,$D6,$AB,$7F,$37,$27,$FC,$89
	.dc.b	$68,$EB,$6F,$0F,$55,$D3,$C6,$C9,$8D,$7F,$CA,$D8,$71,$9F,$44,$E9
	.dc.b	$EB,$87,$A3,$EA,$5D,$74,$93,$EF,$F3,$6B,$EA,$FA,$9F,$19,$8C,$BF
	.dc.b	$CC,$97,$CA,$7E,$34,$E9,$FF,$2F,$46,$5F,$F2,$07,$CF,$FF,$99,$DF
	.dc.b	$F3,$13,$5C,$FB,$1F,$03,$C0,$F9,$B3,$9C,$3F,$CE,$5B,$FF,$A0,$AF
	.dc.b	$EA,$7B,$5A,$C2,$46,$08,$74,$3F,$CE,$1F,$03,$FE,$54,$6F,$A3,$F6
	.dc.b	$A3,$11,$FE,$55,$E8,$FF,$32,$BA,$3F,$CF,$0B,$E1,$FF,$02,$D8,$D1
	.dc.b	$FE,$5D,$9E,$79,$7F,$A2,$67,$53,$D6,$BC,$17,$F8,$AE,$8F,$B1,$3F
	.dc.b	$E6,$56,$DF,$FA,$45,$BD,$3F,$A8,$FA,$97,$95,$C7,$97,$F2,$70,$F9
	.dc.b	$0A,$67,$1F,$F4,$86,$1C,$9F,$E1,$B3,$B7,$FD,$28,$D7,$C5,$5F,$9E
	.dc.b	$FA,$9F,$F1,$4C,$7F,$A6,$5A,$FF,$9E,$DB,$97,$FA,$9E,$BE,$0E,$5F
	.dc.b	$BE,$9C,$BF,$C0,$3E,$A3,$FD,$55,$7F,$EA,$36,$BC,$FC,$FF,$D6,$C6
	.dc.b	$30,$FF,$59,$7D,$4F,$E0,$6B,$FD,$14,$D8,$FF,$5B,$4E,$EF,$B7,$8B
	.dc.b	$FB,$6F,$A9,$7C,$25,$3A,$C5,$C7,$FD,$69,$2F,$A9,$E0,$66,$3F,$D3
	.dc.b	$10,$FA,$5F,$E0,$3F,$52,$FA,$9A,$77,$8F,$E0,$BF,$17,$FE,$20,$BF
	.dc.b	$F6,$03,$5F,$F5,$13,$75,$FF,$67,$7C,$8B,$D5,$E3,$E9,$FF,$0E,$87
	.dc.b	$CF,$53,$38,$74,$9C,$3F,$EC,$D9,$DB,$93,$FD,$A1,$F5,$3F,$49,$FE
	.dc.b	$9A,$FF,$9D,$5C,$FB,$9F,$C7,$23,$50,$69,$B7,$FD,$74,$DC,$7F,$DC
	.dc.b	$CE,$BC,$7E,$44,$9F,$0F,$D1,$57,$8C,$5A,$E3,$3F,$23,$E5,$E5,$7C
	.dc.b	$87,$C9,$F2,$58,$8F,$0F,$D4,$F9,$CC,$5F,$FB,$DF,$75,$E6,$BE,$0D
	.dc.b	$6B,$9F,$FA,$8F,$D4,$3C,$9F,$E7,$6F,$DD,$AF,$F7,$B3,$5F,$F6,$C3
	.dc.b	$B7,$31,$C3,$8E,$46,$0E,$3F,$53,$D0,$8F,$4F,$02,$BD,$22,$E8,$C7
	.dc.b	$CB,$5F,$D3,$BE,$A5,$F3,$54,$98,$7F,$81,$AF,$83,$FE,$08,$FA,$9F
	.dc.b	$E0,$AE,$70,$FE,$DF,$7C,$A3,$06,$FE,$67,$FA,$DD,$1E,$B1,$FE,$F7
	.dc.b	$9E,$0F,$FB,$E6,$74,$F2,$0E,$1E,$18,$74,$F6,$64,$6E,$7F,$C5,$6D
	.dc.b	$EF,$CB,$D6,$6E,$5E,$C8,$EB,$3E,$6E,$BF,$3F,$54,$9E,$91,$97,$B4
	.dc.b	$65,$F4,$47,$F7,$8F,$A8,$7D,$4D,$24,$FA,$FD,$6F,$D0,$CE,$A3,$1E
	.dc.b	$D7,$F8,$02,$7F,$BE,$65,$E5,$AD,$52,$31,$7B,$B9,$3F,$1E,$B7,$5E
	.dc.b	$DB,$75,$F5,$EB,$97,$E3,$8F,$01,$87,$B7,$17,$F8,$AA,$95,$A9,$F5
	.dc.b	$75,$F9,$FE,$A6,$46,$08,$72,$3F,$C9,$F4,$8D,$C6,$44,$27,$FC,$43
	.dc.b	$F5,$3F,$D0,$BE,$0B,$F5,$F5,$FE,$58,$FA,$7F,$F6,$C3,$FE,$BA,$EE
	.dc.b	$FF,$1F,$C6,$23,$FD,$63,$1F,$E4,$2A,$E4,$DF,$2F,$FD,$4E,$F0,$F8
	.dc.b	$9B,$5D,$5A,$AB,$8B,$93,$C0,$53,$B9,$70,$F3,$BC,$6F,$9E,$1E,$59
	.dc.b	$CB,$FC,$AA,$78,$87,$11,$AB,$E3,$F5,$1E,$4F,$A8,$7D,$4F,$EC,$AB
	.dc.b	$BC,$63,$D4,$F2,$E3,$CE,$F8,$B8,$9D,$3E,$06,$D7,$92,$7B,$5F,$5D
	.dc.b	$3F,$CD,$75,$A6,$C6,$6F,$EA,$57,$0C,$3B,$4E,$65,$7C,$5D,$9A,$98
	.dc.b	$F8,$E7,$86,$4B,$E9,$39,$AD,$E3,$E7,$8B,$8F,$92,$6C,$76,$38,$12
	.dc.b	$7E,$3E,$1D,$AF,$AC,$62,$B8,$D7,$C9,$2E,$3E,$97,$D7,$7C,$5C,$BE
	.dc.b	$F8,$D6,$3E,$99,$CD,$74,$8C,$47,$8F,$1F,$43,$4F,$46,$1E,$37,$07
	.dc.b	$82,$BE,$ED,$72,$9D,$12,$B9,$BF,$16,$BE,$09,$FE,$A4,$39,$1F,$EF
	.dc.b	$DB,$78,$AF,$8C,$FF,$74,$47,$E1,$A7,$83,$A8,$FC,$54,$EE,$7F,$33
	.dc.b	$F5,$1F,$E0,$DE,$4C,$D4,$9F,$C7,$56,$F0,$FD,$2D,$EE,$73,$5B,$9F
	.dc.b	$CD,$D6,$F1,$FF,$5D,$DE,$E7,$35,$B9,$FC,$AD,$6E,$FF,$DB,$B7,$B9
	.dc.b	$CD,$6E,$7F,$DC,$38,$FA,$DF,$4F,$57,$F5,$EC,$7A,$C7,$90,$8F,$A2
	.dc.b	$35,$7D,$27,$2D,$9D,$65,$DE,$7F,$43,$97,$95,$B8,$FE,$DB,$2F,$34
	.dc.b	$79,$27,$59,$79,$70,$F1,$9D,$27,$B9,$BB,$ED,$7C,$A3,$A9,$A2,$CF
	.dc.b	$13,$9E,$3E,$72,$47,$C8,$CB,$DA,$F5,$39,$61,$E3,$9F,$A7,$3A,$9E
	.dc.b	$B1,$C9,$A7,$3C,$BE,$15,$71,$8E,$97,$E3,$AE,$13,$99,$E2,$30,$F6
	.dc.b	$46,$49,$97,$CB,$8C,$BF,$20,$99,$7C,$DC,$BC,$D5,$C2,$3E,$3C,$62
	.dc.b	$77,$7F,$1E,$F9,$D7,$A2,$7B,$56,$E3,$07,$5C,$AD,$F1,$9D,$3F,$22
	.dc.b	$B8,$CF,$80,$91,$B7,$93,$FA,$AF,$A9,$F9,$98,$7D,$B3,$FD,$61,$3F
	.dc.b	$87,$F5,$F3,$EB,$E9,$91,$93,$F1,$91,$8B,$FD,$AB,$75,$FD,$8B,$75
	.dc.b	$FD,$6B,$FE,$3E,$EA,$FF,$12,$62,$8C,$61,$FB,$2F,$72,$E3,$FA,$E9
	.dc.b	$BF,$24,$F8,$BF,$F5,$1D,$FF,$90,$A3,$1F,$0F,$A9,$CD,$FF,$81,$67
	.dc.b	$35,$B9,$CE,$5F,$E5,$09,$CD,$6E,$73,$48,$4E,$AB,$85,$3E,$A3,$C1
	.dc.b	$F5,$3E,$16,$E1,$DD,$B8,$75,$6D,$FF,$91,$7E,$A3,$FD,$E9,$5F,$E5
	.dc.b	$CE,$47,$D4,$FF,$A0,$A3,$FC,$AD,$3A,$C3,$FC,$41,$5A,$E0,$F8,$F0
	.dc.b	$9D,$61,$D7,$E0,$7E,$5F,$C8,$BE,$13,$E8,$65,$D2,$78,$DF,$08,$F8
	.dc.b	$18,$73,$9E,$35,$C2,$3F,$C5,$95,$B7,$E2,$ED,$FE,$7E,$BF,$9D,$7F
	.dc.b	$4F,$CD,$FA,$8D,$8C,$9F,$4E,$32,$FD,$5F,$63,$ED,$1F,$6F,$EA,$6C
	.dc.b	$64,$98,$7D,$78,$CB,$FE,$3F,$FA,$6F,$F1,$F4,$FF,$A1,$B5,$7D,$4F
	.dc.b	$DC,$AF,$B9,$5A,$7F,$C9,$53,$AF,$81,$F9,$73,$98,$FF,$3D,$36,$FF
	.dc.b	$09,$BF,$CB,$F5,$37,$E4,$49,$FF,$46,$56,$F1,$5E,$E7,$35,$F9,$7F
	.dc.b	$A9,$91,$F9,$5F,$02,$DF,$E4,$E6,$E5,$F9,$09,$B5,$F9,$5E,$9F,$D2
	.dc.b	$3F,$A1,$FA,$0F,$D1,$34,$FD,$01,$70,$FC,$F3,$6F,$DB,$E8,$7E,$C3
	.dc.b	$A1,$F5,$2E,$DF,$87,$39,$8C,$5F,$BA,$32,$FE,$66,$3E,$76,$0D,$D5
	.dc.b	$A8,$FA,$CF,$3E,$1C,$65,$FB,$39,$7D,$4F,$F0,$06,$E3,$FB,$F5,$F1
	.dc.b	$99,$99,$39,$78,$46,$27,$F8,$10,$FA,$9E,$9F,$1F,$D5,$B7,$D5,$B3
	.dc.b	$C1,$B8,$7F,$AE,$FE,$99,$F5,$1F,$AC,$FA,$97,$8D,$FB,$29,$CE,$3F
	.dc.b	$ED,$6E,$85,$6A,$3F,$DA,$13,$AC,$BF,$D2,$CD,$C7,$FD,$C7,$F5,$0F
	.dc.b	$A9,$7D,$4F,$5D,$26,$3C,$9F,$E4,$F5,$B9,$CE,$3F,$EA,$1F,$A9,$74
	.dc.b	$7F,$06,$6E,$1F,$EE,$1E,$8F,$96,$FF,$B8,$E3,$35,$FE,$A9,$9F,$F5
	.dc.b	$EC,$EF,$DB,$FE,$1F,$F5,$0F,$95,$FE,$23,$5A,$BF,$F5,$84,$E6,$FF
	.dc.b	$DD,$CD,$8F,$F5,$A7,$D4,$BE,$A3,$9E,$1D,$3E,$0A,$61,$C3,$D0,$B8
	.dc.b	$74,$7C,$27,$F2,$0F,$A9,$E5,$3B,$3F,$DF,$8E,$27,$84,$63,$0E,$2F
	.dc.b	$FC,$01,$F0,$5D,$5F,$D4,$21,$F4,$FF,$EC,$87,$51,$98,$C5,$FF,$B3
	.dc.b	$6B,$8D,$68,$FE,$91,$CC,$7F,$A5,$1D,$3E,$A7,$F9,$BC,$F0,$AE,$58
	.dc.b	$60,$ED,$36,$B9,$9F,$CD,$07,$C3,$79,$E3,$E6,$61,$DE,$FB,$CE,$76
	.dc.b	$FE,$8B,$7C,$0E,$73,$A6,$9F,$03,$E8,$77,$52,$37,$3E,$B8,$F8,$53
	.dc.b	$98,$FF,$77,$7D,$4B,$E4,$FE,$31,$FD,$51,$C5,$F6,$D6,$D7,$0A,$D1
	.dc.b	$F3,$7E,$0E,$D7,$82,$F7,$9F,$B6,$D1,$89,$F9,$B8,$2F,$8D,$7F,$C0
	.dc.b	$1F,$80,$FE,$2E,$B0,$F8,$D7,$F7,$BE,$9F,$FB,$4D,$6A,$73,$82,$73
	.dc.b	$E5,$4B,$F5,$74,$5C,$7B,$4D,$C3,$8C,$DA,$FE,$D7,$E1,$7D,$F9,$FA
	.dc.b	$D2,$FA,$87,$C2,$FF,$7A,$AD,$5F,$F8,$46,$B5,$39,$E4,$93,$FE,$07
	.dc.b	$F8,$1F,$EE,$59,$FF,$4E,$9C,$BE,$F9,$F0,$C7,$F3,$D9,$DE,$1D,$AB
	.dc.b	$FA,$DF,$D1,$FF,$8C,$33,$7D,$4E,$67,$FA,$CC,$E6,$31,$48,$F4,$1F
	.dc.b	$D2,$74,$76,$63,$3F,$C4,$D7,$CB,$6E,$FF,$E1,$8C,$7F,$C5,$DF,$03
	.dc.b	$FE,$2F,$FA,$77,$D4,$7C,$8A,$F8,$84,$BF,$C9,$DF,$01,$5C,$1F,$EE
	.dc.b	$B3,$D5,$5F,$5B,$E9,$76,$7F,$DB,$13,$5F,$EE,$F5,$FA,$BB,$E9,$1F
	.dc.b	$E2,$49,$D6,$1E,$B8,$F0,$E1,$FE,$E1,$7B,$98,$3C,$78,$75,$7D,$17
	.dc.b	$D2,$B9,$3D,$A3,$35,$E6,$79,$1E,$12,$CF,$08,$F0,$5F,$93,$0F,$44
	.dc.b	$62,$B7,$18,$7D,$E6,$30,$F2,$3E,$8C,$3C,$93,$B8,$F0,$46,$27,$77
	.dc.b	$DD,$B3,$CC,$F3,$9F,$54,$F8,$87,$C4,$2C,$61,$D1,$FE,$50,$FA,$9F
	.dc.b	$E3,$BE,$EF,$06,$C7,$CD,$6B,$FE,$56,$0E,$1E,$5F,$23,$FB,$F7,$83
	.dc.b	$5E,$10,$8C,$47,$0E,$0F,$04,$71,$C7,$F0,$3F,$52,$FA,$5F,$F3,$2C
	.dc.b	$61,$BA,$F1,$C3,$EB,$E3,$D7,$1F,$C6,$61,$C6,$6D,$F0,$7F,$D1,$D5
	.dc.b	$B7,$8E,$3E,$1C,$3F,$CE,$77,$C6,$7F,$06,$DC,$7F,$2B,$7C,$5B,$7C
	.dc.b	$6B,$FD,$35,$5B,$BE,$78,$78,$67,$E4,$D6,$DB,$8F,$99,$87,$91,$FD
	.dc.b	$36,$3D,$EF,$9C,$EB,$1F,$0D,$6A,$FA,$63,$E1,$AF,$1E,$3E,$1C,$79
	.dc.b	$46,$0F,$E9,$F3,$EF,$BF,$E8,$0D,$D3,$E7,$63,$F4,$6B,$7E,$27,$FA
	.dc.b	$62,$F6,$DA,$F2,$CF,$F2,$33,$FD,$79,$F5,$3F,$E5,$D7,$FA,$EF,$D3
	.dc.b	$FF,$95,$9B,$EB,$F1,$77,$D7,$55,$71,$7A,$0A,$E4,$E1,$B8,$FF,$59
	.dc.b	$9E,$0D,$F9,$8E,$90,$69,$DE,$5E,$16,$2F,$CF,$F2,$59,$30,$F8,$AF
	.dc.b	$93,$A2,$19,$BF,$57,$D4,$7F,$39,$8F,$EF,$AD,$7D,$6D,$C7,$FB,$86
	.dc.b	$2F,$37,$8F,$5F,$97,$EA,$66,$1C,$A3,$2F,$CC,$8C,$BD,$23,$E5,$7C
	.dc.b	$EF,$F9,$A2,$3F,$C1,$AD,$7E,$5B,$70,$F0,$F8,$1C,$3D,$8F,$89,$ED
	.dc.b	$4B,$E3,$19,$7E,$DC,$65,$E7,$1F,$67,$E2,$7F,$CB,$11,$FE,$1C,$6B
	.dc.b	$F6,$9B,$7F,$E0,$FE,$9F,$F3,$4C,$EF,$CC,$FC,$14,$3E,$22,$9D,$62
	.dc.b	$4F,$F8,$A2,$74,$D9,$CB,$F8,$8E,$8F,$AE,$DE,$2F,$C7,$46,$30,$FF
	.dc.b	$10,$7D,$4B,$CA,$CD,$7E,$4E,$1F,$21,$4C,$E1,$FE,$38,$C3,$97,$5F
	.dc.b	$FB,$F2,$B7,$3F,$E3,$C6,$BE,$13,$FC,$7F,$F5,$3F,$E2,$B6,$F9,$BF
	.dc.b	$A1,$8C,$47,$E8,$27,$E3,$D2,$24,$EB,$3F,$F1,$C7,$D4,$7F,$AA,$27
	.dc.b	$FC,$96,$D7,$F5,$33,$BA,$E5,$09,$D5,$7F,$90,$7E,$A7,$FD,$83,$3F
	.dc.b	$E5,$46,$BF,$B2,$9D,$D7,$27,$FC,$CF,$1F,$E5,$AF,$A8,$7C,$27,$EE
	.dc.b	$DB,$1E,$4F,$53,$13,$9F,$A9,$93,$C6,$38,$7D,$1F,$F0,$3F,$A9,$7D
	.dc.b	$4D,$33,$5F,$E6,$8F,$8B,$FF,$11,$B7,$D1,$FE,$15,$18,$8F,$E1,$3F
	.dc.b	$22,$F5,$73,$5F,$C3,$E1,$F5,$34,$CD,$73,$3F,$CE,$B1,$A6,$C6,$9F
	.dc.b	$E2,$BF,$53,$F3,$5F,$AC,$FF,$9D,$DF,$F3,$EB,$5D,$F3,$7F,$1F,$8E
	.dc.b	$10,$75,$C9,$FE,$73,$6F,$1F,$D8,$76,$7F,$07,$9E,$BE,$2F,$33,$A7
	.dc.b	$D1,$58,$C9,$0F,$3B,$F1,$BE,$5E,$47,$CB,$7C,$9F,$25,$88,$F6,$FD
	.dc.b	$4F,$9C,$C6,$5F,$E8,$7E,$05,$E6,$BD,$F5,$AC,$FF,$1F,$F5,$2F,$23
	.dc.b	$F9,$F2,$E5,$A6,$22,$5F,$FA,$92,$74,$D9,$CB,$FD,$12,$1F,$53,$D0
	.dc.b	$91,$C7,$C1,$AE,$E2,$E9,$FE,$9F,$AF,$DC,$CF,$EB,$7E,$A5,$F3,$5F
	.dc.b	$D4,$DB,$1F,$EB,$66,$BF,$D5,$3E,$A7,$F8,$43,$99,$FE,$B5,$0F,$99
	.dc.b	$A7,$38,$93,$FE,$C0,$9D,$36,$72,$F5,$F6,$B9,$91,$E1,$DF,$5D,$1C
	.dc.b	$E3,$C3,$1E,$2E,$BF,$1F,$5C,$AF,$04,$65,$F1,$46,$5F,$64,$7F,$73
	.dc.b	$FA,$87,$D4,$D2,$4F,$FB,$63,$D6,$FC,$CC,$EA,$31,$EC,$FF,$7B,$9F
	.dc.b	$EE,$BC,$BF,$C4,$2B,$55,$FD,$F1,$BC,$3F,$ED,$1A,$DC,$FF,$1F,$6A
	.dc.b	$9D,$D7,$F1,$A9,$FD,$14,$FF,$81,$97,$57,$84,$F1,$75,$C9,$F4,$46
	.dc.b	$5F,$54,$79,$FA,$FD,$1F,$53,$27,$DD,$19,$7E,$04,$65,$F0,$47,$F8
	.dc.b	$67,$EA,$7F,$9C,$FC,$0F,$E9,$27,$F9,$C7,$D3,$FF,$B4,$9F,$F5,$9F
	.dc.b	$77,$F3,$28,$C4,$7F,$AA,$A3,$F9,$AD,$72,$6F,$83,$FE,$A5,$78,$7C
	.dc.b	$3D,$AE,$AD,$55,$C5,$C9,$E0,$29,$DC,$B8,$79,$DE,$37,$CF,$0F,$2C
	.dc.b	$E5,$FE,$45,$3C,$43,$88,$D5,$F1,$FA,$9F,$27,$D4,$3E,$A7,$F4,$F5
	.dc.b	$DE,$31,$E9,$F9,$71,$E7,$7C,$5C,$4E,$9F,$03,$6B,$C9,$3D,$AF,$AE
	.dc.b	$9F,$D7,$AB,$4D,$8C,$DF,$D4,$AE,$18,$76,$9C,$CA,$F8,$BB,$35,$31
	.dc.b	$F1,$CF,$0C,$97,$D2,$73,$5B,$C7,$CF,$17,$1F,$24,$D8,$EC,$70,$24
	.dc.b	$FC,$7C,$3B,$5F,$58,$C5,$71,$AF,$92,$5C,$7D,$2F,$AE,$F8,$B9,$7D
	.dc.b	$F1,$AC,$7D,$33,$9A,$E9,$18,$8F,$1E,$3E,$86,$9E,$8C,$3C,$6E,$0F
	.dc.b	$05,$7D,$DA,$E5,$3A,$25,$73,$7E,$2D,$71,$BF,$49,$E5,$C7,$E6,$47
	.dc.b	$FB,$56,$3F,$08,$DF,$15,$F1,$7F,$EE,$49,$CC,$FE,$0E,$72,$FE,$29
	.dc.b	$B7,$FA,$1F,$A8,$FF,$03,$72,$FE,$42,$73,$97,$1C,$70,$7E,$5E,$B7
	.dc.b	$87,$38,$CB,$4F,$D3,$46,$B0,$FD,$05,$F4,$9F,$F1,$E6,$28,$C4,$7E
	.dc.b	$45,$B5,$C7,$0F,$14,$F5,$BE,$78,$78,$FA,$99,$78,$A7,$35,$BB,$FF
	.dc.b	$20,$E1,$CE,$3C,$38,$F8,$32,$F1,$65,$C6,$F7,$97,$4B,$E4,$70,$9D
	.dc.b	$3F,$E5,$5C,$3A,$3E,$98,$CB,$4F,$87,$3C,$F1,$F2,$92,$FD,$AF,$D3
	.dc.b	$8D,$46,$09,$3E,$13,$9E,$5D,$B1,$F4,$C7,$98,$63,$E4,$8C,$E1,$DE
	.dc.b	$79,$3D,$EB,$96,$5F,$0C,$FA,$38,$F8,$AB,$78,$F8,$9F,$65,$F1,$7C
	.dc.b	$18,$76,$9D,$63,$E1,$C3,$A6,$5F,$12,$30,$F8,$0F,$B5,$95,$C3,$D5
	.dc.b	$3A,$7A,$D7,$92,$75,$8F,$AE,$7D,$F9,$79,$DF,$2E,$3E,$7C,$BC,$91
	.dc.b	$F5,$E7,$51,$88,$D9,$EE,$35,$7E,$79,$DE,$1E,$19,$F2,$C7,$7C,$7F
	.dc.b	$02,$3E,$A7,$11,$9F,$5A,$61,$FA,$9F,$5A,$E1,$FA,$69,$BA,$7E,$CE
	.dc.b	$6E,$9F,$98,$6D,$7E,$3D,$AF,$DE,$7F,$2D,$D7,$F9,$78,$C6,$1F,$EA
	.dc.b	$98,$C6,$6F,$6F,$73,$D4,$D1,$FE,$AF,$F9,$27,$C5,$7F,$A7,$23,$38
	.dc.b	$27,$3F,$0D,$CA,$31,$5B,$58,$FE,$A6,$B7,$7C,$9E,$BE,$A7,$17,$35
	.dc.b	$C2,$7C,$57,$CE,$F9,$2C,$38,$2F,$A8,$BF,$53,$86,$DF,$79,$D3,$67
	.dc.b	$4D,$3D,$B8,$3E,$A3,$FD,$E5,$1F,$03,$99,$F5,$3F,$E8,$08,$DC,$EA
	.dc.b	$BD,$07,$F3,$29,$D3,$83,$A7,$BF,$F1,$DB,$5F,$33,$E4,$1F,$06,$E3
	.dc.b	$D6,$3D,$77,$FB,$8D,$BF,$F6,$DB,$70,$61,$CA,$F8,$E3,$F1,$BC,$FF
	.dc.b	$8E,$DC,$3E,$23,$6F,$D3,$C3,$F8,$08,$C6,$5F,$41,$F9,$FE,$CC,$BF
	.dc.b	$99,$F5,$37,$0E,$3C,$2A,$C6,$1F,$62,$AD,$FF,$BC,$BE,$99,$E3,$7E
	.dc.b	$26,$9F,$53,$F7,$EB,$53,$98,$C6,$41,$7C,$AB,$EC,$B6,$3E,$9E,$FF
	.dc.b	$33,$D2,$FA,$13,$9C,$7F,$4C,$DB,$EE,$7D,$3E,$2F,$F4,$D3,$D3,$2F
	.dc.b	$C7,$FD,$4C,$8F,$C6,$78,$37,$3F,$83,$37,$1F,$CB,$CD,$AF,$CD,$9F
	.dc.b	$98,$FA,$0F,$CC,$B4,$FC,$B1,$6F,$E5,$B6,$BC,$FE,$87,$EB,$FA,$1F
	.dc.b	$53,$FA,$3A,$F8,$33,$98,$C5,$FE,$47,$54,$FC,$23,$5D,$B6,$F8,$92
	.dc.b	$BF,$37,$87,$69,$D4,$7E,$18,$7D,$4F,$EE,$5B,$5F,$BD,$5F,$17,$99
	.dc.b	$93,$97,$84,$62,$7A,$7D,$4F,$3F,$8D,$E8,$DB,$E8,$D9,$DC,$77,$9F
	.dc.b	$D6,$7D,$43,$E9,$FF,$8A,$73,$3E,$A7,$F9,$05,$FE,$FA,$B4,$FE,$C6
	.dc.b	$33,$87,$ED,$E7,$2D,$BF,$E3,$DF,$50,$FA,$97,$D4,$F8,$26,$63,$D5
	.dc.b	$DC,$EA,$30,$7F,$0E,$FA,$97,$37,$F0,$36,$C7,$F1,$2E,$2F,$92,$EE
	.dc.b	$7F,$84,$43,$D9,$A7,$F0,$BF,$A9,$7C,$AE,$2F,$F8,$DC,$61,$B1,$FC
	.dc.b	$47,$EA,$5F,$51,$C6,$39,$7C,$14,$AE,$3E,$4D,$E1,$FE,$35,$8F,$F1
	.dc.b	$59,$C9,$FC,$A3,$EA,$79,$B8,$9F,$E4,$11,$87,$8D,$7F,$21,$8C,$77
	.dc.b	$7F,$57,$BE,$95,$FC,$9F,$C9,$74,$7F,$4D,$C5,$F4,$CF,$82,$FE,$55
	.dc.b	$5A,$9F,$EA,$A6,$A3,$35,$23,$34,$FA,$9C,$61,$DE,$AE,$5F,$E3,$FB
	.dc.b	$EF,$58,$7F,$88,$57,$F9,$3B,$E1,$BC,$93,$F3,$63,$F9,$F4,$A7,$3B
	.dc.b	$FF,$3F,$8C,$1F,$D3,$8B,$3F,$D2,$1F,$6F,$D0,$6E,$A4,$6D,$F8,$59
	.dc.b	$27,$35,$FD,$37,$EA,$1F,$53,$A3,$FA,$73,$8C,$3E,$0C,$60,$8F,$C4
	.dc.b	$F8,$4B,$5C,$E6,$E9,$F0,$F1,$56,$2F,$E1,$FC,$17,$F0,$EA,$D7,$D3
	.dc.b	$FF,$5D,$8C,$4F,$F3,$73,$EC,$C6,$7D,$99,$1F,$53,$AA,$E7,$FD,$9E
	.dc.b	$F8,$BF,$CF,$DB,$3F,$82,$9D,$E1,$FE,$1F,$F0,$FF,$05,$7F,$C6,$3E
	.dc.b	$A1,$F0,$DF,$DB,$E7,$35,$FE,$7B,$B5,$6B,$0F,$F0,$0D,$69,$FB,$ED
	.dc.b	$C3,$FD,$5B,$F0,$3F,$DB,$DF,$ED,$FC,$BD,$49,$5F,$E0,$79,$D5,$71
	.dc.b	$AF,$EE,$BF,$41,$FD,$7F,$EA,$71,$28,$C6,$1E,$76,$61,$FE,$11,$C5
	.dc.b	$DB,$8C,$7F,$C1,$98,$38,$3D,$C7,$F8,$36,$3F,$B3,$7C,$17,$F6,$7F
	.dc.b	$A6,$7D,$4F,$B2,$B7,$97,$C0,$78,$5A,$F8,$56,$EB,$FC,$4B,$3D,$05
	.dc.b	$F4,$7F,$BA,$61,$E3,$9E,$31,$FA,$83,$A1,$C2,$BA,$34,$D5,$F0,$7F
	.dc.b	$C3,$78,$7F,$AC,$70,$ED,$1F,$B2,$9D,$61,$DE,$39,$19,$2B,$99,$E3
	.dc.b	$1E,$97,$9C,$66,$FC,$D3,$E6,$9C,$DF,$86,$73,$7E,$08,$C5,$7A,$0E
	.dc.b	$EE,$1F,$15,$72,$C3,$D9,$1E,$1B,$F8,$33,$E4,$8F,$04,$60,$EE,$F4
	.dc.b	$9E,$2E,$8F,$9F,$5D,$DF,$A8,$7A,$8F,$9E,$5B,$C3,$A3,$F8,$DF,$D4
	.dc.b	$FF,$17,$F7,$38,$36,$3E,$1B,$5F,$B4,$D7,$8F,$C8,$FF,$9F,$9D,$B5
	.dc.b	$DB,$5D,$F1,$F6,$8E,$19,$FD,$CF,$A9,$7D,$2F,$F2,$E8,$C3,$75,$E3
	.dc.b	$87,$DE,$C5,$87,$29,$FE,$6B,$87,$2A,$FC,$53,$5E,$13,$B7,$FC,$95
	.dc.b	$5C,$B1,$DE,$1D,$E7,$FC,$9D,$7C,$71,$FC,$8D,$EE,$79,$DF,$06,$18
	.dc.b	$24,$FA,$9C,$BC,$B0,$F1,$42,$BA,$4E,$E3,$F3,$77,$E1,$C7,$E0,$CF
	.dc.b	$1C,$39,$C7,$C0,$C7,$D1,$7C,$F1,$EB,$87,$C1,$BF,$F6,$23,$75,$E1
	.dc.b	$5F,$EA,$0C,$BC,$38,$E2,$3F,$D3,$15,$FE,$93,$7F,$D4,$AD,$73,$8F
	.dc.b	$F6,$4C,$3F,$DA,$1F,$53,$FE,$B5,$9F,$F5,$5F,$D4,$7F,$94,$1B,$EB
	.dc.b	$EF,$3F,$ED,$75,$B9,$FF,$53,$36,$FF,$DB,$57,$C6,$7F,$D5,$4D,$8E
	.dc.b	$FD,$7C,$E7,$FD,$09,$4F,$9A,$C4,$23,$0F,$FA,$F9,$AF,$4F,$07,$71
	.dc.b	$8B,$F0,$F8,$BC,$B1,$FF,$08,$5F,$1C,$36,$66,$64,$64,$FF,$65,$FD
	.dc.b	$4F,$F3,$D6,$F4,$F9,$23,$11,$EC,$F3,$71,$3D,$FE,$A6,$4F,$02,$1F
	.dc.b	$09,$CB,$FE,$CD,$AE,$79,$7F,$C0,$1F,$9E,$FF,$7F,$4F,$FB,$71,$AF
	.dc.b	$FB,$4D,$B1,$F3,$BB,$7E,$3C,$63,$83,$FC,$3F,$AF,$F9,$EA,$FE,$DF
	.dc.b	$B3,$AF,$8C,$91,$82,$1D,$5F,$A3,$F0,$5F,$E5,$48,$FF,$76,$B5,$FF
	.dc.b	$72,$37,$26,$1F,$E2,$EE,$5F,$F3,$2D,$70,$F2,$FE,$DC,$3E,$1E,$9D
	.dc.b	$A2,$4F,$FB,$EA,$74,$D9,$CB,$FE,$F3,$F4,$7E,$F5,$FF,$8F,$1B,$1F
	.dc.b	$EC,$D6,$AF,$A9,$79,$3F,$8C,$9C,$C3,$E4,$3F,$1E,$DA,$FF,$6B,$67
	.dc.b	$EC,$7A,$73,$F1,$37,$52,$37,$8F,$8E,$7F,$33,$F5,$3F,$E3,$09,$C9
	.dc.b	$FE,$E1,$70,$AF,$B2,$8C,$C6,$30,$F9,$D7,$FE,$5A,$9C,$BF,$E5,$CF
	.dc.b	$A8,$FF,$54,$5F,$60,$E1,$ED,$1C,$7C,$DF,$D6,$CE,$70,$F0,$FD,$4F
	.dc.b	$E1,$EF,$67,$FB,$DD,$C4,$79,$A7,$85,$F8,$7D,$2F,$DB,$FD,$4B,$E1
	.dc.b	$29,$DA,$2E,$3E,$59,$7D,$4F,$03,$31,$E0,$10,$FA,$4F,$E0,$7F,$52
	.dc.b	$FA,$9A,$78,$23,$F8,$3F,$C5,$FF,$8B,$AF,$E1,$35,$E8,$DD,$7E,$7F
	.dc.b	$C8,$BD,$5E,$3E,$9F,$F1,$08,$7D,$4D,$33,$87,$49,$C3,$F3,$E7,$6E
	.dc.b	$4F,$A5,$F5,$3F,$59,$07,$FC,$FE,$EB,$DB,$FE,$3F,$1B,$8D,$72,$E2
	.dc.b	$64,$74,$8D,$F9,$12,$3B,$BC,$BE,$8B,$11,$F2,$EF,$FD,$A0,$E8,$F2
	.dc.b	$C7,$C8,$F9,$B9,$1F,$21,$F1,$FC,$96,$23,$97,$D4,$F9,$8C,$6F,$F9
	.dc.b	$8E,$05,$E6,$BE,$0D,$6B,$A3,$FD,$63,$F5,$0F,$27,$F9,$EB,$F7,$6B
	.dc.b	$F3,$0D,$7F,$04,$ED,$CC,$70,$E4,$91,$83,$8F,$D4,$F4,$23,$D3,$C0
	.dc.b	$AF,$48,$BA,$7E,$C3,$0F,$E0,$FF,$52,$F9,$AF,$EA,$2D,$BF,$D6,$35
	.dc.b	$FC,$B7,$D4,$FF,$0C,$73,$3F,$D6,$21,$F3,$34,$E7,$12,$7F,$6D,$3A
	.dc.b	$6C,$E5,$EB,$ED,$F3,$23,$C3,$C0,$AE,$8E,$71,$E1,$8F,$17,$67,$8F
	.dc.b	$AE,$57,$82,$32,$F8,$A3,$2F,$B2,$3F,$B8,$FD,$43,$EA,$69,$27,$F8
	.dc.b	$77,$AD,$F9,$79,$D4,$63,$DA,$FE,$F3,$3F,$DF,$F9,$7F,$DE,$35,$AA
	.dc.b	$FD,$F3,$78,$7F,$90,$D6,$E7,$F4,$ED,$53,$BA,$FF,$7D,$CF,$E7,$27
	.dc.b	$F8,$42,$EA,$F0,$9E,$2E,$C9,$3E,$88,$CB,$EA,$8F,$3F,$5F,$A3,$EA
	.dc.b	$64,$FB,$A3,$2F,$C0,$8C,$BE,$08,$FF,$0B,$FD,$4F,$F3,$CF,$82,$FD
	.dc.b	$1C,$FF,$00,$FA,$8F,$F6,$DB,$FE,$BC,$EE,$FE,$06,$A3,$2F,$FA,$E6
	.dc.b	$33,$5C,$DB,$E0,$FF,$AA,$5E,$35,$C2,$FB,$DF,$3F,$89,$D3,$FC,$43
	.dc.b	$0E,$4F,$8E,$32,$EC,$F2,$19,$28,$8E,$2E,$CF,$31,$ED,$F1,$79,$DF
	.dc.b	$F2,$28,$C9,$BB,$F2,$C3,$EA,$3C,$3F,$50,$FA,$9F,$D5,$57,$AA,$31
	.dc.b	$EA,$78,$AF,$86,$17,$0E,$38,$70,$D3,$E5,$E1,$C2,$BC,$37,$C1,$F1
	.dc.b	$C6,$71,$ED,$7C,$67,$F9,$AC,$E6,$FE,$6E,$1D,$B1,$F8,$73,$9A,$DD
	.dc.b	$F1,$7E,$D5,$6D,$EF,$19,$9E,$11,$A3,$9C,$64,$95,$ED,$8C,$B1,$E6
	.dc.b	$E5,$A4,$C3,$D7,$19,$6C,$67,$0F,$8B,$5E,$7C,$7E,$25,$68,$93,$AB
	.dc.b	$E4,$DB,$E4,$7C,$08,$CB,$4C,$8C,$3D,$F7,$DA,$B9,$5F,$6A,$E0,$FE
	.dc.b	$13,$1F,$90,$E6,$39,$63,$E7,$9C,$D7,$22,$4F,$EB,$63,$2F,$48,$CB
	.dc.b	$F3,$CF,$2E,$1F,$48,$F2,$CF,$D7,$8F,$D9,$9E,$0C,$7F,$DE,$52,$F1
	.dc.b	$9F,$19,$8C,$7F,$A0,$5F,$0A,$D4,$FE,$1A,$B7,$3F,$A2,$FA,$8F,$F0
	.dc.b	$77,$36,$6A,$4F,$E5,$2B,$7A,$7E,$EA,$F7,$39,$AD,$CF,$E9,$6B,$7A
	.dc.b	$7F,$0A,$BD,$CE,$6B,$73,$FA,$4A,$DD,$FF,$84,$AF,$73,$9A,$DC,$FE
	.dc.b	$8A,$B6,$AB,$6E,$67,$6F,$DF,$9D,$CE,$A3,$13,$AC,$7E,$2C,$7C,$3E
	.dc.b	$BF,$F3,$66,$3E,$78,$F7,$11,$F8,$51,$AC,$7C,$D3,$96,$CE,$B2,$ED
	.dc.b	$97,$A3,$0E,$F3,$9B,$E0,$7A,$4A,$7B,$70,$EC,$E6,$3C,$84,$C3,$E7
	.dc.b	$4E,$A7,$DA,$72,$AE,$F3,$AC,$3C,$78,$FC,$6C,$BE,$25,$6A,$FB,$5F
	.dc.b	$A8,$E0,$F3,$CB,$D4,$7C,$CC,$7C,$15,$C6,$7C,$39,$79,$5F,$B9,$3A
	.dc.b	$AE,$38,$F8,$63,$0F,$92,$7E,$96,$56,$F9,$47,$C3,$8C,$63,$E5,$8E
	.dc.b	$B1,$87,$D2,$60,$F9,$79,$7B,$B1,$F4,$4F,$84,$FE,$97,$1F,$85,$1F
	.dc.b	$53,$F2,$DC,$7A,$D4,$C7,$AF,$6B,$8C,$5C,$3F,$59,$37,$2F,$D4,$37
	.dc.b	$2F,$D2,$36,$7F,$40,$FF,$AE,$FB,$3F,$D5,$F8,$A3,$18,$FE,$AF,$DC
	.dc.b	$B7,$C2,$3F,$D7,$3F,$24,$F8,$BF,$F4,$F4,$61,$7C,$3F,$09,$D7,$AF
	.dc.b	$67,$35,$B8,$C5,$65,$B1,$CF,$EA,$78,$7D,$4F,$16,$C6,$4F,$04,$65
	.dc.b	$A7,$FB,$07,$EA,$3F,$DD,$CF,$FB,$43,$91,$F5,$3F,$E7,$68,$D7,$C0
	.dc.b	$7B,$A7,$3C,$5F,$03,$66,$1C,$9F,$F5,$A3,$78,$BF,$DC,$9F,$22,$F8
	.dc.b	$4F,$95,$97,$42,$5F,$0C,$BE,$5D,$6F,$8F,$B6,$3C,$EF,$FD,$4B,$C5
	.dc.b	$F3,$7D,$0F,$91,$87,$C9,$6C,$FC,$F6,$E3,$FE,$D6,$AE,$13,$F4,$EB
	.dc.b	$E8,$7B,$52,$3E,$B7,$D4,$DB,$E1,$CA,$AC,$6D,$FE,$FF,$C7,$EF,$5E
	.dc.b	$30,$FF,$1B,$FD,$33,$C7,$FC,$2D,$BE,$A7,$F0,$79,$FF,$B6,$6B,$57
	.dc.b	$FF,$00,$2B,$5C,$FF,$36,$73,$5B,$C3,$E2,$F1,$7F,$A2,$30,$FF,$7D
	.dc.b	$56,$B0,$47,$8A,$B9,$1C,$3D,$3B,$1F,$93,$FA,$9F,$C8,$1F,$91,$F0
	.dc.b	$2C,$64,$99,$7C,$28,$CB,$F1,$23,$F3,$1F,$3F,$F9,$96,$9F,$9F,$2C
	.dc.b	$6D,$FD,$14,$7B,$7D,$1F,$C6,$7D,$4F,$E8,$2B,$E2,$46,$3B,$B5,$7E
	.dc.b	$E8,$DE,$F8,$AD,$CE,$4F,$CD,$5F,$CE,$D7,$E1,$CF,$FB,$B3,$66,$1F
	.dc.b	$B1,$87,$D4,$D3,$1F,$19,$71,$FD,$DC,$E9,$B3,$A6,$9A,$FA,$9E,$9F
	.dc.b	$1F,$D5,$B1,$93,$9C,$65,$A7,$FC,$00,$CB,$CF,$6F,$A8,$7D,$37,$E9
	.dc.b	$3E,$A5,$E3,$7E,$C6,$73,$87,$11,$6F,$05,$7C,$27,$53,$AA,$E2,$B1
	.dc.b	$F0,$C6,$1F,$93,$F5,$2F,$A8,$7D,$4F,$87,$49,$8F,$4C,$7D,$11,$FC
	.dc.b	$92,$F8,$56,$A7,$B5,$7F,$A7,$3E,$A5,$CF,$FC,$2D,$B1,$CF,$C7,$71
	.dc.b	$E6,$BF,$88,$43,$DA,$A7,$F1,$3F,$A6,$7C,$B6,$30,$F2,$46,$1B,$1F
	.dc.b	$C6,$7E,$A5,$F5,$3C,$63,$97,$C1,$CA,$E3,$E6,$5D,$BD,$27,$CE,$C7
	.dc.b	$F8,$DB,$80,$E3,$8F,$ED,$FD,$4F,$6B,$E6,$7C,$77,$15,$C6,$B8,$4F
	.dc.b	$FB,$02,$73,$5F,$84,$9C,$D7,$F2,$2E,$9F,$EA,$36,$FA,$77,$C1,$FF
	.dc.b	$35,$F2,$95,$A9,$CD,$7D,$79,$CB,$67,$5F,$53,$C5,$91,$93,$F1,$27
	.dc.b	$78,$CE,$1F,$06,$1F,$0F,$E4,$9F,$A1,$39,$8C,$6F,$FD,$26,$B6,$7E
	.dc.b	$24,$B3,$F3,$9F,$67,$D0,$EE,$A4,$6D,$F6,$46,$27,$E9,$7D,$4B,$E5
	.dc.b	$3E,$01,$FD,$5D,$C5,$FA,$B4,$B5,$C2,$B4,$7C,$7F,$86,$B5,$F9,$DB
	.dc.b	$DF,$07,$ED,$9B,$1F,$B0,$3E,$94,$F0,$BF,$C2,$7C,$0F,$F1,$DB,$FE
	.dc.b	$27,$5F,$5F,$E9,$FF,$B5,$4E,$63,$96,$49,$CF,$97,$27,$D1,$D5,$71
	.dc.b	$EB,$37,$1E,$B8,$7F,$A4,$AF,$F6,$98,$F5,$BF,$E0,$BE,$27,$E0,$1F
	.dc.b	$CD,$E2,$FA,$87,$C4,$7F,$7D,$AD,$65,$FC,$AA,$B5,$39,$E4,$99,$7F
	.dc.b	$82,$7E,$0B,$FB,$B6,$5D,$A7,$2F,$E9,$71,$E7,$5F,$C5,$B0,$5F,$0A
	.dc.b	$FE,$31,$7E,$1F,$A3,$FE,$D3,$F5,$3C,$27,$D3,$18,$9D,$61,$FC,$47
	.dc.b	$14,$FF,$85,$F4,$76,$EB,$0F,$17,$0F,$B4,$FF,$0D,$3F,$CA,$3E,$0B
	.dc.b	$F9,$4F,$D4,$BE,$A3,$F9,$5C,$FF,$2B,$AE,$19,$F7,$78,$C2,$F8,$D7
	.dc.b	$0C,$FE,$D5,$FF,$25,$AE,$C3,$0E,$CF,$D4,$C3,$FA,$BC,$62,$3C,$38
	.dc.b	$74,$75,$55,$C9,$CF,$0F,$05,$F2,$7F,$DA,$F1,$E5,$34,$72,$29,$CA
	.dc.b	$FC,$18,$7A,$23,$27,$A2,$33,$5D,$63,$FD,$C7,$87,$8E,$B8,$57,$A6
	.dc.b	$78,$5F,$93,$0F,$44,$60,$F1,$4E,$DF,$33,$DA,$37,$18,$9E,$4F,$90
	.dc.b	$C1,$C9,$F7,$DF,$8F,$0F,$34,$FB,$67,$8D,$F7,$AF,$8B,$3F,$28,$AE
	.dc.b	$1F,$AE,$7F,$C0,$0F,$A9,$FE,$3B,$EE,$F0,$6C,$7C,$E6,$BB,$EB,$FF
	.dc.b	$41,$39,$F9,$2F,$ED,$CD,$E5,$E5,$18,$8E,$CD,$7F,$23,$48,$E3,$8F
	.dc.b	$E1,$7E,$A5,$F4,$BF,$96,$8C,$37,$4E,$35,$C1,$B8,$FE,$26,$F8,$E3
	.dc.b	$6F,$83,$FE,$0A,$C7,$C1,$5F,$8F,$6C,$7E,$22,$B8,$B6,$B8,$A9,$DE
	.dc.b	$3F,$9B,$6E,$1F,$AC,$BE,$33,$E8,$AE,$D8,$F9,$DF,$13,$54,$F0,$AE
	.dc.b	$58,$78,$30,$EB,$5C,$31,$F6,$30,$FC,$FC,$69,$F3,$61,$D7,$0E,$91
	.dc.b	$FE,$2A,$6B,$AA,$E4,$7F,$8D,$A3,$E0,$B6,$FF,$83,$BF,$E3,$AB,$ED
	.dc.b	$3A,$DB,$E0,$78,$BA,$6E,$1F,$E2,$99,$D3,$FE,$3A,$FA,$9F,$F6,$3B
	.dc.b	$FE,$2D,$FA,$7F,$F2,$CB,$7D,$7E,$4E,$FB,$AA,$AE,$2F,$F8,$56,$B8
	.dc.b	$B6,$B7,$3A,$CF,$FC,$A2,$BE,$6F,$24,$AF,$17,$87,$5E,$9E,$3F,$83
	.dc.b	$1F,$F0,$8C,$7F,$90,$EF,$93,$6F,$55,$23,$07,$F9,$27,$EA,$3F,$9C
	.dc.b	$B7,$AF,$D7,$18,$8F,$57,$9B,$89,$F0,$7D,$4C,$9E,$04,$3E,$03,$97
	.dc.b	$FC,$6D,$5C,$F4,$FF,$37,$FC,$FF,$F9,$B2,$7F,$CB,$4D,$7F,$C7,$8D
	.dc.b	$8F,$F3,$AE,$9F,$E2,$C9,$DF,$B3,$F0,$7D,$29,$18,$21,$F4,$5C,$BD
	.dc.b	$23,$EB,$FC,$1F,$F9,$62,$3F,$CD,$4D,$7F,$C9,$8D,$BF,$F4,$2F,$3F
	.dc.b	$F9,$A2,$77,$E6,$7E,$02,$1F,$0F,$4E,$B1,$27,$FC,$ED,$3A,$6C,$E5
	.dc.b	$FF,$2D,$74,$7D,$86,$F1,$7E,$32,$31,$87,$FA,$5B,$EA,$5E,$56,$6B
	.dc.b	$F2,$30,$F9,$0A,$67,$0F,$F4,$36,$1C,$BA,$FF,$B6,$D6,$E7,$FD,$14
	.dc.b	$D7,$C2,$7F,$A3,$7E,$A5,$E7,$7E,$79,$B1,$98,$F0,$ED,$97,$F4,$7F
	.dc.b	$51,$FE,$AA,$9F,$F4,$B3,$5F,$F3,$C5,$72,$7F,$CF,$51,$9D,$7F,$D6
	.dc.b	$7F,$53,$FE,$92,$9F,$F4,$F3,$5D,$FB,$BD,$BC,$5F,$D8,$FD,$43,$E1
	.dc.b	$A9,$9A,$FF,$53,$46,$5B,$F5,$3C,$0D,$BA,$FA,$5F,$DD,$FD,$4B,$EA
	.dc.b	$69,$DA,$3F,$7F,$F1,$7F,$E1,$EA,$FF,$59,$35,$FF,$4A,$37,$5F,$F6
	.dc.b	$6F,$C8,$FF,$8C,$27,$7E,$97,$F0,$88,$7C,$F5,$3A,$C4,$9F,$F5,$F4
	.dc.b	$E9,$B3,$97,$FD,$45,$F5,$3F,$2E,$FF,$DB,$6F,$F9,$CE,$3F,$D8,$CD
	.dc.b	$77,$C9,$FC,$62,$38,$73,$FF,$8C,$1D,$57,$2F,$77,$55,$BF,$1E,$61
	.dc.b	$D1,$E5,$F4,$58,$8F,$1C,$5B,$F1,$CF,$C8,$F9,$79,$5F,$21,$F1,$7C
	.dc.b	$9E,$23,$D7,$F5,$3E,$53,$17,$FE,$F6,$E0,$5E,$6B,$DB,$5A,$D7,$FC
	.dc.b	$93,$F5,$0F,$27,$F9,$AA,$E3,$A6,$22,$5F,$FB,$AE,$74,$D9,$CB,$FE
	.dc.b	$CF,$87,$D4,$F4,$24,$71,$F0,$EB,$B8,$BA,$31,$F2,$D7,$F4,$8F,$A9
	.dc.b	$7C,$D5,$26,$1F,$EF,$DF,$A9,$FD,$FB,$FE,$FF,$9D,$B9,$F9,$9F,$EA
	.dc.b	$32,$78,$C7,$FB,$96,$76,$FF,$B9,$E7,$4F,$13,$A5,$72,$F6,$64,$78
	.dc.b	$5E,$BE,$06,$23,$94,$5C,$FD,$90,$17,$FE,$D4,$7D,$5F,$07,$E6,$EC
	.dc.b	$91,$82,$1E,$57,$2F,$9E,$3F,$B6,$FD,$43,$EA,$69,$27,$C1,$87,$89
	.dc.b	$FF,$79,$F5,$65,$46,$7D,$6F,$EE,$6D,$F6,$74,$F1,$C3,$B5,$74,$E2
	.dc.b	$FF,$49,$DF,$06,$DF,$07,$3D,$9F,$E0,$29,$76,$53,$11,$2F,$A1,$F1
	.dc.b	$E7,$6E,$03,$87,$D3,$5D,$9D,$43,$B3,$C1,$F5,$32,$73,$1E,$33,$0F
	.dc.b	$23,$C1,$19,$7C,$1F,$53,$FC,$CF,$E0,$BF,$5B,$1F,$6F,$EA,$7F,$DA
	.dc.b	$AF,$FA,$DF,$BF,$FD,$6B,$18,$8F,$F5,$94,$7F,$AE,$2B,$93,$78,$3F
	.dc.b	$D5,$6F,$0F,$8A,$B5,$D5,$AA,$B8,$B9,$3C,$05,$3B,$97,$0F,$3B,$C6
	.dc.b	$F9,$E1,$E5,$9C,$BF,$CA,$67,$88,$71,$1A,$BE,$3F,$53,$E4,$FA,$87
	.dc.b	$D4,$FE,$CA,$BB,$C6,$3D,$6F,$2E,$3C,$EF,$8B,$89,$D3,$E0,$6D,$79
	.dc.b	$27,$B5,$F5,$D3,$FD,$C3,$5A,$6C,$66,$FE,$A5,$70,$C3,$B4,$E6,$57
	.dc.b	$C5,$D9,$A9,$8F,$8E,$78,$64,$BE,$93,$9A,$DE,$3E,$78,$B8,$F9,$26
	.dc.b	$C7,$63,$81,$27,$E3,$E1,$DA,$FA,$C6,$2B,$8D,$7C,$92,$E3,$E9,$7D
	.dc.b	$77,$C5,$CB,$EF,$8D,$63,$E9,$9C,$D7,$48,$C4,$78,$F1,$F4,$34,$F4
	.dc.b	$61,$E3,$70,$78,$2B,$EE,$D7,$29,$D1,$2B,$9B,$F1,$6B,$8D,$FA,$4F
	.dc.b	$2E,$3F,$02,$B6,$49,$F0,$B5,$F0,$1E,$2B,$F0,$47,$A2,$3F,$97,$1F
	.dc.b	$CB,$EB,$F1,$AE,$0F,$C0,$F8,$AF,$8D,$FF,$77,$CE,$A7,$F1,$B3,$97
	.dc.b	$F2,$8D,$C7,$F5,$7F,$50,$EA,$FE,$7B,$1F,$A0,$9D,$63,$CB,$1F,$CF
	.dc.b	$46,$30,$FE,$85,$87,$1E,$4C,$D4,$9E,$EF,$EA,$27,$51,$82,$61,$DE
	.dc.b	$75,$8F,$8B,$1F,$05,$69,$B5,$E8,$9D,$4F,$C5,$C7,$C8,$FC,$4B,$F1
	.dc.b	$56,$F1,$B8,$72,$25,$7A,$B0,$E6,$F1,$3C,$07,$5B,$F2,$D7,$28,$C1
	.dc.b	$A1,$8F,$BB,$0F,$64,$67,$26,$5E,$6A,$D6,$5E,$2A,$F0,$5F,$03,$D0
	.dc.b	$F1,$AF,$6D,$F2,$C7,$C7,$39,$7C,$8F,$8B,$2F,$5C,$EE,$34,$7F,$24
	.dc.b	$C7,$C2,$FA,$B0,$F0,$63,$E3,$9C,$CF,$92,$3C,$D5,$CE,$BC,$A7,$09
	.dc.b	$E5,$3C,$6F,$C3,$8F,$A0,$98,$FA,$2F,$85,$F5,$38,$9F,$4F,$2F,$5E
	.dc.b	$3D,$F1,$E9,$18,$C3,$9D,$7A,$F1,$EF,$8F,$5B,$DC,$FD,$5C,$BD,$64
	.dc.b	$FA,$9F,$BE,$E3,$D9,$A6,$3D,$9B,$9F,$E6,$62,$DF,$C9,$9B,$AF,$EA
	.dc.b	$9B,$8F,$EE,$30,$46,$2B,$F4,$EF,$F8,$AB,$B3,$FC,$35,$3A,$C7,$FC
	.dc.b	$08,$BD,$DD,$47,$3A,$B6,$F9,$37,$C5,$FF,$A8,$EB,$FC,$1F,$18,$F8
	.dc.b	$7D,$BA,$F5,$ED,$26,$C6,$1B,$5C,$A7,$4D,$FA,$9E,$3F,$53,$C9,$B1
	.dc.b	$93,$84,$65,$A7,$F8,$7F,$EA,$3F,$DD,$2F,$F8,$93,$8D,$F5,$3F,$E7
	.dc.b	$E8,$D7,$C1,$7B,$A7,$3C,$5F,$03,$66,$1C,$9F,$F0,$F1,$8A,$F8,$BF
	.dc.b	$22,$F8,$5F,$99,$3A,$AE,$1C,$9F,$E4,$96,$E5,$FC,$B9,$B2,$BE,$5C
	.dc.b	$5E,$19,$CF,$17,$D0,$F3,$BE,$5E,$7F,$31,$B8,$FC,$96,$E3,$FE,$33
	.dc.b	$AE,$13,$F3,$6B,$E9,$FB,$52,$3E,$87,$D4,$DB,$E1,$C6,$AC,$6B,$FE
	.dc.b	$72,$C7,$F0,$37,$8C,$3F,$A3,$FD,$37,$F9,$1D,$FF,$28,$D3,$EA,$7F
	.dc.b	$0D,$97,$C8,$AD,$3F,$E4,$79,$D7,$3F,$CF,$9C,$C7,$F9,$BD,$B7,$FE
	.dc.b	$81,$E0,$F9,$DE,$A6,$FC,$99,$3F,$E5,$DA,$DE,$3F,$2F,$8F,$FD,$31
	.dc.b	$39,$AF,$C9,$FD,$4C,$8F,$C8,$78,$36,$FE,$0C,$DC,$FE,$07,$0F,$F2
	.dc.b	$AB,$FC,$B4,$FE,$71,$FC,$EF,$D0,$57,$F3,$4D,$BF,$96,$DB,$E6,$FB
	.dc.b	$FC,$FF,$C8,$7D,$4F,$E8,$2B,$E3,$46,$3B,$B5,$87,$B6,$37,$C3,$8A
	.dc.b	$DC,$E4,$FC,$F5,$FD,$0C,$7E,$55,$FE,$D2,$9F,$53,$4C,$7C,$6D,$C3
	.dc.b	$F7,$73,$A6,$CE,$9A,$6B,$EA,$7A,$7C,$DF,$56,$C6,$4E,$71,$96,$9F
	.dc.b	$E9,$3F,$A8,$7C,$57,$EA,$BE,$A7,$F8,$94,$E6,$31,$E3,$7E,$CE,$B7
	.dc.b	$39,$3F,$D3,$FF,$50,$FA,$97,$D4,$EF,$09,$5B,$D7,$FD,$2D,$7B,$AC
	.dc.b	$C7,$EF,$3E,$A5,$CB,$FC,$05,$B7,$FE,$A9,$F1,$DC,$39,$AF,$E0,$D0
	.dc.b	$F6,$A9,$FC,$27,$E9,$9F,$29,$8C,$3F,$D6,$51,$86,$C7,$F0,$FF,$A9
	.dc.b	$7D,$4F,$18,$E5,$F0,$52,$B8,$F9,$97,$6F,$E2,$78,$FF,$10,$EA,$FF
	.dc.b	$6B,$36,$BF,$C7,$7F,$53,$D1,$E2,$7F,$B0,$DC,$57,$2B,$FE,$41,$18
	.dc.b	$BE,$71,$FC,$8F,$C9,$74,$FF,$44,$C1,$F4,$FF,$EC,$C7,$FD,$9D,$3A
	.dc.b	$7F,$DA,$66,$BE,$07,$95,$48,$CD,$3E,$A7,$19,$FF,$2E,$C3,$93,$6F
	.dc.b	$C1,$5B,$BE,$F1,$FC,$DD,$7C,$47,$97,$0F,$9D,$7F,$CE,$27,$27,$92
	.dc.b	$33,$9F,$FB,$F6,$33,$3B,$AE,$FF,$43,$D5,$93,$FE,$EB,$37,$45,$57
	.dc.b	$15,$F0,$8F,$E9,$3F,$52,$F9,$2F,$59,$FD,$3D,$C5,$FC,$48,$C1,$1F
	.dc.b	$8F,$F0,$96,$B9,$CD,$C7,$C5,$1E,$19,$C5,$7F,$9F,$3E,$0B,$F8,$75
	.dc.b	$6A,$3E,$CF,$D3,$FF,$5F,$9C,$C7,$A8,$C9,$EC,$9E,$B6,$9D,$BF,$01
	.dc.b	$B3,$F8,$4E,$9E,$9E,$4C,$8E,$47,$8D,$FB,$71,$9A,$FE,$DB,$E8,$FE
	.dc.b	$03,$0F,$5C,$BE,$A1,$F0,$DF,$DD,$6B,$58,$FD,$0A,$D4,$E7,$8E,$4F
	.dc.b	$F8,$07,$E0,$BF,$B7,$E5,$FD,$BE,$72,$F8,$1F,$F0,$14,$6B,$0E,$B3
	.dc.b	$FD,$47,$E8,$FF,$C4,$79,$BE,$A7,$33,$FD,$E6,$73,$18,$A4,$79,$CF
	.dc.b	$F0,$76,$8E,$CC,$67,$F1,$30,$FE,$BD,$C1,$ED,$3F,$C2,$4F,$E3,$3E
	.dc.b	$0B,$F1,$BF,$52,$FA,$9F,$F5,$BE,$6B,$E1,$87,$3C,$3F,$C1,$91,$C9
	.dc.b	$61,$CA,$F8,$CE,$63,$13,$E2,$AA,$FF,$B1,$22,$DF,$79,$CC,$F9,$46
	.dc.b	$1E,$57,$91,$8C,$3F,$D9,$4F,$88,$C1,$A8,$E0,$F7,$32,$7A,$A7,$BB
	.dc.b	$C6,$37,$5E,$38,$D9,$E3,$8C,$E1,$FA,$47,$AC,$66,$FE,$0D,$F4,$C3
	.dc.b	$DB,$7B,$9F,$3C,$62,$75,$5E,$4C,$3C,$18,$74,$8F,$25,$F4,$7C,$66
	.dc.b	$0F,$73,$D1,$E5,$87,$77,$FD,$DD,$87,$C6,$7E,$11,$5C,$3F,$3C,$FF
	.dc.b	$7A,$7D,$4F,$F2,$0F,$57,$83,$63,$E7,$B5,$DF,$5F,$F9,$E9,$CF,$C9
	.dc.b	$38,$78,$B6,$33,$18,$23,$F5,$3E,$A5,$F4,$BF,$C8,$23,$0D,$D7,$ED
	.dc.b	$E2,$C3,$A4,$61,$BA,$F1,$C3,$9E,$1F,$90,$6B,$F8,$2B,$E5,$8F,$E5
	.dc.b	$1B,$1E,$03,$86,$3E,$1B,$F2,$4E,$B1,$F0,$5E,$AE,$1D,$71,$F1,$BF
	.dc.b	$CE,$AB,$85,$F9,$70,$E9,$3F,$CF,$B1,$F2,$E3,$D2,$39,$63,$D2,$3F
	.dc.b	$37,$3A,$BF,$1E,$3D,$AD,$8F,$67,$A6,$3D,$95,$F1,$7C,$95,$C6,$3E
	.dc.b	$2E,$3F,$CD,$1B,$BF,$9E,$BF,$81,$61,$D5,$E7,$87,$D7,$BE,$1E,$15
	.dc.b	$BF,$A4,$DA,$F1,$9F,$C2,$FE,$A7,$FD,$74,$FF,$0D,$FA,$77,$AF,$B8
	.dc.b	$46,$1E,$1D,$99,$AE,$31,$FE,$AA,$9D,$36,$B5,$5B,$F9,$89,$3E,$17
	.dc.b	$B5,$7F,$6C,$8C,$1E,$3F,$1B,$27,$07,$FD,$6F,$82,$B3,$1B,$38,$9F
	.dc.b	$EB,$CF,$A8,$FE,$6E,$FF,$29,$6B,$9E,$AF,$0F,$9B,$EF,$9C,$ED,$FD
	.dc.b	$93,$27,$D4,$F9,$8D,$5C,$C3,$E4,$46,$5E,$D1,$F2,$7E,$7B,$FD,$E5
	.dc.b	$1F,$CE,$5A,$FF,$B4,$1B,$7D,$FC,$0E,$1E,$C7,$C4,$F4,$A5,$F1,$8C
	.dc.b	$BF,$5E,$32,$F3,$8F,$B1,$F0,$3F,$E5,$68,$FE,$90,$D7,$FD,$BE,$DB
	.dc.b	$FF,$00,$F2,$FF,$99,$A7,$7E,$87,$E0,$61,$F0,$F4,$EB,$12,$7F,$AA
	.dc.b	$4E,$9B,$39,$7F,$DD,$BC,$DF,$5D,$FE,$B4,$D7,$3C,$5F,$8F,$FA,$9F
	.dc.b	$F0,$94,$6A,$31,$F2,$12,$30,$43,$FD,$F3,$28,$D4,$74,$3F,$B4,$7D
	.dc.b	$4B,$CC,$FC,$E3,$63,$31,$E1,$CF,$2F,$E8,$7E,$A7,$FD,$37,$3F,$DC
	.dc.b	$5A,$FF,$2D,$EC,$FE,$4D,$7C,$E3,$07,$F9,$22,$77,$B7,$F8,$BB,$EA
	.dc.b	$7F,$D1,$F3,$FD,$ED,$AF,$01,$3C,$2F,$F9,$93,$FE,$62,$8F,$F0,$17
	.dc.b	$D4,$BE,$17,$F6,$CD,$8F,$37,$8B,$89,$CF,$D4,$C9,$E3,$1C,$3E,$93
	.dc.b	$F7,$DF,$52,$FA,$9A,$66,$BF,$C2,$1F,$17,$FE,$1D,$6F,$9F,$FC,$16
	.dc.b	$31,$1E,$9F,$91,$7A,$79,$AF,$E1,$50,$FA,$9A,$66,$B9,$9F,$E1,$D8
	.dc.b	$D3,$63,$4F,$C5,$FA,$9F,$9D,$C7,$FE,$58,$7F,$CE,$11,$FE,$25,$6B
	.dc.b	$BE,$5F,$E3,$51,$C3,$8F,$FA,$E3,$AA,$E5,$EC,$EA,$B7,$E3,$CC,$3A
	.dc.b	$3C,$BE,$8B,$11,$E3,$8B,$7E,$39,$F8,$9F,$2F,$2B,$E4,$BE,$2F,$91
	.dc.b	$C4,$7A,$FE,$A7,$CA,$62,$FF,$CE,$7C,$2B,$CD,$7B,$6B,$5A,$FF,$75
	.dc.b	$FA,$97,$91,$FC,$DD,$73,$53,$11,$2F,$FC,$95,$3A,$6C,$E5,$FC,$94
	.dc.b	$3E,$A7,$A1,$23,$8F,$85,$5D,$C5,$D3,$FD,$17,$5F,$BD,$7F,$D1,$BF
	.dc.b	$52,$F9,$AF,$E9,$8D,$BF,$F2,$E3,$5F,$D1,$7D,$4F,$F0,$27,$F4,$91
	.dc.b	$9A,$FF,$4B,$CB,$E6,$7F,$AB,$36,$B9,$DF,$63,$33,$23,$27,$F9,$AE
	.dc.b	$7F,$C3,$4D,$F6,$39,$9D,$3C,$0B,$19,$21,$E6,$7C,$3D,$9E,$2E,$C9
	.dc.b	$3D,$63,$2F,$78,$CB,$EE,$8F,$ED,$DF,$50,$FA,$9A,$49,$FF,$40,$7A
	.dc.b	$DF,$99,$7F,$DA,$73,$B7,$3E,$C7,$F7,$79,$FE,$E3,$B7,$FA,$F2,$B5
	.dc.b	$5F,$C3,$5B,$1F,$EC,$09,$DF,$2F,$F8,$FD,$B5,$DB,$0E,$9D,$7D,$7A
	.dc.b	$E4,$F9,$23,$2F,$9A,$32,$FF,$1B,$A4,$F9,$7A,$FC,$DF,$53,$2B,$DB
	.dc.b	$19,$7D,$F1,$97,$94,$7F,$84,$FE,$A7,$F9,$77,$C0,$7E,$96,$7F,$DE
	.dc.b	$BF,$4F,$FE,$CA,$7F,$D5,$9D,$DF,$E5,$98,$C4,$7F,$A9,$A3,$FC,$B9
	.dc.b	$5C,$9B,$E0,$7F,$A8,$DE,$1F,$0D,$6B,$AB,$55,$71,$72,$78,$0A,$77
	.dc.b	$2E,$1E,$77,$8D,$F3,$C3,$CB,$39,$7F,$90,$CF,$10,$E2,$35,$7C,$7E
	.dc.b	$A3,$C9,$F5,$0F,$A9,$FE,$D3,$5D,$E3,$1F,$17,$E5,$C7,$9D,$F1,$71
	.dc.b	$3A,$7C,$0D,$AF,$24,$F6,$BE,$BA,$7F,$A1,$6B,$4D,$8C,$DF,$D4,$AE
	.dc.b	$18,$76,$9C,$CA,$F8,$BB,$35,$31,$F1,$CF,$0C,$97,$D2,$73,$5B,$C7
	.dc.b	$CF,$17,$1F,$24,$D8,$EC,$70,$24,$FC,$7C,$3B,$5F,$58,$C5,$71,$AF
	.dc.b	$92,$5C,$7D,$2F,$AE,$F8,$B9,$7D,$F1,$AC,$7D,$33,$9A,$E9,$18,$8F
	.dc.b	$1E,$3E,$86,$9E,$8C,$3C,$6E,$0F,$05,$7D,$DA,$E5,$3A,$25,$73,$7E
	.dc.b	$2D,$71,$BF,$49,$E5,$C7,$E0,$56,$C9,$3E,$16,$BE,$03,$C5,$7E,$08
	.dc.b	$CD,$74,$C7,$E1,$D7,$CF,$3F,$D9,$87,$12,$C3,$C5,$7C,$5F,$FB,$9E
	.dc.b	$3F,$29,$3A,$9C,$BF,$96,$6E,$5F,$AD,$FA,$8F,$F0,$8F,$37,$E7,$27
	.dc.b	$39,$7E,$5D,$B6,$3F,$24,$E1,$57,$27,$B6,$3C,$C8,$FE,$D6,$35,$87
	.dc.b	$EB,$EB,$96,$3E,$BB,$DC,$F9,$7A,$99,$63,$2F,$25,$5C,$BC,$99,$74
	.dc.b	$9C,$C7,$88,$98,$73,$BE,$11,$EA,$BB,$7C,$63,$DD,$8F,$8D,$E3,$87
	.dc.b	$4C,$39,$46,$0D,$E5,$EA,$3E,$25,$F3,$8E,$C6,$CB,$3C,$8F,$85,$87
	.dc.b	$93,$0F,$04,$67,$0F,$01,$2B,$BD,$7F,$15,$BE,$D7,$C2,$B5,$5D,$9F
	.dc.b	$B0,$FD,$6C,$BE,$5B,$E3,$9E,$0F,$43,$EB,$46,$B0,$F2,$C6,$F1,$F6
	.dc.b	$CE,$6F,$BD,$F2,$AF,$01,$F5,$27,$4F,$5B,$E6,$4A,$F2,$E1,$EA,$C3
	.dc.b	$E5,$63,$F5,$30,$E3,$87,$42,$63,$DE,$F8,$57,$C9,$BF,$2E,$5C,$49
	.dc.b	$7F,$46,$BE,$A1,$3E,$A7,$AB,$8F,$5A,$98,$F6,$2D,$FC,$F8,$B7,$F9
	.dc.b	$B9,$BA,$FE,$AD,$BA,$FB,$1B,$3F,$96,$7E,$8F,$5F,$D2,$8F,$D4,$C6
	.dc.b	$30,$F8,$7E,$E5,$9E,$11,$FB,$1A,$FB,$3F,$24,$F8,$AF,$F4,$E4,$E4
	.dc.b	$FD,$B4,$67,$E1,$B9,$3B,$F5,$6D,$26,$C6,$1B,$5C,$F0,$5F,$38,$C3
	.dc.b	$E1,$FA,$9E,$7F,$53,$D1,$B3,$AC,$51,$8B,$FF,$5F,$DF,$1A,$D5,$FB
	.dc.b	$E5,$F5,$1F,$EE,$A9,$F1,$72,$3E,$A7,$FC,$FD,$1E,$5A,$DF,$C0,$FC
	.dc.b	$7A,$D7,$0F,$3C,$39,$57,$FB,$1D,$BB,$FE,$8B,$E4,$1F,$08,$E3,$E1
	.dc.b	$94,$CF,$B6,$1C,$1F,$E4,$57,$C2,$B5,$8F,$B3,$83,$FC,$F1,$5F,$07
	.dc.b	$0F,$9F,$E7,$FD,$06,$C6,$4F,$8D,$19,$7E,$47,$B5,$F5,$8F,$AF,$F5
	.dc.b	$36,$32,$4C,$3C,$11,$97,$FD,$C9,$F5,$0F,$1B,$F1,$5B,$BE,$A7,$F0
	.dc.b	$F5,$A9,$CC,$62,$7F,$29,$5C,$70,$F9,$CD,$BF,$F7,$BE,$DF,$23,$D3
	.dc.b	$FB,$13,$9B,$E3,$8E,$0F,$CD,$56,$F0,$E7,$1F,$92,$FA,$99,$1F,$87
	.dc.b	$F0,$2E,$5F,$2E,$6E,$1F,$A0,$E0,$FE,$C1,$7F,$8D,$9F,$CD,$BF,$9C
	.dc.b	$F9,$FA,$FE,$45,$B7,$F5,$5B,$7C,$DF,$4F,$A3,$F8,$9F,$A9,$FD,$25
	.dc.b	$FC,$08,$C7,$6E,$A3,$07,$E5,$27,$E4,$3F,$CB,$B6,$56,$F1,$FA,$6F
	.dc.b	$5E,$EC,$65,$FB,$A9,$7D,$4F,$ED,$9B,$87,$EB,$97,$C6,$66,$64,$E5
	.dc.b	$E1,$18,$9F,$D6,$8F,$A9,$E9,$F1,$DD,$5B,$7D,$5B,$3C,$1B,$87,$F1
	.dc.b	$5F,$A8,$7D,$3F,$DA,$FA,$97,$8D,$FB,$79,$CE,$3F,$C1,$F9,$D5,$A8
	.dc.b	$FD,$B4,$EB,$26,$3D,$71,$FE,$3D,$F5,$0F,$A9,$7D,$4F,$59,$98,$F2
	.dc.b	$77,$3A,$8C,$1F,$C0,$FE,$A5,$CF,$FC,$2D,$B1,$FC,$1F,$C6,$F9,$6E
	.dc.b	$E7,$F8,$84,$7F,$AC,$23,$5E,$CF,$F1,$3F,$A7,$7C,$A6,$A3,$0C,$9F
	.dc.b	$E2,$5F,$52,$FA,$9C,$70,$7B,$A7,$FB,$8E,$0F,$82,$E8,$75,$F2,$6C
	.dc.b	$FF,$1B,$9F,$E2,$4D,$8F,$E4,$9F,$53,$D6,$30,$7F,$21,$71,$7D,$CF
	.dc.b	$F1,$B5,$72,$70,$23,$7C,$7F,$D4,$67,$4F,$F5,$3F,$41,$D5,$FD,$47
	.dc.b	$07,$D4,$3E,$07,$F9,$B5,$6A,$7F,$A3,$1A,$8C,$D4,$8C,$D3,$EA,$71
	.dc.b	$9F,$93,$0E,$4D,$BF,$05,$6E,$FF,$9F,$56,$B1,$FE,$D1,$0F,$86,$F3
	.dc.b	$63,$F5,$A3,$FA,$21,$93,$CA,$E7,$1F,$87,$19,$6C,$F8,$3E,$A7,$FA
	.dc.b	$94,$E6,$FB,$CE,$0D,$47,$67,$FA,$C7,$D4,$BE,$4B,$D2,$60,$B3,$FD
	.dc.b	$5E,$34,$D8,$F7,$FC,$26,$E3,$04,$C7,$EE,$3F,$4A,$35,$1F,$73,$E0
	.dc.b	$FF,$8B,$BA,$D5,$F4,$FF,$DA,$E7,$39,$FF,$90,$30,$FB,$73,$9F,$2E
	.dc.b	$47,$D6,$EA,$B8,$F5,$9B,$7C,$71,$FF,$47,$DF,$F7,$1C,$7A,$BF,$E4
	.dc.b	$EF,$0F,$EB,$3F,$E8,$7C,$5F,$50,$F8,$6F,$EF,$D5,$AC,$FF,$D3,$33
	.dc.b	$9B,$FE,$F9,$39,$93,$BB,$7C,$1F,$F7,$39,$FE,$E7,$CB,$D0,$97,$FE
	.dc.b	$08,$9D,$4F,$29,$F0,$FD,$07,$F6,$4F,$A9,$ED,$8A,$31,$3A,$C3,$C7
	.dc.b	$1F,$E0,$DC,$BF,$75,$18,$3C,$FC,$1E,$C3,$FC,$35,$8F,$FA,$9F,$E0
	.dc.b	$7F,$D5,$1F,$52,$FA,$9F,$8D,$5B,$BE,$27,$F8,$6E,$F8,$0A,$E0,$FF
	.dc.b	$84,$27,$AA,$BE,$B7,$D2,$EC,$FF,$B2,$26,$BF,$2A,$BF,$D0,$37,$D2
	.dc.b	$3F,$D0,$B3,$AC,$FF,$13,$7C,$27,$C7,$5B,$9C,$C7,$97,$0F,$F6,$AB
	.dc.b	$E3,$30,$79,$EB,$93,$9A,$E5,$87,$8A,$B8,$E1,$E1,$8C,$57,$A4,$E6
	.dc.b	$7A,$6F,$9C,$ED,$B1,$AC,$3C,$B8,$7B,$63,$D3,$87,$C5,$79,$98,$3C
	.dc.b	$C6,$0F,$33,$B8,$E5,$18,$36,$E2,$FE,$0C,$7C,$A2,$9F,$0C,$CB,$AF
	.dc.b	$A8,$FE,$31,$EB,$5E,$9F,$F3,$D5,$7C,$A9,$CB,$63,$2F,$FA,$E8,$43
	.dc.b	$E4,$7F,$B5,$C7,$16,$BB,$8D,$73,$78,$67,$33,$FE,$C4,$9C,$BF,$3B
	.dc.b	$EA,$5F,$4B,$FE,$C8,$CF,$F0,$D1,$89,$FC,$5D,$70,$9F,$C5,$37,$4F
	.dc.b	$F6,$7D,$6E,$BA,$5F,$39,$F0,$61,$E8,$C3,$8B,$EA,$FC,$27,$BD,$7F
	.dc.b	$B6,$1B,$60,$F0,$CB,$0D,$8E,$E7,$E8,$31,$F1,$61,$92,$63,$E6,$C3
	.dc.b	$C9,$5D,$6B,$73,$E6,$C3,$DF,$8F,$9E,$38,$57,$5B,$E2,$FD,$49,$D3
	.dc.b	$E7,$CB,$E2,$3D,$67,$78,$76,$8D,$57,$05,$1F,$27,$0F,$F6,$33,$77
	.dc.b	$F0,$CF,$F7,$CC,$3A,$BC,$30,$FC,$8D,$7D,$18,$FC,$43,$5E,$27,$E2
	.dc.b	$FE,$AB,$FD,$84,$FE,$69,$AF,$EF,$1A,$EF,$AF,$F3,$B1,$AC,$7F,$DB
	.dc.b	$5F,$53,$FE,$60,$AF,$F5,$A3,$5F,$F0,$8C,$EF,$2F,$B5,$08,$7C,$DE
	.dc.b	$2F,$FD,$C9,$18,$9F,$F5,$F3,$5F,$F0,$DD,$BC,$1E,$18,$74,$AD,$57
	.dc.b	$B7,$0F,$39,$1F,$73,$C1,$FC,$65,$FF,$87,$5F,$4F,$89,$97,$CD,$37
	.dc.b	$EA,$BF,$A3,$47,$FB,$45,$AF,$F8,$B9,$B9,$7C,$4F,$37,$87,$57,$F0
	.dc.b	$49,$DD,$7E,$ED,$AF,$C4,$6B,$F9,$F7,$9E,$7F,$23,$27,$D5,$79,$5B
	.dc.b	$5E,$3B,$F4,$7C,$FF,$F9,$DA,$3F,$DC,$CD,$7F,$C8,$6D,$CB,$EC,$F6
	.dc.b	$F0,$F6,$3F,$8C,$CE,$EB,$F8,$83,$5F,$B2,$D7,$F6,$8F,$3C,$BC,$DE
	.dc.b	$D7,$94,$E9,$3F,$EF,$7F,$8A,$FF,$35,$B7,$D2,$F9,$D1,$88,$FF,$2B
	.dc.b	$F3,$7F,$9E,$9D,$1F,$CB,$A3,$4F,$F2,$36,$BF,$8C,$6B,$FC,$1D,$E9
	.dc.b	$9B,$0F,$C8,$D3,$E2,$7E,$F3,$6F,$B9,$FE,$68,$F4,$3F,$21,$7C,$63
	.dc.b	$F8,$94,$62,$FF,$36,$D8,$FE,$2D,$F5,$4F,$27,$F1,$53,$F9,$49,$DF
	.dc.b	$2F,$F4,$68,$C4,$74,$8C,$1F,$91,$A7,$C8,$53,$9C,$7E,$57,$EA,$9E
	.dc.b	$6D,$25,$FF,$27,$CF,$D0,$B0,$F8,$79,$FE,$CA,$F8,$CF,$F5,$A6,$BF
	.dc.b	$DF,$1A,$FF,$51,$6B,$FB,$AF,$A9,$FF,$62,$BF,$D8,$9A,$FF,$81,$1A
	.dc.b	$FF,$59,$6B,$C7,$C4,$FE,$CF,$1F,$E7,$B8,$E5,$18,$3F,$B5,$FD,$4B
	.dc.b	$E0,$E9,$27,$FB,$8C,$65,$BF,$47,$C0,$8E,$0F,$EE,$FF,$13,$FE,$B3
	.dc.b	$7E,$36,$5F,$C5,$23,$11,$FE,$A8,$6E,$9F,$2B,$E3,$BF,$CA,$33,$C0
	.dc.b	$FE,$A9,$3C,$1F,$EE,$CD,$7A,$7B,$5F,$E0,$48,$C4,$70,$8F,$E1,$D2
	.dc.b	$FA,$6A,$7F,$11,$AF,$0E,$1F,$CA,$3E,$9B,$EA,$E3,$F7,$1F,$F4,$86
	.dc.b	$3E,$30,$E1,$FF,$60,$BC,$27,$97,$0F,$F1,$D8,$D5,$73,$F5,$F1,$39
	.dc.b	$F0,$E6,$1D,$6D,$1C,$7E,$7F,$33,$DA,$2E,$1E,$59,$F5,$7C,$A4,$AF
	.dc.b	$36,$68,$F1,$7C,$7E,$67,$FC,$97,$39,$7E,$BC,$FC,$92,$3E,$3F,$A6
	.dc.b	$F5,$1C,$70,$FC,$FF,$D2,$BC,$2F,$E7,$4B,$8E,$98,$89,$7F,$26,$74
	.dc.b	$D9,$CB,$FE,$EA,$87,$D3,$74,$24,$71,$EB,$AE,$E2,$E1,$E4,$9F,$E9
	.dc.b	$9F,$4C,$F9,$7A,$4B,$FB,$5F,$4D,$FC,$25,$FB,$93,$B7,$FC,$9E,$D7
	.dc.b	$FC,$E2,$D7,$FC,$8C,$D7,$FB,$7B,$BB,$FE,$07,$2F,$96,$FE,$BC,$DA
	.dc.b	$EF,$78,$F5,$24,$79,$1D,$F6,$E2,$18,$74,$9C,$3D,$A7,$AF,$BD,$DB
	.dc.b	$B2,$5F,$BA,$32,$FC,$08,$CB,$D3,$AB,$FC,$0F,$97,$EC,$23,$07,$F9
	.dc.b	$0F,$E9,$9F,$4D,$FD,$DD,$B1,$FA,$36,$BC,$07,$A5,$FA,$37,$94,$FF
	.dc.b	$9E,$27,$71,$FE,$6F,$6B,$FE,$18,$6B,$C0,$FD,$84,$FF,$87,$F3,$FE
	.dc.b	$5E,$DC,$38,$5F,$F8,$2E,$3F,$53,$5C,$3A,$70,$67,$AA,$46,$48,$7A
	.dc.b	$67,$8B,$7A,$BA,$7D,$34,$8F,$1C,$77,$30,$F1,$8C,$3F,$C1,$FE,$9B
	.dc.b	$F9,$DB,$FE,$6D,$6A,$69,$3B,$BF,$D8,$FD,$9F,$E3,$AF,$A5,$FF,$7E
	.dc.b	$57,$FB,$7E,$31,$F0,$BA,$8E,$95,$55,$70,$7F,$DB,$55,$C1,$E4,$7F
	.dc.b	$9F,$E3,$5F,$01,$FC,$0E,$3C,$95,$CE,$79,$6F,$FE,$A5,$6C,$F3,$3F
	.dc.b	$B0,$9E,$0C,$3F,$CA,$33,$BF,$A6,$B3,$EB,$6B,$E1,$6B,$E8,$6D,$1F
	.dc.b	$67,$D2,$BE,$9B,$F7,$19,$79,$E3,$1E,$7F,$B3,$0E,$45,$8C,$BF,$08
	.dc.b	$FF,$4F,$38,$9E,$55,$D3,$4F,$35,$71,$8C,$3C,$1A,$AB,$94,$F0,$8E
	.dc.b	$C6,$67,$73,$A8,$E5,$7E,$7C,$7D,$D1,$DE,$23,$CA,$78,$3E,$C3,$15
	.dc.b	$D0,$93,$CF,$1F,$8F,$18,$9D,$CE,$5E,$B8,$FC,$08,$B8,$FA,$63,$13
	.dc.b	$C0,$8F,$E1,$71,$F3,$C6,$27,$FD,$1F,$8F,$98,$F3,$CE,$AB,$99,$27
	.dc.b	$78,$7C,$A9,$D4,$63,$0F,$1E,$1F,$18,$F9,$73,$A7,$17,$CD,$B5,$E9
	.dc.b	$39,$3E,$53,$27,$E3,$CE,$A7,$D5,$AB,$5D,$63,$07,$12,$D7,$C1,$C7
	.dc.b	$E6,$D7,$2C,$7D,$F1,$8A,$E1,$18,$75,$8F,$D3,$BE,$51,$F7,$8F,$F5
	.dc.b	$B9,$E0,$2C,$BC,$47,$C4,$6A,$7E,$21,$A8,$CB,$99,$E1,$3B,$8F,$CB
	.dc.b	$4E,$9E,$11,$99,$DC,$7E,$3E,$79,$D7,$F0,$3F,$A5,$75,$FF,$9D,$63
	.dc.b	$F6,$13,$AD,$7F,$02,$DB,$FD,$6B,$55,$F2,$DB,$F6,$93,$9C,$7C,$38
	.dc.b	$74,$CB,$F7,$57,$BC,$3C,$91,$C0,$8F,$EB,$A3,$53,$C6,$36,$47,$D5
	.dc.b	$97,$A7,$2F,$3D,$F2,$9E,$06,$B2,$C4,$EA,$F9,$3F,$EE,$CC,$3B,$B9
	.dc.b	$AE,$2D,$35,$87,$7A,$F8,$78,$74,$69,$E9,$8D,$3F,$5A,$34,$D3,$58
	.dc.b	$63,$1F,$8F,$7D,$E4,$7C,$EB,$E7,$3A,$8E,$F8,$79,$9D,$E3,$EA,$73
	.dc.b	$7C,$DF,$04,$FC,$37,$E0,$9A,$21,$DF,$0F,$26,$5F,$6C,$FC,$05,$F5
	.dc.b	$9E,$77,$D7,$0F,$24,$71,$3A,$9A,$3A,$CF,$18,$D9,$2B,$AE,$3E,$E6
	.dc.b	$E1,$F2,$2B,$84,$62,$FF,$AA,$C7,$E2,$E7,$F5,$ED,$8F,$AF,$1F,$28
	.dc.b	$CF,$A1,$FE,$D3,$69,$98,$FC,$14,$65,$B1,$97,$E6,$C7,$DC,$9F,$F6
	.dc.b	$6C,$FE,$73,$D1,$C4,$E5,$B1,$93,$F4,$31,$97,$F7,$52,$9F,$ED,$F5
	.dc.b	$F6,$9A,$FE,$B9,$CF,$C6,$BE,$1B,$FD,$79,$7F,$EA,$B8,$C7,$C0,$FF
	.dc.b	$BB,$FB,$7F,$DE,$11,$BF,$47,$15,$F8,$C8,$DB,$C6,$F9,$E3,$FE,$46
	.dc.b	$9C,$B6,$33,$7D,$16,$1D,$25,$5E,$5F,$A5,$C3,$E7,$FA,$5B,$69,$B1
	.dc.b	$86,$E3,$E3,$6D,$FF,$94,$7E,$91,$EC,$F7,$E5,$7D,$2F,$FA,$8F,$0F
	.dc.b	$F5,$D5,$69,$FF,$67,$4E,$B9,$7C,$D8,$2B,$4F,$F3,$16,$BF,$CE,$1A
	.dc.b	$FE,$C8,$E1,$87,$D4,$8E,$3B,$30,$D4,$61,$FD,$CF,$C5,$BD,$CF,$BD
	.dc.b	$1E,$6B,$F8,$7E,$5F,$C4,$6C,$7D,$36,$E3,$F4,$5B,$8F,$D1,$7E,$5F
	.dc.b	$AF,$2B,$ED,$7D,$2D,$BF,$B1,$36,$FE,$C4,$DA,$F6,$7D,$13,$C5,$FB
	.dc.b	$54,$FA,$5F,$BF,$5A,$9C,$C6,$31,$FE,$2B,$7C,$6B,$FA,$A3,$63,$FA
	.dc.b	$E3,$5F,$EC,$6D,$7A,$47,$FB,$C6,$38,$5F,$CB,$96,$BF,$27,$D0,$FB
	.dc.b	$35,$AB,$FC,$D7,$2B,$E9,$7F,$1A,$7E,$3B,$B6,$CF,$E7,$62,$E3,$F9
	.dc.b	$B9,$B9,$7E,$6E,$BF,$2B,$F3,$5F,$96,$69,$FA,$52,$D7,$E8,$DB,$3E
	.dc.b	$78,$62,$FA,$4F,$CB,$7D,$2F,$E9,$AF,$DD,$18,$EC,$F5,$BF,$BE,$AE
	.dc.b	$13,$F2,$A3,$04,$78,$9F,$9C,$AF,$A5,$7E,$09,$FD,$7A,$FA,$5A,$63
	.dc.b	$E1,$AD,$FF,$07,$9D,$36,$74,$D3,$5F,$4B,$D3,$E2,$7A,$B6,$32,$73
	.dc.b	$8C,$B4,$FE,$B3,$A7,$C5,$FA,$47,$D1,$7F,$07,$E4,$7D,$2B,$E3,$3F
	.dc.b	$6B,$5A,$C3,$89,$FA,$A9,$D3,$8D,$3F,$B2,$CE,$DF,$F1,$53,$5F,$F1
	.dc.b	$8B,$5F,$DD,$35,$E3,$87,$80,$FE,$35,$F4,$AF,$A4,$7D,$2E,$70,$F0
	.dc.b	$5B,$E9,$5C,$DF,$CD,$9B,$3F,$E3,$56,$CF,$F8,$F5,$AF,$F9,$13,$C4
	.dc.b	$7B,$3F,$CF,$9B,$5F,$E3,$D6,$CF,$F9,$15,$AF,$F9,$33,$E8,$9F,$23
	.dc.b	$FD,$19,$B5,$FE,$45,$6C,$FF,$93,$5A,$F1,$F8,$0E,$04,$9E,$0F,$3F
	.dc.b	$3A,$47,$36,$CF,$37,$F9,$5D,$7F,$1C,$87,$27,$E8,$FE,$8D,$F4,$9A
	.dc.b	$BF,$E6,$3E,$7B,$A3,$FA,$46,$4F,$A2,$7B,$FA,$9C,$C6,$27,$E8,$D7
	.dc.b	$17,$FC,$C4,$DC,$FF,$CD,$4D,$7F,$CE,$2D,$7F,$9D,$3B,$CF,$F7,$0D
	.dc.b	$9F,$F0,$35,$3E,$93,$C5,$1F,$80,$1F,$0B,$E8,$8F,$9B,$7F,$C0,$67
	.dc.b	$27,$E5,$4E,$5C,$1F,$D4,$1C,$E3,$FE,$7F,$6B,$FE,$86,$6B,$FE,$EE
	.dc.b	$BF,$F0,$0C,$70,$AE,$D1,$F1,$3E,$6F,$A3,$FD,$73,$E9,$1F,$23,$F2
	.dc.b	$0C,$16,$FF,$33,$1A,$7F,$D1,$4D,$C3,$FD,$24,$D7,$FD,$32,$D7,$FB
	.dc.b	$13,$BC,$3F,$90,$36,$7E,$4F,$C1,$78,$1F,$D3,$FC,$07,$F2,$1B,$FC
	.dc.b	$2C,$7C,$EF,$A2,$FE,$E7,$39,$C3,$FC,$69,$B2,$73,$E4,$FF,$88,$AB
	.dc.b	$96,$1F,$EA,$C6,$BF,$DD,$5A,$FF,$9A,$1E,$98,$66,$7F,$17,$D1,$D2
	.dc.b	$3F,$77,$E1,$FE,$06,$FF,$D6,$7F,$46,$F8,$4F,$F0,$3C,$E6,$D3,$9B
	.dc.b	$E3,$CB,$FE,$36,$AF,$C7,$61,$FE,$00,$8C,$12,$76,$7F,$85,$3D,$FF
	.dc.b	$F0,$34,$EE,$3C,$7F,$35,$FD,$CB,$E9,$3C,$93,$FD,$8A,$31,$7E,$47
	.dc.b	$C0,$4A,$FF,$62,$60,$8F,$26,$D2,$72,$DB,$FF,$68,$57,$CF,$9F,$87
	.dc.b	$3F,$E2,$69,$FF,$48,$FC,$1F,$FA,$4B,$E8,$DF,$47,$FE,$23,$C3,$85
	.dc.b	$7D,$C3,$FD,$BF,$5B,$9D,$5F,$F7,$BB,$E3,$5A,$0E,$1F,$EF,$F1,$B9
	.dc.b	$FF,$72,$57,$39,$FF,$15,$CE,$5E,$55,$C6,$FC,$6F,$38,$F4,$9D,$AF
	.dc.b	$C7,$87,$38,$FF,$1E,$61,$6F,$8D,$78,$27,$81,$E0,$AE,$0F,$4B,$F0
	.dc.b	$61,$E3,$8E,$63,$E9,$33,$5F,$EB,$A7,$C9,$0C,$71,$87,$92,$38,$3E
	.dc.b	$1C,$38,$C7,$88,$EA,$E0,$E7,$3E,$9C,$3C,$53,$E5,$AE,$95,$A8,$FF
	.dc.b	$2D,$1F,$44,$A7,$C6,$29,$E5,$33,$F4,$9F,$C8,$3D,$6C,$B6,$33,$D5
	.dc.b	$FE,$93,$6C,$EB,$E4,$7F,$A1,$C7,$26,$BE,$2A,$46,$23,$FC,$DD,$18
	.dc.b	$E8,$E4,$7D,$7F,$A4,$7D,$07,$F9,$CF,$2F,$C4,$46,$2F,$EE,$5F,$17
	.dc.b	$E4,$36,$77,$5C,$73,$FF,$3E,$5F,$1B,$DE,$3E,$8B,$FC,$A6,$1C,$F6
	.dc.b	$E3,$5F,$60,$FF,$7A,$DF,$18,$DD,$FC,$06,$E5,$E9,$CB,$C1,$7C,$1A
	.dc.b	$B0,$F3,$65,$F0,$B0,$F1,$3E,$6A,$DE,$3E,$7C,$BA,$E1,$E8,$C7,$B3
	.dc.b	$0C,$12,$3D,$2F,$E2,$E7,$95,$F7,$AE,$07,$F9,$6E,$3E,$23,$6F,$FD
	.dc.b	$44,$FF,$97,$F1,$F3,$DF,$AC,$FC,$4D,$7C,$B7,$F1,$ED,$7D,$47,$E3
	.dc.b	$BE,$93,$FD,$92,$DE,$9F,$C8,$A8,$CB,$F9,$66,$BF,$9A,$73,$7F,$E7
	.dc.b	$1F,$A2,$FF,$38,$4F,$F8,$D9,$AF,$F1,$53,$A5,$7A,$B8,$BF,$D7,$96
	.dc.b	$F9,$7C,$50,$F1,$38,$69,$D7,$C1,$C4,$F1,$8F,$64,$FA,$A3,$11,$E9
	.dc.b	$8C,$1C,$F0,$FF,$0F,$C7,$9F,$C5,$E7,$3D,$FE,$93,$F9,$D4,$FF,$94
	.dc.b	$1A,$FA,$DB,$9F,$C1,$F3,$38,$74,$FC,$A9,$DD,$7F,$00,$6B,$FE,$DB
	.dc.b	$6B,$EF,$79,$E9,$FE,$DF,$C9,$F4,$9E,$73,$AD,$7F,$98,$BE,$77,$FC
	.dc.b	$F0,$DF,$5F,$E3,$46,$23,$E8,$57,$F8,$FD,$D1,$FC,$3A,$34,$FF,$14
	.dc.b	$6B,$FE,$EF,$6B,$F4,$DE,$99,$FD,$FF,$6B,$E5,$1E,$0B,$FA,$DF,$0F
	.dc.b	$FE,$64,$BF,$F3,$BB,$5F,$C1,$36,$BF,$D7,$1C,$BF,$E7,$A9,$DF,$95
	.dc.b	$F9,$2A,$E1,$1F,$CA,$1A,$FE,$14,$FF,$7E,$46,$CF,$C3,$4B,$E1,$A9
	.dc.b	$E4,$8F,$C5,$79,$DF,$84,$C3,$FD,$90,$DD,$BA,$37,$0F,$C9,$7D,$23
	.dc.b	$C9,$E3,$C9,$FA,$6A,$FC,$6D,$7F,$3C,$6B,$CA,$34,$7E,$66,$9F,$21
	.dc.b	$4E,$73,$FE,$9C,$FA,$47,$9B,$F9,$F6,$C6,$63,$96,$99,$7F,$6F,$3A
	.dc.b	$9F,$E9,$6D,$7F,$A7,$B5,$F9,$1F,$49,$FE,$C2,$7F,$A7,$B5,$FE,$A6
	.dc.b	$D7,$FA,$BB,$5E,$3E,$27,$F0,$68,$FF,$39,$C7,$28,$C1,$FC,$2B,$E9
	.dc.b	$1F,$25,$49,$3F,$D7,$A3,$2D,$FA,$1E,$04,$70,$7F,$64,$F8,$77,$B5
	.dc.b	$FC,$0D,$B1,$FE,$C4,$6B,$FC,$3F,$E3,$5E,$86,$EB,$F8,$ED,$6A,$BF
	.dc.b	$B9,$B5,$FC,$63,$5F,$E2,$4F,$4E,$57,$2F,$E4,$70,$7D,$27,$F1,$56
	.dc.b	$DF,$8F,$0F,$E3,$1F,$49,$F5,$70,$E3,$3F,$C7,$63,$FD,$0E,$FF,$B7
	.dc.b	$1A,$FF,$27,$78,$61,$F9,$F6,$C7,$1F,$63,$A9,$DB,$8B,$BF,$87,$27
	.dc.b	$A3,$E2,$F9,$DC,$47,$58,$B8,$F9,$21,$82,$BE,$77,$C9,$C9,$F9,$CF
	.dc.b	$B3,$E4,$31,$1F,$E1,$AD,$D3,$F3,$89,$1F,$65,$F8,$7F,$49,$23,$99
	.dc.b	$FE,$FC,$FA,$47,$89,$FC,$FD,$72,$53,$11,$27,$84,$65,$E5,$1C,$7E
	.dc.b	$92,$47,$57,$A7,$76,$23,$FA,$A5,$A7,$1B,$7C,$38,$FE,$05,$2F,$A4
	.dc.b	$7C,$B7,$F5,$36,$C7,$F5,$7A,$F9,$51,$88,$FE,$C1,$F4,$9F,$C2,$5D
	.dc.b	$72,$FF,$19,$9D,$47,$F9,$11,$AE,$6F,$FB,$9C,$62,$3F,$B2,$3D,$6F
	.dc.b	$EC,$5B,$E5,$7F,$B3,$36,$BB,$9C,$2F,$3E,$B4,$8C,$89,$FF,$21,$B8
	.dc.b	$7F,$7A,$FA,$A3,$83,$4D,$77,$78,$3B,$3C,$3D,$52,$79,$46,$5E,$91
	.dc.b	$97,$E2,$C7,$F7,$DF,$A4,$7D,$25,$24,$F9,$BD,$1F,$D0,$A9,$DB,$FE
	.dc.b	$68,$6B,$FE,$6C,$6B,$FE,$70,$6B,$FD,$25,$DE,$BF,$A1,$9F,$F0,$BB
	.dc.b	$7E,$53,$C3,$5F,$E1,$99,$76,$7F,$84,$DF,$D3,$47,$97,$AF,$13,$23
	.dc.b	$04,$39,$3C,$7A,$F9,$7D,$24,$9F,$74,$65,$F8,$11,$97,$C5,$1F,$E2
	.dc.b	$3F,$A4,$FE,$79,$1F,$E2,$76,$A6,$93,$BF,$FD,$5F,$DB,$FD,$A7,$E9
	.dc.b	$3F,$DD,$D5,$FE,$CE,$8C,$7C,$36,$A3,$A5,$55,$5C,$1F,$F5,$55,$70
	.dc.b	$79,$1F,$E3,$98,$D7,$C0,$FF,$04,$8E,$66,$8E,$77,$FD,$EF,$AB,$FD
	.dc.b	$4B,$5D,$5B,$19,$9E,$C7,$F1,$73,$C7,$F4,$99,$7C,$0D,$72,$7F,$92
	.dc.b	$E7,$AC,$66,$8F,$B7,$E8,$DF,$49,$FB,$7C,$7D,$18,$27,$3E,$8F,$A7
	.dc.b	$0E,$33,$EB,$9C,$BF,$10,$FF,$2D,$38,$AE,$6D,$CF,$E9,$63,$E1,$C7
	.dc.b	$C1,$1F,$E1,$D9,$D5,$F2,$9C,$BE,$E3,$C0,$E1,$F3,$36,$38,$C7,$C2
	.dc.b	$3C,$86,$B1,$E2,$47,$38,$FB,$1D,$CF,$22,$57,$79,$D6,$3E,$C8,$F0
	.dc.b	$12,$FE,$36,$3E,$D7,$B4,$ED,$CE,$18,$C7,$E0,$5F,$18,$F1,$DF,$17
	.dc.b	$C7,$19,$BF,$AB,$87,$8A,$F8,$E1,$CE,$FE,$43,$D9,$E2,$E5,$F8,$47
	.dc.b	$63,$18,$FB,$A3,$85,$F9,$E7,$89,$1F,$0E,$3E,$57,$44,$BF,$BB,$3D
	.dc.b	$31,$F4,$46,$23,$93,$E8,$69,$F7,$4A,$79,$A7,$A3,$96,$B9,$8E,$51
	.dc.b	$E4,$8F,$B8,$7F,$99,$CF,$79,$70,$FC,$BB,$6B,$FA,$B3,$70,$E0,$FC
	.dc.b	$A2,$CB,$C5,$7C,$4E,$A7,$A9,$F9,$98,$CB,$F8,$A6,$E1,$FA,$CF,$A4
	.dc.b	$FF,$1A,$EB,$FA,$99,$CE,$7C,$70,$C1,$FC,$1E,$B7,$87,$38,$CB,$4F
	.dc.b	$CE,$C6,$B0,$FD,$7D,$F4,$9F,$D9,$B5,$FD,$E3,$67,$85,$F8,$67,$A5
	.dc.b	$72,$BF,$17,$57,$F8,$EA,$31,$3A,$AE,$11,$ED,$9D,$46,$09,$3E,$2A
	.dc.b	$E7,$8F,$AA,$F9,$92,$BF,$D3,$D8,$79,$2F,$81,$DC,$B3,$DE,$B9,$57
	.dc.b	$69,$F6,$DF,$67,$81,$82,$DF,$57,$17,$E1,$7A,$D7,$8B,$0F,$5C,$60
	.dc.b	$87,$08,$D6,$1E,$28,$CE,$5F,$87,$C7,$C5,$87,$59,$E0,$72,$C7,$C5
	.dc.b	$5B,$8F,$3B,$67,$DD,$18,$CB,$ED,$4E,$71,$F0,$CE,$70,$F1,$2C,$3A
	.dc.b	$DF,$DF,$CB,$D7,$3E,$23,$4E,$30,$E9,$03,$39,$7F,$4B,$7B,$BF,$AD
	.dc.b	$BF,$CA,$B6,$BF,$CF,$EE,$7D,$1B,$5F,$71,$AF,$5A,$FF,$62,$CF,$58
	.dc.b	$C4,$E9,$F5,$3E,$89,$DC,$7E,$83,$D3,$C4,$E5,$B5,$FA,$96,$CE,$E3
	.dc.b	$F7,$14,$9F,$ED,$93,$E3,$6B,$F2,$5C,$FC,$73,$E1,$BF,$D7,$B7,$FE
	.dc.b	$4E,$8C,$7C,$0F,$FA,$7B,$B7,$FD,$3F,$1B,$F4,$B1,$5F,$2E,$36,$F1
	.dc.b	$BE,$78,$AF,$8C,$30,$E3,$2B,$E9,$7C,$67,$31,$E5,$FA,$3B,$6F,$A4
	.dc.b	$B1,$86,$DF,$85,$B8,$78,$63,$CD,$5C,$5F,$F3,$3F,$D1,$BD,$AF,$1F
	.dc.b	$1B,$E9,$3F,$D4,$35,$A9,$CC,$63,$0E,$87,$E2,$EF,$95,$7F,$29,$6C
	.dc.b	$7F,$30,$6B,$FC,$E5,$AF,$E4,$C3,$C3,$5F,$F5,$D3,$67,$F3,$1F,$18
	.dc.b	$F7,$BE,$F4,$79,$4F,$A1,$E6,$7C,$6C,$BE,$3B,$63,$27,$D5,$8C,$B4
	.dc.b	$FB,$1E,$BF,$CE,$3E,$7F,$D2,$58,$C9,$33,$F8,$3C,$3F,$89,$9D,$47
	.dc.b	$C4,$1F,$44,$F1,$3E,$FE,$CF,$A4,$FC,$2B,$FE,$D7,$9D,$63,$FE,$9E
	.dc.b	$8D,$73,$7E,$1B,$0E,$78,$7E,$6A,$F7,$1F,$D7,$1A,$FF,$65,$6B,$F3
	.dc.b	$5D,$D7,$79,$FC,$8E,$FF,$2F,$D1,$ED,$5A,$C7,$F0,$FF,$49,$F9,$03
	.dc.b	$F2,$3D,$D6,$3F,$3D,$17,$2F,$87,$CD,$FD,$AE,$72,$7E,$8C,$FC,$CF
	.dc.b	$CD,$FE,$69,$A6,$4F,$D4,$39,$69,$FE,$BC,$C1,$E6,$FE,$5B,$E9,$3F
	.dc.b	$55,$39,$8C,$76,$7A,$9F,$8C,$48,$FE,$F9,$5C,$9B,$1B,$7F,$D9,$BD
	.dc.b	$4B,$FA,$B1,$F2,$67,$F6,$79,$3E,$92,$98,$F8,$7B,$5F,$C2,$67,$4D
	.dc.b	$9D,$34,$D7,$D2,$74,$F9,$5E,$AD,$8C,$9C,$E3,$2D,$3F,$DB,$FF,$46
	.dc.b	$FA,$3F,$B7,$F4,$8F,$8C,$FC,$44,$E6,$BF,$DD,$3C,$CA,$D4,$72,$BF
	.dc.b	$E4,$B5,$BC,$FF,$C4,$ED,$7F,$C6,$0D,$7F,$70,$EF,$A1,$58,$AF,$E3
	.dc.b	$BF,$48,$FA,$47,$D2,$79,$27,$F3,$9F,$48,$E7,$FE,$6A,$DC,$3F,$C6
	.dc.b	$4D,$CF,$FC,$72,$D7,$FC,$85,$CB,$FE,$DA,$9F,$F1,$C3,$63,$FC,$7E
	.dc.b	$D7,$7E,$C7,$F4,$2F,$A4,$7C,$7D,$25,$FF,$90,$63,$2D,$F7,$F8,$11
	.dc.b	$C1,$BF,$37,$A1,$1C,$13,$0F,$E0,$95,$FC,$61,$FE,$0D,$F4,$8F,$A4
	.dc.b	$FD,$F4,$E7,$8B,$E5,$D7,$F4,$DE,$7F,$E8,$F8,$3E,$89,$F0,$1F,$C1
	.dc.b	$7C,$95,$6A,$73,$5F,$DF,$F0,$E7,$87,$F4,$FA,$FE,$C7,$7C,$2B,$FC
	.dc.b	$C0,$D9,$FF,$33,$B5,$FF,$38,$39,$AE,$6D,$9F,$F6,$A5,$BE,$93,$C9
	.dc.b	$87,$F1,$99,$7C,$27,$AA,$7E,$6E,$3F,$C9,$E7,$27,$9C,$FF,$6F,$46
	.dc.b	$B1,$F4,$C6,$A7,$FC,$F6,$D7,$FD,$08,$D7,$FD,$E5,$5F,$E0,$08,$E1
	.dc.b	$3C,$A3,$E3,$7C,$DF,$95,$FE,$5D,$F4,$8F,$91,$F9,$46,$0B,$7E,$5C
	.dc.b	$3F,$C5,$53,$B8,$FF,$44,$36,$3F,$D1,$ED,$7F,$D3,$0D,$7F,$B0,$3B
	.dc.b	$C3,$EA,$B6,$7E,$5F,$C1,$78,$5F,$05,$3D,$EF,$E4,$2F,$F7,$28,$FC
	.dc.b	$27,$D1,$7F,$47,$9C,$E1,$FE,$38,$D9,$39,$F2,$7F,$C4,$75,$CB,$0F
	.dc.b	$F5,$63,$5F,$EE,$8D,$7F,$C9,$0F,$4C,$33,$3F,$8F,$E9,$E9,$1E,$6F
	.dc.b	$0B,$F0,$97,$FE,$AC,$FA,$47,$C2,$7F,$5B,$9C,$DA,$73,$7C,$78,$FF
	.dc.b	$C6,$F5,$F9,$1C,$3F,$C0,$CD,$7F,$C0,$71,$A3,$A1,$FE,$1A,$F7,$BF
	.dc.b	$C1,$11,$AA,$F2,$7C,$EF,$F7,$2F,$A4,$F2,$CB,$04,$E7,$0F,$33,$E2
	.dc.b	$25,$FF,$B1,$67,$6D,$DB,$CE,$47,$04,$BF,$F0,$24,$FD,$28,$FB,$71
	.dc.b	$FE,$73,$F8,$3F,$F3,$A7,$D1,$BE,$8F,$DF,$87,$F8,$06,$B4,$27,$4F
	.dc.b	$F8,$92,$3F,$83,$C6,$23,$FC,$59,$5D,$2B,$9D,$D9,$FF,$75,$4D,$7E
	.dc.b	$15,$F2,$AF,$F7,$3C,$E7,$0F,$45,$78,$67,$2D,$3C,$57,$CD,$F3,$E1
	.dc.b	$E3,$AE,$2F,$96,$33,$5E,$47,$91,$E8,$2C,$F0,$8E,$FF,$49,$CF,$1F
	.dc.b	$E4,$AF,$3A,$63,$FE,$9D,$8D,$63,$6F,$CF,$8E,$2F,$B9,$E2,$3D,$26
	.dc.b	$0E,$67,$8B,$0F,$88,$FC,$83,$15,$EF,$76,$7F,$96,$CF,$51,$4F,$B2
	.dc.b	$53,$67,$F9,$87,$E9,$3F,$92,$FA,$99,$6C,$71,$6B,$D5,$AF,$1F,$90
	.dc.b	$FF,$47,$3B,$6B,$B6,$BB,$E7,$ED,$1C,$31,$FC,$47,$D2,$3E,$83,$FC
	.dc.b	$E9,$18,$6E,$9C,$70,$FA,$F8,$B0,$E5,$3F,$E7,$BC,$39,$57,$E3,$5A
	.dc.b	$F0,$9D,$BF,$ED,$3A,$E5,$8E,$F0,$EF,$3F,$ED,$7B,$E3,$8F,$E7,$6F
	.dc.b	$73,$CE,$F8,$30,$C1,$27,$D4,$E5,$E5,$87,$8A,$15,$D2,$77,$1F,$94
	.dc.b	$BF,$0E,$3F,$06,$78,$E1,$CE,$3E,$06,$3E,$8B,$E7,$8F,$5C,$3E,$0D
	.dc.b	$FF,$72,$6E,$7C,$2B,$F9,$EE,$5E,$1C,$71,$1F,$E8,$BA,$DB,$F0,$F1
	.dc.b	$FF,$54,$61,$FD,$F2,$F8,$4F,$FB,$F7,$E9,$3F,$D9,$CF,$F4,$86,$BF
	.dc.b	$E9,$56,$CF,$F8,$13,$E8,$9D,$9D,$33,$FF,$5A,$56,$9F,$F5,$B4,$FF
	.dc.b	$9C,$AF,$FA,$9B,$5E,$9D,$5C,$5F,$17,$CC,$D5,$68,$C3,$63,$9F,$83
	.dc.b	$89,$91,$E9,$79,$5A,$73,$3F,$EA,$F8,$FF,$11,$BE,$4F,$13,$39,$FF
	.dc.b	$88,$3E,$93,$F9,$EC,$FF,$6A,$6B,$FE,$B7,$6C,$7C,$3F,$33,$D1,$18
	.dc.b	$D3,$FD,$AC,$DA,$07,$0B,$05,$72,$E9,$F4,$FD,$24,$AF,$14,$7C,$5F
	.dc.b	$9C,$FF,$3C,$47,$F7,$E6,$BF,$EC,$B6,$E5,$F5,$FB,$38,$7A,$F2,$B8
	.dc.b	$3F,$1A,$31,$1F,$17,$DA,$F9,$26,$2B,$FC,$87,$F1,$1F,$E6,$29,$FF
	.dc.b	$06,$B5,$FF,$6D,$36,$3F,$C9,$3C,$DF,$E7,$79,$DF,$A3,$F7,$6B,$5C
	.dc.b	$1F,$EF,$B6,$CF,$C6,$AE,$0F,$4A,$FC,$24,$3E,$1E,$9C,$E3,$F1,$1E
	.dc.b	$8F,$E1,$6B,$FD,$D8,$DC,$7F,$CB,$6D,$C7,$E9,$CB,$E9,$3F,$C4,$CF
	.dc.b	$F8,$A2,$74,$FD,$86,$A6,$C7,$08,$EF,$CD,$F9,$98,$7C,$7D,$3C,$B1
	.dc.b	$9C,$7F,$3D,$F4,$9F,$AA,$BF,$F1,$BB,$56,$5F,$E7,$2C,$3E,$2C,$27
	.dc.b	$53,$FE,$76,$6E,$5E,$C6,$CF,$E4,$2B,$8F,$D2,$7F,$B0,$63,$FC,$8A
	.dc.b	$D7,$BD,$7C,$48,$C5,$F5,$EC,$FD,$A4,$7F,$9D,$2B,$8F,$A1,$83,$F7
	.dc.b	$1F,$54,$F8,$8F,$DD,$7D,$2B,$E2,$A9,$2F,$89,$FE,$56,$87,$D1,$70
	.dc.b	$24,$67,$EA,$BF,$82,$7C,$5B,$DD,$A4,$9F,$F2,$FF,$C7,$BD,$2C,$47
	.dc.b	$B0,$FD,$8D,$71,$7F,$CC,$AD,$74,$27,$93,$AA,$FE,$29,$0F,$A6,$A7
	.dc.b	$6E,$1F,$F5,$3F,$D3,$7D,$0B,$3F,$E8,$57,$FC,$E6,$D7,$BB,$B2,$7B
	.dc.b	$7F,$C8,$5D,$4F,$4E,$3C,$78,$92,$F9,$BB,$F9,$FC,$47,$48,$B5,$E4
	.dc.b	$9F,$93,$F3,$52,$BE,$0B,$DF,$8F,$11,$F0,$B6,$4E,$2F,$DB,$57,$9F
	.dc.b	$F5,$39,$BE,$9B,$D0,$67,$1F,$E3,$FF,$4A,$F1,$BF,$9E,$3F,$7A,$3F
	.dc.b	$D2,$AD,$7E,$8B,$B7,$31,$C3,$96,$46,$0E,$3F,$4D,$D0,$8F,$4E,$FA
	.dc.b	$F4,$8B,$A7,$FB,$73,$0F,$DD,$FD,$33,$E6,$3F,$A8,$36,$FF,$D5,$2D
	.dc.b	$7F,$0B,$F4,$DF,$BF,$BF,$F7,$2C,$E5,$B5,$C8,$FE,$B5,$3B,$DB,$F8
	.dc.b	$7D,$6D,$FE,$23,$0F,$97,$D9,$C7,$0F,$F5,$CC,$7F,$89,$DB,$EB,$F4
	.dc.b	$3A,$F8,$16,$32,$43,$CE,$F8,$BB,$3C,$7D,$92,$7A,$C6,$5E,$F1,$97
	.dc.b	$DF,$1F,$DC,$FE,$95,$F4,$D4,$93,$FE,$CE,$F4,$FF,$41,$3A,$8F,$F6
	.dc.b	$8B,$5F,$D6,$B7,$76,$BF,$CC,$AB,$8D,$FF,$80,$5F,$F0,$34,$EA,$FF
	.dc.b	$B4,$C6,$2B,$F9,$CD,$F3,$79,$75,$F1,$EB,$97,$E5,$8E,$86,$1F,$DF
	.dc.b	$DA,$73,$1E,$BE,$9F,$17,$D3,$48,$C1,$0E,$2E,$5E,$71,$FE,$19,$FA
	.dc.b	$6F,$E0,$44,$EE,$C7,$7E,$5F,$F5,$7F,$77,$F5,$2F,$A6,$FF,$6F,$E1
	.dc.b	$FE,$BD,$8C,$7C,$36,$A3,$FC,$13,$87,$F4,$5F,$82,$FD,$AC,$7F,$83
	.dc.b	$2F,$8B,$FE,$9C,$BE,$2F,$33,$57,$CD,$5F,$39,$DC,$71,$36,$71,$15
	.dc.b	$DA,$FC,$31,$CB,$0E,$92,$FA,$6A,$49,$E5,$0F,$B7,$C4,$77,$1F,$4C
	.dc.b	$FA,$7F,$EE,$78,$F6,$8C,$7C,$5F,$2C,$7D,$51,$8B,$E4,$DA,$E1,$7C
	.dc.b	$B3,$F9,$D7,$C2,$31,$7F,$E3,$4B,$E1,$5F,$E0,$46,$C7,$CD,$AE,$18
	.dc.b	$7C,$39,$CC,$FA,$2F,$9B,$F1,$AB,$6F,$92,$33,$3C,$32,$F8,$D3,$CE
	.dc.b	$72,$4B,$F7,$46,$58,$F3,$72,$D2,$63,$F0,$23,$2D,$8C,$E3,$ED,$AF
	.dc.b	$3D,$F8,$A7,$27,$C3,$8F,$13,$F5,$63,$55,$CC,$95,$CE,$FB,$B6,$B9
	.dc.b	$9F,$16,$32,$F5,$7D,$73,$CA,$3E,$71,$89,$E5,$19,$8F,$98,$64,$E4
	.dc.b	$76,$AF,$2C,$F0,$24,$6A,$7F,$2F,$3A,$8C,$4E,$9F,$31,$EF,$C7,$EC
	.dc.b	$D7,$87,$1F,$B1,$18,$AE,$11,$8A,$F2,$1C,$AB,$D1,$7D,$E3,$15,$C3
	.dc.b	$1F,$A3,$87,$E1,$A7,$FC,$A6,$78,$8B,$83,$C7,$7C,$5E,$A7,$E0,$9A
	.dc.b	$8C,$BF,$A2,$6E,$5F,$B6,$FA,$7F,$F1,$27,$3F,$E6,$67,$39,$FE,$7D
	.dc.b	$B7,$FA,$36,$AA,$E2,$F5,$C3,$91,$1F,$DD,$46,$B0,$FD,$C5,$72,$C7
	.dc.b	$D5,$7B,$9F,$27,$5F,$F3,$D7,$71,$9C,$7C,$8F,$F1,$1C,$3A,$46,$31
	.dc.b	$F1,$12,$B9,$E5,$C6,$3D,$57,$5F,$56,$1D,$EF,$95,$F8,$27,$94,$7F
	.dc.b	$A1,$30,$EF,$1C,$8C,$96,$7B,$B8,$3F,$8C,$E5,$EF,$3E,$4C,$F2,$8C
	.dc.b	$E1,$E3,$CB,$E1,$C7,$C8,$C7,$C8,$F5,$C7,$CA,$47,$FD,$93,$7C,$EB
	.dc.b	$8E,$3E,$92,$57,$A6,$F9,$DF,$80,$F3,$57,$09,$D6,$5E,$3C,$BD,$19
	.dc.b	$7C,$FC,$38,$DE,$F0,$F2,$10,$F9,$77,$C6,$BE,$76,$5E,$5C,$3E,$0E
	.dc.b	$1C,$9B,$5A,$8C,$57,$78,$E3,$EA,$7A,$2D,$F5,$B6,$30,$D9,$FC,$24
	.dc.b	$6F,$D6,$C6,$5F,$92,$6E,$BF,$A1,$6E,$BF,$75,$B3,$F3,$1B,$F2,$2F
	.dc.b	$89,$FF,$4C,$63,$FE,$F8,$EF,$FF,$09,$4E,$7E,$17,$FC,$29,$18,$AD
	.dc.b	$E1,$FE,$8F,$AD,$5F,$13,$FE,$00,$BD,$7D,$4E,$31,$8A,$61,$8B,$5E
	.dc.b	$EB,$94,$62,$B7,$18,$9F,$47,$D4,$71,$7D,$5F,$51,$69,$36,$30,$DA
	.dc.b	$E5,$3A,$6C,$F8,$ED,$F4,$FF,$EF,$19,$F1,$F1,$3E,$A3,$FD,$0F,$1E
	.dc.b	$6A,$DF,$BD,$EB,$AD,$70,$73,$C3,$95,$7F,$AD,$67,$37,$C6,$73,$BF
	.dc.b	$F9,$06,$73,$5F,$EB,$69,$CD,$7F,$BE,$9B,$0F,$90,$7C,$0F,$FB,$DE
	.dc.b	$BA,$65,$F0,$FD,$0F,$88,$D8,$F8,$2D,$BF,$B5,$B7,$4E,$0F,$C2,$5F
	.dc.b	$0B,$F8,$FE,$CF,$CD,$3E,$77,$D4,$58,$C9,$30,$E7,$19,$7F,$D9,$FF
	.dc.b	$4D,$FE,$E7,$9F,$BB,$BB,$EA,$3F,$16,$FF,$98,$69,$5B,$7F,$DB,$91
	.dc.b	$BE,$6F,$AD,$39,$8F,$CA,$D7,$2B,$5C,$AD,$47,$E7,$67,$53,$9F,$23
	.dc.b	$99,$F9,$3E,$0F,$97,$E9,$E2,$FF,$0F,$F5,$1F,$8D,$3F,$1D,$E0,$58
	.dc.b	$C9,$30,$F9,$71,$97,$E6,$C7,$E5,$7E,$7B,$F2,$CD,$3F,$26,$58,$DB
	.dc.b	$F9,$68,$F7,$FD,$3F,$E3,$7E,$A3,$F4,$75,$F3,$23,$1D,$BA,$BF,$CB
	.dc.b	$D6,$B7,$F9,$9B,$AF,$75,$F9,$98,$C3,$63,$B7,$07,$3C,$7F,$7C,$3E
	.dc.b	$A3,$F6,$0D,$AF,$DA,$2F,$8A,$CC,$C9,$CB,$C2,$31,$3D,$3E,$A3,$9F
	.dc.b	$C6,$74,$6D,$F4,$6C,$EE,$3B,$CF,$ED,$FE,$9D,$F4,$9F,$67,$EA,$1F
	.dc.b	$21,$98,$C3,$6F,$F8,$6F,$32,$B7,$1F,$BB,$BE,$6E,$A3,$38,$B8,$3F
	.dc.b	$2D,$5A,$8C,$57,$E4,$14,$66,$7F,$D7,$9F,$50,$FA,$87,$D4,$78,$EF
	.dc.b	$89,$F9,$8F,$A8,$73,$FF,$13,$F3,$FF,$2F,$29,$CE,$7F,$C4,$67,$3E
	.dc.b	$33,$DD,$FE,$34,$FD,$39,$FE,$2C,$E5,$BC,$9F,$C7,$FE,$99,$F2,$9C
	.dc.b	$2F,$0D,$AF,$E3,$D2,$F8,$2F,$E4,$DF,$5A,$F6,$79,$46,$1E,$9E,$8E
	.dc.b	$2F,$8B,$76,$F7,$8C,$3F,$8C,$38,$6F,$1F,$F0,$3F,$A9,$7D,$4F,$4C
	.dc.b	$3F,$98,$79,$6E,$CF,$E9,$50,$FA,$77,$C2,$7F,$3B,$AD,$4F,$F4,$C3
	.dc.b	$51,$9A,$E3,$5C,$27,$AC,$62,$75,$18,$39,$4B,$EA,$73,$A7,$B2,$5F
	.dc.b	$11,$E4,$9F,$9B,$1F,$D1,$CC,$9E,$37,$33,$C6,$78,$71,$7F,$53,$AD
	.dc.b	$CF,$E1,$9F,$04,$66,$64,$79,$3E,$7E,$D7,$CE,$FA,$97,$C3,$7B,$8F
	.dc.b	$EB,$CE,$2B,$D7,$18,$34,$FF,$5D,$8D,$4F,$87,$2B,$C1,$FC,$DE,$6D
	.dc.b	$7E,$85,$B4,$8F,$A9,$F0,$9D,$5F,$ED,$1C,$9F,$C5,$5F,$E6,$E7,$D9
	.dc.b	$FA,$6F,$EE,$51,$9A,$E5,$8A,$73,$E6,$7D,$86,$DF,$D4,$6D,$7E,$3A
	.dc.b	$F9,$4F,$E1,$7A,$F3,$9F,$F4,$FF,$13,$F0,$B7,$FD,$D7,$EA,$5F,$11
	.dc.b	$FE,$03,$9C,$D7,$F7,$59,$CE,$1F,$E0,$19,$CB,$FD,$43,$04,$EB,$0F
	.dc.b	$CF,$D6,$9F,$F4,$EC,$66,$FC,$5C,$BD,$CF,$F0,$57,$7F,$F7,$EA,$FE
	.dc.b	$FD,$1E,$3F,$9F,$FF,$14,$53,$EA,$7A,$57,$F8,$0E,$73,$18,$C5,$18
	.dc.b	$BE,$59,$FF,$72,$AD,$6D,$EF,$E1,$F8,$0D,$9F,$73,$87,$FC,$41,$8F
	.dc.b	$F6,$EF,$81,$FE,$DF,$F5,$0F,$A8,$FE,$45,$A7,$D7,$BD,$DA,$F7,$5A
	.dc.b	$C3,$F0,$D2,$C3,$93,$FE,$C6,$AD,$BF,$DF,$30,$E8,$EE,$FF,$D8,$D3
	.dc.b	$A8,$FF,$12,$5F,$4B,$E7,$1F,$CD,$67,$57,$CA,$73,$3E,$B3,$A1,$DE
	.dc.b	$B9,$5F,$68,$C9,$DB,$0F,$03,$C6,$BB,$DF,$5A,$E1,$18,$BE,$75,$ED
	.dc.b	$39,$D6,$DC,$4E,$EF,$AC,$7C,$93,$75,$C9,$F6,$18,$8D,$E1,$EE,$9D
	.dc.b	$4F,$97,$0F,$A1,$5A,$8F,$F7,$71,$F0,$4B,$58,$70,$7F,$5D,$FA,$7F
	.dc.b	$E3,$BE,$EF,$06,$C7,$FB,$DD,$AF,$FB,$F9,$AF,$1F,$92,$FE,$D4,$ED
	.dc.b	$AE,$DA,$EF,$83,$B4,$70,$CF,$ED,$FD,$43,$E9,$3F,$B8,$46,$1B,$A7
	.dc.b	$1C,$3E,$BE,$2C,$39,$4F,$F7,$7C,$39,$57,$E2,$5A,$F0,$9D,$BF,$E4
	.dc.b	$CA,$E5,$8E,$F0,$EF,$3F,$E5,$0B,$E3,$8F,$E4,$2F,$73,$CE,$F8,$30
	.dc.b	$C1,$27,$D4,$E5,$E5,$87,$8A,$15,$D2,$77,$1F,$94,$BF,$0E,$3F,$06
	.dc.b	$78,$E1,$CE,$3E,$06,$3E,$8B,$E7,$8F,$5C,$3E,$0D,$FF,$9C,$DB,$A7
	.dc.b	$0A,$FF,$50,$65,$E1,$C7,$11,$FD,$F2,$B6,$FC,$3C,$7F,$C4,$98,$7F
	.dc.b	$9F,$6F,$84,$FF,$9D,$BE,$A3,$FD,$86,$FF,$AA,$DA,$FF,$81,$9B,$3F
	.dc.b	$E8,$5F,$A7,$70,$F4,$CF,$FC,$61,$5A,$7F,$C6,$53,$FE,$5F,$BF,$F5
	.dc.b	$B3,$5E,$9D,$5C,$5F,$17,$CD,$D5,$68,$C3,$63,$9F,$7E,$26,$47,$A5
	.dc.b	$E5,$69,$CC,$FF,$88,$23,$FC,$2A,$F9,$3C,$3C,$E7,$FE,$A1,$FA,$8F
	.dc.b	$E6,$F3,$FE,$CD,$6B,$FE,$28,$6C,$7C,$3F,$33,$D1,$18,$D3,$FC,$A4
	.dc.b	$DA,$FF,$51,$B5,$61,$FE,$47,$7B,$75,$7A,$3E,$A2,$46,$0F,$F6,$F7
	.dc.b	$CF,$FF,$9B,$1B,$EB,$FC,$78,$C4,$7B,$7B,$71,$32,$36,$7D,$28,$C3
	.dc.b	$F0,$FD,$0F,$8E,$63,$5F,$F6,$17,$C4,$7F,$95,$27,$FD,$DC,$D7,$FC
	.dc.b	$92,$D8,$FF,$64,$73,$7F,$9A,$27,$7E,$4F,$DB,$AD,$6F,$FE,$75,$6C
	.dc.b	$FF,$B1,$AB,$83,$D2,$BF,$05,$0F,$88,$A7,$38,$FC,$37,$A1,$F8,$3A
	.dc.b	$FF,$2D,$B7,$1F,$F6,$D3,$71,$FF,$42,$7D,$3F,$F9,$9E,$BF,$DA,$B5
	.dc.b	$AE,$6E,$64,$8C,$D7,$0F,$23,$11,$FE,$68,$B7,$C7,$FE,$5D,$B1,$C0
	.dc.b	$FF,$70,$E3,$E1,$7F,$DC,$71,$F0,$17,$D4,$7E,$A9,$E0,$63,$B3,$A6
	.dc.b	$9F,$A3,$9D,$4F,$F9,$E1,$BA,$7F,$BD,$DB,$9F,$BA,$BF,$55,$F5,$1F
	.dc.b	$EB,$4B,$EA,$D7,$FD,$02,$DB,$E5,$87,$0F,$3E,$57,$07,$F6,$9F,$50
	.dc.b	$F8,$9F,$DB,$7D,$43,$E1,$29,$2F,$81,$EC,$87,$D2,$70,$24,$67,$EA
	.dc.b	$3F,$80,$7C,$57,$FA,$76,$3E,$1B,$5F,$F4,$BB,$74,$FF,$0B,$C3,$E3
	.dc.b	$DE,$8F,$2F,$76,$4F,$27,$F8,$64,$EA,$30,$7D,$05,$F5,$1A,$38,$C7
	.dc.b	$FA,$9B,$EA,$3E,$6A,$0F,$F9,$F1,$FB,$2D,$78,$70,$FF,$1B,$8E,$31
	.dc.b	$C3,$DA,$C1,$99,$D7,$8F,$27,$9B,$C7,$E8,$71,$1E,$38,$BA,$7C,$2C
	.dc.b	$15,$F2,$BE,$56,$5F,$CA,$7C,$5C,$78,$8F,$97,$16,$F9,$4D,$E8,$FF
	.dc.b	$75,$60,$FA,$8F,$51,$9C,$3E,$17,$D3,$BC,$7F,$E7,$0F,$DE,$8F,$CD
	.dc.b	$B5,$FF,$67,$BB,$73,$1C,$39,$64,$60,$E3,$F5,$1D,$08,$F4,$F0,$AB
	.dc.b	$D2,$2E,$9F,$84,$C3,$F6,$12,$FA,$87,$CC,$7F,$4D,$6C,$7E,$C1,$AF
	.dc.b	$FB,$93,$EA,$3F,$83,$3F,$EE,$68,$CD,$FE,$41,$BB,$7E,$0E,$7B,$5F
	.dc.b	$F5,$9B,$E3,$7B,$8C,$7C,$BC,$AF,$04,$6F,$D8,$91,$FC,$0A,$FF,$C4
	.dc.b	$CD,$F0,$32,$F9,$66,$E1,$FA,$2A,$FF,$6F,$CF,$1E,$BF,$17,$5C,$8C
	.dc.b	$10,$F7,$39,$7E,$04,$7F,$71,$FA,$77,$D4,$52,$4F,$F1,$6F,$53,$F3
	.dc.b	$73,$A8,$FE,$32,$D5,$8F,$ED,$A3,$0F,$EE,$27,$8D,$7F,$7F,$7F,$C0
	.dc.b	$91,$9F,$96,$EF,$7C,$B7,$F9,$D5,$FD,$9E,$30,$7E,$CD,$FD,$0B,$E5
	.dc.b	$EA,$E9,$D7,$23,$04,$38,$89,$E9,$87,$97,$AB,$C9,$F5,$12,$BE,$04
	.dc.b	$79,$CC,$3C,$A9,$39,$8F,$E7,$7F,$51,$FC,$00,$9D,$DF,$E1,$D7,$F8
	.dc.b	$36,$1F,$C6,$9F,$E7,$BF,$00,$EF,$FC,$67,$D3,$FF,$B8,$B0,$FF,$5F
	.dc.b	$46,$3E,$1F,$FD,$54,$FE,$46,$F8,$B9,$9D,$9F,$C5,$EB,$93,$FE,$C1
	.dc.b	$8D,$34,$FE,$39,$3C,$9E,$15,$D9,$A7,$41,$5D,$2F,$D1,$1C,$B1,$E0
	.dc.b	$EB,$E9,$AD,$1A,$63,$EA,$2C,$F5,$FA,$77,$D4,$7F,$77,$C7,$AC,$A7
	.dc.b	$3E,$A7,$8F,$1E,$93,$C6,$F8,$6B,$FB,$5B,$DC,$FB,$CE,$53,$A7,$E3
	.dc.b	$65,$E8,$9D,$E3,$E0,$6D,$F3,$BE,$17,$ED,$3C,$18,$C7,$B5,$70,$7A
	.dc.b	$99,$BE,$95,$CA,$75,$3C,$CE,$31,$92,$57,$BF,$0F,$34,$62,$77,$39
	.dc.b	$7B,$61,$E7,$8B,$1E,$23,$64,$C3,$C7,$1C,$63,$13,$C5,$A7,$19,$DB
	.dc.b	$83,$E5,$CF,$18,$C9,$27,$93,$F6,$B1,$FA,$F5,$C7,$1F,$82,$7C,$28
	.dc.b	$CC,$F1,$9F,$EE,$13,$AC,$3E,$8D,$6A,$FA,$92,$B7,$3A,$C7,$E3,$E3
	.dc.b	$F0,$A7,$CE,$7E,$0E,$F9,$3C,$23,$58,$78,$EB,$8C,$F9,$EF,$A3,$A7
	.dc.b	$E9,$63,$ED,$9D,$BE,$99,$EE,$E5,$E1,$7E,$3C,$3C,$31,$AC,$7D,$4D
	.dc.b	$3E,$83,$F4,$DF,$EA,$07,$A8,$B8,$BC,$A7,$C5,$EA,$7D,$85,$9D,$47
	.dc.b	$EB,$E7,$73,$FA,$5F,$A8,$FF,$15,$F3,$E6,$A4,$FE,$72,$B7,$9F,$ED
	.dc.b	$2F,$73,$9A,$DC,$FE,$A6,$B7,$8F,$F6,$7B,$DC,$E6,$B7,$3F,$B7,$AD
	.dc.b	$DF,$F8,$BE,$F7,$39,$AD,$CF,$E1,$2B,$6A,$B6,$E6,$76,$FF,$09,$9D
	.dc.b	$CE,$A3,$13,$AC,$7E,$2C,$7C,$3E,$B6,$59,$C3,$BD,$6B,$2F,$85,$5A
	.dc.b	$BF,$84,$48,$F7,$3C,$EF,$96,$1E,$4C,$7C,$58,$73,$CB,$DD,$7C,$5F
	.dc.b	$91,$8F,$5C,$79,$92,$36,$F8,$B1,$EB,$5C,$4F,$95,$5C,$6B,$64,$8F
	.dc.b	$8C,$79,$4E,$07,$C4,$2E,$1E,$6C,$BE,$09,$32,$FE,$89,$19,$6D,$F7
	.dc.b	$AE,$B8,$F9,$5E,$71,$97,$B4,$64,$F6,$3F,$81,$F5,$B5,$6F,$A8,$B1
	.dc.b	$86,$E7,$E2,$8D,$FA,$F8,$BF,$CE,$37,$0F,$CC,$37,$4F,$CA,$36,$7F
	.dc.b	$20,$DF,$91,$7C,$57,$FA,$7F,$1F,$F5,$F4,$63,$E1,$BF,$01,$D3,$F8
	.dc.b	$18,$DF,$AB,$8A,$FF,$44,$46,$DE,$37,$CF,$1F,$F0,$DC,$E5,$B1,$9B
	.dc.b	$E8,$B0,$E9,$2A,$F2,$FD,$46,$1F,$3F,$D4,$5B,$4D,$8C,$37,$1F,$1B
	.dc.b	$6F,$FC,$4B,$F4,$FF,$EF,$19,$FF,$6E,$71,$BE,$A3,$FC,$2D,$1F,$ED
	.dc.b	$89,$D5,$7F,$8D,$2B,$5B,$F9,$A1,$3A,$D7,$F4,$D5,$AB,$E5,$EF,$FB
	.dc.b	$23,$17,$E2,$8E,$57,$C6,$BF,$D6,$5F,$20,$F8,$27,$16,$AF,$E0,$79
	.dc.b	$DF,$05,$B7,$F4,$1B,$8F,$CE,$6E,$7F,$39,$F9,$1E,$D4,$AF,$AF,$F5
	.dc.b	$16,$FE,$AC,$DB,$FA,$B3,$6B,$FD,$09,$F4,$BF,$EE,$77,$FD,$91,$3F
	.dc.b	$86,$FA,$8F,$B9,$1A,$F2,$7E,$DC,$A7,$38,$FF,$B3,$EF,$93,$FE,$60
	.dc.b	$9D,$74,$7C,$D9,$CC,$75,$8C,$5C,$C3,$C2,$7C,$8D,$9E,$9E,$71,$FC
	.dc.b	$7F,$D4,$7E,$1C,$FC,$47,$81,$6F,$F2,$F1,$67,$F2,$B3,$72,$FC,$AD
	.dc.b	$7E,$3F,$E7,$FF,$20,$D3,$F3,$E5,$AF,$CE,$B6,$7F,$DF,$7A,$3C,$FF
	.dc.b	$C1,$7D,$47,$E7,$EB,$D3,$7F,$0E,$73,$DD,$F0,$5B,$3A,$95,$F3,$8F
	.dc.b	$CC,$DF,$39,$FF,$40,$CE,$5B,$5E,$0D,$FA,$5F,$EF,$A5,$F5,$1F,$AD
	.dc.b	$6C,$67,$E2,$B1,$32,$38,$3B,$8C,$4F,$EF,$87,$D4,$73,$F8,$CE,$8D
	.dc.b	$BF,$03,$67,$83,$75,$FE,$9D,$0F,$A6,$7D,$47,$F0,$4E,$C7,$D4,$3E
	.dc.b	$43,$F6,$55,$AC,$BE,$5D,$6A,$73,$7C,$6F,$85,$7C,$BB,$E1,$87,$FA
	.dc.b	$C3,$0E,$13,$9B,$E2,$FD,$6F,$A8,$7D,$43,$EA,$33,$87,$FA,$C3,$EA
	.dc.b	$1C,$FF,$C1,$7C,$FF,$CF,$CA,$73,$8F,$DB,$9C,$F8,$AF,$6F,$F8,$62
	.dc.b	$E4,$B1,$87,$F8,$87,$D3,$BE,$4B,$38,$7E,$16,$30,$FE,$1E,$FA,$39
	.dc.b	$F8,$1F,$E3,$1F,$50,$F5,$F9,$C6,$1E,$BE,$56,$2B,$A3,$76,$FF,$66
	.dc.b	$E3,$FC,$52,$72,$7E,$57,$EA,$1F,$51,$CA,$3F,$31,$E9,$BB,$7F,$9E
	.dc.b	$C3,$E9,$BF,$DA,$CE,$A3,$31,$8C,$7F,$DB,$75,$C6,$BF,$4E,$71,$F8
	.dc.b	$2F,$E6,$51,$99,$FF,$74,$CE,$4E,$84,$73,$4F,$A8,$E3,$87,$F3,$BA
	.dc.b	$7C,$4F,$9A,$7E,$5D,$FF,$BB,$27,$27,$8A,$33,$E0,$7E,$62,$33,$3E
	.dc.b	$29,$CC,$22,$3A,$9F,$F1,$63,$E3,$E3,$F1,$5C,$8F,$7E,$2F,$9F,$C6
	.dc.b	$1F,$EF,$EF,$A8,$7C,$A7,$C8,$3F,$AA,$38,$BE,$31,$83,$4F,$F0,$B8
	.dc.b	$D4,$FC,$49,$D4,$E5,$FA,$7F,$09,$89,$67,$FB,$6E,$4F,$E2,$0F,$AB
	.dc.b	$27,$D3,$7F,$65,$9C,$E3,$F0,$27,$2F,$F1,$B8,$CD,$FF,$94,$A7,$2F
	.dc.b	$69,$F9,$0F,$E5,$FC,$CF,$C2,$DF,$56,$4F,$C6,$E9,$B5,$C4,$FE,$D7
	.dc.b	$C3,$FD,$C7,$EA,$1F,$0F,$FD,$E7,$C5,$FC,$46,$BF,$DE,$27,$38,$FF
	.dc.b	$2F,$9C,$BF,$67,$14,$EB,$1E,$C2,$B6,$FF,$A5,$A3,$57,$FC,$72,$73
	.dc.b	$3B,$C7,$FC,$09,$DF,$FE,$01,$8F,$3E,$3E,$1F,$A0,$FE,$D3,$F5,$1C
	.dc.b	$30,$46,$30,$F5,$36,$B8,$1F,$D3,$31,$5F,$1D,$BE,$07,$0F,$C1,$6D
	.dc.b	$7B,$DC,$3F,$D3,$F2,$FC,$C7,$C0,$FE,$67,$EA,$1F,$51,$F5,$35,$FA
	.dc.b	$F9,$FF,$38,$BD,$8A,$DB,$FE,$1C,$C7,$FD,$15,$18,$8F,$BB,$87,$5C
	.dc.b	$3A,$56,$9F,$F5,$F4,$66,$79,$63,$5E,$27,$79,$E3,$19,$8F,$1D,$F7
	.dc.b	$C7,$F1,$AF,$86,$33,$5C,$1E,$47,$FB,$4A,$FB,$57,$09,$F1,$CE,$8F
	.dc.b	$C0,$61,$E7,$78,$19,$9E,$0E,$2F,$C5,$3B,$8F,$4C,$63,$0F,$44,$30
	.dc.b	$F0,$D6,$E3,$CE,$7C,$23,$C4,$60,$F1,$CF,$29,$FF,$73,$47,$C6,$2B
	.dc.b	$87,$EB,$9F,$EF,$0F,$A7,$FE,$43,$F0,$9C,$1B,$1F,$EE,$F6,$BB,$EA
	.dc.b	$FF,$44,$B9,$F9,$3F,$E0,$AD,$D7,$94,$62,$3B,$35,$FD,$ED,$23,$8E
	.dc.b	$7F,$85,$FA,$87,$D2,$7F,$02,$8C,$37,$7E,$35,$C1,$B8,$FE,$16,$F8
	.dc.b	$E3,$6F,$83,$FC,$CF,$1F,$05,$7E,$2D,$B1,$F7,$6B,$8B,$6B,$8A,$9D
	.dc.b	$E3,$F9,$B6,$E1,$FC,$62,$F8,$CF,$A2,$BB,$63,$E7,$7C,$4D,$53,$C2
	.dc.b	$B9,$61,$E0,$C3,$AD,$70,$C7,$D8,$C3,$F2,$91,$A7,$CD,$87,$5C,$3A
	.dc.b	$47,$F5,$36,$BA,$AE,$47,$F8,$CA,$3E,$0B,$6F,$F9,$9B,$FE,$36,$BE
	.dc.b	$D3,$AD,$7C,$F7,$FE,$57,$8C,$3E,$73,$F5,$1F,$51,$FE,$BF,$7F,$CB
	.dc.b	$8D,$7F,$95,$B6,$FF,$C8,$7F,$4E,$E2,$F0,$E1,$FD,$12,$B4,$FF,$46
	.dc.b	$9F,$F3,$75,$FF,$99,$DA,$F4,$EA,$E2,$FF,$73,$A7,$CD,$E2,$11,$87
	.dc.b	$CB,$19,$AF,$EE,$51,$82,$1E,$88,$D7,$81,$CE,$B9,$4F,$F4,$6A,$78
	.dc.b	$9C,$CD,$5F,$87,$EA,$3F,$9E,$D7,$F9,$FD,$AF,$F4,$E6,$E9,$F1,$3C
	.dc.b	$DE,$1D,$52,$B8,$3E,$78,$C4,$78,$7E,$A3,$D4,$62,$BF,$CD,$FF,$3F
	.dc.b	$FE,$6B,$9F,$F4,$73,$5F,$EB,$AD,$8F,$A7,$DB,$F1,$E3,$1C,$1F,$DD
	.dc.b	$9B,$5F,$E1,$F6,$AD,$7F,$C8,$4F,$6F,$63,$E6,$7A,$12,$30,$7F,$A7
	.dc.b	$FE,$0B,$FC,$BC,$DF,$27,$EB,$46,$23,$FB,$8F,$47,$F9,$C1,$D1,$0F
	.dc.b	$EE,$B1,$BD,$3F,$D1,$B3,$B9,$CD,$7E,$06,$1F,$0F,$4E,$F1,$F8,$5F
	.dc.b	$43,$ED,$DF,$FA,$59,$BA,$FF,$85,$9B,$87,$F8,$FF,$EA,$3F,$CD,$CF
	.dc.b	$FA,$E6,$32,$FF,$BF,$39,$3A,$1F,$91,$9D,$C7,$F8,$32,$77,$0F,$8F
	.dc.b	$D1,$C2,$7F,$D8,$91,$ED,$BF,$F1,$3F,$D4,$7E,$9E,$7F,$D9,$0D,$7F
	.dc.b	$C2,$F3,$B7,$FC,$A9,$19,$C7,$FD,$62,$DA,$E7,$D7,$E5,$CF,$F4,$DF
	.dc.b	$51,$FE,$B4,$C3,$FD,$A8,$D7,$FC,$46,$DC,$39,$D7,$0F,$0A,$57,$07
	.dc.b	$F6,$1F,$50,$F8,$9F,$D9,$7D,$43,$E1,$29,$2F,$81,$FE,$E3,$87,$D2
	.dc.b	$70,$24,$67,$EA,$3F,$79,$F1,$4F,$72,$92,$7F,$DD,$9F,$1E,$F4,$71
	.dc.b	$1F,$E3,$CA,$E0,$FF,$BC,$5A,$FF,$90,$27,$83,$9A,$FE,$15,$0F,$A8
	.dc.b	$A7,$4E,$0F,$F6,$EF,$D4,$7C,$DB,$FF,$42,$3F,$E7,$A7,$FD,$FC,$D7
	.dc.b	$FC,$9A,$ED,$34,$D4,$74,$F6,$7F,$8D,$46,$7C,$4A,$4A,$EB,$F4,$35
	.dc.b	$EB,$16,$FC,$B3,$F2,$3E,$56,$57,$C1,$7A,$F1,$62,$3E,$34,$5A,$F6
	.dc.b	$CD,$E8,$FF,$3C,$E2,$FA,$8F,$29,$9C,$3F,$D4,$9F,$4E,$F1,$BF,$99
	.dc.b	$3F,$72,$3F,$D5,$18,$B0,$FF,$41,$CE,$67,$FC,$E2,$F1,$77,$1A,$AE
	.dc.b	$7C,$92,$72,$78,$23,$BF,$D4,$48,$C8,$B4,$79,$3C,$1D,$4F,$82,$2E
	.dc.b	$3E,$58,$F5,$CF,$F4,$BF,$A8,$7C,$C5,$24,$F6,$FA,$8F,$DF,$DF,$18
	.dc.b	$57,$07,$C6,$D7,$FD,$1E,$D8,$E1,$3C,$B4,$FF,$53,$CE,$5F,$F5,$4C
	.dc.b	$3E,$5F,$FA,$F3,$6F,$AE,$18,$F5,$A4,$79,$5D,$F7,$E2,$3F,$D3,$F5
	.dc.b	$FE,$37,$74,$7A,$E3,$AF,$5F,$6E,$A9,$3E,$F8,$CB,$F0,$63,$2F,$48
	.dc.b	$FE,$E5,$F5,$0F,$A8,$A4,$9F,$AB,$E9,$FE,$8E,$75,$1F,$61,$AF,$FA
	.dc.b	$D1,$BB,$FC,$D9,$E1,$5F,$DF,$DF,$F0,$2C,$67,$E5,$BB,$5F,$FA,$F2
	.dc.b	$5D,$7B,$EB,$93,$E2,$8E,$46,$1E,$11,$87,$CF,$D5,$DF,$EA,$24,$60
	.dc.b	$87,$77,$2F,$86,$3F,$C2,$9F,$51,$FC,$04,$9E,$36,$3B,$F2,$FF,$AA
	.dc.b	$BB,$FF,$DC,$FF,$4B,$FE,$DE,$C3,$FD,$73,$18,$F8,$6D,$40,$61,$FD
	.dc.b	$17,$E0,$55,$B9,$E2,$6E,$FA,$3F,$E9,$58,$D3,$4E,$95,$CE,$33,$5D
	.dc.b	$27,$95,$F8,$95,$F4,$9D,$47,$93,$0D,$4B,$E9,$A9,$27,$CA,$72,$FA
	.dc.b	$8C,$3E,$31,$F4,$EF,$A8,$FE,$E7,$8F,$68,$C7,$C6,$72,$C7,$B5,$F4
	.dc.b	$9C,$BC,$DB,$5C,$6F,$9E,$9F,$43,$2F,$65,$6A,$3E,$75,$F8,$AB,$6F
	.dc.b	$D1,$9D,$3F,$36,$B8,$CE,$A3,$86,$3E,$38,$C6,$3E,$69,$F2,$46,$EF
	.dc.b	$9C,$70,$3E,$14,$EF,$1F,$45,$F3,$24,$EB,$0F,$0C,$62,$77,$39,$31
	.dc.b	$8F,$B3,$0E,$E7,$83,$0E,$D1,$9C,$7C,$84,$C3,$E7,$63,$DB,$0F,$4C
	.dc.b	$74,$7E,$11,$4F,$56,$3E,$23,$F0,$B8,$76,$8F,$95,$7D,$5E,$67,$9C
	.dc.b	$B3,$C7,$0F,$25,$F8,$A7,$B6,$3E,$AB,$F9,$B8,$F9,$70,$E1,$5D,$AF
	.dc.b	$83,$D3,$0E,$6D,$3D,$25,$3D,$2E,$A7,$D3,$5D,$63,$04,$9E,$33,$E0
	.dc.b	$78,$9E,$13,$D8,$E7,$0F,$45,$F9,$23,$18,$7C,$37,$BD,$7B,$A3,$17
	.dc.b	$F9,$78,$C5,$7C,$FC,$7E,$64,$7E,$24,$FC,$C9,$4F,$C0,$79,$4F,$8B
	.dc.b	$CC,$79,$4D,$E4,$9D,$47,$EA,$5B,$97,$EF,$BE,$A3,$FC,$61,$E1,$7E
	.dc.b	$BE,$73,$9F,$6C,$70,$7E,$FA,$B7,$87,$38,$CB,$4F,$E0,$11,$AC,$3F
	.dc.b	$7D,$7D,$27,$F6,$0D,$7F,$82,$B6,$78,$5F,$86,$7A,$57,$2B,$F1,$76
	.dc.b	$7F,$7F,$8C,$4E,$AB,$84,$7B,$67,$51,$82,$4F,$8A,$B9,$E3,$EA,$BE
	.dc.b	$64,$AF,$E1,$38,$79,$2F,$81,$DC,$B3,$DE,$B9,$57,$69,$F6,$DF,$67
	.dc.b	$81,$82,$DF,$57,$17,$E1,$7A,$D7,$8B,$0F,$5C,$60,$87,$08,$D6,$1E
	.dc.b	$28,$CE,$5F,$83,$C7,$C5,$87,$59,$E0,$72,$C7,$C5,$5B,$8F,$B7,$8F
	.dc.b	$81,$D8,$9D,$E5,$F6,$67,$55,$C6,$FE,$25,$FB,$BD,$5E,$58,$3E,$A2
	.dc.b	$C6,$1B,$9F,$E3,$E3,$7E,$BE,$2F,$F1,$AD,$C3,$F3,$4D,$D7,$EB,$B6
	.dc.b	$7F,$B7,$62,$8C,$7C,$8B,$E2,$BF,$D3,$D9,$7F,$95,$23,$1F,$0D,$FE
	.dc.b	$4B,$F0,$3F,$C9,$91,$BF,$57,$15,$FD,$E6,$36,$F1,$BE,$79,$2F,$8C
	.dc.b	$30,$E3,$2B,$E9,$7C,$67,$31,$E5,$FA,$7B,$6F,$A8,$B1,$86,$DF,$85
	.dc.b	$B8,$78,$63,$CD,$5C,$5F,$F3,$6F,$D3,$FF,$BB,$9F,$F3,$9F,$2B,$EA
	.dc.b	$3F,$C2,$31,$AF,$7F,$CF,$39,$E1,$F4,$E2,$C3,$93,$DC,$FE,$29,$87
	.dc.b	$27,$06,$EF,$9C,$67,$6F,$F3,$85,$72,$6C,$FC,$6F,$90,$7C,$17,$D9
	.dc.b	$CF,$C8,$7C,$FF,$3F,$E2,$D7,$C6,$6C,$64,$FA,$91,$96,$9F,$5F,$DA
	.dc.b	$F9,$A7,$CE,$FA,$8B,$19,$26,$9E,$DE,$2F,$C4,$4E,$A3,$DB,$F4,$BF
	.dc.b	$EE,$9A,$FF,$4E,$E8,$FA,$8F,$C1,$BF,$E8,$D9,$D5,$7C,$78,$D7,$37
	.dc.b	$CC,$9C,$C7,$E4,$6B,$8D,$FF,$86,$A7,$2F,$FA,$D2,$33,$7F,$E4,$B9
	.dc.b	$CB,$7C,$8F,$95,$B7,$FA,$46,$FF,$D5,$B5,$AC,$FF,$45,$F5,$1F,$8E
	.dc.b	$3F,$1F,$E0,$5A,$FC,$E4,$5C,$7E,$17,$37,$F4,$D9,$C9,$FA,$13,$F2
	.dc.b	$FF,$3F,$F9,$86,$99,$3F,$4C,$E5,$A7,$93,$CF,$FC,$7F,$D4,$7E,$9E
	.dc.b	$30,$7E,$59,$C5,$FD,$09,$DF,$67,$C9,$6D,$70,$9D,$3C,$78,$74,$FE
	.dc.b	$FE,$1F,$51,$FB,$06,$C6,$7E,$2B,$13,$23,$83,$B8,$C4,$FE,$14,$7D
	.dc.b	$47,$3F,$8E,$E8,$DB,$F0,$36,$78,$37,$97,$FD,$A5,$F4,$EF,$A7,$FD
	.dc.b	$17,$D4,$3E,$43,$F6,$33,$9C,$7F,$DD,$7D,$CA,$D4,$7F,$BA,$A7,$59
	.dc.b	$7E,$D6,$B5,$7C,$B0,$EA,$DB,$E1,$1F,$C5,$BE,$A1,$F5,$0F,$A8,$D6
	.dc.b	$3F,$BC,$FA,$87,$37,$F0,$4F,$3F,$F2,$72,$9C,$E5,$FE,$FC,$9C,$F8
	.dc.b	$FF,$EB,$C8,$FF,$7F,$AE,$3F,$E1,$AE,$6F,$F8,$4B,$99,$FF,$54,$70
	.dc.b	$FF,$BD,$FE,$9D,$F2,$5C,$3D,$DF,$E2,$B1,$8A,$EB,$18,$F8,$1F,$E3
	.dc.b	$3F,$50,$F5,$B1,$5D,$0F,$F4,$3C,$65,$C6,$7F,$EB,$7C,$3F,$89,$3F
	.dc.b	$C7,$15,$F8,$3D,$0E,$73,$97,$15,$CA,$1F,$50,$FA,$8C,$C7,$F2,$6F
	.dc.b	$15,$DB,$FD,$1A,$DF,$4C,$F8,$3F,$E6,$35,$A9,$F2,$9A,$8C,$D7,$1A
	.dc.b	$E1,$3E,$1A,$E1,$7F,$CD,$E3,$81,$9A,$E2,$F4,$A7,$D4,$62,$BF,$9D
	.dc.b	$CB,$E1,$FC,$D3,$F3,$E7,$31,$8E,$2F,$E8,$F5,$B3,$CF,$5B,$E3,$47
	.dc.b	$FA,$D8,$D5,$F9,$67,$51,$9B,$E6,$7B,$7E,$82,$C7,$CD,$FA,$87,$C2
	.dc.b	$FC,$13,$FA,$BB,$88,$F6,$46,$0D,$3C,$5B,$87,$F5,$89,$E0,$D8,$F9
	.dc.b	$9F,$07,$C6,$BE,$E7,$37,$F1,$4C,$FD,$77,$EC,$D5,$F4,$DF,$DA,$2B
	.dc.b	$57,$FF,$00,$4F,$BB,$19,$F2,$BE,$C3,$63,$E9,$36,$FF,$B5,$D7,$27
	.dc.b	$B7,$56,$71,$F0,$C3,$C4,$FB,$F1,$E4,$97,$D4,$3E,$1B,$FB,$DD,$6B
	.dc.b	$26,$1F,$E8,$DA,$D4,$7E,$66,$75,$39,$AE,$3C,$9F,$E0,$6B,$DC,$FF
	.dc.b	$80,$2B,$6F,$8E,$7F,$C1,$7D,$FF,$DF,$F0,$FF,$7A,$D6,$87,$D0,$FF
	.dc.b	$89,$B3,$7D,$46,$63,$0D,$AF,$ED,$ED,$C7,$FB,$63,$CF,$6F,$6F,$0F
	.dc.b	$B9,$B5,$EB,$70,$FE,$D6,$BE,$0F,$C0,$FC,$2F,$A8,$7D,$47,$9A,$FF
	.dc.b	$C4,$35,$A1,$3A,$7E,$4E,$9F,$80,$C3,$95,$FF,$AD,$B0,$B5,$AC,$3F
	.dc.b	$0F,$87,$18,$FA,$22,$76,$FD,$B9,$DD,$73,$8F,$0B,$E5,$BD,$C7,$96
	.dc.b	$FA,$61,$E3,$8D,$9F,$EC,$CC,$3A,$C6,$67,$C5,$3C,$6F,$A3,$C0,$F2
	.dc.b	$D7,$0C,$3C,$71,$CC,$D5,$F3,$AF,$35,$F8,$AB,$75,$C8,$DB,$87,$C0
	.dc.b	$7C,$5C,$3D,$6F,$C3,$36,$E3,$0F,$13,$63,$FD,$C6,$7D,$52,$C6,$1C
	.dc.b	$1F,$3B,$EA,$3F,$8D,$7A,$9C,$1B,$1F,$EE,$C6,$BF,$EF,$26,$BC,$7E
	.dc.b	$4B,$FC,$FC,$ED,$AE,$DA,$EF,$7E,$D1,$C3,$2F,$C0,$7D,$43,$E9,$3F
	.dc.b	$03,$18,$6E,$BC,$70,$FA,$58,$B0,$E5,$3F,$88,$C3,$95,$7E,$21,$AF
	.dc.b	$09,$DB,$FB,$EA,$E5,$8E,$F0,$EF,$3F,$C0,$AF,$8E,$3F,$95,$BD,$CF
	.dc.b	$3B,$E0,$C3,$04,$9F,$53,$97,$96,$1E,$28,$57,$49,$DC,$7E,$3E,$FC
	.dc.b	$38,$FC,$19,$E3,$87,$38,$F8,$18,$FA,$2F,$9E,$3D,$70,$F8,$37,$FC
	.dc.b	$91,$BA,$70,$AF,$EE,$99,$78,$71,$C4,$7F,$A6,$2B,$6F,$C3,$C7,$F6
	.dc.b	$B8,$7F,$2E,$BE,$13,$FC,$A3,$EA,$3F,$D7,$CF,$F7,$F6,$BF,$EA,$26
	.dc.b	$CF,$F3,$6F,$A6,$70,$F4,$CF,$F8,$0D,$69,$FE,$07,$2F,$63,$58,$7F
	.dc.b	$36,$8C,$3D,$FA,$F2,$AD,$F3,$38,$6C,$79,$63,$74,$F0,$35,$12,$74
	.dc.b	$7F,$86,$67,$4E,$0F,$F5,$EB,$FE,$15,$75,$E2,$F1,$7F,$D8,$3F,$51
	.dc.b	$FC,$F1,$FF,$11,$35,$CF,$4F,$8B,$CD,$F4,$CE,$77,$FE,$B6,$DD,$BF
	.dc.b	$92,$B6,$FE,$15,$70,$BF,$5F,$D4,$49,$EF,$1F,$13,$E7,$BF,$CD,$51
	.dc.b	$FE,$33,$6B,$FE,$D6,$6E,$3F,$57,$BB,$87,$AF,$2B,$83,$F2,$23,$11
	.dc.b	$F0,$BD,$1F,$96,$62,$BF,$B9,$FC,$0F,$F9,$5A,$7F,$C8,$4D,$7F,$DC
	.dc.b	$AD,$8F,$EF,$3C,$FF,$E6,$C9,$DF,$97,$F7,$2B,$5B,$FF,$46,$6C,$FF
	.dc.b	$78,$AE,$0F,$4A,$FC,$1C,$3E,$1A,$9C,$E3,$F0,$FE,$87,$E1,$2B,$FD
	.dc.b	$EE,$DC,$7F,$C1,$6D,$C7,$FA,$AF,$D4,$7F,$9A,$6B,$FC,$B9,$39,$7F
	.dc.b	$C1,$2D,$4D,$8E,$11,$DB,$97,$F2,$F0,$F8,$FA,$79,$23,$38,$FE,$73
	.dc.b	$EA,$3F,$4F,$7F,$E6,$B6,$AC,$BF,$C4,$38,$7C,$48,$4E,$A7,$FC,$4A
	.dc.b	$DC,$BF,$B7,$B6,$7F,$C4,$35,$C7,$EA,$3F,$D6,$51,$FE,$79,$6B,$DE
	.dc.b	$BF,$BB,$C6,$2F,$AF,$6F,$EC,$A3,$FC,$CD,$5C,$7D,$0C,$1F,$B6,$FA
	.dc.b	$87,$C4,$7E,$E3,$EA,$1F,$0B,$49,$7C,$4F,$F4,$54,$3E,$93,$81,$23
	.dc.b	$3F,$51,$FC,$0B,$E2,$BF,$D2,$B1,$FE,$91,$6B,$E9,$3F,$C7,$EE,$35
	.dc.b	$FF,$24,$FC,$73,$D1,$E5,$EE,$CB,$E4,$FF,$0E,$9D,$46,$0F,$F4,$F2
	.dc.b	$FA,$8D,$1C,$63,$C0,$3E,$A3,$E8,$37,$DA,$FE,$33,$1F,$E7,$A8,$FF
	.dc.b	$54,$3A,$24,$68,$DF,$1F,$0F,$12,$57,$07,$9F,$D0,$62,$3C,$36,$9C
	.dc.b	$69,$FE,$68,$85,$7C,$9F,$95,$97,$F2,$5F,$2F,$1E,$23,$D7,$BA,$71
	.dc.b	$87,$3A,$BC,$FF,$E3,$CC,$5F,$51,$EE,$33,$8F,$F9,$C3,$EA,$1E,$37
	.dc.b	$F3,$A7,$EF,$47,$FB,$25,$AF,$CA,$76,$E6,$38,$72,$C8,$C1,$C7,$EA
	.dc.b	$3A,$11,$E9,$DF,$5E,$91,$74,$FF,$45,$E1,$FB,$BF,$A8,$7C,$BF,$F4
	.dc.b	$E6,$DF,$FB,$75,$AF,$DE,$FA,$7F,$DF,$DF,$FA,$52,$72,$DA,$E4,$7F
	.dc.b	$58,$9D,$ED,$FE,$62,$AD,$BF,$E6,$48,$7D,$46,$CE,$38,$7F,$BB,$63
	.dc.b	$FC,$48,$DF,$63,$A1,$D7,$C0,$B1,$92,$1E,$77,$C5,$D9,$E3,$EC,$93
	.dc.b	$D6,$32,$F7,$8C,$BE,$F8,$FE,$E3,$F5,$0F,$A8,$A4,$9F,$F5,$6E,$DF
	.dc.b	$EB,$6F,$4B,$F4,$35,$B9,$CE,$BF,$E8,$28,$C4,$7E,$A1,$B5,$C7,$B3
	.dc.b	$A4,$FE,$B2,$FF,$C0,$6D,$F5,$F3,$5D,$FB,$7F,$C0,$B5,$F9,$EA,$EE
	.dc.b	$7F,$B0,$1C,$3C,$E3,$67,$0E,$D9,$1C,$04,$FE,$AD,$C3,$FC,$71,$BF
	.dc.b	$51,$E0,$3C,$3D,$92,$32,$43,$D4,$FF,$86,$BE,$A3,$F8,$29,$33,$C7
	.dc.b	$81,$97,$FD,$27,$DF,$FE,$A8,$FA,$5F,$F6,$E6,$1F,$EB,$A8,$C7,$C2
	.dc.b	$EA,$3F,$DB,$B8,$7F,$48,$F8,$1F,$DC,$47,$FB,$82,$F8,$BF,$EB,$1B
	.dc.b	$E2,$F3,$35,$7C,$D5,$F3,$9D,$C7,$13,$67,$11,$5D,$AF,$C3,$1C,$B0
	.dc.b	$E9,$2F,$A7,$A4,$9E,$50,$FA,$8C,$47,$71,$F4,$EF,$A8,$FE,$ED,$8F
	.dc.b	$68,$C7,$C5,$F2,$C7,$D5,$18,$BE,$4D,$AE,$17,$CB,$3F,$9D,$7C,$23
	.dc.b	$16,$3C,$2E,$1B,$87,$77,$E1,$61,$D6,$FF,$DB,$ED,$8E,$4D,$7E,$3D
	.dc.b	$F7,$AE,$37,$D6,$B5,$97,$C7,$BC,$D4,$7D,$75,$E3,$9D,$46,$27,$58
	.dc.b	$FB,$9F,$34,$64,$91,$97,$E7,$4F,$28,$C1,$31,$F5,$DF,$7C,$7D,$17
	.dc.b	$CA,$73,$87,$6B,$E1,$3F,$02,$FD,$87,$17,$CB,$5C,$1E,$47,$E5,$EF
	.dc.b	$AD,$72,$8F,$9E,$79,$A3,$33,$C6,$75,$87,$A6,$30,$4A,$F1,$61,$E5
	.dc.b	$8C,$47,$4A,$E3,$7D,$23,$67,$90,$FA,$E7,$A0,$CC,$70,$70,$73,$9E
	.dc.b	$B7,$E3,$8D,$C7,$83,$1F,$BB,$7D,$E7,$81,$A3,$E1,$1D,$23,$F1,$07
	.dc.b	$07,$18,$FC,$CC,$7D,$91,$C4,$E6,$4C,$3C,$4F,$CA,$3F,$38,$7B,$8A
	.dc.b	$7E,$BB,$C8,$7C,$5E,$23,$47,$E7,$23,$2F,$EB,$1B,$9F,$F0,$0F,$A7
	.dc.b	$FF,$17,$75,$7E,$92,$73,$AF,$1C,$30,$7F,$09,$AD,$E1,$CE,$32,$D3
	.dc.b	$F8,$14,$6B,$0F,$E0,$17,$D2,$7F,$5A,$D7,$D5,$5D,$7B,$18,$F4,$BE
	.dc.b	$33,$71,$F8,$78,$F5,$7C,$D8,$78,$6B,$6F,$8B,$0E,$6E,$67,$73,$A3
	.dc.b	$E1,$4E,$9C,$46,$89,$3F,$C3,$6B,$C0,$FA,$AF,$C8,$D3,$C1,$86,$1F
	.dc.b	$69,$E1,$39,$61,$DE,$FA,$C6,$B0,$F0,$BA,$AE,$5A,$B2,$F9,$0D,$C7
	.dc.b	$E4,$4E,$A7,$27,$CD,$CB,$CF,$7D,$71,$EF,$87,$18,$F3,$9E,$51,$3C
	.dc.b	$23,$58,$FC,$6C,$3D,$93,$9C,$3B,$65,$EB,$C3,$E1,$5F,$E3,$7D,$5C
	.dc.b	$47,$4F,$A8,$B8,$7F,$AB,$F1,$FC,$64,$62,$A4,$7E,$3F,$D7,$B7,$F8
	.dc.b	$C6,$E1,$F9,$96,$E9,$F8,$46,$CF,$D7,$6F,$C8,$BE,$2B,$FD,$3B,$87
	.dc.b	$F8,$42,$31,$F0,$DF,$86,$EF,$FC,$3C,$6F,$D5,$C5,$7F,$4D,$8D,$BC
	.dc.b	$6F,$9E,$3F,$E9,$B9,$CB,$63,$37,$D1,$61,$D2,$55,$E5,$FA,$8C,$3E
	.dc.b	$7F,$A8,$B6,$9B,$18,$6E,$3E,$36,$DF,$FA,$7B,$E9,$FF,$DE,$13,$FE
	.dc.b	$2E,$E5,$7D,$47,$FA,$1E,$3F,$C5,$53,$AA,$FF,$5A,$56,$B7,$F3,$42
	.dc.b	$75,$AF,$EB,$2B,$57,$CB,$DF,$F6,$46,$2F,$C5,$1C,$AF,$8D,$7F,$77
	.dc.b	$F9,$07,$C1,$7F,$BD,$A3,$91,$F3,$7D,$1F,$83,$7F,$09,$B1,$93,$E9
	.dc.b	$46,$5A,$7D,$5F,$6B,$E4,$9F,$2B,$EA,$2C,$64,$99,$7D,$78,$CB,$FE
	.dc.b	$C6,$FA,$7F,$F7,$2C,$FF,$95,$B5,$7D,$47,$DC,$8F,$E6,$B3,$AA,$FF
	.dc.b	$6E,$56,$B9,$BE,$4C,$E6,$3F,$1D,$7C,$6B,$FC,$31,$39,$BF,$F3,$AC
	.dc.b	$E6,$53,$97,$7E,$47,$E2,$76,$F9,$3E,$A7,$68,$FC,$3F,$D4,$4C,$7F
	.dc.b	$2D,$E0,$5C,$3F,$29,$37,$3F,$CA,$4D,$AF,$CF,$74,$7F,$32,$7F,$3D
	.dc.b	$F3,$FF,$92,$7F,$26,$D7,$27,$E8,$CC,$9F,$E7,$5D,$1E,$87,$E1,$3E
	.dc.b	$A3,$F4,$33,$98,$C7,$6F,$C6,$6C,$64,$FC,$C4,$F0,$D3,$EA,$4E,$E7
	.dc.b	$3C,$36,$7F,$64,$3E,$A3,$F5,$6D,$9F,$E0,$0B,$E2,$B3,$32,$72,$F0
	.dc.b	$8C,$4F,$4F,$A8,$E7,$F1,$9D,$1B,$7D,$1B,$3B,$8E,$F3,$FA,$AF,$A7
	.dc.b	$7D,$3F,$D9,$FA,$87,$C8,$66,$30,$DB,$FF,$52,$F4,$2B,$71,$FE,$A1
	.dc.b	$BE,$6E,$A3,$38,$B9,$7F,$B0,$AA,$DB,$9A,$FF,$06,$C6,$27,$FB,$3F
	.dc.b	$D4,$3E,$A1,$F5,$1E,$3C,$38,$1F,$E6,$FF,$A8,$73,$FF,$06,$F3,$FF
	.dc.b	$31,$18,$CF,$FD,$6D,$1F,$9B,$F1,$5E,$EF,$F0,$C7,$E9,$4F,$FA,$ED
	.dc.b	$CB,$79,$3F,$88,$FD,$3B,$E5,$38,$5E,$1B,$5F,$EC,$39,$7C,$17,$F1
	.dc.b	$8F,$A8,$7A,$FC,$A3,$0F,$4F,$47,$17,$C5,$BB,$FB,$86,$1F,$C5,$5C
	.dc.b	$37,$8F,$E6,$FD,$43,$EA,$3A,$61,$FE,$D5,$F4,$DD,$9F,$CF,$E1,$F4
	.dc.b	$DF,$ED,$87,$51,$98,$C6,$5E,$1A,$E3,$5F,$EE,$63,$8F,$C1,$FF,$34
	.dc.b	$8C,$CF,$D4,$9C,$9D,$08,$E6,$9F,$51,$C7,$0F,$E7,$B4,$F8,$8F,$34
	.dc.b	$FC,$F8,$FF,$77,$19,$3C,$4E,$6F,$E3,$46,$5F,$14,$67,$4F,$C8,$CE
	.dc.b	$5F,$F1,$9B,$E2,$E2,$F0,$DC,$8F,$76,$2F,$9F,$C6,$3F,$4F,$EA,$1F
	.dc.b	$0D,$F1,$CF,$EA,$AE,$2F,$8C,$60,$D3,$FE,$C5,$CB,$F4,$73,$BA,$F8
	.dc.b	$95,$BA,$D1,$FE,$70,$9D,$57,$A3,$07,$C2,$6F,$0F,$C3,$72,$7F,$15
	.dc.b	$AD,$7D,$37,$F6,$88,$C6,$3F,$ED,$53,$F0,$31,$9F,$2F,$F0,$0D,$8F
	.dc.b	$82,$DB,$EB,$3F,$85,$EB,$C6,$1F,$BC,$F1,$3F,$02,$FF,$73,$FA,$87
	.dc.b	$C4,$7F,$78,$9C,$E5,$FE,$E5,$B5,$6A,$3D,$F8,$FF,$73,$E4,$FE,$F3
	.dc.b	$39,$8F,$D1,$5F,$27,$FD,$1F,$19,$C7,$C5,$5D,$CF,$F0,$0F,$7F,$F7
	.dc.b	$EA,$E9,$87,$65,$83,$E7,$FF,$C4,$58,$3E,$A3,$B5,$6A,$73,$18,$AE
	.dc.b	$17,$FD,$CE,$31,$3A,$C7,$E0,$6D,$F0,$78,$2D,$F8,$70,$FB,$2D,$9C
	.dc.b	$E3,$FB,$F0,$3E,$06,$FD,$43,$EA,$3F,$D6,$78,$2F,$FC,$4B,$7C,$2F
	.dc.b	$75,$A9,$FF,$48,$DF,$1A,$FF,$5C,$D6,$9F,$F1,$14,$EA,$BF,$D7,$53
	.dc.b	$99,$F8,$F3,$97,$88,$AE,$8E,$B0,$E0,$F8,$CD,$5F,$8E,$33,$1E,$D3
	.dc.b	$27,$89,$CD,$F8,$70,$F7,$46,$27,$67,$72,$9D,$EF,$C7,$3C,$23,$27
	.dc.b	$78,$CE,$1E,$B8,$FF,$6E,$DF,$78,$C4,$73,$C3,$D0,$F2,$9D,$61,$E3
	.dc.b	$BE,$6F,$96,$7F,$DC,$47,$B4,$B3,$87,$E2,$1F,$EE,$AF,$A7,$FE,$41
	.dc.b	$EA,$F0,$6C,$7F,$BA,$DA,$EF,$AB,$FD,$14,$E7,$E4,$BF,$B7,$37,$87
	.dc.b	$94,$62,$3B,$35,$FF,$7D,$D2,$38,$E5,$F6,$BE,$A1,$F4,$9F,$F0,$02
	.dc.b	$30,$DD,$78,$D7,$06,$E3,$F8,$7B,$E3,$8D,$BE,$0F,$E7,$31,$F0,$57
	.dc.b	$E3,$5B,$1F,$72,$B8,$B6,$B8,$A9,$DE,$3F,$90,$6E,$1F,$A8,$95,$F2
	.dc.b	$AF,$4C,$77,$C7,$D0,$F8,$DA,$F0,$C3,$BD,$71,$BF,$6C,$FE,$43,$0E
	.dc.b	$EF,$ED,$A7,$4D,$AF,$3C,$EA,$3D,$17,$DA,$B7,$1D,$CF,$26,$3F,$1D
	.dc.b	$FE,$2D,$1F,$0A,$3F,$C5,$2D,$D7,$C7,$97,$B2,$BF,$C0,$D3,$97,$C4
	.dc.b	$7F,$1F,$C3,$C5,$F5,$1F,$EC,$08,$FE,$44,$D7,$FC,$1C,$DC,$3F,$C6
	.dc.b	$FF,$51,$FD,$3A,$77,$D5,$F3,$67,$3E,$C5,$CF,$E8,$5F,$F9,$BA,$F6
	.dc.b	$76,$F9,$BD,$3F,$C7,$31,$46,$21,$18,$7F,$9B,$4F,$F7,$17,$91,$0F
	.dc.b	$3C,$F1,$8D,$46,$0E,$3E,$2E,$4E,$D5,$D6,$3F,$C5,$9F,$51,$FD,$05
	.dc.b	$FE,$88,$D7,$3D,$3D,$7C,$DF,$34,$E7,$4F,$F2,$DB,$73,$FC,$83,$6F
	.dc.b	$F9,$CE,$5F,$E4,$E7,$AE,$1E,$7F,$A8,$91,$83,$FA,$A7,$CF,$FF,$9C
	.dc.b	$9B,$EC,$7C,$58,$C4,$7F,$90,$EF,$FC,$76,$E8,$87,$F9,$1E,$37,$97
	.dc.b	$DA,$ED,$E5,$7F,$1F,$D1,$97,$C2,$BF,$CE,$BF,$11,$FE,$5D,$8F,$EC
	.dc.b	$CD,$7F,$CA,$4D,$8F,$F3,$DF,$47,$F9,$C2,$77,$E5,$7D,$BA,$D6,$FF
	.dc.b	$B3,$6C,$FF,$5C,$AE,$0F,$4A,$FC,$14,$3E,$1E,$9C,$E3,$F0,$DE,$8F
	.dc.b	$D9,$C7,$FD,$16,$DC,$7F,$80,$37,$0F,$ED,$3F,$4F,$FE,$70,$7F,$C0
	.dc.b	$11,$94,$D3,$11,$B8,$FF,$38,$46,$3E,$3F,$F2,$D2,$7C,$7C,$F8,$C7
	.dc.b	$F3,$5F,$51,$FA,$7C,$3F,$CF,$0D,$CB,$FD,$43,$87,$C2,$94,$E7,$2F
	.dc.b	$E3,$0D,$8F,$F0,$0B,$54,$F1,$FA,$8F,$F5,$CC,$7F,$86,$DA,$FF,$A1
	.dc.b	$DB,$7F,$EB,$0A,$E5,$1F,$E6,$D9,$E1,$E0,$FE,$C6,$31,$1F,$E8,$FF
	.dc.b	$A8,$7C,$47,$ED,$7E,$A1,$F0,$B4,$95,$C8,$FF,$16,$43,$E9,$38,$12
	.dc.b	$33,$F5,$1F,$BF,$F8,$AF,$F5,$2C,$7F,$8C,$DA,$FF,$A7,$DB,$AF,$FB
	.dc.b	$23,$E3,$DE,$87,$1F,$7A,$5F,$17,$F8,$54,$EA,$30,$7F,$8F,$D7,$D4
	.dc.b	$68,$E3,$1F,$EA,$DF,$A8,$F9,$C7,$F9,$F1,$FF,$23,$35,$D7,$A5,$FC
	.dc.b	$5E,$37,$ED,$53,$11,$9F,$1A,$57,$27,$87,$D0,$E2,$3C,$11,$74,$FF
	.dc.b	$72,$E0,$AF,$8D,$F2,$B2,$FE,$33,$E1,$E3,$C4,$7C,$78,$B7,$CA,$6F
	.dc.b	$4F,$F6,$3C,$1F,$51,$E9,$33,$87,$F9,$3B,$E9,$DE,$37,$F3,$37,$ED
	.dc.b	$47,$F9,$9D,$AF,$FB,$4D,$DB,$98,$E1,$CB,$23,$07,$1F,$A8,$E8,$47
	.dc.b	$A7,$7D,$7A,$45,$D1,$9F,$F7,$BC,$7F,$71,$F5,$0F,$98,$FE,$90,$DC
	.dc.b	$3F,$CF,$8D,$7F,$DD,$1F,$51,$FB,$BC,$39,$C7,$F9,$B2,$75,$18,$84
	.dc.b	$63,$0F,$F0,$53,$63,$FC,$E7,$5C,$E7,$37,$FD,$66,$1F,$2F,$4F,$2C
	.dc.b	$67,$D7,$97,$FE,$91,$AF,$F1,$83,$7C,$1C,$BE,$89,$B9,$79,$6B,$F6
	.dc.b	$AD,$EB,$F0,$75,$CB,$E9,$1E,$33,$0F,$C2,$8C,$3F,$EA,$1F,$A7,$7D
	.dc.b	$47,$F7,$16,$C6,$63,$FB,$A7,$A9,$FC,$46,$73,$2C,$3E,$44,$61,$B5
	.dc.b	$C3,$AF,$A3,$CB,$D5,$97,$D4,$FF,$59,$61,$FD,$95,$E9,$C1,$FE,$39
	.dc.b	$8F,$D1,$46,$BA,$FC,$7D,$92,$30,$43,$89,$FE,$A8,$9E,$99,$7D,$D7
	.dc.b	$F8,$DC,$78,$3E,$A3,$DA,$7B,$BB,$24,$64,$87,$41,$3C,$9C,$7D,$47
	.dc.b	$F3,$52,$6B,$FE,$1B,$7F,$84,$61,$F8,$D7,$FA,$27,$BC,$EF,$FF,$2B
	.dc.b	$7D,$2F,$FB,$9B,$0F,$F5,$C4,$63,$E1,$7F,$D2,$6F,$F9,$6E,$F8,$B9
	.dc.b	$9D,$9F,$EC,$6A,$E4,$FF,$B1,$E3,$4D,$3F,$D9,$73,$C9,$E1,$5D,$9A
	.dc.b	$74,$15,$D2,$FD,$11,$CB,$1E,$0E,$BE,$9E,$D1,$A6,$3E,$A2,$CF,$5F
	.dc.b	$A7,$7D,$47,$F7,$5C,$7A,$CA,$73,$F1,$5E,$3C,$7A,$4F,$1B,$E1,$AF
	.dc.b	$F9,$FE,$F7,$3E,$F3,$94,$E9,$F8,$D9,$7A,$27,$78,$F8,$1B,$7C,$EF
	.dc.b	$85,$FB,$4F,$06,$31,$ED,$5C,$1E,$A6,$6F,$A5,$72,$9D,$4F,$33,$8C
	.dc.b	$64,$95,$EF,$C3,$CD,$18,$9D,$CE,$5E,$D8,$79,$E2,$C7,$88,$D9,$30
	.dc.b	$F1,$C7,$18,$C4,$F1,$69,$C6,$76,$E0,$F9,$73,$C6,$32,$49,$E4,$FD
	.dc.b	$AC,$7E,$BD,$71,$C7,$E0,$9F,$0A,$33,$3C,$64,$76,$3C,$11,$AA,$EA
	.dc.b	$48,$DD,$73,$C7,$E3,$E1,$F0,$A7,$96,$1E,$39,$CC,$78,$F1,$FB,$B8
	.dc.b	$F5,$8E,$71,$F3,$EB,$C9,$1F,$5D,$B8,$7A,$A7,$8D,$F4,$9C,$C6,$2B
	.dc.b	$C5,$5B,$C3,$E4,$63,$E9,$9F,$2B,$71,$F1,$46,$0F,$1C,$FA,$E3,$35
	.dc.b	$CE,$7E,$2E,$3E,$1B,$E9,$3F,$0E,$7F,$2C,$7F,$B8,$CF,$98,$55,$E5
	.dc.b	$3E,$2B,$53,$E0,$2C,$EA,$3F,$71,$3B,$9F,$D7,$7D,$47,$F8,$D3,$9F
	.dc.b	$35,$27,$F4,$B5,$BD,$7F,$7B,$7B,$9C,$D6,$E7,$F6,$55,$BC,$7F,$DF
	.dc.b	$57,$B9,$CD,$6E,$7F,$80,$56,$EF,$F1,$37,$B9,$CD,$6E,$7F,$0F,$5B
	.dc.b	$55,$B7,$33,$B7,$F8,$94,$EE,$75,$18,$9D,$63,$F1,$63,$E1,$F6,$7F
	.dc.b	$83,$B1,$F3,$C7,$B8,$8F,$C2,$8D,$63,$E6,$9C,$B6,$75,$97,$6C,$BD
	.dc.b	$18,$77,$9C,$DF,$03,$D2,$53,$DB,$87,$67,$31,$E4,$26,$1F,$3A,$75
	.dc.b	$3E,$D3,$95,$77,$9D,$61,$F1,$72,$F8,$73,$9A,$FA,$37,$CA,$31,$1E
	.dc.b	$2A,$E1,$39,$C7,$E6,$7A,$BE,$0C,$1F,$51,$63,$0D,$BF,$EA,$93,$FE
	.dc.b	$B1,$6F,$AF,$9C,$7F,$26,$DD,$3F,$1E,$DD,$3F,$16,$D9,$FC,$3B,$7E
	.dc.b	$45,$F1,$3F,$E9,$EC,$3F,$89,$F4,$7F,$92,$27,$3F,$0B,$FE,$4A,$8C
	.dc.b	$56,$F0,$FF,$08,$D6,$AF,$89,$FC,$71,$EB,$EA,$F1,$8C,$53,$0C,$5A
	.dc.b	$F7,$5C,$A3,$15,$B8,$C4,$FA,$3E,$9F,$8B,$EA,$FA,$8B,$49,$B1,$86
	.dc.b	$D7,$29,$D3,$67,$C7,$6F,$A7,$FF,$7A,$4F,$8F,$95,$F5,$1F,$E1,$B8
	.dc.b	$F3,$56,$FD,$FF,$5D,$6B,$83,$9E,$1C,$AB,$FC,$6F,$39,$BE,$33,$9D
	.dc.b	$FF,$CE,$F3,$9A,$FF,$19,$4E,$6B,$FA,$3C,$77,$7E,$37,$C7,$BE,$0B
	.dc.b	$FD,$FB,$1D,$32,$F8,$7E,$77,$C4,$6C,$7C,$16,$DF,$DA,$DB,$B3,$F4
	.dc.b	$2B,$85,$7C,$7F,$6B,$E6,$9F,$3B,$EA,$2C,$64,$95,$CE,$32,$FF,$92
	.dc.b	$FE,$9B,$FD,$DB,$3F,$77,$67,$D4,$7E,$2D,$FF,$4A,$4E,$AB,$FC,$BB
	.dc.b	$1A,$E6,$FA,$B3,$98,$FC,$A5,$71,$BF,$8B,$39,$7F,$37,$19,$BF,$ED
	.dc.b	$F3,$96,$F9,$1F,$92,$DF,$E5,$7A,$7D,$A3,$F1,$7F,$51,$31,$FC,$47
	.dc.b	$7D,$C3,$E4,$4D,$CF,$F4,$B3,$6B,$F3,$E7,$E5,$3E,$7F,$F2,$AD,$3F
	.dc.b	$24,$5B,$FA,$4D,$AF,$87,$E7,$FE,$33,$EA,$3F,$45,$5F,$2E,$31,$DB
	.dc.b	$F1,$1B,$7F,$94,$D9,$18,$AE,$13,$98,$E9,$C1,$C6,$3F,$5B,$2F,$A8
	.dc.b	$FD,$6B,$63,$3F,$15,$89,$91,$C1,$DC,$62,$7F,$82,$8F,$A8,$E7,$F1
	.dc.b	$BD,$1B,$7E,$06,$CF,$06,$EB,$FE,$D7,$FA,$67,$D4,$7D,$7F,$A8,$7C
	.dc.b	$87,$EE,$27,$38,$FE,$92,$72,$FE,$DA,$33,$8F,$FA,$A3,$9D,$5A,$8F
	.dc.b	$F5,$04,$EB,$1F,$F7,$4D,$6B,$0E,$D8,$CC,$7F,$8B,$7D,$43,$EA,$1F
	.dc.b	$51,$9C,$FF,$D5,$5F,$50,$E7,$FE,$1B,$E7,$FE,$4E,$53,$9C,$7F,$85
	.dc.b	$CE,$7C,$57,$B7,$FC,$51,$72,$58,$C3,$FC,$63,$E9,$DF,$29,$9C,$3F
	.dc.b	$8A,$C6,$1F,$E2,$F7,$D1,$CF,$C0,$FF,$20,$FA,$87,$B1,$CE,$30,$F5
	.dc.b	$F4,$71,$5D,$1B,$B7,$FB,$1F,$1F,$E2,$2D,$8F,$E5,$7F,$50,$FA,$8E
	.dc.b	$57,$FC,$A3,$D3,$76,$FF,$42,$87,$D2,$FF,$B4,$1D,$46,$63,$18,$FF
	.dc.b	$B5,$6B,$8D,$7F,$47,$38,$FC,$0F,$F3,$A8,$CC,$FF,$B9,$A7,$27,$42
	.dc.b	$39,$A7,$D4,$71,$C3,$FA,$1D,$3E,$27,$CD,$3F,$2E,$FF,$DC,$D3,$93
	.dc.b	$C5,$19,$EE,$47,$F3,$83,$53,$E3,$9D,$46,$6B,$AF,$1F,$81,$91,$ED
	.dc.b	$C5,$F5,$18,$C3,$FD,$ED,$F5,$0F,$94,$F8,$C7,$F5,$A7,$17,$C6,$30
	.dc.b	$69,$FE,$B3,$1A,$9F,$85,$3A,$9C,$BF,$47,$E1,$31,$2D,$3D,$AF,$FA
	.dc.b	$6B,$93,$F8,$AC,$62,$FE,$DF,$D3,$7F,$6C,$9C,$C7,$FA,$80,$C9,$FD
	.dc.b	$A9,$CD,$7C,$8C,$D5,$C7,$CB,$FC,$13,$6F,$84,$7E,$17,$A7,$17,$E5
	.dc.b	$9F,$EC,$FE,$27,$E0,$B2,$FE,$DB,$F5,$0F,$87,$FE,$FF,$39,$AF,$F5
	.dc.b	$9E,$0A,$D4,$7F,$7E,$9D,$61,$FE,$89,$9C,$BD,$CF,$F5,$DC,$6A,$B9
	.dc.b	$5F,$6E,$49,$87,$F8,$2F,$BF,$FB,$7E,$0C,$3A,$B7,$E8,$7F,$C4,$B9
	.dc.b	$BE,$A3,$34,$9C,$C6,$1B,$8F,$B1,$B8,$FA,$76,$F7,$F0,$FC,$06,$C6
	.dc.b	$49,$97,$AC,$FF,$07,$FC,$0F,$FB,$5B,$E9,$DF,$51,$F3,$2F,$FC,$4F
	.dc.b	$5F,$ED,$AA,$D4,$E6,$EC,$FF,$AC,$E6,$BE,$FA,$FC,$4D,$F4,$8D,$07
	.dc.b	$0E,$EB,$8B,$DC,$D5,$F7,$69,$CA,$78,$E9,$F8,$AA,$E3,$2C,$3C,$73
	.dc.b	$98,$E2,$7C,$02,$B9,$AF,$1C,$78,$5E,$B8,$7A,$B0,$FF,$65,$DF,$48
	.dc.b	$C5,$78,$E7,$CD,$5C,$A3,$C4,$76,$9D,$BE,$18,$EB,$87,$A2,$FA,$36
	.dc.b	$78,$35,$FF,$6D,$9D,$0B,$18,$7E,$41,$FE,$E5,$FA,$7F,$E3,$9E,$EF
	.dc.b	$06,$C7,$FB,$91,$AE,$FA,$BF,$D0,$EE,$7E,$4F,$F8,$13,$78,$79,$46
	.dc.b	$23,$B3,$5F,$F2,$CD,$23,$8E,$5F,$7B,$EA,$1F,$49,$FE,$5E,$8C,$37
	.dc.b	$6E,$35,$C1,$B8,$FE,$32,$F8,$E3,$6F,$82,$18,$F8,$6B,$F2,$6D,$8F
	.dc.b	$C0,$D7,$26,$D7,$27,$8E,$1E,$17,$A5,$71,$C7,$CF,$1E,$13,$AC,$ED
	.dc.b	$E3,$86,$5E,$A6,$98,$E0,$93,$FA,$2A,$ED,$8F,$C0,$9C,$BF,$E8,$4C
	.dc.b	$7D,$15,$BC,$3C,$11,$EB,$C3,$F9,$BD,$A3,$1A,$F1,$CB,$FD,$1B,$5A
	.dc.b	$CB,$BD,$FC,$9B,$DB,$6B,$77,$C6,$3F,$D4,$75,$C1,$7D,$47,$FA,$FD
	.dc.b	$FF,$53,$B5,$FE,$1C,$DC,$3C,$BF,$4C,$E2,$EB,$5F,$66,$B5,$87,$FA
	.dc.b	$8A,$BF,$CB,$B3,$FE,$B2,$6B,$D3,$AB,$8B,$E2,$F9,$BB,$66,$9A,$AF
	.dc.b	$ED,$B3,$E6,$8C,$B4,$ED,$E0,$61,$FE,$4D,$2F,$13,$91,$9C,$FE,$7F
	.dc.b	$D4,$7F,$37,$AF,$F6,$3B,$5F,$E5,$CD,$8F,$83,$E6,$FA,$23,$1A,$7F
	.dc.b	$B0,$1B,$5F,$9B,$6A,$C3,$FC,$8E,$F6,$E9,$F6,$7D,$44,$8C,$1F,$ED
	.dc.b	$8F,$9F,$FF,$34,$B7,$D6,$F8,$F1,$88,$F6,$F6,$E2,$64,$6C,$FA,$11
	.dc.b	$87,$E1,$FA,$1F,$2C,$C6,$BF,$90,$F8,$8F,$F2,$AC,$FF,$BA,$1A,$FF
	.dc.b	$4F,$6C,$7E,$5B,$9B,$FC,$D1,$3B,$F2,$7E,$DD,$6B,$7F,$F7,$1B,$67
	.dc.b	$F7,$D5,$C1,$E9,$5F,$82,$87,$C4,$53,$9C,$7E,$1B,$CF,$FC,$25,$7F
	.dc.b	$BF,$1A,$FF,$63,$6E,$3F,$AA,$FA,$8F,$F3,$3C,$EF,$9B,$91,$1F,$E2
	.dc.b	$13,$BF,$2B,$11,$FD,$B6,$9F,$23,$F9,$66,$C7,$33,$8E,$BF,$F0,$02
	.dc.b	$3D,$D0,$FA,$8F,$D3,$3A,$3F,$8F,$38,$7B,$67,$FA,$28,$C6,$3F,$C0
	.dc.b	$9B,$7F,$C9,$6B,$A7,$67,$8E,$FF,$53,$F5,$1F,$EB,$09,$EC,$1C,$3F
	.dc.b	$E0,$76,$E1,$D6,$B8,$F8,$72,$B8,$D7,$F9,$AD,$E7,$F5,$0F,$88,$FD
	.dc.b	$BF,$D4,$3E,$12,$92,$B8,$E1,$AF,$A4,$91,$E4,$A4,$72,$FA,$8F,$E0
	.dc.b	$3F,$14,$F7,$29,$2F,$D7,$F1,$EF,$4B,$11,$FE,$21,$AE,$0F,$C0,$6B
	.dc.b	$FE,$25,$9E,$0E,$6B,$F8,$7C,$3E,$A2,$9D,$38,$7F,$98,$FD,$47,$D0
	.dc.b	$BF,$EA,$8F,$F9,$E5,$F9,$AD,$7F,$C6,$4E,$D3,$4D,$47,$4F,$67,$F8
	.dc.b	$FC,$67,$C7,$A4,$AE,$BF,$43,$5E,$B1,$6F,$CB,$3F,$23,$E5,$65,$7C
	.dc.b	$17,$AF,$1E,$23,$E3,$45,$AF,$6C,$DE,$8F,$C6,$62,$FA,$8F,$29,$9C
	.dc.b	$3F,$BA,$7D,$3B,$C7,$FE,$6E,$FD,$A8,$FC,$7B,$5F,$F2,$A3,$B7,$31
	.dc.b	$C3,$92,$46,$0E,$3F,$51,$D0,$8F,$4F,$06,$BD,$22,$E9,$FD,$9B,$0F
	.dc.b	$DF,$E7,$FA,$5F,$A8,$7C,$C7,$F4,$C6,$E1,$FA,$46,$BF,$E6,$9F,$A8
	.dc.b	$FE,$0A,$FF,$9A,$E3,$38,$7F,$72,$6D,$F5,$BF,$EA,$F7,$C2,$B5,$18
	.dc.b	$F9,$79,$87,$78,$DF,$AF,$23,$F6,$B7,$FE,$2D,$6F,$6E,$5F,$24,$DC
	.dc.b	$3F,$C0,$75,$FB,$39,$F4,$75,$F8,$BA,$A4,$74,$8F,$29,$87,$DD,$18
	.dc.b	$7F,$82,$FD,$43,$EA,$3F,$B8,$B6,$33,$1F,$DD,$3D,$4F,$CE,$4E,$67
	.dc.b	$FD,$20,$DC,$7F,$C2,$ED,$DB,$F8,$B5,$70,$BE,$5F,$2D,$30,$EA,$7F
	.dc.b	$16,$9F,$EC,$6F,$FA,$6D,$75,$78,$4F,$17,$5C,$9E,$11,$97,$94,$79
	.dc.b	$FA,$BD,$1F,$51,$27,$B4,$65,$F0,$46,$5F,$04,$7F,$84,$BE,$A3,$F7
	.dc.b	$44,$F1,$F1,$DF,$97,$FD,$5F,$DF,$FC,$CF,$E9,$BF,$DA,$D8,$7F,$AB
	.dc.b	$63,$1F,$0B,$A8,$FF,$2E,$61,$FC,$FF,$E0,$95,$A9,$E0,$7F,$98,$6F
	.dc.b	$9B,$FE,$A1,$8D,$34,$E7,$87,$55,$3C,$2B,$AC,$73,$C3,$53,$9B,$FE
	.dc.b	$95,$87,$8D,$F5,$7D,$36,$89,$7E,$43,$AF,$D4,$61,$F1,$0F,$A7,$7D
	.dc.b	$47,$F6,$EC,$FC,$33,$98,$C6,$1D,$A3,$A5,$70,$F4,$ED,$6A,$7F,$D0
	.dc.b	$53,$97,$BE,$6C,$BE,$04,$E7,$1E,$ED,$C7,$FB,$33,$6F,$E6,$57,$0C
	.dc.b	$3B,$47,$97,$1F,$09,$EF,$C7,$B4,$60,$F7,$DF,$37,$81,$E4,$BE,$95
	.dc.b	$BC,$7C,$78,$76,$C3,$9D,$F2,$8F,$0E,$1C,$CE,$93,$AA,$E3,$7C,$F1
	.dc.b	$F5,$DF,$3C,$7B,$E3,$CE,$B4,$FB,$8A,$79,$AF,$9B,$E7,$33,$8F,$86
	.dc.b	$FA,$46,$1F,$29,$87,$B1,$27,$99,$F7,$B1,$F4,$63,$E6,$C3,$91,$F2
	.dc.b	$F1,$F3,$D7,$09,$D5,$F8,$CE,$8E,$23,$51,$EB,$32,$7E,$13,$0F,$2D
	.dc.b	$F5,$8E,$E6,$8F,$D1,$DF,$82,$FA,$12,$3A,$5F,$67,$E3,$46,$5E,$27
	.dc.b	$48,$D6,$1E,$E8,$F1,$5E,$6B,$F3,$91,$E1,$C7,$E6,$4F,$96,$31,$8F
	.dc.b	$A1,$F8,$18,$78,$6B,$71,$8B,$F1,$46,$27,$E7,$63,$F8,$0A,$FF,$21
	.dc.b	$1F,$92,$2C,$BC,$A7,$C5,$6A,$7C,$E5,$9D,$47,$EC,$A7,$73,$FB,$CF
	.dc.b	$A7,$FF,$17,$74,$E6,$A4,$FE,$8A,$B7,$8F,$EF,$AF,$73,$9A,$DC,$FE
	.dc.b	$D2,$B7,$97,$F9,$66,$F7,$39,$AD,$CF,$EE,$2B,$76,$17,$C2,$75,$5C
	.dc.b	$27,$F8,$3D,$70,$77,$DA,$C7,$51,$FC,$5E,$75,$97,$C2,$8F,$81,$5B
	.dc.b	$7D,$44,$CB,$CC,$EA,$F9,$D7,$19,$C9,$F0,$EB,$93,$8C,$7C,$F7,$CD
	.dc.b	$F8,$53,$A7,$33,$B9,$F5,$E3,$FC,$C7,$2F,$81,$18,$39,$16,$7C,$18
	.dc.b	$F9,$32,$F6,$9F,$0E,$B8,$D7,$C1,$BE,$98,$F8,$63,$96,$3E,$4C,$3E
	.dc.b	$3D,$F0,$CB,$DD,$39,$AF,$09,$F5,$3D,$5E,$4F,$D6,$FA,$8B,$7F,$EA
	.dc.b	$8C,$3D,$51,$8A,$91,$ED,$F5,$ED,$FE,$45,$BA,$FE,$35,$B8,$FE,$25
	.dc.b	$B3,$F8,$5C,$51,$8F,$91,$7C,$57,$FA,$57,$1F,$F5,$84,$63,$E1,$BE
	.dc.b	$F7,$85,$F7,$E3,$7E,$AE,$2B,$F2,$11,$B7,$8D,$F3,$C9,$7C,$61,$87
	.dc.b	$19,$5F,$4B,$E3,$39,$8F,$2F,$D4,$5B,$7D,$45,$8C,$36,$FC,$2D,$C3
	.dc.b	$C3,$1E,$6A,$E2,$FF,$B0,$3E,$9F,$FD,$D2,$FF,$B3,$F9,$1F,$51,$FE
	.dc.b	$10,$8D,$7B,$FE,$79,$CF,$0F,$A7,$16,$1C,$9E,$E7,$F0,$CC,$39,$38
	.dc.b	$37,$7C,$E3,$3B,$7E,$AE,$73,$8F,$80,$FF,$71,$7C,$83,$E0,$BE,$CC
	.dc.b	$66,$7B,$EF,$F6,$BC,$FF,$8D,$97,$C7,$6D,$7D,$46,$D7,$0D,$BF,$D0
	.dc.b	$33,$F6,$2B,$E7,$7B,$52,$3E,$DF,$D4,$5B,$E1,$37,$4F,$B9,$36,$BF
	.dc.b	$DE,$1F,$4B,$FE,$E1,$7F,$DF,$52,$FA,$8F,$C2,$3E,$F9,$D6,$1F,$22
	.dc.b	$35,$CD,$F3,$67,$31,$F9,$2A,$E3,$7D,$BC,$8F,$C5,$4F,$08,$FF,$7D
	.dc.b	$D7,$07,$47,$CA,$DD,$EA,$75,$BF,$C6,$7D,$44,$AF,$CC,$F8,$37,$4F
	.dc.b	$8D,$37,$2F,$CD,$4D,$AF,$C6,$1F,$95,$F9,$FF,$CB,$34,$FD,$19,$6F
	.dc.b	$E8,$36,$BE,$17,$A1,$F8,$DF,$A8,$FD,$2C,$FA,$A3,$1D,$BF,$0D,$B1
	.dc.b	$93,$F3,$13,$C2,$FF,$1D,$B2,$B8,$56,$B8,$2D,$7E,$20,$7D,$47,$EB
	.dc.b	$DB,$3F,$C0,$97,$C5,$66,$64,$E5,$E1,$18,$9E,$9F,$51,$CF,$E3,$3A
	.dc.b	$36,$FA,$36,$77,$1D,$E7,$F5,$9F,$4E,$FA,$7F,$E0,$DD,$AF,$A8,$7C
	.dc.b	$86,$A7,$31,$8C,$3F,$B1,$71,$7E,$AF,$0E,$51,$FA,$A9,$D6,$5F,$B3
	.dc.b	$9C,$BF,$B8,$C7,$B6,$3D,$6F,$99,$FD,$B6,$55,$C3,$0F,$EC,$5F,$50
	.dc.b	$FA,$87,$D4,$71,$BF,$E0,$1F,$50,$E7,$FE,$11,$E8,$7E,$6A,$7F,$BC
	.dc.b	$4E,$73,$EF,$1F,$DE,$E7,$4F,$F8,$5B,$DB,$FE,$21,$E4,$FD,$4A,$DC
	.dc.b	$E5,$B1,$98,$FE,$21,$F4,$EF,$92,$E3,$7F,$C6,$63,$17,$C2,$31,$F0
	.dc.b	$5F,$C7,$7E,$A1,$EC,$62,$BB,$C6,$1F,$0D,$FF,$14,$C7,$BD,$F5,$F4
	.dc.b	$78,$47,$89,$C5,$7A,$3E,$A1,$F5,$18,$AF,$E4,$DE,$73,$AF,$FA,$35
	.dc.b	$BE,$99,$F0,$9F,$CC,$6B,$53,$F2,$0D,$46,$6B,$8D,$70,$9F,$8F,$5C
	.dc.b	$2F,$F9,$BC,$70,$33,$5C,$5E,$94,$FA,$8C,$57,$F3,$B9,$7C,$57,$9A
	.dc.b	$7E,$5C,$E6,$31,$C5,$FD,$1E,$B6,$79,$EB,$7D,$5F,$D0,$63,$37,$E4
	.dc.b	$9C,$C7,$E4,$4C,$CE,$EF,$DB,$F3,$FD,$5F,$EA,$9F,$50,$F8,$6F,$82
	.dc.b	$60,$B7,$F8,$88,$D3,$EE,$8D,$34,$D3,$8C,$7E,$3E,$6F,$86,$C6,$5F
	.dc.b	$D6,$79,$BF,$89,$3F,$E4,$5C,$3E,$C7,$D3,$7F,$67,$9C,$C7,$F9,$5C
	.dc.b	$C9,$F9,$A7,$38,$7F,$94,$F0,$57,$1F,$33,$ED,$B6,$F8,$47,$DE,$EA
	.dc.b	$C5,$FF,$6F,$F0,$FE,$F6,$5F,$E5,$79,$7D,$43,$E2,$3F,$BD,$56,$B1
	.dc.b	$FF,$35,$56,$A7,$3C,$9F,$E0,$1B,$E1,$5F,$B0,$BE,$18,$FF,$A2,$A7
	.dc.b	$2F,$86,$B8,$D7,$F8,$37,$BB,$FC,$0D,$97,$F7,$FC,$38,$D9,$7D,$07
	.dc.b	$F5,$CF,$A8,$CC,$63,$0F,$ED,$71,$89,$D6,$3E,$A6,$CF,$97,$6F,$6F
	.dc.b	$0F,$B9,$B1,$92,$63,$FE,$15,$3F,$D1,$5F,$05,$FE,$8B,$FA,$77,$D4
	.dc.b	$7F,$21,$D3,$E9,$DF,$F1,$0B,$D8,$AD,$BF,$E9,$39,$DB,$9A,$FF,$57
	.dc.b	$5F,$57,$A6,$1F,$E8,$BB,$E3,$5F,$CB,$EB,$4F,$39,$D5,$72,$8E,$86
	.dc.b	$70,$DB,$C4,$C9,$F3,$67,$8B,$FE,$C4,$AD,$E1,$E6,$C3,$C3,$3E,$C7
	.dc.b	$C7,$19,$C3,$C1,$18,$BE,$75,$DE,$31,$87,$8A,$39,$57,$BA,$31,$0C
	.dc.b	$3C,$37,$CA,$36,$7A,$CF,$49,$83,$D5,$3C,$A7,$8B,$63,$FD,$B4,$7D
	.dc.b	$42,$C6,$1C,$1F,$D6,$FE,$9F,$F8,$CF,$BD,$C1,$B1,$FE,$E3,$6B,$FE
	.dc.b	$EA,$6B,$C7,$E4,$FF,$CF,$AE,$DA,$ED,$AE,$F8,$3B,$47,$0C,$FE,$FF
	.dc.b	$D4,$3E,$93,$FB,$74,$61,$BA,$F1,$C3,$ED,$E2,$C3,$94,$FF,$75,$C3
	.dc.b	$95,$7E,$09,$AF,$09,$DB,$FE,$E7,$AE,$58,$EF,$0E,$F3,$FE,$EB,$BE
	.dc.b	$38,$FE,$52,$F7,$3C,$EF,$83,$0C,$12,$7D,$4E,$5E,$58,$78,$A1,$5D
	.dc.b	$27,$71,$F9,$8B,$F0,$E3,$F0,$67,$8E,$1C,$E3,$E0,$63,$E8,$BE,$78
	.dc.b	$F5,$C3,$E0,$DF,$FA,$35,$BA,$70,$AF,$F0,$4E,$5E,$1C,$71,$1F,$E9
	.dc.b	$7A,$DB,$F0,$F1,$FF,$11,$61,$FE,$97,$BE,$12,$1F,$51,$FE,$BE,$7F
	.dc.b	$C2,$AD,$7F,$D4,$4D,$9F,$F5,$07,$D3,$FF,$4B,$9D,$F4,$F6,$CF,$FC
	.dc.b	$61,$3F,$E5,$B9,$FF,$0F,$B5,$E7,$3B,$7C,$56,$F9,$BC,$47,$91,$B1
	.dc.b	$98,$E7,$E0,$62,$64,$7A,$1E,$36,$9C,$CF,$FA,$EE,$3F,$C2,$6F,$93
	.dc.b	$C4,$CE,$5F,$EB,$BF,$A8,$FE,$77,$3F,$E3,$66,$BF,$EC,$36,$C7,$C2
	.dc.b	$F3,$3C,$F1,$8D,$3F,$CA,$0D,$AF,$9C,$D5,$82,$B8,$F4,$7A,$3E,$A2
	.dc.b	$57,$86,$3E,$1F,$CF,$FF,$9A,$A3,$FC,$92,$D7,$FD,$AC,$DC,$BE,$AF
	.dc.b	$5F,$0F,$56,$57,$07,$E2,$C6,$23,$E2,$7B,$3F,$20,$C5,$7F,$B7,$7E
	.dc.b	$0F,$FC,$A5,$3F,$E5,$86,$BF,$EE,$56,$C7,$FB,$8F,$97,$FC,$D1,$3B
	.dc.b	$F2,$BE,$DD,$6B,$83,$FC,$E8,$D9,$FC,$95,$70,$7A,$57,$E0,$A1,$F0
	.dc.b	$F4,$E7,$1F,$86,$F3,$FF,$09,$5F,$EF,$76,$E3,$FE,$F5,$6E,$3F,$E8
	.dc.b	$3F,$A8,$FF,$32,$D7,$F9,$CA,$72,$FE,$8D,$A9,$B1,$C2,$3B,$73,$7E
	.dc.b	$5A,$1F,$21,$4F,$24,$67,$1F,$CD,$FD,$47,$E9,$ED,$93,$2F,$F4,$B4
	.dc.b	$E6,$31,$3F,$B5,$C7,$E3,$29,$D5,$70,$F1,$3A,$9D,$E7,$FD,$1B,$F5
	.dc.b	$1F,$EB,$36,$FB,$DF,$AC,$8C,$47,$8A,$B9,$3C,$3D,$0D,$11,$FF,$4B
	.dc.b	$FD,$43,$E2,$3F,$6B,$F5,$0F,$84,$A4,$AE,$47,$FA,$7A,$1F,$49,$C0
	.dc.b	$91,$9F,$A8,$FD,$FF,$C5,$3D,$CA,$49,$FF,$53,$FC,$7B,$D2,$C4,$7A
	.dc.b	$4F,$E3,$55,$C5,$FF,$56,$B5,$D0,$9E,$4E,$AB,$F8,$7C,$3E,$A2,$9D
	.dc.b	$B9,$7E,$9D,$3E,$A3,$E8,$BF,$CA,$5F,$F3,$C3,$FE,$BC,$6B,$E0,$76
	.dc.b	$F1,$E4,$FE,$3B,$1A,$9E,$AD,$E2,$FF,$6C,$3C,$0F,$E0,$D5,$D7,$C4
	.dc.b	$E2,$72,$FA,$1B,$19,$21,$E1,$7E,$6F,$CB,$48,$F8,$AF,$8B,$8F,$11
	.dc.b	$F1,$B6,$4E,$32,$F2,$D5,$CB,$CB,$F5,$1E,$A3,$18,$7F,$4B,$FA,$77
	.dc.b	$8D,$FC,$E9,$FC,$0C,$7F,$B8,$1A,$FD,$47,$6E,$63,$87,$24,$8C,$1C
	.dc.b	$7E,$A3,$A1,$1E,$9D,$F5,$E9,$17,$3F,$B3,$87,$EF,$3E,$A1,$F3,$1F
	.dc.b	$D3,$9B,$7F,$EF,$46,$BF,$88,$FA,$8F,$E0,$17,$F8,$29,$CB,$6B,$91
	.dc.b	$FD,$62,$77,$AF,$F6,$DB,$5F,$08,$C4,$BE,$5F,$81,$CA,$BF,$BA,$61
	.dc.b	$F9,$09,$FF,$14,$46,$2B,$8F,$AF,$DC,$F0,$77,$D9,$D1,$FD,$F6,$B8
	.dc.b	$BE,$B9,$DB,$A1,$1F,$A4,$7E,$07,$57,$AC,$F6,$75,$C8,$E1,$3F,$AD
	.dc.b	$C7,$CD,$59,$9F,$D7,$FD,$43,$EA,$3F,$BD,$B6,$3C,$0D,$7F,$6B,$E9
	.dc.b	$FE,$9A,$31,$7F,$B0,$6E,$9F,$2D,$B5,$FE,$19,$AE,$17,$FE,$08,$7F
	.dc.b	$C1,$D7,$FB,$86,$DF,$19,$D7,$56,$FA,$A6,$1F,$0E,$FF,$5E,$F9,$27
	.dc.b	$2F,$9A,$78,$BB,$E8,$F0,$7D,$44,$8C,$10,$F6,$B9,$7D,$F1,$FE,$1E
	.dc.b	$FA,$8F,$E0,$A4,$EE,$C7,$C3,$E5,$FF,$56,$76,$FE,$2F,$E9,$FF,$DC
	.dc.b	$58,$7F,$B0,$23,$1F,$0B,$A8,$FF,$1E,$61,$FD,$3F,$E0,$7F,$77,$1F
	.dc.b	$E3,$FB,$E2,$FF,$A6,$EF,$8B,$CC,$D5,$F3,$57,$CE,$77,$1C,$4D,$9C
	.dc.b	$45,$76,$BF,$0C,$72,$C3,$A4,$BE,$9A,$92,$79,$43,$EA,$31,$1D,$C7
	.dc.b	$D3,$BE,$A3,$FB,$DE,$3D,$A3,$1F,$19,$CB,$1F,$54,$62,$F9,$36,$B8
	.dc.b	$5F,$2C,$FE,$75,$F0,$8C,$5F,$F9,$9E,$F8,$57,$EF,$5B,$1F,$36,$B8
	.dc.b	$61,$F0,$E7,$33,$E8,$BE,$6F,$C6,$AD,$BE,$48,$CC,$F0,$CB,$E3,$4F
	.dc.b	$39,$C9,$2F,$DD,$19,$63,$CD,$CB,$49,$8F,$C0,$8C,$B6,$33,$8F,$B6
	.dc.b	$BC,$F7,$E2,$9C,$9F,$0E,$3C,$4F,$D5,$8D,$57,$32,$57,$3B,$EE,$DA
	.dc.b	$E6,$7C,$58,$CB,$D5,$F5,$CF,$28,$F9,$C6,$27,$94,$66,$3E,$61,$93
	.dc.b	$91,$DA,$BC,$B3,$C0,$91,$A9,$FC,$BC,$EA,$31,$3A,$7C,$C7,$BF,$1F
	.dc.b	$B3,$5E,$1C,$7E,$C4,$62,$B8,$46,$2B,$C8,$72,$AF,$45,$F7,$8C,$57
	.dc.b	$0C,$7E,$8E,$1C,$E7,$C6,$43,$DE,$7C,$AC,$3F,$1E,$77,$71,$3F,$5B
	.dc.b	$0F,$6D,$73,$9E,$E7,$3C,$3C,$57,$CA,$F8,$61,$FA,$4A,$F4,$CE,$71
	.dc.b	$FA,$F1,$F1,$0F,$EC,$05,$3F,$4D,$E3,$3E,$2F,$31,$F5,$0D,$D2,$75
	.dc.b	$1F,$B6,$6E,$7F,$C2,$BE,$9F,$FC,$85,$D1,$FA,$D9,$CE,$5D,$B1,$C1
	.dc.b	$FC,$4E,$B7,$87,$38,$CB,$4F,$E1,$91,$AC,$3F,$85,$5F,$49,$FE,$00
	.dc.b	$D7,$F8,$8B,$67,$85,$F8,$67,$A5,$72,$BF,$17,$6F,$F9,$F2,$31,$3A
	.dc.b	$AE,$11,$ED,$9D,$46,$09,$3E,$2A,$E7,$8F,$AA,$F9,$92,$BF,$C2,$18
	.dc.b	$79,$2F,$81,$DC,$B3,$DE,$B9,$57,$69,$F6,$DF,$67,$81,$82,$DF,$57
	.dc.b	$17,$CD,$CE,$1E,$A7,$BD,$F9,$23,$11,$C2,$31,$87,$8A,$02,$FF,$D5
	.dc.b	$B1,$F6,$FE,$A3,$8C,$62,$7E,$CC,$63,$D6,$95,$F8,$1F,$5E,$D7,$DE
	.dc.b	$6E,$5F,$8B,$6E,$3F,$21,$B3,$E2,$C7,$FD,$85,$7F,$8B,$C3,$D9,$F2
	.dc.b	$2F,$8A,$FF,$51,$4F,$F8,$C2,$31,$F0,$DF,$EB,$AF,$03,$FD,$77,$1B
	.dc.b	$F5,$31,$5E,$E8,$DB,$C6,$F9,$E5,$FE,$9D,$9C,$B6,$33,$7D,$16,$1D
	.dc.b	$25,$5E,$5F,$A8,$C3,$E7,$FA,$8B,$69,$B1,$86,$E3,$E3,$6D,$FF,$AA
	.dc.b	$BE,$A3,$FD,$E7,$3F,$E4,$EE,$37,$D4,$7F,$A2,$23,$FC,$99,$3A,$AF
	.dc.b	$F5,$B5,$6B,$87,$CD,$09,$D6,$BF,$EE,$0A,$D5,$F2,$F7,$FD,$91,$8B
	.dc.b	$F1,$47,$2B,$E3,$5F,$88,$F9,$07,$C1,$7D,$8B,$DC,$FC,$DF,$43,$E0
	.dc.b	$E1,$F0,$9B,$19,$3E,$94,$65,$A7,$D5,$F6,$7E,$49,$F2,$BE,$A2,$C6
	.dc.b	$49,$7F,$5E,$32,$FF,$B4,$BE,$97,$FD,$D5,$3F,$E7,$6D,$5F,$51,$F7
	.dc.b	$23,$FD,$11,$3A,$AF,$F6,$F5,$6B,$9B,$E4,$CE,$63,$F1,$D7,$C6,$BF
	.dc.b	$53,$39,$BF,$F4,$5C,$E6,$53,$97,$7E,$47,$E2,$76,$F9,$3E,$9F,$68
	.dc.b	$FC,$3F,$D4,$4C,$7F,$2D,$E0,$5C,$3F,$29,$37,$3F,$CA,$4D,$AF,$CF
	.dc.b	$74,$1F,$C8,$FC,$FF,$E4,$9A,$7E,$84,$B8,$7E,$7D,$B7,$FE,$F2,$D1
	.dc.b	$F4,$FF,$84,$FA,$8F,$D0,$D7,$FA,$6E,$72,$FE,$56,$33,$7F,$46,$B8
	.dc.b	$76,$FC,$A6,$D7,$17,$0F,$1E,$1B,$5F,$B3,$1F,$51,$FA,$C6,$CF,$F0
	.dc.b	$15,$F1,$79,$99,$39,$78,$46,$27,$A7,$D4,$73,$F8,$DE,$8D,$BE,$8D
	.dc.b	$9D,$C7,$79,$FD,$5F,$D3,$BE,$2B,$ED,$7D,$43,$E4,$33,$18,$6D,$FF
	.dc.b	$B1,$BA,$15,$B8,$FF,$62,$5F,$37,$51,$9C,$5C,$5F,$6D,$FE,$6F,$5B
	.dc.b	$73,$5F,$EE,$18,$C4,$EF,$EA,$1F,$50,$FA,$8F,$1E,$1C,$0F,$E8,$3F
	.dc.b	$50,$E7,$FE,$0F,$E7,$FE,$66,$31,$9F,$FB,$66,$3F,$39,$E2,$BD,$CF
	.dc.b	$E1,$AF,$D2,$9F,$F6,$F3,$96,$F2,$7F,$12,$FA,$77,$CA,$70,$BC,$36
	.dc.b	$BF,$DC,$D2,$F8,$2F,$E3,$3F,$50,$F5,$B9,$46,$1E,$9E,$9E,$2F,$8B
	.dc.b	$77,$F7,$0C,$3F,$8B,$71,$FC,$B8,$C4,$FF,$26,$FA,$87,$D4,$74,$C3
	.dc.b	$FD,$F1,$E9,$BB,$3F,$A0,$43,$E9,$9F,$09,$FC,$BA,$B5,$3F,$DB,$33
	.dc.b	$FB,$0E,$E7,$57,$CA,$F8,$CF,$68,$C4,$EA,$30,$73,$3F,$BB,$9E,$1F
	.dc.b	$A8,$D6,$9E,$E8,$7C,$57,$9A,$7E,$7C,$71,$9E,$33,$9E,$2F,$E8,$B3
	.dc.b	$C0,$D0,$AE,$35,$D6,$B8,$4F,$38,$FE,$1D,$1C,$8F,$67,$CF,$D9,$D7
	.dc.b	$D4,$3E,$1B,$E0,$1F,$D5,$1C,$47,$83,$4B,$87,$D8,$9B,$AF,$CA,$9B
	.dc.b	$5F,$E1,$8A,$56,$1F,$E4,$13,$5C,$DF,$73,$E8,$7C,$25,$BF,$57,$27
	.dc.b	$F1,$8C,$7B,$7D,$37,$F6,$88,$C4,$FD,$0C,$53,$9F,$2B,$ED,$36,$7D
	.dc.b	$4D,$CB,$F2,$37,$C9,$F2,$F5,$E7,$1F,$C9,$78,$9F,$88,$9F,$EF,$3F
	.dc.b	$50,$F8,$8F,$EF,$33,$9A,$FE,$F1,$39,$CB,$E9,$4E,$5F,$F1,$DE,$09
	.dc.b	$D6,$1F,$A2,$AD,$3F,$E8,$58,$CD,$F8,$B9,$7B,$9F,$E0,$2E,$FF,$EF
	.dc.b	$F5,$FD,$FE,$3C,$7F,$3F,$FE,$21,$A7,$D4,$74,$A6,$2A,$D4,$E6,$31
	.dc.b	$9F,$F7,$38,$C4,$EB,$0F,$7E,$DF,$03,$86,$DF,$B9,$B8,$FA,$E7,$FC
	.dc.b	$3A,$FF,$96,$BE,$0F,$FC,$B7,$F5,$0F,$A8,$FC,$36,$9F,$63,$1F,$CC
	.dc.b	$5E,$C5,$6D,$FF,$0E,$61,$FC,$BE,$31,$1F,$8F,$C3,$AE,$1D,$2B,$4F
	.dc.b	$FA,$BA,$33,$3C,$AE,$BC,$4E,$F3,$C6,$33,$1E,$3B,$EF,$87,$9E,$36
	.dc.b	$7F,$AF,$AF,$9C,$66,$7C,$B3,$C6,$FA,$3C,$0F,$04,$74,$C3,$CF,$1C
	.dc.b	$CD,$57,$3A,$F5,$46,$27,$6F,$28,$EB,$5E,$08,$5F,$92,$B7,$1E,$33
	.dc.b	$E1,$1E,$33,$07,$E1,$9D,$BC,$8F,$F6,$B9,$F5,$4B,$18,$70,$7F,$01
	.dc.b	$FA,$7F,$E3,$FE,$AF,$06,$C7,$FB,$81,$AF,$FB,$99,$AF,$1F,$93,$FE
	.dc.b	$DA,$ED,$AE,$DA,$EF,$83,$B4,$70,$C7,$F0,$BF,$50,$FA,$4F,$E2,$91
	.dc.b	$86,$E9,$C7,$0F,$A9,$8B,$0E,$53,$FC,$6F,$0E,$55,$F8,$86,$BC,$27
	.dc.b	$6F,$FA,$E2,$B9,$63,$BC,$3B,$CF,$FA,$F2,$F8,$E3,$F9,$AB,$DC,$F3
	.dc.b	$BE,$0C,$30,$49,$F5,$39,$79,$61,$E2,$85,$74,$9D,$C7,$E4,$6F,$C3
	.dc.b	$8F,$C1,$9E,$38,$73,$8F,$81,$8F,$A2,$F9,$E3,$D7,$0F,$83,$7F,$DD
	.dc.b	$DB,$B7,$0A,$FE,$A5,$97,$87,$1C,$47,$F8,$02,$B6,$FC,$3C,$7F,$A4
	.dc.b	$61,$FE,$01,$BE,$13,$FE,$E2,$FA,$8F,$F5,$EB,$FD,$75,$AF,$F8,$29
	.dc.b	$B3,$FE,$09,$FA,$77,$0F,$4C,$FF,$AA,$56,$9F,$EA,$D3,$FE,$6A,$BF
	.dc.b	$ED,$0D,$7A,$75,$71,$7C,$5F,$37,$55,$A3,$0D,$8E,$7D,$F8,$99,$1E
	.dc.b	$97,$95,$A7,$33,$FE,$24,$8F,$F0,$CB,$E4,$F0,$F3,$9F,$F8,$97,$EA
	.dc.b	$3F,$9D,$4F,$F7,$F6,$BF,$E2,$C6,$C7,$C3,$F3,$3D,$11,$8D,$3F,$BB
	.dc.b	$36,$81,$C2,$C3,$FC,$98,$F7,$E9,$F4,$FD,$44,$8C,$1F,$E1,$1F,$9F
	.dc.b	$FF,$38,$37,$D7,$F9,$11,$88,$F7,$76,$62,$64,$6C,$FA,$71,$87,$E2
	.dc.b	$7A,$1F,$20,$C6,$BF,$E4,$7F,$88,$FF,$2C,$CF,$F8,$85,$AF,$F9,$39
	.dc.b	$B1,$FE,$4D,$E6,$FF,$36,$4E,$FC,$AF,$B9,$5A,$DF,$FC,$34,$D9,$F8
	.dc.b	$B5,$C1,$E9,$5F,$83,$87,$C3,$D3,$9C,$7E,$1F,$CF,$FC,$25,$7F,$98
	.dc.b	$1B,$8F,$F9,$7D,$B8,$FF,$8A,$7E,$A3,$FC,$D9,$5F,$E3,$B9,$CB,$F5
	.dc.b	$9A,$9B,$1C,$23,$B7,$37,$E5,$E1,$F1,$F4,$F2,$46,$71,$FC,$E7,$D4
	.dc.b	$7E,$A2,$FF,$C9,$4D,$7F,$CE,$F7,$F0,$D4,$67,$2F,$F3,$C3,$72,$FF
	.dc.b	$21,$36,$7F,$1B,$5C,$7E,$A3,$FD,$6D,$1F,$E5,$76,$BF,$E8,$46,$DF
	.dc.b	$4E,$CF,$D8,$47,$F9,$AA,$78,$79,$F8,$3F,$69,$F5,$0F,$88,$FD,$B7
	.dc.b	$D4,$3E,$12,$92,$F8,$9F,$E6,$78,$7D,$27,$02,$46,$7E,$A3,$F8,$07
	.dc.b	$C5,$7F,$A8,$63,$FC,$DC,$D7,$FD,$36,$DD,$7F,$D3,$5F,$1E,$F4,$38
	.dc.b	$FB,$B2,$F8,$BF,$C2,$E7,$51,$83,$FC,$F4,$BE,$A3,$47,$18,$FF,$54
	.dc.b	$FD,$47,$CD,$4F,$F9,$ED,$FF,$41,$35,$DF,$17,$F1,$98,$E0,$67,$1F
	.dc.b	$F5,$83,$79,$7F,$CF,$4E,$BC,$5E,$04,$BE,$BF,$43,$5F,$1C,$5B,$E7
	.dc.b	$3F,$23,$E5,$65,$7C,$87,$C5,$C7,$88,$F9,$31,$6B,$C7,$37,$1F,$1F
	.dc.b	$D4,$7A,$4C,$5F,$F1,$8F,$A7,$78,$BF,$CD,$5F,$B7,$1F,$EA,$36,$BF
	.dc.b	$ED,$07,$6E,$63,$87,$24,$8C,$1C,$7E,$A3,$A1,$1E,$9D,$D5,$E9,$17
	.dc.b	$3F,$F6,$D6,$1F,$B6,$FA,$87,$CC,$7F,$49,$6D,$FF,$AD,$DA,$FF,$B9
	.dc.b	$3E,$A3,$F7,$57,$FE,$E3,$9C,$B6,$B9,$1F,$D4,$E7,$7A,$FF,$AC,$6B
	.dc.b	$6F,$FA,$CE,$1F,$2F,$B3,$8E,$1F,$EC,$78,$FF,$10,$37,$D6,$E8,$75
	.dc.b	$EF,$B1,$92,$1E,$77,$C5,$D5,$E3,$EA,$93,$D6,$32,$F7,$8C,$BE,$F8
	.dc.b	$FE,$D9,$F5,$0F,$A8,$A4,$9F,$F6,$E7,$A5,$F9,$F9,$D4,$7F,$B7,$DA
	.dc.b	$B7,$6B,$FE,$D3,$9C,$C6,$27,$FA,$FD,$73,$BF,$EF,$8F,$F8,$06,$75
	.dc.b	$28,$C5,$F3,$AD,$F4,$F0,$E9,$97,$E6,$8C,$BE,$88,$CB,$E5,$AF,$F7
	.dc.b	$35,$FF,$6C,$8C,$1E,$4E,$8F,$1F,$D4,$48,$F8,$31,$E6,$30,$F0,$3C
	.dc.b	$D1,$97,$FB,$DE,$5E,$AF,$A8,$FE,$0C,$4E,$DF,$F0,$D3,$FB,$F9,$FF
	.dc.b	$00,$BF,$CF,$FD,$D7,$6F,$FB,$EF,$E9,$BF,$DC,$15,$FE,$BB,$8C,$7C
	.dc.b	$27,$FA,$C5,$FF,$80,$17,$C5,$CC,$EC,$FF,$06,$D7,$27,$FD,$2B,$1A
	.dc.b	$69,$E2,$C3,$A3,$C2,$EA,$AE,$58,$79,$A3,$47,$71,$87,$8D,$F0,$FD
	.dc.b	$3F,$2A,$34,$C7,$D4,$59,$E5,$F4,$EF,$A8,$FE,$E7,$8F,$79,$4E,$7E
	.dc.b	$27,$C1,$8F,$49,$E3,$7C,$35,$FF,$16,$C7,$08,$C3,$FE,$30,$BE,$4F
	.dc.b	$C2,$3D,$D9,$7C,$08,$CE,$3E,$16,$DF,$46,$D7,$B8,$F0,$E3,$1E,$91
	.dc.b	$A7,$A9,$9A,$E9,$87,$92,$3C,$D5,$EC,$C3,$C5,$39,$BE,$47,$3C,$7E
	.dc.b	$05,$F6,$9D,$12,$F9,$47,$7C,$7D,$13,$98,$F8,$15,$CB,$1F,$1D,$EE
	.dc.b	$7E,$35,$79,$6B,$57,$CA,$B5,$5E,$83,$9E,$3E,$88,$CE,$3E,$78,$C4
	.dc.b	$F9,$E7,$8D,$F4,$8E,$F1,$C3,$0F,$05,$70,$C7,$CB,$5C,$4E,$E6,$EB
	.dc.b	$9C,$EF,$1F,$34,$EE,$BF,$15,$87,$78,$F4,$47,$42,$4F,$EC,$6B,$94
	.dc.b	$F9,$0E,$51,$97,$C0,$60,$FC,$11,$4E,$AF,$8A,$7C,$33,$D2,$72,$75
	.dc.b	$8C,$BD,$1F,$1B,$D1,$AF,$02,$57,$AF,$1F,$C2,$E3,$F7,$6B,$9D,$78
	.dc.b	$E7,$38,$FE,$12,$31,$5B,$C7,$D9,$18,$9F,$96,$F8,$E3,$D4,$6A,$FE
	.dc.b	$9C,$7E,$B8,$FC,$11,$6D,$E4,$3E,$27,$53,$F2,$CD,$46,$5F,$D3,$B6
	.dc.b	$BF,$80,$7D,$3F,$F8,$EF,$97,$F8,$2C,$E7,$1F,$D9,$36,$FF,$78,$D5
	.dc.b	$5C,$5E,$B8,$72,$23,$FC,$06,$35,$87,$EF,$EB,$96,$3E,$AB,$DC,$F9
	.dc.b	$3A,$BF,$AE,$3B,$8C,$E3,$E4,$7F,$92,$61,$D2,$31,$8F,$88,$95,$CF
	.dc.b	$2E,$31,$EA,$BA,$FA,$B0,$EF,$7C,$AF,$C1,$3C,$A3,$F8,$FE,$1D,$E3
	.dc.b	$91,$92,$CF,$77,$07,$F2,$CC,$BD,$E7,$C9,$9E,$51,$9C,$3D,$78,$78
	.dc.b	$A3,$9E,$1E,$08,$C4,$6C,$F9,$8E,$32,$F9,$51,$8C,$BD,$B8,$7F,$A9
	.dc.b	$9F,$8F,$F5,$1C,$A3,$13,$F5,$63,$1E,$A4,$AF,$9B,$EA,$DA,$FB,$4D
	.dc.b	$D7,$F1,$2D,$CF,$E2,$B6,$7F,$DE,$38,$FF,$AE,$EF,$EB,$E1,$ED,$F9
	.dc.b	$07,$C4,$7F,$A6,$E7,$FB,$E9,$FE,$F8,$8C,$FC,$27,$F8,$77,$A7,$FC
	.dc.b	$3D,$1C,$3D,$3C,$56,$ED,$78,$C9,$7B,$AE,$B1,$8A,$DC,$62,$75,$3F
	.dc.b	$E1,$87,$C3,$19,$97,$D3,$F5,$9F,$4F,$D4,$59,$CD,$6E,$31,$59,$6C
	.dc.b	$64,$F1,$DB,$E9,$FF,$DD,$F1,$E6,$E4,$7D,$47,$F8,$62,$37,$3A,$AF
	.dc.b	$F1,$6D,$6B,$87,$E0,$47,$3F,$77,$E1,$4E,$6F,$8C,$63,$0E,$95,$FE
	.dc.b	$2F,$C3,$A6,$FF,$E4,$66,$DA,$5F,$1C,$F8,$0F,$B9,$1C,$71,$F8,$5E
	.dc.b	$7F,$C3,$6D,$FC,$96,$CF,$CD,$DF,$AB,$F4,$EB,$85,$7C,$FF,$63,$E6
	.dc.b	$1F,$37,$EA,$2C,$64,$95,$CE,$32,$FF,$92,$7E,$9B,$FD,$D1,$3F,$80
	.dc.b	$D5,$F5,$1F,$8F,$AD,$79,$1F,$87,$9C,$F3,$FC,$AC,$38,$D7,$F9,$6A
	.dc.b	$73,$87,$D8,$9C,$C7,$E5,$70,$E8,$FF,$9B,$A3,$33,$FE,$26,$6D,$7C
	.dc.b	$AD,$BF,$D3,$33,$CF,$1F,$C5,$7D,$44,$8F,$C8,$78,$16,$FE,$44,$DC
	.dc.b	$7F,$47,$BF,$F4,$2B,$FC,$B4,$FE,$4D,$FC,$A7,$CF,$57,$F3,$4D,$BF
	.dc.b	$A4,$DB,$E6,$FC,$1F,$A7,$FC,$4F,$D4,$7E,$9A,$BE,$64,$63,$B3,$58
	.dc.b	$7E,$5E,$B5,$87,$C1,$AD,$4F,$C7,$77,$7F,$98,$AF,$8F,$3F,$16,$7F
	.dc.b	$59,$2F,$A8,$A6,$3E,$26,$D7,$ED,$E7,$4D,$9D,$34,$D7,$D4,$74,$F8
	.dc.b	$BE,$AD,$8C,$9C,$E3,$2D,$3F,$6B,$F4,$CF,$A8,$FE,$19,$D6,$FA,$87
	.dc.b	$C7,$FE,$C2,$B5,$7F,$C3,$AB,$53,$9B,$E3,$7C,$2B,$AC,$62,$75,$18
	.dc.b	$3F,$8B,$FD,$43,$EA,$1F,$51,$AD,$BF,$D4,$DF,$50,$E5,$FE,$03,$E7
	.dc.b	$FC,$C9,$FD,$C4,$E7,$1F,$E1,$F3,$99,$FF,$04,$7B,$7F,$C2,$3C,$67
	.dc.b	$25,$8C,$3F,$C3,$3E,$9D,$F2,$79,$C7,$F8,$CC,$61,$FE,$37,$6B,$E1
	.dc.b	$F0,$1F,$C5,$3E,$A1,$EA,$E2,$FF,$90,$C6,$1E,$7E,$96,$74,$FF,$61
	.dc.b	$65,$FC,$3B,$B7,$FD,$B2,$B1,$F2,$54,$BF,$E5,$7F,$50,$FA,$8C,$63
	.dc.b	$FE,$CB,$F4,$9D,$9F,$D1,$6D,$F4,$BF,$EE,$37,$F9,$9C,$E9,$FE,$8E
	.dc.b	$6B,$E0,$7F,$96,$C6,$6F,$FD,$BB,$39,$38,$DF,$5B,$E9,$3E,$18,$E2
	.dc.b	$67,$EA,$2D,$7F,$42,$1F,$0D,$E5,$8F,$99,$5F,$EE,$39,$C9,$E4,$8C
	.dc.b	$EF,$FE,$FC,$8C,$CF,$92,$73,$1F,$EB,$63,$32,$7F,$9E,$08,$DC,$E2
	.dc.b	$BF,$A5,$CF,$03,$78,$3E,$7B,$33,$FD,$43,$EA,$1F,$51,$F0,$4C,$16
	.dc.b	$FF,$AF,$C6,$9F,$9B,$77,$0F,$F8,$01,$37,$2F,$34,$79,$E7,$15,$FE
	.dc.b	$D0,$AB,$3F,$D5,$5B,$1F,$73,$E1,$7C,$2F,$F6,$EE,$5F,$E2,$AF,$AB
	.dc.b	$E9,$BF,$B3,$C6,$2F,$B4,$63,$CA,$FB,$4D,$AF,$3B,$65,$97,$99,$FC
	.dc.b	$ED,$7E,$07,$A7,$53,$9F,$17,$F1,$0F,$B2,$1F,$50,$F8,$7F,$EF,$35
	.dc.b	$AC,$BF,$B8,$56,$A7,$3C,$7F,$DF,$2F,$85,$7F,$82,$AF,$86,$5F,$E8
	.dc.b	$39,$CB,$E6,$9F,$F7,$BC,$EA,$3F,$BD,$D7,$F8,$1B,$BF,$FB,$9C,$62
	.dc.b	$BA,$7D,$07,$F6,$5F,$A8,$E1,$82,$31,$87,$A1,$B5,$C0,$FE,$E1,$5C
	.dc.b	$36,$F6,$F0,$7B,$9B,$3E,$97,$0F,$F8,$93,$1F,$F0,$87,$C1,$7F,$84
	.dc.b	$7E,$A1,$F5,$1E,$AB,$FF,$0F,$D6,$84,$E9,$FF,$0C,$68,$9F,$C5,$CE
	.dc.b	$67,$FC,$38,$2F,$B5,$F5,$C2,$CF,$FA,$C6,$6B,$FE,$22,$BE,$75,$FE
	.dc.b	$B0,$9C,$E1,$DE,$F7,$7E,$5C,$3C,$13,$98,$F3,$DF,$16,$9E,$8C,$3A
	.dc.b	$BE,$13,$27,$86,$FA,$BE,$93,$35,$D6,$FD,$D1,$E1,$BF,$63,$C8,$C1
	.dc.b	$DE,$B9,$E1,$EA,$C3,$C3,$3F,$1A,$31,$3A,$8E,$F3,$A7,$99,$E5,$3B
	.dc.b	$1E,$53,$FD,$AA,$53,$E0,$18,$9C,$FD,$3F,$F2,$2F,$52,$F4,$FF,$A1
	.dc.b	$AB,$FD,$BB,$39,$6C,$65,$FF,$73,$43,$E4,$7F,$6B,$1C,$1A,$E9,$AF
	.dc.b	$E0,$A3,$87,$2F,$8A,$78,$E9,$F8,$7F,$A8,$7D,$27,$E2,$23,$0D,$DB
	.dc.b	$95,$70,$6D,$FE,$22,$F8,$E3,$6F,$83,$FE,$6D,$C7,$C2,$FD,$E6,$E3
	.dc.b	$DA,$38,$1F,$92,$8C,$E1,$DA,$3F,$0B,$7D,$9B,$7D,$95,$70,$7C,$18
	.dc.b	$F9,$DF,$F3,$F5,$71,$C3,$C1,$1E,$39,$D3,$C9,$B3,$BC,$3F,$56,$D9
	.dc.b	$EB,$8F,$9E,$B7,$8F,$8E,$B5,$5C,$EB,$DA,$FF,$17,$3C,$73,$FB,$68
	.dc.b	$FE,$5D,$3E,$D7,$F9,$86,$3E,$BA,$4E,$76,$F1,$DF,$16,$CF,$FA,$C2
	.dc.b	$B8,$7D,$47,$FB,$22,$3F,$D6,$6D,$7C,$41,$C6,$1F,$CF,$FE,$9D,$C5
	.dc.b	$D6,$BF,$D4,$D5,$AC,$3F,$86,$D3,$D7,$D5,$23,$0F,$83,$AF,$36,$F9
	.dc.b	$BA,$49,$FF,$63,$D7,$F7,$17,$CF,$19,$7C,$F3,$C5,$DF,$83,$9C,$FD
	.dc.b	$1E,$27,$8B,$03,$9C,$FF,$91,$FD,$47,$F3,$DC,$3F,$DA,$AD,$7E,$1B
	.dc.b	$6F,$E1,$79,$9E,$C8,$C5,$72,$EB,$C1,$DE,$6D,$FA,$BE,$A3,$C4,$61
	.dc.b	$F9,$FF,$3F,$FE,$6C,$7F,$DC,$AD,$73,$EB,$7A,$7B,$FE,$5C,$E7,$3F
	.dc.b	$EE,$4D,$CB,$F9,$EB,$6F,$FD,$B5,$AF,$F9,$09,$EB,$87,$C3,$F4,$64
	.dc.b	$60,$FF,$7A,$FC,$07,$F9,$81,$BE,$97,$DA,$8C,$47,$E1,$BA,$3F,$CE
	.dc.b	$4E,$88,$7E,$2A,$37,$9F,$F8,$1E,$77,$39,$AF,$C2,$C3,$E1,$E9,$DE
	.dc.b	$3F,$13,$E8,$7D,$DA,$FC,$AB,$6F,$FC,$24,$DD,$7F,$AE,$FD,$3F,$F9
	.dc.b	$C2,$BF,$DF,$55,$AE,$7E,$64,$8C,$D7,$0F,$2B,$11,$FA,$2B,$7C,$77
	.dc.b	$E6,$9B,$1C,$00,$C7,$C4,$E2,$3E,$0D,$FF,$70,$FA,$8F,$D4,$CF,$10
	.dc.b	$E1,$F0,$67,$FA,$88,$C6,$7F,$E2,$F6,$D7,$3E,$DF,$25,$7E,$BB,$EA
	.dc.b	$3F,$D7,$18,$75,$6B,$FB,$C6,$E1,$CE,$B8,$78,$92,$B8,$3F,$B8,$FA
	.dc.b	$87,$C3,$7E,$EB,$EA,$1F,$15,$49,$7C,$0F,$64,$3E,$93,$81,$23,$3F
	.dc.b	$51,$FC,$13,$E2,$BF,$D3,$91,$F0,$DA,$FF,$12,$6E,$9F,$E4,$FF,$8E
	.dc.b	$7A,$5C,$7D,$D9,$7C,$5F,$E1,$F3,$A8,$C1,$F3,$D7,$D4,$68,$E3,$1F
	.dc.b	$C8,$3E,$A3,$E6,$BF,$71,$FF,$3E,$BF,$5D,$AE,$F8,$7F,$8E,$C7,$08
	.dc.b	$3A,$C7,$FC,$CC,$DE,$4F,$F1,$8B,$B3,$FD,$91,$3D,$7C,$4E,$67,$4F
	.dc.b	$A0,$B1,$92,$1E,$77,$E5,$7C,$AC,$8F,$94,$F9,$38,$F1,$1F,$2E,$2E
	.dc.b	$3E,$49,$B8,$F9,$3E,$A3,$D6,$62,$FF,$3B,$F4,$EF,$1B,$F9,$DB,$F7
	.dc.b	$63,$F3,$4D,$7F,$A4,$BB,$73,$1C,$39,$24,$60,$E3,$F5,$1D,$08,$F4
	.dc.b	$F0,$6B,$D2,$2E,$7F,$E9,$1C,$3F,$5F,$2F,$A8,$7C,$BF,$F5,$06,$C7
	.dc.b	$EB,$DA,$FF,$5E,$FA,$7F,$E0,$CF,$F6,$08,$CD,$FF,$A7,$5B,$AF,$F9
	.dc.b	$A1,$B3,$FC,$02,$B8,$D6,$EF,$FA,$FC,$3E,$A2,$9E,$38,$CF,$AF,$2F
	.dc.b	$F8,$15,$7F,$8E,$5B,$DF,$97,$CD,$37,$0F,$F5,$85,$7F,$BA,$27,$8F
	.dc.b	$57,$83,$AA,$46,$08,$7B,$DC,$BF,$06,$3F,$BA,$FD,$43,$EA,$29,$27
	.dc.b	$F8,$B7,$A7,$F9,$D9,$D4,$7F,$19,$6B,$FE,$08,$6E,$3F,$EC,$59,$E1
	.dc.b	$5F,$E0,$27,$FC,$03,$5F,$E0,$E6,$CF,$FB,$2E,$B8,$EF,$FE,$96,$9F
	.dc.b	$D2,$D6,$FA,$39,$1C,$FA,$64,$79,$A3,$2F,$A2,$3C,$1D,$1E,$1F,$A8
	.dc.b	$93,$EC,$8C,$BE,$E8,$CB,$E5,$8F,$F0,$CF,$D4,$7E,$F4,$9D,$B8,$F8
	.dc.b	$8C,$BF,$E8,$7E,$DF,$F5,$6F,$D3,$7F,$B7,$70,$FF,$5D,$C6,$3E,$1B
	.dc.b	$51,$FD,$23,$0F,$E8,$1F,$04,$AD,$4F,$03,$FA,$6D,$F3,$7F,$D7,$11
	.dc.b	$A6,$9C,$F0,$EA,$A7,$85,$75,$8E,$78,$6A,$73,$7F,$EC,$4C,$3C,$6F
	.dc.b	$AB,$E9,$F4,$4B,$F2,$1D,$7E,$A3,$0F,$88,$7D,$3B,$EA,$3F,$B9,$67
	.dc.b	$E1,$9C,$C6,$30,$ED,$1D,$2B,$87,$A7,$6B,$53,$FD,$BA,$72,$F7,$CD
	.dc.b	$97,$C0,$9C,$E3,$DD,$B8,$FF,$B8,$1B,$7F,$32,$B8,$61,$DA,$3C,$B8
	.dc.b	$F8,$4F,$7E,$3D,$A3,$07,$BE,$F9,$BC,$0F,$25,$F4,$AD,$E3,$E3,$C3
	.dc.b	$B6,$1C,$EF,$94,$78,$70,$E6,$74,$9D,$57,$1B,$E7,$8F,$AE,$F9,$E3
	.dc.b	$DF,$1E,$75,$A7,$DC,$53,$CD,$7C,$DF,$39,$9C,$7C,$37,$D2,$30,$F9
	.dc.b	$4C,$3D,$89,$3C,$CF,$BD,$8F,$A3,$1F,$36,$1C,$8F,$97,$8F,$9E,$B8
	.dc.b	$4E,$AF,$C6,$74,$71,$1A,$8F,$59,$93,$F0,$98,$79,$6F,$AC,$77,$34
	.dc.b	$7E,$8E,$FC,$17,$D0,$91,$D2,$FB,$3F,$1A,$32,$F1,$3A,$46,$B0,$F7
	.dc.b	$47,$8A,$F3,$5F,$9C,$8F,$0E,$3F,$32,$7C,$B1,$8C,$7D,$0F,$C0,$C3
	.dc.b	$C3,$5B,$8C,$5F,$8A,$31,$3F,$3B,$1F,$5D,$78,$4F,$7E,$3E,$47,$D4
	.dc.b	$66,$7D,$13,$C2,$32,$49,$E1,$5F,$AE,$8F,$F2,$E9,$D4,$A7,$EF,$BC
	.dc.b	$97,$C5,$E2,$34,$7E,$BE,$32,$FE,$9D,$B8,$7F,$03,$FA,$7F,$F1,$DF
	.dc.b	$3F,$F0,$19,$CE,$7C,$70,$C1,$FC,$42,$B7,$87,$38,$CB,$4F,$E2,$51
	.dc.b	$AC,$3F,$81,$5F,$49,$FE,$14,$D7,$D5,$5D,$7A,$D8,$F4,$BE,$33,$71
	.dc.b	$F8,$78,$F5,$7C,$D8,$78,$6B,$6F,$8B,$0E,$6E,$67,$73,$A3,$E1,$4E
	.dc.b	$9C,$46,$89,$3F,$E8,$EA,$F0,$3E,$AB,$F2,$34,$F0,$61,$85,$3C,$1F
	.dc.b	$79,$A3,$C8,$75,$C3,$C5,$7E,$08,$E3,$18,$7B,$61,$F6,$7D,$5C,$47
	.dc.b	$6F,$A8,$B8,$7F,$AA,$31,$FB,$11,$8A,$91,$F6,$FD,$6B,$7F,$61,$B8
	.dc.b	$7E,$15,$B9,$7D,$06,$CF,$C3,$6F,$C8,$BE,$2B,$FD,$2B,$87,$FA,$D6
	.dc.b	$31,$F0,$BF,$4B,$C0,$FA,$71,$BF,$53,$15,$FE,$BA,$8D,$BC,$6F,$9E
	.dc.b	$2B,$E3,$0C,$38,$CA,$FA,$5F,$19,$CC,$79,$7E,$A2,$DB,$EA,$2C,$61
	.dc.b	$B7,$E1,$6E,$1E,$18,$F3,$57,$17,$FD,$95,$F4,$FF,$EE,$97,$FD,$A9
	.dc.b	$C6,$FA,$8F,$F0,$64,$6B,$DE,$F3,$CE,$78,$3D,$38,$B0,$E4,$F7,$3F
	.dc.b	$7B,$87,$27,$06,$EF,$9C,$67,$6F,$CF,$CE,$71,$F0,$1F,$EE,$7F,$90
	.dc.b	$7C,$0F,$FB,$9E,$33,$3D,$F7,$FB,$5E,$77,$C6,$CB,$E3,$B6,$BE,$A3
	.dc.b	$6B,$86,$DF,$E8,$09,$FB,$15,$F3,$BD,$99,$1F,$6F,$EA,$2D,$F0,$9B
	.dc.b	$A7,$DC,$9B,$5F,$C8,$3E,$9F,$FD,$C2,$FF,$BF,$E5,$F5,$1F,$84,$7D
	.dc.b	$F3,$AC,$3E,$44,$6B,$9B,$E6,$CE,$63,$F2,$55,$C6,$FB,$79,$1F,$8A
	.dc.b	$9E,$10,$C3,$E7,$DF,$18,$D9,$F2,$F7,$FF,$47,$3B,$AF,$1D,$FE,$3B
	.dc.b	$EA,$24,$FE,$6F,$BE,$E3,$F2,$26,$DF,$E7,$26,$D7,$E3,$8F,$CB,$FC
	.dc.b	$FF,$E6,$1A,$7E,$94,$B7,$F4,$9B,$5F,$13,$E4,$3F,$1F,$F5,$1F,$A7
	.dc.b	$9F,$31,$F3,$63,$3D,$9F,$19,$B1,$AB,$F8,$5A,$A7,$31,$F1,$1E,$13
	.dc.b	$FA,$0B,$F9,$93,$C2,$7F,$62,$BE,$A2,$98,$F8,$AB,$7F,$C1,$E7,$4D
	.dc.b	$9D,$34,$D7,$D4,$74,$F8,$EE,$AD,$8C,$9C,$E3,$2D,$3D,$94,$FA,$87
	.dc.b	$D3,$FF,$0A,$EC,$7D,$43,$E4,$3F,$6B,$5A,$C3,$F5,$53,$97,$9C,$67
	.dc.b	$0F,$DC,$5F,$2B,$E3,$87,$5C,$38,$57,$EE,$AF,$87,$0F,$EE,$1B,$8F
	.dc.b	$E4,$7E,$A1,$F5,$0F,$A8,$E7,$3E,$AF,$A8,$73,$7F,$0A,$F3,$BE,$8E
	.dc.b	$69,$CD,$7C,$09,$CF,$8E,$F6,$BF,$88,$AE,$5B,$18,$7F,$8A,$FD,$3B
	.dc.b	$E4,$B3,$87,$CA,$8C,$3F,$32,$FA,$39,$F8,$1F,$E3,$BF,$50,$F6,$39
	.dc.b	$C6,$1E,$BE,$86,$2B,$A3,$76,$FE,$3F,$8F,$F1,$76,$C7,$F2,$8F,$A8
	.dc.b	$7D,$47,$2B,$FB,$BE,$6B,$B3,$FA,$2C,$3E,$9D,$F0,$7F,$CC,$2B,$53
	.dc.b	$F9,$F3,$51,$9A,$E3,$5C,$27,$AC,$62,$75,$18,$39,$4B,$EA,$33,$9F
	.dc.b	$F3,$89,$7C,$4F,$92,$7E,$4C,$E6,$31,$C5,$FD,$12,$B6,$79,$6B,$7D
	.dc.b	$9F,$CF,$A3,$37,$FD,$1D,$F0,$57,$1B,$91,$E4,$F9,$FB,$5F,$D2,$BE
	.dc.b	$A1,$F0,$BE,$E3,$FA,$9B,$8A,$F5,$C6,$0D,$3F,$AA,$8D,$4F,$87,$3B
	.dc.b	$97,$F5,$29,$B5,$FE,$5A,$6D,$7F,$8F,$A9,$F0,$BD,$6B,$F6,$BC,$9F
	.dc.b	$C5,$EB,$A7,$D3,$7F,$66,$8C,$63,$FE,$3C,$AF,$B9,$39,$F2,$FE,$BB
	.dc.b	$63,$E9,$B6,$FF,$B4,$DF,$29,$FE,$7C,$7E,$17,$A7,$58,$7F,$6A,$F1
	.dc.b	$BF,$0B,$7F,$DC,$E5,$F5,$0F,$87,$FE,$F3,$5A,$BF,$F2,$7D,$6A,$73
	.dc.b	$CB,$FE,$00,$BE,$18,$FF,$12,$C3,$83,$FD,$3E,$F8,$39,$AE,$2F,$F8
	.dc.b	$33,$BF,$FC,$0B,$8F,$96,$7C,$3F,$43,$FD,$67,$EA,$39,$E9,$FD,$67
	.dc.b	$24,$E6,$B7,$8F,$F6,$BB,$DD,$AF,$73,$9A,$F6,$ED,$EE,$E0,$B3,$E9
	.dc.b	$6D,$FF,$85,$2D,$5F,$E1,$C8,$FE,$D5,$F0,$1F,$DA,$FE,$A1,$F5,$1F
	.dc.b	$E9,$CB,$DE,$3F,$CD,$2F,$62,$B6,$FA,$F0,$FF,$3D,$C6,$23,$FD,$4B
	.dc.b	$87,$5C,$3A,$56,$9F,$F5,$24,$66,$79,$5D,$78,$9D,$E7,$8C,$66,$3C
	.dc.b	$77,$DF,$0F,$3C,$6C,$FF,$59,$DF,$38,$CC,$F9,$67,$8D,$F4,$78,$1E
	.dc.b	$08,$E9,$87,$9E,$39,$9A,$AE,$75,$EA,$8C,$4E,$DE,$51,$D6,$BC,$10
	.dc.b	$BF,$25,$6E,$3C,$67,$C2,$3C,$66,$0F,$C1,$3B,$79,$1F,$EC,$D3,$EA
	.dc.b	$96,$30,$E0,$FF,$6B,$FD,$3F,$F1,$EF,$77,$83,$63,$FD,$AC,$D7,$FD
	.dc.b	$B6,$D7,$8F,$C9,$7F,$9F,$DD,$B5,$DB,$5D,$F0,$76,$8E,$18,$FE,$17
	.dc.b	$EA,$1F,$49,$FE,$EA,$8C,$37,$4E,$38,$7D,$4C,$58,$72,$9F,$F7,$8E
	.dc.b	$1C,$AB,$F1,$0D,$78,$4E,$DF,$F2,$45,$72,$C7,$78,$77,$9F,$F2,$65
	.dc.b	$F1,$C7,$F3,$57,$B9,$E7,$7C,$18,$60,$93,$EA,$72,$F2,$C3,$C5,$0A
	.dc.b	$E9,$3B,$8F,$C8,$DF,$87,$1F,$83,$3C,$70,$E7,$1F,$03,$1F,$45,$F3
	.dc.b	$C7,$AE,$1F,$06,$FF,$D1,$ED,$DB,$85,$7F,$8B,$B2,$F0,$E3,$88,$FF
	.dc.b	$4C,$D6,$DF,$87,$8F,$F9,$BD,$FF,$4E,$D7,$07,$FC,$E5,$F5,$1F,$EB
	.dc.b	$F7,$FC,$72,$D7,$FD,$46,$DC,$7F,$D4,$3F,$4D,$FD,$32,$77,$D5,$D6
	.dc.b	$73,$EB,$DD,$3C,$17,$FE,$5D,$BD,$9D,$BE,$6F,$4F,$89,$AB,$14,$62
	.dc.b	$5E,$06,$A7,$35,$CC,$87,$FA,$EA,$76,$E3,$1F,$4F,$89,$CC,$E5,$87
	.dc.b	$83,$EA,$3F,$9D,$D7,$F9,$69,$AF,$FB,$11,$B5,$F0,$BC,$EE,$1D,$72
	.dc.b	$B8,$3E,$78,$C4,$78,$FE,$A3,$D4,$62,$BF,$D9,$7F,$3F,$FE,$63,$9F
	.dc.b	$F3,$4B,$5F,$F6,$9B,$63,$E8,$F6,$7C,$68,$C6,$FF,$75,$B5,$FE,$AD
	.dc.b	$6A,$D7,$FC,$70,$F6,$F6,$7E,$67,$A3,$23,$07,$F9,$EB,$E0,$7F,$CA
	.dc.b	$8D,$F2,$BE,$AC,$62,$3F,$DC,$FC,$FF,$E6,$67,$44,$3F,$DD,$71,$BD
	.dc.b	$3F,$DD,$33,$B9,$CD,$7E,$02,$1F,$0D,$4E,$F1,$F8,$4F,$43,$EF,$D7
	.dc.b	$FB,$D9,$B7,$FE,$F4,$6E,$BF,$9F,$FA,$8F,$F3,$25,$7F,$A5,$27,$2F
	.dc.b	$FB,$45,$A9,$B1,$C2,$3B,$72,$7E,$52,$1F,$1D,$4F,$24,$67,$1F,$CC
	.dc.b	$FD,$47,$E9,$6D,$93,$2F,$D9,$4E,$63,$13,$FE,$E3,$C7,$E3,$56,$BC
	.dc.b	$5E,$64,$9F,$F5,$6C,$E9,$FF,$75,$7D,$47,$FA,$B9,$FF,$58,$B5,$F0
	.dc.b	$D7,$F0,$18,$C5,$F8,$25,$87,$8E,$B5,$E8,$7E,$CA,$FF,$CD,$75,$D2
	.dc.b	$3F,$6B,$F5,$0F,$87,$FD,$BF,$D4,$3E,$2A,$92,$F8,$1F,$EC,$48,$7D
	.dc.b	$27,$02,$46,$7E,$A3,$F8,$0F,$C5,$3D,$DA,$49,$FF,$66,$FC,$73,$D2
	.dc.b	$C4,$7B,$80,$AE,$4F,$FB,$51,$AE,$C4,$F3,$75,$5F,$C4,$A1,$F5,$14
	.dc.b	$EF,$C5,$F2,$7E,$A3,$E4,$5F,$81,$FF,$3F,$BF,$EE,$26,$BD,$DD,$A6
	.dc.b	$9A,$8E,$9E,$DF,$F2,$28,$CF,$89,$49,$5D,$7E,$82,$BD,$62,$DF,$96
	.dc.b	$7E,$5F,$CA,$CA,$F8,$4F,$5E,$2C,$47,$C3,$D5,$38,$BF,$75,$5E,$6F
	.dc.b	$E9,$59,$3E,$A3,$CC,$67,$1F,$A3,$F4,$EF,$1B,$F9,$DB,$F8,$08,$FA
	.dc.b	$B8,$FE,$46,$31,$1F,$4D,$E0,$EB,$0F,$EB,$11,$88,$D7,$27,$43,$AD
	.dc.b	$F6,$FA,$89,$3C,$4F,$1B,$E2,$EF,$CC,$77,$8B,$8F,$8E,$7F,$A8,$FD
	.dc.b	$43,$E5,$E9,$2F,$A0,$FA,$7F,$E0,$D7,$FA,$BA,$E1,$3A,$6B,$F8,$F6
	.dc.b	$C7,$EC,$67,$8E,$9F,$DD,$67,$2F,$F7,$78,$7D,$47,$F6,$46,$DF,$5C
	.dc.b	$31,$EC,$48,$F2,$CF,$F8,$B9,$BD,$F9,$7F,$41,$19,$6C,$F5,$9E,$DD
	.dc.b	$7D,$FA,$E5,$74,$8C,$BD,$A3,$2F,$48,$FE,$F3,$F4,$EF,$A8,$A4,$9F
	.dc.b	$93,$E9,$FE,$8A,$75,$1F,$31,$AF,$ED,$5B,$BF,$F1,$79,$E1,$5F,$E0
	.dc.b	$67,$FC,$19,$5F,$BB,$6C,$FF,$1D,$AE,$3B,$7F,$88,$65,$D7,$C7,$AE
	.dc.b	$61,$E4,$8E,$86,$1E,$91,$87,$D3,$D3,$E1,$FA,$89,$18,$21,$E3,$72
	.dc.b	$F9,$63,$FC,$3B,$F5,$1F,$C0,$89,$E4,$63,$BF,$2F,$FA,$C3,$BB,$FC
	.dc.b	$71,$F4,$BF,$EE,$3C,$3F,$D7,$F1,$8F,$86,$D4,$7F,$20,$C3,$FA,$87
	.dc.b	$C0,$FE,$EA,$3F,$91,$DF,$17,$FD,$31,$7C,$5E,$66,$AF,$9A,$BE,$73
	.dc.b	$B8,$E2,$6C,$E2,$2B,$B5,$F8,$63,$96,$1D,$25,$F4,$D4,$93,$CA,$1F
	.dc.b	$51,$88,$EE,$3E,$9D,$F5,$1F,$DF,$31,$ED,$18,$F8,$BE,$58,$FA,$A3
	.dc.b	$17,$C9,$B5,$C2,$F9,$67,$F3,$AF,$84,$62,$FF,$A5,$DF,$0A,$FF,$35
	.dc.b	$36,$3E,$6D,$70,$C3,$E1,$CE,$67,$D1,$7C,$DF,$8D,$5B,$7C,$91,$99
	.dc.b	$E1,$97,$C6,$9E,$73,$92,$5F,$BA,$32,$C7,$9B,$96,$93,$1F,$81,$19
	.dc.b	$6C,$67,$1F,$6D,$79,$EF,$C5,$39,$3E,$1C,$78,$9F,$AB,$1A,$AE,$64
	.dc.b	$AE,$77,$DD,$B5,$CC,$F8,$B1,$97,$AB,$EB,$9E,$51,$F3,$8C,$4F,$28
	.dc.b	$CC,$7C,$C3,$27,$23,$B5,$79,$67,$81,$23,$53,$F9,$79,$D4,$62,$74
	.dc.b	$F9,$8F,$7E,$3F,$66,$BC,$38,$FD,$88,$C5,$70,$8C,$57,$90,$E5,$5E
	.dc.b	$8B,$EF,$18,$AE,$18,$FD,$1C,$39,$CF,$8C,$87,$BC,$F9,$58,$7E,$3C
	.dc.b	$EE,$E2,$7E,$B6,$1E,$DA,$E7,$3D,$CE,$78,$78,$AF,$95,$F0,$C3,$F4
	.dc.b	$95,$E9,$9C,$E3,$F5,$E3,$91,$23,$C3,$8F,$C4,$BE,$53,$F6,$4F,$25
	.dc.b	$7C,$3C,$3C,$13,$FE,$6F,$9F,$C2,$9F,$ED,$73,$F8,$07,$8E,$F8,$AC
	.dc.b	$47,$DD,$35,$8F,$F8,$FE,$3F,$63,$3B,$9F,$E1,$3F,$51,$FE,$47,$E7
	.dc.b	$CD,$49,$FE,$09,$5B,$CF,$F8,$C5,$EE,$73,$5B,$9F,$E1,$F5,$BD,$3F
	.dc.b	$DD,$97,$B9,$CD,$6E,$7F,$83,$D6,$EF,$FC,$AB,$7B,$9C,$D6,$E7,$F8
	.dc.b	$7D,$6D,$56,$DC,$CE,$DF,$8A,$FF,$32,$8D,$46,$0F,$49,$2F,$DB,$D7
	.dc.b	$FE,$90,$7C,$F1,$9C,$3C,$6F,$C8,$9D,$46,$27,$8E,$3E,$1B,$EF,$97
	.dc.b	$D1,$C3,$84,$E7,$0E,$31,$E8,$3E,$9F,$A9,$87,$DB,$F5,$17,$0F,$F5
	.dc.b	$6D,$FA,$E3,$15,$23,$DD,$EB,$5B,$FB,$AD,$CB,$EC,$B7,$3F,$AA,$D9
	.dc.b	$FA,$2D,$F9,$17,$C5,$7F,$A6,$70,$FF,$24,$46,$3E,$1B,$FD,$6B,$E0
	.dc.b	$7F,$AD,$A3,$7E,$A6,$2B,$FD,$1F,$1B,$78,$DF,$3C,$7F,$D2,$93,$96
	.dc.b	$C6,$6F,$A2,$C3,$A4,$AB,$CB,$F5,$18,$7C,$FF,$51,$6D,$36,$30,$DC
	.dc.b	$7C,$6D,$BF,$F5,$07,$D4,$7F,$BB,$A7,$FC,$CD,$C8,$FA,$8F,$F4,$0C
	.dc.b	$7F,$97,$A7,$55,$FE,$AD,$AD,$6F,$E6,$84,$EB,$5F,$F6,$DD,$6A,$F9
	.dc.b	$7B,$DE,$C8,$C5,$F8,$A3,$95,$F1,$AF,$F5,$CF,$C8,$3E,$07,$FD,$E9
	.dc.b	$1C,$8F,$9B,$E7,$7C,$1B,$F8,$4D,$8C,$9F,$4A,$32,$D3,$EA,$FB,$3F
	.dc.b	$24,$F9,$5F,$51,$63,$24,$CB,$EB,$C6,$5F,$F6,$47,$D3,$7F,$B8,$67
	.dc.b	$FD,$1D,$AB,$EA,$3E,$E4,$7F,$B4,$A9,$5B,$7F,$DA,$D5,$BE,$6F,$95
	.dc.b	$39,$8F,$C7,$DF,$2A,$FF,$6D,$CE,$6F,$FD,$53,$39,$9E,$BE,$47,$23
	.dc.b	$F1,$3B,$7C,$9F,$4F,$17,$F9,$4F,$A8,$FC,$41,$F8,$9F,$06,$C6,$49
	.dc.b	$87,$E5,$A3,$2F,$E6,$23,$F2,$1F,$3F,$F9,$16,$9F,$A0,$2C,$6D,$FD
	.dc.b	$1C,$7B,$BE,$43,$F0,$5F,$51,$F9,$F9,$FF,$55,$EA,$9C,$F6,$EE,$7F
	.dc.b	$2B,$5B,$C3,$E8,$D6,$A7,$8B,$B8,$FC,$B4,$FC,$5B,$E9,$5F,$BC,$97
	.dc.b	$D4,$7E,$AD,$B1,$9F,$8A,$C4,$C8,$E0,$EE,$31,$3F,$BC,$1F,$51,$CF
	.dc.b	$E3,$BA,$36,$FC,$0D,$9E,$0D,$E3,$4F,$E1,$FE,$99,$F5,$1F,$6B,$EA
	.dc.b	$1F,$21,$FA,$F9,$CD,$7F,$B7,$79,$55,$A8,$FF,$6A,$CE,$B2,$63,$D9
	.dc.b	$5F,$0C,$3A,$63,$DD,$B8,$72,$8F,$E2,$7F,$50,$FA,$87,$D4,$6B,$1E
	.dc.b	$B7,$F9,$FF,$A8,$73,$7F,$05,$F3,$FF,$2D,$18,$9F,$F7,$4C,$7E,$67
	.dc.b	$C4,$FF,$59,$C7,$FB,$B1,$71,$FF,$0D,$73,$87,$60,$7F,$80,$E3,$F8
	.dc.b	$97,$D3,$BE,$53,$87,$BD,$86,$E1,$FE,$F8,$97,$C1,$7F,$19,$FA,$87
	.dc.b	$B1,$CA,$30,$F4,$F4,$71,$7C,$5B,$9F,$8C,$61,$FC,$59,$C3,$79,$BE
	.dc.b	$84,$3E,$A1,$F5,$1D,$70,$F5,$D6,$BD,$37,$6F,$F4,$29,$7D,$33,$E1
	.dc.b	$3F,$98,$56,$A7,$89,$F1,$0D,$C6,$AB,$95,$71,$9D,$E5,$FD,$62,$73
	.dc.b	$5B,$9C,$CF,$F5,$19,$E3,$F5,$19,$8F,$81,$2F,$88,$F3,$4F,$D1,$8E
	.dc.b	$26,$4E,$E1,$D4,$F2,$9E,$3C,$5F,$D2,$EB,$84,$FD,$F7,$84,$66,$64
	.dc.b	$74,$C5,$F4,$18,$BF,$A9,$F5,$0F,$86,$F8,$27,$F5,$77,$17,$C6,$30
	.dc.b	$69,$F8,$31,$A9,$F1,$5D,$DB,$ED,$CD,$AF,$E7,$CD,$AF,$DD,$D3,$E1
	.dc.b	$3A,$D7,$CC,$E4,$FE,$35,$87,$EC,$E3,$EE,$FD,$37,$F6,$B9,$CC,$7D
	.dc.b	$BB,$4E,$7C,$BF,$B6,$D9,$F8,$4D,$C7,$F2,$97,$C9,$F1,$F5,$E7,$1F
	.dc.b	$C9,$F8,$9F,$89,$9F,$EF,$9F,$50,$F8,$8F,$EF,$93,$9A,$FE,$F7,$39
	.dc.b	$CB,$F0,$B3,$97,$F6,$78,$27,$58,$7E,$96,$B4,$FF,$9E,$63,$37,$E2
	.dc.b	$E4,$EE,$7F,$81,$7B,$FF,$C0,$55,$FE,$02,$8F,$1F,$D0,$FF,$89,$29
	.dc.b	$F5,$1D,$2B,$F8,$DC,$E6,$31,$8A,$31,$7C,$B3,$FE,$EF,$5A,$DB,$DF
	.dc.b	$C3,$F0,$1B,$3E,$E7,$0F,$E9,$31,$FE,$23,$F0,$3F,$C4,$BE,$A1,$F5
	.dc.b	$1F,$DF,$F4,$FA,$F7,$BB,$5E,$EB,$58,$7F,$33,$C3,$8D,$7F,$A7,$EB
	.dc.b	$4F,$F8,$8A,$75,$5F,$EA,$09,$CC,$FF,$3A,$9C,$BC,$45,$74,$75,$7C
	.dc.b	$1F,$49,$AB,$F1,$C6,$63,$CE,$64,$F1,$39,$BF,$0E,$1E,$E9,$D1,$FE
	.dc.b	$1F,$8C,$E1,$E4,$8C,$1A,$9E,$18,$79,$63,$C6,$F9,$E3,$38,$7B,$E7
	.dc.b	$85,$73,$77,$1E,$87,$C2,$60,$F3,$98,$3D,$75,$D6,$79,$47,$B8,$FF
	.dc.b	$65,$94,$F0,$18,$BC,$FD,$3F,$F1,$FF,$84,$BD,$5F,$E8,$8A,$FF,$69
	.dc.b	$CE,$5B,$19,$7F,$DB,$50,$F9,$2F,$EE,$71,$C1,$AE,$9A,$FF,$92,$63
	.dc.b	$87,$0F,$8A,$78,$E5,$F7,$7E,$A1,$F4,$9F,$E4,$F8,$C3,$75,$E5,$5C
	.dc.b	$1B,$7F,$8A,$BE,$38,$DB,$E0,$FF,$6E,$C7,$C2,$FD,$E6,$E3,$DA,$38
	.dc.b	$1F,$94,$8C,$E1,$DA,$3F,$0B,$7D,$9B,$7D,$95,$70,$7C,$18,$F9,$DF
	.dc.b	$EF,$F5,$C7,$0F,$04,$78,$E7,$4F,$26,$CE,$F0,$FF,$39,$36,$7A,$E3
	.dc.b	$E7,$AD,$E3,$E3,$AD,$57,$3A,$F6,$BF,$DF,$E3,$D5,$1F,$DF,$1B,$BF
	.dc.b	$BF,$1F,$2C,$E6,$FF,$D0,$4D,$BF,$ED,$B9,$7B,$D7,$D4,$7F,$B0,$5F
	.dc.b	$EE,$2D,$7F,$C1,$CD,$9F,$F0,$5F,$D3,$7F,$4D,$9D,$F5,$F6,$CB,$FD
	.dc.b	$21,$3F,$E6,$B9,$FE,$F6,$D7,$9C,$ED,$FF,$15,$DB,$E6,$71,$1E,$46
	.dc.b	$C6,$63,$9F,$81,$89,$91,$E8,$78,$DF,$F7,$48,$C1,$E2,$F0,$F2,$70
	.dc.b	$7F,$C5,$5F,$51,$FC,$DD,$FF,$08,$B5,$CF,$57,$4F,$33,$CB,$39,$D7
	.dc.b	$FC,$5A,$DD,$7F,$D6,$0D,$BF,$F2,$3E,$2B,$E3,$87,$9B,$EA,$25,$78
	.dc.b	$23,$E0,$FC,$EF,$F9,$C2,$3F,$C4,$6D,$7F,$C8,$0D,$CB,$E9,$F6,$F0
	.dc.b	$F5,$A5,$70,$7E,$1C,$62,$3F,$C9,$19,$7F,$97,$9F,$AB,$E8,$66,$FF
	.dc.b	$C8,$FF,$11,$FE,$5B,$9F,$F1,$9B,$5F,$F2,$83,$63,$FC,$9B,$CB,$FE
	.dc.b	$70,$9D,$F9,$3F,$6A,$B5,$C1,$FE,$D3,$6C,$FF,$9A,$AB,$83,$D2,$BF
	.dc.b	$03,$0F,$87,$A7,$38,$FC,$2F,$A1,$F8,$3A,$FF,$31,$37,$1F,$F2,$FB
	.dc.b	$71,$FF,$71,$7D,$47,$F9,$BA,$BF,$C9,$B3,$97,$FC,$F6,$D4,$D8,$E1
	.dc.b	$1D,$B9,$7F,$2B,$0F,$8F,$A7,$92,$33,$8F,$E6,$BE,$A3,$F4,$D7,$FE
	.dc.b	$77,$6E,$5F,$E7,$4C,$3E,$1C,$A7,$39,$7F,$BE,$5B,$1F,$E9,$06,$A9
	.dc.b	$E3,$F5,$1F,$EB,$88,$FF,$34,$B5,$FF,$43,$36,$FF,$D0,$75,$CA,$3F
	.dc.b	$CD,$53,$C3,$C1,$FD,$8C,$62,$3F,$D1,$DF,$50,$F8,$8F,$DA,$FD,$43
	.dc.b	$E2,$E9,$2B,$91,$FE,$7A,$87,$D1,$F0,$24,$67,$EA,$3F,$7F,$F1,$5F
	.dc.b	$EA,$48,$FF,$40,$B5,$FF,$4F,$37,$5F,$F4,$DF,$C7,$BD,$0E,$3E,$EC
	.dc.b	$BE,$2F,$F0,$A9,$D4,$60,$FF,$46,$AF,$A8,$D1,$C6,$3F,$D5,$9F,$51
	.dc.b	$F3,$8F,$F3,$D3,$FE,$94,$6B,$AF,$4B,$F8,$BC,$6F,$DA,$A6,$23,$3E
	.dc.b	$2C,$AE,$4F,$0F,$A1,$C4,$78,$22,$E9,$FE,$B8,$C1,$5F,$1B,$E5,$65
	.dc.b	$FC,$67,$C3,$C7,$88,$F8,$F1,$6F,$94,$DE,$9F,$F6,$CD,$30,$7D,$47
	.dc.b	$A8,$D4,$7F,$B7,$BE,$A1,$E3,$FF,$34,$7E,$DC,$7F,$AD,$9A,$FF,$B4
	.dc.b	$DD,$B9,$8E,$1C,$B2,$30,$71,$FA,$8E,$84,$7A,$77,$D7,$A4,$5D,$3F
	.dc.b	$DB,$78,$7E,$DF,$EA,$1F,$31,$FD,$21,$B7,$FE,$CA,$6B,$FE,$E7,$FA
	.dc.b	$8F,$DD,$DF,$FB,$92,$72,$DA,$E4,$7F,$52,$9D,$ED,$F2,$2B,$6F,$CA
	.dc.b	$87,$CB,$EC,$E3,$87,$FB,$6A,$3F,$C6,$0D,$F5,$FA,$1D,$7C,$1B,$19
	.dc.b	$21,$E7,$7C,$5D,$7E,$3E,$B9,$3D,$63,$2F,$78,$CB,$EF,$8F,$ED,$7F
	.dc.b	$4E,$FA,$8A,$49,$FF,$77,$FA,$7F,$C3,$27,$51,$FE,$F2,$6A,$DD,$AF
	.dc.b	$E2,$A7,$31,$89,$0A,$E9,$7F,$DF,$1F,$F0,$14,$EE,$75,$F3,$1E,$37
	.dc.b	$B6,$1F,$96,$AF,$ED,$11,$E6,$EA,$E4,$73,$EB,$91,$92,$1C,$8F,$1C
	.dc.b	$F2,$70,$79,$7A,$BC,$9F,$51,$23,$E1,$C7,$9C,$C3,$C2,$53,$98,$F7
	.dc.b	$E5,$EB,$FA,$8F,$E6,$A4,$EE,$FF,$0D,$BF,$C2,$23,$51,$FE,$97,$EE
	.dc.b	$FD,$9F,$D3,$7F,$B8,$6F,$FD,$7B,$18,$F8,$6D,$47,$CC,$C3,$FA,$5F
	.dc.b	$C0,$AB,$53,$C0,$FA,$37,$CD,$FF,$60,$C6,$9A,$73,$C3,$AA,$9E,$15
	.dc.b	$D6,$39,$E1,$A9,$CD,$FF,$0A,$C3,$C6,$FA,$BE,$9F,$44,$BF,$21,$D7
	.dc.b	$EA,$30,$F8,$87,$D3,$BE,$A3,$FB,$B6,$7E,$19,$CC,$63,$0E,$D1,$D2
	.dc.b	$B8,$7C,$5D,$AD,$4F,$E4,$A7,$2F,$7C,$D9,$7C,$09,$CE,$3D,$DB,$8F
	.dc.b	$F2,$66,$DF,$CC,$AE,$18,$76,$8F,$2E,$3E,$13,$DF,$8F,$68,$C1,$EF
	.dc.b	$BE,$6F,$03,$C9,$7D,$2B,$78,$F8,$F0,$ED,$87,$3B,$E5,$1E,$1C,$39
	.dc.b	$9D,$27,$55,$C6,$F9,$E3,$EB,$BE,$78,$F7,$C7,$9D,$69,$F7,$14,$F3
	.dc.b	$5F,$37,$CE,$67,$1F,$0D,$F4,$8C,$3E,$53,$0F,$62,$4F,$33,$EF,$63
	.dc.b	$E8,$C7,$CD,$87,$23,$E5,$E3,$E7,$AE,$13,$AB,$F1,$9D,$1C,$46,$A3
	.dc.b	$D6,$64,$FC,$26,$1E,$5B,$EB,$1D,$CD,$1F,$A3,$BF,$05,$F4,$24,$74
	.dc.b	$BE,$CF,$C6,$8C,$BC,$4E,$91,$AC,$3D,$D1,$E2,$BC,$D7,$E7,$23,$C3
	.dc.b	$8F,$CC,$9F,$2C,$63,$1F,$43,$F0,$30,$F0,$D6,$E3,$17,$E2,$8C,$4F
	.dc.b	$CE,$C7,$D7,$5E,$13,$DF,$8F,$91,$F5,$19,$9F,$44,$F0,$8C,$92,$78
	.dc.b	$57,$6B,$F8,$31,$8C,$3C,$F8,$72,$BF,$D1,$9F,$E1,$23,$F4,$45,$B7
	.dc.b	$94,$F8,$BD,$4F,$9C,$D4,$65,$FD,$63,$6B,$F8,$47,$D4,$7F,$91,$79
	.dc.b	$FF,$83,$4E,$73,$FD,$E3,$6F,$F8,$33,$55,$71,$7A,$E1,$C8,$8F,$F1
	.dc.b	$98,$D6,$1F,$C2,$6B,$96,$3E,$AB,$DC,$F9,$3A,$FF,$B3,$3B,$8C,$E3
	.dc.b	$E4,$7F,$93,$E1,$D2,$31,$8F,$88,$95,$CF,$2E,$31,$EA,$BA,$FA,$B0
	.dc.b	$EF,$7C,$AF,$84,$F2,$8F,$F2,$1E,$1D,$E3,$91,$ED,$AE,$2E,$2B,$89
	.dc.b	$E9,$7E,$17,$AB,$A7,$E4,$FD,$45,$BF,$F5,$36,$5E,$C8,$C5,$48,$F7
	.dc.b	$FA,$F6,$FE,$63,$70,$FB,$6D,$CB,$CE,$D9,$FF,$07,$E3,$FE,$B8,$BF
	.dc.b	$93,$87,$A3,$E4,$5F,$15,$FE,$97,$9F,$EE,$E7,$F8,$5A,$33,$F0,$DF
	.dc.b	$E1,$9E,$9F,$F0,$D4,$70,$F5,$71,$5B,$B5,$E3,$25,$EE,$BA,$C6,$2B
	.dc.b	$71,$89,$D4,$FF,$85,$5F,$0C,$66,$5F,$4F,$D6,$7D,$3F,$51,$67,$35
	.dc.b	$B8,$C5,$65,$B1,$93,$C7,$6F,$A7,$FF,$76,$47,$9B,$8D,$F5,$1F,$E0
	.dc.b	$B8,$DC,$EA,$BF,$C5,$55,$AE,$1F,$81,$1C,$FD,$FF,$85,$39,$BE,$31
	.dc.b	$8C,$3A,$57,$F8,$C3,$0E,$9B,$FF,$90,$5B,$7F,$E4,$8F,$90,$7C,$17
	.dc.b	$DB,$BE,$18,$FC,$1F,$43,$E1,$36,$FE,$43,$67,$E6,$6F,$D1,$FA,$55
	.dc.b	$C2,$BE,$77,$B5,$F2,$CF,$99,$F5,$16,$32,$4A,$E7,$19,$7F,$C9,$3F
	.dc.b	$4B,$FE,$E3,$9F,$BF,$AB,$EA,$3F,$1D,$5A,$F2,$7F,$0D,$39,$E7,$F9
	.dc.b	$38,$71,$AF,$F2,$B4,$E7,$0F,$AF,$39,$8F,$CA,$61,$D1,$FF,$34,$46
	.dc.b	$67,$FD,$76,$DA,$6F,$F2,$BD,$3E,$F9,$7E,$27,$EA,$26,$1F,$8F,$F0
	.dc.b	$6E,$1F,$22,$6E,$5F,$A2,$E0,$FE,$67,$7F,$95,$9F,$C9,$3F,$93,$F9
	.dc.b	$FA,$FE,$65,$B7,$F4,$9B,$7C,$DF,$83,$E7,$FE,$23,$EA,$3F,$4B,$5F
	.dc.b	$32,$31,$DB,$AC,$3F,$2D,$5A,$C3,$E0,$D6,$A7,$E3,$BB,$BF,$CB,$D7
	.dc.b	$C7,$9F,$8B,$3F,$AB,$97,$D4,$53,$1F,$15,$6B,$F6,$D3,$A6,$CE,$9A
	.dc.b	$6B,$EA,$3A,$7C,$6F,$56,$C6,$4E,$71,$96,$9F,$B4,$FA,$77,$D3,$FF
	.dc.b	$0B,$E4,$7D,$43,$E4,$7F,$5F,$5A,$BF,$E1,$B5,$A9,$CD,$F1,$BE,$15
	.dc.b	$D6,$31,$3A,$8C,$1F,$C5,$BE,$A1,$F5,$0F,$A8,$D6,$DF,$EA,$6F,$A8
	.dc.b	$72,$FF,$00,$F3,$FE,$64,$FF,$AF,$A7,$38,$FF,$0E,$9C,$CF,$F7,$9F
	.dc.b	$6F,$F8,$3F,$8E,$E4,$B1,$87,$F8,$5F,$D3,$BE,$53,$38,$FF,$18,$8C
	.dc.b	$3F,$C6,$AD,$7C,$3E,$07,$F8,$9F,$D4,$3D,$9C,$5F,$F2,$08,$C3,$CF
	.dc.b	$D3,$CE,$9F,$EC,$2C,$BF,$87,$76,$FF,$B4,$D6,$3E,$4A,$97,$FC,$AB
	.dc.b	$EA,$1F,$51,$8C,$7F,$D9,$7E,$9B,$B3,$FA,$25,$BE,$9B,$FD,$C4,$FF
	.dc.b	$32,$9D,$3F,$D1,$8D,$7C,$17,$F2,$C8,$CD,$FF,$B6,$67,$27,$1B,$EB
	.dc.b	$7D,$27,$C3,$1C,$4C,$FD,$45,$AF,$E8,$23,$E1,$FC,$B1,$F3,$2B,$FD
	.dc.b	$C7,$39,$3C,$91,$9D,$FF,$DE,$D1,$99,$F2,$4E,$63,$FD,$FF,$07,$53
	.dc.b	$63,$13,$B4,$6E,$BF,$A5,$D7,$03,$27,$C9,$F9,$FC,$C7,$F5,$0F,$A8
	.dc.b	$7C,$A7,$C2,$30,$5A,$FE,$BF,$1A,$7E,$76,$77,$0F,$F8,$03,$37,$2F
	.dc.b	$35,$7D,$7B,$E6,$F2,$7E,$7F,$C3,$62,$D5,$ED,$7F,$21,$E4,$FD,$4F
	.dc.b	$A6,$FE,$CB,$CF,$FC,$5F,$1F,$5E,$69,$CD,$6E,$7F,$BA,$D6,$EF,$E5
	.dc.b	$5E,$E7,$35,$C1,$F5,$F4,$DA,$F1,$C7,$D2,$F1,$7F,$13,$97,$D1,$A7
	.dc.b	$D4,$3E,$23,$FB,$CD,$6A,$FE,$9D,$6A,$73,$C7,$FD,$F2,$F8,$57,$F8
	.dc.b	$2A,$F8,$63,$FE,$75,$9C,$BE,$23,$FB,$D5,$72,$AF,$F0,$77,$7F,$F7
	.dc.b	$39,$FE,$AB,$3E,$2F,$A0,$FE,$CB,$F5,$1D,$27,$FA,$CC,$62,$75,$7F
	.dc.b	$DB,$AB,$58,$AB,$53,$E5,$E0,$F6,$F0,$FB,$9B,$19,$26,$3E,$73,$F3
	.dc.b	$3F,$07,$F9,$4F,$A6,$7D,$47,$CB,$BF,$F0,$FD,$7E,$62,$B5,$39,$BB
	.dc.b	$3F,$E9,$B9,$AF,$F7,$2A,$FD,$15,$F4,$8D,$07,$0E,$EB,$8B,$DC,$D5
	.dc.b	$F7,$69,$CA,$78,$E5,$F8,$78,$C9,$EC,$2B,$98,$EF,$1D,$5E,$37,$E6
	.dc.b	$AF,$F5,$7E,$1E,$AB,$E1,$18,$BE,$75,$B7,$C5,$1E,$5B,$EB,$0B,$E5
	.dc.b	$5B,$8F,$24,$F8,$30,$F7,$C6,$CE,$AE,$2B,$8B,$63,$FD,$86,$7C,$52
	.dc.b	$C6,$1C,$1F,$C5,$3E,$9F,$F9,$07,$A9,$C1,$B1,$FE,$CC,$6B,$FE,$D3
	.dc.b	$6B,$C7,$E4,$BF,$7E,$ED,$AE,$DA,$EF,$93,$B4,$70,$C7,$EE,$FD,$43
	.dc.b	$E9,$3F,$93,$C6,$1B,$AF,$1C,$3E,$CE,$2C,$39,$4F,$F2,$EC,$39,$57
	.dc.b	$E3,$9A,$F0,$9D,$BF,$CB,$AB,$96,$3B,$C3,$BC,$FF,$32,$BE,$38,$FE
	.dc.b	$4A,$F7,$3C,$EF,$83,$0C,$12,$7D,$4E,$5E,$58,$78,$A1,$5D,$27,$71
	.dc.b	$F9,$6B,$F0,$E3,$F0,$67,$8E,$1C,$E3,$E0,$63,$E8,$BE,$78,$F5,$C3
	.dc.b	$E0,$DF,$F3,$36,$E9,$C2,$BF,$D4,$39,$78,$71,$C4,$7F,$01,$AD,$BF
	.dc.b	$0F,$1F,$EB,$78,$7F,$3D,$BE,$13,$FD,$63,$EA,$3F,$D8,$2F,$FA,$B1
	.dc.b	$AF,$F0,$B6,$CF,$F4,$5F,$A6,$70,$F4,$CF,$FB,$3D,$69,$FE,$D3,$2F
	.dc.b	$5F,$58,$7F,$66,$8C,$3D,$FA,$B2,$AD,$F3,$78,$6C,$79,$63,$74,$F0
	.dc.b	$75,$12,$74,$7F,$B1,$A7,$4E,$0F,$E4,$CB,$C4,$E6,$71,$8F,$E5,$5F
	.dc.b	$51,$FC,$CD,$FF,$67,$B5,$CF,$47,$8B,$CC,$F4,$CE,$77,$FE,$D4,$DD
	.dc.b	$BF,$C1,$6D,$BF,$85,$5C,$2F,$D7,$F5,$12,$7B,$C7,$C4,$F9,$FF,$F3
	.dc.b	$44,$7F,$B8,$1A,$FF,$3F,$6E,$3F,$57,$B7,$87,$AD,$2B,$83,$F2,$23
	.dc.b	$11,$F0,$BD,$1F,$96,$62,$BF,$C0,$5F,$11,$FE,$59,$9F,$F7,$7B,$5F
	.dc.b	$E9,$CD,$8F,$F0,$47,$3F,$F9,$B2,$77,$E9,$7D,$CA,$D6,$FF,$E2,$C6
	.dc.b	$CF,$F8,$A2,$B8,$3D,$2B,$F0,$70,$F8,$8A,$73,$8F,$C3,$FA,$1F,$84
	.dc.b	$AF,$EC,$0D,$C7,$FC,$34,$DC,$7F,$C7,$7F,$51,$FE,$6B,$AF,$F1,$BD
	.dc.b	$6B,$97,$99,$23,$35,$C3,$CA,$C4,$7F,$6D,$B7,$C7,$FE,$61,$B1,$C0
	.dc.b	$FF,$21,$E3,$E1,$7F,$C8,$B1,$F0,$17,$D4,$7E,$A9,$E0,$63,$B7,$A6
	.dc.b	$9F,$A4,$9D,$4F,$F7,$D6,$E9,$FE,$37,$6E,$7E,$EA,$FD,$5F,$D4,$7F
	.dc.b	$AC,$AF,$AB,$5F,$F0,$33,$6F,$96,$1C,$3D,$09,$5C,$1F,$DA,$FD,$43
	.dc.b	$E1,$BF,$6F,$F5,$0F,$8B,$A4,$BE,$07,$B2,$1F,$49,$C0,$91,$9F,$A8
	.dc.b	$FE,$03,$F1,$5F,$E9,$68,$F8,$6D,$7F,$C3,$8D,$D3,$FC,$AD,$F1,$EF
	.dc.b	$4B,$8F,$BB,$2F,$8B,$FC,$32,$75,$18,$3E,$7A,$FA,$8D,$1C,$63,$FC
	.dc.b	$55,$F5,$1F,$31,$FF,$3E,$BF,$E7,$67,$EB,$B5,$DF,$0F,$F1,$A8,$E1
	.dc.b	$07,$58,$FF,$9B,$5B,$C9,$FE,$8E,$76,$7F,$B0,$A7,$AF,$89,$CC,$E9
	.dc.b	$F4,$36,$32,$43,$CE,$FC,$AF,$96,$91,$F2,$9F,$27,$1E,$23,$E5,$C5
	.dc.b	$C7,$C9,$37,$1F,$27,$D4,$7A,$CC,$5F,$FA,$7B,$E9,$DE,$37,$F3,$87
	.dc.b	$EE,$C7,$E6,$9A,$FF,$94,$DD,$B9,$8E,$1C,$92,$30,$71,$FA,$8E,$84
	.dc.b	$7A,$78,$35,$E9,$17,$3F,$F4,$C6,$1F,$AF,$97,$D4,$3E,$63,$FA,$6B
	.dc.b	$63,$F5,$ED,$7F,$CD,$1F,$51,$FC,$15,$FF,$35,$46,$6F,$FD,$48,$DD
	.dc.b	$7F,$D8,$AD,$9F,$F6,$0D,$71,$AD,$DF,$F5,$C8,$7C,$B5,$3C,$71,$9F
	.dc.b	$62,$5F,$F0,$2A,$FF,$13,$37,$BF,$2F,$9A,$6E,$1F,$EB,$7A,$FF,$71
	.dc.b	$4F,$1E,$AF,$07,$54,$8C,$10,$F7,$B9,$7E,$0C,$7F,$72,$FA,$87,$D4
	.dc.b	$52,$4F,$F1,$6F,$4B,$F3,$B3,$A8,$FE,$32,$D7,$FD,$24,$DC,$7F,$D9
	.dc.b	$93,$C2,$BF,$BF,$BF,$E0,$2A,$FF,$4C,$36,$7F,$DA,$55,$C7,$7F,$F7
	.dc.b	$9C,$FE,$96,$B7,$D1,$C8,$E7,$D3,$23,$CD,$19,$7D,$11,$E0,$E8,$F0
	.dc.b	$FD,$44,$9F,$64,$65,$F7,$46,$5F,$2C,$7F,$85,$BE,$A3,$F7,$A4,$ED
	.dc.b	$C7,$C3,$E5,$FF,$58,$F6,$AE,$7D,$37,$FB,$5D,$FF,$58,$46,$54,$65
	.dc.b	$FF,$47,$46,$6F,$FA,$0F,$C0,$A3,$85,$6E,$31,$5D,$BE,$1B,$C2,$D3
	.dc.b	$B9,$E3,$BE,$77,$CA,$77,$3E,$26,$DF,$59,$F1,$61,$8F,$A7,$97,$D0
	.dc.b	$F4,$BC,$3E,$A3,$31,$EA,$3E,$17,$D3,$BE,$A3,$FB,$96,$9E,$89,$4E
	.dc.b	$7D,$3E,$99,$B1,$F2,$CE,$6B,$A1,$C3,$0E,$79,$FC,$B9,$D6,$3D,$70
	.dc.b	$E5,$7E,$79,$CD,$6E,$F8,$CF,$99,$AF,$B0,$E0,$E2,$78,$C6,$F1,$F3
	.dc.b	$E3,$E6,$AD,$12,$33,$3C,$98,$FA,$AB,$9B,$DC,$C4,$F1,$6C,$F2,$69
	.dc.b	$31,$F9,$55,$CB,$1F,$7C,$E7,$1F,$75,$F8,$2F,$E2,$63,$E7,$8C,$4E
	.dc.b	$AB,$E6,$DF,$2B,$EB,$8F,$8D,$B7,$E3,$6D,$77,$3E,$21,$CB,$0E,$CE
	.dc.b	$4F,$B1,$3C,$70,$F0,$57,$22,$3F,$9E,$AE,$31,$8A,$EA,$FA,$9F,$49
	.dc.b	$89,$E6,$71,$72,$78,$0F,$16,$1E,$D8,$DC,$71,$21,$F4,$E7,$C0,$F1
	.dc.b	$6C,$F1,$8D,$E3,$F6,$8F,$A5,$7E,$8B,$EB,$3C,$5F,$23,$F0,$E7,$86
	.dc.b	$3F,$90,$BE,$93,$B3,$AC,$66,$FA,$46,$31,$F9,$4D,$3D,$B1,$A7,$91
	.dc.b	$2B,$AD,$F8,$2B,$E2,$C6,$2F,$EF,$61,$E8,$8C,$4F,$CA,$AF,$6D,$DB
	.dc.b	$DB,$FC,$7B,$0F,$04,$F5,$8F,$C1,$1B,$AE,$F3,$F8,$48,$FE,$44,$7D
	.dc.b	$52,$CB,$C9,$7C,$5E,$A7,$F1,$A6,$A3,$2F,$EA,$DB,$9F,$F0,$DF,$A7
	.dc.b	$FF,$23,$73,$FF,$01,$9C,$EB,$FC,$1D,$B7,$FC,$59,$AA,$B8,$BD,$70
	.dc.b	$E4,$47,$F8,$9C,$6B,$0F,$D9,$D7,$2C,$7D,$57,$B9,$F2,$75,$B2,$C6
	.dc.b	$5E,$3A,$B9,$7F,$49,$CB,$A4,$E6,$3C,$44,$C3,$9D,$F0,$8F,$46,$1D
	.dc.b	$5D,$57,$F4,$FB,$E7,$1E,$FC,$3F,$9D,$46,$23,$F9,$9E,$5F,$0A,$31
	.dc.b	$97,$AA,$7D,$FE,$AF,$5C,$FE,$77,$D4,$5C,$BB,$4D,$AF,$54,$D7,$E6
	.dc.b	$45,$C3,$FD,$57,$A7,$8B,$D6,$C4,$E5,$B1,$93,$CF,$19,$69,$9A,$F2
	.dc.b	$BF,$D8,$3E,$45,$F1,$5F,$E9,$78,$FF,$5B,$46,$3E,$17,$E3,$78,$1F
	.dc.b	$1E,$37,$EA,$62,$BF,$BD,$46,$DE,$37,$CF,$05,$F1,$86,$1C,$65,$7D
	.dc.b	$2F,$8C,$E6,$3C,$BF,$51,$6D,$F5,$16,$30,$DB,$F0,$B7,$0F,$0C,$79
	.dc.b	$AB,$8B,$FE,$CC,$FA,$7F,$F7,$53,$FE,$D5,$E4,$7D,$47,$F8,$2E,$35
	.dc.b	$EF,$79,$E7,$3C,$7E,$9C,$58,$72,$7B,$9F,$B4,$C3,$93,$83,$77,$CE
	.dc.b	$33,$B7,$E4,$E7,$38,$F8,$0F,$F7,$47,$C8,$3E,$05,$C7,$E1,$CB,$C1
	.dc.b	$87,$CD,$F3,$BE,$36,$7F,$1D,$B8,$FD,$46,$E5,$E6,$E1,$FC,$0C,$62
	.dc.b	$FE,$B7,$B3,$F3,$CF,$A1,$F5,$17,$0F,$BB,$17,$0E,$5C,$2A,$C5,$FA
	.dc.b	$FE,$9F,$FD,$C9,$87,$FC,$01,$87,$D4,$7E,$15,$D7,$91,$F8,$37,$3D
	.dc.b	$1F,$36,$F8,$E5,$F3,$27,$38,$7F,$92,$5E,$2F,$5C,$3A,$3F,$E4,$9A
	.dc.b	$E3,$3A,$AF,$AB,$C5,$F2,$FD,$3E,$D9,$7E,$37,$EA,$26,$3F,$9A,$F0
	.dc.b	$2E,$1E,$E9,$B8,$FE,$6E,$6D,$7E,$34,$FC,$B7,$CF,$FE,$5D,$A7,$E9
	.dc.b	$0B,$7F,$21,$B5,$F1,$FE,$43,$F1,$DF,$51,$FA,$69,$F7,$1E,$B8,$CF
	.dc.b	$77,$C5,$6C,$6A,$FF,$CD,$9B,$27,$31,$F9,$18,$C4,$EE,$11,$87,$F3
	.dc.b	$3D,$8C,$3E,$5C,$74,$BF,$DA,$2F,$A8,$A6,$3E,$2A,$E3,$FC,$2A,$74
	.dc.b	$D9,$D3,$4D,$7D,$47,$4F,$8C,$EA,$D8,$C9,$CE,$32,$D3,$D5,$F5,$0F
	.dc.b	$A7,$FE,$19,$CA,$FA,$87,$C8,$7E,$DE,$B5,$7F,$AC,$9C,$BC,$AF,$8C
	.dc.b	$63,$3F,$DC,$57,$07,$F1,$31,$9E,$2F,$DC,$4E,$5B,$5F,$C7,$7E,$A1
	.dc.b	$F5,$0F,$A8,$F1,$4F,$F0,$9F,$A8,$73,$FF,$0B,$F4,$3E,$7E,$89,$CE
	.dc.b	$BF,$5E,$73,$CC,$F6,$FF,$89,$2E,$5B,$18,$7F,$8B,$7D,$3B,$E4,$B3
	.dc.b	$87,$E0,$63,$0F,$E0,$EF,$A3,$9F,$82,$FE,$3D,$F5,$0F,$5F,$9C,$61
	.dc.b	$EB,$E6,$62,$BA,$37,$6F,$E4,$18,$FF,$18,$9C,$9F,$91,$FA,$87,$D4
	.dc.b	$72,$8F,$CA,$79,$8E,$CF,$E8,$D0,$FA,$67,$C2,$7F,$31,$AD,$4F,$E8
	.dc.b	$4D,$46,$6B,$8D,$70,$9E,$B1,$89,$D4,$60,$E5,$2F,$A8,$CE,$7F,$CE
	.dc.b	$65,$F1,$5E,$49,$F9,$33,$98,$C7,$17,$F4,$5A,$D9,$E5,$AD,$F6,$7F
	.dc.b	$3F,$8C,$DF,$F4,$87,$C1,$5C,$6E,$47,$93,$E7,$ED,$7F,$4B,$FA,$87
	.dc.b	$C3,$7B,$8F,$EA,$8E,$2B,$D7,$18,$34,$FF,$01,$8D,$4F,$87,$3B,$97
	.dc.b	$F5,$39,$B5,$FE,$F3,$6D,$7F,$8D,$69,$F0,$9D,$6B,$F8,$67,$27,$F1
	.dc.b	$9A,$E9,$F4,$DF,$D9,$E3,$18,$FF,$8C,$6B,$EE,$4E,$7C,$BF,$AE,$D8
	.dc.b	$FA,$6D,$BF,$ED,$57,$CA,$51,$E3,$E9,$D6,$1F,$DA,$FC,$6F,$C2,$CF
	.dc.b	$F1,$CF,$A8,$7C,$3F,$F7,$99,$CD,$7F,$90,$ED,$5A,$C3,$F9,$35,$69
	.dc.b	$EF,$3A,$BE,$5C,$BF,$E0,$5B,$E1,$8F,$F7,$FC,$38,$3E,$66,$DB,$BF
	.dc.b	$FB,$CE,$3E,$AA,$57,$06,$FC,$FF,$F6,$5F,$A8,$C4,$A3,$19,$7A,$DB
	.dc.b	$7E,$86,$DF,$8F,$6F,$81,$C1,$F0,$5B,$5E,$E7,$0F,$F4,$59,$F9,$BF
	.dc.b	$09,$F3,$BE,$A1,$F5,$1E,$7C,$D8,$7F,$88,$EF,$62,$B6,$FC,$18,$FF
	.dc.b	$38,$46,$23,$E0,$5F,$5B,$E9,$5A,$7F,$D3,$91,$99,$E5,$85,$78,$9D
	.dc.b	$E7,$8C,$66,$3C,$77,$DF,$0F,$3C,$E6,$57,$CE,$73,$1C,$CF,$F5,$8C
	.dc.b	$67,$0F,$24,$60,$D1,$F6,$1C,$BE,$4B,$F0,$46,$23,$C7,$87,$99,$E2
	.dc.b	$60,$F6,$98,$3C,$A7,$6B,$F2,$D7,$C9,$8E,$07,$87,$0F,$66,$1E,$76
	.dc.b	$C7,$FB,$14,$FA,$A5,$8C,$38,$3F,$25,$F4,$FF,$C7,$BD,$BE,$0D,$8F
	.dc.b	$F6,$73,$5F,$F6,$AB,$5E,$3F,$27,$FE,$80,$76,$D7,$6D,$77,$C9,$DA
	.dc.b	$38,$63,$F8,$5F,$A8,$7D,$27,$E9,$63,$0D,$D7,$8E,$1F,$4B,$16,$1C
	.dc.b	$A7,$F5,$98,$72,$AF,$C4,$35,$E1,$3B,$7F,$85,$D7,$2C,$77,$87,$79
	.dc.b	$FE,$1F,$7C,$71,$FC,$D5,$EE,$79,$DF,$06,$18,$24,$FA,$9C,$BC,$B0
	.dc.b	$F1,$42,$BA,$4E,$E3,$F1,$F7,$E1,$C7,$E0,$CF,$1C,$39,$C7,$C0,$C7
	.dc.b	$D1,$7C,$F1,$EB,$87,$C1,$BF,$F4,$83,$74,$E1,$5F,$E4,$5C,$BC,$38
	.dc.b	$E2,$3F,$D3,$55,$B7,$E1,$E3,$FC,$67,$0F,$F4,$D5,$F0,$9F,$E6,$BF
	.dc.b	$51,$FE,$C0,$7F,$CA,$0D,$7F,$D4,$8D,$9F,$F5,$17,$D4,$7F,$4C,$9D
	.dc.b	$F5,$75,$CF,$F9,$34,$FF,$97,$67,$FC,$B4,$D7,$9C,$ED,$F0,$FC,$DD
	.dc.b	$58,$23,$0D,$8E,$7D,$F8,$99,$1E,$87,$95,$A7,$33,$FE,$BD,$8F,$F0
	.dc.b	$AB,$E4,$F0,$F3,$97,$FA,$F7,$EA,$3F,$9C,$4F,$F9,$C5,$AF,$FB,$15
	.dc.b	$B1,$F0,$BC,$DF,$44,$63,$4F,$E8,$CD,$AF,$EC,$2D,$58,$7F,$91,$5E
	.dc.b	$DD,$1E,$8F,$A8,$91,$83,$FD,$05,$F3,$FF,$E6,$66,$FB,$1F,$1E,$31
	.dc.b	$1E,$DE,$CC,$4C,$8D,$9F,$4A,$30,$FC,$3F,$43,$E3,$98,$D7,$FD,$BD
	.dc.b	$F1,$3F,$E5,$39,$FF,$49,$35,$FF,$73,$36,$3F,$DC,$9C,$BF,$E6,$69
	.dc.b	$DF,$93,$F6,$EB,$5B,$FF,$69,$6C,$FF,$81,$2B,$83,$D2,$BF,$05,$0F
	.dc.b	$88,$A7,$38,$FC,$37,$A1,$F8,$3A,$FF,$7C,$37,$1F,$F7,$B3,$71,$FE
	.dc.b	$F1,$F5,$1F,$E6,$5A,$FF,$56,$4E,$5F,$F0,$AB,$5F,$F7,$CF,$37,$E4
	.dc.b	$E7,$84,$76,$8C,$7C,$7C,$8E,$71,$BC,$7F,$35,$F5,$1F,$A5,$96,$4C
	.dc.b	$BF,$C1,$53,$98,$C4,$FF,$89,$B1,$F8,$AA,$75,$5C,$3C,$4E,$A7,$79
	.dc.b	$FF,$63,$7D,$47,$FA,$B5,$BE,$CF,$EA,$E3,$11,$E2,$AE,$4F,$0F,$43
	.dc.b	$44,$7F,$DA,$1F,$50,$F8,$7F,$DA,$7D,$43,$E2,$E9,$2B,$91,$FE,$D7
	.dc.b	$87,$D2,$70,$24,$67,$EA,$3F,$7D,$F1,$4F,$76,$92,$7F,$DC,$1F,$1F
	.dc.b	$FE,$31,$7F,$DC,$53,$B6,$CE,$DF,$61,$FE,$4E,$9E,$0E,$04,$F1,$71
	.dc.b	$0F,$A8,$FE,$20,$D8,$EA,$7F,$BC,$7E,$A3,$E8,$BF,$E5,$A7,$FC,$F6
	.dc.b	$E7,$DB,$FE,$31,$1A,$83,$4D,$BE,$36,$F3,$7E,$02,$37,$C9,$CB,$C6
	.dc.b	$91,$E1,$7A,$FD,$0E,$23,$94,$5A,$E5,$6A,$F9,$BF,$2D,$2F,$E4,$BE
	.dc.b	$6E,$2C,$47,$C5,$D5,$38,$C3,$9D,$5E,$5F,$F3,$06,$2F,$A8,$F7,$99
	.dc.b	$C7,$FC,$FB,$F4,$EF,$1B,$F9,$D3,$F8,$28,$FF,$40,$E3,$EF,$8C,$47
	.dc.b	$D5,$78,$3A,$C3,$FC,$DD,$18,$8D,$72,$74,$3A,$DF,$6F,$A8,$93,$C4
	.dc.b	$F1,$BE,$2E,$FC,$C7,$78,$B8,$F8,$E7,$FA,$87,$D4,$3E,$62,$92,$FA
	.dc.b	$0F,$A8,$FE,$07,$7F,$66,$B8,$4E,$9A,$FE,$45,$B1,$F7,$E7,$8E,$9F
	.dc.b	$E8,$B9,$CB,$FE,$8D,$87,$CB,$FF,$63,$6D,$F5,$C3,$1E,$BC,$8F,$2C
	.dc.b	$FF,$8B,$1B,$DF,$97,$F4,$51,$96,$CF,$59,$ED,$EF,$77,$EC,$95,$D2
	.dc.b	$32,$F6,$8C,$BD,$23,$FB,$C7,$D3,$BE,$A2,$92,$7E,$4F,$A9,$FA,$59
	.dc.b	$D4,$7C,$C6,$BF,$B7,$6E,$FF,$A5,$9E,$15,$FE,$05,$7F,$C1,$51,$9F
	.dc.b	$96,$ED,$7F,$BF,$9F,$ED,$8D,$EB,$E0,$71,$EC,$91,$E4,$8C,$BE,$68
	.dc.b	$ED,$D5,$DF,$EA,$24,$FA,$A3,$2F,$B2,$32,$F8,$63,$FC,$35,$F5,$1F
	.dc.b	$C0,$C9,$DF,$8F,$03,$2F,$FA,$B3,$BF,$FD,$99,$F4,$FF,$EE,$0C,$3F
	.dc.b	$D7,$71,$8F,$87,$D4,$7F,$BA,$B0,$FE,$99,$F0,$5F,$B8,$8F,$F7,$6D
	.dc.b	$F1,$7F,$D4,$F7,$C5,$E6,$6A,$F9,$AB,$E7,$3B,$8E,$26,$CE,$22,$BB
	.dc.b	$5F,$86,$39,$61,$D2,$5F,$4D,$49,$3C,$A1,$F5,$18,$8E,$E3,$E9,$DF
	.dc.b	$51,$FD,$DF,$1E,$D1,$8F,$8C,$E5,$8F,$AA,$31,$7C,$9B,$5C,$2F,$96
	.dc.b	$7F,$3A,$F8,$46,$2C,$78,$5C,$37,$0E,$EF,$C2,$C3,$AD,$F0,$8D,$1E
	.dc.b	$1A,$E3,$97,$C2,$8E,$11,$D2,$B8,$C7,$81,$AF,$CD,$C3,$CB,$87,$02
	.dc.b	$47,$C6,$BE,$8F,$78,$CE,$3E,$F7,$CF,$19,$24,$65,$E5,$5D,$9C,$E3
	.dc.b	$EB,$C3,$C7,$87,$6B,$E0,$49,$D6,$3E,$EC,$39,$4E,$70,$F6,$4E,$5F
	.dc.b	$2D,$70,$C6,$C7,$E6,$2F,$8C,$E6,$3C,$66,$4D,$9F,$3A,$FB,$3F,$1C
	.dc.b	$CE,$3F,$28,$F4,$D7,$48,$F1,$DF,$2C,$7D,$98,$77,$C3,$9E,$1C,$2F
	.dc.b	$9E,$3F,$5C,$F5,$3E,$43,$C5,$5C,$4F,$C6,$3C,$30,$ED,$8F,$86,$36
	.dc.b	$78,$8F,$A2,$6C,$E9,$8F,$91,$D6,$3E,$4C,$3C,$2D,$BE,$B3,$C6,$3E
	.dc.b	$1D,$71,$8E,$E4,$C7,$D8,$FA,$5F,$8D,$7D,$31,$F5,$4F,$81,$AF,$CD
	.dc.b	$38,$47,$C3,$C3,$F1,$B8,$78,$E7,$D3,$8F,$BF,$0E,$B5,$AC,$39,$C6
	.dc.b	$2B,$E0,$9D,$27,$D6,$FE,$5C,$FD,$61,$FA,$C2,$C7,$F8,$EB,$E2,$F3
	.dc.b	$1F,$2C,$CF,$92,$8F,$DB,$36,$3F,$86,$FD,$3B,$AF,$FC,$DF,$9F,$F0
	.dc.b	$CA,$D3,$6F,$89,$FC,$82,$B8,$78,$58,$C3,$F8,$75,$F0,$CF,$F5,$0D
	.dc.b	$8F,$E1,$75,$C1,$DD,$F1,$C3,$B4,$79,$31,$F1,$E1,$CA,$B4,$D9,$D6
	.dc.b	$38,$7A,$D7,$07,$F8,$56,$3E,$15,$8F,$87,$1E,$E7,$F3,$77,$18,$75
	.dc.b	$8C,$5F,$3A,$FE,$65,$7C,$ED,$EB,$65,$7D,$45,$8C,$37,$2F,$91,$1B
	.dc.b	$F6,$31,$87,$D2,$6E,$5E,$36,$E7,$FC,$AF,$14,$62,$BE,$5B,$7E,$49
	.dc.b	$F1,$7F,$E9,$3C,$BF,$C8,$11,$8F,$87,$FF,$55,$F8,$1F,$EA,$C8,$DF
	.dc.b	$AB,$8A,$FE,$7F,$1B,$78,$DF,$3C,$7F,$D1,$73,$96,$C6,$6F,$A2,$C3
	.dc.b	$A4,$AB,$CB,$F5,$18,$7C,$FF,$51,$6D,$36,$30,$DC,$7C,$6D,$BF,$F4
	.dc.b	$D7,$D3,$FF,$B8,$E7,$FC,$BD,$C8,$FA,$8F,$F3,$CC,$7F,$95,$A7,$55
	.dc.b	$FE,$A4,$AD,$70,$79,$A1,$3A,$D7,$FD,$A3,$5A,$BE,$5F,$01,$EC,$8C
	.dc.b	$5F,$8A,$39,$5F,$1A,$FE,$D3,$F2,$0F,$83,$FB,$31,$C8,$F9,$BE,$7F
	.dc.b	$C1,$BF,$84,$D8,$C9,$F4,$A3,$2D,$3E,$AF,$B3,$F2,$4F,$95,$F5,$16
	.dc.b	$32,$4C,$BE,$BC,$65,$FF,$61,$7D,$37,$FB,$6E,$7F,$D0,$DA,$BE,$A3
	.dc.b	$EE,$47,$F7,$B9,$D5,$7F,$B3,$2B,$5C,$DF,$26,$73,$1F,$8E,$BE,$35
	.dc.b	$FE,$0A,$9C,$DF,$FA,$82,$73,$29,$CB,$BF,$23,$F1,$3B,$7C,$9F,$53
	.dc.b	$B4,$7E,$1F,$EA,$26,$3F,$96,$EF,$B8,$7E,$52,$6E,$7F,$94,$9B,$5F
	.dc.b	$9E,$E8,$3F,$91,$FA,$0F,$C9,$34,$FD,$09,$70,$FC,$FB,$6F,$FD,$AD
	.dc.b	$A3,$D0,$FC,$27,$D4,$7E,$86,$BF,$D4,$B3,$97,$F2,$B1,$9B,$FA,$35
	.dc.b	$C3,$B7,$E5,$36,$B8,$B8,$78,$F0,$DA,$FD,$98,$FA,$8F,$D6,$36,$7F
	.dc.b	$80,$AF,$8B,$CC,$C9,$CB,$C2,$31,$3D,$3E,$A3,$9F,$C6,$F4,$6D,$F4
	.dc.b	$6C,$EE,$3B,$CF,$EA,$FE,$9D,$F1,$5F,$6B,$EA,$1F,$21,$98,$C3,$6F
	.dc.b	$FD,$AD,$D0,$AD,$C7,$FB,$3A,$F9,$BA,$8C,$E2,$E2,$FB,$6F,$F9,$62
	.dc.b	$B6,$E6,$BF,$CA,$91,$89,$DF,$D4,$3E,$A1,$F5,$1E,$3C,$38,$1F,$E6
	.dc.b	$2F,$A8,$73,$FF,$07,$F3,$FF,$33,$18,$CF,$FD,$C7,$1F,$9C,$F1,$5E
	.dc.b	$DF,$F0,$D7,$E9,$4F,$FB,$A1,$CB,$79,$3F,$89,$7D,$3B,$E5,$38,$5E
	.dc.b	$1B,$5F,$EF,$09,$7C,$17,$F1,$9F,$A8,$7A,$DC,$A3,$0F,$4F,$47,$17
	.dc.b	$C5,$BB,$FB,$86,$1F,$C5,$9C,$37,$8F,$E6,$C3,$EA,$1F,$51,$D7,$0F
	.dc.b	$F4,$3D,$6B,$D3,$76,$7F,$42,$97,$D3,$3E,$13,$F9,$85,$6A,$78,$9F
	.dc.b	$E9,$03,$71,$AA,$E5,$5C,$67,$79,$7F,$58,$9C,$D6,$E7,$33,$FD,$46
	.dc.b	$78,$FD,$46,$63,$E0,$4B,$E2,$3C,$D3,$F4,$63,$89,$93,$B8,$75,$3C
	.dc.b	$A7,$8F,$17,$F4,$BA,$E1,$3F,$8A,$78,$46,$66,$47,$4C,$5F,$3F,$8B
	.dc.b	$FA,$9F,$50,$F8,$6F,$82,$7F,$57,$71,$7C,$63,$06,$9F,$83,$1A,$9F
	.dc.b	$15,$DD,$BE,$DC,$DA,$FF,$5C,$36,$BF,$92,$D3,$E1,$3A,$D7,$CC,$E4
	.dc.b	$FE,$35,$87,$EC,$23,$EE,$FD,$37,$F6,$B9,$CC,$7D,$BB,$4E,$7C,$AF
	.dc.b	$B6,$D9,$F8,$4D,$C7,$F2,$97,$C9,$F1,$F5,$E7,$1F,$C9,$F8,$9F,$89
	.dc.b	$9F,$EF,$9F,$50,$F8,$8F,$EF,$93,$9A,$FE,$F7,$39,$CB,$F0,$B3,$97
	.dc.b	$FD,$BF,$82,$75,$87,$E9,$6B,$4F,$F9,$AE,$33,$7E,$2E,$4E,$E7,$F8
	.dc.b	$17,$BF,$FC,$05,$5F,$E0,$28,$F1,$FD,$07,$F8,$92,$9F,$51,$D2,$98
	.dc.b	$AB,$53,$98,$C6,$7F,$DD,$A3,$13,$AC,$3D,$FB,$7C,$0E,$1B,$7E,$E6
	.dc.b	$E3,$EB,$9F,$F1,$03,$FE,$FE,$F8,$2F,$F7,$FF,$D4,$3E,$A3,$FD,$F5
	.dc.b	$A7,$D8,$C7,$F5,$77,$B1,$5B,$7F,$C3,$F8,$7F,$32,$8C,$40,$E0,$E2
	.dc.b	$BA,$47,$2B,$FF,$4C,$E1,$E2,$7C,$37,$CD,$CC,$EC,$EF,$86,$DF,$F5
	.dc.b	$2D,$57,$E5,$D7,$17,$27,$13,$33,$B7,$CB,$5E,$08,$C4,$2B,$A5,$FA
	.dc.b	$30,$F6,$CF,$03,$C6,$75,$30,$76,$30,$73,$3C,$51,$CE,$FD,$B3,$F1
	.dc.b	$70,$F4,$57,$C9,$38,$46,$70,$F2,$E1,$E4,$AD,$D7,$D5,$3F,$D8,$67
	.dc.b	$72,$C6,$1D,$1D,$7E,$9F,$F9,$17,$AB,$C1,$B1,$FE,$CC,$6B,$FE,$D2
	.dc.b	$6B,$C7,$E4,$BF,$B9,$3B,$6B,$B6,$BB,$E0,$ED,$1C,$32,$FC,$3F,$D4
	.dc.b	$3E,$93,$E0,$46,$1B,$AF,$1C,$3E,$D6,$2C,$39,$4F,$C7,$C3,$95,$7E
	.dc.b	$0D,$AF,$09,$DB,$F9,$5A,$E5,$8E,$F0,$EF,$3F,$99,$BE,$38,$FE,$72
	.dc.b	$F7,$3C,$EF,$83,$0C,$12,$7D,$4E,$5E,$58,$78,$A1,$5D,$27,$71,$F9
	.dc.b	$7B,$F0,$E3,$F0,$67,$8E,$1C,$E3,$E0,$63,$E8,$BE,$78,$F5,$C3,$E0
	.dc.b	$DF,$F7,$A6,$EB,$C2,$BF,$B3,$E5,$E1,$C7,$11,$FE,$05,$AD,$BF,$0F
	.dc.b	$1F,$CA,$61,$FE,$04,$BE,$13,$FB,$FF,$A8,$FF,$61,$3F,$DC,$9A,$FF
	.dc.b	$83,$DB,$3F,$E0,$CF,$A7,$FE,$9D,$3B,$EA,$EB,$9F,$E8,$67,$FC,$DD
	.dc.b	$3F,$DF,$1A,$F3,$9D,$BE,$1F,$99,$AB,$04,$61,$B1,$CF,$BF,$13,$23
	.dc.b	$D0,$F2,$BF,$EE,$B1,$83,$C5,$E1,$E4,$E0,$FF,$8A,$FE,$A3,$F9,$D3
	.dc.b	$FE,$12,$6B,$9E,$8E,$9E,$67,$9A,$73,$AF,$F8,$B9,$BA,$FF,$01,$6D
	.dc.b	$FF,$2B,$C7,$FC,$9A,$F5,$C3,$CD,$F5,$12,$30,$7F,$88,$BE,$77,$FC
	.dc.b	$E2,$DF,$5F,$E2,$C6,$23,$FC,$85,$7F,$E3,$87,$44,$3F,$C8,$D1,$BC
	.dc.b	$BE,$C7,$67,$2B,$F8,$DE,$84,$BE,$15,$FE,$48,$F8,$8F,$F2,$E4,$7F
	.dc.b	$8D,$1A,$FF,$94,$5B,$1F,$E4,$EE,$5F,$F3,$7C,$EF,$C9,$FB,$75,$AD
	.dc.b	$FF,$92,$36,$7F,$A7,$57,$07,$A5,$7E,$0A,$1F,$0F,$4E,$71,$F8,$6F
	.dc.b	$43,$F0,$75,$FE,$63,$6E,$3F,$E6,$06,$E3,$FC,$CF,$EA,$3F,$CD,$F5
	.dc.b	$FE,$4E,$9C,$BF,$D7,$DA,$9B,$1C,$23,$B7,$2F,$E5,$A1,$F2,$14,$F2
	.dc.b	$46,$71,$FC,$DF,$D4,$7E,$9A,$FF,$CF,$0D,$CB,$FC,$EB,$87,$C3,$94
	.dc.b	$E7,$2F,$E9,$4D,$8F,$EE,$8D,$53,$C7,$EA,$3F,$D7,$31,$FE,$6A,$6B
	.dc.b	$FE,$87,$6D,$FF,$A1,$2B,$94,$7F,$9A,$E7,$87,$9F,$FB,$28,$C4,$7F
	.dc.b	$A3,$FE,$A1,$F1,$1F,$B6,$FA,$87,$C2,$52,$57,$23,$FC,$F7,$0F,$A3
	.dc.b	$E0,$48,$CF,$D4,$7F,$00,$F8,$AF,$F5,$2C,$7F,$A0,$9A,$FF,$A7,$DB
	.dc.b	$AF,$FA,$73,$E3,$DE,$87,$1F,$76,$5F,$17,$F8,$5C,$EA,$30,$7F,$A3
	.dc.b	$97,$D4,$68,$E3,$1F,$EA,$DF,$A8,$F9,$C7,$F9,$ED,$FF,$4A,$B5,$D7
	.dc.b	$A5,$FC,$62,$37,$ED,$53,$11,$9F,$1A,$57,$27,$87,$D0,$E2,$3C,$11
	.dc.b	$74,$FF,$5C,$E0,$AF,$8D,$F2,$B2,$FE,$33,$E1,$E3,$C4,$7C,$78,$B7
	.dc.b	$CA,$6F,$4F,$F8,$57,$07,$D4,$7A,$4C,$E1,$FE,$41,$FA,$87,$8D,$FC
	.dc.b	$D1,$FB,$51,$FE,$B6,$6B,$FE,$D3,$76,$E6,$38,$72,$C8,$C1,$C7,$EA
	.dc.b	$3A,$11,$E9,$DF,$5E,$91,$74,$FF,$6D,$E1,$FB,$6F,$A8,$7C,$C7,$F4
	.dc.b	$86,$DF,$FB,$29,$AF,$FB,$9F,$EA,$3F,$75,$7F,$EE,$49,$CB,$6B,$91
	.dc.b	$FD,$4A,$77,$B7,$F8,$EE,$B6,$FF,$8F,$61,$F2,$FB,$38,$E1,$FE,$DA
	.dc.b	$8F,$F0,$F3,$7D,$7E,$87,$5F,$06,$C6,$48,$79,$DF,$17,$5F,$8F,$AE
	.dc.b	$4F,$58,$CB,$DE,$32,$FB,$E3,$FB,$5F,$D4,$3E,$A2,$92,$7F,$DD,$FE
	.dc.b	$9F,$F0,$B9,$D4,$7F,$BC,$9A,$B7,$6B,$FE,$5D,$9C,$C6,$27,$FD,$11
	.dc.b	$5C,$EF,$FB,$DB,$FE,$00,$9D,$4A,$31,$7C,$EB,$7D,$5C,$3A,$A5,$F9
	.dc.b	$A3,$2F,$A2,$32,$F9,$6B,$FC,$E1,$7F,$DB,$A3,$07,$93,$A7,$C7,$F5
	.dc.b	$12,$3E,$0C,$79,$8C,$3C,$0F,$34,$65,$FF,$52,$65,$EA,$FA,$8F,$E6
	.dc.b	$44,$ED,$FF,$0C,$BF,$C1,$63,$51,$FE,$83,$F7,$9D,$DF,$E8,$5F,$A7
	.dc.b	$FF,$70,$61,$FE,$BB,$8C,$7C,$2E,$DF,$F4,$56,$3F,$D2,$6F,$8D,$7F
	.dc.b	$AD,$6F,$8B,$FE,$9F,$AE,$2F,$33,$57,$CD,$FF,$5E,$5F,$39,$DC,$71
	.dc.b	$36,$71,$15,$DA,$FC,$30,$AE,$4F,$8F,$E9,$BB,$3C,$1B,$46,$FD,$46
	.dc.b	$1D,$FD,$3B,$EA,$3F,$BA,$E3,$DA,$31,$F1,$5E,$3C,$39,$63,$E5,$9C
	.dc.b	$BE,$90,$E2,$B9,$5F,$4C,$FE,$86,$3E,$08,$C5,$FF,$B7,$EF,$85,$7F
	.dc.b	$AE,$5B,$1E,$D3,$C1,$3C,$1E,$ED,$8D,$C7,$C1,$37,$8C,$7C,$D1,$AC
	.dc.b	$7D,$71,$8A,$E4,$4A,$F8,$18,$79,$DE,$58,$FB,$1C,$D6,$31,$F4,$5F
	.dc.b	$18,$F8,$98,$76,$BE,$11,$DE,$F8,$3E,$D3,$04,$8D,$9F,$80,$9E,$57
	.dc.b	$DA,$B8,$D7,$C2,$7C,$6F,$18,$D3,$4E,$D5,$D2,$32,$D8,$CC,$60,$8F
	.dc.b	$D5,$BF,$A1,$7E,$28,$C5,$F8,$B1,$F9,$98,$FC,$79,$F0,$1E,$53,$E1
	.dc.b	$B9,$C7,$C1,$19,$7D,$D1,$9C,$7C,$33,$97,$F1,$2F,$7C,$7E,$75,$70
	.dc.b	$8C,$61,$F1,$31,$F3,$47,$13,$CE,$79,$1C,$E3,$F8,$82,$5F,$BB,$1F
	.dc.b	$15,$70,$C3,$CF,$7C,$2B,$C3,$8F,$9A,$31,$8F,$8A,$FD,$E5,$BE,$38
	.dc.b	$74,$8F,$21,$2B,$8B,$F2,$31,$F9,$6F,$1C,$3B,$57,$19,$F3,$E1,$F8
	.dc.b	$0C,$3E,$5C,$62,$FC,$27,$3C,$7C,$EF,$DE,$3D,$98,$7C,$E9,$EF,$1B
	.dc.b	$BF,$95,$87,$DB,$81,$F8,$C2,$9F,$B9,$F2,$5F,$15,$88,$D1,$FB,$B8
	.dc.b	$CB,$FB,$26,$E5,$FC,$73,$E9,$FF,$C9,$7C,$BF,$C3,$67,$39,$F1,$C3
	.dc.b	$07,$F2,$5A,$DE,$1C,$E3,$2D,$3F,$8A,$46,$B0,$FE,$39,$7D,$27,$D1
	.dc.b	$5C,$27,$C5,$D5,$FD,$AD,$DC,$77,$8C,$61,$E3,$AE,$B1,$87,$99,$8C
	.dc.b	$7A,$4F,$C6,$C7,$C4,$F1,$C3,$E0,$7A,$9C,$6D,$4F,$A8,$C3,$FE,$9E
	.dc.b	$C3,$E0,$46,$2A,$47,$C3,$F5,$AD,$FC,$06,$E1,$F3,$5B,$8F,$8D,$B3
	.dc.b	$F7,$31,$FF,$58,$5F,$CD,$C3,$D9,$F2,$0F,$89,$FF,$47,$CF,$F7,$23
	.dc.b	$F0,$71,$9F,$85,$FF,$09,$74,$FF,$84,$E3,$87,$A9,$8A,$DD,$AF,$19
	.dc.b	$2F,$75,$D6,$31,$5B,$8C,$4E,$A7,$FC,$1A,$F8,$63,$32,$FA,$7E,$B3
	.dc.b	$E9,$FA,$8B,$39,$AD,$C6,$2B,$2D,$8C,$9E,$3B,$7D,$3F,$FB,$92,$3C
	.dc.b	$DC,$6F,$A8,$FF,$06,$C6,$E7,$55,$FE,$22,$AD,$70,$FC,$08,$E7,$EF
	.dc.b	$7C,$29,$CD,$F1,$8C,$61,$D2,$BF,$C4,$98,$74,$DF,$FC,$76,$DB,$FD
	.dc.b	$AF,$C7,$BE,$07,$ED,$DF,$0C,$7E,$0F,$9F,$F0,$9B,$7F,$21,$B3,$F3
	.dc.b	$37,$E8,$FD,$2A,$E1,$5F,$3B,$D9,$F9,$67,$CC,$FA,$8B,$19,$25,$73
	.dc.b	$8C,$BF,$E3,$DF,$A6,$FF,$6D,$CF,$DF,$D5,$F5,$1F,$8E,$AD,$79,$1F
	.dc.b	$86,$9C,$F3,$FC,$9C,$38,$D7,$F9,$3A,$73,$87,$D7,$9C,$C7,$E5,$30
	.dc.b	$E8,$FF,$98,$A3,$33,$FC,$69,$B4,$DF,$E5,$7A,$5D,$F2,$FC,$4F,$D4
	.dc.b	$4C,$3F,$1F,$E0,$5C,$3E,$44,$DC,$BF,$45,$C0,$AC,$5F,$E8,$8F,$C9
	.dc.b	$FC,$F7,$E5,$1A,$7E,$68,$B8,$7D,$36,$DF,$A7,$E9,$FF,$11,$F5,$1F
	.dc.b	$A5,$AF,$99,$18,$ED,$F8,$CD,$BF,$88,$D9,$DC,$60,$FC,$DD,$7C,$67
	.dc.b	$E2,$4F,$EA,$A5,$F5,$14,$C7,$C5,$5A,$FD,$AC,$E9,$B3,$A6,$9A,$FA
	.dc.b	$8E,$9F,$1B,$D5,$B1,$93,$9C,$65,$A7,$EC,$FE,$99,$F5,$1F,$C2,$B9
	.dc.b	$1F,$50,$F9,$0F,$D7,$56,$AF,$F8,$65,$6A,$73,$7C,$6F,$85,$75,$8C
	.dc.b	$4E,$A3,$07,$F1,$5F,$A8,$7D,$43,$EA,$35,$B7,$FA,$67,$EA,$1C,$BF
	.dc.b	$BF,$F3,$FE,$64,$FF,$6A,$9C,$E3,$FC,$36,$73,$3F,$DA,$BD,$CF,$E0
	.dc.b	$DE,$3B,$92,$C6,$1F,$E1,$5F,$4E,$F9,$4C,$E3,$FC,$5E,$30,$FF,$19
	.dc.b	$B5,$F0,$F8,$1F,$E2,$5F,$50,$F5,$B1,$7F,$C7,$E3,$0F,$3F,$4F,$3A
	.dc.b	$7F,$AD,$72,$FE,$1B,$DD,$FE,$CC,$58,$F9,$2A,$5F,$F2,$9F,$A8,$7D
	.dc.b	$46,$31,$FF,$5F,$7A,$6E,$CF,$E8,$76,$FA,$6F,$F6,$B3,$FC,$C6,$74
	.dc.b	$FF,$45,$35,$F0,$5F,$CA,$E3,$37,$FE,$D0,$9C,$9C,$6F,$AD,$F4,$9F
	.dc.b	$0C,$71,$33,$F5,$16,$BF,$A0,$0F,$87,$F2,$C7,$CC,$AF,$F6,$B4,$E4
	.dc.b	$F2,$46,$77,$FF,$77,$C6,$67,$C9,$39,$8F,$F1,$49,$99,$FF,$7D,$28
	.dc.b	$D4,$FF,$8C,$23,$47,$0A,$F5,$FC,$FF,$F4,$B8,$CC,$FF,$8D,$BE,$A1
	.dc.b	$F2,$9F,$00,$FE,$A0,$E2,$7E,$56,$76,$BF,$DE,$13,$72,$F1,$CD,$AF
	.dc.b	$F1,$FC,$DA,$47,$D7,$BE,$A4,$7A,$FC,$35,$9F,$F7,$E7,$2F,$F1,$7C
	.dc.b	$7A,$FD,$37,$F6,$38,$C5,$FB,$F2,$4E,$7C,$AF,$B2,$D9,$F4,$B7,$0E
	.dc.b	$B3,$F7,$BA,$71,$8F,$9A,$7F,$B2,$F0,$7F,$6F,$FA,$87,$C4,$7F,$73
	.dc.b	$F1,$FF,$13,$7B,$C1,$5A,$CF,$FC,$05,$5A,$7F,$CC,$73,$AB,$E5,$87
	.dc.b	$56,$CF,$08,$FE,$FD,$DF,$FD,$B3,$2F,$E9,$D1,$E1,$FA,$1F,$EB,$DF
	.dc.b	$51,$D6,$7F,$BD,$C6,$27,$57,$FD,$9E,$B5,$5F,$D5,$EB,$53,$E6,$DB
	.dc.b	$D9,$C3,$ED,$6C,$64,$94,$DB,$E3,$3F,$E7,$CF,$83,$FF,$3F,$FD,$43
	.dc.b	$EA,$3F,$CF,$FC,$1F,$53,$0F,$F1,$25,$FF,$A1,$2F,$75,$AC,$7E,$86
	.dc.b	$1C,$6B,$FD,$09,$5A,$7F,$C2,$F3,$AA,$FF,$44,$42,$74,$FF,$73,$9D
	.dc.b	$3C,$85,$75,$75,$7C,$1F,$21,$AB,$F2,$46,$63,$B1,$93,$C6,$E6,$FC
	.dc.b	$38,$7B,$E7,$47,$C5,$2B,$9B,$F1,$47,$81,$F2,$46,$6F,$DB,$5D,$30
	.dc.b	$F5,$DF,$3B,$DC,$62,$39,$57,$A1,$F0,$98,$3C,$E6,$0F,$5C,$F2,$C3
	.dc.b	$D5,$1F,$30,$F8,$65,$BC,$38,$3F,$D7,$7F,$4F,$FC,$7F,$D4,$E0,$D8
	.dc.b	$FA,$AD,$7F,$D7,$CD,$78,$FC,$97,$E2,$DD,$B5,$DB,$5D,$E9,$DA,$38
	.dc.b	$63,$F6,$FE,$A1,$F4,$9F,$EC,$E8,$C3,$74,$E3,$87,$D4,$C5,$87,$29
	.dc.b	$FF,$6A,$E1,$CA,$BF,$16,$D7,$84,$ED,$FF,$6E,$D7,$2C,$77,$87,$79
	.dc.b	$FF,$70,$DF,$1C,$7F,$21,$7B,$9E,$77,$C1,$86,$09,$3E,$A7,$2F,$2C
	.dc.b	$3C,$50,$AE,$93,$B8,$FC,$8D,$F8,$71,$F8,$33,$C7,$0E,$71,$F0,$31
	.dc.b	$F4,$5F,$3C,$7A,$E1,$F0,$6F,$F7,$ED,$D3,$85,$7F,$A7,$F2,$F0,$E3
	.dc.b	$88,$FE,$0D,$36,$F8,$E2,$9E,$18,$FA,$BE,$A3,$FD,$7D,$1F,$EA,$96
	.dc.b	$BC,$8E,$0E,$2B,$F8,$8F,$D3,$38,$7A,$56,$C5,$6F,$1C,$D3,$D7,$DE
	.dc.b	$1C,$9B,$3C,$AB,$85,$F8,$3E,$6F,$2D,$3C,$CE,$16,$0F,$03,$14,$AA
	.dc.b	$4A,$FF,$63,$56,$DE,$66,$3C,$4E,$90,$78,$47,$B7,$EA,$1C,$FE,$16
	.dc.b	$C7,$FB,$45,$AF,$C3,$AF,$F1,$DB,$99,$E0,$47,$C7,$E6,$57,$87,$D4
	.dc.b	$78,$8C,$6D,$FD,$03,$E7,$FF,$CD,$13,$FE,$DF,$6B,$F4,$5B,$1F,$5B
	.dc.b	$B3,$E3,$46,$37,$FA,$ED,$AF,$C2,$35,$F9,$73,$C3,$D6,$F8,$7E,$8C
	.dc.b	$9F,$04,$7D,$1F,$88,$FF,$2E,$47,$FB,$C5,$AF,$E0,$DB,$8F,$F5,$8E
	.dc.b	$6F,$F3,$7C,$EF,$D2,$E4,$47,$FD,$F5,$1A,$79,$47,$DF,$87,$C4,$53
	.dc.b	$84,$FE,$47,$D0,$FC,$03,$F9,$FB,$FE,$CF,$18,$8D,$72,$7E,$27,$EA
	.dc.b	$3F,$CD,$91,$BD,$FF,$37,$5B,$BE,$37,$F8,$F8,$C4,$FE,$82,$F8,$DF
	.dc.b	$43,$91,$FA,$F5,$F1,$FD,$CE,$57,$E1,$CF,$F3,$5F,$51,$FA,$BB,$E2
	.dc.b	$1C,$3F,$AD,$BF,$97,$87,$EC,$27,$39,$7F,$81,$1B,$19,$EB,$ED,$7D
	.dc.b	$3E,$A3,$FD,$66,$DF,$77,$F5,$51,$88,$FD,$D5,$F2,$7F,$84,$47,$F9
	.dc.b	$B5,$E5,$E8,$F0,$3F,$65,$F5,$0F,$86,$FD,$A7,$D4,$3E,$2E,$92,$F8
	.dc.b	$9E,$E8,$7D,$27,$02,$46,$7E,$A3,$F7,$DF,$13,$FE,$98,$8F,$8A,$D7
	.dc.b	$F8,$83,$78,$3F,$C4,$BF,$1E,$F4,$78,$FB,$D2,$F8,$BF,$C2,$67,$51
	.dc.b	$83,$E8,$AF,$A8,$D1,$C6,$3F,$8F,$7D,$47,$C5,$4F,$F9,$E9,$FB,$2D
	.dc.b	$77,$C9,$FC,$5E,$38,$19,$C7,$FC,$76,$DE,$5F,$E5,$6E,$BC,$3E,$04
	.dc.b	$BE,$BF,$43,$5F,$1C,$5B,$E7,$3F,$23,$E5,$A5,$7C,$87,$C5,$C7,$88
	.dc.b	$F9,$31,$6B,$C7,$37,$1F,$1F,$D4,$7A,$4C,$5F,$F4,$BF,$A7,$78,$DF
	.dc.b	$CC,$DF,$B7,$1F,$9A,$6B,$FD,$15,$DB,$98,$E1,$C9,$23,$07,$1F,$A8
	.dc.b	$E8,$47,$A7,$7D,$7A,$45,$CF,$FC,$CB,$87,$ED,$A7,$FB,$07,$D4,$3E
	.dc.b	$63,$FA,$43,$63,$F5,$ED,$7F,$AD,$7D,$47,$EF,$2F,$FC,$DF,$39,$6D
	.dc.b	$72,$3F,$A9,$4E,$F5,$FE,$C7,$5B,$7F,$B2,$C3,$E5,$B6,$71,$C3,$F7
	.dc.b	$F1,$FE,$24,$6F,$A9,$D0,$EB,$DF,$63,$24,$3C,$EF,$8B,$AB,$C7,$D5
	.dc.b	$27,$AC,$65,$EF,$19,$7D,$F1,$FD,$AF,$EA,$1F,$51,$49,$3F,$C5,$3D
	.dc.b	$3F,$D0,$4E,$A3,$F8,$BB,$5F,$F0,$13,$77,$FF,$4B,$CF,$0A,$FE,$F0
	.dc.b	$FF,$80,$EB,$A7,$AD,$D9,$FF,$05,$47,$F6,$86,$F4,$F0,$38,$F5,$48
	.dc.b	$F2,$46,$5F,$34,$76,$E9,$EF,$F5,$12,$7D,$51,$97,$D9,$19,$7C,$31
	.dc.b	$FE,$0D,$FA,$8F,$DF,$13,$BB,$1F,$11,$97,$FD,$65,$DD,$FE,$22,$FA
	.dc.b	$7F,$F6,$BE,$1F,$EA,$F8,$C7,$C2,$6A,$3F,$C6,$58,$7F,$3C,$F8,$15
	.dc.b	$6A,$78,$1F,$E3,$7B,$E6,$FF,$AB,$E3,$4D,$39,$E1,$D5,$4F,$0A,$EB
	.dc.b	$1C,$F0,$D4,$E6,$FF,$C7,$38,$78,$DF,$57,$D2,$E8,$97,$E4,$3A,$FD
	.dc.b	$46,$1F,$10,$FA,$77,$D4,$7F,$6C,$CF,$C3,$39,$8C,$61,$DA,$3A,$57
	.dc.b	$0F,$8D,$B5,$A9,$FF,$2C,$4E,$5E,$F9,$B2,$F8,$13,$9C,$7B,$B7,$1F
	.dc.b	$F2,$AB,$6F,$E6,$57,$0C,$3B,$47,$97,$1F,$09,$EF,$C7,$B4,$60,$F7
	.dc.b	$DF,$37,$81,$E4,$BE,$95,$BC,$7C,$78,$76,$C3,$9D,$F2,$8F,$0E,$1C
	.dc.b	$CE,$93,$AA,$E3,$7C,$F1,$F5,$DF,$3C,$7B,$E3,$CE,$B4,$FB,$8A,$79
	.dc.b	$AF,$9B,$E7,$33,$8F,$86,$FA,$46,$1F,$29,$87,$B1,$27,$99,$F7,$B1
	.dc.b	$F4,$63,$E6,$C3,$91,$F2,$F1,$F3,$D7,$09,$D5,$F8,$CE,$8E,$23,$51
	.dc.b	$EB,$32,$7E,$13,$0F,$2D,$F5,$8E,$E6,$8F,$D1,$DF,$82,$FA,$12,$3A
	.dc.b	$5F,$67,$E3,$46,$5E,$27,$48,$D6,$1E,$E8,$F1,$5E,$6B,$F3,$91,$E1
	.dc.b	$C7,$E6,$4F,$96,$31,$8F,$A1,$F8,$18,$78,$6B,$71,$8B,$F1,$46,$27
	.dc.b	$E7,$63,$EB,$AF,$09,$EF,$C7,$C8,$FA,$8C,$CF,$A2,$78,$46,$49,$3C
	.dc.b	$2B,$B5,$FC,$18,$C6,$1E,$7C,$39,$5F,$BC,$EB,$5F,$02,$3D,$53,$B8
	.dc.b	$C1,$EA,$71,$3F,$42,$31,$8F,$E5,$AF,$BB,$FE,$CC,$9F,$C2,$97,$07
	.dc.b	$94,$F8,$AD,$4F,$C2,$35,$19,$7F,$6E,$DC,$BF,$89,$FD,$3F,$F9,$17
	.dc.b	$C1,$FE,$23,$39,$CF,$F8,$3B,$6F,$F8,$8B,$55,$71,$7A,$E1,$C8,$8F
	.dc.b	$F2,$48,$D6,$1F,$C5,$2B,$94,$62,$FB,$DF,$F2,$D8,$C5,$F8,$23,$1D
	.dc.b	$8C,$7A,$E3,$EA,$8C,$57,$88,$F0,$7A,$D8,$9F,$7F,$D4,$5C,$BF,$DE
	.dc.b	$F9,$7F,$A5,$31,$F0,$4E,$70,$1E,$EC,$30,$D9,$F3,$7B,$18,$BE,$53
	.dc.b	$EF,$8C,$56,$5B,$5C,$0F,$74,$65,$A7,$F6,$0F,$92,$7C,$57,$FA,$26
	.dc.b	$3F,$D5,$71,$8F,$8F,$F7,$F8,$3F,$02,$37,$EA,$E2,$BD,$31,$B7,$8D
	.dc.b	$F3,$C9,$7C,$61,$87,$19,$5F,$4B,$E3,$39,$8F,$2F,$D4,$5B,$7D,$45
	.dc.b	$8C,$36,$FC,$2D,$C3,$C3,$1E,$6A,$E2,$FF,$B0,$3E,$9F,$FD,$BA,$FF
	.dc.b	$B2,$B8,$DF,$51,$FE,$06,$8D,$7C,$07,$9E,$73,$C7,$E9,$C5,$87,$27
	.dc.b	$B9,$FA,$7C,$39,$38,$37,$7C,$E3,$3B,$7E,$32,$73,$8F,$80,$FF,$6E
	.dc.b	$FC,$8B,$E0,$DC,$7E,$1C,$BC,$18,$7C,$DF,$47,$E3,$67,$F1,$DB,$8F
	.dc.b	$D4,$6E,$5E,$6E,$1F,$C0,$C6,$2F,$EB,$7B,$5F,$3C,$FA,$1F,$51,$70
	.dc.b	$FB,$B1,$70,$E5,$C2,$AC,$5F,$AF,$E9,$FF,$DA,$D8,$7F,$BD,$A1,$F5
	.dc.b	$1F,$85,$75,$E4,$7E,$0D,$CF,$47,$CD,$BE,$39,$7C,$C9,$CE,$1F,$EF
	.dc.b	$89,$CC,$7E,$5B,$0E,$93,$FA,$CA,$56,$A3,$0D,$87,$17,$CD,$F5,$3C
	.dc.b	$19,$7E,$3F,$EA,$26,$1F,$9C,$F0,$2E,$1F,$02,$6E,$7F,$9D,$9B,$5F
	.dc.b	$8F,$3F,$31,$F4,$1F,$99,$69,$FA,$62,$DF,$CA,$6D,$7C,$9F,$43,$F2
	.dc.b	$1F,$51,$FA,$89,$F8,$11,$8E,$EF,$8C,$D8,$C9,$F9,$B9,$E1,$7F,$91
	.dc.b	$D9,$5C,$2B,$5C,$16,$BF,$08,$3E,$A3,$F6,$4D,$9F,$E0,$AB,$E3,$33
	.dc.b	$32,$72,$F0,$8C,$4F,$4F,$A8,$E7,$F1,$DD,$1B,$7D,$1B,$3B,$8E,$F3
	.dc.b	$FA,$FF,$A8,$7C,$5F,$F0,$9E,$57,$D4,$3E,$43,$53,$98,$C6,$1F,$CB
	.dc.b	$B8,$BF,$5D,$87,$28,$FD,$6C,$EB,$2F,$DB,$4E,$5F,$DD,$E3,$DB,$1E
	.dc.b	$B7,$CC,$FC,$75,$6F,$0F,$E8,$BF,$50,$FA,$87,$D4,$70,$CB,$F8,$17
	.dc.b	$D4,$39,$FF,$85,$79,$FF,$98,$9F,$E8,$73,$9C,$FB,$47,$F4,$69,$D3
	.dc.b	$FD,$CF,$DB,$FE,$25,$E5,$7D,$3A,$DC,$E5,$B1,$98,$FE,$25,$F4,$EF
	.dc.b	$94,$E3,$7F,$C6,$E3,$17,$C2,$31,$F0,$5F,$C7,$FE,$A1,$EB,$62,$BB
	.dc.b	$C6,$1F,$0D,$FF,$16,$6D,$75,$F3,$38,$65,$E2,$71,$3E,$8F,$A8,$7D
	.dc.b	$46,$2B,$F9,$47,$98,$EB,$FE,$91,$6F,$A6,$7C,$27,$F3,$3A,$D4,$FC
	.dc.b	$83,$51,$9A,$E3,$5C,$27,$E3,$D7,$0B,$FE,$73,$1C,$0C,$D7,$17,$A5
	.dc.b	$3E,$A3,$15,$FC,$F6,$5F,$11,$E6,$9F,$97,$39,$8C,$71,$7F,$49,$AD
	.dc.b	$9E,$7A,$DF,$57,$F4,$38,$CD,$F9,$27,$31,$FC,$74,$CC,$EE,$FD,$BF
	.dc.b	$41,$D5,$FE,$AD,$F5,$0F,$86,$F8,$26,$0B,$7F,$88,$8D,$3E,$E8,$D3
	.dc.b	$4D,$38,$C7,$E3,$E6,$F8,$6C,$65,$FD,$6F,$9B,$F8,$8B,$FE,$28,$C3
	.dc.b	$EC,$7D,$37,$F6,$99,$CC,$7F,$88,$8C,$9F,$9A,$73,$87,$F8,$CF,$05
	.dc.b	$71,$F3,$3E,$DB,$6F,$84,$7D,$EE,$AC,$5F,$F7,$1F,$0F,$EF,$65,$FE
	.dc.b	$3A,$97,$D4,$3E,$23,$FB,$E5,$6B,$1F,$F1,$AD,$6A,$73,$C9,$FE,$03
	.dc.b	$BE,$15,$FB,$0B,$E1,$8F,$F9,$9E,$72,$F8,$6B,$8D,$7F,$83,$FB,$BF
	.dc.b	$BC,$BF,$D5,$EF,$8D,$7F,$54,$FA,$0F,$EC,$9F,$51,$89,$46,$30,$F3
	.dc.b	$B6,$FC,$4D,$BF,$81,$B7,$B3,$87,$DA,$DA,$F4,$38,$7F,$84,$CF,$F9
	.dc.b	$53,$E1,$3F,$CA,$BF,$50,$FA,$6F,$35,$FF,$88,$6B,$42,$74,$FF,$97
	.dc.b	$33,$FA,$F8,$72,$BF,$F4,$4E,$16,$B5,$87,$F9,$B3,$0E,$31,$FD,$E6
	.dc.b	$74,$FF,$76,$9D,$57,$28,$F0,$3E,$4B,$DC,$79,$2F,$9E,$1E,$28,$D9
	.dc.b	$FE,$9F,$C3,$AC,$66,$7C,$33,$C6,$FA,$3C,$0F,$21,$67,$84,$78,$E7
	.dc.b	$BE,$1E,$6A,$EC,$6D,$C3,$CA,$37,$87,$A2,$18,$79,$2B,$71,$E3,$3E
	.dc.b	$3C,$EE,$74,$D8,$FF,$5A,$9F,$50,$B1,$87,$07,$F8,$97,$E9,$FF,$8B
	.dc.b	$FB,$7C,$1B,$1F,$EB,$E6,$BF,$EC,$56,$BC,$7E,$4B,$FC,$FA,$ED,$AE
	.dc.b	$DA,$EF,$93,$B4,$70,$CB,$EF,$FD,$43,$E9,$3F,$C7,$11,$86,$EB,$C7
	.dc.b	$0F,$A5,$8B,$0E,$53,$FE,$3F,$C3,$95,$7E,$21,$AF,$09,$DB,$FE,$B0
	.dc.b	$AE,$58,$EF,$0E,$F3,$FE,$B4,$BE,$38,$FE,$52,$F7,$3C,$EF,$83,$0C
	.dc.b	$12,$7D,$4E,$5E,$58,$78,$A1,$5D,$27,$71,$F8,$FB,$F0,$E3,$F0,$67
	.dc.b	$8E,$1C,$E3,$E0,$63,$E8,$BE,$78,$F5,$C3,$E0,$DF,$FA,$35,$BA,$70
	.dc.b	$AF,$F0,$3E,$5E,$1C,$71,$1F,$E9,$7A,$DB,$F0,$F1,$FF,$36,$61,$FE
	.dc.b	$97,$BE,$13,$FE,$DA,$FA,$8F,$F5,$EB,$FE,$12,$6B,$FE,$A1,$6C,$FF
	.dc.b	$A7,$FE,$9F,$FA,$44,$EF,$A7,$AE,$7F,$E7,$99,$FF,$2B,$CF,$F8,$71
	.dc.b	$AF,$39,$DB,$E1,$F9,$BA,$B0,$46,$1B,$1C,$FC,$0C,$4C,$8F,$43,$CA
	.dc.b	$D3,$99,$FF,$5D,$47,$F8,$41,$F2,$78,$99,$CB,$FD,$75,$F5,$1F,$CD
	.dc.b	$A7,$FC,$66,$D7,$FD,$84,$D8,$F8,$5E,$67,$A2,$31,$A7,$FA,$65,$B4
	.dc.b	$0E,$16,$0A,$E5,$D1,$E9,$FA,$89,$5E,$28,$F8,$BF,$3F,$FE,$67,$8F
	.dc.b	$F2,$3B,$5F,$F6,$B3,$72,$FA,$DD,$7C,$3D,$59,$5C,$1F,$8D,$18,$8F
	.dc.b	$8B,$EC,$FC,$93,$15,$FE,$DD,$F8,$9F,$F2,$84,$FF,$95,$9A,$FF,$B9
	.dc.b	$5B,$1F,$EE,$3E,$5F,$F3,$34,$EF,$CA,$FB,$B5,$AE,$0F,$F5,$F3,$67
	.dc.b	$E3,$57,$07,$A5,$7E,$12,$1F,$11,$4E,$71,$F8,$8F,$3F,$F0,$B5,$FE
	.dc.b	$F7,$6E,$3F,$EF,$56,$E3,$FE,$CD,$FA,$8F,$F3,$1D,$7F,$9C,$27,$2F
	.dc.b	$D7,$6B,$FE,$F8,$E6,$FC,$AC,$F0,$8E,$D1,$8F,$8F,$91,$CE,$37,$8F
	.dc.b	$E7,$3E,$A3,$F4,$F2,$C9,$97,$FB,$7E,$73,$18,$9F,$C4,$63,$F1,$54
	.dc.b	$EA,$B8,$78,$9D,$4E,$F3,$FE,$8A,$FA,$8F,$F5,$7B,$7D,$EF,$D6,$C6
	.dc.b	$23,$C5,$5C,$9E,$1E,$86,$88,$FF,$A5,$3E,$A1,$F1,$1F,$B6,$FA,$87
	.dc.b	$C2,$52,$57,$23,$FD,$37,$0F,$A4,$E0,$48,$CF,$D4,$7F,$00,$F8,$A7
	.dc.b	$B9,$49,$3F,$EA,$4F,$8F,$7A,$58,$8F,$49,$FA,$DA,$E2,$FF,$AA,$9A
	.dc.b	$E8,$4F,$27,$55,$FC,$42,$1F,$51,$4E,$DC,$BF,$4E,$9F,$51,$F4,$5F
	.dc.b	$DE,$3F,$E7,$57,$FD,$72,$D7,$C0,$ED,$E3,$C9,$FC,$7A,$35,$3D,$5B
	.dc.b	$C5,$FC,$1E,$B6,$74,$F1,$38,$11,$EF,$F4,$35,$EF,$16,$FA,$4F,$CD
	.dc.b	$F9,$69,$5F,$15,$F1,$71,$E2,$3E,$36,$C9,$C5,$F9,$6A,$E3,$E5,$C9
	.dc.b	$F5,$1E,$B3,$38,$FF,$1F,$FA,$77,$8D,$FC,$F1,$FC,$14,$7F,$B7,$9A
	.dc.b	$FD,$57,$6E,$63,$87,$24,$8C,$1C,$7E,$A3,$A1,$1E,$9D,$F5,$E9,$17
	.dc.b	$4F,$B5,$87,$EF,$3E,$A1,$F3,$1F,$D4,$1B,$7F,$EF,$26,$BF,$89,$FA
	.dc.b	$8F,$E0,$17,$F8,$39,$CB,$6B,$91,$FD,$6A,$77,$B7,$DB,$AD,$BF,$76
	.dc.b	$1F,$2F,$B3,$8E,$1F,$D1,$F0,$FC,$84,$FF,$89,$E3,$15,$C7,$D7,$EC
	.dc.b	$77,$EF,$B3,$A3,$FA,$85,$71,$7D,$53,$B7,$42,$3F,$48,$FB,$FA,$BD
	.dc.b	$47,$AF,$AE,$47,$09,$FD,$6E,$3E,$6A,$CC,$FE,$BF,$EA,$1F,$51,$FD
	.dc.b	$F1,$B1,$E0,$6B,$FB,$5F,$4F,$F4,$D1,$8B,$FD,$83,$74,$FD,$23,$6B
	.dc.b	$FB,$65,$70,$BF,$F0,$4B,$FE,$0E,$BF,$DC,$36,$F8,$CE,$BA,$F7,$D7
	.dc.b	$30,$F8,$77,$FD,$CD,$B3,$97,$CD,$3C,$5D,$F4,$F8,$3E,$A2,$46,$08
	.dc.b	$7B,$5C,$BE,$F8,$FF,$0F,$FD,$47,$F0,$52,$77,$63,$E1,$F2,$FF,$AA
	.dc.b	$FB,$BF,$82,$7D,$2F,$FB,$7F,$0F,$F5,$F4,$63,$E1,$75,$1F,$E0,$FC
	.dc.b	$3F,$A7,$FC,$0F,$EE,$E3,$FC,$25,$7C,$5F,$F4,$BD,$F1,$79,$9A,$BE
	.dc.b	$6A,$F9,$CE,$E3,$89,$B3,$88,$AE,$D7,$E1,$8E,$58,$74,$97,$D3,$D2
	.dc.b	$4F,$28,$7D,$46,$23,$B8,$FA,$77,$D4,$7F,$7A,$C7,$B4,$63,$E3,$39
	.dc.b	$63,$EA,$8C,$5F,$26,$D7,$0B,$E5,$9F,$CE,$BE,$11,$8B,$FF,$1D,$5F
	.dc.b	$0A,$FE,$6C,$D8,$F9,$B5,$C3,$0F,$87,$39,$9F,$45,$F3,$7E,$35,$6D
	.dc.b	$F2,$46,$67,$86,$5F,$1A,$79,$CE,$49,$7E,$E8,$CB,$1E,$6E,$5A,$4C
	.dc.b	$7E,$04,$65,$B1,$9C,$7D,$B5,$E7,$BF,$14,$E4,$F8,$71,$E2,$7E,$AC
	.dc.b	$6A,$B9,$92,$B9,$DF,$76,$D7,$33,$E2,$C6,$5E,$AF,$AE,$79,$47,$CE
	.dc.b	$31,$3C,$A3,$31,$F3,$0C,$9C,$8E,$D5,$E5,$9E,$04,$8D,$4F,$E5,$E7
	.dc.b	$51,$89,$D3,$E6,$3D,$F8,$FD,$9A,$F0,$E3,$F6,$23,$15,$C2,$31,$5E
	.dc.b	$43,$95,$7A,$2F,$BC,$62,$B8,$63,$F4,$70,$E7,$3E,$32,$1E,$F3,$E5
	.dc.b	$61,$F8,$F3,$BB,$89,$FA,$D8,$7B,$6B,$9C,$F7,$39,$E1,$E2,$BE,$57
	.dc.b	$C3,$0F,$D2,$57,$A6,$73,$8F,$D7,$8E,$44,$8F,$0E,$3F,$12,$F9,$4F
	.dc.b	$D9,$3C,$95,$F0,$F0,$F0,$4E,$A7,$9C,$62,$7C,$13,$F0,$E6,$C7,$1C
	.dc.b	$7D,$78,$79,$27,$EA,$9A,$3F,$03,$1A,$AE,$D7,$D6,$7F,$C6,$07,$E1
	.dc.b	$CB,$4F,$1D,$F1,$7A,$9D,$1F,$C0,$A3,$2F,$F0,$D6,$E5,$FC,$87,$E9
	.dc.b	$FF,$CB,$3D,$1F,$C1,$27,$3A,$F1,$C3,$07,$F2,$AA,$DE,$1C,$E3,$1D
	.dc.b	$BF,$E8,$5C,$7F,$8F,$E1,$CA,$BF,$D9,$D7,$C2,$73,$7C,$31,$EB,$5F
	.dc.b	$CA,$AB,$58,$F8,$63,$1E,$AF,$58,$FF,$29,$4B,$EA,$31,$5D,$AA,$A9
	.dc.b	$D5,$FA,$67,$30,$6F,$AF,$86,$CE,$CF,$81,$19,$69,$9A,$F4,$46,$5F
	.dc.b	$EA,$3F,$22,$F8,$AF,$F4,$9D,$7F,$88,$63,$1F,$0D,$FE,$AB,$F0,$3F
	.dc.b	$D5,$91,$BF,$53,$15,$FE,$F5,$8D,$BC,$6F,$9E,$1F,$E8,$E9,$CB,$63
	.dc.b	$37,$D1,$61,$D2,$55,$E5,$FA,$8C,$3E,$7F,$A8,$B6,$9B,$18,$6E,$3E
	.dc.b	$36,$DF,$FA,$6F,$E9,$FF,$DC,$73,$FE,$3F,$E1,$7D,$47,$F9,$E6,$3F
	.dc.b	$C7,$B3,$AA,$FF,$53,$D6,$B8,$BC,$D0,$9D,$6B,$FE,$D1,$AD,$5F,$2F
	.dc.b	$7F,$D9,$18,$BF,$14,$72,$BE,$34,$97,$C8,$3E,$09,$C5,$BC,$3E,$0F
	.dc.b	$A1,$F0,$9B,$87,$D1,$6E,$1F,$3D,$B9,$FC,$F7,$E4,$FB,$32,$BE,$C7
	.dc.b	$D4,$5B,$FA,$D3,$6F,$EB,$4D,$AF,$F5,$77,$D2,$FF,$B7,$1F,$F3,$44
	.dc.b	$BE,$A3,$EE,$C6,$BC,$8F,$B9,$39,$E8,$F9,$37,$C7,$2F,$F6,$9C,$E6
	.dc.b	$FF,$CD,$B3,$98,$FC,$96,$1D,$2B,$E3,$4E,$6B,$B1,$FE,$79,$DF,$E4
	.dc.b	$7A,$7D,$B2,$FC,$3F,$D4,$4A,$FC,$B7,$7D,$C7,$F2,$93,$73,$FC,$A4
	.dc.b	$DA,$FC,$F7,$49,$FC,$8F,$CF,$FE,$49,$A7,$E8,$4B,$87,$E7,$DB,$7F
	.dc.b	$EE,$3D,$1F,$4F,$F8,$4F,$A8,$FD,$0D,$7C,$08,$C7,$77,$C6,$6C,$64
	.dc.b	$FC,$C4,$F0,$C3,$FD,$ED,$7D,$27,$51,$F9,$89,$F9,$0A,$74,$FE,$D4
	.dc.b	$7D,$47,$EB,$9B,$5F,$C0,$D7,$C5,$E6,$64,$E5,$E1,$18,$9E,$9F,$51
	.dc.b	$CF,$E3,$7A,$36,$FA,$36,$77,$1D,$E7,$F5,$BF,$4E,$F8,$AF,$E0,$9C
	.dc.b	$AF,$A8,$7C,$86,$A7,$31,$8C,$3F,$D6,$B7,$C5,$FC,$24,$67,$17,$0F
	.dc.b	$DB,$7F,$61,$5B,$73,$5F,$B3,$C7,$C1,$8F,$78,$EB,$8F,$F1,$7F,$A8
	.dc.b	$7D,$43,$EA,$35,$8F,$23,$F7,$9F,$50,$E6,$FE,$13,$E7,$FD,$48,$C6
	.dc.b	$5F,$EC,$C8,$FA,$FE,$33,$DC,$FE,$1E,$FD,$29,$FF,$68,$B9,$6F,$27
	.dc.b	$F1,$4F,$A7,$7C,$A7,$0B,$C3,$6B,$FD,$B3,$2F,$82,$FE,$37,$F5,$0F
	.dc.b	$67,$94,$61,$E9,$E5,$62,$F8,$B7,$7F,$70,$C3,$F8,$BC,$E8,$FF,$73
	.dc.b	$FD,$43,$EA,$39,$C7,$FB,$AB,$D3,$76,$7F,$41,$87,$D3,$3E,$13,$F9
	.dc.b	$7D,$6A,$7F,$DE,$C6,$A3,$35,$C6,$B8,$4F,$58,$C4,$EA,$30,$72,$97
	.dc.b	$D4,$67,$3F,$5C,$BE,$2B,$C9,$3F,$32,$3F,$E0,$09,$93,$C6,$E6,$78
	.dc.b	$CF,$0E,$1F,$E8,$F5,$B9,$FC,$23,$E0,$8C,$CC,$8F,$27,$CF,$DA,$F9
	.dc.b	$BF,$50,$F8,$6F,$71,$FD,$49,$C5,$7A,$E3,$06,$9F,$E9,$9E,$17,$E7
	.dc.b	$E9,$5C,$1F,$1E,$76,$75,$36,$BF,$AA,$57,$0B,$F9,$DF,$09,$DA,$7A
	.dc.b	$9A,$E4,$FE,$29,$87,$EB,$63,$ED,$FD,$37,$F6,$89,$CC,$7F,$5F,$32
	.dc.b	$72,$C3,$FB,$14,$69,$FE,$61,$92,$B9,$79,$7F,$7D,$B7,$C6,$BF,$35
	.dc.b	$3F,$83,$EB,$CC,$7E,$5B,$C4,$FC,$53,$FD,$F3,$EA,$1F,$11,$FD,$F2
	.dc.b	$73,$A7,$F7,$B9,$CE,$3D,$67,$2F,$F7,$7C,$13,$AC,$3B,$72,$FF,$80
	.dc.b	$23,$18,$7D,$1C,$38,$D7,$F0,$DA,$FF,$02,$F7,$FF,$80,$A3,$18,$F7
	.dc.b	$F9,$FF,$ED,$3F,$51,$C2,$7F,$C0,$B1,$8B,$F4,$36,$B8,$1F,$DD,$2B
	.dc.b	$86,$1F,$26,$3D,$5C,$3E,$FD,$FE,$03,$63,$DD,$1F,$E1,$EC,$7F,$C2
	.dc.b	$9F,$09,$FE,$15,$FA,$87,$D4,$7E,$13,$1E,$95,$FC,$08,$D8,$9D,$BF
	.dc.b	$CA,$EF,$F9,$84,$62,$3F,$1B,$7D,$6F,$A5,$69,$FF,$44,$46,$67,$96
	.dc.b	$15,$E2,$77,$9E,$31,$98,$F1,$DF,$7C,$3C,$EF,$86,$33,$5C,$1E,$47
	.dc.b	$FA,$72,$FB,$57,$09,$F1,$CE,$8F,$F1,$BE,$1E,$77,$81,$99,$E0,$E2
	.dc.b	$FC,$53,$B8,$F4,$C6,$30,$F4,$43,$0F,$0D,$6E,$3C,$E7,$C2,$3C,$46
	.dc.b	$0F,$1C,$F2,$9F,$F5,$84,$7C,$62,$B8,$7E,$B9,$FE,$BA,$FA,$7F,$E4
	.dc.b	$1E,$A7,$06,$C7,$FA,$E5,$AE,$FA,$BF,$D1,$4E,$7E,$4F,$FB,$6B,$78
	.dc.b	$B9,$46,$23,$B3,$5F,$E1,$B4,$8E,$39,$FE,$17,$EA,$1F,$49,$FC,$4A
	.dc.b	$30,$DD,$F8,$D7,$06,$E3,$F8,$5B,$E3,$8D,$BE,$0F,$F9,$BF,$1F,$05
	.dc.b	$7E,$2D,$B1,$F7,$6B,$8B,$6B,$8A,$9D,$E3,$F9,$B6,$E1,$FC,$A2,$F8
	.dc.b	$CF,$A2,$BB,$63,$E7,$7C,$4D,$53,$C2,$B9,$61,$E0,$C3,$AD,$70,$C7
	.dc.b	$D8,$C3,$F2,$91,$A7,$CD,$87,$5C,$3A,$47,$FA,$3D,$AE,$AB,$91,$FD
	.dc.b	$1E,$3E,$0B,$6F,$FA,$1B,$FD,$2E,$FB,$4E,$B5,$F3,$DF,$F3,$38,$C3
	.dc.b	$E7,$3F,$51,$F5,$1F,$EC,$07,$F9,$C3,$5F,$F0,$5B,$6F,$FA,$C7,$D3
	.dc.b	$B8,$BC,$38,$7F,$54,$AD,$3F,$D5,$A7,$FC,$DB,$7F,$D0,$DA,$F4,$EA
	.dc.b	$E2,$FF,$AB,$E9,$F3,$78,$84,$61,$F2,$C6,$69,$E0,$6E,$B5,$5E,$8A
	.dc.b	$C1,$B9,$B1,$E3,$F0,$F9,$18,$7F,$C5,$3F,$50,$E9,$F1,$36,$3F,$AE
	.dc.b	$35,$FF,$17,$E3,$FE,$45,$73,$3C,$08,$FC,$1F,$32,$BC,$3E,$A3,$D4
	.dc.b	$63,$7F,$F0,$27,$CF,$FF,$9B,$27,$FB,$53,$5F,$F1,$EB,$63,$E9,$F6
	.dc.b	$FC,$78,$C7,$07,$F8,$19,$B5,$FE,$DB,$6A,$CB,$FC,$7E,$F6,$F5,$BE
	.dc.b	$67,$A1,$23,$07,$F7,$BF,$88,$FF,$2D,$B7,$C9,$FA,$D1,$88,$FF,$28
	.dc.b	$F4,$7F,$9B,$5D,$10,$FF,$2A,$C6,$F4,$FF,$11,$4E,$E7,$35,$F8,$18
	.dc.b	$7C,$3D,$3B,$C7,$E1,$7D,$0F,$BF,$7F,$E2,$B6,$EB,$FE,$26,$6E,$1F
	.dc.b	$EF,$BF,$A8,$FF,$36,$3F,$E1,$58,$CB,$FE,$FC,$E6,$E8,$7E,$46,$77
	.dc.b	$1F,$E6,$C9,$DC,$3E,$3F,$47,$09,$FF,$10,$47,$B6,$FF,$C6,$FF,$51
	.dc.b	$FA,$79,$FF,$12,$35,$FF,$3B,$CE,$D0,$8D,$63,$FE,$47,$6D,$74,$EB
	.dc.b	$F3,$67,$FA,$7F,$A8,$FF,$5B,$61,$FE,$31,$6B,$FE,$84,$6E,$1C,$EB
	.dc.b	$87,$85,$2B,$83,$FB,$1F,$A8,$7C,$3F,$EC,$FE,$A1,$F1,$74,$97,$C0
	.dc.b	$FF,$20,$43,$E9,$38,$12,$33,$F5,$1F,$BD,$F8,$AF,$F5,$04,$7F,$91
	.dc.b	$DA,$FF,$A6,$9B,$A7,$F9,$9F,$E3,$DE,$87,$1F,$76,$5F,$17,$F8,$44
	.dc.b	$EA,$30,$7F,$94,$D7,$D4,$68,$E3,$1F,$EA,$8F,$A8,$F9,$AF,$D2,$7F
	.dc.b	$CF,$CF,$F9,$6D,$AE,$F8,$7F,$8B,$47,$08,$3A,$C7,$FC,$FC,$DE,$4F
	.dc.b	$F3,$CB,$B3,$F8,$44,$F8,$7C,$3E,$67,$4F,$A1,$B1,$E1,$8A,$F9,$E3
	.dc.b	$E5,$7C,$AC,$8F,$94,$F9,$38,$F1,$1F,$2E,$2E,$3E,$49,$B8,$F9,$3E
	.dc.b	$A3,$D6,$62,$FF,$27,$F4,$EF,$17,$F9,$A3,$F7,$63,$FC,$FC,$D7,$FD
	.dc.b	$A2,$ED,$CC,$70,$E4,$91,$83,$8F,$D4,$74,$23,$D3,$C0,$AF,$48,$B9
	.dc.b	$FF,$AA,$70,$FD,$CF,$D4,$3E,$63,$FA,$43,$6F,$FD,$24,$D7,$FD,$CB
	.dc.b	$F5,$1F,$BC,$BF,$F5,$94,$E5,$B5,$C8,$FE,$A5,$3B,$D7,$FD,$63,$5B
	.dc.b	$7F,$D6,$70,$F9,$7D,$9C,$70,$FF,$51,$47,$F8,$AD,$BE,$AF,$43,$AF
	.dc.b	$7D,$8C,$90,$F3,$BE,$2E,$AF,$1F,$54,$9E,$B1,$97,$BC,$65,$F7,$C7
	.dc.b	$F6,$BF,$A8,$7D,$45,$24,$FF,$AD,$FD,$2F,$CD,$4E,$A3,$FD,$74,$D5
	.dc.b	$BF,$FB,$66,$30,$FF,$B6,$E7,$8D,$7F,$79,$7F,$BF,$C6,$7E,$63,$BD
	.dc.b	$F2,$DB,$FD,$B7,$5F,$A2,$AE,$5D,$3C,$4E,$5D,$52,$FC,$D1,$97,$D1
	.dc.b	$1D,$FA,$7C,$1F,$51,$27,$D9,$19,$7D,$D1,$97,$C4,$7F,$30,$C3,$F8
	.dc.b	$FB,$FE,$D9,$FA,$8F,$DE,$13,$BB,$FC,$28,$FE,$EE,$3F,$9B,$BF,$CE
	.dc.b	$FD,$E7,$77,$FB,$DB,$E9,$7F,$DB,$75,$FE,$B8,$8C,$7C,$2F,$FA,$B9
	.dc.b	$FF,$7D,$DF,$17,$33,$B3,$FA,$1D,$72,$7F,$D2,$51,$A6,$9E,$2C,$3A
	.dc.b	$3C,$2E,$AA,$E5,$87,$9A,$34,$77,$18,$78,$DF,$0F,$D3,$F2,$A3,$4C
	.dc.b	$7D,$45,$9E,$5F,$4E,$FA,$8F,$ED,$F8,$F7,$94,$E7,$E2,$FC,$18,$F4
	.dc.b	$9E,$37,$C3,$5F,$EC,$71,$C2,$30,$FF,$66,$BE,$4F,$C2,$3D,$D9,$7C
	.dc.b	$08,$CE,$3E,$16,$DF,$46,$D7,$B8,$F0,$E3,$1E,$91,$A7,$A9,$9A,$E9
	.dc.b	$87,$92,$3C,$D5,$EC,$C3,$C5,$39,$BE,$47,$3C,$7E,$05,$F6,$9D,$12
	.dc.b	$F9,$47,$7C,$7D,$13,$98,$F8,$15,$CB,$1F,$1D,$EE,$7E,$35,$79,$6B
	.dc.b	$57,$CA,$B5,$5E,$83,$9E,$3E,$88,$CE,$3E,$78,$C4,$F9,$E7,$8D,$F4
	.dc.b	$8E,$F1,$C3,$0F,$05,$70,$C7,$CB,$5C,$4E,$E6,$EB,$9C,$EF,$1F,$34
	.dc.b	$EE,$BF,$15,$87,$78,$F4,$47,$42,$4F,$EC,$6B,$94,$F9,$0E,$51,$97
	.dc.b	$C0,$60,$FC,$11,$4E,$AF,$8A,$7C,$33,$D2,$72,$75,$8C,$BD,$1F,$1B
	.dc.b	$D1,$AF,$02,$57,$AF,$1F,$C2,$E3,$F7,$6B,$9D,$78,$E7,$38,$FE,$12
	.dc.b	$31,$5B,$C7,$D9,$18,$9F,$96,$F8,$E3,$D4,$6A,$FC,$77,$D2,$7E,$99
	.dc.b	$C4,$F8,$58,$79,$70,$ED,$1E,$FB,$E4,$74,$70,$43,$F4,$E7,$3B,$F8
	.dc.b	$35,$CF,$0F,$24,$FC,$5A,$E4,$FC,$5C,$3F,$20,$E8,$F8,$F8,$FD,$18
	.dc.b	$F4,$61,$F1,$8F,$C6,$CF,$F0,$83,$F3,$85,$87,$92,$F8,$AD,$4F,$E1
	.dc.b	$4D,$46,$5F,$E0,$2D,$AF,$E3,$5F,$4F,$FE,$52,$E7,$FE,$3F,$39,$CB
	.dc.b	$F8,$BC,$E6,$F8,$E1,$FC,$B6,$31,$87,$F1,$3B,$E1,$5F,$E8,$09,$CD
	.dc.b	$FF,$15,$9C,$CA,$72,$FF,$21,$8C,$C7,$FA,$4F,$0E,$DF,$50,$7C,$1E
	.dc.b	$AD,$BF,$E4,$D4,$AC,$63,$EC,$AB,$8F,$A9,$B7,$E5,$F5,$AB,$E3,$8B
	.dc.b	$82,$6C,$61,$B8,$FC,$16,$D7,$89,$B8,$7F,$53,$F9,$17,$C4,$FF,$A3
	.dc.b	$AF,$FD,$47,$D9,$E9,$9C,$FC,$2F,$AE,$31,$5B,$C7,$FB,$DD,$6A,$F8
	.dc.b	$9F,$EA,$C7,$AF,$A9,$C6,$31,$4C,$31,$6B,$DD,$72,$8C,$56,$E3,$13
	.dc.b	$E8,$FA,$7E,$2F,$AB,$EA,$2D,$26,$C6,$1B,$5C,$A7,$4D,$9F,$1D,$BE
	.dc.b	$9F,$FD,$C7,$3E,$3E,$37,$D4,$7F,$84,$23,$CD,$5B,$F7,$BD,$75,$AE
	.dc.b	$3E,$78,$72,$AF,$F1,$0C,$E6,$F8,$CE,$77,$FC,$44,$E6,$BF,$C4,$E7
	.dc.b	$5A,$DE,$1F,$23,$E3,$DF,$03,$F8,$39,$C9,$ED,$F4,$3E,$1D,$7C,$46
	.dc.b	$C6,$4F,$85,$19,$69,$F1,$7D,$9F,$96,$7C,$CF,$A8,$B1,$92,$63,$F8
	.dc.b	$2E,$04,$EA,$3D,$3F,$4B,$FE,$DC,$AF,$B9,$AB,$EA,$3F,$14,$FE,$9E
	.dc.b	$95,$B7,$FC,$8F,$1B,$E6,$FA,$B3,$98,$FC,$A5,$72,$BF,$F0,$CC,$E5
	.dc.b	$FC,$DC,$66,$FF,$DF,$F3,$96,$F9,$1F,$2F,$6F,$F4,$65,$FD,$CA,$D7
	.dc.b	$37,$E9,$BE,$A3,$F1,$A7,$E3,$BB,$AD,$7E,$2E,$2E,$3F,$0E,$6E,$7F
	.dc.b	$A8,$AF,$CA,$FC,$FF,$E5,$9A,$7E,$4C,$B5,$F2,$DB,$3F,$17,$17,$9B
	.dc.b	$F8,$EF,$A8,$FD,$25,$FC,$E8,$C7,$67,$C6,$6E,$1E,$B6,$CE,$E3,$07
	.dc.b	$E7,$6B,$E3,$47,$D2,$9F,$D6,$AF,$A8,$A6,$3E,$26,$DF,$ED,$27,$4D
	.dc.b	$9D,$34,$D7,$D4,$74,$F8,$DE,$AD,$8C,$9C,$E3,$2D,$3F,$5F,$F4,$EF
	.dc.b	$A7,$FD,$17,$D4,$3E,$3F,$F7,$33,$9A,$FE,$15,$DC,$AD,$47,$2B,$E3
	.dc.b	$7C,$1B,$19,$26,$7F,$E9,$8F,$A8,$7D,$43,$EA,$38,$57,$F0,$2F,$A8
	.dc.b	$73,$7F,$0A,$F7,$7E,$5C,$62,$7F,$83,$C7,$CE,$F1,$1E,$D7,$F1,$05
	.dc.b	$C9,$63,$0F,$F1,$4F,$A7,$7C,$8E,$70,$FE,$23,$18,$7F,$89,$DF,$47
	.dc.b	$3F,$01,$FC,$73,$EA,$1E,$B7,$38,$C3,$D7,$D1,$C5,$74,$6F,$17,$F1
	.dc.b	$2C,$7F,$87,$36,$3F,$93,$FD,$43,$EA,$39,$5F,$F2,$4F,$11,$D7,$FC
	.dc.b	$FA,$1F,$4A,$F8,$0F,$E5,$F5,$A9,$FE,$80,$6A,$33,$5C,$6B,$84,$F5
	.dc.b	$8C,$4E,$A3,$07,$29,$7D,$46,$73,$FE,$57,$2F,$86,$F2,$4F,$C9,$9C
	.dc.b	$C6,$38,$BF,$A1,$D6,$CF,$2D,$6F,$C6,$FE,$31,$19,$BF,$F6,$EB,$E0
	.dc.b	$AE,$37,$23,$C9,$F3,$D6,$BF,$9E,$FD,$43,$EA,$3D,$C7,$F5,$27,$15
	.dc.b	$EB,$8C,$1A,$7F,$A8,$C6,$A7,$C3,$9D,$CB,$F9,$64,$DA,$FF,$5E,$B6
	.dc.b	$BF,$C3,$14,$F8,$5E,$B5,$FD,$73,$93,$F8,$6E,$1F,$E1,$B8,$FB,$3F
	.dc.b	$4D,$FD,$9A,$73,$0C,$39,$CE,$7C,$AF,$B0,$D9,$FA,$8D,$9F,$EB,$B5
	.dc.b	$C9,$F1,$F4,$E7,$0F,$E8,$9E,$2F,$E1,$67,$FB,$4F,$D4,$3E,$1F,$FB
	.dc.b	$BC,$E6,$BF,$B4,$CE,$72,$FE,$EB,$39,$7C,$D8,$27,$58,$7F,$6E,$AD
	.dc.b	$3F,$E5,$78,$CD,$F8,$B9,$7B,$9F,$DF,$BB,$FF,$BC,$63,$FE,$E8,$C3
	.dc.b	$8E,$05,$F4,$3F,$E1,$EC,$1F,$51,$A9,$CD,$FF,$64,$8C,$56,$DB,$19
	.dc.b	$C7,$DF,$B7,$C0,$E0,$B3,$EE,$6E,$5E,$B9,$FF,$0D,$BF,$57,$E0,$3E
	.dc.b	$B7,$D4,$3E,$A3,$D1,$5A,$A5,$6A,$73,$85,$9F,$F3,$F4,$D7,$DD,$5F
	.dc.b	$E5,$FB,$E9,$1A,$0E,$1D,$D7,$17,$C8,$6A,$FB,$B4,$E5,$3C,$74,$5F
	.dc.b	$7A,$FF,$45,$E1,$E3,$9C,$E1,$D7,$0F,$59,$CC,$F0,$94,$F0,$E1,$D7
	.dc.b	$0E,$91,$E0,$C3,$C7,$3E,$98,$DC,$60,$F0,$BD,$6F,$C2,$F5,$3C,$33
	.dc.b	$D6,$FC,$98,$79,$F0,$EF,$5F,$A9,$37,$87,$57,$E5,$1F,$EA,$83,$D6
	.dc.b	$59,$C3,$F2,$8F,$F5,$97,$D4,$7F,$1D,$F6,$F8,$36,$3F,$D6,$2D,$77
	.dc.b	$D5,$FE,$82,$73,$F2,$2E,$1E,$2D,$8C,$C6,$08,$FC,$DF,$A8,$7D,$27
	.dc.b	$AF,$6F,$C0,$46,$27,$F1,$18,$B0,$E9,$5F,$85,$8C,$61,$F8,$DB,$E1
	.dc.b	$7D,$70,$E9,$A7,$C7,$BD,$CF,$DD,$6E,$1E,$0A,$DC,$FF,$15,$BE,$73
	.dc.b	$DF,$1F,$0E,$1C,$DB,$5B,$BE,$2A,$F8,$D6,$F1,$F5,$DF,$0A,$D5,$F0
	.dc.b	$BD,$C7,$F2,$5C,$7C,$B8,$71,$8F,$61,$EF,$AE,$F3,$A8,$E1,$7D,$6F
	.dc.b	$9C,$76,$3C,$38,$71,$7F,$D2,$71,$F2,$23,$FC,$D8,$DD,$BC,$19,$7C
	.dc.b	$5A,$DE,$3F,$91,$BD,$B6,$B7,$97,$E5,$7E,$A3,$FD,$71,$7F,$EA,$06
	.dc.b	$BF,$E9,$66,$DF,$F9,$EF,$E9,$BF,$A6,$4E,$FA,$FA,$3F,$D1,$E3,$FC
	.dc.b	$B6,$FF,$AA,$9A,$F3,$8D,$AA,$7C,$DE,$1B,$1E,$38,$DD,$3C,$0D,$44
	.dc.b	$9D,$1D,$E3,$0F,$FA,$BA,$7F,$C2,$8F,$8F,$C3,$CE,$3F,$E9,$5F,$A8
	.dc.b	$FE,$77,$3F,$EC,$06,$BF,$EB,$66,$C7,$C1,$F3,$7D,$31,$8D,$7F,$7C
	.dc.b	$DA,$FE,$D0,$D5,$87,$F9,$11,$ED,$D7,$EC,$FA,$89,$18,$3F,$DA,$3F
	.dc.b	$3F,$FE,$61,$6F,$AD,$F2,$23,$11,$ED,$ED,$C4,$C8,$D9,$F4,$63,$0F
	.dc.b	$C3,$F4,$3E,$59,$8D,$7F,$D6,$DF,$01,$FE,$52,$9F,$F7,$0B,$5F,$F6
	.dc.b	$BB,$63,$FD,$79,$CB,$FE,$64,$9D,$F9,$3F,$72,$B5,$BF,$F2,$16,$CF
	.dc.b	$F8,$2E,$B8,$3D,$2B,$F0,$70,$F8,$6A,$73,$8F,$C3,$F9,$FF,$84,$AF
	.dc.b	$F7,$4B,$71,$FF,$67,$37,$1F,$E6,$1F,$51,$FE,$64,$AF,$F7,$DC,$E5
	.dc.b	$FF,$0E,$35,$36,$38,$47,$6E,$6F,$CB,$C3,$E3,$A9,$E4,$8C,$E3,$F9
	.dc.b	$CF,$A8,$FD,$2D,$FF,$BF,$9B,$97,$FB,$8F,$0F,$87,$29,$CE,$5F,$D1
	.dc.b	$DB,$1F,$E3,$16,$A9,$E3,$F5,$1F,$EA,$78,$FF,$1B,$5F,$FB,$BE,$31
	.dc.b	$1F,$CF,$EF,$FA,$94,$62,$7F,$DE,$97,$D2,$3F,$CD,$15,$CA,$B7,$E0
	.dc.b	$FE,$DA,$73,$3C,$CF,$F2,$27,$D4,$3E,$1F,$F7,$7F,$50,$F8,$AA,$4A
	.dc.b	$E8,$7F,$92,$EF,$9E,$6C,$BF,$B3,$5B,$E9,$39,$92,$F8,$88,$7D,$47
	.dc.b	$F0,$AF,$8A,$7B,$94,$95,$C3,$E3,$9E,$8E,$23,$D4,$2B,$8B,$D9,$AE
	.dc.b	$A7,$8B,$9A,$FE,$33,$0F,$A8,$A7,$5D,$3D,$95,$FE,$69,$FA,$8F,$B2
	.dc.b	$7F,$A0,$DF,$5B,$5F,$1B,$AF,$4B,$F9,$2B,$99,$E7,$28,$C6,$DF,$E0
	.dc.b	$D7,$AF,$8B,$C0,$93,$DB,$E8,$2B,$DA,$2D,$F4,$A5,$7C,$BF,$94,$97
	.dc.b	$F0,$DF,$1F,$26,$23,$E4,$EC,$9C,$61,$CE,$AE,$5E,$5F,$A8,$F5,$98
	.dc.b	$C3,$FD,$19,$F4,$EF,$17,$FA,$1B,$F8,$18,$FB,$ED,$7E,$BB,$B7,$31
	.dc.b	$C3,$96,$46,$0E,$3F,$51,$D0,$8F,$4E,$EA,$F4,$8B,$9F,$CE,$C3,$F8
	.dc.b	$27,$D4,$3E,$5F,$FA,$BB,$6F,$F2,$6D,$7F,$17,$F4,$FF,$C2,$2F,$EB
	.dc.b	$CE,$5B,$5C,$8F,$EC,$33,$BD,$7F,$C8,$D5,$B7,$FC,$91,$0F,$A8,$D9
	.dc.b	$C7,$0F,$D1,$C7,$F8,$BD,$BE,$B7,$43,$AF,$81,$63,$24,$3C,$EF,$8B
	.dc.b	$AF,$C7,$D7,$27,$AC,$65,$EF,$19,$7D,$F1,$FD,$EB,$EA,$1F,$51,$49
	.dc.b	$3F,$B8,$F4,$BF,$8C,$CE,$A3,$F7,$6D,$7F,$66,$DD,$FF,$3F,$3C,$2B
	.dc.b	$FC,$0E,$FF,$82,$A3,$3F,$2D,$DA,$FF,$77,$3F,$DB,$5B,$D5,$C0,$E3
	.dc.b	$D9,$23,$C9,$19,$7C,$D1,$DB,$A7,$BF,$D4,$49,$F5,$46,$5F,$64,$65
	.dc.b	$F0,$C7,$F8,$73,$EA,$3F,$9B,$93,$BF,$1F,$11,$97,$FD,$09,$EF,$FE
	.dc.b	$A6,$FF,$DD,$DF,$4D,$FE,$E1,$7F,$D7,$9E,$07,$FA,$26,$31,$1F,$EB
	.dc.b	$A8,$FF,$46,$61,$FD,$3A,$B9,$4A,$B7,$18,$6D,$73,$F8,$6F,$0B,$CB
	.dc.b	$0B,$7E,$29,$AF,$36,$D7,$28,$0C,$3C,$57,$E7,$FA,$6D,$92,$7D,$67
	.dc.b	$BB,$EA,$30,$FA,$CE,$7F,$4E,$FA,$8F,$EF,$1A,$7B,$23,$1F,$17,$CA
	.dc.b	$BD,$D8,$79,$A7,$31,$CC,$38,$AE,$57,$D3,$1F,$35,$73,$AD,$47,$FA
	.dc.b	$DD,$B1,$E4,$9E,$0F,$D6,$BF,$04,$62,$3B,$B6,$78,$C7,$98,$ED,$19
	.dc.b	$26,$1E,$BC,$D7,$D2,$36,$4A,$F6,$4E,$9C,$63,$EC,$9B,$8F,$B2,$7E
	.dc.b	$16,$3E,$22,$63,$E7,$C3,$94,$EA,$3C,$6D,$BF,$26,$3E,$38,$C1,$31
	.dc.b	$FC,$64,$65,$F1,$46,$71,$F1,$46,$08,$FC,$DB,$F1,$57,$18,$C5,$FA
	.dc.b	$AF,$E7,$14,$F9,$AD,$C3,$D3,$87,$49,$FA,$D3,$9B,$E9,$18,$3C,$87
	.dc.b	$42,$9D,$0B,$5D,$2F,$C7,$1A,$C7,$ED,$CE,$63,$A6,$1E,$17,$93,$F1
	.dc.b	$1F,$44,$FD,$0C,$3E,$5C,$63,$1F,$9F,$5B,$8E,$D1,$8C,$3E,$49,$E2
	.dc.b	$9E,$D7,$E2,$C7,$DB,$7E,$0A,$D1,$C1,$F2,$B8,$69,$F0,$6B,$9D,$F7
	.dc.b	$8C,$4E,$DE,$87,$03,$F3,$47,$9B,$0F,$9C,$FA,$DA,$E4,$F9,$77,$E6
	.dc.b	$C3,$E1,$47,$E6,$A3,$17,$E1,$8E,$B8,$79,$2F,$9B,$63,$99,$DC,$A7
	.dc.b	$77,$E9,$63,$F5,$63,$D2,$4C,$3E,$43,$C6,$77,$87,$E7,$63,$F8,$D9
	.dc.b	$F8,$72,$9F,$C2,$BC,$77,$C5,$E2,$34,$70,$FA,$7F,$F2,$6F,$83,$FC
	.dc.b	$2E,$B7,$3F,$C4,$E7,$2F,$ED,$23,$38,$7F,$38,$9C,$BF,$C7,$E3,$3A
	.dc.b	$F5,$EC,$FF,$25,$CE,$71,$FE,$45,$7F,$E9,$3C,$6F,$D4,$6B,$0F,$27
	.dc.b	$AD,$73,$FE,$6F,$6A,$C4,$23,$93,$E7,$9C,$C7,$FA,$92,$FB,$C6,$3D
	.dc.b	$7C,$B6,$32,$79,$23,$2D,$33,$5F,$CF,$FE,$45,$F1,$5F,$E8,$79,$FF
	.dc.b	$42,$C6,$3E,$1B,$FC,$D3,$E0,$7F,$9A,$A3,$7E,$B6,$2B,$FA,$8C,$6D
	.dc.b	$E3,$7C,$F0,$FF,$4B,$4E,$5B,$19,$BE,$8B,$0E,$92,$AF,$2F,$D4,$61
	.dc.b	$F3,$FD,$45,$B4,$D8,$C3,$71,$F1,$B6,$FF,$D1,$FF,$4F,$FE,$DD,$9F
	.dc.b	$F5,$1F,$0B,$EA,$3F,$CD,$51,$FE,$9E,$9D,$57,$FA,$BE,$B5,$C5,$E6
	.dc.b	$84,$EB,$5F,$F4,$7D,$6A,$F9,$7B,$FE,$C8,$C5,$F8,$A3,$95,$F1,$AF
	.dc.b	$EF,$7F,$20,$F8,$2F,$C0,$5E,$E7,$E6,$FA,$3F,$07,$0F,$84,$D8,$C9
	.dc.b	$F4,$A3,$2D,$3E,$AF,$B5,$F2,$4F,$95,$F5,$16,$32,$4B,$FA,$F1,$97
	.dc.b	$FD,$6F,$F4,$DF,$EC,$F9,$FF,$65,$EA,$FA,$8F,$B9,$1F,$BA,$9D,$57
	.dc.b	$FB,$5E,$B5,$CD,$F2,$67,$31,$F8,$EB,$E3,$5F,$E1,$C9,$CD,$FF,$BB
	.dc.b	$27,$32,$9C,$BB,$F2,$3F,$13,$B7,$C9,$F5,$3B,$47,$E1,$FE,$A2,$63
	.dc.b	$F9,$6F,$02,$E1,$F9,$49,$B9,$FE,$52,$6D,$7E,$7B,$A3,$FD,$CC,$FE
	.dc.b	$7B,$E7,$FF,$24,$FE,$4D,$AE,$4F,$D1,$99,$3F,$DD,$3A,$3D,$0F,$C2
	.dc.b	$7D,$47,$E8,$67,$31,$8E,$DF,$8C,$D8,$C9,$F9,$89,$E1,$A7,$D4,$9D
	.dc.b	$CE,$78,$6C,$FE,$C8,$7D,$47,$EA,$DB,$3F,$C0,$17,$C5,$66,$64,$E5
	.dc.b	$E1,$18,$9E,$9F,$51,$CF,$E3,$3A,$36,$FA,$36,$77,$1D,$E7,$F5,$5F
	.dc.b	$50,$FA,$7F,$B3,$F5,$0F,$90,$CC,$61,$B7,$FE,$60,$E4,$FF,$7E,$F4
	.dc.b	$2F,$84,$E4,$FE,$97,$87,$57,$19,$2B,$86,$1D,$21,$87,$FB,$FA,$F8
	.dc.b	$46,$2F,$FC,$DF,$F5,$0F,$A8,$7D,$47,$1C,$15,$FD,$6B,$EA,$1C,$FF
	.dc.b	$C1,$FC,$FF,$CC,$C6,$27,$C7,$92,$7F,$3B,$E2,$BD,$BF,$E1,$CF,$D4
	.dc.b	$AD,$B9,$6F,$27,$F1,$3F,$A7,$7C,$A7,$0B,$C3,$6B,$BC,$BE,$0B,$F8
	.dc.b	$D7,$D4,$3D,$8E,$51,$87,$A7,$A3,$8B,$E2,$DD,$7E,$21,$FE,$92,$C3
	.dc.b	$F8,$BF,$07,$C2,$8C,$4F,$F2,$7F,$A8,$7D,$47,$4C,$3E,$0F,$A6,$EC
	.dc.b	$FE,$83,$0F,$A6,$7C,$27,$F2,$FA,$D4,$FD,$A3,$51,$9A,$E3,$5C,$27
	.dc.b	$AC,$62,$75,$18,$39,$4B,$EA,$33,$9F,$B0,$FB,$7F,$11,$E4,$8F,$9B
	.dc.b	$5F,$60,$46,$78,$BF,$A2,$46,$0F,$34,$F0,$AE,$91,$97,$E4,$BE,$19
	.dc.b	$E3,$52,$3C,$BF,$3F,$6B,$C9,$F5,$0F,$86,$F7,$9F,$D4,$DC,$57,$B2
	.dc.b	$30,$69,$FC,$7C,$6A,$7C,$39,$5E,$1F,$B1,$09,$C5,$7F,$8D,$1B,$1F
	.dc.b	$9E,$A7,$C3,$76,$AF,$CD,$F2,$7F,$14,$C3,$F0,$51,$F6,$FE,$9B,$FB
	.dc.b	$44,$E6,$3F,$DB,$26,$4E,$3E,$5F,$DA,$9D,$B6,$32,$4C,$7D,$3D,$7A
	.dc.b	$8F,$D5,$F8,$9F,$86,$9F,$EF,$1F,$50,$F8,$8F,$EF,$13,$9C,$FF,$BB
	.dc.b	$CE,$6F,$8F,$2F,$F7,$78,$C6,$1F,$83,$BE,$15,$FE,$F0,$C5,$5A,$C3
	.dc.b	$E4,$D6,$AF,$FB,$DE,$1F,$E0,$1E,$FF,$EF,$D1,$8C,$7C,$FF,$41,$FE
	.dc.b	$20,$C1,$F5,$1C,$6B,$F8,$B4,$E6,$31,$8A,$31,$8F,$B1,$B8,$7C,$2D
	.dc.b	$BD,$FC,$1F,$01,$B1,$92,$63,$F9,$F0,$3E,$12,$FD,$33,$E9,$BF,$CE
	.dc.b	$B9,$B0,$FF,$11,$5F,$0B,$DD,$6A,$7F,$81,$5F,$1A,$FF,$3D,$D6,$9F
	.dc.b	$F0,$FC,$EA,$BF,$CF,$93,$99,$FE,$6D,$39,$78,$8A,$E8,$EB,$0E,$0F
	.dc.b	$8C,$D5,$F8,$E3,$31,$ED,$32,$78,$9C,$DF,$87,$0F,$74,$62,$76,$77
	.dc.b	$29,$DE,$FC,$73,$C2,$32,$77,$8C,$E1,$EB,$8F,$F4,$FD,$F7,$8C,$47
	.dc.b	$3C,$3D,$0F,$29,$D6,$1E,$3B,$E6,$F9,$67,$FD,$48,$7B,$4B,$38,$7E
	.dc.b	$21,$FE,$AC,$FA,$8F,$E3,$3E,$AF,$06,$C7,$FA,$B5,$AE,$FA,$BF,$D0
	.dc.b	$EE,$7E,$4B,$FB,$3B,$78,$79,$46,$23,$B3,$5F,$F5,$ED,$23,$8E,$5F
	.dc.b	$6B,$EA,$1F,$49,$FE,$C2,$8C,$37,$5E,$35,$C1,$B8,$FE,$1E,$F8,$E3
	.dc.b	$6F,$83,$F9,$CC,$7C,$15,$F8,$D6,$C7,$DC,$AE,$2D,$AE,$2A,$77,$8F
	.dc.b	$E4,$1B,$87,$FB,$62,$F8,$CF,$A2,$BB,$63,$E7,$7C,$4D,$53,$C2,$B9
	.dc.b	$61,$E0,$C3,$AD,$70,$C7,$D8,$C3,$F2,$71,$A7,$CD,$87,$5C,$3A,$47
	.dc.b	$EF,$1A,$EA,$B9,$1F,$D0,$23,$E0,$B6,$FF,$DE,$6F,$F4,$4B,$ED,$3A
	.dc.b	$D7,$CF,$7F,$E1,$A8,$C3,$E7,$3F,$4F,$F5,$1F,$EB,$D7,$FC,$3E,$D7
	.dc.b	$FB,$73,$6F,$FA,$8F,$D3,$FF,$4C,$9D,$F5,$78,$B0,$FE,$2B,$09,$FF
	.dc.b	$35,$3F,$E2,$A6,$BD,$27,$83,$FC,$72,$DF,$37,$88,$46,$1F,$2C,$66
	.dc.b	$9E,$06,$EB,$55,$E8,$AC,$1B,$9B,$1E,$3F,$0F,$91,$87,$FC,$19,$F5
	.dc.b	$1F,$CE,$5F,$F2,$0B,$5C,$F4,$78,$7C,$CF,$34,$E7,$5F,$EE,$AD,$D7
	.dc.b	$E1,$B6,$FF,$9C,$E9,$FE,$49,$7A,$E1,$EB,$FA,$89,$18,$3F,$CA,$3F
	.dc.b	$3F,$FE,$6F,$6F,$AD,$F1,$63,$11,$FE,$23,$BF,$F1,$B3,$A2,$1F,$E2
	.dc.b	$78,$DE,$5F,$6B,$B7,$95,$FC,$BF,$42,$5F,$0A,$FF,$07,$FC,$47,$F9
	.dc.b	$62,$3F,$CC,$6D,$7F,$C6,$4D,$8F,$F0,$AF,$37,$F9,$B2,$77,$E4,$FD
	.dc.b	$BA,$D6,$FF,$86,$6C,$FF,$5C,$AE,$0F,$4A,$FC,$14,$3E,$1E,$9C,$E3
	.dc.b	$F0,$DE,$7F,$E0,$31,$FF,$12,$B7,$1F,$C9,$B7,$0F,$ED,$3F,$51,$FE
	.dc.b	$6B,$7F,$CF,$B1,$94,$D3,$11,$B8,$FF,$28,$46,$3E,$3F,$F2,$D2,$7C
	.dc.b	$7C,$F8,$C7,$F3,$5F,$51,$FA,$7C,$3F,$CB,$0D,$CB,$FC,$77,$87,$C2
	.dc.b	$94,$E7,$2F,$D5,$B6,$3F,$C0,$2D,$53,$C7,$E9,$FF,$D6,$D1,$FE,$98
	.dc.b	$6B,$FE,$67,$6D,$FF,$92,$AB,$94,$7F,$99,$A7,$87,$83,$FB,$18,$C4
	.dc.b	$7F,$9B,$FE,$A1,$F1,$1F,$B5,$FA,$87,$C2,$52,$57,$23,$FD,$53,$0F
	.dc.b	$A4,$E0,$48,$CF,$D4,$7E,$FF,$E2,$BF,$D4,$11,$FE,$B0,$6B,$FE,$7F
	.dc.b	$6E,$BF,$E6,$2F,$8F,$FF,$5A,$4E,$FD,$DE,$44,$7F,$D7,$31,$A7,$94
	.dc.b	$7F,$0D,$87,$D4,$53,$84,$FF,$A3,$7E,$A3,$E7,$1F,$E7,$B7,$FD,$86
	.dc.b	$D7,$5E,$8F,$F1,$78,$DF,$B5,$4C,$46,$7C,$59,$5C,$9E,$1F,$43,$88
	.dc.b	$F0,$45,$CF,$FC,$FF,$82,$BE,$37,$CA,$CB,$F8,$CF,$87,$8B,$11,$F1
	.dc.b	$E2,$DF,$29,$BD,$1F,$CD,$B0,$7D,$47,$A4,$CE,$1F,$E4,$EF,$A7,$78
	.dc.b	$BF,$CC,$DF,$B5,$1F,$EE,$46,$BF,$EB,$37,$6E,$63,$87,$24,$8C,$1C
	.dc.b	$7E,$A3,$A1,$1E,$9D,$D5,$E9,$17,$4F,$F5,$0E,$1F,$B7,$FA,$87,$CC
	.dc.b	$7F,$47,$6D,$FF,$BE,$1A,$FF,$B1,$FE,$A3,$F7,$57,$FE,$AA,$9C,$B6
	.dc.b	$B9,$1F,$D4,$67,$7B,$7F,$5C,$AD,$BF,$D7,$A1,$F2,$FB,$38,$E1,$FE
	.dc.b	$76,$C3,$FD,$6F,$3F,$E2,$B8,$C5,$71,$F5,$7B,$1D,$FC,$0B,$3A,$3F
	.dc.b	$CF,$F5,$C5,$F5,$4E,$DD,$08,$FC,$F5,$F8,$3A,$BD,$47,$AF,$DD,$95
	.dc.b	$C2,$BF,$57,$8F,$AE,$B3,$3F,$EE,$DF,$A8,$7D,$47,$F7,$36,$C7,$81
	.dc.b	$AF,$FB,$CF,$D3,$FE,$29,$18,$CB,$FD,$AE,$DC,$3F,$C0,$AD,$AF,$F4
	.dc.b	$DD,$70,$BF,$F0,$03,$FE,$06,$BF,$F6,$FB,$6F,$8C,$EB,$AB,$7D,$53
	.dc.b	$0F,$87,$7F,$DB,$5F,$24,$E5,$F3,$4F,$17,$7D,$3E,$0F,$A8,$91,$82
	.dc.b	$1E,$D7,$2F,$BE,$3F,$C2,$DF,$51,$FC,$0F,$3F,$F7,$AB,$FC,$D4,$9D
	.dc.b	$DA,$37,$DD,$2F,$EC,$CF,$1D,$7F,$C5,$2F,$19,$D4,$AF,$AD,$E3,$DE
	.dc.b	$77,$FF,$8C,$3E,$9B,$FD,$D5,$7F,$EB,$F8,$C7,$C2,$FF,$A6,$9E,$B8
	.dc.b	$D9,$FF,$65,$D3,$EA,$31,$08,$7D,$36,$1B,$1D,$4E,$33,$93,$53,$23
	.dc.b	$47,$38,$C3,$FE,$D9,$BF,$F2,$3F,$D3,$BE,$A3,$FB,$A6,$7E,$CA,$FF
	.dc.b	$6F,$CE,$7E,$2B,$9E,$5E,$29,$E4,$7A,$B2,$F4,$D6,$E3,$21,$C5,$F5
	.dc.b	$8F,$74,$F9,$23,$13,$AC,$7C,$B8,$F9,$1F,$9D,$5B,$73,$8F,$BA,$B8
	.dc.b	$61,$E3,$3C,$F5,$E4,$BE,$51,$EC,$6B,$E2,$36,$73,$73,$3D,$23,$47
	.dc.b	$B3,$1F,$85,$8F,$A2,$B6,$74,$8C,$E3,$E8,$AE,$98,$C7,$96,$3E,$B7
	.dc.b	$C9,$19,$C3,$E3,$5F,$29,$F6,$1F,$33,$1F,$21,$E5,$8D,$47,$9A,$7C
	.dc.b	$95,$CF,$0F,$14,$F8,$0C,$1E,$03,$04,$9E,$2F,$7C,$7D,$98,$F9,$E3
	.dc.b	$EE,$47,$C0,$BF,$93,$1C,$6F,$EA,$BC,$63,$37,$EC,$BF,$A9,$1E,$29
	.dc.b	$F4,$57,$9A,$BC,$24,$70,$48,$F4,$46,$5F,$8A,$F5,$7A,$63,$F1,$6F
	.dc.b	$D0,$FC,$8C,$3D,$73,$E4,$BF,$3D,$F3,$8F,$19,$D2,$32,$47,$04,$3E
	.dc.b	$1E,$1E,$BC,$7E,$99,$F5,$30,$F0,$47,$79,$E1,$18,$7E,$41,$8C,$3E
	.dc.b	$63,$83,$C6,$76,$3E,$61,$9A,$ED,$1C,$30,$F8,$53,$C6,$BB,$46,$09
	.dc.b	$87,$AE,$7A,$46,$1F,$16,$1E,$BC,$7E,$4E,$1D,$6F,$E9,$57,$93,$0E
	.dc.b	$B1,$8C,$7C,$78,$71,$AF,$80,$76,$9F,$04,$FC,$6C,$3F,$07,$3E,$CC
	.dc.b	$7C,$4F,$5C,$7C,$0F,$CF,$AE,$AE,$EF,$C7,$8F,$F1,$18,$FF,$71,$9F
	.dc.b	$1C,$AB,$C9,$7C,$56,$A7,$F4,$46,$A3,$2F,$EB,$DB,$8F,$F1,$9F,$A7
	.dc.b	$76,$B9,$3F,$89,$56,$B2,$FF,$79,$D3,$EA,$35,$8F,$F2,$48,$C5,$FF
	.dc.b	$36,$7F,$D3,$4D,$9D,$E3,$DD,$B7,$D1,$B6,$17,$C9,$ED,$1A,$73,$EA
	.dc.b	$D9,$F0,$C7,$8E,$31,$39,$66,$1E,$5F,$5A,$D7,$F4,$0F,$90,$7C,$4F
	.dc.b	$FA,$4A,$BF,$C5,$11,$8F,$85,$F7,$F4,$FC,$08,$DF,$A9,$8A,$F4,$46
	.dc.b	$DE,$37,$CF,$0F,$EE,$11,$A8,$61,$C6,$55,$D2,$B8,$CE,$63,$CB,$F4
	.dc.b	$F6,$DF,$51,$63,$0D,$BF,$0B,$70,$F0,$9D,$AB,$8C,$FF,$90,$3E,$9F
	.dc.b	$FD,$C8,$FF,$91,$F9,$1F,$51,$FE,$04,$8D,$7B,$DE,$79,$CF,$1F,$A7
	.dc.b	$16,$1C,$9E,$F3,$AB,$E2,$75,$C3,$A4,$F9,$0E,$38,$73,$BE,$5B,$FE
	.dc.b	$3B,$E3,$DF,$03,$FE,$EC,$C7,$9E,$1E,$CD,$FE,$CF,$A1,$F1,$72,$F8
	.dc.b	$CD,$AF,$A6,$DA,$E1,$B7,$FA,$02,$7E,$BD,$7C,$DF,$66,$47,$DA,$FA
	.dc.b	$8B,$7C,$26,$E9,$F8,$3C,$15,$8B,$FC,$D5,$BE,$9B,$FD,$BF,$1F,$E7
	.dc.b	$3B,$7D,$47,$E1,$5F,$D5,$D6,$FC,$8F,$C2,$46,$B9,$FE,$76,$1C,$AB
	.dc.b	$ED,$CE,$70,$FF,$3C,$CE,$63,$F2,$F8,$75,$9F,$1C,$E5,$B8,$7E,$47
	.dc.b	$83,$E6,$FA,$7D,$A3,$F1,$DF,$51,$2B,$F3,$7D,$F7,$3F,$93,$37,$3F
	.dc.b	$85,$36,$BF,$3C,$7E,$5F,$E7,$BF,$30,$D3,$F4,$A5,$BF,$A6,$DA,$F8
	.dc.b	$79,$BE,$9F,$F2,$1F,$51,$FA,$8A,$F7,$C6,$3B,$7E,$2B,$63,$27,$E6
	.dc.b	$E7,$86,$1F,$8F,$BE,$93,$A8,$FC,$BC,$FC,$77,$F8,$94,$E9,$FC,$70
	.dc.b	$FA,$8F,$DA,$36,$BF,$83,$AF,$8A,$CC,$C9,$CB,$C2,$31,$3D,$3E,$A3
	.dc.b	$9F,$C6,$F4,$6D,$F4,$6C,$EE,$3B,$CF,$EC,$7E,$99,$F5,$1F,$6F,$EA
	.dc.b	$1F,$21,$98,$C3,$6F,$FD,$89,$CC,$AD,$C7,$FB,$0A,$F9,$BA,$8C,$E3
	.dc.b	$FB,$79,$CB,$70,$FE,$35,$F5,$0F,$A8,$7D,$47,$7A,$FD,$F7,$D4,$39
	.dc.b	$BF,$84,$79,$FF,$3E,$53,$9D,$3F,$DA,$73,$9F,$1D,$EE,$7F,$0F,$7E
	.dc.b	$74,$FF,$B6,$1C,$B7,$8F,$F8,$A7,$D3,$BE,$53,$85,$E1,$B5,$FE,$E0
	.dc.b	$97,$C1,$7F,$1B,$FA,$87,$B3,$CA,$30,$F4,$F4,$71,$7C,$5B,$B7,$F1
	.dc.b	$BC,$3F,$8B,$36,$3F,$93,$7D,$43,$EA,$39,$5F,$FB,$C3,$CD,$75,$FF
	.dc.b	$43,$87,$D3,$3E,$13,$F9,$75,$6A,$7F,$DF,$86,$A3,$35,$C6,$B8,$4F
	.dc.b	$58,$C4,$EA,$30,$72,$97,$D4,$67,$3F,$E6,$92,$F8,$AF,$24,$FC,$B8
	.dc.b	$FE,$ED,$E3,$7F,$18,$74,$79,$23,$55,$CA,$B8,$F1,$7F,$47,$AD,$CF
	.dc.b	$F5,$47,$C3,$19,$99,$1E,$5F,$9F,$B5,$FD,$1F,$EA,$1F,$0B,$EF,$3F
	.dc.b	$A9,$38,$AF,$64,$60,$D3,$E8,$3F,$9A,$C6,$E7,$C5,$9D,$CF,$F9,$9C
	.dc.b	$DA,$DD,$FC,$DF,$84,$E9,$1C,$F0,$FF,$0C,$72,$7F,$0E,$C3,$FC,$4B
	.dc.b	$3F,$6B,$E9,$BF,$B3,$4E,$63,$FC,$3C,$64,$E5,$5F,$90,$F2,$FE,$D5
	.dc.b	$70,$6C,$EA,$B9,$57,$82,$7F,$01,$D7,$BA,$E7,$E2,$FE,$1E,$BF,$B9
	.dc.b	$FD,$43,$E1,$FF,$BC,$4E,$73,$FF,$1D,$E0,$AD,$61,$CB,$97,$FC,$01
	.dc.b	$18,$C7,$FB,$C6,$1C,$6B,$FC,$71,$7C,$71,$EB,$84,$9F,$F0,$67,$7F
	.dc.b	$F7,$C7,$FB,$2D,$F1,$C3,$FC,$80,$BE,$87,$FB,$0F,$D4,$66,$53,$9A
	.dc.b	$FE,$D5,$39,$8C,$4F,$F6,$48,$C5,$72,$C3,$DF,$BF,$C0,$E0,$B8,$FC
	.dc.b	$66,$E1,$EC,$9F,$F0,$E3,$FE,$53,$F8,$4F,$F2,$A7,$D4,$3E,$A3,$D3
	.dc.b	$5A,$A5,$6A,$73,$85,$9F,$F3,$CC,$D7,$D7,$5F,$E6,$AB,$E9,$1A,$0E
	.dc.b	$1D,$D7,$17,$C8,$6A,$FB,$B4,$E5,$3C,$74,$5F,$7A,$FF,$43,$61,$E3
	.dc.b	$9C,$E1,$D7,$0F,$59,$CC,$F0,$94,$F0,$E1,$D7,$0E,$91,$E0,$C3,$C7
	.dc.b	$3E,$98,$DC,$60,$F0,$BD,$6F,$C2,$F5,$3C,$33,$D6,$FC,$98,$79,$F0
	.dc.b	$EF,$5F,$E2,$E8,$C4,$EA,$3F,$D4,$47,$C6,$2C,$61,$C1,$FE,$36,$FA
	.dc.b	$8F,$E3,$DE,$DF,$06,$C7,$FA,$AD,$AF,$FA,$C9,$AF,$1F,$93,$72,$71
	.dc.b	$6C,$71,$6B,$C5,$AF,$18,$FC,$27,$D4,$3E,$93,$FC,$93,$A7,$E0,$63
	.dc.b	$15,$F8,$9C,$18,$76,$AF,$C3,$46,$30,$E7,$87,$2D,$BF,$CA,$78,$F2
	.dc.b	$C7,$EE,$5E,$DB,$8F,$3A,$DD,$F1,$9F,$F5,$A5,$F1,$C7,$BD,$EE,$7F
	.dc.b	$D7,$18,$73,$9F,$45,$F1,$95,$F1,$C3,$E0,$47,$83,$1F,$3B,$0C,$12
	.dc.b	$F9,$B9,$7C,$31,$F0,$F1,$F1,$4F,$6C,$3C,$18,$77,$8F,$2C,$EA,$3E
	.dc.b	$15,$7F,$9C,$1B,$A7,$C0,$AF,$E6,$79,$7A,$F2,$C4,$7F,$A5,$EB,$6F
	.dc.b	$C4,$C7,$FC,$FF,$87,$F9,$EA,$F8,$4F,$FB,$6F,$EA,$3F,$D7,$EF,$F3
	.dc.b	$E6,$BF,$EA,$16,$CF,$FA,$0B,$E9,$9C,$3D,$33,$FF,$45,$D6,$9F,$F4
	.dc.b	$6C,$FF,$99,$6F,$FA,$4B,$5E,$9D,$3C,$5F,$17,$CD,$D5,$68,$C3,$63
	.dc.b	$9F,$7E,$26,$47,$A5,$E5,$69,$CC,$FF,$AE,$E3,$FC,$30,$F9,$3C,$3C
	.dc.b	$E7,$FE,$9E,$FA,$8F,$E7,$D3,$FD,$81,$AF,$FB,$0D,$B1,$F0,$FC,$CF
	.dc.b	$44,$63,$4F,$F5,$6B,$68,$1C,$2C,$3F,$C9,$2F,$7E,$9F,$4F,$D4,$48
	.dc.b	$C1,$FD,$CB,$E7,$FF,$CD,$2D,$F6,$3E,$44,$62,$3D,$DD,$98,$99,$1B
	.dc.b	$3E,$9C,$61,$F8,$9E,$87,$C8,$31,$AF,$FB,$03,$E2,$3F,$CA,$D3,$FE
	.dc.b	$02,$6B,$FE,$E5,$6C,$7F,$B1,$B9,$BF,$CD,$53,$BF,$27,$EE,$56,$B7
	.dc.b	$FF,$68,$B6,$7E,$2D,$70,$7A,$57,$E0,$E1,$F0,$F4,$E7,$1F,$87,$F3
	.dc.b	$FF,$09,$5F,$EF,$76,$E3,$FE,$D8,$6E,$3F,$EE,$0F,$A8,$FF,$34,$D7
	.dc.b	$F8,$72,$72,$FD,$66,$A6,$C7,$08,$ED,$CD,$F9,$78,$7C,$7D,$3C,$91
	.dc.b	$9C,$7F,$39,$F5,$1F,$A8,$BF,$F1,$5B,$5F,$E6,$B8,$F9,$67,$7D,$72
	.dc.b	$76,$F3,$7F,$C6,$7F,$51,$FE,$AC,$6F,$B9,$FA,$98,$C4,$74,$3F,$23
	.dc.b	$87,$7D,$7F,$DF,$73,$FE,$69,$9E,$7E,$87,$13,$F6,$5F,$50,$F8,$7F
	.dc.b	$DA,$7D,$43,$E2,$E9,$2F,$89,$FE,$46,$87,$D2,$70,$24,$67,$EA,$3F
	.dc.b	$7D,$F1,$5F,$E9,$C8,$FF,$26,$B5,$F4,$07,$1A,$38,$7F,$57,$2F,$8F
	.dc.b	$7A,$3D,$3D,$E9,$3D,$1F,$E1,$93,$A8,$C1,$FE,$5C,$5F,$51,$A3,$8C
	.dc.b	$78,$BE,$A3,$E4,$61,$E0,$7F,$CF,$B1,$FE,$64,$6B,$BE,$5F,$E3,$51
	.dc.b	$C2,$3F,$80,$BA,$CB,$D5,$18,$F6,$77,$5C,$3C,$79,$3D,$5E,$7F,$43
	.dc.b	$88,$F2,$45,$BE,$70,$C1,$5F,$3B,$E5,$A4,$FC,$E7,$CB,$C9,$88,$F9
	.dc.b	$FC,$09,$C6,$1D,$30,$FC,$15,$71,$3E,$07,$D4,$7B,$8D,$EB,$F5,$3E
	.dc.b	$9D,$E3,$FF,$3B,$7F,$07,$5F,$E8,$A6,$BF,$79,$DB,$98,$E1,$CB,$23
	.dc.b	$07,$1F,$A8,$E8,$47,$A7,$7D,$7A,$45,$CF,$E2,$E1,$FB,$CF,$A8,$7C
	.dc.b	$C7,$F4,$F6,$DF,$FA,$79,$AF,$E4,$BE,$A3,$F8,$05,$FC,$E9,$CB,$6B
	.dc.b	$91,$FD,$66,$77,$A7,$E5,$AB,$6F,$E6,$21,$F2,$FB,$38,$E1,$FE,$AE
	.dc.b	$8F,$F1,$4B,$7D,$8E,$87,$5F,$02,$C6,$48,$79,$DF,$17,$67,$8F,$B2
	.dc.b	$4F,$58,$CB,$DE,$32,$FB,$E3,$FB,$97,$D4,$3E,$A2,$92,$7F,$D8,$7E
	.dc.b	$A7,$E7,$54,$ED,$FF,$63,$B5,$FD,$DB,$77,$FC,$BC,$F1,$8C,$6B,$FB
	.dc.b	$C8,$FF,$03,$4E,$BE,$63,$C1,$7F,$C1,$27,$FB,$3B,$7A,$FF,$C1,$11
	.dc.b	$FA,$08,$F2,$F6,$60,$CC,$C8,$C9,$EA,$EC,$E3,$F5,$12,$30,$43,$C0
	.dc.b	$E5,$F1,$47,$F8,$5F,$EA,$3F,$80,$C7,$F1,$40,$E1,$FE,$2E,$D7,$FC
	.dc.b	$06,$D7,$7E,$07,$3F,$02,$47,$FB,$C2,$7F,$84,$6B,$CA,$73,$3F,$C9
	.dc.b	$A7,$E9,$1A,$7A,$C7,$FA,$2E,$3F,$A7,$47,$95,$AF,$33,$FD,$27,$F4
	.dc.b	$AF,$A8,$FE,$5B,$7E,$B8,$C4,$FF,$86,$A3,$13,$A8,$FE,$67,$3A,$7F
	.dc.b	$89,$87,$1F,$05,$FC,$36,$3F,$C4,$70,$FA,$8F,$F5,$37,$87,$C9,$B3
	.dc.b	$AB,$AF,$38,$B5,$C6,$3F,$DC,$55,$C5,$A7,$12,$9C,$4B,$1C,$5F,$F7
	.dc.b	$2C,$71,$69,$C4,$A7,$12,$C7,$1B,$F6,$46,$5B,$1F,$02,$7A,$46,$2B
	.dc.b	$D2,$7A,$AE,$DF,$92,$2B,$DA,$2C,$F1,$8D,$1E,$A8,$CD,$76,$24,$7A
	.dc.b	$67,$83,$E8,$AD,$B9,$9D,$BE,$96,$DF,$A6,$3E,$BC,$F0,$74,$71,$32
	.dc.b	$79,$0E,$C7,$94,$C5,$78,$89,$5F,$88,$BE,$97,$F2,$E3,$C3,$18,$BC
	.dc.b	$B6,$F9,$CF,$16,$9E,$23,$C2,$7C,$93,$27,$A0,$95,$ED,$6D,$76,$7D
	.dc.b	$58,$7E,$86,$3F,$3D,$1D,$63,$0F,$E8,$CF,$C4,$38,$7F,$43,$3F,$8D
	.dc.b	$8C,$19,$C3,$F0,$FF,$4E,$FA,$8F,$F2,$9C,$61,$B8,$FD,$2C,$7F,$43
	.dc.b	$5B,$F8,$CF,$CD,$B7,$2F,$7E,$1F,$31,$FF,$47,$C6,$63,$A4,$FC,$2C
	.dc.b	$BE,$75,$6B,$1E,$ED,$9D,$61,$E0,$9C,$CB,$0E,$98,$F9,$63,$C7,$3A
	.dc.b	$9C,$C7,$D3,$C7,$C8,$F1,$9D,$1D,$31,$F1,$C7,$D5,$AE,$2E,$F1,$F1
	.dc.b	$BD,$98,$FC,$0C,$3F,$33,$3C,$0E,$35,$D0,$EA,$E6,$FC,$18,$FE,$6E
	.dc.b	$31,$87,$6C,$3F,$7F,$18,$7B,$1E,$BC,$3F,$CA,$31,$9C,$7D,$B8,$F6
	.dc.b	$C7,$DB,$87,$39,$FA,$38,$74,$9F,$0B,$F9,$4B,$E5,$87,$CB,$8D,$5F
	.dc.b	$BA,$FE,$4D,$6A,$7E,$B1,$E4,$3E,$31,$FD,$44,$A7,$F5,$37,$FC,$D5
	.dc.b	$F1,$79,$8C,$9F,$E6,$AF,$A8,$7D,$47,$F9,$53,$5F,$EB,$51,$87,$F3
	.dc.b	$D3,$FE,$DA,$8C,$3F,$CF,$9B,$1C,$27,$FC,$BB,$87,$F2,$49,$DB,$7D
	.dc.b	$5C,$E7,$E1,$AB,$87,$59,$AA,$F8,$D7,$9E,$73,$18,$9E,$18,$79,$9E
	.dc.b	$13,$EB,$78,$46,$67,$FD,$6E,$FB,$3E,$49,$F1,$5F,$E9,$D8,$FF,$7B
	.dc.b	$F5,$FF,$9D,$A7,$3F,$0D,$FE,$78,$8C,$56,$F0,$FF,$57,$56,$AF,$89
	.dc.b	$FF,$00,$1E,$BE,$9F,$18,$C5,$30,$C5,$AF,$75,$CA,$31,$5B,$8C,$4F
	.dc.b	$A3,$EA,$38,$BE,$AF,$A8,$B4,$9B,$18,$6D,$72,$9D,$36,$7C,$76,$FA
	.dc.b	$7F,$F7,$8C,$F8,$F8,$5F,$51,$FE,$87,$8F,$35,$6F,$DF,$F5,$D6,$B8
	.dc.b	$79,$E1,$CA,$BF,$DA,$93,$9B,$E3,$39,$DF,$FD,$4B,$39,$AF,$F6,$1C
	.dc.b	$E6,$BF,$D8,$2D,$87,$C8,$3E,$0B,$FD,$F1,$5D,$32,$F8,$7E,$77,$C4
	.dc.b	$6C,$7C,$16,$DF,$DA,$DB,$A7,$07,$E1,$2F,$85,$FC,$7F,$63,$E6,$9F
	.dc.b	$3B,$EA,$2C,$64,$98,$73,$8C,$BF,$ED,$CF,$A5,$FF,$76,$4F,$DD,$DD
	.dc.b	$F5,$1F,$8B,$7F,$D7,$73,$AA,$FF,$7D,$46,$B9,$BE,$AC,$E6,$3F,$29
	.dc.b	$5C,$6F,$FD,$CB,$39,$7F,$37,$19,$B0,$9D,$38,$6E,$FF,$2F,$D1,$EF
	.dc.b	$E2,$FE,$33,$EA,$26,$3F,$89,$F6,$EE,$3F,$26,$6E,$7F,$A6,$9B,$5F
	.dc.b	$A0,$3F,$2B,$F4,$1F,$96,$69,$F9,$32,$DF,$D3,$6D,$7C,$4F,$37,$F1
	.dc.b	$BF,$51,$FA,$3A,$F9,$91,$8E,$DF,$8A,$DB,$FC,$AE,$C8,$C5,$70,$9C
	.dc.b	$C7,$4E,$0E,$31,$FA,$E9,$7D,$47,$EB,$9B,$19,$F8,$AC,$4C,$8E,$0E
	.dc.b	$E3,$13,$FC,$18,$7D,$47,$3F,$8C,$E8,$DB,$F0,$36,$78,$37,$5F,$F7
	.dc.b	$FF,$D3,$BE,$27,$F8,$4F,$2B,$EA,$1F,$1F,$FB,$AA,$D6,$5F,$C2,$EB
	.dc.b	$53,$9B,$E3,$7C,$2B,$F7,$77,$C3,$26,$29,$FC,$4E,$3C,$9D,$E1,$CD
	.dc.b	$FE,$31,$F5,$0F,$A8,$7D,$46,$71,$E7,$5F,$92,$5F,$50,$E7,$FE,$23
	.dc.b	$EE,$FE,$5E,$73,$18,$E4,$FE,$2B,$E3,$3D,$AB,$CB,$86,$E5,$FC,$53
	.dc.b	$E9,$DF,$23,$98,$C3,$72,$F1,$9F,$A4,$97,$C0,$7F,$22,$FA,$87,$AB
	.dc.b	$FC,$92,$73,$87,$F2,$09,$CB,$DF,$E5,$35,$7B,$6C,$64,$98,$78,$07
	.dc.b	$D4,$3E,$A3,$84,$FF,$2C,$F2,$DD,$DF,$D1,$A1,$F4,$EF,$83,$FE,$6F
	.dc.b	$5A,$9F,$E9,$06,$A3,$35,$C6,$B8,$4F,$58,$C4,$EA,$30,$72,$97,$D4
	.dc.b	$67,$2F,$44,$BE,$1F,$C9,$3F,$3A,$3F,$A1,$99,$3C,$6E,$67,$8C,$F0
	.dc.b	$E1,$FE,$9F,$5B,$9F,$A6,$F8,$23,$33,$23,$C9,$F3,$D6,$BE,$57,$D4
	.dc.b	$3E,$17,$DC,$7F,$5A,$71,$5E,$B8,$C1,$A7,$FA,$CC,$6A,$7C,$39,$5F
	.dc.b	$07,$F4,$11,$FC,$A2,$71,$5F,$99,$6C,$7E,$96,$9F,$0B,$DA,$BF,$B2
	.dc.b	$F3,$7F,$17,$C3,$FC,$5D,$4F,$A5,$FE,$DF,$39,$AF,$F7,$74,$E5,$E3
	.dc.b	$E5,$FD,$9A,$DB,$63,$24,$C7,$D3,$D7,$A8,$FE,$EB,$E2,$7E,$16,$7F
	.dc.b	$B8,$FD,$43,$EA,$3F,$BF,$CE,$74,$FE,$E3,$39,$BE,$3C,$BF,$DF,$23
	.dc.b	$18,$7E,$66,$F8,$57,$F4,$CC,$55,$AC,$3E,$4D,$FF,$76,$C3,$B9,$FE
	.dc.b	$06,$F0,$7F,$BD,$46,$70,$F3,$FD,$07,$F8,$92,$DF,$51,$DA,$BF,$88
	.dc.b	$CE,$63,$11,$FE,$02,$8C,$61,$EC,$6D,$FC,$2D,$BD,$FC,$3F,$01,$B1
	.dc.b	$92,$5F,$ED,$4F,$EC,$FF,$09,$FD,$A3,$EA,$1F,$51,$FC,$67,$83,$EB
	.dc.b	$56,$C5,$6E,$75,$87,$E2,$AF,$8D,$7F,$21,$36,$1C,$BF,$E1,$C8,$E1
	.dc.b	$3F,$E7,$DA,$E9,$3F,$CC,$A7,$2F,$38,$CB,$C7,$0A,$FA,$8E,$D5,$E4
	.dc.b	$8C,$C7,$8A,$FC,$38,$7B,$1F,$1C,$71,$73,$87,$82,$3F,$9D,$61,$E0
	.dc.b	$AE,$18,$79,$E7,$B9,$E7,$C3,$BD,$6E,$F9,$4F,$5A,$F2,$3C,$0E,$A5
	.dc.b	$BE,$AF,$BC,$C4,$F2,$7D,$53,$CB,$0F,$4C,$7C,$E3,$FD,$45,$7F,$EA
	.dc.b	$37,$E2,$18,$3F,$D5,$5F,$4F,$FC,$77,$DD,$FF,$54,$36,$33,$D5,$FE
	.dc.b	$88,$6C,$EB,$E4,$BF,$B3,$C7,$26,$BF,$7E,$91,$88,$FE,$C5,$18,$E1
	.dc.b	$E4,$7D,$AF,$A8,$7D,$27,$F6,$7C,$FF,$0B,$18,$BF,$BF,$7C,$5F,$90
	.dc.b	$D9,$DD,$71,$D3,$FB,$85,$F1,$BD,$E3,$E8,$BF,$C8,$E1,$C6,$F7,$87
	.dc.b	$1B,$FF,$24,$E1,$C6,$F7,$87,$1B,$F5,$B6,$B7,$7C,$55,$F1,$AD,$E3
	.dc.b	$EA,$BE,$15,$AB,$E1,$7B,$8F,$F2,$CE,$3E,$56,$D7,$09,$D4,$71,$3A
	.dc.b	$5C,$7F,$19,$1A,$37,$5C,$E3,$81,$FD,$D6,$3E,$34,$FF,$1A,$6E,$FE
	.dc.b	$89,$E7,$87,$F8,$6E,$B4,$D9,$F9,$33,$FE,$A6,$BE,$DF,$51,$FE,$C1
	.dc.b	$8F,$F5,$4B,$5F,$F0,$23,$6F,$F9,$4F,$D3,$FF,$4C,$9D,$F5,$75,$7F
	.dc.b	$CF,$71,$FE,$5D,$7F,$D6,$6D,$79,$C6,$D5,$3E,$6F,$0D,$8F,$1C,$6E
	.dc.b	$9E,$06,$A2,$4E,$8E,$F1,$87,$FC,$39,$3F,$E1,$57,$C7,$E1,$E7,$1F
	.dc.b	$E8,$7F,$50,$E8,$F2,$36,$BF,$D9,$6D,$79,$F9,$9A,$99,$1A,$7C,$D1
	.dc.b	$88,$F0,$FD,$47,$A4,$C5,$FF,$52,$F9,$FF,$F3,$2C,$FF,$B6,$1A,$FF
	.dc.b	$8C,$DB,$1F,$43,$B3,$E4,$46,$38,$7F,$CB,$0D,$AF,$F5,$5B,$56,$9F
	.dc.b	$E3,$87,$B7,$B3,$F2,$FD,$09,$18,$3F,$DD,$3F,$11,$FE,$56,$6F,$93
	.dc.b	$F5,$E3,$11,$FE,$47,$E6,$FF,$33,$3A,$21,$FE,$4B,$8D,$E9,$FD,$C2
	.dc.b	$77,$39,$AF,$C1,$43,$E2,$29,$DE,$3F,$0D,$E7,$FE,$0E,$BF,$CA,$CD
	.dc.b	$BF,$EF,$6D,$D7,$FC,$F5,$F5,$1F,$E6,$6A,$FF,$66,$D6,$B9,$B9,$92
	.dc.b	$33,$5C,$3C,$AC,$47,$F9,$8E,$DF,$21,$F9,$76,$C7,$03,$FD,$B3,$8F
	.dc.b	$85,$FF,$6D,$C7,$C0,$BF,$F4,$67,$D4,$7E,$9E,$7F,$CD,$CD,$CB,$FC
	.dc.b	$27,$5B,$9F,$F7,$1C,$E6,$B8,$F8,$DF,$A7,$8C,$4F,$6D,$3F,$55,$F5
	.dc.b	$1F,$EB,$18,$F2,$07,$0F,$F9,$F5,$B8,$FF,$88,$AB,$9C,$EF,$C4,$E6
	.dc.b	$47,$9F,$D4,$3E,$23,$F7,$1F,$50,$F6,$A9,$2B,$C3,$97,$F9,$CD,$F1
	.dc.b	$CB,$E9,$38,$92,$33,$F5,$1F,$C0,$FE,$2B,$FD,$3F,$1E,$A6,$BF,$E9
	.dc.b	$56,$EB,$FE,$3A,$F8,$F7,$A3,$C7,$DC,$97,$C5,$FE,$1D,$3A,$8C,$1F
	.dc.b	$19,$7D,$46,$8E,$31,$FE,$A3,$FA,$8F,$8A,$7F,$9E,$DF,$9A,$D7,$5C
	.dc.b	$9F,$C6,$E3,$7E,$D5,$31,$19,$F1,$65,$72,$78,$7D,$0E,$23,$C1,$17
	.dc.b	$4F,$F2,$BE,$0A,$F8,$DF,$2B,$2F,$E3,$3E,$1E,$4C,$47,$C7,$8B,$7C
	.dc.b	$A6,$F4,$FF,$B7,$F0,$7D,$47,$A4,$CE,$01,$F4,$EF,$1F,$F9,$C3,$F6
	.dc.b	$E3,$F1,$ED,$7F,$D9,$4E,$DC,$C7,$0E,$59,$18,$38,$FD,$47,$42,$3D
	.dc.b	$3C,$2A,$F4,$8B,$A7,$F9,$EB,$0F,$D8,$4B,$EA,$1F,$31,$FD,$35,$B1
	.dc.b	$FA,$46,$BF,$ED,$FF,$A8,$FE,$0C,$FF,$B8,$63,$37,$FE,$86,$6E,$DF
	.dc.b	$0E,$7B,$5F,$F5,$9B,$E3,$7B,$8C,$7C,$BC,$AF,$04,$6F,$D7,$91,$FB
	.dc.b	$6B,$FF,$18,$B7,$C1,$CB,$E5,$9B,$87,$FA,$62,$BF,$59,$3E,$9E,$BF
	.dc.b	$17,$5C,$8E,$91,$E5,$30,$FB,$E3,$0F,$F0,$6F,$A7,$7D,$47,$F7,$36
	.dc.b	$C6,$63,$FB,$B7,$A9,$FC,$1A,$73,$2C,$7F,$D5,$71,$86,$D7,$0E,$BE
	.dc.b	$8F,$2F,$96,$97,$D4,$FE,$2F,$83,$B3,$BF,$64,$9C,$C7,$88,$C3,$E1
	.dc.b	$E2,$FA,$95,$FA,$58,$D9,$D3,$AF,$D3,$F5,$12,$30,$43,$91,$F9,$F9
	.dc.b	$46,$E3,$22,$3F,$91,$3F,$CB,$BE,$A3,$FC,$38,$FF,$AA,$5A,$FE,$1D
	.dc.b	$AF,$EA,$DA,$9F,$F1,$26,$BF,$ED,$4F,$D5,$7F,$C3,$B3,$FC,$E9,$AF
	.dc.b	$8A,$77,$8F,$E4,$E7,$35,$C7,$D9,$FF,$16,$46,$27,$C5,$AE,$27,$FD
	.dc.b	$1D,$3D,$30,$FF,$60,$35,$E9,$1F,$E9,$3C,$3F,$AC,$47,$B7,$7C,$11
	.dc.b	$FF,$64,$1E,$D7,$FD,$63,$E0,$7E,$9F,$E9,$FF,$DF,$93,$FE,$D8,$8C
	.dc.b	$7C,$46,$A3,$81,$93,$F8,$84,$F0,$7F,$DB,$33,$C1,$D0,$8D,$7C,$1A
	.dc.b	$7E,$85,$F5,$AD,$5D,$BD,$D7,$39,$EA,$78,$8A,$70,$3B,$C7,$07,$CB
	.dc.b	$3D,$70,$FE,$F9,$87,$86,$72,$70,$9F,$01,$BC,$2C,$64,$87,$82,$B9
	.dc.b	$E1,$E5,$BE,$93,$E3,$3C,$B8,$78,$DF,$C5,$5F,$47,$38,$5A,$E3,$87
	.dc.b	$49,$F2,$9A,$3C,$31,$AA,$E1,$87,$84,$93,$F1,$67,$C7,$87,$84,$98
	.dc.b	$7B,$67,$9D,$70,$7D,$15,$D5,$C9,$E2,$79,$CF,$18,$F2,$12,$32,$FE
	.dc.b	$03,$0F,$53,$D2,$BB,$5F,$C3,$9C,$BE,$49,$EE,$7C,$62,$B9,$7E,$86
	.dc.b	$1F,$01,$F2,$57,$A6,$B9,$E1,$99,$F0,$1F,$0E,$7F,$15,$A7,$F4,$58
	.dc.b	$FA,$CD,$C3,$F3,$4D,$AF,$E9,$4D,$9F,$ED,$CD,$4D,$23,$EE,$F1,$7F
	.dc.b	$34,$78,$EB,$E9,$38,$8F,$6D,$57,$F0,$F1,$B8,$C6,$5F,$6A,$31,$3F
	.dc.b	$9E,$73,$87,$CC,$AB,$3F,$B4,$9D,$BD,$0E,$07,$80,$A6,$4F,$B5,$3E
	.dc.b	$97,$35,$F7,$1C,$9D,$4C,$C7,$85,$A6,$4D,$9F,$48,$FE,$05,$3E,$67
	.dc.b	$17,$F1,$1E,$26,$2F,$51,$FE,$40,$8F,$41,$E7,$BF,$88,$E1,$AE,$6F
	.dc.b	$C8,$D3,$C9,$1E,$B2,$47,$F2,$1B,$EF,$7E,$06,$D7,$C6,$C3,$F0,$D1
	.dc.b	$F8,$48,$E0,$7E,$10,$A7,$E1,$0F,$F0,$97,$D3,$3E,$9F,$FC,$33,$18
	.dc.b	$6B,$FE,$39,$87,$92,$F8,$AF,$E6,$F2,$9E,$5E,$AE,$1D,$CF,$F2,$E6
	.dc.b	$CF,$F2,$4F,$A6,$FF,$37,$E5,$9C,$3A,$4E,$7B,$7C,$DB,$7F,$18,$C3
	.dc.b	$8B,$FC,$6B,$0E,$35,$FC,$BA,$73,$B7,$F8,$B6,$73,$8F,$EA,$DB,$59
	.dc.b	$C3,$E1,$4E,$5B,$1F,$E5,$7F,$8F,$7C,$47,$FA,$56,$BF,$D6,$D1,$8F
	.dc.b	$84,$F9,$5D,$1F,$2E,$37,$E9,$E2,$B9,$5A,$F1,$82,$F7,$5D,$2B,$4D
	.dc.b	$BD,$CF,$78,$D7,$D3,$62,$9F,$4D,$63,$0D,$9E,$8D,$C3,$FC,$2B,$7E
	.dc.b	$C8,$F1,$C3,$E9,$7F,$DD,$15,$FE,$D3,$E2,$7D,$37,$F6,$38,$FF,$61
	.dc.b	$CE,$AB,$FC,$3F,$5A,$E0,$E7,$39,$DB,$F1,$D3,$9B,$E3,$EF,$7C,$28
	.dc.b	$C6,$1F,$E2,$4B,$E1,$5F,$E6,$56,$C3,$E3,$9F,$02,$F7,$FB,$E3,$F0
	.dc.b	$FC,$DF,$88,$DC,$7E,$63,$70,$F8,$3C,$3D,$38,$3E,$ED,$F0,$BF,$A5
	.dc.b	$EC,$7C,$D3,$E7,$7D,$35,$8C,$93,$0E,$71,$97,$FC,$7B,$F4,$7F,$ED
	.dc.b	$D9,$FF,$7E,$70,$3E,$9B,$F1,$55,$AF,$27,$F0,$73,$9E,$7F,$A7,$87
	.dc.b	$1A,$FF,$25,$CE,$70,$FF,$79,$CE,$63,$F2,$38,$74,$7F,$79,$49,$D5
	.dc.b	$7F,$A3,$DB,$1F,$2F,$81,$E9,$78,$32,$FC,$67,$D3,$4C,$3F,$29,$DF
	.dc.b	$70,$F9,$33,$72,$FD,$07,$0F,$F0,$4B,$FC,$D4,$FE,$55,$FC,$B7,$CE
	.dc.b	$57,$F3,$CD,$BF,$A6,$DB,$E6,$FC,$2F,$2B,$F0,$DF,$4D,$FA,$4A,$F9
	.dc.b	$B1,$8E,$ED,$61,$F9,$7A,$D6,$1F,$0B,$75,$EE,$BE,$4B,$C3,$0F,$CC
	.dc.b	$DF,$C9,$9F,$8D,$3F,$B0,$97,$D3,$53,$1F,$13,$6B,$F8,$1C,$E9,$B3
	.dc.b	$A6,$9A,$FA,$6E,$9F,$1B,$D5,$B1,$93,$9C,$65,$A7,$FB,$A7,$3F,$BB
	.dc.b	$F4,$AF,$A5,$FE,$17,$DC,$FA,$67,$C7,$FE,$F2,$B5,$87,$13,$FD,$E5
	.dc.b	$5B,$9D,$5F,$2B,$E3,$5B,$D3,$FD,$F1,$87,$27,$FD,$3B,$87,$19,$D4
	.dc.b	$27,$58,$F9,$B0,$FE,$3F,$F4,$CF,$A6,$7D,$37,$83,$1E,$3F,$4C,$E6
	.dc.b	$FE,$27,$E5,$FD,$3D,$02,$75,$5E,$09,$D4,$FF,$53,$F6,$BF,$8D,$F9
	.dc.b	$0E,$5B,$18,$7F,$8F,$FD,$33,$E4,$73,$8F,$AA,$30,$FB,$2F,$FD,$94
	.dc.b	$E7,$E0,$3F,$93,$FD,$33,$D4,$E7,$18,$7A,$F9,$98,$AE,$8D,$D3,$FD
	.dc.b	$83,$8F,$F1,$99,$C9,$F2,$BE,$99,$F4,$DC,$A3,$E6,$F9,$AE,$CF,$E9
	.dc.b	$30,$FA,$4F,$F6,$53,$A8,$CC,$63,$1F,$F6,$7D,$71,$AF,$B8,$71,$F8
	.dc.b	$0F,$E8,$31,$99,$FF,$6C,$CE,$4E,$84,$73,$4F,$A6,$E3,$87,$F4,$AA
	.dc.b	$7C,$47,$9A,$7E,$5D,$FF,$B8,$27,$27,$8A,$33,$D9,$FC,$62,$33,$3E
	.dc.b	$29,$CC,$7F,$97,$AB,$A7,$1F,$76,$47,$B3,$17,$C0,$E3,$0F,$F7,$7F
	.dc.b	$D3,$3E,$4B,$E2,$9F,$D8,$1C,$5F,$18,$C1,$A7,$F3,$B1,$A9,$F8,$33
	.dc.b	$A9,$CB,$F4,$3E,$0F,$13,$FE,$FE,$E4,$FE,$27,$5F,$E7,$4C,$5F,$49
	.dc.b	$FD,$C2,$73,$5F,$E7,$69,$CD,$7E,$B2,$72,$FD,$FA,$E1,$E5,$FD,$E6
	.dc.b	$C6,$E3,$F0,$3D,$38,$C1,$C1,$FD,$13,$C4,$FB,$CF,$F7,$4F,$A6,$7C
	.dc.b	$1F,$F8,$0A,$73,$8F,$EC,$30,$56,$A3,$F7,$F3,$AC,$3F,$C8,$53,$97
	.dc.b	$BC,$FE,$E1,$FE,$F1,$C9,$FD,$FE,$75,$8F,$8E,$3C,$07,$F8,$0F,$BF
	.dc.b	$FB,$FD,$70,$BE,$FF,$3F,$FD,$9B,$E9,$BC,$13,$FE,$03,$8C,$5F,$A5
	.dc.b	$B5,$C0,$FE,$E3,$5C,$32,$F9,$B7,$E6,$3F,$91,$F0,$FC,$1D,$FE,$13
	.dc.b	$6F,$DF,$1F,$E2,$69,$FE,$27,$F0,$7F,$C5,$3E,$99,$F4,$DE,$FC,$7A
	.dc.b	$ED,$FD,$CA,$F6,$2B,$6F,$F3,$2C,$3F,$CE,$51,$88,$F9,$18,$75,$C3
	.dc.b	$A5,$69,$FF,$62,$4A,$74,$F3,$C6,$BC,$8F,$04,$F2,$8C,$C7,$92,$FC
	.dc.b	$18,$FC,$47,$C5,$19,$AE,$0F,$23,$FD,$A9,$7D,$AB,$84,$F9,$23,$18
	.dc.b	$79,$EB,$67,$13,$FC,$74,$E5,$F3,$5F,$8A,$31,$1E,$13,$D8,$71,$30
	.dc.b	$72,$C3,$D6,$FC,$33,$B6,$1E,$47,$C3,$5D,$67,$77,$FE,$98,$9F,$AA
	.dc.b	$7F,$A6,$9C,$3C,$CF,$F7,$97,$D2,$FF,$1C,$F5,$7F,$D3,$ED,$8F,$A6
	.dc.b	$D7,$7D,$5F,$E8,$87,$3F,$23,$FB,$26,$EB,$CA,$31,$1D,$9A,$FE,$A6
	.dc.b	$91,$C7,$2F,$C3,$7D,$33,$E8,$BF,$5D,$18,$6E,$FC,$6B,$83,$71,$FC
	.dc.b	$2D,$F1,$C6,$DF,$07,$FB,$7E,$3E,$0A,$FC,$5B,$63,$ED,$57,$16,$D7
	.dc.b	$15,$3B,$C7,$F3,$8D,$C3,$F8,$25,$F1,$9F,$45,$76,$C7,$CE,$F8,$9A
	.dc.b	$A7,$85,$72,$C3,$C1,$87,$5A,$E1,$8F,$B1,$87,$E4,$63,$4F,$9B,$0E
	.dc.b	$B8,$74,$8F,$F0,$7B,$5D,$57,$23,$FC,$DB,$1F,$05,$B7,$FC,$7D,$FF
	.dc.b	$38,$DF,$69,$D6,$BE,$7B,$FF,$13,$C6,$1F,$39,$FA,$5F,$A6,$FF,$63
	.dc.b	$3F,$E2,$C6,$BF,$C5,$C3,$8B,$FF,$3F,$FD,$23,$8B,$C5,$87,$F3,$1A
	.dc.b	$D3,$FC,$CE,$7F,$CC,$77,$FE,$38,$6B,$D6,$76,$FF,$8C,$17,$CC,$F4
	.dc.b	$8C,$35,$F1,$B6,$7F,$B8,$C7,$A6,$35,$3C,$A3,$07,$87,$C0,$CC,$F1
	.dc.b	$F1,$24,$E8,$F8,$1F,$4C,$E8,$F0,$36,$3F,$CA,$2D,$7E,$36,$2F,$37
	.dc.b	$75,$27,$6F,$9E,$31,$1C,$3E,$9B,$D4,$63,$6F,$F5,$5F,$CE,$FF,$9B
	.dc.b	$27,$FC,$BE,$D7,$EA,$36,$3E,$9F,$67,$C5,$8C,$6F,$FD,$8D,$B5,$FE
	.dc.b	$5C,$6A,$CB,$FC,$7C,$F6,$F6,$3E,$67,$A1,$23,$07,$F9,$C3,$E0,$BF
	.dc.b	$CB,$4D,$F2,$7E,$A4,$62,$3F,$11,$CF,$FE,$6D,$74,$43,$F1,$91,$BD
	.dc.b	$3F,$D9,$33,$B9,$CD,$7D,$F8,$7C,$45,$3B,$C7,$E0,$FD,$0F,$BD,$5F
	.dc.b	$97,$6D,$FF,$B5,$1B,$AF,$F8,$17,$E9,$7F,$CD,$55,$FE,$8A,$9C,$BF
	.dc.b	$E8,$06,$BF,$EF,$FE,$5F,$C8,$4F,$08,$ED,$18,$FA,$69,$1C,$E3,$78
	.dc.b	$FE,$5F,$E9,$BF,$4D,$3F,$E9,$86,$BF,$AD,$BF,$85,$8F,$F8,$66,$73
	.dc.b	$97,$FB,$B1,$B1,$9E,$AE,$B7,$FE,$A2,$FA,$5F,$F5,$A3,$7D,$CF,$D3
	.dc.b	$46,$23,$F7,$57,$C9,$FF,$4E,$C7,$F9,$A1,$E5,$E1,$70,$3F,$5D,$F4
	.dc.b	$CF,$87,$FD,$87,$D3,$3E,$2E,$92,$F8,$9F,$EB,$48,$7D,$1F,$02,$46
	.dc.b	$7E,$9B,$F7,$5F,$16,$F7,$69,$27,$FD,$7B,$F1,$EF,$47,$11,$FC,$4E
	.dc.b	$B8,$3F,$EC,$26,$BF,$C5,$E7,$83,$9A,$FE,$11,$0F,$99,$A7,$4E,$47
	.dc.b	$27,$F9,$27,$E9,$BE,$36,$07,$FC,$F4,$FF,$B4,$1A,$FF,$25,$76,$4F
	.dc.b	$6B,$F8,$B3,$AA,$EB,$3F,$EC,$C8,$DF,$8B,$A2,$47,$5F,$9F,$AF,$58
	.dc.b	$B5,$CA,$3C,$93,$F2,$BE,$56,$47,$C3,$7C,$3C,$78,$8F,$91,$16,$B9
	.dc.b	$61,$F7,$67,$81,$ED,$FA,$6F,$49,$BC,$FF,$CD,$9F,$4A,$F1,$FF,$99
	.dc.b	$3F,$7A,$BF,$DE,$6D,$7F,$A4,$3B,$73,$1C,$39,$24,$60,$E3,$F4,$DD
	.dc.b	$08,$F4,$F0,$2B,$D2,$2E,$7F,$0F,$0F,$DB,$FD,$33,$E6,$3F,$A3,$36
	.dc.b	$FF,$DF,$75,$F2,$63,$11,$FD,$77,$E9,$BF,$79,$39,$AE,$24,$9E,$3A
	.dc.b	$7F,$A3,$6B,$6F,$FA,$3E,$F9,$DF,$D5,$97,$CB,$FF,$57,$6D,$F2,$02
	.dc.b	$B8,$DF,$5F,$5B,$81,$27,$80,$AF,$DC,$3E,$98,$D3,$4D,$C6,$B0,$F2
	.dc.b	$75,$78,$FA,$65,$6A,$3C,$C6,$1F,$89,$18,$7C,$3F,$4C,$FA,$6F,$EE
	.dc.b	$4D,$8C,$C7,$F7,$5F,$53,$F5,$33,$99,$FF,$05,$37,$0F,$CD,$37,$3F
	.dc.b	$F5,$C5,$70,$BE,$5F,$2D,$30,$EA,$7C,$29,$FE,$DC,$DE,$9F,$01,$E1
	.dc.b	$EA,$91,$92,$1F,$E1,$B9,$E2,$DE,$AF,$47,$D3,$48,$ED,$1E,$B3,$0F
	.dc.b	$18,$C3,$F5,$3E,$97,$FC,$37,$AF,$ED,$5B,$A7,$FB,$4D,$B3,$F8,$07
	.dc.b	$FD,$61,$3B,$CF,$FD,$09,$1F,$E2,$49,$FF,$6D,$07,$0F,$38,$FF,$45
	.dc.b	$7D,$37,$F8,$82,$3F,$C6,$CD,$AF,$E0,$CD,$BE,$13,$F8,$A8,$C3,$F9
	.dc.b	$26,$A6,$DF,$A2,$3C,$5E,$57,$F8,$C6,$3F,$92,$93,$BF,$11,$27,$E2
	.dc.b	$3B,$8F,$F1,$C7,$D3,$7F,$C0,$09,$FF,$6D,$F8,$1F,$EF,$78,$C4,$7F
	.dc.b	$B3,$63,$1F,$0F,$FE,$D4,$7E,$55,$D9,$FD,$4D,$70,$F8,$25,$70,$9E
	.dc.b	$A6,$AF,$83,$EA,$32,$2B,$94,$F1,$8F,$11,$C4,$DE,$3F,$DE,$2F,$BC
	.dc.b	$64,$D9,$CE,$FB,$DF,$48,$FC,$06,$18,$99,$1E,$29,$D3,$E5,$39,$B8
	.dc.b	$24,$7A,$AF,$C5,$87,$9F,$0F,$1D,$F8,$4E,$95,$BC,$3B,$C7,$48,$B3
	.dc.b	$C2,$F9,$CF,$AE,$39,$E1,$E0,$8F,$83,$7C,$30,$F0,$12,$3D,$A7,$80
	.dc.b	$87,$73,$89,$8A,$EB,$87,$38,$E2,$6C,$91,$A7,$89,$89,$F0,$3E,$F3
	.dc.b	$E8,$E1,$EA,$71,$3C,$5E,$07,$67,$04,$9E,$73,$C7,$0F,$81,$5C,$0F
	.dc.b	$A9,$5B,$7E,$35,$70,$C3,$D5,$1E,$08,$C3,$EF,$7C,$F5,$F7,$AF,$B6
	.dc.b	$1C,$23,$17,$63,$9C,$EA,$F9,$1E,$03,$D5,$7D,$27,$75,$C6,$76,$7A
	.dc.b	$27,$4E,$2B,$A3,$ED,$BF,$0B,$C1,$F8,$38,$7E,$16,$FA,$4F,$C3,$AE
	.dc.b	$33,$F0,$E3,$E0,$1E,$43,$CC,$60,$91,$C2,$FC,$71,$F6,$2B,$AD,$73
	.dc.b	$AD,$4F,$2B,$F2,$BE,$7C,$3D,$EF,$9D,$AE,$4F,$75,$F4,$C3,$E9,$C7
	.dc.b	$EB,$3C,$3F,$DA,$91,$F8,$0D,$74,$7C,$AF,$A5,$7D,$37,$FA,$23,$3F
	.dc.b	$D7,$C6,$3E,$2F,$F8,$5B,$6B,$F7,$3F,$51,$FC,$42,$31,$1F,$CF,$1A
	.dc.b	$FF,$80,$82,$F2,$5F,$17,$FD,$0A,$7F,$C0,$F3,$C1,$FF,$04,$FD,$3F
	.dc.b	$F9,$97,$8F,$C3,$7C,$27,$F8,$9C,$E5,$FE,$2B,$5D,$6F,$F9,$0C,$E7
	.dc.b	$2F,$5D,$F1,$7F,$85,$C6,$75,$F3,$5F,$1C,$3F,$92,$FC,$8B,$E2,$7F
	.dc.b	$D0,$99,$FF,$87,$63,$1F,$1D,$FE,$60,$E7,$FF,$30,$C6,$FD,$1C,$57
	.dc.b	$F8,$8A,$36,$F1,$BE,$78,$2F,$8C,$30,$E3,$2B,$E9,$7C,$67,$31,$E5
	.dc.b	$FA,$7B,$6F,$A7,$B1,$86,$DF,$85,$B8,$78,$63,$CD,$5C,$5F,$F1,$9F
	.dc.b	$D3,$7F,$B5,$1F,$F1,$E7,$0B,$E9,$FF,$A9,$C6,$BD,$FF,$3C,$E7,$83
	.dc.b	$D3,$8B,$0E,$4F,$73,$F0,$38,$72,$70,$6E,$F9,$C6,$76,$FF,$42,$57
	.dc.b	$26,$CF,$C6,$F8,$F7,$C1,$3E,$07,$C3,$7B,$D3,$DB,$E7,$FC,$6C,$3E
	.dc.b	$3B,$63,$EA,$37,$3F,$37,$0F,$E0,$63,$19,$FD,$6F,$5F,$E7,$9F,$43
	.dc.b	$E9,$EE,$1F,$76,$2E,$1C,$B8,$55,$8B,$FF,$4F,$7D,$2F,$FB,$47,$0F
	.dc.b	$F3,$44,$3E,$9F,$F0,$AE,$BC,$8F,$C1,$B9,$E8,$F9,$B7,$C7,$2F,$99
	.dc.b	$39,$C3,$FC,$D7,$39,$8F,$CB,$61,$D2,$7F,$69,$39,$6E,$1F,$2F,$87
	.dc.b	$FC,$FB,$1C,$A3,$F1,$BF,$4F,$23,$F3,$5D,$B7,$2F,$74,$DC,$BF,$37
	.dc.b	$36,$BF,$1A,$7E,$5B,$E7,$FF,$2E,$D3,$F4,$85,$BF,$90,$DA,$F8,$FE
	.dc.b	$4F,$E3,$BE,$9F,$F4,$D3,$EE,$8C,$76,$FC,$46,$C6,$4F,$CC,$CF,$0B
	.dc.b	$FC,$7E,$AA,$E1,$5A,$DE,$D7,$E0,$87,$D3,$FE,$C1,$B3,$FC,$0D,$7C
	.dc.b	$56,$66,$4E,$5E,$11,$89,$E9,$F4,$FC,$FE,$3B,$A3,$6F,$A3,$67,$71
	.dc.b	$DE,$7F,$5B,$F4,$CF,$89,$FD,$17,$D3,$BE,$3F,$31,$86,$DF,$FA,$D7
	.dc.b	$C0,$56,$E3,$FD,$63,$7C,$DD,$46,$71,$FD,$AC,$E5,$B8,$7F,$16,$FA
	.dc.b	$77,$D3,$BE,$9F,$BD,$7E,$F3,$E9,$DC,$FF,$C1,$3D,$EF,$C9,$4F,$FB
	.dc.b	$1A,$73,$A7,$FB,$0A,$73,$3F,$D1,$BD,$AF,$E1,$7E,$4F,$CE,$AF,$F6
	.dc.b	$3A,$8C,$F1,$FF,$10,$FA,$67,$C9,$70,$9C,$36,$FF,$D9,$F2,$F8,$0F
	.dc.b	$E2,$FF,$4E,$F5,$39,$46,$1E,$9E,$66,$2F,$8B,$74,$FE,$33,$87,$F1
	.dc.b	$19,$C9,$FE,$DC,$FA,$77,$D3,$F2,$8F,$F6,$FF,$98,$EB,$FE,$7F,$0F
	.dc.b	$A5,$7C,$0F,$F2,$AA,$D4,$FF,$BB,$4D,$46,$6B,$8D,$70,$9E,$B1,$89
	.dc.b	$D4,$60,$E5,$2F,$A7,$CE,$7F,$CC,$A5,$F2,$9E,$49,$F9,$71,$FE,$F6
	.dc.b	$32,$78,$DC,$CF,$19,$E1,$C3,$FD,$0A,$B7,$3F,$D0,$5F,$04,$66,$64
	.dc.b	$79,$3E,$7A,$D7,$F4,$4F,$A7,$7C,$1F,$B8,$FE,$9A,$E2,$BD,$71,$83
	.dc.b	$4A,$79,$46,$1B,$19,$F2,$8F,$F8,$5E,$BA,$4F,$9F,$37,$C0,$75,$AE
	.dc.b	$07,$F8,$77,$93,$F8,$66,$1C,$AF,$9D,$BE,$97,$FB,$1D,$6A,$FC,$87
	.dc.b	$DF,$8C,$F9,$9F,$69,$B1,$F5,$9B,$7F,$D9,$AB,$93,$DB,$AF,$38,$F8
	.dc.b	$61,$E2,$FE,$21,$FE,$E1,$2F,$A7,$7C,$1F,$F7,$6A,$D5,$FA,$6B,$53
	.dc.b	$9E,$5F,$EF,$B7,$C3,$3F,$83,$8F,$CD,$C7,$8C,$7C,$FC,$38,$BA,$9F
	.dc.b	$1D,$F3,$8F,$F0,$5F,$7F,$F7,$59,$E5,$3E,$3F,$A0,$FE,$BF,$F4,$FD
	.dc.b	$27,$B4,$62,$75,$7F,$DA,$2B,$55,$F0,$2B,$53,$E1,$C7,$E9,$37,$87
	.dc.b	$E0,$6F,$F0,$5B,$1B,$8F,$F0,$CD,$FE,$27,$E0,$BF,$15,$F4,$EF,$A7
	.dc.b	$FF,$2E,$63,$D2,$BF,$C3,$26,$C4,$ED,$F8,$D7,$F8,$58,$C4,$7F,$99
	.dc.b	$6F,$AD,$F4,$AD,$3F,$E6,$58,$CC,$F2,$C2,$BC,$4E,$F3,$C6,$33,$1E
	.dc.b	$3B,$EF,$87,$9D,$F0,$C6,$6B,$83,$C8,$FF,$3C,$0B,$EF,$5C,$67,$C9
	.dc.b	$3B,$3F,$1F,$87,$A1,$E3,$5C,$23,$11,$E7,$BF,$1C,$62,$3C,$47,$B4
	.dc.b	$E8,$60,$EB,$87,$B1,$F2,$1E,$1C,$3C,$8F,$80,$C4,$F5,$76,$7F,$A3
	.dc.b	$CA,$7A,$8C,$5E,$7E,$9F,$F9,$0F,$B7,$7A,$BF,$D0,$55,$FE,$97,$9C
	.dc.b	$B6,$32,$FC,$88,$7C,$8F,$F9,$F6,$38,$35,$D3,$5F,$F5,$34,$70,$E5
	.dc.b	$F1,$4F,$1D,$3F,$0F,$F4,$EF,$A3,$FF,$56,$46,$1B,$BF,$2A,$E0,$DB
	.dc.b	$FC,$3D,$F1,$C6,$DF,$07,$F9,$36,$3E,$17,$EE,$37,$1E,$D1,$C0,$FC
	.dc.b	$8C,$67,$0E,$D1,$F8,$3B,$EC,$DB,$EC,$AB,$83,$E0,$C7,$CE,$FF,$37
	.dc.b	$AE,$38,$78,$23,$C7,$3A,$79,$36,$77,$87,$FB,$3D,$B3,$D7,$1F,$3D
	.dc.b	$6F,$1F,$1D,$6A,$B9,$D7,$B5,$FF,$49,$C7,$AA,$3F,$96,$37,$7F,$7E
	.dc.b	$3E,$59,$CD,$FF,$B7,$DB,$7F,$C8,$72,$F7,$AF,$A7,$FF,$5F,$3F,$C9
	.dc.b	$5A,$FF,$A8,$9B,$3F,$CE,$FE,$97,$FA,$8C,$EF,$AF,$B6,$5F,$EE,$D9
	.dc.b	$FF,$2D,$4F,$F2,$F6,$BC,$E7,$6F,$F6,$3B,$7C,$D6,$23,$C8,$D8,$CC
	.dc.b	$73,$F0,$31,$32,$3D,$0F,$1B,$4E,$67,$FD,$77,$1F,$E1,$A7,$C9,$E1
	.dc.b	$E7,$2F,$EA,$DF,$4E,$E8,$F1,$36,$BF,$A4,$35,$E7,$E6,$6A,$64,$69
	.dc.b	$F1,$46,$23,$C3,$F4,$FE,$53,$17,$FD,$97,$E7,$BF,$CC,$D3,$FD,$55
	.dc.b	$AF,$FB,$49,$B1,$F4,$7B,$7E,$34,$63,$83,$FC,$19,$48,$C3,$FE,$0F
	.dc.b	$6A,$D3,$FC,$7C,$F7,$F6,$3E,$47,$A1,$23,$07,$F6,$7F,$82,$FF,$2C
	.dc.b	$B7,$C9,$FA,$D1,$88,$FF,$73,$F3,$7F,$9B,$1D,$10,$FF,$75,$C6,$F4
	.dc.b	$FF,$02,$4E,$E7,$35,$F8,$18,$7C,$3D,$3B,$C7,$E1,$7D,$0F,$C1,$D7
	.dc.b	$FB,$D9,$B7,$FE,$0F,$6E,$BF,$13,$E9,$BF,$CD,$55,$FE,$01,$9C,$BF
	.dc.b	$E3,$36,$BF,$EF,$7E,$6F,$C9,$4F,$08,$ED,$18,$FA,$79,$1C,$E3,$78
	.dc.b	$FE,$67,$E9,$FF,$4B,$2C,$BF,$C4,$51,$87,$FC,$49,$87,$C3,$CB,$EC
	.dc.b	$CE,$5B,$E2,$F3,$3A,$DF,$F8,$67,$E9,$FF,$D5,$CD,$F7,$3F,$51,$18
	.dc.b	$8F,$05,$72,$78,$78,$7A,$23,$FE,$23,$FA,$77,$C4,$7E,$C7,$E9,$DF
	.dc.b	$09,$49,$5C,$8F,$F1,$54,$3E,$93,$81,$23,$3F,$4F,$FB,$BF,$8A,$FF
	.dc.b	$4E,$47,$F8,$C5,$AF,$98,$FF,$32,$38,$E1,$FF,$24,$7C,$73,$D1,$E5
	.dc.b	$EF,$4B,$E4,$FF,$07,$9D,$46,$0F,$F1,$F2,$F9,$9D,$1C,$63,$C0,$3E
	.dc.b	$9F,$E8,$3F,$E7,$57,$FC,$FA,$FF,$91,$DA,$F0,$76,$F1,$E4,$FE,$31
	.dc.b	$1A,$C7,$FC,$C0,$DE,$2F,$DA,$BC,$0F,$F6,$35,$75,$F1,$38,$9C,$BE
	.dc.b	$7E,$C6,$48,$78,$5F,$99,$F2,$B2,$3E,$63,$E2,$E3,$C4,$7B,$B7,$4E
	.dc.b	$32,$F2,$D5,$CB,$CB,$F4,$FE,$A3,$18,$7F,$A5,$FE,$9B,$F8,$63,$FE
	.dc.b	$69,$7F,$00,$DE,$3F,$E7,$11,$A8,$F9,$AD,$C3,$FD,$04,$F0,$8C,$CF
	.dc.b	$0F,$1B,$8F,$D3,$C9,$EC,$F4,$EF,$C4,$74,$8B,$87,$49,$FE,$91,$F4
	.dc.b	$CF,$97,$A4,$BF,$F4,$07,$D3,$7E,$EE,$FF,$D2,$95,$B9,$FF,$42,$35
	.dc.b	$FC,$2B,$63,$FD,$35,$3C,$75,$FE,$5F,$39,$7F,$98,$C3,$E9,$FF,$AD
	.dc.b	$36,$FA,$E1,$8F,$5E,$47,$91,$DF,$7E,$23,$F2,$B5,$FE,$38,$74,$7A
	.dc.b	$A3,$AF,$5F,$6E,$B9,$3E,$E8,$CB,$F0,$23,$2F,$48,$FE,$DD,$F4,$EF
	.dc.b	$A7,$A4,$9F,$F5,$37,$A7,$F9,$E9,$D4,$7F,$AA,$5A,$FE,$A9,$BB,$FF
	.dc.b	$AF,$E7,$85,$7F,$7B,$7F,$C0,$31,$9F,$96,$ED,$7F,$B2,$9F,$D1,$B7
	.dc.b	$AB,$81,$C7,$AE,$47,$92,$32,$F9,$A3,$B7,$4F,$7F,$A7,$93,$EA,$8C
	.dc.b	$BE,$C8,$CB,$E1,$8F,$F0,$87,$D3,$7F,$A8,$B2,$FF,$6C,$3E,$8E,$4F
	.dc.b	$ED,$AD,$AF,$F7,$D3,$B8,$CF,$BB,$CB,$49,$5F,$EC,$FA,$D3,$C2,$3F
	.dc.b	$D0,$0A,$3F,$C4,$AF,$FA,$79,$AF,$82,$37,$1E,$18,$7D,$3F,$F8,$79
	.dc.b	$FF,$6E,$35,$FE,$3B,$1B,$3F,$89,$3E,$8F,$07,$DE,$7C,$0F,$8A,$91
	.dc.b	$E7,$8C,$BE,$98,$AF,$FA,$C7,$C0,$FF,$07,$FD,$3F,$FB,$E2,$3F,$DA
	.dc.b	$91,$8F,$88,$FF,$65,$3F,$22,$EC,$F8,$4D,$81,$5C,$DE,$C6,$5C,$7C
	.dc.b	$1F,$F1,$A8,$E4,$68,$F9,$53,$D9,$E7,$76,$B8,$DF,$8E,$7C,$27,$02
	.dc.b	$C6,$9F,$04,$F3,$C7,$FB,$DE,$1E,$19,$E0,$78,$6B,$83,$E0,$3A,$CF
	.dc.b	$0B,$F0,$46,$EF,$C8,$75,$C2,$47,$8A,$32,$F2,$8C,$BC,$4D,$19,$21
	.dc.b	$92,$61,$E7,$BF,$14,$7C,$88,$F5,$61,$E8,$3D,$33,$A9,$F6,$CE,$AF
	.dc.b	$C7,$1F,$3E,$74,$6E,$B9,$4F,$9A,$BC,$17,$C1,$E5,$7E,$67,$2F,$A8
	.dc.b	$F3,$E1,$F1,$A3,$31,$83,$E9,$9C,$63,$47,$E3,$DE,$51,$C0,$91,$98
	.dc.b	$E4,$7B,$1E,$2F,$43,$38,$7B,$2B,$D1,$3C,$6B,$9D,$7C,$98,$C1,$1C
	.dc.b	$10,$FC,$11,$D2,$7E,$24,$6E,$33,$3D,$1F,$21,$E3,$7E,$3B,$F3,$E7
	.dc.b	$8B,$99,$E5,$5E,$78,$F8,$31,$89,$DC,$76,$8C,$3F,$24,$E4,$43,$C2
	.dc.b	$4C,$3E,$DB,$B3,$18,$7C,$47,$DD,$7F,$81,$8C,$47,$28,$C3,$E4,$37
	.dc.b	$7A,$AF,$C7,$BE,$67,$BC,$7B,$CF,$03,$8C,$3E,$35,$74,$9F,$98,$6A
	.dc.b	$FC,$31,$F0,$E3,$EF,$D7,$67,$47,$8C,$F5,$9F,$C0,$3C,$6F,$E0,$C7
	.dc.b	$96,$AC,$64,$E1,$2F,$A6,$FE,$1D,$F4,$AF,$A7,$FF,$46,$E5,$FC,$16
	.dc.b	$31,$F1,$9F,$C3,$2B,$F8,$91,$C2,$B7,$AF,$FA,$E2,$B7,$8F,$F0,$BA
	.dc.b	$D5,$FF,$AF,$23,$57,$FE,$93,$78,$79,$2F,$8B,$D4,$7F,$2B,$C3,$9B
	.dc.b	$C5,$4E,$E3,$F9,$44,$62,$7F,$91,$FD,$3F,$F9,$BF,$97,$F9,$2C,$E6
	.dc.b	$BF,$99,$FC,$8B,$E2,$BF,$D0,$B5,$FC,$E7,$A0,$FF,$3B,$9D,$7C,$3E
	.dc.b	$2F,$FC,$FD,$1C,$7D,$3C,$5E,$63,$77,$CC,$E5,$87,$6B,$E2,$B0,$E3
	.dc.b	$3A,$C2,$D2,$F9,$BC,$3E,$9F,$31,$E9,$FA,$7B,$8A,$6C,$61,$B8,$78
	.dc.b	$EF,$A5,$6E,$31,$7C,$DD,$4B,$E9,$BF,$DB,$31,$C7,$67,$D3,$FF,$9C
	.dc.b	$A3,$73,$AA,$FF,$51,$56,$B9,$3D,$71,$CF,$E0,$3D,$B3,$9B,$E3,$18
	.dc.b	$C3,$A5,$7F,$A7,$30,$E9,$BF,$D1,$8E,$57,$EB,$F9,$07,$C1,$BE,$0B
	.dc.b	$BE,$1C,$2B,$E8,$F9,$FF,$1B,$3F,$8E,$D8,$F8,$3B,$FF,$9B,$DB,$3A
	.dc.b	$69,$AF,$67,$E7,$9F,$43,$E9,$EC,$64,$9A,$7D,$E8,$CB,$FE,$B6,$FA
	.dc.b	$4F,$F6,$AC,$FD,$3D,$9F,$4F,$F8,$57,$DD,$3A,$AF,$F6,$64,$6B,$9B
	.dc.b	$E5,$CE,$63,$F2,$35,$C6,$FF,$C2,$56,$AD,$47,$E7,$67,$53,$FE,$17
	.dc.b	$9C,$B7,$C8,$FC,$8F,$0F,$CC,$F5,$3B,$C7,$E3,$BE,$9E,$63,$F2,$3B
	.dc.b	$6E,$1F,$9C,$9B,$9F,$E8,$F8,$3F,$6F,$7F,$3A,$7F,$2E,$FE,$63,$E7
	.dc.b	$EB,$F4,$9B,$7F,$A6,$6D,$F3,$7E,$0F,$95,$F8,$AF,$A7,$FD,$55,$FE
	.dc.b	$63,$64,$E7,$BB,$73,$F9,$9A,$DE,$3F,$4E,$B5,$3C,$5D,$C7,$E6,$A7
	.dc.b	$E3,$BF,$2E,$3F,$62,$BE,$9E,$98,$F8,$BB,$7F,$AC,$9D,$36,$74,$D3
	.dc.b	$5F,$4F,$D3,$E3,$FA,$B6,$32,$73,$8C,$B4,$FD,$BF,$D3,$3E,$27,$EC
	.dc.b	$FD,$3B,$E4,$3F,$63,$39,$AF,$D4,$4E,$5E,$35,$C2,$B6,$D8,$C9,$32
	.dc.b	$70,$FF,$23,$FA,$77,$D3,$BE,$9F,$8D,$FF,$0A,$FA,$77,$2F,$F0,$2F
	.dc.b	$7F,$E5,$C6,$27,$F8,$7C,$7C,$EF,$1D,$ED,$FF,$08,$5C,$96,$30,$FF
	.dc.b	$0C,$FA,$67,$C9,$E7,$0F,$E3,$31,$87,$F8,$DD,$F4,$73,$F0,$5F,$C5
	.dc.b	$3E,$9D,$EA,$73,$8C,$3D,$7C,$BC,$57,$46,$F1,$FC,$28,$FD,$7E,$3F
	.dc.b	$C4,$5D,$1F,$CA,$BE,$9D,$F4,$FC,$E3,$F9,$6F,$A2,$EB,$FE,$7F,$0F
	.dc.b	$A5,$7C,$27,$F2,$8A,$D4,$FF,$39,$35,$19,$AE,$35,$C2,$7A,$C6,$27
	.dc.b	$51,$83,$94,$BE,$9F,$39,$FA,$CF,$F4,$D4,$BE,$23,$CB,$3F,$2E,$74
	.dc.b	$7F,$66,$8C,$F1,$7F,$3E,$AE,$07,$9E,$B8,$75,$FF,$39,$8C,$DF,$77
	.dc.b	$C5,$5C,$6E,$47,$9B,$E7,$ED,$78,$3E,$9D,$F0,$9F,$00,$FE,$9A,$E2
	.dc.b	$BD,$B1,$83,$4F,$F5,$C8,$D4,$F8,$73,$B9,$7F,$4D,$9B,$5F,$CA,$9B
	.dc.b	$5F,$E0,$4A,$7C,$27,$5A,$FE,$CF,$C9,$FC,$33,$0F,$F0,$34,$7D,$BF
	.dc.b	$A5,$FE,$C3,$39,$8F,$9B,$69,$CF,$93,$F6,$9B,$3F,$59,$B8,$FD,$0B
	.dc.b	$E4,$F8,$FA,$73,$8F,$F6,$5F,$43,$F0,$F3,$FD,$CB,$E9,$DF,$09,$FD
	.dc.b	$CA,$73,$5F,$E1,$BB,$56,$B2,$FF,$00,$D6,$9F,$F6,$E4,$EA,$F9,$72
	.dc.b	$FF,$7A,$BE,$18,$FE,$13,$0E,$0F,$FB,$8D,$B6,$EF,$FE,$E9,$39,$BF
	.dc.b	$27,$D0,$FF,$88,$33,$7D,$3F,$2A,$62,$AD,$4E,$63,$1A,$7F,$6B,$8C
	.dc.b	$4E,$B0,$F8,$1B,$7C,$1E,$1B,$7E,$E6,$E3,$E9,$9F,$F0,$BB,$FE,$FB
	.dc.b	$F8,$5F,$F7,$E7,$D3,$3E,$2F,$FB,$6E,$9F,$67,$1F,$F1,$3D,$EC,$56
	.dc.b	$DF,$85,$87,$F9,$26,$31,$1F,$DD,$B0,$EB,$87,$4A,$D3,$FE,$62,$8C
	.dc.b	$CF,$2B,$AF,$13,$BC,$F1,$8C,$C7,$8E,$FB,$E1,$E7,$8D,$9F,$E7,$3B
	.dc.b	$E7,$19,$9F,$2C,$F1,$BE,$8F,$03,$C1,$1D,$30,$F3,$C7,$33,$55,$CE
	.dc.b	$BD,$51,$89,$DB,$CA,$3A,$D7,$82,$17,$E4,$AD,$C7,$8C,$F8,$47,$8C
	.dc.b	$C1,$F8,$67,$6F,$23,$FD,$16,$7D,$52,$C6,$1C,$1F,$E2,$8F,$A6,$FE
	.dc.b	$41,$EA,$F0,$6C,$7F,$A5,$1A,$FF,$A6,$DA,$F1,$F9,$2F,$EC,$AE,$DA
	.dc.b	$ED,$AE,$F8,$3B,$47,$0C,$7F,$0B,$F4,$EF,$A3,$FF,$1D,$C6,$1B,$A7
	.dc.b	$1C,$3E,$A6,$2C,$39,$4F,$F9,$0B,$0E,$55,$F8,$86,$BC,$27,$68,$57
	.dc.b	$3C,$78,$61,$E0,$9D,$5F,$2C,$7F,$37,$7B,$9E,$97,$C1,$86,$09,$3E
	.dc.b	$B7,$2F,$2C,$3C,$71,$E2,$6D,$F7,$C7,$D9,$8F,$8E,$B5,$87,$E9,$31
	.dc.b	$F1,$63,$D6,$7F,$31,$87,$17,$DD,$3A,$C3,$9C,$7F,$A8,$27,$E2,$DF
	.dc.b	$FA,$3D,$BB,$7C,$FC,$3A,$4E,$71,$FF,$37,$B6,$FF,$D5,$38,$F5,$7E
	.dc.b	$67,$D3,$FF,$AF,$DF,$F5,$6B,$5F,$ED,$6D,$9F,$F4,$F7,$D2,$FF,$4E
	.dc.b	$9D,$F5,$76,$CB,$FC,$FB,$3F,$E5,$E9,$FF,$5B,$B5,$E7,$3B,$7C,$3F
	.dc.b	$31,$56,$08,$C3,$63,$9F,$7E,$26,$47,$A1,$E5,$69,$CC,$FF,$81,$23
	.dc.b	$FC,$32,$F9,$3C,$3C,$E5,$FE,$B9,$FA,$7F,$E7,$53,$FE,$CE,$6B,$FE
	.dc.b	$0C,$6C,$7C,$2F,$37,$D1,$18,$D3,$FD,$3A,$DA,$FC,$C3,$56,$1F,$E4
	.dc.b	$A7,$B7,$4F,$A3,$E9,$E4,$60,$FF,$6F,$FC,$E7,$F9,$9D,$BE,$B7,$C7
	.dc.b	$8C,$47,$B7,$B3,$13,$23,$67,$D2,$8C,$3F,$0F,$D0,$F8,$E6,$35,$FF
	.dc.b	$6D,$7C,$4F,$F9,$62,$7F,$DD,$ED,$7F,$C5,$AD,$8F,$F7,$07,$2F,$F9
	.dc.b	$AA,$77,$E4,$FD,$BA,$D6,$FF,$EC,$06,$CF,$EE,$EB,$83,$D2,$BF,$05
	.dc.b	$0F,$87,$A7,$38,$FC,$37,$A1,$F8,$3A,$FF,$1F,$B7,$1F,$F7,$93,$71
	.dc.b	$FF,$67,$FD,$37,$F9,$AA,$BF,$85,$56,$B9,$B9,$92,$33,$5C,$3C,$8C
	.dc.b	$47,$F9,$32,$DF,$4F,$F9,$76,$C7,$03,$F8,$B6,$3E,$17,$F8,$C4,$7C
	.dc.b	$05,$F4,$FF,$A9,$7F,$CA,$ED,$C9,$8F,$A3,$3F,$D2,$4E,$63,$FD,$D2
	.dc.b	$DA,$FE,$48,$D5,$7D,$FE,$9F,$FD,$61,$1E,$50,$E1,$FF,$33,$B7,$0F
	.dc.b	$0C,$74,$F1,$BF,$69,$3F,$E6,$F8,$E4,$D7,$8F,$9F,$FB,$9F,$A7,$7C
	.dc.b	$3F,$EE,$FE,$9D,$F1,$74,$97,$CA,$F5,$F4,$72,$3C,$D4,$8E,$5F,$4F
	.dc.b	$FC,$17,$E2,$DE,$E5,$25,$FB,$7E,$3F,$FC,$74,$FB,$E7,$6D,$9D,$BF
	.dc.b	$E8,$69,$DB,$FE,$88,$9D,$B9,$F9,$BF,$E2,$92,$78,$F0,$7C,$69,$7D
	.dc.b	$3F,$D1,$3F,$D0,$2F,$CF,$6B,$FE,$94,$75,$C3,$FC,$7D,$CC,$F3,$6F
	.dc.b	$27,$F5,$FA,$EB,$5B,$F1,$64,$75,$79,$7D,$06,$23,$B4,$5C,$3A,$4F
	.dc.b	$C9,$F9,$69,$5F,$15,$F1,$71,$E2,$3E,$14,$5A,$F2,$4D,$C7,$C9,$93
	.dc.b	$E9,$FD,$46,$70,$FE,$ED,$F4,$EF,$17,$F9,$E3,$F6,$E3,$F2,$4D,$7F
	.dc.b	$D6,$EE,$DC,$C7,$0E,$39,$18,$38,$FD,$3F,$42,$3D,$3C,$0A,$F4,$8B
	.dc.b	$A7,$D7,$C3,$F6,$92,$FA,$77,$CC,$7F,$51,$6C,$7E,$9D,$AF,$FB,$33
	.dc.b	$E9,$FF,$85,$3F,$EC,$E8,$CD,$FE,$0D,$BB,$7D,$96,$CF,$F8,$66,$B8
	.dc.b	$D6,$EF,$FB,$04,$3E,$5E,$9E,$38,$CF,$AF,$2F,$F7,$55,$FE,$2B,$6F
	.dc.b	$66,$5F,$34,$DC,$3F,$2D,$5F,$B7,$6F,$BB,$E0,$EB,$97,$D2,$3C,$66
	.dc.b	$1F,$81,$18,$7F,$85,$7D,$3B,$E9,$FF,$BC,$36,$33,$1F,$DE,$FD,$3F
	.dc.b	$E2,$93,$99,$FF,$7A,$B7,$0F,$D5,$37,$7F,$D0,$57,$0B,$E5,$F2,$D3
	.dc.b	$0E,$A7,$F1,$B9,$FE,$D4,$FF,$C0,$05,$D5,$E1,$3C,$5D,$72,$78,$46
	.dc.b	$5E,$51,$E7,$EA,$F4,$7D,$3C,$9E,$D1,$97,$C1,$19,$7C,$11,$FE,$19
	.dc.b	$FA,$6F,$F4,$34,$7F,$95,$DA,$9A,$4F,$77,$86,$52,$7F,$9B,$4F,$FA
	.dc.b	$CB,$6F,$E1,$CF,$F8,$A5,$B3,$B9,$46,$AB,$11,$FC,$24,$F0,$4F,$FA
	.dc.b	$2E,$BF,$C6,$51,$96,$BF,$C3,$DF,$E2,$EE,$7E,$23,$81,$1F,$F4,$BC
	.dc.b	$7F,$16,$7D,$DE,$07,$C1,$3E,$14,$D8,$F5,$C5,$7C,$FB,$F7,$9C,$63
	.dc.b	$EE,$AB,$39,$AE,$24,$38,$9E,$03,$E3,$D7,$07,$06,$8F,$A1,$3E,$47
	.dc.b	$95,$77,$6D,$77,$78,$5F,$F6,$B2,$57,$18,$CB,$DD,$EA,$78,$8E,$67
	.dc.b	$8A,$BB,$3D,$2F,$85,$74,$8E,$07,$43,$89,$D0,$E0,$61,$E0,$E0,$E0
	.dc.b	$49,$EE,$7E,$22,$7B,$D7,$81,$B3,$CD,$AF,$1B,$F8,$55,$C9,$FC,$25
	.dc.b	$71,$9F,$13,$5E,$07,$BE,$36,$FB,$27,$E0,$BF,$8F,$9F,$C9,$11,$FC
	.dc.b	$5B,$5F,$CC,$1F,$04,$E0,$68,$FF,$61,$7C,$03,$C1,$FE,$CD,$F4,$CF
	.dc.b	$87,$FE,$D1,$39,$AF,$F6,$44,$E5,$E3,$5E,$67,$FD,$F6,$D9,$E1,$09
	.dc.b	$E0,$E8,$E7,$5D,$AB,$81,$23,$27,$62,$9C,$4F,$B9,$5D,$B3,$FF,$21
	.dc.b	$3E,$0B,$EE,$DB,$ED,$5F,$9C,$8E,$18,$62,$3C,$D3,$6B,$8E,$1E,$39
	.dc.b	$EE,$FC,$69,$D1,$C7,$0F,$0C,$6E,$B8,$D7,$A3,$0E,$EF,$AC,$E6,$66
	.dc.b	$7A,$39,$AE,$6F,$9E,$32,$FC,$1A,$EA,$F2,$23,$83,$33,$CA,$33,$7E
	.dc.b	$87,$B6,$1E,$93,$0F,$58,$C3,$A3,$D1,$5D,$63,$31,$F4,$EB,$B4,$E6
	.dc.b	$38,$19,$AE,$B8,$7D,$38,$CB,$F1,$1F,$01,$DA,$FA,$C6,$4E,$11,$A3
	.dc.b	$A1,$CB,$0F,$8B,$1A,$38,$38,$24,$7C,$F3,$E6,$1F,$8C,$CB,$F9,$F6
	.dc.b	$1F,$96,$6D,$FF,$92,$5B,$23,$FC,$B0,$7E,$77,$E9,$B4,$FE,$A9,$AF
	.dc.b	$EA,$1A,$EC,$E5,$1E,$08,$FC,$97,$D2,$3E,$9B,$FD,$5D,$97,$EA,$A3
	.dc.b	$1E,$A7,$D2,$AD,$D6,$B3,$FF,$2F,$56,$B1,$FD,$33,$70,$FD,$25,$F0
	.dc.b	$C3,$F8,$2D,$2B,$83,$FD,$7D,$79,$4F,$89,$DC,$6A,$7E,$17,$C7,$3E
	.dc.b	$1F,$FD,$2B,$AF,$FA,$E6,$31,$F0,$9F,$EA,$2E,$3F,$F5,$1C,$6F,$D1
	.dc.b	$C5,$72,$B5,$E3,$05,$EE,$BA,$56,$9B,$7B,$9E,$F1,$AF,$A4,$C5,$3E
	.dc.b	$96,$C6,$1B,$3D,$1B,$87,$F6,$4B,$F6,$47,$8E,$1F,$49,$FE,$EB,$AF
	.dc.b	$F6,$B6,$AF,$A5,$FE,$55,$1F,$EC,$59,$D5,$7F,$6E,$AD,$71,$F3,$9C
	.dc.b	$ED,$FE,$C6,$9C,$DF,$1F,$77,$E1,$46,$30,$FE,$E5,$7C,$2B,$FD,$38
	.dc.b	$D8,$7C,$6B,$E0,$1E,$F7,$7C,$7E,$1F,$99,$F1,$1B,$8F,$CC,$6E,$1F
	.dc.b	$07,$87,$A7,$07,$DD,$BE,$17,$F4,$BD,$7F,$9A,$7C,$EF,$A5,$B1,$92
	.dc.b	$61,$CE,$32,$FF,$82,$FE,$8F,$FD,$BB,$3F,$EF,$FE,$07,$D2,$FE,$2A
	.dc.b	$B5,$E2,$FE,$0E,$73,$CF,$F4,$F0,$E3,$5F,$E1,$A9,$CE,$1F,$EF,$59
	.dc.b	$CC,$7E,$47,$0E,$8F,$FB,$3E,$33,$3F,$EC,$56,$D7,$CA,$E0,$FF,$45
	.dc.b	$4F,$3C,$7F,$17,$F4,$B2,$3F,$27,$D7,$6F,$E4,$4D,$C7,$F3,$FC,$1F
	.dc.b	$AA,$BF,$CC,$CF,$E5,$1F,$CA,$FC,$E5,$7F,$3A,$DB,$FA,$4D,$BE,$6F
	.dc.b	$C1,$F1,$BF,$0B,$F4,$BF,$A3,$AF,$99,$18,$ED,$D6,$1F,$96,$AD,$61
	.dc.b	$F0,$6B,$53,$F1,$DD,$DF,$E5,$EB,$E3,$CF,$C5,$9F,$D7,$4B,$E9,$69
	.dc.b	$8F,$88,$B5,$FC,$06,$74,$D9,$D3,$4D,$7D,$2F,$4F,$8A,$EA,$D8,$C9
	.dc.b	$CE,$32,$D3,$F9,$96,$47,$FD,$FD,$F4,$AF,$A4,$FE,$15,$DA,$FA,$57
	.dc.b	$C7,$7E,$EE,$B5,$7C,$8F,$E7,$95,$B9,$D5,$F2,$BE,$35,$BD,$3F,$A1
	.dc.b	$61,$C9,$FF,$35,$E1,$C6,$75,$09,$D6,$3E,$78,$E3,$87,$F1,$FF,$A5
	.dc.b	$7D,$2B,$E9,$7C,$38,$F2,$FA,$57,$2F,$F1,$3F,$73,$EA,$E5,$FD,$4E
	.dc.b	$73,$5E,$09,$CC,$FF,$39,$F5,$FF,$8D,$78,$EE,$5B,$18,$7F,$8F,$7D
	.dc.b	$23,$E4,$73,$8F,$AA,$30,$FB,$2D,$7C,$3D,$EF,$E4,$DF,$4A,$F4,$71
	.dc.b	$7F,$0A,$30,$F3,$F2,$F3,$A7,$FA,$37,$2F,$E2,$F3,$DF,$0F,$1D,$48
	.dc.b	$F9,$7F,$4A,$FA,$5C,$61,$FE,$95,$F3,$1D,$9F,$D2,$6D,$F4,$6F,$80
	.dc.b	$FE,$75,$5A,$9F,$B4,$6A,$33,$5C,$6B,$84,$FD,$AA,$E1,$7F,$EA,$88
	.dc.b	$E0,$66,$B8,$BD,$29,$F4,$B8,$AF,$F5,$54,$BE,$1B,$CD,$3F,$32,$73
	.dc.b	$18,$E3,$FE,$A1,$5B,$3C,$F5,$BE,$CF,$C5,$5F,$E8,$6F,$85,$7E,$40
	.dc.b	$E1,$1D,$1F,$27,$CE,$F4,$9F,$EB,$9F,$4A,$F8,$1F,$80,$60,$B5,$F9
	.dc.b	$A8,$D3,$ED,$8D,$34,$D3,$8B,$F8,$D9,$BE,$03,$19,$7A,$39,$BF,$89
	.dc.b	$E1,$FE,$22,$A7,$D1,$FF,$6F,$9C,$E1,$FB,$D9,$CD,$7E,$B2,$72,$FD
	.dc.b	$EA,$E1,$E5,$7D,$96,$C6,$E3,$EE,$75,$63,$0F,$F6,$E7,$89,$F7,$32
	.dc.b	$FE,$E5,$F4,$BF,$EA,$1C,$3F,$73,$39,$7F,$C8,$16,$9D,$5F,$2E,$3F
	.dc.b	$EE,$F1,$8C,$3E,$AE,$1C,$6B,$FC,$55,$39,$C3,$FC,$95,$1F,$DE,$2F
	.dc.b	$FC,$1D,$5B,$C1,$DD,$FD,$F6,$B5,$7E,$4F,$9E,$FE,$C9,$F4,$BD,$67
	.dc.b	$FC,$03,$18,$C7,$9B,$6B,$81,$FD,$BE,$B8,$63,$F3,$2F,$E1,$F0,$78
	.dc.b	$B6,$F1,$B6,$FE,$1C,$7F,$88,$E7,$FD,$92,$3E,$91,$F0,$FE,$CC,$3A
	.dc.b	$6F,$FE,$66,$A7,$C1,$FF,$9A,$6F,$71,$FC,$C3,$0F,$E4,$91,$88,$FF
	.dc.b	$68,$9D,$70,$E9,$3A,$7F,$DB,$B5,$CD,$CE,$36,$BE,$69,$E0,$39,$1D
	.dc.b	$DD,$3D,$AF,$C1,$8F,$C3,$7C,$51,$EE,$73,$87,$86,$3A,$E1,$CA,$B8
	.dc.b	$61,$E4,$9E,$E7,$F8,$D2,$B8,$3D,$E3,$CE,$E8,$EC,$E2,$FC,$31,$E9
	.dc.b	$7C,$38,$7A,$23,$67,$AC,$E4,$6E,$79,$1E,$39,$E3,$1A,$C3,$FD,$02
	.dc.b	$FD,$A3,$FD,$06,$E1,$F9,$07,$CA,$FA,$4F,$E3,$1E,$9F,$FA,$21,$B1
	.dc.b	$F3,$83,$87,$87,$47,$FA,$15,$D4,$3E,$39,$AF,$36,$CE,$9A,$F8,$5A
	.dc.b	$FE,$A2,$79,$4F,$DA,$FA,$57,$D0,$FE,$B3,$3F,$C5,$46,$2F,$EF,$DF
	.dc.b	$07,$EA,$B6,$77,$5C,$74,$FD,$AD,$F1,$BD,$E3,$E6,$BF,$CB,$61,$C6
	.dc.b	$F7,$87,$1B,$FF,$63,$61,$C6,$F7,$87,$1B,$F5,$36,$B7,$7C,$55,$F1
	.dc.b	$AD,$E3,$EA,$BE,$15,$AB,$E1,$7B,$8F,$F6,$B6,$3E,$56,$D7,$09,$D4
	.dc.b	$71,$3A,$5C,$7F,$33,$1A,$37,$5C,$E3,$81,$F6,$B5,$F8,$2F,$F8,$82
	.dc.b	$7E,$45,$7F,$56,$AE,$11,$8A,$DD,$71,$3E,$09,$FE,$28,$9E,$DF,$4B
	.dc.b	$FE,$C4,$8F,$F1,$5B,$5E,$01,$C5,$FF,$64,$FA,$4F,$EA,$13,$C3,$A7
	.dc.b	$F9,$7C,$E7,$D6,$B8,$FF,$32,$B4,$62,$FF,$DF,$D8,$74,$AF,$E6,$EB
	.dc.b	$E5,$F5,$3E,$16,$C6,$63,$FC,$83,$6F,$03,$35,$27,$D6,$F2,$9D,$46
	.dc.b	$0E,$3E,$26,$4E,$0F,$C0,$FA,$57,$57,$89,$B1,$FE,$50,$6B,$F1,$AB
	.dc.b	$FC,$6E,$E6,$78,$11,$F3,$F9,$95,$E1,$F4,$BE,$53,$1B,$7F,$84,$3E
	.dc.b	$6F,$FC,$D7,$3F,$E5,$E6,$BF,$51,$B1,$F4,$FB,$BE,$3C,$63,$7F,$EC
	.dc.b	$2D,$A0,$70,$B2,$FF,$1E,$BD,$FD,$8F,$91,$E8,$48,$C1,$FE,$70,$F7
	.dc.b	$FF,$CB,$4D,$F2,$BE,$BC,$62,$3F,$13,$D3,$FE,$6C,$74,$43,$F1,$B1
	.dc.b	$BD,$3F,$C5,$F3,$B9,$CD,$7E,$0A,$1F,$0D,$4E,$F1,$F8,$6F,$3F,$F0
	.dc.b	$95,$F9,$86,$DF,$F8,$E9,$BA,$FF,$81,$3E,$8F,$FC,$D5,$5F,$E8,$A9
	.dc.b	$CB,$F1,$9A,$FF,$BF,$B9,$7F,$27,$3C,$23,$B4,$63,$E9,$64,$73,$8D
	.dc.b	$E3,$F9,$AF,$A5,$FD,$3C,$FE,$B1,$B9,$7F,$93,$B0,$F8,$59,$7F,$86
	.dc.b	$27,$2D,$F1,$B9,$1D,$2F,$FD,$43,$F4,$BF,$EB,$36,$FB,$9F,$A8,$8C
	.dc.b	$47,$79,$C7,$89,$23,$8B,$FA,$EF,$A5,$7C,$2F,$EC,$3E,$95,$F1,$34
	.dc.b	$97,$C0,$FF,$58,$C3,$E8,$78,$12,$33,$F4,$BF,$BA,$F8,$DF,$EA,$91
	.dc.b	$FE,$B8,$7F,$CE,$CD,$CB,$5E,$4C,$8C,$10,$FE,$27,$3C,$9B,$1D,$66
	.dc.b	$47,$58,$FE,$13,$F1,$4F,$76,$92,$7C,$5E,$9D,$7A,$FD,$2F,$9C,$C7
	.dc.b	$37,$FA,$0B,$E9,$7E,$9C,$FE,$A9,$FF,$3E,$3F,$ED,$26,$BF,$CA,$1D
	.dc.b	$A6,$9A,$8E,$9E,$D7,$F1,$E8,$CF,$8D,$49,$5D,$7E,$76,$BD,$62,$DF
	.dc.b	$96,$6F,$47,$F9,$82,$FF,$89,$DE,$CE,$47,$FB,$A3,$47,$C9,$FC,$62
	.dc.b	$47,$AF,$96,$BE,$BF,$A5,$F3,$18,$CB,$FD,$4B,$F4,$9F,$C2,$E7,$FD
	.dc.b	$E6,$FE,$0D,$BC,$BF,$CE,$A3,$51,$FD,$31,$B3,$CA,$74,$6F,$C8,$E1
	.dc.b	$F4,$B2,$FA,$BC,$FC,$1C,$47,$38,$B7,$E4,$9B,$B7,$FA,$0B,$15,$62
	.dc.b	$FF,$91,$C7,$EF,$BE,$95,$F2,$DF,$D4,$1B,$3F,$CA,$31,$FF,$61,$C6
	.dc.b	$23,$FB,$2F,$D2,$7F,$03,$9C,$D7,$F5,$88,$7D,$2D,$3A,$46,$7A,$65
	.dc.b	$F5,$CF,$FD,$2F,$7F,$9D,$C3,$FD,$A2,$FE,$9A,$7B,$D7,$06,$9C,$4F
	.dc.b	$E7,$B1,$B9,$E3,$D1,$E3,$E9,$95,$B1,$5F,$EA,$27,$0D,$7D,$B1,$C1
	.dc.b	$F7,$CF,$B4,$C5,$79,$A7,$13,$D7,$E9,$5F,$4B,$FE,$62,$6C,$7F,$96
	.dc.b	$C3,$87,$FC,$BC,$D7,$FD,$1A,$E6,$BF,$DE,$7E,$A7,$F8,$2A,$7D,$4D
	.dc.b	$7F,$C2,$F3,$BC,$FF,$D8,$33,$FA,$2A,$E3,$D3,$FE,$0C,$83,$7A,$A6
	.dc.b	$1D,$23,$46,$1F,$F1,$1C,$FF,$26,$7C,$3F,$4B,$CC,$E9,$D1,$23,$24
	.dc.b	$38,$B9,$79,$C7,$17,$D7,$19,$C3,$DF,$39,$9E,$26,$8E,$91,$A7,$06
	.dc.b	$8E,$A6,$B1,$F8,$75,$D5,$F8,$B5,$D6,$BE,$6C,$E6,$BF,$C0,$F6,$AD
	.dc.b	$5F,$29,$EF,$7E,$17,$B9,$E1,$79,$BE,$13,$9C,$F5,$55,$F2,$DE,$33
	.dc.b	$CC,$E2,$F3,$AF,$3B,$CC,$E3,$7C,$EB,$CD,$19,$7A,$57,$23,$A5,$72
	.dc.b	$9E,$87,$BA,$BD,$AF,$98,$F6,$C7,$BD,$F2,$BC,$67,$A3,$97,$AC,$71
	.dc.b	$3A,$9E,$63,$C9,$7E,$7A,$EF,$1F,$0A,$30,$E8,$F0,$9E,$53,$07,$4B
	.dc.b	$F6,$DF,$8C,$91,$B3,$53,$DA,$77,$19,$BF,$3B,$E1,$3B,$9E,$73,$DE
	.dc.b	$7A,$DF,$7B,$C6,$BB,$BC,$4F,$25,$F7,$BF,$5C,$72,$35,$5C,$9E,$CF
	.dc.b	$8A,$FE,$CC,$72,$3D,$F1,$B8,$F4,$C7,$2B,$F0,$C6,$09,$5E,$FB,$FC
	.dc.b	$23,$6B,$DB,$3D,$AF,$C1,$5A,$9F,$C4,$35,$F4,$4F,$58,$F6,$47,$26
	.dc.b	$BE,$73,$75,$F0,$2B,$D8,$DB,$F3,$B6,$7E,$64,$7A,$AB,$C1,$5E,$49
	.dc.b	$F3,$D7,$A5,$FB,$67,$98,$C9,$4C,$9E,$89,$E7,$5D,$63,$66,$4B,$5D
	.dc.b	$DF,$F5,$FF,$D0,$3E,$8B,$FC,$BD,$1F,$E8,$81,$19,$F8,$7F,$E7,$4B
	.dc.b	$B5,$F0,$D8,$BE,$37,$FE,$90,$E0,$FD,$FE,$5F,$D0,$5B,$1D,$4D,$6A
	.dc.b	$9D,$36,$35,$F1,$6F,$83,$FF,$2A,$D7,$F6,$18,$C7,$BF,$BD,$B8,$46
	.dc.b	$FC,$9C,$57,$FA,$8A,$36,$F1,$BE,$78,$7F,$B9,$E7,$2D,$8C,$DF,$45
	.dc.b	$87,$49,$57,$97,$E8,$70,$F9,$FE,$86,$DA,$6C,$61,$B8,$F8,$DB,$7E
	.dc.b	$16,$E3,$E1,$6C,$FD,$5F,$A0,$FF,$52,$BF,$E0,$3D,$DF,$43,$FE,$3E
	.dc.b	$8F,$F0,$04,$EB,$06,$1F,$36,$F7,$E3,$7A,$E5,$5B,$AF,$A5,$18,$F5
	.dc.b	$FE,$07,$C5,$3D,$A7,$B3,$C7,$0E,$19,$FC,$FD,$FF,$C5,$AF,$CF,$C3
	.dc.b	$E1,$F8,$5F,$11,$B1,$93,$EA,$46,$5F,$AD,$E7,$FC,$A3,$E5,$FD,$0D
	.dc.b	$8C,$93,$4F,$B1,$19,$7E,$D4,$DD,$7E,$D4,$DA,$FC,$F7,$D0,$7F,$A8
	.dc.b	$9F,$F1,$64,$BE,$87,$F0,$31,$AE,$EF,$C0,$4E,$7A,$3E,$45,$F1,$CB
	.dc.b	$E6,$DE,$CF,$F1,$B7,$0F,$C7,$F1,$F7,$8F,$E1,$3E,$86,$57,$E4,$FB
	.dc.b	$AE,$1F,$91,$9B,$9F,$E4,$66,$D7,$CD,$3F,$1B,$F3,$3F,$8E,$69,$F9
	.dc.b	$D2,$DF,$E6,$DB,$5F,$9B,$69,$CC,$FD,$24,$FB,$3E,$27,$F3,$DD,$0F
	.dc.b	$A1,$FD,$0D,$7F,$92,$F6,$56,$A7,$38,$FC,$20,$27,$75,$C6,$B8,$6F
	.dc.b	$DE,$3F,$79,$4F,$A1,$FD,$5B,$63,$3F,$07,$89,$91,$C1,$DC,$62,$7F
	.dc.b	$78,$3E,$87,$9F,$C3,$F4,$6D,$F8,$1B,$3C,$1B,$B7,$F0,$C8,$E5,$3E
	.dc.b	$6A,$E4,$D3,$FC,$DD,$F3,$8D,$FF,$59,$E7,$7E,$2A,$FE,$1C,$E6,$BF
	.dc.b	$CE,$93,$99,$FE,$57,$E6,$7E,$CF,$85,$A5,$8C,$3F,$B7,$F9,$B7,$C2
	.dc.b	$E7,$1F,$F3,$F4,$61,$FF,$40,$5A,$F8,$7A,$1F,$BF,$F9,$C7,$9D,$8B
	.dc.b	$FF,$43,$46,$1E,$7E,$1E,$74,$FB,$83,$2F,$E0,$33,$E0,$C3,$C9,$52
	.dc.b	$3F,$D2,$1F,$38,$F9,$CC,$61,$CB,$CB,$71,$FF,$2B,$B7,$CD,$7F,$46
	.dc.b	$E5,$FE,$23,$5A,$9F,$F5,$01,$AF,$3B,$F8,$C4,$67,$0F,$C8,$4E,$4F
	.dc.b	$F5,$04,$66,$BF,$A6,$46,$67,$C5,$87,$23,$C5,$83,$E7,$31,$5F,$EA
	.dc.b	$91,$F0,$5E,$78,$F9,$D7,$F1,$27,$27,$9A,$33,$87,$F2,$A8,$CC,$F9
	.dc.b	$A7,$31,$FD,$BC,$CC,$F9,$A3,$33,$FA,$A9,$C9,$BA,$F3,$FC,$97,$F3
	.dc.b	$98,$CC,$FE,$C3,$E7,$1F,$07,$EC,$3F,$9F,$38,$9F,$97,$95,$AF,$BB
	.dc.b	$37,$2F,$8F,$36,$BF,$C0,$93,$6B,$FB,$C4,$D5,$14,$D4,$FD,$CF,$88
	.dc.b	$F0,$E1,$FD,$3F,$97,$F8,$3E,$3D,$BE,$6B,$1E,$AF,$D2,$6E,$1E,$66
	.dc.b	$E5,$F7,$B3,$46,$2F,$FA,$BE,$1D,$2F,$E4,$57,$F8,$59,$FF,$71,$46
	.dc.b	$6F,$F8,$AF,$93,$F8,$59,$FE,$2F,$D4,$F9,$C7,$D1,$FF,$65,$AD,$5F
	.dc.b	$F8,$82,$73,$7F,$EE,$D9,$CD,$F1,$BE,$1C,$7F,$CA,$AF,$78,$7E,$AA
	.dc.b	$F9,$A7,$8F,$7F,$F6,$A8,$F2,$D7,$87,$E5,$7F,$A8,$FC,$EF,$0B,$46
	.dc.b	$27,$59,$FF,$5A,$AD,$4E,$63,$CD,$BF,$AF,$8B,$D8,$D8,$C9,$31,$FD
	.dc.b	$F9,$FC,$EB,$E7,$7F,$9D,$FC,$EB,$E7,$7F,$C8,$DA,$7D,$2B,$FE,$7F
	.dc.b	$C7,$FE,$08,$C3,$80,$BE,$0F,$D0,$C7,$FC,$7D,$18,$8F,$F2,$86,$3D
	.dc.b	$B1,$EB,$5A,$7F,$C9,$12,$9D,$3C,$F0,$AF,$23,$C1,$3C,$A3,$31,$E4
	.dc.b	$BF,$06,$3E,$98,$D9,$FE,$55,$BE,$71,$99,$F3,$4F,$1B,$E8,$F0,$3C
	.dc.b	$11,$D3,$0F,$44,$73,$35,$5C,$EB,$D7,$7D,$AB,$75,$CA,$7A,$DF,$82
	.dc.b	$3E,$01,$CA,$76,$F8,$CF,$84,$78,$CC,$1F,$8D,$76,$F2,$3E,$59,$4F
	.dc.b	$19,$83,$07,$F7,$AF,$9C,$FE,$43,$EB,$F0,$6C,$7D,$46,$BF,$E7,$D6
	.dc.b	$BF,$E8,$B5,$F0,$BF,$E5,$28,$E0,$D7,$83,$5E,$0D,$7C,$4F,$1B,$FC
	.dc.b	$47,$CE,$BE,$63,$FD,$31,$18,$6E,$5C,$B6,$FC,$2E,$3E,$27,$F0,$F7
	.dc.b	$C7,$1B,$7C,$1E,$F3,$C3,$1F,$06,$1F,$EA,$5B,$E1,$5F,$81,$6E,$3F
	.dc.b	$9B,$C7,$B4,$FE,$15,$B1,$E5,$8E,$E7,$5C,$23,$CF,$1F,$12,$BE,$B8
	.dc.b	$7A,$E3,$A6,$3E,$47,$FC,$5F,$3A,$6D,$72,$C7,$C3,$7C,$71,$F1,$57
	.dc.b	$73,$C2,$E3,$1F,$F4,$0E,$1E,$E6,$E1,$FE,$C5,$7F,$D0,$B8,$74,$BE
	.dc.b	$E7,$F0,$2D,$BE,$35,$6D,$B1,$93,$E6,$9F,$A9,$F9,$DF,$F5,$6B,$90
	.dc.b	$E1,$FF,$08,$B6,$FF,$D2,$BF,$36,$E1,$ED,$5F,$E5,$0A,$D5,$FF,$B6
	.dc.b	$EB,$FC,$AF,$3D,$5A,$F4,$EB,$E2,$F8,$3E,$36,$D9,$A6,$A9,$DF,$8A
	.dc.b	$F4,$56,$DB,$5B,$9D,$47,$F8,$AA,$9E,$1F,$33,$39,$FF,$AC,$3E,$77
	.dc.b	$F9,$ED,$7C,$36,$BF,$E3,$26,$C7,$C2,$F2,$FD,$31,$8D,$3F,$DF,$4D
	.dc.b	$AF,$F3,$9B,$56,$1F,$E4,$57,$B7,$47,$B7,$E7,$64,$60,$FA,$9F,$20
	.dc.b	$F9,$DF,$8E,$D8,$CC,$7C,$5E,$DC,$B2,$32,$7D,$18,$C3,$F0,$FD,$0F
	.dc.b	$98,$63,$5F,$F6,$8F,$C1,$FF,$97,$A7,$F0,$8D,$7F,$CA,$6D,$8F,$F6
	.dc.b	$BF,$2F,$F9,$C6,$77,$E8,$FD,$CA,$D6,$FF,$E9,$49,$FF,$6D,$4F,$E0
	.dc.b	$2B,$84,$74,$8C,$7C,$14,$8E,$67,$E5,$7C,$FF,$BC,$FF,$99,$1B,$8F
	.dc.b	$FB,$A1,$B8,$77,$F9,$CF,$F3,$7D,$71,$E6,$FC,$7D,$6A,$3F,$D4,$ED
	.dc.b	$53,$C1,$ED,$3F,$96,$87,$CE,$D3,$9C,$67,$2F,$CD,$FC,$EF,$E9,$6F
	.dc.b	$F5,$8D,$7F,$CF,$77,$F1,$1F,$F5,$BC,$67,$2F,$F8,$02,$DC,$BE,$43
	.dc.b	$67,$FD,$7B,$5C,$57,$CE,$3E,$77,$F5,$6D,$8F,$DE,$35,$FF,$45,$D7
	.dc.b	$37,$8F,$A1,$C0,$8F,$F0,$3F,$9D,$7C,$1F,$ED,$BE,$75,$EF,$D2,$57
	.dc.b	$6B,$D7,$CC,$C8,$FE,$19,$48,$E5,$F3,$BF,$C0,$3E,$2B,$FA,$AD,$7F
	.dc.b	$11,$7F,$CF,$ED,$BD,$79,$52,$30,$43,$FD,$4B,$3C,$9B,$1D,$66,$47
	.dc.b	$58,$FE,$1F,$F0,$CF,$76,$92,$7C,$5E,$95,$7A,$FC,$EF,$9C,$C7,$2B
	.dc.b	$97,$F3,$9F,$3B,$F2,$CF,$F4,$0D,$7F,$2B,$6B,$FE,$BA,$75,$E5,$FF
	.dc.b	$20,$73,$5C,$E7,$FD,$F3,$18,$AE,$B5,$BF,$1E,$57,$57,$97,$CB,$E2
	.dc.b	$3B,$45,$C3,$A4,$30,$56,$70,$F4,$DD,$AF,$0C,$FC,$6F,$90,$91,$FD
	.dc.b	$17,$54,$7A,$79,$73,$3E,$DF,$9D,$F6,$18,$CF,$C3,$F3,$8F,$1F,$F9
	.dc.b	$FB,$F8,$3A,$FE,$A0,$D7,$FD,$BE,$ED,$CC,$70,$E6,$91,$83,$8F,$CE
	.dc.b	$F4,$23,$D3,$C0,$AF,$48,$B8,$79,$26,$EF,$FC,$57,$15,$62,$C2,$3F
	.dc.b	$83,$7C,$EB,$E4,$FF,$AC,$B6,$7F,$B5,$35,$FF,$7C,$FC,$EF,$F0,$BC
	.dc.b	$BE,$9C,$E5,$FA,$B0,$F9,$2F,$EC,$6D,$AE,$77,$8E,$99,$1F,$DD,$E7
	.dc.b	$F4,$17,$CB,$AB,$33,$E0,$AB,$39,$8E,$1D,$3D,$BA,$A4,$60,$87,$A1
	.dc.b	$CB,$EA,$8F,$29,$89,$F4,$CE,$27,$FC,$15,$F3,$AF,$9A,$FE,$FA,$D8
	.dc.b	$CC,$7F,$80,$3D,$3F,$E2,$53,$99,$5F,$E6,$27,$FC,$0E,$DF,$91,$E7
	.dc.b	$7C,$7A,$F9,$75,$CC,$3F,$C3,$58,$7F,$6D,$79,$F2,$7F,$51,$8C,$63
	.dc.b	$D6,$B8,$56,$8F,$BF,$3A,$AE,$1D,$3E,$7F,$9D,$95,$DA,$3D,$46,$1E
	.dc.b	$47,$F8,$3A,$32,$EC,$F2,$BB,$C3,$DD,$3A,$9E,$36,$AD,$5F,$8A,$79
	.dc.b	$1A,$8C,$BE,$8A,$E5,$7F,$26,$73,$3F,$E2,$09,$CD,$7C,$39,$CD,$FF
	.dc.b	$3E,$AE,$B7,$C5,$F3,$98,$AF,$13,$E1,$32,$7A,$2F,$AB,$F2,$AF,$AC
	.dc.b	$62,$3E,$21,$C2,$BE,$0B,$D4,$E0,$66,$BB,$C6,$A3,$C0,$47,$04,$8F
	.dc.b	$64,$71,$3E,$21,$D4,$F0,$9C,$5E,$A4,$7C,$26,$4E,$27,$B4,$E8,$72
	.dc.b	$8F,$03,$A7,$E4,$3C,$4F,$A0,$78,$6F,$EE,$3C,$9F,$A0,$F5,$38,$3A
	.dc.b	$71,$1D,$9F,$04,$FB,$9C,$BE,$97,$27,$AA,$38,$4F,$67,$CA,$49,$E1
	.dc.b	$19,$7A,$46,$67,$D9,$1A,$BF,$B7,$1E,$B6,$BE,$23,$C1,$1C,$8E,$B7
	.dc.b	$F8,$E9,$FC,$15,$F7,$AD,$90,$F8,$E7,$33,$D6,$47,$83,$AB,$F3,$39
	.dc.b	$AE,$B5,$E4,$9F,$82,$DB,$F5,$57,$D1,$AD,$B4,$F6,$14,$F1,$5F,$D3
	.dc.b	$BE,$EF,$88,$97,$D6,$7E,$39,$EB,$8E,$17,$F0,$A3,$F0,$13,$F0,$AF
	.dc.b	$E2,$47,$E1,$5A,$F8,$1B,$7F,$69,$B5,$CE,$7F,$0F,$5C,$DE,$A7,$32
	.dc.b	$9C,$CB,$5C,$DB,$3C,$E3,$DB,$3F,$B7,$FA,$F7,$CE,$FF,$9E,$6E,$46
	.dc.b	$1C,$1F,$D9,$9C,$38,$6B,$C7,$E7,$B9,$FC,$47,$08,$46,$1E,$6B,$C0
	.dc.b	$7C,$46,$E3,$53,$FE,$8B,$F8,$D7,$C1,$FF,$95,$35,$FF,$6F,$46,$3D
	.dc.b	$FF,$F1,$66,$DF,$E2,$D8,$DF,$97,$8A,$FE,$FB,$1B,$78,$DF,$3C,$3F
	.dc.b	$A2,$4E,$5B,$19,$BE,$8B,$0E,$92,$AF,$2F,$CE,$E1,$F3,$FC,$F5,$B4
	.dc.b	$D8,$C3,$71,$F1,$B6,$FC,$2D,$C7,$C2,$D9,$FF,$57,$7C,$EB,$D7,$F1
	.dc.b	$EC,$F9,$EF,$F1,$CD,$7F,$BF,$EB,$4F,$F8,$7B,$0F,$F2,$95,$6F,$CA
	.dc.b	$F5,$E4,$BD,$BF,$D7,$9B,$5F,$E5,$9F,$8A,$7B,$6F,$6B,$8D,$6B,$2F
	.dc.b	$9F,$BF,$F8,$A5,$F9,$FA,$FC,$3F,$17,$E2,$36,$32,$7D,$48,$CB,$F5
	.dc.b	$BD,$1F,$94,$7C,$BF,$9E,$B1,$92,$6B,$F6,$23,$2F,$DA,$9B,$AF,$DA
	.dc.b	$9B,$5F,$ED,$7F,$9B,$FF,$50,$BF,$14,$F8,$DF,$3D,$F8,$17,$5D,$DF
	.dc.b	$80,$B4,$E7,$1F,$F2,$4D,$F2,$7F,$C0,$AD,$9F,$91,$BF,$F9,$07,$0E
	.dc.b	$D5,$C3,$0F,$C3,$7C,$F4,$9F,$CA,$F5,$DA,$FC,$9C,$DC,$3F,$27,$36
	.dc.b	$BE,$79,$F8,$FF,$9B,$FC,$83,$4F,$CF,$96,$FF,$3A,$DA,$FC,$EB,$4E
	.dc.b	$67,$E9,$A7,$F0,$B9,$3C,$5F,$C7,$FC,$F7,$E8,$EB,$D7,$82,$B5,$1E
	.dc.b	$EE,$75,$6E,$BE,$2D,$F4,$39,$F6,$DB,$FD,$EC,$BE,$7B,$F5,$8D,$8C
	.dc.b	$FC,$1E,$26,$47,$07,$71,$89,$FD,$E8,$F9,$EE,$7F,$0B,$D1,$B7,$E0
	.dc.b	$6C,$F0,$6E,$DF,$C3,$63,$94,$F9,$AB,$93,$4F,$87,$F3,$CD,$FF,$5B
	.dc.b	$E7,$7E,$1E,$D3,$9A,$F9,$33,$9D,$DE,$5F,$ED,$16,$96,$30,$FE,$E3
	.dc.b	$E7,$5F,$0B,$9C,$3E,$9C,$61,$FA,$B7,$D1,$CF,$A1,$FC,$03,$E7,$9E
	.dc.b	$57,$38,$C3,$D7,$C2,$C5,$74,$6E,$DF,$E6,$2C,$7F,$81,$4E,$4F,$C1
	.dc.b	$FC,$F3,$E7,$B9,$47,$E1,$BC,$A7,$1F,$F2,$A8,$7C,$DB,$D1,$FE,$27
	.dc.b	$5A,$9F,$D3,$9A,$8C,$D7,$1A,$E1,$3D,$63,$13,$A8,$C1,$CA,$5F,$3D
	.dc.b	$9C,$FF,$CF,$72,$F5,$FC,$93,$F2,$67,$31,$8E,$2F,$E4,$F5,$B3,$CB
	.dc.b	$5B,$E8,$FF,$79,$C6,$6F,$FD,$20,$F8,$2B,$8D,$C8,$F2,$7C,$95,$AF
	.dc.b	$F4,$6F,$CF,$3E,$7B,$DC,$7F,$36,$71,$5E,$B8,$C1,$A7,$F6,$71,$A9
	.dc.b	$F0,$E7,$72,$57,$FB,$82,$55,$98,$C3,$63,$74,$F8,$5E,$F5,$FC,$03
	.dc.b	$93,$F7,$98,$74,$BD,$FC,$D7,$F4,$C9,$CD,$7F,$BB,$8F,$BF,$19,$F5
	.dc.b	$3E,$D3,$63,$EB,$36,$BF,$D5,$F5,$C9,$F1,$F4,$E7,$0F,$04,$3D,$0F
	.dc.b	$C4,$3F,$EB,$F9,$7C,$EB,$E7,$BF,$AF,$D6,$AF,$FB,$1C,$E7,$2F,$E3
	.dc.b	$73,$95,$87,$F8,$5A,$74,$FF,$B0,$A7,$4E,$6F,$AF,$2C,$BF,$E4,$5D
	.dc.b	$FF,$D9,$6B,$9D,$F0,$7D,$1F,$2B,$FD,$37,$E7,$B1,$28,$C6,$1E,$86
	.dc.b	$E1,$E6,$6D,$F8,$F6,$F6,$F0,$7B,$9B,$5E,$C7,$0F,$F3,$89,$F6,$FC
	.dc.b	$F7,$BB,$E7,$5F,$3D,$F8,$5C,$D8,$7F,$81,$AF,$62,$B6,$FB,$23,$F0
	.dc.b	$51,$88,$FC,$75,$F5,$BE,$95,$A7,$FC,$9B,$19,$9E,$58,$57,$89,$DE
	.dc.b	$78,$C6,$63,$C7,$7D,$F0,$F3,$CE,$65,$7C,$E7,$31,$CC,$FF,$30,$46
	.dc.b	$70,$F2,$46,$0D,$1F,$8D,$72,$F9,$2F,$C1,$18,$8F,$1E,$1E,$67,$89
	.dc.b	$83,$DA,$60,$F2,$9D,$AF,$CB,$5F,$26,$38,$1E,$1C,$3D,$98,$79,$DB
	.dc.b	$1F,$E6,$F3,$99,$FE,$71,$8C,$38,$3F,$80,$7C,$EF,$F1,$FF,$5F,$FC
	.dc.b	$EC,$D8,$FF,$3F,$B5,$FF,$42,$35,$E3,$F0,$8E,$3E,$2D,$8E,$2D,$78
	.dc.b	$B5,$E3,$1F,$77,$E7,$9F,$33,$FC,$4F,$3F,$C4,$46,$2B,$F0,$58,$30
	.dc.b	$ED,$5F,$8D,$8C,$61,$CF,$0E,$5A,$FF,$1D,$C7,$96,$3F,$6A,$F6,$DC
	.dc.b	$79,$D6,$EF,$8C,$FF,$00,$BE,$38,$F7,$BD,$CF,$F0,$4C,$39,$CF,$A2
	.dc.b	$F8,$CA,$F8,$E1,$F0,$23,$C1,$8F,$9D,$86,$09,$7C,$DC,$BE,$18,$F8
	.dc.b	$78,$F8,$A7,$B6,$1E,$0C,$3B,$C7,$96,$75,$1F,$0A,$BF,$93,$B7,$4F
	.dc.b	$81,$5F,$E3,$AC,$BD,$79,$62,$3E,$F4,$DB,$E3,$8F,$F4,$97,$F9,$8D
	.dc.b	$F0,$AF,$E4,$FF,$3D,$FE,$A7,$7F,$C8,$AD,$79,$9A,$E9,$E5,$87,$F5
	.dc.b	$3F,$9D,$FE,$9D,$3C,$27,$2F,$F3,$28,$FF,$29,$3F,$E5,$06,$BD,$05
	.dc.b	$77,$C3,$FA,$2A,$F8,$EE,$73,$A6,$C6,$63,$A7,$7E,$26,$47,$A9,$E3
	.dc.b	$69,$CC,$F8,$63,$FC,$34,$F9,$7C,$3C,$E1,$FD,$5B,$E7,$BF,$9E,$4F
	.dc.b	$F9,$A1,$AF,$A5,$B1,$FD,$72,$BF,$C8,$73,$BE,$7F,$4D,$6B,$4F,$EF
	.dc.b	$4D,$9F,$8D,$E6,$EA,$BC,$FF,$3D,$27,$91,$FE,$78,$F9,$27,$CF,$7C
	.dc.b	$46,$C6,$63,$DF,$D9,$96,$46,$4F,$A5,$18,$7E,$17,$A1,$F1,$8C,$6B
	.dc.b	$FD,$DB,$DC,$FF,$2B,$CF,$FA,$21,$AF,$D9,$6C,$7F,$7C,$E5,$FF,$35
	.dc.b	$4E,$FC,$9F,$B3,$5A,$DF,$FC,$34,$D9,$FE,$EF,$5C,$1E,$95,$F8,$08
	.dc.b	$FF,$74,$47,$FA,$61,$AE,$8F,$F4,$DF,$9F,$F8,$37,$F1,$8D,$C3,$FC
	.dc.b	$1E,$DC,$3F,$C5,$5F,$3B,$FE,$6A,$AE,$3C,$BF,$8E,$AD,$47,$F8,$29
	.dc.b	$AA,$78,$3D,$A7,$F2,$B0,$F9,$EA,$73,$8C,$E5,$F9,$AF,$9C,$FD,$3D
	.dc.b	$FE,$85,$B9,$7F,$89,$30,$F8,$92,$9C,$E5,$FE,$41,$6C,$7F,$87,$9A
	.dc.b	$F9,$6F,$F4,$FF,$3A,$F9,$EA,$49,$FF,$60,$D7,$09,$FF,$35,$47,$FB
	.dc.b	$11,$AF,$ED,$5B,$8F,$F8,$E3,$E7,$9F,$05,$FB,$4F,$9E,$7B,$D4,$98
	.dc.b	$73,$F4,$F5,$F3,$DB,$F9,$B9,$3F,$ED,$2B,$4F,$EF,$BE,$23,$E6,$E7
	.dc.b	$FE,$46,$6F,$07,$F9,$85,$B7,$FE,$3E,$5F,$57,$C7,$D4,$93,$FE,$DE
	.dc.b	$9D,$A7,$6E,$63,$6F,$F1,$38,$D9,$A7,$1E,$4F,$97,$49,$1D,$63,$D0
	.dc.b	$61,$E5,$18,$7F,$DD,$FF,$0C,$F7,$BF,$8D,$B6,$33,$1F,$07,$D3,$CB
	.dc.b	$F0,$FE,$7E,$4E,$8F,$F7,$CF,$CF,$FE,$00,$FF,$48,$38,$F6,$FF,$93
	.dc.b	$C6,$7D,$5B,$12,$78,$C6,$8C,$3F,$CC,$E7,$9B,$63,$9F,$8F,$23,$C2
	.dc.b	$F6,$F9,$7C,$47,$18,$BC,$5F,$E7,$EB,$57,$C0,$F9,$09,$7F,$11,$F3
	.dc.b	$F4,$62,$3C,$5F,$3F,$EA,$31,$87,$FA,$23,$E7,$9E,$4F,$F4,$45,$D5
	.dc.b	$4C,$44,$BF,$F4,$1D,$71,$9D,$46,$3A,$3F,$D5,$97,$D2,$30,$7F,$A2
	.dc.b	$9C,$3F,$D5,$DE,$EE,$83,$87,$57,$E5,$F9,$F9,$1B,$94,$7A,$7C,$3C
	.dc.b	$CF,$58,$B8,$FB,$E7,$FB,$0F,$CF,$BE,$4B,$FC,$90,$DB,$FF,$62,$1F
	.dc.b	$EC,$57,$0E,$03,$87,$6E,$AB,$FD,$67,$F3,$CF,$9F,$FE,$DE,$DA,$F7
	.dc.b	$56,$F4,$FF,$69,$4F,$F8,$F6,$7E,$13,$5F,$EF,$8D,$BE,$BD,$5D,$3A
	.dc.b	$E5,$70,$7B,$C6,$23,$8C,$59,$E3,$1D,$BA,$BC,$BD,$52,$30,$43,$C2
	.dc.b	$E5,$F1,$C7,$F8,$2B,$E7,$DF,$3F,$FE,$0C,$F5,$7F,$C2,$78,$7F,$B5
	.dc.b	$1B,$A7,$FB,$BE,$B7,$87,$DF,$6B,$FE,$19,$9D,$BF,$ED,$06,$A8,$DD
	.dc.b	$75,$EA,$F0,$74,$CA,$FC,$55,$FF,$82,$1E,$93,$97,$B4,$F9,$E2,$47
	.dc.b	$08,$D5,$75,$8E,$73,$8C,$3F,$DC,$70,$BF,$D9,$BA,$21,$D8,$E8,$62
	.dc.b	$BA,$39,$23,$82,$4F,$69,$EB,$3A,$8C,$57,$A2,$30,$7C,$7B,$C5,$F2
	.dc.b	$8C,$1E,$23,$E1,$9E,$3A,$ED,$3E,$D8,$FA,$15,$CA,$73,$1C,$6B,$A3
	.dc.b	$F0,$67,$83,$F0,$EB,$83,$C8,$E1,$5C,$9F,$09,$2B,$F0,$D3,$CA,$FE
	.dc.b	$C4,$76,$7E,$0B,$A9,$F4,$1D,$4F,$82,$78,$4C,$9E,$53,$C2,$7C,$23
	.dc.b	$C1,$5F,$29,$C9,$C4,$C9,$D6,$3C,$0F,$69,$F2,$CF,$18,$C5,$7C,$07
	.dc.b	$C0,$60,$F1,$CF,$07,$C5,$5D,$0F,$75,$FC,$1A,$E5,$5B,$AF,$25,$6E
	.dc.b	$FC,$0F,$C5,$AF,$91,$7C,$1B,$5E,$06,$BE,$C3,$9D,$7D,$3A,$F0,$C7
	.dc.b	$C2,$BF,$9F,$3D,$89,$1E,$87,$81,$87,$DF,$3E,$57,$47,$C9,$8D,$4F
	.dc.b	$48,$C9,$C2,$33,$5E,$6A,$F6,$4F,$8E,$FD,$8D,$3D,$85,$3D,$97,$D5
	.dc.b	$F6,$9E,$12,$18,$23,$F3,$49,$5E,$1B,$F1,$CF,$5B,$F1,$BF,$81,$BF
	.dc.b	$1B,$4F,$19,$4F,$1D,$73,$7C,$37,$E5,$24,$F2,$AF,$91,$3F,$0C,$DD
	.dc.b	$71,$70,$7A,$1F,$C2,$CF,$E1,$E3,$F1,$AD,$7E,$1F,$47,$F3,$3F,$9F
	.dc.b	$91,$FC,$92,$3F,$9B,$18,$7F,$98,$46,$1F,$EE,$5F,$3A,$F9,$FF,$F3
	.dc.b	$AC,$7F,$20,$8C,$78,$7F,$D8,$E1,$0F,$09,$F0,$F9,$8D,$CF,$F6,$9F
	.dc.b	$B0,$7D,$8B,$83,$94,$61,$B1,$99,$E0,$D7,$F9,$8F,$C6,$BE,$13,$FC
	.dc.b	$CF,$5F,$E0,$E8,$C7,$C0,$71,$E2,$E5,$1B,$F3,$71,$4C,$13,$C2,$39
	.dc.b	$61,$D2,$7F,$DE,$F3,$96,$C6,$6F,$AA,$C3,$AC,$F4,$8D,$5F,$7B,$E5
	.dc.b	$7C,$30,$DC,$62,$75,$4F,$9F,$CA,$FA,$0B,$18,$6C,$FA,$1B,$8F,$99
	.dc.b	$F1,$AF,$9F,$7C,$FF,$CB,$8F,$EB,$71,$8A,$FB,$8D,$8F,$F1,$03,$5F
	.dc.b	$F1,$2B,$9C,$FE,$A3,$67,$E6,$7C,$53,$DC,$7B,$7E,$2A,$D6,$1F,$17
	.dc.b	$C5,$F6,$E5,$EE,$6E,$3F,$49,$B3,$F4,$1B,$97,$D0,$6C,$FC,$86,$CF
	.dc.b	$D1,$F4,$7E,$49,$F2,$BE,$7E,$D7,$DE,$8B,$3F,$72,$6D,$FE,$3F,$E7
	.dc.b	$7F,$D6,$B7,$FE,$5A,$87,$CF,$FD,$CB,$FC,$55,$69,$4F,$E4,$67,$50
	.dc.b	$9D,$4F,$5E,$DF,$C3,$EF,$F1,$FC,$9F,$AF,$5A,$C7,$F0,$7F,$3F,$F8
	.dc.b	$23,$F0,$7D,$96,$3F,$2F,$17,$2F,$CA,$CD,$CB,$F2,$B3,$6B,$F1,$B3
	.dc.b	$6B,$F2,$D1,$F8,$EF,$9B,$FC,$7B,$4F,$D1,$96,$3F,$42,$DB,$F8,$DF
	.dc.b	$3D,$FA,$0E,$E7,$CF,$FE,$8E,$BF,$CF,$99,$AB,$53,$9C,$3E,$3D,$7F
	.dc.b	$8B,$2B,$4F,$0A,$DF,$5F,$C4,$7F,$4F,$4F,$9F,$A6,$3E,$16,$DF,$F0
	.dc.b	$39,$D3,$67,$4D,$24,$7E,$F8,$D9,$67,$9B,$E6,$3F,$84,$79,$5F,$70
	.dc.b	$FB,$BF,$3C,$DB,$F5,$F1,$F8,$0D,$57,$FB,$26,$E1,$FE,$9F,$5E,$3E
	.dc.b	$67,$F6,$FF,$3A,$F8,$4E,$13,$86,$DF,$FA,$9E,$7F,$DC,$4F,$FA,$A7
	.dc.b	$E7,$9E,$47,$2F,$3F,$F8,$0C,$62,$78,$78,$19,$69,$CE,$7C,$31,$23
	.dc.b	$FD,$69,$F3,$CF,$9E,$C7,$0F,$C3,$3F,$94,$FC,$5F,$F2,$0A,$7C,$E7
	.dc.b	$F9,$95,$FF,$5E,$79,$29,$D4,$E6,$B9,$4E,$6B,$FD,$8E,$72,$F4,$7F
	.dc.b	$8C,$C6,$67,$E9,$4E,$4E,$B7,$A3,$AF,$CF,$58,$FF,$66,$0F,$5B,$CD
	.dc.b	$1F,$36,$33,$18,$E1,$FE,$4F,$1B,$3C,$73,$BB,$F8,$11,$97,$C7,$19
	.dc.b	$ED,$FE,$AD,$39,$6D,$FC,$0F,$93,$EA,$FF,$34,$F9,$E7,$C2,$7C,$43
	.dc.b	$05,$BF,$F7,$24,$69,$F8,$BB,$5C,$BE,$94,$DC,$7F,$B4,$57,$D2,$BE
	.dc.b	$6F,$27,$E8,$FC,$26,$2F,$ED,$72,$FE,$CE,$F8,$E2,$F9,$BF,$E9,$33
	.dc.b	$9C,$7F,$B4,$1F,$7A,$33,$E9,$7D,$C6,$C7,$99,$B8,$76,$9F,$C0,$F5
	.dc.b	$E3,$1F,$EF,$9E,$3F,$DD,$7F,$C0,$10,$F9,$E7,$CF,$7F,$5A,$AD,$65
	.dc.b	$FD,$E6,$B5,$39,$E4,$FE,$25,$7C,$2B,$FB,$F6,$1D,$A7,$35,$B9,$C9
	.dc.b	$FE,$BC,$C3,$AE,$5F,$E1,$AE,$FF,$EA,$EA,$32,$DF,$95,$FE,$FF,$8B
	.dc.b	$E7,$B3,$49,$CC,$61,$B9,$7A,$DB,$8F,$8B,$0F,$ED,$3B,$7B,$F8,$7E
	.dc.b	$03,$67,$53,$E3,$8C,$C6,$23,$FC,$39,$F3,$DF,$E1,$DF,$9E,$7C,$EF
	.dc.b	$B6,$FF,$C0,$75,$A1,$3A,$7C,$37,$C6,$BF,$DB,$F1,$B9,$D3,$E1,$9D
	.dc.b	$3A,$BA,$E4,$EB,$5C,$94,$6B,$4F,$C7,$57,$09,$ED,$5E,$78,$C4,$70
	.dc.b	$39,$9E,$62,$9F,$EF,$4A,$E9,$87,$F9,$27,$0F,$3C,$F4,$38,$38,$7A
	.dc.b	$3E,$CC,$3C,$71,$D0,$E0,$7A,$E3,$85,$76,$BF,$2B,$67,$81,$D3,$0F
	.dc.b	$53,$88,$D5,$72,$8F,$F3,$51,$ED,$3F,$CD,$91,$87,$07,$F9,$5C,$7C
	.dc.b	$EF,$F1,$BF,$6B,$FC,$E4,$D8,$F9,$CD,$7E,$CB,$5E,$4B,$E0,$FF,$C2
	.dc.b	$F1,$C1,$AF,$06,$BC,$1A,$FF,$9A,$63,$97,$07,$E0,$3E,$79,$F3,$3E
	.dc.b	$7C,$BF,$0F,$18,$BF,$C2,$DF,$07,$EC,$B6,$77,$5C,$73,$F8,$17,$C6
	.dc.b	$F7,$8F,$A2,$FF,$27,$87,$1B,$DE,$1C,$6F,$FC,$FD,$87,$1B,$DE,$1C
	.dc.b	$6F,$D9,$8F,$A7,$0E,$18,$30,$E1,$5A,$8F,$0D,$6D,$86,$09,$87,$A9
	.dc.b	$CB,$E0,$8F,$83,$87,$86,$7C,$B8,$78,$30,$EF,$1F,$36,$74,$F8,$E7
	.dc.b	$F0,$55,$FE,$A3,$AF,$88,$D8,$CE,$9E,$7C,$3C,$57,$FD,$DE,$F7,$3F
	.dc.b	$1A,$3F,$85,$E5,$F2,$5F,$F5,$27,$CF,$7F,$A6,$DF,$E1,$ED,$7F,$C0
	.dc.b	$0D,$C7,$FD,$69,$F3,$CE,$1F,$0D,$72,$E9,$AA,$7F,$C8,$4F,$F1,$A6
	.dc.b	$BD,$23,$76,$F8,$AA,$49,$E1,$E0,$57,$C7,$17,$1F,$1C,$DC,$FF,$51
	.dc.b	$8B,$C4,$F1,$B7,$0F,$6E,$1E,$6F,$9E,$FE,$71,$1F,$CB,$9A,$FF,$87
	.dc.b	$DB,$8F,$C0,$F3,$38,$74,$4A,$E0,$F6,$8C,$47,$F8,$9F,$07,$CF,$78
	.dc.b	$8C,$DF,$FB,$77,$E3,$DF,$3D,$F2,$1B,$7F,$D0,$1A,$F3,$EC,$D4,$C8
	.dc.b	$D3,$EE,$8C,$47,$87,$D0,$F8,$26,$2F,$FD,$D9,$F0,$5F,$E4,$F9,$FE
	.dc.b	$9A,$D7,$FC,$7C,$D8,$FF,$79,$74,$7F,$97,$E7,$7E,$57,$DE,$AD,$71
	.dc.b	$7F,$17,$6C,$FF,$BB,$EB,$83,$D2,$BF,$0B,$1F,$ED,$48,$FE,$BA,$D7
	.dc.b	$47,$F6,$0F,$3B,$F0,$8D,$E2,$FC,$6C,$62,$3F,$CA,$9F,$3B,$FE,$62
	.dc.b	$74,$43,$FC,$B1,$1B,$C9,$97,$F2,$DA,$E1,$5A,$BF,$CC,$C3,$E7,$A9
	.dc.b	$E0,$8C,$E5,$F9,$EF,$9D,$FD,$15,$FF,$76,$6B,$E4,$8D,$75,$48,$D3
	.dc.b	$C9,$FE,$FB,$F3,$AF,$9E,$FD,$4B,$63,$31,$C5,$FF,$39,$47,$F9,$75
	.dc.b	$E3,$E7,$F0,$3F,$63,$F3,$CF,$82,$FD,$9F,$CF,$3D,$EA,$4B,$E2,$7F
	.dc.b	$83,$A1,$F3,$DC,$3E,$66,$4E,$63,$F7,$BF,$11,$F2,$39,$3E,$5C,$84
	.dc.b	$67,$93,$FA,$D3,$63,$4B,$E7,$B9,$FA,$92,$7F,$C3,$F3,$B7,$9C,$E5
	.dc.b	$FF,$49,$46,$49,$1E,$4F,$2B,$CB,$BC,$AE,$71,$97,$AC,$65,$E7,$1F
	.dc.b	$C6,$3E,$15,$EE,$D2,$4F,$93,$D3,$AF,$8B,$E7,$BE,$19,$8D,$7E,$1C
	.dc.b	$3E,$7B,$F0,$71,$87,$FD,$03,$1F,$E3,$C6,$BF,$EA,$D7,$7B,$7F,$80
	.dc.b	$5B,$1F,$35,$DC,$F3,$F6,$BF,$97,$C6,$A7,$C1,$D1,$89,$91,$DE,$37
	.dc.b	$E3,$C8,$F1,$3C,$BE,$5B,$11,$E3,$8B,$87,$A6,$7E,$37,$CF,$4A,$FF
	.dc.b	$2A,$EC,$8F,$1F,$46,$67,$D9,$F3,$DE,$63,$18,$7C,$6F,$9E,$79,$3F
	.dc.b	$D2,$17,$4D,$31,$12,$FF,$CC,$73,$A6,$CE,$5F,$F6,$B4,$74,$72,$43
	.dc.b	$33,$DF,$E7,$A4,$74,$7C,$5E,$16,$23,$9C,$5C,$3D,$13,$FD,$7B,$E7
	.dc.b	$9F,$25,$49,$7F,$E7,$8F,$9E,$FC,$D4,$EA,$31,$F2,$3F,$D9,$A4,$F0
	.dc.b	$8F,$F7,$74,$7F,$8D,$DE,$5D,$5B,$38,$75,$D8,$C9,$0E,$2E,$5E,$71
	.dc.b	$72,$FC,$AD,$7F,$09,$9E,$FD,$5E,$3E,$B9,$1E,$98,$F3,$18,$7C,$B1
	.dc.b	$87,$FD,$2B,$F3,$AF,$9E,$FF,$00,$B6,$33,$1F,$E0,$4F,$57,$F2,$D3
	.dc.b	$99,$FF,$43,$35,$FF,$48,$B5,$FF,$31,$35,$FE,$1D,$0E,$9E,$9D,$32
	.dc.b	$7F,$D4,$F3,$96,$CE,$56,$7F,$B6,$95,$79,$6A,$57,$28,$E1,$5B,$9C
	.dc.b	$CF,$2E,$DF,$F3,$B5,$F2,$8E,$07,$80,$86,$CF,$4D,$F4,$9E,$35,$D2
	.dc.b	$FC,$93,$9A,$F7,$CE,$6B,$BC,$E6,$7F,$C3,$D3,$97,$A9,$E2,$3C,$06
	.dc.b	$8F,$1D,$76,$74,$FB,$5C,$D7,$67,$2F,$99,$C9,$D8,$EB,$7F,$05,$B5
	.dc.b	$E3,$7F,$C6,$57,$E7,$9F,$A9,$1A,$3A,$9C,$EF,$C4,$E4,$EB,$3C,$6B
	.dc.b	$A5,$77,$AF,$55,$6D,$E2,$60,$97,$E9,$8D,$3E,$D7,$A9,$EC,$3E,$31
	.dc.b	$82,$47,$48,$E4,$F4,$8F,$44,$60,$8E,$08,$74,$8E,$4E,$27,$D5,$3C
	.dc.b	$23,$4F,$96,$7B,$46,$0F,$C1,$9F,$20,$F9,$A7,$A6,$7E,$6C,$70,$7A
	.dc.b	$C7,$03,$E0,$9E,$1B,$F9,$77,$F5,$5B,$1D,$67,$8B,$E1,$23,$82,$4F
	.dc.b	$D4,$9F,$AD,$3E,$36,$BE,$03,$CE,$7C,$CB,$F8,$15,$E8,$7E,$01,$23
	.dc.b	$D1,$7F,$21,$E2,$FB,$6F,$C1,$1E,$76,$BD,$4E,$33,$DA,$FC,$75,$CE
	.dc.b	$FD,$51,$F0,$2B,$81,$DA,$BC,$B7,$F1,$63,$E7,$35,$F6,$57,$A5,$F9
	.dc.b	$86,$8F,$95,$5D,$6F,$EA,$36,$FD,$D3,$DC,$F3,$90,$C1,$2B,$B3,$F4
	.dc.b	$A7,$D7,$3E,$D8,$F8,$47,$D1,$29,$E6,$8D,$BD,$67,$E0,$BF,$89,$E8
	.dc.b	$FE,$6D,$F3,$D2,$3F,$95,$C7,$F3,$93,$0F,$F0,$08,$C3,$FE,$41,$C3
	.dc.b	$FC,$43,$F3,$9F,$D6,$63,$FC,$88,$47,$0E,$03,$87,$5C,$19,$78,$FC
	.dc.b	$EB,$E7,$BF,$CE,$F1,$86,$BF,$E5,$18,$77,$3E,$23,$94,$A7,$97,$8D
	.dc.b	$D7,$E3,$1F,$07,$FE,$65,$E1,$F5,$1F,$E5,$B8,$CF,$BF,$FE,$5E,$DB
	.dc.b	$FC,$BF,$1C,$3C,$CC,$56,$ED,$78,$C9,$7B,$AE,$B1,$8A,$DC,$62,$75
	.dc.b	$3F,$D1,$DF,$0C,$66,$7C,$D3,$97,$8B,$6F,$C1,$5C,$23,$5F,$3B,$8A
	.dc.b	$7C,$ED,$8C,$37,$0E,$CD,$AE,$CF,$AD,$7C,$EB,$E7,$7D,$DB,$BE,$83
	.dc.b	$FC,$85,$87,$BA,$F7,$5A,$9C,$C7,$F5,$4B,$E5,$5B,$6C,$7F,$67,$F8
	.dc.b	$A7,$B8,$F6,$FC,$55,$AE,$1F,$A7,$B7,$F8,$CB,$1F,$8D,$9F,$BF,$C5
	.dc.b	$F8,$0D,$8C,$9F,$2E,$32,$D3,$39,$FC,$D8,$CB,$F3,$FD,$2F,$94,$7C
	.dc.b	$BF,$9E,$B5,$F5,$62,$CF,$D5,$9B,$8F,$FA,$13,$E6,$FF,$D6,$97,$F8
	.dc.b	$28,$7C,$F7,$DC,$75,$DD,$F8,$97,$3C,$DF,$46,$F8,$E3,$FD,$F1,$B3
	.dc.b	$FE,$B2,$E1,$F8,$FE,$46,$F0,$FC,$07,$CF,$4C,$FF,$15,$D9,$72,$FC
	.dc.b	$54,$DC,$7E,$5C,$DA,$FC,$E4,$DA,$FC,$C4,$DE,$9F,$D4,$D7,$E3,$7E
	.dc.b	$6B,$F1,$CD,$3F,$30,$5A,$FC,$C3,$70,$F8,$9F,$3B,$F8,$AF,$9E,$FC
	.dc.b	$DC,$62,$7F,$4F,$D4,$9C,$C7,$F0,$CD,$D7,$F1,$4E,$5C,$16,$7F,$59
	.dc.b	$8B,$E7,$BF,$4A,$D8,$CF,$C2,$E2,$64,$70,$77,$18,$9F,$DA,$1C,$4B
	.dc.b	$1A,$9E,$07,$E8,$7C,$CF,$B8,$FF,$22,$6C,$FF,$B9,$FE,$79,$B7,$EB
	.dc.b	$67,$33,$FC,$67,$0F,$D8,$6A,$6F,$95,$87,$F6,$9F,$3C,$F8,$3C,$E7
	.dc.b	$FB,$18,$C3,$FB,$3B,$FF,$34,$39,$F3,$FF,$79,$F3,$CF,$1F,$9C,$61
	.dc.b	$EB,$E1,$62,$BA,$37,$4F,$F2,$6E,$3F,$B5,$9C,$9F,$C0,$7E,$79,$F3
	.dc.b	$DC,$A3,$F8,$27,$43,$E2,$FF,$93,$43,$E7,$1E,$8F,$F0,$FA,$D4,$FF
	.dc.b	$28,$35,$19,$AE,$35,$C2,$7A,$C6,$27,$51,$83,$94,$BE,$7B,$39,$FF
	.dc.b	$99,$A5,$EB,$F9,$27,$E4,$CE,$63,$1C,$5F,$C8,$EB,$67,$96,$B7,$DA
	.dc.b	$8F,$EC,$26,$AF,$FD,$04,$F8,$6B,$95,$C8,$F2,$FC,$9D,$AF,$F3,$CF
	.dc.b	$CF,$3E,$0F,$DE,$7F,$33,$71,$5E,$C8,$C1,$A7,$F9,$94,$6A,$7C,$39
	.dc.b	$DC,$BC,$D0,$13,$9A,$FE,$F2,$D8,$F6,$53,$E0,$FB,$D7,$F3,$FE,$4F
	.dc.b	$DA,$D6,$BE,$6F,$FA,$44,$62,$7E,$F4,$63,$D3,$FB,$2D,$8F,$AA,$DA
	.dc.b	$FF,$4C,$D7,$1C,$3F,$BE,$57,$7B,$FE,$A9,$E4,$7E,$1A,$FF,$D5,$DD
	.dc.b	$4F,$9E,$7C,$F6,$B9,$7F,$AB,$C6,$31,$FE,$AD,$87,$1D,$FE,$1E,$1C
	.dc.b	$70,$E1,$5F,$C4,$AF,$86,$3F,$EA,$CC,$38,$4F,$F8,$79,$B1,$FE,$11
	.dc.b	$EF,$FE,$AD,$39,$CB,$CF,$F2,$DF,$CF,$7E,$7B,$76,$C1,$39,$AD,$DF
	.dc.b	$F4,$7B,$DD,$6A,$31,$3F,$0B,$83,$E1,$F0,$D9,$F6,$B7,$1F,$44,$FF
	.dc.b	$7B,$7F,$01,$F3,$DF,$81,$F9,$D7,$CF,$7A,$EB,$54,$AD,$4E,$70,$B3
	.dc.b	$FE,$3B,$9A,$FA,$EB,$F1,$B7,$D2,$34,$1C,$3B,$AE,$2F,$90,$D5,$F7
	.dc.b	$69,$CA,$78,$E4,$BE,$F5,$FE,$49,$C3,$C7,$39,$C3,$AE,$1E,$B3,$99
	.dc.b	$E1,$29,$E1,$C3,$AE,$1D,$23,$C1,$87,$8E,$7D,$31,$B8,$C1,$E1,$7A
	.dc.b	$DF,$85,$EA,$78,$67,$AD,$F9,$30,$F3,$E1,$DE,$BF,$CD,$31,$89,$D4
	.dc.b	$7F,$97,$0F,$51,$FE,$5E,$8C,$38,$3F,$CE,$3F,$3B,$FC,$73,$D6,$FF
	.dc.b	$31,$B6,$3F,$CD,$CD,$7F,$CE,$8D,$78,$FB,$CE,$4E,$2D,$8E,$2D,$78
	.dc.b	$B5,$E3,$1F,$84,$F9,$E7,$CC,$FF,$A1,$34,$FC,$0C,$62,$BF,$13,$83
	.dc.b	$0E,$D5,$F8,$68,$C6,$1C,$F0,$E5,$B7,$FA,$3B,$1E,$58,$FD,$CB,$DB
	.dc.b	$71,$E7,$5B,$BE,$33,$FC,$86,$F8,$E3,$DE,$F7,$3F,$C9,$70,$E7,$3E
	.dc.b	$8B,$E3,$2B,$E3,$87,$C0,$8F,$06,$3E,$76,$18,$25,$F3,$72,$F8,$63
	.dc.b	$E1,$E3,$E2,$9E,$D8,$78,$30,$EF,$1E,$59,$D4,$7C,$2A,$FE,$7B,$3C
	.dc.b	$0F,$F3,$A4,$67,$4F,$36,$1E,$19,$FB,$D1,$F1,$5D,$E1,$FE,$BA,$BE
	.dc.b	$79,$7C,$2C,$BF,$D7,$BF,$3D,$FE,$A0,$BF,$F4,$03,$5E,$67,$F4,$D7
	.dc.b	$15,$E4,$F9,$DF,$E9,$53,$C3,$A7,$B6,$5F,$EC,$89,$FF,$1E,$CF,$FA
	.dc.b	$2D,$AF,$41,$7D,$27,$C7,$F1,$5C,$A7,$C0,$D7,$6D,$A7,$7E,$6D,$34
	.dc.b	$95,$E1,$9A,$F8,$6B,$FC,$2E,$F9,$7C,$3C,$E5,$F4,$3E,$7B,$F9,$DC
	.dc.b	$FF,$A8,$DA,$FA,$5B,$1F,$0F,$CD,$F3,$46,$33,$FF,$74,$B6,$BF,$BB
	.dc.b	$35,$60,$AE,$3D,$5E,$DF,$9E,$95,$E1,$8F,$85,$F1,$CF,$9E,$A4,$9F
	.dc.b	$9D,$D9,$5F,$9D,$3B,$7F,$D7,$8D,$7E,$14,$6F,$4F,$B9,$E8,$7C,$D3
	.dc.b	$1A,$FF,$83,$6B,$FD,$77,$E4,$7D,$3F,$82,$FF,$2C,$47,$FB,$1D,$AF
	.dc.b	$DD,$E7,$FF,$32,$B9,$9E,$04,$7F,$D9,$D1,$A7,$F0,$93,$CB,$0F,$D0
	.dc.b	$CF,$FB,$5E,$77,$F0,$3C,$5F,$C5,$7A,$1F,$80,$7F,$DB,$0D,$73,$CB
	.dc.b	$F8,$9F,$9E,$FF,$33,$46,$A3,$F1,$93,$BC,$3F,$5E,$DC,$FE,$05,$F0
	.dc.b	$BE,$67,$FB,$95,$7C,$F7,$53,$9E,$1D,$F4,$FC,$C7,$CF,$7E,$9A,$FF
	.dc.b	$40,$DC,$BF,$7B,$87,$BF,$1F,$9D,$39,$6F,$89,$C8,$E9,$7F,$EF,$5F
	.dc.b	$9D,$7C,$F7,$E9,$5B,$19,$8E,$0E,$FD,$0D,$11,$FF,$7F,$7C,$F3,$E0
	.dc.b	$BF,$5D,$F3,$CF,$7A,$92,$B9,$1F,$E4,$FD,$BF,$8A,$CB,$E7,$B8,$FC
	.dc.b	$CC,$AD,$6D,$F8,$9A,$D4,$60,$FF,$2C,$7E,$39,$FA,$32,$76,$FE,$FF
	.dc.b	$F3,$77,$CD,$7F,$01,$9F,$D4,$43,$EC,$B4,$66,$A5,$F5,$13,$B7,$0D
	.dc.b	$7F,$89,$57,$EC,$FE,$37,$EF,$6B,$FC,$C1,$B9,$FE,$85,$B5,$FE,$6F
	.dc.b	$5F,$43,$FC,$55,$CC,$1A,$48,$D7,$9F,$BE,$39,$5E,$7A,$FE,$D0,$D9
	.dc.b	$CB,$FC,$9E,$3F,$D3,$AE,$4F,$77,$C5,$3E,$0F,$F9,$33,$63,$31,$E5
	.dc.b	$F5,$F2,$F9,$FE,$86,$4E,$88,$77,$72,$FF,$37,$AE,$8F,$F3,$9F,$A1
	.dc.b	$FC,$B3,$FE,$91,$7F,$D3,$CF,$D1,$6B,$BE,$BF,$E7,$51,$C2,$0E,$B6
	.dc.b	$FE,$BE,$DD,$BF,$88,$3B,$3F,$93,$4F,$5F,$2B,$99,$D3,$E6,$AC,$64
	.dc.b	$87,$9D,$F8,$BF,$43,$23,$E4,$BE,$4E,$BC,$47,$B7,$E8,$7C,$E6,$32
	.dc.b	$FE,$E7,$A2,$F3,$5E,$FA,$D6,$7F,$C9,$FE,$85,$E6,$7F,$59,$5D,$94
	.dc.b	$C4,$4B,$FC,$C4,$E9,$B3,$97,$FB,$44,$7E,$91,$DF,$5F,$53,$B7,$8B
	.dc.b	$63,$24,$38,$BF,$DA,$FE,$85,$F2,$D4,$93,$FA,$9F,$A1,$FC,$DC,$EA
	.dc.b	$31,$F2,$BF,$DC,$E4,$F0,$8F,$EF,$93,$B7,$FB,$F4,$E9,$E2,$7E,$D6
	.dc.b	$35,$C1,$FE,$1E,$9F,$D0,$CF,$0E,$AE,$C7,$7E,$BB,$1E,$58,$AF,$68
	.dc.b	$F2,$75,$79,$7A,$A4,$FA,$E3,$2F,$B6,$32,$F9,$A3,$FC,$1D,$F4,$0F
	.dc.b	$A1,$FF,$08,$7A,$FF,$E1,$8C,$FF,$C6,$6D,$E4,$FE,$C9,$5B,$C3,$F8
	.dc.b	$83,$5F,$F0,$F4,$ED,$FF,$7B,$35,$46,$EB,$FC,$7E,$DA,$FE,$E1,$5B
	.dc.b	$38,$46,$6B,$FE,$00,$35,$46,$FE,$53,$D7,$84,$9F,$E4,$93,$97,$CF
	.dc.b	$19,$7C,$F3,$F0,$0C,$61,$E4,$9C,$4F,$29,$CE,$1F,$0A,$73,$3C,$8E
	.dc.b	$A4,$30,$4A,$E8,$E5,$F9,$B5,$C9,$CC,$7F,$91,$F0,$F7,$61,$E9,$8C
	.dc.b	$5F,$89,$F9,$B7,$C7,$0F,$24,$75,$7E,$5C,$66,$3E,$4C,$F1,$7E,$4C
	.dc.b	$65,$FA,$B1,$9A,$F0,$61,$E4,$25,$F4,$AF,$88,$77,$3A,$98,$3E,$03
	.dc.b	$E4,$7D,$26,$63,$E0,$BC,$4F,$09,$98,$EC,$F0,$24,$70,$7A,$9E,$03
	.dc.b	$31,$C2,$34,$7A,$23,$4E,$23,$47,$95,$D4,$7C,$43,$C4,$70,$8D,$1C
	.dc.b	$1F,$A2,$7A,$70,$FA,$F8,$7C,$B7,$F2,$C6,$23,$99,$EE,$3C,$E7,$23
	.dc.b	$D6,$66,$BC,$77,$F4,$AF,$C3,$3E,$33,$84,$FA,$DC,$1C,$63,$CC,$FA
	.dc.b	$CE,$E6,$4E,$2E,$AB,$CF,$87,$E0,$EF,$F6,$2F,$E0,$F0,$F6,$36,$FD
	.dc.b	$75,$C1,$F6,$D7,$07,$EE,$3E,$97,$CB,$1F,$0D,$F7,$BC,$5F,$83,$1C
	.dc.b	$5F,$3B,$C5,$F4,$11,$C4,$71,$C3,$E1,$4F,$BD,$FB,$71,$DD,$CB,$BC
	.dc.b	$3E,$2D,$77,$9F,$9F,$87,$BA,$FD,$2D,$7C,$6D,$8D,$C7,$C2,$C3,$E2
	.dc.b	$3E,$C6,$C7,$81,$B7,$F3,$DF,$E5,$FD,$7F,$CC,$3A,$E5,$FF,$18,$8C
	.dc.b	$BF,$C6,$A3,$2F,$F2,$E8,$FE,$6F,$D7,$FC,$E3,$E8,$64,$FF,$20,$8C
	.dc.b	$BF,$C8,$A3,$2F,$F3,$68,$FF,$67,$7C,$FB,$E8,$3F,$D0,$13,$FE,$A1
	.dc.b	$8C,$7C,$67,$FA,$8E,$31,$DE,$F8,$BC,$40,$57,$07,$3F,$42,$FA,$2F
	.dc.b	$F7,$EC,$7F,$AB,$63,$1B,$F2,$84,$61,$E3,$3C,$1E,$D1,$FE,$13,$9E
	.dc.b	$0D,$8E,$FE,$16,$DE,$6F,$1F,$92,$7C,$47,$F9,$FE,$FF,$CE,$91,$8F
	.dc.b	$84,$46,$27,$56,$E4,$CA,$BE,$15,$BF,$47,$35,$A9,$CD,$71,$AF,$8D
	.dc.b	$1E,$2A,$DC,$62,$D8,$62,$FB,$43,$E8,$B1,$87,$AB,$E8,$AC,$E7,$0F
	.dc.b	$44,$62,$B2,$D8,$C9,$FE,$87,$FA,$17,$BB,$DF,$85,$F4,$5F,$2F,$6F
	.dc.b	$7E,$8B,$DE,$1F,$E9,$BB,$DE,$FE,$7A,$D6,$36,$F7,$1F,$04,$7C,$7B
	.dc.b	$E0,$5F,$01,$BB,$99,$79,$5F,$85,$5B,$9C,$DF,$3E,$0F,$B7,$E4,$7C
	.dc.b	$2B,$F8,$6D,$C3,$ED,$36,$B8,$70,$FF,$90,$A7,$EF,$57,$CA,$F4,$A4
	.dc.b	$7E,$07,$E8,$AD,$F0,$9B,$C1,$F7,$70,$56,$2F,$F4,$56,$F9,$FF,$F6
	.dc.b	$74,$7F,$B0,$6D,$F4,$5F,$8C,$7E,$6C,$EA,$BE,$34,$6B,$9F,$E6,$4E
	.dc.b	$63,$7D,$DF,$13,$87,$FC,$A7,$8F,$FB,$06,$B4,$F8,$9B,$87,$ED,$5B
	.dc.b	$C3,$F4,$7C,$CE,$B8,$7E,$2B,$E8,$A6,$1F,$A1,$EE,$B8,$FC,$89,$B9
	.dc.b	$FC,$19,$B5,$F9,$93,$F2,$7F,$3F,$F9,$46,$9F,$AB,$2D,$FD,$26,$D7
	.dc.b	$C2,$F1,$BF,$25,$F4,$5F,$A2,$9F,$64,$63,$BB,$E1,$B2,$3F,$25,$3F
	.dc.b	$2D,$F8,$11,$9B,$FC,$81,$F4,$DE,$BC,$19,$C3,$F3,$BE,$57,$1A,$FE
	.dc.b	$34,$D5,$F4,$2C,$FF,$3D,$39,$ED,$7C,$4F,$E8,$5B,$9F,$FB,$CD,$79
	.dc.b	$19,$9F,$D3,$7D,$03,$E2,$38,$4E,$1B,$7F,$EF,$B9,$FF,$6F,$3F,$EF
	.dc.b	$DF,$A1,$79,$3C,$BD,$0F,$D8,$46,$27,$87,$85,$96,$9C,$E7,$C3,$12
	.dc.b	$3F,$9C,$70,$7E,$FB,$E8,$5F,$43,$9E,$2F,$93,$E2,$3E,$3B,$F8,$A6
	.dc.b	$0F,$9E,$FF,$38,$BC,$8F,$D0,$CE,$DC,$57,$F2,$C9,$DB,$D2,$B9,$C6
	.dc.b	$31,$FE,$1F,$7C,$6B,$F8,$45,$F7,$F4,$FF,$87,$4E,$6B,$78,$72,$FA
	.dc.b	$1C,$BF,$C4,$ED,$EB,$7A,$27,$E8,$47,$93,$1F,$E7,$F3,$9E,$2E,$0D
	.dc.b	$9D,$1C,$8F,$58,$AE,$75,$F8,$E8,$D3,$67,$DB,$F2,$BD,$27,$C1,$F4
	.dc.b	$0F,$A1,$F8,$27,$F2,$97,$11,$DB,$6B,$87,$E6,$A6,$E9,$FD,$4A,$BE
	.dc.b	$95,$F1,$39,$BD,$9C,$3C,$FE,$12,$DF,$E9,$39,$7F,$5C,$FC,$EC,$5F
	.dc.b	$3D,$FC,$EE,$73,$8F,$79,$CE,$1F,$26,$72,$FF,$80,$4F,$C2,$57,$1F
	.dc.b	$07,$EE,$B6,$38,$47,$E0,$3A,$F1,$8F,$ED,$BC,$6F,$B9,$5F,$E0,$B9
	.dc.b	$7D,$03,$E8,$7F,$A7,$56,$AF,$FC,$25,$39,$BF,$B9,$39,$BE,$37,$C3
	.dc.b	$93,$FA,$F5,$EF,$0F,$04,$67,$0F,$EB,$9D,$FF,$D6,$36,$FE,$C3,$97
	.dc.b	$7B,$FE,$B3,$F1,$9F,$DD,$65,$F4,$39,$A4,$E6,$30,$DC,$7D,$2D,$C7
	.dc.b	$CF,$B7,$B3,$87,$DA,$D8,$C9,$32,$FE,$36,$7F,$8B,$FE,$87,$FC,$61
	.dc.b	$F4,$0F,$9F,$FE,$A3,$A7,$D6,$AD,$8A,$DC,$EB,$0F,$F0,$DD,$F1,$AF
	.dc.b	$F1,$F5,$69,$FF,$1E,$5F,$39,$D4,$7F,$5E,$9D,$3F,$DF,$2B,$9D,$FF
	.dc.b	$91,$27,$37,$C6,$31,$4B,$E7,$39,$8F,$09,$99,$F3,$61,$DA,$3E,$01
	.dc.b	$93,$CB,$87,$81,$E3,$7D,$23,$11,$E7,$38,$1E,$79,$E8,$FC,$53,$37
	.dc.b	$E3,$78,$57,$39,$DE,$1E,$CB,$EB,$5E,$4A,$D3,$D8,$C6,$1E,$A7,$07
	.dc.b	$73,$FC,$C1,$5F,$E6,$17,$DA,$60,$FF,$36,$FD,$07,$F1,$AF,$4F,$FC
	.dc.b	$CC,$D8,$CF,$57,$F9,$71,$B3,$AF,$84,$FF,$2C,$47,$26,$BF,$E7,$BA
	.dc.b	$46,$23,$FC,$33,$18,$E1,$E4,$7D,$EF,$A1,$7C,$E7,$F8,$7B,$4F,$C1
	.dc.b	$C6,$2F,$F0,$77,$C5,$F9,$ED,$9D,$D7,$1D,$3F,$C4,$F7,$C6,$F7,$8F
	.dc.b	$A2,$FF,$1F,$87,$1B,$DE,$1C,$6F,$FD,$21,$87,$1B,$DE,$1C,$6F,$D6
	.dc.b	$DA,$DD,$F1,$57,$C6,$B7,$8F,$AA,$F8,$56,$AF,$85,$EE,$3F,$D4,$18
	.dc.b	$F9,$5B,$5C,$27,$51,$C4,$E9,$71,$FC,$9C,$68,$DD,$73,$8E,$07,$F8
	.dc.b	$6A,$3E,$34,$FF,$9D,$1B,$B7,$A2,$79,$E1,$FE,$52,$AD,$36,$7E,$4C
	.dc.b	$FF,$61,$BE,$DF,$43,$FE,$90,$8F,$EC,$8D,$7F,$C5,$2D,$BF,$F4,$1F
	.dc.b	$D0,$38,$BA,$57,$FB,$06,$B5,$87,$F9,$8A,$9E,$7E,$A9,$18,$7B,$F5
	.dc.b	$E6,$DF,$17,$49,$3F,$DD,$E9,$E0,$E2,$3C,$B3,$73,$F2,$CD,$CB,$C5
	.dc.b	$E2,$F8,$70,$39,$C3,$FD,$B1,$F4,$3F,$CE,$1F,$F0,$23,$5F,$F2,$4B
	.dc.b	$70,$F8,$3E,$6F,$AE,$31,$87,$FA,$01,$B5,$D3,$AB,$C1,$5B,$7C,$5F
	.dc.b	$43,$27,$BC,$7C,$5F,$91,$7D,$0D,$24,$FC,$1E,$EA,$FF,$97,$27,$6F
	.dc.b	$F8,$6D,$AF,$F9,$83,$5F,$F1,$C3,$9A,$F8,$5E,$8C,$8C,$1F,$E2,$3F
	.dc.b	$6F,$FC,$A8,$DF,$2F,$EC,$46,$23,$FC,$D7,$D1,$FE,$63,$74,$43,$FC
	.dc.b	$DF,$1B,$E0,$FF,$64,$CE,$E7,$35,$F8,$38,$FF,$68,$C7,$F8,$CD,$AE
	.dc.b	$8F,$F1,$B7,$9F,$F8,$46,$F2,$7E,$2E,$31,$1F,$E7,$EF,$A0,$FF,$32
	.dc.b	$BA,$21,$FE,$82,$8D,$E1,$FE,$DC,$9D,$CE,$6B,$F2,$D0,$FA,$1A,$77
	.dc.b	$8C,$E3,$F9,$BF,$A1,$FD,$15,$FF,$92,$DA,$F8,$E3,$5D,$72,$34,$F2
	.dc.b	$69,$FE,$53,$FA,$17,$D0,$FE,$9D,$B1,$A3,$FD,$37,$F4,$2F,$7F,$5E
	.dc.b	$34,$9F,$F2,$E4,$E5,$FF,$2F,$7D,$03,$DD,$FD,$93,$63,$31,$C5,$FF
	.dc.b	$53,$C3,$E8,$7F,$AA,$36,$3F,$DB,$A7,$FB,$79,$C3,$80,$E1,$DF,$CF
	.dc.b	$FE,$91,$34,$C7,$D0,$3E,$6B,$F8,$34,$7E,$C6,$78,$F6,$FF,$65,$8C
	.dc.b	$47,$FB,$B9,$A8,$EB,$1F,$C3,$E9,$F4,$34,$EB,$1F,$C4,$FE,$2B,$F0
	.dc.b	$3C,$EF,$0F,$FD,$F3,$18,$6F,$AF,$BF,$A1,$E0,$CC,$BF,$D1,$55,$B7
	.dc.b	$C4,$A7,$B7,$9B,$DF,$8A,$57,$08,$CB,$CA,$32,$FF,$B6,$2B,$FB,$AB
	.dc.b	$FE,$97,$F8,$67,$C0,$FF,$2B,$6C,$66,$3D,$9E,$BE,$5F,$F0,$64,$EA
	.dc.b	$7E,$B3,$5F,$B6,$D7,$E5,$07,$86,$9F,$16,$3D,$BF,$43,$F2,$4E,$CF
	.dc.b	$FA,$AF,$E8,$7F,$43,$3B,$84,$7F,$A7,$DD,$7B,$DF,$D0,$E3,$7B,$F8
	.dc.b	$1B,$3F,$6D,$E0,$FF,$BE,$9B,$5D,$5D,$C7,$6F,$2F,$A1,$1F,$27,$CD
	.dc.b	$57,$94,$5A,$E5,$3F,$2F,$E4,$A5,$7C,$F7,$CD,$D9,$88,$F9,$31,$F3
	.dc.b	$08,$FE,$84,$F9,$0F,$B7,$E8,$64,$73,$3F,$DA,$1F,$40,$F3,$7F,$B2
	.dc.b	$2F,$A1,$A6,$22,$4F,$08,$CB,$CA,$38,$F6,$48,$EA,$F4,$F1,$B1,$0D
	.dc.b	$3E,$A6,$1F,$AF,$8D,$9E,$49,$FE,$ED,$F4,$2F,$95,$A4,$9F,$F7,$47
	.dc.b	$D0,$FE,$7D,$4E,$DF,$84,$D7,$F8,$23,$5F,$E1,$2D,$7F,$1E,$EF,$0F
	.dc.b	$C4,$4B,$E4,$BF,$C1,$0D,$AF,$F7,$C5,$6F,$8F,$F2,$F3,$FA,$69,$F7
	.dc.b	$46,$3A,$F8,$9C,$BA,$EC,$FA,$A2,$BC,$0F,$E1,$54,$9F,$0F,$57,$8B
	.dc.b	$AA,$57,$C2,$8C,$BF,$12,$32,$F3,$14,$9F,$F1,$17,$D0,$BE,$86,$92
	.dc.b	$BF,$8D,$6B,$FA,$9F,$63,$F3,$B5,$B9,$CE,$9F,$B6,$8C,$47,$76,$D7
	.dc.b	$1F,$2F,$A4,$E7,$0B,$39,$8E,$67,$F2,$9A,$E3,$7D,$6B,$6F,$F0,$4C
	.dc.b	$3C,$71,$8A,$E0,$FF,$8F,$63,$31,$D0,$57,$47,$66,$63,$6F,$94,$E1
	.dc.b	$1B,$7F,$C6,$F8,$77,$79,$D5,$AE,$38,$7F,$0F,$BE,$07,$18,$CE,$1E
	.dc.b	$BB,$EB,$7E,$7C,$3A,$5F,$08,$C4,$E8,$87,$94,$FB,$97,$D1,$F5,$99
	.dc.b	$C3,$E6,$37,$0F,$03,$4F,$EA,$D3,$C5,$F5,$9D,$A7,$89,$EB,$AF,$29
	.dc.b	$E4,$8E,$47,$02,$CF,$3B,$F5,$47,$CD,$C3,$8B,$D7,$0F,$B5,$18,$8E
	.dc.b	$2D,$C3,$C1,$1C,$4E,$07,$B9,$CB,$F0,$A3,$84,$FB,$A7,$D2,$D2,$4F
	.dc.b	$17,$C2,$FC,$C3,$FC,$17,$5F,$9A,$6E,$1F,$EE,$38,$C4,$7C,$A3,$27
	.dc.b	$F2,$F7,$2F,$D2,$32,$7D,$6F,$73,$F2,$87,$99,$FA,$2E,$1F,$69,$F2
	.dc.b	$EF,$F6,$11,$98,$F9,$F3,$F8,$B8,$C4,$FC,$22,$5F,$E9,$63,$2F,$C0
	.dc.b	$AE,$8E,$5F,$D0,$47,$8D,$FF,$49,$BC,$1F,$8C,$43,$27,$03,$CF,$5E
	.dc.b	$37,$07,$9A,$F9,$3C,$0C,$1A,$BF,$6B,$CA,$37,$8F,$FA,$CE,$36,$7C
	.dc.b	$93,$86,$1E,$BB,$F0,$46,$4D,$DF,$47,$0F,$C4,$7D,$55,$D6,$34,$66
	.dc.b	$3E,$29,$F5,$0E,$E7,$C1,$9E,$F1,$C2,$BC,$57,$E8,$C3,$E5,$61,$D6
	.dc.b	$7E,$76,$1E,$46,$DF,$18,$E9,$7C,$5F,$84,$78,$0F,$91,$1E,$5C,$3C
	.dc.b	$35,$C8,$FB,$96,$EB,$FE,$41,$D7,$30,$FE,$39,$1F,$C8,$CC,$3F,$C2
	.dc.b	$23,$0F,$F2,$1E,$BF,$E5,$1F,$43,$23,$04,$3F,$87,$B9,$7F,$88,$C7
	.dc.b	$FB,$AF,$E7,$BF,$B0,$47,$F9,$D1,$AF,$F8,$E9,$AF,$F9,$95,$AB,$E8
	.dc.b	$1F,$43,$FE,$A2,$8F,$F2,$0C,$63,$C6,$FF,$0E,$3F,$E4,$3F,$8C,$B5
	.dc.b	$BA,$FF,$38,$7D,$6A,$FF,$C8,$11,$FE,$4A,$8C,$6D,$FE,$25,$6C,$7F
	.dc.b	$A1,$DA,$FF,$94,$3E,$45,$F1,$1F,$E9,$7A,$FE,$AB,$18,$F8,$4F,$F2
	.dc.b	$97,$1F,$F9,$4E,$37,$E9,$62,$BF,$D0,$51,$B7,$8D,$F3,$C3,$FB,$F4
	.dc.b	$E5,$B1,$9B,$E8,$B0,$E9,$2A,$F2,$FD,$0E,$1F,$3F,$D0,$DB,$4D,$8C
	.dc.b	$37,$1F,$1B,$6F,$FC,$0D,$F4,$0F,$A0,$EF,$C0,$FA,$17,$97,$E2,$B5
	.dc.b	$EF,$1F,$EF,$37,$BA,$D4,$E6,$BF,$D6,$8D,$7F,$D7,$2D,$7F,$CF,$ED
	.dc.b	$7F,$D5,$4F,$1D,$BF,$CF,$4D,$BF,$F5,$77,$C8,$3E,$09,$F0,$3E,$0D
	.dc.b	$3E,$17,$91,$F0,$DB,$7F,$49,$B9,$7D,$06,$E5,$F4,$1F,$95,$E9,$CA
	.dc.b	$FB,$3F,$43,$6F,$EB,$CD,$BF,$AF,$36,$BF,$D3,$3F,$41,$FE,$E5,$7F
	.dc.b	$C3,$B2,$FA,$1F,$BD,$1A,$ED,$FB,$B3,$9E,$8F,$9D,$7C,$72,$FF,$6E
	.dc.b	$35,$FF,$70,$B5,$E5,$7F,$E8,$18,$E1,$5F,$EA,$16,$C7,$C8,$DB,$FC
	.dc.b	$B8,$FF,$89,$63,$35,$F8,$DE,$67,$D0,$FE,$2C,$FC,$67,$AF,$6F,$F3
	.dc.b	$71,$6B,$F3,$33,$71,$FC,$CD,$7E,$4F,$E7,$FF,$28,$D3,$F4,$65,$AF
	.dc.b	$D0,$B6,$7D,$1E,$3F,$E4,$BE,$87,$F4,$53,$E7,$8C,$76,$FC,$67,$F7
	.dc.b	$F7,$FE,$F1,$D1,$18,$84,$62,$BE,$34,$E6,$31,$3F,$E6,$76,$E1,$FE
	.dc.b	$CF,$E9,$63,$F6,$2B,$B5,$FE,$8F,$CA,$FD,$E4,$7F,$B6,$0F,$F3,$77
	.dc.b	$07,$86,$33,$1F,$ED,$C6,$A1,$6F,$89,$F4,$37,$3F,$F5,$E3,$67,$B1
	.dc.b	$C2,$77,$E3,$FE,$D6,$7F,$A2,$57,$F9,$87,$E8,$1F,$0D,$9A,$FD,$D7
	.dc.b	$D0,$7F,$22,$DB,$FD,$CC,$DB,$FF,$78,$DF,$FB,$F9,$FF,$35,$46,$72
	.dc.b	$F4,$4F,$EC,$27,$B7,$A1,$FC,$16,$31,$3F,$E9,$17,$93,$7B,$74,$7F
	.dc.b	$0B,$FA,$07,$D0,$63,$0F,$F3,$C7,$8C,$F9,$0F,$E4,$30,$F9,$E7,$A1
	.dc.b	$FC,$4E,$B5,$3F,$E8,$A3,$51,$9A,$E3,$5C,$27,$AC,$62,$75,$18,$39
	.dc.b	$4B,$E8,$33,$9F,$FA,$AE,$7F,$D6,$87,$78,$F9,$AF,$FA,$3C,$CC,$7F
	.dc.b	$AE,$DE,$5C,$3F,$CA,$5C,$F0,$1F,$C1,$1A,$39,$7A,$DF,$02,$FA,$DF
	.dc.b	$86,$39,$1E,$CF,$95,$B5,$FE,$9F,$FA,$07,$C1,$FC,$03,$F9,$C3,$88
	.dc.b	$F6,$C6,$0D,$3C,$5B,$7F,$EC,$99,$E0,$D8,$F9,$7E,$C7,$1A,$FF,$57
	.dc.b	$7A,$3F,$B6,$AF,$81,$AB,$E7,$7F,$A4,$4E,$75,$F8,$15,$F0,$6B,$57
	.dc.b	$CB,$C0,$FA,$F8,$70,$6C,$64,$95,$DB,$AB,$51,$FE,$BF,$F3,$3F,$07
	.dc.b	$3F,$EE,$B9,$7C,$FB,$E8,$3F,$AD,$56,$B3,$FD,$F4,$E7,$0E,$5C,$BF
	.dc.b	$D8,$A3,$18,$FF,$BA,$EF,$85,$7E,$E7,$1E,$B8,$F4,$C2,$4F,$F6,$4E
	.dc.b	$FF,$EB,$18,$7F,$58,$9F,$0F,$C4,$FF,$40,$FA,$0E,$73,$FD,$72,$31
	.dc.b	$3A,$BF,$E9,$35,$AA,$64,$BD,$D7,$C1,$DB,$DB,$C3,$EE,$6C,$64,$96
	.dc.b	$DB,$F9,$4F,$D0,$7F,$13,$FA,$07,$D0,$7F,$15,$AE,$F8,$7F,$80,$5F
	.dc.b	$E3,$15,$B9,$D6,$7F,$6A,$57,$C9,$FE,$49,$5B,$7F,$BC,$DF,$E0,$A7
	.dc.b	$51,$DE,$74,$F8,$31,$AF,$80,$EB,$87,$82,$F9,$3F,$C9,$2B,$8E,$1E
	.dc.b	$3C,$3A,$DF,$17,$BC,$66,$B8,$4F,$43,$CB,$87,$81,$5F,$4C,$3B,$C7
	.dc.b	$B0,$E0,$79,$CA,$77,$AF,$35,$73,$AE,$4D,$9E,$37,$E3,$7A,$61,$E6
	.dc.b	$C3,$BC,$7C,$53,$6E,$1F,$88,$7F,$44,$3C,$E5,$29,$9B,$D7,$CF,$BA
	.dc.b	$7F,$CB,$F3,$FD,$32,$31,$E6,$EA,$73,$3F,$D7,$67,$2E,$FE,$0F,$F1
	.dc.b	$4D,$D7,$AC,$62,$38,$35,$FC,$64,$A3,$8E,$5F,$82,$FA,$07,$CD,$FE
	.dc.b	$46,$30,$DD,$F8,$D7,$06,$E3,$F8,$8B,$E3,$8D,$BE,$0F,$F6,$FC,$7C
	.dc.b	$15,$F8,$E6,$C7,$DA,$AE,$2D,$AE,$2A,$77,$8F,$E6,$1B,$87,$E9,$EF
	.dc.b	$8C,$FA,$2B,$B6,$3E,$77,$C4,$D5,$3C,$2B,$96,$1E,$0C,$3A,$D7,$0C
	.dc.b	$7D,$8C,$3F,$23,$1A,$7C,$D8,$75,$C3,$A4,$7F,$83,$DA,$EA,$B9,$1F
	.dc.b	$CD,$1F,$1D,$FF,$01,$8C,$4F,$C4,$75,$7D,$E3,$1B,$FA,$2F,$FC,$39
	.dc.b	$5C,$2F,$94,$FD,$69,$FD,$3F,$D0,$7F,$A7,$5D,$C7,$F0,$F8,$C4,$7F
	.dc.b	$15,$8C,$74,$F2,$8F,$F1,$47,$CF,$38,$F1,$0B,$E3,$8F,$AE,$BF,$DE
	.dc.b	$93,$C3,$CD,$DD,$6B,$1F,$07,$C5,$60,$95,$E4,$3B,$F8,$38,$7C,$87
	.dc.b	$F8,$E6,$93,$9C,$7C,$15,$6F,$F9,$56,$3F,$E1,$8C,$3A,$07,$0F,$43
	.dc.b	$D5,$7E,$CF,$A0,$FE,$7D,$1C,$1A,$ED,$B8,$7B,$FC,$DE,$1D,$32,$B8
	.dc.b	$3F,$1B,$04,$E6,$7C,$10,$FA,$0D,$19,$BF,$C2,$7C,$7B,$E8,$3E,$7B
	.dc.b	$6F,$D6,$D7,$9F,$76,$A6,$46,$9F,$14,$62,$3C,$3E,$8F,$94,$C5,$FE
	.dc.b	$4F,$D8,$FF,$2C,$CF,$C8,$6B,$F1,$DB,$1F,$99,$E5,$7A,$DF,$84,$9C
	.dc.b	$F0,$FF,$5B,$6D,$FF,$9C,$5A,$FF,$10,$9E,$1E,$5F,$E3,$2F,$FD,$C9
	.dc.b	$3F,$61,$AF,$86,$3F,$23,$E8,$7E,$0A,$3E,$EB,$5F,$B8,$DC,$7F,$5D
	.dc.b	$F4,$0F,$2B,$8F,$2C,$BE,$2F,$E7,$27,$51,$83,$F1,$2B,$E8,$34,$71
	.dc.b	$8F,$C5,$47,$B3,$E8,$3F,$4D,$1F,$90,$6B,$AC,$BF,$53,$18,$C7,$F8
	.dc.b	$13,$76,$FF,$01,$36,$BF,$D3,$17,$D7,$AF,$F6,$3F,$3E,$FA,$0A,$74
	.dc.b	$8F,$DA,$FD,$03,$DF,$DF,$8D,$31,$FD,$1C,$E5,$FD,$2E,$1F,$D5,$5F
	.dc.b	$D3,$FC,$FB,$DC,$FD,$FB,$63,$77,$8F,$A0,$91,$C1,$DD,$7F,$06,$FA
	.dc.b	$0F,$D3,$EF,$FE,$1D,$6D,$7F,$AC,$DA,$BE,$81,$F3,$5C,$7C,$29,$3F
	.dc.b	$BA,$AD,$D6,$A3,$13,$FC,$52,$1F,$41,$4E,$33,$FB,$FF,$88,$FB,$AD
	.dc.b	$E9,$FE,$37,$18,$E6,$FE,$68,$BE,$83,$33,$2B,$F8,$5C,$EF,$4F,$F1
	.dc.b	$C3,$B8,$CF,$9B,$D3,$82,$46,$08,$78,$DC,$BE,$58,$FE,$5B,$F0,$CF
	.dc.b	$83,$A4,$9F,$0F,$AF,$5F,$E3,$53,$B7,$F8,$FB,$5F,$81,$1B,$D3,$DB
	.dc.b	$F4,$1F,$00,$C6,$7F,$E5,$1F,$A0,$FC,$85,$9F,$F4,$EB,$FC,$A9,$AF
	.dc.b	$F2,$97,$64,$F7,$BF,$A2,$BA,$BF,$EB,$CE,$BC,$CC,$92,$BA,$7C,$C5
	.dc.b	$7A,$45,$AE,$33,$F1,$3E,$42,$57,$B9,$EF,$DB,$88,$F8,$71,$6B,$C3
	.dc.b	$39,$9F,$F3,$70,$D5,$5F,$E3,$96,$C7,$F8,$F5,$AF,$01,$F3,$CF,$37
	.dc.b	$FB,$1A,$F9,$FA,$62,$24,$FF,$52,$9D,$36,$72,$FF,$53,$8F,$CE,$37
	.dc.b	$AF,$A9,$DB,$C7,$B1,$CE,$2B,$C6,$3F,$BA,$FC,$F3,$E4,$A9,$27,$FB
	.dc.b	$1F,$CF,$7E,$65,$09,$E0,$FF,$68,$6B,$FD,$9D,$BB,$7F,$86,$67,$94
	.dc.b	$F1,$AE,$CE,$DF,$F0,$3C,$68,$FF,$29,$47,$8B,$9F,$26,$BA,$2C,$64
	.dc.b	$87,$F7,$79,$FE,$2C,$DE,$7E,$BD,$12,$3D,$B1,$E0,$30,$F1,$8C,$3F
	.dc.b	$E0,$4F,$9D,$7C,$F7,$F8,$71,$B1,$98,$FF,$0F,$FC,$57,$E9,$27,$33
	.dc.b	$FE,$09,$6E,$9F,$E3,$B6,$F5,$7F,$AE,$EB,$85,$F5,$25,$FF,$85,$23
	.dc.b	$2F,$72,$1F,$E1,$57,$2F,$78,$E2,$E7,$C8,$B8,$78,$66,$DF,$86,$6D
	.dc.b	$7D,$EC,$3C,$AD,$AF,$6C,$EA,$13,$A6,$CE,$9F,$2D,$F7,$6D,$F7,$69
	.dc.b	$D4,$A3,$0F,$5C,$F1,$8F,$61,$F0,$F0,$F1,$61,$E1,$3C,$55,$C2,$FF
	.dc.b	$DC,$D1,$9C,$3C,$97,$C8,$98,$75,$3F,$CD,$D7,$C6,$3C,$86,$AF,$D0
	.dc.b	$D3,$33,$C6,$32,$D8,$CC,$78,$AB,$C3,$39,$9F,$8E,$7C,$0A,$F0,$38
	.dc.b	$3B,$1F,$0F,$0F,$7D,$7D,$19,$DE,$1E,$DA,$F8,$87,$CA,$70,$FE,$04
	.dc.b	$C4,$F1,$AE,$71,$E1,$6B,$E0,$AE,$B3,$B8,$E2,$6C,$A6,$E7,$8D,$FB
	.dc.b	$6B,$8C,$EC,$F5,$DF,$92,$7E,$7D,$70,$C3,$E3,$57,$A8,$DE,$1E,$16
	.dc.b	$9C,$EB,$E5,$5F,$3A,$FA,$27,$29,$D9,$F0,$4B,$3E,$5B,$F0,$47,$D6
	.dc.b	$3F,$8E,$E1,$E8,$7D,$87,$BA,$78,$36,$BB,$47,$AA,$30,$73,$3E,$C1
	.dc.b	$83,$EE,$3B,$7C,$71,$C8,$F4,$BC,$6F,$D8,$F9,$2F,$CE,$F2,$9D,$B4
	.dc.b	$E1,$87,$E3,$A3,$A4,$60,$F9,$97,$F2,$27,$E6,$36,$3E,$64,$FF,$13
	.dc.b	$E9,$FE,$37,$D3,$27,$F8,$94,$7F,$1E,$30,$FF,$11,$8C,$3F,$C5,$FA
	.dc.b	$7F,$91,$FC,$F4,$8C,$10,$FE,$30,$E5,$FE,$35,$1F,$EC,$8F,$9E,$FE
	.dc.b	$73,$1F,$DC,$9A,$FF,$52,$0E,$1F,$EA,$AD,$5F,$5A,$F8,$DF,$EE,$91
	.dc.b	$FD,$6A,$30,$E1,$AF,$F7,$96,$BF,$E0,$2F,$9F,$7C,$FF,$F9,$12,$3C
	.dc.b	$51,$8F,$13,$AB,$FE,$07,$F8,$CB,$5B,$A7,$C8,$3E,$23,$FC,$D1,$C1
	.dc.b	$FE,$DB,$8C,$7C,$22,$31,$3A,$AF,$F6,$ED,$6A,$73,$E7,$6E,$73,$4B
	.dc.b	$C7,$1E,$70,$F0,$5E,$30,$FE,$F7,$1C,$F0,$E3,$F3,$B9,$7C,$F8,$BE
	.dc.b	$7B,$15,$E9,$AB,$39,$BF,$3C,$62,$B2,$D8,$FF,$76,$FC,$EB,$E7,$7B
	.dc.b	$6C,$F9,$EF,$F1,$6D,$6A,$73,$18,$CF,$FD,$E5,$7C,$6B,$FC,$27,$AF
	.dc.b	$C1,$C3,$97,$95,$E9,$B5,$ED,$56,$F0,$E8,$FC,$98,$DE,$3C,$6B,$84
	.dc.b	$FC,$DF,$90,$7C,$0B,$DF,$D5,$FC,$2F,$1B,$E1,$B7,$5F,$B2,$DC,$BE
	.dc.b	$AB,$71,$FA,$9C,$1D,$AB,$E9,$AF,$47,$E6,$9F,$3B,$E7,$AC,$FE,$12
	.dc.b	$2E,$1F,$80,$9B,$7E,$BF,$9C,$FF,$5E,$5F,$C6,$3E,$3F,$CF,$7E,$2E
	.dc.b	$75,$DD,$F7,$D4,$66,$7F,$5D,$5C,$B0,$F9,$31,$CB,$9F,$E7,$4E,$6F
	.dc.b	$F1,$F7,$C6,$FF,$C8,$F3,$9B,$E7,$87,$E3,$78,$3E,$5F,$D0,$72,$8F
	.dc.b	$C5,$FC,$FC,$AF,$D0,$FA,$77,$1F,$CE,$CD,$CF,$E4,$CD,$AF,$92,$7E
	.dc.b	$53,$E1,$BF,$2A,$D3,$F5,$65,$BF,$D3,$B6,$BE,$47,$C7,$FE,$53,$E7
	.dc.b	$FF,$55,$3F,$26,$BD,$73,$9E,$EF,$82,$D9,$D4,$B6,$4E,$63,$E4,$9C
	.dc.b	$A7,$73,$A7,$B7,$5B,$0F,$95,$5C,$EF,$F3,$9E,$6E,$3E,$E3,$FC,$F1
	.dc.b	$87,$82,$31,$D1,$FA,$1C,$5F,$15,$FA,$3F,$A3,$6F,$98,$C3,$BF,$2E
	.dc.b	$63,$EA,$FA,$07,$C5,$63,$6F,$48,$FA,$17,$93,$FA,$F6,$DF,$1F,$4F
	.dc.b	$13,$FB,$3C,$3F,$DA,$D3,$93,$B7,$83,$9A,$D3,$23,$E5,$FD,$0B,$E8
	.dc.b	$71,$97,$8B,$CA,$7C,$87,$F1,$7B,$7D,$07,$F9,$25,$FA,$73,$A7,$ED
	.dc.b	$1A,$F5,$FF,$83,$46,$6F,$EA,$CE,$4E,$37,$D6,$FA,$4F,$86,$38,$99
	.dc.b	$F4,$6D,$7E,$08,$7B,$3E,$58,$F9,$71,$98,$C7,$0F,$F1,$78,$D9,$E6
	.dc.b	$9D,$DF,$D0,$8C,$BE,$68,$CF,$77,$F1,$99,$CB,$6F,$D1,$F2,$DD,$5F
	.dc.b	$E4,$9F,$42,$F8,$4F,$61,$82,$DF,$E6,$A3,$4F,$C8,$D2,$E5,$F8,$29
	.dc.b	$B8,$FF,$51,$AF,$A3,$7C,$DE,$4F,$C9,$F6,$B1,$7F,$8A,$E5,$FD,$6B
	.dc.b	$F4,$31,$7C,$F7,$F3,$99,$CE,$3D,$67,$38,$FE,$BA,$72,$FF,$B5,$0F
	.dc.b	$C0,$D7,$1F,$0F,$EC,$36,$38,$47,$DB,$ED,$C6,$3F,$9C,$F2,$3F,$05
	.dc.b	$97,$F4,$BF,$A1,$7D,$0F,$F4,$B9,$CE,$3F,$EE,$2C,$15,$A8,$FE,$03
	.dc.b	$3A,$C1,$5A,$BE,$5C,$9F,$D8,$30,$E1,$5E,$18,$CE,$1F,$C0,$BC,$2F
	.dc.b	$E8,$93,$FE,$F0,$C7,$A2,$5F,$19,$FC,$E7,$E8,$73,$18,$BF,$E3,$1A
	.dc.b	$27,$35,$BC,$7D,$8D,$AF,$4E,$DE,$FE,$1F,$80,$D8,$C9,$32,$FE,$06
	.dc.b	$07,$D0,$DF,$A1,$7D,$0F,$F8,$DF,$5F,$B3,$5C,$0D,$89,$DB,$EF,$C3
	.dc.b	$FC,$39,$18,$8F,$F2,$05,$F5,$BE,$95,$A7,$FC,$89,$3E,$39,$D3,$CF
	.dc.b	$0A,$F2,$3C,$13,$CA,$33,$1C,$4C,$9C,$4A,$71,$BF,$26,$1E,$C7,$CD
	.dc.b	$1C,$5C,$E1,$E5,$8F,$2E,$1E,$2A,$E1,$87,$A2,$7E,$1D,$F0,$C3,$C7
	.dc.b	$18,$9F,$33,$E2,$3B,$38,$9F,$0C,$6C,$F4,$38,$BF,$06,$1E,$39,$F5
	.dc.b	$57,$13,$C5,$3F,$E6,$68,$F8,$47,$F9,$A1,$C3,$F6,$0F,$9F,$F4,$3F
	.dc.b	$C7,$BD,$5F,$F3,$6B,$63,$D8,$D7,$7D,$3F,$E6,$17,$3F,$0D,$FE,$34
	.dc.b	$6F,$0F,$28,$C4,$76,$6B,$FE,$81,$A4,$71,$CF,$F0,$DF,$42,$F9,$CF
	.dc.b	$F4,$34,$61,$BA,$71,$AE,$0D,$C7,$F0,$B7,$C7,$1B,$7C,$1F,$CF,$63
	.dc.b	$E0,$AF,$C5,$B6,$3F,$0B,$5C,$5B,$5C,$54,$EF,$1F,$CE,$37,$0F,$F4
	.dc.b	$E5,$F1,$9F,$45,$76,$C7,$CE,$F8,$9A,$A7,$85,$72,$C3,$C1,$87,$5A
	.dc.b	$E1,$8F,$B1,$87,$E7,$23,$4F,$9B,$0E,$B8,$74,$8F,$DF,$35,$D5,$72
	.dc.b	$3F,$CD,$51,$F0,$5B,$7F,$EB,$77,$FC,$DD,$7D,$A7,$5B,$79,$EF,$FC
	.dc.b	$53,$18,$7C,$E7,$EC,$3E,$87,$FD,$38,$FF,$8B,$9A,$FF,$85,$1B,$7F
	.dc.b	$E7,$BF,$A0,$71,$78,$70,$FF,$64,$56,$9F,$F6,$54,$FF,$8D,$AF,$FC
	.dc.b	$72,$D7,$A7,$57,$17,$F8,$F5,$3E,$33,$10,$8C,$3E,$58,$CD,$3C,$0D
	.dc.b	$D6,$AB,$D1,$58,$37,$36,$3C,$7E,$27,$23,$0F,$F8,$CB,$E8,$7F,$A0
	.dc.b	$BF,$E5,$06,$B9,$E8,$F0,$F9,$9E,$69,$CE,$DF,$E9,$F6,$EB,$FE,$EB
	.dc.b	$6D,$FF,$3B,$D1,$7C,$70,$F5,$FD,$0C,$AF,$04,$7C,$2F,$92,$7D,$0D
	.dc.b	$24,$FC,$CE,$DA,$FF,$92,$A7,$6F,$F9,$A1,$AF,$F9,$3B,$1F,$F1,$EB
	.dc.b	$9A,$F9,$7E,$8C,$8C,$1F,$E7,$0F,$81,$FF,$2B,$B7,$CA,$FA,$B1,$88
	.dc.b	$FF,$2C,$F4,$7F,$98,$9D,$10,$FF,$2E,$C6,$F8,$3F,$D8,$93,$B9,$CD
	.dc.b	$7E,$02,$3F,$DB,$51,$FE,$80,$6B,$A3,$FD,$07,$E7,$7D,$FA,$FF,$67
	.dc.b	$37,$87,$FB,$44,$FF,$B2,$F9,$3F,$17,$F3,$FF,$E6,$C8,$E3,$1F,$8E
	.dc.b	$AE,$11,$FD,$BD,$AA,$78,$3F,$E7,$68,$CC,$3E,$87,$A9,$CE,$BB,$E5
	.dc.b	$F9,$AF,$A1,$FD,$3E,$1F,$E9,$A6,$BF,$E8,$0B,$F8,$98,$FA,$27,$39
	.dc.b	$7F,$BA,$5B,$19,$EA,$EB,$7C,$DF,$D3,$FC,$FB,$E8,$69,$E1,$8F,$D5
	.dc.b	$FD,$0B,$E0,$77,$E2,$4C,$3F,$D5,$B3,$97,$FD,$5F,$87,$F9,$B9,$FF
	.dc.b	$58,$FD,$0B,$E1,$7F,$6A,$D8,$DD,$E3,$E8,$64,$70,$77,$5F,$BC,$FA
	.dc.b	$07,$55,$F9,$C9,$7F,$EB,$E9,$CB,$67,$2F,$FA,$9A,$13,$CE,$64,$F1
	.dc.b	$8D,$F0,$B8,$3F,$07,$7D,$2B,$FD,$5F,$F4,$3F,$BC,$7F,$C3,$AD,$7F
	.dc.b	$12,$D7,$FC,$64,$D4,$FE,$F1,$43,$E8,$7D,$C7,$F1,$A9,$D7,$C6,$7E
	.dc.b	$2B,$0F,$25,$FF,$8F,$63,$18,$7E,$61,$B1,$9F,$A1,$FE,$47,$18,$83
	.dc.b	$49,$1A,$F2,$F7,$BC,$AF,$F7,$4D,$7F,$7C,$6C,$E5,$FF,$68,$C7,$DF
	.dc.b	$6C,$73,$78,$CE,$F0,$F4,$C7,$8A,$7F,$9B,$FC,$4B,$E0,$3F,$C3,$CD
	.dc.b	$9F,$F1,$90,$70,$FF,$8D,$5A,$FF,$93,$9C,$D7,$6C,$1F,$43,$F0,$8E
	.dc.b	$F7,$FD,$1B,$E8,$7F,$4B,$87,$F9,$9D,$FF,$59,$3F,$E6,$9B,$F3,$C6
	.dc.b	$23,$FD,$DE,$F0,$4D,$35,$1C,$3D,$DE,$DD,$3D,$FC,$99,$7D,$77,$FE
	.dc.b	$1F,$87,$F3,$C8,$C1,$E3,$F9,$BD,$BE,$0A,$B8,$F9,$26,$DF,$29,$B5
	.dc.b	$C4,$E6,$7C,$9F,$94,$F9,$A4,$7E,$6E,$C8,$C4,$7B,$BE,$8F,$E3,$1F
	.dc.b	$5A,$3F,$B6,$FD,$17,$E8,$F2,$FB,$6F,$EB,$2B,$BB,$54,$6B,$A3,$89
	.dc.b	$CB,$A2,$4F,$08,$EB,$F4,$52,$39,$BE,$0F,$13,$11,$D3,$5F,$E6,$F8
	.dc.b	$AB,$38,$F9,$2E,$D6,$A6,$DF,$FA,$96,$FF,$9B,$FD,$13,$E3,$7F,$C1
	.dc.b	$3F,$44,$F9,$4A,$7F,$83,$27,$57,$F3,$1A,$F8,$CF,$F5,$84,$F1,$7E
	.dc.b	$8B,$5D,$08,$E3,$9F,$E9,$27,$F5,$D3,$A6,$C6,$7A,$7C,$BD,$5E,$66
	.dc.b	$57,$67,$2F,$83,$E0,$7D,$7D,$32,$30,$43,$B3,$97,$C1,$1F,$08,$C5
	.dc.b	$F9,$67,$13,$C6,$30,$FB,$A3,$2F,$23,$04,$30,$48,$E2,$F8,$8E,$9B
	.dc.b	$7F,$72,$C5,$7E,$28,$C5,$F6,$BF,$91,$39,$BF,$A3,$39,$AE,$53,$99
	.dc.b	$5F,$F0,$69,$CC,$78,$0E,$27,$9E,$FA,$B9,$38,$DF,$3A,$F2,$C6,$5E
	.dc.b	$8F,$23,$DD,$5D,$9F,$71,$2B,$AB,$A3,$83,$F1,$0C,$92,$3E,$0C,$79
	.dc.b	$5E,$91,$98,$E3,$1A,$7A,$1F,$02,$FD,$17,$F1,$63,$F2,$F1,$87,$D4
	.dc.b	$D7,$8D,$FB,$9F,$45,$75,$BF,$04,$75,$3C,$87,$88,$EA,$7C,$5B,$F9
	.dc.b	$17,$E1,$8F,$96,$70,$39,$1E,$08,$DB,$C4,$D5,$F8,$EB,$C2,$FD,$08
	.dc.b	$E6,$7A,$2B,$D5,$3A,$6D,$FC,$82,$4F,$9C,$E3,$3B,$7B,$DF,$9E,$FC
	.dc.b	$15,$F0,$49,$3E,$E3,$D0,$79,$AF,$85,$7A,$A3,$27,$CA,$38,$5F,$AA
	.dc.b	$34,$D8,$E4,$7C,$13,$81,$F1,$9D,$38,$BF,$2B,$D2,$FC,$AF,$B1,$F7
	.dc.b	$BC,$5C,$C7,$C9,$38,$C6,$4F,$88,$E6,$FD,$71,$E4,$3D,$75,$F9,$7A
	.dc.b	$EB,$5C,$8E,$97,$F2,$6F,$E8,$DF,$94,$87,$94,$EE,$F6,$7D,$8D,$BE
	.dc.b	$D1,$C2,$BB,$57,$C8,$BF,$0B,$64,$3E,$89,$F4,$5F,$EC,$6F,$5F,$FD
	.dc.b	$99,$8F,$F8,$51,$B8,$7F,$A0,$AB,$77,$FD,$A9,$AF,$F3,$A9,$DB,$D9
	.dc.b	$AF,$7B,$ED,$D1,$DF,$A6,$4F,$F7,$4A,$FF,$00,$3C,$E7,$2F,$59,$E2
	.dc.b	$EF,$C8,$F4,$7D,$14,$8C,$10,$F2,$39,$7C,$D1,$EF,$31,$9F,$BE,$71
	.dc.b	$3F,$E0,$6F,$A0,$7D,$07,$FA,$62,$3F,$D4,$71,$8F,$8A,$FF,$19,$BF
	.dc.b	$E3,$3F,$8A,$B5,$BA,$F9,$FF,$46,$BF,$F1,$FC,$7F,$AB,$23,$1C,$1F
	.dc.b	$E3,$C6,$C7,$FA,$C0,$E3,$3C,$17,$C7,$BE,$1B,$FD,$15,$7F,$E2,$68
	.dc.b	$C7,$C1,$7F,$99,$35,$FF,$32,$C6,$FC,$DC,$57,$E0,$A3,$6F,$1B,$E7
	.dc.b	$87,$F7,$99,$CB,$63,$37,$D1,$61,$D2,$7C,$B0,$FA,$0C,$5F,$9F,$E8
	.dc.b	$6C,$E7,$1F,$04,$62,$B2,$D8,$C9,$E5,$8C,$B4,$CD,$FF,$82,$BE,$81
	.dc.b	$F4,$1F,$1E,$11,$89,$FF,$34,$B6,$3E,$71,$A9,$E1,$19,$CF,$FC,$EA
	.dc.b	$DA,$FD,$4F,$C6,$BD,$E7,$BB,$F0,$71,$E5,$5F,$2F,$C3,$F7,$E5,$F0
	.dc.b	$1B,$19,$3E,$84,$65,$A7,$D3,$F4,$3E,$39,$F2,$3E,$82,$C6,$49,$7F
	.dc.b	$56,$32,$FD,$79,$BA,$7D,$79,$B5,$FE,$FC,$F9,$EF,$F6,$D3,$FE,$8C
	.dc.b	$6D,$FC,$46,$BF,$93,$6B,$C7,$AB,$E0,$47,$0B,$FF,$59,$EB,$F1,$7C
	.dc.b	$6C,$D7,$E0,$3E,$7E,$63,$F8,$FF,$3E,$E5,$F8,$D9,$B8,$FE,$36,$6D
	.dc.b	$7E,$5F,$93,$FD,$A4,$FE,$5F,$E6,$FF,$16,$FE,$31,$AE,$4F,$CE,$19
	.dc.b	$29,$9C,$3F,$38,$E5,$FF,$29,$F8,$DF,$9E,$8F,$E4,$F0,$F9,$DF,$F3
	.dc.b	$24,$7F,$A2,$A7,$53,$FC,$B2,$72,$FF,$97,$3B,$9F,$09,$F8,$5F,$9D
	.dc.b	$63,$F8,$79,$CE,$3F,$E9,$49,$CB,$FE,$96,$87,$89,$B3,$F1,$DF,$38
	.dc.b	$F8,$4C,$57,$FA,$73,$E7,$7F,$84,$F0,$7F,$B3,$1B,$1E,$AA,$FF,$66
	.dc.b	$3B,$F0,$B9,$F9,$72,$3F,$D4,$B3,$F9,$C6,$C7,$83,$B2,$47,$FA,$AB
	.dc.b	$E7,$5F,$3B,$8E,$1F,$F3,$DF,$86,$F8,$CF,$E0,$96,$F9,$B7,$9B,$FA
	.dc.b	$7A,$D4,$FF,$AF,$8D,$46,$6B,$8D,$70,$9F,$F5,$CD,$70,$BF,$F4,$74
	.dc.b	$70,$33,$5C,$5E,$94,$F9,$DC,$57,$FA,$42,$5E,$9F,$9A,$7E,$64,$7F
	.dc.b	$B1,$8C,$9E,$57,$33,$C6,$78,$71,$7F,$01,$AD,$CF,$92,$B7,$C5,$FE
	.dc.b	$FE,$9C,$B6,$FD,$BF,$25,$C5,$FE,$15,$F3,$AF,$80,$F8,$26,$0B,$7F
	.dc.b	$ED,$A8,$D3,$EE,$8D,$34,$D3,$8C,$7E,$3E,$6F,$80,$C6,$4B,$FC,$FF
	.dc.b	$2F,$F3,$C7,$E4,$FC,$D7,$F1,$D8,$C6,$1C,$A3,$1E,$77,$D6,$6D,$70
	.dc.b	$3F,$DD,$F5,$C2,$FB,$5F,$F8,$4A,$31,$3F,$E1,$43,$F0,$DD,$5C,$A7
	.dc.b	$FD,$8B,$E2,$FD,$FB,$E3,$2F,$9C,$7C,$EF,$F3,$1A,$D6,$1F,$E1,$EA
	.dc.b	$D4,$E7,$93,$FA,$B4,$62,$7F,$C4,$78,$79,$31,$61,$C9,$56,$DF,$1D
	.dc.b	$F3,$BE,$27,$7F,$0B,$FA,$34,$FF,$46,$9F,$27,$CA,$7F,$65,$C5,$F3
	.dc.b	$BD,$6B,$BC,$E6,$31,$92,$31,$7C,$B1,$EF,$6C,$39,$CF,$C1,$DB,$E1
	.dc.b	$70,$5B,$DD,$FB,$23,$0F,$CC,$9F,$EB,$0F,$F9,$13,$E7,$7F,$C8,$BF
	.dc.b	$3A,$F9,$DF,$E4,$3A,$2F,$76,$BD,$D6,$A7,$EE,$DF,$1A,$FF,$0D,$56
	.dc.b	$9F,$ED,$33,$AA,$FE,$E3,$39,$9F,$A3,$39,$78,$8A,$E8,$EB,$1E,$0F
	.dc.b	$A4,$D5,$F8,$E3,$31,$E7,$32,$78,$9C,$DF,$87,$0F,$74,$62,$76,$77
	.dc.b	$29,$DE,$FC,$73,$C2,$32,$77,$8C,$E1,$EB,$8F,$E9,$77,$DE,$31,$1C
	.dc.b	$F0,$F4,$3C,$A7,$58,$78,$EF,$9B,$E5,$9F,$F0,$C9,$ED,$2C,$E1,$F8
	.dc.b	$87,$F9,$2F,$E7,$7F,$8F,$F9,$3C,$1B,$1F,$E2,$26,$BB,$E9,$FF,$27
	.dc.b	$39,$F8,$0F,$C8,$37,$7E,$51,$88,$EC,$D7,$FB,$4D,$23,$8E,$5F,$6B
	.dc.b	$E7,$5F,$31,$FD,$BA,$30,$DD,$78,$D7,$06,$E3,$F8,$7B,$E3,$8D,$BE
	.dc.b	$0F,$FA,$7B,$1F,$05,$7E,$35,$B1,$F7,$2B,$8B,$6B,$8A,$9D,$E3,$F9
	.dc.b	$06,$E1,$FE,$04,$BE,$33,$E8,$AE,$D8,$F9,$DF,$13,$54,$F0,$AE,$58
	.dc.b	$78,$30,$EB,$5C,$31,$F6,$30,$FC,$9C,$69,$F3,$61,$D7,$0E,$91,$FE
	.dc.b	$BE,$6B,$AA,$E4,$7F,$1F,$8F,$82,$DB,$FF,$0F,$BF,$C8,$EF,$B4,$EB
	.dc.b	$5F,$3D,$FF,$0C,$8C,$3E,$73,$F4,$FF,$3B,$FE,$84,$7F,$88,$B5,$FE
	.dc.b	$31,$3F,$E2,$C8,$C5,$FF,$31,$F9,$BF,$EA,$95,$C3,$AF,$C7,$7F,$E3
	.dc.b	$29,$FF,$65,$9F,$EC,$CB,$E3,$1C,$FE,$23,$77,$FE,$DE,$94,$62,$11
	.dc.b	$87,$F9,$2D,$3C,$0D,$4E,$67,$CE,$73,$AC,$9C,$EB,$84,$7B,$A3,$1E
	.dc.b	$24,$8C,$1F,$CC,$BE,$77,$F9,$E3,$7A,$7C,$91,$88,$EE,$2F,$FD,$C7
	.dc.b	$A7,$F5,$59,$D3,$6B,$89,$E3,$F3,$2D,$FB,$3E,$77,$CC,$61,$EB,$F1
	.dc.b	$6F,$9D,$F8,$8D,$8F,$E8,$ED,$7C,$79,$7F,$8E,$DC,$CF,$02,$3F,$2B
	.dc.b	$B6,$BC,$3D,$2F,$8C,$63,$3F,$ED,$DE,$EF,$F9,$66,$7F,$AA,$B5,$F8
	.dc.b	$2D,$8F,$EE,$9C,$DF,$E6,$89,$DF,$95,$F6,$EB,$5C,$9F,$E7,$66,$C8
	.dc.b	$FE,$1B,$E9,$EE,$BF,$05,$7F,$EC,$69,$FE,$CA,$D7,$C3,$1F,$87,$F3
	.dc.b	$FF,$05,$8F,$F8,$15,$B8,$FF,$A1,$1B,$5D,$FE,$73,$FD,$FC,$FF,$70
	.dc.b	$8C,$BF,$EC,$6E,$5E,$87,$E4,$E7,$71,$F8,$09,$DC,$3E,$77,$47,$09
	.dc.b	$FE,$F7,$1E,$E5,$F3,$BF,$A8,$7F,$12,$DC,$7F,$C3,$93,$B8,$C6,$3F
	.dc.b	$E9,$C6,$D7,$C7,$6A,$9E,$1E,$2F,$EA,$7E,$6D,$F3,$B4,$FD,$65,$F0
	.dc.b	$F9,$D7,$BB,$BF,$16,$65,$FE,$12,$9C,$BF,$E1,$4C,$3F,$CE,$2F,$F8
	.dc.b	$5B,$E7,$5F,$11,$FB,$76,$C6,$EF,$1F,$3B,$23,$83,$BA,$FD,$F7,$CE
	.dc.b	$BB,$2F,$CB,$4B,$FF,$11,$CE,$5B,$39,$7F,$63,$09,$E7,$32,$78,$C6
	.dc.b	$F8,$BF,$C9,$15,$CE,$7F,$73,$F3,$9F,$B6,$E3,$FF,$64,$B6,$3F,$28
	.dc.b	$D5,$2F,$97,$FE,$2D,$39,$87,$CE,$FB,$0C,$F8,$7F,$C6,$FE,$27,$F1
	.dc.b	$39,$7F,$96,$5B,$97,$FB,$61,$B9,$7E,$8D,$7C,$EF,$1F,$66,$4F,$F9
	.dc.b	$1A,$76,$9D,$B9,$F3,$BC,$1C,$52,$30,$43,$F8,$CC,$FD,$C6,$C7,$9E
	.dc.b	$64,$77,$8E,$4F,$08,$FE,$6D,$F0,$CF,$82,$A4,$9F,$37,$B1,$5E,$FF
	.dc.b	$3B,$F0,$CC,$6B,$FE,$7B,$F9,$DF,$CD,$9F,$EA,$B9,$FF,$31,$B5,$FE
	.dc.b	$5E,$EB,$BB,$FA,$33,$99,$E5,$EF,$E2,$73,$E6,$CB,$E6,$F0,$F9,$6C
	.dc.b	$47,$48,$B7,$E3,$9B,$C5,$FC,$56,$98,$FF,$3E,$8C,$36,$3F,$8D,$29
	.dc.b	$F8,$BF,$39,$27,$FC,$FB,$B2,$3D,$3E,$06,$67,$B1,$FC,$8E,$74,$FF
	.dc.b	$92,$1A,$E8,$38,$78,$FC,$E7,$E7,$72,$FF,$53,$3F,$A5,$8F,$F4,$6B
	.dc.b	$5F,$EB,$2E,$F2,$F7,$36,$3F,$9A,$3B,$9E,$7D,$7F,$DD,$63,$52,$7F
	.dc.b	$48,$F6,$EB,$D1,$BF,$1E,$C6,$48,$78,$9C,$BE,$48,$B7,$EA,$9F,$F0
	.dc.b	$27,$CE,$3E,$4A,$92,$FF,$D5,$3F,$39,$FA,$49,$D4,$7F,$9C,$1A,$FC
	.dc.b	$96,$BF,$41,$AF,$DF,$8F,$F3,$13,$D3,$9F,$FC,$29,$18,$8F,$F0,$0B
	.dc.b	$FA,$9C,$3F,$DA,$8E,$5B,$3C,$0E,$3D,$1C,$B9,$F9,$F4,$4C,$3D,$31
	.dc.b	$97,$D7,$19,$7B,$C6,$5F,$0C,$66,$FC,$33,$97,$C7,$19,$DF,$D9,$39
	.dc.b	$7D,$D1,$9A,$FE,$E5,$8A,$B5,$18,$9D,$29,$D3,$E8,$3C,$E4,$3C,$E6
	.dc.b	$4F,$39,$9B,$F3,$B9,$7C,$EE,$6F,$CE,$E5,$F3,$B9,$3C,$F7,$E6,$7C
	.dc.b	$97,$E6,$BF,$2C,$78,$8F,$09,$2F,$C2,$E5,$F0,$B9,$BF,$0B,$97,$C2
	.dc.b	$E4,$F0,$9E,$53,$E3,$1E,$53,$E3,$5F,$96,$BF,$93,$46,$5F,$2D,$78
	.dc.b	$8F,$09,$2B,$C2,$E5,$F0,$B9,$9F,$09,$E5,$3C,$27,$94,$F0,$9E,$53
	.dc.b	$E2,$9E,$1A,$F2,$5F,$C7,$6C,$64,$E1,$7E,$87,$A3,$E8,$BF,$29,$E4
	.dc.b	$24,$79,$2F,$D0,$F9,$1F,$43,$E4,$3D,$07,$90,$EB,$5E,$26,$9E,$2B
	.dc.b	$F3,$34,$F3,$1C,$AF,$CD,$5D,$9F,$34,$78,$EF,$CA,$F7,$9F,$2B,$E3
	.dc.b	$78,$57,$D7,$BF,$C5,$36,$7E,$C3,$5F,$86,$D9,$E3,$1F,$22,$78,$B4
	.dc.b	$E2,$53,$89,$63,$8B,$EB,$8E,$2D,$38,$94,$E2,$58,$E2,$7B,$2F,$E0
	.dc.b	$C7,$BE,$FE,$0C,$60,$97,$F3,$27,$E5,$36,$B9,$36,$BE,$FC,$FD,$B7
	.dc.b	$9D,$FD,$7A,$FA,$0F,$FA,$E7,$E7,$1F,$39,$49,$7F,$E7,$FA,$5F,$F6
	.dc.b	$DF,$5F,$FC,$B7,$7C,$2B,$53,$FD,$C6,$31,$1F,$EB,$06,$D7,$28,$FF
	.dc.b	$45,$57,$F9,$90,$FE,$E8,$9F,$EE,$8F,$43,$7C,$DF,$EC,$F8,$D9,$1C
	.dc.b	$19,$E8,$D7,$36,$FE,$72,$5F,$33,$B4,$69,$C5,$A7,$71,$FE,$C8,$8F
	.dc.b	$F3,$6B,$67,$38,$7A,$E3,$D1,$5C,$DF,$F6,$7F,$D1,$3E,$2F,$FC,$05
	.dc.b	$1F,$EA,$98,$C3,$86,$BF,$E0,$76,$BF,$E0,$EF,$9D,$7C,$EF,$FA,$72
	.dc.b	$3F,$C3,$D1,$8F,$0B,$AB,$FE,$0E,$F8,$9B,$5B,$A7,$C7,$3E,$1B,$FD
	.dc.b	$07,$C5,$F3,$63,$1F,$05,$F5,$76,$FA,$D1,$BF,$37,$15,$FE,$C1,$8D
	.dc.b	$BC,$6F,$9E,$1F,$EF,$C9,$CB,$63,$37,$D1,$61,$D2,$7C,$B0,$F9,$BC
	.dc.b	$5F,$9F,$E7,$2C,$E7,$1F,$04,$62,$B2,$D8,$C9,$E5,$8C,$B4,$CD,$B1
	.dc.b	$FB,$5F,$38,$F9,$CF,$26,$CF,$A2,$FF,$21,$61,$F8,$9A,$D3,$F9,$59
	.dc.b	$D7,$8F,$EB,$C5,$5A,$7B,$57,$F9,$56,$31,$8F,$E8,$FE,$39,$F0,$0F
	.dc.b	$77,$95,$6A,$FE,$86,$FF,$E2,$B7,$E8,$6B,$F1,$3C,$5F,$8A,$D8,$C9
	.dc.b	$F5,$63,$2F,$D7,$F3,$FE,$59,$F3,$3E,$76,$C6,$49,$AF,$D9,$8C,$BF
	.dc.b	$6E,$6E,$BF,$6E,$6D,$7F,$9C,$7E,$77,$FD,$AE,$FE,$E6,$5F,$45,$F8
	.dc.b	$28,$D7,$77,$E0,$67,$3D,$1F,$26,$F8,$E5,$F2,$00,$73,$3F,$27,$81
	.dc.b	$F3,$DC,$70,$FC,$3F,$CF,$4C,$BF,$2D,$E8,$5C,$BF,$29,$37,$1F,$CA
	.dc.b	$4D,$AF,$9E,$7E,$43,$E1,$7F,$22,$D3,$F4,$05,$BF,$CF,$36,$BF,$3C
	.dc.b	$D3,$99,$FA,$79,$FC,$37,$91,$F0,$EB,$DB,$F3,$DF,$75,$F8,$07,$F8
	.dc.b	$F7,$6F,$C6,$46,$70,$FE,$21,$D2,$F8,$7F,$C8,$7C,$F3,$2F,$C9,$4E
	.dc.b	$73,$FE,$41,$39,$7F,$91,$43,$C8,$D9,$F9,$9F,$9D,$7C,$46,$2B,$F9
	.dc.b	$47,$CF,$3C,$7F,$CF,$36,$3F,$96,$57,$FB,$35,$DC,$F3,$F3,$E4,$6B
	.dc.b	$BE,$F0,$FB,$AF,$F4,$13,$93,$F9,$B7,$CF,$3E,$7B,$C9,$7F,$AD,$F1
	.dc.b	$DF,$1B,$FC,$26,$DF,$3B,$FD,$25,$FE,$7D,$3A,$7F,$A5,$1A,$F4,$39
	.dc.b	$57,$1A,$E1,$9F,$D0,$BE,$0F,$E4,$6F,$81,$D2,$B8,$BD,$2B,$F9,$F3
	.dc.b	$93,$FA,$70,$F5,$7C,$DE,$87,$CC,$8C,$C6,$38,$7F,$84,$46,$CF,$3C
	.dc.b	$EF,$0F,$C1,$C6,$5F,$24,$67,$9F,$FA,$94,$E5,$B7,$ED,$F9,$4E,$AF
	.dc.b	$F1,$5F,$9E,$7C,$07,$C1,$30,$5B,$FE,$CD,$1A,$7D,$D1,$A6,$9A,$71
	.dc.b	$8F,$C7,$CD,$F0,$18,$CB,$FB,$37,$2F,$F4,$0C,$3E,$47,$CD,$FF,$27
	.dc.b	$8C,$61,$C6,$31,$E9,$7D,$56,$D7,$03,$FB,$D5,$70,$BE,$B5,$F6,$FA
	.dc.b	$B7,$87,$EA,$BC,$EF,$B7,$5F,$E1,$19,$7C,$F3,$E7,$BF,$9C,$56,$B2
	.dc.b	$FE,$1F,$5A,$9C,$F2,$7F,$44,$BE,$15,$FE,$0B,$BE,$18,$B0,$E1,$39
	.dc.b	$BE,$2F,$F4,$7F,$0F,$F9,$E5,$7F,$1C,$BE,$29,$7C,$37,$F2,$5F,$9E
	.dc.b	$CC,$62,$FF,$96,$46,$27,$58,$7A,$9B,$3E,$5D,$FD,$BC,$1E,$E6,$C6
	.dc.b	$49,$8F,$F1,$03,$F9,$7F,$CF,$7F,$30,$F9,$E7,$CF,$7F,$32,$9F,$35
	.dc.b	$FF,$5F,$76,$27,$6F,$B6,$3F,$C8,$11,$88,$E5,$5D,$6B,$A6,$8B,$DC
	.dc.b	$77,$AE,$B7,$C6,$B5,$87,$19,$D6,$15,$F2,$C5,$78,$DF,$6C,$3A,$C7
	.dc.b	$03,$D1,$19,$9E,$10,$AE,$97,$E0,$8D,$9C,$70,$F2,$61,$D2,$3D,$A6
	.dc.b	$8F,$19,$4F,$19,$F1,$6F,$BD,$F6,$9E,$53,$A3,$8D,$7C,$97,$17,$E3
	.dc.b	$8C,$47,$9C,$F0,$46,$70,$F9,$51,$93,$FC,$A7,$7F,$E5,$47,$EC,$98
	.dc.b	$37,$F3,$DF,$C6,$7D,$5F,$F2,$CB,$63,$3D,$3F,$E4,$B6,$CE,$BE,$0F
	.dc.b	$FC,$59,$1C,$9A,$FD,$BA,$46,$23,$F1,$91,$8E,$5E,$47,$DB,$F9,$E7
	.dc.b	$CC,$FE,$47,$5F,$C5,$46,$2F,$F0,$17,$C5,$FA,$0D,$9D,$D7,$1D,$3F
	.dc.b	$35,$7C,$6F,$78,$FA,$2F,$F2,$D8,$71,$BD,$E1,$C6,$FF,$C2,$78,$71
	.dc.b	$BD,$E1,$C6,$FD,$6D,$AD,$DF,$15,$7C,$6B,$78,$FA,$AF,$85,$6A,$F8
	.dc.b	$5E,$E3,$FC,$47,$8F,$95,$B5,$C2,$75,$1C,$4E,$97,$1F,$CA,$46,$8D
	.dc.b	$D7,$38,$E0,$7F,$84,$A3,$E3,$4F,$FA,$7D,$BB,$FA,$27,$9E,$1F,$C1
	.dc.b	$6B,$4D,$9F,$93,$3F,$EB,$1B,$ED,$F3,$DF,$E8,$28,$FF,$5A,$35,$FF
	.dc.b	$10,$B6,$FF,$D5,$DF,$39,$FD,$32,$77,$D5,$D5,$FF,$22,$C3,$E8,$71
	.dc.b	$08,$C3,$C0,$EF,$F1,$5A,$C0,$D3,$8C,$BC,$1C,$57,$02,$1F,$E3,$68
	.dc.b	$D3,$8C,$7C,$FE,$2F,$23,$8C,$7F,$8F,$3E,$7F,$F9,$EB,$FE,$D3,$6B
	.dc.b	$9E,$9F,$0F,$9B,$EA,$9C,$DF,$FB,$3D,$BB,$FF,$33,$6D,$FC,$2A,$E1
	.dc.b	$7E,$AF,$9F,$93,$DE,$3E,$2F,$CB,$3E,$7E,$92,$7F,$DC,$DA,$7F,$8B
	.dc.b	$9F,$F7,$44,$ED,$B3,$B7,$E1,$77,$5D,$3E,$BF,$A7,$F2,$8C,$6B,$FD
	.dc.b	$27,$DA,$FF,$26,$DF,$FB,$CD,$AF,$F9,$81,$B5,$FE,$ED,$E6,$FF,$2E
	.dc.b	$4E,$FC,$BF,$BB,$5A,$BE,$91,$F8,$0A,$DD,$73,$8C,$7C,$04,$8E,$47
	.dc.b	$FA,$33,$0F,$F7,$D7,$9D,$F8,$28,$C6,$1F,$D7,$E3,$11,$FE,$74,$6C
	.dc.b	$FF,$C0,$1F,$9E,$FF,$32,$4F,$03,$FD,$2F,$3A,$AE,$5C,$DF,$94,$8C
	.dc.b	$4F,$FA,$72,$B8,$BD,$67,$F3,$30,$F9,$FA,$71,$8C,$E5,$F9,$EF,$9E
	.dc.b	$FD,$25,$FF,$A2,$1B,$97,$96,$75,$E3,$4A,$F3,$0A,$DE,$7F,$DF,$72
	.dc.b	$F9,$93,$FE,$B0,$9C,$BF,$EB,$2F,$9D,$7C,$FF,$EB,$9B,$1D,$27,$6B
	.dc.b	$E7,$DF,$01,$BF,$1E,$4F,$8A,$72,$F9,$2F,$FC,$D0,$F9,$BE,$7D,$EF
	.dc.b	$FE,$F1,$B1,$BB,$C4,$E7,$E7,$E4,$F1,$8F,$E0,$7F,$9B,$F6,$8F,$F5
	.dc.b	$6B,$78,$5B,$FF,$87,$63,$0D,$F9,$2E,$3A,$CB,$F9,$35,$A7,$FD,$AA
	.dc.b	$A7,$51,$DA,$6D,$F6,$8D,$CF,$F1,$6F,$9F,$76,$DF,$9B,$95,$F5,$A7
	.dc.b	$2D,$9C,$BF,$EC,$68,$4F,$39,$FE,$43,$F4,$1F,$C1,$28,$D3,$14,$F9
	.dc.b	$BD,$F8,$5E,$3F,$A0,$97,$F8,$6B,$E5,$F1,$DF,$96,$D3,$D5,$28,$C6
	.dc.b	$FF,$E4,$56,$C7,$FB,$E1,$7D,$07,$F3,$47,$31,$C7,$DC,$C1,$99,$B1
	.dc.b	$F9,$69,$D9,$23,$97,$A3,$CF,$96,$56,$E3,$2F,$18,$CB,$FE,$EC,$AF
	.dc.b	$F0,$8B,$FA,$2F,$8A,$7C,$37,$F4,$86,$C6,$63,$D1,$ED,$E5,$F5,$7D
	.dc.b	$04,$9D,$1F,$AC,$FA,$0F,$D0,$9F,$EA,$E7,$1F,$05,$FD,$52,$33,$E1
	.dc.b	$58,$93,$C6,$34,$61,$70,$7D,$0A,$E9,$18,$E0,$FF,$3A,$47,$83,$CF
	.dc.b	$EC,$49,$F2,$7C,$DD,$78,$01,$29,$D5,$F7,$AF,$89,$F2,$BF,$E4,$7D
	.dc.b	$8D,$26,$1F,$E4,$E1,$1A,$7E,$27,$D1,$7C,$13,$9B,$FC,$4B,$E8,$7F
	.dc.b	$3A,$F9,$03,$F9,$C7,$5D,$9F,$DF,$23,$78,$FE,$49,$BA,$7F,$A5,$5E
	.dc.b	$15,$CD,$B3,$BE,$CE,$47,$3B,$95,$E3,$8E,$DD,$72,$3C,$CF,$87,$C9
	.dc.b	$C4,$75,$8B,$7E,$39,$FF,$08,$FD,$0B,$E5,$FF,$C2,$7F,$42,$F9,$5A
	.dc.b	$6A,$26,$3F,$CD,$27,$4D,$9C,$BF,$16,$3F,$32,$DE,$BE,$A7,$6E,$BB
	.dc.b	$1C,$E2,$BC,$62,$E5,$FB,$AA,$E6,$DA,$E5,$5C,$48,$77,$3C,$D5,$CB
	.dc.b	$7F,$F6,$25,$F2,$6D,$72,$AE,$04,$BE,$CF,$23,$AB,$87,$53,$D2,$FB
	.dc.b	$47,$E5,$EF,$B3,$6F,$B5,$F1,$25,$F8,$A7,$B1,$D0,$E0,$78,$1E,$CD
	.dc.b	$32,$6C,$57,$F5,$47,$0F,$89,$CD,$78,$A7,$C1,$1A,$8C,$3D,$2B,$CA
	.dc.b	$D9,$F0,$BB,$3C,$47,$82,$FE,$0B,$D8,$CD,$F8,$EB,$9C,$64,$F2,$CF
	.dc.b	$09,$F3,$11,$C5,$E5,$B3,$D1,$AF,$23,$BC,$6D,$A6,$CA,$6C,$F5,$16
	.dc.b	$BC,$B5,$E2,$9F,$87,$3E,$79,$CC,$62,$75,$3D,$4E,$73,$CA,$BC,$AD
	.dc.b	$8F,$0B,$5D,$DF,$D2,$8F,$CF,$4F,$08,$C1,$F4,$23,$C3,$3D,$DA,$FA
	.dc.b	$63,$6F,$D7,$21,$F5,$4F,$0B,$A7,$07,$88,$DD,$7D,$78,$D3,$EE,$8F
	.dc.b	$DA,$1C,$8D,$D7,$9D,$A6,$4F,$91,$3E,$93,$9D,$FD,$58,$F0,$1A,$3F
	.dc.b	$51,$3C,$DC,$1E,$6B,$F7,$9F,$24,$B5,$E6,$9E,$17,$F8,$9A,$F1,$35
	.dc.b	$EE,$DA,$FE,$19,$D5,$FC,$83,$AA,$57,$F1,$D8,$FE,$46,$61,$FE,$39
	.dc.b	$18,$7F,$CA,$BF,$40,$FA,$1F,$E9,$90,$8C,$3F,$D3,$DA,$FF,$5B,$23
	.dc.b	$87,$07,$18,$FF,$62,$FB,$7F,$EC,$48,$FE,$A0,$DC,$3F,$C3,$B5,$D1
	.dc.b	$FF,$0F,$C7,$F6,$59,$FF,$34,$AE,$8E,$67,$4E,$99,$3A,$8C,$BC,$23
	.dc.b	$2F,$82,$32,$F8,$A3,$CB,$E3,$79,$BC,$69,$3E,$08,$CB,$E2,$8C,$BE
	.dc.b	$A8,$FF,$70,$FD,$03,$E8,$3F,$D1,$B3,$F1,$63,$1E,$27,$F7,$77,$FB
	.dc.b	$BF,$C5,$DA,$DD,$7F,$81,$7E,$89,$7F,$E4,$08,$F9,$D1,$8E,$4F,$F0
	.dc.b	$03,$63,$E7,$87,$0F,$D5,$F9,$27,$C4,$FF,$9F,$2B,$FD,$51,$18,$F8
	.dc.b	$5F,$F3,$37,$0F,$F9,$9E,$37,$E7,$E2,$BF,$C3,$11,$B7,$8D,$F3,$C3
	.dc.b	$FC,$3B,$39,$6C,$66,$FA,$2C,$3A,$4A,$BB,$DF,$49,$CC,$62,$79,$C6
	.dc.b	$29,$F4,$17,$E8,$70,$DC,$3C,$AD,$BF,$1B,$71,$FF,$35,$FD,$03,$DD
	.dc.b	$F1,$F0,$3E,$87,$FC,$5F,$8F,$FA,$E2,$B4,$FF,$B3,$E7,$5E,$4F,$AB
	.dc.b	$25,$69,$FF,$16,$37,$1F,$F4,$97,$C7,$BE,$05,$F0,$1C,$6B,$59,$7C
	.dc.b	$ED,$FF,$C3,$4F,$CE,$D7,$E0,$78,$9F,$05,$B1,$93,$E9,$C6,$5A,$67
	.dc.b	$3F,$A7,$19,$7E,$AF,$A3,$F2,$CF,$99,$F4,$36,$BE,$EC,$59,$FB,$73
	.dc.b	$75,$FF,$49,$FD,$07,$FB,$3A,$FF,$DE,$50,$FA,$1F,$C0,$47,$F0,$59
	.dc.b	$D5,$FF,$98,$AB,$5C,$FF,$26,$73,$1D,$3B,$3F,$0B,$C3,$F1,$FC,$8D
	.dc.b	$47,$E0,$7E,$86,$65,$F9,$2F,$46,$E7,$F9,$09,$B8,$FE,$42,$6D,$7C
	.dc.b	$D9,$B5,$F9,$1E,$1F,$F3,$4C,$FE,$72,$7E,$E7,$D0,$FD,$2A,$FF,$33
	.dc.b	$37,$0F,$F5,$9F,$91,$EE,$9F,$F3,$E4,$E6,$FF,$92,$7D,$03,$E8,$7F
	.dc.b	$0C,$D8,$FF,$41,$61,$FE,$BC,$EA,$7C,$4E,$AF,$F1,$91,$FE,$85,$AD
	.dc.b	$9F,$CB,$DC,$63,$FE,$79,$FA,$07,$C4,$70,$F2,$3F,$29,$F4,$3F,$C1
	.dc.b	$73,$FF,$64,$B7,$1F,$E7,$17,$FE,$B2,$7A,$E1,$FE,$95,$9D,$3F,$E9
	.dc.b	$79,$D3,$DF,$CE,$FD,$04,$7E,$6A,$B8,$CA,$37,$7F,$ED,$87,$51,$E7
	.dc.b	$EE,$FD,$47,$D0,$BE,$87,$9C,$7E,$AF,$C7,$7C,$77,$F0,$6C,$9F,$3E
	.dc.b	$F4,$35,$39,$8C,$4F,$8A,$B8,$BE,$C3,$8D,$70,$71,$3C,$32,$FF,$51
	.dc.b	$57,$06,$DF,$5F,$A1,$E1,$1F,$EA,$7B,$7A,$BE,$59,$FA,$31,$F0,$8C
	.dc.b	$9E,$17,$33,$C6,$78,$71,$7F,$09,$AD,$CF,$FB,$19,$57,$0D,$FF,$E0
	.dc.b	$0C,$69,$B7,$F0,$7E,$57,$93,$FC,$4F,$E8,$5F,$05,$F1,$4C,$16,$FE
	.dc.b	$BC,$69,$F8,$71,$A6,$9A,$71,$8F,$97,$57,$C1,$63,$25,$FF,$80,$FD
	.dc.b	$2F,$E7,$CF,$CE,$F9,$EF,$E4,$D1,$8C,$39,$46,$3D,$2F,$B4,$DA,$E0
	.dc.b	$7E,$32,$B8,$5F,$6B,$FF,$75,$46,$27,$FD,$D8,$7E,$1B,$AB,$94,$FF
	.dc.b	$B7,$7C,$5F,$B9,$7C,$65,$F4,$0F,$A1,$FE,$75,$5A,$C3,$FD,$ED,$5A
	.dc.b	$9C,$F2,$7F,$46,$BE,$15,$FA,$0B,$E1,$8B,$0E,$13,$9B,$E2,$FF,$49
	.dc.b	$F0,$BF,$9F,$D0,$5F,$26,$FC,$57,$F6,$6C,$DF,$43,$9A,$4E,$63,$0D
	.dc.b	$C7,$D6,$DC,$7C,$DB,$7B,$B8,$3D,$ED,$8C,$93,$26,$5F,$E1,$15,$F4
	.dc.b	$3E,$7F,$A1,$7D,$0F,$F9,$07,$56,$3F,$D8,$EF,$D9,$7B,$AD,$4F,$D9
	.dc.b	$BE,$35,$FE,$46,$AD,$3F,$0E,$F9,$E3,$88,$FF,$27,$4D,$9C,$C2,$B9
	.dc.b	$BA,$31,$87,$82,$F9,$47,$90,$C9,$DF,$0F,$24,$FF,$8A,$B0,$ED,$18
	.dc.b	$9F,$31,$C0,$FA,$EE,$5F,$05,$F8,$B0,$F2,$BE,$48,$F0,$39,$9E,$11
	.dc.b	$8B,$F3,$3D,$4C,$1E,$5C,$3D,$2F,$D2,$3C,$57,$DA,$7D,$11,$8A,$EA
	.dc.b	$EE,$FC,$17,$CB,$0F,$F3,$EC,$F3,$3F,$C9,$25,$39,$9F,$E7,$AF,$A0
	.dc.b	$FE,$3F,$E5,$69,$B1,$FE,$50,$6B,$BE,$9F,$F2,$7B,$9F,$83,$FC,$5B
	.dc.b	$75,$E5,$18,$8E,$CD,$7F,$D1,$94,$8E,$39,$7D,$EF,$A1,$7C,$DF,$FA
	.dc.b	$46,$30,$DD,$B8,$D7,$06,$E3,$F8,$BB,$E3,$8D,$BE,$0F,$EC,$F1,$F0
	.dc.b	$57,$E4,$5B,$1F,$81,$AE,$2D,$AE,$2A,$77,$8F,$E5,$1B,$87,$FA,$9E
	.dc.b	$F8,$CF,$A2,$BB,$63,$E7,$7C,$4D,$53,$C2,$B9,$61,$E0,$C3,$AD,$70
	.dc.b	$C7,$D8,$C3,$F3,$11,$A7,$CD,$87,$5C,$3A,$47,$F1,$36,$BA,$AE,$47
	.dc.b	$F3,$17,$C7,$7F,$EC,$18,$C4,$FC,$47,$57,$DE,$31,$AF,$A2,$FF,$C0
	.dc.b	$F1,$87,$D0,$7E,$B7,$E8,$3F,$D0,$CF,$F8,$31,$AF,$F1,$49,$FF,$68
	.dc.b	$46,$2F,$C9,$F3,$DF,$D5,$2B,$87,$5F,$96,$FF,$DA,$D3,$FE,$CF,$6F
	.dc.b	$D0,$E6,$BF,$99,$61,$CE,$7A,$FC,$56,$65,$28,$C3,$63,$FC,$43,$6F
	.dc.b	$07,$35,$27,$D4,$F3,$9D,$46,$0E,$3E,$2E,$4E,$0F,$BC,$FE,$8D,$F4
	.dc.b	$3F,$D0,$1F,$F1,$7B,$5D,$07,$1A,$7C,$4F,$37,$D5,$18,$CF,$FD,$F0
	.dc.b	$DA,$EB,$D5,$E5,$AD,$BE,$7F,$A1,$93,$E0,$8F,$89,$F1,$CF,$A1,$A4
	.dc.b	$9F,$87,$DD,$5F,$2C,$ED,$FF,$23,$B5,$F4,$EB,$FE,$40,$73,$5F,$1F
	.dc.b	$D2,$91,$83,$FC,$9F,$F0,$3F,$E6,$06,$F9,$7F,$5E,$31,$1F,$13,$A3
	.dc.b	$FC,$E4,$E8,$87,$C8,$8D,$F0,$7E,$76,$77,$39,$AF,$C1,$47,$FB,$1E
	.dc.b	$3F,$CB,$ED,$74,$7F,$98,$BC,$DF,$C2,$3F,$51,$B5,$FA,$96,$F0,$FF
	.dc.b	$80,$65,$F4,$1F,$F0,$01,$E5,$C9,$F9,$1A,$DC,$7F,$81,$DA,$F0,$F2
	.dc.b	$75,$3F,$97,$A7,$D0,$D3,$35,$FE,$76,$8F,$7E,$1E,$1F,$A0,$FD,$44
	.dc.b	$FE,$29,$B9,$7E,$FE,$B7,$3F,$E1,$79,$CD,$71,$F1,$BF,$4D,$18,$9F
	.dc.b	$F0,$E5,$70,$51,$95,$F3,$EF,$A1,$FD,$73,$63,$FD,$17,$3B,$FA,$17
	.dc.b	$BD,$AF,$1A,$57,$FA,$42,$72,$FF,$A4,$AF,$FC,$EA,$FF,$A5,$3E,$85
	.dc.b	$F0,$5F,$BB,$6C,$6E,$F1,$39,$FA,$19,$3C,$63,$F8,$17,$D0,$F6,$8F
	.dc.b	$D7,$B7,$83,$F9,$23,$78,$3E,$E2,$F9,$2E,$3A,$CB,$FF,$53,$CE,$DF
	.dc.b	$F2,$13,$B7,$31,$B7,$BB,$E2,$9F,$E2,$9F,$42,$ED,$BF,$35,$2F,$FD
	.dc.b	$67,$39,$6C,$E5,$FE,$0F,$09,$E7,$3F,$C7,$FE,$83,$F8,$25,$1A,$62
	.dc.b	$9F,$37,$BF,$0B,$FC,$13,$5B,$6D,$7F,$15,$8F,$CF,$19,$3F,$C1,$FF
	.dc.b	$19,$F9,$F7,$F8,$CB,$76,$FE,$9C,$DD,$BF,$4C,$BE,$83,$F9,$A3,$9A
	.dc.b	$FF,$35,$35,$34,$E0,$FF,$B5,$9C,$76,$72,$39,$CF,$4F,$3F,$AF,$24
	.dc.b	$BE,$11,$97,$94,$65,$F4,$C7,$F4,$7F,$89,$7C,$25,$24,$F8,$7D,$9A
	.dc.b	$FC,$41,$3C,$1F,$F1,$9B,$5D,$35,$F9,$7F,$3F,$FA,$B3,$FD,$7E,$FF
	.dc.b	$BC,$9A,$FF,$44,$75,$DF,$FD,$6D,$CE,$DF,$DE,$DB,$B7,$F1,$37,$75
	.dc.b	$CF,$DF,$E0,$71,$BE,$5E,$64,$9F,$03,$D7,$E6,$71,$1E,$18,$B8,$F7
	.dc.b	$9F,$8D,$F2,$72,$BF,$D4,$53,$FE,$0D,$CD,$3E,$7E,$AD,$57,$C2,$8F
	.dc.b	$C7,$CE,$A3,$E3,$C7,$E4,$BE,$7B,$F4,$11,$FD,$91,$FC,$F3,$DB,$5F
	.dc.b	$E6,$11,$88,$D7,$4F,$F7,$F8,$E3,$8F,$F8,$7D,$D3,$7A,$B5,$32,$BA
	.dc.b	$1F,$EB,$BA,$DD,$F6,$AD,$4F,$8F,$A7,$B1,$1E,$22,$BF,$9E,$9F,$AE
	.dc.b	$72,$D7,$91,$8F,$1B,$FC,$2D,$F3,$CF,$93,$A4,$9E,$BF,$3D,$FA,$59
	.dc.b	$D4,$78,$1A,$FF,$82,$9B,$A7,$F8,$FE,$7F,$58,$7F,$9C,$E7,$9F,$3F
	.dc.b	$F8,$96,$31,$1F,$E1,$27,$F9,$8B,$63,$44,$7C,$7D,$15,$D1,$D0,$C9
	.dc.b	$4C,$92,$BC,$91,$C5,$F3,$4D,$D7,$FB,$65,$F8,$A7,$37,$CD,$AF,$9E
	.dc.b	$F8,$3F,$EF,$39,$F5,$CE,$8E,$F1,$C6,$BA,$DF,$08,$57,$79,$CC,$6C
	.dc.b	$E9,$3F,$1E,$F8,$57,$6A,$DD,$78,$CE,$07,$4A,$DD,$70,$6C,$6A,$B9
	.dc.b	$3D,$8F,$92,$7C,$7A,$EB,$7E,$B7,$E2,$DF,$39,$E4,$D7,$81,$F0,$DC
	.dc.b	$4F,$18,$C1,$E1,$BF,$1B,$C0,$F4,$1D,$0D,$14,$D5,$FA,$1D,$9F,$04
	.dc.b	$CC,$74,$9F,$04,$69,$B3,$F3,$E3,$E0,$1C,$0F,$83,$1A,$9E,$57,$F0
	.dc.b	$89,$3C,$27,$EE,$BC,$4F,$2D,$FC,$09,$DC,$7D,$E8,$C3,$C6,$3E,$5B
	.dc.b	$EE,$38,$96,$7C,$F3,$C2,$3D,$53,$C1,$F0,$4E,$9E,$11,$E0,$BF,$74
	.dc.b	$70,$3C,$04,$3C,$87,$03,$0F,$03,$E7,$12,$FF,$25,$1B,$3A,$1C,$8F
	.dc.b	$1C,$F8,$6F,$D1,$5C,$9B,$5C,$AB,$67,$82,$32,$FD,$2A,$F5,$4F,$D2
	.dc.b	$8F,$2B,$5F,$0D,$79,$5E,$06,$8F,$C4,$5F,$DB,$6D,$74,$69,$B3,$E8
	.dc.b	$9F,$C4,$F9,$FF,$8B,$74,$48,$FE,$47,$1F,$C6,$4C,$3F,$C6,$63,$0F
	.dc.b	$F3,$4F,$9D,$7C,$F7,$FB,$0D,$B1,$98,$FF,$63,$7A,$DF,$E5,$D9,$CC
	.dc.b	$FF,$AD,$1B,$8F,$FB,$A5,$B8,$FF,$A0,$6B,$FC,$C8,$3A,$39,$F4,$4C
	.dc.b	$3F,$A2,$CE,$5B,$39,$7F,$D8,$30,$9F,$1C,$C9,$E3,$1B,$C3,$FD,$F5
	.dc.b	$5C,$E7,$8F,$47,$AB,$C2,$91,$E0,$8F,$69,$87,$BC,$61,$FE,$B3,$F4
	.dc.b	$4F,$8A,$FF,$21,$B6,$3F,$C3,$C1,$C3,$FE,$21,$6B,$FE,$46,$F9,$F7
	.dc.b	$CF,$FF,$91,$63,$FD,$4D,$18,$F0,$B9,$BF,$E2,$8F,$88,$B5,$BA,$7C
	.dc.b	$6B,$E1,$7F,$D0,$3C,$3F,$DD,$A3,$1F,$02,$8C,$4E,$AB,$FD,$E1,$5A
	.dc.b	$9C,$F9,$9B,$9C,$D2,$F1,$C3,$9C,$3C,$17,$8C,$18,$79,$30,$E1,$1D
	.dc.b	$0F,$0E,$1B,$9D,$55,$BF,$14,$78,$3E,$77,$18,$3E,$7A,$C6,$1B,$8F
	.dc.b	$B1,$B5,$E8,$7B,$2F,$9D,$7C,$EF,$87,$67,$CF,$7F,$8C,$B0,$FF,$0A
	.dc.b	$D6,$9F,$F0,$FC,$EB,$C7,$E7,$39,$C3,$FC,$B8,$D7,$FD,$02,$D7,$FD
	.dc.b	$C0,$D7,$E2,$78,$5E,$CC,$7D,$AD,$C3,$EA,$B6,$7E,$8B,$78,$3E,$86
	.dc.b	$BF,$E2,$1A,$FB,$77,$C9,$A7,$DE,$F3,$7E,$49,$F2,$BE,$7A,$C6,$49
	.dc.b	$8F,$DE,$8C,$AC,$3F,$04,$BE,$77,$FD,$AB,$5F,$E4,$29,$7C,$F7,$E2
	.dc.b	$DD,$CE,$AB,$95,$7F,$98,$2B,$4F,$E1,$A3,$A7,$5F,$67,$15,$C5,$FF
	.dc.b	$55,$B5,$CC,$6D,$FC,$77,$CE,$FE,$18,$FC,$3F,$9B,$6F,$F3,$F1,$67
	.dc.b	$F3,$73,$71,$F9,$73,$6B,$F3,$3C,$BF,$AB,$AC,$5F,$E7,$3E,$77,$E8
	.dc.b	$3E,$C3,$1C,$1F,$8E,$F1,$7D,$B1,$8C,$3F,$CB,$FF,$3A,$F9,$DF,$C9
	.dc.b	$B6,$3F,$CC,$7C,$CF,$85,$CD,$7E,$62,$3F,$93,$4E,$BC,$5F,$CD,$FC
	.dc.b	$E3,$E1,$75,$1F,$9E,$F9,$D7,$87,$4D,$F9,$77,$0F,$F3,$8E,$1F,$EB
	.dc.b	$86,$C7,$4F,$85,$DC,$69,$C3,$FE,$78,$F9,$D7,$CE,$E3,$83,$E1,$1F
	.dc.b	$C4,$BE,$33,$F7,$94,$F9,$C7,$9B,$FB,$2E,$85,$6A,$73,$5F,$E8,$49
	.dc.b	$CB,$FE,$8E,$39,$57,$1A,$E0,$D8,$E1,$1C,$FC,$CC,$61,$E4,$97,$A7
	.dc.b	$E4,$9F,$91,$1F,$E9,$23,$27,$77,$33,$C6,$78,$71,$7F,$0A,$AD,$CF
	.dc.b	$E0,$5F,$E8,$51,$99,$91,$E9,$C9,$F2,$18,$BF,$8F,$F3,$AF,$77,$E1
	.dc.b	$1F,$C5,$9C,$5F,$18,$C1,$A7,$FD,$53,$1A,$9F,$15,$DE,$FF,$DB,$CD
	.dc.b	$AF,$D7,$36,$BF,$A1,$53,$DD,$EB,$5F,$EB,$4E,$5F,$E6,$F5,$D3,$E6
	.dc.b	$7F,$94,$C6,$31,$FE,$89,$5F,$76,$73,$E7,$FD,$B6,$C7,$C1,$6D,$FE
	.dc.b	$1A,$F9,$4F,$F4,$63,$F0,$DD,$5A,$C3,$F8,$7F,$CD,$7E,$0A,$FF,$5D
	.dc.b	$2F,$9C,$7C,$EF,$F3,$CA,$D5,$FF,$57,$AD,$4E,$79,$3F,$9F,$5F,$0C
	.dc.b	$7F,$DA,$38,$70,$7F,$AB,$DF,$07,$35,$C5,$FE,$83,$E0,$FF,$3F,$C7
	.dc.b	$CB,$3E,$1F,$85,$FE,$B5,$0F,$9D,$E9,$A7,$F6,$2C,$95,$A9,$CC,$A7
	.dc.b	$38,$FF,$2B,$9C,$C6,$27,$F8,$75,$F5,$C3,$E0,$6D,$F0,$78,$2E,$3E
	.dc.b	$C6,$DF,$ED,$30,$57,$F6,$28,$FE,$EB,$F3,$BF,$DD,$BE,$75,$F3,$BF
	.dc.b	$CC,$EF,$78,$FF,$BD,$AF,$62,$B6,$FB,$70,$FF,$03,$46,$23,$F9,$E6
	.dc.b	$1D,$70,$E9,$5A,$7F,$A0,$4F,$4A,$E1,$87,$82,$3B,$4E,$95,$F3,$7C
	.dc.b	$18,$57,$B9,$C0,$C9,$4C,$DF,$6C,$3C,$51,$E2,$AE,$58,$78,$6B,$D5
	.dc.b	$3E,$5C,$3D,$38,$77,$8C,$1D,$A7,$6F,$98,$F0,$57,$3A,$F5,$5F,$6C
	.dc.b	$3C,$35,$E4,$9E,$AF,$9C,$EA,$7E,$16,$36,$E2,$BA,$3E,$93,$FC,$9E
	.dc.b	$78,$4F,$F2,$8D,$61,$C1,$FD,$BF,$E7,$7F,$91,$FA,$5F,$E5,$66,$C7
	.dc.b	$CF,$6B,$F5,$9A,$F1,$F8,$1F,$F0,$23,$B6,$BB,$6B,$BD,$BB,$47,$0C
	.dc.b	$FF,$0D,$F3,$AF,$96,$FF,$01,$46,$1B,$AF,$1C,$3E,$D6,$2C,$39,$4F
	.dc.b	$F8,$27,$0E,$55,$F8,$36,$BC,$27,$6F,$F9,$52,$B9,$63,$BC,$3B,$CF
	.dc.b	$F9,$62,$F8,$E3,$F9,$BB,$DC,$F3,$BE,$0C,$30,$49,$F5,$39,$79,$61
	.dc.b	$E2,$85,$74,$9D,$C7,$E5,$EF,$C3,$8F,$C1,$9E,$38,$73,$8F,$81,$8F
	.dc.b	$A2,$F9,$E3,$D7,$0F,$83,$7F,$E3,$86,$ED,$C2,$BF,$CE,$B3,$FE,$3C
	.dc.b	$CB,$C5,$8E,$67,$FB,$DD,$F0,$7E,$26,$3F,$E3,$7B,$E1,$8F,$AF,$E7
	.dc.b	$3F,$D0,$95,$D3,$2F,$F1,$E4,$62,$3F,$C0,$AD,$AF,$F2,$5F,$CD,$7F
	.dc.b	$51,$9E,$07,$FA,$22,$75,$5F,$E4,$59,$7C,$EF,$86,$30,$DA,$E1,$D7
	.dc.b	$55,$BE,$23,$0D,$8E,$E6,$A3,$86,$1F,$DD,$A3,$33,$B7,$D3,$3C,$8D
	.dc.b	$F8,$16,$39,$D7,$F8,$79,$BE,$26,$5F,$F1,$2F,$CE,$FF,$42,$7C,$2D
	.dc.b	$73,$D5,$E2,$F3,$7D,$53,$9D,$BF,$CE,$2D,$DB,$FC,$D0,$DB,$FF,$57
	.dc.b	$E3,$FE,$DD,$AD,$E1,$EC,$F9,$D9,$5E,$08,$F8,$DF,$1C,$F9,$DA,$49
	.dc.b	$F5,$F6,$D7,$FC,$7B,$3B,$7E,$4B,$5F,$F2,$0E,$3F,$E3,$C7,$35,$F3
	.dc.b	$3D,$29,$18,$3E,$8F,$CE,$FF,$96,$5B,$E5,$7D,$98,$C4,$7F,$93,$39
	.dc.b	$FF,$CD,$0E,$88,$7F,$94,$23,$7A,$FF,$A5,$A7,$73,$9A,$FC,$24,$7F
	.dc.b	$B0,$23,$F0,$2D,$74,$7E,$13,$CF,$FC,$05,$7F,$A8,$1B,$C1,$FE,$97
	.dc.b	$6C,$FF,$B7,$BE,$73,$FC,$DE,$F1,$E5,$FC,$94,$EA,$0D,$9D,$C7,$F9
	.dc.b	$A2,$31,$F3,$BF,$99,$93,$CA,$B1,$97,$E7,$7E,$73,$F5,$18,$7F,$9C
	.dc.b	$1B,$97,$FA,$DB,$0F,$87,$97,$FA,$B6,$72,$DF,$13,$91,$D2,$FF,$47
	.dc.b	$3A,$7F,$DF,$3F,$36,$F9,$DF,$D6,$36,$37,$5F,$AF,$F9,$D7,$BF,$AF
	.dc.b	$4A,$57,$EB,$A7,$2F,$EC,$2F,$FC,$E6,$FE,$CB,$E7,$5F,$17,$FB,$96
	.dc.b	$C6,$EF,$13,$9F,$9D,$93,$C6,$3F,$80,$7E,$3E,$F8,$3A,$49,$FD,$FF
	.dc.b	$CF,$7E,$97,$83,$FD,$A1,$5B,$9C,$F7,$FF,$09,$A3,$C6,$3F,$D3,$B3
	.dc.b	$B7,$FD,$3F,$3A,$78,$9F,$C3,$23,$5C,$7F,$EE,$89,$F8,$33,$C3,$7E
	.dc.b	$C7,$7F,$92,$B1,$E6,$8A,$F6,$8F,$E3,$9F,$3C,$EF,$BF,$37,$2B,$F8
	.dc.b	$DC,$E5,$B3,$97,$FD,$6B,$09,$E7,$3F,$C9,$FE,$7B,$F8,$55,$1A,$62
	.dc.b	$9F,$39,$BF,$13,$C7,$F4,$72,$FF,$95,$DF,$09,$C4,$78,$27,$83,$FC
	.dc.b	$C3,$E3,$BF,$43,$B3,$85,$32,$73,$18,$84,$61,$D7,$CF,$FF,$3F,$8C
	.dc.b	$46,$BD,$DC,$19,$9D,$7A,$1B,$E8,$95,$E3,$8C,$BE,$58,$CB,$FE,$DC
	.dc.b	$BF,$F0,$83,$FD,$23,$E3,$1F,$0D,$FD,$3D,$B1,$98,$F2,$7B,$99,$7C
	.dc.b	$DF,$47,$27,$44,$3B,$39,$7F,$DD,$75,$D1,$FF,$76,$FD,$07,$ED,$53
	.dc.b	$FE,$C9,$7F,$AD,$B5,$DF,$83,$FD,$8E,$38,$19,$DF,$E8,$4F,$D0,$8C
	.dc.b	$6F,$F3,$DD,$FA,$1C,$49,$3D,$BE,$6E,$BE,$48,$B7,$D2,$3C,$B3,$F1
	.dc.b	$7E,$57,$FC,$AB,$91,$A4,$9F,$F3,$F0,$8D,$3E,$CF,$A3,$F5,$1C,$DF
	.dc.b	$A5,$38,$74,$60,$FE,$FB,$F4,$3F,$A3,$58,$79,$63,$F4,$51,$AE,$BF
	.dc.b	$F0,$44,$6E,$75,$D7,$89,$95,$D2,$37,$D5,$23,$BB,$CB,$CB,$C4,$78
	.dc.b	$37,$FC,$D6,$3F,$AA,$8D,$9E,$69,$FF,$0C,$7D,$0B,$E4,$BF,$C3,$5F
	.dc.b	$42,$FA,$2A,$6A,$25,$FF,$86,$A7,$4D,$9C,$BE,$13,$F5,$B5,$D5,$D1
	.dc.b	$0D,$07,$4E,$1B,$AF,$DD,$AE,$F3,$C1,$B1,$F8,$19,$DE,$5F,$B9,$6D
	.dc.b	$7E,$FA,$B7,$3F,$BD,$6C,$79,$B0,$F4,$DF,$7C,$3E,$0C,$FF,$8E,$30
	.dc.b	$E6,$F9,$EB,$2F,$8E,$78,$E1,$E1,$9C,$DF,$3A,$E5,$1A,$7C,$95,$AA
	.dc.b	$E1,$1A,$3E,$35,$71,$7E,$25,$71,$58,$7A,$9B,$79,$C3,$CF,$39,$34
	.dc.b	$FA,$8C,$C7,$86,$7E,$11,$E1,$3D,$87,$8B,$0F,$2B,$E2,$30,$7A,$8E
	.dc.b	$18,$7D,$3C,$3A,$47,$90,$CC,$F2,$9F,$61,$1C,$12,$3A,$1E,$23,$C0
	.dc.b	$48,$E0,$E9,$F0,$3D,$8C,$47,$DF,$AF,$89,$5E,$4C,$3E,$1B,$F2,$DF
	.dc.b	$8D,$1A,$AF,$69,$E9,$3C,$15,$E5,$71,$7E,$27,$F0,$87,$84,$CC,$78
	.dc.b	$5C,$61,$F4,$1E,$AF,$D3,$C3,$C9,$7E,$19,$F8,$17,$D6,$B5,$1E,$4C
	.dc.b	$3D,$2F,$77,$D2,$F6,$BE,$6D,$38,$5F,$5B,$F4,$57,$29,$F6,$9F,$20
	.dc.b	$87,$B0,$C9,$EC,$33,$7E,$47,$CF,$87,$BD,$DC,$7B,$DC,$9F,$8E,$39
	.dc.b	$CF,$A1,$DE,$1E,$86,$9E,$83,$D4,$72,$3C,$47,$96,$FF,$02,$E6,$7D
	.dc.b	$91,$83,$DF,$87,$C1,$C3,$CD,$1F,$36,$FF,$05,$3F,$98,$C3,$D9,$3C
	.dc.b	$30,$F0,$B5,$F2,$9D,$E7,$A3,$F2,$30,$F0,$E1,$F8,$77,$DD,$5E,$38
	.dc.b	$FE,$25,$D5,$FC,$C0,$95,$FC,$B7,$AA,$BF,$C7,$E2,$D7,$F2,$39,$FE
	.dc.b	$6F,$D3,$FC,$E3,$A6,$5F,$F3,$58,$CB,$FC,$DE,$32,$FF,$29,$8F,$F6
	.dc.b	$C7,$D0,$BE,$8B,$FD,$B5,$ED,$FF,$B8,$2B,$FD,$AE,$D7,$F9,$BC,$ED
	.dc.b	$FF,$6D,$B5,$FE,$77,$1B,$C7,$FB,$A3,$73,$FF,$16,$56,$E7,$FB,$C3
	.dc.b	$63,$FC,$61,$3B,$AF,$F1,$AB,$54,$ED,$CF,$4F,$A3,$A6,$47,$A2,$3D
	.dc.b	$66,$1F,$2C,$61,$F6,$F8,$BF,$03,$C5,$91,$82,$1E,$C7,$2F,$BA,$3F
	.dc.b	$E0,$07,$D0,$BE,$87,$FD,$5B,$3F,$E4,$98,$C7,$87,$FE,$0E,$7F,$C1
	.dc.b	$DF,$17,$6B,$75,$FE,$52,$FA,$35,$FF,$86,$E3,$FC,$A5,$18,$DB,$FC
	.dc.b	$2E,$D8,$FF,$2B,$1C,$67,$83,$FE,$5C,$FA,$17,$89,$88,$46,$2F,$FC
	.dc.b	$42,$D8,$E6,$6A,$79,$FC,$A3,$E2,$BF,$D4,$F7,$FE,$68,$E3,$FF,$1D
	.dc.b	$CE,$7E,$17,$FC,$7B,$18,$AD,$E1,$FE,$03,$AD,$5F,$13,$FC,$E2,$F5
	.dc.b	$F4,$B8,$C6,$29,$86,$2D,$7B,$AE,$51,$8A,$DC,$62,$7D,$1F,$45,$C5
	.dc.b	$F5,$67,$69,$36,$30,$DA,$E5,$3A,$6C,$F8,$ED,$F4,$2F,$A1,$F3,$70
	.dc.b	$BE,$8B,$FC,$A1,$87,$9A,$F7,$5A,$9C,$C7,$F8,$8A,$F9,$56,$DB,$1F
	.dc.b	$E2,$BF,$90,$7C,$1F,$FB,$F2,$B5,$87,$8E,$F7,$C1,$FE,$63,$6C,$FF
	.dc.b	$A7,$AF,$8F,$C1,$6E,$FE,$17,$8B,$F0,$DB,$5F,$01,$B9,$7D,$9D,$B9
	.dc.b	$BF,$3E,$B8,$57,$C6,$F4,$3E,$61,$F3,$7E,$8A,$C6,$49,$5C,$E3,$2F
	.dc.b	$F9,$0B,$E8,$7F,$DF,$73,$F7,$35,$7D,$17,$E2,$9F,$F3,$F4,$EA,$BF
	.dc.b	$C9,$D1,$AE,$6F,$A9,$39,$8D,$F6,$7E,$27,$7F,$93,$E4,$6A,$26,$5F
	.dc.b	$56,$F8,$3F,$1F,$8B,$FC,$C4,$E6,$7F,$01,$F4,$5F,$E6,$78,$FB,$8D
	.dc.b	$CF,$FC,$B7,$E2,$F9,$E7,$FD,$95,$39,$CF,$FD,$2D,$F4,$4F,$A2,$FC
	.dc.b	$53,$6B,$F0,$DD,$0F,$8A,$CD,$7E,$3E,$1E,$2D,$3F,$25,$F4,$2F,$89
	.dc.b	$C6,$1F,$8F,$FA,$2F,$E0,$D9,$FF,$A9,$5B,$1F,$ED,$DA,$FF,$65,$BB
	.dc.b	$F0,$F9,$F9,$92,$3F,$31,$3F,$8F,$E0,$FF,$3B,$BC,$1B,$DB,$E2,$3F
	.dc.b	$49,$F4,$4F,$A2,$C5,$FE,$87,$C4,$7C,$8F,$EE,$E1,$F3,$EF,$3F,$F5
	.dc.b	$95,$A9,$FD,$E9,$A8,$CD,$71,$AE,$13,$D6,$31,$3A,$8C,$1C,$A5,$F4
	.dc.b	$59,$CF,$FD,$23,$2F,$4B,$C9,$3F,$2E,$3F,$70,$64,$F1,$B9,$9E,$33
	.dc.b	$C3,$8B,$F8,$25,$6E,$7F,$D4,$4F,$82,$33,$32,$3C,$9F,$2B,$6B,$FD
	.dc.b	$47,$F4,$2F,$A2,$F7,$1F,$C3,$DC,$57,$AE,$30,$69,$FE,$1D,$1A,$9F
	.dc.b	$0E,$57,$97,$FD,$CD,$36,$95,$23,$0A,$3E,$AF,$C1,$76,$7F,$8D,$72
	.dc.b	$FF,$2E,$9E,$BF,$3F,$FC,$8E,$31,$87,$FB,$FA,$BE,$EC,$E7,$CF,$FB
	.dc.b	$0D,$8F,$A8,$DB,$FF,$61,$DF,$29,$F2,$9F,$86,$EA,$D6,$1F,$EF,$FF
	.dc.b	$13,$F0,$D7,$FE,$CF,$97,$D0,$BE,$8B,$F9,$B5,$6A,$FD,$96,$BD,$D6
	.dc.b	$B9,$7F,$9C,$E1,$C7,$1F,$E7,$18,$F1,$7E,$3D,$F0,$73,$5C,$5F,$EA
	.dc.b	$70,$F0,$BF,$93,$E2,$BE,$4F,$F3,$3F,$88,$FE,$B9,$9B,$E8,$B3,$3B
	.dc.b	$9C,$C6,$29,$18,$CB,$DA,$DC,$BC,$FB,$7C,$0E,$0F,$82,$D8,$C9,$31
	.dc.b	$FF,$78,$9F,$82,$FA,$2F,$C1,$FD,$13,$E8,$7F,$0B,$3E,$9B,$FE,$C0
	.dc.b	$EC,$4E,$DF,$E9,$D1,$F3,$A3,$11,$CA,$BA,$D7,$4C,$D7,$B8,$FF,$08
	.dc.b	$CE,$AB,$9E,$15,$E4,$78,$27,$95,$71,$8F,$F1,$1D,$6E,$FA,$DF,$19
	.dc.b	$F4,$61,$E4,$9C,$C7,$33,$89,$4F,$81,$3C,$55,$75,$BF,$04,$79,$2F
	.dc.b	$A4,$62,$3B,$E1,$E6,$BE,$57,$C5,$B3,$DF,$0F,$1E,$1F,$22,$38,$9F
	.dc.b	$08,$F2,$18,$3C,$A6,$0F,$29,$FE,$3A,$3E,$B1,$6B,$0E,$0F,$BB,$F4
	.dc.b	$5F,$C8,$7D,$2E,$0D,$8F,$F2,$1B,$5F,$F2,$7B,$5E,$3F,$00,$E0,$E2
	.dc.b	$D8,$E2,$D7,$8B,$5E,$31,$F7,$7E,$89,$F3,$BF,$90,$CB,$F1,$31,$8A
	.dc.b	$FC,$16,$0C,$3B,$57,$E3,$A3,$18,$73,$C3,$96,$BF,$9A,$C7,$96,$3F
	.dc.b	$76,$F6,$DC,$79,$D6,$EF,$8C,$FF,$26,$BE,$38,$F7,$BD,$CF,$F2,$AC
	.dc.b	$39,$CF,$A2,$F8,$CA,$F8,$E1,$F0,$23,$C1,$8F,$9D,$86,$09,$7C,$DC
	.dc.b	$BE,$18,$F8,$78,$F8,$A7,$B6,$1E,$0C,$3B,$C7,$96,$75,$1F,$0A,$BF
	.dc.b	$91,$CF,$03,$FC,$AD,$19,$D7,$CD,$87,$86,$7E,$F4,$7C,$57,$78,$7F
	.dc.b	$06,$BE,$79,$7C,$2C,$BF,$84,$FD,$0F,$F9,$EA,$FF,$CC,$4D,$79,$9F
	.dc.b	$D4,$9C,$57,$93,$E8,$1C,$3D,$72,$E7,$D5,$F9,$AB,$E5,$87,$FA,$89
	.dc.b	$AE,$8F,$D1,$2F,$87,$E5,$1D,$DA,$A3,$4F,$F9,$DF,$07,$7E,$A7,$33
	.dc.b	$E8,$38,$56,$4E,$75,$C2,$3D,$F1,$E7,$23,$E4,$7F,$67,$F4,$3F,$CF
	.dc.b	$23,$FD,$0C,$D7,$B3,$73,$F8,$9E,$67,$A2,$31,$9F,$F3,$66,$D7,$F7
	.dc.b	$C6,$AC,$3F,$C9,$2F,$6E,$7F,$09,$1F,$E1,$13,$E1,$7F,$D9,$BF,$16
	.dc.b	$FA,$1F,$8E,$D8,$FF,$4E,$35,$F8,$97,$FE,$3E,$73,$3C,$08,$FC,$DE
	.dc.b	$AA,$FF,$B4,$2B,$93,$E0,$9F,$E2,$F1,$88,$FE,$A7,$EA,$7F,$97,$DF
	.dc.b	$F5,$63,$5F,$A8,$DD,$3F,$3B,$C9,$FE,$6F,$9D,$F8,$DF,$80,$AD,$6B
	.dc.b	$D2,$3F,$07,$5B,$AE,$71,$FC,$A2,$BB,$47,$FB,$C1,$AF,$13,$F9,$87
	.dc.b	$99,$F8,$6C,$3F,$5A,$DC,$3F,$B5,$B6,$7F,$C6,$3F,$41,$FF,$00,$1F
	.dc.b	$F6,$34,$65,$34,$C4,$6E,$3F,$25,$1F,$94,$7F,$BB,$9C,$38,$7F,$9F
	.dc.b	$C6,$63,$C3,$1F,$D0,$A7,$D3,$1E,$2C,$70,$47,$FA,$57,$D0,$7E,$C9
	.dc.b	$FF,$6D,$35,$FD,$13,$71,$FE,$17,$5F,$32,$7F,$C9,$B3,$B7,$CA,$D7
	.dc.b	$FC,$1A,$D7,$7C,$DF,$D5,$BE,$7D,$EB,$D2,$4F,$F5,$D8,$CB,$FD,$7F
	.dc.b	$E7,$FF,$CF,$4D,$F3,$3F,$AF,$C6,$23,$E2,$87,$0F,$C8,$F9,$F7,$B9
	.dc.b	$FD,$99,$B1,$FD,$AD,$AF,$26,$BF,$43,$E7,$1F,$3D,$FC,$31,$B1,$FF
	.dc.b	$00,$5A,$FF,$09,$F9,$DF,$D9,$BF,$C2,$E3,$2F,$39,$FE,$EF,$18,$8F
	.dc.b	$F1,$9B,$5F,$F3,$CB,$D3,$7F,$E7,$74,$F9,$EF,$E3,$AD,$AF,$15,$E3
	.dc.b	$59,$1F,$E8,$2B,$E3,$1C,$3E,$3B,$33,$E7,$8B,$3D,$A7,$F9,$4F,$CF
	.dc.b	$3A,$6F,$CB,$CC,$3A,$74,$7F,$92,$AB,$51,$83,$FD,$22,$E1,$FE,$58
	.dc.b	$A7,$94,$62,$BF,$9B,$FC,$F7,$F1,$C2,$78,$1E,$2F,$98,$CB,$FE,$1E
	.dc.b	$8F,$E3,$55,$FE,$5A,$6B,$FE,$A0,$6A,$7F,$93,$E0,$F9,$EF,$69,$D2
	.dc.b	$7D,$03,$E2,$7F,$65,$AF,$F6,$A6,$E9,$FE,$68,$6C,$FF,$AB,$97,$CF
	.dc.b	$7F,$50,$73,$1C,$FD,$7C,$19,$9D,$79,$DB,$E9,$95,$E2,$BF,$F1,$13
	.dc.b	$67,$2F,$F5,$58,$FD,$06,$3E,$59,$FE,$BF,$F3,$AF,$82,$A4,$BE,$9E
	.dc.b	$B5,$7F,$4F,$3B,$7F,$C9,$8D,$73,$3F,$E4,$C8,$C5,$FC,$3F,$92,$F7
	.dc.b	$1C,$DF,$F0,$7F,$CF,$7F,$07,$DB,$FD,$1A,$FF,$B7,$F0,$F9,$AD,$7F
	.dc.b	$B7,$3B,$7F,$DA,$ED,$4E,$E7,$AF,$BD,$FD,$F2,$34,$7E,$BA,$3C,$3E
	.dc.b	$6E,$4D,$7C,$BD,$BF,$0C,$57,$C3,$1F,$23,$E7,$64,$7E,$0B,$34,$77
	.dc.b	$EB,$CC,$FC,$88,$F8,$84,$7E,$4B,$E5,$F9,$E9,$1C,$E3,$FC,$27,$F3
	.dc.b	$DF,$A3,$CB,$FC,$6A,$FE,$8E,$7F,$22,$D7,$FC,$24,$EF,$5F,$F5,$BB
	.dc.b	$63,$FD,$FA,$EE,$79,$F6,$76,$EC,$EF,$D9,$2F,$AB,$E2,$F2,$B1,$1D
	.dc.b	$A2,$DF,$92,$7F,$C5,$3F,$3C,$F9,$3A,$4B,$FD,$3F,$CF,$7E,$7A,$75
	.dc.b	$1F,$AA,$6B,$FE,$2B,$6D,$FF,$95,$27,$85,$73,$AD,$3F,$E3,$88,$CC
	.dc.b	$73,$8D,$4F,$1E,$CB,$AF,$FB,$6B,$0E,$AD,$AE,$53,$AB,$0A,$DB,$86
	.dc.b	$AC,$3C,$2F,$90,$F1,$1F,$02,$7A,$3C,$EA,$CF,$0C,$3F,$CE,$37,$D0
	.dc.b	$EB,$19,$C3,$CD,$7D,$0F,$F2,$84,$65,$F2,$5F,$4C,$3B,$DF,$38,$F6
	.dc.b	$CF,$1C,$3B,$C6,$E3,$18,$76,$7D,$47,$6B,$E7,$3B,$8C,$D7,$27,$89
	.dc.b	$F2,$30,$F2,$3E,$F3,$27,$C7,$AE,$6E,$23,$78,$79,$E3,$E4,$1C,$1C
	.dc.b	$3E,$33,$17,$E3,$BF,$14,$79,$30,$F5,$47,$E0,$63,$67,$7C,$3C,$2E
	.dc.b	$CF,$43,$F8,$D7,$FC,$53,$7F,$94,$6D,$FF,$BA,$23,$11,$F1,$CC,$9F
	.dc.b	$EE,$F7,$2F,$D4,$31,$1F,$97,$F9,$29,$1E,$43,$ED,$B8,$7C,$87,$C9
	.dc.b	$32,$7E,$A9,$F4,$DF,$E9,$5E,$0F,$C2,$26,$1F,$9A,$9F,$CB,$4E,$A3
	.dc.b	$E2,$92,$3E,$C4,$65,$F4,$1E,$2C,$51,$98,$F2,$57,$67,$89,$E5,$3D
	.dc.b	$8F,$95,$C1,$EC,$C3,$F3,$B3,$CD,$FB,$B3,$B9,$F7,$9B,$C3,$C1,$7E
	.dc.b	$28,$EB,$87,$B6,$31,$1C,$2B,$C8,$F9,$A7,$58,$77,$BF,$D8,$B6,$7F
	.dc.b	$24,$D7,$E0,$36,$B9,$DF,$67,$CC,$76,$C3,$D7,$5E,$E8,$C5,$FA,$2F
	.dc.b	$EB,$E1,$E2,$AD,$3F,$45,$F3,$1F,$28,$F5,$39,$BE,$B1,$83,$EA,$DF
	.dc.b	$92,$BC,$D5,$F0,$FA,$BF,$92,$12,$BF,$98,$F6,$57,$F8,$4C,$5B,$FE
	.dc.b	$5D,$3F,$CA,$FA,$FF,$96,$75,$CB,$FE,$73,$19,$7F,$9D,$C6,$5F,$E2
	.dc.b	$71,$FE,$EA,$F9,$D7,$CF,$52,$4F,$F8,$A3,$E2,$BF,$BC,$4E,$A3,$FC
	.dc.b	$E6,$1C,$3F,$E7,$66,$BF,$E9,$C6,$BF,$9D,$77,$8B,$3F,$D0,$53,$E8
	.dc.b	$BF,$DF,$AD,$BF,$F1,$CD,$F0,$AD,$F6,$6B,$B2,$57,$9E,$38,$98,$7F
	.dc.b	$E0,$05,$A7,$31,$ED,$EB,$EF,$E3,$48,$C1,$0E,$2E,$5E,$71,$FB,$7C
	.dc.b	$FD,$51,$FD,$4B,$E8,$DF,$19,$F1,$A1,$18,$7E,$4B,$5F,$F4,$EB,$5F
	.dc.b	$F6,$1F,$D0,$3E,$83,$FD,$91,$1F,$C1,$63,$1D,$3F,$49,$78,$2F,$8A
	.dc.b	$C5,$F0,$BF,$EC,$BF,$76,$F9,$1F,$3C,$7F,$B4,$23,$1C,$5F,$71,$B1
	.dc.b	$FE,$B0,$6B,$FC,$57,$E4,$5F,$13,$FE,$B0,$AF,$F4,$24,$63,$E1,$7E
	.dc.b	$1F,$07,$C4,$8D,$FA,$D8,$AF,$EE,$F1,$B7,$8D,$F3,$C1,$7C,$61,$87
	.dc.b	$19,$5F,$4B,$E3,$39,$8F,$2F,$D0,$5B,$7D,$05,$8C,$36,$FC,$2D,$C3
	.dc.b	$C3,$1E,$6A,$E2,$FF,$A2,$BE,$7D,$F3,$FE,$3E,$07,$D0,$7F,$93,$2B
	.dc.b	$53,$98,$C6,$1D,$0F,$F0,$BD,$F2,$AF,$D5,$B6,$3F,$DB,$ED,$7F,$A2
	.dc.b	$07,$0F,$F4,$77,$86,$BF,$71,$B3,$FE,$24,$F9,$07,$C2,$3E,$13,$E5
	.dc.b	$61,$CF,$2F,$8D,$E4,$7C,$76,$E1,$F5,$1B,$3E,$9E,$0F,$C0,$C6,$33
	.dc.b	$FA,$CF,$CF,$F5,$25,$FD,$CF,$A0,$B8,$71,$E0,$FF,$59,$4F,$E0,$2B
	.dc.b	$EB,$56,$B0,$FC,$AF,$CF,$BB,$BF,$01,$0F,$A0,$FC,$33,$A8,$CC,$63
	.dc.b	$0F,$ED,$55,$C5,$FE,$2E,$DC,$BE,$58,$70,$FC,$E6,$BC,$AB,$E3,$C7
	.dc.b	$1B,$FD,$FF,$07,$CC,$F2,$F3,$5F,$33,$E8,$3F,$CD,$98,$FF,$26,$6E
	.dc.b	$3E,$06,$BF,$71,$AF,$F8,$07,$E8,$1F,$3F,$FB,$68,$C3,$5F,$F2,$FC
	.dc.b	$6A,$BF,$99,$B6,$3D,$BC,$EF,$1F,$84,$7F,$05,$6C,$66,$BF,$9D,$B6
	.dc.b	$7E,$2F,$CF,$BC,$3E,$2D,$AF,$E8,$2D,$9D,$CF,$6F,$9F,$FF,$6B,$57
	.dc.b	$F4,$66,$CF,$D0,$6B,$F9,$C6,$BF,$E2,$17,$F2,$D1,$C7,$A3,$A1,$1F
	.dc.b	$EA,$3C,$EF,$8F,$FE,$09,$0F,$9E,$79,$7F,$9E,$AD,$4F,$FA,$1B,$93
	.dc.b	$F9,$BB,$B9,$D5,$F2,$BE,$33,$B3,$F9,$04,$66,$77,$1E,$28,$F9,$B7
	.dc.b	$DB,$2F,$C6,$B5,$FD,$CC,$BD,$1F,$3F,$94,$F8,$EF,$D9,$CE,$4F,$05
	.dc.b	$7F,$65,$9D,$46,$23,$FB,$45,$72,$AE,$2D,$9D,$1C,$E7,$F3,$ED,$7C
	.dc.b	$71,$F8,$48,$C4,$7F,$94,$7E,$79,$EF,$FC,$33,$F8,$3B,$89,$F2,$9F
	.dc.b	$3E,$32,$6D,$F3,$46,$E7,$F9,$04,$6C,$CC,$FE,$52,$75,$3F,$C7,$9A
	.dc.b	$F8,$E3,$F3,$3C,$1F,$D0,$AB,$C3,$AB,$E6,$FF,$8D,$CE,$6F,$BC,$E6
	.dc.b	$BE,$44,$E7,$7F,$C0,$CE,$6B,$F9,$A1,$F9,$FF,$1F,$F0,$58,$74,$7C
	.dc.b	$24,$C3,$FC,$43,$87,$7A,$FE,$6B,$D7,$F8,$1C,$7F,$C0,$7C,$AF,$9D
	.dc.b	$7C,$F7,$F3,$0A,$D6,$3F,$ED,$8A,$D4,$E7,$8F,$FA,$65,$F0,$AF,$C0
	.dc.b	$5F,$0C,$7F,$BF,$4E,$5F,$4C,$7E,$3D,$FF,$45,$35,$FE,$85,$5F,$C0
	.dc.b	$A3,$10,$5F,$11,$FC,$BB,$E7,$BB,$4E,$63,$15,$E9,$6E,$1E,$46,$CF
	.dc.b	$8B,$6F,$77,$C8,$FB,$DB,$5E,$A7,$0F,$E6,$E7,$FD,$F3,$F3,$DF,$EF
	.dc.b	$AF,$9D,$7C,$E7,$AA,$FF,$AB,$D6,$84,$E9,$FE,$BD,$1F,$E3,$28,$C4
	.dc.b	$7F,$AA,$EB,$A5,$73,$BB,$3F,$DB,$26,$BF,$0E,$F9,$57,$F4,$69,$CE
	.dc.b	$9F,$87,$C3,$AD,$F8,$DE,$F1,$9A,$E1,$3E,$67,$C0,$D3,$C0,$53,$C1
	.dc.b	$87,$89,$E9,$1C,$2B,$BC,$2F,$C7,$87,$9A,$39,$CF,$8A,$31,$1F,$04
	.dc.b	$ED,$87,$95,$DD,$F5,$3C,$F5,$C7,$0F,$14,$F7,$38,$DF,$47,$E9,$1F
	.dc.b	$E8,$2C,$3F,$CB,$CE,$CC,$1F,$DB,$BE,$77,$F9,$07,$93,$FE,$88,$6C
	.dc.b	$67,$A3,$FC,$96,$D9,$D7,$BF,$FE,$42,$8E,$4D,$7F,$C4,$74,$8C,$47
	.dc.b	$FB,$A6,$31,$C9,$C8,$FB,$BF,$3C,$F9,$9F,$F7,$76,$BF,$86,$8C,$5F
	.dc.b	$E0,$AF,$8B,$F5,$5B,$3B,$AE,$3B,$7F,$BD,$EF,$8D,$EF,$1F,$45,$FE
	.dc.b	$4B,$0E,$37,$BC,$38,$D8,$61,$CA,$F8,$61,$CA,$FD,$8D,$AE,$15,$B9
	.dc.b	$E9,$1E,$99,$D2,$C3,$C2,$F0,$C3,$C2,$F8,$1B,$1A,$8E,$13,$CA,$FB
	.dc.b	$47,$13,$B6,$31,$FC,$D4,$69,$E8,$67,$1F,$90,$FF,$59,$8F,$8F,$3F
	.dc.b	$E9,$26,$ED,$EE,$8F,$AF,$8F,$FA,$B6,$3D,$B5,$FC,$DA,$B6,$FC,$9C
	.dc.b	$7D,$1F,$3B,$FE,$86,$8F,$EC,$CD,$7F,$9E,$B7,$0F,$F5,$0F,$CE,$38
	.dc.b	$BB,$57,$D3,$AD,$63,$F5,$61,$18,$75,$D5,$C1,$5F,$37,$C3,$F0,$D9
	.dc.b	$A3,$4E,$12,$F9,$EC,$57,$02,$1F,$D5,$23,$4E,$32,$FF,$5C,$DF,$48
	.dc.b	$C6,$3E,$7A,$DC,$7F,$5C,$F9,$EF,$E7,$AF,$F8,$25,$AE,$79,$FC,$7E
	.dc.b	$47,$AE,$73,$8F,$FB,$19,$B8,$7C,$36,$DF,$E6,$70,$FF,$27,$3D,$70
	.dc.b	$E4,$47,$FC,$31,$1A,$7F,$B7,$FC,$53,$E7,$BE,$53,$63,$FC,$3C,$D7
	.dc.b	$E1,$75,$EA,$64,$69,$F9,$71,$88,$EB,$17,$2F,$BF,$3F,$5F,$D6,$FF
	.dc.b	$18,$57,$F8,$AD,$AF,$F8,$09,$B3,$FE,$E6,$E4,$FF,$39,$CE,$FC,$5E
	.dc.b	$44,$7F,$C6,$91,$A7,$94,$65,$A6,$BC,$DF,$C4,$79,$5F,$66,$3F,$C2
	.dc.b	$2D,$BF,$F7,$AB,$77,$FC,$97,$CF,$7F,$9C,$EB,$FC,$85,$39,$7F,$85
	.dc.b	$35,$FF,$5F,$72,$7E,$56,$78,$47,$68,$F0,$91,$FF,$26,$46,$AD,$7E
	.dc.b	$0C,$79,$12,$BF,$CA,$BE,$73,$EE,$FF,$49,$0C,$3E,$4D,$FE,$A7,$E7
	.dc.b	$FF,$49,$3F,$E2,$E6,$E3,$DA,$3F,$91,$57,$07,$F9,$24,$EA,$B9,$74
	.dc.b	$FE,$BE,$31,$3C,$32,$FD,$97,$CF,$FF,$B1,$E3,$FC,$CC,$D7,$FC,$7A
	.dc.b	$DC,$7C,$10,$FD,$C5,$F5,$3F,$BB,$FA,$1F,$F3,$B4,$FF,$9B,$DA,$FF
	.dc.b	$91,$9B,$5C,$AF,$A6,$7F,$C1,$FE,$85,$F0,$7F,$C1,$5B,$87,$F9,$E5
	.dc.b	$AF,$3F,$1B,$F8,$57,$D0,$BE,$7E,$92,$7F,$D0,$1F,$41,$FA,$9C,$FD
	.dc.b	$35,$B9,$FF,$42,$35,$FF,$2D,$B6,$3D,$F3,$C7,$2F,$E3,$93,$BA,$FE
	.dc.b	$3C,$1C,$3A,$BF,$F7,$3B,$FD,$5E,$32,$FF,$59,$AF,$82,$DF,$6F,$F9
	.dc.b	$24,$62,$0B,$E3,$73,$32,$7D,$AF,$08,$FE,$59,$F3,$EE,$FB,$F3,$12
	.dc.b	$FF,$D4,$53,$96,$CE,$5F,$F3,$DC,$27,$9C,$FF,$37,$FA,$0F,$E3,$34
	.dc.b	$69,$8A,$7C,$CE,$FC,$49,$5F,$EB,$1A,$D4,$E2,$3E,$49,$CE,$7E,$3B
	.dc.b	$4F,$93,$3F,$CC,$1F,$9F,$F0,$BF,$AF,$DF,$FA,$73,$63,$FC,$08,$D4
	.dc.b	$BE,$7F,$FA,$83,$92,$7B,$38,$8C,$FA,$1A,$E4,$97,$FE,$C9,$BF,$F1
	.dc.b	$43,$67,$2F,$FA,$8A,$3F,$3B,$87,$92,$7F,$AF,$7C,$83,$E6,$29,$2F
	.dc.b	$A7,$CC,$57,$FD,$57,$3B,$7F,$DB,$0D,$7B,$46,$F9,$FF,$67,$39,$6D
	.dc.b	$7B,$A7,$FB,$77,$CF,$7F,$00,$B3,$FE,$D6,$7F,$DC,$8D,$7F,$D7,$6E
	.dc.b	$C9,$EF,$F2,$39,$F8,$5D,$3C,$F9,$5C,$9E,$FF,$2D,$88,$E7,$17,$2F
	.dc.b	$E1,$38,$2B,$E2,$FC,$EC,$BF,$7B,$E8,$EA,$C4,$7C,$48,$B7,$CA,$6E
	.dc.b	$FF,$E0,$FD,$95,$8B,$FF,$23,$63,$FA,$BF,$9D,$FD,$04,$7F,$B6,$1F
	.dc.b	$CF,$3F,$E4,$BC,$7F,$8E,$C6,$23,$5D,$3F,$E1,$78,$E3,$3A,$EA,$C1
	.dc.b	$99,$D7,$4C,$BE,$99,$FF,$88,$F0,$FD,$2C,$60,$FF,$2D,$3F,$D4,$4C
	.dc.b	$BE,$59,$E4,$F9,$FC,$AF,$F1,$47,$CE,$3E,$42,$92,$7A,$0F,$9D,$FD
	.dc.b	$14,$EE,$32,$D7,$FD,$F2,$DD,$3F,$9E,$CF,$0A,$E9,$5A,$7F,$C7,$11
	.dc.b	$98,$E7,$1A,$9E,$3D,$77,$3F,$F2,$16,$1D,$5B,$5C,$A7,$57,$FE,$7B
	.dc.b	$AD,$36,$74,$F6,$21,$E3,$3D,$F8,$78,$DC,$13,$0F,$1B,$C1,$FE,$B3
	.dc.b	$5D,$A0,$D9,$E5,$3A,$AF,$2D,$7C,$BB,$B7,$C4,$E0,$E2,$77,$7C,$B0
	.dc.b	$F7,$DF,$85,$FF,$32,$46,$63,$DC,$64,$FF,$4E,$39,$7C,$B3,$D2,$BB
	.dc.b	$46,$CF,$21,$6F,$C9,$3F,$22,$3E,$15,$F8,$9A,$76,$AE,$6E,$0E,$65
	.dc.b	$AE,$6F,$CE,$8D,$E1,$F1,$63,$07,$86,$FC,$CD,$3C,$D8,$7A,$EF,$9C
	.dc.b	$2B,$EA,$BD,$EB,$8B,$4E,$37,$E0,$C3,$D5,$5C,$0F,$14,$EC,$E9,$1F
	.dc.b	$3A,$78,$47,$CE,$32,$7C,$13,$CC,$7C,$1C,$3C,$D7,$E2,$BE,$A4,$BF
	.dc.b	$09,$D6,$B8,$C6,$5E,$87,$8D,$C3,$F3,$1B,$87,$C2,$7D,$53,$D2,$7C
	.dc.b	$D1,$C2,$FC,$71,$F0,$8E,$F3,$B7,$A3,$6B,$A3,$E5,$8F,$13,$B9,$E7
	.dc.b	$1E,$FC,$3E,$2B,$F8,$E3,$ED,$9C,$0F,$1C,$F3,$BF,$4E,$1F,$16,$3C
	.dc.b	$0D,$9D,$90,$E8,$7D,$03,$CE,$67,$0F,$3B,$F2,$2B,$AB,$EF,$3C,$47
	.dc.b	$27,$E3,$BC,$67,$C0,$FF,$1F,$E8,$FE,$2A,$47,$F8,$2F,$55,$7C,$59
	.dc.b	$7F,$7A,$6C,$FF,$1C,$9F,$E3,$BD,$1F,$C7,$BA,$65,$7F,$0E,$8C,$BF
	.dc.b	$C4,$23,$2F,$48,$FF,$6E,$7C,$DB,$E7,$69,$27,$F9,$AF,$C5,$FF,$73
	.dc.b	$9D,$47,$F3,$86,$BF,$1D,$BA,$7F,$9C,$27,$85,$73,$AD,$3F,$EE,$B8
	.dc.b	$FE,$FC,$7F,$95,$E0,$1F,$EE,$0F,$8B,$A7,$87,$54,$8C,$10,$FB,$13
	.dc.b	$CD,$E5,$D5,$DF,$C2,$91,$EC,$8F,$11,$87,$8C,$61,$FE,$AB,$F7,$8F
	.dc.b	$88,$FF,$1F,$B6,$3F,$CC,$ED,$7F,$C8,$AD,$7F,$CD,$FF,$3B,$EB,$7F
	.dc.b	$CD,$AD,$7F,$C9,$6D,$77,$87,$F9,$33,$E7,$5F,$3B,$FE,$BC,$9F,$8A
	.dc.b	$23,$3D,$1F,$E7,$85,$E0,$78,$67,$E5,$C6,$9F,$F4,$0F,$C7,$BE,$1B
	.dc.b	$FD,$33,$D7,$FD,$F6,$31,$F0,$5F,$EB,$7E,$4F,$F5,$C4,$6F,$CF,$C5
	.dc.b	$7D,$68,$DB,$C6,$F9,$E1,$FC,$62,$72,$D8,$CD,$F4,$58,$74,$95,$79
	.dc.b	$7E,$6F,$0F,$9F,$E6,$ED,$A6,$C6,$1B,$8F,$8D,$B7,$FB,$BF,$9A,$7C
	.dc.b	$D7,$7E,$47,$CD,$FF,$8C,$B0,$FF,$09,$D6,$9F,$F1,$04,$EB,$C9,$F3
	.dc.b	$60,$AD,$3F,$CD,$A7,$55,$CB,$0E,$D8,$75,$AF,$14,$7F,$B5,$1F,$B7
	.dc.b	$7C,$9F,$F5,$AB,$5C,$C6,$DF,$8B,$9F,$F8,$AE,$FE,$84,$FC,$4F,$0B
	.dc.b	$E2,$B6,$32,$7D,$58,$CB,$F5,$FC,$CF,$96,$7C,$CF,$9B,$B1,$92,$61
	.dc.b	$F6,$63,$2F,$F2,$5F,$9A,$76,$7D,$9E,$37,$CD,$FD,$EC,$3F,$2F,$5A
	.dc.b	$7F,$C9,$53,$AE,$5F,$97,$39,$C3,$FA,$FC,$E6,$BF,$C8,$B3,$98,$FC
	.dc.b	$A5,$F5,$AF,$E0,$F3,$9A,$EF,$87,$E2,$FE,$45,$F0,$AF,$9B,$EF,$3F
	.dc.b	$6B,$E6,$FE,$56,$3F,$CF,$FF,$03,$78,$F9,$E1,$FB,$F1,$8C,$FF,$CA
	.dc.b	$FF,$0A,$F8,$7F,$C1,$DF,$8F,$04,$E6,$BF,$CB,$B3,$9E,$87,$8D,$F8
	.dc.b	$A5,$F4,$B6,$30,$FE,$3B,$E7,$5E,$1F,$9F,$EB,$F0,$DC,$7F,$CD,$3F
	.dc.b	$3D,$FE,$B7,$7F,$CD,$75,$C1,$CF,$9D,$C1,$DF,$C5,$FE,$6A,$B8,$6D
	.dc.b	$FD,$7D,$46,$6B,$F3,$FE,$23,$E3,$7F,$6D,$83,$E7,$3F,$9B,$3F,$E7
	.dc.b	$A9,$D3,$FE,$87,$35,$E5,$F2,$AE,$35,$C3,$2F,$ED,$D7,$C1,$FF,$0F
	.dc.b	$DF,$03,$A5,$70,$BE,$51,$D6,$F9,$76,$A3,$F9,$28,$F4,$BC,$FF,$3C
	.dc.b	$F8,$BF,$DC,$CE,$4F,$34,$65,$FF,$49,$99,$CB,$C7,$19,$7F,$C0,$91
	.dc.b	$9B,$F0,$47,$86,$75,$18,$8F,$25,$F8,$AF,$C2,$DB,$FF,$50,$FC,$F3
	.dc.b	$E0,$3E,$29,$FC,$39,$C4,$7B,$E3,$06,$9E,$2D,$BF,$F0,$C4,$F0,$7E
	.dc.b	$5E,$57,$1F,$24,$DB,$FE,$DE,$7D,$BB,$FE,$3B,$C5,$FC,$D2,$3E,$77
	.dc.b	$CD,$E3,$CE,$FB,$2D,$8C,$9F,$EB,$A9,$E1,$97,$82,$74,$D9,$D5,$72
	.dc.b	$7F,$D8,$3E,$17,$E0,$72,$F9,$7D,$2F,$9D,$7C,$F7,$F3,$0A,$D5,$FF
	.dc.b	$81,$AB,$53,$9E,$3F,$E7,$97,$C2,$BA,$C6,$27,$51,$FC,$F2,$75,$97
	.dc.b	$F8,$FF,$0E,$AF,$F7,$F8,$CE,$3F,$6E,$5F,$0D,$FC,$7F,$E7,$B7,$69
	.dc.b	$CE,$1F,$C7,$A7,$31,$89,$FE,$3B,$18,$AE,$59,$7B,$75,$F7,$7C,$2D
	.dc.b	$C7,$E2,$B7,$0F,$34,$FF,$52,$7F,$C4,$FF,$3D,$FE,$28,$F9,$D7,$CF
	.dc.b	$7F,$81,$B4,$5E,$ED,$7B,$AD,$4F,$E6,$AF,$8D,$7F,$83,$8D,$87,$2F
	.dc.b	$BE,$38,$4E,$E3,$73,$AB,$F0,$E3,$DE,$7A,$61,$5E,$51,$5F,$59,$DE
	.dc.b	$FC,$B1,$98,$EB,$18,$74,$74,$29,$D0,$A7,$4B,$F2,$61,$E8,$8E,$11
	.dc.b	$8A,$ED,$7E,$5C,$3D,$35,$C8,$F0,$9F,$E5,$67,$35,$F1,$23,$B6,$1E
	.dc.b	$58,$C4,$70,$3C,$8E,$2B,$B6,$1E,$39,$FF,$1E,$C7,$CB,$69,$4F,$A6
	.dc.b	$65,$D7,$CE,$FF,$1F,$F4,$AF,$47,$F8,$F6,$BE,$A4,$E5,$B1,$97,$EF
	.dc.b	$43,$E0,$7F,$DB,$D1,$C1,$AE,$9A,$FF,$94,$A3,$87,$0F,$8A,$78,$E5
	.dc.b	$F8,$6F,$9E,$7C,$CF,$F9,$62,$30,$DD,$79,$57,$06,$DF,$E1,$6F,$8E
	.dc.b	$36,$F8,$3F,$E8,$6C,$7C,$2F,$E0,$1B,$8F,$68,$E0,$7E,$3E,$33,$87
	.dc.b	$68,$FC,$3D,$F6,$6D,$F6,$55,$C1,$F0,$63,$E7,$7F,$D2,$F5,$C7,$0F
	.dc.b	$04,$78,$E7,$4F,$26,$CE,$F0,$FF,$3D,$B6,$7A,$E3,$E7,$AD,$E3,$E3
	.dc.b	$AD,$57,$3A,$F6,$BF,$ED,$78,$F5,$47,$F5,$16,$EF,$EF,$C7,$CB,$39
	.dc.b	$BF,$F4,$5B,$6F,$FD,$62,$FF,$59,$CB,$E0,$43,$E7,$7F,$CE,$EE,$B0
	.dc.b	$FF,$49,$C6,$23,$FD,$D0,$DA,$FE,$C5,$F3,$8E,$2E,$F7,$C3,$A6,$E1
	.dc.b	$FE,$C0,$C3,$84,$E6,$3B,$1F,$EC,$4B,$E4,$E0,$F2,$7C,$47,$26,$99
	.dc.b	$A4,$66,$5F,$3D,$B9,$CD,$6E,$B0,$6E,$6C,$64,$83,$2F,$6C,$78,$7E
	.dc.b	$7B,$FA,$2C,$F9,$5A,$EF,$1F,$F5,$CC,$62,$BE,$2F,$97,$C7,$A6,$5F
	.dc.b	$17,$D1,$18,$8E,$47,$FB,$9A,$B9,$5F,$4A,$DC,$FC,$5F,$8D,$7C,$F7
	.dc.b	$C5,$6C,$66,$3E,$27,$5E,$59,$19,$3E,$76,$DF,$E4,$17,$07,$88,$57
	.dc.b	$47,$0D,$CF,$F0,$37,$F5,$7D,$6F,$F1,$95,$7D,$96,$BC,$DB,$3F,$E2
	.dc.b	$9E,$3F,$F3,$74,$EF,$C5,$E4,$47,$F0,$51,$A7,$94,$65,$A6,$BD,$1F
	.dc.b	$C3,$79,$7F,$80,$8F,$53,$6F,$FC,$70,$DD,$BF,$DD,$FF,$3B,$FE,$72
	.dc.b	$AF,$C8,$4E,$51,$FE,$68,$E3,$EA,$7E,$4E,$78,$47,$68,$CB,$C9,$D7
	.dc.b	$8B,$BC,$A4,$FE,$76,$BE,$1E,$1D,$BD,$27,$CF,$7E,$9A,$BC,$35,$A7
	.dc.b	$C7,$F3,$DF,$A6,$7E,$B3,$74,$FF,$2A,$4E,$E5,$39,$C3,$D7,$18,$EA
	.dc.b	$FD,$64,$E6,$3B,$E5,$FA,$FF,$9B,$FF,$60,$C7,$EC,$DA,$FE,$15,$B8
	.dc.b	$FF,$98,$FE,$79,$F3,$DF,$B7,$F9,$EF,$F3,$B5,$FE,$ED,$AF,$E3,$5B
	.dc.b	$5F,$E6,$BB,$E7,$9F,$9B,$E7,$9F,$03,$FC,$05,$B8,$7F,$05,$6B,$CF
	.dc.b	$C6,$FE,$0D,$F3,$9E,$B8,$3F,$E7,$17,$F8,$63,$5F,$CF,$3B,$27,$A1
	.dc.b	$C8,$E7,$C1,$27,$8C,$65,$E7,$1F,$C5,$7E,$75,$F3,$DF,$DE,$DB,$3F
	.dc.b	$DF,$DA,$FF,$8A,$DA,$FF,$8C,$DC,$EB,$FE,$8D,$C1,$F4,$BE,$C3,$B5
	.dc.b	$FB,$BE,$0A,$61,$EC,$7E,$0F,$C9,$E2,$3D,$B1,$70,$E9,$3F,$CA,$FE
	.dc.b	$7F,$F7,$B5,$FC,$BD,$7C,$E6,$4D,$78,$52,$75,$19,$78,$46,$5F,$E1
	.dc.b	$13,$C9,$D9,$A3,$F9,$E7,$CF,$FF,$28,$7F,$CA,$4D,$7F,$2A,$D7,$F6
	.dc.b	$2D,$4F,$F2,$47,$F6,$B0,$FA,$5F,$61,$FD,$1E,$75,$2F,$8C,$FD,$BE
	.dc.b	$5F,$EB,$96,$F2,$FE,$91,$B3,$FC,$01,$7D,$07,$F5,$17,$31,$CF,$C0
	.dc.b	$C1,$99,$D7,$A5,$BE,$69,$5F,$D6,$EF,$FC,$64,$D9,$CB,$FC,$BE,$3F
	.dc.b	$47,$87,$8E,$79,$CF,$1A,$DE,$3E,$99,$FE,$C9,$F2,$AF,$82,$FF,$25
	.dc.b	$36,$FF,$C9,$CD,$7F,$CF,$4D,$7F,$D0,$4E,$6B,$E2,$7D,$2F,$BC,$EB
	.dc.b	$5F,$DC,$7E,$87,$F5,$87,$EA,$67,$FD,$C4,$DC,$BF,$86,$BA,$E9,$E1
	.dc.b	$D5,$32,$FE,$F9,$3C,$BD,$0E,$84,$8C,$FC,$DD,$7F,$A9,$4A,$71,$8F
	.dc.b	$82,$AE,$3E,$09,$B7,$FC,$E2,$D5,$F1,$BE,$57,$D5,$F4,$B2,$BF,$C1
	.dc.b	$DB,$27,$51,$E6,$9D,$3F,$D8,$E7,$4F,$F8,$57,$E8,$BF,$52,$FF,$67
	.dc.b	$7F,$40,$DE,$9F,$F0,$C4,$6B,$46,$9F,$CC,$A3,$0D,$9D,$9C,$3A,$78
	.dc.b	$FD,$14,$C3,$C0,$F4,$F3,$31,$1E,$18,$B8,$F6,$8F,$EB,$38,$FF,$5B
	.dc.b	$4A,$77,$8F,$8D,$E9,$3D,$88,$75,$3E,$05,$F0,$7A,$C5,$7A,$C5,$AE
	.dc.b	$D1,$CA,$FC,$8E,$8C,$5F,$85,$A7,$A2,$B9,$5F,$36,$C6,$8E,$47,$94
	.dc.b	$95,$DA,$FC,$31,$F5,$67,$8B,$AB,$FA,$51,$8A,$E1,$7F,$0A,$B4,$7B
	.dc.b	$8F,$05,$FC,$B3,$AC,$F0,$8C,$D7,$3B,$F6,$47,$84,$E3,$7E,$B6,$9C
	.dc.b	$0A,$78,$0F,$6D,$75,$71,$7E,$0A,$F4,$C7,$09,$D1,$F0,$6F,$C5,$1E
	.dc.b	$93,$C4,$72,$BF,$13,$4F,$11,$4E,$F1,$C9,$FC,$4C,$7C,$59,$EF,$5E
	.dc.b	$E6,$CF,$D9,$3C,$91,$93,$CE,$F5,$8D,$BE,$E8,$D3,$E7,$8C,$DF,$CF
	.dc.b	$AF,$13,$67,$A9,$DC,$E9,$5C,$9E,$87,$37,$11,$F2,$63,$85,$76,$AF
	.dc.b	$86,$D8,$F8,$6D,$7E,$19,$E0,$3E,$19,$FA,$43,$13,$D5,$F0,$9A,$3C
	.dc.b	$A7,$82,$BC,$75,$E0,$6C,$78,$1A,$F8,$0E,$27,$3A,$FB,$CF,$82,$7E
	.dc.b	$A4,$F0,$9F,$93,$7F,$22,$73,$5F,$0C,$87,$3A,$FA,$B5,$EF,$38,$C7
	.dc.b	$53,$DC,$78,$AF,$DC,$7F,$0D,$FA,$27,$C8,$FF,$AD,$FE,$89,$F2,$94
	.dc.b	$FF,$5D,$4E,$AF,$FD,$7C,$D7,$FD,$21,$3C,$1F,$F6,$13,$5F,$F4,$A4
	.dc.b	$6F,$0F,$E1,$73,$FD,$B2,$78,$F4,$78,$4F,$17,$4D,$8E,$71,$5E,$F1
	.dc.b	$E7,$F9,$2F,$47,$4C,$9F,$0C,$65,$F1,$C6,$5F,$3C,$65,$F4,$C6,$70
	.dc.b	$F2,$CF,$FB,$67,$E8,$9F,$45,$FE,$72,$B9,$18,$70,$7F,$A4,$1C,$38
	.dc.b	$6B,$C2,$3F,$DC,$9F,$1E,$FA,$5A,$77,$8C,$F5,$EB,$AA,$5F,$FB,$B6
	.dc.b	$FF,$B6,$36,$72,$FF,$B0,$63,$FB,$4B,$93,$87,$8D,$E0,$F1,$A4,$60
	.dc.b	$87,$17,$2F,$38,$F3,$98,$D3,$FA,$1D,$2B,$35,$FF,$00,$7E,$81,$F4
	.dc.b	$3F,$EA,$E8,$FF,$4C,$C6,$3C,$5F,$8F,$08,$76,$BE,$3F,$31,$B9,$F9
	.dc.b	$BF,$48,$FA,$5F,$EA,$71,$FE,$A4,$8C,$76,$74,$6C,$70,$9D,$AF,$95
	.dc.b	$7C,$67,$FA,$5A,$FF,$D8,$FB,$FF,$71,$9C,$FC,$47,$F7,$38,$C5,$6E
	.dc.b	$FF,$DF,$F5,$AB,$E2,$7F,$B4,$1E,$BE,$A7,$18,$C5,$30,$C5,$F8,$E3
	.dc.b	$A6,$1D,$A7,$38,$79,$BE,$8A,$D3,$E8,$EC,$61,$B8,$F8,$B1,$E9,$18
	.dc.b	$7C,$F5,$C6,$75,$7D,$5B,$19,$9D,$DF,$59,$E9,$0F,$A2,$7D,$17,$CC
	.dc.b	$8C,$36,$33,$3F,$79,$B3,$FA,$66,$E7,$FE,$0C,$6C,$FF,$BE,$3E,$4D
	.dc.b	$F0,$CF,$84,$F8,$F8,$76,$AF,$3F,$8D,$F1,$32,$F8,$AD,$8C,$9E,$C8
	.dc.b	$CB,$4F,$7F,$A5,$F3,$0F,$9B,$F4,$56,$32,$4B,$F9,$31,$97,$E5,$CD
	.dc.b	$D3,$F0,$B3,$6B,$FD,$EF,$F4,$1F,$CB,$1F,$93,$F4,$0F,$13,$D7,$39
	.dc.b	$C5,$87,$F8,$BA,$B5,$7F,$43,$E1,$DF,$11,$F0,$B1,$FC,$3C,$B0,$95
	.dc.b	$A8,$FA,$F3,$A9,$D7,$5F,$F9,$CE,$7E,$DA,$FA,$0F,$9A,$E6,$7C,$73
	.dc.b	$D5,$FD,$A4,$FD,$2F,$9F,$77,$FA,$BE,$83,$0D,$BF,$C3,$FD,$07,$FA
	.dc.b	$75,$FC,$4D,$70,$73,$E5,$F0,$77,$CB,$F7,$AB,$86,$5E,$23,$DC,$FE
	.dc.b	$0E,$78,$9F,$94,$EB,$7C,$77,$EB,$21,$F3,$EF,$1F,$F1,$35,$A9,$FD
	.dc.b	$31,$A8,$CD,$71,$AE,$13,$D6,$31,$3A,$8E,$33,$AC,$3C,$C7,$F0,$DA
	.dc.b	$DC,$FE,$A0,$79,$DE,$5F,$A0,$7C,$67,$E7,$27,$27,$92,$32,$F1,$9E
	.dc.b	$13,$B6,$C6,$4E,$11,$96,$9F,$B8,$FA,$07,$BD,$ED,$3F,$56,$E2,$3D
	.dc.b	$71,$83,$4F,$16,$E3,$EE,$9E,$0F,$D5,$9F,$9F,$3C,$5E,$67,$F0,$8E
	.dc.b	$1F,$E4,$0F,$C8,$F9,$DF,$DD,$C6,$2B,$96,$89,$CF,$8D,$F4,$DB,$7C
	.dc.b	$4E,$91,$96,$99,$AF,$A7,$E1,$7D,$F7,$F1,$5C,$EF,$9F,$7D,$07,$F0
	.dc.b	$CA,$D7,$77,$FA,$C6,$B5,$39,$E2,$FE,$6F,$18,$C7,$F9,$1D,$F0,$7F
	.dc.b	$5B,$87,$47,$35,$C2,$31,$8F,$92,$31,$87,$87,$E5,$FF,$A1,$E4,$FA
	.dc.b	$0E,$97,$FC,$8E,$73,$18,$86,$49,$CE,$1C,$F2,$FE,$85,$87,$38,$F7
	.dc.b	$6B,$EF,$F8,$4B,$5E,$A6,$E3,$F2,$27,$F9,$7B,$FC,$37,$E8,$3F,$87
	.dc.b	$7C,$F3,$E8,$3F,$82,$69,$F5,$F1,$FE,$8B,$7B,$15,$B7,$F9,$E6,$1F
	.dc.b	$ED,$D8,$C4,$7F,$0D,$C3,$AE,$1D,$2B,$4F,$F8,$B2,$33,$3C,$AE,$BC
	.dc.b	$4E,$F3,$C6,$33,$1E,$3B,$EF,$87,$9E,$36,$7F,$8E,$EF,$9C,$66,$7C
	.dc.b	$B3,$C6,$FA,$3C,$0F,$04,$74,$C3,$CF,$1C,$CD,$57,$3A,$F5,$46,$27
	.dc.b	$6F,$28,$EB,$5E,$08,$5F,$92,$B7,$1E,$33,$E1,$1E,$33,$07,$E1,$1D
	.dc.b	$BC,$8F,$F1,$79,$4F,$F1,$8C,$61,$C1,$FC,$FF,$E7,$FF,$8B,$F9,$DF
	.dc.b	$E3,$66,$C7,$F9,$79,$AF,$F9,$8D,$AF,$1F,$77,$FA,$DB,$B6,$BB,$6B
	.dc.b	$BD,$BB,$47,$0C,$7F,$0B,$F4,$0F,$9B,$FE,$A5,$18,$6E,$9C,$70,$FA
	.dc.b	$98,$B0,$E5,$3F,$D6,$70,$E5,$5F,$88,$6B,$C2,$76,$FF,$6B,$AE,$58
	.dc.b	$EF,$0E,$F3,$FD,$BE,$F8,$E3,$F9,$AB,$DC,$F3,$BE,$0C,$30,$49,$F5
	.dc.b	$39,$79,$61,$E2,$85,$74,$9D,$C7,$E4,$6F,$C3,$8F,$C1,$9E,$38,$73
	.dc.b	$8F,$81,$8F,$A2,$F9,$E3,$D7,$0F,$83,$6D,$B7,$5F,$E3,$E9,$CC,$75
	.dc.b	$C3,$33,$FD,$7A,$72,$FC,$3C,$3F,$C0,$D7,$CE,$3E,$56,$1E,$BF,$A0
	.dc.b	$FF,$30,$47,$F9,$29,$AF,$F6,$86,$CF,$18,$7C,$EB,$87,$AE,$7F,$E1
	.dc.b	$3B,$DB,$7A,$33,$7F,$E2,$39,$CB,$FE,$25,$BE,$4F,$87,$E1,$F6,$A5
	.dc.b	$18,$6C,$7F,$98,$25,$F4,$19,$A9,$3C,$CD,$46,$1F,$64,$5C,$BD,$95
	.dc.b	$CA,$7F,$CD,$7F,$41,$FC,$E5,$BC,$FE,$48,$C4,$7F,$83,$27,$FD,$7E
	.dc.b	$FF,$83,$A3,$35,$F1,$DB,$9F,$F8,$DD,$B3,$F0,$7E,$03,$A6,$12,$3B
	.dc.b	$9E,$D8,$F8,$7F,$14,$FA,$0A,$49,$F8,$5D,$55,$FF,$0F,$CE,$DF,$F4
	.dc.b	$2B,$5F,$44,$6F,$2F,$B9,$39,$6D,$7D,$2C,$FF,$C7,$37,$FE,$8E,$F5
	.dc.b	$FF,$C4,$AD,$F1,$3E,$BC,$62,$3F,$C5,$FC,$9F,$E6,$47,$44,$3F,$C6
	.dc.b	$71,$BC,$3F,$19,$3B,$9C,$D7,$42,$3F,$E9,$E8,$D4,$FE,$4F,$CA,$FC
	.dc.b	$1C,$FF,$8F,$9B,$19,$E0,$FC,$5F,$CF,$FF,$99,$2B,$FD,$4F,$5A,$E1
	.dc.b	$FF,$35,$36,$3F,$CF,$0D,$53,$C5,$ED,$5F,$96,$9D,$1B,$9D,$5F,$E9
	.dc.b	$DF,$75,$FF,$9D,$DB,$7F,$E8,$56,$FA,$2F,$A0,$FD,$2C,$F8,$33,$FD
	.dc.b	$1F,$D0,$7E,$9A,$3F,$CB,$6D,$C7,$B4,$EB,$A2,$57,$FB,$12,$B5,$97
	.dc.b	$86,$BF,$5B,$F3,$AF,$86,$A4,$BF,$F6,$5F,$D0,$3E,$83,$F6,$7F,$41
	.dc.b	$FE,$66,$9F,$F6,$7B,$5F,$F3,$6B,$74,$FE,$0D,$87,$59,$FF,$6A,$FD
	.dc.b	$03,$E1,$3F,$7A,$D8,$CC,$76,$F4,$7F,$80,$FC,$FF,$A2,$7F,$D5,$4F
	.dc.b	$F9,$99,$FF,$6F,$B5,$FF,$3E,$BB,$4D,$35,$87,$F1,$96,$ED,$FE,$A9
	.dc.b	$75,$C1,$E0,$F4,$65,$F6,$8F,$E2,$7F,$40,$F9,$FA,$49,$F1,$F7,$D7
	.dc.b	$FD,$68,$7F,$B0,$67,$83,$83,$13,$FE,$00,$8C,$C6,$70,$F9,$0F,$B3
	.dc.b	$8B,$E1,$9F,$13,$E3,$6C,$7C,$28,$AF,$C2,$8F,$E5,$3F,$3E,$F0,$2F
	.dc.b	$CC,$4A,$FF,$6A,$6F,$FC,$EE,$B5,$18,$DF,$FD,$99,$18,$8F,$F5,$4A
	.dc.b	$AE,$51,$8B,$EE,$62,$FA,$4E,$27,$A1,$FE,$E3,$F9,$EF,$E4,$29,$A6
	.dc.b	$2D,$F3,$3F,$D0,$A3,$F9,$7C,$FC,$C6,$BF,$EE,$16,$BD,$7C,$5F,$F1
	.dc.b	$5F,$C3,$FE,$CB,$3F,$F7,$33,$67,$FD,$E8,$D4,$BE,$7B,$FA,$8B,$92
	.dc.b	$7B,$58,$8C,$F9,$BA,$E7,$97,$EE,$39,$C6,$9C,$5F,$E9,$70,$F1,$D6
	.dc.b	$67,$FD,$AB,$3F,$E2,$57,$71,$D1,$CD,$7F,$64,$F9,$17,$C0,$D2,$4F
	.dc.b	$6F,$85,$AF,$FB,$82,$76,$F9,$5A,$F8,$23,$FB,$7F,$CE,$FE,$A4,$FD
	.dc.b	$2C,$03,$86,$F5,$70,$38,$F5,$C8,$F8,$53,$CF,$AF,$A9,$23,$3F,$29
	.dc.b	$5F,$F7,$AC,$FF,$6A,$E6,$FF,$07,$37,$7F,$14,$F8,$AB,$85,$6A,$ED
	.dc.b	$6A,$D5,$F1,$FE,$76,$61,$F0,$5F,$7F,$56,$23,$E3,$45,$C7,$C3,$3F
	.dc.b	$E1,$2F,$9C,$FD,$14,$7F,$88,$5F,$D1,$57,$DF,$6A,$8D,$75,$71,$39
	.dc.b	$75,$49,$E1,$1D,$7A,$64,$73,$7C,$1E,$5E,$23,$A6,$BF,$0B,$1F,$D2
	.dc.b	$C6,$CF,$14,$EF,$1F,$35,$DB,$E1,$36,$BE,$E5,$F2,$7A,$98,$AF,$0B
	.dc.b	$D4,$C8,$8E,$17,$DA,$3C,$26,$DC,$3C,$23,$77,$E2,$8D,$3E,$BA,$F2
	.dc.b	$47,$5B,$B5,$C2,$3C,$87,$63,$A4,$E8,$F3,$B8,$3A,$9C,$0F,$14,$ED
	.dc.b	$C9,$DC,$B3,$DD,$F7,$4F,$4B,$F5,$92,$FD,$13,$E3,$8C,$3D,$DA,$F5
	.dc.b	$3C,$84,$73,$5F,$E0,$D3,$C0,$4A,$F0,$57,$6A,$E0,$DA,$FC,$95,$FA
	.dc.b	$E7,$27,$DC,$71,$1D,$9B,$3C,$4F,$A5,$3C,$0F,$79,$F0,$A3,$6F,$AE
	.dc.b	$37,$3E,$8A,$FA,$B5,$AA,$E2,$7C,$E3,$E5,$C6,$9A,$75,$29,$E1,$BF
	.dc.b	$85,$1C,$DB,$1A,$AF,$7B,$6F,$D4,$FB,$CC,$1E,$73,$07,$5A,$F1,$34
	.dc.b	$F1,$14,$F1,$5F,$86,$3E,$8C,$73,$3D,$B5,$F2,$DE,$E6,$2B,$E2,$9E
	.dc.b	$47,$07,$1B,$F5,$BD,$E7,$D6,$D9,$E1,$3A,$9F,$9D,$3F,$8D,$9C,$D7
	.dc.b	$AA,$73,$3E,$83,$C7,$5F,$8A,$BF,$54,$71,$3C,$55,$C6,$F9,$C7,$C2
	.dc.b	$BE,$67,$B5,$F2,$1E,$9A,$E6,$FD,$D3,$F8,$1F,$CD,$BE,$43,$FD,$6E
	.dc.b	$D8,$FE,$C4,$D7,$F8,$4F,$CD,$FF,$6F,$8C,$5F,$FB,$2D,$B8,$7F,$3E
	.dc.b	$6D,$7F,$51,$AF,$EE,$C7,$F6,$57,$9F,$3F,$FB,$1A,$31,$7F,$ED,$57
	.dc.b	$FB,$74,$F1,$9D,$11,$FE,$ED,$1A,$4B,$97,$A7,$3C,$9D,$C7,$73,$0F
	.dc.b	$B2,$30,$E8,$EA,$EA,$FC,$93,$FE,$DC,$F9,$B7,$CD,$52,$5F,$F8,$27
	.dc.b	$E2,$7F,$CB,$53,$A8,$FF,$06,$35,$FE,$66,$DB,$FF,$7A,$4F,$F9,$84
	.dc.b	$FE,$D1,$3F,$DE,$5F,$EC,$6F,$5E,$5F,$F7,$6C,$6A,$0D,$36,$FF,$40
	.dc.b	$6C,$F6,$E7,$E3,$CB,$CB,$C2,$97,$E9,$8C,$BE,$B8,$CB,$CE,$32,$F5
	.dc.b	$8C,$D7,$AA,$7F,$E0,$0F,$D2,$79,$2B,$FC,$D0,$1C,$3F,$E6,$B6,$BF
	.dc.b	$E7,$06,$BF,$E0,$8F,$9A,$7C,$DF,$FA,$E6,$30,$D7,$FC,$3D,$0E,$E7
	.dc.b	$C4,$F8,$23,$8F,$87,$FE,$79,$F8,$D7,$C2,$7F,$A6,$FA,$3F,$CB,$B1
	.dc.b	$8F,$80,$46,$27,$59,$7F,$6E,$AD,$4E,$7C,$DD,$CE,$69,$78,$DF,$38
	.dc.b	$78,$2F,$18,$7F,$8D,$63,$9E,$1C,$7E,$63,$2A,$3C,$FF,$33,$87,$D3
	.dc.b	$37,$24,$D8,$C3,$6F,$D4,$DB,$F3,$37,$1F,$2D,$F2,$AF,$F3,$17,$CC
	.dc.b	$BE,$67,$C7,$BB,$E9,$1E,$37,$9B,$35,$EF,$0F,$F4,$3D,$EE,$B5,$39
	.dc.b	$AF,$F5,$53,$5F,$F5,$7B,$5F,$F2,$3B,$5F,$85,$8F,$F1,$7B,$FA,$98
	.dc.b	$FC,$1F,$07,$E1,$36,$32,$7D,$98,$CB,$F6,$FC,$BF,$90,$7C,$9F,$9A
	.dc.b	$B1,$92,$61,$F6,$E3,$2F,$DD,$9B,$AF,$C8,$9B,$5F,$E5,$EF,$9A,$F3
	.dc.b	$5F,$F8,$C3,$E6,$BF,$8A,$4F,$FA,$36,$73,$7F,$88,$A5,$69,$FF,$1B
	.dc.b	$FB,$CF,$9A,$F9,$DE,$1F,$E4,$AF,$84,$E7,$1F,$F4,$C4,$E5,$FF,$37
	.dc.b	$F0,$BB,$FE,$B2,$F9,$AB,$18,$7E,$D7,$CC,$BA,$FE,$E6,$CF,$9A,$CC
	.dc.b	$61,$FC,$07,$CD,$7F,$A6,$B2,$FF,$54,$5F,$17,$FD,$55,$19,$DF,$FC
	.dc.b	$A3,$39,$7F,$CA,$70,$F8,$5E,$7E,$4F,$46,$DF,$FA,$CF,$B5,$F1,$5F
	.dc.b	$A4,$87,$CC,$7F,$20,$75,$19,$8C,$63,$FE,$60,$AE,$35,$FE,$DA,$38
	.dc.b	$F9,$1F,$99,$8C,$CF,$F9,$AA,$72,$74,$38,$57,$4E,$4F,$B1,$5C,$1F
	.dc.b	$E0,$33,$C2,$BF,$CD,$EB,$D7,$F4,$7C,$D3,$E1,$7F,$51,$39,$3C,$91
	.dc.b	$97,$FD,$A4,$66,$BC,$91,$97,$AD,$74,$AE,$6D,$AE,$2D,$9F,$F4,$0F
	.dc.b	$CD,$3D,$9F,$8C,$7E,$E5,$C5,$7B,$A3,$06,$9F,$F7,$24,$6A,$7E,$36
	.dc.b	$77,$2F,$CD,$4D,$AF,$F6,$81,$F5,$AF,$9B,$D4,$FF,$77,$71,$FF,$26
	.dc.b	$7E,$6F,$CB,$7F,$0B,$8C,$63,$DA,$31,$E4,$7D,$D6,$D7,$03,$9C,$65
	.dc.b	$A6,$67,$FD,$43,$F2,$FF,$5E,$3F,$DC,$9D,$4F,$9A,$7C,$D7,$F1,$BA
	.dc.b	$D5,$FF,$4A,$9C,$DF,$FB,$AF,$5F,$DC,$56,$B0,$E5,$87,$1E,$4F,$E4
	.dc.b	$F8,$70,$C7,$C3,$18,$9D,$63,$FC,$9A,$B5,$8F,$4F,$80,$FE,$82,$BE
	.dc.b	$6B,$C7,$8B,$35,$6A,$73,$18,$C3,$CC,$79,$E3,$33,$BC,$BD,$DA,$FB
	.dc.b	$FE,$02,$DF,$0C,$47,$78,$CB,$F2,$A7,$FA,$03,$AF,$9A,$DF,$CC,$BE
	.dc.b	$67,$FD,$FF,$A2,$F7,$6B,$DD,$6A,$7F,$B4,$DF,$1A,$F0,$9D,$43,$97
	.dc.b	$E1,$47,$09,$DC,$6E,$75,$7E,$1C,$7B,$CE,$B0,$AF,$28,$AF,$1B,$F2
	.dc.b	$46,$23,$D0,$78,$A3,$2D,$3A,$14,$E8,$53,$A1,$F3,$2B,$C8,$E4,$EB
	.dc.b	$7E,$8C,$3D,$91,$D2,$31,$7E,$3C,$3D,$37,$CA,$7E,$25,$76,$3C,$2E
	.dc.b	$1E,$B8,$7A,$23,$15,$CB,$0E,$F5,$F4,$23,$FC,$6E,$FD,$13,$FC,$72
	.dc.b	$E1,$F8,$67,$D0,$F9,$AF,$E3,$DE,$67,$F8,$F9,$B1,$F5,$03,$87,$87
	.dc.b	$47,$F8,$C9,$D4,$3D,$D7,$0F,$36,$CE,$9A,$F8,$5A,$FE,$22,$79,$4F
	.dc.b	$DE,$F9,$A7,$CA,$7E,$33,$4F,$C5,$46,$2F,$F0,$77,$C1,$FB,$4D,$9D
	.dc.b	$D7,$1D,$7F,$2B,$7C,$6F,$78,$F9,$AF,$F2,$D8,$71,$BD,$E1,$C6,$FF
	.dc.b	$5B,$87,$1B,$DE,$1C,$6F,$D4,$DA,$DD,$F1,$57,$C6,$B7,$8F,$AA,$F8
	.dc.b	$56,$AF,$85,$EE,$3F,$7B,$8F,$95,$B5,$C2,$75,$1C,$4E,$97,$1F,$CF
	.dc.b	$46,$8D,$D7,$38,$E0,$7D,$AD,$FE,$0B,$FC,$34,$CD,$7F,$88,$67,$85
	.dc.b	$7C,$B8,$FF,$12,$36,$B8,$D7,$33,$CF,$3F,$BA,$F9,$AF,$F3,$7C,$75
	.dc.b	$6B,$B0,$E2,$3F,$C6,$1F,$2E,$E2,$FE,$1F,$1F,$E3,$39,$D5,$FF,$12
	.dc.b	$6D,$FF,$24,$6A,$BE,$4F,$87,$E0,$F9,$D7,$16,$C7,$16,$BC,$7E,$6A
	.dc.b	$DA,$69,$27,$E4,$5A,$B1,$1E,$DA,$E0,$F8,$4D,$C7,$6F,$9A,$FE,$8F
	.dc.b	$1F,$45,$AF,$C5,$6E,$3F,$13,$CC,$E1,$D1,$2B,$83,$E8,$C7,$FD,$A1
	.dc.b	$7C,$33,$FE,$6E,$DA,$FE,$92,$D7,$E3,$5F,$C0,$F8,$87,$CD,$52,$4F
	.dc.b	$BB,$AA,$BF,$66,$76,$FE,$25,$AF,$DD,$C7,$FC,$8A,$E6,$B8,$E7,$FE
	.dc.b	$6D,$AD,$BF,$80,$BF,$A3,$F3,$5F,$E3,$3B,$FC,$A3,$5F,$C4,$36,$7F
	.dc.b	$CE,$DC,$9F,$E7,$39,$DF,$8B,$C8,$8F,$E7,$63,$4F,$28,$CB,$4D,$7A
	.dc.b	$5F,$82,$F2,$BE,$CC,$7E,$5D,$B7,$FE,$85,$6E,$FF,$D9,$3E,$6B,$FC
	.dc.b	$ED,$5F,$AB,$9C,$BF,$DE,$1A,$FF,$9B,$79,$3F,$1F,$3C,$23,$B4,$78
	.dc.b	$48,$FE,$D6,$35,$7F,$E9,$57,$E1,$5F,$87,$1E,$67,$E7,$3D,$37,$CD
	.dc.b	$7E,$92,$B7,$87,$F8,$2B,$E6,$BF,$4A,$FE,$CD,$B8,$F5,$C3,$F4,$71
	.dc.b	$8C,$7F,$C0,$2D,$AE,$5D,$1E,$07,$F8,$57,$CB,$BE,$1B,$F5,$6D,$8C
	.dc.b	$C7,$EB,$BE,$69,$F3,$5F,$B0,$F9,$AF,$F3,$A5,$7F,$13,$6B,$FC,$1D
	.dc.b	$BA,$7F,$AD,$70,$E9,$3F,$C2,$FE,$69,$F0,$DF,$BA,$6C,$7F,$1D,$6B
	.dc.b	$CF,$C5,$FD,$FF,$D2,$3D,$DA,$49,$FE,$49,$0F,$63,$11,$FC,$66,$38
	.dc.b	$3A,$9E,$73,$AC,$BF,$D9,$55,$AA,$E1,$5A,$7F,$98,$7C,$CF,$CB,$4F
	.dc.b	$FA,$0D,$CF,$A9,$FC,$52,$34,$47,$06,$78,$7F,$8C,$FC,$D3,$E6,$FF
	.dc.b	$C0,$ED,$BF,$F0,$00,$70,$FF,$80,$DA,$FF,$8E,$1C,$E5,$FE,$E3,$C1
	.dc.b	$F4,$7E,$33,$BD,$F9,$79,$26,$1E,$C7,$D1,$F2,$18,$8F,$E8,$17,$F6
	.dc.b	$30,$E9,$5A,$9E,$04,$36,$E5,$FE,$91,$F3,$8F,$0F,$F9,$BA,$F9,$9A
	.dc.b	$62,$24,$FF,$5D,$9D,$36,$72,$FF,$53,$F9,$CF,$E6,$2F,$F9,$61,$AF
	.dc.b	$F9,$81,$AF,$F9,$C9,$A9,$FE,$59,$2F,$A3,$F2,$9F,$D3,$23,$F9,$87
	.dc.b	$C4,$FE,$EB,$B7,$FC,$D6,$DB,$FF,$3E,$B5,$2F,$9D,$FE,$AA,$E4,$9E
	.dc.b	$F6,$23,$2C,$9F,$EF,$93,$B9,$E3,$E9,$F2,$E9,$95,$A8,$CB,$C2,$32
	.dc.b	$FF,$77,$AF,$F2,$13,$FE,$07,$F9,$3F,$D2,$3D,$BE,$8F,$CA,$7F,$6A
	.dc.b	$8C,$C7,$F7,$F6,$F0,$B3,$F8,$35,$C2,$BA,$75,$FF,$72,$F9,$EF,$D2
	.dc.b	$1F,$A3,$8F,$F0,$43,$6F,$94,$7F,$A9,$23,$7D,$3C,$7A,$A4,$FF,$87
	.dc.b	$67,$2D,$8C,$BD,$27,$C1,$E8,$48,$E0,$F8,$FE,$63,$11,$E2,$8B,$87
	.dc.b	$8A,$18,$2B,$E1,$7C,$97,$AF,$E8,$E5,$7C,$98,$CB,$E5,$8C,$BF,$1E
	.dc.b	$3F,$C2,$7F,$3F,$F9,$F8,$FF,$1B,$3F,$9A,$7F,$C5,$4D,$C3,$C4,$EE
	.dc.b	$33,$D3,$CB,$A6,$47,$28,$EA,$61,$E7,$18,$74,$78,$1D,$57,$69,$D4
	.dc.b	$F0,$34,$70,$3C,$26,$0D,$18,$24,$7A,$27,$AF,$07,$FB,$3A,$FD,$53
	.dc.b	$DE,$B9,$3B,$7B,$39,$7A,$B9,$7A,$CF,$3B,$EB,$1F,$EE,$AA,$F9,$13
	.dc.b	$98,$E6,$F8,$6F,$70,$9F,$1C,$EE,$3B,$BB,$78,$9D,$6B,$8C,$F6,$8C
	.dc.b	$1A,$3A,$56,$A7,$E5,$DF,$A4,$97,$DD,$F7,$18,$AF,$55,$70,$9D,$1E
	.dc.b	$38,$F2,$9E,$03,$E1,$9C,$0F,$96,$F0,$75,$3D,$5F,$29,$E7,$BF,$82
	.dc.b	$F1,$3A,$9F,$16,$FC,$4E,$67,$94,$F0,$8D,$1E,$03,$55,$E1,$72,$74
	.dc.b	$8F,$90,$E2,$7E,$0C,$FB,$5E,$B5,$E0,$9F,$39,$C5,$C5,$72,$23,$8B
	.dc.b	$FA,$13,$ED,$6B,$F1,$8E,$75,$D9,$D1,$E9,$3C,$97,$E9,$BE,$D5,$E0
	.dc.b	$BF,$C0,$B6,$B8,$4F,$89,$E6,$79,$CB,$7D,$DF,$14,$74,$7E,$95,$74
	.dc.b	$69,$D2,$BB,$3F,$25,$B5,$F4,$9B,$7F,$24,$93,$F8,$07,$D5,$1F,$0A
	.dc.b	$7E,$7C,$FD,$62,$1F,$00,$E0,$62,$FD,$8F,$A2,$79,$57,$48,$F8,$37
	.dc.b	$F4,$5A,$7C,$08,$F1,$BD,$49,$5F,$05,$F8,$73,$F0,$5F,$8B,$F3,$FF
	.dc.b	$C9,$88,$FF,$26,$F3,$AB,$FC,$9A,$2E,$3F,$C9,$A7,$FD,$7F,$F3,$EF
	.dc.b	$94,$FF,$60,$FC,$FB,$E8,$E9,$A8,$98,$FF,$B9,$E7,$4D,$9C,$BF,$ED
	.dc.b	$D8,$FE,$CE,$DE,$AE,$A7,$6E,$AB,$1C,$E2,$BC,$62,$E5,$FD,$36,$B9
	.dc.b	$B6,$B9,$57,$8F,$C6,$F4,$F4,$48,$C1,$0F,$23,$97,$CD,$1F,$EE,$0F
	.dc.b	$9F,$7D,$07,$FB,$8B,$E4,$5F,$47,$4D,$44,$B1,$FE,$FB,$9D,$B8,$CB
	.dc.b	$FB,$24,$62,$3F,$E0,$05,$F6,$75,$C9,$C9,$4B,$AB,$C5,$D3,$27,$67
	.dc.b	$82,$74,$E0,$E3,$3B,$7C,$93,$EB,$89,$1B,$9F,$EE,$CD,$9C,$BE,$BF
	.dc.b	$17,$E1,$78,$F2,$38,$47,$C5,$30,$FC,$18,$C3,$E3,$F9,$F7,$D0,$7F
	.dc.b	$AF,$63,$DD,$18,$F1,$7F,$CE,$6B,$C3,$7C,$7E,$2F,$85,$FC,$EF,$A3
	.dc.b	$7D,$1F,$F8,$3E,$3E,$44,$63,$7F,$F3,$EB,$63,$E6,$1C,$67,$82,$F9
	.dc.b	$67,$C6,$7F,$AA,$2F,$E6,$C6,$3E,$4F,$FC,$5B,$C9,$FE,$2E,$8D,$FA
	.dc.b	$B8,$AF,$C0,$C6,$DE,$37,$CF,$05,$F1,$86,$1C,$67,$C5,$1B,$9D,$4E
	.dc.b	$6F,$BD,$F4,$8C,$56,$E3,$C2,$F4,$7C,$3F,$45,$AF,$A2,$DB,$63,$27
	.dc.b	$9A,$32,$D3,$F1,$5F,$42,$FA,$1F,$8F,$9A,$31,$3F,$EA,$A6,$C7,$CE
	.dc.b	$35,$3C,$23,$39,$FF,$91,$5B,$5F,$A8,$F9,$47,$C3,$BE,$17,$E2,$63
	.dc.b	$CA,$BE,$5F,$8B,$F0,$32,$F8,$2D,$8C,$9F,$42,$32,$D3,$33,$F4,$9F
	.dc.b	$8F,$1A,$8C,$1F,$49,$F0,$FD,$07,$FB,$8E,$7F,$58,$D9,$D8,$F9,$FF
	.dc.b	$E5,$13,$AF,$0B,$D2,$7D,$59,$FE,$0A,$D5,$F0,$EF,$88,$DF,$2F,$C0
	.dc.b	$E9,$7D,$05,$31,$E1,$5C,$BF,$7D,$F3,$EE,$FF,$8F,$AB,$E8,$3E,$4C
	.dc.b	$E6,$BF,$82,$4E,$5F,$E0,$DF,$41,$FE,$FB,$79,$DF,$D0,$8C,$6D,$EE
	.dc.b	$BE,$4F,$C0,$8F,$AD,$5C,$BC,$9E,$1C,$9C,$5F,$B3,$DA,$F8,$EF,$CF
	.dc.b	$53,$E7,$FF,$8E,$BF,$C5,$E7,$4F,$F1,$C3,$5E,$47,$2A,$E3,$5C,$32
	.dc.b	$FA,$37,$C1,$FB,$37,$C0,$E9,$5C,$2F,$94,$75,$BE,$5D,$9F,$E9,$78
	.dc.b	$79,$DE,$6F,$A0,$7C,$6F,$E4,$A7,$27,$96,$32,$FF,$2F,$33,$97,$8A
	.dc.b	$32,$FE,$1A,$33,$7F,$EB,$65,$3A,$8F,$15,$F8,$6F,$C0,$DA,$FE,$79
	.dc.b	$F4,$0F,$77,$E1,$9F,$A6,$71,$1E,$D8,$C1,$A7,$8B,$71,$FC,$CC,$F0
	.dc.b	$7E,$4E,$57,$1F,$F5,$D4,$DB,$FE,$BA,$7D,$9B,$FD,$AF,$17,$F0,$FC
	.dc.b	$38,$FC,$EE,$3C,$6F,$AE,$DB,$FA,$0D,$9D,$C7,$F5,$7A,$E5,$8F,$C5
	.dc.b	$C9,$7B,$8C,$3F,$C1,$3C,$2F,$C0,$47,$FB,$83,$A9,$F3,$EF,$A0,$D7
	.dc.b	$1F,$F2,$D9,$CC,$63,$3F,$ED,$31,$87,$FB,$46,$1D,$1F,$EF,$71,$9D
	.dc.b	$3F,$7F,$39,$7F,$AA,$E1,$DE,$B8,$E1,$DE,$31,$3C,$3E,$63,$F9,$F6
	.dc.b	$4F,$A0,$E5,$2C,$55,$A9,$CC,$63,$1F,$EF,$F7,$C7,$1F,$85,$7B,$7E
	.dc.b	$26,$9F,$17,$E0,$AD,$FB,$1B,$97,$AA,$7F,$96,$BF,$E1,$8F,$A0,$FF
	.dc.b	$0C,$FD,$03,$E8,$3F,$C2,$5A,$7D,$2C,$7F,$C1,$97,$B1,$5B,$7F,$9D
	.dc.b	$60,$9E,$77,$FD,$0F,$0B,$5A,$9A,$FF,$87,$E2,$E1,$DE,$3A,$D7,$16
	.dc.b	$9C,$4A,$71,$2C,$71,$C3,$C9,$19,$6C,$66,$7A,$46,$30,$F4,$57,$43
	.dc.b	$DD,$19,$C3,$CF,$18,$AF,$8B,$5C,$2B,$BC,$E4,$F3,$38,$58,$7B,$A7
	.dc.b	$85,$78,$8E,$77,$DA,$B6,$7B,$27,$6F,$F8,$6A,$B9,$4E,$DF,$EE,$11
	.dc.b	$F0,$CA,$E1,$FB,$07,$F8,$8B,$E8,$3F,$8F,$F9,$FC,$1B,$1F,$DE,$5A
	.dc.b	$EF,$A3,$FC,$4A,$E7,$DB,$FF,$0D,$B7,$83,$94,$62,$3B,$35,$FF,$2E
	.dc.b	$52,$38,$E9,$F5,$BE,$81,$F3,$7F,$E6,$18,$C3,$76,$E3,$5C,$1B,$8F
	.dc.b	$E2,$AF,$8E,$36,$F8,$3F,$E7,$4C,$7C,$15,$F9,$06,$C7,$E2,$AB,$8B
	.dc.b	$6B,$8A,$9D,$E3,$F8,$C6,$E1,$FE,$77,$BE,$33,$E8,$AE,$D8,$F9,$DF
	.dc.b	$13,$54,$F0,$AE,$58,$78,$30,$EB,$5C,$31,$F6,$30,$FD,$0C,$69,$F3
	.dc.b	$61,$D7,$0E,$91,$FE,$94,$6B,$AA,$E4,$7E,$81,$F1,$DF,$FA,$46,$31
	.dc.b	$3F,$11,$D5,$F7,$8C,$6B,$E8,$BF,$F4,$84,$61,$F4,$1F,$B6,$FA,$0F
	.dc.b	$F2,$FB,$FE,$95,$6B,$FE,$2A,$6D,$F8,$FE,$79,$C5,$E3,$C3,$FD,$49
	.dc.b	$5A,$6F,$5E,$70,$FF,$58,$CE,$5F,$F5,$9D,$F2,$7C,$1F,$07,$B5,$28
	.dc.b	$C3,$63,$FD,$55,$2F,$A0,$CD,$49,$E6,$6A,$30,$FA,$E2,$E5,$EB,$9F
	.dc.b	$07,$D0,$7F,$3E,$AF,$F5,$B3,$5F,$F2,$5B,$67,$E1,$79,$7C,$3A,$65
	.dc.b	$70,$7D,$58,$FF,$B0,$23,$8C,$69,$A7,$AA,$7E,$07,$C3,$BE,$82,$92
	.dc.b	$7D,$FD,$75,$FF,$2E,$CE,$DF,$F6,$63,$5F,$F3,$0E,$1F,$E3,$D7,$35
	.dc.b	$C7,$1F,$C6,$56,$DF,$B9,$7F,$47,$D6,$FF,$16,$DF,$FB,$5D,$AF,$F9
	.dc.b	$B1,$B3,$F9,$5E,$4F,$F3,$34,$EF,$C5,$E4,$47,$FD,$C1,$1A,$79,$46
	.dc.b	$5A,$6B,$D1,$FC,$17,$97,$F7,$E3,$FC,$F6,$DB,$FD,$23,$79,$3F,$DE
	.dc.b	$9F,$3F,$FE,$6F,$AF,$F7,$74,$E5,$FF,$80,$2D,$7F,$D3,$5C,$7F,$8F
	.dc.b	$9E,$11,$DA,$3C,$24,$7F,$DF,$11,$AB,$FD,$A3,$F0,$AF,$C3,$8F,$33
	.dc.b	$F4,$1E,$9B,$E8,$3F,$41,$82,$F8,$21,$F4,$1F,$A3,$7F,$D3,$2D,$C7
	.dc.b	$BE,$1F,$A5,$8C,$63,$C6,$B8,$F4,$E4,$F1,$3D,$6F,$F8,$57,$CE,$BE
	.dc.b	$13,$F5,$CD,$9D,$5F,$38,$C4,$7F,$AA,$BE,$81,$F4,$1F,$B5,$FA,$0F
	.dc.b	$F3,$9C,$F0,$3D,$4E,$1F,$F5,$8B,$71,$F0,$E3,$FB,$DF,$A0,$7B,$74
	.dc.b	$97,$CC,$5F,$1C,$FD,$DF,$40,$F7,$7F,$83,$36,$B8,$35,$FF,$60,$AF
	.dc.b	$63,$73,$27,$6F,$28,$C4,$70,$8B,$9F,$F3,$1A,$E5,$3E,$5F,$9D,$FA
	.dc.b	$2F,$D9,$7F,$D0,$0E,$7D,$5F,$E3,$51,$A8,$34,$DB,$FE,$D6,$6D,$7F
	.dc.b	$3F,$78,$7D,$03,$E7,$FF,$92,$36,$BE,$33,$5E,$5E,$16,$A7,$FC,$19
	.dc.b	$39,$9F,$C3,$07,$0F,$E2,$5A,$FA,$F8,$3D,$24,$7E,$07,$C6,$57,$D1
	.dc.b	$17,$1F,$44,$DD,$BF,$29,$97,$E2,$A7,$D9,$5F,$12,$BF,$9F,$7C,$FB
	.dc.b	$C2,$BF,$2F,$32,$FC,$14,$E5,$B3,$97,$FD,$FB,$09,$E7,$3F,$D2,$7E
	.dc.b	$7B,$F9,$B6,$7F,$A5,$6D,$FE,$BD,$AA,$DF,$2F,$C7,$C8,$FF,$1A,$57
	.dc.b	$F4,$1A,$F9,$2D,$7A,$47,$F8,$E7,$E7,$7F,$82,$61,$FB,$36,$CF,$EF
	.dc.b	$5A,$97,$CF,$7F,$5F,$72,$4F,$7B,$11,$96,$4F,$E9,$67,$73,$C7,$D2
	.dc.b	$E5,$EA,$CA,$D4,$65,$E1,$19,$5C,$3F,$E0,$DB,$FF,$20,$47,$EC,$7E
	.dc.b	$47,$F3,$EE,$B2,$FE,$1D,$18,$AF,$F0,$A3,$63,$E0,$74,$FF,$79,$8E
	.dc.b	$35,$F9,$06,$BC,$BA,$BF,$CC,$11,$88,$FE,$45,$F3,$BF,$A7,$3F,$47
	.dc.b	$7F,$C7,$5B,$1A,$F0,$BF,$C0,$D1,$BE,$BA,$62,$24,$FF,$0B,$9E,$29
	.dc.b	$A7,$7E,$BE,$64,$8E,$1F,$29,$5F,$84,$79,$A3,$1A,$7F,$8D,$69,$5F
	.dc.b	$07,$E7,$25,$FC,$B7,$CF,$D5,$88,$E6,$34,$4E,$70,$E9,$5F,$E2,$3F
	.dc.b	$9C,$FD,$15,$F0,$7F,$40,$FF,$27,$6B,$9E,$9E,$07,$1E,$A9,$1B,$8C
	.dc.b	$BC,$63,$2F,$13,$46,$48,$64,$97,$FE,$55,$AF,$0B,$67,$AC,$EB,$7F
	.dc.b	$EA,$D5,$A6,$CE,$AB,$95,$F8,$9E,$57,$E2,$39,$1F,$D9,$EF,$93,$6B
	.dc.b	$95,$F6,$BE,$4F,$69,$E5,$7D,$9F,$5B,$D9,$EA,$2F,$E7,$C6,$49,$3D
	.dc.b	$67,$9B,$D5,$FA,$97,$E2,$7C,$D7,$E2,$7C,$C7,$03,$27,$8A,$79,$57
	.dc.b	$96,$3C,$77,$EE,$7C,$97,$D5,$F3,$5F,$57,$27,$53,$27,$53,$C2,$74
	.dc.b	$78,$4F,$53,$DF,$7E,$D9,$F5,$9C,$4E,$67,$92,$B8,$D7,$8E,$B9,$45
	.dc.b	$8E,$AD,$9E,$53,$DE,$78,$4F,$29,$F4,$4F,$26,$CF,$26,$9C,$A7,$A3
	.dc.b	$E6,$25,$FC,$92,$5F,$CC,$24,$7D,$03,$D4,$79,$2B,$C6,$D3,$C7,$5E
	.dc.b	$57,$67,$69,$F8,$D5,$DA,$3E,$6B,$F8,$67,$B3,$97,$EA,$9C,$4C,$C7
	.dc.b	$23,$CA,$EA,$38,$5F,$D4,$8E,$B3,$A8,$C4,$7D,$49,$F0,$3C,$67,$A3
	.dc.b	$D8,$E2,$43,$83,$A7,$81,$99,$EA,$4A,$F8,$55,$E4,$6D,$71,$9F,$B1
	.dc.b	$5F,$8D,$3E,$B1,$F6,$0F,$80,$76,$3C,$27,$94,$C3,$F8,$57,$E8,$9F
	.dc.b	$4C,$FA,$35,$EE,$8F,$2D,$FE,$16,$7C,$07,$68,$E4,$FF,$27,$E9,$FE
	.dc.b	$54,$47,$F9,$47,$97,$5D,$1F,$CA,$E3,$F9,$84,$7F,$B4,$FE,$71,$F2
	.dc.b	$14,$95,$FE,$57,$F9,$BF,$EF,$B3,$A8,$FF,$2D,$35,$FE,$7E,$DC,$BF
	.dc.b	$DF,$93,$FE,$04,$3F,$B8,$CF,$F4,$46,$C6,$BA,$7F,$DC,$91,$FE,$63
	.dc.b	$9F,$F4,$43,$AD,$3F,$D0,$4D,$8C,$CF,$0E,$9E,$2C,$8F,$43,$97,$D5
	.dc.b	$D1,$DB,$A2,$46,$08,$79,$5C,$BE,$78,$FF,$7C,$7C,$CB,$E6,$A9,$27
	.dc.b	$FC,$F9,$F1,$7F,$E6,$39,$D4,$7F,$95,$0F,$F5,$13,$87,$0D,$7F,$CB
	.dc.b	$2D,$7F,$D4,$11,$FE,$ED,$D1,$A6,$39,$FF,$D4,$91,$8A,$FF,$57,$57
	.dc.b	$F7,$67,$C7,$D3,$AB,$E6,$D8,$DD,$D5,$5C,$5F,$EF,$2F,$86,$B8,$C7
	.dc.b	$BA,$78,$4E,$EF,$C5,$3C,$AF,$3C,$DA,$F0,$A4,$FC,$68,$CB,$F2,$23
	.dc.b	$2F,$82,$3C,$9F,$47,$E4,$8F,$94,$1C,$3F,$35,$AF,$FA,$09,$AF,$FB
	.dc.b	$27,$E6,$DF,$37,$FE,$E2,$8C,$35,$FF,$6B,$47,$F8,$5F,$E2,$BB,$BC
	.dc.b	$3C,$3F,$B5,$F1,$EF,$86,$FF,$5E,$F0,$7F,$B1,$63,$1F,$05,$F4,$F7
	.dc.b	$FA,$91,$BF,$3F,$15,$FE,$E3,$8D,$BC,$6F,$9E,$7F,$E3,$29,$CB,$63
	.dc.b	$37,$D1,$61,$D2,$55,$DE,$FA,$4E,$63,$13,$CE,$31,$4F,$99,$BF,$33
	.dc.b	$86,$E1,$E5,$6D,$F8,$DB,$8F,$E0,$3E,$65,$F3,$3E,$3D,$DF,$47,$FE
	.dc.b	$40,$C7,$FD,$C1,$5A,$7F,$DD,$93,$AF,$1B,$D5,$92,$B4,$FE,$A5,$AF
	.dc.b	$FA,$F9,$A8,$8E,$1F,$87,$A7,$F1,$CB,$F9,$D8,$FC,$0F,$07,$E0,$B6
	.dc.b	$32,$7D,$38,$CB,$4C,$D7,$D3,$8C,$BF,$E5,$EF,$9A,$FF,$7C,$4F,$FB
	.dc.b	$F5,$B1,$FE,$62,$F9,$8F,$13,$BE,$EF,$79,$F7,$2B,$D5,$1F,$F0,$07
	.dc.b	$8F,$E1,$FB,$EF,$81,$DC,$7C,$6D,$DF,$35,$4C,$76,$DC,$7E,$27,$1F
	.dc.b	$ED,$3E,$65,$D3,$F3,$B6,$7C,$D7,$D0,$9C,$DF,$03,$C7,$3A,$71,$B7
	.dc.b	$CF,$F9,$A7,$8B,$C7,$1F,$B3,$18,$AD,$D3,$0E,$AF,$FA,$12,$5F,$0B
	.dc.b	$C3,$0E,$2D,$BE,$3D,$AF,$88,$FC,$F4,$3E,$63,$F8,$BB,$A8,$CC,$63
	.dc.b	$1F,$F4,$8D,$71,$AF,$49,$C7,$C5,$FC,$7C,$66,$7F,$D4,$33,$93,$A1
	.dc.b	$C2,$BA,$6F,$FC,$7A,$B8,$3F,$CC,$67,$85,$7F,$A8,$17,$95,$E8,$F9
	.dc.b	$A7,$C3,$7E,$7A,$72,$79,$23,$2F,$C8,$33,$5E,$48,$CB,$D6,$BA,$57
	.dc.b	$36,$D7,$16,$CF,$FA,$CB,$E6,$9E,$C7,$C6,$3F,$5E,$E2,$BD,$D1,$83
	.dc.b	$4F,$DA,$8D,$4F,$C6,$CE,$E5,$FD,$02,$6D,$7E,$CC,$FA,$D7,$CD,$EA
	.dc.b	$7E,$13,$8B,$F8,$CB,$F3,$7E,$5B,$F8,$0C,$63,$1E,$D1,$8F,$1B,$EE
	.dc.b	$B6,$B8,$1C,$E3,$2D,$33,$3F,$ED,$2F,$97,$FA,$F1,$FC,$1F,$A9,$F3
	.dc.b	$2F,$9A,$FE,$23,$5A,$B5,$6A,$73,$C9,$FC,$76,$F8,$61,$F9,$EC,$38
	.dc.b	$5E,$E3,$15,$BB,$E3,$1F,$C8,$2F,$8E,$1E,$5F,$7F,$F9,$D6,$EF,$9A
	.dc.b	$F1,$4B,$35,$6A,$73,$18,$C3,$F6,$B7,$C7,$2F,$5D,$ED,$F6,$EB,$EE
	.dc.b	$F7,$AD,$FA,$5B,$97,$C7,$9F,$E6,$6F,$F2,$DF,$9A,$FE,$5D,$F3,$4F
	.dc.b	$9A,$FE,$2F,$A2,$F7,$6B,$DD,$6A,$7F,$C2,$D7,$C6,$BF,$8F,$9B,$0E
	.dc.b	$5F,$81,$1C,$27,$FC,$55,$5D,$27,$FA,$05,$F8,$71,$C4,$7F,$2A,$9B
	.dc.b	$39,$8E,$B5,$C9,$A7,$2B,$EC,$F9,$2B,$8C,$62,$3A,$1C,$0F,$2D,$F8
	.dc.b	$1F,$8C,$66,$FC,$18,$7A,$E3,$2F,$B6,$33,$87,$8E,$31,$7D,$EB,$73
	.dc.b	$F2,$6F,$B4,$7C,$7A,$DC,$63,$0F,$3B,$67,$C5,$7C,$DF,$9C,$7F,$8A
	.dc.b	$A7,$FC,$56,$FA,$0C,$1F,$E5,$7F,$9A,$FE,$33,$E5,$FF,$8C,$1B,$19
	.dc.b	$E8,$FF,$11,$36,$75,$EC,$B8,$29,$27,$A4,$65,$B1,$F6,$3E,$69,$F2
	.dc.b	$9F,$D8,$B5,$FC,$14,$62,$BE,$EE,$4C,$39,$57,$E1,$E3,$18,$73,$C3
	.dc.b	$96,$DF,$DB,$31,$E5,$8F,$D9,$BD,$B7,$1E,$75,$BB,$E3,$3F,$E0,$DB
	.dc.b	$E3,$8F,$7B,$DC,$FF,$84,$70,$E7,$3E,$8B,$E3,$2B,$E3,$87,$F7,$EB
	.dc.b	$DB,$70,$E7,$5B,$BE,$37,$C2,$3B,$4E,$AF,$8C,$79,$6B,$67,$82,$33
	.dc.b	$06,$D7,$38,$EC,$7C,$CA,$F0,$C7,$C6,$C7,$FC,$F6,$DD,$BE,$25,$7F
	.dc.b	$7C,$BE,$78,$7F,$A2,$6B,$53,$E3,$35,$87,$B3,$E6,$BF,$CA,$F1,$FE
	.dc.b	$04,$6B,$F8,$7C,$3F,$C4,$31,$8C,$7F,$D1,$7F,$30,$E2,$FF,$4B,$4E
	.dc.b	$6B,$7D,$38,$3F,$D3,$55,$BC,$3B,$7C,$07,$02,$3F,$E1,$48,$7C,$D6
	.dc.b	$23,$C4,$7F,$92,$F0,$56,$72,$F1,$D5,$C7,$CF,$36,$FF,$CA,$58,$74
	.dc.b	$7F,$C4,$3F,$35,$FC,$E1,$BD,$1F,$EA,$F8,$C4,$78,$04,$FF,$AB,$DF
	.dc.b	$F2,$E4,$6B,$1F,$F5,$6B,$70,$FF,$23,$B6,$7C,$3E,$5F,$5C,$24,$78
	.dc.b	$0F,$81,$1F,$EB,$FF,$85,$7C,$D5,$24,$FC,$5E,$BA,$F9,$67,$6F,$F8
	.dc.b	$F1,$AF,$A7,$3F,$F1,$B3,$9A,$E3,$9F,$FB,$12,$B6,$FD,$CB,$FF,$68
	.dc.b	$7C,$D7,$F8,$92,$FF,$C8,$ED,$7E,$23,$67,$FD,$9B,$C9,$FE,$63,$9D
	.dc.b	$F8,$BC,$88,$FF,$94,$23,$4F,$28,$CB,$4D,$7A,$3F,$EE,$1F,$2B,$EF
	.dc.b	$C7,$F9,$61,$AF,$D5,$6D,$FF,$B7,$BE,$67,$FC,$D7,$3B,$E2,$E4,$47
	.dc.b	$FC,$C3,$1A,$79,$46,$5A,$6B,$C5,$FF,$79,$4F,$BF,$8F,$FD,$06,$D9
	.dc.b	$FF,$4F,$37,$D3,$7C,$D7,$E8,$E7,$3A,$7F,$BE,$FE,$6B,$F4,$51,$F8
	.dc.b	$D6,$E7,$E9,$9D,$75,$4A,$FF,$3B,$D6,$B2,$F0,$D7,$FA,$52,$FF,$DF
	.dc.b	$BF,$34,$F8,$3C,$57,$FA,$62,$31,$87,$FA,$01,$AF,$E7,$7E,$69,$F3
	.dc.b	$5C,$23,$FD,$71,$F3,$3F,$E6,$67,$FD,$0E,$D7,$F4,$8D,$C7,$C3,$8F
	.dc.b	$19,$7C,$D3,$DB,$C3,$63,$FD,$22,$D7,$9F,$91,$C7,$E6,$3F,$DC,$11
	.dc.b	$FE,$96,$6B,$FB,$36,$E0,$D7,$FD,$6B,$F3,$5F,$D4,$6B,$87,$C2,$F8
	.dc.b	$E1,$F5,$54,$ED,$12,$7F,$D4,$73,$A6,$CE,$5F,$E0,$50,$F8,$0E,$84
	.dc.b	$8E,$3E,$DD,$77,$17,$1F,$34,$7F,$B8,$EB,$FD,$6A,$FF,$AC,$BE,$6B
	.dc.b	$EE,$3F,$EE,$77,$FD,$06,$E7,$D7,$F9,$91,$A8,$34,$DB,$FC,$5D,$B5
	.dc.b	$E0,$8D,$F3,$FC,$DF,$9B,$FF,$06,$EC,$69,$27,$FC,$82,$7F,$BE,$63
	.dc.b	$4F,$9F,$E8,$FC,$67,$37,$EB,$4E,$1D,$18,$3D,$DC,$7E,$A2,$3F,$07
	.dc.b	$E3,$AB,$E9,$8B,$8F,$A7,$25,$7E,$37,$E7,$1E,$35,$F9,$79,$8F,$FB
	.dc.b	$72,$72,$D9,$CB,$FC,$EE,$13,$CE,$7F,$31,$F3,$9F,$D0,$2B,$FC,$C2
	.dc.b	$D7,$FE,$00,$B5,$07,$0D,$F3,$3C,$DF,$47,$23,$FD,$DF,$3F,$D0,$DD
	.dc.b	$1E,$78,$D3,$FD,$3E,$35,$F3,$9F,$C5,$74,$FB,$4D,$E8,$F0,$D2,$30
	.dc.b	$F9,$17,$CE,$FE,$CD,$CC,$6F,$C6,$C1,$99,$D7,$A9,$BE,$C9,$5E,$EC
	.dc.b	$7C,$33,$A8,$C6,$3E,$2C,$7F,$81,$36,$F5,$3F,$DA,$7E,$73,$F3,$EF
	.dc.b	$F6,$BA,$7D,$1F,$94,$FE,$11,$3A,$C7,$F3,$0D,$C3,$B4,$EE,$FC,$27
	.dc.b	$D2,$9D,$37,$B3,$4F,$79,$EB,$1F,$C5,$7E,$7B,$F5,$47,$E9,$AF,$E5
	.dc.b	$B6,$BE,$EB,$AE,$AE,$1D,$72,$7C,$22,$79,$FA,$7D,$49,$1A,$87,$CC
	.dc.b	$62,$3F,$C1,$91,$FE,$0B,$79,$4F,$96,$7E,$1F,$C9,$F9,$FE,$8E,$57
	.dc.b	$83,$54,$EA,$3B,$4E,$9F,$A1,$19,$7B,$C6,$76,$F9,$B3,$97,$F8,$36
	.dc.b	$29,$D6,$9F,$8E,$C3,$BE,$1E,$1C,$32,$FC,$6A,$E5,$5B,$9C,$CF,$C5
	.dc.b	$9C,$BF,$5E,$33,$7D,$E3,$A1,$0F,$39,$EC,$31,$87,$A1,$C1,$A3,$DE
	.dc.b	$6A,$FE,$DD,$F8,$B0,$F4,$CE,$8E,$15,$F0,$27,$CB,$09,$DC,$E6,$3B
	.dc.b	$BE,$17,$93,$97,$9D,$7C,$17,$C4,$7C,$53,$8B,$97,$8B,$E6,$72,$7C
	.dc.b	$53,$89,$F2,$4C,$9E,$5C,$3C,$F8,$7C,$98,$F0,$9C,$4E,$A7,$84,$C9
	.dc.b	$C4,$EA,$F1,$AF,$51,$ED,$25,$79,$67,$8D,$FC,$48,$F5,$4F,$07,$4F
	.dc.b	$09,$F1,$39,$7C,$4E,$5F,$11,$E4,$AF,$14,$F4,$24,$70,$7A,$98,$78
	.dc.b	$CF,$57,$E3,$1F,$00,$CC,$70,$79,$1E,$47,$94,$F8,$23,$27,$03,$46
	.dc.b	$1D,$39,$23,$83,$35,$F1,$63,$89,$E4,$AF,$34,$66,$38,$E1,$F9,$5C
	.dc.b	$3E,$DB,$63,$EA,$B5,$F6,$5F,$78,$FC,$BD,$73,$8C,$4F,$E2,$23,$13
	.dc.b	$C6,$FE,$85,$6A,$BE,$35,$69,$FC,$6B,$E5,$30,$FE,$02,$FC,$AE,$9A
	.dc.b	$E4,$F3,$99,$29,$93,$D3,$7E,$3C,$3E,$8B,$63,$E3,$DF,$97,$0F,$C3
	.dc.b	$61,$E2,$7E,$4E,$1E,$06,$D7,$66,$BD,$8F,$11,$FE,$B9,$F9,$FF,$F0
	.dc.b	$1B,$7A,$7F,$CA,$D1,$FE,$02,$8F,$E7,$6E,$A7,$FC,$50,$DC,$BF,$B7
	.dc.b	$BB,$9E,$47,$0E,$9E,$7F,$3F,$23,$9B,$DB,$D1,$C4,$78,$E2,$E1,$DA
	.dc.b	$7F,$CD,$1F,$3E,$F9,$5F,$F3,$57,$CF,$BE,$8E,$9A,$89,$8F,$F5,$29
	.dc.b	$D3,$67,$2F,$FB,$66,$BA,$39,$21,$9A,$EF,$D5,$23,$A3,$E2,$EA,$C4
	.dc.b	$7F,$B8,$F0,$FF,$37,$36,$74,$F8,$FC,$8F,$4F,$4C,$8C,$10,$E4,$E5
	.dc.b	$E9,$1F,$E8,$9F,$9F,$7D,$07,$FA,$2F,$E4,$DF,$47,$4D,$44,$BF,$EE
	.dc.b	$73,$A6,$CE,$5F,$F7,$D5,$74,$72,$43,$35,$DF,$AF,$C1,$D7,$27,$AC
	.dc.b	$65,$EF,$19,$5C,$5F,$E6,$FC,$7F,$BE,$1F,$EA,$E7,$1E,$37,$AC,$F6
	.dc.b	$79,$12,$BD,$31,$97,$D7,$1F,$EA,$BF,$9F,$7D,$07,$FB,$5E,$7F,$C8
	.dc.b	$31,$8F,$17,$FC,$EF,$08,$FE,$B5,$F2,$58,$75,$1F,$E7,$CF,$A3,$7D
	.dc.b	$1F,$FA,$16,$3F,$C9,$71,$8E,$EE,$6D,$8E,$13,$B7,$FC,$C3,$0F,$89
	.dc.b	$C4,$23,$15,$CA,$73,$18,$9F,$F2,$C8,$F9,$87,$C7,$FF,$B1,$EF,$FC
	.dc.b	$7B,$18,$BF,$F5,$4E,$FF,$EA,$A8,$DF,$B1,$8A,$FF,$39,$C6,$DE,$37
	.dc.b	$CF,$0F,$AB,$1A,$8F,$15,$F3,$74,$DC,$3A,$57,$96,$5E,$BE,$30,$F3
	.dc.b	$FD,$15,$9C,$E3,$E0,$8C,$56,$5B,$19,$3F,$C9,$3F,$42,$FA,$1F,$92
	.dc.b	$D8,$FB,$6D,$C7,$FD,$66,$D7,$E4,$B9,$CF,$FD,$7C,$D9,$FF,$4A,$FC
	.dc.b	$AB,$E2,$5F,$0F,$EC,$C1,$7B,$7F,$D3,$6D,$73,$8F,$C2,$BE,$55,$C6
	.dc.b	$B5,$F4,$3F,$EE,$3C,$7F,$CC,$6D,$C3,$E7,$9F,$EA,$6F,$9F,$FE,$39
	.dc.b	$3A,$F0,$3C,$47,$D0,$7F,$D3,$6D,$9F,$F5,$6F,$C4,$3E,$2B,$7C,$9E
	.dc.b	$9E,$87,$D0,$D3,$1E,$05,$CB,$FC,$EB,$F4,$0E,$CF,$81,$AB,$E8,$7E
	.dc.b	$14,$E6,$BF,$CF,$73,$97,$FC,$FB,$F4,$2F,$1F,$78,$7C,$98,$C4,$F6
	.dc.b	$8C,$71,$E6,$AC,$F2,$7E,$87,$5B,$E4,$3F,$1B,$2F,$A0,$78,$BA,$9C
	.dc.b	$C6,$27,$FD,$1F,$5C,$5F,$F5,$09,$C6,$B8,$38,$9E,$1C,$9F,$48,$FD
	.dc.b	$75,$F4,$AD,$C7,$D6,$AE,$47,$FA,$70,$79,$9E,$6F,$A1,$7C,$77,$E2
	.dc.b	$67,$27,$8A,$32,$F1,$9E,$13,$B6,$C6,$4E,$11,$96,$9F,$EA,$CF,$A1
	.dc.b	$7B,$9F,$08,$FC,$C3,$88,$F7,$46,$0D,$3C,$5B,$87,$C1,$9E,$0F,$BF
	.dc.b	$3B,$E0,$2F,$FD,$D7,$58,$7F,$DF,$A7,$D5,$BF,$D4,$70,$7F,$08,$8C
	.dc.b	$DB,$E7,$B3,$E2,$FD,$86,$CF,$D3,$6E,$5E,$16,$E5,$E1,$8F,$15,$F3
	.dc.b	$AF,$F6,$6E,$1D,$E7,$37,$FB,$9F,$03,$F0,$51,$FA,$EE,$A7,$D0,$3E
	.dc.b	$87,$5C,$7F,$C6,$A7,$31,$8C,$BF,$DD,$31,$87,$FD,$B7,$87,$4D,$3C
	.dc.b	$93,$9C,$7F,$15,$39,$A4,$27,$55,$EB,$85,$6D,$D6,$3E,$5B,$EF,$F0
	.dc.b	$DF,$CB,$AD,$F4,$3B,$BE,$33,$98,$C6,$08,$C6,$7F,$0D,$B8,$7A,$B4
	.dc.b	$F8,$DF,$05,$F1,$DB,$19,$26,$3F,$A5,$3D,$7F,$43,$EC,$FA,$17,$D0
	.dc.b	$FB,$A7,$D7,$7F,$CA,$5D,$89,$DB,$FC,$C2,$3E,$E4,$62,$39,$57,$5A
	.dc.b	$E9,$9F,$DF,$9E,$51,$FD,$91,$4E,$DE,$98,$57,$99,$C0,$C9,$4C,$DF
	.dc.b	$86,$FB,$3E,$38,$F4,$B9,$C3,$CF,$87,$87,$0E,$D7,$C2,$31,$7D,$63
	.dc.b	$0F,$F7,$A9,$DD,$72,$8C,$1B,$70,$F9,$2B,$E0,$C6,$23,$C7,$18,$C3
	.dc.b	$DD,$18,$3D,$B8,$7A,$5F,$8A,$75,$8D,$B8,$3F,$C4,$77,$FE,$24,$7E
	.dc.b	$B9,$83,$E7,$7D,$0F,$F1,$EF,$3B,$FC,$52,$D8,$CF,$4F,$F8,$51,$B3
	.dc.b	$AF,$67,$FC,$7F,$1C,$9A,$FF,$87,$29,$18,$8F,$C5,$46,$38,$79,$1F
	.dc.b	$6F,$E8,$5F,$39,$F8,$FD,$3F,$11,$18,$BF,$C0,$5F,$17,$EF,$36,$77
	.dc.b	$5C,$75,$FC,$C5,$F1,$BD,$E3,$E8,$BF,$CA,$61,$C6,$F7,$87,$1B,$FD
	.dc.b	$EE,$1C,$6F,$78,$71,$BF,$5B,$6B,$77,$C5,$5F,$1A,$DE,$3E,$AB,$E1
	.dc.b	$5A,$BE,$17,$B8,$FE,$1B,$8F,$95,$B5,$C2,$75,$1C,$4E,$97,$1F,$D1
	.dc.b	$46,$8D,$D7,$38,$E0,$7F,$65,$8F,$8D,$3F,$C6,$5B,$AF,$A2,$79,$E1
	.dc.b	$FC,$0E,$B4,$D9,$F9,$33,$FC,$92,$7F,$91,$5F,$7F,$A1,$FF,$33,$4E
	.dc.b	$6B,$F8,$64,$62,$3F,$BB,$36,$BF,$95,$7C,$FB,$8B,$B5,$7F,$30,$AD
	.dc.b	$5F,$2E,$BC,$CA,$72,$FF,$36,$BE,$4F,$84,$E5,$F0,$5C,$24,$D3,$9C
	.dc.b	$BE,$87,$15,$C4,$ED,$38,$7F,$C1,$D3,$B7,$55,$D1,$B3,$BC,$BD,$B7
	.dc.b	$E8,$FA,$1F,$E7,$95,$D5,$AF,$F8,$65,$B3,$F1,$3C,$BE,$1D,$12,$B8
	.dc.b	$3E,$88,$C4,$71,$8B,$3C,$63,$D5,$F0,$CF,$A1,$F9,$2D,$8C,$C7,$C2
	.dc.b	$EB,$CB,$23,$27,$C3,$D7,$FC,$78,$E0,$FF,$18,$57,$36,$CF,$37,$E5
	.dc.b	$7A,$DF,$E2,$C6,$F8,$BF,$6A,$31,$1F,$E3,$9E,$4F,$F3,$13,$A2,$1F
	.dc.b	$E3,$D8,$DE,$5F,$DC,$67,$73,$9A,$E8,$47,$ED,$46,$A7,$FB,$C7,$95
	.dc.b	$F5,$E7,$FC,$90,$D8,$CF,$07,$E3,$BE,$87,$FC,$D9,$5F,$83,$AD,$72
	.dc.b	$7F,$67,$6C,$7F,$81,$9A,$A7,$8B,$DA,$BF,$33,$3A,$37,$3A,$BF,$F0
	.dc.b	$63,$EE,$BF,$EE,$8D,$BF,$F0,$9B,$7D,$37,$D0,$FE,$8A,$7C,$19,$FE
	.dc.b	$9B,$E8,$7F,$47,$1F,$E6,$56,$E3,$DA,$75,$D1,$2B,$F4,$15,$AC,$BC
	.dc.b	$35,$FB,$0F,$A1,$FF,$7F,$47,$F8,$1D,$B5,$FE,$26,$6A,$FA,$17,$D0
	.dc.b	$F1,$F8,$3F,$DC,$FD,$0F,$F9,$C2,$BF,$5C,$D7,$FC,$F5,$7C,$BE,$85
	.dc.b	$ED,$7F,$00,$6C,$FE,$D5,$AF,$FA,$06,$B9,$6D,$FE,$35,$F9,$FF,$F6
	.dc.b	$DC,$FE,$ED,$AF,$FA,$1D,$B1,$FE,$3B,$FA,$0F,$EB,$73,$BF,$83,$FE
	.dc.b	$D9,$5C,$23,$FD,$F6,$D7,$FC,$50,$D7,$FC,$8E,$F4,$DB,$FC,$59,$4F
	.dc.b	$A1,$FE,$32,$DA,$EC,$7F,$0F,$5F,$01,$A2,$46,$7D,$AA,$FF,$A7,$E7
	.dc.b	$6F,$19,$EB,$3F,$C9,$BE,$7F,$ED,$4F,$F9,$69,$FF,$47,$BF,$C7,$5A
	.dc.b	$FF,$AA,$5D,$A6,$9A,$CF,$FC,$C0,$DC,$7F,$C9,$0F,$0C,$9F,$41,$FC
	.dc.b	$D9,$B7,$F4,$63,$C9,$E1,$E5,$F9,$A4,$8F,$F2,$20,$9E,$2F,$F9,$1D
	.dc.b	$B8,$78,$BE,$87,$DC,$73,$9F,$67,$1F,$C1,$23,$EF,$F8,$CA,$FB,$E2
	.dc.b	$D7,$B2,$7F,$A6,$FD,$0B,$C3,$BF,$31,$30,$FE,$89,$39,$6C,$E5,$FF
	.dc.b	$69,$42,$79,$CF,$F5,$8F,$A0,$FE,$7B,$BF,$F9,$BD,$B7,$FE,$8E,$6A
	.dc.b	$B7,$CC,$71,$F2,$7F,$C7,$D5,$FD,$22,$BF,$CF,$01,$C3,$D6,$36,$7B
	.dc.b	$3E,$86,$47,$F5,$D9,$EF,$F4,$1F,$C4,$71,$FF,$41,$36,$7F,$D4,$2D
	.dc.b	$4B,$E8,$7F,$B9,$B9,$27,$BF,$88,$CF,$A9,$AF,$62,$5F,$F6,$DB,$FD
	.dc.b	$7B,$FE,$F9,$6F,$0F,$FA,$B5,$FD,$DE,$3E,$4A,$E7,$5F,$DD,$7E,$79
	.dc.b	$D3,$FE,$04,$6C,$66,$3C,$BD,$59,$7F,$CC,$D3,$A9,$FF,$49,$35,$F9
	.dc.b	$6D,$7F,$D4,$8E,$F1,$F2,$7D,$0F,$C0,$3A,$4F,$F8,$1F,$E8,$3F,$52
	.dc.b	$7E,$95,$FF,$64,$B5,$3A,$F2,$3F,$C3,$6E,$48,$E3,$B2,$4F,$57,$7F
	.dc.b	$2F,$88,$61,$FE,$D3,$B5,$66,$31,$5F,$17,$E8,$25,$FC,$97,$C3,$D7
	.dc.b	$88,$E7,$BF,$C0,$9C,$5F,$5A,$B3,$9A,$E4,$7C,$18,$CB,$FE,$E8,$D1
	.dc.b	$78,$8F,$83,$36,$73,$1C,$8F,$09,$F2,$23,$53,$C7,$2F,$E9,$B5,$C2
	.dc.b	$B6,$1C,$61,$FE,$F5,$6E,$1F,$ED,$DA,$E3,$7D,$2F,$D0,$78,$6F,$84
	.dc.b	$77,$C3,$53,$DB,$0E,$6F,$94,$E1,$87,$C3,$75,$5F,$31,$C9,$F0,$DE
	.dc.b	$8E,$0F,$43,$87,$89,$99,$EE,$FC,$83,$11,$BB,$F3,$C6,$0F,$59,$C8
	.dc.b	$8E,$9F,$8F,$87,$C2,$78,$18,$C3,$DF,$3B,$3F,$01,$1A,$71,$87,$CB
	.dc.b	$9D,$BF,$06,$37,$3E,$08,$C1,$1F,$90,$64,$F8,$87,$C3,$AF,$23,$93
	.dc.b	$E2,$9E,$C9,$F5,$3F,$B6,$BF,$DD,$91,$FE,$93,$87,$E1,$5D,$18,$8D
	.dc.b	$3F,$C0,$1F,$D8,$FD,$0C,$8F,$76,$1F,$87,$8C,$47,$8E,$BE,$9E,$09
	.dc.b	$CE,$1F,$12,$BD,$67,$CE,$8C,$DF,$C1,$8F,$59,$93,$F8,$44,$7C,$A7
	.dc.b	$F8,$3B,$C1,$E7,$3D,$5F,$E1,$31,$97,$9C,$66,$BC,$91,$D0,$F5,$9D
	.dc.b	$0C,$1E,$BC,$39,$47,$08,$C1,$A9,$FE,$19,$5E,$58,$F0,$99,$BF,$66
	.dc.b	$1F,$12,$7C,$8F,$E6,$0C,$4E,$CF,$5C,$FC,$3B,$F0,$5F,$69,$F3,$9D
	.dc.b	$8F,$8C,$79,$0A,$ED,$C3,$49,$3E,$4C,$3D,$B8,$7A,$6F,$BC,$FD,$DB
	.dc.b	$EF,$18,$7E,$31,$D2,$34,$F8,$1E,$8F,$AC,$F3,$5F,$D5,$C3,$DF,$1B
	.dc.b	$3D,$E0,$7D,$0F,$F8,$25,$FE,$32,$FF,$81,$DF,$F6,$83,$5C,$F5,$FF
	.dc.b	$B8,$23,$85,$2F,$A5,$74,$EC,$D1,$BB,$E1,$D7,$27,$AB,$CF,$D1,$C4
	.dc.b	$78,$A2,$E1,$D6,$3D,$18,$FF,$A7,$3E,$81,$F4,$1F,$EF,$26,$CF,$FB
	.dc.b	$9D,$AF,$F3,$4F,$A0,$FF,$04,$46,$23,$FC,$2E,$D7,$D2,$D7,$E3,$35
	.dc.b	$3D,$2F,$C9,$09,$7D,$0F,$C2,$9C,$C6,$27,$C9,$79,$D7,$E3,$CF,$F8
	.dc.b	$39,$BD,$79,$AF,$AD,$39,$6F,$56,$5F,$04,$D9,$D4,$60,$FB,$D8,$78
	.dc.b	$AB,$5D,$5B,$EA,$95,$AC,$7C,$73,$A8,$C1,$E2,$9D,$BF,$2E,$3C,$5F
	.dc.b	$40,$FA,$0A,$49,$E2,$3E,$47,$FC,$F7,$3B,$8F,$F3,$6B,$5F,$F2,$48
	.dc.b	$70,$FF,$93,$1A,$FE,$39,$E1,$97,$D2,$97,$D0,$FC,$56,$D7,$8A,$B7
	.dc.b	$97,$E5,$A7,$FC,$2B,$3E,$65,$D7,$C4,$E5,$D9,$27,$51,$97,$84,$77
	.dc.b	$EC,$F0,$78,$D2,$7A,$46,$5E,$D1,$97,$FC,$0F,$7F,$ED,$6F,$A3,$F8
	.dc.b	$D1,$FE,$90,$6B,$F1,$1A,$FF,$9B,$9A,$FE,$C3,$E8,$5F,$45,$FE,$FE
	.dc.b	$8C,$35,$FE,$01,$1F,$E1,$5F,$90,$FF,$4E,$BC,$3C,$6F,$F4,$FF,$D0
	.dc.b	$BE,$8B,$F3,$B7,$FB,$C8,$C6,$FA,$9C,$CF,$F0,$B9,$CB,$FE,$82,$F9
	.dc.b	$87,$C7,$7F,$BA,$2F,$F1,$71,$8F,$8A,$FD,$E7,$2F,$EF,$63,$7E,$CE
	.dc.b	$2B,$F8,$B4,$6D,$E3,$7C,$F0,$FF,$1A,$CE,$5B,$19,$BE,$8B,$0E,$92
	.dc.b	$AF,$2F,$AF,$87,$CF,$F4,$56,$D3,$63,$0D,$C7,$C6,$DB,$FF,$20,$FD
	.dc.b	$0B,$E8,$BB,$F0,$BE,$8F,$FC,$A3,$87,$E9,$2B,$4F,$EB,$A7,$5E,$6F
	.dc.b	$9B,$05,$69,$FD,$2B,$5F,$F5,$7B,$5F,$E6,$2D,$7F,$9F,$3C,$76,$FE
	.dc.b	$2E,$DB,$FE,$85,$F2,$AF,$89,$7C,$47,$86,$B5,$87,$F4,$7F,$A3,$78
	.dc.b	$5F,$80,$84,$63,$1F,$F7,$3B,$63,$F7,$6D,$7F,$D8,$B3,$F6,$5B,$C9
	.dc.b	$CE,$31,$1F,$D2,$9A,$FF,$57,$D5,$F4,$5E,$16,$E1,$FE,$F6,$6C,$FF
	.dc.b	$0A,$6B,$FE,$D4,$FA,$17,$4F,$CD,$FA,$17,$D1,$7C,$F9,$CE,$1F,$C6
	.dc.b	$27,$2F,$F1,$A8,$CD,$FF,$9F,$9B,$C3,$FD,$0B,$E8,$BF,$D8,$35,$BC
	.dc.b	$7E,$D4,$7C,$C8,$FE,$46,$EA,$FF,$D0,$EE,$5B,$E4,$7E,$07,$A9,$F2
	.dc.b	$5F,$98,$B7,$CF,$FF,$2B,$7F,$96,$CE,$9F,$E6,$E6,$BC,$3E,$55,$C6
	.dc.b	$B8,$61,$FE,$96,$BE,$0F,$FA,$7A,$F8,$1D,$2B,$85,$F2,$8E,$B7,$CB
	.dc.b	$3F,$EC,$30,$F2,$FC,$DF,$44,$F8,$DF,$CD,$4E,$4F,$2C,$65,$FE,$8A
	.dc.b	$67,$2F,$14,$65,$FF,$55,$C6,$6F,$FC,$04,$A7,$51,$E2,$BF,$0D,$F8
	.dc.b	$1B,$5F,$D4,$BE,$89,$F4,$5F,$0C,$FD,$73,$88,$F6,$C6,$0D,$3C,$5B
	.dc.b	$8F,$FA,$FA,$78,$3F,$27,$2B,$8F,$F8,$12,$6D,$FF,$8B,$CF,$B3,$7F
	.dc.b	$BD,$E0,$FE,$29,$1F,$33,$E7,$F1,$E2,$7D,$76,$C6,$4F,$ED,$73,$C3
	.dc.b	$2F,$04,$E9,$B3,$AA,$E4,$FF,$72,$EE,$FB,$F9,$7F,$8A,$F9,$DF,$42
	.dc.b	$FA,$2F,$E2,$15,$AB,$FD,$84,$E5,$E5,$C7,$FC,$72,$31,$97,$FB,$82
	.dc.b	$B8,$2C,$3A,$E1,$D2,$F8,$37,$0F,$E4,$18,$72,$C7,$FC,$57,$8B,$E1
	.dc.b	$BF,$9B,$C3,$E8,$BA,$E6,$AD,$61,$FE,$16,$AD,$65,$EA,$C3,$8E,$1C
	.dc.b	$2F,$78,$7F,$BF,$2F,$78,$7C,$5D,$7E,$0F,$C1,$7C,$26,$C6,$49,$5F
	.dc.b	$EF,$E3,$FC,$D5,$F4,$5F,$E6,$BF,$A2,$7D,$17,$F9,$8B,$26,$1F,$CE
	.dc.b	$EF,$62,$B6,$FF,$32,$8F,$F2,$F4,$62,$3F,$CD,$77,$D6,$FA,$56,$9F
	.dc.b	$F1,$3C,$66,$79,$61,$5E,$27,$79,$E3,$19,$8F,$1D,$F7,$C3,$CF,$39
	.dc.b	$95,$F3,$9C,$C7,$33,$FC,$FE,$68,$C1,$F5,$70,$F3,$3A,$9E,$51,$88
	.dc.b	$F2,$DF,$86,$31,$1E,$4C,$3C,$EF,$43,$07,$B8,$C1,$E1,$9E,$58,$7A
	.dc.b	$6F,$BB,$67,$CB,$87,$AA,$BE,$74,$7D,$43,$FC,$A2,$7A,$CB,$58,$74
	.dc.b	$7F,$A7,$BE,$87,$F9,$07,$9B,$C1,$B1,$FE,$5A,$6B,$F1,$9A,$F1,$F6
	.dc.b	$DA,$F1,$6C,$71,$6B,$C5,$AF,$18,$FB,$DF,$44,$F9,$DF,$F5,$8E,$5F
	.dc.b	$89,$8C,$57,$E0,$F0,$61,$DA,$BF,$1D,$18,$C3,$9E,$1C,$B5,$FF,$5D
	.dc.b	$E3,$CB,$1F,$BB,$7B,$6E,$3C,$EB,$77,$C6,$7F,$D9,$D7,$C7,$1E,$F7
	.dc.b	$B9,$FF,$69,$61,$CE,$7D,$17,$C6,$57,$C7,$0F,$81,$1E,$0C,$7C,$EC
	.dc.b	$30,$4B,$E6,$E5,$F0,$C7,$C3,$C7,$C5,$3D,$B0,$F0,$61,$DE,$3C,$B3
	.dc.b	$A8,$F8,$55,$FE,$C0,$6E,$BF,$02,$BF,$D3,$F9,$7A,$F2,$C4,$7D,$E9
	.dc.b	$B7,$C7,$1F,$F5,$D5,$EE,$BF,$DE,$9F,$45,$FE,$57,$7F,$D5,$2D,$79
	.dc.b	$1A,$E7,$FF,$61,$7D,$03,$8B,$FD,$89,$5B,$C7,$FD,$AA,$DC,$BF,$DE
	.dc.b	$AD,$C0,$2F,$93,$E0,$F8,$5E,$64,$7F,$D7,$70,$FA,$2C,$47,$90,$52
	.dc.b	$B3,$87,$FB,$9A,$F8,$57,$52,$1C,$6B,$83,$DA,$FF,$DB,$DF,$45,$FD
	.dc.b	$0E,$3F,$D9,$6D,$7C,$0D,$BE,$DE,$EF,$0E,$89,$5C,$1F,$8F,$18,$8E
	.dc.b	$31,$67,$8C,$7F,$B6,$3E,$21,$F4,$5F,$EE,$F6,$C6,$63,$E8,$F5,$E5
	.dc.b	$91,$93,$E3,$C6,$1F,$8D,$3C,$9B,$3C,$9F,$F7,$3F,$A7,$FE,$29,$6F
	.dc.b	$8B,$FE,$FF,$8C,$47,$CF,$E4,$7D,$1E,$6A,$4F,$03,$FD,$E3,$3A,$70
	.dc.b	$7D,$7A,$E7,$93,$8B,$F0,$73,$FE,$83,$BB,$39,$8F,$C3,$DF,$37,$F0
	.dc.b	$1E,$A7,$D8,$6F,$1F,$0C,$39,$CE,$6F,$95,$3E,$8F,$FC,$E6,$FE,$4A
	.dc.b	$33,$3B,$AE,$37,$CE,$BF,$2F,$39,$7F,$33,$7C,$9F,$09,$CE,$B8,$B4
	.dc.b	$E3,$5C,$DF,$17,$89,$6B,$E1,$D7,$C2,$6C,$F6,$CB,$C5,$F4,$6F,$A3
	.dc.b	$FD,$54,$FA,$E7,$A7,$D1,$FE,$AA,$3F,$48,$DD,$3E,$0E,$1D,$25,$39
	.dc.b	$C7,$E8,$B6,$3F,$6A,$D7,$B1,$EB,$1F,$46,$FA,$3F,$67,$D1,$BE,$1E
	.dc.b	$9E,$F9,$D5,$71,$9D,$3C,$FE,$8B,$FC,$F4,$DF,$43,$8B,$4F,$07,$D1
	.dc.b	$BD,$CF,$94,$D8,$D1,$FC,$0F,$E7,$FF,$DD,$2F,$9D,$AE,$7E,$0F,$E9
	.dc.b	$7D,$17,$F5,$48,$D4,$7D,$59,$DF,$0F,$E2,$1B,$C1,$F9,$6B,$E1,$7C
	.dc.b	$CF,$EF,$33,$D9,$FF,$0A,$B5,$FC,$EB,$5F,$E4,$31,$FE,$DC,$7E,$63
	.dc.b	$5F,$E3,$8D,$AF,$CC,$AF,$67,$D3,$EF,$CA,$E2,$FB,$23,$F1,$DF,$45
	.dc.b	$F6,$4F,$F4,$8C,$FD,$96,$BF,$CA,$5D,$78,$FC,$3D,$69,$3B,$8F,$89
	.dc.b	$F4,$52,$3E,$63,$DB,$C5,$C4,$7D,$18,$B9,$FE,$C3,$05,$6A,$BE,$2D
	.dc.b	$7F,$87,$74,$34,$FF,$10,$57,$E1,$A7,$CD,$CB,$EC,$3D,$BF,$1B,$63
	.dc.b	$D9,$15,$F6,$47,$EB,$FE,$85,$E2,$DF,$99,$95,$F9,$B9,$CB,$67,$2F
	.dc.b	$F4,$D8,$4F,$39,$FD,$D7,$D0,$7F,$41,$E3,$FE,$32,$DB,$FE,$BA,$D5
	.dc.b	$6F,$9A,$E3,$E5,$49,$FD,$65,$6E,$B5,$18,$9D,$3F,$D7,$E7,$55,$FC
	.dc.b	$2F,$E7,$FF,$89,$70,$FF,$1E,$6E,$3F,$CB,$1B,$7F,$DC,$17,$D0,$71
	.dc.b	$F1,$64,$FE,$F6,$76,$9D,$B9,$F5,$7C,$1D,$B2,$30,$43,$FB,$A4,$FE
	.dc.b	$F1,$B1,$E7,$9F,$E4,$5F,$3B,$F9,$B9,$FE,$18,$D7,$FB,$E3,$72,$F6
	.dc.b	$75,$70,$EA,$95,$C1,$F8,$51,$88,$E3,$16,$78,$C7,$F1,$9F,$9E,$FE
	.dc.b	$38,$7E,$8A,$3F,$A2,$B7,$5F,$EA,$6E,$BA,$BF,$9C,$39,$AF,$F0,$CB
	.dc.b	$53,$D3,$AB,$64,$8F,$07,$CB,$57,$FC,$31,$3F,$A4,$7C,$93,$DA,$7E
	.dc.b	$2F,$C7,$49,$F8,$AF,$8B,$A7,$11,$F1,$A2,$E5,$EA,$9B,$97,$F6,$CD
	.dc.b	$55,$8B,$F4,$D5,$53,$A7,$B9,$F1,$E3,$51,$98,$F0,$19,$36,$60,$97
	.dc.b	$E3,$72,$FF,$72,$8E,$4D,$AF,$64,$EA,$FD,$11,$E5,$CB,$F8,$15,$6E
	.dc.b	$B4,$7F,$92,$AF,$9D,$FD,$8C,$3C,$F8,$71,$6D,$70,$3E,$BB,$87,$CF
	.dc.b	$3C,$5D,$57,$47,$B5,$79,$9D,$98,$3A,$61,$E6,$AF,$4C,$72,$9E,$18
	.dc.b	$7A,$63,$89,$F4,$23,$33,$C2,$33,$7E,$17,$CE,$73,$3B,$E1,$E9,$BE
	.dc.b	$87,$CF,$8E,$37,$F0,$5E,$07,$83,$0F,$33,$E9,$36,$7D,$B2,$9B,$AE
	.dc.b	$D8,$79,$6F,$B1,$E0,$70,$49,$FC,$69,$E0,$8D,$38,$3A,$B8,$7D,$A6
	.dc.b	$23,$93,$88,$D6,$1F,$0C,$F8,$97,$EA,$9F,$94,$77,$AF,$B4,$F3,$3D
	.dc.b	$D5,$E8,$8E,$B5,$CF,$0F,$A3,$5A,$9C,$C6,$23,$A9,$CC,$F2,$46,$9F
	.dc.b	$A6,$78,$4E,$47,$BE,$BD,$33,$F4,$E3,$95,$7A,$B0,$F9,$B1,$C0,$EA
	.dc.b	$58,$F3,$3F,$56,$B9,$BE,$03,$9D,$79,$E7,$71,$F6,$4F,$21,$EA,$9F
	.dc.b	$84,$F5,$9F,$24,$74,$8E,$07,$43,$C1,$7E,$F7,$31,$B3,$ED,$61,$F4
	.dc.b	$EF,$C3,$18,$C3,$FC,$C1,$F3,$DF,$DB,$AF,$FD,$70,$FF,$6C,$7F,$C9
	.dc.b	$8D,$73,$D1,$C0,$E3,$D3,$23,$71,$D3,$9A,$47,$27,$BF,$99,$88,$EF
	.dc.b	$17,$1F,$F4,$1E,$1F,$E7,$B9,$7C,$F3,$E7,$7F,$CE,$CD,$8F,$F3,$03
	.dc.b	$5F,$F6,$87,$CE,$FF,$66,$8C,$5F,$FA,$39,$BB,$FF,$A7,$1B,$5F,$EE
	.dc.b	$0A,$E1,$7D,$89,$1F,$E6,$D8,$FE,$E2,$7F,$93,$AF,$FD,$32,$DB,$FF
	.dc.b	$53,$B6,$7A,$73,$FF,$A3,$63,$10,$69,$23,$5C,$DB,$E6,$95,$E1,$8C
	.dc.b	$BE,$38,$CB,$F0,$23,$FD,$35,$F3,$6F,$9D,$A4,$9F,$F4,$37,$C6,$7F
	.dc.b	$A2,$67,$51,$FE,$89,$6B,$FE,$F9,$6E,$7F,$EB,$69,$E1,$5F,$EC,$4A
	.dc.b	$95,$C1,$B1,$97,$FD,$93,$19,$7F,$D9,$73,$FD,$B5,$57,$5E,$7F,$F5
	.dc.b	$94,$9B,$D1,$2B,$D3,$1A,$30,$F2,$D7,$AF,$2F,$2F,$06,$4F,$C0,$8E
	.dc.b	$A6,$1E,$00,$60,$9D,$C7,$FA,$B7,$E7,$1E,$0F,$B6,$38,$46,$1C,$35
	.dc.b	$F8,$4D,$7E,$43,$D9,$DD,$73,$51,$97,$8C,$67,$E6,$DF,$39,$FE,$DD
	.dc.b	$9F,$F7,$AC,$63,$C1,$F2,$3F,$E0,$8F,$8C,$B5,$BA,$7C,$83,$E1,$FF
	.dc.b	$D7,$1B,$7F,$B8,$E3,$1F,$07,$FE,$FD,$E2,$FF,$7F,$46,$FD,$0C,$57
	.dc.b	$C7,$8D,$BC,$6F,$9E,$1F,$32,$72,$D8,$CD,$F4,$58,$74,$9F,$2C,$3C
	.dc.b	$FC,$5F,$9F,$E6,$AC,$E7,$1F,$04,$62,$B2,$D8,$C9,$FE,$E5,$F9,$87
	.dc.b	$CD,$77,$D9,$F3,$5F,$E3,$2A,$D4,$E6,$31,$9F,$E1,$85,$F2,$AF,$C4
	.dc.b	$6A,$7F,$15,$87,$4F,$23,$D9,$7C,$31,$E9,$18,$C3,$94,$F5,$6E,$1C
	.dc.b	$E1,$F1,$AF,$81,$7C,$07,$1C,$7A,$FC,$D3,$C0,$EF,$F8,$23,$E4,$33
	.dc.b	$18,$7C,$6F,$D2,$DF,$F4,$19,$BE,$67,$F9,$74,$7B,$0C,$CE,$E7,$F4
	.dc.b	$D3,$A7,$E2,$70,$3E,$6F,$DF,$F0,$AF,$87,$F8,$33,$9C,$7A,$5F,$EB
	.dc.b	$EB,$51,$88,$F1,$7C,$DF,$C8,$C4,$FC,$AF,$9B,$77,$E6,$30,$CB,$E5
	.dc.b	$F3,$7F,$EF,$37,$A5,$70,$73,$E5,$7D,$49,$D6,$3F,$75,$F9,$B3,$A3
	.dc.b	$C9,$DA,$F8,$CF,$CB,$43,$E6,$BF,$83,$BA,$8C,$C6,$2F,$F1,$15,$C6
	.dc.b	$BE,$09,$C7,$C4,$FC,$2C,$66,$7F,$89,$CE,$4E,$87,$0A,$E9,$D5,$F8
	.dc.b	$59,$FD,$DE,$1E,$0C,$3F,$1F,$6F,$37,$CF,$F3,$6F,$8B,$CC,$61,$B3
	.dc.b	$E4,$9D,$3F,$4C,$E4,$66,$37,$87,$9A,$76,$FE,$96,$76,$E6,$7F,$43
	.dc.b	$F3,$6F,$84,$F8,$A6,$0B,$5F,$81,$8D,$3E,$F8,$D3,$4D,$3C,$2B,$F6
	.dc.b	$93,$C1,$F9,$3A,$5D,$3E,$37,$0F,$F0,$EC,$3E,$57,$4F,$ED,$23,$17
	.dc.b	$C2,$31,$E3,$7D,$A6,$D7,$03,$F2,$D5,$C2,$FF,$6D,$1F,$83,$9D,$DF
	.dc.b	$38,$FC,$E7,$81,$F7,$6B,$F7,$9D,$0F,$9A,$7C,$DF,$F0,$6A,$D5,$FF
	.dc.b	$0A,$9C,$BC,$B8,$FF,$8C,$46,$31,$FE,$29,$5C,$16,$1D,$70,$E9,$7C
	.dc.b	$1B,$87,$EC,$A7,$4F,$9B,$CD,$FE,$65,$93,$E6,$FA,$63,$FC,$4A,$73
	.dc.b	$18,$CB,$FB,$1D,$FE,$F2,$F8,$65,$EB,$BD,$BE,$DD,$3D,$DE,$FD,$BF
	.dc.b	$8A,$DC,$BD,$73,$FC,$A1,$FE,$E9,$F3,$7F,$DD,$7E,$6D,$F3,$5F,$C6
	.dc.b	$34,$FE,$3D,$8F,$F0,$9B,$D8,$AD,$BF,$DF,$67,$6E,$6B,$FC,$2D,$7D
	.dc.b	$5E,$98,$7F,$31,$BE,$35,$FC,$9A,$B4,$F3,$9D,$57,$28,$E8,$67,$0D
	.dc.b	$BC,$4C,$9F,$CF,$27,$8B,$FE,$2B,$AD,$E1,$E6,$C3,$C3,$3E,$C7,$C7
	.dc.b	$19,$C3,$C1,$18,$BE,$75,$DE,$31,$87,$8A,$39,$57,$BA,$31,$0C,$3C
	.dc.b	$37,$CA,$36,$7A,$CF,$49,$83,$D5,$3C,$A7,$8B,$63,$FC,$39,$1F,$E1
	.dc.b	$D7,$B1,$83,$FC,$8D,$F3,$5F,$C5,$FC,$FF,$F1,$63,$63,$3D,$5F,$E1
	.dc.b	$76,$CE,$BE,$13,$FC,$93,$1C,$9A,$FF,$95,$69,$18,$8F,$EB,$F1,$8E
	.dc.b	$0E,$47,$D6,$F9,$B7,$CB,$7F,$64,$DB,$F0,$31,$8B,$FB,$77,$C5,$FB
	.dc.b	$6D,$9D,$D7,$1D,$BF,$B6,$DF,$1B,$DE,$3E,$8B,$FC,$6E,$1C,$6F,$78
	.dc.b	$71,$BF,$F2,$A6,$1C,$6F,$78,$71,$BF,$5B,$6B,$77,$C5,$5F,$1A,$DE
	.dc.b	$3E,$AB,$E1,$5A,$BE,$17,$B8,$FF,$32,$63,$E5,$6D,$70,$9D,$47,$13
	.dc.b	$A5,$C7,$F3,$F1,$A3,$75,$CE,$38,$1F,$E2,$D8,$F8,$D3,$FD,$75,$BA
	.dc.b	$FA,$27,$9E,$1F,$E1,$8A,$D3,$67,$E4,$CF,$F8,$2E,$FB,$7C,$D7,$F9
	.dc.b	$3A,$3F,$C1,$CD,$7F,$C7,$8D,$BF,$ED,$9F,$36,$E2,$E9,$5F,$E8,$5A
	.dc.b	$D6,$1F,$E2,$86,$C6,$7A,$B6,$AF,$93,$E0,$F8,$4C,$D7,$FB,$74,$FF
	.dc.b	$6F,$4B,$E6,$F3,$32,$B8,$1F,$E2,$59,$D3,$83,$D3,$1B,$7F,$CA,$D1
	.dc.b	$A7,$FC,$B1,$F3,$7F,$E9,$3C,$FF,$C1,$0D,$CB,$FC,$82,$D9,$FF,$51
	.dc.b	$57,$FA,$D6,$B7,$D1,$EC,$AD,$51,$B5,$BB,$E7,$5B,$25,$7A,$7D,$FF
	.dc.b	$8B,$F0,$EF,$9B,$A4,$9F,$77,$5D,$7F,$CE,$13,$B7,$FC,$8C,$D7,$DF
	.dc.b	$1B,$CF,$FC,$49,$39,$6D,$7D,$4C,$BF,$C7,$57,$FE,$4F,$F5,$7F,$C4
	.dc.b	$8D,$F1,$7E,$CC,$62,$3F,$D0,$1C,$9F,$E5,$B7,$44,$3F,$D0,$91,$BC
	.dc.b	$3F,$C6,$93,$B9,$CD,$74,$23,$FE,$5F,$8D,$4F,$F8,$EF,$D5,$70,$7E
	.dc.b	$29,$B5,$FE,$65,$6B,$FE,$93,$F9,$AF,$F3,$1B,$99,$E0,$47,$FC,$D9
	.dc.b	$1A,$7F,$D3,$53,$CA,$BC,$15,$B9,$E9,$E2,$7E,$6A,$FD,$9C,$3F,$E8
	.dc.b	$66,$CF,$FB,$9D,$BE,$83,$E6,$FF,$47,$29,$DA,$F9,$BF,$D1,$BF,$EA
	.dc.b	$B6,$E5,$FE,$5B,$9D,$C6,$30,$FF,$4A,$B6,$BF,$DE,$CD,$7B,$6B,$FB
	.dc.b	$0F,$9A,$7C,$DF,$EC,$BE,$6D,$F0,$D4,$D4,$5B,$FF,$48,$7C,$DF,$F9
	.dc.b	$9D,$BE,$2E,$5F,$F6,$04,$65,$FF,$61,$7C,$DB,$E1,$3F,$7E,$D8,$E2
	.dc.b	$7F,$A6,$FE,$6F,$E8,$70,$7F,$9D,$D7,$CB,$7F,$06,$6C,$EF,$93,$06
	.dc.b	$66,$46,$48,$7F,$B4,$E7,$8B,$63,$8B,$FE,$AB,$F9,$BF,$F7,$53,$FE
	.dc.b	$BE,$3F,$D7,$EE,$1C,$07,$0E,$DA,$F0,$F8,$7F,$47,$D2,$48,$ED,$1E
	.dc.b	$BF,$81,$91,$F0,$5F,$77,$B9,$88,$F4,$45,$CF,$E2,$4F,$F2,$8F,$9B
	.dc.b	$FB,$87,$FA,$1E,$BF,$D8,$8D,$7F,$DE,$4E,$BC,$9E,$1E,$CC,$9D,$C7
	.dc.b	$33,$0F,$08,$C3,$F1,$7E,$73,$E0,$A3,$E1,$38,$69,$AD,$11,$96,$C7
	.dc.b	$9B,$E9,$3D,$47,$F4,$28,$CC,$7A,$39,$BD,$C4,$7D,$BF,$1F,$5D,$1F
	.dc.b	$02,$3E,$24,$7F,$4C,$F9,$D7,$8F,$7E,$66,$5F,$FB,$A6,$72,$D9,$CA
	.dc.b	$DF,$FD,$91,$2A,$E9,$52,$B9,$47,$0A,$DC,$FF,$5C,$F9,$D7,$99,$F1
	.dc.b	$5B,$7F,$29,$AF,$E3,$1A,$FE,$45,$FE,$93,$83,$E9,$3D,$87,$5A,$FE
	.dc.b	$D3,$F3,$DF,$A8,$F0,$FE,$A6,$08,$C3,$F5,$DA,$E7,$A3,$8F,$54,$8F
	.dc.b	$B9,$A7,$FB,$9E,$B6,$7E,$A1,$C7,$AB,$E0,$EF,$95,$C3,$1F,$0D,$FF
	.dc.b	$92,$A3,$07,$86,$B6,$FC,$D9,$DD,$7A,$A6,$4F,$03,$D5,$1C,$2F,$C3
	.dc.b	$19,$F9,$D7,$4F,$C2,$FA,$49,$39,$8F,$F0,$6C,$EA,$3A,$64,$AD,$C7
	.dc.b	$8C,$FC,$E4,$ED,$C0,$F9,$FF,$E5,$47,$EB,$70,$67,$F4,$23,$0D,$EA
	.dc.b	$FF,$0E,$C7,$0E,$AA,$62,$33,$E9,$4C,$3E,$05,$7F,$81,$5F,$D7,$2F
	.dc.b	$98,$D4,$79,$23,$4F,$96,$7B,$CE,$B1,$F2,$CF,$29,$C6,$3E,$7A,$B3
	.dc.b	$93,$C3,$3C,$9A,$43,$C0,$E5,$A6,$6B,$BB,$E3,$3D,$73,$D9,$C1,$C0
	.dc.b	$F5,$39,$BF,$59,$1C,$56,$5B,$3D,$EF,$D4,$D3,$D4,$75,$BF,$1D,$FC
	.dc.b	$98,$D1,$E2,$8C,$DF,$AD,$B1,$A9,$F1,$36,$7C,$4D,$35,$7E,$47,$C1
	.dc.b	$7E,$46,$9E,$4B,$EA,$FC,$03,$91,$A3,$F1,$D3,$CE,$36,$7A,$9C,$4F
	.dc.b	$E3,$EB,$D0,$DA,$E7,$5B,$6C,$65,$F0,$1D,$EF,$E7,$D7,$28,$F3,$4E
	.dc.b	$EB,$C2,$FA,$DB,$1A,$21,$E8,$3B,$98,$AF,$A0,$F7,$AF,$92,$D3,$E4
	.dc.b	$9C,$0F,$A9,$5D,$CF,$11,$1E,$97,$F0,$9A,$71,$AF,$AE,$F5,$3D,$D1
	.dc.b	$B7,$A9,$FA,$C7,$11,$C2,$FD,$B3,$A2,$47,$AA,$32,$FC,$88,$FC,$E9
	.dc.b	$E1,$BF,$94,$F0,$7D,$F3,$C1,$B7,$EB,$9D,$1C,$8F,$44,$F6,$BF,$2D
	.dc.b	$74,$BF,$05,$7B,$0F,$01,$C6,$BB,$3F,$30,$E7,$7D,$E7,$D4,$DA,$F1
	.dc.b	$CF,$D7,$6B,$E4,$3B,$1F,$C8,$B2,$7C,$9F,$F1,$CF,$A4,$93,$FC,$CE
	.dc.b	$3F,$97,$98,$7F,$8B,$46,$1F,$E7,$9F,$3F,$FD,$C1,$FE,$5A,$FF,$6D
	.dc.b	$73,$D5,$FE,$C4,$8D,$67,$FC,$31,$B9,$7F,$5E,$78,$57,$2A,$E1,$D5
	.dc.b	$C7,$E8,$24,$F8,$1E,$9E,$8E,$23,$C3,$17,$1E,$D3,$71,$ED,$36,$BF
	.dc.b	$C4,$18,$FF,$A0,$E7,$FB,$7F,$D0,$3E,$83,$FC,$44,$DA,$FF,$13,$B5
	.dc.b	$FF,$24,$B5,$FF,$27,$39,$C7,$F9,$4F,$D0,$3E,$93,$FD,$CA,$DA,$FE
	.dc.b	$C9,$5B,$C3,$FB,$FC,$FF,$6F,$9F,$EA,$ED,$BE,$7D,$5C,$BA,$B9,$B2
	.dc.b	$BF,$B7,$56,$EF,$9D,$78,$3C,$7F,$0F,$5C,$BD,$C6,$5E,$31,$97,$C3
	.dc.b	$19,$7C,$71,$9C,$7D,$B3,$FF,$00,$3E,$81,$F4,$3F,$E4,$96,$DF,$F9
	.dc.b	$39,$AF,$F9,$C1,$AF,$F9,$D5,$CD,$7F,$50,$F9,$17,$D2,$7F,$88,$A3
	.dc.b	$0F,$7B,$C6,$7F,$E2,$6A,$FE,$D2,$7E,$93,$0F,$F2,$E3,$53,$AE,$9E
	.dc.b	$31,$AE,$BA,$E9,$B3,$D3,$C7,$BE,$3C,$AF,$F0,$E4,$FF,$AB,$9B,$39
	.dc.b	$7F,$C1,$F0,$9E,$73,$27,$8C,$6F,$3F,$81,$3E,$2F,$A1,$7D,$0F,$FB
	.dc.b	$66,$FF,$C8,$71,$8F,$13,$FC,$F2,$BB,$DF,$1F,$8B,$E1,$7F,$E7,$DF
	.dc.b	$A3,$7D,$27,$F7,$C8,$FF,$26,$C6,$38,$3F,$D0,$6D,$8F,$F5,$2B,$5F
	.dc.b	$F2,$9F,$CB,$BE,$3B,$FD,$85,$5F,$E4,$48,$C7,$C5,$7F,$85,$B7,$FF
	.dc.b	$0B,$C6,$FD,$8C,$57,$FA,$3E,$36,$F1,$BE,$78,$7F,$8E,$67,$2D,$8C
	.dc.b	$DF,$45,$87,$49,$57,$97,$D5,$C3,$E7,$FA,$3B,$69,$B1,$86,$E3,$E3
	.dc.b	$6D,$F8,$5B,$8F,$85,$B3,$FE,$76,$FA,$07,$5F,$8E,$DF,$40,$F9,$3C
	.dc.b	$56,$AF,$FC,$C9,$39,$7F,$CC,$D7,$EE,$7F,$D0,$3F,$3D,$FC,$4D,$CE
	.dc.b	$FD,$63,$18,$FF,$97,$38,$DF,$41,$E1,$F8,$77,$D0,$78,$E7,$38,$FF
	.dc.b	$9C,$E7,$2F,$F9,$D7,$07,$D0,$75,$3D,$5F,$40,$EB,$C5,$7F,$9E,$E3
	.dc.b	$0D,$8F,$81,$F4,$1F,$E7,$A8,$C1,$FE,$81,$71,$A7,$F9,$C9,$BC,$7F
	.dc.b	$E4,$E5,$F1,$9C,$BC,$6F,$95,$D0,$F9,$1F,$BD,$8B,$E7,$BF,$76,$FF
	.dc.b	$A3,$67,$4F,$FA,$58,$D7,$87,$CA,$B8,$D7,$0C,$FF,$D0,$77,$C1,$FF
	.dc.b	$44,$5F,$03,$A5,$70,$BE,$51,$D6,$F9,$67,$FE,$66,$87,$93,$E6,$FA
	.dc.b	$07,$C6,$7E,$0E,$72,$79,$63,$2F,$FA,$94,$CE,$5E,$28,$CB,$FE,$98
	.dc.b	$8C,$DF,$FA,$09,$4E,$A3,$C5,$7E,$1B,$F0,$36,$BF,$D6,$5F,$40,$F6
	.dc.b	$FE,$19,$F9,$77,$11,$ED,$8C,$1A,$78,$B7,$1F,$F5,$5C,$F0,$7E,$4E
	.dc.b	$57,$1F,$F4,$4C,$DB,$57,$FA,$BA,$B0,$E3,$87,$F7,$73,$F3,$7B,$7F
	.dc.b	$55,$18,$9E,$B1,$8F,$07,$ED,$B6,$B8,$1F,$EC,$FA,$E1,$8F,$C7,$C3
	.dc.b	$F5,$D8,$7F,$B0,$A7,$35,$E4,$3F,$19,$7F,$B9,$F9,$DF,$C2,$61,$FE
	.dc.b	$B0,$E9,$7D,$03,$E8,$35,$C7,$FC,$46,$73,$18,$C7,$FD,$CF,$18,$7F
	.dc.b	$DC,$58,$74,$7E,$1C,$67,$3F,$F6,$AC,$E5,$FE,$77,$87,$7A,$E3,$87
	.dc.b	$78,$C4,$F0,$F8,$2F,$E4,$59,$3E,$83,$94,$B1,$56,$A7,$31,$8C,$7F
	.dc.b	$DF,$B7,$C7,$1F,$85,$7B,$7E,$26,$9F,$17,$E0,$2D,$FB,$1B,$97,$AA
	.dc.b	$7F,$91,$3F,$80,$FA,$0F,$C0,$FD,$03,$E8,$3F,$07,$3E,$5A,$DA,$AD
	.dc.b	$CE,$B4,$FB,$37,$C6,$B8,$9B,$0E,$5E,$98,$F2,$77,$75,$C9,$C2,$F9
	.dc.b	$BC,$EB,$95,$F8,$27,$35,$C6,$31,$5D,$0E,$47,$E7,$9C,$BE,$E8,$E1
	.dc.b	$87,$FB,$EA,$FC,$D3,$F1,$4E,$27,$91,$CB,$E3,$BF,$85,$3C,$5F,$1C
	.dc.b	$70,$BE,$EF,$3A,$F4,$D6,$E3,$CD,$18,$C3,$E1,$46,$EF,$AC,$62,$3C
	.dc.b	$07,$AE,$B9,$3E,$B3,$E3,$9E,$92,$E1,$87,$07,$DE,$F9,$FF,$E4,$BE
	.dc.b	$5F,$06,$C7,$D0,$6B,$F7,$1A,$F1,$F6,$5B,$F1,$6C,$71,$6B,$C5,$AF
	.dc.b	$18,$FC,$07,$D0,$3E,$6B,$F2,$3A,$FE,$1E,$31,$5F,$85,$C1,$87,$6A
	.dc.b	$FC,$64,$63,$0E,$78,$72,$D7,$F3,$78,$F2,$C7,$ED,$DE,$DB,$8F,$3A
	.dc.b	$DD,$F1,$9F,$E4,$B7,$C7,$1E,$F7,$B9,$FE,$53,$87,$39,$F4,$5F,$19
	.dc.b	$5F,$1C,$3E,$04,$78,$31,$F3,$B0,$C1,$2F,$9B,$97,$C3,$1F,$0F,$1F
	.dc.b	$14,$F6,$C3,$C1,$87,$78,$F2,$CE,$A3,$E1,$53,$5F,$7D,$7F,$42,$8C
	.dc.b	$3F,$02,$FF,$CE,$39,$7B,$72,$DC,$FE,$02,$ED,$F0,$8F,$11,$CE,$3D
	.dc.b	$F9,$7F,$01,$FA,$0F,$F2,$2C,$E4,$FE,$A2,$E1,$E8,$63,$A7,$FD,$03
	.dc.b	$F3,$AE,$2F,$F4,$1D,$6F,$1F,$03,$71,$FE,$32,$DC,$3F,$AE,$DF,$17
	.dc.b	$C0,$23,$FD,$9C,$F8,$BE,$0F,$A9,$E0,$FA,$0B,$1C,$27,$6F,$19,$D3
	.dc.b	$97,$6F,$A8,$D1,$E0,$A5,$7F,$A7,$3E,$83,$FA,$0C,$F9,$1A,$F1,$6E
	.dc.b	$3F,$36,$9E,$E7,$1E,$89,$7C,$5F,$8F,$18,$8E,$51,$6B,$94,$FF,$AB
	.dc.b	$3E,$19,$F4,$14,$97,$E7,$EA,$AF,$AE,$76,$FC,$A6,$BE,$FC,$7F,$C8
	.dc.b	$8E,$6B,$8E,$1F,$85,$AD,$BE,$BB,$FF,$5E,$FD,$07,$F8,$D2,$FE,$AB
	.dc.b	$5F,$94,$D9,$FC,$77,$2F,$F9,$AA,$77,$E2,$F2,$23,$F7,$A3,$4F,$28
	.dc.b	$CB,$4D,$7A,$1F,$EC,$EF,$2F,$EE,$93,$87,$FD,$A3,$F3,$FF,$E6,$FA
	.dc.b	$FC,$4C,$E5,$B3,$AF,$13,$93,$23,$27,$E4,$2B,$93,$F8,$57,$D9,$CB
	.dc.b	$FD,$C1,$B1,$FE,$34,$6F,$A2,$FA,$0F,$CE,$4A,$76,$BE,$83,$F3,$CF
	.dc.b	$E4,$5B,$B7,$ED,$67,$71,$8C,$3F,$BF,$B6,$BF,$C8,$6D,$7B,$6B,$FE
	.dc.b	$F2,$FA,$07,$CF,$FF,$BD,$3E,$83,$FD,$DD,$5F,$9F,$6E,$5F,$C0,$EB
	.dc.b	$73,$FB,$09,$CE,$3F,$E0,$EF,$A0,$FF,$3A,$CF,$ED,$1A,$F3,$F8,$4F
	.dc.b	$F8,$03,$18,$C3,$F8,$77,$D0,$3D,$BF,$F0,$BC,$61,$DF,$A5,$8C,$7F
	.dc.b	$C3,$7F,$41,$F4,$2B,$F8,$B2,$F9,$8D,$67,$FE,$62,$8C,$3F,$C1,$9E
	.dc.b	$15,$CB,$97,$72,$9D,$BF,$BF,$6D,$F6,$9D,$C7,$29,$ED,$78,$33,$3D
	.dc.b	$3E,$73,$FD,$C5,$3F,$C5,$9A,$FF,$0B,$6D,$FA,$BC,$AE,$1F,$05,$FD
	.dc.b	$F6,$77,$5F,$E3,$26,$BF,$E3,$C6,$BE,$FF,$80,$F2,$91,$F8,$1E,$DD
	.dc.b	$7D,$91,$70,$F9,$33,$F0,$BE,$7F,$F0,$C7,$FA,$66,$BF,$96,$B5,$FE
	.dc.b	$44,$EB,$C9,$E1,$EC,$C9,$DC,$73,$30,$F0,$8C,$3F,$0F,$E7,$FE,$31
	.dc.b	$1E,$FE,$3D,$7F,$C8,$F9,$A7,$E2,$12,$3F,$C4,$B8,$7E,$49,$FF,$50
	.dc.b	$B5,$34,$7C,$5F,$BF,$9E,$4F,$B9,$F8,$51,$F8,$4F,$9D,$78,$F7,$E5
	.dc.b	$E5,$FF,$50,$9C,$B6,$72,$FF,$47,$84,$F3,$99,$3C,$63,$7B,$FF,$6F
	.dc.b	$AE,$73,$FD,$3F,$E7,$BF,$A3,$F3,$FF,$9E,$9B,$1F,$EB,$E6,$A9,$7C
	.dc.b	$C7,$E7,$27,$24,$F3,$7D,$91,$FA,$0F,$9D,$FC,$D5,$9A,$4E,$9D,$74
	.dc.b	$C9,$FE,$DF,$39,$6F,$4F,$3E,$B9,$18,$21,$FD,$A6,$7F,$80,$69,$FE
	.dc.b	$94,$6C,$7F,$B6,$1F,$F2,$1B,$C3,$5F,$F0,$94,$6A,$3C,$D5,$E9,$BF
	.dc.b	$DC,$FC,$7B,$A6,$92,$BC,$FD,$55,$FE,$F5,$3B,$7F,$C1,$2D,$7C,$D1
	.dc.b	$FC,$13,$E7,$3F,$94,$1F,$A1,$CB,$FC,$48,$DC,$FF,$D5,$AE,$BA,$F8
	.dc.b	$75,$CC,$3F,$C2,$B3,$CB,$AF,$A1,$23,$3F,$25,$5F,$F0,$74,$FF,$81
	.dc.b	$1E,$33,$E3,$9D,$63,$E2,$8B,$8F,$8A,$6D,$05,$F3,$7A,$D7,$37,$A9
	.dc.b	$0C,$9E,$43,$17,$DD,$B7,$DD,$A7,$7A,$ED,$3B,$8F,$5D,$F9,$1F,$1D
	.dc.b	$DA,$4D,$8C,$36,$7A,$9E,$A7,$0F,$98,$EA,$6A,$7C,$37,$F5,$2B,$DD
	.dc.b	$3B,$BF,$21,$27,$D5,$19,$78,$3C,$EF,$BC,$EE,$3D,$D1,$89,$DC,$78
	.dc.b	$8E,$47,$C3,$AF,$5B,$E5,$6B,$A3,$E1,$9D,$CD,$14,$D5,$FB,$1E,$87
	.dc.b	$D5,$33,$1D,$2F,$CD,$1D,$0E,$37,$E7,$78,$9E,$78,$DD,$75,$8E,$11
	.dc.b	$F9,$E9,$E1,$18,$BF,$8F,$3C,$2F,$C1,$1F,$0A,$31,$3E,$56,$BE,$23
	.dc.b	$DA,$53,$D2,$7B,$DD,$9E,$4A,$F0,$DF,$94,$FC,$3C,$F2,$9D,$36,$BC
	.dc.b	$47,$98,$AF,$C0,$8D,$BE,$03,$C4,$F8,$23,$71,$F8,$F3,$A1,$C5,$E8
	.dc.b	$D2,$1F,$14,$FC,$29,$C0,$C1,$CC,$F0,$47,$88,$F0,$3C,$A7,$DA,$EA
	.dc.b	$7C,$71,$89,$DC,$F4,$8C,$4E,$AF,$E9,$CF,$36,$BB,$6C,$FC,$B6,$9F
	.dc.b	$3A,$FF,$95,$FC,$E4,$AF,$E4,$AF,$F2,$3E,$AC,$47,$F2,$A8,$B8,$FF
	.dc.b	$24,$9F,$F1,$5F,$CD,$FF,$5A,$8F,$F6,$53,$FD,$8A,$BF,$D6,$ED,$7F
	.dc.b	$D5,$8E,$F2,$FE,$31,$08,$C3,$FC,$6D,$E1,$3D,$3A,$7B,$F5,$78,$39
	.dc.b	$E5,$F6,$7C,$7E,$66,$23,$BC,$5B,$F2,$CD,$C7,$C3,$8F,$F6,$AB,$E4
	.dc.b	$7A,$27,$FC,$8F,$F3,$2F,$9B,$A4,$9F,$F6,$9F,$CD,$7F,$6C,$9D,$47
	.dc.b	$FB,$59,$AF,$FB,$2D,$B8,$FF,$BB,$E7,$FB,$99,$FD,$8A,$7F,$DA,$0D
	.dc.b	$8D,$73,$FF,$96,$23,$FB,$21,$9C,$BE,$4B,$6B,$FA,$13,$53,$C3,$97
	.dc.b	$97,$3C,$9F,$4C,$75,$30,$F8,$A3,$0E,$8F,$7B,$AC,$3D,$53,$FE,$6C
	.dc.b	$F9,$A7,$CC,$D2,$5F,$FB,$E3,$E3,$3F,$D2,$93,$A8,$FC,$09,$F8,$27
	.dc.b	$0E,$03,$87,$6D,$7F,$13,$1F,$C6,$CF,$D2,$BF,$D8,$DA,$9D,$73,$7F
	.dc.b	$A0,$9C,$91,$C7,$26,$7C,$29,$5E,$98,$CB,$EB,$8C,$AC,$9A,$F9,$2B
	.dc.b	$73,$98,$F3,$D7,$0A,$DC,$F9,$EB,$FD,$2B,$F4,$9E,$78,$F8,$41,$C3
	.dc.b	$F1,$5A,$FC,$A6,$BF,$AC,$F9,$A7,$CD,$7F,$B6,$23,$0D,$7F,$C1,$10
	.dc.b	$EE,$7C,$5F,$82,$38,$F8,$5F,$4F,$E4,$1F,$0F,$FE,$B9,$DB,$FC,$21
	.dc.b	$AF,$AA,$73,$F0,$7E,$C8,$C5,$6F,$3F,$84,$2B,$77,$C8,$FF,$0D,$3D
	.dc.b	$BC,$FE,$51,$8A,$61,$8B,$E5,$1D,$70,$EF,$39,$C3,$CF,$E7,$5A,$7C
	.dc.b	$C5,$8C,$37,$1F,$16,$3D,$23,$0F,$A2,$FE,$0C,$63,$1F,$6B,$6B,$CC
	.dc.b	$F9,$69,$F3,$0E,$8C,$7C,$C3,$E3,$35,$39,$C3,$C3,$6A,$D4,$62,$FD
	.dc.b	$CF,$F2,$8F,$97,$FE,$28,$E7,$7F,$04,$62,$FE,$79,$F7,$3E,$63,$C9
	.dc.b	$EF,$BE,$07,$CD,$CA,$9C,$C7,$2F,$98,$99,$79,$3E,$61,$CD,$8C,$BA
	.dc.b	$C6,$27,$53,$F0,$7E,$63,$FC,$ED,$18,$3D,$AE,$2B,$CA,$DE,$3F,$D1
	.dc.b	$C2,$1F,$09,$CF,$C3,$F9,$9C,$CF,$87,$FC,$16,$2F,$96,$FD,$F3,$F2
	.dc.b	$A7,$4F,$DC,$35,$E1,$F2,$AE,$35,$C3,$2F,$8F,$7C,$1F,$CB,$DF,$03
	.dc.b	$A5,$70,$BE,$51,$D6,$F9,$67,$EA,$87,$A9,$E6,$F9,$87,$C2,$7E,$16
	.dc.b	$72,$79,$63,$2F,$E0,$8C,$E5,$E2,$8C,$BF,$72,$33,$7F,$C2,$94,$EA
	.dc.b	$3C,$57,$E1,$BF,$03,$6B,$F1,$FF,$30,$F5,$BE,$19,$F9,$97,$11,$ED
	.dc.b	$8C,$1A,$78,$B7,$1F,$C7,$CF,$07,$E4,$E5,$71,$FB,$B3,$6F,$FD,$30
	.dc.b	$7D,$9B,$FD,$47,$0F,$EF,$A3,$E6,$76,$E3,$C1,$FA,$ED,$8C,$9F,$A8
	.dc.b	$9E,$19,$78,$27,$4D,$9D,$57,$27,$F5,$FF,$2B,$F7,$F2,$FE,$67,$D2
	.dc.b	$F9,$77,$CC,$7E,$EE,$B5,$7F,$EA,$FA,$D4,$E7,$8F,$F8,$A4,$62,$7A
	.dc.b	$C6,$2F,$9C,$60,$FE,$1F,$3B,$CB,$F6,$F8,$76,$7F,$99,$D7,$07,$3E
	.dc.b	$EF,$F2,$1C,$9F,$31,$B9,$5A,$B5,$39,$8C,$63,$FC,$7A,$F8,$E5,$ED
	.dc.b	$BD,$BE,$FD,$7E,$07,$B9,$6F,$E3,$B7,$2F,$3C,$FF,$1F,$7F,$DA,$5F
	.dc.b	$31,$FE,$D3,$F9,$57,$CB,$7F,$8F,$B4,$FE,$37,$5F,$C5,$0D,$89,$DB
	.dc.b	$FE,$DB,$8D,$C6,$23,$FC,$8F,$5D,$6B,$A5,$FF,$6A,$AD,$57,$47,$85
	.dc.b	$F2,$AF,$F0,$65,$F1,$AE,$D7,$C6,$73,$1E,$37,$AE,$3B,$8F,$09,$93
	.dc.b	$8B,$97,$C1,$1C,$1D,$1C,$30,$F1,$D7,$78,$FF,$10,$DF,$5C,$3C,$D1
	.dc.b	$83,$8B,$87,$95,$F5,$BF,$7C,$2B,$95,$F5,$8D,$9E,$63,$C6,$79,$8C
	.dc.b	$4F,$29,$FF,$07,$C7,$D4,$3F,$C2,$2E,$1E,$C7,$F8,$D7,$E5,$FF,$8E
	.dc.b	$F9,$7F,$E1,$66,$C6,$7A,$3F,$C0,$ED,$9D,$7A,$DF,$E2,$78,$E4,$D7
	.dc.b	$FC,$87,$48,$C4,$7F,$A1,$A3,$1C,$1C,$8F,$A9,$F3,$0F,$92,$FF,$45
	.dc.b	$E7,$F8,$78,$C5,$FD,$9B,$E2,$FD,$46,$CE,$EB,$8E,$BF,$E9,$4B,$E3
	.dc.b	$7B,$C7,$D1,$7F,$93,$C3,$8D,$EF,$0E,$36,$18,$72,$BE,$18,$72,$BF
	.dc.b	$63,$6B,$85,$6E,$7A,$47,$A6,$74,$B0,$F0,$BC,$30,$F0,$BE,$06,$C6
	.dc.b	$A3,$84,$F2,$BE,$D1,$C4,$ED,$8C,$7F,$33,$1A,$7A,$19,$C7,$E4,$3F
	.dc.b	$E1,$88,$F8,$F3,$FD,$E9,$BB,$FB,$A3,$EB,$E3,$FE,$C0,$74,$D8,$F7
	.dc.b	$D7,$F8,$63,$1F,$47,$CC,$7F,$90,$23,$FC,$4A,$D7,$FC,$3A,$DB,$FF
	.dc.b	$04,$FC,$B3,$8B,$A5,$7D,$3A,$D6,$1F,$ED,$06,$C6,$7A,$76,$AF,$93
	.dc.b	$E0,$F7,$73,$46,$9C,$25,$F3,$18,$AE,$04,$3F,$C6,$71,$A7,$19,$7F
	.dc.b	$88,$2F,$A4,$63,$1F,$F7,$2D,$6E,$3F,$C7,$7F,$31,$FC,$F5,$FF,$23
	.dc.b	$B5,$CF,$47,$8F,$DA,$F5,$CE,$71,$FF,$16,$B7,$0F,$F7,$8B,$6F,$F3
	.dc.b	$18,$2F,$8E,$12,$7B,$BD,$5F,$F2,$BF,$C1,$3E,$63,$E4,$B6,$33,$1F
	.dc.b	$17,$AF,$2C,$8C,$9F,$E6,$0C,$3F,$C7,$8E,$0F,$F2,$CD,$73,$6C,$F3
	.dc.b	$7F,$CC,$FF,$31,$FE,$2C,$6F,$8B,$F6,$A3,$11,$FE,$62,$E4,$FF,$30
	.dc.b	$3A,$21,$FE,$66,$8D,$ED,$FE,$52,$9D,$CE,$6B,$A1,$1F,$F3,$BC,$6A
	.dc.b	$7F,$CB,$3E,$57,$DA,$9F,$F3,$8B,$63,$3C,$5F,$8E,$F9,$7F,$F3,$5D
	.dc.b	$7F,$A0,$2B,$5C,$3F,$C2,$63,$FC,$C3,$5F,$94,$9C,$C7,$F0,$DA,$E2
	.dc.b	$F6,$9D,$11,$FF,$45,$C6,$AF,$FC,$DA,$FB,$AF,$B3,$67,$A6,$3F,$A8
	.dc.b	$F4,$5F,$31,$FA,$69,$56,$D7,$CC,$7E,$99,$FF,$47,$B7,$1E,$F8,$7E
	.dc.b	$B2,$31,$8F,$B5,$B5,$FC,$9D,$AF,$69,$FD,$9F,$CB,$BE,$63,$F6,$BF
	.dc.b	$30,$F8,$2A,$6A,$2D,$FF,$AA,$BE,$63,$FC,$E2,$DF,$13,$2F,$FA,$92
	.dc.b	$32,$FF,$A9,$BE,$61,$ED,$FF,$02,$6C,$71,$3F,$D6,$BE,$23,$E3,$3F
	.dc.b	$83,$B6,$33,$18,$E4,$CB,$23,$27,$28,$C3,$FE,$B6,$9E,$4D,$9E,$4F
	.dc.b	$FB,$13,$E6,$3E,$D7,$0F,$E1,$57,$C9,$FF,$18,$6C,$EF,$D1,$C1,$99
	.dc.b	$91,$93,$FD,$9F,$F3,$0F,$84,$FF,$00,$B6,$3F,$24,$7E,$4D,$C3,$80
	.dc.b	$E1,$DB,$AE,$6F,$F5,$DE,$0F,$A1,$F3,$1F,$CB,$A7,$59,$7A,$FE,$0B
	.dc.b	$D4,$4C,$3E,$07,$B9,$5F,$F6,$DE,$1F,$EE,$B7,$E2,$4F,$C1,$9D,$63
	.dc.b	$E0,$8B,$7F,$1E,$7F,$9E,$FC,$C7,$E3,$8F,$F5,$05,$FF,$BC,$1A,$FF
	.dc.b	$BA,$1D,$79,$7C,$3D,$89,$3B,$8F,$57,$CC,$FC,$82,$47,$C5,$3E,$2B
	.dc.b	$87,$A7,$D0,$F8,$8E,$2F,$87,$AB,$CA,$47,$C9,$F1,$B5,$F2,$45,$D7
	.dc.b	$FD,$CF,$92,$B1,$A7,$F0,$98,$FD,$35,$F5,$39,$9F,$E1,$5A,$FF,$77
	.dc.b	$FC,$D3,$C8,$FE,$C2,$BE,$56,$98,$89,$5C,$70,$FE,$2D,$5B,$8C,$1F
	.dc.b	$E2,$17,$0A,$FB,$D7,$F6,$EF,$9A,$FE,$A1,$2A,$7D,$0F,$88,$D7,$9D
	.dc.b	$27,$A9,$B8,$D3,$8A,$E7,$F3,$7F,$A6,$8F,$F1,$93,$53,$4F,$D3,$BE
	.dc.b	$5C,$BF,$98,$46,$7A,$BF,$C0,$6E,$FA,$AC,$4A,$F1,$CE,$5B,$D9,$C3
	.dc.b	$C1,$91,$82,$1D,$CF,$2B,$A7,$07,$9A,$3F,$C2,$7F,$20,$E8,$F1,$FD
	.dc.b	$0C,$AF,$87,$39,$6C,$E5,$E7,$82,$B9,$CF,$F8,$7B,$E7,$3F,$97,$9F
	.dc.b	$A0,$E2,$6B,$FD,$46,$30,$DE,$9E,$07,$1E,$A9,$97,$D4,$AE,$73,$88
	.dc.b	$F5,$C7,$27,$FC,$C8,$E9,$E1,$3B,$BF,$0C,$6A,$BB,$BA,$21,$D1,$CB
	.dc.b	$DA,$79,$5F,$9E,$3C,$93,$87,$C3,$3C,$2B,$BC,$2F,$E1,$CE,$E3,$15
	.dc.b	$BB,$F8,$93,$9A,$F5,$57,$84,$F2,$DF,$48,$CB,$67,$85,$F3,$AF,$19
	.dc.b	$C2,$FD,$0F,$B4,$C1,$A3,$04,$AE,$8E,$8F,$39,$F0,$0C,$5F,$D1,$72
	.dc.b	$71,$9F,$5B,$AA,$F2,$3F,$0E,$FC,$EF,$33,$DC,$4B,$F7,$4F,$0B,$EC
	.dc.b	$47,$15,$CA,$BC,$07,$AE,$BD,$2E,$0F,$2B,$82,$5F,$C4,$7B,$C7,$B2
	.dc.b	$7C,$15,$F4,$DB,$19,$3D,$13,$C1,$F4,$57,$C0,$7C,$17,$F8,$07,$E1
	.dc.b	$1C,$23,$53,$E2,$8F,$CA,$19,$3D,$51,$A3,$CC,$47,$24,$3D,$AE,$5F
	.dc.b	$9C,$78,$0F,$1D,$7C,$79,$F4,$1E,$18,$D9,$E8,$BF,$80,$F8,$CE,$A7
	.dc.b	$7B,$FA,$4E,$5F,$AB,$5F,$1A,$3E,$19,$B9,$FA,$EF,$8E,$BC,$37,$ED
	.dc.b	$8D,$1C,$EB,$CD,$1E,$BB,$EB,$3D,$EF,$CA,$F9,$CF,$A8,$53,$C2,$7B
	.dc.b	$2F,$BB,$F3,$8C,$CF,$A9,$F8,$A7,$96,$BC,$17,$E2,$8E,$15,$F1,$EB
	.dc.b	$83,$D2,$3B,$57,$AD,$F8,$E4,$9F,$4B,$F8,$4F,$4F,$F9,$31,$1F,$E3
	.dc.b	$DF,$39,$5F,$E3,$B1,$6F,$F8,$04,$FF,$38,$F9,$CF,$E5,$DF,$43,$2F
	.dc.b	$F9,$BC,$65,$FE,$51,$19,$7F,$A0,$EE,$9F,$F6,$9F,$CE,$FF,$75,$9F
	.dc.b	$EC,$2F,$F6,$CC,$3F,$8C,$B7,$3F,$F4,$43,$B9,$FE,$96,$D9,$DF,$57
	.dc.b	$43,$AF,$54,$AE,$B1,$97,$BC,$78,$BC,$59,$1E,$27,$CB,$E3,$62,$3B
	.dc.b	$C5,$BF,$2C,$FF,$BB,$3E,$71,$F3,$BF,$EE,$EF,$9D,$7D,$0D,$35,$13
	.dc.b	$1F,$EF,$F3,$A6,$CE,$5F,$EC,$F0,$E9,$E8,$48,$E3,$D5,$5D,$C5,$CB
	.dc.b	$FA,$0D,$7F,$A0,$9B,$19,$DF,$FD,$65,$3D,$6B,$99,$E4,$F1,$BD,$5D
	.dc.b	$32,$30,$43,$CA,$E5,$F3,$C7,$FA,$F7,$0F,$EA,$3F,$3A,$F9,$EF,$F2
	.dc.b	$45,$23,$0F,$F9,$2D,$AF,$F9,$C5,$AF,$F9,$D9,$CC,$FF,$5C,$F9,$17
	.dc.b	$D0,$F7,$CF,$FD,$99,$18,$BF,$F1,$5D,$F0,$9C,$F5,$6E,$42,$9D,$59
	.dc.b	$8C,$36,$13,$4C,$44,$9F,$F1,$CC,$E9,$B3,$97,$FC,$43,$1D,$9C,$9C
	.dc.b	$3C,$8E,$DE,$44,$8C,$10,$E2,$E5,$E7,$1E,$7F,$9D,$7C,$F7,$FB,$A2
	.dc.b	$7F,$DD,$D1,$8F,$13,$FC,$F0,$BB,$9F,$21,$8B,$E1,$7F,$13,$E8,$1F
	.dc.b	$43,$FE,$0A,$8F,$F7,$C4,$63,$7F,$F4,$13,$63,$FD,$50,$D7,$FD,$FD
	.dc.b	$F2,$EF,$8E,$FF,$68,$D7,$F9,$AA,$31,$F1,$5F,$E2,$4E,$5F,$F1,$2C
	.dc.b	$6F,$D8,$C5,$7D,$68,$DB,$C6,$F9,$E1,$FE,$4F,$9C,$B6,$33,$7D,$16
	.dc.b	$1D,$25,$5D,$EF,$A4,$E6,$31,$3C,$E3,$82,$F9,$E7,$C8,$E1,$B7,$FE
	.dc.b	$7A,$F9,$F7,$5E,$6B,$8F,$CF,$FF,$06,$AF,$F3,$FB,$70,$FF,$32,$D7
	.dc.b	$A2,$7F,$2B,$09,$D6,$9F,$E3,$CF,$85,$7C,$F7,$91,$B7,$FE,$8A,$E3
	.dc.b	$7C,$FE,$6B,$C3,$F3,$F2,$BF,$D1,$FF,$3E,$E9,$C6,$9F,$E7,$68,$C4
	.dc.b	$EA,$7E,$0F,$CF,$FF,$9D,$E3,$07,$FA,$61,$C5,$7F,$9F,$9B,$A7,$F9
	.dc.b	$49,$7C,$5F,$2F,$13,$E5,$F2,$BE,$43,$EA,$62,$F9,$EF,$DA,$3F,$EA
	.dc.b	$49,$D3,$FE,$B2,$35,$E0,$F2,$AE,$35,$C3,$3F,$F4,$7D,$F0,$7F,$D2
	.dc.b	$97,$C0,$E9,$5C,$2F,$94,$75,$BE,$59,$FF,$9A,$A1,$E4,$F9,$BE,$7D
	.dc.b	$F1,$7F,$84,$9C,$9E,$58,$CB,$FE,$BC,$33,$97,$8A,$32,$FF,$A9,$23
	.dc.b	$37,$FE,$86,$53,$A8,$F1,$5F,$86,$FC,$0D,$AF,$F6,$4F,$CF,$BD,$8F
	.dc.b	$86,$7E,$61,$C4,$7B,$63,$06,$9E,$2D,$C7,$FD,$6F,$3C,$1F,$93,$95
	.dc.b	$C7,$FD,$1B,$36,$FF,$95,$9F,$66,$FF,$4F,$C3,$FB,$08,$F9,$9D,$58
	.dc.b	$ED,$FA,$ED,$8C,$9F,$EE,$29,$E1,$97,$82,$74,$D9,$D5,$72,$7F,$DD
	.dc.b	$3F,$39,$F7,$F2,$FF,$57,$F4,$BE,$79,$F3,$FF,$BA,$AD,$5F,$F4,$2A
	.dc.b	$D4,$E7,$8F,$F8,$34,$62,$7A,$C6,$2F,$9C,$60,$FE,$97,$87,$82,$33
	.dc.b	$97,$FB,$83,$1E,$EF,$F3,$08,$CC,$FF,$AD,$A5,$F0,$5F,$C5,$61,$F3
	.dc.b	$FC,$2D,$5A,$C3,$B9,$E0,$9D,$46,$67,$FD,$E5,$5C,$B1,$F8,$18,$75
	.dc.b	$BD,$EB,$F0,$FD,$EF,$88,$DC,$BE,$5B,$87,$BE,$3F,$C5,$07,$CF,$FF
	.dc.b	$6A,$F9,$E7,$CF,$FF,$18,$D7,$F9,$1D,$7F,$C0,$03,$FB,$7D,$6E,$75
	.dc.b	$87,$FB,$92,$F8,$D7,$5A,$D3,$FD,$DA,$F9,$CE,$A3,$F9,$2C,$E9,$F8
	.dc.b	$B5,$CE,$FC,$D1,$E6,$15,$C6,$F9,$CE,$AF,$94,$62,$BC,$47,$09,$F3
	.dc.b	$E1,$DE,$3D,$66,$4F,$36,$1E,$17,$8E,$1D,$E3,$11,$E8,$38,$1E,$89
	.dc.b	$E8,$F9,$E7,$84,$63,$0F,$2D,$72,$AF,$0D,$69,$E6,$62,$FC,$EE,$0E
	.dc.b	$8E,$23,$E5,$47,$A8,$F8,$E7,$C7,$2D,$7F,$8E,$1C,$1F,$D5,$3E,$7B
	.dc.b	$F9,$1F,$91,$C1,$B1,$F4,$1A,$FE,$1D,$AF,$1F,$5D,$C1,$C5,$B1,$C5
	.dc.b	$AF,$16,$BC,$63,$F0,$1F,$3E,$F9,$9F,$EC,$F9,$7E,$16,$31,$5F,$85
	.dc.b	$C3,$FC,$D5,$77,$0E,$F7,$BC,$38,$E7,$FD,$C3,$1E,$38,$FD,$6C,$78
	.dc.b	$B7,$0E,$35,$BB,$E3,$3F,$E4,$CB,$E3,$87,$E5,$F0,$E2,$DC,$38,$D6
	.dc.b	$EF,$8A,$BE,$38,$7B,$E3,$C1,$8F,$C0,$AD,$D6,$A7,$FC,$B7,$8F,$6B
	.dc.b	$E1,$8F,$B6,$78,$1F,$07,$08,$FE,$4A,$F9,$3E,$1C,$3C,$78,$73,$8F
	.dc.b	$31,$FC,$08,$D1,$67,$E3,$D3,$5F,$0E,$3F,$16,$B5,$87,$F9,$B6,$F7
	.dc.b	$5A,$9F,$F3,$9F,$CF,$7F,$8D,$9E,$CD,$7F,$0F,$1F,$E2,$48,$C6,$1D
	.dc.b	$E7,$FC,$F7,$F3,$AE,$1F,$F3,$D5,$6A,$F8,$74,$60,$FF,$3F,$56,$EF
	.dc.b	$BF,$BF,$C0,$8F,$BE,$1F,$3F,$88,$F1,$9E,$89,$C6,$5E,$38,$57,$07
	.dc.b	$A9,$0E,$23,$15,$79,$2F,$FD,$25,$F3,$FF,$D1,$2B,$E8,$35,$E2,$DC
	.dc.b	$3C,$18,$7F,$A4,$67,$3C,$FC,$88,$FA,$FD,$9A,$E8,$EB,$1F,$12,$3F
	.dc.b	$D4,$5F,$06,$F9,$FA,$4A,$F8,$9D,$75,$F5,$CE,$DF,$C3,$35,$F7,$E1
	.dc.b	$FE,$3F,$73,$5C,$71,$F9,$15,$B7,$D7,$7F,$EB,$3F,$9F,$FF,$17,$DF
	.dc.b	$E4,$5A,$FC,$A6,$CF,$D1,$E3,$FF,$32,$CE,$FC,$5E,$44,$7F,$35,$1A
	.dc.b	$79,$46,$5A,$6B,$CE,$FF,$62,$79,$7F,$5A,$3F,$44,$D7,$ED,$B6,$FF
	.dc.b	$05,$F3,$DF,$E6,$D9,$DF,$17,$22,$3F,$AB,$8D,$3C,$A3,$2D,$35,$E2
	.dc.b	$7F,$B5,$67,$DF,$CD,$FE,$86,$6C,$FF,$B7,$1B,$E8,$BE,$7F,$F4,$53
	.dc.b	$9D,$3F,$DB,$DF,$3F,$FA,$18,$FC,$9B,$73,$F4,$CE,$BA,$A5,$7F,$01
	.dc.b	$AD,$65,$E1,$AF,$F7,$47,$CF,$BE,$7B,$FD,$D5,$F3,$EF,$81,$A6,$A2
	.dc.b	$E3,$FC,$2F,$E7,$FF,$CD,$AD,$F4,$32,$FE,$9A,$32,$FE,$A3,$E7,$DE
	.dc.b	$F7,$FB,$DD,$B1,$C4,$FE,$2D,$E3,$BE,$37,$FD,$F8,$D8,$CC,$63,$9B
	.dc.b	$2C,$8C,$9C,$A3,$0F,$EE,$27,$93,$67,$93,$FC,$97,$E7,$3E,$DF,$1F
	.dc.b	$F0,$35,$F2,$DF,$EB,$C8,$C3,$78,$BF,$D7,$F3,$C1,$BD,$19,$B0,$9E
	.dc.b	$6F,$7A,$D3,$99,$BF,$3A,$F9,$FA,$49,$F3,$67,$5F,$F1,$3C,$ED,$FF
	.dc.b	$01,$B5,$CC,$FF,$81,$9F,$17,$C0,$F8,$31,$FF,$6F,$C7,$C1,$8C,$7B
	.dc.b	$99,$64,$64,$F8,$D1,$87,$F9,$1C,$F2,$6C,$F2,$7F,$A8,$7C,$E7,$E4
	.dc.b	$1F,$86,$FF,$A7,$DC,$FB,$1F,$0A,$35,$06,$9B,$7F,$98,$B7,$6F,$E7
	.dc.b	$0F,$0C,$9F,$3D,$F2,$89,$5E,$2F,$22,$BF,$22,$3E,$24,$F0,$8F,$93
	.dc.b	$1F,$1E,$BF,$45,$1E,$AE,$8F,$31,$E7,$F8,$BB,$19,$21,$E6,$72,$FA
	.dc.b	$22,$E3,$E8,$9F,$C0,$FC,$EB,$C8,$BF,$2B,$30,$FE,$F1,$39,$6C,$E5
	.dc.b	$FE,$A7,$09,$E7,$3F,$8C,$F9,$CF,$E9,$DC,$DF,$92,$B4,$61,$FC,$AB
	.dc.b	$55,$FE,$9B,$1F,$EC,$93,$FD,$66,$7F,$D1,$4D,$7A,$35,$E7,$E6,$FE
	.dc.b	$3F,$E7,$3F,$49,$1F,$A2,$6A,$69,$FA,$47,$FB,$6B,$8E,$7F,$D2,$BA
	.dc.b	$E8,$B1,$27,$FC,$39,$39,$6F,$47,$0E,$E9,$18,$21,$D9,$CB,$E0,$8F
	.dc.b	$D9,$FC,$73,$A6,$92,$7B,$F4,$57,$FC,$07,$3B,$7F,$C5,$CD,$7D,$71
	.dc.b	$FB,$EF,$9B,$FC,$E9,$F9,$9E,$0F,$A8,$D9,$FE,$0E,$EB,$C6,$E1,$D5
	.dc.b	$27,$FC,$75,$39,$6C,$65,$FF,$0A,$C6,$5B,$19,$E1,$FF,$0F,$4E,$5B
	.dc.b	$19,$BF,$1D,$F9,$CC,$5F,$92,$70,$F8,$A3,$2A,$FD,$4D,$8D,$BE,$23
	.dc.b	$CE,$71,$21,$92,$5F,$9A,$38,$BE,$68,$E2,$7B,$A3,$37,$F2,$AF,$C7
	.dc.b	$7E,$D8,$C5,$70,$BE,$8F,$81,$E8,$F8,$23,$E0,$B9,$78,$B9,$3E,$AD
	.dc.b	$F8,$1F,$88,$75,$8E,$2F,$94,$C1,$A3,$04,$9F,$09,$23,$8B,$E6,$8F
	.dc.b	$01,$F3,$6F,$C0,$FC,$D3,$A9,$93,$D0,$79,$8C,$DF,$D4,$3E,$23,$E3
	.dc.b	$9F,$03,$B3,$C9,$5F,$02,$33,$1D,$CC,$9C,$63,$47,$A0,$8E,$48,$71
	.dc.b	$8F,$61,$D2,$7C,$97,$EC,$7C,$D7,$63,$ED,$B6,$BD,$6D,$7E,$2C,$F2
	.dc.b	$7E,$09,$F0,$CC,$CF,$59,$FC,$2B,$63,$E2,$B5,$D9,$CA,$7C,$6D,$8F
	.dc.b	$1B,$4F,$19,$4F,$1C,$F9,$6B,$DC,$DB,$F8,$F3,$F1,$2F,$E9,$BE,$7A
	.dc.b	$F8,$97,$E3,$6C,$78,$EF,$D2,$D3,$D2,$53,$D2,$7D,$63,$A5,$F8,$1F
	.dc.b	$A4,$FD,$67,$27,$8E,$BC,$11,$BB,$F2,$9F,$0A,$7A,$3F,$C6,$AB,$C7
	.dc.b	$18,$6D,$FF,$0E,$EA,$FE,$40,$49,$FE,$33,$F2,$35,$FE,$2F,$16,$7F
	.dc.b	$8A,$CF,$F3,$2F,$9A,$95,$FC,$B1,$FE,$55,$D3,$88,$FE,$63,$16,$BF
	.dc.b	$8E,$CF,$F8,$B3,$E6,$BF,$B4,$57,$FA,$E1,$FE,$CC,$FF,$B8,$DA,$E7
	.dc.b	$A3,$81,$C7,$A6,$46,$E3,$2F,$18,$CB,$C4,$E9,$E0,$78,$48,$F7,$F0
	.dc.b	$6B,$E3,$8B,$8F,$FB,$47,$0F,$F4,$96,$FF,$DD,$FE,$69,$F3,$5F,$E4
	.dc.b	$66,$E1,$FE,$FB,$6B,$FE,$D4,$F9,$9F,$ED,$B1,$88,$FF,$00,$07,$0F
	.dc.b	$59,$FF,$22,$46,$23,$FC,$AF,$2E,$7F,$F2,$AB,$63,$FC,$B7,$8F,$FB
	.dc.b	$9E,$31,$1E,$1E,$8D,$D4,$8D,$E3,$FE,$11,$AD,$DF,$1B,$D5,$7A,$6E
	.dc.b	$CE,$63,$89,$E4,$E6,$F1,$F4,$48,$C8,$C3,$FC,$A5,$5E,$5A,$D4,$FC
	.dc.b	$5A,$F5,$C7,$F9,$C3,$E5,$DF,$33,$49,$5D,$3E,$33,$FC,$A7,$3A,$8F
	.dc.b	$96,$7F,$9F,$1C,$38,$6B,$F4,$1A,$FF,$8C,$63,$FD,$FD,$83,$4C,$73
	.dc.b	$7F,$A0,$A4,$DE,$79,$7E,$38,$D1,$85,$7C,$DF,$F4,$7D,$76,$BE,$B1
	.dc.b	$C4,$F6,$5F,$7E,$6F,$07,$85,$27,$E0,$46,$5F,$85,$19,$7B,$D7,$AE
	.dc.b	$7F,$D3,$3F,$42,$C3,$FD,$1C,$DA,$F8,$E1,$C3,$F2,$9A,$FE,$5B,$E6
	.dc.b	$5F,$33,$FE,$EC,$8F,$F2,$C4,$63,$C1,$E6,$BC,$07,$C5,$E2,$F8,$5B
	.dc.b	$E4,$1F,$0F,$FE,$D0,$D7,$F1,$71,$8F,$83,$F9,$1C,$3F,$26,$37,$E8
	.dc.b	$62,$BF,$D7,$31,$B7,$8D,$F3,$C3,$E1,$9D,$D6,$1C,$63,$17,$6F,$94
	.dc.b	$70,$7B,$DF,$4A,$D5,$F5,$C3,$C8,$F4,$9F,$CF,$FC,$BB,$E2,$BB,$62
	.dc.b	$8C,$4F,$E9,$3E,$5D,$D3,$A9,$CD,$7A,$07,$CB,$FF,$0E,$BF,$D5,$B6
	.dc.b	$35,$5E,$39,$FF,$45,$4E,$72,$FC,$85,$3D,$C7,$BF,$E4,$6D,$FE,$DB
	.dc.b	$85,$F2,$F9,$AF,$4E,$2F,$97,$A7,$B3,$E5,$DC,$F8,$C7,$F7,$F1,$86
	.dc.b	$C7,$C2,$F9,$7F,$F3,$C4,$60,$FE,$0A,$E3,$4F,$80,$DD,$7D,$2B,$E0
	.dc.b	$F9,$78,$5F,$33,$99,$F0,$DF,$6F,$17,$CB,$7E,$7D,$FE,$23,$3A,$7F
	.dc.b	$8B,$9A,$EE,$E5,$5C,$6B,$86,$7F,$46,$F8,$3F,$8A,$BE,$07,$4A,$E1
	.dc.b	$7C,$A3,$AD,$F2,$CF,$E8,$C3,$C7,$F3,$7C,$BB,$E0,$FF,$0B,$39,$3C
	.dc.b	$B1,$97,$F9,$49,$9C,$BC,$51,$97,$F0,$D1,$9B,$FE,$1C,$A7,$51,$E2
	.dc.b	$BF,$0D,$F8,$1B,$5F,$CD,$BE,$5D,$E9,$FC,$33,$F3,$2E,$23,$DB,$18
	.dc.b	$34,$F1,$6E,$3F,$99,$9E,$0F,$C9,$CA,$E3,$F9,$49,$B6,$AF,$E6,$15
	.dc.b	$87,$1C,$3F,$AD,$79,$76,$7E,$B2,$31,$87,$C0,$AF,$D4,$BF,$A0,$8C
	.dc.b	$DF,$53,$F0,$B5,$C7,$BB,$F0,$0D,$8F,$B2,$DC,$3B,$C7,$F5,$AF,$94
	.dc.b	$FC,$1E,$1F,$CE,$FA,$9F,$2C,$F9,$7F,$DF,$D6,$AF,$D9,$5A,$9C,$F1
	.dc.b	$FF,$11,$8C,$4F,$58,$C5,$F3,$8C,$1F,$DB,$A7,$78,$FF,$05,$C3,$B3
	.dc.b	$FC,$BA,$B8,$39,$F6,$FF,$8F,$64,$F9,$7D,$CA,$D5,$A9,$CC,$63,$1F
	.dc.b	$EF,$B7,$C7,$2F,$6D,$ED,$F7,$E9,$F0,$3D,$AB,$7F,$1D,$B9,$79,$E7
	.dc.b	$F9,$1B,$F7,$FE,$5F,$F0,$1F,$2E,$F9,$6F,$C1,$4F,$96,$B6,$AB,$73
	.dc.b	$AD,$3F,$93,$DF,$1A,$FF,$15,$D6,$9F,$C6,$4E,$AB,$FC,$0B,$39,$9E
	.dc.b	$D8,$D7,$C8,$2F,$A3,$A3,$15,$C9,$A7,$6C,$3F,$BF,$4F,$08,$E6,$F2
	.dc.b	$C3,$C5,$1E,$63,$89,$D6,$38,$3E,$F3,$31,$C1,$F8,$B8,$7A,$63,$11
	.dc.b	$DA,$FE,$25,$F8,$9F,$35,$F8,$63,$13,$E1,$8C,$61,$F0,$63,$C1,$18
	.dc.b	$7C,$A7,$53,$DD,$1B,$C3,$FC,$13,$1F,$20,$FF,$05,$B8,$7D,$47,$F9
	.dc.b	$7B,$E5,$FF,$91,$79,$1F,$E1,$06,$C7,$F8,$E1,$AE,$FA,$3F,$C0,$AE
	.dc.b	$7D,$5F,$F1,$5B,$78,$39,$46,$23,$B3,$5F,$C5,$D7,$F9,$9E,$7F,$01
	.dc.b	$F2,$EF,$91,$FC,$7E,$9F,$7A,$31,$7F,$85,$BE,$0F,$D6,$6C,$EE,$B8
	.dc.b	$ED,$F9,$8B,$E3,$7B,$C7,$CF,$7F,$8A,$C3,$8D,$EF,$0E,$37,$FC,$87
	.dc.b	$0E,$37,$BC,$38,$DF,$A9,$B5,$BB,$E2,$AF,$8D,$6F,$1F,$55,$F0,$AD
	.dc.b	$5F,$0B,$DC,$7F,$31,$C7,$CA,$DA,$E1,$3A,$8E,$27,$4B,$8F,$E6,$A3
	.dc.b	$46,$EB,$9C,$70,$3F,$C3,$91,$F1,$A7,$FC,$2A,$DD,$7D,$13,$CF,$0F
	.dc.b	$E0,$75,$A6,$CF,$C9,$9F,$F4,$6D,$F6,$F9,$7F,$F1,$EC,$7F,$A4,$1A
	.dc.b	$FF,$8A,$DB,$7F,$E2,$4F,$95,$71,$74,$AF,$EA,$55,$AC,$3F,$89,$B6
	.dc.b	$33,$D3,$B5,$7C,$9F,$07,$BD,$9A,$FF,$63,$C7,$48,$E1,$2F,$97,$D4
	.dc.b	$49,$E2,$7F,$AA,$27,$4E,$0F,$54,$6D,$FF,$23,$46,$9F,$F2,$47,$CB
	.dc.b	$FF,$3C,$7F,$D6,$2D,$73,$D1,$E2,$F6,$3D,$53,$9D,$7F,$C8,$ED,$D3
	.dc.b	$F9,$9B,$6F,$FB,$B6,$1F,$E9,$97,$FB,$C4,$65,$34,$C4,$6E,$1F,$04
	.dc.b	$F9,$7F,$94,$D8,$FF,$63,$B5,$F6,$F5,$EA,$64,$69,$F9,$71,$88,$FF
	.dc.b	$34,$4F,$1C,$7F,$CC,$2D,$C7,$FA,$6E,$1F,$E4,$8A,$FF,$6A,$FC,$BF
	.dc.b	$F8,$CD,$BE,$2F,$DD,$8C,$47,$F9,$CF,$93,$FC,$D2,$E8,$87,$F9,$DE
	.dc.b	$37,$8F,$F9,$C2,$77,$39,$AE,$84,$7F,$DC,$B1,$A9,$FF,$3B,$F9,$5F
	.dc.b	$88,$9F,$F3,$07,$07,$E3,$E3,$1F,$2F,$FE,$70,$9F,$F7,$6C,$6C,$93
	.dc.b	$F9,$59,$D4,$FF,$A2,$E7,$2F,$09,$D7,$89,$BC,$64,$FF,$BE,$2F,$D7
	.dc.b	$A7,$F8,$D7,$D2,$7C,$BF,$E8,$E9,$5A,$7F,$C7,$5F,$2F,$FA,$37,$FD
	.dc.b	$34,$DD,$3F,$D2,$73,$B9,$4E,$70,$FF,$00,$B6,$39,$F4,$F9,$1F,$F5
	.dc.b	$27,$CB,$BE,$5B,$F6,$5F,$2E,$F8,$0A,$7E,$D6,$75,$5F,$E4,$DC,$FF
	.dc.b	$D4,$75,$B8,$C6,$5F,$E1,$0F,$97,$FF,$3B,$46,$0F,$F2,$A3,$87,$FD
	.dc.b	$69,$3C,$5F,$F5,$B4,$66,$BF,$D5,$FF,$2E,$F7,$BF,$83,$36,$BA,$0A
	.dc.b	$E1,$87,$F8,$83,$F2,$1B,$3F,$E6,$36,$A6,$F8,$9F,$AA,$8E,$71,$8F
	.dc.b	$67,$F8,$A4,$FF,$AD,$E7,$3E,$47,$F1,$7A,$D4,$78,$A7,$4F,$FB,$3E
	.dc.b	$3A,$FC,$67,$22,$46,$FA,$2B,$FE,$D5,$A4,$E3,$5F,$F6,$7C,$7E,$67
	.dc.b	$E5,$FF,$23,$1F,$ED,$25,$F2,$5F,$CA,$5B,$7F,$11,$DE,$9F,$E4,$26
	.dc.b	$C7,$FA,$15,$DB,$FE,$E5,$6D,$75,$EA,$FE,$69,$F2,$EF,$98,$FF,$0A
	.dc.b	$B6,$FF,$C9,$A7,$F9,$39,$C3,$80,$E1,$DB,$AA,$F5,$7D,$0F,$A4,$EF
	.dc.b	$52,$BD,$51,$97,$EF,$D7,$1C,$7D,$9F,$07,$F0,$49,$3F,$07,$DE,$AF
	.dc.b	$B6,$2E,$3F,$26,$7F,$A8,$7C,$C7,$E9,$68,$FF,$AE,$1F,$C6,$B5,$4A
	.dc.b	$FF,$CD,$B3,$C4,$9E,$E7,$43,$AF,$67,$CA,$F9,$9F,$92,$4C,$3E,$31
	.dc.b	$F1,$9C,$3D,$AF,$AF,$D0,$F9,$8E,$51,$F0,$A7,$0F,$9A,$32,$F1,$3F
	.dc.b	$D8,$58,$7E,$B7,$1F,$44,$63,$B7,$E0,$9F,$0B,$E3,$6C,$7C,$08,$AF
	.dc.b	$33,$E2,$C7,$F7,$7F,$9A,$FD,$44,$65,$7C,$B6,$4D,$79,$72,$7F,$6F
	.dc.b	$3A,$6C,$E5,$ED,$F3,$5F,$D8,$65,$F4,$3D,$CF,$F0,$34,$7F,$65,$34
	.dc.b	$E6,$7F,$81,$C6,$7D,$2C,$47,$96,$78,$C7,$F8,$3F,$E6,$FF,$47,$26
	.dc.b	$93,$AB,$5D,$32,$7F,$87,$CE,$5B,$D7,$CF,$BE,$46,$08,$7C,$99,$FE
	.dc.b	$22,$49,$FF,$10,$7C,$83,$A3,$C9,$F4,$32,$FF,$8E,$CE,$5B,$39,$7A
	.dc.b	$4F,$28,$91,$E6,$8D,$71,$FE,$02,$15,$CD,$FB,$8D,$8D,$60,$AE,$3C
	.dc.b	$FF,$EC,$AA,$D3,$70,$CE,$03,$E5,$57,$89,$EB,$1A,$72,$7A,$88,$7A
	.dc.b	$8F,$41,$8C,$3D,$17,$E0,$8E,$D8,$79,$5D,$9D,$8E,$18,$7F,$B8,$23
	.dc.b	$51,$D4,$D1,$2B,$DF,$87,$96,$FB,$4F,$C6,$C3,$D1,$7C,$AF,$C7,$1E
	.dc.b	$D7,$8C,$F8,$9C,$BE,$23,$CF,$7D,$67,$86,$1E,$58,$C4,$EC,$F8,$A7
	.dc.b	$9E,$B9,$BE,$79,$E0,$F7,$7E,$93,$C6,$FC,$D1,$A8,$E0,$7D,$5C,$3D
	.dc.b	$B3,$A8,$F2,$9E,$18,$CD,$F9,$23,$89,$E4,$BF,$7B,$89,$DB,$E3,$3B
	.dc.b	$61,$EE,$7A,$9E,$BC,$3C,$EF,$4C,$3C,$4F,$58,$CD,$FC,$03,$B9,$1C
	.dc.b	$92,$3A,$C6,$5F,$09,$F0,$F0,$F5,$47,$39,$F8,$8E,$E7,$EF,$1B,$3D
	.dc.b	$46,$0F,$34,$F2,$9F,$04,$79,$CF,$0B,$8C,$3E,$1C,$7B,$CF,$19,$C8
	.dc.b	$EA,$FE,$38,$DD,$78,$DC,$61,$F9,$0B,$F9,$55,$E3,$AF,$46,$1F,$22
	.dc.b	$F7,$3E,$88,$C5,$F0,$7E,$69,$88,$F7,$1E,$53,$A6,$1E,$C8,$DE,$1E
	.dc.b	$0B,$F2,$CF,$C8,$3D,$D7,$E5,$BF,$C4,$35,$FC,$23,$70,$FA,$11,$F0
	.dc.b	$9F,$4D,$F5,$9D,$57,$4A,$F5,$4E,$5A,$76,$29,$D8,$B5,$D9,$F5,$9F
	.dc.b	$50,$86,$07,$CE,$7F,$7F,$3F,$BE,$47,$F2,$56,$D7,$F1,$B9,$DF,$4F
	.dc.b	$FB,$56,$78,$75,$53,$11,$9F,$5A,$57,$F9,$52,$BF,$D6,$C3,$E5,$73
	.dc.b	$1F,$DA,$E3,$33,$D6,$7F,$A1,$FC,$E7,$7F,$A1,$97,$FD,$0A,$32,$F5
	.dc.b	$8C,$BF,$CE,$A3,$2F,$58,$CE,$7F,$CA,$A7,$FD,$DD,$F3,$BF,$DE,$23
	.dc.b	$FA,$23,$FD,$CE,$BF,$BF,$36,$33,$D5,$C0,$E3,$D5,$2B,$FC,$E3,$5C
	.dc.b	$F6,$F9,$0F,$F9,$6A,$33,$E2,$F8,$0F,$0F,$8D,$63,$24,$3C,$2F,$D1
	.dc.b	$C7,$FA,$C7,$CE,$3E,$77,$18,$FD,$5F,$9D,$7D,$0E,$31,$FE,$A5,$18
	.dc.b	$7F,$D0,$B7,$C5,$B7,$B5,$3B,$7F,$C3,$CD,$C3,$C3,$3B,$AE,$55,$AB
	.dc.b	$E5,$7B,$C3,$F1,$15,$AC,$3F,$B8,$4B,$A7,$64,$C3,$FD,$33,$1F,$E9
	.dc.b	$43,$4E,$23,$4F,$93,$AB,$BF,$8D,$E0,$E9,$93,$CA,$32,$F4,$8C,$BE
	.dc.b	$C8,$F8,$1F,$3A,$F9,$EF,$F2,$8C,$A3,$0F,$F9,$51,$AF,$F9,$ED,$AF
	.dc.b	$FA,$01,$CE,$5F,$E0,$DF,$93,$7D,$0F,$CD,$6D,$7F,$AD,$AB,$7B,$7E
	.dc.b	$B5,$B1,$FE,$09,$9D,$BF,$E4,$F6,$DF,$F8,$66,$77,$29,$74,$F7,$E9
	.dc.b	$93,$FE,$C4,$9C,$B6,$72,$F5,$9F,$34,$79,$FC,$7F,$47,$8F,$2B,$94
	.dc.b	$65,$E9,$19,$7C,$D1,$F8,$BF,$9E,$7C,$F3,$C4,$FF,$3E,$C6,$1B,$E1
	.dc.b	$BE,$4B,$35,$BC,$3F,$D5,$1F,$40,$FA,$1F,$F1,$C4,$7F,$13,$8C,$70
	.dc.b	$FF,$A2,$1B,$1C,$27,$6F,$F8,$CF,$D2,$78,$58,$8F,$F5,$D3,$63,$37
	.dc.b	$FE,$92,$6C,$FF,$AE,$7E,$7D,$F2,$5F,$E8,$8E,$5F,$F2,$54,$62,$FF
	.dc.b	$DE,$3F,$40,$E9,$D4,$7F,$29,$9D,$61,$FE,$89,$F9,$EF,$DD,$4F,$FB
	.dc.b	$E1,$B1,$FE,$8D,$AF,$F7,$F4,$E7,$C1,$E7,$3F,$E4,$DF,$83,$7C,$FE
	.dc.b	$21,$18,$AF,$F2,$96,$7F,$E5,$FE,$27,$D0,$6E,$B5,$7E,$3F,$73,$96
	.dc.b	$87,$A7,$D0,$7D,$9C,$BC,$D1,$8E,$79,$1D,$8F,$E8,$DF,$41,$FE,$79
	.dc.b	$71,$87,$F9,$8E,$F9,$3F,$EA,$99,$D3,$7C,$2F,$54,$3E,$26,$C7,$B7
	.dc.b	$E8,$1F,$1D,$F0,$B1,$7C,$F7,$E8,$5E,$F8,$AB,$71,$E2,$37,$E1,$F3
	.dc.b	$BE,$57,$C6,$FF,$D7,$97,$C1,$FF,$64,$5F,$03,$AC,$67,$8B,$FD,$93
	.dc.b	$19,$9E,$37,$D3,$0F,$A9,$2F,$1F,$D1,$F4,$0F,$8D,$CC,$61,$B3,$E7
	.dc.b	$9D,$3E,$C3,$91,$98,$DE,$1E,$69,$DB,$FE,$D5,$9D,$B9,$9F,$C3,$7D
	.dc.b	$03,$D7,$F8,$26,$0B,$5F,$2A,$34,$FC,$9D,$6D,$FF,$B8,$66,$E9,$F0
	.dc.b	$A6,$D7,$F8,$1E,$6D,$23,$EB,$DF,$77,$A3,$F9,$9E,$27,$CB,$7D,$4E
	.dc.b	$EC,$36,$BE,$EC,$E9,$F2,$4E,$9A,$6A,$78,$5F,$FB,$DE,$B8,$3F,$73
	.dc.b	$1F,$D9,$E5,$FE,$C3,$F9,$DF,$B9,$87,$FA,$E7,$A9,$F3,$EF,$A0,$FD
	.dc.b	$75,$6B,$1F,$E3,$13,$9B,$FC,$8C,$E6,$F8,$DF,$0E,$3F,$DC,$C6,$2B
	.dc.b	$71,$8B,$E5,$83,$1F,$24,$E7,$0F,$0F,$BF,$FC,$0A,$5F,$41,$E3,$B6
	.dc.b	$6A,$D4,$E6,$31,$97,$ED,$2F,$8E,$3D,$23,$15,$BD,$FD,$FE,$EF,$C0
	.dc.b	$6C,$FB,$9C,$3F,$B0,$CB,$FD,$C1,$F4,$1F,$EE,$1F,$9F,$7C,$FF,$F9
	.dc.b	$27,$4F,$9F,$7F,$E4,$BB,$D8,$AD,$BF,$EE,$99,$FE,$4F,$18,$8F,$F2
	.dc.b	$95,$F5,$BE,$98,$7E,$36,$79,$47,$55,$3B,$7A,$4E,$EB,$A4,$78,$9E
	.dc.b	$D7,$E0,$BE,$57,$C6,$BE,$6D,$F1,$9C,$C7,$80,$EB,$86,$DE,$91,$E6
	.dc.b	$8C,$47,$90,$E0,$74,$73,$5D,$EF,$C5,$3F,$1E,$31,$7E,$18,$C1,$CA
	.dc.b	$76,$FC,$63,$11,$B7,$18,$7B,$A3,$11,$E6,$3D,$27,$90,$FA,$A7,$CB
	.dc.b	$3D,$05,$AF,$F1,$3B,$83,$FD,$0D,$F3,$DF,$C7,$FE,$83,$83,$63,$EA
	.dc.b	$35,$FC,$13,$5E,$3E,$C7,$E8,$5D,$B5,$DB,$5D,$EF,$DA,$38,$63,$F8
	.dc.b	$7F,$A0,$7C,$D7,$FA,$66,$30,$DD,$B8,$E1,$F4,$B1,$FF,$2A,$63,$F7
	.dc.b	$F0,$E3,$36,$F8,$21,$5C,$1E,$58,$F8,$B0,$FF,$54,$5F,$29,$FB,$ED
	.dc.b	$C7,$F3,$97,$C5,$B7,$C6,$BB,$56,$EF,$9E,$1E,$29,$F0,$D6,$DB,$8F
	.dc.b	$9D,$87,$91,$FC,$D6,$3D,$EF,$9C,$EB,$1F,$0D,$6A,$FA,$63,$E1,$AF
	.dc.b	$1E,$3E,$1C,$79,$46,$0F,$EE,$33,$EF,$BF,$DA,$42,$31,$B7,$D0,$C7
	.dc.b	$E9,$CF,$0C,$3F,$D2,$55,$A8,$F9,$D8,$76,$5F,$41,$FD,$19,$FE,$FE
	.dc.b	$D7,$F9,$46,$1F,$ED,$78,$C6,$3E,$2F,$9D,$70,$FF,$A7,$E7,$35,$BE
	.dc.b	$AC,$1F,$EA,$3A,$DE,$1D,$BE,$03,$81,$1F,$F0,$7C,$3E,$83,$11,$E2
	.dc.b	$3F,$0D,$82,$B3,$97,$8E,$AE,$3E,$79,$B7,$F8,$DC,$3A,$3F,$E1,$CF
	.dc.b	$A0,$FE,$78,$DE,$8F,$F5,$C4,$62,$3C,$02,$7F,$CF,$EE,$23,$58,$FD
	.dc.b	$46,$E1,$FE,$FD,$6C,$F8,$7D,$BE,$B8,$48,$F0,$1F,$02,$3F,$D9,$1F
	.dc.b	$0C,$FA,$0A,$49,$F8,$BD,$75,$F2,$CE,$DF,$F1,$BB,$5F,$4E,$7F,$E3
	.dc.b	$E7,$35,$C7,$3F,$C5,$D6,$DF,$B9,$7F,$ED,$9F,$A0,$FF,$16,$5F,$F9
	.dc.b	$09,$AF,$C4,$6C,$FE,$53,$93,$FC,$D3,$3B,$F1,$39,$11,$FF,$25,$C6
	.dc.b	$9E,$51,$96,$9A,$F4,$7F,$DD,$5E,$5F,$DF,$8F,$F2,$9B,$5F,$AA,$DB
	.dc.b	$FD,$2F,$CF,$FF,$A7,$E7,$3C,$5C,$88,$FF,$97,$23,$53,$FA,$CA,$E5
	.dc.b	$C5,$FC,$4E,$BF,$CD,$17,$C1,$FC,$1D,$F0,$9F,$F4,$47,$D0,$7E,$82
	.dc.b	$7F,$8D,$4F,$C4,$9E,$4D,$8F,$F4,$5C,$ED,$47,$CC,$7F,$BD,$35,$E3
	.dc.b	$97,$FB,$AA,$31,$1F,$C9,$FE,$83,$F5,$6F,$E5,$9B,$97,$F0,$66,$E1
	.dc.b	$CA,$7F,$CF,$7F,$3E,$FA,$0F,$F7,$AB,$63,$FA,$A3,$5F,$F0,$03,$5F
	.dc.b	$F0,$77,$D0,$3E,$0F,$FD,$FA,$D8,$FE,$B8,$D7,$FC,$10,$D7,$FC,$2D
	.dc.b	$F3,$DF,$E7,$97,$FB,$13,$5F,$F0,$73,$5D,$FA,$1F,$D9,$D7,$CF,$6F
	.dc.b	$4F,$E4,$6D,$CF,$EC,$36,$7F,$A6,$78,$6E,$7F,$B9,$7C,$30,$E9,$7D
	.dc.b	$EF,$8C,$60,$8F,$FA,$8A,$FF,$D7,$4D,$F5,$F9,$E3,$C2,$35,$F1,$D8
	.dc.b	$7C,$2D,$54,$8C,$3F,$90,$8D,$9C,$88,$7F,$12,$74,$BE,$7B,$F2,$4F
	.dc.b	$F6,$95,$F2,$DD,$DB,$1F,$EB,$A7,$6F,$1F,$4B,$53,$23,$51,$E7,$F9
	.dc.b	$D7,$C1,$D2,$4F,$3C,$2B,$FE,$45,$9D,$BF,$E1,$D6,$B9,$9F,$B5,$18
	.dc.b	$DB,$FA,$DE,$2F,$9E,$F4,$1D,$23,$3F,$05,$EB,$26,$1E,$BF,$72,$BF
	.dc.b	$CD,$AD,$38,$AE,$B5,$F4,$7E,$73,$F4,$77,$FE,$0F,$7F,$D7,$0F,$FB
	.dc.b	$79,$AF,$F4,$07,$69,$A6,$A7,$FB,$9B,$77,$FE,$10,$F0,$C9,$F3,$DF
	.dc.b	$2C,$97,$DF,$C6,$AF,$C9,$8F,$8A,$47,$E5,$3E,$9F,$9E,$91,$FE,$F4
	.dc.b	$AE,$AF,$A7,$A7,$DC,$47,$D9,$F1,$D5,$F6,$45,$CF,$D3,$3F,$95,$F9
	.dc.b	$EF,$D3,$57,$FC,$00,$5F,$31,$93,$5D,$52,$75,$19,$78,$47,$E7,$FE
	.dc.b	$7B,$FB,$0C,$1A,$62,$BF,$45,$1F,$E4,$1C,$3F,$C3,$B3,$FD,$C6,$3F
	.dc.b	$D3,$8D,$7F,$9A,$35,$FF,$24,$35,$ED,$E6,$7A,$FE,$7A,$47,$5E,$1E
	.dc.b	$33,$FB,$3F,$9E,$FD,$19,$3A,$71,$D5,$96,$5F,$13,$FC,$AD,$3C,$27
	.dc.b	$97,$57,$3E,$D9,$5B,$15,$FC,$01,$C3,$5F,$F0,$6B,$FC,$5D,$DC,$74
	.dc.b	$79,$7C,$93,$A3,$F8,$53,$63,$31,$EB,$E9,$CB,$23,$27,$82,$30,$FF
	.dc.b	$87,$27,$93,$67,$93,$C0,$C1,$77,$FF,$28,$E1,$FA,$E8,$C6,$BF,$D9
	.dc.b	$EF,$9D,$7F,$95,$9B,$1F,$DA,$EB,$85,$74,$25,$6E,$FC,$31,$8A,$E9
	.dc.b	$1E,$89,$F0,$E1,$67,$35,$E2,$3E,$AC,$65,$FF,$46,$61,$E2,$6D,$F6
	.dc.b	$BF,$04,$71,$13,$B7,$EC,$D7,$37,$35,$70,$F5,$DF,$29,$F4,$1D,$8F
	.dc.b	$A0,$75,$BE,$87,$59,$E3,$87,$AA,$78,$9D,$8E,$26,$27,$8D,$FA,$67
	.dc.b	$85,$F8,$48,$FC,$FC,$3D,$15,$C2,$F9,$D6,$9F,$31,$E0,$C3,$C6,$E2
	.dc.b	$34,$F8,$CF,$CB,$D7,$4C,$3E,$13,$63,$E5,$57,$27,$85,$72,$BF,$81
	.dc.b	$5B,$9F,$03,$67,$BD,$71,$AF,$CA,$D7,$F8,$CC,$F1,$15,$CB,$F2,$4C
	.dc.b	$5F,$F8,$1E,$33,$1E,$E3,$27,$FB,$C5,$FA,$5F,$3D,$FC,$0C,$F1,$63
	.dc.b	$F9,$77,$13,$F6,$6A,$BF,$B8,$8D,$BF,$C1,$1C,$E1,$EF,$9B,$5C,$EB
	.dc.b	$91,$FC,$0A,$B2,$FC,$F8,$FB,$66,$DE,$C7,$02,$92,$BB,$3C,$C8,$7E
	.dc.b	$1E,$BA,$BF,$C4,$AB,$CE,$D3,$73,$C9,$E9,$5E,$18,$C1,$FC,$02,$79
	.dc.b	$DF,$E2,$EB,$67,$C1,$3C,$D3,$B7,$AD,$7B,$1F,$71,$B3,$89,$F8,$67
	.dc.b	$4F,$94,$CC,$F7,$9E,$91,$8B,$FD,$95,$F9,$1B,$1E,$03,$B9,$D4,$F0
	.dc.b	$4F,$23,$DB,$7E,$F2,$47,$0B,$FC,$2B,$CC,$E4,$4C,$18,$78,$63,$FA
	.dc.b	$07,$CF,$7F,$82,$0F,$F0,$22,$0E,$1B,$EB,$7F,$91,$63,$87,$C2,$53
	.dc.b	$11,$9E,$C9,$5D,$9E,$1F,$2D,$88,$FF,$6B,$DF,$FB,$5B,$46,$9F,$E9
	.dc.b	$B8,$C3,$6B,$FA,$5F,$C9,$4B,$FE,$9E,$F8,$FA,$F1,$1F,$B1,$3F,$92
	.dc.b	$C7,$EB,$48,$FE,$AD,$F5,$7C,$F4,$8E,$67,$F6,$3F,$9E,$FF,$06,$BF
	.dc.b	$EE,$C7,$FC,$1A,$E7,$B3,$FC,$CD,$1A,$CF,$D3,$3F,$26,$31,$97,$FA
	.dc.b	$D5,$E3,$1C,$EB,$8F,$67,$2F,$16,$4F,$85,$EB,$E8,$62,$3C,$51,$71
	.dc.b	$EF,$0C,$3F,$CD,$FF,$3A,$F9,$3F,$F3,$D3,$67,$FB,$F3,$55,$72,$F9
	.dc.b	$DF,$F3,$7C,$E6,$7F,$8D,$35,$FF,$67,$35,$FC,$33,$53,$D6,$7A,$7C
	.dc.b	$F7,$FA,$2E,$8F,$92,$F1,$D9,$30,$FF,$09,$57,$F8,$3E,$FC,$9D,$99
	.dc.b	$9F,$3E,$FF,$ED,$FA,$D3,$6F,$B5,$77,$EC,$F0,$76,$4C,$3A,$C6,$5E
	.dc.b	$F1,$97,$A9,$FA,$1C,$3F,$A2,$7C,$E3,$E7,$BF,$D4,$8D,$9F,$F1,$43
	.dc.b	$5E,$23,$E4,$7F,$CE,$71,$98,$FE,$88,$D7,$FD,$F0,$D7,$F6,$0D,$4F
	.dc.b	$5D,$3E,$FC,$BE,$7B,$FD,$6C,$DA,$F1,$DE,$3A,$F3,$D5,$27,$FC,$7D
	.dc.b	$5F,$AC,$7F,$7C,$D7,$D1,$1B,$7D,$71,$93,$FD,$CD,$87,$8B,$C7,$F0
	.dc.b	$9E,$2E,$B9,$5D,$A3,$2F,$82,$3F,$D9,$9F,$43,$F1,$63,$F6,$C1,$C3
	.dc.b	$FB,$96,$BF,$DA,$DA,$FF,$15,$1F,$3E,$F9,$F7,$8D,$D2,$30,$DF,$15
	.dc.b	$F2,$59,$AD,$DF,$F0,$7F,$9F,$7C,$47,$EB,$23,$B4,$63,$A7,$CE,$D8
	.dc.b	$E1,$3B,$7F,$94,$7C,$EB,$E4,$31,$08,$C5,$FC,$06,$C7,$F1,$76,$BF
	.dc.b	$E0,$9F,$9B,$FD,$D4,$FC,$66,$C7,$F1,$D6,$BB,$ED,$EF,$F3,$8F,$84
	.dc.b	$A4,$BF,$9F,$19,$7A,$7B,$5E,$6C,$0D,$24,$7D,$6F,$9C,$F9,$AD,$E1
	.dc.b	$F6,$C6,$32,$FE,$60,$D7,$FC,$45,$F3,$9F,$E8,$3A,$FB,$ED,$8F,$E6
	.dc.b	$CD,$77,$DF,$F2,$3E,$71,$F1,$5F,$CB,$3A,$BF,$96,$E0,$F9,$AF,$D7
	.dc.b	$3F,$E9,$9A,$DB,$FE,$A1,$35,$D9,$C5,$FC,$44,$6B,$1F,$DF,$C6,$8E
	.dc.b	$58,$6E,$3F,$A9,$DF,$67,$F7,$B5,$C7,$0F,$1C,$E5,$FE,$67,$19,$BF
	.dc.b	$ED,$47,$5A,$FE,$0A,$BC,$4E,$CB,$E7,$1F,$11,$E1,$AF,$EC,$75,$A9
	.dc.b	$CD,$7F,$AD,$CE,$47,$BD,$CC,$F6,$9E,$15,$CE,$B7,$3F,$02,$B7,$87
	.dc.b	$F4,$A9,$CB,$F3,$A3,$31,$FA,$FF,$9C,$7A,$9E,$8C,$EC,$FF,$B2,$66
	.dc.b	$E1,$F3,$26,$E1,$FD,$5A,$6D,$7C,$49,$B5,$FE,$0F,$9B,$3F,$76,$FE
	.dc.b	$14,$62,$B7,$48,$C4,$FC,$7E,$1F,$DA,$61,$F5,$3A,$31,$D3,$E7,$6E
	.dc.b	$5F,$85,$CB,$F6,$35,$FD,$AA,$F8,$CF,$96,$B7,$97,$AD,$B5,$D6,$7F
	.dc.b	$DD,$1F,$33,$F7,$EB,$F8,$6F,$CE,$3E,$73,$ED,$4E,$7B,$D3,$9B,$E3
	.dc.b	$C7,$F8,$F9,$CD,$F0,$8C,$56,$F0,$FE,$9B,$87,$38,$C6,$3F,$E0,$28
	.dc.b	$C5,$F8,$3D,$BF,$CD,$4B,$E7,$3C,$D8,$2D,$5A,$9C,$C6,$33,$FF,$21
	.dc.b	$5F,$EC,$67,$35,$BC,$BD,$BA,$7B,$BD,$9B,$87,$C4,$6E,$5E,$89,$FC
	.dc.b	$E3,$FE,$4D,$F9,$CF,$F2,$77,$CD,$3E,$6F,$F9,$E6,$BF,$53,$1F,$F1
	.dc.b	$D5,$6A,$73,$87,$F8,$BA,$F8,$61,$FE,$AF,$AD,$3F,$E5,$9C,$3F,$96
	.dc.b	$CE,$A3,$FA,$44,$E9,$F8,$55,$CE,$FB,$C2,$B8,$5F,$29,$CC,$62,$B8
	.dc.b	$57,$84,$E1,$3E,$6C,$3B,$47,$BC,$C9,$E5,$9E,$2F,$90,$C9,$E4,$8E
	.dc.b	$2B,$0F,$46,$1E,$38,$F0,$61,$E3,$AD,$D6,$AF,$AD,$78,$6B,$4F,$63
	.dc.b	$18,$79,$DC,$1D,$CF,$1E,$1D,$DF,$31,$FD,$E2,$FF,$BC,$BE,$D3,$07
	.dc.b	$F9,$C7,$E6,$BF,$8F,$7C,$E7,$F7,$E6,$C6,$7A,$93,$96,$C7,$1F,$53
	.dc.b	$FC,$AA,$DE,$1E,$51,$88,$FF,$77,$5A,$31,$1F,$E0,$58,$C4,$FD,$FF
	.dc.b	$9C,$7C,$B7,$F8,$27,$5F,$C2,$46,$2F,$F0,$97,$C1,$FA,$6D,$9D,$D7
	.dc.b	$1D,$7F,$C2,$57,$C6,$F7,$8F,$A2,$FF,$21,$87,$1B,$DE,$1C,$6F,$FD
	.dc.b	$5D,$87,$1B,$DE,$1C,$6F,$D4,$DA,$DD,$F1,$57,$C6,$B7,$8F,$AA,$F8
	.dc.b	$56,$AF,$85,$EE,$3F,$D7,$D8,$F9,$5B,$5C,$27,$51,$C4,$E9,$71,$FC
	.dc.b	$BC,$68,$DD,$73,$8E,$07,$F9,$92,$3E,$34,$FF,$A6,$1B,$BF,$A2,$79
	.dc.b	$E1,$FE,$3D,$AD,$36,$7E,$4C,$FF,$14,$78,$65,$FB,$DF,$9C,$FF,$1C
	.dc.b	$CE,$4F,$F6,$D3,$87,$FC,$EA,$DB,$FF,$54,$FC,$D3,$8B,$B5,$70,$EB
	.dc.b	$AA,$F8,$63,$FE,$51,$6C,$7F,$B9,$EF,$93,$E1,$38,$7B,$BC,$A4,$D3
	.dc.b	$AC,$BE,$73,$15,$C6,$B6,$DA,$DC,$EA,$3D,$71,$72,$F5,$CF,$A3,$E7
	.dc.b	$3F,$9E,$D7,$8D,$AF,$FA,$59,$B3,$F0,$BD,$6E,$1D,$32,$B8,$3E,$68
	.dc.b	$C4,$71,$8B,$3C,$63,$E0,$7C,$03,$E7,$3E,$4B,$63,$31,$F0,$3A,$F2
	.dc.b	$C8,$C9,$F2,$75,$FF,$20,$B8,$3F,$D5,$B5,$CD,$B3,$CD,$FA,$1F,$39
	.dc.b	$FE,$2A,$6F,$8D,$F6,$A3,$11,$FE,$B6,$E4,$FF,$33,$3A,$21,$FE,$BA
	.dc.b	$8D,$E5,$FE,$EE,$9D,$CE,$6B,$A1,$1F,$C0,$46,$A7,$FD,$EB,$F3,$9F
	.dc.b	$6A,$7F,$D8,$AD,$8C,$F0,$7E,$3B,$E6,$FF,$D2,$75,$A8,$FC,$8D,$6F
	.dc.b	$8F,$FD,$40,$D9,$0A,$E3,$7D,$23,$F2,$53,$B4,$D9,$DD,$72,$F4,$1F
	.dc.b	$39,$F9,$CB,$E9,$9F,$94,$FD,$2C,$EB,$9B,$13,$F9,$FC,$3E,$3E,$1F
	.dc.b	$EB,$56,$C2,$B8,$3B,$BF,$D6,$FC,$E7,$EA,$23,$FD,$D2,$DC,$D8,$73
	.dc.b	$AF,$D9,$FC,$DB,$E7,$29,$30,$FE,$01,$8F,$AB,$E7,$FF,$73,$F3,$BF
	.dc.b	$ED,$8C,$7F,$D9,$0D,$AF,$92,$D5,$87,$57,$F7,$5F,$3B,$FE,$77,$6F
	.dc.b	$87,$FC,$12,$31,$1F,$F0,$02,$B9,$7C,$EE,$35,$F4,$CF,$D6,$8C,$63
	.dc.b	$FE,$D9,$6C,$7D,$AF,$0D,$D1,$FC,$42,$73,$26,$98,$9D,$C7,$F0,$F8
	.dc.b	$CE,$5C,$4F,$E3,$33,$A9,$61,$E4,$7F,$B4,$7C,$66,$88,$FF,$19,$9F
	.dc.b	$C4,$08,$CC,$72,$C7,$FD,$E5,$39,$8E,$1C,$DF,$CA,$3E,$6F,$F1,$B1
	.dc.b	$F9,$65,$F2,$99,$35,$E8,$49,$FE,$51,$3A,$6C,$E5,$F0,$1F,$A0,$F9
	.dc.b	$87,$CE,$FF,$36,$6C,$7F,$31,$6B,$C0,$57,$F8,$1D,$D4,$F1,$23,$E2
	.dc.b	$C6,$BC,$6B,$9E,$9F,$25,$B9,$FE,$CA,$7A,$D7,$C1,$F8,$D9,$1E,$C7
	.dc.b	$E2,$7B,$58,$8F,$26,$49,$C6,$3F,$56,$15,$FD,$4B,$E7,$3F,$49,$8F
	.dc.b	$77,$FD,$67,$87,$F4,$F6,$BE,$F7,$6F,$F0,$66,$A7,$77,$F8,$36,$CF
	.dc.b	$5F,$27,$E6,$7C,$E4,$BF,$98,$F8,$FC,$7C,$47,$C5,$8B,$87,$39,$B7
	.dc.b	$F1,$75,$56,$2F,$C9,$5E,$DF,$89,$91,$EB,$7E,$07,$C5,$62,$3D,$91
	.dc.b	$71,$F4,$4F,$F7,$6F,$9B,$FC,$F4,$7E,$75,$7C,$A6,$4D,$74,$CC,$3F
	.dc.b	$BB,$4E,$9B,$39,$7F,$11,$F3,$7F,$D9,$B5,$FB,$CD,$8F,$E6,$2D,$55
	.dc.b	$FA,$37,$8F,$99,$FE,$0A,$9D,$1B,$9D,$63,$FB,$38,$CC,$EC,$F6,$74
	.dc.b	$FF,$84,$BE,$6B,$F5,$78,$1A,$4E,$8D,$74,$C9,$FF,$09,$4E,$5B,$D1
	.dc.b	$CF,$AE,$46,$08,$7E,$9E,$7F,$88,$92,$7F,$C4,$7F,$19,$F9,$49,$FF
	.dc.b	$0F,$35,$FD,$73,$74,$FE,$1D,$5F,$9D,$9D,$F4,$72,$23,$E6,$EA,$AF
	.dc.b	$28,$B5,$CA,$6F,$07,$EB,$F0,$EC,$DA,$F8,$D3,$AB,$FE,$CB,$5A,$6C
	.dc.b	$E9,$F1,$90,$F0,$1E,$33,$17,$FC,$9E,$79,$41,$B8,$7B,$63,$89,$27
	.dc.b	$85,$77,$8C,$61,$E7,$7A,$BF,$C2,$6B,$95,$70,$9C,$9E,$83,$C8,$72
	.dc.b	$30,$53,$F0,$98,$7C,$17,$CD,$76,$BA,$DF,$86,$78,$61,$EA,$BE,$0F
	.dc.b	$B8,$C4,$78,$8E,$65,$AE,$6F,$8B,$0F,$43,$E3,$8C,$BC,$0E,$F8,$7C
	.dc.b	$39,$D9,$30,$F2,$3F,$7E,$7C,$17,$E3,$AE,$15,$A2,$4F,$85,$F9,$58
	.dc.b	$7A,$A7,$E0,$36,$7D,$0D,$7C,$CD,$BF,$14,$7F,$8B,$30,$F4,$E1,$E1
	.dc.b	$8E,$27,$92,$FA,$B4,$C9,$EE,$2D,$74,$7E,$13,$6B,$AC,$F8,$70,$F4
	.dc.b	$D7,$33,$CA,$5A,$E1,$86,$E3,$E0,$1E,$77,$0F,$69,$E6,$F8,$CE,$65
	.dc.b	$9E,$6F,$58,$F9,$6F,$E2,$8C,$1D,$8C,$1C,$5D,$DF,$AA,$36,$6A,$7E
	.dc.b	$06,$1F,$26,$3F,$86,$5F,$9A,$3F,$1B,$1C,$CE,$D5,$CD,$C1,$CE,$FC
	.dc.b	$B3,$CE,$FF,$0D,$3A,$8E,$71,$E0,$C3,$EA,$CE,$5F,$C4,$BE,$B9,$E5
	.dc.b	$5F,$1A,$B8,$4F,$D8,$8F,$EF,$9F,$33,$FD,$A0,$FE,$CB,$1F,$E7,$06
	.dc.b	$A7,$5D,$3F,$EC,$87,$24,$71,$CF,$27,$FD,$8D,$28,$E1,$F2,$19,$9F
	.dc.b	$F0,$C4,$E5,$DC,$E9,$FE,$63,$F3,$3F,$CD,$08,$F4,$E9,$AF,$F3,$48
	.dc.b	$B9,$7F,$90,$6F,$FC,$C1,$9A,$73,$9F,$F2,$D7,$D5,$87,$FB,$8B,$E6
	.dc.b	$3F,$B6,$CF,$F8,$7D,$FE,$D1,$8F,$F8,$B5,$B8,$FF,$93,$5D,$C7,$FA
	.dc.b	$59,$A9,$DC,$F3,$E8,$ED,$D3,$DF,$9A,$5F,$FB,$AE,$FF,$B3,$DF,$87
	.dc.b	$C0,$CC,$F8,$6A,$E3,$E1,$9F,$F7,$CF,$CC,$3E,$3A,$92,$FF,$DE,$DF
	.dc.b	$2F,$FD,$B2,$75,$1F,$EF,$96,$BF,$E4,$96,$DF,$F9,$A6,$78,$57,$3A
	.dc.b	$D3,$FE,$43,$9D,$74,$DC,$BF,$C8,$D7,$FE,$51,$BF,$F2,$A3,$6B,$A7
	.dc.b	$46,$E7,$FD,$7B,$3F,$E8,$23,$55,$CA,$75,$29,$CB,$C3,$A7,$31,$FC
	.dc.b	$C8,$FE,$B2,$4E,$9C,$9A,$E8,$95,$F1,$23,$AF,$CB,$BE,$5A,$92,$7F
	.dc.b	$DA,$39,$FF,$A2,$FE,$33,$FC,$9B,$5B,$9C,$E9,$FE,$65,$8C,$47,$F9
	.dc.b	$CD,$B5,$C7,$AF,$A4,$E7,$0B,$39,$8E,$67,$FB,$83,$0F,$D2,$3F,$E7
	.dc.b	$D5,$D7,$F1,$1B,$1C,$03,$C2,$7D,$1D,$7B,$A9,$5B,$AD,$65,$FE,$7D
	.dc.b	$7F,$D8,$C3,$C1,$EC,$77,$E9,$95,$C6,$32,$F3,$8F,$AF,$F5,$AF,$90
	.dc.b	$FF,$44,$4C,$8C,$38,$3F,$DF,$AE,$1C,$35,$E3,$F2,$EF,$0B,$84,$04
	.dc.b	$65,$CB,$5E,$CD,$7C,$7F,$2E,$F9,$77,$47,$48,$C3,$7C,$17,$C6,$66
	.dc.b	$B7,$7E,$9F,$96,$7C,$57,$FB,$8B,$93,$FD,$C7,$EA,$3D,$A7,$D1,$E7
	.dc.b	$97,$FD,$55,$87,$07,$FD,$F7,$3A,$7E,$D7,$CB,$7E,$ED,$FB,$9F,$43
	.dc.b	$FE,$A9,$73,$E1,$73,$D3,$E1,$4E,$5F,$F6,$16,$4F,$98,$D4,$7F,$AC
	.dc.b	$7E,$8E,$CE,$70,$FC,$47,$80,$9D,$E5,$F6,$67,$CB,$F4,$98,$E9,$F0
	.dc.b	$61,$ED,$8E,$BF,$35,$F2,$AB,$D3,$58,$E6,$E1,$1F,$95,$F9,$AF,$F5
	.dc.b	$FB,$F9,$7A,$E1,$A7,$FB,$66,$72,$FF,$B6,$E3,$3C,$7F,$EF,$BF,$9A
	.dc.b	$7C,$4F,$C1,$C1,$F2,$FF,$A0,$7F,$45,$3A,$7F,$56,$6B,$C1,$F8,$71
	.dc.b	$9C,$7F,$DD,$F3,$93,$8D,$F5,$BE,$93,$FE,$ED,$AE,$18,$2F,$F2,$47
	.dc.b	$19,$D5,$F1,$C3,$E8,$CB,$C9,$F4,$7C,$D3,$E0,$F3,$18,$6C,$FA,$27
	.dc.b	$4F,$EF,$4E,$46,$63,$78,$7A,$A7,$6B,$1F,$33,$E6,$BE,$D7,$FC,$1F
	.dc.b	$E6,$9F,$35,$EF,$3F,$0E,$E2,$3E,$4E,$B6,$FA,$47,$EC,$54,$E7,$0E
	.dc.b	$F3,$F5,$2F,$99,$D5,$F0,$36,$B9,$9F,$BB,$AE,$53,$F3,$2E,$CF,$EF
	.dc.b	$F8,$1F,$11,$F5,$7B,$BF,$3B,$18,$BF,$3C,$62,$75,$1F,$C9,$67,$55
	.dc.b	$F8,$0B,$FD,$93,$E2,$3F,$81,$D7,$07,$13,$C3,$2F,$6D,$70,$7E,$6C
	.dc.b	$BE,$5F,$F0,$B8,$7F,$15,$F0,$5F,$34,$F9,$AF,$D9,$D6,$AF,$F8,$6C
	.dc.b	$E6,$FF,$9D,$4E,$6F,$8D,$F0,$E3,$FE,$01,$18,$AD,$C6,$2F,$96,$1E
	.dc.b	$BC,$39,$46,$31,$F3,$C7,$F2,$BB,$FE,$9F,$0F,$9A,$F2,$5F,$AF,$C8
	.dc.b	$FE,$1D,$39,$C7,$D3,$18,$AD,$E1,$CE,$FE,$DD,$EF,$0F,$5E,$FF,$03
	.dc.b	$CF,$F8,$2D,$8C,$92,$BE,$91,$FC,$F7,$E6,$BF,$9F,$7C,$B3,$E6,$7F
	.dc.b	$21,$97,$D1,$BF,$E9,$55,$A9,$CE,$1F,$A0,$BE,$18,$7F,$52,$AD,$3F
	.dc.b	$D2,$70,$FC,$0C,$EA,$3F,$39,$3A,$7E,$E5,$73,$BE,$F0,$AE,$17,$CA
	.dc.b	$73,$18,$AE,$15,$E1,$38,$4F,$9B,$0E,$D1,$EF,$32,$79,$67,$8B,$E4
	.dc.b	$32,$79,$23,$8A,$C3,$D1,$87,$8E,$3C,$18,$78,$EB,$75,$AB,$EB,$5E
	.dc.b	$1A,$D3,$D8,$C6,$1E,$77,$07,$73,$C7,$87,$77,$CC,$7F,$76,$BF,$EE
	.dc.b	$EF,$B4,$C1,$FE,$01,$F9,$9F,$E4,$9F,$35,$FD,$F1,$B1,$9E,$84,$E5
	.dc.b	$B1,$C7,$D8,$FE,$3C,$DD,$B9,$46,$23,$FC,$23,$68,$C4,$7F,$19,$8C
	.dc.b	$4F,$DF,$F9,$A7,$CA,$7F,$1D,$D7,$F0,$91,$8B,$FC,$25,$F0,$7E,$9B
	.dc.b	$67,$75,$C7,$5F,$E4,$D7,$C6,$F7,$8F,$A2,$FF,$21,$87,$1B,$DE,$1C
	.dc.b	$6F,$FC,$3D,$87,$1B,$DE,$1C,$6F,$D4,$DA,$DD,$F1,$57,$C6,$B7,$8F
	.dc.b	$AA,$F8,$56,$AF,$85,$EE,$3F,$C5,$D8,$F9,$5B,$5C,$27,$51,$C4,$E9
	.dc.b	$71,$FC,$BC,$68,$DD,$73,$8E,$07,$F3,$18,$F8,$D3,$FC,$B1,$BB,$FA
	.dc.b	$27,$9E,$1F,$E8,$49,$CC,$7C,$03,$FC,$C5,$7D,$7E,$6B,$FA,$3C,$7F
	.dc.b	$99,$5A,$FF,$3F,$6E,$1F,$CE,$BE,$61,$C5,$FE,$90,$9C,$E5,$FD,$7D
	.dc.b	$B7,$FE,$51,$6A,$AE,$2F,$7F,$7B,$91,$1F,$F3,$B4,$3E,$6B,$11,$E3
	.dc.b	$9B,$8F,$8E,$6E,$1E,$19,$B5,$FE,$61,$BE,$4F,$FA,$07,$E6,$BF,$CC
	.dc.b	$18,$7F,$54,$6D,$FF,$77,$6C,$F1,$F6,$FF,$D5,$B1,$88,$34,$C5,$FF
	.dc.b	$A1,$A3,$FC,$E1,$19,$E8,$E2,$7F,$AD,$A7,$52,$F8,$37,$CD,$7F,$AE
	.dc.b	$9B,$1F,$E9,$76,$BF,$DC,$67,$FC,$7A,$E6,$78,$11,$F9,$3D,$75,$FE
	.dc.b	$F1,$5C,$B3,$FE,$E6,$DC,$FF,$C2,$D8,$74,$AF,$EF,$DF,$B9,$3E,$C7
	.dc.b	$F8,$AB,$63,$FD,$1A,$D7,$FD,$2C,$D7,$FD,$36,$E7,$0F,$F3,$FF,$E8
	.dc.b	$AF,$8C,$FF,$6E,$7A,$08,$C5,$FF,$AD,$6F,$86,$5F,$E2,$96,$C2,$B8
	.dc.b	$5F,$2B,$E3,$3E,$8C,$7F,$DC,$77,$FE,$C4,$8F,$1F,$07,$A8,$8F,$A9
	.dc.b	$E5,$E2,$C8,$F1,$47,$23,$0F,$F8,$CB,$1F,$44,$E6,$3D,$1E,$4F,$C6
	.dc.b	$23,$F1,$9A,$FC,$CE,$BF,$29,$E6,$8B,$1F,$32,$3E,$79,$1F,$F5,$DF
	.dc.b	$CD,$7E,$A1,$FF,$62,$35,$FF,$64,$B5,$FF,$28,$35,$FF,$61,$7C,$BB
	.dc.b	$EC,$3F,$A7,$B6,$3F,$D8,$ED,$7F,$DA,$0D,$7F,$DA,$BF,$30,$FA,$7F
	.dc.b	$EA,$ED,$8F,$F6,$7B,$5F,$F6,$C3,$5F,$F6,$EC,$FF,$9F,$5F,$F7,$93
	.dc.b	$5F,$F2,$63,$78,$BF,$C8,$F2,$F8,$4F,$F3,$54,$7F,$92,$A7,$85,$74
	.dc.b	$F0,$FD,$64,$7D,$6F,$0E,$4F,$F3,$97,$D4,$7F,$AE,$67,$FD,$C9,$7F
	.dc.b	$E5,$48,$C4,$78,$AB,$83,$F2,$1A,$A3,$73,$C7,$D9,$CE,$1E,$9C,$3F
	.dc.b	$D0,$0D,$BF,$F7,$95,$F0,$9F,$1D,$70,$C2,$57,$92,$72,$D3,$8C,$E9
	.dc.b	$F3,$19,$3C,$8F,$9A,$BB,$47,$A4,$F9,$F5,$99,$91,$83,$94,$F0,$79
	.dc.b	$98,$8F,$06,$1E,$F8,$C1,$0F,$19,$FE,$85,$9E,$44,$9E,$37,$EC,$9D
	.dc.b	$47,$89,$E3,$5D,$5C,$9D,$5F,$C0,$5E,$DD,$12,$BE,$7C,$65,$ED,$19
	.dc.b	$9F,$84,$74,$3F,$D6,$87,$CC,$29,$4B,$7F,$84,$76,$FB,$3E,$C3,$E7
	.dc.b	$1D,$E3,$84,$FE,$2E,$33,$1E,$9A,$F1,$9F,$92,$8C,$C7,$8E,$78,$3F
	.dc.b	$4A,$33,$1E,$98,$FC,$84,$62,$79,$47,$53,$F1,$B1,$96,$99,$38,$9E
	.dc.b	$E2,$9E,$E7,$FC,$A3,$46,$9F,$8B,$F9,$09,$3E,$52,$1F,$7D,$CB,$F8
	.dc.b	$18,$F1,$3F,$A7,$8C,$DF,$BA,$38,$1F,$90,$8C,$DF,$B2,$32,$49,$FD
	.dc.b	$B4,$F4,$B3,$C2,$7F,$5A,$E4,$C3,$E1,$3C,$F3,$C9,$F0,$10,$F4,$15
	.dc.b	$D3,$E1,$AF,$24,$60,$8F,$F0,$C9,$F0,$4F,$18,$C3,$E0,$6B,$CC,$F3
	.dc.b	$99,$29,$93,$D0,$79,$0C,$57,$C0,$7B,$1F,$78,$CC,$FC,$08,$FB,$51
	.dc.b	$A3,$9D,$7A,$1E,$0D,$73,$5E,$68,$FA,$C6,$67,$C1,$3E,$2B,$ED,$7D
	.dc.b	$DB,$1D,$E7,$C2,$F9,$0C,$46,$9E,$46,$E7,$93,$4D,$94,$D9,$CC,$B5
	.dc.b	$E5,$8E,$31,$98,$F4,$4F,$76,$C7,$DF,$3F,$77,$3F,$A0,$23,$FA,$5A
	.dc.b	$F7,$BA,$31,$1A,$7F,$82,$BF,$D1,$3E,$FD,$F5,$14,$F0,$62,$A9,$3A
	.dc.b	$7F,$9C,$35,$F4,$46,$D4,$67,$04,$E6,$30,$FD,$72,$4F,$6A,$F1,$3F
	.dc.b	$26,$B8,$B9,$3C,$31,$E6,$70,$23,$6F,$8A,$34,$D3,$89,$D6,$7C,$91
	.dc.b	$E0,$8F,$59,$D2,$79,$D7,$A2,$3F,$3F,$1F,$91,$8C,$47,$53,$89,$63
	.dc.b	$8C,$F8,$A7,$67,$0C,$3F,$2B,$3C,$E7,$C4,$FE,$16,$BC,$E7,$B6,$FC
	.dc.b	$EF,$A4,$F3,$38,$7C,$37,$DA,$BC,$07,$49,$F9,$F3,$FA,$83,$89,$0C
	.dc.b	$12,$79,$3D,$CF,$64,$77,$AF,$34,$79,$4E,$53,$ED,$C3,$F8,$DB,$CA
	.dc.b	$3C,$B7,$E3,$7E,$5D,$F9,$27,$57,$D0,$F9,$33,$F0,$9C,$4F,$84,$91
	.dc.b	$C8,$F0,$D7,$57,$D4,$DA,$F2,$47,$39,$E2,$FB,$0E,$47,$DB,$3F,$B7
	.dc.b	$C7,$E7,$DB,$3F,$D4,$DD,$1C,$BF,$00,$7C,$6F,$F8,$C1,$B1,$DC,$F9
	.dc.b	$CE,$1F,$90,$1C,$61,$F0,$23,$37,$F8,$E9,$CE,$1F,$AF,$6C,$F4,$BF
	.dc.b	$01,$1F,$01,$D5,$C3,$E0,$8E,$2F,$C1,$34,$7F,$8D,$5C,$9A,$AF,$14
	.dc.b	$FC,$18,$F5,$4F,$F8,$51,$AF,$96,$34,$7F,$94,$BF,$02,$7A,$14,$F1
	.dc.b	$E3,$FB,$8B,$F1,$CE,$AF,$FC,$8C,$D7,$D1,$7C,$31,$FE,$21,$39,$BF
	.dc.b	$E0,$2D,$8F,$15,$72,$C3,$C0,$D3,$DD,$78,$7E,$9D,$71,$7F,$91,$35
	.dc.b	$F8,$A6,$A7,$85,$F8,$1F,$24,$70,$7F,$0D,$3E,$98,$FF,$1C,$B5,$F4
	.dc.b	$C6,$8F,$F3,$BF,$CF,$7F,$B7,$63,$F9,$93,$54,$F8,$CF,$24,$64,$D3
	.dc.b	$99,$E7,$E1,$CA,$F8,$07,$E4,$2B,$84,$66,$3D,$87,$F9,$D6,$32,$D8
	.dc.b	$F7,$E3,$FC,$FE,$73,$3C,$E7,$2F,$A1,$FD,$34,$1A,$48,$FF,$1E,$47
	.dc.b	$BF,$AD,$F0,$DF,$E8,$26,$F1,$F9,$73,$FE,$7B,$F1,$6F,$8A,$77,$7F
	.dc.b	$91,$63,$0D,$F8,$B7,$B1,$9A,$DE,$9F,$E6,$8F,$B0,$FF,$5B,$6F,$FE
	.dc.b	$4D,$6C,$67,$AF,$FA,$5B,$FD,$76,$32,$FF,$9C,$BE,$E7,$FD,$A9,$5F
	.dc.b	$E5,$56,$C7,$F9,$D1,$AE,$FB,$F9,$BF,$DB,$71,$78,$78,$8F,$3B,$63
	.dc.b	$37,$FE,$61,$6C,$FF,$A0,$3E,$11,$F0,$3C,$5B,$5F,$E6,$76,$CE,$E7
	.dc.b	$B7,$E0,$0F,$A4,$A4,$BF,$F3,$74,$65,$FF,$38,$7E,$19,$CA,$FF,$D2
	.dc.b	$2D,$7F,$C0,$0D,$7F,$C1,$0D,$4F,$8B,$8F,$FB,$AB,$67,$FC,$1D,$CA
	.dc.b	$F8,$E7,$C9,$FF,$80,$2F,$79,$DB,$EE,$DC,$7B,$B7,$0E,$57,$99,$F0
	.dc.b	$CD,$F4,$5F,$5F,$F3,$DB,$87,$FA,$35,$B5,$FE,$AC,$6B,$FE,$B0,$73
	.dc.b	$8F,$F8,$39,$B3,$EA,$C9,$EB,$77,$C2,$4F,$FB,$7E,$B7,$83,$3F,$81
	.dc.b	$7A,$8F,$E8,$7F,$0C,$F8,$7F,$76,$0F,$86,$C3,$FE,$E9,$9B,$5F,$DC
	.dc.b	$63,$55,$F0,$67,$3A,$7F,$8A,$A7,$27,$19,$EB,$7A,$8F,$F1,$D5,$5B
	.dc.b	$FF,$1C,$CD,$51,$99,$E7,$1F,$D6,$BA,$DF,$0F,$D3,$C2,$FA,$51,$8A
	.dc.b	$FF,$26,$65,$E3,$C7,$FA,$1C,$60,$FF,$25,$4E,$F0,$F6,$4E,$9E,$D8
	.dc.b	$75,$AE,$11,$FE,$57,$AE,$17,$D7,$0F,$0D,$EE,$BF,$B7,$FC,$23,$6F
	.dc.b	$C5,$73,$BE,$19,$F0,$FF,$89,$AD,$69,$DA,$B5,$39,$BE,$37,$C2,$BC
	.dc.b	$78,$30,$E2,$FF,$89,$70,$E3,$3A,$9C,$E7,$FE,$5E,$9C,$D7,$9B,$0E
	.dc.b	$B8,$F8,$FD,$BF,$D0,$5B,$E1,$F8,$E5,$FE,$71,$C9,$5A,$9C,$E3,$FE
	.dc.b	$7B,$9C,$D7,$C8,$BE,$58,$F5,$8C,$5F,$0B,$F5,$6B,$F0,$FE,$03,$E2
	.dc.b	$36,$BD,$4E,$1F,$83,$3F,$E8,$9F,$87,$FF,$45,$7C,$33,$D8,$F3,$5F
	.dc.b	$ED,$6B,$42,$74,$F9,$E3,$FD,$03,$18,$8F,$F4,$75,$74,$AE,$77,$67
	.dc.b	$FD,$24,$27,$0F,$AA,$2B,$E0,$AE,$9A,$7E,$35,$ED,$19,$BE,$AF,$95
	.dc.b	$EA,$F5,$30,$79,$EB,$AB,$99,$EB,$87,$AA,$B8,$E1,$E4,$C3,$C1,$18
	.dc.b	$BE,$67,$9C,$A7,$8C,$EC,$60,$EC,$F9,$AF,$D3,$18,$36,$E2,$BC,$17
	.dc.b	$E3,$9F,$8F,$18,$C3,$D2,$FF,$97,$63,$DC,$7C,$02,$9E,$73,$E7,$FC
	.dc.b	$1B,$D5,$7C,$3E,$E3,$0D,$8C,$CF,$CA,$9C,$B7,$E1,$FF,$D7,$4D,$E0
	.dc.b	$E5,$18,$8E,$ED,$7F,$DA,$12,$8E,$39,$FE,$1F,$E1,$9F,$0F,$FE,$D5
	.dc.b	$8C,$37,$6E,$35,$C1,$B8,$FE,$22,$F8,$E3,$6F,$83,$FE,$E5,$C7,$C1
	.dc.b	$5F,$8E,$6C,$7D,$AA,$E2,$DA,$E2,$A7,$78,$FE,$75,$B8,$7F,$BA,$EF
	.dc.b	$8C,$FA,$2B,$B6,$3E,$77,$C4,$D5,$3C,$2B,$96,$1E,$0C,$3A,$D7,$0C
	.dc.b	$7D,$8C,$3F,$23,$1A,$7C,$D8,$75,$C3,$A4,$0E,$F7,$D6,$74,$E6,$3F
	.dc.b	$D3,$B3,$F0,$AF,$FD,$EA,$DD,$7E,$7C,$FC,$53,$F9,$4C,$65,$EC,$D7
	.dc.b	$C7,$3F,$BC,$F8,$7F,$E9,$F1,$FC,$C1,$AF,$FA,$B1,$B0,$AF,$27,$C2
	.dc.b	$38,$BC,$D3,$9B,$F6,$5F,$0B,$F5,$36,$33,$D3,$C2,$79,$DD,$9E,$3F
	.dc.b	$0F,$55,$A3,$0D,$8F,$E8,$F8,$FF,$82,$1F,$54,$E8,$8E,$BB,$EB,$ED
	.dc.b	$8B,$97,$A6,$15,$CD,$FF,$66,$FC,$3F,$F4,$47,$FA,$AB,$5C,$F4,$F9
	.dc.b	$E7,$FC,$A7,$1A,$8F,$54,$EE,$BB,$37,$3F,$97,$E5,$6A,$A5,$F1,$3D
	.dc.b	$B1,$F0,$FE,$1F,$FD,$34,$4F,$5B,$C1,$D9,$96,$57,$F6,$A9,$FF,$50
	.dc.b	$67,$FD,$DE,$31,$3A,$7F,$DD,$33,$AA,$FA,$1F,$08,$F8,$5A,$4A,$FF
	.dc.b	$03,$F8,$CF,$86,$7C,$E6,$67,$F9,$ED,$F1,$7F,$90,$35,$CD,$FF,$C0
	.dc.b	$18,$C7,$07,$E2,$70,$E9,$F1,$1F,$83,$6D,$FF,$81,$2B,$7C,$1F,$A5
	.dc.b	$7E,$7C,$2F,$ED,$46,$1B,$BF,$3C,$DF,$5B,$EE,$8F,$11,$2F,$CE,$D7
	.dc.b	$A6,$1F,$E6,$EA,$FF,$50,$C6,$0F,$F2,$93,$8B,$EF,$1E,$7F,$89,$F9
	.dc.b	$64,$7F,$C3,$BE,$2F,$F9,$DD,$D1,$F1,$A3,$E2,$11,$C1,$F0,$7B,$BE
	.dc.b	$A9,$CD,$F1,$0F,$89,$FD,$A3,$FE,$2A,$6B,$A6,$DF,$E7,$FE,$A7,$FD
	.dc.b	$43,$3B,$EB,$FE,$AC,$F3,$6D,$72,$8F,$F0,$2B,$5F,$E2,$DF,$10,$F5
	.dc.b	$F8,$B6,$3F,$B0,$35,$FF,$00,$B5,$FF,$07,$FC,$4B,$DB,$FE,$00,$D8
	.dc.b	$FF,$21,$B5,$F7,$FD,$43,$EA,$77,$5E,$8A,$D5,$74,$8C,$11,$FF,$26
	.dc.b	$DF,$F6,$F6,$F9,$1C,$9E,$F3,$FC,$46,$75,$1F,$39,$BA,$FF,$07,$BF
	.dc.b	$23,$6B,$85,$CB,$E6,$43,$93,$97,$EC,$D7,$2E,$2F,$E8,$0D,$8F,$E8
	.dc.b	$D8,$78,$27,$E7,$D7,$9E,$78,$DF,$3C,$3F,$8D,$5E,$E7,$FA,$76,$1C
	.dc.b	$D3,$63,$50,$6C,$66,$FA,$61,$E0,$AF,$55,$F6,$9C,$C7,$99,$F2,$46
	.dc.b	$63,$E6,$90,$F3,$95,$D3,$F1,$0E,$B3,$DD,$C1,$23,$E7,$3C,$63,$51
	.dc.b	$C8,$CC,$70,$3F,$CB,$77,$F6,$9B,$5F,$96,$23,$FE,$90,$F8,$A7,$C1
	.dc.b	$7F,$51,$6C,$77,$3D,$C6,$E3,$47,$C2,$7E,$54,$E1,$D1,$C4,$C9,$83
	.dc.b	$F7,$33,$D9,$F7,$61,$EC,$3F,$C0,$51,$9C,$3F,$AB,$36,$35,$3E,$43
	.dc.b	$9D,$F8,$5E,$67,$F8,$3E,$78,$36,$35,$5F,$26,$BB,$C6,$1B,$3E,$9A
	.dc.b	$E2,$9A,$63,$0F,$C5,$B8,$AE,$53,$E5,$BF,$4C,$62,$39,$3F,$26,$78
	.dc.b	$47,$C5,$9F,$23,$E7,$31,$7E,$47,$DA,$72,$32,$6C,$CC,$F4,$7E,$29
	.dc.b	$EB,$AF,$14,$7A,$70,$B1,$E5,$C3,$D6,$E2,$B8,$1D,$8F,$01,$D9,$D3
	.dc.b	$4D,$57,$56,$C7,$5A,$DE,$1F,$12,$7A,$9E,$62,$3F,$09,$EE,$ED,$CE
	.dc.b	$1F,$82,$72,$FC,$58,$D1,$C6,$FE,$13,$88,$E1,$18,$78,$18,$BF,$2B
	.dc.b	$6B,$D2,$79,$AB,$93,$F3,$DE,$D3,$C2,$3F,$01,$3D,$1D,$35,$FD,$19
	.dc.b	$EA,$32,$43,$33,$D9,$F6,$99,$C3,$EA,$BE,$43,$E8,$1F,$EC,$EB,$FD
	.dc.b	$63,$FE,$DC,$6B,$F8,$93,$17,$FD,$82,$33,$1F,$89,$32,$7F,$AE,$5F
	.dc.b	$CF,$FC,$5F,$F0,$53,$DD,$1F,$A3,$70,$F8,$EF,$F6,$F1,$88,$F7,$1E
	.dc.b	$A3,$81,$FC,$45,$E7,$27,$84,$FC,$18,$F0,$CE,$CE,$D8,$7F,$06,$AD
	.dc.b	$BF,$36,$77,$3F,$C3,$A7,$A9,$A7,$E7,$46,$AB,$CF,$1D,$0F,$E1,$13
	.dc.b	$B7,$D1,$3B,$AF,$04,$6C,$F5,$9D,$CC,$47,$AD,$F4,$1F,$09,$E4,$F6
	.dc.b	$37,$3D,$E3,$F2,$91,$93,$A1,$C6,$F5,$3E,$C8,$EB,$7F,$1E,$73,$1D
	.dc.b	$0C,$DF,$C5,$8C,$C7,$33,$27,$79,$E4,$E0,$F6,$B8,$7C,$6F,$C8,$79
	.dc.b	$18,$9F,$8A,$FB,$EB,$BC,$FD,$B6,$CF,$D7,$6C,$FD,$47,$F2,$84,$7F
	.dc.b	$24,$D7,$FA,$4F,$C5,$FF,$57,$38,$BF,$CD,$5C,$3E,$9C,$3E,$93,$C0
	.dc.b	$FE,$C1,$1F,$3C,$DB,$93,$D4,$7F,$49,$73,$7F,$D4,$9E,$11,$C8,$87
	.dc.b	$03,$C4,$7E,$A5,$FD,$71,$1C,$1F,$EE,$38,$FF,$59,$7C,$53,$E2,$FF
	.dc.b	$D8,$4D,$9E,$B3,$E1,$78,$63,$FE,$E7,$9D,$47,$F9,$91,$B3,$B8,$F6
	.dc.b	$3D,$AB,$A9,$EC,$AE,$4E,$27,$A9,$FE,$F7,$AE,$73,$FD,$D2,$F5,$3D
	.dc.b	$EA,$DF,$73,$B4,$72,$17,$C9,$E9,$3E,$16,$9F,$F0,$04,$F1,$94,$C9
	.dc.b	$63,$DE,$ED,$E8,$BE,$2D,$F1,$7F,$EE,$B6,$C7,$8E,$76,$70,$8C,$C7
	.dc.b	$FA,$31,$B7,$EA,$8F,$71,$EB,$8C,$E2,$89,$18,$7C,$F3,$FE,$ED,$8C
	.dc.b	$47,$96,$FC,$87,$C8,$AE,$51,$93,$B4,$69,$F4,$61,$5F,$34,$5B,$F3
	.dc.b	$1D,$DF,$F4,$DD,$FB,$27,$E6,$35,$EE,$D7,$D3,$1E,$29,$5F,$93,$E2
	.dc.b	$DE,$A6,$63,$0F,$85,$F4,$FC,$F7,$FA,$F7,$0F,$B0,$D7,$EE,$35,$DF
	.dc.b	$BF,$FE,$B0,$F8,$A7,$C4,$FF,$B3,$A7,$E1,$C6,$3D,$DF,$C1,$42,$1F
	.dc.b	$14,$F7,$73,$1B,$9F,$C4,$7C,$57,$FA,$B7,$3E,$2D,$8C,$F5,$7B,$9B
	.dc.b	$3D,$BC,$87,$CA,$E6,$7F,$29,$18,$9D,$4E,$67,$EC,$CE,$5F,$CC,$62
	.dc.b	$F1,$F5,$1F,$87,$8C,$5F,$B1,$B1,$C2,$76,$BE,$29,$F0,$D8,$6D,$FC
	.dc.b	$46,$CF,$E9,$DA,$FE,$AF,$E2,$33,$CB,$EF,$C9,$F1,$9F,$2B,$0F,$DB
	.dc.b	$56,$9F,$CE,$4E,$B8,$3E,$35,$FC,$3A,$EB,$87,$E6,$1A,$FE,$8D,$A9
	.dc.b	$A7,$57,$59,$74,$C3,$BF,$1B,$E2,$71,$5A,$9F,$E1,$5E,$93,$E2,$73
	.dc.b	$A3,$07,$A5,$C3,$05,$EF,$0F,$E3,$77,$BA,$D4,$E6,$BF,$87,$35,$FE
	.dc.b	$24,$D7,$F6,$AD,$7F,$80,$3C,$74,$FC,$33,$6F,$D7,$ED,$76,$CB,$C9
	.dc.b	$F1,$0F,$89,$F6,$D3,$E2,$70,$FF,$27,$9B,$93,$2F,$CB,$4E,$EF,$99
	.dc.b	$CA,$F7,$3F,$97,$BB,$5F,$C5,$4E,$07,$5A,$E0,$72,$7F,$9F,$9A,$BF
	.dc.b	$03,$FC,$53,$A9,$EF,$F7,$EE,$F9,$D1,$89,$E3,$97,$F0,$67,$FA,$04
	.dc.b	$66,$F9,$9F,$C3,$27,$4E,$04,$6F,$2F,$7C,$E9,$FD,$94,$66,$BD,$93
	.dc.b	$97,$CD,$87,$95,$FE,$A1,$F0,$FF,$82,$7F,$AD,$F4,$BE,$25,$EA,$FE
	.dc.b	$1E,$B5,$9F,$9A,$B5,$39,$E1,$FC,$9C,$63,$0E,$B1,$8C,$39,$C6,$0E
	.dc.b	$EE,$2F,$DD,$87,$6D,$3F,$88,$61,$CE,$73,$EF,$7E,$82,$5F,$13,$BC
	.dc.b	$7F,$A2,$E0,$AD,$4E,$6F,$FA,$94,$E6,$BF,$BC,$DF,$2C,$BD,$D8,$70
	.dc.b	$7E,$06,$FF,$07,$D4,$B7,$F2,$1B,$8F,$A2,$7F,$58,$FF,$5F,$F8,$9F
	.dc.b	$EC,$1F,$10,$F6,$3C,$B5,$AA,$56,$A7,$38,$59,$FE,$FD,$35,$69,$F8
	.dc.b	$4C,$3A,$46,$83,$87,$78,$7F,$39,$8C,$47,$94,$DE,$1E,$06,$9C,$E3
	.dc.b	$95,$77,$9F,$F0,$75,$F8,$E3,$15,$B9,$E5,$5D,$A3,$DA,$74,$3C,$78
	.dc.b	$7A,$DF,$14,$76,$C3,$CD,$5D,$A7,$85,$F6,$7F,$C0,$38,$78,$6B,$71
	.dc.b	$E1,$3E,$EB,$87,$67,$27,$0F,$82,$BC,$E7,$77,$FB,$4C,$FF,$6A,$7E
	.dc.b	$31,$83,$FC,$5D,$F1,$0F,$4D,$F1,$3F,$DC,$23,$0D,$8C,$CF,$06,$BF
	.dc.b	$E3,$99,$7C,$4F,$F1,$98,$EA,$D7,$4D,$7F,$C4,$31,$C3,$83,$C1,$3C
	.dc.b	$74,$FC,$07,$C4,$3E,$27,$FC,$4F,$18,$6E,$9C,$AB,$83,$6F,$F1,$D7
	.dc.b	$C7,$1B,$7C,$1F,$F1,$D6,$3E,$17,$EE,$37,$1E,$D1,$C0,$FC,$C4,$67
	.dc.b	$0E,$D1,$F8,$3B,$EC,$DB,$EC,$AB,$83,$E0,$C7,$CE,$FF,$92,$EB,$8E
	.dc.b	$1E,$08,$F1,$CE,$9E,$4D,$9D,$E1,$FE,$4A,$6C,$F5,$C7,$CF,$5B,$C7
	.dc.b	$C7,$5A,$AE,$75,$ED,$7F,$9A,$47,$AA,$3F,$C4,$4D,$DF,$DF,$8F,$96
	.dc.b	$73,$7F,$E4,$5B,$F5,$35,$EA,$7E,$9F,$E2,$7F,$A5,$BF,$E8,$56,$BF
	.dc.b	$D0,$DB,$3F,$E3,$1F,$87,$71,$73,$9C,$E3,$FE,$68,$6D,$FF,$9C,$DA
	.dc.b	$AB,$8B,$DF,$E2,$79,$11,$FF,$4B,$43,$C0,$C4,$78,$E6,$E3,$E3,$9B
	.dc.b	$87,$86,$6D,$7F,$A0,$EF,$93,$FE,$A1,$F8,$9F,$E7,$0D,$E7,$F8,$91
	.dc.b	$88,$FE,$C5,$3F,$E4,$77,$44,$3F,$B3,$46,$F1,$F8,$9E,$5F,$2B,$95
	.dc.b	$B9,$CB,$FE,$B1,$F8,$87,$C4,$FD,$36,$C6,$63,$EB,$63,$FE,$28,$8C
	.dc.b	$10,$FE,$ED,$3C,$1F,$3F,$57,$3A,$91,$BB,$FB,$13,$F8,$1F,$87,$FF
	.dc.b	$1A,$CF,$FB,$0D,$AF,$F8,$05,$B3,$FE,$55,$F8,$67,$C4,$F1,$F2,$25
	.dc.b	$F1,$7F,$1D,$E4,$AB,$73,$93,$FD,$0F,$87,$67,$FD,$98,$D7,$FD,$5E
	.dc.b	$D7,$FD,$74,$FD,$17,$C9,$B7,$A4,$EF,$2F,$88,$C4,$FA,$E6,$BF,$E1
	.dc.b	$EB,$F2,$4E,$23,$C3,$3F,$CC,$6F,$B4,$60,$FF,$4E,$F8,$9F,$EA,$57
	.dc.b	$FD,$98,$D7,$FC,$2F,$E5,$FF,$39,$8C,$4B,$E2,$3E,$71,$2B,$FD,$E1
	.dc.b	$3F,$EB,$3D,$7F,$D3,$55,$B9,$F8,$B5,$E3,$23,$FE,$AE,$F8,$8F,$F4
	.dc.b	$7B,$7C,$FF,$E9,$11,$88,$FF,$7A,$35,$FF,$7C,$7C,$47,$EB,$9F,$F1
	.dc.b	$E3,$70,$FF,$55,$37,$A3,$FD,$BF,$5C,$E7,$3D,$7F,$D5,$6B,$95,$F4
	.dc.b	$8C,$1F,$D6,$7E,$1D,$F0,$FF,$C3,$5B,$3F,$EF,$3B,$FF,$5C,$46,$23
	.dc.b	$EA,$7A,$AF,$8B,$E1,$7C,$C9,$3F,$EC,$EA,$E0,$FF,$BD,$03,$CE,$FF
	.dc.b	$D8,$9F,$10,$F6,$BF,$8F,$36,$BC,$5A,$FF,$BF,$2F,$83,$70,$DD,$CA
	.dc.b	$D8,$3B,$C7,$99,$C8,$BE,$CF,$F9,$A5,$B8,$7C,$09,$E1,$39,$AE,$CF
	.dc.b	$F9,$BF,$0F,$1B,$DD,$AF,$58,$DD,$73,$8F,$0C,$EA,$FB,$E1,$E7,$9E
	.dc.b	$06,$B0,$E3,$19,$9D,$CF,$B5,$AF,$B8,$F6,$B8,$26,$1E,$AC,$3E,$7B
	.dc.b	$D2,$79,$39,$AE,$CE,$63,$C6,$7B,$EF,$17,$D2,$3C,$C7,$82,$72,$73
	.dc.b	$7A,$99,$21,$CC,$A7,$E1,$63,$E9,$B6,$3F,$C5,$D1,$88,$F2,$9F,$6B
	.dc.b	$E2,$3F,$C5,$CD,$F2,$7C,$2D,$21,$F6,$4F,$0E,$2C,$7E,$85,$72,$9F
	.dc.b	$F5,$53,$67,$E3,$C7,$09,$DD,$7F,$73,$AD,$47,$92,$7B,$1C,$A7,$47
	.dc.b	$91,$C1,$E4,$AE,$AF,$07,$C8,$FE,$66,$BA,$B6,$7E,$A4,$62,$0D,$31
	.dc.b	$7F,$84,$71,$3E,$29,$F5,$1C,$CF,$15,$FB,$1C,$9D,$23,$83,$E8,$3C
	.dc.b	$D1,$B7,$17,$E8,$78,$1C,$AB,$C2,$F6,$3F,$06,$70,$9E,$53,$C6,$7D
	.dc.b	$38,$78,$23,$17,$6B,$E7,$46,$2F,$A1,$E0,$3C,$A7,$B4,$EC,$57,$E7
	.dc.b	$9C,$5C,$9C,$5C,$C7,$90,$91,$C5,$F0,$98,$78,$37,$0F,$A1,$5E,$79
	.dc.b	$E0,$7C,$0A,$F0,$12,$B8,$5E,$E7,$99,$C0,$FD,$8C,$ED,$C6,$1F,$19
	.dc.b	$F8,$2F,$CE,$7E,$09,$C8,$8E,$DF,$7B,$B8,$F7,$C7,$4B,$F1,$61,$EA
	.dc.b	$8E,$B1,$F9,$03,$CD,$5C,$EB,$C9,$3D,$A7,$4F,$03,$F0,$4F,$03,$2F
	.dc.b	$89,$B1,$CA,$7A,$61,$F3,$5F,$33,$F4,$2B,$EE,$1F,$D2,$E3,$F7,$0D
	.dc.b	$AF,$F5,$A4,$62,$3F,$2C,$64,$FE,$A4,$72,$32,$7D,$BF,$86,$7C,$47
	.dc.b	$F9,$BD,$B1,$EB,$18,$7E,$7E,$33,$18,$C3,$F9,$D4,$62,$7D,$98,$73
	.dc.b	$B9,$7C,$E7,$94,$62,$3E,$21,$FD,$5A,$B9,$5F,$49,$F9,$53,$97,$85
	.dc.b	$76,$8C,$10,$E6,$79,$4F,$EC,$95,$D1,$B3,$D1,$D0,$8D,$36,$35,$3E
	.dc.b	$88,$EC,$F8,$5E,$33,$C0,$E3,$87,$DA,$8E,$E7,$0B,$FB,$D8,$7A,$9F
	.dc.b	$74,$7C,$56,$BD,$30,$F0,$C6,$30,$F1,$4F,$B9,$A4,$8E,$B1,$DC,$F0
	.dc.b	$9B,$79,$11,$FD,$11,$B3,$E0,$1E,$33,$8C,$75,$76,$79,$4E,$17,$EF
	.dc.b	$78,$99,$3C,$07,$29,$F0,$3F,$A9,$3C,$18,$7C,$8C,$3F,$82,$3F,$2F
	.dc.b	$0F,$83,$3C,$48,$71,$3B,$61,$E5,$70,$D9,$E9,$1F,$CA,$1E,$8D,$9F
	.dc.b	$A4,$F4,$C3,$FA,$74,$EA,$3A,$E1,$FA,$7A,$DC,$7B,$AF,$F8,$1D,$6D
	.dc.b	$F1,$5F,$2A,$DC,$F8,$AB,$99,$EA,$71,$5E,$79,$EB,$5C,$6B,$F7,$27
	.dc.b	$CC,$3F,$C9,$95,$FB,$26,$CE,$9F,$25,$FE,$D3,$E2,$3F,$78,$78,$A6
	.dc.b	$CF,$EB,$E3,$0D,$21,$E6,$3F,$48,$F0,$7F,$3E,$FB,$0C,$61,$FC,$15
	.dc.b	$C4,$72,$8C,$3F,$7A,$3C,$D6,$69,$23,$DF,$19,$7F,$82,$7C,$47,$F1
	.dc.b	$53,$BE,$1E,$D8,$C9,$0F,$29,$E1,$7F,$8A,$C7,$18,$C1,$FC,$11,$C3
	.dc.b	$FB,$A7,$AB,$93,$A5,$FF,$21,$8C,$C7,$F0,$77,$8C,$66,$3A,$99,$3D
	.dc.b	$71,$F0,$08,$E0,$FF,$3A,$FC,$43,$E2,$3F,$CC,$95,$FE,$0D,$8C,$4F
	.dc.b	$8A,$FF,$CD,$1E,$FB,$E7,$7F,$00,$D9,$FC,$3B,$5F,$E8,$2D,$7F,$D2
	.dc.b	$9F,$0D,$FE,$C5,$27,$6F,$E2,$5F,$F6,$47,$C2,$BE,$19,$EF,$7E,$4A
	.dc.b	$30,$DC,$1E,$DE,$6B,$78,$FF,$50,$F9,$56,$5F,$E2,$78,$FF,$40,$46
	.dc.b	$34,$FC,$EB,$63,$84,$ED,$FF,$4E,$78,$AF,$86,$C4,$23,$17,$FA,$66
	.dc.b	$C7,$F6,$06,$BF,$EB,$3C,$5E,$2F,$26,$D7,$EB,$9B,$3F,$DA,$1A,$FF
	.dc.b	$AE,$FE,$15,$E1,$FF,$96,$38,$FF,$CB,$38,$BC,$06,$F9,$CB,$FD,$3D
	.dc.b	$87,$07,$FD,$6F,$5B,$9D,$63,$FE,$B9,$AD,$3F,$ED,$08,$CA,$8C,$E1
	.dc.b	$E0,$C3,$BE,$1F,$EB,$E9,$E3,$5E,$7C,$7E,$3C,$62,$3C,$71,$F2,$BC
	.dc.b	$B7,$C2,$78,$B3,$FA,$FE,$5F,$82,$73,$8F,$A3,$81,$39,$BE,$11,$89
	.dc.b	$FF,$67,$D7,$0C,$7F,$B9,$61,$D2,$73,$7E,$12,$3F,$C7,$AB,$85,$7F
	.dc.b	$BA,$7E,$11,$F0,$9F,$1A,$9F,$09,$87,$FD,$BB,$36,$BF,$C0,$11,$AA
	.dc.b	$F9,$33,$9D,$FF,$D1,$F3,$93,$8C,$F5,$BD,$47,$FB,$56,$AD,$FF,$AA
	.dc.b	$66,$A8,$CC,$F3,$8F,$F0,$5F,$33,$E1,$3A,$7C,$0F,$DA,$8C,$57,$FB
	.dc.b	$E0,$F5,$C6,$67,$71,$FE,$F8,$9D,$D7,$B3,$05,$EE,$B9,$46,$1E,$17
	.dc.b	$CA,$BF,$C3,$BF,$04,$D3,$F1,$FC,$AF,$84,$7A,$1F,$8E,$AD,$62,$C3
	.dc.b	$FC,$55,$7B,$AD,$61,$CB,$0E,$37,$BD,$FF,$D4,$98,$F2,$C7,$8D,$69
	.dc.b	$F1,$DF,$1C,$39,$E1,$C4,$F0,$63,$DD,$C5,$78,$F0,$7A,$FF,$A8,$C9
	.dc.b	$F0,$9D,$DF,$0C,$66,$31,$7E,$BB,$5F,$28,$EB,$7C,$B2,$F8,$58,$71
	.dc.b	$7E,$26,$DF,$17,$CE,$B7,$E4,$C7,$9C,$61,$F9,$F3,$FB,$87,$E5,$7C
	.dc.b	$27,$CB,$F7,$DF,$05,$E3,$D3,$EC,$57,$F9,$48,$D8,$9D,$BF,$06,$FF
	.dc.b	$DC,$F1,$88,$F6,$5F,$5B,$E9,$5A,$7F,$D0,$31,$99,$E5,$8D,$78,$9D
	.dc.b	$E7,$8C,$66,$3C,$77,$DF,$0F,$3B,$E1,$8C,$D7,$07,$91,$F3,$C5,$F7
	.dc.b	$AE,$33,$E4,$9D,$9F,$53,$0F,$43,$C6,$B8,$46,$23,$CF,$7E,$38,$C4
	.dc.b	$78,$8F,$69,$D0,$C1,$D7,$0F,$63,$E4,$3C,$38,$79,$1F,$01,$89,$EA
	.dc.b	$EC,$FF,$19,$DF,$F8,$BD,$C1,$EB,$3E,$07,$C1,$BC,$D7,$C2,$7F,$8E
	.dc.b	$A3,$0D,$8C,$CF,$FA,$DA,$72,$DF,$84,$FF,$58,$B7,$83,$94,$62,$3B
	.dc.b	$B5,$FC,$F4,$A3,$8E,$9F,$87,$F8,$47,$C2,$7E,$8E,$30,$DD,$F8,$D7
	.dc.b	$06,$E3,$F8,$7B,$E3,$8D,$BE,$0F,$F0,$7C,$7C,$15,$F8,$D6,$C7,$DE
	.dc.b	$AE,$2D,$AE,$2A,$77,$8F,$E7,$5B,$87,$EE,$6F,$8C,$FA,$2B,$B6,$3E
	.dc.b	$77,$C4,$D5,$3C,$2B,$96,$1E,$0C,$3A,$D7,$0C,$7D,$8C,$3F,$2B,$1A
	.dc.b	$7C,$D8,$75,$C3,$A4,$7F,$27,$6B,$AA,$E4,$37,$F0,$C7,$F0,$99,$F8
	.dc.b	$73,$98,$F1,$1F,$14,$D6,$1F,$E4,$4A,$E3,$1F,$ED,$8C,$3C,$9F,$09
	.dc.b	$FD,$3E,$3F,$8A,$35,$FF,$6E,$B6,$FB,$43,$E0,$5C,$5D,$63,$FA,$05
	.dc.b	$6F,$0F,$E5,$0D,$8C,$F4,$ED,$E5,$56,$5F,$07,$49,$3F,$C9,$F0,$7C
	.dc.b	$26,$23,$CB,$37,$2F,$2D,$6A,$7D,$73,$9C,$BC,$6F,$86,$75,$3F,$EF
	.dc.b	$AF,$84,$FE,$8A,$FF,$35,$6B,$9E,$8E,$1E,$4F,$A6,$73,$8F,$C4,$6E
	.dc.b	$3F,$D1,$DB,$7F,$D8,$F0,$FF,$24,$BD,$70,$E4,$47,$DD,$87,$C5,$F8
	.dc.b	$47,$C2,$52,$4F,$C3,$EB,$AB,$3F,$AB,$5C,$23,$FA,$73,$5D,$67,$FD
	.dc.b	$6F,$2F,$EF,$91,$AC,$39,$DF,$27,$E3,$61,$F5,$FF,$73,$7C,$1F,$FB
	.dc.b	$1D,$B7,$FE,$CE,$6B,$FC,$E4,$38,$7F,$9E,$3A,$BF,$F3,$0F,$C2,$3E
	.dc.b	$AF,$F1,$5E,$4A,$31,$7F,$D9,$EF,$86,$9F,$DC,$DB,$1F,$E7,$B9,$DB
	.dc.b	$FE,$48,$6E,$1F,$98,$9D,$CF,$F8,$63,$E0,$7F,$DC,$1E,$21,$A4,$8F
	.dc.b	$F7,$4C,$65,$FF,$38,$DF,$CB,$BF,$4E,$BF,$9B,$AD,$57,$CE,$3F,$C4
	.dc.b	$4E,$27,$71,$CA,$78,$5F,$E8,$FE,$09,$F2,$5F,$EF,$36,$D7,$FB,$E9
	.dc.b	$AF,$CD,$6B,$FD,$9D,$CD,$FE,$DB,$F0,$0F,$C8,$C9,$B3,$AD,$3F,$C3
	.dc.b	$91,$F9,$E9,$FE,$E0,$D7,$3C,$9F,$6D,$FF,$1E,$46,$5F,$F1,$F7,$C3
	.dc.b	$BE,$1F,$50,$8C,$3C,$5A,$FE,$15,$AF,$F7,$F7,$38,$FF,$0E,$7F,$C2
	.dc.b	$CD,$F0,$7F,$84,$46,$B1,$FF,$1A,$3F,$E0,$BA,$B3,$E2,$8C,$47,$9A
	.dc.b	$3C,$BF,$0F,$FE,$B2,$8F,$F1,$96,$7F,$C6,$9B,$DB,$FC,$56,$75,$1F
	.dc.b	$6C,$4E,$DC,$37,$1F,$E4,$AD,$C7,$FC,$91,$5C,$A7,$37,$F7,$C9,$7F
	.dc.b	$E4,$39,$D5,$7F,$93,$EF,$94,$FF,$9A,$DA,$F8,$AB,$CD,$E9,$F2,$8F
	.dc.b	$1C,$5C,$3F,$9B,$E1,$D2,$79,$4E,$70,$EC,$F2,$8F,$82,$43,$99,$E5
	.dc.b	$BE,$0F,$23,$57,$DA,$33,$1E,$53,$35,$F0,$A3,$38,$7A,$8F,$5E,$1E
	.dc.b	$5A,$E0,$78,$09,$7D,$5E,$67,$84,$CD,$F5,$72,$72,$8F,$9E,$FA,$4C
	.dc.b	$D7,$87,$0F,$81,$1C,$8E,$93,$C1,$FC,$05,$F9,$6B,$F0,$11,$FA,$FA
	.dc.b	$FD,$B1,$1F,$DB,$3F,$D7,$7E,$0D,$F1,$F4,$F8,$31,$67,$FD,$05,$3A
	.dc.b	$7C,$B1,$97,$F2,$11,$97,$CC,$7B,$71,$FF,$4A,$46,$71,$FE,$EA,$D8
	.dc.b	$FB,$B3,$F1,$23,$0F,$FA,$D5,$AF,$33,$E0,$C7,$4B,$F1,$57,$C3,$24
	.dc.b	$66,$7D,$33,$F8,$0A,$E6,$47,$07,$29,$E1,$3E,$A9,$D3,$C8,$F3,$4E
	.dc.b	$DF,$33,$C9,$D1,$F1,$1C,$34,$CD,$79,$9E,$26,$2B,$C3,$19,$85,$78
	.dc.b	$9E,$47,$96,$BC,$4E,$9E,$F3,$F3,$9C,$E1,$FB,$4A,$E5,$39,$AF,$59
	.dc.b	$E1,$3C,$67,$53,$07,$53,$67,$89,$E4,$EC,$C5,$FA,$8F,$BA,$E0,$EB
	.dc.b	$1C,$0F,$74,$F9,$27,$84,$FD,$E3,$6E,$1F,$11,$DE,$79,$46,$6B,$DD
	.dc.b	$7F,$12,$FF,$13,$18,$BF,$5C,$63,$0D,$57,$C6,$6C,$FC,$0A,$E0,$FA
	.dc.b	$4F,$0E,$1E,$8A,$DC,$78,$F0,$F3,$61,$D6,$3B,$CF,$56,$9D,$70,$E3
	.dc.b	$7E,$A8,$F8,$07,$B8,$FF,$67,$57,$EB,$5B,$5F,$D6,$A3,$11,$F9,$83
	.dc.b	$27,$F6,$37,$2F,$DC,$32,$7F,$47,$F8,$5F,$ED,$C8,$38,$7D,$F8,$7D
	.dc.b	$18,$C1,$99,$FB,$13,$C6,$33,$C3,$EC,$AE,$93,$D2,$3F,$CD,$93,$C6
	.dc.b	$FF,$32,$F0,$9F,$8A,$79,$67,$67,$37,$B3,$87,$78,$7F,$A0,$27,$94
	.dc.b	$7D,$7C,$3F,$84,$C6,$27,$E7,$4F,$B0,$E6,$7C,$17,$95,$77,$8F,$41
	.dc.b	$9C,$3D,$75,$E9,$8F,$D2,$BD,$63,$F0,$38,$79,$67,$D4,$79,$A3,$8D
	.dc.b	$66,$FC,$78,$79,$A3,$13,$CA,$BC,$07,$98,$E6,$7B,$6F,$6E,$6F,$9C
	.dc.b	$62,$3D,$26,$CE,$66,$23,$DE,$F9,$4E,$53,$E6,$70,$6D,$C3,$CC,$F1
	.dc.b	$C7,$27,$0D,$24,$7A,$E3,$6F,$28,$F3,$C7,$93,$0F,$94,$FF,$00,$21
	.dc.b	$E1,$3C,$A6,$30,$F9,$0F,$1C,$3E,$C4,$F2,$8F,$BC,$47,$13,$C2,$3C
	.dc.b	$67,$43,$53,$F2,$67,$EF,$46,$8F,$11,$27,$95,$7B,$A3,$94,$62,$FD
	.dc.b	$C4,$8F,$5D,$79,$4E,$B5,$C6,$3A,$C6,$27,$E7,$C7,$EA,$8F,$B6,$58
	.dc.b	$FF,$36,$BA,$3B,$0F,$93,$7C,$47,$FB,$6D,$B1,$F1,$C8,$7F,$92,$5C
	.dc.b	$B6,$3E,$16,$3C,$59,$1C,$30,$F4,$57,$07,$FD,$E2,$DC,$3C,$78,$7F
	.dc.b	$0B,$8C,$57,$F0,$F6,$BF,$03,$0E,$AF,$84,$ED,$5E,$C9,$D9,$C2,$BA
	.dc.b	$61,$C3,$0F,$0B,$FA,$19,$DB,$9A,$E4,$DB,$F1,$3B,$30,$FC,$97,$0F
	.dc.b	$EA,$27,$9B,$E4,$38,$99,$8E,$37,$E4,$8F,$F4,$A4,$FC,$C7,$ED,$35
	.dc.b	$FE,$4A,$D7,$99,$FE,$B5,$F9,$5F,$F7,$47,$D8,$FF,$2A,$C5,$8F,$F3
	.dc.b	$79,$CD,$78,$4F,$DE,$FC,$3B,$E2,$7F,$D3,$ED,$8F,$F5,$33,$5F,$F7
	.dc.b	$13,$5F,$C7,$2F,$88,$E2,$48,$E2,$FF,$AA,$BE,$49,$F2,$54,$93,$FE
	.dc.b	$B3,$8C,$BF,$EB,$5F,$86,$7C,$33,$DC,$C3,$67,$FD,$E9,$2F,$84,$7B
	.dc.b	$7A,$BF,$F5,$F7,$C8,$BE,$37,$E1,$E1,$FE,$FB,$8C,$70,$6A,$73,$3F
	.dc.b	$A9,$9C,$BF,$F0,$07,$E5,$7F,$CC,$35,$FE,$C9,$6C,$67,$A7,$8B,$FD
	.dc.b	$6A,$32,$FE,$CF,$F0,$E7,$BB,$A8,$FE,$D5,$18,$BF,$F6,$A3,$63,$F7
	.dc.b	$41,$C3,$CE,$7D,$7C,$BE,$89,$E0,$FB,$B9,$43,$D0,$DE,$1F,$DB,$AB
	.dc.b	$6D,$BC,$E1,$E4,$6D,$F2,$BC,$CF,$8E,$6F,$9C,$FB,$BF,$64,$A3,$18
	.dc.b	$7F,$BB,$DB,$1E,$C6,$BF,$C5,$DC,$E5,$F4,$9B,$3F,$E0,$FF,$35,$F0
	.dc.b	$DF,$EF,$58,$FE,$AB,$3A,$BF,$F0,$A3,$5F,$16,$1E,$CB,$E3,$E3,$FA
	.dc.b	$23,$D7,$F0,$8F,$85,$F8,$38,$BE,$17,$0F,$F6,$29,$B8,$B4,$FC,$5C
	.dc.b	$EE,$F9,$9C,$AF,$73,$F8,$CB,$B5,$FE,$29,$38,$1D,$6B,$81,$C9,$FE
	.dc.b	$F6,$6A,$FF,$DF,$1C,$AF,$6B,$B6,$FF,$52,$31,$5C,$32,$FB,$B7,$FE
	.dc.b	$37,$9C,$BE,$99,$F1,$4E,$A3,$F5,$D7,$D2,$7F,$C0,$35,$C3,$2F,$7D
	.dc.b	$EE,$31,$3C,$B0,$FC,$2F,$C1,$FE,$06,$3F,$C2,$7D,$0F,$85,$7A,$5A
	.dc.b	$E2,$FC,$7C,$E6,$31,$5C,$2B,$6F,$F8,$4A,$FA,$6B,$E9,$3F,$7F,$3A
	.dc.b	$70,$23,$7A,$7F,$00,$C3,$B5,$F3,$C7,$C3,$87,$1A,$FE,$03,$6F,$43
	.dc.b	$F4,$70,$F8,$5E,$78,$7F,$95,$AB,$4F,$F8,$C2,$75,$87,$C2,$BE,$37
	.dc.b	$C3,$2F,$F3,$46,$1C,$1F,$55,$F0,$C3,$DD,$A7,$CA,$F2,$FE,$5B,$63
	.dc.b	$24,$AF,$79,$FE,$73,$F8,$5F,$F3,$A7,$C1,$BD,$1F,$35,$FE,$E2,$B4
	.dc.b	$27,$4F,$92,$3F,$92,$C6,$23,$FC,$E7,$5D,$2B,$9D,$D9,$FF,$07,$4D
	.dc.b	$7E,$85,$F2,$AF,$E3,$33,$9C,$BF,$8E,$61,$D6,$FC,$6F,$78,$CD,$70
	.dc.b	$9F,$33,$E0,$69,$E0,$29,$E0,$C3,$C4,$F4,$8E,$15,$DE,$17,$E3,$C3
	.dc.b	$CD,$1C,$E7,$C5,$18,$8F,$82,$76,$C3,$CA,$EE,$FA,$9E,$7A,$E3,$87
	.dc.b	$8A,$7B,$9C,$6F,$A3,$F4,$8F,$E7,$A6,$8B,$86,$1D,$1F,$EA,$4F,$80
	.dc.b	$7C,$23,$E1,$78,$C6,$1B,$19,$9F,$E9,$33,$96,$FC,$2F,$F1,$96,$ED
	.dc.b	$CA,$31,$1F,$E4,$99,$46,$23,$C4,$F1,$CF,$F0,$DF,$0A,$F8,$5F,$F5
	.dc.b	$CC,$61,$BB,$70,$C7,$F0,$D8,$F4,$7F,$11,$7C,$71,$B7,$C1,$FF,$67
	.dc.b	$56,$DF,$BE,$DC,$3B,$47,$03,$AE,$3E,$5C,$7C,$11,$F8,$6C,$3B,$37
	.dc.b	$0E,$CB,$0E,$95,$C3,$0E,$B3,$B3,$AE,$31,$C1,$23,$F5,$15,$CA,$3D
	.dc.b	$38,$F8,$DF,$F7,$46,$3E,$2A,$EA,$70,$9D,$4A,$7C,$F7,$FE,$14,$6E
	.dc.b	$DF,$EE,$66,$CF,$69,$F8,$4F,$F2,$28,$CB,$CF,$2F,$F2,$25,$7B,$72
	.dc.b	$FF,$77,$FC,$2F,$F4,$BA,$FF,$5F,$B5,$FE,$56,$DC,$3F,$C4,$BF,$02
	.dc.b	$E2,$E7,$41,$5B,$C3,$AF,$5E,$A7,$2D,$BE,$51,$E0,$F8,$4D,$59,$A6
	.dc.b	$FE,$16,$CE,$C8,$7F,$40,$A4,$6A,$3D,$33,$72,$F4,$CF,$8F,$E1,$7F
	.dc.b	$9F,$57,$FB,$79,$AF,$F4,$A6,$CF,$C0,$F1,$78,$74,$CA,$E0,$FA,$E3
	.dc.b	$11,$C6,$2C,$F1,$8F,$F7,$5F,$C2,$BE,$07,$E2,$36,$33,$1F,$17,$B3
	.dc.b	$2C,$8C,$9F,$EF,$4D,$3F,$C5,$CE,$0F,$EC,$35,$CD,$B3,$CD,$FF,$7D
	.dc.b	$FC,$1B,$E1,$7E,$A3,$63,$31,$F7,$3C,$87,$C2,$7F,$AC,$2B,$55,$FE
	.dc.b	$54,$6B,$FE,$CF,$6B,$F8,$30,$E1,$FC,$33,$CB,$3F,$F3,$9D,$76,$F8
	.dc.b	$5F,$05,$9F,$2E,$1F,$89,$F1,$1F,$0B,$49,$97,$E4,$71,$E7,$87,$E8
	.dc.b	$2B,$73,$FC,$B9,$AF,$F9,$B1,$AF,$FB,$9D,$AF,$A5,$DC,$F4,$CF,$CA
	.dc.b	$78,$27,$FC,$1F,$F0,$2F,$80,$FC,$F3,$63,$C9,$D1,$F0,$E3,$11,$FE
	.dc.b	$19,$D5,$F8,$87,$FA,$43,$1F,$F4,$8D,$7F,$BD,$AB,$F5,$8D,$4D,$26
	.dc.b	$1F,$AD,$F8,$6F,$CB,$E3,$F4,$5B,$5C,$A7,$F6,$5F,$0C,$F4,$29,$27
	.dc.b	$CE,$7E,$EF,$5F,$F3,$7B,$81,$3C,$5F,$EB,$2D,$7F,$D2,$ED,$7F,$04
	.dc.b	$70,$9F,$EB,$91,$BC,$BF,$D5,$EF,$F8,$3D,$BD,$DE,$3A,$FF,$1D,$FC
	.dc.b	$2B,$E1,$BF,$85,$36,$3C,$8D,$7E,$67,$C5,$7F,$11,$7F,$8A,$46,$5F
	.dc.b	$1B,$5F,$CB,$35,$DC,$FF,$71,$8E,$97,$F5,$3C,$1F,$BA,$78,$A6,$57
	.dc.b	$C5,$DF,$F8,$DB,$F8,$49,$CB,$EA,$9E,$2E,$E3,$35,$FE,$D0,$BE,$FB
	.dc.b	$FE,$1E,$77,$5C,$4F,$7D,$E2,$FC,$91,$82,$1E,$13,$B1,$87,$B4,$71
	.dc.b	$AE,$D3,$97,$E1,$C6,$63,$F2,$35,$E6,$9C,$C7,$43,$B1,$E5,$25,$79
	.dc.b	$1C,$BE,$07,$B1,$D0,$E4,$E5,$E3,$1E,$47,$8B,$E0,$72,$71,$32,$7D
	.dc.b	$83,$9B,$C6,$3C,$47,$52,$B9,$24,$71,$39,$9F,$AE,$9F,$84,$AF,$EA
	.dc.b	$BE,$1A,$E6,$DA,$F3,$46,$23,$91,$E4,$3B,$39,$C3,$F0,$EE,$EB,$AC
	.dc.b	$F0,$3C,$73,$AA,$EA,$7A,$23,$55,$C7,$0F,$3D,$6E,$7C,$87,$82,$B8
	.dc.b	$46,$63,$8C,$74,$76,$60,$B7,$E4,$9E,$04,$C3,$F0,$93,$C2,$30,$FD
	.dc.b	$E3,$91,$1F,$04,$F3,$9F,$39,$CA,$74,$FD,$08,$D4,$7C,$93,$9B,$87
	.dc.b	$E2,$C7,$19,$F0,$C7,$DA,$9D,$61,$F3,$EB,$47,$82,$7C,$D5,$C2,$7E
	.dc.b	$54,$77,$8C,$5F,$C5,$8F,$29,$E3,$BE,$B1,$F3,$2F,$AC,$F1,$6B,$E6
	.dc.b	$3B,$BE,$17,$F3,$67,$43,$D1,$86,$70,$F1,$C7,$13,$27,$53,$C5,$1B
	.dc.b	$AF,$C2,$5F,$92,$7E,$A5,$F8,$AF,$E1,$47,$88,$E4,$78,$B0,$F7,$BF
	.dc.b	$80,$C3,$F8,$8B,$FC,$EA,$3F,$8E,$B6,$BF,$D2,$D1,$88,$FD,$C1,$93
	.dc.b	$FD,$3A,$E5,$FD,$C1,$93,$FD,$EF,$F0,$EF,$67,$FC,$8C,$D8,$F2,$E0
	.dc.b	$CB,$FC,$19,$39,$9E,$35,$F0,$5F,$F2,$94,$66,$B8,$46,$27,$4A,$75
	.dc.b	$5D,$27,$31,$89,$E0,$78,$A7,$4F,$F6,$38,$D3,$F6,$6B,$BB,$D6,$BD
	.dc.b	$67,$42,$BC,$CE,$A7,$A8,$C1,$C4,$EC,$F1,$69,$A9,$F0,$9E,$2B,$FB
	.dc.b	$31,$C4,$E8,$7E,$26,$BC,$B3,$C2,$BA,$4E,$A7,$F0,$AD,$BF,$7B,$C5
	.dc.b	$B1,$CA,$78,$B6,$7C,$91,$EE,$78,$1E,$E8,$D3,$83,$CC,$E0,$F0,$96
	.dc.b	$7B,$9F,$AD,$8F,$82,$79,$4F,$CE,$98,$8E,$0E,$0D,$1D,$A3,$93,$8A
	.dc.b	$F9,$18,$7F,$3E,$7E,$24,$7A,$8E,$93,$C2,$BA,$C7,$33,$B3,$89,$E7
	.dc.b	$1C,$0F,$81,$3E,$A7,$F9,$6C,$76,$7C,$B8,$7F,$33,$8C,$57,$68,$C6
	.dc.b	$1F,$2E,$31,$85,$C3,$A6,$1E,$33,$F9,$7E,$1C,$A3,$38,$72,$9F,$F1
	.dc.b	$99,$F9,$03,$FC,$6B,$3F,$AB,$6E,$1A,$7D,$27,$FA,$F7,$E1,$BF,$CB
	.dc.b	$38,$FE,$79,$B8,$78,$AF,$F4,$71,$8A,$E1,$C3,$E7,$79,$88,$D3,$63
	.dc.b	$4F,$AE,$32,$6A,$7D,$91,$D0,$FF,$2A,$46,$70,$E2,$F0,$AA,$7F,$B4
	.dc.b	$5C,$4F,$D5,$3C,$07,$28,$D1,$C1,$C1,$0F,$C5,$C7,$25,$87,$C3,$9E
	.dc.b	$11,$8C,$3F,$D4,$55,$BA,$FC,$4B,$63,$35,$C2,$79,$57,$77,$71,$F4
	.dc.b	$E3,$CE,$7D,$AA,$FF,$42,$12,$3F,$DD,$DF,$0C,$F8,$7A,$7B,$2B,$F8
	.dc.b	$C7,$99,$CF,$33,$49,$1F,$81,$F8,$57,$C3,$78,$5B,$19,$8C,$19,$F8
	.dc.b	$69,$1C,$5A,$FE,$43,$E3,$7F,$D1,$6D,$F7,$7D,$91,$88,$FF,$43,$B5
	.dc.b	$FF,$65,$7C,$23,$E1,$5E,$DF,$E5,$63,$0D,$F8,$57,$B9,$9A,$DD,$FE
	.dc.b	$83,$E1,$1F,$0B,$FC,$DA,$3F,$DA,$D1,$8E,$6F,$D0,$36,$38,$4E,$D0
	.dc.b	$F8,$47,$B9,$FD,$3B,$A3,$FD,$B7,$DD,$E7,$C7,$FD,$51,$5A,$7F,$D5
	.dc.b	$BE,$2A,$75,$87,$E1,$CE,$B5,$B7,$F6,$ED,$7F,$D5,$6D,$7F,$DD,$CD
	.dc.b	$78,$E1,$E3,$BC,$B6,$1F,$09,$3D,$1C,$47,$AE,$6F,$A0,$F8,$4E,$19
	.dc.b	$B9,$53,$A9,$CC,$63,$2F,$F6,$34,$F1,$7F,$85,$37,$0F,$E2,$2D,$7F
	.dc.b	$D8,$AD,$7E,$69,$DA,$BB,$4F,$C8,$AF,$F6,$C3,$DA,$3D,$3F,$08,$F8
	.dc.b	$4F,$74,$BE,$13,$0F,$FB,$66,$6E,$1F,$3E,$35,$5C,$8E,$35,$AE,$0F
	.dc.b	$EF,$77,$6B,$F0,$27,$03,$A5,$70,$39,$3F,$EE,$A3,$57,$F7,$4F,$CC
	.dc.b	$FB,$5D,$B5,$FA,$3C,$E8,$C5,$7F,$BB,$E3,$0F,$3C,$BF,$09,$97,$F8
	.dc.b	$3A,$73,$87,$AE,$73,$5F,$2A,$73,$5F,$E2,$28,$F8,$98,$78,$30,$EF
	.dc.b	$87,$97,$E0,$BE,$D3,$F9,$3E,$77,$C2,$3C,$DF,$C3,$D6,$B2,$FF,$16
	.dc.b	$D6,$A7,$3C,$3F,$8F,$8C,$67,$FA,$8A,$FF,$12,$DF,$18,$FD,$5C,$EB
	.dc.b	$1F,$15,$F1,$C7,$C1,$5C,$4F,$C2,$DF,$1A,$FD,$FC,$E9,$F5,$7B,$1F
	.dc.b	$9F,$C5,$F0,$9B,$95,$AB,$53,$98,$C6,$1D,$EB,$F9,$EC,$E6,$B7,$8F
	.dc.b	$C2,$D7,$E1,$F9,$57,$0E,$18,$76,$8C,$3F,$36,$7F,$5E,$FF,$0A,$F8
	.dc.b	$4F,$E1,$7F,$04,$F3,$FD,$B5,$AA,$56,$A7,$38,$59,$FD,$39,$A1,$7C
	.dc.b	$DF,$E2,$FA,$7F,$0B,$8C,$47,$EB,$AF,$93,$F1,$30,$FF,$33,$CE,$A3
	.dc.b	$CC,$74,$C3,$C2,$D3,$C3,$5C,$AF,$B4,$FF,$83,$A3,$17,$D2,$31,$5C
	.dc.b	$4F,$06,$1E,$27,$BE,$1E,$6C,$39,$C7,$63,$83,$87,$D4,$62,$FB,$3B
	.dc.b	$9E,$51,$88,$DC,$62,$BC,$11,$F2,$63,$0F,$94,$F1,$DF,$C0,$8F,$89
	.dc.b	$1F,$3C,$FF,$13,$61,$FD,$D5,$C1,$EE,$3E,$AF,$C1,$FF,$6E,$F8,$4F
	.dc.b	$F1,$63,$63,$FC,$66,$D7,$7E,$42,$77,$F0,$9F,$E0,$E9,$DB,$5E,$AD
	.dc.b	$7F,$94,$D2,$31,$1F,$D6,$23,$1C,$3F,$80,$F8,$47,$C2,$7F,$30,$CF
	.dc.b	$F0,$B1,$8B,$FC,$2D,$F0,$7E,$C3,$67,$75,$C7,$4F,$E7,$37,$C6,$F7
	.dc.b	$8F,$A6,$FF,$23,$87,$1B,$DE,$1C,$6F,$FB,$6E,$1C,$6F,$78,$71,$BF
	.dc.b	$6B,$6B,$77,$C5,$5F,$1A,$DE,$3E,$AB,$E1,$5A,$BE,$17,$B8,$F8,$D3
	.dc.b	$AC,$7C,$D1,$D8,$F9,$15,$DE,$33,$06,$D7,$38,$EC,$78,$6F,$8B,$FC
	.dc.b	$1A,$3E,$34,$7F,$B2,$1B,$AF,$A2,$7F,$4D,$18,$7C,$47,$F4,$B7,$0F
	.dc.b	$87,$0F,$6F,$C2,$7F,$4B,$8F,$E2,$6D,$7F,$A9,$36,$FF,$DA,$FF,$02
	.dc.b	$E2,$FE,$E3,$1F,$ED,$A9,$D6,$3F,$DD,$1B,$7F,$E2,$26,$AB,$E4,$F8
	.dc.b	$3E,$13,$9D,$71,$6C,$71,$6B,$C7,$E1,$2D,$A6,$92,$7F,$96,$DA,$B1
	.dc.b	$1E,$C9,$B8,$7B,$27,$AF,$C2,$7F,$3F,$AF,$E6,$AD,$7F,$BA,$36,$7E
	.dc.b	$1F,$89,$C3,$A2,$57,$07,$CF,$8A,$73,$3C,$A6,$D7,$28,$FE,$87,$F0
	.dc.b	$8F,$7F,$DE,$D8,$CC,$74,$EA,$CB,$23,$27,$C0,$8C,$3F,$E0,$B9,$E4
	.dc.b	$D9,$E4,$FF,$52,$F8,$47,$C1,$FC,$E6,$C6,$63,$E9,$F9,$0F,$82,$FF
	.dc.b	$44,$4E,$48,$7F,$86,$6B,$8B,$FE,$57,$9E,$1E,$BF,$5B,$FF,$4B,$56
	.dc.b	$AB,$FA,$AB,$5F,$F1,$EB,$5F,$F3,$5F,$86,$F8,$3F,$C1,$B6,$3F,$B5
	.dc.b	$B5,$F0,$63,$FA,$08,$C6,$AE,$5F,$F2,$4C,$61,$B7,$EE,$AE,$0F,$FA
	.dc.b	$AE,$7A,$5F,$F6,$A6,$BC,$0F,$F7,$17,$BB,$FE,$D6,$27,$B7,$F2,$3E
	.dc.b	$0F,$C1,$3F,$E0,$09,$CB,$67,$58,$FE,$7A,$51,$FE,$BB,$6C,$F6,$8C
	.dc.b	$1E,$5F,$3A,$47,$FB,$CF,$0F,$F5,$23,$F2,$23,$FD,$89,$F0,$2F,$82
	.dc.b	$FD,$5B,$67,$FC,$26,$D7,$FC,$A5,$5F,$EB,$2D,$7D,$30,$9D,$3F,$E0
	.dc.b	$C6,$BF,$E7,$46,$BF,$EA,$26,$F0,$FF,$A6,$FC,$BF,$DD,$FC,$13,$E0
	.dc.b	$A9,$F8,$EF,$57,$F8,$4D,$6F,$0F,$F1,$3D,$6D,$B5,$BB,$E3,$18,$21
	.dc.b	$F9,$1B,$FE,$23,$5C,$63,$FC,$DC,$DC,$7D,$AD,$DF,$FD,$73,$8F,$86
	.dc.b	$B7,$8F,$66,$9F,$EB,$6C,$3A,$C6,$70,$E6,$F9,$67,$C7,$18,$6D,$F6
	.dc.b	$AE,$31,$83,$FD,$93,$87,$8D,$61,$E3,$BE,$51,$E2,$23,$FE,$CC,$C7
	.dc.b	$C5,$87,$FA,$9A,$F9,$51,$E9,$7C,$E7,$58,$F0,$AF,$0B,$E9,$7A,$C6
	.dc.b	$9F,$34,$F1,$76,$7A,$4E,$D3,$C7,$0F,$39,$2B,$94,$FD,$5C,$3C,$0F
	.dc.b	$03,$AE,$3F,$11,$FF,$2C,$99,$2E,$3F,$6D,$FF,$2E,$B5,$EB,$7F,$98
	.dc.b	$DF,$FA,$7C,$7F,$AB,$5B,$1E,$38,$F4,$99,$F0,$3F,$64,$79,$2B,$E4
	.dc.b	$D7,$0C,$7F,$DA,$2D,$AE,$37,$CD,$46,$5B,$7C,$F2,$EB,$1F,$C0,$E3
	.dc.b	$0D,$21,$F4,$0E,$B7,$E0,$8E,$31,$93,$E2,$5F,$BE,$32,$F8,$F0,$F4
	.dc.b	$C6,$2B,$E2,$92,$87,$98,$F0,$D7,$1B,$E1,$1A,$C3,$D9,$87,$E3,$0E
	.dc.b	$B1,$C2,$FA,$3E,$47,$D0,$E4,$F5,$90,$F0,$1F,$44,$FC,$54,$74,$70
	.dc.b	$7B,$DC,$5F,$BE,$37,$7E,$67,$9C,$7A,$5F,$01,$8A,$F2,$D7,$27,$CB
	.dc.b	$1D,$6B,$95,$FB,$9F,$CA,$BF,$6C,$C4,$7E,$7D,$C6,$1F,$6A,$3C,$F8
	.dc.b	$79,$5E,$87,$8A,$78,$BF,$65,$E2,$FA,$8D,$4F,$39,$E4,$4B,$F4,$DF
	.dc.b	$9A,$75,$1E,$A9,$D5,$7C,$DC,$3D,$CF,$D9,$AD,$B4,$F8,$B1,$F6,$EF
	.dc.b	$A4,$78,$CF,$5D,$F5,$3C,$2F,$98,$D5,$F8,$1F,$9C,$78,$67,$9B,$EF
	.dc.b	$BB,$3F,$A7,$C3,$E0,$CF,$63,$F0,$8E,$23,$A6,$1E,$A9,$EA,$47,$13
	.dc.b	$BC,$3F,$22,$D9,$F7,$1E,$13,$D0,$53,$C0,$73,$C3,$CB,$3E,$F8,$F1
	.dc.b	$E1,$86,$9C,$B0,$F5,$46,$63,$F4,$07,$82,$3E,$25,$FE,$01,$B3,$F9
	.dc.b	$7C,$3D,$2D,$C3,$95,$6F,$0F,$16,$1F,$02,$30,$F5,$8C,$C6,$23,$55
	.dc.b	$CD,$F3,$4F,$2B,$EF,$1F,$80,$BE,$F5,$F9,$E3,$D5,$3A,$9F,$8A,$DC
	.dc.b	$3E,$9D,$70,$9F,$D1,$57,$08,$C4,$7C,$D9,$F7,$BA,$3C,$67,$CA,$3E
	.dc.b	$8E,$1E,$93,$EC,$3F,$4C,$D9,$4D,$D7,$B5,$A7,$AC,$DC,$73,$8F,$24
	.dc.b	$61,$F7,$3F,$E3,$87,$FC,$5C,$47,$07,$F2,$49,$FF,$10,$FB,$EF,$82
	.dc.b	$FF,$49,$B6,$7C,$AD,$53,$C2,$7F,$09,$39,$8C,$4F,$09,$E6,$68,$AE
	.dc.b	$4F,$16,$5F,$CD,$95,$73,$8F,$C0,$DF,$09,$F0,$C6,$E7,$51,$F8,$73
	.dc.b	$F9,$FB,$87,$9D,$71,$BF,$DD,$4E,$49,$1F,$CB,$5B,$3E,$43,$9C,$64
	.dc.b	$E5,$5E,$17,$11,$B3,$B9,$67,$C5,$87,$F6,$77,$AC,$7A,$8E,$A7,$F3
	.dc.b	$A7,$E7,$3F,$0C,$E6,$7F,$81,$9C,$BE,$13,$FC,$12,$6A,$FF,$CB,$7F
	.dc.b	$05,$FE,$6C,$8E,$C1,$C4,$7D,$68,$CC,$3D,$9E,$71,$B6,$C6,$63,$BB
	.dc.b	$5F,$27,$99,$C2,$CD,$24,$79,$BE,$05,$F0,$3D,$9B,$19,$8C,$1E,$2F
	.dc.b	$79,$F1,$D4,$93,$F0,$23,$2F,$C2,$F8,$1F,$F4,$2A,$69,$3A,$7D,$BF
	.dc.b	$02,$F8,$17,$B5,$F2,$A3,$0D,$C1,$ED,$66,$B7,$8F,$F8,$E3,$E0,$1E
	.dc.b	$C7,$FA,$8B,$8F,$FD,$47,$D8,$EB,$ED,$D1,$F8,$5B,$DB,$FE,$6D,$C7
	.dc.b	$83,$FA,$18,$CE,$3F,$E7,$D9,$CB,$FE,$1A,$8C,$E1,$DA,$FA,$CF,$DE
	.dc.b	$9E,$2F,$E2,$5A,$F5,$C7,$FD,$3E,$DC,$3F,$13,$E4,$BD,$FE,$F9,$B8
	.dc.b	$13,$E2,$BB,$5A,$9C,$DF,$E8,$6B,$93,$FA,$9B,$E5,$1C,$4C,$DF,$FB
	.dc.b	$A6,$B8,$BF,$96,$8C,$BF,$A1,$6B,$FC,$15,$AF,$F8,$F1,$AB,$DF,$7B
	.dc.b	$FF,$23,$07,$BF,$87,$F6,$D3,$6B,$F6,$51,$AA,$F7,$4E,$76,$FF,$6E
	.dc.b	$CE,$4E,$33,$D6,$F5,$1F,$C6,$6A,$DF,$FB,$CE,$6B,$E5,$AE,$07,$CB
	.dc.b	$E2,$78,$3C,$F6,$C3,$71,$FE,$19,$3A,$7F,$8A,$1E,$C9,$DB,$8C,$3F
	.dc.b	$0A,$FF,$2B,$9E,$18,$7B,$A7,$2F,$F9,$92,$33,$97,$BA,$73,$2F,$77
	.dc.b	$EA,$CF,$FB,$FB,$8D,$EF,$BC,$5F,$C2,$56,$B2,$FF,$37,$D6,$A7,$3B
	.dc.b	$FE,$6A,$31,$3F,$C9,$EB,$86,$3F,$AE,$C3,$A4,$E6,$F8,$46,$27,$F8
	.dc.b	$DD,$70,$BE,$71,$8F,$43,$F4,$F9,$3D,$FE,$B3,$EB,$9C,$C6,$2D,$18
	.dc.b	$C7,$D6,$DC,$BC,$DB,$7B,$BC,$4F,$7B,$63,$24,$C7,$F9,$79,$FE,$90
	.dc.b	$F7,$FF,$D2,$3E,$F3,$DE,$FE,$BD,$9F,$CE,$AD,$09,$D3,$FE,$98,$8F
	.dc.b	$E0,$71,$88,$FE,$CD,$5D,$2B,$9E,$0C,$31,$5F,$D2,$6A,$BF,$6E,$74
	.dc.b	$F7,$AF,$F6,$8C,$EB,$0E,$F1,$89,$D1,$E3,$AE,$38,$78,$E3,$D2,$F9
	.dc.b	$6B,$8C,$78,$4E,$27,$87,$0F,$1B,$C2,$FD,$31,$88,$61,$E3,$C3,$BD
	.dc.b	$78,$F0,$ED,$18,$AE,$D1,$8B,$E5,$7C,$5B,$3E,$2C,$3C,$AF,$CA,$31
	.dc.b	$7C,$DC,$1F,$DB,$2B,$FB,$1B,$E4,$3F,$B2,$BF,$DC,$3D,$DF,$EB,$FE
	.dc.b	$FD,$F1,$0D,$25,$7F,$BF,$A5,$EE,$FF,$8A,$A3,$0D,$7F,$C0,$11,$AE
	.dc.b	$1E,$D3,$99,$FF,$13,$4E,$5F,$9B,$EF,$3D,$FF,$F1,$56,$DF,$87,$8C
	.dc.b	$4F,$E0,$AB,$84,$FE,$2D,$BB,$FF,$8C,$6B,$75,$D2,$F9,$CF,$83,$0F
	.dc.b	$46,$1C,$5B,$1C,$67,$75,$FE,$F0,$BE,$77,$BC,$7C,$58,$70,$AD,$5F
	.dc.b	$05,$7C,$30,$F0,$5F,$03,$F3,$71,$99,$E2,$64,$FC,$7E,$3E,$C7,$27
	.dc.b	$AC,$D6,$3E,$37,$04,$70,$7C,$9C,$3F,$CA,$97,$FA,$89,$F2,$C6,$37
	.dc.b	$F8,$11,$FE,$41,$BF,$03,$A3,$FC,$89,$87,$CC,$BE,$4D,$AF,$6C,$FE
	.dc.b	$9B,$DF,$FE,$86,$FF,$92,$DA,$FF,$8E,$9B,$5E,$3F,$71,$C5,$E3,$A0
	.dc.b	$AD,$DF,$F9,$B5,$B1,$A6,$AA,$E6,$F8,$7D,$FE,$8F,$36,$C9,$CC,$BD
	.dc.b	$FC,$C4,$AE,$47,$F9,$96,$74,$E0,$F6,$46,$DF,$F2,$BC,$69,$FF,$2C
	.dc.b	$FE,$E8,$F8,$BF,$F2,$AB,$63,$FC,$C2,$D7,$FC,$D0,$1C,$3F,$E6,$B7
	.dc.b	$59,$FC,$9F,$80,$F4,$9E,$E8,$CE,$7F,$E9,$36,$D7,$CC,$C1,$5C,$23
	.dc.b	$FD,$01,$3A,$7C,$1D,$1E,$62,$3E,$66,$BF,$E8,$1F,$11,$ED,$FF,$9A
	.dc.b	$23,$FC,$50,$D7,$E0,$CF,$94,$FB,$D3,$C3,$9E,$D6,$F3,$FC,$B6,$1F
	.dc.b	$D7,$30,$F2,$ED,$F2,$08,$FC,$86,$BE,$3A,$FE,$2A,$F8,$B9,$BE,$24
	.dc.b	$C8,$C1,$0F,$AA,$FE,$1B,$EC,$9E,$B7,$FA,$29,$B3,$FE,$96,$6B,$F5
	.dc.b	$9A,$FF,$A7,$DC,$E3,$FA,$FE,$67,$CF,$7E,$49,$B5,$FE,$AF,$AD,$F0
	.dc.b	$FF,$B2,$5B,$1F,$A2,$9D,$BF,$EA,$26,$DF,$EE,$E7,$73,$FE,$1B,$F7
	.dc.b	$7F,$D4,$3E,$19,$A4,$8F,$F5,$7C,$65,$FA,$FF,$03,$F4,$8F,$D3,$C6
	.dc.b	$63,$FD,$62,$DB,$FF,$42,$EC,$AE,$15,$FE,$C9,$AD,$65,$FE,$E3,$F6
	.dc.b	$FF,$1F,$3F,$EB,$E6,$BF,$9E,$6B,$FE,$C7,$6B,$C7,$93,$F0,$BE,$3C
	.dc.b	$8F,$F6,$B4,$F2,$CB,$F8,$E3,$6F,$F8,$4F,$27,$E5,$2F,$85,$FF,$B2
	.dc.b	$AB,$4F,$F0,$EF,$89,$FE,$04,$FF,$B8,$5A,$FF,$B4,$1B,$87,$F2,$9F
	.dc.b	$73,$FD,$53,$39,$EA,$FE,$1B,$5A,$CB,$F8,$CC,$A3,$0F,$F1,$CA,$E3
	.dc.b	$7F,$CC,$E7,$33,$BB,$E6,$7F,$18,$AD,$DF,$4A,$D9,$27,$58,$F0,$BF
	.dc.b	$54,$57,$FD,$D3,$7C,$9F,$81,$19,$C3,$C4,$F2,$8E,$24,$3D,$C7,$F3
	.dc.b	$48,$F8,$55,$C7,$0F,$25,$5A,$F7,$4D,$7D,$87,$3A,$E2,$DA,$E2,$D3
	.dc.b	$89,$4E,$27,$88,$F5,$CF,$5C,$3E,$14,$79,$67,$E8,$E1,$EF,$8E,$37
	.dc.b	$EE,$7A,$46,$6B,$8C,$75,$9F,$2B,$6B,$CA,$D3,$CA,$78,$CF,$BE,$79
	.dc.b	$8B,$3F,$8A,$6E,$1F,$CF,$1F,$A2,$79,$CF,$C2,$7D,$B7,$DF,$3D,$91
	.dc.b	$63,$43,$04,$EA,$33,$18,$6C,$65,$FC,$0B,$CD,$F5,$1F,$12,$FE,$3C
	.dc.b	$67,$0F,$8D,$3A,$AE,$8F,$23,$EA,$61,$F9,$88,$CC,$FE,$96,$BC,$0D
	.dc.b	$AF,$19,$C3,$0E,$D8,$78,$23,$06,$8E,$C7,$9B,$0E,$D7,$C2,$3E,$49
	.dc.b	$E5,$70,$FA,$CC,$DF,$85,$E6,$FB,$09,$5E,$18,$CB,$F7,$A3,$31,$F5
	.dc.b	$0F,$23,$87,$B1,$F4,$E7,$C3,$5F,$8A,$9F,$01,$2B,$A1,$2B,$A1,$E0
	.dc.b	$3C,$EE,$5E,$87,$81,$C4,$FC,$28,$C1,$C0,$87,$03,$A1,$E6,$78,$38
	.dc.b	$7C,$D1,$F0,$67,$CD,$3C,$A7,$71,$88,$E5,$7F,$3B,$0F,$BF,$7E,$48
	.dc.b	$C6,$1F,$80,$8E,$84,$AE,$93,$A7,$EF,$9E,$6A,$E6,$E8,$F6,$9E,$BB
	.dc.b	$FA,$18,$79,$2F,$A4,$F9,$0D,$D7,$27,$07,$5C,$3C,$8F,$D0,$BF,$65
	.dc.b	$FC,$28,$C5,$F4,$8C,$47,$86,$31,$87,$F1,$BC,$3B,$1D,$EB,$6E,$30
	.dc.b	$EF,$7C,$4E,$A7,$B4,$F2,$1F,$0B,$0F,$9D,$3C,$EF,$B5,$7E,$31,$B5
	.dc.b	$F1,$2F,$BC,$E6,$3F,$8D,$B5,$FD,$21,$B9,$E2,$FE,$05,$B7,$F0,$9B
	.dc.b	$87,$BA,$79,$9C,$4A,$71,$2C,$F1,$6C,$71,$7C,$26,$CB,$1B,$6C,$6E
	.dc.b	$3F,$59,$87,$C9,$9F,$89,$3F,$7A,$3E,$24,$76,$3E,$59,$F7,$A7,$E4
	.dc.b	$DF,$D0,$9E,$53,$FC,$26,$FF,$20,$D3,$F0,$F3,$B3,$E4,$57,$89,$B7
	.dc.b	$FB,$3B,$E7,$3F,$92,$BE,$6D,$BE,$67,$A0,$F2,$CF,$A1,$F7,$9E,$8C
	.dc.b	$3E,$CB,$F3,$4C,$9F,$87,$3D,$97,$F0,$70,$F3,$46,$EF,$CC,$D9,$F3
	.dc.b	$4F,$A2,$3F,$81,$DF,$A2,$7F,$6D,$1E,$63,$E1,$1F,$92,$2D,$7C,$A7
	.dc.b	$D4,$78,$E7,$D5,$1E,$08,$FB,$8D,$3E,$E1,$67,$EE,$1E,$48,$F3,$1E
	.dc.b	$4A,$F3,$3F,$CA,$AB,$CC,$FE,$96,$BC,$4F,$92,$BE,$13,$E4,$9D,$BF
	.dc.b	$CC,$BE,$31,$F6,$DF,$DB,$61,$18,$7F,$B8,$35,$FF,$56,$B5,$FE,$F2
	.dc.b	$E6,$D7,$FE,$4A,$F4,$DF,$55,$FE,$98,$6D,$FF,$74,$BE,$15,$FE,$40
	.dc.b	$6C,$37,$57,$08,$FF,$54,$37,$0F,$F7,$A5,$70,$97,$E1,$AF,$86,$F6
	.dc.b	$36,$3F,$C1,$4D,$7F,$D8,$ED,$7F,$C2,$6E,$6B,$D7,$1E,$3B,$7D,$77
	.dc.b	$AC,$FF,$72,$D6,$F2,$FF,$2C,$B6,$B8,$4E,$DF,$F0,$9C,$65,$7B,$4F
	.dc.b	$67,$FC,$CB,$1F,$E6,$28,$C7,$1F,$F8,$85,$7B,$4F,$4B,$17,$C2,$FF
	.dc.b	$E0,$0F,$C6,$3E,$0B,$D3,$1F,$E2,$E8,$C6,$9F,$E2,$B6,$C7,$FB,$A1
	.dc.b	$AF,$F8,$D3,$E4,$1F,$0B,$CA,$3F,$CE,$B1,$8A,$FF,$1A,$36,$3F,$DE
	.dc.b	$2D,$7F,$C7,$BF,$1A,$F5,$B9,$40,$46,$6B,$FC,$7C,$D8,$FF,$7D,$B5
	.dc.b	$E7,$3F,$53,$8B,$DB,$3B,$7E,$BD,$3D,$97,$C3,$EE,$3F,$D1,$B5,$B6
	.dc.b	$DE,$70,$F2,$4E,$5F,$CD,$DF,$2A,$FF,$25,$CE,$65,$F8,$A3,$E9,$3D
	.dc.b	$B1,$86,$C6,$6B,$FC,$B4,$D9,$FF,$30,$9B,$8F,$F2,$BC,$67,$2F,$C8
	.dc.b	$F4,$BF,$11,$7E,$13,$E7,$C3,$80,$AF,$BF,$7C,$70,$FF,$55,$B6,$7F
	.dc.b	$59,$7C,$A7,$BE,$3C,$FF,$15,$FF,$77,$C7,$F9,$BD,$B5,$E6,$6B,$FE
	.dc.b	$78,$6B,$FE,$B8,$77,$87,$E4,$1B,$3F,$EC,$CB,$FF,$70,$3F,$E7,$48
	.dc.b	$CD,$7E,$0D,$A9,$CE,$5F,$07,$C6,$F3,$79,$7F,$07,$DD,$FA,$DC,$EF
	.dc.b	$7D,$E4,$E7,$8F,$EE,$4E,$63,$1A,$7F,$A2,$63,$0F,$FA,$1A,$F9,$DF
	.dc.b	$28,$C5,$F2,$8C,$69,$F3,$E7,$FD,$AF,$5C,$30,$F2,$7A,$7F,$8B,$DD
	.dc.b	$EF,$FE,$1E,$93,$98,$C3,$72,$E3,$87,$FA,$76,$F8,$C7,$BF,$DF,$F8
	.dc.b	$1E,$CD,$AF,$43,$73,$F2,$1F,$AD,$C9,$5F,$87,$8F,$D8,$7B,$FF,$B1
	.dc.b	$F7,$5E,$F7,$AA,$B5,$4A,$D4,$E7,$1B,$3F,$EB,$89,$AB,$0E,$F3,$88
	.dc.b	$EE,$30,$EE,$F9,$70,$FE,$5B,$18,$8F,$F8,$03,$7D,$2F,$9C,$71,$C3
	.dc.b	$F4,$F3,$9B,$E3,$18,$9D,$1D,$70,$F1,$3E,$B3,$27,$9E,$78,$BE,$B3
	.dc.b	$37,$E0,$C3,$C9,$1E,$A7,$CD,$19,$BF,$04,$62,$B9,$D7,$A8,$E6,$E1
	.dc.b	$E5,$3F,$2D,$F1,$98,$3D,$06,$0F,$44,$F5,$BF,$5E,$1F,$DD,$E7,$D8
	.dc.b	$7F,$B9,$5C,$1F,$EE,$7F,$79,$ED,$3D,$FC,$36,$BF,$88,$CE,$A7,$27
	.dc.b	$1F,$6B,$F8,$64,$E9,$AF,$FB,$E2,$51,$88,$FE,$78,$23,$3B,$F3,$3F
	.dc.b	$0F,$EF,$BD,$9D,$E9,$F8,$18,$C5,$FE,$32,$F8,$BF,$61,$B3,$BA,$E3
	.dc.b	$AF,$82,$F8,$DE,$F1,$F4,$DF,$E3,$70,$E3,$7B,$C3,$8D,$FF,$53,$C3
	.dc.b	$8D,$EF,$0E,$37,$EC,$C3,$8D,$EF,$0E,$2B,$0E,$35,$BB,$E3,$3F,$D7
	.dc.b	$EF,$8B,$BB,$E3,$7C,$23,$55,$C1,$F9,$4D,$8D,$47,$13,$E3,$46,$EB
	.dc.b	$AC,$E8,$DB,$8A,$FA,$B8,$FF,$81,$27,$E3,$B6,$33,$AF,$A6,$3E,$BC
	.dc.b	$63,$0F,$80,$7F,$79,$8E,$0D,$9E,$0F,$EB,$BD,$F7,$C0,$71,$84,$61
	.dc.b	$E0,$7F,$80,$DC,$3F,$C6,$87,$BE,$E3,$E5,$3A,$C3,$FC,$3A,$DC,$3F
	.dc.b	$22,$DB,$FF,$07,$5F,$17,$B9,$E2,$F5,$F9,$91,$CC,$3D,$EC,$47,$8E
	.dc.b	$95,$8C,$3F,$C5,$B7,$BA,$EB,$5A,$6D,$6F,$0F,$D2,$D6,$E7,$DD,$EF
	.dc.b	$BE,$03,$E3,$36,$3C,$ED,$7F,$9E,$F8,$0F,$77,$FC,$6B,$1A,$AF,$F3
	.dc.b	$D5,$6E,$3F,$66,$1C,$3F,$B6,$6B,$FE,$33,$7B,$67,$EC,$F7,$7E,$C6
	.dc.b	$27,$CB,$7E,$6E,$C9,$87,$99,$FC,$36,$5E,$5F,$C2,$73,$3F,$E1,$EC
	.dc.b	$BF,$B0,$BF,$E2,$03,$C9,$5E,$82,$3F,$84,$AF,$77,$4F,$4D,$3F,$8D
	.dc.b	$CF,$73,$E2,$F0,$7C,$82,$3C,$61,$EE,$62,$3C,$53,$F1,$88,$F8,$F1
	.dc.b	$FA,$9B,$7F,$8C,$2F,$F8,$EC,$F3,$E7,$E5,$3F,$E6,$CF,$80,$7C,$0F
	.dc.b	$E7,$5B,$5F,$8E,$6B,$F9,$BE,$77,$CD,$EE,$BF,$D6,$15,$AA,$FA,$ED
	.dc.b	$7F,$95,$B5,$FF,$34,$3D,$2F,$FC,$FB,$6F,$1B,$F5,$CD,$AF,$15,$FE
	.dc.b	$C7,$C4,$7C,$15,$24,$FC,$7A,$AF,$F9,$52,$76,$FD,$76,$B9,$9F,$F6
	.dc.b	$44,$63,$0F,$91,$E4,$7D,$73,$9B,$FE,$92,$F8,$2F,$C5,$5F,$EC,$9A
	.dc.b	$FE,$FD,$B1,$F4,$7D,$DF,$E1,$11,$A9,$FC,$DB,$5F,$E9,$ED,$7F,$D1
	.dc.b	$8D,$4F,$4E,$AF,$F4,$87,$2F,$EA,$7D,$AF,$C6,$1D,$BA,$BF,$8B,$7C
	.dc.b	$17,$EE,$70,$FF,$75,$1F,$E9,$7F,$82,$56,$BB,$A4,$FF,$0C,$AE,$1A
	.dc.b	$7F,$AC,$2F,$73,$9A,$DB,$FC,$76,$B7,$97,$F6,$8A,$DC,$F8,$0F,$F7
	.dc.b	$F6,$1C,$30,$FF,$5D,$DF,$09,$FF,$55,$35,$57,$29,$DE,$1C,$70,$FE
	.dc.b	$EB,$7B,$AD,$46,$27,$57,$D1,$FE,$71,$3A,$8C,$1E,$B2,$1F,$CD,$DD
	.dc.b	$2A,$F2,$12,$79,$E1,$DA,$75,$16,$B8,$D6,$E7,$AE,$1F,$2E,$F8,$4E
	.dc.b	$67,$C1,$18,$BE,$4D,$8E,$38,$7C,$D8,$C4,$7D,$19,$F1,$BA,$3D,$A6
	.dc.b	$4F,$E8,$F8,$79,$23,$8E,$1E,$39,$E0,$7A,$8B,$3B,$73,$87,$CF,$BE
	.dc.b	$38,$7C,$F9,$EB,$18,$7C,$37,$D5,$FC,$15,$71,$79,$E1,$EE,$69,$E0
	.dc.b	$AE,$4F,$E2,$0C,$9E,$B3,$E3,$98,$AF,$C2,$BE,$23,$17,$E5,$76,$78
	.dc.b	$AB,$E3,$B4,$E0,$53,$73,$D1,$FC,$D6,$1F,$0D,$B7,$E6,$C3,$D9,$1C
	.dc.b	$6B,$E0,$1C,$49,$87,$8D,$CB,$49,$1F,$21,$FB,$CF,$A8,$C1,$FB,$9A
	.dc.b	$F0,$61,$EA,$AE,$B3,$A3,$E4,$1E,$3A,$E6,$D3,$91,$D4,$DC,$72,$7F
	.dc.b	$04,$7C,$F7,$CE,$F5,$BF,$CC,$46,$23,$6F,$18,$F8,$6D,$AF,$BD,$3E
	.dc.b	$13,$F7,$B7,$EA,$7E,$8C,$ED,$FC,$11,$E5,$AE,$77,$F1,$A3,$BB,$70
	.dc.b	$FA,$91,$FE,$06,$9F,$E4,$0D,$C3,$FB,$B3,$EC,$30,$7F,$79,$8D,$BF
	.dc.b	$87,$32,$7F,$84,$0E,$E7,$F2,$9E,$9F,$E1,$67,$E5,$AB,$FB,$AC,$7E
	.dc.b	$BA,$30,$69,$F9,$D1,$A7,$F8,$74,$69,$E0,$69,$E0,$FE,$41,$B8,$7C
	.dc.b	$58,$E2,$70,$34,$57,$47,$F3,$98,$E3,$87,$C9,$BF,$E5,$11,$8A,$FB
	.dc.b	$98,$57,$F9,$8C,$79,$5E,$E7,$62,$DF,$6B,$FE,$81,$3C,$6F,$F0,$B1
	.dc.b	$8A,$F2,$3E,$1B,$ED,$87,$C3,$C3,$95,$7C,$D3,$67,$90,$A7,$90,$F8
	.dc.b	$C6,$2F,$BB,$C4,$F2,$DF,$77,$8D,$77,$8C,$46,$8E,$15,$E2,$7B,$19
	.dc.b	$3F,$02,$49,$E4,$F0,$9E,$73,$F3,$E3,$C9,$7F,$D0,$F0,$F8,$8D,$8C
	.dc.b	$9F,$CB,$A3,$2D,$32,$7E,$C6,$BE,$53,$E3,$C3,$DE,$F9,$AF,$C0,$FD
	.dc.b	$56,$9E,$BC,$3C,$F1,$BA,$F3,$9D,$27,$6E,$23,$D8,$43,$E0,$D7,$E2
	.dc.b	$1F,$2C,$FF,$07,$AE,$A7,$69,$F2,$B8,$BF,$C5,$4F,$6A,$FB,$87,$CD
	.dc.b	$9E,$18,$7C,$58,$F7,$36,$78,$47,$8A,$3F,$13,$5D,$30,$F5,$1E,$99
	.dc.b	$F7,$1E,$9C,$3C,$73,$F3,$A3,$E6,$1E,$DC,$3C,$2F,$D1,$3C,$25,$C3
	.dc.b	$C3,$5C,$A3,$E0,$DF,$F1,$1B,$E9,$48,$4C,$F8,$07,$C1,$7F,$B2,$5B
	.dc.b	$1F,$EB,$A6,$BF,$EA,$4F,$81,$7C,$F6,$EB,$F8,$CD,$6A,$BF,$D4,$6D
	.dc.b	$7B,$87,$0F,$86,$3F,$D3,$74,$F8,$3A,$75,$9F,$F6,$5F,$C1,$38,$FF
	.dc.b	$DB,$CD,$8C,$C7,$FB,$A7,$D9,$7C,$2B,$E1,$F7,$3F,$CB,$AF,$85,$FF
	.dc.b	$29,$6B,$FE,$BA,$6B,$E8,$77,$B7,$C9,$87,$2B,$37,$C2,$FA,$8E,$F3
	.dc.b	$FE,$E2,$F8,$97,$C4,$7F,$3E,$6C,$7F,$AD,$5A,$FF,$3B,$6B,$FE,$CC
	.dc.b	$F8,$77,$C3,$7F,$96,$23,$E5,$C6,$3A,$7F,$D7,$CB,$E1,$DE,$E6,$2F
	.dc.b	$85,$FE,$03,$E1,$9F,$0F,$EB,$8F,$F6,$CC,$63,$A7,$FD,$90,$D8,$FE
	.dc.b	$9C,$D7,$FD,$BF,$CA,$F7,$3E,$27,$17,$C5,$B7,$C2,$3E,$9F,$31,$F8
	.dc.b	$2B,$E0,$FE,$4A,$B8,$56,$E3,$06,$EF,$CD,$85,$8F,$F6,$E5,$F2,$8F
	.dc.b	$EC,$4D,$7C,$B3,$E0,$7F,$3F,$1B,$BE,$FF,$0D,$99,$4E,$6F,$F2,$FF
	.dc.b	$0A,$F8,$5E,$5C,$FF,$62,$7C,$B7,$FA,$AA,$D4,$7F,$BA,$F9,$7F,$5A
	.dc.b	$EA,$53,$A7,$FB,$C3,$5F,$D2,$9A,$BF,$7B,$D5,$AE,$67,$A4,$F3,$CB
	.dc.b	$F2,$98,$3E,$15,$F0,$DE,$DA,$D5,$FE,$CB,$E1,$BF,$DB,$4F,$FC,$01
	.dc.b	$A7,$AB,$F0,$E7,$39,$32,$EB,$5A,$BF,$65,$6A,$BD,$2D,$7B,$79,$FE
	.dc.b	$88,$C4,$7F,$10,$7A,$DF,$E4,$5B,$5F,$EC,$8B,$79,$7E,$3C,$3E,$B7
	.dc.b	$C2,$7D,$5C,$7F,$3B,$F0,$CF,$3B,$3C,$9F,$7F,$A9,$39,$CF,$E9,$61
	.dc.b	$C6,$73,$7C,$27,$ED,$C6,$1D,$DF,$3B,$E5,$87,$C3,$9F,$E3,$78,$73
	.dc.b	$9C,$EB,$FA,$B8,$7B,$1F,$91,$87,$C3,$7A,$70,$56,$B0,$FC,$5D,$6B
	.dc.b	$1F,$7E,$1C,$70,$E1,$7B,$C3,$C4,$7D,$58,$CF,$C3,$7C,$6F,$7F,$E3
	.dc.b	$B7,$2F,$82,$E1,$FC,$A5,$7F,$42,$F8,$6F,$E8,$7F,$0A,$F8,$5F,$6D
	.dc.b	$FE,$96,$B4,$27,$4F,$F4,$AC,$93,$FC,$96,$73,$3F,$D0,$2F,$AD,$F4
	.dc.b	$C2,$CF,$E8,$26,$BF,$12,$F9,$57,$F9,$76,$73,$87,$6B,$DD,$F9,$30
	.dc.b	$EF,$39,$8F,$35,$F1,$69,$E7,$C3,$AB,$E1,$32,$78,$6F,$AA,$C3,$C3
	.dc.b	$3C,$23,$D8,$70,$BF,$06,$1F,$02,$3C,$98,$7B,$DE,$A6,$0F,$16,$1D
	.dc.b	$B0,$EA,$D8,$E9,$87,$77,$81,$E5,$39,$9E,$53,$FD,$13,$3F,$E8,$A7
	.dc.b	$07,$FA,$14,$A7,$F7,$2F,$85,$FD,$FF,$C3,$65,$B1,$FE,$8B,$6B,$FC
	.dc.b	$51,$AF,$1F,$77,$F8,$23,$B6,$BB,$6B,$BE,$0E,$D1,$C3,$3F,$C2,$FC
	.dc.b	$33,$DB,$FF,$02,$C6,$1B,$AF,$1C,$3E,$AE,$2C,$39,$4F,$F8,$33,$0E
	.dc.b	$55,$F8,$A6,$BC,$27,$6F,$F8,$82,$B9,$63,$BC,$3B,$CF,$F8,$92,$F8
	.dc.b	$E3,$F9,$AB,$DC,$F3,$BE,$0C,$30,$49,$F5,$39,$79,$61,$E2,$85,$74
	.dc.b	$9D,$C7,$E4,$AF,$C3,$8F,$C1,$9E,$38,$73,$8F,$81,$8F,$A2,$F9,$E3
	.dc.b	$D7,$0F,$83,$7F,$DF,$5B,$AF,$0A,$FE,$A9,$97,$87,$1C,$47,$F0,$C1
	.dc.b	$5C,$1F,$89,$8F,$F8,$EB,$0F,$F0,$45,$F1,$9F,$F2,$5F,$C2,$BE,$1B
	.dc.b	$59,$7F,$90,$23,$13,$FD,$8D,$AF,$7F,$86,$71,$70,$8E,$77,$55,$70
	.dc.b	$C7,$FC,$30,$DC,$3F,$C9,$B8,$72,$AE,$DE,$DF,$1A,$34,$E9,$2F,$84
	.dc.b	$C5,$70,$21,$E5,$72,$FC,$5A,$E6,$EB,$0F,$F1,$55,$73,$8F,$F0,$07
	.dc.b	$C2,$3E,$1B,$D4,$D8,$CC,$7B,$FC,$47,$C2,$BE,$23,$73,$2B,$67,$F8
	.dc.b	$42,$B7,$39,$8F,$F0,$48,$BE,$CE,$1A,$FF,$83,$9A,$FF,$9D,$1F,$EA
	.dc.b	$EF,$A3,$AF,$CE,$79,$70,$7C,$3E,$27,$44,$3F,$A9,$E1,$EA,$8D,$47
	.dc.b	$86,$BF,$89,$3E,$56,$DF,$82,$3C,$DA,$FC,$62,$3E,$FC,$3F,$D8,$B3
	.dc.b	$F3,$63,$F2,$CD,$D3,$FC,$B4,$DD,$FF,$D3,$B9,$7C,$CF,$7F,$9D,$F6
	.dc.b	$3F,$92,$57,$37,$E2,$63,$E6,$F8,$67,$C3,$7E,$45,$B3,$FE,$3B,$6B
	.dc.b	$FA,$7E,$67,$C3,$EE,$A4,$ED,$FC,$D4,$EA,$30,$7F,$35,$AE,$6F,$F9
	.dc.b	$09,$AF,$FA,$A1,$AF,$FA,$E3,$C4,$7C,$3F,$E9,$5B,$1F,$E4,$F6,$BF
	.dc.b	$53,$7F,$DD,$3F,$E2,$79,$E0,$47,$E7,$5D,$7F,$17,$5C,$9F,$F2,$AB
	.dc.b	$5F,$F5,$DB,$5F,$F6,$5F,$C3,$BE,$1F,$F7,$0D,$8F,$F3,$13,$5F,$27
	.dc.b	$BF,$A9,$91,$A7,$F2,$11,$88,$FC,$EC,$F1,$7F,$CC,$ED,$7F,$D9,$AD
	.dc.b	$7F,$DF,$3C,$5F,$F0,$07,$DB,$FE,$15,$F0,$DF,$B8,$9F,$F7,$68,$71
	.dc.b	$3F,$EA,$3F,$87,$57,$0E,$C9,$7F,$E7,$BB,$DF,$37,$FB,$72,$F7,$39
	.dc.b	$AD,$BD,$A3,$2F,$82,$B8,$BF,$EF,$26,$A8,$DE,$1F,$EB,$59,$C9,$2B
	.dc.b	$F9,$24,$EA,$FF,$DD,$B3,$9B,$F1,$1F,$E9,$19,$D3,$F1,$27,$75,$CB
	.dc.b	$0F,$3E,$1D,$E2,$CF,$B6,$6D,$71,$07,$9D,$CB,$F2,$E6,$DF,$82,$30
	.dc.b	$79,$0F,$89,$3C,$30,$F4,$47,$F5,$19,$E1,$18,$C3,$ED,$4F,$17,$CC
	.dc.b	$DC,$3C,$CD,$3C,$C7,$94,$C1,$E2,$3C,$26,$27,$B4,$72,$8F,$AD,$87
	.dc.b	$AB,$0F,$33,$63,$83,$5D,$9D,$B0,$F4,$C7,$08,$FA,$53,$5F,$09,$C0
	.dc.b	$B8,$7B,$1C,$3F,$93,$8E,$0F,$C2,$8C,$DF,$DA,$7A,$9E,$43,$C5,$3C
	.dc.b	$23,$51,$F3,$63,$0F,$E4,$6B,$F0,$4F,$13,$C7,$3C,$2B,$C8,$78,$4A
	.dc.b	$E4,$91,$F8,$4B,$F8,$2F,$39,$F5,$36,$76,$FD,$53,$D6,$78,$CF,$59
	.dc.b	$83,$EC,$47,$27,$18,$7E,$A2,$BB,$C6,$DB,$3E,$68,$E8,$7C,$13,$C4
	.dc.b	$61,$E8,$6D,$EF,$5E,$69,$E6,$FA,$2B,$C1,$87,$CA,$AD,$9F,$05,$C4
	.dc.b	$6A,$3B,$9A,$3E,$05,$F7,$8F,$C8,$C7,$13,$6F,$17,$DF,$87,$D6,$69
	.dc.b	$F1,$A7,$94,$7E,$4D,$B1,$F8,$23,$9C,$71,$8C,$C7,$DE,$3F,$4F,$3D
	.dc.b	$E3,$27,$DF,$BF,$81,$5D,$5B,$5D,$4F,$19,$F3,$4B,$3F,$32,$37,$1D
	.dc.b	$CF,$85,$5F,$01,$F8,$71,$A9,$F0,$47,$92,$BD,$13,$93,$EA,$5F,$CE
	.dc.b	$8C,$CE,$CF,$89,$3B,$7E,$F3,$FB,$DA,$F6,$BC,$A3,$E2,$9C,$0E,$47
	.dc.b	$AE,$36,$F2,$3E,$B1,$EE,$8E,$87,$77,$EF,$61,$F8,$87,$F8,$9B,$5F
	.dc.b	$83,$5E,$86,$9F,$2E,$BB,$BC,$AF,$DC,$F9,$4D,$9F,$2F,$0F,$94,$7D
	.dc.b	$9C,$3C,$4D,$8F,$5D,$7D,$36,$CF,$A1,$A7,$B2,$3C,$33,$C6,$3B,$C7
	.dc.b	$D4,$6C,$FC,$97,$E1,$1C,$5F,$E1,$EF,$F4,$63,$F6,$25,$3F,$99,$16
	.dc.b	$7F,$BA,$36,$FF,$AE,$F2,$C8,$FD,$4E,$28,$C4,$7E,$FD,$B8,$7C,$89
	.dc.b	$D4,$FF,$61,$9F,$DB,$46,$23,$FA,$74,$60,$E3,$87,$D9,$9D,$57,$F8
	.dc.b	$0E,$75,$5C,$CF,$1B,$C1,$FB,$AD,$7E,$29,$FC,$8D,$C6,$1F,$9F,$AE
	.dc.b	$B7,$FE,$09,$9E,$B5,$E4,$3A,$9D,$CF,$71,$6F,$F8,$B7,$C3,$3E,$1B
	.dc.b	$FD,$4C,$D9,$FE,$EB,$2B,$FA,$53,$99,$FF,$3B,$FC,$13,$E1,$F8,$E1
	.dc.b	$27,$8D,$F0,$8F,$F4,$B5,$6E,$72,$7F,$80,$AF,$A3,$F7,$5A,$F1,$3F
	.dc.b	$D2,$A1,$CB,$FE,$0A,$F8,$47,$07,$FA,$FD,$B1,$D8,$38,$59,$7B,$57
	.dc.b	$B4,$F8,$67,$C3,$73,$C2,$57,$3C,$39,$61,$C7,$0E,$17,$AA,$FF,$55
	.dc.b	$D7,$77,$F3,$0D,$7F,$38,$D7,$FC,$47,$F0,$6F,$8B,$FF,$4F,$B6,$3E
	.dc.b	$0B,$5F,$F5,$BB,$5F,$D2,$7C,$33,$DF,$FF,$53,$B6,$3E,$4B,$5F,$F5
	.dc.b	$FB,$5F,$D6,$7C,$33,$E1,$BF,$C9,$F1,$FB,$18,$C7,$47,$D2,$58,$3D
	.dc.b	$9C,$5F,$0B,$FF,$1E,$EA,$F6,$3E,$76,$FF,$3F,$95,$F0,$6F,$84,$CC
	.dc.b	$7E,$FE,$F8,$3F,$86,$37,$1A,$C3,$AE,$16,$72,$7D,$FA,$E3,$3F,$7A
	.dc.b	$3F,$C9,$E6,$9C,$1C,$9F,$C6,$C7,$08,$FE,$1C,$D7,$C9,$3F,$87,$7F
	.dc.b	$89,$7C,$0B,$E0,$FC,$9E,$57,$CF,$C3,$F2,$13,$97,$FD,$E1,$C5,$E2
	.dc.b	$BF,$08,$9D,$3F,$EE,$28,$D6,$9D,$E7,$2F,$E5,$EF,$B5,$72,$9E,$2F
	.dc.b	$93,$0F,$33,$FC,$A6,$32,$D8,$CB,$FE,$72,$F8,$17,$BF,$EA,$F8,$17
	.dc.b	$C1,$7C,$FA,$56,$B4,$FD,$CE,$4B,$DC,$7E,$BE,$75,$5F,$EF,$C9,$CB
	.dc.b	$F0,$23,$39,$FA,$E7,$2F,$CD,$8E,$33,$FC,$FD,$AF,$F4,$36,$BE,$9C
	.dc.b	$7D,$6B,$DF,$FC,$0D,$7E,$F7,$8D,$F0,$2F,$23,$D9,$5A,$CF,$FD,$27
	.dc.b	$5A,$9C,$F0,$FE,$3E,$31,$3F,$C2,$EB,$86,$3F,$A1,$C3,$A4,$E6,$F8
	.dc.b	$46,$27,$EF,$D7,$0B,$E7,$18,$F4,$7F,$39,$93,$E0,$7A,$CF,$AE,$73
	.dc.b	$18,$B4,$63,$1F,$5B,$72,$F3,$7C,$0F,$BB,$DA,$F7,$B6,$32,$4C,$7F
	.dc.b	$26,$7F,$AD,$3E,$07,$FD,$6B,$F0,$0F,$7F,$FB,$96,$6C,$3F,$09,$7B
	.dc.b	$15,$B7,$FD,$7B,$3B,$C3,$9D,$FF,$79,$96,$18,$77,$87,$F3,$3C,$39
	.dc.b	$47,$F8,$06,$74,$FD,$D9,$D5,$73,$9C,$E1,$DE,$BC,$73,$96,$9D,$B0
	.dc.b	$E6,$FA,$70,$EF,$5C,$5F,$0C,$66,$BE,$0E,$1E,$08,$FF,$09,$61,$E0
	.dc.b	$8C,$47,$9C,$EE,$79,$F0,$EE,$FC,$03,$38,$77,$BF,$36,$1E,$2B,$E7
	.dc.b	$5E,$13,$D0,$74,$30,$75,$30,$75,$8E,$57,$E0,$C3,$FB,$4C,$FD,$73
	.dc.b	$FB,$5B,$87,$C8,$7F,$8B,$7D,$EF,$E0,$3F,$03,$FD,$C1,$B1,$F6,$5A
	.dc.b	$EF,$9D,$3B,$F8,$1F,$ED,$B3,$B6,$BD,$5A,$FF,$8F,$69,$18,$80,$8C
	.dc.b	$F0,$FE,$0B,$E0,$5E,$A7,$F9,$17,$4F,$C3,$C6,$2F,$F0,$F7,$C1,$FA
	.dc.b	$6D,$9D,$D7,$1D,$BF,$C9,$D7,$C6,$F7,$8F,$AA,$FF,$27,$87,$1B,$DE
	.dc.b	$1C,$6F,$CF,$87,$1B,$DE,$1C,$6F,$DC,$DA,$DD,$F1,$57,$C6,$B7,$8F
	.dc.b	$AA,$F8,$56,$AF,$85,$EE,$3E,$3E,$3E,$56,$D7,$09,$D4,$71,$3A,$5C
	.dc.b	$7F,$2F,$1A,$37,$5C,$E3,$81,$FD,$62,$3E,$34,$FF,$63,$6E,$BE,$89
	.dc.b	$E7,$87,$F9,$E6,$B4,$D9,$F9,$33,$FE,$35,$BE,$DF,$00,$F8,$1D,$4F
	.dc.b	$E0,$A3,$0F,$F8,$ED,$AF,$F6,$AF,$81,$71,$F0,$84,$EB,$2F,$EF,$0D
	.dc.b	$C7,$FD,$18,$D9,$FC,$7D,$F2,$7B,$7A,$BC,$E4,$D3,$B4,$BD,$DC,$57
	.dc.b	$02,$1E,$57,$2F,$F8,$02,$B9,$BA,$C3,$FC,$19,$5C,$E3,$FC,$B1,$D0
	.dc.b	$FA,$0F,$DC,$B6,$3F,$CC,$6D,$7F,$96,$35,$FF,$53,$39,$AF,$F0,$47
	.dc.b	$B6,$F8,$1F,$9B,$E1,$23,$17,$E0,$C3,$E5,$61,$C7,$5F,$D9,$DE,$F0
	.dc.b	$E3,$87,$0C,$3F,$C3,$77,$B7,$F7,$2B,$EF,$FD,$24,$8F,$4B,$5F,$F1
	.dc.b	$9D,$F8,$FF,$08,$EA,$78,$A6,$4F,$F8,$42,$D3,$A8,$F8,$F3,$F8,$1F
	.dc.b	$C6,$DF,$0B,$FC,$81,$B3,$FE,$8A,$6B,$FD,$4D,$AF,$FB,$35,$CE,$BF
	.dc.b	$E2,$FF,$96,$7C,$27,$C8,$6D,$7F,$A3,$6B,$7B,$7F,$B5,$DB,$0D,$15
	.dc.b	$C2,$3D,$ED,$C3,$FC,$99,$5C,$25,$F0,$BF,$83,$7F,$B4,$35,$FF,$71
	.dc.b	$35,$FF,$73,$35,$FE,$52,$F1,$9F,$C0,$F5,$7D,$83,$F4,$D1,$9C,$78
	.dc.b	$F2,$66,$FE,$75,$6A,$3F,$59,$F2,$FF,$E9,$79,$34,$9E,$CF,$F4,$E7
	.dc.b	$FD,$4B,$7F,$88,$E1,$7C,$0F,$E2,$0E,$75,$2F,$EE,$46,$5F,$F5,$C5
	.dc.b	$F2,$7F,$A1,$FC,$13,$E8,$3F,$AD,$B6,$3F,$D8,$AD,$7F,$C1,$CD,$7F
	.dc.b	$A5,$9C,$A3,$F3,$3F,$53,$23,$FD,$8D,$3C,$76,$FF,$3F,$37,$3F,$55
	.dc.b	$6E,$BF,$CF,$33,$97,$FA,$BF,$C3,$BE,$23,$F8,$93,$63,$FD,$A4,$D7
	.dc.b	$FD,$03,$E0,$FE,$07,$0F,$2C,$E7,$1F,$F4,$73,$63,$37,$FC,$76,$B8
	.dc.b	$5F,$82,$3B,$91,$FF,$70,$46,$95,$73,$78,$1E,$49,$E9,$CB,$C6,$3F
	.dc.b	$D2,$B3,$B7,$8C,$E9,$E0,$43,$6E,$5A,$67,$0F,$5C,$65,$B1,$9C,$3D
	.dc.b	$AF,$A2,$FC,$53,$A8,$E6,$7C,$88,$CC,$F0,$8C,$CF,$FA,$D7,$0F,$81
	.dc.b	$87,$A4,$98,$78,$6F,$85,$75,$23,$F8,$6A,$E3,$3D,$AF,$E3,$9E,$13
	.dc.b	$D0,$62,$B8,$BF,$E1,$89,$E2,$E8,$F9,$27,$B4,$F9,$07,$E5,$0C,$9F
	.dc.b	$E6,$5B,$FC,$33,$96,$BE,$8B,$F3,$79,$FF,$83,$3C,$D1,$CA,$FE,$14
	.dc.b	$66,$3C,$55,$F6,$EF,$FD,$7F,$1F,$86,$9E,$33,$97,$84,$7C,$CB,$F0
	.dc.b	$3E,$D8,$CD,$FC,$08,$F0,$4F,$91,$E8,$78,$30,$DC,$66,$3A,$1F,$25
	.dc.b	$C4,$FA,$23,$81,$C8,$FD,$3D,$79,$AB,$9C,$78,$CD,$9A,$38,$4F,$96
	.dc.b	$FD,$91,$8B,$F5,$BE,$83,$91,$C4,$F2,$99,$3F,$C9,$F7,$E0,$71,$3D
	.dc.b	$1E,$15,$DE,$F5,$87,$CE,$9E,$75,$BA,$F7,$46,$2B,$77,$ED,$8F,$31
	.dc.b	$E4,$9E,$8E,$0F,$19,$F0,$1E,$64,$C3,$DB,$7D,$DF,$9A,$62,$FB,$D7
	.dc.b	$18,$CB,$DD,$D3,$DF,$0F,$55,$F8,$A3,$DF,$3C,$1F,$87,$5D,$1F,$61
	.dc.b	$C7,$0F,$63,$C0,$EC,$7E,$04,$87,$B0,$D1,$D0,$E4,$79,$CE,$26,$EB
	.dc.b	$C8,$7C,$17,$38,$7F,$00,$78,$9E,$43,$AE,$1E,$CC,$3D,$F1,$8B,$F3
	.dc.b	$DF,$38,$ED,$3A,$8C,$57,$19,$F4,$DF,$CF,$7C,$38,$7E,$D2,$75,$1C
	.dc.b	$8D,$C7,$C7,$AF,$4D,$FC,$BA,$DD,$FA,$6F,$C8,$7C,$13,$91,$1D,$57
	.dc.b	$93,$0F,$4D,$F3,$6D,$7C,$8B,$EB,$87,$96,$7F,$39,$87,$2B,$E1,$7D
	.dc.b	$EF,$D2,$70,$7F,$8E,$38,$AE,$B1,$E4,$8D,$3F,$AA,$74,$F3,$8E,$F7
	.dc.b	$E1,$8F,$CE,$5F,$85,$F3,$4F,$4A,$E3,$1F,$67,$0F,$73,$CC,$E3,$7C
	.dc.b	$DF,$C4,$DF,$37,$C9,$7C,$CF,$0B,$EA,$6B,$E7,$6D,$7F,$1D,$BF,$0B
	.dc.b	$F5,$A7,$CA,$78,$4E,$F7,$E1,$8F,$C4,$5F,$86,$B8,$E1,$F8,$93,$E6
	.dc.b	$DF,$AD,$B5,$D6,$3C,$E7,$57,$9D,$F5,$9F,$D8,$5F,$5C,$3C,$F1,$FC
	.dc.b	$73,$0E,$F8,$7E,$38,$FD,$0E,$1D,$63,$31,$FA,$D3,$F1,$A7,$8C,$F6
	.dc.b	$94,$F1,$95,$E3,$7E,$18,$FA,$27,$CB,$29,$F2,$CB,$3F,$2E,$FC,$F1
	.dc.b	$FC,$02,$B9,$E1,$ED,$8F,$D4,$9C,$CF,$9A,$5A,$F8,$68,$2C,$3F,$C7
	.dc.b	$FF,$58,$FB,$9F,$ED,$D5,$86,$BC,$63,$4E,$1A,$F0,$6D,$75,$8F,$F3
	.dc.b	$E7,$C2,$7F,$AE,$9F,$F4,$CB,$63,$50,$9E,$37,$E1,$8F,$44,$F3,$E9
	.dc.b	$5F,$1A,$FF,$57,$B6,$BC,$D8,$7F,$B7,$7E,$07,$FC,$D7,$1E,$A6,$BE
	.dc.b	$E6,$BF,$E0,$D6,$BF,$E6,$27,$78,$78,$6B,$FD,$6D,$3E,$4B,$F0,$F6
	.dc.b	$77,$E4,$93,$93,$E0,$C7,$FB,$2A,$B8,$C7,$D2,$AE,$D5,$FE,$F3,$3E
	.dc.b	$34,$7F,$87,$DA,$FF,$9D,$1A,$FD,$B6,$BF,$E8,$8F,$0F,$FD,$F9,$F1
	.dc.b	$0F,$88,$A7,$58,$93,$E5,$13,$B7,$0D,$7C,$C7,$78,$FF,$42,$E5,$F6
	.dc.b	$3E,$01,$EF,$BE,$07,$F0,$F1,$86,$FC,$0B,$D7,$CD,$6F,$1F,$C7,$FC
	.dc.b	$7B,$E4,$3F,$07,$1F,$E9,$28,$C6,$FF,$92,$6C,$70,$9D,$BF,$9D,$F9
	.dc.b	$3F,$F7,$C5,$7E,$61,$B1,$F9,$C6,$BF,$E4,$B6,$BF,$EA,$0F,$80,$7C
	.dc.b	$0F,$48,$46,$2B,$F4,$6D,$8F,$D3,$B5,$FF,$2B,$39,$DB,$F3,$AD,$9F
	.dc.b	$D9,$73,$BE,$03,$F4,$11,$F4,$E7,$57,$FB,$76,$BE,$2C,$3F,$59,$87
	.dc.b	$E8,$EF,$8C,$FD,$63,$57,$F7,$7E,$01,$C5,$EB,$6C,$7E,$F1,$B3,$FC
	.dc.b	$01,$AF,$F9,$BD,$CD,$FE,$D9,$B3,$EA,$BF,$F6,$BB,$DE,$FC,$AF,$FB
	.dc.b	$2A,$77,$7F,$C4,$5B,$97,$C1,$E5,$DC,$E6,$EA,$F7,$1E,$F6,$B8,$9E
	.dc.b	$FB,$E0,$37,$5A,$D3,$F6,$15,$A9,$CD,$F1,$BE,$15,$FA,$1B,$E1,$87
	.dc.b	$E7,$B0,$E1,$39,$BE,$18,$72,$8E,$B8,$72,$BF,$A1,$83,$E0,$3D,$7B
	.dc.b	$BD,$2E,$AF,$F1,$E9,$D6,$1F,$C1,$63,$59,$7A,$AF,$8D,$F0,$AF,$F7
	.dc.b	$9D,$F0,$C3,$F6,$78,$70,$9F,$57,$C0,$7C,$2F,$63,$E1,$B6,$32,$4A
	.dc.b	$FC,$31,$FD,$0F,$E0,$3F,$A2,$7B,$CF,$7B,$BD,$FD,$2A,$D0,$9D,$3F
	.dc.b	$D2,$F2,$4F,$F3,$F9,$CC,$FF,$4F,$17,$DA,$FA,$E1,$67,$FD,$79,$35
	.dc.b	$FC,$05,$F3,$AF,$EA,$73,$9C,$3B,$DE,$EF,$CB,$87,$82,$73,$1E,$7B
	.dc.b	$E2,$D3,$D1,$87,$57,$C2,$64,$F0,$DF,$57,$D2,$66,$BA,$DF,$BA,$3C
	.dc.b	$37,$EC,$79,$18,$3B,$D7,$3C,$3D,$58,$78,$67,$E3,$46,$27,$51,$DE
	.dc.b	$74,$F3,$3C,$A7,$63,$CA,$7F,$76,$9F,$EE,$EF,$D8,$30,$7F,$7C,$F7
	.dc.b	$5F,$00,$F8,$0F,$EF,$B1,$86,$C6,$67,$83,$5F,$F7,$94,$BD,$5F,$F0
	.dc.b	$5C,$75,$6B,$A6,$BF,$83,$8E,$1B,$78,$27,$8E,$9F,$88,$F8,$07,$99
	.dc.b	$F8,$98,$C3,$76,$E5,$5C,$1B,$7F,$89,$BE,$38,$DB,$E0,$FF,$88,$71
	.dc.b	$F0,$BF,$7D,$B8,$F6,$8E,$07,$E4,$E3,$38,$76,$8F,$C3,$5F,$66,$DF
	.dc.b	$65,$5C,$1F,$06,$3E,$77,$FC,$65,$5C,$70,$F0,$47,$8E,$74,$F2,$6C
	.dc.b	$EF,$0F,$D6,$36,$7A,$E3,$E7,$AD,$E3,$E3,$AD,$57,$3A,$F6,$BF,$D1
	.dc.b	$E3,$D5,$1F,$E3,$26,$ED,$EF,$C7,$CB,$39,$BF,$F0,$0C,$F8,$9B,$7E
	.dc.b	$23,$F5,$5E,$FB,$E0,$33,$08,$C3,$FE,$11,$6B,$FD,$67,$E0,$1C,$7C
	.dc.b	$27,$38,$FF,$91,$9B,$8F,$F0,$D6,$DF,$F9,$76,$F8,$BD,$7A,$B9,$11
	.dc.b	$FF,$10,$43,$D9,$C4,$7F,$69,$AF,$F2,$53,$E7,$AF,$4D,$6E,$3A,$10
	.dc.b	$E1,$1A,$AF,$7F,$33,$E0,$29,$2B,$FC,$89,$DC,$F6,$9F,$21,$99,$FF
	.dc.b	$38,$DF,$17,$FC,$E2,$D7,$37,$FB,$E8,$C6,$BF,$E6,$5B,$7C,$0F,$D4
	.dc.b	$C7,$B3,$D6,$7F,$CD,$F5,$FE,$52,$6F,$A1,$E8,$38,$5B,$F0,$BC,$65
	.dc.b	$E2,$9F,$01,$1C,$1E,$C9,$7C,$17,$23,$8C,$7F,$85,$7E,$01,$F0,$5F
	.dc.b	$84,$6C,$7F,$94,$DA,$FF,$94,$34,$7C,$17,$FA,$12,$3F,$D3,$53,$A7
	.dc.b	$FA,$70,$70,$FF,$51,$6B,$C2,$BF,$92,$C7,$6C,$FF,$D1,$94,$F7,$3E
	.dc.b	$43,$6F,$C9,$7F,$9A,$F8,$2F,$C1,$47,$F8,$BD,$B7,$FE,$9A,$6F,$37
	.dc.b	$F8,$FB,$4F,$F3,$54,$E7,$8F,$F9,$AD,$72,$BF,$4B,$5F,$ED,$0D,$7F
	.dc.b	$D7,$0F,$45,$A7,$E2,$7B,$3E,$B9,$E5,$AF,$D7,$FC,$0B,$E0,$A9,$2F
	.dc.b	$EB,$7C,$ED,$7F,$D7,$F3,$B7,$FD,$7E,$D7,$33,$FD,$32,$31,$97,$E3
	.dc.b	$FC,$CF,$C3,$9C,$DF,$F5,$DF,$C0,$BE,$0F,$F8,$0B,$70,$FF,$48,$35
	.dc.b	$FB,$DB,$3E,$A7,$75,$FD,$6A,$B5,$5F,$79,$AF,$F8,$35,$AF,$FB,$75
	.dc.b	$E9,$7F,$A3,$F8,$4F,$CD,$9D,$27,$FD,$41,$F0,$8F,$83,$FE,$2A,$D8
	.dc.b	$CC,$7F,$18,$F1,$3F,$05,$1F,$EF,$D9,$D3,$FE,$6F,$8C,$DF,$FB,$6A
	.dc.b	$72,$DA,$EA,$73,$E7,$B7,$D8,$87,$37,$2F,$F9,$EE,$B9,$F2,$FF,$8A
	.dc.b	$DB,$1F,$EF,$BC,$3C,$33,$9B,$E3,$87,$86,$BF,$D7,$D5,$A4,$2F,$9B
	.dc.b	$6B,$8D,$FF,$8E,$6B,$84,$E4,$F3,$DF,$57,$D0,$62,$7C,$4D,$3C,$D8
	.dc.b	$74,$7C,$38,$78,$0F,$0D,$F6,$71,$86,$E7,$FC,$07,$56,$73,$1E,$C8
	.dc.b	$C4,$EE,$B9,$38,$9D,$1E,$33,$D4,$7C,$D7,$EA,$39,$AE,$AF,$B0,$FC
	.dc.b	$04,$69,$A7,$42,$9B,$C3,$E2,$13,$0F,$87,$19,$6C,$66,$3E,$1C,$F5
	.dc.b	$8C,$47,$99,$B7,$EC,$8F,$34,$EE,$B9,$92,$39,$9E,$5A,$E7,$7E,$C5
	.dc.b	$3F,$77,$0F,$8D,$5F,$0C,$E1,$1C,$DF,$13,$70,$F1,$46,$C9,$5D,$1F
	.dc.b	$BE,$79,$8F,$A1,$1C,$23,$C7,$87,$A6,$7C,$E7,$2A,$ED,$7E,$4A,$F8
	.dc.b	$2D,$8C,$CF,$53,$81,$E0,$3E,$9C,$F8,$DF,$C4,$9B,$79,$B6,$39,$BC
	.dc.b	$1F,$88,$E4,$F2,$C7,$3B,$F5,$4F,$C5,$C3,$F3,$37,$E5,$BD,$CF,$27
	.dc.b	$A1,$E5,$BF,$82,$FC,$49,$E0,$FC,$E3,$53,$D7,$0F,$CE,$4E,$AF,$BD
	.dc.b	$79,$4E,$07,$B8,$EE,$5B,$EE,$FC,$63,$17,$D5,$F1,$1C,$E7,$AB,$CB
	.dc.b	$0F,$BE,$FF,$6E,$8F,$E4,$2D,$C3,$FD,$1F,$18,$8F,$CF,$99,$3F,$BA
	.dc.b	$9F,$78,$C1,$87,$F8,$0F,$C2,$7F,$2F,$3E,$1D,$1C,$4F,$BE,$30,$FF
	.dc.b	$1C,$8D,$9A,$7D,$D3,$D1,$D1,$C6,$35,$87,$C4,$3E,$E3,$8C,$3E,$24
	.dc.b	$FF,$3D,$8C,$3F,$70,$87,$FA,$59,$CB,$F4,$AB,$C0,$F1,$3C,$07,$88
	.dc.b	$F2,$E1,$C3,$0F,$5C,$60,$D6,$1E,$47,$89,$CC,$D1,$5D,$38,$69,$30
	.dc.b	$F5,$B6,$3F,$A4,$61,$F2,$1E,$F3,$A3,$FA,$04,$7B,$30,$FE,$11,$18
	.dc.b	$9D,$9F,$0A,$3D,$F5,$D6,$3E,$16,$1E,$DB,$ED,$1E,$7B,$EC,$F8,$8E
	.dc.b	$67,$02,$E1,$F1,$2F,$EB,$4F,$86,$79,$1E,$1B,$EC,$FC,$69,$F1,$61
	.dc.b	$E8,$8E,$C7,$03,$B1,$6F,$B5,$F9,$23,$EE,$1E,$49,$E4,$E0,$E4,$E2
	.dc.b	$78,$C7,$6A,$F3,$B6,$BD,$11,$88,$E4,$F9,$AF,$FA,$C6,$1F,$3D,$B3
	.dc.b	$E1,$6B,$E0,$6C,$F1,$8F,$E5,$B3,$C5,$A7,$12,$9C,$4B,$1C,$5F,$E0
	.dc.b	$97,$EA,$7C,$B7,$F0,$E3,$06,$70,$F8,$17,$DE,$B8,$B6,$FF,$AB,$B6
	.dc.b	$FF,$A4,$36,$7F,$53,$7D,$23,$F0,$B5,$E3,$38,$CF,$A4,$3F,$CE,$FE
	.dc.b	$13,$FD,$05,$06,$9F,$E8,$5E,$67,$C1,$3E,$8B,$FD,$A3,$2F,$85,$FE
	.dc.b	$27,$7F,$E2,$28,$C3,$86,$BF,$C6,$1A,$FF,$A3,$9E,$D4,$F1,$7F,$D5
	.dc.b	$CF,$FB,$22,$3C,$BF,$0A,$E6,$FF,$70,$B6,$FF,$C6,$39,$FF,$55,$8C
	.dc.b	$47,$FA,$7E,$DE,$DF,$0B,$F3,$CA,$B6,$FF,$A8,$1A,$FF,$8F,$0D,$47
	.dc.b	$08,$EB,$7F,$DD,$74,$F8,$BE,$17,$A0,$ED,$7E,$4E,$0E,$58,$7F,$B8
	.dc.b	$89,$A7,$FB,$EB,$E1,$1F,$07,$FE,$FC,$9F,$85,$7F,$31,$AF,$F4,$06
	.dc.b	$BF,$EB,$66,$BD,$AB,$A7,$C2,$C8,$F1,$1F,$E5,$69,$FF,$79,$FC,$57
	.dc.b	$C4,$8F,$B2,$D7,$FA,$6B,$5F,$F6,$1B,$5F,$89,$F0,$CF,$85,$F4,$C7
	.dc.b	$F5,$48,$C6,$3F,$84,$5F,$0C,$F8,$0C,$5F,$0B,$FF,$67,$FC,$33,$E1
	.dc.b	$BF,$C1,$D1,$F4,$63,$1B,$FE,$39,$B1,$FD,$89,$AF,$FB,$63,$E1,$7E
	.dc.b	$FC,$FC,$0E,$2F,$C0,$E3,$FE,$2B,$E3,$FF,$0D,$56,$AF,$FD,$C9,$CB
	.dc.b	$E5,$C3,$E3,$C2,$74,$F5,$9F,$2B,$FE,$80,$8C,$BF,$87,$6B,$98,$F0
	.dc.b	$4E,$5B,$1F,$ED,$0F,$85,$FD,$84,$F8,$2B,$C9,$7F,$E2,$FF,$85,$FD
	.dc.b	$24,$FA,$E9,$C7,$CB,$05,$69,$FF,$7F,$4E,$AB,$D7,$39,$9F,$F7,$EB
	.dc.b	$5F,$CF,$35,$FF,$4D,$B5,$EB,$C9,$DE,$3B,$61,$FA,$1B,$FD,$A5,$F6
	.dc.b	$AF,$27,$C0,$BE,$0F,$DF,$AB,$E1,$1F,$0B,$A9,$CC,$63,$0F,$A3,$7C
	.dc.b	$5F,$CC,$DF,$1B,$E1,$18,$AE,$1A,$7F,$1C,$8F,$DC,$5F,$18,$C1,$F9
	.dc.b	$19,$DE,$7F,$E5,$B9,$D3,$FE,$6F,$C3,$B5,$F8,$BE,$17,$EB,$CB,$D6
	.dc.b	$DC,$F6,$9C,$C6,$31,$FF,$61,$E0,$9C,$E7,$F1,$1B,$7E,$EF,$85,$F8
	.dc.b	$FE,$EF,$C8,$6C,$64,$98,$7A,$8F,$E2,$9F,$0B,$FE,$D0,$F8,$07,$C1
	.dc.b	$7C,$0B,$FC,$25,$7F,$B5,$2B,$53,$9B,$B3,$FC,$74,$4E,$16,$6B,$FF
	.dc.b	$49,$61,$DA,$78,$56,$EB,$57,$C6,$7C,$02,$B8,$3D,$2F,$9E,$1F,$E5
	.dc.b	$AC,$39,$46,$27,$A9,$C6,$7D,$0D,$33,$5D,$30,$F0,$C6,$4F,$8A,$57
	.dc.b	$31,$E2,$8F,$2B,$C6,$BC,$11,$C2,$31,$87,$AE,$3E,$29,$CE,$76,$F2
	.dc.b	$BF,$2D,$75,$8D,$BC,$9F,$21,$B3,$F0,$B3,$C9,$EA,$7F,$6A,$C3,$FA
	.dc.b	$AB,$83,$E3,$1F,$E8,$0F,$83,$7C,$23,$E1,$7F,$B8,$C6,$1B,$19,$9E
	.dc.b	$0D,$7F,$01,$2F,$63,$FC,$07,$1D,$5A,$E9,$AF,$F6,$38,$E1,$C3,$E0
	.dc.b	$9E,$39,$7D,$EF,$85,$79,$BF,$DA,$63,$0D,$D7,$95,$70,$6D,$FE,$46
	.dc.b	$F8,$E3,$6F,$82,$18,$F8,$9F,$C0,$37,$1E,$F1,$C4,$FC,$E4,$67,$0E
	.dc.b	$F1,$F8,$7B,$EE,$DB,$EE,$F6,$C7,$CC,$F8,$6F,$93,$E3,$AE,$38,$78
	.dc.b	$23,$C7,$5C,$1E,$58,$F8,$DE,$C7,$92,$35,$7D,$21,$7D,$31,$FF,$09
	.dc.b	$36,$3C,$55,$D6,$3D,$37,$FE,$F5,$6E,$DC,$32,$FF,$0D,$4F,$FB,$E7
	.dc.b	$1E,$D5,$AC,$3D,$55,$AB,$E1,$1F,$D8,$AB,$C9,$F0,$8F,$85,$D4,$FD
	.dc.b	$38,$C3,$FD,$A1,$AF,$F1,$7F,$85,$71,$F0,$84,$EB,$36,$3F,$E2,$F8
	.dc.b	$C3,$70,$E1,$87,$2B,$ED,$D5,$CE,$FF,$C4,$B8,$72,$9F,$24,$F0,$F7
	.dc.b	$34,$C8,$E2,$7F,$7E,$B4,$E6,$3F,$97,$5F,$4C,$7C,$97,$CE,$7F,$C0
	.dc.b	$BD,$0F,$84,$F6,$36,$33,$1F,$1B,$C2,$7B,$4F,$85,$DC,$CA,$D9,$FE
	.dc.b	$17,$9D,$38,$3F,$DF,$F7,$D5,$B3,$FE,$5A,$8C,$47,$F3,$B5,$E7,$79
	.dc.b	$09,$1D,$7E,$12,$BF,$D2,$E7,$6F,$B2,$7D,$16,$AD,$CF,$59,$F0,$11
	.dc.b	$FF,$40,$FC,$1B,$E0,$BE,$F3,$63,$FC,$56,$D7,$FB,$9E,$6F,$83,$FF
	.dc.b	$48,$CE,$78,$BE,$D5,$23,$1B,$FF,$9C,$DB,$0B,$E5,$EB,$78,$31,$FF
	.dc.b	$3E,$D7,$7A,$FE,$BC,$D7,$94,$7F,$A0,$BE,$0F,$F0,$31,$FE,$40,$6B
	.dc.b	$FD,$BD,$B8,$FF,$9F,$F5,$FF,$18,$CE,$F8,$FF,$39,$5A,$CF,$FD,$10
	.dc.b	$D9,$F3,$57,$07,$A5,$78,$48,$FF,$AA,$A7,$C2,$FF,$75,$F8,$17,$C1
	.dc.b	$7E,$A5,$B1,$FE,$56,$6B,$F3,$2D,$F0,$7B,$A9,$3B,$7E,$EC,$62,$3F
	.dc.b	$C0,$42,$B9,$BE,$19,$FF,$53,$C6,$23,$FD,$49,$F0,$2F,$82,$FD,$E3
	.dc.b	$63,$FC,$D2,$D7,$C9,$F0,$7A,$99,$1A,$7F,$1F,$18,$8F,$1C,$F1,$7F
	.dc.b	$C2,$CD,$7C,$33,$FE,$BA,$F8,$07,$C0,$FF,$0D,$6C,$FF,$9E,$9A,$FF
	.dc.b	$8A,$7C,$07,$C1,$6E,$A4,$ED,$FE,$2D,$3A,$8C,$10,$FF,$7E,$CF,$07
	.dc.b	$FC,$69,$5D,$B8,$7F,$32,$DE,$1F,$F6,$55,$F3,$AD,$E1,$C6,$F6,$F8
	.dc.b	$30,$C5,$FF,$B4,$70,$E5,$7F,$EF,$E9,$CB,$FD,$8E,$32,$9B,$87,$7C
	.dc.b	$3A,$CF,$A5,$AF,$C3,$AE,$6F,$1C,$2D,$71,$C3,$C5,$1A,$C3,$AD,$D8
	.dc.b	$DB,$6F,$B5,$7C,$08,$C4,$7C,$83,$27,$A6,$78,$B8,$3C,$45,$BF,$13
	.dc.b	$B3,$57,$E2,$6C,$F2,$7E,$69,$83,$FD,$77,$3C,$9E,$A6,$8F,$21,$70
	.dc.b	$F8,$A4,$C3,$D5,$1C,$CD,$C6,$5F,$21,$88,$F8,$EE,$E3,$F8,$DC,$F8
	.dc.b	$6F,$EB,$92,$FF,$01,$5B,$BE,$2F,$D5,$30,$7E,$04,$A7,$AA,$FA,$E1
	.dc.b	$F3,$9F,$24,$7F,$8C,$A3,$13,$D2,$77,$3C,$AF,$E5,$3F,$06,$74,$47
	.dc.b	$FC,$61,$87,$A7,$0E,$71,$8C,$2B,$E3,$AF,$8E,$F3,$3E,$2E,$1E,$4B
	.dc.b	$E5,$1E,$A6,$BB,$39,$1D,$4F,$35,$FC,$67,$58,$7C,$97,$EE,$E1,$E2
	.dc.b	$8F,$5D,$FC,$27,$53,$F2,$CE,$35,$F8,$79,$D3,$70,$F8,$13,$E2,$26
	.dc.b	$1E,$6A,$DC,$78,$0F,$29,$4E,$A7,$E6,$70,$F2,$5F,$A6,$78,$CF,$E1
	.dc.b	$CF,$8B,$7E,$0C,$3D,$2D,$AF,$24,$E8,$87,$92,$FA,$BE,$1B,$EA,$7C
	.dc.b	$F8,$CC,$7C,$18,$CD,$7B,$B0,$F6,$3D,$E7,$93,$5F,$31,$F0,$F0,$F9
	.dc.b	$75,$F8,$A8,$F7,$9A,$8C,$CF,$5B,$F0,$D6,$CF,$0D,$7E,$6C,$E2,$F9
	.dc.b	$DF,$96,$79,$E7,$C5,$19,$8D,$C7,$57,$71,$D6,$36,$66,$36,$77,$8F
	.dc.b	$0B,$F2,$4E,$31,$F9,$88,$CD,$7D,$29,$F9,$EF,$B1,$FC,$A1,$D0,$E4
	.dc.b	$57,$83,$F1,$F0,$FB,$2D,$3F,$08,$78,$0E,$73,$E0,$8F,$C3,$57,$AA
	.dc.b	$7D,$0D,$7B,$CE,$DF,$AE,$79,$A7,$BD,$79,$E3,$75,$C6,$3C,$A6,$4B
	.dc.b	$3F,$BA,$C3,$F0,$75,$CC,$EA,$7A,$4F,$BD,$1B,$C3,$FB,$14,$7F,$AD
	.dc.b	$4F,$E4,$65,$3F,$97,$96,$BF,$B9,$B6,$7F,$B7,$FC,$14,$8F,$E0,$8D
	.dc.b	$7E,$75,$A3,$18,$FD,$A8,$C4,$FF,$3E,$9F,$51,$AC,$7E,$44,$EA,$7F
	.dc.b	$C0,$53,$A7,$81,$E0,$29,$E0,$79,$3F,$0C,$E2,$6A,$BC,$51,$A8,$F8
	.dc.b	$D1,$FD,$02,$31,$87,$D3,$AF,$14,$E6,$BC,$07,$43,$C0,$75,$BF,$E2
	.dc.b	$7F,$02,$E6,$FF,$55,$36,$7F,$DC,$95,$FE,$68,$8C,$47,$F9,$BB,$E0
	.dc.b	$5F,$05,$C2,$E4,$F0,$C3,$EC,$5F,$0C,$3F,$D5,$5D,$2C,$78,$DE,$CF
	.dc.b	$F7,$95,$76,$7E,$F3,$5E,$78,$7F,$93,$1D,$B5,$7C,$13,$E0,$7F,$D8
	.dc.b	$8D,$8E,$C6,$A3,$FD,$55,$6F,$6B,$71,$27,$75,$C5,$F8,$13,$97,$FC
	.dc.b	$BB,$19,$7F,$28,$D7,$F3,$0D,$47,$97,$C3,$FF,$6C,$57,$FB,$0E,$32
	.dc.b	$D5,$F0,$4F,$81,$DD,$49,$F5,$D6,$EB,$51,$89,$F8,$92,$69,$23,$E5
	.dc.b	$7C,$03,$E2,$FE,$53,$63,$31,$83,$E7,$7C,$1F,$C7,$8F,$A4,$D7,$FC
	.dc.b	$FE,$D7,$F6,$CD,$7F,$75,$F0,$6F,$83,$F3,$C7,$EE,$E3,$19,$FD,$D5
	.dc.b	$83,$DB,$C5,$F0,$BF,$E0,$9F,$03,$F6,$79,$7F,$15,$C9,$F6,$EF,$FD
	.dc.b	$F1,$CB,$FE,$C0,$AD,$5F,$FB,$1B,$8B,$A6,$1F,$84,$9C,$D7,$E5,$A7
	.dc.b	$34,$8E,$51,$D2,$BE,$DE,$1D,$DE,$51,$FE,$CE,$AF,$2B,$5F,$C0,$DF
	.dc.b	$89,$FE,$34,$D7,$F8,$F7,$C0,$7D,$7A,$F5,$FC,$03,$E0,$7C,$15,$F8
	.dc.b	$AA,$D3,$FE,$E0,$9D,$57,$E7,$A7,$39,$7E,$C2,$73,$5E,$BA,$EF,$C5
	.dc.b	$D3,$35,$6A,$FF,$DC,$F8,$76,$C3,$F9,$83,$5E,$38,$7E,$CE,$31,$1F
	.dc.b	$9E,$F7,$5E,$FF,$A3,$67,$C0,$3E,$07,$D7,$6B,$DE,$3F,$C0,$EF,$75
	.dc.b	$A9,$CD,$7A,$4F,$C2,$5F,$3B,$FD,$EC,$EB,$24,$7E,$C3,$0E,$6E,$B1
	.dc.b	$F2,$5F,$1C,$3A,$1E,$38,$CE,$3E,$1F,$81,$FA,$38,$BD,$2E,$D8,$FF
	.dc.b	$55,$CD,$5A,$9C,$E3,$F0,$E3,$18,$F3,$BF,$E4,$18,$73,$8F,$77,$C0
	.dc.b	$FB,$FD,$9B,$5E,$C6,$E1,$E6,$C9,$5F,$2A,$3F,$86,$FC,$0F,$F0,$EF
	.dc.b	$7D,$EF,$F9,$6B,$54,$AD,$4E,$71,$B3,$FD,$B6,$6B,$F5,$6B,$E5,$AB
	.dc.b	$EA,$EC,$C5,$F3,$69,$98,$DB,$CE,$FA,$BE,$83,$66,$4A,$66,$38,$E1
	.dc.b	$F0,$27,$8C,$FA,$CE,$31,$97,$89,$99,$E3,$7E,$A7,$C7,$1D,$6F,$D1
	.dc.b	$18,$8E,$11,$83,$FC,$20,$E0,$FC,$F4,$F3,$8C,$E1,$F2,$2F,$C5,$7C
	.dc.b	$E3,$15,$BB,$F8,$97,$D5,$F9,$47,$F5,$D9,$FE,$B8,$FB,$4C,$1E,$2F
	.dc.b	$75,$F0,$0F,$81,$FE,$C9,$18,$6C,$66,$78,$35,$FF,$15,$4B,$E0,$7F
	.dc.b	$8F,$47,$56,$BA,$6B,$FE,$33,$8E,$1B,$F8,$27,$8E,$7F,$7B,$E0,$5E
	.dc.b	$57,$F8,$E6,$30,$DD,$39,$57,$06,$DF,$E4,$2F,$8E,$36,$F8,$3F,$DA
	.dc.b	$B1,$F0,$BF,$88,$6E,$3D,$A3,$81,$F9,$A8,$CE,$1D,$A3,$F1,$F7,$D9
	.dc.b	$B7,$D9,$57,$07,$C1,$8F,$9D,$FE,$F7,$5C,$70,$F0,$47,$8E,$74,$F2
	.dc.b	$6C,$EF,$0F,$F2,$E3,$67,$AE,$3E,$7A,$DE,$3E,$3A,$D5,$73,$AF,$6B
	.dc.b	$FC,$62,$3D,$51,$FB,$68,$46,$37,$F8,$18,$F9,$AB,$53,$9A,$F5,$DF
	.dc.b	$F9,$BE,$79,$65,$F0,$65,$F0,$2F,$86,$D6,$5F,$D8,$BD,$77,$D2,$E2
	.dc.b	$DD,$A9,$CC,$7F,$8C,$B0,$E4,$DC,$38,$56,$EE,$DF,$0A,$FE,$6B,$7C
	.dc.b	$1B,$DF,$FE,$41,$8F,$EC,$D1,$FE,$45,$6B,$FB,$86,$BF,$E8,$97,$39
	.dc.b	$FF,$95,$2B,$FC,$41,$3F,$D3,$70,$F4,$78,$FE,$B8,$F8,$38,$5C,$38
	.dc.b	$E7,$F0,$6B,$CE,$66,$75,$EC,$79,$49,$7E,$56,$BD,$F1,$79,$DF,$00
	.dc.b	$E3,$1F,$1E,$70,$FF,$87,$A7,$6E,$4F,$F2,$EF,$D1,$7F,$52,$7F,$CC
	.dc.b	$CD,$7F,$8C,$B5,$FF,$55,$35,$FF,$57,$3B,$D3,$F1,$5F,$08,$F7,$7E
	.dc.b	$13,$6B,$BF,$77,$C5,$BE,$1A,$FF,$AD,$EB,$55,$FD,$B1,$B8,$7E,$56
	.dc.b	$B7,$3F,$E6,$4E,$F7,$C0,$7F,$9E,$DB,$1F,$E7,$F6,$BF,$CC,$9A,$FF
	.dc.b	$B0,$DC,$D7,$DB,$F8,$8F,$C4,$9F,$66,$33,$8F,$FB,$2D,$B5,$F9,$6D
	.dc.b	$15,$C2,$3F,$D0,$F3,$A5,$EF,$7C,$77,$FC,$F6,$D7,$3A,$7F,$A4,$E3
	.dc.b	$07,$BB,$8F,$F4,$5D,$9F,$9C,$3F,$19,$1F,$0A,$B8,$9F,$76,$34,$FF
	.dc.b	$A6,$23,$5F,$44,$F9,$7F,$CA,$36,$7F,$D3,$AD,$7E,$87,$89,$A9,$91
	.dc.b	$A7,$ED,$C6,$23,$FC,$43,$3C,$76,$FE,$14,$DD,$BE,$E5,$6E,$73,$76
	.dc.b	$73,$7D,$48,$74,$8E,$EE,$27,$8B,$67,$8B,$B3,$07,$FA,$A2,$FC,$4F
	.dc.b	$FA,$AB,$87,$FC,$9A,$E8,$46,$9B,$7E,$07,$A1,$2F,$C2,$FC,$63,$94
	.dc.b	$F1,$39,$57,$68,$E1,$7F,$EB,$A7,$A9,$E1,$8D,$57,$42,$4E,$EB,$9E
	.dc.b	$1F,$1E,$FD,$98,$55,$87,$2B,$E1,$39,$BE,$35,$E6,$8E,$15,$B8,$E6
	.dc.b	$64,$FF,$6F,$B9,$7C,$E7,$A8,$D9,$F0,$4C,$1A,$C3,$D2,$FC,$FC,$3C
	.dc.b	$55,$E0,$6C,$7C,$16,$BC,$AB,$C1,$7F,$93,$8F,$90,$78,$8A,$E5,$F7
	.dc.b	$9F,$EF,$D7,$E8,$FD,$17,$F1,$03,$AE,$1F,$C5,$1C,$4F,$BC,$FE,$1C
	.dc.b	$E1,$F5,$CF,$27,$E1,$C7,$2C,$7F,$8B,$3B,$AF,$7C,$FC,$19,$CB,$C0
	.dc.b	$ED,$1E,$17,$07,$6C,$79,$4F,$08,$C1,$A9,$E0,$E0,$F3,$99,$8F,$82
	.dc.b	$FC,$9A,$E6,$FD,$69,$F2,$E0,$9C,$CF,$9C,$D5,$78,$6B,$B9,$EC,$C3
	.dc.b	$E1,$4F,$80,$F1,$46,$A7,$81,$E2,$7B,$D7,$6A,$FD,$26,$1E,$0C,$3C
	.dc.b	$58,$7C,$4C,$38,$BF,$B5,$BF,$9C,$D3,$DC,$79,$88,$7E,$81,$ED,$5F
	.dc.b	$9C,$6C,$7E,$B5,$AF,$C5,$39,$9F,$04,$95,$E4,$7D,$D8,$7C,$1A,$F4
	.dc.b	$57,$92,$7C,$07,$2C,$3D,$AF,$4B,$E5,$7C,$63,$15,$C6,$73,$87,$C0
	.dc.b	$9F,$15,$FC,$EB,$F9,$91,$EE,$8C,$4E,$DE,$F1,$E6,$8C,$61,$F0,$63
	.dc.b	$C8,$D7,$47,$37,$D6,$FD,$13,$E7,$1A,$3C,$6E,$9F,$BF,$1E,$7B,$F4
	.dc.b	$C7,$33,$DA,$73,$3E,$29,$CE,$BC,$F7,$EC,$8C,$BF,$15,$F0,$4F,$47
	.dc.b	$C2,$74,$3C,$27,$9A,$39,$4F,$4A,$F2,$61,$F3,$4F,$BF,$3F,$92,$AF
	.dc.b	$86,$7E,$0C,$A7,$E0,$CF,$36,$1F,$1E,$3A,$B6,$BE,$6C,$F1,$36,$59
	.dc.b	$DB,$63,$6F,$58,$CB,$63,$31,$E5,$8E,$66,$4F,$74,$7E,$AC,$8E,$08
	.dc.b	$7C,$63,$E4,$D7,$C1,$7D,$B3,$D4,$F0,$96,$7C,$27,$48,$FA,$6E,$1F
	.dc.b	$D6,$E1,$F4,$CF,$A9,$87,$CF,$8F,$0C,$63,$0F,$8D,$1E,$49,$D3,$D0
	.dc.b	$C4,$69,$C1,$E9,$BF,$C2,$57,$38,$F9,$47,$02,$9C,$2B,$AC,$79,$70
	.dc.b	$F8,$13,$F2,$AB,$DC,$F1,$3E,$61,$C6,$FC,$55,$DE,$32,$D9,$E3,$1B
	.dc.b	$3C,$31,$9C,$3E,$43,$E1,$8F,$D0,$35,$FC,$B9,$EE,$78,$3F,$32,$FF
	.dc.b	$2B,$3C,$9B,$19,$3E,$CC,$65,$A6,$4F,$73,$96,$99,$AF,$BF,$19,$6C
	.dc.b	$66,$3E,$F3,$FC,$82,$7C,$2D,$3B,$94,$EA,$59,$E6,$BE,$75,$E0,$FF
	.dc.b	$76,$8F,$F0,$F4,$61,$C1,$B8,$AD,$33,$1B,$7F,$D5,$5E,$73,$E5,$7F
	.dc.b	$D7,$AD,$80,$BF,$F6,$C5,$71,$6D,$F0,$B9,$5C,$27,$2D,$C7,$FC,$69
	.dc.b	$AB,$EA,$BC,$AD,$9F,$4B,$5F,$73,$5F,$F0,$AB,$9A,$57,$9B,$85,$9B
	.dc.b	$E4,$BD,$67,$FB,$16,$F8,$57,$8B,$1C,$E3,$E5,$3F,$DC,$77,$C5,$C0
	.dc.b	$AE,$2F,$C4,$F5,$1E,$9B,$D5,$FA,$31,$86,$FA,$AF,$47,$35,$BB,$FB
	.dc.b	$5F,$1A,$F8,$EF,$5C,$61,$AF,$F8,$D2,$35,$AF,$DE,$6C,$7E,$0F,$E4
	.dc.b	$3F,$E0,$04,$FE,$09,$B1,$9F,$23,$9B,$FE,$65,$8C,$BF,$E4,$0F,$95
	.dc.b	$7C,$F6,$A3,$F0,$31,$8B,$FC,$6B,$63,$F2,$4D,$79,$4F,$AB,$93,$D3
	.dc.b	$3B,$7D,$B4,$F9,$57,$AB,$B8,$FF,$3C,$56,$DB,$79,$C3,$C6,$D8,$E5
	.dc.b	$59,$95,$71,$7F,$33,$F3,$CF,$5F,$D9,$08,$C5,$FE,$A1,$B1,$FA,$D6
	.dc.b	$BF,$E6,$47,$39,$7D,$46,$CF,$E4,$7E,$3D,$F0,$9E,$39,$CD,$F8,$B0
	.dc.b	$C1,$EB,$EA,$5F,$0C,$3F,$1F,$7C,$2D,$F8,$D3,$E2,$BD,$2D,$AF,$DF
	.dc.b	$B6,$7F,$82,$35,$FF,$3D,$B9,$CF,$F1,$6D,$9F,$47,$57,$C9,$8C,$57
	.dc.b	$F0,$0F,$6D,$ED,$FC,$D7,$EB,$DF,$E9,$1C,$B7,$83,$53,$23,$F5,$35
	.dc.b	$FC,$2B,$DB,$FE,$17,$EC,$FF,$11,$C3,$8C,$FF,$0D,$AF,$E3,$95,$AF
	.dc.b	$0B,$FD,$75,$5A,$BF,$E3,$17,$CC,$3C,$F0,$FD,$4B,$A8,$5F,$37,$D9
	.dc.b	$3A,$BE,$F1,$C9,$EF,$19,$8F,$F5,$1C,$F1,$57,$E3,$AE,$37,$E1,$C3
	.dc.b	$55,$E5,$8C,$3D,$4E,$65,$39,$E1,$DD,$FF,$6C,$61,$DF,$0E,$D1,$C2
	.dc.b	$B9,$9E,$D9,$D5,$71,$9F,$06,$1E,$AC,$3C,$71,$F0,$8F,$06,$1D,$1E
	.dc.b	$E6,$2B,$93,$B3,$FA,$5D,$FF,$4A,$70,$7C,$83,$FD,$8F,$EB,$BC,$A7
	.dc.b	$B7,$FD,$4A,$30,$D8,$CC,$FF,$59,$9C,$B7,$DB,$FE,$9C,$DE,$0E,$51
	.dc.b	$88,$EE,$D7,$FD,$FD,$28,$E3,$AF,$F5,$DF,$6D,$ED,$FF,$5A,$BF,$A1
	.dc.b	$18,$7E,$E5,$F1,$C2,$E1,$CB,$7E,$17,$BB,$E3,$87,$3C,$7F,$B5,$63
	.dc.b	$C6,$7E,$E3,$63,$C5,$3B,$BE,$6D,$AF,$ED,$B5,$B9,$F5,$E1,$E1,$9F
	.dc.b	$EE,$38,$78,$70,$E7,$1E,$6C,$63,$E4,$3F,$2D,$3B,$BE,$73,$AA,$E3
	.dc.b	$5B,$3C,$F7,$DD,$61,$E3,$7C,$38,$F9,$D6,$3D,$5F,$4D,$71,$8F,$8D
	.dc.b	$87,$F5,$86,$EF,$F0,$2B,$FB,$5E,$1D,$5E,$78,$7D,$5B,$E0,$DA,$E1
	.dc.b	$3F,$C1,$4E,$F3,$FB,$3F,$69,$DD,$FE,$1E,$6C,$FF,$88,$9A,$FF,$89
	.dc.b	$DA,$FE,$0C,$3A,$9F,$EE,$3D,$0F,$4D,$ED,$F8,$ED,$39,$8C,$47,$2C
	.dc.b	$2E,$0F,$6A,$92,$7F,$C1,$D7,$D7,$E2,$71,$1E,$59,$F5,$11,$F5,$3E
	.dc.b	$19,$F2,$7B,$97,$23,$CB,$3F,$F0,$3E,$2F,$6F,$C6,$4C,$7C,$37,$FD
	.dc.b	$8D,$FF,$0E,$CE,$DF,$91,$3F,$2C,$8E,$0E,$3E,$E7,$84,$E8,$FD,$51
	.dc.b	$ED,$BD,$8F,$8A,$D8,$FF,$15,$B5,$E1,$EA,$BE,$6F,$75,$FE,$58,$AD
	.dc.b	$57,$D7,$6B,$FB,$B6,$BF,$E5,$67,$A6,$7F,$E2,$8B,$7B,$BF,$59,$B5
	.dc.b	$E2,$BF,$B3,$ED,$BD,$DA,$49,$F8,$FF,$25,$5F,$F3,$84,$ED,$FF,$37
	.dc.b	$B5,$CC,$FF,$9C,$23,$18,$FE,$33,$77,$BD,$F6,$4E,$91,$FE,$3E,$F7
	.dc.b	$BD,$3D,$1F,$C5,$DB,$8F,$F9,$C9,$BC,$9F,$95,$F9,$DF,$C8,$C7,$C2
	.dc.b	$93,$49,$AF,$FA,$09,$E3,$97,$F3,$79,$FC,$1F,$81,$CA,$BF,$37,$EF
	.dc.b	$FF,$B3,$07,$BF,$F2,$7D,$AC,$B2,$FF,$CC,$B3,$A7,$FC,$CF,$C5,$FE
	.dc.b	$67,$73,$3B,$7F,$0D,$3B,$7F,$03,$59,$DB,$FC,$BF,$3F,$ED,$08,$C6
	.dc.b	$DF,$EE,$98,$E1,$3F,$76,$73,$5F,$E9,$69,$E0,$F8,$0F,$F6,$3C,$69
	.dc.b	$FF,$71,$57,$F4,$3C,$3C,$D1,$FE,$CA,$AE,$B3,$9A,$E1,$1D,$27,$58
	.dc.b	$7C,$37,$B1,$FE,$D2,$BE,$78,$7A,$A7,$89,$B9,$E2,$7C,$6B,$CB,$CA
	.dc.b	$7D,$8D,$C3,$E0,$57,$19,$E9,$3E,$C3,$47,$C1,$8F,$75,$F8,$AB,$8C
	.dc.b	$E8,$46,$5E,$07,$33,$CA,$7D,$43,$86,$1F,$06,$FC,$31,$EF,$C3,$D2
	.dc.b	$FD,$CC,$3C,$D8,$78,$23,$07,$BC,$E7,$1E,$69,$E3,$5D,$0F,$83,$3E
	.dc.b	$26,$99,$AE,$8E,$1A,$7F,$7D,$BF,$B3,$3F,$ED,$8C,$3D,$CF,$E0,$0F
	.dc.b	$79,$ED,$3B,$4F,$DF,$7C,$A7,$C7,$8E,$33,$FA,$3C,$3D,$B3,$CA,$FC
	.dc.b	$8D,$BE,$4F,$43,$C5,$87,$BA,$3C,$D5,$CA,$7C,$82,$FD,$2F,$D6,$AE
	.dc.b	$0F,$53,$99,$E6,$3E,$69,$EB,$BF,$1B,$D3,$0F,$4D,$F2,$8E,$87,$19
	.dc.b	$DC,$EA,$3F,$05,$1C,$AF,$F0,$18,$7C,$E9,$F1,$1C,$8F,$71,$67,$85
	.dc.b	$FD,$D8,$ED,$3F,$01,$FC,$11,$C2,$FC,$0F,$E2,$CD,$9D,$9C,$D7,$48
	.dc.b	$E2,$D7,$E0,$BF,$83,$7C,$B7,$E8,$6D,$F8,$DF,$06,$1F,$0E,$7E,$7D
	.dc.b	$F8,$A7,$FC,$09,$1F,$C5,$5B,$5F,$E0,$77,$D4,$58,$D3,$83,$66,$CF
	.dc.b	$B9,$15,$A7,$F2,$6F,$7F,$FA,$A1,$F5,$68,$E2,$3E,$19,$FB,$87,$0F
	.dc.b	$F0,$F6,$E1,$F0,$CF,$EC,$0E,$31,$FB,$13,$C4,$F0,$1D,$2B,$BB,$83
	.dc.b	$A1,$6B,$A4,$FF,$27,$AE,$93,$F3,$0D,$46,$5E,$91,$A7,$F2,$4D,$C3
	.dc.b	$F3,$A7,$D0,$71,$87,$F1,$86,$CF,$EC,$5A,$FD,$EA,$EC,$FF,$82,$23
	.dc.b	$E4,$1A,$BF,$CD,$B6,$7E,$BC,$7C,$49,$F3,$9E,$93,$91,$83,$CE,$5B
	.dc.b	$F3,$DF,$AF,$0F,$8F,$3F,$18,$E2,$5A,$EC,$DB,$F3,$C7,$CA,$8F,$E2
	.dc.b	$98,$79,$5F,$CF,$35,$FC,$29,$D2,$39,$9C,$5F,$B4,$66,$7E,$ED,$79
	.dc.b	$27,$71,$8A,$E9,$87,$CF,$BF,$66,$1F,$3D,$A7,$CF,$29,$F3,$CF,$39
	.dc.b	$8B,$F9,$AD,$BF,$9A,$7B,$88,$FC,$D8,$F1,$BC,$8F,$88,$57,$D9,$87
	.dc.b	$A6,$78,$CF,$D2,$3D,$C7,$37,$31,$B8,$C1,$CE,$38,$DF,$D8,$8F,$B9
	.dc.b	$3B,$7A,$57,$91,$B5,$C2,$FE,$24,$78,$0C,$9D,$4E,$E7,$D2,$9E,$57
	.dc.b	$DE,$38,$5F,$78,$C4,$F7,$8C,$57,$68,$E2,$DB,$F1,$5F,$D3,$3C,$57
	.dc.b	$CA,$3E,$0E,$1F,$A0,$9F,$A4,$7E,$28,$A7,$E2,$AF,$C1,$1F,$E2,$1F
	.dc.b	$7D,$E2,$FF,$A6,$1B,$43,$FC,$F4,$E1,$FF,$58,$70,$BE,$27,$FA,$B4
	.dc.b	$6F,$06,$3F,$EE,$DB,$E1,$3F,$EA,$D6,$B9,$C3,$F9,$14,$62,$3C,$0F
	.dc.b	$7A,$FF,$58,$F9,$AF,$80,$FF,$50,$36,$BC,$18,$73,$69,$E9,$DD,$EF
	.dc.b	$FF,$AD,$1B,$1A,$0E,$1F,$6D,$3E,$6B,$85,$7F,$37,$AD,$57,$FB,$1D
	.dc.b	$AF,$C7,$39,$C6,$CF,$3F,$C0,$C8,$F0,$4F,$7C,$BE,$3C,$7C,$FF,$87
	.dc.b	$63,$FD,$15,$B1,$FD,$25,$AF,$F5,$16,$BF,$ED,$9F,$82,$7B,$EF,$82
	.dc.b	$FE,$A1,$18,$6F,$C1,$3E,$0B,$35,$BB,$FE,$B9,$F0,$2F,$82,$F2,$C7
	.dc.b	$FB,$96,$31,$C3,$FD,$6D,$B1,$C2,$76,$FC,$3F,$7D,$F0,$1E,$ED,$FD
	.dc.b	$F8,$3E,$07,$2F,$D3,$AB,$7F,$EF,$D8,$D5,$F1,$7A,$3D,$A3,$F2,$D3
	.dc.b	$C1,$FF,$7D,$35,$DC,$FF,$71,$8E,$B7,$EC,$6C,$7E,$7C,$F2,$7B,$7D
	.dc.b	$EB,$AD,$D5,$EF,$3E,$03,$5D,$8F,$80,$7C,$EE,$EB,$59,$7F,$BF,$A7
	.dc.b	$2F,$E4,$63,$38,$7F,$82,$DA,$FE,$2C,$EF,$87,$BD,$B3,$FE,$0B,$9E
	.dc.b	$3A,$7D,$66,$C7,$EE,$71,$F2,$7C,$17,$8F,$0F,$57,$C0,$BE,$0B,$D3
	.dc.b	$9B,$E3,$B3,$18,$6E,$3F,$E1,$9A,$E0,$D9,$F2,$46,$34,$FF,$12,$7C
	.dc.b	$1B,$E0,$FE,$42,$E1,$B1,$86,$E1,$FB,$39,$E0,$FE,$D6,$BF,$89,$7C
	.dc.b	$1F,$F1,$3F,$82,$FE,$25,$5A,$97,$0F,$E0,$EB,$5E,$2F,$EC,$EB,$57
	.dc.b	$FC,$42,$73,$7F,$C7,$67,$35,$FE,$3B,$9C,$CA,$72,$FD,$38,$CC,$62
	.dc.b	$3F,$80,$46,$32,$FD,$DD,$F7,$93,$E8,$8C,$DF,$79,$CB,$E7,$8C,$C7
	.dc.b	$73,$27,$9D,$CB,$E0,$32,$79,$DC,$BF,$20,$CD,$F9,$DC,$57,$8F,$1F
	.dc.b	$7C,$62,$FC,$51,$88,$61,$ED,$8C,$36,$3C,$27,$73,$FC,$BC,$FF,$97
	.dc.b	$5E,$27,$F9,$7D,$FF,$31,$7C,$03,$CD,$7C,$1E,$1B,$5F,$E6,$D9,$D4
	.dc.b	$E4,$E3,$F0,$7F,$80,$9D,$35,$FD,$44,$A3,$11,$FC,$D6,$31,$AF,$23
	.dc.b	$EF,$FC,$1B,$E0,$FF,$9D,$6B,$FD,$7E,$31,$7F,$84,$BE,$2F,$D0,$6C
	.dc.b	$EE,$B8,$EB,$FD,$0E,$F8,$DE,$F1,$F4,$5F,$F6,$BC,$38,$DE,$F0,$E3
	.dc.b	$7F,$DC,$70,$E3,$7B,$C3,$8D,$FA,$DB,$5B,$BE,$2A,$F8,$D6,$F1,$F5
	.dc.b	$5F,$0A,$D5,$F0,$BD,$C7,$F8,$07,$1F,$2B,$6B,$84,$EA,$38,$9D,$2E
	.dc.b	$3F,$94,$8D,$1B,$AE,$71,$C0,$FD,$CE,$BF,$05,$FE,$E9,$3F,$22,$BF
	.dc.b	$A8,$D7,$08,$C5,$6E,$B8,$9F,$04,$FE,$F9,$3D,$BD,$E7,$C1,$7F,$75
	.dc.b	$C3,$FA,$DC,$62,$BF,$C0,$4D,$7F,$C3,$BF,$A4,$7F,$31,$9D,$D7,$FA
	.dc.b	$DE,$B5,$5F,$CD,$5A,$FF,$80,$5A,$FF,$8A,$57,$C1,$78,$71,$D3,$67
	.dc.b	$4D,$EB,$7A,$CF,$83,$D5,$30,$E6,$FA,$2F,$31,$88,$FF,$1A,$D7,$17
	.dc.b	$47,$FB,$46,$34,$FA,$67,$D8,$7B,$7E,$0B,$99,$23,$FC,$57,$87,$93
	.dc.b	$B7,$11,$F4,$0F,$24,$E3,$4F,$F0,$DD,$6E,$7E,$6C,$F5,$7C,$4D,$7F
	.dc.b	$DD,$5F,$00,$F6,$7D,$ED,$8F,$F1,$B3,$5F,$07,$B2,$F8,$1D,$D4,$9D
	.dc.b	$BF,$22,$75,$18,$3F,$DE,$55,$CD,$FE,$DC,$D7,$FC,$74,$D7,$FC,$C5
	.dc.b	$EF,$3D,$FF,$A6,$D8,$FF,$25,$35,$FA,$3F,$01,$A9,$91,$A7,$EF,$70
	.dc.b	$27,$33,$FD,$FC,$57,$37,$FC,$04,$D7,$FC,$9C,$D7,$FC,$ED,$EF,$7C
	.dc.b	$5D,$BF,$C6,$ED,$E3,$FF,$2A,$36,$5B,$FE,$3B,$DF,$FC,$6C,$E6,$4D
	.dc.b	$38,$4E,$E3,$FC,$D1,$1F,$15,$DE,$DE,$97,$C6,$D7,$26,$7D,$EF,$F4
	.dc.b	$C3,$FE,$8D,$5E,$FF,$D7,$F6,$3F,$3F,$1A,$9F,$F4,$C6,$5F,$E6,$DE
	.dc.b	$1F,$F3,$BC,$7F,$9E,$67,$47,$27,$FC,$F9,$1A,$7F,$DB,$07,$23,$C1
	.dc.b	$7D,$DE,$4F,$80,$8F,$2B,$AF,$2A,$EC,$FE,$4A,$33,$A7,$FB,$61,$B8
	.dc.b	$7F,$BC,$9B,$3F,$EC,$6C,$3C,$95,$FE,$F9,$BE,$6F,$F9,$BA,$33,$7F
	.dc.b	$EE,$AA,$E1,$7C,$9F,$51,$27,$6F,$FB,$4A,$F9,$D7,$C7,$AE,$11,$F2
	.dc.b	$EB,$75,$CA,$31,$3A,$C3,$B5,$6A,$7E,$6C,$7C,$7A,$F8,$4E,$AB,$84
	.dc.b	$7C,$A3,$E4,$14,$D9,$82,$9F,$52,$BB,$4F,$28,$F9,$87,$1C,$3E,$63
	.dc.b	$70,$F9,$8D,$3E,$29,$4F,$64,$F9,$E7,$E7,$36,$7E,$64,$E8,$F2,$4E
	.dc.b	$9B,$1A,$C3,$E2,$C7,$56,$BE,$A9,$FF,$52,$C7,$D0,$8F,$F7,$45,$FB
	.dc.b	$1E,$17,$E1,$7E,$B3,$5F,$11,$F0,$F0,$F8,$57,$D2,$75,$7F,$4A,$BC
	.dc.b	$A7,$22,$9C,$A7,$83,$88,$E0,$FC,$C3,$D1,$87,$C2,$C3,$CD,$18,$3D
	.dc.b	$38,$78,$A7,$4D,$8F,$D4,$9D,$70,$C2,$AE,$78,$7D,$88,$E8,$7E,$5E
	.dc.b	$B7,$87,$C7,$8E,$11,$8B,$F7,$9D,$2F,$9E,$1F,$80,$AE,$77,$C6,$73
	.dc.b	$1F,$82,$33,$5C,$9F,$A2,$79,$CA,$7C,$3C,$3F,$00,$DB,$FC,$04,$FE
	.dc.b	$76,$37,$3F,$31,$D1,$F4,$4A,$7D,$19,$F8,$EF,$06,$BE,$43,$C4,$7C
	.dc.b	$4A,$F4,$BC,$4F,$A2,$71,$3E,$79,$DA,$FE,$DD,$7F,$07,$C3,$E4,$4F
	.dc.b	$48,$FA,$07,$B0,$D3,$CC,$FA,$45,$9F,$95,$3C,$A3,$B6,$1F,$46,$31
	.dc.b	$5C,$A7,$D4,$68,$E9,$7E,$AB,$FD,$2C,$F8,$70,$FE,$25,$5D,$F0,$FA
	.dc.b	$8F,$C1,$C3,$AC,$EA,$3A,$D7,$73,$93,$F1,$0C,$C6,$E3,$31,$F8,$A3
	.dc.b	$B1,$F0,$E3,$F1,$11,$AC,$3C,$D7,$F7,$6B,$99,$D6,$75,$7E,$48,$C5
	.dc.b	$FE,$11,$FE,$1D,$7C,$9B,$3B,$7C,$47,$F8,$97,$0F,$E8,$EF,$FB,$ED
	.dc.b	$A8,$FB,$A6,$70,$FE,$D1,$1F,$9D,$F7,$FF,$C3,$47,$E8,$CB,$3F,$41
	.dc.b	$AF,$E2,$DB,$8F,$E5,$C9,$8F,$F6,$98,$CB,$EB,$9F,$F1,$5C,$61,$FE
	.dc.b	$AC,$D7,$F2,$87,$EF,$5C,$57,$F4,$B6,$DF,$F4,$86,$CF,$C6,$6B,$FC
	.dc.b	$FD,$B7,$E4,$8F,$E4,$F1,$8B,$FE,$28,$D8,$FD,$0C,$FF,$65,$8C,$47
	.dc.b	$F6,$16,$DF,$C7,$8F,$F0,$0C,$7C,$76,$9F,$1C,$A7,$C7,$2C,$F8,$DE
	.dc.b	$86,$2B,$FA,$3F,$BC,$F0,$3F,$CE,$6D,$9F,$B5,$4A,$FB,$90,$9D,$3F
	.dc.b	$E9,$6D,$9E,$FB,$2E,$78,$CA,$E7,$87,$2C,$38,$E3,$FE,$6C,$C3,$85
	.dc.b	$68,$FC,$54,$F6,$7C,$27,$86,$FF,$C6,$51,$98,$E9,$1F,$E3,$5D,$1E
	.dc.b	$EF,$FA,$39,$B3,$D4,$38,$7D,$5E,$E3,$DE,$E1,$52,$78,$56,$F2,$FF
	.dc.b	$41,$DE,$E7,$35,$E1,$23,$FE,$49,$AE,$CF,$AB,$DA,$7C,$1F,$F9,$2D
	.dc.b	$B1,$FE,$4F,$6B,$FE,$69,$6B,$F0,$BD,$E6,$1F,$E5,$36,$C7,$F9,$5D
	.dc.b	$AF,$F9,$C5,$AF,$CA,$F7,$9E,$E3,$DE,$FF,$2F,$46,$1B,$83,$DE,$CD
	.dc.b	$6E,$FF,$CF,$9E,$DB,$DB,$FC,$34,$7F,$B6,$39,$1E,$E6,$5F,$A1,$56
	.dc.b	$75,$37,$87,$8C,$27,$15,$FC,$2C,$E7,$7D,$AB,$73,$AA,$FC,$B5,$D7
	.dc.b	$95,$74,$7F,$33,$3E,$17,$F1,$6D,$73,$1B,$7F,$8D,$7B,$6F,$1B,$C1
	.dc.b	$8F,$F1,$DE,$97,$B4,$F7,$33,$87,$0B,$B5,$A9,$CC,$B0,$E9,$87,$FB
	.dc.b	$4B,$0E,$4F,$E5,$63,$33,$FA,$D9,$CB,$FE,$D5,$8C,$BF,$A4,$6B,$FB
	.dc.b	$66,$BE,$39,$FF,$4D,$7B,$2F,$6B,$CB,$6F,$6F,$CB,$39,$AF,$24,$62
	.dc.b	$B6,$D8,$C9,$FA,$BF,$65,$ED,$7B,$97,$B7,$63,$0D,$9E,$11,$F1,$32
	.dc.b	$FE,$91,$ED,$FF,$49,$F6,$7F,$86,$E1,$D2,$FF,$71,$5A,$13,$AB,$FC
	.dc.b	$8B,$FC,$0E,$F9,$78,$1F,$C6,$30,$E5,$7C,$23,$FD,$A9,$3A,$8F,$EB
	.dc.b	$13,$AA,$E5,$1B,$33,$7D,$AF,$8C,$64,$57,$18,$CD,$79,$70,$ED,$5F
	.dc.b	$CA,$B0,$B7,$B8,$FE,$3B,$18,$C3,$B4,$60,$ED,$3B,$79,$18,$8D,$B8
	.dc.b	$BF,$03,$E5,$3E,$43,$83,$DA,$5C,$3C,$E7,$C7,$9E,$73,$F3,$CF,$F7
	.dc.b	$15,$FF,$B8,$53,$4F,$F7,$37,$AC,$F6,$DE,$DE,$1B,$5F,$CE,$84,$EE
	.dc.b	$74,$72,$F6,$9D,$34,$95,$C6,$93,$A8,$C3,$73,$FE,$AB,$ED,$3D,$BF
	.dc.b	$EB,$31,$86,$EB,$FD,$73,$0E,$78,$FD,$66,$E3,$F7,$70,$E4,$FD,$A6
	.dc.b	$BF,$81,$8D,$E1,$FD,$91,$B7,$FD,$A2,$B7,$3E,$1A,$ED,$8F,$9E,$BC
	.dc.b	$4D,$8F,$F1,$55,$70,$9E,$58,$FA,$27,$FC,$5D,$87,$66,$DF,$15,$7C
	.dc.b	$5E,$75,$C5,$FA,$33,$BA,$E5,$1C,$27,$55,$C6,$37,$8F,$A0,$F8,$53
	.dc.b	$A3,$A5,$72,$AF,$8B,$3F,$CF,$1B,$BF,$AA,$FF,$B9,$56,$EF,$FC,$1B
	.dc.b	$7B,$6D,$6E,$BF,$06,$FC,$69,$FD,$5F,$EE,$BF,$C4,$27,$FC,$11,$D0
	.dc.b	$F7,$31,$1F,$E1,$56,$D7,$F4,$EA,$B7,$FE,$1C,$6D,$2F,$97,$0F,$8A
	.dc.b	$31,$86,$95,$53,$F4,$84,$7F,$3C,$7F,$98,$DC,$3E,$68,$FE,$D8,$DE
	.dc.b	$FE,$CF,$42,$CF,$F8,$83,$1F,$ED,$B8,$7B,$6B,$C8,$DA,$FF,$0E,$56
	.dc.b	$F1,$FF,$1B,$FC,$D3,$E4,$3C,$74,$8C,$3E,$66,$BF,$B5,$6B,$FE,$30
	.dc.b	$73,$8F,$8B,$A9,$E1,$78,$9B,$5F,$E2,$DA,$DE,$1E,$2C,$35,$72,$B5
	.dc.b	$1F,$03,$DC,$7B,$7E,$66,$CF,$CA,$6B,$FC,$25,$AF,$F9,$11,$CE,$5F
	.dc.b	$1B,$D8,$7C,$37,$C6,$6D,$7F,$90,$6B,$79,$7F,$97,$1B,$1F,$81,$9D
	.dc.b	$BF,$E5,$D6,$DF,$F8,$5A,$77,$2F,$61,$F3,$3F,$39,$B1,$F8,$26,$BF
	.dc.b	$C8,$5A,$FF,$97,$1C,$D7,$A7,$E0,$3D,$07,$E1,$23,$38,$FF,$9C,$9B
	.dc.b	$5E,$79,$DB,$FE,$5D,$8C,$AF,$65,$F0,$FF,$82,$6C,$7E,$59,$AF,$F3
	.dc.b	$56,$BF,$E7,$07,$38,$7F,$8D,$BD,$97,$BF,$F9,$86,$D7,$F9,$B6,$B7
	.dc.b	$97,$FA,$21,$B1,$FE,$C5,$9D,$BF,$BC,$31,$5F,$A2,$9D,$C3,$DF,$7B
	.dc.b	$9F,$A6,$6C,$7F,$9E,$DA,$F3,$17,$DB,$1F,$F2,$6C,$E9,$B5,$CC,$F5
	.dc.b	$61,$6F,$A4,$57,$A5,$F3,$C7,$FD,$47,$87,$36,$D6,$F0,$F8,$D8,$59
	.dc.b	$CE,$1D,$0F,$F4,$96,$1C,$DF,$F7,$46,$1C,$DB,$87,$17,$E2,$9C,$4F
	.dc.b	$74,$6D,$F8,$D1,$EA,$75,$87,$B9,$B7,$FE,$BA,$BF,$04,$70,$32,$78
	.dc.b	$0F,$2E,$1E,$17,$47,$82,$FB,$34,$E2,$6A,$FE,$0B,$8C,$3C,$C7,$9E
	.dc.b	$FC,$2E,$2B,$B4,$61,$F0,$46,$EE,$C6,$9B,$5E,$0A,$E0,$F7,$9F,$2D
	.dc.b	$FC,$18,$F3,$11,$F2,$B7,$0F,$AA,$D3,$E7,$94,$F6,$3A,$8E,$27,$52
	.dc.b	$39,$3E,$75,$FE,$26,$7F,$DB,$38,$7E,$59,$B7,$FC,$D2,$31,$1C,$4F
	.dc.b	$D3,$65,$FA,$D3,$DE,$FE,$49,$C3,$EF,$3F,$16,$E1,$F4,$47,$CF,$8C
	.dc.b	$61,$E6,$9F,$BC,$73,$38,$9C,$CC,$11,$E0,$FC,$23,$CA,$70,$3B,$D7
	.dc.b	$6C,$3F,$1D,$5E,$8A,$FE,$10,$E6,$BC,$31,$D3,$1F,$DF,$9F,$C0,$30
	.dc.b	$FC,$24,$F5,$9E,$2F,$95,$AF,$96,$FE,$43,$E5,$3F,$80,$DF,$77,$15
	.dc.b	$F0,$DE,$93,$C2,$31,$1C,$CF,$2C,$74,$BF,$2D,$7A,$E3,$DB,$3C,$8D
	.dc.b	$90,$F0,$99,$3C,$38,$78,$EB,$DC,$FE,$82,$BC,$D7,$F0,$2B,$C4,$72
	.dc.b	$3E,$69,$CE,$B8,$BF,$A1,$9E,$35,$CE,$3C,$15,$CF,$0F,$8E,$4C,$3E
	.dc.b	$05,$78,$EF,$D9,$7F,$70,$F0,$DF,$1A,$F7,$C6,$2B,$B3,$4F,$11,$E0
	.dc.b	$73,$7D,$E3,$E2,$57,$CE,$8C,$47,$C4,$3D,$97,$E2,$7E,$04,$6A,$7C
	.dc.b	$51,$E5,$8C,$57,$6B,$F2,$C7,$E5,$AB,$B4,$60,$F7,$D7,$63,$F1,$4E
	.dc.b	$5E,$8E,$9F,$A3,$87,$CE,$7F,$81,$3F,$82,$78,$9C,$08,$7A,$CE,$38
	.dc.b	$7A,$E3,$84,$7D,$A3,$27,$23,$DF,$3E,$37,$95,$FC,$69,$F1,$3F,$A7
	.dc.b	$AF,$BF,$3F,$40,$FA,$47,$99,$CB,$4F,$A8,$6E,$3E,$4B,$6B,$C1,$7F
	.dc.b	$3A,$3C,$67,$B1,$C5,$FA,$AB,$E7,$61,$F8,$9A,$F5,$4E,$6F,$E5,$57
	.dc.b	$A7,$0F,$84,$79,$1F,$55,$79,$0F,$31,$EA,$72,$D9,$FD,$94,$F1,$6C
	.dc.b	$71,$9D,$9E,$C7,$9B,$4C,$C7,$BA,$FF,$11,$3E,$D6,$BF,$00,$FA,$31
	.dc.b	$F4,$EF,$E8,$BE,$08,$F2,$1F,$BF,$8F,$80,$FE,$7A,$7E,$23,$F7,$EB
	.dc.b	$E1,$3F,$1A,$FD,$4F,$16,$BF,$C6,$6B,$9B,$E2,$3F,$1A,$F8,$9E,$07
	.dc.b	$C1,$BF,$25,$7B,$DA,$7B,$E3,$EF,$1E,$F2,$9E,$FC,$3E,$ED,$F8,$9F
	.dc.b	$D4,$C6,$CF,$97,$3C,$DA,$66,$BE,$6B,$63,$C2,$D3,$CE,$7C,$83,$DD
	.dc.b	$5D,$F0,$F8,$F5,$F8,$C9,$FD,$B1,$F2,$6B,$E9,$57,$E2,$A3,$E0,$1D
	.dc.b	$27,$E0,$21,$EC,$3E,$17,$FD,$42,$D8,$FF,$5A,$B5,$E0,$47,$0E,$0E
	.dc.b	$71,$FE,$B0,$F6,$98,$53,$FD,$73,$5C,$EF,$37,$E3,$8C,$4F,$97,$A9
	.dc.b	$7A,$AF,$F3,$14,$62,$79,$E1,$FE,$C4,$FC,$5D,$F2,$9F,$EB,$A6,$DF
	.dc.b	$FB,$39,$AF,$9C,$38,$7D,$4E,$B2,$FF,$65,$FB,$1F,$E4,$FC,$4F,$F9
	.dc.b	$2E,$75,$8F,$C4,$6D,$FF,$B9,$67,$81,$BC,$3F,$DB,$3E,$EB,$DF,$FF
	.dc.b	$67,$B6,$BF,$DC,$6D,$7C,$CD,$7E,$73,$9C,$BF,$D1,$DE,$FF,$FA,$34
	.dc.b	$EB,$52,$BF,$CD,$31,$97,$F0,$13,$F2,$EB,$96,$5F,$73,$DA,$7B,$2F
	.dc.b	$6B,$FD,$EB,$18,$6F,$81,$FE,$F4,$C7,$F0,$13,$97,$E6,$7C,$3B,$E1
	.dc.b	$FE,$C4,$7E,$22,$31,$8F,$FC,$00,$6C,$70,$9D,$BF,$95,$F8,$B7,$6E
	.dc.b	$21,$18,$BF,$B5,$18,$9D,$4E,$63,$EF,$7E,$38,$F2,$39,$36,$F8,$B6
	.dc.b	$77,$3A,$3F,$2B,$BF,$E4,$67,$4F,$FB,$F7,$ED,$B3,$7F,$A4,$9C,$B6
	.dc.b	$B3,$7F,$AF,$9C,$AE,$67,$90,$C3,$72,$AD,$AF,$65,$C9,$F0,$DB,$97
	.dc.b	$C2,$6C,$FE,$75,$AF,$ED,$5C,$E7,$FA,$B6,$CF,$EB,$F1,$FF,$7A,$BF
	.dc.b	$86,$3F,$7F,$1A,$AF,$E1,$1E,$B3,$CD,$F3,$AF,$5E,$C6,$1B,$7C,$EB
	.dc.b	$8C,$FB,$9F,$E2,$5E,$BF,$F1,$3F,$55,$EA,$FF,$0E,$C7,$B6,$1C,$6F
	.dc.b	$F8,$8D,$EE,$B5,$87,$F0,$AC,$38,$D7,$F1,$3A,$D3,$E0,$BA,$F2,$13
	.dc.b	$C1,$61,$E0,$73,$87,$F0,$7B,$E3,$39,$AE,$15,$EA,$36,$77,$29,$DD
	.dc.b	$ED,$1C,$6B,$A4,$EA,$30,$FA,$4C,$61,$ED,$9D,$D7,$28,$C4,$EE,$31
	.dc.b	$87,$BA,$3C,$91,$87,$C6,$62,$BE,$4C,$EE,$18,$7B,$AF,$AC,$7A,$6B
	.dc.b	$6F,$23,$67,$EC,$8A,$7E,$D2,$70,$E0,$FE,$93,$E9,$BD,$77,$AF,$FB
	.dc.b	$A8,$C3,$63,$33,$FC,$DE,$72,$DF,$5D,$C1,$C5,$B1,$98,$C1,$1F,$A7
	.dc.b	$EB,$BD,$7F,$EB,$7B,$7E,$06,$31,$3F,$86,$C1,$87,$4A,$FC,$34,$63
	.dc.b	$0F,$C5,$DF,$0B,$EB,$87,$4D,$3D,$F8,$F1,$6D,$F0,$9D,$57,$F7,$3B
	.dc.b	$E5,$8F,$6C,$38,$4F,$F7,$7C,$7C,$0F,$63,$91,$E7,$8D,$A8,$DC,$FB
	.dc.b	$6F,$BC,$AF,$BE,$3E,$48,$FC,$BD,$F0,$8F,$EF,$F3,$A9,$F3,$57,$2A
	.dc.b	$D4,$79,$70,$F0,$D6,$E3,$B5,$6E,$72,$FF,$82,$B1,$FE,$F9,$5F,$25
	.dc.b	$B1,$9D,$BC,$99,$7A,$70,$FE,$05,$7B,$9F,$2D,$6E,$7C,$57,$FB,$EF
	.dc.b	$5B,$F8,$B4,$7F,$46,$6B,$FE,$01,$6B,$FE,$19,$6B,$FE,$1E,$77,$1F
	.dc.b	$E1,$6E,$77,$AF,$8C,$3F,$C2,$CD,$BF,$F1,$65,$D4,$FF,$89,$6F,$FC
	.dc.b	$34,$DA,$ED,$3E,$9B,$EB,$C7,$EB,$8E,$27,$EB,$6B,$F5,$64,$70,$71
	.dc.b	$C3,$FB,$A6,$1E,$76,$E1,$DA,$7F,$95,$7A,$EF,$5F,$B3,$63,$31,$F2
	.dc.b	$FF,$14,$7E,$88,$F8,$4D,$CF,$EF,$EF,$85,$FF,$8A,$DA,$FF,$90,$DA
	.dc.b	$FF,$92,$1D,$E5,$F2,$FA,$9E,$27,$C1,$6D,$78,$8F,$E8,$1E,$BB,$DA
	.dc.b	$F8,$8D,$8C,$C7,$C9,$F6,$5E,$FE,$A3,$F8,$A5,$6E,$BF,$C8,$0D,$7F
	.dc.b	$CB,$0D,$7F,$CB,$AE,$F1,$FC,$05,$BD,$BE,$C7,$5A,$FE,$A9,$ED,$3D
	.dc.b	$BF,$B0,$D8,$CC,$79,$7E,$7F,$2F,$F2,$79,$D4,$FF,$94,$DA,$FF,$9A
	.dc.b	$DA,$FF,$9C,$1D,$E1,$E4,$F2,$7E,$01,$D2,$7F,$B3,$7B,$8F,$63,$F1
	.dc.b	$4D,$8C,$C7,$06,$9F,$E2,$A8,$D3,$FE,$65,$6B,$FE,$7D,$6B,$FE,$82
	.dc.b	$77,$5F,$E7,$5F,$6D,$EE,$7E,$65,$B5,$D8,$FF,$4B,$1F,$EC,$38,$FC
	.dc.b	$F7,$B9,$FE,$EB,$1E,$E7,$18,$CB,$27,$FC,$05,$3A,$7F,$C0,$95,$D9
	.dc.b	$EA,$7F,$B1,$6B,$83,$88,$D5,$FF,$AA,$63,$FA,$4F,$57,$FA,$72,$FA
	.dc.b	$46,$0F,$F5,$35,$70,$9F,$F7,$B9,$BC,$3C,$D3,$E4,$7B,$DF,$FA,$5E
	.dc.b	$F9,$BF,$17,$0C,$DF,$2B,$EB,$18,$7F,$DC,$87,$B1,$C3,$FE,$DD,$18
	.dc.b	$78,$DD,$9E,$73,$E0,$1E,$32,$19,$BF,$1D,$7F,$AE,$A7,$94,$F5,$9C
	.dc.b	$BE,$43,$8D,$FC,$73,$D4,$E2,$BD,$4F,$92,$FC,$F5,$CA,$3C,$67,$D4
	.dc.b	$71,$7E,$07,$F0,$18,$7C,$A9,$D3,$70,$F0,$61,$F3,$E7,$E2,$1E,$13
	.dc.b	$E0,$C7,$91,$E2,$68,$A7,$C5,$BF,$05,$79,$8E,$D5,$EA,$BF,$2E,$19
	.dc.b	$BE,$C7,$B4,$B1,$ED,$36,$E2,$BE,$BD,$F6,$8C,$1E,$3C,$3D,$6F,$E8
	.dc.b	$4C,$9F,$56,$7C,$6F,$B3,$0F,$80,$FA,$70,$F2,$3B,$3A,$47,$A6,$78
	.dc.b	$C7,$DB,$6E,$1F,$3A,$36,$75,$3C,$47,$C6,$BF,$94,$79,$9F,$B2,$D8
	.dc.b	$F7,$C6,$EB,$F1,$57,$E4,$AF,$D0,$5F,$80,$C4,$EA,$BD,$F3,$A7,$C6
	.dc.b	$79,$CB,$7E,$77,$E0,$1C,$27,$E1,$CF,$99,$F8,$11,$C2,$BB,$BF,$99
	.dc.b	$3C,$C5,$9F,$54,$79,$63,$A1,$99,$E6,$F8,$0E,$27,$E0,$4B,$87,$DD
	.dc.b	$7F,$86,$61,$E0,$7E,$3C,$FD,$84,$52,$18,$C3,$F8,$7B,$A3,$17,$FE
	.dc.b	$6B,$8C,$C7,$00,$E1,$D9,$FC,$8B,$2F,$E6,$07,$8F,$0F,$E6,$B7,$FB
	.dc.b	$F8,$C1,$A9,$F8,$27,$F0,$C7,$0F,$F2,$76,$E3,$F8,$3A,$E7,$3F,$C9
	.dc.b	$27,$9B,$E1,$39,$C6,$9E,$86,$23,$9D,$7F,$08,$9D,$1B,$BF,$D6,$61
	.dc.b	$AB,$EC,$71,$3B,$1E,$52,$E1,$E5,$C3,$84,$7F,$27,$BE,$B1,$E6,$8F
	.dc.b	$E0,$4D,$9F,$EB,$0D,$C3,$E3,$4F,$73,$65,$9D,$B6,$36,$F8,$CD,$96
	.dc.b	$36,$D8,$DC,$79,$E3,$D3,$1F,$0E,$7D,$38,$7C,$29,$C9,$F0,$8E,$45
	.dc.b	$AE,$4F,$81,$A7,$80,$B5,$E0,$3B,$47,$43,$E3,$57,$83,$0F,$E3,$2D
	.dc.b	$8F,$E4,$38,$7C,$A7,$E0,$1E,$1A,$E6,$F8,$5B,$7E,$16,$9E,$18,$FE
	.dc.b	$20,$7E,$10,$E0,$62,$7B,$3E,$E3,$53,$EA,$7C,$06,$EB,$CC,$FF,$23
	.dc.b	$39,$9C,$2B,$C2,$F0,$3D,$E7,$C1,$C3,$E7,$3C,$2F,$F2,$58,$7C,$28
	.dc.b	$D3,$5E,$B3,$C3,$0F,$74,$75,$8C,$5F,$BE,$3C,$47,$9C,$E6,$7C,$CC
	.dc.b	$3D,$35,$FC,$8E,$3C,$51,$F9,$37,$A5,$F8,$67,$BC,$7A,$8F,$1C,$7F
	.dc.b	$A1,$FD,$C7,$B7,$49,$3F,$C4,$AB,$FD,$35,$EE,$3E,$B3,$33,$FC,$44
	.dc.b	$FE,$2F,$7C,$9C,$18,$BF,$F7,$F4,$66,$32,$DB,$EB,$97,$FA,$6F,$03
	.dc.b	$DA,$2D,$7F,$24,$F7,$5E,$E7,$FA,$7D,$B1,$E7,$AF,$E4,$91,$88,$FF
	.dc.b	$1A,$78,$8F,$75,$E9,$71,$C3,$CF,$5D,$5A,$AF,$93,$EF,$6B,$DA,$3F
	.dc.b	$D6,$55,$FE,$22,$BE,$DB,$75,$AF,$21,$FC,$F3,$DC,$7B,$BF,$EB,$76
	.dc.b	$C6,$83,$87,$FC,$25,$E1,$EE,$7C,$B3,$99,$FA,$4D,$7F,$A2,$35,$F0
	.dc.b	$47,$F9,$B7,$39,$1D,$E7,$B6,$BF,$3A,$1F,$0B,$F1,$DF,$8A,$D7,$EE
	.dc.b	$B5,$FC,$13,$5F,$EA,$9E,$E3,$DE,$7B,$DF,$3A,$30,$DF,$15,$ED,$E6
	.dc.b	$B7,$7F,$8A,$F7,$9E,$F7,$96,$3F,$B0,$46,$39,$7E,$CB,$63,$84,$ED
	.dc.b	$FE,$D7,$EE,$76,$AF,$A7,$C6,$F9,$4F,$2C,$7D,$19,$D5,$7E,$22,$72
	.dc.b	$FF,$99,$E3,$2F,$E8,$5A,$FF,$73,$6B,$E0,$CB,$C4,$A3,$AE,$FF,$3D
	.dc.b	$B7,$FE,$00,$C1,$E5,$3D,$EF,$D0,$DE,$F1,$B7,$FE,$01,$F7,$BF,$36
	.dc.b	$FB,$2D,$F0,$5A,$CF,$FC,$17,$5A,$7F,$47,$3A,$AF,$C8,$4E,$67,$F7
	.dc.b	$8D,$7B,$72,$FA,$A3,$11,$FE,$16,$F7,$DE,$FF,$78,$43,$D6,$C3,$AC
	.dc.b	$3F,$3C,$E5,$B5,$A3,$95,$7F,$88,$BD,$FF,$F1,$1F,$B4,$F7,$7F,$C4
	.dc.b	$18,$73,$AF,$F4,$D1,$B1,$3B,$7C,$D7,$FC,$2A,$31,$1F,$E2,$9B,$EB
	.dc.b	$7D,$2B,$4F,$F1,$A8,$CC,$F2,$C2,$BC,$4E,$F3,$C6,$33,$1E,$3B,$EF
	.dc.b	$87,$9D,$F0,$C6,$6B,$83,$C8,$FE,$59,$7D,$AB,$84,$F8,$E7,$47,$CE
	.dc.b	$C3,$CE,$F0,$33,$3C,$1C,$5F,$8A,$77,$1E,$98,$C6,$1E,$88,$61,$E1
	.dc.b	$AD,$C7,$9C,$F8,$47,$88,$C1,$E3,$9E,$53,$FC,$AA,$39,$9F,$CB,$1C
	.dc.b	$1F,$CC,$3D,$D7,$BC,$F7,$F0,$DA,$FE,$97,$3A,$9C,$9C,$7D,$F7,$1D
	.dc.b	$25,$70,$8C,$B6,$3E,$CF,$BE,$F7,$FF,$CD,$5A,$FE,$12,$31,$5F,$7B
	.dc.b	$16,$1C,$AB,$F1,$31,$8C,$39,$E1,$CB,$7F,$F3,$A6,$3C,$B1,$FA,$F7
	.dc.b	$B6,$E3,$CE,$B7,$7C,$67,$FD,$03,$7C,$71,$EF,$7B,$9F,$F4,$2E,$1C
	.dc.b	$E7,$D1,$7C,$65,$7C,$70,$FF,$43,$5E,$DB,$87,$3A,$DD,$F1,$BE,$11
	.dc.b	$DA,$75,$7C,$63,$CB,$5B,$3C,$11,$98,$36,$B9,$C7,$63,$E6,$57,$86
	.dc.b	$3E,$36,$3F,$D4,$5B,$AF,$C4,$AF,$EC,$97,$CF,$0F,$F5,$05,$ED,$B5
	.dc.b	$BA,$FC,$1B,$EE,$9F,$DB,$FB,$FF,$C5,$A7,$FB,$73,$5D,$9F,$EB,$07
	.dc.b	$11,$FE,$10,$E8,$7B,$EF,$84,$EF,$7B,$E2,$FF,$03,$E1,$CE,$3F,$C4
	.dc.b	$8D,$7F,$D6,$CD,$7B,$1C,$83,$FE,$2B,$C3,$FB,$93,$71,$FE,$89,$87
	.dc.b	$F2,$BA,$F4,$9C,$67,$4F,$43,$FA,$EC,$6D,$F6,$9D,$8F,$39,$0F,$F1
	.dc.b	$0C,$7E,$DF,$DD,$7B,$DE,$56,$CF,$F8,$55,$AF,$C8,$F8,$07,$C0,$3C
	.dc.b	$AE,$17,$2B,$83,$F2,$2B,$73,$99,$FF,$23,$DF,$47,$FC,$98,$D7,$FD
	.dc.b	$BC,$D7,$FD,$CB,$EF,$3D,$FF,$94,$D8,$FF,$15,$35,$FF,$20,$FB,$EF
	.dc.b	$80,$DD,$49,$DB,$F7,$38,$95,$B9,$C9,$FE,$46,$BE,$8F,$F9,$85,$AF
	.dc.b	$FB,$CD,$AF,$FB,$EB,$DD,$7B,$DF,$7D,$B1,$FE,$3F,$6B,$E6,$F7,$F5
	.dc.b	$32,$34,$F9,$6D,$39,$9F,$F3,$7D,$72,$7F,$CE,$8D,$47,$8E,$BF,$CB
	.dc.b	$BE,$F7,$FB,$BC,$9E,$DF,$5F,$7F,$F2,$91,$8A,$FF,$2A,$3F,$EF,$AC
	.dc.b	$FE,$2C,$EA,$B9,$92,$BF,$CD,$74,$8F,$F3,$5F,$BB,$FE,$9E,$9F,$1A
	.dc.b	$F5,$BF,$45,$EF,$3D,$FF,$D2,$4E,$E8,$ED,$FF,$33,$4E,$DF,$F3,$44
	.dc.b	$6C,$D3,$FA,$08,$D3,$FE,$BF,$EA,$FF,$54,$13,$AB,$E2,$C2,$77,$39
	.dc.b	$8F,$F6,$75,$F9,$30,$91,$FE,$DE,$8A,$F5,$BE,$78,$78,$AF,$FC,$D9
	.dc.b	$87,$49,$CE,$1C,$27,$B5,$F8,$70,$C5,$6A,$3B,$1E,$EA,$CB,$F2,$9F
	.dc.b	$93,$38,$8F,$8F,$16,$FC,$11,$FE,$EA,$C3,$C9,$7F,$02,$73,$3C,$0E
	.dc.b	$C7,$80,$F6,$1F,$08,$C9,$83,$81,$F3,$EF,$C5,$5F,$66,$7A,$BF,$56
	.dc.b	$FC,$F5,$CB,$0F,$83,$3E,$48,$FC,$BD,$70,$AE,$67,$08,$F0,$4F,$5B
	.dc.b	$CC,$79,$EF,$9D,$7C,$19,$CC,$F9,$EF,$A3,$E1,$C3,$E7,$3E,$93,$07
	.dc.b	$BC,$EF,$5C,$DC,$9F,$02,$FE,$CD,$74,$8D,$9F,$96,$BF,$0C,$E9,$B7
	.dc.b	$E0,$8F,$C2,$C6,$27,$EE,$1C,$4E,$F8,$7B,$DB,$87,$8D,$EE,$7B,$67
	.dc.b	$93,$E5,$BF,$BD,$18,$AF,$9D,$3F,$8F,$39,$9F,$26,$7C,$84,$7C,$67
	.dc.b	$C5,$3B,$DF,$AD,$E2,$78,$4A,$E5,$DC,$73,$7C,$06,$0E,$B1,$F0,$9E
	.dc.b	$31,$BC,$3E,$64,$FD,$C9,$F8,$87,$13,$DC,$7E,$22,$FD,$D8,$7E,$1E
	.dc.b	$78,$4F,$94,$E1,$87,$C2,$BE,$B8,$79,$63,$18,$72,$AE,$F7,$C8,$F6
	.dc.b	$16,$BC,$D3,$B9,$F8,$D3,$97,$DB,$3F,$49,$F6,$98,$3D,$06,$0F,$59
	.dc.b	$F1,$A3,$93,$8B,$F3,$BF,$95,$8C,$DF,$A2,$39,$35,$F7,$E1,$F0,$EF
	.dc.b	$DC,$FE,$26,$B8,$B9,$9F,$E0,$CD,$7C,$75,$F9,$17,$99,$F1,$2B,$9B
	.dc.b	$F8,$93,$89,$92,$9F,$11,$DB,$F4,$EB,$C2,$EC,$EF,$5E,$59,$DC,$FD
	.dc.b	$06,$9F,$42,$7D,$F3,$E7,$AF,$E1,$D1,$CD,$AF,$84,$F1,$D7,$E3,$DF
	.dc.b	$59,$93,$C2,$7A,$CA,$79,$EB,$C2,$FD,$93,$EB,$38,$7E,$B6,$1F,$83
	.dc.b	$9E,$C7,$BA,$3B,$57,$92,$33,$5C,$63,$EC,$46,$CC,$CF,$A2,$7F,$14
	.dc.b	$FA,$23,$2F,$99,$F8,$CE,$CF,$C6,$9B,$8E,$31,$D5,$A7,$86,$FE,$51
	.dc.b	$FD,$E0,$93,$FD,$98,$FE,$4E,$5A,$FD,$56,$3F,$DE,$8F,$E2,$38,$FF
	.dc.b	$80,$67,$FB,$BC,$60,$D3,$F9,$F8,$D3,$FD,$DE,$34,$F0,$34,$F0,$7F
	.dc.b	$80,$47,$0C,$3F,$B6,$CF,$09,$E4,$70,$39,$1C,$1E,$51,$FA,$C9,$E5
	.dc.b	$87,$F6,$39,$E5,$3D,$2B,$91,$D0,$E4,$74,$79,$47,$49,$FC,$15,$74
	.dc.b	$AF,$EB,$73,$D2,$7B,$47,$4A,$EC,$F4,$3B,$3D,$1E,$D1,$D2,$7B,$47
	.dc.b	$E4,$7D,$D7,$BD,$FE,$8B,$6C,$7F,$90,$2D,$7F,$E1,$E9,$CC,$FF,$A8
	.dc.b	$36,$7B,$CF,$7F,$96,$32,$B9,$61,$C6,$3F,$CC,$D5,$B9,$C9,$FE,$4E
	.dc.b	$C3,$AB,$F8,$76,$BC,$4F,$C5,$1B,$9A,$FF,$95,$BD,$F7,$BF,$FE,$A5
	.dc.b	$6C,$76,$0E,$1F,$57,$BE,$F4,$38,$54,$9E,$15,$BD,$7F,$D7,$97,$B9
	.dc.b	$CD,$78,$9E,$AD,$AE,$D1,$FA,$16,$BF,$E4,$8F,$79,$EF,$FF,$AE,$DB
	.dc.b	$1E,$76,$BF,$E5,$3E,$FD,$4C,$8D,$3F,$E6,$28,$C4,$7C,$69,$E2,$FE
	.dc.b	$C5,$AF,$86,$7D,$1E,$E3,$E1,$E9,$27,$E6,$C6,$5F,$9F,$F0,$0C,$FE
	.dc.b	$03,$63,$31,$83,$E2,$7C,$03,$DF,$7C,$07,$D8,$8C,$37,$07,$B7,$9A
	.dc.b	$DE,$1F,$80,$F7,$73,$CD,$F8,$8E,$87,$BD,$D2,$3E,$F4,$EA,$BF,$2D
	.dc.b	$39,$7F,$DB,$71,$9E,$4F,$F7,$B4,$E5,$FC,$FC,$FF,$A8,$27,$55,$FE
	.dc.b	$A2,$BE,$F9,$EA,$F7,$6B,$F0,$3F,$C7,$DA,$FF,$A1,$9A,$F5,$CB,$8B
	.dc.b	$FE,$8A,$F6,$FF,$25,$8F,$B7,$DC,$7B,$BD,$E3,$5D,$16,$DA,$2F,$75
	.dc.b	$F9,$FC,$39,$D7,$ED,$E7,$2F,$FB,$F2,$33,$9F,$C0,$BE,$58,$F5,$A7
	.dc.b	$B8,$F4,$B0,$BD,$DB,$18,$6E,$7F,$AF,$9E,$0F,$EC,$6B,$FD,$97,$EE
	.dc.b	$FF,$B3,$3D,$B7,$B7,$FE,$AA,$BD,$D3,$4F,$8A,$AF,$85,$6F,$2F,$AF
	.dc.b	$87,$2A,$FE,$9D,$5A,$7F,$DB,$38,$74,$9D,$47,$FA,$E2,$74,$FB,$AB
	.dc.b	$9D,$FF,$0E,$9C,$DF,$18,$C5,$2F,$9C,$E6,$3C,$26,$67,$CD,$8F,$68
	.dc.b	$F8,$26,$4F,$2E,$1E,$07,$8D,$F4,$8C,$47,$9C,$E0,$79,$E7,$A3,$F1
	.dc.b	$4C,$DF,$8D,$E1,$5C,$E7,$78,$7B,$2F,$AD,$79,$2B,$4F,$63,$18,$7A
	.dc.b	$9C,$1D,$CF,$E6,$35,$FC,$85,$C1,$F1,$8F,$F0,$07,$B2,$F7,$5E,$EF
	.dc.b	$F3,$88,$C3,$63,$33,$FC,$FA,$72,$DF,$77,$F0,$8D,$D7,$94,$62,$3B
	.dc.b	$B5,$FF,$7A,$CA,$38,$E5,$F8,$2F,$75,$EE,$FF,$BE,$A3,$0D,$D3,$8D
	.dc.b	$70,$6E,$3F,$88,$BE,$38,$DB,$E0,$86,$3E,$1A,$FC,$7B,$63,$ED,$D7
	.dc.b	$26,$D7,$27,$8E,$1E,$17,$A5,$71,$C7,$CF,$1E,$13,$AC,$ED,$E3,$86
	.dc.b	$5E,$A6,$98,$E0,$93,$FA,$6A,$ED,$8F,$C0,$9C,$BE,$D5,$8F,$A6,$B8
	.dc.b	$61,$E1,$8F,$66,$1F,$C9,$9B,$B7,$16,$DF,$B2,$B8,$65,$D6,$F8,$36
	.dc.b	$F8,$36,$B8,$4F,$EE,$BD,$B7,$D3,$67,$91,$EE,$BF,$07,$C3,$70,$FF
	.dc.b	$02,$5F,$06,$DE,$AA,$46,$0F,$F0,$57,$B6,$F6,$7F,$C0,$0D,$8F,$F0
	.dc.b	$9B,$5F,$F0,$BB,$5F,$F3,$13,$9C,$BF,$A4,$FA,$1E,$39,$FE,$41,$AF
	.dc.b	$CD,$8C,$4F,$F4,$EA,$E1,$3F,$45,$B5,$F8,$FA,$DC,$F1,$C3,$E3,$4A
	.dc.b	$B6,$B2,$FD,$C1,$1F,$F0,$D3,$5F,$ED,$D9,$FF,$74,$6F,$6F,$49,$FF
	.dc.b	$3E,$65,$FD,$89,$B8,$FF,$67,$BE,$17,$FE,$47,$6D,$6E,$3C,$B9,$BE
	.dc.b	$3F,$FC,$54,$D8,$FF,$1E,$B5,$FF,$20,$B5,$FF,$46,$39,$9F,$EB,$5E
	.dc.b	$D7,$F8,$5E,$7F,$CA,$CD,$8F,$EE,$F3,$BC,$3F,$34,$DA,$FE,$D7,$5B
	.dc.b	$9F,$F0,$03,$63,$F7,$93,$B7,$E6,$7C,$BF,$CE,$D0,$D3,$E8,$EC,$FA
	.dc.b	$1C,$57,$E2,$A6,$BF,$80,$3E,$D4,$63,$2F,$C0,$4D,$CB,$F0,$93,$6B
	.dc.b	$F8,$76,$1E,$07,$A1,$FE,$5F,$C3,$C0,$E0,$EC,$7A,$0C,$61,$F9,$2A
	.dc.b	$E9,$19,$37,$5D,$5E,$35,$D1,$CE,$1C,$27,$57,$CE,$7A,$E1,$F9,$7A
	.dc.b	$EB,$7C,$6B,$DB,$5A,$3D,$73,$F2,$30,$E8,$F0,$6C,$6C,$57,$3B,$ED
	.dc.b	$19,$C3,$C7,$1F,$22,$31,$7C,$DA,$73,$AE,$4F,$88,$C9,$F0,$2F,$C2
	.dc.b	$FC,$03,$E7,$38,$C3,$D7,$1B,$38,$B8,$3E,$23,$C4,$8E,$6F,$CD,$18
	.dc.b	$7D,$6D,$3F,$0A,$74,$73,$1D,$0E,$51,$E9,$38,$1F,$1A,$7C,$D3,$C0
	.dc.b	$EF,$3C,$27,$F0,$0D,$AF,$87,$18,$9E,$0F,$59,$F0,$46,$23,$46,$5E
	.dc.b	$A6,$70,$FA,$71,$93,$D6,$7C,$33,$F9,$2E,$1F,$CB,$08,$FF,$2C,$7F
	.dc.b	$57,$E8,$48,$F8,$47,$F2,$17,$0F,$C0,$9E,$EF,$E1,$A3,$37,$EF,$9C
	.dc.b	$E1,$FE,$DC,$8F,$E6,$73,$F1,$AB,$A1,$FA,$B7,$0F,$33,$07,$88,$C1
	.dc.b	$FA,$67,$73,$D7,$0F,$E7,$91,$83,$C6,$FC,$5C,$2D,$78,$0F,$53,$87
	.dc.b	$AD,$FA,$63,$17,$D6,$7E,$61,$F4,$8D,$D7,$81,$F3,$46,$DC,$9C,$30
	.dc.b	$F8,$78,$77,$7E,$49,$88,$F7,$9E,$42,$E1,$F2,$1E,$B3,$C7,$0F,$DF
	.dc.b	$B6,$3F,$5C,$D7,$F2,$B7,$E3,$3C,$97,$E0,$AF,$6D,$7C,$7A,$EF,$87
	.dc.b	$C7,$69,$F1,$F0,$F8,$8E,$0F,$69,$F0,$EF,$D4,$FD,$53,$CE,$6D,$F0
	.dc.b	$3C,$70,$F1,$E1,$EA,$8D,$35,$E4,$7C,$32,$D7,$C3,$70,$D2,$5F,$A6
	.dc.b	$36,$FD,$57,$6D,$8D,$C7,$D5,$9F,$86,$FA,$0E,$26,$4F,$46,$1F,$1D
	.dc.b	$C9,$F0,$63,$8B,$D8,$F9,$15,$DC,$F2,$57,$82,$3E,$4E,$1F,$11,$E7
	.dc.b	$87,$C4,$7E,$D9,$F1,$0A,$78,$23,$84,$F3,$9E,$13,$F8,$0C,$3F,$05
	.dc.b	$3F,$C3,$DB,$7F,$C2,$EB,$F0,$91,$EB,$9F,$25,$F4,$AF,$45,$FA,$8F
	.dc.b	$84,$79,$AB,$84,$6A,$3E,$11,$F1,$8F,$95,$5D,$30,$FB,$8D,$8F,$5B
	.dc.b	$5F,$41,$EF,$29,$EF,$8E,$33,$DE,$FC,$D3,$E5,$3E,$5C,$E9,$C4,$69
	.dc.b	$FC,$BE,$1F,$1A,$FE,$24,$77,$6B,$DC,$F2,$DF,$7B,$E9,$1D,$8C,$9D
	.dc.b	$9C,$D7,$59,$F9,$D8,$7D,$0A,$FE,$13,$3F,$41,$F8,$B3,$D5,$E6,$75
	.dc.b	$3E,$0D,$FC,$53,$CA,$58,$EB,$1E,$96,$BE,$93,$8B,$D5,$A7,$9C,$EF
	.dc.b	$3E,$77,$CB,$5E,$77,$F8,$F4,$79,$DF,$CE,$1F,$2A,$3F,$02,$F3,$3A
	.dc.b	$9E,$9A,$F3,$D7,$85,$B1,$B6,$BB,$8F,$4C,$7B,$63,$C6,$7B,$4A,$7B
	.dc.b	$5E,$35,$E9,$8F,$3D,$7A,$70,$FE,$27,$5C,$E7,$27,$32,$9C,$EF,$C2
	.dc.b	$79,$CA,$F9,$DA,$F9,$CF,$1C,$75,$9E,$F3,$C6,$72,$D7,$BC,$F8,$FE
	.dc.b	$67,$FC,$81,$1F,$EB,$F0,$E1,$FF,$61,$35,$FF,$70,$35,$DE,$3F,$EB
	.dc.b	$3F,$1B,$FC,$D5,$C8,$7F,$CD,$93,$A3,$FD,$D1,$6C,$3A,$DF,$1F,$32
	.dc.b	$56,$A7,$2D,$3F,$DE,$1F,$32,$FA,$9F,$44,$7F,$B5,$E3,$0E,$03,$87
	.dc.b	$6D,$7E,$03,$9C,$3C,$90,$CB,$D5,$D1,$ED,$3F,$D5,$F5,$BC,$3C,$18
	.dc.b	$66,$FC,$D7,$C1,$F4,$1E,$28,$FF,$5B,$F9,$8F,$2D,$E5,$F9,$E3,$0D
	.dc.b	$F3,$3F,$D5,$B7,$FE,$F3,$9C,$BF,$03,$E2,$1F,$11,$E8,$8F,$B5,$18
	.dc.b	$CF,$E1,$36,$38,$4E,$DF,$F7,$F7,$C5,$3D,$FC,$40,$46,$6F,$E6,$36
	.dc.b	$3E,$8B,$5F,$C5,$79,$4F,$57,$9C,$23,$15,$F5,$5B,$1F,$69,$AF,$29
	.dc.b	$FA,$7C,$5E,$D9,$DB,$F5,$BC,$07,$C5,$3E,$2F,$78,$FA,$71,$F0,$63
	.dc.b	$87,$D9,$8B,$CD,$C4,$79,$7C,$EB,$39,$BF,$B2,$A3,$35,$27,$EE,$D7
	.dc.b	$C7,$F3,$BE,$47,$94,$F2,$FC,$58,$F3,$AF,$F5,$A9,$B1,$3B,$7C,$B8
	.dc.b	$7E,$4A,$31,$1E,$BB,$EB,$7D,$2B,$4F,$E8,$A3,$33,$CB,$0A,$F1,$3B
	.dc.b	$CF,$18,$CC,$78,$EF,$BE,$1E,$77,$C3,$19,$AE,$0F,$23,$F8,$45,$F6
	.dc.b	$AE,$13,$E3,$9D,$1F,$81,$C3,$CE,$F0,$33,$3C,$1C,$5F,$8A,$77,$1E
	.dc.b	$98,$C6,$1E,$88,$61,$E1,$AD,$C7,$9C,$F8,$47,$88,$C1,$E3,$9E,$53
	.dc.b	$FC,$1E,$3E,$A1,$FC,$25,$C3,$D8,$FE,$1F,$E6,$3C,$D7,$9D,$FC,$3E
	.dc.b	$30,$D8,$CC,$F0,$6B,$FC,$BE,$5E,$6B,$8B,$93,$67,$BB,$5E,$0D,$7F
	.dc.b	$3D,$3C,$A7,$EF,$F9,$CE,$FF,$D1,$6B,$F8,$A8,$C5,$FE,$12,$F8,$3F
	.dc.b	$59,$B3,$BA,$E3,$BF,$EA,$EF,$8D,$EF,$1F,$35,$FE,$5B,$0E,$37,$BC
	.dc.b	$38,$DF,$F0,$BC,$38,$DE,$F0,$E3,$7E,$A6,$D6,$EF,$8A,$BE,$35,$BC
	.dc.b	$7D,$57,$C2,$B5,$7C,$2F,$71,$FC,$6F,$1F,$2B,$6B,$84,$EA,$38,$9D
	.dc.b	$2E,$3F,$9A,$8D,$1B,$AE,$71,$C0,$FB,$5A,$FC,$17,$FA,$0C,$FC,$8A
	.dc.b	$FE,$6D,$5C,$23,$15,$BA,$E2,$7C,$13,$FA,$4C,$F6,$F3,$1E,$77,$E5
	.dc.b	$DB,$3F,$D7,$DA,$FF,$64,$6B,$FD,$F5,$CE,$1F,$D7,$BC,$07,$9A,$F3
	.dc.b	$BF,$94,$4E,$63,$17,$FD,$6A,$F8,$E5,$FC,$B7,$C8,$79,$9C,$23,$0D
	.dc.b	$AF,$E9,$53,$A7,$FB,$1F,$97,$FC,$AD,$FE,$CB,$19,$7F,$BD,$B5,$FF
	.dc.b	$0A,$35,$EA,$7F,$7D,$78,$63,$FD,$86,$7C,$F3,$F2,$83,$88,$FE,$CD
	.dc.b	$3C,$2F,$FA,$36,$1F,$E2,$19,$FE,$D4,$D8,$CD,$FA,$F3,$F8,$FD,$1E
	.dc.b	$D8,$57,$F0,$62,$3F,$54,$87,$F8,$13,$37,$67,$36,$C7,$CD,$3A,$68
	.dc.b	$F3,$A9,$23,$FC,$1B,$E7,$3E,$57,$11,$FE,$22,$AE,$0F,$F8,$85,$AE
	.dc.b	$6B,$F8,$F4,$63,$5F,$C1,$5F,$F8,$E0,$FF,$1B,$C6,$75,$FF,$63,$9F
	.dc.b	$EB,$7F,$3F,$E0,$6F,$FE,$9B,$1E,$67,$C3,$E1,$4E,$A2,$67,$F7,$9D
	.dc.b	$46,$23,$FC,$F7,$3C,$1D,$10,$FB,$55,$D9,$FC,$4C,$F1,$CB,$ED,$DD
	.dc.b	$C3,$FC,$01,$87,$3B,$FF,$43,$E1,$C5,$FC,$24,$66,$F8,$C7,$76,$BA
	.dc.b	$C3,$A3,$FE,$D0,$8F,$C7,$46,$23,$FD,$9E,$6C,$E5,$7F,$21,$FF,$15
	.dc.b	$57,$78,$C4,$7F,$A7,$27,$95,$FF,$B1,$6B,$53,$DE,$BE,$83,$D0,$C1
	.dc.b	$D2,$B8,$E1,$E3,$AE,$33,$E4,$C3,$D6,$7A,$1E,$46,$6F,$E3,$DF,$28
	.dc.b	$CD,$72,$C3,$EF,$61,$F6,$27,$8C,$79,$EB,$84,$7D,$69,$D3,$F4,$1B
	.dc.b	$87,$8B,$0F,$83,$87,$59,$E3,$3E,$5A,$E0,$7D,$E3,$BC,$7B,$6F,$77
	.dc.b	$F1,$27,$67,$D1,$2D,$73,$3C,$0F,$C8,$3F,$81,$CF,$27,$FD,$C6,$D7
	.dc.b	$EA,$1E,$03,$E1,$B9,$C3,$F1,$53,$D6,$BE,$DC,$61,$F9,$75,$C4,$F8
	.dc.b	$87,$AC,$F2,$BA,$69,$F6,$CE,$13,$E8,$7C,$E6,$23,$E0,$46,$67,$E9
	.dc.b	$4E,$4F,$59,$4F,$59,$D8,$E6,$5C,$3E,$93,$B6,$BF,$B4,$BF,$81,$87
	.dc.b	$DA,$C3,$DE,$D9,$F9,$97,$F0,$5F,$51,$D6,$B8,$BF,$07,$0F,$45,$7D
	.dc.b	$B8,$F3,$57,$17,$CB,$3A,$7E,$89,$A9,$F0,$3F,$99,$3D,$E7,$73,$47
	.dc.b	$7B,$F3,$B4,$F3,$C7,$2A,$FC,$EC,$64,$E8,$7C,$8A,$FE,$06,$FF,$52
	.dc.b	$9F,$F2,$34,$FF,$B8,$48,$E0,$91,$FD,$55,$AF,$F1,$0F,$3F,$F9,$31
	.dc.b	$F8,$FC,$4E,$23,$E9,$35,$FE,$4A,$DB,$F8,$91,$CC,$FE,$D7,$1C,$CD
	.dc.b	$61,$FA,$E6,$C7,$E8,$6B,$BB,$C0,$EE,$53,$B9,$6B,$B9,$FD,$B6,$79
	.dc.b	$C7,$C6,$9E,$6F,$8C,$E2,$59,$E3,$5F,$CD,$2B,$8E,$1F,$C7,$DB,$87
	.dc.b	$F1,$96,$DF,$F1,$26,$BF,$AA,$3A,$57,$37,$D2,$60,$F7,$47,$0B,$FD
	.dc.b	$56,$1E,$67,$D5,$3C,$1A,$FA,$1A,$7A,$1F,$E7,$33,$CE,$BF,$C4,$17
	.dc.b	$F0,$AF,$A4,$7C,$FA,$F9,$0D,$7C,$87,$C7,$AE,$4F,$F8,$92,$B9,$47
	.dc.b	$D9,$8E,$27,$A8,$A7,$12,$E1,$F2,$5B,$3F,$25,$A7,$C9,$29,$F2,$70
	.dc.b	$F6,$DF,$C9,$8F,$80,$D8,$F8,$11,$FE,$28,$BF,$3A,$BF,$DB,$61,$F0
	.dc.b	$23,$E2,$9F,$2E,$F9,$1F,$06,$FB,$C7,$AC,$F8,$75,$E9,$6E,$1E,$96
	.dc.b	$9E,$92,$9E,$9A,$EA,$F8,$0F,$49,$6B,$D3,$7E,$58,$F5,$1E,$53,$53
	.dc.b	$E5,$39,$BF,$34,$EF,$3F,$88,$C3,$E0,$E1,$F0,$1B,$87,$D8,$AE,$87
	.dc.b	$63,$C6,$5A,$F1,$CF,$08,$FB,$78,$79,$6F,$E1,$57,$C6,$9F,$13,$63
	.dc.b	$38,$79,$9B,$3E,$A3,$E6,$4E,$F0,$F9,$4D,$81,$FE,$63,$F3,$FF,$97
	.dc.b	$12,$7F,$CB,$9E,$AB,$C0,$FF,$2E,$4F,$F8,$9E,$B8,$9C,$9C,$38,$6B
	.dc.b	$9A,$FF,$75,$47,$F4,$3E,$57,$A0,$F4,$3F,$CD,$4D,$BF,$2E,$4D,$3A
	.dc.b	$63,$FE,$6D,$F4,$1E,$85,$25,$F0,$AF,$6E,$5D,$BE,$4F,$33,$E8,$3E
	.dc.b	$0D,$F2,$70,$62,$FF,$E0,$04,$66,$32,$DC,$7F,$D4,$D9,$7F,$AD,$A5
	.dc.b	$87,$B3,$C0,$EF,$5E,$99,$F9,$FE,$BB,$83,$BC,$61,$AA,$34,$E9,$AF
	.dc.b	$9B,$CF,$7A,$4F,$4B,$9C,$61,$BE,$97,$FA,$2A,$BE,$04,$E5,$F7,$FA
	.dc.b	$4F,$4B,$CD,$1F,$16,$31,$C5,$E6,$6C,$70,$9D,$BF,$82,$F3,$5E,$77
	.dc.b	$BF,$6F,$81,$E1,$3E,$69,$F1,$39,$CF,$85,$DC,$7F,$D4,$35,$A9,$FB
	.dc.b	$51,$F9,$6C,$7A,$C7,$C8,$6C,$7D,$86,$BF,$75,$AF,$97,$AF,$B2,$87
	.dc.b	$8D,$87,$F3,$D5,$E0,$9D,$4E,$6A,$4E,$67,$F5,$3E,$87,$EA,$BC,$E7
	.dc.b	$9D,$DE,$B5,$4A,$D4,$E6,$15,$C1,$FC,$A4,$E9,$F8,$17,$F8,$99,$D4
	.dc.b	$7E,$C2,$74,$FA,$E7,$55,$F9,$A9,$CE,$3E,$CA,$EF,$39,$69,$DB,$0E
	.dc.b	$6F,$B0,$C9,$D5,$CB,$EB,$33,$3D,$70,$F2,$C7,$17,$CF,$19,$AF,$04
	.dc.b	$63,$0F,$3D,$7C,$03,$9B,$87,$DC,$62,$FC,$38,$7B,$27,$53,$DE,$FE
	.dc.b	$65,$F8,$A3,$CC,$F7,$BE,$31,$FA,$EB,$FD,$7B,$EB,$3F,$66,$FE,$D3
	.dc.b	$CC,$79,$EF,$43,$0D,$AF,$E0,$F3,$A9,$C9,$C7,$D0,$71,$52,$57,$08
	.dc.b	$CB,$63,$ED,$FA,$0E,$CF,$E5,$BA,$7E,$02,$31,$5F,$80,$C5,$87,$2A
	.dc.b	$FC,$2C,$63,$0E,$78,$72,$D7,$F9,$BE,$3C,$B1,$FB,$B7,$B6,$E3,$CE
	.dc.b	$B7,$7C,$67,$FA,$3D,$F1,$C7,$BD,$EE,$7F,$A5,$E1,$CE,$7D,$17,$C6
	.dc.b	$57,$C7,$0F,$E9,$37,$B6,$E1,$CE,$B7,$7C,$6F,$84,$76,$9D,$5F,$18
	.dc.b	$F2,$D6,$CF,$04,$66,$0D,$AE,$71,$D8,$F9,$95,$E1,$8F,$E7,$91,$93
	.dc.b	$ED,$EB,$F1,$1F,$E0,$D5,$CE,$FF,$B1,$E5,$FC,$AA,$F8,$9C,$A7,$C5
	.dc.b	$E7,$3D,$0D,$E3,$FD,$AA,$31,$8F,$F1,$16,$BE,$03,$FA,$6F,$9F,$FC
	.dc.b	$0B,$1F,$E6,$F3,$A9,$FE,$B8,$D7,$A5,$FF,$4C,$8C,$47,$F8,$03,$C8
	.dc.b	$79,$9E,$28,$C3,$5F,$F1,$E3,$7C,$AF,$E4,$F1,$FC,$96,$78,$B6,$78
	.dc.b	$46,$23,$D0,$27,$8B,$88,$D3,$FD,$CD,$AF,$F8,$31,$AF,$90,$FE,$CF
	.dc.b	$5F,$DE,$70,$F8,$31,$E7,$8C,$65,$FE,$00,$CB,$BD,$6A,$BE,$61,$B3
	.dc.b	$E0,$78,$69,$DD,$7D,$39,$F9,$04,$3C,$27,$F7,$6C,$9E,$77,$C7,$6C
	.dc.b	$7F,$49,$6B,$F6,$6D,$E7,$3C,$FE,$17,$2B,$83,$E4,$9D,$46,$0F,$F0
	.dc.b	$3D,$F4,$7F,$C3,$4D,$7F,$C6,$6D,$7F,$DB,$67,$FB,$67,$4F,$F6,$DB
	.dc.b	$F5,$35,$FF,$7D,$E9,$FE,$64,$F3,$FF,$DE,$73,$96,$C7,$F8,$12,$77
	.dc.b	$A7,$FA,$7A,$BF,$DE,$27,$FB,$B4,$93,$FE,$86,$BF,$F4,$1D,$F2,$AF
	.dc.b	$F2,$3D,$F2,$BF,$F6,$84,$E5,$FF,$29,$C6,$5F,$F6,$54,$65,$FF,$80
	.dc.b	$33,$FE,$F8,$7C,$C7,$FB,$F0,$A7,$89,$D5,$FA,$67,$95,$75,$B7,$99
	.dc.b	$F9,$83,$FE,$00,$1F,$EF,$19,$D9,$30,$ED,$7C,$DB,$87,$B2,$75,$1E
	.dc.b	$13,$E8,$9E,$49,$F8,$95,$DA,$3F,$54,$F8,$5F,$AC,$F8,$5E,$A7,$C4
	.dc.b	$9E,$B3,$CA,$36,$72,$3E,$9B,$97,$C1,$5F,$08,$F3,$3D,$0F,$96,$79
	.dc.b	$A3,$D8,$FA,$4F,$F5,$E5,$78,$DC,$5F,$E2,$A7,$67,$72,$D7,$79,$F4
	.dc.b	$1D,$F0,$FA,$17,$EF,$8F,$29,$E1,$9F,$2B,$DA,$3C,$15,$D6,$72,$D9
	.dc.b	$EA,$6C,$F8,$47,$FB,$0C,$CB,$B7,$DE,$7B,$0D,$BF,$E6,$0B,$FC,$BC
	.dc.b	$79,$F0,$F9,$F5,$EB,$8E,$33,$F0,$63,$C6,$67,$0F,$6C,$FD,$28,$C5
	.dc.b	$72,$8F,$71,$E4,$8C,$BD,$C8,$7A,$CF,$51,$E7,$3E,$39,$6F,$F7,$0F
	.dc.b	$07,$DD,$5E,$28,$E2,$D8,$F7,$61,$F9,$19,$EB,$87,$C0,$7C,$13,$D1
	.dc.b	$B7,$FB,$F7,$9E,$1E,$B7,$C5,$19,$9E,$B1,$F9,$43,$AD,$F7,$8E,$87
	.dc.b	$9A,$B9,$5F,$CB,$AF,$88,$7C,$63,$E7,$D7,$E0,$AF,$E3,$C7,$90,$FE
	.dc.b	$12,$7C,$17,$37,$D6,$30,$73,$C3,$E1,$4F,$9C,$FB,$31,$9C,$3C,$77
	.dc.b	$DE,$30,$7E,$30,$FE,$AC,$FF,$AD,$B0,$FF,$6E,$C7,$D5,$32,$7F,$BE
	.dc.b	$BC,$FF,$CF,$63,$F1,$23,$F9,$84,$7F,$BF,$DA,$7F,$A3,$F1,$FF,$80
	.dc.b	$17,$FE,$F6,$C7,$FD,$F8,$D9,$FF,$7E,$1F,$40,$FE,$F4,$FF,$75,$AE
	.dc.b	$97,$FC,$DD,$B5,$FC,$C1,$B5,$CA,$FB,$BF,$A2,$9D,$3E,$39,$F4,$C7
	.dc.b	$C7,$8E,$11,$87,$C2,$7B,$CE,$46,$0F,$69,$2B,$E1,$12,$7B,$C6,$5E
	.dc.b	$CF,$B4,$FF,$04,$4F,$A1,$A7,$94,$EE,$6E,$91,$B6,$C6,$CF,$C9,$9E
	.dc.b	$72,$9E,$13,$E6,$E1,$FD,$11,$EB,$1C,$9D,$1E,$58,$D3,$88,$F9,$6F
	.dc.b	$EC,$27,$C4,$47,$C5,$7F,$05,$F8,$CF,$D7,$73,$7F,$0A,$FA,$CE,$67
	.dc.b	$C3,$18,$94,$4F,$3F,$0D,$C7,$FC,$75,$18,$DD,$E5,$3B,$73,$52,$FF
	.dc.b	$DC,$17,$BA,$D4,$62,$7B,$B0,$EC,$52,$96,$7F,$4B,$E6,$3C,$DF,$F2
	.dc.b	$63,$67,$FD,$DC,$D7,$FC,$4D,$E6,$BC,$ED,$D4,$9D,$BF,$E3,$89,$D4
	.dc.b	$60,$F1,$0A,$E8,$F9,$DA,$FA,$9A,$FC,$AF,$2D,$E7,$F2,$84,$61,$EA
	.dc.b	$D7,$E0,$35,$F8,$9E,$7F,$CE,$7C,$6D,$7E,$33,$5D,$E9,$DF,$CD,$79
	.dc.b	$EF,$3F,$D7,$18,$6E,$0F,$37,$35,$BC,$3E,$97,$92,$F3,$1C,$BF,$86
	.dc.b	$E7,$B5,$C7,$CC,$79,$B8,$9C,$CB,$0E,$57,$C4,$FC,$7E,$4C,$38,$C6
	.dc.b	$6B,$F2,$75,$C1,$FF,$4C,$56,$F0,$ED,$5F,$8A,$AD,$61,$E4,$AD,$4F
	.dc.b	$3B,$F1,$73,$F0,$7A,$F9,$B5,$CC,$EF,$04,$E5,$FC,$F4,$66,$B8,$57
	.dc.b	$EA,$3C,$DF,$D4,$F9,$6F,$2F,$C9,$5A,$A5,$6A,$73,$76,$7F,$45,$35
	.dc.b	$F1,$8B,$E8,$E8,$C5,$72,$69,$98,$DB,$C9,$AF,$9C,$D9,$98,$FD,$84
	.dc.b	$F8,$70,$F7,$CE,$8E,$15,$D2,$32,$F4,$8F,$69,$D2,$33,$5D,$23,$18
	.dc.b	$7E,$22,$B8,$5F,$78,$C5,$6E,$7B,$46,$2B,$D9,$3E,$F9,$E3,$1D,$30
	.dc.b	$F9,$0F,$9A,$74,$78,$5F,$DD,$DF,$EC,$DF,$39,$83,$F8,$87,$90,$F3
	.dc.b	$5E,$6F,$F0,$28,$C3,$63,$33,$C1,$AF,$F1,$B9,$79,$AD,$F9,$36,$7B
	.dc.b	$B5,$E0,$D7,$F9,$2C,$F2,$9F,$C2,$79,$8F,$37,$F9,$4E,$BF,$7A,$31
	.dc.b	$7F,$89,$BE,$0F,$DE,$6C,$EE,$B8,$E9,$FC,$CA,$F8,$DE,$F1,$F3,$5F
	.dc.b	$E2,$B0,$E3,$7B,$C3,$8D,$FF,$48,$C3,$8D,$EF,$0E,$37,$EA,$6D,$6E
	.dc.b	$F8,$AB,$E3,$5B,$C7,$D5,$7C,$2B,$57,$C2,$F7,$1F,$D6,$31,$F2,$B6
	.dc.b	$B8,$4E,$A3,$89,$D2,$E3,$FA,$28,$D1,$BA,$E7,$1C,$0F,$B5,$AF,$C1
	.dc.b	$7F,$86,$4F,$C8,$AF,$E3,$F5,$C2,$31,$5B,$AE,$27,$C1,$3F,$8A,$4F
	.dc.b	$6F,$AE,$79,$FF,$D9,$30,$FE,$55,$FB,$AB,$E1,$B1,$8F,$F3,$18,$C6
	.dc.b	$3F,$C7,$70,$E2,$DC,$36,$FE,$46,$76,$F8,$B0,$46,$1F,$E6,$F5,$CA
	.dc.b	$7C,$E3,$D0,$7A,$3F,$D6,$DB,$1F,$D9,$9A,$FF,$79,$6B,$FD,$F5,$CD
	.dc.b	$7F,$48,$F2,$1E,$8B,$D3,$F3,$C6,$1F,$4D,$F8,$70,$B8,$72,$D3,$FC
	.dc.b	$2F,$7B,$AF,$E9,$0D,$7D,$33,$B7,$FC,$3A,$FF,$84,$9F,$37,$3F,$AC
	.dc.b	$8F,$AD,$F1,$E3,$BC,$BD,$0D,$C3,$FC,$33,$76,$FA,$DF,$02,$3F,$D4
	.dc.b	$63,$4B,$CF,$7A,$FF,$E0,$56,$C7,$F8,$41,$AF,$F8,$8D,$AF,$F8,$9D
	.dc.b	$CE,$3F,$2F,$D1,$7C,$E7,$CA,$6D,$7F,$61,$AD,$E5,$FE,$30,$6C,$34
	.dc.b	$57,$08,$FC,$BC,$62,$FE,$9D,$70,$97,$9F,$EB,$7F,$B7,$35,$FC,$4B
	.dc.b	$71,$FF,$01,$FA,$3F,$DB,$27,$75,$F7,$AB,$59,$7F,$91,$1B,$3F,$E5
	.dc.b	$4A,$E0,$F4,$AE,$04,$7F,$D4,$1D,$F5,$D1,$D8,$C9,$4C,$93,$0F,$2C
	.dc.b	$71,$7F,$DF,$53,$70,$FF,$2B,$DF,$8A,$73,$7C,$DA,$FE,$A2,$35,$3E
	.dc.b	$BA,$E5,$3F,$E6,$EC,$3D,$AA,$B6,$E6,$76,$FC,$E7,$F4,$71,$A8,$C1
	.dc.b	$E1,$AE,$CF,$B8,$F6,$38,$69,$F2,$F0,$EB,$7E,$23,$C7,$5D,$5C,$5F
	.dc.b	$8A,$30,$F0,$3C,$73,$6B,$C4,$75,$29,$E6,$3C,$27,$BD,$E0,$E2,$7A
	.dc.b	$BC,$2B,$C7,$3A,$C3,$E6,$9F,$5A,$75,$7E,$59,$D6,$1E,$FA,$D5,$78
	.dc.b	$2F,$D6,$FC,$58,$E7,$3B,$24,$7A,$4E,$17,$D5,$EE,$7B,$30,$F3,$BA
	.dc.b	$3A,$4F,$27,$E0,$57,$67,$A3,$F8,$67,$FD,$41,$7F,$A7,$6D,$FF,$A1
	.dc.b	$E3,$11,$CE,$FF,$93,$3F,$EF,$DE,$1F,$7B,$FA,$57,$F0,$A6,$1F,$C6
	.dc.b	$38,$7D,$27,$E5,$5C,$3E,$93,$07,$53,$77,$FB,$88,$DB,$EB,$8F,$6E
	.dc.b	$1F,$A4,$9C,$BF,$06,$FF,$17,$18,$9F,$CE,$1F,$25,$C3,$C8,$C9,$E2
	.dc.b	$25,$77,$9E,$71,$E5,$3D,$11,$E5,$70,$EF,$0E,$B8,$7C,$89,$DE,$1E
	.dc.b	$77,$6F,$9E,$77,$87,$C9,$AE,$17,$D6,$F8,$9F,$3C,$E2,$7A,$1F,$89
	.dc.b	$1C,$2F,$B4,$F0,$3E,$99,$C0,$F8,$04,$8E,$0F,$53,$83,$87,$89,$8A
	.dc.b	$F9,$38,$7E,$72,$3E,$3D,$78,$2F,$E1,$47,$7B,$E5,$5B,$9F,$61,$2B
	.dc.b	$E1,$56,$EF,$E1,$1D,$B0,$FE,$04,$D8,$FB,$47,$AE,$FA,$57,$84,$91
	.dc.b	$F1,$5F,$CD,$4F,$18,$FE,$00,$DB,$F9,$2D,$9D,$CF,$5B,$F0,$B4,$F0
	.dc.b	$9F,$00,$EC,$43,$E3,$D7,$B5,$C5,$FE,$29,$E0,$7D,$27,$11,$B8,$C4
	.dc.b	$FE,$11,$FD,$1D,$7B,$1E,$AF,$C8,$79,$E1,$F3,$A7,$CF,$1D,$AF,$E2
	.dc.b	$BF,$76,$FB,$46,$23,$99,$9B,$E9,$3E,$3B,$E9,$3F,$02,$38,$DF,$66
	.dc.b	$C6,$6B,$CD,$7D,$6F,$F1,$47,$29,$FA,$27,$CB,$29,$E7,$BE,$95,$C9
	.dc.b	$B1,$E5,$3E,$5B,$F1,$5F,$CA,$1C,$EF,$ED,$B6,$7E,$DB,$D4,$F3,$1C
	.dc.b	$CF,$9E,$7E,$44,$F9,$E6,$6B,$D1,$3D,$23,$91,$E8,$7E,$0B,$67,$D0
	.dc.b	$F6,$3E,$69,$D8,$FA,$F8,$7E,$79,$F3,$1F,$89,$AF,$8F,$19,$8E,$F5
	.dc.b	$E1,$3B,$96,$BB,$C7,$84,$F6,$61,$F4,$6B,$F3,$91,$D5,$AF,$F0,$16
	.dc.b	$C7,$E4,$A7,$9D,$75,$AE,$78,$7D,$27,$D7,$5D,$5F,$49,$CC,$AF,$37
	.dc.b	$89,$F0,$F0,$F9,$77,$F5,$2F,$D6,$D3,$C2,$59,$E2,$7B,$4F,$0D,$F8
	.dc.b	$EB,$9C,$7B,$4F,$0C,$7E,$16,$BF,$89,$C7,$D0,$3E,$E9,$F5,$27,$8B
	.dc.b	$E9,$3E,$7D,$78,$5E,$E7,$A7,$0F,$A9,$5E,$88,$F0,$1C,$0A,$E5,$E8
	.dc.b	$DC,$3E,$24,$74,$3D,$97,$E2,$8C,$B7,$0F,$56,$1F,$DF,$7D,$0F,$F3
	.dc.b	$35,$88,$1C,$B8,$6B,$B6,$BB,$CF,$FD,$43,$E6,$FF,$9C,$72,$FF,$53
	.dc.b	$36,$F9,$4F,$FA,$92,$F1,$7F,$E7,$9A,$BA,$9C,$C7,$53,$CD,$83,$0F
	.dc.b	$3F,$E3,$BF,$E6,$C9,$F4,$87,$0E,$B6,$CB,$4A,$6D,$F4,$FC,$6C,$8E
	.dc.b	$F5,$B7,$FD,$97,$5F,$E5,$EC,$7C,$D5,$B9,$CD,$FF,$A0,$F2,$FF,$11
	.dc.b	$12,$3B,$ED,$F3,$5B,$1F,$E6,$4F,$4E,$47,$FA,$4A,$39,$3E,$23,$FD
	.dc.b	$15,$1A,$7F,$D2,$B1,$FE,$CE,$F3,$1E,$53,$CC,$F6,$C6,$1B,$E6,$3C
	.dc.b	$9C,$D6,$F0,$F8,$BE,$BF,$FB,$EB,$83,$E2,$B6,$33,$DF,$F4,$5B,$3F
	.dc.b	$2B,$CA,$76,$67,$93,$54,$97,$0E,$6F,$F1,$F1,$C1,$F7,$4E,$E7,$4F
	.dc.b	$E5,$67,$4F,$FB,$F1,$B5,$DA,$3F,$35,$E5,$7E,$6F,$C8,$79,$1D,$B2
	.dc.b	$E7,$5F,$9E,$AD,$4E,$72,$B3,$F8,$A9,$AF,$A6,$BE,$05,$F4,$8D,$07
	.dc.b	$0E,$EB,$8B,$E4,$35,$7D,$DA,$72,$9E,$38,$FB,$9E,$71,$C5,$CD,$F8
	.dc.b	$AB,$F3,$78,$79,$AF,$84,$62,$FA,$9E,$32,$9C,$DF,$15,$F6,$85,$73
	.dc.b	$BE,$D1,$E5,$AD,$9D,$6B,$C7,$87,$BA,$78,$9E,$0C,$3C,$6F,$13,$ED
	.dc.b	$5F,$47,$E9,$1F,$B2,$38,$9F,$B4,$C3,$0E,$0F,$E2,$BE,$2B,$C9,$79
	.dc.b	$5F,$BA,$8C,$36,$33,$3F,$C0,$27,$2D,$F2,$5A,$F1,$6C,$66,$30,$47
	.dc.b	$E5,$F9,$2F,$2B,$F9,$3E,$DF,$80,$8C,$4F,$E2,$30,$61,$D2,$BF,$0B
	.dc.b	$18,$C3,$F1,$B7,$C2,$FA,$E1,$D3,$7F,$7E,$3C,$5B,$7C,$27,$55,$FC
	.dc.b	$E6,$F9,$63,$DB,$0E,$13,$FC,$F7,$1F,$03,$D8,$E4,$79,$E3,$6A,$37
	.dc.b	$3E,$DB,$EF,$2B,$EF,$8F,$92,$3F,$1F,$7C,$23,$FA,$5C,$EA,$7C,$D5
	.dc.b	$CA,$B5,$1E,$5C,$3C,$35,$B8,$ED,$5B,$9C,$BF,$D5,$B1,$FE,$1B,$5F
	.dc.b	$25,$B1,$9D,$FC,$99,$7A,$70,$FE,$1B,$7B,$9F,$2D,$6E,$7C,$57,$FC
	.dc.b	$03,$C8,$71,$7E,$55,$B3,$FD,$21,$AF,$F4,$C6,$BF,$D4,$9C,$CF,$F2
	.dc.b	$3F,$25,$E5,$77,$84,$62,$BF,$90,$D7,$0C,$7F,$B9,$F9,$0F,$2B,$CD
	.dc.b	$48,$C3,$FC,$A1,$AF,$E7,$27,$FB,$E7,$82,$F2,$BF,$8A,$E5,$FC,$D2
	.dc.b	$F7,$7F,$DA,$1A,$F8,$30,$FD,$A4,$62,$3F,$C0,$CF,$69,$FE,$85,$7D
	.dc.b	$7C,$8F,$94,$DA,$F2,$5F,$A2,$FF,$A2,$C7,$9F,$9F,$07,$0C,$5F,$23
	.dc.b	$89,$FE,$EB,$5C,$1F,$EF,$2D,$7D,$15,$E6,$39,$11,$EC,$7F,$4C,$F2
	.dc.b	$1E,$5F,$CA,$6C,$66,$3C,$4E,$5F,$E2,$91,$A8,$FF,$04,$B5,$FF,$07
	.dc.b	$B5,$FF,$14,$BB,$D3,$E2,$F9,$7F,$E1,$C3,$FC,$37,$19,$DB,$FD,$4A
	.dc.b	$7F,$8C,$3C,$64,$7D,$EF,$2F,$FB,$3E,$1F,$E4,$E9,$CB,$FD,$9A,$32
	.dc.b	$FF,$BD,$63,$2F,$E0,$4F,$F1,$E4,$6B,$0F,$F2,$C9,$FE,$FE,$EF,$FC
	.dc.b	$5D,$7F,$AE,$4E,$31,$FD,$E2,$78,$61,$E6,$BF,$F4,$65,$F0,$C2,$47
	.dc.b	$F7,$8C,$38,$D7,$F9,$2A,$F8,$CF,$FA,$B8,$D5,$7B,$6F,$8C,$79,$CF
	.dc.b	$F4,$E6,$1E,$18,$CB,$E8,$BE,$91,$87,$C9,$2B,$EE,$F2,$8F,$99,$7D
	.dc.b	$E3,$E5,$35,$F9,$0D,$BE,$D1,$EE,$36,$7C,$FB,$F0,$3E,$B3,$D1,$7E
	.dc.b	$0C,$3E,$55,$F3,$C3,$CB,$1F,$12,$BE,$9D,$F2,$7C,$C6,$0E,$E7,$47
	.dc.b	$51,$FE,$E1,$6D,$71,$BF,$03,$8A,$F3,$1C,$B0,$F9,$4D,$C3,$E4,$B4
	.dc.b	$F0,$1F,$08,$E2,$53,$C4,$6C,$F9,$D7,$DD,$FB,$18,$7B,$5C,$9B,$C3
	.dc.b	$CA,$FF,$A9,$9A,$F0,$C3,$D5,$5C,$F0,$FC,$A4,$F1,$8C,$61,$F6,$A3
	.dc.b	$CB,$3F,$59,$61,$E1,$C3,$AC,$E5,$B5,$F8,$66,$D7,$D9,$9F,$8A,$FB
	.dc.b	$DA,$75,$AE,$97,$E2,$9F,$C2,$1B,$BF,$96,$D3,$E5,$D7,$1C,$3F,$05
	.dc.b	$1D,$E7,$D4,$FC,$C3,$81,$F2,$0B,$87,$99,$FC,$29,$9A,$E8,$FC,$29
	.dc.b	$FC,$6D,$76,$8D,$E1,$E8,$7E,$51,$0E,$B8,$7A,$DF,$93,$3C,$67,$F3
	.dc.b	$31,$D4,$F4,$1F,$02,$FC,$2F,$A5,$A7,$C8,$BF,$1C,$71,$3C,$75,$D4
	.dc.b	$FF,$1F,$57,$F1,$68,$FF,$5B,$B5,$F8,$A6,$23,$64,$7F,$D5,$2D,$7F
	.dc.b	$DE,$47,$E2,$CF,$F7,$61,$F8,$FE,$1F,$E5,$47,$FB,$BA,$BF,$8A,$36
	.dc.b	$3F,$92,$B5,$FF,$75,$5F,$FA,$62,$FF,$DD,$33,$9C,$7F,$DD,$55,$FC
	.dc.b	$C6,$3A,$BD,$CC,$14,$FF,$7D,$1F,$EF,$8B,$F0,$B9,$BF,$F7,$D6,$1F
	.dc.b	$EF,$48,$F9,$E7,$F1,$BA,$EA,$FB,$4F,$F7,$EB,$E3,$23,$CE,$BA,$BB
	.dc.b	$AF,$55,$75,$8F,$C5,$DA,$FA,$D7,$E9,$DB,$7F,$A7,$6D,$79,$2B,$83
	.dc.b	$EE,$3B,$4F,$E7,$2B,$FA,$45,$70,$9E,$ED,$C3,$F8,$31,$F3,$23,$D0
	.dc.b	$7D,$22,$9F,$4A,$BA,$1E,$CA,$F8,$F1,$E1,$6B,$F8,$27,$D4,$E6,$3D
	.dc.b	$4D,$3D,$55,$E3,$C3,$F9,$53,$63,$F1,$27,$B0,$1F,$64,$F3,$19,$30
	.dc.b	$79,$CD,$DF,$B5,$D3,$6F,$CB,$1F,$84,$8C,$5F,$9A,$31,$7F,$5E,$7E
	.dc.b	$3B,$4F,$3C,$7A,$CF,$9A,$58,$DB,$9C,$3F,$2E,$FC,$5A,$ED,$7F,$20
	.dc.b	$FA,$F1,$C6,$FA,$BE,$23,$96,$1F,$8E,$7D,$71,$E3,$7E,$3C,$FB,$0E
	.dc.b	$95,$F5,$C9,$1F,$40,$EC,$70,$7C,$66,$9F,$59,$E1,$3D,$07,$CC,$9F
	.dc.b	$F3,$0F,$8C,$F2,$E9,$27,$EB,$DF,$F9,$53,$CF,$FF,$01,$D6,$E7,$ED
	.dc.b	$90,$39,$70,$62,$BF,$DA,$91,$98,$E0,$EE,$BF,$CC,$3E,$65,$C7,$FC
	.dc.b	$DE,$DA,$FC,$4D,$55,$E6,$7F,$85,$89,$B7,$F9,$A2,$BF,$C8,$93,$F9
	.dc.b	$09,$FF,$1E,$BE,$76,$BE,$A6,$BB,$9F,$F7,$54,$7F,$8C,$39,$94,$F3
	.dc.b	$3B,$1F,$E8,$A9,$FF,$0C,$61,$FE,$42,$26,$1F,$E4,$DF,$5A,$BF,$13
	.dc.b	$87,$E3,$BF,$18,$FE,$E3,$3F,$CF,$08,$FF,$95,$7C,$DF,$F2,$D1,$CB
	.dc.b	$56,$BF,$4B,$D1,$70,$7F,$C0,$16,$DF,$56,$BF,$51,$AF,$D9,$F3,$9E
	.dc.b	$5B,$CE,$F2,$46,$1B,$E7,$3C,$EC,$D6,$EF,$EF,$F9,$AE,$CF,$2C,$7F
	.dc.b	$94,$E1,$CD,$8D,$B3,$29,$E1,$B7,$E4,$0F,$C6,$BC,$1C,$47,$F0,$A8
	.dc.b	$DB,$E0,$BE,$6F,$06,$BB,$6B,$F9,$8F,$37,$F8,$9F,$90,$F2,$F8,$61
	.dc.b	$87,$F3,$33,$6B,$94,$D7,$90,$AE,$4E,$8C,$57,$26,$99,$8D,$BC,$9B
	.dc.b	$87,$B1,$CC,$7F,$28,$BF,$04,$E6,$BA,$57,$0C,$3E,$7B,$C2,$D7,$C2
	.dc.b	$73,$5C,$0F,$7C,$F9,$0F,$13,$8A,$E5,$5C,$70,$F8,$15,$E3,$38,$61
	.dc.b	$E3,$7F,$81,$4E,$DE,$86,$EB,$CD,$5C,$A3,$CA,$7E,$84,$A7,$E8,$E7
	.dc.b	$0E,$0F,$E1,$FE,$5B,$CC,$79,$BF,$A9,$8C,$36,$33,$3F,$C4,$A7,$2D
	.dc.b	$F3,$5A,$F1,$6C,$66,$30,$47,$E3,$79,$AF,$37,$F9,$0E,$9F,$6A,$31
	.dc.b	$3F,$85,$C1,$87,$4A,$FB,$F1,$8C,$3F,$15,$7C,$2F,$AE,$1D,$36,$F7
	.dc.b	$E3,$C5,$B7,$C2,$75,$5F,$DB,$2F,$96,$3D,$B0,$E1,$3F,$DC,$31,$F0
	.dc.b	$3D,$8E,$47,$9E,$36,$A3,$73,$ED,$BE,$F2,$BE,$F8,$F9,$23,$F1,$57
	.dc.b	$C2,$3F,$A2,$CE,$A7,$CD,$5C,$AB,$51,$E5,$C3,$C3,$5B,$8E,$D5,$B9
	.dc.b	$F2,$D7,$09,$C9,$EA,$3C,$31,$A3,$ED,$FD,$3F,$A6,$3F,$C2,$2D,$4F
	.dc.b	$C7,$CB,$38,$FA,$30,$FE,$BB,$1F,$C5,$B1,$FA,$CD,$8C,$ED,$F0,$73
	.dc.b	$E9,$39,$C7,$F8,$B9,$E6,$3F,$DE,$99,$FA,$70,$E7,$7D,$E7,$71,$87
	.dc.b	$E2,$57,$B4,$E8,$FF,$BF,$63,$4F,$D7,$CF,$E4,$E3,$E4,$8F,$9E,$74
	.dc.b	$70,$EE,$BA,$5F,$86,$3A,$4E,$B0,$F3,$3C,$30,$EE,$78,$23,$33,$D4
	.dc.b	$87,$BB,$0F,$1B,$8C,$3D,$27,$F4,$09,$E3,$1E,$03,$58,$7C,$BC,$BC
	.dc.b	$95,$C2,$29,$99,$F6,$9C,$49,$1D,$6F,$C3,$18,$9E,$2F,$CC,$3F,$44
	.dc.b	$64,$E7,$3C,$4F,$8A,$E2,$38,$BF,$24,$E3,$1E,$57,$47,$5C,$BF,$07
	.dc.b	$3C,$31,$F9,$17,$E4,$BE,$17,$F8,$57,$C0,$F4,$AE,$19,$7D,$3A,$E1
	.dc.b	$18,$9F,$C6,$E5,$F2,$AF,$9E,$1D,$9F,$4C,$F5,$AF,$4D,$6D,$E5,$1F
	.dc.b	$5E,$77,$5F,$7B,$0F,$04,$61,$FC,$11,$89,$F3,$1F,$00,$EC,$7C,$1C
	.dc.b	$BA,$CF,$BB,$0F,$04,$EE,$FC,$CF,$23,$89,$23,$6E,$23,$79,$7C,$C8
	.dc.b	$E5,$87,$8E,$31,$1B,$8C,$61,$E2,$C7,$E0,$E5,$F2,$AB,$89,$F2,$9C
	.dc.b	$3C,$4C,$65,$F0,$5F,$CA,$9E,$D3,$89,$F8,$93,$F0,$98,$78,$B1,$F9
	.dc.b	$51,$C8,$F7,$38,$7E,$81,$8C,$7E,$6B,$B9,$E5,$18,$8D,$C6,$1F,$19
	.dc.b	$E0,$C7,$DC,$FD,$3C,$BE,$4D,$F7,$C3,$F1,$B7,$C6,$BD,$15,$AC,$3B
	.dc.b	$56,$B0,$E7,$5F,$4F,$1F,$8B,$8F,$3C,$BF,$65,$3F,$1C,$E7,$3A,$C3
	.dc.b	$9C,$EA,$B8,$57,$96,$B4,$FA,$B1,$F1,$36,$32,$F4,$C7,$E9,$34,$F9
	.dc.b	$06,$DC,$9F,$8F,$8E,$2D,$33,$7F,$82,$6C,$F8,$DA,$FB,$DB,$7F,$90
	.dc.b	$C7,$E1,$B6,$7F,$03,$3D,$5F,$19,$E5,$CB,$E7,$46,$5B,$7E,$19,$DC
	.dc.b	$E6,$FA,$E5,$F9,$89,$EB,$5B,$BE,$31,$E7,$38,$C6,$6B,$8C,$7E,$25
	.dc.b	$AF,$E2,$1B,$7F,$76,$3F,$94,$9F,$EB,$9C,$3F,$C2,$4D,$C3,$FC,$69
	.dc.b	$18,$8F,$E8,$C6,$4F,$F4,$C3,$97,$99,$FC,$7F,$E7,$FF,$A4,$1F,$82
	.dc.b	$7F,$80,$38,$7F,$33,$1F,$D3,$23,$11,$FA,$B8,$C3,$FD,$92,$39,$3F
	.dc.b	$C0,$DD,$D7,$F2,$0A,$B8,$7F,$84,$27,$77,$CA,$FF,$B4,$46,$2B,$71
	.dc.b	$FE,$29,$8C,$3D,$C8,$78,$E3,$6F,$F8,$86,$36,$F4,$37,$7E,$36,$DF
	.dc.b	$8D,$F7,$D7,$2B,$5F,$2B,$FF,$07,$39,$9F,$4B,$6B,$E3,$B5,$F8,$CD
	.dc.b	$9F,$E8,$2D,$BF,$A5,$3E,$23,$F1,$67,$BC,$C1,$FB,$99,$F0,$B8,$BF
	.dc.b	$DB,$BE,$F3,$17,$E4,$69,$E4,$29,$E4,$2C,$F9,$1B,$5E,$47,$E4,$1F
	.dc.b	$34,$F0,$5F,$C3,$8F,$91,$1A,$9F,$98,$D8,$CC,$60,$F4,$C6,$4F,$79
	.dc.b	$E9,$30,$7B,$E3,$89,$1D,$DF,$D4,$7C,$E7,$E6,$0A,$7E,$44,$AF,$E3
	.dc.b	$67,$C2,$78,$E7,$A4,$F8,$63,$0F,$D1,$3C,$73,$F3,$8F,$03,$8B,$F9
	.dc.b	$CD,$3E,$71,$4F,$9C,$59,$F9,$CD,$AF,$9D,$1E,$47,$F8,$CD,$FE,$16
	.dc.b	$7D,$86,$CB,$3B,$7E,$1C,$6D,$F8,$06,$CB,$1B,$7E,$61,$ED,$71,$7E
	.dc.b	$D9,$E4,$FC,$7A,$F8,$AE,$6B,$E5,$DF,$E9,$6B,$F0,$51,$CE,$B7,$3F
	.dc.b	$12,$FA,$B6,$33,$3F,$1A,$72,$D8,$CB,$F2,$23,$2D,$8C,$C7,$CA,$32
	.dc.b	$57,$27,$CC,$2C,$7C,$86,$BF,$11,$B7,$EF,$8F,$51,$F3,$0F,$79,$FA
	.dc.b	$03,$F8,$29,$4F,$D2,$1F,$86,$BF,$86,$FE,$2E,$7A,$DF,$C7,$8F,$F0
	.dc.b	$1F,$3B,$91,$CE,$98,$49,$D4,$62,$31,$CE,$E7,$C4,$71,$AD,$B8,$E5
	.dc.b	$7C,$63,$3D,$CA,$73,$18,$6C,$71,$9E,$0F,$FC,$01,$E6,$F3,$70,$B8
	.dc.b	$7C,$F6,$E7,$DC,$E6,$FD,$17,$C2,$F7,$5A,$7D,$31,$99,$F8,$57,$CF
	.dc.b	$1C,$47,$C4,$9B,$39,$86,$1D,$DD,$18,$BE,$F7,$CA,$3E,$75,$F2,$C3
	.dc.b	$C5,$1E,$63,$81,$E2,$72,$F4,$BE,$D8,$78,$5F,$04,$74,$38,$9F,$36
	.dc.b	$78,$B9,$8E,$31,$88,$57,$8E,$FC,$D1,$83,$9D,$F7,$C3,$E3,$C6,$2F
	.dc.b	$95,$F1,$8F,$09,$E3,$3E,$FB,$F8,$07,$07,$23,$F1,$9C,$8E,$67,$3F
	.dc.b	$E1,$A3,$0D,$8C,$CF,$E4,$27,$2D,$E6,$73,$71,$6C,$66,$30,$47,$FD
	.dc.b	$7B,$AF,$FB,$FC,$FF,$7F,$3F,$2A,$BF,$DA,$31,$9C,$FF,$CB,$71,$87
	.dc.b	$FD,$C9,$1C,$08,$E0,$93,$FE,$83,$3F,$CE,$47,$FB,$52,$3A,$39,$9E
	.dc.b	$8E,$67,$BD,$FF,$B5,$AB,$73,$C8,$9C,$CE,$5F,$DC,$CE,$B0,$5F,$C2
	.dc.b	$8F,$2D,$F5,$21,$CA,$BF,$84,$D7,$F9,$BA,$3F,$CA,$D8,$7F,$91,$37
	.dc.b	$FF,$3B,$56,$B0,$FF,$34,$C7,$F9,$36,$BF,$B2,$CF,$F9,$55,$FE,$7E
	.dc.b	$7F,$99,$CF,$F2,$99,$FE,$55,$C3,$FC,$AE,$E5,$B8,$FF,$9A,$63,$8B
	.dc.b	$4F,$EC,$53,$D5,$B1,$FE,$5B,$3C,$25,$38,$C6,$CF,$46,$5F,$E7,$87
	.dc.b	$11,$FC,$B0,$FF,$31,$1C,$8A,$72,$3A,$4F,$81,$CB,$67,$FC,$D1,$1B
	.dc.b	$33,$3E,$07,$E0,$9E,$FC,$7E,$63,$F8,$5A,$E5,$87,$CE,$8F,$25,$F4
	.dc.b	$9C,$C6,$EF,$A6,$1F,$21,$F5,$C7,$F4,$4A,$E1,$1A,$C3,$BC,$FD,$F3
	.dc.b	$AE,$1D,$DE,$0D,$AF,$F0,$9C,$6C,$C9,$E1,$38,$94,$C9,$B7,$8E,$5F
	.dc.b	$8B,$7F,$D3,$51,$F3,$5F,$2B,$C8,$E1,$3F,$EC,$C9,$E0,$FE,$36,$3E
	.dc.b	$64,$F0,$7B,$CE,$9F,$C6,$57,$BA,$31,$1F,$3C,$F0,$57,$B4,$ED,$97
	.dc.b	$E6,$A3,$89,$A3,$D9,$7E,$A8,$D4,$73,$3C,$73,$CD,$FF,$25,$D7,$83
	.dc.b	$2F,$CC,$63,$F4,$67,$A1,$EE,$37,$7D,$27,$71,$F0,$EB,$95,$7D,$CC
	.dc.b	$3C,$47,$E6,$67,$F2,$71,$D5,$F1,$9E,$8C,$3D,$B3,$B8,$FC,$99,$D7
	.dc.b	$0F,$05,$75,$AD,$CF,$E5,$B0,$F1,$BF,$3A,$B6,$F8,$0E,$57,$D4,$FA
	.dc.b	$87,$99,$E5,$7D,$5F,$7C,$6E,$FE,$44,$7F,$06,$C3,$DE,$F5,$3C,$13
	.dc.b	$B7,$D8,$72,$3A,$9B,$7C,$19,$7E,$8E,$31,$5E,$C8,$C4,$7C,$23,$A9
	.dc.b	$C8,$C4,$F2,$CB,$F1,$58,$FC,$09,$CD,$F3,$9C,$D7,$96,$7E,$E1,$EB
	.dc.b	$AE,$11,$F1,$2B,$84,$FE,$0B,$0F,$7B,$D0,$F6,$61,$E6,$AE,$11,$8C
	.dc.b	$3B,$BE,$78,$CB,$DD,$F1,$57,$47,$39,$7F,$AD,$A3,$18,$7D,$08,$C6
	.dc.b	$3E,$68,$C6,$5F,$EE,$AA,$DE,$1F,$40,$EC,$78,$67,$6F,$F9,$27,$0F
	.dc.b	$DE,$BF,$92,$3E,$1E,$1D,$A3,$CB,$87,$79,$E9,$39,$7A,$18,$C3,$C5
	.dc.b	$3B,$8E,$71,$8A,$E6,$48,$E2,$68,$F2,$5F,$F0,$9B,$FE,$53,$19,$3F
	.dc.b	$AE,$9F,$4E,$FB,$61,$F0,$23,$57,$EC,$AE,$33,$92,$57,$02,$4F,$8F
	.dc.b	$2F,$B5,$87,$C6,$C3,$B8,$C3,$FB,$C3,$FE,$64,$6B,$FE,$6C,$6B,$FE
	.dc.b	$71,$6B,$FE,$76,$ED,$78,$9F,$E2,$D6,$C7,$F9,$B9,$AF,$F9,$D9,$AF
	.dc.b	$F9,$EF,$8F,$FC,$EE,$DE,$7F,$F4,$1B,$FE,$7C,$87,$37,$F9,$FE,$31
	.dc.b	$3A,$7F,$C7,$A7,$F9,$0F,$1F,$F2,$4F,$17,$F9,$0F,$3F,$F2,$AC,$FF
	.dc.b	$92,$DF,$F2,$8C,$EA,$38,$1F,$E4,$83,$5C,$1F,$E6,$87,$15,$CE,$3F
	.dc.b	$CA,$35,$CA,$72,$72,$7B,$CE,$E3,$FC,$D9,$1B,$32,$EE,$79,$9D,$E3
	.dc.b	$26,$DC,$C7,$F9,$CE,$BD,$0F,$C0,$21,$FE,$5B,$3F,$D5,$3B,$FF,$95
	.dc.b	$1F,$F2,$EF,$0F,$F9,$61,$BC,$5F,$E5,$B3,$FC,$A4,$FF,$98,$B8,$7F
	.dc.b	$CA,$ED,$E1,$CB,$FD,$D7,$2F,$F1,$E4,$FF,$8E,$F2,$C1,$FE,$4E,$9F
	.dc.b	$F2,$26,$3F,$E4,$A7,$F1,$DC,$5F,$E4,$49,$FF,$25,$E1,$FE,$9C,$E2
	.dc.b	$FC,$7E,$3D,$23,$FD,$3F,$87,$19,$CD,$70,$C3,$FC,$CE,$F7,$9F,$F2
	.dc.b	$6D,$70,$C7,$FC,$AD,$39,$7C,$D5,$FE,$05,$C3,$B3,$FE,$6D,$8D,$BF
	.dc.b	$E5,$2B,$E6,$E6,$FC,$D8,$FA,$63,$55,$D1,$FF,$0A,$5F,$43,$FC,$BA
	.dc.b	$FC,$03,$89,$E5,$A8,$FA,$8E,$6E,$23,$FC,$DB,$8F,$C0,$9F,$F3,$A5
	.dc.b	$79,$9E,$87,$94,$93,$C2,$FC,$D1,$F4,$A7,$85,$73,$8C,$5F,$91,$F1
	.dc.b	$18,$2C,$6F,$0F,$87,$87,$B2,$39,$19,$BF,$85,$3F,$5F,$0F,$0B,$C3
	.dc.b	$1F,$C2,$E1,$EE,$3C,$EF,$E1,$EF,$96,$3F,$06,$31,$7C,$B0,$EA,$7A
	.dc.b	$9C,$4E,$EF,$FD,$13,$7C,$9E,$A6,$27,$6E,$0F,$93,$8F,$D7,$C3,$CB
	.dc.b	$18,$7F,$1D,$1F,$2E,$3B,$1F,$7A,$3F,$19,$78,$BF,$87,$7D,$AF,$E3
	.dc.b	$3F,$4A,$3C,$91,$8B,$FA,$0F,$E1,$B1,$F9,$11,$D9,$AF,$9D,$B1,$BC
	.dc.b	$7E,$14,$ED,$AE,$DB,$1B,$C3,$BE,$3E,$8A,$F8,$58,$7A,$1E,$53,$A7
	.dc.b	$EB,$57,$9D,$FC,$2E,$1F,$6D,$FD,$55,$F3,$9D,$1F,$2E,$32,$72,$8F
	.dc.b	$9B,$7F,$09,$B8,$FC,$76,$CF,$88,$FC,$8C,$65,$E6,$7D,$53,$F4,$78
	.dc.b	$79,$DC,$1F,$99,$BF,$B0,$FC,$38,$D5,$F8,$63,$9B,$5F,$B8,$79,$CC
	.dc.b	$D7,$85,$C9,$DB,$0F,$84,$E2,$79,$47,$D6,$C3,$F1,$27,$E6,$A7,$D9
	.dc.b	$87,$90,$FA,$8F,$B8,$CC,$6D,$F4,$19,$2C,$71,$8E,$B1,$F9,$D3,$89
	.dc.b	$4D,$BC,$E3,$B9,$C6,$3E,$49,$D5,$CC,$6C,$EB,$1F,$32,$BD,$B8,$FE
	.dc.b	$32,$73,$1F,$89,$9E,$E7,$86,$B9,$B6,$77,$87,$E8,$CF,$99,$87,$C7
	.dc.b	$9F,$1D,$79,$5F,$B5,$5E,$37,$2D,$72,$75,$32,$53,$27,$A6,$BB,$E1
	.dc.b	$E9,$8D,$99,$3E,$49,$4F,$59,$6B,$D6,$DA,$F5,$BF,$91,$BF,$C1,$9E
	.dc.b	$B7,$C3,$87,$C5,$AE,$2D,$AE,$33,$FA,$CB,$FE,$3B,$8F,$E5,$B1,$F3
	.dc.b	$B6,$7C,$B7,$C5,$F4,$B5,$EF,$3F,$35,$DB,$5F,$5C,$6F,$1F,$3D,$FC
	.dc.b	$39,$E2,$7A,$DF,$1B,$F8,$5C,$3C,$F8,$FE,$F6,$3F,$41,$87,$99,$D1
	.dc.b	$E4,$7C,$11,$F0,$1B,$3C,$8F,$E1,$07,$F0,$09,$F2,$C6,$9E,$47,$8F
	.dc.b	$1F,$E5,$D1,$98,$F3,$9F,$26,$32,$EC,$FE,$3B,$1B,$C3,$DF,$3F,$2E
	.dc.b	$7B,$9B,$BF,$E3,$37,$E4,$8F,$E0,$27,$CA,$34,$F9,$E3,$D9,$18,$36
	.dc.b	$74,$3C,$87,$88,$D1,$F2,$0E,$86,$9F,$0B,$FC,$44,$D1,$CD,$FA,$15
	.dc.b	$F8,$D3,$D7,$1F,$40,$F8,$71,$AC,$3E,$E4,$EA,$3F,$58,$DA,$FD,$53
	.dc.b	$6B,$F5,$0D,$7F,$3C,$7D,$23,$CD,$1F,$04,$FB,$4F,$9E,$3C,$95,$F8
	.dc.b	$8A,$EE,$DA,$EE,$DA,$EE,$D7,$B9,$EA,$8F,$A4,$FC,$29,$F9,$EF,$88
	.dc.b	$F9,$EE,$0F,$81,$7F,$75,$C4,$F2,$BF,$2C,$FE,$1A,$B9,$1D,$67,$D6
	.dc.b	$FF,$3D,$BF,$26,$3F,$A7,$6E,$3F,$8B,$7F,$C8,$1C,$1F,$E9,$8A,$11
	.dc.b	$C3,$78,$3F,$D3,$0F,$FA,$77,$9F,$6F,$FA,$63,$A7,$FD,$14,$7F,$A1
	.dc.b	$0B,$97,$FA,$02,$1C,$CE,$67,$17,$FA,$67,$83,$FD,$0D,$5F,$E8,$EB
	.dc.b	$FF,$45,$1C,$6F,$85,$6E,$33,$3B,$3F,$D1,$87,$FA,$6A,$B9,$BE,$6E
	.dc.b	$3F,$F4,$D3,$4F,$F4,$C1,$C4,$F2,$E1,$E4,$7F,$D4,$67,$FA,$50,$F7
	.dc.b	$88,$AE,$1F,$F4,$93,$8E,$1F,$F4,$C3,$FE,$9A,$E3,$FF,$4C,$37,$8F
	.dc.b	$FD,$26,$FD,$09,$FF,$3B,$BF,$E9,$73,$97,$07,$FA,$69,$FC,$C1,$F9
	.dc.b	$BC,$7F,$0E,$FE,$A6,$7F,$41,$C5,$F7,$F8,$FF,$0D,$5A,$3F,$36,$E3
	.dc.b	$4F,$C6,$61,$FA,$E3,$F0,$B8,$7E,$FB,$0F,$C7,$BF,$C0,$4F,$C7,$B8
	.dc.b	$BF,$DF,$C7,$E3,$EF,$A3,$FA,$BA,$E9,$18,$8D,$1C,$E7,$C9,$7F,$AC
	.dc.b	$9D,$61,$FD,$D9,$FD,$14,$7E,$72,$31,$5E,$FB,$F3,$BE,$47,$BB,$D6
	.dc.b	$7D,$B5,$CA,$3E,$44,$61,$E5,$1E,$0E,$3F,$E8,$58,$7C,$69,$E4,$7A
	.dc.b	$E7,$93,$FB,$D3,$07,$23,$67,$BA,$3E,$2E,$1E,$87,$A1,$8A,$EF,$7F
	.dc.b	$06,$B8,$D6,$E3,$15,$B9,$FE,$5E,$FC,$87,$F8,$8D,$FA,$5C,$CF,$B6
	.dc.b	$7B,$BF,$88,$8F,$41,$E5,$71,$8F,$E2,$30,$FE,$09,$7C,$EB,$F9,$C5
	.dc.b	$73,$3C,$67,$0C,$7E,$8E,$1E,$39,$FE,$43,$87,$37,$F1,$53,$F5,$27
	.dc.b	$D5,$1D,$0F,$92,$F9,$9F,$E4,$75,$F7,$F0,$F8,$B1,$EE,$3E,$2B,$87
	.dc.b	$BD,$78,$A7,$CD,$1B,$AE,$E7,$F2,$A7,$BC,$7A,$9E,$E7,$E1,$0C,$9C
	.dc.b	$5F,$76,$1F,$0E,$79,$4F,$C9,$38,$9F,$33,$1F,$B0,$F1,$33,$8F,$C3
	.dc.b	$7F,$2D,$5F,$8A,$8C,$57,$C9,$AF,$DC,$57,$B8,$FB,$67,$90,$F7,$D7
	.dc.b	$8E,$FE,$34,$68,$F2,$1F,$D0,$CF,$61,$8B,$FD,$C4,$7B,$23,$B1,$F4
	.dc.b	$CD,$61,$F5,$E3,$58,$7C,$A7,$BC,$EF,$0F,$7B,$E4,$39,$CF,$C6,$8C
	.dc.b	$9E,$C7,$94,$79,$23,$38,$FE,$42,$38,$D7,$6A,$E5,$3F,$7E,$B9,$4F
	.dc.b	$90,$95,$F3,$AF,$F0,$4F,$C7,$9D,$B9,$C7,$E4,$3E,$1B,$E8,$78,$70
	.dc.b	$F2,$C7,$DB,$8C,$BF,$9A,$9F,$B1,$7D,$A3,$E3,$C6,$0E,$AF,$13,$EA
	.dc.b	$5F,$B9,$E4,$7F,$07,$BE,$F8,$FE,$0B,$1F,$CC,$CF,$CB,$9C,$E1,$DE
	.dc.b	$F9,$4F,$E2,$8E,$55,$F0,$9F,$C0,$1E,$13,$EB,$63,$EA,$78,$5F,$EB
	.dc.b	$DD,$E1,$E1,$7E,$44,$FC,$3A,$E9,$3F,$88,$9F,$94,$7D,$63,$E5,$1E
	.dc.b	$E9,$ED,$5F,$9B,$79,$47,$08,$F5,$E3,$F1,$AB,$C1,$3F,$28,$F5,$9F
	.dc.b	$D8,$F1,$4F,$F8,$53,$81,$8F,$F8,$43,$26,$5F,$E1,$29,$CE,$1F,$E1
	.dc.b	$A9,$CB,$C6,$BF,$C2,$F8,$B1,$EB,$2B,$E2,$B8,$FF,$C4,$2D,$E4,$71
	.dc.b	$FF,$80,$23,$5C,$9F,$E2,$47,$FC,$47,$5C,$1F,$F1,$5F,$0F,$F8,$41
	.dc.b	$A7,$F8,$40,$FF,$09,$6F,$FE,$16,$72,$D7,$FC,$1C,$6C,$CE,$9F,$E2
	.dc.b	$07,$8B,$5C,$9F,$E2,$28,$F5,$B8,$3F,$C3,$BC,$7F,$E2,$97,$53,$FE
	.dc.b	$1A,$AF,$64,$79,$8F,$F1,$01,$FE,$27,$3F,$C4,$3A,$FF,$87,$1F,$F0
	.dc.b	$C7,$17,$F8,$71,$BC,$5F,$E2,$27,$FC,$34,$43,$E8,$EF,$FE,$1B,$7E
	.dc.b	$F7,$17,$CB,$BF,$C7,$47,$E5,$27,$F4,$1C,$5F,$8C,$CB,$F4,$BC,$5F
	.dc.b	$BC,$AD,$61,$F8,$BB,$DC,$7E,$4C,$87,$E8,$8B,$AF,$F0,$C6,$DF,$F0
	.dc.b	$F2,$3F,$9C,$38,$12,$38,$36,$36,$FE,$BE,$32,$48,$CE,$3F,$9E,$C3
	.dc.b	$F7,$33,$FA,$5A,$FE,$06,$FE,$CE,$7E,$09,$2F,$F9,$94,$FE,$A6,$FF
	.dc.b	$67,$18,$8F,$82,$7E,$CC,$F1,$47,$82,$3E,$1C,$79,$CF,$65,$79,$5B
	.dc.b	$3B,$C3,$E5,$12,$3A,$5F,$A7,$1F,$A5,$1F,$C1,$8F,$4D,$7F,$6B,$7C
	.dc.b	$77,$FB,$E7,$8E,$1E,$28,$F0,$47,$98,$F0,$90,$C1,$2F,$C0,$FD,$63
	.dc.b	$4E,$1F,$61,$B9,$ED,$5E,$B8,$F0,$E1,$EB,$BF,$21,$27,$E5,$C7,$F3
	.dc.b	$99,$D9,$27,$4F,$D1,$21,$EE,$9E,$2F,$F1,$83,$06,$AF,$ED,$57,$5A
	.dc.b	$E7,$3A,$24,$E9,$E4,$D8,$FC,$25,$FD,$F8,$C5,$FC,$F6,$D7,$F8,$3C
	.dc.b	$E7,$3B,$25,$F8,$6B,$CF,$7D,$23,$15,$E7,$AD,$BE,$26,$CE,$E3,$13
	.dc.b	$C1,$EE,$DB,$F1,$C7,$F5,$9B,$EE,$79,$EF,$9B,$6B,$84,$62,$7D,$55
	.dc.b	$CB,$1F,$DF,$37,$1F,$E9,$73,$E0,$3C,$64,$C7,$CC,$4C,$7B,$46,$5B
	.dc.b	$19,$8E,$F8,$78,$5F,$41,$F0,$F1,$F2,$91,$E4,$70,$71,$3C,$E3,$89
	.dc.b	$E5,$71,$8F,$AA,$BB,$36,$B8,$DF,$90,$97,$D4,$EE,$74,$C7,$E1,$3F
	.dc.b	$B0,$35,$87,$F7,$BD,$BF,$AB,$93,$7F,$EA,$93,$FE,$FB,$E1,$FE,$B2
	.dc.b	$FF,$3E,$DF,$FB,$23,$FE,$E8,$E2,$FF,$7C,$65,$FE,$FC,$E4,$FF,$77
	.dc.b	$CF,$F6,$58,$FE,$8C,$D3,$83,$89,$FF,$7F,$C7,$FB,$AE,$78,$E5,$FD
	.dc.b	$4A,$3F,$AD,$BF,$D7,$6F,$C3,$1F,$D5,$0E,$87,$FB,$D6,$B9,$BF,$D4
	.dc.b	$4F,$09,$FD,$50,$FE,$AC,$7F,$BA,$CF,$F7,$87,$0F,$F5,$35,$A7,$F5
	.dc.b	$76,$C6,$63,$FD,$DF,$1F,$D5,$35,$FF,$79,$BC,$37,$FE,$AC,$FD,$BC
	.dc.b	$FE,$AB,$F4,$E3,$ED,$F0,$7E,$52,$FE,$AC,$FF,$8B,$AB,$7C,$1F,$7E
	.dc.b	$3E,$DB,$74,$FC,$2D,$F1,$C3,$9D,$7D,$D3,$B6,$3F,$86,$7E,$FE,$1D
	.dc.b	$2B,$84,$E6,$76,$76,$8C,$B4,$E9,$87,$86,$B9,$46,$27,$78,$79,$23
	.dc.b	$04,$3D,$87,$88,$87,$81,$E0,$E8,$C4,$F6,$71,$5E,$17,$8B,$5D,$98
	.dc.b	$AF,$1C,$F1,$AF,$ED,$13,$CA,$72,$EC,$F3,$46,$89,$3F,$E0,$89,$DC
	.dc.b	$F6,$25,$78,$49,$1A,$6C,$74,$9E,$4D,$3F,$B1,$1E,$33,$07,$F9,$6E
	.dc.b	$BE,$19,$E2,$76,$D7,$51,$ED,$3A,$BC,$C9,$1B,$9F,$F0,$4C,$EA,$31
	.dc.b	$5F,$DD,$2B,$99,$F0,$CE,$33,$B2,$3E,$26,$BC,$0F,$87,$1F,$04,$8F
	.dc.b	$B2,$BC,$2F,$12,$1C,$5F,$8E,$F8,$0C,$BC,$1D,$47,$47,$46,$0E,$CF
	.dc.b	$AE,$38,$11,$C1,$23,$FC,$08,$7F,$80,$6F,$FB,$F5,$7F,$50,$8F,$85
	.dc.b	$7F,$19,$DE,$3F,$E0,$98,$FF,$09,$B6,$7F,$C1,$11,$FD,$DA,$3F,$76
	.dc.b	$F1,$9F,$F0,$7C,$6D,$FF,$09,$E3,$FE,$10,$6D,$F8,$9F,$E9,$47,$EC
	.dc.b	$4F,$7E,$3E,$B7,$C7,$3C,$5F,$51,$FE,$11,$C3,$D4,$FD,$43,$E2,$9C
	.dc.b	$2B,$B3,$FC,$54,$F7,$C7,$DC,$AE,$D1,$E8,$AE,$D7,$E9,$8E,$F8,$7B
	.dc.b	$E7,$F8,$8C,$7F,$B3,$73,$FF,$51,$DF,$FA,$AA,$3F,$9A,$B6,$3F,$D4
	.dc.b	$F7,$FE,$AB,$C7,$F9,$DF,$07,$F5,$2D,$FF,$A7,$E0,$3F,$D4,$BC,$1F
	.dc.b	$D5,$1F,$F5,$67,$1F,$FA,$41,$FF,$4F,$E3,$FD,$01,$FF,$53,$B5,$FE
	.dc.b	$93,$AF,$FA,$7D,$A7,$FA,$68,$FE,$A1,$C8,$E4,$FF,$4F,$47,$F4,$DD
	.dc.b	$FF,$A9,$71,$77,$24,$7F,$63,$3F,$A4,$F2,$7F,$A6,$5F,$E9,$87,$FA
	.dc.b	$83,$8F,$FA,$47,$17,$F4,$4E,$1F,$E9,$33,$FE,$9A,$3D,$C7,$F4,$B2
	.dc.b	$9F,$82,$C3,$FC,$55,$A7,$CC,$E3,$FF,$6C,$6B,$FE,$DB,$6D,$73,$9F
	.dc.b	$A9,$C5,$FC,$F7,$8F,$F5,$B8,$7E,$4C,$FC,$CE,$5F,$70,$93,$F8,$4C
	.dc.b	$BF,$33,$97,$E0,$F2,$FD,$46,$5F,$83,$AD,$5F,$E0,$30,$E9,$7F,$A2
	.dc.b	$3F,$23,$8F,$81,$E5,$7F,$91,$3F,$2C,$7B,$5F,$E3,$4F,$F0,$0C,$3B
	.dc.b	$D7,$9B,$0E,$E7,$AE,$3F,$81,$47,$82,$7F,$52,$74,$3F,$38,$7E,$84
	.dc.b	$FD,$4C,$74,$AF,$34,$78,$4F,$D0,$9E,$13,$CD,$3F,$29,$FD,$93,$DD
	.dc.b	$FD,$0C,$7B,$A7,$C6,$FB,$A3,$E3,$1E,$3C,$7E,$5B,$D4,$CD,$75,$3F
	.dc.b	$9F,$CF,$17,$E0,$47,$1C,$7E,$74,$E4,$F4,$C7,$BB,$1F,$9C,$F5,$AE
	.dc.b	$27,$59,$FF,$2D,$E5,$F8,$99,$FF,$14,$D7,$EC,$E7,$F2,$13,$FB,$4B
	.dc.b	$65,$AC,$BF,$1B,$83,$0E,$0F,$E3,$F0,$F8,$4F,$F0,$E3,$DA,$E1,$FA
	.dc.b	$27,$CC,$71,$8F,$CB,$CB,$F8,$61,$31,$FC,$9D,$7C,$1C,$7F,$00,$F5
	.dc.b	$AF,$6D,$F2,$3E,$24,$7E,$0A,$3E,$03,$F7,$0F,$E8,$C7,$88,$F2,$9F
	.dc.b	$81,$3F,$0C,$F8,$EB,$C2,$F2,$8C,$CF,$67,$F0,$27,$1C,$3E,$8B,$DA
	.dc.b	$79,$5F,$E6,$23,$F1,$35,$C7,$0F,$74,$E6,$BF,$03,$5F,$66,$BE,$ED
	.dc.b	$7E,$22,$3A,$47,$1C,$3C,$AF,$C1,$8E,$07,$C1,$70,$F0,$32,$7A,$0F
	.dc.b	$C4,$63,$F9,$87,$35,$E4,$7E,$81,$E8,$33,$5E,$07,$FA,$AE,$5F,$8B
	.dc.b	$7F,$9E,$D7,$E1,$AB,$F9,$E4,$7F,$9D,$1B,$3B,$25,$FE,$40,$FE,$43
	.dc.b	$C7,$FD,$07,$87,$F9,$0D,$FF,$41,$87,$17,$F4,$15,$C7,$FD,$1A,$7F
	.dc.b	$D1,$BA,$7F,$A1,$53,$49,$87,$F4,$1E,$1F,$E7,$33,$FE,$83,$0F,$F3
	.dc.b	$97,$FA,$86,$7F,$E8,$12,$70,$7F,$3A,$DF,$FC,$FF,$1F,$CE,$A7,$FD
	.dc.b	$01,$1E,$87,$E6,$9F,$97,$BF,$B7,$BF,$EE,$A3,$E5,$CF,$E0,$67,$F9
	.dc.b	$0C,$7D,$72,$70,$7F,$72,$CB,$F3,$F1,$F8,$FE,$0F,$B0,$E9,$FE,$38
	.dc.b	$FE,$04,$D1,$E1,$30,$79,$0F,$D1,$47,$47,$0F,$E6,$1F,$C0,$BF,$94
	.dc.b	$7B,$1F,$B9,$3A,$9C,$AB,$DB,$7F,$B6,$8C,$1E,$0C,$7F,$35,$87,$AE
	.dc.b	$B7,$3F,$8D,$3A,$9F,$B4,$3C,$6F,$E7,$EB,$E3,$47,$EE,$E7,$93,$D7
	.dc.b	$1F,$98,$F2,$C3,$D9,$87,$83,$1F,$8B,$3F,$A1,$8C,$3E,$B3,$B9,$F5
	.dc.b	$0F,$83,$8F,$D4,$C3,$BC,$71,$C3,$BC,$62,$BF,$51,$87,$F0,$78,$FE
	.dc.b	$06,$66,$7F,$51,$5D,$8F,$35,$FC,$2A,$F9,$D5,$BC,$7D,$F5,$FC,$40
	.dc.b	$F2,$4F,$5C,$7F,$03,$5B,$9F,$39,$F1,$63,$DF,$8F,$93,$0F,$E0,$75
	.dc.b	$BB,$E5,$7F,$85,$3D,$98,$7D,$98,$FC,$11,$A8,$FB,$66,$8F,$B1,$3C
	.dc.b	$EB,$AE,$1E,$0C,$7D,$F8,$FC,$EB,$FB,$8D,$C3,$EA,$CF,$CA,$AE,$18
	.dc.b	$FC,$0C,$7E,$4C,$63,$0F,$69,$F0,$CF,$92,$7E,$06,$7D,$18,$FE,$1E
	.dc.b	$37,$5C,$9F,$BD,$8F,$86,$7E,$81,$F7,$6F,$B6,$3E,$F3,$E0,$C7,$B2
	.dc.b	$B8,$E3,$E7,$7F,$CA,$93,$FD,$9F,$1B,$5F,$C6,$B4,$FE,$29,$5F,$C6
	.dc.b	$5A,$B5,$FE,$37,$D1,$FC,$1D,$FE,$37,$C1,$FE,$4C,$E1,$FF,$2C,$70
	.dc.b	$7F,$98,$5B,$7F,$E4,$F3,$F8,$C4,$7F,$17,$5B,$7F,$16,$E2,$C3,$FC
	.dc.b	$54,$9B,$7F,$94,$49,$AF,$C8,$AF,$C6,$EB,$EF,$7E,$BC,$FB,$DB,$3B
	.dc.b	$7E,$89,$F8,$3D,$BF,$33,$BF,$CA,$BE,$51,$F7,$AB,$87,$07,$E1,$EF
	.dc.b	$F8,$4C,$FC,$E6,$D7,$D9,$8F,$A4,$D3,$99,$F6,$23,$F3,$C7,$D4,$AE
	.dc.b	$B7,$E1,$8E,$67,$B0,$A7,$A3,$1F,$D2,$38,$C3,$C8,$FE,$D5,$B1,$F8
	.dc.b	$78,$FC,$71,$FA,$09,$EC,$ED,$B1,$B8,$F8,$97,$F8,$D7,$CC,$71,$2E
	.dc.b	$1F,$21,$B1,$B7,$E7,$9B,$2C,$6D,$B1,$B8,$F9,$D1,$B6,$9B,$2C,$6C
	.dc.b	$F9,$51,$FA,$BB,$F7,$E3,$F7,$EB,$73,$F3,$8E,$E5,$72,$F1,$C3,$E0
	.dc.b	$CF,$1B,$EB,$7F,$B1,$AE,$26,$CF,$79,$CE,$37,$5F,$C7,$0E,$78,$FD
	.dc.b	$C9,$D9,$C4,$E7,$5D,$70,$F1,$B6,$F9,$CF,$F1,$DC,$7E,$FB,$F5,$CE
	.dc.b	$73,$B7,$EE,$D7,$C7,$AE,$31,$FC,$82,$B8,$9F,$74,$B8,$FB,$E3,$8C
	.dc.b	$FC,$D6,$9E,$D2,$9F,$E6,$9C,$7F,$D7,$F7,$FE,$29,$3F,$85,$47,$F8
	.dc.b	$62,$7F,$C4,$8F,$FB,$87,$0F,$F3,$7B,$C4,$FE,$09,$1A,$7D,$87,$13
	.dc.b	$FD,$99,$3D,$34,$FF,$17,$C7,$3B,$FF,$1B,$61,$FE,$7E,$84,$AF,$F8
	.dc.b	$D6,$BF,$C5,$78,$3F,$8C,$EF,$FC,$62,$FF,$C6,$AF,$F1,$87,$F8,$A1
	.dc.b	$FE,$38,$DF,$FC,$5A,$D3,$FD,$CA,$43,$FD,$C9,$BF,$F1,$6C,$BF,$89
	.dc.b	$C7,$F9,$03,$6F,$E2,$FB,$FF,$8B,$A7,$33,$FE,$E4,$9C,$BB,$8C,$BF
	.dc.b	$EE,$93,$4E,$1F,$F7,$71,$89,$E8,$47,$93,$5D,$18,$21,$82,$1D,$C8
	.dc.b	$76,$29,$FE,$44,$38,$98,$9F,$4B,$C8,$E0,$43,$81,$92,$9C,$67,$D8
	.dc.b	$E0,$E1,$1D,$1D,$4F,$AD,$A7,$88,$DC,$F2,$9E,$11,$89,$E0,$D3,$81
	.dc.b	$DC,$91,$C1,$B3,$CC,$93,$CA,$32,$D9,$E0,$49,$DC,$F6,$8C,$4E,$E7
	.dc.b	$33,$F0,$E7,$2E,$E3,$2F,$B8,$D3,$87,$D4,$62,$7A,$11,$E4,$D7,$46
	.dc.b	$08,$60,$87,$72,$1D,$8A,$7C,$03,$89,$89,$F4,$BC,$8E,$04,$38,$19
	.dc.b	$29,$C6,$7D,$8E,$0E,$11,$D1,$D4,$FA,$DA,$78,$8D,$CF,$29,$E1,$18
	.dc.b	$9E,$0D,$38,$1D,$C9,$1C,$1B,$3C,$C9,$3C,$A3,$2D,$9E,$04,$9D,$CF
	.dc.b	$68,$C4,$EE,$73,$3F,$0E,$72,$EE,$32,$FB,$8D,$38,$7D,$46,$27,$A1
	.dc.b	$1E,$4D,$74,$60,$86,$08,$77,$21,$D8,$A7,$C0,$38,$98,$9F,$4B,$C8
	.dc.b	$E0,$43,$81,$92,$9C,$67,$D8,$E0,$E1,$1D,$1D,$4F,$AD,$A7,$88,$DC
	.dc.b	$F2,$9E,$11,$89,$E0,$D3,$81,$DC,$91,$C1,$B3,$CC,$93,$CA,$32,$D9
	.dc.b	$E0,$49,$DC,$F6,$8C,$4E,$E7,$33,$F0,$E7,$2E,$E3,$2F,$B8,$D3,$87
	.dc.b	$D4,$62,$7A,$11,$E4,$D7,$46,$08,$60,$87,$72,$1D,$8A,$7C,$03,$89
	.dc.b	$89,$F4,$BC,$8E,$04,$38,$19,$29,$C6,$7D,$8E,$0E,$11,$D1,$D4,$FA
	.dc.b	$DA,$78,$8D,$CF,$29,$E1,$18,$9E,$0D,$3F,$9A,$70,$AD,$0F,$92,$76
	.dc.b	$E2,$3F,$D2,$75,$B7,$51,$9C,$13,$9A,$E3,$5D,$27,$37,$DA,$73,$3D
	.dc.b	$23,$FD,$3B,$5D,$63,$FC,$16,$7F,$83,$4E,$CE,$5B,$1C,$08,$EC,$F0
	.dc.b	$46,$DB,$1A,$7A,$90,$E6,$71,$3C,$85,$AF,$F5,$93,$67,$A3,$A3,$91
	.dc.b	$D8,$D4,$F7,$7A,$9A,$38,$18,$21,$C4,$D4,$F8,$1B,$3E,$08,$C4,$70
	.dc.b	$39,$9C,$4C,$3A,$70,$43,$64,$74,$D8,$EC,$49,$F5,$47,$19,$D9,$27
	.dc.b	$51,$F0,$08,$7A,$27,$8B,$D5,$B3,$C5,$B3,$CA,$78,$1C,$8F,$5B,$96
	.dc.b	$C7,$02,$3B,$3C,$D1,$B6,$C6,$9E,$A4,$39,$9C,$4F,$31,$67,$CC,$D9
	.dc.b	$E8,$E8,$E4,$76,$35,$3D,$DE,$A6,$8E,$06,$08,$71,$35,$3E,$06,$CF
	.dc.b	$82,$31,$1C,$0E,$67,$13,$0E,$9C,$10,$D9,$1D,$36,$3B,$12,$7D,$51
	.dc.b	$C6,$76,$49,$D4,$7C,$02,$1E,$89,$E2,$F5,$6C,$F1,$6C,$F2,$9E,$07
	.dc.b	$23,$D6,$E5,$B1,$C0,$8E,$CF,$34,$6D,$B1,$A7,$A9,$0E,$67,$13,$CC
	.dc.b	$59,$F3,$36,$7A,$3A,$39,$1D,$8D,$4F,$77,$A9,$A3,$81,$82,$1C,$4D
	.dc.b	$4F,$81,$B3,$E0,$8C,$47,$03,$99,$C4,$C3,$A7,$04,$36,$47,$4D,$8E
	.dc.b	$C4,$9F,$54,$71,$9D,$92,$75,$1F,$00,$87,$A2,$78,$BD,$5B,$3C,$5B
	.dc.b	$3C,$A7,$81,$C8,$F5,$B9,$6C,$70,$23,$B3,$CD,$1B,$6C,$69,$EA,$43
	.dc.b	$99,$C4,$F3,$16,$7C,$CD,$9E,$8E,$8E,$47,$63,$53,$DD,$EA,$68,$E0
	.dc.b	$60,$87,$13,$53,$E0,$6C,$F8,$23,$11,$C0,$E6,$71,$30,$E9,$C1,$0D
	.dc.b	$91,$D3,$63,$B1,$27,$D5,$1C,$67,$64,$9D,$47,$C0,$3F,$9C,$7E,$2A
	.dc.b	$20

;--------------------------------------------------------------------------------
;	FSUB.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
fsub_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#63,-(sp)
	peamsg	'FSUB.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FSUB.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..XRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a3	destination handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp3	destination
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	fsub_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;destination handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected result handle,expected status,...
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsub.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSUB.X FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	lea.l	push_decompressed,a3	;destination handle,...
33:
	lea.l	push_decompressed,a2	;source handle,...
22:
	move.l	#0,d1		;fpcr=(XRN..XRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	([a3]),fp3		;destination
	fmove.x	fp3,fp7			;actual result=destination
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	fsub.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FSUB.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn='
	puthex24	([a3]),(4,[a3]),(8,[a3])	;destination
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#48,d1		;fpcr=(XRN..XRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	addq.l	#4,a3			;destination handle,...
	tst.l	(a3)			;destination handle,...
	bpl	33b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
fsub_data_compressed::
	.dc.b	$00,$0E,$B5,$7C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$E1,$8B
	.dc.b	$E0,$AF,$C7,$73,$D1,$95,$CB,$1F,$75,$9F,$C1,$B5,$F9,$56,$FF,$36
	.dc.b	$E1,$FA,$77,$8F,$89,$F4,$77,$F6,$EF,$37,$2B,$F0,$3B,$F9,$37,$A7
	.dc.b	$A1,$CB,$7F,$44,$F6,$75,$3F,$63,$F7,$76,$3E,$F3,$E0,$77,$7C,$01
	.dc.b	$D9,$8E,$2F,$89,$DE,$F9,$3B,$F9,$BF,$C7,$F0,$BE,$07,$83,$F2,$97
	.dc.b	$EA,$99,$FD,$DB,$E6,$77,$3B,$3E,$77,$8A,$F3,$3E,$87,$A8,$FC,$52
	.dc.b	$FE,$41,$8E,$AC,$FE,$97,$AD,$3E,$B7,$B1,$F5,$8F,$9B,$BF,$DA,$7E
	.dc.b	$0B,$7C,$AD,$FC,$EF,$0F,$BF,$E3,$E3,$F2,$F1,$F9,$F4,$74,$FD,$77
	.dc.b	$AF,$ED,$1D,$BF,$73,$FC,$27,$BD,$E7,$D3,$F0,$EB,$DF,$CF,$D4,$6B
	.dc.b	$CB,$DF,$E2,$7C,$3F,$1D,$E3,$F9,$47,$2F,$D2,$39,$FE,$A5,$D3,$F7
	.dc.b	$AF,$C8,$64,$F4,$BF,$25,$F1,$3F,$33,$62,$F5,$F3,$E5,$EB,$D2,$DF
	.dc.b	$6F,$0E,$DF,$CD,$79,$6F,$5A,$B8,$F4,$B3,$DF,$AF,$1F,$7F,$92,$F0
	.dc.b	$FC,$BF,$8F,$7F,$2E,$FE,$7E,$A7,$4F,$63,$AF,$83,$DB,$C1,$EF,$E0
	.dc.b	$F8,$3F,$7E,$FD,$57,$C6,$BD,$5B,$E6,$E3,$B3,$3F,$29,$AF,$9F,$DF
	.dc.b	$AB,$C3,$C7,$E3,$E3,$F2,$FC,$23,$9F,$E2,$5D,$3C,$9E,$BE,$4F,$6F
	.dc.b	$27,$BF,$E6,$FE,$0F,$63,$C3,$E5,$F8,$BC,$BF,$1F,$EB,$DE,$4F,$73
	.dc.b	$F7,$DF,$28,$F9,$6B,$F4,$18,$F8,$5C,$F9,$BA,$F6,$37,$EC,$70,$FD
	.dc.b	$BF,$8F,$BD,$FC,$1F,$E6,$3E,$77,$A9,$FC,$29,$FD,$4B,$FC,$33,$E6
	.dc.b	$9F,$82,$FF,$0D,$5E,$DD,$F3,$31,$E6,$67,$CA,$D7,$95,$BF,$2B,$F8
	.dc.b	$A7,$CE,$7C,$BC,$FF,$8B,$BF,$93,$4F,$ED,$5C,$F8,$5C,$3C,$0E,$3E
	.dc.b	$07,$2F,$03,$9F,$77,$4E,$EE,$BF,$13,$DB,$DB,$FE,$45,$E3,$BE,$66
	.dc.b	$F7,$63,$BB,$3F,$7F,$AF,$C2,$77,$DD,$C3,$BB,$8F,$B5,$CB,$DA,$E7
	.dc.b	$FB,$37,$F2,$DF,$AC,$72,$DF,$1B,$F9,$87,$D6,$FF,$13,$E8,$FE,$66
	.dc.b	$B3,$FE,$69,$F5,$DF,$32,$4B,$D8,$FE,$71,$E4,$7E,$CE,$7F,$9D,$AE
	.dc.b	$8F,$E7,$9F,$5F,$FA,$B1,$C3,$8F,$1F,$FA,$07,$D8,$7F,$15,$C3,$FA
	.dc.b	$1F,$D8,$BC,$FF,$E8,$9E,$9B,$A2,$F3,$63,$CC,$FE,$91,$F6,$4F,$88
	.dc.b	$FE,$95,$F6,$7F,$5F,$32,$F5,$33,$C3,$AF,$0F,$7D,$1C,$3C,$3F,$EA
	.dc.b	$5F,$69,$FB,$FC,$7F,$AA,$AF,$03,$FA,$B7,$C5,$FE,$42,$4F,$C5,$FB
	.dc.b	$6F,$CE,$47,$F5,$C7,$D0,$DE,$9F,$EB,$D1,$D6,$0F,$57,$87,$46,$BE
	.dc.b	$25,$FC,$03,$0B,$3D,$61,$F3,$7F,$DA,$BD,$E7,$99,$C6,$30,$DA,$FE
	.dc.b	$50,$D9,$DC,$7F,$72,$F8,$17,$BF,$FC,$8E,$3F,$B9,$88,$CF,$C6,$E2
	.dc.b	$31,$E6,$FF,$42,$7F,$BC,$CE,$9E,$2D,$73,$5D,$5D,$CF,$17,$F8,$A5
	.dc.b	$75,$AE,$11,$83,$9C,$ED,$F4,$57,$37,$47,$6A,$E5,$3B,$8F,$54,$ED
	.dc.b	$E3,$3D,$E3,$11,$EC,$F9,$3C,$4C,$8F,$5C,$EA,$D7,$D5,$F3,$CF,$43
	.dc.b	$75,$E2,$BE,$51,$E5,$32,$7B,$1C,$BE,$A2,$1F,$56,$BB,$5F,$84,$95
	.dc.b	$D6,$79,$9D,$48,$7F,$40,$AE,$AF,$94,$C9,$D5,$CB,$F1,$A2,$DF,$BA
	.dc.b	$6D,$78,$2B,$D0,$78,$0F,$39,$C4,$C9,$D8,$C1,$67,$A3,$C4,$F2,$1C
	.dc.b	$9C,$BB,$8F,$03,$F8,$89,$F1,$D7,$9A,$FE,$2B,$89,$F0,$9C,$EB,$8B
	.dc.b	$F0,$CF,$01,$E4,$3C,$13,$C9,$CD,$FD,$C7,$F0,$86,$27,$9B,$F2,$4C
	.dc.b	$57,$8E,$FC,$51,$EF,$BE,$8F,$AE,$7E,$C3,$4E,$F1,$A7,$D1,$1A,$69
	.dc.b	$A9,$EA,$FA,$1A,$FC,$A3,$C8,$75,$3C,$4E,$9F,$80,$68,$EA,$68,$F5
	.dc.b	$9D,$CA,$64,$E8,$78,$4C,$94,$C9,$E3,$3E,$49,$83,$C2,$60,$F0,$9C
	.dc.b	$4A,$71,$BF,$C1,$B8,$38,$B8,$75,$7F,$51,$E8,$62,$BE,$A3,$E6,$3A
	.dc.b	$5F,$CF,$74,$7E,$20,$A7,$BC,$F0,$57,$E0,$CF,$03,$E2,$AF,$04,$60
	.dc.b	$D3,$89,$ED,$3B,$AF,$B8,$F6,$3E,$01,$4F,$2D,$FD,$7B,$F0,$57,$27
	.dc.b	$AD,$72,$AE,$87,$D7,$3E,$74,$70,$AE,$57,$FC,$66,$F9,$3E,$58,$CD
	.dc.b	$F3,$AE,$B7,$CC,$F8,$35,$C6,$B7,$1F,$1C,$F0,$9F,$1C,$F5,$1F,$5A
	.dc.b	$FB,$DF,$D2,$8F,$2D,$73,$7E,$A1,$BA,$EA,$FE,$FC,$CD,$FA,$DB,$3C
	.dc.b	$5E,$75,$E9,$8F,$A6,$66,$3D,$31,$93,$E6,$5F,$92,$3D,$C7,$A2,$BB
	.dc.b	$3D,$0F,$73,$F1,$1F,$71,$F0,$0F,$B2,$53,$EC,$9E,$69,$F0,$3E,$A9
	.dc.b	$F0,$3C,$A7,$A3,$C9,$AF,$23,$C4,$7C,$32,$9F,$C6,$CA,$7E,$EC,$F5
	.dc.b	$57,$C1,$6D,$78,$9F,$C2,$D7,$89,$F0,$4F,$C8,$9D,$47,$E8,$27,$4F
	.dc.b	$8C,$F9,$67,$C4,$9E,$0D,$9F,$9B,$7F,$6A,$3C,$C7,$03,$D4,$53,$27
	.dc.b	$98,$E4,$7C,$77,$93,$ED,$3E,$71,$F8,$73,$E7,$1C,$0F,$6B,$C1,$FE
	.dc.b	$06,$70,$BF,$BA,$F8,$4E,$8F,$D0,$7A,$1F,$30,$E2,$E4,$E2,$F9,$CF
	.dc.b	$59,$4F,$59,$DE,$BE,$93,$6B,$E6,$BF,$8F,$BF,$9A,$F8,$0F,$21,$E0
	.dc.b	$9F,$9A,$F6,$39,$1E,$F3,$EB,$5F,$BE,$FF,$89,$4F,$09,$F1,$C7,$A6
	.dc.b	$3C,$6D,$BF,$1C,$FC,$5B,$F1,$CF,$BA,$3B,$9E,$17,$83,$E1,$3D,$57
	.dc.b	$E1,$9E,$37,$D8,$F1,$4F,$CE,$6B,$ED,$BF,$C3,$BF,$68,$ED,$1A,$BF
	.dc.b	$5C,$EA,$7E,$A4,$61,$FC,$31,$C4,$F7,$CF,$1A,$F1,$D7,$BA,$BE,$D4
	.dc.b	$62,$BD,$47,$C1,$BF,$51,$1E,$87,$A6,$BC,$4F,$D6,$AF,$13,$F9,$03
	.dc.b	$67,$B4,$F2,$9E,$8A,$F0,$3E,$E9,$E6,$FE,$24,$EC,$70,$3B,$47,$07
	.dc.b	$DB,$1C,$1F,$AA,$60,$E0,$73,$3C,$40,$FF,$41,$63,$1F,$F4,$66,$5F
	.dc.b	$E8,$86,$9F,$E8,$82,$9C,$4A,$7F,$A3,$4C,$1F,$E9,$CC,$FF,$D2,$AF
	.dc.b	$FA,$2B,$64,$63,$4F,$F4,$8C,$EE,$BC,$0D,$3C,$05,$8D,$BE,$9F,$95
	.dc.b	$FF,$49,$34,$AE,$13,$48,$E8,$87,$FA,$6D,$CB,$4C,$9F,$EA,$27,$2D
	.dc.b	$32,$7C,$07,$2D,$32,$7C,$77,$2D,$32,$7A,$4F,$01,$0E,$C5,$3A,$14
	.dc.b	$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08
	.dc.b	$7D,$03,$DA,$53,$CC,$53,$C2,$53,$B1,$4E,$A5,$3A,$BF,$29,$A7,$67
	.dc.b	$0E,$C8,$E0,$8E,$C8,$79,$1C,$B4,$C9,$E6,$72,$D3,$27,$A1,$CB,$4C
	.dc.b	$9E,$A7,$2D,$32,$7A,$8F,$01,$0E,$C5,$3A,$14,$E4,$53,$91,$C0,$87
	.dc.b	$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86,$08,$7D,$03,$DA,$53,$CC
	.dc.b	$53,$C2,$53,$B1,$4E,$A5,$3A,$BF,$29,$A7,$67,$0E,$C8,$E0,$8E,$C8
	.dc.b	$79,$1C,$B4,$C9,$E6,$72,$D3,$27,$A1,$CB,$4C,$9E,$A7,$2D,$32,$7A
	.dc.b	$8F,$01,$0E,$C5,$3A,$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34
	.dc.b	$43,$45,$34,$60,$86,$08,$7D,$03,$DA,$53,$CC,$53,$C2,$53,$B1,$4E
	.dc.b	$A5,$3A,$BF,$29,$A7,$67,$0E,$C8,$E0,$8E,$C8,$79,$1C,$B4,$C9,$E6
	.dc.b	$72,$D3,$27,$A1,$CB,$4C,$9E,$A7,$2D,$32,$7A,$8F,$01,$0E,$C5,$3A
	.dc.b	$14,$E4,$53,$91,$C0,$87,$02,$9C,$0A,$70,$34,$43,$45,$34,$60,$86
	.dc.b	$08,$7D,$03,$DA,$53,$CC,$53,$C2,$53,$B1,$4E,$A5,$3A,$BF,$29,$A7
	.dc.b	$67,$0E,$C8,$E0,$8E,$C8,$7C,$BB,$FF,$0A,$D7,$F8,$6A,$3F,$C3,$A4
	.dc.b	$3F,$C2,$87,$F8,$7A,$1B,$BE,$6F,$84,$A7,$31,$86,$C7,$F8,$8E,$78
	.dc.b	$3F,$E3,$61,$AB,$76,$BD,$3C,$E6,$8F,$85,$C3,$70,$F4,$4E,$A7,$27
	.dc.b	$F8,$9B,$E2,$F5,$3E,$D6,$BF,$E3,$86,$BF,$E3,$D6,$BF,$E4,$7F,$8C
	.dc.b	$A9,$A4,$9F,$F1,$9F,$C6,$E8,$8E,$0F,$F1,$BF,$27,$9F,$7F,$0F,$C7
	.dc.b	$59,$FF,$24,$CD,$C3,$FC,$9D,$35,$FF,$2C,$45,$51,$63,$0D,$9F,$F2
	.dc.b	$F5,$76,$69,$D8,$B5,$D9,$AF,$66,$BD,$9B,$5D,$9A,$6A,$BC,$13,$F8
	.dc.b	$38,$FF,$31,$C6,$2B,$C5,$39,$3F,$CD,$55,$CD,$C1,$E1,$29,$E1,$2B
	.dc.b	$B8,$FB,$2D,$AF,$64,$F0,$AF,$04,$78,$8F,$3C,$65,$B5,$E4,$39,$BE
	.dc.b	$B3,$27,$18,$F0,$B9,$69,$CC,$B1,$CE,$3C,$6D,$8F,$7C,$71,$8E,$F1
	.dc.b	$C5,$A6,$DF,$84,$D5,$5E,$F7,$99,$C6,$7B,$9E,$79,$E6,$D3,$31,$F7
	.dc.b	$CF,$3B,$97,$9C,$77,$9E,$D1,$E2,$9D,$34,$E5,$3D,$9B,$1D,$9F,$F5
	.dc.b	$75,$7D,$27,$27,$68,$E6,$FB,$1F,$88,$71,$7F,$01,$3D,$5E,$71,$F5
	.dc.b	$1E,$67,$8C,$C9,$EC,$32,$75,$3E,$31,$E1,$33,$5E,$E7,$53,$F0,$67
	.dc.b	$C7,$1B,$3A,$C7,$8D,$F2,$C7,$9E,$BD,$D1,$A8,$C1,$E7,$70,$73,$38
	.dc.b	$96,$38,$B8,$3D,$85,$3C,$B1,$F2,$2B,$D0,$47,$CF,$1C,$27,$99,$23
	.dc.b	$C5,$1C,$E7,$83,$F1,$0E,$11,$A9,$E1,$1F,$2C,$F4,$9A,$30,$73,$8F
	.dc.b	$0B,$E6,$3E,$6B,$89,$EF,$5F,$36,$7C,$06,$CE,$A5,$3A,$1A,$8E,$33
	.dc.b	$F0,$4E,$84,$8D,$36,$7B,$11,$E1,$1A,$24,$65,$F3,$98,$8D,$C6,$63
	.dc.b	$CB,$3E,$27,$71,$97,$CE,$DA,$F9,$2F,$E1,$DB,$5F,$84,$9F,$93,$18
	.dc.b	$7F,$05,$1F,$44,$8E,$9B,$3C,$DA,$73,$71,$1B,$23,$82,$46,$C9,$1E
	.dc.b	$03,$6E,$1F,$BD,$1C,$1D,$35,$CC,$6D,$FB,$13,$EB,$72,$70,$8D,$57
	.dc.b	$C9,$8F,$97,$1F,$5C,$F9,$E4,$7E,$6B,$5D,$9F,$54,$C9,$F6,$0F,$F0
	.dc.b	$C4,$7F,$76,$24,$7F,$70,$24,$7F,$85,$A3,$2D,$8F,$D2,$BF,$EF,$9F
	.dc.b	$A0,$74,$FF,$13,$AF,$DA,$CE,$63,$15,$C3,$91,$2B,$E2,$FE,$FE,$F8
	.dc.b	$CE,$A7,$33,$FB,$E9,$CB,$C6,$FA,$BF,$AE,$3C,$07,$3B,$F0,$3B,$30
	.dc.b	$6E,$FB,$4F,$2C,$7F,$8C,$4F,$23,$C1,$5C,$B0,$F1,$C7,$EC,$70,$F1
	.dc.b	$C6,$1F,$E2,$A7,$F8,$46,$39,$1B,$9E,$57,$DD,$FE,$75,$1A,$9E,$31
	.dc.b	$FE,$0C,$8C,$3F,$C5,$23,$A6,$1E,$89,$F5,$E1,$E8,$8C,$47,$A0,$F7
	.dc.b	$1F,$04,$C4,$6D,$C1,$E5,$37,$87,$AE,$FC,$07,$8E,$BA,$CE,$EB,$95
	.dc.b	$6E,$3F,$8F,$DF,$99,$EA,$FB,$1F,$F0,$B9,$EC,$37,$7D,$5D,$BE,$07
	.dc.b	$77,$E0,$9E,$57,$FE,$52,$8C,$56,$E3,$15,$C8,$EA,$72,$30,$FF,$55
	.dc.b	$AF,$35,$FA,$63,$15,$E3,$8C,$3F,$5E,$BF,$32,$7E,$6A,$FC,$2F,$6A
	.dc.b	$F0,$D7,$08,$FA,$B5,$FE,$87,$AD,$D7,$69,$F2,$56,$DF,$AB,$5B,$7C
	.dc.b	$EF,$85,$F0,$D7,$28,$C4,$78,$CD,$9E,$03,$CE,$7C,$FA,$F0,$1E,$6C
	.dc.b	$38,$E1,$F0,$9B,$5F,$05,$B5,$E6,$6D,$71,$8E,$C7,$F7,$23,$B4,$F9
	.dc.b	$1A,$68,$F3,$61,$EF,$7F,$15,$1E,$F8,$C5,$FE,$B2,$FB,$CE,$5B,$5E
	.dc.b	$CC,$3E,$13,$D4,$EE,$7C,$A3,$47,$C4,$30,$7C,$A3,$07,$88,$DD,$F8
	.dc.b	$AB,$F6,$F1,$E9,$BF,$8C,$D8,$F8,$CD,$3E,$31,$70,$F8,$CF,$80,$FE
	.dc.b	$F0,$79,$8F,$01,$E3,$AF,$03,$FB,$2C,$3D,$EE,$5A,$FC,$A6,$BC,$EF
	.dc.b	$E4,$3E,$1B,$F9,$0D,$3D,$B3,$C3,$0F,$93,$1E,$53,$94,$FC,$07,$D1
	.dc.b	$5F,$01,$B1,$F0,$1A,$7C,$02,$E1,$F0,$1F,$69,$F0,$2B,$BB,$83,$B9
	.dc.b	$EF,$3F,$C8,$01,$FE,$8F,$F3,$FF,$C3,$4F,$E6,$31,$90,$E7,$0F,$7C
	.dc.b	$7F,$A9,$71,$DD,$7E,$D9,$B1,$FC,$15,$A7,$F1,$92,$CF,$FA,$99,$B1
	.dc.b	$FE,$8A,$78,$D7,$78,$FC,$6B,$5F,$C6,$1F,$EF,$EA,$FF,$7F,$BF,$B5
	.dc.b	$30,$7E,$D0,$72,$39,$9C,$D9,$9C,$C6,$1E,$13,$B9,$D3,$63,$3C,$CC
	.dc.b	$F2,$D8,$CC,$60,$ED,$C4,$F4,$BC,$ED,$9C,$C6,$0F,$5F,$3B,$9D,$CF
	.dc.b	$E7,$8C,$37,$A7,$FD,$67,$7E,$F9,$CB,$F0,$70,$79,$2E,$3F,$1B,$67
	.dc.b	$4D,$7F,$8D,$CF,$09,$F4,$E3,$99,$FE,$6B,$3E,$59,$C9,$23,$E6,$74
	.dc.b	$6D,$FE,$76,$D7,$FC,$2A,$D4,$DE,$6F,$26,$BE,$3B,$67,$FE,$C2,$6E
	.dc.b	$1F,$E8,$E6,$E1,$FE,$89,$6D,$7F,$B6,$5B,$3F,$EC,$86,$CF,$FA,$E1
	.dc.b	$AF,$FA,$31,$AF,$FA,$85,$B7,$E4,$9F,$83,$7E,$32,$3E,$36,$D7,$8D
	.dc.b	$B7,$E3,$6C,$F8,$DB,$3E,$36,$BE,$36,$BE,$36,$D7,$8E,$3F,$C7,$F7
	.dc.b	$E6,$AF,$F2,$0D,$FC,$9B,$DC,$60,$F4,$9E,$72,$57,$9D,$B3,$E7,$69
	.dc.b	$E7,$2C,$79,$DA,$F9,$DA,$F9,$DB,$7E,$78,$5F,$9D,$FD,$54,$F9,$1F
	.dc.b	$11,$2B,$C4,$D8,$F1,$34,$F1,$14,$F1,$14,$F1,$16,$BC,$55,$E1,$AF
	.dc.b	$8D,$5E,$63,$E3,$57,$CC,$BF,$92,$F9,$8C,$46,$DC,$1E,$C8,$DC,$F4
	.dc.b	$8F,$04,$F4,$74,$F1,$8D,$1E,$E2,$3E,$E6,$C7,$B9,$AF,$B9,$AF,$B9
	.dc.b	$B7,$EE,$AF,$1D,$7C,$D9,$CB,$E3,$BF,$84,$E4,$F0,$12,$FC,$0D,$7C
	.dc.b	$0D,$7C,$0D,$BF,$03,$ED,$3A,$47,$22,$1C,$8A,$72,$2C,$72,$7E,$49
	.dc.b	$F3,$A7,$CE,$FD,$23,$35,$F2,$AB,$C4,$7B,$6F,$E0,$CE,$E7,$E0,$BC
	.dc.b	$E3,$2F,$D4,$39,$CF,$A1,$F3,$10,$F3,$16,$FC,$D5,$D0,$F1,$47,$AA
	.dc.b	$7C,$47,$C3,$7C,$4F,$99,$F1,$39,$7C,$51,$CE,$3C,$51,$E0,$24,$F8
	.dc.b	$23,$94,$F8,$27,$C9,$1E,$07,$D6,$78,$0F,$21,$E0,$32,$78,$1E,$67
	.dc.b	$82,$7B,$47,$09,$EC,$F8,$A7,$B4,$7F,$45,$9F,$E7,$51,$FD,$D1,$B3
	.dc.b	$FD,$BD,$AF,$F7,$56,$BF,$E6,$A7,$FD,$E8,$9F,$E1,$0D,$8F,$F7,$B3
	.dc.b	$FE,$86,$8C,$9C,$0B,$1A,$8F,$F7,$1C,$61,$FF,$7C,$18,$8F,$DA,$3D
	.dc.b	$4C,$96,$7D,$91,$CE,$30,$F9,$CC,$4F,$59,$DC,$FF,$04,$1C,$D9,$AE
	.dc.b	$11,$DA,$B0,$FA,$E3,$C7,$38,$8F,$31,$B8,$F3,$CF,$BD,$A7,$88,$B1
	.dc.b	$B3,$C5,$1B,$6C,$6D,$A6,$EB,$D2,$48,$FF,$4E,$D7,$8E,$7E,$09,$E7
	.dc.b	$3D,$CE,$6B,$B5,$71,$9D,$3F,$E0,$99,$D5,$78,$63,$53,$EE,$75,$1E
	.dc.b	$A7,$31,$E0,$07,$CC,$8D,$1F,$32,$37,$3D,$A7,$D9,$39,$23,$FE,$CE
	.dc.b	$8D,$4F,$09,$E7,$3E,$87,$6F,$82,$77,$18,$8F,$11,$F2,$A7,$CF,$5F
	.dc.b	$98,$AD,$CF,$33,$C1,$5D,$9F,$80,$F3,$8D,$C6,$0D,$9D,$63,$C1,$3D
	.dc.b	$5E,$51,$F1,$49,$3C,$9B,$1A,$9F,$AA,$F2,$3F,$1E,$64,$FA,$86,$63
	.dc.b	$EE,$B9,$7E,$59,$EB,$3C,$C7,$C5,$75,$5F,$2E,$3D,$06,$CF,$19,$83
	.dc.b	$89,$87,$E4,$C7,$8A,$38,$4F,$BD,$E0,$7B,$DC,$4F,$08,$F2,$9D,$23
	.dc.b	$33,$EB,$9F,$21,$B2,$B9,$3E,$91,$EA,$2C,$F8,$27,$E4,$BE,$48,$F9
	.dc.b	$2E,$5F,$43,$E1,$73,$3D,$DE,$6D,$8F,$89,$5F,$5E,$BC,$F1,$D4,$F5
	.dc.b	$9E,$FA,$E6,$79,$DE,$2F,$9D,$A7,$8C,$E3,$1E,$87,$C8,$7D,$C3,$55
	.dc.b	$F0,$9F,$03,$5D,$B5,$F1,$1D,$27,$DE,$E2,$3F,$72,$F4,$8F,$79,$E7
	.dc.b	$7E,$FD,$7B,$E3,$D7,$1F,$85,$3F,$DA,$1C,$BF,$EC,$7E,$6F,$F6,$7C
	.dc.b	$7E,$BA,$3F,$D8,$EF,$FB,$46,$33,$5F,$E9,$B9,$FF,$64,$C7,$F1,$8A
	.dc.b	$FF,$6E,$BF,$EC,$3E,$5B,$7F,$AC,$E6,$DB,$67,$4F,$FB,$81,$E4,$D3
	.dc.b	$FD,$FA,$7F,$B5,$DF,$0D,$7F,$14,$BF,$F7,$0C,$FF,$B4,$88,$74,$3F
	.dc.b	$8C,$9F,$C6,$B9,$3F,$D9,$8E,$0F,$F6,$9F,$2F,$FB,$43,$9B,$FD,$A9
	.dc.b	$19,$9D,$BC,$97,$37,$FB,$46,$34,$D7,$4D,$7C,$0D,$7F,$DC,$BE,$B7
	.dc.b	$FB,$AA,$4D,$24,$7F,$B6,$BD,$AF,$F6,$C9,$1C,$1F,$ED,$5E,$4F,$F6
	.dc.b	$BB,$FE,$D3,$6B,$ED,$E6,$FF,$5D,$F0,$6F,$9F,$FD,$CB,$F1,$3F,$EC
	.dc.b	$5E,$63,$49,$C1,$FE,$EA,$EC,$D1,$1C,$1F,$EF,$BC,$5D,$7F,$E0,$8E
	.dc.b	$CF,$F0,$4B,$67,$FB,$FB,$67,$FB,$B3,$79,$FF,$DB,$CD,$9F,$A8,$D9
	.dc.b	$FB,$2D,$7F,$A9,$B5,$F0,$C7,$F8,$32,$BC,$2D,$3C,$25,$8F,$0B,$6B
	.dc.b	$C2,$D3,$C2,$58,$F0,$B5,$F0,$B5,$F0,$9F,$E1,$19,$F0,$B6,$3C,$2D
	.dc.b	$3C,$25,$9F,$0B,$63,$C2,$D3,$C2,$53,$C2,$53,$C3,$08,$FC,$E4,$F8
	.dc.b	$A3,$13,$FA,$28,$C4,$E8,$F2,$CE,$A7,$FB,$FC,$E9,$E1,$3A,$38,$3E
	.dc.b	$A9,$E0,$A7,$83,$CA,$78,$3C,$8E,$0F,$29,$F8,$27,$29,$E3,$3C,$E7
	.dc.b	$8B,$CC,$E2,$73,$9E,$31,$CC,$F9,$2F,$37,$8C,$73,$38,$9C,$E3,$8B
	.dc.b	$CE,$38,$9C,$DF,$A3,$1C,$DE,$27,$37,$8B,$CE,$38,$BC,$CE,$2F,$37
	.dc.b	$EC,$C7,$33,$8B,$CE,$38,$BC,$CE,$27,$33,$8B,$CD,$FC,$0B,$6B,$F4
	.dc.b	$84,$7F,$2C,$D8,$FC,$43,$67,$C5,$3E,$09,$F1,$1E,$03,$C4,$78,$0F
	.dc.b	$11,$E0,$2C,$F3,$8F,$11,$E0,$24,$78,$1B,$3E,$08,$EC,$78,$0E,$C7
	.dc.b	$80,$EC,$59,$EC,$F8,$27,$99,$E0,$7B,$12,$7B,$3D,$0E,$C7,$42,$CF
	.dc.b	$47,$B4,$F4,$7B,$1C,$DE,$D3,$D2,$39,$57,$E6,$09,$3F,$90,$6C,$7E
	.dc.b	$21,$B3,$F8,$06,$CF,$D1,$6C,$FC,$46,$BE,$96,$BE,$08,$EC,$47,$B3
	.dc.b	$67,$B3,$4E,$C5,$8E,$CD,$7B,$35,$EC,$74,$24,$74,$6C,$74,$69,$D0
	.dc.b	$A7,$42,$9D,$23,$91,$23,$93,$63,$93,$5E,$4D,$79,$47,$02,$3C,$1A
	.dc.b	$F0,$6B,$C0,$D1,$0D,$14,$D1,$82,$18,$25,$7E,$B5,$FF,$06,$1F,$D1
	.dc.b	$0A,$FF,$26,$7F,$9D,$98,$8F,$ED,$51,$98,$FE,$7A,$64,$E6,$7E,$5D
	.dc.b	$FE,$77,$3F,$CD,$F1,$7C,$57,$F2,$89,$FE,$63,$18,$BF,$E4,$31,$89
	.dc.b	$D4,$7F,$2C,$9D,$5F,$F1,$CA,$D4,$F0,$35,$5C,$27,$F9,$4D,$70,$BF
	.dc.b	$E3,$37,$C2,$79,$5F,$03,$93,$C2,$F9,$57,$F2,$8B,$E5,$5F,$C4,$6F
	.dc.b	$94,$F4,$8E,$57,$D0,$E4,$F4,$8E,$57,$D2,$3F,$94,$DF,$4A,$FE,$1B
	.dc.b	$7D,$27,$B1,$D2,$FB,$4F,$43,$B3,$D2,$3B,$1D,$2F,$B4,$FF,$25,$BE
	.dc.b	$CF,$F0,$DB,$EC,$F8,$2B,$B1,$E0,$7B,$5F,$80,$EC,$F8,$23,$B1,$E0
	.dc.b	$7B,$5F,$81,$FE,$4B,$7E,$07,$F0,$93,$D9,$F0,$1D,$9F,$05,$76,$3C
	.dc.b	$0F,$68,$F0,$1D,$9F,$03,$DA,$BC,$0F,$F3,$2A,$F0,$5F,$8D,$B5,$C1
	.dc.b	$B3,$D0,$38,$BF,$0C,$0C,$DF,$6B,$FC,$ED,$FF,$9E,$2F,$83,$EA,$31
	.dc.b	$3B,$79,$5F,$E8,$E3,$16,$8C,$4F,$B2,$3D,$53,$A9,$CC,$79,$CC,$CE
	.dc.b	$EF,$C7,$1E,$7B,$F1,$BF,$3A,$B8,$9E,$C7,$E4,$57,$13,$CA,$7A,$0C
	.dc.b	$3E,$57,$0F,$86,$BB,$BF,$12,$37,$3D,$9F,$85,$1F,$5C,$E0,$F6,$7C
	.dc.b	$D7,$ED,$8E,$B3,$E5,$9F,$69,$C0,$F1,$1C,$23,$B9,$C0,$EE,$FB,$0E
	.dc.b	$11,$E3,$7E,$14,$7E,$2E,$3B,$57,$AA,$FE,$3D,$7E,$96,$BC,$CF,$B4
	.dc.b	$E6,$F7,$79,$9E,$DA,$E7,$1C,$27,$9C,$FE,$80,$F4,$9C,$E7,$6E,$23
	.dc.b	$75,$FB,$48,$F1,$1F,$A5,$3B,$9E,$D9,$E2,$7B,$5D,$BF,$A9,$AF,$87
	.dc.b	$1F,$2A,$37,$1E,$98,$DB,$F0,$CD,$CF,$C3,$BF,$F2,$EE,$9F,$CF,$E8
	.dc.b	$D2,$57,$FA,$B6,$3F,$DE,$BC,$DF,$EE,$4C,$3F,$A5,$74,$7F,$B8,$63
	.dc.b	$15,$FE,$EA,$E8,$47,$FB,$A9,$FF,$77,$DF,$EB,$F0,$FF,$41,$C6,$23
	.dc.b	$C3,$5F,$EF,$67,$FD,$CF,$75,$C9,$FE,$F3,$32,$3A,$3B,$BF,$EF,$53
	.dc.b	$FD,$EE,$77,$9F,$33,$FE,$EF,$9F,$61,$1F,$F6,$53,$FE,$F1,$E8,$91
	.dc.b	$FD,$AB,$9B,$FD,$E2,$FF,$BE,$7A,$3F,$DE,$AF,$FB,$BB,$4F,$EC,$B3
	.dc.b	$FE,$F4,$6F,$47,$FB,$BE,$31,$3C,$FC,$DF,$F7,$81,$1E,$2F,$FB,$6F
	.dc.b	$85,$EE,$52,$4F,$F7,$28,$CB,$7C,$0E,$04,$70,$7F,$BB,$CF,$F7,$A7
	.dc.b	$3F,$F7,$AE,$8F,$F7,$9C,$60,$FF,$03,$D7,$FB,$CB,$7F,$F7,$A9,$23
	.dc.b	$3D,$1F,$87,$E7,$FE,$0B,$0E,$EF,$F7,$11,$2B,$FD,$A8,$D8,$FF,$63
	.dc.b	$37,$7F,$F2,$C3,$67,$FD,$14,$D9,$FF,$3E,$B5,$FF,$39,$35,$FF,$34
	.dc.b	$B6,$BC,$6D,$9F,$DE,$CF,$95,$F1,$91,$F1,$B6,$BC,$6D,$3C,$65,$8F
	.dc.b	$1B,$5F,$1B,$5F,$1B,$6B,$C6,$D3,$C7,$5E,$59,$FE,$0E,$79,$4F,$19
	.dc.b	$2B,$C6,$D9,$F1,$B4,$F1,$94,$F1,$94,$F1,$96,$BC,$6D,$8F,$19,$E5
	.dc.b	$7C,$75,$E5,$AF,$F7,$5B,$67,$FB,$34,$FF,$67,$8C,$38,$6B,$98,$E9
	.dc.b	$3C,$9A,$72,$8F,$81,$1E,$E2,$47,$B9,$B1,$EE,$6B,$EE,$6B,$EE,$6D
	.dc.b	$7B,$9B,$3E,$E9,$F8,$07,$B9,$F1,$BE,$96,$BE,$93,$71,$E9,$6C,$7A
	.dc.b	$6B,$DF,$1E,$A8,$F4,$11,$F4,$35,$F4,$35,$F4,$36,$BD,$0D,$3D,$15
	.dc.b	$EA,$9F,$41,$E2,$3D,$05,$3D,$0E,$CF,$41,$63,$D0,$7B,$A3,$D0,$79
	.dc.b	$88,$79,$8A,$79,$8B,$3E,$66,$C7,$98,$F4,$3E,$69,$F0,$3E,$66,$BE
	.dc.b	$68,$DC,$79,$9A,$79,$A3,$D8,$79,$8F,$21,$0F,$21,$67,$C8,$D8,$F2
	.dc.b	$3E,$63,$C8,$F6,$7C,$8D,$7C,$86,$E7,$C8,$D8,$F2,$47,$AA,$3C,$87
	.dc.b	$88,$93,$FC,$96,$7F,$D4,$F0,$8C,$3F,$EA,$A6,$BF,$EB,$66,$BF,$A5
	.dc.b	$7F,$96,$9F,$EE,$E2,$C7,$FB,$8D,$B1,$FE,$DA,$6C,$FF,$B4,$DB,$3F
	.dc.b	$EC,$E6,$CF,$FB,$2D,$AF,$FB,$29,$AF,$FB,$2D,$BD,$18,$6C,$FF,$2C
	.dc.b	$70,$F8,$CC,$16,$7F,$8E,$C7,$A6,$33,$1E,$76,$D7,$9D,$A7,$9C,$B1
	.dc.b	$E7,$6B,$E7,$6B,$E7,$6D,$79,$23,$C3,$1E,$39,$3F,$34,$C1,$63,$C8
	.dc.b	$70,$2C,$70,$7F,$90,$D7,$AD,$D4,$7A,$5B,$3E,$96,$9E,$92,$9E,$92
	.dc.b	$9E,$92,$D7,$A6,$7C,$11,$FE,$95,$AF,$04,$62,$3F,$19,$5E,$A7,$CA
	.dc.b	$D7,$83,$5D,$9E,$C7,$E0,$47,$A4,$E9,$3D,$A3,$27,$AE,$32,$F3,$8D
	.dc.b	$4F,$80,$E8,$59,$E9,$3C,$A3,$E1,$47,$27,$67,$98,$A7,$99,$E2,$F3
	.dc.b	$8F,$11,$CD,$F9,$51,$CC,$DB,$CD,$AF,$38,$E2,$F3,$3C,$6F,$37,$E9
	.dc.b	$47,$38,$DC,$73,$6B,$CC,$E2,$73,$3C,$6F,$3A,$FB,$4D,$3F,$2F,$5F
	.dc.b	$AA,$9F,$C0,$B6,$BC,$11,$CC,$EC,$74,$31,$5D,$1B,$5D,$0F,$D0,$95
	.dc.b	$FC,$5B,$67,$AC,$73,$3D,$87,$A8,$A6,$4F,$09,$F1,$23,$D4,$FA,$1E
	.dc.b	$ED,$3B,$96,$3B,$BC,$4F,$39,$98,$E7,$1A,$AF,$85,$1F,$4D,$B3,$F0
	.dc.b	$FB,$7F,$C9,$E7,$F9,$43,$9E,$D7,$F8,$F9,$FF,$4B,$CF,$F8,$34,$FF
	.dc.b	$40,$14,$FF,$44,$97,$97,$FD,$06,$EB,$0F,$F4,$DB,$FE,$8B,$E8,$FF
	.dc.b	$4D,$BF,$E8,$A3,$FD,$23,$CF,$FE,$8E,$7F,$D2,$2D,$73,$1E,$1A,$EB
	.dc.b	$18,$39,$5F,$F0,$BA,$E1,$09,$FF,$49,$3E,$53,$FD,$26,$7F,$A4,$79
	.dc.b	$BF,$D1,$EF,$FA,$47,$9F,$FD,$2A,$FF,$8A,$F9,$BF,$D2,$A9,$FF,$47
	.dc.b	$B9,$E7,$FF,$49,$13,$3F,$F4,$BF,$85,$FD,$5A,$8D,$25,$7F,$A7,$3B
	.dc.b	$B4,$47,$07,$FA,$4F,$C1,$FF,$4C,$11,$C1,$FE,$92,$3F,$D1,$BD,$1F
	.dc.b	$E9,$56,$9F,$E9,$53,$FC,$89,$CE,$EE,$74,$FF,$A5,$FC,$0F,$F6,$AD
	.dc.b	$7F,$A5,$CF,$BD,$D1,$FE,$FD,$AF,$F0,$0F,$57,$1E,$6F,$F0,$0D,$23
	.dc.b	$0F,$F8,$08,$19,$9E,$24,$9E,$33,$C3,$7F,$F0,$1D,$6E,$73,$5B,$8E
	.dc.b	$86,$4B,$5C,$67,$FC,$12,$7F,$9A,$9C,$47,$F8,$48,$F1,$B8,$7F,$C2
	.dc.b	$67,$3A,$F0,$C6,$63,$FC,$48,$DB,$F8,$55,$E4,$24,$F8,$AF,$CD,$7C
	.dc.b	$9B,$3E,$49,$F0,$C6,$1F,$21,$E1,$8D,$34,$CD,$F8,$63,$E5,$D7,$F5
	.dc.b	$88,$C4,$A3,$15,$D6,$3D,$B3,$A9,$CD,$7A,$1F,$33,$4F,$0C,$6D,$F9
	.dc.b	$55,$CE,$BE,$1B,$6F,$E1,$C7,$C1,$9E,$4F,$A2,$75,$1E,$83,$99,$A8
	.dc.b	$F2,$C6,$9F,$54,$69,$E8,$4B,$F3,$B6,$3C,$F3,$C6,$7D,$B3,$93,$C5
	.dc.b	$3D,$DF,$4B,$6B,$D3,$3D,$09,$1D,$0D,$1E,$99,$FB,$2E,$1F,$39,$83
	.dc.b	$C0,$7B,$1D,$47,$90,$EB,$1E,$43,$67,$90,$B3,$E4,$9F,$14,$61,$F1
	.dc.b	$3F,$85,$3B,$19,$3C,$66,$4E,$93,$E0,$76,$D7,$31,$DA,$31,$3E,$19
	.dc.b	$F3,$46,$27,$C3,$7F,$85,$9F,$81,$39,$7C,$31,$97,$C5,$5C,$DF,$0D
	.dc.b	$77,$23,$D9,$AE,$8C,$10,$C1,$0F,$49,$92,$99,$8F,$03,$82,$38,$24
	.dc.b	$7B,$2F,$F1,$4F,$E4,$6F,$D8,$7D,$97,$CB,$5C,$9E,$07,$DE,$38,$1A
	.dc.b	$AE,$B1,$F1,$2B,$AB,$F2,$27,$47,$27,$ED,$57,$77,$D9,$5C,$A3,$11
	.dc.b	$ED,$9E,$11,$E6,$7C,$31,$C1,$C9,$D9,$F1,$3E,$77,$C4,$6C,$C9,$F8
	.dc.b	$B3,$FA,$0B,$FE,$AE,$9F,$F3,$24,$7F,$B9,$0C,$9F,$E0,$17,$2F,$FB
	.dc.b	$F4,$C9,$A9,$FE,$7A,$FF,$AE,$E3,$FD,$96,$D9,$FE,$96,$D3,$F9,$F9
	.dc.b	$67,$8F,$47,$0A,$E2,$DA,$E2,$DA,$E2,$DA,$E3,$3D,$AB,$8B,$4E,$25
	.dc.b	$AE,$2D,$9E,$31,$E4,$9E,$2D,$9E,$2D,$38,$96,$78,$9F,$0A,$78,$B6
	.dc.b	$78,$B6,$78,$B4,$E3,$1F,$D1,$A3,$FD,$CB,$3F,$EC,$C3,$DD,$1F,$E6
	.dc.b	$F6,$C7,$F9,$7D,$AF,$FB,$F5,$F0,$C7,$36,$C7,$F5,$43,$AC,$7F,$B0
	.dc.b	$A3,$8C,$7D,$38,$F1,$56,$27,$C5,$1F,$1C,$F1,$16,$7C,$4D,$7C,$4F
	.dc.b	$C0,$3C,$45,$7C,$51,$E4,$7F,$BB,$9E,$43,$FD,$7A,$F9,$27,$C5,$3E
	.dc.b	$73,$B4,$F9,$CF,$64,$69,$E3,$1A,$8C,$57,$F9,$B2,$BA,$4F,$27,$DB
	.dc.b	$5F,$6E,$3D,$AD,$3D,$A5,$3D,$A7,$DA,$7D,$AD,$3D,$AF,$BD,$F6,$C7
	.dc.b	$BD,$F7,$4F,$E0,$1D,$9C,$4D,$CF,$17,$6F,$D4,$3D,$07,$C3,$7C,$87
	.dc.b	$C3,$3E,$5D,$7C,$08,$F5,$CF,$E3,$DA,$7A,$1F,$88,$FC,$73,$AC,$FC
	.dc.b	$A3,$AC,$F3,$7E,$1C,$F3,$6C,$F3,$3A,$BE,$C8,$EA,$F2,$8E,$B1,$83
	.dc.b	$94,$7C,$C7,$E2,$3F,$08,$F8,$87,$C2,$3E,$FC,$72,$3B,$BB,$8E,$91
	.dc.b	$FE,$56,$8F,$F2,$BF,$47,$F1,$A4,$D2,$47,$FA,$7E,$32,$FF,$74,$69
	.dc.b	$FD,$C8,$73,$FF,$AA,$9C,$3F,$EA,$A7,$FD,$54,$D7,$FD,$38,$7F,$A9
	.dc.b	$4F,$23,$FE,$9B,$E8,$CF,$47,$FA,$8E,$31,$3E,$93,$FD,$3C,$3B,$75
	.dc.b	$7F,$E7,$27,$FD,$47,$D3,$70,$E1,$6A,$C5,$F3,$AF,$F5,$3B,$E4,$3F
	.dc.b	$D5,$07,$FA,$8F,$A7,$FD,$4E,$FF,$A8,$DB,$CF,$FE,$A6,$27,$47,$FA
	.dc.b	$8E,$7F,$D9,$11,$FE,$A8,$7E,$83,$5F,$F4,$3A,$F0,$FE,$A6,$DE,$59
	.dc.b	$CF,$C3,$E3,$2E,$2D,$8E,$35,$5F,$F5,$B4,$5E,$37,$93,$4D,$62,$9B
	.dc.b	$D7,$8F,$02,$61,$FE,$30,$AF,$F5,$CB,$63,$8F,$64,$8C,$1B,$7C,$1E
	.dc.b	$04,$8C,$1F,$E3,$AE,$BF,$F5,$97,$5F,$FA,$D6,$3F,$25,$0E,$BF,$F5
	.dc.b	$9C,$7F,$AE,$3A,$FF,$D6,$EF,$FB,$C7,$67,$A1,$FD,$C1,$BC,$3F,$EF
	.dc.b	$36,$CE,$E7,$FB,$AF,$17,$FB,$4A,$BF,$C9,$6F,$C1,$6E,$FF,$3E,$3E
	.dc.b	$85,$7F,$A3,$78,$31,$5F,$52,$AC,$FF,$A5,$26,$BF,$E9,$38,$B5,$D6
	.dc.b	$3F,$CA,$AF,$85,$B3,$A9,$DB,$6B,$CA,$D9,$DC,$79,$8C,$95,$C9,$DA
	.dc.b	$79,$C6,$6B,$D6,$78,$CB,$1A,$9E,$13,$D1,$A7,$49,$DC,$65,$B1,$98
	.dc.b	$F8,$86,$4A,$E4,$FF,$5C,$CF,$79,$E2,$79,$2B,$DC,$DA,$F4,$1F,$A2
	.dc.b	$8C,$B4,$CD,$7D,$16,$BF,$5D,$B3,$EF,$7C,$6F,$58,$F6,$B4,$F1,$C7
	.dc.b	$58,$CB,$63,$31,$DC,$C9,$5C,$BE,$13,$25,$72,$78,$DF,$87,$1F,$1D
	.dc.b	$A7,$C7,$3F,$14,$E5,$A6,$63,$F1,$47,$A6,$3D,$A7,$9C,$B1,$E5,$6B
	.dc.b	$E3,$6C,$78,$E3,$AB,$CD,$F1,$B5,$CC,$75,$72,$D7,$27,$73,$25,$32
	.dc.b	$78,$4C,$94,$CC,$78,$CF,$6C,$75,$6C,$73,$6B,$C5,$A7,$1A,$FC,$74
	.dc.b	$FB,$AB,$F1,$35,$C6,$72,$F8,$AB,$85,$7C,$F7,$9D,$70,$9E,$D5,$C1
	.dc.b	$F3,$C6,$67,$8C,$7D,$59,$D4,$72,$AF,$C9,$D6,$C8,$E0,$95,$F8,$89
	.dc.b	$E6,$7D,$13,$93,$99,$E2,$D9,$E3,$3F,$8E,$9C,$B6,$32,$7C,$32,$CF
	.dc.b	$C1,$9D,$C6,$63,$A1,$27,$93,$63,$51,$82,$38,$24,$75,$7B,$BF,$21
	.dc.b	$B1,$EC,$38,$C6,$CE,$C5,$72,$F1,$32,$57,$27,$38,$F0,$C6,$5B,$19
	.dc.b	$8E,$91,$8A,$FA,$E4,$7F,$04,$D8,$FB,$91,$82,$38,$24,$FD,$79,$E0
	.dc.b	$D9,$DB,$67,$71,$F9,$D9,$DB,$4D,$96,$36,$7E,$65,$CB,$4C,$9F,$96
	.dc.b	$72,$D3,$33,$F8,$C8,$E9,$1F,$E2,$98,$FE,$A8,$D9,$FE,$CE,$49,$FF
	.dc.b	$17,$47,$F1,$63,$4E,$1F,$E2,$A6,$0F,$E4,$38,$3C,$BF,$D3,$57,$EA
	.dc.b	$E3,$15,$FC,$DA,$31,$3A,$8F,$E7,$73,$AA,$FD,$25,$6A,$78,$1A,$AE
	.dc.b	$13,$E0,$6D,$FF,$34,$BE,$4F,$F3,$2B,$E5,$5D,$2F,$91,$D1,$E4,$D9
	.dc.b	$E5,$7D,$23,$F9,$5D,$F4,$7F,$95,$DF,$4A,$ED,$1D,$2F,$B1,$D0,$B1
	.dc.b	$D2,$3B,$1D,$2F,$B4,$FF,$26,$BE,$CF,$F1,$7B,$ED,$5E,$03,$B5,$F8
	.dc.b	$27,$B3,$4E,$CF,$82,$3B,$1E,$07,$B5,$F8,$1F,$E6,$F7,$E0,$7F,$9A
	.dc.b	$DF,$81,$F1,$57,$80,$F1,$3E,$06,$DF,$80,$F1,$3E,$08,$F1,$1E,$07
	.dc.b	$C4,$F8,$2F,$C5,$3F,$CF,$AF,$C4,$FF,$3E,$BF,$15,$79,$2F,$C4,$F9
	.dc.b	$0F,$11,$67,$C5,$1E,$43,$C4,$F9,$23,$C4,$79,$0F,$14,$79,$23,$C5
	.dc.b	$7E,$4A,$FF,$1F,$DF,$91,$F9,$0D,$AF,$5B,$6F,$A3,$6F,$C1,$08,$C4
	.dc.b	$F9,$60,$66,$FC,$11,$8A,$FF,$2A,$D7,$0A,$F7,$4F,$FA,$EE,$31,$1F
	.dc.b	$14,$4F,$0A,$E7,$18,$AE,$11,$89,$D4,$7C,$1B,$F7,$CE,$63,$66,$4F
	.dc.b	$21,$8A,$F9,$55,$F2,$23,$51,$DC,$F6,$DF,$9A,$7E,$E4,$F2,$9C,$D7
	.dc.b	$76,$C7,$29,$F2,$C7,$6B,$F2,$CE,$4F,$14,$FC,$E3,$C5,$3B,$8E,$CF
	.dc.b	$CD,$78,$3D,$8D,$1E,$73,$46,$23,$C2,$75,$9F,$0C,$66,$7C,$27,$DB
	.dc.b	$7F,$16,$64,$ED,$3C,$63,$C7,$7F,$0E,$BF,$3B,$3F,$12,$BF,$36,$7D
	.dc.b	$B7,$F3,$6F,$E0,$0F,$CC,$BF,$E5,$2F,$CD,$9E,$3F,$FA,$51,$B8,$7C
	.dc.b	$05,$A6,$67,$4F,$FA,$82,$FF,$22,$27,$75,$FC,$FD,$B1,$FD,$11,$AF
	.dc.b	$F4,$6E,$FF,$F5,$1A,$6F,$7E,$1B,$5F,$4A,$53,$C3,$C3,$CC,$7D,$10
	.dc.b	$E6,$7E,$3F,$8B,$67,$5E,$0F,$80,$F0,$F7,$5A,$EC,$E1,$F0,$4B,$BB
	.dc.b	$F5,$FE,$0B,$C0,$F8,$32,$BF,$5B,$B3,$FB,$01,$4F,$EC,$55,$87,$07
	.dc.b	$FB,$FB,$C0,$78,$9F,$D9,$FB,$91,$89,$FE,$DB,$18,$9D,$4E,$67,$C7
	.dc.b	$2F,$07,$F7,$71,$C1,$AE,$9A,$FD,$A8,$E1,$C9,$E1,$F0,$7F,$E0,$0F
	.dc.b	$4C,$AF,$F7,$F4,$57,$F4,$1D,$0F,$07,$F0,$77,$FC,$41,$D3,$FE,$E6
	.dc.b	$C1,$D3,$AE,$1F,$07,$83,$71,$FE,$37,$48,$C4,$23,$8B,$FB,$47,$4D
	.dc.b	$F0,$77,$06,$CE,$1E,$79,$3C,$1A,$66,$FF,$5B,$F1,$7F,$62,$4D,$24
	.dc.b	$7F,$86,$FC,$1F,$D2,$F8,$57,$B8,$E7,$C2,$FC,$25,$7F,$89,$2B,$6F
	.dc.b	$F8,$9B,$7F,$F5,$66,$2F,$37,$FD,$10,$DD,$FF,$D7,$8D,$AF,$F6,$03
	.dc.b	$6B,$FD,$60,$D9,$FF,$57,$36,$7F,$D5,$8D,$7F,$D5,$4D,$9F,$0D,$78
	.dc.b	$09,$5E,$06,$D7,$81,$B3,$E0,$6C,$F8,$1B,$3E,$06,$BE,$06,$CF,$82
	.dc.b	$7B,$12,$7B,$36,$3B,$34,$EC,$58,$EC,$D7,$B3,$63,$B4,$74,$23,$D1
	.dc.b	$B1,$D1,$A7,$42,$9D,$0B,$1D,$0E,$44,$8E,$4D,$8E,$4D,$79,$34,$E5
	.dc.b	$1C,$09,$1C,$1A,$F0,$6C,$70,$8D,$10,$D1,$5D,$38,$23,$82,$71,$FF
	.dc.b	$85,$AB,$EA,$B6,$BE,$53,$6B,$E0,$B6,$BD,$4D,$9F,$2B,$67,$C4,$D7
	.dc.b	$C2,$D9,$F0,$CF,$F5,$CA,$F0,$B4,$F0,$96,$BC,$2D,$9F,$0B,$4F,$09
	.dc.b	$63,$C2,$D7,$C2,$D8,$F0,$C7,$F1,$A9,$F0,$B6,$7C,$2D,$3C,$25,$9F
	.dc.b	$0B,$63,$C2,$D3,$C2,$53,$C2,$58,$F0,$9F,$E3,$69,$F0,$B6,$7C,$2D
	.dc.b	$9F,$0B,$4F,$09,$63,$C2,$D8,$F0,$B5,$F0,$B4,$F0,$C7,$F9,$0E,$7C
	.dc.b	$2D,$3C,$25,$9F,$0B,$67,$C2,$D3,$C2,$58,$F0,$B5,$F0,$B6,$3C,$31
	.dc.b	$FE,$3E,$94,$FF,$8C,$25,$5F,$E0,$FA,$FF,$2E,$35,$FF,$2B,$B6,$37
	.dc.b	$3F,$EC,$F6,$CF,$F2,$F9,$D1,$27,$4F,$FB,$B2,$3A,$C7,$86,$31,$1D
	.dc.b	$A3,$07,$FB,$A2,$7A,$9E,$88,$E2,$7F,$BA,$A7,$11,$98,$E4,$66,$79
	.dc.b	$B4,$FF,$25,$47,$48,$FF,$7E,$3C,$4C,$47,$67,$13,$DA,$7F,$B8,$F0
	.dc.b	$DA,$D4,$7F,$74,$9F,$5D,$72,$9F,$1B,$CA,$7E,$85,$71,$9C,$C7,$13
	.dc.b	$E0,$56,$6B,$91,$2B,$11,$27,$67,$BA,$7B,$57,$BA,$B4,$FC,$0A,$F2
	.dc.b	$98,$9D,$9F,$06,$71,$5E,$3A,$F9,$13,$A8,$F8,$E6,$5C,$47,$C2,$3F
	.dc.b	$D4,$93,$C9,$C1,$EE,$33,$3E,$73,$91,$D1,$DC,$FC,$0A,$F8,$55,$CE
	.dc.b	$79,$12,$7F,$C3,$F5,$CE,$38,$D7,$08,$C4,$6E,$30,$FC,$38,$CC,$FC
	.dc.b	$D6,$CF,$C6,$69,$F6,$0B,$3F,$11,$F4,$BE,$8A,$F3,$57,$A6,$78,$4F
	.dc.b	$72,$47,$C8,$7F,$D8,$73,$C2,$BC,$11,$AA,$F8,$8F,$A5,$FB,$EE,$4F
	.dc.b	$03,$E4,$AE,$87,$C4,$8D,$1C,$4F,$C2,$46,$CE,$B1,$E7,$8F,$61,$F2
	.dc.b	$A3,$D3,$1D,$27,$CA,$D8,$C9,$F0,$A7,$B4,$7E,$2E,$3B,$47,$98,$F1
	.dc.b	$C7,$73,$4E,$27,$AC,$F3,$9F,$A9,$39,$8F,$83,$1E,$18,$F5,$99,$3D
	.dc.b	$27,$8A,$3F,$6E,$78,$9F,$C4,$57,$CF,$8E,$67,$7A,$FC,$2D,$75,$3E
	.dc.b	$5C,$EE,$BA,$C7,$A6,$BA,$C7,$86,$37,$1D,$DF,$49,$CD,$F9,$F5,$CE
	.dc.b	$3E,$09,$99,$F8,$CD,$7C,$F3,$CE,$7C,$AF,$9C,$F4,$C7,$D7,$3E,$1C
	.dc.b	$F1,$3C,$31,$B8,$F4,$CF,$CE,$78,$4F,$8A,$74,$FA,$A3,$D9,$1E,$83
	.dc.b	$D9,$5E,$97,$11,$A3,$F7,$A5,$3D,$27,$E4,$BC,$0F,$F7,$0B,$FE,$D7
	.dc.b	$F0,$7F,$A7,$46,$23,$FD,$7E,$D7,$FD,$C2,$F0,$BF,$E7,$CD,$AC,$ED
	.dc.b	$FE,$E9,$7F,$DB,$5E,$0F,$FB,$75,$FF,$6E,$78,$37,$0F,$F6,$D3,$FE
	.dc.b	$DC,$C3,$FD,$DD,$E0,$DA,$FF,$67,$BF,$ED,$B8,$FE,$99,$B3,$C1,$C5
	.dc.b	$7F,$B5,$3C,$1F,$F6,$E6,$3F,$EE,$5E,$FB,$87,$FB,$E5,$FF,$6F,$47
	.dc.b	$FB,$A4,$FF,$6C,$78,$3F,$ED,$D7,$FD,$FB,$1F,$ED,$CF,$03,$FD,$B0
	.dc.b	$48,$FF,$6E,$BF,$ED,$2F,$02,$47,$FB,$5B,$C0,$FF,$6A,$3F,$ED,$D8
	.dc.b	$FF,$6E,$78,$1F,$ED,$08,$FF,$5F,$B5,$FF,$5F,$37,$BB,$FD,$A7,$1F
	.dc.b	$EE,$0F,$03,$FD,$97,$AF,$FB,$5F,$9F,$FD,$7F,$18,$90,$E0,$FF,$6A
	.dc.b	$38,$DB,$FD,$A8,$FF,$B2,$6B,$FD,$8F,$CF,$FE,$C6,$8F,$F6,$67,$7E
	.dc.b	$1A,$7F,$B3,$48,$2B,$FD,$B9,$1F,$EC,$98,$FF,$71,$F7,$7F,$B3,$1D
	.dc.b	$CF,$F8,$07,$C1,$72,$FF,$B7,$17,$25,$31,$16,$BD,$1A,$FB,$1F,$67
	.dc.b	$9B,$FE,$DA,$34,$F8,$E3,$5D,$BF,$EC,$E8,$E0,$1C,$3A,$EE,$FF,$6A
	.dc.b	$BA,$8C,$BF,$2B,$E7,$9E,$87,$F8,$BD,$B7,$D2,$75,$39,$35,$94,$9D
	.dc.b	$4A,$BD,$F1,$86,$C7,$36,$E9,$FE,$27,$6C,$FC,$15,$D1,$FE,$15,$73
	.dc.b	$5E,$7B,$C1,$99,$B5,$F2,$A7,$4D,$7C,$93,$B8,$C5,$74,$9F,$45,$72
	.dc.b	$7E,$71,$F1,$5C,$5F,$9D,$B5,$E7,$6C,$EE,$3A,$DF,$78,$E6,$78,$A3
	.dc.b	$C3,$3C,$23,$C3,$1C,$0C,$95,$E3,$7F,$12,$31,$7E,$36,$9E,$32,$CF
	.dc.b	$8C,$E3,$7E,$39,$FA,$0D,$9F,$A0,$D9,$F9,$27,$82,$FC,$D7,$E1,$8E
	.dc.b	$75,$E6,$69,$B7,$C2,$FB,$0C,$DF,$9D,$B1,$E7,$69,$E7,$8F,$2C,$75
	.dc.b	$6C,$75,$36,$75,$72,$75,$2C,$75,$6C,$75,$8F,$79,$D4,$AF,$58,$DB
	.dc.b	$D6,$33,$1D,$5A,$75,$2C,$75,$3E,$4C,$6D,$B1,$B6,$9B,$3F,$08,$78
	.dc.b	$5E,$6F,$84,$D9,$4D,$C7,$56,$C7,$56,$9D,$67,$F3,$C7,$D9,$9F,$6B
	.dc.b	$67,$BB,$67,$BB,$5E,$67,$85,$E7,$5F,$84,$24,$FE,$02,$38,$1A,$71
	.dc.b	$1A,$9F,$B3,$3A,$BF,$C8,$47,$68,$C5,$7D,$58,$F1,$D7,$1B,$F4,$B6
	.dc.b	$7D,$2D,$9D,$C7,$80,$8F,$82,$34,$F7,$6D,$79,$E3,$53,$C2,$BB,$B4
	.dc.b	$EE,$58,$EE,$70,$24,$70,$6C,$70,$69,$C2,$34,$48,$D3,$63,$51,$82
	.dc.b	$38,$21,$EB,$7C,$8D,$3B,$96,$3A,$36,$39,$B4,$E7,$3F,$0E,$76,$D9
	.dc.b	$DB,$67,$77,$F6,$A3,$11,$CC,$8E,$09,$1B,$24,$FD,$F9,$DC,$61,$FD
	.dc.b	$21,$DA,$35,$3C,$9B,$3B,$AF,$D2,$4E,$6B,$89,$D2,$B6,$47,$31,$B9
	.dc.b	$E3,$5C,$EB,$F5,$93,$F8,$5A,$FE,$DE,$FF,$23,$6D,$FF,$69,$24,$FF
	.dc.b	$61,$6C,$7F,$61,$6D,$FE,$E7,$89,$F1,$9F,$26,$7F,$47,$18,$BF,$7C
	.dc.b	$62,$75,$1F,$1E,$75,$7E,$AA,$D4,$F0,$35,$5C,$27,$E1,$D7,$0B,$F8
	.dc.b	$F7,$C2,$79,$5F,$03,$93,$C2,$F9,$57,$A5,$B3,$FC,$46,$FA,$3F,$22
	.dc.b	$FA,$57,$68,$E9,$7D,$8E,$8F,$68,$E8,$D3,$A5,$F6,$9F,$E5,$17,$D9
	.dc.b	$FE,$3D,$7D,$AB,$C0,$76,$BF,$04,$F6,$3C,$0F,$66,$C7,$63,$C0,$F6
	.dc.b	$BF,$03,$FC,$C6,$FC,$0F,$F2,$3B,$F0,$3E,$2A,$F0,$1E,$27,$C1,$7E
	.dc.b	$23,$C0,$58,$F0,$47,$88,$F0,$3E,$27,$C1,$7E,$2C,$3F,$9A,$61,$E2
	.dc.b	$7F,$97,$E1,$E2,$AF,$25,$F8,$9F,$21,$E2,$7C,$98,$78,$9A,$78,$9F
	.dc.b	$24,$78,$8F,$21,$E2,$8F,$25,$78,$B0,$F2,$5A,$7F,$C6,$18,$79,$63
	.dc.b	$E4,$B6,$3D,$8D,$BE,$AD,$C3,$C3,$1F,$E7,$DB,$EB,$5D,$1B,$1F,$E2
	.dc.b	$CA,$E1,$3E,$A8,$FF,$42,$C6,$1F,$83,$5F,$EA,$AA,$E2,$DA,$F5,$DF
	.dc.b	$A6,$BA,$56,$A7,$FC,$E9,$7C,$E7,$E6,$9A,$8E,$27,$94,$AE,$4F,$93
	.dc.b	$87,$C7,$8C,$E1,$E1,$9C,$E1,$E1,$8C,$1F,$0A,$B8,$4E,$AB,$B3,$4E
	.dc.b	$D3,$BA,$E5,$3E,$8C,$3D,$55,$F1,$1A,$FC,$D8,$F1,$9B,$2B,$B6,$BB
	.dc.b	$3C,$15,$E6,$8F,$A1,$1F,$85,$32,$F3,$7F,$20,$7E,$2E,$3F,$20,$D8
	.dc.b	$FC,$81,$1F,$C8,$1F,$E9,$CC,$3F,$8E,$B6,$BE,$B3,$67,$FD,$23,$DF
	.dc.b	$FE,$E5,$7F,$D0,$0D,$8F,$F3,$8A,$F2,$BE,$2C,$7F,$B2,$A3,$58,$7F
	.dc.b	$BF,$27,$AC,$FF,$C0,$06,$BB,$6A,$9E,$95,$FE,$A2,$A7,$2F,$6A,$FF
	.dc.b	$65,$36,$1C,$B6,$30,$CA,$FD,$4F,$8F,$6F,$FD,$12,$3C,$0E,$8D,$9F
	.dc.b	$F4,$F2,$F0,$33,$3A,$64,$EB,$96,$D7,$47,$99,$E7,$DB,$E1,$CF,$FB
	.dc.b	$32,$3B,$F0,$C8,$EF,$C1,$6F,$CA,$F9,$0F,$2A,$EE,$FF,$7B,$F7,$BA
	.dc.b	$3F,$DA,$F5,$A9,$FF,$64,$D7,$FB,$F1,$AF,$13,$2F,$FB,$3B,$B1,$DE
	.dc.b	$EF,$C3,$6B,$FD,$B1,$3A,$9C,$9C,$7B,$FF,$DF,$93,$A6,$BF,$EE,$A9
	.dc.b	$46,$23,$FD,$E9,$18,$E4,$F1,$77,$FF,$BD,$DF,$C9,$C7,$E0,$27,$FD
	.dc.b	$A0,$D7,$FD,$82,$A7,$86,$BD,$39,$6D,$7F,$18,$F1,$EA,$E2,$46,$30
	.dc.b	$E9,$8F,$83,$9B,$1E,$04,$C3,$C3,$15,$F0,$F8,$FC,$4C,$57,$A1,$EF
	.dc.b	$8B,$C1,$A6,$EB,$FD,$D5,$C6,$E6,$F2,$AE,$6A,$62,$2C,$FF,$42,$F0
	.dc.b	$7A,$98,$7F,$DC,$7E,$8B,$D1,$FC,$0B,$60,$42,$39,$70,$69,$C3,$8F
	.dc.b	$09,$E2,$BC,$8D,$CF,$49,$CC,$62,$10,$F1,$5D,$99,$8D,$CF,$2D,$FD
	.dc.b	$1C,$FC,$5B,$19,$CF,$F8,$2C,$62,$3F,$8D,$47,$EB,$4F,$F8,$01,$1C
	.dc.b	$38,$BF,$8D,$B6,$BF,$A7,$CE,$88,$FF,$BB,$23,$55,$FE,$C4,$6B,$FE
	.dc.b	$B6,$6F,$89,$F5,$67,$FB,$9D,$72,$AF,$F7,$4D,$6B,$83,$C9,$5D,$6B
	.dc.b	$58,$7F,$B8,$AB,$53,$C8,$D1,$2F,$9D,$70,$8F,$F6,$CC,$F0,$9F,$EE
	.dc.b	$D3,$97,$D0,$D7,$D0,$DC,$3D,$11,$C2,$BB,$5F,$8E,$B5,$87,$A2,$BC
	.dc.b	$35,$A3,$C3,$3A,$7D,$11,$A2,$57,$38,$E1,$1E,$46,$BE,$46,$DF,$92
	.dc.b	$7E,$BD,$75,$AF,$31,$D2,$BC,$27,$0B,$F3,$57,$79,$D1,$DC,$A7,$72
	.dc.b	$DF,$7A,$FC,$19,$CC,$F5,$CF,$32,$3E,$09,$E3,$5E,$08,$C4,$EA,$7C
	.dc.b	$0D,$7C,$0D,$AF,$04,$7F,$8A,$AF,$DD,$3F,$02,$BC,$53,$F0,$23,$85
	.dc.b	$F7,$8E,$B3,$E0,$23,$82,$1E,$07,$2D,$33,$7E,$27,$F9,$15,$F8,$A7
	.dc.b	$E1,$C7,$B9,$F8,$67,$29,$E1,$87,$E4,$70,$F4,$56,$CF,$45,$6F,$0F
	.dc.b	$44,$78,$49,$7C,$B0,$F0,$47,$C3,$AF,$83,$7D,$0F,$82,$F8,$A3,$D3
	.dc.b	$7D,$2F,$94,$F0,$3D,$35,$C2,$BC,$67,$08,$E4,$F8,$AB,$91,$23,$35
	.dc.b	$D2,$7D,$95,$E3,$9F,$61,$AA,$E1,$3D,$8E,$91,$92,$39,$AE,$95,$E5
	.dc.b	$3A,$47,$94,$D5,$74,$8C,$C7,$43,$35,$D2,$33,$5D,$23,$CB,$3A,$21
	.dc.b	$FE,$EC,$AF,$F1,$6C,$FF,$02,$8C,$4F,$F3,$78,$C4,$7F,$45,$6B,$FD
	.dc.b	$21,$A9,$A7,$E2,$CF,$F7,$F4,$71,$9F,$F1,$B4,$62,$79,$46,$0C,$E5
	.dc.b	$6B,$FA,$AD,$FF,$94,$AF,$A5,$6A,$72,$75,$2D,$F4,$6B,$E5,$8E,$0A
	.dc.b	$BA,$47,$78,$C4,$F3,$69,$EF,$BD,$57,$09,$EB,$7E,$BA,$ED,$18,$BE
	.dc.b	$F5,$B9,$E6,$68,$B5,$C4,$F6,$D7,$4B,$EE,$F8,$EB,$91,$F0,$E3,$35
	.dc.b	$EF,$39,$DF,$82,$B8,$12,$35,$5C,$63,$CC,$48,$E2,$F8,$4C,$47,$81
	.dc.b	$E5,$1F,$7E,$BC,$2F,$AE,$74,$FC,$79,$E2,$70,$BF,$75,$F0,$7F,$C0
	.dc.b	$31,$A9,$F2,$BB,$BD,$56,$DB,$1B,$9F,$13,$6B,$E4,$C7,$CC,$9D,$BE
	.dc.b	$C3,$13,$BA,$E8,$D3,$94,$F0,$BF,$5D,$76,$39,$38,$BF,$5D,$70,$39
	.dc.b	$4E,$9F,$D4,$57,$37,$D8,$6A,$B9,$46,$67,$E0,$9F,$24,$E4,$60,$F9
	.dc.b	$D3,$D0,$E4,$74,$76,$68,$E1,$5F,$39,$CB,$F1,$AB,$F0,$51,$8B,$F8
	.dc.b	$6D,$AF,$63,$4E,$37,$E7,$7B,$1F,$16,$BE,$25,$72,$6D,$72,$6C,$F2
	.dc.b	$8F,$4B,$E6,$AE,$97,$F3,$23,$47,$86,$BE,$3D,$70,$8F,$0C,$FB,$67
	.dc.b	$A9,$A8,$E1,$3F,$16,$32,$D3,$F2,$97,$E7,$69,$E7,$2C,$F8,$A7,$8B
	.dc.b	$E7,$8F,$15,$79,$E7,$C0,$6A,$7C,$B7,$EC,$BF,$DA,$47,$D4,$7C,$0D
	.dc.b	$BF,$03,$4F,$03,$E2,$AE,$AD,$3D,$51,$B8,$CB,$63,$2F,$C6,$9E,$ED
	.dc.b	$8E,$ED,$3B,$9C,$4F,$71,$F2,$8B,$3E,$DB,$FE,$31,$7D,$1B,$5F,$88
	.dc.b	$AF,$F4,$4C,$7E,$45,$B3,$FD,$40,$FF,$7E,$F8,$4F,$1A,$92,$FF,$D1
	.dc.b	$11,$96,$C7,$FB,$FE,$3F,$DD,$35,$FE,$9A,$5E,$26,$67,$FE,$00,$57
	.dc.b	$F4,$2C,$5D,$7F,$EF,$F5,$1F,$EC,$D8,$C3,$FE,$CF,$E8,$FF,$7B,$78
	.dc.b	$98,$CB,$FD,$4D,$E2,$7F,$C0,$0A,$FE,$9B,$0F,$13,$0F,$F3,$E3,$11
	.dc.b	$A8,$FF,$80,$2D,$38,$F4,$62,$7F,$A2,$F1,$7B,$1F,$F5,$AC,$7F,$C0
	.dc.b	$1C,$FC,$F1,$FE,$B7,$E1,$B1,$88,$FF,$5D,$F0,$6F,$0F,$F8,$02,$79
	.dc.b	$9D,$BF,$EB,$6F,$0F,$FD,$83,$E2,$3A,$F7,$7E,$91,$3F,$EB,$97,$FD
	.dc.b	$8E,$D7,$26,$5F,$F8,$02,$FF,$B0,$BB,$DE,$23,$C4,$CC,$FF,$C0,$18
	.dc.b	$C3,$BF,$13,$FB,$A3,$74,$FF,$64,$5A,$73,$3F,$ED,$48,$C4,$EB,$C4
	.dc.b	$FF,$7C,$72,$CA,$FF,$00,$D2,$71,$1F,$3B,$C4,$FF,$79,$98,$BF,$F7
	.dc.b	$F4,$70,$F1,$3F,$DF,$3B,$FF,$B9,$70,$73,$E2,$BF,$DD,$30,$F1,$3F
	.dc.b	$DF,$51,$FD,$FA,$17,$8C,$3F,$DA,$3E,$27,$FC,$00,$AF,$2D,$FF,$BE
	.dc.b	$DF,$F8,$03,$C9,$FE,$F9,$AF,$B9,$C4,$E5,$C6,$9F,$EF,$7E,$67,$89
	.dc.b	$FF,$00,$31,$E2,$7F,$88,$3D,$1F,$CF,$BF,$EF,$16,$BF,$EF,$66,$BF
	.dc.b	$EF,$E6,$A2,$7E,$74,$F8,$4F,$C1,$36,$3F,$DF,$8D,$7F,$E0,$0B,$5E
	.dc.b	$47,$3F,$0F,$FD,$FA,$E0,$38,$76,$BC,$47,$66,$E7,$97,$8C,$F1,$B1
	.dc.b	$0F,$97,$F3,$6D,$86,$C7,$16,$BE,$16,$BF,$E5,$DF,$87,$FF,$31,$92
	.dc.b	$7F,$CC,$4F,$F9,$9B,$DD,$91,$FE,$65,$6C,$F9,$7C,$87,$C4,$7F,$9B
	.dc.b	$09,$5F,$E6,$B7,$F6,$D3,$F1,$36,$7C,$46,$23,$E7,$B5,$46,$AB,$E7
	.dc.b	$CE,$61,$8D,$24,$FD,$47,$5F,$25,$23,$EB,$B6,$FE,$BE,$5C,$C9,$3F
	.dc.b	$71,$ED,$8C,$8F,$BE,$D9,$FB,$FF,$15,$E3,$24,$FE,$11,$F3,$7C,$9C
	.dc.b	$8F,$C3,$B6,$7F,$0F,$F2,$9E,$B2,$4F,$E3,$1F,$77,$A5,$23,$F1,$ED
	.dc.b	$9F,$C7,$FC,$6F,$C3,$24,$FE,$51,$F8,$DF,$1F,$23,$F2,$ED,$9F,$CB
	.dc.b	$FE,$03,$F2,$C9,$3F,$9C,$7E,$77,$6C,$8F,$CF,$B6,$7F,$3F,$E7,$7D
	.dc.b	$32,$4F,$E9,$1F,$AD,$E7,$48,$FD,$3B,$67,$F4,$FD,$FF,$6C,$93,$FA
	.dc.b	$C7,$EF,$78,$32,$3F,$5E,$D9,$FD,$7F,$57,$E0,$C9,$3F,$B4,$7F,$D7
	.dc.b	$BB,$FF,$40,$9F,$DD,$9F,$EB,$27,$0F,$E3,$8C,$4F,$FB,$06,$38,$B4
	.dc.b	$91,$FE,$98,$8F,$F5,$DE,$FC,$23,$13,$FE,$A4,$9C,$B4,$91,$A3,$FD
	.dc.b	$43,$3C,$7C,$1F,$F5,$F5,$5A,$FF,$68,$4D,$AF,$F6,$84,$D7,$FD,$A1
	.dc.b	$17,$A1,$36,$30,$DA,$F0,$DF,$FB,$2E,$71,$7E,$38,$C4,$E8,$F0,$96
	.dc.b	$3F,$DC,$13,$C2,$7C,$F3,$FE,$C2,$8C,$3F,$EE,$7A,$F8,$35,$D6,$39
	.dc.b	$B4,$E6,$F5,$7D,$6D,$B5,$7F,$BD,$89,$1F,$EF,$56,$CF,$CA,$8F,$F6
	.dc.b	$9F,$0F,$83,$0F,$5C,$E4,$FF,$67,$C6,$67,$C7,$3A,$3F,$DD,$D1,$99
	.dc.b	$F2,$9C,$63,$33,$EF,$24,$65,$FE,$7F,$6A,$E5,$3E,$C3,$47,$F4,$D8
	.dc.b	$D4,$F2,$AE,$B3,$A1,$3A,$70,$D8,$F4,$4F,$82,$BB,$CE,$63,$DA,$64
	.dc.b	$EC,$5A,$E7,$3C,$09,$1A,$8F,$34,$61,$DC,$65,$F2,$C6,$6B,$C7,$1C
	.dc.b	$27,$55,$E3,$8F,$97,$3E,$E3,$6F,$29,$DD,$73,$7D,$31,$CC,$F0,$92
	.dc.b	$78,$D7,$58,$F9,$D1,$D6,$78,$C7,$53,$83,$89,$E0,$7E,$39,$FC,$49
	.dc.b	$E2,$24,$70,$3D,$2F,$C9,$9F,$0D,$79,$E7,$27,$9E,$33,$3C,$9B,$1C
	.dc.b	$CF,$3C,$73,$3C,$93,$CE,$32,$73,$8C,$9C,$CA,$F3,$AF,$53,$F9,$E8
	.dc.b	$D3,$E4,$3E,$15,$78,$89,$1C,$9B,$1B,$9F,$8B,$7F,$C9,$9B,$5F,$1A
	.dc.b	$7C,$13,$9B,$FE,$F1,$5F,$97,$9F,$F5,$0C,$60,$95,$FE,$18,$9D,$4F
	.dc.b	$E4,$CD,$C6,$67,$A3,$FE,$33,$BF,$F3,$F5,$38,$73,$7E,$F2,$4E,$A3
	.dc.b	$2F,$E0,$E3,$E5,$1F,$B1,$8C,$D7,$D1,$8C,$9F,$B0,$8C,$B8,$AF,$C3
	.dc.b	$C6,$23,$E5,$CF,$17,$83,$F1,$6B,$C1,$1A,$3C,$11,$93,$85,$74,$71
	.dc.b	$1D,$EF,$E8,$5F,$7A,$FC,$BC,$62,$B7,$5D,$49,$3F,$16,$32,$F0,$7E
	.dc.b	$51,$FC,$5A,$BA,$57,$56,$C7,$06,$D7,$CC,$8E,$07,$49,$E0,$D3,$81
	.dc.b	$63,$81,$F0,$0A,$79,$A3,$D4,$E2,$BC,$CA,$FE,$D5,$73,$7E,$5C,$FB
	.dc.b	$67,$85,$78,$9B,$5E,$26,$9D,$6F,$D7,$7C,$9B,$3F,$49,$B3,$F7,$DA
	.dc.b	$FC,$C6,$DF,$D7,$8E,$B5,$CE,$BC,$C7,$32,$C7,$36,$BC,$DB,$3C,$E7
	.dc.b	$DB,$7E,$B6,$CF,$AD,$B5,$EB,$8C,$1D,$DC,$12,$3C,$8D,$8D,$46,$08
	.dc.b	$E0,$91,$E7,$38,$16,$36,$D8,$DB,$EB,$36,$58,$DB,$63,$75,$EF,$6D
	.dc.b	$7E,$0A,$BC,$F5,$FE,$A5,$AF,$BC,$DB,$FE,$1A,$4B,$FC,$DC,$7E,$36
	.dc.b	$91,$FE,$B0,$FC,$E9,$F0,$DD,$63,$1F,$1B,$FE,$40,$97,$C2,$58,$C3
	.dc.b	$67,$FD,$5B,$29,$FF,$4A,$36,$33,$18,$38,$61,$F9,$98,$F1,$7C,$6E
	.dc.b	$31,$F2,$CF,$FB,$36,$DF,$A1,$78,$FE,$1F,$C8,$DC,$3D,$B4,$AD,$C7
	.dc.b	$92,$7C,$64,$70,$70,$8D,$CF,$8C,$F0,$7E,$99,$DA,$75,$F3,$9F,$0F
	.dc.b	$CA,$F8,$8D,$8C,$9F,$09,$F8,$24,$70,$6A,$BB,$1E,$7F,$8E,$E7,$1E
	.dc.b	$3F,$53,$E8,$7C,$D7,$D1,$6C,$64,$F9,$EF,$CE,$23,$83,$55,$DB,$27
	.dc.b	$AD,$CE,$7E,$07,$A7,$F7,$7D,$5F,$BC,$D8,$C9,$F7,$1F,$B6,$47,$06
	.dc.b	$AF,$B7,$AD,$CA,$3E,$67,$C9,$7E,$27,$D5,$FC,$53,$63,$27,$E2,$1F
	.dc.b	$C3,$91,$C1,$AA,$F5,$7A,$9C,$A3,$EF,$F0,$3E,$4F,$F2,$DF,$0F,$FE
	.dc.b	$96,$8C,$4F,$DD,$8C,$4F,$FB,$9E,$73,$62,$15,$CB,$80,$E1,$DB,$CB
	.dc.b	$9B,$F2,$F1,$A9,$FD,$47,$A1,$E0,$C3,$F2,$BF,$00,$F8,$9F,$F5,$5A
	.dc.b	$F8,$BB,$18,$6C,$F8,$70,$F3,$57,$18,$C4,$23,$0E,$DA,$FF,$05,$F1
	.dc.b	$3F,$C1,$E4,$9F,$F0,$63,$F0,$FE,$16,$47,$4D,$11,$8A,$F1,$68,$E1
	.dc.b	$FF,$61,$B6,$FF,$62,$B8,$B3,$3A,$64,$FD,$8F,$81,$B7,$FA,$21,$E5
	.dc.b	$7F,$B3,$9B,$3F,$BC,$5E,$56,$67,$4C,$9D,$79,$56,$BA,$3C,$CF,$C2
	.dc.b	$7C,$15,$9E,$9E,$5F,$73,$C1,$BD,$8F,$03,$87,$C5,$6F,$81,$FC,$47
	.dc.b	$C1,$3C,$EF,$C3,$5F,$A6,$B4,$7D,$62,$9F,$62,$B0,$E0,$FC,$8F,$C0
	.dc.b	$7E,$DB,$E0,$BE,$EB,$63,$F0,$4D,$7F,$28,$D7,$8F,$83,$FE,$03,$76
	.dc.b	$D7,$6D,$77,$C7,$DB,$BF,$F4,$2F,$E6,$AB,$EF,$4F,$F2,$C6,$BF,$C6
	.dc.b	$14,$F0,$D3,$D3,$C9,$6B,$F5,$5F,$05,$5F,$D4,$B6,$F9,$E1,$DF,$97
	.dc.b	$1E,$04,$BF,$04,$57,$C1,$DF,$C4,$C5,$79,$DE,$DD,$2F,$82,$A6,$EB
	.dc.b	$FA,$2F,$13,$9B,$C8,$B9,$69,$88,$B3,$FC,$3B,$E0,$BA,$98,$7F,$9C
	.dc.b	$FC,$17,$AB,$87,$F8,$4B,$63,$F8,$63,$53,$F2,$1F,$EA,$FE,$E3,$E0
	.dc.b	$BE,$33,$63,$31,$F2,$67,$97,$CA,$3E,$2F,$F1,$8D,$AF,$E0,$6D,$7F
	.dc.b	$84,$35,$FE,$35,$F0,$8F,$83,$78,$DF,$C2,$A3,$0D,$F8,$47,$76,$6B
	.dc.b	$77,$FC,$4F,$F3,$D7,$C4,$FF,$20,$8F,$E5,$31,$8D,$FF,$8A,$B6,$38
	.dc.b	$4F,$FA,$83,$E1,$25,$7F,$A7,$5B,$3F,$C7,$FD,$B7,$C1,$FF,$A9,$89
	.dc.b	$5F,$EA,$37,$C1,$3F,$CE,$70,$F6,$4F,$83,$1C,$CA,$72,$FA,$63,$3C
	.dc.b	$4F,$85,$A4,$BF,$E5,$B1,$96,$E5,$C0,$8E,$0F,$7E,$BE,$0C,$4D,$24
	.dc.b	$74,$CB,$C2,$47,$07,$83,$E1,$3C,$A4,$70,$7C,$CF,$8D,$EC,$47,$07
	.dc.b	$A3,$E0,$BD,$A4,$70,$7B,$3E,$0B,$E0,$91,$C1,$F6,$3E,$67,$D4,$47
	.dc.b	$07,$C4,$F8,$2F,$92,$47,$07,$C8,$F8,$5F,$9A,$47,$07,$E1,$0F,$8C
	.dc.b	$43,$07,$D0,$EE,$FA,$A4,$70,$7D,$4F,$3F,$EC,$91,$C1,$F6,$3E,$6B
	.dc.b	$EE,$91,$C1,$F7,$3B,$FF,$02,$47,$07,$E0,$3C,$0F,$C2,$91,$C1,$F8
	.dc.b	$4E,$6F,$F2,$AF,$89,$FE,$00,$E5,$7C,$DE,$2B,$FC,$17,$1F,$E0,$D7
	.dc.b	$F7,$F1,$9A,$FF,$08,$4E,$5F,$EF,$CD,$7F,$C2,$0D,$79,$9F,$98,$39
	.dc.b	$07,$95,$7F,$C0,$0E,$0C,$C6,$2F,$89,$FC,$42,$B8,$F4,$AB,$71,$27
	.dc.b	$47,$F9,$B2,$74,$E2,$3F,$CB,$35,$D5,$F5,$9D,$2B,$93,$A2,$DF,$96
	.dc.b	$B9,$3F,$26,$33,$1F,$E3,$2B,$EF,$1C,$63,$C5,$5A,$8F,$F1,$C4,$EA
	.dc.b	$B9,$56,$8E,$A7,$3A,$E1,$4A,$E1,$3F,$12,$3F,$CE,$71,$87,$D7,$1C
	.dc.b	$DF,$11,$A9,$E7,$3B,$8F,$A4,$7C,$19,$F0,$BE,$69,$D9,$23,$31,$E3
	.dc.b	$8C,$3F,$C8,$5F,$E3,$33,$F0,$A3,$FD,$7D,$18,$3B,$3E,$8E,$1B,$7F
	.dc.b	$36,$3C,$D7,$F6,$67,$2F,$C3,$3D,$93,$FE,$58,$B4,$EA,$BD,$15,$A3
	.dc.b	$EA,$46,$63,$D1,$19,$7E,$9C,$66,$7E,$21,$1C,$5E,$59,$5C,$6B,$EA
	.dc.b	$46,$1F,$49,$ED,$8D,$57,$99,$B3,$BB,$F8,$07,$AD,$C4,$7B,$EB,$91
	.dc.b	$D2,$75,$1C,$89,$19,$7D,$E6,$23,$AB,$88,$D1,$F0,$1D,$1D,$E3,$AC
	.dc.b	$68,$F6,$C6,$9F,$9A,$79,$23,$AB,$CC,$E1,$5E,$F6,$D7,$B1,$E3,$3D
	.dc.b	$8E,$2F,$63,$75,$DD,$FA,$E7,$77,$CC,$7A,$E3,$8D,$77,$6D,$77,$AE
	.dc.b	$31,$DE,$B8,$C7,$9E,$37,$1D,$EB,$F0,$0D,$BF,$E1,$F7,$CE,$73,$7E
	.dc.b	$62,$5F,$26,$DF,$67,$D9,$19,$24,$66,$7D,$0D,$AE,$D7,$E0,$8F,$11
	.dc.b	$FA,$32,$DF,$E1,$0F,$F5,$F5,$7C,$F6,$D7,$EC,$89,$3F,$EB,$09,$D4
	.dc.b	$FE,$0C,$DC,$66,$7C,$11,$FC,$33,$E1,$AE,$1F,$26,$3E,$45,$7F,$1F
	.dc.b	$9C,$BF,$4C,$91,$FC,$7A,$32,$F0,$21,$FC,$6D,$CB,$F1,$A3,$8B,$97
	.dc.b	$E0,$3F,$81,$38,$DA,$78,$BF,$2A,$33,$18,$BC,$B6,$B9,$DF,$F0,$98
	.dc.b	$C5,$75,$3C,$05,$BE,$8F,$94,$F6,$57,$23,$CF,$1A,$3C,$8E,$63,$CF
	.dc.b	$7F,$D6,$2B,$BD,$FC,$B9,$EE,$DB,$E5,$5B,$BE,$55,$E1,$24,$F6,$7C
	.dc.b	$D3,$EE,$9D,$CF,$27,$FB,$15,$F8,$1F,$3B,$6B,$E3,$C7,$0A,$E5,$1D
	.dc.b	$23,$9C,$EE,$3C,$2D,$AF,$80,$DA,$F4,$4F,$13,$E5,$46,$6B,$C6,$77
	.dc.b	$3A,$96,$36,$E1,$A4,$BF,$8D,$3E,$66,$CE,$DB,$3B,$8F,$79,$92,$B9
	.dc.b	$3E,$09,$D8,$FD,$5C,$FF,$5A,$24,$7E,$A5,$F6,$CF,$F5,$1E,$1B,$3F
	.dc.b	$A3,$3D,$B5,$FD,$D2,$32,$FE,$20,$FD,$64,$7F,$85,$63,$31,$F8,$28
	.dc.b	$E4,$F1,$79,$CF,$E5,$DA,$7F,$83,$E7,$8E,$1F,$83,$9D,$57,$13,$B9
	.dc.b	$1C,$9E,$9A,$E7,$1E,$58,$CC,$75,$33,$1E,$53,$4E,$5E,$8F,$A2,$38
	.dc.b	$C7,$F8,$7A,$BC,$B7,$E6,$6D,$74,$B4,$62,$B9,$9C,$67,$55,$C5,$E1
	.dc.b	$36,$76,$3D,$C7,$C2,$FF,$AA,$9B,$49,$84,$8C,$B8,$34,$E1,$C2,$F8
	.dc.b	$5F,$F2,$CB,$C3,$25,$73,$9C,$DE,$9F,$11,$23,$B5,$F6,$8F,$37,$C2
	.dc.b	$49,$CC,$FA,$E3,$0F,$0C,$1F,$09,$E2,$9F,$1B,$6B,$DD,$8F,$92,$FD
	.dc.b	$73,$F0,$C9,$3D,$AF,$9B,$7C,$AF,$90,$47,$06,$FE,$0F,$E0,$C7,$C2
	.dc.b	$6C,$78,$0F,$98,$FD,$12,$3C,$4F,$07,$C1,$7D,$52,$38,$3C,$9E,$A7
	.dc.b	$CE,$8F,$9E,$D8,$ED,$8F,$D6,$8F,$BC,$48,$E2,$7B,$7D,$4F,$C1,$11
	.dc.b	$C1,$F0,$7D,$0F,$B7,$1F,$71,$B1,$DA,$BF,$19,$1F,$8B,$24,$71,$3E
	.dc.b	$57,$A3,$F8,$F2,$38,$3E,$77,$C0,$FE,$22,$3F,$12,$D8,$ED,$A2,$3F
	.dc.b	$30,$48,$E2,$7D,$7F,$81,$FC,$E1,$1C,$1F,$88,$D5,$F0,$3F,$96,$9F
	.dc.b	$F5,$7C,$62,$7B,$C6,$27,$F0,$D9,$7E,$86,$B4,$FE,$59,$AE,$63,$6F
	.dc.b	$EA,$BD,$F7,$C0,$62,$3F,$5D,$18,$C3,$F3,$6D,$8F,$D0,$B5,$FD,$82
	.dc.b	$F8,$1F,$F6,$2B,$BE,$3F,$2B,$9A,$FA,$8E,$5F,$F5,$87,$9D,$FE,$C4
	.dc.b	$EF,$FF,$62,$B6,$3E,$D6,$6F,$81,$CC,$61,$B9,$31,$F8,$63,$DB,$FF
	.dc.b	$6B,$B6,$BF,$08,$B8,$B3,$3A,$64,$F8,$3E,$06,$DF,$58,$FE,$1D,$E2
	.dc.b	$FF,$B8,$A0,$DF,$81,$C3,$67,$F2,$52,$9E,$1E,$2E,$63,$A1,$FC,$71
	.dc.b	$CC,$FC,$0F,$81,$B3,$A3,$F9,$16,$FE,$13,$C5,$BD,$AE,$EE,$1F,$0C
	.dc.b	$BE,$03,$F9,$57,$C0,$BC,$AF,$E5,$17,$EB,$AD,$1F,$C6,$8A,$7F,$1C
	.dc.b	$AC,$38,$3F,$99,$FB,$EF,$82,$7C,$17,$F2,$28,$C3,$63,$33,$FC,$AA
	.dc.b	$72,$DE,$FF,$B4,$DE,$4E,$51,$88,$FE,$7D,$28,$C4,$78,$BB,$FF,$A2
	.dc.b	$D7,$E9,$2B,$F0,$53,$FC,$0D,$AF,$CB,$53,$C3,$5F,$5F,$2D,$AF,$EA
	.dc.b	$1C,$EF,$82,$C4,$23,$0F,$59,$CE,$1F,$8D,$8F,$07,$81,$83,$3D,$16
	.dc.b	$3C,$9F,$05,$C0,$C3,$E5,$7B,$E0,$F8,$2A,$6E,$7F,$8F,$71,$B9,$FC
	.dc.b	$AB,$9A,$98,$8B,$3F,$AC,$F8,$2E,$A6,$1F,$C7,$7C,$0F,$C6,$E2,$FE
	.dc.b	$1A,$DE,$1F,$EE,$2D,$9F,$EE,$8F,$CA,$7F,$98,$7A,$0F,$82,$F8,$ED
	.dc.b	$8C,$C7,$CA,$9E,$50,$F8,$D7,$7F,$E4,$5B,$1F,$DB,$1A,$FF,$72,$6B
	.dc.b	$FE,$01,$FC,$FB,$E0,$3F,$DC,$9A,$FF,$78,$6B,$BD,$7F,$3D,$F0,$4F
	.dc.b	$80,$7C,$17,$F7,$A8,$C3,$7D,$E7,$81,$9A,$DE,$1F,$E0,$4F,$8C,$F8
	.dc.b	$1C,$1F,$E0,$26,$C6,$77,$FE,$5C,$A3,$2F,$F8,$67,$E7,$DF,$0F,$A6
	.dc.b	$DF,$F8,$31,$B3,$FE,$12,$6B,$F0,$71,$EB,$B7,$9A,$BE,$14,$60,$CC
	.dc.b	$C8,$F8,$CD,$3F,$C3,$BF,$24,$F8,$BF,$3B,$67,$FC,$3A,$D9,$DC,$EB
	.dc.b	$D7,$95,$A6,$CF,$4F,$8B,$E6,$49,$E6,$F6,$F6,$24,$76,$6C,$F9,$3D
	.dc.b	$8F,$19,$27,$C6,$F9,$BE,$2E,$47,$99,$B3,$EC,$F6,$7D,$64,$9F,$5B
	.dc.b	$EE,$F0,$24,$7B,$9B,$3F,$13,$83,$E1,$92,$7E,$1B,$F1,$BD,$99,$1F
	.dc.b	$19,$B3,$F3,$3D,$9F,$96,$49,$F9,$6F,$CE,$F2,$E4,$7C,$E6,$CF,$D4
	.dc.b	$F2,$FE,$99,$27,$E9,$BF,$5B,$CB,$91,$F5,$9B,$3F,$73,$CA,$FB,$64
	.dc.b	$9F,$B6,$FD,$EF,$22,$47,$DE,$6C,$FF,$98,$3D,$C7,$C3,$FF,$95,$7C
	.dc.b	$5F,$F0,$8C,$3E,$23,$18,$7F,$9D,$A3,$6F,$F8,$CD,$D3,$6B,$F4,$71
	.dc.b	$AB,$ED,$18,$24,$EC,$E5,$F1,$DD,$A7,$FD,$09,$5C,$70,$F1,$DE,$65
	.dc.b	$39,$BF,$25,$E6,$31,$3C,$AB,$06,$66,$D7,$80,$F1,$4F,$F7,$3A,$4E
	.dc.b	$A3,$0D,$BF,$8A,$DA,$F0,$9C,$A3,$2D,$33,$3F,$3E,$7E,$29,$23,$FD
	.dc.b	$1B,$5E,$07,$C9,$3E,$3F,$88,$F4,$46,$2F,$AC,$E6,$30,$76,$8C,$9F
	.dc.b	$6A,$78,$CE,$CF,$5C,$FD,$98,$CC,$EE,$39,$CF,$D2,$9C,$C7,$E0,$5B
	.dc.b	$19,$8D,$BF,$52,$3F,$02,$48,$F6,$B6,$3C,$53,$F8,$EF,$88,$FB,$13
	.dc.b	$2B,$C8,$7E,$36,$31,$5C,$23,$C4,$64,$DB,$93,$C8,$7D,$B8,$E3,$5E
	.dc.b	$D8,$D1,$F8,$A8,$CC,$F4,$8C,$DA,$73,$1D,$2B,$91,$E9,$AF,$0B,$63
	.dc.b	$8C,$F0,$7B,$4E,$8F,$D5,$46,$6B,$B4,$66,$7F,$43,$39,$24,$66,$39
	.dc.b	$9D,$9C,$4F,$28,$C1,$2B,$CF,$3E,$03,$75,$D2,$75,$1F,$06,$7D,$47
	.dc.b	$08,$F1,$C7,$02,$3C,$0F,$88,$F5,$70,$7E,$0D,$FA,$CE,$29,$7F,$B7
	.dc.b	$BF,$15,$F7,$AF,$6B,$6B,$D8,$D9,$DD,$F9,$CF,$04,$65,$F2,$B6,$BC
	.dc.b	$17,$DE,$72,$71,$8F,$69,$A7,$CF,$3C,$48,$EE,$3C,$AD,$BF,$2B,$6F
	.dc.b	$CB,$3F,$10,$EB,$5C,$9C,$1C,$8B,$5C,$9B,$3C,$9F,$94,$7E,$6E,$BD
	.dc.b	$2E,$9A,$EA,$BF,$24,$8F,$A8,$53,$DE,$53,$F1,$65,$AF,$C5,$36,$BF
	.dc.b	$9D,$92,$BF,$8B,$BF,$2E,$3F,$94,$7C,$75,$BF,$C4,$D7,$D4,$3F,$A4
	.dc.b	$46,$5C,$4F,$C4,$72,$7D,$03,$2F,$C5,$3F,$9F,$C7,$17,$81,$0C,$C7
	.dc.b	$17,$17,$F2,$DC,$9F,$24,$95,$E6,$7F,$A7,$D7,$57,$C1,$3C,$44,$65
	.dc.b	$C5,$FF,$4F,$8C,$39,$6D,$F8,$EB,$C3,$7D,$DE,$B5,$C6,$75,$7D,$1F
	.dc.b	$ED,$97,$CA,$BC,$F1,$F1,$9A,$7B,$AB,$93,$83,$AC,$72,$3A,$C7,$95
	.dc.b	$A7,$82,$7B,$5F,$96,$FD,$F3,$AB,$E9,$18,$6D,$F3,$6C,$F3,$38,$57
	.dc.b	$3B,$EB,$1F,$72,$F9,$4F,$7B,$E4,$D9,$E5,$7D,$0F,$0D,$F2,$7E,$EB
	.dc.b	$5F,$AE,$F2,$8E,$CD,$8D,$C7,$49,$F2,$34,$D9,$63,$67,$81,$C3,$D9
	.dc.b	$B7,$F8,$5A,$FF,$0F,$92,$3F,$BD,$B6,$7F,$4B,$1F,$B3,$A3,$FB,$3E
	.dc.b	$C7,$9F,$F6,$A7,$FC,$63,$76,$B5,$36,$FF,$2F,$1F,$75,$57,$47,$F8
	.dc.b	$05,$F4,$9D,$4F,$46,$9D,$0B,$3D,$0F,$39,$D1,$D3,$88,$E4,$D8,$CC
	.dc.b	$60,$8F,$E5,$DF,$24,$27,$C8,$F1,$3C,$F5,$ED,$74,$D8,$DC,$F4,$9D
	.dc.b	$D7,$A1,$DB,$67,$6F,$84,$F2,$96,$38,$9E,$13,$89,$63,$8B,$F0,$E7
	.dc.b	$D3,$18,$8F,$DD,$CF,$47,$4F,$C6,$7F,$96,$47,$B9,$B5,$FE,$B0,$24
	.dc.b	$FE,$0C,$E5,$1F,$17,$D0,$FF,$2F,$57,$71,$EF,$FF,$AE,$09,$5F,$EB
	.dc.b	$87,$FD,$77,$EF,$C8,$FF,$5D,$B6,$FF,$D8,$3E,$47,$FB,$00,$93,$FE
	.dc.b	$C0,$7F,$DB,$3F,$14,$F8,$0A,$49,$FE,$8D,$5F,$D2,$E7,$51,$88,$43
	.dc.b	$E4,$39,$92,$75,$19,$6F,$A5,$E0,$23,$83,$7E,$F7,$90,$8E,$0E,$7F
	.dc.b	$25,$E8,$23,$83,$BF,$C9,$7B,$08,$E0,$F1,$FB,$5F,$00,$8E,$0F,$3E
	.dc.b	$FF,$10,$8E,$0F,$5F,$A9,$F2,$08,$E0,$F7,$FB,$3F,$30,$8E,$0F,$87
	.dc.b	$F2,$1F,$40,$8E,$0F,$8F,$F0,$9F,$50,$8E,$0F,$97,$F0,$9F,$60,$8E
	.dc.b	$0F,$9F,$F0,$3F,$70,$8E,$0F,$A7,$F2,$5F,$80,$23,$83,$FD,$27,$F1
	.dc.b	$1F,$EB,$A2,$38,$27,$1F,$FA,$9E,$3F,$D6,$1F,$11,$FE,$B1,$6C,$FF
	.dc.b	$A7,$23,$FD,$86,$47,$F1,$1E,$5F,$E2,$AB,$F1,$AD,$7F,$C2,$8D,$7F
	.dc.b	$C3,$0D,$7F,$C3,$8F,$1E,$CF,$E1,$AD,$9F,$F0,$F7,$C4,$3A,$7F,$D9
	.dc.b	$13,$9B,$FF,$66,$FA,$9F,$EC,$7F,$73,$FD,$90,$DC,$3F,$69,$3D,$B9
	.dc.b	$F3,$18,$85,$A7,$31,$FC,$9C,$FE,$7D,$E4,$7F,$B5,$F3,$37,$C8,$C3
	.dc.b	$70,$E9,$29,$F3,$F2,$E6,$FA,$1F,$E3,$BF,$87,$FF,$71,$26,$FC,$3E
	.dc.b	$1B,$87,$9E,$53,$C3,$E0,$B3,$1D,$03,$99,$F7,$68,$E4,$C5,$6F,$7F
	.dc.b	$09,$E2,$DE,$DF,$77,$0F,$86,$5F,$0D,$FE,$53,$F8,$77,$C2,$FF,$94
	.dc.b	$6F,$D7,$5A,$3F,$86,$14,$FE,$1D,$58,$70,$7F,$96,$FE,$19,$E9,$3E
	.dc.b	$1F,$F8,$A4,$61,$B1,$99,$FE,$37,$39,$6F,$A7,$FD,$A9,$BC,$9C,$A3
	.dc.b	$11,$FE,$67,$94,$62,$3C,$5E,$9F,$F9,$B6,$BF,$AF,$D7,$E0,$27,$F1
	.dc.b	$0D,$7E,$DA,$9E,$1A,$FA,$F9,$AD,$7E,$7B,$C2,$AF,$E4,$1B,$7C,$F0
	.dc.b	$EF,$CF,$8F,$2E,$5F,$82,$2B,$E0,$F3,$38,$98,$AF,$3B,$DB,$95,$E1
	.dc.b	$D3,$75,$FA,$1E,$3F,$C3,$B7,$97,$CB,$06,$F3,$E1,$E9,$E3,$CA,$E0
	.dc.b	$7E,$AF,$E1,$BD,$5C,$3F,$E9,$56,$C7,$FA,$65,$A9,$F9,$0F,$1E,$1F
	.dc.b	$8B,$F0,$EF,$83,$A7,$C8,$AE,$3F,$0E,$E4,$FC,$6B,$70,$FE,$B4,$D7
	.dc.b	$FA,$FB,$5F,$F5,$32,$EF,$E2,$48,$E2,$FE,$63,$E4,$5E,$4D,$24,$FF
	.dc.b	$67,$8C,$BF,$DA,$7E,$21,$F0,$FF,$C5,$E3,$FB,$7C,$63,$C5,$D7,$C4
	.dc.b	$3C,$1E,$17,$BB,$FF,$5D,$7C,$47,$C1,$E2,$FE,$E8,$D8,$FE,$F2,$D7
	.dc.b	$7C,$1F,$E3,$47,$FD,$89,$F0,$FF,$EF,$4A,$FE,$F8,$D8,$FF,$00,$B5
	.dc.b	$FF,$64,$35,$E3,$87,$37,$FD,$A1,$1E,$79,$E7,$39,$94,$E5,$EA,$65
	.dc.b	$FF,$6C,$4F,$8A,$30,$67,$D9,$91,$E6,$6C,$7F,$85,$BD,$07,$C3,$77
	.dc.b	$6D,$7F,$85,$DB,$3B,$9D,$7C,$34,$AD,$36,$7E,$0F,$B5,$E5,$24,$FC
	.dc.b	$67,$B7,$C2,$C8,$EC,$D9,$F2,$7C,$2F,$8C,$93,$E3,$7E,$07,$C5,$C8
	.dc.b	$FA,$2D,$9F,$67,$83,$EB,$24,$FA,$DF,$77,$C2,$C8,$F7,$36,$7E,$CF
	.dc.b	$C5,$FC,$D2,$4F,$DE,$7E,$37,$C2,$C8,$F8,$CD,$9F,$99,$E6,$7C,$B2
	.dc.b	$4F,$CB,$7E,$C7,$AD,$23,$F1,$2D,$9F,$A9,$E6,$7D,$32,$4F,$D3,$7E
	.dc.b	$B7,$85,$23,$EB,$36,$7F,$CA,$3E,$E3,$E4,$7F,$D3,$DE,$5F,$FA,$33
	.dc.b	$9D,$EC,$63,$16,$1F,$E5,$7B,$C3,$F9,$49,$AF,$F9,$62,$3F,$CC,$24
	.dc.b	$9F,$F2,$BB,$63,$FB,$8E,$1E,$19,$FF,$01,$7C,$13,$E1,$B9,$CE,$6B
	.dc.b	$77,$83,$BD,$6E,$FF,$C1,$55,$BC,$BF,$D4,$AD,$8F,$4D,$F9,$5E,$F8
	.dc.b	$73,$69,$CB,$0E,$B1,$99,$FF,$03,$E1,$DD,$F1,$D7,$06,$9A,$9E,$18
	.dc.b	$7F,$AD,$AB,$83,$E4,$BE,$B7,$FE,$7F,$9E,$97,$E6,$0E,$1F,$4B,$5F
	.dc.b	$A6,$F4,$8C,$DF,$C4,$F8,$6E,$B5,$E3,$9F,$36,$1E,$39,$CB,$E5,$AB
	.dc.b	$3D,$A6,$D7,$66,$BE,$46,$D9,$A4,$9E,$E7,$32,$CF,$37,$E0,$47,$F1
	.dc.b	$99,$E5,$18,$24,$FD,$09,$FB,$FF,$0D,$ED,$9F,$C3,$C5,$9F,$AB,$37
	.dc.b	$0F,$75,$68,$FC,$6C,$65,$F9,$6D,$9F,$96,$D7,$E8,$B6,$CD,$24,$F5
	.dc.b	$9F,$0B,$F6,$E3,$D8,$47,$07,$88,$FC,$FF,$C3,$7B,$1C,$4F,$23,$F1
	.dc.b	$F1,$97,$17,$F8,$BA,$FD,$0C,$E6,$3F,$1C,$DA,$FC,$3B,$6B,$F0,$ED
	.dc.b	$7F,$1A,$DB,$FE,$A9,$5D,$6B,$CA,$79,$1E,$AD,$3A,$96,$BA,$B5,$EA
	.dc.b	$D9,$EA,$FD,$03,$AC,$7B,$1E,$6D,$39,$94,$E6,$58,$E6,$FD,$88,$E7
	.dc.b	$3E,$98,$F9,$31,$E1,$AF,$71,$1C,$12,$3A,$46,$5B,$19,$9E,$87,$C2
	.dc.b	$8C,$91,$CC,$7A,$9B,$1E,$A6,$BE,$A6,$D7,$AA,$3F,$1F,$3E,$A3,$E3
	.dc.b	$3E,$A3,$B3,$89,$F1,$1C,$AB,$D0,$D8,$F4,$1B,$3C,$85,$9F,$23,$5F
	.dc.b	$23,$6B,$C8,$7E,$81,$F2,$47,$CF,$3C,$93,$DD,$F2,$99,$7C,$B1,$CE
	.dc.b	$BE,$64,$FF,$1D,$AF,$F0,$F3,$F7,$8C,$47,$F0,$32,$4F,$F2,$56,$C7
	.dc.b	$F2,$53,$CE,$7A,$8F,$E4,$7D,$BF,$A4,$9F,$3C,$68,$FE,$57,$19,$7D
	.dc.b	$31,$F1,$1E,$2F,$0A,$F8,$35,$FD,$3A,$B8,$5F,$B2,$FC,$07,$4B,$E1
	.dc.b	$5A,$7F,$9B,$D5,$9D,$DF,$CC,$8C,$47,$96,$31,$3C,$8F,$14,$EA,$3C
	.dc.b	$71,$FD,$9A,$31,$59,$6C,$70,$39,$57,$B8,$F6,$94,$F0,$4F,$B6,$37
	.dc.b	$7E,$D9,$F6,$1B,$BF,$0B,$FD,$8A,$FB,$B9,$6C,$F1,$94,$62,$BE,$75
	.dc.b	$74,$AD,$5F,$A6,$7C,$26,$AF,$8B,$8A,$F9,$37,$CD,$B1,$97,$E5,$46
	.dc.b	$5B,$19,$9F,$9B,$1E,$BA,$F0,$36,$7E,$43,$4F,$63,$C6,$3E,$34,$60
	.dc.b	$97,$EF,$BE,$6D,$AE,$6D,$9D,$C7,$B6,$76,$D3,$65,$8F,$CF,$5F,$C2
	.dc.b	$6C,$FF,$67,$24,$FE,$0C,$FF,$14,$F5,$79,$AF,$FC,$21,$36,$7F,$C1
	.dc.b	$33,$6B,$F0,$75,$A9,$E2,$73,$8F,$BC,$D8,$FF,$0D,$4F,$19,$FB,$53
	.dc.b	$F9,$E3,$8B,$C0,$93,$D5,$FF,$14,$CF,$1B,$34,$95,$DC,$F3,$96,$7F
	.dc.b	$C3,$F1,$A7,$AC,$69,$A6,$9E,$0F,$F8,$AA,$38,$57,$8C,$F5,$CE,$AB
	.dc.b	$C6,$D9,$DB,$ED,$AF,$4B,$FE,$37,$8F,$F0,$C1,$1C,$1F,$0A,$8F,$F9
	.dc.b	$43,$B5,$E8,$FF,$96,$2B,$FC,$CF,$C2,$F8,$FF,$F5,$B1,$2F,$FD,$8E
	.dc.b	$FF,$AB,$FE,$1E,$47,$FA,$BD,$B9,$7F,$AD,$7E,$1F,$FD,$68,$49,$FF
	.dc.b	$5A,$3F,$EC,$1F,$92,$91,$FE,$D2,$6C,$FF,$A0,$BE,$23,$FD,$95,$CA
	.dc.b	$69,$2B,$F9,$84,$A5,$F1,$1C,$09,$3A,$8C,$BF,$EE,$0F,$5B,$FD,$A4
	.dc.b	$47,$06,$FE,$0B,$C2,$47,$07,$3F,$88,$F2,$91,$C1,$FE,$EC,$F8,$3F
	.dc.b	$F6,$F1,$1C,$1E,$3F,$53,$DA,$47,$07,$9F,$8F,$E0,$91,$C1,$FE,$F8
	.dc.b	$F8,$3F,$F7,$51,$1C,$1E,$FF,$6F,$E4,$91,$C1,$F0,$FD,$4F,$9A,$47
	.dc.b	$07,$CC,$E4,$7B,$7F,$EF,$72,$46,$5F,$AB,$F0,$D2,$3E,$73,$78,$7F
	.dc.b	$D7,$7F,$09,$F6,$89,$3F,$4D,$FF,$37,$79,$12,$3F,$D2,$0D,$9F,$E1
	.dc.b	$FE,$37,$F9,$CC,$93,$FE,$95,$6C,$7F,$A8,$9F,$F5,$8F,$89,$FE,$B2
	.dc.b	$7F,$D8,$3E,$57,$E2,$6B,$FD,$8F,$DD,$FA,$2B,$FF,$64,$FC,$2B,$AF
	.dc.b	$7D,$5F,$EC,$88,$C5,$F0,$3F,$D6,$F8,$73,$BF,$F0,$03,$5F,$F5,$33
	.dc.b	$5D,$D7,$FA,$2F,$DE,$FF,$62,$9F,$EC,$6F,$4A,$DF,$5A,$FF,$6D,$DB
	.dc.b	$A7,$53,$99,$4E,$63,$FD,$AB,$5F,$ED,$EC,$DE,$4D,$31,$EF,$DA,$E5
	.dc.b	$28,$CC,$75,$B7,$B7,$8C,$57,$F1,$DF,$F7,$7F,$5F,$FB,$B1,$37,$A7
	.dc.b	$0D,$C3,$D1,$29,$E1,$CD,$98,$EB,$3F,$EC,$D8,$CB,$E3,$E5,$B3,$A3
	.dc.b	$FD,$CF,$BF,$84,$F1,$6D,$6B,$BB,$87,$C3,$2F,$84,$2F,$85,$77,$FF
	.dc.b	$B7,$6F,$D9,$3B,$3F,$86,$94,$FE,$1F,$58,$70,$7F,$BE,$BE,$0D,$E8
	.dc.b	$3E,$17,$F8,$AC,$61,$B1,$99,$FF,$24,$4E,$5B,$F0,$BF,$41,$BC,$3C
	.dc.b	$A3,$11,$FE,$F3,$94,$62,$3C,$5E,$97,$B9,$F9,$02,$7F,$0A,$FE,$2D
	.dc.b	$AE,$94,$71,$AF,$C0,$46,$9F,$C1,$72,$3C,$3D,$6F,$FA,$46,$E1,$F3
	.dc.b	$AF,$84,$FC,$5E,$6C,$CF,$8F,$E1,$6D,$1B,$38,$73,$E6,$4A,$F4,$45
	.dc.b	$7D,$B8,$BC,$4E,$AD,$8F,$75,$78,$38,$BF,$8D,$C7,$EA,$D7,$26,$4D
	.dc.b	$73,$59,$FD,$27,$C2,$F4,$30,$FC,$CF,$84,$F8,$EF,$ED,$5F,$7B,$F8
	.dc.b	$E6,$A7,$6F,$1E,$CF,$87,$F0,$AF,$7A,$9F,$1A,$B8,$FC,$2B,$E2,$3F
	.dc.b	$26,$DC,$3F,$B0,$B5,$FF,$49,$35,$FC,$10,$7E,$C3,$C3,$93,$99,$1F
	.dc.b	$ED,$7F,$06,$F8,$BF,$D0,$B6,$33,$18,$3F,$4B,$F0,$EF,$36,$92,$7F
	.dc.b	$BA,$C6,$5F,$EE,$FF,$0E,$F8,$67,$C3,$E1,$B3,$FE,$AB,$96,$2F,$0B
	.dc.b	$57,$FE,$00,$FC,$FD,$E2,$FF,$15,$C3,$F4,$11,$8E,$ED,$4E,$67,$F8
	.dc.b	$0C,$E5,$FF,$5C,$7A,$3F,$EC,$04,$D2,$7B,$9C,$63,$38,$D9,$F2,$61
	.dc.b	$FD,$9E,$30,$6E,$DF,$13,$49,$3E,$97,$97,$BB,$23,$FB,$83,$70,$FF
	.dc.b	$66,$7A,$AF,$88,$F1,$36,$BF,$C4,$2D,$9D,$CF,$87,$E0,$A5,$7C,$76
	.dc.b	$CF,$C0,$F7,$BD,$44,$9F,$F0,$03,$DB,$E1,$E4,$76,$6C,$FF,$B7,$F3
	.dc.b	$FB,$53,$F4,$F8,$FE,$01,$1F,$F0,$57,$C3,$7C,$72,$38,$3D,$3E,$27
	.dc.b	$B0,$8E,$0F,$BD,$C9,$F2,$C8,$E0,$FF,$0B,$7B,$3F,$4C,$8E,$0F,$8B
	.dc.b	$F0,$BF,$20,$8E,$0F,$C5,$7C,$9F,$D7,$23,$83,$FC,$45,$ED,$7D,$F2
	.dc.b	$38,$3E,$8F,$9D,$F5,$08,$E0,$FC,$B1,$FE,$0F,$F3,$BF,$B5,$6A,$F9
	.dc.b	$3C,$4F,$E6,$0B,$BF,$E5,$A2,$A9,$FF,$3C,$CF,$42,$47,$D3,$7F,$20
	.dc.b	$70,$8D,$C7,$F4,$2E,$87,$B9,$9A,$87,$F8,$36,$B7,$87,$F7,$FA,$DC
	.dc.b	$E4,$DD,$FF,$4C,$AD,$C7,$F4,$F9,$DC,$E6,$3F,$D1,$77,$C2,$FA,$36
	.dc.b	$38,$36,$7F,$BA,$D7,$2A,$F8,$F5,$C8,$F1,$1F,$E2,$38,$D3,$F2,$4C
	.dc.b	$57,$ED,$48,$FF,$96,$DB,$5F,$0D,$F9,$BE,$FB,$97,$B4,$E6,$B9,$9F
	.dc.b	$E3,$19,$D3,$88,$D1,$30,$FF,$13,$4E,$A3,$13,$FD,$CE,$31,$3F,$0A
	.dc.b	$3A,$57,$82,$38,$B6,$38,$36,$BF,$C3,$35,$CA,$7E,$49,$C8,$F2,$57
	.dc.b	$27,$FD,$41,$5F,$EA,$A2,$38,$38,$57,$F8,$1B,$E1,$1F,$0F,$F1,$2A
	.dc.b	$56,$AF,$FC,$85,$5A,$8C,$1E,$62,$DF,$F8,$BA,$34,$FF,$84,$B0,$F4
	.dc.b	$39,$C3,$AB,$E1,$6D,$F6,$6D,$FF,$86,$EF,$9C,$FB,$6B,$84,$F8,$14
	.dc.b	$EA,$3C,$91,$A3,$FB,$19,$AA,$FE,$79,$3E,$68,$FE,$CE,$D7,$C1,$1F
	.dc.b	$B2,$8E,$91,$ED,$F8,$7C,$4C,$AF,$F3,$B5,$6A,$BC,$75,$97,$C5,$7E
	.dc.b	$DA,$DC,$E6,$3E,$6B,$63,$E0,$36,$7C,$11,$DA,$FC,$93,$99,$EF,$39
	.dc.b	$3A,$95,$FB,$D1,$BB,$34,$93,$BB,$F1,$B4,$F1,$96,$BC,$67,$B1,$CB
	.dc.b	$4C,$C7,$BD,$F0,$C7,$17,$CE,$6C,$B1,$B3,$C2,$E6,$BC,$F3,$C5,$B1
	.dc.b	$9A,$F4,$B6,$75,$18,$AE,$91,$E6,$6C,$F9,$A3,$EF,$9F,$21,$E4,$48
	.dc.b	$DB,$E6,$8D,$BE,$C3,$4E,$23,$D2,$70,$AF,$3B,$6B,$C9,$3B,$3C,$85
	.dc.b	$AF,$21,$FA,$77,$C9,$3C,$5F,$69,$23,$C2,$6C,$F2,$46,$DC,$9F,$A1
	.dc.b	$8E,$67,$77,$D8,$71,$8F,$4C,$66,$BD,$2D,$AF,$D4,$91,$FD,$6D,$7D
	.dc.b	$C3,$F5,$71,$97,$F6,$33,$F6,$1C,$CF,$08,$EB,$1F,$00,$FD,$DF,$C3
	.dc.b	$FE,$7A,$7C,$F1,$99,$4E,$63,$DF,$3F,$29,$FE,$25,$19,$BF,$5C,$71
	.dc.b	$8C,$4E,$E3,$E0,$9E,$07,$BB,$8A,$F2,$B9,$3B,$9E,$3B,$FE,$0D,$5D
	.dc.b	$23,$15,$FC,$12,$31,$1C,$E7,$73,$DA,$BA,$92,$B6,$E2,$77,$5E,$03
	.dc.b	$95,$75,$39,$47,$0A,$E2,$D8,$C9,$DE,$32,$D3,$33,$D6,$BE,$6C,$62
	.dc.b	$FE,$4C,$FC,$16,$CF,$96,$7C,$53,$C6,$FC,$0D,$8E,$CD,$3B,$5F,$89
	.dc.b	$B1,$DC,$FA,$95,$E2,$8E,$17,$E0,$AE,$0D,$AE,$AD,$9E,$64,$70,$48
	.dc.b	$EB,$3F,$D0,$88,$FE,$4D,$B3,$FD,$95,$FE,$85,$DF,$F4,$AB,$FB,$1C
	.dc.b	$5A,$FE,$AD,$36,$BF,$23,$1F,$12,$BF,$B4,$CE,$67,$FA,$BB,$6C,$D2
	.dc.b	$57,$13,$A9,$67,$FB,$74,$69,$FD,$6B,$FD,$70,$91,$FA,$D7,$D2,$FF
	.dc.b	$56,$ED,$B3,$FE,$01,$9B,$7F,$DF,$A6,$D7,$82,$B4,$F1,$3F,$AF,$D6
	.dc.b	$A3,$C7,$19,$76,$4A,$F8,$31,$96,$C6,$63,$B4,$78,$5E,$A4,$8D,$1F
	.dc.b	$E0,$DA,$F0,$3F,$E0,$06,$D9,$A4,$9E,$55,$D6,$B9,$9C,$63,$27,$F8
	.dc.b	$3E,$78,$46,$AB,$A4,$EA,$3C,$4D,$79,$1B,$9E,$2F,$F8,$31,$FF,$07
	.dc.b	$92,$3F,$90,$3F,$E0,$D3,$FC,$1D,$D7,$63,$FC,$BD,$F2,$7F,$EC,$F2
	.dc.b	$3F,$CB,$B8,$11,$FE,$D3,$F8,$89,$39,$6E,$1F,$EC,$AF,$88,$FF,$64
	.dc.b	$92,$7F,$D9,$2F,$FB,$6F,$E4,$E4,$75,$6C,$FF,$B9,$BE,$27,$FD,$C0
	.dc.b	$49,$F1,$3F,$EE,$5E,$27,$C2,$52,$4F,$F4,$48,$CB,$E9,$F4,$3F,$DD
	.dc.b	$44,$70,$7B,$7D,$6F,$F7,$71,$1C,$1C,$FE,$13,$C0,$47,$07,$C5,$E6
	.dc.b	$FF,$7B,$11,$C1,$F2,$7E,$0F,$FD,$F4,$47,$07,$9F,$D0,$F6,$11,$C1
	.dc.b	$F4,$7A,$0D,$26,$BF,$E4,$EE,$2F,$EB,$CF,$FA,$97,$E0,$F6,$E5,$A8
	.dc.b	$D4,$FD,$D8,$F9,$9E,$DC,$8F,$94,$D9,$ED,$D5,$C9,$FF,$59,$35,$FF
	.dc.b	$55,$B5,$5F,$F2,$B7,$C0,$7F,$AC,$3D,$9D,$41,$A4,$BF,$B5,$F0,$FF
	.dc.b	$7C,$8E,$0F,$EF,$5E,$47,$F9,$B0,$8E,$0F,$EF,$DE,$37,$F9,$C0,$8E
	.dc.b	$0F,$F0,$17,$89,$FE,$74,$23,$83,$FC,$1B,$F0,$CF,$87,$F6,$36,$33
	.dc.b	$18,$3E,$07,$95,$F9,$0D,$1E,$0F,$D0,$BE,$11,$8B,$F8,$75,$BA,$D5
	.dc.b	$FF,$A0,$7D,$D9,$5F,$E7,$EF,$46,$E1,$FE,$A9,$7F,$D0,$C7,$FB,$E3
	.dc.b	$07,$3E,$2B,$EF,$63,$F8,$01,$EE,$7F,$BF,$9B,$1A,$3E,$AA,$F8,$0D
	.dc.b	$4E,$D9,$3B,$C7,$F1,$58,$3E,$1F,$31,$ED,$3E,$FF,$4F,$E4,$B1,$FC
	.dc.b	$43,$6F,$F0,$72,$E7,$C3,$C2,$3B,$3F,$97,$8C,$B6,$32,$FE,$25,$B1
	.dc.b	$A8,$C1,$E3,$3F,$41,$C9,$F1,$FD,$5D,$3B,$DE,$5F,$4D,$ED,$F1,$0F
	.dc.b	$53,$B2,$F8,$6F,$D6,$7C,$3B,$BB,$DB,$5A,$9F,$F2,$BD,$7E,$C5,$AF
	.dc.b	$13,$2F,$F9,$73,$E1,$1F,$0C,$F8,$7C,$36,$BF,$D9,$B3,$A9,$C9,$C7
	.dc.b	$E1,$FF,$23,$3A,$6B,$F4,$25,$18,$8F,$DA,$C6,$38,$7C,$5E,$9F,$E1
	.dc.b	$35,$CF,$35,$BF,$E1,$57,$F8,$C7,$EB,$AA,$D4,$23,$4F,$D3,$E5,$78
	.dc.b	$5A,$6C,$72,$BE,$35,$F9,$38,$EB,$F0,$F8,$33,$D1,$63,$C3,$F0,$FC
	.dc.b	$0C,$3F,$81,$A4,$77,$F0,$E4,$F0,$3F,$1F,$C7,$FC,$4D,$BC,$BE,$A8
	.dc.b	$37,$9B,$0F,$A7,$E1,$E5,$70,$3F,$33,$F0,$DF,$0D,$BA,$FC,$48,$F8
	.dc.b	$7C,$5F,$CD,$1D,$4F,$E5,$DC,$F1,$7F,$28,$F8,$77,$93,$F3,$1B,$5D
	.dc.b	$0F,$D3,$FE,$83,$FD,$91,$FF,$52,$B5,$FD,$78,$70,$FE,$C9,$AF,$F3
	.dc.b	$65,$F1,$5F,$98,$6C,$74,$8C,$19,$CE,$47,$81,$AF,$82,$7F,$8A,$BE
	.dc.b	$4F,$07,$A1,$1F,$3F,$C5,$77,$23,$83,$FD,$7F,$F0,$AF,$80,$7C,$2E
	.dc.b	$1B,$3E,$E9,$7C,$0F,$EE,$23,$13,$A7,$F8,$8F,$C7,$7C,$CD,$FF,$D9
	.dc.b	$6D,$8F,$8E,$D7,$7B,$FF,$5D,$F9,$37,$C6,$D2,$5F,$FB,$52,$32,$FF
	.dc.b	$B5,$BE,$9B,$F8,$F3,$7C,$8F,$EC,$D1,$8C,$3E,$AB,$5F,$E4,$BF,$4E
	.dc.b	$F0,$3F,$B6,$36,$FF,$DC,$2D,$9F,$BA,$D7,$EE,$7C,$7F,$F7,$3E,$33
	.dc.b	$49,$5F,$86,$F0,$B4,$47,$07,$E2,$BE,$2B,$91,$1C,$1F,$8E,$F3,$7B
	.dc.b	$11,$C1,$F9,$2F,$03,$C4,$47,$07,$E5,$BE,$37,$CC,$47,$07,$E6,$BE
	.dc.b	$4B,$D4,$47,$07,$E7,$BD,$5F,$71,$1C,$1F,$A2,$F3,$FE,$11,$1C,$1F
	.dc.b	$A6,$F8,$DF,$8C,$47,$07,$EA,$BE,$1B,$E5,$11,$C1,$FA,$EE,$8F,$9C
	.dc.b	$47,$07,$EC,$BA,$7E,$91,$1C,$1F,$6B,$E2,$FF,$AE,$F1,$BE,$4E,$C6
	.dc.b	$1B,$3B,$27,$27,$E0,$27,$F0,$A6,$9F,$EC,$15,$FD,$2A,$7F,$B5,$A9
	.dc.b	$FE,$D7,$C8,$F8,$BA,$63,$E3,$2D,$FF,$0C,$8B,$87,$F6,$BA,$56,$3C
	.dc.b	$6F,$F2,$6F,$C4,$7F,$A1,$B0,$FE,$26,$7F,$78,$71,$7C,$63,$1F,$0F
	.dc.b	$E2,$69,$C0,$F3,$FA,$1F,$13,$D0,$F8,$AD,$8C,$9F,$A2,$7F,$4A,$47
	.dc.b	$06,$AB,$E1,$F9,$FC,$A3,$FC,$1B,$87,$EF,$F1,$F4,$F8,$5E,$7F,$85
	.dc.b	$F4,$32,$B6,$E5,$A7,$2F,$81,$F7,$7C,$17,$BD,$B1,$93,$BC,$65,$A7
	.dc.b	$F8,$7B,$C8,$FE,$5E,$DF,$84,$FF,$0F,$C6,$34,$FF,$0D,$3F,$E4,$79
	.dc.b	$FF,$35,$B9,$8F,$F1,$5B,$5C,$F9,$BF,$E7,$38,$C4,$7F,$A0,$5E,$99
	.dc.b	$FF,$A1,$1B,$5D,$3C,$DF,$0D,$FD,$9F,$8A,$FB,$5F,$09,$6B,$E4,$B7
	.dc.b	$1F,$F1,$76,$E8,$C6,$5F,$84,$7F,$05,$F1,$32,$FE,$8D,$5B,$E3,$36
	.dc.b	$FF,$17,$36,$BD,$FE,$2F,$C4,$F1,$BE,$2B,$63,$E5,$3E,$F2,$3F,$D8
	.dc.b	$30,$FE,$8F,$18,$7E,$5B,$F3,$BC,$6E,$73,$F8,$4F,$83,$FC,$C7,$C2
	.dc.b	$7E,$65,$B1,$93,$F1,$31,$96,$9F,$8C,$F8,$BF,$D0,$1F,$A1,$F8,$5B
	.dc.b	$19,$26,$3F,$22,$32,$FD,$EF,$1B,$F0,$3E,$3F,$E0,$9B,$5F,$87,$6C
	.dc.b	$EE,$3F,$15,$3B,$69,$B2,$C6,$CF,$C5,$C6,$DB,$1B,$69,$BD,$BF,$0D
	.dc.b	$3B,$6C,$ED,$B3,$B8,$FC,$AC,$ED,$A6,$CB,$1B,$3F,$29,$1B,$6C,$6D
	.dc.b	$A6,$CF,$F0,$0C,$6D,$B1,$B6,$9B,$E4,$4E,$DB,$3B,$6C,$EE,$74,$49
	.dc.b	$D3,$63,$51,$82,$38,$25,$7E,$B6,$FE,$BB,$67,$E9,$B6,$7E,$7B,$4F
	.dc.b	$96,$58,$F8,$ED,$8F,$86,$D7,$DE,$DB,$F5,$B6,$FD,$0D,$AF,$3B,$6B
	.dc.b	$CF,$1F,$C3,$2F,$F8,$41,$1F,$D5,$5A,$31,$3F,$E4,$C8,$C4,$EA,$73
	.dc.b	$3F,$C5,$1F,$A1,$E3,$FD,$16,$CF,$D1,$6C,$FD,$76,$DF,$AD,$B3,$E5
	.dc.b	$38,$1F,$CA,$6B,$A3,$F6,$DB,$5F,$81,$6D,$7C,$56,$CF,$39,$FE,$5B
	.dc.b	$18,$7F,$93,$57,$F3,$28,$C4,$7F,$94,$23,$13,$E2,$8E,$CD,$AE,$CD
	.dc.b	$3B,$3F,$CB,$EB,$9D,$FC,$8A,$F0,$D6,$DF,$0B,$ED,$30,$F8,$5C,$3F
	.dc.b	$1D,$F8,$06,$A3,$A5,$77,$8C,$5F,$A2,$B8,$1E,$82,$CF,$A1,$B3,$E8
	.dc.b	$3F,$A7,$DF,$EB,$F8,$1F,$11,$FD,$F7,$E2,$BF,$BF,$36,$BF,$B6,$B6
	.dc.b	$FF,$A2,$B7,$0F,$F0,$3F,$C2,$FF,$81,$4F,$F0,$37,$BB,$6F,$FB,$AC
	.dc.b	$5B,$FE,$9B,$36,$BF,$C1,$FE,$E7,$F8,$37,$D3,$FF,$07,$37,$0F,$EF
	.dc.b	$A9,$FE,$85,$9F,$FA,$E9,$B1,$9C,$3F,$D4,$B3,$96,$C7,$49,$FF,$03
	.dc.b	$FB,$B6,$7F,$AF,$9F,$E1,$DF,$84,$FF,$0F,$36,$7F,$C1,$AD,$36,$FF
	.dc.b	$84,$23,$C1,$8F,$86,$5F,$A0,$BE,$3F,$36,$99,$39,$8C,$57,$18,$7C
	.dc.b	$B7,$0A,$DD,$D7,$AB,$67,$94,$2B,$95,$66,$B8,$7C,$BB,$AB,$87,$CB
	.dc.b	$61,$B9,$78,$8F,$36,$2A,$E1,$1F,$DC,$5B,$1F,$ED,$26,$BC,$0E,$13
	.dc.b	$F0,$DE,$71,$C3,$1F,$8C,$DA,$EF,$93,$E5,$3C,$91,$E3,$BF,$8F,$F2
	.dc.b	$DE,$FF,$94,$F8,$0D,$9D,$4F,$D0,$8C,$3F,$0E,$3A,$C6,$31,$F9,$DE
	.dc.b	$E7,$C9,$3E,$57,$CB,$DA,$E7,$97,$FA,$CE,$B7,$18,$9D,$61,$D4,$FB
	.dc.b	$7E,$A7,$D3,$F8,$0F,$A8,$DA,$E6,$FC,$39,$DC,$7F,$86,$F0,$FF,$3D
	.dc.b	$36,$75,$3F,$8A,$79,$C7,$93,$D7,$B3,$E1,$3F,$1F,$F1,$5F,$83,$F8
	.dc.b	$CF,$C2,$36,$BD,$6D,$9E,$18,$23,$19,$FE,$47,$E2,$7F,$18,$7E,$37
	.dc.b	$E2,$EE,$1F,$12,$2E,$1C,$A6,$D7,$E6,$7D,$4F,$CA,$7A,$DF,$95,$6E
	.dc.b	$1F,$31,$F6,$E4,$69,$23,$E2,$3C,$5F,$A7,$EB,$5C,$7D,$3F,$27,$C8
	.dc.b	$93,$F2,$DE,$BF,$21,$23,$E7,$36,$7F,$2C,$7E,$DB,$C7,$F1,$92,$7E
	.dc.b	$A3,$F4,$BE,$55,$EA,$D2,$4F,$F9,$4E,$32,$DF,$53,$81,$1C,$1B,$F5
	.dc.b	$7A,$11,$C1,$CF,$E2,$FC,$04,$70,$77,$F8,$BF,$21,$1C,$1E,$3F,$87
	.dc.b	$F4,$11,$C1,$E7,$F5,$7D,$84,$70,$7A,$FD,$5F,$80,$47,$07,$BF,$C2
	.dc.b	$F8,$84,$70,$7C,$3F,$87,$F9,$04,$70,$7C,$7F,$87,$F9,$84,$70,$7C
	.dc.b	$BF,$87,$FA,$04,$70,$7F,$86,$3E,$1F,$FD,$44,$47,$07,$D3,$F5,$7E
	.dc.b	$C1,$1C,$1F,$C6,$6B,$F5,$BE,$39,$A4,$8F,$F4,$77,$CA,$7F,$92,$B9
	.dc.b	$7F,$8E,$35,$FE,$4A,$D7,$F9,$E3,$53,$FE,$03,$78,$FA,$DF,$BC,$8C
	.dc.b	$47,$28,$C7,$AD,$2B,$C2,$D9,$FC,$BF,$C9,$BA,$BF,$82,$B6,$BF,$D4
	.dc.b	$8D,$9D,$CE,$8F,$E9,$11,$9E,$FF,$EA,$CF,$EC,$6B,$F9,$4B,$97,$FA
	.dc.b	$0F,$CA,$FF,$69,$9F,$EB,$8D,$8C,$FA,$1F,$D4,$E3,$18,$7F,$5D,$51
	.dc.b	$93,$FA,$C7,$94,$E8,$C5,$FF,$38,$C1,$E2,$7F,$5D,$6C,$67,$C3,$C4
	.dc.b	$E5,$91,$FD,$02,$FF,$A5,$65,$FD,$AA,$7F,$B3,$2F,$5A,$98,$F5,$6D
	.dc.b	$FF,$48,$94,$66,$31,$E9,$D9,$FE,$5C,$1E,$A7,$B3,$86,$CF,$3F,$53
	.dc.b	$C2,$78,$B6,$B1,$DD,$C3,$E1,$B7,$C9,$FF,$81,$7E,$51,$EC,$7F,$6B
	.dc.b	$BF,$5D,$68,$FF,$71,$14,$FF,$72,$56,$1C,$1F,$DE,$BE,$4D,$E7,$3E
	.dc.b	$53,$FD,$D5,$18,$6C,$66,$7F,$92,$CE,$5B,$E7,$FF,$83,$5B,$C9,$CA
	.dc.b	$31,$1F,$E0,$19,$FE,$99,$3D,$BC,$FF,$F0,$5B,$FE,$25,$8F,$BF,$3F
	.dc.b	$DE,$1A,$FF,$5B,$53,$C3,$5F,$57,$2D,$AF,$F1,$57,$2B,$C4,$C4,$23
	.dc.b	$0F,$59,$CE,$0E,$6B,$E6,$CA,$F1,$45,$7C,$5E,$77,$03,$1A,$7A,$5E
	.dc.b	$F8,$BC,$4A,$6E,$BF,$C2,$7C,$7F,$87,$6F,$2F,$9A,$0D,$E7,$C3,$D3
	.dc.b	$C9,$95,$C0,$FF,$0E,$FC,$A7,$C4,$6E,$FF,$02,$3E,$4E,$9F,$E5,$17
	.dc.b	$53,$FE,$05,$73,$C3,$FE,$3C,$F9,$47,$C4,$7C,$86,$D7,$43,$FC,$57
	.dc.b	$FA,$1F,$A5,$FE,$CE,$D7,$FC,$7C,$D7,$FC,$8E,$D7,$FC,$BD,$F1,$BF
	.dc.b	$DA,$1F,$ED,$E1,$C3,$FD,$C9,$AF,$F9,$25,$AF,$4D,$3F,$3F,$3F,$22
	.dc.b	$39,$B5,$F2,$C7,$E9,$25,$A5,$24,$F8,$A3,$2D,$F0,$38,$11,$C1,$E8
	.dc.b	$F9,$7F,$E8,$CD,$F1,$FF,$63,$18,$8F,$F0,$5B,$5F,$F3,$17,$CB,$3E
	.dc.b	$4D,$EA,$FC,$28,$C3,$7E,$4F,$F9,$3D,$7F,$85,$27,$2F,$F8,$5B,$E5
	.dc.b	$BE,$BF,$0F,$CA,$6C,$67,$87,$FC,$8F,$18,$8F,$F3,$97,$CA,$BE,$37
	.dc.b	$FC,$96,$DB,$FA,$6D,$9F,$F1,$1B,$5E,$3E,$47,$02,$4F,$07,$FC,$AB
	.dc.b	$F2,$BF,$AF,$8F,$BA,$D9,$FF,$16,$B5,$FF,$42,$35,$FC,$3F,$8F,$FE
	.dc.b	$5E,$E5,$FE,$8F,$5F,$72,$30,$7F,$98,$B8,$FF,$7D,$1F,$67,$C2,$C1
	.dc.b	$1F,$B9,$EA,$EC,$8E,$0F,$C3,$E1,$DC,$8E,$0F,$C2,$77,$77,$23,$83
	.dc.b	$F1,$1C,$BE,$32,$38,$3F,$3B,$FA,$1F,$FA,$E1,$BF,$0F,$FE,$7A,$8C
	.dc.b	$5F,$F9,$41,$AF,$E7,$FD,$4F,$71,$2B,$F2,$CF,$C1,$F3,$64,$7E,$65
	.dc.b	$B3,$CF,$E1,$FB,$12,$7C,$6F,$C9,$F5,$A4,$7E,$85,$B3,$FB,$1E,$AF
	.dc.b	$9C,$49,$FD,$33,$FE,$26,$D7,$FD,$A1,$93,$E1,$FF,$D9,$71,$86,$D7
	.dc.b	$FB,$16,$38,$34,$97,$FE,$97,$F5,$D3,$D3,$E1,$F8,$CF,$FA,$6F,$07
	.dc.b	$C3,$78,$0F,$0F,$C3,$DC,$7F,$D7,$F3,$A6,$CE,$9A,$6B,$CA,$F4,$FC
	.dc.b	$3F,$A9,$B1,$93,$9C,$65,$A7,$6F,$86,$F8,$3F,$0F,$F0,$9B,$19,$3F
	.dc.b	$02,$FD,$F2,$38,$35,$97,$B7,$E1,$F9,$47,$AB,$D4,$F9,$A7,$CE,$F5
	.dc.b	$6C,$64,$95,$F1,$63,$2F,$C7,$F3,$BE,$AF,$AB,$F5,$9B,$5F,$19,$B3
	.dc.b	$B8,$F9,$32,$7F,$DD,$3E,$07,$FB,$4A,$7F,$D9,$8B,$1D,$46,$32,$BE
	.dc.b	$7E,$21,$5B,$BE,$75,$FC,$18,$8F,$F0,$66,$CF,$38,$F0,$E5,$D1,$44
	.dc.b	$AF,$25,$F7,$EF,$7C,$B6,$EB,$FA,$7B,$5F,$E9,$DF,$2D,$67,$78,$74
	.dc.b	$BE,$57,$93,$55,$67,$47,$8A,$DF,$14,$F9,$6E,$93,$DE,$30,$F0,$9F
	.dc.b	$EC,$2E,$67,$6E,$63,$9C,$61,$F9,$FE,$C7,$A7,$3F,$F0,$B5,$7D,$48
	.dc.b	$C7,$B1,$97,$F9,$89,$23,$F9,$8B,$6B,$9C,$7D,$CF,$63,$83,$E1,$C7
	.dc.b	$FA,$4E,$9F,$65,$B1,$EF,$5F,$2D,$ED,$F3,$A4,$F1,$7E,$04,$63,$05
	.dc.b	$F3,$F9,$6F,$89,$F2,$DF,$15,$B8,$78,$5B,$5C,$23,$C1,$18,$AF,$C6
	.dc.b	$7A,$6F,$8A,$FB,$8F,$FB,$BA,$F8,$1F,$86,$75,$7F,$35,$D5,$72,$AE
	.dc.b	$3F,$19,$C9,$B5,$B6,$DF,$43,$FB,$BB,$8A,$E5,$3D,$6B,$93,$4E,$45
	.dc.b	$AE,$57,$E4,$6C,$F0,$6C,$F0,$69,$C2,$FC,$11,$F9,$1A,$E5,$7F,$12
	.dc.b	$F9,$36,$79,$36,$F9,$47,$8E,$3D,$31,$D9,$A6,$CB,$1D,$E7,$AB,$4E
	.dc.b	$A5,$9E,$A7,$37,$CD,$7E,$88,$FA,$8D,$7E,$5C,$78,$CD,$95,$DB,$5D
	.dc.b	$9D,$4C,$94,$CD,$7A,$63,$9C,$66,$FE,$3C,$79,$E7,$DB,$3F,$87,$BF
	.dc.b	$6B,$67,$DA,$D3,$DB,$3D,$5F,$0B,$5F,$0C,$75,$3C,$31,$9A,$F0,$CF
	.dc.b	$38,$F0,$BF,$E4,$2A,$F0,$B6,$BC,$2D,$3C,$2F,$57,$C2,$D7,$C2,$75
	.dc.b	$7C,$26,$6B,$F2,$11,$FC,$64,$F9,$65,$72,$FC,$83,$17,$FC,$ED,$C1
	.dc.b	$0F,$A2,$6A,$7C,$F3,$FE,$6E,$6C,$FB,$23,$13,$EA,$3D,$24,$3F,$CC
	.dc.b	$7E,$C6,$E7,$83,$6B,$AB,$67,$91,$C2,$3B,$9A,$2B,$AA,$F0,$E1,$F4
	.dc.b	$1F,$A7,$1F,$20,$E4,$6E,$FC,$35,$CA,$B8,$5F,$85,$B1,$93,$E7,$D7
	.dc.b	$F1,$19,$D4,$62,$3A,$CF,$86,$BA,$C7,$79,$F4,$34,$F4,$57,$27,$E3
	.dc.b	$5F,$AA,$7A,$3E,$A2,$4E,$E7,$D9,$1B,$78,$D7,$82,$FC,$57,$E0,$6C
	.dc.b	$F8,$23,$D8,$F8,$2F,$F2,$BA,$BE,$5B,$FA,$F7,$A9,$FD,$7D,$B7,$FD
	.dc.b	$2D,$B7,$FC,$A6,$B8,$4F,$F4,$FA,$FE,$CD,$F2,$D2,$3F,$A9,$4D,$AE
	.dc.b	$13,$71,$FE,$A9,$36,$BF,$91,$78,$DF,$D0,$BD,$1F,$E8,$6D,$8F,$E8
	.dc.b	$0F,$E6,$70,$FE,$88,$D7,$FD,$62,$DD,$7F,$CA,$33,$FE,$D4,$6C,$FF
	.dc.b	$38,$9F,$E9,$FF,$17,$DA,$BF,$B4,$7C,$B7,$F7,$FF,$96,$FF,$00,$36
	.dc.b	$32,$7F,$6D,$8C,$B4,$FE,$E1,$F2,$DF,$E0,$73,$FC,$11,$F2,$D6,$32
	.dc.b	$4C,$7E,$54,$65,$FE,$97,$E9,$7F,$62,$F4,$FF,$B1,$B6,$BF,$B0,$3E
	.dc.b	$D2,$3E,$D6,$CF,$28,$FE,$CD,$F1,$BC,$9F,$F0,$27,$CA,$FF,$86,$FE
	.dc.b	$5B,$FC,$38,$D8,$C9,$FE,$09,$8C,$B4,$FF,$06,$7C,$67,$F8,$8C,$FF
	.dc.b	$12,$7C,$45,$8C,$93,$0F,$C3,$46,$5F,$C4,$7C,$2F,$F8,$AF,$DF,$FF
	.dc.b	$16,$36,$BF,$C3,$0F,$DB,$E5,$34,$91,$FC,$7E,$BF,$C9,$33,$A8,$E2
	.dc.b	$7F,$88,$7E,$23,$AB,$FE,$23,$9F,$F1,$D7,$BF,$FE,$3B,$6C,$64,$FF
	.dc.b	$14,$46,$5A,$7F,$8E,$FE,$59,$EE,$FF,$94,$5B,$19,$F7,$B1,$32,$38
	.dc.b	$3B,$8C,$5F,$FA,$97,$9F,$97,$B9,$FE,$5B,$F7,$11,$8C,$3F,$C9,$51
	.dc.b	$87,$E0,$52,$30,$FC,$36,$BF,$D0,$DA,$FF,$97,$5D,$F6,$FF,$9E,$DB
	.dc.b	$3F,$E6,$1C,$9E,$E7,$8F,$0F,$F6,$1F,$97,$FE,$DD,$25,$7F,$B8,$9F
	.dc.b	$D8,$53,$E4,$29,$27,$CF,$19,$6F,$CB,$70,$23,$83,$FD,$DB,$E6,$FF
	.dc.b	$BA,$C8,$E0,$F8,$3F,$2D,$FE,$45,$D7,$FB,$03,$5F,$F3,$D3,$5F,$F4
	.dc.b	$6B,$53,$FD,$91,$CF,$91,$FC,$26,$31,$1A,$8F,$F7,$E7,$9D,$2B,$FE
	.dc.b	$00,$B6,$7F,$25,$4F,$8A,$E4,$4A,$F3,$BD,$7E,$5A,$47,$A9,$B5,$F9
	.dc.b	$BF,$80,$F8,$CD,$D7,$F8,$E4,$62,$3F,$BF,$35,$F7,$F9,$1E,$92,$5F
	.dc.b	$C6,$7D,$9E,$3C,$8F,$92,$D9,$F0,$6D,$D8,$93,$F8,$EB,$47,$C5,$F8
	.dc.b	$59,$3F,$49,$B5,$F0,$BE,$17,$E5,$12,$7E,$BB,$EA,$F2,$A4,$75,$6C
	.dc.b	$FE,$ED,$7C,$2F,$F3,$92,$4F,$F3,$97,$EB,$78,$F2,$3F,$06,$DA,$FE
	.dc.b	$09,$F3,$0E,$FF,$E8,$8D,$AF,$D4,$B6,$77,$3F,$D2,$7E,$55,$74,$97
	.dc.b	$FB,$08,$CB,$7C,$9E,$04,$70,$7E,$DB,$E5,$7E,$DF,$77,$FB,$A5,$74
	.dc.b	$FF,$B6,$5C,$CF,$FB,$45,$CB,$FE,$C9,$7E,$F6,$5F,$EE,$66,$A6,$98
	.dc.b	$F9,$5F,$E4,$51,$FE,$9D,$6D,$FF,$BE,$57,$87,$FE,$E9,$73,$7F,$EE
	.dc.b	$17,$2F,$FB,$65,$B5,$FE,$A7,$F8,$6F,$CF,$4E,$6B,$FD,$E5,$6F,$83
	.dc.b	$A6,$3B,$ED,$7F,$B8,$E7,$FD,$BF,$87,$FA,$C5,$A2,$3F,$DE,$58,$7F
	.dc.b	$14,$9F,$F7,$D5,$7F,$BF,$D7,$95,$4C,$79,$36,$FF,$DE,$12,$8C,$C6
	.dc.b	$38,$6C,$FF,$B0,$4F,$C1,$63,$FE,$63,$8F,$14,$7B,$F7,$F0,$47,$87
	.dc.b	$C9,$97,$DB,$5B,$7C,$DF,$31,$E7,$5F,$28,$5F,$2A,$F3,$3F,$CF,$15
	.dc.b	$BA,$FE,$7B,$5B,$6B,$C8,$CB,$FD,$0B,$E5,$1F,$06,$F9,$5C,$36,$BF
	.dc.b	$B3,$4E,$A7,$27,$1F,$88,$FF,$7D,$CE,$9A,$FF,$A2,$E5,$18,$8F,$50
	.dc.b	$8C,$EF,$E3,$F8,$8C,$BF,$2A,$75,$CB,$8D,$B8,$DF,$E1,$E7,$E2,$56
	.dc.b	$E1,$7B,$9F,$91,$1B,$7F,$25,$E1,$E5,$73,$23,$13,$BB,$E9,$CD,$8E
	.dc.b	$99,$87,$83,$2F,$1C,$E2,$15,$E3,$F2,$39,$99,$71,$1C,$4F,$B9,$93
	.dc.b	$C1,$F2,$36,$34,$7C,$7A,$F4,$71,$FF,$15,$8C,$BF,$E1,$E6,$F2,$EB
	.dc.b	$9B,$6F,$4F,$72,$4E,$8E,$BF,$2B,$F2,$1B,$C5,$F1,$63,$E1,$6D,$F6
	.dc.b	$1D,$4F,$BA,$31,$3F,$2F,$E5,$5F,$07,$4E,$51,$F3,$FE,$31,$D9,$F9
	.dc.b	$66,$E1,$FE,$14,$6B,$FE,$35,$6B,$FE,$D5,$F9,$0F,$80,$FF,$8E,$5A
	.dc.b	$E7,$5F,$D0,$C6,$23,$F1,$F2,$F6,$BC,$04,$8F,$03,$C3,$69,$1F,$E2
	.dc.b	$6A,$F9,$8F,$78,$FD,$6F,$C5,$37,$A4,$9F,$F1,$6C,$65,$FF,$17,$FC
	.dc.b	$A7,$F8,$91,$BE,$AF,$ED,$A3,$11,$FE,$53,$6B,$FE,$F3,$F9,$57,$99
	.dc.b	$FB,$C6,$C7,$F8,$E9,$AF,$F9,$6D,$AF,$FB,$EF,$E5,$5F,$26,$F3,$BF
	.dc.b	$C8,$31,$86,$E1,$FC,$F2,$BF,$CC,$93,$97,$E9,$79,$DF,$EA,$5E,$0F
	.dc.b	$ED,$BC,$DF,$5E,$30,$7F,$AA,$7E,$51,$F0,$D4,$93,$FE,$50,$8C,$B7
	.dc.b	$CA,$E0,$47,$07,$E0,$F4,$F0,$11,$C1,$F5,$7C,$9F,$F5,$C7,$27,$F4
	.dc.b	$EC,$FF,$0D,$18,$3F,$D7,$9C,$5F,$BD,$AF,$AB,$E2,$60,$8F,$E4,$77
	.dc.b	$F7,$11,$C1,$F8,$3B,$EE,$47,$07,$E0,$3C,$0E,$E4,$70,$7E,$77,$E3
	.dc.b	$BE,$51,$1C,$1F,$E2,$7F,$7B,$FC,$6E,$47,$07,$E3,$3D,$1F,$59,$1C
	.dc.b	$1F,$A9,$F8,$FF,$AC,$47,$07,$F8,$CF,$E0,$3F,$C8,$64,$70,$7E,$63
	.dc.b	$8F,$E3,$91,$C1,$FA,$3F,$8D,$FE,$BB,$E3,$FF,$35,$E6,$FE,$E1,$3F
	.dc.b	$C5,$B8,$95,$A8,$FD,$89,$76,$FE,$B5,$3A,$8E,$84,$8F,$DE,$4F,$DC
	.dc.b	$38,$29,$E0,$FF,$11,$39,$FC,$7E,$A3,$38,$7F,$49,$5F,$1D,$93,$5E
	.dc.b	$25,$C7,$D1,$17,$0E,$87,$79,$C5,$F3,$E9,$ED,$F1,$DD,$DB,$87,$C1
	.dc.b	$6D,$F2,$C7,$F3,$ED,$AF,$37,$C6,$79,$FE,$36,$E1,$F2,$5B,$5F,$C7
	.dc.b	$67,$F3,$A4,$9F,$E2,$8E,$B1,$F7,$7C,$3D,$9F,$ED,$B8,$BE,$37,$E2
	.dc.b	$1F,$17,$E3,$6C,$FD,$78,$B8,$F9,$E6,$E1,$F2,$7E,$37,$E6,$7C,$6F
	.dc.b	$CD,$6E,$1F,$7D,$B7,$EF,$6D,$7F,$B1,$7E,$0B,$EC,$FC,$13,$E2,$FF
	.dc.b	$DA,$D1,$86,$E1,$FE,$D4,$F8,$B4,$FF,$80,$71,$FF,$68,$35,$FF,$11
	.dc.b	$46,$A7,$84,$F2,$9F,$B3,$7F,$ED,$29,$FF,$70,$CF,$8B,$E4,$E4,$7F
	.dc.b	$6C,$97,$CA,$62,$F8,$D5,$CB,$FC,$25,$5A,$BE,$13,$EC,$F5,$FD,$BD
	.dc.b	$56,$7A,$37,$0F,$1B,$70,$E6,$FC,$4F,$92,$F8,$BF,$29,$6F,$C8,$DB
	.dc.b	$FF,$5E,$4F,$FA,$E8,$93,$FE,$AD,$75,$7E,$6F,$94,$B3,$FE,$2E,$CF
	.dc.b	$FC,$62,$FB,$FD,$FF,$A7,$F0,$3F,$51,$91,$93,$E2,$46,$31,$F9,$7E
	.dc.b	$C7,$DA,$F8,$0F,$B6,$DA,$F9,$4D,$9D,$C7,$CD,$9F,$DC,$FC,$A3,$4B
	.dc.b	$9E,$1B,$5F,$E1,$E9,$D3,$67,$4D,$35,$3C,$2F,$F6,$F5,$C1,$B5,$C1
	.dc.b	$B5,$C0,$E4,$F0,$AE,$5A,$7F,$93,$EF,$93,$6F,$93,$6F,$95,$F4,$39
	.dc.b	$3D,$23,$95,$F4,$8F,$F2,$FD,$F4,$6D,$F4,$69,$D2,$BB,$5F,$43,$B3
	.dc.b	$D2,$3B,$1D,$2F,$B4,$FF,$A1,$6F,$B3,$6F,$B3,$6B,$B1,$E0,$7B,$5F
	.dc.b	$80,$EC,$F8,$23,$B1,$E0,$7B,$5F,$81,$5F,$81,$B7,$E0,$6D,$78,$1F
	.dc.b	$11,$E0,$7C,$57,$E0,$3C,$4F,$82,$3C,$47,$81,$F1,$3E,$0B,$F1,$5F
	.dc.b	$84,$5F,$85,$B5,$E1,$AF,$19,$E1,$3C,$77,$E1,$8F,$19,$E1,$7C,$71
	.dc.b	$E1,$3C,$67,$86,$3C,$73,$E1,$AF,$01,$23,$C1,$1E,$29,$F0,$1E,$2A
	.dc.b	$F9,$EF,$FA,$90,$E8,$52,$99,$8F,$CF,$36,$FF,$92,$3A,$3E,$41,$9A
	.dc.b	$E4,$F4,$AF,$73,$F8,$7C,$BE,$35,$7C,$76,$D7,$A1,$B3,$E8,$BF,$F6
	.dc.b	$3C,$FB,$2B,$83,$4E,$05,$9E,$15,$CA,$FC,$31,$F8,$88,$C4,$FC,$F3
	.dc.b	$F4,$EE,$23,$E8,$4F,$ED,$63,$11,$EB,$9D,$9D,$E7,$AD,$F9,$A7,$D1
	.dc.b	$7E,$66,$D7,$99,$A7,$99,$E4,$78,$DF,$8C,$E1,$F1,$BE,$07,$8D,$78
	.dc.b	$2F,$CB,$7F,$93,$6D,$FF,$22,$6D,$FF,$1F,$6D,$FF,$1A,$6D,$FB,$E7
	.dc.b	$C1,$3E,$F3,$C6,$7E,$2B,$CC,$78,$DF,$E2,$66,$CF,$FB,$21,$79,$39
	.dc.b	$99,$39,$78,$46,$2B,$FC,$43,$DA,$F1,$3A,$5F,$F8,$CA,$30,$F5,$31
	.dc.b	$1A,$7F,$DA,$F1,$A9,$FF,$15,$E1,$FE,$46,$7F,$9B,$FC,$07,$F4,$4F
	.dc.b	$80,$FE,$8A,$DB,$FE,$84,$FE,$73,$63,$49,$1F,$9C,$8C,$BC,$0F,$E9
	.dc.b	$5F,$01,$D1,$F1,$F8,$DE,$93,$D5,$E4,$58,$C9,$31,$F8,$31,$97,$E1
	.dc.b	$F8,$DF,$07,$C8,$F8,$4D,$AF,$7B,$67,$71,$F0,$A7,$FC,$B3,$EA,$BE
	.dc.b	$3E,$FC,$A6,$1B,$5F,$EF,$E9,$D3,$FB,$86,$DF,$F0,$36,$BF,$C1,$DA
	.dc.b	$88,$F0,$E4,$4E,$67,$FD,$85,$5E,$4F,$6F,$C5,$3F,$63,$CB,$FB,$DE
	.dc.b	$67,$DF,$6C,$64,$FB,$B1,$96,$9F,$80,$F3,$3F,$0C,$7E,$1F,$96,$C6
	.dc.b	$49,$8F,$E3,$A3,$2F,$F9,$C3,$3F,$F1,$E3,$7D,$3F,$93,$E3,$7C,$A6
	.dc.b	$DF,$C5,$8F,$E4,$35,$C2,$BE,$24,$60,$DB,$D7,$E4,$AE,$DF,$12,$3E
	.dc.b	$46,$3F,$E6,$7C,$3F,$33,$E6,$7E,$7F,$CD,$FD,$03,$63,$27,$E7,$A3
	.dc.b	$2F,$E8,$3C,$CF,$D3,$1F,$A7,$F3,$6C,$64,$98,$7E,$BA,$32,$FC,$1F
	.dc.b	$80,$FC,$3F,$A5,$F8,$86,$D7,$E1,$5F,$79,$1F,$C3,$36,$79,$47,$E2
	.dc.b	$FD,$FE,$4F,$FA,$D7,$E4,$FF,$DC,$E4,$7F,$A2,$FA,$48,$FE,$97,$E5
	.dc.b	$BC,$1A,$4A,$FE,$97,$19,$7F,$A6,$FC,$9B,$E4,$3F,$A9,$B6,$33,$18
	.dc.b	$3F,$DE,$FF,$29,$23,$C2,$DC,$FF,$B0,$5F,$EF,$1B,$E2,$79,$09,$5F
	.dc.b	$D7,$3E,$51,$DF,$FD,$89,$B1,$98,$C1,$FD,$9B,$57,$C4,$52,$4F,$C2
	.dc.b	$7E,$0F,$8B,$23,$DE,$DC,$7F,$BA,$60,$F9,$4F,$01,$2B,$C8,$F7,$F8
	.dc.b	$E9,$1F,$39,$B5,$EE,$DB,$E7,$92,$7E,$6B,$EB,$EE,$91,$EF,$6C,$FF
	.dc.b	$81,$73,$7C,$0F,$B4,$95,$F7,$1F,$B7,$F2,$72,$3E,$CB,$6B,$D7,$EF
	.dc.b	$FC,$92,$4F,$CD,$7F,$C1,$BE,$D9,$A4,$8F,$F5,$C3,$F8,$8F,$0A,$47
	.dc.b	$E1,$9B,$5F,$C6,$7D,$DF,$DB,$92,$7F,$8F,$3F,$E1,$AF,$93,$FF,$73
	.dc.b	$CF,$F8,$8D,$AF,$F1,$96,$BF,$C7,$DA,$97,$C5,$73,$25,$7F,$88,$A3
	.dc.b	$2F,$F8,$93,$E4,$DD,$DF,$E2,$A6,$C6,$63,$07,$F8,$BB,$E4,$FF,$29
	.dc.b	$87,$F8,$DD,$AF,$F2,$D6,$BF,$CC,$DA,$9F,$D2,$2F,$17,$F8,$93,$9E
	.dc.b	$9F,$E2,$91,$89,$DB,$63,$37,$FC,$EF,$E4,$DF,$1B,$8C,$3F,$A1,$46
	.dc.b	$27,$FA,$25,$F6,$87,$C3,$EE,$91,$B7,$C9,$5D,$67,$53,$C2,$39,$9F
	.dc.b	$D2,$FE,$07,$F4,$4E,$2B,$C3,$5D,$6D,$EF,$53,$1E,$15,$BE,$17,$F0
	.dc.b	$E3,$31,$8C,$1E,$5E,$2B,$CA,$7F,$5B,$EB,$F5,$26,$F4,$61,$B7,$DA
	.dc.b	$7F,$4E,$DE,$3C,$C7,$43,$FB,$53,$D0,$F8,$D8,$3C,$6C,$57,$3D,$FC
	.dc.b	$27,$8B,$6B,$1D,$DC,$3E,$1C,$1F,$25,$FD,$DF,$E4,$FF,$89,$C7,$F7
	.dc.b	$5A,$F5,$BA,$3F,$CF,$25,$3F,$CF,$73,$87,$07,$F7,$DF,$92,$7B,$CF
	.dc.b	$93,$FF,$40,$C6,$1B,$19,$9F,$F4,$74,$E5,$BF,$07,$F9,$56,$F2,$72
	.dc.b	$8C,$47,$FA,$5E,$51,$88,$F1,$7C,$1F,$F8,$3E,$BF,$3B,$5F,$82,$9F
	.dc.b	$BE,$D7,$E8,$29,$E1,$A7,$AF,$96,$D7,$F8,$77,$81,$E1,$62,$11,$87
	.dc.b	$AC,$E7,$0F,$EB,$51,$E0,$E8,$C1,$9E,$7B,$1E,$4F,$1B,$81,$87,$CA
	.dc.b	$F7,$C3,$F9,$14,$7E,$57,$C0,$DB,$F9,$AF,$1F,$F8,$8B,$79,$7C,$D1
	.dc.b	$FB,$68,$D7,$3E,$27,$A7,$AF,$27,$81,$FA,$4F,$93,$F8,$8D,$DB,$E0
	.dc.b	$47,$C9,$E2,$FE,$12,$EA,$7F,$34,$E7,$87,$FC,$81,$F2,$6F,$7B,$E4
	.dc.b	$36,$BA,$1F,$B2,$FD,$19,$F2,$BF,$90,$6C,$7F,$B2,$9A,$FF,$B5,$5A
	.dc.b	$FF,$B6,$57,$CC,$71,$24,$71,$7A,$73,$48,$E8,$D7,$A7,$B5,$E1,$23
	.dc.b	$E1,$7C,$9E,$14,$8F,$23,$5F,$F7,$47,$91,$FE,$88,$7F,$DC,$2D,$73
	.dc.b	$D1,$FA,$C8,$C4,$7F,$BC,$7E,$75,$7F,$B0,$6C,$7F,$BA,$9A,$FF,$BD
	.dc.b	$5A,$FC,$1E,$6F,$90,$47,$E4,$3F,$BA,$F9,$47,$C9,$BC,$CF,$F7,$AC
	.dc.b	$61,$BF,$27,$FC,$DA,$FF,$CF,$07,$F9,$F6,$74,$E0,$7C,$C7,$DD,$E4
	.dc.b	$FF,$7F,$B6,$34,$D7,$9B,$5F,$F4,$27,$CD,$3E,$3F,$FA,$9C,$61,$AF
	.dc.b	$FA,$36,$35,$5E,$18,$46,$1F,$C2,$7C,$C7,$05,$C8,$8C,$36,$BF,$19
	.dc.b	$18,$F9,$A9,$1F,$8F,$6D,$7E,$3F,$D0,$E2,$49,$FC,$A3,$D3,$CC,$91
	.dc.b	$F9,$76,$CF,$E5,$FD,$EF,$09,$27,$F3,$8F,$93,$92,$47,$E7,$DB,$3F
	.dc.b	$9F,$E1,$F4,$92,$7F,$48,$FB,$38,$24,$7E,$9D,$B3,$FA,$7F,$8E,$F8
	.dc.b	$24,$9F,$D6,$3F,$13,$E3,$E4,$7E,$BD,$B3,$FA,$F8,$F9,$24,$3F,$68
	.dc.b	$7C,$FF,$D1,$BF,$DE,$11,$F8,$86,$BF,$EC,$16,$BF,$EC,$76,$AF,$99
	.dc.b	$7C,$4F,$F8,$15,$7C,$6D,$8C,$36,$FF,$05,$3F,$E3,$22,$4F,$E0,$5E
	.dc.b	$71,$FD,$F9,$B5,$FD,$BD,$7C,$5F,$38,$FE,$E5,$87,$E6,$7C,$4F,$F1
	.dc.b	$2A,$F8,$EF,$F0,$C3,$9C,$3D,$91,$8F,$13,$2D,$8F,$F0,$6C,$F1,$6E
	.dc.b	$7C,$8E,$7F,$1D,$63,$24,$C7,$C7,$19,$7D,$1F,$11,$F1,$7E,$2B,$E3
	.dc.b	$36,$BD,$CF,$C1,$23,$F1,$9B,$3C,$A3,$E2,$7C,$57,$27,$CF,$AF,$BB
	.dc.b	$7F,$7B,$63,$27,$B2,$32,$D3,$DF,$F1,$BF,$18,$F8,$FF,$1B,$63,$24
	.dc.b	$C3,$E7,$C6,$5F,$A9,$EE,$FE,$07,$DD,$FC,$13,$6B,$ED,$3F,$74,$8F
	.dc.b	$E0,$9B,$3C,$A3,$F0,$1E,$8F,$27,$E9,$EB,$F6,$8F,$B7,$F1,$B6,$32
	.dc.b	$4A,$FC,$1C,$65,$FC,$2E,$BF,$82,$DB,$F0,$6D,$AF,$BE,$D9,$DC,$7E
	.dc.b	$0A,$7F,$CC,$1F,$1E,$F9,$8B,$F3,$38,$6D,$7F,$95,$27,$4F,$F1,$F6
	.dc.b	$DF,$F3,$16,$BF,$CD,$9A,$88,$F0,$E7,$FF,$3F,$B6,$75,$28,$7C,$CF
	.dc.b	$8E,$7F,$33,$F1,$FF,$A0,$3F,$43,$F2,$36,$32,$4B,$FD,$4C,$65,$FD
	.dc.b	$5F,$0F,$EA,$3D,$9F,$D4,$B6,$BF,$4A,$D9,$DC,$7E,$A2,$7E,$3F,$B5
	.dc.b	$F2,$3E,$5E,$CF,$C5,$6D,$7F,$9A,$73,$46,$34,$F7,$B6,$BD,$F3,$A7
	.dc.b	$E7,$5F,$56,$9D,$4B,$7D,$6B,$93,$F3,$1A,$FC,$96,$D7,$58,$F2,$57
	.dc.b	$56,$D7,$56,$9D,$5E,$F7,$E8,$69,$E8,$BE,$A4,$73,$3E,$36,$BE,$36
	.dc.b	$DF,$8E,$BF,$03,$7E,$36,$CF,$8D,$B5,$E3,$3C,$AF,$8D,$B1,$E3,$3A
	.dc.b	$DF,$8E,$32,$78,$CA,$78,$CB,$7E,$39,$FA,$77,$E3,$69,$E3,$2C,$F8
	.dc.b	$EF,$CA,$78,$CB,$1E,$38,$EB,$7E,$39,$CC,$78,$DA,$F8,$DB,$7E,$38
	.dc.b	$FD,$3D,$F8,$DB,$7E,$36,$9E,$37,$CB,$5E,$36,$9E,$37,$AD,$F8,$E3
	.dc.b	$D7,$3F,$B7,$AF,$F0,$B5,$FF,$A2,$A7,$F7,$D1,$82,$38,$24,$FF,$01
	.dc.b	$8F,$C7,$F9,$7F,$90,$6C,$FE,$01,$B3,$F4,$1B,$87,$C0,$6C,$FA,$27
	.dc.b	$DB,$3F,$12,$BA,$4F,$E2,$9B,$3F,$69,$B3,$B9,$F0,$1B,$2C,$ED,$B1
	.dc.b	$B8,$F2,$46,$DA,$6C,$B1,$B3,$D1,$1B,$6C,$6D,$A6,$E3,$D9,$1F,$C7
	.dc.b	$6B,$DE,$FB,$A3,$37,$F4,$63,$F8,$55,$FB,$27,$2D,$8E,$07,$F2,$C7
	.dc.b	$0F,$58,$FE,$71,$18,$9E,$97,$FC,$C2,$3F,$83,$F6,$39,$7F,$C1,$67
	.dc.b	$F8,$33,$9A,$E1,$FE,$08,$8B,$8F,$F8,$52,$6E,$1F,$E1,$4F,$84,$FF
	.dc.b	$09,$FB,$9F,$E1,$46,$E1,$FE,$10,$6D,$FF,$87,$1B,$49,$FE,$83,$FA
	.dc.b	$37,$F4,$3A,$FF,$1A,$19,$13,$54,$6D,$65,$FD,$3E,$BF,$A8,$56,$EB
	.dc.b	$FA,$9D,$7F,$53,$6B,$CB,$93,$FC,$ED,$18,$3A,$13,$97,$C5,$5C,$E6
	.dc.b	$DF,$A3,$E7,$2E,$1D,$5A,$F9,$B9,$AE,$29,$A4,$AF,$39,$F0,$A7,$15
	.dc.b	$F0,$23,$E2,$FC,$D7,$2F,$9B,$E6,$D8,$E3,$3E,$D8,$C3,$C4,$C4,$69
	.dc.b	$F9,$33,$E6,$F9,$8F,$F6,$BC,$E7,$1F,$E2,$54,$9C,$BC,$CF,$A3,$CB
	.dc.b	$C7,$DC,$FF,$4F,$4E,$A7,$FD,$6E,$D7,$FB,$EB,$5F,$F0,$23,$D3,$5F
	.dc.b	$B8,$DC,$3F,$07,$CD,$D5,$F2,$7C,$DF,$CA,$3E,$5F,$35,$8C,$93,$0F
	.dc.b	$36,$7F,$EB,$3C,$7F,$05,$7F,$4B,$E2,$3E,$9F,$CD,$DC,$3E,$03,$70
	.dc.b	$E2,$DB,$FC,$53,$F6,$BE,$0F,$ED,$FC,$4D,$BF,$90,$DB,$FF,$6A,$64
	.dc.b	$AF,$C6,$92,$BF,$D9,$11,$97,$E8,$47,$E1,$BE,$27,$96,$3F,$99,$F8
	.dc.b	$0F,$C4,$9F,$8A,$F8,$8B,$7F,$5E,$2E,$3F,$02,$6D,$7E,$87,$E5,$3F
	.dc.b	$23,$F1,$3F,$92,$6E,$1F,$7D,$B7,$F2,$1B,$5F,$AF,$E4,$7C,$BF,$E6
	.dc.b	$BE,$53,$F3,$B1,$8C,$3F,$11,$18,$9F,$F7,$04,$E7,$0F,$F7,$B3,$5F
	.dc.b	$F7,$DB,$5F,$F2,$0B,$BB,$F0,$4F,$EA,$3E,$0B,$A6,$1F,$99,$A7,$CD
	.dc.b	$76,$24,$7F,$BC,$E3,$2D,$F8,$CF,$19,$1C,$1B,$F9,$7F,$39,$1C,$1C
	.dc.b	$FE,$2F,$D6,$47,$07,$7E,$BF,$79,$1C,$1E,$3F,$8B,$F8,$64,$70,$79
	.dc.b	$FE,$2B,$E3,$91,$C1,$EB,$F8,$BF,$96,$47,$07,$BF,$E2,$BE,$79,$1C
	.dc.b	$1F,$0F,$E1,$BE,$99,$1C,$1F,$1F,$E1,$BE,$B9,$1C,$1F,$2F,$E1,$BE
	.dc.b	$D9,$1C,$1F,$E7,$8E,$3F,$F3,$CF,$0F,$D0,$70,$1C,$3B,$6B,$C3,$87
	.dc.b	$F9,$E3,$CB,$E1,$AB,$FE,$89,$94,$62,$3A,$36,$33,$5E,$09,$7C,$B3
	.dc.b	$4E,$0D,$9D,$36,$7C,$ED,$7D,$9F,$30,$D7,$C0,$DA,$EC,$D9,$F7,$B5
	.dc.b	$E3,$AF,$02,$4F,$07,$9E,$B2,$39,$B6,$7E,$4F,$CB,$73,$7F,$D4,$AD
	.dc.b	$7F,$D5,$8D,$4D,$15,$F1,$DB,$5F,$15,$B3,$DA,$75,$52,$B4,$D9,$FB
	.dc.b	$7F,$2C,$EC,$FA,$4D,$AF,$A0,$D9,$DC,$E8,$FE,$C7,$19,$E9,$FE,$F4
	.dc.b	$FF,$21,$9F,$E9,$8E,$6B,$FD,$87,$F2,$CF,$47,$FB,$5B,$63,$FA,$D2
	.dc.b	$F4,$73,$3A,$64,$FF,$5F,$F2,$5D,$58,$C7,$FA,$D6,$5E,$13,$FC,$18
	.dc.b	$5A,$FF,$69,$BE,$38,$FF,$05,$F8,$7F,$DF,$23,$11,$FE,$14,$D3,$FB
	.dc.b	$FE,$1F,$8B,$9E,$71,$FE,$02,$97,$AD,$4C,$7A,$77,$1F,$EE,$92,$8C
	.dc.b	$C6,$3D,$4B,$3F,$DC,$4F,$F1,$C4,$64,$F7,$74,$77,$9F,$06,$D2,$7A
	.dc.b	$ED,$6F,$93,$E5,$3C,$CB,$E5,$7F,$DF,$1F,$2C,$F6,$3F,$7D,$5A,$9F
	.dc.b	$D7,$57,$FB,$F5,$AF,$13,$2F,$EC,$FE,$55,$E7,$3E,$5B,$0D,$AF,$E0
	.dc.b	$F3,$A9,$C9,$C7,$CE,$FF,$20,$CE,$9A,$FF,$11,$94,$62,$04,$8C,$F1
	.dc.b	$78,$FC,$DF,$BB,$AE,$A9,$5F,$DD,$DF,$F1,$33,$5F,$F1,$02,$9E,$31
	.dc.b	$D7,$96,$AE,$57,$89,$86,$C7,$49,$DD,$FE,$AE,$3B,$F9,$B8,$33,$CF
	.dc.b	$63,$C7,$E6,$F0,$30,$F9,$1E,$DE,$04,$8D,$9F,$E3,$FE,$3F,$F3,$F3
	.dc.b	$79,$7C,$F1,$FA,$D8,$D7,$36,$27,$D1,$E4,$C9,$E0,$7F,$92,$7E,$5B
	.dc.b	$E1,$37,$5F,$83,$1F,$1F,$8B,$FC,$EC,$EA,$7F,$C9,$0E,$78,$BE,$87
	.dc.b	$CB,$3E,$2B,$E4,$B6,$BA,$1F,$E5,$7E,$37,$B1,$F9,$06,$C7,$F3,$F6
	.dc.b	$BF,$D4,$1A,$FF,$55,$F9,$27,$9D,$F9,$86,$C7,$F4,$76,$BF,$D6,$1A
	.dc.b	$FF,$5D,$9F,$F3,$C3,$DF,$DA,$F1,$11,$E5,$1D,$48,$E0,$ED,$E7,$78
	.dc.b	$48,$E0,$FE,$B1,$F2,$2F,$6B,$F5,$AD,$8C,$C6,$0C,$F4,$48,$E2,$D7
	.dc.b	$FB,$8F,$CB,$2F,$F6,$ED,$8F,$EC,$CD,$77,$1F,$BC,$F9,$CF,$E9,$31
	.dc.b	$FD,$AD,$AF,$F7,$C6,$BF,$E0,$16,$BF,$AB,$F9,$77,$CA,$7F,$80,$63
	.dc.b	$FC,$07,$18,$F4,$BF,$BB,$2F,$94,$FE,$75,$3A,$AE,$0F,$F8,$37,$E5
	.dc.b	$FF,$0D,$CB,$FD,$F5,$B1,$FE,$0E,$6B,$BE,$5F,$F0,$37,$07,$CB,$AF
	.dc.b	$C3,$6F,$82,$3F,$81,$F4,$3F,$C1,$9C,$7F,$D9,$30,$FC,$44,$60,$CF
	.dc.b	$A1,$2B,$F1,$0D,$9F,$C8,$7C,$0F,$80,$93,$F9,$57,$AF,$99,$23,$F2
	.dc.b	$0D,$9F,$C8,$7C,$B7,$88,$93,$F9,$57,$D7,$ED,$48,$FD,$03,$67,$F3
	.dc.b	$1B,$FA,$89,$3F,$9D,$7D,$BC,$12,$3F,$40,$D9,$FD,$47,$8F,$F2,$09
	.dc.b	$3F,$AD,$7E,$2F,$CA,$C8,$FD,$43,$67,$F5,$11,$F2,$88,$7E,$B4,$FF
	.dc.b	$41,$F4,$FF,$B9,$38,$7F,$D6,$91,$FE,$80,$8C,$7C,$17,$FA,$31,$DC
	.dc.b	$7F,$9E,$DB,$5F,$E8,$A7,$51,$FC,$17,$E0,$AC,$FF,$B5,$6F,$FA,$7C
	.dc.b	$FF,$09,$F8,$2F,$07,$9F,$E1,$6C,$64,$FE,$1D,$19,$7F,$88,$79,$BE
	.dc.b	$73,$D1,$E7,$58,$C9,$30,$F0,$46,$5F,$17,$7F,$BF,$C6,$F8,$0D,$AF
	.dc.b	$E3,$8F,$E5,$88,$FF,$1B,$6C,$F2,$8F,$E4,$5E,$6F,$27,$F9,$27,$C0
	.dc.b	$FC,$BF,$3B,$E6,$36,$32,$7F,$29,$8C,$B4,$FE,$59,$E0,$7D,$33,$EA
	.dc.b	$7C,$0D,$8C,$93,$0F,$99,$19,$5E,$52,$7F,$B5,$7C,$AB,$E4,$71,$5A
	.dc.b	$9C,$FC,$A6,$2F,$32,$BE,$15,$FD,$E2,$73,$0A,$E1,$3D,$AF,$FD,$05
	.dc.b	$3C,$2F,$FD,$1E,$D7,$FD,$28,$D7,$84,$F5,$2B,$E5,$3F,$C0,$F0,$F9
	.dc.b	$3E,$F1,$8A,$AB,$E2,$BF,$88,$E3,$EA,$5F,$2B,$A7,$38,$FF,$83,$5C
	.dc.b	$BE,$1A,$FF,$0A,$C6,$9B,$3C,$7C,$6E,$4F,$3C,$E5,$F1,$AF,$17,$CA
	.dc.b	$E1,$F5,$1E,$59,$C6,$3E,$59,$B5,$F5,$BE,$43,$D7,$AB,$E4,$EC,$61
	.dc.b	$B3,$FE,$AB,$94,$FD,$02,$46,$5D,$E3,$E2,$97,$C9,$63,$1F,$8D,$BF
	.dc.b	$C5,$DF,$E3,$37,$1F,$66,$FF,$45,$B3,$A6,$9A,$F9,$5F,$9C,$7C,$FF
	.dc.b	$95,$B1,$92,$67,$F1,$2D,$3A,$8F,$37,$C0,$FC,$2F,$82,$F8,$6D,$BF
	.dc.b	$9D,$AF,$F9,$CD,$B3,$A6,$9A,$9E,$15,$D2,$32,$D3,$E2,$4F,$16,$C7
	.dc.b	$13,$7A,$7E,$82,$B9,$B6,$32,$7E,$BA,$32,$D3,$33,$F9,$A8,$CB,$63
	.dc.b	$2F,$EB,$E3,$2D,$8C,$DF,$BE,$3C,$A7,$A6,$BF,$67,$7E,$96,$D7,$A6
	.dc.b	$77,$1E,$76,$C7,$95,$A7,$8C,$A7,$84,$B5,$E1,$9E,$B1,$E1,$9F,$E1
	.dc.b	$B5,$E1,$69,$E1,$9D,$9E,$12,$BE,$16,$CF,$85,$AF,$85,$B5,$E1,$9E
	.dc.b	$B3,$E1,$3F,$82,$D7,$85,$B1,$E1,$37,$5E,$16,$9E,$12,$C7,$85,$AF
	.dc.b	$85,$B5,$E1,$8E,$A7,$84,$FE,$4D,$5E,$16,$CF,$85,$DD,$78,$5B,$1E
	.dc.b	$16,$9E,$12,$9F,$44,$FE,$4F,$7F,$CB,$09,$1F,$C0,$5B,$1F,$CB,$9A
	.dc.b	$F0,$8F,$13,$FC,$EF,$DB,$F3,$61,$EB,$8F,$71,$FC,$BA,$33,$5A,$6D
	.dc.b	$72,$AF,$F6,$ED,$7C,$17,$91,$27,$71,$96,$DF,$C1,$8F,$15,$73,$BE
	.dc.b	$CD,$3B,$57,$07,$E3,$4F,$03,$C1,$3C,$1B,$1C,$0E,$4F,$85,$B5,$E1
	.dc.b	$BF,$4B,$6B,$93,$67,$95,$74,$39,$16,$39,$5F,$96,$3E,$C1,$E5,$8E
	.dc.b	$4F,$06,$9C,$27,$C6,$7B,$2F,$C6,$DB,$F1,$B4,$F1,$CF,$C9,$CB,$FA
	.dc.b	$24,$7F,$34,$E0,$FE,$B9,$F2,$BF,$D7,$59,$19,$3F,$9C,$C6,$31,$FE
	.dc.b	$61,$C1,$FD,$8F,$87,$FB,$23,$6B,$F9,$FB,$67,$71,$FD,$0A,$7F,$B7
	.dc.b	$FC,$9B,$E4,$AF,$C9,$E1,$B5,$FD,$8E,$74,$FF,$BE,$EF,$FD,$F9,$18
	.dc.b	$70,$1C,$3B,$6B,$C2,$7B,$4F,$5F,$6E,$CF,$F4,$BF,$95,$FE,$FC,$7F
	.dc.b	$7F,$F9,$5B,$5F,$D6,$A2,$CF,$F5,$19,$BB,$FF,$81,$F8,$FF,$C0,$DF
	.dc.b	$05,$FE,$07,$6E,$1F,$D8,$9B,$7F,$D6,$5B,$5F,$E2,$8C,$DE,$6F,$C6
	.dc.b	$F4,$3E,$3B,$70,$F8,$8F,$F8,$BB,$0F,$86,$D4,$7C,$37,$0D,$8F,$8E
	.dc.b	$D3,$FC,$1B,$5E,$08,$FA,$1E,$6F,$81,$FE,$F3,$CF,$FE,$21,$3F,$C4
	.dc.b	$5C,$D6,$32,$4B,$FE,$E9,$19,$7F,$BB,$7C,$3F,$F8,$A7,$E0,$FF,$C5
	.dc.b	$4D,$AF,$F0,$2B,$67,$71,$FE,$08,$9F,$F1,$AF,$CA,$7D,$FD,$7F,$C6
	.dc.b	$95,$8E,$3F,$F1,$A3,$6F,$DB,$5F,$80,$8F,$86,$D4,$D3,$A4,$7C,$88
	.dc.b	$CC,$F8,$39,$3A,$47,$F8,$CA,$9F,$01,$E4,$3C,$BC,$B7,$0F,$F1,$D4
	.dc.b	$E9,$B3,$A6,$9A,$F5,$FD,$BE,$DF,$B9,$B1,$93,$9C,$65,$A7,$F9,$1F
	.dc.b	$E5,$5F,$29,$F1,$B9,$7F,$CB,$71,$89,$F1,$C6,$27,$F0,$73,$9C,$FF
	.dc.b	$52,$D7,$F5,$CD,$7E,$23,$BD,$BF,$CF,$AD,$BF,$91,$9F,$E2,$6B,$FC
	.dc.b	$AF,$39,$7F,$27,$EE,$BE,$57,$FD,$E6,$4B,$FC,$9C,$65,$FC,$AE,$4F
	.dc.b	$94,$FC,$CB,$63,$31,$83,$3F,$11,$23,$8B,$70,$FE,$0B,$7C,$5B,$5F
	.dc.b	$A4,$8C,$47,$F0,$D6,$BF,$83,$B7,$C3,$78,$89,$87,$99,$F2,$EB,$23
	.dc.b	$D2,$DA,$F0,$E3,$DC,$93,$F0,$1F,$6F,$C2,$C8,$F8,$2D,$9F,$E3,$B8
	.dc.b	$3E,$53,$E1,$92,$BE,$3B,$EB,$F8,$F9,$1F,$31,$B5,$CB,$E1,$3E,$59
	.dc.b	$27,$E7,$BF,$3B,$E1,$24,$7D,$26,$CF,$F2,$EE,$E7,$B9,$FC,$31,$B5
	.dc.b	$FC,$31,$B3,$B9,$FE,$23,$EA,$7C,$C8,$FE,$2A,$D4,$D2,$7C,$27,$02
	.dc.b	$5F,$F1,$88,$47,$F1,$CF,$9A,$7A,$D4,$95,$FC,$7A,$32,$FF,$20,$F9
	.dc.b	$47,$C6,$FF,$24,$6C,$66,$30,$67,$59,$1C,$5B,$97,$F5,$0F,$99,$FF
	.dc.b	$36,$3F,$D5,$5A,$FF,$0B,$6B,$FC,$D1,$AB,$E5,$7C,$2F,$F5,$A6,$BF
	.dc.b	$D7,$DA,$FF,$15,$6B,$FC,$F9,$4F,$F3,$E6,$D7,$F3,$E6,$CF,$59,$FE
	.dc.b	$87,$F2,$AF,$7B,$FD,$C2,$BE,$0A,$C6,$1B,$87,$F4,$89,$E0,$D9,$E6
	.dc.b	$7F,$4E,$F9,$57,$8F,$FE,$EC,$6C,$FF,$B7,$57,$91,$99,$D3,$27,$FD
	.dc.b	$C5,$E6,$39,$F1,$A7,$FB,$8B,$C3,$7C,$17,$FB,$E1,$B7,$FE,$E9,$7C
	.dc.b	$B3,$B9,$D6,$72,$BA,$F5,$5A,$FF,$76,$8F,$39,$3F,$EF,$38,$C3,$7C
	.dc.b	$CC,$C7,$09,$B8,$FF,$73,$9B,$5F,$E1,$76,$D7,$F4,$DF,$36,$BF,$DD
	.dc.b	$A2,$BF,$E3,$A3,$53,$F0,$BA,$73,$5A,$DE,$4F,$97,$7B,$7E,$A7,$99
	.dc.b	$D1,$7C,$A7,$F8,$13,$E5,$5E,$8F,$F8,$AE,$B5,$3F,$E0,$4A,$FF,$04
	.dc.b	$B5,$E2,$65,$FF,$04,$FC,$A3,$E1,$DF,$2B,$86,$D7,$F8,$D2,$75,$39
	.dc.b	$38,$FC,$43,$82,$92,$B8,$46,$5E,$5D,$3F,$DD,$67,$E8,$C2,$FF,$06
	.dc.b	$FF,$88,$23,$31,$88,$ED,$AF,$F7,$D8,$FC,$F4,$6E,$31,$CB,$D5,$FF
	.dc.b	$01,$73,$BE,$27,$15,$FE,$2C,$8C,$3C,$4F,$F1,$7D,$F7,$C3,$C3,$CF
	.dc.b	$C2,$7C,$DF,$13,$68,$D9,$C3,$9F,$88,$95,$D4,$E2,$5A,$47,$5C,$3F
	.dc.b	$8D,$CF,$EA,$3C,$0E,$2F,$C0,$3F,$C8,$7C,$6E,$6F,$23,$FB,$68,$EC
	.dc.b	$D7,$81,$F8,$68,$DF,$37,$7F,$2A,$4F,$08,$F7,$7C,$AF,$CC,$8E,$2F
	.dc.b	$C3,$E0,$FF,$16,$36,$3F,$CA,$AE,$F7,$FB,$6E,$61,$F2,$AF,$8A,$F9
	.dc.b	$4D,$8F,$89,$3B,$FD,$1D,$F3,$BF,$99,$6C,$FF,$97,$9A,$FF,$A1,$9A
	.dc.b	$FF,$A2,$F9,$1E,$D7,$E8,$5B,$1F,$E6,$66,$BF,$E8,$E6,$BF,$E9,$39
	.dc.b	$7C,$67,$84,$91,$E1,$7A,$7B,$72,$3A,$35,$FF,$0A,$D7,$EC,$7D,$5F
	.dc.b	$C6,$C6,$0F,$F1,$2C,$7E,$D7,$F0,$97,$C5,$52,$4F,$F9,$F2,$32,$FF
	.dc.b	$9F,$BE,$41,$EE,$FE,$F5,$B1,$98,$C1,$9E,$B9,$1C,$5A,$FF,$AB,$BE
	.dc.b	$4B,$FC,$C2,$FF,$A2,$9A,$E7,$3F,$E1,$51,$88,$FF,$5C,$7C,$BF,$F5
	.dc.b	$67,$FD,$22,$D7,$FD,$72,$D7,$7E,$77,$F1,$3F,$97,$7C,$AF,$F8,$1A
	.dc.b	$7F,$D7,$F1,$8F,$5B,$FD,$36,$B0,$FE,$9B,$3A,$AE,$0F,$D7,$E7,$FF
	.dc.b	$51,$F1,$FE,$37,$A7,$F0,$D1,$83,$FD,$4F,$C5,$F2,$23,$F0,$5B,$E0
	.dc.b	$8F,$F9,$63,$4F,$F2,$51,$1C,$1F,$91,$F9,$1E,$E4,$70,$7E,$37,$E0
	.dc.b	$FB,$91,$C1,$FE,$64,$F3,$7F,$CA,$C4,$70,$7E,$77,$E4,$BD,$64,$70
	.dc.b	$7E,$67,$DD,$F5,$91,$C1,$FE,$70,$DB,$FC,$C4,$47,$07,$EA,$79,$FE
	.dc.b	$39,$1C,$1F,$A5,$F1,$3E,$39,$1C,$1F,$E7,$CF,$95,$FF,$37,$11,$C1
	.dc.b	$FB,$5F,$92,$FE,$C7,$C2,$F9,$3B,$18,$6C,$EC,$9C,$7F,$BB,$9F,$B6
	.dc.b	$69,$FE,$33,$4F,$93,$C7,$87,$FD,$22,$1F,$23,$E0,$F9,$1F,$0B,$71
	.dc.b	$FE,$61,$3A,$6C,$E9,$A6,$BE,$47,$D2,$7A,$BE,$46,$C6,$49,$97,$92
	.dc.b	$32,$F9,$BC,$BF,$83,$F2,$1F,$09,$B5,$E2,$7F,$2E,$47,$F8,$E3,$67
	.dc.b	$94,$7A,$3E,$47,$93,$EA,$F6,$BE,$6F,$B7,$F3,$9B,$19,$3D,$F1,$96
	.dc.b	$9F,$0B,$CD,$FA,$A7,$D6,$F3,$EC,$64,$98,$7D,$08,$CB,$FD,$2B,$A1
	.dc.b	$E3,$FF,$88,$3C,$07,$C8,$FF,$88,$A7,$38,$7F,$5F,$9C,$BF,$D8,$63
	.dc.b	$37,$FE,$A2,$BF,$F6,$19,$27,$FD,$8E,$F3,$3B,$7C,$AD,$8E,$9E,$D7
	.dc.b	$87,$E1,$FC,$4D,$8C,$9E,$18,$CB,$4F,$27,$B3,$E9,$3D,$5E,$D5,$8C
	.dc.b	$93,$1F,$ED,$39,$A7,$51,$8F,$37,$E1,$7B,$5F,$0D,$B7,$F0,$5F,$E2
	.dc.b	$24,$7F,$A9,$36,$79,$47,$C7,$F6,$B9,$3F,$27,$E5,$7E,$77,$B5,$F3
	.dc.b	$DB,$19,$3E,$74,$65,$A7,$D1,$F3,$FE,$B1,$F5,$FE,$22,$C6,$49,$87
	.dc.b	$BA,$32,$FF,$82,$3E,$57,$FA,$AC,$FF,$86,$57,$93,$FE,$60,$73,$3B
	.dc.b	$8C,$79,$19,$6D,$71,$6C,$EE,$3A,$4E,$DA,$6C,$B1,$B3,$C9,$1B,$6C
	.dc.b	$6D,$A6,$FC,$3F,$F2,$6D,$F9,$9B,$7E,$26,$D7,$66,$D7,$26,$9C,$AF
	.dc.b	$C0,$D9,$F0,$36,$7C,$10,$65,$F8,$63,$10,$8F,$97,$1E,$13,$2E,$23
	.dc.b	$C5,$18,$BF,$43,$4E,$11,$DA,$32,$D8,$CB,$DE,$32,$D8,$CC,$F7,$AF
	.dc.b	$51,$F0,$23,$C7,$7F,$01,$A7,$C0,$2D,$7C,$03,$DF,$1B,$6C,$6D,$A6
	.dc.b	$CF,$85,$1B,$6C,$6D,$A6,$EB,$E5,$3E,$D3,$E2,$96,$BD,$33,$F1,$63
	.dc.b	$27,$BC,$B3,$EF,$8F,$4C,$FB,$E7,$CA,$D7,$C2,$D9,$F0,$C7,$38,$EE
	.dc.b	$F8,$67,$99,$E1,$8F,$C2,$47,$E0,$0F,$59,$67,$F5,$2D,$AF,$F5,$1B
	.dc.b	$F3,$0C,$57,$17,$E4,$98,$31,$3E,$42,$78,$DF,$79,$C5,$7E,$DE,$30
	.dc.b	$E5,$B5,$F8,$26,$D7,$C6,$6D,$FB,$A3,$F3,$11,$89,$F7,$C7,$E7,$A3
	.dc.b	$0F,$CA,$8F,$CD,$C6,$23,$E5,$C6,$0F,$C8,$D7,$8D,$A7,$8C,$B1,$E3
	.dc.b	$6D,$78,$EF,$FC,$1D,$18,$3C,$73,$E6,$70,$F8,$CF,$33,$87,$C6,$60
	.dc.b	$FC,$2D,$78,$DB,$1E,$36,$9E,$32,$D7,$8E,$FD,$D7,$F4,$AB,$B3,$6B
	.dc.b	$71,$E6,$8F,$4C,$60,$F3,$1E,$93,$F3,$3F,$2B,$FE,$5D,$7F,$D7,$6F
	.dc.b	$FA,$5A,$7F,$D9,$8E,$6F,$FD,$D6,$E6,$77,$E8,$2B,$7E,$BD,$9C,$C6
	.dc.b	$0E,$79,$E7,$0E,$6D,$8F,$E3,$38,$3C,$FC,$5F,$F1,$F8,$FD,$63,$6B
	.dc.b	$2A,$32,$D9,$CC,$63,$27,$89,$FD,$37,$C8,$FE,$9C,$DC,$7F,$A9,$3F
	.dc.b	$A6,$C3,$F9,$6B,$5F,$E5,$8D,$AE,$73,$FA,$99,$CB,$C8,$FE,$B9,$E3
	.dc.b	$76,$7E,$5E,$DE,$4D,$FC,$AD,$8C,$9F,$42,$32,$D3,$E9,$F9,$5E,$C3
	.dc.b	$DB,$E5,$58,$C9,$30,$FB,$71,$97,$FB,$87,$8B,$FD,$B7,$C5,$FE,$DC
	.dc.b	$DA,$FE,$E4,$FB,$48,$FB,$73,$46,$2B,$9C,$FF,$79,$F1,$39,$BF,$88
	.dc.b	$DB,$E8,$9F,$4B,$C9,$B1,$92,$57,$E4,$23,$2F,$E4,$B6,$FB,$3B,$FD
	.dc.b	$A6,$D7,$E5,$1B,$3B,$8F,$CB,$CF,$E1,$3E,$2D,$F2,$B7,$E5,$70,$DA
	.dc.b	$FE,$D5,$CD,$F9,$7C,$17,$C2,$7E,$E3,$63,$F9,$33,$5F,$E6,$0D,$7F
	.dc.b	$B8,$BB,$9E,$8D,$9C,$FB,$7E,$27,$F5,$BE,$5F,$E5,$8F,$CB,$F9,$B6
	.dc.b	$32,$4C,$3F,$6B,$19,$7F,$6F,$C5,$F9,$EF,$37,$F3,$ED,$AF,$DC,$B6
	.dc.b	$77,$1F,$BC,$9F,$8D,$E8,$7C,$7F,$32,$CF,$C4,$3D,$6E,$2B,$E0,$CF
	.dc.b	$E8,$F3,$34,$95,$C6,$7F,$C2,$D3,$A8,$C4,$7C,$EF,$37,$A6,$BF,$EA
	.dc.b	$FF,$84,$FF,$77,$92,$BF,$59,$E8,$23,$F6,$5E,$E3,$E4,$E9,$2B,$F6
	.dc.b	$51,$97,$F6,$9A,$3E,$4F,$F7,$4D,$8C,$C6,$0F,$F7,$FE,$72,$3C,$2D
	.dc.b	$3F,$B3,$57,$16,$CF,$90,$93,$E0,$F9,$3E,$E4,$70,$7F,$07,$F3,$9C
	.dc.b	$DF,$C3,$5B,$19,$8C,$1F,$03,$E0,$E4,$7B,$9B,$87,$F7,$9C,$9F,$25
	.dc.b	$EE,$25,$7C,$27,$B7,$44,$8F,$13,$6B,$DB,$F1,$7F,$38,$93,$F3,$1F
	.dc.b	$93,$EE,$48,$F9,$AD,$9F,$F0,$57,$CC,$F8,$A7,$F9,$23,$63,$38,$7F
	.dc.b	$2B,$8C,$47,$EA,$BD,$5F,$C6,$12,$BF,$62,$FF,$31,$F3,$BE,$84,$FF
	.dc.b	$87,$DA,$FF,$98,$1A,$FF,$99,$5A,$BE,$77,$F8,$CD,$7F,$37,$6C,$7F
	.dc.b	$88,$5A,$FF,$8A,$DA,$FF,$9B,$3E,$4D,$E8,$7F,$43,$6D,$7F,$41,$6C
	.dc.b	$FF,$8A,$DA,$FF,$8D,$7E,$4D,$F1,$5F,$D2,$DB,$5F,$D2,$5B,$3F,$E3
	.dc.b	$36,$BF,$E3,$DF,$93,$FF,$3C,$37,$D1,$FE,$A9,$18,$8F,$F4,$0B,$5F
	.dc.b	$F4,$33,$EA,$6F,$95,$C0,$97,$FD,$67,$E4,$FB,$CF,$FA,$29,$AF,$FA
	.dc.b	$3D,$A9,$BF,$26,$F7,$3F,$B3,$36,$BF,$B2,$36,$BF,$C9,$CD,$7F,$CB
	.dc.b	$0B,$D9,$FE,$F6,$EF,$C1,$FE,$68,$E6,$BF,$84,$B9,$7F,$D3,$FF,$26
	.dc.b	$F8,$1F,$F0,$1B,$63,$F8,$7B,$FC,$25,$B1,$AF,$0A,$47,$F1,$4F,$19
	.dc.b	$D5,$8C,$D9,$7F,$45,$1E,$EF,$F8,$41,$B7,$FC,$71,$F3,$4F,$08,$FF
	.dc.b	$59,$C6,$A3,$07,$F5,$0D,$39,$1F,$E1,$AC,$1F,$27,$8B,$FE,$4F,$4F
	.dc.b	$2B,$FC,$3E,$D8,$CF,$0E,$27,$2C,$8E,$3E,$95,$AF,$09,$FE,$C9,$79
	.dc.b	$9E,$1F,$85,$B3,$E0,$DF,$B9,$E0,$DE,$C7,$91,$C3,$E6,$B7,$C9,$7F
	.dc.b	$B5,$7E,$4D,$E7,$7F,$A5,$6F,$D3,$5A,$3F,$C5,$05,$3F,$C5,$55,$87
	.dc.b	$07,$FB,$73,$E4,$9F,$08,$F9,$3F,$F1,$84,$61,$B1,$99,$FF,$4E,$4E
	.dc.b	$5B,$F0,$BF,$71,$BC,$9C,$A3,$11,$FE,$AB,$94,$62,$3C,$5D,$3F,$73
	.dc.b	$5F,$27,$35,$9F,$F7,$A5,$FE,$19,$FE,$BE,$FE,$95,$B1,$B7,$FB,$4F
	.dc.b	$97,$79,$8E,$6B,$94,$FD,$87,$F4,$CF,$4F,$85,$D1,$BE,$8B,$1E,$0F
	.dc.b	$84,$E4,$61,$E4,$F7,$F0,$64,$6C,$FE,$EB,$C6,$E7,$F3,$3F,$AD,$8C
	.dc.b	$F3,$62,$72,$FA,$7C,$C9,$3A,$3F,$BF,$FC,$97,$C1,$7F,$22,$FB,$38
	.dc.b	$FF,$C6,$0D,$E1,$4C,$8E,$13,$AE,$1F,$91,$F2,$6F,$86,$A7,$CB,$AE
	.dc.b	$57,$CB,$F4,$A7,$CC,$FE,$39,$B1,$FE,$77,$6B,$FE,$E1,$6B,$FE,$EB
	.dc.b	$5F,$17,$C4,$91,$C5,$E9,$F1,$B2,$3A,$35,$E9,$F2,$1E,$12,$3E,$17
	.dc.b	$C9,$F3,$B2,$3C,$8D,$7C,$9F,$1F,$E9,$23,$E9,$7D,$9B,$C8,$F6,$35
	.dc.b	$FF,$7E,$7C,$53,$E5,$BF,$5E,$D8,$FF,$4C,$35,$DC,$6B,$E3,$A4,$69
	.dc.b	$AF,$4D,$F9,$91,$E6,$F6,$F8,$E9,$1D,$9A,$FD,$E2,$7C,$9B,$57,$AB
	.dc.b	$FE,$AB,$8C,$37,$6F,$F2,$15,$71,$3F,$0F,$3A,$70,$3E,$6F,$FD,$B8
	.dc.b	$ED,$AF,$85,$AF,$E4,$1A,$BE,$79,$F0,$9F,$EC,$16,$EF,$FE,$BF,$6C
	.dc.b	$F8,$1A,$F9,$7E,$BF,$FD,$E2,$DE,$0F,$F6,$54,$62,$3E,$13,$5F,$D0
	.dc.b	$BF,$B1,$AF,$F6,$63,$63,$DA,$D7,$B7,$67,$FB,$56,$BF,$C0,$10,$69
	.dc.b	$3D,$5D,$7C,$8C,$C3,$F0,$2D,$AF,$C3,$D7,$32,$4F,$E1,$9E,$DF,$21
	.dc.b	$23,$F1,$2D,$9F,$C7,$FC,$8F,$8C,$93,$F8,$E7,$CD,$F1,$F2,$3F,$22
	.dc.b	$D9,$FC,$BF,$C8,$7A,$C9,$3F,$96,$7E,$4C,$7F,$C0,$0D,$3F,$BE,$3F
	.dc.b	$D7,$1F,$E1,$33,$FE,$8E,$73,$E0,$7F,$AF,$5C,$CF,$CA,$39,$CF,$CE
	.dc.b	$69,$CC,$3F,$E9,$83,$FC,$EA,$E5,$FF,$4D,$E4,$F9,$0D,$CF,$F6,$E8
	.dc.b	$7C,$86,$30,$FF,$67,$46,$EF,$8C,$E5,$FD,$24,$66,$B9,$4E,$5F,$E4
	.dc.b	$B1,$9A,$E7,$39,$9F,$C0,$7C,$7D,$A3,$7E,$3F,$0C,$C3,$D2,$E5,$A7
	.dc.b	$BF,$E5,$7A,$2F,$96,$B1,$86,$C7,$EC,$30,$F8,$F3,$FE,$EB,$7F,$7D
	.dc.b	$08,$C4,$7D,$66,$BC,$23,$C0,$7C,$FF,$8A,$EB,$1F,$2F,$DB,$F6,$79
	.dc.b	$7E,$D6,$D7,$D2,$6C,$EE,$3E,$AC,$FC,$3E,$69,$1F,$63,$DB,$B5,$F5
	.dc.b	$E6,$EF,$F5,$A6,$D7,$AB,$D6,$F8,$7F,$0D,$F1,$1B,$1F,$85,$7D,$6F
	.dc.b	$F8,$58,$F8,$85,$C3,$F8,$AB,$5F,$E3,$4E,$6F,$F1,$31,$F9,$09,$FA
	.dc.b	$3E,$BF,$7A,$FC,$7F,$B5,$F7,$4F,$BD,$E8,$58,$C9,$2F,$F2,$51,$97
	.dc.b	$FC,$83,$FA,$57,$E6,$B1,$FC,$9C,$7E,$26,$31,$5F,$92,$8C,$7C,$67
	.dc.b	$E2,$DB,$1F,$A0,$7F,$0B,$EA,$7E,$1B,$D6,$B3,$FA,$76,$CF,$C0,$9F
	.dc.b	$8A,$49,$F8,$AE,$B0,$F8,$30,$F9,$6C,$3F,$D1,$F1,$7C,$A7,$9C,$F4
	.dc.b	$7C,$75,$9F,$F0,$4D,$FC,$AA,$DC,$60,$E5,$5C,$1C,$63,$FE,$10,$9D
	.dc.b	$BF,$1B,$9F,$E2,$7C,$CF,$C5,$6D,$70,$95,$7E,$96,$30,$AB,$84,$61
	.dc.b	$FD,$B7,$D8,$BD,$9B,$F2,$58,$6D,$75,$9D,$36,$74,$D3,$53,$C3,$0F
	.dc.b	$F3,$ED,$70,$6D,$70,$6D,$70,$39,$3C,$2B,$96,$7F,$92,$BE,$4D,$BE
	.dc.b	$4D,$BE,$57,$D0,$E4,$F4,$8E,$57,$D2,$3F,$C0,$17,$D1,$B7,$D1,$A7
	.dc.b	$4A,$ED,$7D,$0E,$CF,$48,$EC,$74,$BE,$D3,$FE,$4A,$BE,$CD,$BE,$CD
	.dc.b	$AE,$C7,$81,$ED,$7E,$03,$B3,$E0,$8E,$C7,$81,$ED,$7E,$05,$7E,$06
	.dc.b	$DF,$81,$B5,$E0,$7C,$47,$81,$F1,$5F,$80,$F1,$3E,$08,$F1,$1E,$07
	.dc.b	$C4,$F8,$2F,$C5,$7E,$11,$7E,$16,$D7,$86,$BC,$67,$84,$F1,$DF,$86
	.dc.b	$3C,$67,$85,$F1,$C7,$84,$F1,$9E,$18,$F1,$CF,$86,$BC,$04,$8F,$04
	.dc.b	$78,$A7,$C0,$78,$AB,$E7,$BF,$E3,$D8,$FC,$D3,$6F,$FA,$3C,$62,$3E
	.dc.b	$21,$9A,$FE,$A5,$19,$8F,$31,$93,$BB,$FE,$FE,$8F,$1C,$FC,$5F,$86
	.dc.b	$F8,$CD,$8C,$9E,$88,$CB,$4C,$D7,$48,$CB,$63,$37,$F2,$E3,$F2,$93
	.dc.b	$A8,$C4,$7E,$22,$31,$3F,$3C,$FD,$3B,$88,$FA,$13,$FB,$58,$C4,$7D
	.dc.b	$13,$F8,$0B,$8B,$FC,$13,$6F,$F8,$8B,$6F,$F8,$73,$6B,$8C,$7A,$0F
	.dc.b	$13,$E8,$70,$F8,$9F,$29,$87,$C5,$7E,$58,$C5,$F8,$9B,$7E,$26,$9E
	.dc.b	$22,$D7,$88,$FA,$F1,$F4,$DF,$C3,$E4,$EB,$FE,$B0,$7F,$59,$EC,$B7
	.dc.b	$FC,$CA,$2D,$7F,$60,$9B,$8F,$F4,$5F,$2B,$FB,$07,$C0,$7F,$61,$6E
	.dc.b	$1F,$CE,$9B,$7F,$DA,$1B,$5F,$D3,$7A,$7F,$B4,$7C,$CF,$F6,$96,$DF
	.dc.b	$F4,$27,$FC,$3B,$9F,$F7,$FC,$3F,$AD,$46,$09,$3B,$3F,$A4,$F5,$F3
	.dc.b	$7F,$A5,$FC,$17,$F7,$53,$FB,$B7,$65,$8C,$92,$FF,$BE,$C6,$5F,$EF
	.dc.b	$FF,$01,$FD,$F7,$DF,$FE,$FC,$DA,$FE,$AE,$D9,$DC,$7F,$5A,$9F,$F2
	.dc.b	$67,$C8,$3E,$4F,$FC,$0D,$DB,$86,$D7,$F6,$39,$D3,$FE,$43,$6D,$FF
	.dc.b	$AD,$1A,$FF,$AE,$5A,$FF,$B0,$DD,$EF,$FE,$22,$6C,$FF,$80,$7B,$7C
	.dc.b	$0F,$F6,$EF,$90,$FF,$0C,$1F,$E1,$9E,$EB,$19,$26,$1F,$E2,$08,$CA
	.dc.b	$C3,$FC,$5F,$FA,$53,$E5,$BF,$C4,$5E,$C7,$F9,$AE,$31,$87,$F7,$C8
	.dc.b	$C4,$EA,$3F,$C0,$53,$AB,$FF,$16,$D7,$1E,$CE,$5D,$B6,$F6,$DA,$FF
	.dc.b	$6B,$DF,$F1,$36,$D7,$F9,$92,$31,$1F,$ED,$F7,$58,$FA,$27,$CD,$D9
	.dc.b	$CE,$3D,$7F,$3D,$E9,$3D,$5C,$17,$2F,$3C,$5C,$7E,$14,$7F,$BC,$F1
	.dc.b	$F9,$F0,$0A,$7E,$27,$CA,$FC,$5F,$89,$B7,$F0,$9B,$7D,$0E,$56,$8C
	.dc.b	$E3,$F4,$1F,$9D,$F2,$DF,$3F,$E3,$EE,$1F,$31,$B8,$7C,$69,$F8,$84
	.dc.b	$9F,$F4,$86,$08,$DE,$1F,$73,$E2,$AD,$7B,$7E,$7B,$E6,$11,$E2,$FC
	.dc.b	$FF,$A2,$91,$D1,$B8,$FD,$6F,$91,$FA,$84,$9F,$0B,$F5,$FE,$2E,$47
	.dc.b	$91,$B3,$F7,$BE,$2B,$EE,$12,$7D,$2F,$DF,$F8,$49,$1E,$C6,$CF,$E1
	.dc.b	$BE,$13,$F0,$84,$9F,$82,$FE,$1F,$E2,$A4,$7C,$46,$CF,$E3,$BE,$2B
	.dc.b	$F1,$84,$9F,$92,$FF,$6C,$E5,$91,$FD,$A9,$B3,$F9,$68,$FC,$A1,$0F
	.dc.b	$A2,$7F,$AE,$BB,$BE,$F3,$FC,$07,$93,$04,$7F,$D8,$DE,$5F,$D7,$6F
	.dc.b	$1F,$FB,$0A,$31,$7F,$91,$23,$87,$07,$F4,$1E,$1E,$04,$AE,$AF,$3E
	.dc.b	$19,$1E,$06,$CF,$21,$F2,$DE,$87,$0D,$7B,$B5,$F2,$35,$FD,$32,$E7
	.dc.b	$FF,$6D,$B6,$BF,$DC,$8D,$9E,$B3,$AE,$79,$5A,$6C,$F9,$7E,$57,$B2
	.dc.b	$69,$34,$FF,$74,$BD,$63,$E2,$B5,$F9,$4D,$7F,$78,$D7,$A4,$F2,$25
	.dc.b	$7F,$BE,$E5,$1F,$EF,$6F,$95,$78,$D4,$95,$FF,$00,$63,$2D,$83,$FE
	.dc.b	$A7,$8C,$DF,$F0,$50,$FF,$B8,$63,$F0,$0E,$BA,$7F,$DC,$7F,$2A,$F5
	.dc.b	$B9,$DF,$18,$C5,$FF,$AC,$D7,$AF,$AB,$E3,$76,$BF,$19,$5A,$78,$56
	.dc.b	$6F,$F9,$07,$A1,$FC,$BA,$77,$39,$9F,$E9,$1E,$4D,$C4,$DF,$2B,$0D
	.dc.b	$BF,$F6,$34,$A7,$96,$1F,$32,$73,$1D,$6D,$ED,$D3,$1E,$CD,$AF,$F6
	.dc.b	$84,$A3,$31,$8F,$6A,$CF,$CD,$0F,$53,$C3,$E0,$D9,$E7,$B7,$84,$F1
	.dc.b	$6B,$63,$BB,$87,$C3,$6F,$94,$FE,$8D,$F2,$AF,$77,$F6,$B7,$EB,$AD
	.dc.b	$1F,$AF,$3F,$A6,$55,$69,$5D,$7C,$A3,$D4,$7C,$AE,$1B,$5F,$B9,$9D
	.dc.b	$4E,$4E,$3C,$DF,$DF,$67,$4D,$7F,$84,$CF,$E6,$B8,$F6,$78,$3D,$4F
	.dc.b	$E2,$0F,$E3,$A5,$3F,$79,$FF,$80,$0D,$7F,$34,$A7,$8E,$BF,$C6,$63
	.dc.b	$55,$CF,$93,$FB,$15,$FF,$6D,$8C,$75,$70,$8F,$EE,$15,$C2,$F9,$72
	.dc.b	$E6,$7C,$7E,$A5,$A3,$67,$0E,$7D,$49,$5E,$88,$AF,$EC,$72,$79,$7D
	.dc.b	$5B,$1E,$EA,$F0,$71,$7E,$2A,$3D,$19,$7F,$35,$8F,$DA,$46,$0F,$F0
	.dc.b	$43,$87,$F7,$CF,$16,$F2,$77,$F3,$A4,$71,$3F,$15,$1E,$FF,$94,$F9
	.dc.b	$91,$97,$E2,$3F,$E1,$56,$A7,$6F,$F0,$D7,$3F,$2A,$F9,$2F,$94,$D9
	.dc.b	$E9,$DB,$F3,$7F,$4B,$FE,$3B,$1F,$97,$6B,$FD,$1D,$AF,$F5,$26,$BF
	.dc.b	$D5,$FE,$59,$E8,$FE,$75,$B1,$FA,$16,$BF,$D4,$5A,$F3,$E6,$FD,$25
	.dc.b	$7A,$A3,$07,$8E,$3A,$7A,$12,$3A,$35,$E9,$E8,$78,$48,$F8,$5E,$B8
	.dc.b	$C8,$F7,$35,$F2,$79,$FE,$92,$3E,$97,$D9,$C3,$23,$D8,$D7,$E3,$73
	.dc.b	$7A,$88,$FC,$B7,$F8,$07,$C3,$BE,$56,$92,$7F,$83,$46,$5B,$C1,$C0
	.dc.b	$8E,$0F,$E1,$9F,$3F,$FE,$41,$6F,$A1,$FC,$36,$31,$1F,$DF,$DA,$FF
	.dc.b	$82,$BE,$55,$B3,$DA,$FE,$2D,$18,$6E,$FF,$D3,$2B,$FC,$11,$39,$7F
	.dc.b	$C1,$5F,$2B,$F8,$DD,$BF,$8F,$36,$33,$9F,$F0,$B7,$FC,$39,$19,$7F
	.dc.b	$C3,$7E,$23,$D7,$D4,$7F,$26,$6B,$FC,$BD,$B1,$BB,$FE,$59,$F1,$CD
	.dc.b	$F6,$D8,$CC,$60,$CF,$95,$23,$8B,$70,$FC,$F7,$83,$E0,$E7,$FE,$1B
	.dc.b	$5F,$A3,$8C,$1E,$4B,$7C,$9D,$24,$FD,$A7,$51,$23,$EE,$B6,$FF,$59
	.dc.b	$E0,$77,$24,$FE,$C5,$ED,$F2,$52,$3F,$0A,$D9,$FF,$1D,$53,$E4,$BC
	.dc.b	$84,$AF,$C5,$BF,$E5,$5F,$7B,$FB,$D6,$6F,$81,$FF,$6E,$C6,$1B,$5F
	.dc.b	$ED,$98,$E0,$D2,$61,$FE,$19,$F6,$FF,$C0,$CF,$FA,$9E,$DF,$05,$CA
	.dc.b	$3F,$C3,$DE,$07,$FB,$17,$CB,$FE,$99,$5F,$8F,$C3,$F9,$04,$E5,$FF
	.dc.b	$30,$46,$6B,$8F,$C1,$7A,$3A,$BF,$C2,$57,$B9,$F0,$46,$3C,$79,$87
	.dc.b	$4F,$82,$B5,$E2,$9B,$97,$F5,$DC,$15,$8B,$FE,$35,$F0,$5D,$B6,$EE
	.dc.b	$D8,$F3,$BF,$94,$AF,$F6,$24,$7F,$1D,$8C,$11,$D9,$EE,$F8,$0E,$6F
	.dc.b	$C0,$F7,$3D,$7E,$87,$B1,$B1,$93,$E2,$C6,$5A,$7C,$8F,$3F,$E1,$9F
	.dc.b	$13,$DB,$B1,$92,$6D,$F4,$A3,$2F,$F9,$47,$B3,$FD,$BD,$3F,$D4,$5F
	.dc.b	$F2,$24,$64,$FF,$20,$38,$DB,$FC,$08,$F0,$8F,$F4,$53,$5F,$F4,$AB
	.dc.b	$5F,$F4,$D3,$5F,$F7,$63,$C6,$39,$C7,$F9,$36,$5F,$1B,$E0,$8F,$F2
	.dc.b	$87,$43,$E4,$6F,$C8,$E1,$B7,$FD,$A6,$74,$D9,$D3,$4D,$7C,$87,$43
	.dc.b	$AF,$C8,$D8,$C9,$30,$FF,$3B,$5A,$75,$1F,$E7,$99,$7C,$67,$A7,$E3
	.dc.b	$FD,$4D,$8F,$EF,$86,$8F,$F2,$11,$67,$FD,$64,$D7,$FD,$7A,$D7,$FD
	.dc.b	$84,$EF,$4F,$F1,$A3,$4C,$FC,$3F,$81,$F8,$5F,$03,$F2,$3D,$8F,$92
	.dc.b	$D8,$C9,$F1,$E3,$2D,$3E,$57,$8D,$F4,$0F,$A3,$F0,$36,$32,$4C,$7E
	.dc.b	$2C,$65,$F8,$9E,$F7,$DB,$F8,$AF,$B8,$DA,$F8,$6D,$9D,$C7,$C6,$9D
	.dc.b	$B4,$D9,$63,$67,$D1,$8D,$B6,$36,$D3,$79,$FC,$89,$DB,$67,$6D,$9D
	.dc.b	$C7,$D0,$9D,$B4,$D9,$63,$67,$EA,$A3,$6D,$8D,$B4,$D9,$FE,$C4,$8D
	.dc.b	$B6,$36,$D3,$6A,$36,$D3,$65,$8D,$C6,$89,$1A,$6C,$6A,$30,$47,$04
	.dc.b	$9F,$C7,$D7,$D7,$6C,$FD,$36,$CF,$CF,$69,$F2,$CB,$1F,$1D,$B1,$F0
	.dc.b	$DA,$FB,$DB,$1E,$B6,$D7,$A1,$B1,$E7,$6C,$F9,$E3,$F4,$F5,$FE,$08
	.dc.b	$7F,$D5,$47,$EC,$1C,$11,$C1,$23,$F6,$71,$97,$E8,$FB,$DF,$42,$3E
	.dc.b	$8B,$67,$E8,$B6,$77,$1F,$52,$76,$D3,$65,$8D,$9F,$7A,$36,$D8,$DB
	.dc.b	$4D,$D7,$C6,$9D,$B6,$76,$D9,$DC,$7C,$43,$C1,$1F,$C2,$AB,$D0,$FD
	.dc.b	$06,$C6,$E7,$F8,$F4,$61,$FE,$8F,$5F,$CD,$E3,$10,$9F,$9E,$F0,$7B
	.dc.b	$9C,$EB,$C0,$E2,$39,$BF,$C9,$A7,$9B,$CA,$77,$3F,$05,$F3,$47,$18
	.dc.b	$F0,$57,$8E,$30,$71,$3F,$94,$D7,$EB,$B8,$3F,$AD,$42,$A4,$61,$BE
	.dc.b	$E7,$F5,$0F,$80,$95,$FD,$35,$FE,$A7,$18,$96,$3F,$EE,$7F,$8B,$FE
	.dc.b	$B3,$7F,$D4,$E7,$E2,$1A,$7C,$0A,$36,$FF,$4E,$37,$3C,$F9,$FF,$B0
	.dc.b	$C6,$2B,$8D,$70,$AF,$F7,$B6,$3F,$DA,$E7,$FA,$3F,$97,$FD,$AF,$CD
	.dc.b	$FE,$D8,$D8,$FE,$A6,$FE,$94,$8F,$F5,$46,$E1,$FD,$75,$FE,$E9,$E5
	.dc.b	$72,$9F,$F6,$B6,$2F,$90,$F2,$5F,$CB,$8C,$4E,$BE,$47,$33,$EF,$AB
	.dc.b	$5A,$8F,$88,$71,$8C,$CE,$9D,$F4,$70,$F9,$1B,$87,$89,$B7,$C8,$F9
	.dc.b	$8E,$35,$FA,$8F,$8B,$E4,$5E,$4D,$F2,$70,$DC,$3F,$DE,$53,$A7,$EB
	.dc.b	$E1,$F6,$23,$0E,$03,$87,$6D,$78,$4F,$69,$EB,$E4,$D9,$F4,$6A,$EC
	.dc.b	$F8,$A7,$C6,$ED,$B7,$F4,$22,$D7,$09,$BA,$7E,$47,$E4,$7E,$6F,$C2
	.dc.b	$7C,$E6,$E1,$F6,$1B,$7E,$06,$D7,$E6,$7A,$BE,$AF,$C8,$FD,$66,$DF
	.dc.b	$E0,$1F,$F5,$9E,$7F,$BE,$C3,$E2,$B5,$FC,$AC,$68,$E5,$3F,$88,$EC
	.dc.b	$B3,$F0,$3E,$0B,$F0,$67,$E1,$3B,$6D,$7E,$3A,$2C,$FC,$69,$B9,$7E
	.dc.b	$BF,$E0,$3F,$17,$F0,$5F,$8C,$6E,$1F,$96,$6D,$FC,$E6,$D7,$ED,$21
	.dc.b	$EA,$3C,$EF,$CA,$F6,$FE,$62,$31,$87,$E8,$23,$13,$FE,$D4,$9C,$D7
	.dc.b	$E0,$DA,$FE,$21,$AF,$F0,$07,$77,$E1,$6D,$FE,$F3,$B7,$AB,$FE,$A4
	.dc.b	$F8,$0F,$F4,$71,$1F,$C8,$5B,$D1,$4F,$FB,$9B,$E4,$5F,$21,$49,$7C
	.dc.b	$23,$2D,$F9,$0E,$04,$70,$76,$BD,$E1,$F9,$F6,$BF,$A4,$6B,$FC,$71
	.dc.b	$AF,$F2,$DA,$7C,$37,$94,$95,$E3,$7D,$1E,$F4,$8F,$33,$6B,$C1,$87
	.dc.b	$FA,$BC,$93,$E5,$7E,$07,$C1,$48,$F7,$36,$7F,$6D,$83,$E4,$7E,$31
	.dc.b	$2B,$E2,$3F,$EB,$BF,$8D,$91,$F0,$5B,$5F,$BF,$F0,$7E,$DB,$FD,$25
	.dc.b	$AF,$F0,$2E,$2F,$C6,$3F,$C1,$DF,$A9,$50,$FA,$45,$8F,$E1,$5F,$36
	.dc.b	$F0,$78,$C7,$F5,$58,$C4,$FE,$01,$B1,$C2,$76,$FF,$09,$E1,$BE,$0F
	.dc.b	$E1,$2F,$F0,$ED,$9F,$E3,$0D,$7F,$8E,$BF,$61,$FE,$25,$C5,$C8,$93
	.dc.b	$F9,$0F,$9A,$79,$3E,$06,$C6,$63,$07,$93,$E4,$DE,$D5,$24,$FE,$6A
	.dc.b	$32,$DF,$6F,$81,$1C,$1F,$A0,$F9,$A7,$AB,$F0,$1B,$19,$8C,$1F,$13
	.dc.b	$E5,$3A,$E9,$FC,$DD,$AF,$F3,$C6,$BF,$DD,$1A,$FF,$81,$55,$F3,$25
	.dc.b	$7E,$CA,$32,$FE,$D3,$E5,$1E,$17,$F6,$07,$FC,$B5,$1F,$E0,$07,$2F
	.dc.b	$F8,$47,$BF,$F8,$94,$6B,$0F,$F0,$A3,$A7,$FA,$73,$5F,$EA,$6D,$7A
	.dc.b	$57,$E0,$BE,$51,$EA,$7F,$70,$6E,$1D,$CF,$F0,$8B,$FE,$1F,$A3,$A7
	.dc.b	$FB,$53,$93,$5E,$5F,$E4,$31,$FE,$51,$1C,$57,$8B,$FD,$F9,$B7,$DE
	.dc.b	$FD,$DE,$2E,$6B,$4C,$9D,$4B,$07,$BD,$98,$FE,$F8,$7D,$E2,$EF,$FE
	.dc.b	$0C,$5E,$BD,$8C,$32,$FA,$CB,$D8,$C3,$D1,$E4,$7F,$85,$B0,$77,$62
	.dc.b	$B8,$ED,$DC,$F0,$7B,$16,$3A,$B8,$7B,$E0,$F8,$8F,$F2,$77,$CA,$3D
	.dc.b	$BF,$F0,$5D,$FA,$6B,$47,$F3,$C2,$9F,$CF,$AB,$0E,$0F,$F2,$4F,$C9
	.dc.b	$BD,$07,$CA,$7F,$44,$8C,$36,$33,$3F,$E1,$79,$CB,$79,$BF,$CB,$2D
	.dc.b	$E5,$E5,$18,$8F,$F1,$0C,$FE,$72,$7B,$7A,$1F,$E5,$A7,$FB,$54,$7E
	.dc.b	$0A,$7E,$DB,$5F,$F1,$A2,$9E,$1A,$FA,$B9,$AD,$7F,$9C,$F8,$7F,$83
	.dc.b	$BF,$84,$6B,$CA,$36,$AD,$CD,$85,$E8,$53,$11,$67,$CB,$E7,$EC,$C3
	.dc.b	$E6,$7A,$E0,$F8,$3A,$6E,$7F,$2F,$C5,$F8,$B6,$F2,$79,$23,$F6,$31
	.dc.b	$AE,$6C,$4F,$4F,$5E,$4F,$03,$F4,$3F,$29,$F1,$5B,$B7,$BA,$3E,$57
	.dc.b	$0F,$F8,$5D,$D4,$FF,$97,$DC,$F0,$7F,$A3,$3E,$51,$F1,$FF,$19,$B5
	.dc.b	$D0,$FD,$7F,$E9,$8F,$96,$FC,$83,$63,$FC,$1A,$D7,$FC,$BC,$D7,$FC
	.dc.b	$C7,$F1,$9F,$C7,$9F,$F0,$9B,$5F,$F3,$23,$5D,$F3,$F6,$F8,$79,$1D
	.dc.b	$9A,$F0,$E8,$FD,$15,$7B,$A3,$BC,$60,$FD,$37,$B0,$F9,$3A,$49,$FF
	.dc.b	$0C,$3E,$CF,$89,$91,$EC,$6B,$E6,$E1,$EE,$47,$D6,$F6,$F9,$69,$1F
	.dc.b	$E2,$36,$BF,$13,$E2,$7E,$49,$1F,$92,$FA,$BA,$E4,$7C,$66,$BF,$E8
	.dc.b	$9F,$88,$7C,$B7,$F0,$16,$C7,$F8,$F9,$AE,$E3,$F8,$37,$B6,$F8,$6A
	.dc.b	$49,$FF,$22,$C6,$5F,$F2,$3F,$CB,$7F,$91,$9B,$EB,$FF,$0F,$8C,$47
	.dc.b	$FA,$6D,$AF,$FA,$83,$E5,$BE,$77,$C8,$7F,$94,$1B,$1F,$EA,$15,$6F
	.dc.b	$27,$5C,$2B,$FC,$5F,$C7,$7F,$85,$BA,$0D,$27,$67,$F1,$FF,$82,$FF
	.dc.b	$0D,$F9,$66,$92,$3F,$CC,$3F,$1E,$E2,$FF,$10,$B6,$33,$18,$3C,$1E
	.dc.b	$2C,$8F,$CE,$B7,$7F,$CB,$F9,$3F,$E2,$AE,$CF,$E1,$35,$F9,$C8,$C1
	.dc.b	$FE,$2E,$B7,$CA,$D2,$4F,$D7,$7D,$DE,$8C,$8F,$D5,$B7,$0F,$D2,$F9
	.dc.b	$7D,$C9,$3F,$AA,$7B,$7C,$AC,$8F,$C1,$36,$7F,$67,$E4,$FF,$BB,$38
	.dc.b	$7F,$B0,$D7,$EF,$23,$1E,$96,$5F,$A0,$48,$FD,$EB,$6B,$FD,$D1,$5F
	.dc.b	$E0,$1F,$27,$FD,$48,$FF,$67,$AE,$87,$F0,$89,$F0,$73,$78,$5B,$19
	.dc.b	$3F,$86,$C6,$5A,$7F,$7E,$DD,$F2,$08,$C3,$6B,$FD,$F5,$2F,$91,$C6
	.dc.b	$3F,$EB,$78,$E3,$7C,$09,$5C,$2B,$70,$89,$3C,$1E,$BF,$21,$A7,$3F
	.dc.b	$23,$54,$A7,$0A,$6C,$7E,$1C,$95,$FC,$89,$B5,$DA,$3E,$07,$7F,$07
	.dc.b	$E1,$7C,$17,$4F,$90,$EA,$D8,$C9,$F1,$E3,$2D,$3E,$57,$C8,$79,$0F
	.dc.b	$2F,$A9,$63,$24,$C3,$E5,$C6,$56,$BF,$ED,$1F,$85,$F6,$FC,$1F,$FB
	.dc.b	$7D,$B7,$FE,$34,$F8,$DF,$F4,$7D,$FF,$B6,$E7,$FD,$C0,$49,$FF,$69
	.dc.b	$CF,$F9,$15,$FF,$6F,$47,$C8,$F9,$4E,$91,$FE,$3E,$F3,$9F,$29,$F3
	.dc.b	$3C,$4F,$A1,$18,$9F,$C3,$46,$27,$51,$C5,$B8,$FA,$A3,$FD,$A1,$EB
	.dc.b	$F2,$39,$FC,$9D,$8F,$04,$59,$E0,$7E,$17,$05,$67,$5F,$C9,$4F,$FB
	.dc.b	$B7,$E5,$3C,$BA,$79,$9B,$8F,$AD,$FF,$58,$E0,$69,$23,$FD,$D4,$F1
	.dc.b	$7E,$17,$C6,$5B,$F1,$9F,$EF,$BF,$86,$F8,$7F,$13,$F1,$1B,$5F,$25
	.dc.b	$B3,$C3,$0F,$F4,$8B,$6B,$E5,$FB,$D2,$FE,$87,$C2,$5A,$E1,$37,$4F
	.dc.b	$7E,$1F,$E7,$0B,$FA,$D3,$FA,$3F,$27,$F5,$51,$FE,$76,$7F,$D1,$06
	.dc.b	$2F,$D2,$E1,$E1,$DF,$C5,$B3,$C5,$A7,$02,$E1,$F4,$2B,$A3,$6B,$95
	.dc.b	$78,$5B,$5E,$06,$CE,$E7,$B1,$B2,$CE,$DB,$1B,$8F,$14,$6D,$A6,$CB
	.dc.b	$1B,$3C,$CF,$A2,$3F,$87,$2A,$F4,$C6,$1F,$88,$F5,$9E,$2D,$35,$3C
	.dc.b	$CF,$E3,$13,$C9,$B5,$EE,$9E,$8D,$8C,$C7,$B0,$C9,$5C,$9E,$A7,$2D
	.dc.b	$33,$5E,$47,$E0,$34,$F8,$11,$E6,$3D,$11,$E6,$3D,$07,$C1,$2D,$7A
	.dc.b	$5F,$89,$5E,$76,$CF,$95,$A7,$8C,$B5,$E1,$8F,$3B,$63,$CF,$1E,$37
	.dc.b	$CE,$D8,$F3,$C7,$96,$3C,$EF,$95,$B5,$E5,$3E,$AC,$72,$38,$14,$E0
	.dc.b	$57,$85,$7E,$4A,$3F,$02,$68,$B1,$FA,$56,$9F,$E1,$62,$4F,$F8,$69
	.dc.b	$A7,$23,$BD,$7B,$DF,$EC,$DD,$DF,$26,$BD,$11,$FD,$5E,$BF,$B3,$1F
	.dc.b	$2E,$32,$FF,$BB,$0C,$C7,$CD,$8C,$BF,$C4,$23,$35,$DE,$72,$FF,$22
	.dc.b	$8C,$9F,$34,$93,$F1,$A3,$55,$E5,$8F,$BD,$3C,$1D,$3C,$1F,$6B,$FA
	.dc.b	$83,$4E,$0D,$90,$F8,$2F,$B5,$EB,$1D,$E3,$F1,$47,$2A,$F4,$BD,$CC
	.dc.b	$4E,$CF,$81,$36,$BE,$74,$E8,$EC,$7C,$98,$D3,$CD,$B3,$E5,$8E,$0F
	.dc.b	$CD,$3E,$A4,$69,$F6,$1E,$A3,$51,$E8,$8D,$57,$E9,$E3,$9D,$7D,$3A
	.dc.b	$E7,$1E,$47,$F4,$CE,$8F,$99,$3C,$6B,$F5,$DE,$3B,$D7,$FF,$0E,$2F
	.dc.b	$66,$C6,$1B,$7F,$E2,$7C,$7F,$B0,$D6,$E3,$07,$F1,$E7,$A3,$8C,$7F
	.dc.b	$C5,$D1,$B7,$FA,$DB,$FE,$7E,$8D,$47,$F9,$DF,$CE,$F1,$FA,$DE,$46
	.dc.b	$4E,$AB,$85,$72,$AC,$BF,$97,$EE,$FD,$07,$83,$FA,$16,$DF,$E7,$DF
	.dc.b	$EF,$33,$F8,$D6,$BF,$D6,$9A,$FF,$94,$5B,$07,$A3,$FA,$9F,$02,$CF
	.dc.b	$B3,$CA,$F9,$7E,$6F,$CC,$6D,$7A,$5B,$3B,$8F,$64,$FD,$2F,$7E,$47
	.dc.b	$BF,$CA,$B5,$F2,$26,$ED,$F4,$A6,$D7,$E9,$FE,$07,$F8,$07,$C9,$7F
	.dc.b	$01,$6C,$7E,$FD,$F5,$91,$FE,$19,$87,$F9,$BE,$30,$FF,$01,$7F,$85
	.dc.b	$7B,$BC,$E7,$E9,$73,$7E,$20,$FC,$4F,$9B,$63,$24,$BF,$BF,$19,$7F
	.dc.b	$C3,$7F,$27,$FB,$CE,$3F,$E8,$4D,$8F,$F3,$73,$6B,$F0,$AD,$8F,$E2
	.dc.b	$9E,$87,$F1,$DF,$4B,$F8,$F3,$6B,$F8,$E3,$F6,$08,$FB,$83,$89,$E7
	.dc.b	$1F,$C8,$FC,$9E,$6F,$83,$CD,$F1,$9E,$4F,$3E,$C6,$49,$5E,$C8,$CB
	.dc.b	$EE,$E7,$F5,$F9,$DE,$C6,$D7,$B1,$B3,$B8,$F7,$CF,$FA,$3F,$E3,$DE
	.dc.b	$4D,$F2,$70,$DA,$FD,$4C,$E9,$FB,$6D,$BF,$E9,$0D,$7F,$A9,$B5,$FF
	.dc.b	$0A,$BB,$E6,$FF,$2E,$B6,$73,$E4,$F8,$1F,$F5,$B7,$CA,$7F,$B8,$C8
	.dc.b	$FE,$2B,$D1,$47,$F9,$83,$DB,$93,$F8,$EF,$1D,$18,$CB,$FA,$E7,$C9
	.dc.b	$3E,$43,$F6,$2D,$BF,$CA,$B6,$77,$3F,$EF,$3E,$C9,$5E,$16,$CF,$F6
	.dc.b	$7C,$38,$3F,$DC,$9A,$FF,$8D,$1A,$FF,$9F,$EB,$C3,$F2,$92,$FC,$2D
	.dc.b	$9F,$EE,$59,$BB,$7F,$83,$B6,$BE,$14,$60,$FF,$40,$FC,$A4,$8F,$84
	.dc.b	$DA,$F8,$3E,$EF,$C0,$24,$FC,$07,$F8,$9F,$AB,$FE,$89,$8F,$F5,$5F
	.dc.b	$A5,$82,$3F,$37,$E3,$3F,$D1,$E4,$70,$7E,$13,$D3,$FE,$3B,$C8,$69
	.dc.b	$23,$F8,$1F,$A5,$FE,$E3,$CB,$FC,$32,$D7,$FC,$B6,$D7,$FD,$D4,$D5
	.dc.b	$F2,$7F,$EC,$47,$FC,$C2,$D7,$FD,$E0,$D7,$7C,$DF,$CC,$3C,$AF,$C2
	.dc.b	$DF,$F1,$46,$CF,$F8,$85,$AF,$F8,$B1,$AF,$F9,$BB,$E7,$1E,$5F,$F3
	.dc.b	$A6,$D7,$F1,$E6,$CF,$F8,$AD,$AF,$F8,$DB,$E5,$1E,$17,$F4,$26,$D7
	.dc.b	$F2,$66,$CF,$F8,$CD,$AF,$F8,$FB,$E5,$1A,$7F,$4A,$6D,$7F,$2E,$6C
	.dc.b	$FF,$8E,$DA,$FF,$91,$BE,$51,$F3,$7F,$D4,$9B,$5F,$CD,$9B,$3F,$E4
	.dc.b	$36,$BC,$73,$E0,$49,$E0,$FF,$5C,$F9,$5E,$F3,$FE,$94,$6A,$0E,$1B
	.dc.b	$8F,$F6,$4F,$95,$76,$D2,$61,$FD,$2A,$32,$DF,$38,$FF,$7E,$8C,$DF
	.dc.b	$FB,$71,$4E,$A3,$15,$7C,$2F,$F7,$7F,$CA,$BC,$EC,$63,$FE,$AC,$8C
	.dc.b	$3F,$E0,$A5,$E7,$EB,$0F,$64,$69,$FF,$75,$BA,$9E,$13,$B9,$7C,$03
	.dc.b	$9B,$17,$B9,$FF,$79,$E2,$F8,$DE,$0D,$C7,$FC,$36,$BE,$33,$33,$A6
	.dc.b	$4F,$F8,$83,$1F,$F7,$65,$F4,$9F,$13,$F9,$6B,$FF,$0B,$F0,$E3,$D3
	.dc.b	$CB,$70,$FF,$14,$D2,$78,$FA,$99,$85,$FF,$88,$23,$35,$E1,$EE,$B5
	.dc.b	$A8,$FF,$4B,$EF,$E2,$3C,$7B,$D8,$EE,$E1,$F0,$CB,$E2,$3F,$15,$F2
	.dc.b	$AF,$5F,$FD,$43,$7E,$CA,$D1,$FE,$10,$29,$FE,$12,$AC,$38,$3F,$DD
	.dc.b	$3F,$29,$FC,$BF,$E5,$7F,$C2,$ED,$8F,$F5,$23,$5F,$F5,$7B,$5E,$3C
	.dc.b	$DF,$09,$DB,$5D,$B5,$DE,$DD,$BE,$47,$FD,$F2,$FE,$56,$15,$F8,$17
	.dc.b	$FC,$B2,$D7,$E5,$1F,$AB,$53,$CB,$4F,$5F,$25,$AF,$D8,$72,$7F,$08
	.dc.b	$7F,$09,$D3,$97,$BC,$DC,$3F,$19,$1E,$0F,$63,$06,$79,$AC,$79,$3D
	.dc.b	$8E,$06,$1F,$29,$EA,$7B,$62,$F8,$2A,$70,$9F,$F3,$AF,$13,$9B,$C8
	.dc.b	$FE,$CE,$33,$CB,$89,$CB,$D3,$D4,$93,$A3,$FC,$FF,$F2,$BF,$11,$BB
	.dc.b	$7B,$E3,$E5,$71,$7C,$F7,$53,$E8,$73,$C3,$FC,$5E,$1F,$28,$F9,$5F
	.dc.b	$90,$D8,$EA,$7F,$A3,$BF,$4E,$7C,$D7,$E4,$5B,$1F,$E6,$26,$BF,$EE
	.dc.b	$F6,$BF,$EF,$A5,$F1,$FC,$49,$1C,$5E,$9F,$33,$23,$A3,$5E,$9F,$31
	.dc.b	$E1,$23,$E1,$7C,$9F,$0F,$23,$C8,$D7,$C9,$F1,$1E,$92,$3E,$97,$D9
	.dc.b	$F0,$F2,$3D,$8D,$7D,$9F,$11,$F0,$48,$FC,$17,$E2,$72,$C8,$F8,$8D
	.dc.b	$7E,$27,$8B,$F2,$48,$FC,$97,$E6,$78,$B2,$3E,$63,$5F,$B0,$7F,$4F
	.dc.b	$EB,$FD,$03,$83,$F9,$47,$77,$F0,$58,$CC,$70,$6B,$F8,$15,$F3,$3B
	.dc.b	$24,$7F,$A7,$23,$2F,$DD,$EA,$FC,$01,$1C,$1F,$EA,$2F,$96,$7C,$B3
	.dc.b	$E4,$70,$D9,$F3,$4B,$DB,$FE,$9F,$18,$9D,$3E,$EF,$9A,$FC,$A7,$57
	.dc.b	$FA,$B9,$B1,$EF,$6B,$BE,$AF,$F6,$8F,$CE,$3E,$56,$92,$FF,$D6,$F1
	.dc.b	$97,$F3,$1E,$0F,$FB,$63,$AF,$FA,$F6,$1F,$97,$8C,$19,$F2,$E5,$7E
	.dc.b	$7D,$B3,$F9,$F3,$91,$23,$F4,$8F,$FB,$93,$F4,$E7,$C8,$D2,$4F,$FB
	.dc.b	$2A,$32,$DF,$8D,$E0,$47,$06,$FE,$3B,$A1,$1C,$1F,$D0,$FE,$67,$FB
	.dc.b	$A7,$A0,$F8,$7B,$18,$6C,$EC,$99,$7E,$2E,$95,$FE,$FC,$6C,$FF,$87
	.dc.b	$FC,$7F,$E7,$91,$8A,$FE,$4A,$FE,$0A,$74,$7F,$0F,$75,$3F,$D0,$F2
	.dc.b	$7C,$7F,$F8,$25,$B5,$C2,$FF,$C5,$90,$F8,$FD,$54,$AD,$52,$B8,$BF
	.dc.b	$CD,$9E,$11,$A7,$37,$E1,$F8,$1F,$07,$C9,$78,$5B,$5B,$6D,$F3,$6D
	.dc.b	$7F,$5B,$F9,$37,$CF,$7C,$EB,$FF,$11,$46,$1F,$2E,$9F,$E3,$A7,$CF
	.dc.b	$19,$C3,$F7,$0D,$7F,$7A,$D7,$F9,$B3,$5F,$F2,$73,$C6,$FC,$57,$E0
	.dc.b	$F9,$CF,$0B,$F2,$F9,$E6,$1F,$07,$D2,$B8,$FC,$59,$B8,$7D,$39,$B5
	.dc.b	$FD,$17,$5F,$F2,$7C,$FD,$7F,$92,$FA,$DE,$F7,$D7,$6D,$7C,$D6,$E3
	.dc.b	$F4,$5F,$87,$F2,$BF,$13,$E6,$ED,$7C,$26,$D7,$C2,$9F,$F0,$33,$F0
	.dc.b	$1A,$FC,$06,$AD,$91,$CB,$1F,$A9,$E8,$DC,$3D,$5E,$DF,$E3,$7E,$3F
	.dc.b	$F1,$CD,$8F,$C2,$37,$1F,$C4,$36,$FF,$2D,$EB,$7E,$54,$FC,$B7,$B9
	.dc.b	$6F,$F1,$91,$6F,$F2,$13,$6B,$F0,$DE,$8F,$E1,$3D,$2F,$C2,$B7,$0F
	.dc.b	$C1,$BE,$CD,$4D,$24,$7F,$3D,$C3,$E0,$47,$28,$FC,$5F,$A5,$67,$F4
	.dc.b	$1E,$DF,$EA,$FC,$DF,$D6,$36,$BF,$44,$D9,$DC,$7E,$96,$7F,$65,$F1
	.dc.b	$12,$3F,$51,$EE,$5A,$FD,$64,$DD,$3F,$75,$36,$BF,$2F,$E7,$FE,$8B
	.dc.b	$D2,$FD,$1B,$63,$F4,$2D,$C3,$FC,$57,$AF,$F8,$96,$7F,$4B,$5C,$1A
	.dc.b	$70,$2C,$F0,$AE,$58,$7E,$CD,$B3,$E2,$37,$19,$69,$99,$F0,$D7,$76
	.dc.b	$C7,$76,$9D,$DE,$2F,$EC,$AB,$BB,$63,$2F,$E5,$63,$2D,$8C,$DF,$BE
	.dc.b	$3A,$DF,$95,$B3,$E5,$8F,$E1,$D7,$E5,$6C,$79,$5B,$3E,$53,$89,$E3
	.dc.b	$29,$E1,$2D,$78,$63,$89,$E1,$2C,$F8,$4F,$E2,$35,$E1,$69,$E1,$2D
	.dc.b	$78,$67,$8B,$E1,$6B,$E1,$6D,$78,$67,$8C,$78,$5A,$78,$67,$F8,$FD
	.dc.b	$78,$5B,$1E,$16,$9E,$17,$8B,$E1,$6B,$E1,$6D,$78,$63,$8D,$78,$5B
	.dc.b	$3E,$17,$F8,$75,$78,$5B,$1E,$16,$9E,$13,$89,$F4,$4F,$E7,$77,$FB
	.dc.b	$D2,$47,$EE,$EF,$F0,$53,$A8,$C4,$EA,$3F,$A2,$BE,$19,$FF,$79,$5F
	.dc.b	$9E,$7D,$87,$BC,$FE,$8B,$E4,$6E,$B8,$36,$FD,$17,$C9,$F7,$3F,$0E
	.dc.b	$B9,$9F,$CB,$67,$83,$ED,$31,$3C,$AB,$AB,$4E,$B3,$C2,$FC,$D7,$DE
	.dc.b	$FC,$C7,$67,$C0,$65,$C5,$F4,$7C,$11,$A8,$FE,$DB,$5D,$6F,$D2,$DA
	.dc.b	$EB,$5C,$CE,$13,$E4,$3C,$D1,$D2,$78,$3E,$36,$CF,$91,$B5,$E0,$AE
	.dc.b	$37,$EC,$9C,$C7,$F7,$C6,$DF,$F1,$C6,$CF,$F8,$1D,$B3,$FD,$F1,$FC
	.dc.b	$67,$4F,$F9,$9E,$3F,$C2,$DF,$35,$FE,$23,$F4,$7F,$C4,$8D,$8C,$9F
	.dc.b	$E1,$B8,$CB,$FE,$22,$F0,$9F,$2C,$8C,$37,$2F,$F6,$CC,$BE,$6B,$19
	.dc.b	$FF,$BD,$63,$8D,$F0,$25,$70,$AD,$C7,$FB,$26,$77,$3D,$3E,$3B,$28
	.dc.b	$43,$DF,$CC,$62,$6C,$E6,$BD,$51,$FD,$FA,$0D,$24,$7F,$63,$8C,$BC
	.dc.b	$0F,$85,$F0,$DD,$27,$AF,$B1,$DB,$E4,$2D,$7C,$76,$E1,$F1,$1B,$9F
	.dc.b	$C3,$7C,$BF,$33,$2B,$E7,$7C,$D5,$BF,$97,$36,$FE,$54,$DA,$FF,$77
	.dc.b	$78,$4F,$9E,$F7,$FC,$2F,$C2,$8C,$4F,$D9,$8C,$4F,$F8,$7A,$73,$8F
	.dc.b	$FA,$F5,$AF,$FB,$09,$AF,$68,$E3,$5F,$83,$F9,$AB,$3F,$80,$F9,$6F
	.dc.b	$9F,$F3,$1F,$41,$B5,$F8,$96,$CE,$E3,$F1,$93,$F5,$BD,$49,$1F,$8F
	.dc.b	$F8,$CB,$5F,$8C,$9B,$AF,$E2,$E6,$D7,$E1,$FD,$FF,$BF,$EB,$7E,$01
	.dc.b	$B1,$F9,$97,$F0,$04,$7E,$F3,$70,$FC,$CB,$F8,$8F,$7F,$94,$FE,$87
	.dc.b	$E7,$BF,$17,$F0,$5F,$8C,$6C,$64,$FD,$34,$65,$A7,$FA,$2B,$E7,$7F
	.dc.b	$CD,$8F,$E6,$70,$46,$1F,$CE,$74,$FF,$9E,$D6,$9F,$AD,$77,$EA,$FE
	.dc.b	$7A,$B8,$46,$2F,$F6,$C6,$67,$5F,$27,$BF,$5E,$CF,$8D,$B5,$FE,$FD
	.dc.b	$B5,$7F,$98,$9B,$5F,$E4,$56,$CE,$E7,$8E,$1E,$59,$FF,$74,$B6,$3C
	.dc.b	$DC,$9C,$C9,$7C,$DE,$DE,$9C,$8E,$CD,$C3,$C9,$CD,$E3,$24,$F8,$DF
	.dc.b	$37,$A7,$23,$CC,$D9,$F6,$7C,$57,$AC,$93,$EB,$7D,$DF,$15,$23,$DC
	.dc.b	$D9,$FA,$C7,$F0,$F8,$7A,$BF,$E8,$66,$9F,$E6,$92,$C7,$03,$F9,$34
	.dc.b	$7F,$A3,$3E,$63,$F8,$94,$65,$AF,$66,$BF,$85,$6B,$FC,$7F,$E6,$7F
	.dc.b	$89,$BD,$5A,$F9,$9A,$FE,$35,$AF,$F2,$7F,$9C,$73,$FF,$A7,$9B,$87
	.dc.b	$FA,$05,$B3,$E9,$6B,$F0,$97,$2F,$12,$4F,$17,$A7,$34,$8E,$8D,$AF
	.dc.b	$89,$EF,$3C,$0F,$F5,$7B,$6B,$FD,$24,$D9,$DC,$EB,$C1,$95,$A6,$CF
	.dc.b	$4F,$07,$99,$27,$9B,$FE,$BC,$F9,$77,$8D,$49,$3F,$EA,$18,$CB,$7C
	.dc.b	$6E,$04,$70,$6F,$C5,$E8,$47,$07,$FA,$AF,$E5,$B9,$E3,$F8,$36,$BF
	.dc.b	$8C,$6B,$FC,$0D,$AF,$F5,$D7,$EC,$A6,$92,$BF,$DA,$EA,$A9,$2B,$C4
	.dc.b	$FF,$B7,$3E,$59,$EE,$D2,$4F,$FB,$06,$32,$D8,$FE,$D7,$18,$C3,$F8
	.dc.b	$9C,$FF,$19,$87,$C0,$6A,$BF,$B8,$FC,$B7,$F8,$BC,$9F,$11,$8F,$4F
	.dc.b	$2C,$C3,$FB,$D7,$83,$FC,$77,$E0,$BE,$3B,$FC,$7E,$D5,$FD,$F2,$1F
	.dc.b	$01,$4C,$7B,$77,$0E,$67,$F2,$D9,$46,$A3,$3A,$FF,$82,$71,$72,$66
	.dc.b	$3D,$A7,$F3,$5F,$8E,$FF,$05,$A6,$FB,$58,$6D,$F4,$9F,$F0,$A4,$F0
	.dc.b	$F6,$73,$1D,$0F,$F0,$C3,$99,$E1,$83,$CE,$C5,$6C,$FF,$2C,$13,$5F
	.dc.b	$1E,$FE,$46,$5F,$86,$7F,$C4,$9F,$09,$FE,$25,$F9,$67,$C8,$7F,$1A
	.dc.b	$BF,$5D,$2B,$FE,$00,$3C,$0A,$E1,$D1,$FC,$6F,$E5,$7F,$90,$FC,$B7
	.dc.b	$F1,$D8,$46,$1D,$78,$C9,$D4,$E6,$7F,$95,$AF,$87,$FE,$BB,$1D,$5A
	.dc.b	$E9,$AF,$F6,$78,$E1,$C9,$D3,$E1,$FE,$F6,$B9,$E6,$B7,$FE,$45,$BF
	.dc.b	$C3,$BF,$55,$56,$A3,$D3,$37,$85,$ED,$53,$57,$CA,$BE,$D3,$F8,$C7
	.dc.b	$A7,$C2,$68,$DF,$45,$8F,$07,$C1,$F2,$30,$F2,$7C,$1E,$1C,$8D,$9F
	.dc.b	$8B,$E3,$73,$F9,$D7,$35,$31,$16,$7F,$17,$EE,$FA,$CC,$3F,$91,$F9
	.dc.b	$6F,$82,$FE,$5D,$F6,$71,$7F,$83,$5B,$C5,$FE,$15,$77,$3C,$78,$7E
	.dc.b	$47,$CB,$3E,$4A,$9F,$2E,$B8,$D7,$F9,$D7,$F4,$FF,$E7,$CF,$E1,$DA
	.dc.b	$FF,$65,$6B,$FD,$B9,$AF,$F8,$97,$E4,$9E,$A7,$E5,$9B,$1F,$91,$6B
	.dc.b	$FD,$BD,$AF,$3F,$07,$F3,$B5,$FB,$E8,$F1,$46,$0E,$9E,$9C,$8E,$8D
	.dc.b	$7A,$7C,$BF,$84,$8F,$85,$EB,$C7,$23,$DC,$D7,$C9,$F1,$9E,$92,$3E
	.dc.b	$97,$D9,$F1,$92,$3D,$8D,$7E,$37,$C2,$FA,$88,$FC,$B7,$E2,$73,$48
	.dc.b	$F8,$8D,$7E,$27,$8D,$F2,$48,$FC,$97,$F8,$0F,$D6,$FE,$6E,$3F,$C3
	.dc.b	$0D,$7F,$C4,$2D,$7F,$CB,$0D,$7F,$D6,$5D,$5F,$A8,$7F,$82,$35,$FF
	.dc.b	$11,$35,$FF,$15,$B5,$FF,$31,$3F,$E0,$C7,$8F,$A9,$FC,$52,$31,$1C
	.dc.b	$A3,$F8,$BF,$D0,$31,$A4,$9F,$E2,$91,$97,$F8,$B7,$CC,$3E,$11,$F1
	.dc.b	$78,$6C,$FF,$8D,$E7,$FC,$4B,$E4,$E6,$BF,$8F,$7C,$C7,$E7,$FB,$2F
	.dc.b	$57,$F9,$CF,$0F,$F3,$D4,$E5,$FF,$20,$35,$FC,$E7,$C4,$ED,$FD,$23
	.dc.b	$57,$42,$30,$DB,$FC,$F7,$83,$BE,$AF,$E0,$31,$82,$5F,$E8,$73,$D1
	.dc.b	$23,$F4,$CF,$4F,$8A,$91,$FB,$36,$CF,$F9,$5F,$E3,$9F,$2F,$E5,$6D
	.dc.b	$7F,$3F,$6C,$EE,$75,$F2,$F2,$B4,$D9,$FF,$46,$7C,$CF,$FB,$D3,$4F
	.dc.b	$EB,$D5,$FE,$A3,$8C,$7A,$5F,$E9,$B7,$FD,$20,$48,$FF,$48,$36,$BF
	.dc.b	$B5,$7C,$4A,$3A,$7C,$77,$1A,$FE,$E5,$87,$DF,$6F,$2F,$87,$E3,$BC
	.dc.b	$4C,$8C,$9C,$23,$18,$FF,$77,$AF,$F2,$FC,$F4,$F8,$4F,$57,$57,$AD
	.dc.b	$B1,$93,$C3,$19,$7F,$C0,$9C,$5F,$EB,$6A,$7C,$1A,$73,$F1,$BF,$16
	.dc.b	$31,$8F,$F8,$1E,$D3,$F8,$12,$57,$F2,$16,$CE,$DF,$9D,$1C,$BC,$BF
	.dc.b	$F6,$45,$7F,$B1,$8F,$F5,$AC,$7F,$2D,$E9,$D3,$FC,$BF,$E3,$6E,$01
	.dc.b	$5C,$1C,$35,$CD,$70,$77,$36,$74,$3E,$33,$1F,$1B,$96,$E1,$F1,$9B
	.dc.b	$7F,$05,$B8,$FF,$B8,$7D,$A7,$07,$75,$C5,$63,$0D,$C3,$FC,$2D,$3C
	.dc.b	$1B,$3F,$B1,$8C,$61,$FE,$B7,$6D,$F7,$9E,$DC,$5C,$EB,$FC,$75,$D0
	.dc.b	$F9,$6F,$80,$DC,$7F,$C8,$4B,$E5,$F3,$32,$72,$F0,$8C,$57,$4F,$89
	.dc.b	$E7,$F2,$1D,$1B,$7D,$1B,$3B,$8E,$F3,$F4,$7E,$53,$F0,$F1,$FE,$53
	.dc.b	$D7,$F1,$CF,$F9,$56,$33,$5F,$E5,$39,$CB,$FE,$55,$8C,$BF,$EE,$06
	.dc.b	$BF,$83,$0E,$1F,$C3,$35,$CF,$27,$8C,$F2,$71,$DB,$F8,$BF,$09,$F2
	.dc.b	$0F,$93,$EE,$58,$C9,$2F,$E7,$C6,$5F,$A3,$F2,$FF,$43,$E1,$7E,$8B
	.dc.b	$6B,$E8,$B6,$77,$1F,$52,$7E,$37,$C8,$FC,$7F,$92,$B3,$F1,$5B,$5F
	.dc.b	$0D,$BC,$BF,$B4,$8F,$83,$18,$9F,$89,$AA,$31,$3F,$11,$AF,$09,$FA
	.dc.b	$57,$E1,$69,$E1,$2C,$78,$5B,$7E,$1B,$ED,$18,$3C,$37,$DA,$33,$5F
	.dc.b	$6E,$FB,$36,$BB,$34,$EC,$5B,$ED,$7C,$C9,$3D,$E7,$D8,$D7,$D8,$71
	.dc.b	$7C,$66,$6F,$C7,$5E,$9B,$F1,$B6,$7C,$6D,$AF,$1B,$4F,$1C,$73,$BF
	.dc.b	$1C,$F7,$3C,$65,$3C,$73,$C6,$7C,$71,$9B,$F1,$C7,$D8,$BF,$1B,$4F
	.dc.b	$19,$67,$C6,$DA,$F1,$C7,$3B,$F1,$C7,$78,$F1,$B5,$F1,$CF,$13,$C7
	.dc.b	$19,$BF,$1C,$FE,$AA,$FC,$6D,$AF,$1B,$4F,$19,$6B,$C6,$73,$BF,$CC
	.dc.b	$C7,$EE,$6B,$FC,$1D,$66,$92,$FF,$D7,$51,$96,$C6,$67,$F8,$01,$DE
	.dc.b	$3D,$4F,$F0,$0F,$5F,$F2,$13,$F9,$16,$CF,$E0,$5B,$3F,$45,$B8,$7C
	.dc.b	$16,$CF,$49,$FE,$1B,$18,$7C,$07,$78,$FC,$65,$77,$69,$DC,$B1,$DD
	.dc.b	$B5,$DD,$A7,$7B,$FF,$46,$B8,$9E,$E7,$89,$C3,$FC,$82,$BF,$97,$5F
	.dc.b	$96,$31,$3F,$CB,$63,$15,$F2,$E7,$F9,$64,$62,$B8,$D7,$4A,$D5,$FE
	.dc.b	$F6,$B5,$3C,$CD,$57,$0B,$FC,$D3,$F0,$5B,$1F,$05,$A7,$C1,$2D,$7C
	.dc.b	$17,$F4,$DC,$9F,$EF,$19,$FE,$AF,$F0,$9F,$D6,$BE,$0F,$FA,$DB,$63
	.dc.b	$27,$F5,$C8,$CB,$FD,$7B,$DC,$FE,$C2,$7F,$62,$F8,$3B,$19,$26,$1F
	.dc.b	$DA,$23,$2F,$F4,$FF,$5F,$FA,$C7,$CB,$FF,$59,$6D,$7F,$57,$7F,$32
	.dc.b	$47,$F2,$ED,$9E,$51,$FD,$77,$D8,$E4,$FF,$72,$F8,$2F,$EE,$BE,$B7
	.dc.b	$F7,$66,$C6,$4F,$EE,$F1,$96,$9F,$DE,$BE,$5F,$FB,$E9,$FD,$FB,$E0
	.dc.b	$EC,$64,$98,$7F,$81,$23,$2F,$F6,$BF,$83,$FE,$E9,$F1,$1F,$DD,$5B
	.dc.b	$5F,$DC,$DF,$DA,$91,$FD,$9B,$67,$94,$7F,$79,$F5,$39,$3F,$E0,$EF
	.dc.b	$5F,$FC,$23,$F2,$FF,$E1,$26,$C6,$4F,$F0,$9C,$65,$A7,$F8,$F7,$97
	.dc.b	$FA,$E7,$82,$F9,$7F,$F1,$D3,$67,$47,$F8,$05,$7C,$BE,$A6,$4E,$B2
	.dc.b	$FF,$24,$57,$15,$7D,$AF,$71,$8A,$E9,$F2,$1E,$0F,$88,$F0,$B6,$3A
	.dc.b	$4B,$04,$FF,$AA,$5B,$3F,$C5,$A7,$C3,$39,$24,$72,$9D,$CF,$84,$F3
	.dc.b	$FC,$97,$82,$75,$F1,$7F,$0B,$E4,$3E,$1B,$63,$27,$38,$CB,$4E,$DF
	.dc.b	$13,$F2,$8F,$97,$F2,$76,$32,$4C,$3C,$47,$B6,$34,$E3,$C0,$FF,$34
	.dc.b	$4F,$A7,$E2,$3F,$CD,$D1,$89,$E3,$9A,$73,$E9,$7F,$B3,$6B,$55,$FC
	.dc.b	$E9,$AF,$F8,$ED,$AF,$F9,$FD,$E9,$C1,$FE,$1F,$6E,$1F,$81,$F8,$5E
	.dc.b	$AF,$FA,$73,$C6,$FF,$6A,$11,$FA,$57,$FE,$88,$8F,$F6,$F7,$B9,$DA
	.dc.b	$0D,$26,$3A,$99,$5C,$A7,$FC,$37,$5F,$D4,$9A,$FF,$94,$FD,$6F,$E5
	.dc.b	$CF,$F5,$66,$B9,$C7,$FD,$D0,$78,$BC,$79,$7D,$5B,$3F,$89,$B7,$CB
	.dc.b	$FA,$09,$5E,$37,$FB,$5C,$C8,$FE,$D2,$D3,$F2,$4B,$C7,$FF,$7F,$34
	.dc.b	$F8,$A5,$8D,$C7,$F9,$35,$FF,$37,$7C,$BB,$9F,$11,$FE,$B3,$8C,$57
	.dc.b	$CE,$6C,$7E,$85,$AF,$F7,$8F,$9F,$7C,$1F,$28,$46,$2B,$EB,$B6,$3F
	.dc.b	$54,$D7,$FC,$03,$F2,$EF,$80,$E4,$DA,$FB,$ED,$9F,$D9,$35,$E3,$F0
	.dc.b	$1C,$09,$3C,$1F,$27,$CF,$7A,$27,$FD,$1C,$D7,$FD,$A4,$D4,$DF,$61
	.dc.b	$D3,$ED,$6D,$FE,$49,$B5,$FC,$15,$AF,$F8,$71,$75,$71,$24,$F1,$7E
	.dc.b	$47,$CE,$BC,$7A,$49,$FC,$FC,$65,$FD,$0F,$CC,$BC,$0F,$A0,$D8,$CC
	.dc.b	$60,$CF,$83,$23,$8B,$72,$FE,$39,$F3,$AF,$90,$FB,$4D,$AF,$D7,$B6
	.dc.b	$77,$3F,$7F,$E6,$BC,$11,$FC,$95,$AF,$F8,$F1,$AF,$FA,$F5,$AF,$FC
	.dc.b	$01,$7F,$93,$BB,$BE,$84,$BF,$DF,$FC,$D3,$D8,$47,$F3,$88,$7B,$58
	.dc.b	$BF,$E3,$F1,$B9,$D4,$E7,$C3,$FF,$57,$4E,$6B,$F9,$E3,$5F,$F2,$9B
	.dc.b	$5F,$F6,$A3,$50,$F9,$A7,$9B,$D7,$0F,$F6,$D4,$61,$FE,$9A,$BE,$37
	.dc.b	$55,$CA,$34,$FF,$36,$75,$3C,$27,$73,$EE,$DE,$D6,$A3,$F9,$EE,$2F
	.dc.b	$63,$6D,$C3,$FA,$EA,$F4,$F3,$3A,$64,$FF,$62,$D3,$F2,$EF,$08,$F0
	.dc.b	$DB,$E4,$E9,$8F,$4E,$D7,$F6,$A9,$46,$63,$1E,$9D,$9F,$EA,$61,$EA
	.dc.b	$7B,$B0,$7C,$66,$2B,$A6,$FE,$23,$C7,$BD,$8F,$03,$87,$C5,$83,$E0
	.dc.b	$BE,$BF,$CD,$7E,$76,$3E,$C9,$F2,$77,$FE,$43,$3F,$D3,$23,$EE,$B5
	.dc.b	$E4,$65,$FE,$A3,$F3,$3F,$CC,$7E,$6A,$F9,$06,$92,$BF,$C1,$12,$F7
	.dc.b	$FF,$C2,$F1,$86,$BF,$EA,$28,$D6,$1D,$A7,$33,$E8,$97,$BF,$EA,$AF
	.dc.b	$D0,$4A,$7F,$04,$FF,$87,$1A,$FF,$80,$9E,$2E,$79,$39,$69,$F1,$38
	.dc.b	$1E,$B6,$2B,$F3,$F1,$87,$81,$D7,$97,$4F,$87,$DC,$B8,$9B,$38,$73
	.dc.b	$E0,$CA,$F3,$45,$7F,$C2,$F7,$FE,$0F,$7F,$29,$EA,$E7,$4F,$7F,$7F
	.dc.b	$E2,$67,$FC,$72,$FE,$C9,$BC,$BE,$48,$D4,$7F,$88,$9D,$3D,$BC,$DF
	.dc.b	$4E,$1F,$D8,$E3,$89,$FE,$44,$F9,$9F,$8E,$FF,$8A,$9F,$97,$87,$D0
	.dc.b	$6E,$FF,$8C,$77,$19,$E0,$F8,$BF,$34,$F8,$EA,$7C,$8A,$E3,$87,$F0
	.dc.b	$CF,$6B,$FC,$A4,$FF,$87,$9A,$FF,$94,$5A,$FF,$BA,$1A,$FE,$73,$E4
	.dc.b	$3F,$AB,$BF,$E2,$66,$BF,$E5,$96,$BF,$EF,$06,$BF,$A4,$87,$95,$FA
	.dc.b	$36,$C7,$8E,$30,$7E,$9E,$7F,$AE,$C7,$7F,$2B,$04,$7F,$C6,$FE,$BB
	.dc.b	$E6,$BF,$5C,$D8,$CC,$60,$F0,$6F,$23,$E1,$35,$F5,$71,$F8,$48,$FB
	.dc.b	$DE,$BF,$07,$23,$C2,$D7,$E5,$7C,$47,$B4,$8F,$CF,$7D,$9E,$EC,$8F
	.dc.b	$94,$D7,$FD,$05,$8F,$F0,$B6,$F2,$FB,$08,$FD,$9F,$2B,$E5,$11,$C1
	.dc.b	$FE,$5A,$F9,$9F,$E6,$8D,$EB,$FE,$25,$18,$8F,$F4,$83,$5F,$E3,$27
	.dc.b	$F3,$BE,$DF,$E3,$13,$FB,$58,$FF,$4B,$35,$DB,$5F,$E4,$3E,$AB,$E3
	.dc.b	$3F,$90,$36,$3F,$CD,$AD,$7F,$D4,$0D,$7C,$3F,$32,$D3,$F9,$43,$63
	.dc.b	$FC,$EA,$D7,$FD,$50,$D7,$D3,$F3,$2F,$90,$7C,$3F,$F9,$F2,$30,$DC
	.dc.b	$3F,$B4,$57,$FA,$C6,$72,$FE,$8F,$E3,$7F,$CA,$1D,$26,$93,$BF,$F4
	.dc.b	$5F,$31,$FE,$55,$EA,$34,$91,$F9,$6F,$0B,$FC,$B5,$CB,$FD,$D3,$1F
	.dc.b	$F4,$04,$60,$EF,$E4,$CA,$FD,$B3,$67,$F5,$DE,$F7,$72,$4F,$EC,$DF
	.dc.b	$F3,$37,$C6,$BE,$62,$92,$7F,$D3,$B1,$97,$F8,$17,$89,$FE,$02,$DB
	.dc.b	$FA,$54,$7F,$B6,$1F,$EF,$D8,$7F,$AF,$9C,$B6,$7E,$29,$27,$99,$23
	.dc.b	$4F,$3F,$2B,$B4,$F7,$F2,$24,$F7,$E8,$B7,$DE,$6F,$8F,$FD,$4B,$1F
	.dc.b	$F0,$75,$F9,$67,$D1,$F1,$9E,$9F,$2A,$CF,$A5,$B8,$7A,$5B,$87,$37
	.dc.b	$FC,$43,$C6,$F8,$AF,$85,$F1,$98,$6E,$1F,$EF,$A9,$D3,$FC,$86,$3F
	.dc.b	$00,$47,$06,$DF,$99,$D3,$CE,$39,$79,$9E,$03,$C3,$F1,$96,$32,$4C
	.dc.b	$FF,$DD,$B8,$27,$51,$8F,$6F,$D3,$ED,$7A,$9B,$7E,$76,$CE,$E3,$D5
	.dc.b	$3F,$E4,$0F,$87,$7C,$BF,$C2,$F6,$30,$DA,$FF,$29,$F0,$7F,$80,$AB
	.dc.b	$71,$FE,$DD,$9F,$F7,$29,$1C,$1C,$0F,$F1,$1C,$70,$C3,$D7,$F2,$BD
	.dc.b	$A7,$E6,$78,$5F,$48,$FA,$7F,$2B,$63,$24,$AF,$87,$19,$7E,$2F,$C3
	.dc.b	$7D,$AF,$13,$ED,$B6,$BE,$C3,$67,$71,$F6,$E7,$FC,$C3,$E0,$BE,$53
	.dc.b	$F0,$7F,$2F,$86,$D7,$AE,$FF,$CE,$B5,$B8,$FE,$A3,$3F,$BC,$23,$83
	.dc.b	$84,$A7,$83,$F7,$23,$B1,$A8,$FE,$F4,$DF,$5B,$CB,$F2,$BE,$66,$46
	.dc.b	$4E,$11,$8A,$70,$7F,$8E,$2D,$F2,$3E,$EF,$97,$F7,$B7,$0F,$03,$6F
	.dc.b	$8C,$7F,$96,$AB,$8B,$E5,$AF,$F4,$3F,$BC,$DE,$F0,$E1,$B5,$F0,$27
	.dc.b	$4D,$9D,$34,$D4,$F0,$C3,$C2,$7F,$99,$9C,$65,$FE,$52,$BE,$73,$A3
	.dc.b	$FC,$E5,$7D,$1C,$42,$31,$3C,$5B,$7F,$E7,$69,$D3,$E2,$3C,$95,$CA
	.dc.b	$FC,$31,$B6,$C7,$96,$37,$7E,$59,$CC,$FF,$A1,$5B,$5E,$89,$F1,$57
	.dc.b	$7A,$F3,$DF,$91,$FF,$42,$B5,$F6,$C6,$E7,$8C,$79,$63,$C2,$79,$4A
	.dc.b	$F9,$4E,$AF,$13,$E0,$4F,$5A,$F8,$33,$CC,$D9,$6B,$B9,$ED,$AE,$B1
	.dc.b	$ED,$6C,$FB,$5E,$A7,$78,$F5,$B5,$F5,$9D,$5D,$B4,$D9,$4D,$CF,$85
	.dc.b	$F2,$B4,$F2,$C7,$C1,$8C,$B6,$33,$1F,$0D,$AF,$A4,$D9,$98,$F1,$C7
	.dc.b	$56,$9C,$CB,$1B,$3C,$F1,$B6,$C6,$DA,$7D,$33,$FC,$5B,$3F,$9A,$6D
	.dc.b	$FF,$AB,$9F,$51,$89,$FF,$19,$BC,$4C,$90,$F1,$1E,$BA,$FD,$37,$0F
	.dc.b	$DD,$AF,$BC,$DA,$FA,$57,$FE,$25,$6C,$E9,$A6,$A7,$85,$FD,$3A,$E0
	.dc.b	$DA,$E0,$DA,$E0,$72,$78,$57,$2B,$F8,$D1,$FE,$E9,$8C,$4F,$C7,$8C
	.dc.b	$3C,$67,$77,$E0,$6D,$F8,$1B,$3E,$0A,$F1,$1E,$07,$91,$D6,$B9,$36
	.dc.b	$B9,$34,$E5,$5D,$2B,$91,$D1,$F1,$DF,$A2,$3E,$8D,$F4,$6D,$F4,$6D
	.dc.b	$F4,$3B,$3D,$2F,$E1,$F8,$8F,$97,$57,$F4,$88,$C3,$7C,$FC,$DF,$F2
	.dc.b	$19,$CB,$6B,$37,$FE,$F8,$9C,$B6,$B3,$6B,$95,$A8,$C4,$23,$FD,$6F
	.dc.b	$1E,$3F,$94,$E3,$3C,$3D,$BA,$EE,$2D,$70,$8F,$34,$62,$72,$F0,$F9
	.dc.b	$7F,$EC,$78,$FF,$3D,$C3,$1C,$9F,$D3,$9A,$7E,$8B,$05,$7F,$6A,$7F
	.dc.b	$D2,$ED,$7D,$2D,$7D,$CF,$68,$CC,$F8,$78,$FA,$D7,$CB,$F5,$FD,$27
	.dc.b	$AB,$DA,$B7,$F4,$E2,$CF,$D1,$9B,$8F,$CF,$F4,$7E,$0F,$AF,$F0,$9B
	.dc.b	$87,$DB,$6D,$FD,$96,$D7,$DC,$8F,$6F,$CA,$7C,$5F,$95,$F8,$CD,$BF
	.dc.b	$F0,$5E,$A8,$FE,$11,$5E,$E6,$BC,$A7,$E4,$4E,$DB,$3D,$1E,$7F,$2B
	.dc.b	$6B,$F1,$9F,$07,$F6,$BD,$4F,$B6,$DA,$FC,$A3,$67,$71,$F9,$79,$FC
	.dc.b	$0F,$A3,$23,$F3,$5E,$95,$AF,$CC,$4D,$D3,$F2,$D3,$6B,$EC,$FB,$BF
	.dc.b	$85,$F9,$7F,$C3,$36,$3E,$0B,$EE,$D4,$D2,$47,$C0,$78,$BF,$23,$E5
	.dc.b	$EE,$5F,$A9,$F4,$FF,$2F,$EA,$FE,$61,$B5,$FB,$16,$CE,$E3,$F6,$93
	.dc.b	$F9,$EF,$52,$47,$ED,$FD,$4B,$5F,$B5,$9B,$AF,$EC,$E6,$D7,$E5,$BE
	.dc.b	$27,$F4,$5F,$15,$FA,$36,$C7,$DE,$7D,$C4,$7E,$DB,$70,$FF,$11,$6A
	.dc.b	$8F,$D6,$7B,$5C,$EB,$FD,$31,$EC,$7F,$B5,$48,$FE,$AF,$D4,$47,$FB
	.dc.b	$53,$A2,$4F,$EB,$FC,$94,$63,$2F,$EA,$7E,$7F,$F9,$11,$BE,$5F,$E4
	.dc.b	$23,$11,$FC,$28,$38,$7F,$86,$E2,$F7,$FF,$DD,$04,$C7,$C8,$FF,$B9
	.dc.b	$BB,$24,$78,$DB,$5E,$2E,$CF,$01,$27,$F8,$36,$08,$FD,$1F,$CD,$BD
	.dc.b	$4A,$4A,$FE,$19,$19,$7F,$4F,$7F,$8D,$23,$83,$F8,$90,$ED,$FD,$7B
	.dc.b	$4C,$B8,$3F,$67,$F3,$F9,$9F,$EF,$6D,$7E,$33,$5F,$E6,$2D,$7F,$9B
	.dc.b	$7C,$BB,$DE,$FD,$EB,$6F,$F9,$23,$67,$FC,$0E,$D7,$E9,$7C,$BB,$D5
	.dc.b	$FE,$0A,$DA,$FE,$58,$D9,$FF,$07,$B5,$FB,$5B,$BD,$FF,$E1,$AD,$AF
	.dc.b	$E6,$8D,$9F,$F0,$BB,$5F,$C1,$7C,$BB,$DD,$FE,$2A,$DA,$FE,$78,$D9
	.dc.b	$FF,$0F,$B5,$FC,$57,$9C,$F9,$7F,$E3,$AD,$AF,$E8,$8D,$9F,$F1,$3B
	.dc.b	$5F,$C9,$7C,$FB,$E2,$3F,$92,$B6,$BF,$A6,$36,$7F,$C5,$ED,$7F,$35
	.dc.b	$F3,$0F,$07,$F9,$6B,$6B,$FA,$A3,$67,$FC,$6E,$D7,$F4,$5F,$31,$F8
	.dc.b	$D7,$FC,$76,$D7,$3C,$DF,$CE,$23,$11,$FD,$CB,$E6,$3F,$D8,$6F,$F9
	.dc.b	$09,$AF,$EB,$5A,$EF,$DD,$FE,$83,$1E,$98,$E5,$9E,$09,$8F,$F6,$BF
	.dc.b	$98,$EF,$3F,$E0,$36,$A6,$98,$F9,$87,$A7,$FD,$39,$B5,$FD,$D5,$B5
	.dc.b	$FE,$58,$6B,$FC,$0D,$7A,$7F,$E0,$87,$7E,$4F,$F6,$17,$35,$FE,$E8
	.dc.b	$72,$FF,$85,$3E,$61,$E3,$FF,$84,$1B,$1F,$EE,$E5,$F2,$F9,$9D,$32
	.dc.b	$7F,$DE,$79,$7D,$3C,$3F,$B8,$5F,$09,$FF,$7B,$C3,$D3,$FF,$0E,$B6
	.dc.b	$33,$E7,$62,$72,$C8,$EB,$F2,$57,$5F,$F6,$B0,$F9,$8F,$F1,$3B,$67
	.dc.b	$FC,$65,$87,$FA,$07,$CF,$CD,$69,$93,$A3,$F9,$67,$9D,$8B,$EA,$F4
	.dc.b	$3F,$C6,$7A,$3E,$1F,$15,$C6,$BF,$C7,$BB,$F8,$8F,$1E,$F6,$75,$3F
	.dc.b	$E4,$58,$CC,$62,$3F,$C9,$1F,$03,$FE,$49,$F9,$8F,$D6,$47,$F9,$22
	.dc.b	$BD,$AA,$7F,$C7,$0F,$02,$B8,$74,$7F,$A7,$BE,$5F,$F9,$67,$CC,$7F
	.dc.b	$8F,$DB,$19,$F2,$93,$96,$C7,$1F,$6D,$C3,$C5,$B1,$98,$C1,$D7,$DB
	.dc.b	$FF,$2F,$BF,$E3,$1C,$3F,$01,$3F,$09,$AF,$4D,$3F,$CC,$B0,$AE,$31
	.dc.b	$8E,$5B,$5F,$E6,$DE,$0F,$EB,$CF,$E4,$3C,$6C,$AC,$3A,$E3,$E1,$E6
	.dc.b	$CC,$3D,$8A,$62,$2D,$79,$7C,$2D,$98,$7C,$C7,$F9,$F6,$BF,$8C,$3F
	.dc.b	$A0,$F9,$5F,$0E,$47,$8C,$7E,$03,$BB,$FC,$14,$DE,$5F,$3C,$7E,$C2
	.dc.b	$35,$CF,$89,$E7,$E4,$49,$E0,$7E,$2B,$E5,$FE,$33,$78,$3E,$0C,$7C
	.dc.b	$0E,$1F,$F4,$7B,$A9,$F4,$87,$4F,$FA,$5F,$E6,$1E,$1F,$CA,$6C,$75
	.dc.b	$3F,$33,$FA,$97,$FA,$AD,$FF,$44,$35,$FF,$75,$B5,$FF,$0B,$35,$FF
	.dc.b	$50,$AF,$84,$FC,$CB,$63,$9C,$60,$CF,$95,$23,$BB,$5E,$FC,$1C,$88
	.dc.b	$F9,$1E,$BF,$05,$23,$CE,$D7,$CF,$E2,$F8,$88,$FB,$1F,$2F,$C1,$48
	.dc.b	$F7,$B5,$F7,$F6,$7A,$88,$FC,$47,$DB,$D9,$23,$E3,$B5,$F8,$FD,$9F
	.dc.b	$08,$8F,$CC,$7E,$2F,$0C,$8F,$9E,$D7,$E7,$F4,$FC,$A2,$3F,$51,$F9
	.dc.b	$BD,$52,$3E,$BB,$5F,$AF,$D5,$F4,$88,$FD,$C7,$EA,$FB,$F2,$3E,$FB
	.dc.b	$5F,$BF,$F2,$7F,$68,$8F,$E1,$1F,$BB,$F2,$52,$3F,$0E,$D7,$FC,$B4
	.dc.b	$3E,$55,$F1,$89,$FF,$6A,$46,$1B,$EB,$7F,$52,$AE,$33,$97,$9F,$CE
	.dc.b	$FF,$78,$7F,$DE,$4D,$4D,$31,$F4,$0E,$4F,$F5,$EB,$7B,$3F,$DC,$8D
	.dc.b	$AF,$3B,$5F,$F3,$97,$8E,$F9,$FF,$F6,$2B,$6B,$FD,$D8,$D9,$F7,$B5
	.dc.b	$FF,$3D,$7D,$C7,$F8,$ED,$11,$C3,$7D,$9F,$F6,$8C,$62,$1F,$A9,$3E
	.dc.b	$67,$FD,$AA,$DC,$3F,$DF,$8D,$9F,$9E,$D7,$C3,$F2,$09,$FE,$AB,$6F
	.dc.b	$A2,$C5,$7F,$08,$85,$63,$CB,$FF,$76,$46,$EF,$8C,$F9,$88,$FD,$EA
	.dc.b	$4F,$28,$C3,$63,$B4,$F5,$9F,$87,$1C,$3E,$4B,$0F,$C5,$9A,$FC,$2F
	.dc.b	$1E,$E4,$F9,$AB,$18,$6E,$1F,$DC,$A7,$4D,$9D,$39,$F9,$AE,$67,$4F
	.dc.b	$72,$C6,$49,$87,$F2,$CB,$FF,$10,$47,$A1,$57,$9B,$3F,$F1,$8D,$7C
	.dc.b	$BE,$3C,$78,$79,$6E,$1F,$12,$3E,$F9,$1C,$1F,$59,$E9,$E1,$F0,$AF
	.dc.b	$17,$A1,$F0,$BE,$37,$E1,$B6,$FE,$03,$67,$71,$F0,$E7,$E4,$FB,$12
	.dc.b	$3E,$37,$C0,$5A,$F7,$4D,$E3,$FC,$2C,$7E,$4A,$71,$5F,$63,$47,$87
	.dc.b	$F0,$7C,$4F,$84,$D9,$FC,$7B,$F9,$C2,$3F,$84,$6D,$FE,$45,$F9,$5E
	.dc.b	$27,$29,$FB,$5F,$3B,$F8,$03,$F0,$3E,$9D,$8C,$92,$FE,$B4,$65,$FB
	.dc.b	$1F,$0B,$F8,$8F,$9C,$FC,$4B,$6B,$F0,$CD,$9D,$C7,$E2,$67,$F2,$3E
	.dc.b	$C7,$EB,$76,$FD,$1F,$8B,$8E,$EF,$C0,$36,$FF,$45,$5F,$B1,$C3,$F3
	.dc.b	$AD,$7F,$C8,$35,$F3,$1C,$BC,$5F,$EF,$31,$F9,$86,$C7,$E8,$3E,$62
	.dc.b	$57,$E7,$7D,$BB,$5F,$95,$9B,$8F,$D8,$EA,$FF,$2D,$5F,$E8,$E7,$F5
	.dc.b	$3E,$47,$EA,$BE,$62,$CF,$E9,$DB,$87,$E7,$DB,$87,$37,$E1,$F4,$7C
	.dc.b	$4E,$8B,$7F,$E6,$0D,$51,$8C,$3D,$6D,$C3,$F1,$73,$AA,$E1,$7C,$DA
	.dc.b	$73,$AE,$0F,$CA,$6D,$FF,$9A,$9A,$FF,$99,$DB,$5D,$63,$E5,$D7,$56
	.dc.b	$D7,$53,$BB,$E6,$6D,$F9,$AF,$99,$23,$27,$8C,$B3,$E3,$6B,$E3,$6D
	.dc.b	$F8,$EB,$E0,$DF,$8D,$B3,$E3,$AF,$29,$E3,$2C,$78,$E3,$9D,$F8,$E7
	.dc.b	$31,$E3,$69,$E3,$29,$E3,$2D,$F8,$E3,$F2,$B7,$E3,$69,$E3,$BF,$2D
	.dc.b	$78,$DA,$78,$DE,$77,$E3,$8C,$CF,$8D,$B3,$E3,$6B,$E3,$6D,$F8,$E3
	.dc.b	$FD,$4B,$7E,$36,$D7,$8C,$F2,$BE,$36,$D7,$8C,$E7,$7E,$38,$C9,$E3
	.dc.b	$2C,$F8,$DE,$6F,$F5,$23,$FD,$61,$5F,$B5,$7C,$27,$F8,$6A,$D1,$98
	.dc.b	$DC,$FF,$00,$8C,$4E,$A3,$04,$9F,$A3,$CF,$F9,$6A,$FC,$BB,$67,$F0
	.dc.b	$ED,$9D,$C7,$E2,$A7,$6D,$36,$58,$D9,$F7,$A3,$6D,$8D,$B4,$DE,$1F
	.dc.b	$4A,$76,$D9,$DB,$67,$71,$F2,$9F,$82,$60,$F8,$F1,$B3,$F7,$F5,$EB
	.dc.b	$7E,$FB,$67,$95,$A3,$15,$CA,$7F,$94,$C6,$23,$BD,$74,$9F,$80,$F4
	.dc.b	$69,$D2,$78,$57,$D3,$AF,$E5,$B1,$8B,$FE,$7A,$DA,$FE,$34,$FF,$0C
	.dc.b	$D1,$F3,$7F,$E1,$03,$FC,$23,$E5,$5A,$FF,$07,$45,$C3,$FC,$0D,$37
	.dc.b	$2F,$E6,$55,$FE,$18,$F9,$BF,$F0,$CB,$4F,$F0,$B1,$6B,$FC,$1C,$D9
	.dc.b	$FE,$E3,$FA,$97,$FA,$E2,$3F,$C9,$31,$8F,$9A,$FF,$12,$CE,$6B,$FD
	.dc.b	$82,$A3,$3E,$CF,$E5,$E3,$FC,$CC,$47,$06,$AF,$A9,$FE,$34,$87,$CD
	.dc.b	$66,$FA,$FC,$A7,$80,$F0,$FC,$E5,$BE,$33,$FE,$5E,$B5,$66,$3C,$B5
	.dc.b	$72,$F6,$FC,$A7,$AB,$E5,$BD,$6D,$C7,$9D,$B1,$4E,$63,$DE,$DB,$FF
	.dc.b	$0B,$FC,$9F,$F7,$7B,$F5,$7C,$A7,$F9,$2A,$31,$7E,$58,$C7,$5F,$EE
	.dc.b	$E7,$FB,$B4,$9A,$48,$DD,$79,$31,$7D,$16,$2F,$E8,$FA,$7F,$50,$FA
	.dc.b	$BF,$43,$6F,$84,$59,$FA,$F3,$73,$FB,$FF,$39,$F7,$3E,$77,$EE,$B7
	.dc.b	$0F,$03,$6F,$EF,$B6,$BF,$0F,$F1,$1F,$84,$F9,$DF,$C2,$B6,$FD,$0F
	.dc.b	$E0,$C8,$FF,$9E,$74,$46,$30,$FC,$5B,$F8,$EF,$9F,$E7,$5F,$07,$E3
	.dc.b	$BF,$22,$7E,$4B,$E3,$EC,$64,$97,$F9,$48,$CB,$F9,$6F,$8B,$FC,$CF
	.dc.b	$C6,$7E,$69,$B5,$F3,$5B,$3B,$8F,$A1,$3F,$E8,$4F,$0D,$EE,$7E,$8B
	.dc.b	$DD,$C3,$6B,$F1,$33,$FE,$AB,$24,$7F,$39,$6C,$6A,$3F,$57,$7C,$A3
	.dc.b	$8F,$D1,$EC,$8F,$FA,$7E,$91,$CF,$DE,$93,$96,$D7,$FB,$8A,$BF,$D4
	.dc.b	$4F,$F4,$F6,$BF,$EA,$26,$BF,$EC,$66,$AF,$1D,$EA,$FF,$4E,$25,$FF
	.dc.b	$94,$1F,$67,$BD,$23,$CC,$DA,$FF,$79,$F5,$BD,$CF,$D0,$12,$BF,$D7
	.dc.b	$51,$97,$FD,$79,$F3,$8E,$EF,$EC,$0D,$8C,$C6,$0C,$F6,$C8,$E2,$DC
	.dc.b	$3F,$B7,$0A,$76,$74,$25,$78,$1F,$ED,$DF,$34,$F3,$E9,$27,$FD,$A1
	.dc.b	$19,$6F,$9D,$C0,$8E,$0D,$F9,$FD,$08,$E0,$FF,$6D,$7C,$5F,$FA,$F1
	.dc.b	$BE,$F7,$F7,$F8,$C6,$7E,$A6,$BF,$E0,$97,$FC,$A0,$DE,$3E,$04,$AE
	.dc.b	$BE,$8F,$32,$38,$3F,$DD,$5F,$2F,$FC,$48,$9E,$37,$F8,$3D,$C7,$C0
	.dc.b	$D2,$4F,$07,$5C,$92,$39,$B7,$1F,$A3,$F2,$DF,$C7,$89,$87,$F8,$6D
	.dc.b	$C5,$D2,$5F,$FB,$F6,$3F,$D2,$95,$FE,$27,$7F,$0C,$F4,$BE,$44,$9F
	.dc.b	$BF,$0F,$96,$69,$FE,$29,$6C,$E9,$B3,$F8,$66,$BF,$E3,$57,$3E,$37
	.dc.b	$FB,$51,$B2,$99,$0F,$8F,$4F,$F4,$A8,$DC,$EF,$E5,$9E,$16,$A1,$87
	.dc.b	$FB,$82,$73,$7F,$D4,$97,$83,$B9,$D5,$DA,$F1,$5A,$F7,$1C,$6F,$53
	.dc.b	$E7,$F9,$07,$C3,$62,$38,$4D,$C5,$F0,$54,$C7,$C1,$5C,$7F,$B2,$CF
	.dc.b	$FB,$C9,$B1,$CF,$1F,$E0,$F1,$83,$FB,$5E,$0F,$1F,$C0,$D8,$CF,$91
	.dc.b	$89,$CB,$23,$8F,$91,$6B,$E3,$C7,$F9,$A6,$39,$9E,$1F,$72,$C7,$83
	.dc.b	$7E,$E7,$83,$5B,$1E,$47,$0F,$9A,$FF,$90,$0F,$96,$FE,$AA,$FC,$E9
	.dc.b	$F4,$BA,$3F,$DE,$27,$E4,$0C,$1F,$E7,$93,$F8,$8F,$CA,$BE,$49,$F2
	.dc.b	$DF,$96,$8C,$36,$33,$3C,$1A,$FF,$A1,$E5,$F2,$7F,$88,$8E,$AD,$74
	.dc.b	$D7,$F0,$B1,$C3,$87,$C1,$F2,$7F,$88,$7F,$25,$29,$FC,$0B,$FE,$22
	.dc.b	$6B,$F4,$0F,$D0,$29,$E5,$9F,$6E,$3B,$5F,$8F,$EE,$AB,$8F,$FD,$36
	.dc.b	$F5,$9C,$E0,$E4,$BF,$11,$2B,$C5,$15,$F1,$1E,$7F,$93,$E2,$67,$4F
	.dc.b	$5D,$A3,$C3,$8B,$E2,$29,$C2,$FF,$C7,$FE,$A3,$9B,$CE,$FE,$DA,$33
	.dc.b	$CB,$89,$CB,$D3,$BA,$4E,$8F,$F2,$4F,$CB,$7C,$56,$EF,$F0,$A3,$E5
	.dc.b	$EB,$F4,$1D,$4F,$B1,$CF,$0F,$EF,$3E,$59,$F0,$3F,$29,$B5,$D0,$FF
	.dc.b	$2B,$FB,$C3,$F2,$45,$7F,$91,$B5,$FE,$C6,$D7,$FD,$A5,$F2,$CF,$4F
	.dc.b	$F3,$4D,$8F,$E5,$6D,$7F,$B5,$B5,$FF,$6D,$4F,$F9,$99,$EF,$EF,$F8
	.dc.b	$88,$F2,$F1,$7A,$91,$C1,$DB,$8F,$C2,$47,$07,$AB,$D3,$F6,$91,$C1
	.dc.b	$E6,$F8,$FF,$49,$1C,$1E,$AF,$8F,$F6,$91,$C1,$F1,$BD,$5F,$92,$47
	.dc.b	$07,$C2,$EE,$F8,$A4,$70,$7C,$6E,$EF,$92,$47,$07,$D2,$F5,$3E,$A9
	.dc.b	$1C,$1F,$3B,$B7,$E8,$91,$C1,$F4,$BB,$BE,$A9,$1C,$1F,$7B,$E0,$7F
	.dc.b	$02,$47,$07,$DA,$F9,$6F,$BA,$47,$07,$DE,$F9,$6F,$C0,$91,$C1,$FD
	.dc.b	$B3,$E8,$5F,$23,$FC,$7D,$B1,$98,$C1,$FC,$8F,$E5,$DE,$BB,$E4,$3F
	.dc.b	$BA,$C6,$1B,$F1,$BF,$D7,$AF,$FC,$6F,$39,$7F,$C7,$3F,$2F,$FE,$CF
	.dc.b	$7F,$AF,$87,$0F,$F6,$26,$BF,$DD,$5A,$BE,$E9,$F1,$FF,$C8,$23,$FB
	.dc.b	$4C,$63,$C0,$FF,$04,$B6,$3F,$C9,$2D,$7C,$3E,$73,$83,$94,$7F,$80
	.dc.b	$E3,$15,$FE,$12,$6C,$7F,$94,$9A,$FA,$BE,$57,$FD,$17,$3F,$E1,$86
	.dc.b	$C7,$F9,$61,$AE,$FE,$1F,$BA,$FD,$47,$E6,$57,$F0,$2F,$8D,$FE,$F9
	.dc.b	$18,$9F,$E4,$FE,$0A,$73,$C1,$CA,$B5,$5F,$31,$AF,$F8,$31,$AF,$F8
	.dc.b	$7D,$E9,$E3,$FF,$48,$6E,$1F,$E1,$2C,$1F,$1B,$DA,$3F,$8A,$FC,$5F
	.dc.b	$F8,$3E,$5F,$1B,$63,$0D,$9F,$E3,$6D,$C7,$D2,$F1,$F6,$64,$FF,$20
	.dc.b	$F4,$2D,$7B,$A6,$E7,$FD,$73,$1F,$F6,$6D,$FF,$26,$9F,$17,$9F,$E3
	.dc.b	$F6,$6C,$FF,$2A,$6E,$1F,$C9,$A7,$E7,$DF,$F8,$D2,$31,$5F,$CB,$27
	.dc.b	$78,$77,$F4,$6C,$FC,$3A,$F8,$1C,$7F,$05,$B1,$FC,$E1,$A7,$D6,$2D
	.dc.b	$7F,$74,$F2,$FF,$D7,$F3,$FE,$4F,$8C,$7C,$67,$F8,$E6,$7F,$0F,$18
	.dc.b	$9D,$1F,$E4,$48,$E3,$5F,$DA,$1B,$3D,$79,$39,$5F,$F9,$0F,$95,$F3
	.dc.b	$17,$E3,$70,$DC,$BA,$E5,$FE,$61,$AF,$F6,$39,$27,$F2,$AD,$9D,$CA
	.dc.b	$3A,$C6,$1F,$EA,$0D,$9F,$F2,$A7,$CA,$7F,$9F,$B2,$FF,$0A,$AF,$83
	.dc.b	$CF,$AD,$A6,$DF,$69,$D3,$67,$4E,$7E,$1F,$A9,$DB,$E0,$2C,$64,$99
	.dc.b	$FF,$83,$AF,$FC,$CF,$18,$C7,$FC,$4F,$9B,$CF,$F9,$1F,$31,$F2,$5B
	.dc.b	$5E,$D7,$E1,$91,$FD,$F3,$70,$FF,$0D,$D7,$39,$F8,$9E,$87,$37,$D9
	.dc.b	$EE,$FC,$4F,$88,$F8,$AD,$8C,$9F,$06,$32,$D3,$E2,$7C,$77,$CC,$3E
	.dc.b	$6F,$CB,$58,$C9,$31,$F8,$11,$97,$CB,$E8,$7E,$17,$DD,$FC,$33,$6B
	.dc.b	$ED,$36,$77,$1F,$76,$76,$D3,$65,$8D,$9F,$0A,$36,$D8,$DB,$4D,$ED
	.dc.b	$F4,$A7,$6D,$9D,$B6,$77,$1F,$EA,$6D,$3F,$D3,$AE,$8C,$57,$26,$99
	.dc.b	$9F,$CF,$46,$5B,$19,$7F,$39,$19,$6C,$66,$7B,$C7,$88,$E2,$53,$65
	.dc.b	$77,$5F,$8E,$BF,$98,$DA,$F9,$0D,$AF,$88,$D3,$E0,$16,$3D,$71,$D6
	.dc.b	$FC,$D7,$DA,$7E,$25,$F6,$69,$D8,$B5,$D9,$B7,$D9,$A7,$67,$C1,$1E
	.dc.b	$E6,$BE,$E6,$D7,$86,$3F,$85,$57,$85,$B5,$E1,$69,$E1,$2D,$78,$5B
	.dc.b	$1E,$13,$C6,$F8,$5A,$F8,$5B,$5E,$18,$FE,$29,$5E,$16,$CF,$85,$A7
	.dc.b	$F7,$38,$FE,$3A,$48,$FE,$F0,$D8,$D4,$60,$8F,$DD,$F4,$7E,$8C,$FD
	.dc.b	$26,$CF,$E1,$DB,$3B,$8F,$C5,$4E,$DA,$6C,$B1,$B3,$F1,$51,$B6,$C6
	.dc.b	$DA,$6E,$FE,$9C,$ED,$B3,$B6,$CE,$E3,$E4,$CE,$DA,$6C,$B1,$B3,$EA
	.dc.b	$90,$F9,$73,$FC,$AA,$BD,$EF,$E1,$1B,$3F,$59,$B3,$C6,$D1,$88,$E7
	.dc.b	$3F,$BD,$8C,$4F,$FB,$2E,$3F,$A7,$46,$2F,$FA,$F3,$63,$27,$EE,$F8
	.dc.b	$3F,$CB,$AF,$F4,$BF,$3B,$FA,$9F,$BD,$FD,$51,$91,$93,$FA,$84,$63
	.dc.b	$1F,$E9,$1F,$31,$FD,$6B,$E6,$3F,$AD,$B6,$BF,$AB,$36,$77,$1F,$D6
	.dc.b	$64,$FF,$68,$F9,$27,$C6,$7F,$75,$F2,$F3,$7F,$EE,$29,$CD,$7F,$6B
	.dc.b	$9C,$EF,$FE,$FC,$9C,$D7,$F9,$A0,$38,$7F,$CD,$6D,$44,$70,$DB,$EB
	.dc.b	$7A,$8E,$71,$87,$87,$5D,$73,$3F,$DD,$7E,$43,$FB,$E9,$FD,$FB,$E2
	.dc.b	$6C,$64,$93,$FD,$C6,$32,$FF,$73,$F5,$7F,$C0,$BF,$1F,$FE,$06,$6D
	.dc.b	$7F,$80,$5B,$3B,$8F,$F0,$24,$F8,$F2,$74,$DE,$9C,$36,$BE,$83,$FE
	.dc.b	$6E,$9D,$C7,$C9,$BF,$93,$39,$24,$70,$3E,$C7,$47,$47,$FC,$29,$F3
	.dc.b	$1F,$E1,$C3,$FC,$3B,$EB,$D8,$C9,$2F,$FC,$21,$19,$7F,$C2,$51,$FE
	.dc.b	$24,$F5,$7F,$C4,$AD,$3F,$C4,$05,$8D,$BF,$E2,$38,$FA,$5F,$28,$F8
	.dc.b	$6B,$F3,$18,$6D,$7F,$8D,$A7,$FD,$4E,$48,$FF,$47,$B6,$35,$1D,$3E
	.dc.b	$63,$94,$71,$F9,$8E,$E7,$83,$E6,$2C,$64,$99,$39,$FF,$36,$B1,$57
	.dc.b	$08,$D5,$7E,$3E,$5F,$11,$EC,$F4,$BD,$AD,$9F,$53,$6B,$89,$F9,$CA
	.dc.b	$E3,$3E,$FA,$FC,$B7,$CC,$7E,$9B,$1F,$1D,$3E,$1F,$3D,$7F,$29,$B8
	.dc.b	$7E,$0E,$FF,$4B,$89,$A4,$AD,$C7,$88,$F7,$7C,$3E,$31,$FF,$3F,$79
	.dc.b	$FF,$EB,$CF,$48,$D2,$63,$F9,$1E,$9F,$F3,$DE,$1F,$EE,$26,$BF,$B3
	.dc.b	$6B,$FB,$A6,$A7,$AC,$73,$9F,$F6,$5C,$62,$7F,$83,$35,$FF,$77,$D7
	.dc.b	$F9,$E1,$FE,$14,$D7,$3E,$87,$A4,$FD,$47,$95,$30,$FD,$47,$5A,$31
	.dc.b	$5D,$BE,$2B,$C2,$4A,$F7,$3F,$ED,$EF,$A1,$FA,$F3,$FC,$65,$AF,$F1
	.dc.b	$10,$E1,$FE,$28,$D7,$F8,$DF,$CC,$FD,$BA,$FD,$D3,$63,$F9,$23,$5E
	.dc.b	$8D,$7F,$8F,$2D,$BF,$DE,$2D,$9E,$71,$83,$FD,$EB,$F3,$EF,$1A,$92
	.dc.b	$7F,$85,$46,$5F,$E1,$9F,$34,$F1,$7F,$DF,$AD,$8C,$C6,$0C,$F8,$72
	.dc.b	$38,$B7,$2F,$E7,$DF,$3F,$FE,$A2,$6F,$8F,$FA,$E7,$F9,$C4,$65,$F9
	.dc.b	$4D,$7F,$9E,$7C,$7B,$D6,$E1,$1F,$D1,$23,$18,$7F,$23,$6C,$7F,$4D
	.dc.b	$6B,$F5,$5F,$C8,$46,$F9,$79,$CA,$72,$F1,$8C,$FC,$F7,$F7,$EA,$FE
	.dc.b	$60,$D9,$FE,$B4,$D7,$F0,$4D,$7F,$AA,$7C,$E7,$ED,$09,$D5,$E0,$71
	.dc.b	$9D,$25,$FF,$3D,$8F,$85,$5F,$D8,$BE,$79,$F0,$9E,$86,$D7,$F4,$36
	.dc.b	$CF,$F6,$D6,$BF,$95,$F9,$DF,$E0,$A4,$C7,$E0,$B8,$C6,$92,$FF,$A7
	.dc.b	$47,$8A,$9F,$39,$F1,$E7,$FD,$D6,$FF,$BA,$23,$F8,$63,$96,$F4,$7F
	.dc.b	$71,$8C,$47,$E9,$DA,$FF,$79,$6B,$FE,$00,$6A,$F9,$C7,$C1,$FF,$7B
	.dc.b	$6D,$FF,$BF,$17,$C1,$66,$74,$C9,$FF,$80,$3F,$2B,$FE,$EA,$BF,$E4
	.dc.b	$57,$C2,$7F,$C1,$FE,$17,$F4,$23,$FC,$E8,$4F,$6B,$FC,$14,$D8,$D9
	.dc.b	$FE,$11,$7F,$C7,$D5,$FC,$B6,$FF,$87,$4E,$A3,$AD,$7F,$84,$D7,$C3
	.dc.b	$53,$1F,$0D,$6B,$84,$FF,$88,$A3,$31,$88,$7C,$36,$1F,$E7,$86,$23
	.dc.b	$51,$FE,$21,$C1,$EB,$F8,$63,$C5,$AC,$AE,$9A,$DC,$39,$BE,$73,$D0
	.dc.b	$FF,$8E,$07,$CE,$3E,$23,$F9,$14,$EB,$C6,$FF,$0B,$57,$F8,$E4,$AE
	.dc.b	$1E,$47,$F2,$9F,$9A,$7C,$DB,$E7,$3F,$C3,$D1,$86,$C6,$67,$FC,$95
	.dc.b	$39,$6F,$CC,$FF,$7B,$6F,$1F,$28,$C4,$77,$6B,$FE,$49,$97,$CC,$FD
	.dc.b	$BD,$71,$F0,$D6,$BF,$CA,$74,$8C,$42,$38,$BF,$61,$D3,$F9,$0F,$7B
	.dc.b	$7C,$7F,$E4,$08,$D5,$72,$A7,$35,$F8,$59,$5C,$E2,$BF,$DE,$65,$F3
	.dc.b	$9C,$4C,$EB,$FE,$1A,$8C,$CF,$E8,$3E,$13,$C2,$6D,$FC,$6F,$1F,$E2
	.dc.b	$9B,$CB,$EC,$8F,$D7,$C6,$B9,$B1,$3E,$AF,$8C,$93,$C0,$FC,$BF,$CE
	.dc.b	$7C,$27,$FC,$1C,$FC,$26,$ED,$F2,$23,$59,$FF,$8C,$1D,$3F,$E3,$2F
	.dc.b	$9C,$FF,$23,$57,$E8,$9A,$F2,$3F,$4B,$FA,$9F,$F4,$57,$FC,$C8,$D7
	.dc.b	$FD,$1C,$D7,$FB,$B3,$5F,$F4,$47,$B6,$3F,$2E,$57,$FC,$DC,$D7,$FD
	.dc.b	$30,$D7,$9F,$93,$F9,$E9,$FF,$0E,$47,$8A,$30,$74,$F7,$24,$74,$6B
	.dc.b	$FE,$1A,$F2,$FF,$4F,$EC,$7F,$53,$8F,$F3,$F3,$5D,$1E,$EE,$7F,$01
	.dc.b	$1C,$1E,$8F,$6F,$D6,$47,$07,$4F,$8D,$EE,$47,$07,$CA,$F8,$6F,$61
	.dc.b	$1C,$1F,$13,$D7,$F8,$E4,$70,$7A,$3B,$BD,$64,$70,$7D,$6F,$6B,$E4
	.dc.b	$11,$C1,$F4,$3E,$77,$E9,$91,$C1,$F1,$3B,$3E,$39,$1C,$1F,$82,$EA
	.dc.b	$FA,$84,$70,$7D,$CF,$97,$FB,$E4,$70,$7D,$0F,$9D,$FA,$64,$70,$7F
	.dc.b	$AC,$7E,$21,$EF,$FF,$1E,$6C,$66,$30,$7F,$22,$F9,$C7,$C7,$52,$4F
	.dc.b	$FA,$EE,$32,$FF,$AF,$7E,$71,$F1,$AF,$8E,$C3,$67,$FD,$D5,$2C,$7F
	.dc.b	$B0,$C6,$27,$4F,$E8,$30,$FF,$0F,$F5,$78,$BC,$8F,$D3,$C6,$0F,$F1
	.dc.b	$1F,$CD,$BE,$3A,$92,$7F,$DA,$11,$97,$FD,$A5,$E7,$B8,$3F,$C5,$6D
	.dc.b	$8C,$C6,$0F,$0F,$A3,$FE,$F3,$C7,$FC,$A1,$97,$FA,$4D,$FE,$AB,$5F
	.dc.b	$BC,$F8,$4D,$57,$08,$FD,$EB,$6B,$F7,$AE,$A3,$F8,$1F,$C8,$59,$E9
	.dc.b	$F0,$5D,$FE,$0F,$C0,$DA,$FE,$12,$D9,$DC,$7F,$0C,$9F,$37,$04,$8F
	.dc.b	$E1,$FC,$16,$BD,$13,$7C,$6F,$F6,$4E,$9F,$62,$FF,$8A,$CF,$BD,$F8
	.dc.b	$1F,$21,$5F,$E3,$2D,$BF,$E2,$F1,$F5,$88,$7F,$1C,$35,$7D,$B8,$6C
	.dc.b	$7F,$1A,$F8,$E3,$FE,$08,$F1,$DF,$21,$8C,$3F,$D8,$10,$AC,$61,$FE
	.dc.b	$E0,$8D,$DF,$19,$CB,$FC,$9A,$ED,$8F,$E5,$13,$87,$F9,$54,$5A,$FF
	.dc.b	$74,$D7,$8F,$E3,$B1,$F2,$19,$64,$64,$F4,$46,$2F,$FD,$AD,$F3,$5F
	.dc.b	$A1,$9F,$67,$CD,$F7,$8C,$4F,$B2,$31,$D7,$FE,$7C,$9C,$CF,$FA,$D5
	.dc.b	$AF,$A4,$38,$7D,$8F,$4D,$3E,$44,$6A,$7D,$9F,$37,$DB,$0F,$9D,$F0
	.dc.b	$FE,$FF,$9B,$F8,$0D,$8C,$9F,$4E,$32,$D3,$EB,$7C,$0F,$C7,$3E,$47
	.dc.b	$AF,$63,$24,$C7,$EC,$C6,$5F,$B7,$EB,$7C,$FF,$81,$FA,$0D,$AF,$C1
	.dc.b	$3F,$E9,$EE,$1F,$F3,$3D,$7F,$5A,$8C,$12,$76,$7E,$27,$D8,$E6,$FE
	.dc.b	$2E,$7E,$E7,$81,$F7,$5B,$19,$3F,$21,$19,$69,$FE,$5A,$F4,$5F,$35
	.dc.b	$F8,$76,$C6,$7E,$6F,$13,$23,$83,$B8,$C5,$B8,$BE,$AF,$CC,$7F,$32
	.dc.b	$BF,$F3,$3A,$F9,$4F,$CA,$39,$AE,$13,$F8,$18,$D3,$C3,$CA,$DD,$70
	.dc.b	$6C,$64,$EB,$19,$69,$9A,$EA,$7E,$96,$BA,$B8,$0F,$8E,$FD,$35,$CE
	.dc.b	$75,$3C,$27,$DB,$7E,$29,$CB,$63,$37,$E2,$9C,$B6,$33,$1D,$2F,$B4
	.dc.b	$78,$27,$D0,$D9,$E2,$6E,$32,$7C,$79,$E2,$D3,$31,$EF,$8C,$B6,$32
	.dc.b	$FC,$08,$CB,$63,$37,$EF,$7E,$A5,$FB,$5A,$7A,$CB,$3E,$53,$CF,$7E
	.dc.b	$93,$DE,$58,$DB,$63,$EA,$4F,$56,$CF,$56,$9D,$67,$9C,$7B,$E3,$BB
	.dc.b	$63,$31,$EF,$32,$57,$2F,$BC,$C9,$5C,$D7,$C1,$8F,$0C,$79,$DF,$C3
	.dc.b	$D7,$9D,$B5,$E7,$69,$E7,$7D,$E7,$94,$AF,$8D,$AF,$85,$B5,$E1,$8E
	.dc.b	$67,$F2,$52,$9F,$40,$C3,$FA,$B6,$CF,$F2,$67,$44,$3E,$59,$A9,$FB
	.dc.b	$FF,$11,$F8,$2A,$FD,$14,$78,$A3,$15,$96,$C7,$03,$F0,$F1,$96,$99
	.dc.b	$AF,$9B,$5D,$1F,$97,$1F,$BF,$AE,$71,$A6,$DF,$D2,$6C,$F8,$CE,$B3
	.dc.b	$B6,$CE,$DA,$6E,$3C,$37,$E5,$6C,$79,$67,$EF,$CE,$5B,$19,$3E,$9C
	.dc.b	$65,$A6,$67,$D6,$43,$F0,$33,$FC,$8A,$BD,$31,$9B,$46,$23,$75,$F0
	.dc.b	$CF,$4B,$B8,$D4,$FC,$98,$FE,$2F,$1F,$B8,$F9,$87,$BD,$FD,$FD,$7C
	.dc.b	$5D,$8C,$37,$0F,$F0,$36,$3F,$DA,$AB,$71,$83,$FB,$93,$D1,$C6,$1F
	.dc.b	$E0,$D8,$DB,$FC,$A5,$7C,$87,$84,$F1,$7C,$DD,$AE,$34,$AD,$C6,$1B
	.dc.b	$1C,$E7,$83,$FC,$C3,$D9,$FE,$8F,$F3,$7F,$D2,$1B,$7F,$D3,$1F,$CF
	.dc.b	$4F,$F3,$D6,$BF,$CF,$1B,$A9,$A7,$39,$FE,$AB,$F3,$76,$BD,$9E,$EF
	.dc.b	$CB,$F8,$0F,$98,$DA,$F4,$36,$77,$1E,$B9,$FA,$5E,$D4,$8F,$77,$BB
	.dc.b	$6B,$E4,$4D,$DF,$E9,$4D,$AF,$EC,$5E,$D7,$F6,$6F,$83,$FE,$CE,$D8
	.dc.b	$FE,$D6,$FB,$08,$FF,$05,$C3,$FA,$F4,$61,$FE,$B2,$FF,$70,$F8,$6E
	.dc.b	$73,$F4,$7D,$8F,$C4,$1F,$89,$F7,$2C,$64,$97,$F7,$E3,$2F,$E0,$7D
	.dc.b	$5F,$C8,$7A,$FF,$91,$6D,$7D,$D6,$CE,$E3,$F0,$13,$FE,$55,$F8,$B7
	.dc.b	$C8,$7F,$81,$3C,$BC,$36,$BF,$13,$3A,$7D,$D6,$8C,$3F,$05,$AF,$F2
	.dc.b	$76,$BF,$E4,$D7,$7B,$FF,$92,$9B,$39,$EA,$F0,$BF,$95,$F5,$FF,$58
	.dc.b	$7E,$B7,$D9,$B1,$92,$61,$FA,$38,$CB,$FA,$5F,$63,$F6,$9E,$CF,$ED
	.dc.b	$5B,$5F,$A1,$6C,$EE,$3F,$49,$3F,$17,$9F,$E3,$73,$D9,$F8,$6D,$AF
	.dc.b	$81,$3F,$A0,$CC,$D2,$57,$CA,$7E,$87,$37,$1D,$7F,$D1,$DF,$37,$FE
	.dc.b	$53,$25,$7C,$EF,$29,$1F,$AC,$F1,$FF,$C5,$39,$7F,$53,$6B,$FE,$79
	.dc.b	$6B,$FE,$93,$6A,$CF,$FC,$9A,$FE,$5D,$AF,$F5,$D6,$BF,$E8,$36,$BF
	.dc.b	$E9,$DF,$9A,$FE,$22,$FF,$63,$6B,$FE,$89,$6B,$BE,$1F,$47,$8B,$32
	.dc.b	$FC,$4E,$1F,$D9,$EB,$D1,$3F,$27,$B2,$57,$E3,$B7,$46,$2B,$F5,$DF
	.dc.b	$45,$F3,$AB,$F2,$6D,$8C,$F0,$7F,$12,$8C,$47,$F7,$9F,$9B,$7C,$47
	.dc.b	$F1,$66,$DF,$E7,$1B,$3F,$BB,$6B,$FE,$01,$F9,$B7,$C4,$7F,$1E,$6D
	.dc.b	$7E,$91,$B3,$FC,$0D,$AF,$F8,$27,$E8,$9E,$1F,$F2,$66,$D7,$EB,$1B
	.dc.b	$3F,$C2,$DA,$FF,$84,$7E,$71,$DD,$FC,$B9,$B5,$FB,$46,$CF,$F1,$36
	.dc.b	$BF,$E1,$9F,$9C,$76,$FF,$36,$6D,$7E,$F1,$B3,$FC,$6D,$AF,$F8,$87
	.dc.b	$EA,$3F,$DB,$D3,$FC,$09,$B1,$FC,$85,$AE,$CF,$E8,$7F,$3B,$F8,$44
	.dc.b	$D2,$73,$7F,$48,$F8,$EF,$EA,$73,$FE,$2F,$6B,$FE,$D5,$6B,$FE,$EF
	.dc.b	$6A,$FA,$37,$B9,$FD,$49,$B8,$FF,$18,$6C,$FF,$37,$6B,$FE,$3D,$F9
	.dc.b	$E7,$93,$FD,$69,$B5,$FC,$81,$B3,$FC,$FD,$AF,$F9,$17,$E7,$BF,$8F
	.dc.b	$A0,$E1,$BE,$37,$F6,$68,$C7,$CF,$3E,$1B,$FB,$4B,$70,$FE,$5E,$DB
	.dc.b	$FE,$9C,$D7,$FC,$A8,$FA,$9F,$F7,$D3,$5F,$13,$53,$7E,$7B,$C8,$4C
	.dc.b	$3F,$BC,$BE,$7F,$9E,$FF,$13,$DF,$F4,$06,$DF,$F5,$F6,$BF,$E6,$56
	.dc.b	$BF,$04,$2D,$93,$FE,$14,$8E,$7E,$36,$2B,$E4,$46,$E7,$5E,$C6,$6B
	.dc.b	$54,$F9,$EF,$F7,$CC,$61,$B5,$E1,$5E,$EF,$18,$C7,$BF,$6F,$EB,$56
	.dc.b	$A7,$FC,$4E,$D8,$FF,$6B,$7C,$D3,$D9,$E1,$1E,$1C,$2E,$2F,$5A,$98
	.dc.b	$F5,$AE,$5F,$E3,$29,$46,$63,$FC,$6B,$8F,$F1,$E7,$51,$DB,$07,$C1
	.dc.b	$D3,$1F,$05,$6B,$FC,$7B,$28,$CC,$63,$E0,$AC,$FF,$BB,$C3,$D4,$F5
	.dc.b	$E0,$F5,$71,$5D,$37,$F1,$1E,$3D,$6C,$78,$1C,$3E,$2C,$3F,$C0,$83
	.dc.b	$E7,$9F,$09,$FE,$A9,$AF,$64,$E8,$FE,$7C,$7F,$83,$0C,$1F,$E0,$B3
	.dc.b	$FC,$59,$F3,$4F,$9D,$7C,$F7,$F8,$4A,$30,$D8,$CC,$F0,$6B,$FE,$B9
	.dc.b	$97,$CC,$B8,$B9,$36,$7B,$B5,$E0,$D7,$F3,$92,$F9,$EF,$CF,$57,$E5
	.dc.b	$23,$F0,$53,$FE,$70,$6B,$DF,$4E,$35,$C1,$E7,$C9,$67,$F8,$47,$07
	.dc.b	$F8,$29,$FC,$2F,$A3,$95,$87,$4C,$7C,$1C,$B9,$8F,$E3,$33,$E1,$F0
	.dc.b	$71,$39,$75,$F3,$D2,$74,$79,$DE,$D8,$BE,$02,$9B,$9F,$51,$FC,$77
	.dc.b	$8B,$F1,$6F,$FA,$39,$72,$79,$DB,$19,$E6,$C4,$F6,$F8,$79,$3C,$63
	.dc.b	$DD,$F3,$DF,$36,$3F,$D3,$2F,$BF,$87,$F8,$63,$63,$F9,$73,$B7,$CA
	.dc.b	$1D,$2F,$9E,$7C,$8F,$CB,$6C,$76,$E3,$F9,$DD,$EF,$80,$FC,$A3,$67
	.dc.b	$FC,$C2,$D7,$FD,$00,$D7,$F9,$01,$FE,$CE,$F6,$78,$91,$E2,$F4,$DE
	.dc.b	$47,$46,$BD,$3E,$17,$C2,$47,$C2,$F9,$3D,$39,$1E,$46,$BE,$4F,$85
	.dc.b	$F4,$91,$F4,$BE,$CF,$02,$47,$B1,$AF,$B3,$D3,$F8,$24,$7E,$0B,$F1
	.dc.b	$3E,$0E,$47,$C4,$6B,$F1,$3E,$13,$E4,$91,$F9,$2F,$CC,$F9,$69,$1F
	.dc.b	$31,$AF,$CC,$F9,$6F,$A2,$47,$E8,$BF,$53,$E5,$A4,$7D,$46,$BF,$53
	.dc.b	$97,$EC,$91,$FB,$2F,$F0,$FF,$D5,$5F,$2F,$49,$3F,$E9,$E8,$CB,$7D
	.dc.b	$2E,$04,$70,$6F,$9F,$A1,$1C,$1F,$EA,$8F,$B8,$75,$FF,$1F,$6C,$66
	.dc.b	$30,$67,$79,$1C,$5A,$FF,$B3,$F0,$7C,$C7,$AC,$91,$E9,$7F,$96,$7C
	.dc.b	$E3,$81,$F1,$BF,$EB,$A8,$C3,$78,$7F,$B3,$DF,$FB,$6E,$72,$FF,$B7
	.dc.b	$7E,$8B,$F5,$3C,$1F,$EC,$26,$C6,$76,$FF,$80,$31,$88,$FD,$1F,$D2
	.dc.b	$BE,$4B,$FB,$CC,$7F,$81,$63,$17,$FE,$CC,$6C,$7F,$BA,$DA,$FE,$B3
	.dc.b	$E7,$9C,$9F,$DC,$97,$2D,$8C,$36,$FF,$DA,$F3,$C1,$B3,$E2,$AF,$F1
	.dc.b	$B4,$E5,$FF,$7D,$37,$CA,$FF,$07,$B6,$BF,$C3,$8F,$F8,$69,$FF,$1F
	.dc.b	$46,$AB,$FC,$0B,$C2,$F9,$1A,$63,$E4,$6D,$F0,$3F,$C4,$D3,$B7,$18
	.dc.b	$FF,$92,$63,$0F,$F1,$C7,$FD,$87,$1A,$84,$6B,$0F,$F1,$2B,$FE,$92
	.dc.b	$6F,$C5,$F9,$3E,$2B,$CA,$DB,$E5,$5F,$EA,$0A,$E3,$5A,$9E,$DE,$A7
	.dc.b	$83,$0F,$E5,$AF,$E4,$23,$38,$FF,$89,$23,$13,$A8,$F9,$CF,$F8,$C9
	.dc.b	$B1,$CE,$3F,$C0,$91,$83,$91,$CF,$2B,$5E,$6F,$A1,$F9,$A7,$CE,$FA
	.dc.b	$2B,$19,$25,$FC,$78,$CB,$F2,$7E,$7B,$EA,$FD,$07,$D6,$6D,$7C,$26
	.dc.b	$CE,$E3,$E2,$CF,$B4,$F7,$5D,$8F,$8A,$D9,$FE,$82,$7F,$94,$27,$DE
	.dc.b	$49,$CC,$65,$F9,$B1,$F2,$6B,$96,$9F,$6B,$E7,$7F,$17,$F3,$DF,$8C
	.dc.b	$6D,$FD,$86,$E1,$F8,$56,$D7,$E0,$BD,$0F,$C9,$9F,$94,$F8,$8B,$7F
	.dc.b	$80,$8B,$7F,$8D,$9B,$5F,$E6,$ED,$FF,$CA,$37,$F8,$2B,$FB,$DE,$F7
	.dc.b	$DF,$6C,$7B,$27,$78,$7C,$26,$BF,$E4,$1A,$FC,$09,$C1,$FC,$5F,$BD
	.dc.b	$70,$FC,$CB,$FA,$A9,$FD,$5B,$6F,$F2,$ED,$AF,$D2,$36,$BF,$43,$E9
	.dc.b	$7E,$C8,$FD,$9F,$A7,$6F,$F3,$F1,$6F,$F5,$93,$6B,$FD,$1B,$F3,$2F
	.dc.b	$9D,$FC,$E5,$7F,$A0,$23,$18,$7E,$92,$31,$3A,$8F,$DA,$4E,$A7,$F4
	.dc.b	$E6,$A3,$81,$A9,$E1,$1F,$E1,$89,$E1,$5F,$46,$38,$1C,$AB,$81,$C9
	.dc.b	$E1,$5C,$A7,$D2,$D9,$FF,$59,$57,$47,$FD,$69,$9F,$FB,$E2,$B7,$1C
	.dc.b	$DB,$7D,$CE,$B1,$93,$9D,$F5,$8F,$F6,$F5,$F5,$7F,$DC,$17,$D6,$B6
	.dc.b	$71,$8D,$D7,$15,$5C,$5F,$EA,$33,$EE,$7C,$57,$E4,$3E,$24,$65,$B1
	.dc.b	$E6,$37,$1E,$06,$9E,$07,$C4,$F8,$2B,$C5,$5E,$1B,$F4,$4E,$CE,$26
	.dc.b	$EB,$8C,$EE,$3E,$AD,$F8,$8F,$34,$FC,$E8,$CB,$4F,$41,$5F,$44,$71
	.dc.b	$8F,$43,$63,$C4,$D7,$C4,$79,$27,$C5,$3D,$23,$E8,$9E,$02,$B9,$7E
	.dc.b	$69,$E0,$7C,$53,$E8,$8D,$99,$2B,$97,$8B,$6B,$E3,$C7,$90,$FF,$30
	.dc.b	$D7,$E8,$5B,$7F,$D0,$A3,$11,$F5,$4C,$DF,$F2,$48,$CC,$7D,$13,$27
	.dc.b	$8C,$FC,$3F,$CF,$7B,$5F,$87,$87,$96,$53,$97,$F1,$84,$71,$79,$6D
	.dc.b	$FC,$59,$E2,$D9,$D3,$4D,$4F,$0B,$F2,$57,$06,$D7,$06,$D7,$03,$93
	.dc.b	$C2,$B9,$5F,$A5,$B3,$F8,$23,$F7,$2E,$27,$F0,$71,$83,$9C,$F1,$69
	.dc.b	$C6,$3F,$6B,$3C,$5F,$8F,$1C,$5F,$29,$E3,$78,$BD,$8E,$2E,$1E,$26
	.dc.b	$0E,$33,$E1,$6C,$78,$5F,$AE,$F1,$BF,$A4,$DB,$FE,$59,$7F,$6F,$A3
	.dc.b	$F1,$63,$1F,$F1,$5E,$1F,$E2,$C6,$DF,$F8,$61,$B1,$FE,$25,$51,$FE
	.dc.b	$25,$FD,$57,$EA,$65,$8A,$08,$CC,$E2,$31,$8E,$EB,$84,$6F,$1C,$56
	.dc.b	$EF,$8C,$F4,$6B,$CA,$3C,$53,$C2,$35,$E9,$D7,$C5,$F3,$F6,$D3,$63
	.dc.b	$0F,$F8,$AA,$33,$7F,$E0,$A6,$D7,$F8,$41,$AF,$FB,$49,$CE,$1E,$A6
	.dc.b	$CF,$A7,$E6,$7A,$BF,$1F,$E2,$3C,$5F,$4D,$E3,$6C,$64,$F9,$91,$96
	.dc.b	$9F,$3F,$C4,$F5,$1E,$BF,$72,$C6,$49,$97,$FB,$DB,$14,$EA,$3F,$DF
	.dc.b	$3F,$3E,$F8,$CF,$89,$F3,$FF,$1E,$31,$87,$DD,$8C,$4F,$F9,$3A,$73
	.dc.b	$5F,$E3,$26,$BF,$EF,$46,$BF,$7D,$DE,$BF,$79,$B7,$F7,$7E,$63,$AB
	.dc.b	$F8,$BF,$87,$FA,$DF,$11,$F5,$DB,$19,$3F,$21,$19,$69,$F9,$3F,$80
	.dc.b	$FB,$C7,$DF,$F8,$8B,$19,$26,$5F,$16,$32,$FC,$7F,$84,$FC,$37,$C3
	.dc.b	$7E,$1D,$B5,$F9,$F7,$F0,$C4,$7F,$0C,$D9,$E5,$1F,$A5,$F8,$6E,$4F
	.dc.b	$E9,$FE,$9B,$F2,$5F,$4B,$F9,$36,$C6,$4F,$D6,$46,$5A,$7F,$A4,$FE
	.dc.b	$93,$E4,$DF,$EB,$9A,$FE,$BF,$D8,$FD,$04,$60,$9C,$5F,$A1,$8C,$BA
	.dc.b	$F8,$ED,$FC,$B5,$C3,$B3,$70,$FF,$37,$E0,$AF,$F4,$7B,$6B,$FC,$DE
	.dc.b	$D9,$DC,$F1,$E4,$FD,$EC,$FF,$BD,$2B,$D5,$3F,$E3,$FD,$BF,$AC,$12
	.dc.b	$3F,$AB,$3F,$E9,$CF,$9F,$7C,$75,$24,$FF,$9F,$A3,$2F,$FA,$03,$E8
	.dc.b	$7F,$23,$97,$F1,$26,$A6,$98,$FA,$2F,$E9,$91,$FC,$54,$38,$7F,$8C
	.dc.b	$35,$FE,$3A,$D7,$7E,$0F,$FA,$CA,$0D,$25,$7F,$A5,$21,$E0,$6C,$91
	.dc.b	$97,$FD,$77,$F3,$FF,$EB,$9B,$FE,$50,$D7,$F9,$6B,$53,$4F,$F6,$DB
	.dc.b	$EB,$6B,$CB,$D4,$EA,$76,$F5,$E5,$FF,$AA,$29,$18,$AF,$F5,$7F,$CE
	.dc.b	$BE,$2F,$FD,$9E,$DB,$FF,$57,$B6,$77,$3A,$F8,$B9,$5A,$6C,$F4,$F8
	.dc.b	$BE,$64,$9E,$6F,$FB,$73,$C7,$73,$52,$4F,$FA,$FA,$32,$DF,$87,$E0
	.dc.b	$47,$06,$F9,$7A,$11,$C1,$FE,$C8,$FB,$37,$5F,$FB,$B1,$B1,$98,$C1
	.dc.b	$9E,$C9,$1C,$5B,$9F,$FB,$5A,$9D,$9C,$C9,$5D,$DF,$F7,$CF,$C8,$76
	.dc.b	$AF,$F6,$EB,$5F,$CA,$35,$FE,$D4,$D7,$FB,$7A,$BE,$64,$AF,$F7,$1C
	.dc.b	$65,$B5,$E0,$23,$83,$FD,$D3,$F2,$FF,$96,$6F,$2F,$F7,$6C,$3F,$BD
	.dc.b	$4E,$5F,$F7,$9B,$5F,$D5,$39,$ED,$FE,$FA,$DA,$E3,$D1,$FD,$D6,$31
	.dc.b	$5F,$E0,$8F,$97,$7A,$9F,$DE,$5B,$7F,$E0,$CC,$3F,$C0,$F1,$FE,$19
	.dc.b	$8C,$1F,$E0,$D7,$74,$9D,$FA,$3C,$CF,$F0,$27,$97,$FC,$5E,$33,$7C
	.dc.b	$AB,$B7,$8D,$FE,$09,$93,$7C,$1C,$36,$B8,$57,$79,$EB,$F2,$59,$8F
	.dc.b	$14,$FF,$0D,$F5,$FF,$C2,$91,$FE,$FA,$C3,$E6,$7A,$F8,$9C,$B2,$BA
	.dc.b	$D3,$D7,$C3,$D0,$FF,$16,$3D,$0F,$6F,$5D,$9E,$3B,$F8,$0F,$0E,$D6
	.dc.b	$3A,$B8,$7B,$DB,$C4,$FF,$1D,$FC,$BB,$D9,$FE,$4D,$7E,$AA,$D1,$FC
	.dc.b	$A4,$FF,$1F,$55,$69,$4D,$9F,$CB,$FE,$59,$E7,$3E,$5F,$84,$61,$B1
	.dc.b	$99,$ED,$39,$6F,$A1,$F7,$DB,$C3,$CA,$31,$1D,$DA,$FF,$94,$E5,$E8
	.dc.b	$7F,$95,$6B,$FC,$A7,$1F,$80,$9F,$BA,$D7,$E8,$07,$8B,$AE,$5E,$5B
	.dc.b	$7F,$97,$F9,$BF,$AC,$CF,$E1,$9A,$F2,$AE,$97,$E0,$E6,$E3,$3F,$E6
	.dc.b	$A9,$79,$F8,$8F,$03,$61,$36,$31,$E8,$4A,$F4,$C5,$7D,$38,$3E,$57
	.dc.b	$A3,$63,$DF,$5D,$F8,$BF,$21,$1F,$9E,$7F,$6A,$DE,$4D,$73,$6D,$E7
	.dc.b	$E2,$49,$D1,$FA,$6F,$97,$F8,$EF,$CA,$7E,$0E,$FF,$E8,$56,$C7,$FA
	.dc.b	$29,$DE,$FF,$E8,$67,$30,$F9,$77,$C6,$FC,$56,$C7,$7E,$2F,$91,$FA
	.dc.b	$B3,$D8,$FC,$93,$67,$F2,$CD,$7F,$34,$D7,$FC,$09,$EE,$3C,$6F,$CD
	.dc.b	$36,$3F,$3C,$D7,$F4,$4D,$7F,$C1,$92,$F6,$3C,$24,$8F,$0B,$D3,$92
	.dc.b	$47,$46,$BD,$3E,$63,$C2,$47,$C2,$FB,$3A,$E4,$7B,$1A,$F9,$3C,$FF
	.dc.b	$49,$1F,$4B,$EC,$F0,$64,$7B,$1A,$FC,$4E,$DF,$92,$47,$E4,$BF,$13
	.dc.b	$E6,$24,$7C,$46,$BF,$13,$E4,$FE,$49,$1F,$92,$FD,$4E,$F9,$1F,$51
	.dc.b	$AF,$CC,$F8,$FF,$A2,$47,$E8,$BF,$53,$E3,$A4,$7D,$46,$BF,$73,$8F
	.dc.b	$F0,$24,$7F,$02,$FF,$15,$F8,$47,$CC,$52,$4F,$F1,$B8,$CB,$7E,$63
	.dc.b	$81,$1C,$1F,$C8,$7E,$EB,$F8,$EB,$78,$3F,$91,$46,$23,$F9,$43,$5F
	.dc.b	$F2,$97,$CC,$BB,$BF,$94,$36,$3F,$96,$35,$FE,$60,$D7,$8F,$CC,$F0
	.dc.b	$23,$C1,$FE,$65,$F5,$0E,$5A,$49,$FE,$71,$19,$7F,$9D,$7C,$C3,$8B
	.dc.b	$FC,$37,$1F,$CF,$E3,$1E,$AE,$BE,$65,$E7,$70,$BD,$DF,$F9,$B7,$E6
	.dc.b	$7F,$6F,$E2,$FF,$45,$6C,$7F,$49,$6B,$BF,$0F,$FD,$1F,$A7,$FB,$EF
	.dc.b	$BF,$F8,$FD,$FF,$A8,$E3,$1E,$67,$FA,$81,$DC,$7F,$A5,$1B,$7F,$E9
	.dc.b	$A7,$51,$FD,$E2,$BF,$96,$BC,$CC,$7B,$5D,$CF,$07,$97,$63,$24,$BF
	.dc.b	$E8,$74,$9D,$46,$36,$F4,$7B,$3E,$96,$DF,$85,$B3,$B8,$F2,$49,$FF
	.dc.b	$06,$F8,$9F,$EA,$99,$FE,$A2,$FF,$B0,$0C,$79,$3B,$80,$8D,$73,$7F
	.dc.b	$8E,$63,$31,$CD,$FE,$3B,$19,$78,$36,$3A,$D7,$49,$CB,$E4,$F7,$7C
	.dc.b	$17,$E1,$9C,$9E,$6F,$A7,$CD,$71,$B4,$62,$7C,$7E,$ED,$7D,$11,$67
	.dc.b	$75,$F0,$A3,$15,$97,$5F,$0D,$BF,$85,$B5,$F2,$1B,$87,$C8,$6E,$3E
	.dc.b	$7C,$7F,$D7,$F5,$FE,$3A,$F9,$97,$87,$E3,$5E,$25,$8C,$36,$7F,$C3
	.dc.b	$D3,$C1,$B3,$F2,$E3,$18,$FF,$89,$5B,$87,$C7,$6D,$7C,$D7,$F0,$0F
	.dc.b	$8E,$BE,$27,$84,$F8,$6A,$63,$E6,$2D,$7F,$94,$A7,$4D,$9D,$34,$D7
	.dc.b	$CC,$F4,$F8,$7E,$AD,$8C,$9C,$E3,$2D,$3F,$CB,$FD,$CC,$3E,$C2,$F9
	.dc.b	$BB,$18,$6C,$FF,$90,$E7,$FD,$B0,$49,$FC,$83,$CE,$3A,$C7,$4C,$38
	.dc.b	$EB,$FE,$30,$C9,$F3,$3F,$10,$F8,$BF,$35,$70,$F8,$51,$6B,$84,$DD
	.dc.b	$FF,$2D,$6F,$99,$F9,$BF,$35,$F3,$9B,$8F,$CB,$6E,$1E,$16,$DF,$E7
	.dc.b	$F4,$5F,$C2,$CB,$E1,$B7,$0F,$8E,$DC,$3F,$C9,$35,$C2,$7E,$5D,$70
	.dc.b	$69,$C0,$B3,$C2,$B9,$5F,$85,$B3,$FB,$5B,$E9,$5D,$AF,$A3,$67,$A3
	.dc.b	$4E,$8F,$6B,$E8,$D3,$A5,$F6,$9F,$E1,$B7,$DA,$BE,$15,$F6,$6D,$F6
	.dc.b	$6D,$76,$3C,$0F,$66,$C7,$63,$C0,$F6,$BF,$03,$FD,$31,$B7,$E1,$9F
	.dc.b	$81,$7E,$16,$9E,$12,$D7,$86,$FC,$67,$84,$B1,$E1,$8F,$19,$E1,$7C
	.dc.b	$6F,$85,$B5,$E1,$8F,$87,$5E,$16,$CF,$85,$A7,$85,$F1,$D7,$85,$A7
	.dc.b	$85,$F1,$C7,$84,$F1,$9E,$12,$D7,$85,$FE,$DF,$5E,$16,$D7,$85,$A7
	.dc.b	$84,$F1,$D7,$85,$B1,$E1,$3C,$6F,$86,$3C,$6F,$85,$B5,$E1,$3F,$67
	.dc.b	$3D,$DB,$3F,$7C,$FF,$63,$DF,$EF,$89,$3F,$C0,$DB,$1F,$EC,$99,$FC
	.dc.b	$CB,$C2,$74,$75,$C3,$DD,$3E,$1A,$FE,$23,$18,$72,$D9,$E0,$79,$63
	.dc.b	$2D,$33,$5F,$14,$F7,$3F,$0D,$FC,$FC,$F3,$75,$5F,$C2,$67,$94,$62
	.dc.b	$75,$7F,$36,$3F,$C4,$71,$89,$E9,$18,$78,$CE,$EB,$A4,$F6,$BF,$3B
	.dc.b	$6B,$CF,$5B,$9C,$B6,$32,$76,$AF,$8C,$FF,$38,$BF,$15,$FD,$D8,$FE
	.dc.b	$11,$18,$8F,$82,$7C,$08,$F2,$9C,$E3,$53,$F8,$16,$DF,$D2,$9D,$D7
	.dc.b	$53,$91,$E9,$3E,$33,$F9,$5C,$5F,$33,$FD,$7B,$E6,$BF,$AF,$B6,$7F
	.dc.b	$AE,$36,$BF,$A1,$37,$2F,$E7,$7E,$EF,$F6,$63,$FB,$3F,$CD,$5B,$FE
	.dc.b	$C9,$16,$FF,$A5,$4D,$AF,$E7,$5F,$0D,$FD,$6A,$BF,$AD,$B6,$FF,$AF
	.dc.b	$BF,$D5,$88,$FF,$5B,$6D,$FF,$3C,$7F,$B1,$FC,$37,$29,$FE,$E8,$FF
	.dc.b	$7A,$8F,$EF,$6D,$8C,$9F,$DE,$63,$2D,$3F,$BE,$7B,$FF,$E0,$23,$FC
	.dc.b	$07,$F0,$16,$32,$4B,$FE,$CD,$19,$58,$FF,$8D,$7E,$69,$F3,$5F,$DE
	.dc.b	$27,$FC,$61,$18,$BF,$EF,$B1,$89,$FF,$22,$4E,$63,$FC,$84,$7F,$93
	.dc.b	$9C,$38,$6B,$FE,$47,$78,$57,$86,$7B,$FC,$D7,$5B,$FF,$0C,$5F,$F8
	.dc.b	$7B,$8F,$FC,$3E,$D8,$C9,$FE,$1D,$8C,$B4,$FF,$2F,$DF,$F9,$47,$E6
	.dc.b	$17,$FE,$4B,$6C,$E8,$FF,$2D,$3E,$46,$C6,$C8,$EE,$FF,$CD,$73,$B6
	.dc.b	$CF,$5E,$FE,$DF,$1D,$6F,$85,$2D,$39,$8F,$92,$66,$BF,$CE,$13,$FE
	.dc.b	$77,$AC,$CE,$5B,$19,$8F,$B7,$1E,$A9,$F7,$7C,$7F,$5C,$7E,$17,$C8
	.dc.b	$FC,$1F,$9A,$F8,$4D,$BF,$2E,$28,$C6,$3D,$0F,$7B,$8B,$F9,$9F,$35
	.dc.b	$F2,$CF,$99,$F1,$57,$0E,$71,$6F,$91,$F0,$67,$15,$FE,$85,$F9,$A7
	.dc.b	$CD,$7D,$5F,$8A,$FF,$40,$46,$31,$F2,$C6,$27,$FD,$29,$39,$C3,$E4
	.dc.b	$B5,$FC,$9B,$5F,$F4,$B3,$BB,$F0,$4F,$E0,$FE,$27,$A6,$1F,$5B,$91
	.dc.b	$BF,$E7,$89,$1F,$4A,$32,$DF,$5B,$F4,$64,$70,$7D,$8F,$90,$79,$BF
	.dc.b	$EA,$B6,$C6,$63,$07,$FA,$BF,$E9,$FF,$0B,$1F,$80,$6E,$BF,$86,$6B
	.dc.b	$FB,$06,$BF,$EB,$3F,$3B,$F3,$AF,$E3,$1A,$FE,$D9,$AE,$F9,$FF,$D7
	.dc.b	$D1,$CA,$3F,$D8,$ED,$4D,$31,$F4,$DF,$E8,$C8,$FC,$CB,$5F,$E0,$6D
	.dc.b	$7F,$D8,$CD,$7F,$DA,$7F,$35,$FE,$7D,$7F,$42,$D7,$F8,$5B,$5F,$F6
	.dc.b	$73,$5E,$3E,$4F,$FB,$56,$3A,$41,$A6,$3E,$95,$ED,$FF,$B6,$DB,$9F
	.dc.b	$EB,$1B,$5F,$B2,$6B,$FC,$73,$E6,$DF,$37,$FE,$E3,$6D,$7E,$D1,$B3
	.dc.b	$FB,$A6,$BC,$7E,$3F,$81,$27,$83,$FE,$ED,$FA,$4F,$CB,$4F,$FB,$D9
	.dc.b	$A9,$A6,$3B,$1E,$CF,$FB,$D5,$B7,$FC,$19,$B5,$FC,$31,$AF,$F3,$67
	.dc.b	$FD,$FC,$FF,$BE,$5A,$8F,$E0,$5E,$D7,$53,$FD,$F4,$FF,$BF,$FE,$8D
	.dc.b	$CD,$88,$E1,$18,$BF,$E2,$ED,$8F,$E3,$AD,$7F,$A3,$FC,$EB,$E0,$79
	.dc.b	$42,$31,$5F,$C8,$1B,$1F,$C9,$9A,$F2,$F7,$B8,$C1,$A4,$9F,$E5,$3F
	.dc.b	$45,$FE,$0C,$6F,$A1,$E0,$8C,$5F,$F3,$26,$BF,$D7,$3E,$57,$BB,$F1
	.dc.b	$43,$87,$E4,$35,$CE,$3E,$C8,$F2,$5F,$F3,$A6,$CF,$F4,$06,$BD,$F0
	.dc.b	$F8,$7F,$3D,$FC,$3A,$3F,$94,$BF,$BC,$F0,$97,$E3,$8D,$46,$27,$ED
	.dc.b	$46,$E7,$5E,$07,$42,$3F,$D3,$23,$B4,$FF,$7A,$F9,$E7,$9F,$CF,$0E
	.dc.b	$92,$9C,$DF,$F5,$C5,$E6,$EE,$3E,$34,$6A,$7F,$A0,$BA,$8E,$13,$B9
	.dc.b	$FC,$87,$90,$F8,$2C,$47,$79,$B8,$3E,$4A,$98,$E4,$B8,$FF,$6E,$9F
	.dc.b	$C8,$1E,$82,$BF,$67,$17,$CF,$71,$8F,$EE,$B8,$BD,$3F,$0B,$63,$3E
	.dc.b	$A6,$27,$2C,$8E,$3E,$A5,$AF,$EB,$E1,$E6,$7B,$30,$7C,$BE,$2B,$C3
	.dc.b	$C1,$E0,$3C,$3B,$58,$F2,$B8,$7C,$F8,$3C,$AF,$DC,$7C,$F3,$D7,$FF
	.dc.b	$19,$DF,$AA,$B4,$7F,$82,$8A,$7F,$55,$31,$5F,$E0,$EF,$9D,$7C,$A3
	.dc.b	$E7,$B0,$DA,$FF,$0B,$CE,$A7,$27,$1F,$9A,$FF,$26,$CE,$9A,$FF,$91
	.dc.b	$A5,$18,$8F,$D6,$46,$38,$FC,$5F,$33,$FA,$F7,$F2,$31,$F8,$19,$FF
	.dc.b	$15,$B5,$FF,$06,$A9,$E1,$AF,$4E,$6B,$5F,$8E,$F4,$6A,$E2,$FE,$EA
	.dc.b	$F4,$9C,$E1,$FC,$7A,$3B,$FC,$F6,$0C,$F4,$58,$F1,$FC,$77,$03,$0F
	.dc.b	$91,$EF,$83,$E7,$29,$B9,$FF,$22,$F1,$FE,$1D,$BC,$BE,$68,$FD,$9C
	.dc.b	$6B,$9F,$13,$D3,$C6,$93,$C0,$FF,$27,$FC,$F7,$C3,$6E,$DF,$02,$3E
	.dc.b	$47,$17,$EF,$1D,$4F,$F8,$D9,$CF,$0F,$F1,$FF,$9E,$7C,$9F,$C8,$6D
	.dc.b	$74,$3F,$CB,$BF,$AB,$BE,$23,$F1,$CD,$8F,$F2,$CB,$5F,$F2,$FB,$5F
	.dc.b	$F4,$4F,$CA,$3D,$FF,$CB,$36,$3F,$CC,$2D,$7F,$CC,$ED,$7F,$D2,$32
	.dc.b	$F8,$7F,$09,$23,$C2,$F4,$F7,$A4,$74,$6B,$FE,$72,$F8,$4F,$AC,$FF
	.dc.b	$9C,$1A,$E7,$93,$F5,$51,$88,$F0,$F9,$DF,$00,$8F,$C0,$7D,$7D,$F2
	.dc.b	$3D,$6D,$7C,$1E,$A7,$42,$3E,$17,$E5,$FC,$3C,$8F,$96,$D7,$E1,$F8
	.dc.b	$1F,$20,$8F,$C8,$7C,$FF,$3F,$23,$D8,$D7,$E9,$F5,$7D,$82,$3F,$61
	.dc.b	$FA,$7E,$0C,$8F,$A6,$D7,$E4,$7C,$9F,$C4,$23,$F2,$5F,$C1,$F9,$32
	.dc.b	$3F,$06,$D7,$ED,$F5,$7E,$00,$8F,$E0,$1F,$E3,$1F,$14,$F9,$FA,$49
	.dc.b	$FF,$52,$C6,$5F,$F5,$3F,$E3,$9F,$96,$7F,$D5,$AD,$7F,$DA,$ED,$7F
	.dc.b	$BE,$35,$FF,$11,$7C,$FF,$F2,$27,$FD,$62,$D7,$FD,$6E,$D7,$FD,$C2
	.dc.b	$D7,$FC,$0D,$F3,$EF,$03,$F9,$83,$63,$FD,$76,$D7,$FD,$82,$D7,$FD
	.dc.b	$D7,$F3,$9F,$E9,$27,$FD,$82,$D7,$FD,$8E,$D7,$7E,$A7,$F3,$DF,$9E
	.dc.b	$FF,$2F,$C7,$FB,$21,$AF,$FB,$39,$AF,$FB,$E1,$AF,$F8,$77,$E7,$9F
	.dc.b	$3C,$F9,$DF,$F6,$8C,$61,$B8,$7F,$80,$2B,$FD,$B3,$39,$7F,$7B,$E3
	.dc.b	$FF,$7A,$EF,$7A,$96,$30,$DF,$1B,$F8,$2C,$7F,$90,$A7,$FC,$11,$5F
	.dc.b	$BC,$8C,$1B,$7F,$D9,$71,$FD,$B6,$7F,$88,$CF,$79,$91,$FC,$53,$D2
	.dc.b	$B1,$C0,$BE,$4F,$F8,$72,$FF,$C2,$91,$9A,$FF,$7B,$F8,$9F,$E2,$98
	.dc.b	$FF,$32,$3F,$EF,$F4,$E5,$BF,$27,$FE,$18,$8C,$1B,$2B,$9F,$0B,$5F
	.dc.b	$27,$6F,$BB,$6F,$F8,$EC,$FF,$99,$2F,$FD,$FD,$1F,$C9,$A3,$06,$F1
	.dc.b	$FF,$1A,$CF,$F9,$0A,$FF,$CA,$CA,$7C,$9F,$03,$27,$DB,$F2,$77,$0F
	.dc.b	$F7,$AD,$FF,$C0,$19,$F8,$35,$EB,$AD,$61,$DB,$E4,$BD,$DE,$87,$BD
	.dc.b	$B7,$F2,$1B,$8F,$46,$D7,$F9,$2B,$E8,$BF,$D5,$EF,$A3,$D3,$FF,$1F
	.dc.b	$C6,$2F,$D3,$18,$F8,$AF,$F4,$2C,$FF,$AD,$30,$FF,$39,$35,$FF,$4C
	.dc.b	$35,$3C,$DC,$C7,$D5,$F5,$7A,$5F,$DE,$F8,$2F,$B0,$7D,$9F,$5A,$C6
	.dc.b	$49,$7F,$1A,$32,$FC,$8E,$9F,$C0,$7C,$0F,$E0,$5B,$5F,$8A,$6C,$EE
	.dc.b	$3F,$1B,$3F,$E5,$EE,$2F,$E0,$D1,$FE,$B1,$D7,$FC,$DF,$0F,$92,$D7
	.dc.b	$81,$B6,$D7,$F5,$58,$FF,$54,$CF,$FA,$EB,$0F,$C0,$46,$0D,$AB,$EB
	.dc.b	$F0,$F8,$7B,$FC,$F7,$8C,$F2,$7C,$15,$C7,$C8,$7F,$A8,$71,$4E,$70
	.dc.b	$E9,$39,$AC,$5F,$AB,$DD,$F7,$7C,$87,$BD,$B5,$D2,$FF,$DA,$18,$A7
	.dc.b	$31,$D0,$CB,$8B,$FF,$49,$FC,$F3,$E7,$BF,$D0,$EB,$CE,$B1,$86,$E3
	.dc.b	$DF,$05,$6E,$30,$F3,$6C,$F0,$9F,$11,$C1,$C5,$F8,$9B,$3F,$EC,$5A
	.dc.b	$E9,$5C,$1E,$8D,$9D,$BF,$EE,$F8,$DD,$7B,$2F,$D1,$7E,$46,$D7,$36
	.dc.b	$CE,$E7,$CE,$6C,$B3,$B6,$C6,$E3,$C5,$1B,$69,$B2,$C6,$CF,$74,$6D
	.dc.b	$B1,$B6,$9B,$8F,$85,$7E,$E9,$F9,$11,$C4,$C9,$5F,$85,$1B,$BF,$87
	.dc.b	$3F,$2E,$FD,$ED,$3D,$65,$AF,$3B,$6F,$C6,$D3,$BC,$79,$DB,$1E,$78
	.dc.b	$F0,$C7,$8D,$F3,$9B,$32,$53,$31,$C5,$B1,$DD,$DC,$66,$BD,$71,$F7
	.dc.b	$EB,$D6,$DA,$F5,$B4,$F5,$96,$BD,$6D,$8F,$49,$DD,$F0,$99,$3C,$2E
	.dc.b	$6B,$C3,$1F,$A0,$AF,$0B,$64,$C3,$E6,$14,$D1,$4A,$66,$7F,$3A,$D9
	.dc.b	$4E,$E3,$67,$43,$C9,$1F,$B8,$8F,$C0,$3F,$81,$69,$F4,$E7,$9B,$63
	.dc.b	$4D,$34,$70,$9F,$AD,$3C,$1B,$3C,$1B,$3C,$0E,$4F,$07,$95,$7D,$96
	.dc.b	$9F,$2C,$B1,$F3,$CF,$E1,$6E,$1F,$8F,$1F,$C5,$A3,$15,$FC,$0A,$31
	.dc.b	$1E,$07,$B3,$4E,$C5,$9E,$D3,$EB,$9F,$C6,$57,$AD,$B5,$EB,$AF,$69
	.dc.b	$EB,$79,$3E,$47,$F8,$B1,$E4,$9F,$A4,$EA,$BC,$CE,$8C,$46,$A7,$F4
	.dc.b	$5E,$5B,$E4,$7E,$FC,$7F,$26,$8C,$57,$F2,$F5,$F3,$7A,$C7,$74,$9D
	.dc.b	$BC,$6F,$57,$F9,$39,$CD,$70,$79,$C7,$59,$DD,$7E,$6A,$5F,$3D,$88
	.dc.b	$ED,$F3,$D6,$73,$7E,$58,$C5,$65,$B1,$93,$FA,$47,$8B,$E0,$BF,$E8
	.dc.b	$6B,$E7,$BB,$39,$CB,$DA,$E6,$B8,$E7,$F0,$2F,$F5,$D3,$CD,$FD,$73
	.dc.b	$5C,$C6,$DF,$DA,$C7,$59,$C9,$E4,$CF,$C3,$5E,$DF,$9D,$93,$F4,$7E
	.dc.b	$7A,$E1,$F5,$66,$E3,$F2,$E6,$D7,$CF,$F8,$2F,$8D,$F3,$DF,$1D,$B1
	.dc.b	$F6,$DB,$87,$DF,$6D,$FC,$0A,$F8,$78,$7C,$46,$DF,$CA,$7D,$96,$CD
	.dc.b	$18,$6C,$FF,$09,$6C,$FF,$11,$6D,$F7,$9F,$AB,$87,$77,$F2,$1E,$FF
	.dc.b	$DF,$F8,$2F,$C0,$36,$32,$7E,$56,$32,$D3,$F3,$1E,$DF,$E1,$CF,$C4
	.dc.b	$7B,$D6,$32,$4C,$BF,$41,$19,$7F,$0B,$ED,$FE,$12,$BF,$0A,$DA,$FC
	.dc.b	$5B,$EF,$23,$EF,$6C,$F2,$8F,$C8,$FA,$3C,$9F,$D6,$3F,$9B,$9F,$CE
	.dc.b	$36,$32,$7E,$C6,$32,$D3,$F6,$9E,$8F,$E8,$CF,$D2,$7C,$3D,$8C,$92
	.dc.b	$FF,$79,$19,$7F,$C4,$7F,$39,$FE,$A3,$DF,$FA,$4D,$7F,$8B,$63,$13
	.dc.b	$FB,$15,$E9,$7E,$01,$B3,$FD,$15,$B3,$FD,$35,$AF,$F5,$17,$FD,$60
	.dc.b	$4F,$9B,$E9,$5F,$BD,$FA,$AF,$F6,$B5,$FF,$4F,$6D,$FF,$57,$6B,$FD
	.dc.b	$6D,$AF,$F5,$FB,$7B,$5E,$52,$57,$95,$F4,$79,$F2,$3D,$0D,$AF,$EC
	.dc.b	$9D,$3F,$B9,$7F,$B5,$B5,$FE,$4D,$D3,$FA,$08,$C7,$CE,$BD,$6F,$D1
	.dc.b	$B7,$0F,$ED,$0D,$BF,$EE,$0D,$7F,$BA,$78,$AD,$BF,$56,$DA,$FE,$E0
	.dc.b	$D9,$FE,$F0,$D7,$FB,$E7,$D5,$23,$F6,$6D,$9F,$EF,$0D,$9F,$F0,$03
	.dc.b	$5F,$F0,$27,$CF,$7F,$38,$6E,$5F,$BC,$8C,$47,$F8,$2D,$AF,$F8,$3B
	.dc.b	$E7,$BE,$21,$34,$FE,$06,$FC,$7F,$AA,$7C,$67,$F0,$66,$E5,$FE,$0B
	.dc.b	$6D,$FF,$84,$DA,$FF,$85,$FE,$7F,$FD,$84,$4F,$5B,$F8,$7B,$FE,$C5
	.dc.b	$F9,$E7,$CF,$FF,$12,$6E,$1F,$E1,$76,$DF,$F8,$7D,$AF,$F8,$8F,$EA
	.dc.b	$BF,$8C,$4F,$F8,$75,$B1,$FE,$25,$6B,$BE,$9F,$E3,$FF,$41,$FE,$BB
	.dc.b	$AF,$F1,$23,$67,$FC,$58,$D7,$FC,$60,$D7,$FC,$69,$EF,$BD,$4F,$E5
	.dc.b	$2D,$AF,$F1,$73,$67,$FC,$6C,$D7,$FC,$75,$F5,$4F,$17,$F9,$8B,$6B
	.dc.b	$FC,$6C,$D9,$FF,$1F,$35,$FF,$21,$7D,$0B,$AF,$F9,$CB,$6B,$FC,$7C
	.dc.b	$D9,$FF,$23,$35,$FF,$25,$7D,$0B,$DD,$FE,$82,$DA,$FF,$23,$36,$7F
	.dc.b	$C9,$CD,$7F,$CA,$5F,$42,$F8,$DF,$E9,$2D,$AF,$F2,$73,$67,$FC,$AC
	.dc.b	$D7,$8E,$7C,$09,$3C,$1F,$EA,$7F,$43,$DE,$7F,$CC,$4D,$7F,$CC,$AD
	.dc.b	$4D,$FA,$1F,$D6,$47,$F9,$A5,$AF,$F9,$B1,$A9,$A7,$F1,$0A,$FF,$6A
	.dc.b	$3F,$B7,$F3,$BF,$C4,$AE,$67,$FD,$A2,$E6,$7F,$CE,$DF,$42,$F2,$7F
	.dc.b	$C3,$CD,$8F,$F6,$F3,$F9,$4B,$FF,$17,$B9,$BF,$F6,$DB,$99,$E7,$C9
	.dc.b	$FE,$27,$F8,$67,$BB,$D2,$3F,$C5,$58,$3E,$46,$98,$E3,$B8,$7F,$BB
	.dc.b	$67,$F0,$47,$F8,$F8,$AF,$FA,$37,$07,$D0,$F1,$8F,$F7,$A6,$6F,$3B
	.dc.b	$FC,$70,$D8,$CF,$9F,$89,$CB,$23,$8F,$9F,$6B,$FD,$EE,$1E,$67,$B7
	.dc.b	$E8,$6C,$F8,$38,$3B,$9E,$0D,$AC,$79,$1C,$3E,$6B,$7D,$07,$FA,$A3
	.dc.b	$E8,$5E,$9F,$F8,$EA,$B5,$3F,$EA,$1A,$FF,$56,$B5,$E2,$65,$FF,$53
	.dc.b	$7D,$03,$E6,$DF,$43,$86,$D7,$FA,$C6,$75,$39,$38,$FC,$DB,$8E,$92
	.dc.b	$B8,$46,$5E,$5F,$35,$FE,$BF,$9F,$F3,$9D,$FD,$D9,$FF,$61,$C6,$39
	.dc.b	$B5,$9F,$FB,$16,$97,$8B,$FF,$64,$C7,$17,$C5,$3F,$7F,$E1,$B5,$0E
	.dc.b	$64,$62,$B8,$63,$E2,$E7,$C7,$C7,$4C,$3C,$78,$F8,$67,$11,$D8,$F0
	.dc.b	$FC,$77,$33,$37,$EC,$7C,$39,$3E,$76,$9B,$AE,$A7,$FB,$8B,$8B,$F1
	.dc.b	$0E,$0F,$BB,$CB,$EA,$8F,$D8,$46,$5D,$BE,$33,$67,$1A,$7C,$67,$88
	.dc.b	$E5,$5F,$13,$E8,$7E,$5D,$77,$7E,$1E,$9F,$EF,$36,$C7,$FB,$D9,$DB
	.dc.b	$DD,$CA,$FA,$07,$D0,$FD,$06,$C7,$5E,$3F,$A7,$F7,$1F,$BE,$8F,$77
	.dc.b	$EB,$2E,$EF,$AE,$D8,$F8,$53,$B7,$DD,$19,$6F,$7F,$03,$8D,$76,$F0
	.dc.b	$3B,$F0,$CA,$E7,$1F,$84,$F1,$BE,$5C,$74,$F0,$7C,$26,$78,$3C,$7E
	.dc.b	$07,$91,$93,$E8,$8F,$C6,$FD,$0F,$F8,$8E,$3E,$C3,$FE,$2D,$6F,$BF
	.dc.b	$AF,$7B,$F2,$7F,$6D,$FE,$C6,$9F,$82,$7C,$03,$3F,$1B,$F9,$8F,$93
	.dc.b	$E9,$1F,$75,$FF,$2D,$BF,$85,$6B,$9A,$FC,$FF,$DC,$FF,$46,$8F,$C5
	.dc.b	$3F,$C8,$1B,$B6,$BC,$CF,$D2,$FA,$BF,$95,$9E,$99,$F8,$4C,$F0,$F8
	.dc.b	$FE,$53,$C9,$F0,$B2,$BC,$51,$FA,$EF,$B2,$79,$17,$C8,$95,$F9,$B9
	.dc.b	$CB,$F9,$DF,$D6,$1F,$1F,$EF,$F9,$C9,$39,$8F,$85,$E2,$7C,$3F,$6A
	.dc.b	$4F,$C1,$8F,$DE,$F9,$2F,$97,$BF,$2D,$2B,$F5,$53,$97,$F5,$9F,$0F
	.dc.b	$FA,$8F,$31,$E4,$A6,$75,$3E,$73,$E9,$1F,$C2,$AF,$85,$6E,$94,$FA
	.dc.b	$7F,$B1,$F0,$F2,$7E,$B4,$7F,$12,$FA,$5E,$91,$FB,$D7,$FC,$9E,$D9
	.dc.b	$D6,$3F,$C6,$3E,$99,$F1,$7F,$83,$F5,$E5,$FF,$05,$9C,$BF,$83,$E7
	.dc.b	$FC,$47,$44,$8C,$1F,$C2,$FE,$99,$F5,$7F,$C9,$17,$8D,$4C,$47,$F2
	.dc.b	$8F,$9C,$7C,$DF,$E4,$FD,$19,$87,$F1,$59,$CB,$F9,$3F,$63,$F3,$1C
	.dc.b	$F2,$30,$7F,$1B,$FA,$47,$D3,$3E,$8B,$E5,$4F,$CB,$9C,$C6,$23,$EA
	.dc.b	$43,$E9,$BF,$DF,$11,$98,$DB,$BF,$C3,$FF,$CA,$0F,$D5,$6B,$F7,$5A
	.dc.b	$9B,$D2,$F7,$7F,$07,$3E,$18,$C7,$5F,$8E,$3E,$FC,$E9,$F4,$CE,$9E
	.dc.b	$1F,$15,$97,$F0,$D1,$FE,$39,$24,$66,$3F,$1F,$F4,$95,$FC,$5B,$64
	.dc.b	$D3,$19,$CC,$3C,$93,$FE,$CC,$AF,$CC,$CF,$3F,$D6,$7F,$CF,$51,$F9
	.dc.b	$D6,$A6,$93,$E3,$F8,$12,$FE,$04,$7C,$83,$FA,$5B,$8F,$80,$EC,$47
	.dc.b	$7F,$11,$E1,$23,$83,$9E,$3E,$52,$38,$3B,$FC,$CF,$A4,$8E,$0F,$1F
	.dc.b	$C2,$FB,$48,$E0,$F3,$FA,$7F,$04,$8E,$0F,$B5,$F5,$BF,$9C,$6F,$4F
	.dc.b	$D1,$8C,$6B,$F9,$16,$BF,$96,$D3,$FB,$D5,$7C,$4F,$4B,$E7,$11,$C1
	.dc.b	$FD,$E6,$37,$3F,$DD,$5F,$F2,$8C,$66,$7F,$1F,$E4,$7E,$2E,$5F,$2D
	.dc.b	$F6,$49,$5F,$49,$FB,$9F,$2B,$C2,$3F,$89,$4F,$07,$C1,$19,$C3,$F0
	.dc.b	$15,$5F,$C0,$7D,$5D,$B3,$7E,$37,$0D,$C3,$F0,$31,$F8,$B2,$3F,$84
	.dc.b	$F0,$7A,$69,$FD,$E6,$3E,$04,$63,$E0,$BB,$3F,$93,$24,$71,$26,$3F
	.dc.b	$16,$3C,$9E,$17,$09,$F9,$5D,$DE,$73,$D1,$DD,$63,$E7,$36,$FE,$73
	.dc.b	$77,$F9,$9E,$77,$BF,$CF,$F8,$0D,$BF,$AC,$DB,$FA,$CD,$AF,$DF,$7D
	.dc.b	$87,$FC,$00,$6F,$99,$F2,$E3,$17,$FD,$0D,$FE,$90,$FF,$8E,$5C,$DF
	.dc.b	$F4,$D7,$2F,$F8,$DD,$CA,$73,$5E,$0F,$37,$B5,$F7,$6D,$F5,$6D,$F3
	.dc.b	$6D,$71,$9F,$15,$71,$69,$C4,$B5,$C5,$B3,$C6,$3C,$D3,$C5,$B3,$C5
	.dc.b	$A7,$12,$CF,$13,$E3,$C6,$DB,$1B,$69,$B8,$F9,$53,$DE,$33,$1F,$36
	.dc.b	$BD,$AD,$3C,$E5,$AF,$0B,$67,$8C,$7D,$19,$E2,$D9,$E2,$D3,$89,$67
	.dc.b	$8B,$F5,$67,$8B,$67,$8B,$4E,$25,$9E,$37,$F7,$8F,$BB,$3E,$53,$DE
	.dc.b	$7C,$52,$CF,$C5,$6C,$FC,$58,$FC,$55,$FC,$59,$C9,$CE,$32,$F3,$69
	.dc.b	$CC,$B3,$CD,$FC,$84,$F3,$73,$3C,$DC,$BC,$DA,$73,$2C,$F3,$AF,$6C
	.dc.b	$7C,$B3,$FB,$EA,$BF,$EB,$CD,$AF,$ED,$8D,$9F,$F0,$03,$5E,$33,$C0
	.dc.b	$93,$C1,$F4,$47,$F5,$B8,$B3,$F3,$E7,$F8,$F4,$61,$DF,$17,$F9,$02
	.dc.b	$3D,$93,$B7,$F9,$C1,$B3,$E0,$46,$EB,$E0,$4E,$DF,$E7,$93,$B7,$E2
	.dc.b	$4E,$EF,$E2,$5F,$82,$B5,$1F,$46,$75,$1F,$3E,$FA,$4F,$F4,$38,$D9
	.dc.b	$F4,$E7,$77,$F9,$58,$E1,$7E,$E7,$D6,$F7,$69,$EA,$39,$C7,$13,$FC
	.dc.b	$08,$E1,$FC,$C3,$FD,$30,$C5,$FF,$16,$AF,$0C,$E8,$E4,$53,$E9,$4F
	.dc.b	$81,$F8,$6F,$2A,$F3,$47,$CA,$7C,$CD,$8F,$31,$C6,$FE,$33,$C4,$FA
	.dc.b	$A7,$E4,$FE,$AD,$F2,$7F,$CC,$FC,$CF,$EB,$D1,$89,$FE,$65,$18,$9D
	.dc.b	$47,$FB,$B6,$75,$7F,$CC,$EB,$F9,$E7,$87,$2B,$5F,$5F,$6B,$F9,$C4
	.dc.b	$DC,$7F,$C8,$58,$2B,$17,$FE,$85,$AF,$F1,$3D,$FF,$BD,$57,$85,$FD
	.dc.b	$CD,$CE,$1F,$E4,$B7,$35,$FF,$00,$5E,$35,$FE,$66,$6B,$D7,$C2,$FF
	.dc.b	$24,$46,$23,$FD,$34,$F4,$9D,$1E,$FF,$8B,$F0,$57,$C6,$EE,$F1,$78
	.dc.b	$16,$BC,$0D,$C3,$D0,$DC,$7C,$CF,$A7,$BA,$57,$CB,$F4,$6D,$FB,$E6
	.dc.b	$DF,$C6,$9B,$5F,$17,$A3,$E1,$75,$7C,$36,$C7,$C1,$7D,$A4,$7F,$CE
	.dc.b	$5A,$FF,$A2,$9F,$F3,$A7,$03,$E7,$79,$47,$DD,$6E,$6C,$BE,$BF,$D4
	.dc.b	$7D,$1A,$FC,$3B,$5F,$C2,$3F,$E7,$3B,$34,$91,$FE,$83,$FC,$3F,$FD
	.dc.b	$6A,$DD,$3F,$CF,$51,$8C,$3F,$D3,$4D,$7F,$D3,$EE,$70,$FF,$45,$36
	.dc.b	$BF,$D9,$98,$3E,$17,$FC,$71,$39,$BF,$F6,$B2,$FA,$9C,$9A,$F7,$2E
	.dc.b	$1F,$ED,$BC,$7A,$C7,$90,$91,$E9,$7D,$D3,$C6,$3F,$48,$DB,$F9,$11
	.dc.b	$E4,$9F,$1F,$CC,$7E,$B9,$B1,$FA,$B7,$E4,$11,$FA,$AF,$C4,$F8,$B9
	.dc.b	$1F,$25,$B5,$F1,$9F,$79,$FB,$AF,$A9,$B3,$EB,$3F,$DE,$1F,$87,$BE
	.dc.b	$9F,$FC,$B8,$D8,$D5,$FF,$BE,$E1,$F1,$7A,$A9,$5A,$A5,$71,$7D,$E7
	.dc.b	$1A,$DB,$83,$6F,$FA,$DE,$3A,$CF,$FB,$15,$B3,$FE,$D9,$6B,$FE,$DE
	.dc.b	$FA,$6E,$DA,$FF,$B2,$5B,$49,$91,$CA,$75,$87,$F9,$E7,$EC,$1F,$5B
	.dc.b	$4F,$F3,$E7,$D8,$BE,$0F,$19,$7B,$7E,$BD,$EB,$7F,$A0,$5B,$19,$F0
	.dc.b	$31,$3F,$E8,$5F,$9D,$78,$14,$C7,$AB,$6F,$E4,$7D,$97,$89,$BA,$7F
	.dc.b	$A3,$7E,$D7,$F2,$91,$83,$E7,$7B,$1E,$EE,$3F,$73,$54,$FC,$19,$DF
	.dc.b	$0F,$FA,$5F,$ED,$9D,$14,$C7,$CB,$DA,$FB,$5E,$0F,$E9,$71,$FD,$BF
	.dc.b	$83,$FE,$9F,$F8,$DF,$F5,$24,$7D,$99,$D7,$D5,$FF,$AF,$BA,$7F,$DF
	.dc.b	$4D,$C3,$E9,$52,$1E,$57,$FA,$A5,$D4,$7D,$7F,$B8,$FE,$5C,$DE,$7F
	.dc.b	$F5,$74,$7C,$E8,$FC,$43,$A5,$C6,$F7,$BF,$D6,$8D,$C7,$F2,$9F,$72
	.dc.b	$F8,$2C,$D7,$FA,$E7,$EA,$FD,$7A,$AF,$CF,$18,$6F,$C0,$EF,$E0,$7F
	.dc.b	$D8,$1F,$74,$F9,$1A,$63,$DB,$B7,$FA,$3F,$BD,$76,$7F,$B1,$DB,$1F
	.dc.b	$A7,$FB,$EE,$4E,$E7,$FD,$99,$F2,$9F,$83,$E2,$FD,$33,$76,$F3,$47
	.dc.b	$E7,$6D,$F0,$FC,$3C,$0F,$F6,$A2,$F8,$6A,$63,$E1,$2E,$7F,$90,$F9
	.dc.b	$BF,$8E,$7F,$B6,$FE,$AD,$F3,$58,$9F,$DE,$47,$C9,$6F,$3E,$FE,$3F
	.dc.b	$83,$FE,$E3,$FA,$AF,$E5,$D3,$FC,$0D,$BB,$7E,$FD,$FC,$DB,$7D,$8E
	.dc.b	$1E,$BC,$AF,$E1,$5E,$57,$7A,$E1,$3F,$EE,$F8,$C7,$17,$C2,$94,$62
	.dc.b	$BF,$1E,$B2,$E1,$97,$97,$ED,$7C,$DF,$6D,$6F,$CC,$DC,$3C,$D3,$FE
	.dc.b	$FD,$FA,$DF,$D6,$6A,$8F,$77,$93,$30,$F7,$7D,$1D,$C3,$DC,$FC,$33
	.dc.b	$EB,$EB,$E7,$EC,$FC,$D3,$E9,$C5,$3A,$BF,$D3,$CB,$E5,$B1,$87,$F5
	.dc.b	$16,$BF,$DC,$7E,$56,$CE,$EB,$F9,$92,$8C,$9C,$31,$72,$E2,$BF,$9C
	.dc.b	$5B,$E9,$7A,$B6,$33,$F2,$38,$9C,$B2,$38,$F9,$F6,$BF,$9E,$1F,$C3
	.dc.b	$A3,$27,$F4,$5C,$3F,$28,$E2,$37,$B7,$80,$F0,$EB,$67,$CA,$E1,$F3
	.dc.b	$CB,$D4,$FE,$7D,$F5,$AF,$AB,$FE,$1F,$3F,$32,$B7,$58,$8C,$B5,$FE
	.dc.b	$2A,$D7,$FA,$43,$9F,$AC,$7C,$F3,$EB,$70,$DA,$E9,$3A,$9C,$9C,$7E
	.dc.b	$7B,$FC,$4D,$3A,$6B,$E6,$94,$62,$3F,$93,$C6,$38,$FC,$5F,$21,$F8
	.dc.b	$4C,$B3,$CB,$6F,$FA,$F5,$FE,$1D,$FE,$E4,$AB,$50,$8D,$3F,$DC,$78
	.dc.b	$3F,$AA,$BF,$57,$DB,$DC,$EB,$1F,$D7,$3F,$8B,$9E,$BD,$9A,$37,$CF
	.dc.b	$63,$C3,$D9,$C8,$C3,$FD,$FE,$BF,$99,$3F,$83,$F3,$7C,$B9,$1E,$11
	.dc.b	$FE,$0B,$E2,$FE,$AE,$DE,$5F,$5C,$1B,$CD,$87,$9F,$A7,$2B,$81,$FE
	.dc.b	$15,$FA,$DF,$88,$DE,$1F,$8B,$1E,$FE,$2F,$EA,$EE,$A7,$FC,$2C,$E5
	.dc.b	$FF,$06,$7D,$6F,$F5,$A7,$FC,$40,$D7,$75,$F4,$3E,$B7,$FC,$0F,$1F
	.dc.b	$8F,$6B,$F9,$36,$BF,$98,$6B,$FD,$A3,$EB,$5D,$9F,$95,$6C,$7E,$65
	.dc.b	$AF,$E7,$5A,$F3,$EE,$FC,$E5,$7F,$00,$8F,$14,$60,$FD,$07,$D6,$FC
	.dc.b	$B8,$FD,$23,$5F,$D4,$35,$FD,$6B,$5F,$37,$77,$94,$8F,$95,$FD,$67
	.dc.b	$C8,$FF,$92,$23,$F6,$0D,$7F,$68,$D7,$F7,$2D,$7F,$C1,$3F,$5B,$FE
	.dc.b	$79,$7F,$70,$D7,$F7,$8D,$7F,$80,$B5,$FF,$08,$FA,$FC,$DF,$F2,$23
	.dc.b	$5F,$F0,$47,$C5,$7F,$01,$8C,$7D,$6F,$F4,$C9,$FE,$0C,$D7,$F8,$53
	.dc.b	$5F,$E1,$ED,$7F,$C3,$FF,$5B,$F9,$B7,$F8,$73,$5F,$E2,$4D,$7F,$8B
	.dc.b	$B5,$F3,$FD,$67,$8C,$8F,$F1,$7A,$47,$F1,$7F,$AC,$FE,$13,$3F,$C6
	.dc.b	$DA,$FF,$1F,$6B,$FC,$91,$AF,$F8,$D3,$EB,$1F,$23,$FC,$85,$B1,$FC
	.dc.b	$95,$AF,$F2,$96,$BF,$CB,$AF,$FC,$44,$F9,$7B,$3E,$51,$1F,$E5,$DD
	.dc.b	$EF,$A1,$FE,$5C,$D8,$CC,$60,$FE,$65,$F5,$9F,$C0,$23,$F9,$C3,$5F
	.dc.b	$E7,$AD,$7F,$C9,$AD,$7E,$EF,$C7,$7D,$B2,$3F,$25,$FE,$7D,$F5,$8F
	.dc.b	$7E,$92,$7F,$A2,$46,$5F,$E8,$DF,$58,$F8,$BF,$E8,$CD,$8C,$C6,$0F
	.dc.b	$C5,$79,$52,$3E,$C3,$5F,$EA,$5F,$58,$F6,$FF,$A8,$36,$3F,$AA,$35
	.dc.b	$DC,$7F,$55,$FA,$C7,$CF,$D2,$4F,$F5,$B8,$CB,$F8,$7F,$33,$F3,$84
	.dc.b	$70,$7F,$61,$FC,$11,$EC,$FF,$61,$6C,$66,$30,$7F,$65,$FA,$D7,$D0
	.dc.b	$7F,$94,$27,$FB,$64,$63,$E0,$BF,$B6,$2F,$A9,$FF,$0E,$CE,$AB,$83
	.dc.b	$FD,$DB,$EB,$7F,$CB,$CD,$F8,$5F,$F1,$E3,$FE,$88,$8C,$B6,$32,$FF
	.dc.b	$AF,$7C,$8F,$CC,$F9,$BF,$C8,$FB,$B3,$29,$CB,$FC,$96,$33,$0F,$9D
	.dc.b	$A4,$AF,$E5,$8E,$B5,$91,$FC,$BD,$B7,$FC,$BF,$DF,$E6,$49,$FE,$68
	.dc.b	$F6,$F9,$E9,$1F,$CD,$DB,$3F,$E1,$3F,$C1,$FF,$B0,$4F,$F8,$4D,$B1
	.dc.b	$9F,$5B,$D7,$18,$8F,$F0,$EF,$D4,$BE,$9B,$DE,$DB,$FF,$0E,$36,$7F
	.dc.b	$C4,$0D,$7F,$C4,$B5,$FE,$32,$7C,$5F,$15,$D8,$95,$FD,$37,$EE,$3E
	.dc.b	$39,$1C,$1F,$D4,$7D,$AF,$96,$47,$07,$AF,$EB,$3D,$24,$70,$7F,$8C
	.dc.b	$7F,$36,$7A,$3F,$61,$B1,$98,$C1,$F7,$3E,$BB,$FD,$93,$1F,$E3,$96
	.dc.b	$E5,$FE,$3D,$6B,$FE,$42,$6B,$FE,$D2,$77,$F0,$1F,$E8,$96,$CE,$7B
	.dc.b	$FF,$CD,$B1,$8A,$05,$F3,$7E,$32,$57,$8D,$F3,$7C,$6F,$27,$9B,$63
	.dc.b	$9E,$1E,$1E,$5E,$E7,$83,$96,$CE,$8F,$43,$F0,$48,$FC,$17,$E2,$7C
	.dc.b	$4F,$A3,$7F,$4B,$67,$96,$7E,$AC,$FF,$CF,$8F,$C7,$9E,$33,$F2,$DB
	.dc.b	$0C,$7E,$34,$7C,$4F,$6F,$AD,$74,$AF,$89,$EC,$7C,$83,$E4,$B6,$35
	.dc.b	$87,$CA,$6C,$E9,$A7,$3F,$AE,$FE,$BA,$DF,$90,$FF,$4F,$C6,$2F,$C9
	.dc.b	$3F,$51,$47,$FB,$62,$39,$7A,$DC,$2B,$8B,$63,$B7,$9D,$D3,$D3,$EA
	.dc.b	$D8,$C9,$E3,$8C,$B4,$CE,$1E,$23,$F0,$B8,$7F,$BE,$9C,$1E,$4A,$E8
	.dc.b	$E2,$3F,$A8,$4F,$38,$C4,$EB,$5F,$F6,$E5,$6A,$78,$5F,$89,$B7,$E2
	.dc.b	$AD,$CE,$5B,$19,$3E,$35,$79,$63,$AD,$F7,$9C,$B6,$FD,$6D,$7D,$51
	.dc.b	$E7,$9E,$2D,$38,$96,$38,$B5,$E2,$7A,$0F,$7B,$9A,$F7,$4F,$A2,$72
	.dc.b	$6D,$EE,$D9,$EE,$D3,$B9,$4E,$F1,$CC,$E2,$58,$E2,$D8,$E2,$D7,$8C
	.dc.b	$7C,$8B,$F9,$8D,$3E,$64,$7E,$02,$32,$D8,$CB,$F9,$48,$CB,$63,$35
	.dc.b	$F4,$DF,$19,$D6,$3E,$0D,$78,$E7,$E2,$CE,$63,$D6,$D3,$8C,$F8,$4E
	.dc.b	$2E,$5F,$6B,$5F,$69,$B8,$E6,$D8,$E6,$F9,$CF,$A6,$7F,$BB,$27,$F2
	.dc.b	$8D,$BF,$ED,$F1,$88,$FA,$A6,$67,$FD,$DD,$19,$8F,$AE,$64,$F3,$9F
	.dc.b	$AA,$FA,$EF,$8F,$1F,$02,$3D,$26,$67,$E4,$D6,$AB,$6D,$8C,$9F,$3A
	.dc.b	$32,$D3,$33,$E0,$3F,$84,$C6,$9C,$52,$31,$7F,$1A,$77,$5C,$A3,$15
	.dc.b	$B8,$C1,$F2,$9C,$4F,$4B,$F3,$36,$7C,$8D,$3C,$37,$E0,$8E,$4D,$8F
	.dc.b	$86,$78,$8E,$17,$E2,$8D,$47,$DE,$35,$1E,$6A,$F0,$B6,$BC,$2D,$9F
	.dc.b	$0C,$7A,$9E,$F7,$EA,$9D,$1D,$E7,$53,$DE,$75,$1F,$07,$CD,$FE,$73
	.dc.b	$F1,$1F,$CE,$9B,$1F,$CF,$DB,$5F,$C9,$5B,$7F,$C8,$BE,$1F,$FA,$09
	.dc.b	$FD,$0B,$EA,$ED,$FF,$47,$8B,$7F,$CB,$66,$D2,$BF,$A3,$47,$F8,$8E
	.dc.b	$31,$3F,$DF,$21,$E3,$62,$73,$29,$E1,$5C,$AB,$4F,$4F,$07,$FC,$9F
	.dc.b	$3F,$E2,$DA,$C4,$FF,$55,$69,$4E,$93,$C9,$F1,$35,$F0,$7C,$2D,$AF
	.dc.b	$0C,$79,$FE,$57,$06,$7E,$2A,$E1,$FD,$CE,$91,$8A,$FE,$A7,$18,$9D
	.dc.b	$57,$5F,$98,$91,$C7,$EB,$2D,$F0,$6E,$9F,$13,$04,$62,$BE,$6F,$81
	.dc.b	$E7,$F5,$3D,$0D,$8F,$23,$F1,$88,$FC,$3C,$3E,$74,$61,$F3,$BF,$07
	.dc.b	$CE,$E7,$3E,$DF,$13,$E2,$1F,$15,$A7,$C1,$3E,$99,$0E,$D8,$FE,$0D
	.dc.b	$F0,$7D,$57,$FA,$A9,$BE,$27,$FA,$22,$31,$8F,$D7,$6B,$F6,$DF,$F6
	.dc.b	$14,$FF,$88,$5D,$E9,$F8,$87,$33,$FE,$02,$73,$3F,$E1,$75,$F2,$BF
	.dc.b	$7D,$78,$96,$30,$D9,$F3,$C7,$94,$8E,$0F,$13,$CF,$CF,$E1,$7E,$1F
	.dc.b	$13,$B1,$DF,$CD,$B7,$E9,$7E,$01,$1F,$8A,$FC,$3F,$0E,$47,$C8,$6D
	.dc.b	$7C,$5F,$8D,$F5,$47,$AD,$B3,$FE,$45,$FA,$A7,$D2,$7B,$EF,$F3,$D1
	.dc.b	$89,$F9,$51,$89,$FF,$5C,$CE,$63,$F5,$0D,$7F,$58,$D7,$FC,$84,$EF
	.dc.b	$6F,$DC,$B6,$B3,$F0,$FD,$5F,$F2,$C7,$D5,$71,$BF,$F3,$EB,$53,$4C
	.dc.b	$7D,$57,$FB,$4A,$3F,$D0,$6D,$4D,$31,$F5,$2F,$A9,$FF,$69,$B6,$FF
	.dc.b	$40,$DA,$FE,$08,$D7,$F8,$47,$CB,$FE,$DD,$FE,$10,$D7,$3D,$DF,$ED
	.dc.b	$E8,$C4,$7F,$98,$FE,$8F,$C8,$4F,$0B,$FD,$C6,$F9,$FC,$0F,$71,$38
	.dc.b	$3F,$DC,$EF,$C1,$FA,$87,$17,$FB,$A9,$BA,$7E,$DD,$B8,$FF,$1B,$6B
	.dc.b	$FC,$7F,$D2,$FD,$0B,$7E,$0F,$FD,$E7,$18,$8F,$F3,$C3,$5F,$F5,$6F
	.dc.b	$D3,$7E,$B9,$FF,$3E,$35,$CF,$6F,$FB,$F2,$31,$0F,$A6,$78,$DF,$F0
	.dc.b	$01,$B9,$7F,$0D,$6C,$FF,$31,$6B,$FC,$D7,$67,$C6,$7C,$62,$57,$F1
	.dc.b	$58,$CB,$FC,$5F,$E9,$5E,$47,$F4,$6B,$FF,$49,$46,$27,$53,$99,$FF
	.dc.b	$64,$CE,$5F,$E8,$5F,$4B,$FE,$87,$7F,$A3,$35,$FF,$4E,$35,$FF,$68
	.dc.b	$B5,$7D,$2B,$8B,$B4,$61,$B1,$9C,$3F,$97,$36,$7F,$A9,$7D,$2F,$E6
	.dc.b	$49,$CB,$C9,$FC,$E7,$C3,$3D,$EE,$AD,$C3,$F9,$C3,$6F,$FA,$E3,$5F
	.dc.b	$99,$F2,$FF,$90,$24,$FC,$B7,$CF,$F4,$8F,$46,$92,$7F,$A2,$46,$5F
	.dc.b	$E8,$DF,$45,$FC,$99,$BE,$57,$C0,$8C,$61,$FD,$B5,$AF,$F7,$17,$F9
	.dc.b	$7C,$F2,$F2,$B8,$11,$FE,$A5,$F4,$9D,$EB,$FB,$B3,$5F,$EF,$4D,$7F
	.dc.b	$DA,$0D,$42,$9F,$01,$FD,$ED,$B5,$FE,$E7,$5E,$5E,$67,$4F,$AC,$93
	.dc.b	$F0,$1F,$77,$CF,$F3,$3A,$7B,$D6,$7A,$38,$7F,$DE,$9E,$E7,$F0,$DF
	.dc.b	$33,$FC,$37,$8A,$BF,$C1,$D0,$F9,$DA,$63,$E5,$AE,$1F,$EF,$D9,$46
	.dc.b	$63,$AF,$15,$D7,$FD,$BE,$3E,$0B,$FC,$34,$D9,$F8,$B7,$FA,$7F,$43
	.dc.b	$35,$A6,$4E,$8F,$F1,$47,$8F,$8B,$EA,$F4,$3F,$C4,$DA,$FE,$47,$0F
	.dc.b	$D9,$47,$83,$69,$7C,$EF,$EA,$9F,$05,$C3,$F6,$23,$CF,$AC,$9C,$CF
	.dc.b	$DC,$F2,$7E,$EF,$D1,$BE,$13,$FC,$79,$7E,$EA,$57,$F8,$6D,$E0,$57
	.dc.b	$0E,$8F,$F2,$17,$D0,$BE,$8D,$F4,$7F,$E2,$38,$C3,$63,$33,$C1,$AF
	.dc.b	$F9,$36,$5F,$02,$DB,$93,$67,$BB,$5E,$0D,$7E,$2C,$BE,$07,$E3,$D7
	.dc.b	$F9,$8A,$3F,$0D,$3F,$21,$AF,$7D,$78,$D7,$07,$9F,$2D,$9F,$D2,$F0
	.dc.b	$FF,$95,$1F,$CA,$FA,$39,$58,$74,$C7,$C1,$CD,$98,$7C,$9D,$31,$16
	.dc.b	$BC,$9D,$FB,$30,$F9,$4F,$DA,$D7,$F0,$F7,$F4,$BF,$1B,$E1,$C8,$F1
	.dc.b	$8F,$C2,$F1,$FF,$62,$6F,$2F,$9E,$0D,$E6,$C3,$CF,$CD,$95,$C0,$FC
	.dc.b	$7F,$D1,$FC,$D6,$F0,$7C,$18,$F8,$1C,$3F,$96,$75,$3E,$90,$E9,$FE
	.dc.b	$1F,$F4,$6F,$A7,$F9,$4D,$8E,$A7,$E7,$BE,$93,$FC,$24,$DF,$67,$E8
	.dc.b	$46,$23,$B3,$FD,$A9,$FD,$2F,$C9,$BB,$BE,$B3,$63,$31,$CA,$B9,$F8
	.dc.b	$12,$7C,$0F,$6F,$13,$EF,$7D,$27,$C8,$9F,$D8,$B5,$F7,$37,$A7,$FD
	.dc.b	$63,$97,$D4,$9F,$1F,$81,$E2,$23,$E6,$E3,$FC,$57,$C3,$3D,$FA,$49
	.dc.b	$FD,$FC,$7F,$94,$9F,$E0,$3F,$49,$FE,$70,$6F,$95,$F9,$38,$C4,$7D
	.dc.b	$1F,$D6,$9F,$17,$F9,$7F,$63,$B4,$FF,$0A,$6B,$F5,$DB,$AF,$FB,$2A
	.dc.b	$BE,$64,$FD,$BF,$A5,$70,$7E,$85,$B1,$FC,$51,$AE,$FC,$3F,$D3,$7D
	.dc.b	$2F,$C3,$8F,$C2,$37,$0F,$F6,$BB,$7A,$BF,$8F,$E3,$F7,$67,$FD,$B7
	.dc.b	$93,$E4,$7D,$44,$C3,$F9,$1D,$7F,$88,$9F,$E4,$BF,$4B,$FC,$31,$BE
	.dc.b	$47,$ED,$E3,$11,$F9,$57,$F8,$D3,$FC,$B3,$E9,$5F,$15,$FB,$D6,$C6
	.dc.b	$63,$F8,$05,$FF,$86,$63,$E5,$78,$BF,$34,$8F,$93,$6F,$E0,$DE,$43
	.dc.b	$E9,$69,$27,$F9,$DC,$3E,$77,$F8,$67,$D2,$FE,$E6,$7F,$54,$DE,$0F
	.dc.b	$F7,$EB,$78,$7E,$3F,$27,$DD,$FA,$59,$7F,$75,$F3,$F7,$FF,$17,$FA
	.dc.b	$57,$CB,$D2,$57,$F4,$C8,$6F,$FC,$77,$E9,$5F,$15,$49,$5F,$43,$CE
	.dc.b	$FC,$67,$A5,$23,$F1,$8F,$C6,$F8,$5F,$E4,$DF,$4A,$F9,$7A,$4A,$FE
	.dc.b	$B7,$1F,$E3,$77,$FA,$EF,$D2,$BE,$33,$F9,$73,$63,$31,$F4,$BE,$67
	.dc.b	$F3,$9E,$C4,$9F,$CE,$3F,$5B,$EC,$3E,$BF,$C2,$FF,$5D,$ED,$ED,$5F
	.dc.b	$DB,$0F,$F1,$0B,$87,$0D,$7F,$B8,$FD,$37,$F8,$81,$FF,$80,$0D,$79
	.dc.b	$07,$0F,$56,$BF,$DE,$BE,$99,$F3,$8F,$88,$E3,$29,$CC,$61,$FC,$27
	.dc.b	$A1,$6B,$85,$78,$30,$F9,$9D,$9C,$9B,$19,$E1,$FF,$36,$C6,$23,$E0
	.dc.b	$7A,$BF,$9E,$AF,$0B,$63,$D6,$D7,$7E,$27,$2F,$3E,$57,$76,$CF,$0F
	.dc.b	$86,$FF,$3D,$DF,$F9,$22,$31,$5D,$A7,$C7,$94,$8F,$43,$67,$C3,$F3
	.dc.b	$3C,$89,$3E,$67,$FD,$0F,$F4,$7F,$D8,$23,$E4,$B6,$7E,$9B,$5F,$BC
	.dc.b	$D7,$F8,$6F,$8B,$F1,$09,$3F,$25,$EB,$C5,$23,$AB,$67,$E8,$78,$FF
	.dc.b	$30,$93,$F4,$5F,$83,$EF,$C8,$F9,$AD,$9F,$B1,$F3,$BF,$50,$93,$F6
	.dc.b	$5F,$F5,$1F,$D1,$37,$A4,$9F,$C3,$C6,$5F,$C4,$FD,$13,$EB,$FF,$D5
	.dc.b	$6D,$8C,$C6,$0F,$C1,$B0,$FC,$41,$70,$FC,$62,$3F,$96,$95,$FE,$F9
	.dc.b	$2F,$98,$C7,$99,$FD,$1E,$A9,$FE,$58,$9F,$C1,$47,$1A,$71,$BE,$C3
	.dc.b	$71,$FE,$92,$AD,$DF,$F9,$82,$B4,$FF,$98,$23,$F1,$93,$FB,$96,$BD
	.dc.b	$F9,$3E,$09,$F9,$2C,$BA,$C7,$81,$B3,$FE,$63,$AF,$F0,$3C,$63,$6F
	.dc.b	$92,$48,$F9,$AD,$AF,$82,$F2,$F8,$CE,$13,$FE,$10,$F8,$2E,$F9,$7F
	.dc.b	$87,$63,$D1,$D5,$89,$91,$B3,$25,$C3,$FC,$51,$F4,$4F,$A4,$E7,$E1
	.dc.b	$FA,$A3,$15,$FE,$1E,$8C,$4E,$A3,$FD,$0B,$3A,$AF,$F1,$96,$3F,$CA
	.dc.b	$E7,$3E,$5F,$9B,$CA,$E8,$D8,$C9,$FE,$2A,$8C,$B4,$CD,$74,$8C,$B6
	.dc.b	$33,$C5,$F7,$EF,$FA,$4C,$60,$FB,$95,$C2,$74,$D8,$C9,$BA,$FF,$6A
	.dc.b	$D7,$27,$E2,$DF,$9D,$B3,$E5,$69,$E3,$2D,$F6,$8E,$91,$F9,$2B,$F0
	.dc.b	$3D,$9A,$E6,$37,$7E,$1A,$FC,$D6,$1F,$D9,$A7,$C1,$18,$C3,$C8,$D8
	.dc.b	$C9,$FE,$BA,$9E,$17,$E5,$9D,$36,$75,$5C,$9A,$72,$8F,$F7,$BD,$72
	.dc.b	$AF,$64,$7B,$A7,$E4,$B4,$F6,$3E,$AC,$3D,$B8,$7A,$E3,$AC,$F8,$4E
	.dc.b	$B8,$78,$5B,$1E,$15,$87,$86,$3B,$E1,$EA,$3C,$65,$72,$76,$72,$D3
	.dc.b	$33,$D8,$F1,$1D,$AB,$C4,$D8,$F1,$57,$86,$3D,$A7,$BC,$FB,$05,$BF
	.dc.b	$73,$D6,$FC,$2D,$AF,$0B,$D4,$F0,$D7,$D6,$7F,$C0,$CD,$7E,$39,$8B
	.dc.b	$FE,$38,$E0,$87,$AC,$D4,$FF,$7E,$9F,$44,$77,$AF,$C5,$F9,$BF,$90
	.dc.b	$9C,$91,$C5,$E5,$B7,$F9,$A9,$FF,$1D,$B6,$74,$D3,$53,$C2,$FE,$35
	.dc.b	$70,$6D,$70,$6D,$70,$39,$3C,$2B,$95,$FD,$D8,$FC,$9C,$62,$7E,$E4
	.dc.b	$61,$E3,$3B,$8F,$C4,$CE,$DF,$A4,$6C,$F0,$9D,$E7,$89,$D1,$E2,$E1
	.dc.b	$E2,$60,$E3,$3D,$EF,$DC,$DB,$F7,$36,$BD,$D5,$C1,$ED,$5C,$1B,$5C
	.dc.b	$1A,$70,$AE,$55,$E7,$3E,$6F,$AD,$FC,$DC,$FE,$71,$E5,$DA,$FE,$6B
	.dc.b	$16,$7F,$9F,$CD,$AF,$E1,$BF,$15,$FD,$2B,$EC,$3F,$A0,$37,$0F,$E7
	.dc.b	$AD,$BF,$E8,$ED,$AF,$ED,$DF,$48,$F9,$6F,$EC,$2D,$B5,$FF,$BD,$63
	.dc.b	$FD,$8B,$18,$AE,$35,$C3,$DF,$E2,$D9,$D1,$CA,$D7,$D5,$E9,$5C,$27
	.dc.b	$71,$E0,$9E,$EF,$86,$BC,$BF,$31,$FA,$A8,$D3,$5E,$AD,$93,$4E,$35
	.dc.b	$DA,$7C,$7E,$D6,$1F,$77,$A5,$FD,$D3,$BB,$E1,$B6,$B8,$60,$8E,$C4
	.dc.b	$8F,$F3,$AC,$65,$BE,$57,$CC,$3E,$6F,$83,$63,$9B,$70,$F9,$CD,$D7
	.dc.b	$EA,$BE,$B2,$4F,$67,$DD,$E1,$48,$F0,$B6,$BF,$D1,$F8,$23,$EE,$9F
	.dc.b	$7A,$69,$F0,$F9,$9F,$05,$FE,$14,$3F,$C4,$7F,$49,$FE,$03,$C3,$FC
	.dc.b	$0B,$18,$EB,$FA,$13,$99,$FF,$53,$35,$FC,$03,$5F,$C5,$BD,$AF,$FD
	.dc.b	$36,$D8,$FF,$52,$B5,$E7,$5D,$BE,$5B,$C6,$F9,$1B,$5F,$6E,$91,$B8
	.dc.b	$EA,$DA,$EB,$F2,$DE,$82,$4F,$89,$B0,$D7,$FC,$A5,$69,$F8,$7E,$5F
	.dc.b	$C4,$7F,$C7,$38,$4C,$3C,$5F,$17,$71,$E5,$1F,$0A,$3E,$39,$1F,$82
	.dc.b	$FC,$A8,$91,$F1,$1A,$7F,$B2,$3E,$91,$F4,$3F,$52,$3F,$CA,$F1,$89
	.dc.b	$F8,$51,$89,$91,$F9,$F9,$7C,$4E,$2B,$CA,$7E,$FB,$E9,$39,$38,$1E
	.dc.b	$2F,$F0,$2B,$FD,$44,$EB,$6F,$F3,$B4,$E6,$B8,$7D,$27,$EF,$67,$FC
	.dc.b	$F5,$F5,$EF,$A6,$FF,$36,$CE,$71,$FE,$09,$39,$7F,$69,$F4,$CF,$92
	.dc.b	$FF,$39,$BF,$C0,$B0,$FF,$42,$39,$9D,$F8,$1F,$E7,$8F,$9B,$77,$DF
	.dc.b	$03,$0D,$AF,$E0,$1F,$47,$E1,$DB,$F8,$EF,$D7,$FE,$47,$3F,$F4,$84
	.dc.b	$E6,$53,$9A,$FE,$35,$87,$C1,$7F,$8E,$47,$C6,$7F,$86,$46,$67,$AF
	.dc.b	$07,$FA,$33,$E9,$DF,$29,$79,$F0,$DC,$3F,$8A,$78,$9F,$E1,$2C,$1F
	.dc.b	$5F,$FE,$D7,$D7,$FD,$49,$39,$8C,$57,$F2,$DF,$A8,$FD,$B5,$7F,$AA
	.dc.b	$5F,$F5,$E5,$FD,$B7,$33,$BF,$53,$FD,$3F,$F5,$0F,$9D,$BC,$98,$6D
	.dc.b	$7C,$8F,$95,$FF,$52,$FD,$7B,$E2,$3F,$D4,$FC,$A9,$CC,$7F,$AA,$BE
	.dc.b	$A7,$F9,$16,$BF,$8B,$5F,$21,$9F,$3B,$4F,$FA,$C3,$EA,$7F,$C3,$E3
	.dc.b	$EA,$31,$EB,$65,$FF,$5A,$FD,$4F,$F3,$CD,$BF,$9D,$7D,$7E,$DF,$E7
	.dc.b	$91,$9C,$FF,$35,$39,$7F,$39,$5F,$76,$1F,$63,$FE,$C0,$8C,$3F,$EC
	.dc.b	$2F,$8D,$7C,$3D,$F8,$FC,$37,$2F,$C5,$EB,$FE,$C8,$FA,$B7,$D9,$7C
	.dc.b	$48,$C3,$9F,$6F,$5F,$37,$6F,$FA,$95,$A3,$FD,$A3,$F5,$9F,$8F,$DF
	.dc.b	$FB,$8B,$FE,$E9,$AF,$EB,$0E,$60,$E5,$FD,$C3,$FC,$8A,$37,$EB,$74
	.dc.b	$F4,$FA,$BF,$ED,$EF,$9B,$65,$9A,$FE,$C9,$18,$7B,$D7,$93,$E2,$A4
	.dc.b	$78,$BD,$6B,$97,$81,$BE,$1F,$FB,$E7,$3F,$E5,$15,$E8,$BF,$6F,$D4
	.dc.b	$48,$F5,$F9,$57,$1F,$4B,$F0,$8F,$F7,$9F,$4F,$EF,$E3,$33,$FD,$E7
	.dc.b	$17,$D4,$7F,$BE,$1B,$19,$E0,$C4,$E5,$91,$FE,$01,$BF,$C9,$BB,$3F
	.dc.b	$C0,$96,$F8,$A4,$61,$B1,$FC,$94,$79,$78,$C7,$FC,$FB,$36,$BF,$C4
	.dc.b	$15,$A7,$AD,$66,$3F,$C4,$9E,$5D,$7F,$C5,$F1,$5F,$E0,$F3,$E4,$9F
	.dc.b	$87,$CF,$97,$5A,$C9,$F4,$EB,$6F,$CC,$F3,$3A,$2F,$43,$FC,$7B,$F5
	.dc.b	$8F,$8C,$F3,$D6,$A7,$D3,$5E,$A6,$BF,$E3,$B6,$BF,$E4,$67,$3F,$54
	.dc.b	$FA,$B7,$D6,$61,$B5,$D2,$75,$39,$38,$FC,$57,$F7,$19,$D3,$5F,$97
	.dc.b	$28,$C4,$7F,$93,$63,$1C,$3E,$2F,$8A,$FF,$28,$3F,$62,$53,$F8,$87
	.dc.b	$FC,$84,$D7,$FB,$02,$9E,$3A,$F5,$E5,$B5,$FE,$64,$E4,$FE,$9E,$FE
	.dc.b	$2F,$D7,$CB,$DA,$6E,$1F,$BD,$8E,$FE,$56,$0C,$F3,$D8,$F1,$F6,$F0
	.dc.b	$30,$F9,$1E,$B8,$BE,$7A,$9B,$9F,$F2,$E7,$1F,$F4,$E6,$F2,$F9,$20
	.dc.b	$DE,$6C,$3D,$3C,$F9,$5C,$0F,$F3,$37,$D6,$7C,$36,$ED,$EE,$8F,$91
	.dc.b	$C5,$F3,$5D,$4F,$F8,$25,$CF,$0F,$FA,$07,$EB,$1F,$25,$F1,$9B,$5D
	.dc.b	$0F,$F3,$A7,$D5,$FF,$80,$5B,$EF,$7C,$C8,$C4,$7F,$85,$5F,$EB,$0F
	.dc.b	$F9,$67,$BB,$E8,$FC,$4F,$ED,$23,$07,$23,$AF,$93,$CC,$8F,$83,$C6
	.dc.b	$FB,$3F,$59,$F1,$A3,$FD,$0C,$D7,$FC,$48,$DD,$BF,$19,$8F,$D3,$9F
	.dc.b	$F3,$57,$81,$E2,$23,$FE,$6D,$E3,$FC,$2F,$C9,$BE,$46,$92,$7F,$D2
	.dc.b	$90,$37,$E5,$48,$CB,$CB,$DE,$E7,$F3,$92,$78,$BD,$BC,$CE,$FE,$57
	.dc.b	$83,$EB,$E5,$76,$8F,$1F,$C6,$79,$3C,$E9,$3E,$28,$F3,$F7,$7A,$3E
	.dc.b	$BA,$4F,$9A,$3D,$7C,$FE,$CF,$3A,$4F,$AA,$3D,$FE,$07,$C0,$E2,$93
	.dc.b	$EE,$8F,$86,$FC,$4E,$38,$FC,$23,$F5,$7F,$AD,$FF,$96,$1F,$F5,$C3
	.dc.b	$5F,$F2,$B3,$79,$BF,$7C,$FF,$99,$A3,$7F,$31,$C8,$8F,$2F,$4F,$AF
	.dc.b	$C2,$48,$EA,$F8,$3E,$B3,$C3,$E3,$F8,$BD,$99,$5E,$28,$F2,$FC,$67
	.dc.b	$9B,$E6,$64,$F9,$A3,$D3,$E9,$FA,$BE,$A6,$4F,$AA,$3D,$BE,$94,$8F
	.dc.b	$6B,$F0,$3D,$3F,$83,$E8,$49,$F8,$2F,$C4,$F3,$7E,$2F,$A1,$27,$E2
	.dc.b	$BF,$23,$D1,$F9,$3F,$2D,$FC,$DB,$93,$FD,$E3,$5F,$CE,$9A,$FF,$3E
	.dc.b	$6B,$FD,$61,$FF,$12,$BB,$F7,$79,$11,$E5,$EE,$F5,$23,$83,$FA,$3F
	.dc.b	$D1,$BE,$89,$38,$6C,$FF,$4E,$97,$D0,$FF,$A4,$23,$13,$A7,$FB,$5F
	.dc.b	$D8,$BE,$77,$FE,$00,$37,$D6,$FE,$AA,$D9,$FE,$B2,$D7,$FB,$9F,$A6
	.dc.b	$F2,$FF,$A2,$40,$91,$9A,$FE,$BE,$D8,$FE,$C6,$D7,$9F,$95,$C9,$E2
	.dc.b	$D5,$08,$C3,$67,$AF,$95,$82,$38,$3C,$3E,$57,$02,$38,$3F,$B9,$FC
	.dc.b	$47,$CC,$6E,$FE,$18,$C6,$3F,$DE,$9A,$FF,$85,$3E,$8F,$FA,$55,$7F
	.dc.b	$7C,$6C,$7F,$80,$1A,$EF,$DB,$F6,$7D,$8B,$E3,$69,$2F,$FC,$0B,$19
	.dc.b	$7A,$7C,$1F,$C3,$C3,$FC,$D7,$18,$26,$1F,$33,$CF,$F8,$A4,$8F,$A2
	.dc.b	$FC,$8F,$3E,$47,$D4,$6C,$F9,$FE,$77,$B1,$27,$D8,$FD,$3F,$A5,$FC
	.dc.b	$14,$FD,$F0,$E1,$FC,$13,$5C,$F2,$FD,$BF,$A2,$FE,$B3,$7F,$E2,$26
	.dc.b	$CF,$F8,$99,$AF,$F9,$01,$AF,$77,$7E,$6F,$F8,$4D,$B3,$9F,$9F,$FF
	.dc.b	$21,$46,$2B,$F2,$8F,$F9,$FC,$FF,$3F,$57,$1F,$5F,$7E,$BF,$07,$CC
	.dc.b	$4C,$3C,$CF,$A7,$9B,$A9,$DB,$9A,$CF,$E8,$6F,$C9,$1E,$F2,$47,$BD
	.dc.b	$B8,$A7,$73,$C4,$F9,$9C,$4F,$8A,$ED,$19,$AF,$5F,$AD,$EE,$3D,$FE
	.dc.b	$B5,$9D,$4B,$5F,$F3,$94,$61,$B1,$E1,$8F,$99,$3C,$63,$07,$C8,$3E
	.dc.b	$5F,$A9,$F3,$1B,$3C,$DA,$70,$3C,$06,$32,$FF,$24,$7D,$13,$D2,$FF
	.dc.b	$3B,$2F,$56,$C6,$1B,$7E,$A9,$D3,$67,$4D,$35,$3C,$2F,$C5,$5C,$1B
	.dc.b	$5C,$1B,$5C,$0E,$4F,$0A,$E5,$7E,$86,$CF,$E7,$EF,$A4,$FB,$2F,$A3
	.dc.b	$67,$A5,$F6,$3A,$3D,$A3,$A3,$4E,$97,$DA,$7F,$95,$5F,$69,$F0,$5F
	.dc.b	$66,$9D,$AF,$C1,$5D,$8F,$03,$D9,$B1,$D8,$F0,$3D,$AF,$C0,$AF,$C0
	.dc.b	$F8,$AF,$C0,$DA,$F0,$1E,$27,$C1,$7E,$23,$C0,$58,$F0,$47,$88,$F0
	.dc.b	$3E,$27,$C1,$7E,$2B,$F0,$C7,$8E,$FC,$2D,$AF,$0B,$E3,$3C,$2F,$8E
	.dc.b	$FC,$2D,$3C,$2F,$8E,$3C,$27,$8C,$F0,$C7,$8E,$BC,$27,$E5,$DE,$F1
	.dc.b	$E1,$3B,$9E,$19,$EE,$D8,$EE,$78,$5E,$F1,$E1,$7B,$9E,$19,$EF,$5E
	.dc.b	$E7,$C1,$1F,$EC,$86,$BF,$EE,$26,$BF,$44,$C4,$7F,$67,$8C,$C7,$C5
	.dc.b	$33,$5F,$ED,$08,$EB,$5E,$3D,$5F,$4B,$F4,$70,$FC,$1C,$62,$FE,$24
	.dc.b	$62,$75,$1F,$66,$75,$87,$AE,$B5,$3C,$0D,$57,$09,$FA,$15,$C2,$FE
	.dc.b	$26,$1E,$4B,$F9,$14,$AD,$46,$27,$53,$F3,$CF,$F0,$84,$E9,$C1,$F5
	.dc.b	$EB,$DD,$7E,$16,$DF,$96,$7C,$73,$98,$F0,$D7,$39,$E5,$1E,$1C,$0F
	.dc.b	$F2,$E9,$D6,$1E,$4A,$D9,$E2,$8C,$D7,$58,$F3,$DD,$76,$7B,$DC,$57
	.dc.b	$3A,$E3,$7E,$7C,$3F,$DE,$51,$8B,$FA,$13,$E3,$BE,$38,$79,$A7,$F9
	.dc.b	$A5,$6E,$FE,$7C,$E5,$E7,$5C,$2F,$91,$DC,$F8,$6F,$ED,$7D,$2F,$E7
	.dc.b	$27,$F3,$AF,$4A,$CF,$F3,$78,$B8,$FF,$32,$9B,$7F,$D0,$BD,$0F,$E8
	.dc.b	$3E,$87,$F4,$26,$E1,$FC,$FD,$B7,$FC,$E9,$B5,$FD,$2B,$CE,$FE,$92
	.dc.b	$7F,$4A,$F3,$6D,$FF,$47,$6D,$FF,$42,$6D,$7F,$52,$6C,$FF,$3B,$6C
	.dc.b	$FF,$96,$5F,$F1,$83,$FE,$1D,$31,$E0,$FF,$94,$E3,$31,$96,$BF,$E2
	.dc.b	$37,$8D,$7D,$18,$46,$2B,$5F,$0B,$F4,$5E,$17,$6B,$C5,$F0,$F5,$E3
	.dc.b	$F3,$17,$04,$D8,$C3,$E5,$25,$FA,$6B,$9B,$5F,$1F,$C8,$F3,$3A,$7D
	.dc.b	$3D,$9F,$F0,$4E,$08,$C5,$7F,$7D,$8C,$4E,$AB,$8B,$D6,$33,$3E,$16
	.dc.b	$BF,$1E,$BE,$57,$D1,$CA,$E0,$D9,$E0,$FB,$CF,$81,$E7,$D9,$F2,$F8
	.dc.b	$9F,$0C,$F8,$8D,$3D,$27,$D7,$21,$E5,$3F,$D1,$9F,$46,$F0,$E9,$23
	.dc.b	$E3,$46,$5B,$F0,$5F,$4E,$31,$8F,$F8,$7D,$46,$6B,$4F,$62,$57,$23
	.dc.b	$AF,$C1,$78,$88,$E0,$97,$D6,$3C,$1E,$8F,$85,$F1,$78,$72,$BF,$C7
	.dc.b	$F8,$BE,$0B,$18,$79,$5F,$79,$27,$DA,$FC,$2F,$06,$47,$C0,$6C,$FE
	.dc.b	$26,$9F,$05,$F0,$67,$E1,$36,$BC,$2E,$1F,$F2,$8F,$D9,$FF,$AD,$6F
	.dc.b	$E3,$4F,$E6,$FE,$A5,$F0,$94,$C4,$5B,$FF,$2C,$FD,$03,$C0,$FD,$13
	.dc.b	$63,$3F,$05,$89,$FD,$37,$D0,$7F,$B5,$FB,$BF,$60,$DA,$FF,$4B,$5F
	.dc.b	$FA,$EA,$73,$E7,$FE,$B7,$E8,$7F,$DD,$95,$FE,$68,$5D,$39,$E9,$D3
	.dc.b	$FB,$3F,$A1,$FF,$44,$D7,$FA,$85,$7C,$4E,$7D,$2D,$3B,$E0,$E1,$C1
	.dc.b	$6F,$FC,$EB,$7F,$31,$FF,$3B,$FE,$03,$FC,$19,$BD,$FF,$C0,$61,$E2
	.dc.b	$D8,$FE,$09,$F1,$2F,$8E,$A6,$3E,$3A,$E1,$FE,$82,$FA,$17,$D4,$7F
	.dc.b	$0B,$6C,$7F,$A1,$AF,$FA,$DB,$B9,$FE,$21,$F4,$1F,$D6,$CF,$F1,$99
	.dc.b	$7D,$3C,$FC,$5E,$BE,$2E,$DF,$FA,$42,$91,$FC,$67,$E7,$D9,$DF,$33
	.dc.b	$0D,$BF,$F4,$BF,$E0,$7F,$E9,$4E,$BF,$E5,$0B,$DF,$FE,$44,$E6,$7F
	.dc.b	$DA,$0E,$5F,$F4,$FF,$CF,$37,$FE,$50,$FF,$73,$31,$1A,$F5,$F3,$C7
	.dc.b	$AE,$4B,$3C,$9F,$E6,$1E,$FB,$CA,$A6,$3D,$FB,$5F,$EA,$EF,$9E,$7D
	.dc.b	$37,$F3,$76,$C6,$7E,$A3,$13,$9E,$49,$3C,$7D,$9B,$E5,$FF,$AA,$7E
	.dc.b	$81,$A7,$F4,$17,$F1,$F0,$0E,$AB,$FD,$7E,$EA,$33,$E8,$F1,$AE,$4D
	.dc.b	$AF,$F6,$2F,$37,$E3,$1E,$31,$FD,$3E,$31,$5E,$38,$F1,$6B,$2B,$BE
	.dc.b	$D6,$BD,$0D,$E6,$FE,$C1,$97,$C9,$AF,$2D,$FA,$FC,$29,$1E,$8F,$80
	.dc.b	$B8,$FC,$17,$E0,$1F,$D8,$7C,$D7,$CC,$E2,$BF,$DB,$B9,$7E,$39,$BC
	.dc.b	$1F,$D9,$E1,$DD,$63,$0C,$9F,$F7,$2D,$FE,$A5,$E7,$1F,$DB,$F1,$FC
	.dc.b	$71,$F3,$CA,$7D,$1F,$47,$5F,$07,$B6,$57,$FB,$C2,$7F,$78,$7F,$B8
	.dc.b	$C4,$75,$F1,$E4,$F8,$BE,$22,$D6,$F6,$F1,$1E,$3D,$AD,$75,$70,$F7
	.dc.b	$97,$CF,$FE,$17,$E8,$3F,$51,$1E,$9A,$F6,$3A,$3F,$C0,$45,$3F,$C0
	.dc.b	$93,$87,$07,$C1,$F9,$F7,$CB,$BE,$83,$FC,$15,$18,$6C,$66,$7F,$C2
	.dc.b	$13,$96,$FC,$23,$8F,$8B,$63,$31,$83,$AF,$5F,$D1,$7F,$C5,$55,$FC
	.dc.b	$32,$7F,$45,$18,$E4,$D6,$BE,$7B,$B5,$FA,$7B,$E0,$F8,$67,$EE,$72
	.dc.b	$7F,$8A,$A3,$F0,$7E,$37,$05,$5C,$31,$F0,$F2,$E6,$1F,$31,$4C,$45
	.dc.b	$AF,$2E,$3F,$27,$C1,$E0,$66,$3C,$47,$95,$F0,$64,$F8,$FA,$70,$9E
	.dc.b	$67,$F0,$1F,$23,$FB,$9B,$83,$F9,$37,$2F,$A6,$3F,$B6,$C6,$5D,$BE
	.dc.b	$03,$67,$E8,$BE,$1B,$C2,$71,$9F,$85,$F4,$0E,$0B,$C3,$2F,$C1,$39
	.dc.b	$6C,$7C,$9F,$9F,$7D,$05,$38,$4F,$87,$E7,$FF,$76,$FA,$5A,$E7,$BF
	.dc.b	$E9,$43,$3F,$A9,$F2,$DF,$AA,$AF,$80,$D7,$D8,$DE,$5F,$F3,$85,$74
	.dc.b	$6F,$A9,$D0,$95,$E1,$F1,$BF,$01,$F4,$1F,$22,$3E,$33,$6B,$FC,$F4
	.dc.b	$DE,$5F,$E0,$98,$7D,$7A,$F1,$F7,$79,$88,$F9,$BC,$1F,$C6,$7C,$37
	.dc.b	$F8,$1A,$3E,$C3,$5F,$A8,$DC,$7F,$D1,$0F,$F9,$32,$37,$E4,$F2,$23
	.dc.b	$CB,$DA,$EB,$F0,$D2,$3A,$BE,$0F,$2F,$C3,$E4,$F8,$BE,$62,$57,$8A
	.dc.b	$3C,$BF,$0F,$E6,$F3,$64,$F9,$A3,$D3,$DB,$EA,$F9,$59,$3E,$A8,$F6
	.dc.b	$F9,$9E,$EF,$36,$4F,$BA,$3E,$0F,$81,$F0,$B8,$A4,$FC,$28,$F8,$AF
	.dc.b	$C6,$E3,$8F,$C6,$3F,$69,$F3,$BF,$DB,$9F,$D0,$35,$FC,$E3,$79,$3F
	.dc.b	$D6,$0B,$D1,$E3,$ED,$4A,$E2,$F4,$FA,$0E,$BE,$C4,$9E,$AF,$83,$D7
	.dc.b	$F0,$F9,$1E,$2F,$A0,$95,$E2,$8F,$2F,$C9,$79,$BE,$26,$4F,$9A,$3D
	.dc.b	$3E,$97,$AB,$E8,$24,$FA,$A3,$DB,$E6,$48,$F6,$BF,$03,$D3,$F8,$3E
	.dc.b	$84,$9F,$82,$FC,$4F,$37,$E2,$FA,$12,$7E,$2B,$F2,$3D,$1F,$C8,$F9
	.dc.b	$DF,$D0,$3C,$6F,$DA,$D7,$FB,$0D,$AF,$FB,$1D,$AF,$FB,$85,$FF,$53
	.dc.b	$3B,$F8,$EE,$44,$7F,$D9,$9C,$EF,$A0,$FE,$9A,$D8,$CC,$60,$FE,$A3
	.dc.b	$F5,$7F,$E2,$D8,$FF,$6A,$35,$FF,$6C,$35,$FF,$79,$35,$FE,$E9,$F4
	.dc.b	$2F,$8A,$7C,$FF,$FB,$76,$30,$DF,$A1,$77,$E6,$B7,$7F,$EE,$5F,$9E
	.dc.b	$7D,$57,$FB,$DE,$3F,$C0,$11,$8F,$47,$FD,$D0,$D8,$E1,$3B,$7F,$C6
	.dc.b	$9E,$AB,$CF,$C4,$23,$17,$FE,$F1,$6C,$7F,$BD,$5A,$FF,$83,$C5,$FF
	.dc.b	$A4,$A7,$C3,$F1,$DD,$23,$13,$E9,$8E,$F4,$F3,$A9,$2F,$C0,$EB,$CE
	.dc.b	$91,$E3,$6D,$F8,$FC,$EE,$64,$9F,$43,$EC,$FA,$0F,$EA,$D1,$E7,$9F
	.dc.b	$F2,$A4,$61,$C3,$5F,$63,$5F,$9B,$BF,$AC,$93,$F3,$9E,$DF,$05,$23
	.dc.b	$B4,$A9,$39,$8F,$B5,$F3,$9F,$14,$95,$F6,$1F,$8B,$E8,$48,$FA,$4D
	.dc.b	$9F,$3C,$FB,$49,$3E,$0B,$47,$DC,$FA,$BF,$A3,$5F,$5D,$AF,$F9,$1D
	.dc.b	$AF,$F9,$DD,$AB,$E8,$1F,$51,$F8,$66,$E1,$EE,$6C,$FC,$D6,$BF,$85
	.dc.b	$E1,$FF,$5B,$BF,$DE,$DB,$E7,$AB,$FC,$B1,$5F,$E8,$39,$CC,$63,$D7
	.dc.b	$F0,$11,$F0,$3E,$3E,$1E,$58,$7F,$6C,$F8,$FB,$8E,$23,$D5,$38,$BF
	.dc.b	$F3,$37,$1F,$FB,$32,$7F,$C1,$D9,$7E,$4E,$7D,$95,$D5,$A4,$9D,$C7
	.dc.b	$37,$FD,$3D,$3F,$E8,$97,$33,$FE,$1F,$8F,$F6,$04,$76,$DF,$83,$F3
	.dc.b	$7D,$5B,$3A,$C7,$D0,$FF,$9E,$8C,$5F,$C6,$9D,$DF,$0E,$9E,$79,$7F
	.dc.b	$86,$AB,$C1,$18,$F0,$75,$1F,$E1,$26,$E5,$FE,$12,$7C,$7F,$42,$F4
	.dc.b	$B9,$F8,$58,$6D,$FF,$8E,$67,$4F,$F8,$72,$74,$D3,$53,$C2,$FF,$C7
	.dc.b	$55,$C1,$FF,$10,$D7,$06,$D7,$03,$93,$C2,$B9,$6B,$F6,$EF,$93,$F7
	.dc.b	$72,$61,$CE,$30,$DC,$3E,$FB,$70,$FE,$42,$D7,$F9,$03,$6B,$9C,$F8
	.dc.b	$E7,$CA,$78,$DF,$2D,$78,$F0,$F2,$C7,$FA,$C7,$0F,$2B,$E0,$39,$4F
	.dc.b	$91,$A6,$6B,$A1,$DA,$3A,$47,$63,$A5,$76,$9F,$F7,$65,$76,$7D,$73
	.dc.b	$D1,$A7,$4A,$F0,$36,$3C,$11,$DA,$3C,$11,$D8,$F0,$3D,$AB,$C0,$AB
	.dc.b	$C0,$FC,$89,$F0,$34,$EC,$78,$AB,$B1,$E2,$2C,$78,$8F,$04,$78,$8F
	.dc.b	$03,$E2,$7C,$15,$E2,$AF,$0D,$FD,$69,$EE,$78,$6F,$AB,$E1,$BE,$B1
	.dc.b	$E1,$69,$E1,$8E,$E7,$84,$EE,$F8,$6B,$BE,$1F,$85,$3F,$31,$3F,$D8
	.dc.b	$63,$F3,$8D,$C3,$FB,$5C,$62,$3D,$A6,$6B,$FA,$2C,$66,$3C,$26,$4F
	.dc.b	$5A,$F4,$3F,$2C,$FD,$02,$38,$AC,$B6,$BE,$9C,$F1,$6C,$E9,$A6,$A7
	.dc.b	$85,$78,$1B,$87,$CA,$8F,$C0,$46,$27,$E4,$C7,$E4,$A3,$0F,$CA,$9F
	.dc.b	$DB,$46,$23,$E5,$9F,$C0,$9C,$5F,$AA,$BC,$73,$E1,$6D,$78,$67,$D5
	.dc.b	$3F,$12,$FD,$4D,$AF,$51,$C0,$F2,$BE,$87,$0F,$95,$FD,$B5,$F8,$A7
	.dc.b	$F9,$0B,$63,$F8,$F3,$67,$D1,$1C,$9F,$5D,$FB,$A3,$17,$EB,$9F,$21
	.dc.b	$EB,$3E,$DF,$9F,$FC,$B8,$FE,$5F,$E7,$DA,$FE,$59,$16,$7F,$93,$CD
	.dc.b	$BF,$E4,$1E,$5F,$F3,$63,$F9,$BF,$99,$70,$FE,$69,$16,$FF,$97,$CD
	.dc.b	$AF,$E5,$1E,$57,$F4,$6F,$94,$FE,$7E,$DC,$3F,$9E,$3F,$98,$23,$F9
	.dc.b	$66,$DF,$F3,$A7,$FA,$3F,$C6,$F2,$9F,$F7,$E2,$FA,$0F,$EB,$AB,$E5
	.dc.b	$EC,$61,$B3,$C1,$B7,$FE,$FF,$AE,$13,$D3,$E8,$39,$9D,$3E,$2A,$C6
	.dc.b	$49,$5C,$63,$2F,$3F,$A0,$FE,$CB,$DB,$E4,$6D,$7A,$1F,$F3,$0E,$06
	.dc.b	$92,$3F,$C7,$70,$FA,$0E,$51,$F0,$1B,$7E,$9F,$23,$E0,$F9,$5F,$09
	.dc.b	$B5,$F1,$9E,$E4,$7A,$BE,$2F,$A0,$91,$E0,$6D,$7F,$96,$70,$47,$CE
	.dc.b	$3E,$7C,$D1,$2E,$1F,$EC,$DF,$07,$FE,$06,$3F,$C2,$3F,$41,$FC,$A6
	.dc.b	$33,$5C,$1F,$F1,$44,$7F,$2D,$AF,$F3,$F3,$5E,$A1,$C3,$E0,$6B,$9F
	.dc.b	$53,$BF,$67,$81,$B7,$DC,$E7,$4F,$84,$F2,$E3,$E6,$6C,$EF,$0E,$D1
	.dc.b	$FE,$98,$AE,$13,$FE,$70,$6C,$FF,$9F,$DA,$FA,$BC,$EE,$04,$9E,$0D
	.dc.b	$8F,$91,$1E,$48,$F8,$FE,$77,$C8,$7F,$C5,$F7,$27,$C9,$5F,$0F,$E0
	.dc.b	$B0,$F2,$3E,$23,$E7,$24,$79,$DF,$57,$C1,$48,$F5,$36,$7F,$D3,$1F
	.dc.b	$3E,$F8,$3F,$F2,$4B,$6B,$C2,$AB,$33,$FE,$4E,$F9,$E7,$81,$4C,$78
	.dc.b	$37,$0F,$4F,$CF,$FE,$C3,$2F,$40,$71,$C7,$FE,$6B,$C3,$FD,$B3,$5C
	.dc.b	$3B,$7F,$CB,$7F,$3F,$FE,$DD,$8F,$86,$FF,$B0,$2B,$FC,$E4,$E7,$0E
	.dc.b	$AE,$67,$F1,$7F,$3E,$ED,$FF,$32,$AF,$0A,$C6,$1C,$ED,$AD,$ED,$FD
	.dc.b	$2B,$F8,$4F,$D4,$F9,$F7,$5F,$F9,$C1,$B1,$9F,$0F,$13,$FE,$74,$F8
	.dc.b	$67,$C7,$53,$1F,$1D,$6F,$EF,$FC,$FF,$FA,$27,$87,$FD,$1C,$D8,$FD
	.dc.b	$2D,$FF,$A6,$B8,$7F,$67,$39,$7F,$6B,$F3,$FF,$D7,$2F,$FD,$2C,$FF
	.dc.b	$8B,$0C,$5F,$EB,$9C,$7C,$37,$FA,$21,$7C,$35,$8C,$3F,$E8,$CF,$9F
	.dc.b	$67,$7C,$9C,$37,$0F,$CB,$FC,$FB,$D6,$FF,$49,$BF,$E8,$ED,$3F,$84
	.dc.b	$B9,$9D,$FA,$DF,$E9,$8F,$9F,$70,$5F,$4F,$0D,$AF,$D1,$CF,$CE,$C3
	.dc.b	$F1,$6B,$93,$97,$27,$37,$FD,$47,$E3,$FF,$AA,$78,$FF,$D7,$0D,$C3
	.dc.b	$F8,$7B,$FE,$B5,$C3,$F2,$AE,$5F,$E3,$FF,$3E,$F9,$FF,$F5,$6B,$6B
	.dc.b	$F6,$AB,$E4,$73,$3A,$E4,$95,$AF,$52,$E1,$FA,$2F,$9F,$65,$FE,$B8
	.dc.b	$7F,$2D,$5F,$C0,$1C,$B7,$C0,$D7,$93,$BA,$B3,$B7,$FD,$83,$CD,$F0
	.dc.b	$E3,$94,$E3,$7D,$E3,$E1,$F7,$7C,$07,$87,$CB,$B7,$E1,$6C,$F9,$DB
	.dc.b	$C3,$FE,$E1,$D3,$F8,$34,$FA,$FC,$2F,$57,$B9,$EB,$6E,$1E,$B8,$FF
	.dc.b	$6A,$F9,$2F,$98,$CD,$FF,$04,$C7,$F2,$37,$FC,$81,$B1,$9E,$FF,$F6
	.dc.b	$EC,$60,$FF,$79,$07,$27,$6F,$0A,$D7,$F2,$AA,$FC,$81,$FC,$48,$B5
	.dc.b	$FE,$FB,$5E,$6F,$FB,$A9,$CC,$FF,$54,$72,$F3,$8D,$78,$DB,$6C,$77
	.dc.b	$8F,$7F,$C2,$E2,$BC,$1A,$F5,$3B,$6B,$6F,$CC,$E1,$F4,$4B,$E7,$BF
	.dc.b	$B8,$FC,$FB,$C1,$F2,$5F,$9E,$B4,$79,$CF,$ED,$75,$5A,$53,$67,$B7
	.dc.b	$E7,$9F,$24,$F9,$FE,$11,$86,$C6,$67,$B4,$E5,$BE,$FF,$F8,$15,$BC
	.dc.b	$1C,$A3,$11,$DD,$AF,$F8,$02,$5D,$BF,$E0,$2A,$6B,$9E,$EC,$36,$FF
	.dc.b	$B2,$DF,$06,$CE,$9F,$F0,$BF,$33,$C8,$C4,$23,$0F,$03,$9F,$36,$AD
	.dc.b	$F2,$74,$C4,$59,$F0,$F6,$6C,$C3,$FD,$7A,$BF,$8A,$3F,$DE,$AD,$F3
	.dc.b	$FE,$26,$C7,$23,$FB,$FF,$1F,$F6,$F6,$F2,$FA,$23,$FB,$B4,$6B,$9F
	.dc.b	$13,$CF,$E7,$64,$F0,$3F,$C1,$3F,$3F,$F1,$5F,$EE,$2F,$F3,$E7,$3B
	.dc.b	$FC,$48,$D7,$0F,$D0,$9F,$F1,$FC,$6D,$FE,$F3,$F3,$EF,$88,$F9,$6D
	.dc.b	$8E,$67,$F8,$63,$E7,$BF,$7C,$DE,$FF,$A3,$18,$8F,$F0,$1A,$CB,$EA
	.dc.b	$FC,$4F,$EC,$A7,$FC,$42,$D7,$FC,$12,$DC,$BC,$D0,$AE,$91,$87,$FA
	.dc.b	$BC,$78,$FC,$0F,$09,$E2,$F1,$7C,$7C,$5F,$84,$F9,$FF,$9B,$5F,$E1
	.dc.b	$66,$E1,$CE,$7C,$BC,$3E,$6E,$F9,$7E,$68,$FC,$6F,$EB,$BF,$D6,$A3
	.dc.b	$FC,$72,$FF,$BB,$DB,$E8,$6B,$D0,$DF,$3F,$0F,$9D,$95,$C2,$39,$FC
	.dc.b	$EF,$4F,$7E,$4F,$48,$EF,$E1,$C8,$EE,$F8,$BE,$77,$C7,$E1,$C9,$F1
	.dc.b	$BE,$6F,$13,$CF,$E0,$C9,$F3,$BE,$AE,$DF,$5F,$83,$27,$D6,$FB,$BC
	.dc.b	$2F,$7F,$CE,$49,$F7,$BF,$0B,$D4,$FD,$7F,$EB,$BF,$EB,$79,$FF,$2D
	.dc.b	$B5,$FF,$21,$B7,$A3,$ED,$F2,$BE,$6F,$8F,$7F,$2F,$8F,$98,$72,$8E
	.dc.b	$BF,$4B,$DB,$C2,$93,$DA,$3C,$3E,$4F,$8B,$E4,$24,$F8,$A3,$CB,$E1
	.dc.b	$48,$F2,$BE,$8F,$0B,$D3,$F2,$12,$7D,$2F,$B3,$C2,$F6,$FD,$2C,$9F
	.dc.b	$6B,$F0,$3E,$47,$E0,$FC,$8F,$C2,$F4,$E5,$7C,$28,$F8,$BF,$3B,$F1
	.dc.b	$BE,$4A,$4F,$C6,$8F,$93,$F3,$5F,$2B,$E7,$24,$FC,$A8,$F9,$BF,$29
	.dc.b	$FC,$EF,$A7,$FD,$73,$3F,$CF,$DA,$FF,$43,$6B,$FD,$6D,$FE,$C0,$FF
	.dc.b	$AC,$1A,$FF,$AF,$9A,$9B,$F2,$6F,$98,$FE,$94,$D9,$FE,$9C,$D7,$FA
	.dc.b	$93,$5F,$ED,$5F,$40,$EE,$FE,$A4,$D8,$FE,$AC,$D7,$FA,$D3,$5E,$3D
	.dc.b	$7C,$08,$F0,$7F,$AE,$7C,$FB,$5F,$C5,$4F,$F6,$38,$C4,$FF,$63,$5A
	.dc.b	$BC,$FC,$5F,$0B,$FF,$00,$7D,$37,$FA,$65,$BF,$25,$F1,$A3,$11,$FE
	.dc.b	$D8,$6B,$FE,$E6,$F5,$1F,$41,$F3,$1B,$EE,$7F,$73,$6C,$FF,$77,$6B
	.dc.b	$FE,$0F,$5F,$49,$C4,$93,$C5,$FA,$FE,$07,$F9,$8E,$BF,$C2,$CD,$7F
	.dc.b	$DD,$6D,$7F,$DF,$2D,$54,$F8,$3F,$31,$2F,$CC,$F6,$DE,$47,$F8,$2E
	.dc.b	$FF,$D6,$F7,$FE,$0F,$6B,$FE,$27,$FA,$1F,$EB,$F3,$FE,$11,$6C,$7F
	.dc.b	$85,$5A,$EF,$B7,$F2,$3F,$4C,$F1,$A9,$2F,$FC,$35,$19,$7A,$76,$FE
	.dc.b	$63,$03,$49,$1F,$3F,$E9,$7E,$91,$1C,$1F,$98,$FA,$2F,$D0,$F4,$9A
	.dc.b	$48,$F5,$76,$77,$23,$83,$FC,$5D,$F4,$5F,$82,$D4,$D2,$72,$FE,$BB
	.dc.b	$E8,$BF,$B5,$57,$FB,$F9,$B5,$FE,$E6,$5F,$41,$FD,$CF,$CD,$FE,$BF
	.dc.b	$1A,$6C,$73,$25,$7F,$8F,$E7,$8B,$7E,$37,$92,$B9,$3F,$E9,$07,$C1
	.dc.b	$3A,$73,$F1,$D5,$4F,$FA,$51,$FF,$27,$35,$FF,$33,$35,$EC,$70,$9F
	.dc.b	$80,$73,$8E,$58,$7F,$7A,$E1,$79,$19,$AD,$73,$CC,$BD,$17,$BF,$23
	.dc.b	$19,$23,$13,$F2,$EB,$D1,$8A,$F7,$5F,$2B,$E2,$FC,$5F,$29,$E3,$6E
	.dc.b	$1F,$4D,$B8,$FD,$36,$D7,$D2,$8F,$57,$81,$EB,$6D,$7D,$B6,$9F,$6C
	.dc.b	$B3,$C6,$3F,$01,$3C,$5A,$71,$2C,$F1,$6C,$F1,$8F,$C3,$4F,$16,$CF
	.dc.b	$16,$9C,$4B,$3C,$4F,$C5,$CF,$16,$CF,$16,$CF,$16,$9C,$63,$EE,$EB
	.dc.b	$FC,$A5,$E0,$62,$BA,$34,$CD,$7E,$42,$32,$D8,$CB,$F8,$E8,$CB,$63
	.dc.b	$33,$DE,$3C,$91,$D5,$A7,$32,$9C,$4B,$1C,$6B,$D5,$1B,$6B,$B6,$C6
	.dc.b	$E7,$DF,$3F,$69,$A7,$D4,$2B,$F3,$9B,$3F,$21,$A7,$9D,$EB,$5E,$EA
	.dc.b	$ED,$5F,$B4,$3B,$16,$BB,$34,$EC,$5A,$EC,$D8,$EC,$78,$1F,$73,$5F
	.dc.b	$73,$67,$C3,$1F,$BE,$3C,$25,$8F,$0B,$4F,$E0,$27,$F1,$B2,$3F,$C4
	.dc.b	$5A,$69,$C1,$1F,$AB,$97,$F7,$78,$FB,$B1,$F8,$08,$C3,$F6,$0C,$78
	.dc.b	$1F,$1D,$B3,$C0,$FC,$4C,$F0,$BF,$B3,$3A,$39,$4E,$AB,$93,$4E,$51
	.dc.b	$E4,$8C,$B6,$33,$3F,$86,$AE,$D1,$F2,$67,$B4,$F8,$23,$A4,$78,$0E
	.dc.b	$45,$8E,$55,$D9,$B3,$D8,$FA,$D5,$DA,$BE,$E1,$0F,$A5,$3F,$CA,$6B
	.dc.b	$95,$70,$6C,$EE,$B9,$3C,$27,$D0,$7F,$2C,$3F,$9B,$C6,$AF,$FA,$34
	.dc.b	$78,$E7,$FA,$44,$E5,$EA,$1C,$46,$E3,$15,$FB,$AF,$A1,$7C,$0F,$F4
	.dc.b	$B6,$D7,$86,$BA,$F9,$39,$A9,$39,$95,$70,$7F,$A0,$47,$A2,$B7,$18
	.dc.b	$3F,$AF,$4B,$DE,$F0,$1E,$1F,$7A,$DF,$59,$D3,$67,$4D,$35,$EE,$7A
	.dc.b	$7D,$CF,$53,$63,$27,$F6,$0C,$3F,$B0,$DA,$7F,$CD,$AD,$9D,$4F,$9A
	.dc.b	$72,$74,$8F,$81,$F4,$DF,$E6,$F7,$AC,$FB,$4F,$87,$EC,$FC,$BF,$6F
	.dc.b	$E6,$36,$7D,$8D,$9E,$18,$23,$1C,$3F,$EE,$A7,$FC,$59,$3D,$5B,$19
	.dc.b	$F6,$7F,$C0,$51,$8B,$FE,$FA,$DA,$FF,$06,$35,$ED,$3C,$3C,$4E,$3E
	.dc.b	$BF,$2F,$62,$E1,$F7,$70,$46,$30,$F8,$B0,$AF,$F4,$AB,$6B,$E4,$CF
	.dc.b	$AE,$7F,$D3,$3A,$7F,$88,$23,$D5,$1F,$E1,$98,$C6,$5E,$F8,$F5,$FA
	.dc.b	$DE,$CE,$DB,$87,$8D,$B8,$78,$A7,$C2,$49,$F0,$E2,$F5,$7C,$A4,$8C
	.dc.b	$B6,$BE,$63,$F2,$B9,$BE,$5B,$F3,$3C,$39,$5F,$0C,$B5,$F4,$DF,$82
	.dc.b	$43,$C2,$7C,$0F,$03,$EA,$BE,$F6,$BF,$A9,$FA,$27,$8D,$FE,$3B,$7E
	.dc.b	$1C,$7F,$8D,$9C,$BB,$F0,$F5,$3F,$EC,$49,$CD,$FE,$61,$AF,$E7,$DA
	.dc.b	$F3,$F9,$2E,$04,$BF,$AF,$28,$C4,$FE,$6E,$31,$3E,$7F,$43,$CD,$5E
	.dc.b	$7F,$12,$4F,$AF,$E8,$AE,$1C,$5F,$31,$23,$BB,$F8,$4B,$91,$F8,$8E
	.dc.b	$4F,$F5,$7D,$7F,$95,$9A,$FF,$99,$7E,$89,$E2,$7F,$98,$57,$8B,$63
	.dc.b	$0F,$F9,$97,$E8,$5F,$0D,$7E,$1F,$0D,$CB,$F8,$D4,$FF,$B8,$5B,$CD
	.dc.b	$C7,$DA,$E4,$FF,$9B,$FE,$85,$EF,$53,$1E,$95,$BF,$E4,$9F,$41,$F5
	.dc.b	$3B,$D1,$FE,$70,$8C,$37,$A3,$FC,$F5,$18,$E1,$FF,$41,$3E,$E9,$E7
	.dc.b	$C1,$AE,$0D,$BF,$E8,$1F,$A0,$7C,$2D,$F8,$7C,$36,$FF,$9B,$71,$FE
	.dc.b	$D7,$2F,$E2,$CF,$EF,$23,$3E,$17,$FA,$2E,$31,$3A,$EE,$C7,$99,$96
	.dc.b	$DF,$F4,$1F,$9F,$7D,$37,$FA,$4D,$B1,$9F,$9B,$C4,$E7,$CC,$93,$C7
	.dc.b	$CD,$B9,$7F,$29,$FA,$0F,$0F,$2F,$C8,$8F,$F5,$24,$62,$BF,$D3,$2B
	.dc.b	$DD,$FF,$51,$39,$9F,$F5,$57,$D0,$3D,$CF,$F5,$33,$6F,$FA,$C2,$F0
	.dc.b	$F3,$3F,$EA,$BF,$A0,$7D,$25,$31,$B5,$C3,$FB,$04,$FD,$06,$F2,$F1
	.dc.b	$E6,$E4,$FF,$AD,$7D,$B7,$C1,$DF,$1F,$0D,$BF,$ED,$53,$FD,$F3,$6F
	.dc.b	$E9,$4F,$F9,$00,$C7,$0F,$C3,$84,$7F,$91,$3E,$85,$C7,$FE,$C3,$5F
	.dc.b	$23,$63,$0F,$FB,$1F,$E7,$FF,$D7,$B8,$FF,$B0,$9F,$F6,$76,$3D,$5C
	.dc.b	$E3,$FE,$D3,$73,$3B,$F3,$39,$65,$CD,$B1,$FE,$03,$F8,$D7,$B3,$98
	.dc.b	$C3,$70,$FE,$CD,$83,$E8,$3C,$87,$97,$CF,$B7,$FE,$0D,$94,$62,$B8
	.dc.b	$B7,$3F,$47,$CD,$FB,$4F,$77,$C0,$DC,$39,$38,$7F,$C3,$1F,$2A,$F9
	.dc.b	$2C,$63,$FD,$E7,$07,$CA,$7F,$BB,$5B,$7F,$E1,$F5,$CF,$99,$D3,$27
	.dc.b	$AF,$C9,$DB,$FF,$02,$1F,$C5,$A3,$FC,$50,$D3,$3E,$4F,$FB,$ED,$C6
	.dc.b	$1F,$7D,$FE,$04,$6A,$78,$FC,$B6,$EB,$83,$23,$DB,$E7,$5A,$F0,$ED
	.dc.b	$D4,$ED,$B5,$AF,$1B,$87,$CB,$2F,$9F,$2F,$A0,$7C,$F7,$F8,$AE,$FD
	.dc.b	$13,$B3,$FC,$02,$53,$FC,$07,$58,$70,$74,$1F,$3F,$F9,$8F,$A0,$FF
	.dc.b	$04,$B6,$3F,$C6,$6D,$7F,$C7,$4D,$79,$3F,$A2,$DB,$8F,$61,$A4,$9E
	.dc.b	$7E,$4F,$86,$3F,$59,$5F,$80,$9F,$F2,$E3,$5F,$F1,$E3,$FE,$47,$73
	.dc.b	$C7,$C7,$5F,$57,$2F,$F8,$8E,$7F,$1B,$E7,$E5,$F9,$37,$CF,$0E,$FC
	.dc.b	$B9,$9F,$83,$E1,$5A,$FE,$77,$1E,$29,$C4,$E7,$AE,$4F,$9A,$2B,$FE
	.dc.b	$51,$C5,$F2,$3D,$5B,$1E,$CA,$F0,$71,$39,$AF,$2C,$AF,$F3,$E4,$E5
	.dc.b	$B1,$CF,$E1,$A4,$60,$FF,$41,$7D,$07,$C5,$7F,$CC,$4F,$BF,$83,$EF
	.dc.b	$36,$3F,$16,$EF,$7F,$E3,$8E,$61,$F3,$EF,$A0,$F8,$CD,$8E,$FC,$7F
	.dc.b	$27,$E8,$3F,$75,$1F,$E9,$36,$BF,$E7,$36,$E1,$FC,$A5,$65,$F4,$7E
	.dc.b	$5F,$FC,$17,$5F,$E9,$B6,$BF,$E7,$B6,$D7,$1E,$FF,$07,$87,$E1,$F0
	.dc.b	$25,$F8,$63,$C9,$C9,$F7,$7E,$81,$C3,$4C,$D7,$7E,$3F,$17,$74,$8C
	.dc.b	$1F,$EA,$DF,$9F,$7B,$1F,$87,$7F,$D6,$F1,$9F,$91,$C4,$E7,$A7,$5F
	.dc.b	$07,$2B,$94,$71,$F9,$1E,$5F,$23,$27,$B4,$75,$F3,$64,$78,$5F,$07
	.dc.b	$D0,$78,$7C,$39,$3E,$57,$C9,$E2,$79,$7C,$29,$3E,$97,$D1,$D7,$E9
	.dc.b	$F0,$64,$FB,$5F,$67,$85,$ED,$F8,$59,$3F,$05,$F8,$1F,$0B,$FA,$7F
	.dc.b	$A0,$7C,$BD,$25,$7C,$7F,$9B,$D7,$83,$BF,$94,$93,$FE,$D7,$BF,$C0
	.dc.b	$B7,$DF,$E9,$E0,$C8,$C1,$BF,$2F,$C1,$F2,$92,$30,$73,$F0,$FC,$84
	.dc.b	$7B,$78,$7E,$7F,$94,$91,$E1,$7D,$5E,$27,$AF,$E6,$24,$F9,$5F,$77
	.dc.b	$CA,$FB,$FE,$33,$E0,$7A,$92,$BD,$51,$F0,$FE,$57,$E2,$7C,$AC,$9F
	.dc.b	$74,$7C,$7F,$8A,$F9,$1F,$3B,$27,$E1,$47,$E2,$FE,$0F,$F9,$BF,$93
	.dc.b	$FE,$3A,$9F,$F5,$CB,$5F,$F7,$1B,$5F,$F7,$53,$FE,$26,$77,$F0,$9C
	.dc.b	$88,$FF,$B0,$B9,$BF,$AD,$3F,$CF,$DA,$FF,$4C,$6A,$6F,$CA,$BE,$43
	.dc.b	$FA,$5B,$67,$FD,$9A,$D7,$FD,$EE,$D7,$FD,$FD,$F3,$EE,$FF,$EA,$6D
	.dc.b	$8F,$F6,$AB,$5F,$F7,$FB,$5F,$EA,$83,$E9,$BF,$C2,$2F,$FB,$71,$AE
	.dc.b	$8D,$7C,$07,$F6,$08,$CC,$72,$FA,$07,$03,$C0,$FF,$72,$C6,$1B,$C2
	.dc.b	$F4,$73,$5B,$BE,$DF,$3B,$FE,$90,$F7,$3F,$DD,$AD,$8C,$F3,$7C,$C8
	.dc.b	$C4,$7A,$BD,$8F,$F5,$4A,$69,$3E,$27,$E9,$BF,$E5,$6B,$FF,$7C,$B6
	.dc.b	$7D,$CD,$7A,$7C,$4F,$0F,$86,$95,$C1,$B3,$C3,$83,$A9,$27,$AB,$E0
	.dc.b	$F8,$C9,$1E,$06,$CF,$CD,$F9,$FF,$EB,$73,$F3,$E1,$18,$75,$E0,$7E
	.dc.b	$32,$31,$1F,$6B,$47,$07,$C5,$25,$FC,$A7,$87,$74,$8E,$D6,$8C,$5F
	.dc.b	$8F,$7E,$C4,$AD,$BF,$2B,$E0,$A4,$7C,$A6,$CF,$AF,$BB,$CC,$49,$EE
	.dc.b	$FE,$93,$E7,$9F,$3F,$49,$3F,$06,$32,$DF,$83,$E0,$47,$07,$E1,$8F
	.dc.b	$F5,$57,$C0,$FF,$A2,$76,$FF,$69,$C7,$F8,$F2,$3F,$D1,$95,$FE,$7E
	.dc.b	$C7,$FC,$7F,$5A,$7F,$C5,$55,$FE,$37,$97,$CE,$F9,$C9,$1E,$56,$CF
	.dc.b	$03,$FC,$93,$3D,$53,$FE,$8E,$6C,$79,$A7,$B7,$1D,$9E,$2D,$78,$F1
	.dc.b	$DB,$FF,$45,$57,$AC,$E2,$59,$E2,$D7,$8B,$63,$8C,$79,$CF,$2F,$CB
	.dc.b	$F5,$DB,$B3,$6B,$83,$67,$85,$7F,$9E,$27,$E3,$57,$97,$A7,$D1,$CF
	.dc.b	$FE,$1B,$9F,$04,$63,$6D,$47,$F8,$45,$B8,$7F,$84,$5B,$B7,$FB,$06
	.dc.b	$F9,$4F,$82,$F9,$5F,$43,$91,$63,$93,$6B,$95,$F4,$8F,$F3,$95,$F4
	.dc.b	$9F,$C1,$DF,$4B,$ED,$5D,$1A,$74,$2C,$F4,$8E,$C7,$4B,$ED,$3F,$E8
	.dc.b	$AB,$ED,$3F,$8B,$BE,$D7,$E0,$8E,$CD,$AE,$CD,$3B,$3E,$08,$EC,$78
	.dc.b	$1E,$D7,$E0,$7F,$90,$36,$FC,$31,$F9,$4B,$F0,$D7,$8C,$F0,$96,$3C
	.dc.b	$2D,$BF,$09,$E3,$7C,$31,$E3,$3C,$2F,$8D,$F0,$B6,$BC,$31,$FA,$2A
	.dc.b	$F0,$BE,$39,$F0,$B4,$F0,$96,$BC,$31,$E3,$3C,$2F,$8E,$3C,$27,$8C
	.dc.b	$F0,$96,$BC,$38,$FF,$4F,$AE,$F3,$E3,$6C,$66,$3C,$26,$4A,$E6,$BC
	.dc.b	$71,$E5,$9F,$19,$E5,$7C,$71,$E5,$7C,$6F,$DD,$7F,$B7,$CF,$76,$DF
	.dc.b	$F2,$32,$57,$F2,$16,$C7,$F7,$19,$E0,$F7,$3C,$D5,$FC,$71,$FC,$14
	.dc.b	$79,$EB,$4F,$FB,$71,$FA,$CE,$63,$D4,$7F,$18,$38,$D7,$26,$CF,$06
	.dc.b	$9F,$3E,$7A,$36,$78,$4F,$66,$C6,$63,$91,$92,$B9,$3A,$1D,$AB,$E5
	.dc.b	$57,$66,$D7,$46,$D7,$26,$D7,$28,$EC,$D3,$B4,$74,$7B,$47,$C6,$69
	.dc.b	$F1,$0F,$E7,$26,$0F,$9E,$FF,$45,$71,$5F,$2D,$DC,$78,$27,$B3,$4E
	.dc.b	$C5,$8E,$D3,$ED,$8F,$B9,$F3,$CF,$0F,$FA,$13,$63,$F9,$E2,$F8,$1C
	.dc.b	$CC,$9C,$BC,$23,$17,$FC,$FE,$9E,$F7,$4F,$7F,$AB,$70,$EA,$DA,$E0
	.dc.b	$DC,$92,$99,$6F,$6F,$CE,$7A,$3D,$CB,$5E,$88,$B5,$E3,$9B,$7D,$23
	.dc.b	$9C,$62,$BB,$42,$D3,$98,$F0,$FB,$5F,$E6,$E9,$34,$95,$FE,$2B,$8C
	.dc.b	$BF,$2A,$3E,$6F,$B3,$F3,$9F,$EE,$0B,$D7,$BE,$CE,$1B,$5F,$DB,$67
	.dc.b	$4F,$F8,$DD,$B8,$7F,$99,$DA,$FF,$9B,$9A,$FB,$17,$AB,$E0,$3C,$3E
	.dc.b	$B5,$AE,$D1,$6B,$FB,$C5,$2B,$1A,$7C,$99,$4E,$6F,$C9,$5F,$E9,$28
	.dc.b	$C4,$F9,$63,$1E,$A4,$AF,$53,$67,$D5,$E9,$FC,$48,$F8,$AD,$9F,$85
	.dc.b	$E9,$FC,$8F,$07,$E4,$B6,$32,$7C,$57,$C4,$47,$E4,$7C,$BF,$94,$8E
	.dc.b	$09,$87,$C0,$8F,$AB,$1F,$59,$AF,$D1,$3D,$C4,$3F,$CD,$5F,$3D,$FE
	.dc.b	$AD,$6F,$A5,$FE,$B4,$8C,$57,$FA,$75,$AF,$FA,$91,$7A,$5F,$E3,$37
	.dc.b	$7B,$7C,$F7,$37,$FE,$2F,$72,$FE,$61,$7C,$87,$1F,$91,$E4,$D9,$E0
	.dc.b	$F8,$C8,$FA,$1F,$37,$85,$23,$D6,$DC,$7D,$0F,$94,$F3,$79,$96,$7F
	.dc.b	$C8,$5F,$3C,$F3,$F3,$18,$6D,$FF,$90,$FE,$79,$F2,$5F,$E5,$17,$FC
	.dc.b	$7D,$1F,$BF,$73,$5F,$E4,$D7,$30,$F9,$E7,$C4,$7F,$96,$57,$C4,$D8
	.dc.b	$C3,$9F,$6B,$5D,$57,$1F,$F2,$C5,$FE,$69,$FF,$2D,$7C,$F7,$FB,$07
	.dc.b	$8F,$F3,$CD,$8F,$E0,$AF,$FA,$65,$DF,$47,$F9,$A7,$E7,$BE,$AE,$9F
	.dc.b	$E5,$F6,$DF,$E9,$DF,$AF,$1F,$C6,$DC,$D7,$F1,$07,$DF,$1C,$76,$E9
	.dc.b	$E6,$F5,$7F,$CE,$DF,$3C,$F8,$7B,$F1,$38,$6E,$3F,$E7,$3E,$0F,$D8
	.dc.b	$E7,$FB,$77,$F8,$0C,$67,$C0,$FF,$40,$46,$33,$FF,$3B,$2E,$DC,$FA
	.dc.b	$BA,$6D,$FF,$9F,$FE,$79,$F0,$7F,$E8,$96,$C6,$7E,$7B,$13,$9F,$2A
	.dc.b	$4F,$1F,$2E,$E7,$FE,$85,$F9,$EF,$16,$DF,$C3,$5B,$5F,$CD,$17,$C5
	.dc.b	$7F,$A5,$DC,$D7,$F4,$5F,$9E,$7C,$F7,$FA,$6D,$B7,$FE,$95,$5E,$0E
	.dc.b	$67,$FD,$41,$F3,$DF,$D7,$EF,$FD,$2C,$DC,$BF,$8F,$3E,$C8,$CF,$A9
	.dc.b	$C3,$96,$65,$FE,$A1,$A7,$36,$2F,$FD,$41,$E3,$7F,$BF,$F1,$FE,$50
	.dc.b	$FE,$96,$FF,$D5,$0E,$5F,$EA,$2E,$67,$89,$FE,$B5,$7F,$C0,$07,$F7
	.dc.b	$97,$FB,$23,$FE,$03,$F9,$EF,$EC,$93,$FE,$B3,$5F,$25,$FE,$BD,$73
	.dc.b	$5F,$EB,$37,$33,$FE,$B9,$F9,$EF,$F6,$41,$FE,$CB,$2F,$BD,$FE,$C5
	.dc.b	$7F,$86,$C6,$F0,$C7,$95,$97,$FD,$95,$F1,$DF,$D8,$31,$FF,$60,$5F
	.dc.b	$07,$FD,$85,$19,$CF,$FA,$54,$E5,$BE,$FF,$73,$C1,$E6,$5C,$BC,$CD
	.dc.b	$C3,$AB,$6B,$9F,$CB,$79,$CF,$47,$BB,$6F,$DC,$E1,$FF,$6C,$7B,$6F
	.dc.b	$91,$C6,$3F,$ED,$7C,$1F,$27,$FE,$E5,$6D,$FF,$B7,$DF,$BE,$D8,$D7
	.dc.b	$1C,$8E,$BF,$25,$6F,$FD,$C4,$7F,$1A,$27,$8F,$FE,$F1,$6C,$FF,$BA
	.dc.b	$DF,$E0,$78,$7F,$6F,$5F,$2D,$BA,$E0,$CA,$F5,$FC,$3D,$AE,$FB,$75
	.dc.b	$3B,$6D,$6B,$B3,$87,$C1,$2F,$9D,$FF,$18,$7C,$F3,$E1,$FF,$C4,$B3
	.dc.b	$FE,$51,$3F,$C6,$93,$87,$2D,$7F,$C5,$6D,$76,$6F,$E7,$7F,$2B,$F3
	.dc.b	$DB,$6C,$67,$B1,$39,$6C,$71,$F1,$5C,$1C,$5B,$19,$8C,$1D,$7C,$6F
	.dc.b	$F1,$EB,$FE,$19,$C3,$EE,$CF,$F9,$36,$31,$C9,$AD,$7F,$C8,$75,$FE
	.dc.b	$42,$AD,$46,$27,$51,$FE,$55,$E6,$FF,$13,$3F,$8C,$F3,$72,$AF,$BB
	.dc.b	$E8,$BE,$BC,$DA,$9F,$E7,$93,$D8,$FF,$30,$38,$78,$18,$8E,$1E,$0F
	.dc.b	$1F,$83,$93,$C4,$F0,$46,$1E,$23,$17,$C7,$79,$9B,$1C,$70,$FF,$2E
	.dc.b	$BC,$F8,$BF,$22,$DE,$5F,$6C,$1B,$CF,$87,$A7,$CC,$4A,$E0,$79,$7E
	.dc.b	$7B,$E6,$3E,$47,$E2,$F0,$FF,$93,$5B,$BF,$F9,$D1,$DC,$67,$87,$E6
	.dc.b	$FC,$F3,$E3,$69,$F4,$2B,$8E,$1F,$E8,$9F,$9E,$FD,$F3,$F1,$5A,$FC
	.dc.b	$46,$E1,$FE,$5D,$59,$7D,$AF,$8E,$FF,$0B,$D7,$CD,$6B,$F3,$1B,$5C
	.dc.b	$7C,$0F,$C0,$DB,$C4,$BE,$0C,$C3,$C5,$1A,$E4,$FC,$3F,$CF,$7C,$B9
	.dc.b	$FB,$4D,$7C,$D1,$E3,$E2,$F2,$78,$12,$7C,$91,$F9,$0F,$C1,$FC,$7C
	.dc.b	$7F,$E9,$C7,$EC,$CF,$E0,$9B,$AF,$F9,$ED,$DC,$7F,$AB,$FC,$CF,$F3
	.dc.b	$F4,$FF,$9F,$DF,$F0,$24,$7E,$35,$AF,$E3,$1D,$D7,$FA,$AB,$F5,$E7
	.dc.b	$C9,$FE,$89,$7D,$1D,$31,$1F,$A6,$FB,$7F,$F5,$16,$1F,$99,$7E,$8B
	.dc.b	$79,$35,$F1,$FF,$AC,$FB,$87,$C6,$75,$E9,$97,$FA,$19,$CB,$FA,$3C
	.dc.b	$FF,$CD,$AD,$AF,$21,$E5,$F8,$CF,$DB,$FD,$BF,$F9,$32,$7F,$54,$FE
	.dc.b	$9D,$BE,$0E,$BE,$27,$D9,$EA,$7B,$7E,$A6,$57,$78,$F8,$1E,$AF,$C1
	.dc.b	$F5,$64,$F8,$E3,$F8,$2F,$92,$FA,$3B,$E9,$4A,$FD,$D4,$E5,$F5,$7C
	.dc.b	$E7,$CB,$F8,$E9,$18,$3D,$DE,$B7,$CF,$F5,$E4,$60,$FE,$09,$F6,$BF
	.dc.b	$E3,$86,$F6,$FF,$16,$83,$7E,$07,$EC,$7B,$BF,$67,$DE,$97,$F2,$E3
	.dc.b	$EE,$7C,$47,$DD,$F4,$64,$FC,$F8,$FE,$43,$F6,$8F,$2E,$F9,$32,$BF
	.dc.b	$8B,$4E,$5F,$E3,$1F,$93,$7F,$B5,$9F,$B1,$EF,$7E,$24,$FE,$57,$19
	.dc.b	$8F,$E3,$7D,$FF,$A5,$F2,$3F,$DA,$8F,$E7,$3C,$8F,$C9,$79,$5F,$91
	.dc.b	$F5,$3F,$25,$E6,$7F,$53,$A3,$49,$87,$F5,$98,$FE,$B0,$9A,$47,$FA
	.dc.b	$E7,$D3,$3E,$AE,$92,$3F,$B1,$46,$5F,$EC,$9F,$4F,$FD,$ED,$11,$C3
	.dc.b	$72,$FE,$D1,$18,$8F,$F6,$2F,$9D,$FD,$AF,$C7,$FD,$5C,$7F,$6F,$6B
	.dc.b	$FD,$CD,$AF,$F7,$8F,$A7,$7C,$A7,$F7,$36,$C7,$F7,$76,$BF,$DE,$DA
	.dc.b	$F1,$FA,$5E,$04,$78,$3F,$DF,$7E,$99,$F4,$C8,$FF,$01,$46,$1B,$F4
	.dc.b	$DF,$42,$BF,$C1,$13,$97,$FC,$15,$F5,$6F,$A1,$F8,$6D,$DB,$FC,$18
	.dc.b	$D9,$DC,$EB,$D3,$95,$A6,$CF,$F8,$57,$DC,$78,$FF,$39,$B5,$FE,$17
	.dc.b	$6C,$EE,$7E,$9F,$C7,$FF,$B1,$63,$FC,$3C,$D7,$FC,$46,$D7,$ED,$9F
	.dc.b	$E9,$6F,$9A,$ED,$E4,$4C,$3B,$36,$79,$ED,$D8,$93,$FE,$29,$BF,$94
	.dc.b	$FF,$8A,$FE,$9F,$FB,$BB,$7C,$0F,$C3,$C6,$2F,$FC,$64,$D7,$FC,$6F
	.dc.b	$F5,$6D,$7F,$1A,$DB,$FF,$1B,$36,$7F,$C7,$4D,$7F,$C7,$FD,$0F,$93
	.dc.b	$FC,$AB,$6B,$F3,$B1,$83,$E6,$78,$D2,$3E,$63,$6B,$F2,$9E,$E7,$E7
	.dc.b	$BA,$7F,$A2,$D7,$E6,$63,$07,$E8,$BE,$A3,$EF,$57,$F9,$41,$AF,$CE
	.dc.b	$6B,$FA,$A6,$AC,$9F,$13,$F0,$09,$7F,$01,$FD,$7F,$D4,$7F,$82,$63
	.dc.b	$FC,$50,$DA,$FF,$13,$BF,$EF,$84,$E7,$C4,$4F,$E0,$23,$FA,$9B,$FE
	.dc.b	$61,$6C,$FF,$99,$5A,$FF,$9A,$9A,$F1,$F6,$F9,$BE,$06,$D6,$23,$D9
	.dc.b	$C3,$8B,$E0,$7A,$3D,$6C,$4F,$F0,$CC,$11,$88,$F8,$9F,$49,$67,$35
	.dc.b	$C4,$93,$FE,$46,$CD,$5B,$BE,$51,$FD,$5E,$BE,$3F,$36,$3D,$6E,$8D
	.dc.b	$8C,$9F,$E5,$39,$E1,$87,$CC,$9F,$17,$95,$E3,$F3,$6D,$FD,$26,$D7
	.dc.b	$D2,$6E,$BF,$0F,$44,$62,$BE,$93,$6B,$E1,$35,$FB,$2D,$9E,$33,$F8
	.dc.b	$0B,$F1,$34,$F1,$16,$BC,$4D,$BF,$15,$F3,$24,$64,$F1,$96,$7C,$6D
	.dc.b	$7C,$6D,$BF,$1C,$7E,$42,$FC,$6D,$9F,$1B,$4F,$19,$6F,$C7,$1C,$EF
	.dc.b	$C7,$39,$8F,$1B,$4F,$19,$4F,$19,$6F,$C7,$5F,$3E,$FC,$6D,$AF,$1B
	.dc.b	$6B,$C6,$D3,$C6,$F3,$BF,$1C,$66,$7C,$6D,$9F,$1B,$5F,$1B,$6F,$C7
	.dc.b	$3F,$AF,$BF,$1B,$4F,$19,$6B,$C6,$DA,$F1,$9C,$EF,$C7,$19,$3C,$65
	.dc.b	$9F,$1B,$5F,$1B,$6F,$C7,$1F,$AF,$BF,$E1,$64,$9F,$E0,$4D,$8F,$E0
	.dc.b	$4D,$7F,$81,$39,$BF,$81,$8B,$E8,$FF,$2D,$3F,$99,$8C,$5F,$E1,$E3
	.dc.b	$13,$A8,$FC,$64,$EA,$FE,$D5,$6A,$78,$1A,$AE,$13,$F8,$1A,$E1,$7F
	.dc.b	$3E,$F8,$4F,$2B,$E0,$72,$78,$5F,$2A,$F4,$B6,$7E,$45,$F4,$7E,$45
	.dc.b	$F4,$AE,$D1,$D2,$FB,$1D,$1E,$D1,$D1,$A7,$4B,$ED,$3F,$46,$3F,$87
	.dc.b	$46,$23,$F9,$BD,$7F,$18,$8C,$60,$8C,$57,$4A,$E4,$7F,$21,$AD,$CE
	.dc.b	$A3,$CB,$7E,$36,$D7,$8C,$FD,$D3,$FC,$B4,$FE,$4F,$E1,$58,$FE,$9B
	.dc.b	$14,$FE,$5D,$16,$7F,$9F,$62,$F9,$AF,$E7,$1F,$4C,$FA,$8F,$E6,$31
	.dc.b	$86,$E3,$FE,$FF,$9D,$3F,$CB,$E7,$4E,$7E,$A3,$99,$D3,$EA,$EC,$64
	.dc.b	$99,$B2,$FE,$A1,$82,$B7,$39,$AF,$F0,$FC,$E5,$E3,$5C,$2B,$6F,$2F
	.dc.b	$92,$FF,$2E,$D1,$A4,$AF,$F1,$4D,$F4,$8F,$82,$7F,$59,$FA,$6F,$F6
	.dc.b	$4C,$FB,$E3,$E2,$F8,$7F,$1B,$E4,$EC,$FC,$36,$DF,$41,$87,$F8,$DD
	.dc.b	$FF,$38,$B5,$CD,$7F,$9C,$8F,$F1,$26,$3F,$DC,$5B,$7F,$E3,$E5,$F1
	.dc.b	$1F,$DB,$5C,$D7,$F9,$F1,$A9,$A7,$09,$FF,$40,$D7,$5F,$0B,$95,$F8
	.dc.b	$49,$3E,$9B,$E5,$1E,$5E,$C9,$39,$C3,$FC,$3E,$D8,$FF,$45,$E9,$FE
	.dc.b	$19,$9C,$DF,$D9,$3F,$D2,$31,$99,$F5,$7D,$47,$AF,$AE,$E1,$E6,$3E
	.dc.b	$03,$8B,$F9,$98,$2B,$E4,$12,$BC,$9C,$DF,$2C,$8E,$09,$5E,$33,$E3
	.dc.b	$47,$CB,$E5,$F9,$8D,$9F,$74,$7D,$52,$3F,$04,$FB,$04,$30,$7C,$9C
	.dc.b	$9F,$07,$F5,$E3,$F1,$8D,$7D,$D7,$FE,$B4,$FA,$87,$85,$FE,$29,$7F
	.dc.b	$C8,$46,$27,$72,$8C,$CF,$FA,$E9,$D4,$7E,$45,$AF,$E7,$9A,$FF,$A2
	.dc.b	$17,$C2,$71,$25,$7E,$06,$32,$D8,$F2,$1E,$5F,$82,$B1,$E4,$C7,$77
	.dc.b	$F3,$6D,$5F,$AB,$BE,$17,$EA,$F5,$FF,$AD,$DF,$FB,$93,$BB,$FC,$91
	.dc.b	$1F,$C9,$E7,$51,$FE,$4B,$FA,$7F,$E1,$F5,$F6,$5B,$CB,$FE,$57,$5F
	.dc.b	$03,$BE,$DE,$1D,$33,$1F,$93,$4F,$96,$C6,$3F,$83,$F9,$AF,$F5,$F6
	.dc.b	$5F,$E6,$16,$DF,$FA,$B9,$FD,$69,$89,$FF,$75,$FD,$2F,$F9,$D6,$7F
	.dc.b	$D6,$6D,$8C,$F9,$5F,$E6,$38,$C4,$17,$C1,$64,$D7,$99,$6F,$EF,$FD
	.dc.b	$9B,$C8,$FF,$36,$36,$33,$F0,$98,$9F,$F3,$87,$D2,$BD,$3A,$63,$E5
	.dc.b	$AD,$FE,$2A,$5E,$DF,$03,$8F,$75,$9F,$C7,$8E,$4F,$F3,$E3,$4C,$F8
	.dc.b	$38,$8F,$F4,$07,$D2,$3E,$76,$98,$FA,$CB,$7F,$9A,$FA,$56,$9F,$E8
	.dc.b	$76,$C6,$7C,$8C,$4F,$FA,$2F,$EC,$7F,$D8,$5C,$DF,$B9,$6F,$4F,$FA
	.dc.b	$5D,$7C,$66,$FD,$9F,$F4,$97,$D2,$EB,$2F,$F7,$3B,$70,$05,$F0,$FC
	.dc.b	$2B,$8F,$CF,$CB,$FD,$8D,$7F,$4B,$6C,$7F,$A7,$FD,$F7,$CE,$53,$1F
	.dc.b	$1F,$6B,$F7,$3E,$E3,$E3,$3F,$D4,$CD,$8C,$FC,$6E,$27,$FD,$55,$F4
	.dc.b	$AE,$5A,$63,$D9,$B7,$FC,$13,$E9,$7F,$D5,$BC,$3F,$C7,$5F,$E2,$31
	.dc.b	$9C,$BF,$D6,$91,$89,$D7,$9D,$89,$CB,$6B,$F8,$7E,$BF,$7E,$B7,$18
	.dc.b	$EB,$DB,$E1,$F9,$3E,$A7,$6F,$0A,$C7,$95,$B5,$E5,$6F,$47,$FB,$3F
	.dc.b	$3F,$85,$3E,$9D,$FC,$FE,$7F,$A1,$B8,$7B,$E3,$FD,$9D,$F3,$1F,$88
	.dc.b	$AF,$81,$1F,$ED,$1C,$1C,$B4,$C7,$CA,$DC,$3F,$95,$CA,$33,$1F,$CB
	.dc.b	$B0,$72,$E3,$1F,$E9,$07,$F1,$82,$7C,$0F,$FB,$7D,$B5,$FC,$D9,$FE
	.dc.b	$34,$70,$E6,$D4,$ED,$93,$DF,$E5,$AD,$76,$D7,$A1,$D7,$5B,$5D,$5C
	.dc.b	$3D,$E5,$E7,$7E,$73,$E9,$7F,$86,$C7,$E7,$85,$79,$DD,$9F,$DF,$CA
	.dc.b	$7F,$80,$67,$0E,$0E,$5F,$47,$FB,$DF,$A5,$FF,$03,$36,$3F,$C4,$0D
	.dc.b	$7F,$C5,$AD,$78,$FD,$2F,$F8,$FD,$DB,$5D,$B5,$DF,$0F,$6F,$83,$F5
	.dc.b	$2C,$FC,$7C,$78,$6D,$7F,$66,$C3,$FC,$80,$D9,$D3,$FC,$03,$89,$F0
	.dc.b	$38,$84,$61,$E0,$73,$E4,$D5,$FF,$49,$8E,$DE,$AE,$27,$2E,$BA,$64
	.dc.b	$E8,$F2,$3D,$ED,$F1,$34,$DC,$FF,$85,$7D,$47,$37,$95,$FE,$11,$1F
	.dc.b	$E0,$26,$BB,$8D,$72,$F5,$F2,$64,$EE,$3D,$DF,$49,$F1,$23,$FC,$44
	.dc.b	$FE,$99,$FF,$06,$B5,$CF,$17,$0E,$0F,$8D,$F4,$AF,$A1,$A6,$6B,$FC
	.dc.b	$2B,$F4,$BF,$CF,$1B,$E6,$FC,$D8,$C4,$7F,$8C,$5F,$F2,$DB,$A3,$FC
	.dc.b	$3B,$F2,$7F,$EA,$76,$F7,$71,$AE,$5E,$1F,$E5,$EF,$F9,$7D,$7F,$95
	.dc.b	$DA,$FF,$9C,$5A,$FF,$9F,$9F,$9E,$EF,$87,$91,$1F,$F3,$06,$1F,$41
	.dc.b	$BC,$1F,$A5,$8C,$47,$F9,$F5,$AF,$FA,$2B,$E9,$3D,$0F,$FA,$05,$AE
	.dc.b	$72,$FD,$74,$62,$3F,$69,$EF,$FE,$6D,$FF,$38,$35,$FF,$45,$35,$FF
	.dc.b	$4B,$B5,$FD,$E7,$CE,$3D,$EF,$DD,$36,$3F,$CF,$4D,$7F,$D2,$8D,$7F
	.dc.b	$D4,$3F,$48,$F8,$BF,$E0,$4D,$8F,$F4,$13,$5F,$F4,$E3,$5F,$F5,$4F
	.dc.b	$D2,$3E,$4F,$F8,$53,$63,$FD,$14,$D7,$FD,$48,$D7,$FD,$63,$7F,$E6
	.dc.b	$07,$CF,$F1,$3E,$A2,$3F,$E9,$3F,$A4,$7D,$27,$F1,$76,$C6,$63,$07
	.dc.b	$C2,$F8,$C9,$1F,$09,$AF,$C2,$F4,$3E,$39,$1F,$8E,$FF,$23,$F0,$9F
	.dc.b	$2B,$49,$3F,$EA,$58,$CB,$F3,$BE,$67,$E8,$91,$C1,$F4,$BD,$1F,$AA
	.dc.b	$47,$07,$FA,$BF,$E8,$FF,$C3,$2D,$F7,$3F,$9A,$46,$23,$FD,$98,$D7
	.dc.b	$E4,$FC,$47,$E0,$08,$FE,$01,$FC,$1F,$C6,$48,$FC,$1B,$5F,$F6,$B7
	.dc.b	$D1,$FF,$5E,$7F,$D8,$0D,$73,$EA,$FF,$45,$8C,$47,$FB,$A3,$E8,$DF
	.dc.b	$59,$FD,$25,$B1,$FE,$C9,$6B,$FE,$E3,$6B,$F8,$0F,$23,$F2,$E4,$7F
	.dc.b	$2E,$FF,$53,$EB,$7C,$C5,$24,$FE,$99,$FE,$B1,$F8,$67,$E2,$E3,$FA
	.dc.b	$F3,$5F,$F1,$EB,$53,$7A,$3F,$AE,$92,$3F,$52,$7F,$63,$F9,$DF,$F4
	.dc.b	$5B,$FE,$E3,$6B,$FE,$FD,$6B,$FD,$B4,$38,$7F,$B8,$7D,$2F,$ED,$DF
	.dc.b	$F2,$7B,$53,$4C,$75,$79,$C9,$D9,$FD,$D5,$F5,$7D,$1B,$E9,$3F,$BC
	.dc.b	$36,$FF,$DF,$0D,$7C,$81,$C3,$E7,$FA,$EF,$F4,$63,$FE,$FF,$6B,$C1
	.dc.b	$AE,$F9,$FF,$C0,$DF,$4A,$F8,$F7,$C3,$F6,$A4,$E6,$30,$FF,$A3,$BD
	.dc.b	$FB,$5C,$2B,$CB,$F4,$AF,$A5,$F8,$CD,$F7,$7A,$36,$77,$3F,$2F,$CC
	.dc.b	$FF,$66,$D7,$F3,$8F,$9A,$C1,$1F,$E7,$7E,$0E,$C8,$E0,$FE,$7F,$F0
	.dc.b	$DC,$C8,$E0,$FE,$87,$B7,$72,$38,$3F,$A3,$FC,$B7,$8C,$8E,$0F,$89
	.dc.b	$F4,$BF,$DF,$9B,$E9,$7E,$0E,$31,$A7,$D6,$6B,$F8,$B5,$F4,$1B,$25
	.dc.b	$74,$79,$7C,$44,$8E,$ED,$AE,$9F,$59,$DC,$93,$E4,$7C,$5D,$D2,$3C
	.dc.b	$ED,$9F,$27,$D5,$F9,$C9,$3E,$C7,$F3,$9F,$4D,$FD,$46,$3F,$04,$D9
	.dc.b	$FC,$8B,$5F,$D3,$35,$FF,$2A,$3F,$E0,$B8,$E3,$ED,$F4,$23,$83,$FD
	.dc.b	$6F,$F0,$FF,$D9,$67,$FD,$D6,$7F,$51,$8C,$9D,$27,$FB,$64,$EF,$D0
	.dc.b	$B3,$E0,$6D,$FF,$B0,$AF,$FD,$53,$5B,$AE,$73,$9F,$8D,$DA,$AF,$F5
	.dc.b	$55,$78,$E7,$31,$8A,$F6,$F0,$E3,$8B,$2D,$8E,$B8,$73,$8C,$2C,$3D
	.dc.b	$D1,$87,$E0,$79,$7D,$BE,$0F,$BB,$6F,$9E,$28,$C6,$1C,$E7,$E7,$C6
	.dc.b	$2F,$9C,$70,$BE,$6D,$39,$CE,$DE,$26,$6B,$A1,$F2,$63,$4D,$8D,$57
	.dc.b	$68,$D9,$F2,$6B,$BB,$8B,$46,$27,$EF,$36,$BF,$0E,$DA,$FB,$0D,$7A
	.dc.b	$47,$A0,$F0,$9E,$87,$B4,$6E,$BB,$CE,$E7,$2D,$AF,$0B,$4F,$09,$63
	.dc.b	$C2,$D7,$C2,$7A,$EF,$E1,$BF,$87,$6D,$7D,$59,$F2,$CF,$C7,$3C,$27
	.dc.b	$17,$2F,$16,$C7,$18,$DD,$7C,$D3,$CE,$F1,$33,$3F,$06,$32,$6D,$F5
	.dc.b	$C6,$E3,$31,$B3,$2E,$E3,$DE,$6E,$32,$6D,$CC,$6D,$F8,$6D,$9F,$9F
	.dc.b	$1C,$4D,$CF,$17,$6D,$8D,$CF,$1B,$FB,$04,$9F,$B1,$5F,$C3,$2F,$F8
	.dc.b	$99,$23,$F8,$9B,$63,$F8,$9B,$5E,$11,$F6,$6F,$FD,$EF,$7E,$7A,$FC
	.dc.b	$37,$B5,$F8,$76,$D6,$DB,$7E,$B3,$EB,$16,$F9,$CF,$AE,$77,$7D,$63
	.dc.b	$C1,$7D,$5A,$75,$2D,$F3,$AE,$31,$CC,$CD,$73,$9F,$05,$73,$6C,$F3
	.dc.b	$69,$CD,$E2,$73,$73,$3C,$CF,$B1,$3C,$DB,$3C,$DB,$3C,$CE,$31,$CE
	.dc.b	$32,$73,$7E,$E4,$F3,$69,$CC,$B3,$F2,$7B,$3F,$95,$9F,$CB,$1B,$1F
	.dc.b	$C7,$A2,$C7,$F0,$92,$D7,$FA,$A7,$E9,$1E,$4F,$F3,$86,$DF,$F3,$05
	.dc.b	$F4,$19,$9D,$36,$74,$DB,$FE,$33,$4F,$A5,$E9,$F0,$FD,$5B,$7D,$5F
	.dc.b	$F2,$04,$EE,$3F,$C4,$CD,$AF,$F1,$AB,$5F,$F2,$53,$9C,$7D,$55,$E9
	.dc.b	$F9,$1E,$D7,$E9,$F8,$DF,$61,$ED,$F8,$7B,$1E,$D6,$E3,$E8,$6E,$1F
	.dc.b	$17,$E2,$BE,$21,$F1,$7E,$6E,$DF,$C5,$6D,$FC,$06,$D7,$FB,$55,$FF
	.dc.b	$64,$4F,$F5,$A7,$FC,$6D,$19,$F8,$1F,$EC,$F1,$89,$54,$8E,$2F,$F9
	.dc.b	$59,$AF,$F9,$AD,$AF,$00,$E1,$E5,$3D,$8E,$FF,$17,$6B,$C9,$EB,$78
	.dc.b	$8F,$1F,$C7,$58,$F4,$3F,$E7,$5C,$0D,$24,$7F,$97,$63,$FC,$66,$DF
	.dc.b	$07,$81,$1D,$7D,$2F,$C0,$8F,$82,$D8,$F8,$DF,$4B,$F1,$3C,$2F,$8A
	.dc.b	$D8,$C9,$DD,$F2,$11,$C1,$AD,$3F,$B5,$E1,$FE,$73,$8C,$6B,$F5,$7C
	.dc.b	$1F,$A2,$7D,$28,$A7,$D7,$3E,$01,$0F,$71,$FE,$97,$FA,$4F,$F1,$65
	.dc.b	$FF,$A2,$9A,$FF,$A7,$DA,$FD,$56,$BF,$79,$7A,$7F,$E2,$87,$9D,$AB
	.dc.b	$84,$E7,$D4,$D7,$C1,$6F,$C3,$B5,$FE,$26,$C3,$FC,$6A,$F7,$9F,$31
	.dc.b	$1F,$F4,$94,$23,$D5,$E6,$49,$CB,$72,$FF,$52,$F7,$7F,$6B,$7F,$D7
	.dc.b	$4D,$7F,$20,$D7,$F3,$2D,$5F,$49,$FC,$2A,$BF,$D4,$4D,$8F,$F5,$8B
	.dc.b	$5F,$F6,$33,$5F,$CE,$2E,$7F,$F6,$33,$67,$9C,$60,$FF,$65,$7C,$A7
	.dc.b	$FA,$8E,$BF,$DA,$0D,$7F,$4C,$D7,$F5,$ED,$5F,$49,$FE,$6F,$7F,$D8
	.dc.b	$AD,$7F,$DB,$4D,$7F,$60,$D7,$F7,$2F,$F7,$A7,$7E,$2F,$FB,$76,$31
	.dc.b	$1C,$23,$FD,$C1,$F4,$8F,$1E,$93,$1F,$F6,$44,$65,$FF,$65,$7D,$1B
	.dc.b	$E9,$3F,$DD,$0D,$8C,$C6,$0C,$FD,$24,$8E,$2D,$C7,$FD,$C3,$1F,$2D
	.dc.b	$BA,$7F,$BC,$E3,$11,$FE,$F9,$6B,$FC,$3F,$E9,$1F,$49,$FE,$F9,$6C
	.dc.b	$FF,$B7,$DB,$3F,$EE,$E6,$BF,$C5,$C7,$D2,$3E,$5B,$F4,$11,$FE,$F4
	.dc.b	$8C,$57,$FB,$A5,$B1,$99,$DB,$CF,$E9,$1F,$47,$88,$FE,$49,$18,$BF
	.dc.b	$F7,$93,$63,$FE,$00,$B5,$F2,$7D,$1D,$7F,$96,$35,$FE,$6A,$D4,$D1
	.dc.b	$F3,$7E,$18,$DB,$5D,$B6,$37,$5F,$0E,$7F,$A0,$FC,$0B,$E6,$B8,$36
	.dc.b	$75,$28,$C3,$C5,$AF,$CD,$EC,$7C,$57,$85,$B5,$C9,$B3,$E1,$6B,$F5
	.dc.b	$7E,$8D,$F4,$7E,$96,$D7,$89,$B3,$E9,$6B,$F7,$7E,$8F,$FD,$7C,$FF
	.dc.b	$5C,$6A,$7F,$8B,$B9,$F6,$BF,$D9,$31,$FC,$56,$BF,$D8,$2E,$A7,$87
	.dc.b	$81,$AC,$76,$CA,$FF,$63,$E9,$F9,$BF,$13,$F0,$B8,$73,$C3,$FD,$5A
	.dc.b	$BD,$AE,$E7,$83,$B2,$E1,$E0,$6D,$70,$6D,$F4,$DB,$CE,$7A,$3C,$1B
	.dc.b	$7E,$87,$0F,$FB,$6B,$D5,$FC,$CD,$79,$27,$FD,$CE,$FE,$0B,$0F,$F6
	.dc.b	$0B,$4F,$CE,$8F,$94,$DF,$27,$06,$61,$D7,$E4,$6D,$FF,$B2,$8F,$F5
	.dc.b	$15,$FE,$8F,$DB,$FF,$7A,$C7,$F9,$C0,$FF,$04,$38,$F6,$E5,$FF,$BC
	.dc.b	$25,$F3,$98,$7A,$3C,$8F,$F7,$FE,$8F,$80,$C5,$71,$DB,$B9,$E0,$DA
	.dc.b	$C7,$57,$0F,$7C,$1E,$3F,$F8,$8B,$E8,$FF,$DB,$F1,$FD,$3A,$75,$3F
	.dc.b	$A2,$8F,$E7,$25,$70,$F2,$3F,$59,$F4,$5F,$C5,$BE,$8F,$F5,$2D,$8C
	.dc.b	$F8,$69,$CB,$63,$8F,$B6,$E1,$E2,$D8,$CC,$60,$EB,$ED,$FF,$8C,$DF
	.dc.b	$E2,$B8,$7F,$04,$9F,$F1,$EC,$63,$97,$59,$FF,$8E,$69,$78,$BF,$F2
	.dc.b	$1C,$71,$7C,$53,$FE,$3F,$E0,$FF,$19,$C7,$E3,$7D,$8E,$0A,$B8,$63
	.dc.b	$E2,$E6,$CC,$3E,$8E,$98,$8B,$5E,$6C,$7C,$9E,$47,$03,$31,$E2,$3C
	.dc.b	$AF,$83,$27,$C3,$53,$84,$F3,$3F,$CB,$BE,$77,$E3,$1C,$07,$F8,$BB
	.dc.b	$86,$BD,$CD,$73,$FB,$A3,$3A,$7F,$42,$B7,$B9,$E3,$35,$CF,$F1,$3E
	.dc.b	$8F,$E6,$E3,$E0,$7E,$2E,$9F,$E6,$96,$C7,$F9,$C9,$DB,$E1,$72,$BE
	.dc.b	$8D,$F0,$DF,$41,$B1,$D7,$93,$E9,$FD,$1F,$92,$3D,$EF,$F5,$F6,$F9
	.dc.b	$BA,$CF,$6F,$DA,$F8,$C7,$7D,$33,$5C,$BC,$5E,$A4,$7A,$F4,$78,$39
	.dc.b	$24,$78,$1F,$1F,$85,$D7,$96,$4F,$D2,$C7,$EA,$3F,$3F,$EA,$BE,$6F
	.dc.b	$17,$FA,$37,$E9,$3C,$73,$F6,$1A,$FD,$56,$F3,$7F,$74,$5E,$27,$E5
	.dc.b	$3E,$27,$F8,$DD,$FF,$A5,$DB,$7F,$EA,$06,$BF,$ED,$4F,$85,$FC,$DF
	.dc.b	$CC,$3E,$92,$92,$7F,$13,$5F,$35,$FC,$37,$D2,$2F,$F4,$6D,$8F,$C7
	.dc.b	$B5,$DF,$C9,$7E,$A3,$E9,$1F,$05,$49,$3D,$B9,$FC,$7F,$07,$23,$F3
	.dc.b	$3A,$7F,$9E,$9F,$1F,$BB,$FB,$2F,$A4,$7D,$25,$24,$FE,$82,$1F,$01
	.dc.b	$EE,$F5,$E4,$F4,$7E,$47,$C6,$7C,$2F,$5E,$4F,$81,$F9,$9D,$BF,$C0
	.dc.b	$3C,$A7,$D2,$52,$57,$EB,$A1,$F1,$9F,$33,$E5,$A4,$FA,$9F,$AD,$E1
	.dc.b	$7D,$0F,$62,$4F,$B9,$FB,$5E,$17,$F0,$EF,$A3,$7D,$25,$25,$7E,$F2
	.dc.b	$3F,$C6,$6F,$C6,$FA,$4F,$B6,$47,$EC,$F8,$DF,$7B,$C1,$91,$F3,$1F
	.dc.b	$C4,$7C,$8F,$F1,$CF,$A4,$65,$49,$5F,$C2,$A3,$FC,$50,$FF,$0C,$FA
	.dc.b	$47,$27,$F2,$46,$C6,$63,$F2,$9C,$1F,$8E,$F4,$64,$FD,$C7,$F3,$1C
	.dc.b	$7F,$99,$F7,$BF,$A8,$78,$9F,$D7,$EB,$FC,$66,$D7,$FC,$74,$D7,$FC
	.dc.b	$BD,$DA,$F7,$BF,$AC,$36,$3F,$C7,$6D,$7F,$C8,$4D,$7F,$CC,$DF,$49
	.dc.b	$F9,$47,$FC,$82,$D7,$FC,$90,$D7,$7E,$17,$F6,$5F,$A3,$7C,$95,$24
	.dc.b	$FF,$92,$A3,$2F,$F9,$33,$E9,$19,$7F,$6D,$6C,$66,$30,$7F,$71,$FA
	.dc.b	$37,$CA,$D2,$4F,$F9,$5A,$32,$FF,$96,$3C,$6F,$32,$23,$86,$E5,$FD
	.dc.b	$F6,$31,$0F,$A3,$FF,$1C,$3F,$E6,$26,$BF,$E6,$96,$BF,$E8,$D6,$BD
	.dc.b	$CE,$5D,$BF,$E0,$98,$FF,$58,$47,$38,$C1,$FE,$0D,$FA,$37,$D1,$FF
	.dc.b	$9B,$E7,$FC,$ED,$18,$F8,$DF,$F3,$9A,$F9,$3F,$B1,$3E,$D8,$D3,$FE
	.dc.b	$7D,$FA,$87,$CD,$FF,$A0,$DB,$EA,$7F,$9F,$1B,$3B,$9F,$F4,$47,$B0
	.dc.b	$F9,$9A,$4B,$FF,$41,$C6,$5E,$5F,$37,$D4,$8E,$0D,$FC,$77,$42,$38
	.dc.b	$3C,$5A,$F9,$48,$E0,$EF,$F4,$7E,$42,$38,$3F,$47,$8F,$FA,$7F,$97
	.dc.b	$D7,$9F,$FA,$FD,$AF,$F4,$18,$F8,$5B,$4A,$F8,$4D,$9F,$D5,$E5,$CC
	.dc.b	$93,$FB,$07,$B6,$32,$3F,$66,$D9,$FF,$56,$67,$FE,$F0,$7F,$20,$D4
	.dc.b	$D3,$18,$FD,$BA,$FD,$EE,$1E,$B2,$3F,$EB,$2F,$B1,$63,$FE,$BD,$6C
	.dc.b	$66,$30,$47,$FC,$3F,$1A,$C3,$FC,$39,$F4,$5F,$D7,$67,$35,$FE,$E2
	.dc.b	$F7,$7F,$DB,$BC,$09,$C1,$F9,$B7,$0F,$F8,$AA,$BF,$D5,$B1,$8B,$FF
	.dc.b	$14,$36,$B9,$D7,$29,$FD,$95,$70,$8C,$42,$3F,$D4,$D1,$FE,$05,$FA
	.dc.b	$AC,$C6,$27,$FC,$69,$A2,$73,$C1,$CA,$31,$7E,$58,$C4,$F2,$F8,$9D
	.dc.b	$7B,$D6,$BF,$C2,$4D,$C3,$AB,$72,$FF,$7F,$E3,$F1,$E7,$E4,$DF,$2B
	.dc.b	$F8,$4D,$8C,$C7,$43,$25,$72,$76,$72,$D3,$35,$E0,$BF,$14,$FF,$0B
	.dc.b	$BF,$14,$FD,$5A,$F9,$77,$E1,$8C,$47,$CB,$8C,$57,$F1,$3C,$BF,$CC
	.dc.b	$0F,$8E,$B7,$18,$AF,$14,$62,$B8,$C7,$6A,$E0,$73,$8C,$C7,$78,$CB
	.dc.b	$B3,$9B,$B3,$2E,$CC,$9B,$9F,$81,$1F,$11,$B1,$F1,$23,$E1,$61,$F1
	.dc.b	$23,$F5,$AD,$7F,$5A,$DC,$3E,$3C,$F8,$67,$B9,$E1,$AE,$A7,$BE,$3C
	.dc.b	$B8,$7C,$38,$CC,$71,$6C,$F9,$E3,$C6,$79,$CA,$F9,$CF,$2B,$96,$99
	.dc.b	$3C,$EE,$5A,$66,$BD,$2F,$73,$8C,$F9,$DF,$5D,$79,$EB,$C3,$39,$3C
	.dc.b	$F3,$EF,$8C,$9E,$B2,$BD,$4F,$0C,$75,$7F,$AD,$1F,$5C,$B3,$FC,$6C
	.dc.b	$98,$7E,$42,$3E,$71,$A7,$0F,$E0,$CC,$57,$BE,$7F,$CE,$ED,$9F,$B5
	.dc.b	$18,$9F,$8C,$FB,$5F,$33,$9A,$F7,$57,$AE,$73,$E5,$F2,$BE,$6D,$9F
	.dc.b	$03,$67,$AB,$6B,$9C,$7F,$16,$8C,$4F,$58,$FE,$2D,$18,$7B,$C7,$A4
	.dc.b	$EA,$57,$AB,$67,$AC,$72,$8C,$1D,$6F,$D8,$DB,$F4,$BA,$9E,$E7,$C4
	.dc.b	$8C,$B6,$3C,$2D,$3C,$25,$AF,$0D,$76,$8F,$97,$5C,$A7,$2D,$AF,$11
	.dc.b	$D2,$32,$D3,$E2,$61,$FE,$D4,$7F,$D7,$9F,$0F,$FC,$E3,$C3,$FE,$68
	.dc.b	$C8,$C9,$FD,$26,$31,$87,$FB,$BF,$E8,$BF,$C8,$73,$FE,$C5,$ED,$FE
	.dc.b	$8B,$18,$9F,$E8,$31,$8E,$4F,$E8,$0D,$AE,$07,$FB,$FA,$9F,$15,$FD
	.dc.b	$17,$BB,$9B,$67,$4F,$E5,$B0,$57,$09,$FD,$0C,$7F,$92,$63,$0E,$1A
	.dc.b	$FE,$99,$7D,$17,$69,$F1,$36,$7D,$DF,$21,$DE,$8D,$25,$78,$27,$C7
	.dc.b	$1F,$08,$F8,$7F,$21,$67,$DD,$F1,$DF,$D8,$7E,$AF,$E3,$B6,$32,$73
	.dc.b	$C1,$1E,$12,$47,$7F,$15,$F0,$9F,$DD,$1B,$1F,$E6,$9E,$3F,$E9,$B1
	.dc.b	$FE,$DF,$8C,$1F,$E5,$F7,$79,$A7,$7F,$0D,$CE,$FA,$36,$F8,$D7,$F9
	.dc.b	$D6,$FA,$3F,$41,$AF,$F1,$26,$BD,$E7,$EC,$47,$FA,$C2,$7D,$5E,$8E
	.dc.b	$08,$E0,$EF,$5F,$E8,$A8,$F7,$7B,$BE,$F7,$FC,$15,$EF,$CA,$DD,$BD
	.dc.b	$7C,$61,$ED,$9F,$81,$1E,$22,$3F,$2D,$F2,$F8,$92,$3E,$73,$67,$DD
	.dc.b	$E0,$FC,$F8,$FA,$0D,$9F,$0B,$EE,$23,$F5,$DF,$83,$E7,$48,$FB,$4D
	.dc.b	$AF,$94,$FD,$B3,$EE,$45,$3E,$1F,$5B,$E5,$FF,$C5,$47,$F8,$D3,$E8
	.dc.b	$9D,$58,$AF,$93,$18,$7F,$2C,$DC,$3F,$4E,$D7,$FA,$4B,$5F,$F5,$23
	.dc.b	$DE,$38,$D7,$EC,$63,$13,$CA,$31,$F0,$3E,$37,$C8,$DA,$F1,$B8,$7C
	.dc.b	$BE,$55,$AF,$AC,$7F,$B0,$7B,$BF,$90,$3F,$EE,$5F,$03,$FC,$A3,$18
	.dc.b	$3F,$A7,$43,$E8,$9F,$49,$FE,$56,$5F,$01,$63,$0E,$7A,$B5,$F0,$77
	.dc.b	$3F,$C7,$DA,$3F,$CB,$FF,$31,$FA,$6E,$CF,$F1,$BF,$DA,$BD,$2C,$F7
	.dc.b	$7F,$99,$23,$18,$FE,$63,$E9,$3F,$97,$6D,$EC,$0F,$F9,$81,$C3,$AC
	.dc.b	$3F,$C8,$6E,$5B,$C1,$FE,$70,$87,$A5,$63,$0F,$F9,$D3,$E9,$1E,$AD
	.dc.b	$F4,$30,$DC,$BF,$57,$F1,$0F,$8B,$FF,$3D,$AF,$93,$B1,$87,$3E,$B6
	.dc.b	$BC,$0B,$7C,$9F,$F4,$1B,$FC,$58,$FD,$B8,$F9,$8C,$F3,$E9,$FF,$44
	.dc.b	$FD,$23,$E6,$EF,$D2,$E1,$B7,$FC,$03,$E9,$5F,$25,$FE,$90,$7F,$87
	.dc.b	$F2,$FF,$9A,$1C,$CE,$FC,$DF,$F4,$AF,$D2,$3E,$96,$FD,$56,$1B,$5F
	.dc.b	$C3,$3E,$9B,$4F,$F0,$EF,$AD,$76,$7F,$A7,$A7,$2F,$D9,$8C,$CF,$1F
	.dc.b	$9F,$D7,$AF,$B6,$C7,$F1,$8F,$05,$F2,$BF,$EA,$85,$EA,$D8,$C3,$FE
	.dc.b	$AC,$F3,$1F,$1D,$7E,$7B,$0D,$C3,$F9,$1F,$D3,$3E,$9F,$FD,$66,$FF
	.dc.b	$5A,$8C,$C6,$3D,$DF,$F5,$BF,$D3,$BD,$AA,$63,$2B,$7F,$CB,$A7,$F7
	.dc.b	$7D,$1F,$E9,$46,$C7,$C4,$E2,$43,$D3,$E5,$3C,$5E,$E7,$FB,$1B,$7F
	.dc.b	$C2,$57,$E8,$6B,$85,$77,$F1,$BC,$A7,$9B,$C6,$B1,$E3,$6D,$71,$6E
	.dc.b	$3E,$5D,$FD,$A7,$BB,$D2,$B7,$EB,$70,$FF,$4A,$F7,$FF,$0F,$87,$8E
	.dc.b	$7F,$DB,$76,$E5,$A6,$3D,$0B,$87,$F5,$29,$FE,$08,$D8,$FF,$71,$E0
	.dc.b	$F9,$6E,$13,$FD,$63,$1F,$E3,$24,$E5,$FF,$74,$B6,$35,$18,$FA,$1C
	.dc.b	$B6,$39,$47,$6F,$93,$C5,$7F,$0B,$3F,$0F,$3D,$F5,$EB,$AF,$66,$C7
	.dc.b	$8E,$3C,$32,$F4,$FF,$3F,$F4,$EF,$47,$FD,$13,$5A,$9F,$F0,$E5,$7E
	.dc.b	$91,$AF,$13,$2F,$F8,$87,$E9,$BF,$79,$F4,$F7,$C8,$34,$95,$FE,$36
	.dc.b	$9F,$F1,$5F,$17,$1F,$03,$FC,$25,$39,$6B,$FE,$9C,$9D,$BC,$7E,$17
	.dc.b	$F2,$95,$FE,$EA,$9F,$B5,$3F,$E0,$86,$BF,$CC,$54,$F0,$D7,$F3,$91
	.dc.b	$AA,$FF,$79,$F1,$FE,$DA,$FF,$11,$E0,$6D,$FE,$0B,$5B,$BE,$3C,$B9
	.dc.b	$94,$FF,$21,$7C,$5E,$66,$63,$FA,$64,$60,$F3,$79,$B9,$35,$36,$3F
	.dc.b	$31,$F3,$7D,$72,$39,$8F,$E9,$9C,$4E,$6F,$13,$FC,$2A,$3F,$C5,$2D
	.dc.b	$77,$1A,$E5,$E9,$E8,$C9,$DC,$7A,$FE,$9F,$E3,$C7,$F8,$C9,$F8,$0F
	.dc.b	$F2,$66,$A7,$6F,$F5,$F7,$3F,$4E,$F8,$BF,$8C,$D9,$E9,$DB,$F2,$7E
	.dc.b	$9F,$FA,$94,$7F,$90,$1F,$E8,$2D,$F4,$75,$96,$DF,$A1,$F0,$6E,$FA
	.dc.b	$66,$B9,$78,$9D,$48,$F5,$E7,$F0,$70,$C8,$F0,$3E,$3F,$07,$AF,$0C
	.dc.b	$9F,$F1,$8B,$F7,$FE,$9F,$E6,$75,$3A,$BF,$93,$7E,$C1,$FC,$BE,$FF
	.dc.b	$CB,$2D,$7F,$C0,$4D,$9E,$11,$FD,$42,$1F,$19,$F8,$B7,$F9,$74,$72
	.dc.b	$F2,$30,$67,$C9,$D7,$D1,$C9,$EE,$F1,$F8,$5E,$5E,$4F,$3F,$2A,$57
	.dc.b	$92,$3B,$7D,$17,$7F,$80,$93,$E8,$8F,$17,$C9,$78,$FC,$09,$3E,$C8
	.dc.b	$F3,$71,$79,$F9,$E4,$FC,$08,$F5,$79,$7E,$BF,$06,$4F,$C4,$8F,$D4
	.dc.b	$7E,$C0,$F6,$3E,$04,$C3,$FD,$03,$19,$6F,$C3,$F0,$23,$AF,$67,$9F
	.dc.b	$C4,$48,$E2,$F6,$F9,$2E,$FF,$4D,$27,$AB,$E2,$F8,$5F,$1F,$B1,$E4
	.dc.b	$F2,$E5,$78,$A3,$CF,$EC,$FA,$3E,$06,$4F,$9A,$3D,$7E,$9F,$B3,$D5
	.dc.b	$93,$EA,$8F,$7F,$99,$23,$DA,$FC,$2F,$53,$E1,$FA,$32,$7E,$0B,$F1
	.dc.b	$BD,$4F,$8F,$F2,$12,$7E,$2B,$F2,$BD,$4F,$97,$F5,$BF,$CE,$39,$8D
	.dc.b	$25,$7F,$B5,$FF,$08,$FF,$75,$26,$93,$CA,$E5,$EF,$C9,$FF,$6F,$5F
	.dc.b	$F9,$85,$B1,$FD,$1B,$E8,$7F,$D2,$31,$FE,$E3,$6B,$FE,$F4,$6B,$FE
	.dc.b	$0B,$0E,$1F,$F0,$77,$D1,$7F,$4E,$23,$FD,$49,$FE,$A3,$F4,$2E,$9F
	.dc.b	$EA,$AD,$9F,$F7,$93,$5E,$C1,$C3,$E1,$FA,$2F,$82,$4C,$BF,$AF,$BF
	.dc.b	$13,$E6,$7C,$93,$FE,$44,$6A,$69,$8F,$A2,$F9,$F1,$E0,$0E,$1F,$1B
	.dc.b	$5F,$84,$D7,$FC,$5C,$FC,$D7,$79,$FF,$72,$8C,$47,$08,$FE,$EB,$F4
	.dc.b	$2F,$A1,$CC,$FF,$96,$A3,$13,$E3,$5F,$1E,$F1,$31,$7C,$2D,$F5,$CF
	.dc.b	$95,$FE,$2B,$1F,$5E,$31,$AF,$B5,$B1,$F3,$9A,$FD,$DF,$71,$E8,$F2
	.dc.b	$8F,$F2,$54,$62,$BE,$33,$63,$EB,$B5,$E5,$E9,$71,$7F,$CE,$B0,$8C
	.dc.b	$42,$31,$3D,$27,$30,$F4,$A9,$27,$C2,$EB,$D2,$91,$E4,$6D,$FE,$2B
	.dc.b	$93,$DA,$FF,$98,$DA,$FF,$A2,$1A,$FE,$57,$3F,$2F,$D7,$7F,$78,$BF
	.dc.b	$C2,$36,$7F,$2A,$D7,$F4,$8D,$7F,$CE,$0B,$E7,$39,$92,$79,$BD,$BE
	.dc.b	$72,$47,$66,$D7,$CC,$F5,$3E,$19,$27,$E8,$BE,$6F,$AE,$91,$E6,$6C
	.dc.b	$F9,$BE,$77,$D6,$49,$F5,$B6,$3F,$DB,$D8,$A3,$FA,$EF,$B9,$F8,$AF
	.dc.b	$A3,$FF,$75,$5F,$B2,$72,$FF,$A3,$7E,$87,$FD,$C4,$DF,$88,$CE,$3F
	.dc.b	$EB,$49,$7C,$FE,$21,$1C,$1D,$4E,$63,$11,$A9,$C4,$E7,$DD,$D7,$C8
	.dc.b	$DC,$3F,$D8,$39,$FE,$83,$1F,$F5,$44,$60,$F3,$57,$07,$18,$F9,$63
	.dc.b	$0F,$FB,$FE,$7F,$DE,$73,$A8,$F3,$F2,$79,$79,$3C,$CD,$9D,$57,$E5
	.dc.b	$DB,$3A,$69,$A9,$E1,$5D,$DB,$87,$A5,$B3,$F3,$5B,$3F,$2D,$AF,$92
	.dc.b	$3A,$BC,$DB,$3C,$DA,$73,$2C,$73,$6B,$CC,$F8,$55,$E9,$6C,$7A,$4D
	.dc.b	$DF,$BE,$32,$78,$4B,$5E,$16,$CF,$85,$A7,$84,$A7,$86,$3E,$D5,$78
	.dc.b	$5B,$3E,$19,$DD,$78,$67,$31,$E1,$69,$E1,$2C,$F8,$5B,$1E,$16,$BE
	.dc.b	$18,$FC,$E5,$78,$5A,$78,$67,$75,$E1,$8C,$D7,$85,$B3,$E1,$69,$E1
	.dc.b	$2C,$78,$5A,$F8,$4F,$D4,$57,$85,$B1,$E1,$37,$5E,$18,$C9,$E1,$2D
	.dc.b	$78,$5B,$3E,$16,$9E,$12,$9E,$13,$F6,$B5,$E1,$6C,$7F,$65,$AF,$D2
	.dc.b	$C7,$F7,$66,$BF,$E0,$56,$BF,$10,$C4,$FF,$40,$8C,$C7,$A0,$CC,$FE
	.dc.b	$4B,$EC,$1F,$4F,$F9,$EA,$EF,$87,$D6,$9C,$D7,$E4,$B0,$FD,$1C,$63
	.dc.b	$0F,$C2,$C6,$30,$E4,$DA,$E0,$77,$8C,$B4,$CC,$FE,$52,$BA,$5F,$E0
	.dc.b	$2F,$94,$F6,$AE,$0D,$38,$5F,$46,$C7,$4A,$F5,$4E,$5B,$19,$3F,$07
	.dc.b	$7E,$07,$EF,$DF,$82,$7C,$51,$E0,$6D,$F6,$3C,$4F,$46,$C7,$4B,$F0
	.dc.b	$34,$F0,$47,$E0,$AF,$C0,$FE,$13,$0F,$A3,$3B,$8E,$66,$EB,$EB,$7B
	.dc.b	$DF,$CE,$4F,$E7,$5E,$F5,$AF,$E4,$B1,$6B,$F9,$74,$DC,$7F,$9C,$47
	.dc.b	$F4,$1E,$EF,$E8,$4D,$BF,$E5,$AD,$3F,$9B,$16,$7F,$C3,$FF,$46,$F6
	.dc.b	$BF,$A7,$3F,$E4,$FA,$FE,$74,$E5,$BD,$7A,$9F,$F0,$9C,$E6,$FF,$CA
	.dc.b	$2D,$7F,$CC,$0D,$44,$D5,$CF,$CF,$0F,$E9,$11,$87,$C3,$7E,$4F,$5F
	.dc.b	$B9,$23,$FC,$A1,$82,$31,$83,$2F,$6C,$7B,$3D,$7E,$57,$EC,$F4,$FD
	.dc.b	$FE,$AF,$C0,$6C,$64,$EA,$FB,$88,$F8,$3D,$1F,$8E,$7C,$8F,$4A,$C7
	.dc.b	$C8,$6E,$5E,$38,$F8,$2B,$34,$EB,$3F,$A6,$FC,$C2,$57,$B9,$FF,$20
	.dc.b	$8F,$99,$A4,$8F,$F3,$94,$65,$B1,$FE,$09,$8C,$5F,$8C,$FA,$6B,$E3
	.dc.b	$35,$3B,$FA,$39,$7F,$DE,$E9,$F1,$78,$C3,$91,$EA,$7C,$A4,$9F,$1B
	.dc.b	$E8,$F1,$24,$79,$9B,$3F,$E9,$8A,$7D,$47,$9E,$7D,$0D,$AF,$13,$F0
	.dc.b	$88,$FC,$07,$E2,$FC,$F4,$8F,$86,$DB,$EC,$FC,$33,$E2,$79,$96,$7D
	.dc.b	$DE,$1F,$C7,$3E,$43,$4F,$84,$7D,$22,$1F,$40,$FF,$3A,$FD,$27,$F8
	.dc.b	$E1,$FF,$52,$36,$FF,$D6,$AD,$7F,$D8,$8D,$7F,$0A,$FF,$AB,$E3,$FC
	.dc.b	$54,$F1,$DF,$FD,$2A,$E6,$7E,$8B,$99,$E3,$E9,$7F,$91,$E3,$55,$FE
	.dc.b	$9C,$FB,$07,$87,$9A,$FF,$26,$7D,$9B,$E2,$69,$8F,$2E,$E5,$FE,$3B
	.dc.b	$FA,$67,$C3,$7F,$94,$DB,$19,$F8,$8C,$4E,$7E,$0A,$4F,$1E,$CB,$C1
	.dc.b	$F8,$8F,$B3,$FE,$4E,$9F,$C0,$CF,$F9,$25,$C9,$88,$E1,$1F,$E6,$38
	.dc.b	$FF,$38,$3B,$31,$C5,$FE,$67,$FA,$6F,$E6,$35,$FC,$41,$FF,$00,$27
	.dc.b	$2D,$F3,$3F,$CD,$DF,$66,$CA,$98,$CA,$E3,$FE,$5D,$F8,$9F,$D5,$F2
	.dc.b	$FF,$B2,$1B,$1F,$BE,$7F,$D0,$38,$FF,$20,$72,$DF,$53,$FC,$F9,$06
	.dc.b	$FC,$F6,$1F,$F4,$07,$D2,$B8,$7F,$D0,$5E,$97,$F2,$AC,$7F,$CD,$D3
	.dc.b	$97,$FC,$E1,$19,$C7,$F6,$B3,$FC,$96,$77,$F1,$7D,$3C,$9E,$AF,$FA
	.dc.b	$37,$E9,$7C,$F6,$B7,$C0,$67,$DA,$F1,$BF,$E9,$3F,$A3,$F9,$27,$CC
	.dc.b	$2F,$26,$7D,$2D,$7C,$85,$D7,$FD,$09,$48,$FF,$4E,$7D,$23,$E3,$AF
	.dc.b	$C7,$E1,$B7,$FE,$8B,$F0,$DF,$29,$FE,$A3,$6C,$65,$C4,$67,$D9,$93
	.dc.b	$C7,$C3,$BC,$3F,$C6,$3E,$91,$C9,$FE,$AD,$6D,$7F,$A6,$57,$C9,$66
	.dc.b	$7F,$D6,$5F,$49,$FE,$A6,$BF,$E4,$2D,$ED,$61,$FC,$FA,$3F,$5E,$FF
	.dc.b	$A8,$DD,$39,$D3,$99,$D3,$4B,$1B,$70,$FF,$AA,$FE,$71,$EB,$63,$1F
	.dc.b	$0F,$2F,$90,$F2,$FC,$05,$BF,$03,$67,$CE,$DC,$FC,$F1,$FD,$D3,$3F
	.dc.b	$A9,$5E,$7E,$4F,$7F,$37,$C0,$6C,$7A,$E3,$FD,$A9,$B3,$E7,$73,$87
	.dc.b	$F4,$4C,$7F,$55,$7F,$EC,$36,$C6,$7E,$53,$FD,$B9,$18,$3C,$71,$FD
	.dc.b	$E6,$32,$76,$F9,$CB,$5F,$EC,$D9,$72,$FF,$B9,$5B,$19,$F8,$1C,$4E
	.dc.b	$59,$1C,$7E,$62,$D7,$F5,$A0,$F3,$3D,$BE,$25,$9F,$06,$DD,$CF,$06
	.dc.b	$D6,$3C,$8E,$1F,$35,$BE,$8F,$FC,$43,$F4,$9F,$CB,$E3,$FC,$07,$5E
	.dc.b	$97,$47,$F8,$18,$93,$5A,$53,$FC,$56,$DF,$A3,$FD,$57,$D2,$5E,$D3
	.dc.b	$49,$5C,$E5,$F0,$3F,$E2,$78,$C3,$5F,$14,$6B,$8B,$B4,$E6,$7F,$C3
	.dc.b	$72,$F2,$FF,$C3,$B5,$F6,$A3,$F9,$54,$FF,$90,$23,$1E,$66,$B5,$E3
	.dc.b	$76,$BF,$C8,$B7,$C1,$FF,$06,$F3,$7F,$84,$1E,$55,$FE,$2A,$8C,$47
	.dc.b	$F8,$4A,$78,$61,$E1,$E5,$E9,$29,$7C,$DE,$1B,$1D,$0F,$0C,$E6,$75
	.dc.b	$D7,$27,$47,$38,$C3,$E6,$FA,$1E,$F8,$9C,$E3,$F9,$CF,$0B,$F1,$B5
	.dc.b	$C9,$FE,$44,$DE,$4F,$3C,$6A,$3F,$C5,$AE,$9F,$87,$F0,$5D,$8D,$CF
	.dc.b	$BB,$E9,$3E,$94,$F9,$9F,$87,$B7,$F9,$41,$B1,$FE,$72,$76,$FF,$8E
	.dc.b	$9C,$AF,$A4,$7C,$6F,$CA,$6C,$75,$E2,$F9,$BF,$49,$FE,$49,$8F,$8A
	.dc.b	$FF,$03,$6F,$7E,$B3,$DB,$F5,$3E,$49,$DD,$4C,$D7,$2F,$13,$A9,$1E
	.dc.b	$BF,$0B,$E0,$E2,$91,$E0,$7C,$7E,$17,$5F,$06,$4F,$D8,$C7,$E6,$72
	.dc.b	$7F,$0E,$FA,$4F,$98,$FF,$10,$FA,$37,$A9,$F8,$86,$DF,$E0,$1A,$FF
	.dc.b	$98,$17,$D2,$6F,$CA,$93,$B7,$97,$97,$CF,$E5,$A4,$F3,$7B,$7C,$BF
	.dc.b	$7F,$33,$C1,$E6,$CA,$F0,$47,$8F,$E0,$7C,$9F,$49,$27,$C9,$1E,$7F
	.dc.b	$3B,$D1,$E0,$C9,$F4,$47,$AF,$8F,$D9,$D1,$27,$D9,$1E,$FF,$33,$E0
	.dc.b	$78,$52,$7E,$04,$7E,$AB,$E8,$7F,$C2,$51,$F0,$CF,$8C,$67,$DB,$D7
	.dc.b	$D2,$48,$FD,$3C,$A3,$97,$C1,$F3,$F9,$29,$5A,$7B,$7C,$AF,$7F,$5A
	.dc.b	$4F,$27,$C5,$F0,$3E,$3F,$80,$F2,$79,$92,$BB,$C7,$9F,$DA,$F4,$7C
	.dc.b	$9C,$9F,$1C,$7A,$FD,$4F,$67,$AD,$27,$CF,$1E,$FF,$32,$47,$A9,$F8
	.dc.b	$5E,$AF,$C3,$F4,$A4,$FB,$9F,$8D,$EA,$7C,$7F,$90,$93,$F0,$9F,$95
	.dc.b	$EA,$FE,$37,$E6,$3F,$9D,$4B,$C7,$46,$1B,$1F,$CF,$DA,$FF,$43,$5F
	.dc.b	$17,$B2,$46,$DE,$BF,$47,$FD,$B2,$3F,$A6,$35,$FF,$69,$B5,$FF,$7A
	.dc.b	$B5,$7D,$13,$B3,$CE,$D8,$FE,$A4,$D7,$FA,$B3,$5F,$EB,$5F,$47,$E8
	.dc.b	$6E,$3F,$02,$31,$1F,$EE,$46,$A1,$F4,$4F,$8C,$F8,$ED,$8F,$EC,$4D
	.dc.b	$7F,$B3,$35,$FE,$D5,$F4,$6E,$BF,$9E,$D8,$FE,$D4,$D7,$FB,$73,$5F
	.dc.b	$EE,$5F,$47,$FE,$F6,$7F,$B8,$35,$FE,$E8,$D7,$7C,$1F,$73,$6F,$5C
	.dc.b	$7D,$6C,$F0,$47,$9F,$D2,$7E,$0A,$CD,$24,$7F,$7F,$FA,$C7,$97,$F8
	.dc.b	$86,$C6,$63,$07,$E3,$3E,$95,$F2,$FF,$E7,$A9,$FF,$05,$46,$3E,$33
	.dc.b	$FC,$14,$BE,$97,$EC,$4F,$D1,$8D,$3F,$E1,$1F,$A3,$FF,$4B,$7C,$2F
	.dc.b	$F8,$45,$B1,$9F,$92,$FC,$54,$62,$3F,$C3,$5E,$E3,$D5,$FC,$73,$6F
	.dc.b	$FC,$34,$D9,$FF,$0F,$35,$FF,$11,$53,$E6,$BA,$92,$BC,$4F,$4F,$5A
	.dc.b	$47,$46,$D7,$F8,$A7,$E9,$1F,$4F,$F9,$F6,$D7,$F8,$AD,$B3,$B9,$F3
	.dc.b	$FB,$12,$BC,$ED,$9F,$F1,$A7,$4F,$BD,$FC,$C1,$F8,$FD,$7F,$59,$19
	.dc.b	$8F,$CE,$35,$FC,$A6,$5A,$25,$FF,$8F,$67,$E0,$3E,$5F,$8E,$F4,$11
	.dc.b	$C1,$BE,$2E,$E4,$70,$7C,$EF,$63,$E8,$91,$C1,$30,$7B,$FF,$E7,$2E
	.dc.b	$05,$7F,$A1,$A3,$FC,$EB,$B3,$E8,$F1,$87,$C0,$BD,$BF,$E2,$6F,$AC
	.dc.b	$F4,$E3,$D2,$3E,$2C,$63,$0F,$F3,$FF,$D3,$3E,$AF,$9D,$FF,$01,$AD
	.dc.b	$5F,$CC,$AD,$3F,$E6,$16,$CF,$F9,$95,$AF,$F9,$A5,$AF,$1F,$83,$EA
	.dc.b	$B9,$7F,$D6,$F5,$FE,$3C,$AF,$F5,$AC,$FF,$11,$9C,$BE,$93,$F8,$A4
	.dc.b	$69,$E5,$3D,$23,$13,$C3,$1F,$6F,$2F,$83,$37,$D6,$6E,$30,$DC,$3F
	.dc.b	$C2,$CD,$C7,$C0,$2B,$8C,$F9,$EB,$8B,$4E,$25,$9E,$33,$C1,$E7,$8A
	.dc.b	$B8,$C6,$1B,$87,$A2,$F9,$CE,$5B,$19,$8E,$B8,$79,$E3,$F1,$0D,$9F
	.dc.b	$C2,$36,$7F,$9E,$61,$E3,$AE,$47,$0A,$F4,$36,$7C,$07,$68,$C9,$D0
	.dc.b	$AF,$4A,$EC,$FC,$76,$D7,$79,$FC,$55,$FB,$9A,$7B,$0B,$5D,$CF,$0D
	.dc.b	$F8,$CF,$3C,$78,$63,$2D,$8C,$C7,$8E,$3C,$AF,$8D,$B5,$E3,$8F,$C5
	.dc.b	$D7,$8D,$B3,$E3,$69,$E3,$AE,$AF,$33,$C2,$77,$29,$97,$75,$DE,$18
	.dc.b	$7E,$06,$BB,$46,$2F,$C4,$D8,$CC,$EE,$3C,$51,$F1,$AF,$E2,$34,$F8
	.dc.b	$45,$8F,$24,$78,$B0,$FA,$64,$9F,$CB,$4F,$ED,$EF,$FC,$25,$7D,$E3
	.dc.b	$13,$FB,$F9,$C9,$1C,$12,$3E,$0C,$FF,$66,$8F,$A2,$FA,$6F,$E4,$B9
	.dc.b	$9D,$D7,$E2,$23,$13,$A8,$C7,$5E,$5B,$7D,$1B,$7C,$A3,$BC,$ED,$A6
	.dc.b	$CB,$1B,$3C,$F1,$B6,$C6,$DA,$6E,$BD,$33,$B6,$CE,$DB,$3B,$8F,$A4
	.dc.b	$FE,$02,$78,$46,$2F,$DE,$DB,$F5,$DF,$C1,$8F,$3B,$6F,$C6,$D3,$B9
	.dc.b	$63,$A5,$F6,$9F,$CC,$CF,$F3,$E8,$C4,$7E,$34,$FE,$26,$FD,$1E,$0F
	.dc.b	$E5,$47,$F2,$BF,$7E,$DF,$F3,$58,$B5,$FC,$6A,$6D,$FF,$57,$F8,$EF
	.dc.b	$E5,$C7,$F3,$3F,$26,$E1,$FC,$F5,$B7,$FC,$89,$B4,$9F,$EF,$3F,$4D
	.dc.b	$FE,$47,$7F,$D9,$1E,$1F,$F4,$88,$C5,$FF,$2C,$8C,$74,$7F,$9A,$67
	.dc.b	$33,$FE,$02,$0E,$1E,$B1,$B3,$8D,$7F,$97,$EF,$C5,$3F,$E6,$1A,$7B
	.dc.b	$5A,$8F,$E8,$70,$F6,$3C,$04,$BF,$04,$F1,$8F,$27,$B1,$E9,$7D,$4D
	.dc.b	$9F,$01,$F0,$3D,$5F,$71,$EF,$F5,$2C,$6A,$FC,$93,$87,$C7,$37,$2F
	.dc.b	$4C,$7C,$0F,$4B,$FB,$07,$C3,$7C,$96,$E1,$CE,$74,$FA,$1A,$7A,$B3
	.dc.b	$4F,$CD,$3E,$77,$99,$CE,$B9,$4A,$31,$5F,$DA,$17,$D3,$7F,$80,$17
	.dc.b	$CD,$58,$C3,$6F,$C5,$3C,$1B,$3F,$EA,$58,$C5,$7C,$E5,$F4,$D9,$25
	.dc.b	$72,$6C,$F8,$DF,$17,$C6,$F8,$DF,$F0,$27,$74,$AF,$89,$6F,$A6,$C6
	.dc.b	$1E,$B7,$C4,$49,$F3,$BE,$5F,$A6,$F7,$3E,$F6,$C7,$4F,$1F,$E1,$79
	.dc.b	$1F,$0D,$B1,$93,$E0,$BE,$F2,$3F,$13,$D5,$F8,$44,$70,$4C,$3E,$84
	.dc.b	$7C,$F8,$FA,$0F,$F8,$87,$E2,$E4,$7E,$07,$E9,$BF,$DB,$EF,$E0,$FC
	.dc.b	$44,$FF,$8D,$A7,$37,$F9,$C6,$BF,$AD,$6B,$FE,$70,$6B,$FE,$C9,$7F
	.dc.b	$DC,$B5,$F8,$D6,$C7,$3F,$57,$FD,$C1,$18,$8F,$DC,$7D,$2F,$F9,$69
	.dc.b	$FF,$18,$B7,$9F,$F3,$AB,$E3,$3F,$C9,$6E,$70,$FF,$40,$FD,$2B,$E1
	.dc.b	$FF,$C9,$ED,$8F,$D0,$2F,$8A,$CC,$EB,$E0,$E5,$6B,$B6,$E7,$FE,$42
	.dc.b	$FA,$57,$D4,$7F,$96,$97,$2D,$8C,$39,$BD,$71,$D7,$F6,$15,$FE,$EB
	.dc.b	$7F,$65,$F4,$DF,$D0,$B0,$FF,$78,$3F,$D9,$E3,$FD,$36,$E6,$0E,$5F
	.dc.b	$F7,$B7,$D3,$33,$FF,$36,$B6,$BF,$7C,$B3,$CC,$FF,$9C,$BE,$2F,$F5
	.dc.b	$7C,$5F,$C2,$1B,$7F,$EA,$A7,$F6,$EF,$F0,$77,$38,$03,$FE,$D2,$9E
	.dc.b	$5D,$1D,$7D,$7E,$CF,$FA,$03,$E9,$7E,$86,$1F,$C4,$7B,$79,$3F,$C4
	.dc.b	$E3,$38,$FF,$9C,$27,$2F,$F9,$CA,$1E,$37,$3F,$1F,$A3,$6F,$F8,$E7
	.dc.b	$D3,$7A,$2F,$87,$BF,$FE,$90,$8C,$69,$FC,$9B,$14,$FF,$A5,$3E,$9B
	.dc.b	$E7,$1F,$36,$CE,$3D,$3D,$79,$FB,$F9,$2B,$9F,$F2,$BA,$47,$FA,$7B
	.dc.b	$E9,$9F,$1D,$7E,$3F,$0D,$BF,$E6,$5E,$23,$E5,$FF,$D4,$AD,$8C,$B8
	.dc.b	$8C,$FA,$D2,$78,$FA,$F7,$6F,$F4,$BF,$D3,$7F,$55,$9F,$E7,$CD,$8C
	.dc.b	$FC,$9F,$FA,$CA,$31,$C7,$FD,$23,$E9,$BF,$D5,$84,$CF,$FD,$6E,$DC
	.dc.b	$3F,$A4,$3F,$C1,$9B,$9E,$A7,$6C,$8F,$E9,$DF,$3B,$FD,$29,$D9,$CB
	.dc.b	$8F,$A9,$DB,$DC,$B1,$D9,$BA,$7F,$AC,$B1,$FD,$11,$F5,$4F,$9F,$3E
	.dc.b	$4E,$2F,$3F,$27,$A1,$B1,$E8,$8F,$F6,$8F,$4B,$E7,$B3,$87,$FA,$FB
	.dc.b	$6F,$D4,$37,$BF,$FD,$AF,$0F,$94,$B1,$86,$63,$FD,$AF,$07,$CE,$E3
	.dc.b	$0F,$F6,$48,$E5,$FF,$70,$B6,$BF,$B9,$2F,$9B,$CC,$E9,$93,$AF,$9B
	.dc.b	$B5,$D1,$E6,$7B,$3C,$2B,$3D,$37,$EE,$78,$37,$B1,$E0,$70,$F8,$AD
	.dc.b	$F4,$BF,$BD,$FA,$67,$57,$F8,$AE,$FD,$35,$A3,$FC,$10,$53,$FC,$15
	.dc.b	$58,$70,$7F,$07,$1F,$4B,$FA,$4F,$A6,$FF,$08,$36,$3F,$C5,$2D,$7F
	.dc.b	$C7,$4D,$79,$2F,$81,$72,$72,$6C,$F2,$6B,$C9,$AF,$2F,$27,$C9,$3F
	.dc.b	$E2,$8A,$FE,$55,$3F,$E2,$56,$BD,$B4,$E0,$E7,$97,$8C,$7B,$79,$1E
	.dc.b	$E6,$2B,$F9,$34,$61,$E0,$75,$E6,$D3,$F1,$3E,$72,$E2,$6C,$E1,$CF
	.dc.b	$C0,$CA,$F3,$45,$7C,$C7,$83,$E8,$BA,$E0,$75,$A7,$F9,$1B,$89,$CD
	.dc.b	$E2,$7F,$8C,$DF,$F9,$15,$AF,$43,$F0,$71,$BE,$5E,$BF,$47,$27,$84
	.dc.b	$7B,$3E,$9B,$FA,$84,$7F,$93,$DF,$97,$87,$F9,$59,$BC,$1F,$85,$77
	.dc.b	$3D,$DC,$BF,$D3,$FE,$99,$F1,$5F,$21,$B1,$D7,$8F,$E5,$FD,$33,$3B
	.dc.b	$DF,$2B,$FC,$C1,$39,$6C,$7D,$2F,$80,$FD,$94,$7F,$88,$FB,$B8,$3F
	.dc.b	$06,$BE,$D3,$7C,$4E,$E7,$83,$A3,$C3,$E0,$F8,$B9,$25,$78,$23,$EF
	.dc.b	$FD,$37,$CC,$E2,$FF,$3B,$E7,$F4,$AB,$FC,$65,$C9,$AE,$4F,$C2,$EC
	.dc.b	$F2,$6F,$A9,$2F,$FC,$F5,$39,$6F,$A5,$C3,$A6,$46,$0D,$F9,$BD,$3E
	.dc.b	$8E,$46,$0E,$7E,$0F,$80,$8F,$6F,$3B,$C7,$D1,$23,$C2,$F9,$BC,$2F
	.dc.b	$3F,$7C,$9F,$2B,$EA,$F0,$7D,$7F,$0B,$27,$D2,$FB,$BC,$2F,$7F,$A5
	.dc.b	$27,$DA,$FC,$2F,$9C,$FD,$3F,$EC,$3F,$E8,$C9,$FF,$4D,$35,$FC,$B3
	.dc.b	$7A,$3F,$D4,$6F,$FA,$42,$37,$C7,$CB,$C0,$91,$83,$9F,$C4,$76,$F9
	.dc.b	$99,$18,$3B,$F8,$1E,$2F,$02,$46,$0F,$1F,$81,$E6,$23,$E6,$F0,$3D
	.dc.b	$3E,$FC,$8F,$4B,$EC,$F8,$9F,$6F,$9F,$27,$DA,$FC,$0F,$81,$F8,$3F
	.dc.b	$13,$F0,$BD,$E9,$5F,$0A,$3E,$2F,$CE,$7C,$6F,$99,$93,$F1,$A3,$E4
	.dc.b	$FC,$27,$CA,$F9,$89,$3F,$2A,$3E,$6F,$A1,$FC,$F3,$9B,$FD,$97,$3F
	.dc.b	$EE,$B6,$BF,$CE,$DA,$FF,$90,$FE,$91,$F2,$DF,$D1,$1B,$1F,$EC,$D6
	.dc.b	$BF,$EE,$66,$BF,$EF,$75,$F3,$3C,$49,$1C,$5F,$E9,$FF,$49,$E0,$83
	.dc.b	$49,$87,$F5,$4F,$A6,$FF,$1E,$CF,$FB,$75,$AF,$FB,$D9,$AF,$F5,$50
	.dc.b	$E1,$FE,$B1,$F4,$8F,$99,$FE,$BE,$D8,$FF,$74,$35,$FF,$80,$2D,$7A
	.dc.b	$7D,$2F,$F6,$32,$3F,$DA,$1F,$EC,$DF,$48,$F9,$BF,$ED,$6D,$9F,$F7
	.dc.b	$A3,$5F,$30,$70,$FA,$7E,$97,$E2,$13,$2F,$EE,$AF,$C7,$F9,$7F,$9F
	.dc.b	$3F,$DC,$DA,$FF,$9E,$DA,$9B,$F4,$AF,$17,$FB,$FB,$67,$C6,$1C,$3E
	.dc.b	$66,$BF,$35,$F9,$8F,$FA,$25,$AF,$1F,$1B,$A9,$FE,$0C,$FA,$47,$47
	.dc.b	$06,$C7,$81,$74,$3E,$0F,$15,$DB,$EA,$BF,$D0,$58,$7D,$D6,$BF,$E1
	.dc.b	$36,$BF,$E9,$B6,$AF,$69,$F4,$BF,$9E,$6E,$FF,$01,$B3,$F5,$1A,$FE
	.dc.b	$2D,$7D,$47,$12,$4F,$17,$F5,$3D,$3F,$D7,$23,$F4,$DF,$03,$82,$3D
	.dc.b	$FE,$2B,$C4,$47,$07,$D2,$E5,$F4,$93,$E7,$BF,$6A,$FB,$3E,$95,$E3
	.dc.b	$7E,$E1,$B5,$F6,$9B,$A7,$E3,$9A,$FE,$97,$EA,$9F,$41,$FC,$01,$B5
	.dc.b	$F8,$26,$CF,$E5,$9A,$FE,$B6,$9E,$37,$84,$95,$E1,$7E,$77,$CF,$C8
	.dc.b	$F9,$CD,$AF,$9C,$A3,$FD,$A5,$E9,$FF,$6A,$8C,$4F,$F8,$5F,$AF,$F8
	.dc.b	$5D,$68,$FF,$29,$71,$2B,$77,$C2,$D8,$FF,$BD,$5F,$F1,$03,$63,$5F
	.dc.b	$4B,$FE,$CA,$C3,$FC,$B9,$39,$7F,$CB,$D1,$F8,$A2,$6F,$FC,$99,$FC
	.dc.b	$6F,$D2,$FF,$0B,$9F,$26,$1F,$EE,$B8,$FC,$13,$FE,$2F,$74,$DE,$2F
	.dc.b	$3C,$FF,$9A,$E7,$38,$FA,$5E,$2D,$F0,$B7,$7D,$6E,$AA,$FF,$5D,$3D
	.dc.b	$FE,$6B,$31,$8B,$E9,$18,$F9,$5C,$B7,$1F,$4D,$F3,$8C,$3C,$E3,$37
	.dc.b	$CD,$F4,$4E,$AF,$AC,$6A,$FA,$B4,$E3,$3B,$78,$99,$BE,$C7,$C7,$9E
	.dc.b	$2D,$8E,$26,$EB,$BC,$64,$E2,$FC,$99,$E2,$D9,$E3,$3B,$8E,$26,$5E
	.dc.b	$31,$F4,$E7,$8B,$4E,$33,$B3,$8B,$99,$E2,$7E,$06,$78,$B6,$38,$9B
	.dc.b	$9E,$31,$93,$89,$F6,$E7,$8B,$67,$8B,$B9,$E2,$E4,$E3,$7F,$55,$FC
	.dc.b	$95,$FC,$F6,$D7,$C9,$3E,$8B,$F0,$DB,$5E,$D6,$9E,$72,$C7,$85,$AF
	.dc.b	$86,$3C,$75,$E1,$8F,$C5,$4F,$77,$C2,$77,$2C,$F7,$6C,$77,$69,$DC
	.dc.b	$A7,$77,$C3,$3D,$C5,$7B,$E7,$C2,$7E,$14,$FF,$02,$47,$E9,$9B,$7F
	.dc.b	$E2,$08,$C4,$7D,$B3,$35,$FD,$92,$33,$1F,$5C,$C9,$EB,$1C,$3F,$CC
	.dc.b	$30,$F6,$47,$9C,$F5,$1A,$E3,$F3,$57,$A7,$0F,$DD,$DF,$1C,$3E,$15
	.dc.b	$F1,$24,$65,$E7,$5C,$67,$F7,$33,$99,$E1,$39,$7F,$D8,$11,$98,$C6
	.dc.b	$1E,$68,$C5,$F9,$0F,$1E,$17,$0F,$B5,$87,$A5,$B8,$72,$51,$2F,$9C
	.dc.b	$6A,$F9,$B4,$E7,$1A,$9F,$95,$5F,$89,$8C,$61,$F2,$A3,$18,$78,$EB
	.dc.b	$E8,$56,$AF,$E3,$D6,$9E,$D3,$AA,$E1,$7E,$F6,$CF,$CE,$BE,$55,$F3
	.dc.b	$30,$F3,$BE,$47,$F1,$91,$A8,$F9,$B1,$AB,$F3,$3A,$BE,$B3,$F8,$0A
	.dc.b	$E1,$83,$97,$F9,$A9,$F8,$0F,$36,$E5,$FC,$BA,$2D,$FD,$89,$B9,$7F
	.dc.b	$45,$F9,$2F,$E7,$A7,$F3,$EF,$8B,$B8,$7F,$36,$6D,$FF,$51,$6D,$7F
	.dc.b	$4D,$8F,$E8,$BD,$DF,$D1,$9B,$5F,$CF,$9F,$D2,$11,$FD,$13,$4F,$EB
	.dc.b	$87,$E6,$3B,$B9,$47,$F4,$AF,$A3,$7D,$2F,$FA,$31,$B1,$FE,$4E,$6D
	.dc.b	$7F,$98,$9A,$FF,$85,$7D,$0F,$EB,$B1,$AC,$7F,$BE,$6E,$9C,$E1,$AF
	.dc.b	$87,$98,$F8,$7E,$16,$E4,$C7,$FC,$09,$92,$F3,$18,$AB,$7F,$9A,$B5
	.dc.b	$BE,$0B,$C7,$F0,$BE,$46,$D7,$BD,$F7,$11,$F7,$3F,$EA,$CF,$AA,$FF
	.dc.b	$55,$C6,$CF,$ED,$EE,$2F,$CD,$1E,$1B,$F0,$3F,$2F,$E0,$BE,$1F,$C1
	.dc.b	$7C,$46,$C6,$4F,$98,$FC,$F2,$3F,$E7,$C7,$FC,$73,$5E,$D6,$BF,$B8
	.dc.b	$6A,$69,$FE,$42,$CF,$FC,$58,$D9,$E5,$F0,$3F,$E1,$28,$C5,$73,$F8
	.dc.b	$1C,$19,$F7,$ED,$FF,$83,$A9,$1F,$ED,$06,$C7,$9E,$30,$7C,$9E,$EE
	.dc.b	$51,$F2,$1B,$7F,$20,$F5,$7B,$BE,$3E,$CF,$23,$6B,$C0,$F7,$23,$F1
	.dc.b	$5F,$17,$B9,$23,$E4,$36,$BF,$D6,$F8,$A3,$E0,$1F,$07,$9E,$D7,$B7
	.dc.b	$DC,$F8,$87,$C5,$69,$F0,$4F,$80,$43,$EB,$1F,$87,$F0,$23,$FE,$B4
	.dc.b	$E2,$FE,$BA,$D8,$FF,$7B,$FD,$23,$EA,$3F,$C8,$6F,$F0,$D9,$FF,$1D
	.dc.b	$B9,$6F,$81,$AD,$37,$DB,$67,$6F,$F9,$33,$E9,$5D,$37,$E2,$70,$DA
	.dc.b	$FF,$23,$FD,$1F,$F4,$F2,$7C,$27,$F9,$51,$C7,$81,$4C,$7A,$B6,$BF
	.dc.b	$C9,$FF,$47,$FE,$AC,$6F,$A9,$FE,$5C,$8F,$F5,$9F,$6F,$F9,$ED,$D6
	.dc.b	$DF,$7C,$7D,$1F,$9A,$FF,$CB,$2B,$5F,$F3,$2B,$99,$E0,$E6,$77,$EC
	.dc.b	$EB,$DB,$DB,$63,$FC,$C7,$F1,$EF,$2F,$FC,$DE,$BB,$6C,$61,$FF,$39
	.dc.b	$F4,$7F,$B6,$6B,$FC,$D6,$DC,$BD,$0B,$E4,$77,$CD,$FE,$7A,$FA,$27
	.dc.b	$C9,$53,$19,$DC,$7F,$CE,$B3,$FE,$F9,$6E,$BC,$7C,$0E,$4F,$FA,$13
	.dc.b	$E8,$BE,$3F,$1F,$FA,$8D,$B5,$FC,$75,$F9,$AD,$E4,$E1,$F2,$92,$3F
	.dc.b	$D0,$72,$F3,$F1,$C1,$F4,$7E,$89,$E6,$FF,$A4,$9B,$7F,$E8,$A7,$F9
	.dc.b	$73,$63,$FD,$2F,$E9,$FE,$6A,$BE,$D2,$FA,$7C,$F2,$69,$DF,$B3,$C3
	.dc.b	$C4,$B8,$FF,$A5,$AF,$F3,$98,$7E,$0F,$E8,$DC,$9F,$EA,$47,$FA,$45
	.dc.b	$7F,$3B,$72,$FF,$A7,$9C,$CF,$F8,$63,$E8,$7C,$9C,$5F,$EC,$96,$D7
	.dc.b	$F4,$37,$FC,$B8,$62,$78,$61,$FE,$B2,$8C,$FD,$1D,$8C,$32,$BF,$D5
	.dc.b	$BF,$24,$F4,$B1,$87,$E6,$F0,$7D,$1F,$33,$A7,$93,$70,$FF,$5A,$CF
	.dc.b	$EF,$6B,$85,$6A,$72,$F8,$FB,$64,$70,$ED,$B5,$DD,$F4,$1F,$EC,$AF
	.dc.b	$9A,$7C,$B6,$2B,$FD,$87,$BB,$E6,$7F,$D9,$ED,$8C,$FC,$C6,$27,$2C
	.dc.b	$8F,$F6,$5D,$BE,$5F,$19,$7E,$E4,$7B,$DF,$ED,$A6,$D7,$FB,$49,$FC
	.dc.b	$83,$63,$5D,$92,$35,$DD,$6B,$A3,$CC,$F6,$7C,$BD,$9E,$9B,$F7,$3C
	.dc.b	$1B,$D8,$F0,$38,$7C,$56,$FA,$2F,$F2,$EF,$D1,$BE,$87,$FC,$0D,$3F
	.dc.b	$BD,$E1,$5E,$23,$81,$4F,$F3,$2D,$D6,$95,$D7,$D1,$7F,$19,$FA,$3B
	.dc.b	$E7,$1A,$4A,$ED,$3F,$CD,$B0,$E3,$EA,$7F,$87,$E7,$2D,$7C,$F3,$B7
	.dc.b	$8F,$C0,$FF,$87,$EB,$D3,$1F,$E7,$A9,$FC,$03,$FE,$7C,$8C,$A8,$CB
	.dc.b	$FC,$CD,$FF,$13,$3B,$CB,$FC,$55,$1A,$9E,$3C,$9F,$E0,$0A,$FC,$2F
	.dc.b	$BD,$E1,$BD,$F3,$D9,$F1,$73,$D7,$C1,$F3,$D7,$23,$67,$0E,$7B,$65
	.dc.b	$78,$CF,$64,$61,$FE,$95,$93,$D7,$EC,$D8,$E0,$7B,$EB,$C5,$C4,$E7
	.dc.b	$BC,$D2,$B7,$39,$6C,$74,$F8,$19,$18,$3A,$FD,$1F,$C8,$7F,$AE,$3F
	.dc.b	$1F,$7F,$F2,$93,$63,$FD,$5A,$EF,$87,$E9,$B9,$87,$D1,$BE,$27,$E4
	.dc.b	$36,$3B,$F1,$7C,$BF,$A3,$65,$7C,$09,$5F,$02,$72,$D8,$FA,$5F,$1D
	.dc.b	$FD,$0E,$3E,$2B,$5F,$EE,$93,$FC,$29,$E5,$CD,$DF,$C5,$91,$83,$B7
	.dc.b	$85,$E3,$E5,$91,$83,$E8,$FD,$1F,$F6,$77,$BF,$27,$88,$FC,$24,$66
	.dc.b	$3F,$C0,$99,$FF,$6A,$7E,$CB,$FD,$11,$BE,$57,$E2,$E3,$51,$C3,$E2
	.dc.b	$71,$D1,$23,$8C,$6F,$E8,$F8,$7A,$32,$7A,$C7,$3E,$F9,$1E,$07,$B7
	.dc.b	$83,$DF,$92,$4F,$91,$F1,$78,$1E,$3E,$E9,$3E,$87,$CD,$DF,$E7,$F5
	.dc.b	$A4,$FB,$1F,$57,$81,$EB,$F6,$E4,$FC,$07,$DD,$EB,$FE,$A7,$E8,$9F
	.dc.b	$29,$49,$5F,$9E,$87,$C8,$EF,$93,$87,$74,$AD,$47,$3F,$8C,$E9,$F4
	.dc.b	$72,$79,$47,$7E,$DF,$07,$74,$9E,$D1,$E3,$ED,$91,$E1,$7C,$DD,$DE
	.dc.b	$7F,$7E,$4F,$95,$F5,$7C,$67,$AF,$CF,$93,$E9,$7D,$DF,$03,$EF,$F8
	.dc.b	$CF,$81,$F2,$B2,$BD,$D1,$F0,$FE,$1B,$E2,$7C,$6C,$9F,$85,$1F,$1F
	.dc.b	$E3,$7E,$47,$C4,$C9,$F8,$D1,$F8,$9F,$96,$FE,$75,$C8,$69,$27,$FA
	.dc.b	$07,$D0,$B3,$FE,$80,$D8,$CC,$60,$FE,$89,$F3,$0F,$A3,$A4,$9F,$E9
	.dc.b	$51,$96,$FC,$CF,$02,$38,$3C,$3E,$A7,$80,$8E,$0F,$EA,$9F,$40,$F7
	.dc.b	$3F,$AA,$36,$33,$18,$3F,$AC,$7D,$0B,$E5,$69,$27,$FA,$F4,$65,$FE
	.dc.b	$C1,$F4,$0E,$0F,$EC,$0D,$8C,$C6,$0F,$EC,$9F,$43,$F2,$E3,$FB,$43
	.dc.b	$5F,$ED,$8D,$7F,$B8,$B5,$FF,$7C,$4F,$F5,$67,$FB,$AB,$5C,$E5,$E2
	.dc.b	$3F,$B9,$67,$E7,$8F,$EF,$0D,$7F,$BE,$B5,$11,$C3,$8F,$A1,$78,$BF
	.dc.b	$DF,$5B,$1F,$E0,$16,$BF,$E0,$56,$BF,$E0,$AF,$A9,$7C,$3F,$F8,$15
	.dc.b	$B1,$FE,$09,$6B,$FE,$0D,$6B,$FE,$12,$FA,$27,$4B,$C9,$FF,$08,$46
	.dc.b	$1B,$D5,$F7,$6B,$85,$6D,$FF,$0C,$7C,$FF,$E5,$F5,$FF,$0C,$36,$33
	.dc.b	$CF,$F2,$D4,$65,$FF,$11,$FB,$6F,$A1,$D3,$6F,$FC,$44,$D9,$FF,$13
	.dc.b	$35,$FF,$15,$F6,$3E,$7B,$B3,$6B,$FC,$54,$D9,$FF,$17,$35,$F0,$FC
	.dc.b	$57,$80,$93,$E0,$7D,$1F,$42,$F2,$E9,$27,$FC,$6D,$19,$7F,$C7,$1E
	.dc.b	$07,$B6,$FF,$02,$7E,$FF,$5F,$85,$19,$8F,$C2,$B5,$57,$F8,$36,$BA
	.dc.b	$F9,$1E,$22,$3D,$BC,$3E,$64,$70,$79,$3C,$5F,$49,$1C,$12,$FF,$DC
	.dc.b	$96,$9F,$F1,$DF,$C1,$FF,$8E,$23,$ED,$36,$33,$5E,$CD,$FF,$CB,$8F
	.dc.b	$F7,$CF,$85,$CE,$1F,$A0,$C1,$E3,$62,$17,$C1,$F0,$C7,$CE,$25,$FC
	.dc.b	$C7,$F1,$1F,$42,$FA,$4E,$4D,$9F,$F1,$7B,$FE,$19,$6C,$6B,$E1,$76
	.dc.b	$D8,$C9,$C3,$E5,$7C,$9C,$FF,$E7,$28,$C1,$FE,$3A,$71,$8B,$6F,$F2
	.dc.b	$DE,$3F,$EB,$B8,$FC,$DD,$79,$2F,$9B,$4E,$45,$8E,$4F,$E1,$2F,$AD
	.dc.b	$F4,$AE,$33,$E2,$AD,$9F,$C2,$AB,$71,$F8,$CB,$F1,$46,$27,$F8,$7C
	.dc.b	$62,$7C,$17,$FE,$72,$AD,$52,$B5,$5D,$EB,$4F,$03,$CC,$59,$F1,$35
	.dc.b	$F8,$2F,$13,$65,$76,$D7,$67,$B0,$F1,$5F,$C0,$71,$5E,$46,$C6,$6B
	.dc.b	$E3,$57,$CC,$3A,$5F,$B2,$F8,$4E,$AF,$B4,$E6,$F9,$5F,$C9,$6C,$FC
	.dc.b	$86,$BE,$48,$EC,$F8,$9B,$5D,$4F,$8E,$F8,$8F,$89,$5D,$DD,$9D,$CB
	.dc.b	$3D,$DA,$F7,$3C,$EF,$22,$1D,$48,$F8,$CD,$B8,$9F,$23,$F4,$A3,$51
	.dc.b	$E6,$6C,$66,$37,$1F,$58,$E8,$7F,$B5,$AD,$7F,$B2,$25,$7F,$B5,$63
	.dc.b	$2D,$8C,$CF,$FB,$6E,$72,$D8,$FC,$44,$3E,$46,$C6,$1B,$5F,$87,$9D
	.dc.b	$36,$74,$D3,$53,$C2,$BD,$A7,$E8,$9C,$5F,$BA,$3F,$41,$18,$9F,$A9
	.dc.b	$1F,$90,$8C,$3F,$54,$C4,$77,$7A,$9C,$9C,$4F,$5B,$F3,$CE,$9F,$E0
	.dc.b	$C4,$AE,$B1,$F7,$E7,$F8,$5C,$62,$75,$5D,$0F,$2C,$E9,$C5,$79,$63
	.dc.b	$0F,$7B,$F8,$AD,$9F,$8A,$70,$7C,$0F,$F0,$9B,$F0,$4F,$F2,$C6,$C7
	.dc.b	$F2,$C6,$CF,$C2,$3A,$1E,$D7,$35,$E2,$3E,$17,$2F,$F3,$63,$F9,$BF
	.dc.b	$75,$BF,$E6,$F1,$6F,$F7,$F3,$6B,$F9,$6F,$C7,$FF,$3E,$E1,$FE,$7E
	.dc.b	$DC,$3F,$9F,$B6,$FF,$83,$B6,$BF,$9A,$F9,$1F,$D3,$F8,$7F,$A2,$57
	.dc.b	$F0,$C8,$C7,$07,$F4,$A7,$FA,$69,$30,$FC,$E3,$FD,$42,$75,$1F,$D3
	.dc.b	$9A,$7F,$19,$FA,$1F,$F3,$D4,$7F,$7E,$8F,$EF,$F1,$87,$01,$C3,$B6
	.dc.b	$BC,$39,$FF,$B3,$47,$1C,$55,$D3,$AF,$35,$AF,$83,$DF,$C2,$5A,$FF
	.dc.b	$27,$DF,$0C,$7F,$CA,$38,$AB,$51,$E4,$85,$FF,$8D,$5B,$7E,$89,$F4
	.dc.b	$4E,$7E,$0B,$CA,$79,$BE,$02,$D7,$80,$F8,$31,$8A,$F8,$D3,$71,$F2
	.dc.b	$47,$C3,$8F,$01,$2B,$C1,$82,$3C,$9E,$EC,$9C,$B6,$BD,$5E,$F7,$C8
	.dc.b	$8F,$92,$D9,$F2,$4F,$F9,$CA,$3F,$DB,$71,$88,$47,$FB,$6A,$30,$EA
	.dc.b	$3F,$AB,$39,$6F,$BD,$FE,$5B,$8D,$47,$E0,$5A,$FF,$90,$1A,$F8,$7D
	.dc.b	$BE,$04,$AF,$9D,$28,$C4,$F8,$1F,$F1,$44,$3D,$BB,$18,$7C,$44,$AE
	.dc.b	$6F,$97,$D8,$91,$D9,$B7,$F9,$4B,$7B,$3D,$A7,$BB,$6B,$E1,$F8,$5E
	.dc.b	$2F,$13,$C6,$D8,$C9,$EA,$7E,$21,$1F,$6F,$A5,$F1,$C8,$E0,$98,$FC
	.dc.b	$E8,$F7,$C7,$C0,$7F,$C9,$1F,$43,$FE,$2F,$CB,$FC,$32,$D9,$FF,$26
	.dc.b	$7D,$0B,$E5,$AF,$76,$1B,$3F,$7E,$7F,$D6,$6D,$EA,$E3,$97,$27,$FC
	.dc.b	$AD,$F4,$2E,$7B,$F0,$D8,$6D,$FE,$27,$E8,$5C,$5F,$E5,$D6,$C6,$7C
	.dc.b	$BC,$4E,$7C,$99,$3C,$7D,$0B,$A7,$F8,$8F,$E8,$5F,$3D,$FE,$67,$6D
	.dc.b	$7E,$5D,$FF,$51,$77,$7E,$A9,$CD,$7F,$8A,$FE,$87,$D1,$5F,$9D,$5B
	.dc.b	$7F,$9C,$5C,$CE,$E3,$1E,$A6,$7D,$8B,$7C,$5F,$F3,$C7,$C6,$BE,$0E
	.dc.b	$F6,$E1,$B5,$FA,$BF,$39,$CD,$FE,$7F,$6C,$67,$E4,$F1,$3F,$E8,$3F
	.dc.b	$A0,$7D,$0D,$31,$95,$BF,$DB,$4F,$FB,$99,$B9,$F1,$EE,$E4,$FF,$A3
	.dc.b	$7E,$87,$E1,$F1,$FF,$0E,$6D,$7F,$AC,$9F,$92,$DE,$4E,$1F,$27,$23
	.dc.b	$F8,$24,$BE,$6F,$1C,$9F,$E5,$7F,$A1,$7C,$A7,$FA,$6D,$B7,$FC,$2D
	.dc.b	$FE,$53,$3B,$9F,$F5,$07,$AD,$FE,$87,$1F,$43,$8E,$4C,$BA,$FA,$1D
	.dc.b	$F8,$97,$2F,$E2,$D7,$F9,$D7,$F8,$C7,$D0,$B9,$3F,$D5,$8F,$FB,$77
	.dc.b	$87,$F9,$6B,$99,$DF,$A9,$FE,$B1,$FA,$0F,$F4,$C1,$30,$C6,$99,$65
	.dc.b	$7F,$25,$97,$D0,$E3,$7F,$F3,$C4,$62,$75,$3F,$EB,$DF,$92,$7A,$5D
	.dc.b	$E3,$C1,$EE,$4A,$E9,$E6,$5C,$79,$B7,$6F,$E8,$B9,$23,$15,$FE,$EA
	.dc.b	$87,$77,$AC,$F6,$77,$59,$F4,$38,$7F,$9F,$7A,$EF,$83,$C6,$5F,$E9
	.dc.b	$1C,$9F,$31,$FE,$D5,$6D,$FF,$45,$5F,$2F,$99,$D3,$27,$AF,$CA,$5B
	.dc.b	$FF,$4E,$0F,$7F,$FD,$C0,$D9,$FE,$A0,$FE,$35,$B1,$AE,$D9,$1A,$EF
	.dc.b	$B5,$D1,$E6,$7F,$BA,$F4,$FF,$7C,$4E,$CE,$BB,$78,$0F,$0E,$D6,$3C
	.dc.b	$2E,$1F,$1D,$BE,$80,$BE,$85,$F0,$5F,$E1,$DB,$F5,$CE,$CF,$F0,$31
	.dc.b	$4F,$F0,$45,$61,$C1,$D0,$7D,$07,$F1,$4F,$A1,$FF,$07,$36,$3F,$C4
	.dc.b	$ED,$7F,$C5,$8D,$79,$3F,$CD,$78,$B8,$FA,$BF,$EF,$59,$E2,$D7,$8B
	.dc.b	$5E,$3E,$FF,$8E,$7F,$C2,$75,$F8,$29,$FF,$02,$35,$FF,$6B,$3F,$E1
	.dc.b	$37,$3C,$9C,$75,$F6,$6F,$FD,$D6,$7F,$1F,$EE,$E5,$F9,$77,$CF,$0E
	.dc.b	$FC,$B9,$9F,$87,$F3,$96,$8D,$9C,$39,$EC,$95,$E6,$8A,$FF,$BA,$F1
	.dc.b	$7A,$FD,$5B,$1E,$CA,$F0,$71,$39,$6F,$2C,$AF,$F1,$24,$E5,$B1,$CF
	.dc.b	$E6,$A4,$60,$FF,$15,$7D,$0F,$C5,$7F,$DF,$8F,$BF,$83,$F0,$0D,$8F
	.dc.b	$C6,$BB,$DF,$FC,$86,$E6,$1F,$42,$F8,$7F,$8C,$D8,$EF,$C7,$F2,$7E
	.dc.b	$85,$95,$EE,$95,$FE,$3D,$9C,$B6,$3E,$87,$C4,$FF,$3D,$8F,$F2,$13
	.dc.b	$56,$7F,$E5,$6A,$FE,$11,$1C,$F9,$BC,$1E,$24,$8C,$1D,$FC,$0F,$27
	.dc.b	$2C,$8C,$1F,$E4,$FF,$A1,$FE,$32,$F8,$39,$3C,$67,$E0,$E3,$31,$E2
	.dc.b	$E2,$FD,$47,$DA,$3C,$5F,$C4,$3F,$C9,$67,$FC,$B4,$D7,$81,$FB,$A7
	.dc.b	$83,$FE,$5F,$F8,$57,$D3,$FE,$45,$7D,$5D,$31,$1F,$95,$FB,$3F,$F7
	.dc.b	$B5,$7C,$63,$E4,$99,$F2,$F9,$77,$73,$ED,$93,$DE,$3F,$3D,$F6,$8E
	.dc.b	$2B,$C5,$2B,$FC,$E5,$39,$7F,$CE,$9F,$7D,$F9,$87,$D6,$3E,$5B,$CE
	.dc.b	$FF,$9E,$5A,$EA,$38,$73,$FF,$A3,$BE,$D5,$F5,$BF,$AF,$7E,$0D,$7F
	.dc.b	$A0,$5A,$F6,$76,$E7,$C7,$E1,$E3,$C8,$C1,$FE,$89,$FB,$E7,$CD,$7E
	.dc.b	$EF,$3F,$EE,$93,$E5,$F5,$BF,$7D,$5A,$8F,$F4,$87,$93,$FC,$1D,$F6
	.dc.b	$BF,$05,$BC,$1F,$C1,$63,$5A,$FF,$AA,$17,$C4,$E7,$E3,$E5,$F4,$8F
	.dc.b	$E1,$9F,$7D,$FB,$98,$F9,$39,$FF,$AF,$9F,$91,$F2,$BA,$F0,$3F,$89
	.dc.b	$7E,$1F,$BC,$FF,$D6,$AF,$E4,$9B,$E6,$6A,$7F,$8C,$7D,$C7,$3C,$3E
	.dc.b	$AB,$FE,$24,$7F,$D5,$8D,$73,$87,$F1,$FF,$BF,$7D,$1F,$C7,$F8,$D9
	.dc.b	$5F,$EB,$39,$CB,$FE,$B5,$FB,$9F,$DC,$BF,$82,$7F,$66,$DF,$33,$F9
	.dc.b	$54,$6B,$4F,$F6,$6A,$F9,$7C,$F8,$B2,$FA,$47,$F3,$1F,$96,$7D,$47
	.dc.b	$D9,$F4,$65,$7F,$A1,$E3,$F9,$B7,$D6,$7E,$4E,$3F,$DF,$CD,$7F,$AE
	.dc.b	$35,$FE,$D2,$D5,$FA,$0B,$E7,$BF,$B3,$B6,$3F,$DD,$CD,$7F,$B3,$1F
	.dc.b	$DC,$9C,$38,$D7,$F2,$ED,$F2,$BF,$B8,$C6,$23,$FB,$73,$5F,$EF,$7F
	.dc.b	$58,$FA,$DF,$EE,$ED,$8F,$F7,$F3,$5F,$08,$70,$F9,$17,$D1,$F2,$24
	.dc.b	$72,$7F,$C0,$79,$7F,$BF,$A3,$E3,$FA,$38,$23,$D8,$7D,$9B,$E6,$3F
	.dc.b	$C1,$6D,$8D,$35,$F2,$B5,$F8,$FF,$5B,$F6,$9E,$8D,$7D,$6D,$77,$BF
	.dc.b	$F8,$63,$EB,$BF,$C2,$C4,$8C,$1F,$E1,$CF,$AE,$FE,$06,$FC,$46,$BF
	.dc.b	$59,$AF,$F8,$75,$AF,$F8,$AB,$EB,$BE,$13,$FE,$21,$6B,$9C,$BF,$C5
	.dc.b	$31,$88,$FD,$2E,$DF,$E2,$CE,$BF,$A3,$1F,$E2,$96,$BF,$E3,$66,$AF
	.dc.b	$AE,$7C,$C7,$F8,$D1,$B1,$F3,$DA,$FD,$E6,$BF,$90,$7E,$63,$FE,$37
	.dc.b	$6B,$FE,$44,$6B,$DF,$6F,$F2,$0F,$D6,$BE,$B7,$6D,$8F,$BC,$BE,$8B
	.dc.b	$FD,$8B,$5F,$8E,$8D,$3F,$90,$FB,$37,$D8,$7C,$66,$ED,$F8,$76,$CE
	.dc.b	$E7,$E5,$F3,$FF,$5A,$AF,$C6,$36,$7F,$36,$D7,$F5,$5E,$E7,$D2,$F8
	.dc.b	$3F,$F7,$4C,$7E,$C1,$AF,$F9,$5D,$AF,$F9,$91,$AF,$E9,$7E,$1B,$ED
	.dc.b	$F5,$7F,$85,$F0,$FD,$5C,$60,$FB,$FD,$BE,$8A,$FD,$EB,$5F,$F3,$23
	.dc.b	$5F,$F3,$8B,$5F,$EC,$1B,$7E,$23,$CB,$FD,$35,$FE,$A5,$B3,$FB,$86
	.dc.b	$BF,$C2,$BE,$B9,$F4,$DF,$90,$6D,$7E,$C5,$B3,$FC,$01,$AF,$F1,$2F
	.dc.b	$B3,$7D,$67,$E6,$1B,$5F,$B9,$6C,$FF,$08,$6B,$C6,$53,$FE,$8C,$F2
	.dc.b	$2C,$FF,$A3,$AB,$8C,$62,$1D,$7F,$E1,$6A,$E0,$7F,$A2,$6B,$6F,$D5
	.dc.b	$F9,$CE,$77,$C7,$81,$18,$9F,$D4,$78,$57,$1F,$B5,$35,$FE,$A7,$1F
	.dc.b	$B4,$EB,$FE,$3B,$5F,$C5,$5B,$3F,$C9,$1A,$FF,$36,$FA,$FF,$C9,$E3
	.dc.b	$E9,$6D,$79,$FE,$27,$B4,$62,$BE,$04,$63,$9F,$C5,$A2,$31,$7F,$02
	.dc.b	$BF,$D6,$7F,$5F,$FD,$5A,$31,$08,$FC,$83,$FE,$26,$74,$DE,$5F,$74
	.dc.b	$63,$2F,$9C,$F0,$6F,$85,$A9,$DB,$6F,$BD,$71,$7F,$D7,$4F,$16,$C6
	.dc.b	$EF,$C4,$DB,$EB,$3C,$AF,$AB,$4E,$B3,$B7,$99,$9A,$E6,$DA,$E7,$1E
	.dc.b	$D8,$E2,$64,$E2,$E4,$E2,$58,$E2,$7C,$2A,$F1,$C6,$67,$8C,$E5,$EB
	.dc.b	$19,$3E,$9D,$79,$DB,$1E,$16,$9B,$3E,$B4,$6D,$B1,$B6,$9B,$BF,$9C
	.dc.b	$FE,$02,$FC,$B7,$E3,$8F,$5B,$4F,$5D,$7D,$49,$DB,$4F,$4C,$6D,$F6
	.dc.b	$99,$9E,$6D,$8E,$71,$E1,$8E,$26,$4E,$2E,$4E,$25,$8E,$2F,$E5,$67
	.dc.b	$C6,$E6,$78,$B9,$7A,$B9,$3E,$4C,$F9,$DB,$1E,$16,$9B,$3E,$4C,$6D
	.dc.b	$B1,$B7,$E6,$9F,$E2,$CA,$FD,$2B,$6F,$FB,$9C,$62,$3F,$12,$66,$7F
	.dc.b	$C2,$D1,$98,$FB,$A6,$4F,$82,$31,$FF,$78,$CF,$B2,$3C,$67,$A0,$D7
	.dc.b	$C5,$7C,$47,$24,$9E,$2F,$4A,$FC,$1B,$99,$E1,$18,$33,$3C,$30,$FA
	.dc.b	$31,$9C,$3F,$86,$D7,$07,$C1,$5B,$C3,$E9,$D6,$DF,$E1,$B1,$9C,$3D
	.dc.b	$13,$97,$FD,$67,$19,$7D,$31,$9C,$3F,$9D,$4E,$63,$15,$F8,$08,$C5
	.dc.b	$F8,$E3,$BE,$17,$0F,$97,$5A,$9F,$95,$7E,$EA,$D6,$1F,$8E,$AD,$5F
	.dc.b	$C7,$C3,$DF,$87,$07,$DF,$7C,$23,$C4,$D9,$FA,$95,$CA,$FE,$4D,$F2
	.dc.b	$9E,$97,$F6,$EB,$4F,$C8,$9D,$57,$08,$D5,$F5,$BF,$B1,$1E,$AF,$86
	.dc.b	$FE,$66,$7F,$40,$E5,$B1,$FC,$BA,$2C,$FC,$E9,$B7,$F3,$3E,$9F,$F9
	.dc.b	$DF,$CD,$7F,$48,$6E,$1F,$61,$B7,$F5,$9B,$5F,$D0,$3E,$63,$FA,$1F
	.dc.b	$CC,$FF,$44,$6D,$FF,$3E,$7F,$46,$47,$F4,$2F,$F8,$07,$D1,$E1,$1F
	.dc.b	$86,$6E,$1F,$D4,$4F,$D5,$92,$3F,$52,$D3,$4F,$E3,$BC,$4F,$EA,$B1
	.dc.b	$FD,$8B,$DB,$FE,$E9,$F5,$EF,$9E,$C5,$FF,$60,$8C,$3F,$96,$9F,$F0
	.dc.b	$A1,$1F,$F0,$A3,$FE,$6E,$87,$B1,$D0,$EB,$F4,$56,$BA,$45,$C1,$AF
	.dc.b	$F8,$4E,$D7,$98,$C3,$D4,$97,$FE,$9B,$9D,$B7,$E1,$7C,$44,$70,$7F
	.dc.b	$A8,$7E,$BB,$FD,$68,$FB,$63,$FC,$29,$18,$C3,$C7,$26,$52,$BE,$3F
	.dc.b	$D3,$FC,$8F,$8F,$B7,$F1,$DB,$7F,$EB,$CB,$57,$C5,$25,$7F,$AC,$DF
	.dc.b	$F1,$CC,$F7,$37,$E1,$FF,$A6,$23,$31,$E3,$6B,$A9,$FF,$0A,$C7,$F8
	.dc.b	$DA,$35,$F2,$BF,$E2,$F8,$C4,$78,$A3,$A1,$27,$C4,$F7,$ED,$91,$E5
	.dc.b	$6C,$FF,$B5,$B3,$FE,$4B,$7F,$E2,$9A,$FF,$12,$C6,$2B,$C1,$F5,$5E
	.dc.b	$1E,$8B,$5E,$66,$E1,$FE,$E0,$C1,$5E,$F2,$57,$C5,$E4,$F8,$44,$70
	.dc.b	$4A,$FA,$E7,$C5,$8F,$81,$C9,$F0,$5F,$BB,$F4,$B8,$AF,$3B,$6B,$88
	.dc.b	$F9,$F7,$D4,$7F,$95,$DB,$5F,$E4,$25,$F2,$F9,$9F,$F2,$E7,$DC,$BE
	.dc.b	$4A,$98,$F7,$AE,$1F,$E4,$AF,$AE,$FF,$2A,$64,$6F,$D7,$BE,$D3,$FC
	.dc.b	$CD,$1A,$C5,$F1,$FD,$8E,$FF,$33,$71,$FF,$2A,$FD,$27,$D3,$CB,$D6
	.dc.b	$DE,$EF,$F3,$F2,$F9,$2F,$F3,$93,$9A,$FC,$87,$DF,$79,$0F,$2F,$D8
	.dc.b	$BE,$A7,$FC,$F1,$1E,$8F,$BA,$7D,$5F,$F9,$ED,$B7,$FE,$66,$5E,$66
	.dc.b	$67,$5F,$27,$2B,$5F,$29,$76,$FA,$34,$F5,$BF,$D0,$CD,$AF,$F3,$8F
	.dc.b	$CC,$FF,$AA,$38,$3F,$D3,$13,$9B,$FB,$34,$F9,$4F,$F4,$73,$4F,$F3
	.dc.b	$C8,$F9,$BC,$C7,$FA,$4F,$EF,$DE,$3D,$31,$F0,$B6,$FF,$D0,$3C,$2F
	.dc.b	$81,$F8,$67,$FA,$6F,$EC,$5E,$6E,$2B,$FD,$0F,$F8,$73,$D4,$FF,$50
	.dc.b	$36,$33,$C9,$89,$FF,$52,$7D,$E3,$EA,$E9,$8F,$52,$DF,$FA,$4B,$F0
	.dc.b	$07,$DB,$FF,$AA,$DB,$1F,$E9,$6F,$B4,$E4,$EE,$7F,$D6,$1F,$13,$FC
	.dc.b	$08,$9C,$58,$F5,$73,$F7,$12,$FF,$D3,$F4,$F9,$CC,$74,$FE,$9B,$95
	.dc.b	$F5,$1F,$84,$3F,$D7,$9F,$68,$F8,$EC,$63,$FE,$A8,$FB,$4F,$C2,$E1
	.dc.b	$FA,$E6,$F7,$FF,$B4,$5F,$C3,$CE,$FB,$FF,$D8,$D1,$8A,$E1,$E1,$E3
	.dc.b	$C5,$CB,$6F,$B4,$7F,$B3,$7C,$8F,$A7,$18,$3C,$5E,$97,$43,$AF,$C1
	.dc.b	$D8,$F3,$37,$0F,$33,$73,$F3,$78,$FE,$43,$CB,$EA,$5B,$F7,$38,$7F
	.dc.b	$D8,$FD,$6F,$0F,$18,$FF,$0E,$C1,$F3,$DF,$EE,$36,$DF,$FB,$39,$FD
	.dc.b	$FB,$63,$5E,$74,$8E,$BC,$96,$FF,$8C,$8F,$83,$FF,$77,$36,$7F,$DB
	.dc.b	$2F,$E4,$4D,$C6,$BC,$19,$3A,$F0,$AD,$74,$79,$9F,$EF,$9D,$1D,$78
	.dc.b	$AE,$BB,$F8,$0F,$0E,$F6,$3C,$2E,$1F,$1E,$0F,$3B,$FB,$77,$DA,$3E
	.dc.b	$CB,$FB,$14,$FF,$33,$AD,$D6,$23,$2D,$7F,$B4,$35,$FE,$F0,$E7,$EC
	.dc.b	$FF,$82,$7D,$A5,$F3,$4D,$25,$75,$9F,$E0,$DC,$9C,$7C,$EF,$F0,$AC
	.dc.b	$E5,$AF,$9A,$76,$F1,$F8,$4F,$EF,$54,$CB,$FB,$E5,$7E,$02,$BF,$C1
	.dc.b	$93,$97,$FC,$1D,$19,$7F,$A9,$BF,$E0,$F7,$73,$CB,$9B,$AB,$FE,$07
	.dc.b	$E1,$7C,$16,$2B,$FC,$31,$18,$78,$9D,$AB,$4A,$FF,$A6,$BA,$31,$EA
	.dc.b	$EA,$0D,$E8,$C3,$9F,$3A,$57,$1A,$DB,$63,$BE,$0F,$89,$A7,$09,$FE
	.dc.b	$E3,$C4,$E7,$F2,$AE,$6A,$62,$2C,$FE,$BB,$C9,$EA,$61,$FE,$F7,$F6
	.dc.b	$9F,$25,$BB,$7C,$08,$F9,$DC,$3F,$46,$51,$B8,$C3,$AE,$1F,$F1,$A7
	.dc.b	$DA,$3E,$3F,$E4,$B6,$FE,$45,$6D,$FF,$24,$7D,$A7,$F4,$C7,$F2,$CD
	.dc.b	$7F,$34,$D7,$F3,$ED,$7F,$C8,$4F,$F0,$97,$8F,$81,$F9,$D8,$C4,$72
	.dc.b	$8C,$72,$C8,$EE,$D7,$BF,$81,$C4,$8F,$91,$FD,$47,$DA,$7C,$F8,$FD
	.dc.b	$63,$5F,$D8,$35,$FD,$AB,$5F,$F2,$C7,$DA,$7F,$89,$1F,$DA,$35,$FD
	.dc.b	$C3,$5F,$DE,$B5,$FF,$30,$7C,$2B,$E0,$BF,$74,$D8,$FD,$F3,$5F,$E0
	.dc.b	$4D,$7F,$83,$FD,$A7,$F8,$41,$FE,$08,$D7,$F8,$43,$5D,$F9,$5D,$BA
	.dc.b	$E4,$76,$6B,$FC,$3F,$ED,$3F,$C5,$0F,$F0,$F6,$B9,$F2,$BF,$88,$46
	.dc.b	$23,$F8,$CF,$DA,$7E,$7D,$FE,$2C,$D7,$F8,$D3,$5D,$F5,$FF,$19,$FB
	.dc.b	$4F,$F2,$54,$7F,$1D,$6B,$FC,$85,$AF,$F2,$66,$BF,$E8,$97,$FC,$3C
	.dc.b	$F1,$F0,$BA,$11,$FE,$51,$F6,$6F,$B4,$FE,$50,$D8,$CC,$60,$FE,$59
	.dc.b	$E1,$FF,$0B,$8F,$E6,$4D,$7F,$9B,$B5,$FF,$4D,$35,$FF,$55,$2F,$93
	.dc.b	$E6,$48,$FE,$75,$19,$7F,$9E,$7D,$9F,$F0,$36,$F8,$1F,$CF,$63,$11
	.dc.b	$FD,$0D,$AF,$F4,$8F,$B3,$AF,$FA,$AD,$AF,$FA,$ED,$A9,$BF,$67,$FE
	.dc.b	$D8,$8F,$E9,$8D,$7F,$A8,$35,$FE,$AA,$D7,$FD,$71,$F6,$7F,$E5,$97
	.dc.b	$FA,$A3,$5F,$EB,$0D,$7F,$AE,$B5,$FF,$60,$7D,$9F,$F0,$17,$FD,$A0
	.dc.b	$D4,$D2,$7B,$7F,$D7,$DF,$F1,$B4,$7F,$63,$6B,$FD,$9D,$AF,$F6,$CF
	.dc.b	$B3,$7C,$E7,$F6,$76,$C7,$F6,$B6,$BF,$DB,$DA,$FF,$74,$EF,$FC,$7B
	.dc.b	$FE,$E5,$6B,$FE,$1F,$F9,$1F,$EE,$51,$8F,$B2,$7C,$C7,$F7,$56,$D7
	.dc.b	$F7,$96,$BF,$DF,$5A,$FF,$80,$BE,$CF,$FC,$98,$FF,$7F,$6B,$FE,$03
	.dc.b	$6B,$BF,$7F,$B6,$92,$3B,$35,$FF,$05,$7D,$43,$EB,$FF,$C1,$0D,$8F
	.dc.b	$F0,$63,$5D,$C6,$BE,$42,$46,$9A,$FA,$7C,$CF,$71,$1F,$73,$FE,$15
	.dc.b	$FB,$1F,$F6,$74,$7F,$86,$5A,$FF,$87,$5A,$FF,$88,$9A,$89,$F6,$5F
	.dc.b	$51,$C0,$70,$EB,$87,$FC,$49,$18,$84,$FF,$89,$DA,$79,$48,$7F,$8B
	.dc.b	$7E,$AB,$E6,$BD,$9A,$F8,$DA,$9B,$F6,$5F,$DB,$63,$FC,$76,$D7,$CE
	.dc.b	$D7,$D8,$D7,$FD,$85,$7F,$E3,$BC,$7D,$04,$7F,$C7,$DF,$62,$FB,$2A
	.dc.b	$4A,$FF,$22,$46,$5F,$F2,$37,$DC,$BE,$17,$FC,$8C,$D8,$CC,$60,$FF
	.dc.b	$25,$7D,$93,$E8,$5F,$1D,$FE,$4F,$8C,$37,$EC,$FE,$AD,$A7,$2F,$F9
	.dc.b	$5F,$F1,$2E,$3C,$5F,$E5,$76,$C6,$6F,$E1,$46,$23,$FC,$C3,$F6,$3F
	.dc.b	$CD,$6B,$FC,$BC,$D8,$FF,$31,$35,$DF,$8D,$F2,$FB,$1F,$55,$49,$7F
	.dc.b	$E6,$78,$CB,$FE,$69,$FB,$2F,$F6,$E3,$7E,$73,$EA,$46,$30,$FF,$38
	.dc.b	$35,$F2,$EF,$F6,$72,$F9,$97,$E6,$8C,$1F,$77,$B3,$FD,$95,$1F,$E7
	.dc.b	$76,$D7,$F9,$ED,$AF,$FA,$01,$AF,$E7,$2F,$FC,$91,$3E,$7F,$37,$CA
	.dc.b	$47,$07,$F9,$66,$37,$2F,$7B,$FC,$7D,$1F,$C4,$A7,$55,$FE,$71,$9C
	.dc.b	$FC,$65,$9F,$03,$6F,$F6,$17,$E8,$AD,$D7,$7E,$FF,$8E,$47,$E6,$3F
	.dc.b	$2B,$CC,$E5,$6B,$FE,$C5,$5E,$7F,$B6,$CC,$62,$7D,$13,$F0,$67,$35
	.dc.b	$A6,$C7,$F9,$DF,$EB,$FF,$CA,$8D,$FB,$2E,$F1,$8B,$E6,$7F,$09,$51
	.dc.b	$AF,$82,$DB,$6B,$91,$D6,$32,$D3,$35,$D0,$57,$7C,$FF,$CD,$8F,$69
	.dc.b	$E1,$3B,$8C,$36,$BA,$B6,$7A,$D7,$29,$CB,$63,$27,$F2,$0A,$F0,$36
	.dc.b	$FD,$57,$DD,$B5,$E4,$6C,$EC,$F2,$C6,$5A,$66,$7C,$87,$37,$2F,$CB
	.dc.b	$9E,$6D,$8C,$BF,$12,$32,$D8,$CD,$7C,$2A,$F6,$4F,$53,$BB,$E3,$6C
	.dc.b	$F8,$E3,$66,$4A,$E4,$F8,$97,$F2,$1A,$7C,$89,$FC,$14,$65,$B1,$97
	.dc.b	$ED,$C6,$5B,$19,$AF,$9F,$1E,$F3,$CA,$5A,$F1,$B6,$7C,$2F,$B6,$32
	.dc.b	$FC,$33,$89,$5C,$CF,$C5,$9E,$31,$EF,$69,$EF,$9F,$49,$D5,$CB,$E1
	.dc.b	$33,$1C,$5B,$1C,$5F,$69,$FD,$A8,$E6,$59,$FD,$53,$6B,$FA,$C4,$62
	.dc.b	$3D,$A6,$67,$93,$C8,$F1,$9D,$AB,$F4,$BD,$BF,$84,$AF,$97,$39,$AF
	.dc.b	$CE,$4E,$6F,$4D,$8F,$C4,$B6,$7E,$4D,$7F,$8F,$27,$F1,$B5,$C1,$A7
	.dc.b	$02,$CF,$0A,$E5,$7F,$41,$B3,$E8,$BE,$95,$DA,$FA,$36,$7A,$34,$E8
	.dc.b	$F6,$BE,$8D,$3A,$5F,$69,$FA,$67,$F1,$57,$0F,$C3,$31,$1C,$5D,$9F
	.dc.b	$C4,$23,$77,$F2,$8F,$3D,$7E,$2A,$FC,$ED,$BF,$3B,$6F,$CE,$7A,$5F
	.dc.b	$3B,$63,$CE,$7E,$E3,$E5,$7F,$96,$9F,$CB,$BD,$9B,$1F,$CC,$E2,$CF
	.dc.b	$F1,$F9,$B5,$FC,$93,$E4,$FF,$9A,$9F,$CD,$B8,$EE,$1F,$CE,$DB,$7F
	.dc.b	$C9,$DB,$5F,$CB,$3E,$4F,$F9,$EF,$C4,$FF,$3E,$6D,$FF,$43,$7F,$2A
	.dc.b	$47,$F2,$77,$FC,$DA,$30,$FF,$33,$7F,$A3,$FC,$9F,$39,$FE,$99,$F2
	.dc.b	$7F,$9E,$23,$CB,$64,$7E,$8A,$FF,$A7,$BF,$D4,$1B,$3F,$E5,$7C,$4F
	.dc.b	$F9,$5B,$EB,$5D,$1F,$D9,$1B,$1B,$35,$D9,$99,$D3,$FE,$47,$9F,$F0
	.dc.b	$CD,$FE,$B9,$AF,$EE,$5A,$F7,$F9,$DE,$DF,$1D,$DD,$B8,$77,$9F,$F3
	.dc.b	$65,$27,$B1,$27,$E4,$C6,$5B,$1C,$DE,$52,$97,$D7,$66,$3F,$B3,$43
	.dc.b	$DD,$F8,$07,$C1,$F9,$2B,$7C,$A5,$1E,$F2,$47,$89,$F8,$5F,$23,$23
	.dc.b	$CA,$DB,$F8,$9F,$5B,$F3,$63,$E7,$36,$7C,$EB,$E7,$3F,$C2,$AB,$EC
	.dc.b	$6C,$61,$B3,$F8,$A9,$E0,$D9,$FF,$3D,$46,$2F,$F9,$0B,$78,$3F,$03
	.dc.b	$1F,$E9,$98,$C5,$7F,$A6,$17,$C9,$F6,$9F,$F0,$64,$3E,$7B,$CA,$79
	.dc.b	$BE,$42,$DF,$A9,$F0,$91,$F3,$3E,$4F,$6E,$47,$A5,$B7,$E6,$FA,$EF
	.dc.b	$83,$1F,$09,$B3,$F1,$BE,$37,$E2,$F8,$DF,$19,$B1,$93,$E2,$3F,$08
	.dc.b	$8F,$C7,$F4,$7E,$29,$1C,$1F,$B4,$E9,$7D,$6F,$FA,$8D,$B1,$98,$C1
	.dc.b	$FE,$A7,$FA,$A7,$C0,$52,$4F,$EF,$A3,$2F,$F0,$0F,$AB,$FB,$8D,$E4
	.dc.b	$FF,$58,$46,$36,$FE,$AE,$D7,$FB,$15,$3E,$27,$C0,$4B,$EC,$FF,$AE
	.dc.b	$7E,$75,$C3,$49,$3F,$C4,$23,$2F,$F1,$2F,$A2,$ED,$87,$FB,$61,$AF
	.dc.b	$FB,$75,$A9,$BF,$54,$F2,$7F,$D9,$0D,$AF,$E3,$AD,$AF,$EE,$2D,$7F
	.dc.b	$BE,$2F,$8C,$E2,$49,$E2,$FF,$B4,$E9,$EF,$52,$4F,$F2,$A8,$CB,$FC
	.dc.b	$B0,$7C,$FF,$FB,$6D,$A6,$5C,$1F,$ED,$FF,$A9,$FC,$FC,$7F,$34,$6E
	.dc.b	$3F,$E0,$86,$BF,$E1,$36,$BF,$1F,$D7,$FF,$74,$67,$FE,$B3,$8C,$57
	.dc.b	$CD,$9F,$F7,$67,$D4,$3C,$AA,$49,$FE,$87,$19,$7F,$A2,$FD,$43,$C8
	.dc.b	$FF,$7A,$36,$33,$18,$3F,$DF,$1F,$51,$FB,$58,$FE,$9A,$DC,$7F,$C4
	.dc.b	$2D,$7F,$C5,$8D,$44,$F7,$BF,$84,$3F,$E3,$16,$BA,$3F,$DE,$3E,$27
	.dc.b	$F0,$A7,$4D,$75,$0F,$67,$11,$D5,$AF,$FB,$E9,$B1,$BB,$FC,$A7,$CE
	.dc.b	$7C,$C2,$38,$3F,$B4,$FD,$37,$E2,$EF,$C8,$D7,$D6,$D4,$D1,$EA,$F8
	.dc.b	$9B,$3D,$23,$06,$7E,$9A,$47,$81,$B5,$FE,$4C,$A7,$AD,$CC,$95,$E6
	.dc.b	$6C,$FF,$AD,$67,$73,$FD,$59,$FF,$7B,$C6,$67,$FD,$AB,$0F,$A3,$A6
	.dc.b	$3C,$2B,$7C,$5C,$3F,$D8,$37,$7C,$06,$31,$FF,$62,$79,$8F,$A3,$FF
	.dc.b	$70,$B6,$FF,$B3,$BF,$C3,$5B,$1A,$F0,$24,$75,$DE,$DF,$FB,$3C,$7A
	.dc.b	$FF,$EE,$C6,$CF,$F7,$27,$F3,$86,$E3,$5C,$F2,$75,$DD,$6B,$A3,$CC
	.dc.b	$FF,$7B,$E9,$F8,$58,$CC,$F5,$DB,$C0,$78,$76,$B1,$E1,$70,$F8,$E5
	.dc.b	$E0,$7E,$83,$E9,$9E,$CF,$FA,$3E,$FD,$55,$A3,$FC,$6E,$53,$F9,$51
	.dc.b	$8A,$FF,$1D,$FD,$2F,$F1,$0F,$A6,$BE,$09,$A4,$AF,$F4,$9C,$FF,$84
	.dc.b	$78,$F8,$FA,$C6,$92,$77,$EA,$FE,$C6,$3F,$61,$5F,$6E,$7F,$C3,$ED
	.dc.b	$79,$6B,$FB,$78,$FF,$0C,$C6,$B9,$79,$4F,$EE,$F8,$3F,$A3,$CF,$E2
	.dc.b	$7C,$7C,$BF,$C1,$2F,$A6,$1E,$0E,$6C,$CA,$5F,$3F,$86,$C7,$92,$71
	.dc.b	$39,$F5,$A4,$FA,$22,$BE,$83,$F8,$9E,$2F,$84,$EE,$D8,$F7,$D7,$8B
	.dc.b	$8F,$FA,$54,$7F,$90,$1F,$D4,$37,$97,$5C,$DB,$79,$FA,$F2,$74,$7F
	.dc.b	$92,$3E,$97,$E3,$BE,$30,$FC,$2E,$0F,$E4,$4D,$8F,$E5,$2F,$0D,$FF
	.dc.b	$92,$39,$87,$D3,$3E,$07,$E4,$B6,$3C,$1C,$7F,$33,$E9,$BF,$C6,$31
	.dc.b	$FE,$5A,$6B,$E3,$6E,$1F,$CC,$5F,$F2,$44,$6F,$BF,$91,$1E,$59,$F5
	.dc.b	$E5,$91,$D5,$F0,$7B,$5E,$1F,$02,$4F,$85,$F2,$7D,$17,$E0,$3E,$9B
	.dc.b	$E6,$CF,$C5,$6E,$7F,$D0,$DB,$C1,$FD,$1F,$4F,$B1,$5F,$2F,$E9,$BF
	.dc.b	$93,$BF,$E7,$66,$B9,$F1,$7F,$1B,$1F,$E3,$28,$FF,$3D,$FB,$4F,$A6
	.dc.b	$FC,$8B,$63,$31,$8F,$67,$F2,$BF,$4D,$FC,$FA,$7F,$D0,$6D,$7E,$E3
	.dc.b	$74,$FE,$B5,$F6,$5F,$2A,$7F,$01,$87,$E6,$9B,$DD,$E0,$3C,$3D,$FF
	.dc.b	$A2,$FA,$77,$0D,$25,$7F,$A4,$A3,$F5,$AF,$E2,$FE,$9F,$E1,$E5,$FD
	.dc.b	$A5,$BC,$3F,$DB,$5B,$3F,$DD,$33,$FE,$76,$FF,$A7,$3E,$9F,$FC,$9E
	.dc.b	$DD,$BF,$67,$18,$8F,$CC,$7E,$23,$FD,$DD,$FC,$D4,$7F,$A7,$DE,$5E
	.dc.b	$37,$03,$8F,$91,$FB,$CF,$A8,$7C,$A5,$25,$7F,$AB,$32,$FF,$41,$3F
	.dc.b	$EA,$EE,$0F,$6E,$DF,$E0,$36,$ED,$FE,$07,$6C,$FF,$83,$9E,$2F,$EB
	.dc.b	$57,$C9,$7F,$0C,$6C,$75,$8C,$7C,$BF,$F1,$0F,$A8,$FD,$FC,$FE,$D5
	.dc.b	$B8,$7F,$84,$DB,$87,$F8,$5F,$1F,$AB,$5F,$EC,$4F,$A8,$CB,$FE,$20
	.dc.b	$6A,$69,$FE,$37,$7F,$D9,$7F,$51,$FE,$D3,$6F,$C7,$FF,$22,$8C,$47
	.dc.b	$F0,$77,$FC,$70,$FF,$B4,$BE,$A3,$FC,$B2,$DF,$84,$FE,$53,$18,$8F
	.dc.b	$E1,$D8,$7E,$01,$7D,$47,$EF,$B2,$FF,$18,$36,$3F,$C6,$ED,$77,$EF
	.dc.b	$7F,$34,$87,$C6,$52,$4F,$FB,$92,$3F,$C6,$8F,$FB,$9B,$EA,$1F,$2F
	.dc.b	$FC,$F1,$B1,$98,$FE,$7F,$8B,$E3,$FF,$A0,$F9,$7F,$8F,$E2,$FF,$21
	.dc.b	$37,$4F,$F2,$33,$70,$FF,$26,$FE,$C4,$F9,$8F,$E9,$3F,$52,$F9,$3A
	.dc.b	$4A,$FF,$7C,$C7,$C3,$3C,$0F,$E6,$CF,$F2,$BB,$87,$0D,$7F,$40,$D7
	.dc.b	$FC,$B1,$4F,$B9,$FE,$F6,$D8,$E8,$1C,$3D,$DA,$F4,$F1,$BF,$C0,$74
	.dc.b	$69,$23,$9C,$BE,$DB,$B9,$23,$2F,$2F,$9D,$91,$E2,$B4,$62,$3C,$5E
	.dc.b	$CF,$A0,$8F,$67,$FC,$2B,$F5,$1F,$D2,$E3,$E3,$35,$FE,$04,$D7,$FC
	.dc.b	$F4,$D7,$EA,$FC,$2F,$C4,$23,$EA,$7E,$3F,$C2,$C8,$F6,$B5,$EF,$F0
	.dc.b	$FE,$22,$3F,$32,$91,$F4,$3D,$D9,$3F,$19,$AF,$CF,$F7,$7E,$A9,$1F
	.dc.b	$96,$FD,$8F,$85,$91,$F3,$9A,$FE,$17,$E8,$FE,$BB,$FE,$97,$6B,$FE
	.dc.b	$9E,$6A,$6F,$D1,$BE,$65,$F1,$9F,$84,$8C,$37,$C7,$61,$9A,$DE,$1F
	.dc.b	$87,$3F,$D3,$BC,$7F,$EA,$CE,$9F,$C1,$C6,$0F,$F5,$0F,$D1,$7F,$2E
	.dc.b	$8F,$C8,$B6,$3F,$3C,$D7,$F9,$D3,$5F,$F5,$93,$FE,$8A,$77,$EC,$F4
	.dc.b	$24,$FE,$73,$E8,$BF,$DA,$93,$FE,$BE,$6A,$69,$8A,$FD,$34,$F9,$3A
	.dc.b	$7C,$E4,$7F,$4F,$F4,$1F,$C3,$9B,$E2,$7F,$AF,$23,$18,$7E,$DD,$AF
	.dc.b	$F5,$3C,$1E,$F7,$FB,$09,$B7,$E7,$8C,$19,$F9,$04,$7F,$B5,$BC,$1F
	.dc.b	$F6,$86,$1F,$B1,$8C,$1F,$ED,$5B,$FF,$43,$4E,$BB,$B8,$E3,$FE,$3B
	.dc.b	$9F,$F4,$53,$E3,$C3,$B5,$FC,$77,$EA,$F5,$F0,$C3,$FD,$BD,$AF,$FA
	.dc.b	$1A,$BB,$DF,$CB,$F3,$B5,$1F,$52,$EC,$FF,$BB,$2B,$83,$FE,$4B,$6D
	.dc.b	$FF,$BA,$71,$FE,$C3,$7D,$29,$5A,$8F,$F7,$87,$D0,$BE,$8B,$AD,$3F
	.dc.b	$00,$F8,$4F,$F9,$5E,$75,$5C,$E7,$38,$FF,$BD,$E7,$35,$FE,$5D,$9C
	.dc.b	$D7,$6B,$EB,$18,$90,$AE,$37,$FF,$00,$67,$53,$9A,$E3,$AF,$9A,$F7
	.dc.b	$87,$2A,$F4,$4E,$6B,$73,$F0,$6B,$75,$E0,$BE,$B1,$88,$E3,$3B,$AE
	.dc.b	$31,$F0,$5B,$87,$BD,$B3,$ED,$3E,$26,$1E,$16,$9C,$CB,$5C,$DB,$5C
	.dc.b	$DA,$73,$8F,$95,$3C,$DB,$1C,$DA,$73,$2C,$F3,$6C,$73,$8F,$99,$3C
	.dc.b	$DB,$1C,$DB,$3C,$DA,$73,$2C,$73,$3E,$86,$1F,$2A,$FE,$1D,$7D,$CC
	.dc.b	$3E,$64,$E6,$BD,$B3,$9A,$DB,$F7,$63,$2D,$3D,$26,$DC,$D7,$4C,$3D
	.dc.b	$51,$E2,$6C,$FD,$D6,$BF,$70,$F6,$46,$CF,$80,$57,$27,$17,$2D,$32
	.dc.b	$73,$C3,$E2,$47,$16,$C6,$DA,$EE,$3E,$E3,$FE,$69,$AF,$D3,$B6,$FF
	.dc.b	$85,$B6,$7F,$C1,$33,$A8,$E2,$43,$E2,$1F,$E1,$97,$F3,$44,$F8,$8F
	.dc.b	$82,$E2,$BF,$51,$18,$72,$DA,$F7,$D7,$EC,$E3,$0F,$C0,$AF,$F6,$74
	.dc.b	$E9,$E8,$FE,$B6,$77,$7F,$02,$77,$5C,$6F,$BD,$F1,$6C,$66,$3A,$5F
	.dc.b	$27,$F1,$D1,$F8,$B8,$C3,$F0,$2F,$F8,$84,$62,$3F,$11,$18,$9E,$C7
	.dc.b	$42,$C7,$AE,$78,$5F,$C7,$9C,$D7,$A6,$73,$7F,$22,$72,$79,$2F,$E3
	.dc.b	$38,$7C,$8F,$F2,$C9,$E2,$FC,$53,$51,$E6,$74,$62,$FC,$51,$F0,$CF
	.dc.b	$34,$7C,$FF,$9A,$FE,$6B,$E4,$FF,$58,$8F,$E3,$91,$87,$F9,$23,$5C
	.dc.b	$9F,$EA,$D3,$F9,$47,$2F,$E3,$AB,$FA,$0B,$5F,$DC,$35,$FE,$68,$DB
	.dc.b	$FF,$02,$65,$FC,$FE,$3F,$99,$78,$FD,$2F,$F9,$77,$07,$F3,$53,$F9
	.dc.b	$B7,$87,$63,$F9,$93,$71,$FE,$98,$DC,$3F,$A5,$3D,$C9,$3F,$A2,$7F
	.dc.b	$C3,$BA,$7F,$56,$E0,$FC,$6D,$7F,$51,$EA,$C4,$FF,$45,$7D,$04,$BF
	.dc.b	$D5,$BE,$DF,$A1,$73,$52,$4F,$F9,$E2,$32,$D8,$FE,$CB,$18,$C9,$87
	.dc.b	$EE,$21,$F4,$7A,$AD,$FC,$77,$72,$61,$D5,$F1,$7A,$9A,$8F,$EE,$D6
	.dc.b	$8C,$63,$DA,$7F,$AF,$4E,$5D,$CF,$93,$E8,$FC,$AF,$9B,$DC,$93,$D3
	.dc.b	$E4,$AE,$1C,$CF,$EE,$38,$23,$3A,$7A,$DB,$5E,$0F,$90,$FA,$1F,$69
	.dc.b	$F0,$9B,$1E,$87,$0F,$AB,$E9,$2E,$3D,$C2,$FA,$3F,$F0,$7B,$67,$FC
	.dc.b	$2C,$BD,$8C,$CE,$9B,$3A,$6D,$75,$AE,$13,$93,$9F,$C8,$71,$7A,$FC
	.dc.b	$7F,$80,$F0,$FC,$85,$8F,$0B,$6F,$A3,$75,$FE,$6B,$87,$FB,$36,$BF
	.dc.b	$C5,$9C,$EE,$DF,$5F,$AD,$F9,$56,$CF,$B2,$3D,$FE,$C6,$32,$F4,$79
	.dc.b	$1F,$08,$F8,$7E,$1D,$BF,$86,$D9,$F7,$37,$3F,$1F,$97,$C2,$BE,$4B
	.dc.b	$6B,$E8,$76,$3E,$93,$FC,$7A,$FF,$A0,$EF,$FA,$7B,$9B,$FF,$23,$39
	.dc.b	$7F,$C5,$9F,$46,$F8,$9E,$38,$7F,$91,$A3,$13,$C2,$3F,$C9,$3F,$49
	.dc.b	$F6,$DF,$C0,$7D,$A3,$BB,$3D,$1F,$E4,$E8,$C5,$FF,$95,$70,$7C,$37
	.dc.b	$5F,$95,$EC,$D9,$FF,$2D,$7C,$DB,$A3,$FC,$AE,$BA,$6C,$61,$FF,$2D
	.dc.b	$FD,$1F,$8F,$87,$FB,$6F,$D9,$BE,$83,$3B,$7F,$98,$23,$19,$7F,$99
	.dc.b	$FE,$9B,$EE,$37,$C1,$FF,$32,$C6,$38,$BF,$38,$A3,$3E,$EE,$BE,$02
	.dc.b	$DF,$37,$FC,$DB,$DC,$F8,$9F,$F3,$7F,$D9,$BE,$7B,$31,$87,$FC,$E7
	.dc.b	$6F,$91,$BF,$69,$86,$C7,$F9,$F7,$EA,$1E,$3F,$F9,$E9,$FD,$1E,$FF
	.dc.b	$B1,$73,$3B,$F2,$3E,$2F,$D4,$7F,$A0,$3E,$D1,$F3,$78,$AF,$F4,$44
	.dc.b	$61,$FF,$45,$7D,$4B,$87,$FD,$0E,$BD,$6B,$18,$7F,$D1,$7F,$52,$F5
	.dc.b	$2F,$DE,$61,$B8,$7F,$A5,$7E,$AB,$ED,$3F,$E9,$7F,$B5,$57,$FA,$52
	.dc.b	$73,$C1,$FC,$2E,$73,$5F,$E9,$DF,$89,$F7,$F7,$FF,$8A,$9F,$C0,$47
	.dc.b	$FA,$15,$CB,$7B,$BF,$D3,$F1,$FE,$BC,$8D,$7D,$BE,$3E,$1B,$2F,$E2
	.dc.b	$BE,$1F,$FD,$4B,$F6,$CF,$B9,$C5,$FF,$AB,$63,$0F,$FA,$BF,$EB,$BF
	.dc.b	$0D,$9F,$F8,$F1,$FC,$34,$67,$C0,$FF,$57,$C6,$36,$FE,$4C,$FE,$49
	.dc.b	$B1,$AF,$17,$6F,$0F,$91,$C5,$F7,$6D,$72,$F8,$5E,$BE,$A7,$66,$D7
	.dc.b	$95,$C3,$E7,$F8,$5B,$5E,$2F,$91,$F2,$9E,$6F,$5A,$D7,$B5,$C3,$FE
	.dc.b	$CD,$E8,$7C,$B6,$2F,$FA,$06,$EF,$4B,$FD,$9A,$DB,$FF,$6A,$3F,$C0
	.dc.b	$1B,$1A,$F4,$A4,$75,$E4,$B7,$FD,$2C,$FC,$84,$7F,$B7,$5A,$67,$BB
	.dc.b	$FD,$B6,$E3,$0F,$F1,$04,$60,$CF,$97,$6B,$93,$87,$D9,$D1,$6F,$C1
	.dc.b	$B7,$43,$AE,$D6,$BC,$4E,$1F,$24,$BD,$2F,$F6,$FF,$D7,$3E,$23,$FD
	.dc.b	$B1,$2B,$F3,$BB,$32,$7F,$B7,$6A,$B4,$A6,$CE,$7F,$5B,$FC,$1F,$EB
	.dc.b	$B6,$D8,$E6,$D7,$7E,$62,$77,$F0,$DF,$E1,$C9,$DB,$5E,$AD,$7F,$DE
	.dc.b	$14,$8C,$47,$FB,$F2,$3F,$C1,$8F,$FB,$FB,$83,$FD,$3B,$1F,$83,$8C
	.dc.b	$69,$CB,$8F,$FD,$F5,$5A,$F8,$5B,$39,$AF,$C7,$7B,$18,$47,$F0,$D7
	.dc.b	$0B,$0E,$CF,$F8,$45,$FE,$9A,$F8,$0F,$37,$97,$C2,$3F,$90,$C6,$4E
	.dc.b	$5C,$D9,$9D,$7C,$2C,$9D,$57,$07,$B4,$78,$BA,$7F,$9E,$47,$E9,$7D
	.dc.b	$2E,$0F,$F8,$8F,$8D,$CF,$E8,$5C,$D4,$C4,$59,$FD,$77,$A3,$D4,$C3
	.dc.b	$FE,$2D,$FA,$EF,$9E,$DE,$2F,$87,$1F,$0B,$8B,$FD,$9A,$E9,$B1,$F2
	.dc.b	$3E,$B9,$F1,$94,$E3,$1F,$33,$EB,$7F,$A6,$E1,$F1,$5A,$FF,$8F,$5B
	.dc.b	$7F,$ED,$A7,$FC,$75,$1B,$EF,$E4,$47,$96,$5D,$79,$24,$75,$7C,$1E
	.dc.b	$EF,$87,$C0,$93,$E1,$7C,$9F,$55,$F8,$0F,$AE,$F9,$33,$F6,$5A,$FF
	.dc.b	$94,$5B,$9F,$97,$8F,$F0,$FF,$B1,$BE,$5A,$FC,$CC,$C3,$F0,$53,$97
	.dc.b	$F0,$9E,$EF,$F7,$F6,$FB,$FF,$90,$8C,$47,$F9,$7D,$FE,$FC,$F0,$F3
	.dc.b	$74,$47,$9F,$9B,$E2,$F0,$24,$78,$1F,$2F,$CE,$F5,$F3,$3B,$79,$B2
	.dc.b	$BC,$71,$E1,$F9,$9F,$17,$9D,$27,$CF,$1F,$A4,$F9,$A7,$CC,$52,$4F
	.dc.b	$E7,$E1,$F6,$DE,$BF,$47,$D9,$EE,$CA,$F8,$31,$EF,$F1,$3E,$07,$89
	.dc.b	$27,$E2,$C7,$7F,$A1,$F0,$78,$72,$7B,$47,$ED,$BE,$C9,$F6,$14,$93
	.dc.b	$FB,$18,$FF,$28,$37,$E6,$78,$11,$D7,$B5,$ED,$E9,$91,$EB,$7E,$07
	.dc.b	$4F,$7F,$2F,$C1,$F0,$12,$BB,$47,$8F,$E3,$BC,$9F,$33,$27,$C5,$1F
	.dc.b	$27,$B7,$E5,$76,$C9,$F9,$11,$EB,$F4,$E4,$7A,$5F,$77,$A9,$EF,$F5
	.dc.b	$24,$FB,$5F,$85,$E7,$7D,$5F,$6F,$F9,$DF,$59,$A4,$AF,$E8,$3F,$45
	.dc.b	$FE,$61,$6F,$9B,$FD,$0A,$31,$1F,$D2,$1A,$FF,$4E,$A7,$D5,$76,$24
	.dc.b	$7F,$4C,$8C,$BD,$BE,$63,$A1,$1C,$1F,$D4,$BE,$E3,$FD,$D2,$DE,$7F
	.dc.b	$EA,$71,$88,$FE,$B2,$D7,$FA,$FE,$0F,$98,$F4,$12,$3D,$AF,$F5,$CF
	.dc.b	$B8,$FC,$9C,$7F,$60,$6B,$FD,$91,$AF,$F6,$A6,$BF,$D1,$03,$FB,$27
	.dc.b	$87,$3F,$52,$3F,$33,$C9,$F9,$44,$70,$7F,$70,$FB,$C7,$C3,$7F,$70
	.dc.b	$6C,$66,$30,$7F,$74,$FA,$67,$81,$49,$3F,$DE,$A3,$2F,$F7,$CF,$B9
	.dc.b	$FE,$F6,$9A,$47,$FB,$FF,$DC,$FC,$B7,$DC,$D7,$FB,$1B,$53,$7E,$E5
	.dc.b	$97,$F8,$21,$B3,$FE,$0C,$6B,$FE,$10,$6B,$FE,$16,$CF,$CE,$FF,$85
	.dc.b	$1A,$E7,$1F,$F0,$9C,$62,$3E,$97,$84,$FB,$9F,$F0,$BB,$63,$FC,$36
	.dc.b	$D7,$FC,$3E,$D7,$F3,$1E,$7F,$E9,$08,$FE,$5D,$FF,$10,$FD,$4B,$E7
	.dc.b	$FE,$9C,$FF,$8B,$23,$1B,$FF,$8A,$17,$A9,$FE,$2F,$9F,$F1,$6C,$69
	.dc.b	$FF,$17,$FD,$CF,$DE,$DB,$FC,$5E,$D8,$CD,$FC,$47,$FC,$23,$19,$7F
	.dc.b	$C7,$7E,$F3,$E9,$B5,$1F,$E6,$46,$C6,$6F,$FC,$76,$D9,$FF,$1F,$FD
	.dc.b	$27,$FB,$CD,$BA,$72,$8C,$47,$F9,$21,$AF,$E7,$FE,$A9,$F3,$1E,$06
	.dc.b	$E1,$FE,$48,$6C,$FF,$93,$1A,$F3,$F5,$FB,$12,$7F,$C9,$F6,$8F,$5F
	.dc.b	$7B,$D8,$A4,$AF,$E1,$4F,$87,$DC,$FC,$6F,$BB,$F8,$CC,$BF,$DB,$D5
	.dc.b	$FE,$2D,$5E,$16,$A3,$D5,$57,$3F,$F1,$ED,$6A,$7D,$55,$ED,$23,$ED
	.dc.b	$EF,$E1,$4A,$FE,$1D,$3F,$E3,$BD,$1F,$59,$A9,$CD,$32,$FD,$0D,$6A
	.dc.b	$FC,$D8,$74,$8C,$56,$E7,$E8,$4F,$3F,$73,$A7,$BD,$67,$DC,$DA,$E0
	.dc.b	$7B,$1C,$61,$F1,$5B,$3F,$6D,$B3,$B8,$F9,$15,$D9,$A7,$62,$C7,$63
	.dc.b	$E1,$4F,$26,$C6,$AB,$C0,$D3,$C0,$58,$F0,$1C,$F1,$5F,$86,$30,$DA
	.dc.b	$FB,$CD,$AF,$CA,$36,$7E,$DB,$5F,$AC,$DA,$EB,$3E,$B9,$EA,$DA,$EA
	.dc.b	$D3,$A9,$63,$AB,$5E,$AD,$9E,$B1,$EB,$BF,$89,$1F,$E4,$19,$F8,$13
	.dc.b	$F1,$8F,$09,$B7,$31,$B8,$CD,$7C,$A6,$CF,$CA,$69,$F2,$A3,$D2,$F3
	.dc.b	$32,$77,$72,$73,$2C,$73,$6C,$73,$8F,$82,$F3,$8C,$BC,$E3,$31,$CD
	.dc.b	$A7,$32,$C7,$33,$E4,$9C,$DC,$BC,$CC,$C7,$36,$C7,$36,$9C,$CF,$A2
	.dc.b	$7F,$0F,$BF,$E2,$84,$8F,$E2,$8D,$8D,$46,$08,$FC,$6F,$AA,$7C,$AF
	.dc.b	$D8,$AF,$CE,$46,$2B,$EA,$C6,$27,$51,$F7,$27,$55,$F4,$30,$FC,$1E
	.dc.b	$1F,$22,$72,$FC,$A8,$CD,$7C,$29,$CB,$F1,$23,$35,$F0,$04,$E9,$C4
	.dc.b	$69,$F8,$75,$FD,$BA,$75,$18,$AE,$98,$7F,$1A,$9C,$DF,$12,$46,$70
	.dc.b	$F6,$DF,$C5,$6E,$1E,$B6,$E1,$E9,$6C,$F9,$F0,$EB,$78,$8F,$4C,$DC
	.dc.b	$3D,$15,$AB,$F9,$A7,$FB,$95,$C5,$7C,$BB,$F2,$4E,$A3,$C8,$DC,$3C
	.dc.b	$8D,$3C,$85,$8F,$26,$1E,$1A,$E7,$1D,$67,$ED,$4F,$F1,$BF,$AA,$FF
	.dc.b	$6B,$C7,$FB,$DD,$B3,$FC,$85,$FF,$18,$C6,$7D,$9F,$E8,$B1,$AC,$3F
	.dc.b	$C3,$ED,$9F,$F2,$43,$C2,$B9,$57,$0F,$47,$8F,$A5,$6F,$C0,$DA,$F0
	.dc.b	$37,$0E,$AF,$83,$97,$C3,$C9,$6F,$D0,$DB,$FF,$2C,$01,$97,$F7,$19
	.dc.b	$CC,$66,$30,$E1,$AF,$2F,$BB,$7C,$BF,$5C,$BF,$C7,$B1,$8B,$F9,$18
	.dc.b	$FC,$4F,$7F,$FC,$2F,$89,$A4,$9F,$0E,$3E,$48,$E1,$1F,$47,$DF,$B3
	.dc.b	$F4,$4F,$9F,$F2,$7F,$41,$B3,$F5,$DE,$E4,$7A,$DF,$C1,$7B,$FD,$5B
	.dc.b	$D1,$FE,$FA,$D8,$F8,$DA,$7F,$96,$23,$FC,$54,$D9,$DF,$81,$FE,$32
	.dc.b	$9F,$F1,$76,$1F,$39,$AF,$FA,$1D,$AF,$D1,$D7,$F9,$F3,$F2,$F9,$FF
	.dc.b	$C7,$71,$88,$FF,$47,$35,$FB,$12,$F6,$BC,$04,$CB,$D6,$D9,$FE,$ED
	.dc.b	$3B,$9F,$59,$E3,$F6,$BD,$F5,$F0,$3D,$99,$3A,$BE,$FE,$D6,$1E,$6F
	.dc.b	$BC,$91,$F2,$9F,$85,$F1,$F2,$3E,$6B,$67,$F1,$B8,$BE,$AF,$E7,$CF
	.dc.b	$D0,$6D,$78,$69,$5B,$8F,$AF,$3F,$3C,$8F,$E2,$17,$CF,$FF,$99,$1B
	.dc.b	$19,$8C,$11,$ED,$1A,$AF,$F0,$E2,$F8,$DF,$F1,$FB,$9E,$1F,$F4,$CA
	.dc.b	$F8,$DC,$9A,$F7,$2E,$1D,$1F,$11,$1F,$33,$E5,$F1,$64,$7A,$5B,$8F
	.dc.b	$EC,$AD,$F5,$7E,$29,$F1,$B6,$BD,$2E,$1F,$91,$F7,$6F,$53,$18,$7F
	.dc.b	$8D,$FE,$BF,$FC,$D9,$5F,$35,$FF,$41,$69,$FE,$B4,$73,$3B,$F3,$7F
	.dc.b	$CB,$F0,$F8,$8A,$63,$E2,$6E,$1F,$5F,$EA,$BF,$1D,$87,$F9,$15,$B8
	.dc.b	$FF,$B0,$17,$C3,$7F,$9A,$9C,$D7,$FB,$AF,$EA,$DF,$25,$FE,$6E,$6C
	.dc.b	$7E,$0D,$7C,$96,$67,$5E,$D6,$FE,$B6,$DE,$DF,$F3,$BF,$D5,$3C,$BB
	.dc.b	$F1,$B8,$6D,$7E,$3F,$EF,$9F,$53,$FE,$7E,$5E,$1D,$8C,$39,$EA,$95
	.dc.b	$C7,$C0,$BC,$7F,$E6,$5F,$AA,$F4,$B7,$7F,$F4,$4C,$7A,$AB,$F3,$CE
	.dc.b	$B9,$3F,$DC,$BA,$39,$7F,$D1,$ED,$AF,$D2,$2F,$2F,$33,$FE,$94,$FA
	.dc.b	$9F,$E2,$B5,$FC,$F7,$24,$7F,$1B,$7F,$CF,$0E,$9F,$E8,$4E,$70,$FF
	.dc.b	$7A,$AF,$90,$E9,$F2,$1D,$5B,$7F,$B5,$FB,$E7,$C8,$FF,$A9,$1B,$19
	.dc.b	$F5,$71,$39,$F4,$B5,$F3,$B6,$B9,$3F,$EA,$CF,$A7,$70,$DF,$47,$0D
	.dc.b	$AF,$E0,$98,$FE,$75,$BF,$15,$FE,$B4,$8C,$70,$7F,$A5,$5F,$ED,$DC
	.dc.b	$0A,$F8,$91,$FD,$72,$31,$3F,$D8,$E3,$13,$AF,$63,$A7,$AF,$FE,$C0
	.dc.b	$FA,$77,$CB,$D3,$1F,$5F,$6F,$F8,$C7,$D4,$7E,$3D,$BE,$07,$FB,$22
	.dc.b	$31,$9F,$FA,$A5,$46,$73,$D7,$55,$BE,$6F,$1E,$4C,$61,$C5,$BB,$7A
	.dc.b	$A1,$8F,$EE,$A3,$0F,$EF,$AF,$F0,$8E,$5B,$E5,$79,$77,$91,$EA,$DE
	.dc.b	$E1,$E7,$F2,$7D,$47,$AF,$DB,$B5,$F0,$1C,$3F,$CE,$7C,$A7,$C4,$62
	.dc.b	$FF,$D9,$1A,$3B,$3F,$DD,$6D,$BF,$E8,$0F,$E4,$5B,$1A,$EC,$91,$D7
	.dc.b	$C0,$B7,$FE,$D3,$3F,$C1,$91,$FD,$25,$A6,$7C,$0F,$F7,$CB,$8C,$3E
	.dc.b	$43,$FA,$B9,$E1,$3A,$F0,$36,$D8,$E9,$1F,$88,$AF,$A1,$9B,$EA,$35
	.dc.b	$87,$8F,$6E,$E7,$83,$6B,$8F,$95,$C3,$E7,$97,$D3,$FF,$8D,$7E,$A1
	.dc.b	$E0,$7F,$8E,$2F,$D3,$5A,$3F,$C4,$E5,$3F,$C5,$35,$87,$07,$F0,$3F
	.dc.b	$A5,$7D,$43,$EA,$3F,$C5,$F1,$86,$C6,$67,$FC,$87,$39,$6F,$BA,$DF
	.dc.b	$8B,$63,$31,$83,$AF,$BB,$FC,$4D,$FC,$B5,$7E,$06,$7F,$BF,$35,$E9
	.dc.b	$AF,$9E,$B8,$3E,$9E,$5B,$3F,$E5,$9E,$17,$B3,$88,$FE,$9A,$DC,$3A
	.dc.b	$63,$E0,$E6,$C7,$D4,$4B,$F0,$C5,$7C,$27,$F9,$96,$9E,$5F,$33,$57
	.dc.b	$EA,$7C,$18,$BE,$36,$9B,$BF,$F0,$D7,$1F,$F3,$B6,$F2,$F9,$E3,$FB
	.dc.b	$64,$6B,$9F,$13,$D3,$DD,$93,$C0,$FF,$12,$7D,$43,$8B,$E0,$3E,$EE
	.dc.b	$2F,$E8,$0D,$9F,$F3,$F3,$B7,$8F,$17,$C7,$FA,$87,$C4,$53,$9C,$7C
	.dc.b	$BF,$A8,$FF,$11,$57,$F8,$CD,$AF,$C6,$0E,$38,$7F,$89,$3F,$E3,$28
	.dc.b	$E1,$DF,$CC,$8F,$3D,$3B,$73,$48,$EC,$F8,$7E,$07,$C5,$E0,$49,$F1
	.dc.b	$3E,$5F,$6B,$F0,$1F,$51,$F2,$A7,$C8,$DC,$FF,$92,$37,$83,$FB,$26
	.dc.b	$9F,$5E,$BF,$C9,$DF,$50,$F8,$8F,$C5,$2F,$98,$A6,$23,$F1,$DF,$05
	.dc.b	$FC,$E2,$BF,$CA,$CD,$7E,$23,$72,$FE,$68,$FF,$8C,$E3,$7E,$6F,$22
	.dc.b	$3C,$BC,$DF,$0E,$72,$3C,$AF,$93,$E3,$BC,$3E,$5F,$8B,$CC,$95,$E2
	.dc.b	$8F,$2F,$D4,$79,$BC,$D9,$3E,$68,$FD,$37,$CA,$BE,$A2,$92,$7F,$CE
	.dc.b	$10,$FA,$8F,$7F,$A1,$F0,$3D,$B9,$5F,$02,$3E,$1F,$85,$F1,$3C,$39
	.dc.b	$3F,$12,$3B,$FB,$1E,$0F,$06,$4F,$68,$FD,$C7,$CE,$7F,$6C,$8F,$F4
	.dc.b	$03,$5F,$C7,$37,$9B,$FA,$F2,$F5,$F8,$FC,$84,$AE,$2F,$4F,$A8,$F7
	.dc.b	$FB,$32,$7D,$AF,$C2,$EA,$F0,$F9,$9E,$2F,$A8,$95,$E2,$8F,$2F,$C1
	.dc.b	$F9,$BE,$46,$4F,$9A,$3E,$5F,$77,$CC,$EE,$93,$F2,$A3,$DB,$E6,$C8
	.dc.b	$F6,$BF,$03,$D4,$F8,$3E,$9C,$9F,$82,$FC,$4F,$37,$F2,$7F,$11,$F6
	.dc.b	$3D,$D9,$5F,$98,$8F,$E4,$1F,$4D,$FE,$93,$8F,$F5,$43,$FC,$91,$BE
	.dc.b	$76,$BC,$EF,$C0,$FD,$44,$9F,$BE,$F2,$9E,$7F,$23,$27,$9B,$DB,$D6
	.dc.b	$EF,$F2,$3F,$D4,$31,$FF,$4A,$D7,$FB,$E1,$AF,$F9,$44,$07,$2E,$7C
	.dc.b	$47,$8D,$FD,$69,$B1,$C4,$E2,$E1,$C3,$5E,$D2,$D3,$C8,$48,$F2,$3D
	.dc.b	$7E,$72,$47,$51,$48,$CC,$78,$69,$E5,$FB,$C9,$1E,$F7,$FB,$8F,$9E
	.dc.b	$F8,$EA,$49,$ED,$19,$7C,$1F,$4B,$FC,$C1,$BE,$2F,$F7,$A8,$C4,$7C
	.dc.b	$16,$BF,$1F,$C9,$F8,$C4,$7D,$59,$A3,$E5,$79,$32,$72,$D7,$A7,$2F
	.dc.b	$72,$3E,$47,$E9,$79,$32,$3A,$B5,$FA,$3E,$4F,$D7,$23,$E2,$7F,$C2
	.dc.b	$7E,$3B,$E8,$A9,$27,$E8,$46,$5F,$A5,$F6,$3F,$6D,$FF,$4C,$35,$FC
	.dc.b	$4B,$53,$7E,$8F,$E8,$7D,$1F,$DB,$6C,$FE,$25,$FE,$75,$F0,$B9,$E2
	.dc.b	$57,$E0,$7E,$85,$DB,$FE,$42,$C3,$F2,$F1,$8F,$57,$53,$99,$4E,$5F
	.dc.b	$CB,$7B,$AF,$8D,$D3,$6F,$F1,$CD,$9D,$CF,$3A,$FF,$77,$C7,$E9,$DA
	.dc.b	$FF,$AE,$5A,$F9,$3D,$6F,$17,$D0,$3E,$8A,$93,$1F,$CE,$46,$5B,$E2
	.dc.b	$70,$23,$83,$7E,$FF,$42,$38,$39,$F8,$7E,$02,$38,$3B,$FC,$5F,$F9
	.dc.b	$8D,$FF,$68,$5F,$F9,$8D,$FE,$03,$EB,$7F,$A5,$61,$F4,$5B,$D3,$F8
	.dc.b	$2C,$6D,$FF,$06,$36,$78,$5E,$E5,$3C,$AF,$A9,$E0,$9D,$47,$F7,$B8
	.dc.b	$CC,$FD,$CF,$85,$CF,$BD,$CF,$7F,$F3,$6D,$6A,$BA,$CE,$6B,$73,$9A
	.dc.b	$DD,$79,$AB,$53,$C6,$FC,$13,$FE,$25,$AD,$E3,$F0,$6F,$73,$C6,$FA
	.dc.b	$61,$DA,$38,$36,$BF,$D7,$97,$CA,$BE,$6D,$F2,$9E,$91,$C8,$F1,$46
	.dc.b	$5D,$99,$BF,$01,$D9,$D9,$99,$E7,$19,$36,$F8,$AF,$C9,$3F,$95,$6C
	.dc.b	$FC,$D8,$E6,$F1,$9E,$67,$18,$CB,$63,$AC,$73,$32,$57,$27,$52,$9F
	.dc.b	$B2,$9F,$0B,$63,$8B,$B9,$F0,$C7,$36,$C6,$63,$E2,$BE,$76,$C7,$94
	.dc.b	$DC,$75,$3C,$37,$F4,$DF,$CC,$DF,$CF,$6D,$7C,$B3,$99,$E1,$2B,$E1
	.dc.b	$36,$78,$4B,$5E,$18,$DC,$78,$63,$C7,$5E,$18,$FC,$BC,$F7,$78,$BD
	.dc.b	$DA,$77,$36,$77,$2C,$F7,$7C,$67,$F7,$9C,$3F,$36,$DB,$FE,$22,$4A
	.dc.b	$FE,$1E,$D8,$FE,$F1,$3C,$1F,$4B,$FA,$4A,$FE,$05,$E1,$FD,$88,$C5
	.dc.b	$EA,$31,$79,$6C,$7D,$CB,$FE,$03,$18,$7F,$07,$5F,$DD,$A7,$4F,$48
	.dc.b	$F8,$27,$08,$CD,$75,$69,$C2,$BA,$5F,$E0,$2F,$A5,$F6,$BE,$0D,$BE
	.dc.b	$07,$27,$C7,$7C,$A3,$D8,$66,$FA,$46,$67,$CB,$7D,$A7,$C1,$7D,$9B
	.dc.b	$7D,$AB,$C0,$76,$7A,$1D,$9C,$C7,$68,$CC,$76,$BF,$03,$F8,$08,$FE
	.dc.b	$6F,$18,$BF,$91,$3F,$1C,$E7,$3D,$6B,$F4,$7F,$3F,$FE,$C8,$8F,$E5
	.dc.b	$0F,$FA,$2D,$BE,$1F,$F3,$78,$C5,$7F,$B1,$DE,$0D,$F7,$75,$3B,$78
	.dc.b	$7A,$5F,$DB,$E8,$D2,$5F,$F8,$E6,$32,$F8,$A3,$C1,$E9,$72,$8F,$2F
	.dc.b	$2F,$8C,$F2,$7B,$D6,$32,$4C,$BC,$D1,$96,$C7,$78,$F2,$11,$F1,$78
	.dc.b	$BE,$FE,$1F,$80,$D8,$F8,$AF,$AC,$8F,$03,$D5,$DB,$EF,$23,$83,$DD
	.dc.b	$7F,$2E,$3E,$63,$63,$FA,$C7,$CF,$3D,$CF,$F0,$B3,$63,$FC,$BC,$D9
	.dc.b	$FF,$36,$35,$FF,$3B,$39,$E0,$4F,$F9,$E2,$53,$98,$C3,$FE,$1B,$9D
	.dc.b	$CF,$F9,$FA,$34,$FF,$A0,$3E,$7F,$F1,$71,$BA,$E9,$18,$EC,$FE,$F9
	.dc.b	$2F,$62,$FB,$38,$6D,$FB,$9C,$3F,$03,$D9,$B8,$B2,$F1,$BE,$8F,$63
	.dc.b	$A9,$DB,$D9,$B7,$F2,$1B,$3C,$63,$DB,$18,$AE,$79,$3E,$27,$FC,$11
	.dc.b	$EE,$FA,$1B,$7F,$49,$C3,$F5,$3E,$5E,$E3,$D4,$F8,$CE,$30,$EA,$D9
	.dc.b	$FA,$0B,$E7,$FF,$C4,$0B,$E8,$6C,$61,$B3,$FE,$19,$9E,$0D,$F8,$AC
	.dc.b	$E1,$E1,$12,$F1,$BA,$9D,$BD,$DB,$7C,$DB,$5C,$69,$18,$CB,$E7,$C3
	.dc.b	$DD,$FF,$16,$76,$FA,$1B,$87,$95,$C3,$E7,$F7,$6E,$3D,$5F,$79,$FE
	.dc.b	$3F,$FA,$07,$C1,$62,$BF,$C6,$DF,$40,$F0,$3F,$C8,$8B,$E2,$2C,$61
	.dc.b	$CF,$AF,$2B,$8F,$B1,$73,$F8,$9F,$41,$F9,$19,$FF,$21,$36,$33,$DB
	.dc.b	$FE,$51,$8C,$70,$FE,$3B,$E8,$1E,$8F,$F9,$55,$B8,$7F,$92,$D7,$C7
	.dc.b	$66,$75,$EB,$EF,$E0,$ED,$ED,$FF,$2F,$7D,$07,$F1,$CD,$3E,$C2,$EE
	.dc.b	$CF,$3E,$9F,$F3,$27,$D0,$3E,$82,$F8,$98,$6E,$5F,$E5,$D9,$FC,$0B
	.dc.b	$7A,$F8,$F8,$3C,$9F,$F3,$77,$CE,$FA,$B4,$FC,$43,$78,$BF,$5A,$FB
	.dc.b	$2F,$FC,$D0,$E5,$FE,$11,$A3,$97,$FC,$F0,$D8,$FF,$37,$2F,$82,$CC
	.dc.b	$FF,$9F,$7E,$83,$F8,$36,$20,$FF,$0C,$6F,$93,$BF,$2F,$87,$C5,$F1
	.dc.b	$F8,$BB,$9F,$F9,$EF,$04,$7F,$A2,$7E,$81,$F0,$D4,$C7,$A9,$6F,$FD
	.dc.b	$05,$3F,$EE,$B6,$FA,$7C,$7D,$3E,$4F,$FA,$4F,$E8,$1C,$57,$6C,$36
	.dc.b	$FF,$D1,$98,$FE,$69,$BF,$23,$FE,$9B,$8C,$70,$7E,$D5,$46,$7D,$69
	.dc.b	$5C,$BC,$6B,$7F,$BB,$FA,$07,$D0,$7F,$A9,$5B,$5F,$E9,$B5,$F2,$79
	.dc.b	$9F,$F5,$57,$D0,$7E,$37,$AB,$F9,$3B,$F8,$DB,$FF,$51,$39,$7F,$A1
	.dc.b	$39,$C3,$F8,$52,$D3,$9F,$57,$46,$D7,$2E,$4C,$34,$EB,$1E,$2D,$3C
	.dc.b	$7E,$55,$9F,$0B,$71,$EF,$3E,$8D,$B1,$9F,$A3,$CA,$F5,$9E,$CF,$6A
	.dc.b	$E3,$EA,$70,$FF,$AF,$7C,$B7,$C3,$63,$2F,$E4,$98,$7E,$1A,$BF,$D8
	.dc.b	$4D,$8C,$F6,$7F,$B4,$63,$07,$F4,$E0,$E4,$FF,$6B,$5F,$E8,$E7,$FD
	.dc.b	$C3,$1F,$ED,$97,$F1,$87,$F3,$32,$CF,$F5,$87,$F4,$95,$FE,$D3,$72
	.dc.b	$FF,$73,$72,$71,$EF,$E9,$DF,$D5,$B3,$C6,$3C,$7F,$07,$8A,$F0,$69
	.dc.b	$E3,$3C,$9A,$5B,$F0,$38,$7C,$52,$F9,$FF,$DA,$7D,$03,$B7,$C9,$7E
	.dc.b	$DA,$D1,$E7,$3F,$C1,$55,$5A,$53,$67,$B7,$E7,$5F,$3E,$FA,$0E,$11
	.dc.b	$86,$C6,$67,$B4,$E5,$BE,$FF,$F8,$0D,$BC,$1C,$A3,$11,$DD,$AF,$F8
	.dc.b	$7E,$5E,$FF,$F8,$86,$BF,$C1,$71,$FC,$2A,$7F,$8A,$46,$3C,$5D,$6D
	.dc.b	$C6,$ED,$7F,$19,$BE,$0F,$F7,$1E,$4F,$ED,$2F,$E4,$FD,$AD,$A9,$DE
	.dc.b	$3E,$0E,$5C,$C3,$E8,$29,$88,$B5,$E4,$CB,$D5,$E5,$F0,$33,$1F,$E0
	.dc.b	$1B,$FF,$0D,$3F,$A3,$F8,$BF,$0E,$67,$84,$73,$3F,$99,$71,$FF,$3B
	.dc.b	$70,$7E,$07,$97,$D3,$06,$F3,$E1,$EB,$EF,$4A,$E3,$5B,$79,$FD,$07
	.dc.b	$D0,$6F,$17,$C4,$8F,$87,$AF,$F9,$4D,$D4,$FF,$86,$DC,$BF,$D1,$FE
	.dc.b	$81,$F1,$1F,$31,$B1,$D0,$F4,$FC,$FB,$2F,$A2,$D8,$CC,$63,$E1,$33
	.dc.b	$DF,$27,$8B,$C2,$78,$F3,$49,$EA,$F4,$F8,$3E,$BE,$E7,$6F,$72,$57
	.dc.b	$8A,$3C,$3C,$7F,$82,$FA,$0F,$89,$3F,$E3,$76,$F5,$7F,$97,$35,$FA
	.dc.b	$D3,$F4,$BF,$64,$7C,$9F,$E2,$9B,$19,$8C,$7C,$BF,$E3,$BE,$9B,$FA
	.dc.b	$BC,$FD,$A5,$EA,$64,$D7,$A9,$BE,$CE,$1F,$3F,$2F,$84,$7E,$63,$E7
	.dc.b	$3F,$CA,$D1,$F8,$47,$FC,$AC,$DF,$8B,$D7,$3F,$87,$C6,$93,$E1,$7C
	.dc.b	$9E,$FF,$97,$C6,$93,$E5,$7D,$1E,$3F,$7E,$BF,$07,$5C,$AF,$04,$7B
	.dc.b	$7C,$59,$1E,$D7,$E0,$78,$FF,$07,$8E,$4F,$C1,$7E,$27,$BD,$EB,$ED
	.dc.b	$93,$EB,$7D,$DF,$2D,$FB,$6F,$A6,$FC,$34,$FF,$9A,$DF,$82,$DE,$0D
	.dc.b	$69,$BA,$E1,$A4,$AF,$D4,$DA,$7E,$57,$77,$CB,$F0,$25,$7C,$B8,$F0
	.dc.b	$78,$12,$38,$BE,$38,$F2,$7C,$6C,$9E,$AF,$9F,$C1,$FA,$C4,$8F,$AC
	.dc.b	$FD,$9A,$F6,$7C,$BF,$B7,$D1,$95,$E6,$8F,$81,$F1,$1F,$07,$E2,$24
	.dc.b	$FA,$A3,$F0,$5F,$1D,$23,$F0,$4F,$E1,$72,$FE,$3B,$F3,$DF,$1B,$C6
	.dc.b	$FF,$5A,$B7,$CB,$FF,$67,$36,$FF,$C8,$0F,$E2,$9F,$F4,$C3,$95,$DE
	.dc.b	$FA,$4F,$C9,$1D,$B9,$BB,$F3,$F8,$38,$E5,$FF,$18,$C3,$FD,$40,$FF
	.dc.b	$1A,$F9,$E7,$CD,$FF,$32,$6C,$66,$38,$BE,$2D,$7D,$64,$7A,$72,$7B
	.dc.b	$B7,$FE,$B9,$81,$A4,$9F,$EC,$58,$3E,$87,$FB,$13,$63,$31,$83,$9F
	.dc.b	$14,$8F,$ED,$56,$8C,$47,$F6,$FD,$DC,$9F,$DB,$5B,$1F,$DC,$5A,$EE
	.dc.b	$3C,$FF,$35,$23,$AB,$5E,$9C,$9C,$C8,$F3,$7F,$BD,$7D,$07,$F2,$E8
	.dc.b	$FF,$00,$35,$FF,$0C,$35,$FF,$3E,$07,$0F,$F9,$FF,$37,$C4,$FC,$42
	.dc.b	$47,$C8,$7E,$3F,$BB,$23,$E5,$B5,$F0,$FC,$17,$90,$8F,$F8,$46,$FF
	.dc.b	$D0,$2F,$F8,$4F,$B7,$FC,$60,$F9,$1A,$FF,$A4,$5A,$9B,$87,$F3,$68
	.dc.b	$FF,$1A,$35,$F6,$B5,$F6,$7A,$FF,$E1,$FF,$A0,$78,$74,$93,$FE,$25
	.dc.b	$8C,$BF,$E2,$7F,$9F,$6B,$FC,$F2,$3F,$C8,$51,$8F,$47,$5F,$48,$F8
	.dc.b	$0E,$17,$BB,$F9,$FF,$4D,$FD,$1B,$C0,$FF,$19,$36,$3F,$C6,$CD,$7F
	.dc.b	$C9,$CD,$7E,$BF,$D0,$FE,$9A,$7F,$C7,$4D,$8F,$F1,$F3,$5F,$F2,$B3
	.dc.b	$5F,$BF,$E3,$FF,$4E,$9F,$F2,$13,$63,$FC,$8C,$D7,$FC,$BC,$D7,$F0
	.dc.b	$FC,$EF,$13,$EA,$B6,$BF,$39,$18,$3E,$CF,$D0,$3C,$2A,$49,$FF,$29
	.dc.b	$46,$5F,$5F,$C2,$FE,$03,$23,$49,$1C,$3C,$03,$FE,$22,$AF,$F1,$43
	.dc.b	$7C,$1C,$E6,$F1,$AA,$AD,$B9,$AF,$D1,$C6,$27,$6D,$8C,$D7,$49,$CD
	.dc.b	$D9,$E3,$9D,$C3,$FD,$D1,$87,$F0,$29,$F3,$4E,$67,$C5,$DD,$6B,$E5
	.dc.b	$47,$8C,$E3,$19,$9D,$36,$BE,$03,$67,$81,$D2,$3B,$9A,$2B,$AA,$F2
	.dc.b	$36,$BA,$CE,$B0,$F8,$37,$D5,$B1,$93,$E3,$C6,$5A,$66,$BD,$2D,$7C
	.dc.b	$87,$77,$C8,$D3,$C9,$1C,$DE,$26,$67,$C2,$D9,$EE,$D3,$71,$F6,$2B
	.dc.b	$D7,$19,$3C,$31,$99,$F0,$B6,$3C,$31,$E7,$BF,$8B,$39,$3C,$25,$8F
	.dc.b	$0B,$67,$C3,$1F,$82,$3A,$95,$EA,$78,$E7,$AC,$66,$7A,$B4,$EA,$59
	.dc.b	$EA,$7E,$32,$BD,$B3,$E1,$8F,$6C,$ED,$A6,$DF,$1B,$4F,$1C,$71,$AF
	.dc.b	$1C,$66,$3C,$6D,$9F,$1B,$4F,$19,$FA,$BA,$F1,$C7,$86,$B9,$CE,$63
	.dc.b	$C2,$D3,$C2,$FC,$57,$FD,$D9,$1F,$A2,$6C,$FF,$0A,$24,$FF,$08,$6C
	.dc.b	$7F,$B7,$67,$83,$E1,$8F,$D7,$66,$FA,$3F,$C6,$4F,$87,$0F,$91,$39
	.dc.b	$C3,$F0,$17,$A8,$C5,$7D,$78,$C5,$F2,$C3,$F6,$F8,$72,$8C,$4E,$B0
	.dc.b	$E6,$7E,$87,$0E,$75,$F3,$E7,$58,$75,$AD,$5F,$26,$9C,$A3,$F3,$57
	.dc.b	$CA,$3D,$C6,$A3,$91,$A9,$E4,$D8,$E5,$7E,$76,$9F,$87,$AE,$93,$F7
	.dc.b	$4F,$8D,$19,$AE,$75,$C6,$7B,$B4,$F6,$4F,$8D,$B1,$E1,$37,$1C,$CF
	.dc.b	$0B,$DE,$BC,$2F,$80,$F1,$61,$F3,$A7,$FB,$0D,$FD,$1E,$FF,$E5,$1F
	.dc.b	$45,$FC,$A5,$B3,$FD,$0D,$B7,$FC,$7D,$B8,$FF,$42,$FA,$1F,$F6,$AD
	.dc.b	$7F,$88,$AB,$FC,$47,$18,$70,$1C,$3B,$6B,$C3,$DF,$FE,$79,$1C,$6B
	.dc.b	$B7,$C8,$E2,$73,$F1,$B3,$0F,$F7,$CD,$BE,$37,$18,$FF,$3A,$8C,$4E
	.dc.b	$AB,$AD,$6B,$1F,$53,$FE,$26,$AD,$DF,$A1,$AF,$9E,$35,$F1,$9E,$63
	.dc.b	$CF,$C9,$71,$F1,$BC,$C8,$FC,$2A,$F8,$51,$DF,$C6,$93,$9C,$3F,$C8
	.dc.b	$ED,$8C,$D7,$AD,$F8,$BE,$37,$C6,$7E,$3F,$67,$A4,$95,$CD,$F6,$56
	.dc.b	$A3,$E3,$B4,$F8,$C7,$D0,$3F,$B6,$7D,$0F,$F8,$DA,$36,$23,$FB,$2F
	.dc.b	$D0,$3E,$5F,$FB,$93,$67,$5B,$7F,$81,$A3,$FC,$48,$D9,$EB,$C7,$C3
	.dc.b	$E0,$6D,$F0,$3F,$CD,$77,$C1,$BF,$43,$DB,$E0,$3B,$B6,$38,$65,$E5
	.dc.b	$A4,$EA,$73,$29,$7C,$D6,$63,$D7,$EE,$FA,$8F,$5F,$BD,$6F,$8D,$A3
	.dc.b	$15,$F0,$9B,$AF,$BD,$E3,$5E,$0F,$A1,$7D,$2E,$2F,$E4,$FC,$07,$C8
	.dc.b	$3E,$4F,$BF,$6F,$BB,$67,$E7,$B7,$1F,$F1,$8A,$F6,$BF,$C4,$4B,$E0
	.dc.b	$EC,$61,$B7,$F0,$A7,$83,$7E,$03,$38,$7F,$88,$27,$FC,$94,$DF,$23
	.dc.b	$AF,$B9,$D9,$B7,$CD,$B3,$B8,$F8,$1E,$0E,$33,$FF,$23,$EC,$8C,$4F
	.dc.b	$FA,$A3,$E8,$7F,$CE,$15,$FE,$4C,$6C,$67,$D8,$FF,$1F,$46,$30,$FC
	.dc.b	$92,$ED,$CF,$6E,$9B,$7F,$80,$FA,$17,$C3,$7F,$91,$D7,$C3,$D8,$C3
	.dc.b	$FE,$4B,$FA,$17,$05,$E0,$C3,$70,$FC,$5C,$FF,$2B,$D3,$F4,$8D,$4F
	.dc.b	$F7,$83,$1F,$42,$FA,$2F,$F2,$C3,$6B,$F2,$AF,$C5,$6C,$7F,$97,$7E
	.dc.b	$87,$F3,$71,$F9,$A6,$DF,$F9,$59,$77,$EF,$BF,$87,$C4,$F1,$F9,$6B
	.dc.b	$87,$17,$FC,$D1,$B7,$BA,$3F,$4A,$FE,$03,$97,$FC,$FA,$E7,$2F,$F3
	.dc.b	$1B,$99,$FF,$6F,$F9,$7F,$C5,$A7,$FC,$CE,$BE,$53,$FC,$E8,$E6,$FF
	.dc.b	$D0,$EE,$6B,$F6,$7F,$42,$FA,$5F,$F3,$D3,$63,$3F,$3D,$89,$CF,$87
	.dc.b	$27,$8F,$8B,$73,$FF,$3B,$FD,$13,$2F,$F4,$22,$FA,$3B,$18,$73,$EA
	.dc.b	$4C,$3F,$82,$D3,$E2,$B1,$87,$FA,$0B,$E8,$FF,$8E,$37,$CE,$FF,$48
	.dc.b	$47,$9E,$31,$7F,$C3,$BD,$4F,$EB,$0F,$F2,$78,$FE,$A9,$EA,$FF,$A5
	.dc.b	$E5,$F4,$96,$31,$F4,$39,$F8,$FB,$87,$37,$86,$1C,$7D,$5B,$3D,$DF
	.dc.b	$F5,$17,$D1,$7B,$E3,$F8,$FB,$67,$FD,$36,$FC,$58,$CF,$07,$0C,$F8
	.dc.b	$F0,$D9,$E2,$F4,$F2,$71,$87,$26,$F4,$7F,$B0,$B1,$7B,$3E,$2D,$7C
	.dc.b	$6D,$C7,$C6,$E1,$F2,$FB,$37,$1F,$2F,$C5,$7A,$8F,$5E,$D6,$BD,$6E
	.dc.b	$1F,$E7,$7E,$0B,$E0,$F1,$7F,$EB,$8D,$1F,$3B,$FE,$C5,$6D,$FF,$42
	.dc.b	$5E,$0E,$67,$4C,$9E,$BF,$27,$6F,$FD,$86,$7E,$6E,$3F,$C3,$DE,$77
	.dc.b	$FB,$46,$3F,$89,$47,$F4,$E7,$47,$E0,$7C,$ED,$1B,$F9,$EB,$3B,$70
	.dc.b	$FB,$3E,$0E,$D7,$83,$7E,$C7,$7D,$ED,$77,$70,$F8,$65,$F4,$3F,$DD
	.dc.b	$FE,$8B,$F6,$91,$F9,$3A,$F4,$3A,$3F,$C2,$85,$3F,$C2,$D3,$87,$07
	.dc.b	$E6,$FE,$85,$F3,$6F,$A2,$FF,$0E,$46,$1B,$19,$9F,$F2,$0C,$E5,$BE
	.dc.b	$2B,$8F,$8B,$63,$31,$83,$AF,$8D,$FA,$B5,$A7,$8F,$93,$0D,$AF,$EF
	.dc.b	$F8,$7F,$77,$6C,$E9,$FF,$0D,$73,$7F,$3E,$7F,$17,$E9,$65,$57,$E2
	.dc.b	$9E,$7C,$D9,$87,$9B,$4C,$45,$9F,$0F,$BB,$B3,$0F,$BF,$2F,$95,$1D
	.dc.b	$F1,$7C,$4D,$38,$57,$F8,$63,$CE,$73,$F9,$57,$35,$31,$16,$7F,$4B
	.dc.b	$EE,$75,$30,$F8,$4F,$F1,$D7,$D1,$7C,$57,$01,$F8,$0E,$1B,$C3,$F4
	.dc.b	$23,$8F,$17,$C7,$9C,$EB,$FC,$7F,$E8,$9E,$47,$CB,$6D,$FF,$8C,$AB
	.dc.b	$6F,$F9,$2F,$E8,$59,$7D,$16,$C7,$F8,$E1,$AF,$99,$FF,$10,$39,$F0
	.dc.b	$38,$11,$E1,$3C,$F9,$64,$73,$7B,$7B,$3D,$FE,$77,$C1,$E1,$4A,$F0
	.dc.b	$47,$8F,$8F,$F0,$9F,$45,$F4,$27,$FC,$96,$D7,$C9,$1F,$89,$FA,$27
	.dc.b	$CA,$D2,$4F,$F9,$4A,$1F,$31,$F8,$FF,$A2,$FF,$09,$CF,$F9,$59,$7C
	.dc.b	$5E,$4D,$7B,$7B,$EB,$E1,$F0,$32,$F8,$47,$E6,$7E,$53,$E2,$C7,$DC
	.dc.b	$7F,$C8,$8F,$F9,$89,$AE,$7E,$27,$C5,$E2,$C8,$F1,$3E,$5F,$A2,$F3
	.dc.b	$76,$49,$F3,$3E,$9F,$1B,$C1,$D1,$E1,$E8,$95,$DE,$3D,$DE,$14,$8F
	.dc.b	$73,$F0,$7C,$4F,$85,$C5,$27,$E1,$3F,$17,$E2,$7D,$9D,$52,$7D,$4F
	.dc.b	$BF,$E8,$BF,$71,$F3,$BF,$86,$F0,$FF,$2D,$48,$F8,$71,$86,$F8,$7F
	.dc.b	$DD,$5E,$0D,$AE,$7A,$4B,$FF,$44,$D2,$7E,$77,$6F,$CF,$EF,$95,$F3
	.dc.b	$63,$C7,$DF,$23,$8B,$E6,$8F,$3F,$CB,$C9,$EA,$FA,$BB,$FE,$D1,$23
	.dc.b	$EC,$3F,$76,$BD,$FF,$2F,$F0,$3D,$09,$5E,$68,$F8,$7F,$33,$F1,$3E
	.dc.b	$7E,$4F,$AA,$3F,$0D,$8F,$F4,$0E,$C3,$49,$3F,$ED,$7F,$A0,$FC,$3B
	.dc.b	$FE,$FD,$6B,$FE,$1D,$6B,$FE,$C5,$6A,$7F,$26,$F1,$ED,$FE,$9F,$18
	.dc.b	$8E,$51,$8E,$49,$1E,$16,$BF,$E2,$B3,$B6,$FC,$88,$F9,$DE,$BB,$48
	.dc.b	$F5,$35,$E4,$79,$70,$F1,$91,$F8,$0F,$F6,$4F,$C0,$3F,$17,$1E,$16
	.dc.b	$BF,$E3,$D6,$BF,$EE,$56,$BF,$07,$5F,$79,$1F,$98,$FF,$70,$FA,$AF
	.dc.b	$F6,$2C,$79,$03,$87,$CE,$D7,$E1,$35,$FF,$27,$3F,$E8,$D7,$93,$5E
	.dc.b	$3F,$19,$DC,$FA,$3C,$D2,$3E,$F3,$5F,$8B,$F5,$2F,$5B,$FC,$0A,$D8
	.dc.b	$F3,$35,$DC,$7F,$82,$BE,$8F,$FD,$7D,$1E,$B6,$BF,$41,$AF,$DB,$6B
	.dc.b	$FB,$3F,$9C,$FF,$09,$F4,$9A,$48,$F8,$DE,$4B,$E1,$FF,$C3,$2D,$8C
	.dc.b	$C6,$0F,$27,$C0,$C8,$F2,$35,$FC,$1E,$9E,$24,$7F,$AD,$FE,$53,$FC
	.dc.b	$4B,$19,$8C,$1F,$5F,$E8,$FF,$A1,$3F,$95,$6B,$FE,$81,$6B,$D1,$AB
	.dc.b	$EA,$3F,$D2,$0F,$E0,$5A,$FE,$59,$AF,$E8,$5A,$FF,$A3,$7E,$8D,$D7
	.dc.b	$FD,$42,$3F,$3B,$18,$F4,$7F,$18,$BE,$8F,$EC,$4F,$1A,$E0,$FA,$4F
	.dc.b	$CD,$E1,$F0,$B8,$FD,$D1,$83,$F3,$FF,$44,$F8,$BA,$49,$FC,$D4,$65
	.dc.b	$FC,$E7,$CF,$FF,$64,$6E,$7F,$A8,$8C,$5F,$ED,$1A,$FE,$FF,$D5,$79
	.dc.b	$3F,$AE,$6D,$FE,$9D,$B3,$FB,$C6,$BF,$C1,$FE,$8B,$F9,$C4,$FE,$B5
	.dc.b	$B1,$FC,$09,$AE,$FC,$8F,$DD,$F2,$7F,$BE,$CF,$F7,$E4,$BC,$7F,$F1
	.dc.b	$17,$95,$FD,$3A,$73,$1F,$C0,$A7,$58,$77,$8C,$19,$F2,$38,$C7,$FB
	.dc.b	$16,$9E,$46,$32,$FC,$8E,$0B,$C4,$7E,$7E,$6C,$E6,$FF,$2F,$8F,$8E
	.dc.b	$73,$7C,$30,$EB,$5A,$3F,$DA,$97,$C7,$0F,$F2,$0C,$E9,$FF,$21,$C6
	.dc.b	$70,$E9,$5C,$67,$CB,$7D,$70,$EF,$39,$9F,$33,$6F,$93,$EF,$9D,$CF
	.dc.b	$81,$B7,$DA,$3C,$37,$D6,$B9,$36,$33,$1D,$F0,$F4,$57,$F8,$03,$0E
	.dc.b	$F5,$FB,$A9,$CD,$7C,$0B,$E5,$7B,$6C,$64,$E9,$87,$91,$F4,$1E,$23
	.dc.b	$25,$BF,$53,$6B,$D4,$75,$77,$1E,$83,$65,$8D,$C7,$86,$33,$1C,$5B
	.dc.b	$19,$8D,$C7,$D2,$8E,$B1,$B3,$30,$C3,$E1,$BF,$50,$F0,$4F,$B1,$A7
	.dc.b	$C7,$AF,$80,$D8,$F6,$1B,$AE,$CD,$3C,$85,$8F,$21,$E8,$2C,$79,$2B
	.dc.b	$C7,$87,$D4,$9C,$D7,$17,$D1,$87,$D4,$72,$78,$CF,$39,$EB,$BF,$1B
	.dc.b	$96,$B9,$9F,$2B,$96,$B9,$3C,$E7,$F0,$79,$FD,$B4,$7F,$07,$6C,$7F
	.dc.b	$07,$6B,$B8,$D1,$F1,$9F,$DC,$F8,$1F,$29,$F5,$C6,$5F,$A9,$19,$C3
	.dc.b	$EA,$91,$C5,$65,$B5,$C6,$7F,$89,$CE,$2B,$F6,$F5,$6F,$E0,$D6,$AF
	.dc.b	$85,$FF,$08,$AD,$47,$C2,$9D,$5F,$CC,$7C,$51,$C0,$F0,$4F,$09,$FA
	.dc.b	$67,$AA,$34,$F1,$8D,$3F,$00,$F2,$C6,$AF,$CB,$3A,$AE,$F5,$E1,$BF
	.dc.b	$81,$7B,$3C,$D5,$B8,$F5,$46,$67,$C7,$5E,$AB,$F8,$35,$A9,$E9,$5E
	.dc.b	$A3,$83,$E5,$9F,$E1,$17,$D6,$7F,$8C,$3F,$67,$81,$EF,$7F,$32,$6C
	.dc.b	$FF,$1D,$E0,$C4,$FF,$34,$6C,$64,$FE,$49,$3C,$30,$FF,$77,$FD,$13
	.dc.b	$E3,$FF,$C9,$2D,$AF,$F2,$13,$67,$FC,$AE,$D7,$FC,$C4,$FF,$BB,$A3
	.dc.b	$F9,$83,$BC,$BF,$D4,$2E,$67,$BD,$7F,$4A,$87,$C6,$FF,$4C,$FA,$96
	.dc.b	$19,$8C,$34,$F1,$BE,$62,$3E,$6F,$8F,$D6,$1E,$C6,$E3,$E2,$F8,$DF
	.dc.b	$69,$2B,$DA,$DA,$F8,$6F,$97,$E3,$7C,$CF,$9F,$E3,$7D,$1B,$DB,$F7
	.dc.b	$B6,$FE,$2C,$FC,$A2,$4F,$F8,$4B,$E1,$3E,$C1,$F0,$7C,$8B,$1F,$21
	.dc.b	$FA,$24,$7E,$8B,$F5,$2A,$47,$D4,$69,$FE,$8A,$FA,$1F,$F2,$94,$7F
	.dc.b	$9E,$5B,$19,$E6,$FF,$42,$C6,$23,$FD,$34,$FF,$BF,$A7,$87,$C0,$E8
	.dc.b	$8F,$3F,$7F,$89,$1C,$1D,$FE,$67,$A9,$1C,$1E,$3F,$77,$C2,$47,$04
	.dc.b	$C1,$C1,$FE,$E8,$C5,$5F,$E1,$05,$EF,$58,$C3,$EC,$25,$FC,$17,$DF
	.dc.b	$EE,$C8,$F8,$8D,$C3,$E4,$78,$3D,$63,$B3,$67,$FC,$14,$BD,$7F,$F1
	.dc.b	$0A,$FA,$0B,$18,$6C,$FF,$AA,$67,$83,$67,$FD,$6B,$18,$BF,$F6,$0A
	.dc.b	$F6,$32,$4A,$E4,$D9,$F8,$B1,$FE,$84,$7F,$CA,$71,$E4,$F6,$3C,$AF
	.dc.b	$F8,$DB,$E8,$1E,$7E,$6F,$FC,$49,$18,$7D,$B5,$FE,$3C,$FA,$0F,$F3
	.dc.b	$36,$1D,$8F,$F4,$8B,$8E,$0F,$DA,$C6,$3D,$1E,$1E,$5F,$1F,$92,$E5
	.dc.b	$D5,$72,$FF,$19,$62,$8F,$F2,$4F,$D0,$3E,$1A,$FC,$3E,$1B,$7F,$E3
	.dc.b	$9F,$A0,$6D,$FE,$50,$7E,$34,$66,$31,$C3,$9F,$26,$4F,$1F,$2A,$F0
	.dc.b	$7A,$C7,$D0,$3D,$7F,$F2,$DB,$6B,$FC,$90,$FC,$59,$E0,$E5,$E5,$F4
	.dc.b	$0E,$EF,$F3,$1B,$6B,$FC,$A0,$BB,$F3,$3A,$F8,$7D,$FC,$8D,$BD,$BF
	.dc.b	$E6,$CE,$1F,$7C,$7F,$96,$5F,$7E,$3E,$97,$2D,$DB,$FC,$E7,$E5,$7C
	.dc.b	$DA,$FF,$30,$3F,$CF,$7A,$7E,$BD,$7F,$2E,$E5,$DF,$8F,$FE,$7A,$FA
	.dc.b	$07,$CF,$D3,$1F,$41,$72,$FF,$35,$4F,$EC,$1B,$E2,$71,$F1,$79,$3F
	.dc.b	$E8,$4F,$9F,$FE,$77,$3F,$E7,$15,$F3,$B9,$DB,$4E,$FE,$82,$56,$FE
	.dc.b	$82,$EF,$F5,$FE,$83,$F8,$CF,$0F,$A9,$B5,$FA,$17,$D2,$FF,$9F,$DC
	.dc.b	$CF,$FB,$A3,$D2,$FE,$A1,$5F,$09,$B5,$FA,$97,$FA,$B3,$7D,$0F,$F4
	.dc.b	$EC,$3E,$7E,$F8,$D8,$6D,$7F,$A3,$67,$FD,$A9,$87,$E3,$57,$A3,$CA
	.dc.b	$39,$BF,$EA,$8F,$A0,$F8,$75,$FE,$95,$6C,$FE,$55,$F9,$2D,$DB,$85
	.dc.b	$F1,$DE,$DF,$17,$A7,$93,$8B,$E4,$DD,$7F,$03,$8B,$D3,$F1,$69,$E3
	.dc.b	$6E,$3E,$37,$0F,$97,$D4,$B8,$F9,$7E,$6F,$D4,$7A,$F6,$B5,$EB,$70
	.dc.b	$FF,$AC,$7C,$17,$C9,$62,$FF,$65,$A3,$E7,$7F,$D9,$6D,$BF,$F5,$C2
	.dc.b	$F0,$73,$3A,$64,$F5,$F7,$ED,$FE,$F4,$FD,$11,$7C,$0F,$F6,$BA,$F3
	.dc.b	$EC,$61,$95,$FE,$C7,$97,$CE,$61,$E8,$F2,$3F,$DC,$3A,$7F,$A6,$A3
	.dc.b	$73,$C7,$6E,$E7,$83,$6B,$1D,$5C,$3D,$E5,$F3,$E5,$F4,$1F,$E3,$E7
	.dc.b	$F9,$F5,$7A,$9D,$9F,$D0,$49,$35,$A5,$3A,$37,$E7,$DF,$36,$FA,$0C
	.dc.b	$36,$B9,$4E,$A7,$27,$1F,$0F,$FB,$AC,$E9,$AF,$8E,$7F,$BB,$F1,$EC
	.dc.b	$F0,$78,$7F,$D5,$DF,$E2,$71,$F8,$19,$FF,$7B,$35,$FE,$6C,$FF,$7E
	.dc.b	$73,$C9,$C7,$5F,$EC,$1B,$FF,$44,$9F,$C7,$7A,$39,$7E,$55,$F3,$C3
	.dc.b	$BF,$2E,$65,$2F,$2F,$0D,$8F,$1C,$E2,$73,$ED,$C9,$F3,$C5,$7F,$A6
	.dc.b	$62,$F9,$6E,$CD,$8F,$6D,$78,$7C,$BF,$C6,$C7,$DF,$CB,$F9,$AC,$39
	.dc.b	$74,$6F,$9A,$D7,$EA,$FE,$6F,$A9,$87,$FA,$E7,$CF,$FC,$76,$F0,$FB
	.dc.b	$E3,$E1,$70,$FF,$81,$DD,$65,$FD,$11,$CD,$7E,$47,$E8,$1E,$3F,$C7
	.dc.b	$6C,$79,$EF,$96,$1F,$2F,$E8,$19,$52,$56,$CF,$CE,$47,$F8,$A5,$C0
	.dc.b	$7F,$82,$B8,$6B,$FD,$DD,$B9,$7F,$5A,$7B,$C6,$F9,$3C,$44,$79,$7A
	.dc.b	$1E,$5F,$9D,$F3,$78,$12,$7B,$47,$A7,$93,$F0,$9F,$41,$F4,$67,$CA
	.dc.b	$D7,$C5,$1F,$89,$FB,$FF,$EB,$71,$EA,$7F,$D3,$2D,$F2,$75,$E5,$EF
	.dc.b	$C3,$E1,$CB,$2B,$84,$7E,$4F,$E1,$BF,$BA,$F0,$7F,$76,$5F,$2F,$93
	.dc.b	$5C,$9F,$99,$F8,$E7,$D0,$DF,$12,$65,$F3,$27,$2F,$CE,$FB,$EF,$A0
	.dc.b	$FF,$88,$5F,$92,$DE,$7F,$D2,$46,$B3,$FF,$03,$FD,$FF,$F9,$22,$BE
	.dc.b	$BF,$CE,$FE,$09,$FF,$15,$36,$77,$F0,$FF,$AD,$FB,$EF,$F1,$F5,$7D
	.dc.b	$E7,$F9,$83,$78,$35,$E6,$FE,$CF,$EF,$BF,$C8,$53,$F8,$47,$FC,$50
	.dc.b	$DF,$17,$5F,$0D,$FB,$9C,$FE,$7C,$F8,$B9,$B0,$67,$9F,$F7,$DE,$47
	.dc.b	$F9,$D6,$7F,$1E,$FB,$79,$7F,$CD,$0D,$8F,$E6,$0E,$DF,$F2,$32,$FA
	.dc.b	$CE,$7F,$0B,$27,$9C,$78,$FD,$5F,$27,$B3,$27,$E3,$C7,$F0,$FF,$BC
	.dc.b	$7A,$D7,$DC,$95,$FA,$29,$CB,$FA,$4F,$BC,$FF,$19,$B7,$BF,$F8,$C4
	.dc.b	$1B,$F0,$5F,$0F,$DF,$F8,$98,$4B,$FA,$F1,$C3,$D1,$E3,$F1,$B2,$7B
	.dc.b	$47,$F2,$3F,$BB,$79,$57,$C8,$95,$FB,$69,$CB,$FB,$8F,$C0,$FF,$D2
	.dc.b	$CF,$E1,$3B,$7E,$A1,$FC,$B6,$33,$1F,$E6,$FF,$9B,$FE,$1A,$BE,$7F
	.dc.b	$F9,$73,$FE,$73,$6F,$85,$FE,$56,$9D,$C3,$3F,$E7,$0B,$4A,$62,$3F
	.dc.b	$9E,$4F,$F8,$1E,$FF,$86,$B5,$F1,$74,$70,$7F,$88,$FD,$47,$E0,$53
	.dc.b	$49,$F2,$9F,$D1,$DD,$7D,$4F,$E0,$E7,$F8,$CB,$5F,$F4,$64,$71,$F4
	.dc.b	$78,$7A,$52,$7B,$C7,$F5,$1F,$A0,$76,$52,$4F,$92,$BC,$1F,$53,$23
	.dc.b	$CE,$F8,$F9,$3C,$9F,$61,$FD,$F6,$7D,$2D,$C8,$D3,$5B,$7D,$09,$FF
	.dc.b	$04,$35,$FF,$65,$35,$FF,$09,$35,$F8,$BE,$77,$52,$3F,$E0,$DA,$47
	.dc.b	$8F,$57,$83,$49,$5F,$E1,$38,$CB,$C7,$D1,$F4,$11,$C1,$BF,$A3,$E8
	.dc.b	$47,$07,$F8,$77,$EA,$3F,$8A,$3F,$E2,$36,$BF,$EE,$56,$BF,$E2,$D6
	.dc.b	$AE,$0F,$E2,$EF,$F8,$A5,$AF,$F8,$B5,$AF,$FB,$BD,$AF,$F8,$DF,$C8
	.dc.b	$FA,$49,$A4,$7E,$B4,$BE,$2A,$92,$3D,$4F,$DB,$FA,$8F,$F5,$BC,$7F
	.dc.b	$8F,$5A,$FF,$BF,$DA,$FF,$91,$DA,$BE,$C1,$F5,$9F,$85,$6C,$7F,$90
	.dc.b	$DA,$FF,$91,$DA,$FB,$FC,$8F,$C6,$65,$FE,$A1,$8F,$F2,$53,$5F,$8D
	.dc.b	$1F,$91,$FA,$86,$8F,$A8,$FF,$28,$46,$1B,$DD,$FE,$A4,$BF,$F2,$B4
	.dc.b	$E5,$FF,$2C,$7D,$87,$1F,$37,$FC,$B0,$D8,$CD,$FC,$08,$C4,$7E,$6B
	.dc.b	$BF,$E1,$F3,$FF,$07,$BF,$F3,$03,$67,$FC,$C8,$D7,$FC,$D1,$C8,$F9
	.dc.b	$BF,$96,$DA,$FC,$B4,$60,$F9,$FF,$4D,$FE,$C3,$8F,$F3,$63,$6B,$FC
	.dc.b	$E0,$D7,$FC,$E8,$D7,$FC,$F8,$1F,$E8,$71,$FE,$22,$6C,$3C,$3F,$F7
	.dc.b	$BF,$07,$FA,$DE,$B5,$E7,$1F,$F1,$4C,$66,$BF,$31,$F2,$3F,$84,$8C
	.dc.b	$47,$E6,$63,$F0,$FF,$50,$F0,$39,$4F,$7B,$F2,$CF,$F9,$65,$FF,$1A
	.dc.b	$BB,$6F,$81,$C3,$2E,$34,$AD,$3C,$CF,$F4,$D3,$CD,$C1,$BB,$E5,$7E
	.dc.b	$0B,$B3,$C2,$6B,$D6,$2D,$76,$BF,$0B,$D5,$B5,$A9,$DB,$6F,$D6,$DB
	.dc.b	$F1,$DF,$39,$F1,$1A,$3D,$E5,$8D,$D7,$48,$CC,$72,$AE,$97,$E2,$9C
	.dc.b	$B6,$78,$C7,$38,$F4,$D7,$39,$C9,$CE,$78,$9F,$52,$BA,$B4,$CC,$7C
	.dc.b	$76,$BF,$35,$B7,$ED,$8F,$A7,$5D,$EF,$C3,$5E,$33,$E0,$CF,$84,$F1
	.dc.b	$C7,$A4,$DB,$96,$99,$8E,$2D,$8E,$F1,$D5,$A7,$17,$67,$AD,$EE,$D8
	.dc.b	$EA,$D3,$71,$EF,$36,$58,$DB,$63,$6F,$C3,$36,$58,$DB,$63,$75,$F1
	.dc.b	$E3,$F2,$11,$E9,$69,$E9,$3D,$73,$EF,$72,$D9,$FC,$A4,$FA,$67,$6F
	.dc.b	$BD,$FF,$7C,$47,$EA,$1B,$7F,$B6,$24,$FF,$BE,$67,$53,$C4,$87,$BC
	.dc.b	$83,$EC,$1F,$47,$F7,$EB,$C5,$3E,$F0,$6E,$34,$F2,$AF,$85,$5C,$AB
	.dc.b	$E1,$47,$C5,$33,$5C,$2B,$A1,$2B,$83,$F1,$CC,$CF,$0A,$E7,$7C,$6B
	.dc.b	$78,$7C,$98,$C6,$1D,$6F,$C1,$1F,$C1,$2B,$77,$EC,$9C,$BF,$EC,$C8
	.dc.b	$CE,$1E,$BA,$E7,$3E,$03,$EE,$4E,$A7,$CB,$3A,$7D,$15,$D1,$FE,$33
	.dc.b	$3B,$8C,$61,$F8,$18,$C5,$79,$CF,$36,$17,$0E,$ED,$C3,$E2,$57,$09
	.dc.b	$F2,$47,$69,$D5,$F5,$6C,$FD,$AA,$E1,$5E,$6B,$EF,$3C,$AF,$83,$4E
	.dc.b	$17,$F5,$F5,$79,$BF,$CB,$37,$FE,$5F,$18,$AF,$E8,$31,$89,$D4,$7F
	.dc.b	$57,$6E,$3F,$CA,$BB,$FF,$A6,$92,$7F,$30,$D8,$D4,$7F,$C0,$0F,$A6
	.dc.b	$79,$1F,$CF,$2B,$F9,$FF,$D3,$BE,$2A,$98,$F7,$AE,$1F,$98,$DB,$FA
	.dc.b	$8D,$6E,$30,$7F,$6E,$71,$97,$F5,$05,$5C,$A3,$0F,$F8,$AA,$33,$3F
	.dc.b	$DD,$A7,$94,$E4,$D7,$CC,$79,$BD,$9F,$3B,$6B,$C3,$5F,$AC,$B4,$F0
	.dc.b	$9E,$93,$98,$C3,$89,$FF,$00,$BE,$16,$BE,$BE,$7F,$31,$FE,$40,$E3
	.dc.b	$FF,$6A,$5F,$82,$55,$FE,$0A,$CB,$FB,$AD,$6A,$7E,$15,$FC,$EF,$7B
	.dc.b	$E0,$13,$0F,$2C,$F2,$8C,$3F,$25,$FA,$BE,$97,$19,$FD,$77,$95,$86
	.dc.b	$9E,$3C,$93,$F7,$2F,$EE,$FC,$AD,$AD,$36,$3E,$B1,$FB,$DF,$0F,$1A
	.dc.b	$7F,$83,$BE,$9B,$FD,$C1,$7E,$F6,$D7,$F8,$31,$7D,$3F,$F8,$21,$CD
	.dc.b	$7C,$2F,$A9,$C4,$FF,$09,$6D,$F3,$FA,$1E,$1F,$05,$C5,$B1,$92,$63
	.dc.b	$DF,$EC,$35,$1E,$36,$C6,$63,$78,$FA,$7E,$5F,$C9,$F3,$1E,$56,$CF
	.dc.b	$AD,$C3,$F6,$FE,$A5,$F6,$18,$C3,$DD,$E4,$FB,$4F,$77,$B9,$6F,$E2
	.dc.b	$36,$7E,$23,$72,$FF,$1B,$BF,$C7,$CF,$D8,$CF,$E2,$5C,$77,$7F,$8C
	.dc.b	$1D,$78,$77,$E3,$30,$DA,$E1,$1F,$E3,$6C,$FF,$5F,$87,$68,$FF,$1C
	.dc.b	$F2,$BE,$9E,$98,$F8,$8B,$87,$E6,$65,$D7,$C0,$E3,$D3,$67,$F3,$DF
	.dc.b	$53,$F7,$9B,$D1,$FE,$46,$8F,$BF,$1F,$E6,$E7,$5C,$1F,$E5,$0F,$BB
	.dc.b	$73,$7F,$93,$1B,$5F,$A9,$5F,$23,$99,$FF,$28,$FC,$FF,$C6,$BF,$F2
	.dc.b	$CB,$78,$3F,$CC,$2F,$F4,$93,$18,$7F,$9F,$BE,$A7,$E1,$B7,$C6,$FF
	.dc.b	$2D,$C3,$EB,$2C,$7F,$97,$BE,$EB,$EF,$D7,$EE,$DB,$8F,$F9,$A1,$E1
	.dc.b	$19,$EA,$E1,$E2,$48,$E1,$F1,$56,$FF,$CD,$FF,$54,$FA,$EF,$F3,$5B
	.dc.b	$6B,$F8,$42,$FA,$6C,$CF,$F9,$C2,$53,$4C,$7C,$9D,$8F,$E2,$07,$FA
	.dc.b	$B7,$83,$FD,$06,$D3,$FD,$3E,$3B,$BA,$78,$1F,$E7,$CF,$AA,$7C,$CD
	.dc.b	$31,$CD,$6F,$F8,$E7,$DC,$FF,$A7,$71,$FF,$47,$B6,$3F,$D5,$6F,$F1
	.dc.b	$F6,$FB,$3F,$E8,$8F,$B9,$7D,$6D,$31,$F2,$96,$BF,$94,$7D,$66,$9B
	.dc.b	$3F,$E9,$08,$7D,$AD,$8E,$9F,$53,$2B,$F9,$86,$1E,$FE,$5F,$F1,$5A
	.dc.b	$9F,$F4,$CF,$C2,$3E,$8E,$98,$F9,$8B,$87,$F3,$9F,$61,$F5,$1F,$EA
	.dc.b	$06,$C6,$7E,$8B,$13,$FE,$A4,$FA,$D7,$99,$4C,$7B,$96,$FF,$A2,$4F
	.dc.b	$F5,$26,$C7,$1C,$37,$1A,$F9,$0C,$34,$E3,$1D,$F9,$3C,$1F,$1B,$67
	.dc.b	$AB,$78,$7F,$D7,$B8,$FD,$57,$FC,$9A,$FD,$56,$D7,$A7,$39,$7E,$6F
	.dc.b	$3E,$D7,$27,$DC,$7F,$B0,$BD,$8F,$CB,$4E,$CF,$EC,$1D,$2F,$A7,$FF
	.dc.b	$63,$B6,$33,$F3,$58,$9C,$B2,$3F,$B4,$53,$8B,$18,$FF,$B5,$8F,$E2
	.dc.b	$73,$FD,$B5,$A6,$7D,$1F,$F6,$AB,$8C,$7F,$DC,$6F,$F6,$09,$E1,$3A
	.dc.b	$E3,$DB,$63,$A4,$7B,$7E,$57,$15,$E1,$D7,$A9,$DB,$5B,$7E,$37,$0F
	.dc.b	$96,$5E,$77,$F9,$F3,$EB,$5F,$57,$FE,$1B,$BF,$3D,$68,$FE,$FE,$53
	.dc.b	$FC,$03,$58,$70,$7F,$A1,$87,$D6,$7E,$E3,$EB,$7F,$C0,$ED,$8F,$F0
	.dc.b	$EB,$5F,$F1,$53,$5E,$4B,$DE,$6F,$C9,$B3,$C9,$AF,$26,$BC,$BD,$EF
	.dc.b	$24,$B4,$EB,$C9,$86,$D7,$F8,$97,$05,$E6,$3F,$D4,$31,$F8,$97,$F0
	.dc.b	$7F,$39,$95,$C8,$8C,$4E,$E7,$D3,$7E,$CE,$5C,$F7,$C9,$F0,$45,$76
	.dc.b	$78,$3C,$9E,$66,$74,$F5,$46,$4F,$CA,$FC,$7F,$8F,$33,$C2,$3F,$C8
	.dc.b	$1C,$4E,$6F,$43,$FC,$2A,$33,$CB,$89,$CB,$CF,$DD,$93,$A3,$FC,$95
	.dc.b	$F5,$BF,$19,$BC,$5F,$0E,$3E,$06,$7E,$69,$46,$DF,$69,$89,$F9,$5F
	.dc.b	$5B,$FE,$7B,$8F,$F2,$B3,$5D,$CF,$4E,$0F,$C7,$7D,$6F,$FB,$62,$7F
	.dc.b	$CB,$8B,$1F,$A8,$D8,$CF,$C1,$63,$D4,$91,$C1,$DD,$70,$E5,$93,$D1
	.dc.b	$E7,$F4,$3D,$3C,$7F,$D1,$E3,$FC,$DE,$BF,$CC,$6D,$7F,$CF,$2D,$7F
	.dc.b	$D0,$CF,$D4,$77,$C5,$C8,$8F,$F9,$B3,$EB,$3F,$9B,$B7,$C8,$FD,$8C
	.dc.b	$62,$3F,$D0,$AD,$7F,$D2,$4F,$F9,$A9,$E5,$E0,$F0,$23,$FE,$78,$FA
	.dc.b	$5F,$E2,$2F,$ED,$DA,$FF,$BF,$9A,$FF,$9A,$0F,$F4,$9F,$C6,$7F,$88
	.dc.b	$1F,$F4,$0B,$5C,$FA,$5F,$C1,$A3,$11,$FE,$A5,$FA,$CF,$AC,$FF,$A7
	.dc.b	$1A,$E7,$87,$F8,$6C,$62,$3F,$85,$7D,$67,$FA,$C9,$FF,$47,$B5,$FF
	.dc.b	$52,$B5,$FF,$59,$35,$FE,$25,$F5,$9F,$E4,$67,$FD,$2E,$D7,$FD,$5A
	.dc.b	$D7,$FD,$74,$D7,$F8,$D7,$D6,$3D,$AF,$E4,$2D,$8F,$F5,$0B,$5F,$F5
	.dc.b	$BB,$5F,$F6,$27,$D2,$3D,$1F,$E5,$2D,$8F,$F5,$4B,$5F,$F5,$FB,$5F
	.dc.b	$F6,$67,$85,$FE,$E0,$7F,$D5,$ED,$7F,$D8,$AD,$77,$F1,$9F,$CD,$A1
	.dc.b	$F5,$94,$93,$C9,$F1,$FA,$32,$3C,$6D,$7F,$DA,$3F,$58,$F8,$2F,$E8
	.dc.b	$2D,$8F,$F6,$0B,$5D,$C7,$F4,$6F,$AC,$7B,$94,$93,$FE,$C8,$8C,$BF
	.dc.b	$0B,$E6,$7E,$29,$1C,$1B,$F2,$BA,$11,$C1,$FE,$D2,$FA,$CF,$EE,$8D
	.dc.b	$F5,$7F,$AB,$C6,$23,$FD,$D6,$D7,$FD,$F1,$F5,$8F,$07,$FA,$EB,$63
	.dc.b	$FD,$BA,$D7,$FD,$E6,$D7,$FE,$00,$7D,$17,$F3,$D6,$F9,$3F,$D9,$A3
	.dc.b	$11,$FD,$8D,$A8,$7D,$2B,$2F,$ED,$AD,$8F,$F7,$6B,$5F,$ED,$87,$17
	.dc.b	$0E,$1F,$F3,$43,$C3,$E1,$79,$91,$FF,$7B,$FD,$57,$E1,$DF,$EE,$CD
	.dc.b	$7C,$8D,$4D,$3F,$68,$EB,$B7,$FC,$07,$18,$8D,$C6,$33,$91,$E0,$6B
	.dc.b	$EA,$3A,$F8,$FC,$88,$F9,$9E,$BE,$2C,$8F,$4B,$5E,$56,$FA,$1F,$19
	.dc.b	$23,$DE,$FF,$86,$3E,$E1,$E2,$52,$4F,$8C,$46,$9F,$8F,$F4,$3E,$E2
	.dc.b	$38,$38,$7D,$AF,$FB,$C1,$BF,$1D,$FE,$26,$8C,$47,$BD,$AF,$DA,$7F
	.dc.b	$DB,$CF,$2F,$98,$E0,$47,$EB,$FA,$1F,$38,$8E,$0F,$E1,$3E,$F7,$F0
	.dc.b	$C2,$38,$3D,$9F,$59,$FC,$8D,$FC,$83,$53,$4C,$7D,$37,$F2,$18,$FC
	.dc.b	$B3,$53,$4C,$7D,$AF,$C7,$8F,$B8,$D7,$F2,$ED,$7F,$C8,$8D,$7C,$7F
	.dc.b	$0F,$FE,$4C,$A3,$49,$1F,$56,$BF,$C2,$CF,$E2,$5A,$FE,$91,$AF,$F9
	.dc.b	$45,$AF,$EB,$FE,$AD,$F5,$9F,$E5,$A6,$C7,$E0,$5A,$FE,$55,$AF,$EC
	.dc.b	$3E,$D5,$F2,$BF,$E6,$26,$C7,$E2,$5A,$FE,$75,$AF,$EE,$3E,$B5,$F4
	.dc.b	$8F,$8E,$FC,$84,$61,$BD,$AD,$F3,$5B,$BF,$D4,$77,$FF,$95,$38,$7F
	.dc.b	$30,$D8,$CF,$C3,$7F,$7A,$8C,$47,$F0,$8F,$A3,$F0,$57,$E7,$DB,$1F
	.dc.b	$B3,$6B,$BC,$BF,$C0,$7E,$F3,$E8,$29,$2F,$F5,$11,$97,$F5,$5F,$56
	.dc.b	$F9,$AF,$F2,$AF,$C2,$7F,$90,$63,$13,$FC,$86,$31,$E1,$EA,$3F,$D1
	.dc.b	$13,$AC,$BF,$C8,$AF,$F3,$58,$EB,$3F,$E1,$68,$C4,$FF,$0B,$6B,$FC
	.dc.b	$99,$FF,$30,$CF,$73,$AF,$95,$DE,$7F,$99,$4E,$9E,$F1,$95,$19,$BF
	.dc.b	$54,$E6,$31,$5D,$0E,$D5,$D3,$C3,$D5,$6D,$B3,$A9,$EB,$18,$7D,$ED
	.dc.b	$BF,$90,$D9,$F8,$AD,$9E,$D1,$CE,$FF,$DA,$8D,$9D,$4F,$68,$D5,$F8
	.dc.b	$DA,$6C,$B1,$E2,$6D,$F8,$9A,$78,$8B,$1E,$23,$E4,$BC,$DA,$73,$2C
	.dc.b	$73,$6C,$73,$69,$CE,$3E,$7C,$73,$6C,$73,$69,$CC,$B1,$CD,$B1,$CE
	.dc.b	$3E,$B3,$C5,$B1,$C5,$A7,$12,$C7,$13,$F0,$F3,$F2,$5F,$39,$F3,$4B
	.dc.b	$7F,$84,$77,$3E,$93,$9B,$BA,$F2,$B6,$33,$3C,$63,$F2,$35,$DE,$B7
	.dc.b	$39,$3E,$FC,$FA,$CE,$E5,$38,$D7,$85,$CC,$FB,$DB,$1E,$F6,$9E,$F3
	.dc.b	$89,$F6,$4F,$D6,$5F,$EC,$49,$1F,$C0,$1B,$1F,$B3,$6B,$C2,$3C,$4F
	.dc.b	$EB,$3C,$2F,$64,$78,$63,$E1,$1F,$B4,$8C,$BE,$3B,$F8,$11,$88,$D9
	.dc.b	$23,$FC,$05,$3D,$9F,$8D,$5F,$B7,$AE,$11,$89,$FD,$44,$62,$72,$D8
	.dc.b	$E4,$DA,$F9,$95,$D9,$B5,$C2,$BC,$0D,$8C,$C7,$23,$25,$72,$74,$AE
	.dc.b	$D7,$F5,$63,$E3,$4F,$07,$4D,$8D,$5F,$8D,$B7,$E1,$6C,$F7,$6D,$72
	.dc.b	$3C,$2F,$2B,$F0,$CF,$D3,$BE,$8D,$BE,$8D,$3A,$1B,$F8,$FF,$E4,$87
	.dc.b	$F2,$5F,$06,$C7,$F1,$56,$E3,$FC,$F1,$B8,$FF,$45,$E0,$FE,$78,$4A
	.dc.b	$FE,$96,$FF,$3F,$FA,$9F,$E6,$0F,$F3,$16,$C7,$F2,$BF,$A7,$7C,$F7
	.dc.b	$F3,$95,$F3,$F6,$30,$D9,$FF,$20,$4F,$06,$CF,$FB,$8A,$31,$97,$F9
	.dc.b	$5D,$B8,$AB,$FC,$F5,$29,$CC,$61,$7D,$4F,$79,$57,$17,$FD,$05,$3F
	.dc.b	$EC,$5A,$F1,$FB,$7E,$12,$3E,$6F,$97,$F6,$47,$B5,$B1,$F0,$7E,$5B
	.dc.b	$E0,$1F,$07,$D7,$B1,$F1,$5F,$61,$1F,$82,$FB,$FE,$97,$E4,$7C,$87
	.dc.b	$C9,$6C,$FC,$D7,$4F,$C8,$6E,$5F,$17,$E1,$FE,$41,$27,$E6,$BF,$2F
	.dc.b	$C0,$91,$F4,$1B,$3F,$34,$FA,$04,$8F,$AA,$FC,$73,$FD,$43,$F5,$1F
	.dc.b	$E7,$A8,$FF,$3C,$36,$3F,$D1,$8D,$77,$E8,$71,$7E,$BC,$7F,$85,$DD
	.dc.b	$F7,$65,$FF,$5A,$46,$5E,$F1,$95,$0F,$92,$C3,$6F,$C7,$18,$33,$F5
	.dc.b	$12,$3C,$ED,$AF,$3B,$FE,$0F,$A3,$7E,$97,$0E,$7E,$5E,$5F,$F8,$26
	.dc.b	$BF,$C8,$1C,$9F,$15,$4F,$3F,$17,$CA,$4A,$F8,$EF,$A3,$D9,$91,$F2
	.dc.b	$9B,$97,$CA,$F6,$7C,$91,$E5,$6C,$FF,$86,$57,$B3,$FE,$27,$5F,$09
	.dc.b	$63,$0D,$9F,$F5,$74,$F0,$6C,$FC,$D8,$C5,$FF,$B1,$57,$C8,$E4,$95
	.dc.b	$C9,$FA,$9C,$4F,$5A,$92,$7F,$D7,$91,$96,$F2,$F0,$23,$83,$FD,$85
	.dc.b	$F4,$FF,$81,$C7,$F7,$AD,$7F,$DF,$0D,$4D,$FA,$77,$1F,$E1,$DB,$5F
	.dc.b	$EC,$D6,$D7,$FB,$85,$AF,$FB,$F3,$E6,$FE,$4C,$FF,$B4,$DB,$1F,$EE
	.dc.b	$76,$BB,$DB,$F2,$9F,$4F,$FC,$5A,$3F,$89,$87,$0F,$F1,$66,$A3,$5E
	.dc.b	$0F,$E7,$7E,$9D,$EE,$D2,$63,$FE,$E5,$8C,$B7,$BB,$81,$1C,$1F,$EE
	.dc.b	$BF,$A7,$FF,$3D,$61,$E4,$6B,$FC,$A1,$AF,$8D,$AF,$F2,$D7,$F6,$10
	.dc.b	$69,$1F,$F7,$C0,$E5,$D9,$1C,$BF,$B9,$FA,$77,$D3,$D2,$4F,$FC,$01
	.dc.b	$8C,$BF,$2A,$FE,$37,$D3,$BC,$3F,$E0,$4D,$9D,$4A,$30,$F1,$6B,$F5
	.dc.b	$3E,$91,$F4,$BF,$C2,$DB,$5C,$9B,$3E,$16,$BF,$73,$E9,$EB,$FD,$41
	.dc.b	$AF,$DB,$6A,$6C,$3E,$8B,$C0,$4A,$F0,$3F,$C7,$7D,$87,$CF,$52,$4F
	.dc.b	$BA,$32,$FC,$0F,$29,$F4,$3F,$C9,$5B,$19,$8C,$1F,$CA,$7E,$99,$F3
	.dc.b	$F4,$93,$F2,$A3,$2D,$F3,$BF,$D6,$51,$FD,$97,$5F,$ED,$33,$FD,$C2
	.dc.b	$18,$6A,$B7,$F0,$DC,$22,$D7,$FA,$BE,$B6,$DE,$1E,$BF,$29,$D9,$B1
	.dc.b	$92,$63,$E0,$3F,$51,$4E,$7E,$13,$D7,$4F,$39,$E8,$F0,$AC,$78,$5C
	.dc.b	$3F,$EC,$1F,$31,$EF,$63,$1E,$C7,$F8,$1B,$E3,$BF,$DA,$5E,$69,$BF
	.dc.b	$2D,$86,$E1,$FE,$CB,$94,$FF,$B6,$70,$FF,$42,$C7,$F6,$93,$33,$FE
	.dc.b	$D1,$B7,$8F,$FE,$DF,$6C,$67,$D2,$C4,$E5,$91,$C7,$E8,$ED,$7A,$03
	.dc.b	$CC,$F1,$7B,$76,$7C,$1B,$77,$3C,$1B,$58,$F2,$38,$7C,$D6,$F1,$FF
	.dc.b	$8C,$7D,$33,$D6,$FE,$3F,$7E,$9A,$D1,$FB,$42,$9F,$CA,$CC,$57,$EE
	.dc.b	$3E,$97,$F8,$1F,$D3,$5F,$10,$D2,$57,$F0,$39,$FE,$DB,$C1,$C7,$D4
	.dc.b	$FF,$1D,$CE,$5A,$FF,$32,$9D,$BC,$7D,$4F,$F2,$3D,$7F,$86,$E7,$EE
	.dc.b	$4F,$F3,$38,$C7,$2E,$B6,$FF,$26,$D2,$F1,$7F,$CE,$63,$8B,$F8,$8E
	.dc.b	$47,$C9,$EA,$3F,$C2,$CD,$9D,$E3,$E1,$E6,$C7,$85,$2F,$C5,$97,$82
	.dc.b	$71,$1F,$94,$BF,$E5,$CE,$8F,$17,$91,$AC,$3D,$8F,$87,$27,$C0,$53
	.dc.b	$C5,$1E,$5E,$3F,$C3,$DF,$33,$EE,$72,$E4,$D7,$3D,$8F,$D4,$7A,$9D
	.dc.b	$4C,$3F,$A8,$FA,$5F,$8C,$F3,$7D,$FB,$7F,$9D,$9B,$1F,$D8,$5D,$BC
	.dc.b	$78,$BE,$47,$D3,$7F,$A2,$A3,$C8,$D7,$47,$9F,$E9,$BF,$C4,$2D,$C3
	.dc.b	$E8,$47,$F8,$62,$35,$EC,$E2,$73,$DF,$27,$81,$CF,$E1,$F8,$F3,$48
	.dc.b	$C1,$DF,$CA,$EA,$47,$C5,$1E,$0E,$29,$1E,$57,$C7,$F0,$FF,$83,$FA
	.dc.b	$6F,$E6,$33,$F3,$DA,$FF,$10,$6F,$37,$F9,$BD,$FF,$16,$46,$FC,$2E
	.dc.b	$44,$79,$78,$7F,$8F,$F8,$0F,$E1,$DC,$9F,$EA,$46,$CF,$F8,$3D,$A9
	.dc.b	$F9,$2F,$F2,$0F,$A5,$7D,$37,$E6,$5B,$1F,$85,$6B,$BF,$73,$F3,$DF
	.dc.b	$4D,$CB,$4F,$F4,$33,$71,$FF,$5A,$B6,$BF,$C4,$0F,$F9,$59,$FC,$87
	.dc.b	$D3,$7F,$AD,$9B,$F0,$DF,$A9,$8C,$47,$F3,$4C,$FF,$9B,$BF,$97,$FA
	.dc.b	$6F,$E5,$2D,$D7,$F6,$11,$88,$FE,$76,$FF,$17,$7F,$3D,$F4,$AF,$A6
	.dc.b	$FD,$B3,$63,$31,$8F,$73,$F7,$5F,$4C,$F5,$A9,$2B,$F5,$19,$3A,$7F
	.dc.b	$80,$78,$FE,$D9,$FE,$96,$DE,$1F,$F5,$53,$78,$7F,$DB,$3A,$3D,$8E
	.dc.b	$3F,$4D,$30,$FD,$A5,$A3,$AF,$D7,$F9,$3D,$79,$5F,$B9,$D9,$1F,$83
	.dc.b	$EE,$FE,$25,$F4,$FF,$D9,$EB,$F7,$CD,$7F,$AF,$B7,$5F,$F5,$F3,$FE
	.dc.b	$2F,$8F,$E0,$9F,$4C,$FA,$7F,$E3,$8D,$8C,$C6,$3E,$03,$E3,$FD,$3C
	.dc.b	$9F,$23,$F9,$5F,$0B,$8F,$9D,$27,$AB,$D3,$D1,$FE,$51,$F4,$EF,$16
	.dc.b	$92,$BF,$8B,$47,$F8,$99,$FE,$31,$F4,$FF,$45,$BD,$1F,$CC,$63,$11
	.dc.b	$F8,$9E,$3F,$E6,$9F,$35,$FC,$4A,$0D,$27,$C8,$FF,$39,$F9,$CF,$E5
	.dc.b	$F3,$FE,$04,$6F,$1F,$FB,$8D,$BD,$7F,$E7,$B9,$47,$F8,$FD,$FE,$58
	.dc.b	$BE,$47,$A1,$23,$3B,$FF,$47,$FA,$7F,$C2,$47,$FA,$15,$A9,$BF,$5A
	.dc.b	$D3,$87,$A1,$2B,$F9,$D5,$7E,$BD,$E1,$F2,$1D,$78,$24,$69,$FE,$AD
	.dc.b	$F4,$8F,$88,$F1,$7A,$32,$B9,$BC,$1F,$F1,$2E,$1F,$D3,$1E,$FC,$FE
	.dc.b	$93,$FB,$07,$A4,$FA,$9F,$67,$CF,$4A,$F2,$BB,$7F,$A8,$7D,$4F,$F1
	.dc.b	$36,$F9,$1F,$DA,$63,$11,$E0,$E0,$F8,$DF,$3D,$23,$DE,$FA,$BE,$93
	.dc.b	$FB,$87,$D4,$FF,$AA,$A7,$FA,$E3,$5F,$F1,$EB,$7A,$5D,$3F,$67,$77
	.dc.b	$AB,$CB,$E6,$A6,$1C,$9F,$8F,$E1,$7D,$7B,$93,$F4,$1F,$B5,$F3,$FF
	.dc.b	$C2,$3E,$2A,$4F,$ED,$5F,$9F,$F0,$7F,$E0,$5F,$A8,$FE,$3B,$A7,$8A
	.dc.b	$BF,$D8,$91,$8C,$3F,$18,$D8,$CF,$CC,$F1,$CF,$FC,$5D,$3F,$D1,$A7
	.dc.b	$FD,$50,$D7,$FD,$A6,$D7,$F2,$DE,$1F,$FB,$99,$FF,$54,$B5,$FF,$58
	.dc.b	$35,$FF,$6D,$B5,$F1,$FC,$97,$80,$8F,$FA,$CA,$7F,$D0,$6F,$4F,$77
	.dc.b	$FC,$85,$5F,$D2,$23,$9C,$60,$F4,$F1,$48,$E4,$D7,$C7,$F1,$9C,$88
	.dc.b	$FA,$1E,$BF,$17,$23,$D6,$D7,$FD,$95,$F4,$AF,$A5,$FE,$5D,$18,$6B
	.dc.b	$FE,$F8,$8F,$DC,$F0,$FF,$B3,$1E,$1E,$BF,$FB,$3B,$EB,$5F,$61,$F5
	.dc.b	$EF,$FD,$FF,$18,$E2,$D4,$E6,$7F,$82,$4E,$5F,$F6,$E7,$D2,$FF,$3F
	.dc.b	$AF,$F6,$CB,$63,$37,$C5,$08,$D3,$FC,$3C,$7D,$33,$E9,$FF,$15,$CD
	.dc.b	$E5,$E2,$FE,$21,$8F,$F8,$F6,$B4,$FF,$8F,$E3,$37,$FE,$A0,$9F,$F8
	.dc.b	$02,$4B,$F1,$47,$66,$C7,$FA,$72,$BF,$C1,$DD,$BE,$07,$F9,$14,$66
	.dc.b	$BF,$C2,$13,$97,$D7,$19,$7F,$C2,$51,$97,$D7,$E9,$5F,$05,$39,$AE
	.dc.b	$11,$8B,$F6,$56,$EB,$4D,$8C,$9F,$DD,$A3,$2D,$33,$7F,$E0,$D8,$CB
	.dc.b	$63,$39,$7D,$FB,$F0,$4F,$C7,$BE,$CD,$AE,$8D,$39,$16,$B9,$5F,$66
	.dc.b	$CF,$63,$E5,$96,$7E,$4B,$5F,$8E,$DB,$F1,$CF,$E4,$F0,$F8,$B7,$D6
	.dc.b	$B9,$DF,$85,$B5,$BC,$3D,$2D,$8F,$4B,$67,$D2,$D3,$CE,$F3,$3C,$A5
	.dc.b	$3C,$A5,$C3,$CB,$3F,$84,$AE,$ED,$3B,$96,$7B,$B6,$BB,$9E,$17,$D2
	.dc.b	$64,$B1,$F8,$C9,$EA,$D8,$EA,$D3,$A9,$67,$AC,$73,$8F,$5C,$77,$6C
	.dc.b	$65,$F6,$C6,$5B,$19,$AF,$7C,$75,$8F,$2B,$F9,$C7,$C6,$D3,$F7,$45
	.dc.b	$3F,$79,$87,$F8,$86,$33,$19,$3E,$31,$AB,$FD,$99,$23,$73,$FA,$D2
	.dc.b	$78,$BF,$83,$AF,$21,$0F,$E1,$2E,$5F,$26,$1E,$D9,$FF,$64,$D7,$E4
	.dc.b	$5E,$97,$FB,$3B,$E1,$18,$9F,$E1,$D1,$89,$CB,$63,$93,$6F,$F1,$15
	.dc.b	$D9,$B5,$C2,$BC,$0D,$8C,$C7,$23,$25,$72,$74,$AE,$D7,$EB,$6C,$FD
	.dc.b	$BB,$F0,$4F,$BE,$FB,$36,$7A,$36,$FA,$47,$81,$A7,$82,$3B,$46,$5B
	.dc.b	$19,$8F,$05,$F8,$A3,$E9,$5F,$8A,$7C,$97,$E2,$7C,$DF,$2D,$FC,$C0
	.dc.b	$8F,$DE,$6D,$7F,$39,$7F,$93,$65,$FC,$9D,$FE,$59,$BB,$B2,$9F,$CA
	.dc.b	$F7,$B8,$7F,$45,$7F,$9E,$91,$FC,$5B,$FD,$0B,$E9,$9F,$4F,$49,$3F
	.dc.b	$96,$8C,$B7,$BB,$F9,$E4,$63,$36,$FF,$DA,$61,$3A,$BD,$B2,$FF,$B0
	.dc.b	$D6,$FE,$1F,$18,$72,$9F,$F3,$86,$1F,$EB,$57,$67,$F9,$CF,$EB,$BF
	.dc.b	$C7,$AE,$27,$FB,$5C,$ED,$E4,$7F,$6E,$9D,$B8,$11,$BB,$F4,$7C,$1F
	.dc.b	$B3,$C0,$F6,$BE,$EF,$9C,$98,$F7,$C3,$8F,$CD,$E1,$E9,$8A,$31,$8F
	.dc.b	$C3,$6D,$7B,$37,$F9,$27,$CA,$F8,$9B,$1D,$6D,$18,$C3,$AB,$70,$F9
	.dc.b	$DF,$1B,$C6,$F8,$36,$BC,$7D,$DF,$86,$E0,$FA,$ED,$AE,$D3,$86,$9E
	.dc.b	$18,$FE,$F1,$F4,$EE,$AF,$C7,$F1,$61,$B5,$F0,$E5,$F5,$19,$8C,$37
	.dc.b	$0F,$A1,$2F,$98,$FC,$CF,$91,$D9,$B7,$CF,$DD,$C3,$4E,$D1,$E4,$89
	.dc.b	$1E,$1F,$90,$B8,$76,$F1,$BD,$27,$AB,$E6,$AD,$79,$DB,$3E,$46,$F0
	.dc.b	$FF,$90,$73,$7B,$BF,$AE,$F9,$FF,$86,$DC,$3E,$03,$87,$E1,$7B,$F7
	.dc.b	$1F,$C7,$2F,$77,$FC,$44,$D9,$FC,$52,$F7,$B3,$3A,$6C,$E9,$B5,$D3
	.dc.b	$E5,$35,$3C,$9B,$1F,$9D,$E5,$71,$7F,$8C,$57,$2D,$8C,$39,$E8,$D7
	.dc.b	$55,$C7,$F3,$D7,$FC,$35,$FD,$07,$D4,$7D,$E6,$FC,$3F,$F8,$F6,$31
	.dc.b	$97,$EB,$54,$7F,$90,$7E,$A1,$CD,$79,$F0,$DC,$3F,$59,$F3,$FF,$17
	.dc.b	$1F,$DB,$5F,$F9,$A7,$9B,$FC,$E9,$39,$7F,$CE,$D1,$9A,$FF,$7B,$7D
	.dc.b	$47,$F1,$07,$F6,$CB,$EA,$3F,$CA,$4E,$70,$FE,$04,$E6,$7F,$79,$F5
	.dc.b	$16,$7F,$83,$63,$FC,$23,$1F,$F4,$24,$E5,$F1,$D7,$F9,$76,$3F,$80
	.dc.b	$3C,$BC,$3C,$78,$79,$7F,$CC,$5F,$2B,$FE,$76,$C7,$F8,$92,$E6,$CF
	.dc.b	$A1,$A7,$FC,$D1,$82,$6F,$83,$86,$C7,$F1,$33,$F2,$E5,$F8,$CE,$3D
	.dc.b	$BC,$9F,$F3,$97,$D4,$3D,$8B,$C9,$86,$DF,$F1,$DF,$A8,$6D,$FE,$79
	.dc.b	$5F,$27,$63,$0F,$F9,$F7,$E9,$DF,$23,$7E,$23,$0D,$C3,$F9,$47,$D4
	.dc.b	$69,$FE,$55,$F6,$5F,$EA,$27,$33,$FE,$87,$8C,$72,$7F,$37,$C1,$BF
	.dc.b	$7F,$99,$F0,$37,$0F,$E6,$9E,$EB,$E5,$FF,$D2,$0B,$E3,$AC,$61,$FF
	.dc.b	$4A,$7A,$AF,$A5,$BE,$A6,$1B,$87,$F3,$FF,$A9,$79,$FF,$E9,$B7,$F5
	.dc.b	$B1,$98,$C7,$99,$9F,$5E,$4F,$1C,$6F,$07,$F5,$1C,$11,$CC,$E9,$F2
	.dc.b	$36,$BC,$1C,$DD,$CF,$07,$3D,$8F,$23,$87,$CD,$E0,$DC,$7B,$6B,$E7
	.dc.b	$3D,$1E,$5D,$AF,$63,$87,$FA,$E7,$AA,$F9,$1C,$5F,$F6,$7F,$19,$F4
	.dc.b	$9F,$EB,$B6,$DF,$F6,$35,$D1,$99,$D3,$27,$AF,$C0,$DB,$FE,$E4,$7F
	.dc.b	$5E,$8F,$ED,$8D,$33,$CB,$FE,$C9,$71,$87,$FB,$C6,$30,$67,$E9,$AD
	.dc.b	$72,$70,$FB,$3E,$06,$DF,$83,$6E,$87,$5D,$AD,$78,$9C,$3E,$49,$79
	.dc.b	$7F,$EF,$8F,$A9,$FF,$18,$47,$FB,$E8,$57,$9D,$D9,$FE,$7E,$29,$FE
	.dc.b	$80,$9C,$38,$39,$7D,$47,$ED,$7E,$A7,$FD,$0A,$D8,$FF,$45,$B5,$F0
	.dc.b	$07,$0F,$2F,$77,$FC,$66,$F0,$6B,$C2,$11,$88,$E1,$B7,$83,$DD,$F6
	.dc.b	$2C,$FC,$BC,$58,$6D,$7F,$88,$F0,$5E,$60,$79,$6B,$0F,$F8,$07,$BA
	.dc.b	$F2,$FE,$D6,$39,$4E,$AF,$F1,$71,$D7,$C1,$C1,$9E,$5B,$1E,$1F,$2F
	.dc.b	$81,$87,$FC,$7D,$5F,$DE,$9F,$CE,$FC,$3F,$80,$DB,$FE,$3B,$F6,$1C
	.dc.b	$DE,$67,$F8,$24,$67,$97,$13,$97,$D3,$E8,$49,$D1,$FE,$45,$FA,$9F
	.dc.b	$86,$DD,$7E,$0C,$7E,$AA,$35,$C1,$89,$F8,$BF,$53,$FE,$68,$9F,$F2
	.dc.b	$73,$5F,$F2,$C5,$71,$7F,$CB,$5F,$53,$FE,$09,$7F,$CA,$8F,$F4,$B6
	.dc.b	$E1,$F4,$23,$51,$C2,$31,$DB,$23,$67,$2F,$9B,$E1,$C9,$23,$07,$6F
	.dc.b	$23,$A1,$1F,$0C,$77,$E0,$91,$E4,$7C,$5D,$FF,$82,$F7,$BF,$53,$3F
	.dc.b	$E7,$35,$F2,$99,$35,$C7,$F8,$8F,$97,$79,$34,$97,$FE,$73,$8F,$E6
	.dc.b	$AD,$F1,$F8,$11,$D7,$91,$E1,$F1,$64,$7F,$9E,$F2,$47,$97,$E0,$3C
	.dc.b	$1E,$3F,$87,$C8,$97,$DE,$3C,$9A,$F9,$7E,$4A,$4F,$8E,$3D,$DD,$72
	.dc.b	$3B,$3F,$07,$B3,$D8,$7B,$72,$93,$EB,$8F,$81,$E3,$FC,$1A,$93,$EF
	.dc.b	$8F,$94,$FC,$B6,$3E,$B3,$F5,$FF,$71,$FE,$32,$7F,$D3,$0D,$7F,$D4
	.dc.b	$4D,$F1,$7F,$7D,$1F,$B8,$8F,$EF,$6F,$0F,$6B,$99,$1E,$7F,$0F,$FB
	.dc.b	$CF,$90,$DC,$7F,$AB,$5B,$7D,$21,$18,$F2,$3F,$94,$2F,$5F,$CF,$F0
	.dc.b	$5E,$8F,$2A,$5F,$A2,$3D,$7F,$29,$EC,$F7,$24,$FB,$23,$C5,$9F,$8F
	.dc.b	$E1,$24,$FF,$AD,$F2,$FF,$4D,$3F,$0B,$C8,$F8,$A4,$7E,$2F,$93,$F2
	.dc.b	$3C,$D9,$1F,$21,$F9,$7E,$6F,$BF,$DF,$93,$D5,$F8,$5C,$5F,$C7,$BE
	.dc.b	$7B,$F8,$0C,$FF,$B3,$DB,$C7,$F0,$DB,$C7,$FD,$25,$FE,$67,$3E,$7F
	.dc.b	$A6,$F9,$84,$7F,$DA,$11,$C1,$FE,$5A,$FF,$1D,$7B,$7A,$1E,$13,$C5
	.dc.b	$9F,$8F,$E2,$A4,$F2,$74,$FF,$B6,$FE,$6B,$FD,$2C,$DF,$8B,$FE,$6F
	.dc.b	$18,$8F,$F7,$47,$CF,$BE,$73,$3F,$1B,$27,$93,$C2,$F8,$FA,$12,$7B
	.dc.b	$3F,$D0,$BE,$7D,$E5,$76,$F4,$25,$78,$DE,$0F,$FB,$DE,$5F,$31,$C4
	.dc.b	$91,$B3,$D5,$F3,$FE,$82,$3E,$DF,$4B,$FA,$8F,$CF,$FE,$0A,$1C,$DF
	.dc.b	$95,$8C,$37,$BB,$E0,$EF,$F0,$BE,$66,$5F,$C7,$7D,$5E,$5F,$F1,$5F
	.dc.b	$90,$93,$FC,$4D,$F9,$3F,$3F,$FC,$67,$E1,$64,$FF,$17,$7E,$17,$9D
	.dc.b	$F3,$BB,$64,$FD,$37,$E8,$FC,$FE,$23,$E8,$FC,$F7,$F6,$AE,$93,$9D
	.dc.b	$FF,$5D,$F3,$FF,$EB,$EA,$FE,$E3,$18,$F5,$7F,$B8,$CE,$67,$E0,$E3
	.dc.b	$FE,$19,$8F,$E9,$D5,$CB,$3F,$F0,$C4,$7F,$B2,$AB,$FC,$1E,$7F,$7C
	.dc.b	$70,$E1,$AF,$F8,$51,$FF,$2E,$BC,$03,$87,$87,$AD,$D8,$FF,$13,$79
	.dc.b	$0F,$36,$92,$79,$46,$5B,$E5,$F0,$23,$83,$C1,$6E,$2F,$59,$23,$2F
	.dc.b	$F8,$D7,$EB,$1F,$5B,$49,$3E,$98,$CB,$EB,$FA,$1F,$A3,$F1,$B7,$84
	.dc.b	$E7,$C8,$FF,$21,$D7,$C8,$AD,$BF,$2B,$4F,$F1,$3F,$0F,$B3,$CD,$FE
	.dc.b	$4D,$3F,$C9,$1A,$FF,$8A,$A7,$6F,$D2,$FA,$1F,$F7,$33,$FE,$42,$6B
	.dc.b	$FE,$53,$6B,$FE,$32,$6B,$FE,$7D,$F3,$3A,$B6,$FE,$B7,$C0,$FF,$BE
	.dc.b	$A3,$FC,$9A,$D7,$A7,$8D,$FD,$1B,$9B,$F9,$E6,$1F,$ED,$B6,$C6,$6F
	.dc.b	$C5,$F2,$D6,$FC,$53,$55,$F1,$AE,$A7,$E6,$0E,$25,$8E,$2D,$78,$D7
	.dc.b	$39,$55,$8C,$36,$B8,$4E,$9B,$3A,$69,$A9,$E1,$5F,$A6,$3C,$25,$3C
	.dc.b	$31,$B3,$8B,$9A,$EC,$DA,$EC,$D9,$E0,$D3,$85,$78,$2F,$E5,$CF,$46
	.dc.b	$CE,$DA,$7D,$2A,$EA,$D9,$EA,$D9,$EB,$1C,$CF,$9E,$F3,$6C,$6C,$EA
	.dc.b	$FD,$28,$E2,$6C,$F8,$71,$B8,$E2,$6D,$E3,$1F,$91,$9F,$2C,$E4,$DC
	.dc.b	$65,$F3,$99,$7A,$B5,$FC,$3D,$7C,$58,$E3,$1F,$48,$F2,$D7,$8D,$E2
	.dc.b	$75,$33,$1C,$EF,$EA,$CE,$E7,$AC,$FC,$57,$89,$D6,$3D,$6D,$8F,$0C
	.dc.b	$FE,$66,$36,$D7,$99,$F1,$CB,$3D,$E3,$C3,$3E,$3A,$F6,$57,$E3,$8F
	.dc.b	$DE,$3F,$B6,$9F,$C5,$C6,$27,$F9,$64,$62,$3E,$D9,$93,$F8,$34,$F1
	.dc.b	$3E,$8B,$FE,$B4,$FA,$17,$B9,$88,$F2,$DF,$CC,$9C,$BF,$52,$3D,$B5
	.dc.b	$FB,$99,$CC,$7C,$C3,$33,$FB,$58,$CE,$1F,$23,$0F,$7D,$7C,$C9,$CD
	.dc.b	$7C,$AB,$F1,$46,$27,$99,$D4,$39,$7C,$E6,$AF,$FC,$9D,$19,$BE,$F3
	.dc.b	$A2,$46,$6B,$E0,$4E,$A3,$17,$EF,$AD,$E1,$AB,$F2,$B6,$75,$5C,$2B
	.dc.b	$CD,$7C,$EB,$51,$E9,$9D,$61,$F0,$F0,$F9,$55,$A8,$FC,$6C,$EB,$0F
	.dc.b	$88,$7C,$B9,$D3,$F5,$5B,$7F,$2D,$AF,$A2,$3D,$58,$7C,$23,$C8,$E3
	.dc.b	$0F,$5B,$4F,$5D,$7C,$4F,$13,$F2,$04,$8F,$C7,$B6,$FF,$8F,$3F,$CC
	.dc.b	$7C,$5F,$E6,$4F,$F3,$3F,$8D,$95,$FC,$C3,$C3,$B7,$FC,$95,$FC,$E1
	.dc.b	$23,$F3,$6F,$E7,$FE,$86,$47,$E7,$9B,$3F,$D8,$7C,$1F,$E8,$11,$FD
	.dc.b	$05,$B3,$FC,$F9,$C3,$FD,$1F,$E8,$1E,$6E,$2F,$F9,$D7,$D0,$3E,$83
	.dc.b	$FA,$A2,$FA,$1B,$18,$6E,$1F,$D4,$67,$83,$7C,$5C,$DF,$4F,$86,$E6
	.dc.b	$74,$F1,$6D,$71,$6C,$F8,$9B,$CD,$FE,$A7,$C1,$F0,$9E,$43,$CB,$C9
	.dc.b	$70,$F1,$37,$0F,$5B,$6F,$FD,$61,$F6,$0F,$A8,$E3,$7F,$DA,$63,$0F
	.dc.b	$23,$D9,$F0,$FF,$0B,$E2,$7E,$1B,$63,$E0,$7C,$46,$1A,$70,$3E,$27
	.dc.b	$C3,$7F,$05,$E2,$F9,$6D,$8F,$91,$A6,$1A,$70,$3F,$BE,$7D,$1B,$D4
	.dc.b	$FE,$1D,$C7,$FE,$07,$8C,$4F,$F0,$D8,$C7,$85,$9A,$B3,$C9,$FF,$07
	.dc.b	$6A,$E9,$CE,$3F,$DE,$63,$15,$E0,$BE,$DF,$0B,$DF,$CF,$B3,$E4,$6D
	.dc.b	$FF,$80,$73,$57,$97,$E1,$73,$C7,$FE,$DF,$97,$C9,$7A,$4F,$57,$91
	.dc.b	$72,$F7,$B7,$1E,$8D,$BE,$5F,$03,$FC,$BF,$E6,$3E,$13,$6F,$E3,$B8
	.dc.b	$7E,$4F,$C0,$DC,$3C,$CF,$C9,$3F,$C6,$1F,$61,$F9,$D9,$D9,$FE,$2B
	.dc.b	$FB,$67,$77,$F8,$D5,$77,$D8,$C3,$9E,$B9,$5C,$7D,$9B,$97,$F8,$97
	.dc.b	$E9,$5F,$3D,$FE,$3F,$6D,$7F,$8E,$5F,$F8,$00,$D8,$FF,$22,$7D,$B7
	.dc.b	$F6,$92,$7A,$18,$F4,$33,$E1,$6B,$DF,$B8,$7F,$90,$EF,$FD,$A8,$D3
	.dc.b	$9F,$95,$D1,$A7,$F9,$27,$ED,$9F,$13,$FE,$53,$5F,$15,$63,$0E,$7E
	.dc.b	$32,$57,$1E,$AB,$C1,$FE,$43,$A7,$C4,$FF,$97,$5B,$5F,$E5,$77,$F8
	.dc.b	$8B,$63,$FC,$C5,$C8,$E3,$A6,$3D,$EB,$5F,$E5,$FF,$A4,$F8,$F8,$FF
	.dc.b	$94,$6C,$FF,$9A,$E3,$3D,$AA,$CC,$FD,$A3,$5E,$0E,$FC,$2B,$3D,$5F
	.dc.b	$F3,$A7,$D0,$BE,$0E,$98,$F2,$6D,$7F,$9C,$67,$FB,$91,$FC,$59,$C7
	.dc.b	$9D,$CB,$CF,$E7,$E6,$74,$F9,$2B,$5C,$5F,$F4,$17,$82,$F8,$7F,$F4
	.dc.b	$27,$DC,$FA,$6B,$FC,$FD,$39,$7F,$D0,$11,$9D,$FF,$CE,$52,$F8,$DE
	.dc.b	$67,$46,$D7,$2F,$4F,$B9,$E0,$F1,$6C,$7F,$A2,$FE,$85,$E5,$FF,$A5
	.dc.b	$5F,$85,$19,$8C,$71,$67,$9E,$4F,$1C,$6E,$7F,$E8,$5C,$3F,$57,$39
	.dc.b	$3B,$67,$D6,$BB,$7D,$0C,$9F,$17,$0D,$C3,$93,$E3,$3C,$9C,$56,$7D
	.dc.b	$1C,$3E,$73,$D1,$C5,$63,$D8,$E1,$FF,$54,$72,$BC,$6C,$63,$FE,$9B
	.dc.b	$E8,$77,$FF,$AD,$5B,$7F,$EA,$F5,$F2,$59,$9D,$32,$7A,$FC,$C5,$BF
	.dc.b	$F5,$30,$E6,$FF,$60,$36,$7F,$D7,$4B,$D7,$CC,$E9,$93,$AF,$97,$B5
	.dc.b	$D1,$E6,$7F,$B2,$F4,$7B,$D8,$AE,$BB,$F8,$0F,$0E,$F6,$3C,$2E,$1F
	.dc.b	$1E,$0F,$A0,$FD,$6F,$D0,$BA,$FF,$57,$2B,$F5,$BB,$32,$7E,$C2,$AB
	.dc.b	$4A,$6C,$E7,$F4,$0F,$A5,$7D,$17,$08,$C3,$63,$33,$DA,$72,$DF,$5D
	.dc.b	$CB,$C5,$B1,$98,$C1,$D7,$CA,$FE,$0A,$FF,$07,$8F,$E1,$75,$F8,$27
	.dc.b	$FD,$DC,$D7,$FD,$C0,$A7,$8E,$BF,$C3,$A3,$55,$CE,$9C,$0F,$97,$DC
	.dc.b	$7F,$18,$8C,$3D,$B9,$F2,$AF,$F1,$EF,$8B,$E6,$73,$06,$F4,$61,$CF
	.dc.b	$B3,$2B,$81,$EA,$7A,$E0,$F8,$6A,$6E,$7F,$DF,$9C,$6E,$7F,$1A,$E6
	.dc.b	$A6,$22,$CF,$EB,$3E,$7F,$A9,$87,$E9,$F0,$7F,$32,$FA,$1F,$91,$1F
	.dc.b	$CE,$38,$7E,$34,$7C,$89,$D7,$07,$C2,$8D,$D7,$23,$F9,$FB,$B7,$FA
	.dc.b	$0F,$D1,$7F,$96,$5E,$95,$FD,$02,$31,$1C,$44,$F1,$5F,$45,$FD,$61
	.dc.b	$E2,$7F,$4C,$58,$FD,$46,$C6,$8E,$3E,$8E,$67,$5E,$04,$9D,$57,$07
	.dc.b	$C4,$BE,$33,$A7,$34,$9D,$C7,$7F,$A2,$91,$C9,$D3,$CB,$8B,$C8,$47
	.dc.b	$BF,$85,$F8,$AF,$3F,$F7,$11,$E7,$7F,$B7,$37,$D3,$D7,$2F,$E4,$3E
	.dc.b	$8B,$FB,$34,$FC,$76,$F2,$7E,$6D,$BC,$7F,$DB,$1F,$E8,$F3,$F0,$DF
	.dc.b	$EA,$4D,$F2,$BA,$1D,$7C,$AF,$27,$8D,$27,$C8,$F9,$FC,$7F,$17,$91
	.dc.b	$E3,$F2,$65,$75,$8F,$36,$BE,$7F,$52,$4F,$86,$3E,$07,$6C,$8F,$80
	.dc.b	$FC,$3E,$DF,$71,$EF,$CA,$4F,$A6,$3E,$17,$91,$F0,$EA,$4F,$B6,$3E
	.dc.b	$63,$F3,$58,$FC,$D3,$F7,$3F,$44,$FA,$1A,$48,$FC,$44,$7F,$91,$1B
	.dc.b	$F0,$DC,$08,$EB,$E2,$7F,$81,$FC,$9E,$E3,$F3,$4D,$E5,$FE,$1C,$DE
	.dc.b	$5F,$F0,$FA,$FA,$2F,$1F,$B3,$E4,$F1,$A6,$1E,$28,$F3,$FC,$E7,$A3
	.dc.b	$E2,$A4,$F9,$A3,$C3,$9F,$8B,$69,$3F,$A1,$CB,$FD,$2C,$FB,$3C,$8F
	.dc.b	$82,$47,$DF,$E6,$FC,$4F,$2E,$47,$C2,$7E,$3F,$99,$EE,$E1,$F7,$FB
	.dc.b	$B2,$BB,$47,$F2,$4F,$97,$FF,$45,$C7,$EC,$5F,$E0,$BC,$7F,$CB,$DB
	.dc.b	$C7,$FE,$40,$77,$3F,$23,$DF,$F9,$64,$79,$67,$D7,$CF,$91,$D5,$F0
	.dc.b	$69,$88,$FE,$15,$6F,$8A,$F2,$19,$CB,$FA,$07,$CE,$BE,$17,$F9,$DF
	.dc.b	$D1,$7F,$BC,$30,$FE,$10,$D7,$F8,$A4,$ED,$CF,$A3,$C0,$8F,$0F,$93
	.dc.b	$E7,$E7,$48,$E6,$FF,$48,$FA,$2F,$D8,$C7,$F1,$FC,$3F,$87,$3E,$0F
	.dc.b	$43,$C6,$71,$7F,$92,$4B,$E8,$B9,$12,$38,$1E,$8F,$98,$F5,$91,$F5
	.dc.b	$FA,$3E,$ED,$7F,$BE,$E3,$EB,$9F,$F0,$1B,$5F,$F0,$4B,$5F,$F3,$AE
	.dc.b	$0F,$23,$FC,$0A,$D8,$FF,$04,$B5,$FF,$06,$B5,$E7,$F4,$3E,$02,$3F
	.dc.b	$E1,$0A,$47,$4F,$26,$4F,$46,$BE,$6F,$8A,$F4,$91,$E4,$FB,$3A,$64
	.dc.b	$75,$6B,$FE,$20,$FA,$1F,$E9,$EF,$F8,$99,$AF,$FA,$49,$AF,$FA,$61
	.dc.b	$A9,$7A,$7F,$E2,$66,$C7,$F8,$A9,$AF,$78,$FF,$15,$FD,$A7,$EC,$A3
	.dc.b	$FC,$62,$D7,$FC,$6E,$D7,$FD,$46,$D7,$C5,$E9,$77,$23,$DD,$FF,$1B
	.dc.b	$FC,$37,$82,$3F,$C8,$2D,$7F,$D5,$AD,$7F,$D6,$6D,$5A,$FF,$82,$5F
	.dc.b	$B7,$E6,$FD,$E2,$3E,$DD,$FE,$01,$1C,$1F,$E4,$AF,$13,$F7,$6F,$F9
	.dc.b	$49,$AF,$F2,$0F,$7B,$FC,$9B,$18,$8F,$F6,$2F,$CF,$3E,$81,$F4,$1F
	.dc.b	$E5,$48,$C3,$7C,$0F,$F2,$E5,$7F,$96,$A7,$2F,$F9,$73,$7F,$CB,$70
	.dc.b	$FF,$97,$1B,$19,$CF,$FC,$80,$FF,$B4,$23,$2F,$F9,$A3,$E6,$7C,$35
	.dc.b	$FE,$64,$6C,$7F,$9A,$1A,$EF,$1E,$6F,$FB,$7B,$E7,$DF,$29,$FE,$57
	.dc.b	$F0,$BF,$A0,$F8,$77,$1F,$F6,$FD,$69,$FE,$6D,$19,$BE,$57,$56,$FF
	.dc.b	$C6,$EB,$FC,$39,$7C,$67,$FA,$7D,$EE,$31,$87,$82,$B7,$87,$3C,$39
	.dc.b	$61,$FE,$2F,$BD,$D7,$F5,$4C,$39,$5E,$DB,$1E,$63,$99,$6B,$94,$EE
	.dc.b	$33,$1E,$46,$E1,$E3,$8D,$C7,$18,$FF,$39,$5F,$8A,$47,$C0,$BF,$0B
	.dc.b	$CC,$E2,$59,$E3,$7E,$18,$CC,$77,$33,$7D,$CF,$A3,$3B,$8C,$CF,$39
	.dc.b	$C9,$B7,$E9,$B5,$5F,$9D,$F3,$47,$B8,$F9,$45,$7D,$78,$7D,$0A,$EE
	.dc.b	$79,$67,$BC,$F8,$4E,$E5,$3B,$DF,$9E,$37,$3C,$67,$6D,$77,$1E,$7B
	.dc.b	$F3,$5F,$86,$38,$9E,$69,$F8,$ED,$AF,$2B,$4D,$C7,$AB,$0F,$A7,$3D
	.dc.b	$EF,$AC,$62,$B8,$57,$91,$AF,$90,$F6,$4F,$08,$F8,$D1,$F2,$8F,$19
	.dc.b	$DD,$CD,$FA,$63,$E2,$9F,$ED,$E9,$FD,$13,$70,$FD,$9B,$DC,$C5,$7F
	.dc.b	$B7,$A7,$71,$F4,$0C,$9F,$EE,$47,$EE,$13,$C0,$F6,$47,$AC,$87,$E0
	.dc.b	$9C,$BE,$DC,$7E,$4B,$FC,$0E,$33,$3D,$2F,$F6,$D1,$87,$2D,$BF,$87
	.dc.b	$3F,$E9,$16,$CE,$9A,$6A,$78,$57,$C8,$3C,$91,$A7,$15,$CD,$B5,$CD
	.dc.b	$B3,$CC,$E1,$7E,$A8,$E6,$68,$E7,$7E,$08,$D5,$F5,$6D,$F5,$6D,$F5
	.dc.b	$BE,$51,$C2,$F9,$4F,$E3,$A7,$BC,$F0,$7B,$C7,$90,$D5,$F7,$6D,$F7
	.dc.b	$69,$DE,$B9,$1D,$DF,$9B,$EE,$FE,$1A,$8D,$24,$7E,$1F,$BB,$F9,$64
	.dc.b	$7F,$2D,$6C,$7F,$32,$F8,$9E,$44,$7F,$1C,$DC,$7F,$97,$3F,$CD,$7E
	.dc.b	$7F,$F9,$B3,$FC,$F7,$E8,$1F,$01,$4F,$E7,$5C,$37,$0F,$E7,$EF,$9C
	.dc.b	$8F,$E6,$DF,$57,$74,$8F,$CF,$37,$1F,$E5,$64,$F9,$E7,$CF,$FE,$99
	.dc.b	$B5,$A9,$D4,$E4,$D7,$CF,$C9,$D4,$FF,$36,$A4,$E6,$33,$0F,$91,$EC
	.dc.b	$4A,$E3,$29,$D4,$61,$7C,$1F,$F6,$17,$32,$E2,$57,$AE,$B5,$7B,$F8
	.dc.b	$2F,$61,$27,$A3,$6F,$94,$7C,$38,$ED,$F0,$5C,$63,$FA,$8E,$1D,$AB
	.dc.b	$FA,$9F,$CF,$BE,$6B,$75,$A3,$E6,$C6,$5C,$4F,$C5,$BF,$57,$C2,$FA
	.dc.b	$F8,$6D,$6A,$BE,$14,$61,$F2,$9F,$36,$7E,$17,$C2,$E7,$2F,$24,$B8
	.dc.b	$FE,$8E,$1C,$DB,$1C,$8F,$9D,$F0,$9F,$8F,$E1,$F9,$91,$F3,$62,$D7
	.dc.b	$8F,$85,$1F,$E1,$1F,$9F,$7B,$FF,$97,$F0,$70,$DB,$F5,$4E,$9B,$1C
	.dc.b	$7D,$BD,$C7,$BF,$85,$D1,$98,$C3,$73,$F4,$60,$F8,$2F,$17,$C1,$F8
	.dc.b	$DB,$7E,$07,$0F,$8B,$E0,$ED,$F2,$F2,$BD,$47,$AF,$DD,$B5,$E8,$6C
	.dc.b	$F8,$1B,$C7,$F3,$B2,$7B,$BF,$B7,$F9,$5F,$88,$DC,$3E,$0B,$87,$E1
	.dc.b	$F9,$17,$1F,$5B,$F2,$CF,$F1,$E7,$CF,$FE,$C2,$33,$3F,$83,$F9,$FF
	.dc.b	$D0,$B7,$BF,$FC,$85,$18,$9F,$B8,$FE,$93,$3F,$C1,$BF,$C2,$AB,$97
	.dc.b	$67,$0F,$5B,$8B,$FE,$4B,$F9,$F7,$CF,$D3,$1F,$13,$6F,$F2,$9F,$3F
	.dc.b	$FE,$1D,$27,$C8,$7F,$94,$DC,$78,$97,$DE,$C3,$6B,$89,$BF,$27,$83
	.dc.b	$4F,$D0,$7C,$FB,$E1,$FF,$CB,$CB,$E2,$2C,$61,$CF,$C5,$4A,$E3,$D5
	.dc.b	$79,$3F,$3D,$4F,$87,$FF,$34,$36,$BF,$5C,$BE,$1F,$33,$FE,$6D,$E7
	.dc.b	$72,$D3,$1E,$CD,$C3,$F6,$DF,$3E,$E4,$E1,$9F,$F9,$D6,$31,$3F,$BB
	.dc.b	$C3,$FC,$B2,$F0,$F0,$75,$E1,$6D,$B1,$FC,$07,$E7,$DF,$03,$FE,$7F
	.dc.b	$6C,$67,$C9,$C4,$E7,$D0,$D7,$A3,$6B,$93,$C7,$CC,$E5,$F0,$16,$7B
	.dc.b	$3F,$E8,$BF,$4D,$F3,$FF,$E8,$DF,$9F,$FE,$57,$5F,$C4,$E7,$2F,$F1
	.dc.b	$58,$CF,$0F,$F0,$B9,$7C,$27,$33,$A3,$6B,$97,$AD,$DC,$F0,$78,$B6
	.dc.b	$3F,$90,$7C,$FB,$CC,$FF,$4F,$3F,$03,$1F,$E3,$0E,$67,$7C,$1A,$E5
	.dc.b	$91,$CB,$2B,$7C,$1E,$67,$4C,$EC,$F8,$3E,$7F,$B9,$E0,$E1,$B1,$E4
	.dc.b	$70,$F9,$B8,$AE,$1E,$4E,$1F,$39,$E8,$E2,$B5,$EC,$70,$FF,$3D,$E6
	.dc.b	$78,$D8,$BF,$E6,$9E,$53,$C1,$FF,$5F,$36,$FF,$A2,$2F,$90,$CC,$E9
	.dc.b	$93,$D7,$DE,$B7,$FC,$FC,$73,$7F,$B2,$5B,$3F,$D3,$D7,$B5,$99,$D3
	.dc.b	$27,$5F,$29,$6B,$A3,$CC,$FF,$6A,$68,$F7,$71,$5D,$77,$F0,$1E,$1D
	.dc.b	$EC,$78,$5C,$3E,$3C,$1F,$3D,$FE,$C0,$F9,$FF,$F1,$1C,$7F,$86,$2B
	.dc.b	$D4,$E8,$FF,$04,$94,$FF,$05,$CE,$1C,$1F,$E2,$0F,$9D,$7C,$F3,$E7
	.dc.b	$FF,$C2,$31,$86,$C6,$67,$FC,$3B,$39,$6F,$B3,$FE,$11,$6F,$27,$28
	.dc.b	$C4,$7F,$8A,$65,$18,$8F,$17,$99,$FE,$32,$AF,$EF,$35,$F8,$19,$FF
	.dc.b	$01,$35,$FE,$F0,$A7,$86,$9E,$BE,$5B,$5F,$EE,$6E,$07,$CA,$62,$11
	.dc.b	$87,$AC,$E7,$0F,$C6,$47,$83,$E5,$B0,$67,$9E,$C7,$93,$DA,$E0,$61
	.dc.b	$F2,$BD,$F0,$7C,$35,$37,$3F,$E1,$DF,$11,$CF,$E5,$5C,$D4,$C4,$59
	.dc.b	$FD,$47,$CE,$F5,$30,$FF,$86,$FE,$7B,$E2,$3F,$E2,$97,$E4,$71,$7F
	.dc.b	$89,$1B,$C3,$FE,$60,$77,$3C,$78,$3E,$37,$CF,$FF,$96,$23,$FC,$64
	.dc.b	$D7,$47,$F8,$D7,$E7,$FF,$A6,$37,$1F,$9D,$06,$FA,$58,$73,$E0,$4A
	.dc.b	$E0,$73,$F8,$DE,$3C,$D2,$30,$77,$7A,$FC,$FC,$8C,$1E,$3F,$0B,$C3
	.dc.b	$C7,$23,$07,$F9,$2F,$F6,$5F,$F1,$6B,$7E,$17,$F0,$91,$88,$FF,$22
	.dc.b	$2F,$93,$FC,$46,$7F,$DC,$E7,$FC,$A8,$FF,$73,$6F,$B3,$AF,$6B,$7E
	.dc.b	$AF,$0E,$A9,$5C,$23,$BF,$37,$83,$DC,$93,$E5,$8E,$FE,$0C,$8E,$EF
	.dc.b	$8B,$C2,$F1,$F4,$49,$F1,$BE,$6F,$0B,$F3,$DE,$AF,$FA,$6A,$7F,$CD
	.dc.b	$4F,$F5,$C6,$FA,$BA,$F8,$DF,$6F,$C2,$C9,$F6,$BF,$03,$C3,$F8,$3F
	.dc.b	$25,$27,$E0,$BF,$13,$E8,$FB,$E3,$27,$BB,$E2,$F8,$EF,$D8,$7D,$CF
	.dc.b	$F3,$E9,$FF,$3E,$3F,$85,$6F,$7E,$B9,$F7,$DF,$C3,$E4,$E5,$70,$8F
	.dc.b	$6F,$B7,$EE,$F6,$A4,$FB,$A3,$BF,$7C,$8E,$EF,$8B,$C0,$F1,$FC,$2C
	.dc.b	$9F,$1B,$E6,$F9,$CF,$93,$ED,$C9,$F9,$2F,$CC,$F7,$3D,$7F,$37,$EC
	.dc.b	$F9,$49,$5E,$C8,$F7,$F8,$9F,$03,$E6,$24,$FC,$08,$FA,$BF,$3B,$F5
	.dc.b	$BE,$02,$4F,$D6,$8F,$E2,$D3,$FC,$5A,$3F,$D4,$4F,$F9,$FD,$BE,$26
	.dc.b	$BC,$5D,$BF,$80,$F8,$09,$3A,$38,$F8,$9D,$3C,$29,$18,$3A,$BF,$C9
	.dc.b	$7E,$63,$F3,$D1,$FE,$A5,$6F,$A5,$FE,$AC,$CB,$F8,$3B,$7D,$7F,$3F
	.dc.b	$85,$2F,$CF,$15,$FF,$57,$53,$C3,$F6,$99,$BF,$F5,$BF,$CF,$39,$FF
	.dc.b	$9B,$7C,$F7,$F9,$4F,$0F,$F6,$13,$5F,$F5,$C4,$ED,$CF,$C5,$F0,$F4
	.dc.b	$E4,$60,$E2,$FF,$41,$9F,$F6,$B4,$7F,$B3,$5A,$FF,$B0,$EB,$F8,$DB
	.dc.b	$7B,$FC,$7E,$9C,$8F,$1C,$57,$FD,$95,$F3,$8E,$6F,$E9,$CD,$8F,$4D
	.dc.b	$7B,$3B,$FD,$BD,$D2,$7D,$B1,$F0,$35,$91,$F0,$1D,$3D,$BE,$2F,$A9
	.dc.b	$1F,$D5,$F9,$FF,$23,$B6,$47,$C8,$7E,$5F,$77,$F6,$1F,$A2,$FE,$05
	.dc.b	$3F,$EE,$16,$F1,$FF,$BA,$1B,$69,$9C,$0C,$3E,$AF,$A3,$27,$EA,$BD
	.dc.b	$7C,$5F,$B3,$72,$7E,$CB,$B7,$FD,$E1,$F3,$EF,$33,$FB,$A3,$63,$FE
	.dc.b	$00,$35,$DF,$99,$F3,$3E,$1A,$47,$97,$5F,$F7,$FF,$27,$F8,$F2,$38
	.dc.b	$E5,$FD,$FF,$E7,$FF,$DE,$B1,$BC,$FD,$8A,$32,$AD,$3A,$9C,$C6,$27
	.dc.b	$E2,$77,$BE,$63,$D3,$E3,$49,$F4,$BE,$CF,$27,$11,$FA,$BF,$1F,$F5
	.dc.b	$C6,$30,$E9,$0D,$DF,$13,$FE,$3A,$6C,$7A,$0F,$4B,$87,$07,$18,$FF
	.dc.b	$20,$F0,$BE,$2A,$92,$76,$7D,$17,$FC,$8F,$F3,$EF,$2D,$EA,$6A,$30
	.dc.b	$DF,$19,$F1,$19,$AD,$E1,$E2,$C7,$D3,$C7,$D9,$B1,$9C,$FD,$EF,$D2
	.dc.b	$8C,$BE,$DF,$9F,$FF,$69,$57,$D7,$8F,$C9,$7B,$DE,$48,$C4,$27,$6F
	.dc.b	$D9,$8F,$BB,$3B,$8C,$1F,$9B,$9D,$EF,$F4,$27,$4F,$D2,$8C,$D7,$2D
	.dc.b	$3F,$01,$39,$F2,$2C,$E4,$F1,$D7,$0C,$3C,$F7,$A7,$C1,$55,$61,$E0
	.dc.b	$BF,$C9,$DF,$19,$CD,$EE,$31,$5B,$9E,$75,$F0,$A3,$11,$F0,$EF,$B4
	.dc.b	$62,$B8,$DF,$39,$CB,$6F,$93,$6F,$E3,$9E,$28,$F6,$CF,$06,$9C,$0B
	.dc.b	$1C,$2F,$C1,$3B,$8E,$77,$E2,$AD,$E1,$F4,$70,$F0,$36,$32,$7C,$DA
	.dc.b	$E3,$7C,$E7,$E9,$4F,$86,$72,$D8,$CC,$78,$EB,$AC,$F3,$9C,$9D,$E7
	.dc.b	$89,$B8,$F0,$BF,$69,$AF,$C3,$37,$1C,$4E,$71,$D4,$FB,$75,$E7,$69
	.dc.b	$E5,$2C,$F7,$39,$BC,$5F,$BA,$66,$7A,$BE,$33,$65,$72,$F9,$5B,$87
	.dc.b	$D2,$9F,$04,$F4,$9F,$44,$78,$9B,$19,$8F,$13,$F8,$B7,$8B,$4F,$E1
	.dc.b	$F8,$7E,$C1,$FD,$C5,$FE,$F1,$F4,$98,$9F,$E1,$F3,$B8,$F8,$26,$4E
	.dc.b	$73,$F5,$9F,$EA,$1E,$A3,$E7,$31,$1E,$1C,$3D,$F3,$9C,$3E,$AE,$1F
	.dc.b	$A9,$3F,$55,$19,$AF,$8E,$71,$9F,$80,$7F,$42,$8D,$57,$CA,$AE,$58
	.dc.b	$7E,$DE,$73,$5C,$2F,$E3,$10,$46,$9C,$57,$29,$FF,$43,$E1,$EA,$8C
	.dc.b	$4F,$EB,$AB,$83,$E5,$38,$DF,$86,$B8,$3F,$0A,$77,$5C,$B0,$F8,$75
	.dc.b	$AC,$3F,$2F,$5A,$BC,$61,$DE,$AD,$F0,$BF,$93,$5D,$67,$31,$EF,$AF
	.dc.b	$6C,$61,$F8,$17,$D6,$78,$9E,$E2,$C7,$D0,$8D,$3F,$4B,$0F,$AB,$5C
	.dc.b	$E3,$E0,$1D,$AB,$8E,$1E,$58,$E1,$87,$96,$72,$79,$70,$EB,$5F,$CE
	.dc.b	$7D,$5F,$E5,$07,$F2,$9F,$26,$D7,$F3,$67,$F2,$D8,$7F,$49,$6E,$46
	.dc.b	$98,$F1,$BF,$99,$7A,$9F,$CC,$DB,$7F,$D0,$1C,$3F,$D0,$BC,$8B,$87
	.dc.b	$F2,$0F,$8C,$FE,$74,$7F,$3B,$F1,$ED,$7F,$48,$6C,$FF,$26,$6E,$9F
	.dc.b	$E0,$4F,$9F,$7C,$A7,$1B,$FE,$C3,$18,$79,$1F,$D4,$B8,$BF,$A7,$43
	.dc.b	$E2,$7F,$A5,$46,$1F,$E9,$9E,$3E,$2B,$FA,$CB,$72,$FE,$83,$F3,$EF
	.dc.b	$23,$FA,$C2,$F7,$AC,$61,$B7,$FD,$D2,$78,$37,$A7,$37,$D3,$DA,$FE
	.dc.b	$C5,$F4,$7F,$E2,$EA,$E1,$18,$7B,$B1,$E9,$E2,$5A,$F2,$BE,$43,$FB
	.dc.b	$67,$CF,$B9,$31,$5E,$3F,$0F,$D2,$7A,$BB,$EC,$79,$DC,$3F,$E0,$CA
	.dc.b	$7C,$CE,$31,$F8,$7C,$3F,$08,$F8,$7E,$6D,$BF,$80,$E1,$F8,$5F,$27
	.dc.b	$6B,$FB,$5F,$2F,$C4,$6F,$1F,$CC,$F1,$3E,$BB,$6F,$E5,$47,$D0,$9C
	.dc.b	$1F,$50,$C4,$79,$FC,$4B,$1F,$DE,$3E,$7D,$F4,$1F,$E0,$F6,$D7,$F8
	.dc.b	$A1,$FE,$1E,$62,$78,$78,$BB,$3E,$5F,$55,$AE,$1E,$57,$33,$A7,$C0
	.dc.b	$58,$E8,$E1,$ED,$E6,$DC,$3C,$9E,$FF,$8C,$F2,$7B,$F6,$BC,$8D,$9F
	.dc.b	$53,$7C,$1F,$F3,$2F,$D0,$3E,$1B,$8D,$FF,$90,$A3,$0F,$23,$FC,$8B
	.dc.b	$F4,$0F,$84,$FF,$18,$36,$33,$EB,$62,$73,$ED,$EB,$B6,$D7,$27,$8E
	.dc.b	$9C,$B9,$AC,$F6,$7F,$C7,$9F,$3C,$F8,$6B,$F0,$58,$6D,$7F,$96,$27
	.dc.b	$F2,$CD,$F0,$B8,$F9,$BC,$9E,$7C,$F2,$78,$7A,$B7,$3F,$F1,$AF,$CF
	.dc.b	$7F,$48,$27,$67,$F9,$35,$C7,$3D,$F8,$BC,$37,$0F,$F3,$4C,$FF,$92
	.dc.b	$07,$C5,$71,$F6,$39,$3F,$E5,$7F,$2B,$FC,$F3,$D5,$FE,$96,$7F,$D1
	.dc.b	$97,$FE,$72,$72,$FF,$AB,$1C,$E7,$FE,$4A,$7F,$D0,$F1,$BE,$3E,$9F
	.dc.b	$29,$D5,$FF,$32,$7C,$EB,$E3,$A9,$8F,$8F,$B7,$FE,$80,$F9,$DF,$D5
	.dc.b	$67,$FE,$A8,$7F,$8A,$46,$7C,$7F,$F3,$74,$63,$4F,$F2,$DA,$F8,$EC
	.dc.b	$9A,$F8,$AB,$7D,$7E,$5B,$89,$CB,$CC,$B1,$FE,$90,$F9,$D7,$B1,$FE
	.dc.b	$7B,$6C,$67,$AB,$13,$FE,$7F,$F5,$1C,$97,$C3,$C3,$6F,$FD,$37,$3F
	.dc.b	$CE,$F6,$FF,$38,$AF,$0F,$97,$B3,$CD,$FF,$45,$7C,$EB,$C6,$A6,$32
	.dc.b	$B8,$FF,$A9,$A7,$C2,$DC,$B8,$E3,$C9,$B1,$B8,$EB,$F0,$B8,$CF,$FD
	.dc.b	$07,$7F,$26,$BA,$71,$78,$BD,$7F,$1B,$63,$C1,$1E,$6F,$23,$17,$CB
	.dc.b	$B3,$D2,$7A,$BB,$2D,$F9,$DC,$3F,$EB,$EF,$3F,$FC,$5A,$E4,$FF,$60
	.dc.b	$F5,$3E,$4F,$FD,$58,$D8,$CF,$CC,$E2,$72,$C8,$E9,$F0,$17,$2F,$F4
	.dc.b	$F8,$F5,$FF,$D6,$CD,$9F,$F6,$7A,$EA,$CC,$E9,$93,$AE,$BB,$5D,$1E
	.dc.b	$67,$FB,$07,$37,$C0,$62,$BA,$ED,$E0,$3C,$3B,$58,$F0,$B8,$7C,$78
	.dc.b	$3E,$70,$BE,$77,$FC,$00,$E0,$FD,$89,$EE,$38,$19,$24,$56,$94,$EA
	.dc.b	$DF,$9C,$79,$4F,$9D,$C3,$6B,$94,$EA,$72,$71,$F2,$DC,$94,$95,$C2
	.dc.b	$32,$F2,$F5,$BD,$52,$CF,$D9,$87,$E0,$6B,$E0,$CE,$5F,$87,$19,$7F
	.dc.b	$DB,$6A,$B9,$D7,$2E,$5B,$5F,$2B,$81,$F2,$78,$84,$61,$EF,$5C,$5A
	.dc.b	$78,$39,$B8,$62,$F9,$6A,$62,$2C,$F0,$12,$F9,$7E,$26,$A3,$D2,$F8
	.dc.b	$F0,$7A,$74,$DC,$F3,$F8,$07,$3F,$A5,$73,$53,$11,$67,$F5,$DE,$B7
	.dc.b	$53,$0F,$D6,$D7,$F1,$5F,$3B,$F2,$63,$D0,$FC,$DE,$2F,$A1,$08,$C4
	.dc.b	$F0,$E2,$F9,$51,$89,$FC,$BF,$CE,$FF,$93,$5F,$84,$D7,$C0,$7E,$72
	.dc.b	$38,$2F,$9D,$7B,$5F,$55,$FE,$E1,$1F,$2D,$AF,$23,$84,$6F,$2E,$1F
	.dc.b	$1D,$27,$84,$73,$F8,$CE,$9C,$D2,$7A,$45,$7A,$60,$F9,$DF,$11,$9D
	.dc.b	$BE,$1C,$3C,$2F,$37,$1C,$BF,$34,$7E,$37,$E7,$5E,$85,$24,$FE,$16
	.dc.b	$1E,$E7,$E4,$F6,$FE,$CD,$3F,$89,$5E,$B6,$4D,$7A,$FB,$F8,$3E,$1D
	.dc.b	$52,$F8,$47,$87,$9F,$C5,$EA,$49,$F0,$47,$7F,$06,$47,$77,$C5,$E1
	.dc.b	$F8,$FC,$29,$3E,$37,$CD,$D7,$FA,$9F,$8B,$FF,$44,$4F,$E7,$DF,$F1
	.dc.b	$D3,$7E,$0F,$5F,$1B,$ED,$F9,$C9,$3E,$D7,$E0,$78,$7F,$07,$DF,$93
	.dc.b	$F0,$5F,$89,$EF,$F7,$C6,$4F,$77,$C5,$EF,$7E,$EF,$E7,$7F,$DE,$F3
	.dc.b	$FB,$17,$F0,$FD,$BF,$9A,$6F,$0F,$F2,$97,$73,$F8,$B7,$FD,$8C,$F1
	.dc.b	$F9,$2E,$87,$5F,$03,$E0,$7B,$5F,$07,$DB,$95,$F0,$63,$C5,$E0,$48
	.dc.b	$F1,$3E,$5F,$03,$CD,$F3,$92,$7C,$CF,$A7,$E1,$7E,$67,$B9,$27,$E6
	.dc.b	$3F,$3F,$DD,$F7,$7C,$D7,$BF,$E4,$E5,$7B,$E3,$E1,$78,$9F,$0F,$E3
	.dc.b	$A4,$FC,$38,$FB,$1F,$09,$F6,$7E,$06,$4F,$D9,$8F,$E4,$93,$FC,$92
	.dc.b	$3F,$8D,$2F,$0F,$26,$BC,$4B,$3F,$81,$F8,$0F,$C2,$18,$7F,$6D,$4F
	.dc.b	$13,$AF,$85,$27,$AC,$57,$F7,$9F,$31,$FA,$26,$F8,$F9,$EC,$FE,$27
	.dc.b	$87,$F3,$19,$F2,$F8,$3E,$83,$D3,$EC,$58,$F4,$F8,$7E,$C3,$0F,$F0
	.dc.b	$9F,$9D,$74,$7F,$43,$F9,$C7,$CC,$53,$FA,$34,$EA,$BF,$9E,$4E,$9B
	.dc.b	$EA,$71,$38,$38,$3F,$A1,$4F,$FB,$41,$BF,$29,$B7,$9F,$83,$E1,$F5
	.dc.b	$24,$60,$EE,$FF,$57,$F9,$C7,$45,$37,$3F,$C8,$F2,$FE,$92,$F9,$BB
	.dc.b	$FD,$87,$B7,$C1,$B1,$EA,$DB,$E0,$18,$7F,$45,$E3,$77,$F8,$89,$1F
	.dc.b	$AB,$8F,$8D,$E0,$7C,$7F,$02,$4F,$C4,$8F,$ED,$9F,$3B,$FC,$0A,$3F
	.dc.b	$9B,$37,$A3,$FA,$5B,$6F,$FC,$35,$A3,$0E,$3E,$37,$D3,$F4,$66,$3F
	.dc.b	$DA,$F0,$FF,$32,$3F,$4B,$0F,$B0,$47,$EA,$CF,$F7,$DF,$9C,$79,$94
	.dc.b	$93,$FD,$D6,$3F,$AA,$67,$FD,$79,$EB,$1F,$E2,$76,$BF,$E2,$B6,$BF
	.dc.b	$E3,$16,$BF,$E3,$6E,$E7,$CA,$7F,$8B,$61,$18,$7F,$C6,$0D,$7F,$C6
	.dc.b	$8D,$7B,$79,$1C,$68,$D2,$47,$F8,$E6,$1E,$36,$C9,$19,$7C,$5C,$DF
	.dc.b	$E0,$68,$FF,$20,$35,$FF,$22,$35,$FF,$24,$35,$F7,$71,$78,$C8,$F8
	.dc.b	$5F,$77,$AC,$F2,$29,$27,$FC,$9D,$19,$7F,$CA,$1F,$39,$BF,$46,$D1
	.dc.b	$CF,$88,$F8,$6C,$C7,$F9,$5F,$D9,$D7,$27,$F4,$FC,$FF,$C8,$9C,$5F
	.dc.b	$E1,$69,$CE,$7F,$E0,$5E,$3F,$F5,$75,$FF,$8C,$E7,$35,$FE,$1C,$9C
	.dc.b	$D7,$F8,$2E,$73,$5F,$E6,$D9,$CD,$75,$C3,$BC,$E7,$0B,$87,$F9,$DB
	.dc.b	$0E,$2F,$F9,$E3,$0E,$35,$FE,$77,$9C,$CF,$E5,$2B,$8B,$FE,$06,$BB
	.dc.b	$7C,$23,$A4,$62,$77,$5C,$AB,$55,$C6,$FE,$21,$EF,$BE,$0D,$33,$1E
	.dc.b	$F8,$CB,$63,$38,$7B,$DA,$7A,$67,$71,$99,$F9,$D8,$79,$27,$35,$B3
	.dc.b	$8C,$FA,$9A,$F9,$CE,$2F,$95,$A7,$8C,$B5,$DC,$FB,$15,$DE,$79,$E1
	.dc.b	$F0,$E3,$15,$C2,$7C,$4D,$3C,$45,$3C,$55,$D2,$3C,$4D,$8F,$13,$4F
	.dc.b	$14,$7E,$04,$D9,$63,$6D,$8D,$C7,$E0,$E7,$CB,$1C,$E7,$CA,$D9,$F2
	.dc.b	$B5,$F2,$C7,$33,$65,$76,$D7,$67,$77,$CA,$D3,$CA,$58,$F2,$BF,$93
	.dc.b	$8D,$B4,$FA,$47,$FB,$2E,$BF,$5A,$DB,$FF,$42,$BF,$38,$C4,$FE,$DE
	.dc.b	$33,$1F,$40,$C9,$FE,$D3,$9F,$D1,$9F,$A4,$F9,$CF,$EE,$35,$E2,$8C
	.dc.b	$D7,$E9,$A7,$35,$EA,$3D,$E4,$73,$7E,$CA,$E3,$7F,$EB,$39,$CB,$F2
	.dc.b	$AB,$E4,$90,$46,$9C,$4F,$27,$FD,$21,$19,$7C,$31,$9A,$FF,$53,$CE
	.dc.b	$63,$15,$DE,$BA,$F8,$9F,$82,$8C,$4F,$C1,$8C,$4F,$83,$0F,$81,$18
	.dc.b	$8E,$F1,$8A,$F8,$CE,$6F,$C0,$EE,$FA,$4E,$8F,$BB,$3A,$71,$5E,$CB
	.dc.b	$E1,$7C,$DB,$3C,$1B,$3F,$46,$B9,$3F,$4E,$FB,$B6,$F9,$5F,$6C,$3F
	.dc.b	$1F,$7C,$1F,$AD,$3E,$BA,$E1,$5D,$CE,$07,$C9,$AE,$0A,$24,$3D,$1F
	.dc.b	$C7,$C7,$DA,$6C,$66,$78,$4C,$34,$D7,$9C,$FE,$3C,$FC,$87,$85,$67
	.dc.b	$F1,$5E,$87,$42,$3C,$6B,$F1,$31,$E0,$F0,$24,$E7,$0F,$F7,$EB,$63
	.dc.b	$CB,$F1,$12,$34,$DB,$F2,$52,$3F,$44,$7E,$8F,$C0,$B5,$F0,$F4,$FF
	.dc.b	$7B,$FC,$CB,$E3,$F5,$39,$3E,$0C,$65,$FD,$25,$7F,$52,$87,$C2,$D3
	.dc.b	$F6,$1C,$16,$BF,$55,$DF,$86,$9F,$AD,$8F,$EB,$7F,$38,$F6,$2F,$B5
	.dc.b	$86,$D7,$8F,$7F,$B5,$5B,$8C,$42,$30,$FF,$BF,$DF,$F1,$DC,$69,$CE
	.dc.b	$A8,$FB,$B0,$F5,$FF,$B6,$FC,$E6,$67,$84,$61,$F3,$B1,$E9,$E1,$5C
	.dc.b	$3C,$0F,$B0,$FE,$EF,$F3,$8F,$2F,$15,$E3,$F9,$1F,$82,$7C,$2F,$91
	.dc.b	$B1,$E7,$70,$FB,$A9,$E8,$63,$1F,$77,$17,$CA,$3E,$5F,$C8,$DB,$F8
	.dc.b	$0E,$1F,$85,$F2,$36,$9C,$1F,$82,$E1,$FB,$87,$D4,$F0,$AD,$FC,$97
	.dc.b	$0F,$CB,$AA,$7C,$F3,$ED,$9F,$03,$C4,$B1,$F5,$BE,$71,$F3,$1F,$E1
	.dc.b	$A6,$C6,$7E,$73,$13,$96,$47,$C1,$E8,$B8,$7A,$23,$F6,$B6,$F7,$FA
	.dc.b	$1D,$7E,$02,$CF,$85,$C3,$E3,$F4,$2D,$F3,$3F,$83,$5F,$FB,$46,$7D
	.dc.b	$1E,$F7,$9F,$DF,$F4,$36,$32,$7B,$63,$2F,$E5,$3E,$71,$F0,$BB,$C3
	.dc.b	$B0,$8C,$CF,$13,$F3,$7F,$38,$F8,$4F,$F1,$FB,$63,$3E,$C6,$27,$3E
	.dc.b	$7E,$BB,$2D,$72,$78,$E3,$CB,$96,$CF,$67,$FC,$97,$F3,$8F,$8E,$BF
	.dc.b	$03,$86,$D7,$EB,$A7,$F3,$4D,$F0,$78,$F9,$BC,$9E,$7C,$F2,$78,$7A
	.dc.b	$97,$3F,$83,$F3,$8E,$AF,$F2,$DB,$6B,$F7,$6F,$E1,$27,$73,$AE,$6D
	.dc.b	$FC,$5D,$BD,$BC,$BD,$4E,$7F,$15,$67,$9B,$FE,$68,$F2,$5C,$77,$C8
	.dc.b	$C3,$6B,$F8,$5C,$FF,$A2,$DB,$C9,$C7,$E4,$F9,$3F,$E7,$2F,$9C,$7C
	.dc.b	$55,$31,$F1,$D6,$FF,$8A,$FC,$E7,$F1,$07,$FA,$5A,$F4,$7F,$CF,$6E
	.dc.b	$78,$BF,$AB,$39,$9E,$3F,$1B,$A3,$7F,$11,$63,$AF,$CA,$F2,$39,$F9
	.dc.b	$B6,$3F,$92,$FC,$E3,$D6,$FF,$44,$B6,$33,$D7,$89,$FF,$46,$FA,$9F
	.dc.b	$C8,$E7,$F9,$6B,$F9,$1D,$3F,$32,$E5,$BE,$1F,$0F,$67,$8F,$89,$6F
	.dc.b	$8B,$FE,$98,$F9,$C7,$8D,$4C,$5D,$AF,$E7,$73,$D7,$D2,$FE,$6D,$8F
	.dc.b	$F7,$98,$C4,$FF,$74,$7F,$DB,$EF,$0E,$9E,$D9,$C8,$E7,$F0,$B6,$FF
	.dc.b	$63,$6E,$3F,$21,$E5,$EC,$B7,$E2,$70,$F9,$3B,$2D,$F2,$EC,$F6,$1E
	.dc.b	$DE,$CB,$5E,$A7,$0F,$F5,$6F,$41,$F2,$78,$BF,$E0,$7B,$BE,$4F,$FD
	.dc.b	$7A,$DB,$FE,$BA,$BE,$67,$33,$A6,$4F,$5F,$81,$B7,$FC,$38,$7B,$1F
	.dc.b	$EC,$86,$CF,$F6,$85,$D7,$99,$D3,$27,$5D,$96,$BA,$3C,$CF,$F6,$9E
	.dc.b	$9F,$E0,$F9,$D9,$D7,$6F,$01,$E1,$DA,$C7,$85,$C3,$E3,$B7,$CD,$FF
	.dc.b	$8D,$FE,$73,$FC,$1B,$1F,$E3,$2A,$F5,$3A,$3F,$C5,$45,$3F,$C5,$93
	.dc.b	$87,$07,$F8,$EB,$E6,$DE,$43,$E7,$3F,$C6,$51,$86,$C6,$67,$FC,$7D
	.dc.b	$39,$6F,$95,$FD,$A1,$BC,$7C,$A3,$11,$FE,$47,$94,$62,$3C,$5E,$B7
	.dc.b	$F9,$32,$BF,$3B,$0A,$FC,$13,$FE,$1C,$6B,$FD,$21,$4F,$1D,$3D,$9C
	.dc.b	$96,$BF,$CB,$DB,$BE,$4F,$11,$FD,$95,$B8,$74,$C7,$C2,$FE,$36,$7C
	.dc.b	$1F,$2D,$A3,$7C,$D6,$3C,$9F,$2F,$C8,$C3,$E5,$7B,$E2,$F4,$E9,$B9
	.dc.b	$FF,$1B,$FC,$03,$9B,$CA,$B9,$A9,$88,$B3,$FA,$9F,$5B,$A9,$87,$FB
	.dc.b	$5F,$CD,$FC,$47,$FC,$86,$FB,$78,$7F,$CF,$2D,$8F,$F4,$13,$B7,$8F
	.dc.b	$0F,$C6,$F9,$CF,$F2,$94,$7F,$93,$1A,$E8,$FF,$27,$FC,$DB,$DA,$F9
	.dc.b	$AF,$F6,$98,$CE,$38,$9C,$BA,$EF,$93,$A3,$9F,$C6,$72,$E6,$91,$83
	.dc.b	$BB,$DB,$E7,$24,$60,$F1,$F8,$3E,$2E,$29,$18,$3C,$FB,$FE,$0B,$ED
	.dc.b	$7F,$C2,$71,$FE,$68,$6B,$FE,$14,$D7,$F2,$7D,$3F,$94,$97,$89,$C7
	.dc.b	$C4,$97,$C6,$3F,$1B,$F3,$0F,$A0,$BF,$45,$2B,$FC,$EB,$39,$6F,$B9
	.dc.b	$C3,$D8,$91,$83,$7D,$FD,$3B,$A4,$60,$FF,$3F,$FD,$AB,$8F,$F3,$AB
	.dc.b	$8E,$98,$8F,$D0,$FD,$A7,$F3,$FA,$FF,$43,$BE,$C6,$ED,$AF,$33,$F4
	.dc.b	$FF,$68,$FA,$4B,$C3,$2F,$FD,$1F,$39,$6F,$8F,$C3,$C7,$91,$83,$7E
	.dc.b	$37,$4E,$09,$18,$3F,$D3,$31,$FB,$76,$EB,$FB,$88,$F7,$46,$BA,$F1
	.dc.b	$EC,$67,$D8,$93,$CA,$3F,$7F,$F6,$7F,$E9,$B8,$FF,$52,$AF,$53,$26
	.dc.b	$BD,$5D,$F9,$1C,$3C,$99,$7C,$23,$F8,$4F,$D9,$BE,$32,$FC,$64,$AF
	.dc.b	$F5,$8C,$E5,$BE,$87,$0E,$39,$18,$3F,$D6,$FF,$67,$FA,$77,$FC,$C6
	.dc.b	$FF,$64,$73,$E4,$FF,$18,$8D,$78,$BF,$C9,$97,$CB,$67,$C3,$97,$D2
	.dc.b	$3F,$8F,$FC,$43,$C5,$BF,$1F,$2B,$FD,$8F,$39,$6C,$70,$F1,$A4,$60
	.dc.b	$DB,$FC,$AB,$E7,$7F,$D7,$51,$FE,$D0,$6B,$FE,$7C,$AF,$F2,$E3,$DB
	.dc.b	$D4,$F1,$F1,$48,$C1,$E2,$7F,$9A,$FC,$E7,$E5,$63,$FD,$0A,$DD,$BF
	.dc.b	$A2,$DF,$CD,$9E,$DC,$1E,$EE,$F9,$1E,$C8,$AF,$FA,$3B,$E8,$3F,$CA
	.dc.b	$AF,$FB,$99,$AF,$9E,$7E,$3F,$81,$F2,$3C,$99,$3F,$1A,$2B,$ED,$FA
	.dc.b	$5F,$9A,$63,$2F,$2B,$FD,$3B,$0F,$E9,$53,$E1,$F0,$76,$FF,$BD,$BE
	.dc.b	$8D,$F3,$FF,$D5,$1F,$E5,$51,$9F,$57,$13,$97,$FA,$CF,$D2,$FF,$B8
	.dc.b	$27,$FD,$FE,$D7,$FD,$57,$5C,$9B,$EA,$F5,$F1,$A4,$78,$62,$BF,$EB
	.dc.b	$2F,$A2,$7C,$57,$F6,$66,$C7,$9E,$BC,$DE,$BF,$9F,$E4,$A4,$FA,$E2
	.dc.b	$BE,$2F,$A5,$F6,$18,$E1,$FE,$FB,$9A,$3F,$B9,$FD,$67,$FB,$8A,$87
	.dc.b	$FB,$11,$C3,$CC,$F9,$1D,$FF,$1F,$C0,$91,$83,$E6,$3F,$DF,$BE,$9B
	.dc.b	$F9,$DE,$3F,$E0,$C6,$FB,$7D,$A9,$92,$FE,$97,$9D,$F4,$EA,$57,$D7
	.dc.b	$8A,$FF,$B6,$38,$9F,$37,$F5,$0C,$E5,$FE,$1A,$FA,$6F,$EC,$ED,$EF
	.dc.b	$FF,$08,$C1,$BF,$3B,$87,$3F,$1D,$31,$F8,$77,$C5,$B9,$FA,$70,$FF
	.dc.b	$44,$C6,$0F,$93,$CB,$FE,$5E,$70,$1C,$3F,$EF,$08,$E2,$DF,$43,$CB
	.dc.b	$F1,$B2,$38,$CF,$F2,$76,$C7,$E7,$7C,$59,$18,$37,$EF,$7B,$BE,$3E
	.dc.b	$46,$0E,$6F,$EF,$BD,$B9,$18,$3C,$FF,$4D,$F3,$3E,$63,$EB,$47,$D7
	.dc.b	$8C,$38,$0B,$C9,$7C,$66,$67,$3C,$8B,$C7,$99,$F9,$1E,$1B,$DB,$FE
	.dc.b	$50,$BF,$C2,$4E,$5E,$CD,$7C,$6B,$E7,$F8,$47,$E6,$23,$15,$DA,$73
	.dc.b	$29,$CB,$E9,$FB,$7F,$F7,$F3,$7C,$FD,$C6,$23,$E0,$B5,$FC,$C2,$FA
	.dc.b	$BD,$93,$0F,$81,$19,$6C,$7F,$B2,$39,$FF,$CD,$B3,$9A,$46,$3D,$5D
	.dc.b	$5F,$6B,$B5,$FB,$4B,$FF,$58,$4F,$E3,$1F,$F5,$A4,$72,$8F,$2F,$95
	.dc.b	$C2,$13,$AC,$3C,$33,$98,$C5,$70,$3A,$7C,$56,$67,$4D,$8C,$9E,$18
	.dc.b	$CB,$4C,$DF,$8C,$F4,$46,$9C,$1E,$47,$17,$E4,$9E,$2D,$9D,$DF,$99
	.dc.b	$E7,$3A,$3D,$0E,$AF,$CC,$D8,$F2,$34,$E7,$7E,$08,$E3,$7F,$52,$B8
	.dc.b	$B6,$38,$B6,$B8,$C7,$85,$F1,$9D,$63,$D0,$6C,$B1,$F8,$89,$EA,$D3
	.dc.b	$A9,$67,$A9,$CD,$F2,$99,$9E,$71,$F8,$39,$E6,$D9,$E6,$D3,$9C,$F8
	.dc.b	$E3,$9C,$66,$39,$BF,$86,$9E,$6D,$9E,$6D,$3C,$A7,$E9,$2B,$F4,$E4
	.dc.b	$8F,$D0,$B6,$3F,$5C,$D7,$84,$74,$CF,$FD,$ED,$1E,$68,$F2,$BF,$85
	.dc.b	$7F,$C3,$9F,$4D,$E4,$C7,$F4,$33,$B3,$F8,$04,$66,$FD,$37,$D3,$1F
	.dc.b	$6D,$F0,$7D,$F1,$9C,$3B,$CE,$5F,$0C,$67,$0F,$BB,$5C,$AF,$F5,$93
	.dc.b	$99,$F3,$1E,$32,$CF,$FB,$AA,$35,$7D,$67,$9D,$F6,$C3,$D0,$FC,$36
	.dc.b	$C7,$3A,$E3,$86,$30,$E9,$56,$F9,$1F,$5B,$0E,$D8,$78,$A3,$F9,$5C
	.dc.b	$EE,$3C,$27,$C8,$8E,$17,$F2,$27,$53,$C4,$D1,$ED,$AE,$A7,$6C,$3C
	.dc.b	$B7,$FC,$4F,$CF,$F8,$BE,$2F,$F0,$D6,$C6,$4F,$E4,$AF,$F2,$E2,$3F
	.dc.b	$CA,$3D,$0F,$E6,$24,$70,$4C,$3F,$81,$C7,$D2,$F1,$3F,$8D,$3F,$C6
	.dc.b	$FD,$0F,$E4,$1F,$88,$3E,$77,$18,$7F,$72,$8C,$3F,$CA,$67,$F9,$F9
	.dc.b	$1F,$E7,$2F,$F4,$8F,$80,$FE,$B1,$1F,$DE,$9B,$8F,$F1,$AF,$A8,$FF
	.dc.b	$24,$CE,$7A,$3F,$98,$46,$27,$5E,$9F,$F3,$3F,$6F,$FA,$7F,$23,$0F
	.dc.b	$F1,$15,$EE,$31,$7F,$E3,$28,$7D,$4F,$43,$AF,$A9,$6F,$FC,$1B,$28
	.dc.b	$CC,$7F,$84,$3E,$A5,$E2,$63,$0F,$47,$C5,$79,$8F,$3F,$A7,$6F,$9B
	.dc.b	$87,$AF,$A9,$6B,$E0,$3E,$E3,$FA,$77,$D3,$BD,$8C,$57,$93,$B3,$E1
	.dc.b	$9F,$13,$AA,$C7,$A1,$C3,$E7,$F9,$CB,$8F,$CD,$E5,$F9,$67,$CC,$F5
	.dc.b	$2D,$7C,$07,$0F,$F8,$BB,$EA,$3E,$53,$F3,$23,$FB,$17,$D4,$BE,$86
	.dc.b	$98,$F5,$AD,$7F,$8D,$A5,$19,$8C,$7A,$D7,$2F,$EB,$35,$FE,$49,$9C
	.dc.b	$9F,$E3,$ED,$DC,$3E,$0B,$F0,$FC,$44,$AE,$DE,$AD,$C7,$9B,$E6,$3E
	.dc.b	$5F,$1D,$9F,$27,$A3,$EA,$3D,$7C,$D6,$3D,$0E,$1F,$17,$99,$70,$F3
	.dc.b	$FC,$BF,$C2,$3E,$1F,$CE,$DA,$F8,$0E,$1F,$85,$F3,$B6,$BD,$EF,$CA
	.dc.b	$3E,$FF,$8F,$67,$E4,$7A,$5F,$38,$F9,$F5,$63,$E6,$38,$7E,$37,$A5
	.dc.b	$63,$FC,$07,$F5,$2F,$86,$FF,$0C,$36,$BF,$CD,$2F,$EA,$1F,$F4,$43
	.dc.b	$99,$E3,$E9,$F0,$3F,$27,$D7,$6B,$8F,$B1,$D0,$EB,$F0,$D6,$3A,$B8
	.dc.b	$7B,$FC,$55,$C3,$CA,$F9,$0F,$CF,$C5,$3F,$CF,$BF,$88,$BB,$BF,$C6
	.dc.b	$0D,$8C,$98,$8F,$F1,$A6,$CF,$93,$A6,$39,$AD,$7F,$A2,$3E,$D5,$F2
	.dc.b	$9F,$E3,$B6,$C6,$7E,$53,$13,$9E,$29,$3C,$78,$6E,$DF,$E3,$0C,$3F
	.dc.b	$A2,$93,$D5,$E9,$EC,$7F,$91,$FE,$7D,$F1,$34,$C7,$CA,$5C,$BF,$D3
	.dc.b	$73,$FE,$B7,$6F,$4F,$1E,$9E,$4F,$F9,$47,$ED,$1E,$15,$31,$E1,$5B
	.dc.b	$FF,$52,$CB,$E4,$F8,$1C,$7E,$52,$CE,$B8,$3A,$1F,$E5,$D7,$E1,$BF
	.dc.b	$E4,$CA,$FF,$6C,$C7,$FA,$C0,$7C,$0E,$CE,$1F,$23,$63,$FD,$6B,$F7
	.dc.b	$AF,$9D,$FF,$34,$36,$33,$ED,$62,$7F,$CD,$9F,$6A,$FA,$5A,$63,$7B
	.dc.b	$7F,$EC,$09,$78,$DC,$0E,$3B,$D9,$FF,$62,$FD,$A3,$C5,$FF,$3C,$36
	.dc.b	$33,$E2,$E2,$73,$F0,$F2,$78,$F0,$5E,$2F,$F3,$6F,$C0,$7F,$33,$9F
	.dc.b	$F6,$8B,$63,$3B,$FF,$A1,$63,$1E,$FF,$F5,$09,$FF,$5B,$7B,$39,$35
	.dc.b	$EC,$DC,$3F,$DB,$7F,$4C,$F9,$0F,$F4,$73,$63,$3C,$98,$9C,$F2,$49
	.dc.b	$E3,$D5,$72,$FF,$41,$E0,$8E,$67,$4F,$8E,$B5,$E0,$F2,$7B,$9E,$0F
	.dc.b	$26,$C7,$91,$C3,$E6,$F4,$EE,$3D,$BB,$BC,$E7,$A3,$CB,$B5,$EC,$70
	.dc.b	$FF,$BE,$39,$9D,$98,$BF,$F4,$CF,$1F,$FA,$4D,$FF,$5C,$FC,$57,$FA
	.dc.b	$BE,$3F,$C6,$26,$4B,$E3,$7F,$AC,$ED,$EF,$EE,$3F,$D6,$D6,$F4,$E9
	.dc.b	$8E,$FB,$9A,$7F,$D5,$27,$E3,$1C,$63,$FE,$E9,$9E,$7D,$FC,$67,$27
	.dc.b	$FB,$E9,$CB,$ED,$E9,$B7,$D3,$7F,$19,$E4,$DE,$D7,$29,$F1,$1F,$EF
	.dc.b	$E7,$0F,$96,$3D,$6B,$C6,$F4,$7D,$33,$D5,$E3,$3C,$4F,$59,$B3,$27
	.dc.b	$69,$AD,$29,$B3,$9F,$D2,$FE,$FB,$E9,$B6,$D8,$E6,$D7,$7E,$0A,$77
	.dc.b	$E3,$FF,$BC,$A7,$6D,$7A,$B5,$F8,$34,$8C,$47,$D5,$87,$CD,$7C,$6E
	.dc.b	$1F,$D6,$43,$9B,$1B,$72,$6D,$F1,$BA,$AB,$83,$F9,$3F,$37,$3C,$FF
	.dc.b	$CD,$A3,$E8,$CE,$9F,$F7,$6A,$E6,$E6,$BC,$EA,$62,$2C,$F2,$F8,$CD
	.dc.b	$98,$79,$BE,$0F,$76,$46,$CF,$C0,$F1,$7F,$39,$6F,$2F,$AA,$0D,$E7
	.dc.b	$C3,$EA,$F1,$E5,$70,$3F,$17,$F4,$DF,$29,$BC,$1F,$12,$3D,$DC,$5F
	.dc.b	$C5,$DD,$4A,$31,$3F,$2B,$E9,$BF,$C7,$73,$F9,$66,$BB,$E3,$FC,$DD
	.dc.b	$F4,$7F,$3B,$F4,$DF,$E5,$B9,$FC,$EB,$FD,$15,$B8,$FD,$68,$D4,$73
	.dc.b	$3F,$6C,$ED,$C0,$7F,$85,$BD,$3B,$F9,$4F,$3F,$89,$E9,$CB,$2B,$C3
	.dc.b	$15,$E7,$6F,$53,$C4,$67,$5F,$E7,$51,$FC,$3A,$7D,$3C,$3E,$73,$D1
	.dc.b	$E0,$FA,$74,$FC,$7F,$D3,$3E,$BE,$99,$BF,$DE,$C3,$C5,$DF,$8B,$27
	.dc.b	$31,$F9,$8F,$55,$ED,$DF,$A0,$95,$FC,$1A,$72,$DF,$6B,$87,$D3,$C8
	.dc.b	$C1,$BE,$EE,$9D,$B2,$30,$7F,$10,$FA,$7F,$84,$DE,$0F,$D4,$C7,$C1
	.dc.b	$8F,$A2,$E9,$FA,$7F,$4F,$FE,$31,$7F,$8C,$BE,$D6,$EB,$FB,$28,$D4
	.dc.b	$7D,$AF,$A7,$FF,$19,$BF,$C8,$1F,$5B,$9D,$BF,$73,$1A,$8E,$1E,$2E
	.dc.b	$3C,$59,$1C,$63,$7E,$1F,$0D,$64,$F5,$8F,$E0,$67,$EF,$DF,$C5,$BE
	.dc.b	$BE,$8F,$EF,$0D,$E6,$FE,$6A,$EE,$33,$F0,$9C,$BE,$0A,$5F,$F3,$8A
	.dc.b	$4F,$F1,$0F,$A7,$FF,$1B,$CF,$F3,$C7,$FB,$A3,$7D,$3D,$7A,$BB,$F1
	.dc.b	$F8,$78,$F2,$B8,$47,$F1,$AF,$A7,$7B,$97,$E1,$A5,$7F,$48,$9C,$B7
	.dc.b	$CF,$E1,$C5,$23,$07,$F4,$EF,$A7,$FD,$67,$0F,$F8,$5D,$FD,$5C,$7E
	.dc.b	$B9,$B1,$9F,$23,$F9,$5C,$3D,$CB,$E0,$4B,$E9,$1F,$CC,$3E,$75,$E1
	.dc.b	$5F,$86,$95,$FD,$7A,$72,$D8,$E1,$E1,$C8,$C1,$B7,$F9,$DF,$CD,$BE
	.dc.b	$32,$9B,$9F,$E0,$B7,$FE,$41,$7B,$71,$78,$CF,$27,$A9,$63,$FB,$7F
	.dc.b	$C6,$FE,$53,$83,$FC,$70,$DE,$DF,$EE,$B8,$7C,$CA,$EB,$C5,$ED,$EF
	.dc.b	$91,$EB,$8A,$FF,$16,$FA,$77,$D5,$7F,$52,$6C,$7C,$1A,$F8,$DE,$0F
	.dc.b	$C7,$F2,$24,$FC,$58,$AF,$B3,$E6,$FE,$61,$8D,$7F,$C9,$D9,$A3,$B7
	.dc.b	$85,$2B,$E7,$47,$F6,$2F,$A5,$7C,$E5,$F9,$99,$5F,$E1,$09,$CB,$63
	.dc.b	$FB,$4F,$D2,$BE,$52,$9C,$27,$F9,$9D,$3D,$6E,$DE,$BC,$9E,$91,$5F
	.dc.b	$E7,$3F,$4A,$FA,$8F,$EE,$AD,$8F,$15,$79,$FD,$7F,$47,$5C,$9F,$34
	.dc.b	$57,$C3,$F4,$7E,$D3,$1C,$5F,$E7,$1C,$D1,$FE,$02,$FA,$8F,$F7,$25
	.dc.b	$7F,$8B,$9A,$F2,$3E,$17,$85,$F1,$FC,$39,$18,$3E,$33,$FE,$0C,$FA
	.dc.b	$8F,$E7,$D1,$FE,$38,$6B,$F1,$23,$E8,$7A,$1F,$47,$BA,$4F,$CF,$8A
	.dc.b	$FF,$5C,$E3,$7C,$AF,$D0,$33,$AF,$FA,$2F,$EA,$3F,$B7,$B7,$C1,$FF
	.dc.b	$0E,$C1,$BF,$3B,$87,$3F,$21,$31,$FF,$26,$D6,$DB,$1F,$4B,$E3,$64
	.dc.b	$60,$DF,$A7,$DB,$E3,$E4,$60,$E6,$FE,$77,$C7,$91,$83,$BF,$BD,$E6
	.dc.b	$FA,$89,$18,$3C,$6F,$F8,$DB,$3F,$E9,$11,$FD,$F9,$BA,$FC,$19,$FE
	.dc.b	$03,$DB,$FE,$58,$F2,$FF,$21,$7F,$E5,$C6,$BF,$E6,$06,$BF,$E6,$6F
	.dc.b	$A8,$F9,$BF,$29,$FE,$62,$6C,$7F,$99,$97,$8C,$F8,$FD,$72,$2F,$B7
	.dc.b	$81,$FE,$BD,$C3,$FC,$F3,$8F,$FB,$FB,$A3,$24,$8F,$17,$87,$B7,$FD
	.dc.b	$21,$3F,$AB,$9F,$44,$E6,$53,$97,$D5,$F0,$D9,$8F,$F0,$B5,$7B,$E3
	.dc.b	$04,$7F,$D1,$31,$A9,$FF,$13,$FD,$43,$E5,$95,$AB,$B5,$AF,$8C,$BC
	.dc.b	$5F,$ED,$9C,$7F,$D5,$B7,$FE,$94,$C7,$FD,$6D,$18,$BE,$13,$E0,$F8
	.dc.b	$5E,$2F,$FA,$72,$B7,$18,$9F,$F5,$84,$62,$7A,$61,$99,$F8,$35,$6B
	.dc.b	$51,$D8,$FF,$20,$DF,$87,$0F,$F7,$74,$E9,$B7,$CE,$75,$87,$8E,$31
	.dc.b	$87,$56,$C7,$88,$F5,$17,$0E,$F3,$B8,$CC,$78,$EF,$E3,$4E,$63,$73
	.dc.b	$C6,$72,$D3,$F0,$F3,$B8,$CC,$F3,$9C,$9B,$7F,$0F,$3B,$8C,$C6,$CC
	.dc.b	$BB,$7E,$4B,$6F,$E0,$C7,$D0,$BF,$7B,$C4,$D9,$63,$6B,$0F,$99,$5F
	.dc.b	$24,$F8,$38,$7C,$E6,$99,$9F,$82,$73,$C3,$D0,$F9,$8C,$47,$CD,$7F
	.dc.b	$DB,$35,$FE,$BB,$8F,$F6,$AB,$63,$FD,$B2,$D7,$78,$7E,$9E,$31,$1E
	.dc.b	$13,$33,$CA,$7C,$31,$EB,$F3,$2D,$75,$C3,$D5,$3F,$A6,$9C,$D6,$9B
	.dc.b	$7F,$36,$FF,$D8,$4D,$9D,$34,$D4,$F0,$AF,$9D,$5C,$1B,$5C,$1B,$5C
	.dc.b	$0E,$4F,$0A,$E5,$7E,$63,$F3,$EE,$27,$E8,$46,$1E,$2E,$DF,$D9,$46
	.dc.b	$E7,$F0,$51,$FB,$B8,$C4,$7D,$FA,$FD,$84,$62,$FE,$35,$72,$3A,$9E
	.dc.b	$03,$15,$EA,$3D,$0F,$E4,$7C,$2F,$D9,$1F,$BB,$F3,$6C,$7F,$0B,$6C
	.dc.b	$FF,$16,$6E,$3F,$BC,$F3,$BF,$75,$E4,$7F,$03,$6D,$FF,$13,$7F,$04
	.dc.b	$47,$F1,$2F,$F2,$BF,$A7,$7D,$45,$24,$B1,$84,$9D,$C7,$E4,$BD,$0F
	.dc.b	$E2,$91,$89,$FE,$57,$18,$AF,$E3,$1F,$05,$FD,$16,$FA,$65,$FC,$B2
	.dc.b	$FF,$D0,$7C,$DF,$C7,$7E,$9F,$FC,$63,$5A,$3C,$37,$C6,$BF,$DB,$B3
	.dc.b	$A6,$FA,$5F,$C9,$7E,$35,$EC,$F3,$8E,$9F,$51,$31,$E9,$7E,$DC,$3F
	.dc.b	$C6,$91,$88,$42,$7C,$9E,$04,$AD,$7D,$43,$C4,$C6,$1E,$4E,$9F,$51
	.dc.b	$EB,$E4,$B7,$D7,$24,$66,$31,$E9,$DB,$F6,$BF,$0C,$FE,$81,$F4,$8F
	.dc.b	$5F,$15,$CF,$E6,$3E,$41,$F2,$7A,$AC,$77,$70,$FA,$3E,$A2,$E3,$F2
	.dc.b	$B9,$3E,$81,$F4,$7D,$2B,$5E,$77,$0F,$C5,$FA,$8F,$A1,$D1,$F6,$77
	.dc.b	$57,$F5,$5F,$A7,$7D,$45,$31,$EA,$5C,$7E,$74,$A3,$31,$8F,$3A,$E8
	.dc.b	$CB,$F0,$D7,$FE,$39,$AD,$1F,$5F,$87,$E5,$B8,$AE,$0F,$E6,$25,$F1
	.dc.b	$7E,$53,$CD,$E9,$59,$F1,$B8,$7A,$72,$5B,$E6,$7E,$8E,$DF,$21,$EF
	.dc.b	$3E,$07,$45,$BF,$6B,$87,$CD,$E5,$DB,$E6,$2D,$F3,$1F,$24,$F9,$5E
	.dc.b	$AD,$BF,$8E,$E1,$F9,$3F,$3B,$6F,$9D,$A3,$E9,$1F,$88,$F6,$AD,$7D
	.dc.b	$0F,$4B,$EB,$1F,$5E,$EC,$7D,$47,$0F,$CD,$F4,$AC,$7A,$3E,$A1,$F0
	.dc.b	$FF,$E1,$36,$D7,$EA,$5F,$D8,$B6,$35,$E3,$48,$FB,$1D,$96,$FA,$7B
	.dc.b	$1C,$CE,$9F,$0F,$6B,$A3,$87,$B7,$C6,$DA,$F2,$3E,$33,$F4,$F1,$4F
	.dc.b	$DE,$7D,$47,$E5,$9B,$DD,$FE,$2B,$8F,$F1,$FC,$60,$FE,$05,$B3,$E8
	.dc.b	$3F,$C6,$0D,$8C,$F3,$E2,$7F,$C6,$9F,$4E,$FA,$8A,$63,$E5,$AD,$FF
	.dc.b	$0D,$9F,$8F,$A7,$E0,$9B,$DF,$FC,$8D,$F9,$31,$9E,$2E,$BE,$BC,$BE
	.dc.b	$5E,$B5,$AF,$C5,$FD,$0F,$F9,$B5,$BE,$EF,$F9,$1E,$3F,$CB,$D9,$7F
	.dc.b	$2F,$74,$DE,$9D,$1B,$E9,$B8,$FF,$23,$FA,$8F,$D4,$B7,$C3,$FF,$29
	.dc.b	$43,$C3,$B1,$8F,$95,$95,$C7,$E5,$AE,$BF,$A0,$C3,$FA,$64,$F5,$8F
	.dc.b	$F2,$E3,$F2,$4C,$1F,$CC,$C7,$CA,$E8,$DF,$C9,$58,$FE,$71,$F5,$0F
	.dc.b	$96,$FF,$33,$36,$33,$ED,$62,$7F,$CD,$5F,$50,$FA,$7A,$63,$6B,$7F
	.dc.b	$D0,$E7,$F5,$8D,$F1,$B8,$EF,$C9,$FF,$3A,$7D,$43,$C4,$A6,$3C,$4B
	.dc.b	$7F,$D3,$67,$FD,$1C,$DF,$87,$E3,$C1,$C9,$FF,$3F,$7B,$FF,$CF,$E7
	.dc.b	$FA,$A3,$78,$BF,$84,$BE,$D6,$FB,$3C,$3D,$99,$5C,$3D,$9B,$5F,$C4
	.dc.b	$3E,$A1,$F1,$BF,$E8,$B6,$D7,$F6,$17,$E2,$36,$35,$C5,$23,$5D,$56
	.dc.b	$FA,$3C,$CE,$9F,$1B,$67,$A7,$8F,$DC,$F0,$79,$16,$3C,$0E,$1F,$17
	.dc.b	$A9,$70,$F4,$77,$79,$CF,$47,$95,$6B,$D0,$E1,$FE,$F5,$CC,$EC,$C5
	.dc.b	$FF,$30,$F0,$BF,$C6,$EF,$F8,$4B,$E2,$3F,$D5,$91,$FE,$51,$3F,$C2
	.dc.b	$86,$A3,$FA,$97,$8F,$A3,$FD,$65,$7F,$E8,$29,$E1,$87,$F8,$9A,$9E
	.dc.b	$9F,$FA,$DD,$B1,$FE,$0D,$7F,$16,$7F,$42,$32,$5E,$CE,$07,$1E,$DB
	.dc.b	$3C,$5C,$3E,$EE,$7B,$5C,$F6,$F0,$1E,$1D,$AD,$78,$5C,$3E,$39,$78
	.dc.b	$DF,$E3,$FF,$A8,$FE,$A7,$1F,$E4,$0A,$F5,$3A,$3D,$C7,$F9,$12,$6B
	.dc.b	$4A,$6C,$F8,$9F,$4F,$FB,$AF,$A8,$DB,$63,$9B,$5D,$F7,$A7,$7E,$37
	.dc.b	$F8,$0E,$76,$D7,$AB,$5F,$F2,$BD,$23,$11,$FE,$5B,$87,$CC,$FF,$97
	.dc.b	$38,$7F,$CA,$50,$E4,$C6,$BC,$9B,$7C,$6E,$AA,$E0,$FE,$43,$CB,$CF
	.dc.b	$21,$DC,$EA,$7F,$C0,$2B,$97,$3E,$64,$9F,$F2,$94,$E5,$B1,$CB,$E2
	.dc.b	$E4,$60,$DB,$E0,$F8,$39,$18,$3F,$CB,$9E,$D3,$9F,$D2,$B9,$A9,$88
	.dc.b	$B3,$FA,$AF,$2F,$D6,$61,$FF,$0C,$FD,$3F,$CA,$7F,$CD,$0F,$B7,$8B
	.dc.b	$FC,$F8,$DE,$2F,$F4,$4B,$B9,$E3,$C5,$F2,$BE,$A3,$FC,$85,$1F,$E7
	.dc.b	$16,$BA,$3F,$CE,$9F,$51,$FE,$61,$6E,$3F,$4A,$3F,$A0,$46,$BC,$8C
	.dc.b	$4E,$7B,$E4,$F0,$39,$FC,$4F,$1E,$59,$18,$3B,$BD,$7E,$46,$46,$0F
	.dc.b	$1F,$C4,$F8,$78,$E4,$60,$F3,$EF,$F8,$6F,$A7,$7D,$45,$3D,$11,$9F
	.dc.b	$1B,$5E,$34,$BF,$F4,$95,$FF,$93,$9F,$F4,$A7,$AE,$F6,$BF,$22,$BE
	.dc.b	$86,$98,$8C,$FC,$0E,$BE,$A2,$5F,$28,$E3,$DD,$CB,$BA,$4F,$68,$FC
	.dc.b	$EF,$D4,$38,$AF,$14,$AF,$F5,$34,$E5,$FF,$54,$7D,$47,$F8,$9D,$BC
	.dc.b	$1F,$A6,$8F,$54,$7F,$98,$1D,$3F,$E6,$2F,$A8,$FF,$14,$3F,$EB,$17
	.dc.b	$DC,$DD,$7F,$5F,$1A,$8E,$1E,$36,$3C,$69,$1C,$63,$7E,$2F,$0D,$64
	.dc.b	$F5,$8F,$DC,$9F,$B5,$7F,$D8,$0F,$AD,$FF,$3A,$B7,$C1,$FF,$6B,$3B
	.dc.b	$8C,$FC,$2F,$2F,$84,$91,$CA,$3F,$82,$FD,$47,$FA,$42,$3F,$D9,$8F
	.dc.b	$F8,$9D,$BE,$A6,$BD,$5D,$F8,$FC,$3C,$89,$5C,$23,$F8,$7F,$D4,$3D
	.dc.b	$DB,$F0,$F2,$BF,$DB,$13,$96,$FA,$1C,$38,$64,$60,$FF,$6F,$7D,$47
	.dc.b	$EA,$1F,$F4,$9B,$FD,$9D,$CF,$8F,$FC,$7A,$35,$C5,$FE,$F7,$5E,$E6
	.dc.b	$7C,$19,$7D,$23,$F9,$2F,$CE,$7E,$56,$3F,$DD,$8F,$FA,$21,$BF,$0D
	.dc.b	$AF,$0F,$6F,$0F,$0A,$4E,$8E,$2F,$F3,$0F,$9C,$7C,$75,$37,$3F,$EA
	.dc.b	$CC,$3F,$C6,$4F,$83,$8B,$C6,$79,$3D,$5B,$1F,$EF,$FF,$90,$FC,$9E
	.dc.b	$EF,$29,$32,$73,$18,$86,$1F,$3D,$F0,$71,$7C,$0E,$F9,$1F,$02,$2B
	.dc.b	$FE,$C0,$FA,$77,$D4,$7F,$4A,$6C,$7C,$6A,$F9,$3E,$0F,$CA,$F2,$64
	.dc.b	$FC,$A8,$AF,$BF,$E7,$3E,$79,$8D,$7D,$51,$EA,$CD,$3D,$FC,$39,$3F
	.dc.b	$52,$3F,$AD,$FD,$33,$E7,$2F,$AB,$2B,$E1,$E3,$D1,$56,$DC,$F0,$7C
	.dc.b	$58,$FE,$6D,$1F,$ED,$EF,$A3,$7C,$AF,$F6,$86,$C7,$03,$E7,$CF,$0B
	.dc.b	$EC,$BD,$6F,$2F,$AD,$2B,$81,$E2,$8C,$3F,$EE,$CF,$A7,$7D,$47,$F7
	.dc.b	$66,$C7,$0B,$FD,$D6,$1D,$BD,$8F,$83,$D7,$2B,$51,$5F,$F7,$BD,$3E
	.dc.b	$A3,$E3,$99,$BF,$3B,$FE,$04,$FA,$8F,$CF,$57,$FB,$F9,$B3,$A8,$F1
	.dc.b	$78,$5F,$43,$C3,$91,$83,$CC,$FF,$83,$BE,$A3,$FA,$4C,$7C,$06,$BE
	.dc.b	$48,$FB,$3E,$87,$DA,$E3,$93,$EF,$8A,$FD,$3F,$95,$FA,$46,$31,$FB
	.dc.b	$9F,$51,$FD,$C1,$BE,$0F,$F8,$7A,$0D,$F9,$DC,$39,$F8,$F9,$87,$D4
	.dc.b	$AD,$B6,$3E,$97,$C6,$C8,$C1,$BF,$4F,$B7,$C7,$C8,$C1,$CD,$FD,$37
	.dc.b	$8F,$23,$07,$7F,$80,$F3,$7D,$44,$8C,$1E,$37,$FC,$6F,$9F,$F5,$2D
	.dc.b	$7F,$34,$DE,$0F,$DE,$6A,$AF,$E1,$DE,$07,$F9,$6B,$CB,$FC,$F6,$3F
	.dc.b	$E6,$06,$BF,$E7,$F6,$BF,$C1,$5F,$D0,$B9,$F0,$79,$11,$FF,$33,$FD
	.dc.b	$43,$8D,$F2,$78,$6C,$FF,$9B,$A7,$F8,$7F,$C7,$66,$BF,$CD,$F8,$7E
	.dc.b	$4F,$4B,$D7,$FB,$5C,$3C,$4D,$7F,$8D,$CE,$DF,$57,$A7,$95,$1F,$EA
	.dc.b	$7A,$EB,$39,$8C,$4F,$09,$FD,$8C,$63,$83,$F8,$23,$FE,$D3,$6D,$7F
	.dc.b	$12,$7F,$84,$7D,$47,$FB,$EA,$BF,$72,$FF,$B8,$AD,$3A,$F5,$78,$57
	.dc.b	$F2,$FA,$D4,$E6,$38,$FC,$AE,$AF,$F9,$BD,$75,$79,$CF,$19,$D4,$27
	.dc.b	$55,$E4,$9C,$C6,$2B,$81,$E6,$BC,$CE,$9B,$19,$3C,$65,$BF,$09,$E4
	.dc.b	$9E,$11,$A8,$C9,$E3,$9E,$6E,$2B,$C4,$F1,$9E,$97,$E7,$69,$E6,$8E
	.dc.b	$15,$CA,$7E,$34,$73,$BF,$0B,$6B,$8B,$D4,$E2,$F9,$67,$89,$DE,$32
	.dc.b	$71,$2B,$93,$71,$E3,$6C,$75,$6B,$F4,$67,$C7,$1C,$67,$EF,$57,$A5
	.dc.b	$A7,$8C,$B5,$E1,$9E,$77,$F8,$49,$CC,$73,$9F,$C1,$D7,$36,$CF,$36
	.dc.b	$9C,$DE,$27,$A4,$EE,$FF,$BE,$EB,$F1,$8D,$9F,$D4,$46,$23,$64,$3A
	.dc.b	$9B,$9F,$F7,$2C,$7D,$C8,$FC,$04,$7E,$AB,$E9,$BE,$1B,$E2,$8F,$2D
	.dc.b	$FE,$B2,$73,$18,$9D,$3F,$B8,$8C,$BD,$0F,$AB,$1F,$C4,$E3,$35,$E1
	.dc.b	$3C,$59,$23,$38,$72,$8C,$57,$E1,$AF,$A6,$1F,$6A,$73,$1D,$8F,$21
	.dc.b	$0F,$24,$F1,$C3,$C1,$7F,$B2,$AE,$37,$F6,$A3,$C1,$87,$8A,$FC,$15
	.dc.b	$A7,$DF,$7C,$9F,$90,$75,$AE,$2D,$38,$D7,$3C,$3C,$C7,$90,$B5,$CD
	.dc.b	$FD,$9D,$F7,$7D,$51,$9C,$3E,$95,$F5,$9E,$75,$B7,$E2,$C7,$F1,$7F
	.dc.b	$2F,$F8,$59,$EF,$F9,$3B,$1F,$C8,$22,$CF,$82,$6D,$FF,$8B,$7E,$A1
	.dc.b	$E8,$7F,$18,$6E,$1F,$E2,$07,$FC,$76,$D8,$D7,$9D,$FC,$EE,$73,$87
	.dc.b	$F9,$5C,$FF,$2D,$B8,$70,$D7,$F9,$74,$FE,$70,$95,$F9,$86,$C7,$F1
	.dc.b	$57,$ED,$78,$5F,$CA,$1F,$E5,$BF,$01,$FD,$6A,$BF,$C6,$AD,$AF,$E3
	.dc.b	$FF,$49,$FE,$4D,$9C,$F2,$7F,$34,$8C,$4E,$BA,$7F,$9B,$F2,$3D,$8D
	.dc.b	$46,$F8,$E6,$3F,$E4,$2C,$3F,$C8,$CD,$8E,$9E,$74,$8F,$F2,$3F,$D3
	.dc.b	$BC,$0C,$67,$E4,$F8,$8F,$11,$E3,$F3,$6D,$F5,$70,$F7,$F3,$AD,$7B
	.dc.b	$1F,$51,$FD,$3B,$E9,$1E,$0E,$2B,$CD,$D1,$EF,$3E,$07,$9B,$63,$D4
	.dc.b	$E1,$F3,$FD,$3D,$C7,$E4,$F2,$FC,$73,$E4,$79,$B6,$BE,$13,$87,$FC
	.dc.b	$CB,$F4,$FF,$33,$B3,$FC,$F1,$C8,$AF,$EC,$7F,$4C,$FA,$7A,$63,$D0
	.dc.b	$B8,$FF,$9B,$A5,$19,$8C,$79,$77,$4F,$EB,$75,$FE,$46,$9C,$9F,$E7
	.dc.b	$6E,$17,$1F,$82,$FC,$3F,$0F,$2B,$B7,$9F,$71,$E6,$F9,$8F,$99,$CB
	.dc.b	$67,$C9,$F1,$9E,$A3,$D7,$D1,$63,$D0,$E1,$F1,$7B,$57,$0F,$3F,$CB
	.dc.b	$FC,$23,$E1,$FA,$76,$BE,$03,$87,$E1,$7C,$E5,$AF,$7B,$F2,$8F,$C0
	.dc.b	$78,$B6,$7E,$47,$9D,$F3,$8F,$9F,$56,$3E,$63,$87,$E3,$7A,$16,$3F
	.dc.b	$C0,$BF,$4E,$F8,$5F,$F0,$CB,$6B,$FD,$42,$FE,$95,$B1,$AF,$1A,$47
	.dc.b	$D1,$ED,$B7,$D3,$D5,$E6,$74,$F8,$5B,$5D,$1C,$3D,$BE,$26,$D7,$91
	.dc.b	$F1,$9F,$9F,$8A,$7F,$AD,$3E,$9F,$F1,$CD,$EE,$FF,$18,$C7,$FB,$B6
	.dc.b	$30,$7F,$AE,$B6,$7C,$FF,$F8,$D9,$B1,$9E,$6C,$4F,$F8,$EB,$E9,$9F
	.dc.b	$4F,$4C,$7C,$95,$BF,$F6,$2C,$FC,$7D,$3F,$C5,$ED,$F1,$3F,$DB,$4F
	.dc.b	$C9,$8C,$F1,$75,$F5,$25,$F2,$F8,$7B,$5F,$E3,$8F,$A0,$7C,$37,$F9
	.dc.b	$31,$B5,$FE,$D4,$5F,$2D,$99,$D7,$54,$AD,$75,$5C,$3F,$C8,$7F,$4E
	.dc.b	$F0,$FF,$CA,$ED,$AF,$F6,$FA,$F0,$F3,$3A,$F9,$29,$5A,$F9,$3B,$87
	.dc.b	$43,$99,$FE,$62,$8F,$92,$74,$71,$EF,$48,$FF,$76,$8F,$90,$C4,$FF
	.dc.b	$96,$3E,$9F,$FC,$FD,$28,$91,$86,$FB,$5F,$E7,$08,$C4,$3E,$99,$F4
	.dc.b	$FF,$E7,$46,$DF,$FB,$F1,$F8,$13,$FE,$66,$72,$DF,$1B,$81,$C7,$4B
	.dc.b	$5E,$3E,$93,$E5,$A7,$D3,$BC,$4F,$F4,$1B,$67,$81,$B5,$E2,$6A,$77
	.dc.b	$F0,$92,$B7,$7F,$06,$FF,$C7,$87,$F9,$4D,$F8,$9E,$FF,$F4,$29,$DC
	.dc.b	$7F,$AF,$BD,$AF,$F4,$9C,$FC,$43,$51,$FE,$84,$5E,$D6,$8D,$FB,$56
	.dc.b	$F9,$FD,$3B,$E2,$FF,$D3,$AD,$8C,$F1,$62,$73,$D1,$27,$8F,$4D,$CF
	.dc.b	$FD,$27,$87,$E3,$67,$27,$6F,$8B,$EB,$5D,$BC,$79,$3E,$2F,$22,$E1
	.dc.b	$C9,$F1,$9E,$4F,$4E,$CF,$A3,$B7,$CE,$7A,$3C,$AB,$1E,$C7,$0F,$C8
	.dc.b	$E5,$FF,$67,$3F,$6F,$44,$FF,$AF,$39,$DF,$47,$4C,$78,$F6,$FE,$8C
	.dc.b	$A3,$31,$D7,$DF,$BA,$FF,$AD,$07,$A7,$FE,$C8,$6C,$FD,$A5,$DB,$99
	.dc.b	$D3,$27,$5D,$B6,$BA,$3C,$CF,$F6,$9E,$BF,$7A,$76,$75,$DB,$C0,$78
	.dc.b	$75,$B1,$E1,$70,$F8,$ED,$E4,$7E,$CF,$E9,$DE,$AF,$F9,$06,$FD,$55
	.dc.b	$A3,$FC,$88,$7E,$96,$AB,$4A,$6C,$FF,$25,$7D,$33,$C6,$7D,$3F,$08
	.dc.b	$C3,$63,$33,$DA,$72,$DF,$1F,$FD,$FE,$DE,$3E,$51,$88,$EE,$D7,$F7
	.dc.b	$52,$F9,$7F,$DE,$57,$EF,$A7,$F8,$94,$FE,$09,$FD,$7B,$5F,$F7,$7B
	.dc.b	$C5,$CF,$2F,$2D,$BF,$83,$70,$7F,$3C,$9F,$C8,$79,$99,$7A,$D5,$57
	.dc.b	$F8,$B7,$BF,$9D,$98,$37,$9F,$0E,$7E,$22,$57,$03,$CE,$F5,$C7,$FC
	.dc.b	$4F,$1F,$A1,$F7,$76,$B4,$F8,$0F,$DE,$E6,$F2,$72,$39,$A9,$88,$B3
	.dc.b	$FA,$BF,$A4,$EC,$61,$E9,$C1,$FC,$9F,$E9,$FE,$34,$60,$FE,$7C,$FC
	.dc.b	$BC,$7C,$2D,$9E,$1C,$5F,$16,$31,$5F,$D1,$7E,$9F,$FC,$92,$FF,$22
	.dc.b	$6B,$D4,$47,$05,$F4,$FF,$E6,$A7,$F9,$4A,$C7,$E9,$36,$33,$E5,$62
	.dc.b	$73,$E8,$49,$E3,$1C,$3E,$2B,$8F,$3C,$9E,$B1,$5F,$26,$1F,$DE,$9F
	.dc.b	$0F,$AB,$9D,$BE,$8C,$BC,$1F,$1F,$1C,$BF,$3C,$79,$B8,$3F,$11,$F4
	.dc.b	$CF,$A7,$A6,$6B,$FA,$2C,$3C,$6D,$F8,$D2,$73,$1F,$91,$F5,$DE,$D5
	.dc.b	$FA,$09,$5F,$D3,$A7,$2D,$F8,$1E,$1F,$4F,$23,$06,$FB,$FA,$77,$C8
	.dc.b	$C1,$FD,$63,$E9,$FE,$13,$F5,$DF,$80,$E7,$83,$F4,$91,$AE,$3F,$CA
	.dc.b	$FD,$3F,$F8,$F2,$BF,$B0,$BE,$D6,$ED,$FA,$B8,$D4,$7E,$0F,$E9,$FF
	.dc.b	$C7,$AF,$F6,$87,$D6,$E7,$7F,$D9,$46,$A3,$87,$8D,$8E,$F9,$1C,$63
	.dc.b	$7E,$27,$0D,$A4,$F5,$8F,$DE,$1F,$C0,$5F,$CA,$3E,$BE,$6F,$F0,$53
	.dc.b	$63,$FC,$40,$EF,$4F,$D9,$AF,$85,$E7,$F0,$92,$BF,$C0,$33,$97,$FC
	.dc.b	$07,$F4,$FF,$E3,$F6,$FA,$7F,$C3,$63,$FC,$7D,$1A,$F5,$71,$E3,$67
	.dc.b	$C7,$93,$CA,$3F,$8A,$7D,$3F,$F8,$62,$3F,$C1,$EB,$E1,$F2,$6B,$DF
	.dc.b	$DF,$9F,$C3,$8A,$5F,$08,$FE,$3D,$F4,$FF,$AD,$8F,$DB,$2F,$27,$26
	.dc.b	$BC,$9D,$FB,$9C,$3C,$19,$7F,$E2,$0B,$4F,$F2,$9F,$A5,$78,$77,$E1
	.dc.b	$A5,$FF,$89,$A7,$2D,$8E,$1E,$24,$8C,$1B,$7F,$9A,$7C,$D3,$E3,$69
	.dc.b	$B9,$FE,$1D,$7F,$E4,$17,$B7,$07,$8C,$F2,$7A,$76,$3F,$C6,$FF,$1D
	.dc.b	$F9,$4E,$DF,$E5,$ED,$EE,$FF,$26,$61,$F3,$EB,$AF,$27,$B7,$C0,$91
	.dc.b	$EB,$8A,$FF,$1F,$FA,$67,$D3,$FF,$4B,$6C,$7C,$1A,$F8,$DE,$17,$C7
	.dc.b	$F2,$64,$FC,$58,$AF,$B3,$E7,$3E,$61,$8D,$7F,$A2,$E6,$8E,$DE,$14
	.dc.b	$AF,$9D,$1F,$D6,$FE,$91,$F3,$97,$D6,$95,$FE,$58,$9C,$B6,$3F,$B1
	.dc.b	$7D,$3B,$E5,$29,$C2,$7F,$9E,$53,$D7,$ED,$EC,$49,$E9,$15,$FE,$85
	.dc.b	$F4,$CF,$A7,$FE,$DC,$D8,$F1,$57,$9F,$D8,$F4,$75,$C9,$F3,$45,$7C
	.dc.b	$3F,$45,$ED,$31,$C5,$FD,$A3,$34,$7F,$7B,$FA,$7F,$F7,$75,$7F,$9D
	.dc.b	$9A,$F2,$3E,$17,$85,$F1,$FC,$49,$18,$3E,$33,$FE,$05,$FA,$7F,$E7
	.dc.b	$D1,$FE,$80,$6B,$F1,$23,$E8,$7A,$1F,$47,$BA,$4F,$CF,$8A,$FF,$65
	.dc.b	$E3,$7C,$AF,$D0,$33,$AF,$F8,$0B,$E9,$FF,$B5,$B7,$C1,$FF,$0B,$41
	.dc.b	$BF,$3B,$87,$3F,$1F,$31,$FF,$4A,$D6,$DB,$1F,$4B,$E3,$24,$60,$DF
	.dc.b	$AB,$DB,$E3,$E4,$60,$E6,$FE,$77,$C7,$91,$83,$BF,$BF,$E6,$FA,$79
	.dc.b	$18,$3C,$6F,$F8,$BF,$3F,$E9,$11,$FE,$07,$6E,$BF,$06,$7F,$C6,$9F
	.dc.b	$8B,$7E,$F6,$BF,$CB,$AD,$7F,$D8,$41,$C3,$FE,$C5,$6B,$FE,$DB,$F3
	.dc.b	$3F,$3E,$FF,$99,$9A,$F2,$6B,$FE,$CD,$6B,$C7,$BF,$FC,$D3,$1F,$A1
	.dc.b	$8F,$F3,$6B,$5E,$91,$FE,$6E,$FA,$8F,$BB,$F2,$BF,$E7,$26,$CF,$F9
	.dc.b	$D9,$78,$FF,$EA,$08,$C7,$22,$BF,$CF,$3F,$51,$FC,$FF,$A2,$FA,$FF
	.dc.b	$E4,$6B,$FF,$71,$4E,$5F,$F4,$0B,$5F,$8A,$E7,$6F,$F2,$6B,$6B,$DD
	.dc.b	$EB,$7F,$9B,$E7,$37,$67,$AF,$CC,$5C,$3E,$57,$C4,$D5,$8F,$FB,$95
	.dc.b	$6E,$9D,$4E,$63,$17,$FE,$E8,$8C,$4E,$AF,$A5,$6A,$FF,$C9,$55,$AB
	.dc.b	$E1,$5A,$BE,$11,$FE,$F2,$BE,$18,$FB,$AB,$53,$C0,$D5,$70,$C3,$FC
	.dc.b	$AD,$7C,$27,$DE,$68,$E1,$1A,$9E,$17,$FE,$76,$AE,$12,$3E,$1D,$F8
	.dc.b	$1F,$4B,$67,$C9,$7D,$67,$2D,$8C,$C7,$78,$F0,$C7,$7B,$F0,$C7,$E5
	.dc.b	$5B,$3F,$52,$36,$64,$AE,$4E,$E7,$52,$CF,$58,$DC,$7D,$A9,$EA,$D8
	.dc.b	$CD,$FC,$08,$F7,$C7,$A4,$F9,$2F,$5A,$F5,$9D,$5E,$F1,$E5,$69,$E5
	.dc.b	$8D,$C6,$5B,$19,$8E,$AF,$E7,$2F,$F4,$4F,$D1,$C1,$18,$8F,$29,$FA
	.dc.b	$47,$11,$A9,$FD,$6C,$60,$8F,$C2,$8F,$F7,$23,$63,$E1,$C6,$27,$D2
	.dc.b	$76,$AF,$3B,$FA,$4F,$67,$75,$C1,$B7,$D5,$B3,$CA,$FA,$5F,$27,$F5
	.dc.b	$84,$9D,$C6,$5B,$5F,$42,$3F,$7B,$18,$AF,$89,$7F,$B2,$8C,$47,$9E
	.dc.b	$BF,$84,$C6,$2F,$B4,$7B,$CF,$31,$67,$CD,$3B,$AE,$93,$9B,$F2,$34
	.dc.b	$F2,$1D,$8C,$57,$91,$FE,$05,$5F,$33,$CD,$FE,$0F,$F1,$DF,$C2,$1B
	.dc.b	$3F,$C1,$DF,$B2,$47,$F9,$2B,$F7,$FE,$A1,$E8,$D2,$4F,$F4,$6C,$3F
	.dc.b	$6D,$3A,$8C,$42,$30,$FF,$B6,$5B,$8F,$FB,$D1,$FF,$18,$4F,$EF,$E3
	.dc.b	$15,$FC,$78,$FE,$33,$E4,$59,$FE,$28,$DB,$FE,$6B,$3F,$92,$24,$F8
	.dc.b	$BE,$07,$FA,$03,$D4,$FE,$19,$3F,$E9,$1E,$2F,$E6,$1F,$49,$FE,$3C
	.dc.b	$9D,$4F,$F1,$1A,$E2,$FF,$BC,$67,$4D,$F4,$7F,$9C,$72,$3D,$9E,$71
	.dc.b	$D3,$EA,$26,$3D,$2D,$87,$F8,$FE,$30,$A7,$C7,$D5,$27,$87,$D4,$3B
	.dc.b	$F1,$87,$8F,$E2,$BD,$27,$AB,$82,$DF,$47,$0F,$6F,$42,$D7,$AD,$F8
	.dc.b	$27,$F5,$1F,$A4,$7A,$D8,$AF,$2F,$CE,$7C,$63,$E3,$FA,$16,$3D,$2E
	.dc.b	$1F,$3F,$D4,$5C,$7E,$47,$27,$CE,$3E,$7F,$9F,$6B,$E0,$B8,$7E,$37
	.dc.b	$D4,$7C,$EE,$94,$7D,$8E,$45,$7F,$67,$FA,$77,$D4,$53,$1E,$8D,$BF
	.dc.b	$A1,$28,$CC,$63,$CB,$BA,$32,$EE,$FF,$51,$BF,$F2,$25,$6C,$FB,$5C
	.dc.b	$2E,$3F,$16,$1E,$3F,$8A,$95,$E0,$F3,$EE,$3D,$63,$F2,$38,$A7,$D4
	.dc.b	$7D,$1E,$5B,$3E,$8F,$8F,$F7,$1E,$FE,$8B,$1E,$C7,$0F,$93,$DD,$B8
	.dc.b	$F7,$3D,$5F,$2F,$F1,$CF,$91,$E9,$DA,$F8,$AE,$1F,$8F,$F3,$B6,$B9
	.dc.b	$0D,$11,$F4,$4F,$C4,$7B,$16,$BE,$7F,$CE,$7D,$53,$EB,$5D,$8F,$A6
	.dc.b	$E1,$F9,$7E,$85,$8E,$FF,$50,$F8,$9F,$F1,$23,$6B,$F5,$4F,$EB,$DB
	.dc.b	$1A,$F1,$A4,$7D,$6E,$DB,$7D,$3D,$7E,$67,$4F,$89,$B5,$D1,$C3,$DB
	.dc.b	$D4,$B5,$E4,$7C,$67,$E9,$E2,$9F,$BD,$FA,$8F,$CA,$B7,$BF,$FC,$79
	.dc.b	$1F,$9A,$8C,$1F,$C0,$F6,$7D,$0F,$F9,$05,$B1,$9E,$7C,$4F,$F9,$17
	.dc.b	$E9,$DF,$51,$4C,$7C,$B5,$BF,$E1,$D3,$F1,$F4,$FB,$2D,$F0,$7F,$3B
	.dc.b	$08,$F9,$4E,$B8,$BB,$7A,$72,$F9,$FC,$55,$AF,$C1,$FD,$0B,$E2,$3F
	.dc.b	$CB,$0D,$AF,$E3,$CF,$F8,$6C,$C4,$F0,$E9,$D9,$C3,$A6,$D7,$F2,$5F
	.dc.b	$A8,$78,$7F,$E6,$36,$C6,$7C,$3C,$4E,$7E,$56,$4F,$1F,$96,$B9,$7E
	.dc.b	$63,$0F,$F2,$54,$F5,$8F,$F3,$73,$F2,$8C,$1F,$CD,$4F,$F0,$39,$FD
	.dc.b	$18,$AF,$2F,$92,$E3,$EE,$7F,$9D,$FE,$A1,$F2,$D4,$C7,$B5,$70,$FE
	.dc.b	$81,$F5,$0F,$A7,$FF,$3F,$36,$33,$B6,$27,$3E,$34,$9E,$3B,$DD,$7F
	.dc.b	$67,$F5,$0F,$13,$FD,$10,$DA,$FE,$9C,$BC,$4C,$CE,$BE,$0A,$56,$B8
	.dc.b	$2E,$1F,$BE,$F7,$FF,$A0,$CF,$F5,$76,$C6,$78,$3F,$D2,$D1,$8E,$EF
	.dc.b	$C5,$AF,$67,$26,$BD,$9B,$87,$F6,$1F,$A8,$7C,$67,$FA,$7D,$B1,$9E
	.dc.b	$2C,$4E,$78,$E4,$F1,$EA,$BA,$FF,$14,$C3,$F9,$54,$E4,$ED,$F1,$9D
	.dc.b	$6B,$B7,$8D,$27,$C5,$E3,$DC,$39,$3E,$33,$C9,$E9,$D9,$F4,$77,$79
	.dc.b	$CF,$47,$93,$63,$D8,$E1,$FE,$FD,$CC,$EC,$C6,$3F,$CA,$B9,$1F,$0B
	.dc.b	$FE,$BF,$6D,$FF,$81,$57,$8D,$99,$D3,$27,$AF,$BF,$6F,$F9,$A8,$F4
	.dc.b	$FF,$D9,$4D,$9F,$F0,$83,$FC,$E1,$B1,$AE,$F9,$1A,$EF,$B5,$D1,$E6
	.dc.b	$7F,$B5,$74,$FB,$D3,$B3,$AE,$DE,$03,$C3,$B5,$8F,$0B,$87,$C7,$6F
	.dc.b	$1B,$FD,$7D,$F5,$0F,$57,$FC,$6B,$2B,$F5,$BB,$3F,$C9,$A5,$3F,$C9
	.dc.b	$F5,$87,$07,$F8,$FB,$E9,$DE,$03,$EA,$3F,$CA,$B1,$86,$C6,$67,$FC
	.dc.b	$B9,$39,$6F,$8F,$FE,$05,$6F,$27,$28,$C4,$7A,$25,$18,$8F,$17,$CC
	.dc.b	$FF,$94,$2B,$F9,$8D,$7E,$0A,$7F,$C8,$8D,$7F,$B7,$A9,$E1,$AF,$AF
	.dc.b	$96,$D7,$FB,$97,$83,$F9,$93,$F9,$0F,$33,$2B,$0E,$98,$F8,$39,$B3
	.dc.b	$0F,$36,$98,$8B,$5E,$4F,$88,$D9,$87,$CA,$F6,$C7,$FC,$45,$1F,$A0
	.dc.b	$F7,$76,$FF,$97,$F8,$DC,$DE,$45,$CD,$4C,$45,$9F,$D4,$F8,$FD,$4C
	.dc.b	$3F,$E1,$2F,$A7,$F8,$8F,$F9,$AD,$F6,$F0,$FF,$9F,$1B,$00,$F0,$79
	.dc.b	$70,$FC,$6F,$A8,$FF,$22,$47,$F9,$D5,$AE,$8F,$F3,$C7,$D4,$7F,$99
	.dc.b	$1B,$87,$CE,$8F,$E8,$31,$AF,$87,$C4,$E7,$BE,$4F,$03,$9F,$C4,$F1
	.dc.b	$E6,$91,$83,$BB,$D7,$E3,$E4,$60,$F1,$F8,$3E,$1E,$29,$18,$3C,$FB
	.dc.b	$FE,$0B,$E9,$DF,$51,$4F,$44,$67,$C5,$D7,$8B,$2F,$FD,$2D,$69,$FC
	.dc.b	$5F,$AE,$F6,$AF,$D0,$CB,$FF,$4E,$4E,$5B,$F0,$3C,$3E,$A2,$46,$0D
	.dc.b	$F7,$F4,$EE,$91,$83,$FD,$4D,$F5,$0E,$2F,$CD,$AE,$2A,$62,$3F,$3F
	.dc.b	$F5,$1F,$E2,$AA,$FF,$57,$3E,$96,$ED,$AF,$37,$F4,$BF,$51,$FE,$2A
	.dc.b	$9F,$F5,$A3,$E9,$7F,$CB,$ED,$F1,$7F,$C4,$EE,$E3,$3E,$3F,$2F,$1E
	.dc.b	$47,$28,$EB,$E3,$76,$DE,$4F,$68,$FD,$A9,$FB,$87,$FC,$DA,$FB,$DF
	.dc.b	$F6,$33,$5C,$F5,$70,$F8,$5E,$3F,$09,$27,$71,$FC,$03,$EA,$3F,$C5
	.dc.b	$31,$FE,$CF,$7F,$D1,$CD,$F4,$B5,$EA,$EF,$C8,$E1,$E4,$CA,$E1,$1F
	.dc.b	$C2,$BE,$A3,$FC,$15,$1F,$ED,$95,$F0,$F9,$35,$EF,$6F,$D0,$E1,$C3
	.dc.b	$2F,$84,$7F,$14,$FA,$8F,$D3,$C7,$FA,$31,$FE,$CC,$FF,$B9,$5A,$E7
	.dc.b	$C9,$E1,$EE,$71,$F0,$64,$EE,$3F,$8F,$FC,$D3,$C3,$BF,$0D,$2B,$FD
	.dc.b	$E1,$39,$6C,$70,$F1,$24,$60,$DB,$FC,$AB,$E6,$9F,$1D,$4D,$CF,$FA
	.dc.b	$92,$FF,$C6,$4F,$6D,$FC,$67,$93,$D5,$B1,$FF,$00,$7E,$3F,$F2,$CF
	.dc.b	$FA,$B1,$BE,$07,$F9,$D2,$FE,$7C,$F5,$E2,$F6,$F7,$C8,$F5,$C5,$7F
	.dc.b	$D6,$BF,$4E,$FA,$8F,$E8,$2D,$8F,$83,$5F,$1B,$C1,$F8,$FE,$4C,$9F
	.dc.b	$8B,$15,$F6,$7D,$47,$CC,$31,$9F,$95,$EB,$E1,$C9,$F9,$B1,$FD,$43
	.dc.b	$E9,$9F,$39,$7D,$69,$5F,$CC,$F0,$FF,$48,$D6,$A3,$1E,$97,$F3,$75
	.dc.b	$1F,$CB,$DF,$F6,$8F,$D1,$BE,$57,$FA,$EB,$63,$89,$FC,$FE,$78,$5F
	.dc.b	$6F,$5F,$C9,$E5,$4A,$D9,$DA,$30,$FF,$B7,$BE,$9D,$F5,$1F,$DA,$9B
	.dc.b	$1C,$30,$FD,$C6,$1E,$AF,$67,$E0,$75,$CA,$D4,$57,$FD,$D3,$4F,$A8
	.dc.b	$F8,$C6,$6F,$E1,$BF,$DE,$3E,$A3,$FD,$C1,$5E,$66,$BC,$A3,$E7,$78
	.dc.b	$5F,$3F,$C3,$93,$E5,$8A,$FF,$BE,$3E,$A3,$FA,$0B,$F0,$1A,$F9,$27
	.dc.b	$EC,$7A,$1F,$67,$8E,$4F,$BE,$2B,$F4,$BE,$57,$E8,$98,$CF,$FD,$97
	.dc.b	$F5,$1F,$DA,$1B,$E0,$FF,$84,$A0,$DF,$9D,$C3,$9F,$8F,$98,$7D,$4A
	.dc.b	$DB,$63,$E9,$7C,$6C,$8C,$1B,$F4,$FB,$7C,$74,$8C,$1C,$DF,$D1,$F8
	.dc.b	$F2,$30,$77,$F8,$1F,$37,$D4,$48,$C1,$E3,$7F,$C5,$79,$FF,$4B,$8F
	.dc.b	$DC,$7B,$7B,$5B,$7F,$BB,$3E,$A3,$FA,$54,$7F,$96,$9A,$FE,$75,$AF
	.dc.b	$F8,$55,$AB,$F6,$67,$8F,$FE,$5E,$6C,$7F,$99,$0F,$F0,$E3,$87,$0D
	.dc.b	$7F,$CD,$2B,$C5,$E4,$48,$E4,$77,$7F,$CD,$75,$FC,$96,$3B,$35,$F5
	.dc.b	$7A,$3A,$3F,$CE,$3F,$52,$E5,$7C,$B7,$48,$C3,$7C,$FF,$F3,$D5,$FA
	.dc.b	$67,$2F,$A3,$EA,$9D,$5E,$4D,$11,$89,$D4,$E6,$31,$1F,$E8,$6F,$AA
	.dc.b	$7A,$9B,$E0,$CC,$3D,$5C,$61,$FC,$36,$36,$FF,$02,$74,$DA,$DD,$FF
	.dc.b	$92,$67,$2F,$DE,$9E,$53,$A8,$4E,$AF,$C3,$39,$8C,$57,$03,$C1,$79
	.dc.b	$9D,$36,$32,$78,$CB,$7E,$13,$C9,$3C,$23,$51,$93,$C7,$3C,$DC,$57
	.dc.b	$89,$E3,$3D,$2F,$CE,$D3,$CD,$1C,$2B,$94,$FC,$C6,$CF,$CB,$6C,$FC
	.dc.b	$76,$BE,$28,$F2,$CF,$16,$9C,$4B,$1C,$5A,$F1,$3D,$37,$EF,$6C,$79
	.dc.b	$DF,$59,$E7,$78,$DF,$BE,$3E,$AC,$F1,$AF,$7C,$6C,$E6,$57,$27,$9D
	.dc.b	$CB,$4C,$D7,$A6,$32,$D8,$CC,$7A,$CC,$95,$F0,$BF,$99,$9F,$CD,$BF
	.dc.b	$40,$C4,$FE,$DA,$BF,$2F,$1F,$C0,$5A,$E8,$C4,$F2,$7E,$1D,$78,$9F
	.dc.b	$E1,$9A,$7D,$4A,$E7,$39,$CB,$F7,$38,$7C,$8C,$3E,$CC,$63,$0F,$49
	.dc.b	$F2,$A7,$EA,$C6,$6B,$B5,$F4,$9F,$C4,$C7,$ED,$A7,$51,$82,$57,$99
	.dc.b	$FF,$00,$46,$A3,$04,$3C,$07,$E7,$5D,$38,$9F,$0D,$FE,$CE,$73,$7E
	.dc.b	$3A,$E4,$D7,$E1,$DF,$BC,$F5,$E1,$F1,$6F,$E0,$B6,$32,$6E,$7E,$1D
	.dc.b	$F6,$AD,$47,$96,$AE,$1F,$35,$B7,$F0,$27,$C2,$7C,$CC,$3C,$87,$97
	.dc.b	$0E,$93,$DE,$FE,$55,$FD,$9F,$86,$FE,$21,$F5,$5F,$EF,$9A,$FE,$35
	.dc.b	$18,$F4,$7F,$88,$3F,$83,$CB,$F8,$AB,$72,$34,$C7,$A3,$2F,$F2,$0D
	.dc.b	$AF,$C8,$3F,$C6,$CF,$E4,$1E,$BB,$E6,$71,$5F,$CA,$F1,$CF,$8D,$FC
	.dc.b	$85,$B1,$93,$F3,$8F,$8C,$8F,$F5,$4F,$82,$FB,$A7,$DE,$FA,$6F,$F2
	.dc.b	$5C,$7F,$8E,$23,$1E,$AF,$F3,$19,$CE,$7F,$C8,$7E,$37,$EF,$46,$38
	.dc.b	$7F,$A3,$AF,$AA,$E0,$71,$F4,$EE,$5F,$E4,$49,$46,$63,$FC,$8F,$F5
	.dc.b	$4F,$0F,$19,$7F,$2F,$BF,$F6,$A5,$71,$F8,$BF,$27,$A5,$E5,$6C,$75
	.dc.b	$8F,$47,$A7,$8B,$E4,$E2,$7F,$CA,$BF,$4E,$F6,$B1,$5D,$F9,$3E,$01
	.dc.b	$F0,$79,$2D,$FA,$5C,$3E,$8F,$AA,$B5,$E7,$E4,$F9,$07,$C9,$F4,$6D
	.dc.b	$7C,$17,$0F,$F9,$9B,$EA,$BE,$53,$FE,$75,$DD,$3F,$D8,$7E,$A5,$F5
	.dc.b	$54,$C7,$A9,$6F,$FC,$DF,$28,$CC,$63,$CD,$B9,$FF,$59,$AF,$F2,$3C
	.dc.b	$E4,$F8,$7C,$5D,$EB,$C1,$F1,$72,$7A,$FA,$57,$0E,$4F,$94,$FE,$ED
	.dc.b	$C7,$FE,$BA,$9D,$9E,$4F,$3F,$D4,$7A,$F9,$EC,$7A,$1C,$3F,$53,$DF
	.dc.b	$B8,$79,$FE,$67,$E1,$1F,$0F,$D5,$B5,$F0,$1C,$3F,$0B,$E7,$ED,$7B
	.dc.b	$DF,$94,$79,$BD,$9B,$3F,$23,$E7,$BE,$71,$F3,$EA,$C7,$CC,$70,$FE
	.dc.b	$23,$D1,$B1,$FE,$03,$FA,$AF,$EC,$44,$F3,$BF,$C3,$0D,$AF,$F5,$1B
	.dc.b	$FC,$97,$77,$56,$AB,$7E,$C4,$BD,$FC,$45,$C7,$B3,$D0,$EB,$F1,$96
	.dc.b	$7A,$B8,$7F,$3A,$59,$FF,$0C,$7D,$57,$E3,$23,$FD,$6A,$D8,$CC,$7F
	.dc.b	$8B,$E3,$1E,$47,$F9,$0B,$67,$C9,$FF,$8C,$DB,$5F,$EB,$E5,$CD,$99
	.dc.b	$FF,$1C,$7D,$4B,$EA,$A9,$8F,$99,$B8,$7F,$B1,$A7,$FC,$14,$4D,$F8
	.dc.b	$FA,$FC,$BD,$29,$5B,$F4,$6E,$FF,$E3,$3F,$A2,$7C,$57,$F9,$1D,$B5
	.dc.b	$FE,$D2,$5F,$33,$99,$D7,$54,$AD,$75,$5C,$3F,$C7,$DF,$54,$F0,$BF
	.dc.b	$CA,$4D,$AF,$F6,$EA,$F0,$B3,$3A,$F9,$69,$5A,$F9,$7B,$87,$43,$99
	.dc.b	$FE,$5D,$8F,$8C,$74,$71,$F2,$D2,$3F,$DD,$63,$E5,$31,$3F,$E5,$3F
	.dc.b	$AA,$7C,$B7,$F9,$9D,$B5,$FE,$F2,$5E,$D6,$67,$FC,$D9,$F5,$4F,$A8
	.dc.b	$A6,$35,$B8,$7F,$BE,$A5,$E3,$F0,$38,$ED,$67,$FD,$FF,$F5,$4F,$17
	.dc.b	$FC,$F0,$D8,$CF,$8B,$89,$CF,$D5,$49,$E3,$C3,$78,$3F,$CD,$DF,$09
	.dc.b	$FC,$D6,$51,$8F,$6B,$FD,$09,$3E,$E8,$C4,$EB,$C6,$FF,$4C,$CF,$F6
	.dc.b	$98,$C1,$CA,$7F,$D5,$7E,$D7,$2F,$6B,$9B,$FE,$8F,$FA,$A7,$C8,$53
	.dc.b	$1C,$97,$0E,$78,$79,$2B,$F0,$8F,$FA,$15,$B9,$7F,$AA,$9F,$C3,$4F
	.dc.b	$09,$46,$A3,$C8,$EA,$39,$75,$78,$FE,$43,$C8,$F9,$7C,$99,$7C,$EF
	.dc.b	$F6,$2D,$8F,$57,$95,$23,$5E,$A5,$C7,$CD,$DD,$EF,$3E,$07,$99,$6B
	.dc.b	$A3,$87,$D1,$CB,$F8,$A7,$27,$AF,$57,$D3,$7F,$AD,$DB,$19,$F2,$31
	.dc.b	$39,$64,$74,$F8,$0B,$97,$FA,$B4,$7A,$9F,$EC,$16,$CF,$C9,$7F,$26
	.dc.b	$D8,$D7,$6C,$8D,$76,$DA,$E8,$F3,$3F,$D9,$99,$FE,$16,$33,$3D,$75
	.dc.b	$F0,$1E,$1D,$6C,$78,$5C,$3E,$39,$79,$3F,$96,$FA,$A7,$AD,$F8,$F9
	.dc.b	$5F,$AD,$D9,$93,$F2,$95,$5A,$53,$67,$3F,$A9,$77,$3E,$AB,$84,$61
	.dc.b	$B1,$99,$ED,$39,$6F,$90,$E4,$E2,$D8,$CC,$60,$EB,$F3,$BF,$A9,$5A
	.dc.b	$FE,$AE,$FF,$03,$5F,$9D,$6B,$FE,$E7,$55,$C6,$78,$72,$DA,$FD,$EF
	.dc.b	$07,$F3,$A7,$F1,$FE,$6E,$56,$1D,$31,$F0,$C7,$F0,$48,$73,$6D,$79
	.dc.b	$F4,$C4,$59,$F3,$7C,$A6,$CC,$3E,$77,$C1,$8B,$DD,$A6,$E7,$F7,$5C
	.dc.b	$6E,$6F,$32,$E6,$A6,$22,$CF,$EA,$BC,$BE,$A6,$1F,$A5,$C1,$FC,$57
	.dc.b	$EA,$BE,$3C,$7F,$07,$7E,$04,$67,$87,$E2,$46,$A7,$89,$FC,$99,$DB
	.dc.b	$FC,$A3,$EA,$BF,$C8,$8F,$F1,$16,$BC,$04,$70,$5F,$55,$FE,$65,$7F
	.dc.b	$8C,$3F,$D3,$9B,$87,$D5,$8D,$47,$37,$4D,$F3,$F4,$6E,$78,$7C,$57
	.dc.b	$1E,$59,$5D,$A2,$BC,$FD,$5E,$C6,$39,$BF,$03,$1E,$6F,$03,$C5,$C5
	.dc.b	$23,$CF,$1E,$5D,$FF,$15,$F5,$2F,$AA,$A6,$6B,$F9,$A4,$7F,$86,$9B
	.dc.b	$E1,$F0,$38,$F8,$9F,$93,$F5,$DE,$DD,$F9,$79,$7F,$CF,$67,$2D,$F6
	.dc.b	$F8,$7C,$8C,$8C,$1B,$EF,$E9,$DD,$23,$07,$F4,$7F,$AA,$FF,$0C,$B7
	.dc.b	$83,$F4,$90,$6F,$1F,$E9,$FE,$AB,$FC,$6F,$5F,$D3,$DF,$63,$F7,$DB
	.dc.b	$C9,$F2,$9D,$C7,$E0,$FE,$AB,$FC,$70,$FF,$57,$7D,$6E,$77,$FD,$A4
	.dc.b	$6A,$38,$78,$B8,$F1,$64,$71,$8D,$F8,$9C,$36,$93,$D6,$3F,$7C,$7F
	.dc.b	$01,$7F,$28,$FC,$06,$E7,$AE,$AD,$FC,$27,$0F,$8C,$95,$FD,$B6,$FF
	.dc.b	$CD,$AF,$F6,$FF,$AA,$FF,$1D,$37,$D2,$FE,$19,$1F,$E3,$98,$D7,$BD
	.dc.b	$8F,$1B,$3E,$3C,$9E,$51,$FC,$4F,$EA,$9E,$ED,$F8,$79,$5F,$DF,$67
	.dc.b	$2D,$F3,$B8,$71,$48,$C1,$FE,$03,$FA,$AF,$D3,$F4,$7E,$71,$FF,$1B
	.dc.b	$C6,$7C,$8F,$E4,$51,$AF,$17,$FC,$34,$BD,$CC,$F8,$32,$FA,$47,$F2
	.dc.b	$9F,$9D,$78,$77,$E1,$A5,$7F,$84,$E7,$2D,$8E,$1E,$24,$8C,$1B,$7F
	.dc.b	$9A,$7C,$C3,$E5,$69,$B9,$FE,$19,$7F,$E4,$37,$B7,$1F,$8C,$F2,$7A
	.dc.b	$76,$3F,$C4,$9F,$1F,$F9,$2E,$1F,$DE,$B7,$83,$FC,$71,$87,$CC,$AE
	.dc.b	$BC,$5E,$DF,$06,$47,$AE,$2B,$FC,$77,$EA,$5F,$55,$FD,$2D,$B1,$F0
	.dc.b	$6B,$E3,$78,$7F,$1F,$C9,$93,$F1,$62,$BE,$CF,$A1,$F9,$86,$35,$FE
	.dc.b	$27,$9A,$3B,$78,$32,$BE,$74,$7F,$5B,$FA,$77,$CE,$5F,$88,$95,$FE
	.dc.b	$45,$9C,$B6,$3F,$B1,$7D,$53,$E5,$69,$C2,$7F,$9D,$53,$D6,$ED,$E5
	.dc.b	$49,$E9,$15,$FE,$81,$F5,$2F,$AA,$FE,$DC,$D8,$F1,$57,$9F,$D8,$F4
	.dc.b	$75,$C9,$F3,$45,$7C,$3F,$53,$ED,$31,$C5,$FC,$DB,$34,$7F,$7B,$FA
	.dc.b	$A7,$81,$4E,$17,$F0,$7C,$2F,$8D,$E1,$C8,$C1,$F1,$5F,$F0,$27,$D5
	.dc.b	$7F,$3B,$8F,$F3,$63,$5F,$87,$1F,$3F,$CF,$FA,$1D,$F2,$7E,$74,$57
	.dc.b	$FB,$17,$23,$E5,$7E,$79,$9D,$BF,$A7,$FD,$57,$F6,$86,$F8,$5F,$E1
	.dc.b	$58,$37,$E7,$70,$E7,$E4,$26,$3F,$E8,$1A,$DB,$63,$E8,$FC,$64,$8C
	.dc.b	$1B,$F5,$3B,$7C,$7C,$8C,$1C,$DF,$CE,$79,$12,$30,$77,$F7,$FC,$DF
	.dc.b	$55,$23,$07,$8D,$FF,$17,$67,$FD,$13,$87,$FD,$4C,$D9,$FF,$58,$E8
	.dc.b	$AF,$F1,$A7,$D5,$7F,$44,$DF,$FB,$9B,$70,$FF,$57,$36,$FF,$D7,$2F
	.dc.b	$F4,$79,$FE,$F3,$F5,$0F,$23,$FC,$7C,$DA,$FF,$53,$35,$FF,$0B,$57
	.dc.b	$35,$1F,$AF,$77,$E2,$F2,$39,$FC,$4F,$9B,$D1,$93,$FE,$B1,$BF,$E3
	.dc.b	$EF,$F8,$86,$F9,$37,$C8,$FF,$38,$47,$F9,$32,$3F,$CE,$8D,$7F,$CF
	.dc.b	$0D,$7F,$CF,$DF,$55,$F8,$9F,$96,$FF,$3D,$36,$3F,$CF,$CB,$CB,$FF
	.dc.b	$41,$C6,$3A,$55,$FE,$81,$FA,$8F,$D2,$F0,$7C,$2F,$37,$FC,$BF,$C3
	.dc.b	$FE,$20,$9C,$E1,$E9,$9C,$D7,$FB,$76,$73,$5F,$3A,$73,$5F,$EE,$A9
	.dc.b	$CD,$7F,$B8,$F3,$FF,$33,$BF,$ED,$58,$CC,$FF,$BC,$27,$2F,$F9,$1E
	.dc.b	$32,$FD,$2A,$EA,$A3,$33,$C2,$FD,$1E,$0E,$A3,$18,$79,$E3,$18,$72
	.dc.b	$BE,$2D,$8C,$CF,$95,$B8,$78,$DB,$87,$69,$F9,$06,$CB,$3B,$6C,$6E
	.dc.b	$3E,$54,$6D,$A6,$CB,$1B,$3E,$64,$6D,$B1,$B6,$9B,$8F,$9D,$5E,$B7
	.dc.b	$C2,$7C,$12,$CF,$D5,$AF,$5C,$66,$3C,$2D,$AE,$AD,$3A,$9E,$32,$C7
	.dc.b	$8D,$B1,$E3,$8F,$B9,$5E,$37,$A9,$E1,$9F,$1B,$F6,$EB,$C6,$E4,$D9
	.dc.b	$98,$FC,$C4,$7F,$9D,$23,$D0,$7F,$B8,$6F,$F2,$ED,$8D,$3C,$CC,$4F
	.dc.b	$22,$3E,$93,$F3,$1E,$A7,$9F,$0F,$54,$78,$4F,$CA,$46,$67,$E2,$47
	.dc.b	$9E,$7F,$51,$39,$8C,$4E,$9F,$CC,$46,$5E,$93,$DA,$D3,$97,$F4,$24
	.dc.b	$71,$79,$6D,$FD,$3B,$F4,$B5,$EA,$F8,$6B,$AB,$97,$AD,$7E,$8E,$B9
	.dc.b	$D7,$C3,$3C,$F5,$B9,$F0,$34,$E5,$3C,$2F,$D4,$D8,$F0,$36,$BE,$19
	.dc.b	$E6,$3B,$5F,$8A,$7F,$82,$4E,$E7,$D2,$6E,$BE,$9F,$D4,$FF,$06,$81
	.dc.b	$03,$97,$0D,$76,$D7,$F8,$57,$C3,$7F,$0C,$8F,$E0,$F1,$FC,$66,$31
	.dc.b	$97,$F0,$7E,$0F,$E1,$1C,$77,$2F,$E3,$4D,$BF,$E3,$13,$FC,$5B,$D6
	.dc.b	$7C,$C6,$73,$FF,$65,$C7,$F9,$AF,$0F,$4C,$63,$C6,$FE,$29,$97,$16
	.dc.b	$CE,$8F,$75,$71,$C7,$F9,$35,$6E,$75,$1F,$13,$E0,$7F,$9F,$C6,$21
	.dc.b	$18,$7F,$DE,$9F,$4B,$FE,$46,$9D,$4E,$71,$FF,$70,$CE,$6B,$7E,$A7
	.dc.b	$F3,$2F,$6F,$F1,$71,$8C,$3B,$53,$EA,$7F,$9B,$B6,$33,$EA,$62,$72
	.dc.b	$C8,$F3,$7D,$4B,$C3,$C6,$1E,$6F,$89,$E6,$74,$F4,$AD,$F8,$1C,$3E
	.dc.b	$2F,$4E,$D7,$B9,$F1,$9F,$D2,$BE,$91,$EC,$E2,$BD,$3C,$BE,$83,$D3
	.dc.b	$CB,$63,$DA,$E1,$F3,$FD,$4D,$C7,$E5,$F2,$7C,$03,$E0,$FA,$36,$BE
	.dc.b	$2B,$87,$EA,$7D,$4F,$CD,$E6,$F9,$F3,$FD,$7B,$EA,$1F,$53,$4C,$7A
	.dc.b	$77,$0F,$B9,$28,$CC,$63,$CE,$B9,$B1,$4F,$E9,$EF,$FC,$96,$EC,$FC
	.dc.b	$37,$03,$8B,$C5,$87,$8F,$E2,$65,$78,$3D,$2B,$97,$58,$E9,$1E,$93
	.dc.b	$E7,$72,$58,$F3,$FC,$87,$B4,$F7,$73,$D8,$F5,$B8,$7C,$7E,$FD,$C3
	.dc.b	$B1,$FC,$03,$57,$CB,$FC,$73,$E4,$7A,$D6,$FE,$2B,$87,$E3,$FC,$ED
	.dc.b	$BE,$62,$D1,$F4,$4F,$C3,$7B,$56,$BE,$7F,$CE,$7D,$53,$EB,$5D,$8F
	.dc.b	$A6,$E1,$F9,$7E,$7D,$8E,$FF,$53,$FE,$1C,$27,$99,$FE,$1C,$6D,$7E
	.dc.b	$DD,$FD,$6F,$37,$D8,$AF,$E3,$FC,$DF,$BA,$AD,$5F,$1F,$66,$47,$1F
	.dc.b	$88,$B7,$E0,$7B,$1D,$FE,$32,$CF,$77,$0F,$E0,$8B,$3F,$1B,$EA,$5D
	.dc.b	$BF,$E3,$26,$CF,$F0,$D7,$F1,$AD,$8F,$F1,$BE,$CF,$A0,$A6,$39,$ED
	.dc.b	$7F,$15,$FA,$87,$D4,$FF,$8F,$9B,$19,$F9,$6C,$4E,$7D,$69,$3C,$78
	.dc.b	$2E,$BF,$73,$07,$A5,$CC,$E9,$E8,$DB,$FE,$49,$F4,$2F,$89,$FF,$26
	.dc.b	$B6,$33,$F2,$D8,$9C,$F5,$49,$E3,$D5,$73,$FC,$5F,$D4,$BC,$4F,$F2
	.dc.b	$BB,$6B,$F9,$92,$F1,$33,$3A,$F9,$59,$5A,$F9,$6B,$87,$43,$99,$FE
	.dc.b	$62,$8F,$8A,$E5,$FE,$7C,$3E,$57,$46,$FE,$52,$C7,$F4,$3F,$A9,$7C
	.dc.b	$F7,$F9,$B1,$B1,$9F,$6B,$13,$FE,$70,$FA,$97,$D3,$D3,$1A,$DB,$FE
	.dc.b	$9D,$3F,$AA,$6F,$8F,$C7,$6E,$4F,$F9,$EB,$EA,$5E,$2D,$31,$E2,$DB
	.dc.b	$FE,$B1,$2F,$A9,$E0,$71,$E0,$B3,$FD,$77,$E1,$BF,$9D,$B7,$DA,$FF
	.dc.b	$43,$C7,$B3,$8F,$F7,$EE,$9B,$ED,$68,$DF,$B5,$6F,$FB,$47,$D4,$BE
	.dc.b	$47,$FD,$22,$D8,$CF,$1E,$27,$3C,$72,$78,$F4,$DC,$BF,$87,$E1,$F8
	.dc.b	$19,$C9,$DB,$E4,$7A,$D7,$6F,$26,$4F,$8B,$C9,$B8,$72,$7C,$67,$93
	.dc.b	$D5,$B3,$E8,$ED,$F3,$9E,$8F,$2E,$C7,$B1,$C3,$FE,$05,$E7,$76,$63
	.dc.b	$1F,$E4,$BE,$83,$E1,$7F,$D6,$2D,$BF,$F0,$63,$FA,$E6,$C6,$BC,$89
	.dc.b	$1D,$7D,$FB,$7F,$CC,$07,$AB,$FE,$BC,$6C,$FF,$85,$DF,$C5,$1B,$8D
	.dc.b	$77,$49,$D7,$7D,$AE,$8F,$33,$FD,$8F,$A7,$F8,$CA,$76,$75,$DB,$C0
	.dc.b	$78,$76,$B1,$E1,$70,$F8,$ED,$E4,$FF,$94,$3E,$A7,$FA,$8C,$7F,$94
	.dc.b	$6B,$D4,$E8,$FF,$2A,$12,$6B,$4A,$7F,$96,$9B,$F5,$1F,$B9,$FA,$9B
	.dc.b	$E0,$1A,$4A,$E7,$2F,$1F,$FB,$F4,$61,$AF,$8A,$35,$C7,$DA,$73,$3F
	.dc.b	$E6,$99,$7C,$E7,$F9,$AE,$BF,$A8,$4A,$7F,$04,$FF,$92,$DA,$FF,$6A
	.dc.b	$78,$B9,$E4,$E5,$AF,$F9,$D7,$7F,$E5,$D3,$F8,$FF,$33,$2F,$F9,$EA
	.dc.b	$F9,$E1,$E0,$7F,$9C,$C7,$7F,$3B,$6F,$8D,$B1,$A8,$F1,$46,$3E,$4E
	.dc.b	$47,$13,$D0,$F2,$C3,$FC,$3D,$1F,$A1,$F7,$F6,$FF,$81,$3D,$C7,$2F
	.dc.b	$81,$72,$D3,$11,$67,$F5,$5E,$67,$53,$0F,$F8,$3B,$EA,$3E,$2B,$78
	.dc.b	$3D,$71,$F0,$38,$3F,$D3,$4E,$B7,$FA,$2E,$6B,$FD,$3F,$F5,$3F,$E3
	.dc.b	$F7,$FC,$F8,$D7,$A1,$FE,$7F,$FA,$9F,$F2,$EB,$71,$F9,$31,$FD,$02
	.dc.b	$35,$E5,$E2,$73,$DF,$27,$81,$CF,$E2,$78,$F2,$C8,$C1,$DD,$EB,$EA
	.dc.b	$C8,$C1,$E3,$F0,$3C,$3C,$72,$30,$79,$F7,$FB,$BF,$50,$FA,$9A,$7A
	.dc.b	$23,$3E,$2E,$BC,$59,$7F,$E9,$DB,$4F,$E1,$BD,$77,$B9,$7E,$5E,$5F
	.dc.b	$FA,$8E,$72,$DF,$6F,$87,$C8,$C8,$C1,$BE,$FE,$9D,$F2,$30,$7F,$AB
	.dc.b	$7E,$A5,$C5,$F9,$65,$C5,$4C,$47,$E6,$BE,$A7,$FC,$49,$5F,$EB,$57
	.dc.b	$D2,$DD,$B5,$E6,$FE,$83,$EA,$7F,$C4,$B3,$FE,$BB,$7D,$2F,$F9,$69
	.dc.b	$BD,$DF,$B8,$77,$19,$F1,$B9,$78,$D2,$39,$47,$5F,$3B,$B6,$F2,$7B
	.dc.b	$47,$EC,$0F,$DC,$BF,$E6,$87,$DE,$FF,$B3,$5A,$E7,$AB,$87,$C7,$71
	.dc.b	$F8,$39,$3B,$8F,$DD,$FD,$4F,$F8,$9E,$3F,$DA,$CF,$F8,$95,$BE,$96
	.dc.b	$BD,$5D,$F9,$1C,$3C,$99,$5C,$23,$F8,$2F,$D4,$BD,$EB,$F0,$F2,$BF
	.dc.b	$DC,$53,$96,$FA,$3C,$38,$64,$60,$FF,$74,$7D,$4F,$E9,$DF,$F4,$3B
	.dc.b	$FD,$99,$CF,$91,$FC,$52,$35,$CD,$FC,$A1,$7B,$99,$F0,$65,$F4,$8F
	.dc.b	$E3,$7F,$36,$F0,$EF,$C3,$CA,$FF,$7C,$4E,$5B,$1C,$3C,$49,$18,$36
	.dc.b	$FF,$26,$F9,$87,$C7,$D3,$73,$FE,$A0,$BF,$F1,$8B,$DB,$8B,$C6,$79
	.dc.b	$3D,$5B,$03,$FD,$4F,$F2,$1F,$98,$6F,$83,$A7,$A7,$17,$B3,$C0,$91
	.dc.b	$EA,$8A,$FF,$AC,$3E,$A1,$F5,$3F,$CF,$1B,$1F,$02,$BE,$2F,$85,$F1
	.dc.b	$BC,$99,$3F,$12,$2B,$EB,$FA,$0F,$96,$63,$9F,$FA,$CE,$E8,$ED,$E2
	.dc.b	$4A,$F9,$B1,$FD,$33,$E9,$DF,$39,$7D,$69,$5F,$07,$1F,$EC,$75,$A8
	.dc.b	$C7,$A5,$F1,$54,$7F,$30,$7F,$D9,$BF,$48,$F9,$6F,$EB,$2D,$8E,$27
	.dc.b	$CD,$9E,$17,$DB,$D7,$F2,$79,$72,$B6,$76,$8C,$3F,$ED,$AF,$A8,$7D
	.dc.b	$4F,$F6,$66,$C7,$0C,$3F,$73,$87,$AB,$D9,$F8,$1D,$92,$B5,$15,$FF
	.dc.b	$72,$D3,$EA,$7E,$31,$9B,$F8,$6F,$F7,$4F,$A9,$78,$34,$DD,$F8,$7C
	.dc.b	$3F,$9D,$E2,$48,$C1,$E5,$7F,$BE,$FD,$4F,$F4,$18,$F7,$B5,$F1,$C7
	.dc.b	$D7,$F4,$7E,$C7,$24,$9F,$74,$57,$E8,$FC,$AF,$D0,$31,$A7,$F8,$73
	.dc.b	$EA,$7F,$B4,$37,$C3,$FF,$06,$C1,$BF,$3B,$87,$3F,$1F,$30,$FA,$75
	.dc.b	$B6,$C7,$D1,$F8,$D9,$18,$37,$EA,$76,$F8,$E9,$18,$39,$BF,$A3,$F2
	.dc.b	$64,$60,$EF,$F0,$3E,$6F,$A9,$91,$83,$C6,$FF,$89,$B3,$FE,$97,$1F
	.dc.b	$B8,$F6,$F6,$B6,$FF,$1E,$FD,$E7,$F8,$7A,$3F,$1A,$FF,$87,$9B,$EC
	.dc.b	$7F,$8C,$23,$51,$C7,$8F,$F4,$11,$B8,$C1,$FC,$1F,$EA,$3F,$86,$B8
	.dc.b	$0E,$2F,$C3,$7F,$E3,$CF,$AA,$FE,$AF,$1F,$E5,$07,$FA,$DE,$1F,$AD
	.dc.b	$6D,$FF,$13,$77,$19,$F8,$4F,$F2,$2E,$0F,$32,$9D,$63,$B6,$FD,$FD
	.dc.b	$29,$8F,$EB,$31,$F2,$63,$DE,$79,$92,$74,$FF,$94,$BE,$A9,$D0,$F9
	.dc.b	$7F,$F3,$DC,$61,$BE,$73,$E4,$B3,$5B,$C3,$FD,$07,$B3,$DD,$FF,$37
	.dc.b	$C7,$F0,$B8,$C6,$7F,$E8,$66,$C7,$09,$DB,$F3,$FE,$62,$BF,$39,$B3
	.dc.b	$FE,$E9,$AE,$51,$AF,$7D,$EC,$52,$4F,$FA,$52,$32,$DE,$CE,$04,$70
	.dc.b	$4C,$3E,$F4,$FF,$AE,$9B,$F2,$A9,$FE,$33,$39,$E5,$FF,$46,$BF,$E6
	.dc.b	$C8,$CB,$BA,$E7,$CD,$6B,$73,$6B,$F9,$83,$C6,$BC,$2E,$5F,$EB,$0E
	.dc.b	$67,$8C,$F8,$AB,$55,$FE,$7E,$AA,$6E,$7A,$B8,$9C,$B6,$3C,$E7,$32
	.dc.b	$CF,$A1,$B5,$EE,$AE,$31,$B9,$F3,$B6,$7C,$B1,$C4,$C9,$5C,$9C,$E3
	.dc.b	$A9,$F0,$A7,$AC,$EE,$3D,$27,$52,$B9,$3E,$2D,$7A,$DE,$F5,$D5,$B3
	.dc.b	$D4,$EE,$FC,$CA,$EF,$3B,$8F,$F6,$FD,$7D,$36,$CF,$E3,$C9,$5F,$8D
	.dc.b	$6C,$7F,$B9,$27,$83,$DC,$F3,$1F,$C2,$3D,$FF,$35,$F9,$49,$7F,$A7
	.dc.b	$8C,$BC,$63,$33,$FC,$4E,$72,$FA,$67,$E2,$39,$6B,$FE,$C0,$BD,$BE
	.dc.b	$18,$CD,$FE,$16,$39,$1F,$C7,$74,$5F,$09,$D4,$F3,$AF,$EF,$35,$A7
	.dc.b	$9D,$F6,$8C,$4F,$C0,$AE,$27,$28,$CE,$1E,$C9,$E8,$7C,$68,$CD,$FC
	.dc.b	$12,$47,$9A,$7F,$DF,$73,$A8,$C4,$E9,$E7,$3A,$C3,$D5,$BF,$E9,$AD
	.dc.b	$ED,$7E,$9A,$30,$DC,$3F,$84,$47,$F0,$D2,$3F,$C2,$F0,$FE,$20,$47
	.dc.b	$04,$7F,$84,$9D,$FD,$8F,$D9,$BF,$B6,$FA,$A7,$CF,$FE,$E3,$EA,$9E
	.dc.b	$D6,$63,$0D,$C3,$FC,$69,$7F,$E3,$5B,$E3,$18,$8F,$F2,$E4,$61,$C3
	.dc.b	$5E,$2F,$F8,$72,$3F,$C6,$2D,$8F,$EB,$8B,$DF,$F3,$79,$BE,$76,$57
	.dc.b	$E0,$3A,$6D,$F0,$E6,$F6,$1E,$DE,$0B,$5D,$9C,$3F,$E4,$0F,$7D,$F5
	.dc.b	$38,$BF,$0F,$CA,$7C,$53,$E3,$6B,$6F,$CE,$E1,$E5,$F3,$96,$BF,$BA
	.dc.b	$7C,$4F,$F5,$99,$FF,$26,$B6,$3F,$BC,$7D,$57,$F9,$92,$73,$AF,$F2
	.dc.b	$88,$C4,$EB,$E0,$3F,$95,$FC,$3F,$C3,$BF,$F2,$BD,$6B,$A7,$F9,$E5
	.dc.b	$A3,$D7,$3A,$D7,$A7,$BB,$D5,$B1,$CA,$3F,$9B,$FD,$53,$CE,$CE,$1E
	.dc.b	$6F,$8F,$F2,$1E,$5F,$72,$E1,$E2,$70,$F9,$3D,$DB,$5E,$E7,$D8,$7F
	.dc.b	$44,$FA,$57,$C2,$62,$BD,$7C,$7F,$04,$F8,$5E,$E5,$8F,$7B,$87,$FC
	.dc.b	$EF,$A3,$EA,$B1,$97,$CC,$E5,$F9,$47,$CB,$F6,$ED,$FC,$87,$0F,$FA
	.dc.b	$03,$EA,$BE,$53,$FE,$8F,$E3,$4F,$F5,$AF,$A9,$7D,$55,$31,$EF,$5B
	.dc.b	$FF,$44,$CA,$33,$18,$F6,$6E,$7F,$E3,$7A,$FF,$31,$4E,$4F,$F4,$8F
	.dc.b	$0B,$8F,$C1,$7E,$1F,$90,$95,$DB,$DD,$B8,$F3,$7C,$C7,$CD,$E5,$B3
	.dc.b	$E4,$F6,$FD,$47,$AF,$A2,$C7,$A1,$C3,$E2,$F8,$9B,$87,$9F,$E7,$7E
	.dc.b	$11,$F0,$FD,$2B,$5F,$01,$C3,$F0,$BD,$2B,$5E,$F7,$E5,$1F,$81,$F8
	.dc.b	$2B,$3F,$23,$CF,$F9,$C7,$CF,$AB,$1F,$31,$C3,$F1,$BD,$6B,$1F,$E6
	.dc.b	$1F,$AA,$FE,$11,$E1,$FE,$89,$B5,$FE,$C9,$7F,$51,$3B,$F3,$7F,$C2
	.dc.b	$B1,$8A,$E1,$E6,$FE,$57,$B7,$2E,$BD,$C9,$8F,$FB,$1A,$DE,$DE,$32
	.dc.b	$F1,$3D,$0E,$BE,$CD,$AE,$0E,$1F,$BE,$59,$FF,$3D,$7D,$57,$E4,$23
	.dc.b	$FD,$A6,$D8,$CC,$7F,$8B,$23,$19,$FF,$00,$DB,$FC,$97,$5F,$E8,$36
	.dc.b	$C6,$79,$7F,$C6,$71,$8B,$FF,$76,$7D,$57,$F3,$5A,$FE,$12,$D8,$CF
	.dc.b	$BF,$FE,$3A,$8C,$60,$0F,$CE,$70,$D7,$3B,$71,$E0,$E5,$E7,$CC,$3F
	.dc.b	$DD,$F6,$F8,$FC,$61,$FE,$9B,$FA,$37,$C7,$7F,$92,$9B,$7F,$EF,$65
	.dc.b	$F3,$39,$9D,$73,$CA,$D7,$3D,$C3,$FD,$4D,$F5,$4F,$03,$FC,$AA,$DA
	.dc.b	$FF,$80,$2B,$C0,$CC,$EB,$DF,$95,$AF,$80,$B8,$74,$39,$9F,$E6,$08
	.dc.b	$F8,$EB,$0F,$91,$1F,$27,$93,$F9,$BB,$FE,$07,$9F,$F3,$D7,$B3,$C7
	.dc.b	$DF,$E4,$FF,$9A,$BE,$A9,$F4,$14,$C7,$B1,$70,$E8,$7D,$5C,$1F,$54
	.dc.b	$FA,$9F,$F3,$9B,$67,$73,$F6,$E3,$E1,$3F,$EC,$B7,$4D,$F1,$38,$9C
	.dc.b	$B7,$B5,$CA,$5F,$54,$F1,$3F,$D0,$2D,$9D,$3F,$A8,$6C,$6B,$E0,$24
	.dc.b	$6B,$6B,$7F,$ED,$DF,$7B,$FA,$3C,$F7,$6C,$67,$6F,$F4,$7C,$63,$87
	.dc.b	$FB,$1A,$F6,$72,$6B,$D9,$B8,$7A,$7E,$A9,$F1,$FF,$E9,$86,$C6,$78
	.dc.b	$B1,$39,$E8,$93,$C7,$A6,$EB,$FE,$F3,$C3,$F1,$53,$93,$B7,$C7,$F5
	.dc.b	$AE,$DF,$13,$27,$C5,$E4,$5C,$39,$3E,$33,$C9,$E8,$D9,$F4,$76,$F9
	.dc.b	$CF,$47,$93,$63,$D8,$E1,$FA,$DC,$CE,$BC,$62,$E4,$FF,$04,$8F,$83
	.dc.b	$FF,$5C,$B6,$FE,$F3,$FE,$44,$6C,$6B,$C7,$91,$DB,$E0,$2D,$F5,$8F
	.dc.b	$F2,$3F,$A3,$FE,$C4,$8F,$F6,$B4,$7F,$92,$BB,$B1,$39,$64,$7E,$36
	.dc.b	$5D,$D8,$7A,$1F,$ED,$68,$D4,$7F,$B4,$34,$FC,$04,$60,$E5,$B7,$84
	.dc.b	$F1,$6D,$63,$B3,$87,$C1,$6F,$1F,$F9,$07,$D5,$3D,$3F,$F3,$0D,$FA
	.dc.b	$EB,$47,$F9,$84,$FE,$4D,$55,$A5,$75,$F5,$3F,$E9,$CF,$AA,$BE,$09
	.dc.b	$A4,$AF,$F3,$64,$FF,$A1,$B8,$F8,$F8,$DF,$66,$72,$D7,$FC,$EB,$3B
	.dc.b	$78,$FC,$5F,$F0,$5A,$FF,$23,$CF,$DF,$9F,$DD,$35,$F9,$A1,$4F,$1D
	.dc.b	$7F,$87,$46,$EB,$FC,$A3,$BF,$F3,$8B,$FC,$7F,$97,$B7,$FA,$45,$6E
	.dc.b	$F8,$F2,$E6,$7B,$79,$B6,$BF,$A2,$47,$92,$71,$39,$F9,$09,$3E,$88
	.dc.b	$AF,$A3,$1F,$F1,$53,$FA,$0F,$6F,$39,$FF,$99,$B8,$BF,$17,$3F,$C6
	.dc.b	$57,$27,$8D,$B1,$9E,$5C,$4F,$9F,$0F,$E6,$73,$FA,$CF,$1F,$2F,$F2
	.dc.b	$3F,$AA,$F8,$ED,$EB,$F7,$47,$C1,$8F,$ED,$4E,$9F,$9E,$E5,$7D,$57
	.dc.b	$F9,$2D,$FE,$5C,$D7,$73,$F2,$7E,$AB,$FC,$D7,$1F,$CC,$DF,$2B,$F6
	.dc.b	$9B,$A7,$FA,$19,$DC,$67,$0E,$5D,$D2,$39,$47,$5F,$89,$ED,$CB,$27
	.dc.b	$B4,$57,$B7,$A7,$E3,$31,$8F,$7F,$03,$CD,$C5,$2B,$CD,$1E,$9E,$7F
	.dc.b	$C1,$7D,$4B,$EA,$A9,$9A,$FE,$9D,$0F,$13,$7E,$24,$9C,$C7,$E2,$FD
	.dc.b	$57,$B3,$7E,$8E,$57,$F5,$79,$CB,$7D,$AE,$1F,$55,$23,$06,$FB,$7A
	.dc.b	$77,$48,$C1,$FD,$87,$EA,$BF,$C2,$6D,$E0,$FC,$E4,$1B,$C7,$F9,$FF
	.dc.b	$AA,$79,$97,$79,$87,$F6,$B9,$CB,$FD,$B7,$EA,$BF,$DE,$AD,$E0,$FD
	.dc.b	$44,$7A,$A3,$F6,$EE,$9B,$E3,$6B,$C6,$91,$CE,$38,$F8,$BC,$B7,$93
	.dc.b	$DE,$3F,$68,$7E,$ED,$FE,$06,$FB,$9B,$A6,$BA,$F7,$F0,$7C,$3E,$2E
	.dc.b	$57,$F8,$16,$D3,$FB,$FF,$AA,$FF,$19,$4F,$F8,$25,$FF,$18,$37,$D1
	.dc.b	$D7,$A9,$BF,$27,$87,$93,$2B,$84,$7F,$09,$FA,$AF,$F3,$9C,$7F,$84
	.dc.b	$D7,$C2,$64,$D7,$C2,$EF,$D0,$E1,$C5,$2F,$84,$7F,$13,$FA,$AF,$D3
	.dc.b	$F8,$9F,$E2,$B7,$FB,$44,$FF,$87,$DA,$FF,$26,$76,$FF,$99,$57,$B7
	.dc.b	$CF,$C1,$93,$C6,$3F,$90,$7C,$DB,$C3,$BF,$0D,$2B,$FC,$59,$39,$6C
	.dc.b	$70,$F1,$24,$60,$DB,$FC,$AF,$E5,$DF,$17,$4D,$CF,$F3,$CB,$FF,$1B
	.dc.b	$3D,$B7,$F1,$9E,$4F,$4E,$C7,$F8,$F7,$E3,$3F,$26,$FF,$44,$6F,$9A
	.dc.b	$E0,$FF,$28,$61,$F4,$AB,$B7,$17,$BB,$C0,$91,$EC,$8A,$FF,$4E,$FA
	.dc.b	$97,$D5,$7F,$43,$6C,$7C,$2A,$F8,$FE,$17,$C8,$F2,$A4,$FC,$68,$AF
	.dc.b	$B7,$EA,$BE,$69,$8D,$3C,$D1,$FE,$99,$CD,$3D,$FC,$19,$3F,$42,$3F
	.dc.b	$AA,$7D,$3B,$E6,$EF,$CD,$4A,$FF,$30,$4E,$5B,$1F,$D6,$FE,$A9,$F2
	.dc.b	$74,$E1,$3F,$DB,$29,$EB,$76,$F2,$64,$F4,$8A,$FF,$72,$FA,$97,$D5
	.dc.b	$7F,$67,$6C,$78,$AB,$CF,$EC,$7A,$3A,$E4,$F9,$A2,$BE,$1F,$AA,$F6
	.dc.b	$98,$E2,$F8,$B0,$CD,$3F,$DD,$7E,$A9,$E0,$53,$8D,$7C,$2F,$0B,$E3
	.dc.b	$F8,$72,$30,$7C,$67,$FB,$F7,$D5,$7F,$3D,$8F,$F4,$43,$5F,$89,$1F
	.dc.b	$43,$CF,$FA,$3D,$F2,$7E,$7C,$57,$FC,$2B,$C8,$F9,$5F,$A0,$67,$6F
	.dc.b	$5F,$D5,$7F,$67,$6F,$85,$FE,$0F,$83,$7E,$73,$0E,$7E,$3E,$63,$FE
	.dc.b	$9D,$AD,$B6,$3E,$97,$C6,$C8,$C1,$BF,$4B,$B7,$C7,$48,$C1,$CD,$FC
	.dc.b	$F7,$91,$23,$07,$7F,$7F,$CD,$F5,$52,$30,$78,$DF,$F1,$46,$7F,$D8
	.dc.b	$E3,$F6,$DE,$E6,$DF,$F1,$CF,$D5,$7F,$87,$9B,$EB,$7F,$8B,$E3,$FC
	.dc.b	$3B,$1A,$F5,$B1,$3F,$E3,$3F,$A8,$FE,$0B,$3F,$E4,$26,$F1,$7E,$1A
	.dc.b	$F8,$BF,$88,$FA,$AF,$EA,$0D,$F7,$BF,$C7,$91,$FD,$56,$35,$F5,$18
	.dc.b	$9E,$DE,$74,$C7,$FD,$95,$7B,$EA,$FF,$6D,$3E,$06,$FA,$7D,$0E,$BC
	.dc.b	$56,$BF,$DA,$5A,$7F,$11,$CF,$F3,$6D,$BE,$33,$AF,$5B,$FC,$E9,$3F
	.dc.b	$CB,$63,$FD,$D0,$03,$97,$2D,$7F,$DD,$DF,$55,$F9,$1F,$98,$FF,$40
	.dc.b	$36,$39,$85,$E8,$BE,$4F,$7D,$CB,$FF,$44,$61,$FA,$EC,$AF,$89,$FB
	.dc.b	$4C,$BF,$DF,$93,$97,$AB,$5F,$2F,$CC,$ED,$F5,$42,$31,$1E,$C6,$C6
	.dc.b	$6B,$C3,$ED,$60,$8E,$09,$5F,$83,$8F,$E5,$7B,$7F,$08,$E7,$FF,$5D
	.dc.b	$CF,$82,$73,$5F,$2E,$3F,$D7,$33,$87,$CB,$F1,$15,$CC,$FD,$A8,$C4
	.dc.b	$FF,$B0,$23,$07,$D5,$BF,$04,$71,$C1,$5C,$23,$0D,$AE,$37,$FE,$9C
	.dc.b	$9E,$8F,$E7,$67,$4D,$9E,$C7,$A6,$77,$7F,$8C,$AD,$B6,$77,$5F,$89
	.dc.b	$AD,$C7,$D8,$AD,$B4,$DC,$FC,$03,$C0,$79,$67,$D3,$1A,$79,$C7,$C8
	.dc.b	$8C,$46,$A3,$84,$7B,$AB,$E2,$B4,$F5,$4E,$E3,$33,$E0,$69,$EB,$79
	.dc.b	$99,$9E,$27,$37,$DD,$3C,$E3,$31,$FA,$08,$F9,$AF,$EB,$1A,$E8,$C3
	.dc.b	$F9,$43,$2F,$48,$FF,$75,$46,$23,$E3,$4F,$9A,$7E,$A7,$47,$CC,$9E
	.dc.b	$07,$A2,$7F,$29,$19,$8E,$33,$E6,$9C,$BF,$28,$91,$C5,$EA,$62,$3C
	.dc.b	$11,$D9,$E2,$EC,$C4,$75,$38,$B9,$3C,$26,$8F,$C3,$4F,$96,$BE,$EC
	.dc.b	$68,$3F,$ED,$49,$F1,$1D,$63,$91,$A9,$E5,$3C,$DB,$3C,$27,$BB,$5E
	.dc.b	$D3,$E0,$3B,$3E,$7C,$1E,$D7,$EB,$4F,$D9,$EF,$6F,$F2,$8D,$AF,$D6
	.dc.b	$37,$1F,$E0,$91,$FB,$5F,$6B,$F7,$6D,$AF,$CE,$3F,$81,$23,$F8,$07
	.dc.b	$F8,$BF,$D5,$3D,$8A,$49,$FE,$5B,$29,$D4,$63,$EA,$9D,$9F,$C3,$67
	.dc.b	$34,$C7,$FC,$3F,$5A,$BF,$D9,$D6,$A7,$3C,$5F,$C5,$6B,$F9,$84,$70
	.dc.b	$E1,$C4,$E5,$97,$FD,$4B,$8E,$E3,$DE,$3F,$91,$61,$FC,$BE,$30,$7F
	.dc.b	$4F,$71,$87,$93,$DF,$F0,$72,$78,$59,$7F,$C2,$3D,$E7,$D5,$63,$1E
	.dc.b	$63,$17,$C8,$FA,$4F,$57,$05,$C3,$F8,$8C,$FB,$DF,$2C,$7F,$33,$C1
	.dc.b	$F3,$9C,$27,$F8,$BF,$C3,$FF,$57,$6E,$DF,$CE,$7E,$9F,$FC,$CD,$18
	.dc.b	$3F,$8E,$B8,$CF,$CA,$DD,$FF,$C9,$FF,$24,$E1,$E5,$F0,$1C,$F8,$26
	.dc.b	$3C,$BD,$EB,$97,$27,$C0,$7F,$49,$FA,$A7,$99,$8A,$F0,$FC,$7F,$94
	.dc.b	$F3,$7B,$96,$3C,$AE,$1F,$3F,$BB,$71,$F3,$BE,$D3,$FA,$B7,$D3,$3E
	.dc.b	$0F,$15,$EE,$E4,$F8,$47,$C3,$F7,$2C,$7C,$27,$0F,$F4,$0D,$1F,$55
	.dc.b	$8C,$BE,$2F,$37,$CB,$3E,$67,$B7,$6F,$E5,$B8,$7F,$A4,$FD,$57,$CB
	.dc.b	$7F,$B7,$73,$A7,$FB,$67,$D4,$BE,$AA,$98,$F7,$6D,$FF,$52,$94,$66
	.dc.b	$31,$EB,$DC,$FF,$1B,$5F,$E6,$89,$C9,$FD,$6B,$85,$C7,$E0,$BF,$0F
	.dc.b	$C7,$CA,$ED,$ED,$DC,$79,$BE,$63,$E6,$F2,$D9,$F2,$7B,$3E,$A3,$D7
	.dc.b	$D1,$63,$D0,$E1,$F1,$7C,$45,$C3,$CF,$F3,$9F,$08,$F8,$7E,$F5,$AF
	.dc.b	$80,$E1,$F8,$5F,$43,$6B,$DE,$FC,$A3,$F0,$3F,$07,$67,$E4,$79,$FF
	.dc.b	$38,$F9,$F5,$63,$E6,$38,$7E,$37,$AB,$63,$F6,$FF,$55,$FC,$32,$7F
	.dc.b	$C4,$CD,$8F,$F2,$93,$FA,$58,$FF,$03,$B9,$83,$97,$9F,$89,$F9,$6E
	.dc.b	$BF,$F1,$64,$E7,$C6,$FF,$5E,$AF,$6B,$26,$BD,$BB,$8F,$67,$0F,$83
	.dc.b	$E4,$6E,$5E,$87,$A9,$FA,$48,$A7,$F8,$73,$EA,$BF,$24,$DE,$EF,$F1
	.dc.b	$F4,$22,$C7,$F9,$07,$67,$CC,$53,$1C,$D6,$BF,$C4,$FF,$55,$FC,$DF
	.dc.b	$2F,$E3,$ED,$DB,$FD,$B2,$BE,$0B,$FC,$98,$E6,$BF,$CA,$4F,$C9,$27
	.dc.b	$0E,$BD,$5D,$FA,$32,$BB,$FC,$8D,$CB,$F9,$6F,$D1,$BE,$3F,$FC,$B0
	.dc.b	$DA,$FF,$1D,$AF,$98,$CC,$EB,$A6,$56,$BA,$6E,$1F,$CE,$3E,$A9,$E1
	.dc.b	$7F,$98,$DB,$5F,$E4,$65,$E1,$66,$75,$F0,$12,$B5,$F0,$37,$0E,$87
	.dc.b	$33,$FC,$DD,$1F,$0D,$FF,$4F,$61,$FE,$9F,$8F,$F2,$98,$F6,$F6,$70
	.dc.b	$F8,$0B,$1F,$E5,$8F,$AA,$7D,$07,$F9,$E5,$B1,$9F,$6B,$13,$FE,$7D
	.dc.b	$FA,$A7,$D4,$D3,$1B,$5B,$FF,$31,$CB,$C6,$E0,$71,$DE,$CF,$F9,$A3
	.dc.b	$EA,$BF,$34,$DF,$17,$FD,$15,$0F,$16,$C6,$3E,$0A,$57,$1E,$0B,$C5
	.dc.b	$FD,$A3,$DF,$FE,$79,$3F,$E7,$46,$C6,$77,$FF,$4B,$46,$3C,$1F,$CD
	.dc.b	$4F,$FB,$1E,$3F,$B3,$D7,$1F,$6B,$7E,$D7,$07,$FD,$3D,$F5,$4F,$91
	.dc.b	$A6,$39,$2D,$FF,$A1,$27,$F0,$38,$FF,$7C,$6E,$5E,$08,$47,$E1,$27
	.dc.b	$87,$27,$6E,$A9,$1C,$FE,$46,$DF,$3F,$23,$C6,$79,$3C,$8B,$5E,$17
	.dc.b	$0F,$8F,$D2,$B5,$E3,$EE,$F5,$9E,$CF,$2A,$D7,$A5,$C3,$FE,$9E,$E6
	.dc.b	$76,$62,$FF,$C2,$FB,$FF,$85,$AB,$FD,$46,$D8,$CF,$8D,$FE,$C5,$8C
	.dc.b	$1F,$04,$FF,$0C,$47,$FB,$0E,$3F,$D9,$56,$F7,$F8,$47,$FB,$36,$7F
	.dc.b	$DA,$A7,$FB,$F7,$D1,$C7,$76,$5B,$8F,$FA,$D2,$7F,$24,$DE,$EC,$C7
	.dc.b	$F8,$B8,$FF,$3B,$9F,$E7,$8C,$7F,$02,$E6,$3C,$1A,$CA,$E9,$AD,$AE
	.dc.b	$4F,$94,$F3,$2F,$1C,$BE,$A9,$E9,$FF,$BA,$E1,$5C,$2B,$FC,$C7,$1C
	.dc.b	$1A,$F3,$32,$FF,$9A,$3E,$A7,$F7,$DF,$55,$7C,$03,$49,$5D,$C4,$FF
	.dc.b	$A1,$38,$B9,$78,$DF,$E6,$B9,$D4,$E5,$B1,$97,$D1,$0F,$9E,$F6,$C7
	.dc.b	$F9,$62,$3F,$CA,$EF,$FB,$91,$AE,$63,$75,$FA,$19,$CB,$FA,$3E,$0F
	.dc.b	$E6,$95,$F8,$FF,$2F,$7B,$F6,$BD,$B7,$93,$35,$F3,$A5,$E6,$E2,$3D
	.dc.b	$ED,$8F,$F4,$94,$79,$63,$1F,$27,$23,$07,$A5,$E5,$8F,$F8,$92,$3F
	.dc.b	$45,$ED,$ED,$FF,$3B,$F1,$7F,$AF,$9F,$BF,$A7,$EF,$61,$C9,$E3,$6C
	.dc.b	$EB,$4F,$93,$1C,$63,$9C,$F8,$3C,$69,$1B,$3F,$17,$3C,$9F,$F4,$3F
	.dc.b	$D5,$7D,$37,$41,$F8,$EF,$E4,$9A,$E7,$83,$E3,$46,$3E,$AB,$FC,$93
	.dc.b	$3C,$9A,$F6,$F8,$CF,$99,$F5,$5F,$E6,$98,$EE,$FF,$4A,$6E,$3A,$F8
	.dc.b	$8D,$BC,$3B,$E4,$E8,$E3,$F1,$3D,$39,$A4,$60,$EA,$F8,$3D,$59,$18
	.dc.b	$3C,$3E,$0F,$93,$8E,$46,$0F,$2F,$17,$E0,$FE,$A5,$F5,$54,$F5,$C6
	.dc.b	$7C,$5D,$78,$B2,$FE,$5D,$A7,$F1,$9E,$BF,$F8,$56,$7E,$7A,$F9,$7C
	.dc.b	$9A,$F7,$37,$ED,$F0,$F9,$59,$7C,$23,$9F,$7F,$4E,$F9,$3D,$23,$F3
	.dc.b	$5F,$55,$FE,$13,$8F,$BA,$FF,$25,$6F,$0E,$B8,$BF,$41,$F5,$5F,$E3
	.dc.b	$29,$FC,$1B,$E9,$6E,$DA,$F3,$7F,$4D,$F5,$5F,$EE,$E9,$FC,$4B,$ED
	.dc.b	$7F,$DA,$6D,$F0,$BF,$AC,$3B,$8C,$F8,$DC,$BC,$19,$1C,$A3,$AF,$8D
	.dc.b	$DB,$69,$3D,$A3,$F6,$C7,$F0,$07,$FD,$C2,$FB,$DB,$A6,$BA,$B7,$F0
	.dc.b	$BC,$3E,$0A,$57,$E8,$2D,$3F,$C0,$7E,$AB,$FC,$67,$3F,$A4,$7F,$C6
	.dc.b	$2D,$F4,$F5,$EA,$6F,$D1,$E1,$E5,$4A,$E1,$1F,$C2,$FE,$AB,$FC,$29
	.dc.b	$1F,$AE,$5F,$0D,$93,$5E,$F6,$FD,$1E,$1C,$52,$F8,$47,$F1,$5F,$AA
	.dc.b	$FD,$44,$7F,$C0,$07,$FB,$6B,$FB,$A6,$B9,$F2,$78,$7B,$BC,$7C,$39
	.dc.b	$3B,$8F,$E4,$1F,$39,$F9,$38,$FE,$06,$BE,$1B,$26,$BC,$5B,$3A,$F0
	.dc.b	$F8,$98,$5D,$BF,$E1,$7B,$FF,$7F,$47,$F0,$EF,$97,$7C,$67,$F2,$F6
	.dc.b	$C7,$33,$C3,$EA,$F9,$38,$64,$60,$F2,$BF,$CD,$FE,$37,$F3,$11,$E0
	.dc.b	$F3,$3F,$A6,$C6,$32,$F1,$57,$BB,$8F,$DF,$E0,$CC,$3D,$F1,$5E,$47
	.dc.b	$F7,$2F,$AA,$7C,$E7,$F4,$56,$C7,$C6,$AF,$93,$E1,$FC,$AF,$2A,$4F
	.dc.b	$CA,$8A,$F4,$1C,$0F,$AA,$FA,$26,$B4,$F9,$0F,$7F,$0A,$57,$D4,$8F
	.dc.b	$EA,$DF,$50,$F9,$BB,$EB,$4A,$FE,$77,$39,$6C,$7F,$5D,$FA,$57,$C9
	.dc.b	$D3,$84,$F9,$69,$EB,$F6,$F2,$A4,$F4,$8A,$FA,$FE,$A5,$F5,$5F,$DA
	.dc.b	$5B,$1E,$2A,$F3,$FB,$3E,$8E,$C9,$3E,$68,$AF,$87,$EA,$BD,$A6,$38
	.dc.b	$7F,$CC,$19,$A3,$FB,$B7,$D5,$7E,$56,$BF,$AC,$B5,$E4,$7C,$2F,$0B
	.dc.b	$E3,$F8,$72,$30,$7C,$67,$FC,$01,$F5,$5F,$CF,$A3,$FB,$1B,$5F,$89
	.dc.b	$1F,$43,$D0,$FA,$3E,$04,$9F,$9F,$15,$FB,$DC,$6F,$95,$FA,$06,$75
	.dc.b	$FF,$3D,$7D,$57,$F6,$66,$F8,$5F,$E1,$18,$37,$E7,$30,$E7,$E4,$26
	.dc.b	$3F,$DE,$2B,$6D,$8F,$A5,$F1,$92,$30,$6F,$D3,$ED,$F1,$D2,$30,$73
	.dc.b	$7F,$3D,$E4,$48,$C1,$DF,$E0,$7C,$DF,$55,$23,$07,$8D,$FD,$A7,$B9
	.dc.b	$23,$07,$F8,$3B,$EA,$BF,$C3,$1B,$FF,$A6,$DF,$F0,$A3,$7D,$7F,$F1
	.dc.b	$94,$6B,$93,$FC,$8E,$E9,$FF,$24,$FD,$43,$3F,$F1,$C3,$70,$ED,$7F
	.dc.b	$E3,$BF,$AA,$FE,$A1,$8F,$FA,$A5,$FE,$A9,$87,$F9,$49,$B5,$FE,$70
	.dc.b	$77,$19,$F8,$4E,$BE,$5C,$AF,$F1,$3D,$70,$4F,$85,$FD,$F3,$FD,$01
	.dc.b	$BE,$87,$8F,$3F,$23,$E8,$24,$74,$7F,$CA,$5F,$B3,$FF,$8D,$E3,$FD
	.dc.b	$00,$D7,$FC,$D2,$D7,$FD,$0E,$D4,$FF,$A1,$9F,$F3,$D1,$8F,$91,$FF
	.dc.b	$43,$C6,$63,$FC,$FC,$D7,$FD,$1D,$F5,$7F,$99,$F9,$8E,$2D,$8F,$F4
	.dc.b	$23,$FC,$6A,$B3,$D6,$AE,$FC,$1F,$D7,$B6,$BE,$E7,$FB,$67,$1F,$B2
	.dc.b	$D7,$FD,$3F,$3B,$7F,$D2,$0B,$EA,$F5,$08,$C5,$76,$9C,$C6,$27,$8C
	.dc.b	$7F,$BD,$B8,$7F,$DD,$13,$97,$FD,$D5,$A7,$FA,$BA,$FF,$CC,$33,$97
	.dc.b	$FD,$2B,$EF,$FF,$B0,$B0,$FE,$70,$FF,$B2,$8C,$5F,$FB,$CD,$C7,$CB
	.dc.b	$77,$AE,$3D,$76,$FF,$D7,$75,$C2,$72,$7F,$9C,$1C,$46,$A7,$82,$8C
	.dc.b	$B6,$33,$18,$3C,$33,$CE,$33,$5A,$6C,$64,$F0,$46,$5A,$66,$BD,$F1
	.dc.b	$96,$C6,$6F,$E0,$4E,$5B,$19,$3E,$4C,$65,$A6,$67,$E1,$C6,$5B,$19
	.dc.b	$7E,$04,$65,$B1,$9B,$F8,$17,$EC,$9F,$AA,$D7,$FD,$CD,$1F,$65,$B5
	.dc.b	$F8,$A8,$C4,$79,$0F,$F6,$F3,$87,$C6,$62,$38,$1D,$EF,$F1,$7C,$97
	.dc.b	$0E,$51,$F0,$AB,$FB,$F6,$2A,$D4,$74,$9D,$1C,$AB,$E8,$C7,$CE,$24
	.dc.b	$65,$B1,$D8,$8F,$EA,$27,$AB,$F1,$8D,$99,$8E,$B3,$E0,$78,$C7,$EA
	.dc.b	$E7,$8E,$1F,$8B,$C3,$F2,$38,$7A,$6F,$89,$C2,$7F,$27,$19,$BE,$E7
	.dc.b	$87,$0E,$CF,$84,$CC,$F5,$AF,$D7,$E1,$E0,$C3,$D3,$87,$3C,$38,$E1
	.dc.b	$FA,$1E,$5F,$CF,$9E,$4F,$7E,$E3,$FA,$98,$B3,$F4,$26,$D2,$93,$3E
	.dc.b	$AD,$EB,$FE,$B9,$B3,$FE,$E2,$5F,$21,$99,$D3,$FC,$4A,$33,$3F,$CB
	.dc.b	$1B,$5D,$88,$E1,$FC,$99,$F9,$12,$47,$F3,$4D,$7F,$89,$47,$E4,$8F
	.dc.b	$DE,$9F,$4A,$EC,$FF,$BD,$BE,$AF,$FC,$E8,$DF,$80,$FE,$0F,$0F,$7E
	.dc.b	$C6,$19,$3F,$C7,$B9,$6E,$1F,$E5,$3B,$7C,$E7,$23,$9E,$D6,$FC,$0E
	.dc.b	$1F,$F3,$D7,$D5,$DB,$E5,$B7,$88,$F1,$FB,$76,$BD,$0E,$1F,$F4,$17
	.dc.b	$CE,$DA,$FF,$32,$7C,$4F,$F4,$99,$FA,$FC,$5F,$66,$7E,$CC,$E5,$FF
	.dc.b	$35,$7D,$47,$F9,$86,$B6,$7D,$19,$D5,$71,$A5,$6D,$E3,$EE,$FF,$2E
	.dc.b	$F9,$4F,$91,$87,$29,$FC,$3D,$6E,$F9,$2D,$3F,$9A,$B6,$B8,$57,$6F
	.dc.b	$73,$35,$A6,$4E,$BE,$AD,$E5,$62,$FD,$7F,$23,$D0,$EB,$ED,$5B,$EB
	.dc.b	$92,$33,$18,$F6,$ED,$FC,$47,$CA,$7F,$4A,$FA,$77,$C0,$E2,$B9,$F1
	.dc.b	$7A,$8F,$5F,$B5,$63,$BB,$87,$D1,$F5,$77,$1F,$A3,$C9,$F0,$8F,$87
	.dc.b	$EC,$DA,$F3,$B8,$7E,$2F,$D5,$FC,$B7,$F4,$DB,$27,$FA,$FF,$D5,$3E
	.dc.b	$AE,$98,$F6,$ED,$FC,$E9,$46,$63,$1E,$DD,$CF,$FD,$5D,$5F,$E6,$A9
	.dc.b	$C9,$F5,$B8,$3E,$53,$89,$DF,$C8,$F8,$8F,$1F,$B5,$67,$C0,$E1,$E5
	.dc.b	$C7,$6B,$BF,$B1,$EA,$3D,$7C,$F6,$BD,$0E,$1F,$17,$C3,$5A,$F3,$FC
	.dc.b	$F7,$C2,$3E,$1F,$9F,$6B,$E0,$38,$7E,$17,$9D,$6B,$DE,$FC,$A3,$F0
	.dc.b	$1F,$01,$67,$E4,$7A,$BF,$38,$F9,$F5,$63,$E6,$38,$7E,$37,$AB,$63
	.dc.b	$FD,$B1,$F5,$7F,$C4,$1B,$E6,$FF,$85,$A3,$F5,$31,$8A,$FD,$1C,$63
	.dc.b	$C4,$FD,$93,$63,$F9,$37,$1B,$AF,$97,$9B,$CF,$D9,$95,$CB,$D9,$B9
	.dc.b	$78,$9F,$01,$E1,$F6,$2C,$F8,$1C,$3F,$7C,$B3,$FE,$F1,$FA,$BF,$C8
	.dc.b	$47,$EE,$1B,$19,$8F,$F1,$8C,$63,$4F,$E1,$5B,$BE,$63,$FC,$6A,$DA
	.dc.b	$FE,$02,$B9,$F3,$3F,$E3,$AF,$AB,$FF,$26,$63,$FD,$49,$A9,$FE,$C1
	.dc.b	$7F,$D2,$5C,$9F,$3F,$0F,$E3,$4D,$8D,$6B,$D7,$87,$B7,$9D,$27,$F8
	.dc.b	$9D,$3E,$3B,$18,$B1,$FE,$49,$F4,$8F,$8B,$FF,$27,$B7,$0F,$E3,$6B
	.dc.b	$E6,$73,$3A,$E8,$95,$AE,$8B,$8F,$58,$FE,$C5,$F5,$6F,$07,$FC,$B6
	.dc.b	$D8,$FE,$4E,$BC,$1C,$CE,$BD,$A9,$5A,$F7,$2E,$1D,$4F,$EF,$78,$7F
	.dc.b	$68,$DB,$F9,$F3,$6B,$FB,$91,$F3,$9C,$4F,$F9,$FF,$D7,$FF,$36,$C9
	.dc.b	$BF,$29,$87,$FC,$E3,$F5,$6F,$98,$A6,$3D,$8B,$7F,$D0,$3E,$AD,$F5
	.dc.b	$5F,$E7,$86,$C6,$77,$C4,$E7,$C5,$93,$C7,$82,$EB,$ED,$1F,$57,$FE
	.dc.b	$DD,$9F,$E9,$AD,$8D,$2F,$0B,$FD,$0E,$E7,$87,$FB,$03,$FD,$F2,$B8
	.dc.b	$FB,$DA,$E0,$DB,$FE,$8E,$F8,$0F,$E8,$B3,$FD,$69,$B8,$F8,$DF,$85
	.dc.b	$19,$F6,$38,$7B,$12,$38,$7B,$16,$FD,$5F,$56,$F8,$EF,$F4,$DB,$6B
	.dc.b	$FB,$3B,$F1,$A7,$73,$AE,$29,$5A,$EA,$B5,$D1,$E6,$74,$F8,$EB,$3D
	.dc.b	$3C,$6E,$E7,$83,$C7,$B1,$E0,$70,$F8,$BD,$2B,$87,$A3,$BB,$CE,$7A
	.dc.b	$3C,$9B,$5E,$87,$0F,$F8,$03,$CC,$76,$62,$FE,$D7,$3F,$F9,$10,$FF
	.dc.b	$09,$16,$BF,$C8,$CB,$C8,$FF,$5F,$B9,$9F,$F1,$0C,$60,$FF,$61,$DB
	.dc.b	$DD,$C5,$FF,$84,$29,$E8,$FF,$B2,$1B,$19,$ED,$C4,$E5,$91,$C7,$B6
	.dc.b	$D7,$E4,$4F,$F6,$84,$64,$F7,$73,$F5,$9E,$DB,$49,$F1,$6D,$6F,$93
	.dc.b	$E3,$3C,$8B,$C8,$FF,$39,$FD,$5B,$D3,$FF,$34,$4A,$B7,$58,$8C,$B5
	.dc.b	$FF,$37,$35,$FF,$3E,$B9,$FA,$AF,$DE,$7D,$5D,$F0,$0D,$25,$75,$97
	.dc.b	$8F,$FC,$96,$30,$D7,$CB,$1A,$E4,$ED,$39,$9F,$F4,$04,$FF,$A3,$1F
	.dc.b	$F2,$1B,$FE,$8A,$6F,$C2,$FE,$03,$5D,$FC,$3D,$9F,$F4,$AD,$6D,$FE
	.dc.b	$05,$CB,$FC,$D5,$FC,$77,$9B,$B5,$3B,$CB,$FC,$BC,$FE,$2E,$3C,$3E
	.dc.b	$7E,$E3,$FA,$14,$79,$3A,$31,$39,$F8,$F9,$3C,$0F,$4B,$DF,$1F,$F2
	.dc.b	$84,$7E,$7F,$E0,$F6,$FF,$17,$E3,$73,$79,$57,$35,$31,$16,$7F,$51
	.dc.b	$E6,$75,$30,$FF,$21,$FA,$BF,$94,$DE,$2F,$81,$1E,$FE,$2F,$F5,$73
	.dc.b	$AE,$0F,$F3,$F3,$9A,$FF,$5F,$FD,$5F,$F9,$19,$FF,$3A,$B5,$E8,$7F
	.dc.b	$9E,$3E,$AF,$FC,$CA,$DC,$7E,$74,$7F,$41,$8D,$79,$38,$9C,$F8,$12
	.dc.b	$78,$1C,$FE,$23,$8F,$34,$8C,$1D,$DE,$BE,$AC,$8C,$1E,$3F,$0B,$C3
	.dc.b	$C7,$23,$07,$9F,$7F,$C1,$7D,$53,$EA,$E9,$E8,$8C,$F8,$DA,$F1,$A5
	.dc.b	$FF,$A5,$AF,$FC,$2C,$FF,$A6,$3D,$77,$C8,$7E,$35,$7D,$25,$31,$19
	.dc.b	$F7,$35,$F5,$72,$F9,$47,$1E,$FE,$5D,$F2,$7B,$47,$E6,$7E,$AF,$FC
	.dc.b	$19,$1F,$EA,$87,$F8,$E3,$78,$75,$C5,$F9,$FF,$AB,$FF,$18,$4F,$FA
	.dc.b	$B9,$F4,$BF,$D9,$DB,$D9,$FE,$BF,$77,$1F,$DA,$FE,$AD,$F2,$FF,$A9
	.dc.b	$7D,$F1,$FE,$B8,$6B,$B8,$D7,$8F,$BF,$1E,$4E,$E3,$97,$8D,$CF,$59
	.dc.b	$3C,$E3,$F6,$C7,$EE,$5F,$EF,$8F,$C0,$6E,$7A,$EA,$DF,$C2,$70,$F8
	.dc.b	$D9,$5F,$EC,$CB,$FF,$33,$3F,$EC,$EF,$AB,$FF,$16,$37,$D5,$FE,$09
	.dc.b	$1F,$E2,$B8,$D7,$AD,$8F,$23,$3E,$4C,$9E,$51,$FC,$2F,$EA,$FF,$C1
	.dc.b	$51,$FE,$DA,$5F,$0F,$93,$5E,$FE,$FD,$1E,$1C,$52,$F8,$47,$F1,$5F
	.dc.b	$AB,$FD,$44,$7F,$86,$DF,$EC,$AF,$FB,$99,$AE,$7C,$9E,$1E,$EF,$1F
	.dc.b	$0A,$4E,$E3,$F9,$07,$D4,$7E,$52,$3F,$DD,$EB,$E2,$32,$6B,$C4,$B3
	.dc.b	$AF,$0B,$89,$87,$FC,$5D,$48,$FE,$59,$F5,$6F,$8E,$A7,$0A,$EF,$EA
	.dc.b	$F8,$B8,$24,$60,$F1,$BF,$CD,$7E,$43,$F2,$3E,$1B,$CC,$54,$9C,$C6
	.dc.b	$21,$9F,$D2,$7C,$BC,$7F,$03,$C1,$91,$F0,$22,$BF,$E4,$7F,$AA,$7D
	.dc.b	$5F,$F4,$56,$C7,$C6,$AF,$93,$E1,$FC,$AF,$26,$4F,$CA,$8A,$FC,$4F
	.dc.b	$AB,$F9,$E6,$35,$F5,$47,$AB,$34,$F7,$F0,$64,$FD,$48,$FE,$AD,$F5
	.dc.b	$2F,$9C,$BE,$B4,$AF,$87,$8F,$45,$5B,$73,$C1,$F1,$63,$F9,$AC,$7F
	.dc.b	$99,$7E,$99,$F2,$BF,$D8,$DB,$1C,$0F,$9F,$3C,$2F,$B2,$F5,$FC,$BE
	.dc.b	$54,$AE,$07,$8A,$30,$FF,$9C,$FE,$A9,$F5,$7F,$DC,$5B,$1C,$2F,$F7
	.dc.b	$D8,$76,$F6,$7E,$0F,$5C,$AD,$45,$7F,$CF,$D4,$FA,$BF,$8E,$66,$FC
	.dc.b	$EF,$F7,$FF,$AB,$FC,$F5,$79,$9A,$F2,$8F,$9F,$E1,$7D,$0F,$0E,$4F
	.dc.b	$96,$2B,$FE,$8D,$FA,$BF,$E9,$0F,$C0,$6B,$E4,$9F,$B3,$E7,$FD,$AE
	.dc.b	$39,$3E,$F8,$AF,$D3,$F9,$6F,$A4,$63,$3F,$B9,$F5,$7F,$DB,$DB,$E0
	.dc.b	$FF,$86,$60,$DF,$9D,$C3,$9F,$90,$98,$7D,$4A,$DB,$63,$E9,$7C,$74
	.dc.b	$8C,$1B,$F5,$3B,$7C,$84,$8C,$1C,$DF,$D3,$79,$12,$30,$77,$F8,$0F
	.dc.b	$37,$D5,$C8,$C1,$E3,$7F,$7D,$F0,$52,$30,$7E,$2B,$EA,$FF,$C4,$6D
	.dc.b	$F5,$FF,$C7,$11,$FE,$24,$8D,$7A,$F8,$9F,$F1,$E7,$D4,$32,$A6,$6B
	.dc.b	$F2,$BF,$57,$FD,$69,$BF,$09,$FE,$43,$8F,$EB,$3C,$9F,$AB,$75,$3F
	.dc.b	$EC,$B7,$3C,$FF,$C3,$F0,$79,$9E,$03,$96,$1C,$F1,$E9,$E8,$CC,$7C
	.dc.b	$31,$5F,$F6,$B5,$FF,$17,$CB,$F7,$4D,$C7,$F8,$E6,$56,$BF,$90,$FD
	.dc.b	$5F,$F4,$47,$FC,$FE,$D7,$F9,$13,$5D,$FA,$FF,$E7,$DF,$AB,$F9,$11
	.dc.b	$FE,$83,$6B,$FE,$88,$6B,$FC,$AD,$AF,$F3,$0C,$3F,$A4,$3F,$E8,$B6
	.dc.b	$BF,$E9,$06,$BF,$17,$D7,$FF,$46,$5B,$B1,$F3,$1F,$E9,$18,$C3,$7D
	.dc.b	$1F,$F4,$C5,$FF,$A5,$E7,$2F,$FA,$67,$8F,$FD,$3D,$AF,$FA,$65,$B1
	.dc.b	$9F,$8F,$FE,$44,$FF,$43,$8C,$BF,$EA,$47,$7B,$FF,$27,$6D,$7D,$BF
	.dc.b	$83,$FE,$6B,$39,$BB,$3C,$FB,$2E,$1F,$0F,$DA,$AF,$F9,$E2,$2E,$FF
	.dc.b	$D8,$EB,$F9,$1E,$5F,$EC,$AA,$FB,$F7,$C2,$31,$8F,$E0,$23,$18,$72
	.dc.b	$6D,$FF,$2C,$C3,$A6,$0B,$F5,$5E,$EB,$BB,$67,$51,$E9,$39,$5F,$48
	.dc.b	$F1,$C7,$03,$25,$72,$72,$3F,$B8,$E1,$E8,$8F,$D1,$4F,$3B,$F5,$47
	.dc.b	$16,$D7,$13,$E3,$47,$67,$15,$CD,$F9,$91,$C4,$EA,$57,$27,$0B,$F7
	.dc.b	$57,$C2,$9E,$8D,$9F,$4B,$4F,$19,$63,$AC,$73,$7F,$19,$3F,$8A,$8F
	.dc.b	$7B,$5F,$49,$F9,$37,$15,$F7,$63,$F1,$D1,$82,$4F,$A6,$3F,$D7,$BF
	.dc.b	$2D,$F1,$E3,$EA,$47,$84,$F8,$71,$97,$C3,$1C,$69,$39,$9F,$1D,$FA
	.dc.b	$DF,$AB,$19,$9F,$99,$1A,$AF,$D2,$4E,$6F,$4D,$8E,$CD,$9F,$25,$FE
	.dc.b	$9E,$30,$F7,$8F,$5C,$F9,$6B,$9B,$4E,$65,$8E,$6F,$88,$F6,$74,$7E
	.dc.b	$8F,$0F,$D2,$36,$3F,$52,$FC,$E2,$3F,$C0,$5F,$A9,$F5,$6F,$62,$92
	.dc.b	$7F,$C8,$97,$FE,$45,$9D,$46,$23,$FC,$97,$18,$7F,$2C,$DC,$7F,$C5
	.dc.b	$8B,$E4,$7F,$7E,$E6,$BF,$87,$57,$F0,$0F,$6F,$F7,$8D,$8F,$E0,$2F
	.dc.b	$E1,$88,$FF,$19,$7F,$19,$F5,$7F,$E7,$18,$FF,$14,$B6,$3F,$50,$FF
	.dc.b	$00,$C3,$FC,$7C,$E5,$BB,$7F,$12,$8D,$57,$F5,$5E,$4C,$4E,$7E,$6E
	.dc.b	$4F,$6F,$82,$B8,$F6,$79,$1F,$C7,$3E,$01,$F5,$78,$AF,$2E,$DD,$CF
	.dc.b	$07,$BD,$63,$D2,$E1,$FF,$21,$D3,$E6,$B1,$8F,$F0,$3F,$89,$FE,$87
	.dc.b	$5F,$E4,$86,$C7,$F0,$7F,$A8,$FF,$30,$4E,$76,$FE,$5F,$18,$9D,$7C
	.dc.b	$07,$F3,$2F,$93,$F8,$31,$FE,$86,$9D,$71,$7F,$99,$5F,$5E,$1F,$E5
	.dc.b	$66,$CF,$7F,$77,$AE,$FD,$99,$5F,$E5,$CF,$AB,$79,$B8,$BF,$37,$C8
	.dc.b	$79,$0F,$2F,$B9,$6F,$B3,$87,$C1,$EE,$DA,$F7,$3E,$C3,$FA,$5F,$D3
	.dc.b	$3E,$13,$15,$E7,$E3,$F8,$27,$C2,$F7,$2C,$7A,$DC,$3E,$7F,$AB,$B8
	.dc.b	$FC,$BE,$5F,$92,$7C,$AF,$6E,$D7,$C3,$70,$FF,$9E,$BE,$AF,$E4,$BF
	.dc.b	$EA,$DE,$34,$FF,$60,$FA,$A7,$D5,$D3,$1E,$F5,$BF,$F4,$1C,$A3,$31
	.dc.b	$8F,$7A,$E7,$FD,$16,$BF,$CB,$73,$93,$FD,$17,$C0,$E2,$F0,$5F,$87
	.dc.b	$E4,$65,$76,$F7,$6E,$3C,$DF,$31,$F3,$39,$2C,$F9,$3D,$BF,$51,$EB
	.dc.b	$E7,$B1,$E8,$70,$F8,$BE,$26,$E1,$E7,$F3,$FE,$11,$F0,$FC,$DB,$5F
	.dc.b	$01,$C3,$F0,$BC,$EB,$5E,$F7,$E5,$1F,$80,$F8,$1B,$3F,$23,$E8,$3E
	.dc.b	$71,$F3,$EA,$C7,$CC,$70,$FC,$6F,$56,$C7,$F6,$FF,$AB,$FE,$0D,$3F
	.dc.b	$EE,$56,$C7,$FB,$A9,$FD,$3A,$72,$DF,$33,$FC,$3B,$1A,$E4,$75,$E2
	.dc.b	$73,$ED,$CC,$3F,$D8,$36,$F7,$31,$8F,$8D,$E6,$74,$F9,$2B,$5C,$1C
	.dc.b	$3F,$80,$2C,$FF,$82,$7E,$AF,$F2,$31,$FE,$CE,$6C,$66,$3F,$C6,$71
	.dc.b	$8F,$23,$FD,$C9,$B3,$E9,$3F,$C6,$ED,$AF,$F6,$BA,$E6,$CC,$FF,$8F
	.dc.b	$3E,$AD,$ED,$53,$1F,$03,$70,$FF,$6F,$CF,$C8,$27,$07,$1F,$94,$E5
	.dc.b	$E9,$4A,$DF,$C9,$5D,$FF,$C4,$9F,$48,$F9,$1F,$F2,$5B,$6B,$FD,$D6
	.dc.b	$BE,$67,$33,$AE,$A9,$5A,$EA,$B8,$7F,$8B,$FE,$AD,$E1,$FF,$95,$9B
	.dc.b	$5F,$EF,$65,$E1,$E6,$75,$F0,$12,$B5,$F0,$17,$0E,$87,$33,$FC,$C3
	.dc.b	$1F,$18,$E8,$E3,$DB,$90,$8F,$F2,$12,$F9,$7C,$C7,$F9,$17,$EA,$DF
	.dc.b	$11,$FE,$6C,$6D,$71,$EC,$3C,$E9,$2F,$6B,$85,$FF,$9D,$3E,$AD,$F5
	.dc.b	$74,$C6,$D7,$1E,$87,$39,$FF,$16,$B8,$C7,$D1,$18,$7F,$CB,$4B,$6E
	.dc.b	$DE,$3F,$FA,$0B,$EA,$FF,$DA,$17,$C6,$D1,$87,$FC,$C4,$FE,$96,$34
	.dc.b	$E7,$5F,$6C,$FC,$78,$7C,$17,$53,$B7,$05,$BE,$DF,$01,$FD,$1D,$BE
	.dc.b	$DF,$FA,$52,$3E,$4E,$1F,$E7,$17,$4D,$F6,$F4,$6F,$DB,$B7,$E9,$FA
	.dc.b	$B7,$C9,$7F,$A7,$9B,$19,$E2,$C4,$E7,$8A,$4F,$1E,$8B,$97,$FA,$07
	.dc.b	$0F,$C7,$CE,$4E,$DF,$25,$D6,$BB,$78,$F2,$7C,$5E,$45,$C3,$93,$E3
	.dc.b	$3C,$9E,$A5,$9F,$47,$67,$9C,$F4,$79,$56,$3D,$8E,$1F,$AD,$E7,$FE
	.dc.b	$25,$C9,$F6,$79,$FF,$C8,$0D,$F8,$5F,$F5,$FC,$3C,$7B,$18,$64,$F4
	.dc.b	$F8,$0B,$A7,$FA,$84,$FF,$39,$93,$AF,$FD,$92,$D9,$FC,$43,$F8,$FC
	.dc.b	$3F,$DB,$7D,$9A,$AD,$B2,$BD,$3C,$F6,$FB,$6B,$D0,$EB,$AD,$AE,$AE
	.dc.b	$1E,$F2,$F1,$FF,$6D,$F5,$6F,$5B,$FC,$D3,$7E,$6A,$D1,$FE,$69,$3F
	.dc.b	$79,$55,$A5,$75,$F5,$4E,$E7,$D5,$E1,$B5,$FE,$73,$9D,$4E,$4E,$3E
	.dc.b	$37,$FB,$7E,$74,$D7,$FC,$FB,$3F,$C2,$F8,$36,$78,$3E,$33,$F8,$33
	.dc.b	$FA,$C9,$4F,$F9,$ED,$FD,$B3,$5F,$F6,$B3,$FE,$8B,$73,$C9,$CB,$6F
	.dc.b	$E2,$1C,$BF,$CD,$E7,$F1,$3E,$5E,$5F,$E3,$D7,$CF,$0F,$03,$F8,$78
	.dc.b	$EF,$E6,$ED,$F1,$B6,$35,$1E,$28,$C7,$C9,$C8,$E2,$7A,$1E,$58,$7F
	.dc.b	$89,$63,$F3,$7E,$F6,$DF,$F7,$87,$17,$F1,$77,$F8,$A2,$E4,$F1,$36
	.dc.b	$33,$CD,$89,$EB,$E4,$C9,$E3,$1E,$9F,$AB,$F8,$D1,$FE,$FD,$7E,$06
	.dc.b	$FF,$CD,$5B,$1F,$D0,$5D,$ED,$F4,$DC,$C3,$EA,$FF,$C8,$CF,$F2,$96
	.dc.b	$BD,$78,$7E,$47,$D5,$FF,$99,$63,$F9,$73,$FD,$21,$B9,$6B,$C9,$DB
	.dc.b	$C3,$CD,$93,$A3,$8F,$C5,$74,$E5,$91,$83,$AB,$E0,$F4,$E4,$60,$F0
	.dc.b	$F8,$1E,$4E,$39,$18,$3C,$BC,$1F,$7F,$EA,$9F,$57,$4F,$5C,$67,$C5
	.dc.b	$D7,$8B,$2F,$FA,$55,$FF,$85,$5F,$E9,$9E,$AB,$D9,$FC,$52,$F9,$6A
	.dc.b	$62,$33,$EC,$EB,$EA,$E5,$F2,$8E,$3D,$FC,$BB,$E4,$F6,$8F,$CB,$7D
	.dc.b	$5F,$F8,$46,$3F,$AE,$2E,$1C,$9A,$E2,$FC,$E7,$D5,$FF,$8B,$EB,$FB
	.dc.b	$13,$E9,$5D,$BF,$B3,$8C,$37,$CD,$FD,$27,$D5,$FF,$8C,$2B,$FB,$53
	.dc.b	$EF,$6E,$DA,$F7,$F7,$E3,$F0,$F1,$E5,$70,$8E,$7E,$7F,$4D,$E4,$F4
	.dc.b	$8F,$D9,$1F,$BB,$7C,$BD,$1F,$E0,$D8,$F8,$11,$83,$FC,$34,$E3,$4F
	.dc.b	$E1,$8F,$16,$FC,$27,$4F,$8D,$95,$FE,$03,$AF,$F3,$4B,$FE,$05,$FA
	.dc.b	$BF,$F1,$AB,$7D,$3F,$E0,$91,$FE,$34,$8D,$7A,$D8,$F2,$B3,$E5,$49
	.dc.b	$E5,$1F,$C2,$FE,$AF,$FC,$E1,$1F,$E1,$15,$F0,$B9,$35,$F0,$DB,$F4
	.dc.b	$B8,$71,$4B,$E1,$1F,$C5,$7E,$AF,$F5,$91,$F0,$43,$FE,$2F,$7F,$C3
	.dc.b	$CD,$74,$F0,$E1,$FE,$60,$BD,$DE,$9E,$14,$AE,$31,$FC,$8F,$E7,$7F
	.dc.b	$35,$1F,$E2,$95,$F0,$D9,$35,$E2,$59,$D7,$87,$C4,$C3,$E7,$A4,$7F
	.dc.b	$2F,$F9,$87,$C7,$53,$85,$77,$F5,$BC,$5C,$12,$30,$78,$DF,$E7,$1F
	.dc.b	$1F,$F9,$D8,$F8,$8D,$DF,$FA,$76,$5F,$4A,$7C,$5C,$3E,$DE,$E9,$1E
	.dc.b	$D8,$AF,$CD,$FA,$B7,$CD,$FF,$45,$6C,$7C,$3A,$F8,$DE,$07,$C7,$F2
	.dc.b	$64,$FC,$78,$AF,$C1,$FA,$BF,$98,$63,$3F,$2B,$D7,$BE,$4F,$CF,$8F
	.dc.b	$EA,$9F,$50,$F9,$CB,$EB,$CA,$FF,$2E,$4E,$5B,$1F,$D6,$FE,$AD,$F2
	.dc.b	$B4,$E1,$3F,$89,$A7,$AF,$DB,$D7,$93,$D2,$2B,$F8,$FF,$AA,$7D,$5F
	.dc.b	$F6,$76,$C7,$8A,$BC,$FE,$CF,$A3,$AA,$4F,$9A,$2B,$E1,$FA,$AF,$69
	.dc.b	$8E,$1F,$F0,$26,$68,$FE,$E9,$F5,$7F,$90,$AF,$F3,$D3,$5E,$47,$C2
	.dc.b	$F0,$7E,$3F,$85,$23,$07,$C6,$7F,$BF,$7D,$5F,$F4,$08,$FF,$42,$35
	.dc.b	$F8,$91,$F4,$3C,$DF,$A3,$E7,$C9,$F9,$F1,$5F,$D9,$71,$BE,$57,$E8
	.dc.b	$19,$D7,$FC,$3D,$F5,$7F,$DA,$1B,$E0,$FF,$83,$E0,$DF,$9D,$C3,$9F
	.dc.b	$90,$98,$FF,$A6,$6B,$6D,$8F,$A5,$F3,$92,$30,$6F,$D3,$ED,$F2,$12
	.dc.b	$30,$73,$7F,$39,$E4,$48,$C1,$DF,$E0,$3C,$DF,$57,$23,$07,$8D,$FD
	.dc.b	$8F,$C1,$48,$C1,$FE,$AF,$FA,$BF,$F0,$F3,$FC,$5D,$FF,$0E,$37,$D6
	.dc.b	$FF,$18,$46,$B8,$3F,$C8,$8E,$9F,$F2,$37,$D4,$32,$FF,$1B,$36,$FB
	.dc.b	$5F,$F8,$E7,$EA,$FF,$AB,$47,$F2,$57,$FA,$B3,$7D,$DD,$7B,$BB,$79
	.dc.b	$F9,$72,$7F,$D8,$D5,$C3,$2F,$F2,$B3,$DE,$3A,$7A,$3D,$CF,$06,$76
	.dc.b	$3F,$D9,$D9,$FF,$0F,$BF,$4D,$66,$7C,$98,$7F,$9D,$70,$34,$91,$FE
	.dc.b	$7B,$FA,$BF,$F4,$03,$7E,$3B,$FC,$F9,$18,$8F,$F7,$81,$FE,$FB,$70
	.dc.b	$E1,$72,$70,$24,$75,$7C,$5E,$64,$8F,$F4,$56,$08,$C4,$76,$1F,$57
	.dc.b	$F9,$FF,$99,$FF,$48,$36,$34,$BD,$57,$CA,$6B,$B9,$7E,$DF,$97,$FF
	.dc.b	$41,$60,$3D,$4E,$1B,$E1,$6C,$FF,$4F,$7C,$03,$E1,$7F,$DE,$BC,$7F
	.dc.b	$E8,$B8,$FE,$2B,$3F,$EF,$F7,$4D,$F3,$F6,$F7,$B4,$E9,$FE,$D8,$F2
	.dc.b	$6D,$F0,$7C,$D1,$9B,$F0,$4E,$67,$9D,$DA,$E9,$5E,$58,$C5,$65,$B7
	.dc.b	$E5,$3D,$11,$96,$99,$BF,$1B,$6B,$C2,$D3,$D9,$7E,$06,$DF,$6A,$F6
	.dc.b	$B6,$BC,$93,$DA,$72,$D8,$CC,$78,$0C,$95,$C9,$E2,$72,$D3,$31,$E4
	.dc.b	$BF,$64,$F1,$8F,$29,$DC,$AF,$73,$F3,$B7,$F5,$5B,$3F,$85,$7E,$CB
	.dc.b	$63,$51,$DE,$7F,$0F,$39,$23,$DE,$3F,$51,$3F,$0E,$38,$47,$94,$CB
	.dc.b	$B7,$BB,$78,$7F,$12,$E6,$7E,$84,$FE,$1A,$BF,$09,$5A,$8E,$87,$22
	.dc.b	$B9,$BF,$91,$7E,$26,$CF,$88,$E3,$7D,$9E,$51,$FE,$EB,$AE,$55,$C4
	.dc.b	$93,$E3,$7F,$5D,$3B,$AF,$81,$39,$3C,$6E,$1E,$B3,$C9,$57,$67,$85
	.dc.b	$79,$9D,$9E,$77,$15,$E7,$3D,$13,$C6,$76,$7E,$CB,$AF,$F5,$5F,$56
	.dc.b	$F6,$BF,$55,$18,$6E,$3F,$C0,$A3,$F0,$04,$7F,$82,$F6,$7E,$0C,$8E
	.dc.b	$09,$7F,$BE,$8F,$DC,$74,$BE,$03,$F7,$11,$FB,$9E,$2C,$E3,$70,$FE
	.dc.b	$7A,$DC,$3F,$92,$69,$FC,$7A,$7F,$8A,$D7,$E4,$CF,$CB,$7D,$5F,$F9
	.dc.b	$A2,$3E,$89,$FC,$45,$C5,$7F,$9E,$DF,$F3,$2C,$6A,$3F,$92,$46,$B6
	.dc.b	$4F,$1F,$7B,$F8,$84,$F0,$AF,$F4,$26,$9F,$CB,$AB,$94,$27,$57,$CA
	.dc.b	$3A,$FC,$E7,$23,$9E,$B6,$F8,$DF,$C5,$6C,$7F,$1F,$F7,$DF,$57,$C2
	.dc.b	$77,$F2,$9E,$43,$CB,$EE,$D8,$E6,$E1,$F2,$D3,$E6,$F1,$9F,$FA,$77
	.dc.b	$E2,$BD,$4D,$EA,$FE,$5D,$F5,$1F,$E6,$3C,$3D,$B5,$AD,$FD,$96,$AD
	.dc.b	$3F,$EA,$76,$C6,$BE,$53,$E1,$DF,$F4,$39,$CB,$75,$EE,$78,$3D,$EB
	.dc.b	$1F,$32,$7F,$CD,$2D,$8F,$E8,$5F,$56,$F3,$78,$4E,$FE,$47,$D2,$7A
	.dc.b	$BD,$DB,$1C,$DC,$3D,$7D,$EB,$AF,$A5,$F8,$27,$F4,$FF,$A7,$7C,$1E
	.dc.b	$2B,$C9,$C5,$F1,$8F,$8F,$EE,$D8,$F4,$38,$7F,$0B,$A3,$EA,$F1,$97
	.dc.b	$C7,$E3,$F9,$E7,$D0,$F7,$2D,$FC,$17,$0F,$E3,$BE,$AF,$E4,$D7,$F7
	.dc.b	$BE,$25,$7F,$65,$FA,$A7,$D5,$D3,$1E,$F5,$C7,$F2,$F2,$8C,$C6,$3D
	.dc.b	$EB,$8F,$FB,$86,$BF,$CD,$33,$93,$F3,$FC,$2E,$3F,$05,$F8,$7E,$4A
	.dc.b	$57,$6F,$76,$E3,$CD,$F3,$1F,$37,$96,$CF,$93,$C9,$F5,$1E,$BE,$8B
	.dc.b	$1E,$87,$0F,$8B,$E2,$2E,$1E,$7F,$5B,$E1,$1F,$0F,$D6,$B5,$F0,$1C
	.dc.b	$3F,$0B,$D1,$B5,$EF,$7E,$51,$F8,$1F,$82,$B3,$F2,$3C,$BF,$9C,$7C
	.dc.b	$FA,$B1,$F3,$1C,$3F,$1B,$D3,$B0,$F0,$FF,$53,$F5,$7F,$C3,$EB,$FC
	.dc.b	$4C,$D9,$FF,$1B,$3F,$A8,$6F,$C8,$7F,$88,$63,$10,$6F,$2B,$B3,$5E
	.dc.b	$76,$FD,$C9,$8F,$F1,$6B,$7B,$98,$C7,$C9,$3E,$28,$EC,$77,$F9,$2B
	.dc.b	$1C,$5C,$3F,$82,$2C,$F2,$3F,$C9,$DF,$57,$F9,$48,$FE,$4A,$D8,$D0
	.dc.b	$7F,$97,$C6,$BB,$FF,$C7,$DC,$1F,$DC,$49,$CF,$8F,$63,$FC,$85,$F5
	.dc.b	$6F,$81,$A6,$3E,$06,$E3,$FC,$D2,$7E,$66,$66,$9F,$28,$BE,$8F,$43
	.dc.b	$AF,$C7,$DA,$FE,$7B,$F4,$8F,$8E,$FF,$29,$36,$33,$F3,$58,$9C,$F5
	.dc.b	$C9,$E3,$D7,$73,$F8,$3F,$56,$F1,$3F,$CB,$8D,$AF,$E9,$6B,$C4,$CC
	.dc.b	$EB,$DC,$95,$AF,$7A,$E1,$D0,$E6,$7F,$99,$E3,$E3,$1D,$1C,$7B,$52
	.dc.b	$3F,$AC,$9F,$E0,$32,$BF,$E6,$FF,$AB,$7D,$0D,$31,$EC,$DA,$FE,$C3
	.dc.b	$F5,$6F,$AA,$FF,$3B,$36,$33,$A6,$27,$3E,$3C,$9E,$3A,$DD,$3F,$0B
	.dc.b	$F5,$7F,$ED,$39,$FE,$DA,$D8,$CF,$83,$FE,$83,$8C,$79,$1F,$D5,$17
	.dc.b	$C0,$E4,$D7,$05,$C3,$FB,$BF,$BD,$FC,$F1,$BE,$DF,$FA,$36,$3D,$F1
	.dc.b	$83,$7E,$DE,$4D,$7B,$76,$3F,$C0,$3F,$56,$F9,$0F,$F4,$B3,$63,$3C
	.dc.b	$38,$9C,$F1,$49,$E3,$D1,$78,$7F,$4B,$87,$E0,$67,$27,$6F,$90,$EB
	.dc.b	$5D,$BC,$59,$3E,$2F,$1A,$E1,$C9,$F1,$9E,$4F,$52,$CF,$A3,$B3,$CE
	.dc.b	$7A,$3C,$8B,$1E,$C7,$0F,$F8,$7B,$CD,$76,$63,$1F,$DD,$F4,$3E,$0F
	.dc.b	$FD,$6C,$DB,$FF,$12,$AF,$1F,$33,$A6,$4F,$5F,$80,$B7,$FC,$18,$FF
	.dc.b	$76,$C7,$F8,$B9,$A6,$7B,$BF,$D8,$8E,$30,$FF,$6F,$C6,$0C,$F7,$5A
	.dc.b	$E4,$E1,$F6,$74,$5B,$F0,$6D,$D0,$EB,$B5,$AF,$13,$87,$C9,$2F,$1F
	.dc.b	$FD,$DD,$F5,$6F,$5B,$FD,$C5,$3F,$E5,$93,$D8,$6C,$FF,$80,$07,$FB
	.dc.b	$D6,$6B,$4A,$EB,$EA,$BF,$CD,$5F,$57,$7C,$03,$49,$5E,$01,$3F,$E7
	.dc.b	$FD,$F9,$7C,$4F,$F5,$39,$D4,$E5,$B1,$97,$D3,$0F,$9C,$FF,$7E,$57
	.dc.b	$F9,$0A,$3F,$01,$3F,$E6,$26,$BF,$CF,$9E,$0E,$79,$38,$EA,$89,$CD
	.dc.b	$FC,$B5,$FC,$77,$99,$A7,$57,$D3,$0F,$07,$36,$67,$DB,$E7,$5A,$36
	.dc.b	$70,$E7,$E4,$A5,$79,$E2,$BF,$D5,$31,$FF,$0E,$3F,$9F,$F7,$F3,$8F
	.dc.b	$F9,$83,$89,$CF,$E1,$5C,$D4,$C4,$5B,$FD,$47,$97,$E7,$C7,$F9,$5C
	.dc.b	$7F,$A2,$2F,$8B,$FE,$8A,$FA,$BF,$8E,$FF,$66,$7E,$07,$07,$C7,$84
	.dc.b	$61,$F9,$4F,$0C,$BF,$CE,$AE,$61,$F5,$7F,$E3,$F7,$FD,$2E,$D7,$B7
	.dc.b	$17,$CB,$FA,$BF,$F2,$FC,$7F,$A7,$1F,$3B,$FD,$E1,$B7,$FE,$82,$77
	.dc.b	$19,$CF,$97,$74,$8E,$51,$D7,$E2,$BB,$72,$C9,$ED,$15,$ED,$EA,$78
	.dc.b	$CC,$63,$DF,$C0,$F3,$71,$CA,$F3,$47,$A7,$7F,$C2,$7D,$53,$EA,$E9
	.dc.b	$9A,$FF,$5B,$C3,$C4,$DF,$89,$27,$31,$F8,$DF,$59,$EC,$5F,$A4,$95
	.dc.b	$FE,$C0,$9C,$B7,$DF,$E1,$F5,$72,$30,$6F,$B7,$A7,$74,$8C,$1F,$EC
	.dc.b	$AF,$AB,$71,$FE,$75,$70,$D3,$11,$FA,$1F,$AB,$79,$97,$79,$87,$FB
	.dc.b	$52,$72,$FF,$B5,$BE,$AF,$FC,$50,$DE,$0F,$D5,$47,$AA,$3F,$C6,$AE
	.dc.b	$9B,$E3,$6B,$C6,$91,$CE,$38,$F8,$BC,$B7,$93,$DE,$3F,$6C,$7E,$F5
	.dc.b	$FF,$20,$BE,$E7,$FD,$D2,$D7,$3D,$7C,$3E,$13,$8F,$C0,$C9,$DC,$7F
	.dc.b	$01,$FA,$BF,$F1,$54,$7F,$BC,$9F,$F1,$43,$7D,$1D,$7A,$9B,$F2,$78
	.dc.b	$79,$52,$B8,$47,$F0,$BF,$AB,$FF,$32,$C7,$FB,$F1,$7C,$2E,$4D,$7B
	.dc.b	$DB,$F4,$38,$70,$CB,$E1,$1F,$C5,$7E,$AF,$F4,$FE,$27,$EE,$5F,$ED
	.dc.b	$33,$FC,$87,$C6,$FF,$75,$C6,$23,$FC,$CA,$F0,$7F,$93,$2F,$6F,$A7
	.dc.b	$85,$27,$95,$F5,$9F,$E4,$9F,$38,$F1,$2F,$C5,$4B,$D9,$FC,$C6,$74
	.dc.b	$E0,$28,$E7,$E2,$C9,$D4,$E5,$B1,$FC,$CF,$E5,$DF,$1D,$4E,$13,$FE
	.dc.b	$80,$A7,$A9,$E6,$E1,$93,$D2,$2B,$FE,$86,$F8,$FF,$CC,$B7,$C4,$CF
	.dc.b	$0B,$7F,$E4,$98,$B8,$FE,$0F,$81,$30,$F3,$C5,$7F,$D2,$3F,$54,$FA
	.dc.b	$BF,$E9,$4D,$8F,$6D,$7C,$AF,$0B,$E5,$F9,$52,$7E,$0C,$57,$D9,$F4
	.dc.b	$1F,$40,$C6,$9E,$58,$F9,$D1,$DB,$C4,$91,$F2,$A3,$FA,$D7,$D4,$BE
	.dc.b	$AE,$FC,$3C,$AF,$A9,$39,$6C,$7F,$61,$FA,$77,$CA,$D3,$84,$FF,$AB
	.dc.b	$A9,$EB,$76,$F2,$A4,$F4,$8A,$FF,$AD,$3E,$AD,$F4,$7F,$DB,$5B,$1E
	.dc.b	$2A,$F3,$FB,$1E,$8E,$C9,$3E,$68,$AF,$87,$EA,$FD,$A6,$38,$7E,$24
	.dc.b	$33,$4F,$F7,$BF,$AB,$FC,$64,$FF,$B1,$9B,$5C,$2B,$73,$F1,$3C,$3F
	.dc.b	$93,$E2,$48,$C1,$F2,$1F,$F0,$37,$D5,$FF,$43,$8F,$CF,$35,$F8,$D1
	.dc.b	$F4,$BD,$0F,$A7,$DD,$27,$E8,$C5,$7F,$DB,$3C,$6F,$94,$FA,$46,$73
	.dc.b	$F6,$FD,$5F,$F6,$D6,$F8,$5F,$E1,$78,$37,$E7,$70,$E7,$E3,$E6,$3F
	.dc.b	$B3,$AD,$B6,$3E,$A7,$CE,$48,$C1,$BF,$4F,$B7,$C7,$C8,$C1,$CD,$FC
	.dc.b	$FF,$91,$23,$07,$7F,$7F,$CD,$F5,$72,$30,$78,$DF,$F1,$86,$7F,$DA
	.dc.b	$63,$F8,$43,$5F,$F7,$EE,$69,$FF,$1B,$7D,$5F,$F8,$A2,$7F,$87,$3F
	.dc.b	$E2,$05,$B7,$E1,$23,$0D,$F5,$78,$FA,$DF,$E3,$FF,$A7,$FE,$13,$7F
	.dc.b	$C5,$DA,$F2,$E5,$FF,$13,$61,$C9,$FF,$14,$7D,$5F,$F5,$BC,$7F,$22
	.dc.b	$FF,$5F,$6F,$BB,$FE,$4B,$8D,$4F,$08,$F0,$79,$53,$0F,$E4,$D5,$C6
	.dc.b	$BA,$F0,$71,$F4,$65,$6A,$2B,$E5,$3F,$CC,$19,$FF,$4E,$70,$1C,$65
	.dc.b	$E6,$AF,$26,$32,$BC,$2F,$9F,$CB,$FF,$31,$7D,$57,$FA,$3E,$7F,$9D
	.dc.b	$35,$EA,$DB,$F8,$35,$C3,$9A,$57,$07,$9F,$27,$C3,$F2,$E4,$FC,$07
	.dc.b	$6F,$4E,$3F,$F4,$BE,$3F,$7E,$3F,$D3,$6D,$7F,$D2,$4D,$7F,$D4,$7F
	.dc.b	$57,$FA,$9F,$98,$FF,$50,$B6,$3F,$D3,$0F,$F5,$2F,$94,$CF,$6A,$BF
	.dc.b	$D4,$FC,$FF,$C5,$B8,$FD,$FF,$4F,$FC,$BF,$83,$F8,$6C,$E7,$2F,$83
	.dc.b	$39,$AF,$F3,$54,$E6,$BF,$89,$E6,$AD,$5F,$FA,$56,$B5,$7F,$E7,$1A
	.dc.b	$D3,$FD,$9E,$33,$3F,$E9,$B9,$CB,$F7,$23,$2A,$33,$7E,$6F,$2F,$BC
	.dc.b	$E7,$0B,$7C,$2F,$78,$7A,$2F,$71,$8B,$F3,$46,$2F,$8B,$6B,$CC,$D9
	.dc.b	$F2,$37,$0F,$14,$FC,$A3,$65,$9D,$B6,$37,$1F,$32,$36,$D3,$65,$8D
	.dc.b	$9F,$3A,$BC,$B3,$CC,$F6,$16,$7D,$6D,$7F,$D5,$F5,$E7,$8D,$C7,$FB
	.dc.b	$77,$0F,$BC,$FF,$BB,$9A,$E8,$C4,$FF,$50,$70,$43,$B9,$CD,$F2,$BF
	.dc.b	$67,$D6,$B8,$7A,$AF,$C2,$FE,$3A,$33,$18,$24,$FE,$26,$32,$F6,$9E
	.dc.b	$4E,$67,$E4,$C6,$23,$DF,$1F,$8D,$9E,$31,$83,$E6,$CF,$86,$78,$9E
	.dc.b	$08,$D4,$65,$DC,$66,$FF,$26,$7F,$A2,$1E,$31,$E5,$75,$5E,$09,$DD
	.dc.b	$FE,$52,$30,$E5,$B3,$CA,$BF,$DC,$B1,$E8,$FA,$BF,$F1,$A4,$FF,$B6
	.dc.b	$67,$FD,$B5,$18,$70,$1C,$3B,$6B,$FB,$8F,$85,$FD,$3C,$71,$C7,$F7
	.dc.b	$F1,$8F,$5F,$F5,$3D,$BF,$AA,$CE,$E3,$FB,$76,$E3,$FB,$59,$FD,$8F
	.dc.b	$4B,$DE,$CE,$5F,$E1,$68,$ED,$87,$FB,$C5,$B7,$FC,$17,$8B,$8E,$5F
	.dc.b	$B8,$6D,$7F,$05,$6C,$FF,$3E,$FA,$BF,$F3,$64,$FF,$9B,$DB,$19,$F7
	.dc.b	$3F,$81,$C6,$31,$FF,$13,$A8,$C9,$E8,$E3,$B7,$CF,$E7,$38,$9C,$B4
	.dc.b	$B1,$DD,$C3,$FE,$7F,$F7,$5F,$57,$8C,$FC,$DF,$25,$E2,$3C,$7E,$DD
	.dc.b	$BF,$43,$87,$97,$CD,$5A,$FE,$B7,$F1,$1F,$CD,$E7,$FD,$18,$D8,$FE
	.dc.b	$BF,$F5,$3F,$E5,$F9,$CE,$BF,$C8,$A3,$13,$AF,$7B,$F9,$2F,$C9,$FC
	.dc.b	$1B,$FF,$4B,$56,$B8,$BF,$C9,$6F,$AA,$33,$AF,$3F,$77,$A3,$63,$8C
	.dc.b	$7F,$30,$FA,$B7,$89,$9C,$FC,$BF,$1F,$E3,$3C,$9E,$E5,$C3,$C2,$E1
	.dc.b	$F1,$FB,$B6,$BD,$AF,$AC,$FE,$7B,$F5,$0F,$82,$C5,$7A,$B8,$FE,$01
	.dc.b	$F0,$7D,$CB,$1E,$E7,$0F,$FA,$D7,$4F,$E8,$18,$B7,$FF,$62,$4F,$C4
	.dc.b	$E4,$F9,$47,$CB,$F6,$AC,$7C,$87,$0F,$FA,$FF,$EA,$FE,$52,$B9,$6A
	.dc.b	$FE,$AB,$F5,$4F,$AB,$A6,$3D,$CB,$87,$FB,$26,$51,$98,$C7,$B3,$71
	.dc.b	$FF,$0F,$D7,$F9,$8A,$72,$7F,$B4,$78,$1C,$5E,$0B,$F0,$FC,$7C,$AE
	.dc.b	$DE,$E5,$C7,$9B,$E6,$3E,$7F,$25,$9F,$27,$93,$EA,$3D,$7C,$F6,$3D
	.dc.b	$0E,$1F,$17,$C3,$5C,$3C,$FF,$39,$F0,$8F,$87,$EA,$DA,$F8,$0E,$1F
	.dc.b	$85,$F4,$36,$BD,$EF,$CA,$3F,$07,$F0,$56,$7E,$47,$95,$F3,$8F,$9F
	.dc.b	$56,$3E,$63,$87,$E3,$7A,$96,$3F,$C9,$3F,$57,$FC,$1E,$7E,$D4,$23
	.dc.b	$0F,$DD,$7F,$54,$F0,$F8,$EF,$F0,$8C,$63,$C3,$FF,$5A,$B6,$B9,$F2
	.dc.b	$BB,$37,$E6,$70,$F7,$65,$7E,$27,$0E,$3E,$EE,$33,$F2,$BD,$8E,$FF
	.dc.b	$23,$6B,$81,$E8,$CD,$1A,$8F,$C2,$16,$BF,$CC,$FF,$57,$F9,$58,$E3
	.dc.b	$1F,$99,$8C,$3A,$8F,$F1,$6C,$63,$0F,$F6,$8E,$FF,$D8,$6B,$FC,$E2
	.dc.b	$D8,$CF,$37,$F8,$D2,$31,$7F,$33,$EA,$9F,$57,$FE,$38,$6C,$78,$B0
	.dc.b	$43,$E0,$F5,$5B,$E2,$97,$BA,$F9,$D7,$FB,$17,$EB,$BC,$BC,$FE,$C7
	.dc.b	$7F,$82,$B5,$DF,$E9,$1F,$1D,$FE,$4A,$6C,$67,$E6,$B1,$39,$E9,$93
	.dc.b	$C7,$A6,$E5,$FE,$92,$FA,$B7,$85,$FE,$54,$6D,$7B,$57,$85,$99,$D7
	.dc.b	$BB,$2B,$5E,$ED,$C3,$A1,$CC,$FF,$2F,$C7,$C8,$72,$FC,$93,$FB,$91
	.dc.b	$7D,$9D,$FB,$9C,$1F,$F3,$3F,$D5,$BE,$86,$98,$F6,$2D,$FD,$3F,$AB
	.dc.b	$7D,$57,$F9,$B9,$B1,$9D,$71,$39,$F1,$A4,$F1,$DA,$EB,$FE,$B6,$FA
	.dc.b	$B7,$89,$FE,$78,$6D,$7E,$05,$78,$99,$9D,$7C,$04,$AD,$70,$5C,$3F
	.dc.b	$D8,$3E,$F7,$F4,$59,$FC,$53,$63,$3C,$1F,$E8,$68,$C7,$86,$AF,$DF
	.dc.b	$C3,$D9,$D1,$BF,$66,$CF,$E5,$7E,$AD,$F2,$3F,$E9,$06,$C6,$78,$B1
	.dc.b	$39,$E3,$93,$C7,$A6,$EB,$FE,$D8,$C3,$F0,$73,$93,$B7,$C8,$F5,$AE
	.dc.b	$DF,$0F,$27,$C5,$E4,$DC,$39,$3E,$33,$C9,$E9,$D9,$F4,$76,$F9,$CF
	.dc.b	$47,$95,$63,$D8,$E1,$FD,$97,$33,$B3,$18,$FF,$BB,$B9,$1F,$07,$FE
	.dc.b	$B0,$6D,$FE,$E5,$78,$F9,$9D,$32,$7A,$FC,$05,$BF,$F7,$C8,$F4,$FF
	.dc.b	$D7,$6D,$9F,$E0,$8B,$BF,$33,$A6,$4E,$BB,$ED,$74,$79,$9F,$EC,$6D
	.dc.b	$3E,$F4,$EC,$EB,$B7,$80,$F0,$ED,$63,$C2,$E1,$F1,$DB,$C8,$FF,$00
	.dc.b	$7D,$5B,$D5,$FF,$3B,$5F,$AA,$B4,$7F,$9D,$8F,$F0,$35,$56,$95,$D7
	.dc.b	$D5,$7F,$9A,$FE,$AE,$F8,$26,$92,$BF,$D0,$32,$F1,$DC,$9C,$9B,$3A
	.dc.b	$6B,$FE,$8B,$9E,$13,$E0,$F9,$BF,$EC,$2F,$E9,$67,$FC,$39,$3F,$80
	.dc.b	$7F,$9B,$35,$EF,$AF,$F6,$88,$57,$28,$C7,$2D,$AF,$F1,$37,$07,$F3
	.dc.b	$07,$F1,$FE,$66,$56,$1D,$71,$F1,$46,$39,$B5,$2F,$3A,$98,$8B,$3E
	.dc.b	$6F,$92,$D9,$87,$CE,$7B,$1F,$06,$3F,$E2,$88,$FD,$0F,$BB,$C1,$FE
	.dc.b	$A9,$C6,$E6,$F3,$2E,$6A,$62,$2C,$FE,$AB,$C7,$EA,$61,$F0,$8F,$AB
	.dc.b	$F8,$EF,$F6,$17,$E0,$70,$FF,$84,$9B,$1F,$E4,$D7,$6F,$27,$2B,$EA
	.dc.b	$FF,$C9,$2F,$F6,$B6,$BB,$E2,$F9,$DF,$57,$FE,$6B,$8F,$EE,$0F,$F4
	.dc.b	$36,$E1,$AF,$88,$DB,$C3,$BE,$4E,$8E,$3F,$13,$D3,$9A,$46,$0E,$AF
	.dc.b	$83,$D5,$91,$83,$C3,$E0,$F9,$38,$A4,$60,$F2,$EF,$F8,$5F,$AA,$7D
	.dc.b	$5D,$3D,$71,$9F,$17,$5E,$2C,$BF,$F0,$65,$A7,$F1,$DE,$BF,$F8,$36
	.dc.b	$7F,$C2,$0B,$E9,$32,$6B,$DC,$DF,$C0,$70,$FA,$B9,$7C,$23,$9F,$7F
	.dc.b	$4E,$E9,$3D,$23,$F3,$9F,$56,$E2,$BC,$52,$BF,$C4,$13,$97,$FC,$45
	.dc.b	$F5,$6F,$37,$F4,$8F,$A6,$33,$BE,$27,$F5,$1F,$57,$FE,$2C,$9F,$F1
	.dc.b	$53,$ED,$7F,$1E,$DF,$17,$F5,$6E,$E3,$3E,$3F,$2F,$1E,$47,$28,$EB
	.dc.b	$E3,$76,$DA,$4F,$68,$FD,$C1,$FB,$B7,$F3,$AF,$AD,$BB,$6B,$AB,$7F
	.dc.b	$0B,$C3,$E3,$65,$7F,$90,$6F,$FC,$CC,$FF,$90,$FE,$AF,$FC,$56,$DF
	.dc.b	$4B,$F8,$34,$7F,$8A,$A3,$5E,$AE,$3C,$8C,$F9,$32,$79,$47,$F0,$DF
	.dc.b	$AB,$FF,$06,$C7,$F9,$3D,$FF,$33,$B7,$E1,$F5,$EF,$6F,$D1,$E1,$C5
	.dc.b	$2B,$84,$7F,$17,$FA,$BF,$D4,$47,$EF,$1F,$ED,$0F,$F9,$6D,$AE,$7C
	.dc.b	$9E,$1E,$E7,$1F,$0E,$4E,$E3,$F9,$17,$D5,$FE,$4E,$3F,$CC,$8B,$E1
	.dc.b	$F2,$6B,$C5,$B3,$AF,$0F,$89,$87,$F8,$7D,$23,$F9,$77,$CB,$BE,$3A
	.dc.b	$9C,$2B,$BF,$AB,$E2,$E0,$91,$83,$C6,$FF,$37,$F8,$FF,$CB,$47,$F9
	.dc.b	$E5,$AF,$82,$3D,$7C,$3E,$CE,$F9,$3E,$C8,$AF,$F2,$4F,$AA,$7D,$5F
	.dc.b	$F4,$36,$C7,$C2,$AF,$8B,$E0,$FC,$6F,$26,$4F,$C6,$8A,$FC,$0F,$A2
	.dc.b	$F9,$66,$3C,$5F,$E9,$7C,$28,$ED,$E0,$4A,$F9,$F1,$FD,$4F,$EA,$1F
	.dc.b	$37,$7D,$69,$5F,$E9,$79,$CB,$63,$FA,$D7,$D3,$BE,$56,$9C,$27,$FA
	.dc.b	$0D,$3D,$7E,$DE,$4C,$9E,$91,$5F,$E8,$FF,$56,$FA,$1F,$EC,$CD,$8F
	.dc.b	$15,$79,$FD,$9F,$47,$54,$9F,$34,$57,$C3,$F5,$7E,$D3,$1C,$5F,$DB
	.dc.b	$B3,$47,$F7,$3F,$AB,$FC,$85,$7F,$AD,$5A,$F2,$3E,$17,$83,$F1,$FC
	.dc.b	$29,$18,$3E,$33,$FD,$F7,$EA,$FF,$9C,$C7,$FA,$FD,$AF,$C4,$8F,$A1
	.dc.b	$E7,$7D,$1F,$6E,$4F,$CF,$8A,$FF,$6A,$E3,$7C,$AF,$D0,$33,$AF,$F8
	.dc.b	$23,$EA,$FF,$B0,$B7,$C2,$FF,$07,$41,$BF,$3B,$87,$3F,$1F,$31,$FF
	.dc.b	$6A,$56,$DB,$1F,$4B,$E3,$64,$60,$DF,$A9,$DB,$E3,$A4,$60,$E6,$FE
	.dc.b	$77,$C8,$91,$83,$BF,$C0,$F9,$BE,$A6,$46,$0F,$1B,$FB,$1F,$72,$46
	.dc.b	$0F,$F7,$57,$D5,$FF,$85,$DB,$EB,$FF,$8B,$63,$FC,$31,$1A,$F5,$F1
	.dc.b	$3F,$E3,$1F,$A6,$FF,$5A,$CF,$FB,$CD,$AF,$F8,$62,$B8,$BF,$E1,$AF
	.dc.b	$AB,$FE,$95,$C7,$FE,$36,$7F,$A7,$73,$A6,$46,$1B,$F3,$BF,$E3,$F8
	.dc.b	$C4,$30,$79,$9E,$33,$5F,$0B,$BE,$0E,$1E,$94,$BE,$D9,$F8,$67,$11
	.dc.b	$FE,$2E,$C3,$F9,$FB,$6B,$35,$E7,$F8,$AF,$17,$9B,$2B,$9E,$BE,$95
	.dc.b	$A7,$F8,$B8,$FF,$4E,$BA,$3D,$DE,$B7,$F5,$87,$FC,$7A,$DC,$3E,$25
	.dc.b	$FB,$FD,$0F,$81,$B4,$BE,$47,$68,$C3,$FE,$46,$FA,$AF,$F4,$63,$A6
	.dc.b	$BF,$46,$7F,$CD,$1F,$57,$FD,$4B,$0F,$F3,$3B,$FD,$41,$BF,$17,$AF
	.dc.b	$6F,$6F,$F9,$BF,$37,$87,$4C,$E3,$E4,$C3,$F5,$BF,$33,$FE,$99,$6C
	.dc.b	$7C,$53,$EF,$2F,$55,$F2,$9B,$F1,$57,$C7,$0F,$F4,$67,$17,$1C,$3C
	.dc.b	$2F,$E1,$CC,$46,$B1,$FA,$53,$A7,$FD,$06,$7F,$8D,$E7,$4F,$D2,$6B
	.dc.b	$97,$A3,$F6,$F2,$FF,$80,$11,$88,$FC,$B7,$5B,$E6,$3A,$C6,$1A,$FF
	.dc.b	$A0,$5B,$1B,$9F,$F5,$E7,$D5,$7F,$58,$6F,$B7,$C2,$31,$7F,$1C,$38
	.dc.b	$7E,$57,$1F,$F9,$72,$BF,$D8,$CD,$8E,$0D,$77,$F1,$FE,$48,$7D,$2D
	.dc.b	$25,$72,$75,$E5,$48,$EA,$DB,$EB,$E5,$73,$24,$F8,$9E,$DE,$4C,$8F
	.dc.b	$2B,$67,$D9,$A3,$C8,$F8,$A4,$AF,$F7,$04,$65,$B1,$FE,$CD,$E8,$FF
	.dc.b	$5E,$CE,$6B,$FD,$4A,$BC,$8D,$5F,$EC,$E3,$55,$C2,$7E,$99,$27,$BB
	.dc.b	$6B,$CA,$D7,$F3,$15,$D6,$7E,$7F,$07,$69,$FC,$1D,$55,$AF,$DF,$7C
	.dc.b	$B1,$98,$C4,$7D,$A9,$FA,$84,$AF,$B7,$F0,$DB,$38,$65,$63,$F7,$8F
	.dc.b	$DB,$23,$F8,$27,$EF,$61,$C2,$3F,$81,$36,$FF,$83,$7A,$FF,$83,$24
	.dc.b	$FE,$29,$B1,$A8,$FE,$15,$F1,$7E,$99,$F5,$3F,$8B,$24,$7E,$49,$FC
	.dc.b	$77,$93,$A8,$FE,$24,$DA,$FF,$6D,$F1,$BE,$73,$E4,$36,$BF,$8E,$AF
	.dc.b	$86,$CC,$C9,$CB,$C2,$31,$5D,$1F,$D5,$12,$7F,$5E,$FE,$CB,$1F,$AD
	.dc.b	$BF,$F8,$96,$7F,$95,$BF,$EA,$96,$C6,$B2,$91,$FC,$AE,$6D,$FE,$43
	.dc.b	$CE,$E6,$74,$F9,$AB,$5D,$1C,$3F,$D9,$7C,$AB,$5F,$98,$BF,$19,$E4
	.dc.b	$F3,$6D,$79,$1C,$3F,$DB,$73,$B4,$E4,$4F,$F5,$1F,$AA,$FF,$40,$3F
	.dc.b	$D1,$9B,$3B,$8D,$7C,$37,$E5,$E7,$35,$CA,$BF,$B9,$47,$FA,$06,$B9
	.dc.b	$57,$E7,$7C,$CF,$E6,$B1,$83,$B7,$07,$33,$A6,$36,$3C,$4E,$1F,$EB
	.dc.b	$3F,$0C,$FA,$AC,$65,$E5,$3C,$3E,$6F,$94,$F3,$79,$96,$FD,$8E,$1E
	.dc.b	$7F,$3D,$6B,$90,$F8,$EF,$3B,$86,$EF,$EF,$87,$D4,$7F,$9E,$2F,$FB
	.dc.b	$6D,$69,$FE,$F9,$19,$9E,$AD,$AE,$5E,$67,$C7,$9C,$FC,$37,$F0,$3A
	.dc.b	$F7,$5F,$F7,$86,$CF,$2F,$84,$D6,$BB,$65,$7F,$7E,$FA,$A7,$B3,$8B
	.dc.b	$F0,$FC,$AF,$53,$B7,$C1,$DB,$EC,$E1,$F0,$7C,$25,$AF,$4B,$E5,$3F
	.dc.b	$8C,$7D,$43,$E2,$71,$5E,$7E,$2F,$51,$EB,$F8,$3B,$1E,$B7,$0F,$9F
	.dc.b	$EA,$EE,$3F,$1B,$93,$E1,$1F,$0F,$D3,$B5,$F0,$DC,$3F,$E1,$CF,$AB
	.dc.b	$F9,$8F,$D0,$8F,$E5,$9F,$57,$FD,$62,$3F,$C4,$2D,$BF,$BE,$BD,$0D
	.dc.b	$FA,$3C,$19,$5C,$3D,$1B,$7C,$1E,$A7,$C2,$E1,$B3,$DB,$CE,$F0,$9E
	.dc.b	$2F,$3A,$C7,$89,$C3,$FE,$34,$E2,$FF,$73,$39,$3C,$FE,$6F,$A8,$F5
	.dc.b	$F3,$58,$F5,$B8,$79,$7C,$65,$CF,$DB,$F3,$FF,$08,$F8,$7E,$65,$AF
	.dc.b	$86,$E1,$F8,$BE,$65,$AF,$8A,$FC,$A3,$EF,$FC,$3D,$9F,$97,$ED,$FC
	.dc.b	$E3,$E7,$D5,$8F,$9E,$E1,$F7,$7B,$76,$3F,$5F,$F5,$7F,$C2,$A7,$FC
	.dc.b	$C4,$D8,$FF,$3E,$3F,$A5,$F1,$7F,$60,$E6,$0E,$67,$8F,$99,$F9,$4E
	.dc.b	$AF,$ED,$93,$9A,$E7,$F0,$78,$33,$F0,$77,$0F,$F3,$2D,$23,$31,$8F
	.dc.b	$95,$B7,$E8,$7A,$9F,$A2,$8A,$7F,$9B,$BE,$AF,$FC,$95,$9F,$F0,$76
	.dc.b	$C7,$F9,$FD,$FF,$22,$D7,$F1,$07,$2F,$FA,$6F,$5F,$E9,$8F,$F1,$46
	.dc.b	$D7,$F1,$15,$C9,$FE,$09,$73,$7F,$E8,$BF,$AB,$7C,$3F,$F8,$35,$B1
	.dc.b	$FE,$82,$5E,$B6,$67,$5F,$2D,$2B,$5C,$37,$3F,$E4,$B8,$3C,$FE,$87
	.dc.b	$5F,$87,$B7,$FE,$8E,$FA,$57,$C9,$7F,$87,$5B,$19,$F9,$BC,$4E,$7A
	.dc.b	$24,$F1,$E8,$B9,$FF,$34,$FA,$B7,$83,$FE,$27,$6D,$7F,$A7,$1F,$F1
	.dc.b	$BB,$FE,$AB,$73,$3C,$7D,$CE,$07,$1F,$72,$D7,$19,$E8,$7F,$8C,$DF
	.dc.b	$8E,$F8,$4C,$7B,$32,$3F,$D5,$63,$E6,$71,$3F,$D2,$BE,$AD,$F1,$9F
	.dc.b	$E3,$D6,$D7,$FA,$C9,$7B,$59,$9F,$F2,$17,$D5,$FF,$90,$7B,$3F,$A7
	.dc.b	$36,$FF,$D8,$CF,$B2,$33,$F0,$BC,$3C,$49,$5F,$EB,$EA,$6D,$8C,$FF
	.dc.b	$AE,$FD,$5B,$C4,$FF,$28,$36,$FF,$D8,$AF,$E8,$5B,$1A,$F8,$09,$1A
	.dc.b	$E0,$B7,$FD,$A3,$E3,$7F,$A1,$4F,$FB,$39,$B1,$9D,$BF,$CB,$D1,$8D
	.dc.b	$BF,$B5,$2F,$5F,$26,$BD,$7B,$87,$FB,$63,$EA,$DF,$15,$FE,$67,$6C
	.dc.b	$67,$8B,$13,$9E,$89,$3C,$7D,$1B,$AF,$F7,$EC,$3F,$0E,$83,$86,$FC
	.dc.b	$4F,$79,$F0,$7C,$3C,$BF,$27,$93,$6F,$A3,$E5,$3C,$DE,$95,$9F,$57
	.dc.b	$67,$A4,$F5,$79,$56,$3D,$CE,$1F,$F7,$BF,$37,$E1,$9C,$9F,$EF,$AE
	.dc.b	$37,$C2,$7F,$A2,$5B,$19,$F1,$B1,$39,$64,$74,$F7,$EE,$9F,$E2,$01
	.dc.b	$E9,$7F,$A4,$DB,$3F,$17,$0F,$F7,$1F,$76,$6B,$4C,$9D,$1F,$EA,$02
	.dc.b	$D7,$CD,$11,$D7,$B6,$4F,$BB,$A2,$D7,$8F,$6F,$11,$E3,$DA,$D7,$49
	.dc.b	$FE,$4C,$D5,$1E,$95,$E3,$FD,$DF,$AB,$7A,$9E,$39,$DD,$F1,$76,$7F
	.dc.b	$BA,$4F,$C2,$55,$69,$5D,$7D,$57,$FA,$3F,$EA,$EF,$80,$69,$2B,$FD
	.dc.b	$E9,$3F,$C1,$F8,$38,$F8,$DF,$E7,$39,$CB,$5F,$34,$ED,$E3,$F3,$FF
	.dc.b	$2E,$BF,$CA,$53,$F8,$29,$F8,$8D,$7F,$CB,$6A,$78,$65,$F4,$E3,$55
	.dc.b	$FE,$5B,$E6,$FE,$6F,$7F,$90,$F2,$F6,$FE,$96,$B7,$7C,$79,$B3,$29
	.dc.b	$79,$B8,$6C,$79,$27,$13,$9F,$94,$93,$E8,$8A,$FF,$9C,$B1,$7B,$7D
	.dc.b	$9B,$1E,$EA,$F0,$F0,$FF,$1E,$8F,$C2,$2E,$4C,$9A,$E6,$B3,$FA,$CF
	.dc.b	$1B,$A1,$87,$FD,$03,$F5,$7F,$1D,$FC,$73,$F0,$1F,$E0,$CD,$4E,$DE
	.dc.b	$3D,$9F,$13,$EA,$FF,$C9,$91,$F9,$76,$BA,$3F,$35,$F5,$7F,$E6,$96
	.dc.b	$DF,$CC,$8F,$E9,$31,$AF,$86,$C4,$E7,$CE,$93,$C0,$E7,$F1,$3C,$79
	.dc.b	$64,$60,$EE,$F5,$F5,$24,$60,$F1,$F8,$5E,$1E,$39,$18,$3C,$FC,$FF
	.dc.b	$80,$FA,$B7,$CA,$53,$D1,$19,$F1,$B5,$E3,$4B,$FD,$C5,$FF,$86,$1F
	.dc.b	$DD,$7A,$DF,$E1,$26,$FC,$B7,$E3,$20,$DF,$73,$1E,$D6,$7E,$AE,$5F
	.dc.b	$28,$E1,$DD,$C7,$B6,$4F,$68,$FC,$BF,$D5,$B8,$AF,$14,$AF,$E1,$93
	.dc.b	$97,$F8,$77,$D5,$BC,$EF,$CF,$BE,$98,$FF,$63,$B7,$C4,$FF,$62,$3B
	.dc.b	$8F,$E2,$DF,$57,$FE,$32,$6E,$DF,$A8,$8F,$81,$1F,$ED,$07,$4D,$F1
	.dc.b	$F5,$E3,$C8,$E7,$1C,$7C,$6E,$5B,$49,$EF,$1F,$B4,$3F,$76,$FF,$B7
	.dc.b	$1F,$63,$76,$D7,$56,$FE,$17,$87,$C6,$4A,$FE,$5D,$7F,$EE,$37,$F9
	.dc.b	$87,$D5,$FF,$8C,$9B,$E9,$FF,$02,$83,$7D,$6C,$79,$19,$F2,$65,$F2
	.dc.b	$8F,$E1,$3F,$57,$FE,$6F,$8F,$E7,$8B,$E1,$B2,$6B,$E1,$F7,$E8,$F0
	.dc.b	$E2,$97,$C2,$3F,$89,$FD,$5F,$EA,$23,$FD,$F6,$FF,$8B,$9F,$E9,$0D
	.dc.b	$73,$E4,$F0,$F7,$78,$F8,$52,$77,$1F,$C7,$BE,$7F,$F2,$71,$FD,$45
	.dc.b	$7C,$3E,$4D,$78,$96,$75,$E1,$F1,$30,$BB,$DE,$47,$F6,$5C,$15,$FC
	.dc.b	$B3,$E6,$1F,$29,$4E,$58,$78,$7D,$5F,$27,$14,$8C,$1E,$57,$F9,$AF
	.dc.b	$C7,$FE,$4E,$3C,$19,$7F,$73,$8C,$67,$E2,$9F,$F1,$16,$7F,$4D,$D9
	.dc.b	$8E,$3F,$85,$E0,$C8,$F8,$51,$5D,$8F,$AB,$7C,$FF,$F4,$66,$C7,$C9
	.dc.b	$AF,$99,$E1,$FC,$DF,$2A,$4F,$CD,$8A,$F4,$C5,$F5,$7F,$4C,$CE,$7E
	.dc.b	$99,$4E,$9F,$F0,$36,$3F,$E9,$76,$F8,$3C,$5F,$F0,$57,$D5,$3E,$73
	.dc.b	$FA,$DB,$FC,$CE,$3D,$6D,$C3,$B3,$B8,$E3,$EB,$7F,$62,$FA,$77,$CA
	.dc.b	$53,$9C,$74,$F5,$FA,$F9,$52,$FC,$31,$5F,$93,$F5,$6F,$A1,$FE,$DA
	.dc.b	$D8,$F3,$D7,$9B,$D9,$F3,$F5,$C9,$F5,$C5,$7D,$3F,$57,$EC,$31,$B7
	.dc.b	$A9,$FE,$F3,$F5,$7F,$9A,$9F,$F1,$63,$5E,$27,$C6,$F0,$BE,$2F,$87
	.dc.b	$23,$07,$CA,$7F,$C0,$7F,$57,$FD,$12,$3F,$C6,$CD,$7E,$5C,$7C,$EF
	.dc.b	$3F,$E7,$F7,$C9,$FA,$71,$5F,$C5,$F1,$BE,$57,$E7,$19,$D3,$F0,$1F
	.dc.b	$57,$FD,$B1,$BE,$0F,$F8,$52,$0D,$F9,$CC,$39,$F9,$09,$8F,$F9,$2E
	.dc.b	$B6,$D8,$FA,$3F,$1D,$23,$06,$FD,$3E,$DF,$1F,$23,$07,$37,$F4,$1E
	.dc.b	$44,$8C,$1D,$FE,$03,$CD,$F5,$72,$30,$78,$DF,$DC,$FC,$14,$8C,$1F
	.dc.b	$E6,$1F,$AB,$FF,$10,$3F,$B0,$7F,$C4,$0E,$7D,$7F,$F1,$AC,$6B,$83
	.dc.b	$F4,$CE,$A7,$F6,$FF,$4D,$FE,$C3,$7F,$CD,$8D,$7A,$57,$F8,$FB,$EA
	.dc.b	$FF,$AA,$DF,$EB,$DF,$EB,$2D,$F7,$F5,$F3,$BB,$7F,$C8,$B8,$3C,$DA
	.dc.b	$67,$1F,$F3,$D6,$1D,$F0,$E9,$E8,$F7,$3C,$1B,$DA,$F2,$C7,$43,$0F
	.dc.b	$F1,$0A,$FF,$1A,$1F,$BF,$71,$E5,$7A,$7E,$2F,$D4,$FF,$96,$3D,$5F
	.dc.b	$EB,$93,$FC,$61,$B8,$FB,$67,$E0,$7A,$1F,$07,$39,$7E,$88,$AF,$F2
	.dc.b	$0F,$AA,$FF,$46,$BF,$E9,$56,$BE,$59,$FF,$33,$FD,$5F,$F5,$5C,$3F
	.dc.b	$8C,$3F,$D4,$DB,$F1,$7A,$F6,$B6,$F9,$FC,$19,$87,$FA,$86,$F8,$78
	.dc.b	$5F,$EA,$D5,$E2,$FF,$AB,$E3,$FD,$3D,$18,$9F,$F6,$F4,$62,$75,$1C
	.dc.b	$5A,$FF,$B8,$E7,$F6,$1F,$33,$FE,$A3,$6C,$7F,$AA,$43,$FC,$C3,$E5
	.dc.b	$75,$E2,$2B,$87,$C2,$7F,$4D,$C7,$C6,$DC,$3D,$0D,$7C,$0D,$5F,$05
	.dc.b	$FB,$A9,$F4,$C7,$FB,$2E,$30,$FF,$41,$37,$E2,$F1,$8C,$CF,$FA,$FD
	.dc.b	$78,$9B,$AF,$F8,$01,$18,$9F,$F6,$03,$63,$DC,$D7,$97,$CC,$F1,$7C
	.dc.b	$AD,$83,$4D,$7D,$5F,$F6,$2A,$FF,$66,$36,$BE,$53,$5F,$F6,$A3,$5F
	.dc.b	$F7,$47,$3F,$DC,$9F,$F6,$A3,$63,$E9,$35,$FF,$6E,$35,$E3,$E5,$79
	.dc.b	$E0,$D2,$4F,$FB,$7F,$99,$E4,$FB,$5B,$19,$8C,$1E,$CE,$C9,$1E,$C6
	.dc.b	$E1,$F2,$3E,$17,$E2,$12,$7E,$6B,$F1,$FC,$89,$1F,$41,$B3,$F4,$1A
	.dc.b	$FD,$96,$BF,$81,$7F,$C3,$1D,$3F,$E0,$1D,$BF,$CA,$34,$AF,$F0,$9C
	.dc.b	$3C,$8D,$CF,$13,$55,$C6,$3E,$D1,$23,$F0,$6F,$C3,$C3,$8D,$F5,$D7
	.dc.b	$FA,$A5,$7E,$FE,$73,$3F,$E9,$8E,$4B,$AA,$79,$57,$CF,$6D,$7E,$AE
	.dc.b	$75,$3C,$CF,$E1,$59,$F2,$AE,$B8,$7F,$87,$A7,$F0,$7E,$D7,$DD,$23
	.dc.b	$F6,$FD,$4D,$C7,$89,$B1,$FC,$55,$FB,$C4,$78,$13,$1F,$E3,$91,$EA
	.dc.b	$F8,$AE,$53,$F2,$0F,$9D,$2E,$27,$CD,$FC,$27,$FC,$F5,$3B,$9F,$A9
	.dc.b	$1A,$8C,$47,$1F,$86,$E1,$5F,$A8,$24,$FE,$A1,$B3,$FE,$4E,$AE,$4D
	.dc.b	$F8,$6E,$35,$D3,$8B,$FC,$54,$DC,$BE,$AC,$7F,$AE,$E3,$06,$F2,$C9
	.dc.b	$F6,$22,$C7,$3F,$3F,$89,$CB,$E6,$AC,$77,$70,$FF,$6A,$F2,$AE,$5F
	.dc.b	$92,$BF,$09,$E2,$F3,$2D,$79,$DC,$3F,$DC,$B3,$B5,$FE,$6C,$FA,$BF
	.dc.b	$F4,$7C,$FC,$86,$C6,$7E,$1B,$F2,$31,$8E,$8F,$F3,$23,$FE,$8C,$AE
	.dc.b	$13,$F9,$6F,$37,$F9,$B4,$60,$E9,$C3,$C8,$E7,$8D,$8F,$03,$87,$EC
	.dc.b	$FC,$33,$EA,$B1,$9F,$9F,$CD,$F1,$9E,$4F,$36,$DF,$A5,$C3,$CB,$E7
	.dc.b	$2D,$7F,$A2,$BC,$5F,$2B,$7C,$6F,$D6,$FD,$5F,$F9,$DE,$FF,$13,$5A
	.dc.b	$E1,$FF,$41,$4E,$6F,$FD,$26,$D8,$FF,$43,$79,$9F,$0E,$73,$F0,$9D
	.dc.b	$35,$98,$FE,$53,$07,$C0,$E2,$F9,$BE,$23,$F7,$FF,$56,$F6,$71,$7C
	.dc.b	$7E,$4F,$CE,$7A,$3E,$02,$C7,$57,$0F,$7F,$81,$B8,$FA,$1F,$79,$FC
	.dc.b	$33,$E9,$DF,$0B,$8A,$F3,$71,$FC,$43,$E2,$FA,$96,$3D,$4E,$1F,$D7
	.dc.b	$E8,$FA,$BC,$65,$F1,$B9,$3E,$69,$F3,$BD,$6B,$7F,$0D,$C3,$FB,$8F
	.dc.b	$AB,$F9,$8F,$F0,$2E,$24,$FF,$22,$FA,$B7,$D5,$D3,$1E,$9D,$BF,$E0
	.dc.b	$33,$FD,$3B,$2F,$F6,$4B,$7A,$7F,$D7,$8F,$F7,$E6,$D7,$5F,$5E,$5F
	.dc.b	$F0,$AE,$27,$26,$31,$E7,$E7,$F8,$CF,$27,$A1,$6F,$C2,$E1,$E5,$CD
	.dc.b	$6B,$C7,$F0,$3E,$B3,$D9,$D3,$6B,$D2,$E1,$F1,$7C,$55,$AF,$5F,$CF
	.dc.b	$FC,$33,$E2,$79,$B6,$BE,$0B,$87,$E1,$F9,$B6,$BE,$1B,$F2,$CF,$C1
	.dc.b	$7B,$16,$7E,$4F,$B7,$F3,$CF,$A1,$56,$3E,$6B,$87,$E3,$7A,$F6,$3F
	.dc.b	$DE,$FF,$57,$FC,$1E,$7F,$9D,$B6,$3F,$B5,$3F,$A7,$E5,$FF,$77,$39
	.dc.b	$83,$99,$E3,$E6,$FE,$5F,$B3,$FB,$1C,$E6,$B9,$FC,$16,$0C,$FB,$77
	.dc.b	$0F,$E8,$74,$8C,$C6,$3D,$7B,$7E,$87,$A9,$FA,$58,$A7,$F4,$CF,$AB
	.dc.b	$FC,$9B,$7B,$BF,$BA,$C2,$2C,$7F,$78,$DB,$FA,$25,$B5,$D3,$FE,$0B
	.dc.b	$8C,$C6,$21,$0E,$5E,$5E,$B7,$F8,$0B,$EA,$DF,$0B,$4C,$7B,$17,$3F
	.dc.b	$EC,$33,$F2,$C9,$C5,$C7,$E5,$79,$7A,$32,$B7,$F0,$B7,$5F,$34,$7F
	.dc.b	$72,$FA,$57,$C7,$FF,$84,$9B,$1F,$DB,$97,$CD,$66,$75,$D5,$2B,$5D
	.dc.b	$57,$0E,$A7,$F8,$AF,$EA,$FF,$C2,$AE,$03,$8E,$4F,$88,$FF,$8C,$A3
	.dc.b	$4E,$6B,$FC,$0E,$FF,$9F,$1F,$F0,$0B,$63,$7E,$EF,$F8,$AA,$31,$5D
	.dc.b	$27,$07,$F8,$BA,$BE,$69,$C9,$C7,$B5,$23,$FC,$1A,$3E,$63,$13,$E9
	.dc.b	$FA,$B7,$C5,$7F,$8E,$1B,$5F,$E1,$45,$ED,$66,$7F,$C7,$BF,$57,$FE
	.dc.b	$7E,$AF,$F0,$CB,$6F,$E2,$3E,$F7,$FC,$A4,$E7,$2F,$C4,$2F,$0F,$99
	.dc.b	$D3,$E0,$AD,$FF,$89,$3E,$AD,$E2,$7F,$93,$1B,$19,$F0,$F1,$39,$F8
	.dc.b	$19,$3C,$78,$2E,$9F,$5F,$DF,$FE,$93,$3F,$E3,$06,$C6,$76,$FF,$2C
	.dc.b	$C6,$36,$FC,$F2,$F6,$B2,$6B,$DA,$B8,$7F,$8E,$BE,$AF,$FD,$51,$7F
	.dc.b	$A5,$6C,$7F,$91,$9F,$84,$DF,$8C,$FF,$33,$C7,$C7,$C8,$D3,$F1,$55
	.dc.b	$CF,$D2,$DF,$A5,$30,$F0,$7C,$5D,$C3,$F2,$98,$7E,$DA,$7C,$1E,$3F
	.dc.b	$6F,$23,$BB,$63,$D1,$1E,$2F,$4B,$17,$CB,$B7,$CA,$79,$BC,$AB,$7E
	.dc.b	$F7,$0F,$F9,$6B,$97,$F0,$EE,$4F,$F2,$EE,$CF,$83,$FF,$42,$B6,$33
	.dc.b	$E3,$62,$72,$C8,$E9,$F0,$17,$2F,$D8,$0F,$4B,$FD,$1E,$D9,$FF,$35
	.dc.b	$BF,$8D,$6C,$6B,$B6,$46,$BB,$6D,$74,$79,$9F,$E9,$AC,$FF,$01,$3B
	.dc.b	$3A,$EB,$E0,$3C,$3A,$D8,$F0,$B8,$7C,$76,$F1,$FF,$D6,$9F,$56,$F5
	.dc.b	$3F,$D1,$33,$FC,$AA,$B7,$58,$8C,$B5,$FE,$3C,$53,$A1,$BF,$AA,$77
	.dc.b	$3E,$AF,$84,$61,$B1,$99,$E0,$D7,$C7,$3F,$C1,$F8,$F8,$F8,$DF,$D0
	.dc.b	$E7,$B3,$5D,$B5,$FF,$51,$CB,$E7,$BF,$D4,$B5,$FD,$1E,$3F,$05,$3F
	.dc.b	$EA,$06,$BF,$CA,$DE,$0E,$79,$78,$E7,$FE,$AE,$E0,$FE,$7B,$3F,$8F
	.dc.b	$F2,$F2,$FF,$B7,$EF,$9E,$1D,$F9,$B3,$29,$79,$B8,$6C,$78,$E7,$13
	.dc.b	$9F,$92,$93,$E7,$8A,$FF,$41,$C7,$FC,$5E,$FE,$87,$DB,$CE,$3F,$D9
	.dc.b	$78,$9C,$DE,$25,$CD,$4C,$45,$BF,$D5,$78,$BE,$8C,$7F,$9A,$C7,$FB
	.dc.b	$1A,$F8,$BF,$EC,$8F,$AB,$F8,$EF,$F5,$37,$E0,$F0,$7F,$B5,$1B,$00
	.dc.b	$F0,$CB,$FB,$D3,$A8,$7D,$5B,$E0,$3E,$4B,$63,$C1,$C7,$F3,$3E,$AF
	.dc.b	$FC,$E7,$1F,$ED,$A7,$CE,$FF,$65,$6E,$1F,$E0,$77,$71,$9D,$39,$79
	.dc.b	$D2,$39,$47,$5F,$89,$ED,$CD,$27,$B4,$57,$B7,$A9,$E3,$31,$8F,$7F
	.dc.b	$07,$CD,$C9,$2B,$CD,$1E,$9E,$0F,$C2,$FD,$53,$EA,$E9,$9A,$FF,$7A
	.dc.b	$C7,$F8,$75,$BE,$27,$03,$8F,$8B,$F8,$EF,$59,$ED,$5F,$94,$97,$FE
	.dc.b	$FE,$9C,$B7,$DA,$E1,$F5,$72,$30,$6F,$BB,$A7,$7C,$8C,$1F,$6F,$87
	.dc.b	$FC,$65,$F5,$7F,$E1,$C8,$C7,$07,$DF,$50,$E1,$FD,$1B,$63,$67,$E1
	.dc.b	$5E,$4F,$F8,$77,$EA,$DE,$6F,$EA,$9F,$85,$19,$E0,$C4,$FE,$BB,$EA
	.dc.b	$FF,$C7,$F3,$D0,$3F,$01,$FF,$14,$37,$2F,$F2,$53,$C2,$33,$E3,$F3
	.dc.b	$F1,$E4,$73,$9F,$F0,$D3,$7C,$6F,$06,$F2,$30,$78,$63,$F8,$3B,$FE
	.dc.b	$37,$7D,$B8,$7F,$96,$5B,$B7,$C7,$77,$19,$F8,$5F,$E1,$31,$FE,$BC
	.dc.b	$AE,$9F,$05,$89,$FE,$1B,$F5,$6F,$56,$FA,$B2,$FE,$14,$E5,$BE,$57
	.dc.b	$0F,$2E,$46,$0F,$91,$F5,$7F,$E8,$16,$FC,$37,$F1,$A8,$37,$DD,$C7
	.dc.b	$A1,$9E,$49,$7C,$A3,$F9,$0F,$D5,$FE,$CB,$8B,$F0,$8F,$F8,$F2,$7E
	.dc.b	$B3,$5F,$F2,$E3,$B7,$F8,$42,$F6,$F9,$F7,$C9,$E3,$1F,$CB,$BE,$6D
	.dc.b	$E0,$DF,$85,$95,$F8,$39,$CB,$63,$87,$85,$23,$06,$DF,$E7,$3F,$2E
	.dc.b	$F8,$CA,$6E,$7F,$CE,$57,$FE,$46,$7B,$6F,$E3,$3C,$9E,$AD,$8F,$CA
	.dc.b	$7C,$77,$E7,$1B,$E0,$65,$E7,$C5,$EB,$F0,$24,$7A,$62,$BF,$E8,$1F
	.dc.b	$A9,$7D,$5F,$F4,$E6,$C7,$BE,$BE,$27,$83,$F1,$7C,$79,$3F,$0E,$2B
	.dc.b	$EA,$FA,$BF,$94,$63,$C3,$FF,$57,$6E,$8E,$DE,$1C,$AF,$99,$1F,$D7
	.dc.b	$BE,$AD,$F3,$77,$E6,$A5,$7E,$CE,$72,$D8,$FE,$CB,$F4,$EF,$94,$A7
	.dc.b	$09,$FF,$4F,$53,$D5,$ED,$EB,$49,$E9,$15,$FF,$52,$7D,$5B,$E8,$3F
	.dc.b	$B8,$B6,$3C,$55,$E7,$F5,$FD,$1D,$72,$7C,$D1,$5F,$0F,$D5,$FB,$4C
	.dc.b	$71,$7F,$B4,$73,$47,$F7,$EF,$AB,$78,$14,$E1,$7F,$07,$C2,$F8,$DE
	.dc.b	$1C,$8C,$1F,$15,$FF,$03,$FD,$5F,$F4,$38,$FE,$34,$D7,$E1,$C7,$CF
	.dc.b	$F4,$3E,$87,$8F,$27,$E7,$45,$7F,$D8,$DC,$8F,$95,$F9,$E6,$76,$FF
	.dc.b	$75,$7D,$5F,$F6,$DA,$FF,$66,$2F,$0F,$FC,$32,$D8,$FE,$58,$EE,$78
	.dc.b	$CE,$BE,$3E,$56,$A2,$BF,$53,$E3,$7E,$99,$8A,$F0,$FA,$5D,$BE,$3E
	.dc.b	$57,$68,$AF,$E7,$FC,$8F,$CF,$18,$AF,$4F,$BB,$E6,$FA,$B9,$5E,$68
	.dc.b	$AF,$FB,$9F,$2F,$E9,$34,$E0,$FE,$AD,$18,$7F,$AC,$6A,$C3,$FC,$73
	.dc.b	$F5,$7F,$E2,$AC,$3F,$A6,$3F,$E2,$27,$FD,$E4,$DC,$3A,$47,$F8,$7E
	.dc.b	$38,$3A,$F5,$3F,$C8,$7F,$4B,$FC,$1E,$3F,$DF,$4D,$AE,$30,$AE,$98
	.dc.b	$FF,$6B,$FA,$BF,$EA,$D5,$A7,$FA,$D4,$67,$E0,$7F,$C9,$F1,$A8,$C5
	.dc.b	$FF,$63,$C1,$E5,$78,$8D,$DF,$0D,$F8,$FA,$12,$FF,$B5,$63,$E5,$C7
	.dc.b	$C5,$3E,$A2,$4E,$9F,$0F,$C4,$78,$BC,$A9,$3C,$A2,$AD,$3D,$FE,$A3
	.dc.b	$DC,$FF,$33,$B6,$7C,$37,$EC,$F3,$BD,$BB,$49,$F2,$C5,$7A,$67,$FE
	.dc.b	$11,$FA,$9F,$F4,$F4,$7F,$80,$DA,$E8,$FF,$2A,$6B,$F1,$1F,$F0,$47
	.dc.b	$D5,$FF,$51,$4F,$F5,$5D,$3E,$AB,$6F,$FC,$3C,$EC,$C4,$FF,$97,$F3
	.dc.b	$FE,$F0,$F7,$3F,$AA,$F7,$7F,$A1,$27,$71,$FD,$63,$C7,$FF,$14,$5F
	.dc.b	$FA,$D2,$FF,$A8,$4F,$F9,$A6,$31,$3A,$8F,$F1,$74,$E9,$FC,$49,$E8
	.dc.b	$F0,$34,$FF,$9C,$6D,$18,$8D,$1D,$4F,$5B,$98,$FE,$11,$F2,$FF,$EC
	.dc.b	$16,$CF,$FB,$15,$7A,$2F,$94,$D7,$84,$BF,$F6,$47,$BD,$FE,$7B,$D3
	.dc.b	$B6,$3F,$E8,$38,$C3,$E8,$6C,$7F,$B5,$1B,$1F,$E4,$B7,$F8,$A4,$72
	.dc.b	$8F,$F4,$EF,$77,$EC,$23,$83,$63,$83,$5F,$07,$87,$F1,$9B,$3E,$68
	.dc.b	$C4,$7F,$AA,$1B,$4F,$AB,$A4,$AF,$6B,$C3,$6F,$E8,$71,$FE,$E7,$6D
	.dc.b	$FF,$BA,$DA,$FD,$DF,$2B,$87,$91,$2B,$83,$67,$87,$07,$52,$4F,$57
	.dc.b	$C1,$E8,$C8,$F0,$36,$7F,$DF,$5F,$56,$ED,$F6,$B6,$BF,$DF,$8D,$9D
	.dc.b	$CF,$C0,$F8,$29,$5F,$69,$B3,$FE,$7D,$3E,$D3,$87,$EC,$35,$FB,$1E
	.dc.b	$3F,$CA,$24,$FE,$11,$B1,$FA,$5D,$91,$FC,$AB,$A3,$F9,$1C,$3C,$8C
	.dc.b	$5F,$43,$FC,$FD,$BF,$F5,$D7,$E8,$72,$7D,$32,$3E,$7C,$7F,$DA,$5D
	.dc.b	$7F,$E6,$78,$E3,$3D,$27,$39,$D7,$FC,$97,$16,$B8,$47,$36,$BF,$0D
	.dc.b	$AF,$F1,$19,$FC,$27,$CC,$F3,$C0,$DE,$8C,$37,$1F,$C4,$47,$E2,$C8
	.dc.b	$FA,$6D,$87,$F8,$76,$3E,$0C,$F9,$7D,$3E,$55,$FC,$3E,$B4,$F9,$E3
	.dc.b	$F3,$B1,$88,$F4,$47,$E1,$3E,$06,$5F,$E1,$1B,$3F,$8A,$F9,$AF,$89
	.dc.b	$C5,$FE,$82,$AF,$F5,$72,$8C,$D6,$99,$3A,$F8,$2B,$7F,$7F,$0F,$9D
	.dc.b	$C1,$E6,$E1,$FB,$FE,$FE,$27,$2C,$C3,$F9,$D5,$5B,$FC,$47,$9F,$C8
	.dc.b	$E7,$F3,$36,$BF,$D8,$54,$8C,$C7,$F6,$3F,$2A,$DF,$E4,$6B,$C6,$79
	.dc.b	$3C,$DB,$5C,$9C,$3F,$DA,$F4,$B5,$FC,$F7,$EA,$FF,$C8,$33,$FE,$D1
	.dc.b	$6C,$67,$E0,$FF,$19,$18,$DB,$FB,$FA,$8C,$9F,$91,$F3,$BF,$95,$5F
	.dc.b	$38,$E3,$C1,$23,$B6,$37,$1F,$13,$D8,$FC,$D7,$C1,$BE,$AF,$15,$E5
	.dc.b	$F3,$FC,$67,$93,$CE,$B1,$E9,$70,$FF,$BA,$EB,$F3,$2E,$4F,$F7,$77
	.dc.b	$9F,$E6,$C7,$FA,$F3,$76,$FF,$0D,$FD,$47,$FA,$12,$B7,$E4,$FE,$B6
	.dc.b	$31,$7F,$D9,$5B,$5F,$E2,$2F,$93,$F8,$97,$FE,$FB,$9C,$AA,$91,$ED
	.dc.b	$8D,$69,$D7,$DF,$EC,$D8,$E5,$1F,$BD,$FA,$B7,$B1,$9C,$3C,$5F,$1F
	.dc.b	$E5,$3C,$DE,$FD,$C3,$C4,$E1,$F2,$7A,$56,$BD,$4F,$B4,$FE,$15,$F4
	.dc.b	$EF,$83,$C5,$7A,$FC,$5F,$84,$7C,$3F,$3A,$C7,$BD,$C3,$EF,$E1,$FF
	.dc.b	$22,$EA,$FA,$BC,$E7,$F2,$79,$3E,$61,$F3,$7D,$6B,$87,$C9,$70,$F3
	.dc.b	$3E,$54,$BE,$AF,$E6,$BF,$32,$7F,$91,$7D,$53,$EA,$E9,$8F,$9C,$B8
	.dc.b	$73,$BF,$A9,$5F,$DF,$34,$FF,$2D,$B6,$7B,$BF,$DE,$A7,$FC,$48,$E5
	.dc.b	$E3,$EC,$78,$0F,$93,$C7,$6B,$B4,$AB,$F9,$9D,$F6,$F3,$FC,$FE,$97
	.dc.b	$A1,$B7,$C2,$BF,$A6,$BC,$B9,$3F,$DF,$8E,$6B,$DF,$F2,$72,$38,$74
	.dc.b	$5C,$BC,$EF,$C4,$3E,$EF,$AB,$67,$B7,$CF,$7C,$83,$E4,$F9,$96,$3C
	.dc.b	$4E,$1F,$27,$99,$70,$F8,$6F,$D0,$3F,$13,$EB,$D9,$F5,$7B,$7F,$50
	.dc.b	$FA,$B5,$63,$DC,$E1,$F7,$7A,$D6,$3F,$CA,$5F,$57,$FC,$46,$7F,$02
	.dc.b	$D8,$FD,$63,$FB,$07,$7F,$0F,$FD,$8E,$31,$C7,$FE,$96,$6D,$7E,$2B
	.dc.b	$87,$FC,$F4,$FD,$68,$CD,$F4,$F8,$2E,$47,$3F,$6E,$D6,$DC,$3C,$7D
	.dc.b	$6B,$5E,$87,$C4,$7E,$B2,$29,$F8,$4F,$AB,$76,$7F,$79,$6C,$64,$C4
	.dc.b	$7F,$7D,$D9,$EB,$53,$1C,$D6,$BF,$1F,$F5,$4F,$AB,$FF,$02,$36,$33
	.dc.b	$EC,$62,$73,$C5,$27,$8F,$A9,$76,$FF,$41,$60,$F4,$79,$9D,$3E,$12
	.dc.b	$DF,$E7,$BE,$95,$F1,$DF,$E1,$36,$C6,$7E,$7F,$13,$9E,$A9,$3C,$7A
	.dc.b	$AE,$7F,$E9,$1F,$AB,$FD,$3C,$FF,$00,$6C,$7F,$2B,$7F,$4A,$EF,$C3
	.dc.b	$FF,$11,$43,$DD,$A6,$3D,$DB,$7F,$B4,$9F,$F4,$57,$77,$FB,$69,$B1
	.dc.b	$FE,$9C,$E3,$E7,$8F,$F8,$B6,$1E,$D5,$31,$F2,$F6,$3F,$7F,$F5,$6F
	.dc.b	$A2,$FF,$1A,$36,$33,$ED,$62,$7F,$C7,$1F,$57,$FE,$77,$9F,$E0,$ED
	.dc.b	$C7,$FD,$64,$FC,$04,$E7,$4F,$F7,$8B,$FE,$07,$9E,$3E,$2F,$4D,$FA
	.dc.b	$BF,$E4,$5F,$AB,$78,$94,$C7,$89,$6F,$F8,$BC,$FF,$89,$5B,$F0,$1C
	.dc.b	$78,$39,$3F,$E4,$FF,$7B,$FA,$34,$FF,$1E,$6E,$BF,$EC,$97,$E0,$37
	.dc.b	$D9,$E1,$EC,$CA,$E1,$EC,$DA,$FF,$68,$7D,$5B,$E3,$3F,$CB,$8D,$AF
	.dc.b	$E5,$6F,$C1,$7F,$A1,$B9,$9E,$3D,$5C,$0E,$3E,$95,$AE,$2E,$1E,$7F
	.dc.b	$15,$6B,$C3,$E4,$F8,$0F,$0F,$93,$6B,$C2,$E1,$F1,$FA,$76,$BD,$3D
	.dc.b	$FE,$83,$D3,$E5,$DA,$F4,$B8,$7F,$A2,$F3,$3B,$31,$7F,$EF,$1E,$B7
	.dc.b	$C1,$FF,$9F,$5B,$7F,$D3,$1F,$DA,$36,$35,$E4,$48,$EB,$EF,$DB,$FF
	.dc.b	$7E,$0F,$4F,$FD,$10,$D9,$FE,$AE,$FE,$2C,$DC,$6B,$B6,$4E,$BB,$AD
	.dc.b	$74,$79,$9F,$E9,$3D,$3F,$C6,$31,$99,$EB,$B7,$80,$F0,$ED,$63,$C2
	.dc.b	$E1,$F1,$CB,$C9,$FF,$15,$7D,$5B,$D5,$FF,$5E,$5F,$AA,$B4,$7F,$A4
	.dc.b	$4A,$7F,$A4,$EB,$0E,$0F,$F0,$5F,$D5,$7E,$F3,$EA,$FF,$D3,$0D,$8F
	.dc.b	$F5,$93,$5F,$F6,$53,$5E,$3E,$3B,$8F,$8B,$63,$8B,$5E,$2D,$7C,$3F
	.dc.b	$3B,$FE,$1C,$96,$9E,$4E,$4C,$36,$BF,$C1,$78,$2F,$31,$FE,$4D,$8F
	.dc.b	$C3,$BF,$8C,$F3,$72,$FF,$00,$6C,$6A,$3D,$15,$F0,$B9,$73,$E7,$C9
	.dc.b	$EF,$15,$D9,$DF,$E4,$F9,$99,$CF,$D3,$19,$3F,$3B,$EE,$78,$B9,$0F
	.dc.b	$08,$FF,$12,$F1,$7F,$37,$6F,$27,$A2,$0D,$E5,$C3,$CF,$CB,$95,$C0
	.dc.b	$FF,$17,$FD,$5F,$C5,$6F,$17,$C2,$8F,$81,$9F,$9A,$51,$B7,$DA,$62
	.dc.b	$7E,$4F,$D5,$BE,$02,$9C,$E3,$E6,$FD,$5F,$F9,$9E,$BF,$C7,$EF,$F4
	.dc.b	$16,$E3,$AF,$27,$6F,$0E,$F9,$3A,$38,$FC,$47,$4E,$79,$18,$3A,$BE
	.dc.b	$0F,$5A,$46,$0F,$0F,$83,$E4,$E3,$91,$83,$CB,$D1,$F8,$4F,$AA,$7D
	.dc.b	$5D,$3D,$71,$9F,$17,$5E,$2C,$BF,$F2,$DD,$A7,$F1,$BE,$C3,$DC,$BF
	.dc.b	$29,$2F,$FC,$C3,$39,$6F,$B9,$C3,$EA,$E4,$60,$DF,$83,$D3,$C0,$91
	.dc.b	$83,$FC,$D7,$F5,$6E,$2F,$CE,$AE,$2A,$62,$3F,$43,$F5,$7F,$E2,$8A
	.dc.b	$FF,$39,$BE,$96,$ED,$AF,$37,$F4,$FF,$57,$FE,$29,$9F,$F3,$CB,$E9
	.dc.b	$7F,$0A,$DF,$03,$F1,$4E,$E3,$3E,$47,$2F,$22,$47,$28,$EB,$E3,$F6
	.dc.b	$DE,$4F,$68,$FD,$B9,$FE,$2A,$7F,$D1,$2F,$BD,$BA,$6B,$AB,$7F,$19
	.dc.b	$C3,$E0,$E5,$70,$8F,$E0,$3F,$56,$F5,$2F,$A9,$2B,$FD,$2B,$39,$6F
	.dc.b	$95,$C3,$CA,$91,$83,$FD,$37,$F5,$6F,$7B,$F8,$6A,$F8,$7A,$62,$33
	.dc.b	$EF,$6B,$8A,$5F,$28,$FE,$2B,$F5,$7F,$A7,$8F,$D8,$BF,$D8,$DF,$F5
	.dc.b	$4B,$5C,$F9,$3C,$3D,$CE,$3E,$0C,$9D,$C7,$F2,$0F,$9B,$78,$77,$E2
	.dc.b	$25,$7F,$AD,$27,$2D,$8E,$1E,$24,$8C,$1B,$7F,$95,$FC,$B3,$E5,$29
	.dc.b	$B9,$FE,$13,$7F,$E2,$E7,$B6,$FE,$33,$C9,$EB,$58,$FF,$62,$FC,$87
	.dc.b	$E5,$5F,$E2,$2D,$F0,$7F,$8B,$5F,$CF,$9F,$47,$17,$B7,$C0,$91,$83
	.dc.b	$D8,$FF,$40,$FA,$A7,$D5,$D3,$73,$E6,$F0,$BE,$37,$93,$23,$E2,$45
	.dc.b	$7D,$7F,$57,$F2,$CC,$67,$E4,$7B,$78,$92,$7E,$64,$7F,$50,$FA,$97
	.dc.b	$D5,$DF,$5A,$57,$FB,$92,$72,$D8,$FE,$AF,$F5,$0F,$95,$A7,$09,$FE
	.dc.b	$6B,$4F,$5F,$B7,$93,$27,$A4,$57,$F9,$DF,$D5,$BE,$83,$FB,$1B,$63
	.dc.b	$C5,$5E,$7F,$67,$D1,$D5,$27,$CD,$15,$F0,$FD,$1F,$B4,$C7,$0F,$F4
	.dc.b	$BC,$D1,$FD,$C3,$EA,$DD,$F4,$E1,$7F,$07,$C1,$F8,$DE,$14,$8C,$1F
	.dc.b	$15,$FE,$F3,$F5,$7F,$CD,$A0,$7F,$56,$70,$FC,$48,$FA,$1E,$7F,$D1
	.dc.b	$F6,$E4,$FC,$F8,$AF,$F5,$EE,$47,$CB,$7D,$03,$3B,$7F,$65,$FA,$BF
	.dc.b	$EC,$15,$E0,$8C,$78,$5F,$E0,$C9,$37,$E0,$B0,$E7,$E3,$E5,$F1,$CF
	.dc.b	$D2,$AF,$8B,$93,$DA,$FF,$A0,$5C,$0E,$6F,$F2,$C3,$86,$BF,$DE,$23
	.dc.b	$83,$7D,$2F,$37,$C7,$48,$E7,$29,$C4,$7F,$80,$3C,$3F,$DC,$37,$C7
	.dc.b	$CE,$BF,$0F,$E0,$3E,$07,$D5,$CA,$E7,$15,$FD,$CF,$B7,$FE,$30,$DF
	.dc.b	$FA,$74,$79,$6B,$FC,$67,$F5,$7F,$E1,$F8,$FF,$07,$BF,$E1,$E7,$DE
	.dc.b	$D7,$3E,$B7,$0F,$5B,$FC,$77,$F4,$BE,$6C,$FF,$C3,$4D,$E2,$F9,$B8
	.dc.b	$71,$9F,$F0,$F7,$D5,$FF,$55,$6F,$C1,$FF,$91,$23,$FA,$D5,$7D,$47
	.dc.b	$4D,$8E,$FE,$64,$CF,$E9,$D7,$07,$F2,$6F,$85,$BE,$8F,$53,$B7,$0D
	.dc.b	$8F,$B7,$9F,$F1,$B7,$FC,$5E,$DC,$BC,$F5,$E5,$F8,$8F,$37,$9B,$2F
	.dc.b	$C3,$15,$FF,$1B,$FA,$AF,$77,$FC,$BE,$D8,$F3,$D7,$BF,$D0,$F8,$1A
	.dc.b	$C9,$F5,$C5,$7F,$C8,$3F,$53,$FE,$8D,$7F,$22,$D7,$FC,$8B,$B2,$BF
	.dc.b	$CD,$BF,$57,$FD,$67,$5F,$F2,$63,$FD,$5D,$BF,$13,$AF,$6B,$6F,$9B
	.dc.b	$C4,$98,$FE,$7A,$F8,$72,$7E,$99,$FF,$00,$47,$EA,$0F,$E0,$2E,$1C
	.dc.b	$35,$E1,$F0,$3F,$EA,$18,$FF,$37,$C7,$F0,$89,$DF,$9B,$FB,$26,$C6
	.dc.b	$4F,$04,$65,$F1,$79,$D5,$F0,$B5,$57,$F6,$37,$91,$A2,$9A,$F1,$9F
	.dc.b	$2F,$EB,$8F,$5C,$61,$C4,$BD,$A7,$C9,$6E,$79,$5D,$AF,$C0,$65,$D9
	.dc.b	$B3,$FE,$7D,$3C,$BC,$DF,$E9,$49,$F8,$ED,$8F,$9D,$AD,$7E,$6B,$5F
	.dc.b	$6B,$68,$E7,$9B,$FD,$6F,$7E,$E6,$D7,$85,$B1,$F6,$5B,$1F,$C8,$BB
	.dc.b	$BF,$CE,$33,$F1,$9B,$1E,$96,$C7,$E0,$5B,$1C,$7E,$8F,$CD,$06,$92
	.dc.b	$7C,$9B,$F9,$C8,$E0,$F0,$78,$3C,$88,$E0,$F2,$79,$3F,$2F,$0F,$EA
	.dc.b	$38,$23,$10,$8C,$4F,$D7,$6C,$7D,$F5,$F5,$7C,$1B,$3B,$8C,$11,$C1
	.dc.b	$27,$94,$78,$3E,$AF,$FC,$3B,$3F,$85,$6C,$FE,$35,$AF,$EA,$5A,$FF
	.dc.b	$52,$EA,$F9,$EF,$EB,$5A,$E7,$E3,$BD,$B1,$88,$FE,$E3,$AB,$CF,$F8
	.dc.b	$2D,$C3,$F2,$CD,$9F,$CF,$35,$ED,$E7,$74,$24,$FE,$76,$FF,$D9,$CD
	.dc.b	$D7,$C2,$47,$07,$CD,$F8,$7F,$94,$47,$07,$D1,$F3,$3E,$71,$1C,$13
	.dc.b	$0F,$F6,$FE,$DF,$D8,$E3,$FB,$7A,$F3,$BF,$DF,$7D,$DF,$EE,$78,$D5
	.dc.b	$70,$9F,$B4,$4A,$FC,$03,$F2,$3D,$89,$1E,$D6,$CF,$F0,$1C,$FF,$C6
	.dc.b	$93,$FC,$BA,$7B,$BF,$57,$E2,$6D,$7F,$88,$2B,$6F,$F3,$6F,$37,$2A
	.dc.b	$13,$87,$FD,$AD,$5F,$CE,$A3,$78,$F2,$CB,$D9,$8E,$27,$2D,$AF,$87
	.dc.b	$1C,$EB,$18,$7C,$8F,$98,$E6,$74,$F9,$8B,$7F,$29,$C3,$FD,$83,$CC
	.dc.b	$B5,$F4,$2F,$C6,$79,$3C,$EB,$5F,$49,$C3,$FD,$A3,$E1,$ED,$34,$4F
	.dc.b	$F8,$B3,$EA,$FF,$C9,$EF,$F8,$A1,$B3,$B8,$D7,$C3,$FC,$89,$CD,$72
	.dc.b	$AF,$F2,$6C,$27,$47,$CE,$F3,$FF,$9C,$DF,$58,$E5,$F3,$92,$3C,$18
	.dc.b	$DC,$3C,$07,$47,$C0,$7D,$8F,$88,$7D,$56,$2B,$D1,$E8,$F9,$4F,$37
	.dc.b	$A3,$63,$D8,$E1,$FF,$22,$D3,$E7,$71,$8F,$80,$FF,$02,$79,$1F,$86
	.dc.b	$F2,$3D,$31,$9F,$AB,$FF,$4A,$E1,$FE,$4E,$9C,$F0,$FC,$3A,$56,$9E
	.dc.b	$1F,$0F,$F9,$0F,$43,$E3,$61,$FE,$56,$AD,$5F,$17,$DB,$19,$DB,$F2
	.dc.b	$F1,$FE,$99,$8C,$1C,$7E,$17,$27,$E7,$3E,$AD,$ED,$E2,$7B,$7C,$9F
	.dc.b	$23,$9F,$99,$63,$C4,$E1,$F2,$7A,$77,$6F,$80,$F8,$8F,$D4,$FD,$43
	.dc.b	$E2,$B1,$5E,$BE,$3F,$39,$E8,$F8,$4B,$1E,$F7,$0F,$F9,$DB,$47,$D5
	.dc.b	$E3,$2F,$9D,$C9,$F0,$0F,$83,$E9,$DB,$F9,$0E,$1F,$F3,$FF,$D5,$FC
	.dc.b	$A7,$FD,$0F,$C5,$FD,$C1,$FF,$41,$FD,$5F,$F5,$F6,$FA,$BF,$C0,$A3
	.dc.b	$F9,$4E,$3F,$81,$74,$DF,$3F,$46,$FC,$FB,$7D,$1C,$3F,$E9,$0E,$17
	.dc.b	$1E,$2F,$A7,$9F,$DC,$F0,$79,$F6,$FC,$AE,$1E,$5C,$B6,$BC,$9E,$77
	.dc.b	$9C,$F4,$74,$5A,$F6,$B8,$7C,$5F,$1B,$6B,$D9,$F3,$DE,$F3,$E0,$79
	.dc.b	$B6,$BE,$2B,$87,$E3,$F9,$B6,$BE,$23,$F1,$CF,$C0,$FA,$76,$7E,$6F
	.dc.b	$B3,$F2,$CF,$99,$56,$3E,$8B,$87,$E3,$7B,$76,$3F,$61,$F5,$7F,$C3
	.dc.b	$27,$FD,$7E,$D8,$FF,$6C,$BF,$A8,$77,$F1,$7F,$CE,$A3,$1E,$27,$EC
	.dc.b	$5B,$5F,$EC,$8E,$4F,$F1,$EC,$E7,$CD,$E1,$F0,$92,$BF,$D8,$F6,$F8
	.dc.b	$4C,$65,$E4,$7B,$1D,$FE,$56,$D7,$07,$0F,$E0,$4B,$3F,$C2,$FE,$AF
	.dc.b	$F2,$4D,$F0,$3F,$A8,$C7,$F9,$03,$3F,$E1,$0E,$9F,$E1,$5B,$FF,$54
	.dc.b	$BF,$F6,$DB,$6F,$F8,$B3,$FD,$59,$BE,$BF,$F5,$CF,$AB,$7D,$55,$31
	.dc.b	$EC,$DC,$3F,$DC,$B3,$F3,$6C,$D3,$E4,$A3,$3C,$5D,$7D,$29,$5C,$BE
	.dc.b	$22,$D7,$F2,$9F,$A5,$7C,$A7,$F6,$E6,$D7,$FB,$D1,$7C,$3E,$67,$5D
	.dc.b	$32,$B5,$D3,$70,$FE,$69,$F5,$6F,$0F,$FB,$CB,$6B,$FD,$FE,$FF,$83
	.dc.b	$B8,$FF,$99,$39,$AE,$3E,$F7,$03,$8F,$BD,$6F,$8C,$F4,$3F,$C0,$AF
	.dc.b	$CA,$32,$7C,$3B,$FE,$7C,$BD,$BD,$9C,$3E,$66,$C6,$CF,$95,$4F,$AB
	.dc.b	$7C,$5F,$F8,$49,$B3,$B9,$FA,$10,$F6,$B5,$5F,$E1,$8F,$AB,$FF,$30
	.dc.b	$D7,$F5,$76,$E5,$FD,$45,$F8,$0D,$F8,$9E,$1E,$2C,$CB,$B5,$AF,$FC
	.dc.b	$E5,$18,$53,$FE,$26,$FA,$B7,$89,$4C,$78,$76,$F9,$CB,$E0,$B8,$1C
	.dc.b	$78,$2C,$F8,$3E,$03,$FA,$43,$7D,$CF,$F1,$9C,$7C,$2D,$3F,$B9,$3A
	.dc.b	$6F,$B9,$A3,$7E,$E5,$BF,$5F,$D5,$FF,$A2,$38,$7F,$B9,$36,$3E,$8B
	.dc.b	$F1,$9B,$F1,$9F,$E4,$28,$F8,$F1,$AE,$AC,$7A,$99,$6C,$76,$8E,$1F
	.dc.b	$13,$8C,$FF,$C1,$17,$FB,$5C,$7C,$3E,$47,$5F,$53,$B3,$71,$F3,$C7
	.dc.b	$87,$D4,$C6,$1C,$BB,$7C,$87,$97,$CB,$B7,$EE,$70,$FD,$AE,$7F,$C3
	.dc.b	$39,$3E,$EE,$4F,$85,$FF,$2F,$B6,$33,$E4,$62,$72,$C8,$E9,$F0,$37
	.dc.b	$2F,$F1,$20,$F5,$3F,$CD,$2D,$9F,$C5,$BF,$88,$6C,$6B,$B2,$46,$BC
	.dc.b	$3B,$5D,$1E,$67,$F9,$D3,$3F,$F1,$A4,$EC,$EB,$AF,$80,$F0,$EB,$63
	.dc.b	$C2,$E1,$F1,$DB,$C9,$FE,$35,$F5,$6F,$5B,$F6,$72,$BF,$5B,$B3,$27
	.dc.b	$EE,$2A,$B4,$A6,$CE,$7F,$55,$FC,$0F,$EA,$F6,$D8,$E6,$D7,$7D,$E9
	.dc.b	$DF,$91,$FE,$75,$9D,$B5,$EA,$D7,$F8,$5D,$23,$11,$FC,$DE,$1F,$41
	.dc.b	$FC,$43,$97,$FC,$A9,$1F,$83,$9F,$E0,$2D,$7F,$CC,$0F,$07,$3C,$BC
	.dc.b	$76,$FE,$33,$C1,$FC,$F2,$7F,$21,$E6,$E5,$F0,$55,$56,$E6,$C2,$F3
	.dc.b	$E9,$88,$B3,$E4,$F9,$3D,$98,$7C,$AF,$5C,$7F,$C5,$B1,$FA,$1F,$77
	.dc.b	$6F,$F1,$CE,$37,$37,$8D,$73,$53,$11,$67,$F5,$5E,$5F,$53,$0F,$FA
	.dc.b	$0F,$EA,$FE,$23,$76,$F7,$47,$C9,$E1,$FF,$40,$3A,$6C,$7C,$3F,$AB
	.dc.b	$7C,$0D,$38,$C7,$C7,$FA,$BF,$F3,$56,$1F,$CD,$1F,$E8,$AD,$C7,$5E
	.dc.b	$56,$DE,$1E,$84,$9D,$1C,$7E,$2B,$A7,$34,$8C,$1D,$5F,$07,$AD,$23
	.dc.b	$07,$87,$C2,$F2,$71,$C8,$C1,$E5,$E9,$FB,$DF,$56,$F9,$4A,$7A,$E3
	.dc.b	$3E,$36,$BC,$69,$7F,$D4,$2D,$3F,$87,$F6,$1E,$DD,$F9,$49,$7F,$D5
	.dc.b	$A7,$2D,$F8,$2E,$1F,$57,$23,$06,$FC,$1E,$9E,$04,$8C,$1F,$D8,$3E
	.dc.b	$AF,$FC,$22,$DE,$1F,$CC,$41,$BC,$5F,$9B,$FA,$B7,$9D,$77,$98,$7F
	.dc.b	$6A,$9C,$BF,$DB,$3E,$AF,$FC,$62,$DE,$0F,$D2,$47,$AA,$3F,$D9,$CE
	.dc.b	$9B,$E4,$6B,$C8,$91,$CE,$38,$F8,$FC,$B7,$93,$DE,$3F,$60,$7E,$D5
	.dc.b	$FF,$6D,$3E,$E6,$E9,$AE,$9D,$FC,$7F,$0F,$85,$95,$FE,$04,$B4,$FE
	.dc.b	$EF,$EA,$FF,$C6,$13,$FE,$08,$5E,$AE,$4D,$7A,$DB,$F2,$78,$79,$52
	.dc.b	$F8,$47,$F0,$5F,$AB,$FF,$37,$47,$F8,$49,$7C,$46,$4D,$7B,$FB,$F4
	.dc.b	$B8,$71,$4B,$E1,$1F,$C3,$FE,$AF,$F4,$F1,$FE,$F9,$7F,$C5,$0F,$F8
	.dc.b	$79,$AE,$7C,$8E,$1E,$E7,$1F,$06,$4E,$E3,$F8,$D7,$CF,$3C,$3B,$F0
	.dc.b	$F2,$BF,$C5,$33,$96,$C7,$0F,$12,$46,$0D,$BF,$C9,$7E,$61,$F1,$F4
	.dc.b	$DC,$BD,$37,$95,$FE,$48,$C7,$FC,$63,$3E,$0E,$1F,$29,$E6,$F5,$AC
	.dc.b	$7F,$8F,$3E,$43,$F2,$6F,$3B,$FF,$28,$C6,$30,$EB,$3F,$E7,$DC,$3E
	.dc.b	$9B,$B3,$1C,$7F,$0B,$C1,$91,$F0,$22,$BB,$1F,$44,$FA,$BF,$E8,$4D
	.dc.b	$8F,$8F,$5F,$33,$C3,$F9,$BE,$5C,$9F,$97,$15,$E9,$8B,$EA,$FE,$99
	.dc.b	$9C,$FD,$32,$9D,$3F,$E5,$CC,$7F,$A8,$37,$C2,$E2,$FF,$98,$3E,$AD
	.dc.b	$F3,$9F,$D5,$DF,$E6,$31,$EB,$6E,$1D,$9D,$C7,$1F,$5B,$FA,$F7,$D4
	.dc.b	$3E,$5A,$9C,$E3,$A7,$AF,$D7,$CA,$97,$E1,$8A,$FC,$9F,$A8,$7D,$0F
	.dc.b	$F6,$96,$C7,$9E,$BC,$DE,$CF,$9F,$AE,$4F,$AE,$2B,$E9,$FA,$BF,$61
	.dc.b	$8D,$BD,$4F,$F7,$5F,$AB,$78,$14,$DD,$7C,$5F,$0B,$E2,$78,$72,$30
	.dc.b	$7C,$97,$FB,$F7,$D5,$FF,$41,$8F,$F4,$5B,$5F,$95,$1F,$37,$D0,$F9
	.dc.b	$DE,$04,$9F,$A5,$15,$FC,$57,$23,$E5,$BE,$69,$9D,$7E,$FF,$D5,$FF
	.dc.b	$6A,$AF,$F4,$CA,$F0,$BF,$C2,$0D,$8C,$FC,$16,$27,$3F,$1F,$27,$8C
	.dc.b	$57,$E9,$7C,$6F,$D1,$31,$87,$83,$D3,$EB,$F1,$F2,$BC,$31,$5F,$D1
	.dc.b	$78,$1F,$B6,$6F,$91,$9B,$F4,$FC,$0F,$9B,$EA,$E5,$7A,$A2,$BF,$AB
	.dc.b	$CF,$FB,$2B,$FB,$BF,$77,$37,$FA,$0F,$AB,$FF,$0E,$CF,$FA,$E9,$FF
	.dc.b	$0E,$B7,$D7,$FF,$19,$46,$A3,$F6,$CE,$9F,$DC,$7D,$37,$9E,$BF,$56
	.dc.b	$DE,$6F,$F6,$BD,$6E,$7F,$5F,$F5,$7F,$D5,$DB,$EF,$FF,$8F,$E3,$FA
	.dc.b	$E4,$6B,$E7,$71,$3F,$E4,$3C,$1E,$75,$33,$97,$FB,$4B,$0F,$02,$0F
	.dc.b	$F1,$4C,$3C,$3E,$97,$7E,$1F,$03,$D4,$95,$CD,$F2,$7C,$57,$97,$CD
	.dc.b	$93,$DE,$2B,$FC,$67,$D6,$FE,$BE,$FF,$B9,$1A,$FF,$1D,$C7,$F8,$5B
	.dc.b	$7D,$1F,$81,$9C,$8F,$54,$57,$F9,$27,$D4,$FF,$A3,$DF,$F7,$7B,$5E
	.dc.b	$93,$FE,$67,$FA,$BF,$EB,$3A,$FF,$1C,$7F,$AB,$B7,$E2,$F5,$ED,$6D
	.dc.b	$FF,$37,$66,$F0,$E9,$9C,$7F,$DF,$BF,$57,$FB,$E7,$FD,$32,$D7,$FD
	.dc.b	$3A,$D7,$FD,$46,$D7,$E9,$0C,$7F,$C4,$8F,$FA,$89,$AF,$FA,$99,$AF
	.dc.b	$FA,$B1,$AF,$27,$FC,$EE,$FF,$38,$3A,$79,$7D,$B9,$F8,$36,$38,$35
	.dc.b	$FC,$24,$BC,$6C,$47,$95,$B1,$98,$FC,$55,$78,$E7,$3E,$2B,$E6,$7D
	.dc.b	$4D,$AF,$57,$AD,$FE,$C3,$7D,$91,$97,$D9,$F0,$9F,$DA,$30,$A7,$68
	.dc.b	$F7,$46,$5B,$19,$7D,$3A,$65,$B1,$97,$E7,$46,$5B,$19,$CB,$FC,$06
	.dc.b	$D3,$4E,$23,$E6,$36,$3F,$DB,$BF,$05,$FE,$75,$7B,$36,$7E,$93,$63
	.dc.b	$7F,$1B,$E1,$87,$D2,$D2,$4F,$27,$5E,$54,$8E,$AD,$BE,$BE,$1F,$32
	.dc.b	$4F,$89,$ED,$E5,$7E,$0E,$7F,$80,$60,$8C,$47,$FA,$E9,$B1,$9A,$FF
	.dc.b	$79,$B6,$7F,$DE,$ED,$7E,$0C,$F0,$38,$FC,$F4,$AE,$2D,$9F,$D1,$FB
	.dc.b	$DE,$0E,$4F,$F0,$D5,$7E,$A2,$30,$79,$39,$5E,$85,$24,$FE,$A2,$FF
	.dc.b	$53,$3A,$8C,$47,$F1,$D8,$FC,$D4,$62,$3F,$C9,$DC,$BD,$49,$19,$EB
	.dc.b	$FF,$64,$B7,$6F,$8B,$18,$AE,$AD,$7F,$77,$E0,$7A,$C9,$5F,$C0,$9F
	.dc.b	$97,$F0,$D2,$3E,$5B,$67,$E5,$F9,$9F,$40,$93,$F4,$1B,$1F,$E4,$1E
	.dc.b	$0F,$F4,$1C,$FF,$39,$ED,$FF,$1A,$C3,$CD,$FF,$27,$C6,$2E,$D7,$F0
	.dc.b	$EC,$BF,$C4,$2F,$F3,$1E,$AC,$E3,$FD,$0E,$72,$FF,$37,$8C,$A8,$CB
	.dc.b	$F4,$5A,$F5,$F6,$3E,$39,$F2,$3E,$2E,$5F,$CD,$6C,$FE,$2A,$7A,$E7
	.dc.b	$E0,$CB,$D7,$58,$9C,$B2,$FF,$9E,$D5,$BF,$89,$E8,$72,$39,$FC,$CD
	.dc.b	$AF,$8C,$7F,$2E,$A4,$6A,$33,$5F,$D8,$FC,$9F,$F7,$2C,$62,$15,$C2
	.dc.b	$30,$E1,$AF,$AB,$CD,$F5,$B6,$37,$1F,$2A,$9F,$0D,$9B,$FF,$3E,$7D
	.dc.b	$5F,$F9,$1A,$FC,$0D,$8C,$FC,$37,$D2,$8C,$6D,$FE,$74,$51,$93,$B7
	.dc.b	$89,$6F,$9F,$CE,$F1,$39,$63,$63,$BB,$87,$E0,$7C,$3B,$EA,$F1,$9F
	.dc.b	$9B,$CE,$F1,$1E,$3F,$3E,$DF,$A1,$C3,$CB,$E7,$AD,$7F,$A6,$7C,$9F
	.dc.b	$E8,$93,$F3,$5B,$1F,$E9,$DF,$A8,$FF,$43,$CE,$75,$FC,$8C,$62,$75
	.dc.b	$F0,$FF,$94,$F3,$BE,$0D,$FD,$7A,$D7,$17,$FA,$61,$F5,$46,$75,$E7
	.dc.b	$F0,$9D,$1B,$1C,$63,$F3,$FF,$56,$F6,$B3,$9F,$97,$E4,$FC,$67,$93
	.dc.b	$D2,$B8,$78,$5C,$3E,$3F,$84,$B5,$ED,$7D,$67,$EB,$3E,$A1,$F1,$18
	.dc.b	$AF,$57,$1F,$C0,$3E,$0F,$97,$63,$DC,$E1,$FC,$B6,$8F,$AB,$C6,$5F
	.dc.b	$2F,$97,$E4,$9F,$2B,$D5,$B7,$F1,$DC,$3F,$9F,$FA,$BF,$92,$FF,$85
	.dc.b	$38,$FF,$C4,$0F,$E8,$FE,$AF,$FA,$BB,$7D,$5F,$E0,$B1,$FC,$97,$9B
	.dc.b	$FD,$8E,$EB,$2F,$F6,$C3,$9A,$FF,$0F,$B6,$3F,$CA,$4F,$FA,$36,$3B
	.dc.b	$79,$DD,$3D,$59,$1F,$B3,$E2,$72,$63,$0E,$BE,$77,$88,$F1,$F9,$B6
	.dc.b	$FD,$0E,$1E,$5C,$D6,$BC,$BF,$03,$EA,$3D,$7D,$36,$BE,$03,$87,$C5
	.dc.b	$F1,$96,$BD,$BF,$3D,$F0,$8F,$87,$E6,$5A,$F9,$0E,$1F,$95,$E6,$5A
	.dc.b	$F8,$AF,$CA,$3F,$09,$E9,$D9,$FA,$1F,$45,$F3,$8F,$9F,$56,$3E,$A3
	.dc.b	$87,$E3,$7B,$96,$1C,$BF,$EF,$AF,$AB,$FE,$13,$5F,$E5,$06,$CF,$FA
	.dc.b	$0D,$FD,$5B,$C2,$35,$F1,$7F,$D0,$A7,$35,$CF,$CD,$FC,$D7,$6E,$5D
	.dc.b	$7C,$1C,$C3,$F9,$55,$BD,$8C,$65,$D4,$F0,$3D,$4E,$DF,$2B,$6B,$8B
	.dc.b	$87,$F0,$45,$9E,$43,$EA,$FF,$2E,$E1,$B9,$A7,$FC,$DF,$1F,$E3,$B8
	.dc.b	$C4,$FF,$A8,$E3,$73,$FE,$74,$D9,$EB,$7F,$62,$6C,$7F,$45,$5C,$D9
	.dc.b	$9F,$EC,$FF,$56,$F8,$8A,$63,$E2,$2E,$1F,$D3,$67,$E8,$E4,$69,$F3
	.dc.b	$A3,$FD,$0A,$E3,$D1,$EA,$76,$F8,$7B,$3F,$D5,$FE,$95,$F2,$5F,$DE
	.dc.b	$1B,$19,$F9,$BC,$4E,$7A,$64,$F1,$E9,$B9,$FC,$CF,$AB,$78,$5F,$E0
	.dc.b	$26,$D7,$F6,$57,$FC,$27,$3B,$9D,$7B,$D2,$B5,$EF,$5A,$E8,$73,$3F
	.dc.b	$C1,$B1,$F1,$CE,$8E,$3D,$B9,$1F,$DC,$C7,$CC,$E2,$7F,$03,$F5,$6F
	.dc.b	$8C,$FF,$0B,$B6,$BF,$BD,$2F,$6B,$33,$FE,$1C,$FA,$B7,$C5,$53,$1A
	.dc.b	$DC,$3F,$C0,$53,$FE,$01,$C7,$F1,$ED,$ED,$FC,$4C,$23,$FC,$08,$EB
	.dc.b	$C5,$FF,$14,$FD,$5B,$C3,$A6,$3C,$2B,$8F,$F8,$3E,$5F,$05,$C0,$E3
	.dc.b	$C3,$67,$FC,$29,$F0,$1F,$D2,$1B,$EE,$7F,$8D,$A3,$E0,$67,$FA,$47
	.dc.b	$4D,$F7,$34,$6F,$DC,$B7,$FE,$1F,$FA,$BF,$F3,$F7,$07,$B1,$B1,$FE
	.dc.b	$D6,$7E,$3B,$BF,$8D,$FF,$22,$C7,$E1,$30,$FD,$9B,$6B,$97,$A7,$AE
	.dc.b	$3D,$B3,$0F,$07,$C5,$5C,$38,$F8,$FD,$0E,$BE,$45,$AF,$33,$87,$D1
	.dc.b	$E9,$DA,$F0,$F6,$F9,$0F,$2F,$95,$6B,$DC,$E1,$FF,$1F,$73,$FE,$15
	.dc.b	$C9,$FE,$41,$D5,$F0,$9F,$E6,$26,$C6,$7C,$8C,$4E,$59,$1D,$3E,$06
	.dc.b	$E5,$FC,$58,$7A,$7F,$E6,$C6,$CF,$F9,$3D,$FC,$63,$FE,$9D,$8D,$76
	.dc.b	$48,$D7,$87,$6B,$A3,$CC,$FF,$3B,$E6,$E7,$C5,$75,$DB,$C0,$78,$76
	.dc.b	$B1,$E1,$70,$F8,$F0,$79,$25,$F5,$6F,$57,$FD,$65,$7E,$B9,$D9,$FE
	.dc.b	$AF,$29,$FE,$B1,$AC,$38,$3A,$0F,$AA,$FF,$32,$FD,$5F,$FA,$DD,$B1
	.dc.b	$FE,$BC,$6B,$FE,$C1,$6B,$C9,$FE,$0B,$C7,$C7,$C8,$FE,$C3,$3C,$5A
	.dc.b	$F1,$6B,$C7,$E8,$3C,$73,$FB,$CA,$FC,$0C,$FF,$A0,$DA,$FF,$50,$53
	.dc.b	$C3,$4F,$17,$2D,$AF,$91,$C1,$FC,$D1,$FC,$7F,$9B,$97,$E0,$4D,$C1
	.dc.b	$CD,$7C,$F9,$5E,$18,$AF,$87,$E5,$38,$18,$D3,$D0,$F5,$CB,$FC,$4F
	.dc.b	$3F,$9E,$F7,$36,$FF,$A7,$78,$BF,$D3,$AD,$E4,$F2,$41,$BC,$D8,$7A
	.dc.b	$79,$72,$B8,$1F,$EA,$7F,$AB,$F8,$6D,$DF,$DD,$1F,$23,$4F,$E4,$EE
	.dc.b	$A7,$FB,$CB,$9E,$1F,$C1,$FD,$5B,$E0,$7E,$33,$6B,$A1,$FE,$B5,$FA
	.dc.b	$BF,$F3,$4B,$72,$F9,$91,$FD,$0E,$35,$E4,$E2,$73,$E8,$49,$E0,$73
	.dc.b	$F8,$AE,$3C,$F2,$30,$77,$7A,$FA,$D2,$30,$78,$FC,$2F,$0F,$24,$8C
	.dc.b	$1E,$7D,$BF,$01,$F5,$6F,$97,$A7,$A2,$33,$E3,$6B,$C6,$97,$FE,$D3
	.dc.b	$BF,$F0,$7B,$FE,$D5,$F6,$1E,$D7,$E2,$D7,$CA,$D3,$11,$9F,$80,$D7
	.dc.b	$D5,$CB,$E5,$1C,$7C,$1E,$5E,$04,$9E,$D1,$F9,$7F,$AB,$71,$5E,$29
	.dc.b	$5F,$EE,$79,$CB,$FE,$E9,$FA,$BF,$F1,$4B,$78,$3F,$41,$1E,$A8,$FF
	.dc.b	$1C,$BA,$7F,$C7,$7F,$57,$FE,$29,$7F,$DE,$4F,$B9,$BA,$FE,$A6,$35
	.dc.b	$1C,$3C,$7C,$78,$F2,$38,$C6,$FC,$6E,$1A,$C9,$EB,$1F,$B3,$3F,$6C
	.dc.b	$FF,$93,$1F,$5B,$75,$D7,$46,$FE,$17,$87,$C7,$CA,$FE,$4D,$E1,$FF
	.dc.b	$81,$30,$57,$EF,$FE,$AF,$FC,$57,$5A,$EA,$FE,$43,$0F,$53,$46,$FD
	.dc.b	$5E,$1E,$4F,$1F,$26,$61,$C4,$FE,$69,$7F,$BA,$7F,$CC,$7F,$57,$FE
	.dc.b	$6F,$74,$17,$C3,$FF,$11,$6C,$69,$E0,$F2,$F7,$35,$C7,$23,$51,$FC
	.dc.b	$67,$EA,$FF,$5D,$1F,$E7,$27,$FC,$5C,$F8,$1A,$E7,$C7,$E1,$EE,$71
	.dc.b	$F0,$A4,$EE,$3F,$92,$7C,$E7,$E6,$A3,$D2,$BE,$2F,$26,$BC,$4B,$3A
	.dc.b	$F0,$F8,$98,$7F,$D0,$B4,$8F,$E6,$1F,$30,$F8,$EA,$70,$AE,$FE,$A7
	.dc.b	$8B,$E1,$24,$60,$F1,$BF,$CE,$7E,$43,$F3,$1D,$9F,$E4,$26,$F5,$7F
	.dc.b	$5C,$CF,$E5,$CF,$F9,$19,$FF,$2C,$DF,$B3,$C1,$F7,$9F,$03,$92,$C7
	.dc.b	$D8,$FA,$37,$CF,$7F,$47,$6C,$71,$3E,$2F,$87,$F2,$3C,$A9,$18,$3E
	.dc.b	$4B,$F3,$3E,$AE,$46,$0F,$9A,$F5,$F0,$A4,$60,$FC,$4F,$D5,$BE,$73
	.dc.b	$FA,$C3,$FC,$C2,$3F,$D6,$2D,$DB,$FC,$C6,$EE,$38,$FA,$BF,$D7,$FE
	.dc.b	$9D,$F2,$B4,$E7,$1D,$3D,$6E,$BE,$54,$BF,$0C,$57,$FD,$7F,$F5,$6F
	.dc.b	$A1,$FE,$D4,$D8,$F3,$D7,$9B,$D8,$F3,$F5,$C9,$F5,$C5,$7D,$3F,$57
	.dc.b	$EC,$31,$B7,$A9,$FE,$ED,$F5,$7F,$91,$9F,$D5,$B5,$E2,$7C,$6F,$0B
	.dc.b	$E2,$F8,$72,$30,$7C,$A7,$FC,$01,$F5,$7F,$D0,$63,$F6,$AD,$7E,$5C
	.dc.b	$7C,$EF,$43,$E7,$F7,$49,$FA,$71,$5F,$F7,$0F,$1B,$E5,$BE,$71,$9D
	.dc.b	$3F,$D2,$BF,$57,$FD,$A6,$BF,$DC,$EB,$C1,$FF,$09,$36,$3F,$83,$BB
	.dc.b	$9E,$33,$AF,$90,$95,$A8,$AF,$D4,$F8,$DF,$A6,$62,$BC,$3E,$9F,$6F
	.dc.b	$8F,$95,$DA,$2B,$FE,$FA,$F0,$7F,$D4,$CF,$E8,$7C,$7C,$DF,$AB,$DF
	.dc.b	$F3,$FD,$5C,$AF,$3C,$55,$3F,$B3,$F7,$3F,$6C,$65,$FF,$5B,$FD,$5F
	.dc.b	$F2,$76,$FA,$DF,$E3,$38,$FF,$10,$57,$F2,$87,$5C,$9F,$E5,$67,$31
	.dc.b	$DA,$3F,$CB,$3F,$4F,$E8,$8C,$1F,$E6,$77,$17,$DA,$3F,$99,$FD,$5F
	.dc.b	$F5,$88,$FF,$64,$BF,$D7,$63,$88,$71,$1A,$F9,$DF,$F2,$44,$62,$FC
	.dc.b	$3E,$4F,$F9,$2F,$03,$AF,$7F,$7B,$F0,$F4,$26,$7E,$28,$AF,$82,$FF
	.dc.b	$8B,$E1,$EC,$AC,$E7,$E6,$F8,$BF,$0F,$97,$30,$F5,$C5,$7D,$3E,$A7
	.dc.b	$F6,$07,$FA,$8B,$5F,$6E,$BF,$E5,$E7,$E1,$F9,$DE,$CD,$A4,$60,$F8
	.dc.b	$EF,$F9,$9F,$EA,$7F,$D2,$11,$FD,$71,$AF,$48,$FF,$36,$7D,$5F,$F5
	.dc.b	$4C,$FF,$DE,$AF,$F5,$26,$FC,$4E,$BD,$9D,$BF,$E7,$3C,$DE,$15,$33
	.dc.b	$8F,$AB,$CA,$FF,$4E,$4F,$F8,$1A,$3F,$D3,$27,$F8,$C2,$32,$7F,$A8
	.dc.b	$71,$FF,$04,$C6,$9A,$FF,$A8,$1A,$FF,$AB,$5A,$97,$C3,$73,$24,$78
	.dc.b	$0E,$0F,$7F,$1F,$FD,$57,$C7,$DD,$B3,$DD,$AF,$76,$BD,$FC,$6B,$66
	.dc.b	$92,$3F,$D8,$55,$EB,$F1,$5F,$31,$E3,$9F,$44,$61,$F3,$5B,$D8,$7C
	.dc.b	$9E,$E3,$F8,$C5,$DF,$84,$63,$DB,$0F,$8D,$18,$61,$F1,$DC,$47,$CD
	.dc.b	$6C,$7F,$B3,$36,$FF,$2B,$CF,$38,$FF,$6D,$46,$1E,$67,$F9,$A3,$E1
	.dc.b	$78,$BA,$6B,$F4,$E6,$E1,$E6,$7A,$46,$5B,$19,$8E,$BF,$09,$FE,$7F
	.dc.b	$9D,$B6,$33,$F1,$7D,$A3,$13,$F8,$CC,$3F,$02,$F6,$6D,$7A,$1B,$1B
	.dc.b	$E0,$E5,$F1,$D2,$3B,$B6,$BC,$7E,$5F,$72,$4F,$91,$B1,$A8,$C1,$1F
	.dc.b	$5F,$99,$F2,$10,$C5,$19,$8C,$36,$33,$3F,$87,$5F,$39,$92,$46,$5E
	.dc.b	$59,$3D,$2A,$49,$FC,$84,$65,$BC,$BC,$08,$E0,$DF,$2F,$42,$38,$39
	.dc.b	$FB,$3E,$02,$38,$3F,$3D,$F5,$6F,$0F,$DC,$D8,$CC,$60,$F8,$1F,$13
	.dc.b	$23,$E0,$37,$6F,$DD,$6B,$FE,$81,$9F,$89,$E7,$FC,$72,$3E,$EE,$9F
	.dc.b	$86,$47,$07,$EC,$7C,$1F,$F1,$23,$FC,$C2,$3F,$51,$E0,$7E,$A2,$73
	.dc.b	$87,$F3,$69,$CD,$7D,$A8,$4E,$AB,$87,$B3,$A9,$FB,$F2,$F9,$9C,$47
	.dc.b	$09,$B9,$71,$C3,$A3,$67,$F0,$EF,$18,$FF,$4A,$B9,$8F,$C5,$57,$23
	.dc.b	$AD,$FE,$36,$B7,$1D,$BD,$39,$1C,$BE,$66,$DF,$F6,$AC,$11,$E4,$3C
	.dc.b	$FE,$57,$FB,$A2,$B8,$1E,$29,$F4,$9E,$AF,$3E,$C7,$99,$C3,$FE,$AC
	.dc.b	$97,$C4,$63,$0F,$EF,$3F,$57,$FE,$97,$AF,$F5,$8B,$63,$3F,$0F,$F2
	.dc.b	$23,$1B,$FF,$1C,$7F,$D2,$D5,$C2,$7C,$FE,$1E,$DE,$1F,$39,$27,$AE
	.dc.b	$37,$1F,$0B,$D4,$FA,$BF,$10,$FA,$BC,$57,$93,$D0,$F1,$1E,$3F,$46
	.dc.b	$C7,$A1,$C3,$FE,$CA,$A7,$CE,$63,$1F,$F0,$BF,$93,$FD,$0A,$BF,$D9
	.dc.b	$ED,$8F,$F0,$DF,$D4,$7F,$A2,$67,$3B,$7E,$22,$31,$3A,$F8,$7F,$C5
	.dc.b	$FA,$1F,$0A,$31,$7F,$ED,$AA,$F5,$37,$5E,$5F,$0B,$CD,$B1,$B8,$FC
	.dc.b	$AF,$D5,$BD,$CC,$E1,$E4,$F3,$3C,$47,$8F,$E1,$2E,$1D,$DC,$3E,$1F
	.dc.b	$85,$B5,$EC,$7D,$47,$E8,$BE,$9D,$F1,$18,$AF,$47,$17,$BC,$F8,$1E
	.dc.b	$55,$8F,$63,$87,$FD,$EF,$A3,$EA,$F1,$97,$CA,$E4,$F9,$07,$C9,$F6
	.dc.b	$2D,$FC,$57,$0F,$FC,$00,$FA,$BF,$9B,$DD,$FD,$A3,$93,$FC,$3B,$3F
	.dc.b	$CE,$23,$FB,$5F,$D5,$3E,$AF,$F7,$6D,$9D,$1F,$C1,$57,$AB,$A9,$DB
	.dc.b	$27,$79,$FF,$44,$AF,$F4,$74,$62,$11,$CA,$30,$EA,$3F,$CC,$2D,$BF
	.dc.b	$EF,$FC,$2E,$3F,$2F,$AB,$E6,$F3,$66,$3C,$BC,$DB,$97,$27,$D6,$7D
	.dc.b	$5E,$6B,$3E,$0F,$84,$F7,$9F,$03,$A6,$C7,$91,$C3,$E9,$CF,$FD,$B2
	.dc.b	$E4,$F4,$FC,$FF,$C8,$3E,$4F,$97,$63,$DA,$E1,$F7,$F9,$77,$3F,$7B
	.dc.b	$F4,$0F,$C5,$7A,$76,$7E,$2F,$B7,$F5,$0F,$AB,$56,$3E,$4B,$87,$CD
	.dc.b	$EE,$58,$FF,$45,$FD,$5F,$F0,$E9,$FE,$7C,$D8,$FE,$F2,$FE,$BF,$8F
	.dc.b	$FC,$5A,$E6,$0E,$67,$8F,$97,$F9,$FE,$BF,$E8,$33,$9A,$E7,$F0,$78
	.dc.b	$33,$EA,$DC,$3F,$05,$5F,$D9,$EB,$CB,$3C,$BE,$5F,$75,$A7,$0F,$D1
	.dc.b	$2C,$FF,$A9,$BE,$AF,$F3,$6D,$EF,$FE,$A7,$1F,$E3,$1D,$3F,$C8,$EE
	.dc.b	$9F,$F2,$4E,$DF,$D4,$6B,$FD,$5E,$D8,$CF,$2F,$F5,$C8,$C6,$1F,$D9
	.dc.b	$7E,$AD,$F1,$1F,$D8,$1B,$1F,$9D,$7F,$A6,$61,$FE,$BA,$72,$DF,$97
	.dc.b	$E0,$71,$E1,$B5,$C7,$CF,$E8,$75,$F8,$7B,$1F,$A9,$FA,$57,$C3,$7F
	.dc.b	$70,$6C,$67,$E1,$B1,$39,$E8,$93,$C7,$A2,$E9,$FE,$CE,$FA,$B7,$83
	.dc.b	$FD,$E9,$B5,$FB,$77,$FC,$29,$3B,$9D,$7B,$72,$B5,$ED,$DA,$E8,$73
	.dc.b	$3F,$C0,$B1,$F0,$B9,$7F,$CF,$38,$7F,$9F,$5F,$F3,$DC,$7F,$6A,$3F
	.dc.b	$83,$16,$7F,$DC,$6B,$E6,$39,$FB,$3F,$E1,$0F,$AB,$7C,$55,$31,$EB
	.dc.b	$DB,$FE,$21,$F5,$7F,$E4,$E6,$FC,$47,$F8,$62,$3D,$F9,$7F,$BC,$1D
	.dc.b	$37,$C4,$D1,$BD,$ED,$FF,$1B,$FA,$B7,$89,$FE,$22,$6C,$67,$C3,$C4
	.dc.b	$E7,$DF,$93,$C7,$7B,$97,$FC,$00,$F7,$7F,$A4,$CF,$F2,$76,$C6,$77
	.dc.b	$FF,$17,$C6,$38,$BF,$D4,$6F,$F7,$4A,$E1,$EC,$6B,$D8,$DB,$FE,$36
	.dc.b	$FA,$BF,$F3,$8E,$2C,$3F,$D5,$71,$86,$F1,$6F,$E2,$F8,$74,$CC,$BF
	.dc.b	$9D,$5F,$E1,$2F,$8C,$F5,$F4,$65,$6B,$E5,$2E,$5D,$E3,$FC,$9D,$87
	.dc.b	$EB,$23,$FD,$12,$D4,$BC,$8F,$47,$C2,$FA,$59,$5E,$2F,$4A,$DF,$53
	.dc.b	$9F,$77,$BC,$F8,$1E,$55,$AF,$4B,$87,$FA,$A7,$33,$B3,$17,$CC,$F8
	.dc.b	$DF,$07,$FE,$5E,$E0,$37,$C7,$C3,$70,$FE,$BD,$29,$FF,$32,$E1,$FE
	.dc.b	$B8,$8F,$F6,$C9,$99,$FE,$C9,$6F,$47,$FC,$D6,$D8,$CF,$76,$27,$2C
	.dc.b	$8E,$3D,$D6,$BD,$07,$F9,$9C,$AF,$8B,$9F,$AB,$DB,$69,$5E,$2D,$AD
	.dc.b	$F2,$7C,$67,$91,$78,$FF,$EC,$0F,$AB,$7A,$7F,$E8,$99,$FF,$7E,$DF
	.dc.b	$B1,$D9,$93,$FD,$1D,$55,$A5,$36,$73,$FA,$AF,$F3,$6F,$D5,$ED,$B1
	.dc.b	$CD,$AE,$FC,$04,$EF,$C6,$FC,$F4,$ED,$AF,$56,$BF,$EA,$1A,$FE,$1D
	.dc.b	$C3,$DF,$E8,$3F,$D4,$8B,$2D,$72,$E1,$B7,$FE,$1D,$B5,$E6,$3F,$DC
	.dc.b	$B1,$F8,$87,$F1,$FE,$5E,$5F,$E1,$2D,$8D,$47,$E4,$5F,$C5,$3D,$7C
	.dc.b	$DD,$1B,$E7,$B1,$E1,$F9,$3E,$46,$1E,$4F,$7E,$8F,$F1,$54,$7E,$83
	.dc.b	$DB,$DB,$FE,$35,$E2,$FF,$7D,$37,$97,$CD,$06,$F3,$61,$E9,$E3,$4A
	.dc.b	$E0,$7F,$8F,$FE,$AF,$E1,$B7,$87,$E0,$47,$BB,$8B,$E6,$48,$47,$07
	.dc.b	$F6,$4E,$F8,$BF,$DA,$DF,$57,$FE,$52,$9F,$F2,$63,$5D,$DF,$CD,$FA
	.dc.b	$BF,$F3,$9C,$7F,$94,$5F,$E8,$ED,$C7,$5F,$0D,$B7,$87,$7C,$9D,$1C
	.dc.b	$7E,$23,$A7,$34,$8C,$1D,$5F,$07,$A9,$23,$07,$87,$E2,$3C,$9C,$72
	.dc.b	$30,$79,$76,$FC,$27,$D5,$BE,$5A,$9E,$B8,$CF,$8D,$AF,$1A,$5F,$F9
	.dc.b	$B6,$FF,$94,$3F,$E6,$FF,$5D,$F3,$3F,$8F,$5F,$2D,$4C,$46,$7D,$AD
	.dc.b	$7D,$5C,$BE,$51,$C7,$C0,$E5,$DF,$27,$B4,$7E,$6F,$EA,$DC,$57,$8A
	.dc.b	$57,$FA,$06,$72,$FF,$A0,$FE,$AD,$E7,$7E,$8D,$F4,$C7,$F3,$16,$F7
	.dc.b	$7E,$1A,$11,$C1,$FF,$46,$7D,$5F,$F8,$D1,$BB,$7E,$B6,$3E,$0C,$6B
	.dc.b	$E0,$71,$E4,$67,$C8,$93,$CA,$38,$78,$FC,$77,$93,$DA,$3F,$70,$7F
	.dc.b	$8C,$9F,$F4,$E3,$EC,$7F,$A4,$37,$3F,$43,$B8,$CF,$C3,$72,$F8,$49
	.dc.b	$1C,$A3,$F8,$27,$D5,$FF,$8D,$23,$FD,$50,$BD,$6C,$9A,$F5,$F7,$E5
	.dc.b	$70,$F2,$A5,$F0,$8F,$E1,$DF,$56,$F7,$AF,$C3,$CA,$FF,$5A,$4E,$5B
	.dc.b	$E9,$70,$E1,$91,$83,$FD,$75,$F5,$7F,$AA,$7F,$B3,$3F,$DA,$5C,$F9
	.dc.b	$3F,$C7,$63,$5C,$5F,$65,$7B,$D9,$F0,$A5,$F4,$8F,$E4,$9F,$37,$F9
	.dc.b	$68,$FF,$64,$2F,$87,$C9,$AF,$12,$CE,$BC,$3E,$26,$1F,$EF,$54,$8F
	.dc.b	$E6,$1F,$2E,$F8,$DA,$70,$AE,$FE,$B7,$8B,$8E,$46,$0F,$1B,$FC,$E7
	.dc.b	$E3,$BF,$25,$1F,$E0,$76,$EF,$F9,$7C,$BE,$64,$F8,$B8,$7D,$BD,$D2
	.dc.b	$3D,$B1,$5F,$F0,$87,$D1,$BE,$AF,$FA,$33,$63,$E1,$D7,$C6,$F0,$3E
	.dc.b	$3F,$93,$27,$E3,$C5,$7E,$0F,$D5,$FC,$C3,$19,$F9,$5E,$BE,$14,$9F
	.dc.b	$9F,$1F,$D5,$7E,$AD,$F3,$77,$D7,$95,$FE,$FB,$9C,$B6,$3F,$AE,$7D
	.dc.b	$43,$E8,$A9,$C2,$7F,$C5,$94,$F6,$3B,$7B,$12,$7A,$45,$7F,$C6,$5F
	.dc.b	$56,$FA,$2F,$ED,$0D,$8F,$15,$79,$FD,$AF,$47,$54,$9F,$34,$57,$C3
	.dc.b	$F4,$9E,$D3,$1C,$3F,$C3,$73,$47,$F7,$5F,$AB,$FC,$85,$7F,$55,$AF
	.dc.b	$E2,$51,$88,$E6,$7C,$3F,$07,$E4,$78,$52,$30,$7C,$77,$FC,$01,$F5
	.dc.b	$7F,$CF,$63,$A7,$B9,$FD,$85,$46,$5F,$8F,$3F,$4F,$D0,$FA,$95,$27
	.dc.b	$E9,$45,$7F,$CB,$1C,$6F,$96,$FA,$66,$73,$FE,$55,$F5,$7F,$D9,$DB
	.dc.b	$E0,$FF,$84,$E0,$DF,$9D,$C3,$9F,$91,$98,$FA,$0F,$EF,$15,$C1,$C0
	.dc.b	$51,$F6,$3E,$3E,$4E,$A7,$2D,$F4,$FC,$5F,$23,$23,$06,$DF,$D1,$F9
	.dc.b	$12,$30,$73,$F8,$1F,$57,$D5,$C8,$C1,$DD,$FD,$C7,$C1,$C8,$C1,$E8
	.dc.b	$FA,$BF,$F1,$0B,$FE,$81,$7F,$C4,$0E,$7D,$7F,$F1,$C4,$6B,$83,$FA
	.dc.b	$A3,$A9,$FF,$44,$7D,$3F,$FB,$11,$F8,$6D,$7A,$57,$F9,$07,$EA,$FF
	.dc.b	$AC,$47,$FA,$3D,$FE,$AE,$DF,$7F,$5E,$F6,$DF,$F2,$45,$BC,$BA,$66
	.dc.b	$BE,$8D,$F7,$CB,$FB,$43,$FC,$49,$BE,$8F,$87,$7F,$13,$D0,$95,$C9
	.dc.b	$F3,$7C,$57,$9F,$CC,$93,$DA,$2B,$FE,$A7,$F5,$5E,$FF,$F9,$75,$B1
	.dc.b	$E4,$AF,$81,$E7,$FC,$1D,$24,$FA,$22,$BF,$EB,$1F,$A9,$FF,$46,$BF
	.dc.b	$8C,$6B,$F1,$E7,$FC,$D5,$F5,$7F,$D3,$BC,$1F,$D5,$3F,$D4,$E7,$FD
	.dc.b	$74,$DD,$BF,$C1,$4E,$E3,$FC,$A0,$E6,$7F,$65,$93,$C1,$F5,$1D,$BD
	.dc.b	$BF,$67,$C5,$FF,$A5,$E7,$FB,$16,$1F,$EA,$16,$BC,$A3,$FD,$39,$F5
	.dc.b	$7F,$C1,$E3,$FD,$4C,$D7,$F8,$03,$5F,$F3,$13,$56,$3F,$E0,$D7,$FD
	.dc.b	$5C,$D7,$DF,$F1,$3F,$EA,$98,$C4,$7F,$9A,$EF,$FA,$F3,$FE,$6D,$CB
	.dc.b	$CD,$E6,$FF,$0F,$6C,$64,$F8,$31,$97,$E5,$F9,$35,$F8,$ED,$7C,$AD
	.dc.b	$BF,$EF,$B1,$CC,$F9,$BE,$46,$48,$E4,$F6,$BC,$7C,$77,$CD,$7F,$B1
	.dc.b	$6B,$F8,$D3,$7D,$B7,$CB,$66,$B7,$87,$FB,$37,$4F,$F7,$D6,$9D,$5B
	.dc.b	$1F,$E4,$93,$FC,$C3,$C1,$FE,$90,$9F,$B0,$D8,$FC,$0E,$D5,$F9,$4D
	.dc.b	$7F,$CD,$2D,$AF,$F2,$FD,$7A,$F3,$FF,$63,$34,$E0,$E2,$BC,$4D,$8F
	.dc.b	$C4,$7D,$57,$32,$3E,$97,$B7,$C6,$48,$F6,$36,$BD,$9F,$0C,$F3,$7C
	.dc.b	$CD,$AF,$9C,$D9,$DC,$E8,$93,$A6,$C7,$4F,$37,$99,$27,$9B,$DB,$CD
	.dc.b	$F9,$93,$FE,$B9,$BF,$E7,$F5,$FE,$FD,$6C,$7F,$C0,$16,$BF,$A7,$F5
	.dc.b	$F8,$BF,$E9,$46,$E9,$FE,$05,$AF,$EC,$93,$FB,$49,$CF,$A5,$27,$33
	.dc.b	$F9,$F8,$C3,$F8,$3E,$5E,$64,$9D,$C7,$FA,$8A,$3C,$3C,$EF,$0A,$92
	.dc.b	$B7,$19,$78,$E3,$FE,$1D,$6F,$B1,$E8,$23,$C3,$C3,$EA,$47,$07,$C5
	.dc.b	$F8,$7F,$83,$A9,$A4,$8E,$FF,$31,$FE,$E6,$EF,$FF,$51,$62,$8C,$4F
	.dc.b	$F1,$8E,$AB,$B7,$F0,$19,$AF,$FB,$02,$3E,$69,$27,$D6,$FD,$2F,$07
	.dc.b	$FC,$5D,$1F,$19,$B3,$F3,$9A,$FD,$46,$BF,$E1,$9C,$B7,$8F,$FA,$DD
	.dc.b	$B5,$9F,$8C,$F1,$C6,$2B,$F1,$11,$83,$CF,$69,$C5,$7F,$83,$17,$A7
	.dc.b	$C8,$E7,$F3,$36,$74,$E1,$FE,$E3,$E5,$DC,$59,$79,$70,$F1,$9E,$4F
	.dc.b	$36,$DF,$77,$0F,$F7,$AF,$88,$B7,$CA,$3F,$C7,$1F,$57,$FE,$95,$7F
	.dc.b	$C4,$2D,$8D,$00,$FF,$A4,$E3,$7F,$11,$F5,$A3,$13,$C0,$DF,$C4,$7D
	.dc.b	$AF,$43,$F9,$2D,$F6,$8E,$1F,$3B,$23,$C3,$8D,$BE,$58,$23,$B1,$F8
	.dc.b	$4F,$89,$78,$58,$BF,$4F,$A3,$E3,$3C,$9E,$85,$8F,$6B,$87,$FC,$7F
	.dc.b	$5F,$93,$72,$7F,$90,$BD,$1F,$3E,$5E,$B6,$F0,$FF,$BE,$FE,$A3,$FD
	.dc.b	$1B,$5B,$F3,$3F,$2F,$18,$BF,$84,$DA,$57,$F9,$9F,$CA,$F8,$F1,$CF
	.dc.b	$E1,$78,$EB,$2B,$FC,$A7,$7F,$C2,$F0,$E7,$3D,$BC,$59,$5F,$E5,$8F
	.dc.b	$AB,$7B,$58,$BF,$0F,$97,$E4,$3C,$BF,$05,$6F,$AB,$87,$BF,$A5,$6B
	.dc.b	$D2,$FB,$0F,$D9,$FD,$3B,$E1,$F1,$5E,$6F,$13,$E0,$9F,$0B,$CB,$B1
	.dc.b	$EA,$70,$F9,$FE,$AE,$E3,$F1,$B9,$7E,$49,$F2,$BD,$7B,$5F,$09,$C3
	.dc.b	$FE,$76,$FA,$BF,$98,$FF,$A4,$78,$FF,$C4,$0F,$F9,$EB,$EA,$9F,$57
	.dc.b	$FC,$29,$B1,$9F,$53,$13,$96,$47,$1F,$52,$E7,$F9,$BB,$FF,$46,$CE
	.dc.b	$4F,$F4,$3F,$03,$8B,$B5,$F7,$F3,$A5,$78,$FE,$0E,$E3,$CD,$F2,$1F
	.dc.b	$33,$92,$CF,$A7,$E0,$7D,$07,$A7,$9E,$C7,$B5,$C3,$E2,$F4,$AE,$1E
	.dc.b	$7F,$9E,$F8,$07,$C1,$F3,$AD,$7C,$57,$0F,$C7,$F3,$AD,$7B,$DF,$90
	.dc.b	$7E,$03,$DC,$B3,$F3,$7D,$CF,$98,$7C,$DA,$B1,$F4,$5C,$3F,$1B,$DC
	.dc.b	$B1,$FC,$17,$EA,$FF,$87,$78,$9F,$A4,$8F,$E1,$D1,$8A,$FF,$66,$3F
	.dc.b	$A6,$6F,$9F,$FD,$12,$31,$3C,$3C,$EF,$CA,$F6,$E5,$D7,$B3,$30,$FF
	.dc.b	$5E,$DB,$D8,$C6,$1E,$37,$A1,$D7,$E5,$6D,$70,$70,$FE,$00,$B3,$FC
	.dc.b	$7F,$EA,$FF,$22,$DF,$07,$FA,$CC,$7F,$8A,$75,$F7,$4A,$37,$18,$DD
	.dc.b	$EC,$7F,$60,$6E,$1F,$ED,$45,$CF,$99,$FE,$CB,$F5,$6F,$85,$A6,$3D
	.dc.b	$9B,$87,$FB,$72,$7E,$66,$7F,$CE,$DA,$9F,$96,$DE,$3E,$BE,$94,$AE
	.dc.b	$5F,$09,$6B,$F9,$F7,$D2,$3E,$3F,$FB,$BB,$6B,$FD,$D6,$FF,$93,$A7
	.dc.b	$73,$AE,$99,$5A,$E9,$B5,$FD,$27,$EA,$FF,$C1,$93,$FE,$F4,$6C,$67
	.dc.b	$C1,$FF,$02,$C6,$38,$FE,$6B,$FD,$C6,$B8,$7B,$BA,$F7,$76,$F0,$C2
	.dc.b	$4F,$FC,$00,$3E,$51,$DD,$FF,$08,$BF,$D4,$4D,$7B,$58,$F9,$7C,$BF
	.dc.b	$E1,$6F,$AB,$7C,$45,$31,$ED,$5B,$FE,$D5,$A7,$E0,$BE,$AD,$F5,$5F
	.dc.b	$E1,$E6,$CE,$8F,$EE,$19,$A3,$DD,$18,$77,$1F,$DB,$D7,$8B,$CC,$E9
	.dc.b	$AD,$AE,$95,$FD,$EF,$EA,$FF,$C8,$8F,$F7,$76,$E3,$F9,$67,$F4,$D3
	.dc.b	$C2,$51,$A8,$F2,$AF,$82,$FF,$1B,$B6,$38,$3F,$26,$7B,$4F,$F8,$F3
	.dc.b	$DF,$FE,$95,$5D,$5B,$3F,$E0,$97,$E5,$C6,$7D,$AE,$1E,$D4,$8E,$1E
	.dc.b	$D5,$BF,$F0,$87,$D5,$FF,$9E,$33,$FD,$7B,$6B,$E5,$BF,$22,$33,$F1
	.dc.b	$5F,$E4,$E8,$F9,$B5,$FE,$1F,$6A,$5D,$1B,$F4,$F8,$33,$0F,$8D,$2F
	.dc.b	$91,$C5,$F9,$3C,$5E,$C7,$7F,$1A,$DF,$07,$0F,$2F,$4E,$D7,$B3,$B3
	.dc.b	$CC,$79,$FC,$8B,$5E,$07,$0F,$D7,$F3,$BF,$D9,$3B,$FC,$29,$FF,$35
	.dc.b	$AF,$84,$A6,$3C,$6B,$7F,$7E,$51,$98,$E9,$EF,$DC,$BF,$C7,$C7,$F9
	.dc.b	$72,$3F,$0C,$D3,$3D,$5F,$E7,$A7,$17,$FE,$DB,$8C,$19,$EA,$B5,$C9
	.dc.b	$C3,$EB,$E6,$B7,$E0,$D7,$A1,$D7,$4B,$5E,$27,$0F,$92,$5E,$37,$F9
	.dc.b	$6F,$EA,$DE,$AF,$FA,$DA,$FC,$D5,$A3,$FD,$5E,$53,$FD,$63,$58,$70
	.dc.b	$7E,$DB,$EA,$BF,$CA,$DF,$57,$FE,$B6,$6C,$7F,$AF,$9A,$FF,$B1,$1A
	.dc.b	$F1,$F1,$BF,$CF,$AE,$DA,$ED,$AE,$F7,$ED,$F3,$BF,$C1,$9F,$E9,$B5
	.dc.b	$F7,$67,$F7,$0D,$7F,$CE,$8A,$78,$69,$E9,$E3,$B5,$FE,$7F,$E4,$FE
	.dc.b	$78,$FE,$2F,$CB,$CA,$C3,$A6,$3E,$0E,$5C,$C3,$CC,$A6,$22,$D7,$93
	.dc.b	$E4,$B6,$61,$F2,$BD,$B1,$FF,$16,$47,$E6,$FD,$DD,$BF,$C4,$F8,$BF
	.dc.b	$D5,$CD,$E4,$F2,$C1,$BC,$D8,$7A,$79,$32,$B8,$1F,$C7,$7E,$AF,$E2
	.dc.b	$37,$83,$DF,$1E,$EE,$1F,$E6,$6E,$A7,$FD,$3C,$E5,$FF,$53,$7D,$5B
	.dc.b	$E0,$BE,$3B,$63,$A1,$FC,$AF,$EA,$FF,$CD,$4D,$C7,$E6,$C7,$F4,$A8
	.dc.b	$D7,$8F,$89,$CF,$9B,$27,$81,$CF,$E2,$78,$F3,$C8,$C1,$DD,$EB,$F2
	.dc.b	$B2,$30,$78,$FC,$1F,$0F,$1C,$8C,$1E,$7D,$FF,$03,$F5,$4F,$AB,$A7
	.dc.b	$A2,$33,$E2,$EB,$C5,$97,$FD,$26,$FF,$C3,$2F,$F4,$BF,$59,$EC,$FE
	.dc.b	$31,$7C,$A5,$31,$19,$F8,$1D,$7D,$5C,$BE,$51,$C7,$C0,$E5,$DF,$27
	.dc.b	$B4,$7E,$63,$EA,$DC,$57,$8A,$57,$F5,$D9,$CB,$FD,$7F,$EA,$FF,$C6
	.dc.b	$6D,$E0,$FD,$0C,$7A,$A3,$FD,$BC,$E9,$FF,$70,$7D,$5F,$F8,$CD,$FE
	.dc.b	$D0,$FB,$9B,$AF,$EA,$A3,$51,$C3,$C7,$C7,$8F,$23,$8C,$6F,$C5,$E1
	.dc.b	$AC,$9E,$B1,$FB,$43,$F6,$EF,$FB,$C5,$F5,$B7,$5D,$74,$6F,$E1,$B8
	.dc.b	$7C,$1C,$AF,$EF,$D6,$9F,$DF,$FD,$5B,$D5,$BE,$AC,$BF,$F0,$24,$E5
	.dc.b	$BE,$47,$0F,$22,$46,$0F,$F0,$57,$D5,$FF,$9C,$5B,$F0,$DF,$C3,$20
	.dc.b	$DF,$87,$C7,$C0,$E7,$8A,$5F,$28,$FE,$27,$F5,$7F,$A7,$87,$A3,$FE
	.dc.b	$6E,$87,$8F,$A3,$7E,$47,$0F,$73,$8F,$87,$30,$FF,$0F,$5A,$7F,$8F
	.dc.b	$FC,$DF,$E5,$27,$FC,$44,$BE,$1B,$26,$BC,$5B,$3A,$F1,$38,$98,$7C
	.dc.b	$BE,$47,$F8,$F2,$D3,$FC,$B3,$E6,$1F,$1D,$4E,$37,$E0,$F5,$7C,$7C
	.dc.b	$72,$30,$79,$1F,$E6,$BF,$1F,$F9,$3E,$2F,$F4,$7B,$73,$F0,$4F,$B3
	.dc.b	$93,$DB,$E1,$4C,$3D,$B1,$5F,$38,$1F,$54,$FA,$BF,$E8,$6D,$8F,$8B
	.dc.b	$5F,$23,$C4,$F9,$3E,$5C,$9F,$93,$15,$EA,$38,$5F,$57,$F4,$0D,$6D
	.dc.b	$FE,$A8,$8C,$CF,$F9,$4F,$1F,$F0,$AB,$7C,$3E,$0F,$F9,$5F,$EA,$5F
	.dc.b	$39,$FD,$61,$FE,$5F,$1E,$06,$E1,$FE,$B4,$77,$1C,$7D,$5F,$EB,$FF
	.dc.b	$4E,$F9,$5A,$73,$8E,$9E,$BF,$5F,$2E,$5F,$86,$2B,$EE,$FA,$A7,$D5
	.dc.b	$FF,$6A,$6C,$79,$EB,$CD,$EC,$F9,$FA,$E4,$FA,$E2,$BE,$9F,$A4,$F6
	.dc.b	$18,$DB,$D4,$FF,$76,$FA,$BF,$CB,$4F,$F9,$E5,$AF,$13,$E3,$78,$5F
	.dc.b	$17,$C4,$91,$83,$E5,$3F,$E0,$0F,$AB,$FE,$81,$1F,$E8,$36,$BF,$2E
	.dc.b	$3E,$77,$A3,$F3,$FB,$E4,$FD,$38,$AF,$E0,$38,$DF,$2B,$F3,$8C,$E9
	.dc.b	$FE,$E6,$FA,$BF,$ED,$15,$F8,$65,$E1,$7F,$84,$9B,$1F,$E9,$57,$73
	.dc.b	$C6,$75,$F1,$F2,$B5,$15,$FA,$9F,$37,$F4,$CC,$57,$87,$D3,$ED,$F1
	.dc.b	$F2,$BB,$45,$7F,$33,$DC,$F1,$FF,$46,$66,$FD,$5F,$01,$E7,$FA,$B9
	.dc.b	$7E,$78,$AF,$E8,$B3,$FE,$99,$5F,$B8,$F7,$72,$B8,$BF,$D8,$7F,$57
	.dc.b	$FE,$20,$BF,$D4,$BF,$E1,$F7,$FD,$6C,$D7,$3E,$BF,$F8,$E2,$31,$5C
	.dc.b	$E7,$F7,$9F,$4F,$E9,$7F,$D7,$CD,$7A,$FB,$1F,$E4,$0F,$AB,$FE,$AF
	.dc.b	$5C,$65,$1F,$D7,$5F,$DD,$B6,$34,$F0,$C7,$FD,$B6,$E5,$FF,$6E,$DB
	.dc.b	$CB,$FF,$26,$B6,$FF,$DA,$15,$B7,$F8,$93,$E5,$BE,$BE,$7F,$94,$F3
	.dc.b	$70,$5A,$F5,$C7,$43,$0F,$9B,$E2,$BD,$9E,$54,$8E,$F1,$5F,$E2,$BE
	.dc.b	$9F,$F6,$37,$FD,$C4,$D7,$DF,$3F,$17,$CE,$F8,$DA,$49,$F4,$C5,$7F
	.dc.b	$8F,$FD,$4F,$FA,$41,$FF,$76,$35,$F9,$33,$FE,$6E,$FA,$BF,$EB,$9A
	.dc.b	$7C,$F7,$FA,$DC,$FF,$BC,$DA,$FF,$2A,$77,$BF,$F3,$78,$46,$A7,$EC
	.dc.b	$64,$EF,$FF,$3D,$B6,$BA,$7B,$3F,$E7,$EF,$DA,$3E,$94,$73,$3F,$A1
	.dc.b	$38,$70,$D7,$C0,$D7,$7E,$97,$FA,$AA,$3F,$9C,$C7,$30,$E1,$E7,$1B
	.dc.b	$FA,$D9,$1B,$36,$E1,$E5,$97,$F3,$57,$FA,$B6,$7E,$78,$C4,$7E,$5C
	.dc.b	$F8,$7E,$87,$C5,$9D,$36,$3E,$44,$BE,$57,$11,$FD,$86,$30,$F6,$37
	.dc.b	$1A,$7E,$7B,$71,$7C,$CD,$24,$F1,$3E,$5B,$C7,$4F,$EA,$31,$E7,$6E
	.dc.b	$3F,$45,$B1,$F7,$5A,$F9,$7C,$5E,$EF,$C4,$6B,$F0,$FD,$57,$CA,$FC
	.dc.b	$BC,$FF,$98,$E3,$7E,$13,$F8,$26,$5F,$25,$B8,$7C,$46,$BF,$0D,$AF
	.dc.b	$E3,$B8,$6D,$FE,$36,$BD,$D3,$96,$C6,$63,$EA,$3F,$C7,$63,$27,$22
	.dc.b	$1C,$BB,$BF,$D2,$0D,$F8,$DF,$0C,$62,$BE,$F3,$67,$F3,$0B,$E9,$B6
	.dc.b	$48,$E8,$D9,$E2,$D8,$E2,$FB,$B6,$7A,$34,$93,$EE,$8C,$B7,$D7,$E0
	.dc.b	$47,$07,$AF,$CB,$FC,$A6,$1F,$E2,$DC,$51,$88,$46,$27,$F4,$4D,$8F
	.dc.b	$D5,$2F,$9E,$E0,$D9,$DC,$60,$E9,$F5,$BF,$E3,$98,$FD,$63,$6B,$F6
	.dc.b	$8D,$7F,$80,$35,$FC,$87,$A5,$E5,$CB,$FC,$CD,$5F,$B5,$6C,$FE,$F5
	.dc.b	$AF,$E6,$7D,$1E,$04,$9E,$0F,$BF,$BB,$FD,$D1,$1F,$BF,$6C,$FF,$07
	.dc.b	$6B,$FC,$4D,$AF,$F1,$F7,$FC,$0B,$1C,$7C,$9E,$84,$7A,$7C,$47,$72
	.dc.b	$38,$25,$7F,$36,$8F,$EC,$DD,$FF,$D7,$ED,$E7,$62,$FC,$15,$6F,$FC
	.dc.b	$CD,$4F,$3F,$10,$8E,$35,$FD,$02,$73,$18,$9F,$F3,$1E,$FF,$E4,$6A
	.dc.b	$FF,$38,$63,$FD,$BE,$31,$3A,$F9,$BE,$B1,$89,$D4,$60,$EB,$16,$B9
	.dc.b	$9F,$D1,$57,$A3,$C8,$E7,$F3,$76,$77,$87,$E5,$63,$31,$FE,$06,$F8
	.dc.b	$DB,$3E,$8B,$F1,$9E,$4F,$3A,$D7,$29,$FC,$DC,$66,$3F,$C1,$FF,$11
	.dc.b	$67,$FD,$07,$F5,$BF,$EA,$D9,$E0,$D8,$CF,$C3,$7C,$C8,$C6,$BF,$E1
	.dc.b	$37,$FD,$57,$5C,$27,$E8,$FA,$3F,$CC,$A3,$07,$4F,$9E,$E4,$73,$C6
	.dc.b	$C7,$81,$C3,$E9,$F8,$77,$D6,$E3,$3F,$3F,$A3,$E3,$3C,$9E,$8D,$BF
	.dc.b	$4B,$87,$97,$97,$6B,$FD,$43,$E3,$79,$5B,$E3,$FE,$23,$EA,$BF,$D3
	.dc.b	$37,$F2,$6B,$5C,$3F,$E3,$19,$CD,$FF,$AA,$9B,$1F,$E3,$5F,$43,$E1
	.dc.b	$CE,$7E,$1B,$A6,$B3,$1F,$A9,$83,$E1,$71,$7C,$DF,$11,$F9,$AF,$AD
	.dc.b	$7B,$78,$BE,$3F,$2B,$E7,$3D,$1E,$55,$8E,$AE,$1E,$FF,$09,$71,$F4
	.dc.b	$3E,$F3,$F4,$FF,$54,$F3,$31,$5E,$6E,$3F,$88,$7C,$5F,$26,$C7,$A9
	.dc.b	$C3,$F8,$FD,$1F,$5B,$8C,$BE,$37,$27,$CD,$3E,$77,$A3,$6F,$E1,$B8
	.dc.b	$7F,$31,$F5,$BF,$31,$FF,$16,$71,$7F,$59,$E6,$FF,$31,$7D,$6F,$F4
	.dc.b	$D9,$FC,$F3,$71,$FF,$6E,$BF,$C6,$A7,$7E,$8F,$F0,$48,$C5,$7F,$8D
	.dc.b	$DB,$5F,$EE,$F7,$F8,$DB,$CB,$E7,$B8,$79,$B2,$BF,$59,$C5,$FD,$42
	.dc.b	$BB,$CF,$5F,$32,$57,$93,$CD,$B5,$E4,$7C,$47,$D0,$E5,$B3,$EA,$F8
	.dc.b	$2F,$31,$E7,$E9,$B1,$EE,$70,$F8,$BD,$AB,$87,$C0,$FA,$0F,$71,$EF
	.dc.b	$F3,$6D,$7C,$67,$0F,$C8,$F3,$6D,$7C,$87,$E3,$1F,$84,$F4,$6C,$FC
	.dc.b	$EF,$63,$E5,$1F,$2E,$AC,$7D,$27,$0F,$C6,$F6,$AC,$39,$7F,$D3,$5F
	.dc.b	$5B,$FC,$2A,$BF,$CD,$0D,$90,$7F,$57,$3C,$3E,$4F,$F9,$EC,$62,$B8
	.dc.b	$F9,$DF,$9A,$EC,$CB,$AF,$83,$93,$FC,$8E,$9F,$07,$8C,$BC,$93,$E2
	.dc.b	$8E,$A7,$6F,$96,$B1,$C5,$C3,$F8,$22,$CF,$23,$D3,$F5,$BF,$97,$8F
	.dc.b	$F4,$23,$63,$31,$FD,$5A,$31,$1F,$E4,$6F,$5F,$FA,$CE,$E6,$F3,$61
	.dc.b	$FE,$BB,$F5,$8F,$AD,$A6,$3D,$9B,$97,$F4,$19,$F9,$98,$9A,$7C,$97
	.dc.b	$E7,$B8,$F4,$7A,$9D,$BE,$1E,$D7,$F4,$BF,$A7,$7C,$97,$F6,$E6,$C6
	.dc.b	$7E,$77,$13,$9E,$99,$3C,$7A,$6E,$7F,$1C,$7D,$6B,$C2,$FE,$F2,$DA
	.dc.b	$FE,$B4,$FF,$77,$EB,$FF,$71,$39,$AE,$4F,$F8,$E2,$77,$EF,$72,$F7
	.dc.b	$B9,$BD,$2A,$4F,$F6,$53,$E5,$1D,$9F,$F0,$4A,$F6,$E9,$8F,$9A,$B1
	.dc.b	$FD,$B7,$EB,$5F,$3B,$FE,$10,$6C,$67,$DB,$C4,$FF,$85,$3E,$B7,$FC
	.dc.b	$8D,$8F,$C1,$6E,$BF,$EF,$D7,$D8,$DF,$87,$E1,$E2,$CC,$7F,$BE,$57
	.dc.b	$BE,$F9,$4F,$F8,$83,$EB,$5E,$1D,$31,$E1,$5C,$3F,$C0,$92,$F8,$2E
	.dc.b	$07,$1E,$0B,$3F,$E0,$9F,$80,$FE,$88,$DF,$6F,$FC,$5B,$1E,$EC,$FE
	.dc.b	$D3,$A6,$FB,$7A,$37,$ED,$DB,$FF,$0A,$7D,$6F,$F9,$AF,$85,$1F,$EE
	.dc.b	$98,$C3,$78,$7F,$C7,$71,$8C,$3F,$82,$BF,$1A,$78,$74,$EB,$D2,$DB
	.dc.b	$63,$BC,$72,$F8,$7C,$5F,$E4,$6F,$F5,$D5,$E4,$F1,$BB,$F8,$FE,$06
	.dc.b	$C7,$A2,$3C,$7E,$96,$2F,$97,$6F,$98,$F3,$F9,$36,$FD,$EE,$1F,$F1
	.dc.b	$A7,$3B,$AF,$17,$FA,$3E,$07,$C1,$FF,$96,$9B,$7F,$E3,$B5,$E3,$E6
	.dc.b	$74,$C9,$EB,$F0,$16,$FF,$5C,$7F,$94,$23,$FC,$88,$D3,$3D,$9F,$E6
	.dc.b	$87,$18,$7F,$BE,$63,$06,$7C,$1B,$5C,$9C,$3E,$CE,$7B,$7E,$0D,$BA
	.dc.b	$1D,$76,$B5,$E2,$70,$F9,$25,$E3,$FF,$42,$FA,$D7,$A9,$FD,$12,$7F
	.dc.b	$A2,$DF,$9D,$D9,$92,$55,$69,$4F,$E9,$AD,$FA,$CF,$DF,$7D,$6D,$EF
	.dc.b	$34,$95,$CE,$5E,$37,$F5,$B8,$C3,$5F,$14,$6B,$83,$B4,$E6,$7D,$52
	.dc.b	$FA,$1F,$65,$7E,$8A,$7F,$B0,$CF,$E0,$9F,$F3,$D3,$5F,$E5,$0F,$17
	.dc.b	$3C,$9C,$B6,$F8,$DC,$DF,$CD,$A7,$F1,$FE,$5E,$5E,$B5,$55,$FE,$31
	.dc.b	$EF,$E6,$E6,$0D,$E7,$C3,$9F,$95,$95,$C0,$F3,$BD,$71,$FF,$7C,$C7
	.dc.b	$E8,$3E,$17,$6F,$F4,$1E,$3F,$F7,$7B,$FE,$97,$5C,$BE,$56,$C6,$79
	.dc.b	$F1,$3D,$7C,$A9,$3C,$63,$D9,$F5,$BF,$16,$3F,$A7,$3F,$27,$87,$E7
	.dc.b	$42,$31,$C3,$F8,$D7,$84,$E9,$CC,$FF,$85,$3E,$B5,$F0,$1F,$25,$B1
	.dc.b	$FE,$AF,$9D,$AF,$AD,$FF,$36,$3F,$EB,$37,$FA,$3B,$72,$FA,$51,$A8
	.dc.b	$FE,$C8,$E9,$BD,$FA,$37,$3C,$3E,$27,$8F,$34,$AE,$D1,$5E,$7E,$A7
	.dc.b	$63,$1A,$7C,$2C,$5E,$0F,$8B,$8A,$5F,$9E,$3C,$BB,$FE,$1F,$EB,$5F
	.dc.b	$29,$4C,$D7,$FB,$3E,$1E,$2E,$FC,$59,$39,$8F,$C8,$7A,$CF,$62,$FC
	.dc.b	$94,$AF,$F6,$C4,$E5,$BE,$F7,$0F,$AD,$91,$83,$7D,$DD,$3B,$A4,$60
	.dc.b	$FF,$71,$7D,$6B,$93,$F4,$0F,$C0,$8C,$F0,$E2,$7F,$49,$F5,$BF,$E3
	.dc.b	$19,$FF,$75,$3E,$C6,$ED,$AF,$37,$F5,$5F,$5B,$FE,$33,$9F,$F7,$83
	.dc.b	$E9,$7F,$C3,$CD,$E7,$FC,$8B,$B8,$CF,$8F,$CB,$C7,$91,$CA,$3A,$F8
	.dc.b	$BD,$B7,$93,$DA,$3F,$74,$7E,$FD,$FF,$16,$3E,$F7,$FC,$D7,$87,$E8
	.dc.b	$23,$11,$AE,$CE,$3F,$07,$CB,$E3,$A4,$F0,$EB,$FE,$61,$6A,$FE,$11
	.dc.b	$F5,$BF,$E3,$7A,$D1,$FE,$78,$5E,$A6,$8D,$FA,$BC,$3C,$8E,$3E,$44
	.dc.b	$BE,$22,$FF,$81,$BF,$E4,$5F,$AD,$FF,$3B,$BA,$5F,$0B,$FC,$65,$B1
	.dc.b	$9F,$7B,$1E,$8E,$78,$E4,$F1,$8F,$E4,$1F,$5B,$FA,$DE,$3F,$E0,$6F
	.dc.b	$F7,$39,$F1,$B5,$FF,$2B,$3B,$E1,$FE,$A4,$BD,$BE,$7E,$0C,$AE,$31
	.dc.b	$FC,$B7,$E7,$5E,$1D,$F8,$99,$5F,$02,$72,$D8,$E1,$E2,$48,$C1,$B7
	.dc.b	$F9,$C7,$CB,$BE,$36,$9B,$9F,$F3,$65,$FF,$90,$9E,$DC,$1E,$33,$C9
	.dc.b	$EA,$58,$FA,$3F,$1D,$F9,$F7,$FC,$E8,$DE,$0F,$E4,$F7,$F4,$A7,$AF
	.dc.b	$17,$B7,$BE,$47,$AE,$2B,$FE,$7D,$FA,$47,$D6,$FF,$4F,$6C,$7C,$1A
	.dc.b	$F8,$DE,$0F,$C7,$F2,$64,$FC,$58,$AF,$B3,$E8,$BE,$61,$8C,$FC,$AF
	.dc.b	$5F,$06,$4F,$CD,$8F,$EB,$DF,$5A,$F9,$BB,$F0,$F2,$BF,$23,$39,$6C
	.dc.b	$7F,$65,$FA,$87,$CA,$53,$84,$FF,$A6,$69,$EB,$F6,$F2,$64,$F4,$8A
	.dc.b	$FF,$A7,$FE,$B1,$F5,$BF,$DC,$5B,$1E,$2A,$F3,$FB,$1E,$8E,$A9,$3E
	.dc.b	$68,$AF,$87,$EB,$7D,$A6,$38,$7F,$AE,$E6,$8F,$EF,$DF,$5A,$EF,$A7
	.dc.b	$0B,$F8,$3E,$0F,$C6,$F0,$A4,$60,$F8,$AF,$F8,$1F,$EB,$7F,$A0,$C7
	.dc.b	$EE,$1A,$FC,$38,$F9,$FE,$7F,$D0,$F1,$A4,$FC,$E8,$AF,$FB,$07,$91
	.dc.b	$F2,$9F,$3C,$CE,$DF,$DD,$BE,$B7,$FB,$6D,$7F,$B1,$D7,$85,$FE,$19
	.dc.b	$6C,$7F,$0A,$77,$3C,$67,$5F,$4F,$2B,$51,$5F,$A9,$F3,$7F,$4C,$C5
	.dc.b	$78,$7D,$2E,$DF,$21,$2B,$B4,$57,$F3,$DE,$3F,$E7,$4C,$57,$A7,$DE
	.dc.b	$F3,$7D,$6C,$AF,$34,$57,$F5,$FE,$CF,$F8,$CF,$3F,$E9,$D1,$EF,$AF
	.dc.b	$F1,$B7,$D6,$FF,$24,$C3,$FC,$32,$FF,$89,$27,$FD,$D4,$D8,$CF,$AD
	.dc.b	$C3,$D6,$FF,$1F,$7D,$37,$F0,$3B,$FF,$78,$36,$B9,$CF,$F9,$0F,$EB
	.dc.b	$7F,$AA,$47,$F8,$99,$FE,$B1,$CD,$FE,$71,$6D,$83,$C2,$35,$F0,$5E
	.dc.b	$0F,$2E,$57,$F3,$EA,$E2,$F0,$E0,$F0,$7A,$32,$35,$15,$F3,$C7,$A8
	.dc.b	$C6,$3E,$5F,$8D,$F4,$79,$92,$BB,$45,$53,$FE,$40,$F5,$7F,$AD,$47
	.dc.b	$F5,$06,$BA,$D3,$DD,$7F,$0F,$CE,$F8,$9A,$4A,$F5,$45,$78,$1F,$0F
	.dc.b	$EA,$7F,$D2,$EF,$F5,$C6,$BA,$3E,$3C,$FF,$9C,$7E,$AF,$FA,$64,$7C
	.dc.b	$B7,$FA,$86,$5F,$E5,$66,$E3,$FE,$A1,$77,$27,$2F,$D3,$C9,$DF,$FE
	.dc.b	$7D,$6C,$7F,$6E,$9D,$AF,$AC,$FF,$40,$3F,$EA,$76,$BF,$EA,$B6,$BF
	.dc.b	$EB,$16,$BF,$81,$7F,$D0,$AF,$FA,$B5,$AF,$3C,$FF,$D6,$0F,$76,$C3
	.dc.b	$D1,$A4,$9E,$4F,$0A,$FE,$35,$1F,$8F,$C7,$FD,$7D,$18,$8F,$F3,$D9
	.dc.b	$F0,$FC,$7F,$F6,$34,$E9,$B1,$F2,$7C,$FA,$FC,$F6,$BF,$2D,$B5,$FD
	.dc.b	$02,$3C,$07,$D1,$F9,$5C,$91,$C9,$F6,$5E,$78,$BE,$66,$92,$76,$7D
	.dc.b	$17,$BF,$8D,$23,$36,$8C,$47,$BA,$3F,$93,$CF,$FB,$6D,$B1,$FE,$DF
	.dc.b	$5E,$C3,$E5,$75,$E3,$F8,$EB,$5F,$05,$FE,$CF,$DB,$E5,$B6,$3F,$1C
	.dc.b	$D7,$F2,$8D,$7F,$C3,$7F,$0D,$78,$BF,$CD,$95,$EC,$6C,$FD,$46,$C7
	.dc.b	$D8,$39,$64,$69,$0F,$8B,$DF,$FE,$8E,$6F,$C5,$F8,$63,$17,$F8,$06
	.dc.b	$CF,$FC,$01,$6C,$7F,$15,$0E,$1F,$E3,$0D,$73,$3D,$7E,$96,$57,$89
	.dc.b	$B3,$EB,$F4,$BE,$01,$27,$E0,$3E,$5F,$16,$47,$A9,$B3,$E7,$F2,$FF
	.dc.b	$2E,$F8,$9A,$F9,$9B,$89,$BF,$35,$2F,$D9,$5E,$DA,$4E,$63,$27,$F9
	.dc.b	$43,$9F,$AE,$BF,$E0,$3B,$CC,$E5,$B1,$98,$E2,$FF,$97,$63,$3E,$87
	.dc.b	$22,$3C,$39,$7A,$91,$C1,$DB,$BB,$FD,$CE,$DF,$93,$F7,$C6,$30,$F3
	.dc.b	$B5,$F6,$AF,$87,$FC,$34,$7F,$29,$C7,$FC,$0F,$7F,$CB,$67,$3D,$FF
	.dc.b	$CF,$79,$AC,$FE,$07,$0E,$4F,$8E,$BC,$24,$7C,$DE,$6F,$90,$8E,$0F
	.dc.b	$8F,$BF,$F8,$E1,$FF,$4B,$B7,$2F,$E7,$8B,$E2,$7C,$31,$F8,$EF,$6A
	.dc.b	$BF,$92,$8B,$3B,$AF,$E8,$CE,$23,$3E,$8C,$9E,$FF,$33,$6F,$BB,$C8
	.dc.b	$EB,$E6,$FF,$BA,$E7,$67,$9B,$0E,$E7,$83,$CD,$B1,$EA,$70,$FF,$AC
	.dc.b	$A5,$F0,$D8,$C3,$FA,$C7,$D5,$FF,$92,$6B,$FD,$6E,$D8,$CF,$C3,$7B
	.dc.b	$A3,$1C,$5F,$62,$50,$9D,$1E,$8F,$0E,$DF,$4F,$9C,$E4,$73,$C6,$C7
	.dc.b	$81,$C3,$FE,$C8,$F8,$77,$D5,$E3,$2F,$3F,$9F,$E3,$3C,$9E,$7D,$BF
	.dc.b	$4B,$87,$97,$95,$6B,$FB,$B7,$8F,$FC,$E6,$7F,$DA,$AD,$8F,$EF,$5F
	.dc.b	$51,$FE,$8B,$9C,$EB,$F7,$A3,$13,$AF,$85,$FC,$0F,$CB,$7C,$1B,$FF
	.dc.b	$6F,$D6,$B8,$BE,$33,$EA,$8C,$EB,$CF,$E1,$7A,$36,$38,$C7,$E3,$3E
	.dc.b	$AD,$EE,$67,$3F,$2F,$C9,$78,$CF,$27,$95,$70,$F0,$B8,$7C,$7F,$0B
	.dc.b	$6B,$DA,$FA,$CF,$CC,$FD,$43,$E2,$71,$5E,$AE,$3F,$80,$7C,$1F,$26
	.dc.b	$C7,$B9,$C3,$FE,$FE,$D1,$F5,$78,$CB,$E5,$F2,$FC,$93,$E5,$7B,$36
	.dc.b	$FE,$3B,$87,$FC,$A1,$C7,$F8,$DF,$AB,$F9,$51,$FC,$FF,$93,$FC,$42
	.dc.b	$E8,$FE,$85,$F5,$6F,$4B,$F6,$6D,$8D,$65,$FE,$5B,$8F,$E2,$98,$7E
	.dc.b	$65,$D4,$7F,$8E,$5C,$CF,$4F,$37,$9F,$A5,$D1,$B1,$D2,$95,$FE,$98
	.dc.b	$8C,$71,$39,$37,$5D,$7C,$BF,$29,$E6,$F2,$ED,$F3,$C1,$19,$8E,$7C
	.dc.b	$D6,$FC,$DF,$07,$EE,$3D,$FD,$36,$B9,$38,$7C,$7E,$D5,$AF,$77,$CF
	.dc.b	$7C,$63,$E3,$F9,$96,$BC,$4E,$1F,$27,$99,$6B,$E3,$3F,$38,$FC,$4F
	.dc.b	$A5,$67,$D5,$EC,$FD,$23,$E9,$D5,$8F,$73,$87,$E3,$FA,$56,$3F,$CD
	.dc.b	$5F,$57,$FC,$3A,$7F,$BF,$36,$3F,$D2,$8F,$EB,$DD,$FC,$57,$F3,$28
	.dc.b	$C7,$3F,$F1,$56,$D7,$F8,$1F,$95,$D9,$BF,$37,$87,$C1,$4A,$FB,$96
	.dc.b	$F8,$2C,$67,$E4,$7B,$1D,$FE,$56,$D7,$07,$0F,$E0,$8B,$3F,$E8,$5F
	.dc.b	$AB,$FC,$CB,$7C,$1F,$E9,$31,$FE,$27,$9F,$C6,$3A,$D7,$F9,$77,$07
	.dc.b	$F4,$BB,$FF,$47,$36,$33,$CF,$FD,$4A,$31,$7F,$E2,$5F,$AA,$7D,$5F
	.dc.b	$F5,$66,$C7,$E6,$5F,$E9,$98,$7F,$A6,$5C,$B7,$8B,$81,$C7,$82,$D7
	.dc.b	$1F,$3F,$A1,$D7,$E1,$6C,$7E,$97,$E9,$5F,$25,$FD,$99,$B1,$9F,$84
	.dc.b	$C4,$E7,$A6,$4F,$1E,$9B,$A7,$FA,$B7,$EA,$FF,$BF,$B7,$C3,$FE,$E3
	.dc.b	$1F,$A8,$AF,$DA,$3A,$E3,$FE,$AC,$FF,$91,$A3,$3E,$E7,$0F,$73,$8B
	.dc.b	$CB,$09,$3F,$BD,$3E,$1D,$F2,$8F,$EF,$F1,$FD,$A8,$D7,$B3,$8F,$98
	.dc.b	$CB,$FE,$04,$FA,$B7,$D1,$D3,$1E,$C5,$BF,$E1,$5F,$57,$FE,$3E,$6F
	.dc.b	$C3,$7F,$83,$23,$D5,$1F,$E5,$F7,$59,$7F,$B3,$D7,$8B,$B3,$86,$B6
	.dc.b	$BF,$8B,$7D,$5B,$C4,$FF,$0B,$B6,$33,$E1,$E2,$73,$F0,$12,$78,$EF
	.dc.b	$73,$FF,$6F,$7B,$DF,$D1,$E7,$F9,$13,$63,$3B,$FF,$89,$23,$1C,$5F
	.dc.b	$E0,$75,$EC,$E4,$D7,$B3,$70,$FE,$59,$F5,$7F,$E6,$56,$FC,$4F,$F8
	.dc.b	$BA,$3E,$0E,$1F,$E1,$17,$59,$7F,$BC,$5F,$8E,$83,$86,$FA,$5C,$FA
	.dc.b	$BA,$33,$0F,$0F,$CA,$5A,$E9,$F0,$DE,$13,$C5,$E4,$DA,$F3,$B8,$7D
	.dc.b	$3E,$9D,$AF,$27,$6F,$A4,$F5,$79,$56,$BD,$EE,$1F,$E9,$9C,$AE,$CC
	.dc.b	$5B,$7F,$F1,$80,$F8,$3F,$F2,$8B,$6F,$FA,$A2,$F1,$F3,$3A,$64,$F6
	.dc.b	$F8,$0B,$7D,$63,$FD,$23,$E9,$FF,$97,$20,$DE,$EC,$36,$7F,$B1,$4F
	.dc.b	$E4,$9B,$DD,$98,$E8,$7F,$A6,$E3,$27,$BB,$A3,$B4,$F7,$DE,$4F,$4D
	.dc.b	$ED,$F2,$3E,$E3,$E5,$3C,$CB,$C8,$FC,$07,$D5,$BD,$5F,$F5,$5D,$6A
	.dc.b	$7F,$D4,$B5,$F8,$56,$BC,$4C,$BF,$EA,$AF,$AA,$FF,$33,$FD,$5D,$F0
	.dc.b	$0D,$25,$7F,$AD,$25,$E3,$FD,$78,$C3,$5F,$F5,$EC,$6B,$93,$B4,$E6
	.dc.b	$7F,$DB,$92,$F9,$EF,$F6,$F5,$7F,$9B,$E3,$F0,$13,$FE,$11,$6B,$F1
	.dc.b	$5E,$0E,$79,$78,$ED,$FE,$E7,$E7,$FE,$67,$3F,$8E,$F3,$32,$FE,$9E
	.dc.b	$F9,$E1,$DF,$9F,$32,$97,$9D,$86,$C7,$8E,$71,$39,$F9,$49,$3E,$78
	.dc.b	$AF,$DB,$C7,$FC,$4C,$FE,$83,$DD,$CE,$3F,$E8,$DE,$3F,$F7,$EC,$FF
	.dc.b	$8B,$D7,$2F,$91,$B1,$9E,$6C,$4F,$A3,$0F,$E6,$33,$FA,$CF,$1B,$2F
	.dc.b	$F8,$E7,$EA,$BE,$3B,$F8,$D7,$E0,$F0,$7F,$09,$84,$61,$FE,$18,$F0
	.dc.b	$CB,$FD,$44,$E6,$1F,$56,$F8,$0F,$92,$D8,$F0,$71,$FC,$CF,$AB,$FF
	.dc.b	$38,$47,$F9,$2D,$F3,$BF,$9F,$6E,$1F,$EA,$F7,$71,$9C,$B9,$78,$12
	.dc.b	$39,$47,$5F,$89,$ED,$CD,$27,$B4,$57,$B7,$AB,$E3,$31,$8F,$7F,$0B
	.dc.b	$CD,$C9,$2B,$CD,$1E,$9E,$0F,$C2,$FD,$5B,$E5,$29,$9A,$FF,$33,$43
	.dc.b	$C6,$DF,$8D,$27,$31,$F8,$EF,$5D,$F3,$57,$E5,$25,$7F,$9B,$E7,$2D
	.dc.b	$F6,$B8,$7D,$5C,$8C,$1B,$EE,$E9,$DF,$23,$07,$F9,$E7,$EA,$DC,$7F
	.dc.b	$9E,$5C,$34,$C4,$7E,$8B,$EA,$DE,$75,$DE,$61,$FE,$83,$9C,$BF,$E8
	.dc.b	$5F,$AB,$FF,$18,$B7,$83,$F5,$71,$EA,$8F,$E3,$8E,$9B,$E3,$EB,$C7
	.dc.b	$91,$CE,$38,$F8,$DC,$B7,$93,$DE,$3F,$6E,$7E,$F9,$FE,$50,$FB,$9F
	.dc.b	$F4,$C3,$5C,$F5,$F0,$F8,$6E,$3F,$09,$27,$71,$FC,$0B,$EA,$DE,$AD
	.dc.b	$F5,$65,$7F,$A8,$E7,$2D,$F2,$B8,$79,$52,$30,$7F,$AA,$7E,$AF,$FC
	.dc.b	$DC,$DF,$87,$FE,$1F,$06,$FB,$B8,$F4,$73,$C3,$2F,$94,$7F,$16,$FA
	.dc.b	$BF,$D3,$F8,$9F,$B9,$7F,$B5,$4F,$FA,$DD,$AF,$F4,$B7,$6F,$F8,$1D
	.dc.b	$7B,$9C,$FC,$19,$3C,$63,$F9,$1F,$CF,$3C,$3B,$F0,$D2,$BF,$D8,$93
	.dc.b	$96,$C7,$0F,$12,$46,$0D,$BF,$CB,$BE,$61,$F1,$D4,$DC,$FF,$61,$BF
	.dc.b	$F1,$D3,$DB,$7F,$19,$E4,$F5,$2C,$7F,$B5,$7E,$43,$F2,$6F,$F6,$96
	.dc.b	$F9,$AE,$0F,$E3,$D8,$7D,$2A,$ED,$C5,$EE,$EF,$91,$EC,$8A,$FF,$73
	.dc.b	$FA,$A7,$D5,$FF,$46,$6C,$7C,$2A,$F8,$FE,$0F,$C8,$F2,$A4,$FC,$68
	.dc.b	$AF,$B7,$E8,$BE,$69,$8D,$3C,$D1,$FE,$3B,$CD,$3D,$FC,$39,$3F,$42
	.dc.b	$3F,$AB,$FD,$53,$E7,$2F,$CD,$CA,$FF,$7C,$4E,$5B,$1F,$D7,$BE,$A1
	.dc.b	$F2,$D4,$E1,$3F,$E1,$1A,$7A,$DD,$BC,$99,$3D,$22,$BF,$E1,$7F,$AB
	.dc.b	$7D,$0F,$F6,$A6,$C7,$8A,$BC,$FE,$C7,$A3,$AE,$4F,$9A,$2B,$E1,$FA
	.dc.b	$3F,$69,$8E,$2F,$8B,$0C,$D3,$FD,$E3,$EA,$DE,$05,$38,$D7,$C2,$F0
	.dc.b	$BE,$3F,$87,$23,$07,$C6,$7F,$C0,$3F,$57,$FC,$F6,$3F,$CD,$99,$78
	.dc.b	$23,$11,$F1,$67,$E8,$F9,$FF,$4B,$BE,$4F,$D0,$8A,$FF,$8E,$F9,$1F
	.dc.b	$2D,$F4,$4C,$E3,$EC,$FA,$BF,$ED,$15,$FE,$41,$5E,$17,$F8,$51,$B1
	.dc.b	$E4,$C7,$FB,$C4,$70,$AE,$55,$BF,$90,$97,$B3,$FD,$0D,$5C,$9E,$D1
	.dc.b	$F5,$FE,$36,$46,$C4,$A7,$C5,$E8,$F8,$FE,$3E,$56,$E9,$38,$8F,$D4
	.dc.b	$78,$FF,$A7,$31,$97,$BF,$DE,$F6,$7D,$5C,$AE,$51,$5F,$DD,$7B,$5F
	.dc.b	$E3,$4D,$7F,$B6,$47,$7A,$FF,$1B,$FD,$5F,$F8,$A2,$3F,$CD,$0F,$F8
	.dc.b	$99,$BE,$A6,$BD,$5D,$BF,$E3,$EF,$A7,$FE,$1B,$3F,$E6,$F6,$E7,$F8
	.dc.b	$EB,$E4,$FE,$43,$EA,$FF,$AC,$37,$DC,$FF,$23,$C7,$F5,$D8,$D7,$CE
	.dc.b	$E2,$7B,$79,$B3,$1F,$99,$7B,$E6,$FF,$18,$3E,$06,$FA,$5D,$0E,$BC
	.dc.b	$56,$BE,$A6,$9F,$C6,$73,$FD,$1B,$6F,$8C,$F6,$F8,$9F,$2F,$9B,$23
	.dc.b	$07,$89,$FF,$2F,$FA,$CF,$76,$9B,$9F,$F4,$96,$9F,$E6,$06,$EB,$F0
	.dc.b	$0F,$83,$E8,$D8,$FC,$47,$D4,$FF,$A4,$DB,$F1,$D9,$7F,$D3,$9F,$55
	.dc.b	$FD,$49,$3F,$D5,$B7,$FD,$EB,$76,$FF,$2A,$3B,$31,$3F,$ED,$FD,$1E
	.dc.b	$17,$A8,$FF,$3C,$CE,$A5,$F5,$7F,$E6,$07,$FC,$BC,$63,$D6,$FF,$52
	.dc.b	$C6,$63,$FC,$C4,$D7,$FD,$D8,$B5,$D9,$23,$99,$A7,$9E,$B2,$32,$78
	.dc.b	$9C,$3C,$27,$FD,$37,$97,$91,$B1,$E2,$E1,$FF,$3A,$4E,$5B,$1F,$13
	.dc.b	$CA,$AF,$FC,$01,$CB,$F9,$1A,$3F,$93,$9F,$3A,$36,$D3,$55,$F1,$27
	.dc.b	$FA,$1C,$60,$FA,$3E,$97,$08,$CC,$74,$6C,$F1,$F9,$6D,$12,$79,$1E
	.dc.b	$57,$97,$95,$23,$27,$E0,$9C,$3D,$7C,$9F,$01,$1D,$DA,$3C,$9E,$43
	.dc.b	$E6,$7E,$7C,$FF,$2E,$6F,$B0,$F9,$6C,$D6,$F0,$FB,$3C,$8F,$8F,$FE
	.dc.b	$E3,$AF,$E1,$A3,$0F,$99,$B3,$F3,$1B,$1B,$F8,$9E,$33,$FE,$AB,$9F
	.dc.b	$F3,$7C,$F1,$9C,$CF,$E6,$5A,$FE,$8A,$2E,$7F,$E1,$77,$DE,$60,$86
	.dc.b	$09,$1F,$4D,$E7,$E2,$7F,$A5,$5F,$9A,$D8,$FC,$1B,$63,$F4,$6D,$8F
	.dc.b	$F6,$12,$FA,$5E,$64,$9E,$6F,$6F,$93,$91,$D9,$B5,$DB,$C8,$F1,$92
	.dc.b	$7C,$6F,$9B,$CE,$FC,$AC,$FE,$5F,$0F,$E9,$D5,$FA,$86,$C7,$EB,$9A
	.dc.b	$FD,$5F,$6B,$8B,$FE,$55,$6E,$9F,$E1,$8A,$FD,$8B,$67,$EF,$F3,$68
	.dc.b	$93,$A7,$8F,$24,$8E,$2D,$9F,$C3,$F2,$76,$24,$F6,$7C,$3C,$32,$3C
	.dc.b	$2D,$9F,$E0,$FD,$FF,$E8,$C9,$FE,$DD,$8F,$F8,$11,$FE,$53,$F1,$1F
	.dc.b	$E4,$2E,$FF,$F0,$FC,$E6,$D1,$FE,$22,$BE,$51,$8B,$FF,$7D,$46,$27
	.dc.b	$5E,$77,$39,$CE,$16,$FE,$87,$B7,$56,$FF,$E3,$AA,$FF,$01,$3F,$EC
	.dc.b	$26,$CE,$E3,$13,$E1,$3E,$D7,$C5,$F3,$CB,$9D,$F2,$8E,$1E,$9C,$8E
	.dc.b	$BF,$33,$6F,$91,$E2,$7B,$1E,$5F,$27,$FD,$CF,$3C,$0F,$2E,$3E,$33
	.dc.b	$C9,$E7,$D8,$F4,$B8,$7F,$B4,$FC,$3D,$C3,$BF,$D5,$FF,$92,$27,$FC
	.dc.b	$40,$D8,$CF,$C3,$FC,$58,$C7,$0F,$F9,$49,$46,$4F,$37,$83,$6F,$9F
	.dc.b	$CE,$71,$39,$63,$63,$BB,$87,$FC,$61,$F1,$0F,$AB,$C6,$7E,$6F,$1F
	.dc.b	$C4,$78,$FD,$0B,$7E,$87,$0F,$2F,$1E,$D7,$CC,$F4,$3F,$93,$4F,$F8
	.dc.b	$F5,$B1,$F4,$3E,$A3,$FD,$15,$39,$D7,$F0,$91,$89,$D7,$C3,$7E,$1F
	.dc.b	$C8,$F8,$37,$FE,$47,$AD,$71,$7F,$9D,$9F,$54,$67,$5E,$7E,$17,$46
	.dc.b	$C7,$18,$FC,$97,$D5,$BD,$BC,$E7,$E5,$F9,$2F,$19,$E4,$F8,$4B,$87
	.dc.b	$85,$C3,$E3,$F8,$5B,$5E,$D7,$D6,$7E,$7F,$EA,$1F,$13,$8A,$F5,$71
	.dc.b	$FC,$03,$E0,$F9,$16,$3D,$CE,$1F,$F3,$36,$8F,$AB,$C6,$5F,$2F,$97
	.dc.b	$E4,$9F,$2B,$D8,$B7,$F1,$DC,$3F,$E6,$FF,$AB,$F9,$2F,$F7,$9A,$71
	.dc.b	$FF,$87,$A3,$FC,$E9,$F5,$7F,$CD,$5B,$E8,$FE,$E6,$1E,$8D,$8C,$32
	.dc.b	$78,$FA,$37,$0F,$D7,$DF,$FA,$3A,$72,$7F,$A0,$78,$5C,$7D,$6F,$B7
	.dc.b	$97,$2B,$C7,$E6,$5C,$79,$BE,$33,$E7,$72,$D9,$F4,$FC,$1F,$9C,$F4
	.dc.b	$74,$58,$F6,$B8,$7C,$5E,$BD,$C3,$CF,$EE,$7B,$CF,$81,$E6,$5A,$F8
	.dc.b	$AE,$1F,$8F,$E6,$5A,$F7,$BF,$1C,$FC,$17,$A5,$67,$E6,$FA,$FF,$2C
	.dc.b	$F9,$95,$63,$E8,$B8,$7E,$37,$A7,$63,$F8,$CF,$D5,$FF,$06,$9F,$B4
	.dc.b	$D8,$FF,$1A,$3F,$A8,$77,$F1,$3F,$CC,$63,$1E,$27,$FB,$79,$B5,$F8
	.dc.b	$3E,$5F,$F3,$B3,$FE,$B5,$8C,$DF,$4F,$80,$E4,$73,$F8,$1B,$5B,$70
	.dc.b	$F1,$F9,$3B,$5E,$87,$C4,$7E,$96,$29,$FE,$C3,$FA,$BF,$C9,$4F,$F3
	.dc.b	$66,$E5,$FE,$F4,$7F,$C3,$11,$9E,$FF,$E9,$9B,$7F,$46,$AF,$F7,$CB
	.dc.b	$6F,$F3,$CB,$97,$7E,$B7,$F5,$3F,$AB,$7C,$35,$31,$F0,$D7,$0F,$F6
	.dc.b	$B4,$FC,$DC,$CD,$3E,$2B,$8E,$2E,$9E,$84,$AE,$3F,$07,$6B,$FA,$8F
	.dc.b	$D2,$3E,$43,$FB,$23,$6B,$FD,$C8,$FF,$8F,$27,$73,$AE,$89,$5A,$E8
	.dc.b	$B5,$FD,$73,$EA,$FF,$BA,$4F,$EF,$5B,$1F,$E7,$77,$FD,$D3,$1F,$EF
	.dc.b	$17,$30,$7F,$B5,$BC,$FD,$CF,$EF,$10,$F7,$2C,$62,$65,$7F,$BE,$CF
	.dc.b	$96,$75,$7F,$C0,$0B,$D9,$A6,$3E,$5A,$C7,$FA,$2F,$98,$FF,$15,$FA
	.dc.b	$B7,$D0,$FF,$82,$5B,$1B,$35,$EC,$66,$7F,$C1,$DF,$54,$FA,$BA,$63
	.dc.b	$CA,$B7,$C7,$4F,$F4,$D5,$3C,$4E,$27,$2D,$AD,$6E,$9F,$56,$F0,$FF
	.dc.b	$C3,$8D,$9D,$2F,$0F,$33,$AF,$80,$95,$AE,$2B,$87,$F8,$53,$DE,$FE
	.dc.b	$8F,$3D,$DB,$19,$DF,$FC,$55,$18,$E2,$F8,$F3,$FE,$B3,$87,$B3,$A3
	.dc.b	$7E,$CD,$C3,$D5,$F5,$7F,$E6,$0C,$38,$E0,$8C,$4F,$F4,$57,$E5,$3C
	.dc.b	$3E,$27,$FC,$75,$1F,$88,$27,$A3,$8E,$4C,$B2,$BB,$FC,$A5,$C3,$FC
	.dc.b	$63,$7F,$B5,$7F,$D9,$4D,$7C,$5E,$3F,$6F,$1F,$BB,$2F,$D3,$E8,$DA
	.dc.b	$E7,$DB,$E4,$3C,$BE,$4D,$AF,$7B,$87,$ED,$72,$FF,$AB,$B7,$F8,$39
	.dc.b	$FF,$AE,$67,$EF,$43,$E0,$BF,$CB,$0D,$8C,$F8,$D8,$9C,$B2,$3A,$FB
	.dc.b	$F7,$3F,$F2,$78,$F4,$7F,$CC,$4D,$9F,$C6,$3F,$91,$6C,$6B,$B6,$46
	.dc.b	$BB,$6D,$74,$79,$9F,$E6,$DD,$3F,$03,$3B,$3A,$ED,$E0,$3C,$3B,$58
	.dc.b	$F0,$B8,$7C,$76,$F1,$FF,$DF,$DF,$56,$F4,$FF,$C1,$13,$F1,$CF,$69
	.dc.b	$B3,$27,$F8,$36,$6B,$4A,$6C,$E7,$F5,$5F,$C1,$7E,$AF,$6D,$8E,$6D
	.dc.b	$77,$DE,$9D,$F8,$FF,$E9,$89,$DB,$5E,$AD,$7F,$C4,$14,$8C,$40,$43
	.dc.b	$E4,$FF,$C4,$DC,$BF,$A6,$8F,$C2,$4F,$F0,$36,$BF,$E8,$C7,$83,$9E
	.dc.b	$5E,$3A,$FF,$8B,$F8,$3F,$A0,$4F,$E4,$BC,$CC,$BE,$1B,$E7,$87,$7E
	.dc.b	$6C,$CF,$93,$CD,$B4,$6C,$E1,$CF,$C9,$CA,$F3,$45,$7F,$D4,$38,$FF
	.dc.b	$8C,$1F,$D1,$FB,$79,$C7,$F8,$1F,$17,$FA,$DA,$7F,$90,$3F,$D0,$5B
	.dc.b	$CB,$E4,$8D,$47,$FA,$B1,$D3,$DB,$C5,$F4,$E1,$FC,$EA,$38,$9F,$CA
	.dc.b	$FE,$AF,$E3,$BF,$EB,$47,$E0,$EF,$FE,$55,$6C,$7D,$57,$78,$FF,$18
	.dc.b	$73,$0F,$AB,$FF,$2A,$BF,$CD,$DA,$F5,$E1,$F9,$7F,$57,$FE,$74,$8F
	.dc.b	$E7,$8F,$95,$FF,$62,$B6,$FF,$92,$BB,$8C,$E5,$CB,$CE,$91,$CA,$3A
	.dc.b	$FC,$47,$6E,$49,$3D,$A2,$BD,$BE,$53,$C6,$63,$1E,$FE,$07,$9B,$8A
	.dc.b	$57,$9A,$3D,$3B,$FE,$13,$EA,$9F,$57,$4C,$D7,$F5,$78,$FF,$10,$37
	.dc.b	$C4,$E0,$71,$F1,$7F,$1B,$EB,$3D,$9B,$F2,$92,$FF,$B0,$4E,$5B,$F2
	.dc.b	$1C,$3E,$AE,$46,$0D,$F6,$74,$ED,$91,$83,$FB,$57,$D5,$B8,$FF,$3A
	.dc.b	$B8,$69,$88,$FD,$0F,$D5,$BC,$DB,$BC,$C3,$FB,$A4,$E5,$FE,$ED,$F5
	.dc.b	$7F,$E3,$C6,$F0,$7E,$AA,$3D,$50,$E6,$FE,$C3,$1B,$8C,$72,$FF,$A9
	.dc.b	$63,$F8,$34,$74,$F1,$39,$1C,$FC,$5E,$9E,$27,$5E,$19,$5E,$28,$FD
	.dc.b	$C9,$FC,$01,$EA,$7F,$0F,$7E,$13,$FE,$0D,$6B,$A0,$F1,$C7,$FB,$BA
	.dc.b	$F8,$EE,$BF,$05,$2B,$94,$7F,$08,$FA,$B7,$A9,$7D,$49,$5F,$E1,$B9
	.dc.b	$CB,$7D,$0E,$1E,$54,$8C,$1F,$E2,$1F,$AB,$FF,$0E,$37,$E1,$BF,$8A
	.dc.b	$C1,$BE,$E6,$3D,$0C,$F0,$CB,$E5,$1F,$C7,$3E,$AF,$F5,$DC,$BF,$ED
	.dc.b	$35,$E4,$E4,$D7,$93,$BF,$6B,$87,$87,$31,$FF,$1A,$5A,$7F,$93,$FC
	.dc.b	$F3,$C5,$BF,$0B,$2F,$FC,$75,$39,$6C,$70,$F1,$A4,$60,$DB,$FC,$CB
	.dc.b	$E6,$1F,$1B,$4D,$CF,$D7,$BF,$F2,$23,$DB,$87,$C6,$79,$3D,$3B,$1F
	.dc.b	$E4,$BF,$8E,$FC,$C3,$F7,$DB,$DD,$FE,$34,$BF,$9F,$3D,$79,$3D,$BE
	.dc.b	$04,$8F,$5C,$57,$F1,$1F,$54,$FA,$BF,$E9,$2D,$8F,$83,$5F,$1B,$C2
	.dc.b	$F8,$FE,$5C,$9F,$8B,$15,$F6,$7D,$5F,$CC,$31,$9F,$95,$EB,$E2,$49
	.dc.b	$F9,$B1,$FD,$63,$EA,$9F,$39,$7E,$6E,$57,$F9,$AE,$72,$D8,$FE,$BF
	.dc.b	$F5,$0F,$95,$A7,$09,$FD,$25,$3D,$7E,$DE,$C4,$9E,$91,$5F,$D5,$7D
	.dc.b	$5B,$E8,$BF,$B5,$B6,$3C,$55,$E7,$F6,$BD,$1D,$52,$7C,$D1,$5F,$0F
	.dc.b	$D1,$FB,$4C,$70,$FF,$9A,$B3,$47,$F7,$8F,$AB,$77,$D3,$85,$FC,$1F
	.dc.b	$07,$E3,$78,$72,$30,$7C,$57,$FC,$03,$F5,$7F,$CF,$A3,$FD,$24,$D7
	.dc.b	$E1,$C7,$CF,$F4,$7E,$87,$91,$27,$E7,$45,$7F,$86,$72,$3E,$57,$E7
	.dc.b	$99,$DB,$FD,$07,$F5,$7F,$DA,$1B,$E1,$7F,$84,$A0,$DF,$9D,$C3,$9F
	.dc.b	$8F,$98,$FF,$A9,$EB,$6D,$8F,$A3,$F1,$B2,$30,$6F,$D4,$ED,$F1,$F2
	.dc.b	$30,$73,$7F,$37,$E4,$48,$C1,$DF,$E0,$7C,$DF,$57,$23,$07,$8D,$FF
	.dc.b	$15,$E7,$FD,$92,$3F,$61,$EE,$ED,$FE,$61,$F5,$7F,$E1,$CD,$FF,$D4
	.dc.b	$AF,$F8,$71,$BE,$BF,$F8,$D2,$35,$C9,$FE,$49,$A4,$6E,$31,$F4,$FF
	.dc.b	$C1,$27,$F9,$E3,$6F,$9C,$7F,$B3,$7E,$AF,$FA,$94,$7F,$AD,$1F,$EA
	.dc.b	$ED,$F7,$FF,$C8,$71,$AB,$F0,$3A,$7B,$F9,$7F,$E4,$8C,$0F,$0A,$E3
	.dc.b	$BF,$2F,$46,$65,$E1,$8A,$F9,$A3,$91,$8C,$BC,$9F,$13,$E2,$F3,$25
	.dc.b	$7A,$62,$BF,$D6,$7D,$5F,$EA,$CF,$FB,$A5,$AF,$F5,$DD,$7F,$AE,$BF
	.dc.b	$07,$CE,$F6,$E9,$23,$07,$C5,$7F,$CC,$7F,$53,$FE,$93,$8F,$F7,$A3
	.dc.b	$5E,$91,$FE,$68,$FA,$AF,$E9,$B1,$FE,$7D,$7F,$A6,$BF,$EF,$B6,$BF
	.dc.b	$DB,$DD,$A7,$39,$FF,$B9,$B1,$77,$FF,$9D,$1B,$5C,$FD,$AF,$07,$9B
	.dc.b	$FE,$9E,$C3,$F8,$B4,$FF,$A8,$9A,$FF,$A9,$9A,$FF,$AB,$3E,$B3,$FC
	.dc.b	$2A,$DE,$2F,$F5,$3C,$62,$3F,$1C,$D7,$FD,$31,$F5,$8F,$9A,$FF,$56
	.dc.b	$B6,$3F,$D6,$2D,$7F,$D6,$AD,$7F,$D7,$77,$FE,$83,$5C,$D8,$F3,$3F
	.dc.b	$38,$D8,$FF,$53,$35,$D9,$FA,$19,$FE,$CA,$E8,$C7,$9B,$C9,$A6,$4F
	.dc.b	$A8,$E5,$F5,$FB,$D5,$60,$8C,$36,$39,$1A,$8C,$FB,$32,$33,$3F,$E3
	.dc.b	$A7,$CE,$7C,$88,$ED,$6F,$99,$A4,$9E,$B1,$97,$9F,$CD,$78,$88,$E0
	.dc.b	$F1,$F0,$FF,$5C,$7B,$9F,$E1,$3F,$5F,$D9,$19,$D3,$FD,$D6,$D9,$F9
	.dc.b	$5E,$52,$FF,$10,$D8,$FC,$47,$B8,$F9,$8C,$D7,$E3,$7E,$1F,$E9,$E3
	.dc.b	$F7,$1B,$7F,$55,$AF,$E8,$DA,$FE,$7F,$9E,$E9,$F9,$EA,$F8,$31,$82
	.dc.b	$38,$24,$FB,$A7,$83,$F0,$33,$FC,$5C,$FC,$43,$BB,$88,$D4,$78,$B9
	.dc.b	$1F,$1D,$49,$1F,$7A,$32,$FE,$03,$E1,$9F,$4D,$EA,$6C,$66,$30,$78
	.dc.b	$39,$A4,$7D,$66,$E1,$E7,$F7,$7C,$A4,$9F,$B6,$F5,$F3,$FE,$DC,$86
	.dc.b	$08,$CC,$61,$B1,$9A,$FE,$8D,$3F,$D1,$E1,$F3,$DA,$24,$EA,$7B,$4E
	.dc.b	$A3,$10,$CB,$FC,$7D,$18,$6C,$71,$6B,$FC,$37,$E0,$BC,$BE,$0F,$FB
	.dc.b	$22,$B9,$B6,$7C,$0D,$7C,$ED,$7D,$1E,$87,$F0,$0F,$07,$F5,$78,$A3
	.dc.b	$17,$ED,$F4,$2D,$FA,$A6,$BF,$22,$3E,$29,$27,$DB,$48,$F3,$E7,$27
	.dc.b	$CE,$DA,$F8,$D5,$FE,$3D,$9F,$A9,$87,$93,$DE,$EF,$3F,$EC,$99,$C9
	.dc.b	$F4,$63,$3F,$1B,$23,$FA,$5C,$DB,$FF,$59,$2F,$4F,$99,$D3,$E6,$ED
	.dc.b	$74,$AF,$AF,$19,$8C,$43,$CE,$FF,$77,$46,$2D,$5E,$66,$46,$7D,$0B
	.dc.b	$7E,$27,$D6,$7C,$AB,$FE,$49,$3B,$3B,$FD,$67,$F9,$41,$BE,$E7,$D0
	.dc.b	$87,$C3,$D8,$C3,$27,$CF,$5F,$C4,$5E,$71,$C7,$C8,$91,$DB,$C7,$B9
	.dc.b	$7F,$B5,$F1,$F1,$B6,$3E,$FF,$C3,$BE,$B3,$C5,$1E,$3F,$46,$61,$E8
	.dc.b	$F4,$AE,$1C,$DF,$41,$CB,$C9,$B3,$F3,$3C,$4F,$35,$7F,$B9,$DB,$CD
	.dc.b	$FC,$FB,$EA,$7F,$D2,$35,$BF,$1B,$F2,$31,$8B,$FF,$77,$36,$BF,$A2
	.dc.b	$FA,$1F,$1A,$FE,$CC,$E5,$FF,$1F,$3E,$B7,$3A,$F4,$F4,$3A,$B6,$38
	.dc.b	$C7,$E8,$3E,$B1,$ED,$E7,$2F,$0F,$CA,$79,$0F,$2F,$8D,$70,$F0,$B8
	.dc.b	$7C,$7E,$95,$AF,$4B,$EC,$3F,$5B,$F5,$2F,$23,$15,$EA,$E2,$F8,$27
	.dc.b	$C2,$F8,$3B,$1E,$E7,$0F,$E6,$34,$7D,$66,$32,$F8,$FC,$9F,$28,$F9
	.dc.b	$7E,$7D,$BF,$8E,$E1,$FD,$0F,$D6,$7C,$9A,$FE,$51,$C7,$FD,$1B,$97
	.dc.b	$FB,$9F,$D6,$7F,$38,$94,$63,$CE,$FE,$0F,$2F,$83,$B1,$86,$63,$FA
	.dc.b	$E9,$FF,$4A,$B6,$38,$FC,$0C,$8F,$D9,$F1,$39,$31,$97,$82,$7F,$CF
	.dc.b	$58,$3E,$57,$C2,$78,$BC,$DB,$3D,$DC,$3D,$39,$AD,$F3,$3F,$A3,$5B
	.dc.b	$E0,$BD,$67,$B3,$A6,$DF,$A5,$C3,$E6,$F8,$DB,$7C,$C5,$BE,$7F,$E2
	.dc.b	$9F,$1B,$CC,$B7,$F0,$DC,$3F,$17,$CC,$B7,$CE,$FF,$88,$CE,$4F,$8B
	.dc.b	$F1,$1F,$42,$BE,$8F,$B3,$27,$E7,$5D,$39,$1F,$58,$FC,$8F,$B7,$63
	.dc.b	$F9,$0F,$D6,$7F,$13,$9F,$53,$79,$FF,$C6,$6F,$EB,$E3,$3E,$77,$F4
	.dc.b	$68,$C4,$EB,$CD,$FC,$EF,$6E,$5D,$7C,$2C,$C7,$BF,$C2,$DC,$BB,$BC
	.dc.b	$CE,$9F,$2B,$67,$CA,$E1,$FC,$29,$67,$E8,$7D,$67,$E5,$E3,$FD,$7C
	.dc.b	$D8,$FF,$01,$3F,$E1,$BC,$BF,$C9,$AE,$63,$FC,$6D,$EA,$FF,$60,$DF
	.dc.b	$FA,$94,$7D,$B7,$F0,$97,$FD,$19,$CD,$FF,$95,$BE,$AD,$F5,$9F,$DA
	.dc.b	$1B,$5F,$D3,$1F,$E9,$77,$F8,$57,$2D,$E0,$E0,$71,$DA,$D7,$1F,$3F
	.dc.b	$A1,$D7,$E2,$2C,$7F,$58,$FA,$67,$C9,$FF,$78,$6C,$67,$E7,$B1,$39
	.dc.b	$E8,$93,$C7,$9E,$E9,$F9,$9F,$AC,$FD,$64,$FF,$64,$6C,$67,$C1,$FF
	.dc.b	$02,$C6,$38,$7F,$CE,$EF,$F8,$BA,$B8,$7B,$9A,$F7,$36,$F0,$C2,$4F
	.dc.b	$F7,$03,$E1,$DF,$28,$FF,$08,$C7,$F5,$68,$FF,$41,$B4,$FF,$4F,$FC
	.dc.b	$4E,$BD,$9F,$F0,$BF,$D6,$3E,$72,$98,$F6,$6D,$FF,$7E,$FA,$CF,$F1
	.dc.b	$B6,$5F,$B7,$6E,$5F,$E9,$07,$D9,$3F,$E4,$99,$46,$A3,$0F,$ED,$23
	.dc.b	$97,$85,$FE,$28,$8F,$85,$18,$9F,$F0,$6F,$D6,$3C,$1F,$F1,$6B,$63
	.dc.b	$3E,$0E,$27,$3F,$01,$27,$8E,$D7,$0F,$E1,$DE,$EF,$F4,$99,$FF,$0C
	.dc.b	$B6,$33,$B7,$F8,$EA,$31,$AF,$FA,$B1,$FF,$55,$57,$0F,$67,$5E,$CE
	.dc.b	$DF,$F2,$0F,$D6,$7F,$98,$F1,$FE,$38,$DC,$3F,$D6,$AF,$C3,$8F,$F3
	.dc.b	$73,$9B,$F9,$0F,$CA,$8C,$FA,$1D,$3A,$3A,$B6,$BF,$C6,$35,$FE,$55
	.dc.b	$6F,$C2,$E6,$FC,$3E,$37,$90,$F2,$FA,$16,$B8,$B8,$79,$FA,$16,$BD
	.dc.b	$3D,$7E,$C3,$DB,$E4,$5A,$F0,$B8,$7F,$C8,$7C,$8E,$9C,$5F,$F3,$DD
	.dc.b	$DF,$01,$FE,$66,$6D,$FF,$92,$97,$8B,$99,$D3,$27,$AF,$BD,$6F,$FA
	.dc.b	$41,$FE,$59,$8F,$F2,$9B,$4C,$F6,$7F,$9D,$1C,$61,$FE,$D8,$8C,$19
	.dc.b	$EF,$B5,$C9,$C3,$EC,$E5,$B7,$E0,$DB,$A1,$D7,$6B,$5E,$27,$0F,$92
	.dc.b	$5E,$2F,$FA,$B7,$EB,$3F,$AF,$47,$E9,$EB,$CC,$E8,$FF,$51,$94,$FF
	.dc.b	$52,$CE,$1C,$1F,$AF,$FA,$B7,$63,$EB,$3F,$D5,$71,$86,$C6,$67,$FD
	.dc.b	$65,$39,$6F,$87,$FE,$16,$6E,$FC,$A3,$11,$FE,$BE,$94,$62,$3C,$5F
	.dc.b	$41,$FC,$06,$9A,$79,$78,$F0,$DA,$FF,$44,$E0,$BC,$C7,$FB,$46,$3F
	.dc.b	$0E,$FE,$33,$C8,$CB,$FC,$0D,$B1,$A8,$FF,$05,$3F,$8A,$7A,$F9,$5A
	.dc.b	$37,$CB,$63,$C3,$F2,$BC,$8C,$3C,$9E,$FD,$1F,$E2,$B8,$FC,$EF,$B3
	.dc.b	$B7,$FD,$43,$EB,$FF,$B9,$9B,$CB,$E6,$83,$79,$F0,$F4,$F1,$25,$70
	.dc.b	$3F,$D5,$7F,$59,$F0,$DB,$C3,$F0,$23,$DD,$C5,$F3,$27,$FD,$ED,$1B
	.dc.b	$7F,$C5,$4E,$B8,$BF,$99,$7D,$67,$F9,$36,$7F,$D6,$ED,$77,$7F,$33
	.dc.b	$EB,$3F,$CD,$91,$FE,$BC,$7F,$A2,$37,$1D,$7C,$36,$DE,$1D,$F2,$74
	.dc.b	$71,$F8,$7E,$9C,$D2,$30,$75,$7C,$1E,$94,$8C,$1E,$1F,$07,$C9,$C7
	.dc.b	$23,$07,$97,$6F,$C1,$FD,$5B,$EB,$29,$EB,$8C,$F8,$BA,$F1,$65,$FF
	.dc.b	$B6,$2F,$FC,$B0,$FF,$B6,$BD,$47,$CD,$7E,$39,$7C,$B5,$31,$19,$F5
	.dc.b	$F5,$F2,$D2,$F9,$47,$1F,$03,$97,$7C,$9E,$D1,$F9,$AF,$AC,$71,$5E
	.dc.b	$29,$5F,$EE,$C9,$CB,$FE,$EE,$FA,$C7,$9B,$FA,$27,$D3,$1F,$E6,$96
	.dc.b	$F7,$7E,$16,$11,$C1,$FF,$7B,$FD,$67,$F8,$C1,$BB,$7E,$B2,$3E,$0C
	.dc.b	$6B,$DE,$C7,$91,$9F,$22,$4F,$28,$E1,$E3,$F1,$DE,$4F,$68,$FD,$B9
	.dc.b	$FE,$2E,$47,$F9,$FD,$F6,$B7,$7D,$F5,$F0,$F8,$1E,$3F,$1B,$2B,$8E
	.dc.b	$FE,$DC,$15,$FC,$13,$EB,$3F,$D3,$95,$AE,$CF,$F0,$94,$3D,$6D,$1B
	.dc.b	$F5,$B8,$79,$5C,$7C,$A9,$87,$13,$CF,$7F,$BD,$7F,$D2,$DF,$58,$F6
	.dc.b	$FF,$89,$AF,$84,$A6,$23,$3E,$8E,$B9,$25,$F5,$17,$FC,$E1,$C7,$D6
	.dc.b	$7E,$C1,$FF,$52,$3F,$DA,$DC,$F9,$3F,$C8,$A3,$5C,$5F,$7D,$7B,$79
	.dc.b	$F1,$25,$F4,$8F,$E5,$3F,$3D,$F9,$C8,$F1,$2F,$84,$C9,$AF,$1A,$CE
	.dc.b	$BC,$5E,$26,$1F,$F5,$C5,$23,$F9,$AF,$CA,$BE,$32,$9C,$2B,$BF,$AD
	.dc.b	$E2,$E2,$91,$83,$C6,$FF,$3E,$F8,$EF,$CC,$47,$FB,$15,$BB,$FE,$77
	.dc.b	$2F,$A5,$3E,$AE,$2F,$6F,$83,$23,$07,$B9,$FE,$93,$F5,$8F,$9C,$A6
	.dc.b	$E7,$D7,$E1,$FC,$6F,$2E,$47,$C6,$8A,$FC,$0F,$AC,$F9,$66,$33,$F2
	.dc.b	$3D,$BC,$29,$3F,$3A,$3F,$AC,$FD,$5B,$E8,$AF,$AF,$2B,$F0,$93,$96
	.dc.b	$C7,$F6,$0F,$A8,$7C,$9D,$38,$4F,$FB,$A6,$9E,$C7,$6F,$2A,$4F,$48
	.dc.b	$AF,$FB,$BF,$EB,$1F,$3F,$FD,$B1,$B1,$E2,$AF,$3F,$B5,$E8,$EA,$93
	.dc.b	$E6,$8A,$F8,$7E,$B3,$DA,$63,$87,$F8,$96,$68,$FE,$F3,$F5,$9F,$8F
	.dc.b	$AF,$D1,$35,$E4,$7C,$2F,$07,$E3,$F8,$52,$30,$7C,$67,$FC,$07,$F5
	.dc.b	$9F,$D0,$63,$F5,$8D,$7E,$24,$7D,$0F,$43,$E8,$F9,$F2,$7E,$7C,$55
	.dc.b	$9F,$F2,$9E,$37,$C9,$FD,$13,$51,$FC,$B7,$EB,$3F,$B5,$37,$C1,$FF
	.dc.b	$0A,$C1,$BF,$37,$87,$3F,$4F,$31,$FE,$05,$5B,$6C,$7D,$3F,$9B,$91
	.dc.b	$83,$7E,$A7,$6F,$90,$91,$83,$9B,$F9,$EF,$22,$46,$0E,$FF,$03,$E6
	.dc.b	$FA,$C9,$18,$3C,$6F,$EC,$7E,$0E,$46,$0F,$E3,$1F,$59,$FC,$8F,$7F
	.dc.b	$E9,$CF,$F8,$7A,$33,$EB,$FF,$8D,$A3,$5C,$9F,$EC,$37,$53,$F3,$A3
	.dc.b	$FD,$8D,$F4,$DF,$EB,$D8,$FE,$48,$D7,$47,$F4,$89,$E3,$7F,$D7,$7E
	.dc.b	$B3,$FA,$D4,$E8,$3F,$D6,$9C,$36,$38,$7B,$DF,$E4,$A8,$C6,$1E,$2F
	.dc.b	$2B,$FC,$9B,$89,$D7,$C0,$6F,$7E,$1E,$8C,$CF,$F9,$D6,$3E,$7C,$FC
	.dc.b	$93,$C8,$93,$A7,$C1,$F1,$5E,$1F,$32,$4F,$28,$AF,$AF,$D5,$FE,$BA
	.dc.b	$FF,$47,$6B,$EC,$9F,$57,$9D,$EB,$D2,$4F,$92,$2B,$F1,$7E,$A7,$FD
	.dc.b	$1E,$DF,$8E,$CE,$7F,$E0,$7F,$AB,$FE,$97,$C3,$FE,$F9,$7F,$A8,$57
	.dc.b	$CD,$6C,$EF,$E3,$3F,$CE,$71,$8A,$FF,$80,$1A,$3C,$2F,$F3,$BB,$70
	.dc.b	$FE,$C1,$5B,$7F,$C0,$6F,$F1,$87,$3E,$77,$43,$AF,$D4,$76,$F3,$7F
	.dc.b	$D4,$75,$FE,$0B,$AF,$F0,$59,$E7,$8C,$9F,$EA,$8F,$AB,$FC,$14,$68
	.dc.b	$38,$78,$B5,$FF,$0A,$B5,$4B,$E2,$FC,$24,$8E,$87,$A9,$F2,$F9,$DE
	.dc.b	$4D,$3C,$15,$FE,$89,$AF,$F1,$A1,$D2,$74,$FC,$F6,$BF,$13,$CF,$AF
	.dc.b	$89,$B3,$FD,$B6,$3F,$C5,$75,$DA,$77,$F5,$9F,$E0,$B9,$EE,$D9,$F8
	.dc.b	$4D,$8F,$92,$D7,$E9,$FA,$7D,$88,$F6,$3F,$08,$F8,$BE,$62,$46,$4F
	.dc.b	$B6,$E1,$F0,$FC,$DF,$90,$8E,$C5,$23,$DB,$E7,$49,$D6,$48,$C4,$7C
	.dc.b	$A8,$FE,$6B,$3F,$4A,$3E,$BF,$9B,$CD,$FA,$B2,$F7,$5F,$31,$B8,$E3
	.dc.b	$3A,$F8,$AF,$F5,$66,$BF,$61,$B1,$FA,$36,$BF,$98,$6B,$F8,$66,$CF
	.dc.b	$F9,$52,$7D,$31,$83,$3C,$D2,$3D,$AD,$AF,$C2,$F3,$7F,$BA,$27,$E3
	.dc.b	$36,$33,$1E,$08,$C3,$F9,$DF,$89,$FF,$51,$3F,$35,$B3,$F9,$06,$C6
	.dc.b	$FE,$4B,$D3,$0F,$A6,$A4,$9E,$4F,$7E,$49,$1E,$46,$DF,$5F,$4B,$99
	.dc.b	$27,$C4,$F6,$F4,$7F,$43,$3F,$E8,$2C,$11,$88,$46,$2B,$EE,$36,$3F
	.dc.b	$5E,$BE,$B3,$83,$67,$71,$83,$A7,$4F,$F8,$FE,$3F,$0C,$DA,$FD,$CB
	.dc.b	$5F,$E0,$CD,$7F,$86,$D7,$FA,$26,$7C,$7E,$8F,$80,$8E,$0F,$F3,$E4
	.dc.b	$6E,$7F,$AA,$2F,$47,$FD,$F1,$E0,$7F,$AD,$63,$4F,$FA,$DF,$C2,$F5
	.dc.b	$57,$0A,$DD,$FF,$1E,$9C,$D7,$19,$CD,$7C,$BF,$72,$BF,$29,$B8,$7F
	.dc.b	$B2,$30,$F2,$C6,$3E,$46,$47,$FA,$4F,$1F,$A2,$FA,$A3,$8F,$A9,$23
	.dc.b	$DB,$F3,$97,$1F,$6B,$D8,$F0,$F9,$FF,$EE,$A9,$D9,$F1,$32,$F1,$9E
	.dc.b	$4F,$4A,$C7,$C8,$70,$FF,$80,$BE,$32,$E1,$FD,$F7,$EB,$3F,$D2,$D3
	.dc.b	$FE,$A9,$6C,$67,$E2,$7E,$0C,$63,$83,$FD,$C4,$FF,$A5,$6B,$84,$F9
	.dc.b	$BC,$0D,$BC,$3E,$7E,$4F,$5C,$6E,$3E,$17,$A9,$F3,$BE,$29,$E5,$62
	.dc.b	$BC,$9E,$8F,$88,$F1,$FA,$36,$3D,$0E,$1F,$F6,$15,$3C,$8C,$63,$FE
	.dc.b	$1B,$F0,$BF,$92,$57,$FB,$1D,$B1,$FE,$1F,$FA,$9F,$F4,$54,$E7,$6F
	.dc.b	$C0,$C6,$27,$5F,$0D,$F8,$4F,$8E,$F8,$51,$8B,$FF,$69,$57,$A9,$BA
	.dc.b	$F2,$F4,$39,$B6,$37,$1F,$8E,$FA,$C7,$BD,$9C,$3C,$9E,$8F,$88,$F1
	.dc.b	$F8,$B7,$0E,$EE,$1F,$0F,$A3,$6B,$D8,$FA,$8F,$CD,$FD,$53,$E2,$71
	.dc.b	$5E,$8E,$2F,$79,$F0,$3E,$16,$C7,$B1,$C3,$FE,$EF,$D1,$F5,$98,$CB
	.dc.b	$E5,$72,$7C,$83,$E4,$F9,$D6,$FE,$2B,$87,$FD,$F1,$F5,$9F,$37,$B9
	.dc.b	$1F,$E1,$2E,$4F,$F1,$03,$FE,$FD,$FA,$CF,$E5,$8D,$F3,$BF,$6B,$0F
	.dc.b	$86,$B1,$86,$4F,$1F,$3E,$E9,$FE,$5D,$BF,$F4,$64,$E4,$47,$F9,$87
	.dc.b	$89,$C9,$DA,$FB,$FC,$BC,$BF,$27,$9B,$70,$E8,$F9,$0F,$9F,$CD,$67
	.dc.b	$D5,$F0,$BE,$83,$D3,$D3,$63,$DC,$E1,$F1,$F0,$7C,$4C,$3C,$BA,$7F
	.dc.b	$B6,$27,$67,$C5,$F7,$7E,$19,$F1,$3C,$BB,$1F,$25,$C3,$F2,$FC,$BB
	.dc.b	$7F,$0D,$F9,$67,$E1,$BD,$9B,$3F,$47,$D7,$F9,$E7,$D0,$AB,$1F,$55
	.dc.b	$C3,$E8,$3F,$C9,$78,$FF,$B0,$1D,$19,$8F,$9F,$F5,$9F,$C3,$63,$FC
	.dc.b	$A2,$D8,$CF,$95,$FC,$CA,$31,$07,$F6,$3A,$7C,$58,$C4,$FE,$7B,$AB
	.dc.b	$83,$C7,$E0,$E6,$1E,$1A,$E5,$F0,$58,$7F,$D4,$18,$A3,$C0,$78,$7E
	.dc.b	$5E,$D7,$19,$46,$63,$A6,$7F,$E0,$DC,$FC,$95,$A9,$F0,$7D,$67,$E7
	.dc.b	$31,$FF,$57,$36,$BD,$4F,$19,$FF,$39,$B9,$7F,$CE,$BA,$FF,$51,$AF
	.dc.b	$F5,$B3,$6B,$E1,$2E,$3F,$EB,$2E,$6B,$FC,$FB,$F5,$8F,$88,$FE,$BA
	.dc.b	$D8,$F8,$8F,$F4,$FC,$3F,$D8,$4E,$5B,$F2,$FC,$0E,$3C,$16,$B8,$F9
	.dc.b	$DD,$0E,$BF,$0D,$63,$E8,$7D,$33,$E1,$BF,$B6,$B6,$33,$F3,$98,$9C
	.dc.b	$F3,$C9,$E3,$F0,$B7,$4F,$F6,$B7,$D6,$3C,$0F,$EE,$ED,$AF,$BC,$FF
	.dc.b	$81,$1F,$F4,$EB,$99,$E3,$ED,$70,$38,$FB,$56,$B8,$CF,$43,$FC,$06
	.dc.b	$FC,$3E,$5F,$C8,$5F,$F9,$ED,$FC,$94,$3D,$8D,$9C,$3E,$6A,$C7,$E4
	.dc.b	$7E,$B1,$F1,$7F,$E0,$E6,$C6,$7D,$7C,$4F,$F8,$4B,$EB,$3F,$C2,$F3
	.dc.b	$FE,$B6,$6E,$7F,$C6,$1F,$66,$5F,$A3,$73,$05,$E1,$F3,$3A,$6D,$6F
	.dc.b	$F4,$9F,$58,$F0,$FF,$C4,$2D,$8C,$F8,$78,$9C,$FB,$D2,$78,$F0,$5C
	.dc.b	$DA,$7E,$D7,$DC,$FE,$A1,$5F,$B1,$6C,$EA,$3F,$DA,$31,$F0,$A3,$07
	.dc.b	$F3,$47,$73,$CF,$D7,$FF,$1A,$C7,$F7,$FA,$E9,$EB,$E2,$7F,$C7,$3F
	.dc.b	$59,$FE,$5F,$BF,$E0,$CD,$AF,$F6,$FB,$F1,$8C,$61,$FD,$09,$F9,$71
	.dc.b	$AE,$7E,$7E,$77,$46,$C7,$F0,$B9,$FC,$75,$71,$F8,$5C,$BE,$A1,$1F
	.dc.b	$B3,$7B,$7C,$1F,$9B,$C7,$F3,$B6,$FA,$47,$AB,$CF,$C5,$73,$B7,$67
	.dc.b	$B8,$F7,$F9,$17,$0F,$23,$87,$F9,$07,$1B,$AB,$18,$79,$77,$7C,$0F
	.dc.b	$F9,$81,$B7,$FC,$9D,$78,$B9,$9D,$32,$7A,$FB,$B6,$FE,$00,$F4,$3F
	.dc.b	$CD,$6D,$9F,$E6,$4F,$E5,$1B,$1A,$EC,$91,$AE,$CB,$5D,$1E,$67,$F9
	.dc.b	$DB,$4F,$C1,$46,$67,$AE,$DE,$03,$C3,$B5,$8F,$0B,$87,$C7,$2F,$1B
	.dc.b	$FC,$07,$F5,$8F,$4B,$FC,$0B,$3F,$E0,$6B,$F5,$BB,$3F,$DD,$05,$3F
	.dc.b	$DD,$55,$87,$07,$2F,$AB,$FE,$0D,$F5,$9F,$EE,$F6,$C7,$63,$FC,$26
	.dc.b	$E1,$C3,$5E,$5E,$2F,$EB,$9E,$02,$11,$98,$E2,$D7,$8F,$1F,$87,$E8
	.dc.b	$7D,$34,$69,$27,$CD,$C9,$A2,$38,$3D,$5C,$9C,$88,$E0,$F7,$79,$9D
	.dc.b	$88,$E0,$F8,$5F,$27,$E2,$23,$83,$DA,$7D,$3E,$3F,$9F,$BF,$FB,$7A
	.dc.b	$32,$D7,$E1,$B5,$F9,$8F,$EA,$1D,$F1,$72,$23,$DB,$EB,$3E,$13,$73
	.dc.b	$FB,$B1,$88,$F9,$ED,$7E,$C3,$FA,$A7,$96,$7C,$08,$FA,$BE,$AF,$FC
	.dc.b	$E4,$DF,$84,$FC,$4C,$62,$3E,$DB,$5F,$C2,$3F,$D2,$1E,$5E,$57,$02
	.dc.b	$3D,$7E,$17,$99,$1C,$1E,$1F,$43,$B9,$1C,$1E,$5E,$CF,$19,$1C,$1E
	.dc.b	$9E,$9F,$39,$1C,$1F,$5B,$EA,$FF,$C2,$4D,$F0,$7F,$4B,$18,$8F,$CB
	.dc.b	$B5,$FD,$07,$A2,$F9,$8F,$D5,$B6,$3F,$02,$D7,$F3,$ED,$78,$FA,$9C
	.dc.b	$08,$F0,$79,$F3,$48,$E6,$D7,$F5,$3F,$54,$F5,$7F,$76,$D8,$FC,$8B
	.dc.b	$5D,$C7,$F0,$0F,$AA,$79,$14,$93,$F9,$88,$CB,$F9,$AF,$AA,$FF,$19
	.dc.b	$37,$DA,$FE,$13,$18,$8F,$DD,$B5,$FE,$08,$FE,$F1,$E5,$E0,$70,$23
	.dc.b	$D7,$BB,$99,$1C,$1F,$AA,$7F,$C6,$4D,$E3,$FE,$33,$18,$8F,$E1,$6D
	.dc.b	$7F,$8A,$2F,$7F,$64,$8E,$8F,$F2,$2F,$AA,$FF,$1A,$47,$EE,$1A,$FF
	.dc.b	$16,$6B,$FC,$7D,$AF,$F2,$E7,$F7,$0D,$EF,$E8,$47,$F8,$1F,$D5,$7F
	.dc.b	$9D,$DB,$EB,$7F,$32,$8C,$47,$F2,$46,$BF,$CB,$5F,$F1,$B3,$CB,$D5
	.dc.b	$E0,$47,$F8,$7F,$D5,$7E,$BD,$FE,$5A,$D7,$F9,$B3,$5F,$E8,$AD,$7F
	.dc.b	$D9,$6B,$D5,$FE,$8A,$D8,$EB,$18,$3F,$A4,$FC,$DB,$E6,$A9,$27,$F9
	.dc.b	$0C,$65,$BE,$FF,$02,$38,$37,$D1,$D0,$8E,$0E,$7E,$7F,$80,$8E,$0E
	.dc.b	$FA,$79,$08,$E0,$F1,$F1,$7A,$08,$E0,$F3,$F8,$DE,$C2,$38,$3D,$7D
	.dc.b	$5F,$00,$8E,$0F,$7F,$93,$F1,$08,$E0,$F8,$7F,$51,$F2,$08,$E0,$FE
	.dc.b	$8F,$F4,$EF,$93,$FE,$EA,$D8,$CC,$60,$CF,$91,$23,$8B,$5F,$ED,$B2
	.dc.b	$F1,$B9,$92,$3B,$BD,$BA,$A4,$78,$9A,$F2,$F8,$AF,$19,$1F,$3B,$E6
	.dc.b	$F8,$99,$1E,$A6,$BE,$2F,$A7,$F5,$91,$F7,$BE,$EE,$69,$1F,$09,$AF
	.dc.b	$AB,$C4,$F8,$64,$7E,$3B,$F1,$BE,$46,$47,$CA,$6B,$F0,$BC,$9F,$96
	.dc.b	$47,$E7,$BF,$E2,$2F,$A4,$7C,$7D,$24,$FF,$7A,$8C,$B7,$E3,$B8,$11
	.dc.b	$C1,$BF,$9D,$E8,$47,$07,$3F,$82,$F0,$11,$C1,$DF,$E0,$7C,$84,$70
	.dc.b	$78,$FE,$07,$D0,$47,$07,$9F,$E9,$3D,$84,$70,$7F,$85,$3E,$93,$F8
	.dc.b	$BA,$0E,$1B,$E3,$7F,$91,$A3,$07,$F8,$4E,$3F,$C7,$2D,$7F,$C8,$4D
	.dc.b	$7F,$CA,$2D,$7B,$7D,$27,$F4,$57,$FC,$AA,$D7,$3E,$67,$F9,$52,$31
	.dc.b	$07,$FB,$43,$C3,$D1,$D1,$1F,$F1,$7C,$7F,$B0,$DB,$E1,$73,$23,$AF
	.dc.b	$0B,$B1,$1C,$1C,$BD,$7F,$11,$1C,$1D,$BC,$DF,$31,$1C,$1E,$2F,$0B
	.dc.b	$D4,$47,$07,$9B,$C2,$F7,$11,$C1,$FE,$48,$FA,$2F,$E6,$89,$A4,$F5
	.dc.b	$FF,$CF,$2F,$FA,$96,$38,$7C,$26,$08,$F3,$F9,$AD,$91,$C1,$DF,$C2
	.dc.b	$E6,$47,$07,$8F,$C2,$EE,$47,$07,$9F,$D5,$F1,$91,$C1,$EB,$F6,$7C
	.dc.b	$E4,$70,$7B,$FC,$4F,$F4,$BF,$7F,$FA,$2F,$34,$62,$7F,$D3,$71,$87
	.dc.b	$F3,$13,$B9,$D4,$7E,$6E,$5E,$26,$23,$E5,$36,$33,$1C,$8E,$B1,$9F
	.dc.b	$12,$46,$4E,$8E,$15,$F9,$FE,$93,$A1,$23,$81,$F6,$5F,$07,$C4,$C8
	.dc.b	$CE,$08,$C4,$74,$3D,$FE,$37,$9C,$8F,$07,$D5,$E3,$48,$E6,$D7,$91
	.dc.b	$F0,$FE,$41,$F1,$7F,$8D,$6C,$7F,$AC,$B0,$7A,$2F,$86,$D5,$FE,$4E
	.dc.b	$31,$5B,$6B,$F5,$DA,$FD,$66,$BF,$4F,$D8,$F6,$6B,$F7,$5B,$1F,$71
	.dc.b	$AE,$E3,$E2,$F9,$78,$8F,$D1,$37,$13,$4C,$4C,$3E,$69,$67,$E1,$8F
	.dc.b	$0D,$F4,$7D,$5B,$3F,$61,$B3,$C1,$B1,$FE,$DF,$5F,$31,$C4,$91,$C5
	.dc.b	$E9,$E6,$C8,$E8,$DA,$E9,$E6,$78,$49,$3E,$17,$C9,$E6,$7A,$27,$F9
	.dc.b	$B6,$08,$C4,$7A,$DB,$19,$AF,$F6,$8A,$FA,$3C,$92,$32,$F0,$F8,$E9
	.dc.b	$1C,$1B,$7F,$EF,$6A,$79,$DD,$89,$5D,$9F,$0F,$9D,$23,$C2,$DA,$E5
	.dc.b	$F4,$5E,$62,$4F,$99,$F4,$FC,$04,$8F,$4B,$67,$C5,$F1,$DE,$E2,$4F
	.dc.b	$B9,$F8,$3F,$1D,$23,$E0,$B6,$7D,$59,$7C,$62,$4F,$C6,$7E,$4F,$C4
	.dc.b	$48,$F9,$2D,$9F,$EC,$C4,$FA,$07,$BB,$F4,$9B,$5F,$F0,$01,$B3,$C0
	.dc.b	$FE,$DC,$E1,$C2,$F7,$79,$12,$79,$3D,$7E,$46,$47,$56,$D7,$51,$4F
	.dc.b	$81,$F2,$12,$BC,$8F,$9F,$D0,$91,$E7,$6D,$73,$A7,$A1,$ED,$25,$7B
	.dc.b	$5F,$C8,$F6,$7F,$0C,$8F,$15,$7F,$AB,$FB,$B2,$48,$CF,$CE,$7F,$97
	.dc.b	$5B,$EE,$FE,$6A,$31,$5E,$B6,$BF,$25,$79,$DB,$25,$74,$79,$79,$D2
	.dc.b	$3B,$B6,$BA,$79,$9D,$C9,$3E,$47,$C5,$E6,$48,$F3,$B6,$7C,$9E,$67
	.dc.b	$9C,$93,$EC,$7D,$5E,$54,$8F,$7B,$67,$D9,$E5,$7B,$C9,$3F,$11,$F8
	.dc.b	$5B,$48,$F8,$ED,$9F,$89,$E2,$FC,$72,$4F,$CC,$7F,$81,$FC,$E7,$F1
	.dc.b	$48,$FA,$6D,$9F,$C3,$35,$FC,$A3,$5F,$CE,$BF,$E6,$28,$E3,$E2,$74
	.dc.b	$23,$D3,$C4,$EE,$47,$07,$83,$C2,$F1,$91,$C1,$E4,$F0,$BC,$E4,$70
	.dc.b	$7A,$3E,$2B,$D6,$47,$07,$B3,$DC,$F7,$91,$C1,$F0,$3E,$37,$E1,$91
	.dc.b	$C1,$F1,$3C,$2F,$8E,$47,$07,$C8,$F0,$BE,$59,$1C,$1F,$33,$CC,$F9
	.dc.b	$E4,$70,$7D,$0D,$BE,$99,$1C,$1F,$A6,$F9,$AF,$E0,$51,$FC,$31,$AF
	.dc.b	$F9,$CD,$A9,$A7,$F0,$78,$E5,$E0,$FF,$42,$8C,$4F,$F0,$36,$BF,$C4
	.dc.b	$57,$B3,$B2,$57,$8D,$E5,$EC,$C8,$F3,$36,$BA,$7C,$E7,$72,$4F,$AD
	.dc.b	$FE,$A7,$F3,$7F,$DD,$23,$F8,$1B,$67,$F8,$B3,$5F,$E4,$8D,$7F,$96
	.dc.b	$E4,$F2,$7F,$AE,$B6,$BF,$86,$B6,$7F,$90,$35,$FE,$5D,$F3,$6F,$7F
	.dc.b	$FB,$2B,$6B,$F8,$AB,$67,$F9,$43,$5E,$3E,$4F,$02,$4F,$07,$9F,$8B
	.dc.b	$23,$9B,$67,$F9,$97,$CA,$BD,$AF,$EE,$AD,$AF,$E4,$AD,$9D,$CE,$BD
	.dc.b	$79,$5A,$6C,$FF,$3D,$F9,$B7,$57,$F8,$01,$B5,$FC,$C1,$B3,$B9,$D7
	.dc.b	$9F,$2B,$4D,$9E,$87,$32,$47,$37,$FC,$1A,$FF,$4B,$3F,$9E,$95,$FE
	.dc.b	$A0,$D7,$FA,$F3,$5F,$EC,$FF,$36,$F7,$BF,$C2,$ED,$9F,$E8,$ED,$9F
	.dc.b	$EB,$4D,$7F,$B4,$7C,$DF,$F8,$66,$7F,$A6,$B6,$3F,$B0,$35,$DF,$AB
	.dc.b	$FE,$24,$87,$55,$25,$72,$7F,$C5,$3F,$37,$FA,$48,$FE,$B2,$DB,$FE
	.dc.b	$D8,$D7,$FB,$D3,$5F,$F0,$1A,$F4,$F9,$92,$79,$BF,$E3,$6F,$41,$F1
	.dc.b	$34,$93,$FD,$A2,$32,$DF,$2B,$81,$1C,$1F,$DB,$BE,$6F,$FC,$7D,$8F
	.dc.b	$F8,$05,$AF,$F8,$3D,$AF,$F8,$61,$AF,$FB,$FD,$FC,$3B,$BE,$0F,$F2
	.dc.b	$5C,$62,$38,$46,$3C,$59,$5E,$16,$CF,$87,$AF,$89,$27,$CC,$F4,$F1
	.dc.b	$64,$7A,$5B,$3E,$9F,$5B,$C2,$49,$F7,$3E,$4E,$D9,$1F,$05,$B3,$FE
	.dc.b	$21,$F9,$97,$AD,$FE,$64,$6D,$7F,$85,$1B,$3B,$9D,$76,$4A,$D3,$67
	.dc.b	$FC,$37,$E3,$7F,$9B,$93,$49,$3F,$E2,$08,$FC,$7B,$79,$38,$11,$D7
	.dc.b	$4F,$32,$38,$39,$67,$DC,$8E,$0E,$D9,$F8,$C8,$E0,$F1,$5B,$B6,$91
	.dc.b	$F3,$3F,$E3,$6F,$99,$FF,$23,$7C,$D5,$F6,$4D,$25,$7F,$93,$65,$F2
	.dc.b	$0E,$0E,$4D,$9D,$35,$FF,$30,$4F,$09,$F0,$7C,$87,$C7,$83,$49,$3E
	.dc.b	$5E,$1D,$11,$C1,$E9,$E1,$E4,$47,$07,$B7,$E2,$7B,$11,$C1,$F0,$7E
	.dc.b	$27,$C4,$47,$07,$F9,$6B,$CE,$FF,$63,$37,$8B,$EF,$46,$23,$FC,$E0
	.dc.b	$D7,$FC,$FC,$B8,$B6,$48,$E8,$FE,$1F,$E6,$BE,$0C,$7F,$9A,$1A,$FF
	.dc.b	$9E,$DA,$FF,$A2,$5A,$FF,$A4,$9F,$CF,$BC,$72,$E8,$47,$FC,$E7,$F3
	.dc.b	$5F,$C4,$1B,$DD,$F9,$78,$C4,$7F,$A2,$5A,$FF,$A5,$DF,$E7,$0F,$2F
	.dc.b	$13,$81,$1E,$BE,$DF,$32,$38,$3C,$3E,$47,$72,$38,$3C,$BE,$DF,$8C
	.dc.b	$8E,$0F,$4F,$47,$9C,$8E,$0F,$F4,$7F,$CD,$3B,$7F,$64,$D8,$CC,$60
	.dc.b	$FD,$B7,$94,$F8,$EA,$49,$FF,$4C,$C6,$5B,$EA,$70,$23,$83,$7C,$BD
	.dc.b	$08,$E0,$FF,$51,$FC,$D3,$E3,$7F,$83,$36,$33,$18,$3F,$85,$7C,$D7
	.dc.b	$F8,$86,$3F,$D5,$6D,$7F,$D7,$CD,$7F,$D9,$8D,$7F,$DA,$BF,$35,$FE
	.dc.b	$21,$7F,$D6,$6D,$7F,$D8,$CD,$7F,$DA,$8D,$78,$F7,$7F,$1B,$8F,$DF
	.dc.b	$C7,$28,$C1,$9E,$B9,$1D,$9A,$FF,$B4,$DF,$F1,$13,$FE,$C4,$6B,$9E
	.dc.b	$1F,$E5,$11,$88,$FF,$73,$2F,$7B,$24,$8E,$4F,$F2,$FF,$9A,$FF,$11
	.dc.b	$C7,$FB,$41,$AF,$FB,$8D,$AF,$FB,$C5,$AF,$FB,$E9,$FD,$C3,$7B,$3A
	.dc.b	$11,$FF,$6D,$FC,$D3,$D9,$FE,$7C,$D8,$CC,$60,$CF,$89,$23,$8B,$5F
	.dc.b	$F7,$B7,$CD,$7E,$B1,$BD,$5F,$D2,$63,$10,$20,$72,$E1,$7C,$17,$12
	.dc.b	$47,$FB,$CA,$32,$FF,$BD,$3E,$3D,$F3,$5F,$D5,$DB,$19,$8C,$19,$F6
	.dc.b	$64,$71,$6B,$E4,$3C,$D2,$E7,$E8,$48,$F0,$3D,$FD,$E9,$1E,$36,$BC
	.dc.b	$C4,$FC,$D7,$0D,$7C,$0D,$7E,$2B,$5F,$1F,$1F,$B4,$8F,$C1,$7E,$07
	.dc.b	$8F,$23,$E2,$35,$EF,$D3,$F1,$48,$FC,$97,$E4,$78,$F2,$3E,$63,$5F
	.dc.b	$3F,$CD,$7C,$D2,$3F,$45,$FF,$03,$7C,$D3,$E2,$E9,$27,$E1,$08,$D3
	.dc.b	$87,$F9,$AB,$F8,$46,$BF,$87,$6B,$C7,$C6,$EF,$E2,$C8,$E4,$D7,$E8
	.dc.b	$4F,$EC,$5E,$9D,$7E,$62,$3E,$1F,$89,$F4,$91,$C1,$E5,$F8,$9F,$69
	.dc.b	$1C,$1E,$9F,$95,$F8,$24,$70,$7B,$79,$BE,$29,$1C,$1F,$07,$C3,$F9
	.dc.b	$24,$70,$7C,$5F,$8A,$F9,$A4,$70,$7C,$9F,$23,$E8,$91,$C1,$F4,$3E
	.dc.b	$69,$F1,$5F,$E3,$B6,$C6,$63,$06,$7E,$2A,$47,$16,$BF,$9B,$97,$C8
	.dc.b	$F3,$24,$77,$7B,$7B,$72,$3C,$4D,$79,$7C,$8F,$8C,$8F,$9D,$F3,$7C
	.dc.b	$64,$8F,$53,$5F,$17,$CD,$7A,$C8,$FB,$DF,$F2,$C7,$CD,$7F,$22,$8F
	.dc.b	$E2,$0D,$4D,$27,$AD,$C0,$91,$F9,$88,$FF,$10,$3F,$9A,$F9,$AF,$E9
	.dc.b	$8F,$F1,$B6,$A6,$93,$D1,$FF,$35,$3F,$DC,$63,$F8,$23,$5F,$E2,$0D
	.dc.b	$7F,$92,$AF,$07,$89,$23,$C4,$F4,$F2,$24,$79,$5A,$F6,$F5,$BC,$24
	.dc.b	$7D,$4F,$93,$CE,$91,$ED,$6B,$E6,$F0,$7D,$24,$7E,$13,$EC,$F0,$64
	.dc.b	$7C,$56,$BF,$C8,$7E,$5F,$F9,$EB,$FD,$05,$A9,$A4,$F9,$AF,$F4,$82
	.dc.b	$69,$23,$F8,$5C,$3E,$4F,$64,$8C,$BC,$BC,$19,$1C,$1A,$F9,$7C,$0E
	.dc.b	$E4,$7A,$BE,$2F,$52,$47,$81,$AF,$B7,$D8,$F3,$91,$F2,$BE,$AF,$0B
	.dc.b	$FA,$FF,$6F,$F4,$DC,$91,$8A,$FE,$C7,$18,$7F,$B3,$4E,$E7,$51,$F2
	.dc.b	$7C,$2A,$FC,$96,$D1,$A6,$3C,$29,$3D,$0E,$EE,$15,$F9,$7E,$6B,$99
	.dc.b	$23,$81,$F6,$5E,$FF,$0F,$23,$27,$DB,$70,$F4,$3D,$BE,$27,$98,$8F
	.dc.b	$0C,$7F,$BD,$BB,$3F,$0D,$E2,$7B,$08,$ED,$F7,$FC,$D3,$E2,$7F,$06
	.dc.b	$D9,$FC,$83,$5F,$A0,$D7,$E7,$B5,$FC,$A3,$5F,$CA,$79,$AF,$84,$EB
	.dc.b	$5F,$90,$F6,$3D,$9A,$FD,$B6,$DF,$DA,$6B,$F9,$F6,$BF,$9F,$F2,$6E
	.dc.b	$66,$92,$BE,$21,$A2,$18,$3E,$7F,$AE,$F9,$AE,$AD,$8C,$C6,$0C,$FC
	.dc.b	$D4,$8E,$2D,$C3,$E8,$0A,$FE,$8F,$3D,$BC,$CE,$A4,$7C,$3E,$67,$80
	.dc.b	$8E,$0F,$2F,$97,$D2,$FF,$C8,$17,$FD,$22,$7F,$C0,$6D,$8F,$F1,$33
	.dc.b	$5F,$F1,$92,$F8,$FE,$51,$E5,$6C,$66,$79,$C6,$3E,$2E,$47,$66,$D7
	.dc.b	$6F,$37,$89,$27,$C6,$F4,$F3,$64,$79,$9B,$3E,$6F,$9A,$F0,$92,$7D
	.dc.b	$6F,$93,$E3,$24,$7B,$9B,$3E,$EF,$8C,$F4,$92,$7E,$1B,$EC,$F8,$C9
	.dc.b	$1F,$19,$B3,$F1,$B4,$F8,$24,$9F,$96,$FC,$4F,$3E,$47,$CE,$6C,$FF
	.dc.b	$94,$FE,$69,$ED,$FC,$C6,$D7,$F8,$D9,$B3,$B9,$D7,$B7,$2B,$4D,$9E
	.dc.b	$9F,$1F,$CC,$93,$CD,$ED,$F0,$12,$3B,$36,$7C,$9E,$7F,$8C,$93,$E3
	.dc.b	$7C,$DE,$7C,$8F,$33,$67,$FC,$DD,$D9,$FC,$0E,$7F,$C9,$ED,$8C,$F6
	.dc.b	$FE,$2A,$31,$1F,$E7,$FF,$9A,$FF,$2C,$57,$F9,$59,$B1,$FE,$79,$6B
	.dc.b	$BF,$6F,$F2,$71,$FD,$46,$78,$79,$D8,$23,$CF,$CE,$D9,$1C,$1D,$FC
	.dc.b	$CE,$64,$70,$78,$FC,$CE,$E4,$70,$79,$FC,$CF,$19,$1C,$1E,$BF,$2B
	.dc.b	$CE,$47,$07,$BF,$CA,$F5,$91,$C1,$F0,$F5,$F7,$91,$C1,$F1,$FC,$5F
	.dc.b	$86,$47,$07,$FA,$03,$E6,$9E,$2F,$EE,$9B,$19,$8C,$19,$F1,$64,$71
	.dc.b	$6F,$0F,$FA,$AE,$9E,$2F,$32,$57,$77,$B7,$87,$23,$C4,$DA,$E5,$E1
	.dc.b	$F8,$C9,$3E,$77,$CD,$F1,$12,$3D,$4D,$9F,$17,$B5,$EB,$24,$FB,$DF
	.dc.b	$77,$C6,$48,$F8,$4D,$9F,$57,$87,$F0,$C9,$3F,$1D,$F8,$DE,$1C,$8F
	.dc.b	$94,$D9,$F8,$5E,$5F,$CB,$24,$FC,$F7,$E7,$44,$8F,$A4,$D3,$FD,$A9
	.dc.b	$F3,$5F,$BE,$7F,$DC,$0D,$7F,$DE,$AD,$7F,$DF,$0D,$4F,$FB,$96,$3F
	.dc.b	$D7,$AD,$8E,$BE,$0F,$F3,$88,$C4,$76,$F6,$30,$4A,$E2,$EF,$D8,$91
	.dc.b	$D1,$B3,$E8,$F9,$AE,$84,$9F,$0B,$FD,$1B,$E6,$BF,$B6,$47,$FB,$41
	.dc.b	$B3,$FE,$F3,$6B,$FE,$FF,$6A,$26,$4F,$2B,$FA,$8B,$6B,$FD,$B6,$D9
	.dc.b	$FF,$7F,$35,$E4,$73,$F9,$97,$CD,$7F,$5A,$6D,$7F,$B9,$1B,$3C,$0E
	.dc.b	$13,$BF,$26,$56,$DB,$3B,$14,$F1,$BB,$12,$BB,$3F,$DA,$BE,$55,$ED
	.dc.b	$52,$4F,$FB,$DA,$32,$DF,$5F,$81,$1C,$1F,$EF,$CF,$9A,$75,$7F,$76
	.dc.b	$6C,$66,$30,$67,$D0,$91,$C5,$B9,$FA,$EB,$FC,$96,$FD,$36,$A6,$98
	.dc.b	$97,$9B,$FE,$06,$69,$F1,$67,$FC,$BB,$19,$3F,$C1,$7F,$34,$F8,$CA
	.dc.b	$4A,$E1,$19,$79,$7C,$D7,$F8,$3E,$FE,$83,$5F,$B0,$D7,$F0,$2D,$7F
	.dc.b	$18,$BA,$BF,$C3,$6D,$9E,$B1,$83,$FC,$3F,$F3,$5F,$EC,$18,$F5,$36
	.dc.b	$BF,$00,$D7,$F1,$0D,$7F,$22,$BD,$2E,$64,$9E,$6F,$F8,$B3,$D1,$7C
	.dc.b	$4D,$24,$FC,$98,$CB,$7C,$BE,$04,$70,$7C,$FF,$9A,$FF,$1D,$63,$FA
	.dc.b	$26,$BF,$AD,$6A,$69,$F8,$87,$F3,$8D,$7F,$48,$D7,$A7,$0F,$F9,$12
	.dc.b	$1E,$2D,$25,$78,$DD,$75,$C8,$F3,$36,$FB,$F8,$BC,$C9,$3E,$B7,$B7
	.dc.b	$AD,$23,$DC,$D9,$F3,$F6,$F8,$C9,$3F,$0D,$FF,$2C,$FC,$CB,$D6,$A4
	.dc.b	$9F,$CB,$46,$5B,$D9,$C0,$8E,$0F,$CE,$5B,$C6,$FF,$32,$B6,$33,$18
	.dc.b	$33,$C9,$23,$8B,$67,$FC,$E7,$1D,$3A,$B9,$91,$EF,$9F,$62,$38,$3C
	.dc.b	$79,$F8,$88,$E0,$F3,$D3,$B2,$91,$F5,$BF,$B7,$F9,$9F,$DD,$FC,$D5
	.dc.b	$EC,$34,$95,$FC,$86,$5F,$20,$DF,$93,$67,$4D,$7F,$97,$4F,$09,$F0
	.dc.b	$7C,$87,$F3,$87,$FD,$A5,$08,$C6,$5F,$CF,$23,$15,$FD,$1E,$B7,$7D
	.dc.b	$FE,$6D,$E6,$6F,$8F,$8D,$FF,$4D,$C3,$F0,$4F,$1C,$74,$D3,$59,$70
	.dc.b	$53,$FC,$F5,$D1,$8E,$3D,$46,$63,$E1,$57,$F5,$2F,$DA,$7F,$A6,$69
	.dc.b	$EC,$F8,$BD,$92,$B9,$DF,$19,$F8,$3F,$17,$D0,$93,$AA,$F2,$69,$FE
	.dc.b	$C7,$8F,$1B,$5F,$09,$FD,$AD,$C5,$7F,$BF,$7E,$97,$FB,$1C,$F0,$E4
	.dc.b	$E6,$47,$9F,$2F,$C1,$F9,$DF,$9F,$1D,$03,$8B,$6B,$FD,$C2,$30,$FF
	.dc.b	$72,$C1,$F1,$FC,$F8,$66,$1F,$2A,$FE,$9B,$D5,$E2,$FF,$80,$FE,$77
	.dc.b	$F9,$2B,$F4,$1A,$F8,$DB,$8F,$8A,$38,$7C,$17,$D9,$9F,$E9,$D1,$C7
	.dc.b	$C8,$C1,$9F,$13,$5F,$01,$27,$B3,$C7,$E0,$79,$79,$B2,$7C,$4E,$DF
	.dc.b	$17,$C0,$F7,$23,$E7,$F8,$EF,$17,$0C,$8F,$53,$E5,$F7,$BF,$1D,$F3
	.dc.b	$AF,$16,$92,$BF,$1F,$1F,$E2,$37,$F2,$3E,$8F,$CC,$6F,$27,$E5,$E3
	.dc.b	$11,$F5,$DF,$BE,$DF,$6B,$44,$79,$FA,$9C,$7A,$E4,$76,$7A,$77,$FE
	.dc.b	$8F,$E7,$5B,$52,$57,$E9,$23,$FC,$3A,$FE,$9B,$E7,$7D,$0F,$E2,$5B
	.dc.b	$D1,$F5,$E1,$18,$E0,$FF,$1C,$AF,$88,$FD,$97,$CE,$FA,$AF,$83,$6B
	.dc.b	$FC,$7E,$DB,$FF,$25,$5F,$EC,$27,$3E,$37,$22,$4F,$EE,$AB,$FC,$32
	.dc.b	$DC,$FB,$91,$D7,$6F,$F0,$38,$F6,$C7,$E8,$DB,$8F,$AD,$B8,$7F,$94
	.dc.b	$97,$47,$1F,$84,$97,$FC,$2A,$51,$D7,$E3,$FF,$88,$7C,$EF,$F8,$F2
	.dc.b	$BF,$87,$B5,$FD,$93,$71,$F9,$4B,$E3,$78,$F8,$72,$B8,$BD,$3C,$FF
	.dc.b	$E3,$9F,$3A,$F8,$1A,$4A,$FE,$3B,$0F,$82,$DF,$A1,$27,$2F,$2F,$6B
	.dc.b	$F9,$37,$CE,$FE,$BA,$7F,$83,$37,$5F,$BC,$DE,$2F,$F3,$AB,$FB,$86
	.dc.b	$FC,$57,$22,$57,$F2,$E9,$7B,$7F,$CC,$F8,$5F,$3B,$49,$5F,$CD,$3C
	.dc.b	$DC,$BF,$CD,$A5,$DF,$C4,$91,$9A,$E9,$E0,$48,$E0,$E9,$FE,$41,$5F
	.dc.b	$0D,$E9,$BF,$8C,$FE,$8D,$F1,$0F,$90,$F3,$71,$CA,$F1,$3B,$7C,$9D
	.dc.b	$BE,$B2,$3F,$D3,$3E,$AD,$F5,$9E,$FF,$5E,$4F,$AD,$DB,$E4,$F1,$7E
	.dc.b	$21,$1E,$5E,$07,$C7,$F5,$64,$7C,$37,$4F,$B3,$E7,$FE,$61,$1D,$1F
	.dc.b	$D7,$BE,$7D,$EA,$7F,$63,$6C,$66,$3F,$B3,$7D,$43,$E6,$F5,$EA,$4A
	.dc.b	$E4,$F1,$AE,$5E,$3C,$9E,$CF,$F6,$FF,$AB,$F9,$51,$E1,$F5,$3C,$04
	.dc.b	$76,$79,$7E,$57,$C8,$47,$FB,$AF,$CE,$FF,$03,$5A,$3E,$5B,$D4,$48
	.dc.b	$E4,$7B,$F4,$F7,$11,$F8,$51,$F0,$7C,$19,$1F,$15,$F4,$7D,$07,$C5
	.dc.b	$F4,$E4,$FC,$97,$6F,$F6,$9C,$1E,$BF,$CB,$24,$7F,$83,$3E,$A9,$F3
	.dc.b	$7F,$3F,$C7,$93,$F4,$DD,$BF,$DC,$FE,$89,$F3,$3F,$E1,$76,$C7,$F8
	.dc.b	$59,$AF,$F7,$9F,$BE,$7C,$4E,$FC,$39,$3B,$79,$5F,$3C,$64,$F3,$76
	.dc.b	$F8,$7E,$33,$C0,$47,$FC,$4B,$F7,$1F,$B0,$7C,$5E,$2F,$94,$8F,$96
	.dc.b	$7D,$1F,$35,$23,$D0,$FE,$0F,$E1,$BD,$9C,$72,$7D,$8E,$DF,$F0,$A7
	.dc.b	$D2,$7F,$87,$B8,$3F,$E0,$0B,$60,$1C,$2F,$3B,$CC,$F2,$BF,$F2,$04
	.dc.b	$9A,$63,$D6,$FF,$22,$7D,$27,$E3,$49,$F3,$3F,$E4,$7C,$1E,$7E,$BE
	.dc.b	$0A,$65,$FE,$48,$BF,$EE,$BE,$AF,$F6,$17,$97,$A1,$D0,$93,$FC,$6F
	.dc.b	$CD,$EF,$E8,$48,$E4,$E9,$FF,$1A,$57,$F8,$AA,$FA,$79,$9E,$63,$C9
	.dc.b	$F5,$DE,$8F,$7A,$5F,$8D,$D3,$E5,$F3,$FD,$A4,$7F,$96,$67,$F0,$3C
	.dc.b	$F9,$1E,$B7,$4F,$F9,$17,$E8,$FF,$9F,$3F,$41,$A9,$BC,$1F,$D4,$63
	.dc.b	$FC,$D7,$0F,$23,$FC,$E0,$D8,$CC,$63,$E2,$7E,$77,$CE,$EB,$E2,$A5
	.dc.b	$7F,$41,$8E,$3E,$7C,$8E,$AE,$9F,$33,$FE,$BA,$C3,$FB,$AC,$7E,$18
	.dc.b	$87,$9E,$34,$7F,$B0,$3E,$B7,$F0,$F1,$A6,$BC,$5A,$FE,$41,$AF,$C1
	.dc.b	$F8,$8F,$01,$1E,$E7,$C7,$7C,$9F,$17,$23,$27,$C9,$70,$F0,$B7,$A1
	.dc.b	$E9,$EE,$F4,$C6,$27,$F4,$31,$89,$D4,$7B,$E7,$4F,$F8,$CC,$FD,$27
	.dc.b	$26,$9F,$D4,$EF,$FE,$76,$BF,$8A,$D8,$FF,$3F,$4E,$DF,$BD,$E8,$5B
	.dc.b	$FE,$89,$1E,$23,$C1,$19,$35,$F4,$D2,$34,$7E,$25,$C3,$E5,$F7,$3F
	.dc.b	$9D,$3A,$3F,$C0,$7E,$97,$86,$33,$1B,$8F,$17,$A5,$23,$86,$08,$C4
	.dc.b	$7C,$18,$7A,$5E,$92,$47,$06,$BB,$6E,$9F,$41,$F8,$1F,$4C,$F9,$AF
	.dc.b	$B3,$68,$C5,$FD,$9F,$7B,$F9,$9B,$F6,$E3,$2F,$DA,$F8,$8F,$F4,$E6
	.dc.b	$34,$F4,$B8,$8F,$B4,$D7,$F8,$FE,$9F,$E7,$69,$F8,$CD,$8F,$95,$C7
	.dc.b	$B7,$FC,$56,$DC,$7F,$D7,$75,$F1,$8C,$10,$FA,$1E,$33,$E9,$B9,$36
	.dc.b	$33,$18,$3B,$60,$E5,$A4,$9F,$2B,$CB,$D0,$91,$D9,$B9,$7E,$57,$07
	.dc.b	$A3,$E0,$25,$79,$1F,$1F,$A3,$F9,$BA,$FE,$83,$5F,$DB,$2B,$F5,$6D
	.dc.b	$8F,$D8,$B5,$FD,$E7,$8D,$C5,$75,$7F,$99,$EB,$F6,$6D,$8F,$DC,$B5
	.dc.b	$FE,$13,$4F,$9B,$E4,$DA,$F0,$C6,$0C,$FA,$1F,$ED,$9F,$1F,$FD,$49
	.dc.b	$48,$C5,$7F,$15,$8F,$F2,$73,$FD,$72,$77,$3A,$BF,$7F,$C6,$D5,$F5
	.dc.b	$AF,$AA,$C6,$3F,$EA,$DC,$79,$1F,$C8,$67,$4E,$2B,$E4,$FC,$E7,$0C
	.dc.b	$38,$B6,$75,$2B,$7C,$E7,$87,$EB,$7F,$B1,$D6,$A7,$2F,$08,$CD,$34
	.dc.b	$F8,$3F,$1D,$E6,$C3,$CE,$DB,$E5,$87,$5F,$43,$D6,$D8,$ED,$EA,$61
	.dc.b	$A7,$03,$CF,$F5,$3F,$E8,$47,$C5,$E7,$FC,$78,$C6,$7E,$39,$FE,$89
	.dc.b	$A2,$B5,$EA,$6F,$E2,$AD,$74,$79,$7C,$6F,$D1,$FA,$DF,$D8,$70,$7F
	.dc.b	$B7,$F3,$4E,$7C,$0F,$07,$C7,$F0,$C7,$CA,$DD,$7C,$47,$F8,$6A,$DE
	.dc.b	$5F,$5E,$FF,$43,$6F,$DB,$EA,$E1,$A7,$C0,$8F,$C2,$78,$9F,$C5,$63
	.dc.b	$EC,$F8,$0F,$6F,$3E,$47,$E2,$63,$13,$F7,$BE,$AB,$FD,$06,$FE,$02
	.dc.b	$F9,$69,$E6,$13,$A7,$11,$AE,$0F,$F8,$01,$2F,$4F,$F2,$DF,$1A,$F8
	.dc.b	$CF,$CC,$4E,$71,$FF,$52,$46,$3D,$4C,$DD,$9F,$23,$C3,$83,$F4,$1C
	.dc.b	$DF,$CF,$67,$D1,$18,$F3,$FA,$FA,$9B,$9F,$63,$76,$F5,$67,$FE,$87
	.dc.b	$9C,$9F,$07,$D1,$E2,$FC,$3F,$8F,$EF,$E9,$7A,$1B,$1F,$1F,$1C,$34
	.dc.b	$E0,$7C,$BE,$1F,$DB,$7D,$6B,$E7,$3E,$0C,$61,$F8,$7F,$13,$8A,$FA
	.dc.b	$4D,$D3,$E8,$7D,$53,$EA,$FF,$80,$7E,$09,$FD,$6A,$FF,$6B,$39,$7F
	.dc.b	$6F,$19,$AE,$3E,$9F,$F0,$7A,$D4,$62,$B8,$DE,$7A,$75,$76,$7A,$BC
	.dc.b	$7D,$0F,$E2,$1F,$5D,$FD,$2B,$0F,$2C,$62,$BB,$7A,$3B,$7C,$AD,$9F
	.dc.b	$5F,$CF,$7C,$4F,$5F,$C8,$DA,$F5,$FC,$46,$1A,$7B,$A3,$D5,$C1,$F3
	.dc.b	$35,$B3,$FC,$5B,$8B,$DB,$DE,$8F,$F2,$07,$07,$F1,$4C,$FE,$5E,$1F
	.dc.b	$0F,$D0,$E3,$1F,$3B,$D0,$C5,$FC,$C6,$C7,$D2,$F8,$BF,$B7,$C5,$F2
	.dc.b	$DB,$5F,$4A,$70,$D3,$EA,$C7,$F3,$7F,$AD,$FE,$23,$9F,$96,$BF,$D3
	.dc.b	$3A,$7F,$60,$BF,$E5,$F5,$A7,$FC,$B1,$19,$9F,$D9,$4F,$F2,$CA,$E5
	.dc.b	$08,$E0,$FF,$A9,$1D,$57,$F9,$85,$CC,$6F,$E1,$7F,$2B,$ED,$79,$5B
	.dc.b	$5E,$2F,$6B,$0D,$3C,$B1,$EB,$F8,$2F,$64,$D9,$D5,$FF,$07,$FA,$DF
	.dc.b	$F7,$1C,$7F,$4A,$5D,$9F,$D6,$5C,$CE,$E3,$FA,$E6,$AF,$62,$FB,$38
	.dc.b	$6E,$1F,$D5,$3E,$B5,$F1,$7C,$31,$FE,$C9,$18,$9F,$EB,$38,$7F,$95
	.dc.b	$78,$3E,$3C,$FC,$21,$EF,$F0,$F2,$78,$BC,$BD,$1E,$7E,$8D,$C7,$C0
	.dc.b	$FF,$71,$FA,$87,$CA,$5F,$3F,$0D,$AF,$ED,$33,$FB,$3C,$FE,$BB,$FC
	.dc.b	$67,$5F,$E5,$8E,$67,$FD,$2C,$E6,$7E,$EF,$D6,$FF,$83,$6B,$F9,$92
	.dc.b	$F9,$EF,$F0,$13,$99,$DC,$63,$DF,$CF,$6D,$C7,$FB,$DD,$A3,$8E,$1C
	.dc.b	$B8,$6D,$69,$FF,$07,$2F,$62,$FB,$58,$6C,$7F,$81,$BE,$B5,$F1,$7F
	.dc.b	$E1,$55,$F3,$56,$30,$FF,$86,$7E,$B7,$D5,$5F,$E0,$E7,$FC,$71,$BF
	.dc.b	$E7,$1C,$E7,$FD,$45,$CC,$FF,$AE,$DF,$53,$C7,$C4,$E9,$AF,$57,$FC
	.dc.b	$4D,$F5,$AF,$06,$FC,$8E,$1B,$7F,$E1,$E9,$FE,$F2,$DF,$0B,$8F,$07
	.dc.b	$27,$FC,$61,$EE,$FB,$27,$FC,$4A,$FF,$63,$D7,$F6,$6E,$75,$FE,$D0
	.dc.b	$E6,$7F,$DA,$4B,$D5,$E7,$F1,$BD,$1B,$3F,$E3,$0F,$AD,$F7,$D8,$3F
	.dc.b	$C0,$F0,$FE,$02,$EA,$33,$EA,$7F,$91,$63,$F0,$95,$FE,$E1,$7E,$5D
	.dc.b	$70,$F4,$F5,$D5,$BF,$4F,$84,$5B,$F1,$3C,$FC,$5E,$9E,$35,$9F,$33
	.dc.b	$DF,$D1,$C6,$9F,$E0,$7C,$D1,$8C,$FC,$FF,$0F,$FE,$59,$E4,$F8,$D1
	.dc.b	$8C,$7F,$C9,$F1,$87,$D1,$DB,$EB,$AF,$63,$63,$FC,$AB,$5F,$E2,$87
	.dc.b	$8F,$85,$FE,$65,$8C,$65,$FE,$19,$C1,$39,$F1,$35,$E2,$5C,$39,$BF
	.dc.b	$E6,$DB,$FE,$91,$87,$F2,$EC,$3F,$D7,$D5,$FE,$64,$9C,$BF,$55,$FC
	.dc.b	$77,$C6,$F1,$EC,$E5,$D9,$6B,$83,$D7,$C3,$C5,$FF,$8B,$4F,$F3,$74
	.dc.b	$F5,$AF,$16,$DF,$1F,$E3,$AC,$FC,$67,$CD,$BE,$2F,$93,$74,$FF,$53
	.dc.b	$78,$BE,$59,$81,$AF,$F1,$02,$99,$37,$F5,$5F,$C0,$7E,$B7,$6D,$8C
	.dc.b	$F8,$C9,$CB,$63,$8F,$D1,$7F,$A0,$5B,$87,$28,$C4,$79,$6D,$18,$8F
	.dc.b	$99,$0F,$A2,$F9,$D5,$FD,$82,$31,$3F,$E9,$66,$BF,$E5,$77,$9C,$7E
	.dc.b	$A7,$EB,$7F,$C6,$53,$FA,$B9,$FF,$37,$BC,$79,$3A,$ED,$E0,$6C,$6F
	.dc.b	$0A,$A9,$CB,$86,$C7,$8E,$37,$F5,$BF,$47,$2F,$47,$C7,$6C,$95,$B3
	.dc.b	$D5,$F3,$1C,$C8,$FB,$63,$C3,$ED,$7F,$AC,$E3,$FD,$6A,$D7,$FC,$EC
	.dc.b	$DA,$FF,$02,$FC,$BF,$F6,$69,$DF,$2F,$22,$3C,$B9,$FD,$BF,$5B,$F2
	.dc.b	$AF,$FC,$1A,$DE,$0F,$CF,$36,$BF,$8A,$60,$F9,$BE,$3C,$32,$7F,$D9
	.dc.b	$35,$F3,$9F,$F4,$53,$C5,$7D,$6F,$F4,$67,$FD,$9E,$D7,$FD,$20,$DC
	.dc.b	$7F,$C3,$EB,$E9,$7E,$9B,$FD,$42,$78,$FA,$38,$33,$E8,$EB,$C1,$93
	.dc.b	$D9,$E3,$F1,$BC,$BD,$59,$3E,$27,$6F,$8B,$E3,$7B,$91,$F3,$FC,$97
	.dc.b	$8B,$86,$47,$A9,$F2,$FC,$4F,$E2,$3E,$B7,$F9,$1C,$FF,$BB,$1A,$FF
	.dc.b	$AB,$5B,$BF,$F8,$F1,$FF,$11,$47,$FB,$CB,$E3,$3E,$73,$78,$FF,$27
	.dc.b	$18,$8F,$F5,$B3,$F7,$DB,$F2,$BA,$23,$CF,$DD,$E3,$D5,$23,$B3,$D3
	.dc.b	$B7,$F3,$FF,$5A,$DA,$92,$BF,$AA,$69,$FE,$52,$9F,$F1,$1C,$63,$6F
	.dc.b	$E4,$9F,$5B,$EA,$8F,$F6,$53,$6B,$FC,$B2,$D9,$DF,$9F,$FA,$EF,$AD
	.dc.b	$F5,$57,$F9,$79,$B3,$FC,$BD,$B7,$FD,$96,$FF,$5D,$39,$F1,$79,$12
	.dc.b	$7C,$87,$F6,$BA,$FF,$0E,$38,$0F,$EE,$9F,$F7,$03,$6F,$C5,$9F,$97
	.dc.b	$B3,$F8,$1B,$F0,$2B,$FD,$CC,$D9,$FF,$3A,$B7,$0F,$E8,$CF,$F5,$29
	.dc.b	$F0,$3F,$E6,$46,$FC,$1F,$43,$AF,$97,$FC,$43,$EB,$5E,$A5,$25,$7A
	.dc.b	$63,$FC,$88,$DF,$0B,$81,$1D,$79,$DF,$C6,$BE,$B5,$F0,$F4,$93,$F0
	.dc.b	$E3,$FC,$50,$DF,$37,$81,$1D,$7B,$DF,$C9,$3E,$B7,$F6,$50,$E0,$FF
	.dc.b	$4A,$C6,$1B,$E2,$6F,$C4,$E1,$F2,$72,$FE,$95,$A3,$A7,$BB,$FC,$C3
	.dc.b	$87,$FC,$EF,$5F,$5D,$AF,$7E,$4F,$EE,$51,$8C,$BF,$90,$FC,$37,$1E
	.dc.b	$E9,$7C,$5E,$91,$D7,$BE,$4F,$57,$6F,$53,$FC,$65,$7F,$11,$C0,$71
	.dc.b	$7F,$EB,$46,$E1,$FE,$05,$F7,$9F,$4D,$EB,$E3,$98,$7A,$DF,$77,$B5
	.dc.b	$EF,$ED,$93,$EF,$7F,$A8,$FC,$CB,$E7,$BE,$27,$AD,$2B,$E2,$3C,$1F
	.dc.b	$2E,$8F,$17,$E5,$12,$3F,$41,$7F,$CD,$76,$F4,$7A,$9F,$3C,$93,$F3
	.dc.b	$EB,$E9,$7C,$D4,$8F,$A4,$E9,$F8,$7F,$57,$FC,$D5,$FD,$63,$5C,$FA
	.dc.b	$5F,$DA,$BE,$B5,$F4,$BF,$DB,$2B,$F9,$F7,$27,$F8,$AD,$B7,$FE,$61
	.dc.b	$6A,$7A,$67,$FE,$E2,$7F,$59,$1E,$2F,$4F,$A1,$FD,$E3,$EB,$7E,$8C
	.dc.b	$79,$3D,$3F,$01,$1F,$3C,$F8,$FE,$7A,$47,$A9,$FF,$00,$FD,$3F,$F1
	.dc.b	$08,$F1,$7C,$F7,$A4,$8E,$CF,$83,$A7,$B4,$8F,$C4,$8F,$81,$DF,$23
	.dc.b	$E3,$BE,$7F,$A0,$F8,$9E,$9C,$9F,$96,$ED,$F9,$FE,$B7,$C9,$23,$FC
	.dc.b	$6F,$EB,$5F,$2F,$F3,$78,$A4,$FD,$47,$6F,$E6,$BE,$B5,$F3,$7F,$E1
	.dc.b	$E6,$C7,$F2,$56,$BF,$A0,$FA,$D7,$C4,$6F,$C4,$93,$B7,$95,$F3,$C6
	.dc.b	$4F,$37,$6F,$87,$E3,$BC,$04,$7F,$9A,$7D,$6B,$C9,$F2,$78,$F2,$7C
	.dc.b	$8E,$DF,$57,$D0,$7A,$48,$FF,$3D,$EA,$7C,$2F,$B7,$92,$4F,$B5,$DB
	.dc.b	$FB,$CF,$AD,$FF,$15,$71,$7F,$9F,$5B,$1F,$EB,$36,$AF,$86,$F3,$3F
	.dc.b	$D2,$E3,$BB,$6F,$FC,$93,$18,$8E,$DE,$B7,$F9,$33,$EB,$7F,$18,$4F
	.dc.b	$9D,$FF,$27,$E0,$F3,$F5,$F0,$73,$0F,$EB,$17,$FE,$01,$F1,$1A,$7F
	.dc.b	$A4,$E3,$9F,$9F,$D4,$95,$FD,$7F,$1F,$E1,$4F,$2F,$43,$C4,$47,$AC
	.dc.b	$F9,$7D,$E9,$1E,$07,$CB,$F3,$3E,$9F,$7E,$4F,$91,$DB,$FC,$9B,$1F
	.dc.b	$E2,$4F,$A3,$CD,$F8,$07,$BB,$0F,$85,$E7,$49,$F6,$BA,$7F,$BB,$7D
	.dc.b	$6F,$F4,$BD,$7F,$D6,$0D,$8F,$F7,$33,$5F,$A9,$A7,$F9,$E2,$1E,$45
	.dc.b	$24,$F4,$F6,$7E,$7F,$CC,$FF,$A8,$B8,$0D,$24,$FF,$AA,$2D,$F1,$3F
	.dc.b	$EA,$86,$C6,$63,$06,$7C,$DF,$11,$23,$D7,$1F,$E8,$CF,$63,$FC,$15
	.dc.b	$1F,$EB,$76,$BF,$EC,$D6,$A0,$E1,$7D,$6F,$E4,$9F,$F5,$FB,$5F,$F6
	.dc.b	$13,$5F,$F6,$C3,$5F,$93,$F1,$1E,$B2,$3E,$F7,$DD,$F1,$52,$3E,$13
	.dc.b	$5F,$F6,$6F,$C7,$7F,$04,$7C,$DC,$1F,$ED,$18,$C4,$7E,$58,$FF,$74
	.dc.b	$38,$8D,$79,$3F,$EE,$B7,$FD,$B1,$B7,$FA,$25,$BD,$19,$BF,$F7,$84
	.dc.b	$E5,$B1,$97,$F0,$7E,$8D,$5B,$FF,$50,$7C,$A7,$E2,$A3,$51,$98,$F3
	.dc.b	$1F,$3F,$4F,$D9,$3C,$0C,$76,$75,$8C,$C7,$03,$E9,$C7,$87,$67,$A3
	.dc.b	$49,$3C,$CF,$B6,$F8,$BD,$19,$19,$C3,$FC,$48,$D8,$CB,$DE,$32,$D8
	.dc.b	$CE,$9F,$F0,$06,$5E,$8F,$B8,$95,$D5,$F8,$3F,$28,$F9,$AF,$CF,$36
	.dc.b	$7F,$20,$7E,$C3,$DB,$FB,$6F,$EC,$A3,$4F,$E6,$7A,$7E,$F6,$34,$F7
	.dc.b	$69,$FB,$26,$C6,$A7,$F3,$DC,$B5,$FC,$F5,$7F,$A4,$6B,$8C,$E6,$7F
	.dc.b	$42,$DC,$FF,$90,$CF,$C7,$30,$43,$C3,$CD,$C6,$FF,$73,$18,$25,$FD
	.dc.b	$98,$73,$6C,$93,$EA,$7C,$5E,$27,$F6,$88,$FB,$4D,$AF,$C1,$36,$3E
	.dc.b	$4C,$23,$0F,$CC,$5E,$87,$42,$4F,$47,$D1,$F3,$5F,$8D,$AF,$8A,$D7
	.dc.b	$F8,$1B,$70,$39,$E4,$FF,$3F,$DF,$8A,$BF,$39,$DB,$92,$46,$63,$CD
	.dc.b	$E7,$F5,$C8,$D2,$4E,$E1,$F1,$9F,$BD,$F1,$BF,$53,$F5,$8F,$AD,$C5
	.dc.b	$FB,$EA,$E3,$FC,$6E,$B5,$3F,$EB,$3C,$1F,$37,$AA,$F9,$17,$67,$35
	.dc.b	$B8,$C7,$CC,$EF,$EB,$7F,$B4,$4F,$38,$C7,$6F,$1F,$8C,$DC,$F7,$6E
	.dc.b	$3F,$E7,$CC,$7E,$B4,$FF,$B1,$70,$FF,$38,$3E,$58,$ED,$F0,$BE,$3F
	.dc.b	$47,$C9,$16,$74,$7F,$B3,$7E,$B3,$FC,$34,$F3,$F3,$3D,$91,$8C,$7F
	.dc.b	$D1,$78,$FF,$9D,$EB,$86,$7A,$F5,$F6,$D8,$E9,$F1,$7F,$17,$EB,$7F
	.dc.b	$4D,$87,$FA,$53,$14,$E7,$BB,$BF,$C6,$F8,$22,$E1,$C8,$F3,$79,$1C
	.dc.b	$FB,$BC,$AD,$8F,$57,$A5,$86,$9C,$0F,$F7,$57,$C8,$FF,$09,$7F,$DD
	.dc.b	$9E,$03,$D8,$FB,$53,$9D,$7F,$D5,$33,$9A,$FF,$79,$7D,$6B,$E2,$37
	.dc.b	$E4,$FE,$06,$31,$3D,$A3,$1E,$2F,$E1,$7E,$29,$F1,$58,$AF,$F4,$EB
	.dc.b	$C7,$D3,$C6,$7F,$EB,$8C,$51,$8F,$0F,$F9,$FC,$7F,$00,$BF,$ED,$D1
	.dc.b	$FD,$0B,$A3,$FD,$A9,$1E,$6E,$2F,$CA,$46,$30,$E9,$F0,$FE,$79,$F4
	.dc.b	$36,$BD,$BE,$86,$1A,$7C,$08,$F7,$7C,$6F,$4F,$42,$CF,$88,$FE,$63
	.dc.b	$8F,$F3,$A7,$E3,$FA,$3A,$CB,$D8,$DC,$FD,$7C,$9F,$A9,$FA,$87,$CD
	.dc.b	$7C,$B8,$C3,$F3,$7E,$2B,$18,$FD,$26,$E3,$FE,$DE,$FA,$97,$D5,$7E
	.dc.b	$CF,$EB,$7C,$53,$98,$C6,$DF,$B7,$AF,$E4,$5A,$7F,$6A,$78,$E3,$EB
	.dc.b	$BF,$F0,$D4,$6A,$3F,$A6,$BA,$8F,$07,$57,$3F,$8A,$E8,$DA,$E5,$2A
	.dc.b	$7C,$BF,$F0,$7F,$AD,$FE,$87,$3E,$98,$C5,$79,$7D,$2F,$34,$58,$E7
	.dc.b	$7C,$9B,$F4,$7F,$27,$EB,$7D,$8D,$8D,$FC,$5E,$1A,$70,$39,$F1,$7D
	.dc.b	$1E,$0F,$88,$D8,$EF,$E8,$E1,$A7,$03,$C7,$E8,$FC,$B8,$F9,$8D,$8F
	.dc.b	$3F,$C0,$7D,$DE,$1F,$A0,$D8,$F5,$EF,$86,$9C,$0F,$93,$F5,$BF,$C4
	.dc.b	$78,$FF,$C3,$FD,$1F,$D4,$A7,$3D,$5F,$89,$E5,$FE,$65,$7F,$CB,$30
	.dc.b	$61,$FE,$26,$8E,$57,$FE,$06,$75,$19,$BE,$3F,$0D,$F9,$3F,$8F,$B7
	.dc.b	$F6,$B2,$FF,$6B,$BE,$4F,$6B,$39,$76,$8E,$91,$8C,$7C,$1F,$0B,$E4
	.dc.b	$8F,$2B,$67,$F0,$BF,$5A,$ED,$FE,$94,$BB,$AC,$61,$FE,$9D,$C1,$FD
	.dc.b	$33,$2F,$01,$FE,$6F,$7F,$AE,$DF,$E4,$1D,$65,$FE,$46,$73,$3C,$07
	.dc.b	$D6,$BE,$37,$AC,$7F,$5B,$8C,$5F,$38,$C7,$06,$7E,$06,$CF,$E7,$AB
	.dc.b	$FC,$FE,$DF,$3F,$97,$A1,$CD,$B1,$FA,$4F,$A7,$7C,$B7,$F6,$97,$F8
	.dc.b	$FC,$66,$31,$E6,$67,$C5,$D7,$8F,$6B,$93,$FD,$C7,$EB,$7F,$49,$B7
	.dc.b	$B5,$7D,$56,$7C,$6D,$3B,$F7,$78,$78,$17,$1F,$DB,$DA,$3A,$61,$D7
	.dc.b	$92,$D6,$9F,$F0,$02,$F5,$EF,$AF,$86,$C7,$F0,$2F,$AD,$7C,$77,$F8
	.dc.b	$1D,$7C,$FD,$8C,$3F,$E0,$BF,$AD,$F4,$D7,$F0,$A7,$F2,$39,$7D,$97
	.dc.b	$2D,$F0,$B8,$69,$C7,$C7,$B7,$C5,$FF,$0B,$7D,$6B,$C5,$BF,$29,$86
	.dc.b	$D7,$F1,$89,$FE,$EA,$DF,$13,$8F,$07,$27,$FC,$41,$ED,$FA,$A7,$F8
	.dc.b	$FB,$FD,$8B,$4F,$C7,$39,$6F,$B9,$C3,$D5,$E3,$F1,$76,$F8,$BF,$E2
	.dc.b	$BF,$AD,$F7,$F8,$DF,$E9,$E7,$F8,$06,$1F,$99,$73,$5F,$E8,$07,$33
	.dc.b	$BE,$BE,$5D,$7C,$DB,$5F,$CD,$2B,$FC,$8E,$D8,$F0,$7A,$BB,$8D,$7C
	.dc.b	$4F,$8F,$E2,$7C,$8D,$8E,$5E,$AE,$1A,$70,$3B,$7C,$47,$F9,$17,$93
	.dc.b	$E1,$69,$EB,$8C,$75,$7B,$FB,$BE,$04,$5B,$E2,$7F,$4A,$A7,$8F,$CF
	.dc.b	$C4,$FF,$28,$46,$27,$FA,$7E,$3F,$AA,$E1,$FD,$F4,$BC,$7D,$E1,$C1
	.dc.b	$B7,$FD,$5A,$9D,$7E,$7F,$8F,$FF,$2E,$46,$27,$AC,$63,$B7,$35,$67
	.dc.b	$93,$C3,$C4,$C5,$FF,$0F,$3F,$CA,$E5,$AE,$BA,$FC,$5F,$90,$B8,$7A
	.dc.b	$9F,$0E,$F8,$BE,$4D,$D7,$FD,$1D,$E4,$FF,$A0,$E4,$1C,$34,$D3,$FE
	.dc.b	$89,$FA,$BF,$DC,$7D,$6D,$F1,$CD,$25,$7F,$A4,$E5,$F4,$7F,$D3,$23
	.dc.b	$0D,$7F,$D5,$31,$AC,$3B,$4E,$67,$FD,$AB,$2F,$A3,$FF,$6B,$D7,$FB
	.dc.b	$EA,$31,$3F,$E0,$86,$BF,$CD,$DE,$8B,$EB,$7F,$C4,$EF,$C0,$8F,$C2
	.dc.b	$BC,$79,$3A,$ED,$E0,$6C,$6F,$0A,$AB,$F0,$CF,$87,$93,$31,$88,$FE
	.dc.b	$91,$F3,$1F,$EC,$6C,$BD,$1F,$21,$A2,$4F,$13,$D5,$F3,$3C,$88,$FB
	.dc.b	$63,$C3,$EF,$7F,$B3,$E3,$FC,$4E,$D7,$FA,$B3,$69,$A7,$FB,$FB,$E6
	.dc.b	$7F,$AC,$57,$0E,$5E,$64,$79,$F3,$FB,$BE,$B7,$E5,$E1,$CA,$7F,$0F
	.dc.b	$18,$C3,$F1,$30,$8C,$3F,$8C,$C1,$F3,$9C,$F8,$64,$FF,$8F,$AB,$E8
	.dc.b	$BF,$DB,$5E,$2B,$EB,$7F,$9F,$BF,$E4,$46,$BF,$DC,$DB,$8F,$96,$38
	.dc.b	$79,$3F,$5A,$7F,$A4,$C7,$1F,$47,$06,$7C,$1D,$78,$32,$7B,$3C,$7E
	.dc.b	$3F,$97,$AF,$27,$C4,$ED,$F1,$7C,$67,$72,$3E,$7F,$91,$F1,$70,$C8
	.dc.b	$F5,$3E,$5F,$7F,$F1,$7F,$5A,$F2,$29,$2B,$FC,$C7,$1F,$EE,$87,$FC
	.dc.b	$CB,$ED,$FC,$C6,$F2,$7E,$52,$31,$1F,$E1,$77,$EF,$B7,$E5,$74,$47
	.dc.b	$9F,$BF,$C7,$AE,$47,$67,$A7,$7F,$E8,$3E,$B5,$B5,$25,$7F,$9E,$23
	.dc.b	$FC,$2A,$FF,$9E,$BE,$B7,$D0,$FF,$8A,$5B,$D1,$F5,$DB,$C1,$F8,$A5
	.dc.b	$E7,$FE,$BB,$EB,$7D,$35,$F7,$5B,$5F,$90,$6D,$FF,$1A,$BF,$D7,$4E
	.dc.b	$7C,$6E,$44,$9F,$F4,$6D,$7F,$84,$5B,$9F,$72,$3A,$ED,$FD,$FC,$7B
	.dc.b	$23,$FC,$84,$DC,$7F,$20,$DC,$3F,$46,$FF,$B4,$A7,$FD,$36,$FF,$95
	.dc.b	$5B,$F0,$9D,$0E,$BE,$67,$F0,$CF,$AD,$FF,$4F,$CF,$FA,$89,$AF,$F9
	.dc.b	$45,$B8,$7E,$7D,$FF,$1B,$C6,$FC,$2E,$44,$79,$79,$DF,$C6,$3E,$B5
	.dc.b	$F1,$14,$93,$FE,$B0,$8F,$F0,$F3,$7C,$EE,$04,$75,$EE,$FF,$22,$FA
	.dc.b	$DF,$D6,$C7,$F9,$95,$B9,$FE,$CD,$BB,$FF,$06,$7F,$6E,$DF,$85,$E4
	.dc.b	$4A,$FF,$60,$CB,$DD,$FE,$5F,$BB,$E1,$E9,$2B,$FD,$8F,$E5,$E5,$FF
	.dc.b	$64,$CB,$B3,$89,$23,$35,$D3,$B6,$47,$07,$4F,$F9,$FE,$BE,$1B,$7D
	.dc.b	$8F,$19,$FD,$0B,$E3,$1F,$4F,$E6,$E3,$95,$E2,$76,$F9,$3A,$FD,$64
	.dc.b	$7F,$DB,$FF,$32,$F9,$EF,$7F,$AD,$27,$D6,$ED,$F2,$78,$7F,$10,$8F
	.dc.b	$2E,$EF,$8F,$EA,$48,$F8,$6E,$9F,$67,$CD,$7C,$C2,$3A,$3F,$DE,$3F
	.dc.b	$58,$F4,$FF,$AF,$B6,$33,$1F,$D8,$BE,$B5,$F4,$DA,$F4,$E5,$72,$78
	.dc.b	$D7,$2F,$4E,$4F,$67,$FB,$5F,$D4,$FC,$A8,$F0,$FA,$7E,$02,$3B,$3C
	.dc.b	$BF,$39,$E4,$23,$FE,$1A,$F0,$7F,$A1,$7D,$3F,$F0,$48,$6A,$F9,$EF
	.dc.b	$59,$23,$99,$F0,$35,$F7,$91,$F8,$71,$F0,$BC,$29,$1F,$19,$F4,$FD
	.dc.b	$07,$C6,$F4,$A4,$FC,$A7,$6F,$FB,$27,$0F,$F6,$C3,$F3,$BD,$6F,$9A
	.dc.b	$7F,$83,$3E,$B5,$F2,$3F,$43,$C8,$95,$F5,$1D,$BE,$F7,$FB,$8F,$D6
	.dc.b	$BE,$6F,$FC,$2A,$D9,$D1,$FE,$3E,$70,$FF,$B6,$FE,$B5,$F1,$1C,$3C
	.dc.b	$49,$3C,$05,$23,$AD,$F6,$C6,$56,$EB,$D2,$DA,$F1,$FC,$64,$8D,$3F
	.dc.b	$E2,$AF,$AD,$FD,$B4,$71,$F1,$BD,$24,$76,$75,$F9,$FF,$69,$1F,$37
	.dc.b	$53,$E1,$7E,$0F,$24,$9F,$23,$B7,$FD,$F5,$F5,$BF,$E2,$5E,$1F,$F0
	.dc.b	$CB,$63,$FC,$C6,$D5,$E7,$F9,$9E,$37,$FE,$42,$9F,$04,$7C,$76,$BD
	.dc.b	$BD,$6F,$F2,$3F,$D6,$FE,$4C,$9F,$3B,$FE,$4A,$C1,$E8,$6B,$E0,$66
	.dc.b	$1F,$46,$FF,$BD,$F8,$1F,$E1,$D7,$97,$A1,$D0,$93,$FC,$8B,$CD,$EF
	.dc.b	$E8,$48,$E4,$E9,$5F,$3F,$33,$CA,$48,$FE,$57,$F3,$3E,$8F,$76,$47
	.dc.b	$8D,$D3,$CB,$8B,$FC,$83,$9B,$CF,$F7,$92,$7F,$9A,$E9,$F0,$BC,$F9
	.dc.b	$1E,$D7,$4F,$33,$FD,$27,$F5,$BF,$D2,$1B,$F1,$1F,$E7,$08,$C4,$7D
	.dc.b	$4D,$B1,$E5,$48,$FC,$C4,$A3,$85,$7F,$9E,$3E,$31,$EF,$52,$5F,$E7
	.dc.b	$E3,$FD,$76,$D8,$E1,$CF,$23,$06,$DF,$F5,$54,$FF,$9C,$63,$FC,$C6
	.dc.b	$1C,$3F,$E6,$66,$BF,$E6,$5D,$3F,$D6,$5C,$DF,$26,$38,$99,$29,$F9
	.dc.b	$28,$DF,$A7,$23,$32,$8C,$47,$28,$7D,$8F,$62,$47,$07,$D3,$D7,$23
	.dc.b	$BE,$1F,$9A,$7D,$6D,$7F,$CF,$FA,$BE,$27,$FD,$94,$D8,$F3,$B5,$F5
	.dc.b	$B5,$F8,$AB,$D0,$FF,$44,$F2,$61,$B3,$D9,$AF,$E1,$1A,$A7,$F9,$E3
	.dc.b	$FE,$DA,$9E,$BE,$7E,$A7,$33,$FE,$85,$38,$46,$7D,$09,$1C,$4F,$6B
	.dc.b	$87,$84,$BE,$63,$A1,$23,$81,$F9,$87,$C1,$EE,$C8,$C9,$E7,$70,$F4
	.dc.b	$6E,$1F,$11,$B1,$FE,$B7,$F9,$9F,$7E,$E6,$92,$3A,$FB,$BF,$DD,$5F
	.dc.b	$73,$6F,$F2,$6D,$8F,$F5,$BB,$5F,$D7,$FC,$AF,$57,$EB,$B5,$FA,$FE
	.dc.b	$F7,$F4,$67,$37,$F7,$BE,$17,$FD,$89,$87,$DC,$6C,$FE,$19,$AF,$EF
	.dc.b	$1A,$FD,$9F,$52,$EA,$69,$2B,$E4,$C7,$FB,$69,$A7,$02,$1F,$3B,$9D
	.dc.b	$F4,$9D,$9B,$19,$8C,$19,$FA,$49,$1C,$5B,$87,$CA,$E3,$F3,$ED,$F9
	.dc.b	$7A,$F9,$F1,$83,$3E,$C7,$E3,$68,$7F,$95,$DC,$36,$7F,$77,$49,$CF
	.dc.b	$D6,$7F,$79,$9F,$DF,$B6,$BF,$81,$B5,$FE,$1A,$D7,$FD,$E2,$FF,$82
	.dc.b	$23,$8F,$A9,$D0,$8F,$F0,$AC,$FF,$D5,$AF,$F7,$D9,$7D,$6F,$FB,$F7
	.dc.b	$C8,$FF,$75,$56,$B0,$FE,$68,$2B,$77,$FE,$05,$97,$CD,$71,$8F,$83
	.dc.b	$49,$E5,$7D,$2B,$53,$F1,$A7,$35,$C1,$7C,$D6,$A5,$F5,$BF,$D9,$AB
	.dc.b	$C1,$39,$8C,$4F,$9A,$BF,$9E,$D6,$8F,$37,$85,$CE,$3A,$C7,$A5,$E5
	.dc.b	$E6,$65,$B1,$ED,$F0,$30,$D3,$81,$FE,$25,$FA,$CF,$F4,$73,$FE,$27
	.dc.b	$C7,$FB,$13,$9F,$88,$F3,$78,$BE,$BA,$F3,$EC,$AF,$8F,$1C,$B4,$F9
	.dc.b	$EE,$A3,$37,$C7,$E2,$BE,$27,$D6,$FF,$BC,$A7,$C5,$18,$F9,$3E,$DF
	.dc.b	$1B,$B9,$F3,$37,$1E,$BE,$97,$2F,$1F,$C8,$DA,$F6,$7A,$38,$69,$EF
	.dc.b	$8F,$A9,$E0,$7F,$13,$8F,$F2,$37,$BA,$F6,$33,$E1,$FD,$98,$C4,$FF
	.dc.b	$92,$FE,$B7,$FD,$18,$FF,$93,$6F,$96,$7F,$8E,$9C,$BF,$90,$8C,$F1
	.dc.b	$7C,$78,$FD,$6C,$BE,$5F,$F0,$DF,$1B,$FD,$87,$0E,$D1,$8F,$43,$F1
	.dc.b	$54,$F8,$9B,$38,$6D,$79,$78,$3F,$21,$CA,$F8,$8E,$31,$87,$9F,$8B
	.dc.b	$8A,$F6,$37,$4F,$57,$A1,$A9,$F8,$0D,$8F,$77,$9D,$DB,$C6,$F2,$B6
	.dc.b	$BE,$36,$18,$69,$F2,$63,$D7,$C3,$FA,$5F,$A9,$7C,$DE,$2B,$FC,$F1
	.dc.b	$18,$7E,$87,$9B,$9C,$BE,$73,$79,$3E,$64,$3E,$AD,$E6,$7E,$C7,$EB
	.dc.b	$7F,$8B,$61,$FE,$83,$9C,$BF,$E8,$58,$CD,$7F,$06,$9F,$E3,$17,$D1
	.dc.b	$DC,$67,$E2,$3F,$79,$18,$9D,$74,$E2,$72,$DC,$3C,$1F,$45,$FC,$0F
	.dc.b	$EB,$7F,$A1,$63,$DE,$31,$3D,$3E,$27,$AC,$58,$E2,$7A,$7E,$77,$E4
	.dc.b	$7A,$7E,$26,$C7,$B7,$E2,$30,$D3,$81,$F0,$78,$BE,$87,$07,$A9,$B1
	.dc.b	$F1,$7C,$FC,$34,$E0,$7C,$9F,$3F,$E0,$C7,$C2,$6C,$7C,$DF,$8C,$FB
	.dc.b	$9C,$1F,$19,$B1,$F4,$76,$C3,$4E,$07,$FA,$DF,$EB,$7F,$85,$71,$7D
	.dc.b	$CE,$6F,$F5,$C4,$E7,$BB,$F1,$1C,$9F,$CB,$6F,$F8,$76,$1F,$D6,$1E
	.dc.b	$35,$FC,$E9,$CC,$1C,$BB,$F8,$3F,$C9,$7A,$DC,$DB,$5E,$3F,$82,$C3
	.dc.b	$4F,$34,$78,$7E,$9F,$C5,$36,$75,$7F,$D4,$3E,$B7,$FC,$09,$1F,$ED
	.dc.b	$35,$DB,$FD,$15,$CC,$EE,$3F,$A4,$6F,$FD,$52,$BF,$DB,$4F,$F4,$E9
	.dc.b	$FE,$EA,$E5,$FE,$BA,$E5,$7D,$6B,$E2,$F9,$C7,$F5,$38,$C5,$74,$8C
	.dc.b	$6F,$9F,$72,$CF,$FB,$A6,$91,$C7,$CF,$E5,$E7,$DA,$D3,$FD,$7F,$E9
	.dc.b	$DF,$2B,$7C,$BC,$36,$BF,$DE,$93,$FB,$0E,$7F,$DC,$BF,$C5,$E3,$3E
	.dc.b	$47,$3F,$23,$FB,$57,$D6,$BC,$3B,$F5,$58,$6D,$FF,$BF,$E7,$FB,$C3
	.dc.b	$7B,$B8,$FB,$9C,$9E,$7C,$7D,$39,$6D,$71,$7F,$BC,$2F,$56,$FA,$F8
	.dc.b	$6C,$7F,$89,$B7,$FE,$23,$F5,$AF,$8A,$FF,$00,$2F,$9A,$B1,$87,$FC
	.dc.b	$09,$F5,$AE,$0B,$E0,$E1,$B8,$F5,$3F,$C2,$73,$FE,$04,$D7,$FC,$36
	.dc.b	$FF,$80,$9B,$B7,$4F,$1B,$FC,$21,$F5,$AF,$0E,$FC,$8E,$1B,$8F,$5C
	.dc.b	$3F,$C7,$F4,$E1,$E3,$E2,$72,$6D,$6E,$BF,$CF,$5B,$7F,$87,$7D,$BF
	.dc.b	$EB,$FC,$5F,$CC,$9F,$81,$5B,$72,$FF,$8B,$17,$A9,$C7,$E2,$F9,$36
	.dc.b	$7B,$7D,$6F,$F9,$0F,$1F,$F1,$CB,$FC,$09,$BC,$7F,$E2,$D8,$C6,$3F
	.dc.b	$D0,$9F,$C3,$57,$2E,$3D,$75,$6D,$E1,$E9,$62,$7B,$B7,$1E,$FF,$0D
	.dc.b	$D7,$E9,$BB,$36,$BC,$BE,$9E,$1A,$7A,$23,$C9,$F0,$DF,$E4,$1E,$5F
	.dc.b	$87,$87,$AA,$31,$C5,$E9,$ED,$DE,$1F,$01,$B5,$FE,$58,$DD,$E2,$F3
	.dc.b	$F1,$7F,$C9,$91,$8B,$FA,$99,$7F,$81,$5F,$AD,$0F,$1B,$77,$C1,$B3
	.dc.b	$F6,$F0,$FC,$55,$7F,$6A,$B7,$C6,$7F,$96,$67,$35,$FE,$64,$9C,$BC
	.dc.b	$7B,$75,$DB,$B6,$C7,$5F,$0B,$0D,$38,$1E,$1D,$FE,$37,$C6,$77,$6C
	.dc.b	$79,$77,$C3,$4E,$07,$A7,$27,$8F,$FE,$78,$61,$82,$9F,$E9,$C6,$B4
	.dc.b	$DF,$D5,$FE,$FF,$EB,$76,$D8,$CF,$8C,$9C,$B6,$38,$FD,$1B,$87,$8B
	.dc.b	$63,$31,$83,$AF,$D1,$FF,$A3,$9F,$F0,$CE,$08,$C6,$5F,$E9,$28,$C5
	.dc.b	$7E,$C5,$5F,$5F,$AD,$FF,$14,$CF,$FA,$E2,$7F,$0E,$EF,$97,$9B,$96
	.dc.b	$A8,$EA,$FF,$A9,$9F,$C4,$3D,$F9,$77,$18,$8F,$F5,$5F,$CA,$7C,$FC
	.dc.b	$FD,$3F,$1B,$A2,$4F,$13,$D1,$F2,$FC,$88,$FA,$E3,$C7,$EE,$7F,$AB
	.dc.b	$E3,$F8,$73,$5F,$F5,$CB,$6B,$FC,$7F,$F2,$9F,$D7,$67,$7C,$9C,$88
	.dc.b	$F2,$E6,$F7,$FD,$6F,$BA,$FF,$C8,$ED,$E1,$FF,$5F,$B6,$BF,$DC,$38
	.dc.b	$3E,$6B,$8F,$04,$9F,$E4,$55,$F3,$9F,$F6,$6B,$C5,$7D,$6F,$F4,$17
	.dc.b	$F9,$43,$5F,$F6,$9B,$71,$FF,$2C,$AF,$A5,$FA,$AF,$F4,$B9,$E3,$E8
	.dc.b	$E0,$CF,$83,$AF,$06,$4F,$67,$8F,$C6,$72,$F5,$E4,$F8,$9D,$BE,$2F
	.dc.b	$8C,$EE,$47,$CF,$F2,$3E,$2E,$19,$1E,$A7,$CB,$F1,$7F,$8A,$FA,$D7
	.dc.b	$93,$49,$5F,$D1,$E3,$FD,$DA,$FF,$49,$F6,$FE,$63,$79,$3F,$27,$18
	.dc.b	$8F,$F7,$BB,$F7,$5B,$F3,$DA,$23,$CF,$DF,$E3,$D7,$23,$B3,$D3,$BF
	.dc.b	$F3,$FF,$5A,$DA,$92,$BF,$AE,$47,$F8,$61,$FE,$BD,$F5,$BE,$87,$F7
	.dc.b	$9B,$99,$19,$E8,$FF,$48,$B6,$75,$28,$79,$FF,$B0,$FA,$DF,$65,$FF
	.dc.b	$A6,$1B,$7C,$1B,$7D,$8E,$1E,$2C,$7F,$B9,$47,$F0,$26,$F8,$FD,$74
	.dc.b	$91,$A7,$C1,$DD,$FC,$01,$F7,$47,$B0,$F6,$38,$C7,$FD,$56,$DC,$7D
	.dc.b	$6F,$FB,$46,$7F,$C0,$4F,$F9,$69,$BF,$09,$D4,$ED,$E6,$7F,$0E,$FA
	.dc.b	$DF,$F1,$C4,$FF,$82,$5A,$F8,$43,$8C,$3F,$D7,$6F,$F8,$96,$38,$78
	.dc.b	$7C,$C8,$F3,$F3,$BF,$8D,$FD,$6B,$E2,$29,$27,$FC,$31,$1F,$E2,$36
	.dc.b	$ED,$C0,$8E,$BD,$DF,$E4,$BF,$5B,$FB,$18,$F3,$37,$3F,$F6,$7B,$77
	.dc.b	$FB,$EF,$EE,$DB,$F0,$BC,$89,$5F,$E2,$99,$FF,$10,$BF,$E2,$BD,$BF
	.dc.b	$CD,$AD,$F8,$6F,$E6,$B1,$88,$F8,$DE,$55,$C7,$FD,$BF,$2E,$BD,$92
	.dc.b	$B9,$1D,$7B,$39,$91,$F0,$47,$6F,$5A,$47,$8D,$FE,$8B,$F1,$6F,$A7
	.dc.b	$F1,$F1,$4A,$F4,$3C,$1F,$B5,$8F,$E4,$DB,$D5,$EA,$3F,$A7,$FC,$CB
	.dc.b	$E7,$BD,$BE,$B4,$AF,$84,$ED,$F8,$9E,$17,$C2,$23,$FE,$51,$AF,$E6
	.dc.b	$1B,$FC,$2F,$4F,$E3,$92,$78,$9F,$37,$E6,$BE,$59,$1F,$A1,$1F,$D8
	.dc.b	$BE,$B3,$F9,$84,$7F,$97,$5A,$FE,$55,$B4,$E7,$FD,$EF,$D6,$BE,$93
	.dc.b	$97,$A3,$30,$E4,$F5,$F5,$3B,$7A,$52,$7B,$3F,$DC,$BE,$A7,$E7,$47
	.dc.b	$87,$D2,$F2,$11,$D9,$E5,$F9,$DF,$41,$1F,$F3,$A7,$D3,$FF,$08,$5B
	.dc.b	$3E,$7B,$DC,$48,$E4,$7B,$F4,$F8,$44,$7E,$14,$7C,$5F,$06,$47,$C5
	.dc.b	$7D,$1F,$41,$F2,$7D,$19,$3F,$25,$DB,$F2,$7D,$5F,$9C,$47,$FD,$15
	.dc.b	$F5,$AF,$8B,$FA,$5C,$92,$7E,$93,$B7,$F8,$37,$D6,$3E,$6F,$FC,$32
	.dc.b	$D8,$FF,$4A,$35,$FE,$19,$F5,$AF,$87,$DF,$89,$27,$6F,$2B,$E7,$8C
	.dc.b	$9E,$6E,$DF,$0F,$C5,$F8,$08,$FF,$A8,$FE,$B7,$F6,$4F,$8B,$C6,$F2
	.dc.b	$91,$F2,$CF,$A3,$E7,$E4,$7A,$1F,$C1,$7C,$27,$B3,$8A,$4F,$B1,$DB
	.dc.b	$FC,$9F,$EB,$7F,$C4,$5C,$1F,$D3,$9B,$1F,$D5,$5A,$BC,$DF,$2B,$C6
	.dc.b	$FF,$C8,$32,$69,$8F,$53,$FC,$8B,$F5,$BF,$8B,$27,$CE,$FF,$92,$30
	.dc.b	$7A,$1A,$F8,$09,$97,$FB,$16,$FF,$BD,$78,$DF,$A7,$8F,$ED,$11,$CF
	.dc.b	$D0,$EA,$48,$FF,$66,$61,$FC,$15,$E5,$E8,$78,$88,$F5,$9F,$2F,$B7
	.dc.b	$23,$C0,$F9,$7E,$67,$D3,$EE,$49,$F2,$3B,$7F,$A8,$62,$F3,$FD,$E4
	.dc.b	$8F,$1E,$1F,$0B,$CD,$91,$ED,$74,$FF,$59,$FA,$CF,$E8,$4F,$F8,$15
	.dc.b	$A9,$BA,$7F,$55,$8F,$F7,$54,$3C,$7F,$F3,$9B,$63,$31,$8F,$87,$FF
	.dc.b	$3B,$FC,$5B,$E4,$3F,$D3,$2D,$AF,$F4,$EB,$5F,$F5,$0B,$5F,$F5,$22
	.dc.b	$F7,$B8,$92,$38,$B5,$FF,$3E,$F3,$78,$0F,$0C,$7E,$07,$3F,$F5,$4E
	.dc.b	$66,$92,$BF,$D5,$F1,$F3,$5B,$AF,$02,$3A,$F4,$F9,$91,$C1,$CB,$EB
	.dc.b	$7B,$91,$C1,$FE,$BC,$DB,$FC,$78,$DF,$86,$FF,$5E,$C6,$23,$FD,$86
	.dc.b	$D7,$FD,$8C,$BE,$27,$64,$8E,$8F,$2F,$43,$FC,$77,$DD,$C1,$B3,$FE
	.dc.b	$3E,$A4,$62,$3F,$04,$D7,$F6,$5C,$F7,$4F,$F4,$45,$FF,$B5,$DB,$3F
	.dc.b	$E4,$86,$BF,$8B,$F3,$F9,$61,$FD,$3E,$3F,$CA,$55,$EA,$9C,$9A,$F9
	.dc.b	$89,$3A,$3C,$CE,$1F,$DA,$1F,$97,$71,$3F,$11,$B1,$ED,$F9,$8F,$09
	.dc.b	$1E,$87,$E7,$DF,$67,$CC,$C8,$C9,$FC,$45,$C3,$E0,$3F,$49,$E8,$7B
	.dc.b	$08,$F0,$C5,$1F,$03,$E5,$55,$F7,$DB,$5F,$73,$DF,$FE,$7C,$FD,$E8
	.dc.b	$CB,$F7,$BE,$1F,$F8,$96,$74,$EC,$E2,$3E,$D3,$5F,$F3,$F7,$C1,$7F
	.dc.b	$A1,$E7,$D6,$D8,$F5,$36,$3F,$CF,$F0,$8C,$3F,$85,$8E,$8F,$F0,$36
	.dc.b	$E5,$FE,$0B,$7E,$01,$AE,$87,$D2,$52,$47,$CC,$8C,$B7,$E9,$38,$11
	.dc.b	$C1,$E8,$F5,$FE,$86,$1E,$C6,$BE,$76,$E2,$6E,$D2,$BD,$B5,$F1,$23
	.dc.b	$0E,$3E,$B1,$E9,$F4,$6D,$70,$94,$61,$E1,$3E,$2C,$BF,$D4,$B5,$FE
	.dc.b	$AD,$F1,$FF,$2B,$6F,$AD,$FE,$95,$1F,$98,$AD,$E1,$FE,$89,$AD,$5F
	.dc.b	$13,$FD,$63,$2F,$9A,$E3,$18,$F8,$EB,$7F,$EB,$BA,$D4,$F1,$8C,$4E
	.dc.b	$94,$FF,$31,$9D,$43,$EB,$7F,$B1,$C6,$27,$53,$FE,$9D,$9C,$BF,$E9
	.dc.b	$F8,$FE,$65,$3B,$F1,$3B,$46,$27,$51,$88,$F1,$57,$F9,$A6,$BB,$46
	.dc.b	$3E,$1F,$8F,$A1,$B9,$F1,$37,$0F,$F5,$77,$D6,$FF,$A1,$E7,$D3,$87
	.dc.b	$F6,$17,$D7,$19,$9F,$47,$89,$F0,$63,$D3,$B7,$F0,$87,$4D,$F4,$35
	.dc.b	$F1,$5B,$F8,$CF,$91,$F5,$BF,$A7,$AE,$11,$8F,$1B,$A7,$C7,$6F,$4F
	.dc.b	$0B,$74,$F3,$F8,$BC,$BC,$7F,$13,$6B,$CF,$E8,$E1,$A7,$AA,$3E,$C7
	.dc.b	$81,$FC,$22,$3E,$BF,$53,$D7,$CF,$89,$F7,$63,$13,$F7,$3E,$B5,$F1
	.dc.b	$3A,$8C,$37,$4F,$F6,$C4,$BE,$5F,$F0,$DF,$1A,$F3,$FF,$11,$39,$C3
	.dc.b	$FD,$31,$18,$F8,$BC,$DD,$9F,$03,$C3,$83,$F2,$1C,$DF,$B3,$9F,$24
	.dc.b	$63,$C4,$EB,$E7,$EE,$7D,$0D,$D3,$CB,$F1,$1A,$9F,$63,$63,$D3,$E6
	.dc.b	$F6,$F1,$FC,$ED,$AF,$85,$86,$1A,$7C,$58,$F6,$F2,$7E,$97,$EA,$9F
	.dc.b	$37,$8A,$FD,$2C,$61,$F9,$9F,$15,$9C,$BE,$4B,$7B,$7F,$A9,$FD,$63
	.dc.b	$CD,$FD,$87,$D6,$FF,$42,$C3,$F5,$F3,$97,$F6,$31,$9C,$17,$CB,$E2
	.dc.b	$3F,$71,$7B,$9C,$D9,$91,$FE,$8B,$8E,$B8,$70,$EA,$E3,$76,$7C,$2F
	.dc.b	$4F,$A3,$FE,$09,$F5,$BF,$CE,$71,$EA,$74,$8C,$D7,$8B,$E1,$F8,$3E
	.dc.b	$A6,$CF,$01,$93,$E7,$7E,$67,$A9,$EA,$6D,$FC,$2F,$88,$C3,$4F,$8B
	.dc.b	$1F,$03,$8B,$EA,$6D,$67,$E9,$BF,$13,$D0,$C6,$5E,$1C,$BF,$8B,$F0
	.dc.b	$7D,$2C,$3E,$4F,$BD,$BC,$7E,$8B,$6F,$87,$C6,$7E,$07,$87,$E7,$B6
	.dc.b	$BE,$BC,$E1,$A7,$DA,$8F,$E5,$9F,$5B,$FC,$2A,$3E,$A4,$B9,$BF,$A3
	.dc.b	$65,$FE,$16,$AD,$4F,$D9,$9C,$D7,$09,$FE,$2A,$DE,$EE,$BE,$6F,$67
	.dc.b	$BF,$C2,$7E,$5F,$E3,$6D,$7E,$59,$F1,$FC,$46,$32,$FA,$98,$A3,$19
	.dc.b	$F8,$FE,$9F,$78,$7F,$42,$A4,$62,$FE,$E7,$D6,$BB,$7F,$A5,$2E,$EB
	.dc.b	$18,$7F,$A7,$70,$7F,$52,$AF,$C2,$3F,$D7,$B2,$FF,$1B,$B9,$6F,$B3
	.dc.b	$FD,$57,$EB,$5F,$1B,$CA,$F0,$DD,$7F,$AB,$CF,$CA,$93,$FE,$17,$8C
	.dc.b	$F0,$F3,$F7,$FA,$7A,$5D,$7D,$2B,$5C,$9F,$EC,$9F,$4E,$F9,$6B,$E7
	.dc.b	$61,$B5,$FD,$9A,$7F,$6D,$B7,$F9,$41,$FE,$39,$A7,$E8,$5C,$CF,$E0
	.dc.b	$9C,$CF,$F9,$5B,$EB,$5E,$2F,$F7,$35,$F5,$36,$30,$E7,$E0,$35,$DF
	.dc.b	$71,$FE,$EF,$7F,$DB,$70,$E9,$97,$3E,$4E,$8D,$AF,$EF,$A3,$D8,$FF
	.dc.b	$01,$2F,$6A,$C6,$1F,$F0,$37,$D6,$BE,$3A,$FC,$DE,$1B,$7F,$E0,$8F
	.dc.b	$AD,$F4,$B7,$C1,$FF,$07,$46,$32,$FE,$08,$A3,$3B,$6B,$C8,$B7,$CD
	.dc.b	$FF,$0B,$7D,$6B,$C4,$BF,$29,$86,$D7,$F8,$62,$7D,$4D,$F0,$F8,$EF
	.dc.b	$C9,$FF,$10,$7B,$CD,$EF,$BD,$86,$DF,$F8,$8A,$5E,$AF,$0F,$8D,$E2
	.dc.b	$D9,$FF,$13,$FD,$6F,$AD,$BE,$B7,$F8,$B6,$31,$C7,$FC,$8D,$FD,$EB
	.dc.b	$FC,$DD,$FC,$0C,$F2,$EB,$E1,$D9,$C5,$E5,$EA,$E2,$7B,$B6,$FA,$7C
	.dc.b	$4F,$7F,$23,$C0,$DA,$F2,$FA,$98,$69,$E8,$8F,$37,$C4,$7F,$90,$F9
	.dc.b	$7E,$24,$FB,$23,$1C,$3E,$BE,$ED,$CF,$C0,$6E,$5F,$CF,$FB,$DE,$2F
	.dc.b	$3F,$17,$FC,$9D,$18,$BF,$F2,$6E,$49,$CF,$8D,$AB,$B5,$FE,$52,$BE
	.dc.b	$18,$7F,$4B,$3F,$9D,$C7,$F2,$CC,$1F,$1B,$FE,$5B,$9C,$9F,$D8,$E3
	.dc.b	$33,$FE,$5C,$8F,$D3,$3F,$E9,$CE,$DE,$1D,$DC,$5E,$5E,$2E,$27,$83
	.dc.b	$6B,$C5,$BF,$C7,$F8,$EF,$03,$6B,$B6,$F8,$69,$E1,$8F,$35,$3C,$7F
	.dc.b	$F4,$53,$FD,$E5,$AE,$4C,$BF,$E8,$FF,$AB,$FD,$C7,$D6,$DF,$20,$D2
	.dc.b	$57,$FA,$72,$5F,$47,$FE,$0A,$8C,$35,$FF,$5E,$C6,$B8,$7B,$4E,$67
	.dc.b	$FA,$CC,$BE,$8F,$FA,$DD,$7F,$BB,$A3,$13,$FE,$88,$6B,$FE,$01,$7A
	.dc.b	$2F,$AD,$FF,$15,$BF,$E1,$78,$FC,$3B,$C7,$93,$AE,$BE,$06,$C6,$F0
	.dc.b	$AA,$9C,$B8,$6C,$78,$E3,$7F,$31,$FE,$DA,$CB,$D1,$F1,$FB,$25,$6C
	.dc.b	$F5,$7C,$C7,$32,$3E,$D8,$F0,$FB,$BF,$EA,$88,$FF,$53,$B5,$FF,$10
	.dc.b	$36,$9A,$7F,$80,$BE,$5F,$FA,$DD,$70,$E3,$E6,$47,$9F,$37,$BB,$EB
	.dc.b	$7E,$5E,$1C,$A7,$FC,$7D,$18,$C3,$FC,$1A,$D8,$FF,$21,$60,$F9,$CE
	.dc.b	$5C,$32,$7F,$D7,$B5,$F4,$1F,$F1,$B3,$C5,$7D,$6F,$F2,$57,$FD,$86
	.dc.b	$D7,$FC,$78,$DC,$7C,$91,$C3,$CB,$FA,$B3,$FD,$1E,$38,$FA,$58,$33
	.dc.b	$E0,$EB,$C1,$93,$D9,$E3,$F1,$BC,$BD,$69,$3E,$27,$6F,$8B,$E3,$7B
	.dc.b	$91,$F3,$FC,$97,$8B,$86,$47,$A9,$F2,$FB,$DF,$8A,$FA,$DF,$F2,$A4
	.dc.b	$FF,$B8,$1A,$FF,$96,$5B,$BF,$B9,$FF,$78,$47,$FB,$9B,$E3,$7E,$73
	.dc.b	$78,$FF,$2B,$18,$8F,$F3,$13,$F8,$06,$FC,$26,$88,$F3,$F7,$B8,$F5
	.dc.b	$48,$EC,$F4,$ED,$FD,$0F,$D6,$B6,$A4,$AF,$F7,$CC,$7F,$86,$5F,$F7
	.dc.b	$DF,$D6,$FA,$1F,$F3,$A3,$75,$FB,$2D,$DF,$FC,$AE,$BC,$FF,$D7,$FD
	.dc.b	$6F,$A6,$BE,$FB,$6B,$FC,$BA,$DB,$FF,$49,$5F,$EB,$E7,$3E,$2F,$22
	.dc.b	$4F,$FA,$5E,$F9,$F8,$FD,$B4,$93,$9B,$EB,$1E,$2E,$DF,$E0,$2F,$B6
	.dc.b	$7F,$D1,$CD,$BF,$C9,$B6,$BF,$CE,$2F,$F4,$F9,$E8,$7F,$A9,$DF,$F2
	.dc.b	$C3,$83,$F8,$5F,$C2,$76,$3B,$F9,$7F,$C4,$3E,$B7,$FC,$6F,$3B,$6B
	.dc.b	$FE,$9F,$6E,$1F,$A4,$7F,$C7,$71,$BF,$0B,$91,$1E,$5E,$77,$F1,$CF
	.dc.b	$AD,$FF,$3A,$47,$8D,$AF,$FA,$AD,$B7,$FB,$07,$FC,$49,$1B,$D7,$91
	.dc.b	$1E,$5E,$EF,$F2,$8F,$AD,$FD,$A4,$7F,$AD,$5B,$7F,$BA,$6E,$DF,$E9
	.dc.b	$87,$F7,$CD,$F9,$3E,$44,$AF,$91,$3F,$E2,$67,$E5,$6A,$F8,$6F,$E6
	.dc.b	$ED,$8C,$C7,$F3,$AF,$23,$11,$FE,$C7,$97,$57,$02,$47,$03,$AF,$5F
	.dc.b	$42,$3E,$08,$EF,$EA,$48,$F1,$BF,$D2,$3D,$C7,$D3,$79,$3E,$32,$57
	.dc.b	$A1,$E0,$F8,$FA,$7D,$24,$7F,$0D,$F3,$2F,$9E,$F6,$FA,$92,$7E,$0B
	.dc.b	$B7,$D9,$E1,$7C,$22,$3C,$BB,$7E,$2F,$A5,$23,$E4,$BA,$7E,$27,$CC
	.dc.b	$FC,$A2,$3A,$3F,$31,$F5,$6F,$47,$FB,$2B,$63,$31,$FD,$A3,$EB,$5F
	.dc.b	$49,$AF,$46,$57,$27,$8D,$72,$F4,$64,$F6,$7F,$B8,$FD,$6F,$CA,$8F
	.dc.b	$0F,$A3,$E0,$23,$B3,$CB,$F3,$BE,$42,$3F,$AF,$FA,$7F,$E1,$0B,$47
	.dc.b	$CF,$7A,$89,$1C,$8F,$7E,$9E,$E2,$3F,$0A,$3E,$0F,$81,$23,$E2,$BE
	.dc.b	$8F,$A0,$F8,$BE,$8C,$9F,$92,$ED,$78,$9F,$D7,$31,$FF,$6F,$47,$CE
	.dc.b	$F5,$3E,$69,$FE,$12,$FA,$DF,$F4,$D4,$7D,$2F,$1B,$E8,$91,$E0,$7F
	.dc.b	$0D,$FA,$D7,$CD,$FF,$86,$9B,$19,$8F,$F0,$EF,$D6,$BE,$17,$5E,$1C
	.dc.b	$AE,$4F,$1A,$E5,$F2,$92,$7B,$3B,$7D,$78,$99,$8B,$E2,$FC,$44,$8F
	.dc.b	$E4,$5F,$5B,$FB,$27,$D1,$E3,$79,$C8,$FA,$E7,$D5,$F4,$12,$3D,$CF
	.dc.b	$E1,$BE,$0F,$DD,$C9,$27,$E1,$3B,$7C,$A7,$97,$EB,$7F,$C5,$19,$F9
	.dc.b	$E1,$18,$75,$EA,$FF,$09,$F3,$3F,$C8,$72,$C2,$92,$BF,$9E,$47,$15
	.dc.b	$F5,$BF,$90,$27,$CE,$FF,$13,$F3,$FF,$C9,$B1,$FD,$E7,$A7,$FC,$3A
	.dc.b	$DC,$FC,$8D,$BF,$98,$F3,$F3,$F8,$92,$3F,$A6,$7A,$3F,$C1,$DF,$E9
	.dc.b	$D3,$FC,$EA,$B9,$FA,$1E,$03,$84,$FC,$43,$CF,$E5,$F9,$08,$F4,$F9
	.dc.b	$9F,$3F,$B7,$23,$AB,$A7,$96,$0F,$3B,$DA,$48,$F2,$E7,$F0,$3C,$E9
	.dc.b	$1E,$67,$4F,$8F,$EB,$3F,$A2,$BF,$8B,$6A,$6E,$BE,$D8,$D7,$8F,$FE
	.dc.b	$78,$93,$4C,$7C,$37,$F9,$DB,$E2,$7F,$CE,$FF,$1F,$2F,$FB,$A7,$C5
	.dc.b	$FF,$A2,$1F,$88,$BD,$EF,$F3,$E9,$23,$73,$5F,$C5,$1F,$A1,$9E,$0E
	.dc.b	$0C,$4F,$FA,$CA,$33,$19,$CB,$FD,$16,$DD,$FF,$D6,$F1,$88,$EA,$D7
	.dc.b	$FC,$FD,$A6,$88,$F7,$10,$8E,$7B,$49,$D4,$FD,$76,$C7,$83,$D4,$91
	.dc.b	$A6,$BE,$BF,$AE,$FD,$C3,$7A,$FF,$D9,$B1,$88,$F8,$6D,$7F,$5D,$C1
	.dc.b	$FE,$4B,$7C,$CD,$7E,$33,$5F,$96,$D7,$8F,$C4,$FF,$B6,$A0,$D2,$47
	.dc.b	$C1,$D3,$F8,$E3,$FC,$4F,$9F,$EA,$C6,$23,$FD,$16,$1C,$46,$BC,$7F
	.dc.b	$E0,$8F,$D0,$E7,$B5,$FE,$AD,$6B,$FE,$A2,$6B,$F4,$1B,$87,$F6,$28
	.dc.b	$F1,$1F,$0B,$CF,$C9,$1C,$9F,$9F,$70,$F9,$A3,$93,$63,$AF,$CC,$48
	.dc.b	$E2,$7E,$95,$C3,$C7,$E6,$3C,$64,$77,$5F,$E2,$B7,$B7,$A1,$E8,$23
	.dc.b	$A7,$D7,$E8,$48,$E2,$D7,$CD,$F2,$6F,$99,$FB,$AD,$8F,$BB,$EF,$FD
	.dc.b	$17,$F0,$71,$97,$F5,$3F,$01,$FB,$8D,$BD,$6D,$8F,$AE,$D8,$DF,$6F
	.dc.b	$FB,$5D,$E3,$F1,$76,$BF,$9E,$57,$FA,$66,$79,$CE,$67,$FA,$B4,$E5
	.dc.b	$F9,$D3,$97,$F8,$73,$74,$FF,$76,$BF,$20,$D7,$90,$FA,$5A,$48,$FA
	.dc.b	$71,$97,$D7,$C3,$F9,$0B,$F0,$B5,$FF,$2D,$37,$0F,$F1,$4C,$FE,$ED
	.dc.b	$6F,$92,$46,$5E,$5F,$59,$FE,$00,$8F,$E0,$4D,$BF,$E2,$4D,$7F,$8B
	.dc.b	$B5,$FE,$C5,$87,$FA,$CE,$7F,$D2,$D2,$FA,$D4,$7F,$A6,$67,$55,$FD
	.dc.b	$26,$B5,$39,$F9,$8D,$CF,$F4,$DA,$DF,$91,$FE,$FA,$53,$A8,$C7,$C7
	.dc.b	$5B,$F1,$F8,$77,$0F,$F0,$67,$D6,$BC,$BC,$43,$25,$62,$B5,$7E,$7F
	.dc.b	$17,$75,$A8,$C1,$EB,$AC,$C7,$1F,$85,$B3,$FE,$B7,$BF,$F3,$93,$F0
	.dc.b	$7C,$FC,$E3,$ED,$6E,$9F,$02,$1F,$5B,$FE,$8E,$AE,$5E,$2F,$A2,$31
	.dc.b	$3F,$DC,$70,$FE,$D1,$5F,$EC,$59,$F4,$CF,$F7,$57,$77,$D5,$CB,$7C
	.dc.b	$4E,$9F,$15,$F1,$3E,$B7,$F3,$93,$B8,$C7,$57,$87,$E3,$37,$97,$76
	.dc.b	$E5,$DF,$C3,$E5,$E3,$79,$DB,$5E,$5F,$2F,$0D,$3D,$11,$F5,$3E,$3F
	.dc.b	$F8,$14,$7F,$B7,$FD,$D7,$AD,$9F,$17,$EC,$C6,$27,$FD,$C9,$F5,$BF
	.dc.b	$E8,$D7,$FD,$CD,$7C,$B3,$FF,$0B,$4E,$5F,$F0,$C4,$67,$7F,$A1,$2F
	.dc.b	$96,$FC,$2F,$C6,$BC,$FF,$C3,$CE,$71,$FF,$56,$C6,$3E,$27,$37,$67
	.dc.b	$C6,$F0,$E0,$FC,$7F,$2F,$EC,$A7,$CF,$18,$F0,$FA,$FC,$3E,$E7,$D6
	.dc.b	$DD,$BD,$3F,$0F,$A9,$F7,$B6,$3D,$BE,$77,$6F,$3B,$CE,$DA,$F8,$B8
	.dc.b	$61,$A7,$C8,$8F,$6F,$0F,$E9,$3E,$A5,$EB,$62,$BF,$CE,$9C,$7F,$95
	.dc.b	$9C,$C7,$D0,$F8,$7D,$67,$F3,$9B,$5F,$9B,$FA,$A7,$D5,$FE,$C3,$EB
	.dc.b	$5B,$7E,$CA,$73,$8F,$63,$FC,$D9,$3A,$71,$8F,$FA,$16,$7F,$8A,$61
	.dc.b	$D7,$0E,$5F,$11,$CD,$E9,$D3,$8C,$38,$E0,$C1,$39,$8F,$F2,$D4,$3E
	.dc.b	$8F,$F8,$2F,$D6,$FF,$38,$9C,$C6,$2F,$D1,$F0,$DE,$98,$B1,$DA,$F9
	.dc.b	$37,$E8,$FE,$67,$AB,$EF,$6C,$6F,$E1,$F0,$D3,$81,$CF,$8F,$EA,$6B
	.dc.b	$F1,$DB,$1D,$FC,$EC,$34,$E0,$78,$FC,$FF,$9D,$1F,$3D,$B1,$E7,$F7
	.dc.b	$3F,$01,$C3,$F4,$DB,$1E,$BD,$F0,$D3,$81,$F3,$3E,$B7,$F8,$6B,$FE
	.dc.b	$A3,$E7,$F9,$AE,$7B,$BF,$19,$C9,$FC,$C2,$BF,$88,$CF,$FA,$A5,$E3
	.dc.b	$7F,$EC,$17,$2D,$AE,$1F,$07,$F9,$5F,$8C,$B7,$F7,$31,$47,$5F,$87
	.dc.b	$CE,$1F,$E9,$5C,$51,$8E,$8F,$E5,$D0,$FA,$7E,$18,$FA,$DB,$7F,$E9
	.dc.b	$DF,$AD,$76,$FF,$48,$5D,$D6,$30,$FF,$4C,$E0,$FE,$AB,$87,$E4,$1F
	.dc.b	$EB,$2F,$FB,$1D,$CD,$FF,$AA,$5C,$BF,$EE,$1F,$AD,$7C,$5F,$38,$FE
	.dc.b	$AF,$18,$AE,$91,$8E,$1C,$FB,$D6,$7F,$3F,$48,$E3,$E7,$F2,$F3,$ED
	.dc.b	$69,$FE,$C7,$F4,$EF,$94,$BE,$66,$1B,$5F,$AA,$9F,$DA,$67,$FE,$C1
	.dc.b	$7F,$8E,$37,$C6,$E7,$E4,$7F,$6E,$FA,$DF,$D1,$CF,$FB,$1D,$7D,$4E
	.dc.b	$7C,$6D,$3B,$F7,$78,$78,$17,$3F,$DC,$DF,$FA,$AD,$B8,$75,$E4,$EC
	.dc.b	$D8,$FD,$F8,$F5,$7F,$BF,$AF,$5E,$C6,$1F,$F0,$1F,$D6,$BE,$3A,$FC
	.dc.b	$D6,$1B,$7F,$C2,$BE,$B7,$D2,$DF,$0B,$FC,$17,$18,$CB,$FD,$C8,$FE
	.dc.b	$52,$78,$6B,$AF,$1F,$6D,$8F,$E2,$BF,$5A,$F1,$BF,$C2,$AB,$E4,$EC
	.dc.b	$61,$CE,$DA,$F0,$ED,$F2,$7F,$C3,$BE,$DB,$7B,$F3,$98,$6D,$7F,$24
	.dc.b	$97,$AB,$C3,$E3,$F8,$B6,$7F,$94,$FD,$6F,$AF,$87,$FE,$00,$3F,$BE
	.dc.b	$6F,$17,$F8,$B2,$31,$E5,$7F,$86,$DF,$C0,$B7,$AB,$5D,$7B,$78,$7A
	.dc.b	$B8,$C7,$BB,$71,$EF,$F1,$3D,$7E,$27,$B3,$6B,$CB,$EA,$E1,$A7,$A2
	.dc.b	$3C,$9F,$0F,$FE,$40,$E6,$F8,$93,$EC,$8C,$71,$7A,$7B,$B7,$3F,$01
	.dc.b	$B9,$36,$FF,$1B,$0F,$1B,$A7,$8D,$FE,$4C,$8C,$5F,$F5,$0C,$93,$9F
	.dc.b	$1F,$57,$6B,$FA,$AD,$F0,$C3,$B4,$18,$7F,$94,$4C,$99,$C1,$F1,$FF
	.dc.b	$E5,$C8,$D5,$7F,$92,$E7,$35,$FD,$8A,$7F,$54,$72,$3B,$77,$71,$EE
	.dc.b	$E4,$F3,$F1,$31,$5C,$5B,$5C,$46,$FF,$37,$E4,$3C,$4D,$AF,$06,$F8
	.dc.b	$69,$E3,$8F,$45,$3C,$8F,$7B,$0C,$14,$F8,$8D,$69,$BF,$AB,$FE,$01
	.dc.b	$F5,$BB,$6C,$67,$C7,$4E,$5B,$1C,$7E,$8B,$EF,$37,$83,$94,$62,$3C
	.dc.b	$B6,$8C,$47,$D6,$87,$D1,$7D,$8A,$FF,$3E,$46,$27,$FC,$26,$D7,$E9
	.dc.b	$3C,$E3,$F0,$1F,$5B,$FE,$32,$9F,$C1,$4F,$E3,$DE,$3C,$9D,$75,$F0
	.dc.b	$36,$37,$85,$54,$E5,$C3,$63,$C7,$1B,$F9,$6F,$A3,$97,$A3,$E3,$F6
	.dc.b	$4A,$D9,$EA,$F9,$9E,$64,$7D,$B1,$E1,$F7,$BF,$D5,$D1,$FE,$32,$6B
	.dc.b	$F9,$16,$D7,$FA,$87,$E6,$7F,$B0,$CE,$F8,$F9,$11,$E5,$CD,$ED,$FA
	.dc.b	$DF,$95,$7F,$EA,$96,$F0,$7E,$A1,$B5,$FA,$BC,$1F,$57,$C7,$86,$4F
	.dc.b	$F9,$1A,$BE,$73,$FA,$57,$8A,$FA,$DF,$E5,$8F,$F9,$31,$AF,$EA,$DB
	.dc.b	$8F,$FA,$ED,$79,$7F,$4D,$FE,$9D,$3C,$7D,$1C,$19,$F0,$75,$E0,$C9
	.dc.b	$EC,$F1,$F8,$CE,$5E,$BC,$9F,$13,$B7,$C5,$F1,$9D,$C8,$F9,$FE,$47
	.dc.b	$C5,$C3,$23,$D4,$F9,$7D,$FF,$C4,$7D,$6B,$C9,$A4,$AF,$F3,$5C,$7F
	.dc.b	$BC,$9F,$F3,$6F,$B7,$F3,$1B,$C9,$F9,$18,$C4,$7F,$0F,$7E,$EB,$7E
	.dc.b	$5F,$44,$79,$FB,$FC,$7A,$E4,$76,$7A,$77,$7E,$77,$EB,$5B,$52,$57
	.dc.b	$FA,$02,$3F,$C3,$8F,$FA,$0B,$EB,$3D,$0F,$F2,$16,$F4,$7F,$BC,$5B
	.dc.b	$C1,$FC,$C1,$79,$FF,$AB,$FA,$DF,$4D,$7F,$BD,$DB,$5F,$CD,$9B,$7F
	.dc.b	$CE,$6F,$F5,$73,$9F,$1B,$91,$27,$FD,$2B,$5F,$E1,$76,$E7,$DC,$8E
	.dc.b	$BB,$7F,$79,$1E,$C8,$FE,$74,$DC,$58,$7F,$48,$8C,$37,$A3,$8F,$C2
	.dc.b	$4B,$FF,$52,$D7,$F9,$71,$BE,$67,$F0,$9F,$AD,$FF,$52,$47,$FA,$A9
	.dc.b	$AF,$F4,$96,$E3,$E0,$8F,$EB,$11,$FE,$3E,$78,$78,$5C,$C8,$F3,$F3
	.dc.b	$FF,$8B,$7D,$6F,$F9,$DA,$3F,$D6,$ED,$7F,$AB,$36,$FB,$42,$3F,$C5
	.dc.b	$0F,$0F,$33,$99,$1E,$7E,$EF,$F2,$3F,$AD,$FD,$8C,$7F,$61,$6D,$F4
	.dc.b	$6E,$FF,$DC,$1F,$DD,$B7,$E1,$39,$12,$BF,$D9,$B3,$FE,$2A,$7F,$D9
	.dc.b	$FA,$FF,$9C,$5B,$F0,$BF,$CC,$E3,$11,$FD,$C7,$C9,$B8,$FB,$25,$D7
	.dc.b	$B2,$57,$23,$AF,$67,$32,$3E,$08,$ED,$EA,$C8,$F1,$BF,$D0,$FE,$31
	.dc.b	$F4,$DE,$3F,$8C,$95,$E8,$78,$3F,$E0,$5C,$7F,$AF,$3E,$CE,$BF,$51
	.dc.b	$FD,$37,$E6,$5F,$3D,$ED,$F5,$65,$7C,$27,$6F,$C4,$F0,$FE,$11,$1F
	.dc.b	$F7,$AD,$7F,$30,$DF,$E1,$7A,$7F,$1C,$93,$C4,$F9,$BF,$33,$F2,$C8
	.dc.b	$FD,$08,$FE,$C1,$F5,$7F,$CC,$23,$FE,$00,$B5,$FF,$10,$36,$BF,$17
	.dc.b	$F5,$AF,$A3,$E3,$E8,$CB,$E2,$F4,$F5,$3A,$F8,$92,$7A,$BF,$DB,$FE
	.dc.b	$B7,$E6,$C7,$83,$D2,$F1,$91,$D9,$E4,$F9,$DF,$39,$1F,$F2,$06,$7F
	.dc.b	$9E,$FA,$6F,$E1,$70,$D5,$F3,$DE,$E2,$47,$33,$DF,$AF,$C2,$23,$F0
	.dc.b	$A3,$E2,$F8,$12,$3E,$2B,$E9,$FA,$0F,$93,$E8,$C9,$F9,$2E,$DF,$93
	.dc.b	$EA,$FC,$E2,$3E,$8F,$13,$FC,$AF,$F5,$BF,$EF,$48,$FA,$31,$F5,$08
	.dc.b	$FD,$49,$FF,$0B,$7D,$63,$E6,$E9,$27,$A1,$FE,$63,$FA,$D7,$C3,$70
	.dc.b	$F0,$E4,$E8,$5A,$39,$47,$F9,$7E,$3C,$8D,$F8,$EF,$09,$C9,$D7,$C5
	.dc.b	$F9,$08,$F5,$FA,$DF,$DA,$BD,$EB,$EF,$BD,$7C,$6F,$59,$C1,$F0,$7C
	.dc.b	$FF,$B8,$8F,$9F,$AB,$FC,$D2,$F3,$E2,$F8,$64,$7B,$4F,$F8,$EF,$EB
	.dc.b	$7F,$C4,$FC,$3F,$E8,$46,$CF,$FA,$1D,$AB,$CC,$F3,$BF,$E7,$F8,$EE
	.dc.b	$DF,$85,$E8,$75,$F5,$3F,$C9,$1F,$5B,$F9,$02,$7C,$EF,$F9,$2F,$07
	.dc.b	$A1,$AF,$80,$98,$FD,$1B,$FE,$FF,$E0,$7F,$36,$79,$7A,$1D,$09,$3F
	.dc.b	$5F,$0F,$E1,$2F,$1F,$43,$C2,$47,$A4,$F9,$3D,$B9,$1D,$DF,$27,$CC
	.dc.b	$7A,$3D,$B9,$3E,$37,$6F,$FA,$7F,$17,$9F,$EE,$24,$78,$F0,$F8,$3E
	.dc.b	$6C,$8F,$63,$A7,$FD,$51,$F5,$9F,$D1,$9F,$F5,$B3,$53,$74,$FE,$B1
	.dc.b	$C9,$FE,$BA,$87,$91,$FE,$75,$6D,$7E,$65,$AF,$6F,$87,$FF,$3A,$73
	.dc.b	$FF,$A7,$FE,$2C,$D2,$4F,$FA,$96,$1D,$3B,$24,$65,$E5,$F5,$9F,$E7
	.dc.b	$F2,$47,$EA,$67,$8B,$E3,$F3,$7F,$D5,$D8,$FF,$83,$E3,$FD,$6A,$D7
	.dc.b	$D9,$1F,$EB,$3F,$AD,$FF,$1F,$C7,$FA,$E9,$AF,$FB,$69,$AF,$FB,$75
	.dc.b	$AF,$E5,$FD,$CF,$F5,$E6,$C6,$92,$3F,$D8,$53,$FD,$A9,$ED,$ED,$F8
	.dc.b	$48,$FF,$B1,$E9,$EA,$F4,$24,$65,$EF,$F5,$D2,$38,$35,$FF,$69,$67
	.dc.b	$FA,$87,$FD,$AA,$D7,$FD,$EA,$D7,$FD,$F0,$D7,$F0,$7F,$15,$F2,$88
	.dc.b	$FF,$B6,$F4,$47,$CE,$F4,$FF,$DF,$FC,$7F,$39,$B5,$F8,$96,$BF,$90
	.dc.b	$6B,$F9,$0F,$46,$DF,$F4,$68,$1F,$E2,$83,$46,$2B,$ED,$CF,$02,$46
	.dc.b	$4F,$CC,$38,$79,$67,$FE,$57,$6F,$BF,$E2,$8C,$57,$FB,$E9,$B3,$E2
	.dc.b	$3F,$8A,$7B,$9B,$23,$DC,$FF,$5B,$3E,$AF,$99,$91,$93,$F5,$AE,$1F
	.dc.b	$1F,$CD,$F8,$48,$EE,$91,$F1,$B8,$3F,$C0,$58,$78,$E9,$18,$7C,$CD
	.dc.b	$7E,$4B,$5F,$DE,$FC,$B7,$47,$91,$8A,$FB,$9F,$01,$FD,$5A,$33,$AA
	.dc.b	$B8,$47,$EA,$BE,$1B,$FC,$D1,$8D,$3D,$D4,$8C,$CF,$F0,$96,$BF,$C4
	.dc.b	$3D,$5A,$FF,$10,$6C,$9A,$6B,$2F,$F4,$2D,$FC,$B6,$D7,$E1,$5B,$1F
	.dc.b	$15,$B1,$F6,$7C,$97,$CC,$7E,$2E,$3F,$1B,$18,$AF,$87,$08,$C3,$F1
	.dc.b	$9A,$FF,$05,$F1,$39,$2E,$84,$61,$B5,$FC,$02,$37,$F5,$CF,$A6,$A4
	.dc.b	$9F,$0C,$65,$F1,$E1,$FE,$BA,$7F,$C0,$D0,$ED,$FF,$4C,$79,$3F,$E8
	.dc.b	$CA,$D6,$48,$FF,$47,$D6,$EF,$FC,$1D,$4F,$9A,$E3,$1F,$17,$E3,$6E
	.dc.b	$1F,$12,$6B,$F0,$E2,$D3,$E6,$B3,$F5,$DF,$DB,$30,$EB,$18,$F3,$F7
	.dc.b	$F1,$9C,$22,$CE,$8F,$1D,$F2,$F8,$9E,$AD,$8F,$3F,$A7,$86,$9C,$0F
	.dc.b	$F1,$47,$D7,$7F,$A4,$1F,$F1,$56,$5F,$D9,$9C,$FC,$5F,$9B,$CF,$F5
	.dc.b	$D7,$9F,$6F,$8D,$5F,$BD,$8E,$51,$83,$F8,$8B,$BE,$1F,$F5,$CB,$A6
	.dc.b	$F9,$1F,$1B,$EB,$BF,$DF,$77,$E4,$8C,$7C,$9F,$83,$C4,$F0,$C5,$9D
	.dc.b	$1E,$AF,$53,$97,$91,$E6,$6C,$7B,$BD,$4C,$34,$E0,$7F,$92,$7C,$1F
	.dc.b	$E0,$EF,$F9,$2F,$DF,$FE,$D6,$E7,$C5,$FB,$71,$8D,$FD,$C3,$EB,$9F
	.dc.b	$15,$F8,$0C,$39,$CE,$A7,$38,$73,$3F,$CB,$0B,$E5,$FF,$0F,$F1,$DF
	.dc.b	$D6,$27,$31,$8F,$47,$F1,$74,$F8,$CB,$38,$6D,$74,$DF,$F2,$3C,$EF
	.dc.b	$8A,$E3,$18,$79,$FA,$38,$AF,$1B,$74,$F4,$7A,$3A,$9F,$3B,$63,$D9
	.dc.b	$E8,$76,$F8,$AF,$2B,$6B,$DB,$86,$1A,$7C,$08,$F5,$F0,$7E,$9B,$EA
	.dc.b	$9F,$37,$8A,$FF,$3E,$46,$1F,$8F,$E7,$E7,$2F,$98,$DE,$D4,$FF,$4F
	.dc.b	$FA,$CF,$E0,$55,$F5,$BE,$B9,$AF,$ED,$27,$37,$C2,$BF,$B3,$D6,$AF
	.dc.b	$FD,$19,$5F,$E9,$56,$F9,$BC,$7E,$37,$93,$CF,$AB,$13,$C1,$B8,$7E
	.dc.b	$1A,$DF,$45,$FC,$13,$EB,$BF,$98,$E7,$E1,$A5,$6B,$0F,$F4,$E5,$69
	.dc.b	$F6,$7C,$67,$A7,$0F,$53,$63,$CF,$F3,$BF,$2F,$EB,$BD,$CD,$8F,$5F
	.dc.b	$C0,$E1,$A7,$03,$DF,$C7,$F4,$F6,$F8,$CD,$8F,$87,$E7,$61,$A7,$03
	.dc.b	$E3,$F9,$FF,$36,$3E,$73,$63,$E5,$FC,$87,$DF,$E2,$FA,$4D,$8F,$9F
	.dc.b	$C1,$86,$9C,$0F,$F5,$F7,$D7,$7F,$09,$7F,$C1,$9C,$FF,$EB,$F7,$3D
	.dc.b	$DF,$8B,$E5,$FE,$63,$5F,$C4,$67,$FC,$24,$F1,$BF,$F1,$2B,$96,$D7
	.dc.b	$0F,$87,$FC,$A7,$C8,$DB,$FC,$9B,$D3,$E2,$B1,$7F,$C0,$B0,$46,$3A
	.dc.b	$3E,$E4,$3E,$A3,$86,$1E,$A6,$DF,$F0,$AF,$AE,$FE,$F1,$3F,$ED,$B7
	.dc.b	$FC,$04,$DE,$DF,$E9,$71,$89,$FE,$23,$BF,$F5,$3A,$FF,$71,$3F,$D6
	.dc.b	$B1,$FC,$63,$99,$DF,$B1,$FD,$53,$EB,$9F,$1F,$FD,$5A,$BF,$C8,$17
	.dc.b	$F9,$39,$CE,$1D,$30,$E1,$C1,$C7,$E0,$6C,$FF,$BC,$6B,$FC,$56,$DF
	.dc.b	$3F,$B7,$A1,$DD,$B1,$FE,$F7,$FA,$87,$C5,$7F,$66,$7F,$8E,$46,$63
	.dc.b	$1E,$66,$7C,$6D,$79,$36,$B9,$3F,$DB,$BE,$BB,$F5,$1B,$7F,$34,$7F
	.dc.b	$C2,$3B,$FE,$B5,$CB,$7E,$13,$87,$BB,$C7,$BE,$E3,$FE,$68,$AE,$11
	.dc.b	$D7,$0E,$DC,$76,$B4,$7F,$97,$71,$47,$F8,$02,$1E,$BD,$F5,$F0,$D8
	.dc.b	$E3,$7F,$E7,$6F,$AE,$7C,$77,$F8,$21,$7C,$E5,$8C,$3F,$E0,$CF,$AE
	.dc.b	$F6,$DF,$F9,$B5,$78,$39,$E0,$D3,$BD,$B8,$78,$D6,$FC,$39,$B0,$FF
	.dc.b	$0A,$46,$3E,$B9,$E2,$7F,$86,$9F,$F1,$9D,$7F,$60,$73,$3B,$F8,$8D
	.dc.b	$70,$6F,$C3,$B3,$CD,$FF,$11,$FB,$BE,$C8,$F0,$BF,$D9,$30,$FE,$D6
	.dc.b	$E5,$BE,$EF,$0F,$5B,$8F,$D7,$5B,$E2,$FF,$8B,$BE,$BB,$E0,$C7,$B1
	.dc.b	$FE,$09,$7F,$DD,$9C,$B7,$CB,$E1,$EA,$71,$E4,$B7,$C5,$E9,$EA,$62
	.dc.b	$F9,$37,$B3,$F9,$C6,$3F,$AD,$6F,$93,$E3,$F8,$8F,$23,$E5,$F5,$31
	.dc.b	$97,$95,$B8,$F9,$7E,$1B,$FC,$8F,$CD,$F0,$71,$EB,$18,$EB,$F7,$77
	.dc.b	$FB,$E2,$E3,$EF,$7F,$C9,$F4,$F2,$39,$F8,$B8,$6D,$7E,$02,$5E,$46
	.dc.b	$63,$0D,$CF,$FC,$3A,$7E,$10,$FF,$0F,$E7,$FE,$9B,$9F,$C4,$C6,$6B
	.dc.b	$FA,$D4,$E6,$38,$F7,$7F,$98,$9D,$57,$FA,$C6,$31,$DD,$9F,$13,$51
	.dc.b	$67,$AB,$C7,$4F,$8B,$F1,$D6,$7F,$33,$8A,$3B,$6D,$9C,$BD,$2D,$CF
	.dc.b	$FD,$31,$E4,$FF,$A5,$A7,$FD,$CA,$D4,$D3,$FD,$37,$F5,$9F,$BA,$FA
	.dc.b	$EB,$E3,$9A,$4A,$FF,$61,$CB,$E4,$BF,$D1,$51,$86,$BF,$ED,$18,$D6
	.dc.b	$1D,$A7,$33,$FE,$E7,$97,$D2,$7F,$BA,$6B,$FC,$0D,$18,$9F,$DC,$B5
	.dc.b	$FF,$29,$BD,$17,$D7,$7F,$8B,$10,$8F,$C3,$3C,$B9,$3B,$6F,$E1,$6C
	.dc.b	$70,$9E,$D7,$E2,$E4,$CC,$F3,$6C,$FE,$32,$3F,$86,$FC,$AF,$FB,$BE
	.dc.b	$BD,$3F,$19,$A2,$4F,$AA,$F8,$FC,$CC,$8F,$6B,$A7,$FC,$DF,$EE,$7F
	.dc.b	$B3,$5F,$E3,$2D,$73,$CF,$E4,$F9,$6F,$EB,$B1,$AE,$7F,$3D,$1A,$62
	.dc.b	$BD,$7F,$5D,$EE,$DF,$E3,$DA,$30,$FC,$A6,$AC,$1F,$39,$CB,$82,$4F
	.dc.b	$F2,$EA,$F9,$EF,$FA,$21,$E3,$F5,$DF,$CF,$A3,$F9,$9B,$5F,$F4,$6B
	.dc.b	$77,$FF,$23,$3F,$E8,$E8,$DF,$A3,$F4,$A3,$FA,$4C,$60,$E7,$DF,$9E
	.dc.b	$F9,$1D,$9E,$1F,$17,$C7,$D7,$93,$E2,$76,$F8,$BE,$2F,$B1,$1F,$3F
	.dc.b	$C7,$F8,$78,$24,$7A,$9F,$27,$BF,$F8,$6F,$AE,$78,$D4,$95,$FD,$4A
	.dc.b	$3F,$DD,$8F,$F5,$4F,$6F,$E5,$B7,$8F,$F1,$F1,$88,$FF,$58,$BF,$79
	.dc.b	$BF,$2B,$A2,$3C,$FD,$FE,$3D,$92,$3B,$3D,$3B,$3F,$37,$F5,$CD,$A9
	.dc.b	$2B,$FB,$3C,$7F,$86,$1F,$ED,$3F,$5D,$E8,$7F,$D8,$6D,$E7,$FF,$3E
	.dc.b	$37,$83,$F1,$6B,$CC,$FD,$4F,$D7,$7A,$6B,$FD,$04,$DA,$FC,$8B,$6F
	.dc.b	$F7,$37,$FA,$B9,$CF,$8B,$C8,$93,$FD,$F2,$BF,$C2,$2D,$CF,$B9,$1D
	.dc.b	$76,$7E,$EA,$3D,$91,$FE,$DC,$6E,$3F,$E9,$56,$E1,$FA,$47,$FD,$A5
	.dc.b	$3F,$E0,$A5,$F1,$FC,$C9,$19,$F2,$FF,$83,$FD,$77,$FA,$8A,$3F,$C2
	.dc.b	$2D,$7F,$DD,$6D,$C3,$FD,$48,$BD,$3E,$3E,$14,$AE,$2F,$4F,$3F,$F8
	.dc.b	$9F,$D7,$7F,$9C,$A7,$FC,$3A,$D7,$FD,$EE,$DC,$3F,$D6,$0F,$F8,$7A
	.dc.b	$37,$E6,$F2,$23,$CB,$DC,$FE,$3F,$F5,$DF,$AE,$8F,$F8,$02,$DB,$FF
	.dc.b	$5D,$37,$7F,$E1,$2F,$EE,$5B,$F0,$9C,$89,$5F,$E3,$29,$7B,$7F,$CB
	.dc.b	$37,$7C,$35,$25,$7F,$8E,$3C,$AC,$BF,$E3,$A9,$76,$71,$24,$66,$BA
	.dc.b	$76,$C8,$E0,$E9,$69,$FE,$D3,$BF,$89,$18,$D3,$F9,$1C,$62,$75,$EA
	.dc.b	$FF,$42,$F6,$DF,$4F,$E9,$F8,$B9,$8F,$95,$F6,$57,$B7,$B2,$4F,$A5
	.dc.b	$FE,$97,$F3,$0F,$9C,$F8,$5E,$BC,$AF,$73,$C1,$F2,$1F,$D7,$34,$78
	.dc.b	$9F,$24,$91,$FE,$5E,$BF,$E6,$3A,$FB,$7D,$4F,$9C,$49,$F9,$55,$F4
	.dc.b	$7E,$66,$47,$CD,$74,$F8,$07,$D6,$FF,$33,$70,$DB,$FF,$7E,$37,$87
	.dc.b	$FA,$5F,$D7,$3E,$BB,$FB,$4C,$FE,$16,$FF,$BA,$B5,$EB,$E9,$EB,$E1
	.dc.b	$77,$EA,$49,$FF,$3C,$D7,$EC,$9F,$27,$D5,$7D,$27,$3E,$A7,$62,$3C
	.dc.b	$2B,$C3,$F3,$92,$39,$BF,$DF,$3E,$A3,$F8,$84,$7B,$BE,$77,$CE,$47
	.dc.b	$67,$8B,$4F,$59,$1F,$2C,$7B,$BC,$29,$1E,$87,$FC,$13,$93,$EA,$3E
	.dc.b	$1F,$A7,$2B,$DA,$F0,$7E,$6F,$AF,$F2,$08,$FF,$A6,$BE,$BB,$FD,$42
	.dc.b	$FC,$4E,$3F,$9A,$47,$E3,$CF,$F8,$63,$EB,$1F,$35,$49,$3F,$EA,$4F
	.dc.b	$AE,$FF,$5D,$BF,$EA,$69,$78,$7C,$49,$19,$AE,$9F,$29,$23,$83,$A7
	.dc.b	$F1,$14,$F8,$DF,$11,$23,$FD,$67,$F5,$DF,$BA,$7C,$1E,$37,$9C,$8F
	.dc.b	$8E,$7D,$5F,$3D,$23,$CC,$FC,$CF,$85,$F7,$70,$C9,$F5,$3B,$7F,$35
	.dc.b	$F5,$DF,$E2,$9E,$F7,$0F,$F8,$82,$30,$DF,$53,$FC,$59,$E6,$FF,$90
	.dc.b	$27,$FC,$55,$7F,$EC,$A6,$BD,$31,$D5,$7F,$91,$BE,$BB,$F1,$64,$F9
	.dc.b	$CF,$F1,$B7,$9D,$9F,$81,$98,$7F,$B5,$2B,$FB,$F6,$9E,$48,$FF,$2C
	.dc.b	$C7,$2F,$3B,$A1,$23,$F9,$27,$9B,$DF,$CF,$91,$CD,$D3,$FB,$9A,$FE
	.dc.b	$07,$BF,$F8,$F9,$B1,$DB,$DC,$FF,$25,$7C,$BF,$A7,$DD,$93,$E5,$7D
	.dc.b	$93,$ED,$F3,$64,$FA,$5F,$F2,$96,$7F,$07,$CE,$93,$ED,$76,$FF,$0B
	.dc.b	$FA,$CF,$E9,$0F,$F9,$D9,$A9,$BE,$AF,$F5,$9D,$3F,$CA,$B0,$F2,$3F
	.dc.b	$CE,$6D,$AF,$F7,$EB,$5E,$DF,$13,$FE,$78,$F8,$C7,$5D,$24,$FF,$9E
	.dc.b	$3E,$1E,$BF,$C7,$E7,$F6,$CF,$F9,$F3,$B3,$3C,$BF,$56,$13,$A7,$FC
	.dc.b	$D2,$7F,$A4,$1C,$38,$6B,$FE,$6E,$6B,$F0,$7C,$DF,$F5,$B6,$5F,$C7
	.dc.b	$63,$9B,$5F,$83,$1A,$F3,$A4,$78,$0F,$94,$E1,$C7,$D6,$7F,$92,$5F
	.dc.b	$2B,$5F,$F3,$FB,$5F,$F5,$3B,$5F,$F3,$F7,$B9,$FE,$C9,$E0,$34,$91
	.dc.b	$E1,$3F,$D0,$9C,$7B,$23,$99,$FE,$AE,$DF,$57,$A1,$1C,$1B,$FA,$EF
	.dc.b	$01,$1C,$1D,$F3,$FF,$2C,$37,$E2,$3E,$49,$1D,$7C,$57,$CC,$23,$83
	.dc.b	$97,$A9,$FE,$A0,$E6,$F9,$ED,$8F,$C7,$B5,$FC,$B3,$5F,$CB,$36,$CD
	.dc.b	$24,$7F,$B4,$A7,$ED,$FA,$7B,$24,$64,$FC,$EB,$D7,$3F,$F2,$DC,$7E
	.dc.b	$3C,$FF,$58,$7C,$06,$48,$F1,$3F,$55,$F3,$1E,$32,$39,$3F,$58,$F9
	.dc.b	$FE,$02,$46,$72,$FF,$19,$B6,$3C,$5F,$33,$23,$96,$08,$C4,$79,$69
	.dc.b	$EA,$FC,$32,$47,$57,$E3,$7D,$77,$18,$FC,$C4,$A3,$11,$FA,$0F,$7F
	.dc.b	$FA,$9B,$FA,$28,$CB,$FC,$0B,$E2,$3F,$93,$6D,$FA,$66,$C6,$63,$D9
	.dc.b	$18,$8F,$E1,$DE,$BD,$7F,$87,$37,$03,$4D,$79,$8F,$53,$E9,$50,$46
	.dc.b	$70,$FD,$13,$63,$F6,$8D,$7F,$82,$BF,$E7,$D8,$DF,$C3,$F4,$9E,$6D
	.dc.b	$9D,$47,$EF,$7E,$B5,$F4,$DB,$6D,$FE,$C5,$B3,$B9,$E5,$F1,$BF,$D2
	.dc.b	$FC,$1F,$E3,$14,$FA,$EC,$61,$FF,$00,$31,$FF,$5B,$3E,$DA,$DC,$EB
	.dc.b	$1F,$81,$4F,$9B,$C4,$7C,$2C,$6C,$E7,$0E,$47,$58,$CF,$CD,$F0,$FA
	.dc.b	$E7,$A9,$8A,$DE,$1F,$36,$73,$18,$8F,$9F,$F1,$DB,$7F,$CA,$98,$A3
	.dc.b	$1A,$7F,$35,$8F,$AB,$5C,$6F,$FC,$DC,$F3,$8C,$DF,$F9,$76,$72,$F4
	.dc.b	$F5,$FD,$73,$EC,$6D,$78,$FE,$BB,$FD,$22,$F5,$F1,$3E,$1C,$62,$FA
	.dc.b	$46,$33,$CF,$B5,$6F,$93,$C3,$C6,$F9,$7F,$5D,$FE,$F3,$9E,$D1,$8F
	.dc.b	$17,$AF,$85,$B9,$F1,$37,$2F,$F3,$B6,$6F,$4F,$9F,$8D,$E4,$6D,$FA
	.dc.b	$7D,$4C,$34,$F6,$47,$DC,$EE,$FD,$F4,$7D,$1F,$81,$FF,$2C,$3F,$4E
	.dc.b	$33,$8F,$FA,$16,$72,$FF,$A1,$FE,$B9,$F1,$3C,$FC,$5F,$C3,$C6,$30
	.dc.b	$E5,$18,$F9,$7F,$C5,$FC,$6B,$CE,$C5,$7F,$A6,$DE,$3E,$8E,$32,$EC
	.dc.b	$DE,$BF,$EE,$99,$38,$3F,$2B,$CC,$F8,$9F,$0C,$61,$F1,$F9,$F8,$CF
	.dc.b	$F1,$D7,$FC,$E9,$B5,$E8,$F3,$F7,$1A,$6E,$7E,$FF,$3B,$BF,$9B,$EE
	.dc.b	$6D,$74,$CB,$0D,$3B,$C7,$C4,$E4,$FD,$5F,$D5,$3E,$6F,$15,$FA,$18
	.dc.b	$C3,$E6,$F3,$33,$97,$CF,$6F,$07,$F8,$57,$EB,$5F,$59,$FB,$6F,$AE
	.dc.b	$FE,$6B,$87,$EA,$E7,$2F,$EB,$63,$38,$7F,$AF,$E7,$FD,$0F,$3B,$AF
	.dc.b	$DF,$C6,$27,$51,$8E,$AC,$FC,$5D,$BE,$CF,$0F,$A5,$FE,$0F,$F5,$DF
	.dc.b	$E7,$09,$F1,$46,$1E,$BF,$15,$B7,$CC,$DC,$7C,$9E,$F7,$C6,$F4,$BC
	.dc.b	$6D,$AF,$67,$BF,$86,$9E,$F8,$F4,$F1,$7C,$ED,$6C,$FC,$D7,$DB,$E7
	.dc.b	$E3,$1F,$84,$DE,$1F,$99,$E7,$EA,$FE,$53,$6B,$E4,$7D,$2F,$DA,$E0
	.dc.b	$F8,$ED,$AF,$A1,$38,$69,$F4,$E3,$F9,$5F,$D7,$7F,$09,$C3,$FC,$9D
	.dc.b	$CD,$FE,$B2,$AF,$E4,$53,$97,$FA,$95,$A7,$4F,$FB,$DE,$7F,$94,$B7
	.dc.b	$BF,$AF,$9B,$D9,$EF,$F0,$BF,$90,$F6,$2E,$3F,$8F,$7C,$7F,$11,$8C
	.dc.b	$3A,$37,$3F,$F3,$16,$4F,$A8,$DE,$1E,$66,$DA,$BF,$CD,$1F,$5D,$FD
	.dc.b	$BA,$BF,$9E,$2E,$EF,$E9,$6E,$6B,$84,$FB,$A3,$4F,$F4,$3D,$FF,$A7
	.dc.b	$4E,$8F,$EE,$8F,$F5,$A7,$01,$DD,$FF,$9E,$5D,$3F,$E7,$BF,$AE,$7C
	.dc.b	$77,$F5,$C9,$FF,$1A,$4E,$63,$15,$FD,$82,$BE,$93,$FE,$03,$5F,$39
	.dc.b	$9E,$0D,$3B,$F4,$38,$7A,$17,$3F,$EB,$97,$FC,$DD,$FE,$BD,$F5,$2F
	.dc.b	$87,$FE,$DA,$BC,$DB,$18,$73,$E4,$EB,$C9,$B7,$C9,$FE,$EB,$F5,$CF
	.dc.b	$12,$FC,$0E,$1B,$5F,$DB,$67,$FB,$1F,$17,$FA,$75,$FE,$D9,$19,$F7
	.dc.b	$39,$F7,$F4,$E3,$EB,$CB,$6B,$93,$FE,$04,$5E,$BD,$F6,$70,$D8,$FE
	.dc.b	$F9,$F5,$CF,$8C,$FF,$05,$AF,$9C,$B1,$87,$FC,$1F,$F5,$DF,$E2,$4D
	.dc.b	$BF,$D6,$4F,$F8,$BD,$BC,$1A,$F0,$77,$B7,$0F,$1A,$E7,$FE,$0B,$C1
	.dc.b	$1F,$E1,$AF,$AE,$78,$97,$E4,$B0,$DB,$FF,$09,$4F,$FA,$89,$BE,$1F
	.dc.b	$1E,$1E,$4F,$F8,$93,$DC,$F5,$4F,$F8,$65,$FE,$C1,$C9,$F9,$47,$2D
	.dc.b	$F7,$78,$7A,$DC,$7E,$32,$DF,$17,$FC,$5F,$F5,$DE,$F8,$FE,$02,$FF
	.dc.b	$00,$E3,$FF,$6A,$39,$C3,$FC,$AC,$E6,$77,$D7,$CB,$93,$9B,$63,$FC
	.dc.b	$5D,$2F,$4F,$31,$86,$DF,$EB,$27,$F5,$73,$E2,$F8,$6F,$2F,$C3,$F9
	.dc.b	$9F,$3F,$A7,$89,$EA,$DB,$E5,$F0,$DF,$E4,$AE,$4F,$87,$AF,$AA,$31
	.dc.b	$D5,$F0,$3B,$BE,$0C,$5C,$7C,$EF,$F9,$4A,$BF,$55,$87,$F0,$0C,$3F
	.dc.b	$51,$5F,$E4,$99,$C9,$FE,$F9,$8C,$CF,$F0,$68,$79,$1C,$FC,$8E,$8D
	.dc.b	$BF,$F2,$95,$7E,$21,$FF,$2A,$5F,$FA,$65,$CF,$57,$AF,$E2,$FF,$CC
	.dc.b	$B5,$FA,$A3,$F8,$99,$C4,$BD,$BA,$9D,$F8,$5C,$22,$DF,$81,$E7,$A7
	.dc.b	$C7,$F8,$CB,$3E,$27,$BE,$D8,$C7,$BB,$76,$FF,$4C,$F9,$3F,$EF,$79
	.dc.b	$FF,$4D,$35,$34,$FF,$7E,$7D,$67,$EE,$BE,$BA,$F8,$C6,$92,$BF,$A9
	.dc.b	$9F,$EA,$59,$7D,$27,$F4,$18,$C8,$8C,$E1,$DA,$75,$3C,$A7,$2F,$05
	.dc.b	$F4,$9F,$D8,$29,$A7,$F6,$29,$CD,$7F,$A1,$DA,$FF,$42,$79,$FD,$77
	.dc.b	$F8,$B6,$3C,$DC,$BC,$EB,$93,$63,$27,$3C,$30,$FA,$79,$6E,$1F,$88
	.dc.b	$8F,$04,$6B,$E5,$7E,$1E,$5E,$69,$F8,$9F,$1D,$C0,$8F,$03,$D1,$F3
	.dc.b	$1D,$08,$FA,$E3,$C3,$EF,$7F,$B4,$23,$FD,$48,$D7,$FA,$FB,$6B,$C1
	.dc.b	$0F,$94,$FE,$BA,$F0,$E5,$E6,$47,$9F,$3F,$BB,$EB,$BE,$4D,$F2,$6F
	.dc.b	$07,$F8,$41,$B5,$FE,$E3,$C1,$F3,$9C,$78,$64,$FF,$AE,$6B,$FD,$28
	.dc.b	$DE,$1F,$97,$F5,$DF,$CF,$E3,$FD,$7C,$D7,$FB,$CB,$72,$F5,$BF,$E8
	.dc.b	$C8,$DF,$A1,$C8,$8F,$2E,$FE,$BE,$AC,$8E,$AF,$83,$E4,$3C,$3E,$BC
	.dc.b	$9F,$0B,$B7,$C3,$F1,$9E,$62,$3E,$6F,$92,$F4,$F0,$C8,$F4,$BE,$CF
	.dc.b	$8C,$FC,$3F,$D7,$3C,$7A,$4A,$FF,$6E,$47,$FB,$AD,$FF,$6F,$7B,$7F
	.dc.b	$29,$BC,$9F,$90,$8C,$47,$F8,$71,$FB,$6D,$F8,$4D,$11,$E7,$EF,$F1
	.dc.b	$EB,$91,$D9,$E9,$DD,$F9,$CF,$AE,$6D,$49,$5F,$EF,$18,$FF,$0A,$3F
	.dc.b	$EF,$3F,$AE,$F4,$3F,$E2,$E6,$F3,$FE,$41,$BC,$1F,$E5,$65,$E6,$FE
	.dc.b	$AB,$EB,$BD,$35,$F9,$66,$D7,$F9,$71,$B6,$17,$FA,$C9,$D7,$8B,$CC
	.dc.b	$93,$BB,$E9,$E3,$F7,$D2,$4E,$6F,$B4,$78,$FB,$BF,$7A,$FB,$A7,$FC
	.dc.b	$92,$DB,$FD,$5B,$6B,$FC,$E2,$FF,$8D,$67,$A1,$E7,$7F,$CB,$8E,$0F
	.dc.b	$E9,$BF,$0B,$D8,$EF,$E5,$FF,$0C,$FA,$EF,$F1,$B4,$ED,$AF,$F9,$69
	.dc.b	$B8,$7E,$F1,$7A,$BC,$7C,$49,$5C,$5E,$9E,$77,$F1,$8F,$AE,$7C,$4D
	.dc.b	$25,$79,$63,$FC,$42,$DD,$F8,$11,$D7,$BB,$FC,$8B,$EB,$BF,$C7,$11
	.dc.b	$ED,$6B,$FE,$70,$6E,$9F,$C4,$9F,$DD,$70,$7F,$A6,$1F,$F7,$A4,$F1
	.dc.b	$F8,$5E,$87,$5F,$73,$F9,$7F,$07,$F9,$B2,$7E,$53,$5F,$F3,$EB,$70
	.dc.b	$FE,$3F,$E5,$DC,$7F,$D4,$B2,$EC,$E6,$4B,$E7,$F0,$FD,$BB,$A4,$76
	.dc.b	$74,$F6,$F6,$3C,$64,$7E,$E7,$C8,$BE,$9B,$CF,$C7,$27,$CE,$ED,$F3
	.dc.b	$F5,$FB,$08,$FE,$17,$E6,$5F,$4B,$F0,$3D,$09,$3F,$01,$DB,$F0,$3C
	.dc.b	$4F,$8A,$47,$97,$7F,$C8,$F5,$24,$7C,$87,$4F,$C8,$F9,$AF,$9A,$47
	.dc.b	$47,$E5,$FE,$B7,$F9,$73,$FE,$AA,$6F,$37,$F4,$C6,$F2,$FF,$B4,$BE
	.dc.b	$B9,$F5,$DF,$D9,$A7,$EF,$DF,$EE,$5A,$9A,$72,$75,$E9,$71,$23,$FA
	.dc.b	$9F,$AE,$F9,$CF,$97,$D2,$EC,$47,$D1,$3E,$1F,$9E,$91,$EB,$7F,$BC
	.dc.b	$FD,$47,$F0,$C8,$F1,$7C,$F7,$9C,$8E,$CF,$85,$97,$AC,$8F,$C5,$8F
	.dc.b	$77,$81,$23,$E4,$3F,$E0,$6C,$9F,$51,$F0,$FD,$19,$5F,$35,$E0,$FC
	.dc.b	$DF,$57,$E4,$11,$FE,$11,$F5,$CF,$92,$F9,$9C,$52,$7E,$B3,$B7,$FD
	.dc.b	$BF,$F5,$AF,$9B,$FF,$0B,$B6,$3F,$89,$35,$FF,$72,$FD,$77,$FB,$95
	.dc.b	$CF,$83,$C0,$8F,$0A,$E7,$84,$8E,$6E,$9F,$0F,$C6,$F8,$08,$FF,$21
	.dc.b	$FA,$EF,$D9,$3E,$2F,$0F,$CA,$47,$CB,$3E,$8F,$9F,$91,$E8,$7E,$67
	.dc.b	$C2,$7B,$38,$64,$FB,$1D,$BF,$F0,$03,$EB,$BF,$C4,$5B,$FF,$49,$84
	.dc.b	$61,$FE,$98,$D7,$FC,$4D,$E5,$7F,$8F,$A1,$74,$93,$FC,$EE,$38,$7D
	.dc.b	$77,$E2,$C9,$F3,$BF,$E4,$6F,$39,$E6,$7F,$92,$27,$FB,$CF,$85,$FE
	.dc.b	$34,$6E,$7E,$56,$E3,$FD,$7D,$E9,$E6,$71,$24,$7F,$4B,$C3,$F8,$3B
	.dc.b	$FD,$36,$7F,$9E,$B7,$CC,$F0,$1C,$16,$1E,$1F,$1F,$C8,$48,$E9,$F3
	.dc.b	$3E,$7F,$6A,$47,$57,$4F,$2C,$FD,$1E,$57,$B4,$91,$E5,$D3,$E0,$79
	.dc.b	$52,$3C,$CE,$9E,$47,$F7,$DF,$AC,$FE,$8F,$7F,$5D,$B1,$9F,$87,$F9
	.dc.b	$9A,$FF,$9C,$A5,$E3,$52,$57,$F6,$F8,$E2,$BE,$2F,$FC,$64,$FF,$A4
	.dc.b	$8C,$7C,$2F,$FA,$8E,$33,$1F,$E9,$66,$BF,$E9,$C5,$D3,$B2,$47,$38
	.dc.b	$CB,$63,$FD,$05,$D3,$23,$FC,$05,$3A,$7E,$08,$BF,$F0,$53,$86,$F1
	.dc.b	$7F,$9E,$9B,$A7,$C5,$F8,$7F,$85,$E5,$FF,$AE,$E7,$8D,$FF,$AC,$9A
	.dc.b	$FC,$B8,$D7,$99,$23,$E1,$4F,$F8,$7D,$E1,$1F,$EC,$5F,$AE,$FF,$27
	.dc.b	$C7,$FA,$F1,$AF,$EA,$DA,$FF,$8D,$5A,$FF,$A3,$F8,$7F,$D9,$DD,$06
	.dc.b	$92,$3E,$94,$39,$36,$48,$CB,$CB,$D0,$91,$C1,$AF,$FB,$32,$FF,$86
	.dc.b	$B7,$BB,$D4,$47,$DF,$F0,$BF,$1C,$8E,$0F,$87,$F1,$1F,$2C,$8E,$08
	.dc.b	$7F,$B6,$B4,$46,$A3,$0F,$D3,$F4,$3F,$88,$70,$7D,$36,$CF,$E7,$5A
	.dc.b	$FE,$85,$AF,$E0,$FC,$FB,$66,$92,$3F,$CC,$95,$F7,$FD,$0D,$92,$32
	.dc.b	$7E,$AD,$E7,$F2,$F2,$32,$7E,$C5,$C3,$FE,$C8,$3F,$29,$ED,$78,$88
	.dc.b	$F0,$32,$5C,$3F,$C7,$6F,$16,$DF,$E4,$FD,$2F,$59,$EC,$F4,$65,$71
	.dc.b	$6B,$C7,$E5,$1F,$33,$F9,$66,$C7,$D1,$F7,$3F,$AC,$3F,$9F,$8C,$BF
	.dc.b	$A1,$F8,$5F,$F7,$86,$34,$F3,$38,$8F,$E1,$CD,$7F,$D1,$DF,$06,$F4
	.dc.b	$B1,$08,$C5,$7C,$16,$C7,$C7,$6C,$7F,$4F,$8E,$1E,$8F,$CB,$7A,$B5
	.dc.b	$CC,$7F,$13,$9F,$EF,$CF,$F1,$5E,$67,$91,$B6,$C6,$63,$07,$39,$FF
	.dc.b	$56,$C7,$FA,$3F,$EB,$5F,$5A,$8F,$F4,$94,$EB,$0F,$E2,$35,$A9,$CF
	.dc.b	$CC,$6E,$7F,$8A,$D6,$FC,$2C,$47,$C3,$9D,$46,$27,$55,$FD,$BD,$7C
	.dc.b	$C6,$A3,$FB,$97,$D6,$FF,$64,$7C,$71,$9C,$3F,$D7,$D3,$97,$C3,$F1
	.dc.b	$19,$7E,$94,$DB,$4D,$8C,$46,$EB,$FC,$D7,$7F,$02,$31,$F0,$3C,$7D
	.dc.b	$5D,$CF,$C4,$6E,$3F,$CD,$3E,$B3,$FD,$11,$3F,$EB,$9C,$3F,$8B,$3F
	.dc.b	$EB,$B8,$CC,$FF,$3A,$9C,$BC,$7D,$4F,$44,$EB,$4F,$92,$FB,$5B,$E1
	.dc.b	$6B,$E1,$FD,$DF,$5B,$FE,$F2,$CB,$B4,$63,$C4,$E7,$F1,$3B,$9F,$13
	.dc.b	$73,$F4,$7A,$DC,$BC,$6F,$0B,$6B,$D1,$EB,$61,$A7,$AE,$3E,$77,$67
	.dc.b	$EC,$23,$FD,$3B,$E8,$63,$C0,$FA,$5D,$DF,$E5,$C9,$FF,$6E,$7D,$6B
	.dc.b	$E7,$F5,$18,$6E,$9F,$D7,$A5,$E8,$FD,$BF,$88,$79,$BF,$76,$73,$87
	.dc.b	$FA,$8E,$31,$E9,$66,$EC,$F8,$1E,$1C,$1F,$85,$E5,$79,$18,$AF,$F7
	.dc.b	$84,$61,$F3,$F9,$D9,$D3,$C8,$DE,$8F,$CE,$E9,$FC,$BE,$B6,$7B,$BC
	.dc.b	$EF,$35,$79,$FC,$CF,$07,$97,$6B,$BB,$EB,$CF,$19,$7C,$16,$E5,$DB
	.dc.b	$97,$F3,$9F,$50,$F9,$9F,$85,$18,$7E,$27,$C3,$63,$2F,$9E,$DC,$7F
	.dc.b	$C1,$3F,$58,$FA,$AF,$D3,$7D,$6B,$5F,$D4,$4E,$70,$5F,$F8,$42,$B5
	.dc.b	$18,$D3,$F7,$B5,$FE,$89,$8C,$1F,$E6,$97,$95,$78,$3E,$1B,$AE,$3D
	.dc.b	$BA,$BB,$C5,$BE,$35,$E0,$BE,$8F,$A3,$E8,$FF,$79,$F5,$BF,$C5,$6B
	.dc.b	$72,$A5,$6A,$7D,$7F,$0D,$EC,$8B,$1C,$27,$6D,$F7,$FE,$6F,$D6,$FC
	.dc.b	$26,$C6,$FE,$03,$0D,$38,$1C,$F8,$BE,$AE,$DF,$29,$B1,$DF,$CF,$C3
	.dc.b	$4E,$07,$8F,$D0,$FA,$31,$F4,$9B,$1E,$7F,$A4,$FC,$0F,$07,$D6,$6C
	.dc.b	$7A,$F7,$C3,$4E,$07,$C6,$FA,$DF,$E1,$8F,$FA,$77,$A3,$FA,$9F,$27
	.dc.b	$F3,$69,$CD,$7E,$02,$73,$5F,$EA,$39,$FE,$24,$DE,$EF,$E5,$D1,$8A
	.dc.b	$FC,$3B,$C9,$2F,$86,$FC,$CF,$B3,$C1,$B7,$E5,$F8,$5C,$34,$F4,$47
	.dc.b	$6F,$A7,$EF,$36,$75,$87,$F9,$93,$EB,$5D,$BF,$D0,$57,$75,$8C,$3F
	.dc.b	$D1,$77,$7B,$17,$E7,$70,$DC,$7F,$1B,$F5,$AF,$8C,$E1,$9F,$F4,$D8
	.dc.b	$C4,$FE,$4B,$0F,$F1,$4F,$1F,$F6,$09,$F8,$43,$E6,$38,$71,$F1,$79
	.dc.b	$7A,$5C,$FD,$2B,$8F,$81,$FE,$B5,$F5,$0F,$8A,$BE,$7E,$1B,$5F,$A2
	.dc.b	$9F,$DA,$EB,$FE,$86,$7F,$8D,$37,$C8,$E7,$E4,$FF,$65,$FA,$DF,$D0
	.dc.b	$CF,$FA,$31,$7C,$EE,$7C,$7D,$3B,$F8,$0E,$1E,$0D,$CF,$F6,$37,$FD
	.dc.b	$B1,$B8,$75,$E4,$EC,$D8,$FD,$B8,$F6,$3F,$BB,$2F,$6A,$C6,$1F,$EF
	.dc.b	$5F,$5A,$F9,$0B,$F4,$B8,$6D,$FF,$02,$FA,$DF,$4B,$7C,$4F,$F0,$14
	.dc.b	$63,$2F,$F5,$4B,$FE,$28,$6E,$BA,$F2,$36,$DB,$FE,$1B,$F5,$BF,$99
	.dc.b	$6F,$C4,$7F,$83,$63,$17,$FE,$B4,$7F,$C5,$ED,$D7,$5E,$36,$DB,$7F
	.dc.b	$C6,$3D,$E6,$DF,$E1,$77,$FB,$24,$66,$31,$EF,$67,$D7,$D7,$C6,$DA
	.dc.b	$E4,$FF,$87,$FE,$B7,$D9,$E2,$7F,$8C,$1F,$DC,$E9,$FC,$E9,$CD,$7F
	.dc.b	$3C,$84,$69,$BD,$5C,$FA,$BA,$36,$BF,$97,$D3,$D6,$CC,$61,$B7,$FE
	.dc.b	$D6,$9F,$D6,$CF,$0F,$23,$CD,$E4,$79,$DF,$47,$A9,$89,$EA,$DB,$E5
	.dc.b	$F0,$FF,$E3,$BF,$3F,$E3,$D7,$8A,$31,$C3,$F0,$7B,$BE,$14,$59,$D1
	.dc.b	$FD,$1B,$77,$8D,$CF,$C6,$FF,$23,$46,$27,$FA,$5E,$29,$CF,$8F,$AB
	.dc.b	$B5,$CD,$FF,$27,$67,$FE,$58,$8E,$3F,$19,$8E,$DF,$43,$AE,$DC,$4F
	.dc.b	$F5,$8A,$FD,$41,$FE,$FC,$39,$17,$C4,$E1,$1B,$6C,$6F,$5F,$89,$F1
	.dc.b	$BE,$26,$D7,$5D,$F0,$D3,$C1,$1E,$8A,$79,$1F,$E9,$17,$FD,$24,$D7
	.dc.b	$F0,$CD,$7F,$C0,$6E,$7E,$AF,$F7,$5F,$5B,$7C,$73,$49,$5D,$65,$F4
	.dc.b	$4E,$4E,$4D,$9D,$35,$FF,$52,$4F,$09,$F0,$7D,$17,$E6,$5F,$F4,$6C
	.dc.b	$23,$19,$FE,$76,$31,$5F,$E0,$67,$92,$FA,$DF,$F1,$53,$FE,$1E,$8F
	.dc.b	$C2,$BB,$E3,$E6,$E5,$AA,$3A,$BE,$BA,$F0,$72,$6A,$4D,$38,$FC,$C7
	.dc.b	$FA,$FF,$2F,$4F,$CB,$6C,$98,$79,$EB,$97,$CC,$48,$F5,$3A,$7C,$07
	.dc.b	$F8,$DB,$DD,$FF,$68,$3F,$E2,$46,$BA,$3F,$C9,$1C,$BE,$AF,$98,$FE
	.dc.b	$B9,$1B,$E4,$F6,$D1,$A6,$2B,$E0,$FD,$6F,$C0,$CF,$FD,$7A,$DD,$BF
	.dc.b	$86,$37,$0F,$F2,$4E,$2F,$AB,$E3,$C1,$27,$FC,$73,$5F,$41,$F3,$3C
	.dc.b	$57,$D6,$FF,$26,$7F,$C7,$ED,$7D,$8D,$C7,$FD,$9E,$FF,$A3,$63,$7E
	.dc.b	$8F,$D8,$8F,$E9,$51,$83,$9F,$81,$9E,$F9,$1D,$9E,$1F,$1B,$C7,$D7
	.dc.b	$93,$E2,$76,$F8,$BE,$37,$B1,$1F,$3F,$C8,$F8,$78,$24,$7A,$9F,$27
	.dc.b	$BF,$F8,$CF,$AD,$78,$F4,$95,$FE,$5E,$8F,$F0,$BB,$FE,$60,$F6,$FE
	.dc.b	$63,$78,$FF,$2B,$18,$8F,$C0,$3F,$79,$BF,$2B,$A2,$3C,$FD,$EE,$3D
	.dc.b	$92,$3B,$3D,$3B,$3F,$43,$F5,$AD,$A9,$2B,$FC,$E7,$1F,$E1,$77,$FC
	.dc.b	$EB,$F5,$9E,$87,$F2,$4D,$E7,$70,$7F,$53,$8C,$37,$CD,$FD,$7F,$D6
	.dc.b	$FA,$6B,$77,$FE,$94,$8C,$57,$F5,$B6,$C7,$FA,$62,$FF,$5D,$39,$F1
	.dc.b	$79,$92,$7F,$D1,$35,$FE,$13,$6E,$7E,$02,$3A,$EC,$FE,$03,$1E,$D8
	.dc.b	$FD,$63,$71,$F2,$C7,$F0,$0E,$6D,$F4,$70,$F8,$49,$3F,$E9,$8B,$FF
	.dc.b	$2A,$B7,$CB,$FE,$1B,$F5,$BF,$E9,$A8,$FF,$4E,$B5,$FD,$D3,$71,$EE
	.dc.b	$FF,$8E,$23,$7E,$0F,$22,$3C,$BC,$DF,$E3,$3F,$5A,$F8,$8A,$49,$FF
	.dc.b	$55,$C7,$F9,$DD,$BE,$6F,$02,$3A,$F7,$3F,$91,$FD,$6F,$EB,$23,$F8
	.dc.b	$73,$73,$F8,$6D,$E4,$FF,$08,$3F,$B7,$6F,$C2,$F2,$25,$7F,$AF,$25
	.dc.b	$EE,$7F,$30,$DF,$FC,$D5,$3F,$EC,$06,$BF,$C7,$5B,$87,$D2,$F2,$6E
	.dc.b	$3F,$E1,$F9,$76,$F3,$25,$F3,$F8,$6E,$DD,$F2,$3B,$3A,$7B,$7A,$FE
	.dc.b	$32,$3F,$ED,$2F,$92,$7D,$37,$9F,$8E,$4F,$9D,$DB,$E7,$ED,$F6,$11
	.dc.b	$FF,$6D,$FC,$CB,$E7,$BE,$07,$AD,$27,$E0,$3B,$7E,$07,$8B,$F1,$48
	.dc.b	$F2,$F0,$3E,$47,$A7,$23,$E4,$3A,$7E,$47,$CD,$7C,$D2,$3A,$3F,$DD
	.dc.b	$BF,$59,$FC,$AD,$FE,$94,$DE,$6F,$CF,$B7,$97,$FC,$99,$F5,$AF,$A5
	.dc.b	$FE,$CF,$3F,$7E,$C1,$C3,$7D,$2D,$FC,$37,$0F,$1A,$4F,$89,$FE,$E3
	.dc.b	$F5,$BF,$3A,$3C,$BE,$47,$62,$3B,$3D,$3F,$39,$E2,$23,$E2,$D7,$F6
	.dc.b	$FF,$4F,$FC,$22,$3C,$9F,$3D,$E8,$23,$C8,$F8,$7A,$FB,$08,$FC,$68
	.dc.b	$F7,$F8,$72,$3E,$4B,$E8,$FA,$0F,$87,$E9,$49,$F9,$AE,$DF,$9B,$EB
	.dc.b	$7C,$82,$3E,$7E,$3F,$F3,$C7,$D6,$BE,$4B,$E6,$C4,$9F,$AE,$F0,$7F
	.dc.b	$C0,$1F,$5A,$F9,$BF,$F0,$CB,$63,$A1,$F5,$1C,$3F,$E0,$7F,$AD,$7C
	.dc.b	$47,$0F,$12,$4F,$01,$48,$EB,$7D,$B1,$95,$BA,$F4,$36,$BC,$7F,$1B
	.dc.b	$23,$4F,$F8,$BB,$EB,$7F,$6F,$1C,$7C,$6F,$49,$1D,$9D,$7E,$7F,$DA
	.dc.b	$47,$CD,$D3,$FE,$1B,$7C,$5C,$5F,$08,$8F,$13,$D9,$F5,$BF,$E2,$8E
	.dc.b	$8F,$E7,$6D,$E1,$FF,$54,$36,$7F,$C3,$BE,$67,$F9,$12,$3B,$47,$F8
	.dc.b	$BD,$B5,$CB,$E1,$B5,$EB,$7F,$92,$BE,$B7,$F2,$04,$F9,$DF,$F1,$A7
	.dc.b	$9F,$9F,$81,$99,$7C,$FA,$FE,$F9,$8F,$F4,$E7,$8F,$A1,$CC,$93,$F5
	.dc.b	$7D,$3F,$E1,$4F,$1F,$47,$C0,$47,$A4,$F8,$FC,$C9,$1D,$DF,$F2,$37
	.dc.b	$CC,$F9,$FD,$C9,$3E,$37,$6F,$F9,$37,$17,$9F,$ED,$24,$78,$F0,$F8
	.dc.b	$1E,$7C,$8F,$63,$A7,$FC,$AD,$F5,$BF,$D1,$DF,$E0,$ED,$4D,$D3,$FA
	.dc.b	$D4,$7E,$4E,$1E,$47,$F9,$D9,$B1,$98,$C7,$C4,$FF,$9C,$BE,$4F,$FC
	.dc.b	$E7,$F0,$F2,$BF,$39,$F1,$BF,$E1,$27,$F8,$BA,$ED,$FF,$3C,$92,$37
	.dc.b	$35,$FF,$55,$4E,$DF,$F5,$53,$5C,$CF,$FA,$AF,$17,$C5,$52,$4F,$FA
	.dc.b	$C6,$32,$DF,$43,$D8,$47,$E0,$4D,$7F,$D7,$53,$FE,$D9,$78,$10,$CF
	.dc.b	$A1,$E1,$23,$DD,$FF,$5F,$6B,$FE,$98,$8F,$F6,$13,$5F,$F6,$33,$5F
	.dc.b	$E7,$67,$F9,$FB,$8B,$8E,$F2,$38,$B5,$E3,$C3,$D8,$8F,$67,$FD,$9D
	.dc.b	$ED,$7F,$8B,$63,$A8,$70,$F8,$1A,$FF,$4C,$6A,$FA,$DF,$E1,$2F,$FB
	.dc.b	$75,$AF,$46,$BE,$A6,$BF,$90,$F8,$9F,$AC,$47,$EA,$35,$F9,$ED,$7C
	.dc.b	$F9,$FF,$99,$5F,$B5,$F1,$7F,$7C,$8E,$0F,$BD,$EA,$7C,$AF,$0F,$8B
	.dc.b	$63,$E6,$D2,$31,$1F,$A4,$6A,$D7,$FD,$55,$5F,$21,$B1,$F5,$9A,$FF
	.dc.b	$AD,$3A,$B8,$DF,$F7,$78,$FE,$EB,$5E,$AF,$53,$24,$8C,$9E,$4F,$98
	.dc.b	$E2,$47,$27,$EE,$1F,$2E,$AF,$56,$92,$76,$7D,$D7,$C7,$EA,$C8,$CE
	.dc.b	$28,$C4,$7F,$9B,$8F,$E0,$7E,$AF,$A8,$8F,$07,$DB,$F2,$8F,$99,$FD
	.dc.b	$63,$67,$F4,$FF,$03,$FD,$79,$FD,$4C,$65,$FC,$8F,$91,$FD,$A7,$6F
	.dc.b	$3B,$63,$C4,$D8,$DF,$A9,$FE,$96,$78,$FA,$56,$BF,$4F,$5F,$D9,$E7
	.dc.b	$9C,$E6,$7F,$C4,$B3,$97,$AF,$CC,$FB,$DC,$B5,$42,$30,$DA,$FD,$4C
	.dc.b	$F4,$F8,$DF,$F5,$4F,$93,$FE,$96,$FA,$C7,$D6,$E2,$FC,$B5,$6F,$FC
	.dc.b	$6F,$5A,$9F,$F0,$06,$0F,$17,$55,$E2,$BB,$39,$AD,$C6,$3E,$67,$7F
	.dc.b	$5A,$F3,$F1,$5F,$E0,$B8,$C3,$DB,$E3,$73,$8B,$5F,$F5,$6E,$49,$CC
	.dc.b	$62,$3A,$FC,$37,$89,$B7,$E8,$F4,$F0,$D3,$D7,$1E,$AF,$AD,$FF,$42
	.dc.b	$C7,$8F,$C4,$C7,$AD,$E8,$75,$AE,$27,$FC,$45,$5E,$8D,$3C,$D1,$FC
	.dc.b	$92,$39,$C6,$0F,$F3,$0A,$F8,$CF,$93,$F5,$BF,$EE,$BA,$E7,$18,$F1
	.dc.b	$BC,$9F,$1D,$E5,$8B,$3A,$3C,$3E,$27,$2F,$1F,$D4,$D8,$F2,$F9,$58
	.dc.b	$69,$C0,$FF,$1D,$7C,$8F,$E9,$5F,$F4,$D7,$B7,$F7,$23,$1A,$FA,$4F
	.dc.b	$F1,$F4,$7E,$03,$EB,$5F,$0F,$BF,$0F,$0D,$C3,$FC,$8D,$3F,$D7,$9F
	.dc.b	$F4,$EF,$C5,$F1,$F4,$7F,$13,$F1,$A8,$E9,$E9,$E6,$B8,$B7,$3F,$16
	.dc.b	$0D,$FF,$23,$CF,$FC,$16,$FB,$C6,$3C,$9F,$37,$A1,$E7,$8B,$3A,$3C
	.dc.b	$BE,$66,$31,$EA,$DE,$EF,$F7,$DE,$5F,$ED,$2A,$FF,$30,$64,$F2,$FE
	.dc.b	$17,$C7,$F8,$9F,$89,$A6,$2B,$E0,$B7,$2F,$4F,$27,$EA,$3E,$A1,$F3
	.dc.b	$5F,$2E,$30,$FC,$DF,$8C,$C6,$1F,$29,$B7,$F4,$FE,$AD,$F5,$5F,$B2
	.dc.b	$FA,$DF,$25,$FF,$9D,$67,$2F,$F9,$DE,$33,$5C,$7D,$0F,$DC,$D6,$A3
	.dc.b	$15,$C6,$F3,$E8,$6A,$EC,$F5,$78,$FD,$1F,$F0,$2F,$AD,$4E,$0F,$F4
	.dc.b	$33,$87,$C9,$F1,$99,$C7,$D4,$DE,$5E,$31,$FE,$C6,$CD,$F0,$1F,$1F
	.dc.b	$EB,$7C,$ED,$C3,$DD,$F0,$18,$69,$F0,$63,$DB,$C3,$F3,$F7,$B3,$FE
	.dc.b	$99,$E3,$F8,$19,$F8,$33,$66,$F4,$77,$1F,$3D,$B3,$B9,$FE,$25,$7F
	.dc.b	$3E,$7E,$5F,$A1,$F3,$1F,$9B,$F2,$1F,$7F,$8A,$D7,$85,$FA,$38,$E0
	.dc.b	$FB,$85,$C7,$C9,$F5,$BF,$C2,$9F,$36,$7F,$D1,$5C,$F5,$FE,$2F,$97
	.dc.b	$F9,$9D,$7F,$13,$9C,$E3,$FE,$B7,$8C,$3F,$E3,$87,$A3,$7E,$1B,$F2
	.dc.b	$9F,$23,$C9,$B1,$C7,$E2,$B0,$D3,$A4,$78,$3E,$9F,$C3,$36,$75,$A7
	.dc.b	$F0,$9F,$AD,$76,$7F,$48,$5D,$B6,$30,$FF,$4C,$E1,$FE,$95,$5F,$C4
	.dc.b	$57,$B3,$9F,$6B,$4F,$F5,$2F,$AD,$7C,$7F,$1B,$C3,$73,$FF,$6C,$4F
	.dc.b	$C0,$90,$7F,$C9,$D1,$AE,$3E,$9F,$07,$D7,$D1,$ED,$E9,$5A,$E6,$FF
	.dc.b	$62,$FA,$87,$C5,$5F,$3B,$0D,$AF,$F7,$44,$FE,$CB,$A3,$F2,$0F,$F1
	.dc.b	$9D,$3F,$97,$39,$9F,$F3,$6B,$99,$FC,$B7,$D6,$BC,$7F,$EE,$2F,$F7
	.dc.b	$FB,$FE,$6C,$E6,$FF,$DE,$CE,$6B,$8F,$BB,$C3,$C0,$E2,$DA,$DD,$F4
	.dc.b	$E4,$EA,$D8,$FF,$80,$03,$D8,$FF,$00,$BF,$D6,$CC,$B8,$F6,$3F,$C0
	.dc.b	$BF,$5A,$F8,$EB,$F3,$58,$6D,$FD,$2D,$3F,$65,$F5,$BE,$78,$FF,$46
	.dc.b	$AF,$07,$FC,$1E,$E7,$3F,$B2,$E6,$79,$1C,$36,$DF,$8D,$C1,$B1,$AA
	.dc.b	$FC,$0F,$D6,$BC,$3F,$F0,$DA,$F9,$4B,$18,$73,$BE,$BC,$4B,$87,$3C
	.dc.b	$18,$27,$FC,$4B,$EE,$B8,$2F,$BB,$86,$E1,$C6,$7F,$A7,$61,$FD,$BD
	.dc.b	$7C,$6F,$2F,$5F,$9B,$FE,$30,$FA,$DF,$81,$5F,$DD,$5F,$E0,$97,$FE
	.dc.b	$B4,$73,$C7,$FC,$75,$CC,$FE,$81,$FC,$33,$C7,$E2,$FA,$75,$F5,$7B
	.dc.b	$7A,$B8,$C3,$DD,$68,$C5,$F8,$7C,$7F,$27,$8F,$E5,$6D,$F0,$F4,$F0
	.dc.b	$D3,$9C,$7A,$FE,$1F,$FC,$97,$CD,$F1,$E7,$BC,$63,$87,$E0,$77,$6E
	.dc.b	$7C,$6D,$CB,$FC,$2D,$BB,$C4,$E7,$E2,$7F,$95,$A3,$17,$F5,$F2,$4E
	.dc.b	$7C,$5D,$5D,$AF,$B9,$7C,$30,$FF,$11,$1F,$D1,$63,$97,$C6,$7F,$98
	.dc.b	$A3,$07,$F8,$9F,$14,$67,$B3,$5D,$B6,$FF,$11,$7C,$30,$FF,$45,$C6
	.dc.b	$2B,$A7,$87,$BA,$F0,$EF,$F1,$BE,$37,$C0,$DB,$E7,$BE,$1A,$76,$8F
	.dc.b	$35,$3C,$6F,$F4,$2B,$FC,$38,$A6,$4A,$EB,$EA,$FF,$80,$7D,$6D,$F1
	.dc.b	$CD,$25,$7F,$A4,$25,$F4,$7F,$E7,$98,$C3,$5F,$F7,$EC,$6B,$6E,$D3
	.dc.b	$99,$FE,$6A,$25,$F4,$78,$AF,$EF,$11,$89,$FD,$93,$5F,$F2,$C3,$D5
	.dc.b	$7D,$6F,$F8,$C1,$FE,$83,$1F,$87,$78,$F2,$75,$D7,$C3,$39,$9E,$78
	.dc.b	$65,$E3,$CB,$5F,$33,$64,$E7,$E6,$3D,$73,$E4,$F8,$FD,$11,$F4,$CF
	.dc.b	$1F,$98,$91,$EC,$74,$FF,$9C,$FD,$EF,$F6,$0B,$FC,$45,$AE,$79,$7C
	.dc.b	$9F,$2F,$FD,$92,$35,$C9,$E7,$A3,$4C,$57,$AF,$EB,$7E,$57,$0F,$F8
	.dc.b	$95,$BB,$FD,$46,$E1,$FD,$AB,$17,$CE,$71,$E1,$93,$FC,$96,$BE,$56
	.dc.b	$1C,$F8,$7E,$47,$D6,$FF,$25,$AF,$E5,$4D,$7F,$D2,$0D,$9F,$F1,$B3
	.dc.b	$FE,$90,$8D,$FA,$3C,$88,$F2,$F0,$3A,$F8,$12,$3A,$BE,$0F,$8F,$F0
	.dc.b	$FA,$B2,$7C,$2E,$DF,$0F,$C6,$79,$88,$F9,$BE,$47,$D3,$C3,$23,$D2
	.dc.b	$FB,$3D,$DF,$C2,$7D,$6F,$F9,$6A,$7F,$A3,$35,$FF,$56,$37,$6F,$F2
	.dc.b	$93,$FE,$ED,$8F,$E9,$9F,$19,$F3,$1B,$C7,$F8,$F8,$C4,$7F,$AD,$5F
	.dc.b	$BC,$DF,$95,$D1,$1E,$7E,$EF,$1E,$C9,$1D,$9E,$9D,$9F,$9B,$FA,$D6
	.dc.b	$D4,$95,$FD,$7E,$3F,$C3,$EF,$F6,$1F,$AD,$F4,$3F,$EC,$76,$EB,$FE
	.dc.b	$6D,$6E,$FF,$BC,$7F,$C7,$73,$FD,$A7,$EB,$7D,$35,$FE,$73,$6D,$7F
	.dc.b	$02,$6C,$FF,$8E,$2B,$F5,$75,$BF,$0F,$F6,$55,$FC,$42,$BA,$46,$3C
	.dc.b	$5C,$E7,$27,$C0,$F0,$EC,$FD,$D3,$EC,$8F,$F7,$03,$70,$FF,$43,$37
	.dc.b	$2F,$E2,$AF,$F4,$A9,$FF,$01,$3F,$E5,$C6,$FC,$1F,$43,$AF,$93,$FC
	.dc.b	$1F,$EB,$7F,$C7,$53,$FE,$09,$6B,$FE,$EF,$6E,$1F,$E9,$55,$EA,$71
	.dc.b	$F0,$A5,$71,$7A,$79,$9F,$C4,$FE,$B5,$F0,$F4,$95,$FE,$17,$8F,$F1
	.dc.b	$03,$77,$E0,$47,$5E,$F7,$F1,$DF,$AD,$FF,$1D,$47,$F8,$7D,$AB,$4F
	.dc.b	$F5,$54,$61,$BE,$17,$1F,$93,$95,$C5,$E9,$EE,$7F,$29,$E0,$7C,$35
	.dc.b	$25,$7F,$8B,$7C,$AC,$BF,$E2,$F9,$76,$F1,$24,$66,$BA,$77,$48,$E0
	.dc.b	$E9,$F4,$73,$FF,$4E,$BF,$87,$18,$3F,$D0,$AE,$34,$FF,$63,$36,$7F
	.dc.b	$AA,$FC,$5B,$E9,$FD,$5C,$73,$0F,$33,$ED,$F6,$3D,$DD,$92,$7D,$4F
	.dc.b	$F4,$9F,$99,$7C,$EF,$C3,$F3,$A5,$7B,$DE,$0F,$94,$68,$F1,$3E,$51
	.dc.b	$23,$FC,$A9,$7F,$CC,$35,$F7,$7A,$5F,$3C,$93,$F2,$EB,$E9,$7C,$D4
	.dc.b	$8F,$9C,$E9,$F0,$7D,$67,$F3,$16,$FC,$2F,$F6,$18,$C7,$17,$FB,$AF
	.dc.b	$EB,$5F,$49,$9E,$B9,$87,$F9,$9E,$FF,$9E,$69,$FD,$F9,$E2,$DF,$4B
	.dc.b	$A1,$27,$FC,$DD,$F5,$3F,$3D,$E5,$E9,$78,$48,$F5,$9F,$27,$CF,$48
	.dc.b	$F0,$3F,$DD,$FE,$9F,$F8,$5C,$31,$7C,$FF,$AC,$91,$C0,$F4,$69,$EF
	.dc.b	$23,$EB,$8F,$85,$E0,$C8,$F7,$3F,$E0,$5C,$DF,$41,$F1,$FD,$29,$5F
	.dc.b	$0D,$E0,$FC,$EF,$57,$E6,$11,$FF,$47,$7D,$6F,$FA,$8D,$F9,$5C,$1F
	.dc.b	$44,$8F,$CD,$9F,$F0,$A7,$D6,$3E,$6E,$92,$7F,$D3,$1F,$5A,$F8,$6D
	.dc.b	$F8,$92,$7F,$D3,$96,$8E,$11,$F9,$18,$F1,$37,$E5,$7B,$9C,$5D,$7C
	.dc.b	$6F,$88,$8F,$FA,$9F,$EB,$7F,$72,$F6,$F1,$FC,$E4,$7C,$33,$EA,$FA
	.dc.b	$09,$1E,$47,$E6,$7C,$27,$BB,$8A,$4F,$A1,$DB,$FA,$7F,$AD,$FF,$12
	.dc.b	$F7,$B8,$BF,$C9,$B1,$86,$FA,$BF,$E2,$8F,$3B,$FC,$79,$3F,$E2,$6B
	.dc.b	$FF,$5F,$35,$E9,$8E,$AB,$FC,$85,$F5,$BF,$8C,$27,$CE,$FF,$8C,$BC
	.dc.b	$FC,$FC,$0C,$C3,$FD,$91,$5F,$DE,$B4,$F2,$40,$47,$3F,$3F,$A9,$23
	.dc.b	$F9,$27,$9D,$E0,$F4,$24,$74,$74,$FF,$07,$AF,$E0,$4F,$56,$BD,$BD
	.dc.b	$DF,$F2,$3F,$CC,$FA,$BD,$E9,$3E,$67,$DB,$3E,$EF,$3A,$4F,$A9,$FF
	.dc.b	$27,$E5,$F0,$BC,$F9,$3E,$E7,$6F,$F1,$9F,$AC,$FE,$8D,$CB,$FE,$7E
	.dc.b	$6C,$7C,$16,$AF,$5B,$FA,$DB,$FE,$EF,$87,$95,$FE,$71,$6C,$66,$31
	.dc.b	$F0,$FF,$E7,$5F,$C1,$FF,$C0,$F8,$7F,$A3,$1B,$5F,$3D,$A9,$FD,$23
	.dc.b	$FE,$FD,$73,$F5,$DF,$C2,$0F,$E1,$2E,$39,$BF,$D5,$31,$FE,$32,$9F
	.dc.b	$F3,$FB,$5F,$F4,$DB,$5F,$BD,$F1,$AE,$DF,$F5,$8B,$63,$FD,$6A,$D7
	.dc.b	$FD,$12,$1C,$3F,$E8,$C7,$5C,$DF,$8A,$8D,$76,$FF,$A4,$4F,$F4,$95
	.dc.b	$58,$EB,$1E,$02,$3E,$0F,$AE,$FF,$4C,$11,$FC,$A5,$75,$BF,$E3,$71
	.dc.b	$8C,$FF,$D6,$F0,$F8,$1F,$59,$27,$D6,$D7,$FA,$D4,$7E,$C5,$F8,$7D
	.dc.b	$9F,$EA,$79,$FF,$54,$3F,$10,$91,$F1,$1F,$F6,$BF,$07,$F8,$C6,$3F
	.dc.b	$19,$E9,$7E,$0C,$FC,$27,$C5,$D8,$FF,$6F,$CE,$DF,$F5,$9B,$5F,$C9
	.dc.b	$4F,$4F,$52,$47,$06,$BF,$4A,$7F,$C8,$CF,$28,$CB,$4F,$F7,$6E,$FF
	.dc.b	$E7,$67,$F0,$7D,$DF,$86,$23,$FE,$F2,$E5,$7C,$07,$FB,$C9,$B1,$98
	.dc.b	$C1,$9E,$99,$1C,$5A,$FE,$1A,$5E,$CF,$32,$47,$77,$B7,$D7,$C8,$F1
	.dc.b	$35,$F7,$7C,$57,$CE,$23,$F2,$9F,$A3,$F1,$92,$3E,$6B,$5F,$8D,$EC
	.dc.b	$FF,$BE,$B9,$3E,$7B,$63,$F8,$03,$5F,$E0,$0D,$5B,$FF,$74,$7F,$92
	.dc.b	$4F,$D8,$F5,$F5,$39,$9F,$D0,$1C,$23,$3F,$2A,$F9,$CF,$C0,$B6,$7E
	.dc.b	$D7,$C0,$7E,$F9,$FC,$1C,$65,$FC,$2F,$C5,$7F,$34,$C6,$9E,$28,$FE
	.dc.b	$31,$19,$6C,$65,$FE,$75,$F0,$9F,$D3,$E3,$11,$FE,$5E,$78,$3A,$6B
	.dc.b	$97,$7D,$0F,$99,$FE,$8B,$D9,$FB,$DF,$AF,$77,$62,$FF,$A4,$60,$C3
	.dc.b	$84,$7E,$D6,$3F,$A6,$57,$08,$C4,$6F,$E6,$79,$3F,$D4,$6B,$8B,$FE
	.dc.b	$4F,$51,$A8,$ED,$8D,$97,$CC,$E7,$EB,$DE,$86,$63,$0E,$FC,$1C,$65
	.dc.b	$E2,$BF,$E1,$0D,$AE,$91,$CF,$D0,$B3,$E8,$C3,$FC,$CE,$F1,$F4,$73
	.dc.b	$9F,$9D,$BB,$FF,$23,$FA,$FF,$F4,$3D,$71,$EE,$F5,$46,$2F,$FC,$D9
	.dc.b	$87,$F9,$AE,$BE,$2C,$FA,$9B,$EA,$70,$DB,$8B,$CB,$C0,$F8,$9F,$5F
	.dc.b	$FE,$E8,$AE,$31,$8E,$EE,$FF,$1D,$B9,$EA,$DC,$FF,$CF,$99,$3D,$1E
	.dc.b	$7E,$07,$99,$B7,$E4,$F4,$B0,$D3,$CF,$1F,$57,$AF,$F4,$11,$F7,$BD
	.dc.b	$CF,$F2,$8B,$F8,$08,$CE,$3F,$E8,$B9,$CB,$FE,$8D,$FA,$F7,$C5,$73
	.dc.b	$F0,$7F,$01,$18,$C3,$94,$63,$CC,$FC,$27,$C5,$FF,$10,$9F,$C6,$D2
	.dc.b	$73,$E4,$72,$F3,$78,$56,$9B,$97,$87,$7F,$C7,$72,$BE,$1F,$C3,$18
	.dc.b	$7C,$7E,$4E,$30,$EE,$DB,$F5,$79,$1A,$9F,$1B,$63,$DD,$F1,$DD,$BE
	.dc.b	$1F,$DA,$DA,$F4,$E1,$86,$9E,$C8,$F8,$7C,$7F,$A3,$FA,$A7,$CD,$62
	.dc.b	$BF,$5B,$18,$7E,$1F,$C4,$67,$2F,$9C,$DF,$07,$FB,$C7,$D5,$BE,$B3
	.dc.b	$F5,$DF,$5E,$DB,$F6,$13,$9C,$7F,$73,$39,$7F,$77,$1F,$E7,$F6,$D7
	.dc.b	$0F,$8B,$E2,$F2,$EB,$C4,$F0,$6E,$7F,$EC,$DB,$7D,$27,$EF,$FE,$BF
	.dc.b	$F8,$D4,$E6,$31,$7E,$4F,$87,$F2,$C5,$8F,$1B,$E8,$FA,$4F,$8D,$F5
	.dc.b	$F6,$7E,$2B,$EC,$F7,$F1,$87,$85,$BC,$1F,$13,$8F,$E7,$6D,$F0,$9B
	.dc.b	$87,$C0,$F3,$F0,$D3,$E1,$C7,$C8,$F4,$3E,$4C,$D9,$D6,$3E,$FF,$8E
	.dc.b	$FB,$5C,$5F,$39,$B3,$F3,$27,$0D,$3E,$7C,$7F,$23,$FA,$FF,$E1,$59
	.dc.b	$FF,$8C,$79,$FF,$D5,$15,$FC,$BA,$72,$FF,$36,$8C,$DF,$FB,$EE,$7F
	.dc.b	$87,$37,$BB,$A7,$9B,$D5,$ED,$F0,$FF,$8F,$F8,$EB,$8F,$E3,$9F,$17
	.dc.b	$C5,$E3,$0E,$8D,$CF,$FC,$89,$93,$EA,$77,$87,$95,$B6,$AF,$F2,$57
	.dc.b	$D7,$BB,$BF,$A0,$2E,$FB,$18,$7F,$A2,$70,$7F,$46,$C7,$FA,$6B,$FD
	.dc.b	$2A,$FF,$CA,$8E,$70,$ED,$3F,$D6,$23,$4F,$F5,$AF,$AF,$7C,$47,$4A
	.dc.b	$FE,$A1,$18,$9E,$B1,$8E,$0C,$FB,$D6,$7F,$AE,$D7,$F6,$FA,$EE,$7F
	.dc.b	$AB,$57,$A1,$D3,$D0,$EA,$DA,$FE,$CB,$F5,$2F,$87,$FE,$C0,$BC,$CB
	.dc.b	$18,$73,$E3,$EB,$C9,$B7,$C9,$FE,$D1,$F5,$FF,$AC,$8F,$EE,$2B,$E8
	.dc.b	$33,$E2,$E9,$DF,$BD,$C3,$C1,$B7,$C1,$E7,$C1,$D3,$92,$CF,$47,$FB
	.dc.b	$B2,$F5,$EF,$B3,$86,$C7,$F8,$0B,$EB,$DF,$1D,$FD,$F5,$7C,$E5,$8C
	.dc.b	$3F,$E0,$1F,$AF,$F4,$57,$F8,$29,$78,$59,$DB,$4E,$F5,$E1,$E3,$DB
	.dc.b	$E0,$FF,$82,$FE,$BD,$E2,$DF,$94,$C3,$6B,$FC,$2F,$3F,$E6,$66,$F8
	.dc.b	$9C,$78,$39,$3F,$E1,$5F,$73,$D3,$3F,$E1,$E7,$FA,$C7,$1F,$E0,$8F
	.dc.b	$F2,$26,$FB,$F7,$78,$7A,$DC,$7E,$BE,$E1,$C5,$FF,$10,$7D,$7F,$BA
	.dc.b	$3F,$C5,$2F,$EE,$BC,$6F,$F5,$83,$9C,$7B,$39,$9F,$F1,$AB,$F8,$18
	.dc.b	$CF,$1F,$4E,$AE,$AF,$6F,$53,$17,$D9,$B5,$D3,$E2,$3C,$7F,$11,$E4
	.dc.b	$6D,$79,$3D,$6C,$34,$F3,$C7,$AB,$E1,$FF,$C7,$5C,$BF,$1B,$0F,$3C
	.dc.b	$63,$87,$DF,$DD,$BC,$3D,$ED,$AF,$95,$BF,$F8,$5E,$7F,$C9,$78,$7E
	.dc.b	$95,$FF,$26,$C6,$67,$E8,$4E,$5E,$3E,$37,$F9,$2E,$74,$7F,$B6,$4F
	.dc.b	$DA,$D7,$2F,$1F,$FC,$A1,$5F,$E4,$E8,$EF,$F1,$78,$EB,$F5,$BA,$EC
	.dc.b	$C4,$FF,$97,$AB,$FC,$9A,$7E,$04,$E4,$5F,$0B,$84,$6D,$B1,$BD,$BE
	.dc.b	$3F,$C6,$78,$9B,$5D,$76,$C3,$4F,$04,$7A,$29,$E4,$FF,$BE,$9F,$F3
	.dc.b	$BB,$51,$A3,$0E,$FE,$B7,$F7,$5F,$5F,$7C,$63,$49,$5F,$E8,$23,$9C
	.dc.b	$BE,$97,$F9,$E4,$64,$46,$78,$7B,$4E,$A7,$94,$E5,$E0,$BE,$97,$FD
	.dc.b	$1F,$4C,$FF,$D2,$33,$9A,$FF,$46,$B5,$FD,$83,$CF,$EB,$FF,$C4,$F1
	.dc.b	$E6,$E3,$E7,$5C,$9B,$19,$39,$E1,$87,$D3,$C9,$70,$FE,$77,$1E,$08
	.dc.b	$D7,$CB,$FC,$3C,$BC,$D3,$F1,$3E,$3F,$81,$1E,$07,$A3,$E6,$FA,$11
	.dc.b	$F5,$C7,$87,$DE,$FF,$57,$47,$FA,$A9,$AF,$F0,$D6,$D7,$82,$1F,$2D
	.dc.b	$FD,$71,$E1,$C7,$CC,$8F,$3E,$6F,$77,$D7,$FC,$9B,$E4,$DE,$0F,$F5
	.dc.b	$FB,$6B,$F1,$38,$3E,$63,$8F,$0C,$9F,$F5,$FD,$7F,$A3,$5B,$C3,$F2
	.dc.b	$FE,$BF,$F9,$FC,$7F,$B1,$1A,$FF,$27,$6E,$5E,$B7,$FD,$0B,$1B,$F4
	.dc.b	$39,$11,$E5,$DF,$D7,$D5,$91,$D5,$F0,$7C,$77,$87,$D7,$93,$E1,$76
	.dc.b	$F8,$7E,$37,$CC,$47,$CD,$F2,$5E,$9E,$19,$1E,$97,$D9,$EF,$FE,$1F
	.dc.b	$EB,$DE,$45,$25,$7F,$B8,$A3,$FD,$BA,$FF,$B9,$3D,$BF,$94,$DE,$4F
	.dc.b	$C8,$46,$23,$FA,$53,$F6,$DB,$F0,$9A,$23,$CF,$DF,$E3,$D7,$23,$B3
	.dc.b	$D3,$BB,$F3,$9F,$5E,$DA,$92,$BF,$DE,$F1,$FE,$13,$7F,$DF,$3F,$5F
	.dc.b	$E8,$7F,$AE,$37,$9F,$F2,$0D,$E0,$FE,$1B,$28,$79,$DF,$AB,$FA,$FF
	.dc.b	$55,$FE,$5D,$B7,$C1,$B7,$FC,$5A,$FF,$59,$39,$F1,$79,$12,$7F,$8D
	.dc.b	$DF,$3F,$1F,$B6,$92,$73,$7D,$63,$C5,$DD,$FB,$D7,$DD,$3F,$DD,$DB
	.dc.b	$7F,$AB,$6D,$7B,$1F,$F1,$A4,$F4,$3F,$95,$BF,$E5,$77,$07,$F9,$03
	.dc.b	$E1,$3B,$1D,$FC,$BF,$E1,$9F,$5F,$FE,$A0,$9D,$B5,$FF,$04,$37,$0F
	.dc.b	$DE,$3F,$E3,$98,$DF,$87,$C8,$8F,$2F,$37,$F8,$C7,$D7,$BE,$26,$92
	.dc.b	$7C,$B1,$FE,$22,$6E,$DC,$08,$EB,$DD,$FE,$45,$F5,$FF,$E3,$88,$F6
	.dc.b	$B5,$FF,$0F,$37,$3F,$E2,$4F,$EE,$B7,$FC,$1B,$FE,$ED,$9E,$3F,$0B
	.dc.b	$D0,$EB,$EE,$7F,$2F,$DD,$F0,$F4,$95,$F3,$3C,$BC,$BF,$3A,$5D,$9C
	.dc.b	$49,$19,$AE,$9D,$B2,$38,$3A,$7F,$C6,$D5,$F0,$DB,$EB,$F8,$CF,$E8
	.dc.b	$5F,$24,$FA,$7F,$37,$1C,$AF,$13,$B7,$C9,$D5,$EB,$23,$F8,$4F,$99
	.dc.b	$7C,$EF,$BF,$CF,$93,$EB,$76,$F9,$3C,$3F,$88,$47,$97,$77,$C7,$F4
	.dc.b	$E4,$7C,$37,$4F,$B3,$E6,$7E,$61,$1D,$1F,$96,$FA,$EF,$E5,$AF,$F9
	.dc.b	$5D,$BD,$3F,$D2,$DB,$D3,$FB,$CF,$AF,$7D,$7F,$F6,$59,$FB,$D7,$FE
	.dc.b	$0C,$6A,$69,$C9,$D7,$A3,$C4,$8F,$EA,$3E,$BF,$E7,$3E,$5F,$47,$B1
	.dc.b	$1F,$44,$F8,$7E,$76,$47,$AD,$FE,$F1,$F5,$3F,$C2,$A3,$C5,$F3,$DE
	.dc.b	$72,$3B,$3E,$16,$5E,$B2,$3F,$16,$3D,$DE,$04,$8F,$90,$FF,$81,$72
	.dc.b	$7D,$47,$C3,$F4,$25,$7C,$D7,$83,$F3,$7D,$4F,$90,$47,$F8,$3F,$D7
	.dc.b	$FF,$A7,$1F,$A9,$C3,$F3,$48,$FD,$79,$FF,$0A,$7D,$73,$E6,$E9,$27
	.dc.b	$F8,$8F,$D7,$BE,$1B,$7E,$0C,$9F,$E2,$B6,$8E,$11,$FE,$97,$8F,$13
	.dc.b	$7E,$53,$B9,$C5,$D7,$C5,$F8,$88,$FF,$20,$FA,$FF,$D9,$BD,$BC,$4F
	.dc.b	$39,$1F,$0C,$FA,$BE,$7E,$47,$91,$F9,$9F,$09,$EE,$E2,$93,$E8,$76
	.dc.b	$FF,$AB,$7E,$BF,$FC,$43,$C3,$FD,$41,$B1,$FE,$5E,$6B,$FE,$24,$F3
	.dc.b	$3F,$C7,$90,$BA,$49,$FE,$73,$1C,$3E,$BF,$F1,$44,$F9,$DF,$F2,$27
	.dc.b	$9E,$F3,$BF,$C8,$D3,$FD,$DF,$C3,$FF,$19,$37,$3F,$EC,$6D,$C7,$FC
	.dc.b	$E2,$F4,$F3,$78,$92,$3F,$8F,$79,$5D,$3C,$D9,$1F,$D3,$2B,$F9,$E3
	.dc.b	$6B,$C3,$E4,$C8,$D3,$E6,$F9,$9F,$2F,$B5,$27,$93,$B7,$FD,$AD,$6F
	.dc.b	$33,$D6,$48,$FE,$61,$87,$BB,$CC,$91,$E3,$74,$FF,$B7,$FE,$B3,$FA
	.dc.b	$16,$BF,$E0,$26,$C7,$FA,$41,$AF,$F9,$5B,$4F,$F3,$74,$3C,$7A,$49
	.dc.b	$FE,$D7,$1C,$3E,$BF,$FC,$07,$1F,$E9,$A6,$A6,$9F,$A1,$7F,$B9,$BF
	.dc.b	$EA,$17,$47,$FA,$8D,$C1,$1F,$F4,$07,$D7,$FF,$09,$3F,$85,$47,$6F
	.dc.b	$8D,$75,$7F,$A0,$09,$5F,$E0,$08,$FF,$20,$B6,$3F,$D6,$11,$C2,$3F
	.dc.b	$D6,$8D,$7F,$D4,$CD,$7F,$D5,$8E,$78,$FF,$CF,$8F,$F0,$E8,$ED,$D5
	.dc.b	$E0,$53,$A8,$EE,$48,$EE,$F8,$BD,$BF,$F4,$B9,$23,$D1,$F2,$75,$F1
	.dc.b	$C6,$DF,$3B,$5E,$FB,$7F,$8F,$B0,$FD,$69,$27,$ED,$FC,$4F,$FB,$3F
	.dc.b	$53,$49,$7F,$ED,$58,$FF,$15,$BF,$89,$F3,$FF,$06,$72,$7F,$02,$F3
	.dc.b	$23,$C5,$AF,$FB,$3E,$39,$BF,$ED,$56,$BF,$93,$8F,$27,$A1,$23,$C4
	.dc.b	$D7,$FD,$B1,$BF,$F9,$C5,$FF,$6F,$35,$FF,$2E,$B5,$FF,$75,$B5,$FF
	.dc.b	$58,$F1,$7F,$BB,$FA,$CD,$24,$7F,$BD,$63,$F6,$0D,$E5,$E0,$47,$5E
	.dc.b	$9F,$32,$38,$3F,$DF,$D7,$FC,$C1,$F4,$F8,$1F,$EF,$F8,$C4,$7A,$A3
	.dc.b	$EA,$7C,$2C,$8F,$A4,$D7,$E5,$7C,$4F,$D9,$23,$F5,$DF,$B9,$E9,$FF
	.dc.b	$9F,$B8,$BE,$73,$67,$F8,$03,$5F,$E0,$0D,$5B,$BC,$FC,$36,$07,$FB
	.dc.b	$6C,$DC,$68,$DF,$C9,$BE,$6B,$F1,$4D,$9F,$C1,$7B,$7F,$C0,$5F,$C1
	.dc.b	$46,$5F,$C2,$7A,$3F,$EF,$6C,$A9,$E0,$2C,$78,$71,$4E,$E3,$FD,$3B
	.dc.b	$F0,$5F,$E9,$CA,$FF,$3F,$FD,$6B,$EB,$BF,$E0,$0F,$6F,$FB,$CE,$B5
	.dc.b	$7F,$A8,$3D,$B5,$BB,$FF,$43,$CB,$E5,$F8,$C7,$6F,$8A,$B7,$F1,$2B
	.dc.b	$53,$C2,$73,$29,$77,$63,$EB,$BF,$AB,$5F,$68,$C7,$97,$C2,$9D,$F6
	.dc.b	$70,$DA,$F2,$CF,$3F,$2F,$83,$63,$D3,$E7,$61,$A7,$03,$FD,$47,$F5
	.dc.b	$BF,$E7,$67,$FD,$4B,$97,$F4,$F7,$3D,$5E,$7E,$DF,$5D,$7A,$35,$F9
	.dc.b	$01,$E4,$E1,$D7,$0A,$FF,$DA,$31,$A8,$C4,$FE,$32,$1F,$0D,$F2,$3E
	.dc.b	$BB,$FD,$C5,$7E,$68,$C7,$6F,$8B,$E2,$3C,$71,$67,$47,$B3,$D0,$E5
	.dc.b	$DF,$E8,$6C,$7C,$0F,$3F,$0D,$38,$1F,$EC,$6F,$8B,$FC,$D3,$FE,$C8
	.dc.b	$F6,$7F,$C9,$6E,$7C,$0F,$BB,$18,$DB,$E1,$FD,$73,$E8,$3F,$03,$87
	.dc.b	$29,$CC,$62,$FF,$0B,$5F,$CC,$B0,$FF,$49,$78,$D8,$F2,$FF,$13,$F0
	.dc.b	$E8,$DF,$99,$9A,$FF,$6D,$DA,$31,$97,$8B,$7F,$C8,$F3,$BE,$0B,$BC
	.dc.b	$61,$F0,$F9,$38,$C7,$9B,$6F,$D7,$F0,$7A,$9E,$ED,$8F,$7F,$C5,$76
	.dc.b	$F2,$FD,$6D,$AF,$2E,$18,$69,$E8,$8F,$83,$C5,$FA,$6F,$A9,$7C,$CE
	.dc.b	$2B,$FD,$EF,$18,$7D,$FE,$56,$72,$F9,$ED,$EE,$FC,$5C,$3E,$B5,$F5
	.dc.b	$7F,$B2,$FA,$EF,$E3,$98,$7F,$1A,$BD,$D6,$A3,$07,$F1,$EB,$57,$07
	.dc.b	$F3,$F3,$FE,$7D,$9D,$C7,$F2,$48,$E4,$EB,$E1,$FF,$80,$C6,$2F,$BF
	.dc.b	$5E,$27,$2D,$BE,$77,$E9,$FA,$2F,$E1,$3F,$5D,$FA,$99,$CE,$29,$CC
	.dc.b	$76,$F8,$3E,$F1,$63,$6F,$8B,$E6,$BE,$77,$D7,$59,$F9,$AF,$9B,$DD
	.dc.b	$C6,$5E,$E6,$EB,$F1,$B8,$BE,$B7,$07,$B1,$B8,$79,$3C,$FC,$34,$F3
	.dc.b	$C7,$C2,$F3,$FE,$1C,$D9,$D6,$3F,$33,$E2,$BF,$05,$C5,$F2,$1B,$3F
	.dc.b	$02,$70,$D3,$E1,$C7,$F2,$EF,$AE,$FE,$19,$1F,$67,$9B,$FD,$54,$FC
	.dc.b	$C8,$CD,$7F,$5A,$9C,$BF,$C6,$A3,$F8,$BC,$FE,$05,$DC,$1C,$BF,$D8
	.dc.b	$9C,$8F,$83,$FC,$D7,$C5,$78,$DB,$5E,$1F,$84,$C3,$4F,$24,$7A,$7E
	.dc.b	$A3,$D5,$36,$75,$87,$F2,$EF,$AE,$FE,$BF,$1F,$86,$5D,$9F,$D3,$9C
	.dc.b	$CE,$E3,$FA,$8E,$9F,$D5,$23,$F9,$BB,$FD,$83,$8B,$F4,$EE,$5B,$EC
	.dc.b	$7F,$58,$FA,$E7,$C5,$72,$BC,$37,$2F,$CB,$CF,$C6,$1F,$2F,$C7,$C7
	.dc.b	$E4,$F3,$F3,$FA,$7A,$16,$F8,$BF,$D9,$BE,$9D,$F0,$D7,$CC,$C3,$6B
	.dc.b	$F4,$B3,$FB,$3D,$BF,$7E,$FF,$1D,$8C,$F9,$3C,$FC,$9F,$EE,$3F,$5D
	.dc.b	$FA,$79,$FD,$6A,$FA,$9C,$F8,$9A,$77,$EE,$F0,$F0,$2D,$F0,$79,$F2
	.dc.b	$74,$E6,$B3,$D1,$FE,$FE,$BD,$7B,$EB,$E1,$B1,$FB,$EF,$AE,$7D,$2F
	.dc.b	$F8,$19,$7D,$2D,$8C,$3F,$E0,$AF,$AE,$F4,$57,$F0,$65,$E1,$67,$6D
	.dc.b	$3B,$D7,$87,$8D,$6F,$83,$FE,$14,$FA,$EF,$CC,$47,$F1,$05,$F2,$79
	.dc.b	$F1,$74,$EF,$7E,$1E,$2D,$BE,$0F,$F8,$7B,$DC,$F4,$C7,$F1,$A7,$FB
	.dc.b	$13,$76,$D7,$BB,$BF,$5B,$87,$D7,$5A,$E0,$FF,$8A,$3E,$BB,$DB,$1F
	.dc.b	$C9,$1F,$DB,$F7,$7F,$3E,$73,$D1,$FE,$12,$73,$3F,$CE,$5F,$C0,$46
	.dc.b	$78,$FA,$75,$75,$7B,$7A,$78,$BE,$CD,$AE,$9F,$11,$E3,$F2,$7C,$8D
	.dc.b	$AF,$27,$A9,$86,$9E,$78,$F5,$7C,$37,$F9,$0B,$97,$E2,$E1,$E7,$8C
	.dc.b	$70,$FB,$FB,$B7,$87,$BD,$B5,$FE,$2E,$DD,$E3,$73,$F1,$BF,$C9,$B1
	.dc.b	$8B,$FE,$93,$97,$F8,$0D,$FE,$97,$0F,$1F,$77,$C1,$B3,$FD,$43,$0F
	.dc.b	$F2,$AB,$CB,$E3,$3F,$CB,$31,$83,$FC,$7F,$82,$33,$DB,$AE,$DB,$7C
	.dc.b	$DE,$3E,$1E,$30,$FF,$23,$9F,$E6,$39,$F0,$57,$6D,$FE,$37,$C6,$D9
	.dc.b	$F5,$BE,$2D,$F1,$7C,$9B,$B7,$FA,$4E,$1E,$47,$F7,$96,$18,$29,$FE
	.dc.b	$9A,$6B,$4D,$FD,$67,$EE,$7E,$BB,$6D,$8C,$F8,$E9,$CB,$63,$8F,$D1
	.dc.b	$FF,$C0,$06,$E1,$CA,$31,$1E,$5B,$46,$23,$FC,$21,$0F,$A3,$FF,$09
	.dc.b	$57,$F8,$72,$31,$3F,$E0,$66,$BF,$E7,$27,$9C,$7F,$AD,$BE,$BB,$FC
	.dc.b	$57,$3F,$EB,$89,$FC,$4B,$C7,$97,$AE,$DE,$06,$C6,$F0,$AA,$BF,$14
	.dc.b	$F8,$79,$73,$18,$8F,$F4,$1F,$CB,$7D,$1C,$BD,$1F,$1B,$A2,$4F,$13
	.dc.b	$D5,$F2,$DC,$88,$FB,$63,$C3,$EE,$7F,$AA,$63,$FC,$4A,$D7,$FD,$22
	.dc.b	$DA,$FF,$20,$FC,$B7,$F5,$89,$DF,$2F,$22,$3C,$B9,$FD,$BF,$5D,$F2
	.dc.b	$AF,$FC,$92,$DE,$0F,$F2,$13,$6B,$FD,$C3,$83,$EB,$38,$F0,$C9,$FF
	.dc.b	$1C,$D7,$CE,$7F,$D4,$AF,$15,$F5,$DF,$CF,$9F,$F1,$FB,$5F,$F5,$5B
	.dc.b	$71,$FF,$2D,$BF,$E8,$B8,$DF,$A3,$F5,$23,$FA,$4C,$60,$E7,$E0,$67
	.dc.b	$C0,$91,$D9,$E1,$F1,$FC,$7D,$79,$3E,$27,$6F,$8B,$E3,$3B,$11,$F3
	.dc.b	$FC,$87,$87,$82,$47,$A9,$F2,$7C,$5F,$E2,$3E,$B9,$E4,$52,$57,$F9
	.dc.b	$7A,$3F,$DD,$8F,$F9,$83,$DB,$F9,$8D,$E3,$FC,$8C,$62,$3F,$D9,$EF
	.dc.b	$DD,$6F,$CB,$68,$8F,$3F,$7B,$8F,$54,$8E,$CF,$4E,$EF,$CE,$FD,$73
	.dc.b	$6A,$4A,$FF,$39,$C7,$F8,$51,$FF,$3A,$FD,$6F,$A1,$FF,$70,$B7,$9F
	.dc.b	$FD,$22,$DE,$0F,$F4,$52,$F3,$BF,$57,$F5,$DE,$9A,$FF,$4B,$B6,$BF
	.dc.b	$D2,$0D,$B0,$BF,$D6,$4E,$BC,$5E,$64,$9F,$F4,$4D,$7F,$84,$5B,$9F
	.dc.b	$80,$8E,$BB,$3F,$7B,$1E,$D8,$FF,$7D,$B7,$1F,$F5,$5B,$70,$FF,$52
	.dc.b	$3F,$EC,$F9,$FF,$4B,$BF,$E5,$26,$FC,$27,$43,$AF,$97,$FC,$2B,$EB
	.dc.b	$BF,$D3,$13,$FE,$9E,$6A,$C3,$FD,$73,$18,$6F,$A7,$C7,$C2,$95,$C5
	.dc.b	$E9,$E7,$7F,$16,$FA,$E7,$C4,$52,$57,$FA,$B2,$3F,$CE,$CD,$F3,$78
	.dc.b	$11,$D7,$BB,$FC,$83,$EB,$BF,$5B,$1E,$5E,$0F,$F6,$24,$63,$4F,$F6
	.dc.b	$7B,$63,$3E,$27,$1F,$85,$97,$FE,$BC,$AF,$F7,$93,$7D,$CF,$E5,$BC
	.dc.b	$3F,$E6,$88,$FF,$60,$B5,$F0,$1F,$82,$71,$8F,$F1,$EF,$86,$E3,$DB
	.dc.b	$2B,$8B,$D2,$3A,$F7,$49,$EA,$ED,$EA,$2F,$E2,$B8,$6D,$FF,$BB,$1B
	.dc.b	$A7,$FB,$97,$E4,$9F,$4D,$EA,$E2,$98,$7A,$9F,$6F,$B5,$EE,$EC,$93
	.dc.b	$EE,$7F,$A6,$FC,$CB,$E7,$7E,$1F,$AD,$2B,$E1,$BC,$1F,$27,$89,$F2
	.dc.b	$08,$FF,$BA,$2B,$F9,$86,$DE,$7F,$53,$E6,$92,$7E,$6D,$7D,$0F,$9A
	.dc.b	$91,$F4,$1D,$3F,$13,$EB,$7F,$98,$37,$E1,$3F,$B1,$C6,$30,$70,$FE
	.dc.b	$DA,$9F,$5C,$FA,$ED,$F5,$C9,$FF,$7E,$E1,$FC,$F6,$F8,$3C,$DB,$E9
	.dc.b	$F6,$24,$FF,$02,$C7,$BB,$FC,$13,$EB,$BE,$8B,$D3,$D3,$F2,$91,$EF
	.dc.b	$5E,$8F,$9D,$91,$E2,$7F,$BE,$FD,$47,$F0,$F8,$64,$F9,$EF,$79,$23
	.dc.b	$81,$EA,$D7,$E1,$91,$F6,$C7,$C6,$F0,$64,$7C,$07,$FC,$17,$9B,$E9
	.dc.b	$FE,$5F,$A5,$2B,$E2,$BC,$1F,$A5,$EB,$7D,$02,$3F,$0F,$83,$E8,$7D
	.dc.b	$77,$FA,$8A,$3E,$6C,$7D,$62,$3F,$42,$7F,$C3,$7F,$5A,$F9,$BA,$49
	.dc.b	$E8,$7F,$30,$FA,$E7,$C4,$70,$F0,$E4,$E8,$5A,$39,$47,$EB,$23,$C8
	.dc.b	$DF,$94,$F0,$9C,$9D,$7C,$6F,$90,$8F,$5F,$AE,$79,$5E,$8F,$1E,$4F
	.dc.b	$87,$0F,$C0,$BC,$8C,$7C,$FF,$B8,$8F,$9F,$8D,$F0,$DF,$0B,$92,$4F
	.dc.b	$37,$6F,$F0,$4F,$AE,$FF,$16,$71,$FE,$75,$B1,$FD,$51,$AB,$E1,$FC
	.dc.b	$EF,$C5,$8E,$ED,$BF,$F2,$5C,$62,$3B,$7A,$BF,$E4,$EF,$AE,$FC,$91
	.dc.b	$3E,$77,$FC,$A3,$83,$D0,$D7,$C1,$4C,$3E,$95,$FF,$81,$7C,$4F,$D2
	.dc.b	$C2,$39,$F9,$FD,$49,$1F,$67,$0F,$E2,$0F,$2F,$43,$C4,$47,$AC,$F9
	.dc.b	$7D,$D9,$1E,$07,$CB,$F3,$1E,$9F,$7A,$4F,$91,$DB,$FC,$BB,$17,$9D
	.dc.b	$EF,$24,$78,$F0,$F8,$5E,$7C,$8F,$6B,$A7,$F9,$BF,$D6,$7F,$4F,$7F
	.dc.b	$C0,$AD,$4D,$D3,$FB,$1C,$7E,$5E,$1E,$4F,$F9,$F1,$B1,$98,$C7,$C5
	.dc.b	$7F,$A0,$3E,$BB,$F4,$33,$FE,$A8,$6B,$FE,$AC,$6B,$FE,$AF,$6B,$FC
	.dc.b	$8D,$FF,$0B,$BF,$E1,$A6,$A6,$9D,$38,$3E,$CB,$88,$F1,$1E,$3F,$0B
	.dc.b	$FD,$27,$F1,$DF,$E2,$D8,$FF,$12,$BF,$E4,$E7,$F6,$8D,$7F,$D2,$7F
	.dc.b	$2B,$C3,$E2,$BF,$D7,$F3,$99,$FF,$63,$35,$FE,$74,$D7,$AF,$77,$80
	.dc.b	$8F,$69,$AF,$FB,$3A,$53,$D8,$91,$FE,$D0,$8F,$F5,$9B,$FE,$A6,$F9
	.dc.b	$4F,$59,$EC,$F9,$0B,$1B,$8F,$39,$1E,$5F,$07,$F0,$88,$FC,$08,$FD
	.dc.b	$3F,$CD,$7F,$B7,$38,$4D,$24,$FF,$B8,$A3,$2F,$E1,$7E,$3F,$F0,$87
	.dc.b	$27,$07,$FB,$A3,$89,$1F,$EE,$86,$C6,$63,$04,$70,$43,$FD,$E7,$1F
	.dc.b	$E4,$97,$F3,$7E,$AF,$E4,$CF,$0B,$F9,$77,$B1,$1F,$13,$ED,$F7,$24
	.dc.b	$7A,$DA,$FF,$C0,$0B,$FF,$6A,$3F,$DD,$4F,$E2,$DD,$3F,$14,$8F,$C3
	.dc.b	$E6,$F9,$04,$70,$71,$3F,$CE,$7F,$01,$FE,$44,$7A,$35,$FF,$01,$35
	.dc.b	$DF,$37,$17,$8F,$B1,$54,$FF,$9C,$5E,$71,$98,$C4,$7F,$83,$7E,$BB
	.dc.b	$F9,$6B,$E3,$6B,$ED,$6B,$BF,$2F,$EB,$7C,$74,$8F,$B4,$D7,$D9,$5F
	.dc.b	$DF,$DF,$F4,$7F,$91,$EC,$F5,$FF,$C3,$8D,$8C,$9E,$E8,$CB,$FE,$21
	.dc.b	$E0,$FF,$2E,$BF,$CC,$8F,$E2,$7E,$BE,$9F,$E2,$ED,$72,$74,$3F,$8D
	.dc.b	$FC,$BB,$E6,$FF,$20,$D8,$FC,$4F,$C2,$7F,$0E,$7F,$8F,$C6,$5F,$BD
	.dc.b	$E5,$FF,$AB,$F2,$FE,$ED,$F5,$AF,$AE,$FF,$14,$F8,$7F,$D1,$AB,$57
	.dc.b	$FE,$62,$AD,$5F,$F7,$D9,$78,$3C,$23,$FC,$7D,$1C,$2F,$94,$E6,$FF
	.dc.b	$D6,$F3,$9A,$DF,$66,$65,$F5,$DF,$D6,$E7,$B4,$E6,$31,$3F,$E7,$4A
	.dc.b	$77,$71,$55,$63,$0E,$63,$97,$C3,$5C,$3F,$C2,$D8,$7F,$9B,$5F,$67
	.dc.b	$A3,$9C,$7B,$37,$3F,$ED,$3F,$5D,$FE,$10,$AE,$3E,$57,$A6,$31,$7F
	.dc.b	$E8,$8C,$3F,$AD,$D7,$F8,$8E,$7D,$37,$CB,$A3,$87,$A1,$C5,$E5,$F1
	.dc.b	$7F,$0F,$EB,$BF,$DE,$18,$71,$8C,$75,$F7,$F8,$DD,$CF,$56,$E9,$FE
	.dc.b	$98,$CD,$E8,$F3,$F9,$7F,$33,$6F,$C9,$E1,$E1,$A7,$9E,$3E,$A7,$C9
	.dc.b	$7E,$3E,$3F,$C7,$9E,$FF,$F9,$55,$FF,$1F,$46,$71,$FF,$53,$CE,$5F
	.dc.b	$F5,$4F,$D7,$3E,$2F,$9F,$7F,$DF,$8C,$61,$CA,$31,$E5,$FE,$0F,$E3
	.dc.b	$5F,$1B,$8A,$FF,$4D,$3C,$7C,$DC,$65,$D9,$BC,$FF,$E1,$EC,$9C,$1F
	.dc.b	$8D,$E6,$78,$9E,$18,$C3,$E3,$F8,$BC,$67,$FE,$59,$B4,$63,$2F,$5F
	.dc.b	$8F,$BA,$D3,$63,$DF,$F1,$FD,$FE,$2B,$DC,$DA,$E9,$8E,$1A,$77,$8F
	.dc.b	$89,$C9,$FA,$3F,$A9,$7A,$78,$AF,$F3,$64,$61,$F3,$7C,$56,$72,$F9
	.dc.b	$ED,$E0,$FF,$1F,$FD,$63,$EA,$FF,$5D,$F5,$CD,$BF,$61,$39,$C7,$FC
	.dc.b	$F1,$39,$7F,$CF,$51,$FE,$83,$6D,$70,$F8,$DE,$2F,$2E,$BC,$4F,$06
	.dc.b	$E7,$FE,$EA,$B7,$D1,$FE,$FF,$EB,$BF,$D8,$B7,$DA,$30,$F9,$3E,$37
	.dc.b	$CB,$16,$34,$79,$3D,$EF,$8D,$F5,$DE,$A6,$C7,$A3,$DF,$C3,$4E,$07
	.dc.b	$B3,$8F,$E7,$6B,$F0,$9B,$1F,$03,$CF,$C3,$4E,$07,$C4,$F4,$3E,$4C
	.dc.b	$7C,$A6,$C7,$C8,$F8,$0F,$B5,$C3,$F3,$9B,$1F,$33,$83,$0D,$38,$1F
	.dc.b	$EA,$9F,$AE,$FE,$17,$CB,$FE,$78,$E7,$FF,$55,$4E,$7B,$FF,$0D,$CB
	.dc.b	$FC,$A6,$FF,$86,$CF,$FB,$BD,$E2,$DF,$37,$55,$BF,$84,$FC,$77,$C8
	.dc.b	$5A,$FC,$6B,$CF,$E2,$31,$A3,$4F,$F4,$4E,$49,$CC,$63,$E0,$F8,$61
	.dc.b	$E9,$6D,$F1,$83,$3E,$B9,$DD,$FD,$05,$77,$D8,$C3,$FD,$17,$87,$FA
	.dc.b	$7D,$FF,$B3,$5F,$E9,$57,$D1,$CB,$DC,$C7,$CF,$FF,$4D,$FA,$E7,$C6
	.dc.b	$F3,$9C,$36,$FF,$DB,$13,$F2,$27,$CE,$FF,$84,$A3,$3C,$3C,$FE,$07
	.dc.b	$A7,$A3,$D7,$D1,$B5,$C9,$FE,$B9,$F5,$0F,$84,$BE,$76,$1B,$5F,$EE
	.dc.b	$79,$FD,$86,$7E,$90,$FF,$18,$8D,$79,$1D,$3C,$AF,$ED,$1F,$5D,$FA
	.dc.b	$C9,$FF,$78,$2F,$9F,$CF,$8F,$A7,$7E,$FF,$0F,$0A,$DF,$07,$9E,$7D
	.dc.b	$39,$6C,$F4,$7F,$BB,$2F,$62,$FB,$58,$6C,$7F,$C0,$1F,$AE,$7C,$47
	.dc.b	$F7,$D5,$F3,$96,$30,$FF,$80,$7E,$BB,$D1,$5F,$86,$F0,$3F,$DA,$F0
	.dc.b	$F0,$F5,$BE,$DE,$1B,$71,$F2,$6E,$1C,$4F,$C7,$EE,$C3,$FC,$0B,$19
	.dc.b	$8F,$C9,$FD,$73,$C6,$FF,$09,$2F,$93,$B1,$87,$3C,$3A,$F1,$6E,$1C
	.dc.b	$F0,$61,$FE,$8C,$8C,$7B,$BF,$01,$BE,$F7,$F8,$7E,$31,$5F,$7D,$FE
	.dc.b	$C6,$DF,$5F,$5F,$5D,$B6,$DF,$7F,$AE,$F8,$8D,$F3,$BF,$C5,$51,$8B
	.dc.b	$FC,$53,$FC,$21,$BD,$7A,$E4,$DB,$6F,$A7,$AD,$86,$9D,$E3,$AF,$C4
	.dc.b	$76,$F2,$AC,$F8,$DF,$0F,$AD,$8C,$7C,$2D,$E9,$57,$D4,$CD,$F0,$FF
	.dc.b	$E4,$1E,$6F,$87,$5A,$C3,$F8,$3D,$6A,$FE,$75,$69,$E5,$DF,$F0,$30
	.dc.b	$F8,$2D,$8F,$A9,$87,$F8,$89,$E3,$E3,$7F,$93,$E3,$18,$FE,$BF,$04
	.dc.b	$E7,$C8,$D7,$91,$70,$E6,$FF,$95,$EF,$F1,$67,$1F,$8B,$C7,$6F,$AD
	.dc.b	$D7,$76,$2F,$F0,$D5,$FE,$58,$6D,$73,$F1,$37,$1A,$6C,$FF,$03,$3E
	.dc.b	$11,$E7,$F8,$CF,$16,$FF,$25,$F2,$6F,$8B,$EE,$DA,$E4,$7F,$0C,$3F
	.dc.b	$D5,$DE,$4F,$FA,$BD,$C1,$FE,$B3,$7F,$D6,$9F,$59,$FB,$FF,$AE,$BE
	.dc.b	$41,$A4,$AF,$F6,$14,$BE,$93,$FA,$44,$61,$AF,$FB,$3E,$35,$C9,$DA
	.dc.b	$73,$3F,$B8,$97,$CA,$7E,$EA,$BF,$09,$29,$CB,$FB,$16,$BF,$CA,$5E
	.dc.b	$AB,$EB,$BF,$C5,$6F,$F5,$28,$FF,$32,$BC,$78,$FB,$6B,$E1,$6C,$6F
	.dc.b	$0A,$F8,$79,$2D,$9A,$6B,$E5,$BF,$DD,$99,$7A,$25,$F1,$FC,$09,$7E
	.dc.b	$9B,$E7,$F2,$B2,$3D,$8E,$9F,$E8,$9E,$FB,$97,$C6,$D8,$FE,$24,$D7
	.dc.b	$FA,$57,$CB,$7F,$60,$73,$CD,$C0,$8F,$0A,$F6,$FD,$77,$CA,$D3,$E7
	.dc.b	$B7,$83,$F9,$5B,$6B,$FB,$DE,$0F,$9B,$E3,$C3,$27,$F9,$25,$7C,$E7
	.dc.b	$FA,$E3,$C5,$7D,$77,$F4,$47,$F9,$53,$5F,$EC,$4D,$C7,$F0,$2F,$FA
	.dc.b	$2E,$37,$E9,$7D,$48,$FE,$9F,$18,$39,$FA,$59,$EF,$91,$D9,$E1,$F2
	.dc.b	$1C,$7D,$89,$3E,$27,$6F,$8B,$E3,$BB,$11,$F3,$FC,$97,$87,$82,$47
	.dc.b	$A9,$F2,$7C,$07,$E2,$3E,$B9,$E4,$52,$57,$F4,$98,$FF,$75,$BF,$D2
	.dc.b	$FD,$CF,$98,$DE,$3F,$C8,$C6,$23,$FC,$0C,$FD,$D6,$FC,$2E,$88,$F3
	.dc.b	$F7,$F8,$F6,$48,$EC,$F4,$EC,$FC,$EF,$D7,$36,$A4,$AF,$EB,$D1,$FE
	.dc.b	$1B,$7F,$B0,$7D,$77,$A1,$FF,$0C,$37,$9F,$F6,$ED,$E0,$FE,$E6,$BC
	.dc.b	$EF,$D5,$FD,$77,$A6,$BF,$7A,$DA,$FE,$F2,$DB,$FF,$24,$DF,$EB,$27
	.dc.b	$3E,$2F,$22,$4F,$F7,$3A,$FF,$0B,$37,$3E,$E4,$75,$D9,$FB,$C8,$F6
	.dc.b	$47,$F8,$C5,$B8,$FF,$10,$6E,$1F,$E0,$C7,$FA,$3C,$FF,$80,$5F,$F2
	.dc.b	$CB,$7E,$17,$A1,$D7,$CB,$FE,$13,$F5,$DF,$EA,$29,$FF,$04,$35,$FF
	.dc.b	$20,$37,$0F,$E3,$EF,$F8,$EE,$37,$E0,$F2,$23,$CB,$CE,$FE,$2B,$F5
	.dc.b	$DF,$E7,$58,$FF,$0A,$35,$FF,$26,$36,$FF,$95,$BF,$E2,$18,$DF,$99
	.dc.b	$C8,$8F,$2F,$7B,$F9,$0F,$D7,$7E,$BA,$3F,$CA,$CD,$BF,$E6,$AD,$DB
	.dc.b	$FC,$62,$FE,$E5,$BF,$0B,$C8,$95,$FE,$29,$97,$B7,$FC,$BB,$67,$C3
	.dc.b	$52,$57,$F8,$BF,$C9,$CB,$FE,$31,$97,$5F,$12,$46,$6B,$A7,$64,$8E
	.dc.b	$0E,$9F,$F3,$85,$7C,$36,$FA,$DE,$33,$FA,$0F,$B8,$FA,$8F,$37,$C5
	.dc.b	$CA,$F1,$3B,$7C,$9D,$7E,$B2,$3F,$E4,$7F,$99,$7C,$EF,$BF,$D5,$93
	.dc.b	$EB,$76,$F9,$3C,$3F,$88,$47,$97,$77,$C7,$F4,$E4,$7C,$37,$4F,$B3
	.dc.b	$E6,$BE,$61,$1D,$1F,$E5,$9F,$AD,$FE,$58,$FF,$A4,$5B,$D3,$FD,$C1
	.dc.b	$BD,$3F,$E6,$CF,$AE,$7D,$77,$F6,$49,$FB,$D7,$FE,$BE,$6A,$69,$C9
	.dc.b	$D7,$A3,$C4,$8F,$F9,$AF,$EA,$BE,$73,$E5,$F4,$7B,$11,$F4,$4F,$87
	.dc.b	$E7,$64,$7A,$DF,$EE,$FF,$4F,$FC,$26,$3C,$5F,$3D,$E7,$23,$B3,$E1
	.dc.b	$65,$EB,$23,$F1,$63,$DD,$E0,$C8,$F9,$0F,$F8,$13,$27,$D5,$7C,$3F
	.dc.b	$42,$57,$CD,$78,$3F,$37,$D4,$F9,$04,$7F,$D1,$5F,$5D,$FE,$9F,$7E
	.dc.b	$A7,$0F,$CD,$23,$F5,$E7,$FC,$27,$F5,$8F,$9B,$A4,$9F,$F4,$A7,$D7
	.dc.b	$3E,$1B,$7E,$0C,$9F,$F4,$C5,$A3,$84,$7F,$B4,$63,$C4,$DF,$95,$EE
	.dc.b	$71,$75,$F1,$BE,$22,$3F,$EA,$3F,$AE,$FD,$9B,$DB,$C4,$F3,$91,$F0
	.dc.b	$CF,$AB,$E7,$E4,$79,$1F,$99,$F0,$9E,$EE,$29,$3E,$87,$6F,$FB,$9F
	.dc.b	$EB,$BF,$C4,$1D,$DF,$E4,$66,$F1,$7F,$B0,$9B,$3F,$E1,$8F,$33,$FC
	.dc.b	$77,$1F,$E9,$28,$FF,$5D,$35,$E9,$86,$BD,$4F,$F2,$0F,$D7,$7E,$2C
	.dc.b	$9F,$3B,$FE,$44,$F4,$1E,$7E,$BD,$F9,$8F,$FB,$16,$FF,$BC,$6B,$FE
	.dc.b	$5A,$79,$79,$FD,$09,$3F,$C7,$7C,$CE,$FE,$7C,$8E,$4E,$95,$F3,$F3
	.dc.b	$7F,$CA,$1E,$EC,$9F,$04,$57,$7C,$5F,$EE,$4F,$8A,$FE,$E5,$1F,$ED
	.dc.b	$96,$BA,$03,$3F,$F1,$C1,$FE,$3B,$C3,$D5,$E7,$7C,$43,$E1,$7C,$CD
	.dc.b	$AF,$81,$EE,$FC,$83,$0F,$51,$F5,$9F,$C1,$BE,$BB,$E8,$61,$FE,$7E
	.dc.b	$6E,$7F,$EF,$A9,$D7,$37,$FB,$F2,$E5,$7D,$76,$C6,$5F,$F8,$03,$19
	.dc.b	$79,$79,$7F,$E7,$AA,$34,$C7,$C5,$59,$FF,$7E,$4E,$9B,$3A,$69,$0F
	.dc.b	$64,$6D,$B3,$CD,$CC,$DC,$7F,$D3,$75,$FD,$9E,$7E,$7B,$C6,$33,$3B
	.dc.b	$C7,$E1,$5D,$BD,$CD,$7E,$5C,$5B,$E5,$5F,$02,$7D,$73,$9A,$D3,$63
	.dc.b	$D5,$7E,$48,$C4,$7A,$0E,$26,$AB,$A3,$93,$DE,$64,$F2,$94,$F2,$9F
	.dc.b	$1C,$FB,$57,$F1,$DB,$7F,$65,$E2,$7D,$B8,$D3,$E0,$36,$6D,$CB,$C2
	.dc.b	$FF,$D8,$FE,$C3,$C7,$FF,$5C,$B6,$7C,$95,$F1,$9F,$CF,$C6,$5F,$CB
	.dc.b	$36,$BE,$9C,$7A,$E3,$18,$7E,$9A,$3F,$C2,$F7,$F2,$2B,$C2,$F8,$0C
	.dc.b	$4E,$EB,$8D,$78,$27,$2D,$3F,$63,$5C,$9F,$3E,$1C,$DC,$DF,$3A,$E3
	.dc.b	$1E,$23,$27,$33,$EB,$57,$A9,$F6,$C7,$37,$33,$F7,$27,$C1,$1A,$3C
	.dc.b	$27,$43,$B1,$93,$89,$F4,$63,$C8,$EE,$FD,$EF,$F0,$0B,$FD,$97,$C9
	.dc.b	$3E,$63,$ED,$36,$7F,$01,$ED,$7F,$59,$7F,$05,$19,$7F,$81,$7D,$63
	.dc.b	$E0,$59,$7E,$26,$30,$D8,$CC,$FF,$05,$9C,$BF,$C1,$FB,$DF,$1F,$AD
	.dc.b	$77,$3F,$EE,$F8,$C6,$1F,$7F,$EB,$1E,$0E,$2B,$F9,$3E,$1C,$AB,$E3
	.dc.b	$56,$9B,$5B,$A6,$3F,$CB,$A7,$27,$0F,$03,$57,$FD,$32,$3E,$86,$0F
	.dc.b	$95,$CD,$73,$8C,$4E,$A7,$1A,$65,$F0,$7D,$5F,$F9,$9A,$7C,$BE,$F6
	.dc.b	$3C,$6E,$EE,$B1,$C4,$FD,$FA,$F1,$69,$FC,$E5,$E4,$DF,$83,$F6,$7D
	.dc.b	$67,$FB,$26,$B8,$46,$3E,$33,$C3,$F0,$DE,$28,$B3,$A3,$B7,$81,$CB
	.dc.b	$E3,$FC,$ED,$8F,$17,$81,$86,$9C,$0F,$CA,$7C,$57,$E2,$DF,$CB,$7A
	.dc.b	$9F,$E3,$47,$3D,$7F,$46,$31,$AF,$F5,$3F,$AC,$7C,$F7,$D5,$C3,$94
	.dc.b	$E6,$31,$7F,$66,$9F,$25,$F6,$FE,$21,$F0,$F9,$8C,$3B,$F8,$4C,$69
	.dc.b	$FA,$9B,$46,$32,$F1,$6F,$F8,$5E,$77,$C0,$F7,$8C,$3E,$1F,$84,$C6
	.dc.b	$3C,$DB,$7E,$BF,$83,$D4,$F7,$6C,$7B,$FE,$2B,$B7,$C1,$FA,$DB,$5E
	.dc.b	$5C,$30,$D3,$D1,$1F,$07,$8B,$F3,$5F,$4F,$F9,$FC,$3E,$54,$63,$C4
	.dc.b	$F8,$DE,$56,$F0,$F8,$0D,$AF,$F0,$17,$D5,$BE,$A7,$F4,$9F,$58,$D7
	.dc.b	$F4,$D3,$9C,$3F,$88,$4E,$5F,$E2,$51,$FE,$61,$6D,$70,$F8,$6E,$2F
	.dc.b	$2E,$AC,$4F,$06,$E3,$E2,$FA,$0F,$DA,$7D,$67,$F0,$AA,$EB,$18,$7C
	.dc.b	$7F,$05,$E4,$8B,$1A,$3C,$7F,$31,$F1,$3E,$B3,$D2,$D8,$F3,$FC,$26
	.dc.b	$1A,$70,$3D,$7C,$3F,$33,$4F,$82,$D8,$F7,$F9,$58,$69,$C0,$F8,$7E
	.dc.b	$5F,$C8,$8F,$92,$D8,$F8,$FF,$13,$F6,$38,$3E,$6B,$63,$E5,$EF,$86
	.dc.b	$9C,$0F,$E8,$1F,$59,$FC,$03,$D2,$FE,$4B,$D1,$FE,$89,$9C,$F7,$7E
	.dc.b	$13,$93,$F8,$F5,$FF,$08,$BF,$E5,$6F,$1A,$FE,$D0,$E5,$B5,$C3,$E0
	.dc.b	$BF,$1B,$F1,$36,$7F,$A8,$62,$8E,$BF,$05,$9C,$FF,$C9,$98,$A3,$19
	.dc.b	$78,$3E,$9F,$78,$FA,$9B,$7F,$E5,$3F,$AC,$FF,$03,$4F,$F5,$D5,$DD
	.dc.b	$FC,$DD,$CC,$EE,$3F,$9D,$6E,$FA,$5B,$F0,$B8,$6E,$1F,$D9,$FE,$B3
	.dc.b	$FA,$16,$1F,$E6,$1B,$FE,$80,$FF,$98,$E3,$39,$FF,$50,$9C,$BF,$D4
	.dc.b	$A3,$E4,$4F,$F9,$D1,$7B,$DF,$D2,$DC,$C1,$CD,$F8,$7C,$FD,$7A,$1B
	.dc.b	$6C,$7F,$7B,$FA,$57,$CA,$7F,$55,$7F,$8A,$C6,$63,$1E,$66,$7C,$6D
	.dc.b	$79,$36,$B9,$3F,$D7,$7E,$B1,$E2,$5F,$A8,$C3,$6B,$FC,$15,$2F,$73
	.dc.b	$87,$81,$C5,$B3,$AE,$0E,$9C,$7D,$5B,$1F,$E1,$31,$EB,$FF,$6D,$5E
	.dc.b	$BD,$8C,$3F,$DC,$7E,$B1,$F0,$B7,$E8,$F0,$DB,$FF,$0F,$7D,$63,$7F
	.dc.b	$EF,$0F,$E4,$63,$31,$8F,$07,$3B,$6B,$C7,$B5,$C9,$FF,$00,$7D,$63
	.dc.b	$C5,$BF,$17,$86,$D7,$F8,$B6,$7D,$2D,$F0,$F8,$EF,$C9,$FF,$05,$7B
	.dc.b	$8D,$EF,$BB,$86,$DF,$F8,$DE,$5E,$B7,$0F,$AC,$E2,$D9,$FF,$1E,$7D
	.dc.b	$67,$B7,$9B,$FD,$6C,$FE,$F5,$BC,$3F,$E1,$98,$C7,$89,$FE,$26,$7E
	.dc.b	$FD,$72,$E3,$D7,$5E,$DE,$1E,$A6,$27,$BB,$71,$EF,$E3,$75,$F1,$FB
	.dc.b	$36,$BC,$BE,$9E,$1A,$7A,$23,$C9,$F1,$BF,$E2,$DE,$6F,$89,$3E,$C8
	.dc.b	$C7,$17,$A7,$BB,$73,$F0,$1B,$97,$FB,$57,$77,$8F,$CF,$C4,$FF,$1C
	.dc.b	$46,$2F,$FC,$C5,$97,$E9,$AF,$87,$8F,$AC,$36,$DA,$FF,$34,$4F,$F8
	.dc.b	$D8,$FF,$71,$53,$E2,$7F,$C8,$91,$9C,$3F,$CA,$33,$9A,$E3,$DD,$AE
	.dc.b	$ED,$B6,$3A,$F8,$78,$69,$C0,$F0,$EF,$F1,$7E,$2B,$BB,$63,$CB,$BE
	.dc.b	$1A,$70,$3D,$39,$BC,$9F,$F5,$B2,$30,$61,$CB,$83,$FD,$F1,$F5,$5F
	.dc.b	$B6,$FA,$CF,$F5,$F3,$63,$FD,$F0,$D7,$7E,$3A,$77,$F4,$4E,$0A,$4A
	.dc.b	$E3,$19,$79,$FC,$9F,$FA,$D6,$7F,$CF,$78,$23,$19,$7F,$AE,$63,$15
	.dc.b	$FE,$9D,$57,$CF,$EB,$3F,$D0,$B3,$EF,$9F,$C2,$BB,$E5,$E6,$E5,$AA
	.dc.b	$3A,$BE,$9B,$F0,$72,$EA,$4D,$38,$FC,$A7,$FB,$36,$B8,$1E,$1F,$8D
	.dc.b	$E0,$47,$D1,$3C,$FE,$8E,$47,$AD,$D3,$C8,$FF,$4F,$F2,$79,$BD,$CF
	.dc.b	$E6,$51,$FE,$BD,$6B,$BF,$98,$FF,$4F,$B9,$E5,$E0,$47,$85,$7C,$0F
	.dc.b	$AC,$F7,$F1,$7F,$AA,$9B,$B7,$FB,$9D,$B5,$F8,$9C,$1F,$33,$C7,$82
	.dc.b	$4F,$FB,$3E,$BE,$7B,$E9,$78,$AF,$AC,$FE,$44,$FF,$B5,$5A,$FB,$DB
	.dc.b	$8F,$FA,$F1,$FF,$3F,$C6,$FD,$1F,$AF,$1F,$D1,$23,$07,$3F,$03,$3E
	.dc.b	$04,$8E,$CF,$0F,$8D,$E3,$EB,$C9,$F1,$3B,$7C,$5F,$17,$D8,$8F,$9F
	.dc.b	$E3,$FC,$3C,$12,$3D,$4F,$93,$E2,$BF,$17,$F5,$8F,$22,$92,$BF,$DE
	.dc.b	$91,$FE,$13,$7F,$DE,$DE,$DF,$CC,$6F,$1F,$E5,$23,$11,$F8,$37,$EE
	.dc.b	$B7,$E4,$F4,$47,$9F,$C2,$71,$EC,$91,$D9,$E9,$D9,$FA,$0F,$AC,$6D
	.dc.b	$49,$5F,$C1,$F9,$BF,$DD,$B3,$FE,$10,$8C,$6F,$FC,$2D,$7D,$67,$AD
	.dc.b	$FC,$C3,$6B,$FD,$EA,$D9,$E0,$BC,$EF,$D8,$FD,$67,$B2,$BF,$DF,$6D
	.dc.b	$AE,$CD,$BF,$E3,$57,$FA,$FA,$DF,$89,$C8,$93,$E6,$3F,$90,$57,$F8
	.dc.b	$41,$C0,$7F,$76,$E1,$AF,$EC,$5B,$3E,$4E,$CF,$E0,$FA,$7C,$3A,$FD
	.dc.b	$B3,$65,$8F,$C5,$8C,$37,$A3,$8F,$C1,$CB,$F1,$D7,$F9,$69,$BE,$5F
	.dc.b	$F1,$4F,$AC,$FF,$1A,$C7,$A1,$AF,$F0,$56,$E3,$E0,$8F,$E7,$D1,$FE
	.dc.b	$23,$78,$78,$3C,$C8,$F3,$F3,$7F,$90,$FD,$67,$F9,$CE,$3E,$2B,$5F
	.dc.b	$E2,$4D,$BE,$D0,$8F,$F1,$33,$C3,$CC,$E6,$47,$9F,$BB,$FC,$BB,$EB
	.dc.b	$3F,$71,$1F,$C7,$5B,$7D,$1B,$BF,$E3,$5F,$E0,$6D,$F8,$4E,$44,$AF
	.dc.b	$B7,$3F,$E2,$87,$EE,$EB,$FE,$70,$6F,$C3,$7F,$3E,$8C,$47,$F2,$78
	.dc.b	$C7,$54,$8E,$2E,$E3,$87,$5C,$9E,$AE,$DF,$E6,$D5,$F0,$9B,$EA,$77
	.dc.b	$3F,$A6,$FC,$5B,$E9,$3C,$5F,$15,$2B,$CE,$ED,$F4,$F5,$79,$C8,$FE
	.dc.b	$63,$E6,$1F,$45,$EB,$F5,$64,$FC,$07,$6F,$93,$C2,$F8,$04,$79,$76
	.dc.b	$FC,$3F,$4A,$47,$C8,$74,$FB,$3E,$67,$E4,$11,$D1,$FA,$AF,$AB,$FE
	.dc.b	$6A,$FF,$57,$6F,$3F,$E1,$9B,$CF,$FC,$1F,$EB,$1F,$59,$FD,$C2,$7E
	.dc.b	$ED,$FF,$85,$5A,$9A,$72,$75,$E1,$71,$23,$FB,$CF,$AC,$F9,$CF,$97
	.dc.b	$CF,$EC,$47,$D1,$3E,$1F,$9C,$91,$EB,$7F,$C0,$7F,$47,$FC,$3E,$3C
	.dc.b	$5F,$3B,$E7,$23,$B3,$E1,$65,$EB,$23,$F1,$63,$DD,$DF,$23,$E4,$3F
	.dc.b	$E1,$0C,$9F,$51,$F0,$FC,$E9,$5F,$35,$E0,$FC,$DF,$4B,$E4,$11,$FE
	.dc.b	$37,$F5,$9F,$EE,$F7,$EA,$70,$FC,$D2,$3F,$5E,$7F,$C3,$FF,$56,$F9
	.dc.b	$AA,$49,$FE,$4B,$F5,$8F,$85,$DF,$83,$27,$F9,$4D,$A3,$84,$7F,$85
	.dc.b	$E3,$C4,$DF,$8D,$EE,$71,$75,$F1,$5E,$22,$3F,$CD,$3E,$B1,$E4,$F9
	.dc.b	$BC,$59,$3E,$07,$6F,$7F,$9E,$F5,$91,$FC,$17,$C1,$FB,$B8,$E4,$7A
	.dc.b	$1D,$3F,$E2,$DF,$AC,$FF,$14,$F7,$FF,$1C,$6F,$1F,$F5,$D6,$CF,$F8
	.dc.b	$87,$2F,$F2,$3C,$30,$A4,$9F,$E9,$B1,$C1,$7D,$67,$E2,$C9,$F5,$5F
	.dc.b	$E5,$0F,$86,$79,$DF,$E5,$29,$FE,$FF,$9F,$F2,$E6,$E7,$FD,$BD,$B8
	.dc.b	$7F,$9E,$DE,$9E,$6F,$12,$47,$F5,$EC,$3F,$83,$BF,$D8,$27,$F9,$D5
	.dc.b	$73,$F3,$7C,$07,$09,$FF,$28,$4F,$F9,$89,$FF,$2E,$35,$F1,$79,$5F
	.dc.b	$E8,$C8,$C4,$78,$E3,$FC,$C7,$ED,$C8,$F0,$47,$B7,$CA,$91,$E3,$7E
	.dc.b	$17,$CC,$E2,$3F,$CD,$47,$C8,$8E,$C4,$72,$3E,$AF,$FC,$92,$FF,$9C
	.dc.b	$5A,$FF,$85,$5A,$FF,$A8,$5A,$F1,$F8,$5F,$F4,$0F,$D6,$7D,$B8,$FF
	.dc.b	$02,$B5,$FF,$39,$4F,$F1,$43,$FA,$AB,$CF,$E2,$3F,$D1,$31,$88,$6D
	.dc.b	$F3,$9D,$1F,$E1,$08,$D1,$B8,$D4,$7A,$E7,$99,$1F,$B5,$AD,$7C,$71
	.dc.b	$78,$3F,$B9,$D7,$2B,$EB,$39,$DF,$FC,$6D,$39,$8C,$5F,$90,$FF,$44
	.dc.b	$61,$87,$C4,$79,$E3,$F2,$D1,$9B,$EB,$3E,$43,$44,$3C,$A6,$4A,$4C
	.dc.b	$3F,$C1,$31,$F0,$67,$A5,$78,$6F,$CD,$7C,$EF,$73,$E2,$8E,$C7,$DF
	.dc.b	$7F,$DB,$B7,$FE,$0C,$8C,$1F,$86,$70,$FE,$0E,$36,$47,$4F,$DA,$F2
	.dc.b	$24,$78,$71,$E0,$FD,$98,$CB,$EB,$8C,$BF,$86,$8C,$DF,$A2,$39,$1F
	.dc.b	$91,$8C,$CF,$09,$D4,$F1,$3B,$1C,$0F,$A6,$64,$8F,$02,$47,$E7,$E7
	.dc.b	$AC,$FD,$D9,$CB,$E1,$3A,$9E,$93,$53,$C5,$CC,$70,$7C,$2F,$D8,$9E
	.dc.b	$B1,$89,$F5,$1C,$5F,$2B,$87,$A9,$E1,$32,$60,$7D,$67,$F6,$27,$F7
	.dc.b	$AD,$7F,$81,$B5,$FF,$77,$35,$FD,$F7,$C9,$3E,$67,$F0,$0D,$8F,$E1
	.dc.b	$79,$3D,$7F,$EB,$91,$89,$D3,$FE,$F9,$F7,$9F,$1B,$FE,$5E,$8F,$F4
	.dc.b	$F4,$63,$2F,$E2,$2D,$8F,$E2,$ED,$7F,$D4,$83,$EB,$1D,$5C,$E0,$65
	.dc.b	$C3,$6B,$F8,$EC,$F0,$6F,$56,$A3,$FA,$3C,$60,$93,$E2,$F9,$4E,$78
	.dc.b	$1A,$49,$FE,$53,$F5,$9F,$DB,$9F,$F5,$1D,$BD,$9F,$F4,$57,$17,$F9
	.dc.b	$FA,$B5,$8A,$3F,$D0,$35,$BB,$FF,$55,$D7,$92,$B5,$1F,$EC,$58,$E3
	.dc.b	$5C,$F2,$EB,$39,$F0,$2D,$FC,$4B,$7B,$F8,$87,$D6,$7F,$B5,$AF,$BC
	.dc.b	$E6,$31,$3E,$03,$97,$C0,$F1,$BE,$77,$E6,$8C,$5E,$7B,$39,$7C,$75
	.dc.b	$AF,$F6,$2E,$28,$E5,$F0,$D9,$CB,$B0,$C9,$19,$D7,$FD,$4B,$F0,$FF
	.dc.b	$93,$C7,$FD,$9F,$E7,$7F,$8F,$9F,$F6,$8C,$67,$0E,$73,$97,$AF,$D6
	.dc.b	$3E,$7B,$E1,$56,$F0,$FF,$58,$CE,$70,$E7,$D7,$F1,$E1,$F2,$3F,$27
	.dc.b	$E1,$1F,$05,$98,$C3,$BF,$7F,$18,$7F,$B9,$2D,$18,$D3,$C7,$C1,$F4
	.dc.b	$F9,$5E,$E7,$78,$C3,$E1,$F7,$F1,$8F,$36,$DF,$B3,$DE,$D4,$F7,$6C
	.dc.b	$7C,$0F,$9D,$ED,$EF,$7A,$DB,$5E,$5C,$30,$D3,$D1,$1F,$07,$8B,$F1
	.dc.b	$1F,$4F,$FB,$5E,$3F,$8F,$18,$F0,$BE,$37,$BD,$BC,$3E,$03,$73,$FB
	.dc.b	$BF,$56,$FA,$9F,$C9,$7D,$63,$6F,$CA,$4E,$70,$FD,$DE,$7F,$EE,$4A
	.dc.b	$D4,$60,$FD,$FE,$0A,$FF,$38,$46,$63,$F8,$1C,$73,$7C,$3F,$03,$DB
	.dc.b	$1E,$FD,$7E,$08,$B7,$C6,$BA,$47,$93,$E8,$3F,$51,$F5,$8A,$C1,$B7
	.dc.b	$0F,$1C,$5E,$FE,$B0,$F8,$2D,$D3,$FD,$FF,$2F,$98,$F9,$BF,$59,$F0
	.dc.b	$9B,$87,$3F,$7F,$0D,$3B,$47,$C8,$E2,$FA,$BB,$D9,$F5,$72,$7C,$1C
	.dc.b	$BB,$F9,$79,$D7,$E9,$B6,$3C,$7F,$25,$F4,$A3,$E9,$B6,$FD,$9F,$0B
	.dc.b	$F8,$2E,$2F,$AE,$D8,$F8,$18,$61,$A7,$03,$E6,$7D,$67,$F1,$87,$EC
	.dc.b	$E5,$FE,$9F,$73,$DD,$F8,$EE,$5F,$E2,$D5,$FC,$87,$4F,$A1,$87,$F3
	.dc.b	$E8,$E4,$DF,$33,$77,$C3,$DF,$FC,$BF,$C2,$DC,$BE,$1B,$D3,$DA,$C6
	.dc.b	$1F,$C4,$31,$46,$32,$EF,$29,$7D,$37,$1C,$3D,$6D,$9F,$E3,$3F,$58
	.dc.b	$EE,$FE,$5C,$BB,$EC,$61,$FE,$65,$C2,$FA,$4B,$F0,$58,$6E,$5F,$93
	.dc.b	$FA,$CF,$EB,$19,$7F,$26,$BF,$EB,$F3,$9F,$63,$9C,$FF,$3D,$AF,$8A
	.dc.b	$97,$CA,$E7,$8F,$4E,$FD,$1E,$1E,$A5,$C7,$C0,$FF,$47,$FA,$57,$C8
	.dc.b	$DF,$3B,$0D,$AF,$D4,$4F,$ED,$B3,$FE,$76,$FF,$1C,$8C,$F9,$3C,$FC
	.dc.b	$BF,$EA,$5F,$58,$F2,$AF,$D3,$E1,$B7,$FB,$39,$7B,$BC,$3C,$3E,$2D
	.dc.b	$9D,$67,$D3,$93,$AB,$63,$F7,$83,$DB,$FE,$C4,$FF,$5A,$8F,$EA,$2E
	.dc.b	$4D,$FB,$7F,$D9,$FE,$B1,$F0,$97,$E1,$F0,$DA,$FE,$11,$F5,$8D,$FF
	.dc.b	$B7,$2F,$12,$C6,$1C,$ED,$AF,$2A,$DF,$27,$FB,$B7,$D6,$3C,$7B,$F1
	.dc.b	$58,$6D,$7F,$17,$9F,$57,$8F,$F9,$87,$F2,$F1,$9D,$F9,$F8,$DF,$E0
	.dc.b	$1F,$79,$C1,$7D,$FC,$36,$FF,$91,$4B,$D8,$E1,$F5,$9C,$5B,$3F,$C9
	.dc.b	$FE,$B3,$DC,$DF,$3F,$FC,$1D,$18,$E7,$FE,$FA,$FE,$FD,$BD,$9A,$E5
	.dc.b	$DB,$6F,$A7,$B1,$86,$9D,$E3,$AF,$93,$DB,$CA,$B3,$E3,$7C,$3E,$AE
	.dc.b	$31,$F0,$B7,$93,$F7,$B9,$BE,$43,$FC,$49,$CD,$F1,$31,$F6,$46,$38
	.dc.b	$BD,$7D,$FE,$C8,$B3,$A3,$FA,$46,$0F,$1F,$9F,$8F,$FE,$2F,$8C,$4F
	.dc.b	$F4,$DC,$7F,$C0,$6F,$0F,$23,$57,$B6,$C7,$F5,$3D,$7F,$CA,$0F,$1F
	.dc.b	$87,$FF,$1D,$46,$0F,$F1,$4D,$A3,$3D,$DA,$EE,$B7,$CD,$E3,$E2,$63
	.dc.b	$0F,$F1,$89,$FA,$E9,$F0,$57,$6D,$7E,$2F,$C4,$59,$F5,$3E,$2D,$F1
	.dc.b	$7C,$9B,$B7,$F9,$D7,$C9,$FF,$3C,$CF,$F9,$E9,$AF,$F7,$A6,$BF,$E7
	.dc.b	$A7,$3F,$55,$FB,$7F,$AC,$BE,$41,$A4,$AE,$B2,$F9,$FF,$F3,$64,$61
	.dc.b	$AF,$FA,$2A,$35,$87,$69,$CC,$FF,$83,$25,$F2,$3F,$E0,$EA,$FE,$87
	.dc.b	$18,$9F,$F0,$33,$5F,$F2,$CB,$D1,$7D,$67,$FA,$81,$FF,$4A,$C7,$E1
	.dc.b	$5E,$3C,$BD,$76,$F0,$36,$37,$85,$55,$FE,$64,$7C,$3C,$B9,$8C,$47
	.dc.b	$F9,$AB,$EB,$3F,$D4,$F9,$7A,$3E,$3B,$44,$9E,$27,$AB,$E8,$39,$11
	.dc.b	$F6,$C7,$87,$E0,$3F,$98,$C7,$F8,$95,$AF,$F9,$E1,$B5,$FD,$A3,$E4
	.dc.b	$FF,$D6,$B3,$BE,$5E,$44,$79,$73,$FB,$7E,$B3,$E5,$5F,$F7,$06,$F0
	.dc.b	$7F,$8F,$9B,$5F,$EB,$EC,$1F,$55,$C7,$86,$4F,$F8,$E6,$BE,$73,$FE
	.dc.b	$8C,$78,$AF,$AC,$FE,$42,$FF,$8F,$DA,$FF,$A4,$9B,$8F,$F8,$09,$FF
	.dc.b	$44,$46,$FD,$2F,$A9,$1F,$D1,$23,$07,$3F,$03,$3E,$04,$8E,$CF,$0F
	.dc.b	$8F,$E3,$EC,$49,$F1,$3B,$7C,$5F,$1B,$D8,$8F,$9F,$EA,$FC,$3C,$12
	.dc.b	$3D,$4F,$93,$E0,$3F,$11,$F5,$9F,$E4,$D9,$FF,$2E,$35,$FF,$57,$B7
	.dc.b	$6F,$F0,$FB,$FE,$13,$8F,$F3,$17,$B9,$F3,$9B,$C5,$F9,$38,$C4,$7F
	.dc.b	$AE,$1F,$BE,$DF,$85,$D1,$1E,$7E,$FF,$1E,$99,$1D,$9E,$9D,$BF,$9F
	.dc.b	$FA,$C6,$D4,$95,$FE,$75,$8F,$F0,$9B,$FE,$77,$FA,$CF,$43,$FE,$CA
	.dc.b	$6E,$BF,$E3,$A6,$EF,$FE,$88,$5E,$77,$EB,$7E,$B3,$D3,$5F,$E4,$06
	.dc.b	$D7,$FA,$39,$B6,$17,$FA,$E9,$D7,$89,$CC,$93,$FE,$8B,$AF,$F0,$83
	.dc.b	$73,$F0,$11,$D7,$5F,$EF,$E3,$DB,$1F,$EE,$46,$E3,$FE,$54,$6E,$1F
	.dc.b	$EA,$27,$FD,$9B,$3F,$E9,$97,$FC,$A0,$DF,$85,$E8,$75,$F2,$BF,$86
	.dc.b	$7D,$67,$F8,$C6,$7F,$D4,$0D,$7F,$DE,$6D,$C3,$FC,$C6,$FF,$87,$63
	.dc.b	$7E,$07,$22,$3C,$BC,$CF,$E3,$1F,$59,$FE,$6C,$8F,$F5,$63,$5F,$F7
	.dc.b	$FB,$6F,$FC,$DE,$FF,$87,$E3,$7E,$5F,$22,$3C,$BD,$DF,$E4,$9F,$59
	.dc.b	$FB,$18,$5F,$F9,$E2,$30,$DF,$0F,$7E,$17,$0F,$85,$97,$FE,$BF,$BF
	.dc.b	$F7,$8B,$7D,$BF,$E6,$1B,$FF,$9A,$23,$FD,$88,$D7,$BF,$17,$FB,$4A
	.dc.b	$31,$97,$F2,$1F,$86,$E3,$D9,$2F,$8B,$D2,$3A,$F6,$C9,$EA,$ED,$EA
	.dc.b	$7E,$3E,$FE,$23,$80,$E2,$FF,$D2,$ED,$C3,$FD,$CB,$EE,$3E,$8F,$D7
	.dc.b	$F1,$93,$0F,$5B,$EE,$F6,$3D,$FD,$92,$7D,$EF,$F5,$1F,$99,$7D,$67
	.dc.b	$C4,$F5,$25,$7C,$47,$83,$E5,$D1,$E2,$7C,$A2,$47,$FB,$B6,$FF,$9A
	.dc.b	$ED,$E8,$F4,$BE,$79,$27,$E7,$D7,$D2,$F9,$A9,$1F,$49,$D3,$F0,$FE
	.dc.b	$AF,$F9,$AB,$7E,$17,$FB,$44,$63,$0F,$F5,$DF,$D6,$3E,$B3,$3D,$73
	.dc.b	$0F,$F8,$01,$7F,$D0,$39,$7F,$7D,$08,$E5,$18,$7F,$5F,$1E,$1F,$4B
	.dc.b	$B9,$FD,$DF,$EB,$3E,$84,$74,$F4,$BC,$64,$7B,$CF,$9B,$E7,$64,$78
	.dc.b	$9F,$F0,$07,$D2,$7F,$10,$8F,$17,$CF,$7B,$08,$EC,$F4,$E9,$F0,$08
	.dc.b	$FB,$23,$E1,$F8,$32,$3D,$EF,$9F,$EA,$3E,$3F,$A5,$27,$E1,$BB,$7E
	.dc.b	$77,$AB,$F3,$08,$FF,$23,$E2,$FF,$73,$FD,$67,$FA,$7E,$3E,$5F,$0F
	.dc.b	$D2,$23,$F3,$AB,$FC,$39,$F5,$6F,$9B,$A4,$9E,$95,$F4,$BE,$B1,$F0
	.dc.b	$DC,$3C,$49,$5A,$3F,$9A,$60,$8E,$51,$FA,$78,$F2,$37,$E5,$3C,$27
	.dc.b	$27,$5F,$1D,$E4,$23,$D4,$7D,$63,$CA,$F4,$F9,$12,$75,$87,$E0,$1E
	.dc.b	$66,$3E,$7F,$DE,$47,$A7,$5F,$FA,$45,$E3,$C9,$F1,$08,$F1,$3C,$5F
	.dc.b	$59,$FE,$2E,$E0,$FC,$AB,$63,$FA,$A3,$57,$C2,$79,$DE,$3F,$0B,$FE
	.dc.b	$4B,$9F,$04,$7F,$0B,$6F,$32,$FA,$FA,$DF,$E5,$1F,$AC,$FC,$91,$3E
	.dc.b	$77,$FC,$A9,$93,$D0,$D7,$C1,$4C,$FE,$45,$FF,$81,$B2,$E9,$1F,$DA
	.dc.b	$63,$9F,$A1,$D4,$91,$F3,$F0,$FE,$1E,$F2,$F4,$7C,$44,$7A,$CF,$97
	.dc.b	$D1,$FF,$44,$DF,$FA,$06,$7F,$D1,$AD,$7F,$D2,$2D,$7F,$7A,$1D,$61
	.dc.b	$E6,$87,$C4,$FD,$B8,$C3,$76,$FE,$0B,$5C,$1F,$E0,$FF,$59,$F4,$49
	.dc.b	$BF,$F9,$E5,$7B,$76,$3F,$D3,$F1,$FE,$6A,$C7,$C6,$D7,$9C,$7B,$3D
	.dc.b	$F9,$1E,$C6,$BE,$C7,$EB,$7B,$F2,$3E,$8C,$7D,$9F,$4A,$47,$D4,$7E
	.dc.b	$E7,$A5,$88,$FE,$99,$9A,$71,$AF,$C3,$8F,$E7,$73,$C4,$87,$09,$E6
	.dc.b	$E2,$78,$47,$2D,$3E,$3D,$DA,$FC,$14,$D7,$E9,$45,$BF,$14,$73,$AE
	.dc.b	$2D,$38,$94,$E2,$58,$E2,$F9,$4F,$C9,$CF,$5A,$EC,$D8,$E0,$D7,$81
	.dc.b	$F1,$1C,$5F,$B2,$39,$9F,$08,$FC,$5D,$78,$1B,$5F,$94,$23,$F9,$46
	.dc.b	$BF,$94,$6C,$FE,$47,$EB,$3F,$D9,$13,$FD,$FD,$B1,$E1,$8F,$2B,$BF
	.dc.b	$8B,$FF,$63,$4E,$B1,$FC,$AD,$FC,$D8,$C5,$FF,$96,$5A,$FE,$86,$35
	.dc.b	$5C,$8F,$7D,$7F,$62,$AD,$9C,$30,$F9,$B8,$78,$CF,$16,$1F,$C9,$E7
	.dc.b	$58,$74,$8C,$61,$D1,$F1,$3D,$DF,$B1,$15,$F8,$11,$B5,$3E,$28,$C3
	.dc.b	$63,$89,$D8,$87,$C6,$C3,$F6,$FE,$A7,$F1,$B2,$47,$9F,$E4,$2B,$FC
	.dc.b	$1A,$3F,$85,$91,$FE,$0E,$7F,$A0,$7E,$B1,$F2,$4F,$99,$CC,$A7,$31
	.dc.b	$87,$FC,$6F,$E4,$5A,$E1,$5D,$3E,$B1,$F3,$7F,$C4,$63,$FC,$23,$18
	.dc.b	$CF,$AB,$63,$84,$ED,$FF,$4A,$43,$E0,$F1,$08,$C5,$72,$9C,$C6,$27
	.dc.b	$DB,$CA,$F9,$8E,$2D,$BF,$6B,$67,$73,$DB,$EA,$FF,$B2,$57,$F6,$7B
	.dc.b	$FF,$20,$BF,$DA,$63,$35,$FE,$59,$9C,$BF,$E5,$B8,$FE,$95,$3B,$F2
	.dc.b	$7F,$A3,$F2,$7F,$33,$9C,$D6,$E7,$35,$FE,$36,$9F,$55,$7F,$39,$FA
	.dc.b	$CF,$F6,$4C,$F8,$73,$DE,$1F,$EB,$99,$CB,$E1,$F8,$7C,$AC,$39,$DF
	.dc.b	$59,$CC,$62,$7F,$D8,$4B,$A7,$8D,$BE,$47,$71,$87,$8F,$5E,$31,$F8
	.dc.b	$8D,$C3,$FC,$F9,$F1,$7F,$8B,$9F,$F0,$87,$A2,$F2,$BC,$B3,$9A,$FF
	.dc.b	$41,$4E,$6B,$FC,$2D,$F5,$8F,$82,$DF,$4F,$B2,$31,$3F,$E1,$BC,$39
	.dc.b	$70,$FF,$5F,$97,$C9,$FC,$2F,$84,$76,$71,$8C,$3C,$FE,$03,$18,$F8
	.dc.b	$1B,$AF,$FA,$57,$0F,$85,$3F,$E2,$9E,$67,$B9,$E3,$E2,$F9,$F1,$89
	.dc.b	$F4,$FC,$06,$5A,$7B,$27,$D5,$EF,$7A,$E6,$CE,$B1,$F0,$7C,$4F,$4F
	.dc.b	$7F,$E0,$36,$7E,$2E,$58,$69,$F2,$23,$E3,$70,$FE,$0B,$E9,$DE,$36
	.dc.b	$2B,$FC,$85,$18,$7E,$7F,$BF,$9C,$BE,$03,$79,$3F,$C1,$5F,$54,$FA
	.dc.b	$9F,$C6,$FD,$63,$6F,$C7,$CE,$71,$FF,$27,$4E,$5F,$F2,$84,$7F,$99
	.dc.b	$5B,$5C,$3E,$0B,$8B,$CB,$AF,$13,$C1,$B9,$FF,$B1,$ED,$F4,$1F,$9E
	.dc.b	$FA,$CF,$F1,$F5,$F6,$8C,$3E,$4F,$82,$F2,$C5,$8D,$1E,$4F,$A1,$F8
	.dc.b	$DF,$59,$EA,$6C,$7A,$3E,$07,$0D,$38,$1E,$CE,$3F,$9D,$AF,$C2,$6C
	.dc.b	$7C,$0F,$43,$0D,$38,$1F,$13,$D6,$F9,$31,$F2,$9B,$1F,$23,$D0,$FB
	.dc.b	$5C,$3F,$39,$B1,$F3,$38,$30,$D3,$81,$FE,$7F,$FA,$CF,$E2,$FC,$BF
	.dc.b	$E4,$3E,$7F,$F4,$64,$E7,$BB,$F0,$DC,$BF,$C2,$2F,$F8,$AD,$FF,$92
	.dc.b	$5E,$35,$FE,$CA,$72,$DA,$E1,$ED,$7E,$3F,$E1,$AC,$FE,$39,$E9,$EE
	.dc.b	$63,$3F,$F8,$01,$82,$31,$8F,$7F,$4F,$78,$7A,$5B,$69,$FF,$2E,$7D
	.dc.b	$63,$B7,$F8,$FA,$EE,$B1,$87,$F9,$1F,$07,$F3,$FA,$FF,$32,$2F,$9A
	.dc.b	$CF,$A9,$A7,$F9,$4F,$D6,$3E,$0F,$8E,$78,$6E,$BF,$EA,$B9,$F8,$B2
	.dc.b	$0B,$E5,$79,$F1,$F4,$7A,$FA,$3D,$BD,$2B,$87,$37,$F9,$CF,$D2,$3E
	.dc.b	$42,$F9,$D8,$6D,$7F,$AE,$E7,$F5,$D9,$FC,$58,$FF,$7F,$C7,$F1,$38
	.dc.b	$C1,$E5,$77,$C1,$FE,$85,$74,$FF,$A1,$FE,$B3,$F5,$B5,$FE,$C7,$5F
	.dc.b	$4F,$FD,$31,$CD,$F4,$0E,$A7,$87,$B9,$BE,$FE,$0D,$8E,$95,$CF,$93
	.dc.b	$A3,$63,$FD,$AC,$3D,$AF,$EB,$6F,$F5,$78,$FF,$5E,$46,$3D,$BF,$EB
	.dc.b	$FF,$58,$F9,$BB,$F2,$D8,$6E,$1F,$EE,$1F,$AC,$F3,$63,$E9,$5E,$0F
	.dc.b	$F6,$87,$3C,$1F,$EA,$97,$35,$C7,$6D,$78,$DB,$6C,$7F,$BB,$7E,$AD
	.dc.b	$E2,$7F,$72,$5F,$59,$63,$0E,$77,$D7,$87,$6F,$93,$FD,$EB,$DD,$70
	.dc.b	$5F,$7B,$0D,$AF,$F7,$E4,$BD,$6E,$1F,$59,$C5,$B3,$FF,$00,$7E,$B3
	.dc.b	$D7,$C7,$F6,$5F,$E0,$6D,$E3,$FF,$04,$C6,$38,$FF,$D9,$4F,$E0,$1B
	.dc.b	$D5,$AF,$4F,$6F,$0F,$53,$18,$F7,$6E,$3D,$FC,$7E,$BE,$3F,$66,$D7
	.dc.b	$97,$D3,$C3,$4F,$44,$79,$3E,$4B,$FC,$3B,$CD,$F1,$27,$D9,$18,$E2
	.dc.b	$F4,$F7,$6E,$7E,$03,$72,$FC,$C6,$DF,$E1,$F9,$FD,$C3,$FE,$E7,$C1
	.dc.b	$E1,$FF,$8A,$A3,$55,$F9,$F9,$CD,$70,$3F,$80,$47,$F1,$5F,$17,$87
	.dc.b	$8B,$C5,$B3,$B9,$DF,$6F,$AB,$E1,$BF,$C6,$F0,$8C,$4E,$E3,$F8,$67
	.dc.b	$76,$A7,$6D,$9D,$4B,$C3,$CC,$61,$B1,$FB,$71,$BF,$CA,$F9,$DF,$0B
	.dc.b	$6B,$9E,$F8,$69,$DA,$3C,$F4,$F1,$BF,$D0,$EF,$F0,$D2,$B8,$74,$7F
	.dc.b	$A4,$3E,$AB,$F7,$9F,$59,$FE,$8E,$6C,$67,$C7,$4E,$5B,$1C,$7E,$47
	.dc.b	$FA,$13,$78,$F9,$46,$23,$E9,$8B,$46,$63,$EB,$C3,$E8,$3E,$CD,$33
	.dc.b	$CC,$61,$B5,$F4,$EB,$9D,$7F,$3F,$FA,$CF,$F6,$EC,$7F,$9A,$67,$5C
	.dc.b	$9C,$5F,$E8,$91,$97,$F9,$43,$E1,$7A,$72,$6D,$F4,$36,$0E,$7E,$4F
	.dc.b	$DB,$87,$9E,$BA,$FC,$96,$C8,$F4,$3D,$1F,$23,$CC,$8F,$AE,$3C,$1E
	.dc.b	$E7,$E2,$63,$F2,$0D,$7F,$9F,$36,$BC,$10,$F9,$4F,$EC,$CF,$0E,$3E
	.dc.b	$64,$79,$F3,$7B,$7E,$B3,$E0,$47,$F4,$96,$DF,$46,$F0,$7E,$B3,$07
	.dc.b	$CB,$71,$E0,$98,$7E,$96,$FE,$66,$1C,$78,$BE,$57,$D6,$7F,$28,$AF
	.dc.b	$D5,$35,$FE,$B0,$D9,$F5,$BF,$E8,$F8,$DF,$A3,$C8,$8F,$2F,$03,$AF
	.dc.b	$81,$23,$AB,$E0,$F8,$EF,$0F,$AB,$27,$C2,$ED,$F0,$FC,$67,$98,$8F
	.dc.b	$9B,$EA,$FD,$3C,$32,$3D,$2F,$B3,$DD,$FC,$37,$D6,$3C,$8A,$4A,$FE
	.dc.b	$11,$1F,$EE,$17,$F8,$57,$C5,$FC,$A6,$F2,$7E,$3E,$31,$1F,$DE,$DF
	.dc.b	$B6,$DF,$9A,$D1,$1E,$7E,$EF,$1E,$B9,$1D,$9E,$9D,$DF,$9B,$FA,$C6
	.dc.b	$D4,$95,$FC,$76,$3F,$C3,$CF,$F1,$FF,$AC,$F4,$3F,$E0,$D6,$F3,$FE
	.dc.b	$41,$BC,$1F,$CC,$1F,$F1,$D4,$FF,$27,$FA,$CF,$4D,$7E,$59,$B5,$FC
	.dc.b	$D9,$B3,$FE,$37,$AF,$D5,$CE,$7C,$4F,$D9,$57,$F0,$8A,$E9,$18,$F1
	.dc.b	$73,$9C,$9F,$03,$C3,$B7,$F7,$4F,$B2,$3F,$C4,$CD,$CB,$F5,$2D,$C7
	.dc.b	$FA,$4B,$FD,$42,$7F,$A0,$3F,$E5,$B6,$FC,$1F,$43,$AF,$95,$FC,$1F
	.dc.b	$EB,$3F,$D4,$73,$FD,$1D,$AF,$F8,$D5,$B8,$7E,$E1,$FF,$1D,$C6,$FC
	.dc.b	$2E,$44,$79,$79,$9F,$C4,$FE,$B3,$FC,$ED,$1F,$D4,$DA,$FF,$90,$5B
	.dc.b	$7F,$C1,$1F,$F1,$0C,$6F,$5E,$44,$79,$7B,$BF,$C7,$FE,$B3,$FC,$71
	.dc.b	$1F,$D7,$DA,$FF,$93,$5B,$7F,$C4,$1F,$DC,$6D,$FD,$C1,$FF,$79,$CF
	.dc.b	$1F,$93,$E8,$75,$F6,$FF,$96,$6D,$FE,$73,$9F,$ED,$8D,$7F,$CB,$4D
	.dc.b	$C3,$F8,$EF,$91,$71,$FE,$FB,$2E,$9E,$64,$BE,$7F,$0B,$DB,$AE,$47
	.dc.b	$67,$4F,$6F,$57,$C6,$47,$FB,$F7,$C5,$BE,$93,$CF,$F1,$72,$7C,$EE
	.dc.b	$DF,$3F,$57,$B0,$8F,$F8,$1F,$E6,$5F,$59,$F0,$3C,$E9,$3F,$01,$DB
	.dc.b	$F0,$3C,$2F,$8A,$47,$97,$77,$C8,$F4,$A4,$7C,$87,$4F,$C8,$F9,$AF
	.dc.b	$9A,$47,$47,$F8,$67,$EA,$FF,$96,$3F,$E8,$76,$F3,$7F,$4A,$6F,$2F
	.dc.b	$F8,$AB,$EB,$1F,$57,$FD,$8E,$7E,$FD,$FF,$AB,$5A,$9A,$72,$75,$E8
	.dc.b	$71,$23,$FE,$2B,$FA,$7F,$9C,$F9,$7D,$0E,$C4,$7D,$13,$E1,$F9,$C9
	.dc.b	$1E,$B7,$FB,$B7,$D6,$7F,$0A,$8F,$17,$CF,$79,$C8,$EC,$F8,$59,$7A
	.dc.b	$C8,$FC,$58,$F7,$78,$12,$3E,$43,$FE,$03,$C9,$F4,$1F,$0F,$CF,$95
	.dc.b	$F3,$5E,$0F,$CD,$F4,$FE,$41,$1F,$F2,$57,$D6,$7F,$A7,$9F,$A9,$C3
	.dc.b	$F3,$48,$FD,$79,$FF,$09,$7D,$53,$E6,$E9,$27,$FC,$A9,$F5,$8F,$85
	.dc.b	$DF,$83,$27,$FC,$B1,$68,$E1,$1F,$EC,$78,$F1,$37,$E3,$BB,$9C,$5D
	.dc.b	$7C,$5F,$88,$8F,$F9,$8F,$EB,$3F,$64,$F6,$F1,$3C,$E4,$7C,$33,$EA
	.dc.b	$F9,$F9,$1E,$47,$E6,$7C,$1F,$BB,$8A,$4F,$A1,$DB,$FE,$DD,$FA,$CF
	.dc.b	$F1,$07,$0F,$F9,$F9,$B1,$FE,$FF,$6B,$FE,$23,$F2,$FF,$C7,$51,$FE
	.dc.b	$6E,$8F,$F3,$D3,$5E,$97,$AF,$53,$FC,$81,$F5,$7F,$8B,$27,$CE,$FF
	.dc.b	$90,$FD,$07,$9D,$AF,$7A,$63,$FE,$85,$BF,$EF,$3E,$2F,$F8,$D1,$E5
	.dc.b	$E7,$F4,$24,$FF,$A2,$F0,$FE,$0A,$F1,$F3,$FC,$24,$7A,$4F,$93,$CF
	.dc.b	$91,$DD,$FF,$29,$71,$E2,$3F,$CD,$E3,$0F,$ED,$0E,$1A,$FF,$9D,$1A
	.dc.b	$FF,$AA,$3E,$B1,$EC,$7F,$A2,$9B,$1D,$03,$87,$BB,$5F,$F3,$FB,$F4
	.dc.b	$89,$BF,$F9,$AD,$E1,$A7,$C5,$F3,$7E,$79,$2B,$E2,$7C,$CD,$53,$B7
	.dc.b	$E7,$4F,$D3,$9F,$AF,$E7,$C9,$FA,$4F,$D7,$D7,$11,$CB,$9B,$FD,$8B
	.dc.b	$A2,$B3,$8F,$F4,$EB,$B5,$A9,$B6,$3F,$CD,$D3,$87,$E7,$45,$E0,$FE
	.dc.b	$DF,$1D,$CE,$18,$77,$7C,$A4,$34,$74,$31,$87,$E2,$6B,$A5,$F7,$AE
	.dc.b	$55,$C1,$B5,$E9,$8E,$38,$7A,$5A,$7A,$70,$F2,$CF,$58,$C9,$EA,$3F
	.dc.b	$29,$7D,$F0,$F0,$C6,$B0,$EF,$87,$58,$F4,$E1,$D2,$31,$87,$59,$F3
	.dc.b	$5F,$D3,$7E,$01,$F9,$17,$FD,$AF,$7E,$C8,$C1,$F8,$F7,$0F,$73,$27
	.dc.b	$E6,$E3,$89,$FB,$8F,$3E,$47,$9E,$2B,$F1,$E3,$D8,$FE,$DD,$CD,$79
	.dc.b	$C9,$7F,$86,$8C,$BE,$C2,$4F,$DC,$8C,$BF,$7A,$BE,$B4,$62,$3E,$61
	.dc.b	$F0,$8F,$01,$0C,$12,$BF,$84,$47,$1B,$F9,$F3,$D4,$E1,$3A,$7D,$C6
	.dc.b	$2F,$D0,$E0,$F8,$06,$4E,$47,$77,$6E,$8C,$1F,$9F,$FA,$C7,$C8,$7F
	.dc.b	$B8,$DB,$1F,$A3,$AF,$F6,$8C,$62,$3F,$36,$BA,$F8,$57,$F0,$99,$4E
	.dc.b	$DF,$D6,$1D,$E3,$51,$98,$CF,$D5,$FF,$91,$5D,$87,$0E,$BD,$6E,$D1
	.dc.b	$88,$FD,$C7,$D6,$3E,$4D,$F3,$5F,$C6,$E3,$0D,$F6,$BF,$B5,$57,$F2
	.dc.b	$A9,$CB,$FC,$B2,$FF,$CA,$1A,$FF,$23,$6C,$67,$C0,$FF,$79,$BE,$86
	.dc.b	$BF,$C2,$61,$EE,$E2,$3D,$4D,$8C,$D7,$58,$D7,$D6,$76,$E0,$FF,$7F
	.dc.b	$D7,$F1,$9A,$FA,$47,$CA,$9D,$3F,$23,$8C,$EB,$0F,$E3,$53,$FD,$6A
	.dc.b	$31,$08,$E2,$EB,$BF,$8B,$FC,$5A,$32,$EF,$1C,$E3,$A8,$7D,$67,$FB
	.dc.b	$3B,$0E,$B7,$EC,$9F,$E3,$F3,$97,$A7,$C0,$72,$AF,$56,$16,$7C,$F3
	.dc.b	$5F,$04,$F4,$EA,$C5,$BE,$2F,$58,$A7,$31,$8F,$85,$CC,$5B,$E4,$7C
	.dc.b	$2F,$85,$FC,$2B,$FE,$82,$F2,$3C,$11,$8C,$BF,$9B,$F9,$BF,$E4,$2A
	.dc.b	$F9,$3F,$58,$F9,$ED,$F3,$F9,$E3,$13,$D6,$31,$F1,$DE,$BF,$84,$7C
	.dc.b	$16,$2B,$FD,$10,$F1,$F7,$F1,$9F,$F4,$6C,$51,$8E,$5F,$A1,$1F,$06
	.dc.b	$FE,$DF,$4B,$DB,$F1,$71,$FC,$98,$C4,$FA,$3D,$FC,$B4,$F5,$CF,$A7
	.dc.b	$DD,$F5,$4D,$9D,$67,$E3,$F8,$5E,$9E,$EF,$BD,$B3,$F1,$32,$C3,$4F
	.dc.b	$8F,$1F,$17,$93,$EE,$7D,$3B,$E5,$31,$5F,$94,$8C,$3F,$3B,$DE,$CE
	.dc.b	$5F,$0D,$BC,$1F,$91,$FA,$B7,$D4,$FE,$1F,$EB,$1B,$7E,$26,$73,$8F
	.dc.b	$E8,$27,$2F,$E8,$A3,$FC,$DC,$DA,$E1,$F0,$3C,$5E,$5D,$78,$9E,$0D
	.dc.b	$CF,$FB,$DD,$BE,$83,$F3,$1F,$59,$FB,$2B,$ED,$18,$7C,$9E,$BF,$96
	.dc.b	$2C,$68,$F2,$7C,$AF,$C6,$FA,$CF,$53,$63,$D1,$E8,$E1,$A7,$03,$D9
	.dc.b	$C7,$F3,$B5,$F8,$4D,$8F,$81,$E8,$E1,$A7,$03,$E2,$7B,$3F,$26,$3E
	.dc.b	$53,$63,$E4,$7C,$27,$DA,$E1,$F9,$CD,$8F,$99,$C1,$86,$9C,$0F,$E2
	.dc.b	$3F,$59,$FC,$33,$9F,$FD,$2D,$18,$EE,$FC,$2F,$27,$F0,$1A,$FE,$31
	.dc.b	$7F,$E2,$87,$F8,$99,$8E,$7F,$E2,$4E,$3D,$8F,$C6,$FC,$27,$16,$E3
	.dc.b	$E0,$F7,$B0,$D3,$C7,$1D,$FE,$9B,$C1,$36,$75,$8F,$F8,$E3,$EB,$1D
	.dc.b	$BF,$C5,$57,$75,$8C,$3F,$C6,$78,$1F,$47,$7E,$93,$0D,$C7,$F9,$9F
	.dc.b	$D6,$3E,$17,$86,$7F,$C8,$63,$13,$FC,$E3,$0F,$F2,$A3,$C3,$8F,$5E
	.dc.b	$E6,$DB,$1D,$3D,$2E,$5E,$97,$36,$C7,$F4,$2F,$A5,$7C,$8F,$F2,$F7
	.dc.b	$F8,$7C,$66,$31,$E7,$67,$C8,$D7,$95,$6B,$93,$FC,$DF,$EB,$3F,$45
	.dc.b	$1F,$D3,$97,$D3,$E7,$C7,$D3,$BF,$87,$E1,$E1,$5B,$E0,$F3,$E3,$E9
	.dc.b	$C5,$67,$A3,$FD,$15,$7B,$57,$DB,$C3,$63,$FA,$FF,$D6,$3E,$8B,$FA
	.dc.b	$62,$FA,$2B,$18,$7F,$A8,$7D,$67,$A2,$BF,$B3,$AF,$0F,$3B,$69,$DE
	.dc.b	$BC,$3C,$9B,$7C,$1F,$EB,$5F,$58,$F1,$6F,$C5,$E1,$B5,$FD,$D2,$7F
	.dc.b	$B4,$79,$1F,$DB,$57,$8F,$CB,$87,$9B,$FD,$97,$DD,$F5,$5F,$F7,$C7
	.dc.b	$FA,$8F,$27,$FA,$39,$CB,$7D,$EE,$1E,$BF,$1F,$AC,$B7,$C5,$FE,$E1
	.dc.b	$F5,$9E,$F8,$FF,$03,$BF,$B6,$BF,$F4,$CB,$96,$F9,$DC,$3A,$F8,$F2
	.dc.b	$5B,$E2,$F4,$F5,$B1,$7C,$9B,$C1,$FE,$0B,$C7,$F4,$F7,$E0,$F2,$BC
	.dc.b	$7E,$57,$91,$F2,$FA,$B8,$9F,$2B,$71,$F2,$FC,$8F,$F8,$27,$97,$E0
	.dc.b	$E5,$D6,$31,$D7,$EE,$EF,$F7,$C5,$C7,$DE,$FF,$84,$6B,$FC,$19,$1F
	.dc.b	$E2,$8B,$FD,$2E,$1F,$EB,$A9,$CB,$FE,$BC,$8C,$C7,$F8,$94,$78,$FC
	.dc.b	$FC,$8E,$8D,$BF,$F1,$9D,$FF,$9C,$1E,$DF,$15,$FE,$21,$8C,$1D,$5C
	.dc.b	$76,$E7,$B6,$D7,$27,$87,$87,$8B,$FF,$67,$87,$19,$7F,$96,$34,$F8
	.dc.b	$DF,$15,$DD,$B8,$79,$36,$C3,$4F,$3C,$79,$70,$79,$5F,$E7,$F6,$9F
	.dc.b	$E8,$17,$07,$F9,$E9,$FF,$76,$7D,$57,$EE,$3E,$B3,$10,$8C,$3B,$6B
	.dc.b	$FE,$82,$6B,$FE,$8A,$5F,$24,$CF,$9C,$61,$B1,$99,$E8,$D7,$A7,$CF
	.dc.b	$FF,$A5,$27,$FC,$B9,$48,$C6,$5D,$DB,$5E,$AB,$E9,$F5,$9F,$E1,$B8
	.dc.b	$F6,$F1,$E9,$F6,$37,$0F,$F2,$B4,$7F,$9C,$9E,$CD,$E3,$D1,$1F,$F3
	.dc.b	$BF,$D6,$7F,$A8,$B3,$F2,$FC,$77,$32,$4F,$13,$A7,$C9,$F7,$23,$DE
	.dc.b	$3C,$9E,$F7,$E0,$E3,$FD,$06,$D7,$E5,$6D,$FE,$B6,$8C,$5F,$F9,$FF
	.dc.b	$E5,$3F,$AB,$57,$0E,$2E,$64,$79,$F2,$FC,$1F,$AC,$F6,$C7,$40,$1C
	.dc.b	$DF,$FA,$2D,$B8,$FF,$B1,$B0,$7C,$BF,$3E,$09,$87,$FA,$6A,$FE,$5B
	.dc.b	$D4,$E9,$C7,$F4,$7E,$B3,$F9,$F4,$7F,$A8,$1A,$ED,$B9,$7F,$A6,$DF
	.dc.b	$F3,$E4,$6F,$D0,$E4,$47,$97,$77,$5F,$5A,$47,$57,$C1,$F1,$FE,$1F
	.dc.b	$62,$4F,$85,$DB,$E1,$F8,$CF,$31,$1F,$37,$D5,$FA,$78,$64,$7A,$5F
	.dc.b	$67,$C0,$7E,$37,$EB,$1E,$35,$25,$7F,$AF,$A3,$FC,$22,$FF,$B0,$3D
	.dc.b	$CF,$96,$DE,$4F,$CB,$46,$23,$E5,$3F,$79,$BF,$2B,$A2,$3C,$FD,$EE
	.dc.b	$3D,$72,$3B,$3D,$3B,$BF,$45,$F5,$8D,$A9,$2B,$FD,$A7,$1F,$E1,$27
	.dc.b	$FD,$AB,$F5,$9E,$87,$EE,$37,$9F,$FD,$A0,$DE,$0F,$C5,$CA,$1E,$67
	.dc.b	$EC,$7E,$B3,$D5,$7F,$ED,$76,$DF,$06,$DF,$E5,$2F,$F5,$D3,$9F,$17
	.dc.b	$91,$27,$FD,$D3,$5F,$E1,$06,$E7,$DC,$8E,$BB,$7F,$81,$47,$B6,$3F
	.dc.b	$2C,$DC,$7F,$DD,$4D,$C3,$D6,$FF,$B0,$27,$FD,$EE,$FF,$93,$9B,$F0
	.dc.b	$BD,$0E,$BE,$57,$F0,$FF,$AC,$FF,$49,$4F,$FB,$F9,$AF,$E9,$5B,$87
	.dc.b	$FB,$E5,$FF,$1A,$C6,$FC,$2E,$44,$79,$79,$BF,$C6,$FE,$B1,$F1,$34
	.dc.b	$93,$FB,$CC,$1A,$7D,$4A,$FF,$10,$4E,$63,$1F,$07,$CC,$91,$93,$F8
	.dc.b	$2C,$BD,$EF,$E5,$5F,$59,$FE,$39,$9D,$07,$0F,$F0,$16,$E1,$E2,$8F
	.dc.b	$E1,$F1,$FB,$EC,$3F,$0E,$FF,$B9,$E3,$9B,$5E,$5F,$0D,$E0,$F7,$7F
	.dc.b	$9D,$6E,$F8,$7A,$4A,$F1,$F9,$79,$7C,$B2,$EC,$E2,$48,$CD,$74,$ED
	.dc.b	$91,$C1,$D3,$FC,$6E,$BE,$4B,$7D,$7F,$19,$FD,$37,$DD,$7D,$17,$9B
	.dc.b	$8E,$57,$89,$DB,$E4,$EB,$F5,$91,$F9,$7F,$32,$FA,$CF,$7F,$9F,$27
	.dc.b	$D6,$ED,$F2,$78,$7F,$10,$8F,$2E,$EF,$8F,$E9,$C8,$F8,$6E,$9F,$67
	.dc.b	$CD,$7C,$C2,$3A,$3E,$EF,$D5,$FF,$32,$7F,$9E,$37,$A7,$EE,$42,$31
	.dc.b	$D3,$FC,$0B,$EB,$1F,$57,$FD,$C6,$7F,$01,$5F,$D4,$1A,$F1,$F4,$B5
	.dc.b	$F0,$DB,$F4,$E4,$F8,$9F,$EF,$5F,$57,$F3,$A3,$CB,$E9,$F5,$23,$B3
	.dc.b	$D3,$F3,$9E,$12,$3F,$98,$FA,$CF,$E1,$EF,$A7,$E7,$BC,$C4,$78,$9F
	.dc.b	$0B,$4F,$51,$1F,$8B,$1E,$DF,$0A,$47,$C8,$7C,$FF,$41,$F0,$7D,$29
	.dc.b	$3F,$31,$DB,$F3,$3D,$6F,$8C,$47,$F5,$BF,$59,$FE,$9F,$7E,$9F,$1F
	.dc.b	$CB,$23,$F5,$A7,$FC,$3D,$F5,$6F,$9B,$A4,$9F,$DB,$FD,$63,$E1,$F7
	.dc.b	$E1,$C9,$FD,$DD,$A3,$84,$7F,$7B,$8F,$13,$7E,$53,$B9,$C5,$D7,$C6
	.dc.b	$F8,$88,$FF,$06,$FA,$CF,$D9,$3D,$BC,$6F,$39,$1F,$0C,$FA,$BE,$7E
	.dc.b	$47,$91,$FC,$0F,$C2,$FB,$B8,$A4,$FA,$1D,$BF,$E1,$1F,$AC,$FF,$13
	.dc.b	$F0,$FF,$54,$6C,$7F,$88,$1A,$BC,$FF,$2B,$C6,$FF,$C9,$12,$69,$8F
	.dc.b	$5B,$FC,$99,$F5,$7F,$8A,$27,$CE,$FF,$93,$F0,$7A,$1A,$F8,$19,$97
	.dc.b	$F2,$8B,$FE,$FD,$E3,$7E,$A9,$E5,$E8,$74,$24,$FF,$1C,$F3,$7B,$FA
	.dc.b	$12,$39,$3A,$7F,$C6,$15,$FD,$8D,$FF,$19,$7D,$63,$E4,$3F,$CC,$4D
	.dc.b	$8F,$17,$B5,$89,$FF,$40,$C2,$74,$FF,$A1,$DA,$FF,$7E,$35,$1C,$2B
	.dc.b	$FD,$03,$1F,$E8,$A6,$BF,$E8,$F6,$BC,$5A,$FF,$82,$1D,$E5,$FC,$33
	.dc.b	$E3,$3F,$7D,$3F,$E4,$B6,$C7,$C7,$F5,$3E,$8F,$A7,$FE,$7D,$9F,$F0
	.dc.b	$7D,$7F,$9D,$BE,$0B,$2D,$21,$CD,$CB,$F2,$3E,$03,$ED,$91,$FA,$DE
	.dc.b	$A5,$74,$7D,$98,$FC,$0C,$67,$1F,$B5,$17,$A3,$D9,$36,$BF,$CD,$F9
	.dc.b	$FF,$B4,$9E,$47,$62,$9D,$89,$7F,$85,$9E,$50,$9E,$4E,$AB,$BB,$CA
	.dc.b	$BC,$6D,$8C,$BB,$3C,$77,$D2,$BF,$CF,$35,$C2,$74,$77,$9F,$7B,$CC
	.dc.b	$E2,$53,$C9,$7C,$E3,$A3,$63,$55,$DA,$FF,$01,$5F,$89,$24,$FE,$25
	.dc.b	$AF,$E2,$5A,$FE,$25,$CD,$FF,$52,$F6,$3F,$C6,$93,$FE,$91,$6C,$7A
	.dc.b	$C9,$5F,$E5,$68,$CB,$7E,$03,$CD,$5C,$5F,$F6,$44,$EA,$3D,$E7,$C4
	.dc.b	$70,$FF,$84,$E3,$8D,$78,$DE,$26,$2B,$B4,$FF,$8A,$2B,$8E,$1F,$6A
	.dc.b	$FC,$B3,$FE,$B4,$9E,$A6,$B0,$FC,$0B,$8C,$3A,$4E,$DE,$EF,$CE,$9A
	.dc.b	$FA,$CC,$1E,$18,$CD,$78,$23,$EF,$19,$21,$F8,$2F,$AC,$FF,$76,$3F
	.dc.b	$EB,$D6,$C7,$F8,$09,$B8,$FE,$EF,$D7,$FF,$78,$47,$0A,$FE,$10,$D7
	.dc.b	$FD,$0E,$D7,$FD,$A6,$D4,$FF,$31,$7F,$88,$1A,$EB,$FE,$2F,$19,$8C
	.dc.b	$B5,$EF,$E2,$E8,$8F,$23,$AB,$FC,$83,$EB,$1F,$26,$F9,$9D,$46,$1B
	.dc.b	$EE,$FF,$69,$BF,$1C,$E5,$F2,$FD,$63,$E6,$FF,$C4,$B1,$FE,$A4,$8C
	.dc.b	$6B,$E1,$6C,$70,$9D,$BF,$EF,$4F,$AB,$FF,$2C,$57,$F9,$BA,$9E,$FF
	.dc.b	$F9,$AF,$83,$FC,$63,$5A,$BF,$F7,$85,$6A,$FF,$CE,$B3,$FD,$49,$BF
	.dc.b	$03,$C5,$FF,$5B,$C6,$6B,$FD,$F0,$F3,$BE,$BF,$09,$B9,$0F,$AC,$FF
	.dc.b	$65,$5F,$69,$D4,$E7,$E3,$2E,$0F,$02,$D6,$1B,$7E,$6E,$AE,$38,$BE
	.dc.b	$3F,$8C,$61,$E7,$D3,$8A,$F6,$B7,$36,$9F,$E5,$4E,$9F,$B5,$5F,$E9
	.dc.b	$6F,$49,$E6,$79,$A7,$37,$C2,$BF,$DB,$35,$A7,$FD,$3B,$F5,$8F,$81
	.dc.b	$E1,$D1,$EE,$8C,$4F,$FA,$8B,$0E,$79,$7F,$99,$25,$F2,$3F,$13,$E1
	.dc.b	$DF,$0F,$C6,$30,$F3,$F8,$4C,$63,$E0,$6E,$BE,$53,$E8,$E1,$F1,$67
	.dc.b	$FD,$63,$CE,$F7,$BC,$9C,$5F,$4A,$31,$3E,$AF,$82,$CB,$4F,$6C,$FA
	.dc.b	$FE,$03,$D9,$36,$75,$8F,$84,$62,$F8,$8E,$DF,$03,$F0,$DB,$5F,$23
	.dc.b	$3C,$34,$F9,$71,$F2,$78,$7F,$0F,$F4,$EF,$96,$C5,$7F,$B3,$23,$0F
	.dc.b	$D2,$F8,$2C,$E7,$E3,$CD,$18,$E4,$FF,$49,$FD,$5B,$EA,$7F,$27,$F5
	.dc.b	$8D,$FF,$2B,$39,$CB,$FD,$B3,$39,$7F,$DB,$71,$FE,$69,$6D,$70,$F8
	.dc.b	$5E,$2F,$2E,$BC,$4F,$06,$E9,$F1,$ED,$F4,$3F,$A4,$FA,$CF,$F7,$9D
	.dc.b	$F6,$8C,$3E,$4F,$82,$F2,$C5,$8D,$1E,$4F,$98,$F8,$FF,$59,$EA,$6C
	.dc.b	$7A,$3E,$13,$0D,$38,$1E,$CE,$4F,$9F,$AF,$C2,$6C,$7C,$0F,$6B,$0D
	.dc.b	$38,$1F,$13,$E5,$3E,$4C,$7C,$A6,$C7,$C8,$F4,$BE,$DF,$0F,$CE,$6C
	.dc.b	$7C,$CE,$0C,$34,$E0,$7F,$0C,$E4,$FF,$65,$FD,$67,$F1,$18,$FE,$1F
	.dc.b	$2E,$8F,$F4,$A4,$6C,$CC,$EE,$7F,$8A,$CE,$9F,$CC,$47,$F1,$DA,$EB
	.dc.b	$DF,$FC,$5A,$31,$7C,$A3,$1F,$01,$F9,$4F,$89,$B3,$F9,$37,$87,$C0
	.dc.b	$63,$1E,$AD,$C7,$D3,$E9,$EA,$FC,$D5,$FC,$A3,$2F,$E7,$F5,$EE,$BF
	.dc.b	$E4,$FF,$59,$F9,$AB,$DD,$A1,$DF,$AE,$ED,$BF,$CB,$B7,$7D,$1D,$F7
	.dc.b	$B0,$DC,$79,$FD,$63,$E2,$38,$67,$FC,$DA,$31,$3E,$0C,$3F,$9E,$BC
	.dc.b	$38,$B5,$EF,$ED,$B1,$D3,$D1,$E5,$E9,$73,$6C,$7A,$BE,$91,$F2,$7F
	.dc.b	$D1,$17,$9D,$63,$0E,$7C,$8D,$79,$76,$F9,$3F,$D3,$3E,$B3,$F5,$71
	.dc.b	$F1,$97,$D3,$E7,$C6,$D3,$BF,$8D,$E1,$E0,$DB,$E0,$F3,$E3,$E9,$C5
	.dc.b	$67,$A3,$FD,$69,$FE,$B0,$7D,$41,$ED,$67,$DA,$D3,$FD,$83,$EB,$1F
	.dc.b	$45,$7E,$67,$0D,$BF,$BB,$F5,$9E,$86,$F8,$9F,$DA,$23,$19,$FF,$2F
	.dc.b	$51,$9D,$35,$E4,$5B,$E6,$FF,$70,$FA,$B7,$8F,$7E,$B3,$0D,$AF,$C6
	.dc.b	$CF,$A9,$BE,$47,$1D,$F9,$3F,$DE,$BD,$D6,$F7,$DE,$C3,$6F,$F2,$F3
	.dc.b	$FD,$0D,$BF,$59,$C7,$D9,$E4,$FF,$81,$3E,$B3,$D7,$3F,$9E,$7F,$6D
	.dc.b	$E6,$23,$F7,$F1,$AE,$3F,$EA,$0E,$A3,$3D,$5C,$FA,$BA,$36,$3A,$7A
	.dc.b	$B8,$69,$DE,$3C,$5E,$47,$8F,$C9,$B3,$E3,$7C,$DE,$A6,$32,$F0,$B7
	.dc.b	$3F,$34,$7F,$88,$B3,$7C,$5F,$F8,$83,$97,$E3,$E3,$ED,$8C,$70,$FC
	.dc.b	$1E,$EF,$85,$16,$74,$7F,$00,$C5,$E3,$F3,$F1,$3F,$C5,$71,$89,$FE
	.dc.b	$0B,$8F,$EA,$9E,$1E,$3E,$AF,$6D,$8F,$E1,$7A,$7F,$C0,$17,$8F,$C5
	.dc.b	$7F,$8D,$A3,$07,$F7,$DB,$FF,$30,$CF,$0E,$DD,$76,$ED,$B1,$D3,$C3
	.dc.b	$C3,$4E,$07,$83,$5F,$89,$F1,$5D,$DB,$1E,$4D,$F0,$D3,$81,$E8,$C9
	.dc.b	$E4,$FF,$9D,$DF,$F2,$8B,$5C,$99,$7F,$CF,$9F,$55,$FE,$4F,$FA,$CB
	.dc.b	$E3,$9A,$4A,$FF,$45,$CB,$E4,$BF,$C9,$31,$86,$BF,$D4,$84,$6F,$8F
	.dc.b	$BC,$EA,$7F,$AB,$CB,$E4,$BF,$AC,$D7,$E2,$27,$8C,$63,$2D,$B6,$3F
	.dc.b	$CD,$B5,$D5,$7D,$67,$FA,$C2,$3F,$DD,$11,$AE,$5E,$0F,$91,$B1,$E1
	.dc.b	$8F,$E9,$6F,$FB,$36,$FC,$BC,$BA,$38,$FC,$9F,$B7,$2F,$35,$7F,$A3
	.dc.b	$AF,$8F,$C7,$F4,$38,$3D,$3E,$4B,$B9,$1D,$9F,$D6,$BD,$D7,$2F,$99
	.dc.b	$B1,$98,$F4,$FC,$C7,$F5,$58,$E1,$CB,$B2,$3B,$3F,$B2,$FD,$67,$B5
	.dc.b	$FF,$2A,$37,$3F,$9C,$DD,$BF,$C1,$36,$F9,$8F,$8F,$3F,$2A,$CD,$27
	.dc.b	$17,$CC,$FA,$CF,$E4,$33,$FD,$D1,$AF,$F9,$8D,$B8,$FD,$D7,$FD,$0D
	.dc.b	$1B,$F4,$79,$11,$E5,$E0,$75,$F0,$24,$75,$7C,$1F,$1D,$E1,$F5,$E4
	.dc.b	$F8,$5D,$BE,$1F,$8D,$F3,$11,$F3,$7C,$9F,$A7,$82,$47,$A5,$F6,$7B
	.dc.b	$FF,$88,$FA,$C7,$91,$49,$5F,$E1,$28,$FF,$06,$BF,$E1,$4F,$6F,$E5
	.dc.b	$37,$8F,$F2,$31,$88,$FF,$43,$3F,$6D,$BF,$37,$A2,$3C,$FD,$DE,$3D
	.dc.b	$72,$3B,$3D,$3B,$BF,$3B,$F5,$8D,$A9,$2B,$FC,$4F,$1F,$E0,$E7,$FC
	.dc.b	$53,$F5,$9E,$87,$FD,$30,$DE,$7F,$D9,$37,$83,$FC,$9A,$BC,$EF,$D5
	.dc.b	$FD,$67,$A6,$BF,$70,$DA,$FF,$29,$B6,$FF,$DC,$77,$FA,$99,$CF,$8B
	.dc.b	$C8,$93,$FE,$3C,$AF,$F7,$13,$73,$EE,$47,$5E,$27,$EF,$23,$D9,$1F
	.dc.b	$EB,$16,$E3,$FC,$2D,$B8,$7F,$99,$9F,$F1,$6C,$FF,$92,$9F,$F2,$4B
	.dc.b	$7E,$13,$A1,$D7,$CB,$FE,$13,$F5,$9F,$E9,$69,$FF,$28,$B5,$FF,$60
	.dc.b	$37,$0F,$E3,$4F,$F8,$C2,$37,$E1,$F2,$23,$CB,$CD,$FE,$2B,$F5,$8F
	.dc.b	$88,$A4,$9F,$F2,$FC,$7F,$86,$1B,$E6,$F0,$23,$AF,$77,$F8,$FF,$D6
	.dc.b	$7F,$8B,$A3,$FC,$CE,$D7,$FD,$AC,$DC,$FF,$98,$3F,$B6,$DF,$FD,$16
	.dc.b	$FF,$B9,$E7,$8F,$C2,$F4,$3A,$FB,$9F,$CB,$37,$7C,$35,$25,$7F,$9D
	.dc.b	$BC,$BC,$BF,$E7,$89,$76,$71,$24,$66,$BA,$76,$C8,$E0,$E9,$FF,$77
	.dc.b	$57,$C3,$6F,$AD,$E3,$3F,$9F,$FC,$93,$E8,$FC,$DC,$72,$BC,$4E,$DF
	.dc.b	$27,$5F,$AC,$8F,$FA,$3B,$E6,$5F,$45,$EF,$F3,$E4,$FA,$DD,$BE,$4F
	.dc.b	$0F,$E2,$11,$E5,$DD,$F1,$FD,$39,$1F,$0D,$D3,$EC,$F9,$AF,$98,$47
	.dc.b	$47,$FA,$83,$EA,$FF,$94,$AE,$9F,$ED,$11,$86,$FC,$2F,$F6,$0F,$AC
	.dc.b	$7D,$57,$F6,$2A,$FB,$D6,$0E,$1B,$E9,$EE,$F8,$7A,$72,$7C,$4F,$F6
	.dc.b	$CF,$AC,$F9,$D1,$E5,$F4,$FB,$11,$D9,$E9,$F9,$DF,$11,$1F,$F5,$DF
	.dc.b	$D6,$7F,$07,$7C,$7F,$3F,$E7,$23,$C4,$F8,$5A,$7A,$C8,$FC,$58,$F7
	.dc.b	$78,$32,$3E,$43,$E7,$FA,$0F,$85,$E9,$49,$F9,$8E,$DF,$99,$E2,$FF
	.dc.b	$BB,$38,$1E,$B7,$C9,$24,$FF,$B4,$BE,$B3,$FD,$22,$FD,$5A,$F9,$C4
	.dc.b	$7E,$C4,$FF,$84,$BE,$AD,$F3,$74,$93,$FE,$DC,$FA,$C7,$C3,$EF,$C3
	.dc.b	$93,$FE,$E0,$B4,$70,$8F,$39,$F6,$E3,$C6,$E0,$38,$E9,$FF,$17,$37
	.dc.b	$2F,$C1,$4E,$AB,$CD,$F1,$B2,$3C,$0F,$F8,$AB,$EB,$3F,$73,$1E,$3F
	.dc.b	$17,$D8,$47,$67,$9F,$E7,$FE,$01,$1F,$F7,$CF,$4F,$FA,$11,$F6,$71
	.dc.b	$FC,$52,$3C,$4F,$F7,$FF,$D6,$7F,$89,$1F,$63,$75,$FF,$24,$37,$4F
	.dc.b	$F0,$DF,$99,$FE,$40,$8E,$D5,$C3,$E1,$70,$67,$D5,$FF,$22,$FD,$5F
	.dc.b	$E4,$49,$F3,$BF,$E3,$0F,$3F,$3F,$01,$30,$FD,$46,$9F,$E5,$7B,$FE
	.dc.b	$F8,$DB,$E7,$E7,$CA,$DE,$BF,$3A,$FF,$8C,$5F,$F9,$7F,$D2,$FE,$1B
	.dc.b	$5A,$3F,$69,$3F,$D0,$DC,$07,$0F,$DD,$6C,$F8,$EF,$D7,$EB,$7F,$A0
	.dc.b	$E7,$FB,$5D,$7F,$9D,$4E,$51,$92,$3D,$BE,$B1,$F1,$FF,$E6,$86,$C7
	.dc.b	$8A,$B9,$E1,$F0,$7D,$2F,$92,$49,$F3,$FC,$63,$A7,$11,$E1,$BF,$F4
	.dc.b	$53,$E3,$3D,$95,$E7,$3D,$1E,$8C,$8D,$4F,$F9,$6D,$FF,$47,$35,$F9
	.dc.b	$91,$FE,$76,$F8,$19,$1E,$78,$FB,$BE,$8C,$8F,$53,$FE,$6C,$F9,$9C
	.dc.b	$47,$CB,$8B,$9F,$D5,$9B,$C3,$F6,$66,$D7,$F2,$9E,$15,$62,$3E,$B4
	.dc.b	$D9,$CC,$78,$08,$77,$3F,$B3,$4F,$1C,$BF,$CE,$07,$F9,$C6,$78,$CE
	.dc.b	$C9,$87,$E4,$6F,$D5,$1D,$23,$0D,$C3,$FC,$FB,$87,$5A,$F2,$36,$74
	.dc.b	$7A,$DC,$3D,$A7,$8B,$A3,$DA,$5B,$F2,$DF,$82,$3E,$F5,$FE,$44,$8F
	.dc.b	$DA,$6B,$A3,$04,$3F,$39,$F0,$1F,$B4,$3C,$73,$FA,$F7,$11,$E3,$C7
	.dc.b	$E3,$C6,$23,$D0,$7E,$AA,$B1,$5C,$63,$84,$63,$0F,$DA,$57,$66,$CF
	.dc.b	$E7,$E3,$0F,$ED,$1E,$0F,$53,$C4,$E1,$F8,$A7,$88,$FD,$C3,$9B,$FB
	.dc.b	$B5,$63,$AC,$E8,$F0,$1E,$62,$9F,$10,$F8,$1E,$9F,$F1,$82,$3F,$B5
	.dc.b	$F5,$2B,$FA,$F8,$FE,$04,$47,$F6,$1F,$57,$FE,$43,$7F,$DF,$0D,$7F
	.dc.b	$A1,$B5,$FF,$03,$35,$7D,$67,$F2,$C7,$FE,$00,$35,$FE,$96,$D7,$FC
	.dc.b	$1C,$D5,$F5,$9F,$E4,$47,$F8,$A3,$5F,$EA,$66,$BD,$CF,$E0,$CD,$3D
	.dc.b	$3F,$28,$F9,$AA,$7B,$3D,$EF,$ED,$33,$FC,$7F,$15,$6E,$3F,$91,$7D
	.dc.b	$67,$F9,$4B,$83,$F9,$6E,$4D,$51,$FC,$C2,$75,$5F,$DA,$EB,$53,$9E
	.dc.b	$FD,$CF,$F6,$FA,$DF,$16,$21,$09,$DC,$E6,$37,$5F,$DC,$97,$C1,$6A
	.dc.b	$3F,$BA,$FD,$67,$FA,$F5,$F2,$46,$70,$FF,$1D,$4E,$5F,$17,$76,$5F
	.dc.b	$0C,$DB,$4D,$8C,$72,$6E,$DF,$1D,$8C,$7F,$A8,$C6,$1F,$89,$CB,$9C
	.dc.b	$BE,$03,$73,$F8,$1F,$11,$F5,$6B,$FA,$CF,$9A,$F2,$7C,$93,$9C,$3F
	.dc.b	$C2,$D3,$9A,$FE,$BF,$F5,$8F,$7B,$7C,$FE,$B8,$C4,$FF,$64,$C3,$96
	.dc.b	$5F,$3E,$5F,$21,$F0,$7E,$11,$F0,$7C,$63,$0F,$3F,$81,$C6,$3E,$06
	.dc.b	$EB,$FE,$29,$C3,$E0,$3F,$E2,$BF,$3D,$EE,$78,$F8,$FE,$74,$63,$0F
	.dc.b	$4F,$C0,$65,$A7,$B2,$7D,$5E,$F7,$AE,$6C,$EB,$0F,$07,$C3,$74,$F7
	.dc.b	$FE,$03,$67,$E2,$E5,$86,$9F,$22,$3E,$37,$0F,$E0,$7E,$9D,$F2,$B8
	.dc.b	$AF,$F0,$5C,$61,$F9,$FE,$FE,$72,$F8,$0D,$E4,$FC,$D7,$D5,$BE,$A7
	.dc.b	$F1,$9F,$58,$DB,$F1,$D3,$9C,$7F,$C2,$F3,$97,$FC,$33,$1F,$E6,$06
	.dc.b	$D7,$0F,$82,$E2,$F2,$EB,$C4,$F0,$6E,$7F,$E5,$BB,$7C,$FF,$E7,$7E
	.dc.b	$B3,$F6,$F7,$DA,$30,$F9,$3D,$4F,$2C,$58,$D1,$E4,$F9,$7F,$8D,$F5
	.dc.b	$9E,$A6,$C7,$A3,$E0,$70,$D3,$81,$EC,$E3,$F9,$DA,$FC,$26,$C7,$C0
	.dc.b	$F6,$B0,$D3,$81,$F1,$3E,$4B,$E4,$C7,$CA,$6C,$7C,$8F,$86,$FB,$5C
	.dc.b	$3F,$39,$B1,$F3,$38,$30,$D3,$81,$FE,$48,$FA,$CF,$E1,$B1,$FE,$BF
	.dc.b	$E7,$FF,$44,$CE,$7B,$BF,$0D,$C9,$FC,$1E,$FF,$8B,$4E,$4E,$5E,$6E
	.dc.b	$6B,$5E,$F7,$E3,$7E,$9E,$D7,$E3,$1E,$5F,$25,$8C,$FF,$D2,$58,$A3
	.dc.b	$1D,$7F,$C9,$61,$F4,$DC,$30,$F4,$36,$FF,$D3,$5F,$59,$FE,$B2,$9F
	.dc.b	$F3,$1A,$EF,$FE,$3C,$E6,$77,$1F,$C8,$78,$1F,$4B,$7E,$03,$0D,$C3
	.dc.b	$FC,$DB,$F5,$8F,$85,$E1,$8F,$F2,$98,$C4,$FF,$9C,$B0,$FF,$74,$3C
	.dc.b	$38,$75,$EE,$6D,$B1,$D3,$D2,$E5,$EA,$73,$6C,$7F,$9F,$3E,$95,$F2
	.dc.b	$9F,$CD,$D7,$9F,$63,$0E,$7C,$8D,$79,$76,$F9,$3F,$CF,$FE,$B3,$F4
	.dc.b	$B1,$FE,$88,$5F,$51,$9F,$1F,$4E,$FE,$2F,$87,$85,$6F,$83,$CF,$8F
	.dc.b	$A7,$15,$9E,$8F,$F4,$D7,$FA,$59,$FE,$96,$3F,$AC,$97,$DB,$D7,$B7
	.dc.b	$FD,$4F,$EB,$1F,$47,$7E,$67,$0D,$BF,$F4,$FF,$D6,$7A,$1B,$E1,$FF
	.dc.b	$5B,$8C,$67,$FE,$DD,$51,$9D,$35,$E4,$DB,$E6,$FF,$63,$FA,$C7,$8D
	.dc.b	$7E,$2F,$0D,$AF,$F5,$7C,$FA,$B2,$FF,$75,$3F,$96,$8C,$ED,$CF,$C4
	.dc.b	$FE,$DD,$EF,$37,$BE,$FE,$1B,$7F,$EB,$A9,$7B,$1C,$3E,$B3,$8B,$67
	.dc.b	$FD,$7F,$F5,$9E,$C6,$F9,$DF,$DE,$E3,$1A,$FF,$BF,$5F,$DD,$B7,$AB
	.dc.b	$5D,$5B,$6D,$F4,$F5,$B0,$D3,$BC,$75,$F2,$3B,$79,$36,$7C,$6F,$87
	.dc.b	$D5,$C6,$3E,$16,$F9,$DF,$E1,$FC,$DF,$27,$FE,$11,$E6,$F8,$98,$FB
	.dc.b	$23,$1C,$5E,$BE,$FF,$64,$59,$D1,$FE,$E0,$C1,$E4,$73,$F1,$7F,$C3
	.dc.b	$91,$89,$FF,$72,$E3,$FA,$87,$87,$91,$AB,$DB,$63,$FD,$D7,$AF,$FB
	.dc.b	$79,$E3,$F1,$1F,$E2,$78,$C0,$C3,$FC,$75,$87,$F9,$2A,$B8,$F7,$6F
	.dc.b	$BB,$83,$63,$AF,$89,$86,$9C,$0F,$0E,$BF,$17,$E2,$BC,$0D,$8F,$2E
	.dc.b	$F8,$69,$C0,$F4,$E6,$F2,$BF,$CC,$AD,$3F,$CC,$EE,$0F,$F7,$77,$D5
	.dc.b	$7F,$90,$3E,$B2,$F9,$06,$92,$BF,$CE,$22,$5F,$25,$F0,$E3,$31,$8E
	.dc.b	$0E,$B3,$99,$FF,$3C,$CE,$5D,$FC,$91,$95,$FE,$65,$94,$E6,$7E,$86
	.dc.b	$56,$73,$1C,$23,$17,$E3,$FA,$CF,$F0,$EC,$F5,$AF,$C3,$BB,$31,$CB
	.dc.b	$D7,$1F,$03,$70,$ED,$8D,$AE,$5C,$B5,$F8,$4D,$C0,$E7,$E5,$3F,$D2
	.dc.b	$F8,$FA,$FE,$3F,$44,$BF,$11,$C8,$B6,$F9,$2E,$84,$8C,$CF,$87,$DE
	.dc.b	$72,$D2,$4F,$4F,$94,$FE,$B0,$F6,$9F,$D8,$B7,$97,$91,$C1,$F9,$A7
	.dc.b	$D3,$FA,$CF,$9A,$E0,$38,$CF,$FD,$34,$DD,$FE,$B6,$0F,$97,$F8,$D3
	.dc.b	$F4,$2F,$93,$6F,$8F,$0E,$B8,$7E,$67,$D6,$7F,$27,$BF,$8A,$D7,$C4
	.dc.b	$DA,$FF,$55,$3F,$E8,$38,$DF,$A5,$C8,$8F,$2F,$03,$AF,$81,$23,$AB
	.dc.b	$E0,$F8,$FF,$0F,$B1,$27,$C2,$ED,$F0,$FC,$6F,$98,$8F,$9B,$EA,$FD
	.dc.b	$3C,$32,$3D,$2F,$B3,$E3,$3F,$11,$F5,$8F,$22,$92,$BF,$07,$1F,$E1
	.dc.b	$47,$F0,$BE,$E7,$CB,$6F,$27,$E4,$63,$11,$F5,$5F,$BE,$DF,$85,$D1
	.dc.b	$1E,$7E,$F7,$1E,$B9,$1D,$9E,$9D,$FF,$9D,$FA,$C6,$D4,$95,$F9,$88
	.dc.b	$FF,$0A,$BF,$9A,$FA,$CF,$43,$F8,$76,$F3,$FF,$B7,$DB,$C1,$FB,$B5
	.dc.b	$E7,$7E,$AF,$EB,$3D,$35,$FE,$E5,$6D,$7F,$01,$6D,$FF,$07,$BF,$D6
	.dc.b	$CE,$7C,$5E,$44,$9F,$D6,$D7,$FB,$71,$B9,$F7,$23,$AE,$CF,$DE,$47
	.dc.b	$B2,$3F,$40,$DC,$7F,$DF,$0D,$C3,$F8,$A3,$FE,$34,$9F,$DD,$BF,$E5
	.dc.b	$16,$FC,$2F,$43,$AF,$97,$FC,$27,$EB,$3F,$D2,$F3,$FC,$09,$AF,$EB
	.dc.b	$DB,$83,$0F,$E4,$51,$FE,$38,$9E,$1E,$27,$32,$3C,$FC,$DF,$E2,$DF
	.dc.b	$59,$FE,$6D,$8F,$E1,$ED,$7F,$78,$DC,$3B,$47,$F3,$38,$FF,$11,$BC
	.dc.b	$3C,$CE,$64,$79,$FB,$BF,$C8,$FE,$B3,$FC,$73,$1F,$C7,$1A,$FF,$09
	.dc.b	$6D,$F6,$84,$7E,$EF,$3F,$E7,$CF,$FB,$AA,$39,$7C,$2F,$53,$B7,$B7
	.dc.b	$FC,$C7,$67,$C3,$52,$57,$F2,$17,$5D,$52,$3F,$97,$CA,$37,$1F,$C7
	.dc.b	$61,$D7,$D4,$91,$C2,$7C,$1E,$AC,$8E,$6F,$F4,$3F,$71,$F4,$7E,$5F
	.dc.b	$8C,$95,$E0,$78,$3D,$FA,$BD,$44,$7F,$A1,$FC,$CB,$EB,$3D,$DE,$74
	.dc.b	$9F,$4B,$B7,$D1,$E1,$7C,$32,$3C,$BB,$7E,$37,$A7,$23,$E0,$BA,$7E
	.dc.b	$07,$CD,$7C,$B2,$3A,$3F,$AA,$FD,$4F,$F2,$D7,$FA,$13,$7A,$3E,$D3
	.dc.b	$7A,$3F,$BB,$7D,$63,$EA,$FF,$B3,$4F,$DD,$BF,$EF,$CD,$4D,$39,$3A
	.dc.b	$F4,$38,$91,$FE,$CF,$F4,$FF,$39,$F2,$FA,$1D,$88,$FA,$27,$C3,$F3
	.dc.b	$92,$3D,$6F,$F7,$9F,$AC,$FE,$17,$1E,$2F,$9E,$F3,$91,$D9,$F0,$B2
	.dc.b	$F5,$91,$F8,$B1,$EE,$EF,$91,$F2,$1F,$F0,$36,$4F,$A0,$F8,$7E,$7C
	.dc.b	$AF,$9A,$F0,$7E,$6F,$A7,$F2,$08,$FF,$81,$BE,$B3,$FD,$36,$FD,$4E
	.dc.b	$1F,$9A,$47,$EB,$CF,$F8,$57,$EA,$DF,$37,$49,$3F,$E1,$0F,$AC,$7C
	.dc.b	$36,$FC,$19,$3F,$E1,$4B,$47,$08,$FF,$01,$47,$89,$BF,$29,$DC,$E2
	.dc.b	$EB,$E3,$7C,$44,$7F,$C3,$FF,$59,$FB,$17,$B7,$89,$E7,$23,$E1,$9F
	.dc.b	$57,$CF,$C8,$F2,$3F,$33,$E1,$3D,$DC,$52,$7D,$0E,$DF,$F0,$C7,$D6
	.dc.b	$3D,$2F,$F1,$A7,$C8,$49,$FF,$18,$CE,$5F,$F1,$9F,$C4,$FF,$30,$E4
	.dc.b	$FF,$28,$36,$3F,$CA,$CD,$53,$C5,$EF,$EA,$FF,$91,$63,$FC,$6B,$1D
	.dc.b	$A3,$FC,$91,$F0,$FF,$06,$36,$78,$B3,$E0,$71,$9F,$F2,$77,$D5,$BE
	.dc.b	$63,$FC,$A1,$F5,$9F,$E5,$CB,$FF,$1A,$37,$C6,$FE,$59,$5B,$9F,$F2
	.dc.b	$6C,$ED,$FF,$35,$07,$0F,$F9,$B5,$AF,$F9,$CD,$AA,$3F,$A3,$69,$FE
	.dc.b	$73,$6C,$7F,$9E,$5A,$EF,$D0,$F4,$FA,$12,$3F,$CB,$F3,$A3,$FA,$6B
	.dc.b	$FE,$4D,$71,$1B,$8C,$4D,$C3,$DB,$36,$FC,$53,$6B,$C5,$35,$F1,$1E
	.dc.b	$07,$17,$E1,$9B,$18,$6C,$EC,$F3,$B8,$7C,$D1,$F5,$23,$17,$E4,$9D
	.dc.b	$9C,$4D,$95,$FA,$B1,$C2,$FE,$55,$73,$AD,$46,$1A,$7C,$D9,$E0,$FB
	.dc.b	$8E,$E7,$62,$57,$8D,$F9,$27,$23,$ED,$57,$BD,$C5,$FD,$07,$9C,$F1
	.dc.b	$25,$F8,$E3,$99,$DC,$FC,$5C,$FF,$82,$DB,$5F,$EC,$67,$CC,$7F,$B2
	.dc.b	$23,$4F,$33,$66,$DC,$BC,$2B,$FC,$0D,$F5,$5F,$8C,$3C,$D1,$EE,$BF
	.dc.b	$F6,$24,$66,$FE,$2C,$A3,$0F,$BA,$3F,$D7,$11,$88,$E4,$77,$9E,$73
	.dc.b	$A3,$9C,$EE,$BE,$1D,$6C,$FF,$6C,$46,$6B,$E3,$4E,$A7,$C5,$5C,$DF
	.dc.b	$11,$FE,$C0,$AE,$0F,$33,$EA,$99,$8E,$0F,$98,$C9,$FE,$B3,$8E,$2E
	.dc.b	$27,$C1,$18,$BF,$8F,$1F,$2A,$B9,$BE,$58,$E3,$3A,$BF,$49,$F0,$AF
	.dc.b	$C4,$7A,$0F,$05,$F2,$8D,$C7,$A6,$7C,$13,$FB,$16,$C7,$FB,$7B,$EA
	.dc.b	$DF,$26,$F9,$8F,$C3,$46,$1B,$EB,$FF,$54,$AF,$E2,$E2,$74,$E2,$DF
	.dc.b	$5F,$FE,$4B,$8F,$C1,$B5,$FA,$AD,$7F,$08,$DF,$AC,$7C,$E7,$F9,$83
	.dc.b	$6F,$F2,$A4,$63,$5F,$F7,$AF,$D6,$3C,$CD,$43,$19,$5B,$7F,$DC,$D5
	.dc.b	$B9,$E3,$5C,$27,$84,$BE,$43,$10,$8E,$0F,$2F,$8A,$CC,$7F,$AC,$AF
	.dc.b	$C1,$8F,$9B,$E0,$73,$5C,$B2,$45,$31,$F0,$56,$F8,$9E,$9C,$5F,$0B
	.dc.b	$F4,$E3,$31,$ED,$F1,$FF,$C3,$11,$88,$DF,$1F,$A2,$73,$0C,$FF,$CF
	.dc.b	$9F,$58,$F9,$EF,$6E,$3D,$6B,$58,$F8,$6B,$4F,$59,$D5,$F2,$F9,$1F
	.dc.b	$8B,$F0,$0F,$6F,$E3,$E0,$9C,$C6,$3D,$8C,$C5,$C7,$D1,$5B,$6F,$C8
	.dc.b	$FD,$0F,$83,$7A,$FD,$23,$0F,$6F,$6B,$15,$C9,$BB,$7A,$8F,$C1,$60
	.dc.b	$F6,$78,$57,$81,$B3,$B1,$2F,$86,$F1,$7B,$3E,$C6,$DF,$A3,$1C,$34
	.dc.b	$F5,$C7,$C2,$E1,$FC,$4F,$D3,$BE,$5B,$15,$F5,$E3,$0F,$C2,$F6,$B3
	.dc.b	$9F,$8F,$34,$63,$93,$FD,$69,$F5,$6F,$A9,$FC,$AF,$D6,$37,$FC,$BC
	.dc.b	$E7,$2F,$C2,$4E,$5F,$C3,$47,$F9,$41,$B5,$C3,$DD,$E2,$F2,$EA,$C4
	.dc.b	$F0,$6E,$9F,$1E,$DE,$9F,$E9,$BE,$B3,$FD,$C5,$7D,$A3,$0F,$93,$D8
	.dc.b	$F2,$C5,$8D,$1E,$4F,$97,$F8,$FF,$59,$EA,$6C,$7A,$3D,$AC,$34,$E0
	.dc.b	$7B,$39,$3E,$7E,$BF,$09,$B1,$F0,$3E,$83,$0D,$38,$1F,$13,$DA,$F9
	.dc.b	$31,$F2,$9B,$1F,$23,$E0,$3E,$DF,$0F,$CE,$6C,$7C,$CE,$0C,$34,$E0
	.dc.b	$7E,$D7,$EB,$3F,$DB,$8F,$E0,$78,$BF,$86,$C6,$27,$F2,$5D,$1F,$E7
	.dc.b	$CA,$E1,$F1,$FA,$F0,$36,$D8,$E9,$EB,$7E,$3B,$E0,$79,$B6,$3C,$1E
	.dc.b	$B6,$1A,$70,$3C,$9E,$8F,$8A,$3C,$6D,$8F,$E1,$9F,$59,$F9,$26,$F8
	.dc.b	$1F,$C9,$23,$1A,$7E,$A5,$FE,$2F,$3F,$C4,$F7,$7D,$27,$F2,$A5,$F3
	.dc.b	$36,$30,$FF,$2E,$FA,$B7,$D6,$6F,$1C,$37,$0F,$E3,$D3,$F1,$FC,$24
	.dc.b	$7F,$27,$87,$CA,$F3,$E2,$E8,$F5,$F9,$9E,$DF,$25,$6F,$9B,$FC,$FB
	.dc.b	$E9,$5F,$21,$7D,$0C,$36,$BF,$97,$4F,$EA,$F5,$FE,$10,$FF,$16,$8C
	.dc.b	$F9,$1C,$FC,$CF,$E9,$3F,$59,$FA,$69,$FE,$6C,$BE,$A3,$3E,$3E,$9D
	.dc.b	$FC,$37,$0F,$0A,$DF,$07,$9E,$7D,$39,$EC,$F4,$7F,$AB,$AF,$62,$FB
	.dc.b	$38,$6C,$7F,$47,$FA,$CF,$EB,$CD,$FA,$3F,$EB,$F1,$8C,$3F,$92,$28
	.dc.b	$FE,$C7,$F5,$9E,$9A,$FE,$A0,$BC,$2C,$EB,$A7,$7A,$70,$F2,$2D,$F0
	.dc.b	$7F,$B6,$FD,$63,$C4,$BF,$15,$86,$D7,$F5,$D9,$FE,$D2,$DF,$13,$8F
	.dc.b	$07,$27,$FB,$BF,$BD,$E7,$9F,$EC,$8F,$F5,$AE,$0F,$E7,$CE,$5B,$EF
	.dc.b	$70,$F6,$38,$FD,$65,$BE,$2F,$F8,$0B,$EB,$3D,$D1,$FD,$BD,$FD,$B5
	.dc.b	$FF,$49,$72,$DF,$3B,$87,$4F,$1E,$3B,$7C,$5E,$9E,$AE,$2F,$93,$7C
	.dc.b	$6F,$CE,$C7,$F8,$6F,$1F,$D5,$DF,$87,$C8,$F2,$79,$3E,$57,$CD,$EA
	.dc.b	$62,$7C,$CD,$BF,$37,$C6,$FF,$87,$39,$7E,$16,$5D,$A5,$92,$B5,$D7
	.dc.b	$F0,$7B,$BE,$14,$5C,$BE,$13,$FE,$26,$C1,$E3,$75,$F1,$B0,$DA,$FF
	.dc.b	$0B,$4B,$C7,$CC,$61,$B9,$FF,$73,$1D,$9E,$7F,$8B,$FF,$19,$C6,$2B
	.dc.b	$FC,$41,$B7,$F9,$61,$FF,$11,$43,$BB,$77,$C1,$B3,$AF,$0F,$0D,$38
	.dc.b	$C7,$6D,$7E,$37,$C5,$D9,$EE,$F8,$B7,$C6,$3E,$76,$EB,$FE,$5D,$F2
	.dc.b	$3F,$D5,$D3,$FE,$53,$6A,$69,$FE,$B3,$FA,$AF,$DC,$7D,$65,$F1,$CD
	.dc.b	$25,$7F,$9C,$25,$F4,$3F,$E3,$D8,$C3,$5F,$F2,$50,$8D,$E1,$DE,$75
	.dc.b	$3F,$E4,$F9,$7D,$07,$F9,$46,$BF,$09,$18,$9F,$F2,$73,$5F,$F1,$43
	.dc.b	$D1,$7D,$67,$F8,$7D,$FF,$39,$47,$E1,$1E,$3C,$BD,$76,$F0,$36,$37
	.dc.b	$85,$7D,$1C,$B6,$FF,$11,$1E,$28,$D7,$CA,$FB,$B2,$F3,$FD,$0E,$C9
	.dc.b	$5B,$3D,$3F,$43,$CC,$8F,$B2,$3C,$1E,$EB,$9A,$92,$7F,$CD,$DF,$2D
	.dc.b	$FD,$59,$FF,$38,$4F,$E9,$DB,$CD,$C8,$E0,$FF,$93,$7E,$B3,$E4,$65
	.dc.b	$FA,$F6,$F0,$7F,$9F,$DB,$3F,$E9,$9C,$1F,$55,$F0,$E7,$E5,$4F,$3E
	.dc.b	$1C,$19,$E1,$F9,$1F,$59,$FC,$DE,$7F,$D0,$8D,$7F,$CB,$CD,$CB,$F8
	.dc.b	$03,$FE,$83,$8D,$FC,$9F,$22,$3C,$BC,$0E,$BE,$04,$8E,$AF,$83,$E3
	.dc.b	$BC,$3E,$C4,$9F,$0B,$B7,$C3,$F1,$DE,$62,$3E,$6F,$92,$F4,$F0,$C8
	.dc.b	$F4,$BE,$CF,$8C,$FC,$27,$D6,$3C,$8A,$4A,$FF,$51,$47,$F8,$41,$FF
	.dc.b	$52,$7B,$9F,$29,$BC,$9F,$8D,$8C,$47,$FA,$05,$FB,$8D,$F8,$4D,$11
	.dc.b	$E7,$EF,$F1,$EB,$91,$D9,$E9,$DD,$F9,$9F,$AC,$6D,$49,$5F,$EB,$78
	.dc.b	$FF,$07,$BF,$EB,$9F,$AC,$F4,$3F,$E9,$36,$F3,$FF,$35,$6F,$07,$FB
	.dc.b	$15,$FF,$15,$CF,$FB,$07,$EB,$3D,$35,$FC,$ED,$B5,$FE,$CB,$6C,$FF
	.dc.b	$B9,$6B,$F5,$13,$9F,$13,$F6,$15,$FC,$32,$BA,$46,$3C,$6C,$E7,$27
	.dc.b	$C0,$F0,$EC,$FD,$C3,$EC,$8F,$F5,$63,$72,$FE,$9C,$DC,$7F,$DB,$CB
	.dc.b	$A3,$8F,$C3,$4B,$FF,$6E,$4F,$FB,$89,$BE,$57,$F0,$5F,$AC,$FF,$48
	.dc.b	$C7,$FB,$85,$AF,$FA,$ED,$B8,$FF,$5D,$7F,$C6,$11,$BF,$13,$91,$1E
	.dc.b	$5E,$67,$F1,$1F,$AC,$7C,$4D,$24,$FF,$BB,$E3,$FC,$30,$DD,$F8,$11
	.dc.b	$D7,$BD,$FC,$6F,$EB,$1E,$1D,$24,$FF,$BE,$63,$F6,$9C,$3F,$F0,$05
	.dc.b	$FF,$70,$C6,$7E,$1B,$91,$CF,$DD,$FE,$4D,$C1,$FE,$60,$91,$FE,$DB
	.dc.b	$70,$DF,$87,$FE,$59,$E5,$63,$8B,$FC,$01,$4E,$DE,$64,$BE,$73,$AB
	.dc.b	$EF,$DD,$27,$3A,$78,$16,$09,$E5,$1E,$3F,$57,$CE,$47,$89,$E8,$F7
	.dc.b	$5F,$47,$EC,$E4,$93,$A3,$AB,$C5,$F7,$76,$7C,$22,$3B,$F9,$97,$D6
	.dc.b	$7C,$6F,$3A,$4F,$3D,$3F,$97,$BE,$8B,$F9,$7E,$24,$8D,$3F,$D4,$F0
	.dc.b	$FE,$63,$C1,$F2,$3D,$2F,$A2,$4A,$D9,$D7,$E6,$BE,$A9,$1F,$04,$7F
	.dc.b	$5D,$FA,$BF,$E6,$10,$AF,$F1,$94,$61,$BF,$09,$FD,$8F,$EB,$1F,$59
	.dc.b	$C7,$D3,$98,$7C,$5C,$3F,$12,$FE,$09,$E5,$CD,$F5,$5F,$D7,$46,$FC
	.dc.b	$6F,$11,$FD,$BF,$EB,$3E,$8C,$73,$F4,$FC,$C4,$78,$1D,$FE,$7B,$D4
	.dc.b	$47,$EB,$7D,$0F,$F0,$E5,$83,$E7,$FE,$01,$23,$91,$E9,$D3,$E2,$11
	.dc.b	$F6,$47,$C7,$F0,$64,$7B,$DF,$47,$D4,$7C,$BF,$4A,$4F,$C3,$76,$FC
	.dc.b	$DD,$BE,$17,$AD,$F4,$49,$1F,$8D,$FA,$CF,$F4,$D3,$F2,$E7,$EB,$11
	.dc.b	$F9,$D3,$FE,$15,$FA,$B7,$CD,$D2,$4F,$E5,$FE,$B1,$F0,$DB,$F0,$E4
	.dc.b	$FE,$6E,$D1,$C2,$3C,$C7,$EE,$A3,$C6,$E0,$38,$E9,$FF,$3C,$37,$2F
	.dc.b	$D9,$3D,$27,$B7,$C6,$79,$C8,$FE,$AF,$EB,$3F,$80,$3E,$3F,$17,$D8
	.dc.b	$47,$89,$E7,$F9,$FF,$80,$47,$F6,$9D,$3F,$E2,$17,$CD,$83,$8B,$E3
	.dc.b	$12,$39,$1F,$BB,$FA,$CF,$F3,$B3,$FE,$56,$3F,$8C,$B8,$70,$D7,$FC
	.dc.b	$B6,$D7,$F8,$83,$BD,$3F,$D3,$5F,$0D,$FD,$03,$5F,$E2,$8D,$AE,$BE
	.dc.b	$A7,$F9,$36,$FF,$C7,$D1,$E3,$F5,$70,$7F,$26,$F3,$AC,$79,$44,$F0
	.dc.b	$7C,$B3,$FE,$7E,$8C,$47,$F9,$C3,$EA,$DF,$2F,$FE,$82,$6C,$73,$6B
	.dc.b	$F0,$5A,$F1,$8F,$F3,$17,$D6,$7F,$9A,$E3,$F0,$4D,$D3,$FD,$71,$5C
	.dc.b	$9B,$E7,$FF,$9A,$63,$FA,$AD,$7E,$25,$B3,$FE,$BF,$F4,$BF,$92,$CF
	.dc.b	$F2,$D9,$CE,$DF,$CE,$DB,$1F,$D3,$1A,$A7,$8B,$D6,$FC,$04,$7F,$CC
	.dc.b	$FE,$7D,$74,$77,$8F,$EA,$71,$9C,$3D,$33,$93,$89,$4E,$24,$7D,$51
	.dc.b	$C7,$0F,$45,$5A,$FF,$35,$4D,$58,$7A,$8F,$5E,$1E,$AC,$3C,$33,$D7
	.dc.b	$0E,$6F,$88,$C1,$F0,$F0,$E9,$5C,$67,$73,$A3,$BC,$FB,$9F,$55,$76
	.dc.b	$24,$FF,$69,$AE,$53,$D4,$DC,$74,$9F,$7C,$60,$D1,$DE,$78,$BE,$03
	.dc.b	$45,$3F,$85,$3A,$BF,$85,$7E,$AC,$3E,$34,$62,$BE,$05,$F9,$9E,$ED
	.dc.b	$7F,$05,$3F,$75,$FD,$05,$7E,$7C,$8F,$E7,$DA,$FE,$7D,$AF,$03,$F2
	.dc.b	$5E,$87,$E6,$4F,$1C,$58,$C9,$EC,$BF,$B5,$39,$8F,$A1,$18,$76,$62
	.dc.b	$75,$7F,$A8,$AE,$B3,$A8,$C3,$D6,$77,$8F,$E8,$EB,$73,$FA,$B7,$2F
	.dc.b	$C4,$8F,$AC,$6D,$E8,$E1,$D1,$FA,$B8,$E2,$FD,$5A,$F9,$97,$FE,$50
	.dc.b	$8E,$27,$03,$55,$D5,$F1,$1E,$73,$67,$E7,$AB,$CF,$5E,$C8,$D9,$F0
	.dc.b	$CD,$D7,$77,$06,$A3,$81,$E4,$BF,$3C,$EE,$31,$3C,$AB,$F5,$DF,$57
	.dc.b	$FE,$F6,$8F,$E7,$8D,$7F,$C8,$C1,$C3,$FE,$49,$6B,$FE,$5A,$FA,$C7
	.dc.b	$C8,$BE,$63,$F9,$94,$61,$BE,$9F,$F6,$EA,$FE,$95,$39,$7F,$A6,$7D
	.dc.b	$53,$E6,$9B,$7F,$3B,$8C,$36,$33,$3C,$1A,$FA,$3E,$AD,$E5,$E2,$3F
	.dc.b	$CC,$11,$8A,$FE,$8C,$D8,$FE,$B8,$D7,$97,$C3,$F1,$7F,$CE,$F1,$FE
	.dc.b	$C6,$9E,$33,$99,$4E,$5F,$5F,$BD,$9C,$4D,$25,$FF,$57,$F8,$3F,$90
	.dc.b	$FF,$9A,$7C,$4F,$EE,$8F,$F9,$AE,$33,$7F,$E0,$D9,$CB,$FE,$0F,$FA
	.dc.b	$B7,$AB,$F1,$39,$3E,$1D,$F0,$7E,$2C,$EB,$0E,$9C,$3B,$7F,$D0,$90
	.dc.b	$F8,$DC,$63,$FE,$96,$F6,$5E,$CE,$21,$5B,$BF,$F3,$D4,$E6,$3F,$D8
	.dc.b	$D1,$FE,$83,$75,$3E,$7F,$5B,$58,$6F,$E3,$7A,$78,$4F,$4B,$18,$FF
	.dc.b	$A2,$E3,$0F,$3F,$5B,$38,$FC,$16,$ED,$EE,$F5,$3C,$51,$E3,$6D,$78
	.dc.b	$FE,$03,$BF,$AB,$E7,$6C,$79,$F2,$C3,$4E,$07,$AF,$8B,$E1,$7D,$2B
	.dc.b	$E4,$BE,$04,$61,$F8,$5E,$B6,$2B,$E1,$37,$5F,$F2,$37,$D5,$3E,$A3
	.dc.b	$E6,$7D,$5B,$5F,$9D,$39,$C3,$FD,$59,$39,$7F,$D5,$D1,$FE,$5C,$6D
	.dc.b	$70,$F6,$78,$BC,$BA,$F1,$3C,$1B,$8F,$8B,$E7,$FE,$E7,$D5,$FE,$C2
	.dc.b	$BA,$C6,$1F,$1F,$B3,$E4,$8B,$1A,$3C,$7F,$29,$F1,$3E,$AF,$D2,$D8
	.dc.b	$F3,$FB,$38,$69,$C0,$F5,$F0,$FC,$CD,$3E,$0B,$63,$DF,$F3,$B8,$69
	.dc.b	$C0,$F8,$7E,$B7,$C8,$8F,$92,$D8,$F8,$FE,$EF,$D8,$E0,$F9,$AD,$8F
	.dc.b	$97,$BE,$1A,$70,$3F,$DB,$9F,$57,$FE,$F1,$7F,$DB,$DC,$FF,$E8,$47
	.dc.b	$3D,$BF,$84,$E3,$FD,$35,$7F,$BB,$67,$27,$2F,$8A,$CD,$6B,$D3,$FC
	.dc.b	$5F,$D2,$DA,$EA,$F2,$F4,$F1,$9F,$FA,$37,$14,$63,$C1,$FE,$43,$1F
	.dc.b	$C9,$A7,$53,$E2,$F3,$F9,$61,$FE,$F5,$FA,$B7,$77,$EE,$D7,$7D,$8C
	.dc.b	$3F,$BF,$E1,$7D,$1D,$FA,$3C,$37,$0F,$F8,$01,$F5,$4F,$AB,$E1,$AF
	.dc.b	$F0,$78,$C4,$A3,$FD,$4D,$85,$8E,$3C,$5B,$F9,$1E,$0D,$8E,$BF,$2F
	.dc.b	$CF,$E6,$3A,$36,$3A,$71,$FF,$42,$FA,$47,$C7,$7F,$17,$7F,$89,$E7
	.dc.b	$FE,$B5,$73,$5C,$0F,$24,$23,$75,$CF,$C8,$E5,$E4,$F3,$6D,$F1,$BF
	.dc.b	$67,$D5,$BC,$6F,$E4,$CB,$E9,$EC,$61,$CF,$C3,$6B,$C2,$B8,$73,$C1
	.dc.b	$82,$79,$C7,$4E,$2B,$5B,$AF,$F6,$CB,$C4,$7C,$CF,$F3,$77,$FA,$89
	.dc.b	$97,$1E,$CF,$F3,$BF,$AB,$7C,$D5,$F9,$4C,$36,$FC,$9F,$57,$EB,$6F
	.dc.b	$87,$FD,$0E,$31,$A7,$FB,$85,$46,$75,$D7,$93,$6F,$9B,$FD,$2F,$EA
	.dc.b	$DE,$35,$F8,$BC,$36,$BE,$24,$FA,$B2,$FF,$77,$3F,$9F,$E7,$FF,$00
	.dc.b	$B9,$6F,$8B,$FD,$5F,$DE,$6F,$7D,$FC,$37,$1F,$A1,$2F,$5F,$87,$D5
	.dc.b	$F1,$6C,$FD,$5F,$AB,$F6,$B7,$CE,$FE,$C9,$18,$D5,$A7,$F8,$52,$3F
	.dc.b	$83,$46,$2B,$F1,$F1,$F8,$77,$A7,$5F,$2E,$CE,$6F,$4F,$5B,$13,$E1
	.dc.b	$6D,$F3,$3F,$59,$7F,$B6,$9F,$37,$93,$E4,$F2,$7C,$AF,$9B,$D5,$C4
	.dc.b	$FA,$DB,$7C,$C5,$BE,$A3,$FC,$01,$CB,$F2,$AF,$E1,$46,$38,$BE,$0F
	.dc.b	$81,$F0,$A2,$CE,$8F,$CE,$6E,$F1,$F9,$FB,$FF,$E0,$B8,$C4,$FE,$87
	.dc.b	$1F,$D8,$BF,$B4,$87,$91,$BB,$E0,$D9,$FD,$46,$8E,$BF,$3F,$C4,$FF
	.dc.b	$85,$E3,$13,$D6,$31,$DB,$9A,$B3,$C9,$E1,$E2,$E2,$FE,$20,$71,$97
	.dc.b	$F9,$37,$5F,$8B,$F1,$3D,$DB,$87,$93,$7C,$34,$F3,$C7,$97,$0F,$E2
	.dc.b	$07,$F9,$C0,$FE,$27,$E4,$D6,$94,$D9,$FE,$76,$FA,$9F,$DF,$FD,$5E
	.dc.b	$DB,$1C,$DA,$EF,$C8,$4E,$FE,$7F,$F2,$F3,$B6,$BD,$5A,$FF,$26,$A4
	.dc.b	$62,$3F,$9E,$43,$E8,$3E,$C6,$5B,$6C,$E9,$AA,$79,$EB,$FD,$17,$EA
	.dc.b	$FF,$C5,$93,$D3,$93,$75,$F8,$98,$FE,$4C,$D7,$F3,$0F,$85,$CF,$27
	.dc.b	$02,$3F,$CB,$7E,$57,$FD,$3B,$87,$FA,$22,$9F,$43,$D8,$95,$C8,$EF
	.dc.b	$F4,$3E,$22,$3E,$28,$F5,$FB,$FF,$85,$8F,$E7,$4D,$7F,$58,$DB,$FF
	.dc.b	$4B,$FC,$AF,$F6,$79,$DF,$1F,$22,$3C,$B9,$BE,$3F,$D5,$FB,$63,$F6
	.dc.b	$CD,$BF,$F5,$0B,$78,$7F,$AD,$E0,$FA,$9E,$3C,$13,$0F,$E9,$D7,$F2
	.dc.b	$70,$E3,$C5,$F5,$3E,$AF,$F9,$3D,$7F,$53,$6B,$FC,$15,$B3,$FE,$B0
	.dc.b	$7F,$D2,$31,$BF,$47,$91,$1E,$5E,$07,$5F,$02,$47,$57,$C1,$F1,$BE
	.dc.b	$1F,$62,$4F,$85,$DB,$E1,$F8,$DF,$31,$1F,$37,$C9,$FA,$78,$64,$7A
	.dc.b	$5F,$67,$C0,$7E,$43,$EA,$DE,$45,$25,$7F,$71,$8F,$F0,$FB,$FD,$CF
	.dc.b	$DB,$F9,$4D,$E4,$FC,$CC,$62,$3F,$94,$3F,$6D,$BF,$0B,$A2,$3C,$FD
	.dc.b	$EE,$3D,$72,$3B,$3D,$3B,$FF,$4B,$F5,$6D,$A9,$2B,$FC,$0D,$1F,$E1
	.dc.b	$E7,$FC,$11,$F5,$7E,$87,$F9,$D3,$79,$FF,$DC,$AD,$E0,$FF,$0F,$AF
	.dc.b	$3B,$F6,$7F,$57,$E9,$AF,$F7,$5B,$6B,$FC,$4A,$DB,$FF,$1D,$5F,$EA
	.dc.b	$A7,$3E,$2F,$22,$4F,$F8,$6E,$BF,$C2,$6D,$CF,$B9,$1D,$76,$FF,$04
	.dc.b	$8F,$64,$7F,$53,$6E,$3F,$EF,$E6,$E1,$FE,$38,$5D,$1C,$7E,$16,$5F
	.dc.b	$F8,$AE,$7F,$CB,$AD,$F2,$FF,$88,$FD,$5F,$F8,$DE,$3F,$C6,$0D,$7F
	.dc.b	$B1,$37,$16,$3F,$E4,$68,$FF,$1E,$4F,$0F,$0B,$99,$1E,$7E,$77,$F1
	.dc.b	$EF,$AB,$FF,$3B,$C7,$F8,$F5,$AF,$F6,$F6,$E1,$DA,$3F,$CC,$D1,$FE
	.dc.b	$24,$78,$79,$9C,$C8,$F3,$F8,$2F,$E5,$7F,$57,$FE,$3A,$8F,$F2,$53
	.dc.b	$5F,$EF,$8D,$BE,$D0,$8F,$DD,$E7,$FE,$62,$7F,$DE,$51,$CB,$E1,$7A
	.dc.b	$9D,$BD,$BF,$E7,$3B,$3E,$1A,$92,$BF,$CB,$5E,$56,$5F,$F2,$E4,$BB
	.dc.b	$38,$92,$33,$5D,$3B,$64,$70,$74,$FF,$85,$6B,$E2,$37,$D6,$F1,$9F
	.dc.b	$D3,$3D,$C7,$D2,$F9,$BE,$2E,$57,$89,$DB,$E4,$EC,$F5,$91,$FF,$39
	.dc.b	$7C,$CB,$E8,$3D,$FE,$7C,$9F,$5B,$B7,$C9,$E2,$7C,$42,$3C,$BB,$FE
	.dc.b	$3F,$A7,$23,$E1,$BA,$7D,$9F,$35,$F3,$08,$E8,$FF,$42,$7D,$57,$F2
	.dc.b	$E7,$FC,$6C,$DE,$9F,$B6,$DE,$9F,$F4,$DF,$D5,$BE,$AF,$FB,$7C,$FD
	.dc.b	$EB,$FF,$59,$35,$34,$E4,$EB,$C4,$E2,$47,$FD,$2D,$F5,$1F,$39,$F2
	.dc.b	$FA,$3D,$88,$FA,$27,$C3,$F3,$B2,$3D,$6F,$F8,$0B,$E8,$7F,$86,$47
	.dc.b	$8B,$E7,$BC,$E4,$76,$7C,$2C,$BD,$64,$7E,$2C,$7B,$BC,$09,$1F,$21
	.dc.b	$FF,$07,$E4,$FA,$7F,$87,$E8,$4A,$F9,$AF,$07,$E6,$FA,$9F,$20,$8F
	.dc.b	$FA,$D7,$EA,$FF,$D3,$AF,$D4,$E1,$F9,$A4,$7E,$BC,$FF,$87,$BE,$A5
	.dc.b	$F3,$74,$93,$FE,$BF,$FA,$B7,$C3,$6F,$C1,$93,$FE,$C3,$B4,$70,$8F
	.dc.b	$F3,$B4,$78,$9B,$F2,$9D,$CE,$2E,$BE,$33,$C4,$47,$FD,$9D,$F5,$7F
	.dc.b	$B3,$7B,$78,$9E,$72,$3E,$19,$F5,$7C,$FC,$8F,$23,$F3,$3E,$13,$DD
	.dc.b	$C5,$27,$D0,$ED,$FF,$46,$7D,$5F,$F9,$D5,$FF,$2E,$35,$FF,$30,$35
	.dc.b	$FF,$79,$87,$0F,$FB,$D9,$E1,$DD,$FC,$6B,$E2,$3F,$99,$CF,$FA,$65
	.dc.b	$B1,$FC,$7A,$77,$C7,$FE,$FE,$FA,$B7,$D7,$7F,$94,$9B,$87,$FB,$C5
	.dc.b	$AF,$FA,$8D,$FC,$43,$98,$FF,$39,$46,$A3,$CE,$D7,$B4,$7F,$9D,$EB
	.dc.b	$FC,$85,$1E,$C6,$BE,$BF,$5F,$44,$47,$C8,$CF,$FC,$CF,$3F,$EB,$38
	.dc.b	$CE,$3E,$DA,$FF,$35,$FD,$67,$F8,$4E,$3F,$9E,$B7,$2E,$74,$8C,$3D
	.dc.b	$9F,$D8,$98,$75,$F0,$3D,$0E,$BF,$3F,$63,$C3,$5D,$2B,$81,$27,$53
	.dc.b	$FE,$CB,$8C,$D7,$28,$E2,$78,$5D,$DF,$A1,$D4,$71,$73,$7D,$DC,$CE
	.dc.b	$E7,$99,$E8,$2D,$78,$DB,$3C,$9F,$01,$AB,$C3,$C4,$F0,$19,$38,$90
	.dc.b	$C9,$ED,$78,$CF,$BA,$38,$1A,$AF,$33,$6F,$E4,$57,$0B,$E2,$D9,$ED
	.dc.b	$5F,$6A,$31,$1C,$4F,$B0,$E2,$3A,$9B,$8E,$35,$F1,$DA,$70,$29,$EF
	.dc.b	$8E,$31,$E4,$8F,$71,$FB,$2B,$FF,$2F,$36,$FF,$DA,$51,$88,$EC,$7F
	.dc.b	$B6,$5C,$3F,$38,$C4,$70,$3F,$D8,$3F,$58,$F2,$29,$EB,$BF,$CB,$C6
	.dc.b	$6B,$EE,$B5,$F8,$47,$A6,$37,$3C,$A3,$DB,$7F,$9B,$AE,$18,$7F,$88
	.dc.b	$1B,$7E,$88,$F8,$35,$FB,$09,$CB,$6F,$C8,$D3,$C6,$E1,$47,$06,$C6
	.dc.b	$63,$06,$6F,$F0,$D2,$AF,$25,$7A,$63,$DB,$3E,$98,$C4,$F9,$A7,$E0
	.dc.b	$1E,$33,$95,$FC,$27,$E0,$E1,$C2,$31,$1D,$23,$0F,$19,$E9,$7D,$E3
	.dc.b	$DA,$72,$29,$D8,$E1,$5F,$24,$8E,$08,$79,$89,$87,$C2,$C3,$E2,$1E
	.dc.b	$28,$D5,$7F,$91,$DF,$E1,$1E,$9F,$F1,$C3,$F0,$FE,$4D,$8F,$E8,$A7
	.dc.b	$F1,$98,$FC,$99,$1C,$8F,$AC,$7C,$9B,$E6,$73,$18,$6F,$AB,$FD,$CA
	.dc.b	$BA,$CE,$5E,$FF,$59,$FE,$2B,$D7,$AB,$63,$3E,$47,$F9,$B1,$FE,$5F
	.dc.b	$19,$7F,$AD,$AF,$9A,$D4,$7F,$3E,$8C,$57,$49,$CC,$A7,$2F,$C0,$FA
	.dc.b	$C7,$B3,$A6,$DF,$C0,$6C,$EE,$79,$FC,$27,$C1,$AF,$E4,$FE,$1F,$F8
	.dc.b	$8D,$FE,$53,$19,$AF,$BC,$7F,$84,$3B,$7F,$D2,$9C,$BF,$E8,$8F,$AC
	.dc.b	$7C,$EE,$30,$FE,$9B,$87,$1A,$D4,$E7,$0F,$EB,$73,$9A,$FE,$A7,$6F
	.dc.b	$8F,$D5,$3D,$A7,$36,$2B,$57,$FE,$74,$8C,$7A,$79,$C2,$CF,$F4,$2A
	.dc.b	$DE,$1E,$A5,$F1,$FD,$BD,$F7,$A5,$D6,$30,$F7,$F5,$71,$87,$36,$E1
	.dc.b	$CF,$D4,$D4,$F7,$6C,$77,$F8,$1E,$DE,$A7,$A5,$B5,$E5,$C3,$0D,$3D
	.dc.b	$11,$EF,$E0,$F8,$7F,$4E,$F9,$5C,$57,$F5,$E8,$C3,$EF,$F5,$73,$97
	.dc.b	$BD,$BC,$BF,$EB,$EF,$AB,$7D,$4F,$CE,$FA,$C6,$DF,$42,$73,$8F,$F6
	.dc.b	$A9,$CB,$FD,$B2,$3F,$CB,$2D,$AE,$1E,$C7,$17,$97,$5E,$27,$83,$73
	.dc.b	$FC,$5D,$BE,$87,$EF,$FD,$67,$FB,$B2,$FB,$46,$1F,$27,$A9,$E5,$8B
	.dc.b	$1A,$3C,$9F,$2D,$F1,$BE,$B3,$D4,$D8,$F4,$7A,$D8,$69,$C0,$F6,$71
	.dc.b	$FC,$ED,$7E,$13,$63,$E0,$7D,$06,$1A,$70,$3E,$27,$C8,$7C,$98,$F9
	.dc.b	$4D,$8F,$91,$EE,$FD,$AE,$1F,$9C,$D8,$F9,$9C,$18,$69,$C0,$FF,$0D
	.dc.b	$FD,$67,$FB,$9D,$FC,$07,$17,$E9,$A3,$13,$FB,$6E,$7F,$F4,$2D,$70
	.dc.b	$F1,$B5,$DF,$B6,$C7,$4F,$8C,$FC,$6F,$BD,$CD,$B1,$E0,$F8,$CC,$34
	.dc.b	$E0,$79,$3E,$97,$C5,$1E,$36,$C7,$F8,$CB,$EB,$1D,$DF,$BC,$5D,$F6
	.dc.b	$30,$FF,$00,$E0,$7D,$2D,$F9,$9C,$37,$0F,$F1,$EF,$D6,$3D,$CE,$1C
	.dc.b	$5F,$C2,$23,$13,$FE,$42,$C3,$FC,$D2,$F0,$E1,$D7,$B1,$B6,$C7,$4F
	.dc.b	$98,$E5,$F1,$DC,$DB,$1F,$E4,$CF,$A5,$7C,$6F,$F1,$65,$E8,$58,$C3
	.dc.b	$9F,$23,$5E,$5D,$BE,$4F,$F1,$EF,$AC,$FD,$14,$7F,$96,$17,$D4,$67
	.dc.b	$C7,$D3,$BF,$83,$E1,$E1,$5B,$E0,$F3,$E3,$E9,$C5,$67,$A3,$FC,$B1
	.dc.b	$7C,$CD,$F9,$9C,$36,$3F,$CD,$3F,$58,$FA,$3F,$E6,$6B,$E8,$EC,$61
	.dc.b	$FE,$6F,$F5,$9E,$8A,$FF,$38,$AF,$0F,$3B,$69,$DE,$BC,$3C,$9B,$7C
	.dc.b	$1F,$E8,$3F,$56,$F1,$6F,$D6,$61,$B5,$FE,$7E,$96,$FC,$3C,$5E,$2D
	.dc.b	$9F,$F4,$1F,$C3,$7A,$5B,$EF,$FF,$4E,$8C,$72,$7F,$5E,$7F,$AD,$CF
	.dc.b	$0F,$63,$5F,$07,$B6,$C7,$FA,$3B,$EB,$3D,$CD,$F3,$FF,$AC,$C6,$30
	.dc.b	$FE,$D2,$FE,$D1,$BD,$5A,$E3,$DB,$6F,$A7,$AD,$86,$9D,$E3,$AF,$91
	.dc.b	$DB,$C9,$B3,$E3,$7C,$3E,$AE,$31,$F0,$B7,$DA,$57,$FB,$0B,$37,$C7
	.dc.b	$7F,$70,$E5,$F8,$75,$AC,$3F,$DA,$35,$AB,$FF,$56,$D6,$9E,$5D,$FF
	.dc.b	$03,$0F,$82,$D8,$FF,$59,$E0,$F2,$3B,$78,$BF,$DF,$A3,$13,$D2,$31
	.dc.b	$E4,$66,$AC,$F2,$7F,$C0,$B9,$BA,$FD,$1F,$13,$86,$D7,$FB,$0A,$5D
	.dc.b	$B9,$8C,$36,$3F,$C3,$A3,$C4,$EB,$1D,$9B,$3C,$F5,$F8,$DF,$13,$E2
	.dc.b	$6C,$77,$DB,$0D,$38,$1E,$3C,$1E,$57,$F9,$51,$FF,$7E,$B5,$C9,$97
	.dc.b	$FC,$B3,$F5,$5F,$B8,$FA,$CB,$E4,$1A,$4A,$FF,$31,$CB,$E8,$3F,$D1
	.dc.b	$91,$86,$A2,$08,$E1,$C7,$E0,$9D,$CE,$A5,$F3,$FC,$2B,$F0,$D3,$C6
	.dc.b	$31,$96,$DB,$1D,$EB,$AA,$FA,$CF,$F0,$E3,$9E,$3D,$BE,$36,$CF,$E2
	.dc.b	$63,$FD,$FA,$F8,$9B,$C9,$A2,$3E,$DC,$7C,$BF,$2D,$EE,$8F,$F2,$DD
	.dc.b	$3E,$83,$A9,$23,$99,$DB,$E4,$BC,$24,$7C,$31,$E6,$F7,$FF,$11,$1D
	.dc.b	$F5,$F9,$71,$88,$FF,$34,$B7,$1F,$8F,$F4,$5F,$D6,$67,$81,$F4,$E7
	.dc.b	$F6,$6E,$0F,$F2,$07,$27,$73,$C1,$CD,$F1,$BE,$B3,$E1,$4F,$F9,$D9
	.dc.b	$B8,$7D,$46,$E3,$F7,$B1,$7C,$B7,$1E,$09,$87,$7B,$F9,$CF,$3E,$2F
	.dc.b	$A7,$F5,$9F,$D0,$63,$C8,$D7,$FD,$0A,$DC,$BF,$06,$FF,$A1,$23,$7E
	.dc.b	$97,$22,$3C,$BD,$2E,$BD,$F2,$3A,$BE,$0F,$8E,$F0,$FB,$12,$7C,$2E
	.dc.b	$DF,$0F,$C6,$F9,$88,$F9,$BE,$AF,$D3,$C3,$23,$D2,$FB,$3E,$03,$F1
	.dc.b	$FF,$58,$F2,$29,$2B,$E9,$47,$F8,$51,$FA,$9E,$E7,$CA,$6F,$27,$E6
	.dc.b	$23,$11,$FE,$A8,$7E,$F3,$7E,$13,$44,$79,$FB,$DC,$7A,$E4,$76,$7A
	.dc.b	$77,$FE,$93,$EB,$1B,$52,$57,$E1,$E3,$FC,$2A,$FE,$27,$EB,$3D,$0F
	.dc.b	$FA,$E9,$BC,$FF,$B8,$6F,$07,$F0,$25,$E7,$7E,$CB,$EB,$3D,$35,$FB
	.dc.b	$E6,$D7,$F0,$86,$DF,$F0,$BB,$FD,$75,$6F,$C5,$E4,$49,$FC,$E5,$7F
	.dc.b	$84,$9B,$9F,$72,$3A,$ED,$FE,$07,$1E,$C8,$FF,$69,$B6,$FF,$88,$B7
	.dc.b	$1F,$E3,$6F,$F8,$D2,$7F,$54,$FF,$94,$1B,$F0,$BD,$0E,$BE,$5F,$F1
	.dc.b	$0F,$AC,$FF,$4A,$4F,$EC,$5A,$FF,$B8,$9B,$87,$F2,$07,$FC,$6F,$1B
	.dc.b	$F0,$B9,$11,$E5,$E6,$FF,$1C,$FA,$C7,$C4,$D2,$4F,$EF,$E3,$FC,$40
	.dc.b	$DF,$37,$81,$1D,$7B,$BF,$C9,$BE,$B3,$FC,$6D,$1F,$C1,$DA,$FF,$BE
	.dc.b	$1B,$9F,$F3,$67,$F6,$DB,$FF,$40,$7F,$DC,$33,$C7,$E1,$BA,$1D,$7D
	.dc.b	$CF,$E6,$7B,$BE,$1E,$92,$BF,$8B,$79,$79,$7F,$8C,$4B,$B3,$89,$23
	.dc.b	$35,$D3,$B6,$47,$07,$4B,$4F,$E9,$B7,$F1,$23,$1A,$7F,$58,$8C,$4E
	.dc.b	$BD,$7F,$E9,$3E,$EB,$E8,$FD,$3C,$73,$1F,$2B,$EC,$AF,$6F,$5C,$9F
	.dc.b	$4B,$FD,$4B,$E6,$5F,$59,$F0,$BC,$F9,$5E,$E7,$83,$E4,$3F,$B6,$69
	.dc.b	$FD,$35,$F8,$7E,$1F,$CA,$3A,$77,$7C,$CF,$4E,$4F,$C8,$74,$FC,$AF
	.dc.b	$9A,$FA,$24,$74,$7F,$45,$FA,$BF,$E5,$CF,$84,$38,$E1,$FE,$EA,$DE
	.dc.b	$2F,$EF,$FF,$58,$FA,$0F,$ED,$F3,$F8,$4B,$FF,$02,$35,$E7,$E9,$6B
	.dc.b	$E1,$F7,$E9,$C9,$F1,$3F,$DE,$3E,$AF,$E8,$47,$97,$D3,$EA,$47,$67
	.dc.b	$A7,$E7,$7C,$24,$7F,$B0,$7D,$67,$F0,$C7,$D7,$F3,$DE,$62,$3C,$4F
	.dc.b	$85,$A7,$A8,$8F,$C5,$8F,$6F,$85,$23,$E4,$3E,$7F,$A0,$F8,$3E,$94
	.dc.b	$9F,$98,$ED,$F9,$9B,$FF,$B6,$9F,$A1,$EB,$7C,$83,$FC,$2F,$F5,$9F
	.dc.b	$E9,$F8,$FA,$9C,$7F,$30,$8F,$03,$FB,$E7,$D5,$BE,$6F,$FC,$3E,$D8
	.dc.b	$CC,$7F,$88,$BE,$B1,$F0,$FA,$F1,$65,$72,$78,$D7,$2F,$94,$93,$D9
	.dc.b	$DB,$F8,$5C,$1F,$1D,$E0,$24,$7F,$83,$FE,$B3,$F7,$0F,$9B,$C8,$F2
	.dc.b	$91,$F4,$CF,$A3,$E7,$E4,$7B,$1F,$C1,$7C,$2F,$B3,$8E,$4F,$C0,$76
	.dc.b	$FE,$5F,$EA,$BF,$CF,$4F,$F8,$7D,$AE,$7E,$37,$FC,$87,$F1,$5F,$36
	.dc.b	$30,$7F,$8E,$CF,$F2,$D4,$F0,$7F,$C7,$CD,$7F,$CB,$AD,$7F,$CC,$1F
	.dc.b	$58,$BF,$F3,$93,$63,$A0,$70,$F7,$6B,$FE,$4B,$5F,$59,$FE,$58,$6C
	.dc.b	$7F,$8E,$5A,$FE,$D5,$E7,$E2,$7F,$9A,$2F,$FC,$89,$CB,$E6,$F6,$BF
	.dc.b	$CC,$31,$8B,$FF,$3B,$7A,$9F,$51,$F9,$13,$A2,$3F,$CC,$BD,$4F,$F3
	.dc.b	$4F,$D5,$FE,$A4,$9F,$3F,$FE,$6C,$B7,$AB,$A8,$97,$FE,$4F,$AF,$D8
	.dc.b	$56,$A3,$13,$A7,$F8,$74,$EA,$B1,$9F,$FA,$0A,$3B,$CF,$F4,$49,$CE
	.dc.b	$7F,$E8,$E9,$CC,$FF,$17,$C3,$B3,$D3,$0D,$BF,$E9,$59,$EC,$F8,$88
	.dc.b	$6C,$EC,$63,$0F,$14,$F2,$85,$76,$9C,$C6,$EF,$C9,$5F,$17,$0E,$18
	.dc.b	$79,$AF,$FC,$EF,$18,$9E,$98,$73,$BE,$38,$73,$3D,$13,$CD,$C0,$C3
	.dc.b	$A4,$FF,$9C,$EF,$A4,$EA,$FA,$3C,$8E,$35,$F0,$9F,$80,$4C,$3D,$6F
	.dc.b	$CB,$AE,$6E,$6F,$B3,$CB,$0F,$4C,$FD,$B8,$FF,$52,$5F,$F9,$59,$B3
	.dc.b	$FE,$D5,$8C,$47,$B8,$FD,$83,$87,$CC,$62,$38,$1F,$92,$F8,$59,$1E
	.dc.b	$13,$F1,$CE,$1F,$31,$CB,$0F,$C5,$46,$5F,$69,$30,$FC,$54,$65,$F2
	.dc.b	$46,$67,$F1,$13,$97,$A4,$E8,$E2,$F8,$8F,$21,$F9,$C8,$E0,$FD,$7A
	.dc.b	$FC,$1C,$62,$FE,$CC,$F6,$39,$1E,$A3,$81,$8B,$F5,$BD,$9A,$F7,$31
	.dc.b	$7E,$47,$07,$39,$FD,$D5,$78,$6F,$C7,$3D,$6B,$D8,$FB,$8F,$21,$CA
	.dc.b	$FE,$23,$E4,$31,$1D,$E3,$37,$DA,$7A,$1C,$CE,$11,$E8,$24,$F8,$21
	.dc.b	$F5,$8F,$37,$FD,$F2,$D8,$FF,$6D,$35,$FF,$0B,$3F,$EC,$EA,$52,$64
	.dc.b	$ED,$FF,$7C,$35,$D1,$E0,$AF,$E6,$ED,$3F,$9C,$FD,$5F,$FA,$05,$FE
	.dc.b	$8C,$D7,$88,$70,$F4,$6B,$E3,$FA,$C7,$CA,$BE,$6F,$FA,$2C,$61,$BE
	.dc.b	$AF,$F7,$6A,$FE,$A9,$39,$7C,$3E,$6F,$B9,$B9,$F9,$A3,$11,$E4,$C5
	.dc.b	$18,$8F,$27,$CD,$D7,$E3,$4A,$31,$1E,$B6,$C6,$6B,$FA,$DF,$C2,$FA
	.dc.b	$5F,$9D,$B7,$83,$C4,$FE,$C5,$18,$3E,$6C,$EF,$0F,$F2,$64,$E9,$FF
	.dc.b	$27,$7D,$5B,$D3,$F6,$61,$EA,$BE,$0F,$B2,$75,$87,$5B,$FB,$D5,$A8
	.dc.b	$FB,$72,$F9,$0E,$71,$F7,$BD,$CF,$C6,$3C,$6D,$5B,$8C,$4E,$A7,$CB
	.dc.b	$EA,$F2,$7F,$19,$5B,$73,$5C,$E3,$18,$F9,$FE,$43,$85,$7F,$A6,$7D
	.dc.b	$1B,$49,$CF,$62,$73,$EA,$EA,$6E,$3D,$5E,$3E,$9E,$31,$F9,$2D,$CB
	.dc.b	$E3,$7C,$17,$4F,$53,$BB,$70,$F3,$65,$86,$9E,$98,$F2,$F0,$7A,$3E
	.dc.b	$95,$F2,$B8,$AF,$91,$18,$7E,$07,$AB,$9D,$3E,$DB,$73,$FC,$9F,$D5
	.dc.b	$3E,$A3,$E1,$FD,$5B,$6F,$8B,$39,$C7,$EA,$4E,$5F,$AD,$1F,$E5,$D6
	.dc.b	$D7,$0F,$57,$8B,$CB,$A7,$13,$C1,$B9,$FF,$A6,$ED,$F3,$FF,$53,$EA
	.dc.b	$FF,$83,$DF,$68,$C3,$E4,$F4,$FC,$B1,$63,$47,$93,$E5,$FE,$37,$D5
	.dc.b	$FA,$9B,$1E,$8F,$57,$0D,$38,$1E,$CE,$3F,$9D,$AF,$C2,$6C,$7C,$0F
	.dc.b	$53,$0D,$38,$1F,$13,$E4,$3E,$4C,$7C,$A6,$C7,$C8,$F6,$FE,$D7,$0F
	.dc.b	$CE,$6C,$7C,$CE,$0C,$34,$E0,$7E,$97,$EA,$FF,$DE,$0F,$E0,$38,$BF
	.dc.b	$37,$18,$9F,$F6,$37,$3A,$73,$E8,$6B,$BE,$E1,$CD,$E3,$F1,$7F,$8D
	.dc.b	$F7,$6C,$FE,$31,$EB,$F1,$98,$CB,$FD,$A5,$9A,$31,$D7,$FC,$7A,$1F
	.dc.b	$49,$C3,$0F,$43,$6F,$FD,$B5,$F5,$6E,$EF,$D9,$AE,$FB,$18,$7F,$6F
	.dc.b	$C2,$FA,$4B,$F3,$38,$6E,$5F,$C2,$7E,$AD,$EB,$F0,$CF,$F7,$F1,$89
	.dc.b	$FE,$1D,$87,$F5,$37,$87,$16,$BE,$4B,$6D,$8E,$9F,$31,$CB,$E3,$F9
	.dc.b	$B6,$3F,$8C,$7D,$27,$F3,$C6,$F9,$DF,$C3,$A3,$19,$FF,$BE,$54,$67
	.dc.b	$C8,$D7,$95,$6F,$9B,$FC,$5B,$EA,$FF,$4B,$1F,$C9,$57,$D3,$E7,$C7
	.dc.b	$D3,$BF,$80,$E1,$E1,$5B,$E0,$F3,$D3,$A7,$0D,$9E,$8F,$F2,$45,$F3
	.dc.b	$37,$E6,$70,$D8,$FE,$6D,$F5,$6F,$A2,$FE,$56,$BE,$52,$C6,$1F,$E5
	.dc.b	$FF,$57,$E8,$AF,$E7,$CB,$C3,$CE,$DA,$77,$AF,$0F,$26,$DF,$07,$F9
	.dc.b	$CF,$D5,$3C,$6B,$F5,$78,$6D,$7F,$4B,$96,$FC,$3C,$4E,$2D,$9F,$EA
	.dc.b	$1F,$05,$E9,$6F,$BB,$FD,$1A,$31,$C8,$F2,$3F,$BD,$C7,$F6,$08,$C5
	.dc.b	$7F,$86,$21,$EB,$F1,$F7,$F9,$36,$7F,$AF,$7D,$5F,$C1,$6F,$9D,$FD
	.dc.b	$52,$31,$87,$43,$FC,$50,$FE,$EE,$78,$F5,$EF,$93,$83,$63,$AF,$AD
	.dc.b	$86,$9C,$0F,$0F,$91,$DF,$C9,$F0,$36,$3C,$BE,$AE,$1A,$70,$3D,$3F
	.dc.b	$1F,$FD,$AB,$97,$E2,$F0,$FF,$87,$B4,$4E,$7A,$FD,$9D,$FE,$D8,$B8
	.dc.b	$72,$3F,$C0,$16,$F1,$FA,$78,$BF,$DD,$A3,$13,$FE,$05,$CB,$F5,$2F
	.dc.b	$0F,$1F,$57,$B6,$C7,$F8,$2F,$89,$D9,$E7,$F8,$5F,$F0,$0C,$62,$7A
	.dc.b	$46,$3B,$B3,$56,$79,$3C,$3C,$4C,$5F,$D5,$3F,$CB,$8E,$33,$FF,$1E
	.dc.b	$6B,$F1,$7E,$17,$BB,$70,$F2,$6F,$86,$9E,$78,$F2,$E0,$F2,$7F,$CC
	.dc.b	$2D,$3F,$CC,$6E,$0F,$F2,$FB,$FE,$6D,$FA,$9F,$DC,$FD,$5E,$21,$18
	.dc.b	$76,$D7,$FC,$CE,$D7,$FC,$E2,$BE,$83,$F4,$11,$C1,$AF,$06,$BC,$1A
	.dc.b	$F8,$FE,$7F,$FC,$F7,$A7,$F9,$8A,$91,$8C,$BB,$B6,$BD,$57,$E0,$FA
	.dc.b	$BF,$F0,$FC,$7B,$78,$F4,$FB,$1B,$87,$FA,$3F,$0E,$73,$FE,$4B,$87
	.dc.b	$1E,$89,$1F,$E4,$FF,$94,$FF,$43,$E7,$E5,$FA,$0E,$64,$9E,$55,$DB
	.dc.b	$E8,$24,$75,$74,$FE,$0F,$DE,$72,$79,$9B,$1F,$E5,$D6,$BF,$89,$F9
	.dc.b	$2F,$EA,$EE,$78,$F8,$11,$E1,$5F,$03,$EA,$FE,$4C,$7F,$99,$DA,$FE
	.dc.b	$4D,$B9,$FF,$A4,$2D,$F2,$7C,$78,$25,$F1,$7A,$71,$FC,$CF,$AB,$FE
	.dc.b	$75,$3F,$E7,$36,$BF,$9F,$6E,$1F,$E9,$C7,$FD,$0D,$1B,$F4,$79,$11
	.dc.b	$E5,$DF,$D7,$BE,$47,$57,$C1,$F1,$DE,$1F,$90,$93,$E1,$76,$F8,$7E
	.dc.b	$33,$CC,$47,$CD,$F5,$5E,$9E,$99,$1E,$97,$D9,$F0,$1F,$88,$FA,$B7
	.dc.b	$8F,$49,$5F,$E9,$18,$FF,$07,$3F,$E9,$3F,$73,$E4,$B7,$8F,$F2,$31
	.dc.b	$88,$FE,$04,$FD,$A6,$FC,$2E,$88,$F3,$F7,$B8,$F6,$C8,$EC,$F4,$ED
	.dc.b	$FC,$EF,$D5,$B6,$A4,$AF,$F5,$34,$7F,$B2,$9F,$F5,$47,$D5,$FA,$1F
	.dc.b	$E2,$8D,$E7,$FF,$68,$B7,$83,$FD,$DA,$BC,$EF,$D5,$FD,$5F,$A6,$BF
	.dc.b	$DA,$ED,$AF,$F7,$9B,$6F,$FD,$F5,$7F,$A8,$9C,$F8,$BC,$89,$3F,$EB
	.dc.b	$BA,$FF,$05,$B7,$3E,$E4,$75,$D9,$FB,$C8,$F6,$47,$F2,$D6,$E3,$FE
	.dc.b	$EA,$6E,$09,$91,$FE,$2C,$7F,$D9,$4F,$F9,$2D,$BF,$0B,$D4,$ED,$E5
	.dc.b	$7F,$0A,$FA,$BF,$F4,$8C,$FF,$B4,$5A,$FF,$3D,$6E,$1F,$EF,$A7,$FC
	.dc.b	$63,$1B,$F0,$B9,$11,$E5,$E6,$FF,$16,$FA,$BF,$F3,$34,$7F,$B7,$5A
	.dc.b	$FF,$49,$6D,$B3,$F6,$4A,$7F,$C3,$CF,$1F,$82,$E8,$47,$A7,$B9,$FC
	.dc.b	$93,$EA,$FF,$C6,$91,$FE,$EB,$6B,$FD,$5D,$B8,$77,$8F,$95,$1F,$BC
	.dc.b	$C3,$C2,$FF,$B8,$63,$97,$C2,$F5,$3B,$7B,$7F,$CC,$B6,$7C,$35,$25
	.dc.b	$7F,$BE,$FC,$9C,$BF,$EF,$D9,$75,$F1,$24,$66,$BA,$76,$48,$E0,$E9
	.dc.b	$FE,$E1,$5F,$11,$BE,$B7,$8C,$FE,$8B,$EE,$3E,$87,$CD,$F1,$92,$BC
	.dc.b	$4E,$DF,$27,$5F,$AC,$8F,$E3,$78,$7E,$32,$F9,$87,$D5,$FC,$1F,$3E
	.dc.b	$4F,$B5,$E2,$F9,$BC,$3F,$8C,$47,$A7,$2F,$DE,$BF,$EB,$11,$F9,$BD
	.dc.b	$9E,$9F,$CD,$24,$7C,$9B,$FA,$1F,$39,$23,$E6,$3A,$7F,$C2,$9F,$4F
	.dc.b	$FC,$D1,$BF,$0B,$FD,$A2,$31,$AF,$B3,$EA,$DF,$53,$9F,$4E,$61,$E6
	.dc.b	$3F,$5D,$7F,$85,$6E,$1C,$FD,$39,$1B,$13,$FB,$37,$FC,$4D,$F5,$7F
	.dc.b	$4D,$D4,$7E,$E9,$EB,$E9,$F9,$4E,$6F,$AB,$E7,$7D,$04,$7B,$7D,$5F
	.dc.b	$F1,$37,$BF,$CF,$7B,$48,$F6,$9F,$81,$A4,$8F,$0B,$A7,$E3,$F8,$3F
	.dc.b	$14,$8F,$AF,$27,$D3,$FC,$9F,$4A,$4F,$A9,$DB,$F4,$3D,$6F,$9C,$47
	.dc.b	$E2,$7D,$5F,$FA,$7D,$F8,$3D,$3F,$4C,$8F,$C4,$9F,$F0,$EF,$D5,$3E
	.dc.b	$6E,$92,$7E,$77,$D5,$BE,$1F,$7E,$1C,$9F,$A5,$7F,$F0,$05,$B7,$D3
	.dc.b	$E5,$24,$69,$E0,$FF,$96,$E5,$F1,$BE,$22,$47,$DD,$FA,$BF,$DC,$BD
	.dc.b	$BC,$6F,$39,$1F,$0C,$FA,$BE,$7E,$47,$91,$FF,$1C,$F1,$7F,$A1,$A3
	.dc.b	$CF,$C5,$F0,$08,$EC,$FC,$67,$D4,$FF,$90,$9F,$F3,$87,$D5,$2F,$FC
	.dc.b	$86,$D8,$DF,$C4,$62,$7F,$CC,$50,$9D,$3F,$E6,$86,$BF,$CC,$0D,$47
	.dc.b	$0A,$FF,$44,$C7,$F9,$AD,$AF,$F9,$C1,$AF,$16,$BF,$CE,$9D,$F8,$3F
	.dc.b	$C2,$BE,$27,$FA,$4C,$FF,$A2,$5B,$1F,$C3,$A7,$7C,$BC,$7D,$6F,$F2
	.dc.b	$FD,$7F,$52,$8F,$E8,$AD,$BF,$EC,$54,$AD,$FC,$94,$9F,$F2,$F3,$FE
	.dc.b	$68,$9D,$47,$CC,$F8,$AC,$11,$F9,$FE,$CD,$7F,$D3,$D3,$F5,$DB,$3C
	.dc.b	$E7,$59,$7A,$62,$E5,$F0,$66,$D7,$9A,$6A,$BF,$F3,$1D,$F1,$8F,$21
	.dc.b	$0F,$0D,$70,$9F,$14,$EE,$72,$6D,$C1,$2F,$C5,$19,$7E,$A5,$F3,$BE
	.dc.b	$8D,$9D,$1E,$72,$9E,$78,$DB,$8B,$F4,$D7,$48,$C5,$7C,$56,$9B,$AE
	.dc.b	$2F,$CA,$3E,$79,$A3,$07,$B0,$B7,$EB,$7C,$86,$5C,$3F,$2E,$7D,$4F
	.dc.b	$0A,$F2,$34,$FC,$3D,$7D,$D7,$F3,$13,$F9,$F2,$3F,$9F,$6B,$F9,$F6
	.dc.b	$BC,$0F,$E8,$1F,$53,$FD,$CE,$FF,$B3,$B6,$FC,$31,$E6,$8C,$4F,$F8
	.dc.b	$E5,$AF,$A4,$F9,$EE,$0C,$E1,$F8,$0A,$E6,$7F,$89,$A3,$5E,$8F,$CC
	.dc.b	$BF,$0B,$FE,$C6,$8E,$38,$7E,$41,$DE,$1F,$3E,$B6,$FA,$AB,$95,$F5
	.dc.b	$79,$18,$8F,$43,$87,$94,$6D,$E4,$60,$D9,$83,$E9,$1E,$F1,$3E,$26
	.dc.b	$C7,$37,$78,$7C,$17,$99,$A2,$CF,$E0,$A3,$50,$BF,$7B,$CB,$0F,$43
	.dc.b	$C4,$D6,$1E,$D6,$9E,$7B,$E9,$3C,$1F,$31,$88,$D6,$1E,$78,$FB,$DE
	.dc.b	$5F,$D1,$24,$FF,$81,$7E,$AB,$FD,$51,$3F,$86,$9F,$E6,$B8,$FF,$88
	.dc.b	$2B,$F9,$91,$F8,$58,$C9,$FC,$EB,$EA,$7F,$C3,$91,$FC,$DD,$AF,$F9
	.dc.b	$A1,$AF,$FA,$15,$A9,$FF,$6D,$BC,$7C,$AE,$84,$7C,$23,$EA,$9F,$28
	.dc.b	$F9,$9C,$C6,$1B,$E8,$FF,$73,$AF,$4C,$E5,$F5,$E1,$FE,$2E,$D7,$AB
	.dc.b	$63,$3E,$3F,$F9,$39,$F5,$35,$FF,$4F,$7C,$0F,$86,$BF,$C5,$DE,$17
	.dc.b	$F8,$51,$FF,$18,$46,$67,$EE,$4E,$5F,$BD,$F5,$4F,$43,$FD,$01,$C3
	.dc.b	$FE,$5C,$BE,$0F,$F9,$7A,$75,$87,$4D,$76,$FF,$A5,$21,$F1,$D8,$C7
	.dc.b	$FD,$65,$EA,$BD,$4C,$42,$B7,$7F,$E4,$5A,$D3,$FE,$47,$87,$A7,$C1
	.dc.b	$DD,$7A,$23,$18,$67,$E3,$BA,$FB,$AF,$43,$18,$FF,$94,$63,$0F,$5F
	.dc.b	$47,$38,$FC,$26,$ED,$EF,$F4,$BC,$11,$E1,$6D,$79,$7D,$FE,$FE,$9F
	.dc.b	$95,$B1,$E9,$CB,$0D,$38,$1E,$DE,$2F,$87,$F4,$AF,$23,$DD,$18,$7E
	.dc.b	$07,$A7,$8A,$F8,$8D,$D7,$F6,$DF,$52,$FA,$7F,$9B,$F5,$4D,$7E,$7C
	.dc.b	$E7,$0F,$F3,$94,$E5,$FF,$3A,$47,$F9,$2D,$B5,$C3,$D4,$E2,$F2,$EB
	.dc.b	$C4,$F0,$6E,$3E,$2F,$27,$EE,$FD,$57,$FB,$5E,$BA,$C6,$1F,$1F,$A3
	.dc.b	$E4,$8B,$1A,$3C,$7F,$27,$F1,$3E,$AB,$D2,$D8,$F3,$FA,$18,$69,$C0
	.dc.b	$F5,$F0,$FC,$CD,$3E,$0B,$63,$DF,$F3,$98,$69,$C0,$F8,$7E,$87,$C8
	.dc.b	$8F,$92,$D8,$F8,$FE,$CF,$D8,$E0,$F9,$AD,$8F,$97,$BE,$1A,$70,$3F
	.dc.b	$D4,$5F,$55,$FE,$DE,$7F,$D4,$9C,$EE,$2F,$D2,$CE,$6B,$F9,$94,$E6
	.dc.b	$B8,$FC,$36,$BB,$B6,$D8,$EB,$E8,$FE,$33,$DB,$E6,$D8,$F0,$FC,$36
	.dc.b	$1A,$70,$3C,$BF,$45,$E2,$8F,$1B,$63,$FD,$75,$F5,$4E,$DF,$DD,$AE
	.dc.b	$EB,$18,$7F,$7F,$C0,$FA,$1B,$F2,$B8,$6E,$1F,$EC,$5F,$A9,$7D,$57
	.dc.b	$0E,$2F,$E0,$F1,$89,$FF,$65,$60,$9C,$F0,$EB,$E3,$6D,$73,$78,$FC
	.dc.b	$A7,$2F,$95,$B3,$DD,$FE,$27,$F4,$4F,$8A,$BE,$7E,$1B,$5F,$ED,$99
	.dc.b	$FD,$46,$5F,$D9,$5F,$E1,$6D,$F2,$39,$F9,$3F,$C7,$BE,$A9,$E3,$5F
	.dc.b	$A5,$C3,$71,$FF,$72,$4B,$E0,$B8,$78,$5C,$5B,$3A,$CB,$A7,$17,$56
	.dc.b	$C7,$FB,$B4,$7C,$B7,$F2,$E5,$F2,$D6,$30,$FF,$32,$FA,$A7,$D0,$5F
	.dc.b	$81,$C3,$6F,$FD,$F1,$F5,$4D,$FF,$9C,$AF,$0E,$C6,$1C,$ED,$AF,$26
	.dc.b	$DF,$27,$FA,$0F,$D4,$BC,$5B,$F5,$58,$6D,$7D,$DF,$60,$FD,$FE,$15
	.dc.b	$FB,$23,$46,$57,$8D,$DB,$87,$BB,$FD,$43,$DE,$70,$DF,$7F,$0D,$C3
	.dc.b	$9E,$3F,$8D,$A7,$B1,$C7,$EA,$B9,$36,$B7,$4F,$AA,$F8,$91,$87,$F7
	.dc.b	$11,$FE,$2B,$72,$DF,$43,$FB,$1C,$7E,$1A,$78,$72,$63,$B3,$2E,$BD
	.dc.b	$7C,$4F,$56,$E1,$D7,$C9,$E7,$E5,$74,$6D,$78,$FD,$6C,$34,$F3,$47
	.dc.b	$8B,$E2,$BF,$BA,$F3,$7C,$69,$F5,$46,$38,$7C,$FD,$FB,$9F,$73,$72
	.dc.b	$FF,$24,$73,$BC,$7E,$7E,$3F,$F8,$02,$31,$7F,$23,$2F,$EF,$AF,$CA
	.dc.b	$87,$91,$BB,$E0,$D9,$F9,$F8,$3B,$3C,$FF,$0F,$FE,$0C,$8C,$4F,$58
	.dc.b	$C7,$76,$6A,$CF,$27,$87,$89,$8B,$FF,$30,$1F,$B3,$9E,$F5,$D7,$7F
	.dc.b	$8B,$F0,$F6,$7D,$4F,$87,$7C,$5F,$26,$EB,$FE,$37,$F2,$7F,$C9,$33
	.dc.b	$FC,$65,$A9,$A7,$F9,$3B,$EA,$3F,$77,$F5,$57,$C8,$34,$95,$FE,$64
	.dc.b	$97,$CE,$FF,$A2,$23,$0D,$7F,$54,$23,$78,$77,$9D,$4F,$EB,$E5,$F3
	.dc.b	$9F,$B1,$AF,$D9,$CF,$F3,$19,$CB,$FA,$26,$BF,$E8,$97,$AA,$FA,$AF
	.dc.b	$F0,$F3,$C7,$93,$A6,$DD,$DB,$19,$3A,$61,$87,$D1,$C9,$70,$FC,$3C
	.dc.b	$78,$63,$5F,$23,$EE,$CB,$CF,$3E,$FF,$9D,$E0,$47,$81,$E9,$F8,$FE
	.dc.b	$84,$7D,$91,$E1,$F7,$FF,$DF,$71,$FD,$45,$AF,$F8,$E7,$9B,$CD,$F3
	.dc.b	$FF,$D6,$A3,$F8,$2C,$FF,$AC,$09,$CD,$CE,$BD,$9F,$55,$F2,$32,$F3
	.dc.b	$B7,$83,$F9,$03,$70,$FE,$C1,$8F,$FA,$9A,$73,$C1,$C8,$93,$FC,$52
	.dc.b	$B8,$2F,$AA,$FE,$42,$FF,$18,$6B,$FE,$BF,$6E,$1F,$21,$FF,$42,$46
	.dc.b	$FD,$1F,$A5,$1F,$D2,$23,$07,$3E,$FC,$F7,$C8,$EC,$F0,$F8,$CE,$3E
	.dc.b	$BC,$9F,$13,$B7,$C5,$F1,$9D,$88,$F9,$FE,$A3,$C3,$B4,$8F,$53,$E4
	.dc.b	$F7,$FF,$0D,$F5,$4F,$1E,$92,$BF,$9B,$C7,$F8,$4D,$FE,$73,$ED,$FC
	.dc.b	$B7,$FC,$18,$D7,$FC,$D5,$F5,$5F,$91,$8F,$B9,$1F,$EE,$55,$F0,$99
	.dc.b	$24,$74,$F7,$F8,$75,$48,$EE,$F3,$ED,$FC,$E7,$D5,$36,$A4,$AF,$E9
	.dc.b	$B1,$FE,$15,$7F,$A7,$FD,$57,$A5,$FF,$7B,$37,$9F,$F3,$ED,$E0,$FE
	.dc.b	$D0,$BC,$EF,$D5,$7D,$57,$AA,$BF,$4A,$DA,$FE,$DC,$DB,$FF,$17,$5F
	.dc.b	$EA,$E7,$3E,$2F,$22,$4F,$F6,$2A,$FF,$08,$B7,$3E,$E4,$75,$D9,$FB
	.dc.b	$B8,$F6,$43,$1F,$DB,$C6,$1B,$CD,$BE,$8E,$1F,$0B,$2F,$FB,$8D,$FF
	.dc.b	$94,$5B,$E5,$FF,$07,$FA,$AF,$F1,$8C,$7F,$76,$6B,$DF,$5F,$F0,$64
	.dc.b	$63,$2F,$E0,$EF,$F8,$7E,$78,$78,$3C,$C8,$F3,$F3,$7F,$8A,$7D,$53
	.dc.b	$E2,$69,$27,$FC,$0B,$1F,$E1,$F6,$F9,$BC,$08,$EB,$DE,$FE,$3D,$F5
	.dc.b	$5F,$AD,$8F,$31,$FE,$5F,$71,$9F,$F1,$C6,$E9,$FE,$25,$7F,$72,$DF
	.dc.b	$86,$E6,$4A,$FF,$0B,$CF,$F8,$71,$FF,$0C,$EC,$F8,$7F,$E5,$CD,$8C
	.dc.b	$C7,$F1,$C3,$3D,$92,$38,$BC,$23,$8F,$6C,$9E,$AE,$DF,$58,$AF,$88
	.dc.b	$E1,$B9,$7F,$38,$6E,$5F,$E4,$1F,$71,$F4,$5E,$6E,$09,$87,$A9,$F4
	.dc.b	$FB,$1E,$AE,$B9,$3E,$E7,$FA,$57,$CC,$3E,$AB,$DF,$EA,$4A,$F8,$6F
	.dc.b	$07,$C9,$E1,$FC,$42,$3F,$E4,$0A,$FF,$0F,$3F,$2B,$D2,$F9,$24,$7E
	.dc.b	$6C,$FC,$CF,$9A,$91,$F4,$1D,$3F,$13,$EA,$7F,$97,$B7,$E1,$7F,$B0
	.dc.b	$46,$38,$BF,$AD,$FD,$53,$E7,$73,$E8,$CC,$3F,$CA,$77,$F8,$27,$FD
	.dc.b	$4C,$F1,$E4,$FF,$35,$2F,$0F,$A9,$27,$FC,$B7,$F4,$BF,$39,$E7,$E2
	.dc.b	$F8,$88,$F6,$9F,$2F,$CE,$C8,$F0,$BF,$DD,$FE,$AB,$F8,$54,$30,$7C
	.dc.b	$F7,$B0,$91,$C0,$F4,$E9,$F0,$08,$FB,$23,$E1,$F8,$12,$3D,$EF,$F8
	.dc.b	$17,$37,$D3,$7C,$8F,$42,$57,$C4,$78,$3F,$3F,$D4,$F9,$A4,$7F,$CF
	.dc.b	$7F,$55,$FE,$9D,$7E,$5F,$07,$D2,$23,$F3,$A7,$FC,$29,$F5,$2F,$9B
	.dc.b	$A4,$9F,$F4,$2F,$D5,$3E,$1B,$7E,$1C,$9F,$F4,$4D,$FF,$C0,$16,$DF
	.dc.b	$4F,$94,$91,$A7,$83,$FA,$A9,$7C,$5F,$88,$91,$FE,$94,$FA,$A7,$95
	.dc.b	$E6,$F1,$E4,$F8,$1D,$BC,$FE,$7F,$D6,$47,$F0,$7F,$09,$EE,$E3,$91
	.dc.b	$E8,$74,$FE,$F3,$EA,$7F,$C7,$AF,$F9,$38,$38,$7F,$CA,$2D,$7F,$D7
	.dc.b	$8D,$7F,$DF,$8F,$2F,$03,$FC,$73,$F5,$4C,$3F,$C8,$4D,$AF,$F5,$8D
	.dc.b	$6F,$93,$9F,$C6,$FF,$91,$E3,$FD,$0D,$3A,$8F,$F2,$2F,$06,$5F,$F3
	.dc.b	$9C,$EA,$7D,$0D,$7C,$F1,$AF,$52,$47,$B4,$F6,$B8,$78,$BF,$E5,$2F
	.dc.b	$91,$91,$E6,$8F,$27,$AD,$23,$D2,$FF,$92,$38,$31,$1F,$2E,$2E,$BF
	.dc.b	$E5,$AF,$88,$FB,$73,$C0,$86,$CF,$F6,$9B,$A7,$04,$8F,$F3,$1E,$5F
	.dc.b	$E5,$89,$D4,$62,$BB,$E9,$F5,$1F,$25,$F6,$9D,$36,$7F,$DE,$95,$C9
	.dc.b	$E1,$19,$3E,$DD,$72,$7C,$51,$A6,$9F,$E5,$CC,$3C,$2F,$03,$90,$AE
	.dc.b	$CE,$4E,$27,$E3,$AB,$B5,$F9,$A3,$38,$7B,$EF,$AC,$6C,$EB,$7D,$1C
	.dc.b	$1C,$2F,$9E,$1E,$7C,$3A,$11,$F8,$37,$D3,$0F,$F4,$19,$DE,$FD,$2E
	.dc.b	$2F,$9C,$7A,$6F,$93,$6F,$9B,$C2,$7C,$F8,$79,$6D,$1F,$98,$24,$FE
	.dc.b	$4E,$32,$D8,$CB,$F9,$C6,$BF,$B2,$7F,$4D,$C1,$23,$C0,$7E,$B1,$C3
	.dc.b	$F0,$CC,$10,$C1,$FC,$05,$DE,$1E,$C8,$F9,$6F,$E4,$23,$38,$7C,$B2
	.dc.b	$4F,$F0,$08,$CB,$F0,$C8,$7F,$0A,$72,$F4,$8E,$2E,$4F,$24,$F8,$6F
	.dc.b	$E9,$D7,$07,$E9,$4E,$AB,$F8,$8C,$FD,$9B,$D3,$67,$47,$94,$F8,$A7
	.dc.b	$BC,$C9,$E3,$3C,$F4,$8F,$04,$FE,$15,$B3,$F8,$46,$BF,$96,$6D,$FA
	.dc.b	$5B,$69,$9F,$55,$FE,$95,$27,$C1,$7F,$B9,$5F,$ED,$E1,$FE,$2D,$08
	.dc.b	$C3,$FC,$65,$AF,$F3,$76,$BD,$CE,$FE,$8F,$28,$FF,$1B,$C6,$4E,$91
	.dc.b	$93,$BF,$D4,$7F,$86,$A3,$4D,$7A,$35,$F6,$35,$3F,$E1,$57,$F9,$73
	.dc.b	$5F,$E9,$8D,$7A,$F8,$FE,$EF,$A9,$7C,$93,$E6,$3F,$9B,$46,$1B,$EB
	.dc.b	$FF,$73,$BF,$EA,$D3,$97,$FA,$C7,$C0,$71,$D7,$E8,$6E,$7F,$CA,$5D
	.dc.b	$56,$BF,$CB,$75,$B7,$FC,$BB,$E0,$7F,$84,$AB,$89,$F2,$FE,$A5,$E8
	.dc.b	$F1,$C7,$95,$B0,$E5,$18,$AD,$DF,$2A,$FC,$15,$6A,$3A,$7C,$5F,$37
	.dc.b	$EC,$7A,$CF,$53,$17,$F8,$9C,$17,$88,$56,$EF,$B4,$2B,$FC,$8D,$18
	.dc.b	$78,$B7,$87,$94,$62,$7F,$CF,$AB,$E2,$FB,$75,$BC,$DD,$C6,$1E,$3E
	.dc.b	$96,$31,$EC,$DC,$39,$F9,$FA,$9F,$13,$63,$BF,$CC,$76,$F3,$FC,$6D
	.dc.b	$AF,$46,$18,$69,$EB,$8F,$4F,$07,$C3,$FA,$37,$C8,$E2,$BE,$1C,$61
	.dc.b	$F8,$5E,$86,$72,$F7,$B7,$93,$ED,$FD,$43,$E9,$BE,$77,$D4,$B6,$FA
	.dc.b	$13,$9C,$7E,$7C,$E5,$FA,$31,$FE,$4B,$6D,$70,$F4,$78,$BC,$BA,$B1
	.dc.b	$3C,$1B,$9F,$FA,$BA,$DE,$4F,$DF,$FA,$9F,$CD,$5F,$68,$C3,$E4,$F4
	.dc.b	$7C,$B1,$63,$47,$93,$E4,$BE,$37,$D4,$FA,$9B,$1E,$8F,$47,$0D,$38
	.dc.b	$1E,$CE,$3F,$9D,$AF,$C2,$6C,$7C,$0F,$3F,$0D,$38,$1F,$13,$CF,$F9
	.dc.b	$31,$F2,$9B,$1F,$23,$D5,$FB,$5C,$3F,$39,$B1,$F3,$38,$30,$D3,$81
	.dc.b	$FA,$0F,$A9,$FF,$33,$BF,$80,$E2,$FD,$34,$62,$7F,$DB,$3C,$E9,$CF
	.dc.b	$C3,$6B,$BE,$E1,$CD,$E3,$E6,$7E,$37,$D3,$B3,$F8,$C7,$AF,$C3,$63
	.dc.b	$2F,$F7,$2E,$68,$C7,$5F,$F0,$68,$7D,$0F,$0C,$3D,$0D,$BF,$F7,$6F
	.dc.b	$D4,$BB,$BF,$7C,$FF,$0D,$AF,$DE,$B9,$6F,$7F,$F0,$3E,$17,$CF,$DF
	.dc.b	$94,$C3,$6B,$F8,$2F,$D4,$3E,$A7,$86,$7F,$C2,$E3,$13,$FC,$2B,$0F
	.dc.b	$EB,$0F,$0E,$2D,$7C,$6E,$DB,$1D,$3E,$4F,$97,$CA,$73,$6C,$7F,$14
	.dc.b	$FA,$17,$C4,$FF,$19,$7F,$8A,$E6,$E4,$FE,$57,$1A,$AE,$1E,$7E,$FC
	.dc.b	$9E,$1E,$5D,$9E,$8F,$F2,$3F,$A9,$FD,$3C,$7F,$22,$5F,$47,$9F,$23
	.dc.b	$4E,$FD,$FE,$1E,$1D,$BE,$0F,$3D,$7A,$71,$D9,$E8,$FF,$31,$5F,$2B
	.dc.b	$7E,$57,$0D,$8F,$E6,$7F,$52,$F9,$FF,$E7,$0B,$E4,$2C,$61,$FE,$79
	.dc.b	$F5,$3E,$8A,$FE,$76,$BC,$4C,$ED,$A7,$7A,$F0,$F2,$AD,$F0,$7F,$A3
	.dc.b	$7D,$43,$C7,$BF,$53,$86,$D7,$F4,$89,$6F,$C3,$C6,$E2,$D9,$FE,$9B
	.dc.b	$F0,$5E,$96,$FB,$FF,$D4,$E3,$1C,$9F,$6E,$7F,$C2,$51,$FD,$8A,$30
	.dc.b	$7F,$88,$D7,$B1,$C7,$EA,$79,$36,$BF,$AD,$FD,$4F,$C1,$6F,$9F,$FD
	.dc.b	$86,$31,$87,$40,$FE,$F6,$78,$F5,$6F,$8F,$83,$63,$AF,$AF,$86,$9C
	.dc.b	$0F,$0F,$93,$DF,$CA,$F0,$36,$3C,$BE,$B6,$1A,$70,$3D,$3F,$13,$FD
	.dc.b	$DF,$93,$E2,$F8,$BF,$E0,$8D,$3F,$C9,$D5,$FD,$E2,$72,$FF,$8E,$61
	.dc.b	$DF,$EF,$BF,$80,$D9,$FE,$FF,$6F,$23,$B7,$8B,$FE,$05,$8C,$4F,$58
	.dc.b	$C7,$91,$9A,$B3,$C9,$FF,$06,$70,$BA,$FD,$1F,$0D,$86,$D7,$F8,$3A
	.dc.b	$5D,$B9,$8C,$36,$3E,$A1,$FE,$A0,$8C,$9C,$7C,$4E,$F3,$E0,$D7,$E3
	.dc.b	$7C,$35,$AF,$88,$F9,$36,$C6,$1D,$1B,$B7,$F8,$CA,$3F,$CD,$DE,$4F
	.dc.b	$F9,$A1,$34,$8F,$F9,$84,$7D,$3F,$EE,$BE,$A7,$11,$8F,$21,$18,$9D
	.dc.b	$4F,$16,$BF,$E6,$B5,$F3,$BF,$C6,$E3,$4D,$74,$D7,$FC,$ED,$1C,$30
	.dc.b	$ED,$F3,$9F,$E7,$97,$FC,$9B,$08,$C6,$9C,$1B,$5F,$E8,$D9,$E4,$BE
	.dc.b	$A7,$FC,$46,$E7,$8F,$6F,$C0,$6C,$FE,$26,$3F,$C9,$2F,$76,$F2,$68
	.dc.b	$8F,$F9,$3B,$EA,$7F,$D1,$59,$79,$BE,$77,$99,$27,$89,$D3,$E7,$BB
	.dc.b	$91,$EF,$1E,$4F,$7F,$FD,$C3,$1F,$BB,$6D,$7F,$9B,$9B,$AF,$FA,$77
	.dc.b	$E4,$FF,$AD,$57,$F9,$82,$7F,$D7,$0D,$E5,$E8,$75,$E7,$F8,$3F,$53
	.dc.b	$ED,$9F,$E1,$0D,$C7,$FC,$F6,$DC,$7B,$7C,$9F,$0E,$09,$7F,$E6,$EA
	.dc.b	$F9,$0F,$2E,$2F,$9B,$F5,$3F,$CF,$63,$FC,$E6,$D7,$F8,$B3,$71,$FF
	.dc.b	$44,$BF,$E8,$C8,$DF,$A5,$C8,$8F,$2F,$4B,$AF,$7C,$8E,$AF,$83,$E3
	.dc.b	$FC,$3E,$C4,$9F,$0B,$B7,$C3,$F1,$BE,$62,$3E,$6F,$A8,$FC,$2F,$ED
	.dc.b	$5F,$CD,$23,$FD,$16,$BC,$FC,$9A,$F3,$FF,$17,$F4,$5F,$EA,$1A,$FF
	.dc.b	$48,$AF,$13,$26,$BE,$07,$F2,$3E,$C7,$F7,$4A,$FF,$4B,$B5,$FE,$74
	.dc.b	$DE,$2F,$F5,$6B,$FD,$86,$3C,$7E,$EF,$98,$8F,$9B,$DD,$F1,$F4,$C8
	.dc.b	$F1,$BE,$6F,$83,$F0,$F9,$FE,$2F,$42,$57,$8A,$3E,$0E,$92,$3E,$0B
	.dc.b	$F1,$3C,$EF,$D4,$FC,$FF,$FA,$16,$7F,$D5,$ED,$7F,$A9,$37,$5F,$F6
	.dc.b	$2A,$F8,$CF,$83,$E7,$FC,$2F,$42,$5F,$C2,$8F,$A3,$A7,$D2,$E1,$93
	.dc.b	$F4,$A3,$C3,$EB,$F8,$BE,$2A,$4F,$8A,$3F,$7B,$F7,$CF,$12,$92,$7F
	.dc.b	$D8,$B1,$FA,$87,$FB,$47,$97,$E5,$7F,$79,$A7,$F0,$68,$C4,$7F,$6D
	.dc.b	$EA,$FF,$6F,$EF,$EC,$F6,$7E,$19,$27,$E1,$F4,$78,$BD,$6F,$1F,$BF
	.dc.b	$27,$C3,$1D,$76,$ED,$F1,$92,$7F,$86,$47,$CE,$F6,$BE,$7F,$B7,$27
	.dc.b	$E7,$C7,$BB,$C9,$91,$EC,$7E,$0F,$AB,$E9,$E3,$FE,$3F,$F9,$37,$F6
	.dc.b	$AA,$FF,$72,$B5,$F8,$31,$AF,$6B,$71,$0D,$B5,$FF,$07,$4F,$F2,$0E
	.dc.b	$F4,$48,$C3,$67,$F9,$7F,$6B,$E1,$69,$27,$FD,$EF,$0F,$86,$DF,$C6
	.dc.b	$49,$CB,$B7,$FC,$3D,$3F,$E3,$B7,$8F,$0F,$63,$F9,$EE,$6F,$9D,$F0
	.dc.b	$FC,$6C,$AE,$EE,$DF,$0E,$DE,$62,$3E,$CE,$76,$FE,$E8,$7C,$B3,$E9
	.dc.b	$7D,$BC,$52,$7D,$6F,$27,$FC,$69,$2E,$7F,$EA,$2D,$8E,$D7,$D6,$31
	.dc.b	$1E,$EA,$F9,$3E,$DC,$8F,$8E,$FC,$CF,$6F,$E6,$FC,$C4,$9F,$96,$ED
	.dc.b	$FF,$22,$7D,$27,$F2,$D6,$FC,$3F,$F6,$28,$C6,$BE,$E9,$FB,$7E,$73
	.dc.b	$E9,$35,$EB,$C9,$F3,$9F,$80,$BF,$E7,$5B,$7C,$57,$9B,$7D,$7E,$C4
	.dc.b	$9F,$E0,$FC,$9E,$1F,$5A,$47,$41,$7F,$EA,$67,$0A,$B7,$1F,$C3,$BE
	.dc.b	$1F,$F8,$99,$8D,$3D,$51,$CB,$E6,$3D,$C4,$79,$FC,$CD,$7E,$2E,$BF
	.dc.b	$0C,$C5,$72,$C5,$EA,$7C,$7F,$56,$5F,$92,$3F,$C1,$7F,$49,$FC,$1A
	.dc.b	$FC,$2D,$DF,$F1,$0D,$AF,$91,$7C,$3C,$8F,$A5,$E7,$C9,$F8,$55,$FC
	.dc.b	$FD,$FB,$37,$F5,$C8,$FC,$49,$AF,$F8,$EF,$2F,$F2,$13,$F4,$E7,$FC
	.dc.b	$76,$FC,$C6,$FD,$1F,$FC,$01,$8F,$F1,$FB,$5F,$F2,$23,$5F,$D0,$07
	.dc.b	$0F,$E8,$DF,$33,$A6,$B9,$C3,$B9,$E0,$E2,$FF,$18,$CF,$ED,$67,$F0
	.dc.b	$0D,$7F,$D3,$4B,$C7,$DC,$F1,$9F,$3B,$5F,$3F,$CE,$F4,$23,$D3,$E6
	.dc.b	$6B,$D2,$2E,$DF,$86,$C1,$58,$CB,$D9,$56,$FF,$8E,$D6,$9F,$3C,$65
	.dc.b	$F8,$8F,$B2,$B1,$28,$B1,$86,$CF,$22,$1E,$53,$D0,$62,$FC,$D1,$C2
	.dc.b	$34,$48,$CB,$C0,$C5,$FA,$9E,$4F,$D7,$AF,$14,$60,$F6,$B8,$78,$98
	.dc.b	$34,$7B,$8B,$7F,$0D,$D9,$92,$B9,$AF,$2C,$62,$FE,$33,$C9,$F6,$46
	.dc.b	$89,$7E,$99,$FC,$1D,$F9,$67,$31,$E5,$3A,$1E,$48,$D4,$F1,$9F,$BD
	.dc.b	$18,$9F,$B7,$5E,$27,$89,$E8,$8D,$3E,$B9,$E2,$F9,$CE,$C7,$D8,$3F
	.dc.b	$41,$3F,$AC,$23,$FA,$C6,$BF,$96,$30,$43,$FD,$F1,$E1,$3D,$8F,$EB
	.dc.b	$27,$89,$C4,$7F,$25,$6B,$E8,$AF,$E3,$F1,$86,$CF,$13,$9C,$FA,$4E
	.dc.b	$EE,$1E,$58,$7F,$5D,$AE,$4D,$9E,$53,$E1,$6B,$C8,$E9,$5E,$07,$46
	.dc.b	$27,$93,$83,$B9,$FE,$08,$AE,$ED,$AE,$B5,$CE,$36,$27,$6D,$9D,$D7
	.dc.b	$68,$DC,$E8,$DE,$3F,$B3,$8F,$CD,$C6,$6B,$8C,$78,$4F,$41,$4F,$71
	.dc.b	$D1,$F9,$17,$F9,$28,$FC,$35,$F8,$63,$D0,$66,$FC,$13,$CA,$33,$3F
	.dc.b	$95,$24,$77,$8C,$BF,$96,$3C,$F5,$E0,$9D,$C7,$80,$E0,$F8,$4C,$FC
	.dc.b	$F3,$D9,$7C,$7F,$FB,$F2,$30,$DF,$23,$FC,$85,$7F,$E1,$D9,$CB,$FD
	.dc.b	$00,$FA,$1A,$FF,$4C,$6E,$19,$69,$BD,$3F,$95,$C6,$7D,$7F,$0B,$71
	.dc.b	$E5,$18,$AE,$CD,$7F,$A7,$FD,$2B,$CB,$F0,$36,$FC,$0D,$9F,$33,$5F
	.dc.b	$EA,$FF,$3E,$F8,$2F,$43,$6B,$D0,$D9,$F7,$35,$E3,$DF,$C0,$93,$C1
	.dc.b	$F8,$5E,$EF,$ED,$E3,$E1,$36,$7E,$43,$5F,$ED,$2D,$7F,$C9,$54,$F8
	.dc.b	$2F,$29,$2B,$CA,$FA,$3B,$FF,$C6,$9C,$9F,$E1,$CA,$46,$2B,$9F,$81
	.dc.b	$70,$FF,$39,$D6,$A7,$FC,$58,$BE,$03,$52,$8C,$4E,$BC,$0C,$C6,$25
	.dc.b	$8F,$F8,$B3,$E7,$1F,$5D,$AB,$E7,$18,$C3,$91,$FE,$3B,$5F,$17,$A3
	.dc.b	$AF,$1D,$9F,$2F,$83,$86,$9E,$88,$F3,$FD,$9B,$C7,$F4,$FC,$FB,$CB
	.dc.b	$CC,$61,$F6,$FD,$03,$DC,$BE,$EE,$1B,$97,$F9,$3E,$7F,$86,$36,$38
	.dc.b	$F6,$EE,$35,$EE,$F8,$1E,$AD,$79,$78,$F8,$69,$C0,$ED,$E2,$7C,$6F
	.dc.b	$81,$F2,$B6,$3C,$5E,$16,$1A,$70,$3F,$CC,$9B,$BE,$1F,$DB,$3F,$11
	.dc.b	$B3,$EB,$F8,$6F,$83,$C9,$F0,$9B,$1E,$FF,$86,$C3,$4E,$07,$C3,$F8
	.dc.b	$0F,$B7,$BF,$CA,$6C,$7C,$7E,$BC,$34,$E0,$7F,$9E,$7E,$83,$FD,$7C
	.dc.b	$FE,$07,$87,$F1,$F1,$89,$FC,$5C,$FF,$AA,$FA,$7E,$8C,$60,$FF,$65
	.dc.b	$CF,$EF,$7B,$3F,$CF,$6F,$16,$F8,$DD,$3E,$07,$F0,$BD,$9F,$86,$EC
	.dc.b	$B9,$7C,$87,$F3,$F5,$F8,$98,$FF,$47,$3F,$E8,$B6,$F5,$EB,$C5,$FD
	.dc.b	$2F,$DD,$FE,$EF,$2F,$F4,$5A,$F8,$9C,$F4,$69,$DF,$C8,$4C,$7F,$D3
	.dc.b	$54,$F9,$0C,$70,$7B,$47,$E3,$0F,$8C,$FD,$9B,$6F,$FD,$44,$BE,$87
	.dc.b	$33,$FB,$8F,$BC,$FE,$11,$5F,$EA,$76,$DF,$55,$F1,$FB,$F3,$78,$7A
	.dc.b	$9C,$7D,$BB,$87,$17,$F8,$1F,$E1,$4F,$6A,$F6,$61,$B5,$FE,$B5,$FB
	.dc.b	$D7,$CF,$7F,$0A,$5F,$3F,$63,$0E,$7E,$12,$57,$1F,$86,$BC,$1F,$0F
	.dc.b	$F1,$6F,$E2,$CD,$F5,$FF,$8A,$C7,$B3,$AF,$FD,$76,$E9,$FF,$5E,$F9
	.dc.b	$DF,$37,$0F,$F6,$43,$6F,$E7,$2F,$1F,$F8,$FB,$9A,$FF,$62,$FD,$F7
	.dc.b	$E3,$6F,$FD,$F8,$FE,$35,$BE,$0F,$F2,$68,$C6,$03,$F8,$97,$B9,$FC
	.dc.b	$AB,$C2,$FE,$B5,$87,$FB,$41,$7D,$C6,$67,$F9,$7F,$DF,$3E,$0E,$98
	.dc.b	$F4,$EE,$9F,$EE,$19,$79,$DC,$3C,$FE,$2D,$9F,$F7,$37,$E2,$8F,$9C
	.dc.b	$FE,$76,$BE,$36,$C6,$1F,$E7,$F8,$FD,$DA,$FF,$77,$3F,$B0,$DB,$FD
	.dc.b	$CE,$E7,$7F,$CB,$39,$9F,$81,$1F,$EB,$3F,$53,$99,$D3,$CB,$B3,$FE
	.dc.b	$F9,$FA,$4E,$58,$7F,$BC,$9E,$57,$F2,$5C,$CF,$47,$30,$5C,$BF,$D5
	.dc.b	$97,$CB,$D8,$C3,$9C,$B1,$7E,$16,$DF,$EB,$EF,$F7,$13,$E5,$F8,$CE
	.dc.b	$9E,$77,$57,$B6,$78,$9F,$53,$6F,$97,$9D,$FD,$A3,$E5,$5F,$1D,$E5
	.dc.b	$8C,$3E,$7F,$5F,$18,$7C,$36,$DF,$F0,$1D,$FE,$C4,$FE,$2D,$3F,$3B
	.dc.b	$0F,$E2,$93,$BE,$2F,$07,$A5,$FD,$E2,$DF,$3F,$67,$0D,$C3,$A1,$FE
	.dc.b	$F8,$E6,$F4,$79,$7F,$E0,$0C,$7F,$1B,$3A,$7F,$39,$97,$F9,$1E,$31
	.dc.b	$1F,$9F,$F0,$38,$5F,$16,$CE,$A7,$F5,$8F,$E1,$9F,$D8,$39,$F4,$B8
	.dc.b	$3B,$DF,$E5,$7C,$57,$8D,$B1,$CF,$7C,$34,$E0,$77,$B7,$D2,$7F,$94
	.dc.b	$5F,$DA,$35,$C9,$97,$FC,$AD,$F4,$2F,$A4,$7D,$26,$1B,$5F,$E5,$F9
	.dc.b	$D4,$E4,$E3,$F3,$3F,$E2,$29,$D3,$5F,$81,$28,$C4,$7E,$A2,$31,$C7
	.dc.b	$E2,$EE,$FD,$5B,$F8,$09,$FE,$19,$3F,$B9,$7E,$8B,$5F,$E9,$0F,$25
	.dc.b	$F4,$9F,$E2,$C7,$8F,$2F,$4D,$3C,$0D,$8C,$9D,$30,$C2,$BF,$C5,$BE
	.dc.b	$0E,$5C,$C6,$23,$FA,$C7,$CF,$FD,$DC,$BD,$11,$C7,$E6,$36,$47,$91
	.dc.b	$EA,$F3,$79,$91,$F6,$C7,$8B,$C4,$FE,$CD,$1F,$90,$6B,$FD,$9D,$B5
	.dc.b	$E0,$FA,$4F,$EC,$93,$BE,$4E,$44,$79,$73,$7B,$BE,$93,$E5,$5F,$A1
	.dc.b	$BC,$1F,$C7,$1B,$5F,$CF,$70,$FE,$E8,$FE,$8A,$3F,$97,$37,$83,$A1
	.dc.b	$D7,$83,$E5,$FD,$27,$F0,$49,$FD,$4B,$5F,$F0,$03,$71,$F9,$2F,$F9
	.dc.b	$66,$37,$F4,$9C,$88,$F2,$EF,$EB,$DF,$23,$AB,$E0,$F9,$4F,$0F,$9D
	.dc.b	$27,$C2,$ED,$F0,$FB,$7E,$62,$3E,$6F,$73,$F0,$7F,$47,$FE,$85,$8F
	.dc.b	$E0,$2F,$F9,$3D,$BF,$13,$AF,$93,$FC,$4F,$CE,$BE,$1A,$F8,$92,$FF
	.dc.b	$84,$CE,$5F,$E1,$7E,$F7,$F3,$A6,$F9,$FF,$92,$8C,$47,$F8,$95,$FF
	.dc.b	$09,$3E,$2F,$81,$F2,$91,$F2,$FC,$97,$8B,$AA,$47,$81,$F2,$FC,$47
	.dc.b	$5F,$3B,$B7,$A1,$2B,$C7,$1F,$03,$49,$1F,$01,$F8,$7E,$7F,$E9,$BE
	.dc.b	$59,$F2,$34,$95,$FC,$9A,$1F,$21,$EB,$F4,$3D,$9E,$8C,$AF,$83,$1F
	.dc.b	$3B,$4F,$9F,$C7,$27,$E7,$C7,$7F,$85,$F0,$7C,$7C,$9E,$D1,$FB,$8F
	.dc.b	$A4,$FC,$AC,$7F,$92,$DB,$E1,$7E,$D1,$BC,$5F,$DE,$1F,$D3,$CF,$EE
	.dc.b	$B4,$FE,$03,$E6,$7A,$23,$F9,$FB,$5F,$F2,$B7,$5F,$EF,$B0,$FD,$FE
	.dc.b	$FF,$59,$BD,$1F,$14,$F8,$DD,$1E,$5F,$57,$CD,$F4,$13,$1F,$E9,$7A
	.dc.b	$27,$C3,$BF,$8B,$92,$57,$8A,$3E,$87,$B3,$F4,$7D,$A9,$3F,$3E,$3E
	.dc.b	$17,$91,$23,$AB,$F1,$7D,$4F,$6F,$27,$F1,$EF,$A4,$7B,$34,$CD,$FF
	.dc.b	$5E,$8C,$B4,$E0,$6C,$C1,$0F,$F3,$D3,$FC,$8F,$9B,$FC,$4A,$F2,$8F
	.dc.b	$F3,$F7,$63,$E4,$3F,$97,$B6,$3F,$B6,$B5,$DF,$D2,$6B,$E1,$64,$69
	.dc.b	$E3,$5C,$B8,$64,$F2,$7F,$9D,$E7,$FE,$4D,$8E,$BF,$49,$E0,$23,$B3
	.dc.b	$C3,$AF,$90,$8F,$F8,$03,$E8,$DF,$49,$E8,$E0,$93,$E8,$76,$FF,$A6
	.dc.b	$E5,$CD,$FD,$31,$B1,$CE,$38,$BF,$13,$D9,$F8,$64,$7E,$1F,$B3,$F1
	.dc.b	$BE,$2E,$47,$C6,$74,$FF,$AA,$7E,$8B,$F9,$43,$7E,$6B,$FA,$DC,$63
	.dc.b	$BB,$FA,$AF,$98,$FA,$4C,$FA,$B3,$0F,$F0,$ED,$FF,$33,$E9,$FF,$2E
	.dc.b	$BC,$5B,$EA,$F4,$24,$FE,$FF,$E2,$7B,$FA,$B2,$39,$BA,$7C,$CF,$F0
	.dc.b	$9F,$85,$91,$A3,$C3,$F3,$FE,$82,$3E,$4F,$8F,$AF,$BF,$4F,$69,$8D
	.dc.b	$BF,$B8,$E8,$F4,$FE,$0F,$A9,$2F,$DB,$15,$FF,$68,$7D,$27,$EF,$6F
	.dc.b	$FB,$CB,$77,$F8,$37,$FE,$04,$8F,$E6,$35,$F3,$FC,$FF,$9A,$67,$CA
	.dc.b	$B1,$F2,$AF,$E8,$98,$7E,$B3,$87,$47,$D0,$9F,$F0,$8F,$D1,$BB,$E9
	.dc.b	$27,$FC,$A3,$0F,$27,$11,$FE,$E9,$8F,$2E,$7F,$E1,$26,$E1,$FE,$96
	.dc.b	$7A,$4F,$2F,$23,$FC,$8D,$06,$92,$3F,$CB,$07,$FA,$F6,$3F,$54,$F7
	.dc.b	$E4,$F3,$9E,$89,$F4,$FC,$5C,$9F,$3B,$FD,$1F,$E0,$F1,$1E,$78,$B8
	.dc.b	$FF,$2A,$9B,$8F,$9E,$6D,$7F,$B3,$7C,$F5,$62,$19,$2B,$15,$A9,$F0
	.dc.b	$90,$F0,$1F,$D5,$6B,$95,$71,$8D,$61,$E3,$6E,$1E,$36,$9E,$3A,$E3
	.dc.b	$87,$81,$B1,$A1,$87,$8D,$F0,$92,$FE,$B5,$7C,$58,$E1,$87,$8D,$D6
	.dc.b	$1F,$EC,$EB,$E5,$7C,$27,$77,$C5,$D9,$F5,$DF,$23,$A3,$C9,$3C,$AB
	.dc.b	$D9,$1C,$0F,$97,$3C,$EB,$B4,$66,$FC,$91,$F1,$63,$13,$D9,$A7,$6C
	.dc.b	$3E,$1B,$D7,$0F,$03,$E5,$3C,$82,$BA,$1E,$02,$61,$E6,$72,$D2,$47
	.dc.b	$33,$93,$F1,$EB,$C3,$3C,$A7,$EC,$1E,$07,$0F,$D6,$8F,$CA,$BF,$EB
	.dc.b	$7C,$3E,$93,$70,$FD,$39,$1F,$D2,$9F,$20,$B1,$C1,$F9,$A7,$13,$F7
	.dc.b	$3F,$49,$FB,$E3,$CB,$87,$ED,$5C,$47,$96,$30,$F0,$30,$7E,$CE,$36
	.dc.b	$FD,$03,$F8,$13,$8C,$3D,$B5,$F6,$2B,$9D,$68,$E6,$7E,$AA,$B8,$47
	.dc.b	$CC,$8F,$61,$B2,$57,$4B,$DC,$78,$CE,$76,$BC,$C7,$2A,$F1,$C6,$27
	.dc.b	$C9,$3F,$66,$7F,$8C,$4E,$6B,$F6,$AD,$9F,$D2,$B5,$FC,$A3,$67,$AC
	.dc.b	$7C,$73,$E2,$3D,$1C,$3E,$96,$DF,$C4,$8C,$47,$BA,$30,$7C,$3B,$F3
	.dc.b	$C7,$C7,$38,$9E,$98,$D5,$78,$DB,$1A,$3E,$13,$87,$B1,$C4,$87,$53
	.dc.b	$27,$00,$3E,$B1,$EC,$7F,$BF,$DB,$1F,$94,$6B,$FC,$E9,$AF,$23,$9F
	.dc.b	$D2,$F2,$F9,$1F,$CD,$B6,$3F,$A0,$AF,$3B,$FC,$75,$1C,$63,$4F,$33
	.dc.b	$FA,$C6,$3F,$EE,$27,$B7,$17,$E0,$F2,$FD,$3B,$63,$FA,$6B,$5E,$5F
	.dc.b	$48,$F0,$B9,$47,$C0,$8C,$57,$EC,$1B,$1F,$D5,$9A,$F2,$F8,$24,$FF
	.dc.b	$8E,$7C,$4B,$3F,$C1,$8E,$93,$8E,$0F,$F7,$CF,$90,$AC,$5F,$C7,$C1
	.dc.b	$78,$C3,$8C,$FD,$4C,$3F,$E0,$0D,$3C,$3D,$52,$38,$D6,$CF,$A5,$F4
	.dc.b	$1F,$D6,$5C,$4F,$DE,$8C,$3E,$73,$FA,$5F,$BD,$D2,$33,$87,$B2,$7F
	.dc.b	$CD,$9D,$FE,$0F,$43,$FC,$45,$87,$49,$F1,$7C,$1E,$6B,$CB,$93,$C4
	.dc.b	$A6,$3C,$5B,$87,$19,$7C,$0F,$03,$8D,$D9,$D3,$87,$C1,$A3,$E9,$31
	.dc.b	$9F,$AE,$1F,$17,$EB,$3C,$7F,$01,$6F,$C0,$E1,$F1,$78,$96,$FE,$27
	.dc.b	$D2,$7E,$0B,$87,$E1,$36,$BF,$3D,$F4,$6F,$37,$E9,$CE,$6F,$F5,$7F
	.dc.b	$49,$FE,$DA,$BF,$8C,$D8,$FA,$0F,$F8,$BE,$33,$EF,$7D,$C8,$FF,$1B
	.dc.b	$47,$0F,$27,$1D,$B9,$75,$EF,$78,$26,$9F,$4F,$1F,$CA,$BE,$0F,$27
	.dc.b	$39,$78,$1B,$AF,$DD,$C9,$D5,$F8,$1F,$83,$F1,$37,$0F,$4F,$91,$86
	.dc.b	$9E,$C8,$FC,$9F,$03,$E1,$FD,$55,$86,$D7,$C2,$F8,$6F,$77,$2F,$BD
	.dc.b	$B1,$F1,$BE,$1B,$0D,$38,$1F,$2B,$E0,$BF,$2D,$C5,$F1,$DB,$1F,$3B
	.dc.b	$83,$0D,$38,$1F,$97,$FA,$4F,$F6,$33,$F8,$AE,$4F,$D5,$C6,$27,$F1
	.dc.b	$51,$8F,$81,$CF,$85,$6F,$93,$F0,$3A,$7E,$0F,$55,$9F,$D2,$70,$A3
	.dc.b	$F6,$F7,$F8,$79,$FD,$43,$FB,$B6,$F5,$6B,$C3,$FD,$EF,$D2,$7D,$49
	.dc.b	$FD,$72,$F8,$9C,$F3,$69,$DF,$A3,$2B,$7F,$21,$78,$3F,$7D,$F4,$8F
	.dc.b	$A3,$FE,$12,$DA,$FD,$D2,$FA,$4C,$CF,$F0,$EF,$A4,$79,$74,$C7,$C7
	.dc.b	$DC,$3F,$81,$4F,$F4,$86,$FB,$9C,$7D,$7E,$4F,$F1,$8F,$A4,$7B,$97
	.dc.b	$AB,$0D,$BF,$E1,$9F,$48,$FA,$4F,$E3,$EB,$E9,$2C,$61,$CF,$AB,$2B
	.dc.b	$8F,$C3,$DE,$3F,$E3,$DF,$49,$FC,$4E,$7F,$8C,$36,$33,$C3,$FC,$B2
	.dc.b	$31,$E5,$FF,$21,$F3,$9C,$FF,$CC,$1B,$87,$F2,$15,$E3,$66,$7F,$9A
	.dc.b	$FD,$23,$C0,$BE,$26,$1B,$87,$F2,$9F,$05,$EF,$7F,$3C,$6C,$67,$E9
	.dc.b	$31,$3F,$D0,$3E,$91,$F0,$94,$C7,$A5,$6F,$F9,$A4,$BC,$DE,$1E,$77
	.dc.b	$16,$CF,$F3,$9F,$A4,$7C,$DF,$F4,$C5,$F1,$F6,$30,$FF,$50,$C7,$EC
	.dc.b	$D7,$F4,$17,$F8,$BF,$3F,$F4,$B7,$2D,$EE,$E1,$E9,$CA,$E1,$E5,$5A
	.dc.b	$FE,$A5,$F4,$9C,$3A,$7F,$A8,$3D,$1B,$E0,$FF,$60,$8C,$5F,$F8,$3D
	.dc.b	$75,$67,$E6,$34,$D9,$FE,$AD,$3F,$ED,$36,$C7,$3D,$37,$1A,$F3,$3B
	.dc.b	$7C,$6F,$76,$C7,$2D,$30,$D3,$81,$DB,$CD,$FE,$EB,$F2,$AF,$8F,$F4
	.dc.b	$46,$1F,$57,$AD,$8A,$F2,$B7,$83,$FB,$A7,$07,$D4,$9F,$ED,$D8,$7F
	.dc.b	$07,$9C,$F2,$76,$F4,$BF,$C0,$76,$FA,$0B,$38,$6E,$1F,$DD,$EF,$F0
	.dc.b	$8F,$4F,$37,$FC,$17,$18,$3F,$C0,$98,$FE,$ED,$FF,$03,$1F,$9E,$9E
	.dc.b	$5D,$3C,$3C,$4E,$2F,$2F,$13,$13,$DD,$B5,$CB,$7F,$8B,$F1,$5E,$06
	.dc.b	$D7,$97,$7C,$34,$F4,$47,$9A,$9F,$49,$FE,$A6,$7F,$CB,$2D,$72,$65
	.dc.b	$FF,$55,$FD,$0B,$E9,$1F,$49,$86,$D7,$F9,$3C,$4E,$E7,$47,$2F,$99
	.dc.b	$FF,$66,$4E,$E5,$18,$8F,$F2,$C4,$63,$8F,$81,$E2,$F0,$3F,$CC,$CF
	.dc.b	$F8,$B6,$3F,$5F,$3F,$E2,$B6,$BF,$E3,$57,$92,$FA,$4F,$F0,$EB,$FE
	.dc.b	$78,$8F,$DC,$BC,$79,$7A,$EB,$FE,$73,$9C,$CF,$3C,$32,$F1,$E5,$AF
	.dc.b	$C0,$6C,$FE,$2E,$3F,$C7,$DF,$41,$FE,$73,$CB,$D1,$F3,$3A,$24,$FA
	.dc.b	$6F,$8F,$A1,$23,$D8,$E9,$FF,$28,$78,$DF,$CB,$1F,$F2,$53,$5C,$F2
	.dc.b	$F9,$3E,$93,$FA,$A4,$6B,$93,$CF,$5F,$F0,$02,$3F,$C7,$F1,$8A,$F5
	.dc.b	$FD,$27,$CA,$CF,$FC,$B4,$DD,$FF,$D2,$CD,$BF,$F5,$0E,$3F,$D9,$5F
	.dc.b	$F2,$E4,$7C,$8A,$E3,$C1,$D0,$EB,$C1,$F2,$3E,$93,$F7,$D7,$FE,$63
	.dc.b	$6B,$FE,$6D,$6D,$7F,$9A,$DF,$F4,$94,$6F,$E9,$39,$11,$E5,$DF,$D7
	.dc.b	$BE,$47,$57,$C1,$F2,$5E,$1F,$42,$4F,$85,$DB,$E1,$F7,$7C,$DE,$B4
	.dc.b	$8F,$34,$7E,$03,$1F,$EA,$31,$FE,$7B,$6B,$FE,$85,$6E,$7F,$E8,$35
	.dc.b	$E9,$7E,$23,$E4,$F3,$C5,$FE,$C1,$6D,$FF,$B3,$1A,$9E,$2F,$FA,$27
	.dc.b	$EC,$1F,$41,$F9,$27,$FB,$44,$67,$D6,$C4,$FE,$5B,$E9,$7E,$5E,$1F
	.dc.b	$EC,$F7,$E5,$D7,$FA,$6D,$B9,$FF,$A6,$5D,$C7,$FB,$7B,$E8,$2A,$77
	.dc.b	$5F,$ED,$96,$C6,$F8,$BC,$9E,$1C,$8F,$43,$E7,$F7,$BC,$9F,$4D,$E5
	.dc.b	$E0,$95,$FE,$A5,$C9,$3F,$AE,$C6,$CF,$FA,$A5,$56,$4D,$4F,$C3,$EE
	.dc.b	$95,$F1,$DF,$8D,$E0,$7C,$3E,$79,$3D,$DF,$8D,$CF,$DD,$91,$DD,$F1
	.dc.b	$47,$F0,$0F,$C2,$3C,$33,$FE,$BB,$7E,$29,$31,$FA,$1C,$DF,$47,$A2
	.dc.b	$57,$B6,$3D,$87,$B6,$24,$7B,$63,$B7,$44,$8E,$CF,$86,$3E,$E7,$CC
	.dc.b	$C9,$F9,$0F,$DF,$9F,$90,$C8,$F9,$0F,$CB,$8F,$57,$CB,$7A,$FE,$5E
	.dc.b	$57,$AE,$3F,$8E,$FD,$37,$F2,$88,$FF,$70,$37,$9F,$FD,$BC,$DE,$7F
	.dc.b	$C6,$7C,$D7,$0F,$3E,$5F,$FB,$7E,$BF,$C5,$8D,$F1,$B2,$FE,$47,$E8
	.dc.b	$FF,$2C,$63,$1E,$BE,$7F,$8A,$61,$CD,$AF,$90,$B1,$C1,$FF,$77,$7A
	.dc.b	$0F,$82,$FE,$6E,$D8,$CC,$63,$E9,$B2,$EB,$C1,$93,$A3,$FD,$F1,$EF
	.dc.b	$FF,$92,$9E,$BE,$AF,$43,$FA,$1C,$66,$18,$B8,$11,$32,$7D,$37,$90
	.dc.b	$91,$D0,$F4,$BD,$3D,$99,$18,$3D,$BE,$A7,$B3,$CB,$91,$83,$BE,$3E
	.dc.b	$17,$F8,$67,$D3,$7F,$11,$23,$F1,$A7,$E2,$F8,$32,$3E,$4B,$F2,$3C
	.dc.b	$1C,$47,$B3,$4F,$5A,$D1,$F1,$1F,$38,$DE,$5E,$19,$FE,$11,$C9,$FD
	.dc.b	$A3,$E9,$BF,$83,$C1,$BE,$2A,$75,$E4,$CA,$F5,$E5,$E0,$BF,$E2,$67
	.dc.b	$23,$FA,$4C,$79,$0E,$93,$FE,$1F,$8C,$C7,$18,$C7,$8B,$E3,$F0,$A4
	.dc.b	$F6,$3E,$BD,$27,$31,$FE,$1F,$84,$F5,$24,$7F,$88,$CF,$67,$6F,$84
	.dc.b	$8E,$65,$1E,$5C,$64,$E5,$AF,$7E,$EF,$51,$1E,$4F,$B7,$E3,$64,$75
	.dc.b	$6B,$FE,$35,$FA,$67,$CD,$7F,$86,$1B,$1E,$F0,$E1,$FB,$C7,$C2,$9F
	.dc.b	$B1,$AC,$8E,$19,$7D,$DD,$FE,$F2,$F9,$7D,$4F,$C7,$AF,$A4,$47,$E3
	.dc.b	$7D,$2B,$AF,$FC,$56,$FF,$B7,$A3,$C8,$D7,$C6,$1E,$0E,$BE,$1F,$FC
	.dc.b	$B9,$1B,$8F,$C1,$35,$FC,$BB,$5F,$D9,$BF,$B7,$C7,$F2,$CD,$BD,$FC
	.dc.b	$5F,$93,$E1,$7C,$AC,$BF,$2B,$96,$90,$F3,$D7,$37,$BC,$F1,$73,$36
	.dc.b	$FC,$33,$6F,$C3,$5C,$E7,$2D,$8C,$BC,$8F,$8F,$3C,$5F,$6D,$57,$E4
	.dc.b	$45,$BF,$24,$7C,$B9,$F5,$3C,$C9,$7F,$12,$32,$D8,$F1,$1D,$8F,$09
	.dc.b	$83,$A1,$65,$5E,$1A,$F1,$36,$74,$D7,$C3,$7F,$62,$31,$7E,$19,$F9
	.dc.b	$B1,$89,$F8,$73,$F1,$A7,$89,$E5,$29,$F0,$8F,$25,$78,$9A,$78,$8A
	.dc.b	$70,$BF,$4B,$89,$F1,$1D,$CB,$5E,$22,$47,$E0,$0E,$0E,$1F,$11,$D6
	.dc.b	$BC,$4F,$D1,$3F,$32,$FE,$70,$8F,$EA,$1A,$FF,$AC,$6B,$FA,$83,$C0
	.dc.b	$D1,$FB,$DF,$51,$E0,$7F,$94,$4F,$91,$1F,$E4,$F7,$0F,$98,$C1,$C5
	.dc.b	$DB,$FE,$53,$76,$FA,$4D,$10,$C1,$FB,$88,$F0,$B9,$C3,$E3,$BF,$55
	.dc.b	$FE,$8D,$1A,$8E,$13,$E0,$8C,$4E,$A3,$07,$89,$E7,$3F,$90,$38,$9D
	.dc.b	$F0,$FA,$17,$FA,$0B,$E7,$1F,$E7,$1B,$E7,$5B,$3C,$91,$F5,$E7,$71
	.dc.b	$F1,$0F,$6D,$F8,$DE,$07,$85,$F1,$C7,$08,$F9,$07,$CD,$29,$E6,$39
	.dc.b	$CF,$8D,$E7,$7F,$B6,$3D,$17,$EF,$9F,$5E,$16,$BC,$98,$7C,$2B,$DC
	.dc.b	$72,$9F,$E0,$58,$78,$CF,$24,$F5,$8E,$53,$E4,$AE,$51,$8A,$DB,$CB
	.dc.b	$0D,$DF,$97,$E8,$FF,$98,$93,$0F,$E7,$AF,$F6,$37,$F9,$F7,$C2,$DD
	.dc.b	$4D,$24,$7F,$69,$FA,$47,$C3,$7F,$AB,$23,$FC,$1F,$18,$F8,$ED,$77
	.dc.b	$3C,$9E,$17,$BB,$FF,$55,$78,$5F,$E1,$17,$FD,$5C,$D7,$3C,$3F,$DD
	.dc.b	$F8,$13,$99,$7C,$CB,$2C,$36,$FF,$BF,$36,$7F,$C2,$07,$09,$E5,$EA
	.dc.b	$BC,$9A,$4B,$FF,$03,$46,$5F,$F0,$47,$D0,$BC,$9F,$13,$63,$31,$83
	.dc.b	$CD,$F4,$6F,$5E,$92,$7F,$C2,$51,$97,$AF,$C7,$F4,$23,$83,$7E,$AF
	.dc.b	$42,$38,$39,$F8,$9E,$02,$38,$3B,$FA,$9E,$42,$38,$3C,$7E,$27,$A0
	.dc.b	$8E,$0F,$F1,$47,$7B,$E8,$BE,$93,$63,$31,$83,$E0,$F1,$48,$F7,$B7
	.dc.b	$8B,$E9,$FD,$13,$E9,$3E,$E3,$6B,$FC,$6C,$D9,$DC,$FE,$07,$E8,$FC
	.dc.b	$11,$F7,$1A,$FF,$93,$5A,$FF,$98,$9A,$FF,$9F,$97,$D1,$7F,$9C,$7C
	.dc.b	$3F,$F3,$14,$63,$0F,$F1,$83,$FE,$21,$6C,$ED,$FF,$36,$E1,$FE,$89
	.dc.b	$7F,$C6,$CD,$BF,$F1,$1D,$3E,$4B,$84,$7F,$A3,$A7,$58,$78,$3C,$5C
	.dc.b	$CA,$F8,$61,$E2,$9C,$C6,$27,$AC,$60,$E3,$F4,$3F,$E5,$8A,$E0,$7A
	.dc.b	$BC,$1E,$27,$2E,$4B,$1F,$E4,$D9,$7D,$1E,$63,$0D,$C3,$D7,$EE,$78
	.dc.b	$BD,$BF,$1B,$6B,$E3,$B8,$7E,$4F,$C9,$DA,$FF,$22,$F6,$7C,$D6,$F0
	.dc.b	$FB,$3C,$3F,$83,$39,$E0,$FA,$75,$9C,$3E,$9B,$4F,$F2,$7E,$9F,$E9
	.dc.b	$D8,$F5,$78,$7F,$0B,$C4,$F8,$6F,$CC,$F0,$7F,$5D,$7F,$E5,$66,$E1
	.dc.b	$FE,$A1,$5E,$0E,$FC,$0F,$A7,$F4,$4F,$95,$A6,$3C,$CB,$A7,$F9,$D6
	.dc.b	$5C,$3C,$38,$B8,$B6,$7F,$CF,$5F,$2F,$FE,$CE,$6F,$17,$E0,$23,$1B
	.dc.b	$7F,$9A,$9F,$E5,$13,$FE,$81,$F9,$FF,$F4,$DB,$7D,$6F,$C3,$C6,$2B
	.dc.b	$FC,$E2,$FF,$A8,$E7,$87,$85,$A3,$7E,$1D,$8F,$F4,$67,$CF,$FC,$F7
	.dc.b	$FD,$8E,$DB,$FF,$71,$BF,$AE,$E3,$FF,$63,$39,$87,$CF,$FF,$B9,$AB
	.dc.b	$FD,$2C,$D8,$FF,$41,$BF,$E1,$C8,$CF,$C0,$7E,$7A,$1E,$ED,$F9,$4C
	.dc.b	$36,$BF,$D4,$1F,$3C,$F8,$7F,$D2,$BE,$FC,$FF,$DB,$0E,$67,$7E,$EF
	.dc.b	$EA,$BC,$77,$2D,$F0,$70,$DA,$FF,$56,$7C,$F3,$E1,$BF,$62,$D8,$CF
	.dc.b	$D0,$62,$7F,$6B,$8F,$FB,$AA,$7F,$D6,$8B,$E1,$33,$F0,$9A,$7F,$77
	.dc.b	$F3,$FF,$B7,$DB,$FD,$DC,$BA,$B3,$B6,$9D,$FD,$24,$CB,$FD,$81,$4F
	.dc.b	$73,$1C,$FF,$EA,$EF,$A0,$7C,$4F,$F0,$86,$DF,$FB,$1D,$7C,$8E,$67
	.dc.b	$F8,$6F,$A1,$F6,$49,$EA,$E3,$CE,$CF,$B5,$AF,$22,$E5,$FE,$D1,$BF
	.dc.b	$D1,$3F,$ED,$3F,$A0,$FE,$49,$A7,$FB,$01,$BC,$49,$4C,$97,$B1,$FC
	.dc.b	$79,$DC,$EA,$51,$F1,$63,$B7,$16,$F8,$F8,$3C,$71,$C4,$F8,$DB,$7E
	.dc.b	$3C,$FB,$67,$DD,$B5,$E9,$F5,$30,$D3,$D9,$1E,$5F,$9B,$FE,$67,$F0
	.dc.b	$6F,$8A,$C5,$7F,$BC,$E3,$0F,$C3,$F2,$F3,$97,$C3,$6F,$07,$FB,$C7
	.dc.b	$E6,$39,$F9,$9F,$CF,$A1,$18,$BF,$F7,$F4,$3C,$6D,$4E,$DB,$3F,$32
	.dc.b	$7F,$C5,$1E,$17,$9B,$D3,$FE,$93,$97,$E8,$2B,$CD,$18,$F3,$B3,$56
	.dc.b	$79,$1F,$46,$B8,$1F,$EF,$A1,$F3,$7B,$70,$DC,$3F,$C7,$1B,$FC,$6F
	.dc.b	$57,$C2,$DB,$EB,$7F,$77,$6C,$C6,$23,$F0,$B1,$87,$C0,$3E,$83,$FC
	.dc.b	$A8,$D3,$FC,$AC,$E0,$FF,$1D,$AF,$9E,$FC,$FF,$D0,$61,$B1,$FE,$40
	.dc.b	$6B,$EF,$6B,$F9,$17,$F5,$79,$71,$EE,$34,$93,$CF,$E1,$7F,$31,$1F
	.dc.b	$9F,$8F,$CF,$D7,$F9,$59,$F7,$35,$E7,$97,$7B,$EB,$0F,$A0,$79,$9E
	.dc.b	$3E,$4E,$B7,$6B,$9C,$61,$EF,$95,$7E,$17,$25,$CB,$F1,$51,$E2,$8D
	.dc.b	$7D,$07,$DF,$D3,$D1,$3F,$27,$D0,$E0,$47,$81,$EA,$F2,$7A,$11,$F6
	.dc.b	$C7,$87,$4F,$C3,$C7,$DD,$6B,$F0,$0F,$E0,$6E,$2B,$C3,$F4,$0F,$A2
	.dc.b	$E5,$C9,$2F,$93,$D7,$9B,$DD,$F4,$3F,$2F,$0F,$4B,$78,$7F,$85,$36
	.dc.b	$FF,$88,$E2,$F4,$38,$F1,$49,$FC,$95,$7E,$C9,$BC,$5F,$2F,$E8,$7F
	.dc.b	$80,$47,$E5,$DA,$FA,$C3,$8C,$BE,$5B,$FE,$5B,$8E,$1F,$37,$CC,$8F
	.dc.b	$3F,$03,$B7,$7C,$8E,$CF,$87,$DE,$F1,$7A,$52,$7C,$4E,$DF,$16,$1F
	.dc.b	$D9,$5F,$37,$CD,$FA,$4F,$57,$AF,$F8,$5C,$BF,$D4,$D3,$FB,$16,$BE
	.dc.b	$B6,$E7,$F8,$45,$F4,$3F,$8D,$F8,$FC,$F1,$FF,$37,$6D,$FF,$3E,$6A
	.dc.b	$78,$BF,$BF,$FA,$17,$D0,$FE,$5D,$FF,$58,$46,$7E,$17,$13,$F9,$BF
	.dc.b	$A0,$F9,$98,$7F,$45,$7E,$65,$7D,$16,$E7,$F9,$A7,$71,$FD,$43,$E7
	.dc.b	$2A,$77,$5F,$D3,$DB,$1B,$E3,$F2,$7A,$D2,$3D,$0F,$9F,$E8,$7C,$9F
	.dc.b	$11,$E5,$E0,$95,$FC,$6F,$2F,$F2,$4B,$FC,$77,$1A,$FE,$1D,$4F,$ED
	.dc.b	$DB,$1B,$AF,$89,$E0,$C8,$F9,$0F,$C7,$F0,$7E,$27,$44,$9F,$03,$F1
	.dc.b	$FA,$3A,$32,$3C,$4F,$78,$FE,$0D,$F4,$3E,$09,$FC,$D3,$F1,$89,$97
	.dc.b	$D1,$E6,$FA,$5D,$32,$BD,$D1,$E9,$3D,$51,$23,$E0,$47,$6E,$89,$1F
	.dc.b	$CF,$ED,$1E,$29,$FB,$DF,$27,$2B,$E5,$3F,$81,$9F,$8C,$C8,$F9,$8F
	.dc.b	$C9,$8F,$5F,$C3,$FB,$3E,$1E,$57,$78,$FE,$49,$F4,$3F,$CB,$23,$FA
	.dc.b	$83,$5F,$DD,$37,$9F,$F1,$5F,$29,$C3,$D1,$95,$C1,$E7,$E4,$62,$3F
	.dc.b	$25,$F3,$FF,$97,$31,$7F,$CB,$30,$7A,$3E,$2A,$87,$89,$AF,$88,$B3
	.dc.b	$C6,$3F,$9E,$7A,$4F,$A1,$A4,$AF,$ED,$51,$FE,$59,$6C,$70,$F0,$24
	.dc.b	$60,$FE,$E1,$F3,$4F,$8F,$EB,$EC,$49,$CC,$57,$F8,$F6,$0F,$8C,$F1
	.dc.b	$99,$CF,$FA,$4C,$66,$7A,$FB,$5F,$D4,$F1,$3C,$23,$CF,$EB,$7B,$3C
	.dc.b	$B9,$18,$3D,$6F,$F0,$FF,$A1,$91,$83,$DF,$DF,$F1,$3C,$09,$18,$3E
	.dc.b	$1B,$F2,$3E,$26,$46,$0F,$DA,$72,$7F,$67,$F9,$EF,$E0,$30,$6F,$88
	.dc.b	$9D,$7C,$C6,$FC,$99,$7F,$E1,$9B,$E2,$7F,$0E,$24,$70,$8D,$78,$BD
	.dc.b	$FC,$39,$18,$39,$38,$7F,$A6,$52,$3C,$D3,$E7,$ED,$95,$E1,$7D,$5E
	.dc.b	$0E,$23,$8E,$5E,$D3,$1C,$BF,$DF,$B1,$77,$FC,$12,$57,$F8,$CF,$E9
	.dc.b	$BF,$D1,$CF,$B6,$3E,$37,$D0,$48,$C1,$F0,$5F,$F1,$8B,$F2,$5B,$46
	.dc.b	$98,$FA,$09,$3F,$E3,$89,$46,$23,$FC,$A4,$7D,$3F,$37,$EB,$11,$FA
	.dc.b	$11,$FE,$1F,$FA,$07,$55,$24,$FF,$93,$61,$F1,$58,$8E,$B1,$C1,$FF
	.dc.b	$24,$B5,$CC,$FB,$21,$F0,$DD,$C9,$3F,$D6,$FC,$2A,$FF,$80,$E7,$6F
	.dc.b	$86,$7F,$AF,$4E,$B6,$FF,$17,$D6,$AB,$85,$69,$E0,$43,$27,$73,$13
	.dc.b	$CA,$3A,$65,$FE,$A9,$F4,$BF,$83,$56,$A3,$FD,$59,$7D,$EF,$84,$F3
	.dc.b	$3A,$91,$CB,$FE,$1D,$AE,$6E,$6B,$9D,$F8,$23,$FD,$7B,$3A,$AE,$24
	.dc.b	$AD,$CF,$78,$C3,$FE,$C4,$BF,$24,$EA,$3C,$18,$75,$53,$A7,$86,$1F
	.dc.b	$1D,$F2,$18,$3E,$01,$E4,$9E,$53,$C6,$39,$CF,$1B,$F1,$C6,$0E,$23
	.dc.b	$0F,$66,$1E,$59,$DD,$F3,$9C,$D7,$84,$F1,$9C,$CF,$59,$D2,$3C,$47
	.dc.b	$1C,$3C,$B8,$7B,$EB,$9D,$71,$BE,$F3,$E7,$C3,$C4,$49,$F0,$4F,$F9
	.dc.b	$EA,$B8,$E1,$DA,$FA,$61,$DA,$FD,$B5,$A9,$C4,$79,$8F,$4D,$71,$7E
	.dc.b	$C6,$1F,$55,$FF,$63,$47,$E6,$5B,$3F,$A7,$23,$FA,$C6,$BF,$B0,$6E
	.dc.b	$1F,$A2,$F0,$7F,$74,$78,$9C,$4F,$EE,$63,$17,$F9,$A7,$71,$F2,$EA
	.dc.b	$BF,$B1,$8D,$BF,$8D,$73,$5F,$2A,$3A,$10,$F7,$CE,$AB,$B4,$75,$35
	.dc.b	$3E,$47,$66,$23,$6E,$0F,$11,$93,$EA,$BF,$5D,$FC,$DD,$77,$72,$43
	.dc.b	$27,$13,$F0,$51,$C6,$FF,$BC,$46,$5F,$AC,$F8,$09,$5E,$F6,$CF,$9A
	.dc.b	$30,$7A,$C8,$F5,$BF,$24,$F8,$8F,$64,$E9,$F9,$87,$C5,$3A,$98,$37
	.dc.b	$3D,$C8,$FB,$6B,$F4,$37,$D9,$F4,$D7,$59,$F0,$3E,$6A,$E3,$18,$7C
	.dc.b	$CF,$C1,$AF,$E0,$DC,$0F,$6F,$FA,$9B,$67,$FA,$58,$70,$FF,$4E,$6B
	.dc.b	$F9,$CF,$9A,$FE,$80,$47,$FA,$2B,$FD,$7B,$D5,$78,$34,$93,$DA,$32
	.dc.b	$F8,$3E,$81,$F0,$DF,$EA,$E8,$12,$33,$F4,$1B,$F1,$BF,$C8,$F3,$FD
	.dc.b	$8A,$72,$FF,$64,$F9,$B7,$C5,$7F,$BE,$23,$98,$8C,$EF,$F0,$1B,$1C
	.dc.b	$67,$83,$E2,$F9,$F6,$38,$84,$62,$FE,$4B,$63,$FB,$AB,$5F,$09,$FE
	.dc.b	$AD,$EB,$E5,$E1,$FF,$83,$AB,$E8,$B6,$7F,$BE,$B5,$E3,$AB,$C4,$F1
	.dc.b	$36,$BE,$64,$60,$F3,$7D,$03,$C2,$A4,$9F,$BF,$19,$6F,$A7,$C0,$8E
	.dc.b	$0D,$F8,$47,$EA,$F4,$5C,$BE,$BD,$AB,$13,$99,$4B,$C3,$CD,$FD,$48
	.dc.b	$E0,$F4,$31,$E9,$EA,$BF,$07,$DD,$FE,$27,$9D,$47,$82,$74,$FF,$A0
	.dc.b	$7C,$3B,$69,$A6,$38,$2D,$F2,$BF,$C2,$7C,$AF,$2E,$53,$D3,$E4,$5F
	.dc.b	$25,$98,$D7,$AB,$30,$E5,$9F,$E1,$1E,$98,$27,$C7,$E3,$49,$F0,$7D
	.dc.b	$03,$CC,$C6,$8E,$2F,$91,$F5,$DE,$F8,$C5,$FE,$6A,$31,$3F,$37,$E8
	.dc.b	$7C,$B3,$EB,$9C,$BF,$A0,$8F,$F1,$EB,$EF,$6D,$74,$3F,$5E,$FF,$91
	.dc.b	$A3,$5E,$57,$CC,$C7,$F0,$11,$E8,$3E,$7C,$BE,$53,$51,$86,$3D,$FC
	.dc.b	$9B,$87,$71,$5F,$E4,$89,$D1,$F4,$FE,$83,$F0,$CE,$27,$84,$BE,$53
	.dc.b	$C2,$78,$BC,$7B,$5E,$D7,$0F,$3F,$A1,$B7,$CB,$87,$D2,$7A,$BD,$DB
	.dc.b	$5F,$15,$C3,$F1,$FE,$56,$D7,$A7,$93,$E4,$F1,$7E,$75,$B5,$F3,$FD
	.dc.b	$DF,$C5,$71,$7C,$66,$47,$D1,$BA,$7C,$3D,$FF,$90,$C7,$C1,$F2,$BF
	.dc.b	$37,$0F,$2A,$C7,$E7,$BD,$D7,$C1,$53,$1F,$07,$70,$FC,$E7,$D0,$FF
	.dc.b	$1C,$DB,$F6,$4D,$8F,$E4,$6F,$FB,$75,$DF,$9B,$FA,$A8,$70,$5E,$1C
	.dc.b	$36,$BF,$51,$F2,$FF,$CD,$9B,$C1,$FB,$28,$C7,$07,$DF,$7F,$DF,$B3
	.dc.b	$FA,$FF,$A1,$FF,$5C,$37,$CA,$FD,$CC,$62,$BF,$0C,$FF,$15,$6F,$89
	.dc.b	$A3,$7E,$2D,$BF,$DD,$7D,$0F,$D2,$AF,$E9,$2D,$4F,$E9,$9F,$C9,$39
	.dc.b	$FA,$1F,$F7,$C4,$FF,$03,$6C,$FE,$5D,$7C,$27,$F0,$E7,$3C,$3F,$C7
	.dc.b	$57,$C0,$67,$E5,$B4,$DC,$3F,$87,$7D,$0B,$E2,$3F,$8B,$BE,$FC,$3F
	.dc.b	$93,$39,$9D,$FC,$0F,$F1,$CF,$15,$CB,$7C,$0C,$36,$BF,$8D,$7D,$0B
	.dc.b	$E8,$7F,$91,$B6,$33,$F2,$B8,$9F,$E4,$F8,$FF,$49,$E1,$FD,$A2,$F8
	.dc.b	$6C,$FA,$FA,$7F,$96,$FD,$0F,$EF,$F1,$FE,$50,$BA,$B3,$B6,$9D,$FA
	.dc.b	$32,$B7,$F4,$37,$3F,$E0,$5F,$43,$FE,$06,$9F,$E6,$4D,$8C,$FC,$9F
	.dc.b	$F3,$C8,$C7,$27,$F4,$BF,$3F,$EE,$13,$D5,$FE,$82,$E3,$E3,$6F,$8B
	.dc.b	$86,$E5,$FC,$FF,$E8,$7F,$8F,$67,$FC,$4D,$B8,$FF,$57,$5E,$C7,$F4
	.dc.b	$C7,$35,$FE,$0A,$7E,$03,$CF,$7D,$70,$6D,$E1,$9E,$27,$C0,$DC,$7C
	.dc.b	$19,$75,$CF,$B3,$6B,$CD,$E9,$61,$A7,$A6,$3C,$9F,$0B,$FD,$83,$C1
	.dc.b	$7C,$66,$2B,$FA,$E4,$61,$F8,$1E,$46,$72,$F8,$0D,$E0,$FE,$EC,$3E
	.dc.b	$67,$A7,$93,$FD,$B2,$31,$7D,$23,$1E,$26,$6A,$CF,$F6,$BA,$DE,$1F
	.dc.b	$CF,$CF,$CC,$C7,$93,$CF,$FE,$EF,$18,$3F,$A2,$65,$FC,$89,$E1,$E5
	.dc.b	$EB,$C2,$DB,$6F,$FB,$C4,$BE,$6F,$31,$86,$C7,$87,$7F,$8D,$E8,$F8
	.dc.b	$1B,$5C,$B7,$C3,$4E,$B1,$E6,$A7,$D0,$FF,$99,$9F,$F2,$AB,$5C,$99
	.dc.b	$7F,$CD,$7F,$3E,$ED,$7D,$0E,$1B,$5F,$E4,$41,$3B,$9D,$1C,$BE,$2B
	.dc.b	$FC,$09,$3B,$9F,$F3,$6F,$0E,$49,$19,$F9,$3F,$F2,$FC,$7F,$8C,$65
	.dc.b	$18,$CF,$FC,$C7,$18,$AF,$F1,$13,$D1,$7D,$0F,$F8,$65,$FF,$39,$C7
	.dc.b	$E1,$5D,$F2,$73,$72,$D7,$D7,$76,$5C,$B4,$93,$FE,$31,$FA,$1F,$F3
	.dc.b	$2E,$5E,$7F,$3F,$81,$27,$CB,$5C,$FC,$79,$1E,$87,$4F,$F8,$37,$CD
	.dc.b	$FC,$23,$FE,$3D,$6B,$9E,$5F,$37,$D0,$3E,$87,$D1,$5F,$EA,$0C,$BF
	.dc.b	$CB,$0D,$E0,$FF,$44,$B7,$0F,$F4,$AB,$D2,$7F,$CB,$DF,$43,$F2,$E7
	.dc.b	$FC,$3C,$D8,$CE,$FE,$0F,$3F,$E3,$4F,$CB,$AD,$F0,$E0,$CD,$7C,$BF
	.dc.b	$A1,$FD,$EC,$FF,$95,$DA,$FF,$8A,$DB,$87,$F9,$BD,$FF,$26,$46,$FE
	.dc.b	$77,$91,$1E,$5E,$0F,$5F,$06,$47,$57,$C1,$EE,$78,$7D,$19,$3E,$17
	.dc.b	$6F,$87,$D5,$F3,$7C,$6C,$8F,$34,$7E,$0F,$17,$CD,$DF,$16,$57,$F9
	.dc.b	$C6,$72,$FF,$9C,$FE,$E7,$FA,$63,$7C,$DF,$C6,$47,$F9,$3E,$35,$E6
	.dc.b	$E3,$C7,$CF,$54,$9E,$51,$E1,$FA,$0F,$17,$D2,$C9,$F0,$47,$E5,$FE
	.dc.b	$38,$DF,$1A,$4F,$FA,$16,$72,$FF,$A1,$FE,$E7,$E7,$F6,$FF,$B3,$16
	.dc.b	$DF,$A1,$6C,$EF,$9B,$E1,$78,$DF,$0F,$19,$3F,$06,$3F,$51,$F7,$3F
	.dc.b	$DF,$A3,$FD,$28,$FF,$8C,$9B,$AE,$BC,$DF,$D7,$7D,$CB,$E8,$6F,$2C
	.dc.b	$BF,$F4,$E4,$E5,$F9,$DD,$FF,$4F,$BE,$46,$0F,$F5,$17,$8D,$F8,$87
	.dc.b	$FC,$EC,$FB,$9F,$F7,$9B,$5F,$DD,$FC,$87,$EF,$E1,$F1,$97,$E1,$25
	.dc.b	$74,$8F,$C1,$79,$5F,$83,$A9,$3F,$81,$8F,$E1,$1F,$72,$F8,$EB,$EF
	.dc.b	$4A,$FF,$5B,$4E,$5F,$F5,$C7,$DC,$FF,$87,$DB,$F2,$BF,$C4,$60,$DF
	.dc.b	$84,$FC,$97,$C2,$7E,$4F,$D1,$97,$F9,$18,$E1,$1C,$59,$1D,$A3,$F8
	.dc.b	$E7,$DC,$BC,$8B,$F3,$12,$BF,$D9,$53,$97,$FD,$99,$F4,$AF,$7B,$F9
	.dc.b	$23,$FE,$99,$8C,$FB,$F8,$9C,$BA,$E0,$93,$A3,$9B,$87,$FD,$55,$5F
	.dc.b	$0D,$B3,$DB,$AE,$47,$89,$F0,$FA,$98,$8F,$F5,$97,$D3,$BE,$9B,$F9
	.dc.b	$C3,$63,$81,$EA,$F7,$FD,$2C,$30,$7B,$8C,$1E,$43,$DE,$FF,$43,$F8
	.dc.b	$0F,$A7,$1F,$EE,$F6,$BF,$EC,$36,$FA,$CF,$25,$32,$7E,$F3,$BE,$5E
	.dc.b	$9F,$25,$23,$A4,$57,$FD,$9D,$F5,$CF,$9C,$FE,$A8,$D8,$F1,$57,$97
	.dc.b	$C4,$F3,$7A,$92,$7C,$D1,$5F,$F6,$C7,$D3,$7D,$B7,$CB,$A7,$9D,$F3
	.dc.b	$C6,$5F,$65,$7C,$1F,$2B,$E1,$78,$B2,$7E,$14,$57,$FD,$C9,$F3,$DF
	.dc.b	$B7,$6F,$93,$9C,$7B,$47,$FB,$0F,$C3,$F9,$9D,$D2,$3E,$64,$57,$E6
	.dc.b	$77,$FD,$13,$15,$CB,$89,$19,$9F,$A8,$EB,$8F,$E8,$C3,$E9,$FF,$82
	.dc.b	$5F,$89,$FE,$04,$4F,$23,$FC,$00,$E3,$CE,$A6,$23,$3E,$CC,$BF,$93
	.dc.b	$87,$6B,$4F,$2F,$6F,$54,$C7,$E0,$FD,$23,$E2,$7F,$C1,$CD,$9E,$67
	.dc.b	$E2,$30,$E6,$E0,$F5,$74,$F9,$88,$E6,$1E,$2E,$23,$C1,$9F,$97,$BB
	.dc.b	$DA,$67,$2F,$84,$FF,$90,$EF,$F5,$15,$FE,$3B,$6B,$F2,$8F,$95,$1B
	.dc.b	$F5,$64,$7F,$90,$4F,$9C,$E1,$FF,$22,$7D,$5B,$EB,$3F,$C5,$8D,$8F
	.dc.b	$63,$5F,$39,$F0,$67,$E9,$E3,$23,$DC,$FD,$6F,$A1,$C4,$79,$E3,$D4
	.dc.b	$47,$C9,$F5,$1F,$EA,$27,$E5,$35,$EE,$1C,$67,$F9,$65,$DB,$FE,$43
	.dc.b	$73,$5F,$8C,$6D,$7E,$F9,$AF,$39,$E1,$F2,$12,$3C,$2F,$38,$C4,$78
	.dc.b	$A7,$89,$23,$37,$E3,$AB,$7E,$19,$B7,$E1,$9A,$F8,$4E,$E4,$3B,$3E
	.dc.b	$17,$15,$D5,$B5,$D5,$A7,$52,$9D,$4E,$66,$2F,$D0,$DA,$F1,$C7,$CA
	.dc.b	$8C,$1B,$BC,$CE,$9B,$19,$36,$76,$3E,$33,$A7,$17,$F5,$E3,$0F,$AE
	.dc.b	$36,$72,$70,$6D,$C5,$FD,$BA,$F8,$91,$D2,$30,$F8,$CC,$1F,$0C,$A7
	.dc.b	$C3,$9E,$B1,$C2,$33,$7F,$05,$B5,$EF,$7C,$27,$3A,$F2,$3E,$F3,$F6
	.dc.b	$13,$F9,$76,$CF,$FA,$E6,$31,$1D,$CC,$9F,$A0,$8D,$11,$FE,$15,$C1
	.dc.b	$FE,$35,$7F,$42,$DA,$F4,$C7,$8E,$FF,$94,$CE,$5B,$E3,$7C,$E6,$CF
	.dc.b	$48,$C3,$FD,$82,$78,$36,$32,$A3,$2D,$8C,$C6,$0C,$E1,$FA,$4C,$3F
	.dc.b	$93,$CC,$AF,$92,$71,$3C,$46,$2B,$94,$74,$38,$47,$D2,$36,$7C,$27
	.dc.b	$4E,$30,$F8,$6F,$F0,$5A,$E3,$1F,$CE,$A7,$73,$C4,$F0,$3A,$C3,$F0
	.dc.b	$53,$CA,$FC,$75,$A3,$B4,$E8,$8F,$C7,$32,$7D,$32,$1E,$F3,$C0,$7A
	.dc.b	$CB,$3E,$57,$B1,$88,$EA,$FA,$CE,$A7,$38,$F7,$E1,$FC,$6A,$BC,$F3
	.dc.b	$D6,$7A,$3C,$CF,$1C,$BE,$A3,$F2,$84,$8F,$E6,$2D,$7F,$C0,$BE,$0F
	.dc.b	$FB,$B7,$B8,$FF,$2D,$9F,$E5,$DE,$8F,$E7,$49,$5F,$CE,$9F,$E9,$CB
	.dc.b	$EB,$DF,$5B,$86,$C7,$F5,$36,$BF,$D6,$4E,$11,$CB,$EA,$9F,$24,$FA
	.dc.b	$AF,$EB,$51,$86,$F8,$2F,$2F,$35,$BC,$3A,$6A,$F1,$FF,$C3,$B1,$FE
	.dc.b	$10,$8C,$71,$7F,$65,$6C,$70,$9D,$BF,$E4,$68,$78,$B8,$84,$62,$B9
	.dc.b	$4E,$63,$13,$ED,$FA,$AF,$55,$7F,$73,$6C,$67,$3E,$91,$88,$FE,$F9
	.dc.b	$F5,$4F,$A0,$F0,$36,$FF,$BE,$36,$7E,$53,$5F,$F0,$27,$D5,$3C,$4F
	.dc.b	$43,$6B,$FC,$08,$D9,$FA,$4D,$7F,$C1,$9D,$0F,$4B,$E0,$36,$BF,$C1
	.dc.b	$8D,$9F,$B4,$D7,$C3,$EB,$F8,$09,$3E,$07,$C7,$E2,$7F,$97,$7A,$FF
	.dc.b	$CA,$38,$FF,$89,$AB,$FC,$58,$FF,$9B,$5B,$3B,$7F,$C9,$3F,$4A,$F2
	.dc.b	$B1,$8F,$F9,$F2,$31,$3F,$E3,$5F,$AB,$7C,$8E,$A9,$5A,$7F,$C1,$F8
	.dc.b	$BE,$03,$4D,$C3,$C7,$C3,$89,$E5,$F5,$0F,$9F,$A7,$5F,$11,$F3,$D8
	.dc.b	$C3,$FC,$8D,$4F,$6B,$89,$DB,$9E,$CF,$C0,$F2,$3C,$27,$9F,$EA,$9E
	.dc.b	$96,$27,$FC,$A1,$F5,$6F,$33,$D8,$D8,$FF,$2A,$7D,$5F,$89,$DC,$FC
	.dc.b	$18,$7D,$5D,$31,$E7,$5A,$E2,$E1,$FF,$2F,$72,$FD,$5E,$5F,$F1,$AD
	.dc.b	$FF,$8C,$A7,$BF,$CA,$75,$3B,$7D,$0D,$8F,$1B,$87,$9F,$AB,$6F,$9F
	.dc.b	$81,$E5,$3C,$DF,$05,$6B,$D6,$E1,$F1,$FC,$4D,$AF,$1F,$27,$B4,$F7
	.dc.b	$7A,$76,$BE,$1B,$87,$E2,$FC,$65,$AF,$F2,$7E,$DF,$25,$BC,$3F,$87
	.dc.b	$8F,$D7,$78,$5F,$EA,$67,$4F,$FA,$A3,$E0,$1F,$21,$F8,$D6,$E3,$FE
	.dc.b	$85,$7F,$1D,$3B,$9F,$C9,$79,$CE,$DA,$63,$E5,$2E,$1F,$E8,$D9,$6B
	.dc.b	$C0,$E3,$AD,$9F,$F4,$95,$3E,$87,$F3,$CD,$8C,$FC,$AE,$27,$F4,$5F
	.dc.b	$68,$F9,$6A,$63,$E5,$AD,$FF,$A7,$7E,$8B,$F2,$2D,$EF,$FD,$54,$7B
	.dc.b	$23,$FD,$76,$EB,$93,$FC,$F3,$F5,$3F,$E8,$D9,$FF,$53,$36,$3F,$CF
	.dc.b	$BF,$8C,$7F,$BC,$E7,$3F,$03,$FB,$5F,$B5,$D7,$2F,$FB,$51,$B3,$FE
	.dc.b	$C8,$7F,$39,$7F,$E8,$67,$2F,$FB,$F1,$75,$F3,$3A,$7B,$D6,$BF,$D7
	.dc.b	$1F,$55,$FC,$87,$0F,$F7,$0B,$63,$FD,$AA,$BE,$7B,$F8,$43,$9C,$42
	.dc.b	$5F,$61,$FC,$35,$A7,$FB,$14,$7C,$66,$63,$5E,$E4,$9D,$7C,$5D,$BF
	.dc.b	$F5,$0F,$C5,$BE,$33,$F8,$C3,$6B,$FD,$A0,$BE,$63,$33,$FC,$77,$ED
	.dc.b	$3F,$9A,$6D,$FE,$FC,$6D,$FF,$BC,$1F,$8A,$FF,$B6,$5C,$E1,$EC,$FB
	.dc.b	$4F,$E4,$F3,$FE,$DD,$6C,$7F,$AD,$17,$CC,$7F,$2B,$73,$7F,$0E,$7E
	.dc.b	$2F,$D5,$BA,$BF,$98,$37,$0F,$F7,$43,$F9,$16,$C7,$3C,$A4,$78,$76
	.dc.b	$B7,$FE,$C4,$F8,$37,$D1,$7F,$3B,$6D,$7F,$BC,$D7,$C0,$E6,$7F,$A0
	.dc.b	$7D,$5F,$F1,$7B,$FF,$66,$B7,$4F,$44,$23,$F6,$98,$7F,$BF,$5D,$3F
	.dc.b	$80,$8D,$78,$7D,$0E,$BE,$8D,$9E,$1E,$8E,$1A,$73,$8F,$1F,$CD,$79
	.dc.b	$3E,$52,$CF,$77,$CF,$E9,$63,$1F,$3B,$73,$F3,$E7,$FB,$D9,$FF,$73
	.dc.b	$FC,$87,$EC,$DF,$CA,$C6,$67,$DB,$39,$9F,$77,$99,$F1,$3C,$9F,$8A
	.dc.b	$DC,$BF,$3B,$7F,$10,$7C,$F7,$F6,$9C,$3F,$45,$82,$73,$5C,$0F,$8D
	.dc.b	$3A,$8F,$CB,$46,$67,$AD,$7F,$73,$9C,$C7,$FC,$00,$79,$BD,$67,$E8
	.dc.b	$3F,$C8,$9E,$4F,$F9,$5B,$E3,$3F,$BE,$39,$3E,$1B,$8F,$37,$3E,$75
	.dc.b	$AE,$35,$BB,$FF,$22,$F9,$D9,$8F,$01,$F7,$DC,$47,$F9,$2B,$4F,$85
	.dc.b	$F1,$BE,$06,$DF,$4D,$70,$D3,$BC,$79,$B0,$79,$9F,$E6,$86,$9F,$E6
	.dc.b	$A7,$07,$F9,$95,$7D,$4B,$E8,$9F,$57,$98,$C3,$63,$33,$FE,$6D,$9C
	.dc.b	$B7,$E6,$7E,$C3,$74,$E5,$18,$8F,$9F,$28,$C4,$7E,$86,$3F,$CD,$8F
	.dc.b	$D2,$6B,$F2,$0F,$3F,$C8,$FE,$9E,$BF,$5B,$3C,$DF,$E5,$3F,$57,$FE
	.dc.b	$23,$AF,$E5,$73,$F8,$E7,$A7,$1F,$1D,$7C,$0D,$8D,$E1,$5F,$8B,$E3
	.dc.b	$5B,$FE,$49,$1E,$58,$D7,$D5,$FE,$2B,$2F,$1F,$85,$B2,$56,$CF,$67
	.dc.b	$85,$CC,$8F,$BE,$3C,$1E,$E3,$96,$92,$7F,$2B,$F5,$6F,$AB,$DF,$2C
	.dc.b	$9F,$CC,$DF,$F7,$76,$DF,$AF,$EA,$FE,$5E,$7F,$C5,$1B,$C3,$FC,$45
	.dc.b	$B5,$FD,$4B,$2F,$F2,$EB,$FA,$48,$F9,$4F,$D3,$6D,$73,$E1,$ED,$C3
	.dc.b	$F2,$7E,$AF,$FC,$D5,$7F,$AC,$6B,$F6,$DB,$5F,$C9,$9F,$F2,$9C,$6F
	.dc.b	$E4,$B9,$7B,$D2,$30,$73,$F5,$3B,$7C,$AC,$8C,$1D,$FD,$1F,$11,$1F
	.dc.b	$14,$79,$7C,$AF,$F6,$6E,$3F,$D0,$27,$F4,$8D,$7F,$D9,$4D,$7F,$83
	.dc.b	$E0,$FA,$0F,$F6,$AB,$63,$F5,$8D,$7F,$DA,$4D,$7F,$87,$FD,$5B,$CC
	.dc.b	$FF,$6E,$B6,$3F,$68,$D7,$FD,$B4,$D7,$8F,$89,$C0,$8F,$07,$C7,$E7
	.dc.b	$C8,$F1,$B5,$FF,$72,$7C,$27,$F6,$97,$F8,$13,$5C,$C7,$FB,$C2,$31
	.dc.b	$1F,$C9,$BE,$AF,$EC,$B7,$83,$FD,$EB,$18,$8F,$EB,$4D,$7F,$C3,$79
	.dc.b	$7E,$FD,$F8,$BE,$0F,$C8,$23,$FC,$4F,$EA,$FF,$CA,$6D,$F2,$3F,$E0
	.dc.b	$0C,$62,$3F,$DF,$8D,$7F,$9C,$FD,$5F,$F8,$89,$FE,$38,$D7,$F9,$E8
	.dc.b	$70,$F0,$F6,$BE,$CF,$44,$8F,$B2,$D7,$8F,$87,$F9,$46,$FC,$17,$F2
	.dc.b	$97,$FC,$4C,$D8,$D4,$7F,$4C,$5F,$11,$92,$47,$F3,$18,$CB,$F8,$8F
	.dc.b	$0F,$F1,$64,$70,$7F,$37,$FA,$BF,$F2,$8B,$7E,$03,$C5,$18,$8F,$7B
	.dc.b	$5F,$EB,$BF,$57,$FE,$2C,$7F,$A0,$35,$F8,$AD,$77,$EE,$7E,$73,$DE
	.dc.b	$91,$F9,$C6,$BC,$3C,$9F,$85,$46,$92,$3F,$A7,$7D,$5F,$F8,$A1,$BE
	.dc.b	$DF,$C9,$8C,$47,$D6,$6B,$FD,$D3,$EA,$DE,$D7,$CF,$6C,$7F,$59,$6B
	.dc.b	$F7,$9A,$F1,$F6,$78,$11,$E0,$D7,$FA,$64,$7C,$9F,$1B,$FC,$6E,$DF
	.dc.b	$2B,$FC,$68,$AB,$FA,$23,$FD,$36,$B8,$F4,$60,$CF,$99,$63,$CD,$1E
	.dc.b	$92,$3E,$9F,$57,$D8,$47,$07,$B4,$EE,$43,$FC,$2B,$48,$FE,$AD,$EB
	.dc.b	$3E,$AA,$92,$BF,$C3,$71,$F7,$F3,$FF,$29,$BB,$8F,$F0,$13,$5F,$D3
	.dc.b	$B5,$FF,$12,$B5,$FF,$2B,$2F,$8C,$FD,$13,$63,$9C,$60,$CF,$B5,$23
	.dc.b	$BB,$5E,$FE,$47,$22,$3E,$47,$AF,$81,$23,$CE,$D7,$CF,$E7,$78,$88
	.dc.b	$FB,$1F,$2F,$54,$8F,$7B,$5F,$7F,$2F,$A8,$8F,$C4,$6B,$FE,$11,$C9
	.dc.b	$CF,$F4,$0F,$A3,$3F,$E0,$7F,$A9,$FE,$03,$5F,$E7,$B6,$BF,$EB,$C6
	.dc.b	$A6,$9F,$D8,$B9,$38,$78,$BC,$CE,$9D,$FD,$7C,$DE,$DE,$74,$C3,$FC
	.dc.b	$AF,$82,$71,$1F,$C8,$21,$3F,$C5,$DB,$1F,$E4,$16,$BB,$F8,$FF,$51
	.dc.b	$1E,$BF,$3F,$5D,$1C,$63,$A1,$1F,$1F,$D4,$FC,$33,$87,$2F,$FA,$8A
	.dc.b	$DF,$3B,$F1,$FC,$C9,$7E,$A8,$AF,$A6,$3D,$04,$7D,$1F,$31,$F3,$F7
	.dc.b	$91,$F0,$A2,$BF,$E4,$7B,$FF,$47,$4F,$D3,$3E,$B4,$FF,$8F,$7E,$A1
	.dc.b	$F2,$D4,$97,$FE,$82,$87,$CB,$E2,$3F,$CA,$51,$C1,$D4,$F3,$9D,$67
	.dc.b	$FE,$C0,$AD,$57,$0A,$D3,$C0,$86,$4F,$05,$72,$C3,$FD,$89,$5C,$AF
	.dc.b	$A5,$F2,$BE,$04,$8C,$BC,$8E,$AE,$1D,$57,$4B,$ED,$01,$7D,$E7,$C3
	.dc.b	$7D,$EF,$91,$23,$35,$DC,$EA,$73,$AE,$4D,$AE,$53,$C2,$36,$78,$23
	.dc.b	$2E,$CF,$4C,$6E,$BE,$65,$FA,$27,$81,$A3,$C0,$E6,$FC,$31,$A3,$EC
	.dc.b	$C6,$6F,$4D,$8E,$47,$C2,$70,$F9,$A7,$F0,$6E,$9E,$A4,$8D,$D7,$C5
	.dc.b	$9F,$19,$BA,$EE,$D3,$94,$F0,$BF,$71,$F5,$63,$47,$86,$34,$FC,$53
	.dc.b	$D6,$7E,$28,$FF,$6D,$DF,$E4,$1B,$5F,$E1,$78,$C4,$7D,$13,$27,$FB
	.dc.b	$0D,$CB,$D4,$C9,$FA,$0F,$95,$FD,$41,$E2,$7F,$0E,$E1,$F3,$E1,$FA
	.dc.b	$18,$C4,$7C,$9A,$B3,$98,$DE,$1F,$9E,$AC,$BF,$0A,$3D,$D3,$B7,$C4
	.dc.b	$70,$3F,$1D,$3C,$1D,$CF,$17,$CA,$7E,$EE,$3D,$16,$78,$47,$E1,$6B
	.dc.b	$EC,$3C,$0E,$E7,$A4,$E4,$43,$06,$9F,$31,$CA,$7F,$0B,$3D,$9F,$C3
	.dc.b	$DD,$8F,$19,$97,$F3,$0D,$AF,$C7,$35,$FC,$93,$6B,$B3,$6F,$E0,$5F
	.dc.b	$B6,$BC,$D5,$E3,$BF,$04,$7C,$93,$DD,$1E,$8A,$E3,$3A,$9F,$6C,$E5
	.dc.b	$F9,$E7,$E8,$7E,$23,$F1,$84,$7F,$14,$FE,$3F,$EA,$3F,$97,$12,$3F
	.dc.b	$96,$47,$FB,$B7,$E8,$FF,$DA,$D1,$FE,$76,$6B,$FC,$4C,$FE,$28,$E1
	.dc.b	$C3,$5F,$E1,$DF,$23,$F9,$D2,$3F,$9B,$7F,$CE,$DF,$4B,$FE,$91,$8F
	.dc.b	$F4,$1B,$5E,$C1,$C3,$E1,$6B,$FC,$97,$E9,$9F,$14,$F8,$EF,$F4,$6C
	.dc.b	$A7,$31,$85,$DD,$FE,$37,$9F,$F4,$9C,$69,$FF,$4A,$FD,$E7,$3D,$F9
	.dc.b	$B6,$33,$69,$CB,$63,$FD,$3B,$F4,$0F,$8F,$C4,$7B,$23,$15,$E6,$6C
	.dc.b	$70,$9D,$BF,$3F,$EC,$BF,$C7,$95,$EC,$6C,$7F,$AA,$9A,$FC,$86,$BF
	.dc.b	$5F,$E9,$5F,$0B,$D2,$3F,$A6,$46,$2B,$E3,$B6,$3F,$D6,$AD,$7D,$1E
	.dc.b	$BF,$15,$8A,$30,$DA,$F6,$46,$FC,$CF,$F0,$6C,$7D,$46,$D7,$FB,$09
	.dc.b	$AF,$E0,$1A,$FA,$7C,$0F,$06,$06,$92,$BE,$FF,$CF,$3C,$5F,$33,$63
	.dc.b	$31,$83,$9F,$D2,$C8,$F8,$2D,$C3,$EC,$FD,$2F,$B0,$93,$F5,$DF,$7F
	.dc.b	$A7,$23,$ED,$36,$7E,$4F,$81,$E8,$24,$FC,$E7,$E5,$7D,$13,$C3,$A4
	.dc.b	$9F,$CC,$C6,$5F,$CD,$FD,$17,$F8,$81,$BE,$1F,$D3,$8C,$61,$FE,$EC
	.dc.b	$6B,$FA,$A7,$FC,$57,$3C,$BC,$1E,$04,$7A,$FC,$BF,$32,$38,$3C,$3E
	.dc.b	$07,$72,$38,$3C,$BE,$BF,$8C,$8E,$0F,$4F,$2F,$9C,$8E,$0F,$6E,$DE
	.dc.b	$B2,$38,$3E,$0F,$AF,$EF,$23,$83,$F0,$3B,$7E,$00,$8E,$0F,$E0,$7E
	.dc.b	$AB,$E1,$3F,$34,$D8,$CC,$60,$FC,$F7,$AF,$FD,$4E,$3F,$85,$B7,$7F
	.dc.b	$E2,$A7,$FA,$FF,$C9,$D9,$C3,$19,$5C,$1B,$3C,$3E,$25,$F1,$FF,$AB
	.dc.b	$6D,$7F,$18,$6C,$EE,$7F,$61,$F4,$4F,$A2,$A4,$BF,$E3,$F1,$97,$F9
	.dc.b	$0F,$CC,$3D,$5F,$DC,$36,$33,$18,$3F,$79,$F4,$4F,$A1,$A4,$9F,$E5
	.dc.b	$71,$97,$F9,$6F,$D0,$DC,$FE,$1B,$5F,$E7,$2D,$7F,$A4,$35,$FF,$2E
	.dc.b	$3F,$C1,$DE,$3C,$9F,$C3,$63,$11,$CA,$3F,$88,$7D,$0F,$FA,$AE,$BF
	.dc.b	$9E,$B6,$7E,$9B,$5F,$E9,$2D,$7F,$AC,$43,$E8,$7F,$8D,$34,$FE,$8E
	.dc.b	$58,$FB,$AD,$7F,$A9,$AF,$6F,$89,$1E,$2F,$F2,$5F,$69,$EE,$D2,$4F
	.dc.b	$F5,$48,$CB,$FD,$5B,$E8,$5D,$7F,$CB,$5B,$19,$8C,$1F,$CC,$7E,$85
	.dc.b	$F4,$D4,$93,$FD,$82,$32,$FF,$62,$FA,$27,$CA,$7F,$39,$6C,$66,$30
	.dc.b	$75,$CA,$47,$36,$E9,$FA,$0F,$49,$F3,$1F,$D0,$DB,$5F,$DC,$1B,$3B
	.dc.b	$9F,$E9,$1F,$45,$FC,$4A,$3F,$52,$D7,$FB,$FB,$5F,$F0,$73,$5F,$F5
	.dc.b	$12,$EB,$E6,$4B,$FE,$FF,$19,$6F,$89,$FD,$CF,$DD,$FE,$E3,$28,$C4
	.dc.b	$F6,$F1,$6E,$1D,$A6,$AA,$FE,$EB,$7C,$BD,$5C,$C3,$E2,$AC,$63,$8B
	.dc.b	$31,$72,$FF,$5D,$5F,$8A,$38,$F8,$1D,$BE,$27,$FA,$B6,$4A,$F3,$CE
	.dc.b	$7D,$0E,$FE,$2F,$0C,$79,$9F,$19,$C5,$F1,$8F,$F1,$B5,$BE,$4B,$B7
	.dc.b	$C3,$7B,$23,$13,$CA,$BF,$D9,$D9,$7E,$AE,$7F,$C2,$E3,$E4,$78,$61
	.dc.b	$C5,$B7,$C6,$FF,$DA,$DE,$3F,$AF,$DE,$F9,$30,$8C,$4E,$E1,$F2,$FA
	.dc.b	$9D,$B6,$75,$F3,$98,$69,$C6,$3A,$EF,$F2,$7E,$02,$CF,$8F,$5F,$F1
	.dc.b	$EB,$DF,$7C,$E5,$F1,$4F,$F2,$D3,$8D,$BF,$C4,$9E,$4F,$F9,$22,$7F
	.dc.b	$DE,$0D,$4D,$3F,$C9,$BF,$3E,$FA,$17,$D1,$61,$B5,$FD,$1A,$75,$39
	.dc.b	$38,$FC,$7F,$DF,$9D,$35,$FF,$30,$CA,$31,$1F,$E7,$18,$C6,$1E,$2F
	.dc.b	$8F,$FB,$F9,$E5,$B4,$14,$8C,$C6,$2F,$C5,$F4,$4F,$5B,$B7,$36,$B0
	.dc.b	$FC,$55,$FD,$F6,$BF,$2C,$EB,$86,$F9,$A4,$6D,$EF,$F4,$5F,$E7,$CD
	.dc.b	$3F,$CC,$77,$FC,$9A,$BA,$78,$7D,$CE,$2F,$58,$FE,$60,$F8,$7C,$AF
	.dc.b	$21,$B7,$F2,$59,$B9,$FD,$AD,$8C,$C7,$C0,$FA,$17,$D1,$6B,$A2,$57
	.dc.b	$27,$8D,$7C,$6F,$A2,$F9,$13,$F9,$F6,$BF,$05,$BA,$7F,$9F,$B0,$FE
	.dc.b	$6B,$3B,$E0,$E4,$47,$97,$07,$D2,$FA,$2F,$F2,$C4,$7E,$AD,$AF,$CD
	.dc.b	$6D,$FF,$A2,$9F,$EF,$91,$BE,$FE,$5E,$CC,$8C,$1C,$FE,$23,$B7,$83
	.dc.b	$23,$07,$77,$C5,$E8,$48,$C1,$E3,$F2,$BC,$C4,$7C,$D1,$F8,$AB,$7C
	.dc.b	$B5,$24,$FF,$05,$8F,$9E,$FF,$07,$FA,$2F,$E7,$4D,$F3,$3F,$27,$18
	.dc.b	$8F,$C6,$3F,$EE,$76,$F8,$9A,$23,$CF,$C5,$F1,$79,$D2,$3C,$0F,$97
	.dc.b	$D1,$FC,$FF,$BD,$FD,$8A,$7F,$8C,$35,$FC,$D3,$78,$BF,$D6,$EF,$F8
	.dc.b	$BE,$3F,$8F,$7D,$17,$D2,$F0,$FF,$DC,$4D,$8F,$D7,$35,$3F,$59,$FD
	.dc.b	$36,$BF,$ED,$FB,$FF,$61,$B6,$BF,$DD,$8D,$9F,$DC,$2F,$9D,$FD,$BF
	.dc.b	$D1,$7F,$2B,$9F,$E6,$0D,$7F,$64,$DB,$FF,$67,$BF,$BB,$8F,$E6,$DF
	.dc.b	$44,$FA,$2F,$E0,$2D,$8C,$C6,$3D,$9F,$B9,$D7,$27,$ED,$3F,$7F,$E3
	.dc.b	$BF,$84,$F8,$FE,$DC,$BF,$DB,$AD,$D7,$F8,$85,$FF,$49,$F8,$0E,$FE
	.dc.b	$C7,$0F,$A2,$93,$FD,$26,$D1,$D3,$E6,$FF,$19,$E5,$4A,$FC,$53,$F8
	.dc.b	$FF,$86,$FE,$33,$F4,$5F,$ED,$19,$FE,$A4,$D7,$F8,$A3,$74,$FF,$78
	.dc.b	$BF,$E2,$E8,$FE,$B1,$F4,$5F,$35,$BC,$9F,$C9,$23,$11,$FC,$75,$FD
	.dc.b	$DB,$F6,$BE,$7F,$F4,$04,$7F,$3D,$C5,$C7,$CD,$91,$D9,$E9,$9F,$F3
	.dc.b	$0F,$A1,$7D,$15,$25,$7F,$6B,$8F,$F1,$6B,$FD,$B7,$E8,$BF,$C1,$CD
	.dc.b	$F7,$3F,$9C,$C6,$23,$F1,$BD,$18,$F7,$64,$71,$77,$58,$8F,$E7,$15
	.dc.b	$F1,$79,$FE,$BB,$63,$FA,$43,$F1,$DD,$FA,$5E,$0F,$0A,$47,$96,$2B
	.dc.b	$D2,$2F,$1A,$FA,$D7,$A3,$DA,$97,$ED,$7F,$A2,$FC,$6E,$23,$CA,$7B
	.dc.b	$8C,$5F,$37,$FA,$E7,$BB,$F5,$A3,$10,$87,$6F,$F5,$E6,$C7,$F8,$5D
	.dc.b	$E1,$8F,$E5,$9C,$C7,$F5,$B6,$D7,$46,$CE,$F9,$BA,$5F,$5E,$89,$87
	.dc.b	$58,$FE,$DB,$8B,$E7,$BC,$5E,$F4,$AF,$F1,$44,$E5,$F2,$79,$7E,$72
	.dc.b	$3E,$79,$F5,$7B,$72,$3D,$4F,$F7,$AF,$9C,$FF,$4E,$46,$0F,$77,$A9
	.dc.b	$F0,$4E,$0F,$F7,$78,$FE,$F3,$BF,$B7,$A7,$E3,$1C,$FC,$4F,$91,$D7
	.dc.b	$2F,$E4,$3A,$69,$FE,$E2,$3E,$83,$FE,$29,$E2,$FE,$EF,$1F,$EE,$66
	.dc.b	$BF,$C1,$07,$D4,$3C,$1F,$F0,$9F,$D0,$FF,$09,$9F,$F0,$C3,$5D,$C6
	.dc.b	$BC,$6F,$F1,$B4,$FF,$2D,$9E,$8D,$7F,$88,$B5,$E7,$E9,$F0,$23,$E9
	.dc.b	$84,$62,$3F,$C3,$90,$F4,$BC,$A7,$9A,$7C,$FF,$33,$E2,$25,$73,$7C
	.dc.b	$BF,$47,$EC,$24,$7F,$99,$ED,$38,$8E,$F3,$77,$FC,$DD,$6A,$78,$7D
	.dc.b	$1F,$C6,$23,$FE,$71,$FA,$27,$D1,$62,$3E,$70,$A5,$7C,$82,$47,$D9
	.dc.b	$FA,$BF,$ED,$D8,$78,$3E,$6F,$E9,$9F,$E4,$68,$CF,$7F,$F3,$C6,$D7
	.dc.b	$D6,$9D,$BF,$E8,$38,$CA,$FA,$47,$CD,$FF,$94,$1B,$1F,$E8,$96,$BF
	.dc.b	$E4,$BF,$BB,$78,$7B,$A9,$3B,$7F,$CB,$33,$A8,$C1,$0F,$F2,$A5,$74
	.dc.b	$CB,$F8,$2C,$EE,$73,$5D,$1A,$7F,$A6,$E3,$53,$FC,$32,$B9,$65,$FD
	.dc.b	$71,$B5,$FE,$6D,$6A,$AE,$33,$BB,$ED,$7C,$DF,$4E,$18,$9D,$5F,$A6
	.dc.b	$BC,$F3,$98,$ED,$55,$ED,$17,$0F,$09,$E2,$C3,$C3,$3C,$0E,$98,$76
	.dc.b	$7E,$75,$71,$BE,$95,$C2,$7E,$C4,$65,$F8,$93,$5F,$8E,$72,$3A,$9D
	.dc.b	$CC,$E1,$EA,$72,$74,$2D,$77,$C3,$D3,$7E,$2B,$ED,$7F,$1E,$FC,$27
	.dc.b	$2A,$DB,$83,$BE,$1E,$D9,$E1,$1E,$53,$27,$95,$C9,$E5,$72,$79,$4B
	.dc.b	$87,$97,$0F,$1D,$7F,$87,$23,$13,$E3,$9F,$B6,$D3,$DB,$87,$E3,$5F
	.dc.b	$E4,$18,$7E,$79,$B3,$FE,$CF,$8C,$47,$D5,$32,$7F,$2D,$72,$FC,$D3
	.dc.b	$27,$E7,$BB,$FF,$58,$79,$9F,$CA,$38,$7C,$86,$CF,$D8,$3B,$7F,$6C
	.dc.b	$ED,$F8,$71,$F8,$C8,$C6,$3F,$23,$1F,$9C,$7E,$8E,$32,$E2,$BE,$73
	.dc.b	$97,$B9,$E2,$9F,$D6,$4F,$08,$D9,$0F,$19,$C0,$F1,$DE,$DC,$98,$35
	.dc.b	$3D,$1E,$D5,$E1,$7E,$01,$B2,$C6,$DC,$34,$91,$EB,$8F,$C2,$46,$1F
	.dc.b	$9A,$76,$29,$D2,$38,$5F,$C9,$8F,$2C,$62,$BD,$52,$BF,$A4,$DA,$D9
	.dc.b	$F0,$DE,$4F,$53,$13,$CA,$78,$9D,$1F,$E0,$8D,$8F,$D4,$35,$FF,$01
	.dc.b	$7C,$D7,$F7,$95,$D9,$FE,$9F,$7F,$A4,$B5,$FE,$F0,$D7,$FC,$FB,$F5
	.dc.b	$5F,$E2,$44,$1C,$37,$C3,$FF,$0B,$57,$FA,$13,$3F,$F1,$BF,$CE,$7B
	.dc.b	$EB,$FA,$BB,$5F,$F0,$33,$5F,$F4,$5B,$5E,$F8,$7F,$AE,$5F,$2F,$B9
	.dc.b	$EB,$8C,$47,$FA,$49,$AF,$0F,$8D,$D1,$1F,$F0,$D4,$23,$E2,$7D,$D7
	.dc.b	$F6,$C9,$FF,$0F,$35,$FE,$DA,$D7,$FC,$38,$D7,$FD,$43,$F4,$EF,$91
	.dc.b	$7B,$5F,$E2,$78,$C3,$7C,$2F,$EA,$95,$C2,$B6,$FF,$7A,$EB,$F9,$BC
	.dc.b	$9F,$E2,$F6,$C6,$4F,$F3,$09,$FE,$B0,$72,$FF,$8B,$FE,$95,$F2,$7A
	.dc.b	$8F,$AF,$18,$AF,$F1,$D3,$63,$FC,$12,$D7,$97,$A7,$C5,$42,$30,$D9
	.dc.b	$E7,$1B,$FA,$57,$C4,$52,$4F,$F9,$26,$32,$F8,$BC,$0E,$64,$70,$7F
	.dc.b	$93,$F9,$BF,$C3,$6D,$EF,$F2,$46,$32,$FF,$10,$B5,$FF,$28,$5B,$C1
	.dc.b	$F2,$12,$FC,$4F,$2F,$83,$91,$DD,$B5,$FE,$2C,$AE,$8F,$FB,$79,$AF
	.dc.b	$E8,$9A,$9B,$7F,$83,$AF,$56,$FE,$82,$3F,$0F,$EA,$BE,$41,$1C,$1F
	.dc.b	$07,$87,$DE,$47,$07,$F9,$CF,$E8,$9F,$57,$F5,$DB,$19,$8C,$19,$EF
	.dc.b	$FF,$56,$76,$7F,$AA,$AB,$FC,$6F,$5F,$BC,$8C,$77,$EA,$3F,$DF,$B3
	.dc.b	$AD,$55,$FE,$B2,$FA,$BF,$F6,$05,$71,$3F,$82,$CA,$32,$F0,$8E,$7F
	.dc.b	$35,$D6,$3F,$85,$CE,$AB,$C9,$F3,$79,$96,$0B,$C2,$9B,$18,$E6,$E1
	.dc.b	$CE,$FA,$3C,$67,$E5,$C7,$F4,$71,$87,$C8,$7F,$1C,$C7,$F3,$8E,$0F
	.dc.b	$0F,$D0,$EF,$3F,$F6,$8E,$1F,$3A,$74,$D8,$F9,$9E,$DF,$8B,$83,$CF
	.dc.b	$EE,$7A,$22,$E7,$E3,$C3,$DB,$1E,$FF,$A8,$74,$7B,$BE,$3B,$0D,$BE
	.dc.b	$1E,$EB,$E3,$3E,$32,$F5,$2C,$61,$F9,$5E,$6B,$CA,$BD,$18,$6E,$1E
	.dc.b	$39,$6D,$C3,$5E,$2D,$9F,$47,$C8,$3E,$5F,$EB,$2D,$AC,$61,$FB,$5F
	.dc.b	$4A,$F3,$AF,$C0,$61,$B8,$7C,$3F,$9E,$F5,$3F,$16,$BF,$95,$B9,$DF
	.dc.b	$9F,$97,$F8,$5F,$A5,$FB,$D3,$F2,$FE,$AD,$EB,$E7,$BB,$F1,$51,$89
	.dc.b	$FA,$1F,$4D,$C3,$AB,$EF,$69,$F9,$19,$FC,$CC,$E7,$B1,$1F,$9C,$9D
	.dc.b	$57,$F6,$2A,$77,$7E,$5D,$FC,$4D,$F0,$9D,$4E,$7C,$0F,$CE,$7D,$37
	.dc.b	$FB,$4B,$1F,$C1,$2F,$AA,$CF,$BB,$A7,$F4,$52,$F8,$EB,$DF,$86,$C7
	.dc.b	$E2,$C7,$C1,$70,$F8,$4E,$2D,$8F,$C8,$7C,$0A,$7F,$5C,$BE,$12,$C6
	.dc.b	$1F,$D9,$7D,$3F,$B2,$BF,$7A,$BE,$33,$3B,$69,$FD,$C7,$D3,$BE,$66
	.dc.b	$FD,$0E,$1B,$87,$E8,$7E,$9D,$F0,$5F,$BF,$5E,$0D,$8C,$3D,$35,$EB
	.dc.b	$F3,$16,$FF,$55,$92,$3F,$84,$77,$BE,$76,$F0,$E1,$B7,$FB,$1F,$A7
	.dc.b	$FF,$1A,$3F,$B3,$FC,$6B,$F6,$2E,$7C,$DF,$E2,$51,$8E,$CF,$E0,$94
	.dc.b	$F2,$7F,$8B,$3F,$AD,$BE,$91,$8F,$37,$3E,$A4,$9F,$07,$AB,$70,$E7
	.dc.b	$F0,$3C,$8E,$7C,$F6,$BC,$CE,$1F,$47,$45,$AF,$1F,$9B,$FC,$9F,$E4
	.dc.b	$5F,$41,$E4,$8C,$3E,$6F,$4B,$18,$7C,$16,$DF,$F1,$FF,$1B,$F0,$53
	.dc.b	$FC,$57,$0F,$CB,$4E,$7B,$FB,$78,$9F,$CD,$6F,$E8,$5F,$F1,$B7,$8D
	.dc.b	$73,$E8,$D5,$FF,$3B,$8F,$E1,$B5,$D7,$DD,$C7,$A5,$E9,$75,$E9,$62
	.dc.b	$7F,$93,$57,$F0,$D6,$D7,$3F,$A4,$DC,$69,$B3,$FC,$E4,$FE,$2D,$1E
	.dc.b	$BF,$77,$C5,$C1,$F1,$DF,$26,$F8,$9E,$ED,$AE,$43,$CD,$FF,$35,$34
	.dc.b	$FF,$36,$38,$3F,$B0,$BF,$D8,$BE,$97,$FC,$9D,$F5,$18,$8F,$F3,$4C
	.dc.b	$61,$C3,$5E,$2D,$79,$7C,$AF,$F7,$07,$6D,$76,$D7,$7C,$3D,$D4,$BE
	.dc.b	$57,$FC,$E9,$97,$95,$B3,$E4,$6B,$C2,$FC,$1F,$50,$F6,$BD,$9C,$7A
	.dc.b	$C7,$F8,$45,$FF,$5B,$6B,$FD,$A5,$F0,$39,$E4,$E0,$47,$FB,$17,$D4
	.dc.b	$7C,$8D,$3F,$94,$47,$A3,$C7,$EC,$47,$91,$DF,$CE,$F1,$11,$F1,$47
	.dc.b	$AB,$27,$25,$24,$FF,$71,$FA,$77,$9D,$BE,$69,$3F,$DD,$6F,$FA,$F3
	.dc.b	$6F,$E2,$FD,$47,$B2,$3F,$C0,$CD,$CF,$D2,$DE,$3F,$F1,$1E,$5F,$C8
	.dc.b	$EB,$FC,$01,$3F,$1A,$B7,$C1,$D0,$EB,$C7,$F4,$BE,$A3,$F7,$77,$FE
	.dc.b	$07,$6B,$FE,$15,$6D,$7C,$87,$FC,$05,$1B,$EE,$E5,$EF,$C8,$C1,$CF
	.dc.b	$E3,$7B,$77,$C8,$C1,$DD,$F1,$79,$92,$30,$78,$FC,$CF,$31,$1F,$34
	.dc.b	$7E,$2B,$F6,$BF,$E8,$D1,$FE,$1F,$7F,$D2,$AD,$F1,$B5,$E3,$7E,$43
	.dc.b	$CD,$FC,$74,$FF,$89,$57,$D5,$E4,$D7,$37,$E5,$BE,$63,$FC,$89,$5F
	.dc.b	$E2,$B6,$BF,$E3,$96,$F0,$FE,$31,$FE,$71,$1E,$3F,$6F,$CC,$47,$CD
	.dc.b	$E5,$78,$FC,$89,$1E,$37,$CD,$E4,$78,$6B,$C5,$52,$BC,$51,$F0,$72
	.dc.b	$91,$F0,$5F,$89,$F2,$1E,$FE,$D9,$3E,$F7,$E1,$76,$FB,$4F,$76,$52
	.dc.b	$7D,$D1,$F3,$72,$F9,$DA,$C9,$F9,$D1,$F2,$DF,$99,$E4,$47,$E6,$1F
	.dc.b	$BE,$FC,$1D,$F3,$F4,$91,$FE,$55,$8F,$CC,$3F,$E6,$3E,$4F,$D5,$BF
	.dc.b	$6F,$1F,$BC,$7D,$FC,$BF,$84,$FC,$BF,$0E,$2F,$DD,$B7,$C1,$FF,$7B
	.dc.b	$7B,$5A,$F0,$FC,$5E,$C7,$8F,$D0,$99,$F8,$63,$F1,$5B,$FE,$2F,$8A
	.dc.b	$4F,$E2,$E3,$BF,$C8,$78,$38,$64,$FF,$9B,$74,$4F,$BF,$C4,$93,$ED
	.dc.b	$7E,$17,$89,$F9,$78,$FE,$43,$F4,$1F,$88,$DF,$FD,$36,$D8,$0E,$3F
	.dc.b	$E3,$13,$AF,$4F,$89,$27,$FC,$FD,$5C,$53,$FE,$99,$7F,$9B,$35,$FE
	.dc.b	$8C,$D7,$FA,$5B,$5F,$1F,$5F,$F3,$D7,$9B,$63,$FD,$ED,$1D,$23,$07
	.dc.b	$B3,$AE,$47,$92,$3D,$FD,$3F,$03,$A6,$4F,$A2,$2B,$E4,$8E,$C4,$7D
	.dc.b	$3C,$1E,$12,$38,$3D,$A7,$CC,$21,$F1,$1A,$FC,$03,$D4,$43,$FA,$D7
	.dc.b	$A5,$F6,$9B,$BF,$C1,$8C,$47,$F6,$D6,$BF,$DC,$97,$BF,$B2,$47,$47
	.dc.b	$97,$2C,$8E,$ED,$7A,$72,$F7,$23,$E4,$7C,$5E,$8C,$8F,$3B,$5F,$27
	.dc.b	$17,$9C,$8F,$B1,$FB,$5C,$8F,$A1,$FD,$49,$27,$FD,$8D,$97,$F4,$B7
	.dc.b	$F2,$2F,$F9,$F9,$F8,$BE,$77,$C3,$23,$83,$8F,$D2,$7F,$7C,$93,$67
	.dc.b	$FB,$FF,$CF,$7F,$4A,$E6,$FF,$7C,$B6,$FF,$3E,$D4,$FF,$4E,$7F,$DC
	.dc.b	$CF,$F5,$47,$FC,$42,$D7,$9F,$99,$F9,$78,$C4,$74,$F2,$BC,$3E,$4C
	.dc.b	$8F,$F7,$24,$A7,$11,$FE,$29,$84,$F3,$24,$7F,$87,$E3,$2F,$FB,$B7
	.dc.b	$E6,$3F,$A7,$BF,$EF,$B6,$F8,$3F,$E0,$E9,$DC,$FA,$BE,$5F,$E1,$FD
	.dc.b	$04,$8F,$1C,$57,$C5,$1E,$B2,$3E,$1F,$90,$FE,$A9,$F1,$72,$3D,$71
	.dc.b	$5F,$EA,$91,$70,$F1,$4F,$D2,$D6,$57,$C2,$7E,$AF,$C6,$62,$3E,$7C
	.dc.b	$5A,$F5,$4D,$8F,$B2,$59,$FB,$31,$5F,$B3,$17,$2F,$F7,$3C,$77,$21
	.dc.b	$D8,$FA,$73,$B9,$E0,$D8,$DD,$78,$9B,$3E,$26,$9E,$22,$9E,$22,$D7
	.dc.b	$89,$F0,$12,$7E,$B4,$FC,$28,$DB,$E2,$69,$CA,$BC,$8F,$18,$FC,$DC
	.dc.b	$62,$79,$CF,$B0,$E2,$6C,$F3,$14,$F3,$16,$BC,$D1,$CA,$31,$5E,$56
	.dc.b	$9E,$33,$AC,$6C,$F8,$CE,$8E,$95,$E5,$9F,$91,$1F,$AC,$8C,$3F,$B2
	.dc.b	$8E,$D3,$CD,$DB,$63,$D2,$D7,$D2,$DA,$F4,$CF,$69,$FD,$BB,$FB,$52
	.dc.b	$3F,$AB,$6B,$FB,$16,$BC,$0F,$EC,$73,$FE,$78,$FA,$17,$BD,$FE,$8A
	.dc.b	$6C,$FC,$2A,$FE,$D2,$A3,$2A,$B9,$57,$C0,$8D,$3C,$CC,$4F,$09,$DC
	.dc.b	$F5,$3F,$BA,$C6,$6B,$BB,$63,$E5,$1F,$92,$70,$F6,$39,$14,$F2,$47
	.dc.b	$2A,$F0,$9E,$38,$DC,$F2,$39,$DF,$FA,$6A,$FE,$94,$66,$B8,$D7,$C8
	.dc.b	$BE,$8F,$28,$E8,$FE,$16,$BD,$AF,$88,$F6,$C7,$27,$07,$9E,$3E,$F4
	.dc.b	$F2,$78,$98,$8F,$01,$C8,$A7,$18,$FC,$59,$1D,$1D,$AB,$DB,$1F,$C2
	.dc.b	$DF,$64,$7A,$4E,$6E,$9F,$9D,$5E,$39,$E9,$1F,$94,$7E,$0B,$D6,$3D
	.dc.b	$95,$E8,$9D,$DF,$CB,$77,$19,$AF,$B6,$7E,$DE,$FC,$6F,$EE,$7E,$7D
	.dc.b	$F0,$3F,$E2,$76,$C7,$FB,$95,$AF,$FB,$D5,$AF,$FB,$EB,$77,$7F,$FB
	.dc.b	$91,$B1,$FD,$36,$3F,$DC,$F3,$CB,$0F,$F6,$CA,$FA,$2F,$01,$2B,$C0
	.dc.b	$FF,$80,$3B,$64,$7F,$BD,$FA,$91,$88,$F2,$79,$FE,$92,$3E,$97,$FC
	.dc.b	$85,$F6,$0F,$B0,$A4,$9F,$D7,$9F,$B0,$8D,$38,$3F,$CB,$7A,$7F,$AD
	.dc.b	$9C,$35,$F8,$5E,$AF,$F9,$3E,$31,$1F,$BA,$F9,$FF,$DF,$FC,$4F,$38
	.dc.b	$46,$1E,$CB,$BD,$E3,$6E,$35,$3E,$6F,$B0,$5B,$97,$A4,$61,$B1,$99
	.dc.b	$F5,$4E,$5F,$67,$C8,$7F,$90,$2B,$C8,$D8,$CF,$8F,$C5,$F8,$B1,$97
	.dc.b	$E3,$FC,$D3,$5D,$47,$F1,$88,$C5,$FB,$DB,$1F,$2D,$AF,$D0,$F9,$F7
	.dc.b	$BD,$CA,$3F,$D1,$31,$8A,$F9,$0D,$8F,$A8,$D7,$C7,$E7,$71,$56,$8C
	.dc.b	$36,$BC,$91,$BB,$F8,$95,$F8,$26,$BF,$CC,$9A,$FF,$A7,$1A,$AD,$F4
	.dc.b	$BE,$16,$DF,$A2,$30,$79,$7E,$6D,$E3,$D2,$4F,$E1,$63,$2F,$C0,$FB
	.dc.b	$0E,$84,$70,$7C,$DF,$77,$DA,$47,$07,$D1,$F4,$7E,$09,$1C,$1F,$33
	.dc.b	$C4,$F4,$11,$C1,$E3,$F2,$7D,$04,$70,$7E,$6F,$E7,$9F,$41,$F4,$5B
	.dc.b	$19,$8C,$19,$F8,$19,$1C,$5B,$BF,$EB,$E9,$F1,$FC,$C9,$5D,$DE,$DE
	.dc.b	$C4,$8F,$13,$6B,$97,$AF,$E3,$24,$F9,$DF,$37,$24,$8F,$53,$67,$F8
	.dc.b	$07,$03,$83,$EC,$12,$BE,$B3,$F8,$FF,$C6,$FF,$59,$E3,$FE,$9D,$FE
	.dc.b	$1D,$3F,$D2,$1C,$BF,$DF,$1C,$DF,$F5,$4F,$B8,$67,$FB,$7B,$FD,$84
	.dc.b	$7C,$2B,$FE,$A2,$E9,$BA,$6D,$FF,$03,$C3,$8B,$18,$7F,$BC,$A3,$13
	.dc.b	$CB,$8F,$32,$9C,$BF,$D9,$BE,$4B,$F9,$DD,$7F,$60,$6C,$66,$F8,$C6
	.dc.b	$31,$FE,$0D,$F7,$1F,$B5,$6F,$4F,$68,$76,$58,$C7,$D7,$67,$D3,$B9
	.dc.b	$FF,$6E,$B4,$79,$BE,$79,$F4,$B7,$C6,$C3,$6F,$FB,$AF,$DC,$7F,$B1
	.dc.b	$5B,$F3,$DE,$E8,$C7,$07,$F8,$09,$46,$7E,$0E,$57,$2F,$7A,$DF,$F8
	.dc.b	$23,$E7,$DF,$1F,$F2,$1B,$5F,$E0,$47,$F6,$F3,$B9,$F9,$BC,$EE,$7A
	.dc.b	$63,$D0,$B8,$7F,$83,$3E,$E1,$F1,$3C,$31,$FA,$B1,$89,$FF,$08,$E0
	.dc.b	$9F,$B3,$E2,$34,$A6,$3E,$0E,$DF,$F8,$63,$EE,$3F,$BE,$B7,$DA,$FC
	.dc.b	$0C,$3C,$6B,$18,$F1,$33,$E2,$DB,$E4,$FE,$23,$E8,$3F,$B8,$C7,$F8
	.dc.b	$91,$F9,$BD,$FF,$D1,$DC,$EB,$FE,$30,$73,$3F,$E5,$CF,$0D,$F0,$1F
	.dc.b	$92,$FC,$17,$FA,$D4,$FF,$8B,$A7,$2F,$F8,$C2,$33,$7F,$E3,$A9,$FF
	.dc.b	$04,$4E,$FD,$BF,$CD,$C7,$E9,$67,$51,$F9,$EF,$9E,$F7,$57,$F8,$EB
	.dc.b	$E7,$DF,$07,$9E,$0F,$D1,$C6,$27,$FC,$7F,$4F,$1F,$97,$0F,$36,$CE
	.dc.b	$B9,$BB,$1D,$F4,$B1,$CB,$E5,$3C,$47,$8F,$D5,$B1,$D9,$C3,$E0,$E3
	.dc.b	$BB,$FF,$96,$33,$7C,$7F,$EE,$3D,$FF,$C7,$F2,$FF,$75,$9C,$E1,$FE
	.dc.b	$56,$9C,$BF,$E8,$A8,$FF,$05,$BE,$CA,$F8,$9E,$E7,$C5,$7F,$81,$DB
	.dc.b	$C2,$ED,$F0,$F8,$6D,$FF,$98,$A5,$F3,$79,$8C,$37,$3F,$F3,$79,$FC
	.dc.b	$16,$3F,$D3,$79,$FF,$A8,$1F,$F3,$5C,$64,$FC,$E4,$65,$E3,$F2,$5F
	.dc.b	$C5,$9D,$1F,$A1,$71,$F0,$B9,$F9,$3D,$45,$BE,$AF,$1D,$3E,$47,$97
	.dc.b	$67,$E3,$3D,$76,$C6,$3E,$86,$E9,$FE,$13,$F9,$FF,$F1,$C4,$FE,$C5
	.dc.b	$A9,$A7,$F8,$F7,$E7,$5F,$58,$FA,$0C,$36,$BF,$C9,$93,$A9,$C9,$C7
	.dc.b	$C4,$FF,$5F,$4E,$9A,$FF,$95,$A5,$18,$8F,$F6,$3C,$63,$0F,$17,$8B
	.dc.b	$FE,$CA,$7F,$C7,$51,$FC,$1E,$7F,$D3,$6D,$7F,$D4,$AF,$15,$F4,$1F
	.dc.b	$D2,$DF,$E2,$11,$F8,$57,$8F,$27,$5D,$BC,$2D,$8D,$E1,$55,$7E,$21
	.dc.b	$F0,$F2,$66,$31,$1F,$EB,$6F,$A0,$FF,$39,$E5,$E8,$FA,$0D,$12,$78
	.dc.b	$9E,$AF,$92,$E4,$47,$DB,$1E,$1F,$69,$CD,$49,$3F,$EB,$CF,$A0,$FE
	.dc.b	$62,$FF,$AF,$A7,$F5,$ED,$E6,$E4,$70,$7F,$D9,$7F,$41,$F2,$72,$FF
	.dc.b	$33,$37,$83,$FD,$F2,$D9,$FE,$71,$83,$E7,$3E,$24,$FC,$B9,$E7,$C3
	.dc.b	$83,$3C,$3F,$27,$E8,$3F,$C8,$F3,$FE,$D2,$6B,$FE,$DC,$6E,$5F,$E7
	.dc.b	$75,$F4,$DC,$7E,$1F,$97,$AD,$2F,$94,$75,$F7,$FB,$78,$32,$7B,$45
	.dc.b	$7B,$7A,$7E,$33,$19,$77,$FA,$1F,$31,$27,$CD,$3F,$81,$F9,$EF,$F5
	.dc.b	$45,$7F,$A2,$D7,$91,$93,$5E,$47,$E1,$FC,$EF,$C8,$61,$FE,$F8,$7F
	.dc.b	$B5,$3F,$EF,$26,$B9,$E7,$FC,$77,$D0,$FF,$92,$23,$FD,$EE,$D7,$FE
	.dc.b	$00,$B7,$1F,$F4,$E3,$FC,$F6,$3F,$DF,$B8,$FF,$A1,$DB,$EA,$FA,$4F
	.dc.b	$57,$97,$E5,$F2,$24,$F8,$DF,$47,$93,$E2,$AF,$1D,$4A,$F1,$C7,$C3
	.dc.b	$F6,$A4,$78,$1F,$8D,$ED,$7C,$1E,$E9,$3E,$F7,$E2,$77,$FB,$8F,$7E
	.dc.b	$52,$7D,$F1,$F3,$F2,$FA,$1B,$49,$F6,$C7,$CD,$7E,$77,$95,$1F,$98
	.dc.b	$7E,$DF,$E8,$5F,$43,$49,$1F,$E1,$1E,$1F,$F6,$24,$FE,$6E,$3E,$4E
	.dc.b	$3F,$7C,$8F,$CC,$CB,$F8,$17,$CB,$EE,$17,$FE,$CD,$8C,$37,$C1,$DF
	.dc.b	$BB,$E3,$F9,$FF,$27,$A3,$30,$F1,$78,$DF,$E2,$35,$A2,$BF,$21,$B7
	.dc.b	$E4,$78,$65,$7D,$E8,$F1,$FD,$07,$93,$82,$4F,$43,$FC,$5D,$B2,$7E
	.dc.b	$27,$8B,$27,$47,$81,$F9,$1E,$2F,$E7,$E3,$F9,$07,$D0,$FE,$3F,$5F
	.dc.b	$33,$63,$FC,$75,$D5,$FC,$7A,$73,$EB,$70,$F5,$A5,$76,$BE,$09,$FF
	.dc.b	$43,$B9,$8D,$BF,$32,$B9,$BC,$7B,$3C,$64,$7A,$77,$D7,$81,$FE,$4D
	.dc.b	$BF,$D1,$38,$3C,$1D,$BB,$E7,$61,$EC,$EB,$F7,$F5,$CB,$F4,$45,$79
	.dc.b	$60,$9C,$5F,$29,$F0,$57,$C8,$E3,$91,$F0,$5F,$E9,$5E,$6F,$F9,$5E
	.dc.b	$30,$7C,$53,$E7,$9C,$0E,$E6,$0D,$1C,$0F,$67,$AB,$F7,$5F,$B6,$D7
	.dc.b	$3C,$7F,$D6,$63,$F4,$71,$A8,$C7,$CA,$48,$C1,$FC,$57,$DC,$E1,$D7
	.dc.b	$23,$07,$6E,$AE,$84,$7C,$31,$DF,$D4,$91,$E4,$7F,$53,$F0,$78,$8F
	.dc.b	$4F,$47,$94,$C6,$FF,$2E,$1A,$7F,$7D,$7F,$4C,$62,$7F,$C3,$71,$98
	.dc.b	$FD,$53,$5F,$F4,$72,$F4,$BE,$11,$23,$E3,$FA,$55,$EA,$78,$23,$A9
	.dc.b	$1C,$8F,$9D,$FE,$AA,$E1,$AF,$26,$BE,$86,$BF,$BF,$7F,$AF,$F2,$DF
	.dc.b	$3B,$FC,$1D,$18,$9F,$F5,$33,$D3,$2F,$88,$BC,$EF,$F1,$DB,$67,$C5
	.dc.b	$18,$23,$FB,$E8,$79,$9D,$8E,$F3,$67,$51,$5F,$D4,$4F,$56,$FC,$FF
	.dc.b	$98,$8F,$5F,$94,$AF,$C1,$8F,$29,$1F,$8D,$F5,$7E,$EF,$81,$91,$E5
	.dc.b	$8F,$82,$61,$F0,$C6,$1F,$57,$CA,$FC,$62,$3F,$D9,$BE,$06,$BE,$88
	.dc.b	$BA,$7F,$66,$9B,$B7,$D1,$9B,$5F,$ED,$3F,$41,$58,$8F,$BD,$36,$73
	.dc.b	$1E,$22,$1D,$CF,$11,$8B,$FA,$F3,$C6,$BA,$68,$8C,$56,$F0,$F2,$34
	.dc.b	$F2,$57,$09,$DC,$77,$25,$F6,$7C,$26,$2B,$9B,$4E,$67,$53,$DA,$5C
	.dc.b	$3D,$AE,$CC,$95,$CC,$F7,$8D,$61,$EB,$7C,$31,$92,$57,$89,$FA,$51
	.dc.b	$E1,$9E,$0F,$5C,$3D,$CF,$8D,$AE,$4E,$18,$7C,$2A,$EB,$1F,$A6,$8C
	.dc.b	$5F,$72,$46,$4F,$2B,$87,$DB,$3C,$96,$1E,$E9,$D6,$1E,$6A,$D4,$E6
	.dc.b	$BA,$1F,$0A,$FF,$C7,$38,$7E,$4A,$3F,$80,$47,$ED,$88,$FF,$07,$6B
	.dc.b	$FC,$1D,$AF,$03,$F7,$18,$7F,$04,$3E,$1D,$FF,$02,$71,$1F,$10,$D9
	.dc.b	$0F,$21,$63,$83,$B3,$F7,$B1,$C9,$D9,$83,$83,$F0,$A7,$53,$CA,$76
	.dc.b	$D2,$4E,$E3,$89,$FC,$69,$F4,$9F,$C8,$4E,$66,$DF,$01,$23,$EA,$61
	.dc.b	$F4,$67,$84,$F4,$3E,$CB,$F5,$8E,$6F,$D4,$9E,$31,$88,$D9,$97,$9B
	.dc.b	$DD,$F1,$47,$AA,$3D,$07,$C7,$70,$FF,$41,$8F,$5B,$A3,$91,$D4,$86
	.dc.b	$CF,$44,$FA,$DC,$BF,$D7,$E7,$AB,$98,$F3,$9C,$4E,$A6,$1D,$1C,$67
	.dc.b	$D7,$3F,$22,$BE,$04,$E6,$BB,$C6,$2B,$F8,$F7,$CF,$7F,$84,$E7,$FC
	.dc.b	$4C,$DB,$FE,$E4,$D9,$FF,$2B,$2F,$96,$FF,$72,$39,$A0,$70,$DF,$4F
	.dc.b	$73,$FE,$EB,$FA,$DF,$F0,$B4,$FF,$7C,$6B,$FE,$63,$6B,$D1,$AF,$5F
	.dc.b	$1F,$FB,$9B,$4F,$EE,$B7,$FE,$4C,$7F,$BD,$B5,$FF,$04,$F9,$6F,$A5
	.dc.b	$E2,$D8,$FE,$FA,$D7,$FC,$06,$D7,$FC,$23,$F3,$DF,$EE,$C7,$FC,$0E
	.dc.b	$D7,$3E,$7F,$92,$31,$1F,$E7,$DF,$9F,$7D,$07,$A1,$B1,$FE,$0C,$6B
	.dc.b	$FE,$12,$6B,$FE,$20,$FA,$D7,$B1,$F0,$1B,$1F,$E1,$46,$BF,$E1,$A6
	.dc.b	$BF,$E2,$8F,$A1,$7D,$1B,$E8,$BF,$C3,$71,$86,$F8,$4F,$33,$35,$BB
	.dc.b	$FF,$12,$FD,$0F,$FB,$83,$9B,$FC,$46,$D8,$CF,$BD,$FE,$39,$7F,$D3
	.dc.b	$91,$97,$FC,$73,$F3,$6F,$37,$51,$F8,$88,$C5,$FF,$8B,$9B,$1F,$E3
	.dc.b	$46,$BF,$E4,$2A,$7A,$BC,$A1,$18,$AF,$2C,$E6,$31,$3F,$E3,$EF,$A1
	.dc.b	$7C,$17,$16,$DF,$F8,$F9,$B3,$B9,$ED,$E8,$BE,$12,$92,$FF,$C8,$B1
	.dc.b	$96,$F8,$BC,$08,$E0,$ED,$E3,$78,$48,$E0,$E7,$E3,$78,$08,$E0,$F3
	.dc.b	$4F,$A4,$8E,$0F,$1F,$07,$A0,$8E,$0F,$77,$CC,$7C,$12,$38,$3F,$CB
	.dc.b	$FF,$5B,$EB,$D3,$FC,$CA,$D7,$FC,$E6,$D7,$FD,$B4,$D7,$F8,$1F,$D1
	.dc.b	$3E,$B7,$EB,$36,$BF,$CD,$6D,$9F,$F3,$8B,$5F,$F3,$FB,$FC,$3E,$3F
	.dc.b	$DA,$ED,$8F,$F5,$E2,$F1,$7F,$D6,$5E,$37,$FA,$26,$34,$FF,$A2,$F0
	.dc.b	$FF,$5F,$DF,$FB,$81,$B7,$CB,$DD,$E3,$3E,$A2,$C7,$AB,$C5,$AB,$05
	.dc.b	$61,$4D,$8C,$72,$F0,$E5,$7C,$B6,$31,$FF,$78,$C6,$1F,$6C,$E7,$1F
	.dc.b	$6F,$BB,$74,$FF,$52,$7D,$1B,$E8,$3C,$6D,$C3,$FD,$F6,$BE,$8F,$33
	.dc.b	$E8,$FA,$3F,$E0,$15,$FF,$00,$1F,$E2,$6D,$F0,$75,$E3,$7B,$BE,$8F
	.dc.b	$D7,$3F,$CF,$32,$FF,$5A,$47,$C0,$8C,$1F,$D0,$34,$47,$07,$FD,$E0
	.dc.b	$ED,$BF,$01,$D4,$ED,$F1,$96,$B9,$CF,$F4,$AF,$99,$FF,$77,$46,$21
	.dc.b	$18,$7F,$DF,$8A,$BE,$AB,$9C,$3F,$D9,$5F,$47,$FB,$A6,$F8,$3F,$6A
	.dc.b	$1E,$1D,$8C,$6F,$9F,$A3,$B9,$F8,$6D,$1F,$82,$FA,$37,$75,$F2,$70
	.dc.b	$DB,$F4,$7D,$1B,$E9,$FF,$12,$BE,$0A,$C6,$1C,$FC,$64,$AE,$3F,$07
	.dc.b	$78,$DC,$3F,$EE,$5F,$A4,$7C,$AF,$E5,$1B,$7F,$1D,$7C,$1E,$67,$F3
	.dc.b	$3E,$13,$A2,$98,$F5,$2E,$1F,$3F,$E9,$1F,$25,$C3,$4F,$D0,$C6,$27
	.dc.b	$EA,$61,$FE,$00,$7E,$B7,$C2,$36,$FD,$3B,$63,$3F,$11,$89,$FD,$5F
	.dc.b	$D2,$7F,$AC,$27,$EF,$37,$1F,$7C,$7E,$36,$3F,$7C,$EB,$E1,$38,$F8
	.dc.b	$DC,$BC,$7B,$3C,$9F,$DB,$FD,$23,$E2,$EF,$46,$1B,$5F,$8D,$F3,$3F
	.dc.b	$65,$9F,$80,$FF,$0D,$FD,$27,$F0,$6E,$85,$7E,$7E,$B7,$18,$AD,$CF
	.dc.b	$F0,$69,$7C,$75,$30,$D8,$FC,$E7,$D1,$FB,$2B,$C0,$3E,$91,$EB,$FF
	.dc.b	$10,$9D,$63,$DE,$B5,$39,$F2,$35,$C3,$71,$EA,$F1,$E6,$91,$E0,$CE
	.dc.b	$E1,$E0,$CB,$FB,$D6,$5D,$5B,$5F,$BD,$5F,$29,$E3,$F7,$BC,$8C,$9F
	.dc.b	$5F,$1D,$C3,$A7,$C2,$7F,$28,$E2,$FC,$9D,$7C,$08,$C7,$7F,$B7,$E0
	.dc.b	$3D,$D1,$67,$47,$F0,$1D,$1E,$17,$3F,$8E,$FE,$67,$18,$9F,$E0,$D8
	.dc.b	$A7,$3F,$41,$AB,$B5,$CD,$FE,$77,$9F,$F0,$78,$FE,$35,$7F,$E2,$67
	.dc.b	$F8,$84,$64,$FF,$2E,$C6,$5F,$AE,$3E,$6B,$9F,$CA,$F4,$6D,$72,$F9
	.dc.b	$BC,$34,$EB,$1E,$2D,$3E,$37,$97,$67,$D6,$F9,$B6,$C6,$3E,$06,$ED
	.dc.b	$FE,$2A,$FA,$4F,$F3,$1C,$9A,$47,$F9,$B1,$FE,$76,$FA,$17,$D2,$3E
	.dc.b	$93,$33,$98,$C3,$89,$DC,$EA,$11,$87,$5E,$2F,$F1,$66,$E1,$B9,$4E
	.dc.b	$67,$C7,$18,$9D,$78,$BF,$D2,$DF,$CF,$47,$F0,$B9,$FE,$84,$D7,$F4
	.dc.b	$EF,$15,$F4,$9F,$D0,$DF,$77,$1F,$3D,$BF,$AB,$57,$E2,$A7,$9C,$E5
	.dc.b	$F8,$98,$57,$DB,$C9,$68,$D3,$5F,$37,$F3,$32,$F3,$FC,$D6,$C9,$7E
	.dc.b	$8A,$E5,$F3,$52,$3D,$6E,$9F,$E0,$3E,$D3,$9B,$C4,$D8,$FE,$C4,$D7
	.dc.b	$F8,$3F,$D2,$7F,$36,$73,$CD,$C0,$8F,$0A,$F6,$7D,$27,$C9,$CF,$F8
	.dc.b	$53,$78,$3F,$BD,$36,$BF,$D7,$78,$3E,$8B,$8F,$0C,$9F,$EE,$95,$F2
	.dc.b	$DF,$E3,$2F,$15,$F4,$9F,$D4,$9F,$EF,$4D,$7F,$8F,$B7,$1F,$E3,$8F
	.dc.b	$F9,$B6,$37,$DD,$F4,$E3,$FA,$FC,$72,$F5,$31,$39,$F7,$75,$E0,$4A
	.dc.b	$EF,$15,$EF,$E8,$F3,$31,$8F,$81,$ED,$F4,$92,$7C,$F1,$F8,$3F,$DB
	.dc.b	$1E,$05,$24,$FF,$85,$63,$FD,$5A,$FF,$85,$FE,$E7,$FC,$0A,$DF,$73
	.dc.b	$F1,$91,$88,$FE,$7E,$F4,$7F,$C3,$DF,$03,$FE,$07,$6F,$B9,$F9,$38
	.dc.b	$FF,$04,$46,$BD,$4F,$27,$CF,$79,$70,$93,$E3,$8E,$FC,$FE,$06,$47
	.dc.b	$96,$3A,$78,$52,$3C,$2F,$7F,$13,$F4,$5F,$49,$FE,$58,$9F,$75,$7C
	.dc.b	$33,$3F,$17,$FA,$7F,$8E,$FF,$3F,$4F,$F8,$DD,$E0,$FF,$5C,$6F,$1F
	.dc.b	$F5,$97,$71,$FD,$83,$57,$D4,$FB,$C9,$1F,$13,$C1,$F1,$FD,$04,$8F
	.dc.b	$F2,$26,$68,$F3,$FC,$8F,$8F,$19,$5E,$37,$CD,$3F,$BD,$FC,$23,$FC
	.dc.b	$83,$3F,$E4,$D7,$F1,$AF,$F7,$26,$E7,$FD,$BD,$DC,$7F,$76,$C1,$EB
	.dc.b	$FC,$6C,$24,$F9,$23,$E2,$7B,$1F,$15,$91,$F1,$63,$C5,$F4,$92,$3C
	.dc.b	$4F,$97,$EC,$FE,$8F,$B5,$27,$DE,$FD,$4A,$FA,$1E,$D4,$9F,$A0,$FD
	.dc.b	$39,$F7,$7B,$7E,$FF,$96,$95,$EF,$8F,$E3,$DF,$4F,$F4,$B7,$FF,$06
	.dc.b	$37,$9B,$FC,$F8,$DA,$0D,$FF,$D3,$EF,$F9,$D6,$3E,$93,$78,$7A,$9D
	.dc.b	$B8,$AC,$7D,$8F,$7F,$F1,$A6,$1F,$F1,$0D,$7F,$AB,$1E,$9A,$F9,$8F
	.dc.b	$3E,$B6,$32,$43,$8B,$EC,$F6,$E4,$79,$5F,$7F,$B9,$88,$EF,$F0,$5F
	.dc.b	$08,$C6,$DF,$0B,$D1,$F8,$BE,$E4,$AF,$74,$57,$CF,$16,$78,$47,$C3
	.dc.b	$8F,$E9,$5F,$77,$FE,$88,$8F,$F4,$D3,$5E,$27,$C9,$F7,$3E,$9B,$0C
	.dc.b	$1F,$34,$C1,$F1,$8F,$B2,$FF,$57,$F1,$DF,$35,$49,$3F,$EA,$98,$7D
	.dc.b	$4E,$23,$FC,$9F,$0F,$9E,$E4,$67,$2F,$DB,$71,$75,$F8,$19,$7F,$BD
	.dc.b	$8F,$07,$D4,$C8,$EC,$E9,$EA,$FF,$81,$DF,$E5,$2D,$C0,$D3,$1E,$37
	.dc.b	$AF,$DE,$95,$E9,$8A,$FF,$85,$AB,$FC,$04,$FF,$87,$1A,$FF,$B4,$DA
	.dc.b	$FE,$61,$78,$7E,$12,$47,$F8,$6E,$32,$D8,$F9,$9E,$24,$8F,$95,$1F
	.dc.b	$E0,$5F,$A8,$FE,$AD,$1F,$E2,$86,$BF,$E2,$E6,$BF,$EE,$56,$BF,$AA
	.dc.b	$7F,$B0,$76,$E7,$D3,$E8,$75,$F9,$FB,$5F,$EE,$1A,$DB,$C6,$74,$E7
	.dc.b	$D1,$F2,$11,$E1,$EA,$57,$CD,$1F,$4E,$8D,$31,$F3,$FE,$DF,$93,$95
	.dc.b	$E1,$8F,$80,$61,$FF,$4B,$D7,$E6,$F8,$FF,$D2,$8A,$7E,$37,$89,$2B
	.dc.b	$D4,$FC,$9F,$17,$11,$CA,$2D,$F2,$9B,$AF,$C9,$9B,$5F,$26,$6A,$C3
	.dc.b	$FD,$4E,$F6,$3E,$44,$6A,$33,$1D,$CC,$9B,$30,$4B,$F1,$39,$7D,$B5
	.dc.b	$CE,$BA,$36,$75,$7E,$76,$9E,$72,$9E,$78,$E0,$F7,$8E,$0D,$38,$14
	.dc.b	$E0,$72,$B5,$78,$A3,$0D,$78,$46,$9E,$15,$E0,$BF,$5C,$60,$F8,$67
	.dc.b	$08,$D5,$F9,$27,$73,$D2,$FB,$46,$0F,$8F,$1C,$4F,$19,$D4,$E0,$57
	.dc.b	$84,$FE,$96,$79,$BD,$CC,$4E,$DC,$3F,$12,$7C,$6F,$EB,$EB,$F3,$84
	.dc.b	$7F,$D2,$D7,$F8,$07,$46,$23,$4F,$EE,$5F,$F4,$07,$D4,$FF,$85,$AF
	.dc.b	$FA,$FC,$A3,$0F,$EE,$67,$F2,$31,$88,$D7,$9D,$EA,$7F,$B3,$C6,$4D
	.dc.b	$5F,$E4,$B0,$F8,$33,$AA,$F0,$9C,$5E,$67,$F7,$08,$FF,$80,$15,$CA
	.dc.b	$FC,$35,$C2,$BE,$31,$CC,$E1,$87,$AA,$B8,$47,$E4,$9F,$24,$F7,$6C
	.dc.b	$72,$3C,$22,$FC,$33,$B3,$C1,$1F,$7C,$D5,$77,$8C,$CF,$BA,$7F,$32
	.dc.b	$68,$F3,$14,$EE,$EF,$0F,$8D,$5C,$E7,$D6,$70,$9D,$61,$DA,$75,$3E
	.dc.b	$39,$F3,$BC,$0E,$B1,$F8,$A7,$C7,$1A,$7C,$58,$7C,$67,$F6,$0F,$C4
	.dc.b	$7A,$4F,$69,$E0,$78,$A3,$B4,$EE,$31,$1D,$30,$F3,$3F,$89,$3E,$93
	.dc.b	$87,$C7,$5E,$5A,$E4,$F9,$F0,$F8,$73,$F0,$63,$C4,$64,$FE,$A8,$47
	.dc.b	$FA,$47,$A1,$5F,$E6,$F1,$70,$FE,$6F,$3F,$ED,$FF,$A8,$79,$DF,$DD
	.dc.b	$DB,$7F,$EF,$F6,$BF,$A4,$3F,$C8,$6E,$1C,$3F,$E6,$47,$87,$9D,$CC
	.dc.b	$8F,$3F,$1F,$B1,$1C,$1D,$FA,$FC,$44,$70,$78,$4F,$F2,$9F,$DE,$33
	.dc.b	$FF,$07,$36,$34,$1C,$3C,$5A,$FA,$FE,$9D,$F0,$5F,$E1,$76,$C7,$26
	.dc.b	$BE,$16,$BF,$0F,$E9,$DF,$39,$FE,$1F,$6C,$78,$9A,$FA,$5A,$FC,$BF
	.dc.b	$A6,$7D,$2B,$E7,$FD,$31,$86,$F8,$0F,$23,$35,$BB,$F8,$9F,$59,$C7
	.dc.b	$6F,$82,$D8,$CE,$1F,$01,$FF,$3D,$C6,$5F,$B9,$F4,$8F,$92,$D4,$23
	.dc.b	$17,$F3,$1B,$1F,$4D,$AF,$E1,$7D,$BF,$F2,$64,$FD,$26,$C7,$D9,$6B
	.dc.b	$BF,$2B,$BF,$5F,$FA,$EE,$BE,$CB,$67,$F0,$0D,$7F,$24,$D7,$F9,$BF
	.dc.b	$D3,$BE,$13,$D2,$DA,$FC,$1B,$67,$F1,$4D,$7F,$39,$E7,$3C,$7F,$82
	.dc.b	$DA,$FC,$5B,$67,$F2,$4D,$78,$F8,$5C,$09,$3C,$1F,$1F,$87,$23,$C6
	.dc.b	$D9,$E7,$E1,$F8,$09,$3E,$07,$D7,$52,$3D,$6D,$9F,$D2,$76,$BD,$7F
	.dc.b	$B2,$DA,$FD,$1B,$67,$73,$F1,$BD,$F9,$5F,$19,$B3,$FA,$FF,$A7,$F7
	.dc.b	$3F,$D9,$1A,$FF,$B1,$1A,$9B,$F4,$FD,$E3,$F8,$43,$5F,$ED,$AD,$7F
	.dc.b	$D9,$AD,$4F,$E0,$7C,$5F,$DC,$B6,$F8,$F8,$9F,$BF,$9C,$FB,$76,$7C
	.dc.b	$C5,$BF,$33,$5E,$BE,$D5,$C7,$F8,$54,$FF,$47,$C5,$E2,$ED,$55,$8C
	.dc.b	$3C,$FE,$9D,$E5,$DE,$2C,$37,$0F,$EA,$5F,$4C,$F4,$7C,$4D,$8C,$FA
	.dc.b	$38,$9F,$37,$D3,$3E,$02,$98,$E5,$B7,$FD,$76,$5F,$41,$C3,$DD,$E2
	.dc.b	$D9,$FE,$C5,$6F,$A1,$F8,$4F,$08,$CC,$62,$7E,$37,$D2,$BE,$36,$FB
	.dc.b	$18,$6D,$FF,$6D,$9F,$F4,$03,$78,$38,$FC,$37,$27,$E8,$7D,$2F,$F1
	.dc.b	$E9,$FE,$E8,$DE,$8F,$F0,$12,$EF,$DF,$97,$F5,$FE,$97,$F8,$4F,$6F
	.dc.b	$F3,$A6,$E1,$FE,$36,$7F,$DC,$2F,$F7,$F7,$38,$02,$F7,$3A,$7C,$57
	.dc.b	$56,$D7,$F8,$1F,$E9,$7F,$7D,$87,$F4,$8F,$C7,$3F,$95,$3F,$D2,$A3
	.dc.b	$33,$FE,$46,$9C,$BF,$E4,$8E,$D7,$37,$E3,$D7,$C4,$58,$C3,$9F,$4A
	.dc.b	$57,$F8,$5E,$BF,$A9,$B6,$3F,$31,$EF,$B4,$BF,$57,$86,$D7,$F8,$7F
	.dc.b	$E9,$7F,$4D,$C1,$F0,$5F,$E0,$1B,$7F,$62,$73,$07,$33,$FE,$60,$5F
	.dc.b	$15,$FA,$76,$C7,$5F,$53,$13,$FA,$BF,$A5,$7C,$8D,$31,$EA,$5B,$FF
	.dc.b	$19,$7D,$2B,$E0,$FF,$64,$D8,$CF,$9F,$89,$FD,$B7,$97,$FB,$09,$FF
	.dc.b	$1D,$3F,$A3,$6F,$1E,$BC,$4F,$DE,$7D,$2F,$E2,$67,$FC,$80,$DE,$0F
	.dc.b	$F2,$77,$D2,$FF,$87,$A7,$3D,$DF,$C0,$E1,$E4,$53,$18,$5A,$FF,$25
	.dc.b	$4F,$8E,$F9,$E3,$C7,$2E,$4F,$3E,$29,$7C,$38,$AD,$F8,$9F,$01,$E1
	.dc.b	$E9,$B3,$DB,$E0,$BF,$8B,$7C,$5B,$C7,$F3,$46,$1F,$47,$66,$2B,$CA
	.dc.b	$DC,$7F,$CD,$9D,$2F,$9A,$E7,$F1,$5F,$C8,$63,$17,$FE,$65,$C9,$39
	.dc.b	$F9,$ED,$5D,$AF,$F3,$55,$F0,$C3,$FC,$F0,$3E,$43,$CF,$E3,$7F,$2D
	.dc.b	$8C,$57,$58,$C7,$D2,$E6,$AC,$FF,$9D,$6B,$67,$4F,$9F,$CB,$63,$55
	.dc.b	$D7,$7F,$8B,$F0,$B6,$7D,$4F,$87,$7C,$5F,$99,$B9,$FF,$88,$7E,$97
	.dc.b	$FC,$8B,$3F,$C1,$5A,$9A,$7F,$92,$BE,$8F,$F6,$1F,$4B,$7C,$53,$49
	.dc.b	$5F,$E5,$59,$FE,$1D,$7C,$7C,$BF,$F5,$54,$E5,$AF,$F9,$AE,$76,$F1
	.dc.b	$F9,$2F,$F6,$8D,$7F,$96,$25,$18,$D3,$FD,$AB,$18,$AF,$F4,$F3,$CB
	.dc.b	$E9,$5F,$2F,$AE,$4E,$15,$FC,$8E,$BF,$08,$F0,$8C,$A8,$E8,$FF,$AD
	.dc.b	$57,$2E,$9B,$1E,$08,$E3,$F4,$BF,$E8,$5C,$FD,$1F,$4B,$B2,$56,$CF
	.dc.b	$3F,$B7,$CC,$8F,$AA,$3C,$79,$39,$A9,$27,$FD,$79,$F4,$AF,$A7,$DF
	.dc.b	$3C,$9F,$F6,$05,$FF,$9E,$32,$FF,$40,$BC,$E3,$FD,$89,$F4,$BE,$E9
	.dc.b	$FF,$42,$B7,$87,$FE,$00,$B6,$7F,$A5,$62,$FA,$6F,$8F,$3F,$2A,$7F
	.dc.b	$DB,$4D,$BE,$7C,$3A,$E3,$F9,$BF,$4D,$FB,$BB,$FF,$6A,$35,$FF,$71
	.dc.b	$36,$BF,$D2,$EF,$F9,$4E,$37,$E0,$72,$F7,$24,$60,$E7,$F1,$BD,$BC
	.dc.b	$19,$18,$3B,$BE,$2F,$56,$46,$0F,$1B,$E6,$FA,$69,$18,$3F,$DD,$5F
	.dc.b	$4C,$F0,$7F,$0E,$D8,$CC,$63,$C4,$FC,$5F,$D3,$3D,$2A,$4A,$FF,$7A
	.dc.b	$47,$15,$BF,$FA,$DB,$E4,$9F,$4D,$F9,$47,$FA,$6C,$FF,$BE,$DA,$F0
	.dc.b	$E7,$FE,$FC,$A3,$8B,$E8,$F5,$3C,$F8,$C8,$C1,$E2,$AF,$3C,$C8,$C1
	.dc.b	$DF,$C3,$F1,$11,$F1,$78,$BF,$A8,$FA,$47,$D0,$7C,$4C,$25,$7F,$84
	.dc.b	$74,$FF,$68,$D6,$A3,$13,$E8,$F8,$FF,$F3,$C4,$60,$FF,$0C,$3B,$7D
	.dc.b	$C6,$2F,$F6,$F1,$C2,$32,$7C,$1F,$89,$F9,$04,$7E,$47,$89,$E7,$F8
	.dc.b	$F9,$1D,$C6,$28,$F6,$7C,$87,$6C,$25,$6D,$1E,$29,$FE,$13,$F4,$DF
	.dc.b	$46,$7A,$3F,$E4,$96,$F5,$EB,$8B,$E4,$7B,$3F,$27,$09,$5E,$3C,$BF
	.dc.b	$B9,$BD,$FD,$9F,$7D,$48,$C1,$DB,$E2,$3C,$24,$7C,$3F,$4D,$F5,$7D
	.dc.b	$99,$1D,$5F,$B1,$7F,$2B,$DA,$93,$ED,$7E,$6C,$FB,$3D,$CF,$6F,$C9
	.dc.b	$CA,$F6,$C7,$F2,$8F,$A6,$FA,$5C,$EE,$7F,$C9,$C6,$1B,$C9,$F8,$7C
	.dc.b	$B8,$3C,$79,$26,$3F,$4E,$F8,$3F,$E5,$97,$EA,$3F,$9E,$3A,$C7,$F5
	.dc.b	$A8,$CC,$64,$38,$78,$3F,$E9,$A7,$46,$B3,$E4,$47,$31,$FE,$A2,$7C
	.dc.b	$97,$EF,$3E,$04,$FC,$1F,$6E,$4F,$A5,$F8,$9E,$E6,$23,$C9,$7F,$E6
	.dc.b	$27,$AC,$FA,$4F,$57,$9D,$23,$C2,$D7,$8B,$F3,$FD,$B9,$1F,$1E,$2B
	.dc.b	$DA,$3E,$21,$1F,$5F,$B5,$F1,$C8,$E0,$F7,$9F,$6C,$87,$D3,$7F,$B2
	.dc.b	$F8,$8F,$95,$A4,$9F,$D5,$43,$E9,$71,$1F,$4E,$38,$3F,$41,$AE,$67
	.dc.b	$FC,$27,$D2,$FA,$2A,$49,$FA,$F1,$97,$97,$BF,$FE,$1A,$B3,$49,$1D
	.dc.b	$FE,$97,$D2,$47,$CF,$1E,$CF,$13,$91,$23,$CC,$D7,$F8,$6D,$7F,$63
	.dc.b	$7F,$A1,$7A,$3C,$5F,$83,$E1,$7C,$62,$3F,$13,$D9,$AF,$C3,$8F,$41
	.dc.b	$1F,$C9,$FD,$77,$FB,$A9,$FE,$29,$F4,$DF,$D8,$5F,$E3,$AD,$EB,$FD
	.dc.b	$74,$7F,$AF,$67,$83,$AF,$9B,$FF,$1E,$D7,$18,$FD,$43,$5F,$E8,$ED
	.dc.b	$7F,$D8,$CF,$5E,$CC,$FA,$9F,$E2,$48,$FE,$E9,$1C,$2B,$1E,$9E,$7D
	.dc.b	$39,$87,$A2,$32,$FA,$A3,$2F,$AA,$3A,$7C,$94,$8F,$7B,$DF,$C5,$C4
	.dc.b	$7C,$08,$BA,$7C,$08,$64,$AF,$37,$87,$27,$E4,$BF,$E3,$CF,$A4,$7C
	.dc.b	$AE,$67,$E5,$CD,$C3,$A5,$6A,$78,$90,$D9,$FD,$1D,$D3,$82,$47,$97
	.dc.b	$05,$78,$6F,$B6,$BF,$42,$2D,$FF,$BD,$70,$E7,$87,$19,$CC,$FB,$9E
	.dc.b	$4F,$8E,$34,$DA,$FF,$00,$D7,$27,$B4,$64,$FB,$B5,$C9,$F3,$C6,$9A
	.dc.b	$78,$EF,$C3,$87,$AA,$39,$0B,$EC,$F4,$33,$1B,$7F,$B9,$C6,$9F,$51
	.dc.b	$AB,$F1,$61,$EA,$8F,$09,$A3,$8D,$F8,$A7,$E1,$D7,$3B,$ED,$7C,$49
	.dc.b	$87,$C2,$BE,$18,$7A,$2F,$86,$18,$8E,$A7,$8C,$D4,$E6,$F9,$C7,$C3
	.dc.b	$34,$79,$AB,$9B,$88,$FC,$4C,$F1,$BF,$99,$18,$9D,$46,$0F,$33,$83
	.dc.b	$E5,$1E,$43,$F1,$87,$EE,$2F,$F7,$64,$7F,$AC,$E1,$F3,$DD,$18,$8D
	.dc.b	$3F,$D8,$E3,$E3,$CF,$F0,$0F,$17,$F7,$87,$9E,$7F,$80,$B8,$7C,$87
	.dc.b	$EB,$9C,$3F,$5A,$79,$BC,$0F,$E2,$F1,$CD,$E0,$60,$E7,$08,$E7,$7F
	.dc.b	$3A,$7F,$00,$FE,$A6,$33,$5F,$1E,$7E,$2C,$E5,$E8,$78,$EF,$6F,$8C
	.dc.b	$C1,$A8,$E0,$E0,$FE,$57,$5C,$1F,$0D,$77,$7D,$26,$8F,$4C,$6E,$FE
	.dc.b	$0C,$70,$3C,$15,$D8,$F6,$46,$8F,$81,$1E,$3A,$ED,$19,$AE,$97,$FB
	.dc.b	$06,$D7,$EB,$1A,$FD,$16,$DF,$66,$DF,$46,$CF,$C8,$8F,$C0,$3E,$13
	.dc.b	$04,$3D,$F3,$E9,$8F,$2D,$7B,$EB,$F8,$9F,$CF,$7F,$95,$2F,$FA,$9F
	.dc.b	$A3,$FD,$14,$FF,$74,$C6,$70,$FE,$45,$19,$9F,$EB,$6E,$BC,$AF,$F7
	.dc.b	$7F,$D3,$7F,$8D,$67,$FD,$32,$D4,$1C,$37,$C8,$FF,$1C,$B9,$ED,$E6
	.dc.b	$47,$FC,$F9,$3F,$E4,$4C,$3F,$A4,$7D,$63,$ED,$38,$B6,$BF,$C8,$6D
	.dc.b	$7F,$C9,$4D,$7F,$CA,$1F,$5D,$FE,$18,$6F,$85,$E2,$8C,$47,$FA,$C9
	.dc.b	$AF,$AB,$0F,$F0,$C3,$FE,$55,$6B,$FE,$5B,$6B,$F0,$BC,$2F,$6E,$9F
	.dc.b	$E8,$E8,$FF,$2B,$B5,$FF,$2F,$35,$FF,$32,$35,$FE,$87,$F0,$5F,$1B
	.dc.b	$94,$D2,$47,$F9,$8B,$EB,$BF,$AD,$BF,$E6,$D6,$BF,$EC,$C6,$BF,$51
	.dc.b	$AB,$E9,$DE,$E3,$E9,$FF,$CD,$B1,$86,$F8,$8F,$27,$35,$BB,$FF,$3C
	.dc.b	$65,$F8,$0E,$9F,$F3,$AB,$63,$27,$F8,$04,$FE,$BE,$53,$F0,$FF,$4C
	.dc.b	$F9,$0C,$42,$31,$5F,$E8,$06,$C7,$FA,$1D,$AF,$FA,$35,$7A,$1C,$9B
	.dc.b	$3C,$63,$06,$7C,$F9,$1D,$5B,$5D,$7E,$67,$91,$27,$C4,$F5,$F0,$24
	.dc.b	$79,$5B,$3F,$E9,$BF,$A6,$FF,$0F,$4F,$FA,$61,$B1,$9F,$0B,$DF,$18
	.dc.b	$8F,$F5,$37,$D3,$38,$3E,$1B,$6F,$FD,$42,$D9,$FF,$54,$35,$FF,$57
	.dc.b	$53,$EB,$BA,$92,$BC,$4F,$C5,$FA,$29,$1F,$35,$B5,$F3,$7D,$4F,$94
	.dc.b	$49,$FA,$4F,$A3,$C2,$91,$ED,$6C,$FF,$AE,$FE,$F7,$FA,$44,$FF,$AE
	.dc.b	$1B,$19,$F1,$7E,$FC,$62,$3F,$D8,$DF,$51,$FE,$FB,$AF,$F5,$FB,$63
	.dc.b	$FD,$8C,$D7,$7E,$2F,$C6,$98,$7C,$62,$C7,$0F,$63,$F1,$B4,$69,$27
	.dc.b	$FD,$9B,$F5,$0F,$2F,$F2,$6D,$8C,$C6,$0F,$CB,$FD,$3B,$E0,$FF,$CF
	.dc.b	$BE,$4F,$F9,$7E,$31,$5F,$C9,$23,$1E,$9E,$A3,$F9,$3E,$6A,$DA,$AF
	.dc.b	$F3,$47,$D3,$BD,$FE,$13,$FC,$BA,$31,$7D,$4E,$6B,$C6,$DC,$A3,$6F
	.dc.b	$59,$DF,$56,$A5,$D7,$63,$0F,$2C,$1C,$D6,$70,$DB,$F1,$61,$FD,$0B
	.dc.b	$E9,$DF,$09,$E2,$5E,$B5,$8C,$39,$E1,$97,$CA,$D7,$FA,$8D,$BF,$D1
	.dc.b	$4F,$D3,$8F,$51,$FD,$43,$E9,$FF,$8D,$D7,$16,$C6,$83,$F9,$2A,$FF
	.dc.b	$4A,$3A,$7F,$D2,$DF,$4E,$FA,$1F,$92,$DB,$F2,$3F,$EE,$B6,$C6,$BD
	.dc.b	$DD,$FC,$55,$9D,$BF,$4B,$E9,$DE,$E7,$D4,$FA,$7F,$E6,$15,$EF,$9C
	.dc.b	$BF,$06,$33,$A7,$A7,$AF,$EA,$CF,$C5,$5E,$3F,$DE,$73,$3B,$8C,$79
	.dc.b	$B2,$B8,$7A,$36,$BE,$27,$BA,$D3,$F0,$EB,$E9,$EC,$61,$FC,$5F,$D3
	.dc.b	$FE,$CF,$8B,$FD,$78,$BE,$27,$3E,$26,$9D,$F9,$D3,$4F,$B9,$5F,$CE
	.dc.b	$AF,$8C,$FE,$5B,$E9,$DF,$4F,$4C,$73,$5C,$3F,$09,$F4,$EF,$2F,$F3
	.dc.b	$AD,$8C,$FC,$96,$27,$F4,$3F,$1B,$FB,$F9,$FC,$63,$FA,$2C,$FF,$DA
	.dc.b	$8E,$78,$BF,$0C,$E6,$7F,$6B,$F4,$FF,$87,$6F,$9D,$FA,$CF,$A7,$7C
	.dc.b	$1F,$EB,$AB,$58,$7F,$88,$A7,$F6,$57,$F9,$B6,$C7,$E4,$DE,$CE,$FC
	.dc.b	$AE,$9D,$72,$B8,$65,$6B,$87,$B1,$E1,$3C,$5C,$76,$BB,$38,$7C,$1D
	.dc.b	$96,$BC,$1D,$7F,$C1,$3E,$2F,$F6,$75,$E5,$8C,$78,$DE,$DE,$EF,$74
	.dc.b	$59,$D1,$FB,$5E,$27,$CD,$F3,$F8,$9F,$E1,$F1,$89,$FD,$D6,$29,$CF
	.dc.b	$D1,$6A,$ED,$73,$7F,$8B,$E6,$F9,$0F,$37,$BF,$86,$D7,$F0,$59,$7D
	.dc.b	$3E,$63,$0D,$8F,$E0,$03,$E8,$BA,$C7,$66,$CF,$3D,$7E,$27,$BD,$E2
	.dc.b	$6C,$77,$DF,$0D,$38,$1E,$3C,$1F,$4F,$FE,$5E,$69,$FE,$60,$70,$7F
	.dc.b	$40,$7F,$A0,$FD,$2F,$EC,$7E,$9F,$11,$FE,$6A,$8C,$38,$6B,$C5,$AF
	.dc.b	$2F,$37,$FA,$43,$B6,$BB,$6B,$BE,$3E,$EA,$5F,$27,$FD,$52,$58,$F9
	.dc.b	$A3,$0D,$AF,$E6,$D4,$BF,$0F,$D3,$BE,$03,$DB,$C7,$C2,$FD,$8D,$7C
	.dc.b	$0D,$BF,$0C,$74,$E4,$C1,$1F,$E8,$9F,$4F,$F2,$74,$F2,$C7,$C7,$F9
	.dc.b	$5E,$44,$79,$1E,$6F,$7B,$B1,$1F,$4C,$78,$F2,$FC,$34,$7F,$51,$6B
	.dc.b	$FD,$25,$B5,$E0,$87,$D3,$3E,$9F,$97,$24,$9E,$4F,$5E,$6F,$81,$F4
	.dc.b	$DE,$E9,$FE,$A8,$DC,$3A,$37,$8B,$FC,$0B,$8B,$E9,$F8,$F0,$CC,$3F
	.dc.b	$B3,$5F,$EC,$9B,$C9,$F3,$7E,$9F,$FB,$7C,$7F,$6B,$6B,$FD,$8D,$B9
	.dc.b	$7A,$DF,$EA,$B1,$BE,$FE,$5E,$EC,$8C,$1C,$FE,$9F,$B7,$81,$23,$07
	.dc.b	$77,$C5,$EB,$C8,$C1,$E3,$7C,$DF,$4F,$23,$07,$F8,$07,$F0,$EF,$F4
	.dc.b	$23,$FE,$03,$FA,$8F,$F1,$A3,$7E,$07,$F1,$53,$A8,$F6,$71,$FE,$37
	.dc.b	$EF,$9F,$4F,$4F,$89,$19,$F2,$75,$F0,$F2,$FF,$C2,$17,$FD,$B1,$F1
	.dc.b	$7C,$3F,$E5,$F3,$7D,$55,$3C,$31,$F9,$BF,$57,$FC,$11,$5F,$E1,$97
	.dc.b	$FA,$DB,$7D,$AD,$7B,$5B,$F0,$78,$78,$52,$B8,$47,$E9,$7F,$04,$FF
	.dc.b	$68,$F5,$7E,$3F,$47,$D9,$78,$CD,$7C,$D7,$EA,$FE,$FB,$FD,$43,$8F
	.dc.b	$F8,$A5,$F5,$3F,$E2,$16,$DF,$E6,$5D,$C7,$F8,$97,$EF,$BF,$34,$FF
	.dc.b	$8C,$5F,$C8,$37,$C2,$FD,$C4,$6A,$3F,$C5,$9D,$7F,$EF,$1B,$F3,$7C
	.dc.b	$A7,$C7,$3F,$7D,$19,$E7,$FF,$2F,$DB,$CA,$E7,$BC,$C7,$C1,$1F,$C1
	.dc.b	$47,$6D,$ED,$93,$FE,$45,$9C,$BF,$E4,$7F,$CD,$7C,$9A,$7C,$7F,$A4
	.dc.b	$FB,$07,$F0,$F8,$CE,$7F,$E7,$0F,$BF,$FF,$41,$5F,$F9,$3D,$7B,$7F
	.dc.b	$C5,$5B,$19,$F7,$31,$E5,$E7,$CC,$93,$C6,$3F,$8E,$7E,$26,$F8,$8F
	.dc.b	$C4,$77,$4A,$FF,$2E,$4E,$5F,$F2,$F7,$E0,$1F,$E3,$16,$FA,$BF,$C9
	.dc.b	$63,$F9,$64,$6B,$E2,$FF,$94,$7E,$01,$FE,$6C,$8F,$F3,$32,$FB,$0C
	.dc.b	$9A,$F3,$7F,$97,$7E,$55,$FE,$CB,$AF,$F3,$62,$F8,$9C,$9A,$F7,$AC
	.dc.b	$EB,$E2,$78,$98,$7F,$CD,$55,$FE,$31,$74,$43,$35,$FC,$F7,$EB,$FD
	.dc.b	$D1,$FE,$71,$6F,$4F,$F2,$96,$F3,$FF,$A8,$9F,$F2,$54,$FF,$A0,$5C
	.dc.b	$A1,$F6,$3F,$A4,$9F,$F5,$4B,$63,$53,$C1,$3F,$79,$F1,$74,$79,$4F
	.dc.b	$37,$3D,$8F,$04,$57,$D1,$1E,$BF,$3E,$47,$95,$F7,$7C,$36,$23,$DD
	.dc.b	$E8,$7C,$13,$19,$FB,$DC,$47,$B5,$CE,$7F,$07,$E1,$3E,$41,$27,$E1
	.dc.b	$7B,$B5,$FF,$4D,$7D,$CF,$F1,$27,$FD,$4C,$D7,$E7,$4F,$D1,$F8,$4F
	.dc.b	$A5,$10,$F9,$8D,$7E,$A1,$71,$FA,$91,$FD,$D7,$8D,$F5,$54,$95,$FE
	.dc.b	$B5,$87,$D8,$E2,$3F,$D5,$D1,$FC,$89,$BE,$96,$70,$FE,$D1,$39,$9F
	.dc.b	$F1,$13,$5F,$F7,$1B,$5F,$C0,$35,$2F,$B1,$F3,$12,$3C,$9F,$15,$5F
	.dc.b	$F1,$5C,$ED,$FF,$0F,$B5,$CC,$FF,$88,$9E,$BE,$FD,$9F,$EE,$D1,$E6
	.dc.b	$8C,$1B,$FA,$D7,$DA,$7F,$86,$DB,$3F,$ED,$96,$BF,$12,$7E,$67,$AF
	.dc.b	$D4,$91,$E4,$6B,$F2,$BE,$87,$E9,$1C,$1F,$A0,$E1,$A7,$D3,$8F,$F1
	.dc.b	$67,$D8,$FF,$7C,$E4,$FD,$44,$A3,$0F,$EA,$DA,$E7,$E6,$F8,$FC,$1F
	.dc.b	$C3,$9F,$F0,$3C,$FF,$95,$DA,$F4,$6B,$CF,$D1,$FF,$1E,$7D,$57,$F7
	.dc.b	$3E,$1F,$F1,$33,$71,$E1,$3E,$63,$7F,$23,$2F,$FE,$00,$61,$FC,$F1
	.dc.b	$FF,$7C,$3F,$9D,$C3,$FC,$60,$DB,$F1,$7C,$07,$C0,$F8,$4F,$82,$CB
	.dc.b	$E6,$E5,$A4,$3A,$B9,$69,$9B,$EF,$19,$6C,$66,$BC,$2F,$A2,$BC,$0F
	.dc.b	$22,$4F,$A2,$32,$F2,$8C,$CF,$CB,$9C,$BD,$23,$27,$D7,$8C,$BF,$66
	.dc.b	$AB,$F7,$62,$DF,$C5,$8D,$1E,$53,$D0,$60,$EA,$71,$32,$7E,$1E,$79
	.dc.b	$CF,$46,$C6,$A7,$B3,$B3,$B1,$D1,$C9,$27,$94,$71,$6B,$B6,$DF,$C5
	.dc.b	$3E,$44,$77,$AF,$0B,$D5,$B1,$BB,$F9,$0D,$9F,$90,$71,$3C,$05,$3C
	.dc.b	$05,$78,$3F,$0C,$FB,$47,$E8,$67,$F4,$84,$7F,$D5,$37,$FC,$C5,$D1
	.dc.b	$88,$D3,$FE,$BC,$8F,$8F,$F8,$5F,$F2,$C8,$FE,$B9,$5F,$E8,$B8,$C4
	.dc.b	$7B,$4C,$DF,$EE,$A3,$33,$BE,$8F,$8F,$39,$7F,$D5,$93,$A8,$F8,$64
	.dc.b	$AF,$ED,$51,$96,$CF,$26,$9E,$02,$9E,$31,$5E,$17,$42,$7A,$3F,$83
	.dc.b	$3F,$1A,$E0,$CD,$FE,$13,$0F,$3D,$CA,$FA,$58,$7A,$DE,$B7,$C2,$BF
	.dc.b	$3E,$F3,$AF,$44,$F8,$5E,$4F,$9F,$0F,$33,$E8,$BE,$8E,$5B,$1A,$3F
	.dc.b	$03,$87,$D6,$AE,$71,$8C,$2D,$70,$3E,$65,$FB,$E7,$94,$66,$3F,$5D
	.dc.b	$1F,$8C,$C3,$D1,$1E,$E3,$E6,$C7,$07,$73,$D2,$77,$1E,$73,$71,$D5
	.dc.b	$F0,$3B,$BF,$2C,$72,$21,$F1,$4F,$96,$FC,$89,$EB,$5E,$D9,$C9,$0F
	.dc.b	$E0,$BF,$65,$FE,$AA,$7E,$FB,$5F,$98,$7F,$8D,$5C,$47,$F3,$6F,$3F
	.dc.b	$FD,$FD,$18,$C3,$FA,$9B,$6B,$87,$C2,$FF,$58,$AF,$EB,$35,$FD,$44
	.dc.b	$93,$FA,$E6,$BE,$01,$F6,$6F,$94,$FE,$BF,$8F,$FB,$D6,$31,$3F,$D3
	.dc.b	$2B,$27,$F6,$4F,$B1,$7C,$B7,$F8,$05,$B3,$FD,$A0,$D4,$64,$FF,$03
	.dc.b	$71,$BC,$1A,$49,$E0,$7F,$95,$E3,$4E,$07,$E0,$5F,$E0,$F6,$9F,$E1
	.dc.b	$17,$06,$6D,$F0,$9E,$52,$4F,$38,$CB,$D7,$EE,$D9,$7F,$86,$DB,$19
	.dc.b	$8C,$1F,$E1,$FF,$BA,$7D,$DD,$24,$F9,$E3,$2F,$A7,$2F,$F6,$3B,$F1
	.dc.b	$5A,$FF,$83,$9A,$FC,$1F,$2B,$FC,$59,$F6,$9F,$57,$ED,$3E,$13,$67
	.dc.b	$E3,$BF,$E3,$CF,$27,$3C,$6A,$F9,$1F,$77,$74,$B1,$EC,$C3,$EB,$4E
	.dc.b	$5F,$A2,$D7,$ED,$7A,$FF,$E5,$6A,$FA,$4D,$8F,$AE,$D7,$7E,$5F,$57
	.dc.b	$FC,$53,$0F,$2F,$11,$FE,$93,$8C,$57,$49,$CC,$FF,$AC,$DB,$D5,$FE
	.dc.b	$C3,$9F,$C1,$36,$7F,$0E,$D7,$F2,$0D,$7F,$C7,$3F,$64,$F9,$3E,$AD
	.dc.b	$AF,$C6,$36,$7F,$22,$D7,$F3,$5D,$CF,$17,$CA,$DA,$FC,$A3,$67,$F3
	.dc.b	$2D,$78,$FC,$27,$02,$4F,$07,$C7,$E1,$C8,$F1,$B6,$7F,$47,$F6,$EF
	.dc.b	$8C,$F8,$AD,$AF,$D2,$36,$77,$3E,$EB,$95,$EE,$6C,$FE,$BB,$B9,$F4
	.dc.b	$DF,$41,$B5,$FA,$F6,$CE,$E7,$E4,$FB,$52,$BE,$4B,$67,$F7,$3F,$69
	.dc.b	$F0,$5F,$F3,$4B,$5F,$F6,$93,$53,$7E,$CF,$BC,$7F,$09,$6B,$FE,$72
	.dc.b	$6B,$FE,$DB,$6A,$7F,$09,$E2,$7E,$75,$B7,$C7,$E9,$7F,$77,$39,$F8
	.dc.b	$6B,$3F,$E7,$EF,$B3,$7D,$AE,$31,$FF,$74,$46,$27,$FB,$BA,$FB,$9D
	.dc.b	$4A,$34,$FF,$2D,$FC,$69,$C7,$A6,$DF,$F8,$03,$ED,$9F,$25,$9A,$E7
	.dc.b	$F8,$43,$E1,$E9,$8B,$A7,$F8,$23,$C3,$78,$BE,$26,$C6,$7E,$0B,$13
	.dc.b	$E6,$FC,$33,$F7,$DC,$5F,$C0,$5B,$5F,$E9,$EF,$9D,$7C,$26,$FC,$7F
	.dc.b	$6F,$E0,$BF,$ED,$AA,$FF,$0B,$B7,$8B,$FA,$53,$FB,$A6,$F8,$1C,$3E
	.dc.b	$56,$57,$0E,$2B,$5F,$D4,$7F,$1A,$7D,$C7,$CA,$6D,$7F,$89,$3E,$DD
	.dc.b	$E1,$66,$BE,$7F,$D6,$BD,$9A,$63,$CF,$B8,$7F,$8B,$3F,$05,$7C,$67
	.dc.b	$D6,$6C,$67,$E3,$71,$3F,$6B,$F1,$67,$A3,$4F,$BB,$F6,$FF,$CF,$AF
	.dc.b	$FC,$6F,$3A,$7B,$70,$75,$3F,$06,$BE,$B3,$0F,$F8,$F8,$7C,$B6,$8D
	.dc.b	$FC,$C5,$8F,$F2,$1F,$C6,$FB,$5B,$BF,$E3,$BE,$21,$F8,$07,$E4,$2B
	.dc.b	$59,$7F,$7B,$FC,$19,$DD,$F9,$46,$E3,$FE,$4E,$7E,$0C,$EE,$75,$F6
	.dc.b	$72,$B5,$F3,$B6,$BF,$C0,$DB,$BE,$87,$E3,$1F,$A0,$FB,$7F,$C1,$78
	.dc.b	$9F,$D2,$E7,$58,$7F,$97,$3F,$06,$7D,$D7,$E9,$9B,$19,$E3,$C4,$FE
	.dc.b	$AB,$C4,$7C,$8D,$31,$F3,$D6,$FF,$CC,$FF,$71,$F0,$33,$FE,$B4,$D8
	.dc.b	$FF,$7B,$FD,$C3,$D7,$FD,$AC,$E7,$80,$27,$FC,$33,$5F,$E7,$26,$C7
	.dc.b	$F8,$95,$F9,$2D,$F7,$3F,$7F,$1F,$2B,$0F,$ED,$4D,$AF,$0F,$7E,$BB
	.dc.b	$F7,$F2,$92,$7B,$79,$D7,$2E,$CF,$43,$AF,$9F,$67,$C9,$F4,$9F,$C3
	.dc.b	$BE,$73,$E2,$C6,$0F,$F4,$4B,$8B,$F4,$7A,$5E,$6A,$F3,$B6,$BF,$D1
	.dc.b	$FA,$BE,$CF,$A7,$D4,$FF,$1A,$8C,$4F,$48,$C7,$A1,$9A,$B3,$C9,$FE
	.dc.b	$43,$8F,$F8,$EE,$3C,$7E,$56,$3E,$4F,$D0,$EB,$C0,$C4,$FF,$A8,$29
	.dc.b	$18,$7F,$CA,$83,$C0,$E1,$3B,$6C,$6B,$5F,$89,$F3,$BE,$26,$D7,$5D
	.dc.b	$B0,$D3,$C1,$1E,$8A,$79,$BF,$E4,$B2,$3F,$E4,$F3,$27,$F9,$53,$ED
	.dc.b	$BF,$73,$F7,$1F,$E5,$36,$C6,$7C,$04,$E5,$B1,$C7,$CB,$FF,$60,$B7
	.dc.b	$8F,$94,$62,$3F,$CD,$16,$8C,$47,$E6,$04,$3E,$7B,$FC,$E7,$1F,$9C
	.dc.b	$8C,$3F,$EC,$C6,$BF,$E7,$F7,$A5,$7E,$8B,$EE,$1F,$47,$D6,$7F,$09
	.dc.b	$5C,$79,$3A,$BE,$06,$C7,$4B,$FF,$25,$2E,$5C,$92,$BF,$DB,$BF,$71
	.dc.b	$FE,$83,$AD,$FB,$1C,$89,$3E,$7A,$EB,$F4,$12,$3D,$4E,$9F,$F4,$C6
	.dc.b	$5F,$85,$7F,$DD,$2D,$73,$CB,$E8,$FB,$8F,$E6,$51,$AE,$3F,$5D,$1A
	.dc.b	$62,$BD,$FF,$71,$ED,$E5,$FA,$0D,$BF,$E1,$2D,$58,$7F,$8D,$5F,$F7
	.dc.b	$DC,$7C,$A7,$FD,$5C,$DD,$3F,$CE,$0D,$AE,$BC,$3F,$3B,$EE,$3F,$90
	.dc.b	$DF,$F1,$6B,$FF,$3B,$46,$23,$FD,$70,$D8,$CF,$C8,$71,$F2,$7F,$31
	.dc.b	$3B,$AF,$F5,$93,$5F,$C5,$35,$FE,$46,$FF,$99,$5F,$F5,$73,$5F,$F5
	.dc.b	$9B,$5E,$7F,$25,$DB,$BE,$47,$06,$BC,$3D,$5F,$19,$1E,$AF,$EB,$7E
	.dc.b	$E1,$E2,$D2,$4F,$FA,$FE,$32,$F9,$35,$F6,$91,$C1,$FE,$C5,$FB,$87
	.dc.b	$8B,$FB,$96,$C6,$63,$07,$87,$D4,$91,$DD,$AF,$FB,$4B,$D8,$FF,$4C
	.dc.b	$3F,$ED,$16,$B9,$F9,$6F,$E0,$D1,$88,$FF,$70,$3F,$E3,$C7,$87,$56
	.dc.b	$88,$FF,$B6,$FE,$DF,$FD,$A2,$FF,$B8,$1A,$FF,$BA,$5A,$FF,$01,$6B
	.dc.b	$FD,$6B,$ED,$7F,$D8,$E9,$A4,$F2,$FF,$8D,$7D,$A7,$E8,$A7,$FD,$DA
	.dc.b	$D7,$FD,$EC,$D7,$F8,$73,$5F,$EC,$FC,$3F,$E1,$07,$FD,$E4,$D7,$FD
	.dc.b	$EE,$D7,$FE,$00,$35,$E7,$DD,$E2,$23,$FE,$FB,$9F,$E1,$0F,$FB,$F6
	.dc.b	$1C,$5F,$CC,$5B,$19,$8C,$1F,$CD,$7E,$D3,$CD,$06,$93,$DE,$FE,$75
	.dc.b	$F6,$9F,$E4,$29,$FE,$54,$7F,$2B,$70,$E0,$38,$76,$D7,$87,$77,$42
	.dc.b	$3C,$4E,$2F,$F4,$9F,$B4,$7C,$A5,$24,$EE,$32,$F1,$FB,$4F,$F2,$2B
	.dc.b	$7C,$AF,$EA,$51,$88,$F3,$35,$F5,$FD,$A3,$DA,$FE,$B0,$D8,$F2,$35
	.dc.b	$F7,$35,$F8,$7F,$68,$F4,$7F,$B0,$36,$3D,$8D,$7E,$33,$5E,$3F,$35
	.dc.b	$C0,$8F,$06,$AF,$11,$DD,$FD,$6E,$BF,$D3,$D3,$FC,$7E,$31,$F2,$3F
	.dc.b	$D4,$25,$7F,$D4,$BE,$CF,$FD,$AF,$7B,$3F,$C7,$6F,$CA,$7B,$5F,$F6
	.dc.b	$78,$C6,$1E,$09,$DF,$CE,$73,$E0,$FE,$FD,$5C,$67,$F0,$AD,$79,$46
	.dc.b	$BD,$A9,$1F,$7E,$7F,$08,$F8,$A3,$9F,$B7,$E7,$24,$79,$C4,$FE,$43
	.dc.b	$0F,$24,$63,$C2,$F7,$11,$E1,$3F,$EB,$E6,$C7,$A0,$8F,$8D,$AF,$C2
	.dc.b	$3E,$2F,$C9,$7C,$93,$A3,$93,$C7,$E7,$FC,$DF,$6E,$47,$82,$2B,$FA
	.dc.b	$5B,$4F,$C8,$24,$7C,$3F,$67,$E5,$91,$C1,$F1,$CF,$B4,$43,$DA,$FF
	.dc.b	$83,$B8,$9F,$23,$49,$3F,$26,$1F,$43,$88,$FD,$C4,$70,$7F,$64,$D7
	.dc.b	$33,$FE,$3B,$8C,$69,$F4,$1F,$E7,$6D,$F1,$FA,$39,$F7,$B8,$12,$FA
	.dc.b	$BC,$FE,$CB,$D0,$48,$F3,$47,$AF,$DB,$F6,$7A,$72,$7D,$51,$5F,$4C
	.dc.b	$79,$48,$FA,$BD,$7F,$88,$47,$E1,$79,$75,$F8,$31,$ED,$23,$C4,$F9
	.dc.b	$1F,$63,$F3,$4E,$19,$7E,$43,$EC,$5E,$B7,$F8,$F9,$FF,$06,$75,$FF
	.dc.b	$93,$DB,$7F,$EA,$87,$6F,$E6,$DC,$A6,$92,$7A,$79,$12,$78,$3D,$FC
	.dc.b	$BC,$47,$E4,$EB,$F9,$FB,$67,$4F,$7F,$99,$F3,$11,$F1,$7C,$F5,$79
	.dc.b	$45,$CB,$DD,$37,$2F,$64,$DA,$F6,$4D,$7D,$51,$70,$4D,$8C,$36,$7C
	.dc.b	$87,$88,$A7,$88,$B7,$E2,$6D,$F8,$9A,$78,$8A,$78,$8B,$7E,$26,$DF
	.dc.b	$89,$A6,$AF,$C8,$F8,$4F,$98,$E2,$FF,$01,$3F,$7E,$31,$7E,$3A,$D9
	.dc.b	$D0,$E5,$5E,$96,$C6,$E3,$9C,$7C,$13,$27,$E1,$EB,$CC,$E0,$F6,$16
	.dc.b	$FD,$8D,$8D,$CF,$48,$C1,$D5,$C1,$2B,$C4,$7E,$5E,$FD,$73,$E1,$38
	.dc.b	$46,$9C,$1E,$18,$DB,$E4,$9F,$71,$E1,$7F,$C0,$F6,$BF,$F6,$23,$63
	.dc.b	$FC,$00,$D7,$FC,$14,$D7,$8C,$70,$23,$C1,$FD,$87,$D8,$C8,$FA,$15
	.dc.b	$FB,$58,$C4,$7C,$23,$F4,$CE,$23,$E1,$47,$B1,$E0,$FE,$D6,$3D,$B4
	.dc.b	$78,$61,$F0,$E7,$E0,$3F,$BF,$8C,$C7,$38,$D4,$F1,$8E,$87,$9A,$38
	.dc.b	$38,$3C,$D5,$E2,$7C,$E6,$CF,$E1,$55,$D9,$F0,$9D,$CC,$9C,$E3,$DA
	.dc.b	$E8,$FE,$1D,$1C,$6B,$E8,$4E,$9F,$34,$F8,$0F,$1D,$79,$9C,$4E,$EF
	.dc.b	$95,$FC,$63,$A3,$FC,$05,$B3,$FB,$26,$BF,$99,$6C,$F8,$5B,$5F,$21
	.dc.b	$B3,$ED,$3D,$27,$B6,$3E,$6D,$78,$A3,$E3,$1A,$3C,$F1,$F4,$BE,$5B
	.dc.b	$F9,$B7,$A7,$27,$F9,$94,$57,$F9,$54,$5C,$3F,$9F,$4F,$F3,$DF,$17
	.dc.b	$F3,$24,$AF,$CC,$BF,$9C,$BF,$E8,$44,$8F,$E7,$D0,$7F,$C3,$5F,$58
	.dc.b	$F8,$AC,$47,$F8,$2E,$30,$EA,$33,$87,$F8,$7A,$7F,$C6,$77,$C1,$DD
	.dc.b	$7F,$9C,$5A,$FF,$9D,$9A,$EE,$BF,$CD,$CB,$EA,$7F,$C2,$6F,$F9,$C5
	.dc.b	$AF,$F9,$ED,$AF,$FA,$01,$AF,$A3,$E9,$7F,$C0,$CF,$FA,$11,$AE,$7B
	.dc.b	$3B,$C6,$23,$E1,$5F,$FA,$C9,$34,$9F,$5B,$E6,$B7,$C1,$52,$57,$9D
	.dc.b	$F5,$FD,$5B,$EB,$29,$27,$FD,$19,$19,$7F,$D1,$DF,$81,$3E,$FF,$E1
	.dc.b	$B6,$33,$18,$3E,$3F,$D7,$3E,$3D,$F2,$7F,$E9,$78,$C3,$7C,$57,$8F
	.dc.b	$9A,$DE,$1F,$EA,$5F,$C6,$3F,$C0,$FC,$9F,$E9,$F6,$C6,$7B,$FF,$CB
	.dc.b	$6F,$E1,$23,$2F,$FA,$C3,$F0,$3F,$F7,$FD,$7F,$A9,$DB,$1F,$EB,$16
	.dc.b	$BB,$F8,$6E,$6F,$E4,$61,$F0,$38,$84,$62,$BA,$4E,$63,$13,$D7,$D0
	.dc.b	$C1,$2B,$C4,$EF,$D0,$91,$E5,$6C,$F9,$78,$3A,$12,$7D,$4F,$8F,$EB
	.dc.b	$BF,$C0,$F1,$FE,$C3,$6C,$FF,$B3,$5A,$FF,$B4,$9A,$FE,$AB,$EA,$9F
	.dc.b	$17,$ED,$6D,$7F,$B3,$5B,$3F,$ED,$76,$BC,$FE,$E7,$C4,$49,$F1,$3F
	.dc.b	$0F,$E9,$24,$7D,$16,$CF,$D1,$F5,$7E,$41,$27,$EB,$3E,$DF,$0A,$47
	.dc.b	$B5,$B3,$FE,$E9,$FC,$0D,$E2,$FD,$56,$D7,$FB,$95,$B3,$B9,$FB,$5F
	.dc.b	$5D,$FE,$E6,$AF,$F7,$53,$67,$FD,$E8,$D7,$FD,$F2,$D7,$E4,$79,$5F
	.dc.b	$1C,$93,$F1,$DF,$C3,$7D,$6F,$F7,$98,$FF,$7B,$36,$7F,$E0,$03,$5F
	.dc.b	$E3,$61,$C3,$FC,$7B,$EB,$19,$7E,$45,$B5,$FF,$00,$5B,$3C,$83,$87
	.dc.b	$AF,$57,$E5,$1B,$CD,$F9,$B8,$C4,$7F,$2C,$6B,$FC,$C7,$EB,$3F,$73
	.dc.b	$7D,$61,$18,$7C,$0D,$7D,$6D,$7F,$9B,$FD,$67,$FA,$BA,$79,$36,$3D
	.dc.b	$0D,$7E,$1B,$5E,$3F,$17,$FB,$18,$34,$93,$E5,$FA,$C7,$17,$ED,$DB
	.dc.b	$19,$8C,$1F,$BB,$F9,$CF,$C2,$E1,$F4,$9A,$FF,$4E,$6B,$FD,$4D,$AB
	.dc.b	$EA,$DF,$03,$FC,$15,$B7,$F1,$DB,$3F,$55,$AF,$E0,$3E,$91,$F5,$7F
	.dc.b	$C3,$5B,$5F,$3D,$B3,$F7,$5A,$FE,$22,$BF,$B3,$4F,$78,$F1,$11,$FB
	.dc.b	$03,$E2,$FF,$8D,$34,$CB,$83,$F8,$F7,$D4,$FB,$89,$CF,$83,$F9,$17
	.dc.b	$D4,$BE,$5A,$92,$FF,$0D,$19,$6F,$D1,$70,$23,$83,$DD,$F5,$3F,$04
	.dc.b	$8E,$0F,$C8,$7D,$43,$EA,$3F,$99,$36,$33,$18,$3F,$9B,$79,$BF,$C5
	.dc.b	$B5,$FF,$03,$B5,$FF,$06,$35,$FE,$AB,$E1,$FF,$3D,$FA,$8F,$BA,$4D
	.dc.b	$70,$7F,$41,$FA,$77,$D4,$52,$65,$FA,$58,$CB,$79,$78,$11,$C1,$BF
	.dc.b	$33,$A1,$1C,$1C,$FB,$BC,$04,$70,$77,$F5,$4F,$F8,$9F,$DC,$BA,$FF
	.dc.b	$8A,$7E,$23,$F9,$CC,$62,$19,$7F,$81,$AB,$81,$FE,$2C,$AD,$BF,$E2
	.dc.b	$E8,$FE,$EA,$EA,$7F,$C6,$91,$C5,$57,$58,$DC,$62,$79,$5F,$92,$BC
	.dc.b	$43,$E9,$B1,$3E,$1F,$84,$B7,$EB,$9A,$B1,$4E,$2F,$F8,$1D,$F9,$6B
	.dc.b	$C7,$79,$F0,$34,$F2,$87,$C6,$F4,$C7,$FC,$2D,$7C,$67,$2F,$38,$CC
	.dc.b	$7A,$CF,$F2,$8F,$3F,$3A,$E8,$DA,$E5,$F3,$18,$69,$D6,$3C,$5A,$78
	.dc.b	$7D,$7B,$3D,$DF,$37,$8D,$8C,$7C,$07,$F9,$8D,$C6,$BF,$E1,$6F,$0B
	.dc.b	$FC,$AB,$26,$91,$FE,$89,$F4,$DF,$C0,$7E,$A3,$F9,$D3,$63,$3E,$0A
	.dc.b	$72,$D8,$E3,$F1,$BF,$79,$B8,$72,$8C,$47,$F5,$AB,$46,$23,$FC,$EF
	.dc.b	$0F,$8D,$FF,$3A,$E9,$86,$CF,$F9,$FE,$74,$FD,$B7,$9D,$FF,$A0,$FE
	.dc.b	$A1,$E9,$74,$E4,$E3,$6B,$FC,$63,$C6,$33,$1D,$23,$E5,$BF,$D2,$9E
	.dc.b	$DC,$BA,$39,$FD,$47,$F7,$4C,$7F,$51,$4F,$8F,$E6,$4B,$E2,$79,$FE
	.dc.b	$3F,$B9,$1F,$54,$79,$B2,$73,$52,$4F,$E6,$7E,$A1,$F3,$FB,$E7,$93
	.dc.b	$F9,$DB,$FF,$3E,$B6,$FE,$0F,$D4,$7B,$74,$FD,$F3,$78,$BF,$D4,$ED
	.dc.b	$AF,$F5,$66,$4F,$B0,$E3,$C3,$27,$F5,$55,$F2,$9F,$C8,$BC,$57,$D4
	.dc.b	$BE,$0F,$E8,$BF,$E4,$A8,$FD,$93,$5F,$CC,$BB,$73,$E3,$F0,$23,$C2
	.dc.b	$79,$FD,$4C,$8E,$6F,$DE,$FB,$C7,$D0,$F8,$3C,$39,$5E,$07,$83,$E0
	.dc.b	$F9,$6F,$29,$1D,$1F,$C1,$FE,$A9,$E4,$7E,$29,$B1,$98,$F0,$7A,$DE
	.dc.b	$EE,$29,$3E,$E7,$E0,$FB,$9F,$92,$FA,$A7,$C3,$D2,$57,$F1,$58,$FF
	.dc.b	$45,$BE,$1F,$5B,$C0,$47,$F8,$D6,$CD,$FF,$3B,$E8,$3E,$6E,$92,$BF
	.dc.b	$90,$43,$E5,$37,$DF,$27,$2F,$2F,$0F,$F4,$FF,$54,$F9,$3A,$4A,$F0
	.dc.b	$F2,$FE,$B3,$EA,$BC,$F1,$FC,$49,$BD,$FF,$D2,$DB,$C7,$FE,$F4,$7E
	.dc.b	$4C,$FF,$18,$FA,$AF,$CD,$BF,$CD,$5A,$E7,$C4,$EC,$FE,$EB,$07,$D6
	.dc.b	$52,$47,$C4,$EA,$FD,$FD,$FE,$E6,$3B,$6B,$83,$3E,$07,$F0,$45,$54
	.dc.b	$93,$FD,$1E,$3F,$A5,$BF,$D2,$7E,$AF,$CB,$AF,$F6,$76,$EA,$99,$18
	.dc.b	$71,$CA,$F7,$7F,$89,$FD,$5F,$F9,$12,$7F,$AA,$35,$FE,$76,$DC,$3F
	.dc.b	$BA,$3F,$E2,$B8,$DF,$8D,$C8,$8F,$2F,$53,$F8,$FF,$D5,$BE,$9A,$92
	.dc.b	$7F,$13,$F6,$1F,$C9,$3E,$B3,$FC,$81,$1F,$D9,$5A,$FF,$4C,$6E,$5F
	.dc.b	$E0,$57,$FC,$5D,$1F,$DA,$FE,$B3,$F5,$ED,$F1,$BF,$97,$C6,$23,$FA
	.dc.b	$AB,$FB,$FA,$FF,$07,$6C,$F8,$EF,$E6,$BF,$58,$F6,$E9,$FB,$B8,$CF
	.dc.b	$D6,$4C,$7F,$BD,$52,$37,$18,$7F,$D2,$4A,$34,$FF,$A5,$5A,$FF,$A8
	.dc.b	$DA,$FF,$AB,$BE,$A9,$F5,$9C,$9B,$1F,$E9,$D6,$BF,$EA,$B6,$BC,$63
	.dc.b	$FA,$7C,$7D,$18,$FF,$08,$35,$FE,$EE,$F2,$CB,$FC,$61,$F7,$7F,$D2
	.dc.b	$F9,$FF,$15,$4A,$FC,$14,$61,$B1,$F8,$F9,$E0,$FE,$46,$3F,$14,$FA
	.dc.b	$FE,$13,$D4,$47,$D5,$F0,$9E,$D2,$38,$3D,$DE,$5F,$A8,$8F,$AA,$2B
	.dc.b	$F0,$63,$E3,$11,$F8,$DE,$B7,$C9,$23,$82,$1F,$0B,$E0,$BE,$39,$F2
	.dc.b	$23,$FB,$6F,$D2,$FC,$AF,$83,$95,$FE,$39,$F7,$53,$88,$FA,$53,$F5
	.dc.b	$C8,$FF,$B4,$F8,$DF,$59,$F8,$46,$C6,$63,$06,$7C,$CF,$F0,$3C,$FF
	.dc.b	$27,$9F,$F2,$43,$5F,$F1,$7B,$C9,$B1,$C0,$8F,$09,$FF,$08,$DF,$F2
	.dc.b	$B8,$FF,$73,$35,$FF,$7A,$35,$FF,$7F,$B5,$FC,$1B,$BE,$1F,$F3,$B5
	.dc.b	$FF,$97,$E7,$CB,$F0,$BD,$BC,$BF,$4F,$C1,$49,$F4,$BF,$E2,$2F,$BA
	.dc.b	$7D,$EE,$67,$F1,$E7,$F1,$9C,$3F,$C6,$AF,$C9,$AF,$90,$7E,$B2,$BF
	.dc.b	$C0,$18,$F8,$BD,$0F,$8E,$75,$F2,$AD,$FD,$18,$FA,$04,$70,$7C,$CF
	.dc.b	$81,$F9,$E7,$07,$FC,$B7,$F7,$2F,$AC,$FF,$1C,$FD,$6F,$F8,$16,$7F
	.dc.b	$CB,$ED,$C7,$FD,$31,$5B,$9F,$F3,$F3,$5D,$FB,$DC,$8F,$F2,$1C,$66
	.dc.b	$7F,$D4,$15,$FC,$FA,$BA,$FA,$5E,$03,$FC,$93,$F5,$3F,$47,$BF,$F6
	.dc.b	$B5,$BB,$E3,$5B,$BF,$F4,$75,$ED,$FE,$89,$5F,$C7,$DF,$F4,$93,$5F
	.dc.b	$5F,$87,$FE,$56,$F7,$7F,$A4,$42,$75,$87,$9A,$BE,$19,$BF,$56,$D7
	.dc.b	$48,$AF,$FA,$12,$F9,$DF,$2C,$BD,$35,$AC,$3F,$D8,$15,$A8,$C4,$EA
	.dc.b	$FF,$D5,$13,$97,$B1,$0E,$D8,$70,$7D,$F5,$54,$F1,$7D,$E7,$1B,$F0
	.dc.b	$D6,$F0,$F3,$4F,$23,$C3,$19,$BE,$97,$CE,$FE,$25,$73,$C3,$F9,$4C
	.dc.b	$E5,$F5,$47,$80,$F3,$9B,$3B,$19,$9F,$1B,$93,$E9,$57,$82,$FC,$51
	.dc.b	$9B,$EC,$F2,$AF,$1B,$D8,$E9,$7E,$28,$C9,$DA,$32,$71,$8D,$1E,$AA
	.dc.b	$EF,$1A,$3E,$5B,$89,$E7,$3B,$9F,$46,$1F,$29,$F4,$9F,$28,$CE,$1E
	.dc.b	$B6,$9E,$BB,$E9,$5E,$28,$FF,$0E,$61,$F6,$5F,$E0,$F7,$F9,$96,$DF
	.dc.b	$F7,$08,$C4,$7D,$03,$27,$F0,$E3,$E7,$98,$30,$F9,$C7,$D6,$FF,$00
	.dc.b	$3C,$71,$FA,$A7,$0F,$A0,$FE,$06,$E1,$F8,$D1,$CE,$31,$8F,$C6,$9F
	.dc.b	$9C,$73,$3F,$67,$1D,$9F,$D8,$3C,$30,$F8,$51,$F2,$8E,$E7,$03,$C5
	.dc.b	$3F,$C2,$A7,$84,$76,$38,$46,$5E,$07,$9A,$F6,$F8,$0E,$75,$FC,$3A
	.dc.b	$B3,$7F,$16,$AC,$79,$E3,$E3,$D7,$4B,$F7,$3E,$0B,$E4,$E5,$F5,$CF
	.dc.b	$C5,$8C,$1E,$77,$0F,$B8,$F3,$1A,$3C,$C6,$0D,$4F,$B5,$C3,$49,$3E
	.dc.b	$0B,$F4,$CF,$B8,$FA,$47,$28,$D3,$DE,$BD,$AF,$48,$F5,$BB,$3D,$13
	.dc.b	$D1,$FD,$37,$E0,$89,$FE,$73,$F4,$72,$7F,$CE,$D8,$65,$FF,$3C,$55
	.dc.b	$51,$74,$57,$19,$CD,$FF,$8D,$A7,$31,$D2,$3F,$C7,$37,$CE,$F9,$7C
	.dc.b	$CF,$F9,$52,$CD,$25,$FE,$94,$7D,$BB,$E4,$BF,$C8,$77,$E5,$8C,$4F
	.dc.b	$73,$FD,$20,$E1,$F2,$9F,$E9,$3B,$EE,$EC,$7D,$D3,$C1,$EB,$9F,$F9
	.dc.b	$42,$31,$3B,$6B,$D0,$53,$C3,$E3,$1F,$EA,$19,$DB,$FE,$0C,$6B,$D6
	.dc.b	$FF,$CE,$B1,$88,$FF,$54,$7D,$77,$F9,$95,$FF,$09,$B5,$FF,$31,$B5
	.dc.b	$DF,$A7,$FE,$7E,$FB,$A7,$C3,$52,$4F,$CB,$8C,$BF,$37,$27,$AF,$F4
	.dc.b	$C9,$19,$7E,$B7,$B1,$23,$83,$5F,$E3,$5C,$1F,$E5,$27,$EC,$35,$CF
	.dc.b	$8D,$FE,$92,$8C,$47,$F8,$BA,$FE,$03,$F2,$B5,$FC,$21,$1F,$C1,$7D
	.dc.b	$6B,$EA,$7F,$D3,$6D,$8C,$C6,$0F,$F4,$FF,$D6,$BE,$B5,$F3,$7F,$8B
	.dc.b	$8C,$37,$C8,$FA,$11,$8B,$FE,$69,$F7,$2F,$B2,$FD,$14,$7F,$A4,$A3
	.dc.b	$1C,$FF,$95,$6C,$70,$9D,$BF,$EE,$1F,$AE,$7C,$7E,$21,$18,$BF,$CF
	.dc.b	$36,$3F,$A2,$35,$FF,$29,$DB,$E2,$F9,$36,$BC,$51,$83,$B5,$3D,$5A
	.dc.b	$49,$EE,$F2,$D6,$47,$AD,$B8,$78,$BC,$0E,$64,$9F,$3B,$EB,$FA,$EF
	.dc.b	$F6,$C4,$7E,$DD,$B3,$FD,$75,$AF,$F9,$B1,$AF,$CB,$DB,$D6,$49,$F9
	.dc.b	$8F,$5F,$AD,$91,$D5,$B3,$F5,$BD,$EF,$86,$49,$FA,$8F,$C2,$EF,$91
	.dc.b	$F3,$9B,$3F,$7B,$B3,$E5,$92,$7E,$E3,$F5,$FE,$B7,$E7,$4F,$F7,$76
	.dc.b	$BF,$E8,$66,$BF,$ED,$66,$A1,$F5,$BF,$EB,$6A,$FE,$38,$D8,$FF,$01
	.dc.b	$35,$FF,$48,$B5,$FF,$70,$FD,$6F,$F8,$0A,$7F,$8C,$7D,$CB,$E3,$3F
	.dc.b	$9C,$78,$DF,$CD,$EB,$57,$FC,$56,$B5,$7F,$C7,$FE,$BB,$F8,$2B,$75
	.dc.b	$E2,$FF,$BB,$A3,$35,$FC,$75,$4E,$A3,$D9,$DA,$F7,$F1,$1F,$CA,$7A
	.dc.b	$6C,$E7,$0F,$E5,$CF,$E7,$73,$FE,$99,$F5,$DF,$B4,$9F,$E6,$5F,$2D
	.dc.b	$F9,$77,$F9,$A4,$67,$0F,$E5,$D3,$97,$F9,$87,$D7,$3E,$03,$C2,$FF
	.dc.b	$B7,$2F,$F9,$E3,$96,$FC,$06,$B8,$77,$ED,$D9,$DB,$EA,$FA,$EF,$EA
	.dc.b	$71,$FD,$17,$EE,$5D,$39,$F4,$7D,$F1,$89,$FE,$95,$F4,$FF,$86,$E0
	.dc.b	$FE,$D0,$FF,$04,$6F,$85,$F1,$A3,$13,$F8,$5D,$DF,$5C,$F8,$7F,$96
	.dc.b	$B8,$AC,$61,$F9,$FF,$5F,$FE,$8A,$BF,$EB,$5F,$72,$FB,$6C,$FB,$9F
	.dc.b	$52,$31,$3F,$D7,$F1,$FC,$A3,$7A,$3B,$E7,$E0,$7E,$DC,$3E,$AA,$FD
	.dc.b	$76,$1B,$1F,$DA,$7E,$C7,$E0,$3E,$BF,$9B,$7D,$BF,$E1,$27,$38,$FE
	.dc.b	$8E,$73,$5F,$DC,$7E,$CB,$DC,$DF,$5B,$F1,$71,$8A,$FE,$DC,$FF,$80
	.dc.b	$67,$87,$B3,$AF,$90,$DB,$63,$FB,$E7,$13,$E7,$3F,$2B,$F7,$3F,$85
	.dc.b	$C7,$FB,$C4,$E6,$BF,$C0,$53,$9E,$9F,$F1,$07,$D9,$B8,$7F,$3A,$BE
	.dc.b	$86,$C6,$1F,$D0,$FB,$2F,$A8,$BF,$19,$86,$E5,$FE,$0F,$FB,$3F,$87
	.dc.b	$AB,$F1,$6F,$EB,$93,$9F,$03,$F5,$71,$89,$FF,$07,$7D,$A7,$C1,$6F
	.dc.b	$C1,$7E,$C2,$31,$BF,$F8,$DD,$FE,$18,$DF,$07,$5E,$0E,$DB,$97,$F8
	.dc.b	$86,$BF,$68,$FE,$4F,$F1,$1F,$C2,$E3,$E0,$F3,$BF,$7B,$18,$C7,$B7
	.dc.b	$9F,$E1,$9F,$13,$6F,$A7,$D8,$FF,$04,$F9,$9F,$89,$18,$3F,$C6,$0E
	.dc.b	$31,$F6,$F8,$5E,$BA,$F6,$36,$BF,$C6,$D8,$BC,$CE,$9F,$23,$FC,$42
	.dc.b	$31,$3D,$23,$1E,$76,$6A,$CF,$27,$F8,$BE,$2E,$FF,$37,$BD,$86,$D7
	.dc.b	$F9,$16,$7F,$C4,$AD,$8E,$3E,$06,$E3,$5F,$0F,$86,$9C,$63,$BE,$BF
	.dc.b	$0F,$DD,$B3,$F0,$9F,$1E,$D8,$C7,$FC,$98,$7F,$94,$E7,$C5,$5E,$88
	.dc.b	$79,$BF,$E6,$57,$FA,$73,$5C,$99,$7F,$CD,$5F,$65,$FB,$BF,$B5,$BE
	.dc.b	$59,$A4,$AF,$F6,$A4,$BE,$67,$FD,$03,$18,$6B,$FC,$64,$46,$F8,$FB
	.dc.b	$CE,$A7,$F8,$F4,$BE,$D7,$FC,$CB,$1C,$1B,$19,$8F,$F3,$A3,$CD,$7D
	.dc.b	$AF,$F9,$B5,$FE,$D9,$1F,$83,$77,$C7,$C8,$E7,$5D,$39,$24,$F4,$7C
	.dc.b	$3F,$3D,$E9,$CB,$FD,$3D,$83,$E6,$FC,$64,$BE,$07,$93,$E6,$FC,$E4
	.dc.b	$7C,$F1,$EC,$F8,$57,$35,$24,$FF,$A3,$FE,$D5,$F3,$5B,$E4,$93,$FE
	.dc.b	$93,$B4,$70,$8F,$F4,$77,$DA,$7C,$56,$EF,$F2,$A3,$1A,$FF,$AD,$F1
	.dc.b	$7D,$AE,$78,$66,$1F,$E9,$FB,$F9,$0F,$23,$FD,$45,$F6,$AF,$2F,$EA
	.dc.b	$BF,$C6,$E3,$3E,$5E,$27,$2C,$8C,$9C,$A3,$8F,$D9,$C8,$EA,$FE,$03
	.dc.b	$EA,$1F,$55,$DB,$C3,$95,$E2,$78,$3F,$EA,$DC,$5E,$C7,$90,$91,$B3
	.dc.b	$FD,$75,$FB,$6F,$F5,$76,$FA,$7F,$8E,$8F,$E7,$51,$AE,$FF,$C8,$FE
	.dc.b	$01,$FD,$F2,$3F,$D8,$4B,$F0,$1C,$9A,$F8,$4F,$CB,$FD,$27,$F8,$8A
	.dc.b	$BF,$D9,$0B,$C7,$C9,$AF,$8B,$F0,$74,$CA,$F2,$BE,$3F,$13,$C3,$EA
	.dc.b	$C9,$F0,$BE,$4F,$97,$EF,$E4,$C9,$EE,$F8,$BE,$4B,$D9,$E3,$C9,$F8
	.dc.b	$2F,$BF,$B3,$DB,$A4,$9F,$6B,$F0,$3C,$EF,$D8,$FC,$DB,$EA,$E9,$2B
	.dc.b	$FD,$C9,$0F,$C0,$3E,$57,$93,$F2,$FC,$99,$5F,$42,$3E,$5E,$9F,$33
	.dc.b	$8A,$4F,$CC,$8E,$FE,$CF,$83,$E7,$E4,$F6,$8F,$AD,$F7,$B2,$3E,$DB
	.dc.b	$F6,$7E,$2F,$F6,$F9,$C9,$FF,$7C,$79,$FE,$57,$EC,$F1,$FB,$3D,$C9
	.dc.b	$1E,$A7,$DF,$F0,$DF,$91,$F1,$FF,$09,$F0,$12,$BF,$CE,$DE,$6B,$0F
	.dc.b	$F1,$A7,$91,$FA,$39,$EF,$BF,$F0,$8E,$59,$18,$3E,$3F,$DF,$FC,$CF
	.dc.b	$7E,$46,$0E,$B8,$74,$F3,$3F,$90,$FD,$7F,$E4,$65,$9F,$85,$FF,$44
	.dc.b	$46,$63,$10,$CF,$F6,$8F,$0B,$FF,$25,$57,$FB,$5E,$3B,$9C,$BB,$BB
	.dc.b	$9E,$0F,$0E,$C7,$F3,$93,$45,$13,$CD,$D3,$5F,$E8,$0F,$E8,$DC,$F4
	.dc.b	$7F,$53,$8C,$46,$A3,$07,$C0,$E8,$91,$E6,$3B,$C7,$F4,$4F,$AE,$FC
	.dc.b	$A4,$65,$70,$64,$D7,$05,$9F,$44,$E9,$B3,$FD,$96,$31,$1E,$F7,$FC
	.dc.b	$DC,$F0,$DF,$44,$79,$F0,$78,$C8,$F8,$E2,$BC,$E3,$B1,$1F,$1F,$CE
	.dc.b	$F8,$48,$E0,$87,$D9,$3F,$D8,$D4,$FA,$1F,$84,$7C,$39,$F8,$9E,$AF
	.dc.b	$C5,$F7,$A5,$FC,$58,$AF,$B6,$3E,$01,$1F,$8B,$EE,$FC,$32,$38,$3E
	.dc.b	$49,$F4,$48,$7D,$16,$BE,$71,$87,$F6,$9E,$2F,$F4,$D3,$63,$D7,$3F
	.dc.b	$DF,$BC,$3F,$F5,$B4,$7E,$35,$AF,$E4,$5A,$FF,$82,$DA,$FF,$A9,$5D
	.dc.b	$F2,$FE,$89,$7D,$CF,$53,$B7,$CB,$F7,$F8,$F9,$7F,$8D,$AF,$F4,$9B
	.dc.b	$7E,$0B,$2F,$86,$7F,$C5,$53,$A8,$C1,$E7,$F2,$FD,$5E,$64,$8E,$91
	.dc.b	$5F,$54,$72,$23,$EF,$7F,$C3,$FF,$5A,$F9,$0A,$71,$D3,$D9,$E4,$7C
	.dc.b	$7F,$1A,$61,$E9,$8F,$94,$61,$FA,$59,$7F,$AF,$DD,$1F,$AB,$FA,$BF
	.dc.b	$E0,$B7,$FC,$21,$BC,$5F,$EC,$F6,$C8,$3F,$C7,$DF,$BC,$E9,$CF,$83
	.dc.b	$FE,$38,$8C,$1B,$AE,$DE,$B6,$49,$5E,$3F,$62,$BD,$22,$CF,$96,$6E
	.dc.b	$1E,$59,$B5,$E5,$B5,$62,$3C,$F3,$67,$31,$F0,$63,$07,$9E,$6C,$64
	.dc.b	$F2,$14,$F2,$16,$FC,$8D,$BF,$23,$4F,$24,$F4,$78,$9F,$21,$C4,$FC
	.dc.b	$48,$EE,$69,$C3,$67,$EA,$4F,$07,$BC,$73,$3C,$07,$C0,$2D,$FC,$07
	.dc.b	$91,$F1,$EF,$E2,$57,$38,$C1,$E2,$BF,$09,$2F,$BB,$F0,$23,$99,$DC
	.dc.b	$B7,$DE,$38,$47,$43,$35,$DD,$E1,$1C,$CD,$B8,$AE,$E7,$C0,$7D,$77
	.dc.b	$E8,$69,$EE,$AE,$97,$FE,$42,$8F,$55,$72,$7C,$57,$DA,$7C,$D3,$F2
	.dc.b	$A7,$FD,$2D,$1C,$5F,$EC,$0D,$73,$7F,$EA,$B8,$C4,$71,$21,$EF,$36
	.dc.b	$7F,$89,$BE,$1F,$FC,$56,$7C,$1F,$38,$E2,$3D,$67,$F2,$98,$CB,$C4
	.dc.b	$CE,$1F,$DF,$A3,$31,$B9,$E8,$73,$7E,$9D,$DA,$E1,$3A,$9E,$86,$8B
	.dc.b	$1A,$7B,$E1,$FC,$05,$5F,$82,$75,$1C,$4F,$11,$CA,$7E,$83,$F6,$CF
	.dc.b	$E1,$53,$C6,$3A,$36,$3E,$AC,$70,$8C,$47,$82,$BD,$0F,$63,$C1,$5E
	.dc.b	$22,$57,$83,$0F,$B5,$18,$9F,$0D,$72,$8C,$47,$94,$97,$E0,$3C,$35
	.dc.b	$CA,$FC,$97,$F2,$DE,$D8,$5B,$EA,$7C,$82,$9E,$7C,$3F,$21,$1B,$8F
	.dc.b	$61,$B2,$9B,$9E,$AF,$CB,$39,$CF,$27,$07,$07,$F3,$9F,$59,$FC,$C7
	.dc.b	$EB,$64,$FF,$3F,$8A,$FF,$38,$8B,$3F,$D0,$67,$F9,$C7,$A9,$F9,$D2
	.dc.b	$57,$F8,$FF,$24,$7F,$90,$61,$F5,$7F,$E6,$D7,$FC,$00,$DB,$FF,$29
	.dc.b	$B7,$1F,$E9,$95,$FD,$15,$95,$FD,$4D,$FE,$91,$E2,$E2,$3F,$63,$1F
	.dc.b	$AF,$23,$FE,$51,$F5,$9E,$A7,$8D,$B1,$98,$C1,$E7,$FA,$A7,$A9,$49
	.dc.b	$3F,$E5,$B8,$CB,$FE,$5D,$E1,$78,$3E,$F6,$C6,$63,$07,$5F,$AF,$91
	.dc.b	$CD,$AF,$F9,$A7,$EA,$FF,$CE,$AF,$EA,$C3,$87,$F5,$CD,$7F,$B8,$35
	.dc.b	$FD,$8F,$C2,$F8,$C8,$F9,$DF,$A9,$F7,$CF,$62,$92,$7F,$CF,$11,$96
	.dc.b	$FA,$7C,$08,$E0,$FF,$3F,$7D,$5F,$CE,$8B,$B1,$CF,$63,$C7,$CC,$7F
	.dc.b	$A1,$7E,$F9,$EB,$7A,$DB,$C9,$98,$C1,$EF,$FA,$BF,$E4,$71,$FE,$8D
	.dc.b	$6E,$7F,$E9,$16,$BF,$25,$AF,$F1,$55,$F2,$3C,$C9,$3C,$DE,$DE,$74
	.dc.b	$8E,$CD,$AE,$DE,$7F,$8C,$93,$E3,$7E,$A7,$C3,$FC,$18,$FF,$51,$36
	.dc.b	$7F,$D4,$CD,$7E,$F3,$5F,$E5,$5F,$54,$FB,$0F,$BC,$DA,$FF,$56,$36
	.dc.b	$7F,$D6,$0D,$7F,$11,$F7,$CF,$1F,$F0,$CD,$AF,$F5,$A3,$67,$FD,$70
	.dc.b	$D7,$F2,$14,$F6,$BC,$24,$AF,$0B,$F4,$FE,$8E,$47,$D3,$6D,$7D,$3F
	.dc.b	$57,$EC,$12,$7E,$C3,$EC,$F0,$64,$7B,$1B,$3E,$4F,$9E,$F4,$92,$7D
	.dc.b	$2F,$E8,$7E,$A5,$E9,$FF,$B7,$FC,$7F,$F6,$EC,$3D,$4C,$5F,$F5,$58
	.dc.b	$DB,$FE,$C7,$74,$FF,$B2,$7D,$67,$9B,$C3,$2F,$F2,$E4,$7F,$92,$27
	.dc.b	$53,$99,$FF,$44,$F9,$F7,$F0,$77,$3D,$E8,$C3,$70,$FF,$6B,$7C,$17
	.dc.b	$F0,$E6,$F9,$5C,$E3,$FD,$AB,$18,$3F,$DB,$9F,$86,$BE,$47,$C0,$D8
	.dc.b	$CF,$C9,$E2,$7C,$9F,$84,$3E,$E6,$98,$E5,$B7,$FE,$E7,$FA,$87,$7F
	.dc.b	$AD,$7D,$45,$8C,$3E,$FF,$C3,$1F,$15,$77,$C3,$70,$FF,$79,$7E,$12
	.dc.b	$F9,$5F,$8C,$BE,$92,$C6,$1F,$95,$F5,$5F,$53,$A3,$FD,$88,$DE,$D6
	.dc.b	$1F,$E0,$D8,$FA,$F3,$AE,$EF,$A5,$F8,$4F,$E8,$EF,$73,$FE,$93,$8C
	.dc.b	$5F,$F8,$55,$F8,$8F,$FA,$5E,$77,$29,$DB,$FE,$9B,$F8,$97,$CE,$FD
	.dc.b	$F6,$DF,$13,$FC,$40,$BE,$8B,$53,$F8,$5F,$C7,$DE,$BD,$31,$16,$F9
	.dc.b	$61,$FE,$A9,$F0,$7F,$83,$67,$E4,$6C,$FF,$8B,$5F,$E7,$0F,$09,$FC
	.dc.b	$BF,$A3,$F9,$3F,$C2,$9E,$D5,$31,$F1,$B6,$F8,$7E,$3E,$F4,$BF,$34
	.dc.b	$FE,$62,$33,$18,$F1,$3F,$3D,$F5,$6F,$99,$BF,$1B,$86,$DF,$8F,$1F
	.dc.b	$C2,$B7,$B3,$F4,$D1,$8D,$BF,$E0,$03,$FA,$15,$C1,$F4,$BE,$AB,$FC
	.dc.b	$13,$7E,$DF,$AA,$F3,$3E,$F8,$CD,$70,$8F,$F2,$84,$EA,$30,$7F,$B4
	.dc.b	$FE,$AB,$C1,$7F,$79,$F0,$37,$BF,$F7,$51,$8B,$FF,$2D,$2E,$9C,$F9
	.dc.b	$5A,$6C,$FD,$19,$FE,$A3,$7F,$8A,$FA,$AF,$F7,$14,$E7,$E3,$7F,$83
	.dc.b	$46,$32,$F2,$8B,$FE,$A7,$5B,$3A,$79,$BE,$4B,$F2,$F5,$FF,$0F,$F4
	.dc.b	$BF,$9C,$65,$EB,$8C,$7B,$3E,$AF,$7B,$78,$F9,$5B,$9F,$82,$FE,$D4
	.dc.b	$FE,$27,$0F,$E3,$93,$9E,$4E,$DE,$97,$F1,$DB,$7C,$F5,$9C,$37,$0F
	.dc.b	$C9,$5F,$F8,$89,$E9,$F0,$3F,$C9,$63,$07,$C0,$C7,$FC,$67,$3C,$3C
	.dc.b	$3D,$7C,$36,$DB,$1D,$3C,$3C,$34,$E0,$78,$37,$F8,$9F,$05,$DD,$B1
	.dc.b	$E4,$DF,$0D,$38,$1E,$8C,$9F,$55,$FE,$57,$69,$FE,$59,$70,$7F,$8C
	.dc.b	$D0,$FA,$87,$D1,$BE,$AB,$53,$98,$C3,$C6,$77,$3A,$6C,$67,$E5,$FF
	.dc.b	$46,$DE,$3D,$46,$23,$FC,$89,$28,$C4,$7F,$AC,$61,$E0,$FF,$AC,$EB
	.dc.b	$F8,$EC,$62,$7F,$84,$35,$FC,$9B,$CE,$3D,$BF,$54,$F8,$0E,$9C,$DC
	.dc.b	$F5,$B8,$72,$8C,$3F,$13,$0E,$98,$78,$39,$B3,$3F,$1D,$B3,$F8,$C8
	.dc.b	$FE,$35,$F4,$3F,$E5,$FA,$F3,$FD,$1E,$89,$3E,$8B,$E3,$F2,$F2,$3D
	.dc.b	$6E,$9F,$D6,$E7,$FC,$E1,$FE,$4C,$D7,$3C,$BE,$4F,$AA,$7B,$FE,$6A
	.dc.b	$FE,$6F,$5C,$39,$B0,$66,$FD,$7F,$55,$EE,$E0,$FF,$70,$36,$BF,$0A
	.dc.b	$D5,$8F,$F7,$97,$F9,$C4,$7F,$2B,$6F,$07,$43,$AF,$17,$C8,$FA,$9F
	.dc.b	$F3,$2C,$FF,$3F,$7F,$C7,$8D,$EF,$D7,$C6,$ED,$91,$93,$84,$73,$F7
	.dc.b	$64,$73,$7E,$AF,$D2,$BE,$AB,$C1,$E1,$CA,$F0,$3C,$1F,$E3,$38,$FF
	.dc.b	$B7,$DF,$D1,$B5,$34,$47,$E0,$BE,$AB,$F4,$F3,$FD,$65,$FF,$03,$37
	.dc.b	$DD,$D7,$77,$E2,$3E,$AB,$FC,$A5,$3F,$D7,$D7,$D5,$64,$D7,$C7,$7E
	.dc.b	$3B,$E6,$9F,$4F,$7C,$69,$87,$F6,$79,$CB,$E1,$E3,$F2,$11,$F2,$78
	.dc.b	$7E,$1F,$9C,$91,$E1,$7C,$9F,$53,$DF,$AA,$4F,$57,$C5,$F2,$3E,$FF
	.dc.b	$06,$4F,$BD,$F8,$5D,$5E,$DD,$24,$FB,$5F,$81,$E6,$FE,$A3,$E7,$1F
	.dc.b	$51,$49,$5F,$E0,$18,$FF,$74,$3F,$C5,$39,$9E,$27,$CF,$F1,$64,$FC
	.dc.b	$F8,$F9,$9A,$7C,$DE,$39,$3F,$36,$3C,$1E,$D7,$87,$EA,$24,$F7,$8F
	.dc.b	$B5,$F3,$B2,$3E,$D3,$F7,$7E,$93,$F7,$FE,$5F,$92,$4D,$31,$D7,$FE
	.dc.b	$30,$F2,$FF,$C6,$3B,$BD,$9F,$80,$4B,$F7,$7B,$5F,$96,$F1,$BF,$15
	.dc.b	$F0,$92,$7F,$C4,$5E,$37,$E9,$9F,$ED,$B9,$B7,$F1,$F2,$C9,$FE,$1F
	.dc.b	$1F,$37,$DE,$F9,$DE,$E4,$9F,$99,$1F,$A4,$F2,$BF,$9D,$C3,$1F,$D8
	.dc.b	$BF,$D0,$DA,$FF,$BD,$DA,$FF,$46,$1F,$53,$F9,$A7,$FA,$53,$5F,$F7
	.dc.b	$F3,$5E,$0D,$7F,$A7,$AE,$FF,$E5,$6D,$8F,$F2,$13,$5F,$F0,$6B,$DF
	.dc.b	$87,$FC,$BA,$FE,$82,$33,$CF,$C8,$E7,$D9,$63,$FA,$CC,$ED,$F2,$CF
	.dc.b	$C2,$8C,$47,$B7,$EA,$7F,$2A,$FF,$60,$6B,$FD,$9D,$AF,$C3,$0E,$1E
	.dc.b	$5B,$7F,$46,$8F,$B1,$1F,$E5,$D6,$BF,$E2,$87,$AE,$1F,$E7,$8F,$BD
	.dc.b	$FE,$91,$E0,$F8,$F7,$E4,$7F,$50,$8C,$DF,$F7,$8B,$FF,$1E,$3E,$79
	.dc.b	$FB,$B1,$88,$F4,$46,$37,$F2,$12,$3C,$91,$E7,$F8,$39,$1E,$77,$D5
	.dc.b	$EF,$62,$3E,$14,$76,$23,$F1,$FD,$4F,$09,$1C,$10,$F8,$1E,$E7,$C5
	.dc.b	$3E,$34,$7E,$BB,$BB,$E4,$7B,$72,$BF,$B5,$C5,$7E,$94,$7B,$88,$FD
	.dc.b	$77,$E8,$30,$FA,$0F,$F7,$8E,$27,$D5,$52,$4F,$FA,$46,$1F,$63,$88
	.dc.b	$FF,$2E,$C7,$07,$53,$CE,$7F,$C0,$77,$FE,$96,$9E,$5E,$FF,$73,$3F
	.dc.b	$11,$63,$FC,$57,$87,$FB,$ED,$B3,$FB,$48,$C4,$7F,$8D,$5F,$F1,$23
	.dc.b	$E5,$F3,$3D,$47,$F8,$4B,$EA,$DF,$7B,$89,$EB,$3C,$C8,$E0,$F7,$79
	.dc.b	$FF,$0C,$E0,$FB,$BC,$AF,$8D,$E5,$48,$F8,$91,$5F,$87,$17,$A7,$FD
	.dc.b	$9F,$92,$BF,$C4,$BF,$53,$FC,$66,$7F,$DA,$6D,$EA,$FE,$14,$DC,$3F
	.dc.b	$CA,$3E,$8B,$EA,$F8,$F8,$D2,$7F,$D8,$95,$B5,$5B,$73,$EB,$F8,$08
	.dc.b	$FF,$4B,$E2,$AF,$FA,$5A,$BF,$BC,$BC,$0D,$14,$D4,$F0,$CB,$F9,$DD
	.dc.b	$5C,$BD,$33,$6B,$FC,$D3,$E5,$7E,$05,$E2,$73,$BF,$83,$3F,$D8,$6F
	.dc.b	$84,$B0,$F1,$BC,$0F,$F5,$5D,$F3,$BE,$31,$CF,$0F,$0B,$E8,$38,$46
	.dc.b	$9B,$1F,$CD,$E7,$93,$C2,$32,$7F,$A0,$2B,$C3,$5E,$E8,$EC,$67,$0F
	.dc.b	$3D,$F2,$8C,$1A,$3D,$51,$ED,$71,$5D,$2F,$C1,$7D,$67,$C2,$70,$39
	.dc.b	$5F,$57,$06,$AB,$93,$F3,$B0,$F9,$31,$88,$DD,$75,$3E,$11,$2F,$C2
	.dc.b	$7C,$3A,$E3,$87,$B1,$B5,$E2,$6B,$E6,$C3,$C9,$87,$72,$57,$BE,$7C
	.dc.b	$CF,$37,$E0,$B5,$C9,$D4,$C9,$4F,$C0,$1F,$D2,$E3,$F0,$6D,$8F,$EE
	.dc.b	$F1,$88,$FB,$46,$4F,$F5,$AB,$97,$EA,$9F,$EB,$D7,$D1,$F1,$DF,$B4
	.dc.b	$3D,$B8,$FE,$D1,$C4,$F9,$0F,$D3,$38,$7D,$26,$0F,$D9,$BC,$B1,$FD
	.dc.b	$DB,$B9,$FA,$13,$E2,$7B,$3C,$0E,$CE,$1E,$87,$67,$04,$9F,$B3,$1C
	.dc.b	$DE,$EF,$33,$83,$87,$9C,$78,$9F,$C2,$C6,$63,$A4,$73,$21,$FB,$39
	.dc.b	$F5,$4F,$82,$38,$9F,$A7,$8C,$BC,$2B,$E6,$BE,$3A,$F0,$46,$23,$C7
	.dc.b	$5C,$E3,$13,$E2,$9E,$A4,$8F,$F4,$81,$2F,$F3,$F7,$E0,$9D,$45,$9D
	.dc.b	$47,$EC,$27,$13,$C2,$3C,$97,$E1,$8E,$13,$F6,$67,$D8,$68,$F6,$46
	.dc.b	$9F,$DD,$FD,$1F,$F3,$CF,$AB,$91,$FC,$D2,$3F,$9F,$98,$7F,$97,$46
	.dc.b	$1F,$F3,$E7,$DC,$FF,$94,$33,$34,$91,$F9,$AF,$1F,$FA,$49,$1E,$20
	.dc.b	$68,$CF,$FC,$A9,$F5,$8F,$3B,$FD,$EA,$DA,$E1,$5B,$79,$46,$5F,$E5
	.dc.b	$FF,$49,$C2,$BF,$CC,$B3,$99,$FF,$1F,$35,$EF,$1F,$B4,$F5,$24,$7E
	.dc.b	$C9,$AF,$F9,$14,$7D,$E3,$C1,$FF,$38,$B6,$34,$6A,$32,$7F,$9D,$BE
	.dc.b	$B7,$FC,$8F,$1F,$E4,$F6,$BF,$E7,$36,$BF,$15,$AB,$67,$C6,$7D,$12
	.dc.b	$47,$82,$32,$F8,$BF,$08,$FF,$0F,$3E,$96,$BF,$E5,$F6,$BF,$E8,$36
	.dc.b	$BF,$D8,$BE,$1F,$EF,$11,$EC,$FF,$A4,$33,$78,$74,$93,$F0,$A3,$2F
	.dc.b	$C4,$FA,$D7,$B5,$FE,$98,$6C,$66,$30,$7F,$A7,$3E,$F5,$F2,$D4,$93
	.dc.b	$F3,$A3,$2F,$D0,$FA,$E7,$BE,$F6,$F0,$D9,$FB,$52,$F1,$3E,$C4,$7F
	.dc.b	$A0,$63,$8D,$7D,$AF,$BD,$FF,$4E,$74,$DE,$AF,$C8,$5F,$FA,$1E,$72
	.dc.b	$FE,$15,$AF,$FA,$2F,$EB,$9F,$29,$C2,$3F,$D6,$11,$8B,$FC,$3B,$63
	.dc.b	$F1,$CD,$7B,$7C,$2F,$17,$F5,$16,$8C,$42,$31,$5E,$99,$CD,$3D,$3A
	.dc.b	$49,$F0,$BA,$DE,$47,$91,$B8,$79,$3B,$F9,$92,$7D,$2F,$8B,$EB,$5E
	.dc.b	$15,$24,$FE,$92,$32,$FC,$DE,$0F,$4E,$9F,$4E,$30,$4C,$3F,$61,$4F
	.dc.b	$87,$EC,$49,$E8,$F2,$F7,$64,$7D,$96,$D7,$CC,$F5,$7E,$19,$27,$E8
	.dc.b	$BF,$2B,$EB,$9E,$45,$24,$FE,$F2,$32,$FE,$FB,$EB,$9F,$43,$FC,$4F
	.dc.b	$C8,$FC,$2C,$3D,$6C,$57,$FB,$BE,$36,$FF,$1A,$31,$1A,$CB,$F9,$27
	.dc.b	$D7,$76,$7F,$92,$CF,$FB,$DE,$1E,$66,$E5,$1B,$C3,$F2,$AE,$AB,$B4
	.dc.b	$FF,$2F,$FA,$EF,$C1,$46,$21,$0F,$73,$A3,$A8,$F0,$9C,$63,$C1,$F0
	.dc.b	$FF,$C0,$2B,$93,$67,$61,$7C,$37,$0F,$17,$D1,$F5,$CF,$AE,$A6,$3E
	.dc.b	$62,$E3,$E5,$FA,$E7,$D7,$7B,$DB,$19,$FA,$EC,$4F,$C3,$FA,$E7,$CC
	.dc.b	$DF,$AE,$C3,$6F,$E0,$7D,$77,$F4,$D6,$EB,$F2,$E3,$1C,$1E,$A7,$D9
	.dc.b	$C9,$FC,$03,$EB,$9F,$35,$F4,$57,$D7,$58,$C3,$F5,$7E,$BB,$EF,$E1
	.dc.b	$F1,$1B,$7F,$C2,$9F,$C3,$65,$F4,$DC,$BF,$D9,$65,$F5,$DF,$A8,$7E
	.dc.b	$6B,$6B,$F8,$9B,$F1,$23,$3E,$37,$E1,$7E,$39,$EB,$53,$1E,$CD,$C7
	.dc.b	$F0,$5F,$5C,$FA,$3F,$C6,$36,$33,$58,$9F,$C8,$78,$BF,$CB,$67,$F1
	.dc.b	$4D,$CB,$F0,$2F,$F3,$E8,$CF,$A7,$F9,$7F,$AE,$7C,$2D,$31,$F2,$56
	.dc.b	$BF,$29,$F5,$DF,$D0,$1B,$E9,$FE,$7E,$31,$97,$E3,$9F,$CB,$71,$7F
	.dc.b	$34,$FA,$E7,$CF,$FE,$95,$7C,$7D,$8C,$3F,$A9,$CB,$EF,$D7,$F3,$C7
	.dc.b	$EF,$E5,$F4,$5C,$B7,$AF,$F6,$1F,$5D,$FC,$BB,$3F,$CF,$FD,$77,$93
	.dc.b	$0F,$E8,$B3,$97,$FA,$3C,$66,$FE,$E7,$D6,$F4,$BF,$F1,$13,$E0,$6F
	.dc.b	$C1,$FE,$F6,$31,$05,$D1,$9F,$2B,$4D,$BF,$DE,$4F,$F4,$BC,$7F,$AB
	.dc.b	$FD,$77,$F8,$BE,$73,$EF,$7F,$09,$8C,$63,$FB,$8B,$FE,$05,$5A,$39
	.dc.b	$F9,$BE,$3B,$F2,$74,$FF,$11,$F4,$5F,$3F,$8B,$FE,$19,$18,$7C,$7F
	.dc.b	$13,$9C,$BB,$37,$4F,$ED,$67,$D8,$9E,$3E,$9F,$F1,$D8,$C5,$7F,$0C
	.dc.b	$BF,$E3,$8F,$F0,$E8,$7D,$26,$F9,$F8,$37,$0F,$E3,$D4,$F9,$BF,$47
	.dc.b	$C2,$FF,$29,$8C,$4F,$A2,$31,$E1,$E6,$AC,$F2,$78,$78,$58,$BF,$E3
	.dc.b	$E7,$EB,$A7,$BD,$75,$DF,$E3,$7B,$F6,$7E,$2B,$E1,$DF,$17,$C9,$BA
	.dc.b	$FF,$7F,$FA,$EF,$F2,$5C,$FF,$99,$DA,$9A,$7F,$94,$3E,$B1,$F5,$0F
	.dc.b	$AE,$C3,$6B,$FC,$B5,$3A,$9C,$9C,$7E,$6F,$FC,$E7,$3A,$6B,$FC,$4C
	.dc.b	$4A,$33,$1F,$C5,$E3,$18,$78,$FC,$1F,$E3,$4F,$F8,$A2,$3F,$67,$3F
	.dc.b	$D4,$5A,$FF,$48,$78,$AF,$AE,$7C,$0F,$5E,$5E,$7B,$DA,$E5,$18,$7F
	.dc.b	$D1,$97,$D3,$0F,$07,$2E,$67,$D2,$D9,$FC,$4C,$7F,$63,$FA,$6F,$75
	.dc.b	$79,$FE,$73,$44,$9F,$45,$F1,$F9,$D9,$1E,$B7,$4F,$F6,$2C,$FF,$98
	.dc.b	$BF,$DB,$DA,$E7,$9B,$C9,$F5,$CF,$77,$CD,$5F,$CC,$EB,$87,$3E,$0C
	.dc.b	$DF,$AF,$EB,$BD,$DC,$1F,$D1,$DB,$5F,$EA,$D6,$AC,$7F,$B8,$3F,$E0
	.dc.b	$18,$FE,$4E,$DE,$0E,$87,$5E,$2F,$91,$F5,$AF,$9F,$BE,$04,$BF,$F0
	.dc.b	$54,$E5,$B1,$97,$FC,$07,$1F,$C1,$9D,$CF,$4F,$6E,$47,$27,$EB,$7D
	.dc.b	$4B,$EB,$BC,$3E,$1C,$AE,$EF,$07,$AF,$AB,$E6,$23,$A3,$FC,$3D,$F0
	.dc.b	$BF,$E7,$EE,$8F,$F3,$CB,$78,$7F,$B7,$36,$7F,$DA,$8F,$D1,$7F,$C3
	.dc.b	$78,$7F,$A2,$5B,$EB,$7E,$2E,$31,$7F,$E8,$47,$FD,$37,$3F,$E2,$2F
	.dc.b	$A4,$CE,$1F,$DE,$DB,$1F,$EE,$16,$A7,$8D,$7F,$A3,$BE,$B7,$F0,$93
	.dc.b	$EA,$3F,$35,$8F,$E5,$AF,$FC,$0D,$3A,$9F,$F1,$8F,$D6,$7C,$F6,$EF
	.dc.b	$FA,$18,$FA,$D1,$AE,$1F,$D2,$7D,$26,$F2,$FF,$4E,$BA,$6D,$EA,$7F
	.dc.b	$55,$83,$E9,$EF,$C4,$CC,$7F,$C9,$93,$97,$FC,$9D,$93,$E1,$7C,$5B
	.dc.b	$C9,$CC,$78,$8F,$1C,$48,$FF,$2B,$69,$C5,$EB,$F1,$9E,$22,$3E,$0F
	.dc.b	$A8,$F7,$72,$48,$F0,$BF,$07,$9B,$DE,$C8,$EA,$FC,$28,$F6,$FD,$8F
	.dc.b	$BB,$EA,$65,$7B,$23,$E5,$7C,$67,$CB,$E8,$93,$EE,$8F,$96,$7C,$C8
	.dc.b	$91,$EA,$8F,$93,$F2,$32,$3E,$3B,$F3,$3D,$3F,$AD,$F0,$B2,$7E,$4B
	.dc.b	$F6,$7E,$37,$EB,$B2,$3E,$2B,$F6,$A3,$F8,$F7,$D6,$FF,$9A,$27,$F9
	.dc.b	$C0,$0E,$5C,$B5,$FE,$7C,$D7,$FC,$A4,$FF,$AF,$1E,$67,$29,$FE,$AB
	.dc.b	$19,$8E,$71,$FD,$63,$93,$FC,$CB,$1F,$84,$F8,$AF,$C4,$9E,$69,$F3
	.dc.b	$FC,$E4,$9F,$F4,$A5,$A3,$D7,$F2,$D9,$9F,$3C,$78,$08,$F9,$25,$F2
	.dc.b	$DE,$42,$46,$5A,$FF,$A5,$A1,$F4,$BF,$20,$F9,$33,$FD,$1F,$EA,$DF
	.dc.b	$5B,$7C,$29,$87,$FA,$A6,$72,$D8,$FE,$EF,$1F,$EA,$08,$F8,$ED,$7E
	.dc.b	$9B,$5F,$ED,$9F,$5B,$FE,$64,$7F,$D6,$8D,$7F,$D5,$6D,$F0,$9E,$1F
	.dc.b	$FA,$16,$BA,$46,$2B,$EF,$43,$E2,$7D,$44,$8F,$47,$C2,$57,$C7,$1E
	.dc.b	$82,$3E,$8F,$8C,$F5,$91,$C1,$0F,$F5,$FD,$3E,$7B,$E4,$1F,$26,$7E
	.dc.b	$57,$A5,$F2,$FE,$06,$5F,$C9,$8A,$FC,$48,$F9,$24,$7E,$4B,$FC,$67
	.dc.b	$EB,$64,$7D,$18,$FA,$CC,$3E,$A1,$F6,$7E,$03,$0F,$98,$BA,$7C,$75
	.dc.b	$97,$F4,$15,$F0,$2F,$0F,$FC,$10,$FF,$BB,$A3,$FD,$D0,$D7,$FD,$BE
	.dc.b	$ED,$CF,$89,$FE,$38,$8D,$47,$F8,$A5,$AF,$FA,$F5,$AF,$ED,$57,$77
	.dc.b	$80,$91,$E0,$EE,$AF,$EF,$27,$6F,$ED,$1A,$E6,$7C,$47,$F8,$87,$E3
	.dc.b	$3F,$DF,$D1,$FF,$00,$5A,$FA,$A3,$DF,$E9,$7F,$89,$FC,$F7,$89,$4C
	.dc.b	$D7,$F0,$BD,$7E,$54,$EA,$31,$CD,$FD,$7E,$B8,$47,$C8,$BF,$97,$F5
	.dc.b	$D2,$3E,$5B,$F3,$BD,$AC,$47,$CC,$8B,$8F,$D7,$D1,$5F,$E3,$BF,$AE
	.dc.b	$FE,$B1,$97,$FB,$9D,$B8,$7F,$1F,$6A,$7F,$AE,$46,$FD,$1E,$44,$9F
	.dc.b	$49,$C7,$D2,$C2,$C7,$F1,$17,$D0,$93,$A8,$CD,$5B,$F2,$4D,$BF,$24
	.dc.b	$D7,$C9,$16,$FA,$4F,$B6,$30,$E5,$B3,$E6,$3D,$91,$B7,$13,$C9,$B1
	.dc.b	$97,$D5,$5F,$05,$C9,$0C,$9F,$04,$C5,$FB,$1B,$1E,$C6,$9E,$C2,$9E
	.dc.b	$C2,$D7,$8D,$EA,$64,$AE,$67,$B4,$62,$FD,$EE,$AF,$D5,$3A,$25,$FA
	.dc.b	$E7,$81,$E0,$9F,$90,$D3,$C7,$5D,$2B,$8C,$7B,$CE,$C6,$CE,$66,$08
	.dc.b	$F0,$7D,$4D,$BF,$55,$FC,$59,$DB,$E5,$9D,$57,$86,$30,$6E,$3E,$0C
	.dc.b	$FF,$AB,$EB,$FC,$74,$DB,$FF,$61,$C6,$23,$E0,$19,$3F,$BA,$B9,$7E
	.dc.b	$A1,$FE,$02,$7F,$47,$E6,$BE,$BF,$FC,$82,$78,$E7,$FC,$0D,$19,$7F
	.dc.b	$C6,$31,$95,$19,$7E,$DB,$6F,$E1,$5F,$49,$E2,$7C,$B8,$FF,$26,$D6
	.dc.b	$63,$E2,$9E,$68,$CB,$C0,$C9,$F9,$E3,$0E,$9C,$DF,$B2,$7A,$4F,$8A
	.dc.b	$77,$87,$F8,$85,$F4,$CF,$AF,$0F,$44,$E5,$B3,$E3,$8F,$AA,$FC,$49
	.dc.b	$EB,$18,$8F,$21,$F0,$A3,$2F,$12,$1E,$03,$C6,$62,$FD,$0E,$0E,$77
	.dc.b	$F8,$DB,$F1,$C7,$13,$DF,$87,$67,$13,$EF,$7C,$47,$3A,$FC,$9B,$E9
	.dc.b	$31,$3F,$C4,$6B,$A9,$DA,$77,$3D,$30,$F8,$98,$7C,$73,$C1,$3A,$AE
	.dc.b	$31,$9A,$EA,$7F,$02,$8C,$BF,$C0,$BC,$BF,$F7,$97,$D5,$3E,$26,$9F
	.dc.b	$CE,$E2,$4F,$F1,$D9,$D3,$67,$2F,$F0,$8E,$BF,$D1,$91,$FD,$0B,$FE
	.dc.b	$56,$5F,$85,$7F,$75,$7F,$88,$B7,$5F,$F0,$13,$74,$FE,$95,$5F,$87
	.dc.b	$9F,$F0,$37,$D7,$3E,$CF,$15,$F8,$B8,$C6,$1F,$CC,$DA,$FF,$1C,$FA
	.dc.b	$8D,$BF,$E7,$18,$CB,$F9,$30,$E1,$E9,$5E,$98,$C4,$7E,$D3,$CC,$F6
	.dc.b	$6C,$69,$23,$F8,$0E,$1F,$C1,$88,$E0,$FF,$3C,$CB,$E3,$79,$92,$32
	.dc.b	$FC,$9D,$9E,$0D,$24,$FF,$A0,$A3,$2F,$FA,$13,$EB,$FE,$82,$69,$1F
	.dc.b	$A9,$F5,$FF,$E5,$D7,$E1,$35,$FD,$B3,$5F,$F4,$C3,$57,$D9,$FF,$8E
	.dc.b	$5F,$94,$D7,$F7,$CD,$7F,$D4,$0D,$5F,$2B,$F3,$1B,$B7,$E1,$E3,$06
	.dc.b	$7E,$BF,$FD,$F7,$3F,$5D,$AF,$F0,$B6,$BF,$EA,$D6,$A5,$F0,$1F,$93
	.dc.b	$6C,$7F,$AA,$9A,$FF,$AD,$9A,$FE,$17,$EB,$7E,$B4,$FF,$AC,$1B,$1F
	.dc.b	$EB,$C7,$FC,$AF,$E4,$67,$6E,$33,$FE,$BF,$FB,$37,$6F,$C3,$6F,$17
	.dc.b	$FA,$E9,$B3,$FE,$C6,$6B,$F9,$5F,$AE,$FD,$B4,$FF,$B0,$1B,$1F,$EC
	.dc.b	$C6,$BB,$EA,$FA,$10,$F9,$6A,$4A,$E4,$FD,$5F,$79,$E1,$D2,$4F,$FB
	.dc.b	$3A,$32,$F8,$7C,$FE,$A4,$70,$6F,$7E,$84,$70,$7F,$B6,$3E,$B1,$ED
	.dc.b	$7E,$15,$B1,$98,$C1,$F8,$9F,$AC,$78,$B4,$93,$FE,$E2,$8C,$BD,$7B
	.dc.b	$7A,$11,$C1,$F2,$7E,$CB,$E2,$91,$C1,$CF,$E0,$BC,$04,$70,$7F,$BC
	.dc.b	$3E,$B7,$FC,$E6,$FF,$90,$5B,$C5,$FE,$2C,$5F,$07,$FE,$C3,$F1,$3F
	.dc.b	$D3,$D1,$A7,$FD,$41,$E9,$38,$75,$87,$F6,$58,$C3,$FE,$4D,$5F,$0B
	.dc.b	$AC,$42,$36,$FF,$8F,$3E,$B7,$F0,$B8,$7F,$AB,$5B,$7F,$DC,$57,$D1
	.dc.b	$F3,$A4,$69,$EB,$F3,$FF,$EF,$0A,$FF,$2F,$3F,$CC,$5B,$EF,$F1,$8C
	.dc.b	$4F,$F9,$31,$F2,$CF,$0E,$ED,$4E,$DF,$07,$D6,$3C,$EB,$C3,$86,$DF
	.dc.b	$F9,$B7,$EB,$1F,$15,$E7,$6C,$67,$9F,$13,$EB,$FA,$C7,$C1,$D3,$1B
	.dc.b	$5B,$FF,$3C,$7D,$67,$FA,$F9,$BE,$DF,$C3,$87,$BB,$63,$E3,$7D,$5F
	.dc.b	$D9,$AF,$F0,$D3,$F7,$9F,$F4,$13,$9E,$1F,$AA,$E5,$FF,$37,$FD,$5B
	.dc.b	$E3,$BE,$83,$F0,$6B,$FD,$14,$E5,$BC,$1F,$57,$E1,$BF,$A0,$C7,$FA
	.dc.b	$3D,$7C,$26,$7D,$5D,$3F,$6F,$EA,$DF,$49,$67,$0D,$BF,$F4,$C6,$1F
	.dc.b	$BC,$CF,$FD,$06,$FF,$AF,$23,$3E,$4F,$E1,$63,$13,$FE,$88,$FA,$B6
	.dc.b	$7F,$89,$5D,$76,$30,$FE,$37,$EA,$FF,$89,$63,$FE,$A7,$6D,$FF,$A4
	.dc.b	$9F,$C8,$27,$3D,$E9,$FF,$24,$7D,$5B,$E9,$3F,$30,$D8,$FF,$59,$2F
	.dc.b	$A6,$CC,$FE,$77,$E3,$7F,$07,$5F,$EB,$76,$E3,$FE,$A2,$7F,$3C,$DE
	.dc.b	$AF,$D2,$7D,$53,$D0,$A7,$E9,$FE,$AD,$F0,$58,$C7,$FD,$85,$F5,$4F
	.dc.b	$8D,$FD,$63,$63,$3E,$EE,$27,$3C,$52,$78,$F1,$DD,$7F,$D6,$B8,$7C
	.dc.b	$4F,$AB,$FD,$5C,$63,$D9,$FD,$C4,$63,$93,$E3,$4F,$F0,$4A,$F9,$38
	.dc.b	$7F,$9B,$DF,$9A,$62,$35,$7E,$6F,$A3,$FE,$05,$E1,$7F,$2A,$9F,$0C
	.dc.b	$63,$C7,$F5,$F9,$7B,$C7,$DE,$DA,$FF,$65,$E4,$FA,$0E,$7E,$5F,$F0
	.dc.b	$D8,$C5,$FF,$B9,$F2,$4E,$7E,$97,$57,$6B,$FD,$D9,$7C,$30,$FF,$6B
	.dc.b	$9F,$A2,$3F,$DB,$38,$7F,$8A,$67,$FD,$E5,$19,$9F,$86,$FE,$EB,$DF
	.dc.b	$FE,$3C,$FE,$92,$77,$E1,$63,$C2,$CB,$AF,$A4,$C4,$F4,$6D,$78,$B7
	.dc.b	$F8,$FF,$1F,$D1,$B5,$E2,$DF,$0D,$3C,$B1,$E2,$A7,$D5,$FF,$8D,$88
	.dc.b	$FF,$99,$DF,$F2,$A3,$FC,$FB,$EA,$7F,$39,$F5,$78,$84,$61,$DB,$5F
	.dc.b	$F2,$E3,$5F,$F5,$83,$FA,$9D,$B9,$36,$38,$F8,$29,$CB,$63,$AF,$C9
	.dc.b	$FF,$9A,$9F,$0B,$68,$D9,$E1,$F5,$7F,$C5,$AB,$D5,$C7,$B2,$3F,$D6
	.dc.b	$B0,$FA,$F8,$72,$8C,$69,$E8,$E4,$EC,$48,$C9,$FD,$8B,$EA,$FF,$CF
	.dc.b	$57,$CB,$EA,$FC,$84,$9D,$9F,$22,$BD,$1F,$2D,$23,$38,$A3,$84,$7C
	.dc.b	$3C,$7F,$DB,$1E,$73,$93,$E1,$36,$7A,$B5,$E0,$7F,$76,$FA,$A7,$D5
	.dc.b	$F0,$E4,$93,$C1,$E7,$7F,$37,$EA,$FE,$14,$F2,$0E,$36,$FC,$EB,$72
	.dc.b	$FF,$76,$62,$FA,$8E,$5C,$33,$0F,$83,$7F,$B5,$6F,$27,$DB,$FA,$BF
	.dc.b	$F2,$AC,$7C,$67,$FC,$5E,$DE,$FD,$7C,$76,$D9,$19,$38,$47,$3F,$83
	.dc.b	$91,$CD,$FC,$47,$D5,$FF,$7A,$8E,$DE,$17,$84,$8E,$CF,$17,$A9,$E5
	.dc.b	$23,$C0,$FB,$7E,$7B,$BF,$F2,$8D,$8C,$C6,$39,$3F,$31,$EF,$7F,$9F
	.dc.b	$B9,$3F,$82,$B7,$93,$FC,$57,$7F,$E0,$EF,$8C,$DF,$D8,$FE,$7F,$E7
	.dc.b	$F5,$3F,$E3,$06,$A6,$88,$FD,$37,$D4,$FE,$02,$7C,$E7,$98,$CF,$57
	.dc.b	$EA,$FE,$AF,$E7,$C7,$DF,$7E,$9B,$78,$F5,$C9,$FB,$1F,$9B,$DE,$DF
	.dc.b	$C7,$9D,$B6,$75,$3E,$2F,$AC,$F1,$F5,$CC,$3F,$47,$9A,$7C,$5F,$17
	.dc.b	$E3,$E1,$95,$FA,$8D,$3F,$CA,$2F,$EA,$B4,$AF,$E5,$5F,$B8,$D9,$F3
	.dc.b	$57,$C2,$F8,$B9,$3E,$07,$E2,$FD,$1F,$C1,$E8,$93,$DD,$F8,$9D,$5E
	.dc.b	$F6,$47,$85,$F8,$51,$F3,$BE,$8F,$E7,$FC,$7C,$AF,$6C,$7C,$DF,$85
	.dc.b	$F9,$DD,$92,$7D,$91,$F2,$CF,$99,$12,$3D,$D1,$F6,$BE,$1E,$47,$C8
	.dc.b	$7E,$EF,$A7,$F6,$7E,$4E,$4F,$C7,$7E,$E7,$D6,$7D,$76,$47,$C9,$7E
	.dc.b	$D4,$7F,$2B,$FA,$A3,$49,$3F,$9E,$F2,$2B,$FC,$0E,$7F,$D5,$2F,$F1
	.dc.b	$C6,$B9,$9E,$95,$FD,$1E,$13,$A7,$F3,$E6,$23,$51,$98,$FD,$1A,$F9
	.dc.b	$4C,$92,$3F,$A9,$46,$5B,$F5,$7E,$B2,$3E,$59,$F7,$7D,$54,$8F,$43
	.dc.b	$F0,$7D,$6C,$47,$F6,$0B,$4F,$90,$91,$FD,$87,$EA,$1F,$21,$ED,$6C
	.dc.b	$66,$30,$47,$FA,$0C,$7F,$A1,$1B,$E4,$7F,$5B,$8E,$31,$FC,$BD,$D3
	.dc.b	$FD,$C9,$AF,$C8,$6B,$FC,$21,$AF,$FA,$D1,$FF,$32,$BE,$7F,$7F,$CC
	.dc.b	$47,$D5,$F0,$FE,$12,$3E,$68,$F2,$7D,$4C,$8F,$4B,$E7,$F9,$BC,$47
	.dc.b	$96,$3E,$11,$1F,$8D,$E9,$FC,$52,$38,$21,$EF,$F7,$7E,$11,$F0,$E2
	.dc.b	$C7,$C8,$8A,$FC,$58,$F9,$1E,$77,$D1,$24,$7D,$56,$BF,$2C,$F9,$C7
	.dc.b	$03,$FB,$9E,$EE,$DF,$F0,$63,$63,$31,$8F,$0F,$2F,$E6,$E7,$53,$FE
	.dc.b	$26,$6B,$F9,$C6,$BF,$CF,$3E,$13,$FD,$BE,$F8,$3A,$3B,$9F,$E1,$B8
	.dc.b	$CC,$7F,$6F,$9F,$F6,$0B,$FE,$2F,$6B,$DE,$7C,$44,$7D,$7D,$3E,$A3
	.dc.b	$AB,$EE,$F7,$3D,$BE,$9C,$8C,$10,$F0,$39,$7C,$51,$F1,$3D,$79,$1F
	.dc.b	$21,$F8,$FE,$D6,$23,$D1,$17,$CD,$FF,$30,$E4,$AF,$F1,$C7,$D4,$FF
	.dc.b	$46,$F4,$FF,$AF,$B7,$00,$70,$BC,$A7,$D3,$F2,$F4,$24,$FF,$8B,$2B
	.dc.b	$75,$A8,$C4,$E9,$FF,$0D,$4E,$AB,$19,$7F,$9D,$A3,$B9,$99,$E2,$43
	.dc.b	$8D,$7F,$61,$AE,$38,$7A,$A2,$CF,$AB,$0F,$13,$C0,$E5,$87,$79,$E5
	.dc.b	$18,$AE,$4D,$39,$5F,$47,$E2,$4D,$BE,$B7,$CE,$78,$1A,$3F,$B3,$57
	.dc.b	$1C,$3D,$B7,$E0,$9F,$21,$2B,$93,$FD,$AE,$B8,$4F,$76,$C7,$6B,$F0
	.dc.b	$E1,$E8,$8E,$4D,$BF,$F0,$25,$70,$9C,$9E,$C3,$CF,$87,$91,$F3,$98
	.dc.b	$C3,$C0,$D3,$C1,$3C,$EB,$CE,$76,$8D,$1D,$8F,$C4,$46,$EF,$C8,$7F
	.dc.b	$8B,$B0,$F4,$CF,$E4,$30,$F1,$3C,$63,$2F,$CB,$3A,$47,$E1,$5F,$92
	.dc.b	$7F,$5A,$7F,$C4,$ED,$C3,$FB,$04,$62,$3E,$31,$93,$FD,$54,$E5,$F1
	.dc.b	$47,$03,$F4,$FE,$84,$8F,$6E,$3F,$97,$8C,$47,$92,$6C,$E6,$37,$87
	.dc.b	$EB,$AB,$2F,$C2,$8E,$E7,$EA,$63,$2E,$0F,$7B,$87,$E1,$4F,$D0,$9E
	.dc.b	$91,$C4,$EE,$7C,$B9,$E1,$7F,$32,$BC,$93,$FB,$29,$FA,$8E,$CE,$47
	.dc.b	$AC,$B1,$C0,$F0,$3E,$57,$15,$E2,$8E,$07,$53,$8B,$EF,$7D,$2E,$E3
	.dc.b	$8C,$73,$8D,$57,$F9,$71,$D1,$8B,$F5,$DF,$A2,$BF,$8B,$3E,$E8,$FD
	.dc.b	$73,$63,$F3,$AD,$7E,$59,$E4,$3C,$04,$39,$99,$3C,$E6,$6B,$C8,$FC
	.dc.b	$73,$DF,$7E,$A7,$89,$8B,$F4,$BB,$8F,$92,$FF,$2D,$F2,$BF,$99,$F8
	.dc.b	$F2,$3F,$95,$47,$F3,$63,$0F,$F2,$88,$C3,$FC,$BF,$EC,$FF,$9D,$91
	.dc.b	$FE,$65,$E2,$D7,$F9,$7C,$5B,$FE,$5D,$3F,$D0,$BC,$59,$5F,$CE,$5F
	.dc.b	$F7,$E7,$D3,$BE,$C7,$33,$FE,$46,$3F,$A2,$4F,$EB,$08,$FE,$AB,$E2
	.dc.b	$FF,$5E,$47,$07,$11,$F5,$2F,$8C,$FF,$2A,$36,$34,$D7,$C0,$D7,$F7
	.dc.b	$DF,$5B,$FB,$C2,$3F,$BA,$7F,$CB,$DF,$53,$FE,$32,$83,$49,$E4,$6B
	.dc.b	$BA,$4F,$96,$51,$88,$F8,$1A,$3E,$27,$FC,$D4,$D8,$F6,$35,$DC,$7F
	.dc.b	$9B,$FE,$CF,$FC,$23,$1F,$DB,$1A,$9A,$63,$EA,$9C,$BF,$E7,$76,$CF
	.dc.b	$C9,$6B,$F4,$DA,$FF,$9E,$B5,$F4,$BF,$3D,$AF,$D9,$6B,$BC,$7F,$D0
	.dc.b	$9F,$53,$FE,$D7,$8F,$AC,$D7,$F0,$0D,$7F,$D0,$ED,$7F,$DC,$FF,$66
	.dc.b	$F9,$4F,$F4,$73,$63,$F0,$0D,$7F,$14,$D7,$FD,$23,$F5,$4F,$90,$7C
	.dc.b	$27,$E1,$E3,$0D,$F0,$BF,$DF,$95,$C2,$B6,$FE,$57,$EC,$DC,$FF,$12
	.dc.b	$3F,$C3,$71,$8D,$FF,$24,$D8,$E1,$3B,$5F,$54,$F9,$3C,$36,$FF,$34
	.dc.b	$D9,$FD,$23,$5F,$F5,$5A,$F4,$78,$92,$78,$BD,$3D,$19,$1D,$1B,$5D
	.dc.b	$37,$F0,$92,$7C,$2F,$93,$B2,$47,$91,$B3,$FB,$7F,$AA,$77,$7B,$5B
	.dc.b	$5F,$B3,$6C,$EE,$7E,$16,$AD,$E9,$2F,$E4,$BC,$FD,$99,$1C,$DB,$7F
	.dc.b	$C1,$FE,$A9,$F0,$1F,$35,$B5,$FC,$0D,$B3,$B9,$FA,$5F,$52,$F8,$0F
	.dc.b	$D0,$F8,$9F,$A0,$8C,$61,$FE,$56,$8C,$7C,$0E,$A3,$FC,$B3,$82,$B6
	.dc.b	$AB,$F5,$3F,$05,$E0,$6E,$3C,$67,$FC,$C3,$2F,$83,$C4,$23,$86,$1D
	.dc.b	$BE,$A3,$F0,$58,$7E,$D1,$B1,$9F,$A9,$E3,$18,$BF,$F3,$6F,$CD,$FF
	.dc.b	$B6,$2B,$D4,$7A,$17,$D6,$78,$1D,$4F,$A6,$75,$39,$8D,$7B,$96,$79
	.dc.b	$C7,$68,$F4,$FD,$53,$CF,$BE,$86,$1B,$7C,$2D,$F5,$4F,$8A,$F8,$2D
	.dc.b	$9D,$2F,$53,$33,$F1,$BE,$A9,$E8,$53,$1F,$0D,$70,$ED,$F5,$4F,$7F
	.dc.b	$E6,$B6,$33,$F0,$18,$9F,$A3,$F5,$5F,$6B,$9B,$F9,$0A,$EB,$CF,$66
	.dc.b	$9F,$AF,$F5,$4F,$8E,$BC,$78,$6E,$7E,$FF,$88,$7C,$47,$DE,$5F,$11
	.dc.b	$63,$0F,$E0,$BE,$AB,$FA,$15,$7C,$77,$8B,$7D,$9D,$57,$E2,$31,$7D
	.dc.b	$55,$F8,$6C,$36,$FE,$87,$D5,$33,$FC,$7A,$F0,$2C,$61,$FC,$9F,$D5
	.dc.b	$7F,$0D,$AF,$B0,$DE,$DF,$C3,$1F,$EB,$A7,$F2,$09,$D7,$27,$F4,$AF
	.dc.b	$AA,$7C,$57,$E7,$9B,$7F,$83,$5F,$55,$99,$FD,$1F,$C7,$FD,$EA,$FC
	.dc.b	$43,$71,$F0,$87,$F3,$38,$7F,$58,$74,$FF,$5A,$FA,$AF,$E9,$15,$F9
	.dc.b	$16,$D7,$2F,$AA,$7C,$47,$EC,$A7,$37,$FD,$8B,$EA,$7F,$BB,$E1,$F9
	.dc.b	$96,$E1,$E2,$7F,$C1,$D1,$9F,$82,$FD,$DC,$7C,$D8,$E1,$BE,$38,$32
	.dc.b	$EB,$DC,$FE,$03,$F5,$5F,$EE,$4A,$ED,$18,$C7,$9F,$B9,$B9,$F1,$36
	.dc.b	$3C,$5F,$3F,$FC,$2B,$C3,$FD,$45,$79,$E3,$1E,$4F,$93,$CD,$F2,$C5
	.dc.b	$9D,$1F,$B5,$DD,$F4,$7C,$FC,$DF,$E2,$B1,$89,$FD,$D6,$29,$CF,$D3
	.dc.b	$6A,$ED,$73,$7F,$8E,$E7,$F9,$A3,$EC,$DF,$F7,$39,$FE,$09,$19,$C3
	.dc.b	$FC,$23,$39,$8F,$BE,$7E,$90,$BD,$FD,$3C,$0E,$AF,$6F,$9B,$C6,$1D
	.dc.b	$1B,$5C,$B4,$F8,$DE,$D7,$91,$B5,$E2,$DB,$0D,$3C,$B1,$EA,$A7,$D5
	.dc.b	$7F,$93,$D1,$83,$0E,$5C,$1F,$D2,$07,$D3,$FE,$83,$EA,$BF,$CB,$0D
	.dc.b	$8F,$F3,$13,$5E,$0D,$7F,$A8,$BF,$AC,$DB,$8F,$77,$F5,$19,$E2,$D7
	.dc.b	$8B,$5E,$3F,$31,$FD,$6A,$7F,$C7,$F4,$8C,$65,$AA,$FE,$71,$3D,$D7
	.dc.b	$D5,$7F,$1A,$7E,$0C,$7E,$21,$DF,$17,$27,$2D,$7E,$0D,$D9,$71,$D2
	.dc.b	$4F,$F4,$7F,$AA,$F9,$19,$79,$BE,$6B,$81,$27,$CB,$5C,$FC,$89,$1E
	.dc.b	$87,$4F,$F0,$8F,$2D,$C9,$E3,$6C,$7F,$55,$6B,$FC,$3B,$EA,$9F,$4B
	.dc.b	$BE,$49,$3B,$79,$5F,$B7,$EA,$BE,$54,$FF,$5F,$6B,$FC,$61,$BA,$7F
	.dc.b	$9B,$B0,$7D,$37,$1E,$19,$7C,$5E,$9C,$3F,$27,$EA,$9F,$3B,$7B,$E5
	.dc.b	$FF,$6D,$9C,$B6,$32,$FF,$28,$8F,$F6,$6A,$76,$E7,$E1,$3B,$11,$FE
	.dc.b	$F1,$F5,$5F,$DD,$DE,$DE,$0F,$8C,$8F,$86,$7C,$DE,$94,$8F,$23,$A7
	.dc.b	$F9,$CF,$B9,$FC,$E5,$FF,$02,$3F,$D5,$5B,$E4,$7E,$1E,$35,$1F,$D0
	.dc.b	$3D,$EF,$EF,$8F,$F8,$2D,$78,$BF,$8D,$6C,$67,$DB,$FC,$87,$E1,$1F
	.dc.b	$D7,$E3,$FC,$24,$F7,$7F,$A5,$B7,$93,$FD,$42,$EE,$3C,$FE,$4F,$8F
	.dc.b	$E1,$24,$7A,$23,$C1,$F6,$7E,$1F,$B3,$93,$E5,$8F,$D0,$7A,$46,$FA
	.dc.b	$B2,$7F,$C4,$93,$97,$E1,$FC,$8F,$C1,$F3,$E4,$60,$F8,$1E,$4F,$BE
	.dc.b	$E4,$60,$FF,$17,$7E,$0F,$FE,$10,$6F,$A1,$FB,$18,$FF,$65,$6F,$FE
	.dc.b	$BE,$75,$3F,$53,$E4,$7E,$87,$C9,$48,$FA,$B1,$F3,$7B,$FE,$77,$7C
	.dc.b	$9F,$A5,$1F,$BD,$F2,$3F,$17,$8F,$FB,$29,$FC,$79,$3C,$0F,$B9,$E3
	.dc.b	$FD,$DF,$AF,$98,$7E,$0A,$3E,$D7,$95,$F6,$EE,$4F,$DF,$8F,$E1,$5F
	.dc.b	$82,$FF,$AF,$F1,$FF,$6B,$AD,$32,$6B,$5F,$C5,$6F,$F8,$BF,$52,$63
	.dc.b	$F9,$08,$FC,$47,$BF,$F8,$9F,$36,$4F,$E3,$63,$F8,$C7,$E0,$89,$B1
	.dc.b	$27,$FC,$B9,$39,$7F,$CB,$DF,$59,$FE,$94,$6F,$8B,$FC,$86,$31,$1F
	.dc.b	$E1,$A7,$FD,$04,$E8,$E3,$E3,$6C,$E1,$36,$3F,$A5,$1F,$D4,$27,$FB
	.dc.b	$93,$80,$E1,$DB,$5F,$EA,$4F,$FA,$E1,$E0,$D7,$7F,$2B,$FD,$76,$31
	.dc.b	$1F,$D6,$7E,$BD,$DF,$FC,$ED,$B1,$FE,$7D,$6B,$ED,$AF,$81,$F3,$32
	.dc.b	$3E,$2B,$F0,$FE,$5B,$11,$E6,$8F,$01,$1F,$2F,$D6,$F8,$C8,$E0,$87
	.dc.b	$CB,$FA,$5F,$9C,$7C,$F8,$B1,$EE,$8F,$61,$1F,$91,$F5,$BF,$E8,$C6
	.dc.b	$F9,$7F,$DF,$63,$11,$F4,$1A,$FF,$7B,$FB,$67,$91,$FD,$75,$B1,$FE
	.dc.b	$9F,$6B,$FE,$58,$9E,$7D,$2F,$6F,$FD,$97,$1F,$C7,$A3,$13,$FE,$0C
	.dc.b	$8C,$4E,$BE,$1B,$FB,$54,$7F,$17,$8C,$4F,$AA,$7D,$04,$7D,$BE,$3F
	.dc.b	$AC,$8E,$08,$79,$7A,$BA,$9D,$A2,$C7,$FA,$E6,$FF,$61,$7E,$8A,$D4
	.dc.b	$E7,$AB,$E4,$11,$F9,$AD,$7F,$CF,$3F,$60,$FA,$DF,$F0,$23,$63,$91
	.dc.b	$DC,$F7,$90,$F1,$7C,$0D,$7E,$A1,$F5,$C8,$7E,$8F,$E0,$9F,$5D,$FE
	.dc.b	$11,$7F,$DD,$D1,$FE,$D3,$6B,$FE,$8C,$76,$E7,$E6,$BF,$C7,$D1,$A8
	.dc.b	$FF,$1D,$35,$FF,$20,$87,$0F,$F9,$11,$EB,$BF,$DC,$87,$E1,$8F,$AD
	.dc.b	$FF,$12,$36,$3C,$07,$97,$D4,$FF,$14,$7C,$53,$D1,$A7,$AE,$24,$FF
	.dc.b	$BB,$67,$4D,$9C,$BF,$EA,$8F,$B3,$7C,$AF,$F8,$CD,$B1,$FE,$F5,$6B
	.dc.b	$F0,$FD,$0D,$4C,$8D,$3F,$2E,$3E,$67,$8D,$27,$E6,$3F,$3F,$C7,$C4
	.dc.b	$7C,$D8,$BA,$FC,$3D,$BF,$94,$4E,$4E,$24,$3E,$A5,$70,$E6,$F9,$F3
	.dc.b	$FE,$E0,$9D,$57,$3A,$D4,$7D,$1B,$AF,$D5,$8B,$5C,$63,$C0,$76,$39
	.dc.b	$DF,$6A,$DD,$70,$6C,$E8,$EC,$53,$B1,$6B,$B5,$FB,$E7,$E2,$4E,$5E
	.dc.b	$6F,$8A,$34,$F1,$AF,$03,$4E,$05,$38,$5F,$9D,$C4,$F2,$72,$79,$8F
	.dc.b	$71,$83,$F3,$15,$E2,$8E,$04,$9E,$93,$C2,$30,$F3,$6B,$F0,$0E,$97
	.dc.b	$F0,$27,$8C,$FD,$59,$F4,$9A,$BF,$6B,$4C,$C7,$4A,$FC,$2B,$6F,$CD
	.dc.b	$7E,$4A,$FD,$9C,$7E,$D8,$8F,$F0,$16,$BF,$DA,$2B,$F0,$0F,$03,$47
	.dc.b	$E8,$7C,$CF,$D2,$1E,$4B,$FD,$C6,$7F,$E9,$69,$D1,$E6,$C3,$F8,$0C
	.dc.b	$66,$3F,$5D,$5C,$67,$2F,$48,$E0,$7F,$B2,$27,$85,$76,$9E,$13,$97
	.dc.b	$83,$F3,$30,$B3,$B8,$E6,$2B,$9D,$F8,$23,$31,$87,$A1,$8B,$E7,$3F
	.dc.b	$58,$91,$A3,$D6,$6C,$F8,$27,$8C,$B5,$E3,$70,$70,$23,$F6,$5D,$C7
	.dc.b	$95,$DC,$71,$24,$78,$A3,$F2,$A7,$03,$C9,$1C,$DE,$B1,$CE,$30,$72
	.dc.b	$8D,$C7,$3C,$3F,$29,$18,$21,$D8,$FD,$25,$FC,$77,$E6,$5F,$D4,$69
	.dc.b	$E3,$38,$5F,$E8,$A7,$81,$C6,$BB,$CE,$AF,$CC,$47,$F5,$4F,$F4,$AF
	.dc.b	$AF,$FF,$5A,$FD,$83,$ED,$FF,$DD,$0D,$A0,$C7,$FA,$65,$7F,$AF,$0E
	.dc.b	$35,$C0,$FF,$78,$3F,$EC,$0F,$AF,$7B,$5F,$EF,$36,$DF,$4B,$FE,$AD
	.dc.b	$18,$8E,$D8,$F9,$7D,$FE,$17,$27,$84,$FF,$62,$AE,$17,$E4,$BF,$F8
	.dc.b	$01,$F6,$FF,$DB,$2F,$99,$EA,$71,$3F,$D9,$5B,$7F,$0F,$B3,$8F,$57
	.dc.b	$F9,$76,$D5,$CD,$FE,$E2,$62,$35,$19,$8F,$DA,$FA,$7C,$78,$0D,$24
	.dc.b	$7F,$9A,$7E,$C9,$F4,$5D,$DB,$19,$8C,$1E,$3E,$17,$97,$49,$3F,$E7
	.dc.b	$38,$CB,$7E,$AB,$81,$1C,$1F,$E7,$9F,$C3,$9F,$59,$EF,$6C,$66,$30
	.dc.b	$7C,$3E,$2F,$F5,$CC,$7C,$26,$BF,$4F,$C4,$D1,$F2,$7E,$E3,$E5,$41
	.dc.b	$A4,$7E,$77,$D9,$BE,$42,$92,$3F,$D1,$D1,$97,$FD,$21,$F6,$7F,$E1
	.dc.b	$D7,$E9,$B5,$FF,$12,$B5,$FC,$CB,$57,$D9,$FC,$E7,$FD,$3A,$D7,$EE
	.dc.b	$B5,$FF,$17,$B5,$FD,$1F,$D9,$BE,$DF,$F0,$CD,$8F,$F5,$23,$5F,$F5
	.dc.b	$4B,$5F,$C4,$BF,$39,$CE,$9C,$88,$FF,$AB,$7E,$CD,$BE,$E3,$FC,$83
	.dc.b	$18,$9D,$77,$BC,$BE,$17,$BB,$FD,$CF,$DC,$3E,$13,$E2,$37,$B7,$FD
	.dc.b	$74,$D9,$FF,$5F,$B5,$FC,$FA,$F9,$1E,$24,$9E,$2F,$CF,$FB,$4F,$F2
	.dc.b	$44,$7F,$B1,$5B,$5F,$EC,$A6,$BF,$AA,$6B,$FE,$5C,$F7,$DE,$3F,$D9
	.dc.b	$6D,$7F,$B3,$DB,$3F,$ED,$46,$BC,$FD,$2F,$11,$27,$C4,$F3,$DA,$47
	.dc.b	$36,$CF,$FB,$7B,$ED,$1E,$EF,$E2,$5B,$5F,$ED,$F6,$CE,$E7,$F1,$DF
	.dc.b	$67,$F1,$6B,$FD,$C8,$D9,$FF,$74,$B5,$FE,$12,$D7,$FC,$F6,$BE,$CB
	.dc.b	$99,$27,$9B,$F9,$BF,$B3,$7A,$DF,$EB,$8F,$27,$FC,$FD,$18,$AF,$E4
	.dc.b	$B8,$7F,$38,$8F,$F4,$9C,$62,$11,$B7,$F9,$29,$D3,$E4,$BA,$4F,$FA
	.dc.b	$3A,$D1,$89,$FE,$87,$1E,$DB,$FF,$47,$3A,$AE,$DC,$BF,$AE,$6F,$57
	.dc.b	$1B,$FC,$B7,$D1,$5A,$D4,$78,$BF,$02,$FF,$6B,$5F,$9D,$B1,$9F,$81
	.dc.b	$E5,$18,$CB,$FD,$3D,$83,$E7,$7B,$37,$0F,$80,$BE,$77,$33,$E3,$FC
	.dc.b	$01,$F1,$14,$C7,$CB,$5C,$3E,$47,$D9,$FE,$97,$2F,$F5,$63,$63,$FD
	.dc.b	$67,$F1,$9F,$CA,$9D,$F7,$FB,$FE,$CF,$F9,$EC,$7D,$16,$F0,$7D,$0B
	.dc.b	$F5,$4F,$FA,$F2,$73,$87,$FA,$E7,$EF,$EC,$FF,$B0,$1B,$7F,$DD,$17
	.dc.b	$DA,$7C,$D7,$30,$FB,$3F,$BD,$7F,$EC,$36,$DF,$FB,$29,$FC,$2C,$FE
	.dc.b	$09,$CB,$FE,$00,$FB,$26,$9F,$65,$B5,$F8,$B5,$E0,$66,$7E,$F4,$BE
	.dc.b	$CE,$98,$F8,$2B,$1F,$93,$FB,$27,$CC,$7E,$15,$B1,$9F,$99,$C4,$FE
	.dc.b	$27,$DF,$FE,$D7,$3F,$9B,$6E,$9F,$9A,$5F,$07,$BF,$73,$F2,$1F,$7E
	.dc.b	$F8,$EA,$63,$CA,$B8,$FE,$97,$EC,$9F,$25,$F9,$76,$C6,$7E,$47,$13
	.dc.b	$F9,$BF,$B2,$76,$D3,$1D,$D6,$FF,$61,$F7,$DE,$6E,$AF,$F7,$A3,$63
	.dc.b	$FD,$F6,$BE,$2F,$F4,$CE,$78,$BF,$C6,$3F,$64,$DB,$F5,$4D,$BF,$DE
	.dc.b	$7D,$9B,$E9,$33,$5F,$AF,$FB,$37,$C9,$53,$1E,$6D,$C3,$F8,$24,$BC
	.dc.b	$AE,$07,$1F,$A8,$B3,$AF,$8F,$FD,$D7,$D9,$FF,$91,$23,$07,$F0,$D7
	.dc.b	$1B,$7F,$0C,$C3,$FA,$0F,$0A,$BF,$C9,$72,$9F,$F2,$2C,$78,$BE,$53
	.dc.b	$D1,$87,$A7,$CD,$FE,$11,$F5,$3F,$C3,$EB,$8D,$FE,$F7,$0F,$E3,$33
	.dc.b	$97,$FC,$AF,$1F,$D8,$E7,$DB,$5F,$0F,$D2,$E9,$1F,$C8,$F2,$7D,$2F
	.dc.b	$6F,$1F,$F8,$B4,$62,$7F,$94,$65,$FA,$0C,$BF,$93,$CF,$E3,$A7,$95
	.dc.b	$F0,$F0,$78,$BF,$C7,$E9,$F1,$3E,$AF,$81,$C3,$6F,$F9,$A4,$BE,$8F
	.dc.b	$31,$86,$C7,$F3,$63,$F8,$D4,$64,$E3,$EA,$77,$9F,$06,$DF,$23,$E0
	.dc.b	$ED,$7A,$9F,$26,$F8,$C3,$A3,$75,$FF,$15,$7D,$9F,$F8,$E0,$93,$FE
	.dc.b	$63,$7F,$CC,$2F,$FA,$13,$EC,$5F,$52,$FB,$3C,$CA,$73,$18,$6C,$72
	.dc.b	$6B,$FE,$6C,$5F,$57,$FE,$8F,$8D,$35,$D3,$5F,$C5,$C7,$0C,$3B,$7C
	.dc.b	$E7,$E3,$9F,$C8,$4F,$9E,$31,$A7,$06,$C7,$AE,$FC,$1F,$67,$FC,$52
	.dc.b	$3D,$DC,$7A,$7D,$8D,$C7,$F1,$31,$FD,$B5,$5E,$79,$24,$E5,$E5,$F6
	.dc.b	$7F,$E8,$AC,$FC,$B3,$F1,$FC,$EE,$84,$78,$1D,$7C,$6F,$01,$1F,$04
	.dc.b	$79,$7D,$0F,$EB,$D1,$FD,$FD,$AF,$F7,$E6,$D7,$83,$EC,$BF,$CD,$93
	.dc.b	$9E,$5E,$44,$79,$73,$FC,$1F,$B3,$F6,$C7,$F8,$2D,$AF,$F8,$29,$B7
	.dc.b	$D7,$EC,$F8,$70,$4A,$E0,$F3,$DF,$E6,$7D,$9F,$F9,$56,$7F,$C2,$CF
	.dc.b	$F9,$3D,$BF,$1D,$AF,$81,$DB,$23,$27,$08,$E7,$ED,$C8,$E6,$FD,$9F
	.dc.b	$B3,$FF,$0D,$C6,$0E,$FE,$17,$88,$E0,$FF,$89,$30,$7B,$1E,$62,$47
	.dc.b	$03,$FC,$59,$F5,$4F,$5B,$FD,$90,$D8,$FD,$1B,$5F,$D4,$B5,$FF,$65
	.dc.b	$FC,$43,$E9,$7F,$DA,$0D,$8F,$D5,$B5,$FD,$8B,$5F,$F6,$9F,$D9,$FF
	.dc.b	$39,$7F,$62,$D7,$F6,$ED,$77,$E7,$79,$BC,$B9,$1E,$66,$BD,$BD,$7F
	.dc.b	$19,$1F,$1B,$FE,$E5,$F4,$FF,$B0,$C7,$F0,$06,$BF,$C1,$9A,$FF,$BA
	.dc.b	$1A,$FF,$42,$CB,$FB,$A3,$F1,$BD,$BF,$92,$47,$E0,$F8,$7F,$10,$8E
	.dc.b	$0F,$E1,$FF,$67,$FE,$D2,$41,$C3,$7C,$CF,$F7,$E4,$62,$DF,$65,$F6
	.dc.b	$09,$5F,$61,$FA,$7D,$32,$3E,$9B,$5F,$E4,$7E,$27,$E6,$5B,$D1,$FC
	.dc.b	$89,$FF,$09,$B6,$35,$1F,$D8,$85,$FF,$0F,$7F,$0B,$E0,$FE,$28,$8F
	.dc.b	$E0,$BC,$2F,$C2,$91,$C1,$FC,$C3,$EC,$FF,$C0,$CD,$F5,$FC,$51,$88
	.dc.b	$FE,$72,$D7,$C5,$7F,$D4,$5F,$CB,$F9,$FF,$9A,$23,$F9,$2F,$94,$FC
	.dc.b	$A9,$1C,$1F,$D0,$3E,$CF,$FC,$34,$DF,$23,$E1,$46,$23,$FA,$4B,$5F
	.dc.b	$85,$F6,$7F,$EA,$17,$FA,$4B,$5F,$E9,$ED,$77,$F2,$5F,$36,$3F,$CA
	.dc.b	$31,$C3,$C2,$C1,$1C,$1F,$EA,$18,$FF,$2D,$B7,$C1,$FE,$79,$1C,$A3
	.dc.b	$FD,$46,$BE,$F7,$3F,$2B,$27,$94,$57,$CB,$1E,$32,$3E,$AF,$A4,$F3
	.dc.b	$11,$C1,$EE,$FB,$3F,$09,$1F,$34,$79,$3E,$CE,$47,$A5,$F3,$FB,$18
	.dc.b	$8F,$8B,$1F,$0C,$8F,$CA,$7D,$BF,$2B,$23,$91,$FE,$C3,$FB,$2F,$F4
	.dc.b	$4B,$78,$FF,$AF,$47,$E5,$23,$5F,$25,$89,$FC,$E4,$E6,$7F,$C0,$6D
	.dc.b	$7F,$38,$D7,$FC,$7E,$BE,$07,$89,$23,$FC,$0F,$19,$6F,$93,$E2,$23
	.dc.b	$E7,$9F,$2F,$85,$23,$D4,$FA,$3E,$03,$11,$E4,$8F,$39,$1E,$EF,$BB
	.dc.b	$AA,$47,$23,$E2,$76,$7C,$2F,$0A,$46,$0F,$90,$E1,$FF,$74,$E5,$FE
	.dc.b	$3E,$74,$63,$DD,$F9,$6C,$3E,$83,$5E,$25,$F4,$BF,$C0,$72,$9F,$F0
	.dc.b	$87,$0B,$EB,$29,$2F,$FD,$F9,$0F,$BD,$C4,$7F,$8D,$A3,$83,$FE,$34
	.dc.b	$6B,$99,$FE,$2D,$18,$C3,$FC,$26,$BE,$67,$C2,$74,$8C,$FC,$17,$93
	.dc.b	$D0,$98,$78,$A2,$A9,$E7,$38,$78,$C7,$A3,$D5,$F6,$91,$F5,$F9,$75
	.dc.b	$E5,$EF,$7F,$A2,$32,$FF,$06,$B6,$BE,$2D,$7C,$6F,$0E,$5F,$C4,$7E
	.dc.b	$4F,$8F,$88,$E6,$7F,$46,$C3,$F8,$A6,$3F,$02,$B7,$3C,$48,$70,$13
	.dc.b	$C3,$1F,$25,$F5,$BE,$55,$C1,$F9,$71,$99,$F5,$5D,$AF,$14,$D7,$AC
	.dc.b	$5B,$4D,$8C,$36,$7B,$57,$89,$A7,$B6,$76,$E0,$EA,$5A,$EA,$D8,$DB
	.dc.b	$EF,$8F,$73,$CC,$E1,$5E,$17,$53,$F2,$1A,$73,$3F,$2B,$1B,$BF,$8B
	.dc.b	$1D,$63,$17,$F1,$23,$67,$13,$99,$B3,$13,$C8,$F9,$AE,$23,$93,$B3
	.dc.b	$8C,$6E,$32,$F0,$8C,$C6,$08,$FD,$43,$A1,$F0,$4E,$91,$CD,$C3,$5F
	.dc.b	$C2,$CF,$B1,$CC,$FD,$57,$D6,$7F,$80,$6F,$FB,$13,$E8,$31,$7F,$B4
	.dc.b	$23,$FB,$96,$BF,$BD,$7F,$82,$F8,$B2,$3C,$E7,$E5,$9C,$3E,$E3,$C6
	.dc.b	$7E,$85,$CE,$1E,$9A,$FB,$F3,$C4,$FE,$1B,$3E,$B7,$07,$E1,$5C,$41
	.dc.b	$E7,$7F,$2A,$7F,$1E,$6C,$E6,$79,$A7,$83,$83,$CC,$70,$AF,$E2,$75
	.dc.b	$D2,$3D,$26,$6B,$F8,$24,$66,$30,$6A,$BA,$BE,$13,$AB,$EE,$33,$5D
	.dc.b	$A3,$C9,$3C,$E7,$57,$63,$DE,$65,$FD,$B3,$67,$F5,$8D,$7B,$B6,$BE
	.dc.b	$73,$6F,$E3,$5F,$89,$F5,$D7,$88,$F8,$95,$F5,$6F,$C5,$19,$3D,$91
	.dc.b	$97,$73,$C9,$F9,$7F,$5F,$FE,$55,$9F,$F6,$7B,$FE,$AD,$8F,$E3,$8D
	.dc.b	$8C,$F9,$5F,$ED,$F8,$E1,$5F,$E6,$CA,$46,$E3,$0D,$8F,$F3,$8B,$C1
	.dc.b	$BE,$27,$3F,$52,$47,$F9,$DB,$0F,$1C,$E2,$3C,$73,$67,$31,$BB,$EF
	.dc.b	$F6,$5E,$52,$46,$4F,$0F,$89,$87,$FC,$DC,$7A,$23,$FC,$E9,$89,$A6
	.dc.b	$3E,$C2,$4F,$03,$DE,$E1,$E7,$2F,$17,$99,$23,$84,$65,$BE,$8F,$80
	.dc.b	$8E,$0D,$F8,$9E,$42,$38,$39,$E5,$E8,$23,$83,$CD,$F7,$BF,$E2,$F7
	.dc.b	$FD,$42,$D4,$D2,$7C,$57,$FA,$55,$34,$91,$F0,$7E,$AB,$FD,$34,$9A
	.dc.b	$47,$FD,$3D,$F5,$3F,$E0,$C7,$FD,$38,$D7,$F1,$0D,$7F,$D6,$AD,$5F
	.dc.b	$52,$C3,$FD,$50,$D8,$FA,$4D,$7E,$DB,$5F,$F5,$57,$C6,$FA,$9F,$B0
	.dc.b	$D7,$F0,$2D,$77,$87,$FA,$D7,$EA,$1F,$53,$49,$3F,$81,$8C,$BF,$83
	.dc.b	$FB,$DF,$F5,$43,$FA,$36,$BF,$ED,$06,$A6,$FD,$53,$E3,$DF,$51,$F8
	.dc.b	$D8,$C3,$7B,$3F,$C8,$17,$F9,$89,$CB,$F9,$AF,$AA,$76,$7C,$18,$FF
	.dc.b	$6F,$C6,$37,$FC,$C3,$63,$84,$ED,$7D,$53,$E0,$70,$DB,$FD,$03,$67
	.dc.b	$F5,$2D,$7F,$DB,$0B,$CF,$E2,$49,$E2,$F4,$F3,$E4,$74,$6D,$74,$F3
	.dc.b	$FC,$24,$9F,$0B,$E4,$F7,$E4,$79,$1B,$3F,$BC,$FA,$A7,$81,$ED,$6D
	.dc.b	$7E,$E5,$B3,$B9,$F8,$5F,$54,$F8,$6A,$4B,$FE,$01,$19,$7F,$81,$7D
	.dc.b	$57,$F9,$8D,$BE,$77,$CB,$8C,$61,$FC,$39,$AF,$FB,$F6,$39,$3F,$C4
	.dc.b	$9A,$E6,$FE,$A4,$62,$3F,$97,$7A,$2E,$BF,$B0,$DC,$3F,$8A,$36,$7F
	.dc.b	$8E,$B5,$FE,$6C,$3E,$AB,$F5,$93,$FC,$6D,$B1,$FC,$91,$AF,$06,$BF
	.dc.b	$CF,$FE,$1D,$F1,$FF,$89,$6D,$7F,$25,$6C,$FF,$2E,$6B,$E2,$FA,$A7
	.dc.b	$CE,$FE,$45,$B5,$FC,$B5,$B3,$FC,$D9,$AF,$AB,$EA,$BF,$34,$DE,$8F
	.dc.b	$CD,$46,$23,$FA,$A3,$50,$FA,$A7,$D7,$FE,$85,$B8,$7F,$3D,$6C,$FF
	.dc.b	$46,$6B,$F2,$BE,$AE,$BF,$D8,$5A,$F6,$6A,$6F,$D5,$FE,$0A,$3F,$B4
	.dc.b	$35,$F2,$B5,$37,$EA,$DE,$9F,$ED,$DB,$87,$F5,$36,$D7,$F5,$C6,$BF
	.dc.b	$80,$87,$D5,$FE,$FD,$A7,$F5,$B2,$C7,$F6,$46,$BF,$88,$FA,$A7,$C0
	.dc.b	$FF,$07,$6C,$FF,$63,$6C,$FF,$6C,$6B,$F9,$0F,$5B,$FD,$2F,$3F,$DA
	.dc.b	$5B,$1F,$DC,$9A,$EF,$D9,$FE,$29,$F5,$2F,$06,$92,$FF,$B9,$46,$5F
	.dc.b	$EE,$9F,$52,$FA,$9F,$E3,$8D,$8C,$C6,$0F,$E4,$1F,$52,$F9,$6A,$49
	.dc.b	$FE,$FD,$19,$7F,$C0,$1F,$53,$E6,$CF,$FC,$40,$D7,$F1,$AD,$4D,$FA
	.dc.b	$97,$8F,$FC,$BD,$B5,$FE,$0B,$6D,$7F,$85,$1A,$FE,$E3,$E9,$FF,$B9
	.dc.b	$CF,$F8,$45,$B1,$FE,$1A,$6B,$BF,$53,$F9,$E4,$3B,$A9,$2B,$93,$AF
	.dc.b	$33,$FB,$67,$BF,$FD,$AF,$EA,$3F,$4F,$5F,$E2,$9C,$58,$7F,$41,$AD
	.dc.b	$47,$09,$FF,$21,$4E,$5B,$79,$9F,$D8,$CF,$F4,$59,$E8,$EE,$33,$E1
	.dc.b	$6D,$7C,$33,$E7,$B1,$5F,$C3,$31,$FF,$00,$4E,$70,$E7,$E3,$6B,$1D
	.dc.b	$B6,$3D,$54,$F9,$AE,$D7,$FA,$88,$FF,$05,$19,$9F,$81,$19,$7F,$CB
	.dc.b	$D0,$F9,$AF,$33,$BC,$CF,$28,$CD,$FA,$E5,$F1,$BE,$0F,$63,$0D,$C3
	.dc.b	$E7,$4B,$E6,$73,$18,$6C,$76,$8F,$E7,$7F,$43,$D6,$7B,$36,$39,$EB
	.dc.b	$EE,$F6,$7C,$4D,$8E,$FB,$61,$A7,$03,$C7,$83,$EA,$3F,$C7,$E4,$7F
	.dc.b	$CC,$46,$4F,$F3,$5F,$D2,$BE,$A1,$F5,$1F,$E6,$78,$C3,$63,$33,$C1
	.dc.b	$AF,$F5,$69,$7C,$77,$CC,$8E,$AD,$74,$D7,$FD,$25,$1C,$38,$BC,$1F
	.dc.b	$1F,$F6,$32,$CB,$6B,$FA,$1D,$23,$10,$8E,$4B,$EA,$3F,$90,$BC,$38
	.dc.b	$F7,$3F,$8B,$8F,$D3,$35,$FA,$C7,$6A,$DF,$24,$8D,$BE,$0F,$A8,$FE
	.dc.b	$EB,$97,$F9,$6E,$DE,$37,$62,$5F,$03,$BF,$C8,$78,$88,$F8,$A3,$D7
	.dc.b	$E5,$39,$69,$27,$F7,$9F,$50,$FA,$CD,$F3,$49,$FE,$01,$7F,$DF,$9B
	.dc.b	$7F,$1B,$EA,$7E,$44,$7F,$05,$6B,$F0,$03,$8D,$3F,$CE,$D8,$BE,$A7
	.dc.b	$97,$0C,$BE,$4F,$5E,$1F,$A7,$F5,$1F,$EC,$79,$3F,$EC,$B6,$F8,$1A
	.dc.b	$F6,$36,$C8,$FE,$31,$3B,$39,$7D,$4F,$F8,$F6,$33,$F4,$9D,$8F,$C0
	.dc.b	$C6,$63,$D5,$29,$F1,$FA,$B2,$78,$9D,$9F,$3F,$A5,$23,$07,$F2,$9F
	.dc.b	$9F,$7C,$FF,$E3,$9B,$19,$8F,$C8,$FE,$DD,$FC,$4B,$B7,$FD,$D9,$F0
	.dc.b	$6F,$9C,$FC,$AB,$6F,$F9,$A3,$5F,$A0,$FD,$B7,$F9,$C7,$D5,$7F,$41
	.dc.b	$6F,$9D,$F9,$E8,$C4,$7D,$77,$FC,$98,$F9,$FC,$DF,$31,$1F,$57,$BF
	.dc.b	$E2,$F5,$64,$78,$9F,$2F,$3F,$EA,$BE,$63,$FB,$64,$FF,$49,$6B,$F8
	.dc.b	$56,$F5,$7F,$AE,$5F,$F2,$1C,$7C,$7F,$6B,$E3,$11,$F9,$5F,$45,$F0
	.dc.b	$BC,$89,$1F,$09,$F8,$BF,$41,$FB,$AF,$AA,$FE,$5F,$3F,$D6,$5A,$FE
	.dc.b	$55,$B8,$FF,$B2,$5F,$DE,$C7,$E6,$79,$5F,$1D,$F5,$C9,$1F,$6E,$3E
	.dc.b	$9F,$7C,$8F,$A6,$FD,$6F,$A1,$FE,$17,$E5,$FE,$64,$9E,$07,$F0,$EF
	.dc.b	$C2,$BF,$D7,$F7,$F8,$5F,$1B,$F0,$84,$7F,$11,$EC,$FE,$03,$CA,$91
	.dc.b	$F8,$07,$F0,$7F,$19,$FC,$63,$EA,$FF,$C1,$B3,$FD,$DD,$AF,$EC,$DB
	.dc.b	$C1,$FE,$E7,$7F,$C5,$91,$F9,$5F,$5B,$F2,$84,$7F,$31,$F1,$BF,$90
	.dc.b	$FA,$39,$1F,$90,$7F,$27,$F0,$DF,$CA,$3E,$AD,$E9,$52,$57,$F8,$26
	.dc.b	$3F,$8D,$BF,$A3,$F0,$7F,$97,$FD,$53,$EA,$E9,$FA,$58,$CF,$97,$2B
	.dc.b	$FC,$25,$5F,$B3,$6D,$65,$FF,$4E,$88,$D3,$86,$BF,$EA,$26,$BF,$E8
	.dc.b	$43,$87,$D5,$FF,$5C,$83,$49,$1D,$8D,$9E,$4F,$95,$F5,$9E,$C8,$FE
	.dc.b	$17,$E8,$7B,$BE,$AE,$57,$F8,$AF,$D1,$FE,$35,$82,$7C,$04,$9E,$D1
	.dc.b	$96,$FB,$FE,$52,$38,$37,$F5,$7F,$30,$8F,$C8,$8A,$FC,$38,$F7,$11
	.dc.b	$F6,$7D,$53,$C5,$FE,$FA,$D8,$CC,$60,$8F,$F9,$16,$1E,$0F,$F6,$A6
	.dc.b	$C6,$67,$FB,$6F,$DC,$7F,$5E,$F4,$3E,$85,$3E,$B3,$81,$AB,$A7,$91
	.dc.b	$F1,$91,$F3,$FC,$E7,$98,$8E,$0F,$5F,$95,$E3,$23,$E7,$8A,$FF,$57
	.dc.b	$AF,$D1,$B6,$7D,$53,$EB,$F0,$7E,$19,$27,$E3,$B5,$E5,$F0,$BF,$04
	.dc.b	$E0,$FC,$5F,$85,$F8,$84,$7F,$CD,$9F,$02,$F1,$31,$1F,$9A,$C9,$5F
	.dc.b	$48,$91,$F9,$1E,$17,$99,$FE,$18,$5E,$4D,$31,$16,$7F,$CF,$53,$A6
	.dc.b	$CE,$9A,$43,$FB,$D4,$6D,$B3,$FE,$26,$C3,$FD,$5D,$1D,$3E,$1F,$C4
	.dc.b	$67,$E3,$A4,$64,$87,$F8,$16,$78,$BD,$BC,$FF,$51,$1F,$F4,$87,$DA
	.dc.b	$7F,$B6,$1F,$4C,$65,$E0,$6B,$C3,$F8,$5E,$54,$8F,$7B,$FE,$3C,$AF
	.dc.b	$F1,$A7,$92,$E6,$FE,$01,$AF,$F1,$4B,$F7,$DF,$17,$FC,$35,$7C,$2B
	.dc.b	$73,$9B,$E5,$41,$7C,$E3,$30,$BE,$4E,$E7,$57,$D2,$3C,$E7,$C0,$F3
	.dc.b	$BB,$46,$63,$CA,$64,$D9,$82,$5F,$82,$FB,$C7,$93,$0E,$6F,$85,$F8
	.dc.b	$97,$88,$D1,$D2,$B9,$46,$70,$F3,$CF,$13,$C4,$78,$30,$F3,$57,$C0
	.dc.b	$9D,$CF,$C1,$AE,$57,$DF,$0F,$2D,$68,$91,$E4,$AF,$16,$1D,$08,$FC
	.dc.b	$FA,$E3,$3D,$23,$C7,$87,$86,$31,$1F,$6B,$0B,$39,$8D,$DF,$47,$E0
	.dc.b	$9E,$B3,$B4,$F0,$8D,$5F,$82,$7C,$A7,$C3,$3B,$18,$39,$D7,$77,$27
	.dc.b	$53,$33,$D2,$BB,$47,$39,$FA,$2F,$3C,$3E,$2B,$E8,$35,$5E,$98,$F9
	.dc.b	$F7,$E5,$9F,$EC,$71,$F9,$D7,$FA,$F1,$EE,$2B,$97,$F1,$C6,$30,$FF
	.dc.b	$5E,$C6,$63,$DC,$64,$FE,$17,$E3,$7F,$11,$3D,$2F,$E8,$9C,$3E,$99
	.dc.b	$AF,$F0,$C8,$DC,$FF,$82,$23,$99,$B2,$9B,$30,$52,$61,$F3,$49,$87
	.dc.b	$EF,$2B,$BC,$62,$7E,$39,$FB,$18,$CC,$FC,$9B,$B3,$93,$BD,$7E,$6E
	.dc.b	$B3,$1F,$60,$E6,$7B,$CF,$3C,$77,$74,$64,$F3,$D7,$77,$6F,$9D,$EE
	.dc.b	$D7,$67,$6B,$FD,$35,$78,$9F,$C1,$C6,$8E,$D5,$E1,$8D,$C7,$63,$95
	.dc.b	$F1,$AF,$A2,$73,$7F,$7C,$D8,$FC,$83,$5F,$CC,$4F,$D7,$7C,$0F,$8A
	.dc.b	$BE,$34,$77,$3E,$0C,$78,$9D,$D7,$FB,$BB,$EC,$DF,$5B,$FE,$F0,$FA
	.dc.b	$CF,$F5,$F1,$FE,$BF,$8C,$D7,$F2,$DC,$3F,$95,$C6,$23,$FD,$9D,$3C
	.dc.b	$3D,$0E,$97,$CF,$DF,$95,$C7,$4F,$F1,$F5,$A7,$FD,$FF,$F6,$5F,$D4
	.dc.b	$B2,$FE,$6F,$39,$BE,$07,$F3,$D7,$0F,$82,$B8,$F1,$FA,$7C,$89,$3C
	.dc.b	$05,$23,$FC,$9D,$7F,$D1,$3C,$9F,$F5,$B5,$7A,$A7,$FC,$98,$F0,$6D
	.dc.b	$71,$8F,$F2,$83,$5F,$F2,$FA,$FA,$FF,$94,$48,$F1,$46,$EF,$FA,$9F
	.dc.b	$E0,$FF,$D5,$27,$F8,$1B,$5F,$F3,$13,$5F,$EB,$8D,$59,$FF,$A9,$1F
	.dc.b	$F3,$13,$5F,$F3,$8B,$5F,$E1,$ED,$7F,$CD,$CB,$EC,$BD,$ED,$8E,$71
	.dc.b	$83,$E1,$FD,$AF,$FB,$52,$3F,$CF,$AD,$7F,$8D,$B5,$FF,$3D,$35,$FE
	.dc.b	$E9,$3F,$EB,$C7,$C7,$E0,$F8,$08,$FF,$9E,$FE,$BF,$FC,$5E,$DE,$FF
	.dc.b	$A7,$18,$8F,$F4,$6B,$5F,$E5,$7F,$61,$FE,$39,$7F,$D0,$8D,$7F,$D2
	.dc.b	$6D,$77,$DB,$F7,$BE,$BD,$F6,$34,$93,$FE,$8D,$8C,$BF,$E8,$FF,$B1
	.dc.b	$FF,$19,$3F,$CE,$9A,$FF,$A6,$5A,$FF,$84,$DA,$BE,$C1,$EB,$FE,$39
	.dc.b	$B1,$FE,$9B,$6B,$FE,$AA,$6B,$FD,$23,$D8,$F5,$3F,$EA,$76,$BF,$E2
	.dc.b	$16,$A6,$FD,$87,$F9,$B2,$3F,$A8,$B5,$FF,$58,$35,$FF,$15,$35,$3F
	.dc.b	$ED,$A7,$8F,$CE,$7E,$9A,$31,$1C,$A3,$F5,$3F,$5E,$F9,$AF,$F7,$6C
	.dc.b	$FF,$60,$8C,$7D,$87,$FA,$E5,$FF,$1B,$F0,$DA,$DD,$7F,$B0,$FF,$00
	.dc.b	$7D,$9F,$E8,$DB,$E0,$7F,$B0,$5B,$3F,$ED,$06,$BF,$DB,$BE,$C5,$F1
	.dc.b	$9F,$AB,$6D,$7F,$B2,$5B,$3F,$ED,$86,$BF,$DD,$A9,$F1,$3E,$12,$57
	.dc.b	$85,$E9,$EA,$C8,$E8,$DA,$F2,$65,$E9,$24,$FA,$5F,$27,$C4,$48,$F2
	.dc.b	$36,$7F,$DD,$7F,$60,$F8,$0F,$E0,$8D,$AF,$F7,$13,$67,$73,$FC,$27
	.dc.b	$EC,$1F,$11,$FE,$66,$F0,$FF,$CC,$71,$FE,$98,$C3,$F9,$33,$A6,$FA
	.dc.b	$BB,$7F,$DE,$94,$F9,$0C,$63,$FE,$6F,$87,$9B,$88,$EF,$36,$73,$5F
	.dc.b	$CC,$F9,$3F,$28,$DE,$FE,$11,$8B,$67,$FE,$78,$8F,$F5,$A5,$7F,$3B
	.dc.b	$FB,$0F,$F4,$43,$7E,$07,$C1,$18,$BE,$33,$FE,$83,$50,$9F,$37,$C3
	.dc.b	$BE,$6E,$FC,$56,$1B,$87,$F4,$9F,$AF,$7D,$87,$B5,$7C,$AD,$8C,$3F
	.dc.b	$07,$EC,$3D,$FA,$78,$8F,$F4,$A7,$AA,$FB,$3D,$72,$7C,$78,$C6,$5F
	.dc.b	$D5,$FE,$C5,$F2,$FC,$FC,$CF,$99,$18,$9F,$EB,$99,$7F,$18,$7F,$AF
	.dc.b	$7D,$8E,$73,$F0,$0F,$A7,$F8,$9C,$DF,$EA,$29,$D3,$FE,$A4,$8C,$D7
	.dc.b	$C7,$FB,$1F,$C1,$5F,$27,$F0,$CD,$E9,$FB,$D1,$8B,$FF,$56,$FD,$8F
	.dc.b	$F7,$BB,$FE,$E0,$BD,$BF,$C2,$39,$C3,$CA,$E6,$7F,$BB,$0F,$84,$FC
	.dc.b	$53,$E2,$BF,$F5,$CB,$93,$78,$FE,$3F,$EC,$5E,$B5,$F1,$30,$DA,$FF
	.dc.b	$01,$7D,$03,$E0,$BF,$2C,$BE,$1A,$C6,$1F,$CD,$7D,$8B,$CA,$BF,$1D
	.dc.b	$86,$E1,$FE,$0D,$FB,$17,$C2,$7E,$85,$7C,$85,$8C,$3F,$A5,$FB,$17
	.dc.b	$D0,$5F,$07,$0D,$C3,$FC,$31,$F6,$3E,$76,$FC,$FF,$EB,$63,$1D,$1F
	.dc.b	$79,$47,$EC,$7E,$C5,$ED,$FE,$CF,$F2,$1F,$BB,$7F,$EE,$59,$CE,$1F
	.dc.b	$E2,$59,$CC,$BE,$C7,$F5,$EF,$E2,$BE,$D7,$F9,$0D,$7F,$8A,$E7,$3D
	.dc.b	$7F,$EE,$C9,$CD,$7E,$42,$5E,$6F,$F0,$37,$F7,$95,$FE,$F2,$73,$7C
	.dc.b	$FE,$03,$F8,$47,$D9,$FF,$BE,$63,$FC,$70,$DC,$3A,$7B,$BA,$7A,$32
	.dc.b	$7C,$9C,$F6,$7C,$5E,$7F,$F1,$3F,$95,$FE,$49,$9F,$86,$31,$EC,$F9
	.dc.b	$7D,$5F,$34,$5C,$7D,$CF,$F1,$BA,$FC,$34,$7F,$92,$EF,$F3,$50,$8F
	.dc.b	$F2,$5F,$6F,$7F,$3B,$87,$2F,$18,$B3,$FE,$53,$AE,$79,$FE,$D8,$7C
	.dc.b	$CF,$9F,$DC,$FE,$55,$18,$AF,$04,$63,$D5,$CD,$59,$FF,$2F,$56,$CE
	.dc.b	$9F,$5D,$96,$C6,$AB,$AE,$BF,$17,$E5,$2C,$FB,$5F,$0E,$F8,$BF,$33
	.dc.b	$73,$FF,$2B,$79,$DF,$E6,$69,$FE,$1C,$7F,$9A,$1C,$1F,$ED,$E4,$3E
	.dc.b	$C1,$F6,$4F,$B3,$D4,$E6,$30,$F1,$9D,$CE,$9B,$19,$F9,$AF,$E8,$ED
	.dc.b	$C3,$51,$88,$FF,$77,$CA,$31,$1F,$D1,$61,$F5,$FF,$E8,$4D,$30,$D9
	.dc.b	$F4,$CE,$9F,$F4,$63,$CE,$FD,$BF,$67,$FC,$4A,$77,$C7,$C5,$F8,$0D
	.dc.b	$AF,$E2,$11,$CA,$F1,$C9,$23,$AB,$C7,$E7,$FF,$AD,$47,$F3,$CA,$FE
	.dc.b	$34,$F8,$BC,$CE,$87,$17,$5F,$39,$DC,$8E,$CF,$F5,$27,$9E,$E6,$F3
	.dc.b	$36,$33,$1E,$9F,$B1,$FF,$52,$ED,$F9,$F9,$FB,$75,$FD,$A5,$E7,$CD
	.dc.b	$C8,$E7,$7F,$07,$EC,$BE,$44,$FF,$54,$6D,$F3,$84,$63,$87,$FB,$C6
	.dc.b	$2F,$AE,$E5,$C3,$2F,$FD,$79,$7F,$29,$E9,$C3,$F3,$FE,$CB,$FC,$27
	.dc.b	$1F,$EC,$17,$FC,$50,$DE,$FD,$7C,$06,$D9,$19,$38,$47,$D9,$FA,$9F
	.dc.b	$F1,$04,$73,$FA,$BE,$C6,$7E,$26,$C7,$7F,$0B,$C4,$61,$FE,$E7,$8F
	.dc.b	$FB,$63,$83,$E0,$2F,$5F,$D0,$70,$9F,$F6,$C7,$D8,$BE,$B3,$F1,$4B
	.dc.b	$EF,$A9,$88,$FC,$77,$D4,$BC,$9B,$E5,$4C,$3F,$DC,$53,$97,$FD,$C9
	.dc.b	$F1,$BF,$97,$7F,$C0,$EF,$FB,$05,$BF,$41,$F9,$98,$D6,$BF,$5F,$3F
	.dc.b	$F6,$F5,$79,$BE,$8F,$CA,$79,$BE,$93,$C7,$EB,$49,$F0,$BE,$6F,$7B
	.dc.b	$A7,$9B,$27,$C6,$F7,$F3,$BD,$FF,$75,$27,$E1,$BF,$0B,$E9,$BD,$FB
	.dc.b	$49,$F6,$BF,$0B,$E0,$BD,$1E,$24,$9F,$7B,$EB,$F1,$3E,$5F,$CC,$FC
	.dc.b	$CF,$AE,$95,$F4,$23,$E5,$E7,$F3,$3C,$79,$3F,$2A,$3E,$23,$F1,$7D
	.dc.b	$28,$FC,$C3,$EB,$FA,$91,$FB,$6F,$DA,$F9,$8F,$AF,$94,$9F,$AA,$FD
	.dc.b	$AE,$0F,$E1,$1F,$53,$FA,$19,$FE,$53,$CF,$FA,$78,$C4,$7F,$8F,$9B
	.dc.b	$19,$FB,$5F,$C3,$FC,$B7,$E2,$3D,$89,$7F,$8C,$8F,$C3,$E9,$F8,$8E
	.dc.b	$19,$3F,$86,$8F,$0F,$D5,$F8,$BE,$66,$4F,$8B,$D0,$FF,$36,$68,$AF
	.dc.b	$CA,$FC,$A7,$F1,$FF,$CB,$7F,$D7,$FB,$7E,$F5,$B7,$FE,$73,$EF,$FF
	.dc.b	$75,$E3,$FE,$56,$FB,$0F,$DB,$B7,$CD,$FE,$51,$18,$C3,$97,$D0,$E2
	.dc.b	$7F,$A3,$D2,$B8,$BF,$C9,$9A,$FF,$A0,$5A,$FF,$52,$5E,$6F,$26,$C7
	.dc.b	$F5,$26,$BF,$D6,$5A,$F5,$8F,$2F,$99,$23,$E2,$1D,$63,$D3,$E4,$FA
	.dc.b	$BC,$49,$3A,$3D,$11,$87,$B4,$78,$C8,$F8,$FB,$BC,$C4,$70,$79,$FB
	.dc.b	$BE,$39,$1E,$D8,$A7,$11,$EB,$9F,$80,$47,$E0,$73,$7C,$32,$38,$21
	.dc.b	$F5,$F8,$7E,$A9,$F5,$A3,$FA,$F7,$D7,$3E,$BE,$FD,$84,$BF,$81,$39
	.dc.b	$6C,$7D,$A8,$34,$91,$FE,$04,$8F,$F3,$5B,$7E,$3B,$81,$1D,$7D,$B7
	.dc.b	$88,$8F,$82,$2B,$FE,$A4,$94,$E3,$8F,$FA,$94,$76,$9D,$79,$1E,$52
	.dc.b	$3E,$36,$BC,$FC,$9F,$82,$7C,$28,$F8,$7E,$24,$9F,$82,$FC,$6F,$17
	.dc.b	$11,$FE,$25,$BF,$F2,$0B,$EF,$9F,$88,$7C,$5E,$EF,$9B,$10,$F9,$6D
	.dc.b	$7D,$85,$C7,$D7,$17,$AB,$FD,$C5,$7F,$E0,$A9,$F7,$D7,$D8,$AF,$F0
	.dc.b	$B7,$81,$FB,$B8,$FC,$B2,$F9,$4C,$9A,$EE,$93,$A8,$CB,$C2,$39,$FC
	.dc.b	$7C,$8E,$6F,$6F,$86,$C4,$7F,$B5,$AF,$FD,$D6,$E8,$F0,$C7,$93,$D8
	.dc.b	$91,$E4,$7C,$FF,$07,$88,$E5,$17,$3F,$F0,$65,$FF,$8F,$E7,$81,$0D
	.dc.b	$9E,$EA,$E1,$AF,$F9,$5A,$13,$FE,$13,$7F,$CB,$2D,$73,$3C,$1C,$DD
	.dc.b	$72,$77,$BF,$1C,$66,$BC,$93,$E2,$24,$E9,$F2,$1C,$6F,$C9,$3E,$12
	.dc.b	$5F,$3B,$E8,$DA,$E0,$7A,$1C,$3D,$AB,$9C,$62,$FC,$13,$F4,$E3,$13
	.dc.b	$E0,$AF,$9F,$3D,$0F,$8A,$F8,$1C,$5F,$96,$37,$1D,$CF,$8A,$6A,$BA
	.dc.b	$B8,$3C,$07,$28,$E8,$F1,$3B,$9C,$4D,$14,$D4,$F8,$5F,$92,$62,$39
	.dc.b	$DF,$B8,$97,$ED,$9D,$C6,$1F,$23,$5F,$B6,$7D,$73,$FC,$E7,$5F,$E5
	.dc.b	$C6,$D7,$F5,$48,$C4,$75,$3F,$D4,$4E,$1F,$80,$70,$3F,$D4,$2F,$E3
	.dc.b	$7C,$37,$DC,$FF,$80,$8F,$8F,$28,$CB,$FD,$2D,$AF,$BC,$C4,$71,$7E
	.dc.b	$39,$FE,$07,$70,$EC,$C1,$C0,$DD,$FF,$83,$E3,$6F,$C3,$AF,$34,$63
	.dc.b	$0F,$74,$FC,$7B,$F1,$BC,$6F,$9C,$FF,$82,$2B,$84,$7C,$E8,$E2,$65
	.dc.b	$F1,$1E,$59,$E4,$F9,$8C,$46,$DC,$47,$97,$0E,$B1,$F1,$EB,$E4,$DF
	.dc.b	$92,$F8,$5E,$EB,$BB,$D8,$C3,$D2,$36,$79,$9F,$41,$FE,$DC,$9E,$8D
	.dc.b	$AE,$E7,$8B,$0F,$CA,$47,$33,$47,$C4,$29,$93,$F3,$E7,$E8,$67,$AB
	.dc.b	$EB,$38,$14,$E1,$1D,$AF,$D5,$7E,$89,$F3,$9D,$23,$2F,$81,$FB,$9E
	.dc.b	$FF,$F4,$2F,$A9,$93,$FD,$0A,$3F,$A2,$98,$7F,$9C,$46,$1F,$F0,$E7
	.dc.b	$D8,$FF,$AB,$95,$7F,$AF,$63,$0D,$F7,$7F,$DC,$91,$FE,$3D,$C3,$F9
	.dc.b	$83,$A6,$C6,$8D,$F8,$BC,$3D,$A9,$87,$81,$E7,$E3,$62,$3F,$DC,$E3
	.dc.b	$0F,$F5,$73,$83,$13,$FE,$4A,$8C,$C7,$0F,$AF,$7B,$3F,$EF,$96,$C7
	.dc.b	$F8,$E5,$AF,$BB,$2F,$4D,$FB,$3D,$7E,$84,$9F,$06,$1F,$EA,$87,$CA
	.dc.b	$BD,$6F,$11,$23,$6F,$97,$D3,$91,$C9,$AF,$BB,$EC,$1F,$43,$FE,$64
	.dc.b	$6C,$7B,$DA,$EE,$3E,$07,$8F,$23,$CA,$D7,$CB,$9F,$C5,$23,$EA,$7F
	.dc.b	$CE,$3F,$74,$F3,$E9,$27,$E7,$46,$5F,$A1,$F6,$0F,$73,$FC,$F2,$D8
	.dc.b	$CC,$60,$FF,$3E,$FD,$8F,$F9,$FA,$0D,$23,$FE,$82,$FB,$0F,$F0,$C3
	.dc.b	$F8,$06,$BF,$DE,$C3,$87,$FB,$F3,$50,$FB,$16,$5F,$E8,$E6,$C7,$E2
	.dc.b	$5A,$FE,$39,$AF,$7F,$93,$F6,$3F,$92,$6B,$9C,$3F,$D3,$31,$88,$FF
	.dc.b	$08,$7D,$8B,$EA,$3F,$D3,$AD,$8F,$CC,$35,$FC,$EB,$5F,$73,$F3,$DC
	.dc.b	$EB,$C8,$8F,$E8,$7E,$C1,$F0,$1B,$8F,$F1,$04,$62,$75,$E0,$3C,$9E
	.dc.b	$17,$BB,$F8,$FF,$74,$EE,$F8,$31,$86,$A8,$D6,$FF,$B1,$6C,$7E,$DB
	.dc.b	$EC,$7F,$5F,$3F,$B5,$6C,$67,$BB,$9C,$62,$3E,$E2,$F8,$7C,$92,$B9
	.dc.b	$3C,$3D,$09,$1D,$5B,$5E,$1F,$47,$A9,$27,$C4,$FA,$BE,$C3,$CF,$3F
	.dc.b	$8B,$6B,$FE,$47,$6B,$F8,$A6,$AF,$B0,$7C,$87,$E2,$BC,$7F,$C4,$C3
	.dc.b	$E1,$31,$8F,$F9,$46,$36,$FF,$B1,$DD,$37,$8F,$78,$7E,$62,$7F,$11
	.dc.b	$39,$38,$73,$6A,$5F,$81,$7F,$A8,$27,$FD,$A4,$BD,$8D,$39,$9F,$E3
	.dc.b	$87,$E9,$1D,$4F,$FB,$67,$EC,$7F,$CC,$AD,$E7,$ED,$1F,$8C,$8C,$1F
	.dc.b	$ED,$FF,$9E,$7B,$1E,$36,$C6,$7D,$9C,$4F,$9F,$F0,$2F,$F7,$AC,$FF
	.dc.b	$B9,$DB,$97,$90,$2F,$AC,$E1,$F0,$BE,$FF,$41,$E9,$5F,$4F,$0D,$C7
	.dc.b	$FD,$E5,$F5,$EF,$81,$F8,$CF,$B3,$3E,$CE,$67,$7C,$1F,$2F,$F0,$2F
	.dc.b	$F1,$AC,$7F,$BE,$D7,$D6,$E7,$D9,$D3,$F4,$7E,$BF,$ED,$E3,$E9,$6F
	.dc.b	$5F,$F0,$A7,$EE,$C6,$7B,$3E,$C7,$D7,$BE,$5E,$98,$F6,$EE,$9F,$C6
	.dc.b	$31,$FE,$27,$D6,$F9,$3F,$C0,$36,$74,$7F,$A7,$57,$C9,$EA,$7F,$0D
	.dc.b	$F5,$EF,$12,$98,$F1,$6E,$1C,$B6,$FE,$4F,$EE,$BE,$03,$F1,$CD,$9D
	.dc.b	$4A,$1F,$03,$AA,$FC,$A7,$E0,$4F,$4E,$98,$F5,$2E,$3C,$BE,$BD,$E2
	.dc.b	$FE,$6D,$E1,$19,$8C,$57,$E7,$FE,$B9,$F0,$B7,$AF,$0D,$BF,$2F,$C5
	.dc.b	$7F,$68,$7C,$FF,$88,$7F,$40,$73,$F0,$1F,$A9,$8C,$70,$FE,$47,$EB
	.dc.b	$7E,$35,$7B,$BE,$B7,$3D,$5F,$D1,$27,$35,$F9,$79,$CD,$7F,$B3,$7E
	.dc.b	$B3,$FB,$13,$F1,$BE,$B5,$F0,$DF,$B7,$9C,$DF,$E7,$E7,$35,$F3,$27
	.dc.b	$E1,$B7,$83,$87,$0F,$17,$F8,$07,$D6,$FF,$5C,$9F,$A6,$DC,$3F,$55
	.dc.b	$7F,$CB,$B4,$FE,$B7,$39,$7F,$AE,$C6,$6B,$D1,$ED,$79,$7D,$9F,$33
	.dc.b	$E7,$F9,$BF,$E1,$DC,$6F,$0B,$17,$F8,$28,$C3,$E9,$F7,$F3,$A7,$89
	.dc.b	$B9,$FF,$6C,$1F,$49,$D3,$D0,$FE,$35,$18,$BE,$91,$8F,$A9,$CD,$59
	.dc.b	$FC,$8D,$6F,$0F,$E1,$07,$F1,$03,$F8,$56,$3F,$C5,$DC,$FA,$BE,$AF
	.dc.b	$81,$FE,$51,$0F,$9E,$B3,$86,$E1,$F9,$E9,$7C,$FE,$63,$0D,$8F,$0E
	.dc.b	$FF,$1B,$D1,$EA,$DA,$E5,$BE,$1A,$75,$8F,$1D,$3E,$B7,$FC,$48,$47
	.dc.b	$FC,$B8,$64,$FF,$30,$FD,$4B,$EB,$5F,$5B,$FE,$61,$8C,$36,$33,$3C
	.dc.b	$1A,$FF,$8A,$67,$F7,$3C,$1C,$7C,$8F,$F3,$EC,$F6,$6B,$B6,$BF,$E2
	.dc.b	$C1,$3F,$C1,$DF,$E0,$AD,$7F,$9E,$36,$7F,$C5,$AF,$82,$1F,$5A,$F7
	.dc.b	$BA,$4F,$E1,$A7,$9F,$27,$7B,$E2,$DD,$77,$E5,$5A,$FF,$1F,$DB,$9B
	.dc.b	$11,$E6,$6C,$66,$3F,$8D,$FD,$6F,$F9,$2F,$1F,$2F,$D6,$E8,$93,$EC
	.dc.b	$BE,$3E,$EC,$8F,$7B,$A7,$FA,$D6,$9F,$88,$7F,$93,$B5,$CF,$2F,$93
	.dc.b	$EB,$7F,$C0,$D1,$AE,$4F,$3D,$1A,$62,$BD,$7F,$5B,$F2,$F7,$FA,$4D
	.dc.b	$BF,$F2,$FB,$56,$1F,$E1,$67,$F9,$CC,$7C,$D7,$FB,$83,$74,$FF,$30
	.dc.b	$B6,$BA,$EF,$F2,$BE,$B3,$FD,$79,$7F,$E6,$C7,$FD,$78,$DE,$ED,$7C
	.dc.b	$6E,$D9,$1F,$D2,$27,$69,$E6,$BE,$B7,$FB,$23,$BF,$83,$F0,$1F,$6A
	.dc.b	$33,$1F,$E0,$79,$4F,$97,$C1,$93,$C4,$F0,$3E,$9F,$52,$46,$0F,$EB
	.dc.b	$7F,$50,$FA,$8F,$C3,$AF,$AD,$A6,$23,$F1,$7F,$5B,$FE,$55,$AF,$EC
	.dc.b	$6F,$FB,$23,$87,$FD,$12,$DE,$0F,$D3,$3B,$9F,$F0,$F7,$AF,$F9,$F7
	.dc.b	$FB,$5B,$FD,$F5,$BE,$E7,$E6,$23,$51,$FE,$25,$CB,$FD,$9D,$87,$FA
	.dc.b	$59,$B8,$7E,$CD,$B3,$FE,$A0,$5F,$55,$E8,$F7,$24,$FA,$1F,$5F,$B9
	.dc.b	$DF,$C8,$93,$E4,$7C,$5E,$4F,$C4,$FA,$D9,$3F,$21,$F8,$FF,$3D,$F1
	.dc.b	$36,$93,$F1,$1F,$8F,$F0,$1E,$BE,$E9,$3F,$01,$F7,$77,$7D,$0F,$A9
	.dc.b	$FA,$3F,$55,$2B,$EA,$C7,$D0,$CF,$E8,$F8,$32,$7E,$8C,$7C,$77,$E4
	.dc.b	$79,$B1,$F9,$A7,$DC,$F3,$E3,$F8,$07,$EF,$FB,$1F,$73,$29,$3F,$71
	.dc.b	$FB,$FC,$1F,$C2,$BE,$93,$F5,$53,$FE,$57,$6F,$3F,$FB,$21,$BC,$FF
	.dc.b	$C9,$5F,$DE,$CF,$D9,$F8,$CF,$C6,$FA,$52,$3F,$23,$1F,$8C,$D3,$F1
	.dc.b	$BC,$32,$7F,$1B,$1E,$1F,$A6,$F1,$7D,$6C,$9F,$F1,$9E,$3F,$EE,$6C
	.dc.b	$FF,$09,$F4,$5F,$D2,$3B,$0D,$25,$7F,$BA,$4F,$F7,$2F,$D6,$FF,$B0
	.dc.b	$5B,$F3,$1F,$D4,$23,$10,$6F,$D6,$FE,$E0,$9E,$57,$F2,$E8,$FF,$6B
	.dc.b	$63,$FE,$EA,$9F,$DA,$D7,$F9,$21,$AF,$F9,$FD,$B1,$BF,$94,$FE,$6F
	.dc.b	$1A,$9F,$F7,$F3,$5F,$EA,$0D,$7F,$DF,$AD,$4F,$EE,$DF,$71,$8F,$13
	.dc.b	$FB,$5C,$66,$3E,$13,$5E,$F1,$E2,$F0,$A4,$7A,$63,$F8,$97,$97,$E6
	.dc.b	$ED,$93,$FE,$64,$F2,$FF,$8E,$E0,$9E,$E4,$9F,$1C,$65,$BD,$DF,$00
	.dc.b	$8F,$C7,$9F,$87,$CD,$23,$E5,$3F,$1B,$9F,$11,$EA,$8F,$69,$1F,$83
	.dc.b	$F5,$6F,$AD,$FF,$05,$B6,$33,$18,$23,$FE,$7E,$87,$D6,$FF,$6A,$6C
	.dc.b	$66,$7F,$91,$B9,$F8,$F8,$FF,$24,$29,$F8,$F9,$FF,$6A,$3E,$48,$F1
	.dc.b	$91,$F0,$7C,$57,$73,$8F,$13,$B7,$FC,$21,$6F,$23,$C8,$4B,$F4,$7C
	.dc.b	$0D,$7F,$D9,$93,$B7,$D9,$3E,$B9,$F6,$78,$BE,$DF,$16,$57,$C1,$8C
	.dc.b	$BF,$0E,$32,$F4,$8F,$89,$E0,$48,$F9,$0F,$F8,$43,$D8,$79,$99,$9F
	.dc.b	$85,$3F,$48,$8F,$D6,$7E,$73,$0E,$67,$FA,$C3,$8B,$F7,$DA,$7F,$8C
	.dc.b	$5B,$00,$E1,$BE,$0F,$F8,$8A,$30,$4F,$0B,$11,$9F,$80,$95,$DD,$FE
	.dc.b	$A5,$F6,$18,$8F,$F7,$B5,$FF,$AC,$9D,$1F,$D6,$A3,$B7,$BD,$23,$D0
	.dc.b	$FF,$29,$F9,$5C,$47,$28,$BC,$7F,$CA,$66,$E7,$ED,$9B,$5E,$DF,$1B
	.dc.b	$FD,$E6,$F4,$21,$CC,$FE,$47,$5C,$1F,$87,$36,$78,$C6,$C9,$7F,$8D
	.dc.b	$7E,$56,$3F,$ED,$C7,$99,$DA,$34,$F4,$31,$87,$F3,$FC,$3C,$15,$C6
	.dc.b	$7A,$CE,$5D,$9D,$B0,$F2,$57,$06,$C6,$8E,$C7,$23,$E1,$46,$DD,$61
	.dc.b	$EC,$C3,$CB,$87,$85,$B7,$D2,$B9,$1D,$23,$2F,$B8,$CC,$F1,$C3,$E9
	.dc.b	$46,$67,$6E,$AB,$EE,$3E,$7A,$E0,$FB,$4E,$15,$C9,$E0,$60,$D1,$83
	.dc.b	$FB,$66,$1E,$D7,$CF,$87,$B6,$79,$61,$E2,$23,$F3,$67,$94,$61,$F8
	.dc.b	$E7,$97,$0F,$85,$19,$8F,$85,$3D,$0E,$51,$EE,$8F,$F4,$F1,$F4,$8B
	.dc.b	$87,$DD,$6D,$7F,$AD,$A3,$11,$EF,$32,$7F,$52,$72,$FC,$09,$FA,$FE
	.dc.b	$BF,$ED,$4F,$80,$E2,$3F,$80,$46,$2F,$F8,$03,$B8,$F6,$1A,$21,$F0
	.dc.b	$A3,$6E,$27,$E0,$57,$D6,$8F,$E0,$F5,$CD,$E0,$7E,$89,$C4,$F0,$BB
	.dc.b	$7D,$2B,$67,$23,$B3,$F2,$DF,$73,$B9,$F6,$47,$0B,$FF,$4B,$12,$BF
	.dc.b	$53,$7D,$A7,$89,$0E,$C7,$F1,$27,$E2,$3E,$73,$C2,$6E,$7D,$C7,$68
	.dc.b	$DC,$72,$AF,$19,$ED,$78,$99,$BF,$53,$F8,$63,$B1,$E1,$3E,$89,$49
	.dc.b	$1C,$DB,$1F,$04,$E7,$5E,$4A,$F7,$4F,$52,$47,$9D,$EB,$5E,$6A,$E5
	.dc.b	$1E,$DB,$E3,$5F,$26,$3F,$A3,$7B,$9F,$CF,$FD,$B9,$18,$21,$FC,$E9
	.dc.b	$CB,$FC,$F2,$3F,$DC,$FF,$72,$F4,$BF,$DD,$3F,$57,$FE,$B6,$3F,$D6
	.dc.b	$F1,$9A,$FE,$A1,$9F,$F2,$49,$D7,$93,$C0,$E7,$7C,$BD,$99,$7C,$35
	.dc.b	$FE,$B0,$AD,$3F,$EF,$8F,$AD,$71,$70,$AF,$F2,$C5,$69,$FF,$1E,$B5
	.dc.b	$CD,$7F,$90,$1E,$8F,$AB,$DA,$91,$E1,$3F,$B4,$4F,$1C,$FD,$5E,$57
	.dc.b	$18,$FF,$54,$CF,$68,$C1,$9F,$AD,$F8,$A4,$8E,$87,$AE,$3D,$5E,$B7
	.dc.b	$8F,$33,$49,$3E,$AF,$23,$C0,$47,$07,$F9,$6A,$7F,$D8,$CF,$F9,$E5
	.dc.b	$AF,$F1,$66,$BF,$31,$AF,$F8,$17,$EB,$5F,$17,$F1,$5B,$1F,$E6,$66
	.dc.b	$BF,$E8,$56,$BF,$C8,$B2,$FE,$C2,$DD,$FE,$31,$1F,$F3,$7F,$D5,$BB
	.dc.b	$BE,$93,$63,$31,$83,$EB,$7D,$67,$F8,$9A,$3F,$CE,$ED,$7F,$D2,$CD
	.dc.b	$7F,$99,$B5,$FC,$4F,$D5,$FF,$A7,$9F,$F3,$FB,$5F,$F4,$F3,$5F,$E7
	.dc.b	$6D,$7F,$23,$F5,$8F,$07,$F1,$0D,$8F,$F4,$4B,$5F,$F5,$43,$5F,$E8
	.dc.b	$BF,$58,$FA,$2F,$C8,$36,$3F,$D2,$2D,$7F,$D6,$0D,$7F,$A6,$ED,$EB
	.dc.b	$7F,$C7,$4D,$7E,$BF,$89,$F9,$B8,$C4,$BE,$9B,$B9,$2B,$BB,$FA,$1F
	.dc.b	$B8,$7C,$1D,$24,$FF,$A9,$23,$2F,$FA,$9B,$EB,$5F,$05,$FE,$C3,$6F
	.dc.b	$B9,$9F,$13,$ED,$57,$F6,$49,$CB,$FE,$CB,$FB,$87,$4F,$E1,$A3,$FC
	.dc.b	$A9,$18,$E6,$FF,$59,$B6,$38,$4E,$D7,$D6,$BE,$53,$0D,$BF,$F5,$DB
	.dc.b	$67,$FD,$B4,$D7,$FB,$BD,$BE,$17,$A3,$6B,$C3,$18,$3A,$FA,$72,$3A
	.dc.b	$B6,$BF,$DC,$DF,$14,$D7,$CC,$DA,$FF,$66,$B6,$77,$3E,$BF,$AB,$7D
	.dc.b	$2F,$F9,$37,$C4,$FF,$24,$C3,$EB,$31,$8F,$F2,$28,$DB,$FC,$29,$D3
	.dc.b	$6F,$78,$7F,$9F,$E7,$F0,$53,$93,$87,$2E,$A5,$F5,$8F,$6B,$0B,$DB
	.dc.b	$B1,$87,$87,$D6,$7F,$A2,$B0,$FE,$36,$DB,$69,$FE,$62,$8F,$C3,$CE
	.dc.b	$B9,$BC,$1F,$38,$F6,$E9,$8F,$6E,$D7,$F2,$6F,$AC,$71,$F9,$DB,$19
	.dc.b	$F8,$7C,$4F,$AF,$D0,$FD,$74,$FF,$2E,$5E,$B6,$7C,$4D,$3F,$07,$EB
	.dc.b	$1F,$07,$76,$C3,$70,$FE,$6F,$F5,$8F,$A5,$F9,$0B,$DE,$B1,$87,$E6
	.dc.b	$7D,$67,$D6,$E5,$F8,$77,$FD,$6A,$31,$CB,$FE,$86,$7E,$CC,$E8,$FF
	.dc.b	$5B,$3A,$9F,$EC,$3F,$58,$FA,$CF,$B0,$D8,$FE,$96,$BE,$33,$33,$F7
	.dc.b	$7A,$5F,$29,$4C,$7B,$17,$0F,$EA,$7F,$59,$FC,$09,$BE,$6F,$E1,$63
	.dc.b	$FD,$D5,$18,$3F,$AC,$FB,$8F,$83,$FC,$53,$63,$3F,$07,$89,$FC,$77
	.dc.b	$D6,$7F,$4C,$9F,$EC,$2D,$E0,$F8,$01,$FF,$71,$C6,$BD,$3F,$CB,$7D
	.dc.b	$63,$C2,$B1,$86,$CF,$F6,$CF,$AC,$7C,$47,$E7,$5F,$B1,$19,$8C,$74
	.dc.b	$7E,$87,$C6,$7D,$67,$E8,$FE,$B1,$79,$8C,$3F,$A7,$FA,$CF,$8D,$B7
	.dc.b	$A3,$EB,$37,$3F,$DF,$27,$3D,$9F,$EC,$29,$CD,$7B,$BE,$AB,$DA,$FC
	.dc.b	$0F,$AC,$FE,$20,$E7,$DD,$FD,$A4,$62,$FF,$D9,$75,$F0,$71,$FF,$04
	.dc.b	$BF,$26,$33,$C1,$FB,$B8,$C4,$FF,$8A,$3E,$B3,$FE,$00,$3F,$E1,$06
	.dc.b	$E3,$E3,$F6,$7F,$81,$46,$30,$EF,$1C,$CE,$9F,$25,$6F,$C3,$F3,$BF
	.dc.b	$C2,$78,$FF,$21,$18,$3F,$C3,$AE,$31,$F8,$1F,$0B,$E5,$AF,$33,$6B
	.dc.b	$FC,$47,$8B,$EA,$3A,$79,$BF,$C5,$63,$13,$D2,$31,$F4,$F9,$AB,$3C
	.dc.b	$9F,$E3,$98,$FF,$0D,$8F,$1F,$C0,$E3,$D4,$F4,$3A,$FA,$0C,$4F,$F8
	.dc.b	$D2,$91,$87,$F2,$A3,$E8,$78,$4E,$DB,$1A,$D7,$E2,$79,$FE,$26,$D7
	.dc.b	$5D,$B0,$D3,$C1,$1E,$8A,$7D,$67,$F9,$55,$A7,$F9,$5D,$C1,$FE,$D0
	.dc.b	$43,$EA,$5F,$56,$FA,$CD,$4E,$63,$0F,$19,$DC,$E9,$B1,$9F,$1B,$F9
	.dc.b	$DB,$78,$F5,$18,$8F,$F7,$0C,$A3,$11,$FE,$78,$8F,$E0,$CF,$F9,$81
	.dc.b	$AF,$F0,$06,$CF,$EC,$23,$D7,$1C,$E3,$DB,$F5,$8F,$6F,$A7,$1F,$FA
	.dc.b	$12,$78,$5F,$C1,$8C,$6B,$A8,$F0,$79,$18,$7A,$F2,$D7,$E4,$37,$23
	.dc.b	$9F,$9C,$FF,$7E,$E3,$DA,$5F,$35,$B2,$5F,$B6,$F9,$7B,$32,$3E,$03
	.dc.b	$A7,$F8,$FE,$9F,$8B,$7F,$D0,$6D,$73,$CB,$E5,$FA,$CF,$F0,$14,$6B
	.dc.b	$93,$D1,$46,$98,$AF,$67,$D6,$7C,$CD,$FF,$D5,$2D,$BF,$C3,$B5,$61
	.dc.b	$FE,$12,$7F,$D2,$B1,$FB,$36,$F0,$74,$3A,$F0,$7C,$9F,$AC,$7C,$E5
	.dc.b	$F5,$E5,$FF,$A7,$A7,$2D,$8C,$BF,$D0,$23,$FA,$C3,$B9,$FA,$9F,$4B
	.dc.b	$FD,$7E,$39,$7C,$07,$73,$3E,$CD,$8E,$DE,$0F,$8C,$C3,$D5,$F3,$7A
	.dc.b	$92,$36,$7F,$AC,$BF,$6F,$FB,$6D,$E6,$FC,$24,$62,$3F,$AA,$BF,$84
	.dc.b	$7F,$D7,$38,$BC,$0F,$39,$23,$3E,$17,$E3,$7F,$02,$FF,$06,$C7,$FB
	.dc.b	$01,$AF,$F6,$06,$F5,$FE,$69,$E9,$1E,$5F,$82,$F1,$F5,$48,$C1,$E2
	.dc.b	$FA,$AF,$8F,$F0,$52,$30,$77,$E9,$F1,$4C,$8C,$1F,$07,$E2,$BD,$FF
	.dc.b	$13,$23,$07,$FB,$5B,$EB,$7F,$B2,$BF,$02,$FE,$A9,$FA,$98,$CC,$7F
	.dc.b	$7A,$F4,$9F,$4D,$FA,$D7,$3B,$7E,$D9,$B7,$FE,$FC,$77,$5F,$EE,$4F
	.dc.b	$AC,$FC,$1B,$FE,$04,$6D,$7E,$F9,$B5,$FC,$82,$51,$F8,$E7,$FC,$17
	.dc.b	$83,$EA,$3C,$C4,$8F,$66,$3E,$5C,$64,$7F,$BD,$1F,$F6,$66,$3E,$7A
	.dc.b	$FE,$0D,$83,$E2,$AF,$4C,$C3,$D1,$1F,$0F,$DE,$F8,$98,$C9,$F9,$51
	.dc.b	$F0,$FD,$6F,$88,$C8,$F0,$47,$6E,$B9,$1F,$0D,$F0,$F6,$7C,$FF,$85
	.dc.b	$93,$F5,$5F,$A5,$5F,$3F,$DD,$93,$EB,$7E,$94,$FF,$1F,$FA,$CF,$CD
	.dc.b	$4F,$F4,$96,$BF,$2C,$38,$7E,$73,$5F,$E9,$EF,$E9,$5E,$0D,$77,$DD
	.dc.b	$FD,$41,$E6,$DD,$9D,$FF,$30,$93,$FD,$6B,$87,$F9,$DD,$FE,$C5,$FE
	.dc.b	$7B,$F5,$7F,$75,$BE,$0F,$F3,$98,$C7,$37,$9B,$A3,$13,$E1,$9F,$C2
	.dc.b	$61,$ED,$6B,$F8,$96,$BF,$DA,$9F,$F4,$7B,$FD,$A1,$AF,$F6,$F6,$BC
	.dc.b	$F7,$EC,$78,$F8,$24,$7A,$F0,$ED,$3E,$7F,$AC,$F4,$7D,$24,$AD,$1E
	.dc.b	$18,$C3,$DE,$3D,$44,$7F,$BD,$FE,$08,$FB,$AF,$EB,$4D,$8E,$61,$C3
	.dc.b	$FE,$6E,$C9,$5F,$27,$E6,$E4,$F0,$9F,$F7,$AB,$7E,$BB,$2F,$C1,$9F
	.dc.b	$98,$47,$E3,$78,$1F,$3C,$8E,$08,$78,$7C,$0F,$B0,$7D,$98,$FE,$E5
	.dc.b	$F5,$DF,$C0,$A7,$FD,$06,$FE,$E5,$F5,$B5,$CF,$7F,$0E,$FF,$C7,$4E
	.dc.b	$67,$F3,$4D,$7F,$83,$35,$FF,$13,$BC,$F9,$3F,$C1,$3E,$03,$EA,$BF
	.dc.b	$8E,$1D,$6B,$C3,$D9,$E6,$23,$E4,$EC,$AF,$A2,$2D,$FF,$85,$30,$FF
	.dc.b	$75,$4F,$AB,$E6,$FD,$FF,$1D,$23,$04,$3E,$0B,$97,$E1,$C7,$C6,$F7
	.dc.b	$E4,$7C,$47,$E4,$FA,$B8,$8F,$F2,$25,$FF,$BB,$5F,$F2,$0B,$5C,$CF
	.dc.b	$C8,$3F,$C4,$5F,$82,$3E,$B2,$9C,$F2,$F5,$57,$D6,$21,$F5,$3E,$02
	.dc.b	$BF,$60,$B1,$F0,$63,$FC,$5D,$E7,$3E,$DA,$92,$BF,$2F,$0F,$A6,$C4
	.dc.b	$7F,$B1,$23,$83,$A9,$E7,$3D,$3E,$DA,$4F,$27,$BF,$C2,$E2,$39,$45
	.dc.b	$C7,$FC,$8F,$87,$FB,$1A,$78,$10,$D9,$E4,$AE,$1E,$0F,$F4,$F9,$5E
	.dc.b	$2B,$FC,$E3,$1C,$DE,$24,$BE,$F3,$E4,$8E,$D1,$86,$D7,$5B,$E9,$1F
	.dc.b	$E0,$DA,$B7,$D6,$F8,$4F,$5B,$F2,$57,$58,$C1,$E4,$78,$DF,$26,$C7
	.dc.b	$80,$D1,$4F,$F0,$E4,$6D,$C5,$F8,$6B,$BC,$62,$BC,$17,$E4,$24,$EA
	.dc.b	$7E,$D5,$F3,$7C,$A7,$63,$91,$D9,$D4,$F2,$8F,$9B,$18,$9E,$93,$F6
	.dc.b	$A3,$13,$EA,$3B,$9D,$0E,$F7,$E9,$75,$3D,$5F,$9C,$7D,$F3,$83,$98
	.dc.b	$E1,$1A,$39,$DF,$C0,$BF,$19,$23,$AB,$F6,$DF,$C6,$1F,$DB,$67,$F1
	.dc.b	$2F,$FA,$D1,$AF,$FA,$ED,$AE,$E7,$FB,$54,$62,$3E,$61,$93,$94,$7F
	.dc.b	$8B,$FE,$1A,$47,$AA,$BF,$C5,$D1,$88,$F4,$9F,$E1,$F7,$17,$EF,$AB
	.dc.b	$5C,$2B,$91,$FE,$CE,$F3,$17,$A8,$F7,$90,$46,$9E,$C7,$23,$31,$C5
	.dc.b	$F9,$97,$F4,$E7,$84,$7F,$92,$E7,$EE,$BC,$0E,$67,$A8,$D1,$89,$F5
	.dc.b	$36,$BD,$4E,$E3,$FC,$A5,$3E,$1C,$3E,$8C,$78,$63,$18,$7E,$CB,$0F
	.dc.b	$1C,$F0,$BF,$54,$63,$0E,$B3,$E0,$9F,$19,$1C,$1C,$5C,$47,$E1,$DF
	.dc.b	$82,$F2,$23,$F2,$F0,$F2,$C6,$23,$E1,$1E,$B7,$15,$E3,$8E,$17,$F3
	.dc.b	$63,$11,$D8,$FA,$CE,$0C,$D7,$68,$CE,$1F,$09,$FE,$09,$3B,$8F,$E2
	.dc.b	$7E,$C7,$FB,$A7,$EC,$1F,$21,$4F,$E7,$F1,$2F,$FC,$27,$3A,$6C,$E5
	.dc.b	$FE,$7D,$E9,$FF,$46,$F5,$24,$7F,$44,$8A,$FF,$43,$8B,$AF,$F3,$A9
	.dc.b	$FF,$7D,$7D,$67,$FA,$12,$81,$C3,$7D,$BD,$F9,$9C,$1F,$F2,$B4,$FF
	.dc.b	$A4,$27,$FC,$C4,$D7,$FD,$1C,$D7,$AB,$D3,$EC,$1F,$65,$FE,$93,$C5
	.dc.b	$C5,$FE,$95,$9C,$DF,$F8,$D2,$FC,$67,$F9,$89,$FF,$38,$1D,$3D,$1F
	.dc.b	$19,$1C,$8F,$B0,$7C,$6F,$7C,$3F,$D4,$51,$89,$FF,$20,$B5,$63,$E4
	.dc.b	$7B,$1E,$3E,$AF,$82,$47,$BE,$3F,$E0,$37,$C3,$F8,$27,$F7,$56,$F4
	.dc.b	$FF,$9F,$63,$11,$EA,$6B,$F2,$BE,$9F,$E6,$91,$EA,$FF,$A1,$7E,$C7
	.dc.b	$FD,$93,$1F,$E9,$16,$BF,$EB,$D6,$BF,$71,$A9,$FF,$24,$3C,$7C,$9E
	.dc.b	$84,$7D,$7A,$7D,$E2,$38,$3E,$6F,$DB,$BD,$6F,$F4,$C3,$63,$31,$83
	.dc.b	$FD,$39,$F6,$3F,$C7,$A3,$EA,$B5,$FB,$6D,$7F,$D5,$8D,$7F,$DB,$1F
	.dc.b	$65,$FE,$A5,$4D,$23,$FE,$A9,$FB,$1F,$96,$FF,$B7,$DA,$FE,$79,$A9
	.dc.b	$BF,$65,$F7,$E3,$F1,$AD,$7F,$D7,$CD,$7F,$DD,$4D,$7F,$53,$EF,$FC
	.dc.b	$E2,$7C,$DF,$FA,$ED,$FA,$3F,$64,$F8,$4F,$F5,$F3,$6B,$F2,$ED,$7F
	.dc.b	$3A,$D7,$FD,$9C,$FC,$D7,$39,$F2,$23,$FA,$0F,$B1,$7C,$3E,$E3,$FE
	.dc.b	$00,$46,$27,$5E,$33,$CA,$E1,$7B,$BF,$E0,$7F,$6E,$F0,$7D,$CD,$E3
	.dc.b	$FD,$6B,$67,$F6,$6D,$7F,$DC,$9F,$65,$FB,$09,$FD,$9B,$63,$F7,$2D
	.dc.b	$77,$DD,$F2,$E3,$FB,$3C,$F0,$F2,$30,$47,$F7,$BF,$03,$F2,$DB,$C1
	.dc.b	$F4,$E3,$18,$7F,$06,$6B,$FE,$F7,$FC,$67,$F8,$FD,$7F,$B0,$1B,$1F
	.dc.b	$C0,$23,$F9,$1A,$8F,$8D,$1B,$8F,$E0,$9E,$07,$E1,$FE,$C1,$E4,$62
	.dc.b	$B8,$C6,$1F,$F6,$6A,$F2,$35,$7F,$CC,$23,$4F,$84,$FE,$67,$F6,$0F
	.dc.b	$89,$DC,$23,$17,$FE,$D8,$5D,$1A,$73,$3C,$C7,$D8,$3C,$2E,$6B,$E0
	.dc.b	$2C,$61,$EF,$8B,$EA,$AF,$D5,$61,$B9,$FF,$BA,$3E,$C1,$F1,$FE,$65
	.dc.b	$E8,$58,$C3,$EA,$F8,$B7,$C7,$53,$1F,$33,$70,$FF,$79,$FD,$87,$E6
	.dc.b	$9B,$DF,$F0,$A3,$D5,$18,$3F,$DF,$3F,$61,$FE,$A7,$6F,$B9,$F2,$23
	.dc.b	$FD,$4D,$18,$3F,$DF,$FF,$84,$7E,$03,$AB,$F8,$03,$F7,$B9,$7E,$43
	.dc.b	$99,$FC,$1B,$99,$FE,$0D,$F5,$FF,$A1,$7F,$84,$7D,$87,$F2,$3A,$FE
	.dc.b	$D7,$1A,$F1,$3E,$CC,$E7,$0F,$A9,$F0,$EF,$86,$FB,$AB,$E1,$AC,$61
	.dc.b	$FC,$0F,$D8,$7E,$92,$BE,$E3,$FA,$BC,$7F,$8C,$39,$6F,$8F,$F8,$8F
	.dc.b	$6D,$F4,$B7,$EE,$F0,$DD,$BC,$E7,$E8,$3E,$C5,$F1,$DF,$90,$5E,$CD
	.dc.b	$8C,$3F,$94,$FB,$1F,$F2,$CD,$72,$D7,$FC,$17,$18,$7F,$24,$F4,$8D
	.dc.b	$42,$35,$A7,$F3,$1F,$B1,$7D,$97,$E8,$1B,$7C,$1F,$BD,$7F,$9D,$72
	.dc.b	$FE,$7F,$EC,$9F,$29,$FA,$86,$D7,$81,$7C,$A6,$67,$F5,$BF,$63,$FE
	.dc.b	$12,$AF,$2B,$6F,$F5,$5F,$65,$F5,$E7,$27,$ED,$3E,$CB,$F2,$73,$FA
	.dc.b	$F6,$E9,$FD,$2F,$EC,$BD,$B5,$FC,$32,$72,$FF,$0E,$8F,$DC,$BF,$11
	.dc.b	$B5,$FB,$97,$F7,$75,$CF,$CF,$FE,$07,$F6,$5F,$95,$AF,$F1,$CB,$55
	.dc.b	$7E,$16,$FF,$8D,$CE,$63,$13,$FE,$40,$6B,$C3,$D7,$FE,$1F,$18,$1F
	.dc.b	$57,$FC,$4B,$E4,$FF,$7B,$87,$9E,$31,$E5,$70,$F7,$38,$C5,$9D,$1F
	.dc.b	$80,$D1,$F5,$1C,$FC,$CF,$E3,$B1,$89,$FC,$2E,$29,$CF,$97,$AB,$B5
	.dc.b	$CD,$FE,$4B,$AB,$E5,$BC,$DE,$66,$1B,$5F,$90,$97,$D9,$66,$30,$D8
	.dc.b	$FE,$50,$3E,$B3,$AC,$76,$6C,$F3,$D7,$E2,$7C,$0F,$89,$B1,$DF,$7C
	.dc.b	$34,$E0,$78,$F0,$7D,$97,$F8,$74,$8F,$F9,$68,$C9,$FE,$60,$FA,$EF
	.dc.b	$DD,$FD,$97,$F9,$79,$B1,$9F,$11,$39,$6C,$71,$FB,$2F,$F2,$EB,$79
	.dc.b	$39,$46,$23,$FB,$08,$B4,$66,$3F,$B2,$C3,$E6,$7F,$CD,$91,$FE,$8A
	.dc.b	$8C,$3F,$BE,$6B,$FD,$4D,$E7,$5F,$E9,$0F,$B2,$7C,$FF,$4E,$4E,$77
	.dc.b	$E1,$6D,$FE,$22,$3A,$61,$8E,$59,$1D,$DE,$3F,$65,$EA,$CB,$FC,$79
	.dc.b	$7F,$D4,$2B,$C9,$EC,$74,$38,$BE,$5F,$83,$EE,$47,$67,$F1,$8C,$FF
	.dc.b	$08,$DE,$5F,$3C,$62,$3F,$B8,$7D,$93,$EC,$F3,$CD,$27,$93,$C2,$FD
	.dc.b	$FF,$67,$F1,$B4,$FF,$29,$37,$9B,$FC,$26,$DB,$FF,$5B,$E2,$FB,$4E
	.dc.b	$3C,$52,$7F,$96,$D7,$C7,$7F,$C0,$4F,$15,$F6,$9F,$E3,$D7,$F9,$9B
	.dc.b	$FE,$C0,$27,$B3,$F4,$9C,$78,$34,$C4,$7D,$6F,$B3,$FF,$00,$D7,$2F
	.dc.b	$8C,$E0,$67,$E0,$6C,$76,$F0,$FA,$18,$7F,$C2,$D8,$FE,$FD,$F1,$FA
	.dc.b	$DE,$13,$6F,$F4,$8F,$B4,$FB,$4F,$F8,$75,$BC,$3F,$E7,$96,$F0,$FF
	.dc.b	$8E,$97,$57,$9F,$C2,$97,$FD,$4B,$54,$7A,$FC,$5F,$C8,$FD,$A7,$F4
	.dc.b	$6A,$FE,$AE,$D7,$FC,$58,$DC,$BF,$D1,$0F,$28,$F3,$7C,$6F,$E6,$B2
	.dc.b	$FF,$20,$47,$A7,$AF,$13,$F2,$7D,$0F,$95,$E8,$CA,$F2,$47,$8B,$6F
	.dc.b	$1B,$23,$C7,$1D,$7E,$4B,$B7,$C8,$49,$F8,$D1,$FA,$AF,$AB,$7C,$B7
	.dc.b	$D7,$B9,$5F,$DC,$A7,$2F,$F7,$4F,$A6,$FE,$FE,$DA,$FD,$9C,$6A,$3F
	.dc.b	$C9,$8E,$9F,$F2,$77,$D9,$FF,$90,$1F,$EF,$AD,$73,$D7,$FB,$C8,$FC
	.dc.b	$84,$7F,$95,$6F,$FC,$1E,$F8,$FD,$DF,$41,$E9,$F7,$7C,$38,$49,$F3
	.dc.b	$3E,$49,$E9,$F6,$9D,$7B,$25,$78,$A3,$E1,$7A,$FF,$0E,$E4,$FC,$18
	.dc.b	$F6,$FC,$7F,$B9,$91,$F0,$E3,$D1,$DD,$23,$DA,$FA,$FB,$FE,$77,$C3
	.dc.b	$49,$F9,$8F,$D1,$AF,$93,$F5,$92,$7E,$73,$F3,$27,$F8,$FF,$DA,$7E
	.dc.b	$5A,$7F,$C4,$AD,$7F,$CF,$4D,$EC,$FC,$F7,$8B,$C3,$C1,$E3,$E6,$4B
	.dc.b	$E3,$15,$FE,$76,$2F,$F5,$2F,$F3,$E6,$BF,$EA,$06,$A3,$87,$47,$F5
	.dc.b	$38,$FF,$63,$46,$5A,$E8,$EE,$3E,$C3,$EF,$E3,$FE,$ED,$6F,$4F,$FA
	.dc.b	$B9,$B3,$CE,$FF,$A0,$C3,$6A,$49,$FF,$24,$C3,$8B,$11,$EE,$8E,$0F
	.dc.b	$89,$AE,$67,$FB,$64,$3E,$D2,$92,$78,$35,$FF,$55,$43,$EC,$BC,$A7
	.dc.b	$9A,$6C,$F7,$8E,$A4,$7C,$5F,$65,$E0,$23,$83,$CD,$F2,$DF,$00,$8F
	.dc.b	$BA,$3E,$1F,$81,$23,$E0,$BF,$1B,$C6,$C4,$7C,$18,$F6,$91,$FA,$7F
	.dc.b	$63,$FB,$77,$FB,$EB,$5F,$F6,$FB,$5F,$C0,$35,$3F,$BF,$DD,$BF,$0E
	.dc.b	$BF,$EE,$F3,$FB,$AA,$FF,$68,$3B,$9D,$47,$FA,$02,$35,$5C,$A3,$BD
	.dc.b	$72,$EC,$E7,$CD,$FE,$21,$B3,$49,$5E,$C7,$C3,$CF,$23,$B4,$79,$3E
	.dc.b	$36,$47,$85,$F3,$FC,$0E,$23,$F4,$B7,$FE,$B1,$7E,$3C,$F8,$4F,$17
	.dc.b	$A9,$EE,$F7,$E4,$FA,$E3,$E0,$98,$7E,$0C,$61,$F8,$0E,$74,$F8,$11
	.dc.b	$F3,$23,$4F,$C8,$F7,$BE,$5B,$0F,$90,$FC,$E3,$0F,$CE,$30,$7F,$AB
	.dc.b	$38,$1F,$51,$FE,$2A,$6C,$66,$31,$F7,$19,$64,$64,$E5,$1C,$7E,$0A
	.dc.b	$47,$57,$FA,$1F,$D9,$E2,$19,$7B,$30,$56,$63,$DF,$E2,$2F,$1C,$9F
	.dc.b	$C3,$70,$FE,$E3,$78,$7F,$88,$56,$DF,$01,$0E,$F3,$C7,$0E,$F1,$FD
	.dc.b	$66,$F9,$3D,$0E,$18,$78,$A3,$33,$C2,$79,$9F,$C7,$DC,$4E,$E3,$8D
	.dc.b	$77,$C3,$AE,$1E,$59,$E2,$77,$2D,$F2,$AE,$11,$E5,$39,$9D,$EB,$9E
	.dc.b	$1E,$46,$C6,$8E,$98,$7C,$D8,$FA,$31,$9B,$EC,$F3,$3B,$1E,$28,$E8
	.dc.b	$E2,$79,$3C,$63,$9B,$B3,$06,$EF,$D2,$E4,$93,$F1,$5B,$1A,$8E,$B1
	.dc.b	$FE,$06,$BF,$0B,$A3,$EE,$B8,$8D,$9C,$CB,$3C,$C9,$3C,$0F,$29,$CC
	.dc.b	$61,$F4,$E7,$8C,$7C,$33,$51,$99,$E7,$1F,$31,$B1,$98,$E1,$3C,$9D
	.dc.b	$1E,$13,$E1,$1F,$EA,$6C,$3F,$14,$DC,$3F,$D7,$11,$88,$F9,$26,$4F
	.dc.b	$EC,$AE,$5F,$31,$93,$F1,$3F,$65,$FB,$93,$E3,$61,$F9,$F7,$11,$F0
	.dc.b	$AF,$F6,$11,$88,$F8,$95,$F8,$47,$FC,$BB,$19,$9F,$64,$E4,$46,$6B
	.dc.b	$E2,$C6,$63,$15,$F1,$27,$31,$88,$F8,$A4,$9E,$6F,$F1,$38,$F4,$BF
	.dc.b	$AD,$78,$47,$86,$EC,$6A,$7C,$E7,$13,$DD,$5F,$05,$C1,$EE,$9E,$2E
	.dc.b	$8E,$C6,$AF,$DF,$1A,$8E,$06,$67,$A3,$83,$E8,$99,$3E,$01,$0F,$29
	.dc.b	$A3,$84,$FD,$27,$81,$F0,$E7,$C1,$1E,$28,$CD,$FC,$67,$8D,$78,$A7
	.dc.b	$94,$60,$F5,$9D,$DE,$4D,$8E,$07,$D8,$F8,$0F,$E8,$1E,$FC,$9F,$E8
	.dc.b	$91,$FD,$0C,$C3,$FC,$5A,$30,$FF,$49,$F4,$3F,$A4,$11,$FE,$99,$E2
	.dc.b	$57,$F8,$EC,$5C,$3F,$A0,$4F,$FB,$E3,$EB,$3F,$C7,$74,$69,$8F,$AF
	.dc.b	$7D,$76,$FD,$AF,$F2,$3D,$6D,$B5,$FE,$86,$8F,$E8,$C6,$4F,$F2,$6B
	.dc.b	$FE,$8D,$C3,$FB,$5B,$74,$FE,$98,$D9,$FF,$2D,$3E,$2F,$2F,$A9,$23
	.dc.b	$9B,$5F,$E7,$70,$F6,$7D,$87,$B6,$7D,$DF,$71,$2B,$FC,$CD,$87,$8A
	.dc.b	$3E,$17,$89,$99,$FF,$38,$E0,$9F,$69,$23,$FC,$CF,$F6,$0F,$A0,$F8
	.dc.b	$6D,$8C,$C6,$0F,$8B,$E3,$48,$F8,$6D,$7F,$CF,$FF,$5F,$FE,$B7,$7E
	.dc.b	$4B,$5F,$EE,$0D,$7F,$CF,$CD,$4B,$D1,$FA,$AD,$8F,$F3,$EB,$5E,$F1
	.dc.b	$F6,$36,$91,$F5,$9A,$FF,$A4,$BE,$B5,$DD,$F8,$06,$C7,$FA,$21,$AE
	.dc.b	$E3,$F0,$BF,$5D,$FC,$FA,$3F,$D1,$AD,$7F,$D3,$AD,$7F,$0E,$D7,$FC
	.dc.b	$1D,$F5,$AF,$6F,$F1,$ED,$8F,$F4,$BB,$5F,$F5,$3B,$5F,$C8,$FD,$77
	.dc.b	$F8,$A9,$FF,$55,$B5,$CF,$85,$F9,$A8,$C4,$78,$32,$EE,$47,$BB,$FA
	.dc.b	$0C,$FC,$D1,$FE,$AF,$6B,$F4,$7C,$0D,$1F,$A6,$FA,$E7,$BB,$49,$3F
	.dc.b	$EB,$08,$CB,$FE,$B2,$FB,$87,$CF,$FE,$B9,$B1,$98,$C1,$FB,$2F,$AF
	.dc.b	$7C,$27,$FA,$F9,$BE,$C7,$FA,$F1,$78,$BF,$5A,$FF,$C7,$51,$A7,$FD
	.dc.b	$9F,$F7,$0E,$7F,$F2,$FC,$7F,$8F,$E3,$1B,$FF,$B1,$9B,$1C,$27,$6F
	.dc.b	$FB,$47,$EB,$DF,$43,$88,$46,$2F,$FD,$9E,$D8,$FF,$6F,$B5,$FE,$13
	.dc.b	$ED,$3E,$17,$93,$6B,$FD,$AE,$D9,$FF,$73,$B5,$FE,$23,$F5,$FF,$EE
	.dc.b	$39,$FD,$F9,$FE,$59,$71,$1F,$E4,$27,$F3,$B1,$A8,$FF,$75,$46,$BC
	.dc.b	$2F,$F4,$0F,$D7,$F7,$6E,$1F,$EA,$37,$3E,$55,$C3,$FC,$CD,$5B,$7F
	.dc.b	$CD,$1F,$5E,$F8,$3D,$65,$FE,$F7,$8C,$4F,$86,$11,$FB,$B8,$C3,$A8
	.dc.b	$FF,$2C,$FD,$7B,$EB,$F9,$C2,$3E,$AE,$5F,$E9,$87,$57,$D5,$CB,$FE
	.dc.b	$7B,$C1,$F5,$FC,$9F,$E3,$91,$E3,$72,$DF,$17,$BF,$D7,$BE,$22,$FC
	.dc.b	$5E,$1B,$5E,$BF,$35,$E9,$79,$9B,$19,$F8,$8C,$4F,$AB,$EB,$FF,$37
	.dc.b	$3F,$0D,$BB,$FF,$9E,$1F,$5B,$7B,$BE,$0F,$D7,$3E,$BE,$98,$F2,$2E
	.dc.b	$1F,$37,$EB,$FE,$E3,$FE,$9B,$7F,$03,$87,$FA,$11,$CA,$0E,$B9,$3F
	.dc.b	$D8,$7F,$5E,$F5,$BE,$8F,$D7,$BC,$5F,$A7,$39,$C7,$EE,$4E,$5F,$BD
	.dc.b	$F0,$BF,$BE,$6F,$C1,$FD,$A8,$C6,$7F,$E9,$65,$1F,$7B,$EB,$DF,$0D
	.dc.b	$7C,$5C,$37,$0F,$C5,$7B,$6F,$A1,$FC,$2A,$FA,$FB,$18,$7F,$13,$F5
	.dc.b	$EF,$8C,$BF,$19,$86,$E1,$F9,$5F,$AF,$FF,$3C,$B7,$D5,$FC,$8C,$71
	.dc.b	$8C,$1F,$9B,$FA,$E7,$D7,$FE,$59,$B1,$9E,$AC,$4F,$E6,$BE,$BF,$FA
	.dc.b	$04,$FE,$8D,$BC,$9F,$EB,$85,$F2,$3B,$F7,$7F,$43,$F5,$EB,$A7,$E9
	.dc.b	$3E,$BF,$E5,$F6,$FF,$BE,$A7,$53,$FA,$FF,$AF,$FC,$5C,$FF,$BF,$9B
	.dc.b	$1F,$ED,$FF,$AE,$F7,$E7,$FE,$C8,$9C,$BF,$85,$8F,$DB,$3F,$EC,$C6
	.dc.b	$D7,$8E,$35,$E6,$7E,$DA,$73,$7F,$EE,$BF,$AF,$FC,$7D,$1B,$3F,$C9
	.dc.b	$6F,$94,$E4,$E5,$D5,$2F,$F8,$45,$7E,$11,$B1,$DF,$EA,$3F,$82,$FC
	.dc.b	$83,$CD,$C5,$7F,$0F,$8C,$3C,$BE,$1B,$38,$FA,$DB,$97,$88,$7C,$EF
	.dc.b	$4F,$3B,$F8,$8C,$62,$FA,$46,$3C,$EC,$D5,$9F,$E3,$D5,$BC,$3F,$DE
	.dc.b	$63,$E5,$7C,$FE,$6F,$F1,$C8,$C5,$79,$E3,$1F,$5F,$9A,$B3,$FC,$A6
	.dc.b	$B6,$74,$FA,$9C,$B6,$35,$5D,$77,$F8,$BF,$05,$67,$E2,$3E,$1D,$B1
	.dc.b	$7E,$66,$E7,$FD,$83,$EB,$FF,$CC,$72,$69,$1F,$F3,$18,$FA,$CF,$DA
	.dc.b	$7D,$7E,$23,$1E,$3A,$31,$3A,$9E,$2D,$7F,$CD,$AB,$E5,$D8,$70,$6C
	.dc.b	$F0,$6B,$C1,$AF,$F1,$D3,$C9,$F2,$F6,$7F,$1F,$48,$C6,$9B,$8C,$57
	.dc.b	$F5,$77,$C5,$5F,$C9,$FE,$BD,$F3,$9B,$E4,$E3,$7F,$CA,$EF,$F1,$2F
	.dc.b	$18,$CB,$E2,$3A,$72,$E2,$4D,$37,$F5,$FF,$E8,$EC,$7C,$FE,$D6,$C9
	.dc.b	$7C,$8F,$37,$C1,$F3,$23,$E9,$8F,$0E,$4E,$6A,$49,$FE,$E9,$F5,$CF
	.dc.b	$AF,$DF,$3C,$9F,$EF,$16,$8E,$11,$F5,$38,$3E,$77,$D7,$7C,$88,$C1
	.dc.b	$FD,$35,$C6,$9F,$BE,$6C,$FD,$5C,$5F,$5F,$F1,$A7,$E4,$DF,$20,$E2
	.dc.b	$F9,$70,$EF,$8F,$E6,$FD,$7F,$FA,$F8,$FF,$57,$59,$A6,$7C,$0D,$92
	.dc.b	$3F,$C2,$91,$FE,$C6,$FA,$FF,$F0,$04,$63,$E3,$3A,$9F,$6A,$33,$1E
	.dc.b	$59,$4F,$8B,$C1,$93,$C4,$EA,$F9,$BD,$69,$18,$3F,$C4,$9F,$5D,$FE
	.dc.b	$7B,$D3,$F9,$1A,$FA,$4F,$C4,$B6,$77,$F0,$5F,$8C,$FC,$47,$FC,$A7
	.dc.b	$1F,$E2,$DD,$3F,$D7,$8D,$F9,$FD,$79,$3F,$1B,$97,$F2,$7F,$83,$3E
	.dc.b	$C2,$99,$BF,$1F,$AB,$E0,$F5,$A4,$60,$FF,$1D,$67,$FE,$3F,$7C,$1F
	.dc.b	$49,$F1,$4F,$27,$A3,$FA,$0F,$8F,$FF,$0C,$4F,$F9,$05,$FE,$CE,$DC
	.dc.b	$35,$87,$BB,$D0,$F7,$FD,$B4,$AF,$8B,$1F,$AA,$FC,$45,$F5,$3F,$4F
	.dc.b	$ED,$E5,$7F,$93,$67,$2F,$F9,$3F,$F0,$7F,$F0,$5B,$7E,$1F,$F6,$71
	.dc.b	$FE,$10,$8D,$7C,$37,$CE,$FB,$CF,$9F,$F5,$92,$7E,$AC,$7E,$EB,$B8
	.dc.b	$FE,$0F,$EB,$E4,$FF,$90,$E3,$F7,$FF,$83,$FE,$F2,$3F,$CC,$0F,$F9
	.dc.b	$19,$BE,$6E,$BD,$0F,$BB,$EA,$FD,$EF,$8A,$95,$F8,$48,$FE,$13,$F8
	.dc.b	$8F,$ED,$63,$EF,$7C,$FF,$E4,$CC,$FA,$3F,$C3,$BF,$08,$7C,$7D,$FB
	.dc.b	$29,$7F,$E7,$19,$CB,$F9,$0F,$3B,$F1,$BE,$D4,$8C,$1F,$E7,$7F,$C4
	.dc.b	$7F,$C5,$0F,$E3,$7B,$FF,$44,$7F,$1C,$8C,$C7,$ED,$FF,$09,$FF,$1A
	.dc.b	$3F,$E7,$F7,$F9,$D3,$6B,$F9,$1C,$6A,$3F,$7D,$F7,$17,$CE,$FF,$01
	.dc.b	$B7,$C8,$FE,$74,$D9,$FF,$2C,$3F,$E3,$F7,$FD,$18,$E5,$3E,$07,$AD
	.dc.b	$FF,$32,$8D,$19,$8F,$E8,$C7,$08,$FE,$B0,$D3,$FA,$D4,$6D,$F1,$F2
	.dc.b	$6A,$3E,$E4,$7F,$62,$6B,$D1,$AF,$F5,$3F,$BA,$7A,$1F,$D1,$1B,$1F
	.dc.b	$EA,$56,$BF,$12,$BF,$A4,$FD,$C7,$F9,$CE,$3F,$D5,$4D,$7F,$93,$37
	.dc.b	$2F,$93,$F3,$1C,$3E,$67,$D7,$3B,$AF,$7B,$5F,$EC,$CD,$7F,$D0,$8F
	.dc.b	$F1,$C7,$FB,$D3,$5F,$EF,$ED,$79,$FA,$9D,$8F,$3F,$74,$8F,$3C,$57
	.dc.b	$8C,$79,$08,$F6,$FB,$5F,$39,$1C,$1F,$E0,$BF,$C0,$7F,$62,$FF,$B2
	.dc.b	$DA,$FF,$47,$6E,$5F,$E4,$CC,$15,$F2,$BA,$65,$FC,$A7,$E6,$F8,$18
	.dc.b	$8F,$6C,$7C,$82,$3F,$0B,$E4,$7E,$59,$1C,$10,$FA,$7F,$23,$F6,$0F
	.dc.b	$B3,$16,$3E,$CC,$57,$EC,$C7,$F8,$1B,$EE,$3F,$57,$1F,$94,$6B,$FE
	.dc.b	$1B,$6B,$FE,$B9,$6A,$7F,$67,$97,$F9,$B3,$EE,$FF,$91,$CF,$F6,$B6
	.dc.b	$C7,$F9,$C2,$77,$CD,$D7,$B3,$C4,$4B,$FF,$7C,$4F,$F0,$87,$FB,$A2
	.dc.b	$8F,$D3,$B6,$3F,$C8,$0D,$7F,$C8,$AD,$7D,$9E,$47,$BF,$EB,$64,$76
	.dc.b	$8F,$84,$61,$EF,$18,$7C,$6E,$70,$F5,$39,$8F,$2F,$C7,$7F,$8C,$3F
	.dc.b	$01,$7A,$94,$F9,$D1,$27,$FD,$C5,$7F,$E9,$2A,$DC,$63,$3F,$F2,$54
	.dc.b	$62,$3A,$9F,$10,$FB,$04,$3E,$3F,$C6,$D7,$E3,$96,$BE,$84,$7F,$91
	.dc.b	$3E,$3B,$FB,$2C,$FF,$87,$DB,$3F,$EA,$46,$DF,$F9,$6D,$FF,$03,$37
	.dc.b	$D4,$E4,$4A,$F4,$1F,$EF,$D9,$DB,$63,$4F,$77,$C5,$EE,$E2,$FC,$11
	.dc.b	$6F,$C1,$36,$B9,$1D,$85,$F9,$9F,$04,$6D,$CD,$59,$F1,$C6,$0E,$4F
	.dc.b	$17,$07,$C3,$9E,$13,$C5,$B1,$C2,$3D,$2D,$21,$ED,$3A,$C7,$07,$07
	.dc.b	$53,$77,$F0,$1B,$7F,$01,$A7,$C0,$38,$C7,$D5,$9E,$8F,$18,$E8,$6A
	.dc.b	$7E,$B3,$C8,$F8,$F3,$C2,$33,$1C,$27,$91,$23,$2F,$78,$E5,$7F,$1C
	.dc.b	$95,$CE,$3C,$D3,$B7,$C4,$62,$B9,$9E,$D7,$27,$D0,$8F,$03,$F6,$8E
	.dc.b	$F7,$F4,$5A,$78,$AB,$C0,$F1,$AE,$95,$F0,$E3,$FB,$6D,$FE,$79,$B7
	.dc.b	$FD,$BA,$31,$1F,$08,$C9,$FE,$C9,$72,$FD,$13,$27,$FA,$17,$EF,$7F
	.dc.b	$CB,$ED,$F7,$7C,$71,$8B,$61,$FE,$C3,$8C,$3F,$0A,$BE,$20,$11,$A7
	.dc.b	$0D,$7D,$C7,$03,$75,$FE,$BE,$8D,$5F,$D5,$C3,$13,$2B,$E4,$9E,$73
	.dc.b	$F5,$55,$C2,$7F,$50,$F7,$31,$1F,$89,$71,$3E,$E3,$A4,$7D,$AA,$FD
	.dc.b	$54,$EA,$3B,$E1,$F0,$9F,$5D,$70,$8F,$BF,$1E,$C3,$D6,$79,$23,$B3
	.dc.b	$4E,$78,$7C,$B8,$E4,$FC,$B8,$E5,$5E,$17,$C9,$86,$9F,$09,$C8,$A7
	.dc.b	$C3,$36,$77,$31,$7E,$6C,$3D,$B8,$73,$9F,$44,$61,$E1,$3B,$AD,$5F
	.dc.b	$CA,$9F,$CF,$BE,$59,$D4,$70,$C3,$D8,$F9,$4F,$86,$78,$9F,$75,$BD
	.dc.b	$0F,$F7,$EF,$CD,$7F,$A3,$AB,$FD,$B0,$DB,$FE,$75,$3F,$CF,$CD,$FA
	.dc.b	$52,$BF,$90,$C7,$E4,$5F,$F7,$07,$E0,$2F,$67,$17,$F6,$65,$39,$7F
	.dc.b	$99,$35,$FE,$95,$E5,$F0,$CF,$A4,$E6,$B9,$36,$33,$7F,$D4,$EB,$FA
	.dc.b	$A7,$D6,$4C,$3F,$9A,$BF,$D5,$FE,$37,$11,$FE,$68,$3F,$A0,$C7,$EE
	.dc.b	$C8,$E4,$F3,$FA,$9E,$37,$F5,$D8,$FF,$A9,$5F,$7C,$74,$CB,$DD,$F7
	.dc.b	$6F,$87,$FF,$39,$B6,$BA,$35,$F1,$35,$FF,$3D,$B9,$9F,$91,$F7,$8F
	.dc.b	$73,$C5,$48,$C3,$FD,$6E,$B8,$6D,$F4,$FE,$F3,$FD,$DB,$2C,$3E,$94
	.dc.b	$61,$BF,$0B,$D2,$FF,$CF,$31,$8A,$F8,$55,$C9,$FF,$47,$B5,$ED,$5E
	.dc.b	$EE,$C9,$1E,$E6,$BF,$3F,$F0,$17,$57,$FA,$59,$B1,$F3,$DA,$EE,$3F
	.dc.b	$D3,$7F,$7E,$F8,$EA,$49,$FA,$D1,$97,$EC,$7D,$F3,$E2,$3F,$D4,$6D
	.dc.b	$8C,$C6,$0C,$FD,$24,$8E,$2D,$7E,$AE,$FF,$68,$8F,$DA,$7F,$D5,$FF
	.dc.b	$80,$BD,$3A,$49,$FC,$54,$65,$FC,$67,$DF,$3B,$BF,$D6,$ED,$8C,$C6
	.dc.b	$0F,$F5,$DF,$DF,$FF,$AF,$20,$D2,$3F,$EC,$0F,$BE,$FF,$77,$3F,$9D
	.dc.b	$6B,$FE,$C6,$6B,$FB,$76,$BF,$D2,$7E,$FF,$F6,$CF,$FB,$31,$AF,$EF
	.dc.b	$5A,$EF,$2F,$F6,$7F,$DF,$FC,$D8,$FD,$63,$5F,$F6,$AB,$5F,$E0,$CD
	.dc.b	$7F,$AC,$7D,$F3,$EF,$FF,$DB,$2D,$8F,$D9,$35,$FD,$C3,$5F,$F7,$0B
	.dc.b	$F3,$5C,$EB,$C8,$8F,$EE,$FE,$F9,$F4,$9B,$8F,$E3,$11,$89,$D7,$8E
	.dc.b	$F2,$F8,$5E,$EF,$FB,$77,$E0,$2F,$77,$E1,$B7,$5F,$E1,$0D,$9F,$E1
	.dc.b	$CD,$7F,$DE,$FE,$EB,$C7,$F9,$6D,$AF,$E2,$0D,$9F,$E2,$CD,$7F,$DF
	.dc.b	$FF,$7A,$F9,$EF,$EC,$9E,$17,$F2,$D8,$C5,$FF,$97,$57,$D0,$6A,$3F
	.dc.b	$9B,$46,$AF,$F8,$24,$7F,$39,$FB,$B7,$A5,$B9,$FF,$06,$46,$27,$FC
	.dc.b	$D1,$F7,$AF,$83,$D5,$2B,$4F,$43,$FC,$29,$F8,$73,$DE,$DB,$70,$FF
	.dc.b	$39,$3F,$E1,$E6,$C7,$4A,$7C,$05,$31,$E6,$DA,$FF,$3D,$7E,$1D,$FB
	.dc.b	$16,$F4,$79,$3E,$69,$EE,$79,$AB,$5A,$79,$47,$E1,$5F,$C5,$30,$FF
	.dc.b	$42,$36,$34,$BD,$BF,$7B,$9E,$4F,$EB,$9F,$0E,$FA,$2F,$86,$DC,$3F
	.dc.b	$D1,$CB,$D9,$CC,$FC,$8F,$B9,$7D,$FD,$3E,$5F,$DE,$3E,$23,$18,$FF
	.dc.b	$A5,$FE,$B5,$F2,$7F,$41,$B1,$9F,$0F,$13,$D7,$BE,$4F,$FA,$7F,$2F
	.dc.b	$C1,$B6,$3E,$CF,$E1,$CF,$52,$9F,$73,$EE,$FF,$69,$5F,$EA,$79,$D3
	.dc.b	$FE,$A9,$FC,$2B,$FC,$A0,$DE,$5F,$C1,$AF,$99,$FC,$2C,$EA,$7E,$87
	.dc.b	$DE,$3E,$8B,$F1,$2D,$BF,$F5,$A3,$FE,$F5,$6C,$7E,$3B,$E1,$BD,$F5
	.dc.b	$F6,$5B,$CF,$FE,$0D,$F9,$57,$D0,$6F,$87,$F2,$BF,$85,$3B,$E9,$8F
	.dc.b	$9D,$BA,$7F,$B0,$FE,$ED,$EE,$FE,$71,$B1,$9F,$7B,$13,$F2,$3F,$08
	.dc.b	$93,$FE,$CB,$FB,$CF,$C2,$3F,$E2,$3E,$54,$FE,$9B,$F0,$A7,$D2,$53
	.dc.b	$1F,$49,$6F,$FD,$A9,$F7,$7F,$A3,$E5,$FC,$B3,$79,$3F,$C5,$5F,$79
	.dc.b	$F9,$CA,$D9,$FB,$2F,$BC,$F8,$6F,$F8,$CD,$B7,$FE,$85,$FB,$BF,$EA
	.dc.b	$13,$FE,$E1,$9C,$CA,$5F,$03,$FB,$C6,$D7,$FB,$9D,$F9,$4D,$8F,$E0
	.dc.b	$3F,$79,$F7,$71,$FF,$1F,$36,$7F,$C8,$94,$F8,$ED,$E1,$D5,$93,$D7
	.dc.b	$E5,$EC,$FE,$D6,$BF,$8B,$61,$FE,$49,$FA,$DF,$F1,$65,$F8,$FC,$4F
	.dc.b	$E2,$31,$8B,$E1,$F0,$3E,$99,$F5,$36,$BF,$E0,$05,$FE,$D6,$BF,$81
	.dc.b	$5B,$CE,$FE,$37,$39,$3C,$0E,$3E,$57,$3E,$75,$C3,$9B,$FC,$8B,$07
	.dc.b	$CD,$F9,$3C,$BC,$36,$BF,$D6,$D8,$FF,$6E,$FB,$CC,$C6,$0F,$F3,$55
	.dc.b	$A3,$2F,$F1,$73,$F7,$25,$7C,$1F,$01,$E1,$7C,$5B,$7C,$4F,$85,$B7
	.dc.b	$D6,$FF,$D8,$DB,$27,$D1,$BE,$A7,$B3,$72,$FF,$00,$F9,$FF,$E1,$12
	.dc.b	$4F,$F9,$94,$3F,$E6,$7F,$BA,$75,$3E,$F3,$0D,$AF,$F3,$84,$EA,$72
	.dc.b	$71,$FA,$9F,$F2,$B4,$E9,$AF,$C4,$94,$62,$3F,$D2,$11,$8C,$3C,$5E
	.dc.b	$B7,$F9,$9A,$3F,$98,$4F,$C1,$0E,$1F,$E8,$4F,$15,$F7,$8F,$BB,$EB
	.dc.b	$CB,$CC,$93,$CE,$B9,$61,$FE,$9D,$7F,$D0,$8D,$E6,$F1,$1F,$E7,$8F
	.dc.b	$BC,$FB,$B1,$BF,$84,$F3,$12,$FA,$57,$A7,$E2,$64,$77,$74,$FF,$59
	.dc.b	$CB,$FB,$EB,$ED,$6B,$9E,$7F,$F4,$3F,$DE,$7F,$91,$31,$FF,$59,$57
	.dc.b	$F8,$26,$B8,$73,$FF,$A3,$23,$11,$C4,$F9,$1F,$79,$EE,$E8,$FF,$5F
	.dc.b	$36,$01,$C2,$C3,$FC,$E8,$FF,$6F,$8F,$EC,$CF,$4D,$FF,$D3,$71,$88
	.dc.b	$D7,$0F,$FA,$7B,$EF,$3F,$C5,$D1,$FE,$B3,$6B,$FE,$B9,$6B,$FA,$86
	.dc.b	$BF,$ED,$37,$FD,$46,$F9,$9A,$9A,$74,$CF,$FD,$AD,$F7,$4F,$C0,$BF
	.dc.b	$3C,$DA,$FF,$62,$35,$FF,$65,$35,$F1,$7C,$77,$E9,$6C,$D2,$47,$AF
	.dc.b	$C0,$D1,$1C,$1F,$ED,$1F,$BC,$FF,$51,$BF,$C1,$1A,$FF,$BB,$DA,$FD
	.dc.b	$26,$AF,$BB,$FF,$5B,$3F,$59,$AE,$7C,$5F,$DD,$46,$30,$FF,$3E,$CF
	.dc.b	$C5,$E0,$F5,$11,$F2,$7D,$DF,$F0,$1C,$0D,$24,$7F,$BA,$6F,$FB,$83
	.dc.b	$7D,$5E,$E4,$7F,$DD,$DE,$BF,$F8,$B5,$BF,$0D,$FC,$36,$31,$1F,$EF
	.dc.b	$76,$BF,$0F,$C4,$F1,$11,$F9,$0F,$F1,$5F,$BB,$FF,$07,$47,$FB,$ED
	.dc.b	$AF,$FC,$01,$6B,$FC,$C1,$AF,$E7,$07,$DD,$BC,$BF,$E4,$0D,$8F,$E6
	.dc.b	$E7,$E8,$9C,$38,$6B,$F6,$7D,$BF,$90,$47,$EF,$3F,$CA,$F8,$7F,$C3
	.dc.b	$31,$D0,$FE,$A9,$19,$23,$9F,$BB,$77,$FF,$33,$6C,$66,$30,$7D,$7F
	.dc.b	$1A,$47,$E3,$5A,$FA,$BE,$EF,$FD,$86,$9A,$4F,$83,$FE,$81,$F7,$6F
	.dc.b	$7A,$92,$BD,$11,$97,$F3,$5D,$BF,$87,$23,$83,$DD,$F7,$6F,$BF,$FE
	.dc.b	$98,$D8,$CC,$60,$FE,$A1,$F7,$9F,$D3,$A3,$E3,$35,$FA,$6D,$7F,$B5
	.dc.b	$B5,$FB,$9F,$79,$BF,$05,$D9,$8D,$BF,$01,$18,$AF,$E3,$B5,$BA,$E3
	.dc.b	$ED,$E9,$E1,$3F,$5E,$7B,$3C,$2F,$FB,$34,$62,$0F,$37,$AF,$9D,$CD
	.dc.b	$A7,$F5,$78,$D3,$C3,$87,$4F,$2F,$2E,$CF,$FA,$CE,$38,$4E,$27,$C2
	.dc.b	$48,$F3,$3E,$4F,$92,$91,$E9,$6B,$F9,$CF,$BA,$FE,$49,$97,$E8,$DB
	.dc.b	$1F,$D0,$1A,$FB,$27,$CB,$1F,$C8,$23,$FB,$83,$5F,$F5,$83,$74,$F9
	.dc.b	$D0,$8F,$F4,$93,$C3,$C4,$FF,$07,$47,$18,$FF,$12,$B5,$F2,$46,$8F
	.dc.b	$0F,$44,$8E,$47,$8B,$F0,$9F,$2F,$DD,$C8,$C1,$E6,$7B,$91,$FD,$95
	.dc.b	$A3,$C7,$F2,$F2,$72,$D7,$E4,$7C,$8F,$C2,$23,$F0,$A3,$E2,$FD,$DC
	.dc.b	$8F,$8A,$FC,$8F,$67,$11,$F4,$63,$E0,$91,$F1,$BF,$3B,$E6,$A4,$72
	.dc.b	$3F,$C3,$DF,$73,$FE,$9A,$6F,$47,$E2,$23,$F5,$91,$FE,$E1,$74,$D8
	.dc.b	$D1,$1E,$7F,$13,$C7,$B7,$97,$DD,$49,$EF,$15,$E9,$17,$8B,$EE,$E1
	.dc.b	$FC,$0E,$7C,$1F,$75,$E3,$23,$E8,$F1,$6B,$FC,$B2,$53,$FE,$4F,$6C
	.dc.b	$7F,$25,$6B,$B8,$F7,$7C,$DC,$8F,$87,$1F,$07,$E0,$64,$7C,$67,$E2
	.dc.b	$7A,$F8,$8F,$89,$17,$1F,$4C,$FC,$9F,$80,$F9,$55,$0F,$9E,$E5,$FA
	.dc.b	$26,$4F,$9C,$FE,$E3,$85,$F5,$F4,$93,$FE,$55,$87,$DF,$E2,$18,$7E
	.dc.b	$9A,$78,$C6,$EB,$A5,$6E,$B5,$17,$B3,$FA,$D5,$F4,$BE,$04,$3A,$C6
	.dc.b	$A7,$9C,$66,$B9,$B6,$79,$DF,$58,$FF,$20,$DF,$91,$E6,$62,$F9,$B8
	.dc.b	$3F,$D2,$0F,$16,$9F,$06,$71,$3C,$EF,$C3,$3C,$63,$D4,$DA,$D9,$DA
	.dc.b	$76,$4B,$F6,$4F,$53,$D4,$79,$4A,$7F,$A8,$EF,$9B,$8B,$EF,$5C,$E3
	.dc.b	$81,$E1,$3A,$14,$8E,$DE,$87,$3B,$F6,$D7,$68,$DD,$FC,$C7,$47,$86
	.dc.b	$35,$5C,$EF,$C7,$5E,$CA,$E2,$79,$27,$F0,$D5,$C2,$75,$18,$9E,$A7
	.dc.b	$03,$CA,$5A,$FA,$13,$C0,$F2,$46,$6F,$D2,$F2,$AF,$25,$FE,$5E,$3F
	.dc.b	$DB,$37,$FE,$52,$25,$7F,$B1,$DB,$1F,$E1,$D6,$BF,$E2,$A7,$C2,$FE
	.dc.b	$33,$CF,$FD,$41,$EB,$70,$7E,$9D,$C3,$FE,$DB,$76,$F9,$70,$FD,$0C
	.dc.b	$62,$3D,$47,$12,$18,$3F,$1C,$EF,$0F,$A1,$76,$B9,$D7,$53,$C6,$70
	.dc.b	$30,$78,$CE,$B3,$C9,$B1,$C9,$F0,$47,$EF,$6B,$B4,$7D,$B9,$FA,$4F
	.dc.b	$69,$D3,$DA,$FF,$CC,$B1,$88,$E7,$3D,$63,$07,$86,$3C,$C4,$70,$6A
	.dc.b	$38,$1E,$08,$FD,$A9,$B7,$2F,$94,$87,$BC,$D1,$89,$FB,$6E,$67,$B3
	.dc.b	$9B,$F7,$B9,$AF,$05,$70,$3D,$AF,$82,$3C,$75,$FA,$3B,$E3,$5B,$BB
	.dc.b	$1F,$84,$8E,$35,$F2,$6F,$AF,$9D,$FC,$CF,$EE,$64,$FF,$31,$8F,$E6
	.dc.b	$C6,$1F,$E2,$11,$87,$F9,$B7,$87,$FC,$EF,$C8,$91,$83,$F9,$D3,$87
	.dc.b	$F8,$C3,$FD,$06,$3F,$A1,$7B,$D2,$7F,$A0,$3F,$D1,$7C,$3C,$47,$E8
	.dc.b	$E3,$F5,$04,$7F,$50,$FF,$4B,$FB,$99,$1C,$8F,$F7,$9E,$DF,$E2,$CF
	.dc.b	$B6,$FF,$71,$C7,$FB,$D9,$A9,$A4,$FB,$8F,$F2,$6B,$FE,$E6,$27,$D4
	.dc.b	$E3,$C5,$CF,$CB,$CB,$EF,$C1,$FE,$A7,$B5,$66,$7F,$D4,$B0,$AF,$F4
	.dc.b	$E4,$78,$5B,$19,$8F,$F5,$02,$FA,$7F,$19,$23,$B9,$EB,$F1,$70,$F6
	.dc.b	$8E,$64,$7F,$D5,$5F,$6B,$FE,$E4,$27,$37,$89,$FF,$73,$DB,$C6,$F1
	.dc.b	$92,$BC,$6F,$A3,$ED,$1F,$31,$49,$3F,$EB,$78,$CB,$7E,$47,$81,$1C
	.dc.b	$1E,$ED,$BE,$09,$1C,$1F,$EC,$1F,$B1,$76,$FC,$86,$C6,$63,$07,$CC
	.dc.b	$FB,$37,$D9,$D2,$4F,$FB,$2E,$32,$F5,$3A,$10,$C1,$FE,$D1,$FB,$3F
	.dc.b	$EF,$8F,$FB,$8C,$38,$7F,$DC,$CD,$46,$B7,$FB,$FF,$67,$FE,$7E,$8E
	.dc.b	$4D,$7F,$DD,$CD,$78,$B5,$67,$E8,$7F,$DD,$ED,$7C,$8D,$7F,$DF,$0D
	.dc.b	$7C,$7F,$64,$FB,$3F,$C8,$B6,$3F,$DD,$CD,$7F,$DF,$6D,$7E,$07,$E1
	.dc.b	$2F,$23,$F3,$2D,$8F,$F7,$B3,$5F,$8C,$1C,$3F,$27,$ED,$1F,$30,$F5
	.dc.b	$FF,$DF,$D1,$86,$F8,$DF,$7A,$B8,$56,$DE,$BF,$68,$EF,$F9,$0D,$E5
	.dc.b	$F0,$CA,$30,$F0,$9F,$9D,$F6,$8F,$93,$A4,$BE,$31,$97,$9F,$D9,$7E
	.dc.b	$CD,$BD,$DF,$5E,$31,$87,$C0,$6B,$F8,$7F,$B3,$7A,$FF,$75,$B7,$E6
	.dc.b	$6C,$FC,$86,$BF,$8F,$A7,$B3,$DC,$95,$E4,$7F,$11,$F6,$5F,$2E,$3E
	.dc.b	$0B,$6B,$E9,$B5,$FC,$D3,$5F,$CB,$8F,$B2,$FF,$3A,$CF,$CA,$6C,$7D
	.dc.b	$C6,$BF,$A3,$6B,$CB,$C6,$7C,$B7,$E6,$DB,$5F,$51,$B3,$F8,$66,$BF
	.dc.b	$AD,$FB,$07,$D9,$7E,$8D,$B5,$F7,$1B,$3F,$8E,$6B,$FB,$5F,$9F,$7C
	.dc.b	$87,$EA,$DB,$5F,$84,$6C,$FE,$59,$AF,$87,$8B,$C0,$49,$F0,$3F,$B5
	.dc.b	$FA,$87,$D8,$D2,$4F,$E4,$23,$2F,$E4,$BE,$C3,$F8,$1B,$6F,$F7,$D1
	.dc.b	$87,$F4,$ED,$7F,$87,$7D,$83,$EC,$3F,$82,$36,$BF,$3A,$D9,$FD,$7B
	.dc.b	$5F,$E2,$DF,$61,$EC,$7E,$EB,$5F,$E2,$AD,$4D,$FB,$07,$CC,$7F,$13
	.dc.b	$6D,$FE,$B1,$B5,$FB,$E6,$BF,$C9,$7E,$C1,$F0,$FF,$C6,$DB,$5F,$B4
	.dc.b	$6C,$FF,$06,$6B,$FC,$B7,$47,$D7,$FC,$F2,$57,$CD,$7F,$93,$7D,$7B
	.dc.b	$E4,$A9,$27,$F8,$0C,$65,$FE,$07,$F5,$E7,$F9,$73,$5C,$98,$3F,$99
	.dc.b	$7D,$77,$C5,$27,$2E,$0F,$E6,$DF,$5B,$FC,$D2,$7F,$87,$B7,$3F,$E4
	.dc.b	$2D,$7F,$A4,$35,$FD,$7F,$D7,$7D,$F6,$9F,$D0,$5C,$3F,$D2,$9A,$A1
	.dc.b	$F4,$9E,$22,$57,$89,$F2,$F3,$FB,$EB,$FD,$A1,$F0,$DF,$C8,$BE,$03
	.dc.b	$11,$CE,$75,$18,$9D,$5F,$F0,$4C,$3F,$B8,$BF,$E1,$0B,$E0,$D8,$CC
	.dc.b	$7F,$B7,$5F,$E5,$0F,$FB,$7E,$BF,$7E,$75,$ED,$EF,$5F,$C4,$23,$2F
	.dc.b	$F1,$2B,$FF,$14,$BD,$E1,$F2,$BC,$2F,$FC,$7B,$DF,$54,$EA,$B8,$4E
	.dc.b	$65,$F1,$7B,$C3,$FA,$8D,$7A,$E7,$2F,$B6,$33,$1F,$EF,$71,$F5,$DC
	.dc.b	$EB,$A3,$6B,$97,$D5,$61,$A7,$58,$F1,$6B,$E1,$F3,$AC,$F7,$7C,$DE
	.dc.b	$DE,$31,$F0,$37,$4F,$F0,$37,$85,$FE,$48,$93,$48,$FF,$70,$FA,$C7
	.dc.b	$CD,$3E,$BB,$FB,$1C,$61,$B1,$99,$E0,$D7,$FC,$3D,$2F,$8B,$F9,$B1
	.dc.b	$D5,$AE,$9A,$FF,$4A,$8E,$18,$78,$3E,$33,$EF,$69,$96,$D7,$F8,$9A
	.dc.b	$91,$88,$47,$25,$F5,$CF,$91,$ED,$C9,$BA,$FE,$9F,$5F,$95,$6B,$F8
	.dc.b	$4C,$3C,$0E,$79,$78,$11,$FC,$E7,$D7,$7F,$8F,$73,$FD,$2D,$BE,$3F
	.dc.b	$B1,$2B,$91,$DF,$E3,$FC,$44,$7C,$51,$EB,$CD,$CD,$49,$3F,$AB,$FA
	.dc.b	$E7,$D1,$6F,$9E,$4F,$EB,$EF,$FC,$2C,$DB,$F8,$DF,$5D,$F2,$23,$F6
	.dc.b	$8D,$7E,$01,$FE,$52,$71,$A7,$F0,$0C,$5F,$6D,$CB,$86,$5F,$27,$AF
	.dc.b	$1F,$D3,$FA,$D7,$C9,$D2,$57,$F0,$3F,$DC,$3F,$5A,$FF,$05,$9D,$B7
	.dc.b	$DE,$E4,$70,$7C,$A3,$EB,$DF,$61,$F8,$16,$C7,$F0,$F6,$BC,$2F,$F6
	.dc.b	$6F,$7F,$A4,$F3,$11,$E1,$5E,$9F,$12,$47,$95,$D3,$E2,$FB,$0F,$F7
	.dc.b	$A9,$3E,$37,$F2,$58,$7D,$89,$FE,$41,$F5,$FF,$CC,$1B,$E4,$FE,$5E
	.dc.b	$31,$1E,$CC,$FF,$C0,$AF,$C3,$EE,$F9,$44,$7E,$37,$C1,$7E,$7F,$27
	.dc.b	$D8,$52,$4F,$CD,$EA,$ED,$EB,$C8,$F1,$BF,$1F,$83,$F5,$1E,$77,$FB
	.dc.b	$FA,$7F,$C4,$3F,$3B,$83,$3C,$1D,$FC,$D9,$3F,$CE,$F3,$47,$C0,$F2
	.dc.b	$BF,$65,$F6,$1F,$E1,$1A,$FE,$80,$D7,$EC,$B7,$B7,$FA,$5F,$0F,$FB
	.dc.b	$82,$7F,$A3,$7D,$83,$E8,$FF,$78,$D8,$CC,$63,$97,$E1,$7D,$84,$9F
	.dc.b	$23,$F4,$9F,$E0,$9D,$08,$A4,$8F,$C3,$7C,$FF,$F0,$9F,$B0,$FE,$8F
	.dc.b	$1F,$D5,$DA,$FE,$41,$B9,$FF,$62,$7F,$83,$BF,$D9,$37,$FE,$53,$3F
	.dc.b	$17,$CE,$FA,$C7,$D7,$F1,$FF,$8B,$7D,$87,$EE,$27,$FB,$23,$5F,$CE
	.dc.b	$B6,$FE,$2F,$B7,$FC,$77,$EC,$1F,$53,$49,$7F,$DB,$63,$FC,$62,$FD
	.dc.b	$5F,$17,$F0,$C4,$7E,$C7,$8F,$FC,$9F,$EC,$1F,$57,$49,$3F,$A9,$FA
	.dc.b	$EF,$E5,$9F,$61,$EC,$8F,$D8,$B7,$4F,$F0,$1B,$7C,$4F,$F4,$1B,$FE
	.dc.b	$2D,$9F,$F0,$07,$D8,$57,$FD,$7A,$D7,$FD,$C8,$D7,$FD,$16,$D7,$FD
	.dc.b	$C6,$F5,$6F,$0F,$F3,$F8,$C4,$1E,$D7,$FE,$16,$72,$49,$E1,$F0,$5C
	.dc.b	$7D,$69,$7E,$67,$F5,$9E,$76,$23,$CF,$1F,$EC,$47,$B3,$6F,$F0,$B1
	.dc.b	$FE,$FD,$FC,$03,$FD,$6E,$FF,$B2,$9A,$FF,$B7,$1A,$EF,$C9,$E9,$1A
	.dc.b	$CB,$D1,$F3,$1E,$D3,$FA,$E4,$67,$5F,$F5,$7E,$FF,$7B,$0F,$E4,$7F
	.dc.b	$63,$FC,$21,$BF,$1D,$FD,$8E,$75,$8F,$F8,$E2,$B5,$5F,$EB,$80,$E1
	.dc.b	$FF,$5D,$B5,$1A,$AF,$85,$0F,$27,$FB,$83,$67,$FC,$7C,$D7,$C6,$63
	.dc.b	$8B,$64,$7F,$C8,$98,$7F,$27,$7D,$71,$D8,$8F,$FB,$D2,$51,$F4,$FC
	.dc.b	$3F,$F7,$D4,$FF,$BE,$1A,$FB,$4F,$E8,$DE,$86,$C8,$F9,$BE,$4F,$D2
	.dc.b	$7F,$49,$8F,$E9,$5F,$55,$EC,$F8,$19,$5F,$D2,$E2,$BD,$71,$4F,$C4
	.dc.b	$24,$79,$DF,$8B,$F6,$32,$3E,$2C,$7F,$85,$FE,$C3,$FD,$3F,$1F,$E6
	.dc.b	$56,$BF,$D7,$9B,$BF,$F9,$C3,$EF,$DF,$65,$C7,$E3,$E5,$F1,$7F,$C4
	.dc.b	$53,$A8,$FF,$38,$3F,$E1,$5E,$3F,$F7,$BE,$1F,$C2,$A7,$C1,$E0,$F0
	.dc.b	$61,$82,$1F,$DC,$2B,$93,$F8,$49,$FE,$68,$F9,$FC,$BF,$61,$FE,$33
	.dc.b	$F6,$3F,$D8,$B1,$ED,$8F,$D5,$4F,$08,$FC,$93,$5F,$2C,$68,$F8,$DF
	.dc.b	$65,$23,$91,$F1,$FD,$FF,$94,$47,$FD,$1D,$F7,$EF,$9D,$C4,$7C,$B9
	.dc.b	$BB,$FF,$A6,$AB,$53,$FE,$3A,$FA,$AF,$A9,$EA,$48,$C1,$0F,$A8,$E5
	.dc.b	$FF,$09,$FC,$32,$7F,$CA,$5C,$6F,$B3,$A4,$AF,$F1,$C3,$5F,$F0,$DC
	.dc.b	$6D,$FF,$53,$35,$FF,$0F,$FC,$15,$CB,$FD,$63,$37,$8F,$F6,$B5,$F9
	.dc.b	$2B,$E9,$5F,$E3,$76,$CE,$AB,$85,$7E,$F1,$AA,$EC,$4A,$E3,$3F,$C0
	.dc.b	$A7,$0F,$9A,$32,$F1,$6C,$EA,$3D,$31,$83,$73,$9B,$F5,$C6,$0D,$4F
	.dc.b	$53,$E3,$94,$F8,$EE,$DC,$47,$26,$DF,$C4,$AE,$CF,$C1,$AE,$13,$C8
	.dc.b	$D9,$F0,$DF,$F4,$54,$F0,$8C,$1A,$3D,$C5,$BD,$38,$3A,$12,$FC,$8F
	.dc.b	$C2,$38,$57,$8A,$32,$4B,$F3,$4F,$C0,$39,$38,$69,$F4,$E7,$E3,$B8
	.dc.b	$3C,$53,$CD,$F4,$99,$BF,$3D,$FA,$1B,$5E,$3B,$ED,$5F,$D1,$6B,$F2
	.dc.b	$4D,$9F,$EC,$51,$88,$D9,$0E,$E6,$CF,$E6,$47,$E5,$4F,$CE,$FB,$3F
	.dc.b	$A6,$3D,$38,$7E,$71,$C4,$7C,$63,$64,$3D,$D3,$F9,$68,$DC,$62,$AD
	.dc.b	$73,$8E,$47,$EB,$A7,$37,$F2,$E7,$E3,$4F,$F3,$A9,$CB,$F2,$23,$36
	.dc.b	$9C,$CF,$AC,$F1,$B8,$8F,$9A,$DB,$FD,$1B,$67,$F1,$0D,$78,$9E,$DA
	.dc.b	$EF,$3C,$E7,$E2,$DF,$EE,$5E,$13,$E4,$6C,$7A,$CD,$57,$37,$0F,$AD
	.dc.b	$CB,$F0,$CF,$8B,$3F,$1D,$C1,$F2,$CE,$47,$F0,$79,$F1,$B4,$C9,$F0
	.dc.b	$E3,$C8,$E2,$FE,$1B,$6F,$E1,$B4,$FF,$06,$CA,$FF,$8D,$E3,$FE,$A4
	.dc.b	$8C,$4F,$F8,$4A,$3F,$D1,$13,$AF,$98,$FF,$80,$11,$B8,$E3,$EB,$E2
	.dc.b	$65,$76,$9C,$B4,$FF,$0E,$FD,$F7,$FA,$84,$FF,$55,$B9,$0E,$1F,$F6
	.dc.b	$8B,$BF,$80,$FE,$63,$8F,$9A,$77,$18,$AF,$E6,$B5,$C6,$B4,$F7,$F2
	.dc.b	$B5,$7E,$26,$BE,$06,$C3,$ED,$2C,$7F,$94,$E0,$D2,$4F,$EC,$23,$FD
	.dc.b	$2B,$7F,$D0,$7E,$23,$FA,$A9,$CA,$7D,$DF,$7D,$FD,$89,$FF,$14,$35
	.dc.b	$FF,$23,$35,$FE,$9E,$D7,$FC,$8A,$BE,$F7,$D2,$D8,$EB,$18,$E7,$CF
	.dc.b	$C1,$7B,$BE,$FB,$ED,$D7,$F9,$39,$AA,$D3,$BD,$FF,$CE,$55,$FE,$CB
	.dc.b	$AF,$F2,$D3,$5F,$EC,$6D,$7F,$CB,$3E,$5F,$FA,$AF,$3F,$EC,$ED,$8C
	.dc.b	$FB,$5E,$3F,$AA,$FA,$33,$FE,$B3,$9F,$F3,$33,$5E,$33,$FE,$83,$8C
	.dc.b	$47,$F9,$9B,$ED,$5E,$47,$FA,$1D,$B1,$FC,$89,$AF,$F9,$55,$AF,$6E
	.dc.b	$9E,$84,$7F,$94,$D7,$F1,$17,$F9,$5F,$E1,$BF,$8F,$7F,$CF,$4D,$7F
	.dc.b	$C0,$6D,$7F,$CF,$2D,$5F,$69,$FE,$17,$7F,$9A,$B5,$FF,$34,$35,$FF
	.dc.b	$43,$35,$FF,$07,$FD,$FB,$EC,$3F,$D4,$8D,$8F,$E7,$ED,$7F,$CE,$4D
	.dc.b	$7F,$D2,$18,$3C,$EF,$F5,$63,$63,$FA,$3B,$5F,$F3,$D3,$5E,$3E,$67
	.dc.b	$02,$3C,$1F,$F5,$B7,$DA,$BC,$3A,$49,$FE,$A7,$19,$7F,$AA,$FD,$A7
	.dc.b	$F5,$B7,$FD,$3E,$D4,$D3,$1F,$69,$FE,$BA,$8F,$F1,$8B,$5F,$F5,$3B
	.dc.b	$53,$7E,$D3,$FD,$DF,$1F,$D9,$1A,$FF,$A5,$DA,$FF,$AD,$5A,$FF,$8F
	.dc.b	$BE,$D1,$F7,$3F,$ED,$36,$C7,$F6,$E6,$BF,$EA,$16,$BF,$EB,$FF,$C2
	.dc.b	$FE,$53,$FE,$48,$6B,$FE,$BE,$6A,$6F,$DB,$3C,$DF,$F7,$13,$67,$FB
	.dc.b	$EB,$5F,$F5,$83,$5F,$F6,$77,$DB,$3E,$D9,$EF,$7F,$81,$23,$0D,$F3
	.dc.b	$3E,$A5,$70,$AD,$BF,$EB,$BF,$B5,$7C,$B7,$F4,$4F,$2B,$F9,$84,$3E
	.dc.b	$8F,$1E,$0F,$F9,$92,$36,$FF,$3B,$74,$FF,$3D,$FB,$4F,$F4,$33,$FC
	.dc.b	$FF,$F0,$07,$C3,$72,$CB,$FD,$C3,$5A,$58,$FF,$3F,$AD,$DF,$F4,$7F
	.dc.b	$B6,$7C,$37,$1B,$43,$AF,$0E,$AB,$84,$6F,$1F,$D2,$47,$97,$CC,$C7
	.dc.b	$76,$5E,$9F,$6D,$F2,$A7,$FA,$B7,$CD,$FD,$57,$FA,$C4,$67,$2F,$3C
	.dc.b	$7F,$A1,$A7,$51,$83,$FD,$FB,$F6,$DF,$E1,$D9,$FE,$C0,$FF,$A6,$5B
	.dc.b	$E0,$FA,$E3,$13,$C4,$7C,$C7,$F5,$2A,$FE,$D0,$BE,$9B,$E1,$39,$9E
	.dc.b	$0E,$67,$FB,$77,$DA,$BE,$DB,$E4,$7D,$FF,$C2,$BE,$D3,$9A,$FE,$E9
	.dc.b	$39,$E4,$FE,$E1,$F5,$7F,$D1,$A7,$FB,$C2,$FB,$BF,$A4,$E7,$0F,$43
	.dc.b	$99,$F0,$F5,$7D,$7C,$7F,$0B,$5E,$08,$C7,$57,$DB,$FB,$37,$9D,$F7
	.dc.b	$7E,$FF,$F6,$78,$FF,$81,$E7,$2F,$F8,$26,$33,$87,$C7,$FB,$93,$F8
	.dc.b	$57,$F8,$B4,$FF,$83,$DC,$B7,$CC,$FC,$57,$DC,$BD,$7B,$F1,$78,$6D
	.dc.b	$7F,$85,$FE,$E7,$DD,$C5,$FE,$13,$FA,$2F,$F0,$44,$E7,$DB,$FC,$A4
	.dc.b	$63,$4F,$B5,$F7,$2E,$CF,$CC,$2F,$42,$C6,$1F,$CE,$7D,$CB,$E4,$3F
	.dc.b	$3D,$A7,$F8,$E7,$3F,$F1,$44,$E5,$F2,$1F,$27,$E8,$7F,$49,$F7,$FF
	.dc.b	$8A,$DB,$FC,$51,$CA,$9C,$FC,$57,$EA,$7E,$E9,$F3,$B7,$E9,$F0,$DC
	.dc.b	$BF,$C6,$DF,$75,$E1,$E2,$FC,$A3,$FA,$D6,$C7,$EC,$E3,$19,$FF,$8D
	.dc.b	$FE,$EB,$E3,$CF,$FB,$AB,$EF,$DF,$0D,$FB,$99,$CE,$1F,$62,$73,$5F
	.dc.b	$E4,$5F,$BA,$F8,$CD,$F9,$1F,$E0,$11,$8A,$FD,$32,$8F,$E0,$9F,$77
	.dc.b	$F7,$A7,$F5,$4F,$C7,$69,$AE,$5E,$7F,$25,$C2,$2E,$1F,$E5,$5A,$E5
	.dc.b	$A7,$F9,$47,$1F,$E3,$F5,$FB,$5F,$AF,$7C,$F7,$93,$C6,$FE,$29,$18
	.dc.b	$C7,$D5,$E1,$65,$A7,$B6,$7F,$8D,$63,$FB,$93,$FC,$BD,$5F,$C0,$27
	.dc.b	$FC,$CD,$19,$7F,$3B,$82,$74,$FF,$07,$9F,$E4,$ED,$F9,$2E,$BE,$CF
	.dc.b	$67,$F9,$44,$3E,$EF,$CF,$EE,$61,$B8,$FF,$9D,$67,$F6,$F8,$7F,$88
	.dc.b	$E7,$8F,$95,$BA,$D7,$DB,$E1,$A7,$18,$EF,$A7,$C4,$F7,$AC,$FF,$A0
	.dc.b	$75,$47,$93,$6C,$E7,$FC,$84,$FD,$C9,$6F,$D2,$BC,$EF,$F7,$0B,$FC
	.dc.b	$01,$AE,$4C,$BF,$EE,$6F,$B9,$7D,$43,$EE,$F0,$DA,$F9,$A7,$F0,$A9
	.dc.b	$DC,$E8,$E5,$EB,$7F,$56,$9D,$89,$FE,$69,$C1,$A2,$46,$A1,$F5,$BF
	.dc.b	$E9,$26,$C7,$FA,$8D,$AF,$F4,$17,$9A,$FB,$B7,$CF,$FE,$0A,$7F,$D2
	.dc.b	$33,$E7,$E5,$E8,$75,$BE,$DC,$D2,$7B,$3F,$EA,$2F,$9E,$F7,$E5,$FE
	.dc.b	$AD,$C1,$F3,$5E,$52,$5F,$03,$CD,$F7,$7E,$92,$3E,$98,$FF,$56,$E6
	.dc.b	$E7,$A4,$9F,$F6,$07,$CF,$FF,$A1,$5F,$5F,$1F,$02,$3F,$EC,$5A,$E4
	.dc.b	$FF,$B1,$FE,$EF,$E2,$E7,$FE,$C3,$6F,$47,$E9,$1B,$3F,$D0,$B0,$7D
	.dc.b	$87,$FB,$02,$7E,$3C,$9A,$4E,$0F,$F6,$2C,$FE,$52,$7F,$DA,$CF,$F8
	.dc.b	$35,$BF,$0D,$AE,$9D,$BF,$92,$9D,$4F,$E1,$1A,$FE,$31,$AF,$EE,$9F
	.dc.b	$F2,$D1,$3E,$23,$91,$CF,$29,$3F,$8E,$3D,$6E,$1F,$C8,$FD,$CB,$EF
	.dc.b	$BF,$46,$D8,$E0,$D7,$71,$AF,$16,$46,$9A,$F4,$FA,$DE,$64,$79,$BD
	.dc.b	$BC,$79,$1D,$9A,$F9,$3D,$5F,$19,$1F,$1B,$E6,$F5,$A4,$79,$9A,$FB
	.dc.b	$3D,$BF,$59,$1F,$5B,$EE,$FA,$A9,$1E,$E6,$BF,$13,$DC,$F8,$64,$7E
	.dc.b	$1B,$F1,$BC,$C9,$1F,$19,$AF,$CC,$FA,$7F,$96,$47,$E5,$BF,$3B,$E4
	.dc.b	$E4,$7C,$E6,$BF,$53,$D1,$FA,$64,$7E,$9B,$F5,$BF,$09,$91,$F5,$9A
	.dc.b	$FD,$CF,$87,$FB,$64,$7E,$DB,$F7,$BB,$24,$7D,$E6,$BF,$C5,$3E,$D3
	.dc.b	$F5,$2D,$EF,$FE,$4F,$18,$8F,$E3,$ED,$7F,$A6,$FD,$9B,$EA,$3F,$96
	.dc.b	$B6,$3F,$24,$D7,$F9,$0B,$5F,$E5,$3F,$69,$FE,$1D,$7F,$30,$D7,$F9
	.dc.b	$3B,$5D,$F1,$F6,$F3,$3F,$AE,$F8,$4E,$BF,$1C,$61,$B5,$FE,$3E,$9D
	.dc.b	$3F,$E4,$0C,$3F,$C3,$6E,$67,$6F,$9A,$BB,$3E,$A6,$BF,$E6,$F9,$FF
	.dc.b	$36,$3C,$4E,$9E,$5F,$85,$ED,$35,$FF,$08,$72,$DE,$83,$4C,$73,$D9
	.dc.b	$E8,$70,$2C,$F0,$6C,$75,$FA,$97,$C7,$DF,$7A,$57,$29,$CB,$67,$2F
	.dc.b	$FA,$26,$13,$CE,$7D,$19,$BC,$9A,$4B,$F2,$C7,$F4,$26,$C7,$E0,$88
	.dc.b	$FE,$09,$AF,$C9,$F9,$CE,$A7,$37,$FD,$35,$F5,$0F,$A6,$F9,$0D,$8F
	.dc.b	$86,$D7,$7F,$63,$A9,$91,$A7,$8C,$7D,$1F,$A9,$F9,$13,$FE,$A7,$5A
	.dc.b	$E4,$D6,$F2,$7E,$8C,$E9,$B3,$97,$FD,$61,$F8,$93,$EB,$FE,$F3,$63
	.dc.b	$EC,$B5,$DF,$BB,$A9,$FF,$09,$CE,$67,$FC,$16,$D7,$FC,$72,$D7,$FD
	.dc.b	$2E,$F3,$E6,$FC,$DC,$23,$FD,$46,$F8,$3E,$47,$B3,$F8,$AF,$A9,$7C
	.dc.b	$7F,$E4,$DF,$F5,$A4,$67,$E5,$31,$32,$30,$43,$FD,$9D,$3C,$9F,$CA
	.dc.b	$FD,$3F,$F6,$62,$7B,$5F,$9F,$7F,$B6,$C1,$CB,$C2,$35,$E9,$FC,$8E
	.dc.b	$19,$3F,$21,$F9,$7F,$17,$88,$FF,$14,$E8,$9F,$92,$48,$FE,$C1,$19
	.dc.b	$7E,$8F,$1F,$D4,$39,$6D,$F2,$3E,$97,$E0,$CF,$FB,$A5,$7C,$E7,$ED
	.dc.b	$DB,$1F,$B2,$77,$3C,$77,$D5,$4A,$D4,$7E,$FF,$E9,$7E,$09,$34,$C7
	.dc.b	$81,$96,$57,$EF,$A7,$73,$F8,$AA,$93,$F8,$A7,$F1,$B7,$88,$FC,$8C
	.dc.b	$5C,$3F,$07,$9F,$A2,$72,$71,$21,$F9,$8A,$E1,$E0,$FE,$3A,$EC,$FE
	.dc.b	$3A,$6B,$F8,$D8,$B5,$D6,$FB,$C6,$0F,$2C,$F6,$8C,$B6,$FF,$37,$7B
	.dc.b	$8C,$1D,$CA,$77,$2D,$F7,$9F,$34,$62,$FB,$BE,$E2,$5F,$81,$F7,$1B
	.dc.b	$39,$DF,$85,$DD,$79,$1D,$18,$3C,$85,$3C,$85,$BF,$23,$FB,$4B,$E9
	.dc.b	$7F,$4A,$ED,$FA,$2B,$C6,$71,$AF,$97,$3A,$8E,$47,$F0,$87,$0F,$82
	.dc.b	$BA,$3E,$67,$DC,$61,$E8,$7B,$8D,$57,$86,$34,$70,$70,$48,$F7,$C7
	.dc.b	$C9,$7F,$89,$D7,$F1,$62,$3F,$AE,$6B,$FB,$46,$BC,$0F,$E4,$3C,$5F
	.dc.b	$C9,$0F,$5E,$1F,$C2,$63,$F9,$4C,$60,$D3,$E8,$3F,$8F,$B8,$7D,$A7
	.dc.b	$03,$A9,$B3,$F9,$13,$BA,$F7,$47,$CC,$38,$11,$CB,$F3,$6F,$F9,$7C
	.dc.b	$F1,$7E,$24,$64,$F1,$9E,$03,$C7,$3F,$B0,$9E,$11,$DE,$3A,$D7,$BD
	.dc.b	$D4,$79,$E3,$15,$E0,$8C,$57,$27,$C0,$49,$DB,$E2,$3C,$33,$C5,$F6
	.dc.b	$3E,$88,$FE,$2E,$E6,$BC,$33,$C0,$E4,$E0,$F3,$DF,$86,$BB,$CF,$86
	.dc.b	$31,$7F,$5E,$3E,$15,$70,$6D,$F7,$3C,$55,$C2,$3D,$A6,$8F,$8D,$1E
	.dc.b	$7A,$F0,$4F,$F5,$5F,$5F,$FA,$B7,$C3,$4B,$FE,$B5,$19,$7F,$AE,$46
	.dc.b	$5F,$EA,$B1,$FD,$47,$DC,$91,$FD,$59,$FE,$A9,$EF,$E2,$19,$7F,$6A
	.dc.b	$E5,$FF,$0D,$46,$33,$32,$33,$3B,$AF,$F1,$E7,$E3,$CF,$9B,$A4,$BF
	.dc.b	$F1,$CF,$86,$F9,$EC,$47,$42,$57,$17,$FD,$70,$D7,$47,$89,$C3,$87
	.dc.b	$58,$7F,$80,$BE,$A1,$F1,$3F,$E4,$F6,$D7,$F9,$2A,$B7,$A7,$87,$9D
	.dc.b	$C1,$FD,$7F,$8A,$5F,$F5,$F8,$FE,$C4,$61,$FE,$CD,$18,$7F,$AC,$6D
	.dc.b	$FD,$7C,$8F,$F5,$BE,$BA,$FC,$B3,$E5,$F3,$A7,$3A,$FF,$86,$A3,$FD
	.dc.b	$DF,$F8,$63,$CA,$FF,$34,$36,$FF,$CC,$6D,$7A,$0F,$A8,$F9,$CE,$A7
	.dc.b	$FD,$C4,$A7,$83,$F3,$1A,$FD,$36,$BF,$80,$7B,$5F,$F8,$BB,$ED,$59
	.dc.b	$7F,$9E,$DF,$3D,$FF,$8A,$FE,$03,$FC,$59,$F0,$53,$1F,$F1,$84,$65
	.dc.b	$FF,$19,$46,$5F,$F1,$5C,$7F,$89,$FC,$A9,$1F,$E2,$C7,$FC,$55,$E5
	.dc.b	$E2,$3D,$FD,$29,$C6,$7F,$99,$AF,$C0,$7C,$2C,$BF,$F4,$4F,$32,$3F
	.dc.b	$D2,$5D,$2F,$C1,$35,$5F,$8E,$15,$B7,$06,$2B,$E5,$C6,$63,$83,$70
	.dc.b	$FC,$97,$AD,$F9,$13,$AB,$FE,$2F,$D7,$FC,$59,$DD,$2F,$FC,$59,$1F
	.dc.b	$E2,$F3,$0F,$F8,$CE,$30,$FF,$89,$73,$FF,$16,$11,$FF,$14,$77,$57
	.dc.b	$F0,$7C,$BF,$E9,$77,$81,$A3,$FD,$57,$C5,$FB,$AE,$AF,$D9,$36,$3F
	.dc.b	$7E,$D7,$F1,$9B,$7D,$FB,$FC,$5C,$FF,$06,$9F,$DA,$B5,$E7,$1A,$3F
	.dc.b	$0D,$CF,$23,$C3,$1F,$EC,$3F,$A9,$FD,$8F,$D4,$FC,$E6,$CF,$D3,$5A
	.dc.b	$3E,$3B,$5D,$8B,$FA,$BF,$89,$3E,$13,$FD,$07,$8F,$F0,$D8,$C6,$DA
	.dc.b	$9C,$CF,$F1,$C9,$CB,$F8,$1F,$C5,$FF,$DF,$55,$F8,$06,$C6,$7D,$BE
	.dc.b	$2A,$32,$FF,$16,$C9,$F0,$5A,$6D,$FE,$29,$B3,$F8,$F6,$BF,$C7,$BE
	.dc.b	$99,$F6,$3D,$9B,$5F,$92,$6C,$FE,$5D,$AF,$F2,$6F,$C3,$9F,$82,$79
	.dc.b	$9B,$5F,$9A,$6C,$FE,$7D,$AF,$F2,$EF,$C3,$B9,$BF,$CD,$5A,$FF,$42
	.dc.b	$6A,$68,$F8,$EF,$8A,$DA,$FD,$4B,$67,$F5,$CD,$7C,$1E,$F7,$02,$4F
	.dc.b	$07,$E7,$FB,$4F,$8A,$A4,$9F,$DA,$C6,$5B,$F1,$5C,$08,$E0,$DE,$FD
	.dc.b	$08,$E0,$FD,$FF,$E4,$0F,$1F,$EF,$B6,$33,$18,$32,$C3,$89,$72,$FE
	.dc.b	$19,$F8,$73,$E3,$FF,$12,$D9,$FE,$1A,$D9,$DC,$FE,$3B,$F0,$EF,$F6
	.dc.b	$BD,$7F,$12,$6C,$FF,$17,$6B,$FD,$99,$AF,$F6,$D7,$FA,$EC,$71,$EA
	.dc.b	$E8,$47,$A7,$5F,$72,$38,$3F,$90,$FB,$8F,$6B,$F4,$2D,$8C,$C6,$0C
	.dc.b	$F9,$D2,$38,$B7,$1F,$E5,$D5,$FD,$86,$7A,$F8,$3D,$08,$FF,$30,$F9
	.dc.b	$CF,$E7,$0D,$F2,$FF,$63,$18,$BF,$E7,$0D,$7F,$C0,$EF,$F0,$79,$E5
	.dc.b	$D9,$C0,$8F,$F3,$CF,$9C,$E5,$E5,$23,$FC,$29,$F3,$6F,$83,$C6,$3F
	.dc.b	$DB,$AF,$75,$A9,$CD,$EA,$95,$E1,$78,$65,$AB,$FF,$10,$4E,$6B,$83
	.dc.b	$D2,$BF,$C5,$F1,$FD,$BA,$69,$E4,$F6,$B9,$4E,$B6,$AB,$2F,$F1,$95
	.dc.b	$70,$8E,$7A,$EB,$0E,$91,$FE,$08,$7E,$93,$6F,$9A,$F9,$CE,$B8,$7E
	.dc.b	$2A,$3F,$C1,$E6,$70,$EC,$F8,$78,$B8,$FB,$5C,$AE,$D6,$AB,$81,$CB
	.dc.b	$D0,$CB,$63,$55,$E2,$DB,$C3,$C9,$67,$BE,$5F,$D1,$1E,$9C,$39,$C3
	.dc.b	$CA,$DD,$7F,$9F,$FC,$E7,$F8,$8A,$4D,$23,$FE,$01,$F9,$97,$C6,$3E
	.dc.b	$73,$FC,$03,$18,$6C,$66,$78,$35,$FF,$2B,$CB,$E0,$FF,$CD,$B1,$D5
	.dc.b	$AE,$9A,$FF,$98,$E3,$86,$1E,$0E,$9F,$F3,$33,$F8,$D8,$FD,$94,$FD
	.dc.b	$06,$BF,$E4,$27,$8A,$F9,$C7,$C9,$75,$E5,$E7,$B5,$AE,$51,$87,$FC
	.dc.b	$FF,$7D,$30,$F0,$73,$66,$53,$FE,$59,$7C,$51,$9F,$9C,$FF,$3B,$D7
	.dc.b	$9F,$D3,$E1,$7F,$E0,$B2,$56,$3D,$79,$87,$AD,$DD,$77,$CD,$CF,$49
	.dc.b	$5F,$90,$F9,$CF,$EC,$EF,$E4,$A7,$FC,$65,$AF,$FA,$35,$B1,$FE,$9A
	.dc.b	$6B,$F1,$2B,$D9,$F3,$9E,$FA,$FF,$4A,$36,$7F,$D4,$2D,$57,$FD,$7D
	.dc.b	$FD,$0C,$7F,$99,$1B,$B7,$43,$AF,$07,$C9,$F3,$3F,$C3,$73,$FE,$4C
	.dc.b	$6B,$FE,$51,$6B,$FE,$A0,$6B,$FE,$A7,$77,$BF,$F0,$F5,$ED,$75,$3B
	.dc.b	$74,$F7,$F8,$5F,$F4,$0C,$FF,$A1,$2F,$B4,$60,$CF,$8F,$23,$C8,$D7
	.dc.b	$C9,$AF,$22,$3E,$97,$AF,$B5,$23,$D8,$D7,$FC,$D9,$F3,$8F,$9C,$FF
	.dc.b	$49,$B6,$3F,$CD,$AD,$77,$1A,$F6,$64,$69,$AF,$C7,$F8,$AF,$69,$1F
	.dc.b	$98,$F6,$F9,$59,$1D,$9A,$F9,$3E,$73,$C6,$47,$C6,$FC,$8F,$8B,$91
	.dc.b	$F5,$DA,$FB,$3E,$4F,$D6,$47,$D6,$FB,$BD,$59,$1E,$E6,$BF,$7F,$E2
	.dc.b	$BE,$A9,$1F,$C2,$3F,$1B,$D4,$91,$F1,$9A,$FC,$CF,$9C,$F9,$64,$7E
	.dc.b	$5B,$F8,$0F,$32,$47,$E3,$DA,$FD,$4F,$9B,$FA,$64,$7E,$9B,$F5,$BE
	.dc.b	$6E,$47,$D6,$6B,$FE,$A7,$F9,$B7,$B5,$FE,$CE,$6C,$7F,$AA,$1A,$EE
	.dc.b	$35,$E0,$48,$D3,$5F,$F5,$97,$CC,$BE,$0F,$FD,$B2,$D8,$FF,$59,$B5
	.dc.b	$DC,$7F,$B7,$BE,$6F,$FD,$05,$1F,$D5,$1A,$FF,$B9,$9B,$DF,$FD,$DF
	.dc.b	$EF,$9F,$2D,$FC,$76,$77,$42,$07,$2E,$1A,$ED,$AE,$FA,$3F,$94,$53
	.dc.b	$8A,$9D,$E3,$F9,$67,$DE,$FC,$6C,$7F,$C1,$4D,$ED,$F0,$B7,$0F,$22
	.dc.b	$E8,$E3,$BC,$9F,$EE,$73,$B4,$ED,$CF,$A5,$E0,$F5,$24,$60,$83,$1F
	.dc.b	$1C,$6A,$31,$8F,$92,$BF,$A2,$7C,$D7,$FB,$CE,$7F,$C0,$6D,$76,$7C
	.dc.b	$B7,$17,$FE,$27,$5F,$77,$CB,$E4,$7F,$DD,$F4,$9E,$4F,$D2,$31,$1A
	.dc.b	$8C,$C7,$D6,$7A,$E1,$7C,$EF,$29,$D6,$7F,$C2,$FF,$78,$F9,$AF,$EB
	.dc.b	$8D,$8C,$C6,$39,$32,$C8,$C9,$CA,$3F,$B0,$70,$FF,$61,$F0,$E4,$FF
	.dc.b	$89,$3D,$3F,$C7,$E5,$F8,$A6,$C7,$E3,$5F,$E4,$1B,$7F,$92,$5B,$3B
	.dc.b	$BF,$EE,$1D,$1F,$E1,$4B,$FF,$17,$B5,$F9,$A1,$C5,$FF,$94,$D7,$D2
	.dc.b	$72,$F5,$FF,$56,$4A,$FD,$5B,$5E,$73,$FE,$19,$7B,$FA,$FC,$5F,$F1
	.dc.b	$FF,$CB,$3C,$5F,$F0,$23,$63,$31,$8E,$8C,$B2,$32,$72,$8F,$F0,$67
	.dc.b	$D0,$3E,$5A,$FC,$54,$AF,$F2,$6C,$E5,$B3,$97,$E0,$C2,$79,$CF,$C8
	.dc.b	$E5,$95,$F2,$1F,$97,$EC,$62,$3F,$87,$C7,$F0,$E2,$3F,$C3,$9F,$A3
	.dc.b	$D1,$23,$91,$FE,$60,$FA,$0F,$C4,$3F,$49,$BC,$3F,$E8,$36,$F1,$FF
	.dc.b	$0A,$7F,$87,$CF,$1E,$7F,$F1,$64,$6A,$3F,$88,$35,$3C,$1F,$BF,$F4
	.dc.b	$1F,$3D,$BB,$7F,$8D,$63,$F8,$01,$9C,$3F,$D2,$8D,$C3,$F8,$DB,$A9
	.dc.b	$E1,$BF,$F8,$F7,$E7,$BF,$61,$87,$F9,$F1,$AF,$E3,$9B,$87,$FA,$7D
	.dc.b	$FF,$53,$46,$FA,$F9,$11,$E5,$F0,$D5,$E5,$16,$B9,$4D,$CB,$F9,$7D
	.dc.b	$F6,$6D,$79,$67,$57,$FC,$D6,$B4,$D9,$D3,$E3,$21,$E0,$3C,$66,$2F
	.dc.b	$D5,$1D,$21,$3E,$C7,$81,$27,$6F,$23,$17,$E6,$7A,$3E,$AB,$E9,$5E
	.dc.b	$69,$EB,$7E,$67,$89,$E9,$38,$9E,$D9,$E3,$1C,$9B,$1A,$3C,$EE,$1E
	.dc.b	$66,$0D,$4F,$84,$91,$BB,$F5,$CF,$DD,$8C,$5F,$AE,$30,$72,$BF,$13
	.dc.b	$4E,$B3,$C1,$F5,$1C,$0F,$1D,$F8,$67,$E0,$5F,$80,$97,$DD,$FC,$1D
	.dc.b	$75,$38,$DF,$57,$E4,$9F,$46,$3D,$2F,$F4,$5B,$FC,$BB,$6F,$FA,$5C
	.dc.b	$62,$3D,$86,$4F,$E9,$87,$A0,$C1,$87,$E4,$7D,$07,$E9,$4F,$12,$71
	.dc.b	$1E,$7A,$FD,$24,$62,$3E,$0D,$59,$CC,$6F,$0F,$CF,$D6,$5F,$84,$48
	.dc.b	$FC,$F4,$65,$F9,$B3,$F2,$1C,$9F,$B7,$8F,$29,$B7,$E3,$9F,$49,$C5
	.dc.b	$FD,$38,$E0,$43,$CE,$59,$F3,$B8,$3F,$81,$4F,$2B,$FC,$EB,$63,$F1
	.dc.b	$ED,$7C,$07,$3B,$F6,$3E,$4A,$EA,$48,$CD,$FD,$09,$E0,$7C,$08,$CD
	.dc.b	$75,$AF,$85,$3F,$B8,$8C,$4E,$A3,$89,$DD,$C3,$EA,$3E,$91,$F2,$8C
	.dc.b	$1E,$F9,$F2,$39,$7B,$D7,$96,$37,$1E,$2F,$A0,$FF,$1F,$47,$F9,$26
	.dc.b	$15,$FE,$4A,$97,$A1,$FE,$EB,$6D,$6E,$75,$E2,$E4,$D5,$49,$D1,$FE
	.dc.b	$55,$AE,$EF,$F8,$0B,$E7,$3F,$DA,$4D,$F7,$3F,$DF,$11,$8C,$7C,$73
	.dc.b	$FE,$5D,$8F,$F6,$E1,$9C,$7F,$CB,$ED,$8D,$57,$2F,$73,$FA,$37,$85
	.dc.b	$27,$FA,$33,$FD,$27,$E2,$F1,$1F,$A9,$8F,$D8,$91,$FD,$79,$FC,$B9
	.dc.b	$FE,$A1,$CB,$23,$99,$EE,$1F,$3F,$F1,$9F,$F0,$FA,$F5,$FF,$CE,$BA
	.dc.b	$7B,$27,$71,$8D,$3F,$CF,$35,$C6,$B4,$D8,$CC,$EE,$BF,$D0,$13,$D5
	.dc.b	$F1,$3B,$7F,$D0,$9F,$38,$CF,$84,$23,$0F,$A5,$AF,$F8,$D9,$FF,$16
	.dc.b	$36,$38,$C6,$B0,$F5,$37,$0F,$F4,$8B,$C2,$3F,$B8,$6F,$FD,$DC,$97
	.dc.b	$FD,$DF,$1A,$FF,$19,$8F,$E4,$24,$7E,$8F,$0F,$AD,$FE,$BF,$C3,$FE
	.dc.b	$01,$38,$DF,$CB,$E1,$7B,$DF,$97,$F5,$65,$FD,$59,$CB,$67,$2F,$F9
	.dc.b	$52,$13,$F9,$F9,$FC,$9F,$6C,$AF,$BD,$C8,$8F,$CB,$FC,$1E,$67,$F3
	.dc.b	$91,$71,$FB,$D4,$AF,$A3,$F4,$0F,$8C,$A4,$C3,$F1,$5E,$0B,$E0,$31
	.dc.b	$1F,$D0,$2B,$83,$F9,$46,$B9,$AF,$F4,$9C,$63,$1F,$C7,$7C,$F3,$E3
	.dc.b	$3F,$02,$DB,$FC,$C5,$6F,$97,$F2,$3B,$FE,$1B,$9F,$F6,$F8,$FE,$4B
	.dc.b	$8B,$13,$30,$FC,$7C,$6C,$C3,$F8,$9D,$BF,$0E,$47,$F1,$DD,$75,$D1
	.dc.b	$F9,$18,$FC,$6C,$7E,$5B,$E7,$DE,$45,$25,$7E,$BF,$E8,$3E,$33,$FB
	.dc.b	$19,$DB,$FB,$86,$BF,$DA,$1A,$FF,$AD,$5A,$FF,$BA,$1D,$E5,$FB,$5F
	.dc.b	$A0,$FE,$19,$26,$CE,$B5,$FF,$77,$FD,$DB,$EF,$3F,$D8,$6D,$BF,$E1
	.dc.b	$2D,$7F,$BC,$B5,$FF,$63,$7D,$17,$E4,$3E,$8B,$F8,$7B,$63,$FB,$FB
	.dc.b	$F9,$CF,$63,$3C,$2A,$FE,$2B,$F4,$4F,$57,$F8,$66,$02,$46,$73,$DC
	.dc.b	$EA,$72,$D5,$F6,$CF,$8B,$C3,$6B,$F9,$1B,$67,$FC,$26,$D7,$FD,$C1
	.dc.b	$93,$E1,$FA,$36,$BF,$95,$B6,$7F,$C3,$6D,$7F,$DD,$1F,$42,$FA,$2F
	.dc.b	$23,$6B,$F9,$9B,$67,$FC,$46,$D7,$FD,$E1,$F6,$8F,$8E,$F6,$36,$BF
	.dc.b	$9D,$B6,$7F,$C5,$6D,$7F,$DF,$1F,$43,$CD,$F9,$41,$C3,$F3,$5A,$E7
	.dc.b	$0F,$95,$0F,$90,$A4,$BF,$E9,$71,$97,$FA,$6F,$D9,$BD,$9F,$A4,$D8
	.dc.b	$CC,$60,$FA,$DE,$B3,$E3,$E9,$27,$FA,$BC,$65,$BF,$45,$C0,$8E,$0F
	.dc.b	$EB,$BF,$62,$F9,$CF,$C1,$36,$33,$18,$3F,$0D,$F4,$7F,$7F,$6F,$8A
	.dc.b	$D7,$F2,$0D,$4D,$3F,$C1,$B3,$FD,$B1,$B1,$FE,$5C,$6B,$F4,$0E,$31
	.dc.b	$F9,$5F,$AE,$F9,$11,$FE,$62,$6B,$B6,$BF,$61,$AF,$E8,$FE,$89,$E3
	.dc.b	$7E,$7D,$B5,$FD,$F9,$B3,$FE,$6E,$6B,$E3,$7F,$A2,$C6,$FB,$39,$11
	.dc.b	$FF,$03,$FD,$4B,$E5,$BF,$58,$D8,$CC,$60,$FD,$87,$B7,$FD,$D6,$3F
	.dc.b	$C1,$ED,$C3,$FC,$FE,$D7,$E2,$35,$FC,$AA,$F2,$39,$92,$79,$BF,$BD
	.dc.b	$FA,$37,$CA,$52,$4F,$F8,$7A,$32,$DF,$8F,$E0,$47,$07,$F8,$93,$E8
	.dc.b	$D1,$FE,$36,$F2,$BF,$C3,$9C,$4F,$82,$C5,$FE,$AA,$F6,$FF,$89,$E7
	.dc.b	$4F,$F8,$A6,$1E,$07,$0D,$7F,$8D,$47,$05,$87,$F8,$A6,$37,$5C,$71
	.dc.b	$E1,$5A,$D2,$E3,$FD,$BE,$BF,$C6,$D8,$7C,$77,$C5,$E9,$F0,$8F,$EE
	.dc.b	$D1,$AA,$C7,$9F,$97,$AE,$0F,$A3,$EB,$C7,$86,$DF,$F9,$0E,$7F,$B8
	.dc.b	$36,$38,$FA,$9B,$8D,$70,$E1,$A7,$18,$EF,$AF,$AB,$D9,$B3,$F2,$5F
	.dc.b	$1F,$06,$31,$F8,$B3,$FC,$C8,$F1,$57,$A2,$1F,$47,$FE,$5C,$7F,$A3
	.dc.b	$35,$C9,$97,$FC,$C3,$F4,$2F,$8E,$7D,$1E,$1B,$5F,$ED,$49,$D4,$E4
	.dc.b	$E3,$F1,$DF,$62,$74,$D7,$F7,$D2,$8C,$47,$F3,$08,$C7,$17,$8B,$E1
	.dc.b	$3F,$CB,$31,$FE,$4F,$9F,$EB,$CD,$4F,$05,$F4,$6F,$A3,$E9,$C9,$C8
	.dc.b	$93,$FE,$79,$AE,$47,$F8,$DD,$C7,$2F,$80,$FF,$34,$FC,$9F,$F1,$5C
	.dc.b	$BF,$CF,$55,$FE,$43,$AE,$5F,$05,$E6,$38,$3D,$7C,$8F,$49,$1D,$9F
	.dc.b	$E8,$9C,$BF,$6A,$DE,$6F,$F3,$BC,$62,$3E,$E9,$FD,$F7,$E8,$FF,$B3
	.dc.b	$3C,$78,$F6,$47,$91,$FE,$95,$FA,$3F,$8C,$4E,$0F,$F4,$1D,$7F,$57
	.dc.b	$9D,$6F,$FE,$86,$AF,$F1,$09,$2B,$FD,$13,$E6,$FF,$9A,$AB,$C6,$7F
	.dc.b	$5F,$E8,$C9,$AF,$72,$C7,$FA,$AA,$78,$3F,$CF,$5A,$FF,$6E,$6B,$FE
	.dc.b	$19,$7F,$CB,$AE,$FB,$7F,$31,$18,$8E,$11,$8F,$82,$91,$DD,$AF,$F4
	.dc.b	$B3,$FB,$A7,$AD,$C8,$8F,$95,$EB,$E5,$C8,$F4,$35,$E5,$7F,$E0,$D7
	.dc.b	$D9,$E9,$F9,$08,$FF,$B2,$3E,$89,$F2,$FF,$AF,$6C,$66,$30,$7B,$3C
	.dc.b	$89,$1F,$1D,$AF,$83,$E2,$BE,$09,$1F,$98,$F5,$FA,$29,$1E,$16,$BE
	.dc.b	$8F,$8E,$F9,$24,$7E,$A3,$F3,$3E,$32,$47,$D7,$6B,$F2,$71,$7C,$8F
	.dc.b	$AC,$91,$EF,$7E,$AF,$C7,$C8,$FC,$03,$5F,$93,$E9,$7D,$A2,$3F,$85
	.dc.b	$7E,$37,$BB,$23,$E5,$35,$FA,$3F,$09,$F8,$22,$3F,$8D,$7F,$0B,$E6
	.dc.b	$C8,$FC,$83,$5F,$57,$D1,$7D,$32,$3F,$5D,$FE,$49,$F8,$0F,$EB,$23
	.dc.b	$F1,$0D,$7F,$C2,$8D,$7F,$C8,$AD,$7F,$CE,$1F,$44,$F6,$FF,$97,$36
	.dc.b	$3F,$24,$1C,$3F,$95,$6B,$FE,$21,$7F,$80,$37,$C2,$E6,$47,$9F,$E0
	.dc.b	$5F,$C8,$49,$DD,$FE,$FE,$7F,$79,$1F,$B6,$7F,$DF,$AF,$F8,$B5,$AF
	.dc.b	$F9,$79,$AF,$FA,$25,$A9,$ED,$E5,$F5,$3B,$4F,$F9,$9B,$C8,$79,$FF
	.dc.b	$E6,$8F,$C0,$7F,$D9,$98,$7F,$A4,$DA,$9A,$7F,$AB,$5F,$F4,$C4,$F2
	.dc.b	$78,$78,$1D,$6B,$A3,$63,$27,$08,$F0,$78,$39,$8E,$DF,$4D,$70,$FE
	.dc.b	$31,$87,$59,$F1,$78,$79,$A9,$1C,$CF,$44,$71,$32,$FF,$20,$F8,$6F
	.dc.b	$F7,$53,$7D,$4F,$04,$62,$FB,$2E,$DC,$CF,$F7,$18,$E3,$5F,$E5,$F6
	.dc.b	$BF,$E9,$06,$BF,$EB,$B7,$A6,$79,$F5,$7B,$9D,$27,$D0,$3F,$0D,$7D
	.dc.b	$37,$C3,$6C,$69,$AF,$F3,$A7,$FD,$04,$E6,$38,$11,$E1,$EC,$7C,$DF
	.dc.b	$A5,$7D,$5F,$5F,$8C,$95,$E2,$9C,$B6,$72,$F7,$9E,$51,$F5,$FE,$9F
	.dc.b	$FD,$B5,$3E,$A6,$BF,$D3,$9B,$BF,$D9,$B7,$8B,$C7,$6F,$F0,$E5,$70
	.dc.b	$BF,$F4,$53,$5F,$F5,$DB,$5F,$F7,$33,$D3,$6F,$F5,$FD,$FF,$A1,$A7
	.dc.b	$D5,$E0,$F5,$7E,$6F,$D3,$7E,$E3,$2F,$C5,$36,$FF,$D8,$ED,$9F,$F7
	.dc.b	$92,$FA,$7F,$CA,$39,$83,$49,$1A,$F7,$B7,$EA,$4A,$F0,$C6,$5F,$1C
	.dc.b	$65,$FE,$E7,$5F,$E6,$A7,$F0,$7F,$3C,$FA,$1F,$D2,$36,$33,$18,$FA
	.dc.b	$7C,$BF,$E4,$59,$D4,$FF,$69,$6B,$FE,$B5,$6B,$FE,$E6,$77,$9F,$EB
	.dc.b	$5F,$F6,$E4,$78,$BD,$9E,$AF,$E5,$BE,$9D,$E5,$FE,$DD,$B1,$98,$C7
	.dc.b	$1E,$59,$19,$39,$47,$F8,$D3,$8F,$FC,$6B,$E8,$C9,$FD,$1F,$BB,$F7
	.dc.b	$50,$70,$D8,$E0,$FF,$87,$23,$3F,$4D,$FE,$61,$8F,$D6,$B5,$FF,$10
	.dc.b	$B7,$4F,$E1,$8B,$E9,$FF,$89,$39,$E4,$F0,$D2,$30,$F9,$1A,$E6,$78
	.dc.b	$FC,$1C,$8F,$0B,$D2,$31,$1E,$28,$B7,$E2,$9B,$7E,$0B,$EA,$48,$CD
	.dc.b	$F8,$6E,$BE,$13,$B9,$0E,$C7,$C2,$8D,$CF,$49,$F2,$4E,$EF,$E1,$C6
	.dc.b	$AB,$6D,$9D,$5F,$96,$31,$1D,$89,$5C,$9F,$01,$C0,$F4,$14,$F4,$1C
	.dc.b	$27,$A3,$9B,$FA,$8D,$AE,$F5,$D6,$38,$1D,$0F,$BB,$7E,$87,$93,$5D
	.dc.b	$9E,$08,$E4,$4B,$F0,$C7,$23,$C0,$72,$29,$CA,$FC,$AD,$BF,$2C,$F2
	.dc.b	$AE,$93,$E5,$7E,$F3,$C9,$CB,$E3,$8F,$0B,$F9,$03,$F2,$13,$FE,$C1
	.dc.b	$AF,$CD,$46,$5A,$F0,$21,$F0,$CD,$47,$13,$F3,$BE,$3C,$8F,$15,$7E
	.dc.b	$3E,$31,$1E,$B3,$F5,$2E,$2F,$CF,$7F,$94,$8C,$57,$BC,$95,$FA,$58
	.dc.b	$CB,$F0,$63,$2F,$EA,$E3,$2F,$B8,$FC,$03,$89,$FC,$B4,$78,$88,$79
	.dc.b	$4F,$DD,$C6,$EF,$F3,$AD,$8E,$67,$0B,$F4,$CF,$59,$FC,$24,$7F,$02
	.dc.b	$78,$4F,$DD,$6C,$7C,$06,$DF,$AA,$FF,$DC,$55,$D6,$31,$7D,$63,$E6
	.dc.b	$9F,$0C,$D1,$F0,$C8,$7C,$33,$27,$C3,$3C,$46,$27,$E1,$BF,$00,$C4
	.dc.b	$7C,$06,$9F,$00,$D1,$F1,$EF,$DC,$E8,$FF,$61,$FC,$D7,$FA,$19,$FF
	.dc.b	$09,$B5,$CF,$A5,$FE,$35,$8F,$EC,$46,$67,$FC,$68,$DC,$3F,$93,$B6
	.dc.b	$BF,$9C,$BF,$DA,$7E,$9D,$F3,$FF,$E3,$F5,$F4,$F4,$C4,$4A,$FF,$12
	.dc.b	$CE,$9B,$39,$7F,$9D,$F2,$7F,$38,$23,$FD,$03,$E0,$EB,$FE,$CB,$D7
	.dc.b	$FA,$EB,$67,$F5,$84,$7F,$A3,$F3,$FF,$42,$38,$E9,$FE,$51,$FA,$7F
	.dc.b	$EB,$F3,$FE,$EA,$6C,$67,$9B,$FC,$B5,$1F,$E1,$E9,$FF,$1E,$BA,$DB
	.dc.b	$FA,$5B,$63,$FA,$EB,$53,$C5,$FF,$7B,$FD,$3F,$CB,$4F,$F8,$4C,$9C
	.dc.b	$1F,$E6,$97,$46,$7D,$3C,$47,$F9,$B7,$E9,$5D,$74,$97,$FE,$52,$87
	.dc.b	$C6,$E2,$3F,$DE,$51,$C1,$FF,$51,$B5,$CC,$FF,$A9,$94,$67,$4F,$F3
	.dc.b	$C2,$FA,$7F,$11,$D6,$3F,$CF,$9F,$4D,$FE,$39,$A4,$63,$DE,$FF,$42
	.dc.b	$49,$BE,$2C,$AF,$F3,$4C,$E5,$B3,$97,$9E,$FF,$DE,$BE,$99,$F1,$1F
	.dc.b	$E8,$F6,$CF,$F9,$C9,$AF,$03,$FC,$22,$BD,$7E,$13,$27,$83,$CE,$3E
	.dc.b	$77,$85,$27,$E6,$3F,$47,$E0,$B1,$1F,$32,$3E,$46,$DF,$7E,$3F,$A0
	.dc.b	$35,$FF,$08,$BF,$83,$D7,$E5,$F2,$FD,$B3,$9B,$FE,$A5,$FA,$4F,$85
	.dc.b	$5E,$90,$BE,$0F,$FD,$5C,$D8,$FF,$48,$BC,$27,$96,$DA,$A9,$5A,$8F
	.dc.b	$F5,$B7,$D3,$FC,$38,$FF,$4C,$B5,$F0,$36,$BF,$D7,$0F,$F2,$68,$CE
	.dc.b	$9C,$88,$F2,$F0,$BF,$2D,$72,$3F,$28,$FE,$67,$0C,$47,$E5,$23,$E4
	.dc.b	$91,$FC,$AB,$F9,$FE,$29,$1C,$8F,$F5,$7F,$AF,$FC,$49,$FF,$7E,$B5
	.dc.b	$FE,$E0,$1C,$3F,$DC,$DA,$FF,$91,$3E,$8D,$F0,$7F,$A1,$6C,$72,$0E
	.dc.b	$1E,$AD,$7F,$BF,$7D,$1B,$E1,$1E,$AF,$18,$C3,$70,$FF,$43,$57,$0A
	.dc.b	$DB,$E8,$FB,$AF,$1F,$47,$85,$B1,$9C,$BF,$DD,$4F,$F9,$72,$32,$FF
	.dc.b	$83,$FE,$8D,$F1,$9A,$8F,$F3,$8C,$62,$FD,$8D,$8F,$8E,$D7,$FC,$31
	.dc.b	$EA,$3E,$23,$94,$23,$15,$F1,$5B,$1F,$41,$AF,$F8,$87,$E0,$DE,$77
	.dc.b	$26,$D7,$CD,$6C,$FD,$86,$BF,$E2,$9F,$A2,$7C,$5F,$89,$B5,$F5,$5B
	.dc.b	$3F,$80,$6B,$FE,$31,$FA,$17,$D1,$7A,$9B,$5F,$75,$B3,$F8,$86,$BF
	.dc.b	$E3,$9F,$B9,$7C,$1F,$C2,$6D,$7E,$15,$B3,$F9,$06,$BF,$E4,$1F,$A3
	.dc.b	$79,$FF,$29,$B5,$F8,$D6,$CF,$E6,$1A,$FF,$92,$7C,$C7,$D1,$FD,$26
	.dc.b	$D7,$E5,$5B,$3F,$A0,$6B,$FE,$51,$FB,$8F,$F7,$14,$FE,$6D,$B1,$FA
	.dc.b	$66,$BB,$F8,$0F,$BF,$F2,$6F,$85,$A4,$BF,$D2,$46,$5F,$D3,$7D,$1F
	.dc.b	$E2,$6F,$FD,$5E,$D7,$FD,$82,$D4,$DF,$B7,$9D,$3F,$8E,$6D,$7E,$C7
	.dc.b	$E9,$1F,$2D,$F9,$26,$C6,$63,$07,$E5,$BE,$91,$E2,$52,$4F,$EE,$E3
	.dc.b	$2F,$EF,$7E,$CD,$F2,$1F,$9E,$6C,$66,$30,$7E,$8B,$E9,$5F,$4B,$49
	.dc.b	$3F,$C1,$E3,$2F,$F0,$9F,$4F,$FA,$83,$7C,$DF,$D5,$C6,$38,$3F,$8B
	.dc.b	$B5,$FF,$44,$7D,$2B,$EB,$BF,$62,$DB,$FE,$28,$D9,$FE,$3E,$D7,$8F
	.dc.b	$CC,$70,$24,$F0,$7F,$77,$F4,$CF,$8F,$4F,$FB,$DB,$D4,$B7,$FE,$F8
	.dc.b	$E8,$FF,$05,$CE,$4F,$F7,$5D,$F1,$71,$3B,$F7,$B8,$DF,$28,$57,$F0
	.dc.b	$9A,$DC,$63,$0D,$C7,$4F,$22,$C8,$D4,$70,$BF,$F0,$0B,$CB,$17,$8F
	.dc.b	$C2,$BF,$C2,$3E,$D5,$9C,$D7,$1F,$A3,$D6,$1C,$AF,$FB,$72,$8C,$DF
	.dc.b	$39,$FD,$EC,$FA,$EF,$F9,$F5,$FA,$A7,$2F,$9D,$FD,$47,$9F,$C7,$E5
	.dc.b	$B9,$78,$F6,$B8,$3D,$78,$31,$7F,$DF,$C3,$8C,$7F,$9F,$ED,$EB,$F4
	.dc.b	$7C,$8D,$C3,$C5,$BE,$1A,$79,$63,$D5,$83,$E9,$BF,$C7,$CD,$3F,$C8
	.dc.b	$0E,$0F,$EC,$1F,$49,$FC,$E3,$E9,$AF,$A0,$69,$2B,$FB,$7C,$BD,$AF
	.dc.b	$F2,$0C,$61,$AF,$FB,$8E,$35,$A7,$69,$CC,$FC,$99,$FF,$08,$65,$7A
	.dc.b	$BE,$65,$FE,$2E,$73,$5F,$85,$79,$AF,$A6,$7D,$37,$3E,$5E,$38,$DA
	.dc.b	$FC,$8D,$FF,$91,$B0,$E9,$8E,$A7,$A7,$35,$55,$F8,$57,$8C,$67,$E4
	.dc.b	$7F,$DF,$D7,$E5,$F8,$CE,$16,$69,$8A,$CF,$A1,$30,$F1,$BA,$7F,$4B
	.dc.b	$A7,$E0,$9B,$CB,$E3,$8C,$47,$F9,$17,$E9,$BF,$AE,$BC,$39,$34,$47
	.dc.b	$89,$FB,$0F,$A6,$F9,$29,$A6,$27,$F9,$7C,$6F,$83,$E1,$4F,$C7,$E0
	.dc.b	$FC,$F3,$67,$F6,$CF,$4F,$DC,$7F,$A4,$47,$EF,$97,$CE,$FC,$B6,$C6
	.dc.b	$7B,$71,$3F,$E9,$19,$CC,$FF,$8E,$9A,$FE,$B4,$38,$7F,$60,$BE,$D7
	.dc.b	$91,$23,$FC,$75,$19,$6F,$CB,$77,$23,$83,$7D,$3E,$32,$38,$39,$F4
	.dc.b	$79,$C8,$E0,$EF,$CF,$EB,$23,$83,$C7,$AF,$BC,$8E,$0F,$3F,$1F,$C3
	.dc.b	$23,$83,$D7,$C9,$F1,$C8,$E0,$F7,$F2,$7C,$B2,$38,$3E,$1F,$C8,$FC
	.dc.b	$F2,$38,$3E,$3F,$E0,$1F,$4C,$8E,$0F,$97,$F0,$5F,$5C,$8E,$0F,$F3
	.dc.b	$47,$9E,$F9,$FF,$F6,$63,$63,$31,$83,$3E,$3C,$8E,$2D,$7F,$CF,$73
	.dc.b	$FE,$D7,$7F,$04,$D7,$F9,$83,$5F,$E9,$0D,$5F,$4A,$FA,$2F,$F6,$F3
	.dc.b	$63,$FC,$FC,$D7,$FD,$12,$D7,$BF,$C8,$FA,$88,$FB,$9F,$F7,$47,$E0
	.dc.b	$FF,$E0,$48,$FF,$44,$35,$FF,$49,$B5,$FC,$D3,$5F,$E9,$1F,$52,$FA
	.dc.b	$9F,$F7,$93,$63,$FD,$24,$D7,$FD,$3A,$D7,$F4,$8B,$EC,$B8,$92,$38
	.dc.b	$BD,$3C,$09,$1D,$1A,$FC,$0F,$7F,$F9,$06,$46,$98,$F7,$AC,$FC,$88
	.dc.b	$F9,$64,$70,$7F,$7E,$97,$CB,$F4,$38,$C7,$FB,$77,$E7,$97,$FC,$BD
	.dc.b	$FA,$5E,$3F,$F8,$25,$BE,$1F,$F6,$D7,$75,$FE,$0A,$6B,$CF,$83,$91
	.dc.b	$CE,$FF,$9D,$7C,$EF,$F1,$C9,$FF,$76,$37,$0F,$F0,$B3,$70,$FE,$F4
	.dc.b	$FF,$2C,$9D,$F1,$72,$23,$FE,$1A,$8D,$57,$F8,$7F,$37,$DB,$7C,$A2
	.dc.b	$57,$C3,$F6,$6B,$F8,$E8,$FC,$99,$1F,$C9,$BF,$4B,$8A,$47,$CB,$8F
	.dc.b	$EA,$9F,$38,$F6,$6F,$C1,$CA,$FF,$16,$CE,$5B,$39,$59,$7F,$8D,$65
	.dc.b	$5D,$2B,$FB,$07,$CE,$7C,$CA,$FF,$1D,$37,$6F,$F0,$F3,$70,$FF,$1B
	.dc.b	$3F,$59,$BB,$F2,$23,$FE,$40,$8D,$3F,$B4,$F3,$7F,$1E,$47,$F1,$1E
	.dc.b	$B5,$7F,$67,$1F,$B8,$23,$FB,$87,$F2,$FE,$74,$8F,$C8,$47,$F7,$9F
	.dc.b	$9B,$7C,$C5,$24,$FF,$94,$A1,$F3,$58,$8F,$8B,$9F,$F8,$EE,$78,$C6
	.dc.b	$EB,$A5,$7F,$81,$7E,$6F,$FA,$BE,$BF,$E5,$F7,$FB,$0D,$72,$3F,$CB
	.dc.b	$4E,$23,$5E,$7F,$13,$97,$A1,$2B,$FC,$CB,$5D,$32,$FF,$26,$7C,$EB
	.dc.b	$E7,$BF,$C2,$8D,$AF,$F3,$5B,$5F,$00,$7F,$38,$FF,$9A,$DD,$3F,$DA
	.dc.b	$DA,$F6,$8F,$ED,$CD,$29,$CD,$F0,$FC,$FC,$8F,$0C,$7F,$89,$FE,$77
	.dc.b	$E5,$C7,$F9,$F9,$AF,$91,$B9,$FF,$A1,$5F,$CB,$46,$74,$E4,$47,$97
	.dc.b	$27,$F8,$CF,$E7,$7E,$56,$7F,$E6,$F5,$F1,$79,$35,$B4,$BF,$F4,$7C
	.dc.b	$E9,$B3,$97,$E3,$EB,$F3,$DF,$8D,$BF,$CE,$3E,$7D,$D8,$F9,$F1,$5E
	.dc.b	$31,$74,$FF,$50,$57,$39,$FA,$55,$C9,$E2,$43,$C2,$E5,$A6,$67,$9C
	.dc.b	$65,$B1,$9C,$7F,$0B,$39,$6C,$64,$F5,$9F,$86,$BF,$34,$6A,$3B,$9E
	.dc.b	$A3,$81,$E8,$2D,$FD,$07,$99,$C2,$FC,$6F,$13,$13,$D6,$32,$7C,$4A
	.dc.b	$F0,$CF,$28,$CC,$FE,$7A,$73,$3E,$13,$99,$F9,$97,$35,$F3,$48,$E2
	.dc.b	$F2,$D9,$F4,$B5,$E6,$7E,$95,$C5,$F8,$A3,$B1,$F0,$27,$9B,$4E,$65
	.dc.b	$39,$D7,$C5,$7D,$13,$CD,$B7,$ED,$78,$46,$67,$81,$EB,$AE,$F3,$F5
	.dc.b	$0F,$DB,$3F,$B2,$7E,$21,$83,$F6,$B1,$FE,$63,$7E,$A9,$93,$F7,$6E
	.dc.b	$5F,$19,$FC,$0F,$D6,$FE,$16,$7C,$2B,$FE,$1E,$E2,$3C,$F1,$87,$F8
	.dc.b	$7C,$6D,$E4,$60,$D9,$83,$F5,$AE,$DF,$E2,$8F,$57,$E4,$D5,$BF,$D4
	.dc.b	$CE,$EF,$FC,$F1,$5C,$8E,$A7,$23,$C0,$72,$2C,$F2,$77,$1C,$A3,$11
	.dc.b	$B3,$A1,$1F,$A9,$5E,$3B,$F8,$13,$B9,$E2,$48,$FC,$F1,$23,$E0,$46
	.dc.b	$5F,$CF,$46,$5F,$41,$DD,$C5,$7E,$6E,$FC,$13,$C0,$87,$0A,$E8,$FD
	.dc.b	$A2,$4F,$B6,$BC,$F3,$EA,$9F,$87,$18,$3B,$DF,$68,$E9,$18,$76,$7A
	.dc.b	$9C,$3E,$73,$07,$92,$BD,$0F,$03,$C9,$C4,$FA,$CF,$EE,$44,$8F,$EE
	.dc.b	$1F,$35,$5F,$EA,$F3,$B7,$FB,$A4,$FF,$79,$9F,$ED,$DE,$BF,$F6,$FC
	.dc.b	$E5,$7F,$66,$8C,$BF,$DA,$23,$2F,$F6,$2E,$44,$FF,$BE,$FE,$CF,$FC
	.dc.b	$CD,$23,$FB,$3B,$86,$F9,$3F,$F0,$07,$9B,$FC,$03,$18,$3F,$B1,$CE
	.dc.b	$DF,$F6,$A3,$5E,$81,$C3,$DD,$ED,$BB,$6D,$C7,$F9,$F6,$31,$3B,$F0
	.dc.b	$3D,$18,$7F,$6D,$D2,$4F,$F6,$D7,$FB,$87,$46,$23,$FB,$74,$5B,$FE
	.dc.b	$DD,$3F,$DA,$75,$FE,$D5,$C3,$2F,$FB,$2C,$65,$FE,$CF,$19,$7F,$C1
	.dc.b	$7C,$9F,$8D,$7E,$47,$D4,$7F,$8E,$89,$E3,$7C,$0C,$7D,$32,$FB,$2F
	.dc.b	$93,$18,$AF,$F5,$64,$FC,$66,$BD,$2B,$F8,$F4,$62,$0F,$F9,$23,$EB
	.dc.b	$1E,$DF,$7E,$1F,$9F,$18,$AF,$B9,$87,$7C,$3F,$C8,$5F,$45,$27,$FC
	.dc.b	$80,$FF,$91,$7E,$73,$11,$F9,$28,$B5,$EA,$E8,$57,$F8,$FF,$DD,$FF
	.dc.b	$20,$67,$30,$FF,$1B,$C6,$5F,$F1,$CC,$65,$FC,$94,$7E,$47,$E2,$24
	.dc.b	$79,$B4,$FC,$0C,$7E,$3F,$8F,$33,$FE,$32,$E6,$4F,$F4,$16,$C7,$E4
	.dc.b	$A3,$07,$E7,$7D,$19,$1D,$4F,$C8,$47,$F8,$D7,$29,$1F,$E3,$57,$FC
	.dc.b	$71,$D7,$88,$FF,$1B,$45,$D7,$FC,$6D,$3F,$E3,$1D,$3F,$C6,$5D,$52
	.dc.b	$FF,$C5,$B1,$97,$FC,$5F,$19,$7F,$13,$1F,$5D,$FD,$87,$0B,$DE,$C4
	.dc.b	$23,$0F,$A4,$C4,$6B,$91,$B7,$0A,$DD,$7F,$96,$1A,$FE,$C8,$E2,$7F
	.dc.b	$96,$9C,$BF,$88,$AE,$5C,$12,$3C,$13,$D3,$3F,$CD,$FD,$1B,$E9,$3F
	.dc.b	$B9,$36,$BF,$CC,$AD,$7F,$CD,$2D,$7C,$4B,$E9,$38,$92,$38,$BF,$DF
	.dc.b	$7E,$C9,$EA,$52,$4F,$F9,$CA,$32,$FF,$9D,$3E,$91,$F1,$CF,$7B,$0D
	.dc.b	$9F,$F3,$DC,$BE,$39,$F1,$DA,$BF,$F3,$EF,$D9,$3E,$A3,$F7,$38,$7F
	.dc.b	$1A,$8C,$76,$6A,$73,$3F,$CA,$67,$2F,$FA,$23,$ED,$9E,$1E,$A1,$18
	.dc.b	$BF,$F4,$5B,$63,$84,$EF,$F1,$37,$A7,$49,$7F,$E9,$28,$CB,$73,$E0
	.dc.b	$47,$06,$EF,$A1,$1C,$1F,$E9,$CF,$7D,$EB,$F9,$DB,$19,$8C,$19,$89
	.dc.b	$1C,$5B,$3F,$EA,$85,$EF,$F3,$24,$F7,$7E,$27,$CF,$7D,$88,$FF,$56
	.dc.b	$B6,$BF,$D6,$2D,$7F,$34,$D7,$FA,$6B,$F6,$E3,$8F,$07,$42,$3D,$3E
	.dc.b	$67,$B9,$1C,$1F,$EB,$CF,$9B,$FE,$3A,$DF,$0F,$EC,$46,$30,$FF,$61
	.dc.b	$B5,$FD,$62,$F2,$36,$4A,$E8,$F2,$E9,$91,$DD,$B5,$FE,$CE,$F5,$7F
	.dc.b	$BA,$CF,$FB,$39,$B1,$9F,$13,$F1,$51,$88,$FD,$E3,$FD,$0A,$78,$78
	.dc.b	$9A,$23,$CF,$C1,$E2,$47,$07,$FB,$73,$E6,$3F,$96,$B7,$C7,$FC,$C4
	.dc.b	$63,$0F,$F7,$1B,$5F,$E1,$8B,$DB,$D9,$2B,$A3,$FA,$1F,$98,$FC,$BD
	.dc.b	$7F,$BD,$3D,$77,$C4,$7F,$7E,$F1,$BF,$B3,$D6,$9F,$ED,$33,$AC,$3F
	.dc.b	$91,$4F,$EE,$1B,$D7,$C5,$FF,$07,$C6,$70,$FF,$80,$2F,$38,$FE,$E9
	.dc.b	$F1,$FB,$9F,$F1,$05,$BE,$63,$10,$C5,$58,$AD,$5F,$F3,$39,$7C,$87
	.dc.b	$07,$71,$98,$C5,$74,$BF,$95,$5E,$CF,$17,$1B,$F5,$75,$B6,$31,$FE
	.dc.b	$81,$7F,$35,$B5,$CF,$E0,$37,$1A,$6C,$FA,$63,$FA,$06,$DE,$9E,$AF
	.dc.b	$13,$63,$AF,$7E,$1A,$78,$23,$D1,$4F,$98,$FF,$58,$34,$FF,$59,$38
	.dc.b	$3F,$A7,$3F,$E4,$BF,$95,$FF,$0A,$FC,$C6,$21,$18,$76,$D7,$FA,$B3
	.dc.b	$5F,$F2,$4A,$E6,$FA,$91,$C1,$AF,$06,$BC,$1A,$F8,$FC,$FF,$ED,$B3
	.dc.b	$FD,$6F,$1E,$B1,$8A,$EB,$F3,$0F,$53,$C3,$C7,$C0,$98,$FF,$73,$C7
	.dc.b	$93,$EF,$E4,$EA,$47,$FB,$C7,$CC,$7F,$71,$E5,$FF,$21,$D3,$E5,$7C
	.dc.b	$64,$AE,$97,$E6,$F9,$59,$1D,$DD,$3F,$74,$FF,$3C,$64,$E4,$FF,$04
	.dc.b	$B6,$3F,$C1,$0D,$78,$0F,$98,$7D,$27,$0E,$49,$3C,$1E,$77,$FE,$14
	.dc.b	$F9,$9F,$7C,$9A,$62,$DF,$25,$BE,$09,$5F,$E1,$9A,$FD,$5B,$78,$FF
	.dc.b	$C3,$FF,$32,$ED,$FF,$7E,$36,$7F,$10,$D7,$F1,$AD,$7F,$C3,$8E,$78
	.dc.b	$BF,$CD,$EF,$F7,$08,$FF,$14,$35,$F8,$2E,$F9,$BF,$D2,$AF,$A2,$3B
	.dc.b	$E5,$E2,$23,$E2,$F8,$2F,$29,$1C,$1E,$6F,$17,$D2,$47,$07,$AB,$E3
	.dc.b	$3D,$A4,$70,$7B,$BA,$7E,$09,$1C,$1F,$0B,$4F,$8A,$47,$07,$C6,$F8
	.dc.b	$7F,$92,$47,$07,$CA,$E6,$F9,$A4,$70,$7C,$EE,$7F,$A2,$47,$07,$D2
	.dc.b	$FC,$27,$EA,$91,$C1,$F5,$BE,$6B,$EC,$91,$C1,$F6,$BE,$6B,$EE,$91
	.dc.b	$C1,$F7,$BE,$B7,$F0,$24,$70,$7F,$09,$DD,$F3,$7F,$C2,$E3,$F1,$ED
	.dc.b	$8C,$C6,$E3,$1E,$4C,$8E,$4D,$7F,$17,$F1,$FF,$90,$23,$F9,$07,$B7
	.dc.b	$CB,$FF,$AC,$49,$F1,$38,$3C,$1F,$37,$2B,$CC,$D7,$F9,$1E,$3F,$E5
	.dc.b	$47,$D5,$E5,$F9,$C8,$FF,$25,$F9,$A7,$C8,$7B,$5B,$19,$8C,$19,$F9
	.dc.b	$09,$1C,$5A,$FF,$32,$FB,$3F,$DE,$BF,$E8,$96,$BF,$C0,$9B,$DB,$FE
	.dc.b	$D6,$FB,$07,$AB,$FE,$92,$9D,$D7,$FA,$25,$AF,$FB,$F1,$A8,$38,$6F
	.dc.b	$8D,$FE,$9C,$AF,$F4,$04,$7F,$A6,$DA,$F8,$E3,$FD,$43,$F6,$0D,$EF
	.dc.b	$C4,$4A,$FF,$51,$4E,$5B,$39,$7F,$8E,$C7,$FA,$11,$B1,$D3,$C7,$91
	.dc.b	$C9,$EF,$E9,$E2,$38,$C5,$D3,$FD,$F3,$7F,$E1,$D9,$FF,$59,$7C,$57
	.dc.b	$F3,$96,$F6,$7F,$AE,$23,$11,$FC,$B9,$76,$66,$7F,$31,$1A,$9F,$F5
	.dc.b	$A3,$5F,$94,$7F,$AB,$63,$67,$5E,$D9,$1E,$03,$FD,$89,$F6,$1F,$61
	.dc.b	$BF,$35,$FE,$CA,$83,$78,$A4,$E8,$87,$F4,$49,$E4,$FF,$46,$F4,$1F
	.dc.b	$41,$FE,$D4,$6C,$7F,$5C,$8E,$DE,$4E,$59,$19,$3C,$B1,$FE,$DC,$EC
	.dc.b	$7C,$1D,$24,$FF,$B7,$63,$FD,$22,$D8,$FD,$89,$1F,$F0,$54,$72,$6C
	.dc.b	$75,$F8,$39,$1C,$4F,$F7,$47,$2F,$47,$FB,$0B,$F8,$96,$F2,$FF,$BC
	.dc.b	$23,$5C,$6C,$7E,$9C,$61,$B3,$B3,$87,$C2,$7F,$BE,$3E,$7F,$FA,$56
	.dc.b	$1F,$DB,$9B,$87,$38,$FD,$34,$62,$FE,$FB,$FB,$26,$F7,$F3,$24,$FF
	.dc.b	$C0,$19,$D3,$FA,$FD,$1F,$3D,$F3,$49,$1F,$2F,$E7,$EB,$FE,$34,$9D
	.dc.b	$BF,$2A,$7F,$8E,$91,$FA,$5D,$FF,$5C,$E3,$97,$A6,$1F,$3F,$F8,$68
	.dc.b	$C3,$F8,$E8,$FE,$05,$18,$E5,$FC,$DB,$AE,$5B,$AF,$E7,$AF,$93,$6B
	.dc.b	$37,$FC,$2E,$7A,$3F,$E1,$87,$6F,$F0,$FF,$9F,$FA,$8D,$DF,$8A,$8C
	.dc.b	$B7,$83,$31,$82,$4E,$23,$F2,$DD,$92,$BF,$28,$FE,$67,$B7,$11,$FE
	.dc.b	$2A,$DF,$FC,$59,$AF,$C5,$6E,$9F,$BA,$7F,$84,$4E,$EF,$8D,$6D,$E9
	.dc.b	$1D,$88,$6C,$FD,$25,$F0,$BE,$93,$67,$F0,$53,$55,$F8,$9F,$C3,$4F
	.dc.b	$67,$A1,$2F,$C4,$F6,$3C,$2E,$1D,$5F,$AE,$FB,$5F,$59,$EF,$7D,$6F
	.dc.b	$81,$2F,$CF,$3D,$4E,$67,$90,$EE,$F5,$69,$93,$67,$B2,$FC,$35,$E2
	.dc.b	$6C,$70,$34,$5B,$F3,$12,$37,$1D,$23,$EA,$DF,$9E,$3C,$0D,$79,$9B
	.dc.b	$AF,$03,$D5,$AE,$4F,$05,$76,$9E,$55,$DA,$FE,$1D,$6C,$FC,$55,$6E
	.dc.b	$33,$5F,$8B,$8C,$4F,$92,$7E,$33,$F6,$8F,$E8,$E7,$D2,$3F,$92,$DF
	.dc.b	$F8,$60,$8E,$08,$7F,$03,$72,$FE,$6F,$D6,$FE,$6C,$78,$1F,$C5,$B8
	.dc.b	$7E,$01,$FC,$DD,$C3,$EB,$BF,$D1,$46,$1F,$83,$36,$BF,$49,$3D,$70
	.dc.b	$39,$BF,$71,$0F,$09,$DC,$F0,$9E,$73,$07,$86,$7A,$B8,$BF,$0B,$63
	.dc.b	$2E,$CF,$E5,$53,$E1,$7E,$51,$F8,$57,$15,$F8,$AB,$B1,$E9,$32,$FE
	.dc.b	$91,$B3,$F9,$16,$BC,$CF,$11,$ED,$9F,$CA,$C6,$AF,$F0,$ED,$BF,$C1
	.dc.b	$9E,$88,$F9,$B5,$F0,$9A,$FC,$33,$D8,$7C,$83,$0F,$C2,$6C,$73,$3A
	.dc.b	$BE,$39,$F0,$3F,$E3,$BF,$98,$7C,$DF,$F2,$C6,$C7,$F3,$EA,$FF,$0B
	.dc.b	$C6,$23,$FD,$9A,$BE,$7B,$85,$49,$E0,$75,$7A,$47,$F3,$8F,$9D,$FF
	.dc.b	$5B,$63,$FD,$3D,$FF,$5B,$57,$F9,$35,$A9,$DE,$1F,$E2,$86,$C6,$FD
	.dc.b	$7E,$87,$5F,$8F,$FE,$8B,$F3,$CF,$5A,$92,$BC,$A2,$3F,$BB,$3F,$E5
	.dc.b	$D7,$83,$FB,$03,$11,$FE,$93,$8C,$C6,$5A,$FE,$D9,$CE,$3F,$DA,$D7
	.dc.b	$A9,$E4,$3C,$53,$FE,$F6,$F9,$FF,$94,$F9,$1A,$E7,$4F,$EB,$F1,$FD
	.dc.b	$92,$0E,$AF,$FB,$C3,$74,$FF,$26,$B6,$7F,$CF,$0F,$1F,$9F,$F9,$A3
	.dc.b	$D1,$FE,$D6,$D9,$F8,$AE,$B5,$CC,$C9,$CC,$7C,$ED,$64,$7C,$CD,$3E
	.dc.b	$AB,$F3,$67,$2F,$C7,$9F,$94,$47,$E7,$1F,$4D,$FB,$3E,$D4,$8E,$91
	.dc.b	$F9,$5F,$9D,$91,$F6,$B8,$91,$F9,$9F,$9C,$CC,$FE,$5A,$2E,$DF,$E1
	.dc.b	$DA,$57,$E4,$FD,$CF,$CA,$63,$30,$FC,$2F,$1A,$75,$18,$9D,$2C,$3F
	.dc.b	$35,$3F,$E1,$4F,$9E,$7A,$74,$95,$F9,$0E,$4F,$D5,$5F,$F8,$BE,$B7
	.dc.b	$3F,$09,$AF,$E4,$5A,$FF,$42,$6B,$FE,$8A,$77,$1C,$FC,$1F,$11,$E0
	.dc.b	$9F,$C7,$67,$F8,$A2,$3F,$8F,$E7,$AE,$8F,$C9,$47,$E4,$23,$F0,$B9
	.dc.b	$FE,$1B,$79,$5F,$86,$8C,$BF,$88,$8C,$BF,$17,$97,$FD,$3F,$C8,$AF
	.dc.b	$F1,$CF,$CF,$FF,$9D,$73,$FF,$28,$36,$BF,$D4,$AD,$BF,$F6,$87,$CF
	.dc.b	$FC,$89,$CF,$77,$E2,$25,$5F,$E7,$F7,$FB,$3B,$5D,$9F,$8E,$7A,$2F
	.dc.b	$9F,$7C,$8F,$F9,$31,$B1,$FC,$1A,$77,$D1,$88,$FE,$01,$5F,$C0,$C8
	.dc.b	$FE,$6B,$E7,$DE,$9F,$95,$B1,$98,$C1,$9F,$93,$91,$C5,$AF,$F1,$1F
	.dc.b	$AF,$FC,$BB,$FA,$66,$B9,$E3,$F8,$51,$88,$FF,$03,$FC,$FB,$DD,$7C
	.dc.b	$FF,$EB,$A3,$0D,$F9,$27,$C9,$66,$B7,$7F,$C8,$7E,$79,$F3,$FF,$E8
	.dc.b	$58,$FF,$6C,$C6,$39,$3F,$72,$D8,$E1,$3B,$43,$EB,$3F,$6F,$5F,$C0
	.dc.b	$1B,$1F,$CB,$DA,$FF,$88,$1A,$FF,$BA,$3E,$7F,$FC,$17,$3F,$C2,$1B
	.dc.b	$1F,$CD,$DA,$FF,$8A,$1A,$F1,$C7,$C8,$A3,$E0,$CF,$39,$CC,$62,$7F
	.dc.b	$A1,$7D,$53,$DF,$E2,$DB,$FE,$32,$D9,$DC,$F6,$F6,$DE,$8D,$25,$FF
	.dc.b	$20,$8C,$B5,$E0,$47,$07,$F2,$6F,$53,$F9,$C3,$7C,$BF,$5C,$63,$0F
	.dc.b	$EB,$0D,$7F,$C9,$3F,$41,$F8,$3A,$FE,$5E,$D8,$FE,$BC,$D7,$7C,$1F
	.dc.b	$1A,$1C,$34,$95,$C9,$F9,$7F,$4E,$F8,$4A,$49,$FE,$7B,$19,$7F,$9F
	.dc.b	$FC,$FB,$E8,$3E,$9B,$63,$31,$83,$3F,$41,$23,$8B,$73,$FE,$ED,$F4
	.dc.b	$AF,$AD,$FB,$AD,$AF,$E9,$AD,$9D,$CF,$E0,$BD,$DF,$E7,$D5,$FD,$49
	.dc.b	$B3,$FE,$01,$6B,$FE,$72,$6B,$F9,$C0,$FF,$0A,$8E,$5E,$0F,$52,$3F
	.dc.b	$D7,$BE,$89,$F3,$5F,$92,$6C,$66,$30,$67,$E8,$24,$71,$6E,$1F,$E1
	.dc.b	$5F,$A2,$FD,$0F,$97,$FE,$F9,$F6,$1F,$24,$C3,$FC,$3F,$5A,$7D,$13
	.dc.b	$A7,$3D,$DB,$C3,$F9,$C1,$FB,$8A,$E0,$D8,$D6,$1B,$84,$E9,$FE,$C9
	.dc.b	$F3,$7A,$BE,$FF,$45,$67,$F8,$1D,$AB,$10,$C5,$58,$A7,$C7,$5B,$C3
	.dc.b	$70,$FF,$1D,$E0,$D3,$9E,$3F,$9D,$C3,$F8,$C4,$63,$AB,$BD,$3E,$0E
	.dc.b	$CE,$1B,$87,$4F,$83,$C3,$4E,$F1,$CB,$E0,$BB,$F6,$59,$FF,$26,$E7
	.dc.b	$FA,$87,$CB,$C1,$9C,$BE,$69,$FC,$91,$C6,$9F,$C9,$FE,$8B,$FA,$E4
	.dc.b	$9A,$47,$FC,$A9,$F4,$1F,$D6,$3E,$8B,$FC,$52,$D8,$CF,$3A,$72,$D8
	.dc.b	$E3,$F3,$BF,$E1,$96,$E1,$CA,$31,$1F,$ED,$0B,$46,$23,$F4,$10,$F9
	.dc.b	$DF,$D1,$57,$E0,$E3,$13,$FE,$6E,$6B,$F0,$1E,$71,$FD,$13,$E8,$BF
	.dc.b	$47,$3F,$D1,$A7,$F0,$8F,$1E,$5E,$BA,$F8,$1B,$1B,$C2,$AA,$FC,$3B
	.dc.b	$E1,$E5,$CC,$62,$3E,$8F,$C9,$FF,$BB,$32,$F4,$79,$3A,$24,$F1,$3D
	.dc.b	$5F,$45,$C8,$8F,$B6,$3C,$3D,$8E,$7A,$49,$FF,$47,$FD,$17,$F2,$57
	.dc.b	$FD,$23,$3F,$EE,$46,$F3,$F2,$38,$3F,$86,$FA,$2F,$93,$97,$EB,$DB
	.dc.b	$C1,$FC,$45,$B3,$FD,$BB,$0F,$F4,$DA,$8F,$96,$F3,$E0,$F8,$F1,$8B
	.dc.b	$E9,$C1,$F2,$BE,$8B,$FD,$62,$4E,$8F,$9B,$FB,$97,$E9,$F1,$FF,$54
	.dc.b	$CF,$ED,$DF,$F3,$CB,$5F,$F5,$43,$5F,$E4,$C1,$C3,$FC,$A6,$3E,$43
	.dc.b	$FE,$82,$6B,$FE,$B1,$6B,$C9,$AF,$F3,$18,$7C,$AF,$F9,$0D,$B1,$E4
	.dc.b	$8C,$19,$F8,$19,$1E,$86,$BD,$F6,$FF,$26,$5F,$D5,$8C,$1E,$58,$ED
	.dc.b	$F1,$F2,$3E,$0B,$5F,$83,$D1,$E3,$23,$F1,$9E,$BD,$92,$3E,$03,$5F
	.dc.b	$93,$F1,$1E,$B2,$3F,$39,$F7,$74,$C8,$FA,$2D,$7E,$41,$EA,$21,$F3
	.dc.b	$4F,$8D,$F3,$51,$FB,$2D,$7E,$CF,$D1,$FC,$B2,$3F,$79,$F9,$DE,$7C
	.dc.b	$8F,$C0,$B5,$FC,$0F,$D1,$7D,$32,$3F,$86,$7F,$CF,$1B,$BE,$8A,$92
	.dc.b	$7F,$D8,$71,$97,$FD,$8B,$F4,$FF,$DF,$5B,$E6,$FF,$A0,$63,$11,$FE
	.dc.b	$F7,$6B,$F9,$0E,$3F,$F6,$E3,$F9,$4F,$57,$F0,$C4,$7F,$DA,$3F,$2B
	.dc.b	$FE,$87,$7F,$CA,$ED,$4D,$31,$87,$F9,$BE,$3F,$C0,$4D,$7F,$CB,$ED
	.dc.b	$7F,$2B,$F0,$3F,$03,$CB,$91,$EE,$6B,$FA,$61,$F4,$4F,$8E,$FF,$4F
	.dc.b	$36,$3F,$DC,$ED,$78,$35,$FD,$7F,$DA,$7F,$A5,$FB,$7F,$BB,$37,$C1
	.dc.b	$FF,$0C,$36,$7F,$CE,$AF,$F1,$57,$FA,$AB,$FE,$9F,$79,$F9,$3F,$EA
	.dc.b	$F8,$C4,$72,$8C,$1F,$CC,$A3,$FD,$D1,$1F,$DF,$A7,$BF,$C9,$66,$BF
	.dc.b	$9B,$FD,$A3,$E9,$6F,$C4,$CC,$38,$1F,$E8,$89,$D3,$80,$E1,$FE,$D6
	.dc.b	$FF,$AB,$9B,$1D,$A7,$BE,$B2,$79,$57,$FA,$95,$BF,$57,$97,$9C,$DD
	.dc.b	$3F,$C3,$D4,$AF,$E9,$FF,$0B,$FD,$62,$FF,$BD,$B6,$BF,$C4,$8D,$D3
	.dc.b	$FC,$8C,$BE,$83,$8F,$95,$FE,$D9,$8E,$57,$FC,$99,$AF,$08,$D1,$E0
	.dc.b	$F9,$69,$1F,$0E,$7B,$3F,$EA,$EF,$B4,$FC,$FE,$3F,$E3,$55,$D9,$FD
	.dc.b	$99,$B3,$F2,$DD,$E5,$FE,$5B,$6C,$73,$9E,$07,$1E,$8E,$5F,$51,$27
	.dc.b	$93,$D7,$BF,$11,$FE,$1E,$B4,$E3,$1F,$25,$7F,$78,$FA,$8F,$C8,$47
	.dc.b	$F9,$2D,$B9,$7F,$9B,$5B,$87,$FB,$2B,$EA,$3F,$CD,$EF,$E0,$E3,$F3
	.dc.b	$0F,$F4,$E6,$BC,$7C,$1F,$D6,$3F,$ED,$28,$CB,$63,$FC,$13,$2F,$90
	.dc.b	$A7,$96,$3F,$C1,$BF,$50,$F9,$6B,$C7,$30,$FC,$A4,$E5,$B3,$97,$FC
	.dc.b	$7F,$1F,$81,$72,$70,$F4,$FA,$91,$E7,$EE,$57,$8C,$5D,$7F,$CE,$15
	.dc.b	$FD,$CE,$7F,$C9,$7D,$5F,$ED,$57,$F4,$CD,$73,$EA,$FF,$89,$A1,$E4
	.dc.b	$58,$F8,$B2,$9E,$0F,$F7,$86,$BF,$E9,$F6,$BF,$EF,$C7,$B5,$FF,$A0
	.dc.b	$DF,$F1,$73,$7C,$9E,$D3,$FB,$9F,$A8,$FB,$2F,$F9,$89,$7C,$57,$F8
	.dc.b	$E9,$B1,$BF,$84,$C1,$99,$91,$93,$F8,$27,$B4,$FA,$8F,$F2,$1B,$63
	.dc.b	$31,$D7,$9B,$2C,$8C,$9E,$58,$FF,$25,$F9,$4F,$8C,$A4,$9F,$E2,$51
	.dc.b	$FE,$FC,$6C,$65,$FF,$3E,$C7,$17,$7F,$1B,$9D,$7F,$D5,$13,$99,$DD
	.dc.b	$78,$2B,$57,$FE,$CD,$AD,$5F,$0A,$D5,$F8,$08,$72,$3B,$98,$8E,$0D
	.dc.b	$AE,$97,$DA,$3F,$AF,$D5,$7F,$AF,$C5,$BF,$0D,$78,$EB,$C3,$5D,$C9
	.dc.b	$7D,$9F,$83,$7C,$67,$93,$63,$C4,$7F,$6E,$70,$F2,$AF,$2B,$A3,$FB
	.dc.b	$8C,$F2,$71,$7E,$47,$D2,$7C,$03,$91,$82,$9E,$C3,$8C,$73,$73,$1C
	.dc.b	$DE,$E7,$3B,$F6,$12,$BE,$14,$65,$AF,$2A,$F8,$B3,$B8,$F1,$35,$F1
	.dc.b	$1B,$3D,$77,$F0,$A3,$57,$F8,$D9,$D1,$E6,$9D,$5F,$82,$BF,$17,$1F
	.dc.b	$91,$9F,$F3,$7D,$7E,$35,$B7,$FB,$E8,$C4,$7A,$0C,$9F,$E2,$28,$D1
	.dc.b	$1F,$CF,$FD,$47,$E9,$0F,$0B,$FE,$26,$70,$FC,$C3,$64,$3D,$37,$FA
	.dc.b	$08,$DC,$63,$0F,$CF,$46,$27,$E0,$D5,$9C,$D7,$53,$F5,$B5,$97,$E6
	.dc.b	$C7,$C3,$36,$E8,$E1,$5E,$17,$C8,$F6,$6C,$70,$8C,$47,$12,$38,$39
	.dc.b	$38,$8F,$03,$E2,$3E,$21,$D0,$C1,$F8,$C3,$81,$A3,$E1,$16,$3A,$9D
	.dc.b	$E3,$C8,$48,$E7,$7F,$97,$15,$F0,$E3,$B5,$7E,$2A,$31,$3B,$BB,$1F
	.dc.b	$6A,$BF,$CB,$AE,$A3,$F8,$03,$63,$F5,$ED,$7F,$4E,$DA,$ED,$1E,$F3
	.dc.b	$C3,$66,$73,$7E,$6E,$53,$99,$FF,$12,$73,$7F,$BE,$AB,$FB,$54,$2B
	.dc.b	$6F,$2C,$B4,$6E,$A4,$7F,$91,$6B,$B5,$F5,$FA,$8F,$F7,$19,$3C,$EF
	.dc.b	$06,$3D,$2D,$3F,$E0,$E6,$CE,$AB,$93,$63,$D9,$3D,$3A,$7F,$A3,$FD
	.dc.b	$3C,$9F,$E8,$AF,$F4,$AF,$A8,$C4,$7F,$9F,$77,$4F,$EC,$C9,$1F,$E9
	.dc.b	$27,$FA,$8F,$9F,$23,$99,$FE,$61,$FA,$8F,$ED,$3A,$7A,$5F,$EC,$F1
	.dc.b	$F1,$F4,$F9,$91,$89,$D7,$2F,$A6,$93,$CA,$B7,$CB,$98,$E1,$1C,$AB
	.dc.b	$87,$D4,$7C,$CB,$FB,$8D,$4D,$3E,$CE,$1F,$35,$C7,$17,$23,$9F,$17
	.dc.b	$F6,$EE,$D9,$7F,$DB,$1F,$EE,$3D,$D8,$8F,$91,$1F,$C6,$48,$FF,$AE
	.dc.b	$1F,$EE,$FE,$DC,$8E,$47,$FA,$1F,$E9,$5F,$47,$F0,$5B,$19,$8C,$7D
	.dc.b	$46,$59,$19,$39,$47,$C9,$FA,$67,$D2,$DF,$83,$95,$FE,$95,$9C,$B6
	.dc.b	$72,$FD,$53,$F4,$90,$9E,$93,$F5,$7E,$85,$F0,$54,$97,$FE,$A1,$87
	.dc.b	$66,$23,$FD,$91,$1C,$1F,$F2,$FB,$5C,$CF,$F8,$FA,$31,$C5,$FA,$15
	.dc.b	$F0,$1E,$13,$A4,$7E,$97,$EA,$1A,$7E,$25,$B8,$7F,$AC,$DA,$FB,$43
	.dc.b	$F8,$D7,$F8,$13,$A9,$E9,$3C,$CE,$9C,$BF,$95,$FA,$6F,$89,$7F,$EB
	.dc.b	$F7,$FD,$26,$9A,$69,$F0,$36,$BF,$78,$EB,$D1,$E9,$B7,$CD,$DA,$5F
	.dc.b	$CB,$7E,$84,$62,$3E,$5C,$7C,$72,$3F,$31,$FA,$DF,$0D,$23,$91,$FE
	.dc.b	$D5,$FA,$47,$AB,$FE,$6C,$6C,$7F,$A4,$9A,$FF,$BF,$DA,$FF,$23,$1B
	.dc.b	$FF,$BB,$9F,$E5,$4D,$73,$EE,$7D,$B3,$FC,$F3,$F4,$9F,$EE,$18,$E4
	.dc.b	$D7,$F9,$73,$5F,$E6,$8D,$5F,$25,$F8,$27,$FD,$4E,$D7,$C8,$6B,$8B
	.dc.b	$FD,$0F,$19,$8F,$47,$D2,$3E,$0F,$FD,$16,$D8,$FF,$58,$B5,$E6,$D7
	.dc.b	$E0,$7D,$23,$E0,$1F,$49,$FE,$B8,$8C,$37,$07,$C5,$66,$B7,$7E,$6F
	.dc.b	$97,$7C,$2F,$F3,$D8,$FE,$9D,$18,$F1,$BF,$D8,$6D,$8E,$13,$B7,$FA
	.dc.b	$C7,$D1,$BE,$A3,$10,$8C,$5F,$FB,$31,$B1,$F1,$9A,$FD,$BF,$A4,$78
	.dc.b	$9C,$9B,$5F,$ED,$46,$CF,$CE,$6B,$F8,$3F,$B0,$7C,$37,$89,$B5,$FE
	.dc.b	$DC,$6C,$FD,$66,$BF,$8B,$FA,$57,$BB,$EA,$6D,$7F,$B9,$1B,$3F,$79
	.dc.b	$AF,$E4,$FD,$87,$9B,$F0,$9B,$5F,$EE,$C6,$CF,$E1,$9A,$FE,$6F,$EB
	.dc.b	$FF,$DD,$13,$FE,$F2,$6C,$7E,$31,$AE,$FC,$EF,$9F,$E8,$BC,$3A,$4B
	.dc.b	$FF,$7D,$C6,$5F,$F7,$EF,$D3,$7E,$25,$BC,$BF,$62,$31,$87,$E6,$9A
	.dc.b	$FE,$BE,$DF,$57,$F7,$9B,$7F,$A0,$9F,$14,$E4,$FC,$1F,$D3,$BD,$9A
	.dc.b	$4A,$E1,$19,$79,$7D,$3B,$E9,$FF,$16,$D8,$CC,$60,$FC,$7F,$D7,$7E
	.dc.b	$D2,$3C,$4D,$C7,$F6,$6D,$7F,$84,$35,$FF,$16,$7D,$3B,$E5,$FF,$32
	.dc.b	$DA,$F6,$36,$7F,$7A,$D7,$F8,$97,$9F,$FA,$56,$F6,$FE,$8A,$31,$1F
	.dc.b	$E3,$96,$BF,$E3,$FF,$BA,$FF,$1E,$DF,$C9,$6C,$7F,$0A,$6B,$FC,$7D
	.dc.b	$AF,$F9,$0F,$EA,$3F,$32,$DE,$BF,$D8,$C6,$23,$FC,$98,$D5,$F4,$FF
	.dc.b	$AA,$F4,$BF,$C7,$9E,$8B,$EA,$3F,$C9,$31,$FE,$3F,$9D,$65,$FE,$51
	.dc.b	$AD,$3F,$E5,$79,$D7,$8B,$C2,$3F,$CA,$D5,$C3,$0F,$F2,$E3,$FC,$4A
	.dc.b	$38,$1A,$F0,$B8,$4F,$3F,$A1,$AF,$F7,$FF,$1A,$E2,$9C,$E4,$9C,$FC
	.dc.b	$56,$A6,$E3,$DB,$0E,$8D,$7F,$CD,$DF,$51,$CB,$B7,$BC,$23,$13,$B8
	.dc.b	$7B,$FA,$9D,$B6,$75,$EF,$E1,$A7,$18,$EB,$AF,$A3,$CB,$B3,$E3,$D7
	.dc.b	$F2,$DB,$7F,$31,$E0,$CE,$7F,$E0,$53,$E4,$17,$0F,$3A,$FA,$8F,$D8
	.dc.b	$3F,$E8,$96,$B9,$32,$FE,$D7,$E9,$9E,$A3,$EA,$30,$DA,$FE,$B3,$3A
	.dc.b	$9C,$9C,$7E,$8B,$FD,$1D,$3A,$6B,$FD,$CE,$51,$88,$FF,$4C,$46,$32
	.dc.b	$F1,$7D,$27,$F9,$B6,$0D,$31,$5D,$BE,$A3,$F9,$DD,$FF,$A9,$6B,$F0
	.dc.b	$4F,$DB,$6B,$FE,$2B,$6D,$FF,$A5,$DE,$B1,$BE,$5F,$01,$1F,$C2,$7C
	.dc.b	$87,$F8,$2E,$B8,$79,$FE,$52,$4F,$13,$C1,$F5,$1E,$92,$3E,$38,$FF
	.dc.b	$44,$76,$B9,$A9,$27,$F2,$5F,$51,$FE,$A9,$9F,$EB,$F1,$F9,$56,$BF
	.dc.b	$E4,$56,$E1,$FE,$B1,$7A,$47,$E6,$FE,$A3,$DA,$FF,$92,$DB,$3F,$EB
	.dc.b	$76,$F3,$7F,$B2,$6F,$FC,$C7,$5C,$7A,$FF,$D3,$D5,$F2,$2B,$FD,$9C
	.dc.b	$D9,$FF,$6A,$3C,$9F,$F6,$B7,$D4,$3E,$4B,$FD,$4C,$FF,$A6,$AB,$FD
	.dc.b	$AC,$D9,$DC,$EB,$D0,$FC,$BC,$FF,$B6,$27,$FC,$74,$D7,$8C,$6B,$D3
	.dc.b	$91,$FD,$36,$7F,$B6,$B6,$39,$FA,$72,$34,$D7,$FC,$67,$3F,$F0,$01
	.dc.b	$E5,$F2,$7E,$22,$3D,$7E,$C7,$CA,$47,$07,$87,$ED,$BD,$24,$70,$79
	.dc.b	$7C,$CF,$69,$1C,$1E,$9F,$94,$F8,$24,$70,$7B,$7E,$4F,$E2,$91,$C1
	.dc.b	$F0,$7E,$4F,$E4,$91,$C1,$F1,$7E,$8B,$E6,$91,$C1,$F2,$79,$7E,$89
	.dc.b	$1C,$1F,$37,$C7,$FA,$A4,$70,$7F,$7D,$FC,$27,$FB,$D3,$7D,$6F,$E2
	.dc.b	$31,$88,$FF,$30,$B5,$FF,$3B,$3F,$C1,$1E,$5E,$07,$02,$3F,$E0,$9F
	.dc.b	$AA,$6D,$FC,$71,$B1,$98,$C1,$FC,$83,$E4,$5F,$57,$49,$3F,$E1,$38
	.dc.b	$CB,$7D,$0E,$04,$70,$6F,$E7,$3A,$11,$C1,$FE,$1D,$FA,$AF,$F0,$33
	.dc.b	$78,$FF,$98,$C6,$23,$FD,$02,$D7,$FD,$2C,$B8,$F6,$48,$E8,$F2,$E3
	.dc.b	$91,$DD,$AF,$CC,$F1,$0F,$F5,$BD,$D1,$88,$FF,$66,$46,$1D,$77,$67
	.dc.b	$C0,$77,$6E,$B5,$5D,$4F,$BD,$7F,$70,$8F,$7F,$0B,$95,$71,$6C,$79
	.dc.b	$2B,$9D,$F6,$8F,$3F,$7F,$F0,$77,$FB,$93,$5F,$F7,$13,$63,$84,$7F
	.dc.b	$0D,$F8,$EE,$AF,$F0,$F9,$A6,$F0,$F8,$0D,$8E,$F3,$A7,$7E,$6C,$7C
	.dc.b	$26,$4F,$1B,$97,$C6,$78,$32,$FC,$4B,$FE,$F7,$6E,$1E,$66,$E1,$F2
	.dc.b	$FE,$1F,$87,$27,$F7,$49,$E3,$7F,$ED,$76,$BF,$5D,$AF,$DC,$79,$B7
	.dc.b	$8F,$C4,$74,$7F,$C3,$5F,$33,$FE,$21,$6F,$A9,$F1,$A3,$10,$CB,$E5
	.dc.b	$C3,$EB,$B5,$52,$B8,$9D,$23,$E7,$FC,$D7,$CB,$83,$FE,$F2,$BF,$AA
	.dc.b	$DD,$3F,$20,$EC,$9B,$72,$39,$EF,$F2,$FE,$12,$61,$FE,$33,$D5,$1F
	.dc.b	$3F,$3C,$CF,$FB,$AA,$72,$FF,$A3,$5A,$E6,$7E,$71,$F5,$FD,$B9,$1D
	.dc.b	$63,$F1,$3F,$34,$FA,$AB,$D7,$2B,$FC,$89,$39,$6C,$E5,$F9,$5A,$FE
	.dc.b	$B6,$55,$D2,$BF,$2B,$F3,$4F,$9A,$A4,$C3,$FC,$A1,$1F,$DC,$1E,$67
	.dc.b	$EE,$9E,$0E,$A3,$83,$A1,$1C,$78,$3F,$21,$F0,$7F,$95,$24,$FC,$BF
	.dc.b	$8E,$AF,$C9,$8F,$C9,$11,$F9,$4F,$E7,$7E,$B6,$47,$D3,$8F,$D7,$FC
	.dc.b	$DF,$BA,$3F,$CD,$0F,$FA,$15,$BC,$FA,$D2,$46,$08,$7A,$67,$93,$EB
	.dc.b	$F9,$AF,$56,$FF,$C5,$DB,$1F,$C7,$1A,$9F,$E9,$99,$FF,$04,$72,$4F
	.dc.b	$0F,$F8,$2B,$AA,$FD,$67,$C5,$7E,$E4,$97,$F8,$FF,$80,$AF,$EE,$E2
	.dc.b	$E1,$FB,$B9,$B5,$CA,$BF,$2D,$5F,$C0,$E7,$31,$FB,$78,$AF,$F3,$28
	.dc.b	$BC,$BF,$EE,$68,$E2,$78,$4A,$78,$4B,$7E,$16,$DF,$85,$D1,$FC,$46
	.dc.b	$BA,$DF,$66,$C7,$03,$F8,$53,$87,$8D,$75,$6D,$75,$8F,$E3,$55,$ED
	.dc.b	$9E,$47,$F4,$C7,$B3,$8B,$F6,$3C,$8E,$E6,$63,$6E,$4E,$47,$1B,$F2
	.dc.b	$57,$82,$3E,$21,$BA,$EB,$19,$8F,$80,$75,$39,$14,$E1,$3D,$2F,$D8
	.dc.b	$7F,$2D,$71,$1F,$5C,$EF,$3D,$49,$3C,$23,$D6,$78,$EF,$D7,$3A,$8E
	.dc.b	$47,$80,$FB,$67,$F3,$D8,$FE,$80,$47,$F8,$C3,$5F,$E3,$8D,$78,$1F
	.dc.b	$C8,$FE,$CF,$FA,$81,$E8,$7F,$BE,$B8,$7E,$04,$FF,$51,$8C,$47,$9C
	.dc.b	$E4,$43,$07,$F5,$97,$71,$EB,$8F,$AA,$64,$E0,$E7,$0F,$A7,$7F,$D5
	.dc.b	$EB,$8B,$F1,$EB,$8D,$78,$DE,$31,$9A,$E3,$1F,$D7,$6B,$8B,$F2,$AB
	.dc.b	$B4,$62,$39,$D7,$18,$C5,$72,$9C,$9C,$CE,$E7,$39,$E0,$E0,$D9,$21
	.dc.b	$3D,$2F,$F6,$ED,$9F,$CF,$35,$FC,$A3,$6F,$CF,$1D,$6B,$C8,$49,$D5
	.dc.b	$F8,$A7,$E6,$1F,$4A,$BE,$33,$87,$E7,$5F,$BA,$BF,$01,$7E,$6A,$ED
	.dc.b	$3E,$66,$9F,$D6,$39,$FF,$AE,$F5,$C8,$C1,$0F,$E9,$5C,$DF,$E0,$17
	.dc.b	$FA,$FC,$7F,$64,$9F,$EB,$3F,$45,$27,$FA,$BB,$FD,$73,$E5,$F1,$1F
	.dc.b	$D4,$F9,$93,$8D,$3F,$C6,$F0,$AF,$F1,$BF,$D8,$FF,$98,$E7,$FC,$76
	.dc.b	$DD,$BF,$CA,$CD,$C3,$FC,$B7,$4E,$BE,$3E,$4F,$F9,$6E,$65,$7F,$9A
	.dc.b	$DE,$31,$96,$94,$47,$56,$F9,$5E,$43,$AB,$FD,$73,$5F,$EB,$FD,$92
	.dc.b	$30,$43,$FA,$D3,$97,$FA,$E4,$7F,$50,$C6,$47,$F9,$D3,$3F,$EA,$D1
	.dc.b	$FD,$53,$B3,$33,$FD,$86,$2C,$FF,$5B,$9F,$F2,$EF,$CA,$7F,$BF,$2B
	.dc.b	$FC,$00,$D9,$FF,$31,$B7,$5F,$F3,$F7,$D8,$FC,$7A,$F0,$47,$FA,$06
	.dc.b	$79,$47,$C9,$6B,$9F,$73,$E3,$28,$CB,$FE,$8C,$7B,$3F,$E8,$EC,$9F
	.dc.b	$3D,$FE,$72,$6C,$78,$0E,$15,$FE,$2E,$E0,$FF,$17,$F8,$32,$7F,$C5
	.dc.b	$B1,$97,$FC,$5F,$19,$7F,$2B,$1F,$E2,$AF,$92,$91,$FE,$28,$7F,$C5
	.dc.b	$BF,$59,$88,$FC,$94,$5D,$7F,$23,$3F,$90,$F3,$A5,$79,$4F,$C1,$BF
	.dc.b	$E8,$AE,$07,$D0,$6A,$7F,$15,$C8,$AF,$8A,$D8,$FC,$63,$5D,$C7,$E7
	.dc.b	$BD,$09,$1D,$E7,$58,$FF,$A6,$BC,$1F,$F2,$BB,$7A,$BF,$C5,$A7,$F8
	.dc.b	$BB,$69,$7F,$E2,$E8,$CB,$FE,$30,$8F,$F0,$EE,$92,$3D,$9E,$02,$3F
	.dc.b	$C4,$3D,$79,$9F,$F1,$74,$5C,$3F,$C5,$13,$F6,$7E,$42,$57,$C5,$E3
	.dc.b	$47,$FA,$BB,$9F,$E2,$CE,$63,$F0,$F3,$A8,$FD,$73,$5F,$D9,$35,$FD
	.dc.b	$C3,$9B,$FF,$5D,$AF,$A0,$FC,$51,$E2,$9F,$F2,$F7,$D9,$BE,$97,$F2
	.dc.b	$AD,$8F,$DE,$35,$FE,$00,$D7,$F8,$32,$FB,$7E,$24,$8E,$2F,$4F,$36
	.dc.b	$47,$46,$BF,$C2,$FF,$16,$7D,$6F,$E9,$9B,$1F,$C3,$5A,$EE,$35,$F5
	.dc.b	$D2,$34,$D7,$A7,$A9,$CC,$8F,$37,$F6,$7F,$42,$F9,$54,$7F,$1B,$8C
	.dc.b	$37,$CC,$7A,$F9,$AD,$DF,$F2,$3F,$B4,$69,$FE,$EA,$81,$23,$3E,$2F
	.dc.b	$F2,$76,$C7,$19,$E0,$F0,$FB,$5F,$E0,$35,$FC,$B5,$B1,$FC,$C5,$AF
	.dc.b	$F3,$76,$BE,$01,$F8,$0B,$E2,$7A,$C6,$1B,$19,$AF,$E7,$4D,$9F,$E7
	.dc.b	$CB,$C6,$D9,$27,$6F,$4A,$7C,$5D,$24,$FF,$45,$8C,$BF,$D1,$FE,$81
	.dc.b	$EC,$79,$1B,$19,$8C,$19,$FA,$F9,$1C,$5B,$97,$F5,$3A,$7D,$9F,$32
	.dc.b	$57,$77,$E0,$FD,$0B,$EB,$29,$27,$FA,$D4,$65,$BF,$53,$C0,$8E,$0D
	.dc.b	$FB,$DD,$08,$E0,$FE,$C9,$E7,$BE,$77,$E8,$B6,$33,$18,$33,$E8,$C8
	.dc.b	$E2,$DD,$3F,$B8,$57,$EE,$E7,$AF,$83,$D0,$8F,$F7,$3F,$A0,$7C,$7F
	.dc.b	$E0,$1B,$19,$8C,$19,$F8,$F9,$1C,$5B,$87,$F7,$FF,$A0,$7C,$FA,$BF
	.dc.b	$9F,$68,$F8,$3C,$5F,$FB,$A6,$F7,$5A,$9C,$F9,$3A,$A5,$7E,$B2,$BF
	.dc.b	$A4,$3C,$23,$97,$7F,$09,$FC,$E4,$E6,$B9,$0F,$8B,$D5,$FE,$C3,$17
	.dc.b	$A3,$88,$56,$F0,$F4,$CE,$61,$F0,$37,$0C,$36,$BF,$AE,$5B,$E4,$3A
	.dc.b	$63,$FC,$26,$73,$18,$F0,$FC,$14,$FA,$0B,$38,$6D,$74,$F0,$30,$D3
	.dc.b	$BC,$72,$E0,$EF,$E2,$D9,$FA,$2F,$6E,$0C,$63,$EE,$9F,$F1,$D4,$63
	.dc.b	$6F,$E1,$BF,$41,$FE,$12,$A3,$48,$FF,$8F,$7E,$71,$F4,$0F,$A0,$FF
	.dc.b	$1F,$46,$1B,$19,$9E,$0D,$7F,$C9,$52,$F9,$4F,$A3,$1D,$5A,$E9,$AF
	.dc.b	$F1,$98,$E1,$87,$83,$E8,$3F,$8E,$3F,$89,$84,$63,$3E,$0D,$AF,$E6
	.dc.b	$33,$C9,$7D,$03,$E6,$3B,$73,$6E,$55,$F8,$B7,$FC,$34,$D7,$E6,$1D
	.dc.b	$2F,$84,$7E,$22,$37,$CD,$83,$A7,$D0,$7F,$9A,$F3,$F4,$7C,$B7,$02
	.dc.b	$57,$03,$BF,$D0,$74,$23,$E2,$8F,$36,$7F,$E4,$68,$F5,$73,$60,$F5
	.dc.b	$7D,$03,$E6,$F7,$CF,$2B,$FC,$69,$7F,$BA,$6D,$FC,$1F,$A0,$F9,$71
	.dc.b	$FE,$39,$6B,$F0,$03,$8C,$FF,$75,$8B,$E4,$F9,$70,$CB,$E4,$F5,$E1
	.dc.b	$F9,$DF,$3E,$F9,$7F,$F3,$1B,$6B,$FD,$14,$D7,$FD,$1C,$D7,$FD,$28
	.dc.b	$E7,$1F,$E1,$4F,$F5,$08,$FF,$28,$35,$F4,$3B,$E5,$FE,$C9,$FB,$9F
	.dc.b	$F6,$29,$34,$91,$E4,$FA,$19,$3E,$46,$BE,$4F,$4F,$D2,$47,$D2,$FB
	.dc.b	$3E,$02,$47,$B1,$AF,$B3,$E8,$3E,$09,$1F,$82,$FC,$4F,$93,$91,$F1
	.dc.b	$1A,$FC,$4F,$81,$F9,$24,$7E,$4B,$F3,$3E,$32,$47,$CC,$6B,$F3,$3E
	.dc.b	$37,$E8,$91,$FA,$2F,$D4,$F7,$64,$7D,$46,$BF,$53,$CD,$FB,$24,$7E
	.dc.b	$CB,$F7,$3A,$24,$7D,$C6,$BF,$EC,$9F,$C3,$3F,$D9,$8F,$F8,$99,$A9
	.dc.b	$A6,$3E,$7F,$F3,$91,$FE,$CF,$6B,$FE,$D3,$6B,$FE,$D9,$6B,$FE,$23
	.dc.b	$7F,$89,$3C,$7C,$3F,$F5,$9C,$62,$39,$47,$FA,$DB,$EE,$3F,$CF,$D1
	.dc.b	$FE,$DD,$6B,$FE,$E1,$6B,$FE,$E7,$6B,$FE,$31,$E9,$7D,$07,$FB,$01
	.dc.b	$B1,$FE,$E7,$6B,$FE,$EB,$6B,$FE,$F1,$7F,$B6,$3B,$F4,$79,$11,$E5
	.dc.b	$ED,$F5,$23,$83,$FD,$E9,$F4,$3F,$30,$9F,$33,$FE,$D2,$71,$C3,$49
	.dc.b	$3C,$1F,$F6,$BF,$A7,$FE,$89,$8F,$F5,$E3,$5F,$E0,$8D,$EC,$FE,$F9
	.dc.b	$F6,$9F,$EB,$0E,$BF,$F2,$94,$EE,$BF,$CA,$87,$F7,$4F,$63,$A9,$1C
	.dc.b	$9F,$E6,$88,$E9,$3F,$E5,$BA,$FF,$23,$CF,$FB,$35,$AF,$8A,$BF,$9D
	.dc.b	$7D,$A7,$F9,$CA,$3F,$DA,$4D,$7F,$8D,$37,$0F,$F0,$93,$FE,$99,$8D
	.dc.b	$F3,$72,$23,$CB,$A3,$AF,$A9,$DB,$9E,$4F,$68,$CB,$E0,$8C,$BE,$08
	.dc.b	$FE,$9D,$F4,$1F,$D2,$F1,$FF,$11,$37,$4F,$F3,$C3,$6B,$FD,$12,$BE
	.dc.b	$63,$8F,$8B,$F1,$A7,$2D,$53,$FD,$5D,$F9,$ED,$7A,$C7,$8B,$C5,$91
	.dc.b	$FE,$F7,$9F,$02,$FB,$4F,$F1,$4B,$FE,$FB,$6B,$FD,$0D,$B8,$FF,$8E
	.dc.b	$D7,$C8,$7F,$6B,$73,$97,$FA,$4D,$B5,$FE,$A4,$6B,$DE,$78,$7D,$0F
	.dc.b	$1F,$A1,$93,$E1,$8C,$BE,$38,$CB,$E3,$8F,$EF,$5D,$DF,$DC,$E3,$FA
	.dc.b	$C3,$71,$FF,$28,$37,$1F,$E0,$9A,$70,$F4,$BF,$81,$92,$FF,$87,$35
	.dc.b	$E7,$E8,$76,$38,$3F,$EB,$EC,$3F,$CB,$9E,$5F,$F4,$F8,$FF,$60,$B5
	.dc.b	$FE,$C0,$FF,$83,$CF,$C9,$61,$FE,$B7,$74,$43,$5C,$9B,$9E,$1E,$C4
	.dc.b	$7C,$15,$C5,$D1,$83,$C3,$1B,$7F,$BD,$57,$F8,$4E,$7F,$D8,$D8,$FF
	.dc.b	$12,$C3,$FC,$E8,$D8,$CF,$81,$F3,$3E,$37,$FC,$4D,$3F,$C5,$B0,$C9
	.dc.b	$F9,$37,$0F,$FB,$81,$AF,$F9,$B2,$78,$F8,$12,$3C,$C6,$4B,$F4,$9F
	.dc.b	$84,$27,$6F,$F8,$D5,$FC,$6C,$6A,$30,$47,$8F,$1F,$F8,$EF,$E9,$3E
	.dc.b	$81,$F3,$A3,$FD,$E6,$D4,$EB,$5E,$1B,$C9,$F2,$CF,$D9,$CF,$F8,$E4
	.dc.b	$7B,$7E,$6B,$E1,$9F,$13,$1B,$1E,$C9,$DB,$FE,$8C,$9D,$39,$76,$FF
	.dc.b	$8B,$63,$67,$E9,$0B,$1A,$DF,$EA,$D5,$DF,$EC,$CD,$BF,$A1,$35,$F0
	.dc.b	$1D,$CF,$23,$9B,$EB,$18,$34,$78,$E7,$53,$B6,$C6,$AF,$F0,$F1,$88
	.dc.b	$E0,$66,$FC,$AF,$53,$24,$9E,$2E,$5A,$E6,$38,$BF,$16,$BC,$4F,$53
	.dc.b	$DE,$53,$DE,$71,$32,$53,$33,$D8,$F9,$6F,$1B,$F8,$6F,$73,$17,$ED
	.dc.b	$9E,$17,$E4,$6C,$F6,$6B,$DA,$34,$7E,$BA,$FD,$31,$E3,$BF,$BD,$78
	.dc.b	$BF,$1C,$E6,$3C,$17,$CA,$3F,$20,$79,$EF,$F1,$2F,$F6,$CB,$57,$F4
	.dc.b	$36,$C7,$ED,$63,$04,$70,$43,$FB,$8C,$7E,$4B,$E1,$3F,$7E,$79,$6F
	.dc.b	$F8,$3B,$89,$F8,$A7,$E7,$DC,$3E,$93,$04,$30,$7E,$7D,$DE,$1F,$22
	.dc.b	$BC,$13,$B8,$C5,$EA,$FC,$33,$E0,$39,$1B,$7E,$33,$F4,$4F,$E2,$57
	.dc.b	$C5,$F9,$27,$D1,$70,$EC,$96,$8C,$BF,$84,$BF,$E4,$B1,$8B,$FD,$53
	.dc.b	$6B,$F1,$4D,$7B,$1E,$53,$C3,$3E,$F8,$D3,$63,$47,$8A,$7C,$33,$F4
	.dc.b	$63,$D6,$6A,$BC,$97,$F1,$23,$E1,$9F,$3A,$FC,$EE,$EF,$CE,$D3,$CE
	.dc.b	$53,$CE,$72,$8F,$CD,$FD,$1B,$E8,$BF,$DC,$CF,$F8,$3A,$7F,$9C,$B5
	.dc.b	$FF,$63,$3B,$73,$1C,$3A,$A4,$60,$FE,$83,$F4,$6F,$A3,$FF,$79,$36
	.dc.b	$33,$18,$F2,$72,$C8,$C9,$CA,$3F,$AA,$FD,$24,$8F,$E6,$4F,$F5,$8F
	.dc.b	$8B,$C4,$7E,$7A,$3F,$4C,$47,$FA,$D3,$FD,$7B,$E4,$64,$72,$3F,$AB
	.dc.b	$FD,$07,$BF,$56,$E9,$93,$FE,$78,$7F,$DD,$0D,$9D,$C6,$B4,$E0,$E9
	.dc.b	$B1,$A8,$4F,$1A,$DB,$63,$91,$FD,$AB,$E8,$3F,$DE,$4D,$CF,$9D,$76
	.dc.b	$9C,$CF,$FB,$ED,$FF,$54,$DF,$77,$2D,$AE,$33,$ED,$9C,$D7,$3F,$7F
	.dc.b	$FC,$07,$E1,$4A,$FE,$C0,$FF,$81,$BE,$1F,$11,$FC,$32,$3F,$8A,$11
	.dc.b	$FF,$04,$3F,$E0,$CF,$0E,$47,$23,$F1,$BC,$FF,$90,$EB,$91,$82,$1F
	.dc.b	$90,$72,$FE,$4A,$3F,$17,$F3,$F2,$3F,$1E,$FE,$3B,$C7,$C4,$36,$FA
	.dc.b	$1C,$AA,$CC,$62,$BE,$1D,$3E,$85,$E8,$E1,$B1,$FE,$1C,$6B,$D7,$7F
	.dc.b	$BF,$5F,$C3,$63,$55,$FE,$79,$AF,$F7,$09,$1C,$1D,$7D,$89,$3C,$0F
	.dc.b	$C9,$E9,$E4,$E3,$FD,$84,$7E,$43,$A7,$13,$23,$F3,$31,$B3,$0F,$E0
	.dc.b	$F1,$F6,$F1,$FF,$A3,$E3,$F1,$91,$8E,$9C,$B2,$32,$7E,$46,$3E,$2F
	.dc.b	$D2,$3E,$0A,$92,$7F,$C7,$FF,$43,$F0,$5F,$88,$7E,$A6,$78,$3F,$E9
	.dc.b	$D6,$BF,$EC,$46,$BF,$EC,$A6,$BF,$AA,$77,$C1,$FA,$9C,$DF,$25,$F6
	.dc.b	$1B,$5F,$E5,$0A,$DF,$0F,$EC,$BF,$05,$FF,$53,$4F,$EE,$DA,$9A,$63
	.dc.b	$E8,$5F,$01,$FE,$B1,$6C,$FF,$AD,$5A,$FF,$B7,$1A,$FF,$B8,$9F,$F6
	.dc.b	$83,$9F,$53,$91,$1F,$F5,$F7,$C1,$7E,$29,$FF,$74,$B5,$FE,$12,$D7
	.dc.b	$F8,$93,$57,$D0,$BE,$87,$FD,$8C,$D8,$FF,$65,$35,$FF,$78,$B5,$FF
	.dc.b	$7B,$AF,$A1,$E2,$48,$E2,$FF,$B4,$7E,$C1,$F3,$34,$93,$FE,$D7,$8C
	.dc.b	$BF,$ED,$9F,$A2,$7C,$B3,$D4,$C3,$67,$F9,$29,$FC,$B6,$5E,$23,$D5
	.dc.b	$DD,$FF,$B8,$FE,$89,$E9,$7F,$29,$C3,$F9,$6C,$63,$93,$53,$99,$FE
	.dc.b	$71,$39,$7B,$8F,$96,$7C,$26,$E3,$0D,$8C,$DF,$FB,$CD,$B3,$CB,$E8
	.dc.b	$5F,$19,$C5,$B5,$FE,$F9,$6C,$EE,$75,$DD,$2B,$4D,$9F,$2F,$A6,$FA
	.dc.b	$4F,$23,$6B,$D3,$3F,$E4,$29,$C9,$E9,$FA,$07,$A5,$49,$5C,$23,$2D
	.dc.b	$FA,$1E,$04,$70,$76,$FC,$0F,$FB,$03,$70,$F8,$D1,$87,$E5,$B5,$FB
	.dc.b	$7F,$42,$FA,$2F,$98,$DA,$F4,$B6,$7E,$9B,$5E,$3D,$3C,$09,$3C,$1F
	.dc.b	$AD,$F5,$6F,$36,$92,$7E,$2C,$65,$F8,$FE,$E3,$E8,$3E,$F3,$63,$31
	.dc.b	$83,$3E,$44,$8E,$2D,$C7,$F1,$3F,$46,$F9,$4F,$C4,$36,$BE,$AB,$67
	.dc.b	$73,$AF,$A3,$95,$A6,$CF,$E5,$3E,$8D,$F3,$9F,$E0,$AF,$3B,$FC,$0D
	.dc.b	$9B,$E1,$B1,$8F,$F8,$32,$F6,$FF,$BB,$A7,$4F,$FB,$C2,$3F,$54,$DB
	.dc.b	$E3,$87,$F8,$6E,$72,$FF,$36,$0F,$47,$18,$3E,$27,$8C,$7C,$BF,$6E
	.dc.b	$CE,$70,$F1,$46,$3E,$13,$37,$67,$FC,$55,$5F,$E2,$13,$85,$F8,$47
	.dc.b	$C6,$F6,$CF,$F8,$4D,$70,$3F,$83,$C6,$7C,$8F,$1D,$3E,$8E,$CE,$1B
	.dc.b	$5C,$67,$FC,$67,$DF,$98,$C4,$7F,$8F,$63,$0F,$8F,$6F,$17,$91,$D9
	.dc.b	$B5,$C2,$7E,$4E,$7F,$C4,$E7,$C7,$E4,$6E,$B5,$5F,$C3,$CF,$DC,$94
	.dc.b	$FE,$18,$63,$C4,$FD,$EF,$D0,$3E,$89,$F4,$78,$6D,$7F,$1A,$9D,$4E
	.dc.b	$4E,$3F,$33,$FE,$1F,$9D,$35,$FE,$5F,$28,$C4,$7F,$95,$A3,$19,$F8
	.dc.b	$BE,$4B,$FC,$47,$06,$98,$AE,$DF,$46,$F9,$4E,$5C,$BC,$09,$8F,$CE
	.dc.b	$C3,$93,$79,$BA,$91,$FA,$7F,$47,$FD,$13,$3F,$F3,$35,$3E,$4F,$C6
	.dc.b	$4A,$E4,$75,$FA,$3F,$39,$1F,$04,$7F,$8F,$32,$FF,$01,$C7,$F9,$ED
	.dc.b	$AF,$A7,$97,$FC,$83,$F4,$5F,$E6,$8E,$3F,$F3,$CD,$7E,$D2,$BF,$0E
	.dc.b	$D7,$F5,$0D,$CB,$FC,$F8,$F4,$FA,$3F,$7C,$FE,$B5,$B3,$FE,$81,$6D
	.dc.b	$7F,$A1,$AF,$FC,$E3,$5F,$E8,$F9,$FD,$83,$F9,$66,$BC,$F6,$FF,$2B
	.dc.b	$C7,$28,$FF,$4B,$7D,$1B,$E5,$7F,$CB,$6D,$8F,$CF,$B5,$FE,$01,$F4
	.dc.b	$7F,$D8,$9C,$CF,$E5,$A7,$51,$FD,$D9,$AF,$FA,$71,$AF,$FA,$96,$7F
	.dc.b	$BF,$BF,$D5,$9A,$FF,$70,$6B,$FE,$00,$6B,$FE,$A6,$7F,$B9,$26,$93
	.dc.b	$1F,$27,$BB,$27,$FB,$1D,$23,$11,$DF,$AF,$F5,$F7,$FE,$0C,$8F,$24
	.dc.b	$60,$F7,$7C,$04,$8E,$6D,$7F,$C0,$F6,$FA,$3F,$88,$48,$F0,$BD,$BE
	.dc.b	$42,$47,$C1,$6B,$CB,$E0,$7E,$61,$1F,$4B,$F3,$FE,$7E,$47,$B1,$AF
	.dc.b	$C9,$F9,$1F,$59,$1F,$9C,$FD,$7F,$76,$47,$C4,$6B,$EA,$FA,$3F,$B8
	.dc.b	$47,$E4,$BF,$C5,$3E,$A7,$DD,$1F,$E0,$E6,$BF,$E2,$86,$BF,$E3,$86
	.dc.b	$BF,$ED,$EF,$A4,$FF,$2A,$13,$E4,$7F,$8F,$BF,$E5,$7F,$9E,$7C,$B7
	.dc.b	$F2,$26,$D7,$F8,$71,$AF,$F8,$D9,$AF,$F9,$1B,$E9,$2B,$FE,$EC,$6B
	.dc.b	$FE,$F5,$6A,$69,$FE,$2B,$7F,$C5,$2D,$7F,$C8,$6D,$7B,$7B,$FF,$CD
	.dc.b	$BE,$61,$F3,$F4,$93,$FE,$32,$8C,$B7,$DD,$E0,$47,$06,$FE,$17,$A1
	.dc.b	$1C,$1F,$E3,$CF,$A4,$FF,$3E,$13,$5F,$E8,$EF,$F9,$FF,$EE,$9F,$49
	.dc.b	$FE,$D6,$6D,$7F,$5A,$6B,$FD,$EE,$7F,$95,$76,$A2,$4E,$9F,$F3,$8B
	.dc.b	$5D,$07,$0F,$0A,$FE,$B3,$1D,$BC,$25,$70,$8F,$F5,$8D,$7F,$82,$DB
	.dc.b	$3A,$8E,$47,$F7,$0F,$BA,$FF,$0E,$37,$D9,$FF,$78,$C6,$23,$FC,$28
	.dc.b	$FC,$E6,$C6,$88,$F3,$DB,$8F,$C2,$72,$F8,$69,$3D,$E3,$2F,$86,$32
	.dc.b	$F6,$8F,$F8,$03,$D0,$F7,$29,$27,$FC,$0F,$1F,$E9,$26,$C7,$F7,$C8
	.dc.b	$E1,$1F,$E7,$66,$BF,$E9,$06,$BF,$EA,$07,$39,$7C,$58,$FF,$4E,$FB
	.dc.b	$BE,$23,$C0,$FF,$8D,$7E,$E9,$F3,$BF,$5A,$B8,$46,$27,$FC,$36,$D7
	.dc.b	$87,$CC,$EF,$83,$EE,$CE,$6B,$8D,$E6,$FF,$0D,$39,$AD,$DE,$65,$2F
	.dc.b	$3F,$0F,$F7,$58,$F0,$F9,$F8,$99,$18,$21,$EB,$7A,$7B,$FF,$8B,$C3
	.dc.b	$F2,$2D,$9F,$CC,$35,$7D,$3B,$E2,$F8,$F3,$7F,$8B,$2B,$F2,$F4,$69
	.dc.b	$23,$D5,$5F,$EA,$69,$FF,$20,$B5,$FF,$31,$4F,$7F,$A7,$FB,$D1,$FE
	.dc.b	$65,$6E,$DF,$5D,$B8,$FE,$A1,$FF,$7F,$C6,$78,$3E,$44,$6A,$BF,$64
	.dc.b	$D4,$F0,$AE,$5F,$01,$C3,$D3,$95,$FE,$56,$AF,$F3,$2B,$67,$2F,$96
	.dc.b	$7B,$C7,$D7,$FA,$1F,$F0,$3E,$3F,$8D,$6E,$1F,$BE,$6D,$FF,$05,$5C
	.dc.b	$3C,$7D,$3F,$F2,$BC,$7F,$84,$27,$FD,$EE,$D7,$94,$68,$F0,$7B,$72
	.dc.b	$3F,$CD,$B3,$D9,$7D,$3F,$D5,$7F,$CE,$2D,$7F,$D2,$4D,$C7,$F4,$8B
	.dc.b	$E6,$FF,$24,$E7,$2F,$E2,$ED,$AF,$E3,$CD,$7B,$CF,$0F,$2F,$8F,$9D
	.dc.b	$27,$C3,$19,$7C,$71,$97,$C7,$1F,$A1,$F2,$1E,$DD,$24,$FF,$A2,$A3
	.dc.b	$DC,$69,$C4,$69,$FF,$57,$46,$A3,$19,$7F,$01,$8E,$13,$AD,$73,$9F
	.dc.b	$F3,$49,$CD,$6E,$FC,$95,$AB,$FE,$7D,$5A,$BE,$15,$A9,$F1,$90,$F1
	.dc.b	$B9,$69,$9B,$EC,$FB,$6F,$B2,$BE,$CF,$82,$B9,$57,$08,$E6,$47,$DD
	.dc.b	$5C,$AB,$8C,$F3,$AE,$37,$E8,$8F,$13,$5F,$88,$60,$F0,$14,$F2,$57
	.dc.b	$59,$EC,$D8,$D5,$7D,$16,$CF,$81,$E6,$79,$EB,$C1,$7E,$88,$D9,$CA
	.dc.b	$FA,$3B,$3E,$8B,$88,$F8,$47,$18,$D1,$EF,$39,$1D,$0C,$DF,$95,$B7
	.dc.b	$E5,$78,$18,$3E,$A1,$C6,$7B,$DF,$E1,$EF,$F2,$13,$FD,$F0,$E2,$5B
	.dc.b	$FC,$7B,$6F,$FC,$68,$47,$FC,$68,$D7,$FC,$68,$FE,$8F,$A2,$47,$86
	.dc.b	$FF,$2D,$18,$8F,$31,$FA,$07,$18,$79,$A3,$E7,$BF,$A1,$8C,$CF,$D0
	.dc.b	$24,$FE,$72,$32,$F0,$8C,$BF,$93,$8C,$BF,$1A,$7B,$39,$3F,$6D,$1F
	.dc.b	$56,$7C,$F1,$83,$D4,$77,$30,$7A,$A7,$CC,$E1,$EB,$7F,$9A,$3F,$B1
	.dc.b	$47,$BC,$E6,$78,$4C,$1C,$E7,$C2,$E2,$79,$DF,$AA,$B9,$C6,$27,$C1
	.dc.b	$5C,$23,$11,$C6,$30,$7C,$49,$F2,$5F,$3B,$FD,$83,$6B,$F4,$6D,$7F
	.dc.b	$22,$DB,$F6,$B6,$FA,$36,$7E,$91,$F1,$CF,$11,$83,$E9,$47,$91,$D4
	.dc.b	$FD,$4F,$5B,$FD,$A4,$FF,$88,$5B,$3F,$E5,$26,$CF,$F9,$65,$FE,$D0
	.dc.b	$61,$FF,$6F,$3F,$E1,$53,$06,$9C,$7C,$0C,$7F,$CC,$55,$DD,$E0,$64
	.dc.b	$3E,$38,$CC,$70,$AF,$F3,$2E,$1F,$E5,$3A,$FF,$1C,$36,$33,$E0,$FF
	.dc.b	$47,$F9,$3F,$F1,$44,$FF,$96,$6F,$F6,$4D,$76,$7E,$D4,$FF,$3B,$39
	.dc.b	$6C,$F2,$F0,$E4,$78,$CC,$94,$7D,$3F,$F6,$12,$75,$7F,$8E,$5F,$ED
	.dc.b	$51,$B7,$2D,$8D,$71,$7F,$8F,$FE,$9F,$E7,$1F,$E1,$72,$6D,$93,$5D
	.dc.b	$72,$FC,$53,$F5,$73,$FF,$1B,$47,$95,$AF,$CE,$D7,$E5,$F7,$7C,$6C
	.dc.b	$BE,$3B,$F0,$E7,$E1,$B5,$FE,$E2,$DA,$E7,$1F,$32,$3E,$7F,$C6,$48
	.dc.b	$F0,$47,$F9,$67,$E9,$DF,$45,$7A,$A5,$7C,$99,$CB,$67,$2F,$9C,$FF
	.dc.b	$55,$42,$7A,$4F,$F9,$97,$E9,$5F,$45,$49,$7F,$56,$1E,$8E,$23,$A0
	.dc.b	$8E,$2E,$E7,$A4,$FE,$2B,$E9,$A4,$FC,$B7,$F1,$BF,$4F,$88,$FC,$56
	.dc.b	$A9,$F8,$C4,$8F,$99,$19,$7E,$A7,$9B,$F9,$63,$97,$4F,$FA,$4F,$E9
	.dc.b	$BD,$D3,$F8,$D7,$FD,$69,$19,$E8,$FF,$42,$46,$A7,$D6,$D9,$D4,$71
	.dc.b	$9D,$ED,$FE,$8B,$FA,$6F,$87,$3F,$99,$6B,$F1,$1B,$5F,$EA,$55,$EC
	.dc.b	$71,$D6,$57,$17,$A4,$7E,$C3,$C2,$93,$F9,$17,$F6,$7F,$03,$88,$F8
	.dc.b	$F1,$F8,$B2,$3F,$19,$FD,$CF,$87,$23,$91,$FA,$FF,$C1,$BF,$CE,$6F
	.dc.b	$EE,$5A,$FF,$6F,$6B,$FE,$EC,$6B,$FE,$F3,$FA,$5F,$F3,$53,$7C,$5F
	.dc.b	$CF,$46,$20,$DF,$A5,$7C,$27,$E8,$DB,$3F,$A7,$6B,$FC,$2D,$AF,$F8
	.dc.b	$07,$EF,$9E,$E7,$EA,$DB,$1F,$AF,$6B,$FC,$4D,$AF,$F8,$27,$E1,$DE
	.dc.b	$E7,$EC,$DB,$1F,$B7,$6B,$FC,$6D,$AF,$F8,$47,$E9,$5F,$4D,$FB,$B6
	.dc.b	$C7,$EF,$DA,$FF,$23,$6B,$FE,$19,$E0,$FF,$4A,$BF,$C1,$5A,$FF,$29
	.dc.b	$6B,$BF,$43,$F8,$47,$D2,$BE,$7A,$92,$7F,$87,$46,$5F,$E2,$1F,$4A
	.dc.b	$F9,$97,$9F,$86,$CF,$F3,$79,$62,$F4,$35,$7F,$C6,$7E,$D5,$F1,$3F
	.dc.b	$EF,$DC,$04,$8C,$F9,$9B,$9D,$4E,$5A,$BE,$61,$F0,$D8,$6D,$7F,$27
	.dc.b	$6C,$FF,$4B,$6B,$FE,$41,$FA,$57,$D2,$F4,$6D,$7F,$2F,$6C,$FF,$53
	.dc.b	$6B,$FE,$49,$FB,$47,$A9,$E4,$6D,$7F,$37,$6C,$FF,$5B,$6B,$FE,$51
	.dc.b	$F4,$1F,$15,$EC,$6D,$7F,$3F,$6C,$FF,$63,$6B,$FE,$59,$FA,$67,$97
	.dc.b	$F1,$1B,$5F,$D1,$DB,$3F,$DA,$DA,$FF,$98,$7E,$C3,$EB,$CF,$F4,$D6
	.dc.b	$C7,$F7,$16,$BB,$E3,$FA,$3F,$4F,$F6,$2B,$FA,$A3,$67,$FB,$C3,$5F
	.dc.b	$F3,$73,$5F,$BE,$3E,$9F,$F6,$33,$FD,$75,$B1,$FE,$01,$6B,$FE,$77
	.dc.b	$6B,$CB,$EC,$FF,$29,$3F,$D9,$5B,$1F,$E0,$96,$BF,$E7,$F6,$BE,$2F
	.dc.b	$A7,$7D,$47,$E3,$1B,$5F,$DB,$DB,$3F,$E1,$36,$BF,$E8,$9F,$65,$F4
	.dc.b	$BF,$94,$6D,$7F,$77,$6C,$FF,$86,$DA,$FF,$A4,$7E,$CD,$E6,$7E,$71
	.dc.b	$B5,$FD,$FD,$B3,$FE,$23,$6B,$FE,$99,$FA,$87,$C7,$FE,$91,$B5,$FE
	.dc.b	$07,$6C,$FF,$8A,$DA,$FF,$A8,$7E,$A1,$F4,$EA,$FF,$14,$66,$F5,$F1
	.dc.b	$7F,$DD,$2F,$75,$A9,$CF,$AD,$AA,$57,$F0,$1A,$FF,$19,$3C,$23,$97
	.dc.b	$85,$C2,$7F,$07,$39,$AE,$43,$D1,$D5,$FF,$06,$C5,$F0,$18,$8F,$F2
	.dc.b	$AC,$D9,$CE,$1F,$E4,$1B,$7B,$5B,$7A,$46,$63,$15,$CE,$DF,$15,$CF
	.dc.b	$C8,$C3,$6F,$FC,$99,$2F,$A8,$CC,$61,$B1,$FE,$1A,$1F,$45,$D6,$3B
	.dc.b	$36,$79,$F0,$7A,$7A,$FC,$4D,$8E,$FC,$18,$69,$C0,$F1,$E1,$FC,$8C
	.dc.b	$FF,$22,$78,$BF,$E4,$57,$07,$F2,$33,$FC,$D9,$F4,$6F,$A8,$7D,$47
	.dc.b	$F9,$36,$30,$D8,$CC,$F0,$6B,$FE,$E9,$97,$D4,$7F,$A1,$A3,$AB,$5D
	.dc.b	$35,$FE,$03,$1C,$38,$BC,$1F,$21,$FC,$11,$FC,$54,$23,$19,$70,$6D
	.dc.b	$7F,$48,$9E,$4B,$EA,$1F,$25,$DB,$93,$72,$AF,$C6,$3F,$E8,$56,$BF
	.dc.b	$E5,$07,$C0,$E7,$97,$8C,$7E,$2A,$30,$7F,$A3,$7E,$A3,$F8,$DB,$C7
	.dc.b	$E4,$76,$47,$67,$5F,$97,$E6,$47,$C5,$1E,$6C,$7F,$0F,$1F,$E9,$96
	.dc.b	$BF,$E6,$46,$EB,$FA,$FF,$A8,$FF,$20,$CE,$F9,$39,$11,$E5,$CD,$F0
	.dc.b	$BE,$A3,$DB,$1F,$E6,$E6,$DF,$EE,$5B,$C3,$FC,$DB,$0F,$F2,$7B,$FC
	.dc.b	$E2,$3E,$3C,$EF,$7E,$87,$5E,$1F,$9F,$F5,$1F,$E8,$8C,$7F,$CF,$6F
	.dc.b	$FA,$B9,$BF,$27,$AF,$90,$DB,$FE,$9B,$9D,$4F,$CC,$6B,$FD,$21,$AF
	.dc.b	$F8,$21,$FE,$06,$FF,$A9,$5A,$FF,$B2,$5A,$F3,$EF,$ED,$E7,$C8,$E0
	.dc.b	$D7,$87,$8F,$E3,$23,$D5,$F3,$7D,$94,$8F,$03,$5F,$07,$C1,$FA,$C8
	.dc.b	$F9,$5F,$77,$BB,$23,$D0,$D7,$D1,$F1,$FF,$0C,$8F,$B5,$F8,$DF,$1F
	.dc.b	$23,$E0,$35,$F8,$1E,$A7,$CB,$23,$F1,$5F,$F6,$47,$D2,$3E,$A2,$92
	.dc.b	$7F,$D9,$D1,$97,$E6,$7D,$0F,$D5,$23,$83,$FD,$A9,$F4,$CF,$B3,$FF
	.dc.b	$6A,$36,$33,$18,$3F,$DB,$1F,$55,$FE,$42,$8F,$F6,$E3,$5F,$F8,$00
	.dc.b	$D7,$F5,$AD,$7F,$C1,$2F,$EF,$9E,$3D,$BD,$08,$F4,$F2,$FB,$91,$C1
	.dc.b	$FE,$EB,$FA,$8F,$F3,$5B,$FE,$14,$6B,$FE,$54,$6A,$68,$FA,$DF,$F7
	.dc.b	$93,$63,$A4,$60,$CF,$C2,$C8,$F0,$35,$FE,$12,$3E,$A7,$E4,$38,$6B
	.dc.b	$FC,$39,$AF,$F8,$A9,$AF,$F9,$A9,$FB,$0F,$1D,$BA,$B9,$6B,$DD,$B1
	.dc.b	$FE,$22,$8D,$B5,$F1,$7B,$3A,$3F,$13,$F1,$BF,$CB,$38,$3F,$D3,$BD
	.dc.b	$FF,$E0,$16,$F8,$3F,$E3,$E7,$94,$66,$78,$DF,$E3,$9A,$FC,$29,$E2
	.dc.b	$FF,$A1,$17,$D5,$77,$3C,$13,$FC,$F3,$E0,$DF,$57,$7E,$4A,$61,$FE
	.dc.b	$8C,$C7,$FC,$29,$5A,$8C,$65,$FE,$52,$8C,$42,$77,$5C,$A3,$15,$BB
	.dc.b	$F0,$79,$12,$79,$9F,$E9,$AC,$3F,$C5,$6E,$0F,$F4,$2F,$91,$B7,$99
	.dc.b	$B9,$C6,$3E,$69,$F9,$9C,$72,$B9,$6B,$FE,$EC,$AF,$9D,$C9,$99,$F9
	.dc.b	$15,$FB,$C2,$3F,$BA,$7E,$B7,$AB,$23,$9C,$7F,$64,$FA,$0F,$E6,$D7
	.dc.b	$FE,$39,$5F,$35,$93,$5E,$3C,$BF,$64,$E9,$B3,$97,$DA,$3E,$7F,$E6
	.dc.b	$DF,$F9,$15,$BC,$1F,$E8,$26,$CF,$FB,$09,$FF,$7B,$CE,$F8,$3F,$BC
	.dc.b	$C6,$A7,$A4,$63,$87,$FB,$EF,$D1,$7F,$62,$9F,$0B,$6F,$FC,$A8,$DC
	.dc.b	$3F,$29,$3C,$3E,$6B,$F4,$73,$C6,$3F,$BD,$B5,$FF,$4B,$B5,$FF,$6B
	.dc.b	$BC,$DB,$F3,$7E,$23,$A3,$F8,$1D,$7F,$DF,$38,$7F,$99,$5B,$87,$FA
	.dc.b	$89,$B3,$FE,$E2,$5E,$8F,$F8,$61,$CC,$1A,$48,$FF,$0E,$FE,$13,$FE
	.dc.b	$04,$AF,$F5,$6B,$FE,$55,$AF,$C9,$35,$FA,$AE,$DF,$F7,$8B,$53,$BB
	.dc.b	$FF,$3E,$36,$7A,$F2,$7D,$0F,$89,$97,$F9,$FC,$D1,$F5,$3E,$2F,$33
	.dc.b	$FC,$7A,$72,$FF,$14,$6B,$99,$FA,$67,$DC,$F8,$09,$1D,$63,$FC,$77
	.dc.b	$F8,$4F,$C6,$8F,$D7,$35,$FC,$8B,$73,$FF,$4A,$2F,$5B,$8E,$B2,$B8
	.dc.b	$BD,$23,$FC,$8F,$F6,$7E,$F9,$07,$F6,$5A,$FF,$B6,$9B,$1A,$78,$3B
	.dc.b	$E8,$E7,$D1,$23,$F8,$24,$EB,$5F,$F3,$4C,$7C,$9F,$73,$F3,$67,$E7
	.dc.b	$3D,$DB,$19,$21,$F9,$E7,$2F,$E8,$22,$EB,$FA,$09,$B9,$7E,$6A,$6D
	.dc.b	$7E,$6A,$6A,$BF,$1B,$CC,$FB,$91,$A8,$CD,$FE,$B2,$BC,$D5,$E4,$8C
	.dc.b	$36,$FF,$53,$7C,$6B,$93,$67,$47,$8C,$A7,$8E,$36,$E2,$FC,$35,$EB
	.dc.b	$8F,$29,$99,$DB,$E9,$25,$F9,$1F,$90,$6C,$E9,$7E,$37,$75,$E8,$74
	.dc.b	$60,$F3,$14,$F3,$1C,$8D,$CF,$73,$E9,$38,$8F,$43,$E2,$39,$5F,$B1
	.dc.b	$FE,$1B,$7F,$31,$CB,$EC,$AE,$95,$C9,$B3,$E4,$6B,$E4,$3A,$DF,$B6
	.dc.b	$33,$7F,$EC,$19,$FC,$59,$FC,$82,$FF,$91,$91,$FD,$B3,$5F,$DD,$35
	.dc.b	$E0,$7F,$01,$F5,$BF,$99,$9D,$F0,$FE,$52,$E2,$3E,$11,$FC,$A9,$C3
	.dc.b	$EE,$39,$10,$C1,$FC,$F1,$DD,$FA,$AA,$D7,$6A,$EA,$7E,$6E,$B2,$FC
	.dc.b	$C3,$F1,$4E,$1F,$8D,$5F,$12,$31,$7F,$4E,$73,$5D,$5F,$11,$DA,$3F
	.dc.b	$D4,$EF,$AC,$C4,$F2,$BF,$D4,$36,$3C,$27,$B6,$FD,$4D,$BE,$D1,$88
	.dc.b	$EE,$47,$07,$F5,$C9,$E9,$1F,$8A,$6C,$7A,$E3,$A1,$F0,$4D,$18,$3D
	.dc.b	$A6,$5E,$8E,$6F,$E1,$12,$7A,$D7,$E4,$4F,$39,$27,$93,$97,$E2,$1E
	.dc.b	$97,$0F,$F5,$BF,$17,$FA,$F1,$1F,$ED,$1E,$35,$7F,$C1,$7C,$C9,$C6
	.dc.b	$3F,$DA,$EB,$FB,$0F,$C0,$7F,$62,$F0,$66,$1F,$D8,$23,$2F,$F6,$28
	.dc.b	$CB,$FD,$8A,$3F,$9E,$BF,$E1,$8F,$B3,$7C,$8E,$30,$67,$A9,$CD,$7F
	.dc.b	$8F,$5A,$FF,$88,$3E,$5F,$FD,$A3,$5C,$23,$FA,$4C,$E9,$EE,$1C,$3D
	.dc.b	$6F,$FC,$EB,$18,$8E,$31,$E3,$BF,$17,$CD,$59,$EA,$D7,$A4,$FF,$66
	.dc.b	$CE,$47,$F6,$67,$FB,$4F,$CD,$62,$3F,$B3,$C5,$C7,$FB,$3C,$FF,$62
	.dc.b	$D3,$FB,$1F,$64,$BF,$EB,$31,$97,$FA,$DC,$65,$FF,$20,$F2,$7F,$98
	.dc.b	$9F,$F2,$1F,$A1,$FE,$82,$D3,$E7,$B6,$3F,$B6,$B5,$7D,$9F,$CA,$9D
	.dc.b	$F9,$9F,$EC,$C2,$4F,$FB,$41,$AF,$F9,$43,$DD,$78,$1F,$0D,$B1,$FE
	.dc.b	$73,$C3,$B7,$D2,$BE,$2F,$E4,$36,$7F,$CF,$0D,$7F,$CB,$6B,$E9,$F7
	.dc.b	$32,$76,$F2,$8F,$A5,$F4,$DF,$E4,$59,$FF,$15,$9F,$E3,$63,$3E,$3C
	.dc.b	$7F,$C6,$A6,$4F,$F4,$44,$71,$E6,$FF,$5D,$F1,$FE,$2E,$7F,$11,$E9
	.dc.b	$FD,$F9,$CE,$FF,$9A,$E8,$7B,$F9,$AF,$C5,$92,$FF,$D8,$77,$FE,$C6
	.dc.b	$7F,$26,$F2,$7F,$2F,$EB,$F5,$39,$3F,$E9,$BF,$A5,$F8,$4D,$D7,$F2
	.dc.b	$11,$88,$FF,$18,$F6,$E2,$64,$6C,$FF,$1B,$C7,$F8,$B9,$FF,$17,$F6
	.dc.b	$C7,$FC,$4A,$64,$FF,$13,$B9,$7F,$1A,$7D,$DF,$26,$3F,$EA,$FE,$44
	.dc.b	$7E,$8F,$9F,$E1,$CE,$63,$FD,$C3,$3A,$8F,$F7,$2B,$5F,$F7,$9B,$5F
	.dc.b	$DB,$B9,$E0,$FA,$4F,$F9,$61,$BE,$4F,$8A,$7F,$D4,$3F,$5A,$FA,$5F
	.dc.b	$F7,$7B,$63,$FD,$E6,$D7,$FE,$00,$B5,$FE,$0A,$BE,$53,$89,$23,$8B
	.dc.b	$D3,$92,$47,$46,$BF,$C3,$87,$D0,$BE,$33,$FE,$00,$B6,$34,$6A,$32
	.dc.b	$6B,$CC,$91,$A9,$FF,$35,$36,$39,$F9,$B2,$34,$D7,$C5,$5F,$49,$BC
	.dc.b	$FE,$07,$C8,$D8,$D4,$7F,$26,$7E,$23,$FC,$D9,$AF,$F9,$D1,$AF,$B7
	.dc.b	$3E,$9F,$66,$F0,$A9,$27,$DF,$19,$7E,$0F,$D0,$BE,$4F,$FA,$74,$7C
	.dc.b	$98,$C7,$36,$BE,$4D,$CF,$C2,$F7,$7F,$CF,$3E,$CD,$F5,$5F,$B7,$8F
	.dc.b	$E9,$B1,$8E,$2F,$9E,$D8,$FA,$CD,$7F,$A2,$FD,$A3,$ED,$39,$47,$FA
	.dc.b	$4A,$31,$5F,$61,$B1,$F7,$DA,$FF,$4E,$FA,$07,$C5,$F2,$84,$62,$BF
	.dc.b	$02,$D8,$FC,$43,$5E,$5F,$17,$C6,$0D,$24,$F2,$F8,$BD,$11,$C1,$F8
	.dc.b	$EF,$9F,$FE,$2A,$DE,$6F,$24,$63,$0F,$CC,$35,$FE,$CE,$BE,$3F,$64
	.dc.b	$AE,$8F,$2F,$8F,$91,$DD,$B5,$FA,$39,$76,$FC,$36,$D7,$E8,$9B,$3B
	.dc.b	$9D,$79,$D2,$B4,$D9,$E9,$DF,$CC,$93,$CD,$FA,$1F,$36,$F8,$9A,$49
	.dc.b	$FD,$8C,$65,$BF,$15,$C0,$8E,$0F,$DB,$FC,$DB,$D5,$FF,$07,$78,$BF
	.dc.b	$D1,$7C,$7F,$EB,$95,$FE,$D8,$8F,$E9,$15,$B8,$C1,$FE,$20,$9D,$E6
	.dc.b	$AD,$F5,$F4,$AF,$F1,$2D,$F1,$AD,$46,$27,$51,$FE,$E6,$9D,$57,$F5
	.dc.b	$71,$F2,$9A,$9F,$E6,$BF,$27,$67,$38,$79,$27,$FD,$DF,$8A,$B5,$E2
	.dc.b	$6F,$0B,$7B,$BF,$44,$75,$5F,$03,$DF,$0F,$DE,$56,$E7,$2F,$18,$CC
	.dc.b	$7B,$0F,$F2,$1F,$C1,$73,$AE,$8D,$AE,$5E,$2E,$1A,$75,$8F,$16,$DE
	.dc.b	$1F,$0A,$CF,$77,$CD,$BE,$31,$F0,$03,$F8,$6C,$F9,$EB,$D9,$0F,$9B
	.dc.b	$FE,$50,$D3,$F9,$53,$83,$F9,$DF,$CC,$BE,$11,$F3,$78,$6D,$7F,$45
	.dc.b	$9D,$4E,$4E,$3F,$0B,$F8,$39,$D3,$5F,$F2,$CC,$A3,$11,$FE,$6A,$8C
	.dc.b	$71,$F8,$BE,$13,$FC,$A9,$1F,$EA,$69,$34,$DA,$F9,$B7,$C7,$F3,$E4
	.dc.b	$E2,$49,$FC,$05,$71,$E5,$EA,$48,$FC,$27,$CD,$FF,$9B,$B2,$FF,$14
	.dc.b	$57,$F5,$89,$F9,$ED,$8C,$FA,$3C,$67,$3E,$8F,$A0,$8F,$82,$3F,$CD
	.dc.b	$D9,$7F,$49,$8F,$B2,$DA,$FF,$1B,$37,$8B,$FD,$1B,$F3,$7F,$E4,$38
	.dc.b	$FF,$32,$4E,$B8,$F9,$9D,$39,$FF,$CF,$5F,$37,$F1,$C9,$C3,$FE,$7D
	.dc.b	$C1,$F1,$BA,$E1,$9A,$7E,$8E,$FF,$5E,$DB,$FF,$42,$7C,$DB,$E2,$FF
	.dc.b	$B8,$36,$7F,$C0,$2D,$7F,$C1,$0D,$7F,$D2,$EE,$74,$FF,$29,$2F,$67
	.dc.b	$A9,$DA,$BF,$D2,$5F,$07,$FD,$92,$FF,$D3,$ED,$7F,$D5,$ED,$7F,$40
	.dc.b	$D5,$5F,$C0,$5F,$2F,$81,$E8,$23,$E8,$F2,$FD,$64,$70,$7B,$3D,$CF
	.dc.b	$79,$1C,$1F,$03,$CB,$F8,$64,$70,$7C,$4F,$9B,$F8,$E4,$70,$7C,$8C
	.dc.b	$7E,$59,$1C,$1F,$33,$CF,$F9,$E4,$70,$7D,$0F,$3F,$E9,$91,$C1,$F5
	.dc.b	$3E,$4F,$EB,$91,$C1,$FE,$38,$F9,$5F,$F3,$D3,$FE,$DA,$6B,$FC,$2D
	.dc.b	$A9,$BA,$BE,$6B,$F0,$64,$9F,$C1,$BF,$E3,$5F,$9A,$FF,$0D,$C7,$F9
	.dc.b	$19,$AF,$F9,$35,$AF,$FB,$71,$AF,$FB,$B1,$7D,$27,$32,$47,$37,$B7
	.dc.b	$5C,$8E,$CD,$7F,$CB,$5F,$37,$FD,$E1,$FF,$2B,$B5,$CF,$A7,$FE,$4B
	.dc.b	$8C,$47,$FB,$B9,$7C,$7E,$49,$1C,$9E,$1F,$21,$23,$AB,$5F,$F3,$57
	.dc.b	$67,$C7,$7F,$CC,$ED,$73,$9F,$F9,$6E,$31,$1F,$EF,$EF,$9B,$7C,$47
	.dc.b	$F9,$7D,$B1,$FE,$70,$6B,$FE,$77,$6B,$C7,$E2,$B8,$11,$E0,$FF,$9A
	.dc.b	$7E,$3F,$FD,$71,$06,$98,$FA,$FF,$F7,$54,$7F,$BB,$A7,$81,$1F,$F4
	.dc.b	$0C,$FE,$18,$E1,$C9,$FD,$4E,$3F,$AA,$D2,$31,$DD,$FE,$D3,$AF,$0F
	.dc.b	$D7,$BE,$67,$3E,$3F,$1C,$3F,$C0,$4D,$33,$8F,$F5,$FA,$E0,$FF,$7B
	.dc.b	$6B,$E0,$8D,$A9,$E6,$F1,$AC,$D7,$E3,$A3,$97,$26,$0C,$FD,$95,$8C
	.dc.b	$90,$FE,$E3,$28,$EF,$D3,$FD,$76,$7F,$C1,$CD,$7F,$BB,$B7,$3F,$F7
	.dc.b	$A2,$FA,$FE,$3E,$87,$FA,$76,$38,$57,$CB,$0E,$1F,$9C,$D7,$FC,$04
	.dc.b	$F5,$D3,$3E,$97,$90,$EB,$3F,$E2,$1F,$B0,$FE,$69,$3F,$5D,$EB,$19
	.dc.b	$AF,$93,$1A,$8F,$F0,$73,$53,$C1,$FF,$09,$B7,$15,$75,$F1,$79,$FA
	.dc.b	$D2,$FC,$4F,$6F,$88,$C4,$72,$8B,$7C,$A3,$C3,$3F,$73,$C0,$FF,$32
	.dc.b	$47,$F8,$ED,$AF,$F8,$91,$B3,$C6,$3F,$8D,$7C,$B7,$1F,$7F,$FD,$95
	.dc.b	$5C,$27,$E7,$35,$FE,$49,$1A,$3C,$1E,$F4,$8F,$04,$7E,$3B,$E8,$1F
	.dc.b	$3D,$79,$A5,$7F,$93,$E7,$2D,$9C,$BF,$E3,$A8,$FC,$CE,$5E,$77,$71
	.dc.b	$9F,$4F,$B1,$1E,$9E,$9D,$74,$77,$8F,$24,$7C,$7F,$7A,$4F,$C7,$7E
	.dc.b	$57,$8D,$88,$FF,$80,$1A,$FF,$98,$9F,$E8,$D3,$F1,$88,$FC,$FF,$7B
	.dc.b	$E9,$1C,$72,$F9,$DF,$41,$F8,$53,$F0,$5C,$9F,$C0,$DB,$5F,$E5,$77
	.dc.b	$5C,$9F,$B6,$73,$06,$9F,$8F,$FA,$0F,$87,$7F,$6D,$B7,$FC,$3D,$B5
	.dc.b	$FE,$63,$7E,$5C,$67,$7F,$E0,$71,$A9,$34,$DB,$FE,$6B,$F9,$E7,$CD
	.dc.b	$7F,$0A,$6C,$7F,$A2,$5A,$FE,$3C,$D4,$47,$4F,$E5,$DC,$1F,$E7,$88
	.dc.b	$E3,$87,$E5,$5B,$87,$F7,$39,$DC,$E6,$AC,$E6,$7A,$90,$E9,$1D,$DC
	.dc.b	$4F,$16,$CF,$17,$66,$0F,$81,$3C,$5C,$92,$79,$3E,$F3,$80,$AF,$4B
	.dc.b	$C8,$FD,$A1,$E8,$30,$73,$AF,$14,$F4,$23,$FD,$26,$79,$CF,$F8,$44
	.dc.b	$E4,$77,$AB,$AF,$F9,$DE,$B8,$46,$AB,$8C,$E6,$BC,$37,$D9,$B7,$E2
	.dc.b	$38,$BE,$CA,$E2,$F3,$BE,$ED,$3E,$0D,$F5,$7C,$87,$18,$F9,$57,$DE
	.dc.b	$32,$7C,$13,$FC,$5C,$73,$30,$7B,$CE,$47,$63,$4F,$39,$F0,$57,$C6
	.dc.b	$25,$78,$2F,$D7,$7D,$27,$D6,$FE,$1C,$FE,$98,$FE,$69,$B3,$FD,$3D
	.dc.b	$E2,$60,$C1,$EA,$30,$6E,$39,$3E,$0B,$FC,$E7,$8D,$FA,$93,$D3,$7F
	.dc.b	$75,$C4,$FB,$8F,$D2,$38,$7D,$17,$6B,$9C,$6D,$AF,$03,$15,$D5,$FD
	.dc.b	$B5,$78,$1F,$DE,$3C,$E3,$CC,$48,$C9,$C0,$E6,$60,$EC,$79,$0D,$18
	.dc.b	$9F,$BD,$7F,$4E,$BB,$4F,$E3,$88,$7E,$9D,$CB,$F8,$49,$E3,$18,$9F
	.dc.b	$B7,$3D,$A7,$DE,$76,$BD,$57,$B6,$7A,$1B,$BF,$63,$D5,$F8,$8F,$9D
	.dc.b	$CD,$F9,$E3,$81,$E6,$70,$EC,$C5,$78,$1F,$8A,$78,$0D,$1D,$CE,$55
	.dc.b	$EB,$8F,$DA,$36,$3F,$5C,$D5,$7F,$DD,$F4,$FF,$38,$C0,$CE,$1F,$E4
	.dc.b	$D8,$CC,$60,$D3,$89,$E5,$ED,$FF,$90,$A3,$15,$C7,$CA,$C1,$99,$FF
	.dc.b	$23,$FD,$07,$FB,$C7,$0F,$02,$E0,$C9,$AE,$39,$3E,$5C,$BF,$C1,$D5
	.dc.b	$B8,$C6,$3F,$E6,$38,$C4,$79,$C3,$DD,$FE,$55,$F0,$FF,$D3,$C8,$FF
	.dc.b	$4F,$F8,$8A,$FE,$BE,$3F,$66,$F0,$3F,$6B,$BA,$32,$D8,$FE,$B1,$F0
	.dc.b	$DF,$D6,$CE,$78,$FF,$89,$7E,$81,$AF,$F9,$A1,$B5,$EA,$3E,$93,$87
	.dc.b	$D4,$BE,$03,$84,$C9,$E0,$27,$85,$77,$AF,$F3,$AF,$CF,$7C,$48,$FF
	.dc.b	$1B,$3F,$E1,$1A,$F8,$CD,$8C,$EF,$C0,$E3,$D1,$2B,$BD,$73,$DF,$FA
	.dc.b	$57,$8B,$FD,$E4,$93,$FD,$E7,$C5,$AF,$F1,$98,$F8,$E4,$7E,$5B,$FE
	.dc.b	$01,$F7,$A4,$7F,$80,$63,$F1,$9F,$27,$23,$DF,$CC,$8F,$C8,$7A,$79
	.dc.b	$9F,$C0,$4E,$5D,$CE,$9F,$8F,$CA,$F8,$9F,$C9,$F8,$12,$73,$1F,$96
	.dc.b	$30,$FE,$2E,$53,$98,$FA,$3F,$36,$F9,$FF,$F5,$1B,$63,$31,$FE,$A6
	.dc.b	$F9,$DF,$F5,$FC,$27,$4F,$F4,$36,$BF,$E9,$36,$BF,$EB,$F6,$BF,$B7
	.dc.b	$78,$F1,$FF,$9C,$57,$D0,$78,$8F,$1D,$FE,$3F,$59,$7F,$88,$E6,$47
	.dc.b	$E4,$FB,$73,$3F,$90,$8B,$8F,$E6,$67,$F1,$79,$FE,$33,$B6,$5F,$E1
	.dc.b	$23,$2F,$E1,$A3,$2F,$E3,$A3,$FD,$8F,$E9,$BC,$AA,$49,$FC,$DF,$CE
	.dc.b	$FC,$69,$D4,$7F,$B8,$A7,$FD,$D8,$47,$07,$E8,$BE,$15,$E0,$FF,$B5
	.dc.b	$1B,$1F,$A6,$AE,$6F,$F0,$5F,$AE,$74,$7E,$F9,$B1,$FD,$C9,$AF,$F7
	.dc.b	$76,$BF,$EC,$7F,$75,$F3,$9F,$C1,$9B,$1F,$DE,$9A,$FF,$7F,$6B,$C7
	.dc.b	$E7,$B8,$11,$E0,$FF,$10,$E4,$F3,$C7,$FB,$55,$AF,$FC,$01,$6B,$FC
	.dc.b	$E5,$AB,$E7,$9F,$1F,$FC,$69,$B1,$FE,$0E,$6B,$FE,$13,$6B,$FE,$E3
	.dc.b	$7F,$88,$3B,$F0,$F9,$11,$FE,$47,$3F,$C8,$88,$E0,$FF,$0D,$AF,$89
	.dc.b	$FE,$56,$D8,$CC,$60,$CE,$52,$38,$B5,$FF,$14,$7D,$63,$E6,$7F,$9A
	.dc.b	$B6,$3F,$C5,$2D,$77,$1F,$CE,$BE,$79,$EB,$FF,$05,$9F,$F8,$01,$18
	.dc.b	$F0,$7F,$C6,$2B,$D8,$FE,$0D,$3F,$E3,$78,$D3,$FE,$39,$F9,$D7,$CF
	.dc.b	$7F,$B8,$63,$FB,$50,$8C,$F9,$1F,$E3,$C6,$C7,$19,$E0,$FF,$71,$D5
	.dc.b	$C7,$88,$F9,$0D,$8C,$DF,$F9,$19,$B3,$FE,$4B,$F9,$CF,$89,$3F,$E4
	.dc.b	$A6,$C6,$75,$E9,$18,$8F,$38,$5F,$17,$A2,$57,$37,$C9,$F5,$1F,$CA
	.dc.b	$63,$FC,$AE,$DA,$FF,$2E,$35,$EA,$D7,$E2,$FC,$EF,$E0,$E7,$FC,$BE
	.dc.b	$D8,$FF,$32,$35,$F2,$B5,$E3,$F3,$BF,$16,$0D,$24,$FF,$9A,$BE,$9D
	.dc.b	$F1,$3F,$2D,$B1,$98,$C1,$F3,$E7,$E5,$DF,$C4,$6B,$F7,$1A,$FF,$89
	.dc.b	$DA,$F8,$BB,$F9,$12,$BF,$CF,$73,$FC,$51,$FF,$3E,$FC,$F7,$F2,$16
	.dc.b	$F8,$3F,$7E,$31,$7F,$E8,$56,$BF,$59,$7C,$0E,$C9,$5D,$1F,$C4,$7C
	.dc.b	$F7,$F5,$68,$FF,$02,$FA,$EF,$92,$47,$F8,$22,$75,$87,$F8,$9E,$B5
	.dc.b	$39,$F4,$77,$3F,$E2,$BA,$DF,$97,$88,$47,$E0,$63,$0E,$AB,$F4,$70
	.dc.b	$F9,$8E,$6F,$1C,$1F,$2F,$8A,$DD,$5C,$3F,$CA,$95,$A9,$FF,$0F,$DF
	.dc.b	$F1,$C6,$D6,$EB,$FC,$BB,$E0,$DA,$F0,$57,$F8,$9A,$5E,$C6,$F0,$FE
	.dc.b	$07,$39,$8C,$78,$1C,$EB,$FC,$A9,$39,$39,$7B,$39,$AD,$79,$DB,$8B
	.dc.b	$3A,$3C,$1E,$77,$7F,$0B,$A3,$63,$C9,$C1,$86,$9C,$0F,$46,$2F,$9E
	.dc.b	$FF,$29,$B4,$FF,$2A,$B8,$3F,$D9,$3F,$36,$F6,$9F,$3D,$86,$D7,$FB
	.dc.b	$46,$75,$39,$38,$FB,$7F,$E8,$69,$D3,$5F,$E0,$F2,$8C,$47,$F2,$08
	.dc.b	$C7,$0F,$8B,$D0,$FC,$0E,$59,$6D,$7F,$A3,$EB,$FB,$6D,$7E,$2D,$E3
	.dc.b	$0F,$9E,$7C,$3F,$5E,$6E,$32,$AF,$C6,$3F,$E6,$26,$BF,$E8,$17,$AE
	.dc.b	$3E,$5E,$6E,$31,$FD,$83,$1F,$04,$74,$F9,$EF,$E4,$71,$E3,$F3,$F6
	.dc.b	$4C,$3B,$57,$2F,$80,$91,$E1,$74,$FF,$A4,$73,$FF,$48,$13,$9B,$D3
	.dc.b	$F3,$CF,$85,$F5,$CF,$F9,$32,$BF,$CF,$6D,$79,$73,$6A,$FE,$0F,$CF
	.dc.b	$7B,$27,$FD,$70,$D4,$DC,$1F,$0D,$C3,$86,$57,$FA,$22,$BE,$3E,$DF
	.dc.b	$BB,$79,$7C,$DB,$E1,$3E,$7B,$6B,$FD,$1E,$D7,$FD,$5D,$E1,$7F,$81
	.dc.b	$9F,$F5,$84,$65,$FE,$9A,$D7,$FB,$2B,$5F,$F6,$83,$5F,$E2,$DB,$FF
	.dc.b	$AC,$27,$FD,$45,$1F,$D5,$9A,$FF,$6E,$6B,$FE,$D9,$7F,$BE,$3B,$F5
	.dc.b	$39,$11,$FE,$BE,$78,$3C,$2F,$F5,$ED,$FF,$12,$8C,$87,$0F,$06,$BE
	.dc.b	$3F,$7B,$CC,$47,$8D,$FF,$BA,$1F,$83,$F3,$DD,$88,$F6,$A4,$78,$BE
	.dc.b	$7A,$4E,$5A,$FA,$2B,$FC,$EE,$F6,$F8,$1F,$8C,$47,$A4,$7B,$08,$E0
	.dc.b	$F0,$7A,$3F,$00,$8E,$0F,$83,$EF,$FF,$A3,$93,$49,$F0,$7F,$23,$E4
	.dc.b	$A4,$FA,$DA,$FC,$EF,$8B,$FE,$E2,$FC,$C6,$B9,$F4,$BF,$DD,$B1,$88
	.dc.b	$FC,$3F,$C0,$FD,$42,$3F,$1D,$FF,$79,$FC,$EF,$F8,$C2,$3E,$BB,$5F
	.dc.b	$BA,$D7,$FC,$3A,$D7,$FC,$6A,$BE,$5F,$99,$23,$9B,$D9,$87,$62,$9F
	.dc.b	$8A,$F9,$DF,$F1,$BB,$F8,$86,$B9,$F7,$7F,$16,$A3,$2F,$F8,$C1,$FC
	.dc.b	$33,$FE,$48,$6A,$38,$71,$F2,$8C,$C7,$EF,$38,$7A,$F5,$9A,$48,$FC
	.dc.b	$DC,$3E,$77,$64,$8C,$BE,$6F,$9C,$7C,$ED,$24,$FE,$92,$32,$FE,$9B
	.dc.b	$ED,$1E,$AF,$F3,$46,$C7,$A7,$BF,$FC,$0D,$18,$84,$7F,$82,$6B,$FC
	.dc.b	$DB,$18,$F2,$BF,$CD,$35,$B8,$FF,$2A,$B5,$CD,$FC,$C8,$C4,$7C,$77
	.dc.b	$B5,$F8,$3E,$4B,$C8,$79,$6B,$CD,$DD,$E7,$EB,$98,$79,$F1,$F0,$CE
	.dc.b	$A3,$07,$59,$DB,$E7,$3C,$11,$FD,$53,$ED,$1F,$3F,$49,$3C,$84,$7F
	.dc.b	$9B,$9C,$3C,$1B,$1C,$1E,$C7,$0E,$CF,$EC,$3F,$0A,$F9,$AA,$49,$ED
	.dc.b	$0F,$97,$C4,$7F,$20,$8E,$0F,$F1,$E6,$B9,$9F,$CB,$C6,$36,$FF,$1F
	.dc.b	$3F,$E5,$09,$EB,$EE,$F4,$9F,$F2,$17,$DA,$3E,$77,$FB,$AB,$FD,$42
	.dc.b	$BE,$03,$5F,$5B,$B7,$31,$C3,$9A,$46,$0E,$3E,$87,$42,$3D,$3C,$BA
	.dc.b	$F4,$8B,$87,$92,$7F,$C0,$FE,$27,$F1,$08,$FF,$2C,$37,$8F,$FD,$6E
	.dc.b	$DB,$FD,$87,$D0,$7F,$40,$AD,$F8,$7F,$BA,$AF,$E2,$D5,$C2,$30,$47
	.dc.b	$EF,$57,$DB,$7E,$CB,$6B,$9C,$7E,$0F,$E8,$1F,$3D,$FE,$1C,$AF,$F3
	.dc.b	$65,$7F,$9B,$5B,$8F,$FB,$32,$B7,$7C,$6F,$84,$C8,$C1,$D7,$C2,$FF
	.dc.b	$13,$D7,$F1,$F8,$C1,$E1,$FA,$0C,$CC,$8C,$9F,$90,$94,$FF,$8C,$3C
	.dc.b	$7F,$F2,$04,$FE,$6D,$AF,$E5,$9B,$97,$FA,$15,$FB,$D1,$9F,$6B,$F9
	.dc.b	$1C,$70,$8F,$F6,$8B,$5F,$F7,$4B,$5F,$E4,$4F,$4C,$73,$EE,$78,$CF
	.dc.b	$25,$7F,$91,$3E,$83,$FC,$29,$87,$FA,$55,$5E,$4D,$7A,$33,$1F,$DA
	.dc.b	$4E,$9B,$39,$7F,$63,$0F,$80,$EC,$48,$E3,$F0,$75,$DC,$5C,$3C,$53
	.dc.b	$FE,$58,$F3,$7F,$55,$5F,$BC,$6C,$FF,$AA,$DB,$A7,$F4,$19,$43,$E1
	.dc.b	$B9,$75,$49,$FE,$19,$3C,$1F,$E9,$0E,$DC,$C6,$DF,$03,$E3,$9C,$5F
	.dc.b	$86,$2D,$F8,$6F,$95,$F0,$39,$90,$F5,$5F,$07,$C3,$55,$F4,$C5,$9F
	.dc.b	$34,$7B,$27,$C4,$F3,$BF,$13,$6F,$C5,$7C,$6B,$93,$67,$47,$C0,$70
	.dc.b	$F9,$CF,$29,$A3,$C0,$5B,$F6,$DF,$82,$38,$34,$FA,$93,$D1,$C5,$F9
	.dc.b	$27,$95,$F9,$1A,$75,$AE,$13,$F2,$E7,$DF,$3D,$A3,$0E,$CE,$26,$4F
	.dc.b	$61,$8A,$F0,$B8,$34,$7B,$0B,$7E,$C9,$EB,$18,$AF,$0D,$72,$8E,$0D
	.dc.b	$AE,$11,$F6,$8F,$05,$7B,$DF,$98,$7E,$54,$F0,$16,$3F,$3A,$DA,$FF
	.dc.b	$1E,$91,$FF,$21,$35,$FF,$23,$3F,$A5,$EC,$91,$E0,$3F,$36,$E1,$F5
	.dc.b	$1F,$A0,$71,$7E,$99,$B7,$F9,$99,$FA,$35,$C8,$E0,$75,$3F,$5A,$F0
	.dc.b	$7E,$01,$F2,$9C,$3F,$1A,$3E,$5D,$7E,$DE,$73,$7F,$9C,$23,$8B,$FD
	.dc.b	$03,$6F,$F2,$0D,$7B,$1E,$43,$DA,$7F,$05,$75,$5F,$55,$B7,$E2,$3D
	.dc.b	$D3,$C6,$7E,$13,$63,$2F,$48,$F4,$39,$3D,$EF,$57,$07,$BC,$CC,$72
	.dc.b	$8E,$31,$CE,$30,$47,$64,$3E,$69,$D2,$FE,$35,$FC,$8B,$FA,$B3,$A3
	.dc.b	$E0,$1F,$56,$7F,$C2,$55,$D6,$3C,$83,$E8,$3F,$C3,$CF,$F8,$B5,$FE
	.dc.b	$CA,$FF,$8A,$1B,$82,$7F,$96,$47,$09,$D7,$77,$F2,$FA,$D4,$7F,$9C
	.dc.b	$67,$8D,$70,$E5,$D3,$CA,$ED,$6F,$D9,$AF,$95,$B7,$FE,$FB,$C3,$AC
	.dc.b	$FA,$7D,$9C,$D4,$8F,$53,$C2,$3F,$A4,$7C,$2C,$9F,$E7,$EF,$F4,$BF
	.dc.b	$6F,$11,$FE,$AF,$8F,$F5,$59,$1F,$F5,$83,$FD,$47,$E1,$64,$72,$3F
	.dc.b	$CA,$9F,$41,$FD,$A3,$4F,$EA,$0F,$F6,$38,$FF,$28,$B7,$2F,$4B,$B8
	.dc.b	$FF,$4B,$35,$3B,$9E,$7C,$BF,$12,$35,$F4,$1F,$46,$FE,$1B,$6B,$FB
	.dc.b	$03,$6F,$FD,$44,$FD,$26,$ED,$F3,$A3,$50,$69,$B5,$F4,$0F,$80,$FA
	.dc.b	$AD,$8F,$F3,$A3,$5F,$EE,$86,$A3,$FD,$73,$19,$8F,$E4,$0D,$7F,$BB
	.dc.b	$35,$FF,$5C,$3C,$F0,$FB,$EF,$F9,$85,$BE,$27,$59,$FF,$42,$75,$3D
	.dc.b	$5F,$C4,$B6,$33,$18,$F7,$B2,$C8,$C9,$CA,$3F,$23,$F3,$EE,$3B,$C3
	.dc.b	$2B,$FD,$25,$39,$6C,$E5,$FF,$47,$C2,$79,$CF,$C8,$F4,$65,$7C,$87
	.dc.b	$E5,$FC,$6E,$23,$FA,$21,$F4,$23,$E4,$91,$C9,$93,$E8,$FA,$5F,$50
	.dc.b	$E5,$AF,$E9,$BE,$75,$9F,$EB,$1B,$5F,$EA,$B6,$BF,$EA,$65,$F0,$9B
	.dc.b	$99,$3B,$79,$47,$ED,$BE,$83,$DD,$3F,$EB,$55,$D1,$93,$5A,$C9,$D4
	.dc.b	$65,$E1,$1F,$8C,$F1,$24,$7E,$31,$FC,$7F,$8B,$88,$F8,$BC,$9F,$E5
	.dc.b	$D7,$F2,$B3,$F8,$E2,$3F,$94,$F8,$9F,$CB,$9C,$75,$FE,$13,$F6,$4E
	.dc.b	$DF,$97,$48,$C3,$F3,$9A,$FD,$46,$BF,$DE,$7E,$75,$7C,$9B,$1F,$59
	.dc.b	$AF,$DC,$6B,$FE,$01,$F7,$DF,$37,$E2,$6C,$7D,$E6,$BF,$84,$6B,$FE
	.dc.b	$09,$FB,$17,$C0,$7A,$9B,$1F,$86,$6B,$F8,$C6,$BF,$E1,$1F,$9E,$FF
	.dc.b	$2A,$B7,$CA,$F8,$71,$88,$FF,$35,$35,$0F,$9E,$7A,$DF,$29,$B1,$F9
	.dc.b	$66,$BF,$9C,$6B,$FE,$21,$F8,$A7,$DA,$7D,$26,$C7,$E7,$9A,$FE,$91
	.dc.b	$AF,$F0,$AA,$FE,$1A,$47,$F4,$BC,$7D,$9F,$F4,$14,$3E,$73,$EF,$B6
	.dc.b	$3F,$56,$D7,$F6,$2D,$7F,$C6,$DE,$8B,$E8,$FF,$0E,$D8,$FD,$9B,$5F
	.dc.b	$DC,$B5,$FF,$1F,$7D,$03,$E7,$BF,$1E,$D8,$FD,$DB,$5F,$E0,$2D,$7F
	.dc.b	$C8,$DF,$40,$F5,$5F,$09,$FC,$0A,$30,$DC,$1F,$31,$9A,$DD,$FF,$0C
	.dc.b	$FA,$F7,$C1,$FF,$6A,$8F,$F5,$34,$63,$D1,$FE,$1A,$D8,$E1,$3B,$7F
	.dc.b	$03,$BB,$A3,$11,$F3,$5B,$19,$BF,$E2,$ED,$9F,$E3,$BF,$41,$F0,$E7
	.dc.b	$F8,$DB,$63,$3B,$F4,$8C,$47,$F9,$9F,$C9,$F8,$55,$FC,$89,$B1,$FC
	.dc.b	$A1,$AE,$F3,$F2,$FD,$07,$C4,$AF,$E5,$2D,$9F,$E5,$ED,$7F,$CE,$8D
	.dc.b	$7F,$D9,$5F,$41,$F8,$99,$FE,$62,$D8,$FE,$6E,$D7,$FC,$F8,$D7,$FD
	.dc.b	$A5,$F5,$EF,$B1,$F8,$ED,$AF,$E7,$6D,$9F,$E8,$2D,$7F,$D0,$DF,$44
	.dc.b	$F8,$5F,$9E,$DA,$FE,$86,$D9,$FE,$92,$D7,$FD,$1D,$F4,$5F,$45,$BC
	.dc.b	$1F,$62,$31,$1F,$EE,$76,$BF,$BE,$FB,$0F,$E2,$B7,$FD,$41,$B1,$FD
	.dc.b	$59,$AF,$FA,$6D,$AF,$FB,$C7,$E8,$FF,$DE,$13,$FD,$61,$B1,$FD,$79
	.dc.b	$AF,$FA,$8D,$AF,$FB,$E7,$E8,$DE,$FF,$F9,$33,$CB,$FF,$20,$FA,$8F
	.dc.b	$9A,$C6,$3B,$AB,$87,$F1,$6A,$D4,$FF,$18,$BF,$F2,$4E,$1F,$81,$6D
	.dc.b	$F0,$B4,$E6,$BA,$1C,$4F,$F2,$F7,$C6,$6F,$07,$87,$C7,$14,$E6,$31
	.dc.b	$E1,$66,$2D,$F3,$C7,$C9,$1E,$17,$FB,$4C,$65,$61,$F9,$EC,$3F,$71
	.dc.b	$E9,$78,$79,$F8,$7A,$5C,$62,$E1,$C7,$0E,$62,$FF,$80,$7B,$1A,$8C
	.dc.b	$4F,$FB,$4E,$31,$3F,$C1,$76,$F7,$78,$3E,$56,$D7,$5F,$07,$0D,$3C
	.dc.b	$11,$EB,$C1,$F4,$7F,$DB,$9A,$7F,$70,$70,$7F,$87,$3E,$85,$EA,$BE
	.dc.b	$8F,$0D,$AF,$F1,$2C,$EA,$72,$71,$F5,$BD,$D3,$A6,$BF,$EE,$F9,$46
	.dc.b	$23,$FA,$D8,$8C,$E9,$E3,$F5,$3F,$D2,$71,$FD,$EA,$3F,$D2,$2D,$CF
	.dc.b	$8B,$C2,$9F,$47,$FE,$08,$8F,$C1,$C7,$DB,$6B,$F4,$0F,$ED,$6E,$2B
	.dc.b	$AD,$F8,$39,$65,$71,$7F,$D4,$BF,$47,$E8,$8E,$47,$A7,$DA,$F3,$91
	.dc.b	$E0,$78,$7D,$BF,$59,$1F,$24,$7F,$AC,$3D,$AF,$F2,$E4,$74,$6E,$3F
	.dc.b	$EB,$06,$E9,$F2,$3E,$8F,$FC,$EC,$FC,$A8,$FF,$5B,$37,$8B,$A1,$D7
	.dc.b	$97,$FD,$7B,$F4,$7F,$27,$1F,$F0,$5B,$53,$70,$FF,$40,$5F,$D6,$8F
	.dc.b	$8C,$DE,$BE,$67,$4D,$FF,$D9,$3D,$CF,$AA,$BE,$AC,$CB,$F5,$93,$96
	.dc.b	$C7,$F9,$0A,$3F,$C1,$B1,$F8,$46,$BF,$E2,$06,$BF,$C4,$7E,$D7,$81
	.dc.b	$1F,$F3,$AC,$23,$A7,$44,$9C,$B5,$FF,$3F,$CF,$F9,$6C,$9C,$FE,$33
	.dc.b	$C9,$8C,$9E,$AD,$78,$E5,$E9,$23,$E2,$7D,$9C,$B2,$3C,$AD,$7F,$D1
	.dc.b	$FF,$89,$BE,$3F,$FC,$C6,$D8,$FF,$47,$B5,$DC,$7F,$9A,$3C,$B7,$C3
	.dc.b	$D2,$4F,$FA,$5A,$32,$FF,$A6,$3F,$0E,$7D,$07,$F9,$C1,$B1,$98,$C1
	.dc.b	$FE,$74,$FA,$77,$CE,$52,$4F,$FA,$8A,$32,$DF,$9D,$E0,$47,$06,$FE
	.dc.b	$CB,$A1,$1C,$1F,$EA,$CF,$A5,$F7,$3F,$AD,$6B,$FC,$19,$AF,$F9,$89
	.dc.b	$A9,$7D,$9F,$FA,$31,$B1,$D6,$30,$67,$E6,$A4,$78,$5A,$FF,$B0,$BE
	.dc.b	$9B,$FC,$B4,$FF,$9C,$1A,$9A,$4F,$07,$FD,$36,$FF,$AE,$A3,$FD,$8E
	.dc.b	$D7,$A4,$6B,$C2,$91,$A6,$BF,$ED,$2F,$A5,$7C,$6F,$FA,$99,$B1,$FE
	.dc.b	$D2,$6B,$B8,$D7,$C3,$C8,$D3,$5E,$9F,$51,$CC,$8F,$37,$F1,$9E,$BF
	.dc.b	$F3,$4F,$1D,$E1,$7F,$36,$E6,$FF,$28,$5F,$F6,$56,$BF,$C8,$9F,$F3
	.dc.b	$94,$F0,$8F,$F4,$BC,$69,$FE,$50,$D7,$A5,$78,$1F,$CC,$3E,$1A,$F0
	.dc.b	$7C,$85,$9E,$6D,$7F,$C7,$93,$7F,$13,$F0,$47,$F3,$45,$F5,$39,$35
	.dc.b	$72,$7F,$BF,$4E,$9B,$39,$7F,$9E,$FC,$DF,$FB,$B9,$BE,$A7,$86,$31
	.dc.b	$9F,$81,$FF,$78,$CF,$F8,$25,$CE,$1F,$D7,$9B,$1F,$EB,$A6,$A7,$8B
	.dc.b	$F2,$FC,$5F,$76,$9F,$C6,$E3,$E8,$46,$3C,$6C,$BF,$C0,$C9,$1F,$C0
	.dc.b	$5A,$FD,$5F,$1B,$99,$C1,$FF,$0F,$7C,$D7,$F8,$ED,$BF,$55,$F3,$60
	.dc.b	$DF,$22,$4E,$88,$7F,$62,$9E,$4D,$CF,$91,$1E,$FB,$57,$47,$8A,$3C
	.dc.b	$11,$F0,$39,$E4,$FE,$3D,$F8,$7E,$2E,$23,$FB,$91,$FE,$7D,$8F,$E4
	.dc.b	$84,$7F,$91,$3F,$2B,$A2,$47,$33,$FC,$7D,$F3,$1F,$E1,$46,$F9,$3F
	.dc.b	$8D,$8C,$47,$F7,$E7,$FC,$4A,$E8,$86,$63,$86,$BF,$95,$F9,$7F,$E2
	.dc.b	$33,$FE,$06,$7F,$9D,$71,$FE,$59,$BA,$7F,$88,$9D,$C6,$7B,$B9,$78
	.dc.b	$F2,$FF,$CA,$95,$AB,$FF,$14,$FE,$37,$F4,$30,$FD,$0B,$63,$3C,$3F
	.dc.b	$A7,$8F,$B7,$7F,$E5,$D7,$EB,$BF,$E1,$B6,$BC,$0F,$9A,$1C,$A3,$11
	.dc.b	$D7,$E7,$64,$75,$8F,$DB,$7C,$CF,$C8,$8F,$F1,$3B,$70,$FD,$BB,$70
	.dc.b	$FF,$21,$BF,$9F,$9D,$E9,$C8,$8F,$F9,$D6,$35,$5F,$C0,$3B,$FE,$5C
	.dc.b	$F9,$57,$D0,$7F,$07,$6C,$7F,$9F,$5D,$BF,$CB,$B5,$D4,$C8,$D4,$7C
	.dc.b	$9F,$83,$91,$F2,$5F,$99,$F1,$98,$8F,$F2,$1E,$C9,$C6,$3F,$52,$AE
	.dc.b	$9F,$62,$6D,$FC,$BB,$56,$23,$E8,$4D,$9C,$C7,$84,$87,$82,$3C,$6E
	.dc.b	$2F,$F0,$71,$C6,$B9,$36,$75,$7E,$36,$9E,$39,$DB,$83,$F0,$4E,$1F
	.dc.b	$01,$F8,$B7,$15,$E0,$8C,$1C,$6F,$E2,$36,$77,$5E,$18,$F9,$31,$87
	.dc.b	$E2,$5F,$B1,$EE,$4B,$EC,$F0,$3B,$47,$B5,$D5,$75,$9D,$9F,$4E,$36
	.dc.b	$78,$CA,$70,$3E,$11,$87,$8C,$7E,$C2,$BC,$D3,$F5,$8F,$79,$C4,$C1
	.dc.b	$6F,$EC,$B6,$FC,$D7,$D2,$3D,$E7,$8C,$F7,$9F,$B5,$8F,$D9,$BD,$0C
	.dc.b	$1F,$B6,$9F,$9A,$FF,$24,$6B,$A3,$11,$CA,$37,$3F,$C2,$3E,$5F,$F8
	.dc.b	$31,$EC,$83,$8A,$F8,$58,$7F,$10,$8C,$47,$AC,$FE,$18,$E2,$BD,$91
	.dc.b	$F0,$9E,$CE,$8E,$67,$E8,$63,$84,$FD,$72,$46,$4E,$07,$73,$F5,$F1
	.dc.b	$C1,$F9,$27,$CF,$70,$F8,$63,$9B,$FC,$6E,$33,$5F,$0A,$72,$FC,$09
	.dc.b	$EF,$5E,$A8,$D4,$78,$CF,$09,$EA,$3A,$1C,$27,$9D,$F6,$BF,$98,$78
	.dc.b	$1F,$DE,$36,$7F,$4C,$D7,$D4,$73,$3D,$B7,$E3,$8D,$36,$34,$7B,$CD
	.dc.b	$D7,$D7,$73,$7E,$A7,$47,$A0,$F5,$9F,$36,$7E,$BB,$87,$E1,$5F,$AE
	.dc.b	$3F,$82,$71,$FF,$B1,$1F,$E8,$8D,$4D,$3F,$D9,$C5,$F4,$3F,$B7,$1F
	.dc.b	$DB,$FD,$1F,$ED,$1F,$47,$FD,$A7,$C9,$97,$FD,$82,$32,$FF,$62,$8C
	.dc.b	$BF,$DB,$63,$FC,$89,$E8,$7F,$8A,$E3,$FA,$B6,$5F,$E4,$48,$C4,$7F
	.dc.b	$BE,$1B,$1F,$E4,$8F,$94,$F3,$CF,$0D,$7F,$AF,$2A,$E2,$FF,$C0,$16
	.dc.b	$BA,$6B,$FD,$95,$AF,$1F,$0F,$FC,$B1,$F5,$EF,$91,$A7,$A2,$3F,$B3
	.dc.b	$E9,$2B,$FA,$F3,$FD,$AB,$B3,$11,$FD,$AA,$2E,$7F,$DA,$A7,$FA,$F6
	.dc.b	$BF,$D7,$FB,$25,$FF,$5F,$8C,$BF,$D8,$63,$2A,$BF,$B5,$4F,$F9,$B7
	.dc.b	$E4,$1F,$61,$49,$5F,$14,$FF,$01,$7D,$87,$C8,$79,$6F,$F2,$2B,$8C
	.dc.b	$7F,$82,$9A,$FF,$8E,$1A,$FC,$E0,$E1,$E5,$5F,$E7,$FF,$B0,$7D,$55
	.dc.b	$3C,$73,$FE,$35,$F0,$FF,$C7,$04,$7F,$C7,$BE,$25,$74,$7F,$90,$23
	.dc.b	$FC,$7D,$1F,$E3,$7F,$88,$FF,$1C,$7C,$0C,$AF,$F1,$7C,$65,$FF,$18
	.dc.b	$C6,$5F,$F1,$E4,$7D,$CF,$02,$47,$C5,$1C,$BF,$EB,$6E,$0F,$C7,$67
	.dc.b	$A9,$FF,$2C,$CF,$FA,$E9,$F8,$4D,$7F,$22,$D7,$FC,$60,$F3,$DB,$3C
	.dc.b	$9F,$8C,$3A,$5F,$F8,$B3,$3F,$F1,$89,$2F,$FC,$51,$D3,$5F,$F1,$9C
	.dc.b	$59,$FF,$19,$CF,$F8,$9F,$4F,$F1,$47,$4C,$BF,$F1,$44,$65,$FF,$15
	.dc.b	$46,$5F,$C0,$9F,$E3,$18,$FB,$1E,$7C,$8F,$89,$A7,$FB,$23,$0F,$F5
	.dc.b	$BF,$1A,$35,$3F,$E7,$C7,$FD,$8C,$FF,$93,$DA,$EE,$BF,$B1,$C7,$3D
	.dc.b	$3F,$6E,$FF,$B4,$9B,$D5,$D3,$0F,$F6,$2D,$3C,$DF,$ED,$6D,$C3,$FC
	.dc.b	$BE,$D7,$FD,$1A,$D7,$F2,$CB,$E2,$78,$92,$38,$BD,$3E,$26,$47,$46
	.dc.b	$BF,$E9,$8F,$9B,$FF,$46,$BF,$A1,$6B,$FC,$41,$AF,$F9,$BD,$A9,$76
	.dc.b	$7F,$81,$1B,$1F,$E7,$86,$BD,$E3,$5C,$12,$34,$D7,$FD,$57,$F6,$4E
	.dc.b	$0F,$F0,$7B,$63,$FD,$06,$D7,$71,$AE,$69,$1A,$6B,$D3,$D2,$E6,$47
	.dc.b	$9B,$FE,$1C,$F5,$5F,$4B,$49,$3F,$E9,$28,$CB,$7C,$1E,$04,$70,$6F
	.dc.b	$A3,$A1,$1C,$1F,$E9,$CF,$9B,$FB,$5F,$37,$78,$4E,$78,$5F,$1D,$98
	.dc.b	$FF,$51,$FD,$8F,$F3,$9E,$8B,$99,$A4,$CB,$FD,$AB,$F6,$4F,$87,$D4
	.dc.b	$61,$AF,$F2,$18,$D4,$FF,$AC,$1B,$1F,$ED,$EF,$C2,$DC,$3C,$23,$FA
	.dc.b	$AC,$62,$BF,$D6,$ED,$8E,$13,$B7,$FD,$71,$F8,$53,$E4,$F1,$08,$C5
	.dc.b	$FF,$B0,$1B,$1F,$EE,$E6,$BF,$CD,$FE,$6D,$F2,$BC,$9B,$5F,$EC,$86
	.dc.b	$CF,$FB,$D9,$AF,$F3,$FF,$3D,$F1,$7E,$26,$D7,$FB,$41,$B3,$FE,$FE
	.dc.b	$6B,$C7,$C9,$E0,$49,$E0,$F3,$F0,$A4,$73,$6C,$FF,$4F,$1F,$2E,$F8
	.dc.b	$CF,$86,$DA,$FF,$70,$B6,$78,$35,$FE,$B0,$BE,$6B,$89,$27,$8B,$F3
	.dc.b	$BE,$61,$F2,$7F,$82,$F1,$FF,$03,$C2,$F7,$71,$87,$F8,$42,$F6,$FF
	.dc.b	$4B,$9D,$3F,$D3,$63,$F2,$AD,$BE,$39,$7F,$BD,$67,$2F,$F3,$40,$F4
	.dc.b	$71,$83,$C4,$E3,$1F,$85,$E2,$B3,$9C,$3C,$51,$8F,$33,$37,$67,$FB
	.dc.b	$05,$6E,$FC,$03,$C3,$EB,$9F,$F2,$9A,$FE,$CD,$39,$7D,$DD,$FE,$28
	.dc.b	$FC,$D5,$FE,$7E,$F1,$CF,$97,$5F,$27,$8C,$39,$B6,$FC,$3B,$78,$3C
	.dc.b	$0E,$8D,$AF,$0F,$81,$86,$9E,$48,$F1,$53,$E6,$3F,$D6,$2D,$3F,$D6
	.dc.b	$6E,$0F,$C6,$2F,$96,$7C,$03,$E6,$33,$18,$6C,$66,$7F,$29,$39,$6F
	.dc.b	$B3,$FE,$35,$6E,$FC,$A3,$11,$FE,$29,$94,$62,$3F,$CB,$31,$FD,$0A
	.dc.b	$7F,$87,$B7,$2C,$F9,$9F,$E5,$E9,$CD,$F7,$F9,$8F,$E7,$D7,$FE,$7B
	.dc.b	$27,$26,$F0,$E3,$8F,$F8,$7E,$5C,$B9,$81,$F0,$DC,$34,$FF,$13,$7C
	.dc.b	$C7,$F9,$07,$2F,$2F,$AF,$AB,$34,$C5,$E7,$E1,$A6,$1E,$17,$4F,$F8
	.dc.b	$C7,$CF,$72,$F8,$1B,$19,$8F,$1F,$CC,$3E,$A3,$5C,$B2,$B9,$3C,$6B
	.dc.b	$D3,$F3,$3F,$26,$7D,$AD,$4D,$BF,$E3,$B1,$FE,$43,$8F,$F5,$21,$38
	.dc.b	$39,$F1,$FC,$4E,$FF,$E8,$D5,$F8,$A6,$BF,$C5,$DA,$FF,$99,$DA,$FF
	.dc.b	$A5,$9D,$F0,$7F,$4C,$5F,$1D,$D4,$ED,$E6,$77,$F9,$EF,$CB,$DF,$F1
	.dc.b	$16,$D1,$A6,$3E,$3E,$4F,$95,$AF,$97,$A7,$89,$1F,$53,$D3,$C1,$91
	.dc.b	$ED,$6B,$ED,$CF,$C2,$47,$E1,$3E,$4D,$24,$7C,$56,$BF,$CE,$FF,$11
	.dc.b	$FE,$B6,$9A,$4E,$9F,$7F,$CD,$BE,$2A,$92,$BF,$73,$19,$7F,$77,$EE
	.dc.b	$7D,$A7,$FA,$43,$5F,$F4,$C3,$5F,$F6,$3B,$5F,$D0,$FC,$D7,$F3,$D7
	.dc.b	$F8,$2B,$5F,$EA,$4D,$7F,$D4,$4D,$7F,$DA,$2B,$E6,$FE,$BB,$63,$9C
	.dc.b	$60,$CF,$CD,$C8,$EE,$D7,$FB,$07,$CD,$7F,$8A,$9B,$EE,$FE,$0A,$31
	.dc.b	$1F,$EB,$36,$BF,$EE,$27,$E2,$3C,$B2,$E0,$47,$F9,$0C,$3E,$27,$A1
	.dc.b	$23,$2F,$E4,$3E,$69,$E9,$52,$4F,$F2,$A8,$CB,$7C,$1E,$04,$70,$6F
	.dc.b	$C2,$E8,$47,$07,$F3,$4F,$55,$F3,$7F,$A0,$6C,$66,$30,$7E,$93,$E6
	.dc.b	$FF,$D6,$10,$69,$3E,$27,$5F,$37,$27,$FA,$0C,$A3,$11,$FE,$0F,$F8
	.dc.b	$EC,$3F,$EC,$56,$BF,$DD,$5B,$E0,$FF,$91,$3E,$BF,$FD,$49,$3B,$AF
	.dc.b	$DB,$D7,$08,$FF,$74,$35,$FE,$5E,$1C,$3F,$CC,$9E,$BD,$1A,$AF,$C5
	.dc.b	$4F,$F8,$15,$B7,$FE,$54,$AD,$CF,$F3,$A9,$7C,$C7,$F3,$A5,$F4,$F4
	.dc.b	$C4,$49,$FF,$71,$4E,$9B,$39,$7F,$C2,$9F,$5F,$F8,$77,$FD,$D0,$D7
	.dc.b	$3C,$7F,$D2,$23,$F2,$38,$FF,$9A,$1D,$36,$34,$6F,$97,$FA,$7F,$CC
	.dc.b	$BC,$2A,$4C,$3F,$DE,$D1,$FB,$DB,$FF,$39,$BB,$8F,$EF,$6D,$7F,$C9
	.dc.b	$0D,$7E,$78,$70,$FD,$27,$84,$74,$F1,$BC,$67,$86,$7F,$C6,$9F,$5F
	.dc.b	$FC,$0E,$7F,$D0,$6B,$E7,$7F,$B3,$B6,$7F,$B5,$CE,$BC,$7C,$9A,$A9
	.dc.b	$3A,$E1,$FC,$1D,$F5,$6F,$81,$D0,$93,$B3,$FB,$BC,$BE,$D7,$31,$E6
	.dc.b	$9B,$8F,$9A,$7F,$BD,$C7,$FB,$C6,$7F,$BF,$B5,$FF,$77,$7C,$7E,$BB
	.dc.b	$76,$FF,$50,$9F,$F7,$D9,$1C,$1D,$3E,$82,$57,$94,$E6,$5F,$A0,$FF
	.dc.b	$54,$E3,$FE,$AA,$7F,$2B,$19,$E6,$FF,$07,$C6,$B4,$FD,$2B,$67,$FC
	.dc.b	$2A,$F0,$7F,$CC,$CD,$AE,$93,$FE,$18,$F8,$EF,$F5,$8D,$78,$DA,$FF
	.dc.b	$9B,$5B,$3F,$EB,$C7,$FC,$C1,$1B,$F7,$B9,$11,$E5,$EF,$7C,$AE,$49
	.dc.b	$1E,$FD,$7F,$13,$8F,$F4,$CE,$3C,$CF,$F5,$2A,$F9,$44,$7F,$AB,$BF
	.dc.b	$53,$CD,$91,$CE,$3F,$C6,$3F,$41,$F8,$38,$FF,$1A,$BF,$8D,$7F,$D0
	.dc.b	$CD,$D3,$FD,$AA,$EE,$0D,$35,$87,$F1,$06,$C7,$4E,$5E,$F3,$23,$ED
	.dc.b	$4F,$D1,$79,$EF,$97,$C3,$36,$FC,$33,$F8,$EF,$0A,$57,$C5,$7F,$23
	.dc.b	$E1,$E2,$3F,$D3,$DA,$FF,$92,$1D,$1F,$2E,$35,$7F,$EF,$09,$D4,$F0
	.dc.b	$AE,$6F,$03,$05,$9E,$15,$CB,$0F,$EF,$F5,$6B,$F4,$33,$55,$F6,$BF
	.dc.b	$05,$76,$9E,$A4,$8C,$BE,$73,$C1,$7D,$9A,$76,$29,$DA,$7C,$15,$FE
	.dc.b	$72,$8C,$BE,$19,$E5,$1F,$1A,$F7,$5C,$1B,$1C,$1A,$F9,$67,$FC,$3B
	.dc.b	$18,$BF,$3C,$F3,$3B,$D7,$36,$9C,$CA,$73,$3D,$45,$3D,$46,$CC,$5F
	.dc.b	$A5,$FE,$1B,$1F,$26,$7C,$24,$BF,$6B,$F5,$A7,$BB,$97,$BB,$5E,$E7
	.dc.b	$6B,$F6,$4F,$23,$DF,$1C,$9C,$3C,$9F,$96,$FF,$90,$AB,$F8,$E1,$1F
	.dc.b	$F2,$4B,$5F,$F2,$4B,$5E,$07,$EF,$FD,$7F,$E5,$67,$89,$FE,$56,$E1
	.dc.b	$F1,$18,$38,$18,$3F,$1E,$ED,$F9,$31,$6F,$F4,$13,$D6,$76,$72,$36
	.dc.b	$58,$DB,$D7,$04,$F3,$BF,$87,$3F,$2A,$B6,$FE,$CA,$7B,$57,$08,$E9
	.dc.b	$5C,$23,$EE,$47,$88,$E0,$77,$9E,$CF,$E0,$0F,$CD,$B8,$7C,$13,$CE
	.dc.b	$31,$3D,$08,$E2,$FF,$4A,$DA,$E6,$79,$0E,$C7,$E4,$1D,$34,$D1,$E0
	.dc.b	$37,$7F,$00,$97,$ED,$AF,$1D,$7C,$23,$9C,$FE,$46,$FE,$25,$FC,$03
	.dc.b	$E8,$96,$39,$BD,$AF,$FC,$7B,$F3,$5F,$EF,$E8,$FE,$C8,$D5,$29,$DF
	.dc.b	$B9,$FE,$F0,$A6,$F6,$25,$F2,$75,$1F,$EF,$6F,$A0,$FF,$16,$DF,$F9
	.dc.b	$31,$FF,$18,$B7,$B7,$5D,$D3,$0F,$EE,$D5,$A7,$FE,$00,$B5,$3D,$10
	.dc.b	$F9,$CF,$E9,$C4,$E2,$FF,$2D,$E9,$EB,$9D,$CF,$F4,$E7,$84,$FF,$B6
	.dc.b	$A3,$FD,$51,$1A,$8C,$11,$FF,$05,$47,$F8,$DD,$BD,$FD,$B2,$FF,$34
	.dc.b	$CF,$F5,$FF,$A0,$F8,$F5,$FE,$11,$6B,$F3,$A7,$FC,$DF,$18,$C3,$CD
	.dc.b	$28,$FE,$F2,$FB,$9D,$4F,$0D,$3C,$13,$96,$BC,$79,$6F,$8F,$F2,$47
	.dc.b	$D0,$60,$CE,$F2,$BF,$C4,$D3,$A6,$CE,$5F,$41,$F2,$BE,$33,$BE,$86
	.dc.b	$98,$FA,$0B,$3F,$E5,$E3,$81,$4F,$B1,$1F,$44,$8F,$CE,$FA,$0E,$A7
	.dc.b	$27,$C2,$39,$1E,$B7,$E5,$49,$1F,$9C,$F5,$EB,$C2,$76,$FE,$82,$7F
	.dc.b	$39,$3F,$97,$F9,$CF,$CC,$7D,$04,$AF,$C9,$C6,$5F,$CA,$C6,$5F,$1F
	.dc.b	$32,$7E,$8F,$9A,$F0,$E9,$2B,$FC,$A5,$E3,$FA,$5F,$F5,$D4,$7F,$AD
	.dc.b	$08,$FC,$59,$E4,$D8,$FB,$BF,$3C,$F9,$6A,$75,$9F,$8B,$D0,$D3,$F0
	.dc.b	$6D,$8F,$C9,$46,$3B,$72,$C8,$C9,$F9,$78,$FC,$4E,$BF,$8A,$ED,$93
	.dc.b	$F8,$68,$CB,$F8,$88,$CB,$F5,$B9,$93,$F9,$4F,$A0,$7D,$05,$25,$7F
	.dc.b	$9D,$7E,$83,$E2,$3F,$7A,$76,$FE,$1D,$AF,$E2,$DA,$FF,$AA,$9A,$FF
	.dc.b	$B6,$9D,$F7,$7E,$B3,$E8,$1F,$2D,$FA,$66,$D7,$FA,$16,$B7,$E0,$7F
	.dc.b	$B8,$BE,$AF,$F9,$24,$FF,$AD,$9A,$E7,$6F,$CD,$28,$CB,$FC,$53,$CF
	.dc.b	$FF,$1C,$BF,$9F,$6B,$FA,$36,$BF,$EC,$26,$BF,$EF,$25,$F2,$5D,$9B
	.dc.b	$1C,$E3,$07,$8B,$D7,$7C,$AD,$24,$FE,$BE,$32,$FE,$C7,$E8,$7C,$E9
	.dc.b	$A4,$C7,$DA,$FE,$26,$3F,$74,$D7,$FD,$B2,$D7,$F9,$99,$CF,$93,$E2
	.dc.b	$FD,$FB,$E9,$E9,$27,$F8,$14,$65,$FE,$09,$F4,$6F,$90,$F9,$AD,$8C
	.dc.b	$C6,$0C,$FA,$92,$38,$B5,$FE,$21,$F7,$FF,$E5,$B6,$FA,$5F,$62,$31
	.dc.b	$1F,$EF,$26,$BE,$AF,$A4,$7C,$2F,$DE,$6C,$7F,$19,$6B,$FC,$75,$AF
	.dc.b	$FB,$F1,$FD,$B3,$BE,$FE,$44,$7F,$91,$7D,$57,$C1,$7F,$93,$35,$F9
	.dc.b	$61,$C3,$F3,$9A,$FF,$62,$FA,$5F,$C3,$FD,$2F,$F2,$D6,$C7,$F3,$15
	.dc.b	$F3,$8F,$9C,$D7,$3A,$FF,$9A,$7D,$2F,$F0,$DE,$BB,$9F,$F5,$4C,$BE
	.dc.b	$EC,$E5,$FE,$78,$D7,$CF,$F4,$8E,$9E,$11,$FD,$DA,$31,$7F,$D0,$5B
	.dc.b	$1F,$D1,$5A,$F2,$F7,$F8,$A8,$46,$1B,$3C,$E3,$7F,$48,$F9,$3A,$49
	.dc.b	$FE,$A1,$19,$7F,$A9,$7D,$3B,$E9,$3B,$B6,$33,$18,$3C,$7E,$AB,$E3
	.dc.b	$A9,$27,$FA,$E4,$65,$BE,$37,$02,$38,$3F,$B1,$7D,$23,$9F,$DE,$D8
	.dc.b	$CC,$60,$CE,$12,$38,$B7,$6F,$ED,$DF,$48,$F9,$DF,$F1,$07,$95,$FE
	.dc.b	$0C,$DD,$F0,$38,$C7,$F6,$F7,$B7,$FC,$57,$5B,$9D,$61,$FE,$2E,$AF
	.dc.b	$C4,$56,$8E,$38,$6A,$53,$A7,$FB,$BB,$D5,$FE,$F3,$7F,$DE,$2B,$94
	.dc.b	$7F,$8E,$B0,$7D,$27,$3A,$D4,$FF,$80,$27,$35,$C2,$3F,$CA,$FD,$FB
	.dc.b	$F6,$78,$36,$75,$2F,$07,$CD,$7F,$CB,$27,$31,$8E,$BF,$55,$7E,$30
	.dc.b	$FF,$07,$9C,$67,$9F,$36,$AB,$7F,$3D,$C2,$2C,$E8,$F0,$ED,$E0,$EE
	.dc.b	$EA,$D8,$F2,$F7,$61,$A7,$03,$D3,$8B,$E9,$3F,$A7,$B4,$FE,$A2,$E0
	.dc.b	$FF,$16,$7D,$0F,$F9,$DB,$E9,$2F,$C2,$9A,$4A,$FF,$1A,$4B,$CD,$FF
	.dc.b	$40,$C6,$1A,$FF,$12,$8D,$71,$F6,$9C,$CF,$F4,$09,$FE,$8D,$3F,$E5
	.dc.b	$A6,$E3,$9F,$33,$FC,$AB,$1C,$A9,$F4,$9F,$CF,$54,$7F,$8D,$9F,$F1
	.dc.b	$FB,$6F,$A7,$2F,$2B,$E7,$CB,$2F,$F0,$97,$FC,$FE,$FF,$CD,$7F,$49
	.dc.b	$FC,$A9,$F2,$7C,$5F,$88,$95,$C2,$FC,$BF,$17,$23,$9B,$A7,$FC,$A3
	.dc.b	$E6,$B9,$BF,$CD,$ED,$8F,$CA,$35,$FF,$2B,$FD,$1B,$E9,$37,$CD,$27
	.dc.b	$6F,$2B,$FF,$3C,$FD,$27,$C8,$9F,$F4,$EB,$53,$6F,$F8,$94,$7E,$8E
	.dc.b	$3E,$37,$17,$F7,$26,$C7,$2E,$1E,$BC,$3F,$E8,$6E,$F7,$D2,$52,$57
	.dc.b	$EB,$61,$F0,$98,$8F,$F3,$D4,$70,$7F,$CF,$4D,$73,$3F,$E2,$48,$7B
	.dc.b	$14,$93,$C1,$D7,$D2,$48,$E6,$D7,$BF,$C7,$73,$23,$E0,$7B,$7C,$AC
	.dc.b	$8F,$1B,$5F,$3F,$87,$E3,$23,$E8,$7C,$D9,$C8,$F5,$B5,$FF,$4C,$7C
	.dc.b	$3B,$EB,$3F,$C7,$ED,$8F,$F3,$B4,$60,$FF,$21,$FD,$2F,$F5,$88,$34
	.dc.b	$9C,$3F,$E4,$8F,$A5,$7C,$95,$25,$7F,$A9,$63,$2F,$CE,$FB,$6F,$98
	.dc.b	$47,$07,$FA,$AF,$E9,$BF,$DB,$29,$A4,$E1,$FF,$2B,$AF,$A6,$A4,$9E
	.dc.b	$4E,$BE,$82,$47,$56,$BF,$EB,$BF,$A5,$FF,$0F,$BF,$E4,$06,$BF,$EF
	.dc.b	$46,$A6,$9E,$C7,$96,$5F,$E6,$A8,$C4,$7F,$88,$1A,$F5,$F2,$74,$47
	.dc.b	$FD,$95,$1F,$E1,$B7,$FD,$99,$F4,$BF,$98,$6F,$47,$F9,$DA,$31,$1F
	.dc.b	$ED,$36,$BF,$E2,$E7,$FC,$82,$83,$86,$F8,$3C,$CF,$F4,$0F,$AD,$FE
	.dc.b	$E1,$8F,$F6,$E3,$5F,$F7,$03,$5F,$F1,$DB,$5F,$F3,$3A,$FA,$5E,$64
	.dc.b	$8E,$6F,$FA,$3B,$E9,$5F,$1F,$49,$3F,$EE,$B8,$CB,$FE,$ED,$FB,$77
	.dc.b	$57,$F9,$D3,$C8,$73,$7F,$9D,$A7,$35,$FD,$AD,$AF,$FB,$BA,$FF,$C1
	.dc.b	$4E,$6F,$FD,$3F,$3A,$8F,$F2,$AB,$5E,$91,$A3,$C1,$5D,$FB,$AC,$FC
	.dc.b	$DA,$FE,$0A,$F5,$3C,$5F,$63,$A5,$0F,$23,$0D,$8D,$90,$65,$EE,$7B
	.dc.b	$4F,$03,$8D,$75,$FB,$7F,$F3,$85,$6E,$91,$8C,$7F,$D3,$6D,$8F,$C2
	.dc.b	$3F,$EC,$66,$FC,$57,$32,$3F,$E1,$58,$D3,$FE,$F7,$ED,$F6,$7C,$1F
	.dc.b	$F0,$98,$F3,$DF,$FA,$A2,$31,$87,$8D,$7D,$A7,$0B,$FF,$67,$C7,$1B
	.dc.b	$FF,$42,$B5,$E5,$1A,$3B,$FC,$6C,$8F,$F1,$44,$F6,$C3,$FD,$6F,$F6
	.dc.b	$FF,$D5,$67,$27,$98,$FA,$4E,$5F,$39,$FA,$57,$18,$F9,$E3,$81,$FE
	.dc.b	$35,$8C,$CF,$6F,$0F,$99,$27,$5E,$9D,$7A,$8A,$FE,$C0,$48,$FC,$27
	.dc.b	$89,$FE,$8D,$9F,$F2,$13,$5E,$2D,$C3,$EE,$7D,$53,$E4,$B8,$FC,$3F
	.dc.b	$F3,$A2,$5F,$FA,$B8,$FE,$01,$7C,$DF,$F5,$8B,$CF,$3F,$F6,$ED,$FF
	.dc.b	$29,$9F,$F2,$93,$5F,$0D,$7E,$66,$1F,$47,$7C,$29,$5C,$23,$46,$1F
	.dc.b	$83,$29,$CC,$7F,$97,$FE,$A9,$F3,$1F,$A5,$6C,$66,$31,$E2,$65,$91
	.dc.b	$93,$94,$7E,$B7,$E8,$DE,$6D,$24,$FF,$9B,$63,$FC,$1A,$FD,$47,$6F
	.dc.b	$C7,$0E,$1F,$94,$D7,$FD,$B2,$D7,$F9,$61,$C2,$3C,$5F,$4D,$23,$C1
	.dc.b	$1F,$C0,$7E,$A7,$FD,$2F,$1F,$E7,$F5,$F5,$59,$35,$E6,$49,$D4,$65
	.dc.b	$E1,$1C,$FE,$9E,$47,$37,$B7,$D5,$62,$3F,$23,$69,$C7,$2F,$F0,$58
	.dc.b	$57,$F1,$59,$FE,$33,$5F,$98,$6D,$7F,$0B,$6F,$1F,$C0,$7F,$A8,$C6
	.dc.b	$7C,$2E,$44,$AF,$F4,$E4,$EA,$B8,$4E,$67,$B3,$E1,$F2,$B1,$7C,$E7
	.dc.b	$8C,$E5,$A4,$38,$9D,$A7,$8D,$76,$8D,$57,$A6,$73,$1F,$D8,$EF,$D6
	.dc.b	$DB,$EC,$F4,$25,$78,$DF,$89,$5C,$9F,$99,$36,$BD,$D1,$CC,$F0,$DF
	.dc.b	$8D,$E2,$60,$D1,$D9,$F9,$95,$CE,$BA,$36,$78,$1F,$6D,$C3,$C4,$F6
	.dc.b	$38,$BF,$41,$F0,$63,$8D,$FC,$07,$D2,$7C,$93,$35,$E0,$69,$C2,$BA
	.dc.b	$5F,$AE,$FA,$4F,$84,$E6,$5B,$F8,$6E,$23,$C7,$3C,$4F,$7C,$76,$38
	.dc.b	$57,$67,$E2,$CF,$63,$8B,$89,$E4,$FC,$03,$F0,$87,$ED,$0F,$19,$6B
	.dc.b	$F3,$AD,$BF,$E0,$0E,$8F,$AC,$64,$FE,$09,$1C,$1E,$E7,$EA,$FC,$4F
	.dc.b	$D4,$1E,$77,$07,$E8,$1C,$47,$F8,$4A,$FF,$55,$18,$38,$CF,$C1,$8C
	.dc.b	$BC,$4C,$5F,$EB,$E7,$71,$F3,$6E,$D7,$08,$C1,$C4,$FD,$D4,$F7,$71
	.dc.b	$1F,$29,$FA,$95,$D5,$CB,$EE,$BF,$C3,$1E,$97,$51,$D8,$EA,$70,$3E
	.dc.b	$A1,$0F,$86,$68,$E1,$3E,$72,$47,$7B,$F5,$BD,$A3,$CD,$5E,$2B,$CD
	.dc.b	$F5,$8C,$3E,$13,$D6,$66,$36,$E0,$91,$EB,$7F,$81,$47,$C5,$7C,$D3
	.dc.b	$E0,$BE,$35,$B9,$E3,$18,$AE,$15,$EA,$9F,$24,$E5,$FD,$BF,$D5,$7F
	.dc.b	$8C,$E7,$FC,$68,$FF,$6A,$8F,$F3,$A4,$63,$9B,$FC,$8D,$3B,$C7,$FA
	.dc.b	$6B,$6B,$FC,$F4,$F1,$7F,$C1,$AD,$AE,$B3,$FE,$4F,$FA,$27,$C3,$52
	.dc.b	$5F,$82,$7F,$AC,$CB,$E1,$F3,$3F,$E1,$88,$E2,$EC,$FF,$45,$38,$75
	.dc.b	$1F,$D2,$78,$A4,$7F,$40,$7F,$A6,$71,$E2,$3F,$C2,$9B,$7F,$69,$7F
	.dc.b	$99,$4F,$EC,$A3,$11,$FD,$5E,$3F,$AA,$7C,$2C,$8E,$87,$8C,$7D,$57
	.dc.b	$F6,$C7,$FC,$5E,$FF,$6C,$74,$D7,$3C,$9F,$E7,$38,$E1,$AF,$F9,$9D
	.dc.b	$D3,$79,$75,$3B,$D6,$5F,$47,$97,$06,$23,$C3,$16,$FA,$4F,$F6,$DF
	.dc.b	$82,$95,$FD,$91,$FE,$E1,$E8,$62,$3E,$2C,$7C,$32,$3F,$C4,$9F,$EE
	.dc.b	$DE,$5C,$8E,$47,$C3,$FA,$67,$87,$FE,$8F,$6C,$66,$31,$B6,$59,$19
	.dc.b	$39,$47,$FA,$5F,$E9,$BF,$C6,$71,$FE,$5A,$7F,$DB,$D9,$FF,$A3,$5B
	.dc.b	$A7,$F8,$89,$DC,$67,$BB,$97,$74,$BF,$B1,$5A,$BF,$F1,$4F,$D4,$FD
	.dc.b	$26,$F0,$7F,$A9,$E3,$15,$FE,$D9,$7E,$AC,$EA,$3F,$C7,$11,$F8,$17
	.dc.b	$FA,$03,$5D,$1F,$E6,$37,$9A,$7F,$C7,$EF,$E2,$DA,$FF,$9E,$1D,$FD
	.dc.b	$2B,$3F,$F5,$D3,$67,$F2,$4D,$7F,$CF,$EF,$F9,$27,$3F,$F5,$93,$96
	.dc.b	$CF,$23,$9F,$27,$FB,$17,$D2,$F9,$77,$FE,$F7,$5F,$55,$93,$5A,$C9
	.dc.b	$FD,$14,$E9,$B3,$97,$D5,$F2,$BF,$2C,$8F,$BF,$EA,$AB,$F1,$63,$FB
	.dc.b	$1E,$9F,$EF,$B8,$FA,$11,$83,$EA,$7D,$54,$8F,$91,$1F,$ED,$DF,$B4
	.dc.b	$7D,$5F,$F6,$C6,$CF,$F9,$B5,$AF,$FA,$55,$AF,$F7,$7F,$A6,$FC,$BB
	.dc.b	$FE,$99,$6B,$9E,$5F,$EE,$F1,$88,$FF,$38,$F8,$7F,$E5,$D7,$FC,$F4
	.dc.b	$D7,$FD,$44,$D7,$FC,$08,$D7,$FC,$F3,$F6,$9E,$F4,$D2,$63,$FE,$06
	.dc.b	$F3,$5E,$9D,$25,$7F,$A2,$63,$2F,$FA,$2F,$E9,$FC,$C8,$38,$6E,$5F
	.dc.b	$E1,$28,$C7,$CB,$BE,$27,$FC,$2A,$DA,$FF,$4B,$35,$FF,$5F,$35,$FF
	.dc.b	$10,$7D,$3B,$DB,$FF,$0E,$B6,$3F,$D3,$CD,$7F,$D8,$CD,$7F,$C5,$1F
	.dc.b	$4E,$F9,$2F,$F1,$2B,$63,$FD,$4C,$D7,$FD,$9C,$D7,$FC,$61,$F7,$8F
	.dc.b	$AC,$FF,$16,$B6,$3F,$D5,$CD,$7F,$DA,$CD,$7F,$C7,$1F,$52,$F1,$FF
	.dc.b	$C6,$AD,$8F,$F5,$B3,$5F,$F6,$F3,$5F,$F2,$07,$CF,$BE,$A7,$FC,$7A
	.dc.b	$D8,$FF,$5F,$35,$FF,$73,$35,$FF,$24,$7C,$EB,$D5,$FF,$22,$B6,$3F
	.dc.b	$D8,$CD,$7F,$DD,$CD,$7F,$CA,$1F,$53,$F2,$DF,$F6,$2B,$5F,$C1,$35
	.dc.b	$37,$EA,$5F,$40,$FA,$9F,$F6,$A4,$61,$B6,$F8,$FC,$D6,$F0,$FF,$80
	.dc.b	$1F,$2D,$FE,$BF,$F2,$FF,$DB,$8D,$8C,$FA,$A9,$CB,$63,$FC,$D0,$3E
	.dc.b	$A1,$D9,$98,$FF,$36,$46,$2B,$FD,$D0,$D8,$E2,$D7,$7E,$D7,$17,$FD
	.dc.b	$CF,$1F,$E1,$49,$E3,$39,$9F,$D3,$4E,$5F,$1F,$D9,$7F,$B5,$AB,$FD
	.dc.b	$EC,$D8,$CF,$BD,$C5,$46,$5F,$F4,$17,$D4,$3E,$9B,$4D,$BF,$F8,$00
	.dc.b	$D9,$F8,$2D,$7F,$D1,$5E,$1B,$E4,$BB,$36,$BE,$3C,$A3,$0F,$CA,$6B
	.dc.b	$FE,$93,$FB,$27,$17,$9D,$B5,$C9,$B3,$F4,$9A,$FF,$A6,$FE,$A1,$D1
	.dc.b	$EF,$6D,$78,$9B,$3F,$69,$AF,$FA,$8F,$EA,$1F,$45,$F7,$FD,$6F,$B9
	.dc.b	$F3,$CF,$67,$18,$FF,$17,$C1,$87,$08,$FF,$18,$56,$DF,$F1,$94,$7E
	.dc.b	$4D,$B8,$72,$AF,$E4,$13,$95,$87,$E2,$E3,$A4,$62,$BC,$12,$F9,$FE
	.dc.b	$52,$C1,$E4,$62,$B5,$7E,$48,$C7,$85,$9C,$2C,$FF,$91,$EB,$78,$77
	.dc.b	$3D,$DF,$09,$DB,$1F,$D8,$D7,$F9,$36,$7F,$CC,$8F,$C0,$EC,$F1,$C7
	.dc.b	$E5,$6F,$F3,$97,$8E,$9C,$BA,$FA,$2C,$61,$CD,$B7,$E1,$3F,$A2,$6D
	.dc.b	$E1,$E3,$EA,$DA,$F1,$77,$61,$A7,$96,$3C,$74,$FA,$8F,$F6,$DB,$4F
	.dc.b	$F6,$EB,$83,$F2,$FF,$4D,$FE,$1B,$FA,$8B,$EF,$1A,$4A,$FE,$07,$3F
	.dc.b	$E4,$CD,$B8,$FA,$3F,$6A,$72,$D7,$FD,$E1,$3B,$78,$FC,$A7,$DC,$CF
	.dc.b	$F1,$92,$8C,$4F,$F6,$58,$C3,$FE,$7D,$79,$7D,$47,$F4,$F7,$F0,$91
	.dc.b	$9E,$4E,$2A,$32,$FC,$D7,$9B,$79,$77,$1F,$D2,$72,$FE,$38,$DC,$7F
	.dc.b	$41,$F5,$1F,$A1,$BE,$5F,$0F,$C0,$8F,$97,$0E,$7F,$0F,$23,$D0,$E9
	.dc.b	$FB,$DE,$93,$97,$CE,$D8,$FF,$4D,$35,$FC,$27,$D3,$BE,$A3,$7C,$B2
	.dc.b	$76,$F2,$BF,$83,$F5,$1E,$ED,$FF,$9B,$37,$6F,$DD,$36,$FF,$C2,$58
	.dc.b	$FF,$1B,$7F,$D5,$51,$F2,$2B,$8F,$07,$43,$AF,$17,$CE,$FD,$D3,$F8
	.dc.b	$4D,$FF,$AC,$5F,$EA,$4D,$E8,$D6,$9B,$7F,$92,$C2,$76,$FF,$4E,$6B
	.dc.b	$F3,$DA,$FF,$11,$7F,$36,$47,$9B,$F9,$EF,$03,$E3,$47,$F3,$26,$BF
	.dc.b	$D6,$9A,$FD,$B6,$BF,$C7,$5F,$84,$EF,$C5,$F3,$C6,$23,$84,$7A,$BE
	.dc.b	$A5,$F2,$F4,$93,$FD,$0A,$32,$F8,$FE,$AB,$99,$1C,$1F,$D2,$3E,$E7
	.dc.b	$FB,$6B,$7C,$FF,$8F,$18,$8F,$EE,$6D,$7D,$5A,$F9,$88,$FC,$17,$D3
	.dc.b	$B4,$8F,$88,$D7,$E2,$73,$FB,$88,$FC,$97,$EB,$FD,$4B,$E1,$A9,$27
	.dc.b	$FA,$FC,$65,$BF,$61,$C0,$8E,$0D,$FB,$FD,$08,$E0,$FE,$D3,$F5,$3F
	.dc.b	$E6,$87,$FC,$22,$D7,$F6,$4D,$7F,$AA,$35,$FF,$36,$2F,$A2,$FC,$7B
	.dc.b	$63,$AC,$60,$CF,$C6,$48,$F0,$B5,$FF,$10,$FD,$4F,$F9,$D4,$9E,$27
	.dc.b	$E6,$9C,$6D,$49,$3F,$E0,$28,$FE,$1E,$FF,$09,$7F,$89,$B7,$6E,$44
	.dc.b	$7F,$C1,$3F,$51,$FE,$51,$7F,$B8,$35,$FF,$44,$35,$34,$7D,$27,$EB
	.dc.b	$9B,$1D,$23,$06,$7E,$6E,$47,$81,$AF,$F9,$03,$97,$FC,$7D,$D9,$FE
	.dc.b	$3F,$6F,$7F,$F8,$0D,$B3,$FE,$9B,$5F,$31,$FC,$C9,$FF,$19,$41,$A4
	.dc.b	$9F,$D5,$1F,$CE,$3E,$B1,$F4,$D4,$E5,$1F,$CF,$3E,$7B,$F5,$77,$FD
	.dc.b	$9D,$B8,$FF,$94,$DB,$3F,$E1,$77,$FC,$5B,$3F,$EA,$F9,$D7,$67,$32
	.dc.b	$46,$6B,$87,$9F,$FD,$2B,$4F,$F4,$E5,$FF,$97,$DF,$F6,$BB,$7E,$4B
	.dc.b	$5E,$44,$C3,$91,$FE,$BD,$AD,$E1,$FE,$29,$6C,$66,$7E,$07,$74,$8E
	.dc.b	$23,$5F,$E0,$18,$FF,$9B,$D7,$CA,$EE,$7F,$CE,$B1,$FC,$0D,$FF,$1B
	.dc.b	$35,$FF,$67,$35,$E9,$5F,$3F,$D9,$91,$E0,$9E,$F9,$7F,$8F,$FF,$05
	.dc.b	$FF,$45,$4F,$F8,$55,$B8,$FF,$A0,$5B,$3B,$FA,$DF,$EE,$31,$8A,$FF
	.dc.b	$6F,$35,$34,$47,$F7,$8F,$9F,$FB,$33,$FE,$E5,$7E,$CB,$FE,$25,$6D
	.dc.b	$FF,$A4,$5D,$C1,$A6,$AF,$FC,$B0,$D8,$E9,$C1,$F8,$FF,$99,$91,$F0
	.dc.b	$5F,$C9,$76,$62,$3E,$6C,$7F,$58,$CC,$D2,$47,$C9,$F9,$DF,$CD,$1C
	.dc.b	$72,$FF,$53,$7C,$F3,$E2,$3F,$C2,$AB,$E8,$69,$88,$97,$F8,$29,$D3
	.dc.b	$67,$2F,$F9,$1F,$E7,$FF,$D0,$2F,$E2,$1A,$E7,$87,$FC,$47,$1F,$72
	.dc.b	$01,$DE,$1F,$EB,$B6,$EF,$FE,$7F,$6C,$F0,$79,$3F,$95,$F7,$7F,$70
	.dc.b	$47,$F1,$5E,$ED,$7F,$2B,$1F,$24,$8F,$FA,$9B,$64,$7F,$01,$F5,$24
	.dc.b	$FE,$4E,$3F,$C7,$7F,$3F,$F0,$A3,$FC,$CC,$BE,$6B,$26,$B5,$93,$FA
	.dc.b	$A9,$D3,$67,$2F,$F9,$BF,$A3,$D6,$4E,$8F,$F2,$4B,$FA,$E8,$FD,$9B
	.dc.b	$96,$C6,$A3,$FC,$9F,$F1,$9F,$E0,$8D,$FF,$DC,$8D,$DF,$FD,$50,$DB
	.dc.b	$FA,$2F,$F8,$52,$77,$E8,$F2,$24,$FF,$02,$9D,$57,$29,$CC,$FF,$A1
	.dc.b	$E7,$A2,$76,$F6,$8D,$D7,$68,$E9,$EC,$62,$FA,$47,$7A,$F5,$CE,$63
	.dc.b	$91,$93,$89,$4F,$14,$6D,$F1,$45,$BF,$29,$1C,$4E,$5B,$3C,$EF,$E4
	.dc.b	$C6,$23,$A9,$E0,$29,$23,$D8,$FC,$38,$E1,$18,$39,$C7,$5B,$F7,$BC
	.dc.b	$5F,$AB,$3C,$1C,$CF,$09,$D1,$F0,$8A,$7B,$8E,$86,$5F,$04,$74,$8C
	.dc.b	$9F,$56,$BE,$03,$8B,$F8,$B1,$F4,$23,$0E,$9E,$F1,$A7,$E1,$57,$67
	.dc.b	$C1,$5E,$83,$C0,$F0,$AE,$53,$B8,$F1,$1E,$1B,$FC,$5B,$FE,$9D,$BF
	.dc.b	$F2,$F3,$F1,$8C,$57,$F9,$E2,$33,$1F,$0C,$C9,$FE,$A5,$30,$78,$CF
	.dc.b	$CF,$FA,$5F,$A6,$3E,$73,$83,$F3,$CE,$1F,$CD,$3B,$7E,$19,$A2,$1E
	.dc.b	$D2,$DF,$EC,$23,$94,$9E,$53,$B3,$F5,$0E,$27,$E2,$4F,$E5,$A3,$18
	.dc.b	$7A,$6B,$E3,$9C,$0E,$EE,$67,$84,$60,$FA,$64,$3D,$66,$4E,$75,$E0
	.dc.b	$AE,$F1,$B9,$E6,$F7,$3D,$13,$ED,$79,$47,$E0,$DD,$46,$5D,$D7,$F1
	.dc.b	$0A,$EB,$1E,$08,$F4,$46,$2B,$95,$71,$24,$FB,$6F,$55,$F3,$0E,$0F
	.dc.b	$EF,$5B,$7F,$9B,$6B,$F9,$86,$D7,$BE,$3D,$33,$FD,$8B,$E5,$FF,$B1
	.dc.b	$FC,$1C,$9F,$EC,$11,$97,$FB,$14,$65,$57,$F6,$C9,$FE,$BB,$F3,$32
	.dc.b	$7F,$B1,$3F,$D8,$3E,$2F,$11,$FD,$9E,$2D,$7F,$5F,$9F,$EC,$73,$D3
	.dc.b	$2B,$F2,$5F,$E5,$C8,$34,$9E,$C7,$0B,$57,$03,$FD,$BB,$3F,$E6,$02
	.dc.b	$38,$38,$1D,$BC,$AB,$7F,$DD,$27,$FC,$D2,$FF,$90,$9F,$17,$1B,$4F
	.dc.b	$EC,$7D,$92,$BF,$AF,$47,$F6,$63,$0F,$F6,$68,$C3,$FD,$6B,$3F,$EB
	.dc.b	$A4,$7F,$AE,$F5,$D7,$FB,$24,$5C,$FD,$B1,$FE,$08,$E3,$FF,$01,$AF
	.dc.b	$A6,$57,$AA,$30,$D8,$CC,$7F,$97,$3E,$C5,$BE,$6B,$F8,$44,$FF,$85
	.dc.b	$2B,$6F,$F0,$B6,$BF,$58,$EA,$70,$77,$87,$9A,$1F,$8B,$EE,$30,$DE
	.dc.b	$0F,$BF,$5C,$1F,$C0,$F7,$BE,$8B,$FC,$6B,$F4,$93,$1F,$F1,$94,$7F
	.dc.b	$8E,$0C,$3F,$87,$3F,$C5,$72,$9D,$47,$F8,$CF,$C4,$FF,$18,$91,$FF
	.dc.b	$1B,$7A,$D5,$FF,$1D,$45,$CF,$FC,$65,$3F,$2F,$EC,$DF,$33,$49,$7F
	.dc.b	$85,$FA,$7A,$FF,$31,$9F,$F6,$21,$27,$E6,$C2,$39,$BF,$E8,$E7,$FD
	.dc.b	$5E,$DF,$4B,$9E,$1F,$2F,$9F,$D9,$3F,$E2,$BE,$BF,$F1,$99,$FE,$34
	.dc.b	$D2,$46,$48,$7F,$8D,$9F,$F1,$5E,$72,$3F,$C5,$6F,$F8,$BB,$AF,$11
	.dc.b	$FE,$33,$8B,$87,$DB,$E6,$57,$E3,$7E,$AB,$FC,$3E,$4E,$DF,$C8,$74
	.dc.b	$38,$35,$1F,$02,$7F,$AB,$E4,$69,$23,$A7,$81,$27,$FD,$63,$3F,$E2
	.dc.b	$CF,$09,$F6,$9F,$29,$B3,$F9,$86,$BF,$D8,$5A,$FE,$5D,$7D,$87,$12
	.dc.b	$47,$17,$A7,$CC,$C8,$E8,$D7,$FB,$67,$D1,$FD,$26,$F0,$FD,$C8,$C4
	.dc.b	$7E,$95,$AF,$F1,$C7,$F3,$8F,$EB,$9A,$F4,$E9,$E4,$73,$F9,$19,$1B
	.dc.b	$6B,$FD,$F7,$E8,$9E,$7F,$E3,$1B,$1F,$BA,$6B,$B8,$D7,$D1,$48,$D3
	.dc.b	$5E,$9E,$C7,$32,$3C,$DF,$CC,$FE,$2F,$FC,$92,$3F,$C1,$CD,$7F,$82
	.dc.b	$35,$FE,$76,$D7,$FD,$3E,$FF,$05,$77,$E0,$74,$23,$FC,$4B,$C5,$FD
	.dc.b	$1B,$FE,$1B,$6B,$FC,$45,$AF,$F4,$86,$BF,$E7,$CF,$90,$FF,$13,$11
	.dc.b	$FE,$3B,$E6,$23,$F6,$7F,$50,$F9,$CA,$4A,$FE,$47,$19,$7F,$92,$FD
	.dc.b	$AB,$E2,$7F,$76,$D8,$CC,$60,$FD,$FF,$D1,$3E,$6B,$FD,$F3,$3F,$D7
	.dc.b	$A3,$1F,$45,$FC,$C1,$7C,$13,$E4,$B1,$7C,$2F,$FD,$8F,$F6,$AF,$87
	.dc.b	$F9,$51,$86,$A8,$D7,$0F,$F3,$C6,$C7,$F9,$17,$ED,$DE,$4F,$06,$D7
	.dc.b	$F4,$26,$CE,$E7,$AF,$D1,$79,$6B,$FA,$3B,$67,$FC,$A0,$D7,$FA,$33
	.dc.b	$5F,$EF,$AF,$FB,$82,$38,$C7,$42,$3F,$D4,$B8,$5E,$1F,$B1,$B1,$98
	.dc.b	$C1,$F0,$3E,$85,$E4,$D2,$4F,$F5,$C8,$CB,$7D,$BE,$04,$70,$7F,$62
	.dc.b	$FA,$17,$CC,$7F,$70,$F3,$7F,$AE,$7B,$5F,$DA,$2B,$FD,$F3,$1F,$D7
	.dc.b	$EB,$71,$83,$FC,$37,$3B,$D3,$FA,$B0,$9F,$91,$5C,$6F,$B5,$FF,$88
	.dc.b	$A7,$2D,$8C,$D7,$83,$E3,$AB,$FE,$13,$FA,$1B,$82,$73,$8A,$73,$F2
	.dc.b	$FA,$9B,$8F,$9F,$0E,$8D,$43,$E1,$79,$F3,$78,$23,$13,$BC,$7F,$3C
	.dc.b	$F1,$87,$C1,$EE,$F8,$36,$75,$F0,$98,$69,$C6,$3B,$6B,$EA,$EA,$B3
	.dc.b	$DD,$F1,$75,$63,$1F,$9A,$7C,$83,$FC,$97,$5E,$75,$F4,$3F,$DC,$9A
	.dc.b	$7F,$74,$70,$7F,$88,$BE,$7D,$E5,$3E,$87,$0D,$AF,$F5,$74,$EA,$72
	.dc.b	$71,$F9,$6F,$F3,$2C,$E9,$AF,$F8,$A6,$51,$88,$FF,$30,$C6,$39,$3C
	.dc.b	$5F,$2D,$FE,$3B,$83,$4C,$57,$6F,$A1,$7C,$97,$2E,$5E,$04,$C7,$EE
	.dc.b	$61,$C9,$BC,$DD,$48,$FE,$07,$E8,$7F,$C6,$CF,$E0,$A7,$F9,$43,$CB
	.dc.b	$C9,$F2,$1C,$1F,$CE,$C3,$E8,$7D,$04,$8E,$07,$E3,$71,$73,$7F,$95
	.dc.b	$DB,$19,$8F,$F2,$D7,$D0,$FF,$02,$8E,$1C,$7B,$23,$B3,$F3,$1F,$43
	.dc.b	$ED,$7F,$5E,$DE,$1F,$F2,$A3,$79,$FF,$D1,$F6,$F7,$3F,$CD,$13,$FA
	.dc.b	$9B,$E7,$C1,$83,$3C,$7F,$E6,$DF,$A1,$FE,$0B,$3F,$E6,$86,$BF,$EF
	.dc.b	$16,$BF,$E6,$76,$BF,$E9,$E7,$79,$7F,$99,$DF,$E8,$11,$FB,$16,$BF
	.dc.b	$C2,$1D,$E5,$C7,$AB,$FC,$97,$1F,$AE,$9F,$14,$60,$CB,$0F,$31,$4F
	.dc.b	$F3,$D1,$FE,$82,$F0,$FF,$CA,$BF,$17,$F3,$E3,$2D,$7F,$D0,$0D,$7D
	.dc.b	$5B,$F8,$88,$FC,$37,$FC,$BF,$F3,$4F,$95,$A4,$9F,$F4,$7C,$65,$FF
	.dc.b	$48,$D3,$E4,$7C,$64,$8C,$BE,$6F,$06,$47,$06,$BE,$8D,$BD,$64,$7A
	.dc.b	$BE,$ED,$E4,$78,$1A,$FC,$4F,$9F,$7D,$6F,$F9,$E1,$B1,$FE,$A6,$6B
	.dc.b	$B8,$D7,$D0,$C8,$D3,$5E,$9E,$9F,$32,$3C,$DF,$F4,$2F,$D0,$7F,$86
	.dc.b	$A0,$D2,$7C,$5E,$BE,$26,$4F,$FA,$E2,$7F,$CE,$AD,$8E,$7F,$15,$23
	.dc.b	$4D,$7F,$03,$F4,$1F,$3D,$FF,$60,$B5,$CE,$FF,$E9,$88,$C4,$7F,$B0
	.dc.b	$5F,$E3,$2F,$0F,$13,$44,$79,$EF,$C4,$8E,$0F,$F6,$7F,$5F,$F9,$D9
	.dc.b	$FF,$80,$2D,$7F,$A8,$B5,$FD,$B7,$C4,$7F,$AA,$BE,$83,$FD,$53,$1F
	.dc.b	$9A,$6B,$FE,$D5,$6B,$FD,$68,$38,$7F,$AE,$BF,$E2,$B7,$FD,$C2,$D7
	.dc.b	$A7,$B1,$DC,$FF,$5C,$FC,$93,$4F,$F7,$61,$27,$FB,$47,$D8,$3E,$4F
	.dc.b	$11,$FE,$E9,$9E,$24,$7F,$DD,$0F,$FB,$DE,$5F,$45,$4E,$3E,$0F,$FB
	.dc.b	$AE,$3F,$C8,$B5,$DB,$D0,$E0,$71,$F4,$AC,$64,$87,$F8,$42,$7F,$CA
	.dc.b	$4F,$F7,$DF,$B0,$7C,$AF,$C5,$C3,$FD,$F1,$1E,$D9,$D6,$59,$F1,$7E
	.dc.b	$54,$F0,$9E,$57,$97,$FC,$0F,$1C,$5F,$F0,$FB,$6F,$AC,$EE,$7F,$D8
	.dc.b	$DE,$E5,$F9,$1F,$F2,$9C,$7F,$84,$5A,$FF,$8A,$1F,$E2,$94,$C3,$EE
	.dc.b	$46,$A7,$EF,$1D,$0F,$E3,$A5,$3F,$8F,$BB,$78,$D7,$87,$C0,$91,$E1
	.dc.b	$8F,$27,$C0,$C8,$F2,$3E,$7F,$A0,$C4,$79,$62,$E3,$E5,$8F,$F1,$4E
	.dc.b	$0A,$F8,$1F,$61,$FE,$D2,$9F,$F1,$73,$FE,$D8,$AE,$10,$8C,$65,$F9
	.dc.b	$97,$84,$7F,$92,$9B,$3C,$1D,$C7,$F8,$E1,$AF,$3F,$91,$FA,$1E,$37
	.dc.b	$6A,$EE,$DB,$FD,$33,$6B,$FC,$82,$FF,$AF,$E7,$73,$FC,$F0,$93,$FC
	.dc.b	$E9,$AF,$F9,$7B,$E7,$BF,$D7,$33,$EC,$6C,$F6,$3F,$05,$F4,$9F,$D0
	.dc.b	$6B,$FC,$CA,$DA,$F8,$6D,$9F,$DD,$3F,$A2,$9D,$F9,$3C,$88,$FF,$96
	.dc.b	$63,$55,$F3,$3E,$93,$FC,$4B,$3F,$E5,$D5,$F1,$BF,$93,$6C,$67,$D3
	.dc.b	$C1,$99,$91,$93,$FC,$CD,$93,$E0,$7F,$38,$D8,$CC,$63,$E0,$F2,$FE
	.dc.b	$52,$13,$B7,$FB,$9B,$5F,$CD,$B5,$FE,$34,$EF,$1F,$99,$F1,$9E,$03
	.dc.b	$AC,$FF,$9E,$7E,$91,$C9,$FA,$F6,$C6,$63,$1C,$79,$64,$64,$E5,$1F
	.dc.b	$B7,$FA,$4F,$A5,$1F,$E8,$47,$E6,$F0,$FF,$2D,$6C,$67,$87,$81,$C7
	.dc.b	$B6,$47,$FA,$36,$79,$BF,$0B,$E9,$3E,$49,$1F,$9B,$D5,$5F,$C6,$45
	.dc.b	$D7,$F1,$93,$71,$F8,$B3,$6B,$FD,$33,$B7,$F1,$77,$89,$F6,$23,$4F
	.dc.b	$D2,$AB,$29,$B1,$86,$CF,$12,$1D,$0F,$D0,$46,$DC,$DF,$99,$B7,$E6
	.dc.b	$69,$E6,$9E,$0F,$39,$E0,$D9,$E0,$D3,$81,$C8,$F9,$45,$BF,$81,$3F
	.dc.b	$B1,$BE,$AF,$43,$13,$B8,$FC,$CD,$EA,$FB,$D6,$DA,$4A,$F9,$B5,$C1
	.dc.b	$E4,$78,$6B,$93,$A3,$ED,$9A,$3D,$47,$D1,$31,$5E,$47,$E5,$98,$AF
	.dc.b	$03,$B3,$8D,$FD,$77,$C0,$6C,$F0,$1C,$8D,$DF,$DA,$7E,$C1,$FA,$E7
	.dc.b	$F6,$91,$CE,$3F,$67,$18,$23,$FD,$32,$34,$F1,$31,$1F,$C0,$BE,$4A
	.dc.b	$47,$8C,$FD,$FB,$87,$C6,$7F,$0A,$71,$7F,$04,$97,$FC,$02,$32,$FC
	.dc.b	$2B,$F3,$4F,$F2,$D9,$CD,$7C,$08,$E2,$FE,$8A,$32,$FE,$18,$F9,$2E
	.dc.b	$27,$9D,$75,$8C,$5F,$E9,$1B,$5F,$8D,$6B,$C8,$F5,$1E,$FA,$46,$AF
	.dc.b	$C7,$3F,$08,$D9,$F0,$8F,$E9,$11,$BB,$F1,$46,$A7,$8D,$79,$A3,$D8
	.dc.b	$66,$7E,$25,$FA,$EF,$F2,$33,$DC,$EA,$79,$5E,$CF,$C7,$9F,$9C,$F3
	.dc.b	$30,$43,$95,$FC,$F7,$91,$B3,$E6,$3E,$69,$EB,$18,$9E,$AF,$88,$FD
	.dc.b	$0F,$82,$FA,$0F,$F1,$22,$FA,$3A,$62,$24,$86,$1F,$E0,$8A,$E1,$19
	.dc.b	$8C,$57,$08,$DD,$FF,$98,$FE,$91,$C7,$FE,$33,$6C,$F1,$9D,$4E,$79
	.dc.b	$34,$C9,$D1,$E3,$AE,$2F,$63,$13,$FD,$93,$D9,$91,$FD,$A1,$FE,$CF
	.dc.b	$F1,$18,$8F,$EC,$31,$FB,$22,$3F,$AF,$7F,$B6,$FB,$32,$39,$1E,$21
	.dc.b	$F3,$DF,$09,$F1,$18,$F9,$CF,$F2,$84,$1B,$A4,$9E,$53,$96,$CE,$5E
	.dc.b	$7D,$5F,$DC,$75,$FF,$46,$36,$3E,$A3,$53,$1F,$33,$92,$67,$FE,$63
	.dc.b	$75,$1E,$8F,$95,$79,$FF,$E6,$76,$C7,$BD,$AE,$FD,$DD,$4F,$F8,$36
	.dc.b	$73,$3F,$E6,$46,$BF,$EA,$36,$BF,$8B,$79,$F0,$FF,$94,$9F,$F6,$13
	.dc.b	$7D,$4E,$B3,$F3,$F8,$FF,$CE,$B3,$F9,$1F,$23,$F3,$27,$E6,$BC,$89
	.dc.b	$5B,$8C,$BC,$63,$F2,$1E,$AC,$8F,$B5,$CB,$FE,$DA,$7E,$CA,$FA,$2D
	.dc.b	$46,$26,$E1,$FE,$68,$AD,$4F,$E0,$FE,$8F,$FC,$B4,$9A,$4F,$7F,$FD
	.dc.b	$24,$FF,$9C,$23,$FD,$12,$D7,$FD,$90,$D7,$F6,$0D,$7A,$71,$72,$FA
	.dc.b	$49,$1E,$18,$FC,$7E,$9F,$90,$EB,$93,$F8,$E8,$CB,$F9,$08,$CB,$F9
	.dc.b	$8E,$64,$FE,$17,$59,$3F,$8A,$7F,$11,$D7,$88,$FC,$AC,$5D,$7F,$D1
	.dc.b	$7C,$BF,$B4,$7F,$D1,$BE,$9F,$C6,$AF,$D2,$B6,$33,$5F,$EB,$5F,$A1
	.dc.b	$6F,$FE,$B7,$AE,$17,$FE,$AE,$6B,$FE,$ED,$6B,$FC,$55,$A9,$ED,$AF
	.dc.b	$FB,$CF,$E9,$3F,$D6,$F8,$FF,$A8,$1B,$1C,$F1,$FF,$63,$7D,$73,$E3
	.dc.b	$FF,$CD,$0D,$C7,$FD,$7A,$D7,$FD,$82,$1C,$3F,$EC,$4E,$C7,$B5,$FE
	.dc.b	$71,$6C,$7F,$B1,$9A,$F2,$6B,$C7,$D1,$E0,$47,$83,$FE,$7B,$EE,$F7
	.dc.b	$93,$1C,$1F,$E7,$FF,$A0,$7D,$0D,$25,$FF,$B5,$A3,$2D,$FA,$1E,$04
	.dc.b	$70,$7F,$B7,$3E,$B5,$F0,$9F,$E8,$B6,$C6,$63,$07,$FA,$3F,$E8,$BF
	.dc.b	$B9,$C7,$FB,$91,$AF,$C8,$6B,$FE,$EA,$6B,$FD,$3C,$2F,$A2,$E8,$48
	.dc.b	$E8,$FF,$A7,$7E,$B1,$F1,$74,$93,$FE,$F4,$8C,$BF,$EF,$6F,$A3,$77
	.dc.b	$FF,$A9,$5B,$19,$8C,$19,$E5,$91,$C5,$AF,$E0,$BC,$6F,$E7,$4D,$E3
	.dc.b	$FE,$48,$47,$F0,$E2,$1F,$27,$FC,$A4,$91,$A8,$CB,$C3,$E8,$DF,$3D
	.dc.b	$FE,$BB,$6C,$66,$30,$7F,$AF,$FE,$B1,$F0,$D4,$93,$E2,$8C,$BE,$4F
	.dc.b	$A3,$7C,$3F,$FA,$E2,$3E,$90,$8C,$FB,$9B,$F6,$9F,$3B,$C6,$F8,$5F
	.dc.b	$D9,$FA,$37,$C5,$7F,$1B,$6F,$0F,$C2,$6C,$FE,$A9,$AF,$83,$E8,$BF
	.dc.b	$C0,$F3,$F2,$1B,$1F,$B0,$6B,$BF,$2B,$F9,$44,$79,$27,$86,$38,$23
	.dc.b	$CE,$B6,$47,$07,$D6,$FA,$27,$C2,$FF,$33,$6C,$66,$30,$67,$CC,$91
	.dc.b	$C5,$B5,$FC,$1B,$E8,$9F,$34,$AF,$F0,$57,$9F,$FC,$81,$FF,$06,$4E
	.dc.b	$9C,$F9,$DB,$BF,$F0,$6C,$FB,$5B,$79,$F3,$75,$09,$D4,$61,$E9,$1F
	.dc.b	$E1,$68,$FD,$3D,$F8,$BE,$8B,$93,$FB,$18,$CC,$FF,$87,$31,$56,$A3
	.dc.b	$FC,$81,$F3,$F5,$FF,$22,$F8,$36,$D3,$63,$52,$F7,$F8,$61,$F9,$0C
	.dc.b	$7F,$C5,$93,$97,$E2,$F2,$F0,$F7,$F8,$DD,$9D,$BD,$3D,$FC,$5F,$F8
	.dc.b	$C8,$38,$CF,$FB,$86,$DE,$0E,$8F,$1B,$70,$F0,$F2,$61,$A7,$92,$3D
	.dc.b	$38,$3E,$8B,$F7,$6F,$F9,$95,$AE,$4C,$BF,$C0,$3E,$7F,$FA,$67,$D1
	.dc.b	$5F,$50,$D2,$57,$F4,$C9,$7C,$B7,$F9,$C2,$30,$D7,$F6,$B1,$AD,$3B
	.dc.b	$4E,$67,$F9,$54,$BE,$5B,$EE,$67,$86,$CF,$F9,$F6,$74,$FF,$97,$5E
	.dc.b	$8B,$E8,$9F,$1F,$D3,$97,$8C,$AB,$F9,$6B,$FE,$62,$6E,$5F,$94,$7A
	.dc.b	$C6,$79,$B8,$C7,$E1,$A3,$C1,$1D,$3E,$8B,$FB,$AC,$79,$7C,$9D,$91
	.dc.b	$D9,$E7,$FA,$2E,$64,$7D,$51,$E6,$CB,$F0,$71,$FE,$79,$6B,$FE,$78
	.dc.b	$6E,$3F,$AA,$FA,$27,$C4,$71,$E4,$97,$C5,$E9,$CF,$F0,$BE,$8B,$DB
	.dc.b	$87,$EC,$9B,$C5,$FD,$6D,$B7,$FE,$AB,$C7,$FC,$C3,$39,$E0,$E4,$49
	.dc.b	$FF,$47,$57,$05,$F4,$5F,$C2,$9F,$F4,$93,$5F,$F4,$8B,$70,$FE,$06
	.dc.b	$FD,$88,$DE,$5F,$98,$8E,$11,$FC,$F9,$AF,$F7,$16,$BF,$EC,$17,$F7
	.dc.b	$0F,$F5,$A6,$BF,$E2,$26,$BC,$FA,$BB,$75,$C8,$E0,$D7,$81,$E3,$21
	.dc.b	$D4,$FD,$6F,$D4,$FF,$A5,$9F,$EA,$6D,$7F,$C0,$CD,$7F,$DA,$CD,$7E
	.dc.b	$3E,$BE,$F2,$3E,$97,$F7,$3F,$2E,$F9,$3A,$49,$FE,$EF,$19,$7F,$BC
	.dc.b	$F8,$7F,$EB,$F7,$FB,$2B,$5F,$F0,$C3,$5F,$F7,$43,$5F,$F2,$EF,$83
	.dc.b	$FC,$22,$3F,$CD,$F1,$D2,$30,$67,$AA,$47,$81,$AF,$F9,$42,$5F,$67
	.dc.b	$CC,$91,$E6,$7F,$8A,$FD,$23,$E9,$29,$27,$FC,$27,$19,$6F,$D2,$70
	.dc.b	$23,$83,$7E,$27,$42,$38,$3F,$C3,$BF,$46,$FA,$4F,$E4,$AD,$8C,$C6
	.dc.b	$0C,$FC,$7C,$8E,$2D,$7F,$CD,$B3,$F9,$24,$D2,$72,$FF,$32,$FA,$47
	.dc.b	$C9,$D2,$57,$F8,$C2,$32,$DF,$17,$81,$1C,$1F,$E3,$6F,$94,$7D,$8F
	.dc.b	$F3,$E6,$C6,$63,$06,$7E,$2E,$47,$16,$BF,$E8,$6F,$A5,$7D,$5F,$F4
	.dc.b	$86,$C7,$F9,$11,$AE,$E3,$FA,$6F,$DA,$7F,$B2,$E0,$3B,$FF,$BF,$C6
	.dc.b	$63,$1E,$17,$F9,$A6,$31,$3F,$E0,$4B,$FE,$3B,$39,$F9,$5E,$A4,$BF
	.dc.b	$F2,$AD,$F3,$7F,$C8,$2F,$29,$E9,$B7,$F4,$08,$FF,$7B,$C7,$39,$F2
	.dc.b	$E1,$E4,$F1,$F4,$48,$CC,$F9,$7C,$AC,$47,$A0,$F6,$DF,$F9,$81,$C0
	.dc.b	$71,$7F,$02,$7F,$A8,$FD,$A3,$9E,$FC,$04,$C3,$B9,$F2,$27,$4E,$03
	.dc.b	$87,$BB,$FE,$6C,$6C,$76,$9F,$EB,$FE,$E7,$FB,$32,$7E,$9B,$5F,$F2
	.dc.b	$FF,$C7,$79,$33,$DB,$FE,$06,$9F,$F6,$A9,$1C,$1D,$3D,$C9,$5E,$59
	.dc.b	$E8,$7F,$9B,$9C,$7A,$BE,$23,$C7,$D1,$63,$24,$3E,$4C,$FF,$43,$C3
	.dc.b	$FD,$59,$F6,$9F,$ED,$9D,$BF,$C8,$EB,$E3,$7F,$BD,$B6,$FE,$33,$BA
	.dc.b	$E3,$EB,$6A,$A5,$6A,$3F,$DF,$1F,$51,$23,$E7,$7C,$AA,$3F,$C0,$FF
	.dc.b	$52,$F9,$CD,$57,$FB,$AA,$B4,$FF,$A9,$9A,$E6,$BF,$C5,$F1,$8C,$FF
	.dc.b	$CB,$F7,$FE,$9A,$AF,$C3,$B6,$3F,$CC,$73,$BC,$BF,$D9,$92,$F0,$FC
	.dc.b	$64,$BF,$C2,$47,$FA,$89,$B1,$97,$AB,$EA,$F7,$3F,$C4,$3F,$52,$FA
	.dc.b	$2B,$9C,$BF,$C7,$4E,$5B,$39,$7F,$3B,$1E,$57,$F7,$AE,$DF,$D0,$FD
	.dc.b	$4B,$E7,$FF,$C5,$ED,$8F,$CD,$35,$DF,$B3,$A9,$FF,$2A,$CE,$67,$FD
	.dc.b	$76,$D7,$FD,$DC,$D7,$F8,$6B,$CF,$6F,$F4,$5B,$FD,$5A,$3C,$7E,$7F
	.dc.b	$57,$C3,$F1,$1E,$42,$3E,$4F,$A5,$AF,$AA,$2D,$F4,$9F,$F2,$4F,$D4
	.dc.b	$BE,$6E,$F8,$F3,$0F,$DB,$4E,$5B,$39,$7F,$82,$C2,$79,$CF,$F9,$53
	.dc.b	$D4,$FF,$51,$E1,$FE,$A4,$6F,$27,$F2,$AC,$11,$87,$F9,$6A,$F9,$2E
	.dc.b	$53,$27,$F8,$44,$ED,$E3,$39,$7F,$8B,$C6,$49,$19,$AF,$15,$53,$C5
	.dc.b	$3D,$EB,$AC,$65,$F1,$BE,$BA,$C4,$7A,$22,$CF,$18,$F0,$C6,$0E,$0F
	.dc.b	$42,$1A,$3D,$73,$B7,$35,$F1,$E3,$11,$B3,$DA,$E2,$3D,$71,$C8,$E3
	.dc.b	$3E,$C3,$CB,$3B,$71,$1B,$3A,$4F,$26,$C7,$C4,$9E,$11,$F0,$8E,$13
	.dc.b	$B2,$4F,$48,$F2,$9E,$23,$81,$D0,$FB,$B3,$D1,$F9,$47,$48,$E5,$3D
	.dc.b	$23,$8B,$F0,$DE,$87,$03,$A1,$F8,$23,$35,$F8,$19,$FA,$31,$A8,$C3
	.dc.b	$DC,$EB,$1E,$02,$3E,$63,$9B,$F2,$9F,$B2,$7E,$64,$F1,$16,$3F,$E0
	.dc.b	$0B,$63,$FC,$BA,$47,$FC,$B6,$D7,$91,$EB,$94,$7E,$97,$E9,$7F,$C7
	.dc.b	$13,$FD,$C9,$B3,$F1,$63,$84,$EF,$D1,$FF,$61,$D6,$A3,$D7,$3A,$3F
	.dc.b	$C3,$31,$96,$CF,$D3,$8C,$41,$B7,$F2,$23,$C5,$3D,$8E,$05,$7E,$39
	.dc.b	$FD,$F6,$79,$9C,$AB,$F5,$2E,$27,$8C,$ED,$AF,$63,$99,$49,$7F,$06
	.dc.b	$FF,$81,$57,$39,$DC,$FC,$A7,$DA,$75,$33,$7F,$1D,$C1,$F8,$57,$D4
	.dc.b	$F6,$7E,$4D,$76,$BC,$D7,$86,$69,$CE,$B9,$1D,$23,$C4,$FC,$B9,$F6
	.dc.b	$C6,$2F,$E2,$4F,$E7,$1F,$53,$C8,$F1,$1E,$A3,$11,$C6,$7C,$31,$F9
	.dc.b	$A2,$1D,$CF,$15,$7A,$8D,$B9,$3F,$53,$F5,$3F,$EA,$D7,$FC,$7E,$DA
	.dc.b	$F4,$B6,$7F,$9C,$3F,$DE,$67,$FC,$74,$E5,$FF,$32,$B5,$34,$E0,$FF
	.dc.b	$93,$3E,$2F,$FD,$C8,$DF,$A9,$FE,$89,$7F,$22,$51,$8B,$FE,$91,$1C
	.dc.b	$27,$5F,$07,$9B,$FF,$3A,$CF,$07,$FC,$9C,$D7,$C1,$1F,$A6,$F9,$E9
	.dc.b	$1F,$97,$7F,$53,$F1,$B8,$8F,$DA,$E8,$9F,$DB,$C7,$EA,$9A,$FE,$9D
	.dc.b	$B1,$B8,$FD,$17,$8D,$FB,$43,$A6,$3F,$84,$FA,$97,$2F,$A6,$91,$87
	.dc.b	$FC,$CC,$D7,$FC,$E6,$B9,$78,$39,$AE,$95,$83,$33,$D7,$EA,$7E,$A1
	.dc.b	$F4,$A7,$F2,$AD,$DB,$FB,$A3,$AF,$7F,$87,$81,$32,$FF,$3C,$4F,$DD
	.dc.b	$D1,$F5,$3F,$C4,$09,$3F,$C0,$7B,$EB,$F2,$A3,$E2,$91,$F9,$73,$C9
	.dc.b	$FE,$11,$F5,$3F,$C6,$8E,$39,$FE,$AB,$C5,$7D,$3F,$C7,$5F,$53,$4C
	.dc.b	$44,$BF,$F4,$74,$E9,$B3,$97,$FD,$3B,$F5,$2E,$3F,$A4,$D8,$FF,$4B
	.dc.b	$35,$DF,$26,$A6,$46,$9E,$31,$FC,$CB,$E1,$E4,$FF,$23,$7F,$9A,$FC
	.dc.b	$46,$23,$F1,$B1,$F8,$82,$3F,$11,$FE,$77,$F0,$F2,$39,$1F,$EA,$9F
	.dc.b	$A9,$F6,$3F,$EB,$65,$F3,$DF,$8B,$6C,$6F,$3C,$19,$99,$19,$3F,$D6
	.dc.b	$9D,$5F,$99,$7F,$D8,$2D,$DF,$F8,$DB,$79,$FF,$C7,$48,$FC,$D1,$99
	.dc.b	$FF,$73,$B5,$D6,$9A,$8F,$D0,$7C,$F3,$E0,$69,$2B,$FD,$8D,$0F,$2B
	.dc.b	$11,$FD,$A2,$38,$3F,$B4,$6B,$99,$FD,$AC,$63,$1F,$E6,$6F,$FB,$A2
	.dc.b	$7F,$DF,$6D,$4E,$F1,$FF,$2B,$7D,$A3,$E8,$BF,$7E,$DA,$FE,$06,$D7
	.dc.b	$FB,$DB,$5F,$F0,$37,$D4,$3D,$3F,$E0,$ED,$8F,$E1,$6D,$7F,$C0,$6D
	.dc.b	$7F,$C1,$DB,$7F,$9B,$5F,$E2,$2D,$7F,$C1,$2D,$77,$E5,$7F,$14,$FB
	.dc.b	$47,$C2,$D2,$4F,$F1,$A8,$CB,$FC,$73,$85,$ED,$FF,$1C,$6C,$66,$30
	.dc.b	$7F,$20,$FA,$87,$D4,$D2,$4F,$F2,$68,$CB,$FC,$A3,$E7,$3F,$D3,$4F
	.dc.b	$F8,$A5,$AF,$FA,$0D,$A8,$38,$5F,$52,$FA,$9F,$E6,$2D,$8F,$E6,$AD
	.dc.b	$7F,$C5,$AD,$7F,$C7,$1F,$52,$F0,$BF,$9C,$B6,$3F,$9E,$B5,$FF,$1A
	.dc.b	$B5,$FF,$20,$7D,$A5,$7F,$D3,$4D,$7D,$0D,$4D,$FA,$A7,$CD,$7F,$48
	.dc.b	$6C,$FF,$4C,$6B,$FE,$44,$6B,$FE,$4F,$FA,$AF,$F4,$73,$7D,$6F,$EA
	.dc.b	$31,$88,$FF,$56,$35,$F9,$5F,$01,$FB,$B7,$FC,$B0,$D7,$FD,$64,$D7
	.dc.b	$7C,$3F,$A8,$E5,$91,$FA,$D6,$BF,$B1,$F8,$BF,$D6,$91,$FD,$93,$FD
	.dc.b	$93,$CE,$7C,$6D,$24,$FF,$6A,$8C,$BF,$DB,$3E,$A1,$F5,$3F,$DB,$1B
	.dc.b	$19,$8C,$1F,$DC,$3E,$A5,$F0,$9F,$D9,$67,$FD,$99,$18,$F5,$BF,$BB
	.dc.b	$AC,$1F,$15,$8B,$E1,$7F,$8F,$FB,$3F,$F7,$17,$89,$FD,$F9,$B1,$FE
	.dc.b	$76,$6B,$FE,$81,$6B,$FE,$DB,$FA,$9F,$E9,$53,$FE,$06,$6C,$7F,$9F
	.dc.b	$9A,$FF,$A2,$5A,$F1,$BF,$F4,$4B,$FA,$18,$F3,$CF,$39,$CC,$A7,$2F
	.dc.b	$FA,$2B,$EC,$FE,$1D,$7F,$85,$9B,$19,$C3,$8C,$62,$3F,$D3,$5F,$54
	.dc.b	$F7,$BA,$B6,$FF,$C3,$ED,$9F,$F4,$BB,$5F,$F5,$17,$D5,$3E,$3F,$FC
	.dc.b	$E1,$E6,$7F,$98,$3E,$1D,$F3,$F8,$C7,$FE,$00,$5E,$DF,$DF,$56,$E7
	.dc.b	$58,$7F,$01,$AF,$6C,$FF,$9A,$9E,$10,$8C,$6D,$C2,$13,$A9,$FF,$38
	.dc.b	$9F,$B0,$BE,$E7,$F0,$D9,$61,$FC,$3A,$33,$1E,$4E,$DE,$93,$98,$F0
	.dc.b	$4F,$F9,$5B,$E5,$3C,$17,$D6,$72,$7B,$23,$37,$E5,$87,$AF,$D3,$83
	.dc.b	$0D,$C3,$C3,$3F,$C8,$7D,$7C,$C6,$21,$18,$7B,$9F,$C9,$FD,$7E,$D1
	.dc.b	$DD,$B3,$C7,$7F,$5F,$2F,$8D,$B1,$D7,$7C,$34,$E0,$78,$70,$7D,$57
	.dc.b	$F9,$81,$A7,$F9,$89,$C1,$FE,$66,$FA,$8F,$EC,$1F,$55,$7D,$23,$49
	.dc.b	$5F,$EE,$29,$7C,$E7,$DD,$8C,$35,$FF,$78,$46,$B8,$BB,$4E,$67,$EE
	.dc.b	$4B,$E8,$FF,$CD,$8D,$8F,$9A,$D7,$E4,$BC,$D7,$D5,$7F,$A5,$23,$F9
	.dc.b	$9D,$FF,$AE,$9B,$9F,$6E,$3E,$B8,$76,$E4,$98,$F6,$7F,$CF,$9F,$55
	.dc.b	$F8,$F8,$EF,$E1,$F9,$08,$EC,$F2,$FD,$57,$A0,$8F,$A2,$3F,$D0,$DF
	.dc.b	$16,$E5,$A4,$9F,$C4,$FD,$57,$F8,$9D,$FC,$5C,$FE,$D5,$BC,$BC,$8E
	.dc.b	$0F,$E2,$BE,$AB,$E2,$70,$7F,$B6,$1B,$BF,$E9,$DB,$3F,$DD,$70,$FF
	.dc.b	$13,$3F,$DD,$E3,$E2,$B7,$7F,$F4,$E4,$62,$0F,$25,$F5,$5F,$E2,$9C
	.dc.b	$3F,$DC,$EF,$FA,$DD,$BF,$1F,$FE,$A5,$8D,$63,$FB,$97,$FD,$79,$5F
	.dc.b	$35,$AE,$67,$FD,$6B,$18,$8F,$F0,$67,$C9,$7F,$14,$7F,$D7,$6D,$7E
	.dc.b	$B0,$70,$FD,$96,$BF,$C2,$57,$CA,$FF,$B0,$9B,$1D,$23,$06,$7E,$26
	.dc.b	$47,$81,$AF,$82,$BF,$58,$47,$D2,$7F,$B3,$BC,$2F,$B8,$F9,$9A,$FE
	.dc.b	$35,$AF,$F1,$D6,$BF,$E2,$B7,$FC,$CA,$EF,$E0,$FA,$11,$FF,$6E,$C3
	.dc.b	$E6,$BC,$04,$8C,$BE,$3E,$C9,$1C,$1A,$F9,$7A,$FD,$04,$7A,$BF,$EE
	.dc.b	$8F,$A7,$FF,$1E,$C7,$FB,$B1,$AF,$D2,$6B,$FA,$96,$BF,$CE,$5F,$F2
	.dc.b	$CB,$BF,$80,$E8,$47,$A7,$CF,$F7,$23,$83,$FD,$F7,$F4,$DF,$E2,$14
	.dc.b	$D2,$7A,$9F,$EF,$E7,$FD,$85,$1F,$BB,$6B,$FD,$2D,$AF,$F9,$65,$AF
	.dc.b	$4F,$81,$E4,$47,$C2,$FE,$41,$FF,$30,$7D,$37,$F9,$A1,$FC,$9B,$5F
	.dc.b	$E1,$4D,$7F,$AD,$35,$E3,$A7,$35,$1F,$6D,$EC,$D7,$AC,$6E,$3F,$D2
	.dc.b	$11,$D3,$4C,$11,$F3,$6D,$FE,$4D,$7F,$D7,$1E,$CF,$78,$C1,$39,$70
	.dc.b	$49,$E2,$F9,$7E,$95,$F3,$34,$93,$F0,$E3,$2F,$ED,$BE,$0B,$FB,$C7
	.dc.b	$4A,$FF,$65,$18,$6C,$FF,$6A,$E8,$7D,$96,$63,$F8,$0D,$7E,$DC,$8F
	.dc.b	$52,$1F,$DB,$27,$F1,$B7,$DE,$3F,$84,$F1,$F3,$7F,$83,$55,$AF,$CC
	.dc.b	$D6,$E7,$F8,$55,$64,$D5,$48,$F3,$46,$5F,$44,$65,$F1,$46,$5F,$24
	.dc.b	$66,$7B,$57,$9B,$0F,$E2,$97,$97,$E1,$DF,$23,$B9,$D8,$A6,$49,$5D
	.dc.b	$2F,$C0,$F6,$AB,$5D,$23,$15,$C9,$AF,$98,$C4,$79,$1A,$7D,$22,$9F
	.dc.b	$18,$B7,$F2,$A3,$84,$79,$1B,$3E,$58,$DD,$79,$1E,$87,$AC,$FE,$70
	.dc.b	$FF,$35,$23,$FD,$01,$FB,$1C,$72,$39,$9F,$89,$70,$F8,$4F,$C0,$99
	.dc.b	$21,$93,$FA,$43,$C2,$7E,$AB,$E5,$8C,$9C,$23,$2F,$C0,$8E,$15,$FD
	.dc.b	$35,$E1,$67,$FA,$71,$27,$FA,$CB,$5F,$CC,$7D,$1F,$E4,$CE,$6F,$F6
	.dc.b	$08,$C3,$FA,$58,$CB,$E4,$38,$99,$3C,$D5,$FD,$7E,$38,$C6,$8F,$D3
	.dc.b	$4E,$2F,$F2,$73,$6B,$D5,$5C,$CE,$13,$A7,$A9,$FD,$D2,$78,$34,$C9
	.dc.b	$FD,$CD,$CB,$4C,$DA,$32,$D8,$CC,$60,$F7,$C6,$63,$C7,$5E,$D7,$67
	.dc.b	$9E,$BC,$CF,$B8,$C4,$6D,$E2,$7C,$D3,$27,$A4,$C9,$F3,$63,$99,$23
	.dc.b	$93,$F8,$28,$DC,$F9,$63,$51,$D4,$CD,$7B,$AF,$D9,$1E,$03,$B5,$F4
	.dc.b	$23,$E3,$AE,$C7,$47,$E5,$1E,$09,$E6,$D3,$A5,$F8,$EB,$9C,$76,$3A
	.dc.b	$38,$8D,$47,$9E,$BB,$46,$23,$BC,$78,$A7,$85,$FA,$E7,$BB,$5F,$79
	.dc.b	$E2,$BF,$2D,$74,$6D,$78,$5F,$CA,$CF,$F8,$DE,$BF,$C6,$84,$7F,$C7
	.dc.b	$ED,$7F,$C7,$F7,$F8,$37,$81,$A3,$FA,$57,$C2,$7F,$3D,$3B,$BF,$E4
	.dc.b	$A7,$0F,$A0,$D9,$0F,$49,$FD,$2D,$DB,$8C,$3F,$9C,$C6,$1F,$8F,$57
	.dc.b	$0F,$F2,$74,$F8,$2B,$FC,$A8,$E5,$FA,$24,$32,$78,$6B,$9B,$67,$9B
	.dc.b	$E7,$3C,$86,$4E,$57,$F5,$C5,$77,$8F,$B3,$3F,$7E,$31,$3B,$BB,$1E
	.dc.b	$78,$E3,$5D,$4E,$27,$68,$E3,$3F,$B8,$6C,$7E,$41,$AF,$E9,$1B,$7F
	.dc.b	$8C,$6D,$F8,$E7,$B5,$F8,$DB,$3E,$33,$6F,$84,$DD,$F8,$5A,$78,$4A
	.dc.b	$78,$4B,$7F,$61,$EA,$6C,$F9,$64,$7C,$CE,$8E,$47,$E4,$27,$FD,$10
	.dc.b	$47,$FD,$26,$FF,$83,$7D,$6F,$F5,$C4,$7F,$BF,$9B,$3E,$18,$ED,$81
	.dc.b	$93,$A8,$C4,$23,$0F,$E6,$4F,$A8,$E1,$D7,$83,$F7,$E3,$04,$70,$79
	.dc.b	$49,$1C,$9A,$F9,$DD,$BF,$E5,$D8,$F5,$BF,$EE,$36,$BE,$66,$BE,$A8
	.dc.b	$FF,$17,$7D,$77,$FB,$7E,$3D,$D7,$FE,$2D,$8C,$47,$73,$37,$CE,$B8
	.dc.b	$F8,$92,$72,$75,$29,$F0,$0B,$5D,$2F,$B4,$FB,$1B,$3A,$8E,$8F,$84
	.dc.b	$46,$9B,$7F,$4A,$BA,$36,$BA,$36,$7A,$10,$E9,$5C,$DF,$B6,$2B,$E0
	.dc.b	$B8,$6B,$CC,$FF,$6D,$BE,$07,$F0,$CB,$CE,$F9,$A7,$AA,$7D,$07,$CD
	.dc.b	$73,$3E,$38,$CD,$7A,$E7,$33,$EE,$3A,$DE,$2B,$94,$7C,$C9,$ED,$5C
	.dc.b	$63,$04,$39,$1E,$DA,$E5,$1B,$8C,$DF,$A5,$F2,$46,$5F,$85,$5F,$EA
	.dc.b	$19,$D4,$7C,$28,$C1,$F8,$CC,$24,$78,$70,$FF,$3C,$7B,$DF,$C6,$23
	.dc.b	$E7,$87,$0F,$D2,$6B,$FA,$E6,$BF,$B3,$5E,$E7,$FA,$CD,$B1,$FE,$BF
	.dc.b	$6B,$D9,$AF,$5F,$B5,$E0,$47,$83,$FE,$BD,$F8,$77,$CA,$D2,$4F,$FB
	.dc.b	$2E,$32,$DF,$63,$81,$1C,$1B,$FA,$0E,$84,$70,$7F,$B5,$7E,$1D,$F4
	.dc.b	$5F,$EC,$D6,$C6,$63,$06,$7E,$6A,$47,$16,$BF,$32,$5F,$35,$CC,$91
	.dc.b	$DD,$FF,$6D,$7C,$43,$DE,$A4,$9F,$F7,$44,$65,$BE,$E7,$02,$38,$37
	.dc.b	$F3,$9D,$08,$E0,$FF,$79,$7D,$93,$BB,$FD,$D4,$D8,$CC,$60,$CF,$74
	.dc.b	$8E,$2D,$7F,$13,$3F,$E0,$D7,$F9,$9B,$5F,$E7,$0D,$4D,$FB,$77,$A3
	.dc.b	$FE,$FB,$6C,$FE,$50,$38,$7F,$2E,$D7,$F8,$3A,$F4,$79,$12,$39,$46
	.dc.b	$5B,$E1,$F7,$23,$83,$C1,$F0,$FF,$4F,$E2,$2E,$E7,$3B,$7E,$C2,$BF
	.dc.b	$A8,$D6,$DF,$D5,$FD,$83,$EA,$3F,$DC,$D1,$FD,$5A,$31,$E0,$FB,$5B
	.dc.b	$1C,$27,$6F,$F5,$FF,$B3,$74,$62,$11,$8B,$F8,$CD,$8F,$DD,$B5,$FE
	.dc.b	$57,$F0,$EF,$9E,$E4,$DA,$F9,$CD,$9F,$E0,$6D,$78,$FA,$5C,$09,$3C
	.dc.b	$1E,$7E,$1A,$3F,$A7,$F6,$BD,$7C,$5F,$F2,$0C,$17,$8F,$13,$32,$BE
	.dc.b	$18,$7F,$58,$BF,$E7,$4D,$BE,$0F,$F8,$02,$33,$5D,$67,$2F,$F6,$AF
	.dc.b	$6E,$BD,$7C,$DB,$8A,$73,$8A,$73,$F0,$3A,$9B,$97,$F6,$AB,$E0,$FF
	.dc.b	$6C,$C1,$F1,$1C,$B9,$BB,$C6,$27,$E1,$C6,$3E,$23,$35,$67,$93,$C3
	.dc.b	$CC,$C5,$FF,$46,$0E,$30,$FC,$5E,$BE,$8E,$7E,$ED,$C3,$C9,$E5,$E1
	.dc.b	$A7,$9E,$3C,$B8,$3E,$23,$F5,$CF,$F8,$B5,$AE,$4C,$BF,$B3,$F8,$67
	.dc.b	$AE,$F8,$8C,$36,$BF,$A5,$CE,$A7,$27,$1F,$7B,$FC,$13,$3A,$6B,$FE
	.dc.b	$00,$94,$62,$3F,$C7,$B1,$8D,$3C,$5F,$0F,$FE,$40,$7F,$13,$08,$C6
	.dc.b	$7C,$1B,$5F,$E4,$99,$E4,$BE,$21,$F0,$BD,$B9,$37,$2A,$FC,$33,$FE
	.dc.b	$2A,$6B,$FD,$E9,$F0,$39,$E5,$E3,$06,$9D,$3E,$23,$FC,$3B,$9F,$A3
	.dc.b	$E1,$F6,$4B,$E4,$77,$F8,$8E,$64,$7C,$51,$E6,$CB,$FC,$41,$1E,$AE
	.dc.b	$5C,$1E,$AF,$88,$7C,$DE,$F9,$65,$7F,$91,$2F,$FC,$C8,$72,$A7,$C0
	.dc.b	$7C,$2F,$89,$F6,$D7,$F8,$61,$BA,$7E,$E2,$B7,$3F,$E4,$E9,$F8,$B3
	.dc.b	$FB,$C6,$F1,$FF,$9D,$9B,$3F,$E7,$B7,$9C,$7F,$9E,$BE,$21,$F3,$5F
	.dc.b	$94,$6D,$7F,$43,$6B,$FE,$18,$6B,$FE,$48,$73,$7F,$C3,$17,$C5,$7D
	.dc.b	$C6,$CF,$F9,$91,$DD,$FF,$A2,$BE,$0B,$F8,$C4,$FF,$9A,$1A,$FF,$8E
	.dc.b	$1B,$7F,$C6,$17,$D7,$FE,$21,$DC,$FF,$96,$9A,$F5,$F1,$7F,$5F,$18
	.dc.b	$8F,$F4,$E3,$FD,$6D,$F1,$7A,$3A,$23,$F0,$7C,$2F,$88,$47,$07,$C5
	.dc.b	$3E,$41,$0C,$1F,$DA,$EB,$F0,$6D,$E2,$FE,$03,$18,$8F,$F2,$1B,$5F
	.dc.b	$F3,$9B,$FE,$2E,$73,$EE,$70,$23,$D7,$E2,$39,$91,$C1,$E1,$F0,$3B
	.dc.b	$91,$C1,$E5,$F4,$3C,$64,$70,$7F,$80,$FE,$2F,$F9,$B3,$7C,$8F,$E3
	.dc.b	$31,$88,$FF,$2D,$B5,$FF,$45,$BF,$E1,$D7,$3E,$BF,$02,$3D,$78,$79
	.dc.b	$91,$C1,$FE,$13,$F8,$A7,$C5,$FF,$28,$6C,$66,$30,$67,$CF,$91,$C5
	.dc.b	$AF,$F9,$C3,$E2,$FF,$52,$DE,$BF,$E6,$71,$88,$FF,$51,$B5,$FF,$6F
	.dc.b	$AF,$97,$D9,$23,$FC,$51,$19,$6E,$FD,$08,$E0,$FF,$17,$53,$E0,$FF
	.dc.b	$A1,$B6,$33,$18,$33,$F1,$92,$38,$B5,$FF,$45,$CF,$F8,$5D,$FE,$17
	.dc.b	$A7,$42,$3F,$E3,$DE,$77,$B3,$FA,$92,$47,$FB,$63,$EA,$7F,$C8,$1D
	.dc.b	$7F,$AF,$9F,$D4,$92,$79,$35,$FD,$B7,$1F,$EC,$CE,$6F,$EE,$63,$0F
	.dc.b	$EF,$63,$33,$D6,$30,$43,$81,$FC,$0E,$78,$BD,$8C,$10,$C1,$23,$A3
	.dc.b	$C8,$C6,$1F,$C3,$EB,$C2,$FF,$11,$AE,$B5,$FC,$36,$73,$5F,$C1,$9C
	.dc.b	$D7,$18,$F0,$4E,$DF,$43,$5F,$03,$49,$1C,$63,$2D,$8C,$C7,$73,$AC
	.dc.b	$64,$8E,$4E,$25,$8E,$2B,$0F,$DA,$47,$C2,$3C,$91,$FC,$96,$BF,$91
	.dc.b	$91,$FC,$23,$FC,$AF,$8A,$47,$B2,$FC,$2F,$58,$CB,$F3,$0F,$14,$7F
	.dc.b	$3B,$8C,$D7,$92,$FC,$73,$9A,$FE,$7D,$5F,$31,$E2,$7F,$34,$78,$C7
	.dc.b	$82,$78,$CE,$4E,$27,$94,$C1,$FD,$1E,$78,$BD,$0F,$41,$C0,$D1,$C4
	.dc.b	$C9,$82,$57,$A6,$39,$C6,$A3,$0F,$37,$FA,$A4,$7A,$E3,$55,$EF,$79
	.dc.b	$BE,$B8,$E7,$5F,$C9,$6B,$9D,$68,$F5,$B8,$3A,$BD,$EB,$93,$4D,$B8
	.dc.b	$79,$1E,$23,$E1,$4F,$78,$F2,$4F,$49,$FB,$46,$C9,$3B,$78,$D7,$A0
	.dc.b	$F7,$46,$63,$71,$83,$4F,$CD,$8D,$12,$35,$1E,$93,$BD,$F0,$7D,$33
	.dc.b	$F8,$07,$A9,$B7,$07,$37,$11,$C6,$BD,$87,$C9,$BF,$79,$27,$D6,$73
	.dc.b	$BF,$1B,$E0,$6C,$F2,$9E,$13,$FA,$DA,$ED,$18,$BF,$1D,$E2,$BA,$9C
	.dc.b	$CF,$84,$E5,$E4,$7E,$4C,$FF,$0C,$3F,$E1,$62,$3F,$CD,$DA,$8D,$99
	.dc.b	$21,$FE,$27,$F7,$BF,$C4,$87,$79,$38,$8F,$80,$7F,$8B,$DC,$3E,$73
	.dc.b	$91,$93,$66,$09,$87,$C5,$7F,$22,$7F,$8B,$23,$9C,$66,$3F,$17,$18
	.dc.b	$BF,$91,$76,$78,$1D,$AB,$FC,$7D,$59,$8F,$C0,$9E,$0A,$FF,$24,$47
	.dc.b	$B4,$D5,$FE,$0E,$74,$74,$3D,$95,$D1,$F6,$18,$AE,$6E,$23,$D1,$3F
	.dc.b	$98,$AE,$B7,$FB,$86,$D7,$E4,$DA,$FE,$61,$B5,$C9,$B7,$F8,$76,$CF
	.dc.b	$C8,$3F,$9C,$38,$7E,$39,$E3,$8F,$8E,$D3,$D9,$3E,$6A,$F8,$35,$F0
	.dc.b	$E7,$47,$C8,$9F,$80,$FC,$43,$A9,$6F,$F0,$AF,$AC,$E2,$E1,$F6,$18
	.dc.b	$9F,$84,$FF,$82,$DF,$F3,$69,$1C,$1F,$E7,$BF,$8B,$FF,$3B,$1E,$67
	.dc.b	$04,$3C,$51,$A7,$15,$E8,$3F,$D0,$13,$98,$FF,$42,$3F,$80,$31,$1A
	.dc.b	$8F,$4C,$63,$0F,$F4,$73,$B8,$FF,$11,$3F,$E8,$42,$38,$3F,$D2,$5F
	.dc.b	$17,$FE,$8D,$39,$3F,$E8,$88,$C3,$A3,$CF,$3D,$9C,$1F,$E9,$99,$E2
	.dc.b	$FF,$75,$78,$B9,$38,$9F,$E9,$47,$35,$FE,$9F,$78,$E0,$73,$5F,$EA
	.dc.b	$97,$27,$FA,$78,$95,$FE,$3F,$6A,$A4,$FF,$0B,$87,$C5,$BC,$6F,$F7
	.dc.b	$9B,$63,$D7,$5E,$37,$F8,$93,$C9,$FE,$29,$1B,$7B,$57,$58,$C4,$F1
	.dc.b	$9F,$80,$68,$E3,$5E,$03,$DA,$E0,$EF,$5C,$E3,$51,$E4,$8C,$3A,$38
	.dc.b	$1D,$4F,$6C,$74,$69,$B2,$9B,$AF,$3C,$FC,$D8,$D4,$7B,$5F,$90,$70
	.dc.b	$2B,$97,$C6,$68,$EB,$5F,$1D,$7C,$5B,$C8,$FA,$53,$98,$C4,$F8,$30
	.dc.b	$CE,$1F,$5F,$E6,$DC,$BD,$6B,$FA,$24,$61,$C3,$5F,$13,$5F,$4F,$C5
	.dc.b	$FE,$2D,$E4,$D7,$D0,$D7,$7A,$77,$50,$F8,$CC,$36,$39,$C6,$0E,$1D
	.dc.b	$6E,$BA,$49,$F5,$46,$5F,$67,$C6,$7F,$B8,$5F,$9E,$D7,$F3,$0D,$7F
	.dc.b	$B2,$35,$2F,$73,$D0,$D8,$EB,$18,$3D,$9F,$32,$E4,$A4,$9F,$9D,$19
	.dc.b	$7E,$87,$C5,$BE,$37,$E2,$36,$33,$18,$33,$F1,$B2,$38,$B5,$FC,$37
	.dc.b	$CB,$BD,$AF,$9C,$D8,$FB,$CD,$77,$1F,$4F,$E3,$5F,$05,$49,$3F,$85
	.dc.b	$8C,$B7,$E0,$78,$11,$C1,$F8,$BE,$E7,$CB,$7D,$F6,$C6,$63,$07,$E0
	.dc.b	$FE,$37,$FD,$FD,$1F,$A4,$6B,$FC,$35,$AF,$F8,$65,$A9,$FE,$9C,$F1
	.dc.b	$F0,$3A,$11,$FC,$F7,$DC,$7E,$B5,$FF,$11,$35,$34,$C7,$C6,$BD,$3F
	.dc.b	$CC,$36,$7F,$50,$D7,$F6,$CD,$7F,$7A,$BE,$3B,$89,$23,$8B,$FA,$3F
	.dc.b	$97,$FD,$04,$7E,$FD,$AF,$F0,$66,$BF,$CB,$DA,$FF,$8E,$FE,$3F,$EC
	.dc.b	$FA,$3F,$BB,$6C,$7F,$09,$5B,$3E,$23,$5C,$2B,$FE,$05,$F1,$EF,$8F
	.dc.b	$FE,$1D,$8F,$F3,$D8,$C7,$89,$A9,$CC,$FF,$92,$E7,$2F,$F1,$7F,$8C
	.dc.b	$7C,$FE,$A1,$18,$BF,$E2,$0D,$8E,$13,$B8,$FD,$F4,$EB,$C1,$C1,$1E
	.dc.b	$3E,$3F,$F4,$9F,$27,$F9,$27,$82,$F6,$71,$5F,$D5,$70,$FE,$A8,$FF
	.dc.b	$37,$9D,$3F,$CE,$63,$39,$7F,$3F,$6D,$FF,$2F,$7E,$44,$EF,$0E,$B5
	.dc.b	$FD,$6B,$E2,$2B,$FC,$1F,$CF,$B8,$27,$38,$A7,$3F,$0F,$A9,$B8,$FF
	.dc.b	$4B,$BE,$0F,$F4,$DC,$1F,$1B,$CB,$8B,$BC,$62,$7D,$91,$8F,$7B,$35
	.dc.b	$67,$93,$C3,$D0,$C5,$FF,$5D,$3F,$A5,$4F,$7A,$EB,$AF,$A3,$9A,$CF
	.dc.b	$C9,$7C,$3C,$D8,$BE,$4D,$CF,$F0,$BF,$1B,$FC,$FE,$4D,$23,$FD,$77
	.dc.b	$E2,$9E,$F3,$E3,$7F,$AA,$46,$1B,$19,$9E,$0D,$7F,$B5,$CB,$E3,$7F
	.dc.b	$C2,$D1,$D5,$AE,$9A,$FF,$86,$63,$86,$1E,$0F,$1F,$FB,$3C,$FF,$A9
	.dc.b	$E7,$F9,$44,$FF,$71,$6E,$9C,$6F,$BF,$C6,$BE,$2B,$A7,$27,$2C,$58
	.dc.b	$7E,$0D,$FF,$04,$35,$FF,$07,$3D,$1F,$07,$2F,$18,$34,$E9,$F1,$BF
	.dc.b	$E0,$AC,$FC,$9F,$19,$B2,$5F,$6A,$E5,$F1,$B2,$3C,$2E,$9F,$F0,$F6
	.dc.b	$5F,$E0,$86,$F3,$7A,$63,$18,$7F,$84,$7E,$35,$F1,$19,$E7,$98,$7F
	.dc.b	$88,$EF,$FC,$42,$FF,$B3,$DE,$3C,$3F,$E4,$7B,$7C,$0F,$C5,$F8,$DF
	.dc.b	$81,$66,$CE,$A3,$FC,$5D,$1A,$3E,$46,$5F,$E2,$36,$DF,$F9,$4D,$E7
	.dc.b	$F1,$6F,$8D,$FA,$4D,$BF,$F1,$C3,$5F,$F1,$F2,$F9,$BD,$CF,$F8,$5E
	.dc.b	$35,$3F,$E3,$86,$BF,$E6,$16,$BF,$EE,$DF,$83,$7C,$6F,$F8,$7D,$B1
	.dc.b	$FE,$2F,$6B,$FE,$3D,$6B,$FE,$44,$7F,$8D,$3F,$E4,$D6,$BF,$E5,$36
	.dc.b	$E5,$FE,$43,$79,$B7,$C5,$F0,$12,$7C,$0F,$AF,$C1,$91,$FE,$41,$C3
	.dc.b	$F8,$6B,$63,$E0,$78,$32,$34,$D7,$FC,$A9,$3F,$BD,$23,$F1,$CF,$F1
	.dc.b	$F5,$7F,$8F,$9F,$F2,$73,$5F,$F2,$E3,$5F,$F3,$1B,$5F,$F4,$63,$FE
	.dc.b	$33,$27,$C6,$F4,$3A,$FC,$1C,$9E,$AD,$7A,$F3,$78,$88,$F8,$9F,$2F
	.dc.b	$99,$23,$CA,$D7,$FC,$E1,$F1,$BF,$CD,$DF,$F3,$33,$5C,$F8,$FF,$E5
	.dc.b	$B8,$C4,$7F,$9F,$1F,$F1,$3B,$CB,$C7,$D1,$1E,$7C,$9C,$48,$E0,$FF
	.dc.b	$3A,$7C,$6B,$E0,$7F,$CD,$0D,$8C,$C6,$0C,$F6,$48,$E2,$D7,$FD,$15
	.dc.b	$F1,$9F,$AB,$6F,$57,$F9,$CE,$31,$1F,$E9,$46,$BF,$EB,$D5,$F1,$BB
	.dc.b	$24,$7F,$A2,$63,$2F,$FA,$2E,$7E,$1B,$FE,$C5,$6A,$3F,$80,$E9,$FE
	.dc.b	$83,$8C,$C7,$F0,$4F,$8D,$C1,$1E,$8E,$FE,$36,$47,$76,$BF,$EA,$7F
	.dc.b	$8D,$FF,$1A,$74,$7F,$A9,$DB,$B7,$FB,$5D,$B3,$E3,$F9,$47,$47,$F3
	.dc.b	$79,$FF,$1D,$4E,$FD,$AF,$F4,$C4,$62,$3D,$4D,$4F,$45,$D6,$F8,$EF
	.dc.b	$E8,$4D,$8F,$F6,$44,$EC,$E1,$19,$8F,$DF,$4F,$8E,$A4,$7F,$03,$8E
	.dc.b	$66,$1F,$F6,$B5,$72,$8C,$C7,$69,$D6,$5F,$C1,$E7,$33,$C2,$73,$97
	.dc.b	$FB,$EF,$0F,$15,$EE,$B9,$1E,$12,$1E,$63,$25,$23,$C5,$EC,$70,$70
	.dc.b	$49,$F9,$D1,$97,$C5,$7E,$69,$FB,$57,$67,$35,$B3,$E9,$C6,$5F,$E6
	.dc.b	$55,$5F,$AB,$16,$FD,$B5,$CA,$7C,$A7,$A4,$FE,$65,$5F,$CE,$48,$FF
	.dc.b	$39,$7E,$DE,$D2,$3C,$47,$E0,$23,$0F,$F3,$78,$F0,$46,$30,$FE,$71
	.dc.b	$18,$9F,$E7,$CE,$6B,$C9,$3F,$16,$7F,$A6,$4E,$5F,$69,$E0,$BF,$E2
	.dc.b	$93,$D2,$31,$7B,$7F,$27,$5F,$04,$C5,$70,$BE,$E7,$E0,$63,$26,$E7
	.dc.b	$CA,$43,$EE,$C6,$DC,$5F,$85,$C9,$E8,$33,$1B,$72,$79,$0C,$9B,$33
	.dc.b	$5E,$57,$13,$D4,$F8,$8F,$13,$E7,$1C,$8F,$A4,$58,$E4,$EE,$38,$C6
	.dc.b	$09,$5E,$F9,$E6,$49,$DC,$7C,$48,$FA,$0D,$7C,$ED,$8D,$C7,$86,$3A
	.dc.b	$47,$A6,$BE,$13,$A9,$ED,$18,$3E,$9B,$F0,$1C,$34,$FC,$09,$C0,$C1
	.dc.b	$ED,$78,$B4,$CD,$7B,$EF,$CD,$3D,$27,$E3,$9C,$CE,$04,$9F,$D3,$BC
	.dc.b	$AF,$F1,$35,$CA,$3A,$B5,$E0,$D9,$F0,$47,$C3,$3E,$A4,$66,$7E,$2D
	.dc.b	$78,$EB,$75,$F4,$8F,$F1,$24,$FF,$8A,$C8,$FF,$8A,$DA,$FF,$8A,$DA
	.dc.b	$F0,$3F,$C5,$5F,$01,$FC,$C8,$F2,$15,$F2,$1F,$E2,$D7,$0F,$C3,$9F
	.dc.b	$C3,$46,$30,$FB,$53,$F3,$23,$17,$E9,$BB,$3F,$E4,$39,$EE,$F8,$8E
	.dc.b	$F5,$FE,$39,$8D,$C7,$CF,$9B,$69,$E3,$3E,$53,$8D,$FF,$37,$8D,$D7
	.dc.b	$EF,$1B,$3F,$A1,$6B,$EB,$6D,$FA,$A3,$FA,$BC,$62,$BE,$93,$63,$DC
	.dc.b	$7A,$4F,$89,$5E,$68,$F2,$D7,$AE,$7C,$A7,$C1,$38,$3D,$EF,$E5,$C6
	.dc.b	$2F,$E4,$34,$E0,$76,$30,$79,$8E,$A5,$3E,$21,$B2,$CF,$8D,$C3,$49
	.dc.b	$1E,$08,$DB,$F2,$DE,$B3,$F4,$5F,$0B,$FC,$19,$B5,$F3,$DA,$7C,$F2
	.dc.b	$9F,$3C,$B5,$F0,$63,$E4,$46,$CF,$09,$F0,$23,$4F,$0A,$FA,$6D,$3E
	.dc.b	$99,$FE,$89,$9F,$F4,$71,$1F,$F4,$73,$FE,$10,$F7,$A4,$71,$3F,$C2
	.dc.b	$AE,$1F,$09,$F2,$EB,$FD,$19,$19,$7E,$3C,$7F,$A7,$63,$15,$FE,$07
	.dc.b	$7F,$2E,$62,$34,$7C,$87,$0F,$80,$C1,$FE,$A3,$7F,$D4,$C4,$70,$7F
	.dc.b	$A8,$7E,$1F,$FC,$52,$72,$C1,$38,$9D,$1F,$EA,$77,$53,$98,$FF,$54
	.dc.b	$45,$BF,$F5,$44,$DB,$F5,$57,$23,$FD,$5D,$19,$7D,$66,$0D,$4F,$27
	.dc.b	$C0,$71,$3F,$D7,$51,$FE,$C1,$23,$FE,$C1,$7F,$C8,$7F,$17,$23,$9D
	.dc.b	$7F,$99,$57,$CF,$7F,$B4,$DC,$19,$70,$E0,$DF,$89,$5F,$F3,$74,$69
	.dc.b	$E3,$F1,$AE,$FD,$42,$30,$F1,$6B,$D5,$AF,$8B,$E3,$58,$F2,$6C,$78
	.dc.b	$5A,$F9,$5A,$FA,$BE,$77,$FC,$FE,$4E,$0F,$1B,$FE,$82,$EC,$7B,$9E
	.dc.b	$66,$D7,$B5,$AF,$C1,$6B,$F1,$BE,$35,$E2,$7B,$9B,$1F,$15,$AF,$C9
	.dc.b	$6B,$F3,$B0,$AF,$D3,$6B,$FE,$99,$6A,$6F,$C6,$B8,$7E,$6B,$67,$EA
	.dc.b	$35,$FB,$0D,$7E,$EF,$C6,$BE,$37,$EA,$B6,$3E,$E3,$5F,$C0,$35,$FC
	.dc.b	$2F,$CE,$BE,$5F,$EE,$B6,$3F,$08,$D7,$F1,$0D,$7F,$1B,$F1,$CF,$73
	.dc.b	$F0,$AD,$8F,$C6,$35,$FC,$83,$5F,$CA,$FC,$77,$FB,$81,$BE,$4F,$E3
	.dc.b	$A3,$11,$F9,$D6,$BF,$EC,$3F,$B9,$7D,$87,$E5,$1B,$1F,$9B,$6B,$F9
	.dc.b	$F6,$BF,$A4,$EF,$7C,$87,$E7,$1B,$1F,$A3,$6B,$FA,$76,$BF,$AC,$F9
	.dc.b	$0F,$E7,$CF,$EA,$DA,$E6,$3F,$4F,$18,$8F,$DB,$72,$3E,$DB,$F5,$6D
	.dc.b	$8F,$D9,$35,$FD,$B3,$5F,$DD,$FC,$8B,$DB,$FD,$9B,$63,$F7,$4D,$7F
	.dc.b	$7C,$D7,$F8,$1F,$C8,$FF,$B4,$49,$E3,$FE,$F1,$FF,$6A,$7C,$4F,$BA
	.dc.b	$7F,$86,$35,$FF,$79,$35,$37,$E4,$1F,$27,$FC,$19,$B3,$FC,$35,$AF
	.dc.b	$F1,$16,$BF,$C5,$BE,$45,$C6,$F5,$FF,$8A,$46,$1B,$83,$E3,$33,$5B
	.dc.b	$BF,$E3,$9F,$3C,$FA,$27,$97,$FC,$76,$30,$D8,$CC,$F0,$6B,$FC,$A3
	.dc.b	$E4,$1F,$23,$88,$FE,$5D,$18,$AF,$E5,$0D,$8F,$E5,$8D,$7F,$98,$FC
	.dc.b	$8B,$DF,$EF,$E8,$76,$8C,$5A,$93,$E2,$BF,$BD,$39,$8E,$17,$CA,$3C
	.dc.b	$7C,$B6,$F9,$5F,$07,$9C,$FA,$35,$D4,$27,$58,$7C,$01,$F1,$9A,$CB
	.dc.b	$F8,$B6,$1E,$13,$FA,$74,$E9,$C5,$FA,$67,$6F,$80,$F4,$7C,$4F,$82
	.dc.b	$B8,$46,$4F,$7C,$66,$FC,$B0,$F8,$3F,$17,$06,$1B,$8F,$82,$7E,$47
	.dc.b	$B1,$98,$C4,$23,$0F,$71,$EF,$F6,$8E,$ED,$9E,$3B,$7C,$1E,$3F,$1B
	.dc.b	$63,$AE,$F8,$69,$C0,$F0,$E0,$F9,$1F,$EC,$ED,$3F,$B4,$B8,$3F,$B8
	.dc.b	$7C,$6B,$E4,$1F,$23,$86,$D7,$F7,$79,$D4,$E4,$E3,$F2,$3F,$88,$9D
	.dc.b	$35,$FF,$00,$CA,$31,$1F,$E0,$78,$C7,$17,$8B,$C9,$FB,$B9,$65,$B5
	.dc.b	$F9,$5A,$FE,$75,$5F,$A6,$78,$C3,$E4,$5F,$1D,$D7,$93,$8C,$AB,$F1
	.dc.b	$AF,$D9,$6B,$F8,$37,$AE,$3E,$5E,$5E,$30,$69,$D3,$E4,$7F,$C3,$B9
	.dc.b	$78,$FE,$47,$64,$C3,$B5,$72,$F8,$B9,$1E,$17,$4F,$E4,$F3,$FC,$2B
	.dc.b	$F8,$D6,$B9,$E5,$F5,$FC,$83,$E4,$7D,$B5,$FC,$BF,$2F,$E1,$2D,$D7
	.dc.b	$FC,$68,$DC,$3F,$7C,$F4,$9F,$E1,$F6,$F8,$9F,$8F,$F2,$3F,$0A,$FF
	.dc.b	$82,$35,$4E,$DF,$D1,$46,$4F,$97,$5F,$C6,$9B,$87,$F9,$15,$E5,$F6
	.dc.b	$8F,$94,$FA,$AB,$E8,$29,$88,$B8,$7F,$91,$67,$4D,$9F,$F2,$1C,$62
	.dc.b	$3F,$C9,$9F,$24,$F9,$4F,$F2,$3B,$63,$FC,$9C,$D7,$71,$A3,$F0,$F0
	.dc.b	$F9,$AA,$62,$2C,$F2,$8E,$A4,$7A,$F7,$78,$08,$E0,$87,$F1,$1A,$FE
	.dc.b	$94,$FF,$0B,$6B,$98,$F1,$F7,$7A,$89,$1E,$A7,$FC,$C7,$E2,$BE,$76
	.dc.b	$92,$7F,$CD,$91,$96,$C7,$E9,$21,$F2,$74,$C4,$59,$E3,$1D,$88,$FF
	.dc.b	$9D,$FE,$55,$F3,$1F,$E7,$46,$C6,$63,$04,$7F,$94,$43,$E4,$3F,$6A
	.dc.b	$D8,$CC,$E2,$7A,$12,$3C,$0F,$FA,$0B,$E5,$BF,$01,$1F,$E8,$E6,$BF
	.dc.b	$E9,$36,$BF,$D7,$5A,$9F,$C4,$72,$7F,$70,$7F,$B7,$C7,$F3,$D6,$BF
	.dc.b	$D1,$DD,$F2,$75,$F3,$3C,$44,$9F,$F4,$D4,$F0,$7F,$D4,$3F,$2C,$F3
	.dc.b	$3F,$D3,$2D,$8F,$F5,$03,$5F,$F5,$23,$5E,$31,$FC,$6A,$3E,$D4,$7F
	.dc.b	$59,$6E,$1F,$E0,$87,$93,$75,$EE,$49,$EE,$FF,$AA,$BA,$7F,$C5,$31
	.dc.b	$FE,$B1,$6B,$FE,$B5,$6B,$FE,$BA,$6B,$FE,$BF,$77,$3F,$E1,$85,$F3
	.dc.b	$DF,$CC,$1B,$3F,$D9,$DD,$E7,$C6,$7C,$44,$9F,$13,$FE,$BE,$F9,$37
	.dc.b	$CC,$7F,$3A,$7F,$C3,$53,$FD,$C5,$AF,$F7,$B7,$6E,$78,$B8,$11,$E1
	.dc.b	$35,$E1,$17,$1F,$F1,$8D,$72,$9C,$9C,$C8,$77,$AE,$18,$7F,$B8,$67
	.dc.b	$C7,$5C,$AF,$D1,$5A,$BF,$F1,$75,$6A,$F8,$56,$A7,$AD,$F1,$6C,$F5
	.dc.b	$BE,$50,$BE,$4F,$4B,$E4,$FA,$09,$7C,$DF,$41,$D0,$E0,$53,$A5,$76
	.dc.b	$AE,$B1,$DE,$BA,$D7,$32,$5F,$AA,$7E,$69,$CE,$B8,$92,$FA,$5F,$56
	.dc.b	$CE,$DF,$A8,$FF,$94,$CF,$29,$4A,$59,$FC,$0B,$67,$F0,$DE,$C4,$8F
	.dc.b	$14,$57,$F0,$D1,$E0,$69,$F8,$B7,$18,$7E,$09,$E0,$FB,$A3,$C6,$E5
	.dc.b	$E0,$76,$29,$D0,$FA,$38,$7C,$57,$11,$C5,$C9,$F8,$49,$EF,$3F,$8B
	.dc.b	$7E,$91,$88,$D1,$F3,$5C,$3F,$83,$76,$F8,$A7,$B1,$E9,$74,$66,$7B
	.dc.b	$3C,$CC,$1D,$8F,$29,$C0,$F4,$BC,$1C,$DA,$BC,$6D,$8E,$87,$03,$A5
	.dc.b	$78,$E3,$51,$93,$B4,$66,$79,$9E,$C8,$D4,$F0,$9D,$47,$A0,$E3,$5F
	.dc.b	$3E,$3E,$0B,$E7,$6C,$7D,$A8,$C1,$1F,$2B,$C4,$C3,$D0,$C9,$C4,$A6
	.dc.b	$49,$3F,$A0,$AF,$4C,$7C,$2A,$F0,$BA,$BC,$BE,$C6,$C6,$AB,$94,$FB
	.dc.b	$5F,$81,$7E,$58,$F0,$C6,$1B,$5C,$A7,$D3,$3C,$A7,$E4,$DF,$CB,$BE
	.dc.b	$CD,$AD,$B6,$7A,$B5,$F6,$BF,$4D,$FD,$93,$FC,$DC,$8F,$ED,$CF,$C5
	.dc.b	$1A,$30,$7E,$2C,$C1,$BA,$F5,$3F,$C7,$78,$7F,$A0,$1E,$B7,$F9,$4B
	.dc.b	$87,$E2,$4D,$7F,$A1,$46,$DE,$47,$F4,$77,$13,$FE,$16,$8D,$9C,$8D
	.dc.b	$10,$F8,$E5,$BF,$E8,$B3,$C8,$F0,$3F,$2A,$35,$7F,$D0,$E7,$51,$83
	.dc.b	$E5,$94,$F2,$9D,$8F,$19,$27,$A3,$67,$A3,$E7,$30,$7C,$43,$11,$E6
	.dc.b	$7B,$57,$CC,$25,$7A,$2B,$EF,$CE,$A3,$CE,$F4,$3E,$F9,$93,$AC,$78
	.dc.b	$9D,$4F,$5B,$E9,$1E,$33,$D9,$7E,$37,$71,$FC,$1A,$F7,$1D,$8E,$6F
	.dc.b	$8A,$78,$1F,$08,$B7,$E1,$70,$D2,$4F,$CE,$39,$9E,$83,$C9,$1C,$9F
	.dc.b	$D4,$CF,$8A,$FF,$09,$3F,$B1,$8C,$1E,$79,$F8,$47,$F1,$27,$2F,$ED
	.dc.b	$5A,$E4,$DC,$73,$6D,$7E,$D2,$7F,$C3,$C4,$8F,$D2,$D7,$E5,$5F,$EB
	.dc.b	$D1,$FE,$DA,$F9,$B7,$9F,$4E,$F8,$30,$FE,$D1,$3A,$B0,$9D,$38,$8D
	.dc.b	$3F,$18,$D4,$F2,$AE,$D1,$E8,$38,$38,$78,$57,$45,$5D,$23,$0D,$AF
	.dc.b	$AB,$3F,$A0,$9C,$D7,$37,$EA,$46,$67,$C4,$FA,$A3,$6F,$EA,$6B,$A3
	.dc.b	$E4,$35,$3C,$1F,$CD,$46,$9F,$2D,$77,$9F,$1C,$E4,$91,$F9,$46,$CF
	.dc.b	$43,$B1,$D2,$39,$38,$AF,$4C,$F0,$8F,$96,$66,$BB,$D7,$F2,$28,$F1
	.dc.b	$9F,$C9,$27,$E9,$BF,$E1,$F6,$BE,$BA,$F9,$7E,$87,$E0,$CF,$0C,$3E
	.dc.b	$47,$FD,$76,$FF,$2E,$6B,$FC,$CD,$AF,$F8,$AD,$AF,$F8,$BF,$E4,$1F
	.dc.b	$2B,$E5,$6C,$7F,$38,$6B,$FC,$F5,$AF,$F8,$D5,$7C,$67,$12,$47,$17
	.dc.b	$E0,$7C,$78,$A4,$8F,$E8,$AE,$5F,$E8,$FF,$1E,$AF,$90,$D8,$CC,$60
	.dc.b	$CF,$D2,$48,$E2,$D7,$FA,$A7,$C7,$FA,$1F,$F2,$73,$5F,$D3,$B5,$37
	.dc.b	$E3,$9F,$3D,$F6,$5B,$3F,$D7,$1A,$FF,$61,$6B,$FE,$59,$5F,$67,$C4
	.dc.b	$91,$C5,$FC,$27,$C7,$3D,$8A,$49,$FE,$D3,$19,$7F,$B5,$FC,$73,$E2
	.dc.b	$7F,$18,$D8,$CC,$60,$CF,$C0,$C8,$E2,$D7,$FB,$C7,$C5,$BE,$CF,$F2
	.dc.b	$CD,$8F,$EF,$2D,$77,$1F,$9B,$F8,$CF,$F7,$D4,$7F,$80,$5A,$FF,$9E
	.dc.b	$5A,$FF,$9F,$5A,$FF,$16,$5F,$27,$CC,$91,$FE,$08,$8C,$BF,$E0,$AF
	.dc.b	$8C,$7B,$1F,$AB,$6C,$66,$30,$7E,$BF,$E3,$3F,$D7,$91,$FE,$8B,$6B
	.dc.b	$FE,$8F,$6B,$FC,$A1,$A9,$FF,$47,$BB,$F0,$BA,$11,$FE,$73,$97,$FB
	.dc.b	$64,$8E,$0F,$E7,$BD,$1F,$ED,$D2,$38,$3F,$C4,$BF,$40,$F4,$3F,$83
	.dc.b	$36,$33,$18,$3F,$85,$7C,$5B,$E2,$1F,$17,$FE,$2E,$8C,$37,$E2,$5E
	.dc.b	$A6,$6B,$78,$7F,$8D,$FE,$7D,$ED,$FF,$3B,$8F,$F5,$94,$63,$C8,$FF
	.dc.b	$1C,$B6,$38,$4E,$DF,$E9,$FE,$8B,$83,$FA,$7F,$8F,$FD,$16,$31,$87
	.dc.b	$F2,$BA,$7C,$5F,$BE,$31,$3A,$9F,$E5,$26,$B7,$E7,$3F,$EC,$5A,$E1
	.dc.b	$3A,$7D,$B1,$D5,$CD,$FF,$55,$1F,$0D,$CA,$BC,$FD,$15,$FE,$CD,$17
	.dc.b	$1F,$C9,$E1,$E7,$B7,$C1,$ED,$E7,$19,$8C,$57,$3B,$79,$5C,$F7,$C3
	.dc.b	$6F,$E1,$4F,$F7,$C6,$C7,$1F,$33,$71,$AE,$9C,$34,$E3,$1D,$F6,$F4
	.dc.b	$71,$D9,$F9,$6F,$8F,$8F,$18,$FC,$C3,$E3,$96,$BD,$0B,$E2,$FF,$AB
	.dc.b	$34,$FE,$B0,$E0,$FF,$3D,$FC,$3B,$E2,$DF,$17,$86,$D7,$FA,$0E,$75
	.dc.b	$39,$38,$F8,$BF,$E1,$59,$D3,$5F,$F8,$01,$3F,$C8,$F8,$B6,$78,$3E
	.dc.b	$2F,$FC,$3B,$3F,$86,$94,$62,$7F,$C2,$91,$89,$FC,$83,$CD,$7C,$5B
	.dc.b	$E1,$F9,$CF,$E1,$A7,$7C,$9C,$F8,$32,$D8,$FF,$16,$CF,$57,$C5,$CB
	.dc.b	$95,$1F,$88,$7C,$31,$9F,$8B,$F4,$C7,$A7,$E2,$F8,$51,$A6,$2B,$3F
	.dc.b	$17,$2B,$D6,$E9,$FD,$66,$5F,$C8,$9F,$D4,$B7,$4F,$F1,$B3,$73,$F9
	.dc.b	$3F,$14,$F8,$BF,$44,$FF,$0F,$BE,$3C,$98,$33,$D1,$ED,$B7,$BD,$FE
	.dc.b	$1B,$6D,$7F,$AF,$9A,$FF,$B0,$DA,$FC,$0F,$8B,$F9,$0D,$E3,$FF,$12
	.dc.b	$46,$23,$EE,$B5,$FF,$2D,$FC,$57,$F1,$67,$FD,$96,$D7,$FD,$A0,$D7
	.dc.b	$E6,$B5,$FC,$3B,$BD,$7F,$CB,$0B,$E2,$FE,$C3,$67,$F8,$B3,$BD,$BF
	.dc.b	$1F,$F1,$2F,$8B,$FB,$CD,$AF,$E3,$8D,$7F,$8C,$AF,$9B,$DC,$FF,$90
	.dc.b	$63,$53,$F8,$26,$BD,$A3,$47,$6F,$8C,$91,$DA,$3C,$3C,$12,$3C,$2F
	.dc.b	$93,$87,$11,$FE,$F3,$BF,$B8,$F7,$9F,$11,$1F,$4F,$77,$B0,$E3,$AF
	.dc.b	$F9,$F7,$C6,$7C,$37,$E7,$5B,$5F,$CE,$DA,$FF,$38,$5F,$23,$B9,$FF
	.dc.b	$30,$46,$A7,$F4,$2D,$7F,$76,$D7,$FD,$32,$F4,$AF,$17,$C6,$F8,$4E
	.dc.b	$93,$FD,$2F,$E3,$7F,$D7,$3C,$1F,$C1,$1B,$1F,$EA,$26,$BF,$D5,$7A
	.dc.b	$FF,$6D,$1F,$EA,$F8,$E1,$F0,$BF,$E7,$58,$C4,$71,$8C,$1C,$3E,$36
	.dc.b	$47,$86,$3F,$80,$7C,$67,$E2,$63,$FA,$CB,$72,$FF,$53,$37,$1F,$37
	.dc.b	$B3,$C3,$9F,$FD,$09,$3C,$6F,$F8,$2B,$5F,$E3,$ED,$7F,$D7,$AF,$36
	.dc.b	$FC,$6F,$88,$E8,$FF,$6F,$F8,$D7,$9D,$FC,$55,$B1,$98,$F4,$F9,$99
	.dc.b	$7F,$D2,$D3,$A9,$FE,$2E,$1C,$3F,$C6,$9A,$FF,$34,$78,$67,$FE,$C5
	.dc.b	$7E,$D6,$79,$E2,$ED,$5F,$E0,$3E,$F7,$99,$FC,$A5,$B1,$98,$C7,$06
	.dc.b	$5F,$F5,$54,$EA,$7C,$AD,$7F,$96,$B5,$FE,$8E,$EF,$1F,$F6,$C2,$F8
	.dc.b	$DF,$01,$D6,$7F,$C2,$DF,$16,$F8,$DF,$E7,$8D,$8C,$C7,$F3,$FF,$9D
	.dc.b	$FE,$FF,$18,$3F,$DF,$71,$FE,$BF,$9F,$F2,$59,$1E,$7D,$12,$78,$1F
	.dc.b	$E2,$09,$EC,$D8,$D3,$99,$F1,$98,$C7,$FD,$E5,$5F,$E3,$19,$E7,$18
	.dc.b	$9D,$6B,$FD,$A2,$B5,$5C,$CD,$C6,$6F,$70,$13,$E4,$BE,$91,$F1,$A7
	.dc.b	$6E,$AF,$CA,$E4,$87,$72,$4F,$A2,$78,$D7,$5B,$F3,$5E,$27,$9D,$F0
	.dc.b	$9F,$33,$D2,$FC,$6E,$4F,$45,$FB,$A3,$B4,$7B,$8C,$DF,$57,$07,$A4
	.dc.b	$E9,$5D,$9E,$87,$C2,$BE,$0E,$2F,$B4,$EE,$FB,$CE,$6B,$71,$E6,$32
	.dc.b	$49,$E0,$7C,$09,$F0,$9E,$A9,$F8,$B7,$CA,$77,$3E,$A3,$67,$94,$B7
	.dc.b	$F2,$DE,$46,$2B,$E5,$DF,$86,$32,$7E,$3C,$F8,$73,$EC,$73,$1B,$7E
	.dc.b	$31,$B2,$39,$8F,$04,$78,$CC,$D7,$A1,$E3,$3B,$AF,$6C,$79,$4F,$4D
	.dc.b	$FB,$DF,$0D,$7C,$57,$99,$82,$1D,$1E,$D7,$E9,$8F,$15,$FD,$19,$CD
	.dc.b	$F8,$27,$99,$A8,$E3,$7E,$18,$C1,$F3,$CD,$DF,$52,$46,$DB,$1E,$B3
	.dc.b	$71,$9A,$FA,$35,$E2,$9F,$79,$BB,$F9,$55,$B8,$FA,$75,$D6,$31,$1F
	.dc.b	$8D,$38,$57,$C5,$AF,$99,$1D,$0F,$3D,$7B,$E3,$31,$E4,$38,$9F,$2A
	.dc.b	$FE,$C9,$ED,$9E,$15,$E7,$9E,$07,$43,$B5,$7E,$81,$FF,$2E,$D7,$FB
	.dc.b	$A1,$FC,$69,$8A,$FE,$48,$47,$FA,$69,$F8,$73,$46,$23,$E7,$7C,$77
	.dc.b	$F3,$23,$E7,$16,$3D,$27,$F1,$67,$0F,$94,$FE,$58,$E2,$BE,$09,$FC
	.dc.b	$A1,$C5,$FB,$63,$13,$FD,$0A,$7B,$B4,$EF,$48,$E3,$1F,$26,$EB,$E3
	.dc.b	$37,$18,$9F,$99,$1C,$CF,$29,$A3,$15,$E6,$BF,$E7,$15,$C6,$FF,$93
	.dc.b	$B6,$BF,$7E,$D7,$F2,$AD,$AF,$DA,$36,$BF,$4A,$D9,$FC,$03,$5F,$CF
	.dc.b	$9E,$93,$CA,$7C,$B9,$F8,$AE,$1F,$33,$FC,$C2,$39,$5F,$D8,$8F,$5D
	.dc.b	$7C,$D3,$D6,$E0,$F9,$05,$3C,$15,$D1,$B1,$C1,$F1,$36,$3C,$66,$E7
	.dc.b	$EA,$4F,$37,$E7,$BE,$79,$E0,$78,$DF,$41,$BB,$FA,$D7,$EA,$2C,$7C
	.dc.b	$E3,$A3,$F4,$9A,$64,$EC,$74,$3E,$99,$A2,$9F,$E2,$CA,$F3,$B6,$3F
	.dc.b	$C1,$04,$7F,$C2,$7C,$DF,$E1,$23,$C6,$FF,$81,$DC,$3D,$70,$FF,$01
	.dc.b	$C6,$23,$C6,$7C,$A3,$FB,$2B,$98,$F9,$38,$7C,$49,$C9,$E0,$3A,$3F
	.dc.b	$E0,$99,$E0,$EE,$3F,$C0,$B1,$87,$89,$E0,$BF,$F3,$89,$23,$F4,$46
	.dc.b	$E3,$89,$F5,$DC,$47,$0B,$B1,$86,$CF,$94,$ED,$18,$69,$FC,$25,$FF
	.dc.b	$19,$11,$C1,$FE,$1D,$B5,$FF,$87,$E1,$F1,$CE,$3F,$F7,$E3,$63,$C1
	.dc.b	$5C,$1F,$F1,$37,$CE,$3E,$5F,$F8,$DC,$23,$0F,$F1,$E6,$BF,$E6,$C6
	.dc.b	$BE,$11,$F2,$0E,$DE,$6D,$8F,$E4,$ED,$7F,$CE,$AD,$79,$7C,$6B,$E4
	.dc.b	$3C,$6D,$8F,$E5,$ED,$7F,$CF,$AD,$7C,$5F,$34,$E7,$F5,$B6,$3F,$9B
	.dc.b	$B5,$FF,$42,$B5,$F5,$7C,$88,$F8,$65,$7F,$9F,$B5,$FF,$46,$B5,$F8
	.dc.b	$5F,$22,$BF,$96,$D8,$FE,$8E,$D7,$FD,$2A,$D7,$E5,$7B,$AF,$91,$FA
	.dc.b	$6D,$8F,$E9,$ED,$7F,$D3,$AD,$7E,$97,$C8,$FC,$37,$F0,$8D,$7F,$CC
	.dc.b	$8D,$4D,$F9,$07,$C8,$FE,$09,$B3,$FD,$79,$AF,$FA,$B1,$AF,$E0,$7E
	.dc.b	$61,$F4,$1F,$8A,$6C,$7F,$66,$6B,$FE,$B4,$6B,$F8,$9F,$92,$7B,$FF
	.dc.b	$92,$6C,$7F,$6E,$6B,$FE,$BC,$6B,$F9,$1F,$92,$7C,$97,$E6,$9B,$1F
	.dc.b	$DD,$9A,$FF,$B1,$1A,$FE,$67,$E6,$1F,$45,$FA,$26,$C7,$F7,$E6,$BF
	.dc.b	$EC,$C6,$BF,$A1,$F9,$07,$CA,$7E,$A9,$B1,$FE,$06,$6B,$FE,$D4,$6B
	.dc.b	$FA,$9F,$94,$FC,$93,$FB,$B6,$BF,$E9,$B6,$A6,$FC,$CB,$EB,$BF,$72
	.dc.b	$D9,$FF,$0A,$B5,$FF,$71,$B5,$FD,$C7,$CB,$3E,$0B,$F8,$0B,$63,$FC
	.dc.b	$3A,$D7,$FD,$D6,$D7,$F8,$07,$CB,$7F,$B0,$1B,$E5,$7F,$0A,$8C,$47
	.dc.b	$F1,$57,$FD,$97,$CF,$FE,$CD,$24,$7F,$89,$DF,$F6,$86,$72,$3F,$C5
	.dc.b	$4D,$7F,$A8,$7C,$B7,$FB,$54,$8F,$F8,$BD,$FE,$33,$F3,$2F,$9A,$A4
	.dc.b	$9F,$F1,$BC,$65,$FF,$1C,$FC,$BB,$E4,$DF,$2F,$86,$CF,$F1,$D1,$3F
	.dc.b	$F0,$03,$DF,$D5,$7F,$90,$BD,$87,$A5,$FC,$B6,$FF,$92,$46,$3C,$FD
	.dc.b	$4E,$67,$F9,$C4,$E5,$ED,$F2,$AF,$81,$D4,$7F,$B6,$E3,$17,$FE,$50
	.dc.b	$6C,$70,$9D,$AF,$3D,$F2,$1F,$CF,$DF,$26,$0E,$2C,$63,$FD,$26,$76
	.dc.b	$FF,$4D,$3C,$B1,$AB,$FE,$A1,$F1,$D5,$ED,$85,$B4,$D8,$D4,$BE,$57
	.dc.b	$86,$3F,$3F,$1E,$D3,$FD,$66,$73,$C1,$C3,$D4,$E3,$16,$B7,$5F,$EF
	.dc.b	$D8,$EB,$D3,$98,$F8,$13,$FD,$4E,$31,$9F,$CF,$D7,$C3,$ED,$F9,$1B
	.dc.b	$5D,$2D,$C5,$98,$C3,$63,$A9,$FD,$7F,$E5,$7F,$B3,$A6,$91,$FC,$47
	.dc.b	$C8,$BE,$55,$F2,$BF,$E8,$88,$C3,$63,$33,$C1,$AF,$F6,$D9,$FE,$4B
	.dc.b	$9F,$1F,$23,$EF,$4F,$66,$BB,$6B,$FE,$05,$97,$C8,$7D,$FC,$F0,$D9
	.dc.b	$09,$DB,$F4,$1F,$0A,$F9,$57,$C6,$75,$E5,$E5,$3C,$5B,$5F,$92,$8E
	.dc.b	$57,$8E,$69,$1D,$5E,$3F,$2B,$FE,$0B,$D3,$F5,$56,$F8,$BE,$64,$BE
	.dc.b	$07,$97,$E5,$7B,$91,$F4,$47,$93,$3F,$E3,$57,$FB,$56,$EB,$FE,$29
	.dc.b	$6D,$7C,$2F,$94,$7C,$AF,$1E,$59,$3F,$92,$AF,$DB,$37,$9F,$E0,$FC
	.dc.b	$A3,$E6,$E9,$27,$F3,$31,$FD,$2B,$3F,$E0,$CE,$E3,$FD,$96,$D7,$F8
	.dc.b	$83,$5F,$F1,$73,$5F,$F2,$07,$CB,$7D,$67,$F8,$B3,$5F,$F1,$A3,$5D
	.dc.b	$F1,$7F,$8F,$A3,$85,$7F,$17,$7E,$B4,$FE,$78,$38,$8D,$4F,$DE,$8E
	.dc.b	$3A,$F8,$B8,$F0,$66,$6D,$FE,$E2,$B9,$56,$E3,$13,$A7,$C1,$3A,$7F
	.dc.b	$26,$D7,$C7,$1B,$BF,$45,$DB,$F6,$D7,$78,$C3,$C9,$F2,$39,$21,$E0
	.dc.b	$72,$D3,$33,$D9,$F3,$98,$9F,$53,$B3,$D3,$5E,$07,$E4,$D5,$BF,$8F
	.dc.b	$1E,$33,$91,$4E,$26,$0E,$25,$78,$B8,$3C,$17,$E5,$78,$98,$BF,$2C
	.dc.b	$64,$FB,$31,$CE,$BC,$97,$E5,$6C,$64,$F0,$46,$5A,$64,$EC,$7D,$67
	.dc.b	$4E,$0F,$AE,$6E,$7C,$55,$F5,$A3,$57,$F7,$49,$5C,$AB,$73,$D4,$E5
	.dc.b	$5E,$C3,$94,$6A,$3E,$23,$6B,$E3,$9D,$6F,$DB,$3E,$06,$CE,$DB,$5D
	.dc.b	$67,$EA,$4F,$3B,$F2,$4E,$67,$DA,$73,$3E,$29,$4F,$6C,$70,$79,$C7
	.dc.b	$A0,$E5,$7E,$58,$FC,$69,$C6,$BC,$0F,$C0,$38,$9F,$0C,$D4,$F2,$AE
	.dc.b	$D1,$F0,$A3,$B3,$C0,$C1,$6B,$E2,$DF,$06,$C7,$73,$DB,$7F,$15,$F6
	.dc.b	$B5,$F6,$9B,$8F,$B2,$D9,$FD,$5C,$EC,$F5,$C7,$A0,$F1,$5F,$86,$77
	.dc.b	$7F,$11,$FA,$73,$E7,$6B,$EA,$25,$F8,$1E,$86,$2B,$A3,$F3,$2F,$C5
	.dc.b	$7F,$1E,$3C,$C6,$5F,$03,$EE,$6B,$EE,$3A,$D7,$79,$E7,$1E,$A3,$99
	.dc.b	$FB,$27,$2F,$8C,$CB,$FB,$27,$9D,$FC,$07,$DC,$78,$4A,$78,$63,$9C
	.dc.b	$F8,$6B,$CB,$1F,$30,$CD,$73,$AF,$8A,$FC,$8A,$FC,$74,$73,$AE,$27
	.dc.b	$C3,$7C,$67,$C3,$3C,$35,$D6,$37,$1E,$F3,$F9,$01,$FD,$3A,$BF,$A6
	.dc.b	$11,$FF,$34,$1F,$C5,$CD,$18,$AF,$E9,$EE,$0F,$EB,$31,$FE,$0A,$FA
	.dc.b	$47,$CC,$53,$F3,$AE,$27,$FC,$86,$D7,$F0,$2D,$C3,$FC,$E4,$DC,$3D
	.dc.b	$32,$8C,$3F,$E2,$06,$BF,$20,$EB,$5E,$17,$F4,$91,$A3,$78,$7C,$CB
	.dc.b	$EF,$5B,$AF,$1B,$63,$F0,$75,$C6,$3D,$66,$6F,$CB,$19,$AE,$F3,$97
	.dc.b	$E9,$CF,$F8,$D9,$F5,$46,$5F,$41,$FD,$E6,$79,$3E,$23,$D6,$74,$29
	.dc.b	$93,$84,$79,$1C,$D7,$A2,$3B,$5F,$BE,$FC,$57,$DA,$B5,$7C,$27,$D7
	.dc.b	$7C,$30,$FE,$C9,$7B,$9C,$D6,$CF,$69,$F4,$EB,$83,$FD,$8E,$78,$3A
	.dc.b	$3E,$09,$C4,$F8,$37,$F4,$9C,$61,$E7,$7A,$18,$C3,$C7,$7D,$A3,$FA
	.dc.b	$C5,$F6,$79,$E1,$EE,$9D,$61,$FD,$3E,$B5,$87,$5B,$F1,$9F,$28,$F9
	.dc.b	$C5,$3E,$74,$F9,$5B,$7D,$5D,$C6,$4F,$96,$F1,$69,$98,$F9,$31,$96
	.dc.b	$C6,$63,$E3,$57,$9B,$0F,$7C,$7C,$13,$E6,$61,$EC,$BF,$EB,$95,$B9
	.dc.b	$C9,$B3,$E2,$CE,$DF,$E9,$91,$B7,$26,$EF,$E3,$4E,$E3,$F9,$C4,$EE
	.dc.b	$73,$1B,$AF,$0D,$6F,$0F,$F2,$05,$EE,$73,$5B,$3C,$35,$BC,$3F,$45
	.dc.b	$5F,$B4,$7D,$F3,$F1,$4F,$87,$3B,$7F,$11,$3B,$7D,$E6,$CF,$C0,$9F
	.dc.b	$88,$31,$7F,$7E,$3C,$91,$E2,$3C,$87,$BA,$7B,$3C,$E3,$B3,$F1,$8E
	.dc.b	$75,$E6,$C3,$EC,$47,$80,$FC,$5B,$8B,$F0,$5F,$48,$DD,$F4,$6C,$7C
	.dc.b	$F8,$FC,$D1,$9C,$3F,$1D,$7F,$C2,$4F,$61,$63,$F9,$53,$67,$F9,$CA
	.dc.b	$F1,$5E,$03,$C5,$FE,$7B,$18,$6F,$7F,$FB,$AA,$BB,$1C,$6B,$83,$8F
	.dc.b	$11,$F0,$AC,$FF,$A3,$C6,$1B,$19,$9E,$0D,$7F,$D8,$DB,$67,$5C,$78
	.dc.b	$8F,$17,$18,$7F,$27,$95,$62,$FF,$DB,$F7,$BA,$FF,$70,$D7,$51,$DB
	.dc.b	$BA,$47,$29,$D9,$D6,$75,$19,$9F,$F5,$50,$F1,$39,$5F,$CA,$BE,$33
	.dc.b	$FC,$EE,$73,$A7,$78,$78,$B6,$70,$DA,$E3,$3F,$EF,$8E,$7C,$C6,$21
	.dc.b	$18,$7C,$7D,$3E,$2E,$0E,$CD,$AE,$1E,$2E,$1A,$73,$8F,$25,$3C,$5F
	.dc.b	$C2,$B4,$FC,$3B,$83,$FC,$03,$E0,$3C,$47,$8B,$86,$D7,$E1,$A7,$53
	.dc.b	$93,$8F,$8B,$F4,$27,$4D,$7F,$35,$28,$C4,$7F,$65,$8C,$70,$F8,$BB
	.dc.b	$3F,$05,$9E,$5B,$41,$5F,$EA,$E3,$8A,$F1,$5E,$07,$5E,$6E,$33,$C1
	.dc.b	$B5,$FE,$D7,$39,$73,$60,$93,$F6,$3C,$5F,$D9,$67,$F9,$BA,$78,$9C
	.dc.b	$89,$5F,$7A,$3F,$26,$DF,$17,$B9,$1D,$4F,$93,$2F,$C0,$C7,$E1,$9A
	.dc.b	$FF,$BF,$5B,$7F,$AB,$F1,$5E,$2F,$1E,$49,$7C,$5E,$9C,$FF,$03,$C4
	.dc.b	$7A,$7F,$C1,$9B,$5F,$E4,$06,$BF,$C0,$9A,$FF,$11,$5E,$37,$32,$47
	.dc.b	$E7,$23,$8E,$5E,$5F,$8C,$FE,$27,$99,$A4,$AF,$F2,$7C,$65,$FF,$2D
	.dc.b	$1F,$E9,$AE,$5F,$F1,$44,$66,$3F,$77,$18,$9F,$F3,$47,$8F,$FE,$EE
	.dc.b	$AD,$5F,$86,$F9,$CE,$6B,$7C,$DF,$E6,$FB,$DE,$1F,$EB,$0B,$F0,$1F
	.dc.b	$EB,$7C,$39,$3F,$E8,$58,$D6,$1C,$E3,$AD,$F3,$C3,$AC,$7A,$B0,$EA
	.dc.b	$F3,$BE,$35,$F9,$29,$CC,$F8,$09,$7D,$5F,$01,$C4,$82,$38,$C7,$FB
	.dc.b	$A2,$7C,$53,$E4,$3C,$57,$E4,$79,$9C,$5C,$61,$FE,$9D,$3D,$E5,$C3
	.dc.b	$CF,$3F,$C6,$30,$EF,$7C,$27,$E9,$4F,$73,$AC,$E9,$C5,$74,$9E,$71
	.dc.b	$AB,$E4,$FB,$2B,$83,$E6,$C3,$E1,$5F,$8E,$3E,$0C,$62,$F9,$CE,$67
	.dc.b	$D8,$78,$23,$EE,$57,$49,$D5,$71,$AE,$78,$7C,$87,$C0,$D9,$DB,$6F
	.dc.b	$B9,$D3,$0F,$4D,$7D,$9C,$3B,$4E,$A3,$F8,$B4,$61,$EE,$7C,$B9,$E3
	.dc.b	$87,$8E,$3F,$03,$7D,$DD,$9F,$1C,$F6,$1F,$1E,$78,$C7,$E2,$DB,$1D
	.dc.b	$4D,$9E,$57,$2F,$EA,$8F,$2B,$8C,$3E,$B4,$78,$1F,$99,$7E,$98,$F2
	.dc.b	$9D,$4F,$8E,$7E,$38,$EF,$7E,$B3,$F2,$8E,$4D,$9E,$B2,$CF,$6A,$F1
	.dc.b	$61,$F8,$C9,$CB,$F3,$DE,$B8,$7B,$DA,$78,$EB,$EB,$BD,$DA,$F7,$69
	.dc.b	$DD,$E2,$7C,$33,$F1,$27,$80,$F9,$C6,$9F,$8A,$FC,$23,$A4,$FD,$F7
	.dc.b	$E7,$1B,$7D,$51,$E0,$7D,$51,$D4,$D9,$6F,$E4,$57,$E3,$30,$FA,$11
	.dc.b	$F2,$A7,$CD,$18,$9F,$C3,$47,$8E,$FD,$8F,$38,$F8,$CD,$3E,$35,$F9
	.dc.b	$EF,$BC,$7E,$1A,$FA,$3F,$26,$7F,$DB,$12,$AF,$F2,$6B,$67,$FC,$8E
	.dc.b	$1C,$3F,$E4,$A6,$BF,$E5,$96,$BD,$23,$99,$1E,$71,$CA,$FF,$C9,$0F
	.dc.b	$FA,$DC,$F8,$4E,$39,$7F,$BD,$C7,$F9,$1A,$7F,$23,$18,$3F,$5A,$57
	.dc.b	$27,$F8,$CA,$38,$3F,$46,$34,$D3,$4F,$26,$9C,$8F,$C4,$47,$29,$FB
	.dc.b	$D1,$FE,$DF,$8F,$3D,$FC,$AA,$EB,$5A,$AF,$F4,$45,$69,$E1,$7F,$D7
	.dc.b	$2B,$77,$F5,$DB,$3F,$B5,$8E,$8F,$90,$F4,$1E,$48,$F4,$34,$F4,$5F
	.dc.b	$C9,$6C,$F2,$78,$C6,$5B,$19,$8F,$45,$FD,$3B,$F3,$57,$86,$73,$7E
	.dc.b	$68,$FC,$A5,$F1,$23,$E9,$3D,$E7,$F7,$68,$E1,$7F,$0E,$3E,$2D,$F6
	.dc.b	$7D,$47,$90,$FF,$21,$47,$D1,$AF,$8D,$1E,$36,$BC,$2F,$CA,$E4,$F1
	.dc.b	$1D,$8F,$11,$93,$D2,$79,$8F,$EA,$97,$E6,$BF,$24,$FC,$73,$AC,$F7
	.dc.b	$AF,$0C,$65,$A7,$C2,$8E,$A7,$84,$EA,$78,$6F,$E1,$1F,$3D,$F2,$9D
	.dc.b	$4F,$4B,$97,$89,$93,$AB,$99,$FE,$C3,$7E,$99,$E7,$1E,$57,$9C,$65
	.dc.b	$E7,$19,$79,$C6,$63,$9C,$79,$EF,$E1,$DF,$B6,$FD,$37,$B8,$FF,$14
	.dc.b	$B6,$BF,$BD,$B5,$FF,$10,$B6,$7F,$81,$4F,$38,$FC,$DD,$F8,$E7,$AD
	.dc.b	$78,$4F,$5D,$78,$6F,$C7,$1F,$86,$BE,$33,$98,$D9,$93,$F4,$65,$3F
	.dc.b	$1C,$7F,$1D,$8E,$2E,$6F,$EF,$9E,$3A,$FC,$CC,$FA,$4B,$3D,$8F,$BD
	.dc.b	$3F,$AD,$7F,$92,$47,$F2,$C2,$3F,$CB,$3E,$29,$E4,$FF,$BF,$1B,$19
	.dc.b	$8C,$1F,$F0,$03,$C7,$FF,$7E,$47,$FB,$E9,$AF,$FA,$9C,$FF,$54,$2E
	.dc.b	$FF,$F8,$01,$19,$3F,$E0,$07,$8F,$FC,$F3,$1F,$F7,$21,$FE,$F6,$F1
	.dc.b	$7F,$DC,$F9,$7F,$B9,$1F,$F7,$4F,$5E,$35,$FF,$6E,$D7,$FB,$C9,$B5
	.dc.b	$FE,$E1,$C3,$FB,$77,$8B,$FE,$EB,$7F,$DD,$FE,$27,$FB,$AA,$BF,$DC
	.dc.b	$5E,$26,$F0,$47,$96,$30,$F9,$F9,$FF,$DC,$D3,$AB,$F7,$46,$27,$D9
	.dc.b	$D1,$FE,$E7,$9F,$F7,$5C,$7F,$BC,$B1,$F8,$4F,$FB,$A2,$32,$7F,$80
	.dc.b	$C7,$89,$FE,$ED,$70,$7F,$BA,$4F,$EC,$3E,$27,$FB,$A5,$FF,$75,$9F
	.dc.b	$EE,$5F,$13,$FD,$CE,$FF,$B8,$1A,$FF,$B7,$DA,$FF,$B6,$FC,$0F,$F7
	.dc.b	$5B,$FE,$E8,$CF,$FD,$D6,$47,$07,$FB,$C7,$9B,$FD,$B2,$47,$FD,$A1
	.dc.b	$E1,$FF,$BC,$7C,$3F,$F7,$0C,$75,$E1,$FF,$6F,$91,$FF,$6B,$57,$FB
	.dc.b	$79,$FF,$11,$61,$FE,$D8,$6F,$17,$FB,$A3,$C3,$FF,$70,$ED,$FE,$F7
	.dc.b	$F0,$BF,$DB,$F5,$C1,$E5,$A7,$FB,$97,$C3,$FF,$6F,$4F,$47,$71,$F5
	.dc.b	$E1,$C3,$63,$0D,$BF,$CC,$36,$7F,$E0,$02,$FA,$2F,$F7,$DB,$CE,$3F
	.dc.b	$34,$D7,$8C,$FF,$85,$9E,$ED,$F1,$1C,$9F,$EE,$89,$FF,$77,$3F,$EE
	.dc.b	$87,$FD,$D4,$43,$27,$FB,$D7,$97,$86,$3E,$BA,$E7,$3B,$8C,$4E,$AF
	.dc.b	$F6,$55,$AA,$E6,$7F,$BF,$67,$48,$46,$EB,$B5,$F9,$67,$8C,$61,$F0
	.dc.b	$5F,$6B,$E7,$1C,$AA,$DA,$6C,$61,$B3,$C8,$F1,$4E,$DC,$1C,$CA,$E5
	.dc.b	$F1,$D7,$BA,$BA,$46,$9A,$E8,$EA,$E2,$FD,$51,$93,$B1,$F1,$4E,$47
	.dc.b	$82,$FC,$EE,$EF,$C5,$3C,$89,$7E,$39,$FB,$07,$23,$85,$72,$9D,$C7
	.dc.b	$C0,$8C,$4F,$18,$F1,$1D,$EF,$D1,$19,$AE,$33,$E8,$7F,$02,$F7,$73
	.dc.b	$7E,$69,$E0,$49,$F8,$67,$98,$EC,$7B,$CE,$4E,$DF,$C2,$DF,$99,$E2
	.dc.b	$79,$AB,$BD,$F9,$A3,$31,$87,$B3,$5F,$86,$76,$37,$5F,$8B,$70,$7B
	.dc.b	$5C,$3E,$63,$89,$63,$DF,$5E,$58,$E0,$7C,$D3,$9D,$FE,$76,$B9,$D7
	.dc.b	$8C,$E3,$3A,$BF,$D0,$46,$67,$E1,$57,$CB,$AF,$39,$0E,$E7,$C6,$39
	.dc.b	$1F,$02,$38,$D7,$C5,$8D,$CF,$8D,$E4,$F5,$73,$1B,$8D,$46,$0E,$87
	.dc.b	$08,$D1,$F0,$63,$91,$C0,$8F,$B2,$7C,$73,$F7,$CF,$94,$FB,$AF,$94
	.dc.b	$EE,$7B,$1B,$BF,$88,$FA,$27,$6F,$90,$F3,$1F,$5A,$36,$FC,$08,$DB
	.dc.b	$F8,$9A,$F1,$BB,$6C,$FD,$48,$F2,$9D,$CF,$9A,$71,$3C,$A7,$12,$BC
	.dc.b	$4F,$0D,$FC,$3A,$F0,$C7,$EA,$27,$A9,$E5,$3A,$B9,$38,$DF,$C3,$8F
	.dc.b	$6C,$75,$BF,$39,$F2,$63,$8B,$F8,$E3,$2F,$A4,$EF,$7E,$37,$67,$E5
	.dc.b	$CC,$D7,$78,$FB,$31,$F4,$67,$9B,$E7,$3F,$9E,$9F,$E2,$D8,$FE,$80
	.dc.b	$D9,$FE,$A4,$47,$FD,$D7,$1A,$7F,$88,$1B,$33,$1F,$D8,$3E,$2E,$47
	.dc.b	$ED,$31,$46,$23,$E4,$C6,$1F,$F1,$04,$7E,$CE,$31,$1F,$E3,$48,$C1
	.dc.b	$C7,$0F,$C2,$CE,$9F,$EE,$B1,$A9,$E6,$78,$5E,$18,$7C,$2A,$E0,$FF
	.dc.b	$7A,$8E,$13,$DC,$E0,$72,$3C,$EF,$27,$F0,$F1,$C5,$D5,$79,$27,$51
	.dc.b	$F8,$09,$D3,$D4,$E6,$70,$7C,$91,$C3,$0F,$B3,$3B,$AF,$6B,$E4,$31
	.dc.b	$5C,$1E,$47,$90,$FF,$09,$D7,$3C,$3F,$B7,$E1,$F0,$27,$27,$3B,$EA
	.dc.b	$F9,$0E,$A7,$33,$FC,$47,$5C,$DF,$ED,$F3,$CD,$C9,$CE,$FD,$F5,$F0
	.dc.b	$E3,$CA,$7C,$33,$C2,$76,$30,$72,$31,$87,$C2,$8F,$29,$27,$E0,$4E
	.dc.b	$B0,$F6,$E1,$E5,$8F,$EE,$D8,$79,$4F,$36,$1C,$EF,$CA,$49,$F7,$4E
	.dc.b	$EB,$F0,$78,$77,$9C,$C7,$58,$FE,$D7,$3B,$9F,$80,$7B,$23,$E0,$3E
	.dc.b	$53,$E3,$CF,$1A,$FE,$CD,$5C,$67,$E8,$57,$77,$E1,$C6,$63,$99,$9B
	.dc.b	$F5,$4F,$7A,$F5,$3C,$2F,$B3,$FD,$3E,$F9,$4F,$C8,$BE,$58,$7C,$8A
	.dc.b	$FA,$B8,$74,$7E,$94,$E9,$F4,$9A,$9E,$0F,$B4,$F9,$F8,$78,$1F,$EB
	.dc.b	$B8,$78,$1F,$92,$71,$39,$9C,$6B,$E5,$CF,$1B,$FE,$B3,$5C,$6B,$D8
	.dc.b	$71,$3B,$96,$B8,$B7,$0F,$9B,$7C,$F0,$FD,$9D,$FE,$09,$F4,$DF,$B6
	.dc.b	$F9,$9E,$3B,$E2,$F3,$3E,$55,$F3,$7F,$37,$7C,$DF,$5D,$73,$3C,$6F
	.dc.b	$33,$89,$CC,$F8,$B3,$CD,$FD,$14,$F3,$7C,$73,$CE,$BC,$71,$CC,$FF
	.dc.b	$71,$9F,$EE,$BF,$1D,$D7,$FE,$E7,$8F,$F7,$5F,$67,$FB,$A9,$FF,$77
	.dc.b	$76,$FF,$BC,$1F,$F4,$CF,$C3,$7F,$A4,$DF,$F4,$FB,$72,$34,$D4,$3B
	.dc.b	$7F,$DC,$8F,$FB,$9B,$2F,$F6,$C6,$5F,$ED,$D8,$FF,$7E,$4D,$7F,$D1
	.dc.b	$0F,$FB,$8F,$3F,$F5,$26,$F9,$9F,$F4,$05,$FF,$BB,$8F,$F4,$74,$FF
	.dc.b	$AA,$76,$FF,$73,$36,$3F,$DC,$DA,$73,$7F,$D2,$95,$FE,$95,$EC,$FF
	.dc.b	$72,$57,$FA,$C5,$B3,$F1,$7B,$3F,$DC,$95,$FE,$E6,$3F,$DC,$A7,$FB
	.dc.b	$8F,$C8,$F9,$4B,$A3,$FD,$C2,$E4,$FF,$75,$47,$FB,$85,$F0,$9B,$EC
	.dc.b	$63,$98,$F2,$D2,$73,$1F,$EE,$73,$FD,$BE,$43,$FD,$BE,$FF,$B9,$F9
	.dc.b	$38,$F5,$FF,$B9,$67,$FD,$C6,$FF,$BB,$5F,$F7,$19,$F2,$32,$EF,$D5
	.dc.b	$FE,$E0,$7E,$2F,$57,$FB,$9A,$7F,$DF,$9D,$5F,$EE,$09,$FF,$74,$47
	.dc.b	$13,$FE,$00,$57,$FB,$9B,$3F,$F7,$44,$64,$F8,$9D,$5C,$5F,$F7,$0F
	.dc.b	$57,$FB,$6F,$9F,$FD,$C5,$1F,$29,$AF,$E5,$5A,$FF,$B9,$30,$FF,$73
	.dc.b	$91,$ED,$1C,$2F,$FC,$59,$3F,$E2,$59,$FF,$10,$47,$FB,$DF,$C3,$79
	.dc.b	$1F,$E2,$1E,$6F,$F1,$84,$62,$BF,$C9,$DB,$FF,$97,$A7,$F0,$F7,$C6
	.dc.b	$F8,$57,$F9,$4E,$F8,$3F,$E4,$73,$FC,$EC,$57,$33,$E4,$8F,$1E,$1E
	.dc.b	$55,$3A,$8C,$61,$E1,$9F,$95,$39,$78,$90,$EF,$5E,$57,$D3,$87,$B3
	.dc.b	$0F,$3C,$78,$DB,$3E,$38,$F0,$C6,$30,$F2,$5F,$79,$EB,$1E,$4B,$F8
	.dc.b	$6F,$EA,$A7,$58,$79,$EB,$B1,$E2,$3E,$35,$F5,$7D,$13,$C2,$F1,$1E
	.dc.b	$23,$81,$C8,$F9,$95,$D7,$0F,$74,$62,$7D,$CF,$08,$F6,$57,$58,$C6
	.dc.b	$1E,$7C,$3B,$E1,$FE,$A3,$AD,$57,$33,$E7,$CE,$B0,$F7,$C0,$F3,$1B
	.dc.b	$3C,$D8,$7C,$BB,$F3,$61,$DE,$7C,$27,$C0,$BE,$8F,$63,$27,$CE,$3B
	.dc.b	$61,$E4,$BE,$98,$78,$A7,$C3,$18,$7E,$69,$CB,$0F,$65,$F6,$8F,$E1
	.dc.b	$B8,$78,$DF,$45,$7C,$69,$D3,$F3,$0E,$91,$EC,$9F,$C0,$3D,$8E,$2F
	.dc.b	$E1,$A7,$8B,$D6,$78,$9D,$67,$C9,$7E,$9C,$3D,$95,$D2,$7C,$A7,$C0
	.dc.b	$BE,$6D,$38,$9F,$1C,$F0,$9F,$2C,$C9,$C8,$F8,$B1,$D6,$FC,$51,$C0
	.dc.b	$F9,$C6,$8F,$D5,$61,$F6,$1F,$01,$B7,$E5,$BD,$A3,$E3,$4F,$53,$E1
	.dc.b	$46,$5F,$84,$65,$F3,$1C,$F0,$F8,$57,$EE,$8F,$11,$E4,$BF,$13,$86
	.dc.b	$9E,$27,$91,$F0,$EF,$E7,$DF,$8A,$32,$7C,$3C,$3E,$7B,$EE,$C3,$C2
	.dc.b	$F9,$CF,$7D,$72,$7C,$ED,$7E,$3C,$F2,$7C,$D3,$C9,$EF,$3F,$11,$EC
	.dc.b	$7E,$1D,$E5,$7E,$96,$CF,$16,$CF,$AE,$FE,$43,$E0,$C3,$E3,$61,$E1
	.dc.b	$C3,$E2,$B7,$0F,$E0,$35,$E2,$BF,$A2,$FF,$B7,$CF,$F6,$C4,$7F,$76
	.dc.b	$7F,$9A,$98,$3F,$DA,$EE,$68,$D2,$46,$8F,$EB,$91,$FE,$61,$3E,$AF
	.dc.b	$37,$ED,$5F,$F7,$84,$7F,$18,$8C,$1F,$C4,$0A,$FE,$DD,$AF,$E3,$DB
	.dc.b	$8F,$C8,$6D,$FC,$4A,$F0,$5F,$FB,$E6,$73,$7F,$40,$FF,$80,$31,$97
	.dc.b	$FD,$F8,$7E,$7A,$36,$FD,$A3,$D6,$E2,$7D,$71,$C0,$FC,$A4,$66,$7A
	.dc.b	$47,$EB,$E7,$DF,$7F,$C8,$6B,$53,$FD,$A9,$AF,$C3,$3A,$4F,$CE,$7C
	.dc.b	$26,$2B,$C5,$3F,$1E,$30,$68,$F0,$39,$69,$F2,$CD,$9F,$08,$F1,$18
	.dc.b	$A5,$FF,$90,$AF,$E5,$CE,$8E,$67,$D3,$38,$46,$EB,$D7,$5E,$F9,$C9
	.dc.b	$FE,$E1,$AF,$92,$F2,$8F,$B4,$E8,$FF,$6C,$C6,$A3,$C7,$1F,$D3,$63
	.dc.b	$E5,$D7,$F8,$0A,$BB,$B6,$79,$4F,$A1,$F9,$87,$78,$FC,$34,$F0,$8F
	.dc.b	$C0,$1C,$1C,$3F,$70,$C1,$0E,$47,$E2,$EB,$CA,$FA,$2F,$F0,$31,$8B
	.dc.b	$F3,$CF,$63,$89,$F8,$D3,$E9,$1E,$68,$E1,$5F,$65,$B1,$F6,$4E,$33
	.dc.b	$F1,$6F,$E8,$46,$67,$6F,$D5,$37,$7E,$49,$DC,$66,$76,$7D,$27,$67
	.dc.b	$DF,$73,$1F,$22,$3E,$14,$61,$F8,$51,$D2,$31,$1E,$D6,$D7,$C2,$8F
	.dc.b	$98,$76,$3E,$04,$F1,$3B,$39,$9F,$C5,$36,$3E,$F3,$4F,$99,$1D,$AB
	.dc.b	$DD,$1F,$87,$BF,$95,$5D,$63,$CE,$7E,$4E,$76,$AF,$CC,$7E,$C5,$FA
	.dc.b	$B7,$E2,$3B,$3F,$E6,$CF,$86,$FF,$3B,$11,$FF,$3B,$3F,$CA,$FA,$FF
	.dc.b	$D8,$D1,$FE,$25,$6A,$5D,$1F,$EC,$D8,$D7,$57,$FB,$49,$FF,$54,$F5
	.dc.b	$7F,$B1,$B0,$FC,$7C,$7F,$B2,$73,$FF,$63,$9F,$EB,$F7,$FD,$94,$7F
	.dc.b	$7F,$1E,$0F,$FA,$F5,$FF,$65,$C7,$FA,$45,$FF,$63,$9B,$8D,$3F,$E3
	.dc.b	$BD,$3F,$D9,$9A,$FF,$B2,$2B,$FD,$6A,$D9,$FF,$63,$F5,$FF,$B1,$1B
	.dc.b	$3D,$67,$9C,$7F,$91,$7A,$FF,$D8,$EF,$FB,$24,$FF,$63,$9F,$EB,$CF
	.dc.b	$05,$D9,$FE,$C8,$EC,$FF,$63,$CE,$63,$13,$A8,$FF,$7B,$4F,$FB,$0F
	.dc.b	$4F,$F6,$34,$7F,$94,$E3,$0E,$BB,$3F,$D8,$EF,$FB,$3A,$33,$F5,$9F
	.dc.b	$81,$8E,$1E,$0F,$FB,$8A,$31,$1F,$ED,$98,$C7,$2F,$FB,$4A,$FF,$DA
	.dc.b	$35,$AB,$EF,$E1,$5B,$FF,$75,$B7,$0F,$F3,$33,$BB,$FF,$6A,$47,$80
	.dc.b	$F8,$13,$FD,$CE,$8D,$31,$DF,$FE,$D9,$CB,$F8,$8C,$A3,$FD,$B5,$18
	.dc.b	$77,$5F,$2D,$AF,$C8,$EC,$FF,$6A,$70,$7F,$BB,$1F,$F6,$94,$7F,$B5
	.dc.b	$0F,$F6,$B5,$A3,$71,$FE,$E0,$32,$E1,$E0,$7F,$93,$E1,$DD,$97,$94
	.dc.b	$70,$3A,$C6,$67,$4D,$8D,$1C,$DC,$3D,$63,$CF,$3D,$23,$07,$0C,$3E
	.dc.b	$2D,$70,$7F,$DE,$31,$CC,$D5,$FC,$16,$9E,$D3,$99,$23,$89,$D1,$D9
	.dc.b	$F0,$1C,$C7,$FA,$8E,$15,$F0,$9D,$BE,$79,$DC,$77,$6B,$CE,$7C,$8F
	.dc.b	$9C,$F1,$38,$8D,$CF,$59,$C9,$FA,$1B,$F9,$6F,$81,$AE,$8E,$C6,$CF
	.dc.b	$A6,$6C,$F3,$DF,$C9,$BF,$01,$ED,$AF,$05,$F8,$AF,$AB,$6F,$89,$CC
	.dc.b	$F0,$4F,$0A,$E7,$1E,$13,$37,$DD,$C5,$7C,$F7,$8D,$F6,$AE,$31,$EA
	.dc.b	$25,$F5,$78,$C6,$A3,$E8,$3E,$9B,$ED,$3C,$1B,$3A,$8F,$A9,$3C,$6B
	.dc.b	$C1,$1D,$08,$71,$3D,$D7,$F5,$2F,$9C,$FE,$58,$EA,$E2,$7E,$79,$EC
	.dc.b	$3E,$39,$4F,$86,$74,$AF,$54,$71,$8F,$C2,$57,$CF,$76,$71,$7C,$4F
	.dc.b	$13,$CF,$7F,$15,$F8,$F7,$EC,$69,$E2,$9E,$87,$27,$A5,$F8,$9E,$87
	.dc.b	$C1,$32,$76,$BF,$34,$FD,$93,$9C,$F1,$9D,$9C,$63,$2F,$13,$AD,$FD
	.dc.b	$37,$F5,$75,$D2,$BE,$D4,$6A,$FE,$73,$C2,$74,$FD,$A3,$E0,$C7,$BE
	.dc.b	$BA,$11,$E3,$1E,$48,$F8,$C6,$5F,$89,$5F,$83,$69,$D0,$A7,$CA,$3E
	.dc.b	$9D,$FB,$1F,$74,$71,$72,$D8,$F9,$C6,$E7,$CB,$7E,$AA,$F8,$27,$BC
	.dc.b	$FC,$4C,$EE,$FD,$2F,$19,$E1,$5E,$A9,$FC,$8D,$71,$3F,$13,$7E,$13
	.dc.b	$EA,$5F,$38,$EB,$19,$7E,$A1,$FC,$90,$FE,$D5,$3F,$D5,$C8,$FF,$79
	.dc.b	$6B,$FD,$3C,$C1,$04,$FF,$9A,$BB,$1F,$43,$FC,$A6,$30,$FE,$DA,$33
	.dc.b	$23,$8D,$FF,$2F,$8C,$C7,$E3,$1B,$5C,$A3,$F2,$F1,$89,$E1,$5D,$4F
	.dc.b	$03,$89,$DC,$61,$F0,$4F,$17,$F0,$07,$88,$CC,$6D,$46,$DF,$C2,$57
	.dc.b	$8D,$E0,$6A,$7C,$4F,$B6,$34,$7E,$26,$79,$1C,$F0,$FC,$A4,$F2,$8C
	.dc.b	$D7,$28,$C3,$6B,$9D,$FD,$B9,$F0,$1D,$DF,$80,$FC,$99,$F5,$57,$7C
	.dc.b	$3F,$BF,$3D,$A7,$E2,$9C,$4F,$82,$60,$F0,$96,$FC,$2F,$C8,$6B,$E2
	.dc.b	$3C,$07,$33,$0F,$C1,$AF,$19,$EB,$BF,$3C,$66,$78,$57,$9D,$FE,$EA
	.dc.b	$F9,$DE,$A7,$90,$B1,$E2,$C3,$E5,$BE,$49,$E5,$3F,$D9,$DE,$51,$DE
	.dc.b	$39,$34,$E5,$87,$C6,$C3,$CF,$3F,$07,$0F,$39,$F4,$6F,$9B,$67,$73
	.dc.b	$E4,$8D,$BF,$1D,$AF,$C4,$78,$3F,$02,$3E,$23,$F4,$8E,$45,$BF,$34
	.dc.b	$F2,$C3,$DE,$F4,$BF,$7C,$FF,$84,$9F,$2D,$78,$4E,$53,$C7,$0F,$74
	.dc.b	$78,$4F,$71,$A3,$81,$E7,$30,$7E,$8A,$3C,$18,$7A,$23,$46,$4D,$D7
	.dc.b	$92,$7F,$BF,$BE,$48,$D1,$C1,$D3,$FD,$D1,$D6,$1E,$E8,$E7,$5D,$E7
	.dc.b	$71,$FE,$09,$77,$7E,$C8,$FB,$B7,$F3,$9A,$7B,$27,$6F,$E3,$AB,$EC
	.dc.b	$E1,$F3,$A3,$DF,$87,$85,$F3,$4F,$84,$8F,$C8,$3E,$0C,$6E,$BA,$C7
	.dc.b	$E4,$67,$AD,$F8,$EF,$F1,$F3,$E1,$3E,$91,$F8,$03,$9E,$1E,$B8,$DC
	.dc.b	$73,$36,$F9,$DD,$9C,$CF,$79,$F3,$AF,$C7,$1F,$85,$BF,$1C,$FC,$36
	.dc.b	$BB,$6E,$1E,$B6,$DF,$F2,$5E,$AF,$F7,$19,$3A,$BF,$DC,$57,$FE,$E5
	.dc.b	$EA,$FF,$71,$3F,$E6,$B3,$F9,$5F,$D4,$7F,$B8,$9F,$E6,$6D,$73,$A6
	.dc.b	$DC,$B5,$FE,$67,$F2,$1B,$D4,$D2,$4E,$E7,$FD,$DC,$7F,$BC,$9C,$76
	.dc.b	$77,$8C,$9D,$FA,$F0,$49,$E4,$EE,$1D,$3F,$EE,$27,$D8,$D7,$FD,$0F
	.dc.b	$D1,$FE,$DD,$7F,$DA,$CD,$7F,$DC,$47,$FB,$81,$C3,$FE,$DF,$E5,$FF
	.dc.b	$6E,$BF,$ED,$5E,$2F,$F7,$1B,$FE,$92,$8C,$BB,$8F,$F6,$9F,$47,$F9
	.dc.b	$EF,$9B,$FD,$AE,$D9,$C9,$B1,$CF,$FE,$D9,$AC,$C7,$FB,$94,$BC,$BF
	.dc.b	$ED,$86,$E1,$FE,$DE,$7C,$53,$AC,$3F,$DA,$FA,$79,$E7,$AE,$3E,$17
	.dc.b	$FD,$FB,$3F,$EE,$17,$FD,$CA,$7F,$B6,$CF,$F6,$AF,$3F,$FB,$8E,$B8
	.dc.b	$5F,$F9,$3D,$FF,$22,$CF,$F8,$78,$FF,$04,$62,$F0,$7F,$D9,$3D,$1F
	.dc.b	$EC,$A6,$CF,$F8,$80,$DB,$87,$91,$FE,$BC,$71,$7D,$E3,$67,$13,$FC
	.dc.b	$6E,$70,$AF,$03,$97,$FC,$9B,$5F,$E0,$77,$31,$B8,$F2,$3B,$6B,$93
	.dc.b	$95,$78,$AF,$DF,$67,$FD,$6D,$19,$9F,$F3,$11,$E9,$3A,$1C,$8A,$79
	.dc.b	$DE,$0F,$AA,$7A,$3E,$E3,$C4,$6C,$E2,$FC,$37,$B1,$A7,$17,$EE,$6D
	.dc.b	$FB,$1B,$3E,$33,$85,$FB,$0F,$5B,$D9,$C3,$E3,$8E,$2F,$8C,$E2,$7C
	.dc.b	$0A,$F0,$4F,$08,$F6,$36,$33,$1B,$AF,$24,$E5,$B1,$93,$CD,$5E,$03
	.dc.b	$DB,$7F,$22,$FF,$D1,$2D,$AF,$BC,$DA,$FF,$0F,$4F,$81,$F1,$9D,$4F
	.dc.b	$F2,$2B,$93,$F0,$EF,$82,$FC,$8D,$38,$5F,$89,$F8,$C7,$C1,$AF,$03
	.dc.b	$D1,$B1,$F7,$23,$EB,$BE,$B9,$F8,$93,$A7,$E8,$CF,$CC,$3C,$E7,$07
	.dc.b	$C6,$FD,$19,$E4,$F8,$4F,$19,$C2,$BD,$97,$F1,$EF,$B4,$FD,$BB,$E7
	.dc.b	$7E,$B8,$F0,$5F,$CC,$79,$B5,$CD,$FC,$E7,$2D,$72,$7D,$0B,$EE,$FA
	.dc.b	$4F,$93,$7F,$01,$E2,$6C,$A7,$C9,$2D,$FC,$9B,$F5,$C7,$92,$BA,$C6
	.dc.b	$23,$D3,$18,$23,$CA,$BE,$6D,$7C,$88,$F8,$D3,$A7,$D2,$78,$4B,$1E
	.dc.b	$18,$D3,$C6,$BE,$23,$6F,$E1,$36,$BD,$87,$2B,$F0,$9F,$40,$F8,$4E
	.dc.b	$4F,$A8,$7C,$63,$EE,$19,$3E,$51,$F7,$67,$F4,$ED,$3F,$2C,$7F,$24
	.dc.b	$3F,$BE,$52,$3F,$B8,$36,$3F,$BC,$35,$FE,$F8,$D7,$8C,$70,$23,$C1
	.dc.b	$FF,$6F,$BF,$CC,$79,$FF,$E0,$04,$62,$FF,$89,$C6,$27,$F6,$6D,$7F
	.dc.b	$62,$D8,$FF,$76,$47,$F3,$98,$C5,$FF,$AA,$EF,$F6,$B3,$C0,$E6,$7E
	.dc.b	$5A,$76,$FF,$99,$23,$75,$D2,$3C,$C7,$03,$F3,$F3,$C1,$FF,$2F,$CF
	.dc.b	$0A,$F0,$5F,$FC,$00,$7F,$DF,$27,$F1,$BA,$E1,$3F,$E5,$13,$85,$78
	.dc.b	$6B,$83,$FE,$FC,$BF,$86,$F8,$5F,$C5,$57,$22,$46,$6F,$F0,$CF,$33
	.dc.b	$E2,$19,$3F,$0C,$7A,$2B,$BA,$B5,$7F,$27,$3A,$C7,$C9,$3E,$09,$CC
	.dc.b	$FC,$51,$9B,$FD,$84,$71,$8D,$4F,$B9,$F0,$1E,$43,$85,$FA,$1C,$9F
	.dc.b	$40,$C9,$E8,$8F,$91,$7E,$87,$CF,$1F,$6C,$EF,$19,$AF,$14,$72,$3C
	.dc.b	$F7,$F7,$4F,$9A,$E6,$BC,$C7,$67,$CD,$1E,$C7,$B1,$BA,$EC,$E2,$34
	.dc.b	$FC,$63,$47,$2B,$FC,$1B,$FD,$86,$FE,$6C,$F0,$3E,$EC,$7A,$8E,$07
	.dc.b	$E0,$E7,$F1,$4E,$4F,$8D,$1C,$63,$E5,$5F,$A6,$73,$1E,$07,$F9,$F5
	.dc.b	$78,$23,$13,$FB,$BA,$F8,$91,$DA,$BD,$27,$23,$C4,$7A,$4F,$14,$7C
	.dc.b	$A3,$C4,$FD,$7A,$EC,$F0,$3C,$27,$09,$ED,$3C,$23,$EA,$C7,$EA,$9F
	.dc.b	$82,$7E,$56,$7E,$DC,$EE,$BE,$BC,$76,$9D,$1F,$4E,$BD,$B7,$FC,$35
	.dc.b	$44,$E3,$7A,$7F,$E8,$D6,$C7,$FA,$55,$AF,$FA,$65,$AF,$FA,$87,$A3
	.dc.b	$FD,$4F,$CF,$FE,$A4,$8E,$A2,$1D,$1F,$EA,$A5,$3B,$9F,$F5,$19,$D7
	.dc.b	$A3,$FD,$44,$4C,$7C,$53,$FE,$A5,$23,$83,$FD,$49,$DD,$FE,$A1,$23
	.dc.b	$83,$FD,$35,$5F,$03,$9A,$BF,$EA,$03,$FD,$31,$D9,$FC,$F3,$9F,$FD
	.dc.b	$33,$1F,$EA,$08,$C1,$FE,$98,$E0,$73,$7F,$A6,$27,$FD,$12,$7F,$42
	.dc.b	$71,$1A,$E2,$FF,$46,$4C,$3F,$D1,$67,$FA,$67,$9F,$FD,$30,$BB,$1D
	.dc.b	$1F,$E8,$F6,$C7,$FA,$1B,$A3,$E6,$C7,$02,$4E,$C9,$3F,$EA,$58,$EB
	.dc.b	$5C,$A3,$FD,$48,$7F,$8E,$AB,$FC,$FD,$8F,$FA,$8F,$1F,$F5,$34,$FF
	.dc.b	$A7,$8F,$F4,$E9,$4F,$EB,$FC,$FF,$DE,$36,$FE,$EB,$1F,$8C,$F0,$FF
	.dc.b	$DC,$9A,$7F,$80,$63,$FB,$DB,$53,$FE,$3D,$8D,$CF,$29,$FE,$EF,$7F
	.dc.b	$E5,$19,$CC,$7F,$7E,$33,$7F,$8E,$BE,$51,$E3,$35,$5C,$5C,$1C,$0F
	.dc.b	$D0,$C7,$07,$CB,$7E,$19,$E1,$18,$39,$D7,$79,$FC,$ED,$F9,$27,$A1
	.dc.b	$FE,$21,$BE,$8D,$9E,$95,$C6,$3E,$31,$E4,$BF,$F3,$54,$66,$17,$D6
	.dc.b	$73,$1E,$33,$27,$19,$F6,$CE,$DC,$47,$BA,$BC,$EE,$0F,$21,$E2,$3A
	.dc.b	$57,$67,$EA,$C7,$07,$33,$D9,$E5,$7F,$6C,$F7,$96,$7D,$F1,$E1,$3B
	.dc.b	$1E,$03,$E3,$C7,$08,$DC,$66,$FD,$4F,$80,$E4,$7C,$03,$06,$CF,$1C
	.dc.b	$7E,$26,$38,$C7,$7B,$FB,$6D,$BF,$53,$CA,$FD,$0F,$99,$AF,$9A,$FE
	.dc.b	$13,$C4,$FA,$57,$DA,$77,$7D,$9B,$7D,$AB,$81,$E7,$3C,$17,$E7,$7D
	.dc.b	$27,$E7,$6F,$A4,$7C,$4A,$F8,$D3,$F9,$23,$C2,$70,$3F,$CB,$B3,$C1
	.dc.b	$D1,$F6,$8F,$64,$F3,$3C,$0F,$88,$F0,$1C,$EF,$D5,$1B,$8C,$9B,$3B
	.dc.b	$9B,$BF,$CC,$57,$86,$3E,$41,$E0,$3C,$77,$F3,$23,$57,$F8,$F8,$F6
	.dc.b	$CF,$03,$A9,$DE,$3C,$6F,$2B,$F4,$BF,$32,$7A,$3E,$33,$F0,$F5,$DA
	.dc.b	$FE,$94,$FD,$49,$F9,$27,$58,$F4,$BC,$4F,$90,$7E,$3C,$ED,$5E,$09
	.dc.b	$E5,$1F,$6C,$C9,$F1,$8A,$7A,$8B,$7E,$A7,$DC,$D3,$E1,$4F,$53,$E8
	.dc.b	$47,$19,$EE,$7B,$9F,$41,$EE,$3D,$0F,$CA,$9E,$71,$E3,$8E,$33,$D6
	.dc.b	$3E,$51,$97,$BC,$F8,$EB,$F1,$97,$F2,$6B,$E3,$9E,$A8,$F8,$EF,$88
	.dc.b	$F8,$F1,$B7,$E2,$5F,$81,$EC,$78,$0F,$65,$F8,$2B,$9C,$7F,$B2,$4F
	.dc.b	$F5,$8B,$FE,$B0,$23,$FE,$B0,$6B,$FD,$48,$C1,$0F,$F6,$D9,$FC,$73
	.dc.b	$9F,$FD,$7F,$87,$FB,$7E,$FF,$DF,$07,$EE,$23,$35,$FE,$FF,$AE,$07
	.dc.b	$FB,$12,$B8,$5F,$FB,$42,$F8,$4E,$6B,$83,$FC,$3E,$F9,$BF,$EE,$F3
	.dc.b	$99,$F8,$CB,$FF,$70,$DF,$F4,$F9,$DC,$FA,$0E,$47,$53,$91,$83,$93
	.dc.b	$FB,$73,$8D,$FF,$BB,$6F,$FA,$64,$62,$BF,$CC,$71,$8A,$E8,$77,$AF
	.dc.b	$5B,$D0,$FD,$85,$71,$7F,$D2,$13,$C5,$D1,$F9,$C8,$D3,$FE,$58,$3E
	.dc.b	$F5,$FF,$A2,$A7,$85,$7D,$0B,$F8,$15,$E2,$8F,$43,$F1,$9D,$3F,$DE
	.dc.b	$AF,$AD,$F3,$8D,$57,$D3,$39,$BE,$4A,$FC,$83,$E3,$3F,$5F,$5C,$88
	.dc.b	$F8,$67,$E2,$BD,$0F,$F3,$2D,$7D,$67,$C4,$6C,$F8,$66,$CF,$F6,$44
	.dc.b	$FD,$37,$47,$86,$3F,$66,$DB,$F8,$75,$E9,$8E,$A7,$A5,$FD,$9D,$79
	.dc.b	$A3,$07,$CC,$71,$7F,$1A,$3F,$C4,$F1,$89,$E9,$18,$3C,$77,$F2,$5F
	.dc.b	$C5,$57,$18,$FC,$EC,$7C,$0A,$F5,$3E,$03,$E4,$9C,$9E,$CE,$CF,$C3
	.dc.b	$99,$21,$C6,$3E,$09,$E1,$9F,$C7,$9E,$1B,$FD,$4B,$8B,$F4,$CF,$92
	.dc.b	$BE,$69,$D8,$E9,$5F,$29,$F1,$1F,$93,$3C,$35,$D5,$FD,$4A,$AF,$F4
	.dc.b	$AF,$33,$DF,$FF,$50,$12,$BF,$D4,$0F,$FA,$8F,$A2,$47,$FA,$8D,$AF
	.dc.b	$FA,$78,$FF,$55,$F3,$FF,$A8,$BC,$6C,$46,$3A,$7F,$D4,$1D,$1F,$E9
	.dc.b	$F8,$62,$F9,$6F,$24,$7F,$AB,$5A,$F5,$8D,$67,$FE,$B3,$9C,$CF,$87
	.dc.b	$BA,$BF,$EB,$06,$BF,$EB,$E6,$A6,$BF,$EA,$13,$CE,$53,$D3,$D3,$87
	.dc.b	$FD,$50,$7F,$A9,$88,$7C,$1E,$07,$4F,$FA,$A1,$BD,$3F,$EA,$7E,$5F
	.dc.b	$F5,$19,$38,$BB,$CF,$FA,$96,$FF,$D4,$67,$FA,$53,$9F,$FD,$2D,$1F
	.dc.b	$E9,$27,$FD,$33,$D3,$FE,$CC,$8E,$1D,$3A,$E9,$FF,$51,$CF,$FA,$82
	.dc.b	$FF,$D6,$73,$FE,$A4,$7F,$D5,$07,$FB,$81,$E0,$D8,$E9,$8F,$FA,$52
	.dc.b	$3F,$D6,$51,$8C,$7F,$D7,$18,$F0,$9F,$3E,$3E,$98,$F2,$C7,$63,$38
	.dc.b	$FF,$B8,$A3,$30,$97,$6F,$F9,$FD,$B1,$FE,$5E,$6B,$FE,$B5,$31,$1B
	.dc.b	$23,$BA,$AF,$F4,$D8,$B7,$F8,$AA,$FF,$5D,$B6,$33,$C1,$F2,$DF,$15
	.dc.b	$E2,$FF,$90,$C5,$7F,$D8,$F5,$C5,$F1,$90,$F3,$1E,$13,$0E,$DC,$B4
	.dc.b	$95,$E1,$8F,$51,$C1,$C3,$CC,$E2,$E2,$38,$C7,$98,$E2,$6D,$CB,$F3
	.dc.b	$E7,$DA,$F5,$38,$1B,$9F,$C6,$57,$BE,$FE,$8D,$F2,$7E,$4C,$F8,$49
	.dc.b	$5C,$A7,$E7,$57,$1B,$E4,$D9,$F0,$1F,$2A,$B8,$BE,$53,$37,$D9,$E9
	.dc.b	$1A,$8F,$1B,$E6,$8D,$D7,$66,$9C,$AB,$C0,$FB,$4F,$CA,$5F,$AE,$78
	.dc.b	$46,$4F,$11,$6F,$C4,$E2,$78,$3F,$39,$AE,$EB,$BC,$7C,$76,$DF,$A5
	.dc.b	$B3,$EC,$3C,$71,$ED,$BF,$1D,$74,$39,$4F,$B0,$EC,$57,$2F,$9E,$38
	.dc.b	$CF,$38,$F8,$46,$DF,$13,$4F,$11,$F4,$CB,$7F,$39,$B7,$E8,$7E,$14
	.dc.b	$FD,$17,$C0,$76,$29,$E6,$2D,$F9,$8F,$64,$78,$A3,$2D,$8C,$C7,$80
	.dc.b	$F3,$1E,$03,$E3,$9E,$6B,$F7,$B4,$F1,$47,$27,$81,$ED,$2D,$FB,$5B
	.dc.b	$7E,$D8,$C1,$F8,$02,$D7,$69,$E2,$D8,$CC,$78,$2F,$E8,$3F,$06,$FC
	.dc.b	$53,$D2,$FB,$5F,$EC,$AB,$E0,$9E,$F2,$9E,$43,$E0,$CF,$16,$C7,$1A
	.dc.b	$EA,$DB,$F0,$C7,$CF,$69,$F3,$4B,$3D,$E3,$8B,$F9,$B8,$CB,$E1,$3C
	.dc.b	$75,$E1,$6D,$77,$7E,$81,$EF,$9E,$33,$BA,$EB,$3F,$C6,$0D,$BE,$56
	.dc.b	$CF,$95,$A7,$95,$E7,$5D,$E3,$E8,$9F,$3C,$D9,$E7,$9E,$E7,$F1,$27
	.dc.b	$E3,$9E,$D3,$F8,$F9,$FD,$DE,$3F,$BE,$57,$F4,$E8,$C1,$FD,$D2,$32
	.dc.b	$47,$04,$3F,$DC,$BC,$2F,$5D,$12,$FF,$C3,$33,$98,$FC,$DB,$5E,$33
	.dc.b	$FA,$D8,$C4,$F0,$8C,$1D,$27,$67,$27,$11,$B8,$C1,$E1,$C3,$FC,$7B
	.dc.b	$3C,$A3,$07,$9A,$BB,$CF,$E0,$1B,$19,$9E,$0F,$80,$FC,$AC,$69,$FF
	.dc.b	$1C,$C6,$AB,$C5,$1A,$8F,$2D,$78,$A3,$CA,$73,$9F,$DB,$57,$33,$A6
	.dc.b	$1F,$AC,$9D,$57,$17,$A1,$EA,$3C,$27,$4A,$F0,$9D,$2F,$C3,$1E,$B3
	.dc.b	$53,$CE,$75,$1C,$EF,$C3,$87,$F7,$F7,$C9,$7E,$DB,$E0,$7C,$77,$AD
	.dc.b	$78,$8F,$29,$E8,$3F,$2B,$3C,$9F,$09,$E3,$37,$7E,$8A,$E9,$1E,$3A
	.dc.b	$F0,$3B,$9F,$3B,$D4,$FD,$19,$6B,$D2,$FF,$6E,$7C,$0F,$6A,$F4,$BE
	.dc.b	$4A,$F4,$CF,$08,$FE,$C4,$F0,$AE,$E7,$96,$38,$1E,$53,$D6,$7C,$17
	.dc.b	$75,$DA,$3C,$B1,$C5,$D4,$7B,$9F,$56,$1F,$7E,$75,$5E,$37,$FC,$32
	.dc.b	$F8,$EB,$84,$73,$9E,$11,$DE,$78,$3F,$E0,$C7,$85,$7A,$0E,$F8,$7D
	.dc.b	$DC,$3E,$09,$F3,$A3,$CA,$6F,$0E,$D3,$B8,$F8,$07,$CB,$3E,$A5,$F9
	.dc.b	$0F,$81,$87,$92,$32,$43,$E1,$C7,$9E,$FC,$18,$7A,$27,$FB,$9B,$E8
	.dc.b	$76,$71,$74,$FF,$5C,$75,$5D,$E3,$D2,$47,$11,$AB,$F5,$4F,$C2,$3F
	.dc.b	$0D,$5F,$09,$FB,$73,$DA,$BF,$B3,$3C,$63,$67,$C2,$C3,$EF,$BE,$D9
	.dc.b	$EF,$5C,$E3,$E3,$1E,$3B,$F0,$D7,$CC,$8E,$67,$8C,$FC,$8C,$F1,$8E
	.dc.b	$B1,$C4,$DB,$C5,$D9,$E3,$C3,$E9,$3E,$3A,$EF,$1F,$51,$B1,$93,$9C
	.dc.b	$65,$FF,$48,$77,$7F,$44,$23,$FE,$B6,$7F,$D8,$58,$BC,$1A,$49,$FF
	.dc.b	$64,$C6,$5B,$F4,$BC,$08,$E0,$FF,$67,$F4,$FF,$B9,$DF,$F7,$3A,$E8
	.dc.b	$FF,$73,$F4,$F4,$F0,$BF,$DB,$78,$1A,$4C,$FF,$DA,$D8,$7F,$B6,$4F
	.dc.b	$F7,$A7,$3E,$08,$FF,$B6,$0F,$F6,$57,$3F,$FB,$5C,$9C,$FF,$ED,$7E
	.dc.b	$09,$3F,$EF,$AE,$2F,$F6,$7F,$2F,$FB,$35,$72,$7F,$B4,$27,$FB,$0C
	.dc.b	$69,$BC,$DF,$EC,$BE,$6F,$F6,$41,$2F,$FD,$99,$3F,$EC,$A9,$D4,$7F
	.dc.b	$93,$67,$53,$CB,$0F,$ED,$B1,$FE,$CA,$BF,$F7,$25,$6C,$E7,$1F,$ED
	.dc.b	$FA,$FF,$67,$61,$E5,$69,$FE,$DF,$3A,$1F,$ED,$E3,$FC,$9A,$7F,$91
	.dc.b	$EB,$FC,$85,$87,$92,$33,$1F,$E4,$59,$FF,$6C,$4E,$67,$84,$7F,$85
	.dc.b	$E7,$FD,$C9,$7F,$E6,$7C,$3F,$C2,$95,$FE,$23,$3F,$DC,$57,$C6,$BA
	.dc.b	$47,$F8,$5C,$CC,$FC,$0A,$E5,$A4,$9F,$1D,$F1,$C3,$C0,$79,$2F,$84
	.dc.b	$FA,$23,$C9,$3B,$8E,$D8,$7B,$AB,$71,$EA,$33,$3B,$7D,$13,$B7,$FD
	.dc.b	$FA,$72,$AF,$03,$83,$6E,$31,$63,$E9,$9E,$C3,$0F,$F7,$B6,$1D,$DF
	.dc.b	$8E,$4C,$7E,$ED,$71,$C7,$E3,$DF,$19,$CB,$EF,$3D,$35,$EC,$9D,$BC
	.dc.b	$E3,$F0,$38,$79,$1E,$07,$BB,$0F,$15,$F7,$C7,$D5,$8D,$AE,$A7,$BF
	.dc.b	$0E,$8F,$E2,$6F,$8B,$D2,$3D,$A7,$8E,$BE,$BC,$7F,$86,$F1,$FC,$3B
	.dc.b	$BC,$3D,$D7,$CA,$3E,$3D,$F2,$7C,$13,$E3,$7E,$11,$EC,$C3,$13,$E4
	.dc.b	$BE,$B3,$CA,$BA,$CF,$8E,$3E,$D1,$E3,$BF,$05,$7C,$8A,$F9,$B3,$C4
	.dc.b	$FC,$8D,$71,$7E,$F9,$C5,$C1,$E1,$C3,$E4,$C7,$B4,$E2,$7A,$8C,$1F
	.dc.b	$A0,$30,$77,$31,$8F,$D7,$71,$7F,$93,$8C,$3E,$78,$EB,$18,$9F,$9F
	.dc.b	$1F,$18,$E7,$8F,$C6,$C3,$E1,$C7,$82,$F9,$3A,$3C,$AF,$79,$E1,$1D
	.dc.b	$9A,$FE,$07,$0F,$83,$58,$AF,$A2,$74,$39,$4F,$4B,$F8,$AF,$B1,$B1
	.dc.b	$B6,$9E,$AA,$FC,$16,$1E,$DA,$FB,$27,$B6,$BB,$D6,$EF,$EC,$C7,$39
	.dc.b	$FA,$C7,$BE,$7F,$82,$DF,$A2,$3D,$F8,$FE,$0A,$73,$1B,$3F,$43,$87
	.dc.b	$85,$FC,$09,$E1,$9E,$0E,$B1,$F8,$55,$EA,$8F,$3D,$FA,$A7,$31,$E4
	.dc.b	$BE,$93,$F8,$98,$C4,$E9,$ED,$1E,$77,$F0,$77,$E0,$7F,$75,$19,$C3
	.dc.b	$E8,$4E,$63,$F4,$8D,$C7,$F0,$38,$78,$1F,$B1,$19,$BF,$BC,$7D,$73
	.dc.b	$C6,$53,$C6,$7C,$53,$C2,$5A,$EE,$78,$F1,$F9,$77,$DE,$7E,$7C,$E5
	.dc.b	$F9,$07,$F6,$C9,$FF,$6B,$91,$FF,$62,$1F,$EC,$5C,$3F,$B9,$B8,$38
	.dc.b	$14,$FE,$C2,$7F,$7B,$8E,$4D,$3F,$89,$72,$FF,$A7,$27,$FD,$F5,$8F
	.dc.b	$F2,$F8,$C1,$F8,$B9,$FF,$78,$CF,$E6,$E3,$31,$F7,$27,$C9,$18,$8F
	.dc.b	$C2,$DA,$73,$1F,$EF,$B3,$FB,$84,$EE,$7F,$01,$1F,$9B,$8C,$4F,$EA
	.dc.b	$DB,$1F,$B4,$6B,$FC,$11,$B5,$E5,$8E,$CF,$C6,$7A,$18,$BD,$CF,$B6
	.dc.b	$3F,$CE,$27,$33,$47,$C6,$29,$C8,$97,$E8,$3D,$CF,$C6,$9F,$F0,$24
	.dc.b	$F9,$1B,$1C,$A7,$6F,$82,$BD,$0F,$4B,$F8,$0E,$DF,$92,$E4,$F8,$46
	.dc.b	$4F,$01,$FD,$FA,$FC,$0D,$34,$77,$BF,$1B,$CC,$F8,$05,$AF,$1D,$F9
	.dc.b	$9F,$C2,$C7,$F8,$92,$BA,$BF,$E0,$66,$CF,$F5,$16,$DF,$E2,$E7,$C3
	.dc.b	$5F,$E9,$93,$AC,$7B,$2B,$E5,$34,$93,$EE,$3C,$97,$E8,$9F,$25,$FD
	.dc.b	$9B,$E9,$39,$BE,$37,$EC,$BD,$BF,$97,$3E,$39,$E9,$3C,$B7,$E8,$7D
	.dc.b	$C6,$08,$78,$8F,$98,$53,$91,$63,$E1,$3C,$1F,$0B,$FA,$49,$F2,$DF
	.dc.b	$CC,$8F,$D1,$9B,$8C,$DF,$AE,$3F,$CD,$F7,$EF,$3C,$CF,$03,$C1,$3F
	.dc.b	$3D,$A7,$A8,$A6,$EF,$C6,$7D,$07,$E1,$BE,$87,$77,$FA,$98,$DC,$F9
	.dc.b	$27,$E0,$C6,$1F,$AC,$D7,$E8,$46,$DF,$8B,$19,$9F,$C0,$C7,$B0,$FA
	.dc.b	$87,$E5,$4E,$B3,$C6,$7C,$87,$1B,$F8,$92,$AF,$94,$79,$EB,$E6,$DF
	.dc.b	$C7,$9F,$98,$77,$3F,$CB,$BD,$BF,$C9,$DE,$4D,$A3,$4C,$77,$49,$FE
	.dc.b	$5D,$28,$C4,$7F,$AA,$79,$FF,$D9,$6D,$3F,$D9,$87,$FB,$23,$BB,$FD
	.dc.b	$9A,$47,$07,$FB,$1B,$B5,$CB,$4F,$F6,$4F,$37,$FB,$1A,$7D,$1C,$9F
	.dc.b	$EB,$E7,$FA,$03,$5F,$F5,$FF,$1F,$9B,$93,$FD,$7F,$1F,$EC,$33,$FD
	.dc.b	$46,$7F,$4C,$D1,$C7,$FE,$75,$8F,$F6,$13,$5D,$35,$FF,$5D,$98,$E2
	.dc.b	$FF,$64,$B4,$E2,$E1,$D9,$0F,$09,$4F,$F2,$11,$C4,$86,$67,$CB,$1C
	.dc.b	$E3,$89,$0E,$26,$E7,$CC,$E6,$78,$BE,$42,$1D,$8D,$47,$07,$99,$89
	.dc.b	$F0,$38,$24,$F4,$79,$18,$9E,$51,$98,$F2,$4F,$28,$C4,$75,$3C,$66
	.dc.b	$8A,$64,$EA,$F0,$7B,$47,$39,$EE,$47,$04,$8E,$67,$02,$CF,$32,$4F
	.dc.b	$18,$CB,$5F,$0B,$5F,$31,$89,$EC,$D3,$8B,$87,$64,$3D,$85,$3C,$87
	.dc.b	$12,$19,$9F,$2C,$73,$8E,$24,$38,$9B,$9F,$33,$99,$E2,$F9,$08,$76
	.dc.b	$35,$1C,$1E,$66,$27,$C0,$E0,$93,$D1,$E4,$62,$79,$46,$63,$C9,$3C
	.dc.b	$A3,$11,$D4,$F1,$9A,$29,$93,$AB,$C1,$ED,$1C,$E7,$B9,$1C,$12,$39
	.dc.b	$9C,$0B,$3C,$C9,$3C,$63,$2D,$7C,$2D,$7C,$C6,$27,$B3,$4E,$2E,$1D
	.dc.b	$90,$F6,$14,$F2,$1C,$48,$66,$7C,$B1,$CE,$38,$90,$E2,$6E,$7C,$CE
	.dc.b	$67,$8B,$E4,$21,$D8,$D4,$70,$79,$98,$9F,$03,$82,$4F,$47,$91,$89
	.dc.b	$E5,$19,$8F,$24,$F2,$8C,$47,$53,$C6,$68,$A6,$4E,$AF,$07,$B4,$73
	.dc.b	$9E,$E4,$70,$48,$E6,$70,$2C,$F3,$24,$F1,$8C,$B5,$F0,$B5,$F3,$18
	.dc.b	$9E,$CD,$38,$B8,$76,$43,$D8,$53,$C8,$71,$21,$99,$F2,$C7,$38,$E2
	.dc.b	$43,$89,$B9,$F3,$39,$9E,$2F,$90,$87,$63,$51,$C1,$E6,$62,$7C,$0E
	.dc.b	$09,$3D,$1E,$46,$27,$94,$66,$3C,$93,$CA,$31,$1D,$4F,$19,$A2,$99
	.dc.b	$3A,$BC,$1E,$D1,$CE,$7B,$91,$C1,$23,$FD,$8D,$2A,$5A,$5C,$AE,$EF
	.dc.b	$F0,$4C,$61,$B1,$98,$E1,$7C,$EB,$94,$E6,$7F,$C6,$97,$FE,$10,$7F
	.dc.b	$C1,$CD,$7A,$98,$9F,$F0,$B1,$1F,$F0,$B1,$C9,$C3,$B3,$35,$E1,$6C
	.dc.b	$F5,$7C,$AD,$79,$1A,$9E,$6F,$88,$87,$07,$2D,$38,$10,$F1,$10,$E2
	.dc.b	$76,$30,$71,$29,$93,$73,$E9,$72,$D7,$33,$E9,$8C,$47,$33,$33,$E6
	.dc.b	$7A,$19,$21,$82,$47,$26,$CF,$52,$4F,$39,$D1,$23,$33,$F0,$4F,$09
	.dc.b	$4E,$F1,$B9,$E5,$3C,$27,$DA,$D7,$89,$88,$D9,$1E,$87,$99,$C3,$B3
	.dc.b	$33,$DD,$B3,$D5,$F1,$B5,$E4,$6A,$79,$BE,$22,$1C,$1C,$B4,$E0,$43
	.dc.b	$C4,$43,$89,$D8,$C1,$C4,$A6,$4D,$CF,$A5,$CB,$5C,$CF,$A6,$31,$1C
	.dc.b	$CC,$CF,$99,$E8,$64,$86,$09,$1C,$9B,$3D,$49,$3C,$E7,$44,$8C,$CF
	.dc.b	$C1,$3C,$25,$3B,$C6,$E7,$94,$F0,$9F,$6B,$5E,$26,$23,$64,$7A,$1E
	.dc.b	$67,$0E,$CC,$CF,$76,$CF,$57,$C6,$D7,$91,$A9,$E6,$F8,$88,$70,$72
	.dc.b	$D3,$81,$0F,$11,$0E,$27,$63,$07,$12,$99,$37,$3E,$97,$2D,$73,$3E
	.dc.b	$98,$C4,$73,$33,$3E,$67,$A1,$92,$18,$24,$72,$6C,$F5,$24,$F3,$9D
	.dc.b	$12,$33,$3F,$04,$F0,$94,$EF,$1B,$9E,$53,$C2,$7D,$AD,$78,$98,$8D
	.dc.b	$91,$E8,$79,$9C,$3B,$33,$3D,$DB,$3D,$5F,$1B,$5E,$46,$A7,$9B,$E2
	.dc.b	$21,$C1,$CB,$4E,$04,$3C,$44,$38,$9D,$8C,$1C,$4A,$64,$DC,$FA,$5C
	.dc.b	$B5,$CC,$FA,$63,$11,$CC,$CC,$F9,$9E,$86,$48,$60,$91,$C9,$B3,$D4
	.dc.b	$93,$CE,$74,$48,$CC,$FC,$13,$C2,$53,$BC,$7F,$59,$FC,$A4,$40

;--------------------------------------------------------------------------------
;	FTAN.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
ftan_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FTAN.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FTAN.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	ftan_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1953+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	ftan.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTAN.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1953+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	ftan.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTAN.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
ftan_data_compressed::
	.dc.b	$00,$02,$FA,$EC,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$F3,$FE,$4D,$CD,$F8,$3E,$27,$13,$D5,$F5,$7A,$BB
	.dc.b	$AF,$C3,$52,$F8,$FA,$59,$F4,$B5,$E9,$6F,$D2,$E1,$E9,$71,$F4,$B9
	.dc.b	$7A,$5C,$FD,$2E,$9E,$97,$5F,$4B,$B7,$A5,$DF,$D2,$F0,$7A,$5E,$1F
	.dc.b	$4B,$C5,$E9,$7A,$7F,$58,$FF,$77,$F1,$1F,$F5,$4B,$7E,$37,$5F,$15
	.dc.b	$BF,$80,$E1,$EE,$71,$F5,$F9,$7A,$FC,$FD,$6E,$9E,$B7,$5F,$5B,$B7
	.dc.b	$AD,$DF,$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$78,$FC,$2F,$27,$AD,$E5
	.dc.b	$F5,$BC,$DE,$B7,$E0,$FD,$AF,$CA,$7C,$7F,$E5,$7C,$7F,$F6,$B7,$C4
	.dc.b	$FA,$FD,$9C,$CE,$BE,$27,$7F,$0B,$C3,$E0,$B8,$FB,$DC,$BC,$DE,$7E
	.dc.b	$6F,$4F,$37,$AF,$9B,$DB,$CD,$EF,$E6,$F8,$3C,$DF,$0F,$9B,$E2,$F3
	.dc.b	$7C,$7E,$6F,$93,$CD,$F2,$F9,$BE,$6F,$37,$CF,$E6,$FA,$3C,$DF,$D6
	.dc.b	$7C,$AF,$FB,$60,$FF,$48,$9F,$E4,$B3,$FC,$06,$7E,$D4,$FD,$A7,$49
	.dc.b	$79,$1C,$BC,$8E,$7E,$47,$4F,$23,$AF,$91,$DB,$C8,$EF,$E4,$78,$3C
	.dc.b	$8F,$0F,$91,$E2,$F2,$3C,$7E,$47,$93,$C8,$F2,$F9,$1F,$9D,$FD,$67
	.dc.b	$FD,$AB,$C0,$F2,$7D,$3E,$3F,$F0,$5F,$97,$7C,$AE,$FE,$27,$87,$C7
	.dc.b	$71,$F8,$CE,$5F,$0F,$CF,$E1,$7A,$7C,$2F,$5F,$85,$ED,$F0,$BD,$FE
	.dc.b	$17,$C1,$F0,$BE,$1F,$85,$F1,$7C,$2F,$8F,$E1,$7C,$9F,$0B,$E5,$F8
	.dc.b	$5F,$37,$C2,$F9,$FE,$17,$D1,$F0,$9E,$9F,$84,$F5,$7B,$DF,$B7,$FD
	.dc.b	$6B,$F8,$E7,$C2,$17,$C0,$67,$E0,$35,$F0,$1B,$F8,$0E,$1F,$01,$C7
	.dc.b	$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7,$C0,$77,$F8,$0F
	.dc.b	$07,$C0,$78,$7E,$03,$C5,$F0,$1F,$27,$C6,$F9,$5E,$37,$97,$E0,$3E
	.dc.b	$67,$8F,$F3,$7C,$7F,$9D,$E3,$FC,$FF,$1F,$E8,$78,$FF,$47,$C7,$FA
	.dc.b	$5E,$3F,$D3,$F1,$FE,$A7,$8F,$F5,$7C,$7F,$AD,$E3,$FD,$7F,$1F,$EC
	.dc.b	$78,$FF,$67,$C7,$FB,$5E,$3F,$DB,$F1,$FE,$E7,$8F,$F7,$7C,$7F,$E8
	.dc.b	$5F,$33,$FA,$F3,$FE,$00,$1F,$EB,$83,$FC,$56,$7F,$65,$3F,$91,$1F
	.dc.b	$C7,$BD,$C2,$F0,$79,$F8,$3D,$3C,$1E,$BE,$0F,$6F,$07,$BF,$83,$E0
	.dc.b	$F0,$7C,$3E,$0F,$8B,$C1,$F1,$F7,$F9,$3E,$07,$CB,$F0,$3E,$6F,$81
	.dc.b	$F3,$FC,$0F,$A3,$E0,$7F,$01,$5F,$D6,$FC,$57,$CA,$5F,$8B,$FC,$26
	.dc.b	$3F,$85,$C7,$F0,$D8,$FE,$1F,$1F,$E7,$D8,$FF,$26,$3F,$92,$8F,$87
	.dc.b	$CF,$C3,$EB,$E1,$F7,$F0,$FC,$3E,$1F,$8F,$C3,$F2,$F8,$7E,$7F,$0F
	.dc.b	$D3,$E1,$FA,$FC,$3F,$6F,$87,$EF,$F0,$FE,$0F,$87,$F0,$FC,$3F,$8B
	.dc.b	$E1,$FC,$7F,$0F,$E4,$F8,$7F,$2F,$C2,$F9,$BE,$17,$FA,$A7,$EB,$7F
	.dc.b	$B4,$F2,$7F,$A7,$2F,$73,$3D,$FA,$F7,$37,$EE,$70,$F7,$38,$FB,$9C
	.dc.b	$BD,$CE,$7E,$E7,$4F,$73,$AF,$B9,$F2,$FC,$0F,$99,$E0,$78,$3D,$CF
	.dc.b	$9D,$E0,$FC,$FF,$07,$E8,$78,$3F,$47,$C1,$FA,$5E,$0F,$D3,$F0,$7E
	.dc.b	$A7,$83,$F5,$7C,$1F,$AD,$E0,$FD,$7F,$07,$EC,$78,$3F,$67,$C1,$FB
	.dc.b	$5E,$0F,$DB,$F0,$7E,$E7,$83,$F7,$7C,$1F,$BD,$E0,$FF,$86,$7E,$6F
	.dc.b	$E1,$78,$47,$FD,$C2,$7F,$A6,$CF,$F2,$99,$FE,$0A,$75,$DB,$C7,$B7
	.dc.b	$97,$6F,$3E,$DE,$9D,$BD,$7B,$7B,$75,$F7,$F7,$7C,$1E,$EF,$87,$DD
	.dc.b	$F1,$7B,$BE,$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$E7,$F7,$7E,$CE
	.dc.b	$5F,$E3,$1F,$9C,$FE,$BC,$7F,$16,$E3,$FF,$1F,$2F,$83,$D7,$C1,$FE
	.dc.b	$03,$6F,$C0,$ED,$F8,$2D,$BF,$07,$B7,$E1,$36,$FC,$2E,$DF,$86,$DB
	.dc.b	$F0,$FB,$7F,$86,$FF,$5C,$FE,$4A,$77,$F8,$73,$DF,$7F,$F0,$C3,$FE
	.dc.b	$F0,$8C,$FC,$2F,$1F,$85,$E5,$F0,$BC,$FE,$13,$A7,$C2,$75,$F8,$4F
	.dc.b	$F2,$67,$EE,$6F,$84,$F0,$7C,$2F,$85,$F1,$7C,$2F,$8F,$E1,$7C,$9F
	.dc.b	$0B,$E5,$F8,$5F,$37,$C2,$F9,$FD,$FF,$47,$C1,$FA,$7E,$0F,$D5,$F0
	.dc.b	$7E,$BF,$83,$FF,$1B,$7E,$BD,$FE,$1E,$3F,$C3,$A7,$EE,$7A,$3F,$C6
	.dc.b	$4B,$DE,$DF,$BD,$F4,$7A,$3E,$97,$47,$D3,$E8,$E7,$EF,$7D,$5E,$9F
	.dc.b	$AD,$D3,$F5,$FA,$7E,$C7,$47,$D9,$E8,$FB,$5D,$1F,$6F,$A3,$EE,$74
	.dc.b	$7D,$DE,$8F,$BD,$D1,$F7,$FA,$3F,$01,$D1,$F8,$1E,$8F,$C1,$74,$7E
	.dc.b	$0F,$A3,$FC,$E9,$F4,$1F,$C6,$BA,$7F,$D0,$0B,$E1,$B3,$F0,$3A,$EA
	.dc.b	$DF,$57,$0E,$7E,$3F,$03,$CB,$E0,$79,$FC,$0F,$4F,$81,$EB,$F0,$3D
	.dc.b	$BE,$07,$BF,$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93
	.dc.b	$E0,$7C,$BF,$03,$E6,$F8,$1F,$3F,$C0,$FF,$A2,$F9,$3F,$66,$7F,$13
	.dc.b	$E7,$FF,$4C,$7E,$37,$FE,$C4,$87,$C2,$EF,$E1,$78,$7C,$2F,$1F,$84
	.dc.b	$E5,$F0,$9C,$FE,$13,$A7,$C2,$75,$F8,$4E,$DF,$09,$DF,$E1,$3C,$1F
	.dc.b	$09,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$FA,$87,$93,$F8,$27,$23,$E0,$7D,$5F,$03,$FE,$A5,$FA,$2F,$F2
	.dc.b	$91,$FB,$5E,$83,$FA,$76,$FC,$1E,$FD,$FE,$1E,$FF,$1F,$7F,$97,$BF
	.dc.b	$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BF,$E1,$F7,$FC
	.dc.b	$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF,$F9,$FD,$FF,$47,$BF
	.dc.b	$FE,$BD,$FA,$3F,$C5,$73,$1F,$C5,$3F,$EC,$B7,$F4,$CE,$7D,$EE,$1E
	.dc.b	$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$7C,$9F,$3F,$B7,$BD,$DF,$DE
	.dc.b	$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$E5,$F7,$BC,$DE
	.dc.b	$F7,$9F,$DE,$F4,$7B,$DE,$9F,$7B,$FD,$AD,$F4,$9F,$CF,$BC,$F7,$B9
	.dc.b	$8F,$89,$CF,$C0,$6B,$E0,$37,$F0,$1C,$39,$78,$DF,$2F,$80,$E7,$F0
	.dc.b	$1D,$3E,$03,$AF,$C0,$76,$F8,$0E,$FF,$01,$E0,$F8,$0F,$0F,$C0,$78
	.dc.b	$BE,$03,$C7,$F0,$1E,$4F,$80,$F2,$FC,$07,$E2,$F5,$FC,$66,$BF,$8D
	.dc.b	$D7,$F1,$DA,$FE,$3F,$5F,$5F,$BF,$EC,$F7,$FF,$DE,$5E,$F7,$FA,$7C
	.dc.b	$FC,$71,$F8,$D0,$6B,$DE,$DF,$BD,$C3,$DE,$E3,$EF,$72,$F7,$B9,$FB
	.dc.b	$DF,$9D,$E6,$EB,$EF,$76,$F7,$BB,$FB,$DF,$A1,$E9,$FD,$17,$4F,$E8
	.dc.b	$FA,$7F,$49,$D3,$FA,$5E,$9F,$D3,$74,$FE,$9F,$93,$F5,$1C,$8F,$52
	.dc.b	$7E,$39,$FE,$B7,$3F,$4C,$7E,$94,$F4,$BA,$E5,$E0,$A3,$FD,$D8,$7F
	.dc.b	$92,$CF,$F2,$53,$D3,$8B,$70,$E2,$BC,$38,$E1,$F9,$3E,$FF,$CA,$F7
	.dc.b	$FE,$5F,$BF,$F3,$3D,$FF,$9B,$EF,$FC,$EF,$7F,$E7,$FB,$FF,$43,$DF
	.dc.b	$FA,$3E,$FF,$93,$E9,$BF,$4E,$7E,$9B,$84,$FE,$98,$FE,$94,$FC,$15
	.dc.b	$EF,$71,$F7,$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE
	.dc.b	$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27,$BD,$F0,$BC,$8F,$87
	.dc.b	$D5,$F1,$3A,$BE,$2F,$57,$C6,$EA,$F8,$FD,$5F,$23,$AB,$E4,$F5,$7C
	.dc.b	$AE,$AF,$97,$D5,$F4,$BD,$0F,$E0,$A7,$BC,$FE,$06,$7B,$BA,$CF,$B9
	.dc.b	$FC,$03,$FE,$D8,$3F,$DA,$EE,$39,$FA,$73,$F5,$E7,$F9,$9E,$F7,$CD
	.dc.b	$F7,$BE,$77,$BD,$E1,$EA,$FA,$1E,$F7,$E1,$3E,$9F,$FC,$76,$7F,$49
	.dc.b	$CC,$FF,$48,$7F,$20,$BD,$BE,$1E,$DF,$1F,$6F,$97,$B7,$CF,$DB,$E9
	.dc.b	$ED,$F5,$E2,$ED,$ED,$FC,$1F,$0F,$E1,$78,$7F,$0F,$C3,$F8,$9E,$1F
	.dc.b	$C5,$F0,$FC,$9E,$DF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$D1,$EB,$FA,$7D
	.dc.b	$7F,$57,$AF,$EB,$F5,$FD,$9E,$BF,$B7,$D7,$FD,$0F,$D4,$7E,$8B,$AC
	.dc.b	$FF,$4D,$3F,$A6,$1F,$A8,$73,$EC,$70,$F6,$F8,$FB,$7C,$BD,$BF,$B5
	.dc.b	$E9,$74,$F6,$FA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$D6,$F0,$FA,$DE,$2F
	.dc.b	$5B,$C7,$EB,$79,$3D,$6F,$2F,$AD,$E6,$F5,$BC,$FE,$B7,$A3,$D6,$FC
	.dc.b	$47,$A5,$EA,$F5,$BF,$77,$E7,$7F,$4E,$3F,$50,$7F,$4C,$3F,$14,$7E
	.dc.b	$27,$9F,$EC,$7B,$1C,$3D,$8E,$3E,$C7,$2F,$63,$9F,$B1,$D3,$D6,$EB
	.dc.b	$E6,$F6,$F3,$7B,$F9,$BE,$0F,$37,$C3,$E6,$F8,$BB,$BC,$7C,$3E,$4F
	.dc.b	$63,$CB,$EC,$79,$BD,$8F,$E2,$3F,$53,$FE,$88,$3F,$1C,$7E,$37,$8C
	.dc.b	$FE,$35,$FE,$3C,$FF,$A2,$47,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7F,$D7,$69,$E0,$F5,$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F
	.dc.b	$5F,$CB,$EB,$F9,$BD,$7F,$3F,$AF,$E8,$F5,$FD,$3E,$BF,$F2,$AF,$AA
	.dc.b	$FC,$51,$FE,$15,$3F,$C2,$87,$E2,$47,$AF,$BF,$5F,$87,$AF,$FC,$07
	.dc.b	$AB,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F,$07,$AF
	.dc.b	$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB,$F9,$FD
	.dc.b	$7F,$47,$AF,$FD,$0B,$EA,$FF,$C2,$A7,$F2,$F3,$FC,$26,$7F,$11,$3F
	.dc.b	$10,$3D,$7E,$1E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB
	.dc.b	$7A,$FD,$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$79,$DE
	.dc.b	$5F,$03,$CD,$EB,$FF,$58,$F8,$4F,$F6,$79,$FE,$5D,$F8,$C7,$C5,$FF
	.dc.b	$22,$F6,$BF,$D1,$DF,$07,$BF,$73,$87,$BD,$FE,$96,$E6,$FF,$4B,$F3
	.dc.b	$7F,$A6,$39,$BF,$D3,$3C,$DF,$E9,$AE,$6F,$F4,$DF,$37,$FA,$73,$9B
	.dc.b	$FD,$3B,$CD,$FE,$9E,$E6,$FF,$4F,$F3,$7F,$A8,$39,$BF,$D4,$3C,$DF
	.dc.b	$EA,$2E,$6F,$F5,$1F,$37,$F7,$5E,$57,$A0,$F9,$6C,$3E,$93,$F4,$1F
	.dc.b	$97,$3C,$CF,$EE,$AF,$E2,$3A,$17,$CB,$9F,$8F,$3F,$2F,$7B,$9F,$BD
	.dc.b	$D3,$DE,$EB,$EF,$76,$F7,$BB,$FB,$DE,$0F,$7B,$C3,$EF,$78,$BD,$EF
	.dc.b	$1F,$BD,$E4,$F7,$BC,$BE,$F7,$9B,$93,$CF,$EF,$7A,$3D,$EF,$A5,$C1
	.dc.b	$F4,$F8,$3E,$A7,$07,$D5,$E0,$F6,$FB,$DF,$5F,$87,$EC,$70,$FD,$9E
	.dc.b	$1F,$B5,$C3,$F6,$F8,$7E,$E7,$0F,$DD,$E1,$FB,$DC,$3F,$7F,$87,$F0
	.dc.b	$1C,$3F,$81,$E1,$FC,$16,$BF,$E1,$AF,$AD,$FE,$74,$7F,$A1,$4F,$F4
	.dc.b	$21,$FC,$E4,$F9,$5C,$C5,$EE,$F1,$E9,$E5,$EE,$F3,$F7,$7A,$7B,$BD
	.dc.b	$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7,$7C
	.dc.b	$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$D1,$EE,$FF,$8C,$BE,$BB,$FD
	.dc.b	$0A,$7F,$87,$CF,$F4,$19,$FE,$1E,$F2,$1F,$AF,$FF,$57,$8F,$E7,$47
	.dc.b	$EC,$F9,$FF,$0B,$0F,$99,$FC,$82,$F9,$4B,$F1,$58,$F8,$3C,$F5,$6B
	.dc.b	$AB,$7D,$5C,$3A,$B8,$F5,$72,$EA,$E7,$D5,$D3,$AB,$AF,$57,$6E,$AE
	.dc.b	$FD,$5E,$0E,$AF,$0F,$57,$F9,$3B,$F1,$FF,$F5,$E7,$2F,$F9,$5D,$77
	.dc.b	$F9,$BB,$3C,$FD,$9E,$8E,$CF,$4F,$67,$EA,$7F,$60,$FF,$79,$76,$9F
	.dc.b	$F5,$4B,$7E,$4B,$5F,$07,$BF,$8B,$E1,$F0,$DC,$7E,$1B,$97,$C2,$73
	.dc.b	$F8,$4E,$9F,$09,$D7,$E1,$3B,$7C,$27,$7F,$84,$FA,$98,$7D,$5C,$3C
	.dc.b	$5F,$09,$F5,$F1,$FB,$18,$FD,$9C,$7E,$D6,$3F,$6F,$1F,$B9,$7F,$77
	.dc.b	$E1,$3E,$F7,$C2,$7D,$FF,$84,$FC,$07,$C2,$7E,$07,$E1,$3F,$05,$F0
	.dc.b	$9F,$83,$F8,$4F,$F3,$AF,$D9,$7F,$07,$EE,$FF,$3F,$3F,$E3,$33,$FB
	.dc.b	$31,$FD,$95,$2F,$82,$E3,$F0,$5C,$BE,$0B,$9F,$C1,$74,$F8,$2E,$BF
	.dc.b	$05,$DB,$E0,$BB,$FC,$17,$83,$E0,$BC,$3F,$05,$E2,$F8,$2F,$1F,$C1
	.dc.b	$79,$3E,$0B,$CB,$F0,$5E,$6F,$2F,$CF,$F0,$5E,$8F,$82,$FF,$46,$FB
	.dc.b	$9F,$E2,$D3,$FD,$F0,$7F,$BD,$CF,$F1,$61,$FB,$43,$F6,$7C,$BF,$C7
	.dc.b	$BC,$A7,$6F,$2F,$6F,$9F,$B7,$F8,$3E,$FF,$C2,$77,$FE,$17,$BF,$F0
	.dc.b	$DD,$FF,$87,$EF,$FC,$47,$7F,$F3,$1F,$D8,$7E,$57,$9A,$7F,$A4,$1F
	.dc.b	$E5,$B7,$1F,$09,$BF,$07,$87,$83,$C7,$C1,$E5,$E0,$F3,$F0,$7A,$78
	.dc.b	$3D,$7C,$1E,$DE,$0F,$7F,$07,$C1,$E0,$F8,$7C,$1F,$17,$83,$E3,$F0
	.dc.b	$7C,$9E,$0F,$97,$C1,$F3,$78,$3E,$7F,$07,$FA,$47,$4F,$FA,$DB,$94
	.dc.b	$FE,$F7,$C0,$FE,$86,$FE,$F5,$BE,$FF,$0F,$0B,$8F,$BF,$CB,$C1,$E7
	.dc.b	$E0,$F4,$F7,$FE,$A9,$F5,$BD,$FF,$AF,$EF,$F8,$1F,$B3,$EF,$FD,$AF
	.dc.b	$7F,$ED,$FB,$FF,$73,$DF,$FB,$BE,$FF,$DE,$F7,$FE,$FF,$BF,$F8,$0F
	.dc.b	$7B,$F0,$3E,$F7,$E0,$BD,$EF,$C1,$FB,$DF,$ED,$0F,$B4,$7C,$ED,$F9
	.dc.b	$7C,$7C,$0E,$7E,$07,$5F,$03,$BF,$81,$E1,$F0,$3C,$7E,$07,$97,$C0
	.dc.b	$F3,$F8,$1E,$9F,$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7B,$FE,$1F
	.dc.b	$7F,$C5,$EF,$F8,$FC,$8F,$27,$BF,$E5,$F7,$FC,$DE,$FF,$FB,$9B,$ED
	.dc.b	$7F,$65,$E6,$1F,$F5,$21,$FE,$57,$3F,$64,$7E,$C5,$FF,$7A,$3F,$E5
	.dc.b	$66,$FB,$9F,$73,$C7,$E9,$EE,$75,$F7,$3B,$7B,$9F,$80,$F2,$7F,$03
	.dc.b	$E4,$FE,$0B,$C9,$FC,$1F,$93,$F8,$4F,$27,$F0,$BE,$4F,$E1,$BC,$9F
	.dc.b	$C3,$F9,$3F,$88,$F2,$7F,$13,$E4,$FE,$2B,$C9,$FF,$09,$7E,$C5,$FC
	.dc.b	$A8,$FE,$52,$3C,$7C,$F8,$FA,$F1,$F7,$E3,$F0,$F1,$F8,$F8,$FC,$BC
	.dc.b	$7E,$7E,$3F,$4F,$1F,$AF,$8F,$DB,$C7,$EF,$E3,$F8,$3C,$7F,$0F,$8F
	.dc.b	$E2,$F1,$FC,$7E,$3F,$93,$C7,$F2,$F8,$FE,$6F,$1F,$FC,$51,$FB,$1F
	.dc.b	$EE,$0F,$F1,$F1,$FB,$73,$E7,$FB,$DF,$57,$DE,$DF,$C0,$7D,$7F,$7F
	.dc.b	$EC,$7B,$FF,$67,$DF,$FB,$5E,$FF,$DB,$F7,$FE,$E7,$BF,$F7,$7D,$FF
	.dc.b	$BD,$EF,$FD,$FF,$7D,$E9,$CF,$B8,$FF,$63,$1F,$B1,$4F,$EC,$5B,$EF
	.dc.b	$F0,$73,$EF,$D9,$FE,$3B,$1C,$3D,$EC,$7B,$D9,$F7,$B5,$EF,$6F,$D9
	.dc.b	$E1,$EB,$71,$F5,$B9,$7A,$DC,$FD,$6E,$9E,$B7,$5F,$5B,$B7,$AD,$DF
	.dc.b	$D6,$F0,$7A,$DE,$1F,$5B,$C5,$EB,$79,$7B,$FF,$63,$C7,$EE,$5E,$E6
	.dc.b	$7D,$6D,$7A,$DB,$F5,$B8,$7A,$DC,$7D,$6E,$5E,$B7,$3F,$5B,$A7,$AD
	.dc.b	$D7,$D6,$ED,$EB,$77,$F5,$BC,$1E,$B7,$87,$D6,$F1,$7A,$DE,$3F,$5B
	.dc.b	$C9,$EB,$79,$7D,$6F,$37,$AD,$E7,$F5,$BE,$67,$DC,$FE,$C8,$7C,$C6
	.dc.b	$3E,$43,$3F,$0B,$AE,$AD,$FC,$2F,$0F,$80,$E3,$F0,$1C,$BE,$03,$9F
	.dc.b	$C0,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$05,$F7,$5F,$83,$F3
	.dc.b	$0F,$E2,$4F,$C4,$0F,$7F,$7E,$FF,$0F,$85,$E3,$EF,$F2,$F7,$F9,$FB
	.dc.b	$FD,$3D,$FE,$BE,$FF,$6F,$7F,$BF,$BF,$E0,$F7,$FC,$3E,$FF,$8B,$DF
	.dc.b	$F1,$FB,$FE,$4F,$7F,$CB,$EF,$F9,$BD,$FF,$3F,$BF,$E8,$F7,$FF,$35
	.dc.b	$F7,$7F,$86,$E7,$3F,$86,$7F,$4A,$FF,$09,$6F,$BD,$C3,$DE,$E3,$EF
	.dc.b	$72,$F7,$B9,$FB,$DD,$3D,$EE,$BE,$F7,$6F,$7B,$BF,$BD,$E0,$F7,$BC
	.dc.b	$3E,$F7,$8B,$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD
	.dc.b	$E8,$F7,$BF,$65,$C5,$FC,$E0,$FC,$3F,$27,$EF,$17,$B9,$AF,$73,$7E
	.dc.b	$E7,$0F,$73,$8F,$B9,$CB,$DC,$E7,$EE,$74,$F7,$3A,$FB,$9D,$BD,$CE
	.dc.b	$FE,$E7,$83,$DC,$F0,$FB,$9E,$2F,$73,$C7,$EE,$79,$3D,$CF,$2F,$B9
	.dc.b	$E6,$F7,$3C,$FE,$E7,$F0,$9F,$BC,$FC,$41,$FC,$3C,$FE,$1C,$7E,$1F
	.dc.b	$9F,$F8,$AB,$FE,$9C,$3F,$86,$9F,$EC,$97,$3E,$C7,$3F,$63,$A7,$B1
	.dc.b	$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1,$7B,$1E,$3F,$63
	.dc.b	$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7,$F2,$1F,$BD,$FC
	.dc.b	$41,$F8,$7E,$23,$F8,$76,$FB,$1B,$F2,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3F,$9C,$7D
	.dc.b	$F7,$F7,$E3,$F0,$FC,$A7,$F0,$EF,$F4,$47,$FC,$E0,$7F,$7F,$75,$EB
	.dc.b	$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF,$AD,$E0,$F5,$BC
	.dc.b	$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79,$BD,$6F,$3F,$AD
	.dc.b	$E8,$F5,$BF,$A8,$7B,$9F,$C3,$CF,$EA,$C7,$F5,$53,$F1,$1C,$7F,$EE
	.dc.b	$0F,$77,$71,$C3,$DD,$E3,$EC,$72,$F6,$39,$FB,$1D,$3D,$8E,$BE,$C7
	.dc.b	$6F,$63,$BF,$B1,$E0,$F6,$3C,$3E,$C7,$8B,$D8,$F1,$FB,$1E,$4F,$63
	.dc.b	$CB,$EC,$79,$BD,$8F,$3F,$B1,$FD,$9F,$EF,$FF,$87,$1F,$87,$3F,$0C
	.dc.b	$3D,$8D,$7B,$1B,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7,$3F,$63,$A7,$B1
	.dc.b	$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1,$7B,$1E,$3F,$63
	.dc.b	$C9,$EC,$79,$7D,$8F,$37,$B1,$FD,$F7,$F0,$0F,$F2,$19,$F8,$63,$F0
	.dc.b	$BE,$41,$FF,$5A,$1F,$85,$3F,$C8,$4E,$7D,$7E,$5E,$BF,$3F,$5F,$A7
	.dc.b	$AF,$D7,$D7,$ED,$EB,$F7,$F5,$FC,$1E,$BF,$87,$D7,$F1,$7A,$FE,$3F
	.dc.b	$5F,$C9,$EB,$F9,$7D,$7F,$37,$AF,$E7,$F5,$FF,$C2,$9F,$80,$FF,$53
	.dc.b	$3F,$C3,$47,$F8,$64,$FC,$28,$F6,$37,$EC,$70,$F6,$38,$FB,$1C,$BD
	.dc.b	$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F,$B1
	.dc.b	$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7F,$8A
	.dc.b	$FD,$9F,$EA,$47,$F0,$B3,$F8,$51,$F8,$5F,$17,$FC,$6E,$BD,$7E,$1E
	.dc.b	$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD,$FD,$7F
	.dc.b	$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F,$CD,$EB
	.dc.b	$F9,$FD,$7F,$F2,$17,$E0,$5F,$E8,$83,$F8,$59,$F8,$63,$FD,$F0,$7F
	.dc.b	$09,$3F,$0A,$3D,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6
	.dc.b	$F5,$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC
	.dc.b	$BE,$BF,$9B,$D7,$FF,$2C,$7E,$07,$FE,$17,$3F,$CB,$C7,$F9,$74,$FE
	.dc.b	$12,$7E,$17,$C5,$FF,$32,$BF,$E5,$C3,$FA,$82,$5E,$A7,$4F,$53,$AF
	.dc.b	$A9,$DB,$D4,$EF,$EA,$78,$3D,$4F,$0F,$A9,$E2,$F5,$3C,$7E,$A7,$93
	.dc.b	$D4,$F2,$FA,$9E,$6F,$53,$CF,$EA,$7A,$3D,$4F,$4F,$A9,$FE,$71,$FC
	.dc.b	$13,$FC,$30,$7F,$51,$3F,$A8,$1F,$87,$E3,$3F,$87,$7F,$CF,$EB,$D3
	.dc.b	$E5,$E9,$F3,$F4,$FA,$7A,$7D,$7D,$3E,$DE,$9F,$7F,$4F,$C1,$E9,$F8
	.dc.b	$7D,$3F,$17,$A7,$E3,$F4,$FC,$9E,$9F,$97,$D3,$F3,$7A,$7E,$7F,$4F
	.dc.b	$D1,$E9,$FF,$A2,$78,$7F,$D9,$E7,$F5,$23,$F8,$68,$F8,$1D,$7C,$0E
	.dc.b	$FD,$7E,$1E,$AF,$1F,$57,$97,$AB,$CF,$D5,$E9,$EA,$F5,$F5,$7B,$7A
	.dc.b	$BD,$FD,$5F,$07,$AB,$E1,$F5,$7C,$5E,$AF,$8F,$D5,$F2,$7A,$BE,$5F
	.dc.b	$57,$CD,$EA,$FF,$A7,$FF,$05,$FF,$2F,$1F,$EA,$53,$FD,$48,$7F,$50
	.dc.b	$3F,$11,$E3,$1F,$F5,$19,$F8,$83,$F0,$EF,$FA,$C1,$7A,$9D,$7D,$4E
	.dc.b	$DE,$A7,$7F,$53,$C1,$EA,$78,$7D,$4F,$17,$A9,$E3,$F5,$3C,$9E,$A7
	.dc.b	$97,$D4,$F3,$7A,$9E,$7F,$53,$D1,$EA,$7A,$7D,$4F,$F5,$C7,$E0,$DF
	.dc.b	$E5,$F3,$FC,$30,$7F,$85,$CF,$C4,$1F,$87,$E2,$3F,$87,$7F,$D8,$EB
	.dc.b	$D4,$E7,$EA,$74,$F5,$3A,$FA,$9D,$BD,$4E,$FE,$A7,$83,$D4,$F0,$FA
	.dc.b	$9E,$2F,$53,$C7,$EA,$79,$3D,$4F,$2F,$A9,$E6,$F5,$3C,$FE,$A7,$A3
	.dc.b	$D4,$FF,$68,$7E,$0E,$F9,$EB,$F1,$58,$F8,$7C,$FA,$FA,$F5,$F7,$EB
	.dc.b	$F0,$F5,$F8,$FA,$FC,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF
	.dc.b	$EB,$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE
	.dc.b	$6F,$5F,$FD,$CD,$EF,$7F,$A9,$4F,$F7,$61,$FE,$EB,$3F,$C2,$E7,$E2
	.dc.b	$0F,$F7,$49,$FC,$38,$FC,$3F,$92,$7F,$0E,$FF,$BE,$17,$AD,$DB,$D6
	.dc.b	$EF,$EB,$78,$3D,$6F,$0F,$AD,$E2,$F5,$BC,$7E,$B7,$93,$D6,$F2,$FA
	.dc.b	$DE,$6F,$5B,$CF,$EB,$7A,$3D,$6F,$4F,$AD,$FE,$FF,$FC,$23,$FD,$4C
	.dc.b	$7F,$97,$CF,$F2,$F1,$FC,$3C,$FC,$41,$F8,$7E,$13,$F8,$74,$63,$D5
	.dc.b	$E9,$EA,$F5,$F5,$7B,$7A,$BD,$FD,$5F,$07,$AB,$E1,$F5,$7C,$5E,$AF
	.dc.b	$8F,$D5,$F2,$7A,$BE,$5F,$57,$CD,$EA,$F9,$FD,$5F,$47,$AB,$E0,$E0
	.dc.b	$7E,$13,$F8,$97,$C6,$FF,$98,$8F,$F0,$E9,$FC,$45,$CF,$A3,$C6,$3F
	.dc.b	$C3,$A7,$F5,$98,$E3,$E7,$E7,$CF,$D7,$9E,$FE,$01,$85,$AE,$F0,$F9
	.dc.b	$7F,$87,$F2,$0E,$5E,$31,$86,$D2,$74,$F2,$9D,$34,$D4,$F0,$9D,$C7
	.dc.b	$C0,$AF,$03,$A3,$84,$6B,$24,$EA,$30,$DA,$E8,$4A,$7A,$EF,$9F,$F8
	.dc.b	$90,$FC,$29,$F3,$1F,$80,$AD,$6D,$F4,$E7,$35,$CA,$D3,$9B,$E1,$1D
	.dc.b	$2B,$77,$DF,$6D,$61,$C7,$2B,$3E,$36,$DA,$AF,$AD,$8F,$5A,$61,$F9
	.dc.b	$4C,$7A,$DA,$B8,$FA,$D2,$B5,$9F,$92,$3A,$FA,$DD,$BD,$69,$5A,$3D
	.dc.b	$8F,$8B,$D6,$F1,$FA,$DE,$4F,$5A,$56,$F8,$3E,$54,$FA,$3D,$6F,$8B
	.dc.b	$B7,$91,$BA,$7C,$4F,$5A,$E1,$F2,$66,$E1,$F4,$FD,$6B,$F3,$9F,$B0
	.dc.b	$C3,$B7,$0A,$73,$86,$BE,$83,$7F,$3F,$6F,$C9,$82,$39,$FC,$A4,$9C
	.dc.b	$FC,$65,$C3,$F0,$11,$F7,$ED,$F1,$9E,$23,$C7,$EE,$59,$EB,$F1,$3E
	.dc.b	$63,$CF,$F0,$D6,$3C,$3E,$F7,$A8,$F5,$FB,$D6,$3C,$BE,$F7,$BB,$DE
	.dc.b	$F7,$B6,$3D,$3E,$F7,$C2,$3E,$1F,$BD,$63,$DB,$EF,$7C,$63,$E3,$FB
	.dc.b	$D6,$3E,$0F,$BD,$F2,$BD,$EF,$96,$D8,$F8,$BE,$F7,$CE,$3E,$7F,$BD
	.dc.b	$63,$E4,$FB,$DF,$94,$E4,$34,$FC,$BF,$37,$E8,$B9,$BE,$49,$FA,$C5
	.dc.b	$F2,$FA,$F3,$64,$FE,$1E,$72,$DF,$90,$E7,$F1,$D2,$30,$6F,$E2,$FF
	.dc.b	$0B,$5F,$E4,$1C,$F9,$7C,$4E,$27,$3F,$05,$AF,$82,$B5,$DD,$E3,$F0
	.dc.b	$52,$3C,$5E,$A5,$CB,$D3,$E3,$7F,$AB,$7E,$0B,$DB,$F0,$52,$7D,$11
	.dc.b	$F0,$3E,$0B,$E0,$FC,$14,$9F,$64,$7C,$4F,$82,$91,$EF,$7E,$3F,$C1
	.dc.b	$7C,$8F,$82,$F5,$9E,$CF,$82,$BA,$FC,$5F,$82,$F7,$FC,$17,$C0,$6C
	.dc.b	$7E,$BF,$E7,$3F,$C1,$4D,$DF,$F5,$31,$8C,$FD,$6F,$F1,$E9,$E1,$F1
	.dc.b	$1F,$C3,$A3,$FD,$C7,$EA,$FF,$14,$6C,$FF,$1F,$7F,$A5,$57,$2F,$84
	.dc.b	$E0,$47,$C5,$F0,$FC,$FE,$1E,$47,$95,$ED,$F0,$7D,$FE,$0B,$CA,$79
	.dc.b	$BE,$0E,$E5,$EB,$F8,$2F,$49,$EA,$F8,$3B,$1E,$FF,$83,$F6,$9E,$EF
	.dc.b	$83,$B1,$F0,$FE,$0F,$E0,$9F,$0B,$E0,$EC,$7C,$7F,$82,$F8,$BF,$07
	.dc.b	$F1,$9B,$1F,$2F,$E0,$BE,$49,$F2,$BE,$0E,$C7,$CF,$F8,$2F,$E2,$BF
	.dc.b	$39,$FC,$4F,$8F,$E6,$C6,$1C,$FC,$1E,$BE,$0E,$DF,$F2,$5B,$FE,$24
	.dc.b	$DF,$83,$E6,$74,$F3,$6C,$6F,$E0,$FB,$FC,$1F,$81,$B1,$CF,$E0,$FC
	.dc.b	$7F,$07,$E4,$6C,$77,$F8,$3F,$39,$E8,$F8,$3B,$1E,$3F,$82,$F5,$FC
	.dc.b	$1F,$B1,$B1,$E7,$F8,$3F,$79,$F0,$3E,$0E,$C7,$AF,$E0,$FE,$1F,$C1
	.dc.b	$FC,$46,$C7,$BF,$E0,$BE,$39,$F2,$3E,$0E,$C7,$C3,$F8,$3F,$97,$F0
	.dc.b	$7F,$31,$B1,$F1,$FE,$0B,$E7,$9F,$D1,$3E,$6B,$FB,$9C,$64,$FE,$A0
	.dc.b	$FF,$0E,$6F,$C1,$EF,$E0,$F8,$3C,$7E,$0E,$4F,$0F,$83,$BD,$DF,$C0
	.dc.b	$30,$FE,$18,$DF,$82,$F0,$7C,$1F,$85,$F1,$7C,$1F,$F7,$6A,$34,$C7
	.dc.b	$C1,$E7,$E0,$F5,$F0,$73,$1F,$3C,$71,$F8,$29,$1E,$A7,$A7,$C1,$F5
	.dc.b	$F8,$29,$3E,$E7,$C1,$F0,$7E,$1F,$83,$93,$F0,$9F,$27,$C1,$F9,$7E
	.dc.b	$0E,$4F,$C6,$7D,$1F,$07,$E9,$F8,$2F,$57,$C1,$CA,$F9,$71,$FE,$12
	.dc.b	$F7,$FF,$C9,$D1,$FC,$91,$BF,$03,$FE,$1A,$6F,$3F,$F8,$95,$FF,$15
	.dc.b	$CF,$F8,$0D,$FE,$16,$FF,$8B,$9B,$F0,$66,$9F,$C3,$AB,$A3,$FE,$02
	.dc.b	$1F,$03,$5C,$39,$F8,$1D,$7C,$0D,$C7,$C0,$F1,$F8,$1E,$5F,$03,$67
	.dc.b	$C8,$F5,$F8,$19,$1E,$7F,$81,$BA,$FB,$FE,$07,$C5,$F0,$3E,$36,$D7
	.dc.b	$B7,$E0,$7C,$C7,$9F,$E0,$6C,$7C,$1F,$80,$F5,$7C,$0F,$AD,$B1,$F1
	.dc.b	$7E,$07,$DC,$7B,$FE,$06,$C7,$C9,$F8,$1F,$F1,$1F,$CD,$7F,$0F,$E6
	.dc.b	$FA,$11,$8D,$B3,$F0,$5F,$E4,$8A,$34,$C7,$C1,$67,$E0,$65,$7F,$8D
	.dc.b	$E9,$1C,$7E,$07,$97,$C0,$4A,$D3,$D7,$E0,$7B,$7C,$04,$9E,$4F,$87
	.dc.b	$E0,$7C,$5F,$03,$27,$B3,$E5,$F8,$1F,$37,$C0,$49,$F1,$3E,$9F,$81
	.dc.b	$F5,$7C,$0F,$C0,$3E,$0F,$C0,$DE,$4F,$47,$C0,$7C,$43,$E2,$FC,$0D
	.dc.b	$8F,$67,$C0,$FC,$83,$E4,$FC,$0D,$8F,$81,$F0,$1F,$33,$E0,$7E,$6B
	.dc.b	$63,$E2,$7C,$0F,$F9,$4B,$E6,$BF,$87,$EF,$FE,$04,$E8,$FF,$10,$D7
	.dc.b	$F9,$7E,$73,$D9,$FE,$7C,$7F,$89,$57,$F8,$29,$FF,$50,$4E,$FE,$07
	.dc.b	$A7,$C1,$75,$F8,$2F,$F3,$FD,$3E,$0A,$98,$8C,$FC,$0C,$C7,$C0,$F0
	.dc.b	$F8,$2E,$3F,$03,$27,$C8,$F4,$F8,$1E,$BF,$01,$DB,$E0,$65,$7A,$63
	.dc.b	$C3,$F0,$3E,$2F,$81,$93,$ED,$8F,$2F,$C0,$F9,$BE,$06,$4F,$C1,$8F
	.dc.b	$4F,$C0,$FC,$73,$E4,$7C,$0D,$E4,$F8,$DF,$03,$F2,$CF,$99,$F0,$36
	.dc.b	$3E,$57,$C0,$7C,$FF,$81,$FA,$0D,$8F,$F4,$67,$CD,$7F,$0E,$CF,$FD
	.dc.b	$6E,$FF,$0E,$7F,$C6,$ED,$EA,$FF,$56,$BB,$F8,$23,$FC,$40,$FE,$1D
	.dc.b	$5F,$EB,$B6,$D7,$4F,$86,$FF,$58,$C7,$F1,$58,$F0,$FC,$16,$0C,$FC
	.dc.b	$1E,$BE,$0F,$FD,$2F,$5F,$EC,$0C,$3C,$F1,$8F,$83,$CF,$C1,$EB,$E0
	.dc.b	$ED,$7B,$1E,$3F,$07,$23,$DF,$F0,$77,$4F,$6F,$C1,$76,$F8,$3E,$ED
	.dc.b	$AF,$8B,$F0,$7E,$23,$C7,$F0,$76,$3E,$4F,$C1,$79,$BE,$0F,$CE,$D8
	.dc.b	$F9,$BF,$07,$F0,$09,$97,$D0,$F8,$3F,$87,$F0,$72,$3E,$9B,$FE,$D6
	.dc.b	$F2,$91,$7E,$0E,$3F,$EC,$38,$FE,$26,$DF,$83,$E1,$F0,$72,$34,$F3
	.dc.b	$F8,$3E,$9F,$07,$27,$93,$DF,$E0,$FC,$1F,$07,$E1,$F8,$39,$5D,$E3
	.dc.b	$C9,$F0,$72,$3C,$4F,$9F,$E0,$FD,$1F,$07,$27,$CC,$FA,$FE,$0F,$D9
	.dc.b	$F0,$7E,$DF,$83,$95,$EB,$8F,$81,$F0,$7F,$07,$E0,$E4,$FB,$E3,$E2
	.dc.b	$7C,$1F,$C5,$F8,$39,$3F,$0E,$3E,$47,$C1,$7C,$9F,$83,$93,$F1,$E3
	.dc.b	$E6,$7C,$1F,$FB,$8B,$9F,$F8,$7C,$FF,$A5,$3E,$6B,$F8,$7F,$6F,$D0
	.dc.b	$F3,$BF,$84,$5F,$FB,$A2,$7F,$88,$F9,$EA,$D4,$61,$BF,$05,$BF,$83
	.dc.b	$E1,$F0,$5C,$73,$B5,$FE,$FB,$BF,$E2,$8D,$F8,$2F,$BF,$A7,$8F,$E0
	.dc.b	$AE,$7D,$E1,$AB,$05,$6B,$0F,$1E,$16,$F7,$5F,$E2,$A7,$C1,$F0,$59
	.dc.b	$C3,$93,$65,$F0,$57,$E0,$71,$F0,$53,$0F,$44,$6F,$E0,$78,$7C,$14
	.dc.b	$9F,$64,$73,$F8,$19,$1E,$F7,$B7,$C1,$77,$F8,$29,$3F,$0D,$F1,$7C
	.dc.b	$17,$8F,$E0,$BE,$87,$C1,$7D,$16,$F1,$7C,$9F,$81,$FA,$87,$D5,$F8
	.dc.b	$2B,$1F,$A8,$E6,$F9,$1C,$8F,$9A,$FF,$5C,$46,$1B,$3F,$7D,$FF,$5D
	.dc.b	$69,$FE,$D6,$72,$FD,$07,$FC,$5A,$4F,$83,$F9,$4F,$F1,$EB,$ED,$F0
	.dc.b	$58,$25,$78,$EF,$B7,$C1,$F0,$F8,$2E,$3F,$05,$2B,$47,$78,$EB,$F0
	.dc.b	$52,$33,$8F,$AE,$3C,$3F,$05,$E2,$F8,$29,$5A,$D1,$1E,$6F,$82,$F3
	.dc.b	$FC,$0C,$AD,$3E,$AF,$82,$F5,$FC,$17,$E3,$F8,$BE,$33,$78,$3E,$37
	.dc.b	$C1,$5C,$3E,$24,$DD,$BE,$77,$C0,$DD,$4D,$F8,$2C,$39,$F8,$29,$97
	.dc.b	$9F,$E0,$AE,$5F,$47,$E0,$B9,$1D,$7E,$6B,$FC,$6D,$5F,$5E,$31,$F0
	.dc.b	$79,$F8,$3B,$3F,$1A,$BF,$D8,$76,$E0,$FC,$1C,$7F,$8D,$27,$5F,$05
	.dc.b	$D7,$E0,$FB,$3D,$FE,$0F,$F4,$D4,$69,$8F,$83,$CF,$C1,$6B,$E0,$E6
	.dc.b	$7E,$18,$E3,$F0,$72,$3C,$8F,$4F,$83,$EB,$F0,$72,$7D,$0F,$83,$E0
	.dc.b	$FC,$3F,$05,$E2,$F8,$39,$5E,$D8,$F2,$FC,$17,$9B,$E0,$E4,$FC,$18
	.dc.b	$F4,$FC,$17,$AB,$E0,$E4,$FC,$58,$F6,$FC,$17,$BB,$E0,$E4,$FC,$98
	.dc.b	$F8,$3F,$07,$FC,$1F,$8F,$F8,$94,$FC,$BE,$7F,$E4,$D1,$8E,$8F,$9B
	.dc.b	$5B,$F8,$2F,$C2,$4F,$F2,$5B,$FE,$32,$FF,$8E,$1E,$46,$23,$7F,$05
	.dc.b	$C3,$E0,$AE,$3F,$A0,$BF,$E4,$F7,$C3,$C4,$FA,$9F,$03,$D8,$EF,$F0
	.dc.b	$56,$F8,$7C,$17,$A7,$E0,$66,$18,$3A,$7C,$07,$B4,$8F,$7F,$81,$F8
	.dc.b	$1F,$01,$23,$C4,$FC,$3F,$81,$F8,$9F,$03,$EE,$F8,$1F,$7B,$73,$F4
	.dc.b	$7C,$0F,$C2,$F8,$1F,$86,$D8,$F6,$7C,$07,$C6,$F8,$1F,$8E,$D8,$F8
	.dc.b	$1F,$03,$F2,$BE,$07,$E5,$B6,$3F,$81,$FC,$D7,$F1,$6E,$53,$4F,$E3
	.dc.b	$7D,$7F,$C8,$9C,$7C,$17,$ED,$D7,$C8,$D8,$C3,$9F,$83,$D7,$C1,$DC
	.dc.b	$7B,$BC,$7E,$0A,$47,$8B,$E0,$EF,$37,$D5,$C9,$F0,$7D,$CF,$07,$C1
	.dc.b	$DB,$F4,$FC,$1F,$8C,$F2,$7C,$1D,$8F,$6F,$C1,$F9,$CF,$47,$C1,$D8
	.dc.b	$F8,$3F,$05,$EB,$F8,$3F,$63,$63,$E2,$FC,$0F,$F5,$FD,$1F,$05,$4C
	.dc.b	$46,$7E,$0B,$5F,$05,$38,$3E,$6C,$71,$F8,$2F,$ED,$1F,$35,$FD,$DA
	.dc.b	$7F,$24,$DE,$9F,$9D,$3A,$F8,$3F,$E4,$93,$FE,$C8,$D3,$F9,$CB,$FC
	.dc.b	$52,$BF,$AA,$B9,$BE,$7F,$05,$C3,$E0,$B8,$B6,$F8,$F9,$DD,$3E,$07
	.dc.b	$AB,$63,$AF,$C0,$79,$89,$87,$83,$E0,$7D,$3F,$03,$23,$C6,$FB,$3E
	.dc.b	$07,$DB,$F0,$32,$7C,$EF,$C0,$F8,$1F,$83,$F0,$1E,$C3,$DB,$F0,$37
	.dc.b	$3F,$6F,$C0,$7C,$03,$E0,$FC,$0D,$8F,$83,$F0,$1F,$10,$F8,$BF,$03
	.dc.b	$63,$E2,$FC,$07,$CE,$F8,$19,$AE,$0F,$93,$F0,$3F,$48,$8F,$F6,$4F
	.dc.b	$9A,$FE,$25,$7F,$E3,$95,$F0,$5F,$E1,$F6,$CE,$E3,$1F,$0F,$9F,$83
	.dc.b	$93,$CD,$E1,$F0,$7C,$7E,$0F,$97,$C1,$CA,$F0,$47,$5F,$83,$91,$E3
	.dc.b	$7C,$1F,$07,$E1,$F8,$3F,$EC,$DA,$1B,$F0,$78,$73,$F0,$7A,$F8,$3B
	.dc.b	$C1,$ED,$78,$FC,$17,$2F,$83,$B3,$F0,$5E,$BF,$05,$EF,$F8,$39,$B7
	.dc.b	$C5,$8F,$85,$F0,$52,$3E,$43,$F1,$7E,$0F,$E3,$7C,$17,$C7,$F8,$39
	.dc.b	$5F,$36,$3F,$C9,$1E,$FF,$F0,$F3,$F8,$6E,$7F,$E6,$77,$1F,$07,$FE
	.dc.b	$03,$5F,$07,$7E,$0F,$0D,$AF,$F0,$DC,$BE,$0F,$83,$C5,$B1,$AF,$83
	.dc.b	$F0,$93,$2E,$3F,$07,$E4,$F8,$3F,$2F,$C1,$C9,$EB,$1E,$8F,$83,$91
	.dc.b	$E0,$7D,$7F,$07,$EC,$F8,$3F,$6F,$C1,$CA,$F2,$C7,$C0,$F8,$29,$1E
	.dc.b	$87,$E1,$FC,$1F,$C4,$F8,$1F,$8B,$F0,$52,$BD,$B1,$F2,$3E,$07,$E2
	.dc.b	$1F,$17,$E0,$AF,$0F,$C2,$F8,$2F,$F1,$57,$CD,$7F,$87,$B0,$FF,$3F
	.dc.b	$F2,$A7,$31,$AF,$83,$DF,$C3,$5C,$3F,$C8,$B8,$7F,$0D,$7F,$D3,$4F
	.dc.b	$F8,$7E,$BF,$B4,$2F,$82,$EF,$F0,$5E,$07,$C3,$F0,$3E,$2F,$82,$B8
	.dc.b	$74,$7C,$BF,$05,$E6,$F8,$2B,$3E,$07,$D3,$F0,$5F,$E8,$DC,$9F,$05
	.dc.b	$4C,$46,$7E,$0B,$5F,$05,$36,$F4,$C7,$1F,$82,$E5,$F0,$52,$7D,$B1
	.dc.b	$D7,$E0,$BB,$7C,$14,$9F,$83,$1E,$1F,$82,$F1,$7C,$14,$9F,$8B,$1E
	.dc.b	$5F,$82,$F3,$7C,$14,$9F,$93,$1E,$9F,$82,$FA,$67,$F9,$77,$E6,$BF
	.dc.b	$D8,$B3,$FE,$B2,$E1,$FF,$64,$FC,$2F,$FA,$3A,$7F,$C3,$DE,$79,$FF
	.dc.b	$0F,$46,$7E,$13,$97,$C2,$73,$F8,$4E,$9F,$09,$7A,$7F,$CF,$B8,$7F
	.dc.b	$87,$B0,$E9,$F0,$7E,$1F,$84,$F1,$36,$B7,$F0,$9E,$53,$CD,$F0,$96
	.dc.b	$39,$FC,$17,$A7,$E0,$FD,$4D,$8E,$FF,$07,$FE,$C3,$CC,$D3,$1F,$07
	.dc.b	$9F,$83,$D7,$C1,$CE,$5F,$3C,$71,$F8,$2E,$5F,$07,$27,$D7,$1D,$7E
	.dc.b	$0E,$47,$B9,$F0,$7C,$1F,$87,$E0,$BC,$5F,$07,$2B,$E1,$C7,$FB,$5F
	.dc.b	$E6,$9F,$2F,$7B,$25,$7F,$A9,$27,$2D,$F8,$4E,$1F,$0B,$23,$06,$FE
	.dc.b	$17,$A7,$AF,$23,$07,$3F,$85,$F0,$7C,$2C,$8C,$1D,$FE,$13,$FD,$41
	.dc.b	$97,$F8,$7B,$A3,$C5,$F0,$B8,$9C,$FC,$2C,$9F,$37,$C2,$DC,$3F,$CB
	.dc.b	$58,$3E,$13,$99,$D3,$E1,$6D,$FB,$7E,$17,$BF,$C2,$F8,$1B,$1F,$07
	.dc.b	$E1,$3C,$7F,$0B,$E4,$6C,$7C,$5F,$85,$F3,$FC,$2F,$A1,$B1,$F2,$7E
	.dc.b	$17,$E7,$13,$5F,$F6,$7F,$93,$FE,$1D,$94,$3A,$52,$F8,$5C,$CF,$F9
	.dc.b	$D6,$1F,$0B,$8F,$37,$FD,$E7,$18,$9E,$B5,$9C,$3A,$E1,$D3,$E1,$71
	.dc.b	$5D,$EF,$F8,$C3,$E4,$F8,$5C,$D6,$E3,$CF,$F0,$BF,$ED,$3A,$7C,$2D
	.dc.b	$8C,$39,$F8,$59,$97,$BB,$E1,$6E,$1E,$9F,$84,$E5,$F0,$BC,$DB,$5F
	.dc.b	$13,$E1,$3B,$7C,$2F,$76,$C7,$C8,$F8,$4F,$17,$C2,$F8,$DB,$1F,$33
	.dc.b	$E1,$3C,$C7,$9F,$E1,$6C,$7D,$0F,$84,$F5,$7C,$2F,$AD,$B1,$F5,$3E
	.dc.b	$13,$DD,$F0,$BE,$F6,$C7,$D8,$F8,$5F,$84,$7F,$BD,$7E,$6B,$F8,$CF
	.dc.b	$3F,$D8,$8C,$7C,$2E,$7E,$16,$DF,$DE,$ED,$3F,$53,$0F,$E3,$6D,$F8
	.dc.b	$5E,$9F,$0B,$D5,$B1,$C4,$E1,$D9,$E1,$F8,$5F,$13,$63,$5D,$3F,$8B
	.dc.b	$C1,$F0,$BE,$7F,$85,$F4,$36,$B7,$BA,$BF,$C5,$51,$85,$F0,$BE,$FF
	.dc.b	$85,$F8,$0F,$C1,$F8,$49,$5B,$F8,$4B,$9F,$FB,$2F,$07,$C1,$FE,$31
	.dc.b	$7C,$25,$31,$19,$F8,$3D,$7C,$24,$BF,$14,$71,$F8,$4E,$5F,$09,$27
	.dc.b	$CD,$1D,$7E,$0F,$EA,$1F,$57,$E1,$2E,$BE,$EF,$9A,$FF,$5E,$13,$8B
	.dc.b	$E9,$E1,$FE,$E1,$8C,$FC,$26,$3A,$B3,$49,$FF,$5E,$DF,$3A,$D3,$FA
	.dc.b	$E8,$CD,$F3,$F8,$3F,$D5,$C7,$F8,$CE,$3C,$51,$8F,$83,$CF,$C1,$EB
	.dc.b	$E0,$E5,$79,$E3,$8F,$C1,$48,$F5,$3D,$3E,$0F,$AF,$C1,$F8,$A7,$F8
	.dc.b	$56,$48,$C6,$BE,$9F,$83,$B3,$9D,$7E,$27,$C1,$60,$CF,$C1,$58,$F9
	.dc.b	$1F,$03,$C0,$E3,$F0,$56,$3E,$67,$C0,$F4,$F8,$2E,$AD,$8F,$A1,$F0
	.dc.b	$3E,$0F,$82,$F0,$B6,$3E,$A7,$C1,$79,$BE,$6B,$FC,$6D,$BF,$CF,$8C
	.dc.b	$70,$E7,$E0,$BF,$8A,$D7,$F2,$58,$FC,$CB,$57,$57,$D8,$8F,$E4,$91
	.dc.b	$8C,$3F,$91,$C3,$E0,$B9,$92,$79,$FC,$17,$6F,$82,$91,$D9,$F0,$FC
	.dc.b	$17,$8B,$E0,$BD,$1F,$05,$E9,$6E,$5E,$4F,$81,$F6,$7C,$17,$B5,$B1
	.dc.b	$E8,$F8,$1F,$80,$7C,$1F,$82,$B1,$EC,$F8,$1F,$79,$34,$F7,$FC,$17
	.dc.b	$C2,$F8,$2F,$87,$F0,$52,$7E,$1C,$7C,$6F,$81,$91,$F1,$9F,$93,$F0
	.dc.b	$5F,$2B,$E0,$64,$FC,$A7,$E6,$FC,$17,$CE,$F8,$2F,$E8,$7C,$7F,$EC
	.dc.b	$2A,$FE,$AC,$DE,$B7,$42,$B7,$F0,$5C,$3E,$0A,$61,$FC,$3B,$0F,$F6
	.dc.b	$3B,$F4,$CE,$7F,$03,$DB,$E0,$A4,$6E,$3C,$3F,$03,$E2,$F8,$29,$3C
	.dc.b	$E3,$CB,$F0,$32,$3B,$3E,$8F,$82,$F4,$FC,$17,$AB,$E0,$A5,$78,$E3
	.dc.b	$DB,$F0,$52,$3C,$CF,$C0,$F8,$2F,$83,$F0,$12,$7D,$4F,$C4,$F8,$1F
	.dc.b	$8B,$F0,$1F,$1B,$E0,$65,$7B,$E3,$E4,$FC,$04,$8F,$84,$FC,$CF,$81
	.dc.b	$F9,$BF,$03,$F3,$BE,$06,$57,$C7,$8F,$EE,$7F,$35,$FE,$32,$8F,$91
	.dc.b	$D6,$8C,$79,$DF,$C5,$9B,$E2,$FF,$80,$1F,$F0,$65,$7F,$3E,$7F,$D9
	.dc.b	$2D,$F8,$6E,$A7,$6F,$83,$EF,$F0,$7E,$0F,$83,$95,$C6,$3C,$7F,$05
	.dc.b	$23,$A3,$E6,$F8,$3F,$3F,$C1,$FA,$3E,$0E,$57,$86,$3D,$7F,$07,$23
	.dc.b	$C8,$FB,$BE,$0F,$DF,$F0,$5F,$C6,$36,$7C,$15,$8C,$39,$F8,$29,$C3
	.dc.b	$ED,$F8,$2B,$97,$C6,$C1,$F0,$3C,$CE,$9F,$05,$6F,$E2,$7C,$0F,$73
	.dc.b	$C1,$F0,$56,$3E,$47,$C1,$78,$CF,$E5,$7F,$31,$FC,$5B,$1F,$F6,$54
	.dc.b	$FF,$6A,$8F,$F1,$C3,$7E,$1B,$7F,$09,$C1,$E3,$F0,$9C,$BE,$12,$D7
	.dc.b	$17,$AF,$C2,$48,$E9,$E9,$5E,$0F,$C0,$E4,$F8,$3F,$1F,$C2,$79,$1B
	.dc.b	$7E,$3F,$83,$F3,$9E,$8F,$84,$B1,$E7,$F8,$4F,$5F,$C2,$7B,$1B,$1E
	.dc.b	$BF,$84,$F7,$9F,$03,$E1,$2C,$7B,$FE,$0F,$E1,$9F,$13,$E1,$2C,$7C
	.dc.b	$3F,$84,$F8,$FF,$09,$F2,$1B,$1F,$1F,$E1,$3E,$59,$F3,$3E,$12,$C7
	.dc.b	$CB,$F8,$3F,$EB,$3F,$85,$FF,$8B,$B8,$FF,$D8,$B3,$FE,$5C,$8F,$F1
	.dc.b	$89,$FE,$C6,$27,$C1,$F0,$F8,$4E,$3F,$07,$2B,$FC,$3F,$5F,$EC,$37
	.dc.b	$94,$76,$F8,$2E,$FF,$07,$67,$6F,$8B,$E0,$FC,$7F,$07,$67,$9B,$E6
	.dc.b	$F8,$29,$1D,$BE,$0E,$F6,$FE,$E3,$3F,$F6,$1C,$F8,$BE,$0B,$DB,$F0
	.dc.b	$7E,$E7,$DF,$F0,$7F,$03,$E0,$ED,$7A,$1F,$87,$F0,$52,$3D,$7F,$07
	.dc.b	$72,$ED,$F0,$7F,$20,$F9,$3F,$07,$6B,$E0,$FC,$17,$CC,$3E,$6F,$C1
	.dc.b	$D8,$FF,$21,$FB,$BF,$EC,$46,$FC,$1F,$F8,$2A,$31,$8F,$F1,$27,$FD
	.dc.b	$31,$3C,$3E,$07,$46,$FE,$0A,$C7,$4F,$81,$E4,$73,$F8,$2B,$1E,$0B
	.dc.b	$EC,$77,$F8,$2B,$1E,$4F,$81,$F1,$1E,$3F,$82,$B1,$E8,$F8,$2F,$37
	.dc.b	$C1,$79,$DB,$1E,$CF,$82,$F5,$7C,$17,$AD,$B1,$F0,$3E,$0B,$DC,$7B
	.dc.b	$FE,$0A,$C7,$C4,$F8,$2F,$85,$F0,$5F,$0D,$B1,$F2,$3E,$0B,$E3,$7C
	.dc.b	$17,$C7,$6C,$7C,$CF,$82,$F9,$47,$F8,$D3,$F0,$CF,$F0,$FF,$17,$F0
	.dc.b	$AE,$7F,$E7,$B1,$FC,$29,$BF,$03,$BF,$82,$E0,$F1,$F8,$2E,$5F,$05
	.dc.b	$72,$FF,$3D,$E0,$8E,$DF,$05,$27,$3F,$05,$72,$EF,$F0,$3E,$3F,$82
	.dc.b	$F2,$36,$B9,$FC,$17,$9F,$E0,$BD,$0D,$8E,$FF,$03,$EB,$F8,$2F,$63
	.dc.b	$63,$C7,$F0,$5E,$FF,$82,$F8,$0D,$8F,$3F,$C1,$7C,$3F,$82,$F8,$8D
	.dc.b	$8F,$5F,$C0,$FC,$73,$E4,$7C,$15,$8F,$7F,$C1,$7C,$B3,$E6,$7C,$15
	.dc.b	$8F,$87,$F0,$5F,$E6,$2F,$C3,$7F,$D6,$FC,$FF,$C2,$B9,$3F,$DE,$11
	.dc.b	$FE,$BA,$1F,$03,$BF,$82,$E0,$DC,$3F,$D4,$F2,$F8,$2E,$87,$5F,$86
	.dc.b	$B3,$AF,$83,$F0,$1E,$1F,$83,$B1,$CB,$E0,$FC,$87,$97,$E0,$EC,$76
	.dc.b	$F8,$3F,$47,$C1,$FA,$5B,$1E,$2F,$82,$F6,$7C,$17,$B5,$B1,$E6,$F8
	.dc.b	$2F,$81,$F0,$5F,$05,$B1,$EA,$F8,$2F,$88,$7C,$5F,$82,$B1,$EE,$F8
	.dc.b	$2F,$91,$F0,$5F,$25,$B1,$F0,$BE,$0B,$E6,$7C,$17,$CD,$6C,$7F,$B3
	.dc.b	$3F,$0E,$FE,$0F,$F0,$FF,$EA,$A7,$F8,$3F,$57,$F8,$69,$CC,$A1,$F0
	.dc.b	$89,$DD,$71,$F8,$4F,$F4,$DC,$7F,$AB,$A3,$17,$DF,$E1,$33,$F0,$92
	.dc.b	$3C,$5F,$09,$72,$F0,$7A,$DC,$8E,$7F,$09,$6B,$D1,$F0,$9D,$BE,$13
	.dc.b	$BB,$63,$D9,$F0,$7E,$7C,$3D,$4D,$D3,$DC,$D9,$F6,$6F,$EB,$F8,$4C
	.dc.b	$5F,$BB,$77,$C1,$E2,$BE,$03,$63,$33,$F2,$FE,$13,$1F,$09,$9F,$84
	.dc.b	$93,$F3,$E3,$87,$C1,$C8,$FA,$4F,$3F,$84,$E9,$F0,$7D,$7E,$12,$57
	.dc.b	$D7,$8F,$07,$C2,$48,$FB,$4F,$97,$C8,$F8,$11,$FE,$52,$7F,$D6,$1C
	.dc.b	$5F,$EE,$66,$F0,$FC,$06,$CF,$DC,$7F,$D6,$73,$BF,$82,$E8,$49,$FC
	.dc.b	$14,$73,$F8,$2F,$07,$C1,$49,$CE,$1D,$63,$C9,$F0,$5E,$5F,$81,$95
	.dc.b	$A3,$C0,$FA,$7E,$0B,$D5,$F0,$32,$75,$CE,$D5,$3E,$FF,$82,$F8,$1F
	.dc.b	$03,$2F,$74,$8F,$89,$F0,$5F,$17,$E0,$65,$69,$F9,$1F,$05,$F2,$7E
	.dc.b	$07,$E5,$7C,$14,$AE,$71,$F3,$7E,$0A,$47,$67,$E8,$7C,$17,$D1,$F8
	.dc.b	$29,$3E,$27,$EA,$7C,$17,$D5,$F8,$2F,$C5,$7E,$1F,$FE,$B5,$3F,$D6
	.dc.b	$DB,$FF,$AD,$6F,$DF,$1F,$EB,$2E,$97,$47,$E5,$A1,$F0,$5C,$8E,$7F
	.dc.b	$05,$D3,$E0,$66,$1C,$9E,$FF,$03,$E0,$F8,$1F,$0E,$72,$BB,$C7,$93
	.dc.b	$E0,$64,$78,$9F,$3F,$C0,$FA,$3E,$02,$4F,$99,$F5,$FC,$0F,$B3,$E0
	.dc.b	$24,$FA,$9F,$7F,$C0,$FC,$0F,$80,$93,$EE,$7E,$1F,$C0,$FC,$4F,$81
	.dc.b	$F8,$BF,$03,$2B,$E1,$C7,$C8,$F8,$09,$1F,$19,$F9,$7F,$03,$F3,$3E
	.dc.b	$02,$4F,$CA,$7E,$7F,$C0,$FE,$E7,$F1,$0F,$F5,$8C,$FF,$1A,$6F,$90
	.dc.b	$69,$FE,$B4,$8F,$F5,$DB,$FE,$B4,$6F,$C1,$73,$3A,$7C,$17,$5F,$82
	.dc.b	$93,$F9,$8A,$47,$87,$E0,$BC,$5F,$05,$2B,$4F,$97,$E0,$BC,$DF,$03
	.dc.b	$27,$93,$E9,$F8,$2F,$57,$C1,$49,$EC,$FB,$7E,$0B,$DD,$F0,$52,$7C
	.dc.b	$4F,$C1,$F8,$1F,$85,$F0,$12,$7C,$CF,$C5,$F8,$1F,$8D,$F0,$3F,$1F
	.dc.b	$E0,$65,$7A,$E3,$E5,$7C,$04,$8F,$73,$F3,$7E,$07,$E7,$7C,$0F,$F2
	.dc.b	$7F,$C4,$7F,$89,$F9,$FF,$9A,$BF,$E9,$31,$8F,$3F,$F9,$8F,$3A,$F8
	.dc.b	$7C,$3F,$1F,$87,$98,$7F,$02,$C3,$F8,$BF,$61,$7C,$27,$7F,$84,$95
	.dc.b	$B8,$F1,$7C,$24,$8E,$4F,$97,$E1,$3C,$DF,$09,$27,$B3,$E9,$F8,$4F
	.dc.b	$57,$C2,$49,$F1,$3E,$DF,$84,$F7,$7C,$24,$9F,$33,$F0,$7E,$13,$E1
	.dc.b	$7C,$24,$9F,$53,$F1,$7E,$13,$E3,$7C,$1F,$C7,$F8,$49,$5E,$F8,$F9
	.dc.b	$5F,$09,$23,$E1,$3F,$37,$E1,$3E,$77,$C2,$7C,$FF,$84,$95,$F1,$E3
	.dc.b	$E9,$7C,$27,$F5,$9E,$1F,$A5,$2E,$DF,$A9,$D1,$FE,$D7,$ED,$DC,$FF
	.dc.b	$30,$AF,$E2,$51,$FD,$99,$B1,$FE,$02,$6C,$FF,$B7,$FE,$0F,$B7,$C1
	.dc.b	$4A,$E4,$F8,$7E,$0B,$C5,$F0,$32,$7B,$3E,$5F,$82,$F3,$7C,$17,$9F
	.dc.b	$E0,$A5,$78,$E3,$D5,$F0,$5E,$BF,$82,$93,$E7,$8F,$77,$C1,$7B,$FE
	.dc.b	$0A,$4F,$AE,$3E,$17,$C1,$48,$F7,$3F,$17,$E0,$BE,$37,$C0,$C9,$F8
	.dc.b	$4F,$C9,$F8,$2F,$95,$F0,$5F,$2F,$E0,$A5,$7C,$78,$F9,$DF,$05,$23
	.dc.b	$E5,$3F,$47,$E0,$BF,$C1,$5F,$89,$7F,$10,$3F,$87,$75,$7D,$26,$F5
	.dc.b	$7F,$85,$1D,$7C,$1F,$0F,$83,$98,$FF,$64,$9F,$E2,$97,$C7,$E0,$FA
	.dc.b	$FC,$1C,$AD,$47,$83,$E0,$FC,$3F,$07,$27,$94,$79,$3E,$0F,$F6,$D8
	.dc.b	$78,$FE,$0E,$EB,$E3,$9B,$7F,$E4,$2C,$D5,$8F,$05,$0F,$83,$C3,$70
	.dc.b	$F4,$7C,$1E,$27,$3F,$07,$27,$D9,$F0,$77,$2F,$CB,$47,$C3,$F8,$3E
	.dc.b	$7F,$07,$D1,$B1,$F0,$FE,$0B,$BF,$C1,$F8,$1B,$1F,$1F,$E0,$FC,$67
	.dc.b	$93,$E0,$EC,$7C,$BF,$82,$F3,$9E,$BF,$C4,$FF,$CC,$11,$93,$FC,$26
	.dc.b	$FF,$B0,$79,$3F,$52,$BE,$0B,$87,$C1,$F1,$79,$7C,$0F,$3F,$82,$B8
	.dc.b	$72,$7B,$7C,$17,$7F,$82,$B3,$D9,$F1,$7A,$9E,$3F,$82,$B3,$E2,$7C
	.dc.b	$DF,$05,$E7,$F8,$2B,$3E,$67,$D5,$F0,$5E,$BF,$82,$B3,$EA,$7D,$DF
	.dc.b	$05,$23,$DB,$F0,$57,$AB,$E8,$9F,$E8,$0E,$27,$C0,$FC,$53,$E3,$7C
	.dc.b	$15,$BF,$8B,$F0,$5F,$24,$F9,$5F,$05,$63,$E4,$FC,$0F,$CD,$F8,$2F
	.dc.b	$9C,$D8,$F9,$BF,$05,$F8,$BF,$C5,$3F,$C4,$31,$83,$FC,$86,$E3,$4F
	.dc.b	$2F,$90,$8F,$F6,$2C,$EB,$E0,$F9,$7C,$37,$37,$A7,$C2,$49,$E7,$F0
	.dc.b	$97,$1E,$78,$3E,$13,$C5,$F0,$9E,$36,$DF,$8B,$E0,$FC,$C7,$9F,$E1
	.dc.b	$2C,$79,$BE,$0F,$D4,$7A,$FE,$12,$C7,$AB,$E0,$FD,$DF,$09,$EF,$6C
	.dc.b	$7B,$BE,$13,$E1,$7C,$27,$C3,$6C,$7C,$2F,$84,$F8,$DF,$09,$F1,$DB
	.dc.b	$1F,$1B,$E0,$FE,$57,$C2,$7C,$B6,$C7,$CA,$F8,$4F,$9D,$F0,$9F,$3D
	.dc.b	$B1,$FE,$70,$F8,$0F,$F6,$0F,$2F,$FB,$01,$FF,$5F,$C7,$FA,$A1,$CF
	.dc.b	$51,$CC,$EF,$E1,$3F,$82,$41,$BF,$09,$87,$3F,$07,$AF,$84,$B8,$F8
	.dc.b	$5E,$3F,$09,$CB,$0B,$3E,$57,$AF,$C2,$48,$F4,$7C,$25,$CF,$F0,$B5
	.dc.b	$FC,$45,$BF,$05,$E4,$F8,$3F,$2B,$E6,$F8,$39,$87,$C1,$F8,$3B,$7C
	.dc.b	$BE,$0B,$D7,$F0,$7E,$C6,$D7,$C7,$F8,$2F,$7F,$C1,$FC,$06,$C7,$CB
	.dc.b	$F8,$3F,$87,$F0,$7F,$11,$B1,$F3,$FE,$0F,$F9,$57,$E2,$BF,$C4,$33
	.dc.b	$F4,$C6,$1C,$FC,$1F,$FB,$4E,$7F,$C4,$64,$F8,$3C,$7C,$1E,$7E,$0E
	.dc.b	$5F,$FA,$AA,$BF,$C4,$ED,$F8,$3E,$44,$75,$F0,$7D,$7E,$0F,$C7,$F0
	.dc.b	$7E,$46,$EB,$CF,$E0,$FC,$E7,$A3,$E0,$EC,$77,$F8,$3F,$5F,$C1,$FB
	.dc.b	$1B,$1E,$3F,$83,$F7,$FC,$1F,$C0,$6C,$79,$FE,$0B,$E1,$FC,$17,$C4
	.dc.b	$6C,$7A,$FE,$0B,$E3,$9F,$23,$E0,$AC,$7B,$FE,$07,$E5,$9F,$33,$E0
	.dc.b	$AC,$7C,$3F,$82,$F9,$E7,$F5,$9F,$C5,$BF,$86,$7C,$2F,$F8,$93,$9D
	.dc.b	$39,$F8,$49,$7F,$ED,$4B,$FF,$59,$B6,$39,$7C,$1B,$A3,$5F,$07,$5F
	.dc.b	$E2,$F6,$8C,$65,$FE,$32,$85,$63,$0E,$5E,$92,$7F,$8A,$61,$E1,$F8
	.dc.b	$2D,$29,$D4,$61,$E1,$F0,$56,$9F,$05,$7E,$0A,$61,$EB,$8D,$7C,$17
	.dc.b	$C3,$3E,$27,$C1,$5C,$FE,$07,$C0,$FC,$7F,$82,$F9,$0D,$8F,$89,$F0
	.dc.b	$3F,$2F,$E0,$BE,$63,$63,$E4,$7C,$0F,$CF,$F8,$2F,$A0,$D8,$F9,$9F
	.dc.b	$03,$F4,$FE,$0B,$EA,$36,$3E,$87,$C0,$FD,$73,$EC,$7C,$15,$8F,$A1
	.dc.b	$CF,$FE,$5C,$FC,$5F,$FD,$7A,$7F,$8B,$38,$BE,$33,$79,$8E,$BE,$0B
	.dc.b	$E4,$3F,$E2,$CB,$FA,$2D,$AE,$3F,$07,$AF,$83,$DF,$C1,$C9,$F0,$DF
	.dc.b	$09,$E7,$F0,$72,$72,$78,$9E,$FF,$03,$E0,$F8,$3F,$0F,$C1,$CA,$DE
	.dc.b	$BE,$69,$F2,$FC,$14,$9C,$EB,$FE,$2F,$6F,$C1,$FA,$BE,$0E,$46,$9F
	.dc.b	$6F,$C1,$FB,$BE,$0B,$DF,$F0,$72,$B9,$C7,$C2,$F8,$29,$1D,$9F,$8B
	.dc.b	$F0,$7F,$1B,$E0,$BF,$C8,$19,$9B,$F0,$78,$73,$F0,$73,$97,$D7,$E3
	.dc.b	$7F,$1D,$EA,$FA,$D3,$FE,$49,$8F,$F1,$64,$FE,$81,$B9,$FE,$75,$FF
	.dc.b	$17,$47,$2F,$83,$E1,$F0,$5C,$7E,$0A,$CF,$67,$A7,$C1,$48,$F0,$FC
	.dc.b	$15,$C7,$FC,$13,$8F,$FA,$85,$BF,$03,$E3,$F8,$2F,$23,$E5,$F8,$2F
	.dc.b	$D6,$51,$A6,$3E,$0B,$3F,$03,$AF,$82,$98,$FC,$08,$E3,$F0,$32,$3E
	.dc.b	$1B,$D3,$E0,$BA,$FC,$0F,$C3,$3E,$27,$C1,$5D,$FE,$4F,$C1,$7C,$73
	.dc.b	$E4,$7C,$15,$8F,$9B,$F0,$5F,$2F,$E0,$BE,$63,$63,$E8,$FC,$0F,$CF
	.dc.b	$3F,$CF,$1F,$8C,$7F,$20,$8F,$A1,$18,$F8,$2C,$FC,$35,$9F,$C9,$D7
	.dc.b	$F2,$0D,$FF,$C6,$2F,$F1,$88,$CF,$C1,$F4,$F8,$3E,$AF,$6F,$82,$EF
	.dc.b	$F0,$76,$B9,$BD,$70,$E7,$8D,$F0,$7F,$DA,$B9,$77,$6F,$4F,$A0,$F4
	.dc.b	$FC,$15,$C7,$CB,$F0,$5E,$CF,$82,$F6,$B6,$3D,$3B,$FB,$1A,$7B,$32
	.dc.b	$F8,$7F,$03,$F1,$3E,$0A,$CF,$C0,$7E,$3F,$C0,$FC,$8F,$82,$B3,$F1
	.dc.b	$1F,$97,$F0,$3F,$33,$E0,$AC,$FC,$87,$E7,$FC,$0C,$8F,$97,$F0,$57
	.dc.b	$8F,$E5,$FC,$17,$D4,$F8,$2F,$AA,$DA,$FA,$38,$F5,$69,$D4,$FE,$0F
	.dc.b	$F8,$CF,$F8,$D0,$9E,$E7,$FA,$D5,$C7,$C1,$DF,$83,$C3,$6B,$9E,$5F
	.dc.b	$73,$5F,$4B,$63,$AF,$C1,$73,$F8,$3E,$8D,$8F,$0F,$C1,$F7,$3C,$1F
	.dc.b	$07,$63,$CB,$F0,$5E,$3F,$83,$F2,$36,$3D,$3F,$07,$E7,$3D,$1F,$07
	.dc.b	$63,$DB,$F0,$7E,$BF,$83,$F6,$36,$3A,$9F,$0B,$0E,$AD,$3E,$2F,$C1
	.dc.b	$7C,$3F,$83,$F8,$8D,$8F,$93,$F0,$7F,$1C,$F9,$1F,$07,$63,$AF,$0F
	.dc.b	$FA,$43,$97,$FD,$C3,$18,$F1,$8F,$F7,$88,$CC,$EF,$E0,$7E,$97,$C1
	.dc.b	$7D,$36,$D7,$D8,$F8,$2F,$AC,$7F,$BA,$FF,$1A,$FE,$4D,$1C,$E3,$1F
	.dc.b	$07,$23,$FA,$2C,$FF,$24,$C3,$FD,$46,$FF,$28,$EC,$2F,$83,$E8,$75
	.dc.b	$F8,$3B,$7C,$BE,$0F,$C0,$78,$7E,$0E,$C7,$6C,$BC,$19,$76,$F8,$7B
	.dc.b	$C1,$F0,$23,$FC,$25,$9F,$F2,$B8,$C6,$9E,$2F,$83,$F6,$7C,$27,$B5
	.dc.b	$F7,$7C,$24,$CF,$D7,$F0,$97,$2F,$5E,$9D,$A7,$FC,$47,$9F,$BF,$3E
	.dc.b	$AF,$BC,$7C,$1F,$C9,$3E,$57,$C2,$5A,$F8,$FF,$09,$F3,$7E,$13,$E7
	.dc.b	$36,$3E,$5F,$C2,$7D,$1F,$84,$FA,$4D,$8F,$9F,$F0,$9F,$54,$FA,$DF
	.dc.b	$09,$63,$E9,$FC,$1F,$D9,$3E,$D7,$C2,$58,$EA,$7D,$8C,$3A,$B4,$FB
	.dc.b	$7F,$09,$F2,$F7,$F9,$3C,$BF,$CD,$78,$BE,$6C,$E6,$BF,$C8,$B3,$9E
	.dc.b	$63,$FE,$E2,$AF,$F7,$2B,$FC,$A6,$77,$ED,$F1,$9F,$F2,$44,$E5,$FB
	.dc.b	$3B,$7B,$DF,$C5,$B7,$1F,$1F,$C2,$71,$C7,$BF,$C2,$5C,$3F,$A1,$61
	.dc.b	$E0,$8C,$7C,$24,$CF,$FC,$37,$4F,$84,$C5,$F9,$BE,$0F,$91,$CF,$E0
	.dc.b	$ED,$F0,$F8,$2E,$C7,$7F,$83,$B1,$D3,$E0,$BC,$5F,$05,$E3,$6C,$78
	.dc.b	$30,$E7,$8F,$C4,$C6,$F0,$7C,$EF,$82,$F5,$1E,$BF,$82,$B8,$7A,$3D
	.dc.b	$EF,$77,$B9,$EF,$6C,$75,$26,$3D,$5B,$87,$DA,$F8,$2F,$8C,$7C,$7F
	.dc.b	$82,$B3,$F1,$3E,$07,$FC,$F1,$C6,$69,$8F,$82,$FF,$3E,$7E,$36,$F9
	.dc.b	$9A,$4C,$FF,$C8,$F1,$FE,$5D,$3F,$CB,$1D,$25,$F0,$9C,$8E,$7F,$09
	.dc.b	$D3,$E0,$FA,$FC,$24,$C3,$9C,$78,$30,$ED,$0E,$0F,$DE,$E3,$E3,$8F
	.dc.b	$0C,$F9,$BE,$13,$CF,$F0,$92,$7C,$B1,$EA,$F8,$4F,$5F,$C0,$49,$F4
	.dc.b	$C7,$77,$D7,$B7,$83,$2E,$F8,$FC,$2F,$84,$95,$F0,$1F,$8B,$F0,$9F
	.dc.b	$1B,$E0,$E4,$FC,$47,$E4,$FC,$27,$CA,$F8,$39,$3F,$21,$F9,$BF,$09
	.dc.b	$F3,$BE,$12,$4F,$CC,$7E,$8F,$C2,$7D,$2F,$82,$93,$F4,$1F,$AB,$F0
	.dc.b	$7D,$23,$E3,$76,$7F,$AE,$B9,$59,$67,$83,$E4,$E7,$F7,$3E,$0F,$FD
	.dc.b	$71,$F8,$E7,$FB,$1F,$1E,$27,$FA,$F5,$C5,$7D,$09,$DF,$B9,$2B,$FD
	.dc.b	$0F,$48,$E7,$F0,$1D,$F3,$F1,$65,$6F,$FD,$91,$9B,$3F,$0F,$C2,$78
	.dc.b	$BE,$12,$57,$48,$F2,$FC,$24,$8E,$EF,$A3,$E1,$3D,$3F,$09,$EA,$F1
	.dc.b	$65,$79,$23,$DB,$F0,$9E,$EF,$84,$93,$E8,$8F,$83,$F0,$7F,$0B,$E0
	.dc.b	$E4,$FB,$23,$AB,$EE,$C3,$AB,$F1,$BB,$9C,$8C,$BE,$5F,$C1,$4A,$F8
	.dc.b	$AF,$CE,$F8,$3F,$9F,$EE,$49,$F9,$2F,$D2,$F7,$FB,$C7,$5C,$3F,$96
	.dc.b	$BE,$13,$D8,$7D,$0F,$83,$FB,$24,$7E,$9F,$C1,$FD,$CF,$83,$FB,$BF
	.dc.b	$07,$27,$EB,$C7,$E0,$3E,$0E,$47,$DA,$7F,$DF,$5F,$8E,$FF,$6D,$8F
	.dc.b	$72,$F8,$4C,$92,$3F,$D8,$F0,$F8,$7E,$3F,$0B,$27,$2F,$4F,$85,$EF
	.dc.b	$1E,$DC,$2F,$89,$FD,$E3,$0F,$03,$7E,$17,$C6,$4A,$EF,$F0,$BE,$6F
	.dc.b	$81,$F3,$FC,$0C,$9F,$1C,$74,$F1,$51,$F9,$BC,$B0,$BA,$59,$F8,$2B
	.dc.b	$71,$8F,$85,$F8,$7F,$0B,$2F,$DF,$1F,$1B,$E1,$3E,$3F,$C2,$C9,$F8
	.dc.b	$71,$F2,$BE,$16,$47,$C6,$7E,$6F,$C2,$FC,$EF,$84,$93,$F2,$9F,$A3
	.dc.b	$F0,$BF,$4B,$E1,$64,$FC,$E7,$EA,$FC,$2F,$48,$FA,$B8,$5E,$5F,$95
	.dc.b	$8F,$DA,$F8,$59,$5F,$59,$FB,$BF,$0B,$F1,$23,$F3,$5E,$F7,$F2,$B6
	.dc.b	$FB,$D9,$C3,$FC,$57,$39,$6F,$17,$7E,$3F,$8F,$C8,$D3,$18,$FC,$D8
	.dc.b	$69,$CE,$BE,$6E,$3F,$D1,$89,$F0,$99,$BF,$A9,$1F,$D2,$1F,$B3,$97
	.dc.b	$8B,$E0,$F5,$9A,$9F,$09,$86,$DF,$0C,$7B,$DE,$7E,$0F,$5F,$09,$2B
	.dc.b	$47,$82,$39,$5F,$3F,$84,$93,$AC,$BC,$13,$DF,$E1,$3C,$1F,$09,$2B
	.dc.b	$52,$9F,$04,$63,$1F,$03,$F4,$31,$F4,$7C,$27,$A7,$E1,$24,$F2,$8F
	.dc.b	$67,$BB,$ED,$F7,$24,$F6,$8E,$AF,$DA,$C3,$AF,$0F,$D8,$C7,$E2,$7C
	.dc.b	$24,$AF,$2B,$F1,$FE,$13,$E4,$7C,$24,$9F,$4B,$F2,$FE,$13,$E6,$7C
	.dc.b	$24,$9F,$6B,$F4,$BF,$1E,$FE,$8B,$CD,$FA,$87,$FA,$14,$FC,$75,$F0
	.dc.b	$9B,$38,$7C,$2F,$1F,$83,$95,$B7,$A7,$C2,$77,$8F,$CA,$E3,$77,$FF
	.dc.b	$3E,$E3,$E0,$9F,$DD,$2F,$84,$F2,$12,$3C,$5F,$09,$E7,$F7,$E4,$79
	.dc.b	$5F,$57,$C0,$75,$DB,$F2,$B9,$5A,$EA,$7B,$FE,$0F,$E0,$7C,$24,$8F
	.dc.b	$74,$7C,$3F,$83,$91,$F0,$5F,$8D,$F0,$9F,$1F,$E1,$3E,$47,$C2,$4A
	.dc.b	$F8,$D1,$F2,$FE,$12,$47,$C9,$7E,$77,$C2,$7C,$FF,$83,$FA,$1F,$07
	.dc.b	$2B,$E7,$47,$57,$E8,$61,$D5,$FE,$AD,$C6,$C7,$EC,$7C,$14,$9F,$AC
	.dc.b	$FD,$BF,$83,$FB,$9E,$DC,$9F,$B4,$FE,$C3,$F1,$FF,$E4,$D1,$F4,$2F
	.dc.b	$B7,$47,$D4,$CB,$C5,$3F,$D9,$6B,$FA,$43,$FA,$97,$FA,$33,$7E,$0F
	.dc.b	$A7,$C2,$75,$F8,$4F,$F3,$F5,$3E,$12,$C6,$1C,$FC,$24,$E9,$E9,$F0
	.dc.b	$97,$2F,$07,$C1,$F2,$F8,$4E,$6D,$AF,$17,$C1,$F6,$F8,$4E,$ED,$8F
	.dc.b	$36,$3E,$2F,$83,$E9,$C3,$FE,$CC,$D9,$1E,$6F,$84,$F3,$FB,$B7,$0F
	.dc.b	$73,$EA,$F7,$24,$7C,$1F,$84,$BB,$FC,$5C,$FB,$4F,$80,$F8,$D8,$FC
	.dc.b	$2F,$84,$F8,$7F,$09,$67,$E5,$3F,$1B,$E1,$24,$7C,$DF,$84,$B8,$FD
	.dc.b	$1F,$84,$FF,$54,$7E,$41,$FE,$79,$AF,$A5,$18,$F8,$5C,$FC,$04,$9F
	.dc.b	$E3,$D4,$F8,$4C,$61,$FE,$1F,$8F,$F3,$DC,$EB,$E0,$FA,$FC,$47,$67
	.dc.b	$EC,$E5,$D6,$7B,$9D,$32,$F1,$FC,$2F,$93,$DC,$B3,$E0,$7C,$FE,$EC
	.dc.b	$8F,$1F,$C1,$DD,$3D,$91,$CC,$EA,$F9,$70,$F7,$FC,$2C,$9F,$5F,$C2
	.dc.b	$5A,$F5,$7C,$2F,$C4,$F8,$4F,$8A,$DA,$F8,$3F,$0B,$F2,$0F,$93,$F0
	.dc.b	$96,$3E,$2F,$C2,$FC,$C3,$E6,$FC,$2D,$8F,$93,$F0,$BF,$43,$E1,$7E
	.dc.b	$8B,$63,$A9,$33,$F9,$98,$75,$7E,$77,$C2,$FD,$83,$EC,$FB,$B6,$BF
	.dc.b	$A9,$79,$5F,$E1,$C6,$D7,$FB,$96,$3F,$CE,$51,$83,$B7,$8E,$7F,$C3
	.dc.b	$BD,$4D,$BD,$B1,$A3,$FC,$43,$5F,$CE,$1E,$FF,$07,$D7,$E1,$7B,$3D
	.dc.b	$FE,$12,$4F,$93,$E1,$2E,$7F,$E6,$5C,$1F,$09,$E5,$3C,$DF,$09,$6F
	.dc.b	$D7,$F0,$9E,$93,$D5,$F0,$76,$3D,$FF,$09,$ED,$3D,$DF,$09,$63,$E1
	.dc.b	$E1,$CF,$2F,$06,$37,$87,$E0,$47,$F8,$FF,$37,$C2,$7C,$73,$E4,$7B
	.dc.b	$97,$1F,$9B,$EE,$7C,$BF,$84,$F9,$8D,$8E,$E4,$C3,$BB,$71,$EF,$E5
	.dc.b	$32,$7C,$27,$D5,$F8,$3F,$AC,$DB,$FB,$7F,$09,$FE,$1C,$A7,$C1,$D3
	.dc.b	$11,$9F,$83,$D7,$C1,$CD,$3F,$07,$1F,$E2,$4F,$C8,$7F,$9A,$E8,$7F
	.dc.b	$98,$CF,$FA,$5D,$BA,$F9,$9D,$9F,$CD,$23,$5F,$09,$CC,$E9,$F0,$7D
	.dc.b	$7E,$0E,$4F,$F8,$8E,$BF,$9B,$37,$DA,$F1,$7C,$1F,$8F,$E0,$E4,$EE
	.dc.b	$3C,$DC,$1F,$53,$3F,$17,$C1,$5E,$2F,$F3,$8E,$5F,$25,$ED,$19,$C3
	.dc.b	$FC,$F1,$2F,$83,$C6,$FE,$FF,$7A,$CE,$70,$F3,$46,$3E,$0B,$3F,$05
	.dc.b	$67,$D4,$F0,$F8,$2F,$94,$4C,$BD,$DF,$05,$F3,$7E,$0B,$E7,$7C,$14
	.dc.b	$9F,$85,$1F,$47,$E0,$BE,$97,$C1,$49,$F8,$D1,$F5,$7D,$FF,$F2,$CF
	.dc.b	$E4,$5F,$E6,$BD,$3F,$D1,$ED,$AF,$6C,$EB,$82,$47,$F9,$46,$7F,$D3
	.dc.b	$0D,$F8,$3E,$7F,$07,$23,$4F,$6F,$83,$EF,$F0,$7F,$1A,$DF,$07,$63
	.dc.b	$0E,$7E,$0E,$69,$DF,$E0,$EE,$5F,$ED,$FC,$1F,$05,$CF,$E0,$BA,$36
	.dc.b	$FC,$DF,$03,$DF,$E0,$BC,$0D,$8F,$57,$B5,$E3,$F7,$BC,$8D,$8F,$77
	.dc.b	$C1,$79,$CF,$47,$C1,$58,$F8,$5F,$03,$EB,$F8,$2F,$63,$63,$A9,$F1
	.dc.b	$F0,$EA,$D3,$E5,$7C,$17,$C3,$3E,$27,$C1,$58,$FF,$3F,$7E,$47,$FC
	.dc.b	$B7,$87,$FD,$FC,$DF,$1F,$F7,$AF,$F2,$E8,$C9,$D8,$F8,$D8,$59,$FE
	.dc.b	$07,$C4,$8F,$CB,$D7,$F2,$DA,$EB,$F0,$38,$F8,$2C,$BA,$F8,$2F,$F4
	.dc.b	$FD,$1A,$63,$E0,$B3,$EF,$4C,$3D,$2F,$0F,$7B,$8F,$C1,$72,$F8,$29
	.dc.b	$5E,$E8,$EB,$F0,$5D,$BE,$0A,$4F,$C2,$8E,$8F,$C2,$C2,$F0,$7C,$2C
	.dc.b	$7C,$BF,$05,$E6,$F8,$29,$7F,$2A,$3D,$3F,$03,$EA,$F8,$29,$3F,$3A
	.dc.b	$39,$F2,$FF,$BC,$38,$7F,$A3,$CF,$A8,$FA,$78,$FC,$2F,$81,$91,$F5
	.dc.b	$9F,$8B,$F0,$5F,$1B,$E0,$64,$FD,$A7,$E4,$FC,$17,$CA,$F8,$19,$3F
	.dc.b	$79,$F9,$BF,$05,$FE,$CC,$E5,$FE,$52,$4D,$31,$F0,$D9,$F8,$39,$87
	.dc.b	$FB,$0E,$BF,$94,$71,$FF,$40,$5F,$07,$DA,$7F,$13,$95,$AF,$C4,$E5
	.dc.b	$DF,$E0,$FC,$1F,$07,$2F,$A4,$78,$FE,$0B,$C9,$F0,$72,$7C,$11,$CF
	.dc.b	$6F,$91,$95,$97,$AA,$E5,$67,$ED,$F8,$39,$5E,$97,$E0,$7C,$1F,$C1
	.dc.b	$F8,$3F,$85,$F0,$52,$BD,$D1,$F1,$7E,$0A,$47,$C1,$7E,$47,$C1,$7C
	.dc.b	$9F,$82,$93,$F1,$5F,$99,$F0,$5F,$37,$E0,$7E,$77,$C1,$4A,$F9,$51
	.dc.b	$D1,$73,$7C,$BC,$70,$FC,$DC,$7E,$B7,$C0,$CA,$FA,$4F,$D9,$F8,$2F
	.dc.b	$B5,$F0,$52,$7E,$B2,$7F,$DF,$5F,$92,$7F,$2F,$E4,$76,$33,$EF,$CF
	.dc.b	$B9,$CC,$7B,$AF,$8C,$73,$7C,$F3,$BB,$FE,$6B,$18,$3E,$9B,$8D,$3F
	.dc.b	$C0,$AF,$F9,$B6,$3C,$7F,$05,$CE,$73,$F0,$56,$FA,$36,$B9,$D5,$C1
	.dc.b	$83,$E0,$B0,$DA,$F7,$C6,$3E,$0B,$3F,$05,$27,$E1,$BC,$3E,$0B,$8F
	.dc.b	$C0,$C9,$F8,$EF,$4F,$82,$EB,$84,$9E,$A7,$CC,$C3,$C3,$F0,$32,$3E
	.dc.b	$7B,$E4,$F8,$2F,$2F,$C1,$79,$BD,$B9,$5F,$52,$3A,$3F,$27,$6F,$6E
	.dc.b	$38,$E2,$F9,$39,$7B,$BE,$0B,$DF,$F0,$53,$0F,$BB,$1F,$0B,$E0,$7E
	.dc.b	$1F,$C1,$49,$FC,$0C,$7C,$6F,$82,$F8,$FF,$05,$27,$F0,$B1,$F3,$7F
	.dc.b	$05,$FB,$D1,$E3,$5F,$15,$93,$51,$BF,$83,$8F,$E3,$F5,$FF,$21,$D2
	.dc.b	$7A,$7C,$1F,$86,$BF,$27,$8F,$85,$F8,$1D,$2C,$BC,$5F,$07,$E3,$F8
	.dc.b	$39,$5D,$7B,$BF,$2D,$9F,$F4,$58,$C1,$FC,$09,$FE,$56,$FA,$7D,$BF
	.dc.b	$61,$ED,$F8,$0F,$14,$78,$33,$F1,$6D,$F9,$6C,$FE,$17,$C1,$CA,$EB
	.dc.b	$CE,$C9,$3F,$1F,$C3,$F9,$1F,$05,$2F,$74,$8F,$99,$F0,$7F,$37,$E0
	.dc.b	$BE,$77,$C1,$CB,$DC,$7D,$1F,$82,$FA,$5F,$07,$27,$9C,$7D,$5F,$83
	.dc.b	$FA,$DF,$07,$27,$BC,$75,$7C,$38,$75,$7E,$D6,$1F,$77,$E0,$FE,$F7
	.dc.b	$C1,$C9,$F3,$C7,$E0,$7D,$C9,$1E,$A7,$F0,$9E,$DF,$58,$FC,$26,$1D
	.dc.b	$79,$3E,$FE,$3F,$B5,$FC,$17,$FC,$E6,$4E,$9C,$7C,$26,$7E,$13,$FC
	.dc.b	$B3,$26,$FC,$26,$1C,$FC,$1E,$BE,$12,$E1,$F4,$AD,$1C,$BE,$0E,$4E
	.dc.b	$7E,$12,$ED,$FA,$BB,$FF,$36,$37,$E0,$BC,$3F,$07,$E2,$7A,$C6,$F1
	.dc.b	$BB,$FF,$A7,$2F,$BE,$3E,$1F,$82,$F4,$FC,$1F,$A9,$B1,$E5,$F6,$FD
	.dc.b	$BE,$FF,$B9,$B1,$E9,$DF,$D8,$D3,$D9,$97,$C4,$F8,$39,$1E,$FF,$83
	.dc.b	$BA,$FA,$FE,$0B,$E4,$FC,$1F,$CA,$6D,$7C,$5F,$83,$F9,$BF,$07,$F3
	.dc.b	$9B,$1F,$27,$E0,$FE,$89,$FE,$8A,$F7,$DF,$13,$89,$FC,$FC,$7F,$6C
	.dc.b	$6F,$97,$BF,$87,$E0,$F5,$8F,$A9,$8D,$D3,$E8,$3D,$7E,$13,$B7,$C2
	.dc.b	$DA,$EA,$F8,$7D,$EF,$17,$BD,$67,$C2,$F3,$EA,$FA,$58,$59,$FE,$AD
	.dc.b	$D4,$8F,$57,$C2,$C9,$F4,$FC,$2D,$D3,$F9,$66,$FF,$E7,$A9,$F6,$BF
	.dc.b	$07,$E1,$7E,$13,$F0,$FE,$16,$4F,$C2,$F8,$5B,$7C,$BE,$13,$E4,$7C
	.dc.b	$2F,$C9,$6D,$7C,$8F,$84,$F9,$9F,$0B,$F3,$5B,$1F,$33,$1E,$7C,$1F
	.dc.b	$EE,$8C,$D9,$E2,$BF,$B9,$43,$E1,$3E,$B1,$F5,$FE,$16,$CF,$D6,$F7
	.dc.b	$3E,$D1,$F6,$FD,$DB,$1D,$C9,$BF,$C2,$C7,$B3,$FE,$28,$FB,$BF,$F3
	.dc.b	$F4,$FF,$80,$5B,$19,$F8,$6F,$F6,$7C,$63,$AC,$FF,$9E,$F5,$FF,$80
	.dc.b	$2B,$E1,$77,$F0,$DC,$1E,$3F,$0B,$CB,$E1,$6E,$7F,$CB,$2D,$1D,$BE
	.dc.b	$13,$BF,$C2,$DA,$D3,$E2,$F8,$59,$1C,$7E,$16,$EF,$FD,$A7,$1F,$F4
	.dc.b	$44,$F5,$C7,$BC,$E5,$F5,$FC,$24,$9F,$17,$C2,$5A,$E5,$ED,$FC,$0F
	.dc.b	$73,$E0,$B6,$BD,$1B,$7A,$DA,$77,$3D,$9F,$07,$F2,$3E,$13,$E4,$B6
	.dc.b	$3E,$07,$C1,$FC,$C3,$E6,$FC,$25,$8F,$89,$F0,$7F,$43,$E1,$3E,$8B
	.dc.b	$63,$E4,$7C,$27,$D4,$F8,$4F,$AA,$D8,$FE,$AF,$F9,$29,$9B,$FF,$83
	.dc.b	$63,$11,$AF,$85,$CF,$5E,$BB,$15,$AB,$F9,$99,$78,$5B,$87,$D9,$CF
	.dc.b	$8B,$FE,$49,$AF,$E7,$8F,$8B,$E1,$33,$5F,$E7,$69,$FF,$43,$CE,$4F
	.dc.b	$3F,$B9,$AC,$7C,$2D,$9F,$E6,$35,$E9,$8F,$87,$5E,$13,$E3,$96,$3D
	.dc.b	$DF,$09,$C4,$E5,$F0,$96,$7E,$2F,$C2,$7F,$92,$F0,$34,$C7,$C2,$67
	.dc.b	$E0,$E6,$3F,$31,$E1,$F0,$9C,$7E,$0F,$97,$C2,$4A,$FA,$31,$D7,$E0
	.dc.b	$E4,$7D,$47,$C1,$F0,$7D,$23,$EA,$61,$79,$3E,$96,$8C,$BC,$DF,$05
	.dc.b	$E7,$F8,$39,$87,$DE,$8F,$57,$C1,$FA,$FD,$B9,$3F,$82,$8E,$9C,$3F
	.dc.b	$57,$3B,$3F,$4F,$1F,$85,$F0,$7F,$E6,$DF,$C9,$BF,$D0,$C7,$FA,$1F
	.dc.b	$44,$67,$9E,$5F,$F8,$92,$BF,$CF,$D8,$FF,$A2,$5F,$F3,$F4,$7F,$96
	.dc.b	$D7,$C1,$75,$F8,$29,$3C,$63,$C1,$F0,$52,$3A,$3E,$3F,$82,$F2,$7C
	.dc.b	$0F,$97,$E0,$65,$78,$63,$D1,$7D,$23,$C1,$86,$1F,$37,$B7,$ED,$F8
	.dc.b	$19,$18,$3D,$5E,$EF,$C1,$23,$ED,$F6,$79,$3D,$0F,$81,$78,$E8,$FD
	.dc.b	$46,$CF,$80,$F9,$5F,$03,$2F,$E3,$47,$CD,$F8,$09,$1F,$25,$FA,$1F
	.dc.b	$03,$F4,$7E,$06,$4F,$CD,$7E,$A7,$C0,$FD,$5F,$81,$FA,$DF,$03,$2B
	.dc.b	$E9,$47,$FA,$97,$C5,$FF,$3C,$EB,$FB,$B7,$FC,$F3,$3F,$E5,$A6,$BF
	.dc.b	$E8,$F7,$7D,$1F,$67,$1F,$0C,$FD,$97,$C2,$F1,$F8,$0E,$DF,$01,$23
	.dc.b	$07,$5F,$5B,$C4,$47,$C1,$ED,$F4,$CB,$E3,$E1,$71,$FC,$46,$BE,$9F
	.dc.b	$80,$98,$79,$DF,$67,$C0,$7B,$7E,$02,$4F,$AD,$F8,$1F,$01,$F0,$7D
	.dc.b	$F9,$3E,$F7,$E2,$7B,$FF,$17,$DE,$F8,$DE,$FC,$AF,$89,$1F,$27,$DE
	.dc.b	$91,$F1,$DF,$99,$EF,$F4,$8F,$91,$85,$E0,$F9,$D8,$FD,$1F,$7F,$E9
	.dc.b	$7B,$F2,$FE,$84,$7D,$5F,$73,$EB,$7B,$92,$7E,$A4,$73,$E3,$FF,$79
	.dc.b	$74,$FF,$A2,$67,$D4,$7D,$9C,$7E,$F7,$BD,$F7,$FD,$F9,$3F,$76,$3F
	.dc.b	$DC,$DF,$93,$FF,$40,$8F,$F6,$63,$75,$FE,$7C,$DC,$7F,$E0,$02,$F8
	.dc.b	$0E,$3F,$01,$CB,$E1,$E6,$1F,$EA,$2A,$FF,$3C,$37,$E0,$7B,$91,$D7
	.dc.b	$C3,$F8,$BE,$0A,$47,$17,$CB,$F0,$5D,$A3,$A6,$1D,$9F,$3E,$3E,$AF
	.dc.b	$82,$F5,$F0,$C9,$F1,$47,$BB,$DA,$F7,$FB,$52,$7C,$D1,$D5,$F8,$38
	.dc.b	$75,$E0,$F8,$38,$FC,$6F,$81,$F8,$FF,$05,$2F,$DD,$1F,$2B,$E0,$BE
	.dc.b	$5F,$C1,$49,$F8,$51,$F3,$BE,$07,$E7,$FC,$14,$9F,$8D,$1F,$4B,$E0
	.dc.b	$7E,$9F,$C1,$49,$F9,$51,$F5,$BE,$06,$47,$CD,$4F,$FB,$43,$F2,$8F
	.dc.b	$F3,$AB,$D0,$FA,$58,$77,$7E,$EE,$1C,$16,$DF,$DE,$EB,$F9,$AC,$3E
	.dc.b	$09,$18,$69,$C6,$78,$7C,$3D,$EC,$F1,$CF,$0F,$7F,$18,$73,$C1,$ED
	.dc.b	$E2,$BA,$B6,$3F,$A0,$CF,$A4,$FC,$9E,$7E,$9C,$BF,$27,$9E,$FE,$0F
	.dc.b	$87,$C1,$CB,$F8,$51,$CF,$E0,$BF,$C0,$D8,$9B,$F0,$78,$73,$F0,$73
	.dc.b	$7F,$95,$F0,$77,$1F,$8B,$1E,$CC,$1F,$07,$D3,$E0,$FA,$B6,$7E,$97
	.dc.b	$C0,$F8,$3E,$0B,$C2,$D8,$FA,$DF,$05,$E4,$F8,$2F,$2B,$63,$ED,$7B
	.dc.b	$9E,$8F,$73,$D2,$D8,$FB,$DF,$03,$EC,$F8,$2F,$6B,$63,$F0,$7C,$DF
	.dc.b	$E3,$5F,$CA,$7F,$CE,$43,$E1,$FF,$C5,$0E,$1B,$F0,$92,$B8,$9F,$6A
	.dc.b	$7F,$CE,$CE,$38,$BF,$09,$1F,$E7,$B8,$D7,$C1,$F4,$F8,$3E,$AD,$8E
	.dc.b	$1D,$CA,$7C,$1F,$88,$F1,$FC,$1D,$AD,$CB,$E0,$FC,$E7,$A3,$E0,$EC
	.dc.b	$EB,$DF,$FC,$55,$3D,$BA,$62,$33,$F0,$53,$6E,$AF,$0F,$83,$E3,$F0
	.dc.b	$72,$7C,$2F,$4F,$83,$EB,$F0,$5D,$BE,$0E,$57,$9A,$3A,$3F,$86,$3E
	.dc.b	$2E,$18,$E6,$FA,$58,$F9,$BE,$0B,$CF,$F0,$72,$FD,$F1,$EA,$F8,$2F
	.dc.b	$AA,$7D,$6F,$83,$BC,$3F,$2B,$7F,$F3,$76,$5F,$E7,$16,$C7,$EC,$1F
	.dc.b	$E6,$8E,$CF,$F3,$33,$FE,$6E,$C3,$CD,$CC,$8C,$57,$3F,$67,$7E,$CF
	.dc.b	$0F,$81,$FD,$B5,$3E,$06,$98,$8C,$FC,$0C,$CB,$CC,$F0,$F8,$1E,$3F
	.dc.b	$01,$CB,$E0,$65,$7A,$E3,$AF,$C0,$79,$8F,$3F,$C0,$DE,$0F,$81,$F0
	.dc.b	$1E,$A3,$D7,$F0,$36,$3E,$27,$C0,$7B,$BE,$07,$DE,$D8,$F9,$1E,$D7
	.dc.b	$C2,$3E,$1F,$B7,$63,$A9,$34,$F8,$BC,$4C,$7B,$46,$3E,$03,$E5,$9F
	.dc.b	$33,$E0,$6D,$FD,$5F,$81,$F9,$E7,$FA,$57,$F2,$A7,$C5,$62,$7B,$75
	.dc.b	$7D,$9E,$0F,$5C,$60,$FE,$53,$5F,$CC,$5F,$D4,$B7,$2F,$B2,$BE,$0F
	.dc.b	$B7,$67,$7F,$83,$FE,$59,$3F,$CB,$AB,$8F,$C1,$E2,$73,$F0,$52,$7A
	.dc.b	$BC,$3E,$0F,$8F,$C1,$72,$F8,$39,$5E,$28,$EB,$F0,$72,$3C,$AF,$83
	.dc.b	$E0,$FA,$47,$9B,$0B,$C1,$FE,$C7,$D1,$97,$9B,$E0,$65,$FB,$9F,$4F
	.dc.b	$C1,$7A,$BE,$06,$4F,$C2,$7D,$BF,$05,$D3,$8F,$E3,$E7,$6B,$FA,$C6
	.dc.b	$5D,$9F,$93,$F0,$3F,$10,$8F,$CC,$F8,$2F,$8F,$F0,$32,$3E,$7B,$F2
	.dc.b	$BE,$0B,$E5,$FC,$0C,$9F,$A6,$FC,$EF,$82,$FE,$B5,$F9,$5F,$F9,$92
	.dc.b	$7F,$B4,$3F,$E6,$8E,$2F,$F7,$33,$75,$FE,$CA,$EF,$A4,$BE,$0B,$99
	.dc.b	$2F,$F9,$4C,$3E,$1F,$C5,$1F,$82,$CA,$E3,$F3,$72,$F1,$FC,$14,$AE
	.dc.b	$2F,$9B,$E0,$FC,$FF,$07,$E8,$F8,$39,$5D,$A3,$9E,$DF,$92,$CA,$CF
	.dc.b	$CC,$C7,$DF,$F0,$72,$3C,$AF,$C2,$F8,$3F,$87,$E4,$C9,$F4,$BF,$1B
	.dc.b	$E0,$BE,$3F,$BD,$F2,$3D,$E9,$5E,$E8,$F9,$7F,$05,$F3,$3E,$0A,$4F
	.dc.b	$C2,$8F,$9F,$F0,$3F,$43,$E0,$A4,$FC,$68,$EA,$FC,$8C,$3A,$BF,$C8
	.dc.b	$39,$98,$FD,$8F,$81,$93,$F3,$9F,$B7,$F0,$5F,$73,$E0,$BE,$EF,$C1
	.dc.b	$4A,$FA,$71,$D9,$E9,$8F,$6E,$9F,$BF,$8F,$F8,$77,$F2,$CF,$F3,$45
	.dc.b	$7F,$8A,$DF,$F3,$43,$FD,$91,$AF,$F3,$47,$7D,$A7,$FC,$D3,$19,$F8
	.dc.b	$3E,$87,$5F,$83,$ED,$F0,$52,$7A,$3E,$1F,$83,$F1,$7C,$1C,$9F,$03
	.dc.b	$E5,$F8,$3F,$37,$C1,$79,$FE,$0E,$57,$96,$3A,$3E,$5C,$2E,$CE,$FF
	.dc.b	$F1,$FE,$BF,$D1,$6B,$D9,$F0,$7F,$07,$E0,$E4,$60,$F8,$1E,$D7,$C5
	.dc.b	$F7,$E4,$60,$F4,$90,$F0,$1F,$1F,$7F,$99,$F0,$52,$3E,$53,$F3,$FE
	.dc.b	$0F,$E8,$7C,$1F,$D1,$F8,$39,$5F,$3E,$3E,$A7,$C1,$FD,$5F,$83,$93
	.dc.b	$F4,$E3,$EC,$7C,$1F,$F9,$57,$DC,$FF,$35,$CF,$F6,$86,$F1,$7C,$18
	.dc.b	$FB,$93,$BE,$A9,$1F,$E2,$0A,$FF,$34,$1F,$CD,$DE,$C7,$E0,$70,$F0
	.dc.b	$BF,$81,$C3,$C1,$F0,$12,$39,$3E,$3F,$81,$F2,$7B,$32,$7B,$3E,$7F
	.dc.b	$7B,$DB,$1D,$48,$75,$3C,$9F,$01,$ED,$23,$E7,$F8,$1F,$81,$F0,$3F
	.dc.b	$07,$49,$3E,$B8,$F8,$9F,$03,$23,$DC,$FC,$7F,$81,$F9,$1F,$03,$27
	.dc.b	$E1,$3F,$2F,$E0,$7E,$67,$C0,$49,$F8,$CF,$CF,$F8,$0E,$91,$F3,$F0
	.dc.b	$BC,$BF,$3F,$91,$A3,$3F,$AD,$F0,$13,$0F,$A0,$FD,$9F,$80,$FB,$5E
	.dc.b	$D7,$DB,$F6,$A5,$7D,$58,$EB,$E2,$7F,$A6,$31,$FF,$3E,$CF,$B0,$FB
	.dc.b	$58,$FF,$A5,$FF,$2D,$FE,$70,$7F,$3B,$8F,$F5,$33,$73,$3A,$F8,$1E
	.dc.b	$1F,$03,$2B,$FC,$C7,$3F,$CF,$DB,$F0,$3D,$7E,$06,$46,$0D,$FC,$07
	.dc.b	$87,$E0,$64,$60,$E7,$F0,$3E,$52,$3D,$BE,$07,$D1,$87,$48,$FF,$20
	.dc.b	$E2,$C7,$31,$E4,$F8,$1F,$71,$1F,$3F,$C0,$FC,$1F,$5F,$E1,$7B,$12
	.dc.b	$7D,$71,$C9,$EB,$7D,$FA,$3E,$C6,$2C,$BE,$57,$C0,$4B,$F8,$6F,$CD
	.dc.b	$F8,$1F,$9D,$F0,$1F,$AD,$C3,$8B,$79,$FF,$D9,$B8,$BE,$07,$18,$FD
	.dc.b	$2A,$BC,$DF,$4B,$E0,$2D,$BE,$02,$C6,$1C,$FB,$F3,$2F,$AB,$F0,$17
	.dc.b	$2F,$F3,$3F,$E5,$DF,$E6,$E7,$FD,$C8,$DC,$8F,$F9,$C2,$37,$5D,$8F
	.dc.b	$BF,$85,$8F,$A2,$F9,$A7,$FC,$E3,$3F,$EE,$86,$CF,$6F,$81,$D7,$C0
	.dc.b	$EF,$DD,$97,$FE,$AC,$BF,$F3,$A3,$63,$B7,$07,$D0,$C6,$CF,$C8,$7C
	.dc.b	$1F,$03,$23,$97,$C0,$DC,$FD,$FF,$01,$FE,$F5,$C9,$F0,$34,$C4,$67
	.dc.b	$E0,$35,$F0,$12,$FC,$91,$C7,$E0,$39,$7C,$04,$9F,$44,$75,$F7,$FB
	.dc.b	$7B,$F2,$7D,$91,$D5,$F7,$61,$D5,$FA,$78,$79,$7D,$E9,$1F,$0D,$F4
	.dc.b	$7B,$FE,$9F,$76,$4F,$C7,$7D,$9E,$F7,$58,$FB,$B8,$75,$E4,$FE,$1D
	.dc.b	$C0,$CB,$E1,$7B,$EB,$FF,$6D,$FE,$5F,$FD,$0B,$93,$1D,$48,$C4,$FE
	.dc.b	$1A,$B8,$3F,$EB,$08,$4E,$31,$FE,$39,$18,$9E,$B3,$9B,$EB,$3F,$D2
	.dc.b	$67,$97,$C0,$65,$E7,$08,$C5,$78,$DB,$1F,$E8,$39,$F1,$7B,$F8,$F7
	.dc.b	$E4,$7B,$23,$7E,$FF,$0F,$7E,$4F,$C0,$8E,$78,$48,$F8,$6F,$6F,$80
	.dc.b	$EF,$F0,$1E,$0F,$5E,$57,$C8,$8F,$1F,$B9,$23,$E5,$BE,$6F,$63,$A4
	.dc.b	$7B,$8E,$A7,$D0,$BF,$5F,$C0,$7B,$3E,$02,$4F,$D4,$8F,$7F,$C0,$7C
	.dc.b	$0F,$80,$93,$F6,$23,$E1,$FC,$04,$8F,$B6,$FC,$6F,$80,$F8,$FE,$FF
	.dc.b	$C8,$F8,$09,$5F,$80,$8F,$9B,$D5,$FE,$79,$24,$63,$E2,$3A,$3F,$7F
	.dc.b	$0B,$D5,$F7,$B1,$E3,$F0,$3C,$BE,$06,$5F,$E3,$6B,$FA,$F6,$3F,$D1
	.dc.b	$63,$FD,$EE,$FF,$29,$75,$D9,$F9,$08,$C4,$6B,$D8,$F2,$FB,$3E,$0A
	.dc.b	$FC,$4E,$76,$BF,$0F,$9F,$AB,$E0,$64,$74,$3F,$2B,$8A,$3D,$FF,$03
	.dc.b	$F0,$3E,$07,$F9,$6C,$BE,$06,$C6,$1C,$FC,$0C,$CF,$A7,$87,$FA,$BF
	.dc.b	$81,$C6,$7E,$CF,$80,$E6,$74,$F8,$1B,$7C,$33,$7C,$0F,$80,$F0,$FC
	.dc.b	$0D,$9D,$67,$D3,$93,$E4,$E7,$67,$E3,$7C,$07,$A0,$F4,$FC,$0D,$8E
	.dc.b	$DF,$01,$EC,$F6,$FD,$AD,$8E,$A4,$DF,$E9,$E1,$D5,$F8,$FF,$01,$F1
	.dc.b	$3E,$03,$E2,$B6,$BE,$0F,$E6,$1F,$D1,$CF,$E8,$DA,$7F,$86,$DC,$7C
	.dc.b	$67,$F5,$87,$FA,$0F,$21,$A3,$E0,$B5,$F0,$5B,$F8,$29,$8F,$6F,$84
	.dc.b	$BA,$B3,$ED,$F0,$7D,$4E,$DF,$07,$6F,$CB,$8F,$47,$CF,$95,$C3,$94
	.dc.b	$7F,$1E,$CD,$F0,$5E,$7F,$83,$F4,$36,$7D,$FD,$5E,$BF,$6F,$D8,$D8
	.dc.b	$EE,$4C,$3A,$9F,$C9,$30,$F7,$46,$75,$F8,$DF,$05,$F1,$4F,$8D,$F0
	.dc.b	$76,$7E,$6F,$C1,$7C,$93,$E5,$7C,$1D,$8F,$A3,$F0,$7F,$34,$F9,$DF
	.dc.b	$07,$63,$EA,$FC,$1F,$D1,$3E,$97,$C1,$D8,$FB,$3F,$05,$FE,$0C,$FC
	.dc.b	$C5,$F1,$5F,$5E,$30,$F3,$E0,$F5,$F5,$FD,$9C,$FB,$E1,$E2,$3F,$47
	.dc.b	$87,$F4,$26,$CF,$5F,$8A,$EC,$F7,$F8,$0F,$07,$B9,$6B,$8B,$DF,$8F
	.dc.b	$F1,$58,$F6,$CB,$C3,$C8,$8F,$47,$C1,$CC,$3C,$3F,$09,$6F,$CF,$AB
	.dc.b	$E0,$FD,$C7,$BF,$E1,$2D,$FA,$3E,$0F,$E1,$7C,$27,$C3,$6C,$7B,$3E
	.dc.b	$13,$E3,$1F,$1F,$E1,$2C,$7C,$0F,$83,$F9,$47,$CB,$F8,$4B,$1D,$4F
	.dc.b	$8B,$87,$56,$9F,$23,$E0,$FE,$97,$C2,$7D,$36,$C7,$CC,$F7,$7E,$B1
	.dc.b	$F5,$FD,$DB,$1D,$78,$3D,$F8,$75,$6C,$7D,$4F,$83,$FB,$DF,$09,$F7
	.dc.b	$DB,$1F,$C6,$BF,$32,$FF,$01,$36,$BF,$C9,$F0,$F8,$5B,$18,$F8,$59
	.dc.b	$5C,$7E,$16,$CF,$E1,$AF,$FC,$C9,$3D,$7E,$13,$9F,$C2,$F4,$7A,$FC
	.dc.b	$2F,$6F,$85,$B5,$E3,$79,$F7,$FF,$A9,$70,$67,$E4,$CF,$37,$FE,$16
	.dc.b	$97,$C2,$FA,$3E,$17,$D2,$D9,$F7,$7C,$0F,$B0,$F6,$FB,$B6,$3E,$16
	.dc.b	$3D,$37,$FF,$0D,$E2,$CB,$1E,$32,$1F,$0B,$F1,$BE,$17,$E3,$B7,$1F
	.dc.b	$99,$F0,$BF,$28,$F9,$7F,$0B,$63,$E8,$7C,$2F,$CE,$3E,$7F,$C2,$D8
	.dc.b	$FA,$9F,$09,$F4,$8F,$A7,$F0,$B6,$3E,$C7,$C2,$7D,$63,$EB,$FC,$2D
	.dc.b	$8E,$A4,$E0,$FA,$98,$75,$7F,$77,$C5,$FE,$7F,$27,$C2,$FF,$A3,$9C
	.dc.b	$7C,$0D,$F8,$1C,$37,$0E,$9D,$3F,$57,$6F,$4C,$60,$FF,$24,$DB,$E1
	.dc.b	$7A,$7C,$2F,$56,$CF,$F6,$A9,$7C,$2F,$87,$E1,$7C,$4D,$9D,$7C,$27
	.dc.b	$94,$F3,$7C,$2D,$8E,$5F,$09,$E9,$3D,$5F,$0B,$63,$B7,$C2,$FF,$92
	.dc.b	$B2,$34,$C7,$C2,$E7,$E1,$75,$F0,$B3,$97,$CD,$1C,$7D,$CE,$5F,$0B
	.dc.b	$27,$D5,$1D,$7E,$16,$47,$B5,$F0,$7C,$2F,$87,$E1,$7F,$CB,$7F,$99
	.dc.b	$FF,$9B,$3C,$63,$76,$4E,$A6,$61,$FE,$10,$AF,$F3,$63,$FC,$D1,$FF
	.dc.b	$35,$74,$B8,$3F,$CF,$31,$FE,$69,$9E,$1F,$07,$E0,$3C,$3F,$09,$E2
	.dc.b	$F8,$39,$3D,$5F,$2F,$C1,$F9,$BE,$0A,$4F,$85,$F4,$FC,$17,$AB,$E0
	.dc.b	$A4,$F9,$5F,$6F,$B5,$EE,$F8,$29,$3E,$97,$E0,$FC,$17,$C2,$F8,$2F
	.dc.b	$87,$F0,$52,$BD,$D1,$F1,$BE,$07,$E3,$FC,$14,$9F,$85,$1F,$2B,$E0
	.dc.b	$BE,$5F,$C1,$49,$F8,$D1,$F3,$BE,$07,$E7,$FC,$14,$9F,$95,$1F,$E8
	.dc.b	$CF,$CD,$3F,$D3,$71,$FD,$79,$FE,$40,$FF,$8F,$9A,$E4,$E1,$ED,$C7
	.dc.b	$4F,$3F,$6F,$A7,$C1,$49,$E4,$F7,$F8,$2F,$07,$C1,$78,$7E,$0A,$57
	.dc.b	$78,$E8,$F8,$F0,$BD,$DF,$76,$3F,$D5,$F9,$7F,$15,$7C,$D8,$77,$7D
	.dc.b	$58,$F7,$7D,$5F,$03,$EF,$23,$ED,$F8,$2E,$BE,$07,$FB,$03,$0F,$F3
	.dc.b	$B6,$3D,$FC,$14,$62,$75,$B7,$5C,$3E,$46,$77,$0E,$16,$F8,$2F,$9D
	.dc.b	$E2,$4B,$F9,$8F,$D1,$F8,$0F,$A5,$F0,$3F,$4F,$E0,$65,$7D,$18,$FA
	.dc.b	$DF,$03,$23,$EA,$3F,$67,$D9,$FB,$5E,$EF,$DB,$F8,$19,$5F,$66,$3A
	.dc.b	$BF,$6F,$0E,$AF,$DF,$C3,$A3,$F8,$1C,$2E,$BE,$AC,$7F,$0D,$F0,$3F
	.dc.b	$87,$F8,$19,$7F,$85,$8E,$AF,$E2,$71,$EB,$7F,$0B,$1F,$34,$FA,$08
	.dc.b	$7E,$3F,$0F,$C9,$7C,$0C,$8F,$C8,$3F,$EE,$0F,$C3,$BF,$01,$1F,$E3
	.dc.b	$25,$F0,$59,$F8,$19,$3F,$E9,$49,$CB,$7E,$0B,$91,$1D,$7C,$0F,$5F
	.dc.b	$82,$91,$C5,$F0,$7C,$0F,$87,$1E,$B3,$E1,$C7,$0F,$23,$B1,$0F,$35
	.dc.b	$FA,$7E,$0A,$47,$95,$F6,$7C,$17,$28,$E8,$7A,$6B,$1B,$FE,$47,$4F
	.dc.b	$E9,$53,$D6,$31,$A7,$C5,$C7,$E3,$7C,$14,$AF,$84,$FC,$9F,$82,$F9
	.dc.b	$5F,$05,$27,$E3,$3F,$37,$E0,$BE,$77,$C1,$7C,$FF,$81,$95,$F2,$E3
	.dc.b	$E9,$7C,$14,$8F,$9C,$FD,$5F,$81,$E9,$1F,$57,$0B,$AF,$A7,$1E,$B3
	.dc.b	$CC,$87,$DB,$F8,$2F,$BD,$F0,$52,$3E,$D3,$F8,$1F,$82,$E7,$E3,$A5
	.dc.b	$E4,$38,$D9,$EA,$73,$1F,$D7,$9F,$C2,$63,$F8,$B2,$3F,$87,$F8,$2E
	.dc.b	$8F,$F9,$FB,$D9,$FF,$41,$E3,$FE,$DE,$7F,$CF,$91,$9F,$88,$FE,$25
	.dc.b	$B7,$F0,$59,$CB,$7E,$23,$37,$E7,$D1,$49,$CE,$1E,$A8,$7C,$16,$2B
	.dc.b	$75,$65,$F0,$77,$E0,$F0,$DC,$7C,$DF,$07,$B3,$87,$C1,$D8,$F5,$61
	.dc.b	$CC,$E9,$95,$8F,$77,$C1,$F7,$F8,$3F,$03,$63,$E1,$7C,$1F,$8F,$4F
	.dc.b	$23,$63,$A9,$36,$F9,$58,$75,$7E,$67,$C1,$FA,$CF,$67,$C1,$5A,$F9
	.dc.b	$DF,$07,$EF,$F8,$2F,$80,$D8,$FA,$5F,$07,$F0,$FE,$0B,$E2,$36,$3E
	.dc.b	$B7,$C1,$FC,$73,$E4,$7C,$15,$8F,$B5,$F0,$7F,$37,$F3,$5F,$F3,$FE
	.dc.b	$BF,$7E,$30,$F4,$9F,$C0,$63,$63,$F0,$8F,$37,$FD,$87,$5F,$E8,$59
	.dc.b	$F0,$9F,$95,$E9,$FE,$91,$19,$9D,$FC,$17,$84,$F1,$7B,$F6,$FD,$A7
	.dc.b	$80,$B3,$E0,$7A,$1F,$8B,$AE,$97,$DB,$E1,$FC,$8D,$9D,$D7,$E9,$AB
	.dc.b	$FD,$26,$FA,$3E,$03,$E1,$7C,$1F,$C3,$7E,$27,$C2,$7C,$5F,$85,$B8
	.dc.b	$72,$F0,$18,$A7,$E5,$7C,$2C,$AD,$4B,$E1,$71,$9F,$C6,$F8,$4F,$A0
	.dc.b	$7D,$1F,$84,$B7,$C3,$3F,$03,$4E,$79,$72,$DB,$E5,$E7,$F6,$72,$C5
	.dc.b	$7C,$3F,$84,$FB,$9E,$F7,$DD,$6C,$73,$D3,$EB,$E3,$EC,$6C,$7A,$30
	.dc.b	$E7,$1F,$1F,$0B,$B7,$C0,$F7,$FF,$11,$F0,$5F,$89,$6E,$1F,$03,$E0
	.dc.b	$3F,$18,$7E,$37,$E0,$EC,$7C,$4F,$84,$FC,$87,$C1,$FE,$45,$B1,$F2
	.dc.b	$3E,$13,$F2,$87,$E5,$7E,$12,$C7,$52,$6B,$F9,$0C,$3A,$BF,$92,$BE
	.dc.b	$6D,$9F,$A7,$87,$F0,$BF,$CD,$BF,$CC,$91,$F6,$DF,$ED,$2D,$35,$F0
	.dc.b	$7D,$39,$3F,$33,$A7,$AE,$7C,$31,$F9,$CB,$EE,$FE,$7B,$1B,$3F,$C7
	.dc.b	$F0,$7C,$0F,$84,$F1,$7C,$0D,$9F,$17,$C0,$F9,$7E,$17,$CC,$D8,$F3
	.dc.b	$7C,$2F,$A4,$F5,$7C,$2D,$8F,$57,$C2,$FB,$7E,$17,$DC,$D8,$EA,$7B
	.dc.b	$F0,$EA,$D3,$A9,$38,$5B,$7F,$8F,$F5,$63,$DE,$73,$5E,$DF,$85,$F9
	.dc.b	$9E,$F7,$CD,$6C,$78,$27,$67,$F5,$6B,$63,$E2,$8C,$F4,$A7,$35,$E3
	.dc.b	$CB,$C5,$AF,$C0,$8C,$1F,$37,$E0,$7E,$E1,$F7,$7E,$12,$C7,$DD,$F8
	.dc.b	$5F,$C0,$7C,$2F,$E0,$5B,$1F,$81,$F8,$5F,$C2,$7C,$17,$E1,$5B,$1F
	.dc.b	$85,$F8,$5F,$C4,$7C,$2F,$E2,$5B,$1D,$4F,$C5,$61,$D5,$A7,$E3,$6F
	.dc.b	$9E,$FE,$ED,$98,$E3,$6F,$9D,$F0,$9F,$D7,$FF,$37,$FE,$DF,$7B,$8C
	.dc.b	$7C,$2F,$58,$E5,$9D,$BF,$AC,$F3,$CF,$FB,$76,$1F,$DC,$9F,$C9,$67
	.dc.b	$D7,$8B,$B7,$C0,$FF,$7E,$C3,$FC,$F9,$7F,$EC,$5C,$7E,$9B,$63,$33
	.dc.b	$BC,$3A,$FC,$25,$BF,$F0,$8D,$FF,$74,$7F,$93,$35,$E7,$F0,$BB,$C5
	.dc.b	$2F,$85,$C2,$F8,$5A,$62,$33,$F0,$93,$0E,$AF,$0F,$85,$E3,$94,$9F
	.dc.b	$0B,$D3,$2E,$BF,$09,$DB,$E1,$65,$79,$A3,$C3,$F0,$9E,$2F,$84,$93
	.dc.b	$EA,$8E,$8F,$C4,$C2,$F0,$FC,$8C,$7D,$3F,$07,$EA,$F8,$49,$7F,$0A
	.dc.b	$3D,$BF,$09,$23,$E2,$BF,$03,$E0,$7E,$0F,$C2,$49,$F9,$2F,$C4,$F8
	.dc.b	$4F,$8B,$F0,$92,$7E,$6B,$F2,$FC,$7F,$F6,$BE,$9F,$EB,$57,$FD,$AA
	.dc.b	$DF,$82,$D1,$BF,$84,$ED,$7F,$7B,$1B,$5F,$82,$C7,$A4,$F7,$3F,$07
	.dc.b	$8E,$1F,$EC,$58,$FF,$B3,$DF,$DC,$B6,$BC,$5F,$09,$E5,$F8,$4E,$BD
	.dc.b	$9F,$94,$CA,$E1,$F8,$8C,$BA,$3D,$8F,$69,$D8,$F7,$7C,$24,$7C,$2F
	.dc.b	$C1,$F8,$1F,$85,$EF,$C9,$F2,$BF,$17,$DF,$F8,$DE,$FC,$9F,$4B,$F2
	.dc.b	$7E,$0F,$E5,$7C,$1F,$CB,$F8,$29,$5E,$E8,$E8,$FB,$B0,$BC,$1F,$3F
	.dc.b	$1E,$93,$F0,$B1,$B3,$F4,$F1,$FA,$DF,$07,$2B,$E4,$BF,$67,$E0,$B9
	.dc.b	$F0,$7F,$99,$78,$FF,$C1,$95,$EA,$3E,$7E,$5E,$77,$D2,$43,$E8,$E1
	.dc.b	$F8,$3F,$83,$FC,$27,$BF,$27,$EB,$C7,$E1,$FE,$06,$47,$DA,$7F,$15
	.dc.b	$EF,$FE,$2F,$E0,$3F,$19,$F0,$72,$BE,$FC,$7E,$3F,$E0,$FF,$75,$F9
	.dc.b	$C7,$FB,$4B,$85,$BF,$F9,$F6,$30,$FA,$E7,$BD,$FE,$1B,$2B,$7F,$97
	.dc.b	$CB,$C1,$3F,$8C,$CB,$C0,$FE,$33,$2E,$FF,$09,$E0,$F8,$49,$3F,$E3
	.dc.b	$3D,$3F,$DB,$AF,$53,$D8,$43,$F2,$B8,$76,$7F,$2F,$87,$6C,$7F,$51
	.dc.b	$8F,$B3,$E1,$25,$75,$7D,$FF,$09,$F0,$3E,$03,$E0,$FC,$04,$AF,$14
	.dc.b	$7C,$4F,$82,$F8,$BF,$07,$27,$CD,$1F,$23,$E1,$3E,$4F,$C2,$49,$F5
	.dc.b	$47,$57,$D9,$87,$57,$FB,$2E,$FF,$DA,$1E,$67,$C1,$C3,$B3,$AC,$3E
	.dc.b	$AF,$4F,$D6,$F8,$49,$3F,$1E,$3B,$3D,$30,$ED,$BF,$DA,$8F,$F6,$5E
	.dc.b	$3D,$DE,$87,$80,$A7,$93,$E1,$3F,$07,$F0,$92,$3E,$A3,$F8,$6F,$7B
	.dc.b	$F0,$FF,$03,$F8,$8F,$7E,$57,$D9,$8F,$C5,$FC,$1F,$E3,$3E,$12,$4F
	.dc.b	$DD,$8F,$C7,$FC,$1C,$8F,$C0,$3F,$D2,$FF,$39,$7C,$27,$69,$FC,$26
	.dc.b	$3D,$9F,$93,$8F,$67,$A1,$0F,$F0,$8D,$7F,$BA,$9F,$F3,$F3,$FE,$D5
	.dc.b	$6F,$C2,$77,$3C,$1F,$0F,$DE,$3E,$AD,$F7,$E0,$F9,$5F,$0A,$E1,$FE
	.dc.b	$FF,$5D,$E7,$30,$D7,$C9,$3E,$2A,$F6,$7B,$92,$3C,$6F,$BF,$DC,$F8
	.dc.b	$1F,$07,$F0,$7E,$12,$57,$A2,$3E,$27,$C2,$48,$F5,$BF,$1F,$E1,$3E
	.dc.b	$47,$C1,$C9,$F7,$BF,$2F,$E1,$3A,$C7,$C2,$C3,$AB,$F1,$B0,$FA,$1F
	.dc.b	$07,$23,$E3,$BF,$4F,$E1,$3E,$A7,$C1,$C9,$F9,$6F,$D7,$F8,$4E,$B1
	.dc.b	$F4,$30,$EB,$C5,$F2,$39,$BF,$C0,$93,$F4,$BE,$13,$EF,$11,$FA,$BF
	.dc.b	$09,$F8,$1F,$7F,$F3,$17,$C5,$BE,$92,$9C,$AC,$5E,$FE,$A7,$30,$BC
	.dc.b	$56,$FE,$0A,$C9,$CF,$E7,$5F,$EE,$8C,$BD,$11,$8F,$85,$CF,$C2,$DD
	.dc.b	$3F,$D9,$16,$8E,$3F,$0B,$27,$3F,$0B,$75,$FC,$9E,$3E,$03,$BB,$67
	.dc.b	$9E,$3D,$9A,$76,$C3,$CB,$F0,$BE,$6F,$85,$B3,$E2,$79,$5F,$68,$F2
	.dc.b	$B8,$D3,$F3,$3A,$7F,$BB,$27,$E5,$6C,$8F,$54,$78,$2B,$FC,$F3,$18
	.dc.b	$6F,$C1,$FC,$6F,$80,$F8,$ED,$AF,$87,$F0,$7F,$2B,$E0,$7E,$5B,$63
	.dc.b	$E3,$FC,$27,$CE,$F8,$5F,$9E,$D8,$F9,$7F,$0B,$F4,$BE,$17,$E9,$B6
	.dc.b	$3E,$7F,$C2,$7D,$63,$EB,$FC,$2D,$8F,$A7,$F0,$9F,$6B,$E0,$BE,$DB
	.dc.b	$63,$EB,$FC,$0F,$DE,$F8,$2F,$BE,$D8,$FB,$7F,$0B,$F8,$23,$F0,$FA
	.dc.b	$FF,$B3,$38,$BE,$1C,$63,$E1,$73,$F0,$B6,$FF,$29,$CF,$FC,$1B,$0F
	.dc.b	$F6,$2C,$60,$F9,$EB,$E1,$7A,$FC,$2F,$66,$CE,$F2,$FC,$ED,$7F,$AA
	.dc.b	$1F,$F5,$5A,$F7,$BC,$BF,$0B,$E6,$7C,$FF,$05,$2F,$97,$62,$DF,$07
	.dc.b	$9C,$79,$FC,$27,$BB,$E1,$7D,$ED,$C3,$8D,$F8,$76,$F6,$67,$E0,$9C
	.dc.b	$BC,$E3,$E0,$67,$F2,$32,$C5,$7C,$0F,$84,$F9,$67,$CC,$F8,$5B,$1D
	.dc.b	$B6,$F5,$34,$EA,$78,$AF,$9E,$BF,$13,$0B,$B7,$B7,$E1,$7E,$BF,$C2
	.dc.b	$FD,$86,$E1,$EA,$F8,$5F,$B7,$F0,$5F,$71,$B1,$EE,$F8,$0F,$BE,$7E
	.dc.b	$03,$DF,$B1,$F0,$BE,$17,$F0,$7F,$0B,$F8,$46,$C7,$53,$E3,$E1,$D5
	.dc.b	$A7,$52,$6B,$F6,$AF,$AB,$FE,$E6,$FC,$EF,$FD,$8B,$C7,$FD,$39,$FF
	.dc.b	$5F,$47,$F1,$E7,$35,$FB,$87,$3B,$17,$C2,$FF,$1E,$5F,$0B,$63,$0E
	.dc.b	$7E,$12,$5F,$E1,$A9,$F0,$98,$C7,$C9,$F0,$1C,$FD,$CE,$8D,$BE,$19
	.dc.b	$74,$69,$CF,$2E,$5A,$7A,$F3,$F2,$65,$8A,$F5,$7C,$27,$9C,$F4,$7C
	.dc.b	$25,$8F,$27,$C2,$7A,$CF,$67,$C2,$58,$F4,$7C,$27,$BC,$F8,$1F,$09
	.dc.b	$63,$D9,$F0,$9F,$0F,$E1,$3E,$23,$63,$A9,$38,$BE,$96,$1D,$5F,$A9
	.dc.b	$87,$37,$EB,$61,$76,$FA,$DF,$09,$F3,$CF,$A1,$F0,$97,$1E,$93,$F6
	.dc.b	$F4,$F4,$C6,$0F,$B5,$8F,$37,$EE,$61,$71,$FC,$07,$C1,$FD,$BF,$84
	.dc.b	$FB,$8D,$AF,$B1,$F0,$3F,$D7,$3F,$3C,$FE,$B1,$1F,$85,$8C,$5E,$7E
	.dc.b	$0A,$4F,$F0,$4A,$FE,$79,$A7,$F8,$71,$FE,$89,$5F,$DB,$17,$C2,$75
	.dc.b	$F8,$4E,$CD,$BF,$09,$CE,$BF,$21,$8F,$82,$3B,$63,$E0,$96,$3D,$FE
	.dc.b	$13,$0F,$87,$E1,$3D,$5F,$09,$EB,$6E,$1E,$7F,$82,$F7,$1E,$FF,$83
	.dc.b	$B1,$EB,$F8,$2F,$85,$F0,$7F,$0D,$B1,$EF,$F8,$3F,$8C,$7C,$7F,$52
	.dc.b	$C7,$52,$6B,$E6,$7F,$B2,$F1,$31,$EF,$39,$3E,$36,$1D,$9C,$78,$CA
	.dc.b	$4E,$76,$F9,$DF,$05,$F5,$CF,$B1,$F0,$77,$2F,$A9,$F0,$7F,$6F,$E0
	.dc.b	$FE,$E3,$63,$EC,$7C,$17,$DF,$F8,$3F,$C0,$36,$3E,$E7,$C1,$7E,$0C
	.dc.b	$FC,$27,$C1,$D8,$EA,$7E,$07,$0E,$AD,$3A,$9F,$85,$BE,$AD,$3F,$11
	.dc.b	$F0,$7F,$E2,$AF,$CF,$7F,$B2,$6B,$F8,$98,$C4,$F4,$E1,$F8,$7C,$0C
	.dc.b	$71,$7F,$8F,$7D,$13,$AC,$7B,$B6,$FF,$AC,$E3,$FE,$C2,$BF,$F3,$8A
	.dc.b	$F8,$3F,$F2,$15,$7D,$08,$C3,$C3,$E1,$33,$5C,$3E,$0E,$E1,$FE,$20
	.dc.b	$B7,$C2,$63,$0F,$F2,$5C,$61,$F0,$D6,$79,$53,$FE,$A9,$C3,$C7,$F0
	.dc.b	$59,$8F,$EE,$97,$E8,$F8,$0C,$7C,$04,$8F,$5B,$BF,$80,$E1,$97,$1C
	.dc.b	$E5,$78,$0E,$44,$3D,$D7,$DF,$E0,$FC,$1F,$09,$E1,$F8,$49,$5F,$22
	.dc.b	$3C,$9F,$09,$E5,$F8,$39,$3F,$32,$3D,$1F,$09,$E9,$F8,$49,$3F,$42
	.dc.b	$3D,$9F,$03,$ED,$F8,$49,$3F,$52,$3A,$BF,$5B,$0E,$AF,$D3,$C3,$AB
	.dc.b	$F6,$B0,$EA,$FD,$7C,$3E,$47,$C1,$48,$FB,$EF,$CB,$F8,$3F,$99,$F0
	.dc.b	$92,$7F,$06,$FD,$3F,$57,$FD,$8B,$1F,$E1,$E7,$FD,$7C,$FE,$C9,$BC
	.dc.b	$FF,$EB,$86,$F4,$FF,$98,$1F,$F6,$14,$EF,$E0,$3A,$11,$E9,$F0,$1E
	.dc.b	$17,$F1,$D8,$5C,$3A,$7C,$47,$8F,$B2,$57,$8D,$F3,$7C,$17,$9F,$E0
	.dc.b	$A4,$F9,$DF,$57,$C1,$F3,$D7,$F2,$99,$5A,$FF,$4E,$EB,$EE,$7D,$98
	.dc.b	$76,$7D,$D8,$F6,$7E,$17,$C2,$7C,$5F,$84,$91,$83,$E3,$7C,$27,$C9
	.dc.b	$23,$F2,$7D,$FF,$99,$F0,$52,$3E,$63,$F3,$F2,$FA,$1F,$01,$F4,$7E
	.dc.b	$0E,$57,$D1,$8E,$6F,$D1,$C2,$F1,$7D,$1C,$7A,$4F,$D5,$C6,$CF,$D5
	.dc.b	$C7,$EE,$7C,$27,$DD,$F8,$49,$7F,$76,$3F,$01,$F0,$92,$3F,$00,$FE
	.dc.b	$0F,$E0,$FF,$09,$F0,$9F,$85,$F8,$29,$5F,$85,$8F,$DD,$7E,$7D,$FC
	.dc.b	$22,$3F,$9B,$AF,$83,$C9,$A8,$EC,$7E,$2F,$0B,$A7,$E2,$F1,$EF,$3F
	.dc.b	$91,$C7,$BB,$F9,$1C,$7B,$FC,$1C,$8F,$F3,$9E,$5F,$EC,$76,$FC,$27
	.dc.b	$93,$E1,$64,$69,$F3,$FC,$27,$A3,$E1,$7D,$3F,$0B,$2B,$9C,$7B,$3E
	.dc.b	$0B,$DB,$F0,$52,$7B,$C7,$47,$DF,$85,$DD,$DF,$FE,$C2,$D1,$9F,$7B
	.dc.b	$F3,$E7,$DD,$F8,$F9,$FC,$AF,$85,$91,$EC,$7E,$6F,$C2,$F5,$C7,$A6
	.dc.b	$76,$BF,$DE,$F9,$76,$7D,$27,$62,$1D,$8F,$8F,$F0,$BF,$63,$E1,$64
	.dc.b	$60,$F9,$7F,$0B,$F7,$08,$FC,$EF,$85,$FB,$FF,$01,$F8,$0F,$85,$93
	.dc.b	$F4,$A3,$F0,$7F,$03,$F8,$4F,$85,$93,$F5,$A3,$9B,$F8,$6C,$2F,$1F
	.dc.b	$C8,$8F,$5E,$3D,$9F,$BB,$8F,$67,$F1,$B8,$FE,$43,$E1,$7F,$23,$F0
	.dc.b	$B2,$7F,$07,$1F,$D3,$7F,$3F,$E9,$1F,$E0,$A6,$E3,$DF,$CA,$46,$38
	.dc.b	$BF,$DF,$2F,$F6,$56,$FC,$2F,$3F,$80,$98,$7F,$B0,$AB,$FB,$2B,$D7
	.dc.b	$E0,$3C,$04,$75,$F0,$1E,$3C,$24,$71,$7C,$DF,$0B,$E7,$C7,$AC,$F9
	.dc.b	$F1,$C3,$DF,$E1,$7D,$9F,$09,$23,$07,$8F,$E1,$3E,$01,$1F,$37,$C2
	.dc.b	$7C,$3F,$83,$91,$E9,$7E,$37,$C2,$7C,$7F,$81,$F9,$1F,$05,$2B,$DD
	.dc.b	$1C,$DF,$95,$85,$E3,$F9,$5C,$4C,$BA,$D7,$CF,$CA,$D6,$B2,$FA,$9F
	.dc.b	$03,$2B,$E5,$3F,$5F,$E0,$BA,$71,$7D,$2C,$ED,$39,$7E,$BE,$7D,$2B
	.dc.b	$EB,$E7,$6B,$5A,$33,$FC,$17,$C1,$C9,$FB,$0F,$E1,$7E,$13,$F0,$DF
	.dc.b	$09,$27,$EE,$3F,$89,$F8,$4F,$C5,$7B,$F2,$7F,$00,$FF,$86,$7F,$40
	.dc.b	$FE,$95,$C5,$F8,$B7,$FB,$1C,$F9,$DF,$EC,$71,$9F,$80,$E1,$C1,$DE
	.dc.b	$BF,$29,$8D,$AF,$97,$8F,$79,$FC,$C6,$36,$7B,$7C,$1F,$87,$E1,$65
	.dc.b	$0F,$F3,$3E,$7F,$ED,$67,$15,$E5,$F8,$5F,$47,$C2,$FA,$7E,$16,$56
	.dc.b	$B9,$39,$6C,$C7,$FD,$B3,$39,$3D,$9F,$0B,$F0,$48,$E6,$78,$7C,$2F
	.dc.b	$C5,$F8,$39,$39,$B4,$7C,$9F,$83,$F1,$CE,$72,$F1,$BE,$AC,$FA,$BE
	.dc.b	$CC,$AE,$BF,$3F,$2F,$A5,$F0,$7F,$4F,$E0,$E5,$F7,$8F,$AD,$F0,$92
	.dc.b	$3C,$4F,$D9,$F8,$5F,$B5,$F0,$9F,$6F,$E1,$65,$79,$E3,$EF,$7C,$27
	.dc.b	$DF,$F8,$49,$3E,$B8,$E8,$FD,$0C,$2E,$9F,$83,$C7,$AC,$FC,$1C,$7A
	.dc.b	$BF,$5B,$1F,$C5,$7C,$1F,$E2,$FE,$12,$4F,$C7,$8E,$7C,$DF,$53,$95
	.dc.b	$9E,$2B,$F1,$D9,$76,$7F,$03,$97,$6C,$FF,$29,$9F,$F9,$BF,$C7,$FE
	.dc.b	$ED,$1F,$C8,$5F,$EE,$D5,$F6,$5A,$FF,$AC,$1D,$E5,$F8,$47,$FB,$74
	.dc.b	$67,$E0,$FA,$1D,$7E,$13,$B7,$BB,$27,$A3,$E1,$F8,$2F,$17,$C0,$F8
	.dc.b	$FE,$0A,$57,$86,$39,$3E,$3C,$3A,$BE,$5C,$3A,$3E,$5C,$2E,$7E,$BC
	.dc.b	$7D,$DF,$07,$EF,$F8,$49,$7E,$D8,$F8,$5F,$09,$F0,$FE,$12,$4F,$C1
	.dc.b	$8F,$8D,$F0,$7F,$1F,$E0,$E4,$FC,$58,$F9,$5F,$03,$23,$E4,$3F,$37
	.dc.b	$E0,$7A,$C7,$CB,$C3,$AB,$F2,$F0,$EA,$FC,$FC,$3A,$BF,$3F,$0F,$AD
	.dc.b	$F0,$12,$3E,$A3,$F6,$7E,$07,$ED,$7C,$14,$9F,$B0,$FD,$DF,$83,$FB
	.dc.b	$DF,$05,$27,$EE,$3F,$81,$F8,$3F,$F5,$87,$E8,$3F,$D7,$E7,$F9,$63
	.dc.b	$7C,$C3,$4F,$EC,$51,$FC,$C7,$A9,$8F,$86,$3F,$0C,$F8,$5E,$E4,$3F
	.dc.b	$15,$87,$85,$FE,$0D,$5F,$DA,$39,$7F,$B9,$37,$AF,$F8,$8B,$67,$CB
	.dc.b	$F0,$1D,$EB,$F3,$18,$5B,$FC,$A6,$9D,$6F,$B9,$E1,$AF,$CF,$69,$F0
	.dc.b	$3E,$12,$4F,$89,$F8,$7F,$09,$F1,$3E,$0B,$E2,$FC,$14,$AF,$3C,$7C
	.dc.b	$8F,$7E,$47,$A9,$F9,$7E,$FF,$CC,$F7,$E4,$FB,$9F,$9F,$EF,$75,$8F
	.dc.b	$83,$87,$55,$A7,$C2,$C7,$AC,$63,$6F,$87,$96,$2B,$E3,$E5,$F6,$B8
	.dc.b	$FE,$DF,$C1,$4C,$3E,$6C,$7D,$EF,$82,$91,$F4,$1F,$C0,$FC,$17,$E0
	.dc.b	$BE,$07,$F0,$7F,$03,$2B,$EA,$C7,$E1,$BE,$02,$47,$D8,$7F,$13,$F0
	.dc.b	$1D,$23,$CC,$7E,$2B,$0C,$69,$F7,$31,$ED,$3F,$80,$C7,$B3,$C3,$AF
	.dc.b	$F1,$F9,$7E,$53,$E0,$24,$FE,$19,$FE,$B8,$EF,$F4,$2F,$ED,$FC,$3F
	.dc.b	$D5,$9F,$ED,$F6,$69,$3E,$0B,$8E,$7F,$E3,$F8,$FE,$C9,$7C,$FE,$03
	.dc.b	$38,$73,$6C,$7F,$83,$E7,$75,$C7,$DD,$CE,$3D,$FD,$EB,$46,$CE,$26
	.dc.b	$5F,$F6,$5A,$F2,$37,$E0,$78,$E3,$23,$06,$FD,$FE,$A4,$79,$7B,$FD
	.dc.b	$1F,$89,$85,$CF,$DD,$AF,$93,$1E,$95,$EF,$3A,$9F,$2F,$E0,$7D,$24
	.dc.b	$7C,$DF,$03,$EC,$F8,$0F,$6F,$C0,$C9,$F5,$47,$C0,$F7,$3E,$0F,$B9
	.dc.b	$27,$DD,$1F,$13,$DB,$F8,$BE,$E4,$9F,$85,$1D,$5F,$89,$87,$57,$EF
	.dc.b	$61,$D5,$F9,$18,$75,$7F,$05,$87,$D0,$F8,$0F,$F4,$46,$DF,$77,$E0
	.dc.b	$6F,$3F,$E1,$66,$DF,$DB,$9B,$C3,$F8,$7F,$80,$BC,$C6,$FC,$0E,$1D
	.dc.b	$FE,$87,$FE,$2D,$CB,$FC,$3C,$FF,$B7,$5B,$8E,$BE,$23,$7E,$FC,$9D
	.dc.b	$FC,$05,$AF,$CD,$61,$D6,$7F,$09,$8D,$9F,$D2,$63,$DD,$B8,$F8,$9E
	.dc.b	$D5,$E2,$F8,$0F,$2F,$C0,$79,$9B,$1E,$6F,$81,$F4,$FC,$17,$A9,$B1
	.dc.b	$EA,$F8,$1F,$69,$EE,$F8,$2B,$1E,$EF,$7F,$E0,$9F,$0B,$E0,$2C,$75
	.dc.b	$3E,$1E,$1D,$5A,$75,$26,$7E,$CB,$EA,$FB,$BD,$FF,$9B,$F0,$1F,$39
	.dc.b	$B5,$D6,$7E,$1D,$F5,$6C,$74,$E8,$FF,$5D,$70,$34,$F5,$4E,$4F,$F2
	.dc.b	$DC,$3E,$0B,$ED,$9F,$73,$E0,$AC,$FD,$CF,$82,$FB,$E7,$E0,$3E,$0A
	.dc.b	$C7,$E0,$3D,$FF,$C1,$9F,$84,$F7,$EC,$7E,$13,$E0,$BF,$0E,$7E,$23
	.dc.b	$E0,$2C,$75,$26,$DF,$7B,$0E,$AF,$E0,$30,$E6,$FE,$0B,$0B,$B7,$FB
	.dc.b	$5F,$85,$87,$E6,$9A,$FF,$9C,$57,$C3,$E6,$35,$F0,$F1,$D7,$C1,$DC
	.dc.b	$7F,$99,$5B,$E0,$FA,$1D,$7E,$0E,$DF,$5F,$80,$F0,$1E,$1F,$81,$B1
	.dc.b	$E1,$CB,$A6,$9E,$7C,$EC,$FF,$A3,$32,$EC,$F9,$F2,$EC,$D3,$D5,$F0
	.dc.b	$3E,$DF,$81,$F7,$36,$3D,$DF,$07,$F0,$7E,$0F,$E1,$36,$3E,$17,$C1
	.dc.b	$7C,$5F,$83,$F8,$CD,$8F,$8D,$F0,$3F,$24,$F9,$5F,$03,$63,$A9,$F2
	.dc.b	$F0,$EA,$D3,$A9,$F3,$EF,$AB,$4F,$A5,$F0,$1F,$57,$E0,$7E,$B3,$63
	.dc.b	$AF,$2F,$D0,$BE,$AD,$8E,$AF,$D4,$BE,$AD,$8F,$BD,$F0,$7F,$81,$F8
	.dc.b	$3F,$C1,$36,$3F,$05,$F0,$7F,$85,$3F,$8F,$FE,$88,$F8,$7C,$4F,$FA
	.dc.b	$DE,$3F,$DD,$8D,$F8,$0D,$FA,$BC,$1E,$3F,$01,$CB,$E0,$ED,$71,$79
	.dc.b	$E9,$F8,$AC,$BB,$E5,$8E,$AF,$CB,$CF,$FB,$E7,$3E,$EF,$F8,$05,$65
	.dc.b	$E0,$C7,$C4,$79,$7E,$13,$D2,$7A,$BE,$16,$C7,$A7,$E1,$7D,$BF,$0B
	.dc.b	$EE,$6C,$7B,$7E,$17,$E0,$FC,$2F,$C2,$6C,$7C,$1F,$82,$F8,$A7,$C6
	.dc.b	$F8,$3B,$1D,$4F,$8D,$87,$56,$9D,$4F,$95,$7D,$5A,$7C,$DF,$82,$FA
	.dc.b	$3F,$07,$F4,$9B,$1F,$47,$E1,$3E,$AF,$C2,$FD,$66,$C7,$D5,$F8,$5F
	.dc.b	$B2,$7D,$AF,$85,$B1,$F6,$7E,$0B,$EE,$9F,$7B,$E0,$AC,$75,$3E,$F6
	.dc.b	$1D,$5A,$75,$27,$2F,$D3,$BE,$AF,$F8,$D7,$F4,$5F,$EA,$5C,$FF,$6B
	.dc.b	$B1,$1F,$D8,$67,$47,$F6,$C7,$13,$BE,$86,$BE,$CA,$FC,$03,$67,$5A
	.dc.b	$F8,$1E,$18,$DC,$96,$CB,$C3,$F0,$BD,$CF,$07,$C2,$DA,$DE,$1E,$7D
	.dc.b	$1F,$07,$E5,$F8,$4F,$33,$6B,$67,$6C,$7D,$BF,$09,$EB,$3D,$9F,$01
	.dc.b	$67,$7A,$78,$F1,$FE,$A1,$18,$BF,$8B,$F0,$1F,$0F,$E0,$BE,$23,$E0
	.dc.b	$9D,$53,$3C,$61,$F3,$DE,$CF,$D1,$CB,$C3,$5C,$1F,$9D,$F0,$9F,$3F
	.dc.b	$E1,$2C,$F4,$7E,$97,$C2,$7D,$3F,$56,$CF,$81,$FA,$DF,$07,$F5,$FE
	.dc.b	$0E,$CF,$91,$FB,$5F,$07,$23,$CF,$F0,$37,$4F,$C2,$63,$D6,$7D,$6D
	.dc.b	$8F,$C4,$61,$D6,$7D,$ED,$8F,$C6,$7C,$0F,$E1,$CF,$C4,$7C,$0D,$AF
	.dc.b	$8B,$F0,$7F,$8B,$3F,$19,$F0,$76,$3E,$77,$E8,$DF,$D8,$30,$FF,$45
	.dc.b	$3F,$D8,$23,$27,$F8,$F9,$C4,$6B,$E0,$30,$67,$E0,$2D,$78,$09,$3F
	.dc.b	$A1,$BF,$03,$DA,$3A,$7C,$27,$56,$9E,$4F,$81,$F0,$1E,$1F,$84,$B1
	.dc.b	$D3,$BB,$F2,$9A,$7B,$23,$07,$E4,$74,$F6,$4F,$EB,$DF,$4B,$63,$BE
	.dc.b	$3F,$03,$E1,$3D,$C7,$BF,$E1,$2C,$FC,$6F,$83,$F8,$5F,$09,$F0,$DB
	.dc.b	$1F,$2B,$E1,$3E,$37,$BF,$F1,$DB,$1F,$3B,$E1,$3E,$57,$C0,$7C,$B6
	.dc.b	$C7,$53,$E9,$E1,$D5,$A7,$53,$EB,$DF,$56,$9F,$6B,$E1,$3E,$B1,$F5
	.dc.b	$FE,$12,$C7,$DE,$F8,$3F,$B5,$F0,$9F,$6D,$B1,$F8,$2F,$83,$FB,$C7
	.dc.b	$DF,$F8,$4B,$1F,$86,$F8,$2F,$C1,$1F,$E8,$0F,$5F,$FB,$14,$64,$EC
	.dc.b	$7E,$33,$0E,$CD,$3B,$13,$8F,$F1,$F7,$D9,$FF,$6D,$57,$F5,$CE,$4F
	.dc.b	$E4,$CF,$F5,$CA,$FE,$08,$E5,$FE,$1C,$E6,$78,$FC,$27,$94,$F3,$7C
	.dc.b	$25,$AE,$3F,$05,$E9,$F8,$2F,$53,$63,$AF,$C1,$7B,$7D,$FF,$73,$63
	.dc.b	$C3,$87,$3C,$7C,$F8,$DC,$FC,$F8,$F3,$9F,$5B,$63,$D7,$F0,$5F,$27
	.dc.b	$E0,$3E,$53,$71,$EA,$4B,$EA,$DA,$EA,$FC,$2D,$3D,$2D,$3E,$2F,$C1
	.dc.b	$7D,$53,$EB,$7C,$15,$8F,$93,$BF,$D9,$F8,$2F,$B4,$D8,$F9,$BE,$F7
	.dc.b	$DD,$F8,$0F,$BC,$D8,$FA,$3E,$FF,$E0,$4F,$C1,$7B,$F6,$3A,$9F,$5B
	.dc.b	$0E,$AD,$3A,$9F,$6A,$FA,$B4,$FB,$BF,$05,$F8,$D3,$F1,$DF,$05,$63
	.dc.b	$FA,$27,$E8,$FF,$DA,$9F,$E9,$CF,$F6,$66,$F0,$7F,$B5,$23,$0F,$F5
	.dc.b	$F5,$F0,$59,$35,$F0,$56,$BA,$FC,$17,$1F,$77,$93,$63,$C3,$87,$5F
	.dc.b	$87,$EC,$D8,$F2,$E1,$CE,$1C,$BF,$DF,$F2,$65,$98,$C7,$BD,$E7,$3D
	.dc.b	$1F,$07,$71,$F7,$FC,$0F,$AF,$E0,$BD,$8D,$8F,$87,$F0,$3E,$FF,$81
	.dc.b	$F8,$0D,$8F,$8F,$EE,$FC,$3F,$81,$F8,$8D,$8E,$A7,$CC,$C3,$AB,$4E
	.dc.b	$A7,$D0,$BE,$AD,$3E,$9F,$BD,$F3,$CF,$A1,$F0,$36,$3A,$F0,$7C,$27
	.dc.b	$EB,$DF,$68,$C1,$F5,$5E,$AE,$3E,$0D,$C2,$C7,$3C,$1F,$2D,$FB,$DF
	.dc.b	$05,$31,$FC,$2F,$C1,$5C,$79,$60,$F8,$2F,$C2,$1F,$85,$F8,$2B,$7F
	.dc.b	$8C,$F7,$3F,$10,$7E,$27,$E0,$2C,$7F,$84,$FF,$48,$FE,$96,$FF,$87
	.dc.b	$5F,$E9,$6D,$3F,$CE,$BA,$FF,$8F,$63,$39,$7E,$47,$1E,$EF,$93,$1B
	.dc.b	$5F,$85,$C7,$BC,$FA,$31,$B3,$CF,$E0,$FA,$E5,$E1,$F8,$3B,$7F,$E3
	.dc.b	$BC,$D5,$8E,$04,$2B,$0F,$FB,$22,$BE,$73,$70,$F4,$7C,$0E,$BE,$0B
	.dc.b	$7F,$03,$C3,$E0,$6E,$9F,$0D,$E7,$F0,$32,$3E,$37,$BD,$73,$EB,$F0
	.dc.b	$3E,$03,$C3,$EF,$DA,$F9,$9F,$01,$E4,$F8,$1F,$2B,$63,$E8,$7C,$0F
	.dc.b	$A3,$E0,$7D,$2D,$8F,$A9,$87,$3C,$7D,$78,$DE,$0F,$A5,$8F,$39,$F7
	.dc.b	$B6,$3E,$B7,$C0,$FC,$43,$E2,$FB,$F7,$1F,$C0,$7C,$0F,$C8,$F8,$0F
	.dc.b	$92,$D8,$FC,$27,$C0,$FC,$FF,$D2,$7F,$B1,$E7,$F3,$23,$1F,$11,$9E
	.dc.b	$AD,$7C,$45,$C3,$FC,$BD,$87,$F5,$9D,$7F,$D0,$8B,$E0,$BA,$1D,$7E
	.dc.b	$0A,$D7,$0F,$82,$F0,$1E,$1F,$80,$B1,$D3,$E0,$FC,$9F,$09,$E5,$6C
	.dc.b	$7F,$9C,$32,$76,$39,$7F,$17,$18,$EA,$B3,$9A,$E0,$A3,$E0,$7C,$1C
	.dc.b	$9F,$5F,$C2,$5E,$4F,$DD,$70,$3E,$13,$E3,$7C,$27,$C7,$6D,$F4,$F0
	.dc.b	$7B,$9E,$0C,$EC,$7C,$6F,$84,$F9,$DF,$03,$F3,$DB,$1F,$2B,$E1,$3E
	.dc.b	$91,$F4,$FE,$0A,$C7,$A0,$F5,$FC,$37,$B1,$A7,$D2,$F8,$5F,$B4,$7D
	.dc.b	$BF,$85,$B1,$F5,$BE,$17,$EF,$1F,$7F,$E1,$6C,$7C,$BF,$B9,$7A,$FF
	.dc.b	$29,$B1,$9F,$5F,$13,$FB,$5E,$7F,$EB,$5C,$BF,$EB,$36,$F8,$3F,$37
	.dc.b	$E1,$35,$3B,$F8,$49,$8F,$FA,$96,$FF,$A5,$B6,$3C,$1F,$01,$23,$C9
	.dc.b	$ED,$DB,$FE,$5F,$FA,$0B,$95,$ED,$7F,$0A,$8C,$7B,$76,$73,$87,$FA
	.dc.b	$DA,$72,$F8,$7E,$1B,$D4,$7A,$FE,$1E,$C7,$C3,$8F,$85,$1F,$0D,$B1
	.dc.b	$CF,$DE,$FE,$57,$92,$FF,$9F,$F4,$78,$E7,$31,$8A,$E1,$EE,$64,$F9
	.dc.b	$5F,$0D,$63,$F3,$9F,$76,$E9,$FC,$DA,$E9,$B1,$87,$B7,$B7,$DF,$DC
	.dc.b	$B7,$E5,$7C,$5E,$17,$8F,$C2,$B3,$E9,$7A,$F8,$5D,$B8,$EC,$F8,$9F
	.dc.b	$57,$B9,$23,$E0,$7B,$97,$8F,$E9,$F1,$3D,$AF,$81,$ED,$7C,$16,$DF
	.dc.b	$ED,$FC,$BF,$F0,$CB,$7B,$FF,$6F,$18,$C3,$FB,$5F,$98,$8F,$F1,$0F
	.dc.b	$3F,$F6,$AF,$37,$E6,$1F,$37,$C1,$B8,$FD,$3F,$6F,$E8,$1F,$47,$DC
	.dc.b	$B1,$F5,$FD,$CF,$A8,$7D,$5F,$72,$C7,$DB,$F6,$FE,$C7,$B9,$F6,$5B
	.dc.b	$1F,$7F,$DC,$FC,$77,$BB,$FC,$9B,$74,$B8,$3F,$05,$C8,$BD,$FB,$BC
	.dc.b	$3C,$B9,$7F,$93,$C7,$F9,$43,$C7,$8B,$FA,$FC,$F6,$F7,$3B,$FB,$96
	.dc.b	$B8,$3E,$2F,$72,$47,$3F,$72,$EB,$FA,$6C,$9E,$DF,$A3,$DC,$F4,$B6
	.dc.b	$FC,$5D,$3E,$C3,$DB,$EE,$58,$F3,$7B,$5F,$03,$DB,$F8,$2D,$8F,$57
	.dc.b	$85,$F1,$3C,$3F,$8A,$D8,$F7,$7B,$7F,$20,$F9,$3E,$DD,$8F,$85,$ED
	.dc.b	$FC,$C3,$E6,$FB,$76,$3E,$37,$9B,$F4,$3C,$DF,$A2,$D8,$FD,$E7,$B9
	.dc.b	$FC,$4B,$A4,$FE,$20,$C1,$7D,$9F,$E0,$CF,$F1,$0E,$5F,$E1,$6F,$E4
	.dc.b	$7B,$BE,$DC,$7E,$EE,$35,$18,$AE,$BE,$A7,$23,$9F,$AB,$71,$F1,$FA
	.dc.b	$7D,$8E,$FF,$0F,$63,$CF,$F0,$FE,$23,$C7,$F0,$76,$3D,$7E,$E7,$9B
	.dc.b	$DC,$F3,$B6,$3D,$FF,$0D,$EA,$F8,$7F,$5B,$63,$E1,$FC,$3F,$E4,$F8
	.dc.b	$7D,$AD,$E0,$F5,$70,$A3,$15,$F9,$18,$7C,$3E,$2F,$E1,$7C,$0D,$E3
	.dc.b	$F9,$BF,$05,$6C,$D9,$C7,$C0,$E7,$E0,$A6,$1F,$5A,$38,$7C,$3C,$8F
	.dc.b	$B2,$F3,$F8,$7E,$DE,$BF,$D2,$9F,$E7,$EB,$E1,$32,$6B,$6D,$FC,$27
	.dc.b	$0F,$87,$97,$FC,$FE,$FF,$C9,$6D,$F8,$4E,$BF,$09,$23,$06,$FE,$13
	.dc.b	$C2,$47,$97,$C2,$79,$3E,$12,$47,$57,$CF,$F0,$9E,$8F,$83,$93,$E1
	.dc.b	$7D,$7F,$09,$EC,$F8,$39,$3E,$57,$DF,$A7,$C0,$F8,$3F,$83,$F0,$72
	.dc.b	$BD,$51,$F1,$3E,$0F,$E2,$FC,$1C,$9F,$74,$7C,$8F,$82,$91,$F0,$5F
	.dc.b	$97,$F0,$7F,$33,$E0,$BE,$D7,$E9,$4F,$83,$A6,$6F,$FB,$34,$3E,$0F
	.dc.b	$64,$8C,$FC,$47,$2F,$84,$E7,$F0,$92,$78,$C7,$6F,$84,$91,$D1,$F0
	.dc.b	$FC,$27,$8B,$E0,$BC,$7F,$07,$2B,$C3,$1E,$6F,$83,$FC,$97,$E8,$EE
	.dc.b	$27,$A7,$86,$DF,$A2,$BD,$BF,$09,$23,$D6,$FC,$0F,$83,$F8,$3F,$09
	.dc.b	$27,$DE,$FC,$4F,$83,$FD,$07,$A6,$F0,$1E,$1E,$1B,$7F,$E5,$16,$BF
	.dc.b	$D6,$17,$89,$F3,$BE,$13,$E7,$FC,$24,$AF,$9B,$1F,$4B,$E1,$24,$7D
	.dc.b	$07,$EA,$FC,$27,$8B,$F0,$1F,$F0,$EC,$FF,$97,$DA,$FF,$66,$6E,$7F
	.dc.b	$EB,$87,$FD,$8D,$1F,$6B,$E1,$7E,$FF,$C2,$48,$C1,$F7,$BE,$17,$F0
	.dc.b	$64,$7F,$C4,$DF,$A5,$FF,$09,$7B,$7B,$BE,$3F,$7A,$47,$8E,$3F,$82
	.dc.b	$4B,$E2,$69,$27,$9F,$C3,$FF,$0B,$F0,$FF,$CE,$1F,$09,$7C,$D3,$4C
	.dc.b	$7C,$27,$C1,$F8,$5F,$85,$13,$0F,$85,$1E,$0F,$86,$91,$E8,$7C,$71
	.dc.b	$D7,$C0,$ED,$F0,$B2,$BF,$CA,$11,$F3,$58,$7C,$D3,$E8,$69,$EC,$9F
	.dc.b	$6C,$C9,$F8,$B1,$E9,$F8,$6F,$56,$F2,$7F,$CB,$91,$F8,$AC,$9D,$34
	.dc.b	$93,$FE,$99,$87,$47,$C3,$E2,$F8,$9F,$0D,$2B,$FC,$CD,$1F,$1F,$E1
	.dc.b	$A4,$7F,$9A,$5F,$95,$F0,$7D,$FC,$4F,$07,$89,$2B,$B4,$7C,$FF,$84
	.dc.b	$FA,$1F,$07,$27,$FC,$E5,$1F,$4F,$E1,$3E,$A7,$C2,$49,$FF,$3B,$47
	.dc.b	$AF,$CF,$91,$E9,$7D,$DE,$87,$DB,$F8,$49,$3F,$E7,$D7,$F0,$1E,$9F
	.dc.b	$F3,$BF,$13,$FB,$6F,$72,$3F,$9E,$57,$F8,$A6,$73,$1F,$CE,$21,$18
	.dc.b	$F8,$2E,$7F,$09,$FB,$6F,$C0,$DE,$D5,$F6,$A6,$1F,$EC,$59,$CB,$D7
	.dc.b	$E1,$7F,$C6,$D9,$F8,$7E,$17,$CB,$F0,$B2,$B4,$7F,$21,$D5,$3E,$AF
	.dc.b	$3E,$4E,$AF,$D5,$1E,$EF,$47,$DF,$F0,$B2,$B5,$87,$FB,$59,$BF,$0B
	.dc.b	$F1,$3D,$39,$1A,$7E,$3F,$C2,$FC,$8F,$85,$93,$C9,$F9,$7F,$0B,$F3
	.dc.b	$3C,$EF,$9B,$F0,$92,$BB,$C7,$D0,$F8,$5F,$A3,$EE,$49,$F1,$C7,$D4
	.dc.b	$F8,$5F,$AB,$CF,$27,$CF,$1F,$63,$9F,$EC,$FC,$2C,$9F,$5C,$7E,$EF
	.dc.b	$D5,$FE,$95,$1F,$0D,$FE,$90,$DE,$5D,$7C,$2E,$FE,$17,$87,$83,$2B
	.dc.b	$84,$77,$7E,$3E,$3D,$DF,$C0,$F5,$FD,$8E,$66,$7D,$27,$C1,$8D,$76
	.dc.b	$7F,$59,$CD,$E0,$F9,$FE,$16,$4F,$9D,$F5,$7C,$2F,$86,$38,$ED,$F9
	.dc.b	$5D,$3C,$5E,$17,$E5,$33,$EC,$FE,$6B,$2F,$83,$7C,$1B,$19,$CB,$E3
	.dc.b	$7C,$27,$C7,$F8,$59,$7F,$1E,$3E,$57,$95,$23,$E5,$3F,$37,$CB,$F9
	.dc.b	$DF,$09,$F3,$FE,$12,$57,$CF,$8F,$A5,$F0,$9F,$4F,$E1,$24,$FD,$38
	.dc.b	$FA,$DE,$B7,$D7,$F5,$64,$FD,$78,$FB,$5F,$09,$23,$ED,$3F,$77,$E0
	.dc.b	$FE,$F7,$C1,$C9,$FB,$CF,$E0,$7E,$0F,$AC,$7E,$07,$0E,$AF,$C3,$C3
	.dc.b	$A3,$F1,$70,$BC,$5F,$27,$1F,$E8,$9F,$A6,$7F,$4F,$8F,$E0,$AF,$F4
	.dc.b	$BA,$FD,$73,$5F,$AA,$EF,$B3,$F7,$6B,$E1,$39,$92,$78,$FC,$77,$6F
	.dc.b	$8A,$91,$D1,$F0,$FC,$2F,$8B,$E1,$64,$F8,$1F,$2F,$C0,$73,$8F,$0E
	.dc.b	$17,$3F,$3E,$3D,$27,$CB,$8D,$9F,$5E,$3E,$EF,$86,$95,$EC,$7E,$0F
	.dc.b	$C2,$FC,$2F,$85,$93,$F0,$1F,$8B,$F0,$9F,$1B,$E1,$64,$FC,$47,$E4
	.dc.b	$FC,$27,$CA,$F8,$5F,$97,$F0,$B2,$BE,$4C,$74,$7E,$4E,$17,$4F,$9F
	.dc.b	$8F,$49,$F9,$B8,$D9,$FA,$78,$FD,$6F,$85,$95,$F5,$1F,$B3,$F0,$9C
	.dc.b	$F9,$7E,$9E,$56,$BF,$C9,$5C,$AC,$FA,$57,$D8,$CE,$D6,$B2,$FC,$1F
	.dc.b	$C2,$FE,$13,$E1,$24,$FE,$0A,$3F,$0F,$F0,$B2,$3F,$0A,$FE,$2B,$E1
	.dc.b	$7F,$C1,$DC,$1F,$D3,$F2,$FE,$7E,$FF,$4E,$AF,$DE,$B7,$83,$FA,$9B
	.dc.b	$B8,$FF,$30,$3F,$D4,$63,$3F,$03,$D0,$EB,$F0,$BE,$1A,$FC,$9E,$36
	.dc.b	$BF,$33,$8F,$8B,$1F,$1F,$C2,$CB,$FE,$91,$97,$7F,$3F,$FD,$65,$F0
	.dc.b	$9E,$82,$4E,$BE,$13,$D7,$F0,$9E,$CF,$84,$93,$CA,$3D,$FF,$09,$23
	.dc.b	$AB,$F0,$BE,$13,$E1,$FC,$27,$C4,$F8,$49,$5E,$28,$EA,$F9,$30,$EA
	.dc.b	$FC,$6C,$3A,$3F,$23,$0B,$C1,$F2,$31,$F9,$FF,$09,$F4,$3E,$0E,$5F
	.dc.b	$BA,$3E,$9F,$C2,$48,$F8,$2F,$D6,$F8,$3F,$AF,$F0,$9F,$63,$E0,$E5
	.dc.b	$7C,$68,$FB,$7F,$09,$F7,$3D,$D9,$3F,$2A,$3A,$3F,$73,$0B,$A7,$DC
	.dc.b	$C7,$AC,$FD,$0C,$7A,$BF,$82,$C7,$F0,$FE,$F4,$8F,$AA,$FF,$95,$FF
	.dc.b	$4D,$FE,$77,$D3,$F8,$AC,$BA,$B6,$BE,$DE,$3D,$9F,$C6,$E3,$DB,$2F
	.dc.b	$C6,$E5,$CF,$9F,$A7,$C2,$4B,$FF,$05,$67,$FD,$A1,$FE,$A2,$BE,$17
	.dc.b	$C4,$48,$DF,$C0,$79,$7E,$0E,$47,$27,$D1,$F0,$7E,$9F,$81,$F5,$7C
	.dc.b	$14,$AE,$F1,$CD,$EF,$85,$D7,$FC,$6B,$97,$F6,$89,$E8,$43,$CB,$8F
	.dc.b	$67,$D3,$F0,$5F,$1F,$E0,$A4,$60,$F6,$FC,$0F,$CB,$23,$F0,$3E,$0B
	.dc.b	$E7,$7C,$0C,$8F,$86,$FD,$1F,$82,$FA,$5E,$FC,$9F,$8E,$FD,$5F,$80
	.dc.b	$E7,$1F,$23,$0B,$AF,$AB,$1E,$93,$F3,$31,$B3,$EE,$C7,$EF,$7C,$07
	.dc.b	$DF,$F8,$09,$7F,$52,$3F,$05,$F0,$5F,$83,$F8,$29,$3F,$62,$3F,$D3
	.dc.b	$9F,$A7,$7F,$53,$8F,$F7,$43,$FD,$5F,$1F,$F2,$FB,$7B,$7F,$CA,$2E
	.dc.b	$FA,$8F,$F5,$9A,$FF,$26,$2F,$80,$EA,$76,$F8,$1F,$1C,$FE,$23,$2F
	.dc.b	$1B,$F8,$8C,$BC,$6F,$93,$E2,$23,$D0,$F3,$FC,$17,$A3,$E0,$A4,$78
	.dc.b	$23,$AB,$F9,$2B,$EB,$AF,$E1,$30,$EB,$3C,$C8,$7F,$8A,$33,$7C,$27
	.dc.b	$C4,$F8,$49,$3E,$C7,$E3,$FC,$27,$C8,$F8,$49,$3F,$01,$F9,$7F,$03
	.dc.b	$F3,$3E,$03,$E6,$FC,$04,$AF,$8B,$1F,$43,$E0,$3E,$8F,$B3,$27,$E4
	.dc.b	$C7,$47,$C9,$85,$DB,$E9,$E3,$D6,$7E,$7E,$3D,$5F,$56,$3F,$73,$E1
	.dc.b	$24,$7D,$47,$EF,$FC,$27,$E0,$3E,$12,$4F,$D8,$7F,$07,$F0,$5F,$84
	.dc.b	$F8,$39,$3F,$71,$FC,$3F,$C0,$FE,$23,$E0,$64,$FE,$01,$FF,$73,$7E
	.dc.b	$9F,$FD,$17,$95,$8F,$6A,$F9,$CF,$67,$A1,$0F,$A5,$87,$67,$FD,$2D
	.dc.b	$5F,$D7,$FB,$BF,$D4,$2D,$87,$72,$30,$DF,$83,$F1,$FC,$27,$93,$E0
	.dc.b	$E5,$F4,$8F,$3F,$C1,$FA,$3E,$02,$4F,$82,$3D,$7F,$03,$23,$C6,$FB
	.dc.b	$BE,$07,$AC,$79,$B0,$EA,$FF,$9C,$B3,$FF,$58,$BC,$CF,$5E,$1D,$9D
	.dc.b	$61,$F2,$3E,$07,$E4,$FC,$14,$9F,$83,$1D,$9E,$98,$76,$EA,$FF,$69
	.dc.b	$E3,$DA,$79,$9D,$88,$60,$F2,$7C,$1F,$D6,$F8,$29,$1F,$3A,$3E,$CF
	.dc.b	$C1,$C8,$FA,$2F,$DC,$F8,$0F,$BB,$F0,$1F,$7B,$DF,$95,$F5,$A3,$F0
	.dc.b	$3F,$01,$23,$EC,$BF,$84,$F8,$2E,$91,$F1,$70,$BA,$7A,$B1,$EB,$3F
	.dc.b	$80,$C7,$AB,$F2,$F1,$FC,$6F,$C1,$C8,$FC,$2B,$F9,$0F,$83,$47,$FB
	.dc.b	$77,$D4,$FE,$C2,$FE,$19,$BD,$26,$9F,$D6,$72,$FF,$7E,$AF,$83,$E7
	.dc.b	$5F,$EA,$C8,$FE,$B8,$F5,$F7,$F3,$5C,$2A,$F6,$7F,$BD,$AF,$FA,$95
	.dc.b	$F1,$F7,$B2,$F8,$DB,$8F,$8F,$E0,$ED,$79,$3E,$1E,$98,$9C,$E3,$AC
	.dc.b	$E6,$3E,$F8,$E3,$F0,$92,$3E,$13,$D3,$E0,$FA,$FC,$27,$6F,$83,$95
	.dc.b	$F1,$E3,$C3,$F0,$7E,$2F,$81,$93,$F2,$E3,$CB,$F0,$32,$3E,$73,$E8
	.dc.b	$F8,$1E,$B1,$F4,$70,$EA,$FC,$9C,$3A,$BF,$57,$0E,$AF,$CD,$C3,$E0
	.dc.b	$EF,$23,$ED,$3F,$13,$E0,$FA,$C7,$D3,$C3,$AF,$27,$D8,$C7,$AC,$FD
	.dc.b	$7C,$7A,$CF,$DC,$C7,$E6,$FC,$1C,$AF,$C3,$3F,$43,$E0,$BE,$8F,$C1
	.dc.b	$C9,$FC,$53,$F5,$3D,$DF,$AB,$EF,$7D,$8F,$D4,$3F,$B1,$1F,$D8,$F7
	.dc.b	$46,$7E,$0A,$61,$F9,$4D,$1D,$7F,$96,$E8,$63,$FD,$B6,$BF,$8D,$BF
	.dc.b	$DA,$DF,$C8,$67,$E3,$69,$F9,$DC,$FC,$6F,$5B,$E9,$5E,$48,$C8,$C3
	.dc.b	$C2,$F6,$F8,$4F,$51,$1E,$47,$8B,$E0,$BD,$DF,$07,$23,$39,$FA,$23
	.dc.b	$E1,$7C,$1F,$C3,$F8,$3F,$89,$F0,$72,$F7,$9A,$7E,$47,$C0,$C9,$CB
	.dc.b	$F2,$FE,$07,$B4,$79,$F6,$F5,$E5,$8D,$BD,$3A,$7C,$8D,$3C,$17,$DB
	.dc.b	$4F,$03,$C4,$F9,$79,$FD,$7F,$80,$FB,$1F,$03,$27,$CB,$1F,$6F,$E0
	.dc.b	$BE,$E7,$C1,$C9,$F4,$C7,$DF,$F8,$39,$1E,$C7,$F0,$5F,$07,$F8,$3F
	.dc.b	$7F,$F0,$9F,$07,$2B,$E0,$C7,$E1,$FE,$03,$F1,$1F,$07,$27,$E2,$C7
	.dc.b	$E2,$FE,$0E,$47,$C8,$7F,$82,$FE,$A3,$FE,$AF,$3F,$AE,$9F,$EB,$DE
	.dc.b	$5F,$C5,$36,$7F,$B1,$BF,$D8,$BA,$17,$DE,$7E,$27,$27,$E4,$31,$F2
	.dc.b	$47,$72,$1A,$A6,$3E,$2C,$7C,$7F,$05,$2B,$F0,$79,$7F,$51,$E6,$FA
	.dc.b	$AB,$DC,$EB,$3C,$F2,$F1,$57,$DA,$CB,$CB,$3E,$08,$C6,$5F,$0A,$3C
	.dc.b	$F8,$7C,$1F,$7F,$E1,$FC,$04,$9F,$0C,$7C,$6F,$82,$F8,$FF,$05,$27
	.dc.b	$CB,$1F,$2B,$E0,$64,$7A,$1F,$9B,$F0,$5F,$3B,$E0,$3E,$7F,$C0,$CA
	.dc.b	$F6,$C7,$D2,$F7,$FE,$9F,$BF,$27,$E0,$C7,$D6,$F7,$A4,$7C,$47,$EC
	.dc.b	$FB,$FF,$6B,$E0,$64,$FC,$87,$EE,$FC,$0F,$DE,$F8,$1F,$BF,$F0,$32
	.dc.b	$BE,$6C,$7F,$46,$FD,$4B,$FA,$C9,$3E,$03,$1D,$B9,$F7,$E5,$FE,$DA
	.dc.b	$BF,$AD,$37,$E0,$39,$7C,$0F,$3F,$82,$93,$B8,$ED,$F0,$5D,$FE,$0A
	.dc.b	$4F,$38,$F1,$7B,$DE,$3F,$7E,$4F,$78,$F3,$7B,$D2,$3C,$4F,$A7,$DF
	.dc.b	$F5,$7C,$17,$AF,$E0,$A5,$79,$E3,$DD,$EF,$C8,$F5,$3F,$07,$E0,$BE
	.dc.b	$17,$C1,$49,$F7,$3F,$17,$E0,$BE,$37,$C0,$49,$F8,$4F,$C9,$F8,$0F
	.dc.b	$95,$F0,$52,$7E,$33,$FD,$CB,$C0,$FF,$4E,$1F,$CC,$FA,$0F,$F3,$3F
	.dc.b	$2F,$F8,$8B,$7D,$0F,$E7,$4E,$CF,$F5,$05,$FF,$26,$5F,$01,$D4,$ED
	.dc.b	$F0,$3D,$FD,$F9,$5B,$7C,$5E,$FF,$8F,$DE,$F2,$7B,$F2,$BA,$47,$9F
	.dc.b	$E0,$7D,$1F,$03,$27,$C1,$1E,$BF,$80,$F6,$7C,$0C,$9F,$24,$7B,$FE
	.dc.b	$02,$47,$9D,$F8,$5F,$03,$F0,$FE,$06,$4F,$AD,$F8,$DF,$03,$F1,$FE
	.dc.b	$03,$E4,$7C,$0C,$AF,$81,$1F,$2F,$E0,$64,$7C,$37,$E7,$7B,$FF,$3F
	.dc.b	$DE,$FF,$10,$FE,$A7,$FD,$52,$BF,$8A,$B7,$A7,$E7,$CE,$BE,$26,$57
	.dc.b	$F5,$2A,$FF,$3F,$9F,$C8,$7A,$CF,$F1,$3E,$BF,$ED,$CF,$F1,$39,$E1
	.dc.b	$F0,$5E,$0F,$83,$F0,$FC,$07,$8B,$E0,$65,$75,$8F,$2F,$C0,$48,$F0
	.dc.b	$3E,$8F,$81,$F4,$FC,$0C,$9F,$23,$EC,$F8,$1F,$6F,$C0,$7B,$BE,$06
	.dc.b	$57,$A6,$3E,$0F,$C0,$FC,$2F,$1A,$4F,$B6,$3E,$2F,$C0,$C8,$F8,$0F
	.dc.b	$C8,$F8,$1F,$93,$F0,$32,$7E,$23,$F3,$3E,$07,$E6,$FC,$07,$CE,$F8
	.dc.b	$19,$5F,$26,$3F,$CA,$3F,$AA,$7F,$99,$0F,$F3,$17,$3F,$F8,$55,$B3
	.dc.b	$FE,$AE,$75,$F0,$3C,$09,$1F,$E0,$48,$FF,$32,$BF,$D3,$9F,$F3,$0F
	.dc.b	$67,$F7,$86,$D7,$5F,$7B,$C3,$EF,$F8,$BD,$E9,$87,$37,$CB,$EF,$F9
	.dc.b	$BD,$E9,$3D,$DF,$4F,$BF,$EA,$F7,$E4,$F8,$DF,$6F,$BF,$EE,$F7,$A4
	.dc.b	$F9,$DF,$83,$EF,$FC,$2F,$7E,$4F,$AD,$F8,$BE,$FF,$C6,$F7,$FE,$3F
	.dc.b	$BF,$2B,$E0,$47,$CA,$F7,$E4,$7C,$37,$E6,$FB,$FF,$3B,$DF,$93,$F1
	.dc.b	$DF,$F4,$0F,$EA,$BF,$E8,$E8,$FE,$EA,$FF,$11,$3F,$87,$3F,$ED,$E6
	.dc.b	$F9,$87,$5E,$FF,$22,$4F,$F8,$DA,$3F,$88,$B7,$DF,$EE,$47,$5E,$FF
	.dc.b	$8B,$DE,$FF,$22,$F6,$7F,$8B,$AF,$FC,$71,$D7,$CF,$DF,$CB,$FE,$59
	.dc.b	$8C,$E1,$E8,$F7,$B2,$F9,$DB,$A7,$9F,$DF,$B4,$9B,$EF,$63,$DF,$B8
	.dc.b	$FA,$5D,$FB,$F2,$3D,$9E,$FD,$CF,$E3,$FB,$FD,$0E,$BE,$FD,$AF,$85
	.dc.b	$EF,$78,$0F,$0F,$BF,$63,$E3,$7B,$FE,$4F,$7F,$CA,$D8,$F9,$5E,$FF
	.dc.b	$A0,$F5,$E3,$FE,$7A,$8C,$9F,$E6,$45,$EF,$EB,$E2,$36,$D9,$D7,$BD
	.dc.b	$C8,$E7,$EF,$D8,$E5,$EF,$76,$3B,$FB,$F6,$3B,$7B,$FE,$2F,$7F,$C6
	.dc.b	$D8,$F1,$7B,$FE,$6F,$7F,$CE,$D8,$F3,$7B,$DE,$AF,$7F,$D6,$D8,$F5
	.dc.b	$7B,$DE,$EF,$7F,$DE,$D8,$F7,$7B,$FF,$0B,$DF,$F8,$6D,$8F,$85,$EF
	.dc.b	$FC,$63,$E3,$FB,$F6,$3E,$37,$BD,$F2,$BD,$FF,$96,$D8,$FF,$46,$FE
	.dc.b	$AC,$F8,$5F,$C4,$AF,$22,$C6,$1C,$FC,$0E,$BE,$12,$DF,$27,$8F,$C0
	.dc.b	$C8,$EB,$F0,$37,$C5,$FF,$31,$E4,$F8,$1E,$FF,$03,$E0,$6D,$F9,$3E
	.dc.b	$07,$C7,$F0,$3E,$46,$C7,$A3,$E0,$7C,$FF,$03,$E8,$6C,$7B,$3E,$07
	.dc.b	$D6,$7B,$3E,$06,$C7,$C0,$F8,$1F,$7F,$C0,$FC,$06,$C7,$C4,$F8,$0F
	.dc.b	$87,$F0,$3F,$11,$B1,$F2,$3E,$07,$E3,$FC,$0F,$C8,$6C,$7C,$CF,$81
	.dc.b	$F9,$67,$ED,$3F,$57,$72,$62,$7F,$D7,$B0,$F8,$9D,$1B,$F8,$CB,$3A
	.dc.b	$F8,$4E,$5F,$0F,$CD,$B1,$CB,$E1,$7B,$7C,$2F,$76,$C7,$6F,$85,$F1
	.dc.b	$1E,$3F,$84,$B1,$E2,$F8,$2F,$37,$C2,$79,$DB,$1E,$6F,$83,$F5,$7C
	.dc.b	$1F,$AD,$B1,$EA,$F8,$3F,$77,$C1,$FB,$DB,$1E,$EF,$77,$E1,$7C,$1F
	.dc.b	$C3,$6C,$7C,$2F,$83,$F8,$C7,$C7,$F8,$3B,$1F,$1B,$E0,$BE,$57,$C1
	.dc.b	$FC,$B6,$C7,$CA,$F8,$3F,$E4,$3F,$AC,$7F,$07,$EA,$FF,$58,$F5,$7F
	.dc.b	$AB,$2B,$FD,$D7,$39,$F8,$6F,$F6,$EB,$FC,$2A,$33,$F0,$9C,$BE,$17
	.dc.b	$9B,$D3,$E1,$64,$F2,$F8,$5B,$97,$FA,$FB,$07,$C2,$F8,$8F,$1F,$C2
	.dc.b	$DB,$F0,$FC,$27,$9B,$E1,$7C,$ED,$8F,$2F,$C2,$7A,$8F,$5F,$C2,$D8
	.dc.b	$F4,$FC,$27,$BB,$E1,$7D,$ED,$8F,$6F,$C2,$7C,$2F,$85,$F8,$6D,$8F
	.dc.b	$83,$F0,$9F,$18,$F8,$FF,$0B,$63,$E2,$FC,$2F,$CA,$F8,$5F,$96,$D8
	.dc.b	$F9,$3F,$09,$F3,$BE,$17,$E7,$B6,$3F,$3B,$D2,$7F,$3F,$F0,$8F,$5F
	.dc.b	$F8,$74,$68,$CB,$FC,$3D,$C7,$3F,$F1,$18,$DB,$FE,$EA,$74,$DF,$84
	.dc.b	$FE,$BF,$0F,$84,$B1,$87,$3F,$09,$AF,$84,$B8,$76,$C7,$C9,$1C,$BE
	.dc.b	$0A,$4E,$71,$7C,$1E,$33,$F1,$7C,$17,$83,$E0,$FC,$2D,$BE,$1F,$05
	.dc.b	$E4,$F8,$3F,$2B,$63,$A7,$C1,$FA,$3E,$0F,$D2,$D8,$F0,$7C,$1F,$B0
	.dc.b	$F6,$E9,$63,$A9,$E5,$C3,$AB,$4E,$A7,$A6,$FA,$B4,$F6,$7C,$17,$C8
	.dc.b	$3E,$4F,$C1,$D8,$F8,$1F,$05,$F3,$3E,$0F,$E6,$B6,$3E,$27,$C1,$FD
	.dc.b	$03,$FC,$1F,$FA,$CF,$F1,$C8,$C9,$F3,$5F,$E3,$8D,$F8,$3D,$FC,$27
	.dc.b	$07,$8F,$C1,$F2,$F8,$4B,$5C,$5E,$BF,$07,$23,$A7,$C2,$5E,$77,$7F
	.dc.b	$E3,$DF,$EB,$19,$FF,$89,$A3,$1D,$88,$C5,$FE,$51,$7C,$1F,$AB,$E1
	.dc.b	$3C,$38,$74,$AD,$67,$8C,$79,$E1,$FE,$CE,$9F,$76,$7E,$39,$CE,$76
	.dc.b	$7B,$7C,$1F,$C6,$F8,$4F,$8E,$D8,$F8,$DF,$07,$F2,$8F,$97,$F0,$96
	.dc.b	$3E,$57,$C1,$FC,$EF,$84,$F9,$ED,$8F,$9D,$F0,$7F,$48,$FA,$7F,$09
	.dc.b	$63,$A9,$38,$3D,$87,$C6,$C3,$B3,$8E,$3F,$8D,$87,$58,$C7,$02,$31
	.dc.b	$BF,$C2,$F8,$4F,$C0,$FC,$27,$E0,$9B,$97,$79,$F8,$FB,$7B,$23,$07
	.dc.b	$D1,$BE,$6D,$3F,$11,$B7,$E3,$3E,$0F,$F1,$BF,$09,$67,$F1,$8F,$E4
	.dc.b	$3E,$12,$47,$EE,$FF,$5A,$FE,$78,$FD,$D8,$FF,$31,$C7,$F3,$DD,$BD
	.dc.b	$CD,$CF,$F0,$AB,$E1,$37,$F0,$93,$0E,$8F,$2F,$84,$C6,$1F,$13,$1F
	.dc.b	$0B,$D2,$AE,$3D,$2F,$6F,$6C,$2C,$FF,$24,$C5,$F0,$9E,$63,$CF,$F0
	.dc.b	$76,$7D,$9F,$07,$EA,$3D,$7F,$05,$63,$E0,$7C,$1F,$BB,$E0,$FD,$ED
	.dc.b	$8F,$89,$F0,$7F,$0B,$E0,$BE,$1B,$63,$A9,$37,$F8,$B8,$75,$7E,$3D
	.dc.b	$F3,$6C,$FC,$FC,$3E,$6F,$C1,$FC,$EF,$82,$B3,$F4,$DF,$A3,$F0,$7F
	.dc.b	$4B,$E0,$EC,$FD,$77,$EA,$FC,$1C,$8F,$B5,$F0,$57,$2F,$AF,$F0,$7F
	.dc.b	$6B,$E0,$BE,$DB,$6B,$F0,$18,$F5,$69,$D4,$FC,$26,$1C,$F9,$BE,$FE
	.dc.b	$36,$7E,$97,$C1,$FF,$AA,$3F,$5B,$FE,$9F,$5F,$84,$8C,$7C,$47,$A3
	.dc.b	$4F,$C4,$EB,$6B,$F1,$2F,$38,$FF,$34,$74,$33,$FC,$8E,$79,$F1,$59
	.dc.b	$FF,$5A,$C7,$F8,$AB,$74,$FE,$CE,$BE,$17,$C9,$5E,$5F,$85,$95,$FD
	.dc.b	$66,$9F,$0B,$8D,$3F,$75,$0F,$85,$F6,$1E,$DF,$85,$B7,$C7,$E1,$7E
	.dc.b	$07,$C2,$7C,$16,$C7,$5C,$BA,$6B,$E6,$CE,$CF,$9B,$1E,$79,$F8,$32
	.dc.b	$B7,$EC,$F8,$5F,$99,$F0,$BF,$35,$B8,$7A,$7E,$0B,$E8,$1F,$47,$E1
	.dc.b	$6C,$7B,$7E,$0F,$EA,$1F,$57,$E1,$6C,$7C,$1F,$85,$FB,$07,$D9,$F8
	.dc.b	$5B,$1D,$49,$AF,$C4,$C3,$AB,$F1,$B0,$E6,$FC,$8C,$2E,$9F,$53,$E1
	.dc.b	$7F,$08,$7E,$17,$E1,$2E,$3F,$47,$E1,$7F,$11,$F0,$BF,$89,$6C,$7F
	.dc.b	$83,$7B,$7F,$B2,$65,$FC,$B5,$B8,$FF,$54,$7F,$B2,$C6,$4F,$F7,$CA
	.dc.b	$F8,$4B,$F0,$98,$6D,$F8,$4F,$C1,$63,$E1,$69,$E1,$BE,$67,$4F,$85
	.dc.b	$B1,$E5,$F8,$3E,$E7,$83,$E0,$EC,$75,$AF,$C5,$DF,$56,$C7,$52,$3E
	.dc.b	$DB,$EA,$FB,$FE,$0F,$D6,$7B,$3E,$12,$D7,$C5,$F8,$5F,$7F,$C2,$7C
	.dc.b	$06,$C7,$C9,$F8,$5F,$87,$EF,$FC,$46,$C7,$CD,$F8,$3F,$8E,$7C,$8F
	.dc.b	$80,$B1,$D4,$FA,$58,$75,$69,$D4,$FA,$D7,$D5,$A7,$D9,$F8,$0F,$A6
	.dc.b	$7D,$4F,$85,$B1,$F7,$7E,$17,$EB,$FC,$1F,$D8,$6C,$7E,$07,$E1,$3E
	.dc.b	$D9,$F7,$3E,$12,$C7,$E1,$7E,$13,$EF,$9F,$80,$F8,$3B,$1D,$4F,$C5
	.dc.b	$61,$D5,$A7,$52,$71,$FE,$36,$FA,$BF,$DE,$FF,$5C,$FE,$C6,$FF,$BE
	.dc.b	$DF,$EC,$13,$FE,$74,$73,$E7,$9C,$CE,$FE,$0F,$FC,$59,$DD,$FE,$2A
	.dc.b	$9F,$EC,$D7,$FE,$69,$5F,$0B,$A8,$E9,$56,$FC,$1F,$09,$71,$EC,$DC
	.dc.b	$3F,$D4,$F7,$CE,$31,$F0,$79,$F8,$1B,$3D,$DE,$18,$C8,$F1,$65,$72
	.dc.b	$F8,$18,$75,$3B,$7C,$25,$AF,$47,$C1,$78,$7E,$0B,$C4,$D8,$F6,$63
	.dc.b	$CF,$6F,$81,$95,$9F,$91,$8F,$3C,$BE,$76,$56,$FE,$37,$C1,$FB,$4F
	.dc.b	$77,$C2,$5C,$3E,$47,$C2,$7C,$1F,$84,$F8,$4D,$8F,$99,$F0,$5F,$17
	.dc.b	$CA,$F8,$CD,$8F,$A1,$EF,$FC,$93,$E5,$7B,$F6,$3A,$9F,$57,$0E,$AD
	.dc.b	$3A,$9F,$66,$FA,$B4,$FB,$9F,$07,$F5,$4F,$AD,$F0,$96,$3F,$01,$F0
	.dc.b	$9F,$64,$FB,$9F,$AE,$FF,$68,$8F,$C6,$46,$3A,$B3,$F0,$76,$7F,$D6
	.dc.b	$15,$FD,$75,$BF,$01,$C8,$E7,$F0,$16,$3B,$93,$97,$FC,$AD,$5D,$DF
	.dc.b	$CE,$63,$D5,$CE,$17,$93,$FD,$03,$5D,$E7,$8F,$C0,$7A,$0F,$4F,$C0
	.dc.b	$5C,$7C,$BF,$03,$EC,$3D,$BF,$05,$63,$D3,$F0,$3F,$03,$E0,$BE,$0B
	.dc.b	$63,$DB,$F0,$5F,$13,$E0,$3E,$2B,$63,$A9,$33,$F5,$E1,$D5,$F6,$DF
	.dc.b	$36,$CF,$C8,$C3,$E7,$FC,$07,$D0,$F8,$0B,$3F,$31,$E5,$BF,$C0,$D3
	.dc.b	$D3,$3D,$E3,$E1,$E3,$D5,$B1,$D4,$FA,$BF,$05,$F7,$3E,$0B,$EE,$B6
	.dc.b	$3E,$CF,$C1,$7E,$03,$E0,$BF,$02,$D8,$FB,$BF,$05,$F8,$43,$F0,$BE
	.dc.b	$F5,$8F,$F7,$87,$EB,$DF,$D8,$97,$03,$AD,$3F,$D9,$34,$FF,$5C,$BB
	.dc.b	$9D,$39,$AE,$34,$E7,$65,$E3,$BF,$C3,$B7,$0F,$C1,$E5,$E3,$9F,$C5
	.dc.b	$E5,$67,$BF,$C1,$7F,$17,$CD,$F0,$76,$30,$E7,$E0,$E6,$BE,$CF,$83
	.dc.b	$B9,$7F,$AA,$F0,$7C,$17,$3F,$82,$E8,$DB,$F8,$7F,$05,$DF,$DF,$F0
	.dc.b	$36,$3E,$3E,$5D,$35,$F8,$79,$D9,$F8,$39,$76,$7E,$6E,$5D,$9A,$7D
	.dc.b	$0F,$82,$F6,$1E,$DF,$80,$B1,$F5,$3E,$0B,$E0,$7C,$17,$C1,$6C,$7D
	.dc.b	$8F,$82,$F8,$87,$C5,$F8,$2B,$1F,$73,$E0,$BE,$47,$BF,$F2,$5B,$1D
	.dc.b	$4F,$C0,$E1,$D5,$A7,$52,$71,$FC,$9B,$EA,$FC,$BF,$82,$FA,$9F,$01
	.dc.b	$F5,$5B,$5D,$67,$E8,$5F,$56,$C7,$4E,$1F,$C5,$E7,$E9,$8C,$1F,$2B
	.dc.b	$E0,$BF,$01,$F0,$5F,$81,$6C,$7E,$5A,$8F,$E4,$3E,$77,$FD,$72,$E4
	.dc.b	$C7,$B9,$FD,$7E,$3F,$AE,$E3,$FE,$ED,$76,$DF,$82,$D7,$C1,$6D,$B7
	.dc.b	$CE,$7F,$49,$C1,$FD,$D8,$9F,$11,$FD,$AD,$C7,$C0,$D3,$1F,$03,$6F
	.dc.b	$97,$8A,$AF,$0E,$BE,$AC,$7A,$FC,$0F,$3D,$BA,$3D,$74,$ED,$AD,$CB
	.dc.b	$9E,$28,$ED,$83,$1E,$99,$77,$D3,$C7,$87,$07,$B3,$CB,$2B,$7C,$63
	.dc.b	$F7,$78,$FC,$8C,$39,$AF,$82,$F8,$07,$C1,$F8,$3B,$5E,$EE,$07,$B7
	.dc.b	$F4,$7D,$BF,$A4,$D9,$D7,$B3,$FE,$13,$F9,$CF,$E9,$5B,$B0,$4E,$7E
	.dc.b	$27,$E7,$7C,$2F,$CF,$F8,$5B,$8F,$C2,$79,$7B,$F2,$3E,$7E,$5F,$EB
	.dc.b	$46,$C7,$7F,$76,$46,$BD,$DB,$C7,$F0,$4F,$BF,$F0,$BF,$74,$FB,$DF
	.dc.b	$05,$6B,$F0,$9C,$BF,$46,$7F,$C5,$CF,$D3,$9B,$D0,$9B,$1F,$87,$F8
	.dc.b	$5F,$C4,$7C,$1D,$C3,$ED,$BF,$8B,$F8,$59,$1F,$7B,$E1,$6E,$7E,$5C
	.dc.b	$FF,$01,$DB,$D8,$EF,$A5,$BF,$C2,$F3,$FE,$54,$FC,$B7,$45,$8F,$C4
	.dc.b	$F3,$7E,$64,$FF,$22,$7E,$BF,$F4,$BB,$3E,$8C,$FB,$B0,$FF,$2C,$E5
	.dc.b	$F8,$FF,$07,$DB,$3E,$E7,$8F,$5F,$2F,$85,$B3,$FA,$0C,$7F,$C0,$17
	.dc.b	$F9,$45,$CF,$E0,$3C,$3F,$0D,$6B,$E6,$FD,$0F,$E9,$DB,$A7,$CE,$87
	.dc.b	$99,$63,$D3,$D5,$EA,$EA,$B7,$E1,$7D,$BF,$03,$EE,$F8,$5B,$3E,$57
	.dc.b	$A7,$A3,$23,$C3,$E8,$DE,$8F,$C2,$7C,$FB,$F7,$4F,$F6,$25,$7F,$05
	.dc.b	$8C,$7B,$9F,$E6,$F7,$5E,$E5,$AD,$DF,$E3,$79,$9C,$8F,$84,$FA,$3F
	.dc.b	$0F,$F4,$9B,$9F,$C2,$F6,$FE,$07,$B7,$F0,$5B,$1E,$1F,$DD,$1F,$03
	.dc.b	$E1,$5F,$03,$63,$0E,$7E,$22,$57,$D5,$F8,$9B,$B7,$E3,$7E,$8F,$9C
	.dc.b	$FE,$79,$E6,$DC,$3F,$3F,$18,$9F,$8B,$8B,$E2,$3C,$47,$8F,$E2,$6E
	.dc.b	$1E,$7D,$BC,$DB,$F9,$DB,$1D,$FE,$2B,$AE,$BD,$9B,$1E,$FD,$7D,$DF
	.dc.b	$17,$EF,$6C,$7C,$3E,$DF,$84,$7C,$3E,$DB,$1F,$B8,$C7,$FC,$31,$B7
	.dc.b	$F2,$76,$F8,$7F,$4F,$AB,$F7,$71,$AE,$A7,$0B,$BB,$E7,$1F,$3F,$BA
	.dc.b	$E5,$F4,$FD,$4F,$A4,$7D,$3E,$BB,$1F,$5F,$C5,$F9,$27,$D7,$2B,$F6
	.dc.b	$FC,$5F,$B5,$E2,$FD,$B6,$C7,$E3,$FF,$74,$79,$FF,$8F,$5E,$7D,$8C
	.dc.b	$39,$F3,$E5,$71,$F3,$AF,$27,$E7,$30,$79,$DC,$FC,$EE,$8D,$BF,$07
	.dc.b	$AF,$DC,$F0,$7A,$D6,$3C,$9E,$B7,$8F,$D5,$F2,$36,$3D,$1E,$AF,$9C
	.dc.b	$F4,$7A,$B6,$3D,$9E,$E7,$AF,$DC,$F6,$36,$3E,$07,$B7,$EF,$3E,$07
	.dc.b	$B1,$63,$E2,$7B,$1F,$0C,$F8,$9E,$DD,$8F,$DB,$79,$AF,$82,$FD,$B2
	.dc.b	$F7,$AC,$61,$CF,$BF,$2B,$E8,$17,$9B,$E4,$7B,$9F,$47,$DF,$FA,$4D
	.dc.b	$9E,$FF,$03,$DB,$E0,$7B,$B6,$3C,$7F,$07,$E2,$F7,$BC,$6D,$8F,$3F
	.dc.b	$BF,$E6,$F8,$3F,$3B,$63,$D7,$F0,$BE,$A3,$D7,$F0,$F6,$3D,$FF,$11
	.dc.b	$EE,$F8,$3F,$7B,$63,$E1,$FC,$1F,$C2,$F8,$7F,$86,$D8,$F8,$FF,$1D
	.dc.b	$FE,$84,$E2,$FE,$43,$CC,$8C,$7A,$BF,$E1,$9F,$4F,$F8,$8C,$62,$FF
	.dc.b	$D1,$71,$FC,$53,$D5,$C6,$08,$FE,$2F,$F0,$B8,$7F,$8C,$FC,$15,$C8
	.dc.b	$D9,$C7,$C1,$E7,$E2,$A6,$1F,$72,$39,$7E,$E8,$F8,$2A,$49,$FE,$6D
	.dc.b	$1F,$E9,$46,$FC,$77,$02,$3A,$CB,$9F,$C3,$48,$E2,$F9,$5E,$FF,$09
	.dc.b	$1E,$AF,$8B,$E0,$7C,$7F,$05,$E4,$F8,$49,$5E,$28,$F3,$FC,$24,$8F
	.dc.b	$2B,$EA,$F8,$5F,$5F,$B9,$27,$D2,$FB,$BE,$17,$DF,$F0,$BF,$03,$E0
	.dc.b	$E5,$7B,$A3,$E1,$FC,$1F,$C4,$F8,$59,$3F,$0A,$3E,$87,$EE,$9F,$C0
	.dc.b	$23,$FA,$D3,$FE,$30,$6F,$B9,$AF,$73,$E6,$FC,$2C,$9F,$96,$FD,$0F
	.dc.b	$85,$FA,$3F,$0B,$F4,$BE,$16,$57,$D0,$8E,$FE,$E4,$8E,$EF,$8B,$DC
	.dc.b	$F1,$FC,$2F,$93,$E1,$65,$79,$23,$CF,$F0,$B2,$3C,$EF,$AB,$E1,$7D
	.dc.b	$7E,$E7,$B3,$DC,$95,$EC,$8F,$7F,$C2,$48,$F7,$BF,$0B,$E1,$3E,$1F
	.dc.b	$C1,$C9,$F8,$6F,$C6,$F8,$4F,$8F,$F0,$7F,$23,$E1,$25,$7C,$88,$FC
	.dc.b	$EF,$A2,$C5,$F0,$98,$5F,$09,$4C,$47,$D2,$3E,$9B,$2F,$E9,$C7,$E9
	.dc.b	$7F,$74,$7C,$1D,$24,$FD,$8F,$53,$C3,$F4,$BF,$C7,$63,$FC,$AA,$D7
	.dc.b	$FC,$90,$DC,$BF,$D2,$01,$7B,$DF,$84,$F8,$39,$5F,$84,$7F,$0F,$F0
	.dc.b	$9E,$3F,$80,$93,$FE,$21,$CF,$FC,$66,$FF,$8F,$FC,$BF,$1F,$B9,$E4
	.dc.b	$F8,$39,$3E,$48,$F6,$7C,$37,$B7,$E1,$A4,$F3,$8F,$E0,$FE,$1F,$F8
	.dc.b	$9B,$D0,$BF,$14,$69,$8F,$47,$E0,$FB,$5F,$0B,$E1,$26,$3F,$0A,$3E
	.dc.b	$57,$C2,$C8,$F4,$3F,$37,$E1,$BA,$FA,$B2,$7F,$CB,$0F,$83,$DC,$F9
	.dc.b	$B1,$0F,$9A,$7D,$0F,$82,$FA,$DF,$07,$23,$E2,$3E,$8F,$90,$F4,$FB
	.dc.b	$DE,$AF,$7A,$57,$F9,$96,$3F,$13,$AF,$F8,$A6,$3F,$D1,$AD,$7F,$D0
	.dc.b	$0D,$ED,$FB,$2F,$F1,$F8,$FF,$37,$7B,$DF,$23,$F6,$6F,$E7,$11,$FE
	.dc.b	$4F,$6D,$7F,$9C,$67,$59,$CA,$FF,$6A,$D7,$F3,$D6,$FC,$2F,$93,$DA
	.dc.b	$F2,$FC,$04,$9F,$2C,$77,$F8,$3F,$07,$C2,$49,$E7,$1E,$3F,$84,$91
	.dc.b	$D9,$F3,$73,$7C,$0F,$66,$4F,$C0,$7E,$1F,$C1,$7A,$F8,$3D,$9F,$11
	.dc.b	$2B,$CF,$1E,$FF,$87,$91,$EA,$7E,$17,$C4,$7E,$D3,$E9,$BF,$DA,$73
	.dc.b	$FE,$B8,$7F,$90,$3F,$EA,$B6,$F2,$7E,$91,$DC,$7C,$9E,$8F,$98,$47
	.dc.b	$E3,$F8,$7F,$3F,$C4,$91,$F2,$9F,$A5,$F1,$5F,$9A,$FA,$6F,$F7,$B4
	.dc.b	$8F,$F4,$13,$86,$FB,$1B,$F6,$78,$7B,$3C,$7D,$59,$5C,$75,$FF,$7E
	.dc.b	$E0,$AF,$C0,$78,$9F,$81,$F1,$25,$FD,$C8,$F1,$7B,$FE,$3F,$7E,$4F
	.dc.b	$3F,$53,$C9,$E5,$B2,$5F,$A7,$DE,$F5,$7B,$F3,$0E,$02,$F8,$4F,$E4
	.dc.b	$7D,$29,$3F,$8C,$7F,$29,$EA,$FE,$57,$D0,$FC,$B7,$A3,$2B,$F2,$31
	.dc.b	$F9,$9F,$5E,$47,$E5,$1F,$E0,$7F,$B3,$B6,$BE,$BE,$3D,$79,$7E,$4D
	.dc.b	$56,$AF,$04,$FE,$7B,$2F,$03,$FA,$5C,$BA,$E3,$E0,$8F,$D4,$65,$87
	.dc.b	$A6,$0F,$6F,$C7,$EE,$49,$CC,$7F,$18,$FA,$7F,$F5,$E4,$78,$DA,$FF
	.dc.b	$B0,$5B,$D1,$EB,$8F,$95,$0F,$33,$97,$C5,$F3,$F6,$A5,$73,$8F,$87
	.dc.b	$EE,$C8,$F2,$BF,$1B,$DE,$F8,$FE,$07,$C8,$F0,$25,$7A,$A3,$E5,$F8
	.dc.b	$12,$3D,$AF,$CE,$F0,$3E,$7F,$C0,$49,$F8,$2F,$D2,$F8,$0E,$B1,$F1
	.dc.b	$30,$EA,$FF,$B7,$78,$BF,$39,$C1,$ED,$E8,$FA,$79,$77,$AD,$3D,$DF
	.dc.b	$9B,$DD,$F7,$BC,$09,$18,$3D,$07,$72,$1D,$F5,$E9,$87,$7C,$7F,$B9
	.dc.b	$4F,$B0,$FB,$3A,$FE,$2B,$BA,$47,$DC,$7F,$1B,$DF,$F8,$EF,$82,$93
	.dc.b	$F8,$07,$F2,$3F,$05,$F9,$2F,$85,$93,$F8,$47,$F2,$BF,$0D,$FD,$63
	.dc.b	$DE,$7C,$35,$25,$7E,$56,$3F,$50,$F5,$21,$F2,$70,$EE,$F5,$21,$F3
	.dc.b	$6F,$BB,$CF,$E1,$BC,$1E,$A4,$8C,$1D,$FD,$4F,$27,$C5,$C8,$C1,$E3
	.dc.b	$F8,$AF,$41,$1F,$37,$C5,$FA,$FE,$2F,$D9,$EE,$C9,$F5,$47,$BF,$0E
	.dc.b	$90,$E5,$FC,$06,$CC,$B5,$3D,$C8,$6F,$C7,$FE,$23,$2C,$FE,$56,$7F
	.dc.b	$2F,$C3,$97,$F2,$5F,$9D,$E1,$7C,$FF,$0E,$4F,$CD,$7E,$97,$C3,$FD
	.dc.b	$3F,$87,$FA,$9F,$0F,$2B,$E9,$47,$D7,$CF,$EC,$7C,$3C,$9F,$AD,$1D
	.dc.b	$1F,$31,$FC,$7D,$61,$9D,$FD,$F8,$F7,$AF,$BD,$8F,$77,$8F,$85,$FE
	.dc.b	$95,$CB,$F0,$DF,$0D,$F8,$7F,$87,$95,$F8,$58,$E9,$C3,$F1,$33,$B3
	.dc.b	$E1,$CB,$B3,$D3,$4F,$F0,$25,$7C,$AC,$F3,$A7,$D2,$F8,$7F,$F1,$5F
	.dc.b	$ED,$1F,$E0,$2A,$FE,$7E,$DC,$3D,$93,$AF,$88,$91,$FC,$AE,$7F,$C0
	.dc.b	$8F,$A9,$7C,$3F,$4F,$86,$EB,$F0,$F2,$B8,$47,$83,$E1,$FC,$3F,$0F
	.dc.b	$27,$A4,$76,$7A,$63,$74,$FE,$11,$97,$69,$E6,$78,$F1,$EC,$F8,$F1
	.dc.b	$F6,$FC,$34,$8F,$43,$F0,$3E,$1F,$E0,$FC,$3C,$9F,$63,$F1,$3E,$1F
	.dc.b	$E2,$FC,$37,$C6,$F8,$69,$5F,$06,$3E,$4F,$C3,$7C,$AF,$86,$93,$F1
	.dc.b	$63,$A3,$F1,$70,$BB,$7A,$B1,$EB,$3F,$2F,$1E,$AF,$CB,$C7,$EA,$FC
	.dc.b	$2F,$D6,$F8,$69,$3F,$46,$3E,$CF,$C2,$C8,$FA,$8F,$DC,$F8,$5F,$BB
	.dc.b	$F0,$5F,$7B,$E0,$E5,$7D,$98,$FC,$0F,$C1,$FF,$9B,$3F,$69,$FE,$A1
	.dc.b	$3F,$94,$6E,$DF,$42,$7B,$4F,$DF,$C6,$CF,$D4,$C7,$BC,$FE,$17,$1E
	.dc.b	$EF,$E1,$71,$ED,$F0,$B2,$3F,$BA,$65,$FD,$CD,$BF,$0D,$D3,$BB,$F2
	.dc.b	$79,$D9,$FF,$06,$74,$FF,$A5,$1F,$51,$DC,$87,$73,$AF,$C2,$FB,$7E
	.dc.b	$1A,$46,$0F,$0F,$C2,$FC,$1F,$85,$91,$83,$CB,$EF,$FC,$5F,$83,$91
	.dc.b	$83,$D3,$F0,$7F,$24,$8F,$B3,$E0,$3E,$66,$1D,$23,$E6,$61,$87,$91
	.dc.b	$F3,$EF,$AB,$F4,$70,$FA,$9F,$0B,$F5,$7E,$16,$4F,$C8,$8F,$B1,$F0
	.dc.b	$B2,$3E,$5B,$F6,$FE,$17,$EE,$7C,$24,$9F,$9E,$FD,$FF,$84,$FC,$07
	.dc.b	$C0,$C9,$FA,$6F,$E0,$FE,$13,$A4,$7E,$0F,$0B,$D9,$FD,$E7,$A1,$97
	.dc.b	$5A,$FC,$46,$56,$B5,$97,$E3,$7E,$0B,$F1,$DF,$09,$2F,$F0,$31,$FE
	.dc.b	$BC,$FD,$AB,$FB,$2C,$7F,$8D,$DF,$EC,$6F,$F2,$66,$DF,$F8,$25,$DC
	.dc.b	$67,$E1,$39,$7C,$24,$8E,$51,$D7,$E0,$BB,$7C,$14,$9E,$D1,$E1,$F8
	.dc.b	$5F,$17,$BD,$27,$C5,$1E,$5B,$E9,$1E,$3C,$30,$F9,$FE,$0F,$D4,$47
	.dc.b	$D5,$F0,$7D,$5F,$66,$1D,$5F,$66,$1C,$9E,$87,$C1,$BC,$69,$EF,$F8
	.dc.b	$4F,$8C,$49,$F8,$DF,$0B,$F2,$7E,$17,$E5,$7C,$2C,$9F,$95,$1F,$37
	.dc.b	$E0,$64,$7C,$D7,$E8,$7C,$0F,$D1,$F8,$49,$3F,$45,$FA,$9F,$03,$D6
	.dc.b	$3E,$A6,$1D,$5F,$A9,$87,$47,$EB,$61,$73,$FA,$D8,$FD,$DF,$82,$95
	.dc.b	$F7,$5F,$C0,$7C,$2F,$E0,$7E,$13,$F0,$5F,$0B,$2B,$F0,$51,$F8,$5F
	.dc.b	$85,$FF,$80,$3D,$9F,$D5,$79,$0D,$EA,$4E,$BE,$0B,$7F,$0B,$31,$FF
	.dc.b	$41,$5F,$F5,$8D,$BF,$15,$97,$79,$FC,$76,$7E,$37,$F2,$58,$F8,$18
	.dc.b	$60,$E7,$F0,$5E,$4F,$83,$91,$83,$BF,$C2,$7A,$08,$F8,$BE,$17,$D7
	.dc.b	$F0,$B2,$3C,$AF,$BB,$E1,$7D,$FF,$07,$F0,$3E,$0E,$57,$AA,$39,$BF
	.dc.b	$E0,$AD,$D8,$E3,$7F,$F3,$74,$32,$ED,$5B,$CA,$D6,$B2,$F9,$BF,$05
	.dc.b	$2B,$E3,$BF,$43,$E0,$FA,$7B,$7F,$EC,$4C,$BB,$D7,$B8,$F6,$1E,$A2
	.dc.b	$1D,$CF,$A3,$7F,$6B,$E1,$3E,$DF,$C2,$C9,$FA,$B1,$F7,$BE,$13,$EF
	.dc.b	$FC,$2C,$9F,$B3,$1F,$82,$F7,$E4,$7D,$C7,$F0,$BF,$09,$F8,$6F,$83
	.dc.b	$93,$F8,$07,$F1,$3F,$09,$D6,$3F,$07,$87,$57,$E5,$61,$D1,$F9,$98
	.dc.b	$5E,$0F,$97,$8F,$E4,$BE,$17,$FD,$F1,$5F,$E3,$AF,$D7,$3F,$B3,$F0
	.dc.b	$7F,$BA,$1B,$1F,$EE,$3E,$05,$70,$C3,$FD,$45,$5F,$D8,$6B,$97,$C2
	.dc.b	$E7,$0E,$4D,$8F,$F7,$64,$EE,$B8,$FC,$26,$7C,$24,$FF,$B4,$EB,$FD
	.dc.b	$C0,$D7,$CD,$F0,$9B,$87,$C1,$52,$4F,$0F,$82,$D7,$C3,$6F,$39,$3D
	.dc.b	$23,$96,$52,$3B,$BD,$71,$ED,$F0,$5D,$FE,$0E,$57,$92,$3C,$5F,$0B
	.dc.b	$E3,$F8,$59,$3E,$88,$F3,$7C,$27,$9F,$E1,$64,$FB,$23,$D5,$F0,$7E
	.dc.b	$BF,$82,$93,$F0,$23,$AB,$F0,$B0,$EA,$FD,$7C,$3A,$BF,$1B,$0E,$AF
	.dc.b	$DB,$C3,$E3,$7C,$14,$8F,$96,$FC,$9E,$7E,$B1,$F8,$1C,$3A,$F2,$FD
	.dc.b	$CC,$7A,$CF,$E1,$71,$EB,$3F,$80,$C7,$E9,$7C,$2F,$D3,$F8,$59,$7F
	.dc.b	$62,$3E,$E7,$D3,$BC,$1B,$E1,$4A,$1F,$E8,$A9,$D3,$8C,$7F,$DD,$D0
	.dc.b	$EF,$E6,$49,$D7,$67,$2E,$C5,$3C,$0F,$C3,$7C,$24,$BF,$C5,$BF,$89
	.dc.b	$F8,$5F,$C8,$7E,$B9,$FD,$CB,$3F,$F7,$E3,$72,$E8,$DB,$64,$8F,$EF
	.dc.b	$73,$AF,$84,$E7,$F0,$32,$FC,$A7,$0A,$4F,$83,$E1,$7C,$3F,$0D,$2B
	.dc.b	$78,$F5,$9F,$2F,$C2,$49,$CE,$1E,$3A,$F1,$1E,$AC,$3C,$13,$FE,$B4
	.dc.b	$96,$39,$8E,$3F,$0D,$F0,$3E,$1A,$46,$0E,$BF,$0D,$F1,$3E,$1A,$46
	.dc.b	$0F,$0F,$C1,$FC,$82,$3E,$4F,$84,$F9,$7F,$09,$23,$CE,$FC,$EF,$84
	.dc.b	$F9,$F8,$74,$9F,$36,$18,$7A,$90,$F5,$61,$F5,$B0,$FA,$FF,$0D,$F6
	.dc.b	$3E,$1A,$57,$C4,$8F,$E1,$5F,$50,$F2,$2F,$93,$2B,$E6,$CE,$5B,$E2
	.dc.b	$70,$23,$AF,$17,$F0,$BF,$0F,$23,$E8,$BF,$88,$F8,$7E,$FE,$57,$83
	.dc.b	$CB,$95,$DA,$3C,$7E,$54,$8F,$0B,$F9,$0F,$84,$FC,$8F,$C2,$7E,$4B
	.dc.b	$E1,$25,$7D,$E8,$F5,$F8,$9E,$CF,$12,$4F,$AA,$3F,$94,$7E,$B5,$FE
	.dc.b	$0B,$24,$63,$E1,$F3,$F0,$FA,$F8,$79,$3F,$8E,$AF,$F6,$93,$FE,$FA
	.dc.b	$FA,$97,$8D,$FC,$CD,$78,$F4,$C4,$78,$3E,$23,$C3,$F0,$D2,$FA,$47
	.dc.b	$93,$E1,$A4,$77,$7C,$FF,$0D,$D7,$C0,$93,$FA,$0C,$91,$E1,$F0,$3D
	.dc.b	$BF,$09,$EE,$F8,$79,$7E,$98,$F8,$3F,$11,$23,$D8,$FC,$4F,$88,$F5
	.dc.b	$76,$C9,$EC,$FB,$7B,$7E,$4F,$C4,$7C,$AF,$5E,$57,$C5,$8F,$9B,$EB
	.dc.b	$C8,$F9,$0F,$D0,$F5,$FF,$AC,$78,$3F,$E3,$3F,$53,$F8,$93,$F3,$3D
	.dc.b	$54,$63,$AF,$7D,$9F,$63,$D6,$99,$7D,$37,$ED,$FC,$37,$DC,$F8,$69
	.dc.b	$3F,$5D,$FB,$FF,$11,$E0,$E9,$F0,$F2,$CA,$FE,$13,$AA,$7F,$0B,$F1
	.dc.b	$1F,$86,$F8,$89,$5F,$81,$8F,$EE,$19,$7F,$7E,$F6,$8D,$27,$C2,$7B
	.dc.b	$B9,$65,$F6,$7E,$0F,$27,$2F,$85,$E7,$F0,$B2,$BF,$8D,$E0,$9E,$FF
	.dc.b	$0B,$E0,$F8,$79,$5A,$8F,$97,$C3,$23,$D6,$FC,$ED,$7C,$FF,$0F,$E8
	.dc.b	$F5,$65,$76,$8F,$5F,$AB,$EC,$F5,$24,$F8,$A3,$DF,$EA,$48,$F2,$BF
	.dc.b	$0B,$E1,$7E,$1F,$C3,$7C,$4F,$5A,$57,$AA,$3E,$3F,$44,$8F,$6B,$F2
	.dc.b	$BE,$1F,$E5,$FA,$FF,$33,$E1,$A5,$7C,$28,$F9,$FF,$0F,$23,$E2,$BF
	.dc.b	$4B,$D5,$FA,$7E,$B7,$F8,$8F,$F5,$9F,$E7,$C7,$F3,$FE,$84,$67,$B6
	.dc.b	$61,$FD,$22,$BF,$A9,$BF,$D5,$9F,$E9,$EE,$7E,$13,$A7,$AB,$23,$07
	.dc.b	$1F,$5B,$C1,$F0,$92,$30,$75,$F8,$4F,$21,$1F,$07,$C1,$79,$FE,$0A
	.dc.b	$47,$8D,$F5,$7A,$DE,$BF,$5B,$D9,$F0,$92,$BD,$11,$EF,$F8,$49,$1E
	.dc.b	$B7,$E1,$7C,$1F,$C3,$F8,$4F,$89,$EB,$CA,$F8,$11,$F1,$FD,$79,$1F
	.dc.b	$0D,$F9,$5F,$07,$F2,$FE,$0B,$E6,$7C,$1C,$AF,$91,$1F,$E5,$1F,$D5
	.dc.b	$FF,$D6,$84,$8C,$7A,$D9,$F5,$F5,$F0,$92,$7F,$BB,$D7,$EF,$FD,$9F
	.dc.b	$F0,$92,$F8,$2E,$9F,$07,$2B,$71,$DF,$E0,$FC,$1E,$CC,$9E,$71,$E3
	.dc.b	$F6,$A4,$76,$7C,$DF,$05,$E7,$F8,$29,$3E,$27,$D5,$F0,$5E,$BF,$82
	.dc.b	$93,$E6,$7D,$DE,$E7,$BF,$DC,$FE,$2B,$2F,$81,$BE,$3B,$83,$FC,$61
	.dc.b	$C3,$EE,$BD,$D6,$A7,$37,$C6,$BE,$0F,$C0,$E9,$A7,$C6,$F2,$B1,$3F
	.dc.b	$E2,$AA,$FF,$1B,$4F,$CA,$F8,$1C,$E5,$E1,$F6,$ED,$26,$C6,$3D,$BC
	.dc.b	$FC,$0D,$C3,$E9,$BC,$BF,$4C,$FE,$63,$1F,$E2,$C6,$E5,$E7,$3F,$C7
	.dc.b	$6B,$E0,$B8,$4F,$1F,$81,$8F,$1F,$82,$B7,$C7,$56,$0F,$81,$F0,$7C
	.dc.b	$17,$85,$B8,$78,$7E,$07,$C9,$F0,$5E,$56,$C7,$97,$E0,$7D,$07,$A7
	.dc.b	$E0,$AC,$7A,$7E,$0B,$D9,$F0,$5E,$D6,$C7,$B7,$E0,$7E,$01,$F0,$7E
	.dc.b	$0A,$C7,$C1,$F8,$2F,$88,$7C,$5F,$82,$B1,$F1,$7E,$07,$E4,$7C,$17
	.dc.b	$C9,$6C,$7C,$9F,$82,$F9,$87,$CD,$F8,$2B,$1F,$37,$E0,$BE,$81,$F6
	.dc.b	$BD,$37,$C4,$E2,$7F,$CB,$B1,$FE,$6A,$6F,$3E,$FE,$27,$83,$C7,$E0
	.dc.b	$E4,$F0,$F8,$3B,$D7,$F4,$70,$FF,$0A,$CF,$4F,$82,$F0,$7C,$1F,$85
	.dc.b	$F1,$7C,$1F,$8F,$E0,$ED,$78,$9F,$37,$C1,$F9,$FE,$0E,$CF,$99,$F5
	.dc.b	$7C,$17,$AF,$E0,$EC,$FA,$9F,$77,$C1,$FB,$FE,$0E,$CF,$B9,$F8,$5F
	.dc.b	$05,$23,$FC,$FB,$F5,$2C,$B1,$C1,$EB,$F8,$4F,$90,$7C,$9E,$1B,$7F
	.dc.b	$23,$E1,$3E,$61,$F3,$7E,$0E,$C7,$4C,$BF,$53,$91,$B9,$E1,$FA,$9F
	.dc.b	$09,$2B,$FD,$1F,$FA,$5F,$FA,$C9,$FA,$B1,$FB,$28,$7C,$2D,$31,$F0
	.dc.b	$B6,$BD,$75,$E0,$3F,$73,$F5,$5C,$F7,$FD,$36,$69,$E9,$F0,$D2,$FC
	.dc.b	$3F,$0B,$6F,$F5,$99,$BE,$1B,$C5,$F0,$BE,$36,$DF,$FA,$9B,$3D,$36
	.dc.b	$7C,$33,$E2,$7D,$5F,$0D,$27,$DF,$F0,$D7,$1E,$FF,$0D,$EF,$3E,$07
	.dc.b	$C3,$5A,$ED,$F0,$DE,$CF,$86,$F6,$B6,$3E,$4B,$F1,$E3,$E4,$36,$3E
	.dc.b	$6F,$C3,$7C,$BF,$86,$F9,$8D,$8F,$57,$C2,$7C,$83,$E4,$FC,$1D,$8F
	.dc.b	$AB,$F1,$1F,$4C,$FA,$9F,$0D,$63,$EC,$D7,$D7,$3E,$C5,$58,$FF,$65
	.dc.b	$FD,$57,$F1,$D6,$FC,$1F,$F2,$E8,$FF,$1A,$46,$0F,$C1,$FC,$37,$E0
	.dc.b	$7E,$1B,$F0,$4D,$8F,$C3,$FC,$37,$F3,$6F,$D2,$FF,$95,$FC,$4F,$F4
	.dc.b	$8E,$A4,$E6,$BB,$7C,$1F,$7F,$82,$B1,$E4,$79,$69,$23,$FD,$C3,$8F
	.dc.b	$93,$9F,$F9,$73,$E9,$6F,$C3,$F8,$7E,$17,$C4,$DA,$F7,$FC,$07,$B7
	.dc.b	$E0,$3D,$CD,$8E,$BF,$0D,$E9,$3D,$5F,$0D,$63,$C3,$AF,$B7,$5F,$73
	.dc.b	$63,$E5,$FB,$DF,$27,$DE,$F9,$4D,$8F,$4F,$C3,$7C,$53,$E3,$7C,$2D
	.dc.b	$8F,$6F,$C3,$7C,$9F,$86,$F9,$4D,$8F,$F8,$03,$E7,$FF,$AC,$9B,$EE
	.dc.b	$FF,$64,$8C,$71,$FF,$51,$7F,$90,$B7,$DD,$D1,$BF,$76,$DF,$CC,$E2
	.dc.b	$FB,$1C,$7F,$65,$B1,$F4,$3E,$1B,$EE,$7C,$2F,$DD,$6C,$7D,$4F,$86
	.dc.b	$FC,$01,$F8,$1F,$86,$B1,$F6,$38,$BF,$09,$C7,$F8,$56,$C7,$53,$EE
	.dc.b	$E1,$D5,$A7,$52,$6B,$F8,$0B,$EA,$FF,$6F,$FD,$2F,$FA,$5C,$A3,$1F
	.dc.b	$0B,$FE,$08,$9C,$C6,$27,$C0,$FE,$2A,$FC,$0D,$8E,$FE,$07,$F8,$83
	.dc.b	$B5,$A7,$B6,$72,$77,$C5,$8B,$E1,$BC,$1F,$0D,$E1,$6D,$7D,$AF,$77
	.dc.b	$EC,$7B,$BF,$65,$B1,$F7,$BD,$9F,$B8,$7F,$88,$3E,$B3,$F9,$46,$BF
	.dc.b	$E3,$2E,$8F,$E5,$15,$EF,$8F,$F1,$AE,$7F,$82,$6F,$C3,$FC,$1E,$AF
	.dc.b	$84,$FC,$3E,$AF,$89,$D5,$70,$E4,$7A,$F0,$47,$C9,$F8,$7F,$95,$F0
	.dc.b	$76,$B7,$CA,$E3,$F8,$33,$F3,$FE,$1F,$E8,$7C,$3D,$BD,$D7,$C6,$8F
	.dc.b	$05,$FD,$9C,$6E,$DE,$A7,$B4,$7D,$DC,$6C,$FB,$9F,$B9,$F0,$FF,$77
	.dc.b	$E1,$EE,$5D,$B5,$47,$E0,$7E,$17,$F0,$5C,$56,$B4,$FE,$17,$E1,$FF
	.dc.b	$0D,$F0,$F6,$79,$3F,$89,$E7,$91,$D7,$A2,$ED,$F9,$1C,$7A,$CF,$85
	.dc.b	$B1,$F9,$5C,$3A,$CF,$95,$B1,$F9,$9E,$8F,$CB,$1F,$E5,$EE,$BF,$BF
	.dc.b	$5F,$4E,$31,$F0,$F2,$3E,$5C,$BE,$1F,$17,$FE,$51,$87,$C3,$73,$3A
	.dc.b	$7C,$35,$BE,$3F,$0B,$DC,$F0,$7C,$2D,$8E,$B9,$74,$E6,$FF,$40,$E3
	.dc.b	$DE,$72,$78,$B1,$E9,$97,$87,$2B,$7C,$3E,$17,$D8,$7B,$7E,$16,$E1
	.dc.b	$CA,$FD,$19,$7C,$1D,$7C,$13,$FA,$DC,$FD,$6F,$73,$E0,$E5,$D9,$D7
	.dc.b	$C2,$FC,$93,$E5,$7C,$2D,$8F,$8F,$F0,$9F,$34,$F9,$DF,$0B,$63,$E5
	.dc.b	$FC,$27,$D1,$3E,$97,$C2,$D8,$F9,$FF,$0B,$F5,$4F,$AD,$F0,$B6,$3A
	.dc.b	$93,$7F,$A5,$87,$57,$EA,$61,$CD,$FA,$D8,$5E,$2F,$9B,$F0,$BF,$81
	.dc.b	$3F,$05,$F0,$B7,$1F,$BF,$F0,$7F,$85,$3F,$0D,$F0,$96,$3F,$07,$F0
	.dc.b	$7F,$89,$3F,$D4,$FF,$A5,$FF,$72,$8F,$95,$18,$F8,$4C,$FC,$5D,$9F
	.dc.b	$DA,$D7,$F6,$DC,$FE,$FE,$3D,$75,$FC,$66,$56,$7F,$07,$97,$7C,$38
	.dc.b	$96,$BF,$D4,$58,$3E,$13,$C7,$F0,$BE,$46,$D7,$83,$E1,$7C,$FF,$0B
	.dc.b	$E8,$6C,$79,$3E,$13,$D7,$F0,$BE,$C6,$C7,$A3,$E1,$7D,$E7,$C0,$F8
	.dc.b	$5B,$1D,$4F,$6E,$1D,$5A,$75,$3E,$0D,$F5,$69,$F1,$3E,$13,$E5,$FC
	.dc.b	$2F,$CC,$6C,$75,$DB,$DF,$7D,$5B,$1D,$5F,$87,$7D,$5B,$1F,$43,$E1
	.dc.b	$3E,$BF,$C2,$FD,$86,$C7,$D4,$F8,$5F,$B6,$7D,$CF,$82,$B1,$F6,$3C
	.dc.b	$4F,$BE,$7E,$03,$E0,$EC,$7D,$CC,$3F,$07,$F0,$9D,$32,$FA,$AF,$27
	.dc.b	$EB,$E3,$D6,$BC,$07,$52,$47,$DC,$BE,$AF,$DE,$F8,$5F,$C7,$9F,$EF
	.dc.b	$6F,$D2,$FF,$B7,$F4,$AF,$FB,$37,$42,$B5,$EE,$C9,$FE,$8B,$7F,$DA
	.dc.b	$5B,$1C,$FE,$17,$A7,$C1,$59,$DB,$DF,$E0,$FC,$1F,$07,$67,$9B,$D7
	.dc.b	$4F,$F7,$A6,$3F,$EC,$D6,$FC,$2E,$75,$F3,$47,$F5,$B9,$65,$E0,$C3
	.dc.b	$CA,$D9,$E7,$F0,$3E,$FF,$84,$F8,$0D,$AF,$5F,$C1,$7C,$33,$E2,$7C
	.dc.b	$25,$8F,$7F,$C2,$7C,$7F,$84,$F9,$0D,$8F,$87,$F0,$7F,$2C,$E9,$97
	.dc.b	$A9,$E4,$FB,$30,$EB,$3E,$03,$A9,$F3,$2F,$AB,$4F,$9F,$EF,$FD,$73
	.dc.b	$EC,$7C,$1D,$8E,$57,$D5,$FA,$B9,$78,$37,$F8,$FC,$8E,$06,$BE,$9C
	.dc.b	$3E,$7B,$6B,$72,$F8,$0F,$C3,$7C,$1F,$E1,$DB,$97,$E0,$BE,$0F,$F1
	.dc.b	$47,$E2,$FE,$0E,$C7,$E1,$BE,$0F,$F1,$C7,$D8,$C7,$C7,$FB,$5F,$F6
	.dc.b	$8D,$3E,$D4,$E6,$BF,$C6,$D3,$FE,$D8,$7F,$C2,$6F,$F5,$B9,$57,$2F
	.dc.b	$7F,$84,$7F,$87,$67,$53,$EA,$C7,$C7,$3F,$8C,$C6,$CF,$BB,$1F,$2E
	.dc.b	$3D,$AA,$9D,$9F,$1F,$C1,$76,$DD,$93,$DF,$C3,$FD,$82,$F8,$1E,$38
	.dc.b	$CF,$C2,$4C,$3D,$7F,$09,$6B,$78,$74,$F8,$4E,$7F,$09,$D1,$B7,$F0
	.dc.b	$B1,$EF,$F0,$7E,$06,$C7,$C6,$CB,$A6,$DF,$3B,$3B,$3F,$33,$0E,$79
	.dc.b	$F8,$34,$46,$2F,$EA,$FC,$1F,$B0,$F6,$FC,$1D,$C3,$E9,$FC,$0F,$C0
	.dc.b	$F8,$1F,$82,$D8,$FA,$FF,$03,$F1,$0F,$8B,$F0,$76,$3E,$DF,$C1,$FC
	.dc.b	$8F,$83,$F9,$2D,$8E,$A7,$E0,$30,$EA,$D3,$A9,$36,$F5,$5F,$57,$D9
	.dc.b	$7C,$DF,$75,$DD,$3F,$23,$97,$9E,$7E,$13,$63,$F2,$B9,$73,$7F,$2F
	.dc.b	$85,$C7,$F2,$BF,$07,$F8,$0F,$81,$FC,$0B,$63,$F2,$FF,$03,$F8,$43
	.dc.b	$F0,$BF,$03,$63,$FC,$DB,$3F,$DF,$F3,$FF,$3A,$3F,$EE,$5C,$7F,$6E
	.dc.b	$E7,$A0,$E6,$BF,$D3,$4F,$F8,$05,$BF,$05,$E1,$BF,$C5,$CE,$5F,$D2
	.dc.b	$63,$E1,$7F,$01,$8D,$C7,$BE,$1F,$9D,$BE,$D5,$E4,$8C,$7C,$1E,$7E
	.dc.b	$0B,$5F,$07,$67,$D6,$F1,$F8,$39,$1E,$EF,$7A,$E1,$E4,$F8,$3E,$C7
	.dc.b	$7F,$7E,$D7,$23,$E2,$FC,$1F,$66,$9F,$23,$0E,$7A,$FC,$6C,$6C,$FC
	.dc.b	$4C,$39,$E3,$F0,$B1,$B7,$F3,$3E,$0F,$DD,$F0,$5E,$F6,$E1,$F5,$3E
	.dc.b	$0F,$E1,$1F,$0F,$E0,$EC,$7D,$8F,$81,$F8,$DF,$01,$F1,$DB,$1F,$73
	.dc.b	$0F,$95,$F0,$7D,$34,$FA,$6F,$27,$EA,$E3,$D6,$BC,$07,$53,$F0,$B8
	.dc.b	$75,$69,$F8,$8A,$EA,$F7,$30,$72,$AE,$AF,$E3,$B1,$F0,$67,$F8,$CC
	.dc.b	$BC,$F5,$F8,$66,$C7,$E4,$3E,$0F,$F0,$5F,$05,$F8,$36,$E1,$F9,$8F
	.dc.b	$83,$FC,$37,$C1,$7E,$1D,$B1,$FE,$C8,$FD,$B3,$FC,$16,$FF,$BB,$DF
	.dc.b	$EF,$58,$7F,$52,$73,$97,$FB,$4D,$CC,$FE,$73,$1F,$E5,$D1,$FE,$0D
	.dc.b	$9F,$CD,$C6,$3E,$03,$C1,$5F,$AA,$C6,$D7,$E8,$1E,$3F,$1B,$CB,$BE
	.dc.b	$D7,$95,$EB,$F0,$32,$3D,$1F,$09,$70,$FE,$67,$9B,$E1,$3C,$67,$93
	.dc.b	$E1,$2D,$FB,$FE,$0F,$CF,$F0,$9E,$86,$C7,$C3,$F8,$3F,$5F,$C0,$7B
	.dc.b	$1B,$1F,$1F,$E0,$3D,$FF,$07,$F0,$1B,$1F,$2F,$E1,$3E,$1F,$C2,$7C
	.dc.b	$46,$C7,$CF,$F8,$4F,$8F,$F0,$5F,$21,$B1,$F4,$FE,$0F,$E5,$FC,$17
	.dc.b	$CC,$6C,$7D,$7F,$82,$F9,$E7,$D0,$F8,$2B,$1F,$6F,$0F,$EC,$1F,$B6
	.dc.b	$FF,$77,$E8,$72,$FF,$81,$69,$CD,$FD,$6A,$B5,$2C,$BB,$CF,$CC,$6F
	.dc.b	$6F,$E0,$F3,$EF,$3F,$43,$3B,$3F,$E1,$8C,$3F,$C0,$53,$D7,$E1,$3C
	.dc.b	$7F,$0F,$E4,$7C,$BF,$0B,$E6,$F8,$3B,$AF,$93,$1E,$51,$EA,$F8,$3F
	.dc.b	$5F,$C1,$DA,$D1,$EA,$E0,$68,$9F,$83,$F0,$5F,$0B,$E1,$6D,$F0,$D9
	.dc.b	$87,$09,$FF,$09,$FE,$9A,$FB,$3F,$E7,$37,$BA,$EB,$F0,$FA,$9C,$E1
	.dc.b	$F2,$63,$E7,$FC,$34,$BE,$1A,$3E,$1B,$17,$F3,$FE,$0F,$EA,$FC,$37
	.dc.b	$D6,$6D,$F9,$74,$7C,$1F,$F4,$36,$FB,$BE,$68,$C1,$FE,$C1,$9E,$FF
	.dc.b	$09,$F8,$08,$FC,$0B,$63,$C7,$F0,$BF,$84,$3F,$0B,$F0,$D6,$3C,$14
	.dc.b	$F6,$FB,$7B,$FD,$DB,$3E,$CF,$82,$FC,$6F,$C3,$7E,$39,$B1,$F0,$3E
	.dc.b	$17,$F2,$27,$E4,$BE,$16,$C7,$6F,$6B,$D5,$EE,$FA,$DB,$1F,$3F,$F6
	.dc.b	$EF,$ED,$44,$F8,$6F,$F1,$D3,$8F,$87,$A6,$38,$AD,$7A,$BD,$AF,$B7
	.dc.b	$F2,$3F,$C6,$39,$7E,$5C,$63,$DB,$E7,$F1,$12,$FC,$1F,$11,$7A,$FF
	.dc.b	$C7,$BA,$3E,$23,$C2,$78,$BE,$22,$DF,$D7,$D3,$F8,$83,$F8,$DB,$7E
	.dc.b	$FB,$FE,$5C,$8F,$D7,$EA,$AD,$D7,$F9,$40,$DF,$9B,$EE,$E8,$97,$F0
	.dc.b	$FA,$6D,$79,$3E,$27,$E1,$FC,$4F,$C4,$6D,$78,$73,$FF,$22,$BE,$3E
	.dc.b	$4F,$F4,$4B,$9F,$27,$FC,$E3,$82,$72,$FC,$0F,$89,$F9,$E7,$D0,$F8
	.dc.b	$8B,$7F,$57,$A7,$E9,$F5,$7D,$46,$C7,$7F,$73,$AE,$27,$BE,$6F,$13
	.dc.b	$31,$F7,$3E,$26,$5F,$CB,$F6,$AE,$BF,$4B,$E2,$7F,$03,$F1,$1F,$82
	.dc.b	$6D,$7B,$7D,$DE,$E7,$6F,$1E,$C7,$D5,$F6,$7F,$06,$7E,$13,$DD,$B1
	.dc.b	$F8,$FF,$89,$FC,$6F,$87,$F8,$E6,$C7,$C9,$F7,$FD,$67,$C8,$FD,$FB
	.dc.b	$FC,$85,$AF,$C9,$8F,$E1,$3F,$27,$FC,$83,$DB,$FF,$6F,$F8,$28,$C6
	.dc.b	$1F,$82,$7F,$09,$69,$D7,$BB,$CB,$DD,$FF,$4C,$6B,$F9,$9B,$FF,$4B
	.dc.b	$63,$FE,$C4,$8F,$EE,$78,$7D,$9E,$3F,$9E,$79,$7C,$7B,$9F,$F0,$BC
	.dc.b	$1E,$E7,$98,$F3,$FB,$96,$75,$E0,$7A,$BD,$CF,$5B,$63,$F1,$5F,$03
	.dc.b	$F7,$0F,$85,$E1,$D8,$ED,$ED,$FC,$23,$E1,$FB,$76,$3C,$5E,$D7,$C6
	.dc.b	$F0,$3E,$3B,$63,$F2,$DF,$01,$F8,$83,$E7,$78,$B6,$3D,$5E,$DF,$CE
	.dc.b	$3E,$7F,$B7,$63,$DD,$ED,$7D,$2F,$6F,$E9,$B6,$3F,$45,$CF,$F9,$43
	.dc.b	$ED,$77,$D8,$F8,$DE,$0F,$DA,$3E,$DF,$B5,$63,$E5,$7B,$7F,$78,$FB
	.dc.b	$FE,$DD,$8F,$E0,$1F,$BF,$BE,$2B,$F5,$AD,$8C,$FC,$56,$27,$F0,$FE
	.dc.b	$F4,$9F,$E8,$1F,$29,$FE,$8D,$F1,$7F,$11,$3F,$D5,$23,$FC,$FB,$07
	.dc.b	$FA,$21,$83,$FD,$11,$E1,$7E,$EF,$C9,$FD,$89,$E2,$F8,$8B,$87,$6F
	.dc.b	$7F,$A1,$D7,$DE,$B1,$E2,$F1,$7C,$1E,$EF,$85,$B1,$FC,$2F,$E1,$7F
	.dc.b	$7A,$7B,$BC,$8B,$1E,$AF,$7B,$D0,$7A,$7D,$CB,$1E,$EF,$73,$D9,$E2
	.dc.b	$FB,$5B,$1F,$CA,$7F,$5C,$FF,$0A,$B7,$83,$E5,$E5,$FC,$62,$73,$1F
	.dc.b	$EF,$8E,$2F,$89,$EE,$FC,$73,$E4,$7B,$B7,$0F,$9B,$EE,$FC,$BF,$77
	.dc.b	$E6,$36,$3C,$1E,$9F,$53,$F9,$2F,$43,$E1,$B1,$3F,$5B,$C7,$FA,$87
	.dc.b	$D5,$F6,$EC,$7D,$AF,$6B,$EC,$7B,$7F,$65,$B1,$ED,$F7,$FD,$47,$6F
	.dc.b	$42,$C7,$E0,$BD,$BF,$C0,$1F,$81,$F2,$2C,$7F,$81,$7A,$BF,$AE,$1F
	.dc.b	$E7,$2E,$4F,$E1,$4E,$3E,$1B,$E3,$7B,$52,$BF,$AA,$70,$3C,$EC,$61
	.dc.b	$F2,$70,$FF,$3C,$7B,$5F,$C6,$1B,$7F,$D6,$17,$C6,$FF,$89,$9C,$D7
	.dc.b	$83,$C7,$C7,$91,$96,$DF,$DB,$E5,$FB,$07,$CC,$F2,$EC,$79,$3C,$BF
	.dc.b	$9E,$75,$F8,$DB,$1E,$FF,$8C,$F0,$7C,$6F,$85,$B1,$F8,$7F,$77,$F0
	.dc.b	$87,$F5,$CF,$E0,$0F,$F7,$1E,$AF,$0F,$F9,$0D,$6C,$91,$FD,$C6,$7F
	.dc.b	$53,$3A,$8F,$89,$F1,$DE,$FF,$3B,$E0,$36,$FF,$2D,$EF,$FE,$4C,$F0
	.dc.b	$79,$96,$32,$4B,$ED,$E6,$E5,$FA,$1F,$1D,$F2,$FE,$17,$E6,$36,$BF
	.dc.b	$45,$F0,$1F,$9F,$3D,$9E,$6D,$8C,$92,$FD,$5E,$76,$5F,$B9,$F0,$BF
	.dc.b	$5F,$E1,$3E,$C3,$6B,$F5,$DF,$07,$FA,$B3,$E4,$79,$96,$32,$4B,$F8
	.dc.b	$DE,$6E,$5F,$C4,$77,$7C,$FF,$2F,$F0,$8D,$AF,$DD,$79,$1F,$B7,$3E
	.dc.b	$A7,$95,$63,$24,$BF,$A5,$E5,$E5,$FA,$9E,$57,$EF,$FE,$0B,$EE,$36
	.dc.b	$BF,$83,$45,$9F,$B5,$E5,$DF,$1D,$5F,$E9,$E9,$FC,$15,$66,$31,$FC
	.dc.b	$03,$FD,$AB,$B5,$3F,$C3,$7E,$0A,$57,$F8,$EE,$BF,$5D,$7F,$8B,$9C
	.dc.b	$BF,$E3,$9B,$FF,$3F,$56,$FE,$43,$38,$F8,$AB,$FD,$5F,$5A,$78,$5F
	.dc.b	$F8,$CB,$0F,$D6,$E1,$FC,$86,$3F,$8F,$7C,$24,$8F,$17,$93,$6F,$C1
	.dc.b	$37,$1E,$D0,$C3,$FC,$9B,$18,$C7,$FC,$61,$87,$F2,$DB,$E0,$FC,$2F
	.dc.b	$1A,$E1,$E6,$F8,$3F,$E5,$C7,$97,$C6,$B5,$FC,$D6,$2C,$FF,$96,$72
	.dc.b	$78,$F8,$CF,$C1,$56,$FE,$7F,$89,$70,$F8,$7E,$6F,$F3,$B3,$E1,$78
	.dc.b	$D6,$BF,$A1,$45,$9E,$DE,$15,$C7,$D9,$36,$BE,$DF,$87,$6F,$E7,$FC
	.dc.b	$17,$F4,$73,$E7,$78,$16,$BF,$A7,$45,$9F,$57,$83,$71,$F9,$13,$6B
	.dc.b	$F0,$FD,$B6,$FE,$DF,$C1,$FF,$53,$3E,$D7,$C2,$DA,$FE,$B5,$16,$7E
	.dc.b	$37,$C2,$DC,$7E,$A4,$DA,$FC,$9F,$C2,$DB,$FC,$3F,$C2,$FF,$5F,$3F
	.dc.b	$0D,$F0,$56,$BF,$B3,$45,$9F,$A5,$EF,$5C,$7F,$01,$36,$BF,$45,$FA
	.dc.b	$AF,$F6,$89,$CC,$7F,$6F,$F8,$6F,$ED,$87,$FA,$12,$7F,$CA,$F8,$7E
	.dc.b	$62,$33,$3B,$3F,$D3,$D3,$FE,$5A,$9F,$E5,$D3,$A7,$F4,$8F,$3F,$67
	.dc.b	$17,$F9,$3F,$57,$FB,$D1,$DF,$D6,$B8,$FF,$80,$62,$E1,$FE,$A5,$C7
	.dc.b	$F9,$A5,$F1,$AD,$12,$75,$E9,$F0,$9F,$D0,$7A,$DF,$E0,$73,$D9,$E9
	.dc.b	$5B,$FF,$07,$45,$BE,$DE,$85,$BE,$35,$A3,$DF,$E6,$E5,$FE,$19,$E2
	.dc.b	$BD,$9F,$F0,$A1,$F2,$7C,$AB,$3F,$E1,$B8,$B8,$FA,$FC,$9B,$8F,$ED
	.dc.b	$66,$D7,$D6,$F2,$2D,$87,$EB,$DF,$E9,$03,$FD,$23,$C0,$F1,$7F,$D7
	.dc.b	$71,$9B,$A7,$DB,$F0,$AB,$FE,$D5,$8D,$3F,$6B,$BF,$38,$7F,$B4,$E7
	.dc.b	$FB,$49,$FD,$AB,$07,$67,$6F,$8F,$98,$73,$7F,$11,$D5,$8C,$39,$45
	.dc.b	$DB,$F8,$AE,$0E,$AC,$61,$FC,$D7,$2F,$F2,$C9,$FE,$5A,$97,$47,$B3
	.dc.b	$E3,$25,$7A,$5F,$CB,$73,$E3,$3F,$44,$5C,$BB,$F1,$5B,$ED,$F0,$FF
	.dc.b	$18,$FE,$69,$3F,$EB,$6B,$F9,$91,$99,$F8,$CE,$9F,$F7,$CC,$7C,$48
	.dc.b	$D3,$F1,$DD,$3F,$EB,$B3,$D6,$FD,$4F,$87,$93,$F6,$EB,$C1,$38,$AF
	.dc.b	$A1,$56,$7C,$8D,$AF,$8C,$D9,$ED,$ED,$57,$F0,$AF,$E0,$8F,$1F,$C2
	.dc.b	$FE,$27,$DB,$C5,$FE,$06,$6D,$7F,$9A,$47,$C3,$E3,$1F,$E9,$7C,$0A
	.dc.b	$C5,$FC,$2F,$87,$B8,$7E,$56,$98,$FE,$48,$F0,$C3,$DD,$C5,$FE,$46
	.dc.b	$6C,$F8,$7D,$DB,$97,$82,$6D,$7D,$0F,$87,$B7,$DF,$4F,$CF,$9D,$BE
	.dc.b	$1E,$D7,$E7,$62,$CF,$B7,$DF,$B8,$FB,$26,$D7,$DC,$F7,$ED,$FA,$F6
	.dc.b	$FD,$59,$EA,$F8,$6B,$5F,$A9,$8B,$3F,$27,$E0,$2E,$3F,$22,$6D,$7E
	.dc.b	$23,$E0,$2D,$FE,$EF,$F2,$AF,$EE,$B3,$FE,$15,$6C,$7E,$27,$E0,$7F
	.dc.b	$16,$A3,$3C,$BF,$C0,$FE,$07,$E6,$9F,$3B,$E0,$6E,$1F,$8D,$3F,$84
	.dc.b	$48,$7C,$87,$E1,$1B,$1B,$C3,$51,$FE,$8B,$38,$4F,$F0,$CE,$69,$5B
	.dc.b	$9E,$9F,$05,$8B,$FE,$11,$57,$6F,$F1,$05,$78,$7E,$1B,$18,$7E,$FE
	.dc.b	$AD,$F8,$3E,$0E,$BF,$C7,$BA,$7F,$8D,$9E,$7F,$86,$B7,$FC,$66,$2C
	.dc.b	$FF,$B0,$71,$7C,$26,$33,$FC,$CD,$5B,$F8,$74,$F8,$4C,$61,$F9,$9E
	.dc.b	$BF,$E5,$27,$C5,$F8,$3B,$7F,$C9,$E2,$CF,$7F,$83,$B9,$7E,$9A,$6D
	.dc.b	$76,$F8,$3B,$7F,$A5,$EE,$FE,$6A,$7D,$1F,$83,$B5,$FC,$CE,$2C,$FA
	.dc.b	$FE,$0E,$E3,$FB,$29,$B5,$EA,$F8,$3B,$7F,$B1,$F0,$7F,$A0,$9F,$77
	.dc.b	$E0,$ED,$7F,$3F,$8B,$3F,$1F,$E0,$EE,$3F,$BE,$9B,$5F,$1B,$E0,$ED
	.dc.b	$FE,$F7,$C3,$FE,$9A,$7E,$27,$E0,$ED,$7F,$4B,$8B,$3F,$4F,$E0,$EE
	.dc.b	$3F,$C2,$A6,$D7,$D2,$F8,$3B,$7F,$C2,$7C,$6F,$EB,$27,$E5,$7E,$0E
	.dc.b	$D7,$F5,$78,$B3,$FE,$FF,$C3,$FC,$35,$3F,$D2,$67,$31,$FC,$61,$FE
	.dc.b	$42,$FF,$86,$A7,$FB,$4C,$7F,$66,$F1,$E5,$7E,$6E,$BF,$C3,$5E,$52
	.dc.b	$73,$7F,$D8,$DE,$7F,$07,$8A,$FE,$4B,$16,$FF,$5F,$0F,$83,$C5,$FF
	.dc.b	$25,$F2,$7F,$BA,$9E,$4F,$86,$B7,$FD,$CE,$2C,$FB,$7E,$12,$E1,$FC
	.dc.b	$CA,$6D,$76,$F8,$4B,$7F,$CC,$7C,$CF,$F0,$09,$F0,$3E,$12,$D7,$F7
	.dc.b	$F8,$B3,$F2,$7E,$12,$E3,$FC,$FA,$6D,$7A,$BE,$0E,$DF,$F3,$DF,$3B
	.dc.b	$FC,$1A,$7C,$CF,$83,$B5,$FE,$0B,$8B,$3F,$57,$E0,$EE,$3F,$D2,$A6
	.dc.b	$D7,$C6,$F8,$3B,$7F,$D2,$7D,$1F,$F0,$C9,$F6,$3E,$0E,$D7,$F8,$5E
	.dc.b	$2C,$FE,$07,$E0,$EE,$3F,$D5,$A6,$D7,$D2,$F8,$3B,$7F,$E2,$9F,$D7
	.dc.b	$BF,$8F,$AB,$FF,$53,$C6,$1F,$E2,$E3,$E1,$3F,$98,$C6,$6F,$FD,$94
	.dc.b	$DA,$E9,$EB,$7E,$3B,$E0,$F6,$CA,$FE,$59,$2A,$E1,$E8,$78,$7E,$0B
	.dc.b	$53,$E3,$AF,$E8,$5E,$9F,$F4,$0F,$67,$F8,$DC,$7E,$57,$E0,$B1,$3F
	.dc.b	$9C,$F8,$1F,$EF,$BF,$B9,$FF,$B7,$AB,$F3,$51,$87,$3F,$0B,$AF,$84
	.dc.b	$B1,$FC,$E6,$91,$EE,$F8,$99,$39,$F8,$5B,$A7,$94,$F9,$53,$8E,$3F
	.dc.b	$E7,$B8,$3E,$17,$18,$7C,$6A,$B7,$FD,$46,$BF,$C1,$53,$F1,$38,$7F
	.dc.b	$DD,$63,$E0,$7E,$79,$EC,$F8,$4B,$3B,$8B,$7E,$33,$E6,$61,$8C,$BE
	.dc.b	$DC,$DA,$FE,$C3,$8F,$F8,$71,$E4,$78,$7E,$0B,$EE,$9F,$DB,$BE,$57
	.dc.b	$CB,$8F,$5C,$FF,$DE,$6D,$9D,$D7,$28,$C3,$CC,$8F,$C7,$E5,$60,$F8
	.dc.b	$4E,$33,$D7,$E0,$FF,$1A,$7D,$AF,$84,$B8,$7C,$58,$B3,$CE,$DF,$07
	.dc.b	$8D,$BF,$33,$56,$FD,$FF,$01,$70,$FD,$07,$C2,$7E,$68,$F7,$7C,$2D
	.dc.b	$AF,$A5,$16,$7C,$91,$82,$3F,$0E,$3F,$1F,$EF,$EF,$1F,$D4,$F5,$7E
	.dc.b	$98,$C5,$FE,$99,$B7,$FB,$1F,$82,$4F,$FA,$6F,$F7,$07,$17,$EB,$32
	.dc.b	$FD,$7D,$6A,$D5,$A7,$D7,$7F,$89,$F8,$6E,$B1,$FA,$D7,$F6,$E4,$8F
	.dc.b	$CF,$7B,$F6,$7D,$51,$F9,$1B,$ED,$7D,$67,$F6,$D8,$78,$AB,$CD,$8F
	.dc.b	$F0,$0C,$79,$67,$FC,$22,$AD,$7F,$04,$F8,$6F,$E0,$E7,$ED,$3D,$DB
	.dc.b	$1D,$CE,$78,$FC,$36,$BE,$07,$4F,$C0,$C7,$A4,$7E,$9E,$BF,$7F,$ED
	.dc.b	$F4,$AF,$CB,$7C,$37,$F1,$63,$F3,$5E,$D5,$BE,$7A,$FB,$7A,$9B,$FC
	.dc.b	$0C,$FE,$85,$69,$FE,$3B,$1F,$07,$1F,$D1,$67,$76,$FE,$4F,$36,$FF
	.dc.b	$91,$E7,$6F,$D8,$7A,$4B,$97,$87,$6E,$8F,$8E,$F8,$4F,$F2,$5C,$6E
	.dc.b	$6F,$86,$FE,$62,$7E,$D7,$D8,$B9,$FF,$09,$8F,$55,$FF,$3D,$6B,$FC
	.dc.b	$BF,$0B,$7C,$EB,$E7,$B6,$7F,$7D,$5F,$CA,$B6,$EB,$1F,$C0,$7E,$1B
	.dc.b	$FA,$09,$FC,$BB,$1B,$87,$F1,$98,$EA,$78,$BB,$7F,$7F,$EB,$62,$7B
	.dc.b	$CF,$0B,$FD,$D7,$A7,$78,$7F,$A8,$FC,$37,$F4,$E3,$F9,$DF,$A3,$70
	.dc.b	$FE,$4D,$1E,$A3,$A5,$FC,$D3,$FA,$63,$98,$D9,$FE,$75,$F3,$BF,$AB
	.dc.b	$4E,$A3,$F9,$87,$C3,$7F,$5C,$3F,$D8,$B3,$FC,$C9,$C4,$7C,$67,$A4
	.dc.b	$F7,$E0,$47,$CC,$37,$1F,$C7,$EB,$F9,$D5,$76,$3F,$9F,$7C,$37,$F6
	.dc.b	$A3,$BF,$3D,$8C,$9F,$52,$3E,$54,$BC,$0F,$B3,$12,$77,$7D,$F9,$F9
	.dc.b	$3F,$CD,$FE,$1B,$FB,$B9,$FE,$A8,$F9,$5F,$83,$A2,$33,$7F,$82,$7A
	.dc.b	$57,$B5,$FC,$39,$A7,$D9,$DF,$78,$3F,$C0,$FF,$0F,$FE,$04,$31,$7F
	.dc.b	$E0,$47,$87,$5F,$F0,$BF,$D8,$1E,$73,$E3,$FC,$17,$FC,$02,$73,$18
	.dc.b	$CB,$FA,$87,$C4,$F4,$7F,$C3,$0F,$F8,$48,$95,$F7,$7C,$3B,$3F,$06
	.dc.b	$3A,$61,$DE,$3D,$91,$8B,$FD,$0D,$F0,$CF,$FC,$43,$AD,$AF,$F6,$96
	.dc.b	$3F,$E1,$13,$F0,$BE,$05,$8F,$F0,$DE,$7E,$8C,$BF,$24,$DD,$3F,$87
	.dc.b	$61,$FB,$28,$C7,$2F,$FB,$4B,$95,$E0,$FE,$40,$FC,$8F,$83,$72,$FF
	.dc.b	$13,$FA,$AD,$3F,$C4,$47,$E5,$BC,$2B,$3E,$2E,$2F,$76,$BF,$3D,$B9
	.dc.b	$FF,$8D,$B5,$F1,$46,$36,$FF,$6D,$F7,$7E,$83,$B3,$F4,$2D,$C7,$C3
	.dc.b	$B7,$F8,$C4,$FD,$2F,$65,$8F,$07,$2F,$CB,$C3,$E6,$9F,$0B,$1E,$DE
	.dc.b	$07,$FB,$BB,$B3,$F5,$A7,$EB,$B4,$B8,$7B,$B8,$BF,$C7,$C7,$EC,$BA
	.dc.b	$2C,$76,$C7,$A6,$5F,$85,$E0,$F7,$3F,$D2,$B2,$FF,$7D,$F4,$7E,$E8
	.dc.b	$FD,$DF,$C4,$DC,$3E,$5F,$1F,$F9,$2C,$FD,$FF,$C4,$D8,$ED,$8F,$4C
	.dc.b	$BD,$6D,$3F,$B8,$62,$FD,$CF,$FD,$5F,$28,$C3,$FD,$DF,$AF,$FD,$FF
	.dc.b	$3F,$9F,$B4,$FF,$88,$E3,$8B,$FE,$63,$CB,$FC,$DA,$71,$9D,$DF,$7F
	.dc.b	$90,$CC,$79,$B9,$F9,$DB,$AF,$F1,$AF,$F9,$F7,$8B,$EF,$F0,$7F,$3E
	.dc.b	$CB,$8C,$F8,$7E,$2A,$EB,$FE,$80,$DF,$FC,$FA,$78,$67,$D9,$B7,$FA
	.dc.b	$E7,$4F,$F2,$B4,$E6,$7C,$1C,$BF,$8B,$6E,$5F,$57,$5F,$2E,$BE,$49
	.dc.b	$FF,$40,$5F,$6C,$3F,$CC,$1A,$59,$FF,$2D,$69,$F1,$0F,$8B,$E0,$3B
	.dc.b	$BC,$11,$F2,$31,$E8,$75,$F0,$3F,$8E,$63,$F9,$3E,$3F,$F0,$D5,$73
	.dc.b	$C7,$B7,$07,$F8,$AD,$47,$7D,$BA,$9F,$07,$8F,$EF,$7C,$4E,$31,$FA
	.dc.b	$17,$F6,$B3,$F0,$71,$FC,$E8,$FE,$A5,$8F,$FB,$B4,$FC,$03,$F7,$FE
	.dc.b	$27,$93,$EE,$E5,$F5,$9F,$CD,$7B,$2E,$3D,$A7,$E5,$9F,$92,$E4,$F9
	.dc.b	$7C,$BF,$3F,$16,$1C,$2B,$E4,$ED,$77,$F7,$57,$BB,$1F,$F5,$8B,$6B
	.dc.b	$C5,$18,$8F,$B9,$AF,$CC,$C7,$EE,$70,$FC,$CB,$FD,$89,$EE,$CB,$E9
	.dc.b	$EF,$FE,$0E,$C3,$C7,$8F,$68,$C7,$C5,$7D,$83,$FA,$97,$2D,$9F,$76
	.dc.b	$7F,$29,$FB,$1B,$7E,$73,$06,$1D,$4E,$D8,$F8,$2B,$FC,$8F,$3A,$7C
	.dc.b	$07,$BB,$1D,$65,$FE,$A5,$C7,$D9,$B7,$E5,$63,$ED,$6B,$F2,$F2,$ED
	.dc.b	$87,$4C,$B8,$F1,$FE,$2B,$87,$F2,$27,$F6,$BE,$6B,$5E,$AC,$3E,$3F
	.dc.b	$1F,$E0,$EF,$84,$F9,$AF,$AE,$5F,$86,$7E,$47,$C5,$7E,$70,$FC,$87
	.dc.b	$41,$C5,$75,$D7,$D1,$7F,$84,$C3,$F3,$B5,$CA,$BF,$91,$5B,$E2,$B1
	.dc.b	$7F,$94,$7F,$52,$76,$F8,$AB,$5D,$67,$F1,$C7,$C4,$C7,$E6,$D7,$DD
	.dc.b	$39,$E3,$FA,$C7,$F0,$17,$FB,$43,$D3,$F1,$57,$0E,$95,$FD,$89,$FB
	.dc.b	$B8,$7F,$22,$E3,$5F,$29,$EE,$7D,$C7,$FC,$B3,$E2,$BC,$3F,$E2,$0D
	.dc.b	$9F,$E5,$7D,$3F,$CB,$BE,$2B,$35,$F9,$AE,$2F,$E2,$9C,$DF,$8D,$CF
	.dc.b	$F3,$3A,$7F,$3A,$E0,$7C,$57,$F9,$5D,$B9,$FC,$AF,$8A,$C4,$FE,$77
	.dc.b	$83,$F8,$89,$F5,$BE,$2A,$D7,$BB,$AF,$FB,$73,$EE,$CB,$F8,$6D,$71
	.dc.b	$8F,$AB,$F1,$56,$7F,$83,$EF,$FC,$70,$FE,$59,$F2,$FF,$B5,$E1,$FD
	.dc.b	$CC,$7A,$F9,$BA,$77,$FE,$CF,$83,$FA,$A6,$9F,$A7,$AF,$D9,$1D,$63
	.dc.b	$F9,$0B,$FB,$5D,$BF,$94,$9D,$F4,$FE,$06,$F5,$CF,$E2,$DF,$A9,$A7
	.dc.b	$D7,$AF,$CE,$47,$7A,$F7,$72,$3C,$AF,$AA,$7E,$6B,$E2,$EC,$FC,$5C
	.dc.b	$3F,$AA,$3F,$17,$93,$F6,$71,$F0,$A3,$9F,$97,$F7,$8F,$D1,$FC,$5D
	.dc.b	$8E,$52,$DB,$E1,$67,$FC,$93,$7E,$9C,$FF,$C9,$76,$F3,$DF,$F7,$0D
	.dc.b	$9E,$67,$E2,$8F,$D7,$FC,$5D,$CF,$AE,$7D,$2F,$F4,$47,$C2,$3F,$61
	.dc.b	$9F,$C1,$71,$87,$F2,$5E,$3F,$E9,$8D,$BF,$EF,$3D,$2F,$8B,$FC,$CF
	.dc.b	$83,$F3,$F9,$BE,$CE,$3F,$10,$ED,$C7,$F1,$23,$EC,$E1,$F3,$BC,$DF
	.dc.b	$D0,$9F,$C1,$BE,$2E,$C7,$A3,$0F,$7E,$1E,$63,$D3,$7C,$9F,$3F,$9D
	.dc.b	$FA,$A3,$F8,$7F,$C5,$DB,$E4,$7C,$AE,$5F,$CD,$78,$4E,$5E,$BC,$0F
	.dc.b	$2B,$F6,$67,$F6,$7F,$DC,$D8,$60,$E7,$9F,$B3,$A3,$F6,$67,$D5,$8F
	.dc.b	$C4,$F9,$1F,$BC,$39,$7C,$65,$8E,$3B,$FC,$6B,$F6,$13,$4F,$24,$FE
	.dc.b	$BF,$F3,$0F,$F1,$59,$FE,$34,$F3,$D9,$BE,$37,$FC,$31,$3F,$E0,$9D
	.dc.b	$3C,$7C,$1F,$CB,$39,$BE,$87,$57,$E2,$F4,$FC,$A1,$F8,$1B,$FE,$F7
	.dc.b	$D5,$FE,$09,$6E,$9F,$E3,$6C,$FF,$D3,$AF,$F1,$43,$FC,$6B,$1A,$AF
	.dc.b	$A7,$7D,$76,$FE,$49,$1F,$17,$2F,$9F,$C7,$F6,$AF,$FB,$5C,$B2,$F0
	.dc.b	$79,$BF,$3F,$E3,$7D,$6F,$E2,$F2,$E8,$7E,$E7,$2F,$91,$AF,$B3,$0F
	.dc.b	$DD,$69,$F1,$F8,$7E,$19,$F6,$7E,$3E,$C7,$D0,$C3,$9D,$7A,$08,$7F
	.dc.b	$8A,$35,$F3,$7A,$3F,$30,$FC,$17,$C7,$5B,$E5,$5C,$32,$F1,$3C,$63
	.dc.b	$EC,$7A,$5F,$50,$FC,$57,$C3,$58,$E3,$3F,$80,$AE,$3A,$7E,$02,$3E
	.dc.b	$A7,$A7,$F7,$0F,$C9,$7C,$5D,$BE,$37,$E0,$3F,$CA,$57,$C7,$7F,$C6
	.dc.b	$71,$FF,$6F,$AF,$B1,$8F,$F9,$DA,$30,$FD,$18,$FB,$DF,$17,$F8,$AF
	.dc.b	$57,$E8,$4F,$C4,$F1,$BF,$84,$3E,$7E,$9F,$C6,$BF,$8A,$C7,$BF,$AB
	.dc.b	$F9,$23,$F5,$1F,$15,$72,$F3,$DF,$0A,$FD,$06,$9E,$7E,$5F,$A7,$8B
	.dc.b	$5F,$ED,$38,$7A,$23,$1E,$B7,$EC,$FE,$2B,$F3,$EF,$9A,$3F,$53,$AF
	.dc.b	$9B,$83,$E2,$69,$FB,$39,$FC,$F7,$A9,$FA,$63,$F7,$9F,$15,$72,$F2
	.dc.b	$F7,$7F,$05,$AF,$66,$BE,$4D,$7E,$A7,$77,$E9,$7A,$FF,$82,$7C,$57
	.dc.b	$EC,$1F,$16,$BF,$DC,$F8,$BF,$2B,$B7,$BF,$1F,$E5,$91,$F0,$3D,$8F
	.dc.b	$F3,$F9,$FD,$93,$D3,$BB,$78,$B5,$FD,$41,$E2,$3F,$2B,$C8,$F6,$3F
	.dc.b	$D0,$E7,$08,$FF,$46,$BE,$2C,$BF,$C6,$F5,$EB,$C7,$E7,$E9,$FC,$AB
	.dc.b	$07,$B1,$FE,$90,$39,$D7,$F9,$A9,$E5,$C3,$FC,$27,$0F,$56,$5E,$2C
	.dc.b	$39,$67,$FE,$98,$3F,$D5,$1F,$87,$3D,$6C,$65,$EB,$D3,$E1,$E5,$FE
	.dc.b	$5C,$C3,$C5,$EB,$7F,$A8,$0D,$E5,$FE,$A6,$7C,$2F,$F8,$E0,$FC,$84
	.dc.b	$FF,$0A,$BF,$E7,$3E,$AF,$F9,$D4,$FF,$3B,$57,$FA,$A9,$FE,$8F,$C5
	.dc.b	$F4,$F8,$BF,$1A,$FE,$BB,$1F,$85,$EA,$7F,$9F,$4E,$58,$7F,$9C,$1E
	.dc.b	$3B,$FC,$AC,$FE,$9E,$9F,$2A,$BF,$A4,$FA,$7F,$E8,$53,$FA,$2C,$A5
	.dc.b	$97,$FB,$6B,$D2,$76,$EF,$0F,$A9,$3F,$C6,$CF,$2E,$3F,$3B,$0F,$E1
	.dc.b	$47,$F7,$5E,$47,$A5,$FE,$93,$39,$47,$FA,$D1,$F3,$F0,$FD,$73,$AC
	.dc.b	$7F,$24,$3F,$B8,$E1,$E6,$CF,$CB,$1F,$EB,$F8,$E8,$7F,$BE,$A3,$31
	.dc.b	$FE,$F1,$8E,$AF,$D3,$DF,$F5,$B5,$F6,$E7,$FC,$41,$E7,$F5,$CF,$FD
	.dc.b	$CA,$DB,$0D,$7F,$B5,$D7,$7A,$DB,$E8,$9F,$F1,$17,$A7,$E0,$CB,$E8
	.dc.b	$47,$F8,$2B,$CF,$E8,$B3,$FE,$C5,$7C,$E3,$FD,$B5,$96,$A7,$31,$F6
	.dc.b	$B1,$FA,$0F,$F3,$9D,$BE,$86,$DF,$C4,$E3,$1A,$FA,$18,$7F,$B6,$0B
	.dc.b	$6A,$4E,$A7,$FC,$5A,$6A,$32,$E6,$78,$4E,$E3,$A1,$FD,$D4,$E0,$E6
	.dc.b	$37,$1C,$67,$9B,$A3,$A7,$9C,$F8,$1E,$0D,$8C,$C6,$08,$E0,$87,$B2
	.dc.b	$38,$36,$36,$D3,$67,$C5,$A7,$9D,$F1,$F6,$FC,$4C,$62,$7E,$A1,$F5
	.dc.b	$E7,$59,$F1,$F3,$6C,$FE,$1A,$98,$FA,$70,$E6,$DB,$F0,$36,$BF,$1B
	.dc.b	$FB,$17,$F9,$DD,$EF,$7C,$B2,$B5,$B9,$FC,$43,$65,$F1,$1E,$29,$FC
	.dc.b	$66,$56,$4D,$35,$C6,$F2,$35,$87,$32,$63,$F0,$32,$F2,$63,$FC,$17
	.dc.b	$3F,$4C,$F9,$09,$19,$3E,$16,$7E,$1F,$3A,$47,$B1,$F2,$79,$BE,$0A
	.dc.b	$E3,$B7,$C6,$7C,$1E,$D6,$6F,$A1,$F0,$23,$78,$FD,$9C,$79,$CF,$C8
	.dc.b	$C2,$BC,$37,$79,$DC,$F2,$ED,$39,$D7,$E6,$6B,$EE,$9C,$93,$2F,$7D
	.dc.b	$73,$CB,$E6,$61,$23,$EB,$BF,$47,$F7,$47,$C2,$F7,$BE,$38,$73,$BF
	.dc.b	$BF,$B6,$5F,$C4,$67,$F4,$1E,$C7,$E0,$B8,$3E,$D1,$1F,$C2,$F6,$71
	.dc.b	$7E,$21,$EB,$E0,$F9,$2F,$C6,$3A,$B5,$FE,$21,$AF,$E1,$66,$47,$43
	.dc.b	$F2,$3E,$5F,$F0,$9F,$2B,$FA,$2C,$76,$CF,$D4,$DD,$3F,$87,$70,$7C
	.dc.b	$97,$DE,$48,$F0,$9E,$C3,$EA,$5F,$5E,$BF,$D9,$36,$3F,$54,$D7,$F9
	.dc.b	$37,$95,$78,$FF,$49,$1E,$F2,$63,$FD,$03,$1F,$14,$78,$23,$CF,$3C
	.dc.b	$1F,$DE,$E9,$FA,$5B,$E8,$4A,$FF,$1C,$78,$CF,$6B,$C1,$C3,$F5,$70
	.dc.b	$F8,$39,$76,$BF,$84,$77,$D7,$E0,$3F,$38,$FD,$DF,$C6,$FA,$48,$FF
	.dc.b	$91,$74,$7C,$6F,$6E,$2F,$AD,$1E,$7B,$F9,$BC,$9F,$7E,$31,$7D,$CF
	.dc.b	$D7,$9F,$C2,$2F,$E3,$11,$ED,$F1,$BD,$5F,$8F,$7E,$6C,$BD,$57,$D2
	.dc.b	$7D,$64,$7F,$86,$BF,$C5,$2B,$E9,$11,$FF,$2F,$73,$BE,$1B,$B6,$BF
	.dc.b	$43,$6F,$D6,$E9,$E9,$7E,$56,$7F,$B9,$C3,$F9,$6F,$A0,$F8,$1F,$C0
	.dc.b	$36,$3F,$CD,$4D,$7F,$D0,$03,$DC,$FF,$65,$F0,$F8,$2F,$E8,$ED,$EB
	.dc.b	$DF,$DB,$97,$83,$0F,$90,$7D,$53,$F8,$6E,$2F,$4B,$C2,$48,$FF,$48
	.dc.b	$E5,$FE,$D8,$E1,$ED,$1C,$B1,$ED,$AF,$B7,$8B,$EC,$F3,$7F,$33,$CB
	.dc.b	$FC,$C1,$7E,$AF,$4B,$DA,$7E,$7F,$2F,$F7,$1F,$27,$49,$FA,$1A,$FC
	.dc.b	$87,$F0,$6F,$A5,$F4,$71,$7F,$87,$B0,$FF,$31,$C7,$C6,$F4,$BE,$49
	.dc.b	$D7,$DC,$7C,$6F,$5D,$FE,$F1,$E8,$AE,$BB,$7A,$1F,$A7,$EA,$7D,$22
	.dc.b	$3F,$0A,$FF,$06,$FD,$23,$CF,$3F,$C3,$78,$FE,$7B,$D2,$3F,$80,$67
	.dc.b	$E7,$F4,$FE,$E9,$23,$A7,$C6,$F9,$76,$F7,$61,$CA,$7F,$2E,$7C,$FE
	.dc.b	$2F,$99,$EA,$7E,$18,$91,$E5,$F8,$DE,$59,$F8,$F9,$7E,$A4,$F2,$3E
	.dc.b	$4F,$47,$F8,$97,$8D,$EA,$7E,$40,$93,$F8,$0F,$C3,$BF,$C1,$7E,$27
	.dc.b	$F2,$2E,$4F,$93,$BF,$9B,$2E,$BC,$5F,$C3,$B7,$FC,$07,$AB,$F9,$A2
	.dc.b	$65,$CF,$E4,$BC,$AF,$8F,$8B,$DD,$7F,$A7,$E0,$FF,$00,$E1,$F8,$9F
	.dc.b	$5B,$F4,$64,$BF,$27,$C9,$72,$BF,$E2,$79,$F8,$B8,$BF,$43,$87,$F8
	.dc.b	$0F,$E2,$BF,$9C,$F0,$7F,$3A,$6F,$93,$FE,$45,$6E,$5F,$AB,$E2,$FF
	.dc.b	$34,$D7,$BF,$87,$F2,$1C,$5F,$87,$AF,$94,$FC,$2D,$FC,$F1,$FE,$26
	.dc.b	$C1,$EB,$7F,$39,$6C,$7F,$B8,$DA,$FE,$EB,$BD,$93,$E4,$3D,$9B,$7F
	.dc.b	$15,$DF,$D9,$7F,$81,$3B,$7B,$1D,$48,$FD,$7F,$90,$E2,$FE,$FB,$97
	.dc.b	$E6,$DF,$F0,$4D,$FF,$C5,$1C,$5F,$17,$D7,$F3,$12,$FF,$01,$E8,$F2
	.dc.b	$7F,$91,$E1,$C9,$FE,$01,$7C,$9F,$80,$FB,$C8,$FE,$27,$E3,$79,$71
	.dc.b	$FC,$9E,$6F,$BD,$3F,$2B,$2F,$AD,$EC,$7C,$72,$47,$E4,$7E,$43,$8E
	.dc.b	$3F,$C0,$E3,$E8,$BF,$52,$BE,$F7,$B3,$F3,$C9,$3F,$99,$E5,$E3,$CF
	.dc.b	$FE,$04,$8E,$2F,$EF,$DF,$F1,$6C,$3D,$8F,$B0,$48,$FD,$17,$C8,$72
	.dc.b	$7F,$C6,$7B,$FC,$37,$F8,$09,$FB,$EE,$2F,$C6,$7B,$3F,$81,$24,$7E
	.dc.b	$AF,$D1,$E5,$C1,$FD,$DF,$9F,$EE,$CF,$F4,$FD,$BF,$31,$EC,$FE,$24
	.dc.b	$91,$FB,$3F,$8F,$E3,$97,$F6,$48,$FE,$1D,$87,$EC,$5F,$F3,$4F,$99
	.dc.b	$FE,$07,$23,$FD,$BF,$0E,$2F,$F8,$C8,$FC,$C9,$F1,$0F,$51,$FD,$B0
	.dc.b	$FE,$3D,$CB,$FC,$83,$BF,$F8,$14,$FF,$85,$7D,$46,$5D,$2D,$BF,$F6
	.dc.b	$CA,$FF,$39,$12,$3F,$CF,$BE,$57,$F9,$D4,$FF,$0E,$11,$FE,$21,$B7
	.dc.b	$3A,$FE,$EB,$3B,$9E,$27,$FB,$1F,$C0,$FD,$4F,$C0,$7E,$FF,$DB,$FD
	.dc.b	$81,$27,$F8,$D7,$C7,$73,$C7,$FD,$2F,$7F,$B0,$F8,$0B,$E2,$1A,$63
	.dc.b	$D9,$FF,$45,$79,$DF,$E3,$12,$63,$FC,$97,$DE,$E8,$FF,$A6,$23,$6D
	.dc.b	$3F,$D1,$5D,$BF,$E9,$5D,$FF,$95,$F9,$3F,$E4,$88,$FF,$1F,$46,$27
	.dc.b	$F8,$97,$C6,$74,$7F,$1F,$3C,$48,$FF,$A5,$7C,$DE,$87,$22,$3F,$CE
	.dc.b	$7E,$33,$97,$57,$F9,$16,$77,$3C,$4F,$95,$8F,$FB,$9A,$7F,$C9,$5E
	.dc.b	$6F,$90,$92,$3F,$B0,$7C,$7B,$E2,$FA,$F7,$7F,$98,$AB,$DC,$4A,$FF
	.dc.b	$55,$47,$C1,$3D,$E4,$7A,$76,$7F,$9B,$B6,$FE,$45,$1F,$E4,$5F,$1B
	.dc.b	$F9,$89,$FC,$07,$2F,$F2,$F5,$F3,$9E,$A7,$FA,$DF,$CA,$F9,$7E,$74
	.dc.b	$8E,$BA,$FF,$9D,$A3,$F3,$1D,$1E,$89,$FF,$68,$DF,$37,$F7,$AF,$D7
	.dc.b	$BF,$AB,$E6,$C9,$E6,$7F,$B6,$1F,$B3,$1C,$E3,$F4,$2F,$B5,$E6,$79
	.dc.b	$4E,$26,$8B,$9F,$FA,$A1,$B1,$FA,$DE,$CE,$93,$99,$FF,$08,$37,$5F
	.dc.b	$91,$5F,$8A,$F3,$25,$74,$6B,$FD,$3F,$47,$C1,$F4,$9C,$DF,$D1,$7F
	.dc.b	$C6,$92,$F8,$DF,$CA,$F9,$B3,$0F,$3B,$FE,$95,$FD,$D7,$FC,$7B,$1F
	.dc.b	$C1,$B8,$BE,$BF,$47,$F2,$FE,$3F,$F4,$A5,$F1,$AF,$4B,$FA,$0F,$36
	.dc.b	$47,$C2,$6B,$E7,$F8,$6F,$83,$18,$D7,$F0,$CF,$EA,$3C,$D9,$5F,$29
	.dc.b	$AF,$BF,$E3,$BE,$4C,$62,$BF,$1C,$FB,$BA,$7F,$BB,$7C,$3C,$BF,$EE
	.dc.b	$CF,$C2,$F8,$EF,$C3,$78,$AE,$27,$C7,$7C,$62,$5F,$F9,$9A,$1F,$1D
	.dc.b	$FE,$01,$F8,$A9,$3F,$E0,$17,$E6,$FC,$77,$4A,$7C,$6F,$D0,$25,$73
	.dc.b	$F8,$DF,$F0,$57,$C6,$C8,$FF,$05,$3F,$5B,$E3,$39,$7C,$5F,$D8,$24
	.dc.b	$F8,$FE,$2B,$FC,$25,$F1,$92,$3F,$C2,$4F,$DE,$F8,$AF,$17,$C5,$7E
	.dc.b	$03,$E2,$65,$7A,$E3,$FC,$35,$F1,$92,$3F,$C3,$4F,$E1,$BE,$2B,$D5
	.dc.b	$F1,$3F,$88,$24,$FC,$3F,$89,$FE,$3B,$F0,$32,$3F,$3E,$FE,$3B,$E2
	.dc.b	$7F,$90,$F9,$DF,$D8,$BC,$BF,$F6,$61,$FE,$AB,$AD,$90,$FD,$0C,$7F
	.dc.b	$28,$F7,$A4,$7E,$A9,$E6,$FE,$B2,$3A,$1F,$CC,$4F,$13,$FA,$AB,$DB
	.dc.b	$D0,$E0,$43,$FC,$FD,$EF,$BD,$CD,$F8,$FF,$DE,$67,$F3,$DE,$8F,$F2
	.dc.b	$57,$FB,$0F,$A1,$DD,$E5,$E2,$FF,$1D,$CB,$FD,$41,$FB,$AF,$F9,$42
	.dc.b	$76,$7F,$02,$AF,$E8,$66,$BE,$2A,$C7,$F3,$1C,$7F,$DF,$D3,$FD,$2C
	.dc.b	$FE,$B7,$4F,$8A,$C5,$73,$21,$FE,$4F,$87,$C5,$7F,$50,$D2,$4F,$F1
	.dc.b	$27,$CB,$E7,$E2,$3F,$C3,$F1,$FD,$64,$FE,$B5,$EF,$D8,$E0,$43,$B7
	.dc.b	$C5,$7F,$5C,$89,$1F,$C7,$9F,$83,$F1,$58,$8F,$F1,$5C,$74,$7D,$DC
	.dc.b	$1F,$2F,$91,$F0,$39,$9F,$E4,$35,$C7,$8B,$F6,$91,$FE,$E3,$D7,$FC
	.dc.b	$74,$FF,$2C,$AF,$ED,$A4,$7F,$DB,$70,$F8,$9C,$47,$F8,$F2,$7F,$BB
	.dc.b	$1F,$33,$E2,$6C,$73,$21,$F2,$FE,$17,$FB,$CF,$C2,$48,$FE,$76,$FE
	.dc.b	$0F,$1F,$2C,$3E,$49,$F0,$1B,$8F,$F2,$5D,$7F,$81,$4F,$F0,$37,$C2
	.dc.b	$58,$E6,$43,$ED,$7A,$7F,$E0,$88,$91,$FD,$29,$FF,$05,$FC,$26,$23
	.dc.b	$FC,$AF,$1D,$1F,$9A,$7F,$7B,$FD,$95,$ED,$6F,$3F,$EE,$55,$7D,$0F
	.dc.b	$CE,$57,$E7,$6F,$FA,$8E,$1D,$9F,$EB,$2F,$F9,$0F,$8B,$C1,$D8,$E5
	.dc.b	$FB,$58,$F8,$6D,$39,$7C,$59,$F7,$7F,$C1,$DC,$AA,$C4,$32,$F1,$69
	.dc.b	$F3,$1B,$A7,$9A,$7E,$83,$5F,$BF,$E9,$79,$33,$F1,$C7,$C2,$3F,$CA
	.dc.b	$F8,$7E,$E7,$3F,$24,$77,$CF,$9B,$F1,$CA,$78,$CB,$8F,$89,$FC,$97
	.dc.b	$9F,$8F,$80,$91,$FE,$00,$9F,$EE,$2F,$E4,$BC,$DE,$37,$E5,$8F,$83
	.dc.b	$8F,$8F,$2F,$A3,$3C,$A5,$AF,$C1,$C7,$0F,$C3,$9F,$1F,$BF,$2B,$C0
	.dc.b	$F6,$7C,$51,$DD,$F8,$27,$C3,$CB,$F0,$6E,$8F,$45,$F6,$DB,$C7,$19
	.dc.b	$3C,$6E,$1A,$7D,$C3,$FC,$3F,$E2,$7E,$E9,$E1,$8F,$9D,$B3,$F8,$EF
	.dc.b	$6F,$CD,$1C,$8F,$A4,$68,$FD,$F7,$99,$F4,$22,$47,$CD,$7F,$96,$FC
	.dc.b	$7E,$23,$96,$49,$C6,$1F,$C2,$3D,$97,$7B,$E3,$33,$9F,$8F,$1E,$87
	.dc.b	$F0,$8F,$8B,$FB,$FE,$EC,$8F,$B8,$FF,$88,$FD,$3C,$47,$82,$2E,$5F
	.dc.b	$97,$F7,$6E,$1F,$AC,$9B,$5F,$C6,$7E,$2F,$F1,$75,$23,$F1,$0F,$F4
	.dc.b	$8F,$8A,$C4,$7B,$22,$E1,$FA,$3F,$7E,$E3,$FB,$89,$B4,$DB,$F6,$3F
	.dc.b	$07,$F9,$8B,$93,$F9,$57,$F3,$F3,$FB,$6F,$8A,$CC,$FC,$A9,$B8,$FC
	.dc.b	$9E,$DF,$E3,$73,$C4,$86,$CF,$F3,$3D,$FF,$00,$EA,$FD,$4F,$C1,$B9
	.dc.b	$A9,$2B,$77,$FE,$82,$7F,$17,$7F,$C2,$3E,$23,$CB,$3F,$E6,$1D,$BF
	.dc.b	$D2,$93,$FC,$33,$E2,$33,$5A,$F8,$8D,$FC,$3C,$C7,$AC,$7E,$E3,$E1
	.dc.b	$A4,$7E,$D5,$EB,$A6,$23,$F1,$31,$76,$F8,$1F,$0F,$74,$FE,$4B,$36
	.dc.b	$BE,$76,$BF,$E1,$97,$F2,$3F,$0F,$FC,$18,$8F,$BB,$DC,$AE,$8F,$E0
	.dc.b	$D1,$EE,$F6,$F3,$97,$F1,$F8,$B8,$FC,$0C,$1E,$D7,$F1,$2F,$87,$98
	.dc.b	$FF,$0F,$7E,$3F,$B1,$88,$FD,$1C,$5C,$7E,$D7,$9B,$71,$FE,$7F,$36
	.dc.b	$BD,$7E,$AF,$F1,$EF,$87,$91,$FC,$6D,$FA,$7E,$96,$23,$F5,$F1,$71
	.dc.b	$FC,$37,$99,$71,$FE,$97,$36,$BE,$3F,$A3,$FC,$AB,$E1,$E4,$7F,$27
	.dc.b	$7F,$27,$FB,$AF,$F2,$D8,$FE,$85,$1A,$7F,$7A,$EB,$CD,$C6,$5F,$D1
	.dc.b	$62,$E3,$F8,$6B,$FF,$4E,$5F,$F3,$3F,$88,$FE,$72,$47,$F2,$F1,$FC
	.dc.b	$89,$FE,$13,$19,$69,$AF,$2F,$18,$7F,$53,$9B,$7D,$FC,$7F,$E8,$BF
	.dc.b	$11,$30,$FE,$84,$FB,$3C,$6C,$47,$F1,$68,$B8,$7F,$15,$CB,$FD,$1F
	.dc.b	$3C,$08,$6C,$F3,$F5,$7F,$52,$F8,$89,$1F,$D3,$DF,$93,$D3,$88,$FE
	.dc.b	$47,$16,$FB,$72,$DD,$3F,$BF,$CD,$AF,$E1,$18,$B8,$BF,$AF,$FC,$44
	.dc.b	$9F,$EB,$8F,$D6,$E1,$C4,$7F,$30,$8B,$97,$AF,$82,$E5,$FE,$0C,$9B
	.dc.b	$5D,$B6,$FE,$D7,$E5,$C8,$FE,$D0,$FE,$0B,$4C,$47,$F3,$C8,$B8,$FC
	.dc.b	$7B,$B8,$7F,$86,$26,$D7,$A9,$FE,$EF,$F1,$12,$3F,$BA,$3F,$8E,$30
	.dc.b	$FF,$48,$6D,$FD,$3F,$88,$B7,$FE,$24,$9B,$5F,$1B,$E2,$3F,$C0,$75
	.dc.b	$23,$FC,$00,$FE,$6B,$E2,$31,$1F,$D5,$22,$E1,$FD,$4F,$9F,$FD,$09
	.dc.b	$E9,$A7,$39,$7F,$8C,$9F,$E6,$B2,$F8,$7F,$F0,$8E,$53,$1F,$F0,$73
	.dc.b	$AF,$87,$C4,$7F,$62,$8B,$97,$7F,$87,$B9,$7F,$8F,$A6,$D7,$F4,$7C
	.dc.b	$5F,$0F,$FC,$E3,$D6,$92,$3F,$BB,$79,$3F,$DA,$5F,$E7,$F4,$F8,$7F
	.dc.b	$3F,$C3,$C9,$CB,$EA,$F8,$EF,$5F,$C4,$49,$E2,$FB,$BE,$23,$FC,$5B
	.dc.b	$C9,$27,$FC,$54,$FC,$2F,$88,$C4,$7F,$7E,$8B,$C3,$F2,$FE,$1A,$F0
	.dc.b	$FF,$96,$A6,$D7,$B7,$E1,$7F,$C7,$7E,$C7,$FA,$CC,$95,$F3,$BE,$17
	.dc.b	$2C,$AF,$81,$8F,$F9,$F2,$BF,$CF,$AF,$D9,$F8,$5D,$4F,$F8,$66,$79
	.dc.b	$BF,$13,$E0,$BB,$F4,$C8,$E7,$87,$FB,$26,$BE,$E4,$7C,$AF,$6F,$4F
	.dc.b	$3A,$B8,$FE,$1F,$9A,$E1,$C6,$6D,$7D,$9F,$81,$F6,$F5,$C8,$EE,$7D
	.dc.b	$9C,$BF,$D9,$0F,$D8,$F8,$1D,$3E,$99,$B9,$7E,$57,$E0,$6E,$5E,$79
	.dc.b	$B5,$FE,$0D,$FD,$D5,$F0,$5F,$33,$BE,$4F,$80,$FF,$42,$3F,$E0,$E9
	.dc.b	$7C,$16,$A7,$E4,$4D,$CB,$FC,$6D,$5F,$E0,$E9,$C9,$CF,$4F,$F6,$DB
	.dc.b	$6B,$A4,$7D,$CE,$F9,$3B,$7C,$1F,$19,$88,$FA,$D1,$70,$EF,$F0,$17
	.dc.b	$5F,$A9,$36,$BD,$DF,$03,$F8,$8F,$02,$47,$8D,$F6,$7C,$06,$23,$F0
	.dc.b	$51,$71,$F5,$FC,$0D,$C7,$F0,$13,$6B,$E5,$78,$9F,$94,$F0,$E4,$7B
	.dc.b	$DF,$91,$F0,$38,$8F,$C7,$45,$C7,$E3,$FC,$5D,$C7,$F1,$93,$6B,$EB
	.dc.b	$7C,$17,$E8,$3C,$69,$1F,$2D,$FA,$9F,$05,$88,$FC,$D4,$5C,$7E,$9F
	.dc.b	$C0,$DC,$7F,$31,$36,$BF,$05,$F0,$3F,$AC,$F2,$64,$7D,$77,$F0,$1F
	.dc.b	$03,$88,$FD,$34,$5C,$7E,$FF,$C0,$DC,$7F,$49,$36,$BF,$1D,$F0,$3F
	.dc.b	$B8,$F3,$24,$7E,$0D,$FC,$67,$C0,$E2,$3F,$65,$17,$1F,$C5,$FC,$0D
	.dc.b	$C7,$F6,$13,$6B,$FC,$CD,$F8,$0F,$F8,$3F,$B5,$E7,$FF,$06,$24,$FF
	.dc.b	$97,$27,$FD,$76,$FE,$FE,$32,$D3,$5F,$13,$8C,$BF,$84,$4D,$C7,$FC
	.dc.b	$B3,$7F,$E0,$F9,$FE,$19,$E8,$7F,$14,$26,$1D,$FE,$0E,$BF,$C3,$62
	.dc.b	$CF,$FA,$6E,$9F,$07,$8C,$3F,$86,$55,$BF,$6F,$C1,$FF,$20,$F4,$A4
	.dc.b	$7E,$AD,$FF,$40,$E1,$FE,$10,$8F,$E3,$51,$A6,$9F,$07,$E0,$F1,$97
	.dc.b	$F2,$39,$B8,$FB,$3E,$0F,$F9,$6F,$A9,$30,$FD,$C3,$DB,$E0,$F1,$1F
	.dc.b	$C9,$E2,$E1,$EB,$F8,$2B,$87,$F2,$59,$B5,$F5,$BE,$0B,$F9,$CF,$AF
	.dc.b	$23,$F8,$23,$EA,$F8,$2C,$47,$F3,$38,$B8,$FC,$7F,$82,$B8,$FF,$31
	.dc.b	$9B,$5F,$82,$F8,$2F,$E8,$BE,$D4,$8F,$E2,$0F,$C6,$F8,$2C,$47,$F3
	.dc.b	$F8,$B8,$FD,$3F,$82,$B8,$FF,$3D,$9B,$5F,$8E,$F8,$2F,$EA,$3E,$EC
	.dc.b	$8F,$E3,$8F,$D2,$F8,$1C,$47,$F4,$B8,$B8,$FD,$FF,$81,$B8,$FF,$49
	.dc.b	$9B,$5F,$9A,$F8,$0F,$EB,$BF,$01,$23,$F9,$43,$FE,$BE,$FD,$D7,$FB
	.dc.b	$03,$98,$FE,$B0,$E7,$E0,$71,$97,$F5,$F8,$B8,$FF,$AC,$2F,$FA,$4D
	.dc.b	$FF,$5F,$F8,$1F,$ED,$64,$7F,$D8,$31,$FC,$D9,$D1,$FC,$E6,$34,$FF
	.dc.b	$AF,$A3,$FC,$1E,$E4,$87,$03,$5E,$A7,$F8,$45,$60,$BC,$72,$B8,$7F
	.dc.b	$BE,$FC,$0C,$BF,$EA,$53,$97,$FA,$54,$65,$FE,$F0,$F5,$BE,$91,$FE
	.dc.b	$1F,$97,$0E,$E7,$FC,$05,$57,$0F,$0F,$0F,$FB,$AB,$CD,$79,$BF,$3F
	.dc.b	$CE,$99,$FF,$B9,$27,$2F,$F5,$BF,$81,$FF,$09,$91,$F9,$3E,$75,$7F
	.dc.b	$C1,$91,$73,$F2,$FC,$33,$85,$E6,$78,$7D,$0F,$17,$A1,$31,$F0,$47
	.dc.b	$B7,$7F,$95,$FB,$47,$F7,$89,$FE,$EC,$D7,$E6,$C6,$BE,$0B,$F1,$1E
	.dc.b	$3F,$19,$FF,$21,$56,$F4,$F8,$F0,$8C,$5F,$18,$FC,$1F,$89,$F1,$7D
	.dc.b	$2F,$8D,$E9,$49,$F8,$91,$D5,$E5,$F0,$57,$0F,$F0,$3C,$79,$A3,$F3
	.dc.b	$5E,$06,$59,$19,$3F,$23,$DD,$E7,$E2,$7D,$5F,$E2,$67,$DA,$23,$3F
	.dc.b	$0B,$B9,$FE,$7B,$B7,$CE,$7F,$DA,$59,$7F,$84,$E3,$DE,$DD,$BE,$5F
	.dc.b	$C7,$7E,$B7,$9B,$D0,$CC,$BE,$6B,$E3,$D3,$E2,$C7,$87,$1F,$F6,$AB
	.dc.b	$E4,$CF,$CF,$F1,$BE,$33,$C9,$EB,$7C,$3F,$5E,$57,$D8,$7A,$18,$7D
	.dc.b	$A7,$D3,$71,$AF,$E2,$31,$7C,$27,$BF,$D7,$F8,$1E,$BC,$C3,$FC,$8F
	.dc.b	$9F,$C8,$9F,$A5,$52,$7F,$0E,$FA,$FE,$07,$D9,$F0,$32,$78,$E2,$8F
	.dc.b	$99,$EC,$7C,$DF,$8B,$F9,$DE,$C4,$BE,$FC,$3F,$D4,$76,$F9,$BF,$0B
	.dc.b	$F8,$12,$4F,$4F,$81,$FA,$DE,$D4,$8E,$2F,$D9,$F6,$BE,$D7,$B5,$F6
	.dc.b	$FD,$A9,$5D,$A3,$F1,$9F,$09,$23,$F4,$2F,$D5,$F3,$3F,$05,$E8,$49
	.dc.b	$F2,$BF,$85,$F6,$FF,$0D,$ED,$7E,$1F,$DA,$95,$EA,$8F,$CC,$73,$48
	.dc.b	$FD,$6B,$F9,$3F,$DD,$7F,$25,$C3,$F8,$DF,$57,$F2,$04,$AF,$87,$EA
	.dc.b	$FE,$4F,$D4,$FC,$A7,$B5,$27,$E3,$C7,$E9,$7A,$A4,$7E,$E9,$ED,$DF
	.dc.b	$F9,$BF,$6A,$4F,$CE,$7F,$49,$F8,$87,$FA,$6B,$C7,$FF,$5E,$78,$48
	.dc.b	$FF,$49,$D7,$E4,$DF,$E7,$FA,$FF,$10,$6D,$70,$77,$1F,$A4,$F8,$6F
	.dc.b	$DC,$12,$BF,$17,$BB,$A7,$C3,$F1,$12,$7F,$1F,$87,$F4,$7C,$3B,$7C
	.dc.b	$47,$9B,$DC,$F3,$FC,$3C,$BD,$C7,$F0,$5F,$88,$91,$FC,$69,$F9,$DC
	.dc.b	$FE,$EF,$76,$4F,$67,$E0,$FB,$9F,$0B,$DB,$93,$E2,$7E,$2F,$C2,$FD
	.dc.b	$7E,$89,$3E,$A7,$ED,$72,$FC,$AF,$85,$93,$EA,$7E,$6F,$B5,$F3,$BD
	.dc.b	$AF,$9F,$F0,$92,$BD,$F1,$FC,$77,$C8,$91,$FC,$C9,$FE,$31,$FB,$AF
	.dc.b	$F6,$48,$FA,$5F,$0B,$F6,$08,$FC,$8F,$6F,$ED,$FB,$7F,$73,$E1,$24
	.dc.b	$FC,$C8,$FE,$55,$F1,$32,$3F,$9F,$BD,$B8,$7F,$07,$F0,$B2,$7E,$9B
	.dc.b	$F8,$6F,$73,$F9,$57,$E2,$3F,$C6,$3C,$33,$FC,$D2,$BF,$8C,$37,$B7
	.dc.b	$FA,$AB,$FC,$7A,$3F,$A2,$7C,$57,$F3,$B2,$3F,$C3,$B7,$71,$76,$F7
	.dc.b	$A4,$FF,$12,$BF,$E5,$CF,$F3,$75,$EE,$79,$3D,$BF,$EA,$9E,$DB,$E1
	.dc.b	$7C,$5E,$4B,$0F,$F6,$7C,$61,$E5,$5F,$D3,$7E,$2A,$47,$F6,$27,$E8
	.dc.b	$F0,$FC,$0F,$56,$4F,$8D,$F8,$7E,$8F,$83,$0F,$2E,$5E,$1C,$F0,$67
	.dc.b	$5F,$EB,$5F,$11,$31,$FE,$DE,$FF,$55,$FD,$D7,$D7,$1E,$FF,$1B,$E7
	.dc.b	$91,$F8,$5D,$BD,$F7,$F0,$7B,$72,$7F,$DF,$B9,$FC,$19,$FE,$D1,$EC
	.dc.b	$49,$FE,$FA,$F7,$F4,$BE,$DE,$72,$7E,$73,$F7,$B8,$BD,$9C,$3E,$DE
	.dc.b	$19,$5D,$B8,$53,$FD,$D7,$E2,$24,$FF,$82,$9F,$67,$47,$C5,$CA,$4F
	.dc.b	$17,$F1,$7F,$0D,$F2,$7D,$3F,$95,$C5,$2B,$B4,$7C,$5C,$BF,$C0,$44
	.dc.b	$8F,$EA,$BC,$7F,$7D,$F2,$71,$FD,$32,$3E,$7E,$FF,$AD,$9C,$8F,$53
	.dc.b	$F6,$77,$FF,$03,$78,$7F,$EA,$A9,$FE,$C3,$F1,$5F,$EB,$5E,$AF,$E4
	.dc.b	$AD,$8F,$F0,$44,$ED,$FF,$05,$F3,$7E,$47,$93,$F0,$84,$8F,$91,$CB
	.dc.b	$F8,$7E,$39,$1F,$2D,$FC,$56,$FE,$2F,$4F,$C7,$EE,$4A,$FE,$E7,$97
	.dc.b	$F2,$97,$E9,$72,$7E,$44,$8F,$D5,$F8,$8F,$D1,$FC,$97,$F4,$ED,$FF
	.dc.b	$45,$B2,$31,$48,$C3,$C3,$E2,$7E,$1F,$BB,$F1,$3D,$D9,$3E,$28,$EB
	.dc.b	$F0,$F2,$3F,$23,$87,$F8,$6B,$C6,$FF,$34,$E3,$DF,$E2,$3C,$7F,$11
	.dc.b	$2F,$6F,$9B,$DE,$FF,$1B,$7E,$25,$FD,$4E,$7F,$C6,$8F,$F4,$CD,$BF
	.dc.b	$AA,$B6,$33,$1E,$EF,$89,$F7,$FC,$4C,$9F,$1C,$7C,$2F,$89,$F8,$7F
	.dc.b	$11,$27,$CF,$1E,$0F,$7F,$C3,$EE,$C9,$FF,$10,$E4,$9F,$97,$D3,$F3
	.dc.b	$3A,$65,$7C,$08,$F9,$FF,$11,$F4,$3E,$22,$4F,$C4,$8F,$6F,$B5,$F5
	.dc.b	$3E,$22,$4F,$68,$FA,$FF,$03,$23,$E5,$BF,$6B,$A7,$ED,$F5,$49,$F9
	.dc.b	$EF,$DE,$F8,$8F,$BF,$F1,$1F,$80,$F8,$19,$5F,$52,$3F,$07,$F0,$3F
	.dc.b	$C9,$BE,$5B,$E7,$76,$1B,$DC,$9D,$79,$1B,$F9,$09,$8F,$F0,$EB,$FA
	.dc.b	$58,$71,$F8,$9E,$9F,$11,$2B,$51,$DF,$E2,$24,$71,$7C,$5D,$1E,$3E
	.dc.b	$AE,$F3,$D7,$3C,$3C,$8F,$06,$1D,$5F,$D7,$E5,$FE,$07,$EF,$FC,$EA
	.dc.b	$C7,$BC,$E9,$EE,$FA,$7E,$23,$E1,$7C,$14,$8C,$1E,$DF,$88,$F8,$DF
	.dc.b	$11,$23,$07,$C1,$E9,$F9,$5F,$11,$23,$07,$C5,$F8,$2F,$9C,$47,$E4
	.dc.b	$79,$DF,$47,$E2,$24,$7C,$B7,$EA,$7C,$3F,$D5,$F8,$8F,$AD,$F0,$92
	.dc.b	$BE,$84,$7D,$9F,$88,$FE,$BF,$FB,$B7,$FA,$2B,$A4,$DE,$34,$EA,$77
	.dc.b	$F0,$B3,$0F,$E8,$F7,$FD,$8E,$78,$FC,$37,$4F,$85,$95,$A8,$EF,$F0
	.dc.b	$DE,$18,$FC,$16,$38,$79,$DF,$57,$F9,$1E,$1F,$D9,$0E,$C7,$F5,$E7
	.dc.b	$38,$76,$77,$87,$B3,$E1,$64,$79,$9F,$7F,$C3,$7C,$0F,$86,$93,$EA
	.dc.b	$7E,$1F,$C2,$FC,$4F,$86,$F8,$BF,$0B,$2B,$DF,$1C,$9E,$84,$3C,$F7
	.dc.b	$D1,$F4,$DD,$F6,$BF,$8D,$E8,$C7,$B5,$7C,$AC,$7B,$3F,$03,$1F,$AB
	.dc.b	$F0,$BD,$E3,$A9,$F4,$1E,$9E,$07,$F3,$4C,$3F,$D4,$D3,$EB,$3E,$AE
	.dc.b	$1D,$9F,$97,$87,$6C,$7C,$78,$FE,$0B,$E1,$7F,$07,$F0,$B2,$FE,$EC
	.dc.b	$7E,$1B,$E1,$64,$7E,$01,$FC,$4F,$C2,$FE,$2B,$E1,$64,$FE,$11,$FC
	.dc.b	$6F,$C2,$FE,$3B,$E1,$7B,$4F,$43,$F1,$2F,$47,$EF,$DD,$CF,$E3,$E1
	.dc.b	$D2,$7F,$07,$85,$9F,$97,$87,$E6,$BE,$16,$57,$E5,$1F,$F1,$DF,$67
	.dc.b	$F5,$6E,$83,$FE,$F4,$E2,$FE,$B6,$DA,$FE,$B4,$D9,$DF,$C2,$72,$F8
	.dc.b	$4E,$7F,$09,$2F,$FC,$4F,$7F,$DE,$EF,$AF,$C2,$78,$26,$1A,$7C,$7F
	.dc.b	$0B,$23,$8B,$E6,$C7,$B4,$74,$C3,$B3,$FE,$04,$C3,$B0,$F8,$4F,$69
	.dc.b	$23,$C7,$F0,$BF,$03,$E1,$3E,$0F,$C2,$C9,$F3,$C7,$C4,$F8,$39,$1E
	.dc.b	$A7,$E3,$FC,$27,$C8,$BE,$93,$E1,$C3,$0F,$23,$E0,$E1,$D5,$F3,$61
	.dc.b	$D1,$F4,$61,$78,$FF,$BD,$E4,$CB,$EA,$FC,$1C,$BF,$96,$FD,$8F,$84
	.dc.b	$F5,$47,$62,$1D,$8F,$41,$0E,$E7,$D4,$9F,$C0,$FC,$27,$E0,$BE,$12
	.dc.b	$4F,$D8,$8F,$C2,$FC,$24,$8F,$B6,$FE,$23,$E1,$3F,$13,$F0,$9F,$8A
	.dc.b	$F8,$39,$5F,$80,$8F,$C6,$FC,$27,$68,$E8,$7E,$11,$E8,$FD,$9B,$BB
	.dc.b	$7C,$2C,$3A,$CF,$E2,$B0,$EA,$FD,$FC,$3F,$33,$F0,$7F,$E9,$4F,$DD
	.dc.b	$FF,$B7,$9F,$DC,$38,$3F,$B7,$57,$F9,$E2,$3F,$B5,$93,$E0,$38,$7C
	.dc.b	$27,$1F,$84,$95,$B7,$A7,$C1,$75,$F8,$2E,$DF,$09,$2B,$A4,$78,$70
	.dc.b	$91,$DD,$EA,$F3,$21,$FE,$5A,$C7,$FB,$63,$FE,$45,$7F,$DA,$5B,$FF
	.dc.b	$8F,$17,$C2,$7B,$49,$3E,$BF,$84,$F8,$1F,$09,$23,$DC,$FC,$3F,$84
	.dc.b	$F8,$9F,$09,$F1,$7E,$12,$57,$C3,$8F,$91,$F0,$92,$3E,$33,$F2,$FE
	.dc.b	$13,$A4,$79,$F0,$BA,$79,$71,$EB,$3F,$37,$1E,$AF,$B7,$1F,$A9,$F0
	.dc.b	$9C,$E3,$E9,$63,$87,$CE,$43,$B1,$F5,$EF,$AB,$F1,$EF,$AE,$1F,$17
	.dc.b	$0F,$C0,$7C,$27,$E0,$7E,$12,$5F,$DF,$8F,$C2,$73,$48,$FC,$13,$F8
	.dc.b	$7F,$84,$FC,$47,$C0,$FF,$B7,$3F,$78,$FE,$FC,$7F,$7F,$E3,$FE,$FB
	.dc.b	$7F,$EC,$08,$FE,$F4,$4F,$3B,$87,$C2,$75,$9F,$B9,$8D,$D7,$ED,$65
	.dc.b	$DA,$79,$7C,$87,$73,$0F,$FA,$73,$1F,$EF,$15,$FE,$7C,$5F,$09,$E5
	.dc.b	$3C,$DF,$09,$E7,$F8,$59,$7E,$57,$D5,$F0,$BE,$BF,$85,$93,$E9,$7D
	.dc.b	$DF,$0B,$EF,$F8,$3E,$73,$ED,$C3,$0F,$32,$1E,$5C,$3A,$BC,$C8,$7A
	.dc.b	$6F,$AB,$F1,$FE,$0B,$E6,$7C,$1C,$8C,$1F,$2F,$E1,$7E,$87,$C1,$C8
	.dc.b	$C1,$F3,$FE,$13,$EA,$7C,$1C,$8C,$1F,$4F,$E0,$FE,$C1,$1F,$AD,$F0
	.dc.b	$7F,$6F,$0E,$91,$F3,$70,$C3,$C8,$FA,$17,$D5,$FA,$58,$7E,$0F,$E0
	.dc.b	$FF,$09,$F0,$92,$7F,$05,$1C,$9E,$87,$E1,$6F,$1C,$BF,$66,$FA,$3F
	.dc.b	$6F,$0F,$C7,$47,$E3,$FE,$11,$72,$BF,$D7,$7F,$BC,$FF,$73,$3F,$BA
	.dc.b	$73,$AB,$75,$FE,$E2,$9F,$EE,$44,$F8,$2C,$DF,$2F,$82,$BF,$0B,$FE
	.dc.b	$D4,$AF,$EE,$44,$F8,$5C,$E1,$FE,$D5,$87,$C2,$63,$C1,$47,$FB,$C7
	.dc.b	$1F,$2F,$C2,$E6,$31,$8F,$5F,$84,$C6,$93,$2F,$64,$F8,$63,$86,$9C
	.dc.b	$74,$95,$A3,$C4,$F5,$D3,$B7,$C2,$C9,$D6,$BE,$48,$F1,$7C,$27,$8F
	.dc.b	$E1,$65,$6B,$54,$79,$FE,$17,$D1,$F0,$72,$B4,$FA,$FE,$17,$D9,$F0
	.dc.b	$B2,$79,$3E,$FF,$84,$EB,$1F,$47,$0B,$C9,$F6,$31,$E9,$3F,$57,$1B
	.dc.b	$3F,$73,$1F,$91,$F0,$B2,$BC,$CF,$CB,$F8,$4F,$99,$F0,$92,$7D,$4F
	.dc.b	$CF,$F8,$3F,$A1,$F0,$92,$7D,$CF,$D3,$F8,$5F,$A9,$F0,$B2,$7E,$13
	.dc.b	$F6,$E7,$FD,$A7,$C2,$C3,$B5,$7E,$2D,$EC,$F4,$21,$F8,$F9,$EC,$FD
	.dc.b	$27,$FD,$AA,$4F,$80,$ED,$F0,$BD,$36,$FC,$DD,$DE,$BF,$C9,$69,$E9
	.dc.b	$C7,$A1,$DA,$BF,$43,$A7,$9F,$E1,$64,$F7,$7D,$5F,$0B,$EB,$F8,$4F
	.dc.b	$67,$C2,$CA,$F2,$47,$BF,$E1,$7E,$07,$C2,$C9,$F4,$47,$C3,$F8,$4F
	.dc.b	$89,$F0,$52,$7D,$91,$D5,$F7,$61,$D5,$68,$DB,$F7,$3D,$EE,$3F,$F7
	.dc.b	$25,$76,$3E,$36,$7E,$17,$8D,$70,$9F,$EF,$B1,$F2,$FE,$07,$EB,$7C
	.dc.b	$14,$8C,$1F,$3F,$E1,$7E,$D7,$C2,$48,$C1,$F4,$FE,$17,$EF,$7C,$24
	.dc.b	$8C,$1F,$5F,$E1,$7F,$04,$47,$ED,$7C,$17,$E1,$70,$E9,$1E,$61,$B7
	.dc.b	$B7,$1D,$47,$43,$51,$F0,$70,$F0,$46,$34,$65,$FC,$2F,$EF,$DE,$0F
	.dc.b	$E4,$FE,$0E,$5F,$EF,$27,$2D,$F1,$F8,$78,$F2,$30,$6F,$C4,$E8,$47
	.dc.b	$97,$89,$F9,$FF,$86,$FE,$35,$F8,$57,$FB,$E4,$FF,$7D,$6B,$FD,$F2
	.dc.b	$BD,$31,$FE,$F5,$1F,$0D,$C0,$E3,$C1,$CB,$E0,$65,$7F,$12,$AF,$EE
	.dc.b	$FD,$1F,$82,$7F,$BB,$C6,$7E,$07,$C3,$F0,$92,$30,$71,$C3,$B3,$CF
	.dc.b	$1E,$EF,$5C,$7B,$BE,$67,$D7,$F0,$BE,$CF,$83,$F6,$FC,$2C,$AF,$24
	.dc.b	$72,$7D,$B7,$D9,$F7,$E1,$D3,$BF,$F3,$51,$F7,$B4,$F5,$B7,$2E,$EF
	.dc.b	$BF,$E1,$7E,$5F,$C2,$C8,$C1,$FC,$D3,$EF,$FF,$36,$FE,$11,$78,$5F
	.dc.b	$D2,$DB,$1B,$F1,$B1,$E2,$C8,$E4,$EF,$C6,$FB,$3F,$0D,$F6,$BE,$12
	.dc.b	$57,$D1,$8E,$9E,$1F,$5F,$0E,$4F,$8E,$3C,$1E,$1C,$8F,$33,$FD,$77
	.dc.b	$F0,$07,$C2,$7E,$1B,$E1,$3F,$0F,$F0,$72,$FE,$F4,$7A,$7C,$5F,$57
	.dc.b	$81,$27,$E0,$47,$E3,$BE,$12,$47,$E1,$5F,$C8,$FC,$3F,$E4,$BE,$1F
	.dc.b	$F2,$79,$4A,$FC,$54,$7F,$6F,$FB,$FF,$C4,$C7,$EB,$17,$85,$93,$5E
	.dc.b	$0F,$E7,$3E,$22,$57,$E4,$DF,$CF,$FC,$3F,$F7,$9F,$C3,$3F,$DF,$84
	.dc.b	$9C,$7C,$37,$7E,$D9,$1F,$DB,$32,$47,$8F,$9F,$97,$C3,$73,$F8,$69
	.dc.b	$7F,$DC,$B1,$4F,$7F,$86,$F0,$7C,$44,$AD,$47,$B7,$8A,$47,$77,$E0
	.dc.b	$70,$F9,$FE,$22,$4F,$57,$D5,$F1,$1E,$BF,$88,$93,$E1,$7D,$DF,$11
	.dc.b	$FE,$17,$EF,$71,$5D,$65,$FF,$82,$27,$2F,$A7,$E2,$3E,$37,$C4,$48
	.dc.b	$C1,$ED,$F6,$3E,$57,$AF,$23,$07,$3D,$3C,$1A,$48,$C1,$F1,$7D,$7F
	.dc.b	$A4,$47,$E4,$7A,$FF,$57,$D7,$91,$F2,$DF,$B1,$F0,$9E,$BC,$64,$FA
	.dc.b	$5F,$75,$FD,$DF,$84,$93,$F4,$DF,$C0,$7B,$3F,$81,$F8,$49,$3F,$5D
	.dc.b	$FC,$27,$C4,$7C,$79,$93,$F1,$5F,$95,$3F,$E4,$1F,$C3,$7F,$82,$CF
	.dc.b	$F3,$E6,$FA,$DF,$C2,$3D,$74,$63,$DC,$FE,$C6,$BE,$23,$97,$C4,$73
	.dc.b	$F8,$89,$8F,$F8,$EA,$93,$FE,$4F,$E0,$7C,$2D,$25,$6F,$E2,$3C,$9F
	.dc.b	$11,$23,$93,$E7,$F8,$8F,$47,$B5,$E9,$F7,$25,$77,$8F,$67,$B7,$ED
	.dc.b	$F6,$E4,$F8,$E3,$E0,$65,$F0,$7E,$22,$4F,$9E,$3E,$27,$B7,$23,$D4
	.dc.b	$FC,$7F,$0B,$E4,$7C,$34,$9F,$73,$F2,$FD,$CF,$99,$F1,$1F,$37,$E2
	.dc.b	$25,$7C,$38,$FA,$1E,$D7,$D1,$F6,$E4,$FC,$78,$FF,$3A,$7A,$EF,$8B
	.dc.b	$A4,$9F,$F3,$2C,$7F,$78,$6F,$C2,$F0,$23,$AF,$84,$E7,$ED,$F4,$F7
	.dc.b	$24,$F2,$8E,$FF,$11,$23,$AB,$E2,$F8,$7F,$1F,$C4,$49,$F0,$BE,$6F
	.dc.b	$87,$F3,$FB,$DE,$8F,$7A,$57,$9A,$3D,$7F,$0D,$EC,$F8,$49,$3E,$A8
	.dc.b	$F7,$FC,$2C,$8F,$6B,$F0,$BE,$17,$E1,$FB,$DF,$13,$DE,$95,$F0,$A3
	.dc.b	$E3,$FC,$2C,$8F,$8A,$FC,$AE,$4F,$97,$F0,$B2,$7E,$4B,$FE,$A1,$FC
	.dc.b	$3B,$FD,$5F,$1F,$E1,$E7,$F8,$4B,$7D,$ED,$1B,$F7,$F8,$7C,$2C,$9F
	.dc.b	$F4,$BD,$23,$A7,$C1,$F5,$F8,$4E,$DF,$09,$2F,$71,$E1,$F8,$0F,$EE
	.dc.b	$BE,$BF,$EB,$F1,$46,$3D,$6F,$F2,$8C,$3E,$03,$18,$7F,$8E,$23,$73
	.dc.b	$E2,$F8,$4C,$FB,$29,$FF,$52,$CF,$9B,$E1,$33,$18,$BF,$81,$F0,$98
	.dc.b	$33,$F0,$96,$FD,$BF,$05,$C3,$E0,$B8,$B6,$3E,$0F,$C2,$74,$F8,$4E
	.dc.b	$AD,$8F,$8B,$EE,$78,$3E,$13,$C2,$D8,$F9,$3F,$07,$E4,$3C,$BF,$07
	.dc.b	$63,$E6,$FC,$27,$B3,$F0,$FF,$F3,$A4,$60,$FF,$63,$B8,$DF,$FC,$AE
	.dc.b	$FF,$15,$8C,$FC,$2F,$1F,$83,$E4,$F3,$F8,$5E,$9F,$09,$6B,$9B,$DF
	.dc.b	$E1,$24,$76,$F8,$4B,$97,$6F,$83,$F2,$7C,$27,$95,$B5,$E4,$F8,$3F
	.dc.b	$47,$C2,$7A,$5B,$1E,$8F,$83,$F6,$1E,$DF,$84,$B1,$EC,$F8,$4F,$80
	.dc.b	$7C,$1F,$84,$B1,$F0,$3E,$13,$E2,$7C,$27,$C5,$6C,$7C,$4F,$83,$F9
	.dc.b	$07,$C9,$F8,$4B,$1F,$23,$E0,$FE,$67,$C2,$7C,$D6,$C7,$FB,$CB,$F1
	.dc.b	$1F,$E2,$4D,$F8,$5F,$C6,$C7,$FA,$22,$30,$6F,$E1,$73,$F0,$BA,$6C
	.dc.b	$73,$F8,$4E,$27,$2F,$85,$B1,$DF,$E1,$7A,$FC,$2F,$66,$C7,$8F,$E1
	.dc.b	$7C,$27,$E7,$BD,$E7,$BF,$89,$F4,$7C,$2F,$9B,$E1,$3C,$ED,$8F,$67
	.dc.b	$C2,$FA,$BE,$13,$D6,$D8,$78,$1F,$EA,$0F,$BA,$EB,$18,$F6,$BF,$B2
	.dc.b	$46,$27,$5F,$0B,$FB,$08,$C4,$FF,$69,$E4,$7C,$37,$C8,$AF,$92,$DC
	.dc.b	$3E,$76,$BF,$30,$F9,$BF,$0D,$63,$C0,$7F,$B4,$B1,$7C,$27,$73,$F7
	.dc.b	$FF,$71,$E8,$9D,$E9,$FE,$BC,$AD,$C7,$DB,$E1,$FB,$3F,$0F,$F6,$9B
	.dc.b	$1F,$7F,$E1,$FE,$E9,$F7,$B1,$B1,$D3,$DE,$64,$F8,$2F,$0F,$C1,$F8
	.dc.b	$9B,$5B,$EF,$F2,$6B,$FE,$CA,$7F,$C7,$31,$98,$FE,$2F,$18,$75,$F0
	.dc.b	$F8,$33,$F0,$D6,$BC,$58,$BE,$03,$E0,$FC,$07,$C2,$6C,$F8,$7E,$1B
	.dc.b	$A9,$DB,$3B,$1E,$5D,$3C,$27,$8B,$E1,$AC,$76,$F7,$7E,$69,$F3,$BD
	.dc.b	$EB,$1E,$DF,$86,$F4,$FC,$37,$A9,$B1,$F0,$7E,$1B,$DA,$7B,$B4,$B1
	.dc.b	$F2,$FE,$DF,$F8,$B7,$17,$FB,$B9,$BD,$FF,$E3,$17,$FC,$53,$3B,$F7
	.dc.b	$7E,$3F,$0F,$C8,$F8,$5B,$3F,$3D,$F9,$7F,$0D,$F3,$3E,$16,$CF,$D3
	.dc.b	$7F,$A2,$E0,$F6,$EF,$B9,$86,$D7,$D9,$F8,$6F,$AB,$C7,$F5,$9B,$1F
	.dc.b	$77,$8F,$EC,$FC,$37,$DA,$6C,$7B,$7D,$AE,$FE,$D7,$81,$B1,$F8,$5F
	.dc.b	$86,$FC,$09,$F8,$2F,$86,$B1,$F8,$FF,$DC,$7F,$9E,$72,$7F,$94,$DB
	.dc.b	$1F,$B6,$F0,$91,$F2,$23,$0E,$B9,$92,$E1,$78,$5F,$0B,$D9,$F8,$6D
	.dc.b	$C3,$F3,$DF,$80,$BE,$07,$FB,$5B,$63,$3F,$05,$89,$FE,$DF,$9F,$D6
	.dc.b	$CD,$A7,$9F,$E1,$B3,$F0,$FA,$7F,$BA,$F9,$EF,$6E,$FB,$78,$6E,$9F
	.dc.b	$AF,$97,$B5,$C0,$E3,$ED,$D9,$F8,$DF,$0F,$E2,$3C,$7F,$0F,$63,$E5
	.dc.b	$7C,$3F,$9B,$BF,$CE,$D8,$F9,$DE,$07,$A8,$F5,$F1,$58,$FA,$5C,$7E
	.dc.b	$EF,$87,$F7,$B6,$3A,$9F,$5F,$0E,$AD,$3A,$93,$8F,$E2,$BF,$5A,$FB
	.dc.b	$46,$0F,$AF,$C8,$F8,$7F,$9B,$DB,$F3,$9B,$7F,$85,$EE,$FA,$27,$D2
	.dc.b	$F8,$6B,$1F,$89,$F8,$7F,$AB,$F0,$FF,$59,$B1,$F8,$DF,$87,$FB,$27
	.dc.b	$DA,$ED,$B1,$D4,$9A,$79,$7C,$0F,$C7,$E3,$DA,$31,$AB,$1E,$B1,$86
	.dc.b	$F6,$FE,$67,$BF,$FC,$6B,$FB,$D7,$F6,$CA,$E3,$18,$F8,$8F,$C2,$FB
	.dc.b	$9F,$86,$F6,$EE,$BF,$83,$7F,$13,$ED,$C8,$FE,$69,$F4,$1F,$CB,$B3
	.dc.b	$FF,$5C,$78,$1F,$D0,$33,$F4,$69,$FD,$DD,$BC,$1E,$2F,$88,$F1,$BE
	.dc.b	$4F,$89,$F2,$F8,$77,$3F,$E8,$18,$7E,$5D,$BE,$27,$AB,$C4,$F5,$B6
	.dc.b	$37,$E2,$7B,$BE,$27,$DE,$D8,$E7,$8F,$66,$9D,$B1,$E6,$F8,$30,$BA
	.dc.b	$7A,$9F,$93,$F1,$1F,$2B,$E2,$6D,$79,$9F,$45,$7C,$0B,$BC,$1F,$ED
	.dc.b	$BD,$7D,$53,$F1,$35,$F4,$37,$0F,$01,$F0,$3E,$23,$EB,$9F,$63,$E2
	.dc.b	$6C,$7C,$4F,$87,$FB,$7F,$11,$F7,$1B,$1F,$23,$93,$EF,$9F,$80,$F8
	.dc.b	$8B,$1F,$33,$E1,$FF,$07,$D5,$F8,$46,$C7,$53,$E8,$E1,$D5,$A7,$52
	.dc.b	$67,$F7,$AF,$AB,$F8,$0E,$AF,$C7,$9F,$90,$EA,$B5,$FD,$EF,$F7,$BF
	.dc.b	$C4,$3F,$CA,$DF,$EE,$5C,$3F,$F0,$05,$CD,$FF,$2A,$73,$3B,$F8,$8F
	.dc.b	$F4,$7C,$1B,$F1,$18,$73,$F0,$FA,$F8,$7B,$7E,$17,$8E,$1C,$BE,$36
	.dc.b	$CF,$95,$E4,$F9,$EF,$AB,$4F,$57,$C4,$78,$8F,$1F,$C4,$58,$F7,$7C
	.dc.b	$3F,$98,$F3,$FC,$45,$8F,$85,$F1,$1E,$A3,$D7,$F1,$16,$3E,$37,$C3
	.dc.b	$FB,$BE,$1F,$DE,$D8,$EA,$4D,$7F,$83,$70,$FF,$AB,$78,$BE,$56,$7C
	.dc.b	$E7,$39,$76,$BF,$09,$F4,$FE,$1F,$E5,$FC,$3F,$CC,$6C,$75,$C3,$9B
	.dc.b	$F6,$75,$F5,$61,$D2,$39,$BE,$2D,$7C,$15,$F6,$FE,$1B,$EB,$FC,$3F
	.dc.b	$D8,$6D,$FE,$0F,$E1,$BE,$DF,$C3,$FD,$C6,$C7,$E1,$FE,$1F,$EF,$FC
	.dc.b	$3F,$E0,$1B,$1F,$E4,$57,$FB,$87,$9C,$69,$FD,$D7,$4F,$EE,$6E,$B8
	.dc.b	$98,$76,$9F,$92,$DD,$3F,$25,$5D,$A7,$E6,$B6,$18,$7F,$80,$E3,$FB
	.dc.b	$94,$62,$7B,$FC,$2F,$FB,$92,$5F,$0B,$63,$1F,$0B,$9F,$85,$B8,$FA
	.dc.b	$DE,$1F,$0B,$23,$DD,$F0,$57,$2F,$E8,$99,$3E,$17,$B1,$DF,$E0,$ED
	.dc.b	$FC,$5C,$BB,$34,$EC,$7C,$9C,$7A,$6B,$E9,$E0,$F9,$59,$E2,$BD,$DF
	.dc.b	$09,$EB,$3D,$9F,$0B,$63,$E9,$7C,$27,$BF,$E1,$7E,$03,$63,$EB,$7C
	.dc.b	$2F,$C3,$F8,$2F,$88,$D8,$FB,$5F,$03,$F1,$FE,$17,$E4,$36,$3A,$9F
	.dc.b	$7F,$0E,$AD,$3A,$9F,$83,$BE,$AD,$3F,$0D,$F0,$BF,$4F,$E1,$7E,$A3
	.dc.b	$63,$AF,$2F,$CC,$BE,$AD,$8E,$AF,$D0,$BE,$AD,$8F,$C9,$7C,$27,$DF
	.dc.b	$3F,$05,$C3,$F2,$FF,$7C,$FE,$E0,$FC,$DE,$74,$EA,$BF,$C9,$95,$FD
	.dc.b	$9E,$B8,$71,$E7,$0F,$F2,$5C,$BE,$0F,$12,$8E,$17,$CB,$E0,$73,$06
	.dc.b	$FB,$D8,$73,$EE,$EB,$E0,$6D,$FF,$AB,$EF,$FC,$82,$DC,$39,$9D,$31
	.dc.b	$B1,$BC,$3A,$3C,$71,$B9,$7F,$86,$70,$7C,$17,$90,$F2,$FC,$1D,$BF
	.dc.b	$07,$C1,$FA,$3E,$0F,$D2,$D8,$F2,$7C,$17,$B3,$E0,$BD,$AD,$8F,$47
	.dc.b	$C1,$7C,$0F,$7F,$E0,$B6,$3A,$9E,$DC,$3A,$B4,$EA,$7C,$1B,$EA,$D3
	.dc.b	$E2,$7C,$17,$CC,$3E,$6F,$C0,$58,$F9,$1F,$03,$F4,$0F,$A3,$F0,$56
	.dc.b	$3E,$67,$C1,$7D,$4F,$82,$FA,$AD,$8F,$F6,$F7,$EF,$BF,$D8,$39,$CF
	.dc.b	$F6,$0E,$4F,$F4,$5B,$93,$11,$F8,$1C,$7B,$4F,$CF,$C6,$F5,$7E,$47
	.dc.b	$1E,$D3,$F4,$DB,$1F,$BF,$C3,$FB,$56,$3E,$2F,$83,$FC,$6C,$62,$FF
	.dc.b	$CA,$4A,$3A,$CF,$E6,$31,$F7,$BF,$08,$F7,$9F,$8B,$CF,$DE,$D3,$DF
	.dc.b	$F0,$7D,$BE,$0F,$BB,$63,$E1,$FC,$0F,$88,$F1,$FC,$0D,$8F,$8F,$EF
	.dc.b	$F9,$BD,$CF,$3B,$63,$E5,$FB,$FE,$A3,$D7,$EE,$D8,$EA,$4D,$FD,$B8
	.dc.b	$75,$7D,$F7,$CD,$B3,$F5,$B0,$F8,$BF,$01,$F1,$BE,$06,$CF,$DA,$7E
	.dc.b	$4F,$C0,$48,$FB,$BF,$03,$70,$F9,$7F,$01,$F3,$BE,$07,$E7,$B6,$BF
	.dc.b	$07,$EF,$7D,$2F,$81,$FA,$6D,$8F,$C3,$E3,$CF,$A3,$ED,$65,$67,$FD
	.dc.b	$AD,$D0,$C7,$A6,$7E,$09,$FC,$7E,$7F,$7B,$CD,$FB,$FF,$03,$67,$F2
	.dc.b	$6F,$F0,$EF,$DF,$BF,$B6,$93,$E2,$F1,$39,$F8,$1D,$7C,$15,$86,$5F
	.dc.b	$E9,$9B,$FE,$C3,$18,$DB,$FA,$74,$3D,$FE,$DF,$05,$DD,$B5,$B3,$FD
	.dc.b	$E1,$3E,$27,$C9,$97,$89,$D6,$1F,$EF,$09,$FE,$E7,$1C,$31,$F2,$4F
	.dc.b	$8D,$F6,$7C,$04,$9E,$3E,$22,$DF,$01,$9C,$3D,$55,$DD,$F6,$63,$74
	.dc.b	$F5,$EB,$F0,$AB,$9E,$1E,$2B,$F2,$67,$F2,$FE,$0B,$E6,$7C,$15,$BD
	.dc.b	$BF,$3F,$E0,$BE,$87,$C1,$59,$E6,$FD,$3F,$82,$91,$DB,$DD,$B9,$7C
	.dc.b	$FF,$82,$FB,$1E,$F7,$D9,$6D,$79,$31,$EA,$D3,$A9,$E8,$C3,$9E,$FF
	.dc.b	$5B,$1B,$3F,$4B,$E0,$7F,$08,$7E,$17,$E0,$AC,$7C,$0F,$81,$FC,$41
	.dc.b	$F8,$9F,$82,$B1,$F1,$3E,$0B,$F1,$87,$F5,$AF,$3B,$FB,$2C,$7D,$E8
	.dc.b	$C7,$44,$8F,$AB,$3F,$D8,$5B,$1D,$38,$3F,$1D,$87,$79,$F1,$1D,$C9
	.dc.b	$1F,$EE,$BA,$EE,$FE,$1E,$FC,$5E,$FF,$8F,$DF,$B7,$E0,$7C,$DF,$05
	.dc.b	$E7,$F8,$2B,$3E,$47,$D5,$F0,$12,$3C,$FF,$01,$72,$F0,$7C,$07,$BC
	.dc.b	$F8,$1E,$ED,$AF,$6E,$3D,$5A,$75,$3E,$0E,$1C,$F9,$3F,$C6,$DB,$32
	.dc.b	$C5,$7E,$C3,$81,$2F,$80,$F9,$C7,$CF,$F7,$AD,$76,$D7,$E2,$6D,$EC
	.dc.b	$6C,$78,$08,$FC,$9C,$7B,$AC,$7B,$FB,$FF,$6F,$E0,$3E,$E3,$6F,$F5
	.dc.b	$BF,$8D,$7E,$FD,$BE,$3F,$F8,$6A,$1E,$45,$8C,$79,$19,$F2,$2D,$F2
	.dc.b	$7F,$11,$F0,$3F,$89,$F8,$1B,$3F,$86,$7F,$19,$EF,$7E,$37,$E0,$6C
	.dc.b	$FE,$29,$FC,$87,$C0,$FE,$47,$96,$CF,$E3,$9F,$F1,$6F,$EF,$FF,$DC
	.dc.b	$C9,$F0,$F8,$E3,$E6,$7E,$2F,$0B,$D1,$F5,$1E,$98,$73,$3F,$33,$8E
	.dc.b	$29,$5F,$E6,$FC,$3F,$B7,$37,$E0,$FC,$3F,$0F,$E2,$7E,$37,$85,$F1
	.dc.b	$FC,$3B,$97,$CC,$7E,$57,$87,$FE,$44,$FC,$6F,$F3,$BD,$4E,$8F,$F2
	.dc.b	$7D,$FF,$1E,$9C,$E8,$F8,$4F,$79,$F0,$3E,$16,$D6,$FC,$6E,$87,$5F
	.dc.b	$1A,$C7,$3F,$13,$E3,$FC,$2F,$85,$B1,$FC,$B3,$17,$C1,$FC,$C3,$E6
	.dc.b	$FC,$1D,$9F,$E7,$3C,$9F,$CC,$3C,$DF,$F3,$03,$FC,$CB,$CD,$BD,$E9
	.dc.b	$B1,$F5,$BE,$0E,$5F,$7F,$85,$BD,$5F,$3A,$3F,$49,$A3,$E1,$7E,$E9
	.dc.b	$F7,$BE,$16,$CF,$8B,$37,$89,$E0,$3C,$2D,$8F,$67,$C2,$FE,$1B,$E0
	.dc.b	$FF,$0E,$D8,$F8,$1F,$07,$F8,$A3,$F1,$7F,$05,$63,$B1,$EC,$F7,$FD
	.dc.b	$AD,$3E,$47,$C1,$7E,$4B,$E1,$3F,$26,$D8,$F9,$9F,$0B,$F9,$6F,$84
	.dc.b	$FC,$BB,$63,$E5,$FE,$37,$FE,$01,$6E,$3F,$32,$17,$63,$F3,$FF,$0B
	.dc.b	$FE,$8C,$FE,$00,$FF,$75,$F1,$A3,$18,$67,$48,$FF,$68,$AF,$EB,$DE
	.dc.b	$21,$FF,$78,$75,$33,$73,$F8,$4F,$17,$35,$C3,$FB,$CF,$D6,$3E,$93
	.dc.b	$F8,$E4,$61,$BD,$3F,$6C,$FE,$21,$97,$F8,$31,$C5,$71,$F5,$FD,$5E
	.dc.b	$A4,$C3,$CD,$EA,$DB,$E5,$AF,$97,$C7,$F8,$9E,$3F,$C5,$6D,$F8,$7C
	.dc.b	$6F,$F5,$96,$7E,$56,$EA,$F1,$6C,$6B,$E0,$9E,$65,$3E,$67,$83,$71
	.dc.b	$FC,$FF,$E3,$8F,$47,$F3,$EF,$F3,$88,$CC,$63,$D4,$EF,$F2,$1E,$0E
	.dc.b	$9B,$5F,$15,$F0,$65,$23,$E5,$AF,$82,$C7,$1F,$C0,$C9,$ED,$78,$0F
	.dc.b	$0F,$B7,$70,$E3,$F0,$5E,$B3,$D9,$EA,$D8,$EB,$F0,$BE,$FF,$3F,$E0
	.dc.b	$36,$3F,$77,$F0,$7F,$E9,$9E,$A6,$7F,$69,$FF,$72,$C6,$63,$13,$FE
	.dc.b	$48,$E8,$6E,$F4,$3E,$67,$C2,$FC,$D6,$D7,$BB,$CF,$FA,$1F,$0F,$F4
	.dc.b	$5B,$1F,$E4,$4F,$B6,$7E,$2A,$F4,$7F,$DD,$91,$86,$C3,$D2,$B1,$F6
	.dc.b	$FE,$26,$5F,$CC,$F8,$9B,$D1,$F9,$0D,$9E,$67,$E0,$4F,$C1,$7C,$65
	.dc.b	$BF,$A7,$F1,$7F,$85,$F2,$7F,$0C,$D8,$FA,$FE,$4F,$E2,$4F,$C5,$78
	.dc.b	$76,$3F,$4B,$F5,$3F,$B8,$6F,$AD,$F9,$48,$C6,$9F,$A2,$51,$9F,$56
	.dc.b	$57,$2F,$5A,$DE,$FD,$3F,$CC,$4F,$36,$BE,$0F,$47,$B1,$DF,$D8,$B1
	.dc.b	$E4,$E4,$F1,$72,$F8,$DB,$1E,$8F,$63,$CD,$EB,$F9,$DB,$1E,$CF,$4B
	.dc.b	$D4,$7A,$FD,$CB,$1F,$03,$DC,$F7,$7B,$9E,$F6,$C7,$C4,$F7,$3E,$11
	.dc.b	$F0,$FD,$9B,$1F,$C3,$3E,$B3,$F5,$9C,$3F,$C1,$9B,$E4,$27,$F2,$1C
	.dc.b	$DF,$E9,$87,$87,$A5,$F3,$7E,$16,$47,$D2,$F8,$4B,$8F,$CF,$F8,$4F
	.dc.b	$A4,$7D,$3F,$6E,$D7,$87,$D2,$FE,$13,$A1,$BF,$11,$87,$3F,$11,$AF
	.dc.b	$88,$B7,$E9,$78,$FC,$47,$2F,$6E,$CF,$B5,$EB,$F0,$1F,$EB,$0D,$7D
	.dc.b	$CD,$E2,$F7,$43,$D2,$6B,$FE,$49,$8F,$8D,$EA,$6F,$1F,$D9,$4B,$E3
	.dc.b	$71,$5C,$63,$FC,$98,$FC,$DF,$8C,$CE,$1D,$2A,$D3,$E3,$2F,$B9,$30
	.dc.b	$FA,$91,$AF,$91,$DF,$C4,$C9,$FB,$11,$CB,$E0,$F9,$FC,$24,$9F,$B9
	.dc.b	$1D,$BE,$37,$BF,$C7,$49,$FC,$04,$79,$3E,$51,$F0,$77,$E0,$E5,$7F
	.dc.b	$50,$9C,$B7,$E1,$F8,$11,$D7,$C7,$FB,$3E,$23,$DB,$F1,$12,$7F,$1D
	.dc.b	$1D,$FE,$1B,$C1,$F0,$F2,$7B,$47,$8F,$E1,$64,$78,$5F,$37,$C3,$79
	.dc.b	$FE,$02,$4F,$95,$F5,$7C,$47,$AF,$DD,$93,$E9,$7D,$DE,$EF,$BF,$E1
	.dc.b	$BE,$07,$C3,$CA,$F7,$47,$C3,$F8,$59,$1F,$05,$F8,$DF,$0B,$F6,$BC
	.dc.b	$BF,$F0,$04,$FC,$5F,$0F,$E5,$19,$F7,$BE,$67,$C2,$C9,$F9,$4F,$CF
	.dc.b	$F8,$5F,$A1,$F0,$9F,$47,$E1,$65,$7C,$F8,$ED,$EE,$C8,$FF,$00,$E4
	.dc.b	$8F,$17,$BD,$E3,$F8,$DF,$27,$C3,$4B,$DC,$79,$FE,$1B,$D1,$F0,$D2
	.dc.b	$79,$C7,$AF,$E1,$A4,$76,$7D,$DF,$0D,$F9,$CF,$49,$C9,$FE,$92,$AF
	.dc.b	$F2,$9B,$5F,$F0,$3B,$63,$F5,$63,$F5,$86,$FF,$27,$E1,$A5,$FB,$1F
	.dc.b	$99,$F0,$FF,$37,$E1,$3E,$77,$C2,$CA,$F8,$31,$E3,$FC,$77,$F9,$CC
	.dc.b	$7F,$99,$16,$D9,$35,$C3,$F5,$FE,$1B,$EC,$7C,$34,$BF,$97,$1F,$6F
	.dc.b	$E0,$A4,$7C,$E7,$F7,$7E,$4F,$F8,$DF,$BF,$C1,$C7,$E1,$DE,$5F,$86
	.dc.b	$3F,$09,$F1,$12,$3E,$BB,$F8,$7F,$88,$E7,$F1,$AF,$0E,$F3,$CB,$FF
	.dc.b	$23,$47,$B6,$BD,$DC,$52,$7D,$D1,$F9,$1F,$87,$FC,$97,$C3,$C9,$FC
	.dc.b	$2C,$76,$F1,$7B,$F9,$92,$7F,$CA,$51,$F2,$58,$7C,$93,$E6,$4F,$F1
	.dc.b	$6F,$99,$7C,$25,$24,$FF,$94,$A3,$FC,$68,$FF,$97,$7C,$FF,$69,$1F
	.dc.b	$F3,$07,$9F,$F8,$9D,$1E,$2D,$24,$FF,$A7,$21,$F0,$DF,$1B,$D0,$F8
	.dc.b	$FE,$84,$AF,$F3,$54,$7C,$AF,$42,$47,$F9,$B5,$F9,$BE,$B7,$7F,$83
	.dc.b	$F0,$7C,$0C,$AE,$D1,$F4,$BD,$8F,$A7,$EB,$C9,$FF,$3B,$47,$D6,$F5
	.dc.b	$E4,$7F,$9E,$5F,$C1,$FF,$00,$BD,$8F,$67,$BF,$2B,$D4,$FB,$FD,$DD
	.dc.b	$F8,$72,$47,$F9,$D5,$59,$EB,$EF,$F6,$B9,$5C,$3C,$FF,$EA,$B5,$C2
	.dc.b	$7C,$78,$7E,$D7,$F1,$C6,$97,$59,$8F,$FB,$22,$72,$F3,$F6,$BC,$78
	.dc.b	$F8,$BE,$03,$DB,$EF,$CA,$D1,$E5,$DB,$FC,$F4,$E2,$BD,$3F,$01,$F1
	.dc.b	$3E,$1A,$4E,$69,$3F,$AA,$FC,$71,$E6,$52,$57,$38,$FD,$23,$7C,$3E
	.dc.b	$04,$7A,$73,$7D,$2E,$79,$1D,$DF,$AB,$F0,$FD,$FC,$0F,$07,$64,$AE
	.dc.b	$D1,$E3,$ED,$F2,$76,$C9,$F1,$47,$DE,$F8,$4F,$BF,$F0,$B2,$7D,$91
	.dc.b	$F8,$2F,$8A,$91,$EF,$7F,$0B,$E1,$FE,$1B,$C2,$FC,$3F,$C5,$4A,$F8
	.dc.b	$91,$F8,$AF,$1E,$47,$C7,$7F,$1B,$F1,$5D,$23,$F0,$DC,$DF,$E2,$37
	.dc.b	$B1,$F3,$70,$EC,$FC,$FC,$3B,$3C,$30,$FC,$BF,$C6,$7F,$1A,$F1,$7F
	.dc.b	$C4,$13,$F9,$16,$F8,$7F,$8C,$C5,$5F,$C8,$3F,$1C,$73,$52,$61,$F8
	.dc.b	$18,$FF,$10,$37,$97,$87,$04,$8C,$1F,$87,$C9,$E5,$F8,$FC,$D9,$39
	.dc.b	$8F,$37,$AB,$E7,$F8,$D9,$3C,$63,$D5,$EA,$48,$E8,$FB,$7D,$1F,$77
	.dc.b	$9D,$EF,$F7,$25,$78,$63,$AB,$E3,$C3,$AB,$F0,$F0,$E8,$FC,$5C,$2F
	.dc.b	$17,$EC,$34,$65,$F2,$FD,$89,$7E,$D7,$E7,$7B,$3D,$A3,$E7,$65,$DB
	.dc.b	$0E,$59,$74,$F2,$7F,$63,$93,$3C,$61,$FC,$0E,$3B,$BF,$27,$E0,$3E
	.dc.b	$D7,$B3,$23,$07,$CD,$F6,$BE,$F1,$1F,$A1,$EF,$7E,$07,$E1,$3F,$05
	.dc.b	$F0,$92,$7E,$A4,$7E,$17,$E1,$3F,$0D,$EF,$C9,$FB,$11,$CD,$FC,$46
	.dc.b	$17,$7F,$89,$8F,$59,$FB,$D8,$F5,$7F,$1D,$8F,$E4,$7E,$03,$F2,$5F
	.dc.b	$09,$27,$F0,$91,$F9,$5F,$88,$FE,$DF,$F8,$77,$F9,$CA,$7F,$93,$37
	.dc.b	$2F,$A5,$3A,$F8,$99,$5F,$C4,$EB,$FD,$9A,$DF,$89,$E7,$F0,$FD,$3E
	.dc.b	$26,$4E,$E3,$C0,$F1,$E6,$F0,$98,$7C,$38,$76,$69,$D4,$F0,$7C,$5F
	.dc.b	$A3,$E2,$E4,$60,$F2,$7A,$DE,$C2,$3E,$7F,$87,$F7,$FC,$47,$C0,$F5
	.dc.b	$E4,$FA,$E3,$E1,$FC,$47,$C4,$F8,$79,$3E,$F8,$E6,$FC,$6C,$2F,$17
	.dc.b	$F4,$6E,$1F,$E5,$73,$F2,$B0,$EA,$FC,$CC,$7C,$2F,$C9,$F5,$FE,$89
	.dc.b	$1F,$99,$EC,$74,$E0,$FE,$8F,$87,$F3,$A9,$F6,$9E,$B3,$D0,$7D,$3B
	.dc.b	$EF,$8F,$8F,$1F,$BB,$EB,$CA,$FB,$0F,$E0,$3D,$8F,$C0,$FC,$3C,$9F
	.dc.b	$B8,$FE,$13,$E0,$FF,$0B,$F0,$7F,$86,$F1,$25,$7E,$06,$3F,$C6,$FF
	.dc.b	$87,$7F,$75,$8F,$EA,$CD,$7F,$B8,$37,$3F,$ED,$6F,$F7,$18,$EE,$43
	.dc.b	$F2,$38,$78,$5E,$C7,$E3,$EF,$C2,$FE,$5A,$FC,$3F,$0F,$23,$C2,$F9
	.dc.b	$3E,$1F,$CB,$F0,$FE,$6F,$86,$95,$E6,$8F,$4F,$C3,$7A,$BE,$12,$4F
	.dc.b	$AA,$3D,$BF,$0B,$23,$DA,$FC,$0F,$6B,$AC,$7C,$0C,$3A,$BF,$03,$0E
	.dc.b	$AF,$C4,$C3,$AB,$F1,$30,$F9,$3F,$09,$F2,$BD,$A9,$3F,$2A,$3E,$6F
	.dc.b	$B5,$23,$E6,$BF,$43,$E1,$3E,$8F,$C2,$49,$FA,$2F,$D4,$F8,$2F,$AB
	.dc.b	$F0,$3F,$5B,$E0,$A5,$7D,$68,$E8,$FD,$6C,$2F,$3F,$D6,$C7,$A4,$FD
	.dc.b	$AC,$6C,$FD,$AC,$7F,$03,$F0,$9F,$82,$F8,$49,$7F,$82,$8E,$7E,$07
	.dc.b	$F8,$83,$AB,$FD,$E1,$3E,$A3,$F0,$F9,$74,$9D,$65,$67,$FC,$6D,$5F
	.dc.b	$E5,$A7,$F1,$BF,$07,$F9,$02,$3F,$E2,$5F,$C3,$BF,$B6,$DF,$F8,$B9
	.dc.b	$FE,$C3,$19,$F8,$3F,$F4,$6C,$1B,$F0,$78,$F7,$65,$73,$77,$EF,$70
	.dc.b	$F7,$A4,$F7,$79,$FC,$0F,$28,$EC,$43,$FC,$A1,$8F,$89,$B7,$D9,$A7
	.dc.b	$A6,$FC,$BF,$05,$23,$D8,$FA,$3E,$0F,$D3,$F0,$7E,$AF,$83,$95,$F0
	.dc.b	$63,$DB,$F0,$7E,$EF,$83,$93,$F1,$63,$E0,$FC,$0C,$8F,$90,$FC,$4F
	.dc.b	$81,$EB,$1F,$2F,$0E,$AF,$AB,$0E,$AF,$CF,$C3,$AB,$EE,$C3,$E6,$FC
	.dc.b	$04,$8F,$A8,$FD,$0F,$81,$FA,$3F,$05,$F4,$BE,$0E,$57,$D9,$8F,$AB
	.dc.b	$F0,$52,$3E,$E3,$F6,$3E,$0F,$EC,$FC,$1F,$FB,$18,$FE,$AD,$D3,$FE
	.dc.b	$6C,$6D,$7F,$9F,$79,$51,$E0,$8F,$C3,$63,$E0,$7E,$AE,$3D,$DF,$AF
	.dc.b	$8D,$F0,$7F,$D0,$38,$32,$F0,$7C,$04,$BF,$F3,$A6,$5F,$D8,$9F,$F4
	.dc.b	$72,$F8,$3F,$04,$7E,$0B,$3F,$06,$3D,$33,$ED,$A7,$E5,$F4,$B7,$E9
	.dc.b	$CF,$DD,$F0,$52,$3B,$BF,$07,$E0,$FE,$17,$C1,$C9,$F1,$BF,$17,$E0
	.dc.b	$FE,$37,$C0,$49,$F3,$BF,$27,$E0,$FE,$57,$BB,$F2,$FD,$E9,$5E,$C8
	.dc.b	$E8,$FB,$30,$BB,$FF,$A9,$38,$99,$75,$AF,$83,$95,$AD,$65,$F5,$FE
	.dc.b	$0B,$EC,$7C,$1C,$BF,$93,$1F,$6F,$E0,$E4,$7C,$C7,$EF,$7C,$1F,$DF
	.dc.b	$F8,$29,$3F,$41,$FC,$17,$C1,$7E,$0F,$E0,$64,$FD,$47,$F0,$DF,$05
	.dc.b	$D2,$3E,$AE,$17,$4F,$56,$3D,$27,$EC,$E3,$67,$DD,$8F,$E3,$FE,$07
	.dc.b	$F2,$1F,$05,$2F,$F0,$31,$FE,$CE,$7F,$D9,$DF,$87,$7F,$5F,$EA,$FF
	.dc.b	$5F,$BF,$D7,$E3,$3F,$11,$BC,$7F,$D9,$B3,$B9,$FF,$6A,$2F,$84,$D5
	.dc.b	$F2,$6C,$E9,$B6,$A7,$C2,$61,$7C,$0D,$F8,$29,$87,$FB,$2E,$72,$DF
	.dc.b	$8A,$E1,$84,$8C,$1B,$C7,$A1,$1E,$58,$F7,$F8,$2F,$07,$C1,$C9,$ED
	.dc.b	$1E,$3C,$3A,$47,$C2,$3A,$9F,$22,$FA,$34,$EA,$79,$FE,$17,$D9,$F0
	.dc.b	$B2,$30,$7A,$FE,$13,$E0,$11,$F7,$7C,$2F,$C3,$F8,$3F,$89,$F0,$12
	.dc.b	$7E,$14,$7C,$7F,$80,$91,$F1,$5F,$95,$F0,$7C,$E3,$EE,$E1,$79,$7E
	.dc.b	$CE,$3D,$67,$E6,$63,$D5,$FC,$1E,$3F,$4F,$E1,$24,$7D,$17,$EB,$7C
	.dc.b	$2F,$D7,$F8,$49,$3F,$55,$FB,$5F,$0B,$C2,$3F,$1B,$57,$2F,$C4,$FC
	.dc.b	$27,$E0,$BF,$0E,$FE,$C3,$4E,$74,$7F,$51,$A7,$5A,$B5,$39,$8C,$57
	.dc.b	$09,$F1,$9F,$9C,$CA,$DF,$E6,$32,$F1,$5F,$E7,$B2,$B7,$F9,$AC,$BC
	.dc.b	$9F,$0B,$E5,$F8,$29,$3E,$1C,$7C,$6F,$8A,$7D,$5F,$05,$23,$58,$F9
	.dc.b	$35,$4F,$BF,$E1,$3E,$07,$C2,$7C,$1F,$84,$98,$70,$3D,$B1,$F1,$7E
	.dc.b	$12,$46,$78,$7E,$2C,$7C,$9F,$81,$F0,$4B,$5F,$2D,$7C,$AC,$B3,$18
	.dc.b	$CF,$FD,$95,$5C,$DF,$F6,$4B,$C7,$4F,$46,$7E,$48,$C1,$E4,$38,$7C
	.dc.b	$17,$D9,$F8,$19,$18,$3A,$63,$E0,$BF,$84,$FC,$18,$F0,$6B,$D7,$2F
	.dc.b	$26,$3F,$12,$7C,$9F,$05,$F8,$52,$3E,$7F,$84,$FC,$47,$C2,$7E,$27
	.dc.b	$E1,$24,$FA,$E3,$F1,$9F,$09,$23,$DC,$FE,$3F,$E0,$3F,$21,$F0,$3F
	.dc.b	$91,$F8,$19,$5F,$0E,$3A,$BF,$17,$0E,$AF,$C2,$E3,$FC,$AE,$3D,$63
	.dc.b	$1C,$6C,$B1,$B7,$CF,$CF,$F3,$DF,$03,$2F,$E8,$3F,$A1,$F7,$3F,$91
	.dc.b	$FD,$73,$7B,$9C,$BF,$C0,$4E,$5B,$AF,$02,$3A,$CF,$F9,$6F,$E1,$BF
	.dc.b	$E0,$CD,$8D,$27,$C2,$EB,$E1,$77,$F0,$B3,$0F,$C7,$57,$F8,$0D,$BF
	.dc.b	$0B,$D3,$E0,$E4,$60,$DF,$C1,$F8,$3E,$0E,$46,$0E,$78,$F5,$7E,$56
	.dc.b	$0C,$7C,$51,$CC,$CE,$1D,$9D,$E3,$EC,$F8,$3F,$6F,$C2,$49,$F3,$C7
	.dc.b	$F4,$CF,$AE,$FF,$68,$47,$D7,$7F,$DB,$BC,$FF,$A4,$6F,$7F,$F0,$87
	.dc.b	$71,$9B,$E4,$4A,$FD,$84,$BE,$1B,$E6,$9F,$3B,$E1,$64,$FC,$88,$F1
	.dc.b	$7C,$1F,$8F,$E0,$E4,$F2,$8F,$37,$C1,$FD,$6F,$86,$93,$DA,$3E,$CC
	.dc.b	$FD,$AA,$93,$F5,$23,$DD,$F0,$5E,$FF,$83,$93,$E6,$8F,$C0,$DC,$8F
	.dc.b	$B6,$FE,$13,$E1,$3F,$0B,$F0,$D2,$7E,$FB,$F8,$8F,$86,$FE,$EB,$F5
	.dc.b	$BF,$DB,$A7,$F4,$0B,$E1,$B2,$6B,$E1,$FF,$1F,$94,$AF,$C4,$BF,$92
	.dc.b	$CF,$FC,$03,$F7,$3F,$CE,$7E,$12,$D7,$F8,$37,$97,$F9,$5C,$FF,$3A
	.dc.b	$8F,$E4,$4E,$DF,$E4,$19,$3D,$FF,$29,$23,$3E,$FF,$4F,$84,$91,$C8
	.dc.b	$FF,$0A,$17,$E1,$FC,$3A,$F8,$B6,$93,$B8,$F8,$3E,$E4,$8F,$0B,$F1
	.dc.b	$3D,$FF,$4F,$C3,$7A,$BE,$1E,$57,$78,$F6,$FC,$3F,$BB,$E1,$E4,$F8
	.dc.b	$E3,$FC,$43,$D5,$FD,$66,$3F,$B0,$3F,$EB,$D7,$F9,$F3,$5F,$E2,$0E
	.dc.b	$F8,$7F,$B1,$70,$7C,$AE,$09,$5E,$F8,$F9,$BF,$09,$F3,$BE,$16,$4F
	.dc.b	$C3,$8F,$0F,$B3,$E2,$F7,$E4,$F8,$23,$EA,$FC,$3C,$8F,$94,$FD,$8F
	.dc.b	$87,$FB,$3C,$5F,$6B,$8A,$57,$CF,$8F,$6F,$AB,$23,$D6,$FC,$0F,$57
	.dc.b	$F0,$3F,$0F,$27,$EB,$3F,$84,$F8,$7F,$C2,$FC,$3C,$9F,$B4,$FE,$2B
	.dc.b	$EE,$7F,$C0,$FE,$B3,$97,$FD,$3D,$5F,$23,$DF,$F9,$A7,$CE,$F7,$38
	.dc.b	$7C,$5F,$1F,$88,$95,$F8,$EC,$7E,$60,$F8,$9E,$DF,$0B,$27,$51,$E1
	.dc.b	$F8,$9F,$F2,$CF,$D7,$3D,$7A,$66,$BA,$7C,$3F,$9F,$A6,$46,$0F,$07
	.dc.b	$57,$AF,$E2,$A4,$60,$F2,$7C,$57,$BF,$E2,$A4,$60,$F4,$7C,$57,$C3
	.dc.b	$23,$EB,$EA,$F8,$DD,$5F,$1F,$E1,$64,$FB,$E3,$E5,$7C,$2C,$8F,$84
	.dc.b	$FC,$DF,$88,$F9,$DF,$15,$F3,$FA,$65,$7C,$78,$FA,$5D,$3F,$4F,$E2
	.dc.b	$64,$FC,$B8,$FD,$FF,$A5,$FE,$DC,$F2,$7F,$C2,$9F,$11,$83,$3F,$03
	.dc.b	$AF,$86,$DF,$C2,$CC,$7F,$80,$E1,$FD,$45,$FF,$08,$E1,$FE,$1D,$E9
	.dc.b	$EA,$48,$DF,$C2,$F8,$3E,$22,$47,$27,$C7,$F1,$1E,$4F,$87,$F2,$FC
	.dc.b	$3C,$AE,$F1,$E8,$F8,$5F,$4F,$C3,$49,$F1,$C7,$B3,$E1,$A4,$79,$9F
	.dc.b	$7F,$C3,$7C,$0F,$2B,$E0,$FC,$34,$AF,$5C,$7C,$4F,$86,$91,$EE,$7E
	.dc.b	$3F,$C3,$7C,$8F,$86,$93,$F0,$9F,$97,$F0,$DF,$33,$E1,$A4,$FC,$67
	.dc.b	$F9,$6F,$E1,$7F,$EB,$3E,$53,$FC,$B7,$A7,$FC,$D1,$18,$F4,$B6,$70
	.dc.b	$F8,$5E,$3F,$0B,$31,$FE,$63,$87,$FA,$D1,$BF,$09,$DB,$E1,$7B,$FC
	.dc.b	$24,$9D,$C7,$8B,$E1,$7F,$CC,$BD,$DF,$CF,$30,$7C,$26,$36,$F1,$C2
	.dc.b	$B1,$83,$6F,$4D,$E1,$F5,$FC,$2D,$B3,$7E,$13,$0E,$7E,$16,$65,$EA
	.dc.b	$F8,$4B,$97,$E2,$B0,$7C,$2F,$3F,$85,$E8,$DB,$F8,$3F,$0B,$DC,$F0
	.dc.b	$7C,$25,$8F,$8B,$F0,$BE,$33,$C9,$F0,$B6,$3E,$4F,$C2,$F9,$CF,$47
	.dc.b	$C2,$D8,$FE,$BB,$F0,$FF,$E9,$B6,$EB,$EF,$87,$C3,$58,$C7,$C3,$4A
	.dc.b	$E3,$F0,$96,$7F,$4B,$7F,$E8,$E9,$EB,$F0,$BC,$FE,$17,$A3,$D7,$E1
	.dc.b	$24,$F8,$BE,$12,$E1,$CB,$E1,$3C,$47,$8F,$E1,$2D,$7A,$3E,$13,$CD
	.dc.b	$F0,$9E,$76,$C7,$B3,$E1,$3D,$5F,$09,$EB,$6C,$7C,$0F,$83,$F7,$1E
	.dc.b	$FF,$84,$B1,$F1,$3E,$13,$E1,$7C,$27,$C3,$6C,$7C,$8F,$83,$F8,$C7
	.dc.b	$C7,$F8,$4B,$1F,$33,$E0,$FE,$51,$F2,$FE,$12,$C7,$F8,$33,$F0,$BF
	.dc.b	$F6,$5B,$4F,$C7,$2F,$85,$B1,$8F,$85,$8F,$1F,$85,$BC,$BF,$C5,$2F
	.dc.b	$FA,$14,$F5,$F8,$4E,$7F,$0B,$D1,$EB,$F0,$B2,$7C,$5F,$0B,$70,$E5
	.dc.b	$F0,$9E,$2F,$85,$F1,$B6,$BD,$1F,$09,$E6,$F8,$5F,$3B,$63,$D9,$F0
	.dc.b	$BE,$AF,$85,$F5,$B6,$3F,$D9,$1F,$5A,$F3,$BF,$5F,$EF,$A3,$13,$AF
	.dc.b	$47,$31,$8D,$7D,$DF,$0B,$F1,$BE,$1B,$E3,$B6,$FE,$67,$C3,$7C,$AF
	.dc.b	$86,$F9,$6D,$8F,$0F,$A1,$D0,$F0,$FD,$6F,$E8,$71,$FF,$56,$64,$9F
	.dc.b	$A7,$F0,$F2,$FE,$C7,$C3,$DB,$FA,$1F,$0F,$F6,$0F,$E0,$DF,$84,$3D
	.dc.b	$2C,$5F,$FB,$9F,$37,$A5,$E0,$F4,$7C,$2D,$9F,$F2,$7D,$B8,$F9,$FC
	.dc.b	$47,$46,$CE,$BE,$23,$B9,$E0,$F8,$8B,$1D,$FD,$1F,$6F,$9F,$EE,$6C
	.dc.b	$76,$F8,$7F,$3F,$C4,$7A,$1B,$1E,$2F,$87,$F5,$9E,$CF,$87,$B1,$EB
	.dc.b	$F4,$7E,$49,$F2,$BD,$0B,$1E,$AF,$85,$F8,$67,$C4,$F8,$6B,$1E,$EF
	.dc.b	$86,$F8,$FF,$0D,$F2,$1B,$0E,$9F,$F6,$F7,$CC,$7E,$8E,$31,$F1,$27
	.dc.b	$5C,$89,$D5,$70,$F4,$BF,$1B,$18,$BF,$F3,$EF,$3F,$F3,$0C,$3E,$7F
	.dc.b	$8B,$F6,$3E,$1F,$EC,$BF,$6B,$C7,$93,$F5,$3E,$1E,$E7,$E4,$F3,$FA
	.dc.b	$F9,$DD,$9B,$5F,$6B,$E1,$FF,$07,$F0,$FF,$84,$6C,$7D,$EF,$87,$FC
	.dc.b	$39,$F8,$8F,$22,$C7,$B8,$F6,$F0,$3C,$FF,$5F,$9B,$EC,$6C,$FF,$A5
	.dc.b	$BE,$87,$F3,$5C,$7F,$E9,$A6,$C7,$F7,$07,$F9,$DB,$BD,$BF,$AE,$43
	.dc.b	$E1,$EF,$C3,$E1,$B5,$E3,$AF,$CA,$64,$F3,$3E,$6F,$99,$F3,$9B,$5B
	.dc.b	$E9,$FD,$F7,$99,$FB,$EE,$6F,$D2,$4E,$7E,$B9,$EB,$66,$BC,$7F,$11
	.dc.b	$E4,$F3,$6D,$FB,$5E,$5E,$5F,$3F,$22,$CF,$7C,$BF,$D8,$8D,$F2,$7C
	.dc.b	$07,$87,$C9,$B1,$F1,$FC,$EF,$82,$7C,$2F,$88,$B1,$F2,$FE,$23,$E2
	.dc.b	$FC,$47,$C6,$6C,$7C,$FF,$88,$F9,$27,$CA,$F3,$EC,$7D,$3F,$43,$E6
	.dc.b	$FC,$47,$CE,$6C,$75,$3E,$C6,$1D,$5A,$75,$27,$4F,$F9,$43,$B1,$87
	.dc.b	$68,$C7,$C4,$7D,$AF,$88,$FB,$6D,$BE,$D3,$C7,$0E,$CD,$8E,$BD,$FF
	.dc.b	$87,$D3,$D5,$18,$3F,$11,$F1,$1F,$CD,$BF,$13,$7A,$FF,$88,$8C,$3F
	.dc.b	$77,$C8,$FB,$DE,$3D,$BF,$A8,$FE,$07,$C7,$91,$F8,$5F,$9F,$FC,$AE
	.dc.b	$9F,$C7,$FA,$7F,$98,$4F,$E6,$FD,$6F,$CA,$9F,$96,$F5,$AC,$7E,$7F
	.dc.b	$E2,$BF,$32,$7E,$6B,$E2,$AC,$7E,$8F,$E2,$BF,$C7,$0F,$F6,$EC,$7E
	.dc.b	$54,$62,$33,$F1,$51,$FF,$7F,$8F,$5B,$15,$F4,$70,$EA,$FD,$3C,$2F
	.dc.b	$27,$EB,$F1,$E9,$87,$84,$E6,$78,$BE,$2A,$3D,$BE,$26,$E1,$FE,$8C
	.dc.b	$87,$C4,$FA,$0F,$4F,$C4,$D8,$F2,$FA,$BE,$CF,$5B,$DA,$D8,$F4,$F9
	.dc.b	$9F,$00,$F8,$3F,$11,$63,$DB,$87,$3C,$7C,$18,$DD,$7F,$96,$47,$F9
	.dc.b	$86,$98,$F5,$8F,$33,$67,$7F,$1B,$FC,$CF,$E2,$BE,$79,$F4,$3E,$2A
	.dc.b	$E1,$F3,$3D,$8F,$A7,$EC,$FD,$46,$C7,$D0,$F8,$AF,$AF,$F1,$3F,$61
	.dc.b	$B1,$F5,$3E,$2B,$ED,$9F,$73,$E2,$6C,$75,$3E,$CE,$1D,$5A,$75,$3E
	.dc.b	$ED,$F5,$69,$F8,$0F,$67,$F0,$E7,$E2,$3D,$AB,$1D,$78,$3E,$B5,$F5
	.dc.b	$6C,$75,$7E,$D5,$F5,$6C,$7F,$29,$D9,$F8,$27,$F2,$29,$FF,$65,$BF
	.dc.b	$DF,$E7,$4E,$78,$3E,$6B,$99,$DF,$8B,$FE,$93,$8F,$D9,$D7,$0F,$89
	.dc.b	$C4,$E7,$D7,$D7,$AF,$6B,$AB,$C7,$D7,$91,$E0,$F5,$EE,$5D,$A3,$C7
	.dc.b	$87,$7C,$7A,$C7,$9B,$1E,$AD,$3A,$9E,$AF,$89,$F3,$1E,$7F,$89,$B1
	.dc.b	$EE,$F8,$9F,$51,$EB,$F8,$8B,$1F,$0B,$E2,$7D,$DF,$11,$EF,$6C,$7C
	.dc.b	$6F,$88,$F8,$5E,$6F,$C3,$6C,$75,$26,$DF,$3B,$0E,$AF,$D0,$BE,$6D
	.dc.b	$9F,$A3,$87,$CD,$F8,$8F,$9D,$F1,$16,$7E,$AB,$F4,$7D,$3F,$A5,$E9
	.dc.b	$D9,$FB,$2F,$D5,$F4,$FF,$D9,$1F,$82,$7F,$1E,$EA,$FC,$0C,$63,$3C
	.dc.b	$FC,$74,$C3,$F0,$D7,$FD,$CD,$B1,$D6,$7F,$0F,$85,$EB,$FC,$2B,$DE
	.dc.b	$B8,$CD,$13,$FB,$6B,$F1,$0F,$8A,$F2,$9E,$6F,$85,$B7,$F9,$2D,$3D
	.dc.b	$AD,$3D,$A7,$E5,$B3,$F5,$34,$F0,$9E,$AF,$85,$F8,$27,$C2,$F8,$5B
	.dc.b	$1E,$EF,$85,$F8,$A7,$C6,$F8,$5B,$1F,$0B,$DF,$F9,$27,$CA,$F7,$EC
	.dc.b	$7C,$6F,$82,$F9,$A7,$CE,$F8,$4B,$1D,$49,$C3,$F0,$9F,$F2,$67,$23
	.dc.b	$85,$97,$82,$B4,$7F,$C0,$18,$65,$E0,$8C,$D7,$1C,$B1,$C5,$F3,$7E
	.dc.b	$17,$EF,$FC,$1F,$E0,$1B,$9F,$DC,$F8,$4F,$C1,$9F,$84,$F8,$3B,$1F
	.dc.b	$80,$F8,$2F,$C3,$FC,$17,$E2,$1B,$1F,$84,$F8,$0F,$C5,$9F,$8C,$F8
	.dc.b	$0B,$1D,$49,$B7,$9B,$81,$8F,$68,$C6,$1D,$1B,$5F,$90,$C7,$F2,$FF
	.dc.b	$01,$FD,$0B,$6F,$E7,$BF,$82,$7F,$69,$7F,$15,$19,$AF,$97,$39,$6F
	.dc.b	$C1,$E7,$07,$12,$BF,$B6,$57,$F2,$37,$FC,$E8,$F2,$F8,$3E,$13,$BF
	.dc.b	$83,$B4,$6F,$C1,$E1,$CF,$C0,$6B,$E0,$2E,$3D,$5E,$39,$F2,$CA,$CF
	.dc.b	$85,$E4,$F8,$FE,$02,$E9,$F7,$B3,$FE,$C3,$3E,$7F,$83,$F1,$FC,$1F
	.dc.b	$91,$B1,$EB,$C7,$B3,$4E,$C7,$BF,$4F,$5E,$3D,$B8,$3E,$6B,$EE,$F8
	.dc.b	$39,$5F,$17,$E0,$EE,$9E,$3F,$83,$F8,$67,$C4,$F8,$3B,$5F,$2F,$DE
	.dc.b	$F8,$E7,$C8,$F8,$3B,$1F,$3F,$DF,$F9,$7F,$07,$F3,$1B,$1F,$4F,$0E
	.dc.b	$78,$FC,$1C,$6E,$BF,$82,$CB,$9C,$FC,$5C,$AC,$FE,$1B,$E0,$FE,$BF
	.dc.b	$C1,$FD,$86,$E5,$F7,$FE,$0F,$ED,$9F,$7B,$D2,$FE,$BD,$1F,$8A,$8C
	.dc.b	$7C,$4C,$8F,$ED,$13,$FD,$7B,$2F,$D8,$BF,$D8,$F1,$FF,$08,$AF,$7F
	.dc.b	$A7,$BF,$D5,$B7,$E1,$25,$F2,$F8,$8F,$1C,$76,$C7,$C1,$3D,$BE,$06
	.dc.b	$E5,$DB,$E1,$3D,$07,$A7,$E1,$2E,$1E,$6F,$84,$F6,$1E,$DF,$84,$B1
	.dc.b	$EA,$F8,$4F,$81,$F0,$9F,$05,$B1,$EE,$F8,$2F,$89,$F0,$5F,$15,$B1
	.dc.b	$D4,$9A,$FC,$2C,$3A,$BF,$12,$F9,$B6,$7E,$4E,$1F,$3F,$E0,$BE,$87
	.dc.b	$C1,$D9,$F9,$AF,$2D,$BE,$5E,$9E,$99,$EF,$1F,$3B,$1E,$AD,$8E,$A7
	.dc.b	$D6,$F8,$4F,$B9,$F0,$9F,$75,$B1,$F6,$BE,$13,$F0,$1F,$03,$F8,$16
	.dc.b	$C7,$DE,$F8,$4F,$C2,$1F,$85,$F8,$2B,$1F,$E2,$9F,$C0,$FF,$B5,$13
	.dc.b	$AF,$F8,$3B,$FD,$B5,$5F,$4E,$1F,$C3,$B7,$4F,$C3,$E3,$D2,$7F,$17
	.dc.b	$8D,$9F,$F2,$DE,$1F,$DC,$67,$F9,$FB,$FD,$9F,$1F,$17,$C6,$78,$3E
	.dc.b	$1B,$C3,$F0,$D2,$7C,$DF,$0D,$70,$E3,$F0,$DE,$6F,$86,$F3,$B6,$BD
	.dc.b	$9F,$09,$EA,$F8,$4F,$5B,$63,$E0,$63,$CF,$5F,$2E,$56,$7F,$AD,$EA
	.dc.b	$C7,$A6,$7E,$09,$F9,$19,$FC,$6F,$84,$91,$F2,$FE,$1A,$D7,$BB,$E1
	.dc.b	$BE,$59,$F3,$3E,$16,$D7,$D1,$F8,$6F,$9E,$7D,$0F,$86,$B1,$F5,$7E
	.dc.b	$0F,$E9,$FC,$1F,$D4,$6C,$7D,$9C,$39,$E3,$F3,$31,$BC,$1F,$43,$2E
	.dc.b	$73,$F4,$32,$B3,$F5,$3E,$0F,$EF,$9F,$80,$F8,$6B,$97,$52,$5F,$56
	.dc.b	$D7,$57,$D4,$7E,$2E,$FA,$BF,$2B,$E1,$BF,$16,$7E,$33,$E1,$6C,$7F
	.dc.b	$9E,$7E,$97,$FA,$F6,$1F,$D7,$1B,$F0,$C7,$FA,$DC,$66,$7F,$AD,$C7
	.dc.b	$F6,$08,$FF,$04,$B7,$1F,$F5,$4B,$FD,$76,$7B,$7C,$1F,$0F,$86,$E3
	.dc.b	$F0,$D6,$7C,$4F,$46,$1E,$5F,$82,$B8,$F6,$BF,$09,$E2,$F8,$2B,$5E
	.dc.b	$BF,$86,$F2,$9E,$6F,$83,$B1,$EF,$F8,$4F,$4F,$C2,$FA,$9B,$1F,$0F
	.dc.b	$E1,$7D,$A7,$BB,$E1,$6C,$7C,$7F,$83,$F8,$3F,$07,$F0,$9B,$1D,$4F
	.dc.b	$99,$87,$56,$9D,$49,$AF,$C8,$BE,$AF,$CA,$F8,$3F,$9B,$F0,$7F,$39
	.dc.b	$B4,$9F,$F1,$DF,$BA,$FA,$D7,$27,$F8,$56,$30,$D8,$FF,$78,$D7,$1E
	.dc.b	$6F,$F7,$1D,$6A,$30,$75,$7A,$70,$F1,$E1,$E4,$D9,$FC,$47,$C2,$FE
	.dc.b	$08,$FC,$1F,$C3,$58,$FC,$67,$C2,$FE,$1B,$E1,$BF,$0E,$D8,$FC,$87
	.dc.b	$C3,$7E,$28,$FF,$12,$FE,$07,$FD,$A3,$7F,$C9,$46,$34,$CF,$C2,$5B
	.dc.b	$FF,$68,$5F,$F6,$9D,$BF,$1F,$8F,$5C,$7F,$26,$D8,$FC,$9E,$3D,$EF
	.dc.b	$8F,$C2,$DB,$FF,$30,$E1,$FD,$BE,$7A,$FC,$3F,$93,$E1,$FC,$AF,$AA
	.dc.b	$3F,$51,$AD,$BF,$D0,$BC,$F2,$4F,$F9,$D3,$07,$C3,$E6,$30,$FC,$0F
	.dc.b	$87,$91,$EC,$F8,$1B,$8F,$DE,$E5,$6F,$F6,$F3,$FB,$8D,$BF,$B3,$4E
	.dc.b	$1F,$C2,$70,$7E,$0F,$EB,$BE,$64,$62,$B6,$A3,$E8,$78,$AD,$9F,$0D
	.dc.b	$F5,$0F,$AB,$F0,$77,$0E,$37,$F1,$F2,$FC,$9B,$78,$7F,$1A,$B8,$BC
	.dc.b	$5C,$7F,$95,$FA,$EF,$E7,$98,$78,$9B,$87,$2E,$4D,$7A,$DF,$84,$F8
	.dc.b	$7F,$C2,$FC,$45,$C3,$F0,$2F,$F9,$CF,$47,$6B,$C3,$C2,$F1,$2C,$61
	.dc.b	$FC,$7F,$C4,$FE,$43,$E2,$6E,$3F,$8C,$7F,$27,$F1,$3F,$94,$F8,$8B
	.dc.b	$3F,$90,$7A,$F8,$72,$3E,$8E,$7E,$2C,$F1,$C1,$F8,$CF,$89,$FC,$E9
	.dc.b	$F9,$EF,$87,$B7,$F9,$AF,$89,$FD,$09,$FA,$2F,$88,$B1,$D3,$0F,$E6
	.dc.b	$B8,$FB,$BD,$1F,$6C,$FB,$9F,$D4,$FC,$47,$FA,$3B,$BF,$FD,$E9,$DC
	.dc.b	$8C,$6B,$9F,$89,$D7,$C4,$DC,$7F,$13,$E0,$FE,$26,$3E,$67,$D7,$7F
	.dc.b	$AB,$A7,$DA,$DB,$FF,$65,$E8,$87,$C0,$F0,$F8,$1F,$17,$C5,$78,$FE
	.dc.b	$22,$E3,$DB,$74,$79,$FE,$23,$D1,$E0,$5A,$D3,$E2,$F8,$69,$1E,$2F
	.dc.b	$86,$BB,$7F,$B7,$3D,$E7,$BF,$F9,$8A,$FD,$1C,$63,$D0,$FF,$5C,$3A
	.dc.b	$F3,$2D,$27,$E4,$7C,$3C,$BF,$37,$C4,$DC,$7D,$FC,$7E,$89,$F8,$7B
	.dc.b	$7C,$6D,$FE,$3B,$D3,$C7,$93,$E2,$CD,$E2,$E3,$1F,$44,$3C,$5F,$19
	.dc.b	$E4,$F1,$6D,$FE,$33,$EB,$5E,$0F,$E7,$DF,$B9,$19,$8C,$77,$7B,$3C
	.dc.b	$69,$3E,$1F,$16,$E7,$E0,$F1,$7E,$09,$F0,$BC,$5B,$5D,$F4,$EC,$77
	.dc.b	$D2,$C7,$AF,$3F,$93,$9F,$CA,$6C,$7D,$0F,$A0,$FE,$AC,$DC,$BE,$8C
	.dc.b	$3C,$3B,$1E,$BE,$89,$5F,$AA,$F4,$34,$FC,$18,$E3,$E1,$F2,$F0,$2C
	.dc.b	$F7,$7A,$F3,$F6,$E7,$B3,$E3,$7F,$83,$60,$E9,$F1,$F7,$E1,$B5,$E9
	.dc.b	$EB,$F3,$9E,$88,$B1,$ED,$F8,$CF,$59,$EC,$F0,$2C,$7C,$1F,$03,$DF
	.dc.b	$F1,$7F,$01,$B1,$F1,$7E,$2F,$E1,$9F,$13,$E2,$EC,$7C,$9F,$89,$F8
	.dc.b	$FE,$37,$C8,$6C,$7C,$DF,$1F,$E5,$9F,$33,$E2,$EC,$7D,$1F,$8B,$F9
	.dc.b	$FF,$17,$F4,$1B,$1F,$57,$C4,$FA,$67,$E9,$3E,$83,$FC,$83,$E5,$3B
	.dc.b	$7F,$57,$5F,$AA,$9F,$F2,$0C,$9A,$7F,$8F,$E3,$FC,$29,$19,$F3,$FA
	.dc.b	$7C,$57,$5F,$84,$B5,$CD,$F0,$7C,$67,$87,$D4,$B3,$DD,$F2,$7A,$92
	.dc.b	$3C,$5F,$17,$7C,$3F,$E0,$1A,$3E,$2B,$D4,$7A,$FE,$2E,$DF,$A7,$E2
	.dc.b	$FD,$C7,$BF,$DA,$B1,$ED,$F6,$BE,$17,$C5,$7C,$36,$C7,$C1,$F8,$BF
	.dc.b	$8D,$F1,$7F,$1D,$B1,$F1,$7E,$2F,$E5,$1F,$2F,$E1,$AC,$7C,$9F,$87
	.dc.b	$F9,$DF,$0F,$F3,$DB,$1F,$C5,$BE,$7F,$FC,$41,$D2,$7F,$BD,$46,$7E
	.dc.b	$2F,$F8,$CC,$63,$8F,$F9,$02,$F8,$CF,$F2,$24,$61,$BF,$05,$FE,$12
	.dc.b	$8F,$F3,$D4,$ED,$F0,$7C,$46,$B0,$FB,$33,$FD,$DE,$7C,$9F,$0F,$97
	.dc.b	$8B,$72,$F3,$7C,$4D,$A7,$C6,$5F,$85,$98,$7B,$63,$5E,$C6,$FD,$49
	.dc.b	$3F,$06,$39,$7C,$4C,$8F,$88,$F5,$F8,$5E,$DF,$0B,$27,$E4,$3E,$1F
	.dc.b	$80,$F1,$7C,$04,$9F,$98,$F9,$7E,$13,$CD,$F0,$9E,$7F,$84,$95,$F4
	.dc.b	$63,$D5,$F0,$9E,$BF,$83,$93,$F5,$63,$DD,$F0,$9F,$0F,$E2,$7E,$7C
	.dc.b	$FF,$8C,$1A,$FD,$48,$D3,$FC,$7C,$D0,$F8,$4C,$9A,$F8,$5D,$FC,$2C
	.dc.b	$9D,$BC,$BE,$17,$9F,$C2,$C9,$E6,$F6,$F8,$5E,$FF,$0B,$E0,$F8,$59
	.dc.b	$5E,$08,$F1,$FC,$2C,$8F,$1B,$E6,$F8,$5F,$3F,$C2,$7A,$3E,$16,$57
	.dc.b	$A2,$3D,$7F,$0B,$23,$D6,$FB,$BE,$17,$DF,$F0,$9F,$03,$E1,$65,$7C
	.dc.b	$08,$F8,$7F,$0B,$23,$E1,$BF,$1B,$E1,$7E,$3F,$C2,$49,$F8,$EF,$CA
	.dc.b	$F8,$5F,$97,$F0,$BF,$47,$E7,$7F,$39,$4F,$3F,$F3,$DD,$69,$CC,$62
	.dc.b	$7F,$C0,$A7,$D7,$87,$C2,$F3,$F8,$E9,$39,$8F,$D0,$53,$EA,$9F,$0F
	.dc.b	$8F,$81,$7C,$35,$37,$3A,$C3,$B5,$79,$C8,$F8,$7E,$1F,$D5,$5E,$BF
	.dc.b	$87,$93,$E5,$8F,$0F,$D5,$38,$AF,$14,$AF,$F2,$7C,$E5,$F6,$7C,$47
	.dc.b	$C5,$23,$EF,$F8,$8F,$91,$AC,$8F,$84,$FE,$F7,$E3,$7F,$D1,$1D,$DE
	.dc.b	$0E,$DF,$0F,$64,$BE,$F1,$F4,$7E,$1F,$E9,$7C,$3C,$9F,$99,$1C,$FE
	.dc.b	$35,$E0,$D2,$4F,$F9,$07,$C1,$F6,$F8,$9E,$EF,$0E,$4F,$B2,$3E,$F7
	.dc.b	$C2,$FD,$FF,$86,$93,$F6,$63,$B6,$92,$3F,$C9,$8F,$87,$5F,$93,$10
	.dc.b	$F8,$E7,$CC,$F2,$BF,$15,$F0,$9F,$8B,$F8,$49,$3F,$85,$8F,$57,$91
	.dc.b	$23,$FC,$B0,$FB,$7C,$7F,$C2,$EA,$F4,$E9,$2B,$FD,$07,$0F,$4B,$E2
	.dc.b	$79,$52,$7F,$CC,$4F,$C7,$F2,$BE,$47,$74,$9F,$F3,$33,$F2,$FB,$BB
	.dc.b	$FB,$5E,$0F,$66,$57,$68,$FA,$1E,$64,$8F,$F3,$73,$F4,$FC,$CF,$BD
	.dc.b	$DD,$FA,$4F,$81,$52,$27,$95,$A3,$73,$EE,$F8,$29,$5E,$C7,$E0,$FC
	.dc.b	$07,$4F,$37,$AF,$97,$2B,$FD,$17,$8A,$7C,$3E,$5F,$8B,$CA,$95,$A8
	.dc.b	$FD,$7F,$D4,$FF,$B7,$A3,$FD,$56,$BA,$F2,$6B,$E1,$BD,$7E,$6C,$AE
	.dc.b	$EF,$BB,$CC,$F7,$F9,$9F,$03,$CC,$95,$E4,$8F,$EB,$9E,$4B,$EA,$7F
	.dc.b	$D9,$33,$FE,$E4,$6B,$FC,$37,$8F,$E3,$CE,$F0,$E3,$4B,$BF,$17,$30
	.dc.b	$E4,$F6,$FA,$97,$4F,$D2,$89,$3F,$19,$FA,$BE,$6F,$4F,$8B,$EB,$F1
	.dc.b	$92,$BC,$91,$E0,$F2,$FC,$3E,$64,$9F,$44,$7D,$EF,$43,$EF,$D4,$9F
	.dc.b	$A7,$1F,$82,$B9,$1F,$59,$FC,$2D,$7E,$1A,$FF,$0F,$F1,$92,$BE,$DC
	.dc.b	$7E,$2B,$92,$47,$DE,$7F,$1B,$CB,$D2,$3E,$41,$F8,$DE,$A6,$39,$F5
	.dc.b	$7F,$C2,$91,$F9,$1C,$BC,$15,$F8,$8C,$BC,$0F,$29,$FC,$B6,$5F,$9A
	.dc.b	$F3,$3F,$8C,$7E,$04,$F8,$DA,$66,$FF,$3B,$97,$F8,$87,$9D,$F5,$5F
	.dc.b	$4A,$73,$D3,$F7,$23,$11,$E8,$9D,$F2,$FE,$1A,$5E,$35,$F8,$7C,$7C
	.dc.b	$3C,$C3,$98,$C3,$BC,$F9,$9F,$3F,$C7,$49,$FF,$77,$EB,$B6,$F8,$DE
	.dc.b	$C2,$3A,$F1,$BD,$FF,$1D,$23,$8B,$F0,$BE,$2F,$E1,$FC,$64,$9E,$AF
	.dc.b	$C6,$F8,$CE,$71,$F2,$F1,$63,$8E,$2F,$A7,$1F,$E4,$19,$D6,$3D,$DA
	.dc.b	$7A,$31,$FA,$3C,$32,$3D,$6F,$D4,$E0,$FA,$BB,$FD,$6E,$09,$5F,$02
	.dc.b	$3E,$CF,$C6,$7D,$AF,$8C,$93,$F1,$23,$EE,$FA,$DF,$7B,$D5,$93,$F2
	.dc.b	$23,$AB,$F2,$B0,$EA,$FB,$70,$EA,$FC,$EC,$3A,$BF,$07,$0F,$C4,$FC
	.dc.b	$64,$8F,$A6,$FE,$33,$E2,$FA,$C7,$C7,$C3,$AF,$07,$C6,$C7,$AC,$FC
	.dc.b	$BC,$7A,$CF,$CA,$C7,$F2,$BF,$19,$F9,$6F,$8C,$97,$F8,$08,$FE,$E1
	.dc.b	$F8,$17,$F9,$2A,$FF,$46,$FF,$91,$E7,$F4,$0D,$8C,$C7,$0F,$8D,$8F
	.dc.b	$E4,$3C,$05,$7F,$93,$63,$0B,$E3,$7B,$FC,$6C,$9C,$BE,$2F,$8D,$F1
	.dc.b	$C7,$93,$E0,$A3,$C8,$EE,$79,$31,$B9,$7A,$30,$F5,$FC,$2C,$AF,$0B
	.dc.b	$EE,$F8,$1F,$7F,$C6,$FC,$0F,$89,$95,$E6,$8F,$87,$F1,$3F,$13,$E2
	.dc.b	$64,$FA,$A3,$E3,$FC,$4C,$8F,$6B,$F2,$BE,$27,$AC,$7C,$0C,$3A,$BF
	.dc.b	$2B,$0E,$AF,$C4,$C3,$AB,$F3,$B0,$FA,$7F,$13,$F5,$3E,$26,$4F,$CA
	.dc.b	$8F,$AF,$F1,$32,$3E,$6B,$F6,$BE,$27,$ED,$FC,$4F,$DC,$F8,$99,$5F
	.dc.b	$4A,$3E,$FF,$C4,$FE,$03,$E2,$64,$FD,$68,$E8,$FE,$03,$0B,$C9,$F8
	.dc.b	$4C,$7A,$CF,$DC,$C7,$AB,$F8,$6C,$7F,$C7,$5F,$13,$FD,$F6,$3F,$BD
	.dc.b	$B8,$33,$E9,$74,$3F,$C0,$3D,$9F,$88,$CB,$15,$FE,$03,$EC,$65,$E0
	.dc.b	$8F,$C9,$E5,$E0,$CF,$F2,$F9,$F8,$3E,$22,$57,$E4,$F4,$FF,$00,$37
	.dc.b	$E2,$7C,$BF,$11,$E6,$F8,$99,$3B,$8F,$4F,$C4,$48,$E4,$FB,$3E,$1F
	.dc.b	$DB,$F0,$D2,$7B,$3F,$03,$E2,$3A,$C7,$87,$0E,$AF,$F4,$9C,$98,$F5
	.dc.b	$8C,$63,$75,$FE,$A7,$5F,$E0,$06,$FC,$37,$CC,$F8,$69,$7E,$D8,$F9
	.dc.b	$FF,$0B,$F4,$3E,$1A,$4F,$C1,$8F,$A7,$F0,$9F,$53,$E1,$24,$FC,$58
	.dc.b	$FA,$FF,$07,$23,$E4,$3F,$6B,$E1,$3A,$C7,$CB,$C3,$AB,$EB,$C3,$AB
	.dc.b	$F3,$F0,$EA,$FB,$F0,$FC,$1F,$C3,$48,$FA,$8F,$E1,$BE,$1B,$F0,$FF
	.dc.b	$0D,$27,$EC,$3F,$E8,$3F,$E0,$1F,$ED,$91,$FC,$35,$FE,$AC,$FF,$73
	.dc.b	$5F,$09,$B3,$87,$C2,$F1,$F8,$29,$5C,$5E,$9F,$0B,$DE,$3E,$A6,$17
	.dc.b	$8F,$FC,$07,$8F,$82,$7B,$61,$D5,$F1,$63,$D9,$F2,$7C,$17,$9C,$8F
	.dc.b	$9F,$E0,$BA,$78,$BF,$9E,$E7,$67,$8A,$FF,$30,$65,$DF,$3F,$0E,$7D
	.dc.b	$E7,$DD,$9F,$77,$E0,$7C,$2F,$C5,$23,$F1,$7E,$17,$E4,$7C,$27,$C9
	.dc.b	$F8,$49,$3F,$26,$3E,$67,$BF,$F3,$7E,$02,$4F,$CD,$8F,$A1,$EF,$FD
	.dc.b	$1F,$80,$93,$F4,$63,$AB,$F4,$F0,$EA,$FD,$3C,$3A,$BF,$5F,$0E,$AF
	.dc.b	$D7,$C3,$EE,$7C,$1C,$8F,$B8,$FD,$FF,$84,$FC,$07,$C2,$7E,$07,$E1
	.dc.b	$25,$7E,$06,$3F,$09,$F0,$72,$3F,$08,$FE,$1F,$E1,$3F,$D9,$5F,$C0
	.dc.b	$3F,$D6,$FA,$3F,$86,$B6,$BF,$D1,$CD,$F5,$0F,$F5,$D9,$FE,$27,$CE
	.dc.b	$C7,$C3,$7F,$8F,$CB,$CA,$F8,$08,$7E,$4F,$2F,$13,$FD,$8F,$0F,$EC
	.dc.b	$4F,$F9,$25,$FE,$C4,$E7,$E0,$7C,$FF,$05,$E8,$F8,$5F,$4F,$C2,$CA
	.dc.b	$E5,$1E,$CF,$84,$91,$D5,$F7,$FC,$27,$C0,$F8,$19,$3E,$17,$E1,$FC
	.dc.b	$17,$58,$F2,$61,$D5,$F2,$61,$D1,$F3,$61,$7B,$FF,$CC,$BA,$32,$F9
	.dc.b	$BF,$03,$F3,$BE,$0A,$61,$EF,$8E,$CF,$BF,$2E,$D8,$72,$CB,$A7,$1F
	.dc.b	$D4,$CE,$DF,$D4,$CB,$EC,$FC,$24,$8F,$94,$FD,$CF,$84,$FB,$BF,$07
	.dc.b	$F7,$BE,$12,$57,$CF,$8F,$C0,$FC,$17,$E0,$BE,$06,$4F,$D3,$8F,$C2
	.dc.b	$FC,$04,$8F,$AC,$FE,$23,$E0,$7A,$C7,$D9,$C3,$AB,$F6,$70,$EA,$FD
	.dc.b	$DC,$3A,$BF,$77,$0F,$C8,$FC,$1F,$E4,$BE,$12,$4F,$E0,$E3,$FD,$BE
	.dc.b	$FF,$A2,$FF,$80,$7F,$B3,$77,$9A,$4F,$3B,$7C,$9F,$E9,$68,$FE,$B7
	.dc.b	$5C,$7E,$0F,$3C,$9F,$DA,$A5,$18,$9F,$F7,$64,$70,$AE,$5E,$FE,$71
	.dc.b	$F0,$7C,$05,$A3,$67,$18,$CC,$3F,$C7,$D5,$E5,$6F,$C2,$71,$CA,$46
	.dc.b	$0D,$FC,$0F,$5F,$84,$91,$83,$9F,$C2,$78,$7E,$12,$46,$0E,$FF,$07
	.dc.b	$E5,$F8,$49,$18,$3C,$7F,$03,$E9,$23,$E6,$F8,$2F,$66,$1D,$23,$E8
	.dc.b	$61,$87,$91,$F4,$EF,$AB,$F5,$70,$F8,$9F,$03,$F1,$7E,$0A,$4F,$C2
	.dc.b	$8E,$4F,$43,$E2,$DE,$39,$3E,$C5,$F4,$7E,$D6,$1F,$3E,$3E,$87,$C2
	.dc.b	$4B,$F9,$AF,$D3,$F8,$4F,$A9,$F0,$92,$7E,$8B,$F5,$FE,$13,$EC,$7C
	.dc.b	$04,$9F,$AA,$FD,$DF,$3B,$FA,$67,$3F,$F8,$B9,$FE,$D3,$A7,$FB,$81
	.dc.b	$B1,$9F,$80,$E1,$F0,$1D,$F0,$FC,$B6,$57,$0F,$C9,$65,$DE,$7F,$35
	.dc.b	$95,$9E,$9F,$13,$E1,$F8,$59,$5F,$E8,$5D,$3F,$D7,$ED,$F8,$5F,$37
	.dc.b	$C2,$C8,$D3,$E9,$F8,$5F,$57,$C2,$C9,$E4,$FB,$7E,$17,$DD,$F0,$52
	.dc.b	$7B,$3F,$07,$E0,$FA,$47,$FB,$E7,$26,$38,$D7,$E1,$65,$D6,$B5,$95
	.dc.b	$9F,$8D,$97,$CB,$F8,$2F,$99,$F0,$72,$FD,$91,$F3,$FE,$13,$E8,$7C
	.dc.b	$2C,$9F,$81,$1F,$4F,$E1,$64,$7C,$37,$EB,$7C,$2F,$D7,$F8,$39,$3F
	.dc.b	$1D,$FB,$5F,$07,$D6,$3E,$56,$1D,$5F,$66,$1D,$1F,$76,$17,$4F,$BD
	.dc.b	$8F,$E0,$FE,$0F,$F0,$9F,$0B,$2F,$EA,$47,$57,$E3,$63,$D6,$FF,$0D
	.dc.b	$8F,$3E,$DF,$C5,$65,$6D,$E7,$B9,$1A,$7F,$0D,$FE,$01,$FE,$DC,$4A
	.dc.b	$C7,$C4,$66,$A4,$7F,$B2,$67,$FB,$3E,$6C,$BF,$D0,$B1,$FD,$A2,$53
	.dc.b	$97,$C3,$0F,$7F,$C1,$F0,$12,$7A,$47,$8F,$E0,$3C,$9F,$01,$27,$C1
	.dc.b	$1E,$27,$AE,$3E,$0C,$B1,$97,$7D,$59,$F8,$E3,$0F,$86,$FD,$59,$F8
	.dc.b	$5E,$67,$9B,$3F,$89,$F0,$BE,$9B,$E3,$F0,$B7,$3F,$7C,$DB,$F6,$CD
	.dc.b	$CB,$C3,$F0,$77,$37,$C2,$58,$C3,$9F,$84,$D7,$C2,$5C,$FE,$73,$C7
	.dc.b	$E0,$64,$7D,$1F,$82,$BA,$7A,$33,$EF,$3F,$57,$1E,$CD,$3E,$C6,$3D
	.dc.b	$9A,$76,$3E,$E7,$C0,$F9,$CF,$47,$C1,$58,$FC,$07,$C2,$7A,$CF,$67
	.dc.b	$C2,$58,$FC,$27,$C1,$FB,$FE,$13,$E0,$36,$1A,$79,$FF,$80,$7F,$11
	.dc.b	$18,$E5,$FB,$10,$F8,$5F,$94,$EA,$B8,$47,$77,$E0,$72,$33,$F1,$E1
	.dc.b	$E6,$3C,$67,$E5,$F2,$F1,$B4,$F0,$1F,$ED,$CC,$1F,$05,$E0,$3C,$3F
	.dc.b	$05,$67,$A6,$3E,$0C,$3B,$B6,$3B,$BE,$4C,$3B,$B6,$3B,$BF,$B2,$C3
	.dc.b	$FD,$9B,$18,$E2,$46,$27,$E3,$7C,$2F,$C2,$F8,$5F,$87,$F0,$93,$0E
	.dc.b	$16,$F8,$4C,$61,$F3,$7E,$0B,$E4,$FC,$07,$CA,$6D,$F0,$F7,$FE,$6F
	.dc.b	$C0,$7C,$E6,$C7,$4C,$B9,$F2,$7D,$4C,$AC,$FC,$7C,$39,$E5,$DE,$3C
	.dc.b	$79,$7D,$8F,$84,$91,$E6,$F8,$4B,$5F,$81,$F8,$3F,$BB,$F0,$9F,$79
	.dc.b	$B5,$EC,$F8,$4F,$C0,$9F,$82,$F8,$4B,$1F,$03,$E0,$BF,$0A,$7E,$1B
	.dc.b	$E0,$AC,$7C,$4B,$E7,$8F,$78,$F8,$F8,$F2,$8F,$C9,$63,$D6,$7C,$07
	.dc.b	$CC,$F8,$2F,$C8,$9F,$BB,$FE,$01,$7C,$1E,$27,$EC,$43,$E1,$74,$6F
	.dc.b	$E1,$6C,$EB,$E1,$79,$7C,$2F,$36,$C7,$2F,$83,$EC,$77,$F8,$3B,$1D
	.dc.b	$B0,$E7,$AF,$EF,$F2,$FF,$75,$CE,$4F,$1E,$3D,$27,$58,$DE,$8F,$E4
	.dc.b	$96,$F8,$3F,$61,$ED,$F8,$3B,$8F,$3C,$3B,$65,$EB,$8C,$61,$CB,$2E
	.dc.b	$91,$FB,$CF,$05,$93,$3C,$E3,$FD,$42,$1F,$0B,$F2,$CF,$99,$F0,$B7
	.dc.b	$0F,$97,$F0,$9F,$3F,$E1,$7E,$83,$63,$E7,$FC,$07,$D3,$F8,$1F,$A8
	.dc.b	$D8,$FA,$7F,$03,$F5,$FE,$07,$EC,$36,$3A,$9F,$63,$0E,$AD,$3A,$9F
	.dc.b	$72,$FA,$B4,$FB,$FF,$0B,$F8,$3F,$85,$FC,$23,$63,$F0,$7F,$09,$F8
	.dc.b	$7F,$85,$FC,$43,$63,$F7,$DD,$5F,$C8,$9B,$C3,$FD,$1E,$1F,$0B,$63
	.dc.b	$1F,$07,$9F,$88,$B7,$C9,$EA,$FE,$4B,$1B,$5F,$85,$79,$47,$63,$C3
	.dc.b	$78,$9F,$E9,$18,$7F,$73,$9F,$2F,$C2,$78,$FE,$13,$C8,$F9,$7E,$16
	.dc.b	$4F,$AB,$E1,$AE,$1C,$BE,$1B,$D5,$F0,$DE,$B6,$D7,$C0,$F8,$6F,$77
	.dc.b	$C1,$FB,$DB,$1F,$13,$1E,$7E,$07,$F7,$9D,$7F,$DB,$53,$93,$E4,$E1
	.dc.b	$D3,$2F,$04,$7C,$CC,$BE,$57,$C1,$FC,$BF,$86,$B3,$F4,$1E,$AF,$E5
	.dc.b	$F0,$65,$F4,$75,$CE,$3F,$E2,$69,$65,$DF,$3F,$23,$70,$E5,$F0,$DF
	.dc.b	$64,$FB,$5F,$0D,$63,$EF,$FC,$37,$DD,$F8,$6F,$BC,$D8,$FC,$1F,$C0
	.dc.b	$7E,$04,$FC,$17,$C3,$58,$FC,$3F,$C3,$7E,$14,$FC,$37,$C1,$58,$EA
	.dc.b	$7E,$33,$0E,$AD,$3A,$93,$6F,$BF,$7D,$5F,$F0,$47,$F0,$0F,$F6,$DD
	.dc.b	$BF,$C5,$0F,$F7,$1A,$FE,$A8,$E5,$FD,$F3,$99,$E3,$F0,$BF,$E1,$F8
	.dc.b	$FE,$D3,$3C,$3E,$17,$13,$9F,$84,$93,$D3,$E1,$6E,$1E,$2F,$82,$E4
	.dc.b	$73,$F8,$2B,$5E,$2C,$39,$BD,$4C,$1E,$6C,$39,$E7,$EA,$C6,$CF,$A3
	.dc.b	$E0,$BC,$C7,$9F,$E0,$AC,$7B,$BE,$0F,$D4,$7A,$FE,$12,$C7,$C2,$F8
	.dc.b	$4F,$71,$EF,$F8,$4B,$1F,$1B,$E0,$7E,$17,$C0,$FC,$36,$C7,$52,$70
	.dc.b	$FD,$0C,$3A,$BF,$4B,$0E,$6F,$D4,$C2,$EB,$F5,$3E,$07,$E7,$7C,$27
	.dc.b	$CF,$6E,$3D,$27,$EC,$E9,$E9,$8C,$1F,$63,$1E,$6F,$DA,$C2,$E3,$FE
	.dc.b	$84,$E5,$7C,$27,$DB,$3E,$E7,$C2,$5B,$FC,$1F,$C2,$7D,$F3,$FC,$BD
	.dc.b	$FC,$03,$FD,$92,$32,$7F,$86,$DF,$EC,$8D,$F0,$F7,$F0,$1C,$1E,$3E
	.dc.b	$F7,$2F,$80,$B5,$C5,$EC,$F4,$D3,$FC,$BD,$87,$77,$F3,$38,$F4,$9E
	.dc.b	$C7,$87,$1E,$CF,$8F,$E1,$3C,$FF,$09,$E8,$6C,$79,$FE,$13,$D6,$7B
	.dc.b	$3E,$12,$C7,$AF,$E0,$BD,$FF,$05,$F0,$1B,$1E,$FF,$82,$F8,$7E,$FF
	.dc.b	$C4,$6C,$75,$26,$BE,$AC,$3A,$BE,$CB,$E6,$D9,$F9,$58,$7C,$EF,$82
	.dc.b	$91,$F3,$7E,$02,$D7,$C4,$F8,$1F,$A6,$7D,$4F,$82,$B5,$F4,$FE,$07
	.dc.b	$EB,$FC,$17,$D8,$6C,$7D,$7F,$82,$FB,$7F,$01,$F7,$1B,$1F,$6F,$0E
	.dc.b	$78,$FD,$2C,$6F,$81,$FE,$AF,$E7,$65,$D2,$BE,$BB,$67,$FD,$69,$FC
	.dc.b	$03,$FD,$86,$7F,$D3,$0D,$8C,$FC,$0F,$FA,$A6,$31,$97,$F8,$37,$06
	.dc.b	$5E,$08,$C6,$57,$1F,$F6,$ED,$78,$A7,$F2,$5A,$FB,$DA,$7F,$9B,$F0
	.dc.b	$FF,$71,$37,$E0,$BC,$1F,$07,$E1,$7C,$5F,$07,$27,$87,$C1,$DC,$BD
	.dc.b	$3E,$F7,$9F,$DE,$F4,$36,$BB,$7B,$DE,$BF,$81,$F6,$36,$3C,$58,$F3
	.dc.b	$E2,$F8,$B9,$59,$FF,$16,$6B,$FD,$59,$F4,$65,$D9,$A7,$AF,$E0,$BE
	.dc.b	$47,$C1,$FC,$96,$C7,$BF,$E0,$FE,$61,$F3,$7E,$0E,$C7,$C3,$F8,$3F
	.dc.b	$A0,$7D,$1F,$83,$B1,$F1,$FE,$0F,$EA,$7C,$07,$D5,$6C,$75,$27,$1F
	.dc.b	$DC,$C3,$AB,$F7,$B0,$E6,$FE,$03,$0B,$BF,$E1,$3E,$0F,$F0,$07,$E0
	.dc.b	$7E,$06,$E3,$F5,$FE,$0F,$F0,$87,$E1,$7E,$0E,$C7,$FA,$A7,$E0,$3F
	.dc.b	$AB,$64,$F8,$5F,$F5,$DF,$C4,$FF,$BC,$E7,$31,$8A,$E1,$19,$35,$F0
	.dc.b	$B6,$3C,$04,$BF,$CB,$E1,$E0,$7F,$DE,$15,$FD,$8F,$45,$7F,$98,$E3
	.dc.b	$0D,$3C,$58,$F8,$FE,$0E,$47,$A7,$E0,$EE,$1F,$A4,$D3,$B6,$3F,$A8
	.dc.b	$CA,$DF,$5C,$79,$BF,$04,$B7,$E4,$F8,$5F,$81,$F0,$BF,$05,$B3,$F1
	.dc.b	$FE,$17,$E2,$7C,$2F,$C5,$6C,$7C,$BF,$80,$F9,$1F,$07,$F2,$5B,$1F
	.dc.b	$3F,$E0,$7E,$67,$C0,$FC,$D6,$C7,$53,$EA,$61,$D5,$A7,$53,$EC,$5F
	.dc.b	$56,$9F,$6F,$E1,$3E,$C1,$F6,$7E,$16,$C7,$DF,$F8,$4F,$B9,$F0,$BF
	.dc.b	$75,$B1,$F8,$3F,$84,$FC,$01,$F8,$1F,$84,$B1,$F8,$7F,$81,$FA,$FC
	.dc.b	$4E,$5B,$D6,$FC,$63,$FA,$74,$63,$AD,$7C,$23,$73,$AC,$3F,$01,$9F
	.dc.b	$8B,$1F,$C2,$B6,$B9,$67,$C6,$BF,$D5,$17,$FD,$9A,$FF,$17,$39,$7F
	.dc.b	$35,$19,$AE,$FF,$0A,$99,$1E,$2F,$85,$C4,$7A,$F0,$EB,$18,$9E,$33
	.dc.b	$C2,$7C,$3F,$0B,$97,$C7,$F0,$76,$7C,$2D,$DB,$E2,$55,$C9,$F0,$94
	.dc.b	$97,$D4,$F4,$FC,$26,$FE,$0A,$46,$70,$F9,$D1,$CF,$E0,$BA,$63,$D7
	.dc.b	$29,$7B,$C9,$3E,$1C,$7B,$CF,$BC,$F8,$F9,$65,$E1,$F0,$5E,$72,$3C
	.dc.b	$FE,$0F,$BB,$D7,$1E,$EF,$3D,$3D,$EF,$87,$0E,$FB,$7E,$03,$1E,$D3
	.dc.b	$BC,$DF,$09,$F1,$49,$1E,$9F,$84,$F9,$1F,$09,$23,$D8,$FC,$BF,$80
	.dc.b	$F9,$9F,$01,$F3,$7E,$0A,$57,$C1,$8F,$A1,$F0,$1F,$47,$E0,$E4,$FC
	.dc.b	$58,$E8,$F8,$F0,$BA,$FE,$5B,$1E,$B3,$F2,$F1,$EA,$FA,$71,$FB,$9F
	.dc.b	$09,$F7,$FF,$18,$FC,$5C,$FC,$36,$E1,$FA,$09,$D7,$C2,$EF,$E1,$25
	.dc.b	$FD,$DB,$FF,$6D,$B7,$E1,$3A,$11,$D7,$C0,$F7,$F8,$1F,$07,$C2,$49
	.dc.b	$E5,$1D,$9E,$98,$76,$7C,$98,$F3,$7C,$D8,$5C,$FF,$77,$93,$2F,$67
	.dc.b	$C0,$CB,$F3,$3E,$FF,$83,$F8,$1F,$09,$27,$D4,$FC,$3F,$82,$F8,$9F
	.dc.b	$07,$27,$DC,$FC,$7F,$83,$F9,$1F,$03,$F2,$7E,$06,$57,$C3,$8E,$8F
	.dc.b	$CB,$C2,$EB,$E6,$C7,$A4,$FC,$FC,$6C,$FA,$B1,$FA,$9F,$03,$2B,$E7
	.dc.b	$3F,$5F,$E0,$B9,$F9,$1F,$C1,$7C,$27,$0B,$4C,$E1,$FD,$96,$1A,$76
	.dc.b	$BE,$10,$D3,$15,$BC,$FF,$09,$F0,$7F,$85,$F8,$39,$3F,$7E,$3F,$11
	.dc.b	$F0,$7F,$89,$F8,$09,$3F,$83,$8F,$C6,$7C,$04,$8F,$C3,$3F,$C3,$7F
	.dc.b	$16,$FE,$67,$1F,$B1,$5F,$03,$9F,$56,$4F,$F0,$99,$CB,$D4,$FC,$86
	.dc.b	$1E,$07,$F3,$18,$77,$7F,$35,$85,$E4,$FE,$7F,$7E,$06,$FC,$27,$8F
	.dc.b	$E1,$25,$F7,$8F,$37,$C2,$79,$FE,$0E,$4F,$8E,$3D,$5F,$09,$23,$CC
	.dc.b	$FB,$7E,$13,$DD,$F0,$3E,$FF,$84,$95,$EB,$8E,$AF,$B7,$0E,$AF,$C3
	.dc.b	$C3,$A3,$F1,$70,$B9,$FA,$B1,$F9,$5F,$05,$2B,$E3,$3F,$37,$E0,$FE
	.dc.b	$77,$C2,$7C,$FF,$84,$95,$F2,$E3,$E9,$7C,$24,$8F,$9C,$FD,$5F,$81
	.dc.b	$FA,$DF,$09,$F5,$FE,$0A,$57,$D3,$8E,$AF,$D5,$C3,$AB,$F6,$F0,$E8
	.dc.b	$FD,$DC,$2E,$9F,$4B,$1F,$C1,$7C,$27,$E0,$FE,$12,$5F,$DF,$8E,$AF
	.dc.b	$E1,$71,$EB,$7F,$6B,$1E,$7E,$0F,$E8,$79,$D9,$E3,$0F,$F2,$24,$33
	.dc.b	$FE,$B9,$E6,$3E,$2A,$FC,$44,$BF,$EA,$D3,$96,$FC,$37,$02,$3A,$F8
	.dc.b	$6E,$7F,$09,$D3,$E0,$A4,$F2,$8E,$FF,$03,$E0,$F8,$29,$3D,$A3,$C7
	.dc.b	$F0,$D2,$3C,$2F,$9B,$E1,$BC,$FF,$0D,$E8,$F8,$69,$5E,$68,$F5,$FC
	.dc.b	$2F,$B3,$E1,$64,$FA,$A3,$DF,$F0,$7F,$03,$E0,$E4,$FB,$A3,$E1,$FC
	.dc.b	$17,$C4,$F8,$39,$3F,$0A,$3E,$3F,$C2,$7C,$8F,$85,$93,$F1,$A3,$E5
	.dc.b	$FC,$2C,$8F,$92,$FF,$82,$7F,$14,$FE,$93,$DB,$FC,$55,$FE,$8B,$3F
	.dc.b	$DB,$D7,$C2,$6C,$E1,$F0,$9C,$7A,$A5,$7F,$82,$69,$1D,$7E,$13,$C1
	.dc.b	$3F,$DE,$61,$8E,$3C,$2F,$F1,$ED,$F8,$AB,$FD,$0C,$F8,$A3,$38,$F8
	.dc.b	$67,$B6,$3E,$9F,$85,$91,$E1,$7D,$9F,$0D,$D7,$4F,$E6,$19,$F8,$2B
	.dc.b	$E0,$9E,$73,$D6,$7A,$70,$EC,$FC,$1C,$3E,$37,$C3,$7F,$90,$7E,$05
	.dc.b	$A2,$30,$BE,$1A,$C6,$3E,$0B,$3F,$05,$74,$F8,$CF,$0F,$81,$91,$F2
	.dc.b	$7E,$06,$E9,$F4,$7E,$17,$A9,$DB,$E1,$AD,$7C,$FC,$7A,$B4,$EA,$7D
	.dc.b	$3F,$86,$F2,$9E,$6F,$86,$B1,$F5,$FE,$13,$D2,$7A,$BE,$16,$C7,$3C
	.dc.b	$BF,$05,$AF,$A5,$B1,$F7,$FE,$0B,$E0,$9F,$0B,$E0,$EC,$7E,$0F,$E0
	.dc.b	$FE,$29,$F2,$7F,$12,$FF,$5F,$F2,$23,$1F,$0D,$9F,$86,$B7,$FE,$53
	.dc.b	$B4,$71,$F8,$59,$39,$F8,$6B,$A7,$F9,$16,$DF,$07,$DC,$F0,$7C,$25
	.dc.b	$BE,$99,$77,$69,$DF,$2E,$7E,$0F,$FA,$43,$06,$7E,$9C,$F3,$7F,$DE
	.dc.b	$E5,$F0,$9E,$D3,$DD,$F0,$96,$7A,$EB,$E5,$CB,$DA,$D8,$F7,$63,$F1
	.dc.b	$7E,$17,$AC,$FF,$B9,$F1,$47,$C9,$F8,$6F,$95,$F0,$D7,$0F,$8C,$FC
	.dc.b	$DF,$86,$F9,$DF,$03,$67,$E5,$3F,$47,$E0,$A4,$7C,$DF,$82,$B9,$FC
	.dc.b	$9F,$81,$FA,$DF,$0B,$F5,$DB,$5F,$4F,$1E,$AD,$3A,$9F,$5F,$0F,$BD
	.dc.b	$F0,$DD,$39,$3E,$53,$F8,$1F,$85,$95,$F7,$7E,$1A,$E7,$F6,$71,$EA
	.dc.b	$D9,$EA,$FD,$DC,$7A,$C7,$73,$F0,$D8,$7E,$D7,$F1,$2F,$F6,$05,$3C
	.dc.b	$B4,$7F,$AD,$EB,$FC,$64,$EA,$33,$F0,$BC,$7E,$16,$57,$FA,$8E,$9F
	.dc.b	$0B,$8C,$3F,$D2,$51,$FE,$A0,$9D,$7C,$2F,$87,$E1,$7C,$4F,$8F,$E1
	.dc.b	$3C,$9F,$0B,$6B,$A3,$CF,$8B,$FD,$BB,$83,$3F,$56,$79,$BF,$3E,$5D
	.dc.b	$1F,$23,$73,$F1,$FC,$2F,$C1,$F8,$5F,$84,$DA,$F5,$61,$CE,$BE,$26
	.dc.b	$3E,$C9,$C9,$D0,$F5,$96,$3D,$FA,$FC,$DF,$84,$93,$F1,$7E,$16,$D7
	.dc.b	$C4,$F8,$4F,$A4,$7D,$3F,$85,$B5,$F2,$FE,$13,$EB,$7C,$2F,$D7,$6C
	.dc.b	$7C,$FF,$84,$FB,$5F,$0B,$F6,$DB,$1F,$4F,$0E,$78,$FD,$2C,$6F,$1F
	.dc.b	$CA,$CB,$9C,$FD,$6C,$AC,$FC,$EF,$84,$FC,$31,$F8,$7F,$85,$B9,$75
	.dc.b	$25,$F5,$6D,$75,$7F,$09,$87,$A5,$A7,$E1,$FE,$13,$F2,$47,$F4,$7E
	.dc.b	$65,$E2,$7F,$DD,$D0,$F8,$0D,$1B,$F8,$1B,$3A,$F8,$2E,$5F,$0B,$CD
	.dc.b	$B1,$CB,$E1,$7B,$1D,$FE,$16,$C7,$6C,$39,$E5,$FE,$E9,$CB,$FB,$DC
	.dc.b	$E4,$EC,$79,$31,$EC,$D3,$CF,$F0,$9E,$B3,$D9,$F0,$B6,$39,$5F,$5C
	.dc.b	$3D,$51,$8D,$CF,$FA,$87,$1F,$F6,$03,$D6,$BC,$9A,$7B,$1F,$37,$C0
	.dc.b	$FC,$9F,$77,$E5,$36,$BE,$4F,$C2,$FC,$D3,$E7,$7C,$2D,$8F,$9B,$F0
	.dc.b	$BF,$47,$E1,$7E,$93,$63,$E8,$FC,$2F,$D5,$3E,$B7,$C2,$D8,$EA,$4C
	.dc.b	$FE,$46,$1D,$5F,$95,$F0,$BF,$77,$E1,$7E,$F3,$6B,$EE,$FC,$2F,$E0
	.dc.b	$7E,$17,$F0,$4D,$8E,$9C,$5F,$13,$0F,$E2,$DC,$5F,$EB,$D9,$FC,$2E
	.dc.b	$5D,$EB,$5E,$77,$F1,$2F,$84,$FF,$09,$7E,$25,$F3,$2F,$F1,$71,$87
	.dc.b	$3F,$07,$F9,$0C,$98,$72,$BF,$ED,$71,$F9,$48,$C7,$C0,$6D,$D5,$5B
	.dc.b	$FF,$6C,$5F,$F6,$2B,$E1,$18,$F8,$3D,$47,$F2,$F9,$46,$32,$47,$19
	.dc.b	$F1,$D7,$4F,$85,$D4,$BE,$16,$98,$8C,$FC,$2E,$BE,$16,$61,$E2,$3C
	.dc.b	$11,$CB,$E1,$79,$E7,$27,$59,$F8,$A7,$A4,$63,$2F,$F0,$3B,$71,$F1
	.dc.b	$FC,$27,$93,$E1,$24,$ED,$F3,$FC,$27,$78,$E5,$97,$77,$DA,$7C,$3C
	.dc.b	$79,$BD,$48,$6B,$2F,$8B,$87,$C2,$F7,$BE,$1F,$C2,$4A,$F3,$47,$C6
	.dc.b	$F7,$FE,$3F,$C2,$49,$F5,$47,$CA,$F8,$39,$1E,$D7,$E6,$FC,$27,$CE
	.dc.b	$F8,$2F,$9F,$F0,$92,$BE,$14,$75,$7E,$26,$1D,$5F,$37,$13,$1E,$B1
	.dc.b	$8C,$6F,$47,$E3,$72,$FB,$7F,$09,$2B,$E7,$3F,$83,$FE,$01,$FF,$5C
	.dc.b	$47,$4C,$3B,$5F,$E5,$72,$F1,$4F,$53,$D4,$43,$C7,$87,$3F,$85,$E9
	.dc.b	$F0,$12,$7E,$FE,$29,$F0,$78,$12,$72,$F8,$FE,$0F,$C9,$F0,$12,$78
	.dc.b	$BE,$7F,$85,$F4,$7C,$2C,$9E,$AF,$AF,$E1,$3A,$47,$DB,$CB,$FC,$0E
	.dc.b	$F6,$3C,$78,$73,$7A,$90,$E1,$7F,$17,$E1,$7E,$37,$C2,$49,$F5,$C7
	.dc.b	$6E,$0F,$E0,$78,$76,$9F,$06,$9E,$B7,$E0,$E1,$EB,$7C,$9A,$78,$5F
	.dc.b	$37,$C2,$7D,$3F,$82,$91,$F2,$23,$EB,$7C,$17,$D7,$F8,$29,$3F,$32
	.dc.b	$3E,$D7,$C0,$48,$F9,$EF,$DD,$F8,$3F,$BD,$F0,$92,$7E,$9B,$F8,$1F
	.dc.b	$84,$EB,$1F,$5B,$0E,$AF,$C9,$C3,$A3,$F2,$F0,$BB,$FC,$0C,$7F,$15
	.dc.b	$F0,$9F,$8B,$F8,$49,$7F,$80,$8E,$AF,$D3,$C7,$AD,$FC,$8C,$79,$F9
	.dc.b	$BF,$DD,$3C,$57,$1B,$4C,$E3,$FE,$49,$86,$9F,$C6,$FF,$80,$7F,$DA
	.dc.b	$D6,$69,$3D,$FD,$7C,$2E,$FE,$02,$63,$FC,$4A,$BF,$DD,$6D,$F8,$3E
	.dc.b	$84,$75,$F0,$BD,$FE,$17,$C1,$F0,$B2,$79,$47,$67,$A6,$1D,$9F,$E1
	.dc.b	$98,$76,$6D,$F5,$73,$87,$AE,$3D,$9F,$0B,$ED,$F8,$49,$5E,$78,$EC
	.dc.b	$F4,$C3,$B7,$1F,$F5,$4C,$7B,$37,$0E,$95,$EA,$3E,$07,$C2,$FC,$9F
	.dc.b	$85,$F9,$5E,$EC,$9F,$89,$1F,$37,$E1,$24,$7C,$77,$E8,$7C,$2F,$D1
	.dc.b	$F8,$59,$3F,$2D,$FA,$9F,$0B,$F5,$7E,$12,$4F,$CF,$7E,$C7,$C2,$F4
	.dc.b	$8F,$8B,$85,$D7,$D5,$8F,$DD,$F8,$4F,$BD,$F0,$B2,$FE,$C4,$7E,$07
	.dc.b	$E1,$7F,$AE,$7B,$9F,$DE,$67,$F9,$D3,$70,$F8,$93,$D7,$BF,$F1,$DC
	.dc.b	$6C,$F1,$87,$FA,$56,$19,$F4,$F8,$4E,$BF,$0B,$2B,$FA,$CE,$69,$F0
	.dc.b	$BF,$E0,$9C,$FF,$C0,$AF,$FB,$5B,$2D,$FB,$F9,$7F,$3D,$19,$BF,$3F
	.dc.b	$C1,$65,$FF,$53,$47,$FB,$A2,$7B,$FC,$2E,$7C,$1F,$F7,$0D,$DA,$FF
	.dc.b	$6F,$D7,$93,$E1,$31,$F0,$B9,$75,$F0,$72,$7D,$3F,$0B,$73,$F0,$C7
	.dc.b	$B3,$1E,$99,$F5,$7B,$7C,$2C,$9E,$C7,$C2,$7C,$5F,$09,$23,$E2,$FC
	.dc.b	$2D,$C3,$A9,$38,$3E,$2E,$3D,$6B,$91,$F3,$3E,$03,$B5,$7D,$A7,$DD
	.dc.b	$F0,$12,$7E,$8F,$C1,$5A,$FB,$FF,$03,$F0,$FE,$13,$E2,$36,$BE,$BF
	.dc.b	$C2,$FC,$7F,$84,$F9,$0D,$8F,$F0,$9F,$F0,$0F,$FB,$44,$9F,$09,$F3
	.dc.b	$DF,$F6,$A7,$3B,$0E,$99,$78,$2F,$F0,$79,$74,$7F,$0B,$95,$9F,$C9
	.dc.b	$BC,$FE,$16,$4F,$83,$E0,$2D,$73,$3F,$33,$7E,$A9,$EA,$74,$24,$7E
	.dc.b	$7B,$4F,$53,$FE,$19,$CF,$FD,$C4,$DF,$7F,$D5,$F0,$7E,$B7,$D9,$F0
	.dc.b	$3E,$DF,$85,$B8,$FC,$27,$E0,$7C,$2F,$C1,$F8,$4B,$3F,$19,$F8,$9F
	.dc.b	$0B,$F1,$7E,$12,$CF,$CA,$79,$E7,$E3,$C2,$F4,$FF,$9A,$35,$47,$5B
	.dc.b	$E8,$7D,$2C,$B1,$7E,$BF,$85,$FA,$47,$D3,$F8,$0B,$5F,$63,$5F,$53
	.dc.b	$4F,$51,$D2,$BE,$26,$BE,$A6,$C7,$E0,$3E,$07,$EF,$1F,$7F,$E0,$6C
	.dc.b	$7E,$13,$E0,$3F,$04,$7E,$0F,$E1,$6C,$7E,$23,$E1,$7F,$0C,$7E,$1F
	.dc.b	$E1,$2C,$7E,$33,$E1,$7F,$71,$FC,$03,$FE,$D0,$8C,$1F,$E6,$27,$19
	.dc.b	$FD,$9C,$7A,$BF,$6F,$1B,$C7,$F2,$72,$E7,$F0,$BD,$AB,$FC,$B1,$87
	.dc.b	$FB,$67,$1E,$9F,$0B,$E0,$F8,$0F,$0B,$63,$D8,$4A,$EC,$DC,$7F,$D3
	.dc.b	$B7,$CE,$7A,$E1,$DE,$72,$FB,$3E,$0F,$DB,$F0,$37,$0F,$63,$F0,$3E
	.dc.b	$13,$E0,$FC,$2D,$9F,$80,$FC,$4F,$84,$F8,$BF,$0B,$67,$E2,$3F,$23
	.dc.b	$E1,$64,$7C,$7F,$84,$BA,$FC,$0F,$85,$F9,$A7,$CE,$F7,$AD,$7C,$DF
	.dc.b	$77,$E8,$FC,$2F,$D2,$6C,$7D,$1F,$85,$FA,$BF,$0B,$F5,$9B,$1F,$57
	.dc.b	$E1,$7E,$C9,$F6,$BE,$12,$C7,$D9,$F8,$5F,$BA,$7D,$EF,$85,$B1,$FE
	.dc.b	$9E,$FE,$01,$FF,$62,$6D,$FE,$8D,$6F,$47,$FB,$29,$FE,$95,$3B,$3F
	.dc.b	$9F,$2F,$85,$BE,$AE,$1B,$3D,$7E,$13,$67,$0F,$7E,$C7,$87,$DC,$E6
	.dc.b	$74,$F8,$2B,$1E,$5F,$83,$EF,$F0,$7E,$06,$C7,$A7,$E0,$FC,$7F,$07
	.dc.b	$E4,$6C,$7B,$7E,$0F,$FD,$99,$B3,$E0,$E9,$88,$CF,$C1,$CE,$2F,$88
	.dc.b	$F0,$F8,$3E,$3F,$07,$27,$E4,$3D,$3E,$0B,$AF,$BF,$DB,$DF,$95,$F3
	.dc.b	$63,$C3,$F0,$52,$3E,$83,$FE,$DF,$F2,$BF,$D3,$BD,$3F,$EF,$87,$FD
	.dc.b	$25,$3F,$ED,$25,$F0,$5B,$38,$7C,$1F,$1F,$83,$95,$B7,$A7,$C1,$F7
	.dc.b	$8F,$B9,$8D,$C7,$B7,$C6,$78,$7E,$12,$57,$77,$C9,$EE,$F9,$7E,$06
	.dc.b	$4F,$8D,$F4,$7C,$0F,$3E,$8F,$F3,$27,$43,$3C,$5B,$E1,$5C,$3E,$EB
	.dc.b	$FF,$35,$2F,$80,$F8,$44,$AF,$81,$F0,$BF,$17,$E1,$64,$7C,$37,$E4
	.dc.b	$7C,$27,$C9,$F8,$49,$3F,$1D,$F9,$9F,$09,$F3,$7E,$13,$E7,$7C,$14
	.dc.b	$AF,$99,$1F,$47,$E0,$BE,$97,$C1,$49,$FA,$11,$D1,$FF,$78,$6A,$C7
	.dc.b	$1C,$FF,$5F,$2F,$B5,$F0,$9F,$6F,$E1,$26,$1F,$66,$0C,$A7,$0F,$FA
	.dc.b	$2F,$F8,$07,$FD,$0A,$49,$D9,$FE,$FF,$5F,$05,$E0,$9F,$93,$95,$AF
	.dc.b	$9D,$9B,$3E,$38,$F6,$9F,$A5,$8F,$F9,$C3,$0F,$07,$C2,$E9,$B3,$E5
	.dc.b	$7C,$9F,$0B,$98,$E3,$8D,$E6,$61,$F7,$27,$81,$A2,$C6,$BE,$13,$84
	.dc.b	$BE,$12,$FC,$24,$BD,$CE,$5B,$F0,$9C,$3E,$12,$46,$0D,$FC,$27,$42
	.dc.b	$3C,$BE,$13,$BE,$3D,$63,$E5,$63,$87,$BF,$C2,$79,$3E,$12,$46,$0F
	.dc.b	$1F,$C1,$FA,$08,$F9,$BE,$13,$93,$D0,$F4,$DE,$39,$3E,$87,$07,$C6
	.dc.b	$F8,$3F,$82,$4A,$F7,$FC,$27,$C4,$F8,$4F,$8B,$F0,$92,$7E,$1C,$7C
	.dc.b	$8F,$84,$91,$F1,$9F,$97,$F0,$9F,$4F,$F8,$07,$FD,$15,$D5,$FE,$D5
	.dc.b	$6D,$7E,$C5,$B9,$9F,$F4,$24,$7D,$67,$F9,$D3,$7E,$17,$A1,$D7,$E1
	.dc.b	$7C,$31,$F9,$3C,$6E,$DE,$EE,$96,$7E,$3F,$85,$97,$D5,$F3,$7C,$2F
	.dc.b	$9F,$E1,$64,$F8,$5F,$57,$C2,$F4,$DF,$F4,$99,$DA,$FC,$CE,$3E,$FF
	.dc.b	$84,$91,$E9,$7E,$17,$C2,$FC,$3F,$84,$F8,$9F,$0B,$2B,$DD,$1F,$1F
	.dc.b	$E1,$24,$7C,$17,$E5,$7C,$2F,$CB,$F8,$59,$3F,$15,$F9,$DF,$0B,$F3
	.dc.b	$FE,$16,$4F,$C9,$7E,$97,$C2,$F4,$87,$2F,$E0,$BA,$19,$66,$31,$97
	.dc.b	$D9,$F8,$4F,$B5,$F0,$B3,$0F,$A9,$1F,$77,$DB,$FD,$87,$F0,$0F,$F2
	.dc.b	$99,$FE,$58,$DE,$DF,$B9,$3D,$79,$BF,$05,$9D,$BF,$C1,$65,$CB,$E1
	.dc.b	$64,$7E,$DB,$1F,$E9,$EF,$A2,$FF,$8B,$E2,$9F,$EA,$33,$BF,$85,$F1
	.dc.b	$FC,$2C,$8C,$1C,$FE,$13,$CE,$47,$B7,$C2,$7A,$BE,$13,$D7,$F0,$92
	.dc.b	$7C,$51,$EE,$F8,$39,$1E,$57,$E0,$FC,$1F,$38,$F3,$1F,$0B,$0C,$72
	.dc.b	$7E,$C3,$46,$5F,$23,$E0,$E5,$FB,$DF,$97,$F0,$7F,$33,$DC,$F9,$BE
	.dc.b	$E4,$AF,$89,$1D,$38,$BE,$1E,$2C,$F1,$5F,$4B,$2F,$AB,$F0,$52,$3E
	.dc.b	$63,$F6,$3E,$0F,$EC,$FC,$17,$FB,$97,$03,$7E,$0F,$0E,$7E,$0A,$67
	.dc.b	$F5,$7E,$0E,$E5,$F0,$3A,$1C,$FF,$E7,$4E,$3F,$E1,$ED,$BF,$B2,$FF
	.dc.b	$99,$63,$35,$FE,$F5,$87,$C0,$D3,$1F,$05,$6F,$F0,$99,$78,$30,$F9
	.dc.b	$99,$59,$EB,$F0,$5C,$FE,$0B,$A3,$63,$FA,$0E,$5F,$CF,$DB,$EC,$F8
	.dc.b	$7D,$EF,$13,$CF,$5F,$CD,$E7,$F1,$72,$C6,$DE,$EF,$82,$F4,$7C,$17
	.dc.b	$A5,$B9,$76,$F8,$2F,$61,$ED,$F8,$2B,$1E,$2F,$82,$F8,$07,$C1,$F8
	.dc.b	$2B,$1E,$6F,$82,$F8,$9F,$05,$F1,$5B,$1E,$AF,$82,$F9,$1F,$05,$F2
	.dc.b	$5B,$1D,$49,$C1,$F5,$78,$D8,$F6,$8C,$7C,$0F,$D1,$F8,$2F,$A4,$DB
	.dc.b	$F8,$FE,$DF,$D5,$3E,$B7,$B7,$63,$AF,$57,$DE,$DF,$D5,$18,$3F,$6F
	.dc.b	$FC,$03,$FE,$78,$6F,$C5,$7F,$80,$6B,$FA,$9C,$E7,$3F,$F1,$A3,$FE
	.dc.b	$7C,$7F,$18,$FF,$9E,$EB,$FB,$CB,$63,$3F,$09,$CE,$7B,$FC,$25,$9E
	.dc.b	$7F,$07,$7A,$58,$3E,$0F,$0A,$6C,$63,$E0,$F3,$F0,$77,$2F,$0B,$C3
	.dc.b	$E0,$F8,$FC,$1D,$9F,$2B,$D3,$0E,$71,$E8,$6E,$7E,$CF,$82,$F0,$9E
	.dc.b	$2F,$83,$B5,$EE,$F6,$FC,$BE,$DF,$99,$B1,$CC,$F8,$7A,$F7,$69,$F1
	.dc.b	$BE,$0F,$DB,$F0,$7E,$E6,$C7,$CA,$F8,$3F,$F1,$47,$09,$A6,$3E,$0F
	.dc.b	$3F,$03,$AF,$82,$9B,$FD,$28,$E3,$F0,$52,$3E,$AB,$FE,$33,$FE,$01
	.dc.b	$FE,$84,$7F,$3A,$EA,$FF,$08,$B7,$3F,$E7,$6D,$AF,$F3,$A3,$FC,$F1
	.dc.b	$FC,$0D,$F8,$1F,$C1,$E1,$E4,$7F,$C7,$72,$F8,$2F,$01,$23,$3E,$1F
	.dc.b	$8F,$DA,$91,$C1,$F3,$7B,$5C,$E3,$A9,$F8,$EC,$31,$9F,$A3,$E0,$7D
	.dc.b	$84,$9F,$0F,$C1,$7B,$FE,$0B,$F1,$DA,$3E,$0A,$C6,$1C,$FC,$16,$BE
	.dc.b	$0A,$F0,$FA,$9E,$3F,$03,$23,$DB,$F0,$57,$3F,$E5,$D9,$3E,$0B,$B9
	.dc.b	$E0,$F8,$2B,$7F,$13,$2E,$CD,$3B,$1F,$23,$E0,$3C,$E7,$A3,$E0,$6C
	.dc.b	$7C,$CF,$6F,$D7,$ED,$FB,$1B,$1C,$F3,$F1,$ED,$EA,$6C,$7F,$9B,$BF
	.dc.b	$80,$7F,$A0,$47,$FA,$BD,$FE,$81,$1F,$DE,$9C,$F9,$67,$33,$BF,$82
	.dc.b	$FD,$C4,$3E,$0A,$C6,$1C,$FC,$0C,$C3,$C1,$F0,$57,$0F,$EA,$98,$7F
	.dc.b	$3F,$6F,$C0,$F4,$F8,$2E,$AF,$6F,$81,$98,$7A,$7E,$0A,$DF,$2C,$BA
	.dc.b	$76,$FF,$9A,$B2,$67,$8C,$3D,$9F,$05,$E8,$3D,$3F,$05,$63,$E2,$7B
	.dc.b	$3E,$CF,$7B,$DA,$D8,$F9,$19,$73,$CF,$E4,$63,$78,$BE,$37,$C0,$FC
	.dc.b	$43,$E2,$FC,$15,$C3,$E8,$7C,$0F,$C8,$3E,$4F,$C1,$58,$FA,$9F,$03
	.dc.b	$F3,$3E,$0B,$E6,$B6,$3E,$C7,$C1,$7D,$03,$E8,$FC,$15,$8F,$F5,$4F
	.dc.b	$B7,$FC,$DF,$4F,$E5,$11,$FC,$B6,$3F,$9C,$72,$FF,$A3,$9D,$72,$B1
	.dc.b	$ED,$C1,$F5,$1B,$5B,$F8,$0E,$47,$3F,$81,$B8,$F8,$31,$ED,$EC,$F7
	.dc.b	$6C,$7B,$0F,$C8,$6D,$EC,$69,$E8,$F8,$1F,$37,$C0,$F9,$DB,$1E,$CF
	.dc.b	$81,$F5,$7C,$0F,$AD,$B1,$F0,$3E,$07,$DC,$7B,$FE,$06,$C7,$C4,$F8
	.dc.b	$0F,$84,$7C,$3F,$81,$B1,$F2,$3E,$03,$E3,$7C,$0F,$C7,$6C,$75,$27
	.dc.b	$27,$F8,$AF,$76,$3D,$A3,$1F,$01,$F3,$FE,$03,$E8,$36,$FE,$AF,$BF
	.dc.b	$F4,$FD,$BF,$A8,$D8,$EB,$CF,$F5,$77,$F5,$46,$0F,$F3,$6E,$6F,$80
	.dc.b	$FB,$87,$F6,$9F,$E0,$17,$C5,$62,$BF,$DD,$70,$F8,$6D,$7C,$5E,$DB
	.dc.b	$3A,$F8,$4E,$5F,$09,$CD,$B1,$CB,$E1,$3B,$7C,$27,$76,$C7,$6F,$84
	.dc.b	$F1,$7C,$27,$8D,$B1,$D4,$F1,$E1,$D5,$A7,$9B,$E1,$3D,$47,$AF,$E1
	.dc.b	$2C,$7A,$BD,$FF,$71,$EF,$F6,$EC,$75,$E9,$FF,$2F,$70,$7F,$97,$A3
	.dc.b	$1C,$BF,$EE,$C8,$C4,$EB,$E1,$3E,$47,$C2,$7C,$96,$DF,$C8,$F8,$3F
	.dc.b	$99,$F0,$7F,$35,$B1,$F3,$3E,$0F,$E8,$1F,$47,$E0,$EC,$7D,$0F,$82
	.dc.b	$FA,$87,$D5,$F8,$3B,$0C,$FF,$CE,$9F,$C0,$3F,$CB,$89,$F1,$7F,$E2
	.dc.b	$37,$F9,$8E,$DF,$E8,$0A,$F0,$1F,$73,$1E,$F7,$CB,$E1,$78,$1C,$7E
	.dc.b	$16,$D7,$7E,$13,$FE,$B6,$C5,$EE,$F7,$3C,$1E,$ED,$9E,$F8,$FC,$CD
	.dc.b	$FD,$D1,$8C,$39,$CB,$D7,$63,$FC,$EE,$73,$18,$F8,$4F,$67,$C2,$FB
	.dc.b	$5F,$77,$C2,$CC,$7C,$C7,$97,$E1,$71,$97,$C6,$F8,$5F,$88,$7C,$5F
	.dc.b	$85,$B5,$C2,$7C,$5A,$3E,$13,$E5,$1F,$2F,$E1,$6D,$6F,$5F,$3F,$C2
	.dc.b	$FC,$F3,$E8,$7C,$2D,$9D,$60,$CF,$BE,$5F,$63,$3B,$C1,$EE,$A7,$C2
	.dc.b	$FD,$93,$ED,$7C,$2D,$C7,$9F,$C1,$7D,$DF,$7F,$EF,$36,$3B,$E7,$D1
	.dc.b	$FA,$D9,$5C,$FD,$BF,$0B,$F8,$5F,$85,$FC,$33,$6B,$D9,$FC,$03,$F5
	.dc.b	$2B,$FD,$80,$DD,$BE,$9B,$67,$FD,$EE,$BE,$27,$F2,$95,$C3,$E1,$AD
	.dc.b	$F0,$6D,$F0,$6C,$FE,$06,$DF,$0D,$8E,$44,$7F,$9E,$67,$CB,$EB,$E6
	.dc.b	$31,$1E,$6F,$86,$C7,$C2,$C8,$F5,$BB,$F8,$6E,$19,$49,$F7,$BC,$F2
	.dc.b	$E9,$F0,$B2,$7E,$1B,$DF,$E1,$BC,$1E,$EC,$9F,$8E,$F8,$FE,$0B,$A4
	.dc.b	$7C,$13,$A9,$F3,$2F,$D1,$F0,$DE,$9F,$86,$93,$F4,$23,$D9,$F0,$BF
	.dc.b	$E9,$BD,$DF,$0D,$63,$0E,$7E,$16,$72,$7D,$8F,$86,$B8,$FD,$8F,$85
	.dc.b	$E5,$F0,$DC,$DB,$5F,$7B,$E1,$BE,$51,$2B,$F3,$9E,$6F,$F4,$26,$FC
	.dc.b	$27,$D3,$8F,$F3,$D7,$57,$F8,$92,$31,$F0,$B9,$F8,$59,$7C,$9E,$1F
	.dc.b	$03,$C7,$E1,$3F,$D6,$D8,$BE,$16,$C6,$1C,$FC,$24,$D7,$C7,$F0,$B7
	.dc.b	$27,$85,$FC,$13,$17,$C2,$F4,$F8,$5E,$AD,$C3,$D7,$F0,$BE,$0F,$85
	.dc.b	$F0,$B6,$3D,$F9,$F5,$C7,$A6,$76,$7F,$97,$71,$7F,$90,$BE,$13,$D2
	.dc.b	$7A,$BE,$16,$CF,$C8,$F8,$4F,$8D,$F0,$B3,$83,$07,$CC,$F7,$7E,$51
	.dc.b	$1F,$9F,$F0,$3F,$37,$E1,$64,$7D,$27,$E8,$7C,$2F,$E8,$FF,$80,$7F
	.dc.b	$5B,$3F,$0B,$2F,$2C,$3E,$1B,$46,$E7,$87,$C2,$F1,$F8,$59,$1F,$C6
	.dc.b	$69,$3D,$7E,$16,$4E,$5F,$07,$4F,$87,$E1,$7C,$5F,$0B,$2B,$94,$79
	.dc.b	$7E,$17,$CD,$F0,$B2,$7B,$47,$A7,$DC,$F5,$7B,$92,$7C,$51,$ED,$F8
	.dc.b	$59,$1E,$57,$E0,$7C,$2F,$C1,$F8,$59,$3E,$97,$E2,$7C,$2F,$C5,$F8
	.dc.b	$4F,$8D,$F0,$92,$BD,$D1,$F2,$7E,$0F,$FA,$7F,$83,$FD,$33,$91,$F0
	.dc.b	$98,$E8,$F9,$75,$70,$FE,$A9,$5A,$9F,$91,$F0,$79,$F0,$92,$F8,$4C
	.dc.b	$29,$B1,$BF,$E0,$1F,$F3,$61,$FE,$6A,$BF,$C6,$B8,$F8,$5D,$7C,$0C
	.dc.b	$9F,$E9,$D3,$FC,$91,$B1,$DA,$FA,$C7,$90,$FB,$97,$DF,$E1,$64,$72
	.dc.b	$F8,$5B,$A7,$E2,$4F,$EE,$18,$BE,$17,$CD,$F0,$BE,$76,$DF,$5A,$E1
	.dc.b	$B7,$C0,$CB,$C5,$3E,$03,$CF,$F0,$BE,$FF,$85,$F8,$0D,$8F,$5F,$C2
	.dc.b	$FC,$3F,$85,$F8,$8D,$8F,$7F,$C1,$FC,$7F,$84,$F9,$0D,$8F,$87,$F0
	.dc.b	$7F,$2F,$E0,$FE,$63,$63,$E3,$FC,$1F,$CF,$3E,$87,$C1,$D8,$EA,$7C
	.dc.b	$CC,$3A,$B4,$F9,$FF,$05,$F5,$FE,$0F,$EC,$36,$3E,$9F,$C1,$FD,$B3
	.dc.b	$F0,$5F,$C0,$3F,$E6,$88,$C9,$DB,$8F,$E3,$F3,$31,$EF,$18,$78,$FC
	.dc.b	$24,$AF,$F0,$65,$BE,$13,$17,$F4,$21,$F0,$9E,$0F,$84,$F0,$B6,$F8
	.dc.b	$FC,$27,$90,$F2,$FC,$25,$8E,$BF,$01,$E8,$F8,$0F,$4B,$63,$C3,$F0
	.dc.b	$7E,$C3,$DB,$F0,$96,$3A,$93,$4F,$26,$1D,$5F,$37,$C1,$FC,$4F,$84
	.dc.b	$F8,$AD,$AF,$6F,$C2,$7C,$83,$E4,$FC,$25,$8E,$9E,$6F,$F9,$03,$85
	.dc.b	$BF,$AA,$72,$7F,$34,$87,$C2,$7D,$2F,$84,$FA,$6D,$9F,$97,$F0,$9F
	.dc.b	$58,$FA,$FF,$09,$63,$E7,$FC,$27,$DA,$3E,$DF,$C2,$58,$FF,$1F,$65
	.dc.b	$FE,$6F,$6F,$07,$F0,$18,$FE,$61,$18,$37,$F0,$59,$F8,$3D,$36,$3B
	.dc.b	$92,$BE,$E5,$F7,$7F,$5B,$7F,$E6,$C9,$F0,$FC,$17,$7F,$83,$F0,$3E
	.dc.b	$1F,$82,$93,$E6,$F8,$3B,$87,$E1,$F4,$E9,$C5,$F2,$F3,$B7,$FD,$73
	.dc.b	$37,$C1,$7A,$FE,$0F,$D8,$D9,$F8,$3F,$05,$EF,$3E,$07,$C1,$D8,$F8
	.dc.b	$BF,$07,$F0,$FE,$0F,$E2,$36,$3E,$4F,$C1,$FC,$7F,$83,$F9,$0D,$8F
	.dc.b	$9B,$F0,$5F,$2F,$E0,$FE,$63,$63,$A9,$F4,$B0,$EA,$D3,$EA,$FC,$17
	.dc.b	$D3,$F8,$3F,$A8,$D8,$FB,$3F,$01,$F5,$CF,$B1,$ED,$D8,$EB,$CB,$F4
	.dc.b	$30,$EA,$D8,$FF,$3F,$7F,$00,$FF,$A8,$1A,$7F,$41,$7F,$9D,$47,$FA
	.dc.b	$65,$D3,$FC,$99,$7C,$16,$FE,$1F,$83,$6B,$AF,$C1,$F3,$3A,$7C,$1D
	.dc.b	$8F,$0F,$C1,$77,$F8,$3F,$03,$63,$CB,$F0,$3E,$33,$C9,$F0,$56,$3A
	.dc.b	$93,$4F,$26,$1D,$5F,$37,$C0,$FA,$FE,$0B,$D8,$DA,$F8,$3F,$05,$EF
	.dc.b	$3E,$07,$B5,$63,$A7,$83,$FE,$B6,$E1,$6F,$EA,$9C,$9F,$E2,$A8,$7C
	.dc.b	$0F,$C9,$F8,$2F,$94,$D9,$F9,$FF,$05,$FE,$81,$A3,$4C,$7C,$16,$7E
	.dc.b	$07,$5F,$05,$38,$3E,$BC,$71,$F8,$29,$1F,$69,$FF,$46,$7F,$00,$FF
	.dc.b	$97,$63,$FA,$DB,$FE,$5C,$27,$C1,$6B,$E0,$BB,$4F,$E0,$B0,$EC,$FD
	.dc.b	$DC,$39,$FC,$0C,$8F,$F6,$26,$1F,$CD,$9B,$F0,$5E,$0F,$32,$46,$9F
	.dc.b	$1F,$BD,$DA,$3F,$11,$8F,$6E,$4F,$99,$97,$A3,$E0,$7D,$3F,$05,$2F
	.dc.b	$BC,$7B,$3E,$0B,$DB,$F0,$52,$7C,$71,$F0,$3E,$07,$E0,$FC,$14,$9F
	.dc.b	$3C,$7C,$4F,$82,$F8,$BF,$05,$27,$D7,$1F,$23,$E0,$24,$7B,$9F,$97
	.dc.b	$F0,$3D,$63,$E0,$E1,$D5,$F9,$B8,$7D,$0F,$81,$91,$F1,$9F,$A7,$F0
	.dc.b	$3F,$53,$DB,$FA,$BE,$DC,$AF,$97,$1D,$5F,$AF,$87,$5E,$27,$7F,$F9
	.dc.b	$DB,$A1,$9F,$FB,$27,$F8,$07,$FC,$CF,$5A,$C1,$F0,$A9,$FF,$34,$47
	.dc.b	$FB,$ED,$AF,$F8,$79,$E3,$87,$FA,$09,$FF,$33,$C6,$FE,$0B,$B1,$DF
	.dc.b	$E0,$FC,$1F,$07,$27,$A3,$E3,$F8,$3F,$27,$C1,$C9,$F0,$3E,$7F,$83
	.dc.b	$F4,$7C,$17,$A7,$E0,$E5,$79,$63,$A3,$E5,$C2,$ED,$ED,$DB,$D9,$97
	.dc.b	$C1,$F8,$39,$7E,$D7,$E2,$7C,$1F,$C5,$F8,$39,$3F,$05,$F9,$1E,$DF
	.dc.b	$4E,$2F,$81,$9D,$AF,$93,$C4,$CB,$E7,$7C,$17,$CF,$F8,$39,$5F,$2E
	.dc.b	$3E,$97,$C1,$7D,$3F,$83,$93,$F3,$E3,$EB,$7C,$17,$D7,$F8,$39,$3F
	.dc.b	$4E,$3E,$D7,$C1,$C8,$FA,$CF,$DD,$F8,$34,$4E,$4F,$85,$EE,$FD,$F8
	.dc.b	$FF,$2B,$B6,$33,$C1,$C2,$73,$F0,$DE,$27,$F0,$59,$78,$9F,$C1,$65
	.dc.b	$BB,$EB,$F0,$77,$0F,$F6,$2E,$2F,$84,$C6,$1F,$8E,$8C,$3E,$59,$DD
	.dc.b	$6B,$A1,$5F,$CA,$2B,$47,$17,$46,$9C,$C7,$0F,$73,$D1,$1F,$97,$CF
	.dc.b	$D1,$97,$E0,$F9,$1A,$72,$F8,$39,$7C,$B2,$FF,$40,$37,$E1,$3B,$FC
	.dc.b	$27,$83,$E1,$24,$EE,$3C,$7F,$07,$E4,$F8,$49,$3C,$E3,$CF,$F0,$92
	.dc.b	$3B,$3E,$AF,$84,$F5,$FC,$1F,$B3,$E1,$25,$78,$E3,$AB,$E5,$C3,$AB
	.dc.b	$F6,$B0,$F8,$7F,$07,$F1,$3E,$0E,$4F,$AE,$3E,$3F,$B7,$23,$DC,$FC
	.dc.b	$AF,$73,$AC,$7E,$13,$0E,$BC,$9F,$88,$C7,$E7,$FC,$1C,$AF,$8C,$FD
	.dc.b	$7F,$E0,$1F,$E9,$15,$FA,$F5,$F0,$9F,$EC,$36,$F6,$7D,$67,$FC,$F4
	.dc.b	$D8,$D7,$C1,$EF,$E0,$EC,$ED,$E5,$F0,$72,$39,$7C,$1D,$CF,$FD,$63
	.dc.b	$8B,$E0,$BC,$1F,$07,$E1,$6D,$F8,$72,$EC,$D3,$B1,$E5,$F8,$2F,$47
	.dc.b	$C1,$FA,$5B,$1E,$9F,$67,$D8,$7B,$7D,$BB,$1C,$F3,$F1,$ED,$EA,$6C
	.dc.b	$7C,$1F,$82,$F8,$9F,$07,$F1,$5B,$1F,$17,$E0,$FE,$47,$C1,$FC,$96
	.dc.b	$C7,$C9,$F8,$3F,$99,$F0,$7F,$35,$B1,$F3,$7E,$0B,$E8,$1F,$47,$E0
	.dc.b	$EC,$7E,$9B,$F8,$07,$FC,$F1,$A7,$FB,$61,$BD,$3F,$C9,$DF,$F3,$CC
	.dc.b	$64,$ED,$3F,$67,$1B,$1F,$41,$FF,$80,$12,$F8,$3B,$F0,$78,$6D,$F8
	.dc.b	$7D,$BD,$FB,$7C,$1B,$1E,$D3,$B1,$71,$EC,$FA,$7E,$0F,$BF,$C1,$F8
	.dc.b	$1B,$1E,$DF,$82,$F1,$FC,$1F,$91,$B1,$F0,$7E,$0F,$CE,$7A,$3E,$0E
	.dc.b	$C7,$C5,$F8,$3F,$5F,$C1,$FB,$1B,$1F,$27,$E0,$BD,$E7,$C0,$F8,$3B
	.dc.b	$1D,$49,$AF,$CD,$C3,$AB,$F3,$FE,$0B,$E3,$9F,$23,$E0,$ED,$7D,$5F
	.dc.b	$6B,$E5,$FB,$9F,$31,$B1,$D3,$A3,$EB,$6D,$E9,$8C,$1F,$D0,$3D,$E7
	.dc.b	$43,$E0,$FE,$A9,$F5,$BE,$0E,$D7,$F2,$8F,$E0,$1F,$F3,$AE,$3F,$D9
	.dc.b	$DB,$1C,$DF,$F3,$DE,$07,$3B,$7F,$82,$1F,$F3,$DC,$F3,$F8,$2F,$F0
	.dc.b	$65,$F3,$F8,$3B,$3C,$E6,$DF,$1F,$83,$B8,$F5,$6E,$3F,$D0,$AF,$CD
	.dc.b	$18,$F8,$2C,$FC,$0D,$9F,$53,$C3,$19,$1E,$DC,$AE,$4C,$FD,$DF,$01
	.dc.b	$D8,$EF,$F0,$36,$FE,$27,$B1,$E2,$F6,$7C,$6D,$8F,$71,$30,$E8,$FC
	.dc.b	$7C,$7B,$E3,$F2,$FE,$07,$D6,$7B,$3E,$06,$D7,$D1,$F8,$0F,$7F,$C0
	.dc.b	$FC,$06,$C7,$D5,$F8,$0F,$86,$7C,$4F,$81,$B1,$F6,$7E,$07,$E3,$9F
	.dc.b	$23,$E0,$6C,$7D,$DF,$81,$F9,$67,$CE,$F2,$1F,$1F,$89,$EC,$4E,$0F
	.dc.b	$59,$F7,$71,$EE,$E3,$A9,$7F,$CF,$23,$0F,$F3,$C6,$FC,$17,$7F,$87
	.dc.b	$F0,$7B,$D3,$1F,$EE,$F6,$F6,$B1,$87,$5E,$EF,$F4,$ED,$78,$F9,$7F
	.dc.b	$29,$A7,$89,$FF,$1E,$60,$F8,$3F,$61,$ED,$F8,$3B,$7D,$BE,$07,$E0
	.dc.b	$7C,$17,$C1,$6C,$78,$BE,$07,$E2,$1F,$17,$E0,$AC,$79,$BE,$0B,$E4
	.dc.b	$1F,$27,$E0,$AC,$7A,$BE,$07,$E6,$7C,$17,$CD,$6C,$75,$27,$0F,$C6
	.dc.b	$C3,$AB,$F2,$3E,$07,$EA,$7C,$17,$D5,$6D,$7C,$6F,$82,$FB,$07,$D9
	.dc.b	$F8,$0B,$1D,$38,$FE,$1E,$DE,$98,$C1,$FC,$0F,$8D,$F0,$5F,$81,$F8
	.dc.b	$2F,$C1,$36,$7F,$C5,$3F,$C0,$3F,$E7,$CD,$7F,$D0,$6D,$ED,$34,$FF
	.dc.b	$41,$E5,$FE,$6D,$7F,$D0,$33,$F8,$55,$F0,$7F,$E3,$B6,$C7,$6F,$83
	.dc.b	$C7,$C1,$67,$E0,$E4,$F8,$A3,$86,$BC,$7E,$0E,$4F,$9A,$3A,$5F,$5C
	.dc.b	$64,$F5,$3D,$7F,$05,$E1,$23,$EE,$F8,$3F,$27,$B5,$E5,$F7,$E4,$FC
	.dc.b	$28,$EA,$FC,$4B,$EA,$FB,$70,$F6,$7C,$14,$8F,$92,$FB,$FE,$0F,$E0
	.dc.b	$7C,$1F,$C1,$F8,$39,$5F,$3A,$3E,$27,$C1,$C8,$FA,$2F,$C7,$F8,$3F
	.dc.b	$91,$F0,$5F,$27,$E0,$E5,$7D,$68,$F9,$9F,$05,$FC,$A3,$99,$F0,$56
	.dc.b	$30,$FF,$2C,$FE,$01,$FE,$6F,$D4,$69,$FC,$F0,$BF,$01,$B3,$87,$B9
	.dc.b	$7B,$7F,$CD,$B2,$F8,$3E,$87,$5F,$86,$B3,$AF,$84,$F0,$7C,$27,$85
	.dc.b	$B1,$CB,$E1,$3C,$9F,$09,$E5,$6C,$76,$F8,$3F,$F4,$16,$4F,$84,$A6
	.dc.b	$23,$3F,$09,$AF,$84,$9C,$BE,$68,$E3,$F0,$7C,$BE,$12,$4F,$AA,$3A
	.dc.b	$FC,$27,$C5,$3E,$37,$B7,$74,$F7,$FC,$27,$C9,$3E,$57,$C2,$58,$F8
	.dc.b	$7F,$09,$FD,$63,$F8,$07,$F9,$5C,$60,$FF,$49,$B8,$CB,$F9,$43,$FC
	.dc.b	$9F,$CF,$FF,$6A,$3F,$C9,$DB,$F0,$9C,$FE,$13,$A7,$C2,$75,$F8,$4B
	.dc.b	$8F,$47,$C1,$F0,$7E,$1F,$83,$B3,$E0,$7C,$9F,$05,$FE,$AB,$CD,$F0
	.dc.b	$74,$C4,$67,$DA,$9B,$7A,$1E,$1E,$FF,$1F,$83,$E5,$F0,$72,$BD,$B1
	.dc.b	$D7,$E0,$A4,$7C,$07,$C1,$F0,$7D,$23,$E0,$E1,$78,$FE,$0E,$3E,$5F
	.dc.b	$82,$95,$F2,$1F,$47,$C1,$FA,$7E,$0E,$4F,$CC,$7D,$9F,$07,$CF,$8B
	.dc.b	$FC,$41,$C2,$CF,$16,$F3,$12,$CF,$E2,$7C,$14,$AF,$AE,$FC,$7F,$83
	.dc.b	$FF,$69,$72,$FF,$21,$24,$E3,$E3,$F3,$ED,$C8,$FF,$68,$CF,$F2,$2E
	.dc.b	$16,$BF,$E3,$38,$FE,$43,$3F,$DA,$57,$C1,$F6,$F8,$49,$3C,$A3,$C3
	.dc.b	$F0,$7E,$2F,$84,$93,$DA,$3C,$0F,$3C,$AE,$5E,$08,$F8,$18,$F8,$23
	.dc.b	$1C,$08,$7C,$27,$B8,$95,$EA,$F8,$4F,$83,$F0,$7F,$0B,$E1,$24,$FB
	.dc.b	$A3,$BE,$FE,$DC,$EC,$F8,$3E,$13,$E4,$FC,$24,$8F,$8A,$FC,$CF,$84
	.dc.b	$F9,$BF,$07,$27,$E4,$BF,$43,$E1,$3E,$8F,$C2,$7D,$2F,$84,$95,$F3
	.dc.b	$A3,$EA,$FC,$1C,$8F,$A2,$FD,$8F,$84,$FB,$3F,$07,$F6,$BE,$0E,$57
	.dc.b	$D6,$8E,$AF,$D8,$C3,$AB,$F5,$70,$33,$9B,$E4,$FF,$00,$FC,$D8,$FF
	.dc.b	$11,$2F,$8B,$DC,$61,$B3,$C3,$8A,$9F,$E7,$5A,$7C,$2E,$2F,$F2,$D1
	.dc.b	$87,$C1,$3B,$9D,$74,$A9,$F0,$BE,$AB,$FC,$66,$7E,$AC,$7E,$FF,$53
	.dc.b	$4E,$38,$7F,$39,$BE,$A7,$83,$E1,$72,$6B,$E1,$77,$F0,$92,$77,$A7
	.dc.b	$86,$39,$FC,$2F,$4F,$84,$EB,$F0,$B2,$F7,$92,$7C,$3F,$07,$E2,$F8
	.dc.b	$49,$5A,$8F,$2F,$C1,$F9,$BE,$12,$4F,$28,$EC,$F4,$C7,$B3,$F4,$B0
	.dc.b	$F6,$FC,$1F,$BB,$E1,$24,$F8,$A3,$E0,$FC,$27,$C2,$F8,$49,$3E,$68
	.dc.b	$EA,$F3,$3D,$37,$F2,$3E,$13,$E4,$FC,$27,$CA,$F8,$49,$5E,$E8,$FA
	.dc.b	$BF,$C0,$3F,$CF,$FA,$7F,$5E,$DE,$DF,$F5,$A3,$FD,$05,$BF,$13,$FE
	.dc.b	$A0,$8F,$E8,$51,$AF,$84,$C7,$C2,$E5,$D7,$C2,$EF,$E1,$6D,$7D,$4B
	.dc.b	$FF,$31,$37,$BF,$A7,$C2,$75,$6C,$6F,$1E,$8D,$39,$63,$E3,$F8,$59
	.dc.b	$1D,$7D,$DB,$BF,$FA,$07,$2F,$E5,$53,$E1,$F7,$3D,$5F,$09,$EB,$7D
	.dc.b	$31,$D0,$ED,$5C,$71,$F8,$1F,$0B,$27,$D5,$F0,$B6,$BB,$7C,$2F,$C5
	.dc.b	$3E,$37,$C2,$DA,$F8,$1F,$0B,$F2,$4F,$95,$F0,$B6,$3E,$27,$C2,$FC
	.dc.b	$DF,$84,$F9,$CD,$8F,$91,$F0,$9F,$44,$FF,$68,$FF,$00,$FF,$00,$8C
	.dc.b	$9D,$89,$9F,$C1,$C3,$B3,$FE,$A3,$97,$C2,$F2,$39,$FB,$B6,$FF,$53
	.dc.b	$3F,$E6,$46,$FB,$BE,$0F,$84,$F0,$BD,$3B,$BF,$13,$B7,$B2,$BC,$31
	.dc.b	$FC,$EA,$DF,$0B,$E8,$F8,$4F,$4B,$6B,$BF,$C2,$FB,$3E,$17,$DA,$D8
	.dc.b	$F1,$FC,$2F,$C0,$F8,$4F,$82,$D8,$F3,$FC,$2F,$C4,$3E,$2F,$C2,$D8
	.dc.b	$F5,$FC,$2F,$C8,$F8,$5F,$92,$D8,$EA,$7C,$0C,$3A,$B4,$F8,$7F,$0B
	.dc.b	$F4,$3D,$DF,$A2,$D8,$F8,$FE,$F7,$D4,$F8,$5F,$AA,$D8,$EB,$D1,$F4
	.dc.b	$70,$EA,$D8,$F9,$FF,$0B,$F7,$3E,$13,$EE,$B6,$3E,$9F,$C2,$FE,$00
	.dc.b	$FE,$D1,$CD,$FD,$9B,$DF,$FF,$33,$F3,$FF,$9B,$2B,$F9,$0C,$7F,$3F
	.dc.b	$6F,$C2,$F0,$7F,$A3,$FC,$0A,$7F,$9E,$C6,$8E,$3F,$09,$BD,$3F,$C2
	.dc.b	$93,$FC,$EA,$7A,$FC,$26,$6F,$95,$DA,$7C,$25,$30,$DB,$ED,$87,$FB
	.dc.b	$D6,$3C,$CD,$F8,$5E,$59,$73,$F8,$59,$7E,$CF,$6E,$DF,$1F,$80,$F0
	.dc.b	$7C,$1F,$85,$B5,$DB,$83,$E3,$63,$E5,$CB,$AC,$FC,$6F,$84,$F4,$1E
	.dc.b	$9F,$83,$B1,$F2,$BE,$13,$D9,$F0,$7E,$D6,$C7,$CE,$F8,$4F,$80,$7C
	.dc.b	$1F,$83,$B1,$F4,$BE,$13,$E2,$7C,$27,$C5,$6C,$7D,$6F,$83,$F9,$07
	.dc.b	$C9,$F8,$4B,$1D,$49,$C5,$F1,$F0,$EA,$B8,$3F,$C4,$FF,$C0,$3F,$CE
	.dc.b	$63,$FC,$94,$DE,$93,$FC,$EE,$33,$F0,$BF,$66,$3F,$9D,$C7,$0F,$6F
	.dc.b	$1F,$03,$97,$BE,$1F,$8F,$DF,$DD,$9F,$92,$3C,$3F,$0B,$CC,$E9,$F0
	.dc.b	$B6,$7F,$B9,$E5,$FE,$3B,$6F,$C2,$78,$7E,$17,$C4,$F8,$FE,$17,$C9
	.dc.b	$F0,$B6,$B8,$BE,$7F,$83,$91,$D3,$E1,$2F,$07,$C3,$8F,$CE,$E4,$F8
	.dc.b	$4F,$77,$C2,$7B,$DB,$3E,$4C,$BB,$B4,$EE,$7A,$3E,$13,$E3,$1F,$1F
	.dc.b	$E1,$2C,$7B,$3D,$CF,$94,$7C,$BF,$81,$B1,$CF,$3F,$21,$F5,$36,$F6
	.dc.b	$38,$E8,$FA,$B1,$F4,$FE,$0F,$EA,$7C,$25,$BF,$92,$FD,$7F,$84,$91
	.dc.b	$FE,$1E,$FE,$01,$FE,$85,$8F,$3E,$15,$7E,$D6,$7F,$9F,$D6,$FE,$17
	.dc.b	$1F,$0B,$97,$5F,$09,$BF,$85,$B5,$D5,$E5,$F0,$92,$3C,$1F,$0B,$74
	.dc.b	$F5,$65,$D9,$B8,$F6,$AF,$01,$E7,$F8,$4F,$27,$C2,$F9,$5B,$1E,$BF
	.dc.b	$85,$F4,$1E,$9F,$76,$C7,$2C,$3A,$BF,$07,$6F,$53,$F0,$FE,$17,$E0
	.dc.b	$1F,$07,$E1,$6C,$7C,$7F,$84,$F8,$9F,$0B,$F1,$5B,$1F,$2F,$E1,$3E
	.dc.b	$47,$C2,$FC,$96,$C7,$CF,$F8,$4F,$98,$7C,$DF,$85,$B1,$F4,$FE,$13
	.dc.b	$E8,$7C,$2F,$D1,$6C,$75,$3E,$C6,$1D,$5A,$7D,$BF,$85,$FE,$59,$F7
	.dc.b	$1F,$EB,$48,$C1,$FE,$5C,$71,$C1,$F8,$97,$F9,$4F,$9D,$FE,$A8,$6C
	.dc.b	$FF,$3E,$7F,$9E,$CE,$FD,$DF,$5B,$F6,$30,$F1,$36,$BB,$7C,$27,$8B
	.dc.b	$E1,$7C,$6D,$8F,$17,$C2,$FF,$9F,$B1,$FE,$95,$7E,$5F,$85,$C4,$E7
	.dc.b	$E1,$64,$FA,$5E,$1F,$0B,$C7,$E1,$24,$FB,$5E,$9F,$0B,$D7,$E1,$3B
	.dc.b	$7C,$24,$AF,$85,$1D,$1F,$85,$85,$E1,$F8,$58,$F9,$7E,$0E,$57,$C9
	.dc.b	$7D,$1F,$09,$E9,$F7,$24,$FC,$D7,$D9,$F0,$1C,$F8,$7E,$56,$56,$BF
	.dc.b	$DE,$DC,$3F,$E8,$77,$EA,$7C,$1F,$C3,$23,$F5,$FE,$13,$E3,$7C,$27
	.dc.b	$C7,$F8,$49,$3F,$6E,$3F,$D4,$FE,$3B,$E1,$2F,$C5,$4A,$FF,$4D,$CE
	.dc.b	$5B,$F0,$BC,$3E,$16,$46,$0D,$FC,$2F,$42,$3C,$BE,$17,$BE,$3D,$63
	.dc.b	$F8,$96,$0C,$73,$1E,$0F,$84,$F2,$FC,$2C,$8C,$1E,$4F,$81,$F4,$FB
	.dc.b	$92,$30,$7A,$31,$ED,$D5,$FE,$FC,$CB,$B4,$E4,$F6,$FC,$2F,$C2,$23
	.dc.b	$F0,$3E,$17,$E2,$FC,$24,$8F,$86,$FC,$8F,$85,$F9,$3F,$0B,$27,$E3
	.dc.b	$BF,$33,$E1,$7E,$6F,$C2,$C9,$F9,$6F,$D0,$F8,$5F,$A3,$F0,$92,$7E
	.dc.b	$7B,$F5,$3E,$13,$A4,$7C,$6C,$2E,$DE,$BC,$7E,$CF,$C1,$CA,$FA,$EF
	.dc.b	$FB,$8B,$EE,$7F,$DD,$F1,$FC,$FD,$FF,$38,$BF,$ED,$66,$B9,$F7,$F8
	.dc.b	$7B,$9D,$BC,$3F,$F1,$BF,$1B,$3C,$5B,$E1,$92,$CF,$C1,$F0,$B2,$BB
	.dc.b	$3E,$3F,$85,$F2,$7C,$2C,$9F,$13,$E7,$F8,$5F,$46,$1E,$9F,$84,$95
	.dc.b	$E7,$8F,$67,$C2,$7B,$7E,$12,$4F,$AE,$3E,$07,$C2,$7C,$1F,$84,$93
	.dc.b	$EF,$8E,$AF,$C1,$C3,$AB,$F1,$70,$F9,$1F,$07,$F2,$7E,$12,$4F,$C7
	.dc.b	$8F,$99,$F0,$92,$3E,$53,$F3,$FD,$EE,$B1,$F3,$F0,$EB,$D2,$E5,$FF
	.dc.b	$30,$F4,$7C,$8A,$FA,$5F,$09,$F5,$C8,$FD,$5F,$84,$FB,$5F,$07,$F6
	.dc.b	$FE,$12,$4F,$D9,$8F,$BD,$F0,$88,$99,$77,$7E,$B7,$DD,$57,$FC,$00
	.dc.b	$D8,$D7,$5B,$35,$F3,$BD,$FC,$2D,$BD,$E0,$F8,$8C,$6D,$FE,$7C,$8C
	.dc.b	$5F,$8A,$31,$7A,$8F,$1F,$C2,$DC,$D6,$F8,$6C,$36,$B9,$1C,$0F,$CC
	.dc.b	$67,$AC,$64,$60,$DD,$F4,$F8,$5E,$64,$8C,$DF,$F4,$87,$E1,$FB,$9E
	.dc.b	$0F,$81,$91,$A8,$F9,$AF,$0C,$3B,$BF,$AB,$C7,$CF,$F0,$B2,$3A,$BE
	.dc.b	$AF,$85,$F5,$FC,$27,$B3,$E1,$65,$78,$A3,$DF,$F0,$BF,$03,$E1,$64
	.dc.b	$F9,$A3,$E1,$FC,$2F,$C4,$F8,$59,$3E,$A8,$F8,$FF,$09,$F2,$3E,$12
	.dc.b	$4F,$BA,$3A,$BF,$03,$0E,$AF,$E0,$B0,$F9,$FF,$07,$23,$E2,$BF,$4B
	.dc.b	$E1,$3E,$B7,$F0,$0F,$F4,$19,$FF,$24,$37,$AF,$F1,$2D,$E2,$FF,$5E
	.dc.b	$BF,$CB,$A7,$B1,$E1,$21,$83,$9F,$C2,$F6,$F8,$59,$1F,$6A,$53,$E2
	.dc.b	$F8,$59,$39,$7C,$BF,$0B,$E6,$F8,$49,$3C,$5F,$4F,$C2,$7A,$BE,$13
	.dc.b	$D7,$F0,$92,$BB,$47,$BB,$E0,$FF,$D9,$99,$9B,$F0,$78,$73,$F0,$7A
	.dc.b	$F8,$3B,$C1,$E7,$78,$FB,$92,$3D,$5E,$E5,$CF,$E4,$FC,$1F,$CC,$F8
	.dc.b	$39,$1E,$F8,$F9,$FF,$07,$23,$E1,$3F,$4B,$E0,$FE,$9F,$C1,$C9,$F8
	.dc.b	$CF,$EA,$F8,$BF,$9F,$F4,$7E,$F1,$FE,$7A,$DF,$83,$D1,$BF,$83,$E1
	.dc.b	$F0,$73,$0F,$D4,$DA,$3A,$7C,$1F,$5F,$83,$95,$A7,$C1,$F0,$7E,$1F
	.dc.b	$7F,$FD,$E9,$87,$F4,$4E,$2E,$7E,$DE,$27,$3F,$05,$AF,$83,$B5,$E0
	.dc.b	$78,$FC,$14,$8F,$1F,$C1,$DD,$3F,$D3,$19,$3E,$0B,$B9,$E0,$F8,$3B
	.dc.b	$7E,$AF,$83,$F1,$FC,$1F,$91,$B1,$EE,$F8,$3F,$39,$E8,$F8,$3B,$1F
	.dc.b	$0B,$E0,$BD,$7F,$07,$EC,$6C,$7C,$6F,$67,$DE,$7F,$2F,$EA,$FE,$57
	.dc.b	$FC,$03,$FE,$66,$7E,$94,$66,$7F,$9B,$7B,$EA,$FF,$33,$F0,$7F,$6B
	.dc.b	$7F,$CC,$F7,$CB,$E0,$B8,$4E,$7E,$0E,$DF,$1F,$83,$FE,$55,$87,$F2
	.dc.b	$7A,$FF,$5E,$37,$0F,$1C,$EB,$E0,$B7,$F1,$12,$FF,$93,$DA,$79,$FC
	.dc.b	$0C,$9C,$BD,$BE,$0B,$BF,$C1,$49,$E2,$F8,$BE,$0B,$C7,$F0,$3E,$4F
	.dc.b	$82,$95,$DA,$3C,$FE,$EF,$A3,$DE,$93,$E2,$8F,$5F,$C0,$FB,$B5,$7C
	.dc.b	$15,$8C,$39,$F8,$1D,$7C,$15,$E7,$F5,$BC,$7E,$0B,$97,$C1,$59,$F7
	.dc.b	$BF,$2B,$F0,$3F,$E3,$1A,$7F,$86,$DB,$D0,$69,$FC,$6F,$A3,$FB,$3B
	.dc.b	$FC,$52,$7F,$A7,$B5,$FF,$76,$36,$37,$F0,$7F,$D7,$E1,$F0,$54,$95
	.dc.b	$CB,$E0,$B5,$F0,$5B,$F2,$24,$F6,$8E,$5F,$03,$CF,$E0,$A4,$F8,$A3
	.dc.b	$B7,$C0,$48,$F2,$BE,$1F,$81,$F1,$7C,$0F,$8F,$E0,$65,$7A,$A3,$CD
	.dc.b	$F0,$32,$3D,$AF,$A7,$E0,$7D,$5F,$01,$EB,$F8,$19,$5F,$0A,$3D,$DF
	.dc.b	$03,$F3,$4F,$9D,$F0,$37,$A3,$E3,$FC,$0F,$E6,$BF,$04,$FF,$0C,$46
	.dc.b	$0F,$EF,$2E,$30,$74,$7F,$85,$A3,$FD,$FC,$DF,$82,$E5,$F0,$5C,$DE
	.dc.b	$9F,$05,$FE,$0D,$A7,$C1,$53,$11,$9F,$82,$D7,$C1,$4C,$BC,$31,$C7
	.dc.b	$E0,$79,$7C,$14,$9F,$2C,$75,$F8,$29,$1E,$87,$C1,$F0,$5E,$1F,$82
	.dc.b	$93,$EC,$7C,$9F,$05,$E5,$F8,$2F,$37,$C1,$4A,$F8,$31,$E9,$F8,$1F
	.dc.b	$57,$C1,$49,$F8,$B1,$ED,$F8,$2F,$77,$C1,$49,$F9,$31,$F0,$7E,$06
	.dc.b	$47,$CC,$7E,$27,$C1,$7C,$5F,$82,$FF,$19,$7B,$BF,$E4,$3E,$BF,$E2
	.dc.b	$AD,$BF,$F3,$F7,$47,$F9,$11,$FF,$16,$C3,$E0,$B8,$FC,$14,$9C,$C7
	.dc.b	$4F,$82,$91,$C1,$EF,$F0,$DE,$0F,$82,$F0,$FC,$1C,$AE,$B1,$E4,$C7
	.dc.b	$CB,$F0,$72,$7C,$31,$E8,$F8,$3F,$4F,$C1,$C9,$F2,$C7,$B3,$E0,$A4
	.dc.b	$7A,$1F,$7F,$C1,$FC,$0F,$82,$93,$EC,$7E,$1F,$C1,$FC,$4F,$82,$F8
	.dc.b	$BF,$05,$2B,$E0,$C7,$C8,$F8,$2F,$93,$F0,$52,$7E,$2C,$7C,$CF,$82
	.dc.b	$91,$F2,$1F,$9F,$F0,$5F,$E6,$1F,$E0,$1F,$E1,$E7,$F0,$DE,$9F,$E8
	.dc.b	$AD,$AF,$F5,$73,$AF,$86,$E0,$48,$FF,$2F,$C3,$E0,$FA,$7C,$17,$5F
	.dc.b	$82,$95,$A8,$F0,$7C,$0F,$87,$E0,$A4,$F2,$8F,$27,$C0,$F9,$7E,$0A
	.dc.b	$4F,$68,$F4,$7C,$14,$8F,$0B,$EB,$F8,$2F,$67,$C0,$C9,$F2,$BE,$FF
	.dc.b	$82,$F8,$1F,$05,$F0,$7E,$0A,$57,$AA,$3E,$27,$C0,$FC,$5F,$82,$93
	.dc.b	$EE,$8F,$91,$F0,$5F,$27,$E0,$A4,$FC,$28,$F9,$9F,$05,$23,$E2,$BF
	.dc.b	$E8,$FF,$E0,$1F,$F1,$2F,$47,$F7,$27,$FC,$4A,$DF,$82,$D1,$BF,$82
	.dc.b	$E1,$F0,$3C,$7E,$0A,$63,$FE,$8E,$C3,$FC,$4C,$DF,$81,$ED,$F0,$52
	.dc.b	$30,$6F,$E0,$7C,$5F,$05,$23,$07,$3F,$82,$F3,$11,$ED,$F0,$5E,$9F
	.dc.b	$82,$F5,$7C,$14,$9F,$14,$7B,$7E,$06,$47,$95,$F8,$1F,$05,$F0,$7E
	.dc.b	$06,$4F,$A5,$F8,$9F,$05,$F1,$7E,$07,$FC,$1B,$C0,$F8,$2B,$18,$73
	.dc.b	$F0,$5A,$F8,$2B,$D1,$F0,$DE,$3F,$05,$FE,$16,$FE,$01,$FF,$11,$4F
	.dc.b	$CF,$8C,$39,$F8,$1D,$7C,$15,$8F,$F5,$CD,$7F,$88,$9F,$F1,$3A,$F8
	.dc.b	$2E,$87,$5F,$82,$B3,$C3,$E0,$7C,$1F,$05,$E1,$6C,$74,$F8,$1F,$27
	.dc.b	$C1,$79,$5B,$1E,$0F,$81,$F4,$1E,$9F,$82,$B1,$E4,$F8,$2F,$61,$ED
	.dc.b	$F8,$2B,$1E,$8F,$82,$F8,$07,$C1,$F8,$2B,$1E,$CF,$81,$F8,$9F,$05
	.dc.b	$F1,$5B,$1F,$03,$E0,$7E,$47,$C1,$7C,$96,$C7,$C4,$F8,$2F,$99,$F0
	.dc.b	$5F,$35,$B1,$FE,$E0,$F1,$DF,$17,$FE,$46,$6C,$67,$E3,$31,$39,$F8
	.dc.b	$49,$3C,$7E,$0E,$F8,$BF,$DC,$23,$FC,$AF,$87,$F8,$61,$ED,$F0,$7D
	.dc.b	$7E,$0F,$B3,$DF,$E0,$A4,$F8,$FE,$0E,$E1,$CC,$7C,$43,$17,$C1,$F9
	.dc.b	$FE,$13,$D0,$DC,$3D,$9F,$09,$EB,$3D,$9F,$09,$63,$E0,$7C,$27,$BC
	.dc.b	$F8,$1F,$09,$63,$E2,$7C,$1F,$C3,$3E,$27,$C2,$58,$F9,$1F,$07,$F1
	.dc.b	$FE,$13,$E4,$36,$3E,$67,$C1,$FC,$B3,$E6,$7C,$25,$8F,$A1,$F0,$9F
	.dc.b	$3C,$FF,$3A,$FF,$00,$FF,$86,$7E,$21,$3F,$2A,$73,$F0,$B2,$1C,$BF
	.dc.b	$17,$95,$83,$E1,$B5,$39,$8F,$F5,$A5,$70,$C7,$AC,$FC,$C9,$FF,$0F
	.dc.b	$CE,$4F,$F4,$7B,$C2,$39,$7C,$2F,$9B,$E1,$7C,$FF,$0B,$6F,$85,$73
	.dc.b	$8F,$5F,$C2,$C9,$CE,$7E,$6E,$AC,$5F,$97,$E1,$7E,$09,$F0,$BE,$16
	.dc.b	$DF,$0D,$1F,$09,$F1,$BE,$17,$E3,$B6,$75,$F0,$BF,$28,$F9,$7F,$0B
	.dc.b	$63,$97,$C2,$FC,$E3,$E7,$FC,$2D,$8E,$DF,$09,$F4,$8F,$A7,$F0,$96
	.dc.b	$3C,$5F,$09,$F5,$8F,$AF,$F0,$96,$3C,$DF,$09,$F6,$8F,$F5,$7F,$F0
	.dc.b	$0F,$F1,$AE,$2F,$9D,$18,$F8,$59,$5F,$12,$BF,$8C,$57,$E7,$9F,$F5
	.dc.b	$EF,$29,$FE,$35,$3B,$F8,$2E,$BF,$07,$D9,$EF,$F0,$52,$7A,$7C,$1D
	.dc.b	$C7,$F0,$D8,$3E,$0B,$CA,$79,$BE,$0E,$DF,$8F,$E0,$FD,$27,$AB,$E0
	.dc.b	$EC,$79,$FE,$0F,$DA,$7B,$BE,$0E,$C7,$AF,$E0,$FE,$09,$F0,$BE,$0E
	.dc.b	$C7,$BF,$E0,$BE,$29,$F1,$BE,$0E,$C7,$C3,$F8,$2F,$93,$F0,$7F,$29
	.dc.b	$B1,$F1,$FE,$0B,$E6,$9F,$3B,$E0,$EC,$7E,$3B,$F8,$07,$F8,$CF,$1F
	.dc.b	$E9,$5B,$E6,$FE,$83,$A5,$0F,$84,$FF,$72,$BA,$BF,$F7,$7A,$F8,$3C
	.dc.b	$FC,$1E,$9B,$3D,$FE,$0F,$8F,$C1,$F2,$6C,$78,$FE,$0F,$A9,$DB,$E0
	.dc.b	$EC,$79,$FE,$0B,$C3,$F0,$7E,$26,$C7,$AF,$E0,$FC,$A7,$9B,$E0,$EC
	.dc.b	$7B,$FE,$0F,$D2,$7A,$BE,$0E,$C7,$C3,$F8,$3F,$69,$EE,$F8,$3B,$1F
	.dc.b	$1F,$E0,$BE,$09,$F0,$BE,$0E,$C7,$CB,$F8,$3F,$8A,$7C,$6F,$83,$B1
	.dc.b	$F3,$FE,$0F,$E4,$9F,$C3,$3A,$7E,$76,$FF,$C6,$E7,$E5,$CE,$5F,$E3
	.dc.b	$3D,$3F,$CE,$BE,$0F,$1D,$7F,$CF,$AF,$FD,$6F,$5A,$7F,$8E,$3F,$C2
	.dc.b	$9C,$C6,$FE,$23,$8D,$FF,$44,$6C,$FF,$1C,$9F,$E4,$CF,$F1,$BC,$7C
	.dc.b	$BF,$05,$AF,$82,$DB,$71,$ED,$F0,$5C,$8E,$7F,$05,$63,$C5,$F0,$5D
	.dc.b	$8E,$FF,$05,$63,$CD,$F0,$1E,$23,$C7,$F0,$36,$3D,$5F,$03,$E6,$F8
	.dc.b	$1F,$3B,$63,$DD,$F0,$3F,$CF,$B8,$3D,$2D,$DF,$E6,$B6,$7E,$37,$C0
	.dc.b	$DA,$F9,$5F,$03,$79,$3F,$B3,$B7,$8C,$DF,$81,$C7,$C0,$E7,$E0,$66
	.dc.b	$1F,$3E,$38,$7C,$0F,$3F,$E0,$1F,$F5,$F6,$46,$E5,$FE,$BF,$BF,$EA
	.dc.b	$91,$FE,$BF,$7F,$A0,$AF,$88,$E4,$73,$F8,$3E,$9F,$07,$D7,$E0,$E5
	.dc.b	$F0,$8F,$07,$C1,$F8,$7E,$0E,$4F,$48,$F2,$7C,$14,$8E,$EF,$9F,$E0
	.dc.b	$BD,$1F,$05,$E9,$F8,$29,$5E,$48,$F6,$7C,$14,$8F,$3B,$EF,$F8,$2F
	.dc.b	$81,$F0,$5F,$07,$E0,$A5,$7B,$23,$E2,$7C,$17,$C5,$F8,$29,$3F,$02
	.dc.b	$3E,$47,$C0,$FC,$9F,$82,$93,$F1,$23,$E6,$7C,$14,$8F,$8E,$FC,$FF
	.dc.b	$82,$FB,$5F,$C0,$3F,$E3,$69,$FE,$D6,$D7,$FC,$04,$DF,$0F,$F8,$F3
	.dc.b	$FE,$C3,$EB,$3F,$C6,$E7,$8F,$C0,$74,$3A,$FC,$0F,$6F,$80,$93,$D9
	.dc.b	$F0,$FC,$0F,$8B,$E0,$7C,$7F,$03,$2B,$C7,$1E,$6F,$07,$CF,$F0,$32
	.dc.b	$7C,$F1,$EA,$F8,$1F,$5F,$C0,$C9,$F5,$C7,$BB,$E0,$7D,$FF,$03,$27
	.dc.b	$DF,$1F,$0B,$E0,$64,$7C,$27,$E2,$FC,$0F,$C6,$F8,$1F,$8F,$F0,$32
	.dc.b	$BE,$3C,$7C,$AF,$81,$F9,$7F,$03,$27,$E5,$C7,$CE,$F8,$19,$1F,$39
	.dc.b	$FD,$57,$F0,$0F,$FB,$0A,$3F,$CA,$6F,$FB,$09,$BF,$11,$A3,$7F,$07
	.dc.b	$C3,$E0,$F8,$FC,$44,$AF,$F1,$1D,$FF,$B0,$9F,$F1,$D2,$F8,$3E,$E4
	.dc.b	$8D,$FC,$27,$8B,$E1,$3C,$7F,$09,$27,$9C,$79,$BE,$13,$CF,$F0,$92
	.dc.b	$7B,$C7,$AB,$E1,$24,$78,$9F,$6F,$C2,$7B,$BE,$0F,$DF,$F0,$92,$BC
	.dc.b	$F1,$F0,$BE,$13,$E1,$FC,$1C,$9F,$5C,$7C,$6F,$83,$91,$EE,$7E,$4F
	.dc.b	$C1,$FC,$AF,$82,$93,$F0,$9F,$9B,$F0,$7F,$3B,$E0,$E4,$FC,$67,$F8
	.dc.b	$D7,$B7,$FC,$40,$FE,$19,$1F,$E6,$16,$F9,$FF,$DF,$5B,$F0,$9F,$D3
	.dc.b	$9F,$E2,$13,$FE,$52,$7F,$D8,$8D,$F8,$0E,$A7,$6F,$81,$EF,$F0,$1E
	.dc.b	$0F,$81,$95,$C6,$3C,$7F,$03,$E4,$C2,$4F,$58,$F3,$FC,$07,$A3,$E0
	.dc.b	$64,$F8,$63,$D7,$F0,$32,$3C,$8F,$BB,$E0,$7D,$FF,$01,$F0,$3E,$06
	.dc.b	$57,$A6,$3E,$1F,$C0,$7C,$4F,$81,$93,$ED,$8F,$8F,$F0,$12,$3E,$03
	.dc.b	$F2,$BE,$07,$E5,$FC,$0F,$CC,$F8,$19,$5F,$16,$3F,$A4,$7F,$00,$FF
	.dc.b	$0B,$E6,$FE,$EE,$D9,$FF,$59,$B7,$E0,$8F,$F0,$F3,$F8,$74,$7F,$A5
	.dc.b	$5B,$D1,$FE,$31,$6D,$74,$F8,$0E,$FF,$01,$2F,$FD,$0F,$48,$F1,$FC
	.dc.b	$07,$93,$DF,$F2,$FC,$04,$BD,$C7,$A3,$DF,$91,$C9,$F5,$FC,$07,$B3
	.dc.b	$DF,$93,$D9,$F7,$FC,$07,$C0,$F7,$FE,$0F,$C0,$4A,$F1,$C7,$C4,$F8
	.dc.b	$09,$1E,$67,$E3,$FC,$07,$C8,$F7,$FE,$4F,$C0,$4A,$F5,$C7,$CC,$F7
	.dc.b	$E4,$7B,$9F,$9F,$F0,$1F,$43,$E0,$3F,$BC,$7F,$00,$FF,$0F,$E9,$37
	.dc.b	$99,$3A,$F8,$09,$8F,$FA,$BA,$BF,$8A,$5F,$0F,$81,$E7,$F0,$32,$B4
	.dc.b	$F6,$F8,$1E,$FF,$01,$E0,$F8,$09,$5C,$E3,$C7,$EF,$F9,$3E,$02,$4F
	.dc.b	$78,$F3,$FB,$F2,$3C,$4F,$AB,$E0,$3D,$7F,$01,$27,$CC,$FB,$BE,$03
	.dc.b	$DF,$EF,$FF,$8E,$B8,$0D,$F8,$0C,$39,$F7,$F5,$F0,$17,$C8,$F8,$0F
	.dc.b	$1F,$80,$91,$F0,$FE,$02,$E9,$FE,$E7,$C9,$F0,$1D,$CF,$07,$C0,$5B
	.dc.b	$F9,$5F,$01,$FE,$4C,$FE,$01,$FF,$5D,$75,$7D,$28,$C4,$E7,$E1,$75
	.dc.b	$F0,$D6,$3F,$DB,$F7,$FE,$BC,$B3,$FE,$BA,$CF,$FD,$F0,$BE,$0B,$B7
	.dc.b	$C1,$F7,$7C,$1F,$05,$30,$E7,$F0,$77,$0E,$3F,$05,$E5,$F8,$3F,$33
	.dc.b	$6B,$C3,$F0,$5E,$9F,$83,$F5,$36,$3C,$BF,$07,$ED,$3D,$DF,$07,$63
	.dc.b	$D3,$F0,$5F,$04,$F8,$5F,$07,$63,$DB,$F0,$5F,$17,$E0,$FE,$33,$63
	.dc.b	$E0,$FC,$1F,$C9,$F8,$3F,$94,$D8,$F8,$BF,$05,$F3,$4F,$9D,$F0,$76
	.dc.b	$1C,$2F,$83,$FF,$5C,$79,$1F,$EB,$58,$74,$A3,$0D,$F8,$9F,$F4,$1C
	.dc.b	$60,$FF,$5D,$5F,$4B,$E7,$F0,$9A,$F8,$6D,$BC,$3E,$17,$8F,$C2,$DB
	.dc.b	$E0,$7F,$A6,$30,$47,$6F,$84,$93,$AD,$3C,$B4,$F8,$5C,$C7,$AE,$79
	.dc.b	$7C,$2F,$9B,$E1,$7C,$ED,$8E,$5B,$FB,$7E,$17,$D6,$7B,$3E,$16,$CE
	.dc.b	$B2,$7C,$27,$F9,$FD,$7C,$2D,$31,$19,$F8,$5D,$7C,$2C,$D7,$94,$71
	.dc.b	$F8,$49,$1D,$5E,$9F,$0B,$D7,$E1,$7E,$69,$F3,$BE,$16,$E9,$E3,$F8
	.dc.b	$4F,$A2,$7F,$AD,$3F,$80,$7F,$D8,$58,$FC,$48,$C7,$C2,$CA,$F7,$D7
	.dc.b	$FA,$FA,$BE,$31,$F7,$9F,$F5,$FF,$11,$FF,$14,$75,$FF,$BE,$D7,$C1
	.dc.b	$FF,$A7,$DB,$5D,$3E,$13,$1F,$07,$9F,$84,$93,$E0,$8E,$1F,$07,$C7
	.dc.b	$E1,$24,$F9,$23,$A7,$3C,$8F,$3B,$DF,$E1,$3C,$1F,$09,$27,$D6,$F8
	.dc.b	$FE,$13,$C9,$F0,$72,$7D,$EF,$9F,$E1,$3D,$1F,$09,$F1,$CF,$91,$F0
	.dc.b	$97,$93,$E2,$FC,$27,$CB,$F8,$4F,$98,$D8,$F9,$3F,$07,$F3,$CF,$A1
	.dc.b	$F0,$76,$3E,$6F,$C1,$FC,$52,$69,$F8,$EF,$D1,$7F,$D7,$67,$F8,$A1
	.dc.b	$FD,$6B,$7A,$8D,$3F,$C5,$71,$F5,$A3,$3F,$03,$FE,$CF,$9F,$E3,$93
	.dc.b	$D7,$E0,$71,$39,$F8,$1F,$F7,$BD,$3E,$06,$C6,$1C,$FC,$06,$BE,$06
	.dc.b	$E9,$E7,$78,$FC,$07,$2F,$81,$B3,$EB,$7A,$FC,$0F,$8F,$E0,$66,$DE
	.dc.b	$F8,$F3,$7C,$0F,$9F,$E0,$64,$FC,$38,$F5,$7C,$04,$8F,$8C,$FB,$7E
	.dc.b	$07,$DD,$F0,$1E,$FF,$81,$95,$F2,$E3,$E1,$7C,$0F,$C3,$F8,$19,$3F
	.dc.b	$3E,$3F,$DE,$9F,$A2,$7F,$8A,$4F,$F5,$B1,$3E,$3F,$3F,$05,$AF,$82
	.dc.b	$97,$FC,$12,$BF,$C5,$2F,$F2,$27,$FC,$4C,$E7,$E0,$7A,$9D,$BE,$0B
	.dc.b	$BF,$C1,$49,$E4,$F8,$BE,$0B,$C7,$F0,$3E,$4F,$82,$95,$DE,$3C,$FF
	.dc.b	$05,$E8,$F8,$29,$3E,$38,$F5,$FC,$0C,$8F,$33,$EE,$F8,$2F,$7F,$C0
	.dc.b	$FC,$0F,$82,$95,$EB,$8F,$87,$F0,$52,$3D,$CF,$C6,$F8,$2F,$8F,$F0
	.dc.b	$5F,$23,$E0,$A5,$7C,$38,$F9,$7F,$04,$F0,$FE,$33,$67,$F2,$DC,$1F
	.dc.b	$C4,$BC,$C7,$37,$F8,$AA,$FF,$97,$CF,$F0,$E3,$F8,$97,$C0,$F1,$A4
	.dc.b	$67,$E2,$3A,$4E,$61,$F0,$18,$BF,$EB,$37,$C3,$E0,$B1,$97,$F5,$8B
	.dc.b	$DD,$78,$27,$35,$DB,$E0,$3F,$B0,$DF,$A3,$E0,$6E,$5D,$DB,$87,$6C
	.dc.b	$7F,$D9,$15,$EA,$F8,$0C,$BD,$FE,$06,$E5,$EA,$6D,$2A,$FC,$0E,$3E
	.dc.b	$06,$61,$F1,$23,$7F,$03,$23,$E3,$BC,$BE,$07,$9F,$C0,$C9,$F9,$6F
	.dc.b	$6F,$81,$EF,$F0,$32,$7E,$7B,$E2,$F8,$1F,$1F,$C0,$7C,$2D,$5F,$03
	.dc.b	$63,$0E,$7E,$07,$5F,$03,$78,$FE,$CB,$C7,$E0,$7E,$5F,$F0,$0F,$F8
	.dc.b	$E3,$14,$62,$B3,$F1,$11,$FE,$DF,$5F,$E3,$71,$1C,$BE,$0B,$9F,$C1
	.dc.b	$DA,$DB,$DB,$E0,$BB,$FC,$15,$9E,$6F,$8B,$E0,$A4,$76,$F8,$2B,$C7
	.dc.b	$FC,$A7,$37,$C1,$7A,$3E,$0B,$D2,$DB,$F2,$FC,$0F,$B3,$E0,$BD,$AD
	.dc.b	$8F,$4F,$C0,$FD,$EC,$1F,$05,$4C,$46,$7E,$0A,$67,$F0,$1E,$1F,$05
	.dc.b	$C7,$E0,$79,$7C,$14,$AF,$8B,$1D,$7E,$06,$47,$C8,$7C,$1F,$05,$F8
	.dc.b	$EF,$E0,$1F,$F6,$20,$F8,$7C,$19,$F8,$4D,$7C,$26,$FE,$12,$63,$FE
	.dc.b	$20,$BF,$E3,$BE,$69,$FF,$61,$71,$67,$E1,$3F,$4F,$18,$AF,$EA,$4F
	.dc.b	$F8,$D2,$79,$7C,$1E,$BE,$0F,$6D,$8E,$DF,$07,$CB,$C3,$E6,$D8,$F1
	.dc.b	$7C,$17,$6F,$83,$EE,$D8,$F3,$7C,$1F,$8B,$E0,$FC,$6D,$8F,$57,$C1
	.dc.b	$79,$BE,$0F,$CE,$D8,$F7,$7C,$1F,$C7,$F8,$39,$C1,$83,$E1,$7C,$17
	.dc.b	$CB,$F8,$39,$18,$3E,$37,$C1,$FE,$F3,$F8,$07,$FD,$8C,$7F,$B1,$F9
	.dc.b	$3F,$D8,$B3,$FE,$4A,$8F,$F6,$19,$3E,$23,$87,$C2,$71,$F8,$4E,$5F
	.dc.b	$09,$2F,$84,$75,$F8,$4E,$DF,$07,$27,$A4,$78,$7E,$0B,$F8,$CE,$4F
	.dc.b	$83,$B1,$87,$3F,$05,$38,$FC,$9F,$07,$71,$FF,$05,$60,$F8,$3E,$7F
	.dc.b	$07,$D1,$B7,$EB,$F8,$3E,$E7,$83,$E0,$EC,$7B,$FE,$0F,$C7,$F0,$7E
	.dc.b	$46,$C7,$C3,$F8,$3F,$39,$E8,$F8,$3B,$1F,$1F,$E0,$FE,$67,$C1,$CD
	.dc.b	$B0,$7C,$BF,$83,$FE,$57,$ED,$FF,$4D,$8F,$F1,$13,$7E,$13,$FA,$47
	.dc.b	$0F,$F1,$09,$FF,$3D,$CF,$F8,$AC,$FF,$14,$CF,$F4,$D6,$E3,$FD,$21
	.dc.b	$D7,$C1,$FF,$44,$7F,$D8,$D1,$CA,$31,$F0,$79,$F8,$39,$3D,$BE,$0E
	.dc.b	$E3,$E4,$F8,$3E,$5F,$07,$CD,$B5,$E4,$CB,$D6,$4A,$F3,$FC,$1F,$BB
	.dc.b	$E0,$BD,$FF,$07,$27,$D7,$1F,$0B,$E0,$7E,$1F,$C1,$49,$F7,$C7,$C6
	.dc.b	$F8,$2F,$51,$EB,$F8,$2B,$9F,$C4,$F8,$1F,$77,$C1,$7B,$DB,$1F,$23
	.dc.b	$E0,$BE,$11,$F0,$FE,$0A,$C7,$FA,$9B,$F8,$07,$FD,$82,$7F,$0F,$71
	.dc.b	$F0,$5F,$DB,$23,$1A,$7F,$95,$DF,$F6,$1B,$7E,$0B,$5F,$05,$B7,$87
	.dc.b	$C3,$CC,$3B,$7C,$15,$BE,$5F,$03,$FD,$B3,$13,$4C,$7C,$16,$7E,$06
	.dc.b	$57,$9D,$E1,$F0,$5C,$7E,$06,$4F,$AD,$E9,$F0,$5D,$7E,$07,$D4,$7A
	.dc.b	$FE,$0A,$EB,$F0,$7E,$07,$DC,$7B,$FE,$0A,$C7,$C5,$F8,$2F,$84,$7C
	.dc.b	$3F,$82,$B1,$F2,$7E,$0B,$D4,$4D,$3E,$67,$C1,$7B,$7E,$0B,$FC,$13
	.dc.b	$FC,$03,$FE,$14,$9F,$F4,$03,$7C,$CF,$F1,$37,$2F,$F1,$09,$CF,$C1
	.dc.b	$70,$F8,$29,$1F,$ED,$CB,$4F,$4F,$81,$FF,$10,$C7,$FA,$EB,$4D,$46
	.dc.b	$3E,$0B,$3F,$03,$AF,$82,$B5,$CD,$E3,$F0,$5C,$BE,$0A,$CF,$77,$AF
	.dc.b	$C1,$7A,$89,$A7,$8F,$E0,$BD,$BF,$03,$23,$CC,$FC,$0F,$82,$F8,$3F
	.dc.b	$05,$F0,$BE,$0A,$57,$AE,$3E,$2F,$C0,$FC,$6F,$82,$93,$EF,$8F,$93
	.dc.b	$F0,$52,$3E,$13,$F3,$3E,$0B,$E6,$FC,$14,$9F,$8C,$FD,$0F,$82,$FA
	.dc.b	$3F,$05,$FE,$3A,$FE,$01,$FE,$1D,$C5,$FE,$4E,$6D,$BE,$15,$D3,$FC
	.dc.b	$46,$3F,$D6,$D3,$FE,$BB,$27,$C4,$73,$F8,$DE,$9F,$0B,$D7,$E1,$65
	.dc.b	$8E,$31,$E1,$F8,$5F,$17,$C2,$C9,$D7,$87,$FE,$67,$C9,$5E,$7F,$84
	.dc.b	$F4,$7C,$2C,$BD,$EF,$E9,$9F,$67,$C2,$FB,$7E,$16,$56,$8F,$66,$3F
	.dc.b	$E2,$67,$1C,$28,$7C,$27,$F9,$A9,$7C,$2D,$8C,$39,$F8,$49,$9F,$2A
	.dc.b	$FF,$13,$B6,$39,$7C,$24,$8D,$7C,$25,$CB,$E5,$FC,$1F,$D3,$23,$CF
	.dc.b	$E1,$3E,$B7,$C2,$7D,$7F,$84,$93,$DE,$3F,$CE,$7E,$33,$E1,$29,$27
	.dc.b	$DB,$1F,$E2,$96,$FC,$1F,$FA,$1A,$1F,$09,$63,$0E,$7E,$13,$5F,$09
	.dc.b	$76,$E6,$F1,$F8,$4E,$5F,$09,$67,$BB,$D7,$E1,$24,$78,$BE,$12,$E9
	.dc.b	$FE,$FF,$D1,$F0,$9E,$3F,$84,$F2,$36,$FD,$3F,$09,$E7,$F8,$4F,$43
	.dc.b	$63,$DB,$F0,$9E,$B3,$D9,$F0,$96,$3E,$0F,$C2,$7B,$FE,$13,$E0,$36
	.dc.b	$3E,$2F,$C2,$7C,$33,$E2,$7C,$25,$8F,$93,$F0,$9F,$38,$9B,$FE,$23
	.dc.b	$F8,$07,$F8,$C4,$FF,$AD,$9F,$E2,$71,$9F,$84,$FF,$53,$C1,$BF,$09
	.dc.b	$8F,$84,$CF,$C2,$4B,$E9,$1C,$3E,$12,$47,$77,$9F,$C2,$74,$F8,$49
	.dc.b	$3E,$37,$BF,$C2,$78,$3E,$12,$4F,$9D,$F1,$F4,$F9,$3E,$13,$FD,$8F
	.dc.b	$B7,$F8,$9F,$7F,$67,$C2,$62,$73,$F0,$72,$7E,$07,$C2,$5C,$9E,$BF
	.dc.b	$FB,$77,$17,$C1,$F4,$3A,$FC,$25,$C3,$E4,$7C,$27,$80,$F0,$FC,$25
	.dc.b	$8F,$99,$F0,$9E,$43,$CB,$F0,$96,$3E,$87,$C2,$7F,$B7,$BF,$80,$7F
	.dc.b	$DE,$D1,$83,$F7,$6E,$35,$F2,$C7,$FB,$AA,$3F,$D6,$71,$AF,$83,$FF
	.dc.b	$6D,$C7,$F8,$A2,$3F,$90,$3A,$78,$98,$F8,$2D,$91,$EF,$F0,$7C,$BE
	.dc.b	$0A,$47,$89,$EB,$F0,$7D,$BE,$0B,$D1,$F0,$7E,$96,$E7,$E8,$F8,$3F
	.dc.b	$61,$ED,$F8,$3B,$1E,$CF,$82,$F8,$07,$C1,$F8,$2B,$1F,$03,$E0,$7E
	.dc.b	$27,$C3,$FC,$56,$C7,$C4,$F8,$2F,$90,$7C,$9F,$82,$B1,$F2,$3E,$0B
	.dc.b	$E6,$7C,$17,$CD,$6C,$7C,$CF,$81,$F8,$E4,$E2,$F9,$FF,$05,$FC,$F5
	.dc.b	$F8,$FF,$08,$FE,$01,$FE,$29,$1F,$CB,$9A,$E9,$BE,$13,$0F,$E8,$D5
	.dc.b	$C7,$E2,$FA,$4E,$6F,$FA,$A5,$FF,$AF,$0F,$E3,$2F,$5F,$83,$DE,$1C
	.dc.b	$A7,$5F,$07,$C6,$F9,$63,$E5,$F8,$3C,$5E,$EA,$E3,$DA,$6D,$7A,$FE
	.dc.b	$0E,$E2,$F8,$3B,$18,$73,$F0,$7A,$F8,$3B,$9F,$C1,$78,$FC,$14,$8F
	.dc.b	$89,$F0,$77,$4F,$89,$F0,$5D,$8E,$FF,$07,$6B,$E5,$7C,$1F,$8B,$E0
	.dc.b	$FC,$6D,$8F,$9D,$F0,$5E,$63,$CF,$F0,$76,$3E,$97,$C1,$FA,$8F,$5F
	.dc.b	$C1,$D8,$FA,$DF,$07,$F0,$36,$FE,$F1,$18,$3F,$B3,$B8,$D7,$EF,$BF
	.dc.b	$E3,$28,$CF,$C1,$71,$F8,$3E,$4F,$3F,$82,$93,$CB,$E0,$EE,$1C,$BE
	.dc.b	$0F,$C0,$78,$7E,$0E,$D7,$83,$E0,$FE,$ED,$7F,$85,$73,$FF,$63,$3E
	.dc.b	$4F,$83,$CF,$67,$F4,$CA,$B4,$F8,$3A,$4B,$F5,$FC,$1E,$BE,$0B,$7F
	.dc.b	$07,$27,$DF,$1C,$BE,$0F,$9F,$C1,$C9,$F8,$71,$DB,$E0,$BB,$FC,$14
	.dc.b	$9F,$8F,$1E,$2F,$82,$91,$F2,$9F,$2F,$C1,$79,$BE,$0B,$E8,$7C,$17
	.dc.b	$D1,$6F,$1F,$D0,$F8,$2F,$C9,$FF,$00,$FF,$1D,$F3,$7F,$37,$C6,$9C
	.dc.b	$D6,$BE,$23,$7F,$09,$63,$FC,$3F,$82,$39,$FC,$1F,$C8,$83,$4C,$7C
	.dc.b	$1E,$7E,$0E,$70,$72,$78,$7C,$1F,$1F,$82,$E5,$F0,$72,$BB,$C7,$5F
	.dc.b	$83,$F4,$FC,$1F,$A9,$BA,$F9,$3E,$0F,$DB,$F0,$7E,$E6,$C7,$A3,$E0
	.dc.b	$BE,$0F,$C1,$FC,$26,$C7,$B3,$E0,$FE,$2F,$C1,$FC,$66,$C7,$C0,$F8
	.dc.b	$2F,$92,$7C,$AF,$83,$B1,$F1,$3E,$0F,$E6,$9F,$3B,$E0,$EC,$7F,$91
	.dc.b	$BF,$80,$7F,$8B,$BF,$C5,$9F,$F1,$67,$0F,$F2,$06,$F6,$7F,$0E,$77
	.dc.b	$1F,$E6,$B5,$F0,$FF,$96,$7F,$8C,$C7,$1F,$84,$C4,$E7,$E1,$35,$F0
	.dc.b	$92,$BC,$51,$C7,$E0,$E4,$79,$5E,$9F,$07,$D7,$E0,$E4,$FA,$5F,$07
	.dc.b	$C1,$F8,$7C,$49,$3E,$D7,$C9,$F0,$7E,$5F,$83,$FE,$43,$B1,$BF,$07
	.dc.b	$87,$3F,$05,$38,$BE,$37,$C1,$DC,$BE,$27,$C1,$F2,$39,$FC,$1D,$AF
	.dc.b	$99,$F0,$7D,$8E,$FF,$07,$63,$E8,$7C,$1F,$88,$FE,$67,$FC,$03,$FC
	.dc.b	$47,$A7,$E7,$46,$3E,$0E,$57,$F9,$F2,$BF,$C6,$0D,$8E,$3F,$07,$FC
	.dc.b	$1A,$5F,$07,$4C,$46,$7E,$0E,$6D,$CD,$E1,$F0,$7C,$7E,$07,$97,$C1
	.dc.b	$CA,$F0,$47,$5F,$82,$91,$E3,$7C,$1F,$05,$E1,$F8,$1F,$17,$C1,$4A
	.dc.b	$F4,$47,$97,$E0,$64,$7A,$DF,$47,$C1,$7A,$7E,$07,$D5,$F0,$52,$BE
	.dc.b	$04,$7B,$7E,$0B,$DD,$F0,$52,$7E,$24,$7C,$1F,$81,$F8,$5F,$05,$27
	.dc.b	$E4,$47,$C5,$F8,$2F,$EC,$3E,$CF,$F0,$FE,$9F,$EE,$BD,$1F,$C3,$DF
	.dc.b	$F5,$4C,$66,$FF,$D9,$0B,$E0,$B8,$9C,$BE,$0A,$D7,$0F,$81,$EA,$76
	.dc.b	$F8,$1B,$1D,$3E,$07,$C2,$78,$BE,$06,$C7,$83,$E0,$3C,$A7,$9B,$E0
	.dc.b	$6C,$79,$3E,$07,$D2,$7A,$B4,$B1,$E8,$F8,$0F,$69,$EE,$F8,$1B,$1E
	.dc.b	$CF,$80,$F8,$3F,$03,$F0,$9B,$1F,$03,$E0,$3E,$29,$F1,$BE,$06,$C7
	.dc.b	$C4,$F8,$1F,$EA,$79,$1A,$63,$E0,$73,$F0,$33,$9B,$E5,$BC,$3E,$07
	.dc.b	$FA,$DF,$F0,$0F,$F0,$AF,$84,$34,$9F,$0F,$AF,$83,$98,$7F,$B7,$27
	.dc.b	$F8,$7B,$7E,$13,$9F,$C1,$F4,$F8,$49,$3B,$8E,$FF,$09,$23,$93,$E2
	.dc.b	$F8,$4F,$1F,$C2,$79,$3E,$12,$57,$78,$F3,$FC,$24,$8F,$13,$EA,$F8
	.dc.b	$4F,$5F,$C1,$7B,$3E,$0E,$57,$9E,$3D,$FF,$07,$F0,$3E,$0E,$4F,$AE
	.dc.b	$3E,$1F,$C1,$FC,$4F,$83,$93,$EF,$8F,$8F,$F0,$72,$3E,$13,$F2,$BE
	.dc.b	$0F,$E5,$FC,$17,$F8,$1B,$F8,$07,$FD,$73,$4E,$94,$61,$F9,$F3,$BF
	.dc.b	$83,$E1,$F0,$F2,$C7,$F9,$82,$FF,$D8,$2D,$F8,$3E,$C4,$77,$F0,$BF
	.dc.b	$E5,$29,$FE,$23,$18,$F0,$FF,$CC,$30,$F8,$4F,$31,$27,$47,$9E,$1F
	.dc.b	$09,$EB,$F8,$3F,$67,$C2,$4A,$DE,$DE,$C9,$F8,$1F,$09,$27,$38,$7F
	.dc.b	$89,$5B,$F0,$9F,$17,$E0,$E4,$69,$F9,$1F,$09,$F2,$7E,$12,$4F,$27
	.dc.b	$E6,$7C,$27,$CD,$F8,$39,$3D,$9F,$A1,$F0,$9F,$47,$E0,$FF,$CD,$B9
	.dc.b	$9B,$F0,$98,$73,$F0,$73,$93,$CF,$F0,$97,$2F,$A1,$F0,$9C,$8F,$F3
	.dc.b	$BF,$F0,$0F,$F1,$3E,$4F,$B7,$18,$F8,$4F,$F2,$2C,$FF,$AF,$7A,$7F
	.dc.b	$D0,$6D,$C7,$FD,$22,$FF,$B0,$EC,$FF,$89,$30,$F9,$6B,$E0,$B9,$9D
	.dc.b	$3E,$0F,$AF,$C1,$F6,$F8,$39,$7C,$23,$C3,$F0,$7E,$6F,$83,$F3,$B7
	.dc.b	$3E,$BF,$07,$EA,$F7,$7D,$6D,$8F,$0F,$C1,$FB,$8F,$7F,$C1,$D8,$F2
	.dc.b	$FC,$17,$B8,$99,$FA,$3E,$0F,$E0,$FC,$1F,$C2,$F8,$39,$3E,$C8,$F8
	.dc.b	$BF,$07,$23,$DE,$FC,$8F,$83,$F9,$3F,$07,$F3,$8F,$9F,$F0,$77,$3F
	.dc.b	$C7,$F8,$AF,$8C,$FF,$54,$AF,$84,$B1,$87,$3F,$09,$AF,$83,$B7,$C9
	.dc.b	$E3,$F0,$5F,$E7,$CC,$0D,$31,$F0,$79,$F8,$3D,$7C,$1C,$E0,$F1,$C7
	.dc.b	$1F,$82,$91,$E6,$7A,$7C,$1F,$5F,$83,$93,$EA,$7C,$1F,$07,$E1,$F8
	.dc.b	$2F,$17,$C1,$CA,$F7,$C7,$97,$E0,$A4,$7C,$27,$D1,$F0,$7E,$9F,$82
	.dc.b	$93,$F1,$9F,$67,$C1,$FB,$7E,$0E,$4F,$CA,$7E,$07,$C1,$FC,$1F,$83
	.dc.b	$93,$F3,$9F,$F4,$FF,$F0,$0F,$F1,$73,$F8,$8F,$9F,$FE,$DE,$6F,$47
	.dc.b	$FB,$B1,$B5,$FC,$7D,$7C,$1F,$2F,$83,$E7,$F1,$12,$BF,$84,$D2,$7B
	.dc.b	$FC,$27,$83,$E1,$25,$6A,$3C,$7F,$07,$FE,$E8,$B3,$7E,$13,$0E,$7E
	.dc.b	$12,$69,$DB,$E1,$2E,$2D,$BD,$5E,$07,$3F,$84,$E8,$DB,$F2,$FC,$27
	.dc.b	$7F,$84,$F0,$36,$3D,$3F,$09,$E3,$F8,$4F,$23,$63,$DB,$F0,$5E,$7F
	.dc.b	$83,$F4,$36,$3F,$97,$E8,$FB,$07,$86,$FC,$7B,$FA,$0F,$C4,$BD,$AF
	.dc.b	$E8,$77,$BA,$D3,$FE,$AF,$8C,$DF,$2B,$E3,$18,$AD,$DF,$3B,$FE,$9B
	.dc.b	$7B,$79,$18,$9D,$D7,$1B,$E7,$39,$6B,$E3,$6D,$73,$BE,$D3,$93,$A9
	.dc.b	$5C,$C7,$86,$3D,$A6,$CA,$E5,$F1,$1A,$34,$60,$DD,$FC,$46,$D6,$DD
	.dc.b	$9E,$93,$94,$F5,$70,$7F,$70,$FC,$80,$40

;--------------------------------------------------------------------------------
;	FTANH.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
ftanh_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FTANH.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FTANH.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	ftanh_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	ftanh.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTANH.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	ftanh.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTANH.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
ftanh_data_compressed::
	.dc.b	$00,$02,$84,$2C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CF,$ED,$1F,$8A,$7F,$87,$A7,$FC,$D0,$5F,$8A,$E1
	.dc.b	$F0,$FC,$7D,$FE,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$1F
	.dc.b	$07,$9F,$DB,$F4,$7B,$7F,$DF,$7E,$6F,$FA,$B0,$F8,$8C,$7C,$46,$7E
	.dc.b	$23,$5F,$11,$BF,$88,$E1,$F1,$1C,$7E,$17,$97,$C0,$F3,$F8,$1E,$9F
	.dc.b	$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03
	.dc.b	$E3,$F8,$1F,$27,$C0,$F9,$7E,$07,$CD,$F0,$3E,$7F,$81,$FF,$0A,$7C
	.dc.b	$E7,$E1,$BE,$13,$FC,$40,$BE,$07,$3E,$66,$BE,$07,$7F,$03,$C3,$E0
	.dc.b	$78,$FC,$0F,$2F,$81,$E7,$F0,$3F,$63,$D2,$EB,$F0,$3D,$BE,$07,$BF
	.dc.b	$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93,$E0,$7C,$BF
	.dc.b	$03,$E6,$F8,$1F,$3F,$C0,$FA,$3E,$07,$FC,$59,$F3,$BF,$88,$E9,$FF
	.dc.b	$1C,$7C,$07,$F0,$A5,$1F,$EF,$63,$FD,$60,$7F,$9A,$0F,$EB,$47,$F1
	.dc.b	$23,$81,$FC,$41,$FF,$21,$2F,$7B,$BF,$BD,$E0,$F7,$BC,$3E,$F7,$8B
	.dc.b	$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD,$E8,$F7,$BF
	.dc.b	$3B,$EF,$FE,$7B,$5F,$CF,$EB,$FA,$0D,$7F,$43,$AF,$BB,$DE,$F7,$FB
	.dc.b	$DF,$E4,$BF,$6F,$FC,$48,$7E,$3F,$8B,$FC,$B0,$BD,$CD,$3B,$F7,$38
	.dc.b	$7B,$9C,$7D,$CE,$5E,$E7,$3F,$73,$A7,$B9,$FB,$0E,$8E,$DE,$E7,$7F
	.dc.b	$73,$C1,$EE,$7E,$D3,$AF,$F6,$BD,$7F,$B6,$EB,$FD,$BF,$5F,$EE,$3A
	.dc.b	$FF,$73,$D7,$FB,$AE,$6F,$DD,$F3,$7F,$98,$FF,$15,$FF,$1E,$1E,$9E
	.dc.b	$73,$E9,$7F,$CE,$4F,$FB,$38,$FF,$44,$1F,$DE,$0F,$E6,$07,$F2,$F7
	.dc.b	$8F,$0F,$5E,$1E,$DC,$1D,$F8,$3E,$47,$BB,$F2,$7D,$DF,$95,$EE,$FC
	.dc.b	$BF,$77,$E6,$7B,$BF,$37,$DD,$F9,$DE,$EF,$CF,$F7,$7E,$87,$BB,$FE
	.dc.b	$7B,$F9,$FF,$D4,$70,$1F,$D4,$3F,$E8,$95,$EE,$6F,$DC,$E1,$EE,$71
	.dc.b	$F7,$39,$7B,$9C,$FD,$CE,$9E,$E7,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B
	.dc.b	$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$F3,$7C,$9F,$9D,$D9,$F3
	.dc.b	$FB,$3E,$87,$67,$D1,$EC,$FA,$5D,$9F,$4F,$B3,$EA,$76,$7D,$5E,$CF
	.dc.b	$AD,$D9,$FE,$99,$F4,$FD,$E7,$BB,$B4,$FB,$9F,$F5,$42,$EB,$E1,$D7
	.dc.b	$C7,$AF,$97,$5F,$3E,$BE,$9D,$7D,$7A,$FE,$67,$B7,$F3,$7D,$BF,$9D
	.dc.b	$ED,$F8,$7B,$BE,$87,$B7,$FE,$AB,$FA,$0F,$E0,$1C,$05,$F0,$99,$F8
	.dc.b	$AD,$7C,$AE,$FE,$17,$87,$C0,$F1,$F8,$6E,$5F,$01,$CF,$E0,$3A,$74
	.dc.b	$75,$F8,$0F,$81,$E4,$FC,$1F,$27,$E1,$79,$3F,$0F,$C9,$F8,$9E,$4F
	.dc.b	$8F,$E0,$3C,$9E,$D7,$97,$DA,$F3,$7B,$5E,$7F,$6B,$D1,$ED,$7A,$7D
	.dc.b	$AF,$57,$B5,$EB,$F6,$BD,$9E,$D7,$FB,$1F,$E8,$7F,$D4,$07,$E9,$4F
	.dc.b	$D2,$76,$97,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1F,$67,$D9
	.dc.b	$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BB,$E1,$F7,$7C,$5E,$EF
	.dc.b	$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE,$F9,$FD,$DF,$47,$BB,$F8,$7F
	.dc.b	$57,$D5,$EE,$FF,$B8,$7C,$FF,$D3,$9F,$89,$3F,$11,$D2,$7F,$10,$FF
	.dc.b	$BC,$5F,$F7,$61,$FE,$A5,$3F,$CB,$AE,$FD,$9E,$9E,$CF,$5F,$67,$B7
	.dc.b	$AF,$DF,$CC,$F0,$79,$9E,$1F,$33,$C5,$E6,$78,$FC,$CF,$27,$67,$97
	.dc.b	$7F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$FB,$E7,$E8,$BF,$1E,$7E,$3B,$84
	.dc.b	$17,$B1,$BE,$3E,$1E,$C7,$1F,$63,$97,$B1,$CF,$D8,$E9,$EC,$75,$F6
	.dc.b	$3B,$7B,$1F,$AF,$DF,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E
	.dc.b	$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7A,$7D,$8F,$3F,$3B,$E8
	.dc.b	$FF,$DD,$87,$F3,$13,$F1,$A7,$E3,$1F,$6A,$F5,$F8,$BF,$EA,$83,$FC
	.dc.b	$C1,$C2,$EB,$CF,$AB,$AF,$57,$7E,$AF,$0F,$57,$8F,$AB,$CB,$D5,$E7
	.dc.b	$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E,$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F
	.dc.b	$57,$C7,$EA,$FB,$3E,$93,$F9,$99,$FC,$54,$FC,$69,$F8,$CE,$62,$F7
	.dc.b	$F8,$7B,$FC,$7D,$FE,$5E,$CF,$3F,$67,$A7,$B3,$D7,$D9,$ED,$EC,$F7
	.dc.b	$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9,$EC,$F9,$7C,$FF
	.dc.b	$37,$85,$E7,$F6,$7E,$9F,$C0,$7F,$1A,$3F,$19,$CB,$FC,$A7,$D9,$CF
	.dc.b	$B3,$AF,$67,$7E,$CF,$0F,$67,$8F,$B3,$CB,$D9,$E7,$E6,$74,$F6,$7A
	.dc.b	$FB,$3D,$BD,$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F,$67,$C7,$EC,$F9
	.dc.b	$3D,$9F,$2F,$B3,$E6,$F6,$7F,$13,$F4,$BF,$8A,$E6,$3F,$DE,$4F,$E2
	.dc.b	$A7,$E2,$5C,$FB,$5C,$3D,$DE,$3E,$EF,$2F,$6F,$9F,$B7,$D3,$DB,$EB
	.dc.b	$ED,$F6,$F6,$FB,$FB,$7E,$0F,$6F,$C3,$ED,$F8,$BD,$BF,$1F,$B7,$E4
	.dc.b	$F6,$FC,$BE,$DF,$9B,$DB,$F3,$FB,$7E,$8F,$6F,$F4,$3F,$4D,$FD,$E0
	.dc.b	$FE,$B4,$7F,$12,$3F,$0C,$7E,$17,$97,$F7,$AB,$DA,$E3,$ED,$72,$F6
	.dc.b	$B9,$FB,$5D,$3D,$AE,$BE,$D7,$6F,$6B,$BF,$B5,$E0,$F6,$BC,$3E,$D7
	.dc.b	$8B,$DA,$F1,$FB,$5E,$4F,$6B,$CB,$ED,$79,$BD,$AF,$3F,$B5,$FB,$8F
	.dc.b	$A7,$FE,$BE,$7E,$18,$FE,$1F,$C9,$FC,$49,$FE,$54,$DF,$63,$87,$B1
	.dc.b	$C7,$D8,$E5,$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1
	.dc.b	$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E
	.dc.b	$7F,$63,$F8,$7F,$AF,$F8,$73,$FC,$7C,$7F,$5A,$3F,$0C,$7F,$0D,$E4
	.dc.b	$FE,$56,$BD,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6,$F5
	.dc.b	$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC,$BE
	.dc.b	$BF,$9B,$D7,$F3,$FA,$FE,$8F,$5F,$F9,$3F,$D4,$7F,$8F,$8F,$F1,$01
	.dc.b	$FD,$60,$FE,$18,$7E,$1F,$94,$FE,$1D,$FE,$8A,$FF,$58,$1E,$AF,$4F
	.dc.b	$57,$AF,$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E
	.dc.b	$AF,$93,$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FA,$3D,$5F,$E7,$FF,$53
	.dc.b	$FE,$26,$3F,$86,$9F,$88,$E2,$3F,$88,$6F,$AB,$C3,$D5,$E3,$EA,$F2
	.dc.b	$F5,$79,$FA,$BD,$3D,$5E,$BE,$AF,$6F,$57,$BF,$AB,$E0,$F5,$7C,$3E
	.dc.b	$AF,$8B,$D5,$F1,$FA,$BE,$4F,$57,$CB,$EA,$F9,$BD,$5F,$3F,$AB,$FD
	.dc.b	$5B,$EA,$BF,$87,$1F,$E8,$53,$FC,$42,$7F,$0D,$3F,$11,$C8,$7F,$10
	.dc.b	$FF,$74,$5E,$A7,$3F,$53,$A7,$A9,$D7,$D4,$ED,$EA,$77,$F5,$3C,$1E
	.dc.b	$A7,$87,$D4,$F1,$7A,$9E,$3F,$53,$C9,$EA,$79,$7D,$4F,$37,$A9,$E7
	.dc.b	$F5,$3D,$1E,$A7,$A7,$D4,$FE,$DF,$E2,$7F,$A1,$8F,$F3,$41,$FE,$20
	.dc.b	$3F,$AA,$1F,$89,$3F,$11,$C8,$7F,$C3,$E7,$E2,$1F,$F0,$7A,$F4,$BA
	.dc.b	$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA,$5E,$2F,$4B,$C7,$E9,$79
	.dc.b	$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$A3,$D2,$FF,$03,$FD,$5F,$F9
	.dc.b	$B8,$FE,$AA,$7E,$24,$FC,$41,$E7,$E1,$FF,$11,$AF,$4B,$8F,$A5,$CB
	.dc.b	$D2,$E7,$E9,$74,$F4,$BA,$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA
	.dc.b	$5E,$2F,$4B,$C7,$E9,$79,$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$F8
	.dc.b	$7B,$EB,$3F,$AB,$1F,$EC,$D3,$FC,$D2,$7F,$55,$3F,$12,$7E,$23,$90
	.dc.b	$FE,$21,$FF,$1E,$BF,$EC,$D1,$E8,$75,$F4,$3B,$7A,$1D,$FD,$0F,$07
	.dc.b	$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F,$43,$CD,$E8,$79
	.dc.b	$FD,$0F,$47,$A1,$E9,$F4,$3F,$C6,$FF,$5B,$FE,$CE,$3F,$D6,$07,$F9
	.dc.b	$A0,$FF,$0E,$1F,$C4,$4F,$C4,$9F,$E6,$63,$F1,$1C,$47,$F1,$0F,$F9
	.dc.b	$61,$79,$FD,$BC,$FE,$FE,$7F,$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7
	.dc.b	$F9,$3C,$FF,$2F,$9F,$E6,$F3,$FC,$FE,$7F,$A3,$CF,$FF,$27,$FC,$17
	.dc.b	$FA,$D8,$FF,$0E,$9F,$C4,$4F,$E2,$07,$E2,$78,$3F,$C6,$9E,$7F,$0F
	.dc.b	$3F,$8F,$9F,$CB,$CF,$E7,$E7,$F4,$F3,$FA,$F9,$FD,$BC,$FE,$FE,$7F
	.dc.b	$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7,$F9,$3C,$FF,$2F,$9F,$E6,$F3
	.dc.b	$FF,$CC,$BF,$5D,$FE,$1D,$1F,$27,$8F,$8F,$CF,$C4,$EB,$DE,$DF,$BD
	.dc.b	$C3,$DA,$E3,$ED,$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$FE,$7C,$FA,$F7,$CB,$DF,$94,$C7,$C8,$67,$E2,$B5
	.dc.b	$E7,$6F,$E1,$B8,$7C,$37,$1F,$84,$E5,$F0,$3C,$FD,$EE,$9E,$F7,$5F
	.dc.b	$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27
	.dc.b	$BD,$E5,$F7,$BC,$DE,$F7,$FA,$4B,$EC,$3F,$DF,$27,$F9,$80,$FE,$A8
	.dc.b	$7F,$53,$3F,$08,$7F,$0B,$F7,$FF,$D5,$EB,$DE,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$FD,$53,$EC,$FF,$98
	.dc.b	$4F,$C2,$F1,$FF,$B0,$9F,$F7,$A9,$FE,$61,$3F,$AA,$9F,$C3,$4F,$E1
	.dc.b	$8F,$F8,$77,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8E,$FE,$C7,$83
	.dc.b	$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79,$3D,$8F,$2F,$B1,$E6,$F6,$3C
	.dc.b	$FE,$C7,$A3,$D8,$FF,$60,$FD,$8F,$E1,$B8,$BF,$DB,$2B,$D7,$CF,$AF
	.dc.b	$AF,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D
	.dc.b	$7F,$2F,$AF,$E6,$F5,$FC,$FE,$BF,$FB,$63,$EC,$9F,$35,$7E,$4F,$1F
	.dc.b	$1F,$9F,$8F,$D7,$C4,$EF,$E1,$B8,$7B,$BC,$7D,$DE,$5E,$EF,$3F,$77
	.dc.b	$A7,$BB,$D7,$DD,$ED,$EE,$F7,$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE
	.dc.b	$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$FE,$EE,$FB,$3F,$F5,$31,$F8
	.dc.b	$6F,$38,$BE,$17,$5F,$0B,$BF,$85,$E1,$F0,$3C,$7E,$17,$97,$C1,$73
	.dc.b	$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F,$82,$F0,$7C,$17,$87,$E0
	.dc.b	$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E,$0B,$CD,$F0,$5E,$7F,$82
	.dc.b	$F4,$7C,$13,$9E,$78,$FF,$87,$7F,$BE,$7C,$0E,$5F,$EF,$DF,$03,$67
	.dc.b	$FD,$DC,$7F,$98,$CF,$F0,$D8,$F7,$B5,$EF,$6F,$DE,$E1,$EF,$71,$F7
	.dc.b	$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE
	.dc.b	$1F,$7B,$C5,$EF,$78,$FD,$EF,$37,$DA,$7E,$23,$93,$D8,$FF,$BB,$8F
	.dc.b	$F3,$11,$FE,$61,$3F,$A9,$1F,$C3,$D3,$FC,$3D,$E3,$EC,$74,$F6,$3A
	.dc.b	$FB,$1D,$BD,$8E,$FE,$C7,$83,$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79
	.dc.b	$3D,$8F,$2F,$B1,$E6,$F6,$3C,$FE,$C7,$A3,$D8,$F9,$DF,$6B,$FE,$EF
	.dc.b	$3F,$14,$7E,$27,$87,$EB,$3F,$EE,$F3,$FC,$C6,$7F,$86,$CF,$E1,$E3
	.dc.b	$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD,$3E,$FE,$9F,$83,$D3,$F0,$FA
	.dc.b	$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F,$A7,$E6,$F4,$FC,$FE,$9F,$A3
	.dc.b	$D3,$FC,$37,$DB,$7E,$24,$FE,$1D,$C7,$F8,$D5,$E9,$EB,$E1,$37,$F0
	.dc.b	$9C,$3E,$13,$8F,$AB,$CB,$D5,$E7,$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E
	.dc.b	$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F,$57,$C7,$EA,$F9,$3D,$5F,$2F,$AB
	.dc.b	$E6,$F5,$7C,$FE,$AF,$E7,$78,$7F,$11,$CD,$FF,$00,$7E,$0B,$1F,$05
	.dc.b	$9F,$82,$D7,$C1,$6F,$DA,$E1,$F0,$5C,$7D,$AE,$5E,$D7,$3F,$6B,$A7
	.dc.b	$B5,$D7,$DA,$ED,$ED,$77,$F6,$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F
	.dc.b	$6B,$C9,$ED,$79,$7D,$AF,$37,$B5,$FB,$3F,$B7,$FD,$A7,$39,$FE,$20
	.dc.b	$7E,$1D,$C7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$D1,$EF,$7F,$0B,$FB
	.dc.b	$8F,$E1,$C7,$E1,$B9,$8F,$E1,$9F,$E2,$8B,$DD,$E1,$EE,$F1,$F3,$B9
	.dc.b	$7B,$BC,$FD,$DE,$9E,$EF,$5F,$77,$B7,$BB,$DF,$DD,$F0,$7B,$BE,$1F
	.dc.b	$77,$C5,$EE,$F8,$FD,$DF,$27,$BB,$E5,$F7,$7C,$DE,$EF,$9F,$DD,$F4
	.dc.b	$7B,$BF,$C8,$FE,$E7,$F0,$E3,$DE,$C7,$BD,$9F,$7B,$5E,$F6,$FD,$EE
	.dc.b	$1E,$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$75,$F7,$BB,$7B,$DD,$FD
	.dc.b	$EF,$07,$BD,$E1,$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD
	.dc.b	$EF,$7F,$39,$F6,$BF,$0E,$7F,$55,$3F,$0C,$3D,$DD,$7B,$BB,$F7,$78
	.dc.b	$4F,$1F,$77,$97,$BB,$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7B,$BD,$FD,$DF
	.dc.b	$07,$BB,$E1,$F7,$7C,$5E,$EF,$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE
	.dc.b	$F9,$FD,$DF,$EA,$1F,$75,$FD,$4C,$FC,$31,$F8,$5F,$78,$FE,$15,$FE
	.dc.b	$B4,$BD,$CE,$3E,$E7,$2F,$73,$9F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B
	.dc.b	$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7
	.dc.b	$9B,$DC,$F3,$FB,$9E,$8F,$73,$FB,$47,$DD,$FF,$0C,$3F,$0D,$CB,$FD
	.dc.b	$BD,$7B,$7A,$F6,$F7,$ED,$F0,$F6,$F8,$FB,$7C,$BD,$BE,$7E,$DF,$4F
	.dc.b	$6D,$FC,$03,$0B,$A7,$F7,$BF,$A6,$78,$5F,$DE,$E3,$FB,$D4,$63,$E3
	.dc.b	$BF,$84,$C6,$E7,$4D,$8C,$A9,$E0,$F0,$6B,$FC,$49,$AF,$69,$E6,$F1
	.dc.b	$6C,$6A,$3B,$1C,$E3,$27,$F9,$5F,$EE,$FF,$C2,$52,$F1,$31,$7D,$4F
	.dc.b	$F0,$C4,$EA,$33,$7D,$23,$97,$E2,$0E,$5C,$CF,$8E,$F8,$9C,$A7,$4E
	.dc.b	$23,$93,$DB,$F1,$17,$BB,$CB,$2F,$87,$8F,$9E,$73,$5C,$6F,$B0,$9F
	.dc.b	$34,$66,$B8,$BF,$53,$2E,$99,$FA,$A9,$97,$93,$0E,$57,$F0,$5A,$75
	.dc.b	$8D,$4F,$C8,$AE,$58,$F9,$2B,$67,$92,$7D,$19,$78,$30,$F2,$63,$6B
	.dc.b	$79,$7D,$19,$F4,$63,$E5,$CB,$BC,$78,$F1,$F9,$F9,$F1,$7A,$C7,$03
	.dc.b	$E6,$2C,$FC,$8F,$9E,$3C,$26,$F1,$F0,$E5,$CE,$3E,$B5,$D9,$F8,$0D
	.dc.b	$AE,$99,$DC,$3A,$DF,$94,$F9,$C7,$59,$FA,$77,$EE,$CB,$D9,$87,$3C
	.dc.b	$BD,$B3,$F8,$23,$ED,$61,$D1,$C5,$F8,$30,$E8,$76,$8E,$8F,$B1,$A7
	.dc.b	$D8,$8F,$BB,$7D,$CF,$46,$32,$B9,$D7,$EE,$F2,$F8,$0F,$EC,$DE,$11
	.dc.b	$AB,$E6,$DC,$3C,$D5,$F7,$E3,$F3,$57,$CB,$3F,$B9,$3E,$4B,$F8,$67
	.dc.b	$D3,$7C,$D9,$7C,$27,$91,$F4,$2F,$B5,$F9,$E3,$94,$7C,$1A,$E4,$74
	.dc.b	$7E,$94,$71,$36,$58,$DC,$79,$8E,$A7,$C2,$8F,$80,$DB,$F4,$CF,$09
	.dc.b	$E5,$3D,$A3,$E1,$C7,$67,$EB,$37,$0F,$8B,$5F,$B2,$8F,$9D,$5D,$71
	.dc.b	$F9,$B1,$C7,$1F,$85,$5B,$C3,$99,$E1,$2D,$FB,$AB,$CE,$7B,$8B,$1B
	.dc.b	$AE,$79,$FE,$42,$BD,$F5,$E5,$CB,$EB,$63,$27,$EF,$5F,$07,$F5,$6D
	.dc.b	$AF,$C6,$65,$EC,$C3,$F8,$0E,$1C,$67,$C5,$97,$C8,$C3,$E3,$61,$C7
	.dc.b	$2F,$DC,$C7,$89,$F6,$63,$CE,$3F,$35,$1F,$3E,$FD,$71,$CB,$3F,$8F
	.dc.b	$3C,$31,$FC,$23,$BC,$7C,$17,$FA,$B3,$89,$FA,$AC,$7C,$58,$7E,$F8
	.dc.b	$F1,$1F,$6F,$1F,$14,$FF,$04,$8F,$AC,$7C,$4C,$7C,$2E,$F1,$F0,$BF
	.dc.b	$4F,$0F,$0D,$7E,$68,$F0,$9F,$9C,$C7,$F0,$E7,$D5,$BE,$31,$FB,$EC
	.dc.b	$3C,$33,$F9,$B8,$F5,$57,$43,$F0,$98,$FA,$A3,$F2,$67,$1C,$79,$BE
	.dc.b	$19,$E1,$97,$F4,$BC,$FD,$33,$F0,$F2,$DD,$7E,$3F,$2F,$0E,$7F,$AD
	.dc.b	$3D,$97,$C2,$FD,$59,$78,$73,$FC,$9C,$78,$6B,$E3,$BF,$82,$AE,$18
	.dc.b	$7A,$31,$EF,$97,$E2,$2F,$78,$7D,$9B,$DC,$FE,$5F,$0F,$CB,$57,$D6
	.dc.b	$7E,$1C,$FE,$3E,$BE,$54,$FD,$53,$53,$F3,$F2,$F5,$BF,$B0,$3C,$91
	.dc.b	$B7,$F6,$18,$79,$23,$FA,$4C,$F0,$9F,$B3,$8F,$EE,$5B,$8F,$97,$0F
	.dc.b	$0C,$79,$88,$FA,$EF,$DA,$7E,$64,$F0,$3F,$99,$7C,$F5,$D4,$FA,$EF
	.dc.b	$E8,$8E,$13,$E6,$77,$5F,$26,$3D,$64,$AF,$19,$27,$F2,$C7,$03,$A4
	.dc.b	$7C,$A7,$C6,$76,$8F,$24,$7A,$A3,$F0,$27,$A9,$FC,$43,$C1,$FC,$01
	.dc.b	$B3,$A1,$E3,$7A,$11,$DC,$76,$3E,$33,$B3,$A9,$EF,$3A,$1F,$6F,$3F
	.dc.b	$C0,$4F,$92,$35,$3E,$F7,$AC,$7C,$07,$A1,$FA,$3C,$BF,$1B,$1F,$C7
	.dc.b	$32,$FC,$04,$FF,$1C,$BE,$AF,$AA,$FA,$3F,$A0,$9D,$C7,$F8,$42,$3A
	.dc.b	$E3,$F6,$70,$EB,$1F,$D7,$2B,$78,$7F,$4F,$8D,$E3,$FB,$38,$F9,$51
	.dc.b	$D5,$FA,$CF,$8D,$FA,$0F,$A5,$DD,$FE,$AD,$DE,$1E,$33,$E5,$91,$FA
	.dc.b	$13,$D4,$F9,$73,$FB,$FC,$7D,$F1,$D4,$F9,$73,$C2,$74,$F6,$3F,$82
	.dc.b	$47,$C5,$75,$7F,$A3,$7E,$85,$FD,$D7,$EB,$3E,$39,$FB,$AF,$CC,$AE
	.dc.b	$B1,$E4,$AF,$88,$FC,$6A,$FA,$C7,$C0,$8F,$D0,$57,$59,$F1,$10,$F7
	.dc.b	$3D,$63,$73,$E4,$3C,$64,$9F,$1B,$EA,$3F,$80,$DF,$D8,$8F,$E1,$15
	.dc.b	$E9,$7A,$CE,$DF,$AC,$7A,$5F,$D6,$BF,$2E,$3C,$51,$EA,$8D,$1C,$23
	.dc.b	$A5,$FE,$26,$7D,$C6,$E3,$DD,$3F,$A1,$8F,$C1,$CE,$E3,$EB,$9E,$A3
	.dc.b	$F4,$24,$3C,$51,$EF,$7D,$26,$CE,$87,$92,$7F,$A8,$3C,$0D,$4F,$E3
	.dc.b	$DD,$4F,$E6,$E7,$D3,$7F,$61,$F7,$CF,$E8,$67,$F1,$0F,$F9,$72,$FF
	.dc.b	$1F,$7E,$9A,$D9,$FE,$04,$9E,$A6,$EB,$DF,$7D,$5F,$A1,$3B,$3E,$2D
	.dc.b	$7E,$29,$F1,$C7,$D2,$8F,$44,$76,$8F,$41,$F8,$3A,$FC,$44,$F8,$E3
	.dc.b	$A9,$FE,$05,$8F,$1C,$FF,$5E,$7E,$21,$27,$AC,$F0,$8F,$41,$EA,$3F
	.dc.b	$A3,$BE,$3A,$FD,$34,$7C,$C7,$F2,$AF,$D7,$9F,$13,$E4,$8D,$BF,$60
	.dc.b	$FA,$F5,$F0,$23,$E2,$46,$EB,$F1,$07,$8E,$7D,$15,$F1,$63,$F0,$4F
	.dc.b	$69,$E8,$FA,$9F,$15,$79,$1F,$73,$D9,$DB,$E4,$8E,$91,$C2,$7D,$C4
	.dc.b	$7F,$04,$7E,$58,$F4,$BF,$45,$EA,$EE,$BE,$E3,$E9,$CB,$FB,$CC,$7A
	.dc.b	$AF,$C5,$7E,$93,$D1,$7E,$4B,$FD,$C9,$F9,$07,$E8,$9F,$E3,$B3,$C5
	.dc.b	$1C,$0F,$4B,$E4,$3E,$51,$BA,$FC,$43,$FE,$2E,$34,$48,$F1,$57,$90
	.dc.b	$F8,$84,$AF,$1C,$FE,$0E,$7F,$BE,$BD,$63,$EB,$67,$FE,$2B,$C3,$DE
	.dc.b	$7F,$84,$6B,$78,$7E,$D1,$F4,$61,$D7,$0F,$E1,$58,$7C,$B9,$F1,$BF
	.dc.b	$2C,$FC,$D5,$7A,$8F,$94,$F9,$1F,$D3,$CF,$8C,$FC,$A4,$7E,$29,$FB
	.dc.b	$87,$63,$E2,$C7,$8E,$BC,$4F,$D8,$8D,$BF,$D4,$E7,$F1,$F3,$F1,$5E
	.dc.b	$93,$F5,$CE,$87,$CB,$8D,$3E,$33,$53,$D5,$D4,$FA,$4E,$A7,$BD,$FA
	.dc.b	$D5,$F1,$A3,$51,$E4,$7F,$AC,$BF,$90,$3F,$2B,$3F,$16,$FE,$8E,$3F
	.dc.b	$C5,$B0,$F1,$CF,$87,$0E,$97,$E8,$7F,$12,$7D,$07,$D0,$EA,$38,$4F
	.dc.b	$EA,$E3,$6F,$4C,$3C,$71,$B7,$EB,$9F,$6E,$38,$1B,$AF,$EA,$71,$F8
	.dc.b	$8C,$BF,$AB,$5F,$DC,$9F,$0D,$6E,$3F,$D1,$F5,$FA,$BC,$3F,$AE,$57
	.dc.b	$E9,$27,$ED,$4F,$D4,$C3,$C5,$3F,$C6,$A7,$F8,$74,$FC,$D8,$E0,$F9
	.dc.b	$9F,$C3,$9F,$0E,$FC,$F1,$BA,$FC,$CB,$B9,$FB,$6F,$F3,$27,$EA,$1E
	.dc.b	$C3,$F1,$AE,$E3,$C7,$5E,$D3,$F1,$93,$A3,$81,$D0,$FC,$0D,$78,$9F
	.dc.b	$25,$78,$F3,$FF,$1A,$D7,$D4,$8F,$C2,$CF,$E3,$0F,$92,$FE,$2E,$3E
	.dc.b	$4C,$7B,$9F,$0B,$B7,$E4,$BC,$0F,$C5,$C7,$CF,$3F,$89,$5F,$C2,$8F
	.dc.b	$3B,$EC,$CF,$F8,$06,$3F,$52,$35,$3F,$38,$F5,$D7,$D5,$C7,$D9,$87
	.dc.b	$8A,$BF,$20,$F8,$8F,$65,$78,$8F,$9A,$FC,$9A,$FC,$7D,$7E,$17,$0F
	.dc.b	$F5,$D1,$E2,$8F,$3E,$3F,$A1,$7C,$04,$BF,$96,$4B,$F0,$47,$43,$4F
	.dc.b	$88,$F3,$5F,$A2,$35,$3F,$1E,$75,$1F,$7E,$75,$3F,$91,$8F,$DE,$BB
	.dc.b	$74,$76,$3F,$14,$FD,$37,$AC,$F9,$23,$E4,$9F,$0C,$D1,$D4,$FC,$51
	.dc.b	$F3,$DF,$BE,$FE,$2D,$F9,$0E,$CF,$E1,$0F,$53,$78,$7E,$5B,$3F,$E1
	.dc.b	$35,$FE,$18,$3F,$15,$8F,$CE,$CF,$F6,$F3,$FC,$F7,$2F,$9D,$8F,$97
	.dc.b	$0E,$59,$FE,$0E,$73,$97,$C6,$AF,$5B,$F3,$32,$FC,$CD,$FE,$09,$F8
	.dc.b	$27,$C6,$9E,$99,$FC,$CA,$F5,$C7,$E3,$4F,$81,$9F,$96,$BF,$C6,$F3
	.dc.b	$AB,$FC,$99,$F9,$97,$DF,$7F,$D0,$27,$E8,$CF,$F1,$C7,$F5,$C7,$F9
	.dc.b	$6A,$FC,$2F,$9A,$7E,$EC,$78,$67,$A5,$F8,$63,$FC,$5F,$97,$C5,$BF
	.dc.b	$95,$1C,$89,$87,$C1,$7F,$25,$1F,$01,$FD,$6E,$1E,$57,$E3,$C7,$9A
	.dc.b	$3C,$38,$7B,$31,$FE,$07,$3F,$1E,$FA,$11,$E9,$7F,$1B,$3D,$BF,$83
	.dc.b	$BF,$80,$FE,$54,$F0,$CF,$E5,$5F,$09,$BC,$3C,$38,$7D,$C9,$F0,$12
	.dc.b	$74,$FC,$73,$E5,$BF,$E1,$89,$F6,$61,$F9,$43,$EB,$57,$39,$EF,$1F
	.dc.b	$5A,$B9,$47,$C6,$3C,$A7,$9E,$34,$FC,$C8,$F8,$C7,$08,$FD,$54,$7C
	.dc.b	$18,$FB,$91,$F1,$63,$CE,$7C,$13,$47,$9E,$39,$1D,$1D,$5F,$E0,$30
	.dc.b	$F9,$75,$EB,$C7,$F0,$15,$FE,$C1,$3C,$87,$B1,$FD,$6B,$98,$FD,$96
	.dc.b	$5F,$89,$7F,$AB,$C7,$BC,$FC,$76,$3D,$A7,$F8,$F4,$7F,$8C,$E3,$E4
	.dc.b	$4F,$C2,$3F,$5C,$78,$F3,$F9,$93,$EC,$8F,$F0,$34,$7E,$D7,$2F,$CA
	.dc.b	$C7,$0B,$EA,$F8,$8F,$93,$97,$BE,$7E,$49,$F0,$E3,$E5,$5F,$9A,$32
	.dc.b	$FB,$EF,$B3,$EF,$8F,$C3,$4F,$9D,$E4,$48,$FB,$B1,$F1,$32,$F9,$B9
	.dc.b	$7C,$88,$FF,$1F,$99,$AE,$EF,$BD,$F5,$9F,$AB,$7B,$57,$8E,$3F,$A2
	.dc.b	$9A,$7F,$1F,$8F,$C0,$8F,$F0,$C1,$F9,$89,$FF,$28,$46,$67,$FB,$F9
	.dc.b	$EF,$3F,$0C,$7D,$E8,$F1,$3E,$F3,$D7,$3F,$02,$79,$47,$F3,$A3,$94
	.dc.b	$F6,$BF,$E5,$B8,$7D,$8C,$7E,$85,$76,$AF,$CC,$1F,$A8,$C3,$D4,$FB
	.dc.b	$89,$8F,$C7,$8F,$EA,$31,$F1,$E7,$94,$7C,$F8,$E5,$1E,$77,$96,$3E
	.dc.b	$6B,$FE,$0B,$1F,$82,$CB,$F2,$59,$79,$B0,$F9,$58,$7E,$35,$FF,$04
	.dc.b	$3F,$95,$7C,$B8,$7E,$75,$F9,$B1,$F9,$DC,$3F,$8D,$47,$DA,$9F,$06
	.dc.b	$7F,$C3,$A3,$8D,$FE,$0B,$3F,$C5,$E5,$F8,$29,$F9,$93,$E3,$C7,$E8
	.dc.b	$1E,$33,$FC,$41,$1E,$FC,$FF,$6B,$5C,$67,$F0,$35,$C6,$31,$1C,$4F
	.dc.b	$ED,$53,$F2,$DF,$ED,$F9,$7D,$49,$FD,$5D,$FE,$32,$7E,$56,$5E,$F7
	.dc.b	$FC,$33,$7E,$09,$FA,$67,$A0,$FE,$AE,$E6,$3F,$C8,$95,$F4,$B0,$F2
	.dc.b	$57,$09,$FC,$59,$E8,$78,$E3,$F8,$7C,$FE,$AC,$60,$FA,$B9,$79,$DF
	.dc.b	$AB,$87,$EF,$5F,$E8,$B1,$9B,$F8,$E7,$F2,$49,$FD,$EC,$7D,$68,$FC
	.dc.b	$49,$FA,$B3,$EF,$1F,$03,$0F,$AD,$5F,$0E,$BE,$71,$EB,$7C,$93,$F5
	.dc.b	$AB,$C0,$FC,$EC,$3E,$EC,$FB,$EB,$84,$FF,$6C,$9F,$74,$FE,$1E,$FE
	.dc.b	$2C,$FD,$4C,$3E,$0B,$B9,$F8,$31,$8B,$F8,$13,$ED,$9F,$83,$39,$7F
	.dc.b	$2B,$3F,$02,$3F,$96,$E5,$F6,$EB,$FC,$63,$3F,$A8,$3D,$B1,$9C,$7F
	.dc.b	$4F,$7F,$C6,$EB,$F7,$AF,$FA,$7C,$F8,$35,$F8,$E9,$F8,$27,$F8,$A2
	.dc.b	$3E,$41,$F0,$6B,$A6,$7F,$E2,$1C,$FF,$0D,$87,$F8,$EA,$BF,$14,$FB
	.dc.b	$CF,$99,$8F,$AD,$FC,$8E,$7F,$21,$FB,$39,$7F,$06,$8F,$9F,$1F,$90
	.dc.b	$C7,$BD,$FF,$60,$8F,$8C,$7E,$40,$FD,$26,$5E,$77,$F0,$37,$F0,$B3
	.dc.b	$FC,$0C,$62,$BD,$51,$89,$FA,$33,$F1,$0F,$C7,$BE,$39,$F8,$31,$E3
	.dc.b	$9E,$11,$F5,$0F,$AE,$7E,$04,$F9,$67,$E9,$63,$CB,$9F,$F8,$3B,$3F
	.dc.b	$DC,$47,$C7,$3F,$53,$7F,$8C,$3F,$06,$7D,$BA,$F3,$9F,$D3,$27,$8E
	.dc.b	$1F,$E5,$5A,$E3,$8F,$C6,$7F,$C2,$97,$C5,$F4,$5F,$1C,$BE,$6C,$78
	.dc.b	$67,$C1,$3E,$B8,$F9,$A7,$C3,$3F,$CF,$15,$FC,$8A,$3E,$94,$66,$FF
	.dc.b	$D6,$19,$7E,$82,$3D,$87,$C0,$9F,$F1,$C5,$FF,$52,$3F,$1D,$9F,$E0
	.dc.b	$E7,$F3,$39,$7E,$2B,$0F,$65,$79,$27,$F0,$F7,$FA,$83,$E7,$3F,$8D
	.dc.b	$8E,$71,$87,$E4,$D7,$3C,$FF,$D6,$98,$FD,$8B,$FE,$CF,$87,$F0,$9B
	.dc.b	$FE,$BD,$87,$D8,$9F,$DC,$3C,$CF,$89,$1F,$2E,$3E,$46,$5F,$3E,$3F
	.dc.b	$B0,$E3,$DE,$BF,$D2,$B8,$77,$BF,$ED,$98,$77,$9F,$D8,$47,$A5,$FC
	.dc.b	$34,$7E,$C3,$1F,$C3,$67,$F5,$A3,$E1,$E5,$DE,$BF,$A4,$4F,$EA,$72
	.dc.b	$F4,$E1,$CC,$F4,$9F,$AC,$AF,$B0,$FF,$5D,$CF,$F8,$4E,$5E,$5C,$7F
	.dc.b	$82,$CF,$EA,$DF,$16,$3F,$63,$3F,$C7,$E6,$8F,$F0,$91,$EA,$CB,$E7
	.dc.b	$E7,$E3,$8C,$1F,$22,$BE,$26,$1F,$E6,$5C,$FD,$4F,$EE,$AF,$E6,$63
	.dc.b	$EE,$CB,$C9,$97,$F6,$FA,$FE,$75,$8F,$E6,$4F,$D9,$BF,$4A,$77,$1F
	.dc.b	$94,$AF,$9D,$3F,$BA,$C3,$E0,$E1,$FC,$A6,$BF,$BC,$9C,$E7,$F8,$21
	.dc.b	$FA,$63,$F8,$9E,$3F,$02,$3D,$F1,$CF,$0F,$CF,$61,$FB,$18,$FC,$F1
	.dc.b	$FE,$2F,$AE,$0F,$F9,$D3,$3F,$CF,$4F,$CE,$C3,$DF,$97,$E5,$31,$F1
	.dc.b	$47,$F1,$29,$F4,$DF,$EC,$31,$EF,$7F,$46,$3A,$5F,$E5,$5F,$D2,$9F
	.dc.b	$CB,$9C,$3F,$AC,$CF,$EA,$3E,$98,$FC,$F3,$FB,$6B,$FB,$EF,$A6,$7F
	.dc.b	$21,$5E,$58,$F8,$71,$8A,$FC,$E9,$E8,$3F,$84,$47,$F0,$39,$FF,$55
	.dc.b	$E7,$F4,$27,$FA,$85,$FB,$0F,$E0,$EF,$89,$F9,$47,$8B,$1F,$F4,$3E
	.dc.b	$7F,$29,$FF,$1F,$4F,$B2,$7E,$54,$FE,$32,$7F,$C9,$47,$F8,$83,$0F
	.dc.b	$B6,$7E,$16,$BC,$CF,$F0,$0C,$FF,$43,$18,$8F,$DA,$C6,$1F,$7C,$7E
	.dc.b	$3A,$FF,$28,$ED,$FD,$3C,$7B,$E3,$F8,$EE,$1E,$F8,$FE,$53,$3E,$FC
	.dc.b	$7E,$1E,$7F,$6A,$FF,$17,$5F,$75,$F8,$53,$F7,$71,$F7,$CF,$EE,$6B
	.dc.b	$83,$F8,$03,$15,$C0,$E7,$7F,$32,$3F,$81,$1F,$77,$1F,$C8,$E1,$FD
	.dc.b	$DF,$3F,$C1,$CF,$84,$DD,$7D,$2A,$DD,$FF,$6F,$C7,$F2,$EF,$E0,$2B
	.dc.b	$E2,$9F,$3A,$BE,$2B,$89,$FC,$D5,$7C,$B7,$EA,$18,$AF,$D8,$CE,$DF
	.dc.b	$C4,$D7,$C7,$7E,$5E,$3F,$02,$BE,$19,$FC,$78,$FE,$1F,$1F,$D0,$EB
	.dc.b	$E1,$9E,$AC,$7E,$54,$7F,$74,$8C,$CF,$07,$F4,$C6,$23,$F8,$B6,$3F
	.dc.b	$13,$0F,$95,$5B,$3F,$A6,$D6,$CF,$D7,$B8,$3C,$95,$C8,$F2,$38,$8D
	.dc.b	$C7,$F3,$EC,$7E,$6C,$7F,$76,$CF,$F3,$53,$FE,$C5,$CF,$F9,$2D,$FF
	.dc.b	$67,$8E,$07,$E8,$72,$FB,$4F,$63,$F9,$7D,$F6,$3F,$A0,$CF,$85,$FE
	.dc.b	$05,$1B,$7F,$81,$1F,$1D,$FC,$4C,$FE,$59,$FC,$64,$7E,$3E,$79,$63
	.dc.b	$FB,$0C,$BE,$35,$EE,$B9,$1B,$BF,$1B,$E0,$3E,$7D,$F2,$9F,$4C,$62
	.dc.b	$77,$7D,$5F,$34,$7E,$56,$BE,$94,$7C,$28,$FA,$47,$58,$FA,$51,$CA
	.dc.b	$BA,$C6,$27,$DE,$F2,$7E,$2C,$EC,$FC,$F6,$3F,$4A,$3F,$C5,$79,$FF
	.dc.b	$4C,$7F,$B4,$4F,$E6,$2B,$F3,$6F,$E6,$33,$FC,$F3,$FE,$9A,$30,$6C
	.dc.b	$C4,$FF,$11,$70,$FF,$93,$AF,$F8,$F4,$7E,$5E,$BE,$33,$E4,$CB,$E7
	.dc.b	$BF,$A4,$3E,$AC,$7E,$91,$F5,$57,$E2,$E7,$6F,$FC,$01,$CF,$F8,$A5
	.dc.b	$7E,$EA,$3F,$B9,$63,$ED,$8F,$EC,$91,$FD,$B8,$FD,$A1,$E7,$30,$FC
	.dc.b	$07,$0F,$F9,$B6,$FF,$92,$65,$FD,$DE,$31,$3C,$8F,$E5,$B3,$F0,$1F
	.dc.b	$F5,$63,$FC,$B6,$7F,$A7,$4F,$59,$D9,$E0,$9F,$92,$E2,$3A,$9B,$CF
	.dc.b	$F9,$99,$FC,$26,$3E,$05,$FF,$09,$AF,$98,$7C,$6A,$FF,$1B,$E3,$F1
	.dc.b	$9F,$E6,$18,$FE,$C3,$2F,$A7,$87,$F8,$8F,$3F,$DB,$D7,$0C,$3F,$2B
	.dc.b	$97,$F0,$67,$F8,$05,$7E,$6C,$FC,$7B,$89,$FC,$94,$7E,$6E,$30,$7E
	.dc.b	$6E,$3E,$FC,$7D,$98,$EB,$9F,$EB,$67,$F3,$13,$FC,$EE,$7F,$81,$1F
	.dc.b	$85,$36,$7E,$15,$FD,$9C,$7E,$15,$F1,$9F,$C1,$4E,$47,$EA,$AB,$F6
	.dc.b	$0F,$A4,$F3,$1B,$7F,$A2,$4F,$82,$3C,$C6,$E3,$E1,$3B,$AF,$E6,$2F
	.dc.b	$C9,$9E,$15,$F2,$63,$84,$FB,$67,$E1,$C7,$ED,$0F,$39,$FB,$D7,$E3
	.dc.b	$1F,$64,$FF,$30,$9F,$01,$FB,$EF,$E9,$0F,$BE,$F2,$AF,$6D,$7D,$27
	.dc.b	$CF,$1D,$4F,$01,$F0,$0C,$1F,$1A,$3E,$A9,$E9,$39,$57,$A6,$3E,$05
	.dc.b	$7F,$2A,$3E,$94,$FD,$8A,$F9,$33,$F3,$EB,$F8,$AD,$7B,$23,$CD,$18
	.dc.b	$7C,$73,$E6,$3D,$33,$B8,$F3,$46,$1D,$CF,$8D,$F4,$C7,$82,$BD,$35
	.dc.b	$F2,$27,$E8,$BE,$63,$E8,$BE,$F7,$E7,$3F,$E7,$5C,$3F,$92,$3F,$E5
	.dc.b	$E9,$FF,$1A,$E5,$FE,$43,$8F,$E1,$75,$FC,$08,$FF,$39,$63,$FC,$92
	.dc.b	$3F,$8F,$4E,$EB,$FC,$6F,$5F,$BC,$C3,$F7,$37,$BB,$FE,$0A,$7F,$83
	.dc.b	$71,$FE,$0B,$5B,$7F,$87,$3F,$EE,$48,$FE,$1D,$5C,$A3,$F8,$72,$9F
	.dc.b	$F5,$5C,$7F,$0F,$3F,$C3,$57,$FC,$26,$39,$D7,$FA,$06,$31,$3E,$A3
	.dc.b	$9C,$F2,$9F,$24,$72,$3C,$86,$DF,$E3,$07,$C0,$7C,$0F,$C6,$3C,$F1
	.dc.b	$E0,$7F,$CA,$C6,$23,$94,$FF,$1E,$7A,$C7,$F5,$23,$A9,$F1,$A3,$FC
	.dc.b	$58,$E2,$BE,$04,$FF,$AE,$70,$FA,$0F,$F8,$37,$3F,$E8,$0F,$F9,$D3
	.dc.b	$1F,$A1,$5C,$2F,$FD,$05,$3E,$8C,$3E,$BC,$7A,$CF,$82,$7A,$DF,$8E
	.dc.b	$F9,$5F,$43,$B7,$CA,$61,$F4,$47,$F6,$67,$D1,$8F,$CD,$9D,$CF,$DD
	.dc.b	$7A,$9F,$D5,$8E,$55,$E7,$7F,$27,$1E,$73,$DC,$7E,$4E,$BC,$8F,$33
	.dc.b	$E5,$C7,$37,$11,$E7,$7C,$91,$DA,$BC,$91,$83,$73,$EA,$3F,$B1,$B8
	.dc.b	$AF,$22,$85,$33,$FE,$E7,$9F,$EB,$2F,$FD,$41,$7B,$AD,$65,$FD,$67
	.dc.b	$0F,$E5,$47,$F0,$99,$FE,$80,$FE,$E8,$FE,$42,$7F,$40,$7F,$90,$BF
	.dc.b	$C1,$B1,$FE,$18,$FE,$F4,$FE,$1F,$1C,$AB,$F8,$DB,$FC,$0E,$3F,$8D
	.dc.b	$9F,$E3,$B3,$67,$F4,$08,$DB,$E0,$7F,$8F,$4F,$F4,$08,$FF,$52,$19
	.dc.b	$9F,$E1,$B7,$FC,$4D,$FE,$1B,$5F,$E5,$D9,$FE,$33,$1E,$17,$F8,$CD
	.dc.b	$73,$AF,$EF,$95,$FD,$DA,$FF,$33,$5F,$EA,$EC,$BF,$33,$7F,$5A,$3E
	.dc.b	$99,$FB,$BC,$3E,$F9,$F9,$F3,$F7,$67,$E9,$4F,$C4,$3F,$E4,$07,$F1
	.dc.b	$17,$F8,$58,$E4,$7E,$14,$EA,$7E,$E1,$FC,$AC,$75,$8F,$CD,$BD,$5F
	.dc.b	$D3,$C7,$EA,$9F,$C4,$1F,$E1,$B8,$FC,$44,$FA,$63,$07,$80,$DE,$1F
	.dc.b	$C7,$5E,$A7,$ED,$8E,$B3,$FA,$07,$F8,$51,$FA,$03,$94,$7E,$51,$FC
	.dc.b	$9E,$5F,$92,$CF,$FC,$45,$5F,$91,$8F,$F1,$14,$7F,$08,$9F,$88,$74
	.dc.b	$C3,$E2,$3F,$06,$BE,$6B,$F4,$DF,$9A,$F9,$63,$F7,$6F,$8B,$0F,$76
	.dc.b	$1E,$B9,$FD,$75,$7C,$77,$75,$D4,$FA,$07,$C4,$30,$7C,$4A,$FE,$55
	.dc.b	$5F,$ED,$29,$FA,$73,$E3,$3E,$3B,$FC,$19,$F8,$E7,$B1,$D9,$FE,$65
	.dc.b	$3E,$B4,$7C,$77,$F6,$75,$FE,$0D,$9F,$D9,$C7,$F6,$C7,$FB,$35,$FC
	.dc.b	$A7,$E6,$E5,$F9,$69,$E9,$5F,$55,$EF,$9F,$F0,$1C,$FF,$C9,$B3,$FD
	.dc.b	$D6,$4F,$F0,$9F,$B3,$7E,$4D,$F6,$5B,$5F,$E2,$E7,$3A,$59,$FE,$17
	.dc.b	$97,$8A,$FF,$86,$1C,$E1,$E4,$F1,$FC,$9F,$BC,$EA,$98,$78,$2B,$53
	.dc.b	$BF,$8E,$8E,$E5,$6F,$8B,$CE,$1F,$EE,$79,$CF,$C4,$F7,$3C,$1E,$65
	.dc.b	$C3,$96,$0C,$78,$ED,$F0,$B1,$F1,$CF,$5A,$DD,$7F,$05,$D7,$FC,$65
	.dc.b	$97,$69,$E7,$97,$66,$9C,$09,$87,$F3,$47,$E7,$61,$D6,$30,$78,$30
	.dc.b	$ED,$3F,$C9,$6F,$B4,$E9,$B8,$F9,$DF,$87,$97,$B6,$75,$83,$CC,$FA
	.dc.b	$07,$D1,$F3,$2E,$3F,$46,$BF,$CE,$38,$7C,$CF,$33,$EB,$79,$9F,$5D
	.dc.b	$B5,$F2,$30,$E3,$7D,$B0,$B9,$FC,$98,$E7,$93,$0E,$31,$E4,$6C,$FA
	.dc.b	$65,$E6,$7E,$14,$FC,$37,$99,$72,$F3,$74,$7F,$A0,$35,$FC,$0F,$99
	.dc.b	$F8,$BF,$33,$F1,$8D,$AF,$BB,$87,$1D,$3E,$1E,$37,$3F,$36,$5C,$F3
	.dc.b	$63,$C6,$FE,$4B,$6B,$CD,$E6,$7E,$67,$CC,$FC,$D3,$73,$F2,$F2,$FE
	.dc.b	$13,$17,$99,$F9,$F3,$F4,$1E,$65,$AF,$C8,$DF,$1C,$FE,$BB,$72,$F2
	.dc.b	$E3,$CF,$0E,$15,$F7,$30,$B5,$E4,$C3,$F3,$79,$79,$1A,$79,$0F,$C5
	.dc.b	$65,$E4,$7F,$43,$87,$91,$A7,$E9,$3C,$CF,$DB,$1F,$B7,$EE,$B1,$CF
	.dc.b	$2F,$ED,$C7,$E9,$BC,$CF,$DE,$79,$9F,$BD,$6C,$F3,$24,$FE,$52,$B9
	.dc.b	$BF,$B7,$C5,$5C,$23,$F3,$2D,$CF,$F3,$DF,$94,$3E,$4B,$F8,$92,$F8
	.dc.b	$AB,$18,$7F,$53,$7E,$6E,$75,$3C,$EE,$3E,$77,$26,$EB,$FC,$36,$7B
	.dc.b	$37,$0E,$AD,$3E,$16,$5F,$E0,$56,$9F,$D6,$F5,$E8,$F3,$C2,$E5,$FB
	.dc.b	$9C,$11,$EA,$F3,$A5,$78,$7C,$EB,$7F,$C6,$ED,$E7,$7C,$0F,$3B,$E0
	.dc.b	$B6,$FD,$18,$70,$CF,$E3,$E3,$67,$FC,$89,$8F,$EF,$B1,$E7,$1E,$FC
	.dc.b	$79,$B4,$E6,$49,$F4,$57,$37,$FD,$FB,$B7,$CC,$CB,$C5,$1E,$D6,$E3
	.dc.b	$F6,$3C,$EF,$AE,$7D,$8F,$3A,$E3,$F2,$2F,$FA,$41,$FC,$88,$76,$7F
	.dc.b	$63,$F3,$5C,$1F,$80,$F8,$AF,$C0,$FC,$55,$C3,$EC,$3D,$30,$F9,$D9
	.dc.b	$77,$9F,$B8,$FE,$13,$3F,$B1,$3D,$F2,$ED,$8F,$38,$FA,$D8,$5A,$FC
	.dc.b	$8B,$F9,$0F,$8A,$FC,$8F,$C5,$5C,$BF,$10,$F5,$7F,$04,$7B,$6B,$FB
	.dc.b	$DD,$FE,$07,$D1,$FC,$CF,$C5,$7E,$69,$B5,$F9,$1C,$3A,$B4,$EA,$73
	.dc.b	$DB,$E1,$B7,$3E,$4F,$E6,$6E,$E1,$F8,$F7,$F4,$FF,$15,$FA,$8F,$8A
	.dc.b	$B5,$F9,$F7,$D1,$5F,$4B,$3F,$E2,$33,$FD,$D5,$F4,$9F,$B2,$F8,$AF
	.dc.b	$D8,$9F,$B2,$F8,$AB,$3F,$A9,$C3,$B3,$4E,$C7,$9C,$97,$FA,$3A,$E2
	.dc.b	$FE,$96,$AD,$FF,$00,$BE,$73,$FA,$96,$C7,$9A,$3F,$7D,$D0,$BE,$38
	.dc.b	$7F,$86,$E1,$86,$2B,$FC,$1F,$0F,$8A,$FE,$16,$7F,$87,$78,$7F,$D5
	.dc.b	$57,$FC,$0E,$3E,$2E,$3F,$0A,$3F,$57,$C5,$F8,$87,$FC,$29,$4F,$97
	.dc.b	$E4,$73,$F9,$2B,$3C,$C9,$97,$F7,$BA,$EE,$FF,$05,$C7,$BF,$0F,$DE
	.dc.b	$8F,$CB,$79,$1F,$E1,$5C,$FF,$32,$7F,$80,$6F,$C1,$87,$8B,$2F,$03
	.dc.b	$4F,$2F,$C6,$7B,$4F,$77,$C6,$58,$FA,$77,$F2,$CE,$AF,$F9,$0E,$DF
	.dc.b	$19,$F1,$4F,$8D,$F1,$96,$7A,$93,$3F,$66,$1D,$5F,$F1,$A5,$F2,$9F
	.dc.b	$8F,$7C,$9A,$7C,$AF,$8C,$FA,$5F,$19,$F4,$DB,$1F,$93,$DB,$E2,$67
	.dc.b	$E6,$C7,$FD,$C1,$C3,$F4,$DF,$99,$F1,$9F,$6F,$E3,$3E,$E3,$70,$EC
	.dc.b	$7D,$8C,$BB,$34,$ED,$D5,$FE,$3E,$D5,$5C,$70,$EB,$5F,$80,$C3,$8C
	.dc.b	$7D,$BC,$2E,$5F,$55,$FB,$15,$FC,$12,$7F,$AA,$61,$D2,$39,$1F,$8A
	.dc.b	$C3,$B5,$FD,$DF,$8C,$FC,$A7,$C6,$7E,$55,$B5,$E6,$3F,$C8,$D6,$F8
	.dc.b	$CF,$CD,$7C,$67,$E6,$DB,$3F,$96,$AE,$39,$7E,$45,$B8,$FC,$B3,$F3
	.dc.b	$D1,$F3,$39,$BD,$2F,$F0,$5A,$FC,$25,$F4,$8F,$CD,$36,$BF,$29,$F1
	.dc.b	$9F,$AB,$3F,$59,$F1,$96,$FD,$17,$F0,$BE,$33,$F6,$07,$EC,$7E,$32
	.dc.b	$CF,$EB,$2F,$85,$7E,$9E,$EE,$1E,$7E,$4F,$C1,$DF,$ED,$33,$F3,$CF
	.dc.b	$AB,$CA,$6E,$D3,$AE,$5F,$B1,$D2,$DF,$EA,$F0,$FE,$05,$F1,$1F,$C0
	.dc.b	$DB,$97,$A2,$7F,$21,$7F,$D5,$A7,$BE,$7E,$28,$FF,$4D,$E5,$FE,$A0
	.dc.b	$BF,$F4,$C7,$E5,$3F,$EA,$09,$FE,$01,$F1,$1F,$EA,$B8,$C6,$3E,$1A
	.dc.b	$FE,$17,$97,$85,$A7,$84,$FC,$54,$7E,$DD,$A7,$ED,$EB,$A3,$FC,$6D
	.dc.b	$AF,$F1,$6C,$BD,$78,$7F,$11,$6D,$7C,$BE,$7F,$E2,$27,$1A,$F4,$96
	.dc.b	$BF,$D2,$58,$FA,$23,$FD,$90,$FF,$AA,$DF,$7F,$C3,$7B,$7E,$23,$DC
	.dc.b	$FC,$6B,$FF,$54,$D3,$E2,$3E,$11,$F0,$FE,$22,$E1,$C8,$98,$F8,$31
	.dc.b	$E7,$1F,$0E,$3F,$28,$7E,$5D,$C7,$13,$0E,$D1,$F4,$B0,$EC,$D3,$EA
	.dc.b	$FC,$3F,$D3,$3E,$A7,$C3,$D8,$ED,$F8,$0B,$E3,$5E,$67,$AE,$7F,$DB
	.dc.b	$53,$98,$FB,$9F,$13,$F7,$7E,$2A,$D7,$E0,$9E,$9A,$FC,$AC,$BB,$CF
	.dc.b	$88,$EF,$8F,$66,$E1,$CA,$BE,$8D,$DE,$2F,$F6,$16,$A8,$FC,$67,$C4
	.dc.b	$FE,$37,$E2,$6E,$3F,$92,$7D,$18,$7F,$B4,$35,$F4,$4F,$63,$F2,$FF
	.dc.b	$13,$F9,$53,$F2,$DF,$15,$63,$F3,$78,$75,$69,$D4,$E5,$7F,$ED,$CC
	.dc.b	$FF,$85,$E1,$DA,$3F,$13,$87,$37,$F1,$78,$5A,$F5,$E1,$D6,$7F,$1E
	.dc.b	$D8,$FE,$81,$7C,$DB,$AF,$C7,$AE,$B8,$FA,$8F,$D8,$E1,$D5,$FC,$BF
	.dc.b	$C5,$7E,$C4,$FD,$97,$C4,$D8,$FB,$77,$CF,$04,$30,$63,$E1,$8F,$D8
	.dc.b	$C6,$2B,$84,$7E,$93,$E2,$7F,$7D,$F1,$5F,$BF,$6C,$7F,$04,$CB,$C1
	.dc.b	$7F,$AC,$CA,$E5,$E0,$8E,$9C,$3F,$CC,$B3,$F0,$47,$E8,$1B,$CD,$F9
	.dc.b	$FC,$BA,$CF,$ED,$9B,$7F,$86,$F8,$AF,$E2,$1D,$5F,$C4,$5B,$A7,$C0
	.dc.b	$CB,$B6,$DE,$AC,$7B,$37,$2F,$D6,$7C,$47,$F1,$9F,$8A,$FE,$34,$DC
	.dc.b	$3F,$8F,$E3,$D3,$0F,$E0,$ED,$CF,$D9,$AF,$A8,$F5,$BF,$86,$CB,$B7
	.dc.b	$37,$FB,$13,$0F,$E5,$39,$76,$69,$F4,$EB,$E3,$66,$FC,$AB,$FD,$73
	.dc.b	$9A,$73,$1F,$EC,$7A,$E0,$7D,$BA,$E1,$F1,$5F,$D7,$9E,$5F,$13,$F0
	.dc.b	$B7,$FE,$C3,$F7,$8F,$CE,$3F,$A5,$E1,$F1,$CE,$79,$A7,$5E,$37,$3C
	.dc.b	$FF,$DA,$15,$AC,$BC,$DF,$19,$9F,$8C,$D3,$BD,$78,$6B,$6F,$86,$1E
	.dc.b	$9C,$FF,$71,$C3,$F5,$74,$EF,$5E,$39,$F8,$0F,$E5,$E3,$31,$81,$9F
	.dc.b	$8A,$7A,$E5,$E3,$69,$E0,$CB,$C6,$D3,$C7,$97,$C5,$7E,$0E,$1D,$B7
	.dc.b	$FB,$B9,$79,$76,$F8,$F9,$F8,$73,$F0,$37,$4F,$8A,$DD,$7E,$1E,$5F
	.dc.b	$03,$E3,$3E,$51,$F2,$FE,$2E,$C7,$83,$1E,$B9,$FE,$9F,$4F,$14,$FE
	.dc.b	$DF,$93,$F4,$F1,$8C,$BE,$0E,$7F,$EE,$27,$C5,$F1,$9F,$5F,$E3,$3E
	.dc.b	$C3,$DA,$3C,$DA,$78,$71,$F2,$1E,$1E,$3F,$C2,$E9,$E2,$8F,$07,$07
	.dc.b	$C1,$9E,$91,$F6,$B0,$B8,$78,$23,$EE,$FC,$57,$E1,$FE,$33,$F1,$0D
	.dc.b	$8E,$C7,$BF,$E3,$3B,$E1,$E0,$7F,$1C,$7E,$3D,$B1,$F8,$87,$9B,$F8
	.dc.b	$AA,$B7,$F4,$5E,$71,$D4,$EF,$1F,$87,$DB,$F4,$17,$D2,$3E,$C5,$DE
	.dc.b	$0F,$D6,$BF,$9F,$F8,$A9,$87,$E6,$3E,$32,$DF,$3D,$98,$FA,$6B,$BE
	.dc.b	$3F,$66,$73,$1F,$A0,$F8,$CF,$D5,$1F,$AB,$F8,$CB,$1F,$A4,$CB,$AD
	.dc.b	$FE,$FB,$2B,$3F,$A1,$C3,$9E,$5E,$1C,$BD,$ED,$9E,$C4,$7F,$29,$8F
	.dc.b	$67,$F2,$D8,$F4,$7F,$13,$85,$D3,$E2,$DF,$C9,$9F,$76,$9E,$D7,$F3
	.dc.b	$DA,$73,$D3,$F8,$9E,$56,$FF,$89,$71,$FF,$2D,$27,$C6,$7F,$09,$EC
	.dc.b	$E5,$7F,$C6,$BE,$32,$DF,$07,$F8,$77,$C6,$7F,$0F,$F8,$AB,$3F,$C2
	.dc.b	$1E,$4F,$F0,$AB,$FE,$29,$86,$34,$F8,$47,$EC,$F0,$EA,$FE,$99,$E6
	.dc.b	$FE,$DE,$EE,$3F,$CB,$1F,$E5,$1F,$96,$7F,$B6,$A3,$F9,$9B,$FE,$D4
	.dc.b	$9F,$E0,$0E,$71,$FE,$60,$E6,$7D,$91,$FC,$96,$1F,$19,$FC,$A7,$0E
	.dc.b	$AF,$6F,$8B,$EF,$F1,$56,$F8,$3E,$2C,$7C,$7F,$19,$67,$A3,$CD,$ED
	.dc.b	$76,$7C,$AF,$37,$CF,$57,$2F,$5B,$EE,$F8,$B9,$7E,$7F,$8C,$B7,$CA
	.dc.b	$3D,$5F,$19,$E8,$69,$ED,$F8,$AF,$8E,$7C,$8F,$8A,$B1,$F0,$70,$E7
	.dc.b	$97,$CA,$C6,$CF,$C8,$CF,$E7,$E3,$D7,$2F,$9A,$F2,$7E,$7E,$5D,$6B
	.dc.b	$C0,$75,$24,$7D,$4B,$EA,$FF,$44,$DF,$E8,$CF,$A2,$3D,$87,$67,$9B
	.dc.b	$D9,$FB,$19,$F8,$72,$FC,$0F,$C5,$FE,$13,$E2,$7F,$0A,$DA,$E9,$3F
	.dc.b	$86,$F8,$BF,$5C,$7E,$05,$FC,$5F,$C5,$48,$FC,$27,$C5,$DC,$BF,$17
	.dc.b	$87,$36,$CF,$E2,$71,$E7,$1F,$90,$C2,$E3,$DE,$BF,$1D,$8F,$38,$FC
	.dc.b	$BE,$36,$7F,$2B,$F1,$7F,$9C,$F8,$BF,$CE,$B6,$3F,$BF,$D3,$E2,$5E
	.dc.b	$DF,$FB,$08,$FD,$1E,$56,$BF,$39,$F1,$7F,$A6,$3F,$4F,$F1,$76,$3F
	.dc.b	$41,$8F,$5D,$7F,$4E,$DC,$BF,$C8,$47,$7C,$7A,$4F,$B6,$BC,$79,$78
	.dc.b	$9E,$E4,$8F,$D8,$E1,$DD,$FD,$9F,$C4,$FE,$E0,$FD,$CF,$C5,$DA,$E7
	.dc.b	$EF,$7E,$F3,$E2,$FF,$DB,$AD,$8F,$F0,$D7,$C1,$7F,$B0,$70,$FF,$05
	.dc.b	$B7,$D6,$3F,$EC,$38,$FF,$00,$39,$E7,$65,$DE,$7F,$82,$E5,$76,$F7
	.dc.b	$EA,$CF,$C1,$5F,$E8,$B8,$C6,$7F,$C1,$F4,$FF,$1C,$67,$E1,$27,$C5
	.dc.b	$E2,$7A,$1F,$C4,$33,$F7,$BF,$C3,$DF,$6B,$CC,$F8,$25,$C3,$D2,$F3
	.dc.b	$3E,$0C,$D8,$F6,$FC,$57,$88,$F1,$FC,$55,$AF,$8F,$F1,$7E,$6F,$89
	.dc.b	$F3,$B6,$3E,$5F,$C5,$FA,$BE,$2F,$D6,$D8,$F9,$F8,$73,$C7,$E5,$E3
	.dc.b	$78,$3F,$CC,$3A,$B2,$E9,$5F,$41,$B3,$AC,$7A,$4F,$D4,$6C,$72,$F8
	.dc.b	$BF,$97,$F1,$5F,$31,$B9,$FE,$03,$E2,$FE,$7E,$5F,$41,$B1,$F8,$4E
	.dc.b	$8F,$A7,$C9,$F5,$1B,$1F,$88,$BE,$79,$FE,$11,$B9,$7A,$B0,$E7,$3F
	.dc.b	$88,$6C,$7B,$AF,$9C,$FE,$31,$B1,$F0,$BE,$2B,$F0,$67,$E1,$3E,$2A
	.dc.b	$E3,$F9,$8F,$8B,$FC,$39,$F8,$8F,$89,$B1,$F9,$CF,$8B,$FC,$59,$F8
	.dc.b	$CF,$8B,$B1,$FA,$0C,$39,$E7,$F9,$CC,$6E,$5F,$4B,$2E,$73,$FA,$0C
	.dc.b	$AC,$FD,$6C,$B9,$CF,$E9,$32,$B3,$E9,$CB,$B1,$FA,$EC,$7A,$BF,$E9
	.dc.b	$7D,$D9,$74,$8C,$65,$71,$FD,$9E,$5D,$27,$86,$56,$7F,$6F,$F1,$5F
	.dc.b	$A8,$3F,$D4,$1F,$96,$FF,$AA,$34,$FC,$5C,$63,$E2,$66,$1F,$E9,$9A
	.dc.b	$FF,$3B,$D7,$FA,$C1,$FF,$4A,$37,$E2,$79,$FC,$37,$46,$D7,$F0,$8C
	.dc.b	$3C,$0F,$3F,$8A,$B9,$7F,$A1,$71,$E7,$3F,$EB,$56,$C7,$FA,$61,$F0
	.dc.b	$C6,$72,$EC,$F9,$71,$B5,$CD,$F6,$FC,$54,$9F,$57,$C4,$DA,$F0,$7C
	.dc.b	$47,$C2,$3E,$1F,$C4,$5A,$F8,$1F,$11,$F1,$BE,$1F,$E3,$B6,$3E,$26
	.dc.b	$3C,$F8,$3E,$26,$56,$7D,$58,$F3,$CB,$DD,$95,$BF,$95,$8F,$39,$F8
	.dc.b	$4D,$8F,$9D,$F0,$9F,$5B,$E1,$FE,$BB,$71,$FA,$9F,$0F,$F6,$8F,$B7
	.dc.b	$F0,$D6,$3E,$C7,$C3,$7D,$EF,$84,$FB,$ED,$8F,$B9,$7C,$F3,$FA,$4D
	.dc.b	$CB,$EF,$63,$CE,$7E,$B6,$36,$7F,$05,$87,$39,$FB,$4D,$8F,$4E,$1F
	.dc.b	$89,$CB,$D2,$D3,$D2,$4A,$EA,$D9,$EA,$FE,$47,$2F,$4B,$4F,$CA,$7C
	.dc.b	$27,$E6,$8F,$CD,$FC,$1D,$8F,$CC,$7C,$2F,$E7,$8F,$CF,$FC,$2D,$8F
	.dc.b	$F7,$A7,$E5,$DF,$E5,$5C,$BF,$DC,$AF,$F9,$56,$11,$83,$FB,$9B,$8F
	.dc.b	$51,$3B,$9F,$CB,$63,$E1,$7F,$39,$8D,$CF,$F5,$19,$78,$67,$F4,$19
	.dc.b	$59,$65,$D7,$1F,$26,$5D,$BE,$17,$C3,$87,$8E,$BD,$31,$89,$F1,$FC
	.dc.b	$1D,$B7,$C3,$58,$C3,$9F,$86,$98,$7C,$2F,$86,$B9,$7F,$BD,$F0,$7C
	.dc.b	$27,$3F,$86,$E8,$DB,$F9,$39,$74,$69,$F3,$32,$E7,$AF,$B3,$1F,$9B
	.dc.b	$9F,$97,$3C,$DF,$BF,$1E,$95,$E1,$6D,$F1,$D3,$E7,$FC,$37,$B4,$F7
	.dc.b	$7C,$35,$C3,$ED,$FC,$37,$C1,$F8,$6F,$84,$D8,$FB,$FF,$07,$F1,$4F
	.dc.b	$8D,$F0,$D6,$3F,$07,$87,$3D,$3E,$16,$37,$3F,$17,$13,$2E,$95,$F1
	.dc.b	$DB,$3A,$CB,$A4,$FC,$BC,$AC,$F2,$C7,$A6,$5F,$7F,$2B,$7F,$92,$C7
	.dc.b	$9C,$FE,$0D,$B1,$F9,$6C,$79,$CF,$E1,$F1,$B3,$F9,$AF,$86,$FC,$17
	.dc.b	$C2,$7E,$0D,$B5,$FA,$0F,$86,$FC,$37,$C3,$7E,$1D,$B1,$FA,$4F,$86
	.dc.b	$FC,$57,$C1,$7E,$2D,$B1,$FA,$8B,$E7,$8F,$E1,$DB,$BF,$C6,$C3,$9C
	.dc.b	$FE,$2D,$B1,$F2,$AF,$9C,$FE,$3D,$B1,$F3,$BE,$17,$F3,$5F,$09,$F9
	.dc.b	$B6,$E3,$FB,$49,$FC,$27,$E5,$FF,$E5,$13,$FC,$A0,$E2,$3F,$49,$18
	.dc.b	$7F,$70,$BE,$0B,$26,$BE,$0A,$CF,$F0,$4B,$EF,$97,$E8,$DB,$87,$59
	.dc.b	$F1,$1D,$3E,$17,$BB,$E0,$79,$D2,$7C,$F8,$77,$AE,$32,$E8,$65,$E5
	.dc.b	$9C,$C6,$3E,$0F,$D6,$7B,$3E,$16,$CE,$A9,$F0,$9F,$00,$F8,$3F,$0B
	.dc.b	$67,$5F,$07,$F1,$3E,$17,$E2,$B6,$39,$63,$D3,$3F,$44,$7C,$8C,$B1
	.dc.b	$AF,$D0,$CB,$A5,$68,$F5,$66,$CF,$35,$F5,$B2,$ED,$5B,$D3,$DF,$18
	.dc.b	$9F,$B7,$9F,$6C,$FC,$07,$90,$F7,$65,$D1,$FB,$78,$DC,$7C,$38,$73
	.dc.b	$9F,$BE,$D8,$F2,$FC,$27,$E1,$3E,$07,$F0,$AD,$AF,$95,$F0,$9F,$88
	.dc.b	$3F,$13,$F0,$B6,$3E,$77,$C2,$FE,$33,$E1,$7F,$1A,$D8,$FA,$57,$CF
	.dc.b	$1F,$77,$02,$31,$BF,$E6,$32,$E9,$5A,$CA,$CF,$E7,$31,$E9,$3C,$9B
	.dc.b	$1F,$A0,$F8,$5F,$CE,$9F,$9E,$F8,$5B,$9F,$E0,$FE,$17,$F4,$27,$E8
	.dc.b	$BE,$16,$C7,$E1,$FE,$13,$F4,$A7,$F4,$9F,$13,$FC,$85,$1F,$A6,$8E
	.dc.b	$79,$FA,$B1,$ED,$3F,$8F,$79,$C7,$BB,$0E,$D3,$E1,$3B,$12,$3F,$01
	.dc.b	$7D,$9F,$C3,$DF,$87,$E0,$BC,$5F,$1B,$6F,$F4,$37,$FE,$40,$BF,$D3
	.dc.b	$2F,$86,$F4,$7C,$27,$A5,$B5,$C3,$E0,$FD,$9F,$07,$ED,$6C,$74,$CB
	.dc.b	$BB,$4E,$F9,$73,$7C,$38,$5C,$BD,$2F,$38,$F2,$E1,$67,$DA,$FA,$5F
	.dc.b	$7D,$5E,$0F,$EB,$59,$FA,$FB,$0F,$AF,$0E,$E7,$C0,$D3,$D7,$87,$C8
	.dc.b	$D3,$D2,$D9,$F0,$9F,$1B,$E0,$FE,$D1,$F6,$FE,$0E,$C7,$CA,$F8,$3F
	.dc.b	$BC,$7D,$FF,$84,$B1,$F3,$BE,$13,$F0,$5F,$03,$F8,$36,$C7,$52,$63
	.dc.b	$F6,$B0,$EA,$FD,$CC,$39,$BF,$7B,$0B,$B7,$C0,$C3,$9C,$FE,$09,$B1
	.dc.b	$F1,$3E,$13,$F2,$47,$E4,$FE,$12,$E5,$F8,$2F,$84,$FC,$B1,$F9,$7F
	.dc.b	$84,$B1,$F8,$6F,$84,$FC,$D7,$C2,$7E,$6D,$B1,$F8,$AC,$39,$E7,$F9
	.dc.b	$2C,$6E,$5F,$53,$2E,$73,$F9,$6C,$AC,$FD,$8C,$79,$CF,$E6,$9B,$1F
	.dc.b	$73,$E1,$3F,$55,$F0,$9F,$AB,$6E,$7F,$CF,$BF,$08,$7E,$DA,$F5,$FF
	.dc.b	$C7,$71,$8F,$56,$CE,$6B,$82,$8F,$F1,$F7,$E6,$3F,$E4,$43,$FC,$A1
	.dc.b	$D6,$71,$F0,$FE,$1B,$FD,$4E,$77,$6F,$D8,$BD,$B3,$EF,$8D,$AE,$EF
	.dc.b	$6A,$EF,$76,$7F,$A9,$E1,$E0,$EC,$6B,$EE,$BF,$DE,$B7,$6F,$F2,$06
	.dc.b	$DF,$03,$3E,$F8,$FB,$1B,$7F,$D0,$F1,$F4,$CF,$68,$F0,$BE,$0F,$86
	.dc.b	$F8,$27,$C2,$F8,$6B,$1E,$4F,$86,$F8,$A7,$C6,$F8,$6B,$1E,$8F,$86
	.dc.b	$F9,$3F,$0F,$F2,$9B,$1D,$49,$8F,$8F,$0E,$AF,$97,$0E,$6F,$9F,$0B
	.dc.b	$AF,$F9,$57,$3F,$F3,$3E,$1F,$1B,$2E,$CD,$3E,$4F,$C2,$FD,$AF,$85
	.dc.b	$FB,$6D,$8F,$9B,$F0,$BF,$78,$FB,$FF,$0F,$63,$E8,$F1,$FE,$0B,$E1
	.dc.b	$BF,$06,$D8,$EA,$7D,$6C,$3A,$B4,$EA,$4D,$7E,$85,$F5,$7E,$95,$73
	.dc.b	$6C,$FE,$02,$FF,$23,$F0,$9F,$92,$F8,$6B,$3F,$84,$7F,$2B,$F0,$B2
	.dc.b	$3F,$0F,$F0,$D7,$0F,$BD,$D5,$F9,$A3,$F3,$7C,$F6,$BF,$1B,$8F,$56
	.dc.b	$9D,$4F,$C8,$E1,$CF,$8B,$E8,$63,$67,$FD,$69,$C5,$FA,$77,$F2,$F8
	.dc.b	$F6,$69,$D3,$2F,$C8,$67,$EA,$6C,$76,$23,$F9,$5B,$EC,$FF,$8D,$EF
	.dc.b	$B3,$EC,$9F,$D2,$E7,$EB,$8F,$CE,$7C,$27,$EE,$3E,$1B,$F7,$2D,$AF
	.dc.b	$F1,$EF,$E6,$5F,$EC,$06,$9F,$EB,$B7,$FC,$B1,$5F,$E4,$57,$4F,$F8
	.dc.b	$A5,$FF,$23,$37,$E1,$F8,$7C,$37,$17,$BD,$7E,$E3,$0B,$5F,$AF,$7B
	.dc.b	$44,$3C,$7F,$0F,$67,$B5,$F3,$9F,$3B,$63,$C5,$F0,$BE,$73,$D1,$F0
	.dc.b	$D6,$BD,$BF,$0D,$EB,$F8,$6F,$63,$63,$E0,$FC,$37,$BC,$F8,$1F,$0F
	.dc.b	$63,$E2,$E1,$CF,$1F,$81,$8D,$E5,$FF,$62,$6F,$FE,$60,$BF,$97,$9F
	.dc.b	$66,$9D,$89,$7F,$27,$1E,$CF,$CB,$F8,$4F,$A1,$F0,$FF,$45,$B5,$F5
	.dc.b	$BE,$17,$EA,$7C,$2F,$D5,$6C,$7D,$AF,$85,$FB,$07,$D9,$F8,$7B,$1F
	.dc.b	$7A,$F9,$E3,$DE,$3F,$03,$8F,$37,$F0,$78,$D9,$FB,$CF,$27,$F0,$18
	.dc.b	$75,$9F,$01,$D4,$E6,$58,$FC,$35,$73,$D7,$E3,$DD,$A5,$7E,$36,$3B
	.dc.b	$4F,$CC,$BE,$D3,$E0,$AF,$CB,$FC,$27,$E5,$7E,$1B,$F2,$CD,$8F,$CD
	.dc.b	$FC,$2F,$E6,$7E,$1B,$F3,$4D,$8F,$CF,$FC,$31,$FF,$4E,$FE,$67,$FE
	.dc.b	$43,$F5,$1C,$8A,$D6,$7F,$E8,$7A,$D4,$B2,$F0,$4F,$E8,$DB,$AF,$E1
	.dc.b	$32,$F0,$4F,$E9,$DB,$1F,$88,$C3,$AB,$DB,$16,$3D,$A7,$39,$7C,$69
	.dc.b	$CF,$C5,$E3,$1F,$F5,$C5,$BE,$23,$15,$C9,$B1,$FE,$3F,$35,$7F,$03
	.dc.b	$E1,$73,$F1,$1A,$6C,$7C,$4F,$86,$E3,$F1,$1C,$9B,$1F,$23,$0E,$A7
	.dc.b	$6F,$88,$B1,$D4,$F9,$B8,$75,$69,$D4,$FA,$37,$D5,$A7,$D4,$F8,$8F
	.dc.b	$49,$EA,$F8,$7B,$1F,$63,$E1,$FD,$A7,$BB,$E1,$EC,$7D,$CF,$87,$F8
	.dc.b	$3F,$11,$F0,$9B,$1D,$49,$C3,$F7,$70,$EA,$FD,$FC,$39,$BF,$81,$C2
	.dc.b	$F2,$7D,$7E,$16,$3D,$2B,$F0,$CD,$9D,$67,$D1,$E1,$8D,$CB,$EE,$60
	.dc.b	$CB,$AC,$63,$2B,$97,$DE,$CB,$A4,$F0,$6C,$7E,$0B,$E1,$FE,$FF,$C3
	.dc.b	$FE,$01,$B5,$F9,$EF,$87,$FC,$19,$F8,$4F,$87,$B1,$FA,$2F,$87,$FC
	.dc.b	$39,$F8,$8F,$88,$B1,$FA,$6B,$E7,$8F,$E8,$DB,$BF,$C3,$C7,$9C,$FE
	.dc.b	$9F,$1B,$3F,$1F,$1E,$73,$FA,$BC,$6C,$FC,$BF,$85,$FC,$B9,$F9,$8F
	.dc.b	$87,$B9,$FF,$06,$9F,$D3,$71,$FF,$08,$8C,$1F,$A5,$7F,$CA,$66,$70
	.dc.b	$FF,$7F,$38,$F8,$5F,$D2,$A6,$FC,$26,$1E,$F3,$FB,$FC,$7C,$57,$FC
	.dc.b	$19,$E3,$1C,$BE,$1E,$CF,$8A,$1D,$AC,$7C,$13,$98,$C6,$1E,$08,$C4
	.dc.b	$F8,$DB,$F0,$BE,$63,$CF,$F0,$B6,$37,$F0,$DE,$AF,$86,$F5,$B6,$39
	.dc.b	$FC,$37,$B8,$F7,$FC,$2D,$8E,$A4,$D3,$E2,$E1,$D5,$F8,$F7,$CD,$B3
	.dc.b	$E6,$C3,$94,$7C,$CC,$3A,$CF,$80,$E9,$C1,$F5,$B2,$F4,$C6,$0F,$AD
	.dc.b	$87,$59,$FB,$6D,$8F,$B5,$5D,$A7,$EF,$E5,$67,$EF,$7C,$37,$DA,$F8
	.dc.b	$6F,$B6,$D8,$F9,$7F,$0D,$F7,$8F,$BF,$F0,$D6,$3E,$7F,$C3,$7E,$08
	.dc.b	$FC,$1F,$C2,$58,$FA,$77,$CF,$6F,$C4,$B7,$3F,$C8,$E3,$CE,$7F,$1B
	.dc.b	$8D,$9F,$CA,$E3,$CE,$7F,$23,$8D,$9F,$CC,$FC,$37,$E4,$8F,$C9,$FC
	.dc.b	$2D,$CF,$F0,$7F,$0D,$F9,$63,$F2,$FF,$0B,$63,$F0,$FF,$0B,$F9,$A3
	.dc.b	$F3,$7F,$09,$63,$F1,$78,$73,$CF,$F4,$38,$DC,$BF,$53,$97,$39,$FD
	.dc.b	$2E,$56,$7F,$5B,$97,$39,$FD,$4E,$56,$7F,$AB,$7E,$69,$FE,$42,$6F
	.dc.b	$C2,$FF,$50,$8C,$69,$FD,$5D,$FF,$24,$F5,$9F,$F2,$3D,$72,$F8,$6E
	.dc.b	$1F,$0D,$C5,$E5,$F0,$93,$0F,$D4,$5B,$E1,$31,$A3,$1F,$1D,$78,$73
	.dc.b	$F1,$34,$E3,$97,$56,$9D,$32,$E7,$5E,$CC,$BD,$58,$E2,$BD,$47,$3C
	.dc.b	$7C,$0E,$0F,$2E,$3E,$A7,$A1,$DA,$36,$7F,$04,$C9,$8F,$58,$C6,$1E
	.dc.b	$0A,$F7,$BF,$2F,$E1,$A4,$7C,$2F,$85,$B5,$D3,$5F,$99,$F0,$DF,$44
	.dc.b	$FA,$5F,$0D,$6F,$E4,$FC,$27,$D5,$3E,$B7,$C2,$58,$F9,$B9,$74,$E0
	.dc.b	$FB,$B9,$D9,$FB,$58,$F3,$CF,$C1,$B3,$3C,$61,$F8,$4C,$FA,$56,$B3
	.dc.b	$B3,$F8,$8F,$86,$FC,$31,$F8,$7F,$85,$B9,$FD,$EF,$85,$FC,$51,$F8
	.dc.b	$BF,$86,$B1,$F8,$2F,$83,$FC,$71,$F8,$FF,$86,$B1,$F8,$6C,$39,$E7
	.dc.b	$E6,$C6,$E5,$F9,$8C,$B9,$CF,$AB,$2B,$3F,$9C,$CB,$9C,$FB,$B2,$B3
	.dc.b	$FA,$0F,$84,$FC,$F7,$C3,$7E,$7D,$BA,$7E,$5B,$E1,$7F,$C6,$3F,$9A
	.dc.b	$FF,$92,$F8,$BF,$C9,$53,$FD,$06,$3F,$C9,$C7,$F8,$CB,$2F,$B4,$DC
	.dc.b	$7F,$C8,$4B,$E0,$B9,$FC,$1F,$7B,$FD,$1E,$77,$0F,$A6,$FE,$C7,$2F
	.dc.b	$0F,$C3,$5A,$F0,$3C,$9F,$15,$F9,$B0,$C6,$1F,$73,$4F,$5B,$6B,$B9
	.dc.b	$ED,$39,$95,$F4,$DF,$2C,$FD,$38,$75,$AE,$E7,$C1,$F8,$6F,$8B,$F0
	.dc.b	$BF,$19,$B1,$F1,$7E,$1B,$E4,$FC,$37,$CA,$6C,$7C,$9F,$86,$F9,$BF
	.dc.b	$0D,$F3,$9B,$1D,$4F,$9D,$87,$56,$9D,$4F,$A5,$7D,$5A,$75,$3E,$B5
	.dc.b	$75,$69,$F6,$7E,$17,$EE,$FC,$2F,$DE,$6C,$7D,$DF,$85,$FC,$0F,$B5
	.dc.b	$F8,$26,$C7,$E0,$7E,$13,$F0,$A7,$E1,$BE,$16,$C7,$53,$F0,$D8,$75
	.dc.b	$69,$D4,$FC,$55,$F5,$69,$D4,$9C,$5F,$8C,$AE,$AF,$E3,$BE,$17,$F2
	.dc.b	$A7,$E5,$BE,$16,$D7,$E5,$7E,$17,$F3,$27,$E6,$BE,$16,$C7,$E6,$7E
	.dc.b	$0F,$F3,$BF,$0B,$F9,$E6,$C7,$E7,$70,$E7,$8F,$E7,$31,$BC,$FF,$9C
	.dc.b	$CB,$9C,$FE,$83,$2B,$3F,$A0,$C7,$9C,$FE,$91,$B1,$FA,$4C,$39,$E3
	.dc.b	$FE,$A3,$ED,$65,$8C,$3F,$C6,$30,$CB,$AD,$6F,$2B,$5A,$CB,$AC,$F3
	.dc.b	$6C,$7F,$8E,$BF,$36,$FF,$25,$67,$FE,$B5,$5F,$0B,$FE,$AE,$73,$7F
	.dc.b	$EB,$57,$2F,$F8,$FD,$FF,$1E,$57,$F3,$76,$D7,$3F,$86,$E5,$F0,$DC
	.dc.b	$FE,$13,$A7,$C3,$5C,$79,$3D,$F1,$F0,$65,$67,$B3,$C9,$F0,$5F,$97
	.dc.b	$0C,$57,$87,$0E,$B5,$E5,$6C,$79,$7E,$17,$D9,$F0,$DE,$D6,$D7,$AF
	.dc.b	$E1,$BE,$01,$F0,$7E,$1A,$C7,$BF,$E1,$7E,$27,$C3,$7C,$56,$C7,$C3
	.dc.b	$BE,$78,$F7,$8F,$8D,$8F,$28,$F8,$D8,$F5,$9F,$01,$D4,$F9,$98,$75
	.dc.b	$69,$F3,$FE,$1B,$EA,$1F,$57,$E1,$AC,$7D,$3F,$86,$FB,$1F,$0D,$F6
	.dc.b	$5B,$1F,$5F,$E1,$BE,$E1,$F7,$7E,$1A,$C7,$52,$67,$F6,$F0,$EA,$FD
	.dc.b	$DB,$E6,$D9,$FC,$16,$1C,$A3,$F0,$58,$75,$9F,$01,$D3,$97,$F1,$D9
	.dc.b	$7A,$63,$07,$E3,$70,$EB,$3F,$93,$6C,$7E,$46,$BB,$4F,$E5,$F2,$B3
	.dc.b	$F9,$5F,$86,$FC,$C7,$C3,$7E,$65,$B1,$F9,$7F,$86,$FC,$E7,$C3,$7E
	.dc.b	$75,$B1,$FE,$99,$F6,$BF,$C7,$BA,$FF,$A7,$9B,$F0,$27,$FC,$7E,$63
	.dc.b	$B1,$8F,$6A,$FD,$13,$74,$FD,$66,$3D,$A7,$F4,$CD,$86,$29,$F0,$38
	.dc.b	$7F,$C4,$58,$78,$8F,$0C,$66,$75,$8F,$97,$E1,$B1,$3C,$E3,$D1,$F0
	.dc.b	$98,$C3,$D9,$56,$73,$5F,$07,$E1,$31,$8F,$F9,$4E,$30,$7C,$5F,$86
	.dc.b	$CF,$C2,$EB,$E1,$6C,$FC,$97,$8E,$3C,$B3,$B3,$F3,$5E,$B8,$F6,$C2
	.dc.b	$CF,$53,$E9,$61,$D5,$A7,$53,$EB,$7C,$2F,$98,$F3,$FC,$2D,$8F,$B5
	.dc.b	$F0,$BE,$AF,$85,$F5,$B6,$3E,$F6,$BE,$E3,$DF,$F0,$B6,$3A,$93,$83
	.dc.b	$E8,$E1,$D5,$FA,$77,$CD,$B3,$F8,$9C,$39,$47,$D8,$C3,$AC,$F8,$0E
	.dc.b	$9C,$3F,$62,$19,$7A,$A7,$27,$D9,$86,$1D,$EB,$83,$67,$53,$E0,$9E
	.dc.b	$99,$59,$E5,$F0,$BF,$70,$FB,$BF,$0B,$63,$F4,$1F,$0B,$F8,$03,$F0
	.dc.b	$3F,$0B,$63,$F4,$9F,$03,$F8,$43,$F0,$BF,$09,$63,$F5,$18,$73,$DB
	.dc.b	$F2,$F8,$DC,$FD,$D9,$73,$9F,$CD,$E5,$67,$E1,$65,$CE,$7F,$3F,$95
	.dc.b	$9F,$8D,$F0,$9F,$97,$FC,$DD,$F0,$9F,$96,$8C,$39,$F8,$4D,$7C,$2D
	.dc.b	$DF,$F5,$98,$7E,$9B,$0F,$F2,$84,$60,$FD,$32,$F8,$5E,$DF,$07,$DD
	.dc.b	$B3,$FC,$17,$2E,$ED,$39,$68,$EE,$61,$E6,$9C,$C6,$33,$F2,$C6,$27
	.dc.b	$FC,$C8,$F8,$B3,$F0,$39,$CA,$E1,$FE,$A3,$A4,$7C,$3F,$85,$95,$CF
	.dc.b	$E1,$2D,$EF,$E1,$7E,$41,$F2,$7E,$16,$D7,$87,$E1,$7E,$61,$F3,$7E
	.dc.b	$0E,$C7,$97,$1E,$7C,$3F,$27,$2B,$3F,$23,$1E,$79,$79,$72,$B7,$F3
	.dc.b	$F2,$E7,$3E,$9C,$AC,$FD,$3C,$B9,$CF,$C9,$D1,$9E,$37,$F9,$79,$F4
	.dc.b	$D3,$EB,$E9,$70,$E1,$A7,$49,$FB,$7A,$59,$E9,$F0,$BF,$89,$3F,$15
	.dc.b	$F0,$96,$BE,$97,$C2,$FE,$34,$FC,$77,$C2,$D8,$FA,$DF,$09,$F9,$13
	.dc.b	$F2,$5F,$09,$63,$ED,$61,$F9,$5F,$83,$E9,$97,$D7,$79,$3F,$67,$1E
	.dc.b	$B5,$E0,$3A,$92,$3E,$F5,$F5,$7F,$01,$F0,$7F,$A1,$3F,$45,$F0,$76
	.dc.b	$BF,$15,$F0,$BF,$CE,$FF,$38,$FF,$1B,$DF,$CB,$8C,$39,$F8,$4D,$7C
	.dc.b	$25,$BF,$CA,$DF,$F8,$E1,$FD,$1E,$3D,$6F,$F2,$18,$DE,$9F,$D5,$65
	.dc.b	$D2,$FC,$27,$2A,$E8,$7F,$BE,$70,$E8,$FE,$95,$FF,$24,$B9,$C7,$D1
	.dc.b	$F0,$9E,$9F,$84,$B7,$E4,$7D,$9F,$0B,$ED,$F8,$3B,$3E,$87,$E0,$7C
	.dc.b	$1C,$8F,$5F,$C1,$DC,$BC,$58,$F5,$9F,$7B,$63,$CD,$87,$59,$F8,$6D
	.dc.b	$8F,$55,$F3,$D7,$FB,$7E,$8C,$71,$87,$F6,$F8,$63,$D6,$B6,$D9,$D6
	.dc.b	$3D,$67,$9B,$63,$96,$1D,$31,$F9,$18,$DE,$C6,$4F,$85,$FC,$07,$C2
	.dc.b	$FE,$05,$B8,$FD,$CF,$85,$FC,$27,$C1,$FE,$15,$B1,$F8,$0F,$85,$FC
	.dc.b	$41,$F8,$9F,$84,$B1,$F8,$4B,$E7,$A7,$DB,$6E,$FF,$0B,$0E,$73,$F7
	.dc.b	$DB,$1F,$1B,$0E,$73,$F8,$3C,$2C,$FC,$AF,$85,$FC,$C1,$F9,$9F,$85
	.dc.b	$B9,$7E,$53,$E1,$7F,$39,$F0,$5F,$9D,$6C,$7E,$63,$E1,$7F,$41,$F0
	.dc.b	$7F,$A1,$6C,$7E,$73,$0E,$79,$FE,$4F,$1B,$97,$DA,$C7,$9C,$FE,$5D
	.dc.b	$B1,$F7,$B0,$E7,$3F,$9B,$6C,$7E,$0B,$E1,$7F,$60,$7F,$8A,$FF,$39
	.dc.b	$FF,$23,$F1,$A3,$1F,$13,$9F,$85,$BA,$FF,$44,$B4,$71,$F8,$5E,$5F
	.dc.b	$0B,$6B,$4F,$67,$F6,$F8,$77,$71,$A3,$3F,$F2,$1E,$2C,$BC,$15,$DD
	.dc.b	$B5,$B8,$FF,$23,$4B,$1F,$0C,$79,$5B,$3F,$BB,$C3,$BE,$1D,$AA,$9C
	.dc.b	$B0,$EB,$1E,$26,$C7,$6B,$E7,$3E,$66,$C7,$8B,$E1,$7E,$51,$F2,$FE
	.dc.b	$16,$C7,$C7,$F8,$6F,$9C,$7C,$FF,$84,$B1,$F2,$FE,$1B,$E9,$7C,$2F
	.dc.b	$D3,$6C,$7C,$FC,$39,$EB,$F0,$4F,$EF,$1B,$32,$CE,$9F,$2B,$3E,$B5
	.dc.b	$BE,$E5,$FF,$8E,$31,$F2,$1E,$13,$EE,$67,$E1,$69,$F7,$FE,$1B,$F0
	.dc.b	$C7,$E1,$FE,$1A,$C7,$E0,$FE,$1B,$F1,$5F,$0D,$F8,$B6,$C7,$E1,$FE
	.dc.b	$1B,$F1,$DF,$0D,$F8,$F6,$C7,$52,$69,$E7,$CF,$FD,$31,$18,$DB,$F0
	.dc.b	$B8,$74,$AD,$36,$3F,$13,$7D,$27,$93,$63,$F1,$BF,$0D,$F9,$F3,$F4
	.dc.b	$1F,$0D,$70,$FC,$E7,$C3,$7E,$8F,$E1,$3F,$48,$D8,$FF,$11,$F8,$9F
	.dc.b	$E2,$5C,$BE,$63,$71,$FF,$4D,$BF,$E2,$48,$FF,$16,$39,$9F,$CB,$E3
	.dc.b	$DE,$7D,$D8,$DB,$FD,$0E,$1D,$E7,$E1,$36,$3A,$FC,$37,$FA,$87,$1E
	.dc.b	$95,$DB,$E1,$71,$3C,$F2,$EE,$7E,$DF,$1C,$57,$E2,$31,$E9,$3F,$AA
	.dc.b	$6C,$7A,$2B,$9C,$FC,$0B,$B3,$F0,$3E,$17,$C4,$78,$FE,$1A,$D7,$C6
	.dc.b	$F8,$6F,$31,$E7,$F8,$6B,$1F,$2B,$E1,$7D,$47,$AF,$E1,$6C,$7C,$EC
	.dc.b	$39,$E3,$F3,$B1,$BB,$FC,$EC,$B9,$CF,$D2,$CA,$CF,$D2,$CB,$9C,$FD
	.dc.b	$6C,$AC,$FD,$6F,$85,$F9,$5F,$0B,$F2,$DB,$A7,$DE,$F8,$5F,$9C,$7C
	.dc.b	$FF,$82,$B1,$F8,$2F,$85,$FA,$47,$D3,$F8,$2B,$1F,$86,$C3,$9E,$7F
	.dc.b	$86,$C6,$E5,$F8,$6C,$B9,$CF,$E2,$B2,$B3,$F8,$AC,$B9,$CF,$E3,$B2
	.dc.b	$B3,$F8,$EF,$85,$FC,$17,$C2,$FE,$0D,$BA,$7E,$5B,$E1,$3F,$0C,$7E
	.dc.b	$1F,$E1,$6C,$7E,$6B,$E1,$3F,$15,$F0,$9F,$8B,$6C,$7E,$7A,$F9,$E7
	.dc.b	$F9,$E6,$E5,$F9,$EC,$79,$CF,$E8,$B1,$B3,$FA,$2C,$39,$CF,$E9,$9B
	.dc.b	$1F,$A6,$CB,$9E,$3F,$EF,$FE,$B6,$58,$C3,$F5,$39,$74,$AD,$65,$67
	.dc.b	$F5,$B9,$74,$9E,$4D,$8F,$D8,$FC,$22,$7F,$D3,$7F,$2D,$FE,$4E,$F2
	.dc.b	$19,$AB,$57,$FC,$9E,$B4,$7F,$90,$35,$CD,$A3,$FC,$8A,$DF,$86,$E3
	.dc.b	$3A,$9C,$9F,$E2,$CC,$7C,$3F,$07,$A5,$87,$FA,$C6,$5F,$0B,$98,$C5
	.dc.b	$7E,$FF,$2C,$65,$E1,$9C,$61,$EC,$73,$87,$9E,$3A,$CF,$9F,$E1,$B3
	.dc.b	$9E,$9D,$E3,$2B,$E2,$E5,$71,$F8,$5F,$0B,$D3,$E1,$BA,$B6,$BE,$5F
	.dc.b	$C3,$78,$0F,$0F,$C3,$58,$F9,$FF,$09,$E4,$F8,$6F,$2B,$63,$E9,$DF
	.dc.b	$3C,$7B,$C7,$D6,$C7,$94,$7D,$3C,$7A,$CF,$80,$EA,$7D,$CC,$3A,$B4
	.dc.b	$FB,$FF,$0B,$F1,$3E,$1B,$E2,$B6,$3F,$07,$C9,$F2,$0F,$93,$F0,$D6
	.dc.b	$3F,$0F,$F0,$9F,$30,$F9,$BF,$09,$63,$F1,$78,$73,$CF,$F1,$18,$DE
	.dc.b	$7F,$C1,$65,$CE,$7F,$19,$95,$9F,$C3,$65,$CE,$7F,$21,$95,$9F,$C5
	.dc.b	$63,$CF,$2F,$B7,$CC,$CF,$18,$7E,$27,$3E,$95,$AC,$EC,$FE,$37,$3E
	.dc.b	$93,$CB,$3B,$3F,$91,$F8,$4F,$C4,$9F,$8A,$F8,$6B,$5F,$A8,$F8,$5F
	.dc.b	$C6,$9F,$8E,$F8,$6B,$1F,$AC,$F8,$3F,$CA,$FE,$75,$FA,$F8,$C1,$F9
	.dc.b	$EE,$1F,$D4,$C6,$72,$FD,$76,$5D,$9F,$D8,$65,$6B,$F5,$59,$76,$9F
	.dc.b	$DA,$65,$67,$F5,$D9,$78,$32,$FD,$0B,$5F,$01,$FA,$2C,$BF,$C9,$8E
	.dc.b	$2D,$D8,$9C,$C7,$F8,$9D,$FF,$2E,$CE,$E7,$5F,$09,$EE,$F8,$49,$39
	.dc.b	$A7,$C2,$E3,$0E,$5F,$07,$F1,$0F,$8B,$F0,$76,$F8,$63,$F2,$0F,$93
	.dc.b	$F0,$D6,$3A,$65,$CF,$5F,$93,$95,$9F,$66,$3C,$F2,$F4,$65,$6F,$E7
	.dc.b	$FC,$27,$D4,$3E,$AF,$C3,$5C,$3D,$1F,$0D,$F6,$3E,$1B,$EC,$B6,$3D
	.dc.b	$9F,$0D,$F7,$3E,$1B,$EE,$B6,$3E,$06,$1F,$80,$F8,$3E,$9A,$7C,$97
	.dc.b	$9C,$FC,$5C,$6C,$FC,$D7,$9C,$7C,$9C,$2C,$FD,$17,$9E,$7F,$88,$BB
	.dc.b	$5F,$6D,$E7,$1F,$8C,$BB,$3F,$7D,$E7,$1F,$90,$AB,$3F,$83,$7F,$2F
	.dc.b	$F0,$D2,$3E,$D7,$C2,$5C,$7F,$05,$F0,$DF,$9C,$F8,$4F,$CE,$B6,$BF
	.dc.b	$01,$F0,$9F,$A0,$3F,$43,$F0,$96,$3F,$09,$8F,$3C,$FF,$43,$95,$9F
	.dc.b	$CA,$61,$CF,$2E,$F1,$F8,$BC,$B9,$47,$E4,$F2,$EB,$3E,$03,$F2,$1F
	.dc.b	$0B,$FB,$03,$FA,$2F,$E7,$7F,$E4,$4E,$BF,$CE,$46,$3E,$16,$57,$E6
	.dc.b	$E9,$F0,$B8,$C7,$FD,$0D,$0F,$85,$E7,$F0,$BD,$1B,$7E,$02,$57,$0F
	.dc.b	$86,$F1,$47,$5C,$7A,$70,$7F,$68,$C7,$BC,$E4,$F1,$61,$D3,$0F,$0B
	.dc.b	$6B,$87,$C3,$7B,$0F,$6F,$C3,$5B,$F5,$7C,$37,$C0,$F8,$5F,$82,$D8
	.dc.b	$F7,$7C,$37,$C4,$F8,$6F,$8A,$D8,$F8,$58,$73,$CF,$E0,$E3,$72,$F4
	.dc.b	$63,$CE,$7E,$2B,$63,$D9,$87,$39,$F9,$2D,$8F,$4E,$1F,$3F,$E1,$BA
	.dc.b	$3F,$DD,$35,$47,$A2,$3E,$A6,$97,$0F,$A6,$FA,$23,$EC,$69,$67,$EB
	.dc.b	$BF,$80,$F8,$69,$3F,$77,$E1,$AD,$7D,$CF,$84,$FC,$29,$F8,$6F,$85
	.dc.b	$B5,$F8,$3F,$85,$FC,$49,$F8,$AF,$85,$B1,$F8,$7C,$7F,$1B,$F0,$BD
	.dc.b	$37,$FB,$8F,$2C,$BB,$57,$E3,$71,$C5,$7E,$07,$1E,$75,$F8,$EC,$6C
	.dc.b	$FE,$17,$E1,$BF,$33,$F0,$DF,$9A,$6E,$3F,$CC,$75,$FF,$1C,$13,$E1
	.dc.b	$BF,$C4,$6E,$3E,$1A,$98,$F8,$6B,$5F,$9F,$C3,$9E,$7F,$9E,$C6,$E5
	.dc.b	$FE,$3C,$C7,$AC,$F4,$CB,$AB,$C8,$FF,$75,$47,$F9,$00,$FF,$1E,$38
	.dc.b	$CB,$C7,$F0,$DE,$4F,$84,$B9,$FA,$5F,$3F,$C2,$7A,$3E,$12,$CF,$B5
	.dc.b	$F5,$FC,$34,$8F,$81,$F0,$D7,$2F,$0E,$3D,$67,$E2,$36,$3C,$B8,$75
	.dc.b	$9F,$90,$D8,$F4,$DF,$3D,$7E,$06,$16,$FF,$C8,$FA,$33,$E9,$5F,$17
	.dc.b	$2B,$5A,$CB,$A4,$FC,$96,$C7,$2C,$7A,$65,$FD,$BB,$16,$58,$F1,$50
	.dc.b	$F8,$6F,$B6,$7D,$CF,$84,$B8,$FE,$07,$E1,$BE,$F9,$F8,$0F,$85,$B1
	.dc.b	$F8,$5F,$85,$FC,$1F,$C2,$FE,$11,$B1,$F8,$9C,$39,$E9,$F6,$B1,$BB
	.dc.b	$FE,$07,$2E,$73,$F7,$B2,$B3,$F8,$5C,$79,$CF,$E0,$9B,$1F,$89,$F8
	.dc.b	$5F,$C9,$9F,$94,$F8,$5B,$9F,$E6,$7E,$17,$F2,$E7,$E6,$3E,$16,$C7
	.dc.b	$E7,$7E,$17,$F3,$67,$E7,$3E,$16,$C7,$E8,$70,$FF,$40,$FE,$79,$FE
	.dc.b	$29,$CB,$C1,$3F,$E1,$58,$C3,$F6,$B1,$EB,$A7,$E6,$32,$B7,$F9,$FC
	.dc.b	$BA,$CF,$E7,$32,$B3,$FE,$9E,$C7,$FC,$46,$DF,$84,$F0,$FC,$2F,$89
	.dc.b	$F1,$FC,$2F,$93,$E1,$6E,$BE,$37,$CF,$F0,$B2,$3C,$DF,$0B,$74,$65
	.dc.b	$D6,$7D,$58,$F6,$69,$ED,$C7,$B3,$4E,$C7,$C1,$C3,$A5,$7A,$70,$B3
	.dc.b	$FE,$A2,$CF,$FC,$77,$8F,$82,$31,$F0,$5E,$B7,$E5,$3E,$A8,$D1,$FE
	.dc.b	$46,$9F,$F1,$EC,$F8,$63,$3D,$49,$CC,$61,$63,$E2,$8E,$58,$5A,$E1
	.dc.b	$F0,$BF,$70,$FB,$BF,$0B,$71,$FB,$9F,$07,$F8,$03,$F0,$3F,$0B,$63
	.dc.b	$F0,$1F,$09,$F8,$43,$F0,$BF,$0B,$63,$F0,$9F,$09,$F8,$83,$A6,$9E
	.dc.b	$57,$93,$E7,$C3,$AC,$F8,$0E,$A4,$8F,$65,$75,$7D,$DF,$09,$F9,$4F
	.dc.b	$84,$FC,$AB,$6B,$F2,$9F,$09,$F9,$83,$F3,$3F,$09,$63,$F3,$1F,$0B
	.dc.b	$F9,$CF,$85,$FC,$EB,$63,$F3,$98,$7F,$B8,$3F,$3D,$FF,$0E,$72,$A3
	.dc.b	$1F,$09,$CF,$4F,$A1,$9F,$6C,$3C,$27,$63,$F4,$D9,$76,$69,$FE,$7D
	.dc.b	$C5,$F0,$9D,$CF,$07,$C3,$59,$D7,$C2,$F8,$FE,$1B,$C8,$D8,$E5,$F0
	.dc.b	$DE,$73,$D1,$F0,$D6,$3B,$7C,$37,$AC,$E9,$AF,$E2,$B3,$47,$49,$F2
	.dc.b	$63,$6B,$4F,$37,$86,$1D,$A7,$C2,$76,$3D,$9F,$07,$DD,$A7,$A8,$EA
	.dc.b	$59,$EA,$F2,$7A,$BF,$17,$1F,$07,$80,$84,$FF,$AA,$B2,$7C,$37,$D7
	.dc.b	$F8,$4F,$B0,$DA,$F9,$DF,$0B,$F6,$CF,$B9,$F0,$96,$3E,$97,$AB,$F7
	.dc.b	$CF,$C0,$7C,$2D,$8F,$AD,$87,$3C,$BE,$46,$37,$3F,$24,$78,$F2,$E8
	.dc.b	$FC,$DC,$AC,$EB,$C3,$69,$FD,$4B,$2F,$09,$F8,$3C,$FC,$18,$7D,$5F
	.dc.b	$84,$FC,$91,$F9,$3F,$84,$B5,$F8,$9F,$84,$FC,$B1,$F9,$7F,$85,$B1
	.dc.b	$F8,$DF,$83,$FC,$D1,$F9,$BF,$84,$B1,$F9,$1C,$3D,$3A,$F9,$7F,$3E
	.dc.b	$F9,$B3,$93,$FD,$A4,$E3,$3F,$C4,$65,$D9,$FC,$56,$56,$FD,$DA,$B3
	.dc.b	$EF,$5F,$8F,$CE,$D7,$7F,$85,$F0,$E7,$F7,$37,$68,$9C,$E1,$FE,$E8
	.dc.b	$87,$C2,$E2,$B7,$56,$7F,$C4,$B8,$7A,$FE,$1B,$1F,$09,$97,$5F,$0B
	.dc.b	$33,$F6,$FC,$2D,$D3,$F9,$6E,$3E,$19,$F8,$3F,$0D,$D7,$3E,$CF,$38
	.dc.b	$F8,$B8,$F6,$69,$F2,$31,$E9,$9F,$CD,$CA,$CF,$BF,$16,$1D,$23,$E7
	.dc.b	$E3,$72,$F1,$E5,$DD,$F2,$E7,$72,$FB,$19,$F4,$9F,$8D,$9D,$D3,$E5
	.dc.b	$FC,$37,$C5,$3E,$37,$C3,$5D,$3E,$EF,$C2,$FC,$93,$E5,$7C,$2D,$8F
	.dc.b	$C0,$FC,$2F,$CD,$3E,$77,$C2,$D8,$FC,$2E,$1F,$47,$E1,$7A,$69,$F2
	.dc.b	$9E,$4F,$CC,$C7,$AD,$78,$0E,$A4,$8F,$A3,$7D,$5F,$A7,$F0,$DF,$77
	.dc.b	$E1,$7E,$F3,$6B,$F2,$BF,$0D,$F8,$1F,$86,$FC,$13,$63,$F3,$3F,$0D
	.dc.b	$F8,$5F,$86,$FC,$33,$63,$F3,$B8,$7E,$27,$E1,$BA,$65,$F8,$27,$9C
	.dc.b	$FE,$8B,$1B,$3F,$86,$79,$C7,$E9,$B0,$B3,$F8,$A7,$F2,$DE,$CF,$F8
	.dc.b	$9A,$1C,$68,$C3,$7E,$1B,$77,$C3,$E1,$78,$FC,$2D,$8F,$E4,$3D,$3F
	.dc.b	$C2,$F0,$4F,$6F,$86,$95,$A3,$F5,$93,$FE,$04,$71,$97,$8A,$3F,$C1
	.dc.b	$37,$CF,$E1,$3C,$FF,$0D,$E8,$6D,$F0,$D3,$D1,$9F,$92,$31,$87,$F8
	.dc.b	$7B,$0F,$26,$7E,$2C,$BF,$C2,$F3,$DA,$BC,$7F,$0D,$9B,$F5,$67,$F1
	.dc.b	$BE,$1B,$BE,$1C,$1F,$80,$F2,$C6,$D7,$6C,$3C,$3A,$F3,$7A,$98,$3C
	.dc.b	$BF,$0D,$F4,$BE,$0F,$E9,$B6,$3D,$3F,$0D,$F5,$BE,$1B,$EB,$B6,$3D
	.dc.b	$BB,$7D,$AF,$86,$FB,$6D,$8F,$83,$87,$DE,$F8,$6E,$9A,$FC,$07,$9C
	.dc.b	$FC,$6C,$6C,$FC,$47,$9C,$7C,$AC,$2C,$FC,$87,$F1,$3F,$07,$27,$E7
	.dc.b	$FC,$15,$AF,$9B,$F0,$DF,$8E,$3F,$1F,$F0,$D6,$BE,$AF,$C2,$7E,$48
	.dc.b	$FE,$51,$F9,$FB,$E1,$B1,$3F,$76,$3F,$C2,$6D,$F8,$6E,$DC,$9F,$D6
	.dc.b	$67,$39,$78,$2B,$F0,$39,$78,$1A,$78,$09,$1F,$E5,$89,$F0,$46,$2B
	.dc.b	$A7,$C3,$79,$3E,$1B,$CA,$D9,$F2,$7C,$2F,$A0,$F4,$FC,$2D,$8F,$47
	.dc.b	$C2,$7B,$3E,$1B,$DA,$D8,$F6,$7C,$37,$C0,$F8,$4F,$82,$D8,$EA,$4D
	.dc.b	$7C,$D8,$75,$7D,$17,$CD,$B3,$F1,$F0,$E5,$1E,$DF,$85,$BA,$FC,$77
	.dc.b	$93,$F2,$73,$F4,$DF,$78,$F9,$98,$75,$6C,$75,$3E,$A7,$C2,$FD,$8F
	.dc.b	$85,$FB,$2D,$8F,$B1,$F0,$DF,$73,$E1,$BE,$EB,$63,$EE,$7C,$2F,$E0
	.dc.b	$3E,$1B,$F0,$2D,$8F,$C0,$61,$F8,$4F,$86,$E9,$B7,$D7,$79,$CF,$E1
	.dc.b	$71,$B3,$F6,$DE,$71,$F8,$9C,$2C,$FD,$F7,$F1,$FF,$0D,$27,$F1,$DF
	.dc.b	$0B,$6B,$F0,$9F,$0B,$F9,$43,$FB,$E7,$E8,$1F,$E0,$FD,$FF,$21,$18
	.dc.b	$F8,$3C,$FC,$35,$C7,$F8,$A5,$FF,$81,$1B,$F0,$9C,$BE,$13,$9B,$63
	.dc.b	$7F,$09,$DB,$E1,$3B,$B6,$39,$FC,$2F,$8B,$E0,$BC,$6D,$8E,$FF,$0B
	.dc.b	$E6,$3C,$FF,$05,$63,$C7,$F0,$BE,$AF,$85,$F5,$B6,$3C,$FF,$09,$EE
	.dc.b	$F8,$5F,$7B,$63,$D7,$F0,$BF,$08,$E9,$CB,$FE,$C0,$E1,$47,$49,$F8
	.dc.b	$38,$DA,$D3,$D2,$3E,$2E,$16,$79,$3F,$3B,$E0,$FE,$7F,$C2,$DA,$F9
	.dc.b	$6F,$D2,$F8,$3F,$A7,$F0,$B6,$7E,$7B,$F5,$BE,$12,$47,$D2,$F8,$5B
	.dc.b	$97,$FB,$5B,$F4,$1F,$EB,$A7,$FB,$F2,$7F,$C7,$CE,$3E,$1B,$FC,$7B
	.dc.b	$F2,$CF,$C9,$9E,$E7,$F8,$F6,$BF,$A6,$5F,$FA,$02,$B5,$39,$F8,$6D
	.dc.b	$7C,$1C,$AF,$EA,$D7,$FD,$3F,$2F,$ED,$EF,$F6,$68,$CF,$C3,$75,$3B
	.dc.b	$7C,$45,$8F,$EC,$B9,$78,$FD,$BF,$2F,$E4,$FF,$C6,$31,$FF,$7A,$CE
	.dc.b	$6B,$73,$9F,$1B,$FC,$FD,$93,$E2,$3D,$5F,$11,$EB,$6D,$F9,$BE,$1F
	.dc.b	$DC,$7B,$FE,$16,$C7,$4F,$63,$C3,$EA,$F8,$9B,$1E,$EF,$87,$F8,$C7
	.dc.b	$C7,$F8,$6B,$1F,$0B,$E1,$FE,$57,$C3,$FC,$B6,$C7,$A3,$D9,$F6,$FA
	.dc.b	$FE,$E6,$C7,$CA,$E9,$FA,$5F,$0F,$F4,$DB,$1F,$3B,$E1,$FE,$B7,$C2
	.dc.b	$7D,$76,$C7,$C4,$F6,$BE,$4F,$B1,$F2,$9B,$1F,$5B,$E1,$FE,$F7,$5F
	.dc.b	$DF,$6C,$7F,$83,$FC,$8F,$98,$FF,$A1,$5F,$EA,$3D,$3F,$EA,$B7,$3D
	.dc.b	$9F,$E2,$C7,$33,$F8,$7E,$8F,$E0,$93,$F6,$1F,$F4,$2C,$67,$D6,$FC
	.dc.b	$8C,$62,$79,$7C,$1F,$FA,$36,$DF,$09,$63,$0E,$7E,$03,$5D,$37,$2F
	.dc.b	$43,$D3,$D5,$EB,$EA,$59,$FC,$6E,$68,$ED,$D9,$DF,$E0,$AD,$7C,$17
	.dc.b	$C5,$F0,$7E,$3F,$84,$B3,$F1,$5F,$F5,$2E,$EF,$25,$D9,$EC,$E1,$CC
	.dc.b	$62,$FC,$31,$ED,$F8,$4F,$77,$6D,$9F,$A0,$FC,$1B,$F8,$5F,$09,$67
	.dc.b	$EA,$3D,$0F,$93,$E8,$58,$F6,$BF,$33,$D0,$F9,$BE,$85,$9F,$82,$FC
	.dc.b	$DF,$83,$F9,$DF,$09,$67,$F0,$0F,$A2,$3F,$82,$FE,$51,$FC,$0B,$7F
	.dc.b	$F8,$03,$BA,$72,$EB,$D1,$98,$7E,$EA,$FF,$81,$79,$E7,$F8,$16,$1F
	.dc.b	$BA,$5E,$87,$DF,$F8,$5E,$ED,$BF,$F3,$3D,$7C,$DC,$FC,$7D,$3F,$31
	.dc.b	$B3,$DF,$CF,$F3,$9E,$8F,$3E,$C7,$8F,$CF,$F5,$F9,$FE,$C6,$C7,$6D
	.dc.b	$DA,$7C,$0F,$3F,$EC,$B6,$7D,$9E,$77,$C4,$3E,$2F,$9F,$63,$E0,$79
	.dc.b	$DF,$23,$CF,$F9,$2D,$8E,$DE,$67,$CC,$F3,$BF,$0A,$D8,$F9,$1E,$7F
	.dc.b	$D0,$3E,$8F,$9B,$63,$E6,$79,$DF,$CE,$3F,$29,$FD,$8F,$4F,$CD,$8C
	.dc.b	$7C,$4F,$E4,$3D,$1D,$7C,$4D,$C3,$F9,$6E,$1F,$B6,$EA,$62,$87,$A1
	.dc.b	$D4,$ED,$E8,$5B,$E3,$E8,$78,$7D,$0F,$13,$63,$EB,$FE,$FE,$8F,$D2
	.dc.b	$B6,$33,$38,$9F,$57,$9F,$EB,$F3,$ED,$79,$9F,$77,$9F,$EF,$F3,$EC
	.dc.b	$FA,$9F,$D6,$FC,$4F,$C3,$F3,$EC,$FB,$9F,$8D,$E7,$C8,$F8,$3E,$7D
	.dc.b	$E2,$FE,$97,$C2,$F3,$FE,$67,$99,$F3,$5B,$7F,$85,$F8,$9F,$A1,$E4
	.dc.b	$FE,$F1,$B1,$F3,$3C,$3F,$A8,$7D,$5E,$EB,$1F,$43,$AB,$EC,$6F,$F6
	.dc.b	$5B,$1F,$9E,$FC,$37,$FC,$60,$FF,$6F,$D1,$8F,$F0,$E9,$CE,$BF,$62
	.dc.b	$73,$5F,$DA,$3F,$2B,$CF,$59,$A3,$E2,$FF,$BF,$B8,$F5,$EF,$C5,$61
	.dc.b	$B7,$E3,$AD,$FC,$5F,$85,$B1,$DF,$E2,$B9,$9D,$3E,$2E,$C7,$8F,$E2
	.dc.b	$FB,$FB,$9E,$06,$C7,$BF,$1F,$1F,$B9,$ED,$6C,$7A,$FE,$2F,$CE,$7A
	.dc.b	$3E,$2E,$C7,$BF,$E2,$3D,$7F,$17,$EC,$6C,$7C,$BD,$FD,$FF,$01,$F2
	.dc.b	$5B,$1F,$1F,$E2,$BE,$19,$F1,$3E,$2A,$C7,$CB,$F8,$AF,$8F,$F1,$5F
	.dc.b	$21,$B1,$F5,$F9,$3E,$5F,$C2,$7D,$56,$C7,$D3,$F8,$4F,$9E,$7D,$0F
	.dc.b	$84,$B1,$FE,$1E,$E4,$3F,$E3,$27,$FC,$6D,$D7,$FB,$47,$31,$BF,$8A
	.dc.b	$FC,$17,$46,$FD,$DB,$3C,$DE,$5F,$0B,$23,$F1,$93,$70,$FF,$14,$E1
	.dc.b	$FC,$D0,$7C,$4F,$87,$E3,$7C,$4F,$E3,$FB,$3C,$9F,$0D,$71,$F4,$37
	.dc.b	$0E,$7F,$0D,$7E,$23,$D7,$1E,$98,$F6,$7C,$3F,$B7,$E1,$EE,$1F,$05
	.dc.b	$FC,$E7,$83,$F0,$7E,$1A,$CF,$C5,$6E,$1E,$8F,$87,$B9,$76,$F1,$9C
	.dc.b	$09,$F9,$5F,$0F,$F2,$FE,$1E,$E5,$F3,$DF,$D5,$57,$CF,$F8,$7B,$3F
	.dc.b	$4D,$B9,$7C,$6F,$86,$B9,$F7,$BF,$F3,$2D,$FD,$9F,$87,$FB,$1F,$0F
	.dc.b	$F6,$5B,$1F,$BB,$F2,$BE,$E7,$C3,$FE,$E5,$B1,$93,$F0,$5F,$0F,$96
	.dc.b	$9F,$85,$F8,$7F,$F3,$AF,$E5,$9F,$CF,$B2,$46,$3E,$23,$F8,$27,$9D
	.dc.b	$AF,$88,$B8,$7F,$9A,$B0,$FE,$7D,$C0,$F5,$17,$D3,$E2,$37,$3F,$E6
	.dc.b	$CA,$FE,$83,$7F,$C4,$FD,$1F,$0F,$C4,$7F,$11,$6C,$64,$ED,$F0,$D9
	.dc.b	$69,$9B,$E9,$F0,$F9,$6C,$7F,$1F,$F4,$FD,$BF,$0F,$FC,$75,$B1,$93
	.dc.b	$D5,$F0,$F9,$69,$9A,$F4,$7C,$3E,$5B,$1F,$CA,$FD,$7F,$93,$F0,$FF
	.dc.b	$CA,$5B,$19,$3E,$37,$C3,$E5,$A6,$6B,$E2,$7C,$3E,$5B,$1F,$CD,$FD
	.dc.b	$AF,$AB,$F0,$FF,$CD,$5B,$19,$3E,$97,$C3,$65,$A6,$6B,$E8,$7C,$3E
	.dc.b	$5B,$1F,$D0,$FD,$CF,$C0,$FC,$3F,$F4,$16,$C6,$4F,$BD,$F0,$F9,$69
	.dc.b	$9A,$FB,$9F,$0F,$96,$C7,$F4,$FF,$7F,$F1,$BF,$0D,$FD,$35,$B1,$93
	.dc.b	$F1,$5F,$0F,$96,$99,$AF,$C4,$7C,$3E,$5B,$1F,$D6,$FE,$07,$FD,$9B
	.dc.b	$F9,$6F,$F5,$89,$FC,$94,$61,$FE,$B8,$EB,$E1,$AD,$7F,$B0,$AB,$FA
	.dc.b	$C3,$CC,$C7,$C3,$6A,$FF,$D8,$73,$FE,$8E,$F3,$7F,$88,$C7,$E7,$FE
	.dc.b	$17,$C3,$F0,$DF,$DA,$5B,$87,$F6,$F7,$CB,$F0,$B9,$8F,$03,$6B,$AF
	.dc.b	$C0,$5C,$3F,$88,$62,$8F,$EE,$9F,$0D,$EE,$F7,$EE,$3E,$96,$E1,$E6
	.dc.b	$F6,$EE,$3C,$9B,$5F,$EE,$DC,$7F,$8D,$BF,$E0,$3C,$3E,$4F,$9F,$FE
	.dc.b	$01,$F8,$6F,$97,$E9,$5A,$F8,$CD,$CB,$E1,$F9,$B7,$3F,$3B,$6B,$B7
	.dc.b	$95,$6B,$FC,$83,$9A,$3F,$C1,$DF,$0D,$F6,$3C,$8B,$7F,$4D,$B7,$F4
	.dc.b	$3C,$5B,$97,$C4,$6D,$7A,$FC,$2B,$5D,$9F,$F5,$3F,$E2,$1F,$D7,$23
	.dc.b	$EC,$F8,$18,$AF,$C3,$B5,$FC,$17,$6D,$9F,$BE,$DC,$7E,$9F,$5D,$AF
	.dc.b	$94,$F6,$F8,$7F,$C8,$F4,$D9,$FC,$63,$70,$FC,$4F,$2D,$C3,$EE,$B6
	.dc.b	$BE,$A7,$25,$AF,$8E,$FA,$BE,$1F,$F3,$BC,$36,$BF,$30,$DA,$61,$FD
	.dc.b	$5F,$F2,$EF,$80,$FE,$5B,$7F,$F5,$67,$C0,$27,$38,$2E,$56,$A7,$FC
	.dc.b	$0B,$08,$C6,$BF,$33,$E2,$3F,$B8,$E3,$DF,$3C,$63,$CE,$30,$FF,$AB
	.dc.b	$2F,$FD,$67,$1E,$5A,$DE,$1E,$8A,$F1,$65,$E5,$9C,$D7,$AF,$E2,$2D
	.dc.b	$6B,$34,$7D,$E3,$C7,$19,$F8,$8C,$3E,$33,$07,$8E,$37,$8F,$FA,$52
	.dc.b	$BF,$D8,$B3,$93,$84,$67,$E2,$39,$BE,$73,$87,$C4,$63,$1F,$A0,$FE
	.dc.b	$43,$1E,$B9,$FC,$FF,$88,$C6,$3D,$23,$0F,$C5,$9D,$E4,$F8,$8D,$E5
	.dc.b	$DE,$FE,$C4,$63,$2F,$BD,$F1,$17,$1F,$88,$FE,$83,$4E,$7F,$11,$67
	.dc.b	$C1,$97,$F8,$65,$FA,$93,$E0,$F8,$8D,$67,$DE,$91,$8C,$7F,$19,$F1
	.dc.b	$17,$0F,$A4,$FE,$BB,$7F,$4F,$C4,$59,$EE,$DC,$7A,$7C,$45,$C3,$F2
	.dc.b	$CD,$AF,$B9,$F0,$F6,$BE,$A3,$FB,$AE,$2F,$8B,$F1,$16,$BD,$6D,$AF
	.dc.b	$47,$C4,$5C,$7F,$44,$DA,$FC,$47,$C4,$5A,$FC,$03,$F8,$25,$F8,$8B
	.dc.b	$E0,$7E,$9F,$C4,$E6,$31,$7F,$2A,$3E,$ED,$2E,$98,$8B,$3F,$7F,$9F
	.dc.b	$FD,$D2,$4F,$88,$FC,$2A,$F8,$9F,$E2,$FD,$3F,$88,$F8,$2C,$5F,$EA
	.dc.b	$63,$13,$A8,$FC,$54,$FF,$8C,$3B,$BF,$11,$18,$C3,$59,$FE,$26,$BF
	.dc.b	$C5,$EF,$E6,$4F,$DE,$61,$DF,$E2,$3F,$93,$F6,$78,$7E,$1E,$E3,$D9
	.dc.b	$B7,$CF,$E1,$EF,$1F,$F9,$97,$2F,$F7,$84,$F5,$D1,$8F,$AF,$8B,$79
	.dc.b	$7E,$FA,$FE,$5D,$EF,$BB,$E0,$7C,$3F,$F3,$86,$D7,$F0,$97,$E2,$7C
	.dc.b	$3E,$63,$E0,$B6,$BD,$D7,$73,$F2,$3F,$D0,$BB,$BE,$67,$C3,$DC,$3E
	.dc.b	$4B,$6F,$E3,$7C,$3D,$C3,$D8,$DA,$F4,$7C,$3D,$A7,$33,$24,$FF,$50
	.dc.b	$F0,$7E,$D7,$C3,$DC,$3E,$BB,$70,$FA,$9F,$0F,$73,$F9,$4D,$AF,$8D
	.dc.b	$F0,$F6,$BB,$E4,$8F,$EB,$BE,$1F,$E1,$FE,$1E,$DF,$E1,$1B,$7F,$81
	.dc.b	$F8,$7B,$97,$D7,$6D,$7D,$3F,$87,$B5,$D9,$FE,$D3,$E3,$7E,$4F,$E1
	.dc.b	$ED,$7E,$41,$B5,$F8,$DF,$87,$B8,$FE,$0D,$B5,$F7,$FE,$1E,$D7,$A9
	.dc.b	$FE,$EB,$E4,$FF,$26,$FA,$7F,$F1,$C7,$72,$31,$5A,$BF,$E5,$15,$FE
	.dc.b	$36,$8F,$C7,$4F,$1F,$88,$D4,$7F,$27,$9F,$F1,$B5,$7F,$5D,$8F,$EB
	.dc.b	$55,$DF,$E2,$3F,$C0,$BE,$5F,$87,$E2,$2D,$F6,$6D,$F3,$F8,$8B,$B7
	.dc.b	$FB,$0F,$24,$62,$FF,$AC,$C7,$F8,$DD,$FE,$C1,$87,$B7,$E2,$3F,$C2
	.dc.b	$7E,$77,$BF,$E2,$2D,$7B,$1B,$97,$A7,$E2,$2E,$7D,$DB,$5D,$BE,$22
	.dc.b	$D7,$E4,$B3,$FF,$1D,$BF,$B1,$F4,$3E,$67,$C4,$7F,$88,$9B,$7F,$DC
	.dc.b	$DF,$A1,$F1,$19,$8F,$9A,$DA,$F9,$5F,$0F,$73,$F8,$0F,$F8,$B3,$D3
	.dc.b	$FB,$1F,$0F,$70,$FA,$AD,$BF,$A5,$F0,$F7,$0F,$90,$DA,$F8,$FF,$0F
	.dc.b	$6B,$D4,$FF,$33,$FC,$47,$F9,$0C,$7D,$BF,$88,$C7,$B1,$F8,$76,$DF
	.dc.b	$E1,$3E,$22,$D7,$E0,$5B,$97,$D7,$F8,$8B,$5F,$59,$FE,$79,$E9,$FE
	.dc.b	$47,$E2,$2C,$FE,$39,$FE,$F3,$4F,$86,$FE,$F7,$F5,$BF,$E2,$8C,$BF
	.dc.b	$80,$CE,$6F,$FB,$E4,$E7,$D1,$2F,$87,$E3,$F0,$FC,$9B,$5F,$E0,$EC
	.dc.b	$1C,$0F,$7F,$F4,$9F,$13,$F0,$DB,$5B,$9F,$F7,$79,$FC,$F3,$3F,$F4
	.dc.b	$9B,$8F,$7F,$84,$FE,$A3,$E2,$6E,$3F,$A5,$7E,$74,$FF,$61,$F9,$BE
	.dc.b	$15,$E3,$6D,$7F,$47,$B7,$C0,$F0,$AF,$13,$67,$F6,$1C,$17,$2F,$D3
	.dc.b	$3F,$6E,$7F,$B7,$7E,$CF,$FE,$11,$AE,$91,$8F,$7F,$58,$78,$62,$CF
	.dc.b	$F5,$7C,$BC,$33,$E0,$3F,$C5,$5F,$69,$F4,$E7,$E4,$D7,$F8,$C2,$31
	.dc.b	$9F,$F5,$4C,$FF,$D6,$F9,$7F,$8F,$AF,$D9,$F0,$7E,$4F,$89,$FC,$83
	.dc.b	$F9,$1F,$83,$B9,$FC,$F6,$D7,$F8,$CF,$2F,$F3,$2E,$5D,$6B,$FC,$B3
	.dc.b	$18,$AF,$CD,$74,$73,$9F,$B0,$DB,$F9,$9F,$00,$9F,$F3,$97,$E4,$1F
	.dc.b	$3A,$7F,$CE,$98,$FF,$66,$D7,$EE,$DE,$E5,$5B,$7F,$13,$8F,$EA,$7E
	.dc.b	$13,$B5,$FE,$11,$F1,$E3,$F8,$FD,$BF,$C7,$AF,$90,$C7,$C2,$70,$7F
	.dc.b	$94,$66,$CF,$C1,$9F,$78,$F2,$4E,$A7,$3C,$1C,$33,$F1,$4F,$C9,$3F
	.dc.b	$31,$96,$34,$F9,$5D,$37,$5F,$67,$5B,$81,$3E,$3A,$F8,$79,$F6,$7E
	.dc.b	$2E,$5F,$C1,$E3,$11,$FA,$AC,$7F,$05,$F1,$BD,$73,$FA,$AF,$F0,$DF
	.dc.b	$80,$FC,$37,$C6,$D9,$FC,$5B,$CD,$FA,$7B,$7B,$5B,$87,$81,$FD,$96
	.dc.b	$3E,$0D,$FF,$C0,$51,$83,$F6,$D8,$73,$AF,$D5,$7C,$35,$E0,$FD,$BB
	.dc.b	$6B,$B4,$FD,$CC,$6F,$0F,$CF,$DB,$93,$F4,$73,$E1,$5F,$86,$CA,$BF
	.dc.b	$C2,$A3,$F5,$FF,$0B,$8E,$43,$CB,$1F,$BB,$F0,$DE,$9A,$F5,$37,$3F
	.dc.b	$D6,$FC,$6F,$6E,$5F,$A8,$FF,$35,$9F,$D7,$74,$DA,$FD,$9B,$73,$F5
	.dc.b	$1F,$E2,$6E,$67,$C6,$76,$AF,$97,$8F,$DF,$8C,$57,$EB,$7E,$33,$97
	.dc.b	$1F,$E1,$30,$47,$F4,$5F,$89,$FD,$F7,$C5,$DC,$BF,$81,$B7,$0F,$09
	.dc.b	$FC,$B3,$E2,$BE,$B5,$7D,$76,$D7,$F0,$AF,$8B,$E1,$97,$77,$FA,$94
	.dc.b	$FF,$0B,$F0,$6D,$7F,$10,$6E,$1D,$BA,$7F,$C6,$93,$F8,$68,$F0,$36
	.dc.b	$33,$1B,$AF,$E2,$DD,$DC,$B1,$F7,$3F,$D8,$7E,$2F,$F8,$D7,$35,$C7
	.dc.b	$F8,$FB,$70,$F9,$BC,$3F,$A7,$F8,$BF,$C9,$5F,$E9,$63,$17,$E1,$8F
	.dc.b	$E4,$9F,$11,$C7,$2F,$98,$FF,$6F,$F8,$BF,$E8,$5E,$25,$BF,$E8,$4D
	.dc.b	$CB,$DD,$F1,$7F,$9F,$AF,$D0,$36,$BF,$99,$FC,$2E,$F1,$FB,$0F,$F7
	.dc.b	$B9,$FE,$F9,$F1,$76,$BF,$AD,$37,$0F,$AF,$F8,$93,$C5,$7C,$47,$C9
	.dc.b	$C3,$F7,$B3,$98,$C6,$1F,$D8,$FE,$33,$96,$7F,$D8,$9E,$F1,$FD,$B7
	.dc.b	$F6,$8F,$E6,$D9,$7F,$2C,$CF,$FD,$81,$18,$8F,$EE,$39,$76,$CF,$C3
	.dc.b	$1F,$02,$FE,$E6,$7E,$18,$C1,$FD,$97,$1F,$0C,$FF,$28,$7F,$C3,$47
	.dc.b	$F5,$3F,$87,$B1,$FD,$61,$F8,$1E,$37,$F3,$9B,$E8,$FF,$2D,$C7,$C5
	.dc.b	$AF,$E2,$70,$FE,$B9,$F0,$DC,$B2,$FE,$A4,$FF,$B5,$7C,$BF,$EC,$1F
	.dc.b	$03,$74,$FE,$CA,$FF,$B7,$73,$EB,$E3,$FF,$98,$DB,$87,$F4,$EC,$FB
	.dc.b	$BF,$8C,$F6,$DE,$67,$F4,$FE,$3F,$E6,$B1,$8C,$94,$9C,$F9,$3B,$9C
	.dc.b	$F5,$7F,$B7,$70,$FF,$75,$79,$7F,$DD,$7D,$FB,$5F,$DE,$9F,$37,$2F
	.dc.b	$D0,$E4,$F8,$8D,$DF,$FA,$9E,$DE,$69,$FE,$CD,$97,$FB,$A7,$C4,$F0
	.dc.b	$6D,$FE,$B4,$DD,$E0,$E2,$B5,$A7,$A3,$83,$D7,$3F,$EF,$DD,$F5,$F9
	.dc.b	$8F,$F9,$42,$7F,$B7,$62,$F8,$FF,$80,$FB,$F0,$F6,$D6,$DF,$99,$BF
	.dc.b	$9E,$7E,$2F,$1F,$CE,$7F,$1B,$3F,$8A,$9D,$67,$FF,$00,$7B,$7C,$DA
	.dc.b	$9B,$C1,$87,$DB,$87,$8F,$93,$CF,$EB,$E3,$5F,$EE,$D9,$FB,$4F,$2C
	.dc.b	$7C,$CE,$AF,$01,$FA,$8B,$FD,$7E,$9E,$2C,$DD,$5E,$23,$C7,$F1,$F7
	.dc.b	$1F,$F2,$B6,$5F,$ED,$E7,$B6,$1F,$EE,$2E,$0F,$F2,$94,$7F,$AB,$E7
	.dc.b	$53,$F1,$CD,$7A,$FE,$37,$C3,$9F,$8F,$9B,$EF,$BF,$D1,$3F,$26,$FC
	.dc.b	$6E,$1D,$B8,$91,$8B,$E9,$18,$AF,$BD,$39,$E4,$FF,$31,$E9,$F7,$3A
	.dc.b	$3F,$D1,$8D,$C7,$FC,$E1,$CB,$F5,$4F,$AD,$E4,$39,$BF,$1D,$7F,$E2
	.dc.b	$2B,$FF,$4A,$7C,$87,$2C,$3F,$CE,$4E,$1F,$F3,$47,$A9,$76,$FF,$2C
	.dc.b	$3E,$FD,$FD,$B1,$FC,$21,$A7,$F8,$AE,$BE,$EF,$C7,$7A,$6F,$FC,$63
	.dc.b	$18,$A9,$1F,$C6,$FC,$8B,$1F,$34,$78,$9F,$D3,$B7,$FC,$CE,$FE,$DE
	.dc.b	$8F,$CA,$57,$E4,$8F,$A2,$7F,$66,$CB,$C0,$FE,$F3,$D9,$78,$7D,$B0
	.dc.b	$FD,$1C,$63,$3F,$F4,$C5,$FF,$BC,$E7,$F4,$F9,$7F,$88,$B0,$F2,$4E
	.dc.b	$6B,$FD,$E5,$39,$AF,$2E,$3E,$4C,$BE,$8F,$0F,$D0,$C7,$F8,$6E,$3C
	.dc.b	$5B,$F1,$FF,$1B,$97,$E3,$B3,$3F,$1D,$5A,$F0,$BF,$47,$8F,$F6,$B7
	.dc.b	$F8,$F6,$CF,$C1,$BF,$A3,$F1,$DF,$E4,$AB,$46,$30,$FE,$F9,$03,$FA
	.dc.b	$4F,$7F,$D8,$6C,$7E,$33,$4F,$61,$F4,$39,$BF,$19,$1F,$B2,$C7,$C0
	.dc.b	$FC,$AC,$7C,$18,$F7,$9F,$A9,$F1,$F7,$7F,$E2,$79,$23,$18,$7A,$B9
	.dc.b	$7F,$17,$AF,$F2,$68,$ED,$1F,$E8,$5A,$65,$ED,$3B,$CF,$84,$F6,$BF
	.dc.b	$D9,$37,$EB,$97,$F3,$EE,$1F,$4B,$F8,$FE,$1F,$C6,$70,$FC,$B8,$FE
	.dc.b	$3F,$C1,$F8,$EC,$BB,$E5,$F2,$3E,$3F,$C0,$D2,$57,$EE,$BB,$7F,$B9
	.dc.b	$52,$BF,$99,$F9,$9F,$66,$BA,$E3,$DD,$65,$FC,$B6,$F8,$63,$FE,$4C
	.dc.b	$C5,$D3,$F6,$F2,$FF,$12,$46,$27,$ED,$FC,$7F,$F4,$0F,$7B,$F6,$2D
	.dc.b	$8F,$B5,$9F,$CA,$E2,$F9,$9A,$7E,$A3,$8B,$E6,$5F,$E1,$FE,$3F,$F1
	.dc.b	$0D,$C3,$F0,$F1,$FD,$33,$AB,$F7,$AD,$8E,$B7,$CF,$0E,$B8,$7D,$A8
	.dc.b	$FD,$0D,$7E,$4F,$E3,$BA,$47,$ED,$DC,$5F,$F2,$5D,$3F,$58,$FF,$74
	.dc.b	$9F,$CB,$4F,$F5,$CF,$2F,$E5,$47,$C8,$E3,$FC,$A7,$2F,$EB,$E3,$77
	.dc.b	$F2,$7C,$8E,$B1,$FE,$4E,$C7,$FB,$0E,$7F,$EC,$FF,$11,$E5,$7F,$68
	.dc.b	$F2,$7F,$A9,$37,$1F,$6D,$7A,$F0,$FE,$20,$FB,$7C,$5F,$E7,$7F,$19
	.dc.b	$FE,$57,$F1,$8D,$F8,$CC,$3F,$B3,$F8,$EF,$EC,$3E,$2F,$BB,$1F,$E0
	.dc.b	$B1,$8E,$2F,$6F,$6F,$F0,$AD,$3E,$C5,$69,$FD,$CF,$C7,$73,$69,$23
	.dc.b	$F8,$BE,$3F,$CC,$A7,$F9,$9B,$D3,$93,$E3,$D7,$4A,$FE,$05,$5F,$27
	.dc.b	$B8,$DB,$FE,$EF,$FB,$4B,$AF,$9E,$1F,$DA,$30,$4E,$63,$E3,$F5,$7E
	.dc.b	$EC,$FC,$19,$67,$F9,$1E,$1D,$CF,$B1,$CB,$FE,$1E,$AF,$C1,$7C,$0F
	.dc.b	$F8,$0F,$D0,$FF,$0D,$BE,$9E,$EF,$D4,$6F,$FB,$8C,$7F,$20,$74,$BF
	.dc.b	$66,$1F,$D2,$5F,$F0,$EF,$3F,$F8,$37,$CA,$FF,$12,$BC,$EB,$FA,$D6
	.dc.b	$9E,$57,$F1,$98,$FC,$5E,$0F,$87,$87,$4F,$2B,$FC,$5B,$AF,$F8,$5D
	.dc.b	$B1,$F0,$2F,$F7,$A7,$42,$F4,$FE,$27,$5F,$DF,$6F,$F0,$3B,$7F,$C6
	.dc.b	$FB,$7F,$88,$5B,$87,$3C,$3C,$67,$32,$E1,$CF,$93,$F8,$B6,$1F,$2B
	.dc.b	$D5,$FF,$20,$E7,$FE,$2B,$6E,$1C,$F3,$FC,$66,$3F,$20,$E7,$BF,$C1
	.dc.b	$DB,$DF,$19,$E1,$FF,$15,$DF,$ED,$75,$FF,$1B,$79,$FF,$E4,$E7,$E5
	.dc.b	$6D,$FE,$7F,$9F,$C6,$65,$F5,$63,$F9,$C6,$BF,$7B,$CE,$FF,$2A,$EB
	.dc.b	$FE,$3F,$6D,$F9,$77,$FC,$AE,$9E,$5C,$79,$6D,$FC,$0F,$6F,$2E,$7F
	.dc.b	$B3,$D9,$E7,$7F,$98,$37,$FF,$25,$37,$3E,$9C,$7F,$88,$D3,$FA,$DB
	.dc.b	$F7,$77,$FE,$11,$F8,$BF,$F8,$23,$BD,$E0,$B8,$7F,$CA,$F3,$98,$C6
	.dc.b	$3D,$38,$7E,$A7,$0F,$99,$FB,$FC,$3E,$7C,$BA,$E7,$FE,$74,$D7,$9B
	.dc.b	$70,$E5,$C1,$F2,$70,$F8,$1B,$79,$9F,$F0,$E6,$1E,$6F,$3B,$FD,$05
	.dc.b	$EF,$FF,$87,$A5,$B7,$F7,$EC,$7F,$C9,$D3,$9D,$7C,$95,$FC,$63,$5F
	.dc.b	$CF,$46,$38,$FE,$5C,$7E,$87,$0F,$F0,$26,$1C,$B5,$F8,$35,$E0,$7E
	.dc.b	$FD,$7E,$F3,$0F,$79,$0F,$C5,$9F,$B3,$3F,$1D,$7F,$DF,$EF,$CB,$97
	.dc.b	$CC,$F3,$BD,$6F,$4A,$FF,$5D,$4F,$E2,$B5,$F8,$5B,$7E,$2A,$3F,$8C
	.dc.b	$61,$FD,$D2,$BF,$CD,$3A,$FA,$1F,$F2,$54,$67,$CF,$57,$FA,$EA,$B5
	.dc.b	$1F,$E3,$8D,$3F,$91,$7A,$1F,$E0,$4C,$7F,$59,$1F,$85,$BE,$FE,$67
	.dc.b	$D0,$D7,$F0,$6D,$8F,$B3,$B7,$C3,$D3,$E0,$93,$83,$F7,$F3,$E8,$F0
	.dc.b	$3E,$CF,$A1,$F8,$C6,$C7,$2A,$F8,$27,$E0,$B2,$F1,$3F,$2B,$CB,$FC
	.dc.b	$0E,$5F,$94,$6D,$71,$9F,$BC,$7D,$3C,$7E,$F7,$27,$E7,$BC,$BF,$C4
	.dc.b	$E5,$F9,$C6,$DF,$1D,$FE,$16,$BF,$86,$E3,$FA,$2F,$E6,$E7,$E8,$F8
	.dc.b	$DF,$92,$D3,$F4,$AD,$8E,$5D,$DF,$DC,$F4,$F0,$C7,$D8,$D3,$F3,$39
	.dc.b	$FE,$B1,$B1,$C0,$FC,$06,$FF,$03,$8B,$FB,$87,$3B,$C8,$FF,$46,$FC
	.dc.b	$87,$FA,$39,$B7,$C7,$8B,$E4,$1F,$C7,$6B,$F3,$AF,$E5,$55,$7F,$A5
	.dc.b	$BE,$43,$8C,$FF,$9F,$9F,$C9,$3F,$28,$FB,$5C,$3F,$09,$FC,$E6,$3E
	.dc.b	$49,$FE,$F7,$E2,$F3,$C7,$FB,$27,$95,$FE,$0C,$7C,$95,$F8,$CB,$F2
	.dc.b	$6B,$FD,$BF,$E4,$3F,$D4,$FF,$21,$FE,$A8,$6D,$F0,$E5,$FA,$5C,$1F
	.dc.b	$0D,$FE,$6D,$5F,$EA,$EF,$90,$FF,$3F,$36,$BE,$4E,$3F,$C5,$36,$F4
	.dc.b	$3F,$94,$E0,$F8,$6D,$BF,$F4,$37,$C7,$7F,$A1,$DB,$7C,$B5,$FE,$28
	.dc.b	$7A,$4F,$A7,$AF,$F8,$1E,$BF,$5B,$69,$67,$FE,$39,$F1,$5F,$05,$D2
	.dc.b	$B5,$39,$8F,$C3,$69,$FD,$02,$FB,$4F,$F5,$3C,$7F,$C2,$9F,$1D,$FE
	.dc.b	$9B,$F8,$EF,$F4,$E3,$74,$F3,$57,$EC,$7A,$3F,$8F,$9F,$B0,$CF,$F0
	.dc.b	$57,$E6,$C7,$CB,$03,$FC,$E3,$7C,$DF,$29,$9A,$FE,$13,$1D,$A3,$F6
	.dc.b	$B3,$E6,$7F,$19,$8F,$E7,$E7,$B6,$1F,$07,$E1,$2D,$73,$9A,$B5,$4F
	.dc.b	$C2,$E2,$FE,$AC,$77,$9F,$F1,$AD,$7F,$6C,$F8,$FE,$39,$BE,$3E,$C6
	.dc.b	$1E,$55,$F8,$1B,$E5,$C7,$F8,$0B,$70,$7F,$B5,$EF,$F6,$C7,$0D,$BF
	.dc.b	$CB,$DD,$FF,$C0,$B1,$FE,$15,$5F,$8A,$C3,$93,$FD,$56,$3F,$CE,$C7
	.dc.b	$F8,$58,$B8,$76,$C7,$31,$FE,$24,$33,$28,$CC,$71,$3F,$DB,$66,$9C
	.dc.b	$CF,$38,$E6,$D7,$46,$27,$A3,$AF,$8F,$7B,$FC,$9B,$19,$8C,$11,$C1
	.dc.b	$0F,$5C,$70,$6C,$6D,$A6,$CF,$F4,$C5,$3E,$3B,$FD,$35,$9F,$E4,$27
	.dc.b	$FD,$7F,$7E,$28,$C5,$23,$1C,$3C,$AF,$8B,$D2,$33,$7F,$95,$6B,$D3
	.dc.b	$0F,$CE,$61,$DF,$F1,$77,$A9,$CA,$B8,$56,$B0,$E8,$D8,$C9,$FA,$23
	.dc.b	$FD,$AB,$F0,$BC,$5F,$F6,$BF,$C7,$DD,$3D,$D1,$C4,$FF,$6D,$E9,$99
	.dc.b	$D6,$9E,$6A,$F7,$61,$B6,$DF,$7C,$FC,$EE,$6B,$81,$2B,$86,$3E,$1F
	.dc.b	$8F,$F2,$7E,$D5,$F5,$76,$F9,$CD,$D7,$E9,$CF,$8A,$FA,$13,$1F,$06
	.dc.b	$D9,$7E,$06,$9E,$4C,$7E,$AE,$7E,$9C,$FE,$07,$9D,$E1,$BF,$56,$58
	.dc.b	$7A,$5F,$0C,$AE,$7F,$ED,$BD,$7D,$B5,$DB,$2F,$63,$6B,$DA,$FC,$5F
	.dc.b	$8F,$E8,$F9,$32,$FA,$11,$E7,$C7,$B4,$F1,$3D,$67,$80,$FB,$6E,$8F
	.dc.b	$C7,$E7,$F6,$7E,$3F,$E4,$FC,$0C,$AF,$A9,$1D,$74,$FC,$2C,$75,$BF
	.dc.b	$C6,$6F,$98,$EB,$3A,$9E,$B1,$F7,$7E,$3F,$B3,$D0,$C1,$E5,$3F,$30
	.dc.b	$75,$9F,$8D,$5D,$63,$35,$E9,$9F,$D6,$E3,$F9,$1F,$03,$F9,$D5,$3D
	.dc.b	$2F,$EA,$6B,$89,$97,$E5,$CE,$11,$FE,$D9,$8F,$BD,$87,$A4,$91,$FD
	.dc.b	$17,$87,$EB,$E5,$E1,$8F,$79,$0F,$D2,$9F,$9B,$ED,$F1,$E3,$E2,$C3
	.dc.b	$FB,$0C,$F7,$CB,$3D,$BF,$C1,$AE,$CF,$3A,$F1,$BD,$71,$F0,$E5,$E4
	.dc.b	$8F,$86,$D9,$69,$FB,$3F,$8F,$E7,$4F,$1F,$FD,$34,$FF,$85,$3D,$8C
	.dc.b	$DA,$AF,$37,$E4,$B7,$F7,$5B,$3F,$0B,$E5,$9D,$67,$F4,$A3,$F7,$5F
	.dc.b	$1F,$D3,$09,$1F,$36,$3A,$F3,$FD,$FC,$BD,$B3,$D7,$8F,$EF,$46,$09
	.dc.b	$B7,$F0,$DF,$8F,$F3,$E1,$2F,$EC,$47,$47,$F1,$07,$C7,$BE,$79,$F9
	.dc.b	$A3,$FB,$D5,$FF,$19,$F5,$DF,$05,$F4,$3D,$C7,$5B,$D4,$C3,$70,$FF
	.dc.b	$03,$74,$7E,$33,$0E,$31,$F8,$9C,$BB,$3F,$50,$FC,$46,$7E,$2D,$FE
	.dc.b	$06,$9E,$1F,$4B,$AF,$93,$2F,$AF,$4F,$E6,$B2,$FF,$2F,$47,$5C,$3E
	.dc.b	$46,$5F,$CF,$B9,$7E,$9C,$F7,$27,$1F,$C8,$7F,$99,$5F,$F9,$79,$FF
	.dc.b	$61,$F0,$FB,$0F,$C2,$C6,$63,$F5,$18,$FE,$2F,$0F,$5E,$1D,$B0,$FA
	.dc.b	$91,$EE,$7F,$1F,$5D,$BD,$1F,$8F,$A4,$8F,$59,$FE,$2D,$C5,$3F,$5F
	.dc.b	$9F,$F1,$B7,$F8,$2C,$7B,$63,$D4,$FE,$5D,$87,$AB,$D0,$FA,$7E,$94
	.dc.b	$8E,$A7,$FB,$82,$3A,$74,$FA,$77,$FE,$A5,$87,$47,$E4,$9F,$CB,$2B
	.dc.b	$FA,$E6,$8F,$47,$EF,$EB,$27,$AC,$E5,$FA,$47,$E3,$8F,$A1,$7D,$5C
	.dc.b	$69,$F4,$5E,$BE,$8F,$E2,$B3,$91,$D2,$3C,$EF,$EA,$F2,$F3,$F2,$7E
	.dc.b	$1E,$39,$F2,$7E,$5F,$D1,$FC,$9E,$B2,$7C,$B1,$C9,$FC,$0F,$07,$E1
	.dc.b	$6F,$97,$0F,$CE,$C3,$F4,$1E,$8F,$E7,$33,$98,$7B,$A3,$96,$BF,$7B
	.dc.b	$4F,$1C,$FC,$83,$C6,$7E,$33,$D1,$FD,$2E,$92,$3E,$3C,$72,$D3,$F1
	.dc.b	$3A,$78,$B8,$BF,$8B,$69,$FE,$09,$F1,$7F,$9A,$F4,$FF,$A2,$3B,$3F
	.dc.b	$63,$B7,$EC,$0F,$E3,$33,$AE,$8F,$B9,$A7,$D1,$E4,$FA,$1C,$5F,$8A
	.dc.b	$D7,$F0,$F3,$FE,$17,$B7,$A1,$FB,$AD,$E4,$FE,$22,$72,$FD,$CB,$E5
	.dc.b	$C3,$F0,$23,$ED,$E1,$F9,$FD,$BD,$1E,$87,$F0,$7F,$CD,$3C,$94,$DD
	.dc.b	$18,$86,$5F,$E7,$57,$B7,$0F,$AB,$3F,$ED,$31,$F3,$5F,$AC,$FF,$8A
	.dc.b	$67,$FC,$EA,$FE,$AA,$FA,$EB,$E1,$3E,$2F,$9D,$F3,$F5,$FE,$D5,$BF
	.dc.b	$C4,$CB,$B1,$FC,$D7,$0F,$89,$B7,$A7,$59,$1F,$0E,$39,$3F,$37,$8F
	.dc.b	$EC,$3F,$D4,$F7,$FB,$1E,$77,$C1,$CE,$4F,$CB,$8E,$3B,$FF,$48,$CF
	.dc.b	$C5,$3F,$9D,$3F,$C5,$77,$F8,$1F,$3B,$E5,$65,$27,$EA,$47,$2D,$3F
	.dc.b	$AF,$5F,$29,$FC,$EE,$BF,$B9,$C7,$FC,$D5,$C0,$F3,$7E,$A6,$92,$FE
	.dc.b	$F4,$73,$7F,$A2,$1F,$3F,$1F,$C9,$BF,$3F,$C2,$FB,$99,$C8,$FC,$34
	.dc.b	$71,$7F,$7F,$BF,$C4,$7F,$97,$ED,$F9,$BF,$37,$F0,$98,$C9,$FC,$74
	.dc.b	$71,$EA,$FE,$3D,$7C,$5F,$E0,$DA,$FF,$A0,$3E,$43,$FB,$BF,$A7,$F9
	.dc.b	$4C,$3F,$99,$9F,$ED,$57,$1D,$7F,$E0,$B9,$E9,$C7,$FB,$FA,$F8,$9B
	.dc.b	$FE,$01,$FE,$BE,$7F,$1F,$AE,$0F,$F8,$7E,$BF,$C0,$3A,$F2,$E2,$FD
	.dc.b	$06,$5F,$E3,$8A,$FD,$0C,$74,$7F,$DC,$5F,$21,$FE,$0C,$FA,$77,$BD
	.dc.b	$4F,$F0,$5C,$73,$9F,$F2,$89,$FE,$6A,$7F,$D0,$D8,$48,$FF,$40,$47
	.dc.b	$F0,$3F,$4B,$F5,$9C,$72,$7F,$6D,$1C,$F2,$FD,$3C,$7F,$BD,$3A,$DF
	.dc.b	$94,$3D,$AF,$F4,$E4,$61,$B1,$99,$FF,$4E,$4F,$0F,$83,$FF,$4C,$7C
	.dc.b	$97,$EE,$F9,$E5,$7F,$90,$E3,$C1,$C7,$FE,$52,$8F,$F4,$B4,$76,$F0
	.dc.b	$1A,$53,$9C,$FF,$50,$E2,$FE,$0D,$EB,$48,$FE,$AF,$DD,$FE,$C9,$E6
	.dc.b	$FF,$22,$73,$FF,$5F,$BF,$A3,$9F,$FA,$15,$E7,$7F,$E9,$2C,$FF,$D6
	.dc.b	$18,$7F,$A6,$E3,$FC,$A9,$F2,$5F,$C4,$FD,$79,$1D,$A3,$D2,$FF,$84
	.dc.b	$8F,$EC,$6F,$2B,$91,$83,$FB,$3F,$C9,$7F,$1D,$F6,$24,$7A,$23,$9B
	.dc.b	$FE,$37,$36,$7C,$CD,$FF,$C9,$4D,$F5,$BC,$F3,$E8,$D3,$FC,$A7,$F1
	.dc.b	$4F,$82,$A7,$F9,$5B,$5F,$F7,$27,$07,$C4,$D3,$F6,$D9,$7A,$4F,$F6
	.dc.b	$26,$1D,$1F,$09,$83,$FB,$FF,$C8,$F3,$89,$1F,$42,$3D,$37,$FE,$28
	.dc.b	$3F,$DE,$5D,$3F,$EE,$5A,$F3,$57,$F8,$23,$E4,$3C,$9F,$23,$FB,$4F
	.dc.b	$2A,$4F,$DA,$8E,$7E,$2F,$D4,$AE,$F3,$E3,$3F,$DB,$4F,$B4,$E6,$F1
	.dc.b	$3F,$06,$F2,$7C,$07,$E1,$AF,$C6,$F2,$3F,$13,$F2,$3F,$E7,$CF,$DA
	.dc.b	$FF,$CA,$31,$FC,$18,$38,$DB,$FD,$43,$A2,$BB,$57,$FB,$B5,$F2,$3E
	.dc.b	$23,$A9,$F9,$6F,$90,$ED,$EC,$C8,$C1,$FD,$68,$FE,$79,$B7,$ED,$06
	.dc.b	$5F,$ED,$DA,$FE,$0D,$F0,$72,$B8,$9E,$7C,$7F,$B2,$BF,$E3,$AF,$1D
	.dc.b	$F1,$3F,$C3,$4F,$C2,$79,$FE,$8C,$BA,$D7,$FA,$CE,$3F,$11,$8C,$8F
	.dc.b	$D6,$C7,$E2,$FE,$4B,$FB,$B7,$C0,$C9,$FD,$8C,$73,$D3,$FD,$E4,$F6
	.dc.b	$8E,$26,$8F,$F0,$37,$83,$FD,$FF,$E0,$64,$60,$F8,$47,$F5,$43,$F8
	.dc.b	$FE,$5E,$6F,$19,$EB,$3E,$2B,$FD,$1C,$DA,$FE,$F8,$D7,$F9,$FC,$F0
	.dc.b	$F8,$AF,$F1,$0F,$89,$FE,$0F,$F4,$25,$7F,$86,$63,$D4,$FF,$84,$6F
	.dc.b	$FD,$2C,$F2,$6F,$3F,$F8,$B5,$A2,$38,$7C,$17,$F8,$BB,$C5,$FF,$0E
	.dc.b	$7A,$12,$7F,$C4,$D1,$DB,$3F,$5F,$4F,$A8,$93,$EA,$F8,$1F,$F1,$BF
	.dc.b	$8D,$FE,$25,$F4,$64,$FF,$8B,$E3,$96,$6F,$7F,$E1,$FB,$F2,$BE,$1C
	.dc.b	$7F,$90,$3A,$3F,$C5,$DE,$64,$9F,$F1,$C4,$72,$F7,$BE,$59,$23,$E5
	.dc.b	$FB,$DF,$C2,$FE,$47,$FD,$79,$F9,$B7,$F2,$FC,$FE,$07,$02,$31,$7F
	.dc.b	$EF,$DA,$FE,$25,$E3,$FF,$7D,$F0,$FF,$83,$47,$2F,$42,$57,$F3,$A8
	.dc.b	$FE,$31,$F1,$7F,$EE,$EF,$94,$34,$CC,$FF,$40,$C3,$A1,$FC,$E8,$F3
	.dc.b	$3C,$CD,$10,$FD,$34,$72,$7F,$51,$1E,$67,$F9,$29,$FC,$CC,$FE,$DF
	.dc.b	$D3,$FC,$02,$FA,$C7,$03,$FA,$8F,$C9,$FF,$79,$F4,$24,$60,$87,$E6
	.dc.b	$BC,$4F,$EC,$0B,$C1,$FD,$8A,$D5,$7F,$D5,$0F,$9F,$E9,$49,$E7,$97
	.dc.b	$F1,$2F,$93,$FA,$78,$C9,$FE,$C7,$1F,$D0,$4F,$EF,$3C,$0F,$46,$9F
	.dc.b	$D8,$22,$4F,$F9,$8E,$5E,$87,$DE,$39,$47,$F6,$EF,$6B,$F0,$3E,$8C
	.dc.b	$8C,$10,$F6,$63,$FA,$37,$D9,$87,$67,$27,$E9,$F5,$FF,$05,$38,$3E
	.dc.b	$15,$FE,$3B,$D0,$91,$D7,$4F,$E6,$3E,$CF,$E4,$BC,$E9,$3F,$E0,$28
	.dc.b	$FE,$C0,$7F,$7C,$B9,$1F,$ED,$88,$FE,$FE,$7E,$6B,$CE,$91,$C4,$FF
	.dc.b	$05,$FB,$BF,$9E,$F3,$E4,$60,$87,$F8,$9F,$81,$CD,$F3,$8F,$F0,$5D
	.dc.b	$49,$F9,$9D,$1F,$A8,$3F,$BA,$47,$F7,$5F,$8A,$FC,$F7,$0C,$AF,$F0
	.dc.b	$DC,$65,$FF,$77,$CF,$37,$E3,$70,$27,$FC,$32,$7E,$D3,$A6,$4F,$23
	.dc.b	$FC,$4B,$F2,$3C,$DF,$E8,$D3,$FE,$05,$8C,$74,$FE,$4A,$FA,$57,$FB
	.dc.b	$FA,$BE,$D4,$FE,$5B,$F1,$C7,$AB,$CF,$93,$FB,$1D,$5F,$03,$F0,$15
	.dc.b	$F8,$1C,$3F,$B5,$36,$BF,$DA,$9C,$1F,$9E,$9F,$C8,$76,$7E,$1E,$7F
	.dc.b	$9E,$C7,$F2,$BA,$E1,$7F,$8B,$CF,$CB,$3F,$0B,$1F,$82,$FF,$31,$CF
	.dc.b	$CD,$9F,$3B,$F1,$CF,$94,$FC,$85,$65,$EC,$51,$83,$E3,$FF,$8F,$F4
	.dc.b	$C9,$FE,$7D,$9F,$99,$DC,$F1,$AF,$8B,$1F,$44,$F3,$1E,$5E,$1F,$8B
	.dc.b	$96,$EB,$F9,$5F,$3C,$AE,$B9,$FC,$A9,$E9,$F0,$3F,$C8,$F8,$A5,$68
	.dc.b	$F8,$D1,$E1,$C3,$ED,$E3,$F1,$F5,$FB,$2F,$82,$B8,$1E,$E9,$F1,$38
	.dc.b	$CF,$E0,$36,$3A,$39,$BF,$1B,$92,$1F,$8C,$3F,$10,$7B,$CF,$B5,$1C
	.dc.b	$9F,$C1,$E3,$E0,$7F,$0B,$1F,$27,$E4,$BF,$C6,$5F,$19,$27,$C9,$19
	.dc.b	$7B,$ED,$F4,$71,$F9,$F1,$F8,$E9,$D3,$FE,$00,$E3,$75,$7F,$58,$33
	.dc.b	$7C,$7D,$8F,$F2,$07,$95,$2F,$E1,$47,$DB,$3F,$D6,$9E,$B4,$8C,$90
	.dc.b	$ED,$D3,$FD,$90,$E0,$FC,$BF,$13,$FD,$79,$F2,$52,$30,$43,$F4,$B7
	.dc.b	$C1,$FD,$3C,$FE,$2C,$FE,$DF,$C7,$23,$86,$5F,$07,$1F,$EE,$7D,$72
	.dc.b	$7E,$D4,$7E,$4C,$E4,$F0,$E6,$FD,$AC,$66,$3F,$15,$E4,$3F,$37,$FD
	.dc.b	$7C,$23,$0E,$CF,$F3,$27,$82,$F6,$7F,$55,$3C,$79,$ED,$FE,$2A,$F7
	.dc.b	$5A,$ED,$FE,$9B,$1E,$05,$E2,$E4,$D7,$8D,$BF,$1B,$87,$8D,$2F,$84
	.dc.b	$7E,$BF,$D9,$EF,$DD,$27,$F3,$51,$97,$DB,$9F,$E1,$B8,$BF,$86,$ED
	.dc.b	$EF,$EE,$D5,$7F,$16,$BF,$4F,$64,$8E,$67,$E9,$7A,$3D,$BD,$12,$30
	.dc.b	$43,$FC,$9F,$97,$FA,$CD,$F0,$F4,$7F,$05,$A9,$1E,$4E,$5F,$8E,$61
	.dc.b	$FE,$4F,$F2,$1F,$2B,$29,$1F,$B5,$8F,$E1,$C7,$CE,$A9,$19,$21,$E2
	.dc.b	$8F,$A6,$70,$7F,$7D,$EC,$7D,$62,$38,$21,$EA,$F9,$0F,$B6,$7F,$1D
	.dc.b	$89,$1F,$33,$D7,$FB,$E6,$1F,$E7,$F3,$F8,$2F,$5E,$47,$F0,$D8,$FE
	.dc.b	$4C,$7E,$1B,$C3,$91,$92,$1F,$2B,$C0,$FC,$59,$C1,$FE,$2D,$7F,$8E
	.dc.b	$F5,$64,$60,$87,$D2,$F8,$FF,$C9,$9F,$CD,$66,$47,$E1,$3D,$4F,$CB
	.dc.b	$98,$7F,$AB,$E5,$F9,$AF,$52,$47,$F2,$58,$FE,$7C,$7E,$7B,$E3,$E4
	.dc.b	$64,$87,$FA,$43,$7F,$E8,$DD,$1F,$90,$6F,$85,$FD,$9A,$FF,$A4,$6D
	.dc.b	$FA,$A9,$7C,$7D,$33,$87,$F3,$59,$D7,$C1,$48,$C1,$0F,$07,$C7,$F3
	.dc.b	$38,$BF,$DC,$B8,$7B,$7C,$3C,$8F,$E8,$11,$FD,$64,$F1,$7C,$7C,$8C
	.dc.b	$90,$F6,$7C,$7F,$9C,$E0,$FF,$49,$E4,$F5,$7C,$7C,$8C,$10,$F8,$9F
	.dc.b	$1D,$EF,$3F,$B4,$4C,$8F,$E2,$7B,$3E,$37,$E2,$19,$8F,$F0,$5F,$3F
	.dc.b	$C7,$F8,$D9,$1F,$D6,$63,$FB,$91,$F2,$FE,$22,$46,$48,$7D,$5F,$87
	.dc.b	$FA,$07,$07,$FB,$17,$4F,$D3,$F8,$69,$18,$21,$F7,$7E,$17,$EC,$1F
	.dc.b	$E0,$19,$91,$EA,$F8,$4F,$B8,$61,$FF,$11,$F6,$7D,$FF,$3E,$47,$F7
	.dc.b	$18,$FF,$05,$1F,$83,$F8,$89,$19,$21,$F8,$DF,$88,$FC,$41,$C1,$FE
	.dc.b	$FB,$E9,$A7,$F1,$9F,$11,$27,$31,$FC,$E3,$D4,$73,$D2,$4F,$F4,$48
	.dc.b	$73,$EF,$9E,$4E,$5E,$5C,$BF,$E2,$1F,$0B,$F3,$5F,$11,$2B,$FC,$19
	.dc.b	$19,$7F,$A7,$FE,$71,$FE,$C6,$F4,$0F,$FB,$22,$3F,$C6,$33,$B7,$FD
	.dc.b	$F3,$0F,$89,$FE,$9E,$D8,$EA,$7F,$87,$3D,$EF,$D6,$BF,$D6,$61,$F1
	.dc.b	$1C,$0F,$F1,$F4,$EA,$4F,$0C,$3C,$6F,$63,$7E,$3C,$9F,$4F,$C4,$78
	.dc.b	$8C,$2F,$27,$FC,$89,$5F,$BA,$8F,$7F,$B1,$E8,$3C,$3E,$4C,$8F,$87
	.dc.b	$EC,$FB,$0C,$3C,$F1,$FF,$55,$5F,$38,$F8,$7E,$C7,$C2,$37,$8F,$83
	.dc.b	$CA,$F8,$BE,$CC,$AF,$89,$1F,$00,$F9,$3D,$52,$32,$43,$E7,$FC,$47
	.dc.b	$CE,$38,$3F,$37,$CC,$FA,$3F,$11,$23,$04,$3E,$BF,$C4,$7D,$63,$E7
	.dc.b	$CC,$8F,$C0,$FC,$27,$DA,$30,$FB,$C1,$F5,$3C,$FF,$BF,$F1,$12,$3E
	.dc.b	$F4,$7D,$A3,$F0,$7F,$0F,$23,$24,$3F,$13,$F0,$FF,$88,$38,$3F,$87
	.dc.b	$F4,$7F,$17,$F0,$B2,$30,$43,$F2,$3F,$0B,$F9,$03,$F1,$33,$23,$F3
	.dc.b	$7F,$0B,$F9,$43,$0F,$AC,$FC,$6F,$8B,$F9,$8F,$85,$91,$F9,$78,$FC
	.dc.b	$99,$FE,$2A,$F9,$6F,$F1,$8F,$71,$FF,$19,$CF,$23,$81,$FE,$38,$7F
	.dc.b	$C5,$DC,$FF,$E1,$C6,$F5,$7E,$22,$FF,$3F,$E3,$75,$F8,$59,$87,$E9
	.dc.b	$E3,$2F,$3F,$85,$F1,$18,$7D,$5D,$BF,$C0,$9F,$07,$C2,$F9,$CD,$67
	.dc.b	$FA,$0F,$27,$D5,$F0,$B2,$BF,$69,$1F,$B0,$3D,$DF,$09,$23,$24,$3F
	.dc.b	$05,$BF,$C1,$F8,$4F,$88,$71,$8F,$DF,$79,$9F,$1F,$E1,$24,$60,$87
	.dc.b	$2E,$27,$B9,$F3,$4F,$E0,$35,$27,$E5,$7C,$27,$D1,$30,$FC,$0F,$77
	.dc.b	$EA,$7C,$24,$8F,$E1,$D1,$FC,$28,$FB,$1F,$09,$23,$24,$3C,$5F,$09
	.dc.b	$F7,$4E,$0F,$F1,$7F,$7F,$F0,$1F,$09,$23,$04,$3D,$5F,$09,$F8,$53
	.dc.b	$F8,$CC,$C8,$FC,$17,$C2,$7E,$24,$C3,$F6,$3E,$0B,$F1,$9F,$09,$23
	.dc.b	$F9,$34,$7F,$22,$3F,$21,$F0,$92,$32,$43,$E5,$7C,$17,$E5,$4E,$0F
	.dc.b	$F2,$FF,$84,$FC,$C7,$C0,$C8,$C1,$0F,$A5,$EE,$7E,$74,$FE,$63,$32
	.dc.b	$3F,$D2,$3F,$25,$FC,$C4,$FE,$69,$DE,$9F,$E6,$FF,$0B,$FE,$8E,$9F
	.dc.b	$E4,$5D,$FF,$E7,$A6,$F8,$BF,$C9,$A7,$78,$74,$F5,$78,$18,$7F,$37
	.dc.b	$1C,$FD,$19,$1F,$D2,$63,$FA,$2F,$C2,$F7,$F4,$24,$FF,$4D,$8C,$BE
	.dc.b	$8F,$33,$C8,$61,$FD,$3C,$79,$FC,$A9,$1C,$4F,$EA,$DF,$0B,$EB,$F2
	.dc.b	$A4,$60,$87,$E8,$B5,$78,$FF,$04,$FE,$B3,$52,$7E,$4F,$83,$F1,$4C
	.dc.b	$3F,$B8,$F8,$3F,$91,$E0,$48,$FE,$CD,$1F,$D8,$8F,$99,$D9,$23,$24
	.dc.b	$3C,$5D,$5F,$44,$E0,$FF,$6F,$F8,$5F,$A9,$CD,$23,$04,$3D,$5C,$5F
	.dc.b	$64,$FE,$E3,$32,$3F,$03,$C1,$F7,$4C,$3F,$07,$4F,$C0,$65,$23,$F0
	.dc.b	$F1,$FD,$EB,$E1,$7F,$09,$72,$7F,$C0,$51,$97,$F1,$AF,$E2,$4C,$3F
	.dc.b	$37,$E1,$3F,$DB,$3E,$DF,$F8,$4C,$FF,$0A,$70,$A3,$FC,$0E,$6B,$E1
	.dc.b	$25,$7F,$83,$E3,$2F,$FB,$AA,$BF,$C1,$ED,$F8,$3C,$BF,$C7,$23,$FC
	.dc.b	$1E,$EF,$E0,$FC,$27,$8B,$B7,$FB,$0C,$7F,$49,$FC,$7B,$C9,$F0,$9E
	.dc.b	$63,$3D,$EE,$FD,$CF,$F8,$A4,$F5,$7C,$1F,$B0,$DC,$6B,$E3,$FD,$FF
	.dc.b	$09,$F0,$3E,$12,$65,$E9,$8D,$BF,$DC,$B0,$FF,$17,$3E,$9D,$7E,$39
	.dc.b	$A5,$1A,$CB,$C9,$1C,$78,$3E,$6F,$C2,$48,$FE,$FB,$3A,$78,$CF,$D2
	.dc.b	$F8,$4F,$A7,$F0,$92,$7E,$64,$77,$C3,$E6,$C7,$CB,$F8,$4F,$B2,$67
	.dc.b	$8E,$47,$D3,$F8,$4F,$BA,$61,$FA,$5F,$09,$F8,$0F,$84,$91,$F6,$63
	.dc.b	$B7,$2F,$E1,$3E,$12,$4F,$A3,$93,$E7,$CF,$E2,$3E,$12,$4F,$E0,$5E
	.dc.b	$2F,$E0,$FE,$13,$F1,$BF,$09,$23,$07,$33,$E7,$75,$7D,$0F,$7D,$F9
	.dc.b	$CF,$F8,$08,$FF,$00,$F0,$A7,$51,$FD,$FA,$95,$C3,$E2,$32,$6B,$E1
	.dc.b	$B7,$F0,$DC,$3E,$1A,$5F,$86,$FB,$CF,$CD,$F0,$3A,$FC,$34,$AF,$31
	.dc.b	$F7,$A3,$C3,$F0,$D2,$39,$66,$8F,$2F,$C3,$79,$BE,$1B,$CF,$F0,$B2
	.dc.b	$F7,$1F,$6F,$C5,$F5,$FC,$2C,$9E,$FC,$49,$F7,$FC,$2C,$9E,$EF,$C2
	.dc.b	$F8,$5F,$87,$F0,$BF,$13,$E1,$65,$79,$23,$F1,$1E,$47,$C8,$F8,$59
	.dc.b	$3D,$A3,$E5,$FC,$2C,$8F,$5B,$F3,$BE,$17,$E7,$FC,$2F,$D0,$F8,$59
	.dc.b	$5F,$02,$3F,$29,$E7,$7D,$4F,$85,$93,$EA,$8F,$AF,$F0,$B2,$3E,$3B
	.dc.b	$F7,$FF,$3A,$FF,$48,$1F,$D6,$A3,$ED,$AF,$86,$D1,$BF,$88,$FD,$17
	.dc.b	$A7,$C7,$E1,$E5,$FC,$88,$E9,$F0,$F2,$3E,$F6,$08,$F0,$7C,$3F,$87
	.dc.b	$E1,$FC,$5F,$0F,$2F,$71,$FA,$FF,$63,$CD,$F0,$F2,$7E,$AC,$7A,$7E
	.dc.b	$1E,$47,$67,$D9,$F0,$FE,$DF,$87,$F7,$7C,$3C,$AF,$1C,$7E,$7F,$EB
	.dc.b	$5C,$B4,$93,$F9,$58,$72,$7C,$7F,$88,$F9,$1F,$11,$2B,$DB,$1F,$2F
	.dc.b	$E2,$3E,$67,$C3,$C9,$F8,$31,$FC,$27,$DF,$FA,$1F,$11,$27,$F1,$F1
	.dc.b	$F4,$FE,$22,$47,$C8,$7E,$B7,$C4,$7D,$7F,$87,$FD,$B7,$E7,$7F,$ED
	.dc.b	$93,$FD,$B5,$D5,$FD,$1E,$FF,$33,$F0,$BA,$F8,$69,$18,$3F,$61,$5F
	.dc.b	$C6,$BE,$1F,$F9,$3F,$B0,$8C,$7C,$0F,$FA,$8A,$31,$F0,$DE,$0F,$82
	.dc.b	$F0,$FB,$FE,$2F,$6A,$65,$D2,$3F,$93,$7C,$3F,$9B,$D4,$93,$FA,$C8
	.dc.b	$F4,$F9,$F2,$3C,$6F,$B3,$CB,$F6,$F8,$5F,$D0,$BB,$5E,$93,$E1,$30
	.dc.b	$DC,$3D,$75,$FC,$D7,$E1,$FE,$2F,$54,$9F,$DD,$C7,$C8,$E6,$91,$F0
	.dc.b	$9F,$97,$A7,$4F,$85,$EB,$F0,$B2,$BF,$93,$69,$E2,$9F,$E8,$5F,$0D
	.dc.b	$F4,$BE,$1E,$57,$F0,$88,$FA,$BB,$48,$F9,$EF,$D8,$AF,$4F,$C3,$FB
	.dc.b	$7F,$48,$FE,$DF,$5F,$ED,$1C,$51,$8E,$BF,$EB,$71,$F0,$AB,$FA,$96
	.dc.b	$5F,$CD,$FC,$BF,$F5,$B9,$FC,$D9,$C7,$C4,$67,$E2,$26,$5F,$CB,$EB
	.dc.b	$F9,$C6,$1D,$AB,$C5,$52,$BD,$F9,$7F,$3D,$7F,$B6,$E9,$DF,$E1,$A4
	.dc.b	$7F,$26,$8F,$57,$C3,$FA,$FE,$22,$4F,$28,$F7,$57,$BF,$E1,$E4,$F6
	.dc.b	$8F,$57,$C2,$7A,$FE,$0E,$4F,$9E,$3E,$37,$C3,$FC,$7F,$86,$93,$E6
	.dc.b	$8F,$95,$F0,$FF,$2E,$A4,$FA,$A3,$E3,$7C,$1F,$C7,$F8,$39,$3F,$0E
	.dc.b	$3E,$96,$3F,$4F,$E1,$A4,$FC,$28,$FA,$DF,$0D,$F5,$FC,$19,$3F,$1A
	.dc.b	$3F,$B4,$7E,$7B,$FE,$0B,$EA,$FD,$5B,$FC,$AF,$CA,$34,$D7,$96,$B7
	.dc.b	$4F,$E0,$3E,$1F,$F0,$9A,$C9,$FA,$71,$F8,$7D,$BF,$11,$F0,$D2,$7E
	.dc.b	$BC,$78,$BD,$FF,$1F,$C0,$49,$FE,$DD,$9A,$7F,$21,$F0,$FF,$91,$F7
	.dc.b	$65,$7E,$02,$3F,$29,$EE,$C8,$FC,$1B,$FB,$3F,$D2,$7D,$31,$FE,$00
	.dc.b	$C1,$F0,$1F,$0B,$DE,$93,$E0,$8E,$1F,$0F,$23,$F5,$77,$FD,$3F,$84
	.dc.b	$FF,$4F,$8C,$FC,$3F,$72,$3C,$3E,$1B,$E7,$FB,$DF,$43,$DE,$93,$EE
	.dc.b	$8F,$37,$C3,$79,$F8,$64,$F8,$23,$D5,$C5,$EB,$F8,$69,$3E,$48,$FF
	.dc.b	$11,$7E,$7F,$FE,$1F,$8F,$3F,$99,$F0,$4C,$FB,$5F,$0F,$E1,$E4,$FB
	.dc.b	$1F,$8D,$F0,$FF,$1F,$8F,$E4,$72,$4A,$F8,$31,$F2,$FE,$1F,$E6,$7C
	.dc.b	$34,$9F,$8B,$1F,$3F,$E1,$E4,$7C,$87,$E9,$7C,$3F,$D3,$E2,$93,$F3
	.dc.b	$1F,$AD,$C7,$F5,$FE,$1F,$F9,$57,$E9,$4F,$8D,$A6,$6F,$F8,$EC,$7F
	.dc.b	$9A,$7C,$4F,$CA,$75,$A3,$FA,$1C,$FE,$59,$B1,$9F,$88,$EF,$3F,$73
	.dc.b	$1A,$FF,$1B,$AF,$F4,$7B,$FC,$77,$B3,$FA,$B3,$9C,$7C,$6F,$E1,$32
	.dc.b	$F3,$E1,$E8,$F8,$79,$3E,$68,$F5,$F2,$FB,$3E,$1A,$4F,$AA,$3D,$FF
	.dc.b	$0D,$F0,$3E,$1A,$4F,$BA,$3E,$1F,$C3,$7C,$4F,$8A,$93,$F0,$A3,$E3
	.dc.b	$FC,$3C,$8F,$8A,$FC,$AF,$87,$F9,$7F,$0D,$27,$E4,$BF,$3B,$E1,$BE
	.dc.b	$7F,$C3,$FD,$0F,$86,$95,$F3,$A3,$E9,$FC,$27,$D4,$F8,$19,$3F,$4A
	.dc.b	$3E,$BF,$C3,$7D,$8F,$87,$93,$F5,$A3,$ED,$FC,$3F,$F6,$5C,$3F,$A3
	.dc.b	$76,$7D,$86,$D7,$D7,$9D,$7C,$3F,$82,$3B,$1F,$D7,$27,$FA,$1F,$5B
	.dc.b	$0E,$F3,$F6,$71,$F1,$BF,$86,$C3,$C0,$C3,$07,$3F,$87,$F2,$7C,$2C
	.dc.b	$8C,$1D,$FE,$1B,$D0,$47,$C5,$F0,$BE,$BF,$86,$F6,$7C,$2C,$9F,$34
	.dc.b	$7B,$FE,$1B,$E0,$63,$27,$D5,$1C,$DF,$ED,$5B,$31,$C7,$C0,$7F,$6A
	.dc.b	$86,$5D,$EB,$E1,$65,$DD,$E3,$97,$CD,$F8,$5F,$9D,$F0,$D2,$7E,$44
	.dc.b	$74,$E3,$FA,$19,$D9,$FE,$77,$97,$67,$D2,$76,$21,$D8,$FA,$5F,$0B
	.dc.b	$F6,$C8,$FD,$5F,$86,$FB,$DF,$0B,$23,$EC,$3F,$81,$F8,$6F,$C1,$7C
	.dc.b	$24,$9F,$B8,$FE,$17,$E1,$7F,$0D,$72,$7A,$1F,$81,$F8,$4E,$8F,$C8
	.dc.b	$C2,$F1,$7C,$7C,$7A,$4F,$CC,$C6,$CF,$CB,$C7,$F2,$5F,$0B,$F9,$3F
	.dc.b	$84,$97,$F8,$D8,$FC,$B7,$C2,$FE,$5F,$E1,$64,$FE,$46,$3F,$C7,$DF
	.dc.b	$A5,$FF,$4C,$F4,$CF,$F4,$52,$71,$EB,$E3,$3B,$DF,$53,$FC,$6B,$3F
	.dc.b	$D2,$5F,$C0,$DF,$67,$F0,$78,$78,$9D,$DF,$83,$E1,$A4,$60,$E7,$F0
	.dc.b	$DE,$4F,$86,$91,$83,$BF,$C2,$FA,$3E,$1A,$46,$0F,$1F,$C2,$FB,$08
	.dc.b	$F9,$BE,$1B,$DF,$F0,$BF,$03,$19,$3E,$A8,$F8,$78,$74,$8F,$F2,$26
	.dc.b	$DF,$DE,$5E,$C6,$30,$EA,$D3,$B1,$F1,$3E,$17,$E6,$91,$F8,$FF,$0B
	.dc.b	$D9,$F9,$39,$F7,$E6,$FF,$27,$E0,$C7,$AC,$63,$1B,$BF,$F4,$6A,$6B
	.dc.b	$F6,$BE,$13,$ED,$FC,$2C,$CF,$EA,$C7,$DE,$F8,$4F,$BF,$F0,$B2,$7E
	.dc.b	$CC,$7E,$0B,$E1,$7F,$07,$F0,$B2,$7E,$EC,$7E,$1B,$E1,$7B,$47,$43
	.dc.b	$F0,$2F,$57,$F0,$77,$D5,$F9,$37,$D5,$FC,$3D,$F5,$7E,$6D,$FE,$4B
	.dc.b	$E1,$7F,$27,$F0,$72,$7F,$1B,$1F,$E9,$3F,$D3,$3F,$B5,$47,$FA,$11
	.dc.b	$AF,$F9,$E5,$BC,$3F,$E7,$87,$FB,$64,$7F,$8F,$9F,$ED,$24,$F8,$4E
	.dc.b	$9F,$17,$E1,$9E,$E4,$3A,$7C,$1F,$87,$E1,$A4,$78,$5F,$27,$C3,$79
	.dc.b	$6E,$4F,$95,$EA,$F9,$FE,$1B,$D4,$47,$D5,$F0,$BE,$DF,$86,$91,$ED
	.dc.b	$7E,$07,$C2,$FC,$1F,$86,$93,$F0,$5F,$89,$F0,$DF,$16,$FA,$4F,$C3
	.dc.b	$C3,$0F,$32,$1F,$1F,$0E,$AF,$32,$1F,$2E,$FA,$BE,$62,$1F,$3E,$BE
	.dc.b	$A7,$C3,$7A,$23,$A1,$DB,$83,$E9,$65,$E8,$8E,$87,$68,$FA,$D9,$7D
	.dc.b	$DF,$86,$FB,$DF,$0D,$2B,$EF,$47,$E0,$7D,$59,$1F,$81,$7F,$09,$F0
	.dc.b	$DF,$85,$F8,$59,$3F,$85,$7F,$11,$F0,$DD,$63,$99,$0F,$C4,$7C,$27
	.dc.b	$47,$F1,$57,$73,$FC,$56,$1D,$27,$F1,$D8,$59,$FC,$76,$1F,$95,$F8
	.dc.b	$49,$5F,$95,$7F,$31,$F0,$9F,$EE,$BF,$D3,$7F,$C0,$BE,$99,$A4,$F8
	.dc.b	$7D,$7C,$2C,$C3,$FD,$C3,$28,$E5,$F0,$DD,$27,$B1,$0F,$D1,$61,$DD
	.dc.b	$87,$23,$C5,$7D,$5F,$F7,$05,$F5,$7B,$DF,$67,$C3,$F0,$1E,$9F,$85
	.dc.b	$91,$83,$CB,$F0,$9E,$DF,$85,$91,$83,$D3,$F0,$DF,$04,$8F,$B3,$E1
	.dc.b	$7E,$26,$12,$3A,$1E,$4F,$86,$F9,$18,$74,$8F,$3E,$18,$7A,$90,$F5
	.dc.b	$E1,$F3,$F0,$FA,$17,$27,$E5,$BF,$4F,$E1,$B9,$47,$43,$E7,$DE,$38
	.dc.b	$FF,$E0,$0E,$AC,$7D,$35,$D8,$F5,$1F,$5B,$E1,$BE,$F7,$C2,$C8,$C1
	.dc.b	$F6,$BE,$17,$F0,$44,$7E,$EF,$C2,$FE,$17,$E1,$7F,$0D,$F0,$D2,$7F
	.dc.b	$03,$1F,$89,$F8,$5F,$C5,$63,$27,$F0,$B1,$CD,$FB,$78,$5D,$3D,$38
	.dc.b	$F5,$9F,$C5,$E3,$D5,$FC,$0E,$3F,$95,$F8,$69,$1F,$90,$7F,$31,$F0
	.dc.b	$BF,$99,$F8,$59,$3F,$94,$7F,$DD,$4B,$17,$B9,$FE,$9A,$F5,$BF,$C1
	.dc.b	$87,$FB,$3C,$9F,$0B,$C2,$77,$5F,$EF,$29,$FF,$05,$47,$D0,$C3,$C3
	.dc.b	$3F,$A2,$78,$5F,$5F,$84,$F1,$CF,$E6,$B1,$B7,$D3,$BD,$8F,$8B,$2E
	.dc.b	$37,$DD,$B3,$A6,$CF,$8B,$1F,$57,$C2,$E3,$E0,$66,$03,$E0,$4E,$9C
	.dc.b	$63,$E1,$BF,$F0,$44,$68,$F7,$7C,$0F,$63,$BF,$C1,$78,$3E,$06,$4F
	.dc.b	$2C,$3E,$54,$79,$3E,$17,$CB,$9C,$AD,$1F,$25,$F4,$FC,$2F,$48,$CE
	.dc.b	$DF,$0B,$2F,$04,$63,$3F,$2E,$9E,$7C,$BC,$33,$E7,$BF,$93,$1E,$18
	.dc.b	$E1,$87,$85,$D5,$B1,$F0,$C7,$5D,$3B,$BA,$CB,$C8,$F0,$C3,$BE,$5E
	.dc.b	$18,$F2,$67,$F4,$BE,$17,$E9,$FC,$1C,$9F,$44,$7D,$6F,$84,$91,$EB
	.dc.b	$7E,$CE,$9F,$6B,$E0,$E4,$FB,$DF,$BB,$F0,$5D,$63,$99,$0F,$5F,$C1
	.dc.b	$F4,$7D,$B7,$73,$FB,$98,$74,$9F,$83,$85,$9F,$C0,$61,$F8,$AF,$84
	.dc.b	$95,$F3,$DF,$C6,$FC,$1F,$E3,$BE,$12,$4F,$D3,$7F,$27,$FA,$77,$F8
	.dc.b	$6B,$9B,$F1,$DC,$C8,$FF,$0D,$B7,$E0,$36,$70,$F8,$1E,$F8,$7D,$4C
	.dc.b	$AE,$9F,$93,$CF,$BC,$FD,$8C,$EC,$FE,$9F,$2F,$04,$B2,$F0,$63,$DA
	.dc.b	$3F,$17,$2C,$BC,$11,$8F,$84,$F4,$91,$DF,$C2,$7B,$3E,$0F,$DB,$F0
	.dc.b	$32,$79,$C7,$C0,$F8,$39,$1D,$9F,$87,$F0,$3F,$12,$FA,$4F,$E5,$72
	.dc.b	$63,$98,$E6,$79,$B1,$EC,$EF,$1E,$AF,$1C,$6E,$BF,$07,$2E,$53,$D7
	.dc.b	$0E,$F3,$F1,$72,$F5,$4F,$52,$1D,$4F,$41,$0E,$C7,$C8,$AF,$BB,$F0
	.dc.b	$72,$3E,$5B,$F8,$0F,$82,$FC,$0F,$C1,$C9,$F9,$EF,$E1,$3E,$0B,$F0
	.dc.b	$BF,$07,$27,$E9,$BF,$88,$F8,$2E,$51,$D0,$87,$C7,$BE,$8F,$C9,$BB
	.dc.b	$9F,$E0,$70,$E9,$3F,$37,$0B,$3F,$85,$C3,$F2,$BF,$05,$F9,$6F,$82
	.dc.b	$97,$F8,$48,$FC,$CF,$C1,$48,$FC,$3B,$F9,$CF,$83,$FC,$EF,$C1,$C9
	.dc.b	$FC,$5B,$FA,$0F,$82,$EB,$1C,$C8,$7F,$13,$FD,$3F,$FC,$53,$CE,$C3
	.dc.b	$B4,$FE,$0D,$EC,$F4,$21,$F8,$7B,$EC,$FF,$24,$A7,$C1,$FF,$42,$7F
	.dc.b	$C4,$51,$DF,$E0,$F1,$39,$F8,$4D,$7C,$1C,$AE,$51,$C7,$E1,$39,$7C
	.dc.b	$24,$9E,$D1,$C9,$F0,$5D,$7D,$9F,$03,$D5,$F2,$61,$D5,$F8,$18,$75
	.dc.b	$7D,$18,$75,$7E,$26,$1C,$F8,$FF,$B0,$67,$EC,$9E,$E7,$BF,$1E,$CF
	.dc.b	$CB,$C7,$B5,$70,$C7,$B4,$FC,$FC,$7B,$4F,$4C,$7E,$37,$C0,$4A,$F9
	.dc.b	$4F,$C9,$F8,$4F,$95,$F0,$92,$7E,$73,$F3,$7E,$13,$E7,$7C,$17,$CF
	.dc.b	$F8,$39,$5F,$4E,$3A,$3F,$7F,$0B,$97,$B3,$1E,$93,$F8,$3C,$6C,$FC
	.dc.b	$0C,$7A,$CF,$DD,$C7,$AB,$F8,$9C,$7E,$F7,$C1,$48,$FC,$13,$F8,$1F
	.dc.b	$82,$FC,$17,$C0,$FE,$0F,$E0,$A5,$7E,$1E,$3F,$0D,$F0,$72,$3F,$14
	.dc.b	$FE,$27,$E0,$FA,$47,$E5,$F0,$BA,$7D,$4C,$7A,$CF,$E4,$71,$EA,$FE
	.dc.b	$77,$1E,$8F,$E7,$F1,$B7,$F7,$31,$FC,$B7,$C1,$FE,$5F,$E0,$A5,$FE
	.dc.b	$6E,$3F,$C3,$7F,$A8,$7F,$8B,$23,$FB,$D2,$F8,$3C,$9A,$F5,$77,$F1
	.dc.b	$3C,$3E,$16,$5F,$F8,$22,$FF,$C4,$4F,$F4,$3E,$C6,$3E,$08,$FD,$A6
	.dc.b	$5E,$47,$86,$3D,$9E,$59,$77,$7A,$65,$D9,$F2,$E5,$E2,$7D,$44,$3F
	.dc.b	$80,$61,$EC,$F8,$3E,$51,$EC,$BE,$CF,$BB,$0E,$4F,$43,$D5,$78,$D7
	.dc.b	$FC,$89,$A3,$E1,$7E,$39,$2B,$E0,$7C,$27,$CA,$F8,$49,$1F,$0D,$F9
	.dc.b	$BF,$09,$F3,$BE,$12,$4F,$C7,$7E,$8F,$C2,$F3,$8F,$A3,$85,$CB,$C3
	.dc.b	$8F,$49,$FA,$B8,$D9,$F2,$E3,$D2,$7E,$CE,$36,$7D,$38,$FD,$EF,$84
	.dc.b	$95,$F5,$DF,$C0,$FC,$2F,$E0,$BE,$12,$4F,$DB,$7F,$0B,$F0,$9F,$86
	.dc.b	$F8,$4F,$C3,$FC,$24,$AF,$C0,$47,$47,$F1,$38,$5C,$BE,$4E,$3D,$27
	.dc.b	$F1,$B8,$D9,$F9,$B8,$F5,$9F,$C5,$63,$D5,$FC,$9E,$3F,$96,$F8,$59
	.dc.b	$1F,$8F,$7F,$33,$F0,$BF,$9A,$F8,$59,$3F,$93,$7F,$3B,$F0,$BF,$E7
	.dc.b	$8F,$07,$FC,$2F,$F0,$A6,$93,$E1,$7A,$DF,$E8,$B1,$BA,$FE,$0B,$2E
	.dc.b	$B3,$FA,$6C,$AC,$FE,$1B,$2E,$DF,$0F,$2B,$FC,$D7,$97,$57,$F4,$D8
	.dc.b	$74,$EA,$FF,$1B,$F4,$B3,$F5,$D7,$A4,$87,$73,$F5,$D8,$F5,$7F,$CE
	.dc.b	$18,$7A,$B1,$F7,$69,$EF,$F8,$6F,$81,$F0,$72,$FC,$71,$F0,$FE,$0E
	.dc.b	$47,$99,$F8,$DF,$0B,$F1,$FE,$1B,$E4,$7C,$34,$AF,$5C,$75,$7D,$B8
	.dc.b	$75,$7C,$78,$74,$7C,$B8,$5E,$1F,$F5,$76,$8C,$BB,$57,$C6,$CB,$B3
	.dc.b	$EC,$CB,$EC,$7C,$37,$D9,$F8,$69,$3F,$32,$3E,$E7,$C3,$7D,$DF,$86
	.dc.b	$93,$F4,$23,$F0,$1F,$0D,$23,$E9,$BF,$83,$F8,$6E,$B1,$F5,$B0,$EA
	.dc.b	$FC,$CC,$3A,$BF,$6B,$0E,$AF,$D0,$C3,$A3,$F4,$B0,$BA,$FB,$B1,$FC
	.dc.b	$87,$C3,$4A,$FC,$1B,$F9,$3F,$85,$FC,$A7,$C3,$49,$FC,$3B,$F9,$7F
	.dc.b	$83,$FC,$C7,$C3,$7E,$67,$E1,$25,$7E,$32,$39,$3D,$08,$7E,$12,$FA
	.dc.b	$3F,$86,$BB,$97,$D2,$C3,$AC,$FE,$5B,$0E,$AF,$E3,$30,$EA,$FE,$3B
	.dc.b	$1F,$53,$D4,$F3,$90,$EC,$7E,$82,$BA,$BF,$96,$AE,$B7,$F8,$2B,$FD
	.dc.b	$A7,$C3,$4A,$FD,$3B,$FE,$EE,$FD,$47,$FD,$D1,$1F,$ED,$47,$FC,$2C
	.dc.b	$FF,$B3,$DB,$5F,$EF,$96,$F7,$7F,$A8,$9F,$F1,$AC,$67,$E0,$BA,$12
	.dc.b	$BA,$7C,$1F,$77,$C4,$7E,$B3,$0C,$3E,$1F,$85,$F2,$11,$F2,$7C,$37
	.dc.b	$37,$C9,$77,$0F,$26,$1E,$BF,$84,$95,$EB,$7D,$DF,$0B,$EF,$F8,$69
	.dc.b	$3E,$F7,$E1,$7C,$37,$C3,$F8,$69,$3F,$0D,$F8,$DF,$0D,$D2,$3E,$26
	.dc.b	$17,$2F,$89,$8F,$49,$F9,$18,$D9,$F9,$18,$F4,$9F,$99,$8D,$9F,$99
	.dc.b	$8F,$D3,$F8,$69,$5F,$4D,$FA,$DF,$0B,$F5,$FE,$1B,$EC,$7C,$34,$AF
	.dc.b	$B1,$1F,$6F,$C4,$91,$F6,$DF,$BD,$F0,$BD,$23,$EE,$61,$72,$FB,$98
	.dc.b	$F5,$9F,$C1,$63,$D5,$FC,$16,$3D,$1F,$C2,$63,$6F,$F0,$98,$F3,$E6
	.dc.b	$FC,$26,$56,$FF,$DB,$1C,$CC,$FB,$47,$E3,$73,$ED,$8F,$E3,$73,$E9
	.dc.b	$8F,$E3,$73,$B7,$CB,$3F,$CC,$7C,$34,$8F,$CC,$3F,$9B,$F8,$6F,$CE
	.dc.b	$7C,$37,$E7,$7E,$0E,$57,$E7,$63,$F4,$1F,$0B,$FA,$0D,$19,$FF,$BF
	.dc.b	$3F,$52,$FF,$1B,$46,$2F,$FD,$E1,$18,$9F,$D0,$57,$86,$7F,$4D,$95
	.dc.b	$BF,$D3,$65,$E2,$9F,$D6,$65,$E2,$7F,$59,$97,$89,$0A,$F0,$7A,$CC
	.dc.b	$FC,$13,$E2,$CF,$83,$B7,$D1,$F0,$9E,$B6,$C7,$4B,$E7,$5F,$E4,$F8
	.dc.b	$C4,$FB,$FE,$1B,$84,$7B,$BE,$16,$E3,$E1,$6D,$9B,$F0,$D8,$F8,$6C
	.dc.b	$FC,$34,$C3,$A9,$F2,$23,$8E,$92,$33,$8F,$B6,$3A,$E5,$DB,$19,$5A
	.dc.b	$CB,$C0,$DC,$BC,$0E,$72,$F0,$3F,$0C,$F0,$1C,$7E,$1B,$D2,$47,$A7
	.dc.b	$C3,$7B,$3E,$1B,$DB,$F0,$D2,$7C,$11,$F0,$3E,$1B,$E0,$FC,$34,$9F
	.dc.b	$24,$73,$7C,$1B,$FD,$7C,$71,$C5,$F4,$32,$ED,$5E,$BC,$BB,$3C,$35
	.dc.b	$67,$DE,$3E,$06,$7D,$DD,$67,$D3,$8B,$F2,$D9,$D9,$FC,$AE,$5D,$9E
	.dc.b	$99,$76,$CF,$F0,$39,$75,$9F,$06,$5D,$67,$F0,$B9,$7D,$EF,$86,$95
	.dc.b	$F4,$5F,$C0,$FC,$37,$E0,$BE,$17,$F0,$7F,$0D,$2B,$EB,$47,$E1,$BE
	.dc.b	$0F,$F0,$FF,$0B,$27,$ED,$47,$47,$E0,$61,$72,$FC,$AE,$3D,$67,$F0
	.dc.b	$18,$F5,$7E,$36,$3D,$5F,$C2,$63,$D5,$F9,$58,$FE,$5B,$E1,$BF,$2F
	.dc.b	$F0,$B2,$7F,$15,$1F,$9C,$FD,$4F,$FC,$AA,$3E,$17,$06,$7E,$1B,$5F
	.dc.b	$0B,$2F,$F2,$F4,$8E,$5F,$0D,$DE,$7F,$2D,$8F,$77,$EE,$63,$CD,$E3
	.dc.b	$8F,$67,$9E,$3D,$5F,$C4,$E3,$D5,$EF,$8F,$67,$C3,$F0,$DE,$9F,$86
	.dc.b	$91,$83,$CB,$F0,$BE,$DF,$86,$91,$83,$D3,$F0,$BF,$04,$8F,$B3,$E1
	.dc.b	$BE,$26,$1D,$23,$C7,$86,$1E,$47,$9A,$FA,$BE,$8C,$39,$BE,$AC,$2F
	.dc.b	$2F,$F1,$AC,$D9,$73,$AE,$A7,$CC,$C7,$15,$BC,$3A,$BC,$71,$FB,$11
	.dc.b	$D3,$B7,$F9,$3F,$7B,$26,$79,$C9,$39,$CF,$F0,$1F,$0B,$27,$ED,$BF
	.dc.b	$83,$F8,$6F,$C2,$7C,$37,$E1,$7E,$1A,$57,$E0,$23,$F1,$1F,$09,$F8
	.dc.b	$9F,$86,$93,$F8,$48,$E8,$FD,$AC,$2F,$07,$C1,$C7,$AC,$FE,$2B,$1E
	.dc.b	$AF,$E0,$31,$EA,$FE,$3B,$1E,$AF,$E1,$31,$FC,$C7,$C3,$7E,$67,$E1
	.dc.b	$A4,$FE,$52,$3F,$39,$F0,$B2,$3F,$2E,$FE,$7F,$E1,$BF,$41,$F0,$D2
	.dc.b	$7F,$36,$FF,$3E,$C7,$FC,$83,$1F,$92,$BE,$EF,$53,$F9,$BC,$7F,$90
	.dc.b	$1F,$9B,$C4,$C3,$C1,$1F,$9A,$7C,$0F,$62,$1F,$9E,$BF,$03,$E1,$F8
	.dc.b	$58,$F6,$7C,$9F,$0B,$E5,$F8,$59,$3E,$27,$D1,$F0,$BE,$9F,$85,$93
	.dc.b	$E6,$7D,$9F,$0B,$D2,$3F,$9A,$E9,$FE,$56,$E9,$FC,$4B,$63,$F0,$EB
	.dc.b	$1E,$F1,$C3,$1B,$CB,$FD,$D2,$99,$78,$2B,$AE,$56,$B5,$97,$85,$E1
	.dc.b	$9D,$CF,$C3,$9F,$47,$C7,$9D,$C7,$C7,$97,$39,$F3,$E5,$67,$CF,$97
	.dc.b	$D9,$F8,$5F,$B5,$F0,$B2,$7E,$BC,$7D,$DF,$85,$FB,$DF,$0B,$27,$ED
	.dc.b	$C7,$E0,$7E,$13,$F0,$5F,$07,$27,$EF,$C7,$57,$F0,$38,$75,$7E,$5E
	.dc.b	$1D,$5F,$C2,$E1,$D5,$F9,$F8,$75,$7F,$13,$87,$57,$E9,$E1,$F9,$1F
	.dc.b	$84,$91,$F8,$E7,$F2,$9F,$09,$F9,$5F,$84,$93,$F9,$27,$F3,$1F,$07
	.dc.b	$F9,$9F,$83,$93,$F9,$67,$F3,$9F,$05,$D2,$3F,$03,$85,$E6,$FB,$58
	.dc.b	$F4,$9F,$C2,$E3,$67,$EF,$63,$D2,$7F,$13,$8D,$9F,$C1,$63,$FA,$9F
	.dc.b	$83,$95,$FA,$67,$F5,$9F,$09,$FE,$2A,$FD,$53,$FC,$73,$E3,$9F,$F1
	.dc.b	$95,$7F,$82,$1B,$87,$F9,$01,$D8,$F8,$3E,$44,$9F,$F1,$4C,$7F,$89
	.dc.b	$1F,$CE,$E1,$E2,$69,$FB,$8C,$3C,$5C,$FE,$3F,$84,$93,$CA,$3A,$3E
	.dc.b	$3C,$2E,$3E,$3F,$84,$E8,$FE,$87,$1B,$7F,$AA,$C7,$AE,$1E,$4C,$7A
	.dc.b	$BE,$D9,$E6,$7F,$38,$CB,$D9,$1E,$AC,$BE,$36,$5F,$1F,$E1,$3E,$47
	.dc.b	$C2,$CA,$F7,$C7,$CB,$F8,$5F,$99,$F0,$92,$7E,$1C,$7C,$FF,$84,$FA
	.dc.b	$1F,$07,$27,$E3,$C7,$57,$E4,$E1,$D5,$FA,$58,$75,$7E,$6E,$1D,$5F
	.dc.b	$AD,$87,$57,$E8,$E1,$D5,$FB,$58,$7D,$FF,$84,$FC,$07,$C1,$C9,$FA
	.dc.b	$F1,$F8,$3F,$84,$FC,$27,$C2,$49,$FB,$71,$F8,$7F,$84,$91,$F7,$9F
	.dc.b	$C5,$7C,$1F,$58,$FC,$0E,$1D,$5F,$C5,$61,$D5,$FC,$2E,$1D,$5F,$C7
	.dc.b	$61,$D1,$FC,$86,$17,$9B,$F2,$18,$FE,$5F,$E1,$25,$7E,$39,$FC,$D7
	.dc.b	$C1,$FE,$6F,$E1,$3F,$39,$F0,$52,$BF,$27,$1F,$9F,$F8,$39,$1F,$96
	.dc.b	$7F,$D1,$7F,$AA,$BE,$0F,$AC,$FE,$87,$1B,$9F,$E8,$72,$ED,$3F,$9E
	.dc.b	$CB,$B3,$FA,$7C,$BA,$BF,$A9,$CA,$DF,$FA,$A3,$2E,$D3,$E9,$21,$DC
	.dc.b	$FD,$4E,$1D,$9F,$D9,$61,$DA,$BF,$65,$87,$4E,$AF,$CC,$63,$6F,$FC
	.dc.b	$1B,$8F,$67,$FD,$02,$FF,$92,$DF,$F4,$EB,$FE,$3B,$73,$F0,$7F,$0B
	.dc.b	$E1,$3E,$1F,$C2,$C9,$DB,$F1,$BE,$17,$E3,$FC,$2F,$C8,$F8,$49,$5D
	.dc.b	$23,$93,$DB,$0F,$9A,$61,$F1,$DF,$56,$9E,$4B,$EA,$F9,$AF,$AB,$EA
	.dc.b	$BF,$AF,$F0,$BF,$63,$E1,$64,$FB,$23,$ED,$FC,$2C,$8F,$7B,$F7,$BE
	.dc.b	$17,$EF,$FC,$2F,$E0,$3E,$12,$57,$C4,$8E,$AF,$C6,$C3,$AB,$F2,$B0
	.dc.b	$E8,$FC,$CC,$2F,$1F,$DD,$C7,$AC,$FC,$EC,$7A,$BF,$4B,$1F,$C7,$FC
	.dc.b	$2C,$8F,$A6,$FE,$4B,$E1,$7F,$27,$EF,$C9,$FA,$EF,$E5,$BE,$0F,$F2
	.dc.b	$FF,$0B,$F9,$8F,$83,$95,$F7,$23,$A3,$F8,$0C,$2E,$9F,$91,$C7,$A4
	.dc.b	$FE,$13,$1B,$3F,$95,$C7,$AC,$FE,$1B,$1E,$AF,$E2,$B1,$E7,$D1,$FE
	.dc.b	$5E,$E7,$67,$8A,$FC,$8E,$7D,$2B,$59,$D9,$FC,$AE,$5D,$9F,$CA,$E5
	.dc.b	$DB,$3F,$D3,$67,$FB,$4F,$85,$FD,$AF,$C2,$4B,$FC,$CC,$1F,$F7,$BF
	.dc.b	$EA,$DF,$E4,$6B,$67,$FE,$72,$8F,$F2,$27,$A6,$69,$B8,$D7,$C2,$F3
	.dc.b	$BF,$F7,$3C,$FF,$8C,$6B,$84,$E6,$FA,$D7,$F9,$EA,$BF,$C6,$8F,$90
	.dc.b	$FD,$6E,$3E,$77,$F6,$38,$F1,$C7,$C9,$14,$E5,$F1,$37,$2F,$33,$70
	.dc.b	$F3,$4F,$7C,$B0,$47,$E0,$FC,$36,$FE,$1B,$87,$C3,$49,$F8,$B1,$CF
	.dc.b	$E1,$BA,$7C,$34,$9F,$93,$1D,$FE,$17,$C1,$F0,$D2,$7E,$6C,$75,$7E
	.dc.b	$7E,$1D,$5F,$8F,$87,$57,$E9,$E1,$D5,$F9,$78,$75,$7E,$BE,$1D,$5F
	.dc.b	$9F,$87,$BF,$E1,$7E,$07,$C3,$49,$FB,$B1,$F0,$FE,$17,$E2,$7C,$34
	.dc.b	$9F,$C0,$C7,$C7,$F8,$69,$1F,$84,$7E,$57,$C3,$75,$8F,$C3,$E1,$D5
	.dc.b	$FB,$F8,$75,$7F,$17,$87,$57,$F0,$78,$74,$7F,$0B,$85,$E7,$FC,$1F
	.dc.b	$27,$E3,$72,$ED,$5F,$8A,$CB,$B5,$6B,$8D,$9F,$7A,$FC,$7E,$7D,$EB
	.dc.b	$59,$F5,$E5,$FC,$D6,$96,$FF,$21,$9F,$E1,$7E,$1A,$47,$E8,$9F,$C4
	.dc.b	$7C,$37,$E2,$7E,$1A,$4F,$E9,$9F,$C6,$7C,$37,$E3,$7E,$1A,$4F,$EA
	.dc.b	$9F,$C8,$7C,$37,$48,$FD,$1E,$17,$5F,$56,$3D,$27,$F4,$F8,$D9,$F7
	.dc.b	$63,$D2,$7F,$57,$8D,$9F,$85,$8F,$E7,$FD,$7F,$F2,$89,$35,$C7,$C3
	.dc.b	$67,$E1,$66,$3F,$BF,$C3,$E7,$DF,$F8,$DA,$11,$8F,$85,$E9,$F0,$92
	.dc.b	$78,$63,$FE,$F4,$B4,$F8,$7E,$17,$C5,$32,$B6,$7F,$BD,$AB,$CB,$97
	.dc.b	$8A,$3B,$CF,$3C,$B1,$97,$86,$78,$F4,$30,$EF,$18,$6B,$C8,$ED,$97
	.dc.b	$8D,$D7,$C2,$7C,$5F,$85,$91,$83,$97,$C2,$7C,$9F,$85,$91,$83,$B7
	.dc.b	$C2,$7C,$D2,$3E,$1F,$85,$FA,$18,$74,$8F,$16,$7E,$7C,$73,$1C,$CC
	.dc.b	$E6,$C7,$BC,$63,$2E,$AE,$B2,$BA,$7D,$6C,$F9,$CF,$2C,$7C,$13,$F6
	.dc.b	$B3,$FC,$16,$1D,$23,$EE,$E3,$8B,$F5,$90,$EC,$7A,$CF,$99,$F0,$9F
	.dc.b	$8B,$23,$F3,$FE,$17,$F1,$DF,$0B,$23,$E9,$3F,$91,$D7,$F2,$5F,$09
	.dc.b	$27,$EB,$3F,$95,$F8,$4E,$71,$F0,$B0,$B9,$7E,$4B,$1E,$93,$F1,$B1
	.dc.b	$B3,$F9,$6C,$7A,$4F,$CA,$C6,$CF,$E6,$B1,$FD,$17,$C1,$FE,$8F,$E1
	.dc.b	$25,$FE,$1E,$3F,$4D,$F0,$92,$3F,$14,$FF,$47,$FD,$5F,$FC,$7D,$87
	.dc.b	$F1,$27,$FC,$7D,$3F,$71,$B1,$9F,$85,$E1,$F0,$BE,$0A,$FC,$A6,$3E
	.dc.b	$07,$EF,$E3,$DD,$FC,$0E,$37,$0E,$DF,$0D,$E1,$C6,$57,$E8,$32,$EF
	.dc.b	$D5,$FD,$D7,$A1,$1F,$E4,$FC,$3C,$DF,$0D,$E9,$25,$F1,$F8,$6F,$67
	.dc.b	$C2,$FB,$7E,$1A,$4F,$58,$F8,$1F,$0B,$23,$C0,$FC,$3F,$86,$E7,$1F
	.dc.b	$07,$0B,$8F,$C5,$C7,$AC,$F9,$F1,$EA,$FC,$7C,$7A,$3F,$27,$1B,$7F
	.dc.b	$37,$1E,$7A,$7B,$73,$67,$8C,$1A,$7C,$0D,$3B,$C7,$D3,$D3,$BE,$5F
	.dc.b	$5F,$4E,$B9,$73,$D2,$DF,$2C,$DA,$7D,$FF,$86,$93,$F3,$DF,$C1,$7C
	.dc.b	$37,$E0,$FE,$17,$F0,$9F,$0D,$2B,$EA,$47,$E1,$FE,$16,$47,$D7,$7F
	.dc.b	$15,$F0,$BD,$23,$F1,$18,$5D,$BF,$19,$8F,$59,$FB,$D8,$F5,$7F,$1D
	.dc.b	$8F,$47,$F2,$18,$DB,$FC,$A6,$3F,$97,$F8,$5F,$CC,$7C,$2C,$BF,$C4
	.dc.b	$47,$E6,$FE,$12,$47,$E2,$DF,$CF,$7C,$2F,$E7,$FE,$12,$4F,$E3,$DF
	.dc.b	$D1,$7C,$2F,$58,$FC,$96,$1D,$5F,$D1,$61,$D1,$FD,$26,$17,$2F,$D4
	.dc.b	$63,$D2,$7F,$51,$8D,$9F,$D6,$63,$FE,$35,$FD,$63,$FC,$67,$D0,$69
	.dc.b	$3E,$1B,$5F,$0D,$31,$FE,$87,$3F,$E4,$B3,$FC,$8E,$FF,$30,$5F,$09
	.dc.b	$D7,$E1,$64,$F0,$8F,$03,$0E,$67,$8E,$FB,$3C,$EE,$E5,$FE,$14,$C3
	.dc.b	$BB,$6F,$AD,$78,$70,$F6,$47,$58,$F2,$16,$79,$61,$D3,$A7,$F7,$98
	.dc.b	$7A,$F1,$F8,$9F,$0D,$CB,$0F,$7E,$5D,$B1,$F0,$47,$C2,$F8,$5F,$94
	.dc.b	$47,$E2,$FC,$37,$CD,$F8,$5F,$9D,$F0,$D2,$7E,$4C,$7D,$1F,$84,$FA
	.dc.b	$5F,$0D,$27,$E6,$C7,$37,$E6,$61,$71,$F8,$58,$F5,$9F,$A7,$8F,$57
	.dc.b	$E9,$63,$D5,$FA,$F8,$F5,$7E,$B6,$3F,$81,$F8,$4F,$C1,$7C,$2C,$9F
	.dc.b	$BB,$1F,$85,$F8,$5F,$C3,$7C,$2C,$9F,$C0,$C7,$E2,$7E,$17,$F1,$5F
	.dc.b	$0B,$27,$F0,$B1,$D5,$FC,$3E,$1D,$5F,$C3,$61,$D5,$FC,$5E,$1D,$5F
	.dc.b	$C5,$61,$D5,$FC,$7E,$1D,$5F,$C7,$61,$F9,$9F,$85,$FC,$D7,$C2,$C9
	.dc.b	$FC,$AC,$7E,$77,$E1,$7F,$3D,$F0,$B2,$7F,$33,$1F,$E9,$5F,$D6,$7F
	.dc.b	$D4,$31,$FE,$08,$6F,$1F,$F9,$89,$BD,$3F,$D0,$57,$C2,$78,$6B,$F4
	.dc.b	$79,$78,$5F,$D1,$65,$E1,$7F,$4F,$97,$85,$FD,$36,$5E,$1C,$64,$7F
	.dc.b	$8B,$B1,$EE,$FF,$30,$7B,$BE,$C3,$B9,$0E,$E7,$B0,$EE,$43,$B9,$D6
	.dc.b	$3B,$BE,$F7,$E1,$63,$F0,$FE,$12,$4F,$8D,$F8,$DF,$0D,$F1,$FE,$17
	.dc.b	$E4,$7C,$34,$AF,$44,$7C,$BF,$83,$91,$EB,$7E,$77,$C2,$74,$8F,$9D
	.dc.b	$85,$E0,$F9,$D8,$F5,$9F,$85,$8F,$57,$EA,$63,$D1,$FA,$D8,$DB,$FA
	.dc.b	$D8,$FD,$BF,$84,$FB,$9F,$09,$2F,$E6,$47,$DF,$F8,$69,$1F,$3D,$FC
	.dc.b	$17,$C3,$7E,$0F,$E1,$BF,$09,$F0,$D2,$BE,$A4,$75,$7E,$B6,$1D,$5F
	.dc.b	$C4,$61,$D1,$FC,$56,$17,$2F,$C5,$63,$D6,$7E,$F6,$3D,$5F,$C8,$63
	.dc.b	$F9,$3F,$86,$FC,$A7,$C3,$49,$FC,$24,$7E,$5F,$E1,$A4,$7E,$1D,$FC
	.dc.b	$D7,$C3,$7E,$6F,$E1,$24,$FE,$2D,$FC,$F7,$C2,$F5,$8F,$C7,$61,$D5
	.dc.b	$FD,$06,$1D,$1F,$D1,$61,$74,$FD,$16,$3D,$27,$F4,$D8,$D9,$FD,$36
	.dc.b	$3D,$67,$F5,$58,$F5,$9F,$D5,$63,$CF,$9F,$F5,$59,$5B,$FC,$F6,$5C
	.dc.b	$E7,$F5,$D9,$59,$FD,$16,$5F,$B7,$F8,$54,$7F,$A5,$3B,$BF,$C8,$F7
	.dc.b	$FE,$35,$6C,$7E,$A6,$76,$FF,$A1,$23,$FC,$76,$FF,$87,$1B,$E8,$1A
	.dc.b	$7F,$8F,$6B,$A7,$C1,$F2,$D7,$FD,$CB,$2F,$83,$C5,$78,$5B,$87,$5B
	.dc.b	$F4,$3E,$7F,$85,$CE,$3D,$DC,$C7,$97,$E1,$73,$8F,$89,$BF,$0D,$8C
	.dc.b	$33,$8C,$BF,$85,$1C,$3E,$16,$47,$C5,$79,$FC,$37,$4F,$86,$93,$F2
	.dc.b	$5E,$FF,$0B,$E0,$F8,$5F,$0F,$C2,$CA,$F9,$D1,$D1,$F9,$58,$5E,$1F
	.dc.b	$87,$8F,$49,$F9,$D8,$D9,$F8,$F8,$F5,$9F,$B1,$8F,$57,$EA,$63,$F0
	.dc.b	$3E,$13,$E0,$FC,$2C,$9F,$BD,$1F,$13,$E1,$7E,$2F,$C2,$C9,$FC,$14
	.dc.b	$7C,$8F,$84,$F9,$3F,$0B,$27,$F0,$D1,$D5,$FC,$46,$1D,$5F,$C2,$61
	.dc.b	$D5,$FC,$66,$1D,$5F,$C4,$61,$D5,$FC,$86,$1D,$5F,$C6,$61,$D5,$FC
	.dc.b	$76,$1D,$74,$FC,$3E,$3D,$67,$F2,$58,$F5,$9F,$C5,$E3,$D6,$7F,$2D
	.dc.b	$8F,$59,$FC,$7E,$3F,$84,$F8,$59,$5F,$A1,$7F,$0F,$F0,$BF,$88,$F8
	.dc.b	$59,$3F,$A5,$7F,$17,$F0,$BF,$90,$FD,$6B,$FC,$7E,$3E,$17,$06,$6B
	.dc.b	$A9,$FA,$7C,$2E,$5F,$A0,$C7,$AC,$FE,$AF,$1B,$3F,$A4,$C7,$94,$FE
	.dc.b	$DA,$1D,$BF,$EA,$9C,$BC,$13,$98,$C6,$5C,$23,$FD,$F3,$2F,$84,$F4
	.dc.b	$12,$38,$1F,$E9,$48,$FF,$21,$D7,$5F,$86,$F7,$12,$77,$DE,$A7,$C2
	.dc.b	$7C,$3F,$84,$F8,$9F,$09,$2F,$74,$9F,$04,$79,$32,$BA,$7B,$23,$C5
	.dc.b	$F0,$BD,$DF,$4E,$76,$75,$9F,$87,$3F,$05,$76,$CF,$C0,$FC,$0C,$BE
	.dc.b	$B7,$C2,$7D,$7F,$85,$93,$E4,$8F,$B5,$F0,$9F,$6F,$E1,$64,$FA,$23
	.dc.b	$EF,$7C,$27,$DF,$F8,$59,$3E,$C8,$EA,$FB,$B0,$EA,$FD,$0C,$3A,$BF
	.dc.b	$0B,$0E,$AF,$D4,$C3,$AB,$F1,$B0,$EA,$FD,$8C,$3A,$BF,$6B,$0E,$BA
	.dc.b	$FC,$2E,$16,$5D,$AB,$EF,$E5,$DA,$B5,$97,$69,$FC,$1E,$5D,$A7,$96
	.dc.b	$5F,$9B,$F8,$49,$5F,$61,$FC,$F7,$C2,$FE,$7F,$CA,$93,$F7,$1F,$D1
	.dc.b	$7C,$2F,$E8,$FE,$12,$4F,$E0,$1F,$D3,$7C,$2F,$48,$FC,$9E,$17,$2F
	.dc.b	$56,$3D,$27,$F2,$F8,$D9,$F7,$63,$D2,$7F,$37,$8D,$9F,$85,$8F,$EC
	.dc.b	$FE,$13,$F6,$9F,$0B,$2F,$F2,$31,$FD,$2B,$F4,$17,$EF,$EF,$6B,$0D
	.dc.b	$AF,$D1,$CE,$A7,$3E,$E7,$F4,$EF,$D7,$3F,$C9,$7D,$67,$FC,$97,$6D
	.dc.b	$3F,$2B,$18,$6F,$7F,$96,$7F,$47,$A7,$95,$FD,$AE,$9E,$2C,$A4,$78
	.dc.b	$B4,$F2,$3D,$6F,$FB,$BE,$9E,$AD,$3C,$91,$E5,$F8,$79,$5F,$B6,$D3
	.dc.b	$CE,$3E,$1F,$D4,$48,$D7,$C3,$FB,$7E,$1A,$47,$17,$E0,$7C,$3F,$C1
	.dc.b	$F8,$7F,$85,$F0,$F2,$BB,$47,$37,$CC,$38,$FE,$36,$39,$E3,$F8,$B9
	.dc.b	$76,$AD,$D7,$CB,$CB,$15,$F2,$B2,$F0,$57,$AF,$2F,$03,$C3,$36,$7D
	.dc.b	$75,$FF,$0D,$E5,$E0,$AF,$85,$A7,$D9,$D3,$AD,$6B,$06,$98,$BF,$44
	.dc.b	$33,$F0,$C7,$7C,$FC,$3A,$7D,$ED,$3F,$09,$F0,$DF,$85,$F8,$79,$7F
	.dc.b	$46,$3F,$C3,$3F,$68,$FD,$DF,$F9,$BC,$FA,$7E,$27,$F1,$86,$A3,$38
	.dc.b	$FE,$2B,$8F,$F2,$04,$AF,$B9,$CD,$E0,$7E,$F6,$5E,$07,$E2,$65,$CF
	.dc.b	$CC,$EB,$1F,$83,$CB,$0F,$E1,$70,$E8,$FC,$B7,$F3,$D8,$7E,$7F,$E2
	.dc.b	$7F,$41,$F1,$12,$BF,$19,$1F,$A3,$F8,$9F,$D2,$7C,$4C,$9F,$C8,$47
	.dc.b	$E9,$FE,$27,$F5,$1F,$13,$27,$F2,$91,$D5,$FC,$B6,$1D,$5F,$B9,$87
	.dc.b	$57,$F3,$58,$75,$7F,$01,$87,$57,$F3,$D8,$75,$7F,$09,$87,$ED,$FE
	.dc.b	$26,$47,$E8,$DF,$DD,$7C,$4F,$F9,$23,$F5,$EF,$F2,$EF,$69,$FF,$2C
	.dc.b	$93,$E2,$B5,$F1,$33,$0F,$EC,$33,$FE,$5B,$7F,$3C,$BE,$23,$BC,$7E
	.dc.b	$57,$1B,$CD,$FC,$13,$2F,$06,$3E,$1F,$8A,$97,$D2,$3B,$3F,$A4,$C7
	.dc.b	$B6,$7E,$5C,$FD,$19,$F4,$EF,$FC,$D6,$97,$0F,$CC,$E7,$D9,$F3,$E7
	.dc.b	$D9,$E9,$9F,$6D,$3F,$B5,$E9,$D8,$99,$7C,$5F,$8A,$F8,$DF,$15,$30
	.dc.b	$F8,$31,$F2,$7E,$2A,$47,$C4,$7E,$67,$C5,$7C,$DF,$8A,$F9,$DF,$13
	.dc.b	$2B,$E4,$C7,$57,$E5,$E1,$D5,$F8,$18,$74,$7E,$16,$17,$3F,$3E,$3D
	.dc.b	$67,$E9,$E3,$D5,$F9,$18,$FD,$DF,$8A,$91,$F6,$1F,$C0,$7C,$4F,$E0
	.dc.b	$7E,$2A,$4F,$DC,$7F,$D1,$F8,$BF,$5A,$67,$F8,$7F,$8A,$FC,$47,$C5
	.dc.b	$CC,$3F,$07,$1D,$9F,$B1,$95,$C3,$E6,$E5,$DA,$7E,$E6,$56,$7F,$C4
	.dc.b	$07,$7C,$7B,$4F,$E3,$B1,$EC,$FE,$0F,$1F,$CC,$7C,$57,$E6,$7E,$2E
	.dc.b	$4F,$E5,$23,$F3,$9F,$17,$23,$F2,$EF,$E7,$FE,$2F,$F4,$1F,$17,$FA
	.dc.b	$1F,$8B,$95,$F9,$C8,$EA,$FE,$7B,$0E,$AF,$E4,$F0,$E8,$FE,$57,$0B
	.dc.b	$A7,$E2,$B1,$EB,$3F,$A6,$C7,$AB,$F9,$BC,$7A,$BF,$9D,$C7,$AD,$FE
	.dc.b	$4B,$1E,$7D,$9F,$CE,$BA,$19,$E3,$0F,$D7,$E7,$D9,$FD,$36,$7D,$B1
	.dc.b	$FC,$DE,$7F,$EC,$4C,$FF,$DA,$F5,$FD,$39,$B3,$FE,$4A,$6E,$5F,$D2
	.dc.b	$1F,$E6,$2D,$F8,$BE,$44,$AF,$F3,$0C,$FF,$99,$9F,$EB,$CB,$E2,$BC
	.dc.b	$1F,$15,$27,$71,$E3,$61,$C8,$F3,$79,$7D,$5F,$35,$DC,$7C,$DF,$17
	.dc.b	$EB,$BE,$95,$EB,$C3,$0F,$93,$E2,$BE,$01,$1F,$3F,$C5,$FC,$3F,$89
	.dc.b	$91,$EA,$7E,$37,$C5,$FC,$7F,$89,$F9,$1F,$15,$2B,$DF,$1C,$DF,$95
	.dc.b	$85,$D3,$E5,$63,$D2,$7E,$76,$36,$7E,$76,$3D,$67,$E4,$E3,$D5,$FA
	.dc.b	$98,$FD,$7F,$89,$91,$F3,$9F,$B5,$F1,$5F,$6F,$E2,$A4,$FD,$27,$FB
	.dc.b	$47,$C4,$7D,$FF,$8A,$93,$F5,$9F,$C1,$7C,$57,$48,$FC,$16,$17,$4F
	.dc.b	$C1,$63,$D2,$7F,$0D,$8D,$9F,$F4,$A7,$B9,$FD,$E7,$2E,$B5,$F8,$BC
	.dc.b	$AD,$7E,$2F,$2E,$9C,$FF,$8B,$CE,$DF,$E1,$B3,$E9,$3F,$8F,$CE,$CB
	.dc.b	$3E,$CF,$A4,$87,$73,$F2,$38,$7E,$77,$3F,$CF,$7C,$54,$9F,$CA,$C7
	.dc.b	$E8,$7E,$2B,$FC,$19,$32,$7F,$33,$1F,$A5,$F8,$9F,$D3,$7C,$54,$9F
	.dc.b	$CE,$C7,$57,$F3,$F5,$D5,$FD,$55,$F5,$7F,$D9,$38,$FF,$87,$67,$B4
	.dc.b	$7E,$BF,$0E,$CF,$EA,$30,$EC,$FE,$CF,$03,$FE,$CC,$FD,$7F,$FC,$DA
	.dc.b	$4D,$33,$1F,$E4,$68,$FF,$36,$47,$F8,$F9,$FF,$2E,$47,$1F,$87,$DF
	.dc.b	$07,$FB,$1A,$51,$8A,$E1,$F1,$16,$FF,$CA,$D5,$FE,$42,$7C,$5F,$0D
	.dc.b	$9F,$61,$3F,$E4,$39,$F3,$7C,$46,$63,$CD,$1E,$9C,$31,$F1,$39,$C7
	.dc.b	$59,$4A,$F7,$47,$1C,$39,$65,$27,$A9,$F0,$FE,$1F,$B1,$1F,$8D,$F1
	.dc.b	$1F,$D5,$BF,$1F,$F2,$47,$C5,$F8,$7F,$21,$9B,$F3,$7C,$44,$9F,$9C
	.dc.b	$FA,$7E,$23,$A4,$7C,$DC,$6F,$0F,$C9,$CB,$94,$FD,$5C,$7A,$BF,$62
	.dc.b	$7F,$AC,$E5,$E2,$7E,$86,$3D,$E7,$EC,$3D,$DF,$61,$0F,$01,$EC,$3A
	.dc.b	$6F,$D6,$BA,$3F,$7E,$FC,$33,$FE,$8E,$B5,$6A,$3E,$DD,$F8,$67,$C7
	.dc.b	$1F,$8D,$CB,$EC,$7C,$47,$D9,$F8,$89,$3F,$91,$8F,$99,$B4,$8F,$19
	.dc.b	$FE,$BD,$C7,$F0,$CF,$87,$D7,$7A,$BF,$83,$F8,$79,$3F,$9B,$7F,$0D
	.dc.b	$F1,$1E,$08,$FC,$F6,$5E,$07,$F3,$99,$77,$7D,$98,$F6,$D3,$E6,$E7
	.dc.b	$FE,$14,$CF,$BD,$75,$21,$FA,$5C,$BB,$BF,$AB,$F8,$7F,$CB,$7C,$44
	.dc.b	$8C,$1E,$CC,$3F,$C3,$DE,$BF,$5D,$A4,$E7,$14,$FE,$7F,$E1,$E4,$FE
	.dc.b	$DD,$FD,$17,$C4,$7E,$8F,$2E,$D3,$FB,$7C,$B0,$F3,$BE,$AF,$C9,$C7
	.dc.b	$D9,$1D,$23,$F7,$D8,$5C,$BF,$75,$8F,$EB,$FE,$1E,$57,$F0,$77,$F6
	.dc.b	$5F,$0F,$FB,$7F,$D8,$1F,$0F,$D2,$BC,$07,$DA,$F0,$13,$FE,$80,$1F
	.dc.b	$11,$C4,$93,$B8,$FF,$38,$F8,$2C,$7C,$33,$FC,$4D,$F0,$BB,$C3,$C0
	.dc.b	$FE,$17,$5F,$24,$74,$8F,$F6,$FD,$F8,$1E,$D8,$7A,$7E,$27,$AE,$BF
	.dc.b	$C6,$71,$BD,$9F,$C7,$F5,$E7,$87,$B0,$EC,$78,$4E,$64,$3B,$E1,$D5
	.dc.b	$F6,$7C,$4F,$C7,$F8,$89,$18,$3E,$07,$C4,$F6,$7C,$FA,$FC,$D8,$F9
	.dc.b	$DF,$09,$F3,$FE,$26,$57,$C7,$8F,$A5,$87,$48,$F7,$E1,$87,$A9,$E7
	.dc.b	$3A,$98,$3E,$2C,$F4,$7E,$3D,$78,$1F,$A9,$F1,$1F,$7C,$8F,$D7,$F8
	.dc.b	$9F,$F7,$7F,$EF,$1F,$D1,$23,$E7,$E5,$E9,$7F,$AB,$C6,$78,$1B,$FF
	.dc.b	$52,$E5,$FC,$2F,$C5,$7E,$2F,$E2,$A6,$1F,$83,$8E,$EF,$DA,$CE,$E1
	.dc.b	$CA,$17,$F8,$EC,$FC,$7E,$57,$86,$7C,$67,$89,$FC,$C6,$1D,$A3,$F0
	.dc.b	$B8,$E1,$FC,$97,$5F,$E7,$48,$FE,$57,$A3,$AB,$CC,$83,$2F,$F0,$CE
	.dc.b	$BF,$D8,$A3,$F3,$9F,$15,$FA,$72,$3F,$9F,$F8,$AE,$AF,$E5,$B0,$B7
	.dc.b	$ED,$D7,$F5,$38,$F8,$2B,$B6,$FF,$DE,$BD,$36,$7F,$36,$30,$D9,$FF
	.dc.b	$0F,$3E,$37,$F4,$79,$F8,$9A,$78,$CF,$82,$43,$C8,$7C,$13,$DF,$8F
	.dc.b	$BB,$6F,$23,$F0,$3C,$EE,$8F,$A7,$B1,$78,$C7,$F0,$10,$DB,$F8,$57
	.dc.b	$C2,$FF,$0B,$F8,$99,$9F,$F0,$38,$FC,$6E,$5D,$B0,$F6,$E7,$8A,$FE
	.dc.b	$11,$F1,$3F,$C4,$FE,$22,$46,$0E,$47,$F0,$DC,$7A,$BF,$C1,$31,$E4
	.dc.b	$FC,$C3,$E3,$DF,$77,$E4,$E3,$FC,$7F,$0E,$B1,$FC,$33,$0C,$3F,$C6
	.dc.b	$7E,$27,$F9,$21,$1F,$E4,$5F,$B1,$7F,$A8,$1E,$67,$E9,$2F,$C3,$BF
	.dc.b	$FB,$53,$C8,$63,$FE,$23,$BF,$F5,$2D,$FE,$79,$7C,$4E,$4D,$7C,$57
	.dc.b	$79,$FE,$3F,$95,$C7,$F1,$57,$E1,$E4,$FC,$3E,$7E,$0E,$0F,$C3,$BF
	.dc.b	$D7,$F3,$E7,$8F,$A3,$3E,$CF,$AB,$3F,$27,$C5,$79,$7E,$2A,$4F,$BA
	.dc.b	$39,$3E,$D2,$1F,$DE,$6F,$D9,$F1,$52,$3E,$2B,$EF,$F8,$AE,$B1,$F2
	.dc.b	$2F,$AB,$F5,$7A,$9B,$7F,$B0,$F7,$70,$FC,$8A,$EF,$1F,$93,$7F,$B6
	.dc.b	$65,$DD,$E5,$4C,$B1,$5F,$1B,$2F,$0D,$6B,$2F,$0C,$FC,$AC,$FD,$B3
	.dc.b	$F8,$7E,$0F,$96,$F8,$38,$3F,$C6,$DA,$78,$E3,$BE,$5D,$A7,$F9,$06
	.dc.b	$7E,$19,$FC,$2E,$7F,$7F,$E2,$BF,$01,$F1,$52,$7F,$13,$1E,$0D,$3F
	.dc.b	$0D,$B5,$C7,$FC,$EF,$F5,$AF,$33,$F1,$1F,$17,$F8,$9F,$8A,$99,$7E
	.dc.b	$4A,$3B,$3F,$94,$CF,$B3,$F1,$33,$E8,$FF,$1B,$DF,$D5,$A7,$82,$3F
	.dc.b	$C6,$A3,$2F,$03,$D4,$FC,$FE,$3E,$07,$E7,$E3,$F9,$BF,$8B,$91,$FA
	.dc.b	$47,$F3,$DF,$15,$E0,$C3,$ED,$ED,$E2,$AE,$DA,$FE,$8F,$E2,$BF,$49
	.dc.b	$F1,$52,$7F,$5B,$1D,$1F,$BB,$8D,$C3,$F4,$38,$F5,$D7,$D7,$C1,$F5
	.dc.b	$B3,$FF,$4A,$61,$D8,$E6,$7E,$E3,$2E,$CF,$E2,$32,$FD,$A7,$C5,$FE
	.dc.b	$D7,$E2,$E4,$FE,$FA,$39,$F0,$7C,$48,$FD,$01,$F9,$23,$E4,$B9,$D3
	.dc.b	$F6,$BF,$17,$FB,$F2,$4F,$F0,$8F,$8B,$EE,$FF,$0A,$C3,$BB,$F9,$AC
	.dc.b	$3B,$F0,$7B,$74,$B3,$FC,$43,$0E,$6F,$E8,$1F,$E1,$98,$74,$8F,$D1
	.dc.b	$E1,$D3,$1F,$E0,$F8,$F2,$9F,$49,$23,$FB,$83,$F1,$CE,$87,$72,$1F
	.dc.b	$AE,$AF,$E3,$5F,$17,$23,$F9,$1B,$FC,$77,$E2,$BB,$57,$E3,$67,$B6
	.dc.b	$FF,$9C,$D3,$F9,$17,$C5,$FF,$B1,$7D,$DF,$0D,$FF,$65,$6D,$7E,$B2
	.dc.b	$7B,$5F,$EF,$B1,$B7,$FC,$8B,$1E,$F7,$F8,$1A,$FA,$87,$F4,$13,$D9
	.dc.b	$3F,$CA,$2F,$C1,$32,$7F,$9D,$C7,$8F,$E2,$E4,$7F,$97,$F1,$FF,$73
	.dc.b	$BF,$E1,$47,$FB,$93,$D2,$BD,$8F,$E0,$B2,$FC,$8F,$E7,$EF,$87,$F7
	.dc.b	$78,$B2,$BA,$3F,$07,$E0,$FB,$C7,$F2,$3C,$7B,$F8,$86,$9E,$0E,$5F
	.dc.b	$C7,$67,$E0,$AF,$74,$61,$0C,$BC,$4F,$80,$87,$7C,$7C,$4F,$BF,$E3
	.dc.b	$3E,$A7,$C6,$48,$C1,$ED,$21,$E0,$D7,$DF,$AF,$DB,$F8,$CF,$B9,$F1
	.dc.b	$92,$BE,$54,$7D,$FC,$3A,$47,$BF,$0C,$3D,$38,$BF,$27,$A7,$AE,$3E
	.dc.b	$71,$EC,$D3,$D6,$FC,$7C,$3B,$3F,$27,$1F,$13,$E8,$3B,$1F,$6F,$0E
	.dc.b	$EF,$A4,$F0,$6B,$F2,$F1,$FD,$1E,$1D,$9F,$CA,$61,$D9,$FC,$B6,$1D
	.dc.b	$AF,$EC,$7C,$67,$E6,$C8,$FE,$27,$4F,$CA,$BF,$53,$3F,$61,$ED,$C7
	.dc.b	$B6,$5F,$E6,$9D,$DF,$15,$FA,$52,$5F,$E4,$FE,$2F,$AB,$E0,$3F,$2D
	.dc.b	$97,$EA,$F2,$EF,$1F,$20,$98,$F6,$7F,$35,$7D,$1F,$C8,$3F,$B3,$BF
	.dc.b	$DA,$7C,$57,$ED,$7E,$32,$57,$E8,$E3,$AE,$9E,$AE,$2F,$C0,$CF,$E9
	.dc.b	$FE,$2F,$F7,$9A,$C8,$C1,$C4,$FD,$66,$1D,$1F,$D0,$61,$D2,$FD,$39
	.dc.b	$F9,$F9,$7F,$45,$8F,$3B,$FD,$36,$3F,$C2,$63,$F8,$57,$C5,$FF,$0B
	.dc.b	$F8,$B9,$5F,$B9,$8E,$8F,$DE,$3F,$15,$BF,$F1,$1F,$8A,$FE,$25,$F1
	.dc.b	$52,$7F,$80,$47,$47,$F8,$15,$F4,$7F,$6D,$7D,$1F,$C5,$1F,$C3,$F3
	.dc.b	$F4,$65,$FE,$DF,$8F,$E1,$58,$75,$7F,$7D,$87,$57,$F8,$06,$1D,$79
	.dc.b	$7F,$91,$63,$D5,$FD,$1E,$BF,$3A,$B8,$9D,$08,$60,$FF,$16,$7E,$C7
	.dc.b	$FE,$FA,$7F,$DF,$0D,$73,$F1,$5F,$C1,$9D,$FC,$5F,$5B,$EF,$1F,$1C
	.dc.b	$FE,$15,$B6,$E3,$FD,$35,$7F,$C3,$F8,$0D,$27,$C5,$EF,$A3,$FD,$89
	.dc.b	$1E,$1A,$E3,$96,$71,$FF,$5F,$CE,$6B,$79,$78,$DF,$DA,$12,$7E,$C9
	.dc.b	$EC,$CF,$96,$5E,$26,$DF,$8A,$7E,$0F,$C5,$E0,$8F,$C4,$F8,$9E,$CF
	.dc.b	$ED,$6B,$F2,$3A,$F2,$F8,$B9,$3F,$29,$EB,$F1,$7C,$A3,$E7,$E1,$76
	.dc.b	$F8,$B8,$F4,$9F,$D9,$74,$FE,$FF,$1E,$71,$F5,$B1,$B7,$F3,$31,$F4
	.dc.b	$FC,$5F,$AB,$E2,$E5,$FD,$A8,$E9,$8F,$EC,$63,$FC,$73,$D1,$FA,$7D
	.dc.b	$9F,$17,$F0,$89,$3F,$82,$F8,$BE,$AF,$E1,$31,$EA,$FE,$23,$1E,$BC
	.dc.b	$1F,$0A,$3F,$13,$7C,$9F,$C6,$BF,$3A,$F9,$C7,$E4,$2F,$9E,$9F,$86
	.dc.b	$C3,$9C,$FE,$2F,$2F,$36,$1C,$EB,$F1,$6F,$39,$FC,$BF,$C5,$FD,$AF
	.dc.b	$8B,$91,$83,$DF,$C1,$FA,$5E,$4F,$CD,$63,$F0,$7E,$2B,$F0,$3F,$13
	.dc.b	$2F,$F4,$31,$F8,$4C,$79,$C7,$E9,$B1,$C3,$D0,$FD,$07,$0F,$D7,$78
	.dc.b	$9F,$AA,$BE,$8F,$EB,$AF,$F1,$FF,$11,$23,$F5,$EF,$E4,$BE,$27,$E0
	.dc.b	$C7,$F0,$6D,$BF,$A6,$EB,$F9,$6F,$57,$F2,$FF,$11,$2B,$F6,$F1,$CD
	.dc.b	$FD,$DE,$16,$FF,$63,$87,$47,$F6,$F3,$FC,$47,$0E,$8F,$F0,$0C,$3A
	.dc.b	$3F,$C1,$30,$FD,$2F,$C4,$48,$FE,$0A,$FE,$AB,$F6,$4F,$F6,$E4,$7F
	.dc.b	$00,$3E,$1F,$37,$E9,$BB,$13,$FE,$DA,$6F,$C4,$F1,$39,$7C,$4F,$5B
	.dc.b	$FE,$27,$85,$D1,$CB,$FE,$16,$C3,$9D,$7A,$E3,$F7,$CF,$84,$FF,$63
	.dc.b	$63,$DC,$FF,$73,$BF,$E1,$EA,$E6,$AB,$C6,$E6,$BB,$57,$9F,$A9,$87
	.dc.b	$8E,$7F,$8A,$BF,$06,$BC,$47,$F0,$33,$F0,$87,$8B,$1F,$73,$D4,$F0
	.dc.b	$5F,$91,$D1,$9C,$3E,$3F,$C4,$7C,$DF,$89,$93,$9A,$4F,$82,$3F,$BB
	.dc.b	$F0,$7D,$3F,$88,$FA,$9F,$0F,$27,$84,$7D,$7A,$E7,$1F,$1B,$0C,$3E
	.dc.b	$8D,$BF,$0A,$4D,$7F,$4F,$8F,$59,$F1,$63,$D5,$F9,$F8,$FE,$0F,$E1
	.dc.b	$A4,$79,$DF,$C3,$7C,$3F,$9F,$A7,$F4,$D3,$F5,$BE,$1B,$F1,$5F,$0F
	.dc.b	$23,$DD,$1C,$9F,$B7,$77,$4F,$B3,$C1,$F8,$CC,$7A,$72,$FE,$3F,$85
	.dc.b	$8F,$3C,$7F,$07,$97,$E5,$E3,$F3,$1F,$0F,$F9,$9F,$87,$95,$F3,$63
	.dc.b	$9B,$F0,$F8,$7F,$07,$F0,$BF,$9F,$24,$7D,$3F,$86,$E6,$FD,$5B,$E6
	.dc.b	$FE,$4A,$F9,$BF,$AF,$3E,$D5,$72,$7F,$2E,$FE,$AE,$B9,$C7,$E6,$AB
	.dc.b	$9E,$DF,$03,$36,$1D,$23,$F8,$2D,$74,$21,$D0,$F3,$1F,$89,$F8,$4F
	.dc.b	$DC,$11,$FC,$67,$C2,$F3,$CF,$EF,$C7,$E3,$FE,$1B,$F7,$DF,$0D,$23
	.dc.b	$07,$32,$1F,$B0,$BF,$E0,$97,$CE,$3F,$B8,$9F,$B4,$AE,$4F,$ED,$AB
	.dc.b	$F8,$54,$7F,$0B,$F8,$69,$3F,$9F,$7F,$87,$7C,$37,$38,$FD,$F6,$5F
	.dc.b	$0F,$E1,$BF,$BD,$FE,$CB,$FE,$C2,$9F,$DE,$37,$9B,$F8,$8C,$F4,$9F
	.dc.b	$E0,$B8,$5B,$F9,$98,$75,$F2,$BF,$C9,$5C,$0C,$3A,$EF,$31,$FD,$94
	.dc.b	$78,$3E,$1E,$47,$EE,$74,$FF,$59,$9F,$EC,$0C,$3F,$CD,$51,$FD,$F3
	.dc.b	$9F,$F9,$24,$BE,$1B,$D2,$49,$E5,$F0,$FD,$16,$BF,$A0,$AE,$B8,$FC
	.dc.b	$0F,$8B,$F4,$78,$1C,$6F,$7A,$FF,$27,$BE,$D1,$E5,$CB,$E3,$E5,$D7
	.dc.b	$2E,$99,$5C,$38,$65,$D3,$0E,$04,$3A,$9F,$3B,$0E,$6F,$C1,$F8,$8F
	.dc.b	$A7,$F1,$12,$30,$7E,$AC,$FE,$4C,$7F,$5D,$BF,$C7,$BD,$38,$DA,$3E
	.dc.b	$23,$EE,$FC,$44,$AF,$9D,$1F,$E2,$7E,$17,$C4,$7F,$BA,$8C,$70,$FF
	.dc.b	$7F,$AF,$C8,$F0,$7E,$99,$FF,$70,$FB,$DF,$E4,$B5,$E2,$23,$8C,$7F
	.dc.b	$89,$AF,$A3,$E9,$3F,$23,$AF,$F5,$0A,$FE,$37,$7D,$4F,$F6,$9D,$F5
	.dc.b	$8E,$95,$C7,$1F,$C1,$9E,$87,$B1,$9C,$7D,$0F,$63,$D3,$87,$F9,$C6
	.dc.b	$B9,$C7,$A8,$E0,$72,$3F,$D1,$98,$FA,$9F,$9C,$7D,$0C,$7C,$D5,$F4
	.dc.b	$1F,$3C,$7F,$55,$3E,$1E,$3C,$63,$FD,$BB,$5F,$0A,$7F,$8D,$12,$3F
	.dc.b	$A3,$3D,$6B,$A1,$1F,$F6,$D5,$7C,$33,$A4,$7F,$8C,$AB,$A5,$79,$2B
	.dc.b	$DE,$73,$21,$F0,$C9,$7F,$0A,$3D,$31,$E3,$AF,$02,$C7,$E7,$3F,$10
	.dc.b	$F8,$E7,$C4,$3F,$B4,$9F,$76,$BA,$4F,$CF,$C3,$CF,$23,$A6,$1E,$89
	.dc.b	$EF,$7F,$BE,$BF,$F3,$75,$78,$CF,$59,$F3,$0F,$9C,$F6,$73,$1F,$84
	.dc.b	$AE,$91,$8C,$7F,$07,$3F,$88,$7E,$54,$72,$C3,$CA,$F9,$0F,$C1,$63
	.dc.b	$F2,$DF,$8F,$C9,$FE,$75,$7C,$87,$8C,$FC,$51,$8B,$F1,$3C,$B2,$F6
	.dc.b	$E1,$E2,$3F,$D1,$78,$78,$A7,$84,$7A,$1F,$11,$83,$FC,$86,$7D,$0C
	.dc.b	$3E,$B9,$2B,$AC,$77,$3E,$75,$FD,$89,$F5,$3F,$37,$0F,$54,$7B,$72
	.dc.b	$F9,$55,$EB,$8D,$E1,$E3,$BE,$58,$76,$AF,$C0,$5F,$EE,$23,$F0,$EF
	.dc.b	$23,$9C,$FC,$F9,$F2,$1E,$CB,$EA,$FB,$0F,$71,$F2,$9F,$63,$F6,$4E
	.dc.b	$31,$E2,$70,$78,$B0,$FB,$35,$E1,$C7,$ED,$D6,$EB,$C1,$87,$A7,$0F
	.dc.b	$D7,$C7,$E6,$0F,$B7,$7C,$5F,$9D,$3F,$AD,$7C,$38,$FF,$3B,$77,$87
	.dc.b	$D9,$BF,$5C,$7E,$94,$8E,$0E,$F7,$F1,$E3,$F3,$6F,$77,$E0,$C7,$F0
	.dc.b	$67,$B9,$FA,$BC,$7E,$BE,$1F,$3A,$FC,$0F,$AC,$F3,$4F,$81,$C3,$FC
	.dc.b	$02,$FC,$0F,$F0,$09,$E2,$FC,$AB,$F0,$3C,$4C,$57,$F3,$5B,$F9,$B5
	.dc.b	$C2,$7F,$85,$63,$F9,$E7,$81,$8C,$7F,$3C,$FF,$03,$9E,$13,$F8,$38
	.dc.b	$DB,$F5,$4C,$61,$F1,$AF,$BC,$FE,$D4,$FA,$67,$E7,$32,$FC,$A6,$1F
	.dc.b	$4E,$BF,$35,$87,$68,$FE,$B9,$3A,$AF,$94,$7F,$B8,$E3,$F6,$2F,$53
	.dc.b	$E8,$E3,$F2,$AF,$E1,$9F,$65,$F8,$51,$E2,$3F,$4C,$FF,$1D,$71,$8F
	.dc.b	$AE,$FF,$27,$8F,$D8,$9F,$BF,$8F,$95,$FE,$B5,$3E,$57,$F2,$33,$B7
	.dc.b	$EB,$1E,$5C,$3C,$4F,$C1,$3F,$0A,$70,$3D,$87,$CF,$38,$61,$F8,$27
	.dc.b	$F4,$27,$13,$DE,$7E,$16,$7E,$7B,$FB,$18,$E1,$87,$AE,$3E,$06,$5F
	.dc.b	$CD,$2F,$C7,$1F,$08,$FE,$AA,$7C,$2C,$3D,$91,$FC,$37,$0F,$65,$FF
	.dc.b	$3D,$C7,$E1,$5F,$F2,$7B,$DE,$3F,$C2,$AF,$79,$7E,$3F,$2F,$E1,$6F
	.dc.b	$DC,$BF,$83,$97,$57,$E0,$9F,$B4,$3E,$09,$F9,$A3,$E0,$3F,$72,$FE
	.dc.b	$25,$7A,$DB,$8F,$92,$7F,$93,$E3,$E4,$BF,$DB,$E3,$E4,$7A,$9F,$04
	.dc.b	$FD,$01,$FC,$5A,$3F,$02,$F9,$27,$F2,$AF,$0C,$BF,$3F,$87,$E3,$C9
	.dc.b	$1E,$73,$FC,$BC,$7D,$0C,$3C,$91,$FD,$9A,$FC,$93,$FD,$50,$ED,$1F
	.dc.b	$C1,$5E,$D9,$7C,$77,$F5,$EF,$E9,$E3,$51,$F7,$CF,$05,$FD,$F7,$F0
	.dc.b	$A7,$88,$B9,$7F,$17,$C7,$F0,$33,$B7,$F0,$71,$F2,$E7,$D4,$EE,$BF
	.dc.b	$17,$87,$A2,$FF,$A9,$57,$A0,$FD,$29,$E8,$3F,$4A,$7D,$E9,$FC,$23
	.dc.b	$B3,$F4,$A7,$A4,$FA,$27,$F0,$C9,$FA,$2F,$F8,$56,$7E,$7B,$F9,$63
	.dc.b	$83,$87,$F4,$A4,$71,$5E,$B8,$EB,$7F,$D5,$1F,$C0,$4F,$13,$E1,$5F
	.dc.b	$E3,$32,$FC,$34,$F9,$6F,$F3,$2F,$B0,$FA,$67,$F1,$43,$CD,$8F,$F1
	.dc.b	$47,$CA,$FD,$9C,$3F,$31,$1F,$32,$BA,$C7,$8E,$BA,$CF,$EA,$CF,$90
	.dc.b	$7E,$60,$F9,$18,$7E,$C5,$F9,$C6,$CF,$19,$FC,$2C,$F1,$9F,$C2,$CF
	.dc.b	$19,$E9,$9F,$1B,$F1,$30,$FE,$4A,$FE,$1B,$2F,$CB,$BF,$E3,$4C,$BF
	.dc.b	$01,$97,$F4,$48,$FD,$B1,$D0,$F1,$5F,$4C,$3E,$8E,$5E,$EC,$7F,$97
	.dc.b	$56,$EB,$F6,$AF,$B8,$FD,$11,$E8,$3A,$BF,$5E,$BE,$B6,$1E,$29,$FF
	.dc.b	$08,$4F,$E6,$31,$F3,$C7,$F5,$C9,$D3,$FE,$0D,$3F,$64,$7C,$89,$F0
	.dc.b	$BF,$21,$FE,$24,$F8,$5F,$DF,$BE,$7B,$FD,$C3,$E7,$BF,$D0,$BB,$8F
	.dc.b	$D9,$CF,$9F,$1F,$C0,$47,$F0,$49,$F9,$C7,$F3,$73,$C6,$75,$3E,$71
	.dc.b	$D6,$3E,$74,$F8,$8E,$91,$F6,$67,$A5,$7C,$E8,$FC,$2D,$7F,$2F,$C3
	.dc.b	$FC,$53,$87,$C2,$C3,$FB,$A5,$78,$1E,$6F,$D8,$C3,$9D,$FF,$0D,$8F
	.dc.b	$DD,$9E,$53,$E6,$1E,$53,$E8,$C7,$F3,$07,$E6,$3F,$19,$F9,$8F,$F0
	.dc.b	$07,$E3,$1D,$27,$C2,$7D,$23,$F1,$13,$E1,$8F,$E0,$B1,$A9,$E8,$EA
	.dc.b	$3F,$60,$FF,$90,$B0,$FD,$FE,$1F,$82,$7F,$C3,$91,$B9,$FD,$23,$BB
	.dc.b	$FD,$B3,$D6,$3E,$89,$E8,$3F,$60,$7A,$0F,$1D,$FE,$6A,$3F,$94,$E1
	.dc.b	$F3,$63,$F5,$31,$F9,$39,$F8,$67,$47,$C4,$7E,$38,$F1,$3F,$B5,$7C
	.dc.b	$4F,$CE,$3C,$47,$E9,$1D,$BF,$A4,$8F,$11,$FC,$D1,$FA,$87,$E7,$CF
	.dc.b	$DD,$1E,$23,$A4,$78,$A3,$EA,$3D,$0F,$C7,$3D,$23,$FA,$5C,$FB,$5F
	.dc.b	$CC,$1D,$0F,$C2,$5F,$CE,$9F,$C4,$C7,$D4,$7F,$DE,$D8,$FF,$85,$2F
	.dc.b	$DB,$5B,$AF,$EF,$75,$FC,$BD,$FC,$C9,$FD,$D3,$2F,$E4,$75,$F8,$59
	.dc.b	$EB,$3F,$69,$EA,$FE,$6D,$F4,$3F,$AE,$3F,$A1,$61,$FD,$2A,$7D,$D3
	.dc.b	$FB,$63,$77,$FC,$32,$BE,$4C,$FE,$99,$F7,$3E,$8C,$7F,$D5,$57,$CE
	.dc.b	$FC,$35,$A7,$9C,$FF,$00,$8F,$69,$FC,$FC,$F6,$9F,$D2,$F2,$FD,$AB
	.dc.b	$E0,$9F,$C4,$9F,$AB,$7E,$6C,$F8,$1F,$F1,$CE,$5F,$CA,$DF,$EC,$58
	.dc.b	$79,$5F,$F0,$C1,$A3,$BB,$FD,$36,$FB,$C7,$E3,$31,$FE,$33,$3F,$12
	.dc.b	$3F,$69,$87,$A9,$FE,$1C,$7A,$8F,$99,$97,$F6,$BA,$FD,$7C,$7E,$35
	.dc.b	$FB,$8F,$F7,$1C,$7F,$75,$1F,$D5,$E3,$D6,$79,$8F,$F0,$9E,$3E,$B7
	.dc.b	$FC,$65,$5F,$80,$BF,$DF,$9F,$80,$C3,$F5,$0F,$D0,$7F,$5D,$3D,$0F
	.dc.b	$F0,$41,$F0,$4F,$DC,$5F,$86,$3E,$34,$68,$F0,$9F,$1A,$7F,$13,$5E
	.dc.b	$18,$FD,$C4,$78,$4F,$D3,$9E,$13,$EB,$4F,$43,$FB,$06,$5F,$85,$91
	.dc.b	$FE,$D5,$8F,$D6,$BF,$62,$7F,$70,$FD,$3C,$7F,$9F,$C7,$E5,$63,$47
	.dc.b	$EC,$8F,$C3,$1F,$B2,$9F,$82,$78,$63,$FA,$1D,$78,$67,$C1,$1C,$DF
	.dc.b	$F0,$FD,$7E,$55,$FA,$C7,$ED,$CF,$3D,$FF,$B3,$B0,$F9,$15,$F8,$39
	.dc.b	$FE,$8C,$7E,$0E,$3F,$A8,$3F,$3A,$BF,$A8,$3F,$E8,$93,$CE,$7F,$94
	.dc.b	$70,$F3,$D7,$37,$F7,$91,$CE,$7F,$0E,$72,$7F,$CB,$35,$FE,$34,$7F
	.dc.b	$54,$77,$3F,$83,$65,$F9,$CA,$FE,$39,$1F,$CE,$1F,$CE,$3F,$CE,$0F
	.dc.b	$96,$7F,$05,$9E,$8F,$E7,$32,$FC,$96,$3F,$E5,$F8,$FA,$07,$D5,$3F
	.dc.b	$AF,$9F,$E6,$4B,$FE,$9D,$87,$ED,$63,$F1,$67,$94,$FD,$F1,$F0,$CE
	.dc.b	$47,$F3,$5C,$3E,$1B,$E4,$C3,$F8,$93,$A8,$FE,$24,$69,$FB,$27,$ED
	.dc.b	$0F,$B2,$7F,$23,$3E,$CE,$1F,$CB,$F1,$F9,$C7,$EA,$A3,$E7,$1F,$7C
	.dc.b	$FC,$29,$FC,$00,$F1,$5F,$E1,$6F,$FD,$F8,$FD,$48,$F9,$E7,$AE,$FE
	.dc.b	$7B,$FD,$56,$BF,$48,$FC,$53,$F8,$24,$FC,$59,$FF,$46,$DF,$CE,$C3
	.dc.b	$E4,$D7,$F0,$39,$CB,$FB,$33,$EB,$E3,$FB,$37,$FA,$CE,$1F,$B3,$7E
	.dc.b	$F3,$E0,$7F,$8A,$D7,$47,$F7,$D5,$D1,$FA,$8F,$F2,$37,$D5,$3C,$DF
	.dc.b	$51,$F2,$8F,$E9,$87,$C4,$3E,$86,$3F,$11,$F9,$57,$FA,$7C,$7F,$31
	.dc.b	$7F,$E0,$49,$FD,$0C,$FF,$27,$9F,$F0,$7D,$FD,$49,$FE,$1A,$7D,$43
	.dc.b	$ED,$1F,$50,$FC,$C1,$F8,$03,$F6,$27,$E0,$2B,$F9,$1B,$E7,$3F,$A9
	.dc.b	$1E,$73,$77,$E7,$7A,$DF,$9D,$FB,$67,$EB,$B2,$FE,$FC,$FE,$DE,$FF
	.dc.b	$8E,$9E,$77,$E8,$9E,$7B,$F0,$BE,$7B,$FE,$0E,$F8,$63,$F7,$AF,$9E
	.dc.b	$7F,$94,$BE,$78,$FC,$E5,$FD,$DA,$FE,$92,$79,$E3,$AC,$79,$E7,$F5
	.dc.b	$67,$9C,$F9,$B9,$7E,$2D,$FC,$79,$F8,$BB,$FF,$06,$3F,$80,$7F,$9E
	.dc.b	$E5,$FC,$1D,$FC,$06,$1F,$D8,$9F,$C5,$3F,$D7,$27,$FA,$66,$1F,$CF
	.dc.b	$1F,$E9,$8F,$F0,$C8,$F7,$C7,$D5,$8F,$7C,$7A,$67,$DF,$3F,$17,$2F
	.dc.b	$C2,$C7,$F4,$0A,$FD,$4D,$78,$6B,$E2,$4F,$D0,$7E,$53,$FB,$97,$E5
	.dc.b	$1F,$A0,$36,$7E,$BC,$F3,$BF,$D0,$9F,$3B,$EC,$AF,$C8,$C7,$43,$F2
	.dc.b	$27,$43,$F4,$27,$9C,$FC,$2D,$7E,$6D,$FE,$29,$3F,$E1,$67,$C1,$1F
	.dc.b	$D4,$5F,$E8,$99,$7F,$51,$CB,$EC,$4F,$EB,$E7,$EC,$1C,$8F,$EB,$55
	.dc.b	$EF,$7E,$53,$EF,$C3,$E5,$47,$BC,$F2,$1F,$82,$9F,$21,$F2,$B0,$FE
	.dc.b	$D3,$1F,$C6,$CE,$67,$BA,$79,$BE,$EB,$F2,$DF,$E1,$23,$F8,$19,$F9
	.dc.b	$D7,$EC,$C6,$A3,$F8,$9E,$3F,$89,$74,$79,$63,$4F,$F8,$6A,$34,$FC
	.dc.b	$DA,$FB,$CF,$AB,$0F,$88,$7A,$AF,$F2,$78,$7A,$A7,$F8,$0E,$1F,$A5
	.dc.b	$8F,$E5,$53,$F9,$CA,$FC,$94,$7E,$32,$BF,$AB,$C7,$C3,$7F,$AB,$9F
	.dc.b	$C9,$0F,$D2,$9F,$C9,$27,$91,$F8,$57,$FD,$24,$FE,$FE,$FF,$67,$8F
	.dc.b	$DB,$CB,$FA,$65,$7E,$24,$FE,$06,$7B,$70,$F4,$57,$E9,$31,$F4,$4F
	.dc.b	$EC,$CF,$EF,$38,$FA,$1F,$C6,$E3,$FA,$33,$E0,$BE,$8A,$FD,$94,$FE
	.dc.b	$36,$3F,$CF,$67,$E7,$0F,$D0,$9E,$D3,$F4,$27,$E4,$8D,$1F,$3E,$7E
	.dc.b	$F3,$F9,$28,$F8,$CF,$F7,$38,$FC,$95,$7A,$23,$E3,$1E,$88,$F8,$CF
	.dc.b	$FA,$C4,$F8,$55,$F1,$9E,$F3,$F8,$23,$FC,$14,$FF,$83,$F1,$FE,$5B
	.dc.b	$1F,$D0,$E3,$E7,$9C,$B0,$FB,$AF,$73,$DB,$5F,$0A,$FE,$A4,$EA,$7F
	.dc.b	$C3,$07,$F6,$73,$F2,$07,$E6,$8F,$A8,$7C,$B7,$F8,$0B,$FA,$43,$EC
	.dc.b	$CF,$E9,$0F,$CA,$D7,$A7,$2F,$ED,$58,$FF,$21,$AF,$F0,$EB,$F6,$5F
	.dc.b	$D9,$61,$F6,$5F,$D1,$9F,$AD,$CB,$F0,$EF,$7C,$3E,$33,$CA,$7E,$5D
	.dc.b	$7F,$99,$23,$58,$FB,$6B,$FC,$9B,$3D,$CF,$D3,$D7,$7A,$F6,$9F,$64
	.dc.b	$FE,$26,$FC,$67,$F2,$07,$C6,$3F,$A1,$1F,$BA,$3E,$34,$FC,$1A,$FC
	.dc.b	$C9,$FA,$B8,$FB,$07,$F4,$E3,$E7,$63,$F7,$27,$FC,$6F,$7F,$8A,$3F
	.dc.b	$C6,$F7,$F8,$AC,$7F,$80,$CF,$5B,$FA,$47,$E0,$0F,$79,$E6,$7D,$F3
	.dc.b	$F1,$0F,$F3,$C5,$FE,$4E,$FE,$91,$FE,$1B,$3E,$24,$7E,$71,$FE,$15
	.dc.b	$5F,$4A,$BE,$DE,$3F,$4A,$F9,$5F,$F4,$5A,$FA,$47,$E3,$B1,$FC,$2C
	.dc.b	$EA,$3F,$54,$68,$FD,$51,$F2,$4F,$D5,$4F,$EE,$1F,$A4,$7E,$CF,$0F
	.dc.b	$DC,$61,$FE,$6E,$3D,$B1,$DC,$F8,$55,$DD,$FD,$E9,$DD,$FC,$93,$FC
	.dc.b	$62,$7E,$0D,$7F,$18,$8E,$47,$EC,$CF,$96,$7E,$E0,$FC,$04,$7C,$67
	.dc.b	$D0,$FC,$67,$F1,$2F,$D8,$7F,$12,$7E,$CC,$FA,$97,$FE,$61,$7F,$3E
	.dc.b	$7F,$2E,$9F,$81,$3C,$63,$E0,$47,$E4,$5F,$15,$FC,$07,$E4,$9F,$B6
	.dc.b	$3E,$49,$F5,$EB,$F2,$2F,$F2,$43,$B1,$93,$F1,$27,$BE,$FE,$44,$FC
	.dc.b	$D9,$F3,$5F,$E9,$AB,$89,$EE,$BF,$EF,$0F,$98,$F7,$1E,$69,$E3,$1F
	.dc.b	$8B,$3F,$5D,$5F,$99,$9F,$BA,$FF,$13,$78,$D7,$D4,$7F,$8C,$1E,$EB
	.dc.b	$FD,$35,$FF,$A2,$F2,$FE,$95,$3F,$E6,$C3,$EB,$D7,$17,$EB,$E1,$FA
	.dc.b	$17,$FB,$D5,$FB,$AB,$F9,$4E,$3F,$E3,$F9,$FE,$7E,$7E,$22,$FF,$49
	.dc.b	$7C,$63,$FB,$4E,$5F,$B2,$BF,$0C,$7F,$99,$8F,$0C,$FE,$8E,$B8,$C7
	.dc.b	$E3,$1F,$E2,$98,$78,$63,$F8,$11,$FC,$28,$FB,$51,$F8,$BC,$BE,$33
	.dc.b	$F8,$37,$E3,$47,$63,$F5,$71,$D9,$F8,$D8,$7F,$2E,$BF,$11,$F9,$D8
	.dc.b	$C9,$FA,$17,$2F,$CE,$BF,$D2,$BF,$0A,$FF,$04,$FF,$A5,$31,$FE,$01
	.dc.b	$7F,$06,$3F,$AA,$C6,$1F,$E4,$38,$FC,$4C,$7F,$CE,$75,$B9,$FD,$95
	.dc.b	$6D,$FC,$B9,$F8,$B3,$FC,$72,$7E,$A0,$F7,$1F,$4F,$1F,$56,$3E,$6B
	.dc.b	$FE,$07,$87,$DA,$AF,$E5,$78,$7C,$17,$E2,$DF,$0C,$3C,$17,$F3,$6F
	.dc.b	$CD,$3F,$CB,$6B,$D2,$7B,$0F,$C6,$61,$F5,$DF,$F2,$71,$FE,$9B,$3C
	.dc.b	$CE,$6B,$F9,$C4,$FE,$93,$1F,$AF,$39,$8F,$DE,$99,$CB,$F9,$2E,$5F
	.dc.b	$16,$FE,$D5,$7B,$6B,$F8,$04,$7E,$6E,$FF,$80,$4E,$70,$FD,$3D,$7D
	.dc.b	$A7,$F3,$E7,$E3,$DF,$E7,$F7,$D5,$F8,$51,$EC,$8C,$3F,$81,$CB,$E2
	.dc.b	$CF,$ED,$B2,$F8,$B3,$EE,$AD,$BD,$8F,$C5,$9F,$87,$8F,$A4,$FE,$6E
	.dc.b	$FF,$6E,$FF,$B7,$A3,$9B,$FC,$1C,$FB,$B8,$7E,$39,$FF,$65,$E1,$E2
	.dc.b	$7F,$87,$C7,$E8,$B0,$F9,$11,$FC,$FB,$0F,$1C,$FC,$4A,$E1,$3F,$DB
	.dc.b	$8F,$AF,$87,$F8,$95,$FE,$3D,$5C,$32,$FE,$2B,$1F,$CE,$B0,$F8,$0F
	.dc.b	$E0,$4F,$DE,$57,$D9,$7F,$94,$1F,$C6,$27,$EC,$BF,$88,$3F,$3A,$7C
	.dc.b	$D3,$19,$7D,$CC,$3C,$B8,$FF,$7E,$C3,$F0,$75,$FA,$03,$E2,$61,$DD
	.dc.b	$FD,$21,$F8,$F3,$38,$FA,$5F,$DA,$1D,$CF,$D5,$D7,$C7,$7A,$3F,$92
	.dc.b	$8E,$11,$F3,$8E,$18,$7E,$8D,$F8,$0F,$EE,$1F,$4D,$FE,$21,$F8,$73
	.dc.b	$CA,$7F,$DD,$78,$7B,$31,$FE,$69,$5C,$AF,$F5,$8F,$93,$2F,$C0,$E3
	.dc.b	$E3,$8E,$A7,$E4,$4F,$1C,$FD,$27,$E5,$9E,$03,$67,$E4,$72,$FD,$13
	.dc.b	$F8,$AA,$FC,$8E,$3F,$DD,$63,$F9,$5D,$73,$8F,$A5,$1D,$8F,$C6,$CF
	.dc.b	$93,$0F,$B7,$1E,$73,$D4,$7E,$84,$F0,$E1,$F8,$F7,$91,$F8,$FA,$FE
	.dc.b	$02,$FD,$09,$E7,$1F,$36,$3D,$6F,$CD,$7B,$18,$BF,$8F,$1B,$3C,$F1
	.dc.b	$F0,$4D,$B8,$39,$1B,$8F,$AC,$E1,$F8,$06,$0E,$A6,$23,$E8,$1D,$63
	.dc.b	$93,$87,$93,$F1,$8F,$86,$7E,$48,$F8,$73,$FA,$A7,$E7,$9B,$3D,$87
	.dc.b	$C9,$3D,$33,$E3,$71,$3E,$07,$EE,$63,$F8,$B7,$F8,$41,$E0,$BF,$C4
	.dc.b	$3C,$8F,$35,$7B,$1F,$05,$7C,$58,$C4,$7D,$57,$F5,$EF,$EA,$E7,$9C
	.dc.b	$F8,$5F,$B9,$1F,$0E,$7E,$33,$E4,$3E,$FC,$7E,$E5,$FE,$07,$3F,$DD
	.dc.b	$6F,$EF,$D7,$0A,$E7,$3F,$2A,$BC,$47,$09,$FE,$36,$FF,$14,$3E,$59
	.dc.b	$FB,$C3,$C4,$7B,$1F,$90,$FD,$13,$E1,$1E,$53,$67,$63,$F8,$4E,$3F
	.dc.b	$C8,$67,$FC,$23,$97,$EE,$69,$52,$FF,$CF,$AF,$BB,$0F,$DD,$BF,$68
	.dc.b	$F0,$5F,$C7,$8F,$C0,$61,$FC,$FA,$52,$BE,$B1,$FE,$C4,$CB,$F5,$CF
	.dc.b	$F3,$93,$DE,$7F,$13,$3D,$38,$7C,$D7,$D3,$97,$F8,$D6,$FF,$BF,$C7
	.dc.b	$F8,$A1,$FD,$F6,$1E,$17,$C1,$1E,$BC,$7F,$D8,$07,$9B,$2F,$F0,$24
	.dc.b	$7F,$82,$CF,$C4,$9F,$92,$CB,$EC,$4E,$E3,$FC,$4C,$7F,$82,$8D,$9F
	.dc.b	$0A,$36,$FD,$7A,$FB,$D8,$FA,$4F,$E6,$51,$F8,$FA,$FC,$2B,$F6,$CE
	.dc.b	$71,$F9,$2A,$E7,$1F,$0E,$B9,$D7,$D2,$7F,$39,$1E,$73,$91,$F8,$17
	.dc.b	$6F,$EA,$0F,$B1,$5D,$5D,$BF,$94,$AF,$D2,$BE,$7C,$BE,$AC,$FE,$4C
	.dc.b	$FC,$6C,$6D,$F0,$1F,$8D,$7D,$B8,$FE,$36,$BF,$C1,$F3,$F7,$2F,$FA
	.dc.b	$2B,$F8,$B3,$C8,$FA,$9F,$94,$79,$30,$FA,$0F,$F8,$2C,$FB,$D3,$FE
	.dc.b	$71,$B4,$FF,$BA,$DF,$44,$FF,$05,$8F,$9A,$F9,$70,$F5,$DF,$E8,$23
	.dc.b	$C9,$7F,$94,$BF,$CF,$C7,$E4,$2F,$E7,$4F,$EB,$A7,$E4,$4F,$33,$B1
	.dc.b	$F9,$93,$E2,$CF,$A2,$BF,$2B,$1F,$E1,$0C,$BF,$31,$7F,$E3,$A3,$FC
	.dc.b	$6D,$7F,$2E,$7E,$09,$D4,$FA,$F7,$EA,$8F,$6C,$60,$F0,$5F,$EC,$5F
	.dc.b	$C8,$57,$DA,$AF,$4C,$7E,$31,$E4,$61,$FC,$34,$6D,$FA,$6F,$BE,$37
	.dc.b	$18,$39,$1E,$08,$FC,$AC,$FD,$C8,$F9,$06,$CE,$A7,$E1,$AB,$D3,$1F
	.dc.b	$7D,$F9,$66,$27,$C1,$3D,$4F,$C1,$1E,$C3,$95,$7A,$67,$E4,$1C,$9C
	.dc.b	$1F,$46,$7E,$74,$FE,$DC,$F1,$9F,$3A,$7C,$0F,$B0,$FC,$F5,$7C,$B7
	.dc.b	$F6,$13,$F1,$5F,$E3,$6F,$E1,$EB,$F8,$5C,$79,$4E,$8F,$89,$C5,$79
	.dc.b	$5F,$88,$FA,$23,$E6,$C7,$43,$F8,$B1,$FC,$CA,$BF,$CF,$D1,$EB,$8F
	.dc.b	$F3,$E9,$EB,$9E,$4F,$C7,$3F,$11,$3F,$36,$7A,$BF,$C8,$A3,$F9,$81
	.dc.b	$F9,$83,$DF,$5F,$68,$F6,$3E,$17,$FA,$9C,$7E,$9E,$3F,$3A,$7B,$27
	.dc.b	$AC,$F0,$8F,$F7,$F1,$93,$C9,$5F,$39,$FA,$2F,$C3,$7E,$01,$E4,$3F
	.dc.b	$38,$79,$0F,$C0,$9F,$48,$FC,$09,$F2,$A7,$E2,$3F,$E3,$F3,$F8,$29
	.dc.b	$FE,$9F,$3F,$DF,$B9,$7F,$5D,$7E,$69,$FE,$77,$3C,$B1,$F5,$DE,$C7
	.dc.b	$DA,$3B,$1E,$B3,$C8,$7C,$EA,$FC,$BB,$8A,$F7,$1F,$51,$C4,$F5,$BF
	.dc.b	$D3,$1F,$CD,$23,$E0,$CF,$F1,$A7,$E8,$1F,$76,$7D,$B9,$7F,$A0,$67
	.dc.b	$F9,$6E,$3F,$C2,$AB,$F0,$67,$F5,$93,$E6,$D7,$D4,$7D,$F7,$F8,$A7
	.dc.b	$D4,$FD,$B3,$F3,$C7,$93,$0F,$CF,$3F,$87,$BF,$24,$FD,$C3,$E2,$BF
	.dc.b	$66,$35,$5F,$A2,$9C,$BE,$17,$EE,$47,$E4,$DF,$9E,$7C,$C3,$66,$0F
	.dc.b	$87,$1C,$E3,$C9,$1F,$80,$32,$7C,$63,$E0,$4F,$CF,$8F,$65,$7E,$6E
	.dc.b	$7E,$0B,$F9,$A3,$FA,$9B,$F9,$A3,$C2,$7D,$63,$C9,$1E,$07,$F0,$47
	.dc.b	$80,$F8,$4E,$67,$E1,$4F,$E8,$63,$E7,$4F,$47,$C6,$FE,$69,$E1,$5F
	.dc.b	$16,$BE,$6C,$70,$33,$3F,$0D,$F9,$AF,$BE,$3F,$DC,$91,$A8,$F4,$1E
	.dc.b	$57,$95,$78,$CF,$81,$1F,$A2,$3F,$AF,$9F,$10,$FF,$4A,$1D,$DE,$0F
	.dc.b	$E4,$4F,$C3,$BF,$22,$3F,$0E,$EA,$3F,$C3,$47,$13,$A1,$F5,$2B,$F3
	.dc.b	$38,$FE,$C6,$17,$FD,$BA,$FF,$44,$7B,$A3,$E6,$DF,$5C,$7F,$C1,$51
	.dc.b	$E9,$3E,$C4,$FC,$13,$E6,$BF,$B3,$3F,$9D,$38,$AF,$D9,$E1,$E2,$9F
	.dc.b	$82,$78,$9F,$F2,$9C,$FF,$9D,$E3,$D4,$7E,$62,$FF,$7B,$1C,$E7,$F7
	.dc.b	$51,$E5,$C7,$F6,$55,$FC,$44,$FE,$FC,$7E,$57,$0F,$F6,$99,$FB,$9C
	.dc.b	$BF,$C2,$35,$F2,$CF,$CB,$4F,$8C,$E1,$5D,$2B,$84,$7F,$37,$C3,$C4
	.dc.b	$FE,$5A,$BC,$4F,$03,$FC,$58,$7E,$0D,$FF,$06,$E1,$F0,$63,$E4,$C7
	.dc.b	$07,$F9,$8D,$79,$B2,$FA,$D9,$7D,$5A,$FE,$06,$7D,$F7,$F2,$2F,$1C
	.dc.b	$3F,$CD,$F3,$FE,$6A,$9C,$DF,$E0,$AF,$EC,$BE,$D3,$BC,$EF,$2F,$F7
	.dc.b	$AE,$1F,$E3,$BC,$7F,$C7,$47,$CE,$AE,$59,$7C,$3B,$F2,$CF,$D8,$AF
	.dc.b	$CE,$D7,$5C,$BF,$AF,$63,$FE,$81,$9F,$46,$3D,$AB,$FA,$21,$B8,$FF
	.dc.b	$49,$CE,$E3,$F3,$39,$7A,$9D,$1F,$A6,$3F,$5F,$7F,$97,$3F,$5E,$F9
	.dc.b	$8F,$A4,$7F,$88,$8F,$A4,$7E,$AC,$FE,$09,$5F,$81,$7F,$82,$3F,$2C
	.dc.b	$F8,$4F,$E2,$71,$F7,$3F,$EB,$4B,$46,$23,$CA,$7C,$D3,$FD,$5A,$53
	.dc.b	$FB,$A1,$8B,$FE,$76,$DC,$76,$E8,$87,$F6,$63,$46,$8E,$86,$4D,$8F
	.dc.b	$8F,$7B,$0F,$A9,$E1,$18,$6C,$66,$7C,$33,$96,$C6,$5F,$5C,$65,$B1
	.dc.b	$97,$FD,$97,$19,$6C,$65,$FF,$6D,$C6,$5B,$19,$7D,$0F,$3F,$CB,$04
	.dc.b	$00

;--------------------------------------------------------------------------------
;	FTENTOX.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
ftentox_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FTENTOX.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FTENTOX.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	ftentox_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	ftentox.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTENTOX.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	ftentox.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTENTOX.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
ftentox_data_compressed::
	.dc.b	$00,$02,$84,$2C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CF,$ED,$1F,$8A,$7F,$87,$A7,$FC,$D0,$5F,$8A,$E1
	.dc.b	$F0,$FC,$7D,$FE,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$1F
	.dc.b	$07,$9F,$DB,$F4,$7B,$7F,$DF,$7E,$6F,$FA,$B0,$F8,$8C,$7C,$46,$7E
	.dc.b	$23,$5F,$11,$BF,$88,$E1,$F1,$1C,$7E,$17,$97,$C0,$F3,$F8,$1E,$9F
	.dc.b	$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03
	.dc.b	$E3,$F8,$1F,$27,$C0,$F9,$7E,$07,$CD,$F0,$3E,$7F,$81,$FF,$0A,$7C
	.dc.b	$E7,$E1,$BE,$13,$FC,$40,$BE,$07,$3E,$66,$BE,$07,$7F,$03,$C3,$E0
	.dc.b	$78,$FC,$0F,$2F,$81,$E7,$F0,$3F,$63,$D2,$EB,$F0,$3D,$BE,$07,$BF
	.dc.b	$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93,$E0,$7C,$BF
	.dc.b	$03,$E6,$F8,$1F,$3F,$C0,$FA,$3E,$07,$FC,$59,$F3,$BF,$88,$E9,$FF
	.dc.b	$1C,$7C,$07,$F0,$A5,$1F,$EF,$63,$FD,$60,$7F,$9A,$0F,$EB,$47,$F1
	.dc.b	$23,$81,$FC,$41,$FF,$21,$2F,$7B,$BF,$BD,$E0,$F7,$BC,$3E,$F7,$8B
	.dc.b	$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD,$E8,$F7,$BF
	.dc.b	$3B,$EF,$FE,$7B,$5F,$CF,$EB,$FA,$0D,$7F,$43,$AF,$BB,$DE,$F7,$FB
	.dc.b	$DF,$E4,$BF,$6F,$FC,$48,$7E,$3F,$8B,$FC,$B0,$BD,$CD,$3B,$F7,$38
	.dc.b	$7B,$9C,$7D,$CE,$5E,$E7,$3F,$73,$A7,$B9,$FB,$0E,$8E,$DE,$E7,$7F
	.dc.b	$73,$C1,$EE,$7E,$D3,$AF,$F6,$BD,$7F,$B6,$EB,$FD,$BF,$5F,$EE,$3A
	.dc.b	$FF,$73,$D7,$FB,$AE,$6F,$DD,$F3,$7F,$98,$FF,$15,$FF,$1E,$1E,$9E
	.dc.b	$73,$E9,$7F,$CE,$4F,$FB,$38,$FF,$44,$1F,$DE,$0F,$E6,$07,$F2,$F7
	.dc.b	$8F,$0F,$5E,$1E,$DC,$1D,$F8,$3E,$47,$BB,$F2,$7D,$DF,$95,$EE,$FC
	.dc.b	$BF,$77,$E6,$7B,$BF,$37,$DD,$F9,$DE,$EF,$CF,$F7,$7E,$87,$BB,$FE
	.dc.b	$7B,$F9,$FF,$D4,$70,$1F,$D4,$3F,$E8,$95,$EE,$6F,$DC,$E1,$EE,$71
	.dc.b	$F7,$39,$7B,$9C,$FD,$CE,$9E,$E7,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B
	.dc.b	$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$F3,$7C,$9F,$9D,$D9,$F3
	.dc.b	$FB,$3E,$87,$67,$D1,$EC,$FA,$5D,$9F,$4F,$B3,$EA,$76,$7D,$5E,$CF
	.dc.b	$AD,$D9,$FE,$99,$F4,$FD,$E7,$BB,$B4,$FB,$9F,$F5,$42,$EB,$E1,$D7
	.dc.b	$C7,$AF,$97,$5F,$3E,$BE,$9D,$7D,$7A,$FE,$67,$B7,$F3,$7D,$BF,$9D
	.dc.b	$ED,$F8,$7B,$BE,$87,$B7,$FE,$AB,$FA,$0F,$E0,$1C,$05,$F0,$99,$F8
	.dc.b	$AD,$7C,$AE,$FE,$17,$87,$C0,$F1,$F8,$6E,$5F,$01,$CF,$E0,$3A,$74
	.dc.b	$75,$F8,$0F,$81,$E4,$FC,$1F,$27,$E1,$79,$3F,$0F,$C9,$F8,$9E,$4F
	.dc.b	$8F,$E0,$3C,$9E,$D7,$97,$DA,$F3,$7B,$5E,$7F,$6B,$D1,$ED,$7A,$7D
	.dc.b	$AF,$57,$B5,$EB,$F6,$BD,$9E,$D7,$FB,$1F,$E8,$7F,$D4,$07,$E9,$4F
	.dc.b	$D2,$76,$97,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1F,$67,$D9
	.dc.b	$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BB,$E1,$F7,$7C,$5E,$EF
	.dc.b	$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE,$F9,$FD,$DF,$47,$BB,$F8,$7F
	.dc.b	$57,$D5,$EE,$FF,$B8,$7C,$FF,$D3,$9F,$89,$3F,$11,$D2,$7F,$10,$FF
	.dc.b	$BC,$5F,$F7,$61,$FE,$A5,$3F,$CB,$AE,$FD,$9E,$9E,$CF,$5F,$67,$B7
	.dc.b	$AF,$DF,$CC,$F0,$79,$9E,$1F,$33,$C5,$E6,$78,$FC,$CF,$27,$67,$97
	.dc.b	$7F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$FB,$E7,$E8,$BF,$1E,$7E,$3B,$84
	.dc.b	$17,$B1,$BE,$3E,$1E,$C7,$1F,$63,$97,$B1,$CF,$D8,$E9,$EC,$75,$F6
	.dc.b	$3B,$7B,$1F,$AF,$DF,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E
	.dc.b	$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7A,$7D,$8F,$3F,$3B,$E8
	.dc.b	$FF,$DD,$87,$F3,$13,$F1,$A7,$E3,$1F,$6A,$F5,$F8,$BF,$EA,$83,$FC
	.dc.b	$C1,$C2,$EB,$CF,$AB,$AF,$57,$7E,$AF,$0F,$57,$8F,$AB,$CB,$D5,$E7
	.dc.b	$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E,$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F
	.dc.b	$57,$C7,$EA,$FB,$3E,$93,$F9,$99,$FC,$54,$FC,$69,$F8,$CE,$62,$F7
	.dc.b	$F8,$7B,$FC,$7D,$FE,$5E,$CF,$3F,$67,$A7,$B3,$D7,$D9,$ED,$EC,$F7
	.dc.b	$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9,$EC,$F9,$7C,$FF
	.dc.b	$37,$85,$E7,$F6,$7E,$9F,$C0,$7F,$1A,$3F,$19,$CB,$FC,$A7,$D9,$CF
	.dc.b	$B3,$AF,$67,$7E,$CF,$0F,$67,$8F,$B3,$CB,$D9,$E7,$E6,$74,$F6,$7A
	.dc.b	$FB,$3D,$BD,$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F,$67,$C7,$EC,$F9
	.dc.b	$3D,$9F,$2F,$B3,$E6,$F6,$7F,$13,$F4,$BF,$8A,$E6,$3F,$DE,$4F,$E2
	.dc.b	$A7,$E2,$5C,$FB,$5C,$3D,$DE,$3E,$EF,$2F,$6F,$9F,$B7,$D3,$DB,$EB
	.dc.b	$ED,$F6,$F6,$FB,$FB,$7E,$0F,$6F,$C3,$ED,$F8,$BD,$BF,$1F,$B7,$E4
	.dc.b	$F6,$FC,$BE,$DF,$9B,$DB,$F3,$FB,$7E,$8F,$6F,$F4,$3F,$4D,$FD,$E0
	.dc.b	$FE,$B4,$7F,$12,$3F,$0C,$7E,$17,$97,$F7,$AB,$DA,$E3,$ED,$72,$F6
	.dc.b	$B9,$FB,$5D,$3D,$AE,$BE,$D7,$6F,$6B,$BF,$B5,$E0,$F6,$BC,$3E,$D7
	.dc.b	$8B,$DA,$F1,$FB,$5E,$4F,$6B,$CB,$ED,$79,$BD,$AF,$3F,$B5,$FB,$8F
	.dc.b	$A7,$FE,$BE,$7E,$18,$FE,$1F,$C9,$FC,$49,$FE,$54,$DF,$63,$87,$B1
	.dc.b	$C7,$D8,$E5,$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1
	.dc.b	$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E
	.dc.b	$7F,$63,$F8,$7F,$AF,$F8,$73,$FC,$7C,$7F,$5A,$3F,$0C,$7F,$0D,$E4
	.dc.b	$FE,$56,$BD,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6,$F5
	.dc.b	$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC,$BE
	.dc.b	$BF,$9B,$D7,$F3,$FA,$FE,$8F,$5F,$F9,$3F,$D4,$7F,$8F,$8F,$F1,$01
	.dc.b	$FD,$60,$FE,$18,$7E,$1F,$94,$FE,$1D,$FE,$8A,$FF,$58,$1E,$AF,$4F
	.dc.b	$57,$AF,$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E
	.dc.b	$AF,$93,$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FA,$3D,$5F,$E7,$FF,$53
	.dc.b	$FE,$26,$3F,$86,$9F,$88,$E2,$3F,$88,$6F,$AB,$C3,$D5,$E3,$EA,$F2
	.dc.b	$F5,$79,$FA,$BD,$3D,$5E,$BE,$AF,$6F,$57,$BF,$AB,$E0,$F5,$7C,$3E
	.dc.b	$AF,$8B,$D5,$F1,$FA,$BE,$4F,$57,$CB,$EA,$F9,$BD,$5F,$3F,$AB,$FD
	.dc.b	$5B,$EA,$BF,$87,$1F,$E8,$53,$FC,$42,$7F,$0D,$3F,$11,$C8,$7F,$10
	.dc.b	$FF,$74,$5E,$A7,$3F,$53,$A7,$A9,$D7,$D4,$ED,$EA,$77,$F5,$3C,$1E
	.dc.b	$A7,$87,$D4,$F1,$7A,$9E,$3F,$53,$C9,$EA,$79,$7D,$4F,$37,$A9,$E7
	.dc.b	$F5,$3D,$1E,$A7,$A7,$D4,$FE,$DF,$E2,$7F,$A1,$8F,$F3,$41,$FE,$20
	.dc.b	$3F,$AA,$1F,$89,$3F,$11,$C8,$7F,$C3,$E7,$E2,$1F,$F0,$7A,$F4,$BA
	.dc.b	$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA,$5E,$2F,$4B,$C7,$E9,$79
	.dc.b	$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$A3,$D2,$FF,$03,$FD,$5F,$F9
	.dc.b	$B8,$FE,$AA,$7E,$24,$FC,$41,$E7,$E1,$FF,$11,$AF,$4B,$8F,$A5,$CB
	.dc.b	$D2,$E7,$E9,$74,$F4,$BA,$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA
	.dc.b	$5E,$2F,$4B,$C7,$E9,$79,$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$F8
	.dc.b	$7B,$EB,$3F,$AB,$1F,$EC,$D3,$FC,$D2,$7F,$55,$3F,$12,$7E,$23,$90
	.dc.b	$FE,$21,$FF,$1E,$BF,$EC,$D1,$E8,$75,$F4,$3B,$7A,$1D,$FD,$0F,$07
	.dc.b	$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F,$43,$CD,$E8,$79
	.dc.b	$FD,$0F,$47,$A1,$E9,$F4,$3F,$C6,$FF,$5B,$FE,$CE,$3F,$D6,$07,$F9
	.dc.b	$A0,$FF,$0E,$1F,$C4,$4F,$C4,$9F,$E6,$63,$F1,$1C,$47,$F1,$0F,$F9
	.dc.b	$61,$79,$FD,$BC,$FE,$FE,$7F,$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7
	.dc.b	$F9,$3C,$FF,$2F,$9F,$E6,$F3,$FC,$FE,$7F,$A3,$CF,$FF,$27,$FC,$17
	.dc.b	$FA,$D8,$FF,$0E,$9F,$C4,$4F,$E2,$07,$E2,$78,$3F,$C6,$9E,$7F,$0F
	.dc.b	$3F,$8F,$9F,$CB,$CF,$E7,$E7,$F4,$F3,$FA,$F9,$FD,$BC,$FE,$FE,$7F
	.dc.b	$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7,$F9,$3C,$FF,$2F,$9F,$E6,$F3
	.dc.b	$FF,$CC,$BF,$5D,$FE,$1D,$1F,$27,$8F,$8F,$CF,$C4,$EB,$DE,$DF,$BD
	.dc.b	$C3,$DA,$E3,$ED,$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$FE,$7C,$FA,$F7,$CB,$DF,$94,$C7,$C8,$67,$E2,$B5
	.dc.b	$E7,$6F,$E1,$B8,$7C,$37,$1F,$84,$E5,$F0,$3C,$FD,$EE,$9E,$F7,$5F
	.dc.b	$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27
	.dc.b	$BD,$E5,$F7,$BC,$DE,$F7,$FA,$4B,$EC,$3F,$DF,$27,$F9,$80,$FE,$A8
	.dc.b	$7F,$53,$3F,$08,$7F,$0B,$F7,$FF,$D5,$EB,$DE,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$FD,$53,$EC,$FF,$98
	.dc.b	$4F,$C2,$F1,$FF,$B0,$9F,$F7,$A9,$FE,$61,$3F,$AA,$9F,$C3,$4F,$E1
	.dc.b	$8F,$F8,$77,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8E,$FE,$C7,$83
	.dc.b	$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79,$3D,$8F,$2F,$B1,$E6,$F6,$3C
	.dc.b	$FE,$C7,$A3,$D8,$FF,$60,$FD,$8F,$E1,$B8,$BF,$DB,$2B,$D7,$CF,$AF
	.dc.b	$AF,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D
	.dc.b	$7F,$2F,$AF,$E6,$F5,$FC,$FE,$BF,$FB,$63,$EC,$9F,$35,$7E,$4F,$1F
	.dc.b	$1F,$9F,$8F,$D7,$C4,$EF,$E1,$B8,$7B,$BC,$7D,$DE,$5E,$EF,$3F,$77
	.dc.b	$A7,$BB,$D7,$DD,$ED,$EE,$F7,$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE
	.dc.b	$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$FE,$EE,$FB,$3F,$F5,$31,$F8
	.dc.b	$6F,$38,$BE,$17,$5F,$0B,$BF,$85,$E1,$F0,$3C,$7E,$17,$97,$C1,$73
	.dc.b	$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F,$82,$F0,$7C,$17,$87,$E0
	.dc.b	$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E,$0B,$CD,$F0,$5E,$7F,$82
	.dc.b	$F4,$7C,$13,$9E,$78,$FF,$87,$7F,$BE,$7C,$0E,$5F,$EF,$DF,$03,$67
	.dc.b	$FD,$DC,$7F,$98,$CF,$F0,$D8,$F7,$B5,$EF,$6F,$DE,$E1,$EF,$71,$F7
	.dc.b	$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE
	.dc.b	$1F,$7B,$C5,$EF,$78,$FD,$EF,$37,$DA,$7E,$23,$93,$D8,$FF,$BB,$8F
	.dc.b	$F3,$11,$FE,$61,$3F,$A9,$1F,$C3,$D3,$FC,$3D,$E3,$EC,$74,$F6,$3A
	.dc.b	$FB,$1D,$BD,$8E,$FE,$C7,$83,$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79
	.dc.b	$3D,$8F,$2F,$B1,$E6,$F6,$3C,$FE,$C7,$A3,$D8,$F9,$DF,$6B,$FE,$EF
	.dc.b	$3F,$14,$7E,$27,$87,$EB,$3F,$EE,$F3,$FC,$C6,$7F,$86,$CF,$E1,$E3
	.dc.b	$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD,$3E,$FE,$9F,$83,$D3,$F0,$FA
	.dc.b	$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F,$A7,$E6,$F4,$FC,$FE,$9F,$A3
	.dc.b	$D3,$FC,$37,$DB,$7E,$24,$FE,$1D,$C7,$F8,$D5,$E9,$EB,$E1,$37,$F0
	.dc.b	$9C,$3E,$13,$8F,$AB,$CB,$D5,$E7,$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E
	.dc.b	$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F,$57,$C7,$EA,$F9,$3D,$5F,$2F,$AB
	.dc.b	$E6,$F5,$7C,$FE,$AF,$E7,$78,$7F,$11,$CD,$FF,$00,$7E,$0B,$1F,$05
	.dc.b	$9F,$82,$D7,$C1,$6F,$DA,$E1,$F0,$5C,$7D,$AE,$5E,$D7,$3F,$6B,$A7
	.dc.b	$B5,$D7,$DA,$ED,$ED,$77,$F6,$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F
	.dc.b	$6B,$C9,$ED,$79,$7D,$AF,$37,$B5,$FB,$3F,$B7,$FD,$A7,$39,$FE,$20
	.dc.b	$7E,$1D,$C7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$D1,$EF,$7F,$0B,$FB
	.dc.b	$8F,$E1,$C7,$E1,$B9,$8F,$E1,$9F,$E2,$8B,$DD,$E1,$EE,$F1,$F3,$B9
	.dc.b	$7B,$BC,$FD,$DE,$9E,$EF,$5F,$77,$B7,$BB,$DF,$DD,$F0,$7B,$BE,$1F
	.dc.b	$77,$C5,$EE,$F8,$FD,$DF,$27,$BB,$E5,$F7,$7C,$DE,$EF,$9F,$DD,$F4
	.dc.b	$7B,$BF,$C8,$FE,$E7,$F0,$E3,$DE,$C7,$BD,$9F,$7B,$5E,$F6,$FD,$EE
	.dc.b	$1E,$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$75,$F7,$BB,$7B,$DD,$FD
	.dc.b	$EF,$07,$BD,$E1,$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD
	.dc.b	$EF,$7F,$39,$F6,$BF,$0E,$7F,$55,$3F,$0C,$3D,$DD,$7B,$BB,$F7,$78
	.dc.b	$4F,$1F,$77,$97,$BB,$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7B,$BD,$FD,$DF
	.dc.b	$07,$BB,$E1,$F7,$7C,$5E,$EF,$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE
	.dc.b	$F9,$FD,$DF,$EA,$1F,$75,$FD,$4C,$FC,$31,$F8,$5F,$78,$FE,$15,$FE
	.dc.b	$B4,$BD,$CE,$3E,$E7,$2F,$73,$9F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B
	.dc.b	$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7
	.dc.b	$9B,$DC,$F3,$FB,$9E,$8F,$73,$FB,$47,$DD,$FF,$0C,$3F,$0D,$CB,$FD
	.dc.b	$BD,$7B,$7A,$F6,$F7,$ED,$F0,$F6,$F8,$FB,$7C,$BD,$BE,$7E,$DF,$4F
	.dc.b	$6D,$FC,$03,$0B,$A7,$F7,$BF,$9F,$F8,$5F,$15,$FC,$1A,$31,$DF,$FC
	.dc.b	$79,$FE,$F9,$3A,$74,$D7,$F8,$6B,$5D,$C7,$17,$9C,$F5,$53,$B7,$27
	.dc.b	$49,$E0,$E9,$B1,$B8,$C4,$FF,$84,$FE,$D5,$B7,$82,$30,$D9,$4E,$9F
	.dc.b	$81,$E0,$7F,$85,$E3,$11,$D0,$DC,$F1,$8C,$DF,$C1,$77,$F4,$0F,$77
	.dc.b	$95,$6C,$97,$EC,$AE,$3C,$DA,$8C,$4E,$9F,$04,$F6,$C7,$51,$89,$D6
	.dc.b	$3E,$17,$FC,$5D,$8F,$86,$7D,$E7,$42,$1C,$4F,$BD,$87,$4A,$E5,$1A
	.dc.b	$6B,$F5,$2F,$59,$FF,$89,$71,$F8,$2D,$C7,$D1,$8F,$86,$FE,$39,$CE
	.dc.b	$74,$F5,$C7,$9E,$7E,$68,$F2,$13,$26,$5A,$8E,$66,$6F,$9A,$C3,$C1
	.dc.b	$97,$93,$0E,$98,$7A,$89,$8F,$39,$D5,$F4,$9D,$63,$9C,$BE,$6E,$5D
	.dc.b	$B2,$EB,$18,$CB,$AC,$F8,$AD,$97,$1C,$7E,$05,$78,$B1,$C4,$7A,$63
	.dc.b	$E0,$63,$D2,$36,$78,$F0,$E9,$7C,$30,$F5,$47,$7C,$30,$F8,$4F,$36
	.dc.b	$78,$C7,$CA,$7C,$2B,$C6,$1D,$63,$C8,$E8,$C1,$33,$FC,$DD,$70,$7E
	.dc.b	$39,$F9,$88,$E0,$E3,$3F,$C9,$E3,$B7,$F0,$75,$6B,$F0,$07,$08,$F9
	.dc.b	$4F,$D1,$9D,$3F,$82,$CF,$E2,$3F,$1C,$F2,$4F,$F1,$0C,$77,$9F,$D8
	.dc.b	$9E,$87,$CA,$C7,$D2,$E0,$E4,$FA,$5E,$66,$EB,$F8,$DE,$1D,$33,$F9
	.dc.b	$73,$C6,$FE,$8E,$7F,$2E,$31,$3F,$46,$3E,$5C,$74,$3B,$DF,$ED,$33
	.dc.b	$E9,$3C,$67,$57,$D1,$F8,$47,$48,$F4,$67,$F0,$A7,$59,$7C,$FA,$FD
	.dc.b	$C9,$B3,$9E,$5D,$9F,$D3,$D7,$1C,$BC,$D1,$E0,$CF,$57,$FC,$36,$39
	.dc.b	$3D,$5F,$D6,$65,$AC,$FE,$84,$69,$FB,$ED,$3F,$86,$E1,$DF,$3C,$47
	.dc.b	$F3,$CC,$39,$3F,$CB,$8C,$E7,$FC,$3A,$B8,$47,$C4,$CB,$F4,$2E,$9F
	.dc.b	$D0,$36,$FF,$53,$8F,$48,$EC,$64,$F9,$67,$38,$F9,$71,$BB,$FD,$BC
	.dc.b	$EF,$2F,$84,$EC,$FB,$58,$79,$23,$E2,$4F,$F0,$A9,$E1,$97,$E2,$1A
	.dc.b	$7F,$B8,$EB,$84,$7E,$7A,$BF,$20,$72,$C7,$E9,$BE,$03,$18,$F9,$AF
	.dc.b	$99,$EC,$3C,$91,$E2,$CF,$E0,$61,$DA,$7F,$34,$7E,$10,$DE,$1F,$59
	.dc.b	$A7,$D6,$CB,$E1,$67,$F0,$EF,$9C,$62,$7C,$39,$FF,$32,$7C,$78,$77
	.dc.b	$7F,$3A,$77,$C7,$D2,$FE,$48,$DD,$73,$7C,$27,$F3,$DB,$EE,$F6,$C3
	.dc.b	$C4,$F0,$30,$78,$A3,$6F,$2C,$3E,$93,$73,$FC,$2B,$73,$FC,$2B,$C8
	.dc.b	$C1,$F8,$4C,$7C,$E7,$E7,$F2,$F6,$5F,$EB,$B3,$FC,$FD,$70,$AF,$AD
	.dc.b	$1F,$6C,$FE,$3B,$1D,$63,$9C,$7C,$F6,$C7,$CF,$35,$1C,$5F,$ED,$D8
	.dc.b	$FB,$27,$71,$FD,$87,$2F,$86,$F8,$6B,$F8,$53,$C8,$FB,$E7,$CB,$3E
	.dc.b	$41,$B3,$EA,$3D,$DD,$19,$3F,$7A,$73,$3E,$D9,$A9,$F1,$DF,$ED,$73
	.dc.b	$FB,$51,$8C,$FE,$7B,$73,$F9,$F7,$D6,$FF,$5A,$DC,$BD,$D8,$F8,$E7
	.dc.b	$EC,$DF,$1A,$FA,$93,$9C,$3F,$98,$C7,$3C,$BE,$1D,$6C,$FE,$7B,$19
	.dc.b	$3F,$0C,$7C,$27,$55,$F6,$9B,$1F,$C4,$4F,$85,$9F,$E0,$2F,$F6,$75
	.dc.b	$B3,$C5,$97,$F3,$CB,$E3,$7F,$20,$F0,$1E,$D3,$C1,$3C,$1F,$F0,$66
	.dc.b	$1E,$88,$C4,$FA,$0E,$78,$FC,$66,$E5,$F1,$AF,$EE,$CF,$C6,$AE,$27
	.dc.b	$F0,$F8,$F4,$E7,$F5,$23,$31,$ED,$BF,$3E,$7E,$DC,$BF,$AD,$65,$E3
	.dc.b	$AE,$B9,$FC,$EB,$F1,$CE,$67,$F0,$79,$7E,$05,$FC,$33,$FA,$57,$F3
	.dc.b	$27,$C0,$39,$DF,$F5,$DA,$F5,$C6,$4E,$77,$FC,$F1,$F9,$07,$A8,$EA
	.dc.b	$FD,$27,$A9,$93,$E7,$C7,$47,$0F,$A1,$FC,$09,$8C,$7E,$1B,$FC,$A6
	.dc.b	$38,$1F,$13,$2F,$91,$1E,$83,$F2,$25,$8F,$AD,$3C,$B0,$FA,$CD,$CF
	.dc.b	$EB,$57,$B3,$1F,$81,$39,$8F,$81,$5F,$1D,$FD,$E1,$E6,$BF,$13,$F0
	.dc.b	$2B,$FB,$1D,$F4,$AE,$07,$E0,$67,$C9,$7F,$81,$BE,$31,$F1,$23,$C0
	.dc.b	$FC,$13,$67,$F0,$22,$9F,$E7,$B3,$CB,$8F,$CF,$7C,$A7,$D4,$9E,$A7
	.dc.b	$CD,$71,$3F,$33,$3F,$C6,$47,$F8,$2C,$F6,$46,$72,$F9,$B1,$F1,$4F
	.dc.b	$9A,$5A,$F9,$AF,$93,$0F,$9A,$DC,$7E,$16,$3E,$59,$F8,$53,$C7,$0F
	.dc.b	$84,$77,$7E,$51,$FD,$16,$B8,$E5,$FE,$0E,$CF,$F3,$97,$F3,$30,$FD
	.dc.b	$06,$1F,$76,$73,$1D,$CC,$E1,$E5,$8C,$DF,$F8,$73,$3F,$83,$97,$F8
	.dc.b	$12,$73,$5F,$D7,$A7,$38,$7F,$AD,$A7,$35,$FC,$FA,$73,$5F,$E6,$43
	.dc.b	$E6,$9F,$59,$CB,$F5,$23,$F3,$35,$EF,$8F,$C0,$1F,$1C,$FB,$77,$EF
	.dc.b	$7E,$99,$CF,$3F,$A7,$97,$C1,$9F,$E3,$95,$CE,$FE,$F1,$DA,$F8,$1F
	.dc.b	$31,$FE,$87,$9F,$CC,$8F,$EE,$59,$7B,$2B,$FA,$76,$5E,$8C,$7E,$B9
	.dc.b	$DE,$BF,$02,$DC,$3F,$03,$1F,$6C,$FC,$B1,$4F,$CB,$1F,$9B,$31,$1F
	.dc.b	$86,$6E,$1E,$39,$FE,$AF,$87,$8E,$7F,$B2,$DF,$8F,$3F,$E2,$15,$ED
	.dc.b	$9E,$D3,$9C,$BF,$73,$1F,$2C,$F0,$63,$E7,$CB,$FA,$24,$66,$FF,$4B
	.dc.b	$1F,$89,$8F,$9E,$74,$CF,$E7,$E3,$F8,$BC,$3E,$7C,$FC,$C3,$EC,$9E
	.dc.b	$A3,$EC,$9F,$30,$E7,$87,$85,$FF,$26,$57,$49,$FE,$0F,$1C,$30,$F8
	.dc.b	$2F,$0B,$F0,$3F,$E8,$2A,$E0,$FE,$94,$E1,$87,$F7,$3C,$BE,$44,$7F
	.dc.b	$80,$B2,$F8,$8F,$F5,$A3,$A6,$3F,$A1,$BE,$78,$FE,$12,$79,$9F,$D6
	.dc.b	$2F,$9E,$5F,$C8,$70,$E7,$5F,$E6,$2C,$7F,$8B,$BF,$2A,$FF,$5C,$76
	.dc.b	$3F,$55,$87,$69,$FF,$04,$67,$F7,$DF,$44,$EE,$BF,$59,$5B,$8F,$ED
	.dc.b	$35,$B7,$C8,$6E,$BF,$2F,$3B,$8F,$EB,$F3,$B7,$FB,$0E,$3E,$AC,$FF
	.dc.b	$93,$63,$EA,$C3,$F0,$75,$F6,$4F,$C3,$CF,$D0,$C7,$DD,$1F,$87,$9F
	.dc.b	$D7,$1F,$87,$7F,$7E,$7E,$1C,$FD,$73,$EE,$7E,$59,$E2,$3D,$07,$E7
	.dc.b	$31,$F4,$65,$FD,$1B,$2F,$43,$FD,$23,$2F,$46,$7F,$E8,$BB,$FC,$7C
	.dc.b	$FE,$2D,$FD,$6E,$3F,$45,$FD,$6D,$70,$7F,$D4,$38,$76,$CB,$F9,$96
	.dc.b	$3D,$9F,$E6,$95,$BB,$FF,$35,$56,$F1,$FF,$74,$5E,$DF,$F0,$8E,$1E
	.dc.b	$A3,$F1,$D3,$B7,$FD,$10,$7C,$A8,$FE,$34,$FC,$6C,$7D,$19,$FD,$18
	.dc.b	$FE,$0F,$1E,$B3,$FD,$2C,$F1,$3F,$54,$71,$CB,$EA,$67,$FE,$6A,$8F
	.dc.b	$87,$18,$7F,$C5,$11,$F5,$9F,$C9,$9C,$6F,$F4,$0F,$D6,$3F,$1A,$7D
	.dc.b	$6A,$FA,$0F,$B5,$FD,$06,$1E,$DC,$BF,$D7,$13,$BC,$3E,$4B,$B7,$F1
	.dc.b	$B3,$BC,$3F,$D1,$47,$8E,$7E,$A9,$BC,$3E,$E5,$FF,$38,$9F,$D5,$9B
	.dc.b	$C3,$E2,$67,$F5,$63,$FD,$2F,$7F,$02,$FF,$84,$BF,$1E,$FF,$84,$BE
	.dc.b	$F3,$F6,$27,$BC,$FD,$29,$E9,$8F,$B6,$EE,$3F,$62,$EE,$FF,$28,$EC
	.dc.b	$FF,$03,$1B,$AF,$D8,$BB,$C7,$F1,$AE,$EB,$FB,$63,$EF,$3F,$26,$6E
	.dc.b	$FF,$1A,$EE,$BF,$B7,$3B,$9F,$C9,$BB,$3F,$1A,$FE,$7E,$FE,$2C,$6E
	.dc.b	$7E,$2C,$FC,$B8,$FD,$C1,$FC,$DD,$C3,$FB,$F7,$95,$FE,$26,$3F,$C9
	.dc.b	$D3,$F4,$9E,$0F,$EF,$E7,$FA,$85,$7E,$6B,$1F,$BD,$5F,$E5,$E9,$E7
	.dc.b	$87,$E8,$1D,$E1,$D7,$2F,$D0,$57,$63,$CA,$7F,$09,$3E,$E1,$FC,$F4
	.dc.b	$FB,$98,$FF,$26,$7E,$81,$FC,$24,$F8,$95,$F9,$47,$67,$F2,$63,$67
	.dc.b	$ED,$8D,$C7,$E5,$1F,$69,$FA,$68,$F3,$DF,$E9,$A3,$C7,$1F,$D1,$5E
	.dc.b	$A7,$F1,$B3,$AE,$3F,$A6,$7A,$9F,$0E,$BF,$35,$87,$CD,$9F,$F0,$01
	.dc.b	$FB,$FC,$FF,$86,$57,$E3,$2F,$C2,$FE,$60,$F0,$DF,$F2,$B7,$C3,$5F
	.dc.b	$B6,$7C,$37,$FB,$07,$C2,$7F,$B8,$67,$C3,$87,$D3,$7E,$E4,$7C,$FB
	.dc.b	$FC,$1D,$F8,$E3,$F2,$47,$F8,$EE,$3C,$67,$E9,$1F,$19,$FC,$20,$F1
	.dc.b	$DF,$F0,$27,$C7,$7F,$BC,$BF,$C4,$1E,$0B,$FC,$44,$F9,$95,$7F,$BC
	.dc.b	$A7,$CF,$3F,$C8,$2F,$C3,$1F,$AE,$8F,$09,$FB,$A3,$C3,$7F,$BA,$9D
	.dc.b	$C7,$E1,$4D,$DF,$EE,$9D,$9F,$3E,$BC,$97,$F9,$5C,$BF,$37,$3F,$24
	.dc.b	$FC,$29,$FC,$60,$F6,$9F,$BA,$3A,$C7,$F0,$07,$AD,$FF,$00,$76,$FE
	.dc.b	$E9,$DE,$1F,$C0,$1D,$DF,$F0,$07,$6F,$EE,$8F,$92,$FE,$9B,$0F,$92
	.dc.b	$F2,$3F,$81,$C7,$C8,$AF,$E4,$27,$C2,$8F,$D0,$1B,$7F,$4C,$7E,$D9
	.dc.b	$F9,$D1,$F0,$AB,$F4,$13,$F8,$4B,$FC,$C6,$3F,$96,$8F,$61,$D7,$0F
	.dc.b	$D0,$61,$D7,$0F,$D6,$3F,$8D,$7D,$2F,$F0,$F7,$CC,$7E,$EE,$79,$5F
	.dc.b	$E6,$67,$91,$FA,$93,$E8,$D7,$F1,$E7,$E2,$80,$FF,$BF,$5E,$27,$D9
	.dc.b	$3C,$47,$EF,$6B,$DC,$FE,$B6,$BD,$D8,$7F,$4A,$7E,$21,$F2,$A7,$6F
	.dc.b	$E8,$4D,$CF,$F1,$57,$71,$EB,$36,$7E,$5D,$D9,$FC,$24,$DC,$7F,$15
	.dc.b	$7E,$0C,$FE,$9E,$3C,$C7,$DF,$BF,$CA,$3F,$A7,$7F,$4C,$FC,$47,$F8
	.dc.b	$66,$7F,$D9,$63,$E2,$61,$F0,$EF,$C3,$87,$F0,$97,$C3,$3F,$C6,$5F
	.dc.b	$0B,$F9,$B7,$6F,$EF,$5D,$CF,$F0,$F7,$67,$E6,$CD,$C7,$EC,$DD,$DF
	.dc.b	$F0,$37,$75,$F9,$B9,$F6,$9F,$83,$3D,$58,$7F,$15,$7E,$03,$F9,$28
	.dc.b	$FC,$4D,$FF,$15,$9F,$D3,$9F,$C5,$63,$E7,$BF,$BF,$AF,$1C,$FE,$FD
	.dc.b	$F1,$D7,$F0,$5A,$DD,$7E,$92,$77,$7F,$B9,$77,$1F,$C1,$4F,$19,$F8
	.dc.b	$79,$F1,$CF,$E9,$23,$F9,$39,$FE,$DD,$8E,$B7,$FA,$4A,$EB,$7F,$BF
	.dc.b	$7A,$9F,$AA,$9D,$DF,$E8,$63,$77,$FB,$C3,$E2,$9F,$B6,$3C,$6F,$E4
	.dc.b	$E7,$C7,$5F,$99,$7F,$5B,$3F,$8D,$7A,$C7,$E3,$63,$F8,$D4,$FD,$47
	.dc.b	$F0,$F4,$9F,$D6,$4F,$FB,$72,$FF,$60,$EE,$BF,$5F,$5B,$7F,$D6,$87
	.dc.b	$E5,$6B,$FD,$8A,$FE,$C0,$EF,$7E,$27,$F1,$CF,$A1,$FD,$53,$E8,$7F
	.dc.b	$28,$EE,$BF,$1C,$ED,$FD,$53,$BB,$FC,$A3,$B3,$F1,$D3,$E0,$3E,$3B
	.dc.b	$FB,$33,$91,$FA,$33,$95,$FE,$5D,$E5,$1F,$12,$FE,$1C,$7F,$9B,$EF
	.dc.b	$EA,$9F,$CA,$9E,$31,$F8,$D9,$EE,$7E,$4E,$3B,$BF,$AB,$76,$FE,$34
	.dc.b	$FD,$11,$F1,$EF,$E7,$C7,$8A,$3E,$7C,$F8,$9F,$F7,$49,$F3,$C5,$27
	.dc.b	$F9,$24,$7F,$A9,$1E,$F1,$F8,$F8,$F2,$9F,$A1,$8F,$2B,$FB,$97,$67
	.dc.b	$F5,$38,$FC,$05,$7F,$0B,$8D,$BF,$AD,$77,$3F,$B3,$3A,$C7,$E8,$63
	.dc.b	$77,$FA,$D7,$73,$FA,$73,$71,$FA,$18,$DD,$FE,$9D,$DD,$7F,$81,$A3
	.dc.b	$E1,$57,$F9,$24,$F9,$31,$FD,$92,$FE,$AD,$7F,$52,$8C,$3D,$27,$F5
	.dc.b	$87,$CF,$BF,$E5,$2F,$F9,$A4,$FC,$01,$F9,$C3,$FA,$5C,$F8,$5F,$9F
	.dc.b	$3F,$3A,$3F,$D6,$F5,$FB,$C8,$F9,$07,$D9,$8F,$91,$3F,$C8,$1F,$E3
	.dc.b	$53,$B8,$F9,$07,$AA,$BF,$9F,$BE,$07,$F9,$FB,$E7,$BF,$E4,$CE,$DF
	.dc.b	$E4,$CF,$58,$FE,$28,$FF,$26,$3E,$71,$B3,$F8,$D1,$B8,$FE,$10,$FF
	.dc.b	$0E,$9F,$EC,$0F,$D4,$8F,$E1,$15,$F9,$48,$C1,$FB,$A3,$91,$FB,$A3
	.dc.b	$95,$FF,$63,$77,$5F,$BA,$76,$FE,$E9,$D9,$FC,$5C,$DC,$7E,$EA,$7C
	.dc.b	$11,$E9,$9F,$E4,$0F,$29,$FD,$D3,$C8,$FE,$12,$72,$3E,$DC,$7F,$1B
	.dc.b	$BF,$BF,$1F,$E0,$A3,$F5,$A7,$DF,$8F,$C4,$BF,$E1,$13,$E8,$1F,$CC
	.dc.b	$4F,$14,$FF,$0B,$7C,$47,$EF,$4F,$15,$7F,$14,$7C,$47,$69,$F1,$5F
	.dc.b	$DA,$9F,$C1,$D7,$E7,$AF,$F9,$65,$7E,$EE,$FF,$BD,$E1,$FD,$F5,$E5
	.dc.b	$5F,$C4,$CF,$CE,$4F,$F1,$38,$F0,$D7,$F1,$B7,$C3,$7F,$92,$7F,$92
	.dc.b	$1E,$A3,$F8,$B1,$EA,$3F,$81,$1C,$B0,$FE,$38,$EC,$FE,$10,$6E,$3F
	.dc.b	$81,$3B,$3F,$8E,$3E,$C7,$15,$FC,$45,$E5,$1F,$C1,$5E,$55,$FC,$15
	.dc.b	$DD,$FE,$A9,$F4,$C7,$C4,$8F,$81,$1C,$2F,$F8,$33,$C0,$FD,$5D,$78
	.dc.b	$5F,$B3,$7F,$9C,$7C,$37,$FC,$3D,$E4,$7F,$04,$9F,$E0,$EE,$27,$E6
	.dc.b	$3C,$AF,$F3,$8F,$2B,$FD,$2D,$FE,$32,$BF,$93,$BE,$79,$FE,$6E,$7E
	.dc.b	$EE,$7F,$B2,$DF,$68,$FA,$6F,$9C,$C1,$FA,$03,$13,$FC,$12,$BF,$2B
	.dc.b	$3F,$E5,$D9,$DC,$F8,$6F,$F8,$83,$E1,$96,$7F,$EF,$0A,$FF,$68,$CF
	.dc.b	$F0,$6B,$FE,$25,$5F,$C1,$A7,$BE,$1F,$C3,$23,$CB,$5B,$C3,$F2,$D3
	.dc.b	$FC,$CA,$B9,$BF,$C2,$AB,$E3,$1F,$BB,$3E,$31,$FC,$1C,$E6,$EE,$3D
	.dc.b	$2F,$E8,$EB,$B4,$FD,$B3,$C8,$E1,$EA,$7E,$F8,$EB,$7F,$C3,$5E,$A7
	.dc.b	$F1,$33,$75,$FB,$37,$77,$FC,$36,$BE,$44,$7E,$3A,$FF,$A6,$E1,$F1
	.dc.b	$F1,$FD,$86,$1F,$A9,$3E,$A5,$F8,$DF,$4D,$FF,$05,$7E,$4B,$FC,$B8
	.dc.b	$F1,$E3,$FC,$7D,$F1,$CF,$ED,$5F,$19,$FC,$5B,$1F,$CE,$BC,$1F,$E3
	.dc.b	$AF,$0C,$7F,$67,$7D,$AF,$F9,$1C,$EC,$FE,$24,$6C,$FD,$9D,$F9,$CF
	.dc.b	$C5,$5F,$F0,$83,$FB,$33,$88,$FD,$04,$62,$FF,$05,$7F,$01,$FB,$15
	.dc.b	$BA,$F0,$1F,$C2,$0F,$05,$FE,$FE,$BC,$D7,$FC,$9E,$77,$1F,$C2,$1D
	.dc.b	$C7,$EF,$DF,$86,$7F,$1E,$3F,$53,$3F,$A0,$9F,$EE,$4F,$F2,$58,$FF
	.dc.b	$0F,$1F,$87,$8C,$CF,$F7,$5C,$3E,$CD,$FE,$8F,$1F,$B9,$7F,$E2,$BF
	.dc.b	$B1,$7E,$B6,$F0,$7F,$C5,$3F,$78,$F9,$3F,$F6,$75,$6A,$73,$18,$AE
	.dc.b	$17,$FE,$D3,$BE,$0A,$FA,$F8,$7C,$23,$CD,$97,$ED,$75,$69,$FE,$38
	.dc.b	$DB,$FA,$06,$1E,$4F,$0F,$83,$70,$ED,$A7,$86,$73,$A3,$C3,$ED,$5F
	.dc.b	$30,$FF,$06,$F2,$B4,$F1,$7D,$EB,$F1,$47,$C6,$61,$B7,$D8,$FF,$77
	.dc.b	$5F,$17,$07,$53,$FC,$75,$F6,$CF,$DD,$1E,$F7,$BE,$30,$D8,$CC,$F5
	.dc.b	$9E,$1E,$FF,$A3,$8F,$F5,$7C,$5F,$E1,$2E,$2F,$AB,$E5,$7B,$89,$87
	.dc.b	$96,$3E,$CF,$95,$C7,$83,$F3,$51,$F7,$3C,$AF,$8B,$E5,$4B,$E1,$C8
	.dc.b	$9F,$93,$E5,$49,$CB,$F3,$3A,$3A,$77,$7F,$8B,$B1,$E9,$5C,$AF,$E8
	.dc.b	$9F,$4B,$CA,$FA,$7E,$54,$9E,$EF,$DA,$F4,$FF,$D5,$10,$EE,$FC,$3B
	.dc.b	$F9,$F3,$FB,$6E,$1D,$9D,$76,$FE,$3F,$1E,$F1,$EA,$C7,$A7,$C5,$48
	.dc.b	$F8,$98,$A3,$C1,$E5,$FA,$A7,$FC,$83,$8A,$BF,$55,$97,$E6,$23,$1E
	.dc.b	$5F,$9C,$F4,$78,$9E,$9F,$2F,$D5,$E5,$CC,$7B,$C7,$5E,$8F,$F1,$AE
	.dc.b	$1D,$6F,$8D,$79,$1F,$85,$E5,$FC,$3F,$2E,$4F,$A1,$F8,$DE,$07,$3A
	.dc.b	$FF,$17,$5F,$3A,$E4,$43,$E2,$9F,$AF,$D5,$87,$CF,$F8,$A9,$3F,$11
	.dc.b	$FA,$5F,$15,$D6,$3F,$C4,$37,$F0,$FE,$2B,$EB,$12,$3E,$5F,$9B,$F6
	.dc.b	$7E,$2B,$ED,$7C,$54,$9F,$9F,$1C,$DF,$B3,$9A,$B9,$47,$9B,$0E,$CF
	.dc.b	$43,$EB,$FC,$57,$E1,$48,$FD,$AF,$8A,$E9,$E2,$7F,$89,$EB,$EC,$5F
	.dc.b	$E2,$C9,$5F,$80,$F8,$AF,$C7,$7C,$54,$8F,$C1,$BF,$91,$F8,$AE,$7E
	.dc.b	$2F,$F8,$87,$07,$EB,$BF,$EC,$8B,$F3,$37,$AF,$F0,$8D,$F0,$3F,$91
	.dc.b	$AF,$9A,$FE,$2A,$BE,$56,$F9,$73,$0F,$DB,$4E,$5F,$E1,$3B,$FF,$8B
	.dc.b	$B2,$F0,$E1,$CA,$61,$C1,$EB,$E7,$76,$F8,$D9,$3D,$23,$D7,$E3,$FF
	.dc.b	$84,$F6,$FE,$7D,$C5,$FC,$AF,$B3,$CF,$8F,$98,$98,$79,$3E,$37,$D3
	.dc.b	$F1,$BE,$AF,$8D,$93,$E8,$8E,$AF,$FA,$7F,$5F,$8F,$F1,$BF,$03,$E3
	.dc.b	$64,$FB,$A3,$E1,$FC,$6F,$C4,$F8,$D9,$3F,$0A,$39,$F8,$DF,$EA,$3C
	.dc.b	$39,$CF,$12,$1F,$4B,$E3,$7E,$6F,$C6,$FC,$EF,$8D,$93,$F3,$23,$9F
	.dc.b	$07,$FB,$4F,$8B,$E0,$CF,$F5,$DE,$76,$BF,$58,$91,$F5,$31,$FB,$3F
	.dc.b	$1B,$F6,$BC,$29,$3F,$62,$3A,$BF,$E1,$53,$FB,$9E,$3F,$82,$F8,$DF
	.dc.b	$C0,$FC,$6C,$9F,$BF,$1F,$84,$F8,$D9,$1F,$82,$7F,$0F,$F1,$BD,$23
	.dc.b	$F9,$07,$C3,$7E,$9D,$F4,$57,$37,$F2,$77,$F8,$E8,$FE,$C7,$FA,$F7
	.dc.b	$F2,$8F,$0F,$F8,$C2,$F9,$6C,$9A,$F1,$3B,$78,$5F,$EF,$2E,$6F,$CF
	.dc.b	$E1,$CF,$2F,$39,$FD,$2F,$0F,$EC,$2F,$ED,$D7,$C8,$F8,$09,$1B,$E5
	.dc.b	$EA,$7D,$05,$D1,$FD,$EA,$9F,$15,$E8,$F8,$F9,$5D,$E3,$D7,$F1,$F2
	.dc.b	$3C,$4F,$BB,$E3,$FA,$F8,$BF,$E5,$4F,$17,$F3,$95,$FE,$0D,$C7,$AB
	.dc.b	$CC,$87,$A7,$E3,$7E,$3F,$C6,$C8,$F7,$3F,$2B,$E3,$7A,$C7,$FB,$87
	.dc.b	$AB,$FC,$3F,$7D,$23,$E1,$69,$F4,$0F,$A3,$87,$D2,$F8,$D9,$3F,$25
	.dc.b	$FA,$BF,$11,$D3,$83,$FC,$6D,$D5,$F7,$6B,$E2,$E5,$DB,$E2,$3E,$DF
	.dc.b	$C1,$C9,$FA,$51,$F7,$BE,$22,$47,$D5,$7F,$03,$F0,$FD,$6F,$FD,$87
	.dc.b	$D2,$DD,$87,$E1,$89,$87,$DD,$F8,$7F,$C4,$FC,$34,$8F,$C0,$3F,$8C
	.dc.b	$F8,$7E,$91,$F9,$E3,$E7,$57,$27,$F0,$8F,$E4,$AB,$FC,$6B,$FA,$FF
	.dc.b	$F7,$CF,$07,$FB,$53,$7C,$0F,$F2,$4B,$6F,$FC,$88,$FF,$75,$7F,$91
	.dc.b	$F6,$FF,$86,$21,$F0,$FD,$36,$95,$FE,$1F,$9C,$B7,$E1,$FC,$24,$75
	.dc.b	$F0,$DE,$A7,$FD,$83,$D9,$F8,$8E,$15,$F2,$9F,$EA,$8F,$F8,$0B,$3F
	.dc.b	$3E,$3E,$CC,$BD,$BF,$0F,$EE,$F8,$59,$87,$8E,$3B,$73,$FD,$DE,$8F
	.dc.b	$F2,$CE,$5D,$78,$3E,$1E,$3C,$F0,$F5,$7C,$3F,$C8,$F8,$79,$1E,$D7
	.dc.b	$E5,$FC,$3F,$4E,$0F,$46,$2F,$87,$F9,$E4,$BF,$89,$F0,$FF,$4A,$BE
	.dc.b	$9F,$C3,$49,$F9,$11,$CD,$FF,$40,$5F,$37,$E6,$5F,$37,$E1,$DF,$DC
	.dc.b	$F8,$79,$1F,$45,$FB,$FF,$0F,$CE,$3F,$36,$7D,$C9,$E5,$B7,$E0,$AF
	.dc.b	$F0,$B1,$F8,$6F,$87,$95,$F6,$DF,$C4,$FC,$3F,$3E,$1F,$92,$FE,$33
	.dc.b	$E1,$E5,$7E,$06,$3F,$1F,$F0,$FF,$E8,$4B,$FE,$E9,$3F,$E8,$F6,$F4
	.dc.b	$FD,$E9,$E9,$D1,$F9,$2B,$EF,$7C,$CF,$F3,$55,$FF,$7B,$6F,$C3,$F5
	.dc.b	$F8,$79,$18,$3F,$2F,$E6,$FF,$98,$F8,$98,$74,$9F,$C5,$BE,$5F,$85
	.dc.b	$F3,$7C,$34,$9E,$EF,$A7,$E1,$FC,$B1,$E9,$F8,$7F,$67,$C3,$48,$F2
	.dc.b	$C7,$BF,$E1,$E4,$7A,$1F,$85,$F0,$3C,$B8,$9C,$0B,$F8,$C4,$BF,$7F
	.dc.b	$C3,$FC,$9F,$86,$91,$F0,$9F,$99,$F0,$FC,$EF,$94,$F8,$88,$7F,$8E
	.dc.b	$D5,$7D,$3F,$84,$FA,$9F,$0D,$2B,$E6,$C7,$3F,$87,$FB,$1F,$0B,$27
	.dc.b	$E8,$C7,$DB,$F8,$69,$1F,$51,$FB,$DF,$03,$C6,$36,$43,$C5,$F0,$9F
	.dc.b	$83,$F8,$69,$1F,$71,$FC,$37,$C3,$71,$8F,$46,$1F,$89,$3F,$15,$5F
	.dc.b	$EC,$EF,$D8,$3F,$B6,$7A,$1F,$EE,$07,$FB,$65,$7F,$AB,$DA,$FF,$96
	.dc.b	$DD,$BF,$E0,$53,$FD,$32,$FF,$6A,$F1,$7F,$D1,$8D,$9E,$9F,$1D,$DF
	.dc.b	$E1,$BC,$1E,$AF,$87,$E0,$66,$3D,$E3,$EE,$61,$FD,$CA,$3C,$57,$E7
	.dc.b	$39,$3E,$5F,$86,$F5,$11,$F4,$7C,$37,$2C,$78,$47,$23,$D8,$FC,$1F
	.dc.b	$86,$91,$EF,$7E,$27,$C2,$78,$BE,$1B,$E3,$12,$7E,$27,$C3,$7C,$9F
	.dc.b	$86,$91,$F1,$DF,$99,$F0,$DC,$63,$E5,$57,$17,$E5,$57,$D1,$F8,$09
	.dc.b	$1F,$3D,$FA,$9F,$0D,$F0,$A7,$87,$1F,$CC,$BF,$B1,$1F,$67,$E1,$A6
	.dc.b	$1F,$5D,$FB,$9F,$0D,$F2,$BE,$0F,$EF,$7C,$2C,$AF,$B9,$1F,$81,$F8
	.dc.b	$59,$1F,$7D,$3F,$EE,$FF,$D8,$7F,$AD,$3F,$ED,$7F,$88,$FF,$4A,$73
	.dc.b	$AF,$A4,$FE,$17,$0F,$C6,$47,$28,$FF,$51,$5F,$F6,$16,$FC,$5E,$76
	.dc.b	$FF,$69,$4E,$5B,$F0,$7E,$9A,$FC,$96,$1D,$A7,$F1,$B8,$71,$EC,$F3
	.dc.b	$F9,$6C,$7F,$AE,$60,$8C,$7A,$FB,$8C,$4E,$A7,$87,$C4,$7F,$0B,$87
	.dc.b	$C4,$6E,$7A,$7C,$3D,$93,$67,$1F,$11,$9F,$87,$98,$FC,$98,$E9,$3C
	.dc.b	$6A,$4F,$CD,$8E,$9F,$0F,$23,$E8,$3D,$FE,$1F,$8C,$7D,$3A,$E2,$FC
	.dc.b	$CA,$F2,$7C,$1F,$97,$E0,$E4,$FD,$98,$F6,$7C,$37,$A7,$E1,$E4,$FD
	.dc.b	$D8,$F6,$7C,$3C,$8F,$C0,$3E,$FF,$87,$E3,$1F,$83,$AE,$2F,$DC,$AF
	.dc.b	$89,$EB,$48,$FC,$43,$F1,$FE,$03,$8C,$7E,$0A,$B8,$F3,$FE,$02,$FE
	.dc.b	$67,$C3,$4A,$FC,$83,$F3,$FE,$1B,$F9,$F7,$EC,$5F,$D7,$89,$3F,$55
	.dc.b	$FE,$B4,$AD,$D3,$F9,$1E,$37,$85,$F8,$1A,$46,$1F,$CF,$D7,$07,$F4
	.dc.b	$31,$97,$A5,$7E,$93,$E4,$7F,$CF,$FD,$5F,$8B,$7F,$D0,$3C,$97,$E2
	.dc.b	$7F,$0B,$7D,$2F,$84,$62,$39,$8A,$E2,$EC,$C4,$3A,$3A,$74,$C9,$D5
	.dc.b	$7F,$A4,$27,$F3,$9F,$1E,$F5,$DE,$BE,$1B,$7D,$67,$53,$C3,$F7,$7F
	.dc.b	$E7,$70,$7F,$9F,$EF,$F0,$E3,$1E,$A6,$CE,$1F,$1D,$FA,$BF,$C1,$5F
	.dc.b	$9E,$7B,$6F,$E5,$46,$3C,$0D,$4E,$78,$7F,$86,$CF,$0D,$BE,$D7,$C7
	.dc.b	$4D,$BE,$17,$88,$E8,$4F,$DF,$F8,$EF,$DD,$7C,$63,$F4,$CF,$EE,$78
	.dc.b	$7C,$98,$C7,$C3,$EA,$4D,$F8,$6C,$7C,$3E,$7E,$1A,$61,$C6,$D3,$FC
	.dc.b	$1F,$F0,$57,$E0,$3F,$D6,$E9,$18,$E5,$D4,$FC,$8E,$2F,$E5,$33,$C3
	.dc.b	$9B,$C1,$7D,$74,$F2,$18,$7F,$07,$AF,$AF,$2F,$1C,$63,$36,$9E,$88
	.dc.b	$F8,$BC,$0F,$8F,$FD,$11,$23,$3F,$1F,$D7,$A3,$F3,$99,$7C,$49,$F1
	.dc.b	$1E,$FE,$1E,$8F,$7C,$7C,$35,$C3,$2F,$87,$3E,$0F,$8C,$FD,$87,$C7
	.dc.b	$C8,$C1,$E4,$F8,$FF,$DA,$7C,$7C,$8C,$1E,$8C,$BF,$71,$C1,$23,$07
	.dc.b	$B3,$83,$F7,$9A,$48,$C1,$C8,$FD,$4F,$47,$E1,$70,$ED,$18,$E1,$65
	.dc.b	$FC,$17,$2E,$D3,$9C,$B0,$FC,$AF,$8F,$FE,$14,$47,$E6,$FC,$67,$F0
	.dc.b	$DF,$8F,$91,$F4,$1F,$E2,$1F,$1F,$FC,$47,$E3,$E4,$FD,$47,$F8,$A7
	.dc.b	$C7,$FF,$15,$F1,$A4,$FD,$87,$F8,$C7,$8D,$D2,$3D,$18,$5E,$6F,$E2
	.dc.b	$5C,$3F,$A1,$CB,$AD,$7B,$72,$B5,$AE,$06,$7F,$D5,$FE,$0B,$FC,$6B
	.dc.b	$89,$FC,$A6,$1F,$D8,$DB,$19,$ED,$E1,$D9,$C7,$E2,$E5,$FE,$CA,$FF
	.dc.b	$41,$7C,$FE,$3B,$B7,$C7,$CA,$D4,$78,$7E,$3E,$47,$17,$C9,$F1,$FD
	.dc.b	$A3,$A6,$3D,$9F,$B7,$8F,$61,$7D,$63,$38,$7B,$63,$DD,$F1,$F2,$7C
	.dc.b	$CF,$C1,$F8,$FE,$D1,$D3,$0E,$DC,$1F,$33,$1E,$D8,$75,$8C,$63,$D6
	.dc.b	$FE,$07,$C5,$FC,$BF,$8B,$91,$83,$E2,$7C,$7F,$CF,$23,$F1,$FE,$3F
	.dc.b	$E9,$7C,$7F,$D3,$F8,$F9,$3F,$2E,$3E,$B7,$B7,$F5,$F5,$93,$F3,$E3
	.dc.b	$9B,$F1,$30,$BA,$7A,$71,$EB,$3F,$57,$1E,$AF,$CA,$C7,$F0,$5F,$13
	.dc.b	$23,$ED,$3F,$85,$F8,$AF,$C3,$7C,$7C,$9F,$BC,$FF,$87,$FF,$64,$7C
	.dc.b	$95,$F8,$FC,$7C,$7C,$BF,$E6,$B5,$FE,$37,$EC,$FF,$18,$BF,$E2,$B1
	.dc.b	$F0,$3C,$CE,$9F,$1F,$DE,$BE,$FE,$57,$83,$F3,$F7,$E0,$9E,$B9,$77
	.dc.b	$7F,$7F,$0C,$BC,$91,$E2,$F8,$2F,$41,$1F,$2F,$BB,$EB,$F6,$FD,$9F
	.dc.b	$03,$27,$D3,$1E,$FF,$83,$F8,$1F,$1D,$27,$DB,$1F,$0F,$E3,$64,$7C
	.dc.b	$07,$E3,$7C,$77,$58,$F8,$78,$75,$7C,$F8,$75,$7E,$3E,$1D,$5F,$5E
	.dc.b	$1F,$3F,$E2,$64,$7C,$C7,$E9,$7C,$57,$58,$F8,$38,$75,$E0,$F8,$F8
	.dc.b	$F5,$9F,$8B,$8F,$59,$F9,$78,$FD,$BF,$8E,$95,$F6,$1F,$BD,$F1,$DF
	.dc.b	$7F,$E3,$A4,$FD,$C7,$F0,$5F,$1D,$F8,$3F,$8E,$FC,$27,$C7,$4A,$FC
	.dc.b	$0C,$7F,$99,$7F,$65,$FF,$3F,$47,$F3,$47,$FC,$67,$9F,$F1,$A6,$F5
	.dc.b	$FF,$9D,$9D,$C7,$E0,$31,$F0,$D7,$DE,$7C,$2F,$63,$F1,$F8,$78,$5F
	.dc.b	$C1,$E1,$E1,$F8,$99,$1F,$E0,$8C,$3F,$C0,$ED,$F8,$9F,$37,$C3,$48
	.dc.b	$D3,$E9,$F8,$6F,$57,$C2,$7A,$FE,$12,$57,$38,$F7,$7C,$0C,$8E,$CF
	.dc.b	$C1,$F8,$2E,$91,$F0,$70,$BB,$FC,$1C,$7A,$CF,$97,$1E,$AF,$C7,$C7
	.dc.b	$E5,$7C,$14,$8F,$53,$F3,$7E,$0B,$E7,$7C,$27,$CF,$F8,$49,$5E,$F8
	.dc.b	$FA,$5F,$09,$F4,$FE,$12,$4F,$C3,$8F,$AD,$F0,$5F,$5F,$E0,$A4,$FC
	.dc.b	$78,$EA,$FC,$9C,$3A,$BF,$6F,$0E,$AF,$CD,$C3,$AB,$F7,$F0,$FC,$17
	.dc.b	$C1,$7E,$0F,$E0,$64,$FD,$38,$EA,$FE,$17,$0E,$BC,$1F,$85,$C7,$AC
	.dc.b	$FE,$27,$1E,$B3,$F8,$9C,$7F,$1D,$F0,$9F,$8F,$F8,$49,$7F,$7E,$3F
	.dc.b	$D6,$DF,$B3,$7F,$60,$3F,$AE,$93,$B7,$3F,$09,$2B,$FC,$9B,$3F,$D7
	.dc.b	$07,$BF,$CB,$DF,$E7,$EF,$CA,$DC,$76,$C3,$BF,$C5,$49,$E7,$1D,$5F
	.dc.b	$F0,$AE,$1D,$71,$F2,$E3,$E6,$F8,$49,$5E,$27,$D3,$F0,$9E,$AF,$84
	.dc.b	$F5,$FC,$24,$AF,$3C,$7B,$BE,$12,$47,$A9,$F8,$3F,$09,$F0,$BE,$0A
	.dc.b	$4F,$B9,$F8,$BF,$05,$D6,$3E,$0E,$1D,$5F,$3E,$1D,$1F,$4E,$17,$97
	.dc.b	$FC,$27,$B3,$2F,$9F,$F0,$32,$FE,$5B,$F4,$BE,$0B,$FC,$EB,$EA,$7F
	.dc.b	$7A,$9F,$F3,$A3,$5F,$F2,$FB,$70,$FF,$2A,$AE,$5E,$3C,$1C,$B8,$25
	.dc.b	$F2,$8F,$F3,$7C,$C8,$FF,$54,$7C,$4F,$F9,$DA,$3F,$CE,$73,$E2,$E6
	.dc.b	$95,$E2,$7C,$B3,$DA,$3C,$B7,$D9,$F2,$D7,$E3,$3E,$13,$F1,$BF,$0B
	.dc.b	$27,$F1,$31,$D1,$F5,$E1,$73,$F5,$E3,$D6,$7D,$F8,$F5,$9F,$7E,$3F
	.dc.b	$98,$F8,$5F,$CC,$FC,$2C,$BF,$CA,$C7,$E7,$3E,$0A,$47,$E6,$13,$FE
	.dc.b	$A6,$F0,$BF,$B4,$3F,$E6,$47,$FB,$A2,$F8,$64,$7F,$69,$8E,$1F,$03
	.dc.b	$C9,$E1,$F0,$77,$3F,$F2,$8D,$3E,$13,$1D,$FE,$0C,$79,$FC,$2E,$31
	.dc.b	$ED,$7F,$E0,$07,$C9,$7E,$87,$F4,$79,$DC,$7A,$FC,$37,$1C,$7C,$B9
	.dc.b	$DC,$7C,$D8,$7A,$F4,$C4,$E7,$E1,$75,$F0,$B2,$BE,$14,$71,$F8,$19
	.dc.b	$1F,$15,$E9,$F0,$5D,$7E,$06,$4F,$C9,$7C,$1F,$05,$E1,$F8,$5F,$17
	.dc.b	$C2,$CA,$F9,$D1,$D1,$F9,$D8,$5E,$0F,$8F,$8F,$49,$FA,$58,$D9,$F9
	.dc.b	$78,$FB,$7E,$16,$57,$D9,$7E,$07,$7F,$C1,$F8,$5F,$85,$F0,$B2,$BE
	.dc.b	$F4,$7C,$5F,$84,$F8,$DF,$0B,$27,$F0,$51,$F2,$7E,$13,$E5,$7C,$2C
	.dc.b	$9F,$C3,$47,$57,$F1,$18,$75,$7F,$11,$87,$57,$F1,$98,$75,$7F,$19
	.dc.b	$87,$D7,$FD,$9F,$FB,$CF,$6F,$E0,$DF,$EF,$B3,$F8,$F6,$F4,$9D,$C7
	.dc.b	$E1,$B0,$F0,$6F,$F8,$EC,$BC,$93,$D8,$F0,$90,$C1,$FE,$C0,$AF,$EF
	.dc.b	$AD,$F8,$6F,$21,$E5,$F8,$5F,$37,$C1,$C9,$FD,$3E,$1F,$EA,$CB,$FF
	.dc.b	$74,$46,$33,$F5,$7C,$1F,$B7,$E0,$A4,$E4,$F6,$3F,$07,$E0,$FE,$17
	.dc.b	$C2,$FC,$3F,$85,$95,$BE,$0F,$81,$3E,$18,$F4,$69,$74,$F1,$65,$DA
	.dc.b	$7D,$99,$59,$F3,$65,$F3,$FE,$16,$57,$5D,$51,$F4,$FE,$17,$EA,$7C
	.dc.b	$0C,$AD,$3F,$5F,$E0,$BE,$C7,$C0,$C9,$E4,$FD,$BF,$82,$FB,$9F,$09
	.dc.b	$27,$B3,$F7,$FE,$17,$A4,$7D,$1C,$2E,$BF,$2F,$1E,$93,$F5,$71,$B3
	.dc.b	$F3,$F1,$FC,$47,$93,$F8,$9F,$85,$97,$EB,$8F,$C6,$7C,$24,$8F,$73
	.dc.b	$F8,$FF,$85,$FE,$0D,$FB,$47,$FB,$70,$FF,$6C,$93,$A7,$3F,$0B,$2F
	.dc.b	$E9,$4F,$FB,$60,$FE,$FA,$F4,$21,$F8,$BB,$EC,$F0,$AE,$FF,$0D,$23
	.dc.b	$07,$4F,$67,$C6,$47,$BF,$B3,$E9,$7A,$1D,$B6,$FC,$86,$7E,$AC,$3A
	.dc.b	$4A,$7F,$DF,$F9,$7A,$E3,$D3,$EC,$7C,$02,$3E,$CF,$67,$E1,$FC,$17
	.dc.b	$C4,$F8,$39,$3F,$02,$3E,$3F,$C1,$C8,$F8,$6F,$CA,$F8,$2F,$97,$F0
	.dc.b	$B2,$7E,$3B,$F3,$BE,$17,$AC,$7C,$AC,$3A,$BE,$7D,$7F,$A1,$68,$CB
	.dc.b	$B4,$E4,$FE,$A1,$0C,$B3,$AF,$BB,$3F,$B7,$F0,$92,$FE,$BB,$F7,$BE
	.dc.b	$13,$EF,$FC,$17,$E0,$3E,$0E,$57,$DC,$8F,$C1,$FC,$14,$8F,$BE,$FE
	.dc.b	$1B,$E0,$FF,$0F,$F0,$32,$7F,$06,$FF,$4F,$FD,$A7,$FB,$3E,$9F,$43
	.dc.b	$0E,$B5,$E9,$F3,$1B,$B3,$F1,$CF,$62,$1B,$86,$7E,$18,$FE,$05,$8F
	.dc.b	$F7,$2B,$3F,$D7,$31,$70,$FF,$6B,$8C,$37,$E0,$3C,$DF,$07,$E7,$F8
	.dc.b	$49,$8F,$48,$F5,$7C,$0F,$AF,$E0,$A4,$F8,$23,$DD,$F0,$1E,$FF,$80
	.dc.b	$93,$E4,$8E,$AF,$9B,$0E,$AF,$C1,$CF,$FC,$00,$F3,$3D,$78,$76,$75
	.dc.b	$87,$CB,$F8,$09,$1F,$01,$F9,$DF,$01,$DA,$3A,$61,$DB,$6F,$83,$1F
	.dc.b	$E1,$5C,$7B,$BD,$0E,$E4,$30,$79,$7E,$07,$ED,$11,$F9,$FF,$05,$F7
	.dc.b	$7E,$0B,$EF,$7C,$14,$9F,$A7,$1F,$81,$F7,$BF,$05,$EF,$49,$FA,$F1
	.dc.b	$F8,$5F,$7E,$47,$DA,$7F,$11,$F0,$1D,$63,$EE,$E1,$D5,$F9,$18,$75
	.dc.b	$7F,$03,$87,$57,$E6,$61,$F9,$1F,$81,$FC,$97,$C1,$49,$FC,$3C,$7F
	.dc.b	$88,$F1,$FE,$CF,$D2,$7F,$B5,$92,$75,$F0,$51,$FE,$A1,$1F,$D9,$38
	.dc.b	$3E,$6D,$71,$F8,$2E,$9F,$01,$2F,$6F,$7F,$80,$F0,$47,$87,$E3,$23
	.dc.b	$D0,$EC,$7E,$97,$1E,$DD,$DF,$E4,$36,$C7,$F9,$3D,$61,$EB,$F8,$19
	.dc.b	$3E,$67,$DD,$F0,$3E,$FF,$82,$93,$EA,$7E,$17,$C1,$FC,$3F,$82,$F8
	.dc.b	$9F,$05,$2B,$DF,$1F,$1F,$E0,$3E,$47,$C0,$49,$F8,$71,$D1,$F4,$61
	.dc.b	$78,$BE,$2E,$AC,$BB,$57,$CA,$CB,$B3,$EF,$CB,$EA,$7C,$04,$8F,$9E
	.dc.b	$FD,$7F,$80,$E9,$C1,$F0,$F3,$B5,$F0,$F1,$EA,$FC,$9C,$7A,$E7,$E2
	.dc.b	$CB,$F0,$1F,$05,$F8,$1F,$82,$97,$F7,$23,$F0,$9F,$03,$F8,$5F,$82
	.dc.b	$93,$F8,$08,$FC,$47,$C0,$FE,$27,$DE,$93,$F8,$48,$FF,$37,$7E,$D5
	.dc.b	$FB,$08,$FE,$3C,$BE,$0B,$26,$B8,$3B,$CF,$E3,$31,$EE,$FE,$07,$1E
	.dc.b	$EF,$E4,$31,$EE,$FE,$17,$1E,$FF,$07,$23,$FC,$49,$92,$3C,$7F,$13
	.dc.b	$E4,$F8,$4F,$2F,$C2,$4B,$DC,$7A,$3E,$12,$47,$27,$D7,$F0,$9E,$CF
	.dc.b	$81,$93,$D9,$F7,$FC,$17,$58,$F0,$E1,$D5,$F8,$38,$74,$7E,$1E,$17
	.dc.b	$97,$F5,$3A,$32,$F9,$3F,$05,$F2,$BE,$0A,$61,$EC,$8F,$9B,$F0,$92
	.dc.b	$3D,$EF,$D0,$F8,$4F,$A3,$F0,$9F,$4B,$E1,$25,$7C,$48,$FA,$BF,$05
	.dc.b	$F5,$BE,$0A,$4F,$C8,$8E,$8F,$D8,$C2,$EB,$E6,$C7,$AC,$FC,$EC,$7A
	.dc.b	$BF,$7B,$1F,$C0,$FC,$0F,$E0,$BE,$12,$4F,$D4,$8E,$7D,$7F,$EB,$DE
	.dc.b	$06,$78,$AF,$F6,$54,$32,$EF,$1F,$8A,$CB,$BE,$7F,$23,$3F,$C7,$FC
	.dc.b	$24,$AF,$C0,$BF,$EB,$FF,$DA,$FF,$B8,$E7,$FE,$CB,$7F,$B3,$13,$E1
	.dc.b	$35,$F0,$FB,$F8,$59,$87,$F9,$B6,$BF,$A3,$E5,$FC,$79,$FF,$68,$47
	.dc.b	$F7,$95,$F0,$5D,$FE,$0E,$4F,$18,$F1,$63,$23,$A3,$E0,$7B,$61,$E0
	.dc.b	$7A,$3E,$9F,$81,$F5,$7C,$27,$AF,$E1,$65,$79,$63,$DD,$F0,$BE,$FF
	.dc.b	$85,$93,$E9,$8F,$85,$F0,$92,$3D,$8F,$C5,$F8,$4F,$8D,$F0,$5F,$1F
	.dc.b	$E0,$E5,$7C,$18,$EA,$FC,$3C,$3A,$BF,$07,$0E,$8F,$C3,$C2,$F0,$FC
	.dc.b	$7C,$7E,$97,$C1,$4A,$F9,$8F,$D5,$F8,$3F,$AD,$F0,$92,$7E,$83,$F6
	.dc.b	$7D,$8F,$B5,$F0,$72,$7E,$A3,$F7,$7E,$13,$FD,$F9,$FB,$67,$F5,$A3
	.dc.b	$FD,$86,$49,$E8,$7D,$7B,$B9,$FD,$BC,$3A,$4F,$DB,$C2,$CF,$DF,$C3
	.dc.b	$AF,$C2,$F6,$F8,$19,$7F,$EB,$1C,$BF,$A9,$FC,$2B,$AD,$A7,$6E,$DF
	.dc.b	$C8,$3E,$BB,$F4,$90,$F0,$1F,$8F,$CB,$D7,$F0,$B2,$3B,$3E,$EF,$85
	.dc.b	$F7,$FC,$24,$9F,$13,$F0,$BE,$13,$E1,$FC,$04,$9F,$33,$F1,$BE,$03
	.dc.b	$E3,$FB,$F2,$7D,$4F,$CA,$F8,$0E,$91,$FE,$55,$E0,$63,$8E,$3F,$EF
	.dc.b	$F0,$CB,$B5,$6F,$2B,$5A,$CB,$EA,$FC,$1F,$D6,$F8,$49,$7F,$26,$3E
	.dc.b	$CF,$C2,$48,$F9,$8F,$DC,$F8,$4F,$BB,$F0,$72,$7E,$83,$F8,$0F,$84
	.dc.b	$FC,$0F,$C1,$49,$FA,$8F,$E1,$3E,$0F,$A4,$7A,$F0,$BA,$7A,$B1,$E9
	.dc.b	$3E,$FC,$6C,$FB,$B1,$FC,$6F,$C1,$4A,$FC,$03,$F9,$0F,$83,$47,$FB
	.dc.b	$CF,$D6,$FE,$C1,$7F,$E2,$46,$F5,$6B,$B3,$FC,$F2,$F0,$AF,$F4,$14
	.dc.b	$FF,$65,$79,$7C,$1E,$6F,$93,$6B,$95,$7F,$5D,$AF,$07,$C1,$E5,$61
	.dc.b	$FE,$86,$F8,$0C,$36,$3C,$D1,$8A,$7C,$0D,$C2,$57,$AD,$D7,$C2,$6F
	.dc.b	$29,$3E,$63,$C2,$7C,$1F,$82,$E9,$F0,$52,$30,$73,$21,$F0,$2F,$AB
	.dc.b	$CC,$F8,$75,$D9,$F8,$BA,$79,$FE,$12,$47,$CF,$7D,$5F,$09,$EB,$F8
	.dc.b	$39,$3F,$4D,$F7,$7C,$27,$BF,$E0,$7E,$07,$C0,$4A,$FB,$11,$F0,$FD
	.dc.b	$FF,$89,$F0,$12,$7E,$E4,$74,$7E,$BE,$17,$97,$EF,$63,$D6,$7F,$05
	.dc.b	$8F,$57,$EE,$E3,$F3,$FE,$0F,$E8,$7C,$24,$9F,$C4,$47,$D3,$E5,$91
	.dc.b	$F8,$B7,$EB,$7C,$27,$D7,$F8,$3F,$B7,$FB,$6F,$F5,$CA,$FB,$ED,$EA
	.dc.b	$FF,$62,$CE,$BE,$06,$57,$E5,$B6,$FF,$3B,$DF,$FB,$1C,$FF,$61,$C7
	.dc.b	$E2,$31,$FF,$66,$BE,$12,$1F,$96,$C7,$C4,$F3,$E6,$FC,$E5,$78,$4F
	.dc.b	$15,$31,$F3,$FC,$0F,$A3,$E0,$66,$1C,$4F,$4C,$7B,$3E,$0B,$DB,$F0
	.dc.b	$72,$75,$B3,$2F,$ED,$F1,$85,$F0,$5F,$10,$91,$9F,$83,$F8,$FF,$01
	.dc.b	$23,$83,$F2,$BE,$07,$C1,$1C,$F1,$F0,$3F,$EF,$5C,$3B,$B7,$0E,$AD
	.dc.b	$3B,$1E,$3F,$81,$FA,$BF,$03,$23,$07,$9F,$4E,$BC,$7F,$4B,$1F,$0C
	.dc.b	$F3,$24,$7D,$6C,$3E,$FE,$9F,$80,$F8,$29,$5F,$05,$FC,$1F,$C1,$FE
	.dc.b	$13,$E0,$FF,$0B,$F0,$72,$BE,$34,$7E,$23,$DE,$FC,$4F,$C1,$C9,$F9
	.dc.b	$51,$F8,$CF,$83,$FC,$6F,$C0,$49,$F9,$D1,$D5,$FA,$18,$75,$7E,$36
	.dc.b	$1D,$5F,$A9,$87,$57,$E5,$61,$FC,$4B,$F6,$EF,$F6,$79,$23,$1F,$0F
	.dc.b	$9F,$85,$91,$F8,$09,$FE,$DA,$7F,$AF,$9C,$7C,$1F,$3F,$83,$91,$83
	.dc.b	$87,$C0,$F7,$F8,$09,$18,$3A,$61,$D9,$72,$7F,$B6,$B1,$EF,$F0,$BE
	.dc.b	$7C,$BA,$D6,$BC,$2F,$E7,$54,$CF,$53,$E7,$F8,$1F,$77,$C0,$C8,$C1
	.dc.b	$EB,$F8,$4F,$85,$F0,$92,$30,$7B,$FE,$0B,$E3,$11,$F8,$5F,$07,$F2
	.dc.b	$7E,$06,$47,$C5,$7E,$67,$C0,$FC,$DC,$3A,$4F,$9C,$FE,$87,$A3,$1D
	.dc.b	$47,$72,$1B,$F1,$92,$CB,$EC,$65,$F6,$7E,$06,$5F,$D5,$7E,$E7,$C0
	.dc.b	$F4,$8E,$57,$DD,$FB,$58,$E7,$8F,$E7,$63,$ED,$9D,$BD,$9F,$C0,$7C
	.dc.b	$1F,$E1,$C8,$FE,$0F,$E0,$FF,$15,$F0,$72,$3F,$0C,$FE,$37,$E0,$FF
	.dc.b	$AF,$FE,$DF,$FD,$C2,$7F,$B1,$BF,$DB,$5F,$73,$71,$73,$7F,$75,$8E
	.dc.b	$1C,$07,$FB,$75,$70,$F7,$FA,$9D,$BD,$FF,$14,$7C,$0C,$AE,$3F,$8B
	.dc.b	$CB,$C7,$97,$93,$E0,$A5,$FF,$1C,$CB,$B8,$F8,$4F,$4F,$C2,$49,$D4
	.dc.b	$7B,$3E,$12,$47,$17,$DF,$F0,$9F,$03,$E1,$3E,$0F,$C2,$4A,$ED,$1F
	.dc.b	$13,$E0,$BE,$2F,$C1,$49,$F1,$47,$47,$E2,$E1,$77,$F8,$B1,$F2,$B1
	.dc.b	$EC,$FA,$71,$EC,$FC,$CC,$7E,$8F,$C0,$FD,$2F,$82,$93,$EF,$8F,$AB
	.dc.b	$F0,$72,$3E,$13,$F6,$3E,$13,$EC,$FC,$1C,$9F,$8C,$FD,$CF,$84,$FB
	.dc.b	$BF,$05,$F7,$BE,$0E,$57,$CB,$8E,$8F,$DE,$C2,$ED,$EA,$EB,$FB,$99
	.dc.b	$75,$AF,$C1,$E5,$6B,$5C,$AC,$FF,$17,$F0,$72,$FE,$C3,$FE,$2C,$ED
	.dc.b	$FE,$E1,$4C,$FB,$F4,$FE,$2D,$F6,$5F,$A8,$87,$80,$FC,$16,$1C,$FE
	.dc.b	$13,$A7,$C3,$C9,$FE,$C3,$8A,$7C,$1F,$09,$27,$2F,$8F,$E1,$7C,$9F
	.dc.b	$01,$27,$8B,$E7,$F8,$0F,$47,$C0,$7A,$7E,$02,$57,$68,$E8,$FF,$03
	.dc.b	$CB,$FD,$C3,$B7,$F9,$65,$B1,$FE,$4E,$7F,$DC,$7C,$3F,$35,$FE,$CE
	.dc.b	$DC,$7C,$53,$CF,$1F,$91,$F0,$BF,$27,$E1,$64,$FB,$A3,$E6,$7C,$27
	.dc.b	$CD,$F8,$49,$3F,$0A,$3E,$87,$C2,$48,$F8,$AF,$D3,$F8,$4F,$A9,$F0
	.dc.b	$3F,$57,$E0,$A5,$7C,$A8,$EA,$FC,$CC,$3A,$BE,$FC,$3A,$3F,$07,$0B
	.dc.b	$AF,$B7,$1F,$C0,$7C,$0C,$AF,$AA,$FE,$0F,$E0,$BF,$09,$F0,$72,$7E
	.dc.b	$CB,$F8,$7F,$84,$FF,$39,$FE,$E1,$FE,$81,$9F,$E6,$0D,$C7,$F2,$CD
	.dc.b	$E4,$FF,$44,$2F,$85,$E3,$F0,$72,$FF,$C4,$B3,$FE,$D2,$3F,$B3,$3F
	.dc.b	$67,$1F,$13,$4F,$C5,$E3,$E2,$61,$C8,$F2,$DF,$9B,$E0,$A4,$76,$7D
	.dc.b	$3F,$05,$D3,$3F,$C0,$61,$73,$FC,$36,$3D,$30,$F6,$96,$BD,$B8,$FC
	.dc.b	$2F,$85,$93,$EA,$7E,$2F,$C2,$FC,$6F,$84,$F8,$FF,$09,$2B,$DF,$1F
	.dc.b	$2B,$E0,$A4,$7C,$27,$E6,$FB,$DF,$3B,$DE,$93,$F1,$9F,$A3,$F0,$5D
	.dc.b	$63,$E4,$E1,$D5,$FA,$78,$74,$7E,$AE,$17,$4F,$AB,$8F,$DA,$F8,$49
	.dc.b	$5F,$49,$FB,$BF,$09,$F7,$BE,$13,$EF,$FC,$24,$AF,$AF,$1F,$82,$F8
	.dc.b	$39,$1F,$69,$FC,$2F,$8B,$F8,$6F,$81,$93,$F7,$9F,$C4,$FC,$0F,$58
	.dc.b	$FC,$0E,$1D,$5F,$C5,$E1,$D1,$FC,$6E,$17,$4F,$C6,$E3,$FE,$C4,$FD
	.dc.b	$C7,$FA,$F9,$FD,$8F,$9F,$FD,$AC,$DB,$FF,$65,$BA,$F8,$4E,$04,$8F
	.dc.b	$F3,$AC,$3E,$13,$A7,$C1,$49,$CB,$DF,$E0,$FC,$1F,$03,$E1,$F8,$19
	.dc.b	$5C,$A3,$A3,$FE,$13,$C3,$FB,$8E,$9F,$CE,$5B,$9F,$9B,$2E,$D5,$BC
	.dc.b	$AD,$EF,$2F,$77,$C0,$4A,$F3,$BF,$07,$E0,$7B,$D7,$5D,$6E,$5D,$6F
	.dc.b	$D2,$43,$B1,$F0,$30,$F9,$5F,$05,$23,$E1,$BF,$37,$E0,$FE,$77,$C1
	.dc.b	$7C,$FF,$83,$95,$F2,$23,$E9,$7B,$92,$3E,$5B,$F5,$7E,$03,$EB,$7C
	.dc.b	$04,$9F,$9E,$FD,$9F,$7B,$AC,$7D,$2C,$3A,$BF,$27,$0E,$8F,$CB,$C2
	.dc.b	$ED,$F2,$F1,$FC,$17,$C1,$4A,$FB,$6F,$E1,$7E,$0F,$F0,$DF,$04,$9F
	.dc.b	$F5,$9F,$EE,$5F,$EC,$F3,$FD,$A1,$CF,$FE,$BF,$8C,$FC,$0E,$DF,$F2
	.dc.b	$64,$FF,$B3,$BE,$10,$D3,$FA,$E4,$74,$F8,$2E,$1A,$7F,$77,$9F,$E9
	.dc.b	$F5,$C2,$73,$CB,$E3,$8E,$7F,$05,$E6,$C3,$CB,$F1,$FA,$53,$E0,$8D
	.dc.b	$47,$93,$2D,$4F,$93,$E1,$2E,$5E,$59,$BF,$05,$2F,$E0,$BA,$F8,$2D
	.dc.b	$FC,$1C,$9F,$8A,$F2,$F8,$4E,$7F,$07,$D3,$E1,$25,$7C,$A8,$EF,$F0
	.dc.b	$32,$3E,$6B,$E2,$F8,$2E,$91,$F2,$B0,$BC,$7F,$17,$1E,$B3,$F5,$31
	.dc.b	$EA,$FD,$0C,$7D,$7F,$05,$23,$EC,$BE,$EF,$82,$E7,$C9,$F2,$F8,$3F
	.dc.b	$DD,$75,$EA,$3E,$FE,$3D,$9F,$B7,$8F,$6C,$BE,$B6,$5F,$23,$E0,$FE
	.dc.b	$4F,$C2,$4B,$FC,$3C,$7C,$CF,$84,$F9,$BF,$09,$27,$F1,$71,$F4,$3E
	.dc.b	$03,$E8,$FC,$04,$9F,$C7,$C7,$D6,$3F,$B2,$3F,$71,$BD,$06,$9F,$D9
	.dc.b	$E3,$A9,$F9,$8B,$F0,$3F,$94,$BF,$03,$F9,$CB,$F0,$3F,$98,$AF,$07
	.dc.b	$C2,$78,$7E,$22,$4F,$E8,$F5,$FF,$26,$52,$BC,$DF,$09,$E7,$F8,$49
	.dc.b	$7B,$E3,$FC,$C5,$AB,$D9,$F0,$7E,$DF,$83,$97,$B3,$DD,$1F,$07,$E0
	.dc.b	$3E,$17,$C0,$C9,$D7,$0F,$C1,$9F,$0C,$63,$1F,$F6,$B3,$73,$CB,$FE
	.dc.b	$BF,$86,$1E,$08,$C6,$3E,$47,$97,$C0,$FD,$1F,$81,$91,$83,$B7,$C1
	.dc.b	$FD,$52,$3E,$1F,$83,$FB,$1F,$07,$23,$C8,$FD,$BF,$83,$FB,$9F,$01
	.dc.b	$27,$D0,$FD,$FF,$81,$E7,$1E,$7C,$2F,$2F,$D0,$C7,$A4,$FA,$F1,$B3
	.dc.b	$F5,$31,$FC,$47,$C0,$CA,$F8,$8F,$E2,$FE,$07,$9F,$37,$D4,$CA,$D7
	.dc.b	$E3,$72,$E7,$3F,$63,$2B,$3F,$91,$C7,$F2,$9F,$07,$F9,$5F,$83,$93
	.dc.b	$F4,$63,$F8,$6F,$EE,$7F,$EC,$C3,$FB,$67,$39,$A6,$7D,$ED,$92,$BE
	.dc.b	$EC,$7F,$6D,$6F,$C0,$F4,$23,$AF,$7F,$BE,$9E,$0F,$88,$93,$CA,$3A
	.dc.b	$3F,$EF,$FB,$FE,$D6,$FF,$26,$6E,$3E,$5F,$83,$F4,$FC,$24,$AF,$1B
	.dc.b	$EC,$F8,$5F,$6F,$C2,$FB,$BE,$16,$57,$A2,$3E,$0F,$C2,$C8,$F5,$BF
	.dc.b	$13,$E1,$7E,$2F,$C1,$49,$F7,$BF,$23,$E0,$FA,$C7,$C2,$C3,$AB,$E9
	.dc.b	$C3,$A3,$EB,$C2,$F5,$FF,$AF,$F6,$65,$F4,$BE,$0F,$E9,$FC,$1C,$C3
	.dc.b	$E6,$C7,$D6,$F8,$5F,$AF,$F0,$B2,$7E,$8C,$7D,$AF,$85,$FB,$7F,$0B
	.dc.b	$27,$EA,$C7,$DE,$F8,$59,$1F,$61,$FE,$9B,$FB,$A3,$E0,$FB,$4F,$DC
	.dc.b	$C7,$B3,$F4,$F1,$EC,$FE,$03,$1E,$CF,$D7,$C7,$9F,$C1,$74,$F8,$39
	.dc.b	$3F,$C0,$B2,$FE,$D6,$F3,$3B,$10,$C1,$CC,$EC,$43,$03,$83,$FD,$D1
	.dc.b	$1F,$DC,$8F,$F5,$FF,$57,$F7,$16,$CF,$F6,$E7,$5F,$07,$F0,$09,$1E
	.dc.b	$5F,$83,$F8,$7F,$03,$23,$D0,$FC,$6F,$7B,$E3,$FB,$B2,$7D,$8F,$CA
	.dc.b	$F7,$B9,$C7,$97,$2F,$16,$38,$CF,$CF,$97,$5A,$D1,$E5,$CB,$13,$EC
	.dc.b	$CB,$EA,$FC,$1C,$AF,$94,$FD,$8F,$83,$FB,$3F,$07,$F6,$BE,$0E,$57
	.dc.b	$CF,$8F,$BB,$F0,$32,$3E,$93,$F8,$0F,$82,$FC,$0F,$C0,$C9,$FA,$CF
	.dc.b	$E1,$3E,$0B,$AC,$7D,$9C,$3A,$BE,$8F,$21,$C8,$CB,$A4,$F6,$21,$B8
	.dc.b	$63,$F9,$0F,$6F,$F2,$3F,$05,$2B,$F0,$D1,$FE,$20,$FD,$D7,$FB,$49
	.dc.b	$23,$1F,$11,$9F,$83,$D7,$C2,$49,$FE,$9F,$5F,$D7,$DB,$F0,$9C,$C8
	.dc.b	$EB,$E0,$BB,$30,$E2,$76,$39,$DF,$57,$E0,$57,$FB,$6B,$99,$D5,$FE
	.dc.b	$3F,$8F,$04,$EB,$E0,$BD,$44,$7C,$BF,$05,$DD,$E6,$5B,$E6,$FF,$91
	.dc.b	$2B,$FC,$0C,$F8,$30,$F1,$4E,$B8,$95,$E3,$C7,$C5,$5F,$27,$E0,$E6
	.dc.b	$1F,$11,$F9,$9F,$09,$F3,$7E,$12,$4F,$C8,$7E,$87,$C2,$7D,$1F,$80
	.dc.b	$93,$F3,$1F,$A9,$F0,$1F,$57,$E0,$3E,$B7,$C0,$4A,$FA,$31,$D1,$F8
	.dc.b	$F8,$5D,$FD,$1B,$32,$EB,$5F,$33,$2B,$5A,$CB,$F0,$5F,$09,$2B,$EE
	.dc.b	$BF,$85,$F8,$4F,$C3,$7C,$27,$E1,$FE,$12,$57,$E0,$A3,$F1,$5F,$07
	.dc.b	$23,$F0,$AF,$E3,$7E,$13,$FC,$DB,$E8,$7D,$09,$FC,$E3,$73,$F9,$CD
	.dc.b	$E5,$FF,$3E,$3F,$D9,$B7,$63,$E0,$BF,$C2,$E5,$E4,$7B,$90,$FC,$4E
	.dc.b	$5E,$17,$FC,$4B,$7F,$D7,$0F,$ED,$4F,$8B,$E0,$3C,$BF,$09,$E6,$F8
	.dc.b	$49,$3C,$1F,$4F,$C2,$7A,$BE,$0F,$D7,$F0,$72,$BA,$C7,$BB,$E0,$3D
	.dc.b	$FF,$03,$27,$C3,$1D,$1F,$83,$85,$E1,$F8,$38,$F5,$9F,$3E,$3D,$5F
	.dc.b	$8F,$8F,$CA,$F8,$19,$1E,$C7,$E6,$FC,$07,$3E,$3F,$93,$95,$AF,$F5
	.dc.b	$0E,$EC,$BB,$47,$D2,$CB,$B6,$7F,$4B,$3F,$AF,$F0,$7F,$63,$E0,$E5
	.dc.b	$FC,$A8,$FB,$7F,$03,$F7,$3E,$0A,$4F,$CE,$8F,$BF,$EF,$C8,$FA,$2F
	.dc.b	$E0,$BE,$03,$F0,$7F,$01,$27,$EA,$BF,$86,$DB,$AC,$7D,$8C,$3A,$BF
	.dc.b	$88,$C3,$A3,$F8,$AC,$2E,$9F,$8A,$C7,$F1,$FF,$05,$2B,$F0,$2F,$FA
	.dc.b	$FB,$EA,$3F,$B9,$47,$F9,$A1,$FE,$CF,$5F,$BF,$6F,$17,$FB,$35,$B3
	.dc.b	$FE,$CA,$7F,$DB,$47,$FB,$50,$9F,$11,$D7,$E2,$3B,$7C,$04,$AE,$CF
	.dc.b	$87,$E0,$3C,$5F,$09,$27,$C4,$F9,$7E,$13,$94,$79,$70,$F4,$18,$7F
	.dc.b	$A2,$63,$FE,$B6,$6F,$C1,$7B,$4F,$77,$C1,$7B,$FE,$13,$E0,$7C,$24
	.dc.b	$AF,$81,$1F,$0F,$E1,$3E,$27,$C2,$49,$F8,$91,$F1,$FE,$03,$E4,$7C
	.dc.b	$0C,$9F,$91,$1D,$5F,$95,$87,$57,$CB,$87,$57,$E7,$61,$D5,$F4,$E1
	.dc.b	$F4,$FE,$12,$47,$D3,$7E,$B7,$C0,$F5,$8F,$7E,$1D,$79,$7E,$B6,$3D
	.dc.b	$67,$E1,$E3,$D6,$7E,$D6,$3F,$7F,$E1,$3F,$01,$F0,$92,$FF,$01,$1F
	.dc.b	$83,$F8,$29,$1F,$83,$7F,$0D,$F0,$68,$FF,$56,$FE,$ED,$FE,$CE,$23
	.dc.b	$93,$FB,$B3,$FD,$2B,$1F,$E9,$0F,$F5,$66,$FB,$BB,$6C,$F4,$F8,$3F
	.dc.b	$0B,$F7,$30,$B3,$F9,$3C,$38,$76,$78,$FD,$F6,$5E,$2C,$3C,$A6,$84
	.dc.b	$6B,$E1,$30,$F0,$F8,$4B,$5F,$ED,$1A,$F6,$C6,$3E,$13,$3F,$09,$27
	.dc.b	$E0,$BC,$3E,$13,$8F,$C2,$72,$F8,$49,$5F,$1A,$3A,$FC,$14,$8F,$92
	.dc.b	$F8,$3E,$0B,$A4,$7C,$8C,$2F,$07,$C8,$C7,$AC,$FD,$0C,$7A,$BF,$3B
	.dc.b	$1F,$4F,$C1,$7A,$BE,$0A,$4F,$D6,$8F,$6F,$C1,$FB,$BE,$12,$4F,$DA
	.dc.b	$8F,$83,$F0,$92,$3E,$EB,$F1,$3E,$13,$E2,$FC,$17,$C6,$F8,$29,$5F
	.dc.b	$82,$8E,$AF,$E1,$30,$EA,$FE,$0B,$0E,$8F,$E1,$30,$BB,$7E,$13,$1F
	.dc.b	$A3,$F0,$5F,$4B,$E1,$25,$FE,$3A,$3A,$BF,$8C,$C7,$AD,$FE,$33,$1E
	.dc.b	$7D,$3F,$83,$E7,$67,$8C,$3F,$03,$0C,$FF,$03,$FB,$BF,$F6,$D9,$FC
	.dc.b	$2B,$7A,$4D,$3F,$DA,$A7,$FB,$08,$FE,$CF,$97,$F8,$8D,$B8,$FE,$4D
	.dc.b	$D7,$BD,$D8,$97,$FA,$9A,$E1,$EF,$78,$BD,$D9,$39,$C3,$94,$79,$BD
	.dc.b	$EF,$3F,$C1,$FA,$3E,$12,$5E,$CF,$4C,$76,$7F,$CC,$D9,$B3,$C6,$7E
	.dc.b	$09,$E1,$A7,$BB,$E0,$FE,$1F,$C2,$49,$CE,$09,$F8,$FF,$09,$F2,$3E
	.dc.b	$12,$56,$A3,$E5,$FC,$1F,$CC,$F8,$49,$3C,$A3,$E7,$FC,$0C,$8E,$AF
	.dc.b	$D2,$F8,$2E,$B1,$E0,$C3,$AB,$EA,$C3,$AB,$E4,$C3,$AB,$EE,$C3,$ED
	.dc.b	$FC,$17,$DC,$F8,$29,$3E,$A8,$FB,$FF,$09,$23,$DA,$FE,$0B,$E1,$3F
	.dc.b	$07,$F0,$9F,$C1,$3B,$FF,$B1,$D7,$FA,$3D,$BC,$7F,$7E,$75,$F0,$3B
	.dc.b	$F8,$19,$7F,$32,$FF,$B2,$F2,$FE,$2B,$2F,$04,$FD,$47,$C0,$FC,$EC
	.dc.b	$3C,$1F,$0F,$23,$07,$3F,$82,$F2,$11,$ED,$F0,$BD,$34,$FC,$96,$17
	.dc.b	$BB,$F2,$5B,$7B,$2F,$C9,$8F,$57,$F0,$B9,$7B,$FE,$16,$47,$A5,$F8
	.dc.b	$5F,$0B,$F0,$FE,$12,$4F,$B5,$F8,$DF,$09,$F1,$FD,$F9,$3F,$05,$F9
	.dc.b	$5F,$01,$F2,$FD,$F9,$3F,$15,$F9,$DF,$01,$D2,$3E,$76,$17,$83,$FD
	.dc.b	$1D,$C8,$CB,$AD,$7D,$3C,$AD,$6B,$2F,$B1,$F0,$72,$BE,$93,$F6,$FE
	.dc.b	$13,$EE,$7C,$27,$DD,$F8,$49,$5F,$5E,$3F,$01,$F0,$72,$3E,$D3,$F8
	.dc.b	$3F,$84,$FC,$27,$C1,$7E,$17,$E0,$A5,$7D,$F8,$EA,$FE,$07,$0E,$AF
	.dc.b	$E2,$70,$E8,$FE,$2F,$0B,$A7,$BB,$1F,$EA,$7F,$41,$FD,$A6,$BF,$DB
	.dc.b	$CD,$9F,$F5,$83,$7A,$7F,$B0,$3F,$DB,$5B,$F0,$BC,$89,$5F,$BE,$9F
	.dc.b	$F5,$B3,$FD,$89,$FE,$BC,$F6,$F8,$3F,$09,$E2,$F8,$3F,$1F,$C0,$C9
	.dc.b	$E4,$F9,$BE,$07,$AC,$79,$B0,$B9,$79,$BE,$17,$9C,$F8,$71,$AF,$B3
	.dc.b	$1F,$7F,$C0,$FC,$0F,$82,$93,$E7,$8E,$7A,$FB,$23,$D5,$97,$C6,$C7
	.dc.b	$9E,$5F,$13,$2B,$7F,$13,$2F,$97,$F0,$9F,$33,$E1,$26,$5F,$0E,$3E
	.dc.b	$7F,$C1,$C8,$F8,$CF,$D2,$F8,$4F,$A7,$F0,$12,$7E,$53,$F5,$BE,$07
	.dc.b	$EB,$FC,$0C,$9F,$9C,$FD,$AF,$80,$E9,$1F,$6B,$0B,$B7,$DA,$C7,$A4
	.dc.b	$FD,$EC,$6C,$FD,$EC,$7F,$07,$F0,$7F,$84,$F8,$49,$7F,$6E,$3F,$0F
	.dc.b	$E1,$FE,$23,$E1,$24,$FD,$F8,$FF,$0E,$FE,$F1,$FD,$6C,$7C,$5E,$0C
	.dc.b	$F1,$EB,$E0,$77,$F0,$53,$0F,$E8,$B6,$9E,$F1,$F8,$BC,$7B,$BF,$96
	.dc.b	$C7,$9B,$C3,$1E,$CF,$2C,$7C,$7F,$05,$23,$AB,$E6,$F8,$5F,$3F,$C2
	.dc.b	$C9,$F0,$BE,$AF,$85,$F5,$FC,$1F,$B3,$E1,$65,$79,$A3,$DF,$F0,$32
	.dc.b	$3D,$2F,$C2,$F8,$2E,$90,$E5,$69,$FE,$49,$E1,$67,$A9,$CD,$7F,$92
	.dc.b	$25,$9F,$8A,$3E,$2E,$7E,$27,$A6,$1C,$F3,$FA,$3F,$03,$27,$E5,$BF
	.dc.b	$53,$E0,$BB,$57,$87,$3B,$5D,$F1,$EC,$F4,$3C,$58,$F7,$CF,$CB,$EF
	.dc.b	$34,$FE,$ED,$5F,$63,$E1,$3F,$02,$47,$ED,$FC,$27,$E1,$3E,$0F,$F0
	.dc.b	$BF,$09,$27,$EF,$C7,$E2,$3E,$02,$47,$E0,$9F,$C5,$FC,$07,$E3,$3D
	.dc.b	$F9,$3F,$86,$7F,$CD,$7F,$BC,$FF,$66,$E1,$F9,$98,$75,$AF,$56,$BE
	.dc.b	$7C,$BC,$51,$D4,$86,$B5,$67,$E0,$8F,$F0,$BE,$3F,$DB,$09,$F0,$BE
	.dc.b	$0F,$85,$F0,$FC,$2F,$8B,$E1,$65,$F0,$8F,$2F,$C2,$79,$BE,$12,$4F
	.dc.b	$48,$F4,$FC,$14,$8E,$EF,$B3,$E0,$BA,$C7,$8B,$0E,$AF,$B3,$36,$3D
	.dc.b	$A3,$CF,$8F,$67,$58,$FC,$6F,$82,$F8,$FF,$05,$27,$DB,$1F,$2B,$E1
	.dc.b	$24,$7C,$07,$E6,$FC,$27,$CE,$F8,$4F,$9F,$F0,$92,$BE,$2C,$7D,$2F
	.dc.b	$81,$91,$F2,$1F,$AB,$F0,$5D,$23,$D1,$85,$EE,$FC,$5E,$EC,$BB,$57
	.dc.b	$D0,$CB,$B3,$EF,$CB,$EF,$FC,$17,$E0,$3E,$0A,$4F,$D6,$8E,$9D,$9F
	.dc.b	$EB,$5C,$59,$E2,$BF,$D6,$71,$DD,$F4,$9D,$C8,$77,$3F,$03,$F0,$7F
	.dc.b	$8E,$23,$F8,$4F,$84,$FF,$5D,$9F,$D9,$5F,$F5,$FB,$FD,$95,$BF,$11
	.dc.b	$A7,$7F,$01,$C3,$E0,$24,$7F,$99,$2B,$FD,$AD,$97,$FB,$35,$FE,$CF
	.dc.b	$C1,$FB,$05,$EF,$F8,$0F,$0F,$C0,$75,$BE,$59,$79,$0C,$3F,$A2,$AE
	.dc.b	$AF,$9B,$1F,$0B,$9C,$BD,$7F,$09,$23,$C8,$FB,$BE,$13,$DF,$F0,$7F
	.dc.b	$03,$E1,$25,$7A,$63,$E1,$FC,$1F,$C4,$F8,$49,$3E,$D8,$F8,$FF,$05
	.dc.b	$F2,$3E,$0A,$4F,$C1,$8E,$AF,$C3,$C3,$AB,$E9,$C3,$AB,$F1,$F0,$EA
	.dc.b	$FB,$70,$FA,$7F,$05,$F5,$3E,$0A,$4F,$CD,$8F,$AF,$F0,$9F,$63,$E1
	.dc.b	$24,$FD,$18,$FB,$7F,$07,$F7,$3E,$12,$4F,$D5,$8F,$BF,$F0,$32,$3E
	.dc.b	$C3,$F8,$2F,$82,$EB,$1F,$6F,$0E,$AF,$D1,$C3,$AB,$F7,$F0,$EA,$FD
	.dc.b	$5C,$0F,$FA,$9B,$F7,$AF,$F6,$21,$3A,$F3,$1F,$C7,$23,$FB,$6B,$E9
	.dc.b	$3C,$04,$3C,$07,$A4,$F0,$10,$F0,$1D,$FE,$13,$AF,$67,$F3,$E9,$FE
	.dc.b	$E5,$39,$3C,$9F,$0B,$AB,$F3,$FA,$AA,$FF,$60,$57,$A7,$E1,$34,$A7
	.dc.b	$51,$87,$87,$C2,$5A,$FE,$C7,$5F,$07,$DE,$C1,$9F,$7F,$5F,$05,$27
	.dc.b	$E3,$3C,$7E,$03,$96,$52,$7C,$47,$CB,$C7,$93,$D4,$87,$CB,$BF,$17
	.dc.b	$C2,$78,$FE,$12,$4F,$D3,$8F,$37,$C2,$79,$FE,$12,$4F,$D7,$8F,$57
	.dc.b	$C2,$7A,$FE,$12,$4F,$DB,$8F,$77,$C1,$7B,$FE,$0A,$4F,$DF,$8E,$AF
	.dc.b	$E0,$70,$EA,$FD,$FC,$3A,$BF,$85,$C3,$AB,$F8,$3C,$3E,$57,$C0,$FC
	.dc.b	$BF,$84,$93,$F8,$B8,$F9,$DF,$09,$F4,$7F,$7B,$FE,$D1,$3F,$55,$BC
	.dc.b	$DF,$6E,$75,$F0,$92,$BF,$DF,$35,$FE,$CE,$1F,$09,$CF,$E0,$7A,$7C
	.dc.b	$14,$AD,$C7,$81,$E3,$F1,$1E,$13,$0F,$F7,$6B,$E9,$87,$78,$C6,$1E
	.dc.b	$7F,$82,$91,$E2,$7D,$5F,$07,$CB,$C0,$FF,$0A,$E1,$EA,$3D,$7D,$CD
	.dc.b	$7A,$4E,$B2,$BD,$FF,$9A,$BF,$5E,$5E,$EF,$85,$F8,$DF,$0B,$23,$07
	.dc.b	$C2,$F8,$4F,$95,$F0,$B2,$30,$7C,$6F,$7B,$E7,$7C,$1C,$8C,$1F,$2B
	.dc.b	$E0,$BE,$97,$C1,$C8,$C1,$C8,$F8,$97,$D5,$F8,$D8,$7D,$9C,$3A,$47
	.dc.b	$C9,$C3,$0F,$D6,$F8,$5F,$BC,$47,$EC,$FC,$2F,$E0,$7E,$13,$F0,$5F
	.dc.b	$0B,$27,$EE,$C7,$E1,$7E,$12,$47,$E0,$1F,$C4,$7C,$0F,$EE,$7F,$7C
	.dc.b	$FE,$C1,$DC,$69,$3E,$0F,$B5,$FE,$23,$1E,$CF,$DF,$C7,$AB,$F8,$1C
	.dc.b	$6F,$3F,$DA,$CB,$B7,$C1,$77,$F8,$39,$7F,$C9,$AB,$FC,$A7,$9F,$F5
	.dc.b	$CD,$3C,$5F,$09,$E5,$25,$6F,$A7,$F8,$BD,$7F,$6A,$8C,$1F,$C5,$57
	.dc.b	$C2,$7B,$7E,$12,$4E,$B0,$F7,$4F,$C1,$F8,$19,$39,$B4,$7C,$5F,$82
	.dc.b	$F0,$CF,$F9,$4A,$7F,$BD,$E3,$E2,$CB,$E5,$7C,$2F,$CB,$C7,$BD,$EF
	.dc.b	$3C,$3D,$7E,$0B,$E8,$7C,$1C,$8C,$1E,$1C,$3B,$ED,$F1,$1E,$B1,$E5
	.dc.b	$C3,$AC,$FC,$8C,$3A,$CF,$97,$E1,$3E,$E7,$C2,$C8,$F6,$47,$DF,$F8
	.dc.b	$49,$1E,$F7,$F0,$5F,$0B,$F8,$3F,$82,$93,$F0,$DF,$C3,$7C,$1F,$E1
	.dc.b	$FE,$0B,$F1,$1F,$07,$2B,$E4,$47,$47,$E3,$61,$74,$FB,$B8,$F4,$9F
	.dc.b	$95,$8D,$9F,$C0,$E3,$F9,$3F,$85,$95,$F4,$DF,$CB,$7C,$2F,$F4,$FF
	.dc.b	$33,$FD,$5B,$3F,$E8,$46,$DF,$FA,$71,$BC,$BF,$D7,$1F,$ED,$B3,$BF
	.dc.b	$85,$E5,$F0,$B2,$3F,$09,$3F,$D9,$9E,$9F,$05,$DF,$E0,$A4,$6A,$3C
	.dc.b	$58,$48,$E2,$F9,$7E,$23,$CD,$8F,$59,$F3,$63,$87,$BF,$C1,$FA,$FE
	.dc.b	$1A,$46,$0F,$1F,$C3,$7B,$C8,$F9,$BE,$1B,$E1,$7C,$2F,$C3,$F8,$69
	.dc.b	$3E,$A8,$F8,$DF,$07,$23,$DA,$FC,$9F,$84,$E7,$1F,$27,$0B,$C5,$F2
	.dc.b	$71,$EB,$3F,$13,$1E,$AF,$CF,$C7,$E9,$7C,$1F,$D3,$F8,$49,$3F,$2A
	.dc.b	$39,$F0,$FD,$1C,$AC,$FF,$82,$79,$19,$76,$8F,$B5,$97,$6C,$FE,$D6
	.dc.b	$7F,$7F,$E1,$7F,$01,$F0,$D2,$FE,$BC,$7E,$0F,$E1,$7F,$09,$F0,$D2
	.dc.b	$7E,$DC,$7E,$1F,$39,$1F,$79,$FF,$0C,$7E,$FB,$FD,$98,$7C,$06,$3E
	.dc.b	$12,$57,$F0,$79,$CB,$D0,$FC,$3E,$1D,$DF,$C9,$61,$D9,$FC,$A6,$17
	.dc.b	$4F,$F7,$7D,$F7,$F8,$6F,$09,$2B,$B7,$C3,$F9,$3E,$1B,$CB,$F0,$D2
	.dc.b	$7C,$51,$E8,$F8,$49,$1E,$57,$D7,$F0,$9E,$CF,$81,$93,$E9,$7D,$FF
	.dc.b	$05,$D6,$3D,$98,$75,$7F,$DD,$B9,$B1,$EB,$18,$C6,$E9,$EA,$CB,$E4
	.dc.b	$FC,$14,$AF,$8C,$FC,$CF,$82,$F9,$BF,$09,$27,$E5,$3F,$43,$E1,$3E
	.dc.b	$8F,$C2,$7D,$2F,$84,$95,$F3,$E3,$EA,$FC,$14,$8F,$A4,$FD,$8F,$83
	.dc.b	$E9,$1E,$8C,$2E,$9F,$3B,$1E,$B3,$F6,$71,$EA,$FB,$B1,$FC,$0F,$C1
	.dc.b	$C8,$FB,$CF,$E1,$3E,$0F,$9F,$85,$FE,$64,$E2,$67,$8B,$FF,$31,$43
	.dc.b	$2E,$F1,$F2,$B2,$EF,$9F,$E0,$33,$FF,$33,$7E,$FD,$FD,$B8,$FF,$66
	.dc.b	$7C,$11,$A6,$7E,$13,$64,$C3,$FA,$C4,$3E,$13,$9F,$BD,$27,$2F,$6F
	.dc.b	$81,$EF,$EF,$C9,$E2,$F8,$BD,$FE,$90,$D1,$C3,$FE,$82,$CB,$FB,$91
	.dc.b	$32,$ED,$5B,$9D,$65,$8B,$E3,$3F,$E8,$5B,$FE,$E8,$EC,$C7,$C2,$7C
	.dc.b	$2F,$84,$91,$EC,$8F,$8B,$F0,$7F,$1B,$E0,$E4,$FC,$08,$F9,$3F,$05
	.dc.b	$23,$E1,$BF,$33,$E0,$FE,$6F,$C0,$C9,$F8,$EF,$D0,$F8,$1E,$B1,$F2
	.dc.b	$B0,$EA,$F9,$79,$3E,$96,$3D,$63,$1A,$32,$C6,$BE,$EB,$FB,$19,$FD
	.dc.b	$CF,$81,$98,$7D,$67,$EF,$FC,$0F,$E0,$3E,$0B,$F0,$3F,$07,$2B,$ED
	.dc.b	$C7,$E1,$3E,$0F,$F0,$BF,$07,$27,$EF,$C7,$FA,$B7,$F7,$FF,$EB,$71
	.dc.b	$FE,$B1,$7F,$AD,$E9,$E3,$6F,$37,$FA,$E9,$DE,$BF,$8D,$C7,$C1,$7E
	.dc.b	$CC,$BC,$8F,$63,$F1,$B9,$78,$5F,$83,$97,$87,$E0,$64,$7F,$8A,$71
	.dc.b	$FE,$CD,$9F,$87,$E0,$79,$CF,$53,$F2,$F8,$E2,$7E,$47,$0B,$1E,$D1
	.dc.b	$8C,$BE,$15,$F5,$F8,$3F,$80,$47,$C1,$F0,$9F,$0F,$E0,$E4,$78,$DF
	.dc.b	$8D,$F0,$7F,$1F,$DF,$93,$E7,$7E,$57,$BF,$F2,$FD,$E9,$3E,$B7,$E7
	.dc.b	$7B,$FD,$23,$F0,$9C,$3F,$ED,$A6,$E3,$D9,$E8,$43,$78,$76,$7E,$37
	.dc.b	$C1,$FD,$9F,$83,$91,$83,$E5,$7C,$17,$DD,$F8,$39,$18,$3E,$77,$C1
	.dc.b	$7E,$07,$E0,$E4,$60,$FA,$5F,$01,$F8,$52,$3F,$57,$E0,$7F,$11,$87
	.dc.b	$48,$F5,$E1,$87,$91,$EE,$BE,$AF,$C0,$C3,$F2,$1F,$03,$23,$F0,$0F
	.dc.b	$FB,$FF,$97,$FD,$9A,$7F,$64,$8F,$D5,$37,$9B,$F4,$CD,$EE,$4B,$B9
	.dc.b	$3F,$EB,$68,$E3,$F0,$3D,$48,$FF,$9D,$E1,$F0,$3E,$1F,$7B,$C5,$EF
	.dc.b	$CA,$D4,$79,$7E,$0E,$47,$17,$D1,$F0,$7D,$23,$B9,$D7,$1F,$66,$1E
	.dc.b	$DF,$81,$93,$E1,$7E,$07,$C0,$F4,$8F,$7E,$17,$6F,$7E,$3D,$5F,$83
	.dc.b	$8D,$DB,$E0,$ED,$F0,$33,$F9,$5F,$07,$32,$F7,$3F,$37,$E0,$FE,$77
	.dc.b	$C1,$FC,$FF,$83,$95,$F0,$E3,$E9,$7B,$FF,$4F,$DE,$93,$F1,$E3,$EB
	.dc.b	$7B,$B2,$3E,$53,$F6,$7D,$EE,$B1,$F3,$70,$EA,$FD,$AC,$3A,$BF,$47
	.dc.b	$0E,$AF,$DE,$C3,$F0,$5F,$07,$F8,$3F,$83,$93,$F5,$E3,$F0,$DF,$07
	.dc.b	$23,$ED,$3F,$89,$F8,$3F,$C1,$C7,$F3,$1F,$E0,$0F,$F6,$89,$1C,$F2
	.dc.b	$7F,$2F,$8F,$EC,$E7,$FB,$14,$9F,$07,$AE,$4F,$9B,$1F,$D9,$6B,$B9
	.dc.b	$0F,$CA,$E1,$E1,$7F,$D8,$51,$DD,$A7,$33,$F2,$CF,$83,$DF,$C7,$2A
	.dc.b	$3F,$D8,$33,$A9,$CC,$7F,$60,$C7,$D1,$F0,$58,$25,$78,$3E,$0B,$7F
	.dc.b	$01,$C3,$E0,$64,$F9,$23,$9F,$BD,$D3,$DF,$93,$E8,$8E,$6F,$C9,$C2
	.dc.b	$E3,$F1,$F1,$EB,$3E,$EC,$7A,$BF,$3F,$1F,$3F,$BF,$E8,$F7,$E4,$FC
	.dc.b	$48,$E7,$C1,$F3,$32,$B3,$F1,$31,$EA,$FD,$9C,$7A,$E5,$F5,$F2,$F8
	.dc.b	$7F,$03,$F1,$3E,$06,$5F,$D0,$8F,$8F,$F0,$32,$3E,$9B,$F2,$BE,$07
	.dc.b	$E5,$FC,$0F,$CC,$F7,$E5,$7D,$88,$F9,$FE,$F4,$8F,$B6,$FD,$2F,$33
	.dc.b	$A4,$7E,$37,$0B,$A7,$E2,$F1,$EB,$3F,$82,$C7,$AB,$F9,$3C,$7E,$FF
	.dc.b	$F0,$0F,$F6,$A2,$70,$E3,$E1,$B3,$F0,$72,$FF,$8E,$E5,$FE,$05,$BF
	.dc.b	$EC,$B7,$FC,$01,$7C,$1F,$4F,$82,$97,$C2,$B5,$1E,$0F,$82,$F0,$FB
	.dc.b	$B2,$B4,$7F,$29,$C3,$FA,$A6,$3D,$FE,$0B,$C5,$3F,$E1,$B9,$F0,$BB
	.dc.b	$CF,$D7,$9F,$82,$35,$95,$C3,$D1,$97,$C0,$F7,$A5,$78,$3C,$06,$49
	.dc.b	$F8,$BF,$05,$FA,$7F,$88,$7E,$D7,$FE,$89,$C3,$CF,$18,$F6,$35,$26
	.dc.b	$FB,$1F,$AF,$FC,$9D,$FA,$9B,$D1,$C2,$F3,$29,$88,$FD,$B6,$4F,$51
	.dc.b	$F0,$1F,$B9,$F8,$E7,$F0,$0F,$3A,$F9,$31,$8B,$E3,$7C,$27,$CF,$AF
	.dc.b	$A5,$FF,$2C,$6D,$E9,$73,$B7,$A1,$FA,$7A,$FA,$30,$F2,$C6,$35,$F4
	.dc.b	$36,$3F,$17,$F1,$1E,$5D,$33,$3F,$E6,$5D,$3C,$FB,$6A,$3C,$7B,$FC
	.dc.b	$8C,$E4,$60,$E8,$77,$9F,$0B,$FE,$6F,$CF,$C8,$FD,$9E,$27,$C4,$7E
	.dc.b	$73,$E1,$E4,$E6,$3F,$3F,$F0,$FF,$A0,$F8,$69,$3C,$63,$F4,$7F,$0D
	.dc.b	$FA,$4F,$85,$93,$D6,$3F,$4F,$F0,$DF,$A8,$F8,$79,$3E,$18,$EA,$F8
	.dc.b	$F0,$EA,$FE,$5F,$0E,$AF,$9F,$0E,$AF,$E6,$F0,$FD,$9F,$44,$8F,$63
	.dc.b	$FB,$6F,$87,$FE,$6B,$F7,$4E,$3A,$4A,$FC,$D4,$7F,$84,$DF,$F0,$93
	.dc.b	$FE,$11,$EF,$FF,$3E,$37,$C5,$FE,$EE,$D9,$DF,$C2,$76,$F8,$4F,$04
	.dc.b	$7C,$DC,$70,$F6,$C7,$C0,$FF,$A3,$6F,$C8,$F8,$BE,$1B,$CF,$F0,$D2
	.dc.b	$3C,$AF,$AB,$1F,$5F,$4C,$9F,$4B,$EE,$E9,$F7,$FC,$44,$9F,$6B,$F0
	.dc.b	$BE,$23,$E1,$F8,$5F,$13,$E2,$25,$7C,$28,$F8,$FF,$0D,$CE,$3E,$2E
	.dc.b	$18,$79,$90,$F5,$61,$D5,$E4,$7C,$CB,$EA,$FB,$EF,$E9,$FC,$3C,$8F
	.dc.b	$A2,$FD,$6F,$86,$F4,$47,$43,$B7,$27,$D2,$CF,$D3,$1D,$48,$75,$3E
	.dc.b	$DF,$C3,$FE,$03,$E2,$24,$60,$FB,$FE,$2F,$E1,$08,$FE,$0B,$C3,$FC
	.dc.b	$3F,$C4,$48,$FC,$2B,$F8,$AF,$86,$FC,$5F,$C4,$7F,$80,$FF,$24,$FF
	.dc.b	$0B,$D7,$F3,$26,$BF,$76,$3A,$BF,$7F,$0B,$A7,$E3,$31,$ED,$3F,$83
	.dc.b	$C6,$CF,$E4,$31,$ED,$2F,$2E,$91,$F0,$0F,$89,$DF,$C6,$48,$CB,$CB
	.dc.b	$E2,$B9,$FC,$6F,$4F,$86,$95,$CA,$3B,$DC,$8E,$AF,$8B,$E1,$BA,$47
	.dc.b	$FB,$E7,$83,$FC,$F3,$8E,$34,$FF,$7F,$C7,$F9,$F3,$2E,$F5,$E6,$CB
	.dc.b	$BB,$C4,$FF,$12,$78,$7F,$E8,$2A,$FF,$41,$FC,$2F,$C0,$24,$FF,$5E
	.dc.b	$FB,$07,$C8,$3D,$9F,$F1,$A4,$63,$DE,$B3,$FD,$8B,$1F,$F4,$6C,$E6
	.dc.b	$F9,$4B,$DE,$FF,$1E,$7D,$7B,$F5,$F7,$66,$1F,$E9,$57,$FD,$BA,$75
	.dc.b	$DF,$9A,$EF,$EF,$78,$35,$97,$DA,$3C,$7A,$C8,$F0,$BE,$6C,$B9,$47
	.dc.b	$5C,$BB,$AE,$26,$DF,$EA,$6D,$3C,$33,$BC,$FC,$2F,$72,$1F,$37,$E2
	.dc.b	$BC,$0F,$3A,$F8,$7B,$7C,$58,$6F,$F1,$5D,$18,$8F,$8C,$F7,$3C,$DB
	.dc.b	$FC,$53,$D2,$FD,$BF,$8D,$FC,$BF,$C7,$48,$FC,$AC,$7E,$6B,$E3,$7F
	.dc.b	$37,$F1,$D2,$7F,$33,$1F,$9E,$F8,$D9,$1F,$9C,$7F,$43,$F1,$DF,$A2
	.dc.b	$F8,$CF,$D1,$E1,$2B,$F4,$31,$D5,$FD,$1E,$1D,$5F,$C7,$61,$D1,$FC
	.dc.b	$86,$17,$87,$F2,$F0,$C7,$F5,$FF,$19,$23,$F5,$AF,$EC,$BE,$2F,$F6
	.dc.b	$7F,$19,$FB,$4F,$8E,$95,$FB,$28,$FD,$BF,$C7,$48,$FD,$AB,$FC,$1F
	.dc.b	$EC,$BE,$DC,$7E,$E2,$FA,$34,$FE,$15,$0F,$8F,$EF,$1F,$BE,$C3,$BB
	.dc.b	$FA,$9C,$3B,$3F,$AB,$C2,$E7,$FB,$2C,$3B,$CF,$ED,$DF,$F2,$0F,$79
	.dc.b	$FF,$28,$C7,$83,$E3,$BC,$DF,$1F,$E7,$EF,$F4,$78,$12,$BC,$B1,$EB
	.dc.b	$F0,$24,$7A,$1F,$77,$81,$EF,$F8,$EE,$73,$EC,$C3,$0F,$23,$DF,$87
	.dc.b	$57,$CD,$87,$47,$D1,$85,$D3,$D9,$8F,$29,$F6,$61,$D9,$F7,$63,$E9
	.dc.b	$7A,$90,$EA,$79,$C8,$76,$3E,$8C,$FD,$7F,$8E,$91,$F5,$1F,$B5,$F1
	.dc.b	$9F,$6F,$E2,$24,$FD,$87,$EF,$7C,$4F,$DF,$F8,$69,$3F,$71,$FC,$17
	.dc.b	$C3,$F2,$8E,$84,$3E,$B5,$F4,$7E,$C5,$DD,$3E,$E6,$1D,$27,$EE,$61
	.dc.b	$67,$F0,$18,$7E,$3F,$E1,$BF,$21,$F0,$F2,$FF,$1B,$1F,$93,$F8,$89
	.dc.b	$1F,$90,$7F,$2D,$F1,$3F,$97,$F8,$9F,$CC,$7C,$44,$AF,$CA,$C7,$F6
	.dc.b	$1C,$BF,$C3,$31,$FD,$0D,$FF,$15,$74,$7F,$19,$F8,$6D,$7C,$46,$FE
	.dc.b	$1E,$5F,$F6,$7B,$FF,$15,$5F,$F1,$FF,$87,$E9,$F1,$32,$B4,$F7,$F8
	.dc.b	$8F,$07,$C3,$F8,$7E,$1E,$57,$38,$F2,$7A,$72,$3B,$3E,$7F,$3F,$D1
	.dc.b	$F0,$F2,$7C,$4F,$AF,$E2,$3A,$C7,$32,$1E,$DF,$88,$E8,$FB,$EE,$ED
	.dc.b	$FD,$80,$FE,$72,$5B,$ED,$3F,$15,$EC,$FC,$1F,$88,$F9,$44,$7E,$27
	.dc.b	$C4,$7C,$DF,$88,$F9,$DF,$0F,$27,$E4,$47,$D1,$F8,$79,$1F,$2D,$FA
	.dc.b	$9F,$0F,$F5,$7E,$1E,$4F,$CF,$7E,$C7,$C4,$7F,$86,$7E,$2F,$FC,$33
	.dc.b	$AF,$F7,$86,$FB,$DF,$DC,$DB,$6E,$84,$7F,$84,$E3,$FC,$40,$FF,$84
	.dc.b	$5B,$F0,$BD,$4E,$DF,$0F,$DF,$E1,$A4,$F0,$7C,$5F,$11,$E3,$F8,$79
	.dc.b	$3D,$1F,$37,$C4,$79,$FE,$23,$D1,$F0,$D2,$BC,$31,$EB,$F8,$3F,$67
	.dc.b	$C2,$49,$F2,$C7,$BF,$E0,$BE,$07,$C1,$C9,$F4,$C7,$C3,$F8,$39,$1E
	.dc.b	$C7,$E3,$7C,$1F,$C7,$F8,$49,$3F,$01,$F9,$5F,$09,$F2,$FE,$12,$4F
	.dc.b	$C4,$7E,$77,$C1,$7C,$FF,$83,$FA,$1F,$07,$2B,$E4,$C7,$F9,$2F,$CE
	.dc.b	$FE,$6F,$D7,$FE,$2A,$6F,$5F,$F8,$95,$B5,$F4,$DF,$E8,$CF,$F9,$45
	.dc.b	$FE,$7A,$4F,$7B,$A7,$C1,$75,$F8,$1E,$DF,$05,$2F,$84,$78,$7E,$0A
	.dc.b	$47,$37,$C9,$F0,$3E,$5F,$82,$F3,$7C,$14,$AF,$04,$7A,$7E,$0B,$D5
	.dc.b	$F0,$52,$7C,$91,$ED,$F8,$29,$1E,$77,$E0,$7C,$17,$C1,$F8,$2F,$85
	.dc.b	$F0,$32,$BD,$91,$F1,$7E,$0B,$E3,$7C,$14,$9F,$81,$1F,$27,$E0,$A4
	.dc.b	$7C,$37,$E6,$7C,$0F,$CD,$F8,$29,$3F,$1D,$FF,$3D,$7C,$3F,$F1,$AE
	.dc.b	$9F,$F2,$CB,$FC,$68,$FE,$2A,$4F,$83,$DF,$C1,$70,$F8,$29,$87,$FA
	.dc.b	$06,$93,$D3,$E0,$BA,$FC,$0C,$AD,$47,$83,$E0,$BC,$3E,$FC,$9E,$51
	.dc.b	$E4,$F8,$29,$1D,$5F,$3F,$C0,$7A,$3E,$0A,$4F,$85,$F5,$FC,$17,$B3
	.dc.b	$E0,$BD,$BF,$05,$2B,$CD,$1F,$03,$E0,$64,$7A,$5F,$87,$F0,$5F,$13
	.dc.b	$E0,$A4,$FB,$5F,$8F,$F0,$5F,$23,$E0,$BE,$4F,$C1,$4A,$F8,$51,$FE
	.dc.b	$A5,$FC,$AB,$F8,$4F,$C1,$9A,$4F,$8A,$D7,$C1,$6F,$E0,$E6,$1F,$EA
	.dc.b	$BA,$FE,$18,$FF,$A3,$D7,$C1,$F5,$F8,$39,$3B,$8F,$07,$C1,$F8,$7E
	.dc.b	$0E,$4F,$38,$F2,$7C,$14,$8E,$CF,$9F,$E0,$FD,$1F,$05,$E9,$F8,$39
	.dc.b	$5E,$38,$F6,$7C,$14,$8F,$33,$EF,$F8,$34,$4F,$1B,$FC,$7F,$F2,$0F
	.dc.b	$DE,$FF,$D3,$F2,$8C,$6D,$C2,$3F,$CE,$31,$87,$97,$BB,$D2,$39,$70
	.dc.b	$7C,$DF,$85,$F9,$DF,$0B,$2F,$E4,$47,$D1,$F8,$59,$1F,$2D,$FA,$9F
	.dc.b	$0B,$F5,$7E,$17,$EB,$7C,$2C,$AF,$A1,$1F,$67,$E1,$24,$7D,$37,$EE
	.dc.b	$7C,$2F,$FB,$87,$F2,$BF,$E2,$7D,$C7,$F8,$9D,$7F,$B1,$5B,$C5,$FE
	.dc.b	$C4,$76,$7F,$94,$89,$F0,$BC,$FE,$17,$A7,$C2,$75,$F8,$59,$87,$FB
	.dc.b	$B6,$D3,$E1,$F8,$59,$39,$7C,$9F,$0B,$FE,$BC,$F9,$D7,$83,$49,$5F
	.dc.b	$EB,$B8,$78,$1E,$BF,$85,$93,$DD,$F7,$7C,$37,$BF,$E1,$64,$F8,$DF
	.dc.b	$85,$F0,$DF,$0F,$E1,$BE,$D7,$17,$E0,$3C,$F7,$8D,$8F,$03,$ED,$46
	.dc.b	$31,$F6,$5F,$CB,$F8,$69,$1E,$F7,$E7,$7C,$2F,$CF,$F8,$6F,$A1,$F0
	.dc.b	$B2,$BE,$24,$75,$7F,$D9,$1F,$80,$7C,$67,$F1,$26,$3A,$35,$AA,$7F
	.dc.b	$DD,$53,$A8,$C7,$3F,$DD,$F8,$7F,$BD,$F0,$F2,$BE,$94,$7E,$07,$E1
	.dc.b	$E4,$7D,$57,$F0,$9F,$0F,$E3,$F2,$5F,$BF,$BE,$0F,$0D,$BF,$2C,$EA
	.dc.b	$7F,$DE,$7F,$3F,$FE,$96,$E5,$FE,$68,$7E,$B3,$C3,$FF,$71,$FC,$4F
	.dc.b	$BD,$B8,$7B,$9F,$F7,$3F,$C5,$66,$BF,$C9,$A7,$F3,$49,$D3,$5F,$F6
	.dc.b	$FC,$FF,$BB,$BE,$33,$E2,$FE,$A0,$F7,$BC,$7E,$3C,$BF,$8D,$39,$79
	.dc.b	$FC,$67,$7D,$64,$60,$EF,$F1,$5E,$3F,$89,$91,$83,$EA,$8F,$D1,$59
	.dc.b	$F8,$B3,$93,$A9,$F5,$FC,$5F,$B7,$E2,$E4,$FB,$23,$E0,$7C,$5F,$C1
	.dc.b	$F8,$B9,$3F,$02,$3F,$0C,$F7,$C6,$E1,$E0,$8E,$D1,$AF,$C6,$9E,$47
	.dc.b	$CB,$F8,$CF,$99,$F1,$72,$FE,$5C,$7C,$FF,$8C,$91,$F3,$9F,$A5,$F1
	.dc.b	$9C,$EF,$BE,$32,$BE,$0E,$3D,$9F,$AB,$E5,$7D,$92,$3F,$63,$E3,$3E
	.dc.b	$E7,$0F,$DD,$F8,$B9,$3F,$72,$3B,$3E,$A2,$0F,$17,$FD,$C7,$87,$78
	.dc.b	$FC,$0F,$5B,$E3,$3F,$10,$48,$FC,$67,$D9,$FF,$5A,$3F,$CC,$0F,$E2
	.dc.b	$1B,$7F,$EF,$CE,$D4,$62,$FF,$30,$FF,$5A,$9F,$0E,$DF,$06,$FC,$95
	.dc.b	$FD,$07,$8F,$E9,$46,$0F,$E9,$79,$79,$EF,$B6,$0F,$8B,$FD,$13,$FD
	.dc.b	$92,$BC,$FF,$09,$89,$CF,$C2,$6B,$E2,$A5,$7B,$23,$E3,$BD,$CF,$C2
	.dc.b	$1F,$23,$0F,$3B,$F0,$BD,$9E,$E4,$7E,$2F,$B1,$E2,$F5,$64,$7C,$87
	.dc.b	$CB,$F1,$3D,$AD,$BF,$BF,$9F,$E1,$47,$6B,$F9,$9F,$15,$EC,$23,$F4
	.dc.b	$BE,$2B,$DF,$F1,$5F,$03,$E2,$24,$FD,$68,$ED,$8F,$58,$F5,$3F,$DB
	.dc.b	$78,$FE,$53,$F5,$3E,$27,$E4,$92,$7E,$FF,$C4,$FC,$CF,$89,$91,$F8
	.dc.b	$27,$E7,$FC,$47,$68,$FE,$FF,$D1,$F6,$36,$F6,$C6,$37,$D7,$57,$E1
	.dc.b	$3E,$27,$EB,$FC,$4C,$AF,$C6,$C7,$DA,$F8,$99,$1F,$90,$7F,$8F,$7D
	.dc.b	$DF,$FA,$02,$1C,$1E,$A7,$B1,$EE,$3F,$0F,$DD,$C3,$E4,$E4,$7F,$25
	.dc.b	$AF,$EE,$7C,$5F,$CA,$D7,$C5,$F6,$F4,$E5,$6D,$F0,$FB,$BE,$F8,$FC
	.dc.b	$29,$F8,$7E,$1F,$81,$87,$CD,$CB,$CF,$EF,$CA,$F0,$3E,$AF,$8A,$F5
	.dc.b	$FC,$57,$B3,$E2,$65,$79,$63,$B6,$3D,$23,$DA,$7E,$57,$7F,$81,$C6
	.dc.b	$DD,$F1,$3F,$1B,$49,$7E,$F8,$F9,$3F,$13,$23,$E1,$3F,$33,$E2,$7B
	.dc.b	$F3,$FE,$4B,$83,$D5,$5F,$1F,$E2,$7E,$8F,$C4,$C8,$C1,$F2,$FE,$23
	.dc.b	$EA,$91,$F9,$DF,$13,$D1,$FA,$87,$D8,$D3,$ED,$FC,$3F,$DC,$F8,$89
	.dc.b	$3F,$56,$3E,$FF,$C4,$48,$FB,$0F,$E0,$BE,$23,$A4,$7D,$1D,$7F,$1B
	.dc.b	$F9,$77,$F6,$68,$FF,$7B,$37,$8C,$D1,$F0,$DF,$DA,$5F,$9D,$1A,$F9
	.dc.b	$1C,$19,$BE,$FE,$1F,$E7,$30,$FC,$0B,$FB,$73,$F9,$05,$BE,$17,$A1
	.dc.b	$27,$FB,$A4,$7F,$7F,$7F,$B7,$AF,$84,$F1,$12,$37,$F0,$9E,$0E,$5F
	.dc.b	$DD,$47,$E0,$78,$7D,$1F,$07,$E9,$F8,$49,$3D,$A3,$D9,$F0,$72,$3C
	.dc.b	$2F,$BF,$E1,$3A,$47,$E3,$35,$F5,$63,$FE,$0D,$D1,$F0,$7F,$14,$95
	.dc.b	$EA,$F8,$4F,$91,$F0,$92,$3D,$AF,$CB,$F8,$4E,$B1,$F1,$B0,$FC,$5B
	.dc.b	$F0,$3E,$13,$E8,$11,$F8,$BF,$09,$F4,$FE,$0F,$EA,$7C,$24,$9F,$93
	.dc.b	$1D,$3C,$7F,$D2,$47,$E2,$CF,$9B,$F0,$9F,$6C,$8F,$F8,$8F,$E4,$7F
	.dc.b	$D6,$27,$F5,$35,$D6,$8C,$35,$FF,$16,$1F,$D5,$F3,$F4,$F0,$FE,$07
	.dc.b	$0F,$09,$E2,$F2,$7F,$67,$5F,$D4,$F0,$F0,$7C,$27,$23,$9F,$C2,$74
	.dc.b	$F8,$39,$3E,$57,$BF,$C2,$7B,$7B,$FF,$93,$E1,$FE,$78,$AF,$57,$C2
	.dc.b	$79,$0F,$2F,$C2,$79,$BE,$13,$CF,$F0,$92,$BE,$04,$73,$7F,$C1,$BE
	.dc.b	$07,$B3,$E1,$3D,$BF,$09,$27,$E3,$47,$C0,$F8,$3F,$83,$F0,$92,$7E
	.dc.b	$54,$73,$7F,$24,$7C,$DF,$84,$F8,$E4,$7E,$87,$C2,$7C,$AF,$84,$F9
	.dc.b	$7F,$09,$27,$EA,$47,$3E,$8F,$C8,$C7,$D4,$E1,$7C,$1F,$D2,$24,$7D
	.dc.b	$BF,$84,$FA,$BF,$09,$23,$EF,$3F,$E6,$3F,$CB,$BF,$A3,$75,$FF,$45
	.dc.b	$9F,$C9,$1F,$9F,$ED,$FB,$67,$F9,$AE,$7F,$A1,$C7,$E1,$1F,$E8,$6F
	.dc.b	$F9,$AD,$AF,$F9,$99,$DF,$71,$7C,$1F,$84,$93,$C7,$E0,$FD,$D7,$E0
	.dc.b	$8E,$DF,$07,$E7,$F8,$39,$1D,$E3,$D5,$F0,$52,$3C,$4F,$B7,$E0,$F9
	.dc.b	$C7,$F5,$E3,$FD,$3B,$F2,$6F,$D8,$D5,$FE,$A3,$8C,$55,$93,$67,$E4
	.dc.b	$7C,$27,$C9,$ED,$98,$7C,$18,$EE,$FE,$BC,$FF,$1A,$61,$DE,$B9,$47
	.dc.b	$8B,$0F,$0B,$FE,$AB,$FD,$4F,$FD,$1D,$CD,$FE,$45,$E0,$7C,$2F,$D7
	.dc.b	$25,$7C,$FF,$85,$EF,$C9,$FD,$F6,$DC,$AF,$3B,$FD,$73,$97,$F2,$B9
	.dc.b	$FF,$07,$87,$1C,$6C,$7F,$CC,$17,$C5,$FB,$9F,$0B,$F8,$72,$3F,$7F
	.dc.b	$E1,$7C,$3C,$5F,$E8,$4F,$7B,$FD,$8F,$7C,$6B,$FC,$A9,$97,$93,$3D
	.dc.b	$47,$E1,$FE,$17,$F2,$44,$7F,$15,$F0,$BD,$1F,$F4,$D6,$1C,$1F,$F0
	.dc.b	$53,$FE,$26,$E1,$7C,$27,$E6,$FE,$16,$57,$E4,$E3,$93,$FB,$23,$89
	.dc.b	$D0,$FD,$F9,$F9,$9F,$85,$FD,$21,$1F,$CE,$7C,$2F,$2D,$3F,$63,$1F
	.dc.b	$EB,$1C,$7C,$91,$FB,$4E,$F7,$E6,$1F,$DA,$CF,$ED,$35,$FE,$CF,$6B
	.dc.b	$FE,$CA,$7F,$B9,$3F,$87,$9F,$E5,$1E,$87,$F8,$36,$3F,$30,$7A,$63
	.dc.b	$F9,$65,$F8,$39,$3F,$4D,$87,$82,$FF,$DB,$D8,$7F,$76,$25,$7A,$6B
	.dc.b	$F5,$98,$7A,$4F,$84,$75,$7F,$DD,$13,$FD,$D6,$3D,$1F,$0B,$D0,$EB
	.dc.b	$F1,$7C,$E3,$CB,$97,$91,$E6,$FC,$3A,$FE,$F1,$1F,$07,$E1,$7C,$A7
	.dc.b	$9B,$E1,$B9,$47,$C1,$C7,$FD,$39,$97,$B6,$3A,$C7,$93,$E1,$7D,$BF
	.dc.b	$0D,$27,$E6,$47,$2E,$CF,$E3,$5E,$DF,$EA,$AF,$9D,$7D,$E3,$F6,$7F
	.dc.b	$81,$BD,$2F,$E0,$0D,$8F,$E0,$AD,$7F,$D3,$DC,$AF,$7F,$B7,$6F,$96
	.dc.b	$7E,$51,$F2,$4F,$D5,$7E,$93,$FA,$1E,$2F,$F1,$17,$DE,$7C,$B9,$FF
	.dc.b	$54,$36,$3F,$C4,$D3,$BC,$FE,$AE,$9F,$DA,$39,$BF,$CD,$5B,$7B,$38
	.dc.b	$7C,$0F,$F3,$72,$79,$1F,$EF,$13,$7A,$7E,$27,$4F,$E7,$67,$CD,$3E
	.dc.b	$59,$FE,$92,$AE,$71,$DA,$79,$4F,$C7,$CF,$F3,$F8,$F8,$E3,$19,$79
	.dc.b	$B0,$E3,$18,$3F,$DA,$37,$C9,$F8,$87,$84,$F1,$D7,$F8,$0F,$0F,$F3
	.dc.b	$CD,$72,$9D,$E3,$FE,$76,$7F,$DF,$F7,$EC,$7F,$CC,$B7,$C5,$F0,$1F
	.dc.b	$4B,$1F,$C9,$3C,$B0,$F1,$47,$FB,$EA,$B9,$4F,$F9,$8C,$FF,$37,$DF
	.dc.b	$CB,$AF,$B9,$7F,$89,$9F,$8E,$7A,$67,$C8,$FF,$A3,$6B,$9D,$FE,$E2
	.dc.b	$39,$DF,$B1,$FD,$1E,$3F,$B0,$BF,$AF,$1E,$C3,$D2,$FE,$63,$2F,$B0
	.dc.b	$F8,$E3,$89,$F5,$CD,$D7,$B2,$BF,$DA,$D8,$78,$9F,$F3,$0E,$1E,$3A
	.dc.b	$F8,$47,$23,$E3,$47,$8E,$B9,$47,$D8,$3A,$9F,$3B,$0E,$8F,$84,$FE
	.dc.b	$52,$74,$3E,$11,$EB,$31,$5E,$12,$57,$78,$F1,$63,$F3,$5F,$99,$7E
	.dc.b	$27,$FD,$AF,$87,$81,$FC,$3E,$16,$B8,$1C,$71,$F8,$D3,$E2,$26,$3F
	.dc.b	$CD,$A3,$E6,$47,$D4,$3F,$0F,$3C,$B2,$ED,$3F,$72,$BF,$0D,$5F,$08
	.dc.b	$FC,$31,$1E,$98,$FE,$1A,$B8,$E3,$E9,$BF,$1D,$7C,$0B,$FF,$15,$1F
	.dc.b	$8A,$8F,$DD,$3F,$4D,$DE,$5F,$6A,$3E,$54,$F6,$7E,$91,$E3,$7F,$3A
	.dc.b	$48,$F9,$91,$F0,$1E,$4F,$D2,$7F,$06,$47,$C5,$87,$A5,$E9,$1C,$27
	.dc.b	$E0,$C7,$13,$D2,$7F,$0A,$AF,$3B,$F8,$38,$F1,$3D,$0F,$4C,$F3,$3C
	.dc.b	$B1,$F0,$E3,$95,$7E,$95,$FC,$5C,$7B,$DF,$51,$F9,$33,$C4,$7E,$08
	.dc.b	$F3,$9E,$19,$FE,$7A,$EA,$BE,$D4,$78,$63,$C8,$7D,$97,$E1,$3F,$90
	.dc.b	$3F,$27,$87,$DA,$7D,$27,$0A,$F4,$63,$FC,$3E,$3F,$8D,$BF,$89,$7C
	.dc.b	$67,$E6,$0E,$27,$53,$C0,$7C,$AB,$F9,$2F,$EC,$8F,$51,$DE,$FD,$0F
	.dc.b	$33,$E0,$CF,$F3,$16,$DF,$F2,$7B,$FC,$F6,$3F,$91,$C7,$FA,$AD,$7B
	.dc.b	$0F,$2E,$3F,$11,$F9,$95,$F3,$CF,$6E,$5F,$AD,$C3,$A6,$5F,$C5,$F1
	.dc.b	$E9,$7F,$A8,$AF,$E2,$B8,$FD,$9C,$3F,$03,$3E,$F7,$F1,$24,$BF,$C0
	.dc.b	$9F,$43,$2F,$95,$1F,$D3,$31,$EF,$5F,$A5,$C7,$BB,$F3,$B1,$EE,$FE
	.dc.b	$E2,$76,$FF,$33,$9D,$D7,$F5,$6A,$DC,$FD,$6A,$DE,$5F,$CC,$23,$F1
	.dc.b	$A7,$9F,$1F,$25,$7C,$18,$FD,$A6,$5F,$D8,$DF,$B7,$5F,$63,$2F,$A5
	.dc.b	$8F,$C9,$C3,$F2,$2F,$C2,$9F,$BA,$7C,$FA,$FB,$51,$B7,$F5,$75,$F0
	.dc.b	$AB,$FB,$54,$78,$0F,$E8,$79,$7B,$23,$FC,$07,$8F,$B2,$FE,$45,$6C
	.dc.b	$FE,$CF,$5B,$7F,$AB,$DF,$9A,$FE,$8C,$7E,$38,$FC,$81,$F0,$4F,$C1
	.dc.b	$BF,$90,$C3,$DD,$87,$78,$F2,$CF,$ED,$9E,$6F,$B1,$E1,$8F,$E4,$23
	.dc.b	$47,$7A,$F8,$78,$79,$63,$FC,$05,$87,$95,$FD,$01,$E5,$7F,$9E,$57
	.dc.b	$D1,$8F,$AF,$5E,$47,$F9,$D1,$F7,$4F,$85,$8F,$E8,$9F,$AF,$3E,$38
	.dc.b	$FC,$AD,$74,$9F,$9A,$68,$FC,$E1,$D2,$3E,$6B,$EA,$3E,$31,$EF,$3E
	.dc.b	$39,$E8,$3F,$8A,$1F,$B4,$3C,$78,$7E,$A5,$EC,$7E,$6C,$EC,$FC,$C8
	.dc.b	$FC,$45,$7B,$DF,$D4,$4F,$5A,$F8,$CF,$89,$FE,$20,$FF,$1F,$C3,$E6
	.dc.b	$61,$FD,$7A,$3F,$87,$9B,$CB,$F4,$38,$7F,$55,$AF,$ED,$59,$7D,$DA
	.dc.b	$FC,$3D,$69,$FC,$FC,$E9,$F3,$9A,$3D,$A7,$ED,$CF,$6E,$3F,$B1,$76
	.dc.b	$7D,$E9,$E8,$FD,$A9,$E8,$F9,$B0,$FD,$64,$EA,$BE,$0C,$FC,$F3,$E0
	.dc.b	$5F,$9D,$E8,$78,$8F,$E0,$47,$8B,$2F,$D3,$BE,$2B,$FE,$DE,$FD,$B7
	.dc.b	$F2,$6E,$CF,$D3,$CF,$58,$DD,$7D,$49,$FC,$39,$F9,$2C,$7F,$19,$7F
	.dc.b	$C0,$09,$1F,$83,$AF,$E0,$04,$9D,$C7,$E7,$70,$FE,$9C,$F3,$3F,$22
	.dc.b	$7C,$38,$FE,$59,$5D,$23,$F2,$35,$D3,$1F,$93,$87,$F6,$B8,$F9,$33
	.dc.b	$F0,$DF,$E3,$D1,$F2,$9F,$D4,$47,$53,$F9,$51,$BB,$FC,$23,$FD,$6B
	.dc.b	$2F,$C7,$CF,$48,$F9,$13,$F9,$5C,$3F,$7F,$1E,$8B,$F1,$4F,$A2,$3F
	.dc.b	$80,$3B,$7F,$4E,$7D,$A3,$EA,$3F,$E1,$17,$F3,$47,$D3,$3F,$82,$E5
	.dc.b	$F4,$E3,$53,$FC,$A6,$7F,$09,$7F,$35,$F1,$5F,$F7,$C8,$F4,$4F,$D9
	.dc.b	$9F,$E4,$A7,$58,$FD,$34,$7A,$23,$E1,$9E,$E3,$EF,$BF,$26,$BE,$CC
	.dc.b	$7F,$28,$C3,$F8,$53,$D0,$FB,$47,$EF,$4F,$9C,$7F,$15,$3A,$1B,$9E
	.dc.b	$8F,$A0,$FF,$02,$4F,$C9,$7F,$9A,$D7,$E5,$9E,$B3,$FA,$D7,$AB,$FC
	.dc.b	$58,$F7,$1F,$C8,$5D,$DF,$E8,$9D,$BF,$C3,$CF,$C6,$4F,$59,$F1,$CF
	.dc.b	$A5,$F8,$8F,$A5,$EB,$3F,$C6,$E3,$A9,$FA,$83,$ED,$9F,$E0,$F3,$ED
	.dc.b	$9A,$3E,$D9,$A3,$F9,$65,$FC,$A9,$F4,$1E,$23,$DC,$78,$8F,$74,$FA
	.dc.b	$2F,$F1,$95,$F8,$27,$EA,$D7,$F2,$4A,$FF,$25,$48,$82,$D5,$88,$91
	.dc.b	$87,$07,$17,$0E,$1A,$F0,$8D,$BB,$6B,$DD,$B1,$C2,$3C,$6D,$78,$1C
	.dc.b	$9C,$3C,$4C,$47,$22,$38,$21,$E5,$72,$D3,$27,$9C,$A7,$F7,$A8,$F6
	.dc.b	$CA,$FC,$BE,$2F,$FA,$92,$FE,$4C,$E6,$FE,$66,$0F,$90,$D4,$7C,$F9
	.dc.b	$D2,$97,$C5,$EA,$13,$AC,$98,$E6,$38,$DF,$5A,$FF,$2C,$E3,$E2,$CB
	.dc.b	$ED,$CF,$49,$C9,$E1,$8F,$4B,$83,$BB,$C3,$2E,$CF,$2C,$AC,$FD,$FC
	.dc.b	$6B,$E9,$AB,$7D,$E3,$FD,$1D,$87,$8D,$E0,$77,$C7,$D3,$5C,$30,$E7
	.dc.b	$97,$BA,$B5,$97,$8D,$A7,$CC,$9D,$E5,$E0,$BE,$53,$A3,$F3,$51,$97
	.dc.b	$19,$79,$71,$FE,$4D,$8F,$DC,$C7,$E1,$D7,$A0,$F8,$C7,$DA,$C3,$AB
	.dc.b	$93,$E3,$9E,$C3,$15,$EC,$7B,$E3,$E6,$7C,$33,$CA,$7C,$39,$7C,$59
	.dc.b	$F3,$9A,$38,$9E,$BC,$3E,$23,$DB,$0F,$88,$D3,$E2,$1B,$CB,$E0,$39
	.dc.b	$3F,$1D,$8F,$8D,$C6,$5F,$89,$6C,$67,$1F,$2C,$7C,$23,$EF,$1E,$43
	.dc.b	$A6,$1E,$48,$DC,$7C,$43,$67,$43,$C8,$7B,$4C,$1C,$AF,$CF,$1C,$8F
	.dc.b	$CE,$D7,$8F,$0F,$D9,$63,$F1,$F1,$F4,$56,$E7,$F2,$D8,$76,$3D,$B9
	.dc.b	$7C,$F8,$E3,$1E,$BC,$BD,$11,$88,$F4,$5F,$48,$C5,$F8,$A7,$A3,$DD
	.dc.b	$FC,$21,$F0,$DC,$3F,$68,$F0,$99,$8E,$67,$E3,$A3,$73,$DE,$7F,$C4
	.dc.b	$98,$7B,$71,$F8,$11,$83,$E9,$4E,$DF,$F0,$FC,$ED,$F2,$DF,$67,$CA
	.dc.b	$7C,$EB,$F2,$E1,$F0,$23,$A1,$F0,$31,$F8,$CE,$27,$F8,$86,$5F,$52
	.dc.b	$FA,$4F,$E2,$8E,$E6,$DF,$C9,$CF,$B3,$2F,$4C,$F9,$8F,$4D,$F6,$70
	.dc.b	$7E,$1F,$0F,$23,$8A,$FD,$53,$E9,$BF,$0C,$E8,$FF,$64,$E3,$F8,$3B
	.dc.b	$E9,$18,$AF,$BD,$8F,$A5,$F3,$1F,$A6,$3F,$20,$7E,$4F,$0F,$43,$EC
	.dc.b	$3C,$58,$7B,$0F,$83,$97,$F0,$97,$C7,$5D,$67,$67,$5A,$F9,$F1,$9B
	.dc.b	$F2,$4F,$E3,$B2,$F8,$B1,$8A,$F7,$9E,$29,$F2,$9F,$8A,$3E,$11,$EA
	.dc.b	$3F,$1C,$FE,$2B,$1F,$83,$97,$F0,$BA,$FF,$76,$E1,$E1,$8C,$12,$FF
	.dc.b	$17,$87,$8A,$31,$1F,$4D,$F9,$8F,$E6,$4F,$05,$7D,$18,$DE,$1F,$D5
	.dc.b	$5D,$CF,$98,$F8,$67,$9B,$1F,$93,$8F,$9A,$3F,$65,$3C,$A7,$F4,$86
	.dc.b	$A3,$C0,$72,$BF,$03,$FE,$71,$AE,$4F,$9C,$E6,$7E,$00,$FB,$93,$F6
	.dc.b	$30,$F7,$47,$13,$AD,$7D,$19,$F0,$C7,$CD,$CB,$F5,$D3,$98,$FB,$93
	.dc.b	$F8,$07,$A9,$BA,$EA,$F8,$C9,$8F,$DC,$9F,$14,$7E,$36,$BA,$3F,$C9
	.dc.b	$A7,$F0,$07,$E2,$1E,$A7,$CA,$BF,$C8,$3F,$4B,$2F,$B8,$FF,$A4,$F0
	.dc.b	$F0,$D7,$FA,$06,$BD,$39,$7C,$DA,$FF,$14,$9F,$91,$3C,$4E,$5F,$F1
	.dc.b	$51,$FC,$0C,$F8,$AE,$70,$FE,$CD,$1F,$D7,$EB,$F2,$0E,$8F,$A2,$7C
	.dc.b	$0C,$7E,$5B,$E9,$CB,$EB,$CF,$F7,$0C,$BD,$33,$B3,$CB,$1F,$CB,$DF
	.dc.b	$D0,$BD,$9F,$C6,$5F,$C8,$3F,$06,$F1,$3D,$31,$F3,$DF,$A8,$7B,$8F
	.dc.b	$C0,$1C,$B0,$F6,$1F,$A7,$AF,$1C,$FC,$9C,$BE,$5E,$3F,$E5,$2A,$F8
	.dc.b	$D9,$7C,$CC,$3E,$81,$D4,$F1,$63,$F1,$9F,$B9,$5F,$80,$7D,$6F,$E7
	.dc.b	$5F,$11,$F2,$30,$FD,$53,$F9,$93,$F8,$03,$F8,$48,$FC,$84,$61,$E9
	.dc.b	$1F,$87,$8C,$1F,$00,$EE,$7E,$6F,$2F,$A3,$3F,$E5,$BC,$BD,$EF,$B6
	.dc.b	$7C,$4F,$F6,$A7,$F9,$A4,$7C,$57,$C9,$5F,$E5,$8E,$17,$C1,$72,$9F
	.dc.b	$B3,$7C,$DF,$C7,$1F,$E0,$39,$DB,$FE,$0B,$6B,$99,$FF,$11,$7D,$F3
	.dc.b	$DC,$E9,$3F,$72,$FA,$61,$F8,$53,$FC,$DC,$3F,$7B,$7B,$5F,$E4,$F8
	.dc.b	$C7,$45,$93,$67,$AE,$DF,$5C,$FB,$1B,$FE,$BB,$CF,$7D,$73,$D9,$FF
	.dc.b	$2D,$46,$1B,$19,$D3,$FC,$8F,$E1,$BC,$17,$C7,$76,$E6,$FC,$4F,$3F
	.dc.b	$F9,$8E,$B5,$F1,$D6,$CD,$31,$F1,$BC,$71,$FA,$A7,$F1,$21,$5D,$FE
	.dc.b	$37,$97,$C6,$C9,$F1,$E0,$9E,$51,$FA,$7D,$FF,$85,$DF,$9F,$E3,$7C
	.dc.b	$5F,$1B,$2B,$8C,$70,$E6,$FC,$AC,$7D,$C7,$FD,$3D,$5E,$FF,$47,$D4
	.dc.b	$49,$F0,$FC,$6F,$1E,$9F,$B2,$7E,$36,$F8,$E8,$F1,$7E,$17,$6C,$C3
	.dc.b	$D3,$1C,$5F,$C6,$9F,$07,$27,$D0,$7F,$B5,$70,$F1,$B7,$1F,$F6,$E4
	.dc.b	$EA,$BE,$84,$BE,$0B,$A4,$7E,$34,$E9,$0F,$9D,$DF,$C7,$49,$FB,$B3
	.dc.b	$97,$F9,$AF,$0F,$F3,$47,$91,$1F,$9B,$0F,$8F,$F0,$FC,$7C,$AE,$71
	.dc.b	$FE,$09,$AE,$44,$E8,$FC,$C6,$1E,$8F,$8F,$F4,$FC,$7C,$C3,$C7,$1F
	.dc.b	$E1,$DD,$FF,$45,$5C,$27,$FD,$33,$18,$6F,$86,$CF,$E1,$FC,$7C,$9F
	.dc.b	$6B,$F1,$BC,$5E,$95,$B9,$E3,$87,$57,$FD,$43,$87,$EC,$70,$F9,$BF
	.dc.b	$1F,$27,$E3,$BF,$43,$E3,$BF,$C6,$5E,$23,$EB,$9F,$11,$FA,$C5,$F1
	.dc.b	$34,$C4,$7D,$8F,$90,$FB,$3F,$21,$32,$FA,$71,$E6,$E7,$FD,$F1,$23
	.dc.b	$EB,$F9,$9F,$80,$F3,$3F,$03,$F2,$12,$7E,$DC,$7F,$91,$7B,$3F,$17
	.dc.b	$E6,$7E,$18,$93,$F8,$0F,$90,$FC,$4F,$C8,$48,$FC,$1B,$F8,$CF,$90
	.dc.b	$E5,$DE,$F8,$7F,$DF,$77,$7F,$BF,$BA,$3F,$4E,$DF,$0B,$F2,$B5,$BC
	.dc.b	$7F,$84,$CB,$E5,$38,$92,$33,$F2,$9E,$79,$7C,$97,$6E,$79,$5C,$A3
	.dc.b	$C3,$F2,$52,$3A,$BE,$4E,$8E,$5F,$23,$E6,$24,$F8,$BE,$4B,$D3,$EE
	.dc.b	$FA,$BD,$59,$3E,$68,$F1,$7B,$7E,$EF,$91,$93,$EA,$8F,$83,$F2,$1F
	.dc.b	$0B,$E4,$24,$FB,$A3,$D5,$F2,$1F,$18,$91,$F0,$BE,$43,$E4,$FC,$04
	.dc.b	$8F,$8A,$FC,$CF,$8F,$F8,$5F,$1F,$F3,$BE,$3E,$57,$CA,$8F,$E5,$DF
	.dc.b	$31,$FE,$AD,$F0,$4F,$F2,$02,$7C,$2F,$E0,$57,$C3,$70,$F8,$79,$87
	.dc.b	$F3,$C9,$CB,$7E,$43,$AF,$C8,$48,$C1,$FE,$44,$C7,$BF,$C7,$F8,$BE
	.dc.b	$42,$4F,$38,$F2,$FC,$7F,$9B,$E4,$24,$F7,$8F,$27,$C5,$7A,$89,$1E
	.dc.b	$3F,$8A,$F6,$FC,$54,$8F,$33,$F0,$3E,$27,$D9,$F1,$5F,$0B,$E2,$65
	.dc.b	$7A,$E3,$E2,$FC,$57,$C6,$F8,$A9,$3E,$F8,$F8,$9F,$15,$F2,$BE,$22
	.dc.b	$4F,$C3,$8F,$9B,$F1,$3F,$D6,$F8,$BF,$98,$4F,$F5,$46,$9F,$42,$3F
	.dc.b	$5C,$FF,$31,$8F,$EC,$ED,$5E,$87,$F4,$C8,$C3,$7D,$9E,$DF,$03,$DF
	.dc.b	$E2,$25,$F2,$8E,$BF,$05,$E3,$24,$76,$F8,$6F,$37,$C3,$C8,$F0,$BE
	.dc.b	$9F,$87,$F2,$FC,$3F,$AC,$93,$E6,$F8,$6F,$77,$C3,$7B,$FE,$0E,$4F
	.dc.b	$AA,$3D,$BF,$09,$F0,$FE,$12,$4F,$BA,$3E,$37,$C1,$FC,$7F,$84,$93
	.dc.b	$F0,$A3,$E2,$FC,$27,$CB,$F8,$49,$3F,$1A,$3E,$77,$C2,$7C,$FF,$84
	.dc.b	$93,$F2,$A3,$E6,$FC,$27,$D3,$24,$7F,$85,$3F,$02,$FE,$20,$FF,$82
	.dc.b	$1F,$E1,$CE,$7E,$0B,$FC,$29,$06,$FC,$1F,$F2,$58,$FE,$1F,$3C,$FE
	.dc.b	$0B,$67,$0F,$82,$E3,$F0,$5C,$BE,$0A,$57,$86,$39,$7C,$17,$6F,$82
	.dc.b	$93,$E5,$8F,$0F,$C0,$C8,$F4,$3E,$4F,$82,$F1,$7C,$0F,$9B,$E0,$A5
	.dc.b	$7B,$63,$D3,$F0,$3E,$AF,$82,$93,$F0,$63,$D5,$F0,$5E,$EF,$82,$93
	.dc.b	$F1,$63,$E0,$FC,$17,$C2,$F8,$29,$3F,$26,$3E,$17,$C0,$FC,$6F,$82
	.dc.b	$93,$F3,$63,$FC,$79,$CD,$FC,$3C,$9F,$05,$8F,$83,$FE,$9D,$1F,$C2
	.dc.b	$AB,$F3,$FF,$0B,$FE,$16,$87,$C1,$F2,$39,$FC,$27,$4F,$83,$97,$FE
	.dc.b	$4E,$A4,$78,$3E,$0F,$FD,$E3,$E9,$FF,$B0,$73,$F1,$FC,$14,$C3,$83
	.dc.b	$E6,$F8,$3F,$3F,$C1,$7A,$3E,$0E,$57,$58,$E7,$B2,$27,$C4,$B1,$CC
	.dc.b	$7E,$E3,$F7,$A7,$95,$F0,$BE,$17,$E1,$FC,$2C,$AF,$44,$7C,$6F,$85
	.dc.b	$F8,$FF,$0B,$27,$D9,$1F,$E7,$9C,$39,$C7,$F8,$69,$AF,$F8,$41,$BC
	.dc.b	$0D,$3C,$DF,$13,$FE,$94,$F6,$7E,$9F,$C2,$CC,$3E,$43,$F5,$BE,$1B
	.dc.b	$EB,$FC,$2C,$9F,$98,$FD,$AF,$86,$F1,$7A,$FE,$3F,$6A,$57,$8E,$3E
	.dc.b	$FF,$C2,$C8,$FA,$8F,$E0,$BE,$1B,$F0,$7F,$0D,$FE,$7F,$FC,$CF,$F9
	.dc.b	$19,$FC,$93,$BF,$FD,$43,$C3,$E9,$C1,$C1,$FE,$AC,$F7,$7F,$D1,$F8
	.dc.b	$7F,$A6,$27,$E9,$56,$EB,$51,$FE,$90,$C3,$F9,$3F,$17,$F9,$81,$FF
	.dc.b	$2B,$47,$E3,$8F,$F4,$3A,$F8,$4F,$21,$E5,$F8,$6F,$37,$C3,$79,$FE
	.dc.b	$1A,$5F,$58,$FF,$50,$FD,$87,$F0,$88,$E9,$92,$31,$9F,$95,$B9,$FD
	.dc.b	$47,$F8,$65,$79,$FE,$1F,$E2,$74,$48,$C1,$EB,$F8,$7F,$90,$47,$DD
	.dc.b	$F0,$FF,$2F,$E1,$FE,$67,$C3,$C9,$F8,$51,$F3,$FA,$3E,$87,$44,$9F
	.dc.b	$8D,$1F,$4F,$97,$EA,$7C,$34,$9F,$95,$1F,$5F,$E1,$FE,$C7,$C3,$C9
	.dc.b	$F9,$D1,$F6,$FC,$59,$1F,$45,$FB,$DF,$0D,$F7,$FE,$1E,$4F,$D5,$7F
	.dc.b	$05,$F0,$FF,$EC,$1F,$CC,$FF,$99,$FC,$49,$F8,$F5,$FE,$BC,$6C,$67
	.dc.b	$9B,$87,$C4,$F1,$F8,$79,$7F,$ED,$0B,$F9,$AF,$DF,$BE,$9F,$11,$DF
	.dc.b	$E1,$E4,$EE,$3C,$5F,$11,$23,$93,$E5,$F8,$7F,$37,$C4,$79,$FE,$22
	.dc.b	$57,$78,$F5,$7C,$44,$8F,$13,$ED,$F8,$7F,$77,$C4,$7B,$FE,$1E,$57
	.dc.b	$9E,$3E,$17,$C4,$7C,$3F,$87,$93,$EB,$8F,$8D,$F1,$12,$3D,$CF,$C9
	.dc.b	$F8,$7F,$95,$F1,$12,$7E,$13,$F3,$7E,$1F,$E7,$7C,$47,$FB,$BF,$F3
	.dc.b	$3F,$F2,$11,$FE,$43,$E6,$46,$7E,$26,$61,$FE,$FA,$AF,$F7,$29,$FC
	.dc.b	$DF,$CC,$FF,$6F,$BF,$CD,$A3,$5F,$13,$D4,$ED,$F1,$1D,$FE,$22,$4F
	.dc.b	$37,$C5,$F1,$1E,$3F,$88,$F2,$7C,$3C,$AF,$04,$79,$FE,$22,$47,$8D
	.dc.b	$F5,$7C,$2F,$AF,$E1,$A4,$F9,$DF,$77,$C2,$FB,$FE,$1A,$4F,$AD,$F8
	.dc.b	$5F,$0B,$F0,$FE,$16,$4F,$BD,$F8,$DF,$0B,$F1,$FE,$17,$E4,$7C,$2C
	.dc.b	$AF,$89,$1F,$2F,$E1,$7E,$67,$C2,$49,$F9,$11,$FD,$1D,$FC,$97,$C1
	.dc.b	$7F,$90,$8F,$E2,$AB,$CB,$FF,$77,$7C,$7E,$FA,$7F,$91,$47,$F1,$89
	.dc.b	$E3,$59,$F3,$FF,$DF,$6F,$F1,$98,$FE,$47,$18,$F8,$4D,$47,$59,$FE
	.dc.b	$43,$7D,$1F,$F7,$94,$70,$35,$F0,$9C,$30,$FF,$80,$0D,$9D,$CF,$AB
	.dc.b	$E0,$70,$47,$95,$25,$F0,$3C,$7E,$06,$56,$A5,$1D,$7E,$07,$B7,$C0
	.dc.b	$4A,$D3,$E1,$F8,$1F,$17,$C0,$78,$FE,$06,$57,$38,$F3,$7C,$04,$8E
	.dc.b	$CF,$A7,$E0,$7D,$5F,$01,$EB,$F8,$19,$5E,$38,$F7,$7C,$0C,$8F,$33
	.dc.b	$F0,$7E,$07,$E1,$7C,$0C,$9F,$53,$F1,$7E,$07,$E4,$7E,$67,$FC,$68
	.dc.b	$9D,$38,$F8,$7C,$FC,$1E,$BE,$0E,$63,$F2,$2F,$F8,$BF,$5F,$E0,$1F
	.dc.b	$F8,$03,$19,$F8,$3E,$A7,$6F,$83,$EF,$F0,$72,$79,$3E,$2F,$83,$F1
	.dc.b	$FC,$14,$9E,$CF,$9B,$E0,$FC,$FF,$07,$E8,$F8,$39,$5E,$38,$F5,$FC
	.dc.b	$14,$8F,$33,$EE,$F8,$3F,$7F,$C1,$7C,$0F,$83,$95,$EB,$8F,$87,$F0
	.dc.b	$72,$3D,$CF,$C6,$F8,$3F,$8F,$F0,$72,$7E,$13,$F2,$BE,$0F,$E5,$FC
	.dc.b	$14,$9F,$8C,$FE,$57,$F3,$3F,$F6,$D0,$F8,$3C,$7C,$64,$AF,$C9,$CE
	.dc.b	$5B,$F0,$7C,$48,$EB,$E0,$FA,$7C,$1F,$5F,$83,$93,$CA,$3C,$1F,$05
	.dc.b	$23,$AB,$E3,$F8,$3F,$25,$F4,$9F,$EA,$18,$FF,$25,$6C,$76,$7C,$B8
	.dc.b	$76,$77,$87,$57,$8E,$17,$BB,$F8,$1E,$0C,$7E,$17,$C1,$FC,$3F,$83
	.dc.b	$98,$7C,$08,$F8,$DF,$07,$F1,$FE,$0E,$4F,$C4,$8F,$95,$F0,$7F,$2F
	.dc.b	$E0,$E4,$FC,$88,$F9,$DF,$07,$F3,$FE,$0E,$4F,$CC,$8F,$A5,$F0,$5F
	.dc.b	$4F,$E0,$E4,$FD,$08,$FA,$DF,$05,$23,$E9,$BF,$67,$E0,$FF,$87,$7E
	.dc.b	$67,$FE,$E4,$3F,$91,$FC,$01,$A6,$7D,$FD,$FC,$1C,$C3,$F8,$7C,$FF
	.dc.b	$1C,$F4,$BF,$CE,$97,$C7,$E2,$7A,$92,$B7,$F0,$9E,$06,$1C,$8F,$1F
	.dc.b	$C2,$79,$32,$ED,$1F,$E0,$9C,$7F,$97,$36,$3A,$39,$BB,$D3,$FC,$E2
	.dc.b	$98,$7B,$BE,$12,$5F,$A1,$F8,$3F,$09,$F0,$BE,$13,$E1,$FC,$24,$AF
	.dc.b	$6C,$7C,$6F,$84,$91,$F0,$1F,$93,$F0,$7C,$A3,$A1,$0F,$45,$F5,$7E
	.dc.b	$3D,$F5,$7D,$97,$D1,$F7,$5D,$CF,$D5,$87,$D6,$F8,$4E,$D5,$D0,$FA
	.dc.b	$2F,$57,$E3,$61,$D6,$BE,$16,$1C,$F8,$7E,$AE,$AC,$B1,$87,$F6,$18
	.dc.b	$65,$F8,$4F,$84,$FC,$2F,$C2,$4A,$FB,$F1,$F8,$8F,$83,$91,$F8,$27
	.dc.b	$F1,$7F,$09,$F8,$CF,$84,$FC,$6F,$C2,$4A,$FC,$3C,$7E,$43,$E1,$3B
	.dc.b	$47,$43,$F1,$6F,$47,$F0,$17,$77,$FB,$98,$75,$9F,$C8,$E1,$D5,$FC
	.dc.b	$36,$1F,$9C,$F8,$4F,$ED,$5E,$77,$F9,$70,$FF,$2E,$EC,$8C,$FC,$26
	.dc.b	$BE,$12,$61,$FD,$AA,$FF,$CB,$E3,$E1,$39,$92,$35,$F0,$5D,$BE,$0F
	.dc.b	$C1,$1F,$A4,$C7,$0F,$3C,$7A,$BF,$E6,$3C,$3F,$A7,$F2,$FE,$A1,$EF
	.dc.b	$8F,$59,$D3,$EA,$C7,$FC,$17,$EA,$30,$FA,$F5,$FE,$09,$8C,$72,$EA
	.dc.b	$73,$3F,$EA,$F9,$E1,$CD,$9E,$9D,$73,$4A,$E5,$69,$FF,$0C,$FC,$4B
	.dc.b	$F6,$D7,$56,$17,$5D,$31,$1E,$1C,$A6,$3D,$1F,$27,$5F,$86,$3B,$E7
	.dc.b	$E1,$7C,$D9,$F8,$5F,$F1,$26,$5E,$1C,$3B,$46,$32,$F0,$B6,$3E,$FF
	.dc.b	$C2,$F6,$8F,$5E,$38,$7D,$BA,$77,$EA,$FF,$6F,$6D,$F2,$F4,$F2,$57
	.dc.b	$FB,$2B,$2F,$7B,$DF,$6F,$0B,$9C,$FF,$1B,$F0,$BF,$8E,$F8,$59,$3F
	.dc.b	$89,$8F,$C8,$FC,$2C,$8F,$C6,$3F,$94,$F8,$4F,$CA,$FC,$2F,$E5,$BE
	.dc.b	$16,$57,$E4,$63,$F3,$3F,$0B,$F9,$AF,$84,$93,$F9,$58,$E8,$FE,$2F
	.dc.b	$0B,$C1,$FE,$58,$EA,$65,$DA,$BF,$39,$97,$67,$F2,$59,$7E,$9B,$E1
	.dc.b	$7F,$4F,$F0,$B2,$7F,$45,$1F,$AA,$F8,$59,$1F,$A5,$7F,$5B,$F0,$BF
	.dc.b	$AE,$F8,$5F,$D7,$FC,$2C,$AF,$D5,$47,$F9,$6F,$F3,$47,$C7,$77,$9E
	.dc.b	$A7,$F9,$62,$32,$F5,$21,$FA,$7C,$3B,$BD,$0F,$DB,$5F,$77,$F5,$97
	.dc.b	$DF,$A7,$C1,$BC,$9E,$D1,$E3,$F8,$79,$1E,$17,$CD,$F0,$FE,$7F,$86
	.dc.b	$F4,$7C,$3C,$AF,$34,$7A,$FE,$1F,$D9,$F0,$D2,$7D,$51,$D1,$FF,$70
	.dc.b	$6D,$FE,$11,$E6,$FF,$45,$B7,$C9,$FE,$CE,$B2,$F0,$57,$C3,$CB,$C0
	.dc.b	$F2,$CB,$E5,$7C,$2C,$8F,$90,$FC,$DF,$86,$F0,$65,$DB,$5B,$9F,$6C
	.dc.b	$BA,$E1,$E4,$CB,$AD,$F9,$32,$FA,$DF,$0B,$F5,$FE,$1A,$5F,$D5,$8F
	.dc.b	$B5,$F0,$B2,$3E,$C3,$F7,$7E,$1B,$EF,$7C,$2F,$DF,$F8,$69,$5F,$76
	.dc.b	$3F,$05,$F0,$BF,$83,$F8,$69,$3F,$81,$8E,$8F,$C8,$C2,$E9,$F2,$31
	.dc.b	$EB,$3F,$87,$C7,$AB,$F3,$B1,$FC,$77,$C3,$7E,$3F,$E1,$A4,$FE,$36
	.dc.b	$3F,$D5,$9F,$9A,$7F,$B0,$E1,$8F,$FA,$FA,$30,$D8,$DF,$C5,$70,$F8
	.dc.b	$5E,$3F,$11,$27,$FD,$5F,$7F,$E2,$17,$FD,$94,$FF,$AD,$AF,$EF,$E1
	.dc.b	$E4,$9D,$FC,$37,$89,$87,$23,$B9,$D3,$0E,$CF,$9B,$9B,$FC,$5B,$87
	.dc.b	$F8,$8B,$0F,$45,$7F,$B3,$A9,$F9,$1B,$DD,$F7,$FC,$3C,$9F,$F6,$71
	.dc.b	$A8,$E1,$2F,$5F,$97,$E7,$AE,$6A,$6A,$25,$F2,$96,$DC,$0C,$C7,$9C
	.dc.b	$1E,$49,$CE,$6D,$7F,$DE,$98,$F8,$6B,$CA,$D7,$91,$C4,$CD,$72,$9E
	.dc.b	$2D,$8E,$F3,$C6,$BB,$46,$27,$5F,$14,$D9,$F1,$5F,$EF,$6F,$25,$F9
	.dc.b	$97,$F8,$17,$0F,$F7,$A4,$63,$E0,$BF,$13,$5D,$E7,$C8,$79,$B3,$F0
	.dc.b	$69,$F0,$21,$B3,$A3,$85,$6A,$7C,$87,$CB,$EB,$FC,$AF,$C5,$C8,$FC
	.dc.b	$6B,$F9,$8F,$8B,$FC,$CF,$B3,$27,$F2,$2F,$E7,$3D,$AF,$CE,$FC,$0F
	.dc.b	$E7,$BE,$0A,$57,$E5,$A3,$F4,$3E,$FF,$E8,$BD,$F9,$3F,$9A,$8E,$8F
	.dc.b	$D9,$E5,$F8,$D8,$E3,$9F,$F4,$D9,$76,$AF,$D0,$E5,$D9,$E1,$CD,$FE
	.dc.b	$4D,$7F,$4B,$F0,$1F,$B0,$23,$FA,$8F,$8A,$FD,$9F,$C5,$7E,$D3,$E2
	.dc.b	$64,$FE,$B2,$3F,$6F,$F1,$52,$3F,$5E,$FE,$EB,$E2,$FF,$77,$D5,$27
	.dc.b	$F6,$6F,$D1,$D7,$F0,$3B,$BF,$35,$FC,$1D,$79,$B1,$EE,$FA,$32,$F1
	.dc.b	$BD,$88,$7B,$32,$F0,$3F,$9C,$C3,$AD,$FF,$8C,$E3,$1D,$5F,$EA,$28
	.dc.b	$C4,$7C,$B6,$CF,$0F,$8B,$F3,$56,$FE,$2E,$DB,$3F,$47,$C5,$E1,$B1
	.dc.b	$D8,$D6,$3F,$E5,$57,$11,$EF,$F1,$38,$4F,$E3,$E3,$E1,$78,$58,$9C
	.dc.b	$FC,$3E,$BE,$22,$61,$CB,$04,$F2,$CB,$9E,$52,$B5,$1C,$DE,$18,$F8
	.dc.b	$0C,$3F,$07,$E1,$BC,$7F,$0D,$23,$B4,$75,$7C,$17,$D5,$F9,$17,$CB
	.dc.b	$7F,$9F,$87,$59,$F3,$69,$EE,$F8,$59,$1E,$97,$E0,$FC,$37,$C2,$F8
	.dc.b	$7F,$87,$F1,$12,$BD,$D1,$F1,$BE,$1E,$47,$C1,$7E,$4F,$C4,$7C,$AF
	.dc.b	$87,$93,$F1,$5F,$9B,$F1,$1D,$63,$E4,$61,$D5,$FC,$06,$1D,$1F,$C1
	.dc.b	$61,$75,$FC,$4E,$3F,$5B,$E1,$E5,$7D,$17,$EC,$FC,$47,$E0,$7A,$7F
	.dc.b	$C4,$B3,$FB,$56,$DF,$E7,$9B,$D3,$FE,$F2,$5F,$0F,$C7,$E2,$39,$7C
	.dc.b	$1C,$C3,$F0,$F4,$9E,$DF,$07,$27,$2F,$87,$E0,$BC,$5E,$A4,$9E,$2F
	.dc.b	$97,$1E,$D1,$D3,$1E,$CF,$9F,$0F,$57,$C1,$FA,$FE,$1E,$4F,$8A,$3D
	.dc.b	$DF,$11,$23,$CA,$FC,$1F,$87,$F8,$5F,$0F,$27,$D2,$FC,$5F,$86,$F8
	.dc.b	$DF,$0F,$27,$DA,$FC,$9F,$87,$E9,$1F,$27,$0B,$C7,$F2,$71,$E9,$3F
	.dc.b	$37,$1B,$3F,$37,$1F,$A5,$F0,$FF,$4F,$E1,$E5,$FC,$A8,$E7,$CB,$F4
	.dc.b	$72,$B3,$FB,$BE,$56,$7D,$2B,$EB,$67,$6B,$59,$7D,$FF,$72,$47,$D6
	.dc.b	$7F,$05,$ED,$FE,$0F,$E0,$64,$FD,$A7,$F0,$DF,$03,$F8,$7F,$87,$93
	.dc.b	$F7,$9F,$E1,$5F,$9A,$FF,$87,$B3,$FE,$EB,$E7,$A1,$F1,$1A,$F8,$69
	.dc.b	$87,$F0,$C9,$CB,$DC,$87,$E4,$F2,$F0,$BD,$C8,$7E,$5F,$1F,$0B,$E2
	.dc.b	$F8,$88,$F5,$7C,$BF,$11,$E6,$F8,$8F,$3F,$C3,$CA,$F1,$47,$AB,$E1
	.dc.b	$64,$79,$5F,$6F,$C2,$FB,$BE,$0B,$DF,$F0,$52,$BD,$51,$D5,$F6,$61
	.dc.b	$D5,$FE,$C1,$AB,$1E,$B1,$8C,$6F,$67,$F4,$CA,$65,$F3,$3E,$16,$5F
	.dc.b	$C7,$7E,$7F,$C1,$FD,$0F,$84,$93,$F2,$DF,$A7,$F0,$BF,$53,$E1,$3E
	.dc.b	$AF,$C2,$4A,$FA,$11,$F6,$3E,$0A,$47,$D3,$7E,$DF,$C1,$74,$8F,$4E
	.dc.b	$17,$5F,$36,$3D,$67,$ED,$63,$D5,$F7,$E3,$F8,$4F,$82,$91,$F7,$DF
	.dc.b	$C3,$FC,$27,$3E,$CF,$C3,$F0,$7F,$BD,$6B,$B9,$F8,$5C,$7B,$3F,$2B
	.dc.b	$1E,$D9,$7C,$7C,$BF,$23,$F0,$92,$BF,$18,$FF,$5D,$FC,$D7,$FB,$AD
	.dc.b	$7F,$78,$7F,$B3,$E5,$F9,$66,$D7,$F6,$56,$F7,$17,$8B,$CF,$DF,$95
	.dc.b	$FD,$92,$7F,$B3,$3F,$E1,$D5,$F0,$1E,$1F,$7F,$C5,$F0,$12,$B8,$47
	.dc.b	$97,$E3,$64,$73,$7D,$18,$F6,$8E,$98,$DD,$3D,$58,$FB,$7E,$12,$57
	.dc.b	$8D,$F8,$1F,$0D,$F0,$7E,$13,$E1,$7C,$2C,$AF,$44,$7C,$5F,$85,$F8
	.dc.b	$DF,$0B,$27,$D9,$1F,$27,$E0,$A4,$7B,$DF,$99,$F0,$7D,$63,$E1,$61
	.dc.b	$D5,$F8,$58,$75,$7E,$36,$1D,$5F,$8D,$87,$D5,$F8,$39,$1F,$2D,$FB
	.dc.b	$1F,$07,$FE,$18,$E6,$7E,$34,$C3,$FC,$37,$E7,$BA,$9D,$D8,$6E,$5F
	.dc.b	$E1,$BC,$38,$DF,$3C,$39,$4F,$D9,$F8,$6F,$C3,$19,$99,$1C,$67,$6F
	.dc.b	$FA,$46,$B9,$B9,$9D,$B9,$9D,$F2,$7F,$A1,$AB,$73,$9A,$DC,$E6,$B7
	.dc.b	$5F,$E9,$C3,$71,$99,$FC,$D7,$C3,$4B,$FC,$83,$F9,$DF,$86,$E7,$7C
	.dc.b	$70,$E7,$5F,$EA,$36,$F6,$FE,$6F,$E1,$B7,$39,$C6,$CB,$F3,$5F,$EF
	.dc.b	$F3,$FE,$47,$7F,$BF,$BF,$E4,$27,$FB,$E3,$7E,$0B,$FC,$95,$1A,$8F
	.dc.b	$F3,$F2,$F8,$3C,$92,$39,$7C,$3F,$0F,$87,$91,$D5,$E7,$F0,$9D,$3E
	.dc.b	$17,$AF,$C0,$CA,$F1,$47,$37,$BE,$17,$8B,$FC,$EB,$93,$2E,$B5,$E4
	.dc.b	$CA,$D6,$B2,$F4,$FC,$0F,$AB,$E0,$65,$FB,$E3,$DB,$F0,$32,$3E,$13
	.dc.b	$F0,$3E,$17,$E0,$FC,$2C,$9F,$8C,$FC,$4F,$84,$F8,$BF,$0B,$27,$E5
	.dc.b	$3F,$23,$E1,$7A,$47,$C8,$C2,$E9,$EA,$C7,$A4,$FC,$CC,$6C,$FB,$B1
	.dc.b	$FA,$3F,$0B,$F4,$BE,$0E,$5F,$D7,$8F,$AB,$F0,$BF,$5B,$E1,$24,$FD
	.dc.b	$B8,$FF,$41,$7C,$27,$F7,$98,$FF,$43,$37,$A4,$D3,$FB,$BC,$7F,$AE
	.dc.b	$D7,$C2,$F2,$24,$7F,$A2,$23,$8B,$DC,$FC,$55,$F8,$9F,$C3,$DF,$89
	.dc.b	$87,$13,$CB,$E2,$79,$BE,$17,$CF,$F0,$52,$7B,$47,$B9,$EA,$43,$A9
	.dc.b	$CC,$87,$A4,$EA,$79,$FE,$0F,$E1,$91,$F5,$7C,$1F,$C6,$F8,$39,$1E
	.dc.b	$D7,$E4,$FC,$27,$CA,$F8,$59,$3F,$05,$F9,$BF,$0B,$F3,$BE,$17,$E7
	.dc.b	$FC,$1C,$AF,$8D,$1D,$1F,$99,$85,$E6,$FF,$67,$F0,$B2,$EB,$5F,$47
	.dc.b	$2B,$5A,$CB,$ED,$FC,$0F,$DC,$F3,$A5,$FD,$38,$FB,$FF,$03,$23,$EB
	.dc.b	$3F,$82,$F8,$5F,$C1,$FC,$2F,$E1,$3E,$16,$57,$DB,$8F,$C3,$FA,$72
	.dc.b	$3E,$F3,$F8,$AF,$84,$E9,$1F,$85,$C2,$E9,$EA,$C7,$AC,$FE,$17,$1E
	.dc.b	$AF,$E2,$F1,$FF,$6A,$7D,$27,$F7,$8F,$4B,$FC,$96,$FF,$74,$23,$FE
	.dc.b	$DC,$FB,$07,$E4,$CF,$0F,$0D,$C3,$FD,$B9,$3A,$9C,$F8,$7A,$F1,$F7
	.dc.b	$E3,$CB,$DE,$09,$E7,$E3,$F4,$F1,$E5,$6A,$3B,$E3,$23,$8B,$D9,$E7
	.dc.b	$5D,$57,$03,$44,$34,$F5,$E7,$E5,$8F,$F7,$8E,$BF,$E1,$26,$FC,$1F
	.dc.b	$C6,$3E,$3F,$C1,$F8,$63,$AE,$1D,$30,$F8,$11,$E4,$3D,$BB,$F5,$76
	.dc.b	$78,$77,$65,$E7,$CF,$CD,$39,$D3,$EA,$FC,$1F,$D6,$F8,$59,$3E,$68
	.dc.b	$FB,$3F,$0D,$23,$D2,$FD,$CF,$85,$FB,$BF,$0B,$F7,$BE,$0A,$57,$BA
	.dc.b	$3F,$03,$F0,$72,$3E,$0B,$F8,$4F,$83,$E9,$1E,$8F,$3D,$CA,$CB,$3B
	.dc.b	$7A,$27,$39,$F8,$2F,$E5,$E7,$E0,$79,$60,$CF,$F2,$3E,$E4,$9F,$A2
	.dc.b	$FE,$53,$E1,$7F,$2B,$F0,$B2,$7E,$AB,$F9,$8F,$85,$FC,$CF,$C2,$FE
	.dc.b	$6B,$E1,$65,$7D,$A8,$FC,$EF,$C2,$FE,$7B,$E1,$64,$FD,$E8,$E8,$F9
	.dc.b	$30,$BC,$1E,$BD,$59,$76,$AF,$C2,$E5,$D9,$F5,$E5,$FA,$AF,$84,$91
	.dc.b	$F8,$A7,$F1,$EF,$C9,$FC,$6B,$FB,$FF,$11,$FE,$F5,$97,$C9,$6F,$6F
	.dc.b	$FA,$F1,$B4,$AE,$1F,$0B,$D2,$F8,$7C,$1D,$C3,$83,$6F,$86,$1F,$E0
	.dc.b	$57,$F4,$35,$DF,$E0,$B5,$8F,$8F,$E0,$AD,$9A,$63,$E0,$B3,$F1,$53
	.dc.b	$1E,$79,$F9,$E3,$8E,$9C,$B4,$95,$A3,$C6,$F6,$CF,$BF,$C2,$C9,$D6
	.dc.b	$9F,$16,$3C,$7F,$0D,$DE,$7E,$56,$37,$6F,$91,$97,$69,$F9,$D9,$59
	.dc.b	$F7,$47,$4D,$BF,$DF,$2D,$F8,$5F,$79,$F0,$3E,$17,$E0,$F9,$72,$78
	.dc.b	$3F,$13,$CC,$F8,$BE,$64,$9E,$8F,$C8,$F2,$7E,$4F,$8F,$F2,$BC,$89
	.dc.b	$5E,$18,$E8,$FE,$0F,$0B,$A7,$E0,$71,$E9,$3F,$87,$C6,$CF,$E1,$71
	.dc.b	$FA,$BF,$09,$F5,$BE,$16,$5F,$B6,$3E,$CF,$C1,$C8,$F8,$0F,$DC,$F8
	.dc.b	$4F,$BF,$F9,$A7,$F8,$3C,$93,$8F,$85,$CF,$C3,$48,$FA,$73,$FE,$10
	.dc.b	$6F,$BF,$D5,$FC,$E6,$17,$4F,$CD,$63,$DA,$61,$C8,$EC,$74,$F8,$5F
	.dc.b	$19,$1E,$FF,$0D,$FE,$90,$FD,$DB,$FD,$8F,$A1,$FE,$69,$3F,$CA,$A3
	.dc.b	$1E,$BE,$FE,$2A,$4F,$A3,$E2,$3D,$E6,$1E,$17,$D0,$93,$9C,$FF,$98
	.dc.b	$E1,$DE,$31,$E1,$7F,$48,$8C,$5F,$E6,$DA,$F4,$9D,$A8,$CD,$74,$BF
	.dc.b	$F6,$FE,$EF,$88,$FA,$7F,$0D,$30,$F9,$F1,$CB,$0E,$39,$78,$70,$E3
	.dc.b	$F0,$D9,$7D,$37,$F5,$BE,$1B,$EE,$FC,$34,$8F,$B2,$FE,$03,$E1,$FF
	.dc.b	$03,$F1,$1F,$82,$F8,$79,$5F,$7A,$3F,$0B,$F1,$12,$3F,$02,$FE,$23
	.dc.b	$E1,$BF,$13,$F0,$DF,$8A,$F8,$59,$5F,$86,$8E,$AF,$E2,$30,$EA,$FF
	.dc.b	$49,$E2,$63,$D6,$31,$8D,$E3,$FE,$95,$4C,$BF,$2F,$F0,$D2,$FF,$26
	.dc.b	$FE,$6B,$E2,$3F,$37,$F1,$1F,$C6,$BC,$BF,$F0,$D5,$7F,$27,$6F,$A0
	.dc.b	$C5,$5B,$E9,$97,$FC,$42,$D1,$CF,$AF,$A7,$C4,$4A,$D3,$DF,$E1,$7C
	.dc.b	$31,$C7,$A7,$C4,$61,$FE,$D1,$87,$47,$B6,$1E,$07,$38,$7A,$7E,$1B
	.dc.b	$D5,$F0,$92,$7C,$91,$ED,$F8,$4F,$77,$C1,$C9,$F4,$47,$C1,$F8,$4F
	.dc.b	$85,$F0,$D2,$7D,$91,$F1,$7E,$1B,$E3,$7C,$34,$9F,$81,$1D,$5F,$85
	.dc.b	$87,$57,$D3,$87,$57,$E3,$61,$D5,$F6,$E1,$F4,$7E,$1B,$E9,$7C,$24
	.dc.b	$9F,$99,$1D,$5F,$87,$87,$5E,$9F,$F0,$47,$03,$2E,$D5,$F2,$32,$ED
	.dc.b	$5A,$CB,$EF,$7C,$1F,$DF,$F8,$29,$7F,$66,$3F,$05,$F0,$72,$3E,$E3
	.dc.b	$F8,$5F,$85,$FC,$37,$C3,$7E,$1F,$E1,$65,$7E,$06,$3F,$B1,$FE,$6B
	.dc.b	$FD,$D8,$FE,$E3,$DA,$69,$FD,$EB,$76,$3D,$EF,$EF,$BD,$DE,$87,$E3
	.dc.b	$70,$EE,$FE,$13,$0E,$FF,$01,$E0,$B9,$3F,$D8,$B0,$FF,$80,$2F,$F8
	.dc.b	$C5,$FF,$77,$39,$F8,$6F,$39,$E8,$F8,$5F,$4F,$C3,$7A,$BE,$1A,$57
	.dc.b	$38,$F6,$FC,$37,$BB,$E0,$E4,$F7,$8E,$4F,$52,$1E,$1B,$EA,$F9,$6F
	.dc.b	$AB,$E5,$BF,$93,$F0,$9F,$2B,$E0,$A4,$FA,$E3,$E6,$FC,$14,$8F,$73
	.dc.b	$F4,$3E,$0B,$E8,$FC,$17,$D2,$F8,$39,$5F,$0E,$3E,$AF,$C2,$7D,$6F
	.dc.b	$83,$93,$F1,$E3,$A3,$F1,$F0,$BD,$9F,$E5,$5E,$2F,$F4,$B4,$FC,$BC
	.dc.b	$3A,$BF,$37,$1F,$0B,$F4,$7E,$13,$F0,$7F,$05,$23,$07,$A0,$87,$73
	.dc.b	$EC,$69,$C9,$E8,$7D,$BA,$C6,$3E,$3F,$82,$FF,$18,$FE,$6B,$FF,$00
	.dc.b	$4F,$F2,$3D,$7D,$F8,$CF,$BD,$23,$FC,$4B,$3F,$D3,$5B,$F0,$BC,$BE
	.dc.b	$16,$46,$9E,$BF,$0B,$DB,$E1,$7B,$FC,$2C,$AE,$71,$D5,$FF,$3F,$61
	.dc.b	$D5,$B8,$F9,$B1,$EB,$1A,$C6,$E7,$FE,$81,$B6,$5E,$CF,$84,$F6,$FC
	.dc.b	$2C,$C3,$D3,$1F,$03,$E1,$7E,$0F,$C2,$49,$F6,$C7,$C4,$F8,$4F,$8B
	.dc.b	$F0,$92,$7E,$0C,$7C,$8F,$81,$91,$F1,$1F,$97,$F0,$5D,$63,$E3,$E1
	.dc.b	$D5,$F6,$E1,$D5,$F9,$78,$75,$7E,$0E,$1F,$53,$E0,$A4,$7D,$07,$EB
	.dc.b	$FC,$17,$D8,$F8,$39,$3F,$51,$FB,$7F,$09,$F7,$3E,$13,$EE,$FC,$24
	.dc.b	$AF,$B3,$1F,$80,$F8,$39,$1F,$71,$FF,$39,$FD,$47,$F5,$8E,$46,$3D
	.dc.b	$AB,$EA,$BD,$9E,$67,$E1,$30,$EC,$FD,$AC,$39,$FC,$0F,$4F,$81,$93
	.dc.b	$FE,$6E,$C7,$FA,$A3,$D0,$F5,$10,$FC,$1E,$7E,$B7,$B1,$0E,$C7,$3F
	.dc.b	$84,$F4,$91,$ED,$F0,$9E,$CF,$82,$F6,$FC,$1C,$9F,$14,$7C,$0F,$7A
	.dc.b	$47,$95,$F8,$7E,$FF,$C4,$F7,$BE,$2F,$BF,$2B,$D5,$1D,$5F,$66,$1D
	.dc.b	$5F,$F6,$C6,$8C,$7A,$C6,$31,$BD,$FF,$ED,$9A,$FF,$6A,$4F,$C5,$F8
	.dc.b	$2F,$A7,$F0,$72,$30,$7C,$9F,$82,$FA,$FF,$07,$23,$07,$CD,$F8,$2F
	.dc.b	$B6,$47,$E8,$7C,$1F,$DE,$F8,$19,$1F,$4D,$FC,$0F,$C1,$FE,$0B,$0E
	.dc.b	$93,$EA,$C3,$0F,$52,$1E,$EC,$3F,$13,$87,$E2,$BE,$06,$4F,$DF,$7F
	.dc.b	$1B,$F0,$5F,$E6,$DE,$1F,$F3,$77,$0C,$AF,$F2,$E7,$C3,$7F,$BF,$5B
	.dc.b	$3F,$EC,$6E,$7F,$F6,$C8,$F9,$0E,$53,$A2,$1F,$E6,$E9,$CB,$4C,$F1
	.dc.b	$FF,$98,$E3,$2D,$8C,$FC,$37,$F9,$7E,$72,$A3,$FD,$C5,$7F,$DD,$09
	.dc.b	$F1,$3E,$5F,$86,$DC,$FF,$9C,$72,$F5,$3E,$11,$8B,$0E,$9F,$0E,$C7
	.dc.b	$FD,$CD,$5B,$AF,$23,$6B,$73,$FD,$CA,$39,$C6,$3E,$1B,$E3,$7C,$14
	.dc.b	$9F,$F6,$DD,$7F,$BA,$DB,$F0,$7F,$2C,$8E,$BE,$17,$E7,$7C,$37,$CF
	.dc.b	$F8,$49,$3C,$A3,$E9,$7C,$2C,$8E,$AF,$D5,$F8,$3E,$71,$E4,$3E,$3E
	.dc.b	$18,$D7,$E0,$47,$CA,$C7,$BB,$E6,$C7,$BB,$C7,$6F,$89,$97,$E0,$7D
	.dc.b	$FF,$C1,$7C,$14,$AF,$6C,$7E,$17,$E0,$3F,$0D,$F0,$92,$7E,$0C,$7E
	.dc.b	$27,$E1,$7F,$15,$F0,$92,$7E,$2C,$7E,$37,$E1,$24,$7C,$87,$F2,$1F
	.dc.b	$09,$D6,$3E,$5E,$1D,$5F,$37,$1B,$1E,$D1,$F4,$31,$EC,$EB,$1F,$C5
	.dc.b	$3E,$DF,$CD,$7F,$B7,$F0,$FC,$4D,$D1,$8E,$7F,$C2,$B6,$93,$AF,$83
	.dc.b	$E7,$7C,$6B,$FD,$CD,$5D,$7E,$0B,$38,$72,$6C,$72,$9D,$D7,$1F,$82
	.dc.b	$CE,$3E,$26,$CF,$F6,$CA,$FD,$15,$77,$F8,$4C,$FC,$2C,$9C,$E5,$E8
	.dc.b	$8E,$3F,$07,$CB,$E1,$79,$E9,$2F,$67,$8E,$3B,$E7,$23,$3A,$78,$DB
	.dc.b	$A3,$3F,$2F,$C2,$C9,$CB,$E8,$F8,$2F,$04,$70,$CB,$C0,$FD,$1C,$BB
	.dc.b	$3F,$4F,$2B,$B7,$B6,$3B,$FC,$1F,$C2,$F8,$29,$18,$3C,$7F,$07,$F1
	.dc.b	$BE,$16,$46,$0F,$3F,$C2,$FC,$A2,$3E,$AF,$85,$F9,$BF,$0B,$F3,$BE
	.dc.b	$0E,$4F,$BA,$3E,$8E,$1D,$23,$F1,$18,$61,$EA,$43,$F1,$98,$7D,$8C
	.dc.b	$3E,$CF,$BF,$F6,$BE,$0A,$57,$CA,$8F,$BB,$F0,$7F,$7B,$E1,$64,$FC
	.dc.b	$E8,$FC,$1F,$E6,$BF,$ED,$A3,$FB,$BF,$07,$E5,$DB,$D7,$FC,$69,$D7
	.dc.b	$C2,$70,$F0,$66,$1F,$6A,$53,$E0,$8F,$B7,$8F,$81,$FD,$26,$3E,$0C
	.dc.b	$3C,$3F,$0B,$27,$94,$79,$3E,$16,$47,$57,$CF,$EB,$7A,$3D,$5F,$4F
	.dc.b	$C1,$4A,$F1,$47,$B3,$E1,$64,$79,$5F,$7F,$C2,$7C,$0F,$85,$F8,$3F
	.dc.b	$09,$2B,$D5,$1D,$5F,$66,$1D,$5F,$E3,$1C,$1F,$D0,$38,$BE,$36,$3D
	.dc.b	$A7,$2F,$67,$E2,$7C,$27,$CE,$F8,$29,$18,$3C,$E7,$62,$1D,$B4,$E9
	.dc.b	$7D,$A3,$D4,$7C,$FD,$3E,$CF,$C1,$7D,$AF,$82,$93,$F4,$E3,$EE,$FC
	.dc.b	$14,$8F,$AC,$FE,$03,$E1,$3F,$03,$F0,$92,$7E,$D3,$F8,$4F,$84,$FC
	.dc.b	$2F,$C2,$7F,$08,$FC,$D7,$FB,$A5,$7E,$E1,$AF,$C4,$8E,$CF,$C6,$C2
	.dc.b	$E9,$F8,$BC,$7B,$4F,$CA,$C6,$CF,$E3,$F1,$ED,$F0,$92,$BB,$3E,$1F
	.dc.b	$84,$F1,$7C,$54,$9F,$13,$E5,$F8,$AF,$37,$C2,$F9,$FE,$16,$57,$9E
	.dc.b	$3D,$5F,$07,$23,$D4,$FB,$7E,$0F,$A4,$7F,$55,$E1,$FE,$EF,$AF,$F3
	.dc.b	$96,$F7,$FF,$7C,$7F,$BB,$46,$70,$EF,$5C,$71,$F1,$BF,$1F,$E0,$BE
	.dc.b	$51,$1F,$95,$F0,$7F,$37,$E1,$7E,$77,$C2,$C9,$F9,$D1,$F4,$7E,$13
	.dc.b	$E9,$7C,$2C,$9F,$A5,$1F,$57,$E0,$A4,$7D,$57,$EC,$7C,$1F,$58,$FB
	.dc.b	$18,$75,$7D,$D8,$75,$7E,$E6,$1D,$5F,$85,$87,$E0,$7E,$0B,$F0,$5F
	.dc.b	$07,$27,$F0,$51,$D5,$F9,$18,$75,$DF,$E4,$63,$D6,$7E,$66,$3D,$67
	.dc.b	$E6,$63,$F8,$DF,$85,$95,$F8,$D7,$F2,$1F,$0B,$FD,$63,$DB,$FE,$CD
	.dc.b	$F0,$5F,$D5,$EB,$FA,$4C,$60,$FF,$66,$61,$FD,$BD,$FE,$C3,$CB,$FE
	.dc.b	$32,$5E,$F7,$33,$A7,$BD,$D6,$3B,$7C,$0C,$C3,$AC,$78,$58,$78,$0F
	.dc.b	$27,$C2,$F2,$7C,$77,$D5,$F2,$DF,$A7,$E1,$64,$7A,$1F,$67,$C2,$FB
	.dc.b	$7E,$13,$DD,$F0,$92,$BD,$B1,$F0,$7E,$13,$E1,$7C,$24,$9F,$83,$1F
	.dc.b	$17,$E0,$BE,$37,$C1,$49,$F8,$B1,$D5,$F8,$F8,$75,$7F,$C6,$5B,$31
	.dc.b	$ED,$1F,$33,$1E,$CE,$B1,$FA,$5F,$05,$F4,$FE,$0A,$4F,$D2,$8F,$AD
	.dc.b	$F0,$7F,$5F,$E1,$24,$FD,$68,$CB,$F6,$3E,$0F,$EE,$18,$7C,$71,$79
	.dc.b	$BE,$D3,$79,$BE,$D3,$67,$FC,$4F,$AA,$31,$CE,$8C,$4F,$84,$87,$63
	.dc.b	$F1,$3F,$09,$F8,$D2,$3F,$8C,$F8,$1E,$73,$CA,$38,$1F,$23,$E0,$B3
	.dc.b	$87,$83,$F3,$5F,$F3,$0C,$FF,$97,$5F,$ED,$2F,$F8,$F1,$FE,$CA,$FF
	.dc.b	$88,$1A,$EF,$E0,$FF,$C6,$B1,$FE,$87,$8D,$7C,$2E,$0C,$FC,$26,$BE
	.dc.b	$12,$4F,$37,$8F,$BB,$CB,$E0,$B9,$FC,$0C,$AF,$04,$73,$7A,$90,$F4
	.dc.b	$5F,$47,$D5,$77,$8B,$FC,$B5,$97,$F6,$1E,$2F,$F4,$EA,$F7,$FD,$24
	.dc.b	$BF,$77,$C1,$FB,$3E,$12,$47,$C1,$7D,$FF,$07,$F0,$3E,$0E,$4F,$C5
	.dc.b	$7E,$1F,$C1,$7C,$4F,$83,$F8,$BF,$05,$2B,$E5,$47,$47,$E8,$61,$75
	.dc.b	$F3,$E3,$D2,$7E,$A6,$36,$7D,$78,$FD,$0F,$82,$FA,$3F,$05,$2F,$EB
	.dc.b	$47,$D4,$F8,$29,$1F,$65,$FA,$FF,$07,$FE,$6C,$FC,$D7,$FB,$1E,$3F
	.dc.b	$E7,$F6,$F1,$7F,$B0,$9B,$7F,$E7,$E5,$F0,$9C,$7E,$12,$4F,$F9,$92
	.dc.b	$51,$D7,$E1,$3C,$33,$F8,$DC,$AE,$3F,$53,$2F,$17,$C4,$CA,$E4,$F9
	.dc.b	$72,$F3,$7C,$2F,$9F,$E0,$A5,$77,$8E,$7A,$FD,$CC,$AC,$FE,$67,$1E
	.dc.b	$8F,$FA,$E7,$5F,$5F,$8A,$69,$EC,$8F,$57,$C1,$7C,$52,$3E,$DF,$85
	.dc.b	$F9,$1F,$0B,$23,$E0,$3F,$2F,$E1,$3E,$67,$C2,$49,$F8,$8F,$CF,$F7
	.dc.b	$FE,$87,$C0,$49,$F9,$0F,$D3,$F7,$FA,$47,$9F,$0B,$C9,$F5,$71,$E9
	.dc.b	$3E,$BC,$6C,$FD,$9C,$7E,$E7,$C0,$4A,$FA,$8F,$DF,$F8,$4F,$C0,$7C
	.dc.b	$24,$9F,$B0,$FE,$0F,$E0,$FF,$09,$F0,$92,$7E,$E3,$F8,$7F,$84,$FC
	.dc.b	$47,$C2,$7F,$AD,$3F,$35,$FE,$BA,$7F,$5E,$E4,$47,$5B,$FA,$18,$DD
	.dc.b	$7F,$1D,$97,$59,$FA,$99,$59,$FC,$96,$5D,$7E,$12,$57,$FA,$C3,$3F
	.dc.b	$EC,$FD,$28,$FF,$68,$C7,$F6,$86,$FC,$27,$97,$D5,$91,$C6,$3D,$1F
	.dc.b	$0B,$23,$A3,$EB,$F8,$4F,$67,$C2,$49,$F0,$3E,$FF,$81,$F8,$17,$D2
	.dc.b	$5A,$FF,$B9,$F2,$FE,$FC,$D9,$ED,$3A,$CF,$D1,$DA,$BF,$F6,$44,$F8
	.dc.b	$8F,$76,$5F,$37,$E0,$A4,$7C,$17,$E8,$7C,$0F,$8A,$3E,$26,$DE,$3C
	.dc.b	$FC,$11,$CC,$F0,$1F,$1B,$6F,$1C,$79,$F0,$FB,$5F,$05,$F6,$FE,$0E
	.dc.b	$5F,$CF,$8F,$BD,$F0,$9F,$7F,$E0,$E4,$FD,$38,$FC,$17,$C2,$48,$FA
	.dc.b	$CF,$E1,$7D,$FF,$C3,$7C,$24,$9F,$B4,$FE,$27,$E1,$3A,$C7,$DD,$C3
	.dc.b	$AB,$E8,$E2,$F8,$D8,$F5,$8C,$71,$32,$C6,$DF,$47,$3F,$CA,$7C,$27
	.dc.b	$E5,$7E,$12,$61,$F8,$98,$FC,$C7,$C2,$48,$FC,$63,$F8,$B7,$C1,$E6
	.dc.b	$FF,$64,$C1,$CF,$F6,$23,$FB,$2E,$A6,$93,$E2,$39,$57,$0B,$56,$FE
	.dc.b	$13,$38,$73,$6C,$EB,$0F,$ED,$4F,$8B,$E0,$7C,$CF,$C0,$CA,$E3,$D7
	.dc.b	$1F,$16,$3C,$9F,$27,$C2,$DC,$7C,$AD,$CF,$1F,$05,$32,$FD,$56,$3C
	.dc.b	$ED,$3C,$7E,$17,$97,$C2,$4B,$DE,$7F,$0E,$3B,$7C,$2F,$7F,$84,$F0
	.dc.b	$7C,$24,$BD,$9E,$D8,$F2,$7C,$17,$97,$E0,$A4,$EB,$83,$E6,$4F,$78
	.dc.b	$FA,$78,$DD,$BC,$D9,$77,$9E,$1B,$33,$F0,$47,$DA,$CF,$E1,$7C,$17
	.dc.b	$C3,$F8,$29,$3C,$23,$A6,$DE,$B8,$E7,$9F,$B1,$EB,$7D,$33,$F8,$D8
	.dc.b	$F8,$67,$F0,$FF,$07,$F3,$FE,$12,$47,$92,$3E,$97,$C1,$C8,$F3,$BF
	.dc.b	$57,$E1,$3E,$B7,$BF,$27,$D6,$FD,$9F,$7F,$EE,$7E,$6B,$FD,$60,$FE
	.dc.b	$AD,$3F,$A5,$6E,$7F,$83,$7F,$AE,$F0,$B1,$F0,$57,$E7,$1F,$03,$F2
	.dc.b	$2F,$C0,$FE,$83,$0F,$03,$F4,$EF,$FB,$DB,$7E,$17,$C6,$79,$3E,$17
	.dc.b	$CB,$8F,$9B,$E0,$E5,$71,$8F,$4F,$C1,$C8,$E8,$FB,$3E,$0F,$DB,$F0
	.dc.b	$3E,$EF,$81,$95,$E1,$8E,$AF,$8F,$0E,$AF,$C1,$C3,$A3,$F0,$F0,$BB
	.dc.b	$FC,$3C,$7E,$4F,$C0,$CA,$F6,$3F,$33,$E0,$7E,$6F,$C1,$FC,$EF,$83
	.dc.b	$95,$F0,$63,$2F,$C3,$F8,$3F,$A6,$61,$FA,$51,$71,$FA,$4D,$EB,$FE
	.dc.b	$7F,$C4,$8C,$57,$F0,$48,$C3,$DC,$87,$42,$D7,$47,$EA,$7C,$1F,$E1
	.dc.b	$3E,$06,$46,$0D,$E1,$E0,$C3,$BF,$84,$C5,$F0,$38,$7C,$37,$DC,$FC
	.dc.b	$67,$C0,$FE,$43,$E0,$7F,$23,$F0,$52,$7F,$09,$1F,$C3,$FF,$35,$FF
	.dc.b	$6E,$92,$31,$F0,$79,$F8,$8D,$7C,$14,$9F,$DF,$57,$F7,$7D,$0F,$F6
	.dc.b	$8D,$BF,$B2,$B6,$7F,$94,$B6,$77,$EF,$F8,$66,$1D,$0F,$09,$DB,$0F
	.dc.b	$03,$FD,$2E,$BF,$DD,$6E,$4F,$1F,$BF,$EB,$23,$E6,$F8,$4F,$77,$C2
	.dc.b	$48,$F4,$BF,$07,$E1,$3E,$17,$C2,$49,$F6,$BF,$17,$E0,$BE,$37,$C1
	.dc.b	$C9,$F8,$2F,$C9,$F8,$0E,$91,$E5,$C2,$F5,$7F,$3D,$D1,$97,$5A,$F5
	.dc.b	$65,$6B,$59,$7D,$3F,$81,$FA,$9F,$01,$2F,$E7,$C7,$D7,$F8,$09,$1F
	.dc.b	$49,$FB,$5F,$07,$F6,$FE,$0F,$EE,$7C,$0C,$AF,$AF,$1F,$7F,$E0,$BF
	.dc.b	$01,$F0,$12,$7E,$DC,$74,$7E,$76,$17,$4F,$56,$3D,$67,$F0,$38,$F5
	.dc.b	$7E,$A6,$3F,$D6,$7E,$6F,$FB,$87,$4F,$F8,$99,$B3,$FD,$A9,$BD,$27
	.dc.b	$FB,$9F,$17,$FB,$7F,$2F,$91,$9F,$87,$0F,$C5,$E7,$E1,$C7,$F0,$59
	.dc.b	$F8,$6F,$E7,$67,$E1,$F8,$0F,$17,$C1,$CB,$FE,$A9,$A7,$FB,$82,$FF
	.dc.b	$B8,$2F,$83,$F4,$7C,$14,$AD,$C7,$AF,$E0,$BD,$9E,$EC,$9E,$71,$EF
	.dc.b	$F8,$09,$1D,$9F,$85,$EF,$7C,$3B,$E9,$3F,$E2,$3C,$7F,$BA,$BD,$8E
	.dc.b	$67,$9B,$0E,$CE,$F0,$F9,$9E,$F4,$8F,$5B,$F3,$FE,$07,$E8,$7C,$17
	.dc.b	$D1,$F8,$19,$5F,$02,$3E,$A7,$C1,$7D,$5F,$81,$93,$F1,$23,$EC,$7C
	.dc.b	$14,$8F,$8E,$FD,$BF,$81,$EB,$1F,$2B,$0E,$AF,$AF,$0E,$AF,$CE,$C3
	.dc.b	$AB,$EF,$C3,$F0,$9F,$05,$23,$E9,$BF,$87,$F8,$1F,$C4,$7C,$17,$F8
	.dc.b	$87,$87,$FD,$7A,$7F,$AF,$B8,$51,$9F,$87,$D7,$C1,$CC,$7F,$C2,$17
	.dc.b	$FD,$AB,$B7,$FC,$EA,$BE,$0F,$A1,$27,$7F,$01,$E0,$78,$FC,$47,$84
	.dc.b	$C3,$FE,$59,$AE,$8F,$F9,$09,$FE,$D6,$E7,$0F,$23,$E2,$F8,$2F,$51
	.dc.b	$1F,$2F,$C1,$74,$ED,$FF,$20,$5F,$BB,$0C,$67,$E8,$D3,$D9,$87,$53
	.dc.b	$DA,$FC,$5F,$82,$F8,$DF,$07,$27,$E0,$C7,$C9,$F8,$49,$1F,$11,$F9
	.dc.b	$9F,$09,$F3,$7E,$12,$4F,$C8,$7E,$87,$C0,$7D,$1F,$84,$FA,$5F,$09
	.dc.b	$2B,$E6,$C7,$47,$E2,$61,$77,$F5,$E3,$D2,$7E,$46,$36,$7D,$F8,$FD
	.dc.b	$DF,$84,$95,$F6,$1F,$C0,$7C,$27,$E0,$7E,$13,$F0,$5F,$07,$2B,$EE
	.dc.b	$C7,$E1,$7E,$0B,$F0,$DF,$07,$27,$F0,$31,$FE,$69,$FC,$D7,$FD,$05
	.dc.b	$1F,$E9,$17,$FD,$9E,$FF,$9E,$9B,$EA,$1D,$E9,$F5,$30,$F0,$57,$E0
	.dc.b	$71,$F2,$3D,$8F,$C7,$E3,$E1,$7F,$0D,$8F,$87,$E0,$64,$7F,$98,$F1
	.dc.b	$FE,$CC,$7F,$58,$7C,$9F,$11,$E7,$F8,$39,$1B,$8F,$57,$C1,$FA,$FE
	.dc.b	$0E,$4F,$38,$F7,$7C,$0F,$BF,$E0,$64,$F7,$8F,$85,$7D,$23,$E1,$61
	.dc.b	$87,$91,$E5,$C3,$AB,$F1,$F0,$F9,$5F,$03,$F2,$FE,$06,$4F,$AE,$39
	.dc.b	$3D,$0F,$75,$E3,$87,$FD,$39,$A3,$0E,$B1,$8C,$7D,$B7,$F1,$7E,$0F
	.dc.b	$EC,$11,$F9,$1F,$07,$F6,$FE,$0A,$47,$CB,$7E,$F7,$C1,$FD,$FF,$80
	.dc.b	$93,$F3,$DF,$C1,$7C,$07,$E0,$FD,$F9,$3F,$4D,$FC,$37,$C0,$74,$8F
	.dc.b	$C3,$61,$74,$F4,$E3,$D2,$7F,$15,$8D,$9F,$6E,$3F,$8F,$F8,$29,$5F
	.dc.b	$7D,$FC,$97,$C1,$FF,$AE,$FF,$35,$FE,$BE,$7F,$B4,$09,$E4,$E7,$E0
	.dc.b	$E5,$FF,$AA,$67,$FA,$B4,$A3,$FD,$A3,$1F,$EC,$CC,$7F,$D9,$0F,$F5
	.dc.b	$98,$E1,$E6,$77,$3C,$1F,$05,$E1,$BF,$17,$C4,$4A,$ED,$1E,$07,$FD
	.dc.b	$ED,$7E,$07,$BE,$3E,$9F,$84,$F5,$7C,$0C,$9F,$33,$ED,$F8,$2F,$77
	.dc.b	$C2,$7B,$FE,$12,$57,$AE,$32,$FB,$7E,$13,$E2,$18,$7C,$31,$77,$F6
	.dc.b	$B7,$4F,$85,$E4,$35,$4E,$6F,$E2,$B6,$74,$D9,$DC,$F8,$CF,$9D,$F0
	.dc.b	$9F,$57,$E0,$64,$60,$D1,$E4,$2E,$3F,$43,$E0,$BA,$61,$D7,$E0,$7E
	.dc.b	$F7,$C1,$7D,$FF,$84,$98,$7D,$B8,$FC,$17,$C2,$7E,$0F,$E1,$24,$FD
	.dc.b	$F8,$FC,$37,$C2,$48,$FC,$13,$F8,$9F,$83,$FC,$14,$34,$70,$FC,$4F
	.dc.b	$CD,$5E,$FE,$A7,$34,$E3,$56,$A3,$FD,$EC,$D7,$A5,$F3,$F7,$FC,$F1
	.dc.b	$F9,$3D,$3C,$EF,$E1,$B4,$F3,$3F,$88,$D2,$E5,$F9,$4D,$3C,$58,$F9
	.dc.b	$7E,$0A,$DF,$8F,$3F,$EE,$8F,$8B,$E1,$33,$AF,$9A,$AD,$3E,$12,$FC
	.dc.b	$24,$C3,$D3,$1A,$F8,$49,$1E,$C7,$8F,$C1,$F2,$F8,$4E,$7F,$05,$2B
	.dc.b	$E0,$C7,$6B,$E9,$1F,$08,$F3,$61,$97,$A1,$0D,$67,$F3,$31,$EF,$1F
	.dc.b	$33,$E0,$BD,$3F,$05,$23,$07,$D0,$F8,$2F,$69,$1F,$A7,$F0,$7F,$03
	.dc.b	$E1,$24,$7D,$67,$E1,$FC,$27,$C4,$F8,$49,$3F,$69,$F8,$FF,$09,$CE
	.dc.b	$3C,$BC,$5F,$EC,$4E,$7F,$C1,$ED,$F8,$2C,$FE,$6E,$7D,$AF,$79,$DB
	.dc.b	$DE,$0C,$FE,$9F,$C2,$4B,$FC,$4B,$F5,$BE,$0F,$C1,$7C,$32,$BD,$1F
	.dc.b	$95,$CF,$A3,$F9,$7C,$EE,$3F,$83,$CB,$EF,$FC,$24,$8F,$CA,$BF,$84
	.dc.b	$F7,$BF,$BA,$DF,$F0,$27,$FB,$C9,$3E,$23,$5F,$0B,$BF,$88,$E1,$F0
	.dc.b	$93,$1F,$E5,$5B,$7C,$2C,$3F,$B9,$35,$EC,$48,$DE,$3D,$FE,$07,$C5
	.dc.b	$94,$9C,$9F,$AD,$AF,$1E,$5E,$18,$CF,$1A,$7F,$DC,$04,$CF,$D9,$F0
	.dc.b	$3E,$DF,$85,$97,$C2,$93,$F0,$7E,$1A,$4E,$5F,$89,$F0,$BF,$17,$E1
	.dc.b	$A4,$F1,$7E,$47,$C2,$FC,$9F,$85,$93,$D5,$F9,$9F,$07,$D2,$3F,$37
	.dc.b	$97,$FB,$59,$EC,$78,$F0,$EA,$EB,$0B,$C5,$F2,$F1,$FA,$DF,$07,$2B
	.dc.b	$D4,$FD,$9F,$83,$FB,$5F,$07,$27,$DC,$FD,$DF,$85,$FB,$DF,$0B,$27
	.dc.b	$E1,$3F,$81,$F8,$5F,$C1,$7C,$2F,$F0,$6F,$97,$FF,$64,$9F,$EC,$AD
	.dc.b	$D1,$D6,$FD,$78,$DD,$7F,$01,$97,$59,$F7,$E5,$67,$F0,$99,$75,$F8
	.dc.b	$49,$5F,$6B,$34,$78,$7E,$13,$A7,$67,$E4,$F3,$B7,$F8,$3C,$FA,$4F
	.dc.b	$E5,$F3,$B3,$F8,$7C,$BD,$5F,$09,$EB,$F8,$69,$3D,$E3,$DD,$F0,$D2
	.dc.b	$3C,$4F,$C1,$F8,$6F,$85,$F0,$DF,$0F,$E0,$25,$79,$E3,$E3,$7C,$34
	.dc.b	$8F,$53,$F2,$7E,$1B,$A4,$7F,$1A,$E2,$FF,$69,$B7,$1E,$CF,$33,$E1
	.dc.b	$61,$D9,$E1,$87,$D3,$F8,$5F,$A9,$F0,$B2,$7E,$44,$7D,$7F,$85,$91
	.dc.b	$F2,$DF,$B5,$F0,$BF,$6F,$E1,$24,$FC,$F7,$EF,$7C,$2F,$DF,$F8,$2F
	.dc.b	$C0,$7C,$1C,$AF,$A9,$1D,$1F,$56,$17,$AF,$F8,$FE,$EC,$BA,$D7,$BF
	.dc.b	$2B,$5A,$CB,$F1,$9F,$07,$2B,$F0,$0F,$E3,$FE,$0F,$FA,$AF,$E6,$BF
	.dc.b	$D9,$EF,$F9,$5B,$7A,$BF,$C2,$AD,$BF,$EE,$0F,$FB,$23,$CC,$3F,$EC
	.dc.b	$8A,$E3,$F0,$7D,$0E,$BF,$09,$DB,$E0,$7B,$FC,$0C,$C7,$F8,$5E,$1F
	.dc.b	$D8,$DB,$F0,$DE,$42,$3A,$F8,$6E,$8F,$0C,$6B,$FD,$62,$BF,$B8,$B7
	.dc.b	$E0,$FD,$A4,$7B,$FC,$37,$47,$86,$17,$4F,$83,$8F,$47,$D7,$3D,$CF
	.dc.b	$3E,$9F,$23,$E1,$7E,$4F,$C3,$49,$F5,$C7,$CC,$F8,$6F,$9B,$F0,$D2
	.dc.b	$7D,$F1,$F4,$3E,$06,$47,$C2,$7E,$9F,$C0,$FD,$4F,$81,$FA,$BF,$03
	.dc.b	$2B,$E3,$C7,$57,$E4,$E1,$D5,$F8,$58,$74,$7E,$26,$17,$4F,$BB,$8F
	.dc.b	$E0,$3E,$1B,$F0,$3F,$0D,$2F,$E9,$C7,$E1,$3E,$16,$47,$D6,$7F,$0F
	.dc.b	$F0,$BF,$88,$F8,$59,$3F,$69,$FF,$0F,$FE,$6B,$FD,$72,$3F,$C6,$CF
	.dc.b	$FB,$50,$9F,$09,$AF,$84,$EB,$3F,$67,$0B,$A7,$E4,$B1,$EB,$3F,$77
	.dc.b	$1B,$3F,$E0,$0C,$7B,$76,$9F,$EC,$3D,$DF,$E6,$C6,$DF,$8B,$E1,$3C
	.dc.b	$BF,$0D,$FE,$38,$D5,$F3,$8D,$FD,$3E,$E4,$C3,$FB,$1D,$6A,$78,$56
	.dc.b	$E7,$FB,$1E,$3F,$EC,$36,$FC,$3E,$AA,$47,$2A,$E0,$FF,$8C,$EF,$A3
	.dc.b	$99,$DB,$99,$DE,$BF,$E6,$0A,$DC,$E6,$B7,$39,$AD,$E9,$FE,$AD,$BD
	.dc.b	$CE,$6B,$EA,$FC,$34,$C3,$D2,$FD,$8F,$87,$E7,$87,$1C,$79,$D3,$1E
	.dc.b	$73,$F3,$7E,$1F,$73,$9A,$DD,$6A,$3F,$03,$F0,$BF,$84,$F8,$5F,$C2
	.dc.b	$FC,$2C,$9F,$93,$1F,$88,$F8,$7F,$C4,$FC,$3C,$9F,$9B,$1F,$8C,$F8
	.dc.b	$7F,$C6,$FC,$3C,$9F,$A3,$1F,$90,$F8,$4F,$C8,$FC,$24,$9F,$AB,$1D
	.dc.b	$5F,$AF,$87,$57,$FC,$BF,$CA,$C7,$B4,$7D,$CC,$7B,$3A,$C7,$F3,$BF
	.dc.b	$09,$23,$F0,$2F,$E8,$3E,$13,$F4,$3F,$0F,$27,$F0,$AF,$FA,$0B,$D1
	.dc.b	$FF,$4C,$C7,$F9,$B9,$FE,$E2,$FF,$8B,$1A,$E7,$E1,$F8,$7C,$3F,$1F
	.dc.b	$86,$93,$B7,$A7,$C3,$F7,$8F,$5E,$17,$B3,$FC,$CB,$8F,$81,$B8,$F8
	.dc.b	$C9,$5E,$1C,$7C,$DF,$09,$E7,$F8,$79,$3E,$58,$F5,$7C,$24,$8F,$43
	.dc.b	$ED,$F8,$5F,$77,$C3,$C9,$F6,$3F,$07,$E1,$FE,$17,$C3,$49,$F8,$0F
	.dc.b	$C5,$F8,$7E,$91,$F1,$70,$BA,$FA,$B1,$E9,$3F,$27,$1B,$3E,$EC,$7E
	.dc.b	$77,$C2,$FC,$FF,$84,$97,$F3,$63,$9F,$87,$FE,$F0,$E1,$67,$8A,$FF
	.dc.b	$66,$47,$FB,$B1,$EC,$7A,$88,$7D,$9C,$3E,$E7,$C2,$7D,$DF,$84,$93
	.dc.b	$F6,$E3,$F0,$1F,$0D,$F8,$1F,$87,$93,$F7,$E3,$F0,$9F,$0D,$23,$F0
	.dc.b	$4F,$E1,$FE,$1B,$FD,$93,$F9,$AF,$F7,$1F,$48,$D2,$7C,$0F,$6B,$FC
	.dc.b	$56,$3D,$9F,$C7,$63,$D5,$FC,$86,$37,$8B,$F0,$B9,$76,$F8,$69,$5F
	.dc.b	$E7,$EC,$BF,$BE,$70,$A3,$FD,$E1,$1F,$DB,$AF,$FD,$62,$FF,$79,$8F
	.dc.b	$F6,$4A,$E0,$F4,$9E,$AF,$87,$F5,$FC,$34,$AE,$CF,$BB,$E1,$BD,$FF
	.dc.b	$07,$27,$C4,$FC,$2F,$84,$E5,$1E,$5C,$3A,$BE,$7C,$3A,$3E,$2C,$2E
	.dc.b	$5E,$6C,$7E,$5F,$C0,$FC,$CF,$82,$97,$EF,$8F,$9F,$F0,$5F,$43,$E0
	.dc.b	$A4,$FC,$38,$FA,$7F,$07,$F5,$3E,$12,$4F,$C7,$8F,$AF,$ED,$C8,$F9
	.dc.b	$4F,$DA,$F8,$4E,$B1,$F3,$70,$EA,$FC,$8C,$3A,$BF,$47,$0E,$AF,$CC
	.dc.b	$C3,$F0,$7F,$07,$23,$EB,$3F,$86,$F8,$3E,$B1,$F4,$B0,$EB,$B7,$D6
	.dc.b	$C7,$AC,$FD,$6C,$7A,$CF,$DA,$C7,$F1,$FF,$07,$2B,$F0,$4F,$E0,$54
	.dc.b	$FF,$BC,$3F,$35,$FE,$E2,$7F,$45,$8F,$F6,$2A,$F8,$5E,$13,$B6,$F4
	.dc.b	$7E,$2F,$BD,$5C,$6F,$AC,$E6,$3F,$BC,$CE,$BE,$17,$8D,$23,$0F,$83
	.dc.b	$E1,$73,$1D,$3E,$16,$DF,$FB,$5E,$BC,$B1,$8F,$85,$CF,$C0,$EB,$39
	.dc.b	$5E,$A8,$E3,$94,$8F,$6B,$D3,$2E,$BF,$0B,$27,$E0,$BE,$0F,$85,$F0
	.dc.b	$FC,$2C,$9F,$8A,$F9,$3E,$13,$CB,$F0,$72,$7E,$4B,$E8,$F8,$4F,$4F
	.dc.b	$C0,$49,$F9,$AF,$B3,$E0,$BA,$47,$C8,$3E,$27,$0F,$FB,$7D,$C7,$17
	.dc.b	$CE,$CB,$C1,$1D,$88,$6A,$58,$F8,$63,$ED,$FC,$07,$CA,$23,$F7,$BE
	.dc.b	$07,$E6,$FC,$1C,$8F,$C0,$BF,$43,$E1,$3E,$8F,$C1,$FD,$2F,$83,$95
	.dc.b	$F8,$68,$FA,$BF,$03,$23,$F1,$2F,$DA,$FC,$D7,$FA,$C7,$5F,$E0,$F1
	.dc.b	$ED,$5E,$77,$B3,$CC,$FC,$8E,$1D,$9F,$66,$1C,$FE,$0B,$A7,$C1,$49
	.dc.b	$FE,$39,$DC,$C7,$F2,$99,$7F,$72,$8F,$CF,$65,$E0,$6D,$7E,$7B,$4F
	.dc.b	$6B,$E0,$21,$E0,$3A,$DF,$4F,$84,$F6,$7C,$24,$9C,$9F,$9B,$C5,$3F
	.dc.b	$07,$E0,$FE,$17,$C2,$4A,$DF,$32,$FF,$DB,$B9,$FA,$67,$97,$BD,$F2
	.dc.b	$48,$F0,$9F,$EB,$93,$BA,$E5,$F0,$3F,$3C,$95,$BF,$7F,$C4,$F1,$C7
	.dc.b	$C4,$FE,$33,$0F,$0E,$BF,$23,$E0,$BB,$CE,$73,$FB,$51,$F6,$FE,$12
	.dc.b	$4F,$8D,$FB,$DF,$09,$F7,$FE,$0E,$4F,$9D,$FC,$17,$C2,$7E,$0F,$E0
	.dc.b	$FF,$09,$F0,$92,$BD,$91,$F8,$7F,$82,$FC,$47,$C1,$C9,$F8,$11,$D1
	.dc.b	$F4,$61,$78,$FE,$16,$FF,$EE,$B9,$E6,$7C,$7C,$7B,$3E,$FC,$7F,$29
	.dc.b	$F0,$7F,$95,$F8,$39,$3F,$36,$3F,$8A,$FE,$02,$ED,$7C,$0F,$F1,$4E
	.dc.b	$9F,$EF,$47,$FB,$AE,$75,$1C,$27,$77,$F7,$AB,$FD,$DC,$3E,$17,$55
	.dc.b	$27,$F1,$37,$C9,$B7,$C9,$FE,$09,$19,$6C,$67,$5F,$C4,$4E,$5B,$19
	.dc.b	$E6,$FF,$00,$4E,$5B,$19,$A4,$E5,$B1,$98,$C7,$53,$2F,$F0,$5D,$7F
	.dc.b	$39,$6C,$67,$E1,$7E,$37,$C4,$7B,$6B,$CF,$9D,$CF,$F3,$18,$7F,$83
	.dc.b	$DF,$3C,$67,$3F,$2B,$F3,$FE,$13,$E8,$7C,$1F,$D1,$F8,$39,$8F,$CD
	.dc.b	$8F,$A9,$F0,$BF,$56,$E4,$FD,$18,$FB,$1F,$07,$23,$EA,$3F,$6F,$E1
	.dc.b	$3E,$E7,$BF,$27,$EC,$3F,$7F,$E0,$3A,$C7,$DB,$C3,$AB,$FC,$D3,$66
	.dc.b	$3D,$63,$18,$DD,$BD,$F9,$7E,$27,$DF,$95,$F8,$57,$F1,$9F,$01,$F8
	.dc.b	$DF,$84,$FC,$77,$C2,$4A,$FC,$54,$7E,$47,$E1,$3F,$25,$F0,$92,$7F
	.dc.b	$1D,$1F,$95,$F8,$4F,$ED,$3F,$9A,$FF,$BB,$0F,$F7,$6E,$E8,$ED,$5F
	.dc.b	$95,$C7,$B3,$EB,$C7,$B3,$F9,$9C,$7B,$3E,$FC,$7A,$FC,$14,$8F,$E3
	.dc.b	$79,$7F,$80,$76,$FE,$28,$FF,$82,$63,$3F,$05,$E4,$F8,$29,$18,$38
	.dc.b	$FC,$2F,$A0,$8F,$4F,$85,$F5,$FC,$27,$B3,$E1,$24,$F8,$23,$DF,$87
	.dc.b	$48,$FE,$CF,$8B,$1C,$C7,$62,$1A,$C7,$E3,$E3,$F2,$3E,$02,$4F,$B1
	.dc.b	$F9,$7F,$09,$CA,$3A,$DF,$7D,$3D,$F9,$7A,$27,$A1,$DA,$3E,$1E,$1F
	.dc.b	$53,$E0,$24,$FC,$87,$EB,$FC,$27,$D8,$F8,$4F,$B3,$F0,$92,$BE,$6C
	.dc.b	$7D,$CF,$83,$91,$F4,$1F,$BF,$F0,$7F,$80,$F7,$FF,$03,$EF,$CA,$FA
	.dc.b	$B1,$D5,$FA,$F8,$75,$7E,$3E,$1D,$1F,$93,$85,$D3,$ED,$E3,$F8,$CF
	.dc.b	$81,$95,$F8,$07,$F1,$FF,$05,$F9,$0F,$83,$FF,$1C,$FE,$6B,$FE,$00
	.dc.b	$AF,$F1,$F3,$7C,$E7,$3A,$B7,$E8,$4B,$FE,$CB,$7F,$DE,$7C,$DF,$F2
	.dc.b	$72,$F8,$4E,$BF,$07,$DB,$E1,$25,$F0,$8F,$13,$CA,$FC,$66,$1F,$F1
	.dc.b	$5D,$F5,$7C,$B8,$78,$5C,$E1,$EA,$F1,$64,$79,$1F,$6F,$C1,$7B,$BE
	.dc.b	$0A,$4F,$A1,$F8,$3F,$05,$F0,$BE,$12,$4F,$B1,$F8,$BF,$09,$F1,$BE
	.dc.b	$12,$4F,$C0,$7E,$4F,$C2,$74,$8F,$3E,$17,$9F,$FC,$01,$B7,$F8,$0B
	.dc.b	$7F,$8B,$97,$6B,$F6,$BD,$9F,$97,$F0,$3F,$50,$8F,$CE,$F8,$4E,$95
	.dc.b	$F3,$B1,$B3,$F6,$35,$E9,$97,$83,$1B,$7F,$53,$5F,$BF,$F0,$13,$0F
	.dc.b	$B2,$FE,$0B,$E0,$7F,$07,$F0,$7F,$84,$F8,$49,$5F,$7A,$3F,$0F,$F0
	.dc.b	$52,$3F,$02,$FE,$2B,$E0,$FF,$D0,$7F,$9A,$FF,$BA,$49,$F1,$98,$F8
	.dc.b	$5E,$B7,$F8,$8C,$7A,$BF,$7B,$1E,$8F,$E0,$31,$BB,$FC,$AC,$BA,$FC
	.dc.b	$24,$AF,$F1,$AE,$7F,$D7,$0F,$F7,$36,$FF,$E9,$F7,$FA,$C3,$7E,$27
	.dc.b	$CA,$79,$BE,$13,$CF,$F0,$72,$F9,$BE,$AF,$84,$F5,$FC,$0F,$B3,$E0
	.dc.b	$A5,$78,$23,$9B,$E2,$C2,$BF,$E8,$2C,$BF,$DE,$8F,$42,$19,$C3,$B3
	.dc.b	$EA,$F8,$2F,$92,$47,$DB,$F0,$5F,$33,$E1,$24,$7C,$07,$E7,$FC,$27
	.dc.b	$D0,$F8,$4F,$A3,$F0,$92,$BE,$2C,$7D,$4F,$82,$FA,$BF,$07,$27,$E4
	.dc.b	$C7,$47,$D1,$85,$E5,$FF,$2C,$6A,$CB,$B5,$7D,$0C,$BB,$3E,$FC,$BF
	.dc.b	$03,$F0,$5F,$82,$F8,$39,$3F,$5A,$3A,$72,$FD,$5C,$EC,$FF,$B3,$72
	.dc.b	$EC,$FA,$0E,$C4,$3B,$1F,$80,$F8,$4F,$C7,$91,$FF,$3B,$FB,$9F,$EE
	.dc.b	$97,$FD,$66,$FF,$5F,$BF,$F7,$13,$FD,$7E,$33,$F1,$3F,$ED,$A8,$D4
	.dc.b	$7F,$B7,$DF,$F7,$19,$3D,$ED,$7B,$FB,$F8,$2E,$13,$2B,$C3,$1C,$FE
	.dc.b	$17,$A6,$12,$7C,$B1,$DF,$09,$1E,$87,$AB,$EA,$F8,$4F,$21,$1F,$6F
	.dc.b	$C2,$F9,$FE,$16,$47,$C0,$7D,$5F,$0B,$EB,$F8,$5F,$67,$C2,$CA,$F8
	.dc.b	$B1,$EF,$F8,$2F,$81,$F0,$72,$7E,$4C,$74,$7E,$4E,$17,$AB,$FD,$35
	.dc.b	$C0,$CB,$B5,$7D,$0C,$BB,$3F,$43,$2F,$99,$F0,$7F,$37,$E0,$E4,$FD
	.dc.b	$68,$FF,$39,$75,$FD,$1F,$84,$93,$FE,$6C,$8C,$BF,$73,$E1,$7E,$B1
	.dc.b	$87,$FC,$DB,$17,$4F,$F3,$5B,$79,$FF,$CD,$2D,$9F,$F3,$3B,$7E,$1F
	.dc.b	$FD,$10,$DF,$7C,$EF,$E1,$1C,$0F,$86,$FC,$31,$30,$FC,$7F,$C3,$71
	.dc.b	$BE,$15,$FD,$9E,$7D,$44,$3C,$E7,$FB,$33,$F3,$5F,$EE,$1F,$5A,$F7
	.dc.b	$5E,$EA,$BF,$DA,$F3,$98,$C4,$37,$FF,$69,$4E,$63,$75,$C9,$FF,$6F
	.dc.b	$F6,$BE,$39,$E5,$72,$C3,$FD,$B5,$39,$F1,$B1,$C7,$E1,$8C,$61,$CB
	.dc.b	$0E,$2F,$6C,$F6,$B3,$F1,$3E,$5B,$F5,$46,$31,$F3,$E7,$E5,$AD,$1E
	.dc.b	$FF,$17,$FD,$60,$FF,$BD,$82,$F6,$7B,$C6,$1B,$1C,$23,$98,$FC,$27
	.dc.b	$FC,$D4,$F3,$FC,$DF,$F8,$26,$1F,$EF,$98,$46,$32,$F9,$B5,$C2,$79
	.dc.b	$CF,$F0,$47,$7E,$2F,$8A,$3F,$85,$5F,$37,$F8,$4E,$DF,$3F,$CA,$F8
	.dc.b	$AD,$BF,$95,$18,$DF,$B5,$F8,$F1,$FF,$4A,$BF,$C5,$AB,$E2,$7A,$1F
	.dc.b	$80,$E8,$FC,$9C,$60,$D1,$EF,$F4,$BF,$08,$FF,$19,$19,$3B,$7C,$FB
	.dc.b	$7D,$D5,$E4,$79,$9B,$5F,$88,$76,$B0,$73,$FF,$8C,$DB,$F0,$FF,$82
	.dc.b	$6C,$FE,$77,$B7,$EB,$ED,$EC,$8F,$21,$F7,$35,$F6,$75,$7E,$73,$91
	.dc.b	$18,$9F,$D1,$57,$1C,$3E,$09,$E2,$F8,$8F,$06,$66,$98,$F8,$9C,$FC
	.dc.b	$4E,$BE,$26,$63,$FB,$18,$F2,$F1,$FC,$8C,$7E,$6C,$F9,$4F,$9B,$A7
	.dc.b	$91,$F9,$F9,$F8,$F5,$FD,$B5,$78,$E7,$F7,$99,$F8,$9F,$DF,$67,$E6
	.dc.b	$EC,$F3,$FC,$2C,$9F,$E0,$51,$E2,$E6,$EF,$C1,$F0,$67,$C4,$7C,$8C
	.dc.b	$FC,$0F,$D3,$E0,$63,$DE,$3F,$57,$9F,$8D,$B3,$A7,$F8,$87,$C2,$FC
	.dc.b	$8F,$85,$91,$83,$F6,$18,$76,$7C,$37,$F1,$37,$F8,$51,$E3,$E0,$F9
	.dc.b	$1C,$1F,$0B,$27,$C4,$FD,$42,$57,$F1,$FF,$89,$FA,$FF,$11,$F6,$3C
	.dc.b	$19,$3F,$C8,$E3,$E1,$3F,$86,$CB,$BE,$7E,$1E,$0F,$89,$AF,$C2,$8F
	.dc.b	$C9,$E5,$DA,$FD,$EF,$E1,$32,$FC,$2F,$C4,$7E,$1B,$E2,$25,$7F,$31
	.dc.b	$8E,$F7,$D6,$3E,$89,$F7,$B7,$F6,$BF,$3F,$0E,$DC,$6E,$DF,$81,$1D
	.dc.b	$E7,$F9,$E7,$C3,$7E,$53,$E1,$E4,$60,$F9,$B8,$74,$7E,$6F,$1F,$E8
	.dc.b	$77,$F6,$CF,$83,$3F,$D7,$FC,$3F,$E7,$89,$7F,$D2,$7E,$1F,$F6,$1C
	.dc.b	$3F,$C9,$63,$FA,$46,$DE,$57,$F7,$11,$97,$F0,$AD,$8D,$5F,$F2,$7F
	.dc.b	$88,$F8,$79,$F8,$EF,$E1,$9F,$7E,$BF,$8A,$57,$92,$3F,$AD,$63,$E4
	.dc.b	$7F,$1F,$8F,$93,$E2,$24,$79,$38,$3F,$4F,$9A,$7D,$3F,$19,$E0,$AF
	.dc.b	$E3,$87,$E0,$76,$F8,$79,$F8,$3A,$19,$77,$8F,$4C,$33,$FE,$07,$3F
	.dc.b	$13,$E1,$A6,$1E,$13,$E1,$BF,$23,$E2,$7C,$11,$F4,$8F,$0E,$DF,$C0
	.dc.b	$B5,$FC,$71,$FA,$0C,$BE,$3A,$F8,$9F,$A2,$7D,$2D,$3E,$9F,$17,$D4
	.dc.b	$F8,$99,$8F,$87,$1F,$E6,$CF,$E0,$B1,$F9,$B1,$F9,$18,$F9,$C7,$F3
	.dc.b	$97,$FA,$D5,$F6,$8E,$58,$F8,$5B,$1F,$82,$F8,$9F,$C1,$FC,$4C,$9D
	.dc.b	$C7,$BB,$B3,$F5,$7C,$BF,$76,$7B,$E3,$FC,$B3,$1F,$77,$2F,$D6,$DB
	.dc.b	$C1,$5F,$88,$CF,$F1,$FF,$0D,$F9,$0F,$88,$93,$E4,$8E,$D3,$FC,$67
	.dc.b	$3E,$F3,$FC,$EB,$83,$F5,$7E,$6F,$F5,$3F,$83,$FE,$A1,$F0,$32,$BF
	.dc.b	$A8,$47,$F5,$3F,$86,$FE,$A9,$F0,$D2,$7F,$AA,$47,$67,$E4,$F0,$FF
	.dc.b	$78,$C7,$AF,$A0,$C3,$A5,$7F,$83,$B2,$FE,$C5,$97,$F5,$F9,$FD,$4F
	.dc.b	$C4,$7E,$AB,$DB,$97,$F3,$A3,$B3,$F9,$33,$F8,$27,$3F,$F7,$5C,$3D
	.dc.b	$51,$FD,$9F,$0E,$F3,$FE,$2C,$C7,$C5,$3F,$63,$E2,$7F,$6F,$EE,$48
	.dc.b	$C1,$F8,$33,$B1,$39,$FF,$25,$AF,$F0,$3E,$77,$B9,$FC,$03,$D8,$97
	.dc.b	$F8,$48,$FE,$07,$F1,$32,$3F,$0E,$FF,$75,$F6,$BF,$C3,$D1,$F4,$F5
	.dc.b	$F9,$2F,$E8,$38,$FE,$54,$7E,$FB,$1E,$F3,$D4,$87,$F4,$EA,$FF,$53
	.dc.b	$E7,$FE,$24,$8F,$CF,$47,$F8,$29,$D7,$B3,$E1,$F6,$FE,$5D,$7E,$7C
	.dc.b	$F6,$E1,$DF,$3F,$E3,$64,$BE,$7C,$5E,$9E,$69,$1D,$9F,$67,$3F,$4D
	.dc.b	$7F,$39,$8F,$4A,$FD,$56,$9F,$DF,$32,$FE,$3E,$BE,$27,$E2,$1F,$16
	.dc.b	$FE,$37,$44,$BF,$5B,$F2,$79,$BB,$47,$C7,$3E,$51,$DA,$79,$3F,$09
	.dc.b	$FA,$1F,$0F,$F4,$7E,$1A,$4F,$C6,$7E,$A7,$C3,$F4,$8F,$E2,$47,$F8
	.dc.b	$EF,$3F,$E5,$F9,$FA,$7C,$1F,$F0,$7B,$67,$FC,$68,$DD,$7F,$CB,$78
	.dc.b	$FD,$FF,$88,$FC,$07,$C4,$CB,$FA,$B1,$DF,$BF,$F6,$B1,$DC,$F8,$FF
	.dc.b	$11,$F8,$7F,$89,$91,$F7,$23,$F1,$5D,$FF,$8B,$F8,$99,$3F,$80,$8E
	.dc.b	$9E,$1F,$F3,$88,$E6,$7E,$12,$F9,$BF,$89,$BF,$CA,$7C,$44,$8F,$C4
	.dc.b	$BF,$97,$F8,$8E,$71,$FB,$03,$99,$C4,$F2,$1F,$90,$F8,$BF,$F2,$96
	.dc.b	$1F,$E7,$DE,$2F,$EB,$AF,$F9,$5C,$9C,$5D,$2B,$F9,$0F,$2F,$F8,$EA
	.dc.b	$5F,$0D,$CB,$E0,$E5,$7F,$39,$9C,$B7,$E0,$7B,$FC,$24,$8C,$1F,$BB
	.dc.b	$E6,$FC,$F6,$3F,$94,$3F,$5B,$1F,$CF,$6F,$9B,$FE,$38,$6E,$7E,$8C
	.dc.b	$BD,$7F,$0B,$EC,$F8,$4F,$6F,$C2,$CC,$3C,$91,$EB,$CB,$F4,$07,$F9
	.dc.b	$9B,$1E,$7B,$7B,$71,$F8,$A7,$4B,$F6,$7C,$2F,$C8,$23,$EF,$F8,$5E
	.dc.b	$9A,$FA,$31,$E9,$F0,$1F,$3B,$E0,$E5,$FC,$58,$FA,$3F,$09,$23,$E4
	.dc.b	$3F,$53,$E1,$79,$C7,$EA,$2F,$9B,$C4,$87,$C3,$F8,$4F,$B7,$F0,$BF
	.dc.b	$73,$E1,$64,$FD,$38,$E6,$FE,$66,$B9,$BF,$2F,$1F,$C1,$9F,$84,$BF
	.dc.b	$C2,$FC,$2F,$E1,$BE,$06,$57,$DD,$8E,$7E,$8F,$F9,$37,$8F,$FC,$1F
	.dc.b	$3F,$81,$F8,$1F,$C6,$9F,$E9,$FF,$CE,$3F,$C5,$5C,$BF,$E8,$27,$FD
	.dc.b	$A5,$D3,$FE,$2C,$6C,$67,$E1,$B8,$7C,$37,$6F,$63,$FC,$7D,$97,$6C
	.dc.b	$BF,$05,$97,$67,$FC,$11,$8F,$67,$FD,$64,$BE,$1B,$C5,$F0,$B2,$77
	.dc.b	$1D,$74,$FF,$1B,$D7,$1E,$2F,$F5,$6E,$1D,$6F,$FD,$0F,$39,$1F,$0B
	.dc.b	$ED,$CA,$4F,$8A,$3A,$3F,$E1,$59,$FF,$09,$3C,$B3,$F8,$87,$C5,$F8
	.dc.b	$2F,$8D,$F0,$B2,$7D,$6F,$C9,$F8,$4E,$79,$FF,$AF,$30,$E7,$A7,$C9
	.dc.b	$C7,$9E,$1F,$0B,$1F,$A1,$F0,$BF,$47,$E1,$24,$FC,$68,$E7,$DD,$F4
	.dc.b	$EB,$C5,$3C,$4F,$97,$F0,$BF,$67,$E1,$24,$7C,$E7,$EE,$7C,$2F,$28
	.dc.b	$F9,$FF,$0B,$F7,$C8,$FD,$4F,$85,$FC,$17,$C0,$FE,$0F,$E1,$64,$FD
	.dc.b	$88,$E5,$C2,$E1,$61,$CA,$BE,$8B,$F8,$BC,$3F,$19,$F0,$B2,$7F,$02
	.dc.b	$FF,$BB,$7F,$39,$FE,$F3,$1F,$9F,$DF,$F8,$63,$FD,$4C,$C1,$F8,$6A
	.dc.b	$ED,$C7,$F9,$3C,$3B,$4F,$F9,$4E,$FF,$BC,$93,$E1,$3B,$7C,$2F,$C1
	.dc.b	$5E,$C3,$DC,$5F,$F9,$AE,$73,$EC,$62,$3F,$C5,$D4,$8C,$65,$FE,$23
	.dc.b	$BB,$E1,$72,$C3,$AF,$BB,$8C,$51,$E0,$F8,$0C,$36,$3C,$19,$2B,$CC
	.dc.b	$F9,$63,$39,$70,$8F,$14,$70,$C7,$E5,$7C,$34,$8F,$84,$FC,$DF,$86
	.dc.b	$E7,$3C,$49,$3F,$1F,$E1,$BE,$97,$17,$D3,$F8,$69,$3F,$2E,$3E,$FF
	.dc.b	$0F,$F9,$02,$3F,$DE,$6D,$E0,$FF,$57,$B7,$83,$E5,$F3,$21,$F0,$BF
	.dc.b	$7F,$82,$61,$F5,$DF,$C1,$71,$76,$9F,$C2,$7C,$34,$AF,$B9,$1F,$87
	.dc.b	$F8,$69,$1F,$7D,$FC,$57,$C2,$F1,$8F,$C1,$57,$17,$CF,$5F,$8F,$E0
	.dc.b	$FC,$87,$24,$9F,$C4,$47,$BB,$E0,$FF,$29,$CB,$27,$F1,$91,$F9,$7F
	.dc.b	$84,$91,$F8,$F7,$F3,$5F,$09,$C6,$36,$43,$E3,$FC,$27,$E7,$FE,$12
	.dc.b	$47,$E5,$DF,$C3,$71,$3F,$3A,$FE,$FD,$C9,$FD,$57,$9B,$F4,$77,$FE
	.dc.b	$2A,$F4,$DE,$57,$FB,$AF,$9D,$9F,$EA,$F6,$97,$FA,$38,$EB,$93,$E2
	.dc.b	$69,$27,$FD,$75,$1F,$DB,$1F,$EB,$59,$78,$3E,$1F,$FC,$57,$9F,$83
	.dc.b	$AB,$FD,$CF,$6F,$87,$EB,$F0,$F3,$1E,$71,$E0,$F8,$79,$1D,$9F,$1F
	.dc.b	$C3,$FF,$0A,$BE,$53,$C0,$83,$0F,$F7,$DC,$3E,$1B,$D7,$F0,$F2,$BD
	.dc.b	$2F,$BB,$E1,$BA,$4F,$C0,$CA,$57,$BA,$3E,$1F,$C3,$7C,$4F,$81,$93
	.dc.b	$F0,$A3,$C9,$F0,$DF,$20,$91,$F1,$BE,$1B,$E5,$FC,$34,$8F,$92,$FC
	.dc.b	$EF,$84,$F6,$4F,$09,$F5,$75,$7E,$93,$66,$1F,$52,$BE,$AF,$C2,$FD
	.dc.b	$6F,$86,$98,$7D,$48,$E6,$F8,$48,$74,$1D,$BF,$89,$F8,$5F,$E0,$DD
	.dc.b	$9F,$C1,$FC,$29,$5F,$C3,$A3,$FA,$13,$FC,$B0,$75,$BD,$0F,$E4,$18
	.dc.b	$62,$7F,$92,$CE,$5F,$F2,$C4,$65,$E3,$5F,$C7,$DA,$7F,$2C,$E7,$6F
	.dc.b	$FC,$AE,$D5,$B7,$F9,$9D,$7F,$93,$5F,$C8,$7C,$07,$F9,$BF,$0F,$F2
	.dc.b	$0C,$71,$C7,$CA,$DA,$F1,$BF,$CC,$7F,$38,$76,$7F,$2E,$5D,$94,$C4
	.dc.b	$7C,$2E,$DF,$CE,$5F,$CA,$BE,$8D,$3E,$0E,$1E,$38,$ED,$CF,$D7,$A2
	.dc.b	$57,$F4,$9C,$93,$D7,$C2,$FC,$D4,$F3,$3E,$7E,$1E,$A7,$A9,$EA,$3F
	.dc.b	$62,$7F,$54,$C7,$FC,$27,$A7,$CF,$6E,$5F,$C1,$67,$67,$FA,$BB,$BF
	.dc.b	$F6,$F8,$78,$2F,$E4,$CF,$63,$C5,$5F,$6D,$FF,$43,$9C,$AF,$AB,$E1
	.dc.b	$F7,$3E,$03,$DB,$0F,$F5,$69,$DB,$8B,$D1,$1E,$C3,$FA,$0D,$FC,$AC
	.dc.b	$3C,$31,$F8,$FC,$3A,$A6,$9F,$4A,$37,$7D,$B8,$3C,$35,$F9,$7B,$F4
	.dc.b	$57,$E4,$5A,$7C,$38,$ED,$8F,$CD,$7B,$3C,$B0,$ED,$3F,$8E,$9F,$4C
	.dc.b	$61,$F8,$D5,$D7,$3E,$19,$FC,$BB,$FF,$02,$5F,$A3,$0C,$E7,$F0,$31
	.dc.b	$FF,$0C,$67,$F1,$5F,$9A,$7C,$53,$E8,$E3,$C1,$FC,$25,$F3,$7C,$78
	.dc.b	$F7,$79,$E7,$BC,$FC,$91,$DC,$E8,$FC,$EA,$F5,$47,$C0,$34,$FC,$EA
	.dc.b	$E9,$1F,$C0,$70,$F9,$CF,$83,$1E,$99,$FC,$C8,$F8,$6F,$CD,$BE,$A7
	.dc.b	$9E,$3D,$AF,$82,$B9,$C6,$2F,$BD,$73,$8F,$3B,$CE,$3C,$75,$CF,$0F
	.dc.b	$54,$79,$AF,$9C,$63,$0F,$2E,$5F,$77,$0E,$98,$79,$30,$E7,$9F,$E0
	.dc.b	$63,$E0,$5F,$E5,$B1,$E7,$7E,$18,$E7,$94,$3D,$0F,$C8,$BE,$47,$D1
	.dc.b	$9F,$C4,$4E,$8F,$B6,$71,$CB,$CC,$FD,$C9,$E2,$FC,$D3,$F8,$15,$F8
	.dc.b	$9F,$D8,$63,$F1,$B2,$F6,$1D,$E3,$F6,$47,$8C,$E5,$7F,$EF,$27,$FC
	.dc.b	$CD,$97,$C1,$CB,$C5,$97,$EC,$4F,$74,$F8,$CE,$47,$A6,$B9,$3F,$E0
	.dc.b	$DC,$3E,$99,$E3,$79,$9E,$6A,$E3,$3F,$C4,$F2,$F3,$4F,$F1,$1C,$EE
	.dc.b	$1F,$7C,$FA,$B8,$74,$C3,$8C,$7E,$AA,$F8,$E7,$F0,$F0,$FE,$0D,$7E
	.dc.b	$17,$E6,$9E,$7A,$E0,$F2,$3C,$87,$CC,$C3,$F8,$05,$7E,$C7,$1F,$E2
	.dc.b	$35,$E3,$CF,$CF,$9F,$E3,$E3,$E2,$CF,$C2,$AE,$11,$FC,$2C,$E1,$87
	.dc.b	$D0,$7C,$2F,$CB,$3C,$87,$ED,$8D,$1F,$A8,$21,$EC,$CF,$D0,$FE,$D4
	.dc.b	$E2,$7C,$93,$89,$FA,$63,$8D,$FD,$D7,$C1,$7F,$45,$DD,$FC,$99,$D4
	.dc.b	$FF,$06,$9F,$92,$FE,$50,$E0,$F2,$BF,$82,$7E,$72,$34,$FE,$26,$3C
	.dc.b	$39,$7D,$97,$81,$F9,$43,$D3,$7E,$18,$D5,$78,$49,$5D,$AB,$84,$78
	.dc.b	$8E,$EF,$C9,$3E,$1C,$7D,$07,$E2,$DF,$D7,$7E,$FC,$FD,$7C,$BF,$1F
	.dc.b	$97,$EA,$27,$E4,$D6,$DF,$A0,$7E,$14,$FE,$51,$8F,$DE,$7F,$B4,$E1
	.dc.b	$DF,$0F,$1C,$7D,$57,$CD,$1E,$98,$FA,$C7,$9A,$FC,$0F,$9B,$3F,$DF
	.dc.b	$BF,$13,$1F,$C9,$9E,$67,$F5,$67,$92,$FA,$BF,$12,$FB,$61,$F2,$F3
	.dc.b	$FE,$DB,$1E,$53,$C5,$97,$B4,$FB,$EF,$D9,$3C,$C6,$EB,$E3,$BF,$98
	.dc.b	$BE,$F1,$E1,$8E,$0E,$8F,$CB,$BF,$B8,$AF,$EC,$B8,$FB,$2B,$E2,$E1
	.dc.b	$E5,$BF,$EB,$B7,$DD,$E8,$7A,$C8,$60,$F9,$33,$F3,$A3,$86,$1E,$A9
	.dc.b	$DB,$FC,$E1,$FE,$D3,$5D,$EB,$F7,$79,$7D,$07,$EB,$D7,$C3,$CF,$F2
	.dc.b	$38,$7B,$30,$F0,$3D,$A3,$31,$DB,$1F,$E2,$39,$FD,$68,$FD,$8D,$6F
	.dc.b	$3F,$83,$1F,$B8,$C3,$CC,$FA,$A7,$E7,$E5,$E3,$7E,$AE,$1F,$1D,$FB
	.dc.b	$78,$75,$AF,$C4,$1F,$82,$BE,$67,$A7,$0F,$EA,$D3,$BC,$BE,$36,$3D
	.dc.b	$6F,$EB,$1F,$95,$C3,$EE,$1D,$9F,$CE,$D7,$E3,$DF,$E0,$31,$E4,$7D
	.dc.b	$59,$7C,$27,$D4,$74,$8F,$80,$FF,$7C,$CB,$D4,$F8,$33,$FA,$53,$FC
	.dc.b	$97,$0F,$24,$FD,$DC,$FD,$4F,$D8,$C7,$D5,$5F,$1A,$3E,$06,$3E,$1A
	.dc.b	$DE,$5F,$E4,$0B,$DD,$78,$F0,$EB,$5F,$1B,$2F,$1C,$FE,$2F,$2F,$1B
	.dc.b	$FB,$AA,$DE,$3D,$6B,$77,$F6,$2F,$77,$FE,$4B,$BD,$E7,$F7,$A3,$F8
	.dc.b	$E9,$D7,$1F,$1B,$F8,$A9,$DB,$F1,$4F,$97,$3F,$61,$FA,$27,$E2,$A3
	.dc.b	$79,$7F,$8D,$71,$F1,$E5,$F4,$71,$F1,$E5,$FE,$38,$CF,$E8,$DE,$F1
	.dc.b	$EA,$7C,$B3,$A9,$F8,$A3,$EE,$9F,$84,$9F,$97,$3F,$E6,$3A,$EB,$7F
	.dc.b	$DA,$EF,$F9,$9D,$EF,$0F,$1E,$5E,$98,$FE,$4C,$75,$9D,$BF,$76,$76
	.dc.b	$F8,$CD,$C7,$E5,$9F,$BB,$3F,$2E,$3F,$87,$E3,$F4,$73,$F9,$6F,$EA
	.dc.b	$B2,$F7,$9F,$E4,$DC,$3A,$D7,$E0,$F0,$F7,$BD,$5F,$79,$E3,$3E,$2B
	.dc.b	$B7,$C7,$3F,$8F,$77,$5F,$89,$77,$1F,$16,$FC,$78,$FE,$8E,$3C,$73
	.dc.b	$FE,$A3,$C3,$C7,$1B,$8F,$1B,$D7,$0F,$8B,$87,$59,$FC,$4B,$EF,$76
	.dc.b	$7E,$6C,$F8,$A7,$E2,$4D,$DF,$E0,$F0,$F1,$E3,$FD,$03,$3F,$C4,$E3
	.dc.b	$F1,$6F,$D2,$7F,$AC,$EB,$6F,$F3,$29,$DB,$E9,$36,$7F,$9F,$E3,$78
	.dc.b	$FF,$2A,$8F,$1B,$F8,$F7,$C7,$7F,$83,$8F,$F2,$2E,$3F,$5E,$FA,$D7
	.dc.b	$FA,$B8,$EB,$1F,$E8,$1B,$EB,$87,$F0,$59,$EA,$FA,$5E,$A7,$E2,$6F
	.dc.b	$AE,$3F,$47,$0E,$B8,$7F,$A3,$B2,$FA,$37,$EF,$3E,$59,$FB,$33,$A9
	.dc.b	$EF,$3A,$DF,$E0,$CE,$B9,$7C,$B8,$D9,$E9,$37,$8F,$DD,$7F,$12,$7D
	.dc.b	$73,$F1,$38,$FF,$0B,$7D,$EF,$FA,$57,$2F,$7C,$7F,$11,$3A,$E3,$E9
	.dc.b	$77,$7E,$8C,$FF,$C2,$11,$FC,$44,$DD,$FF,$A6,$1D,$9E,$23,$FC,$E9
	.dc.b	$3E,$97,$C6,$FE,$B7,$0F,$E8,$31,$FC,$D7,$0F,$1B,$D4,$D9,$F8,$FC
	.dc.b	$7F,$67,$7F,$8F,$8F,$97,$8F,$DD,$7C,$78,$7F,$5F,$CB,$FB,$2C,$6A
	.dc.b	$BE,$B3,$A3,$F3,$F1,$A7,$D1,$5C,$DF,$0D,$FE,$34,$F0,$DF,$E3,$71
	.dc.b	$4F,$FA,$6B,$0E,$AF,$08,$FC,$45,$70,$8F,$C4,$47,$07,$E2,$9F,$A1
	.dc.b	$3C,$91,$FC,$43,$0F,$23,$F8,$28,$D1,$D4,$FC,$43,$D4,$FC,$44,$F5
	.dc.b	$CB,$EB,$E3,$FC,$D7,$3F,$CA,$E3,$F5,$F0,$FC,$49,$F9,$53,$E0,$BE
	.dc.b	$AC,$3F,$82,$BF,$89,$8D,$9F,$B3,$3E,$66,$5F,$76,$76,$FD,$D8,$FF
	.dc.b	$11,$CF,$D7,$7A,$C6,$F1,$FF,$3B,$47,$F2,$33,$AE,$1F,$B3,$7A,$9F
	.dc.b	$02,$FA,$BF,$83,$9F,$C4,$BF,$D7,$CD,$C7,$F0,$57,$E3,$1F,$A7,$7E
	.dc.b	$34,$7E,$9D,$F9,$86,$CE,$A7,$DF,$7A,$CF,$F2,$39,$F4,$9E,$38,$F4
	.dc.b	$C7,$F2,$99,$EB,$3F,$2D,$EB,$3F,$C2,$E3,$C7,$3F,$AD,$7D,$F7,$FB
	.dc.b	$37,$D3,$3B,$AF,$1B,$F8,$A3,$C7,$3F,$2E,$3A,$CF,$EE,$4E,$B1,$F7
	.dc.b	$63,$C7,$1F,$5C,$F4,$BE,$F9,$F4,$CF,$DD,$8D,$D7,$D1,$AD,$D7,$8C
	.dc.b	$DD,$FD,$D8,$FF,$1D,$1F,$B9,$3D,$27,$8E,$BD,$2F,$E5,$A7,$D3,$3F
	.dc.b	$A9,$8F,$D4,$3F,$83,$7F,$D6,$EF,$5A,$FF,$35,$57,$E7,$27,$67,$D1
	.dc.b	$36,$7E,$0C,$FC,$7C,$F5,$7F,$5B,$1E,$F7,$6F,$DD,$76,$7E,$D0,$EA
	.dc.b	$ED,$FF,$6D,$1F,$95,$7E,$2D,$FD,$78,$F8,$A7,$BE,$FE,$2B,$E3,$3E
	.dc.b	$29,$F9,$53,$67,$F1,$D9,$F8,$AF,$E5,$63,$E2,$BE,$33,$D2,$FD,$17
	.dc.b	$D2,$FE,$0D,$F4,$C7,$C5,$76,$78,$CD,$9F,$44,$DC,$7E,$84,$FE,$0D
	.dc.b	$9F,$F9,$BE,$7C,$78,$7E,$55,$F7,$E1,$FD,$06,$75,$5E,$E9,$F9,$57
	.dc.b	$EE,$BF,$D3,$47,$E5,$23,$DD,$3B,$3D,$0E,$F0,$FC,$D4,$78,$A3,$FC
	.dc.b	$A0,$7F,$22,$7F,$63,$9F,$F2,$27,$51,$FE,$3D,$7A,$3E,$D3,$A6,$3F
	.dc.b	$81,$CF,$F9,$0D,$FF,$92,$6B,$75,$F3,$CF,$86,$7E,$DD,$EB,$87,$F8
	.dc.b	$DC,$F8,$73,$FA,$B8,$FA,$B1,$F3,$EB,$AD,$FF,$B9,$27,$FD,$A1,$7F
	.dc.b	$81,$C7,$F3,$31,$F3,$E3,$F1,$AF,$DB,$7F,$1B,$1F,$C2,$63,$A9,$B8
	.dc.b	$FB,$67,$CA,$3E,$D9,$E9,$3D,$E7,$E1,$EB,$C6,$FF,$93,$A3,$C7,$7F
	.dc.b	$5A,$BC,$73,$EF,$3C,$67,$E1,$CF,$1E,$1F,$C0,$EB,$C7,$1B,$3C,$73
	.dc.b	$FD,$2D,$FA,$C7,$FB,$4C,$FE,$C7,$5F,$EE,$EC,$BF,$BD,$CE,$E7,$F8
	.dc.b	$AE,$3F,$DD,$5D,$DF,$FB,$05,$FD,$2C,$EE,$FE,$8D,$6E,$51,$D5,$FE
	.dc.b	$1E,$7C,$53,$D3,$1F,$C2,$72,$FC,$44,$7B,$8D,$5F,$C4,$3F,$27,$7F
	.dc.b	$12,$3F,$93,$5F,$E6,$63,$C5,$3F,$88,$7F,$48,$FC,$49,$DB,$F8,$83
	.dc.b	$FC,$D8,$6F,$1F,$C4,$47,$E3,$8D,$9E,$39,$EA,$FD,$C7,$AB,$B7,$F9
	.dc.b	$95,$78,$E7,$F3,$53,$B7,$AB,$F8,$28,$FC,$A6,$3F,$B6,$8D,$63,$F2
	.dc.b	$67,$E4,$47,$CE,$AE,$6F,$C8,$3F,$31,$8F,$C8,$7E,$FC,$6B,$1F,$91
	.dc.b	$1F,$AF,$C7,$E4,$1E,$68,$F8,$51,$FE,$51,$BF,$B2,$FF,$8B,$8F,$BE
	.dc.b	$7F,$47,$3C,$58,$7E,$A7,$3F,$F1,$24,$AB,$FE,$00,$57,$F6,$6A,$FE
	.dc.b	$24,$FF,$06,$3F,$16,$75,$3F,$0B,$1F,$8B,$9F,$91,$3D,$67,$C2,$7F
	.dc.b	$3B,$9F,$B5,$7E,$98,$FA,$98,$7B,$F0,$F1,$D7,$8B,$0F,$E4,$75,$FE
	.dc.b	$5D,$CF,$F5,$F1,$FE,$CE,$9F,$8B,$3F,$B5,$C3,$E2,$E1,$FD,$1E,$BC
	.dc.b	$73,$FB,$5C,$3A,$BF,$ED,$B3,$FC,$D3,$1F,$05,$FF,$26,$E1,$FE,$E2
	.dc.b	$C7,$D5,$1E,$E3,$81,$F1,$4F,$26,$7F,$81,$7E,$E5,$7F,$01,$CF,$EE
	.dc.b	$5F,$A3,$3F,$F1,$89,$E8,$CF,$F2,$37,$F8,$B8,$FE,$25,$3B,$8F,$F2
	.dc.b	$FC,$7F,$B3,$9F,$75,$74,$C3,$C1,$3C,$F3,$FC,$AE,$1F,$0B,$2F,$B3
	.dc.b	$87,$F3,$93,$F2,$A7,$D9,$CB,$F2,$A7,$C6,$3E,$66,$7F,$1A,$3E,$C1
	.dc.b	$FA,$5B,$F9,$98,$7E,$42,$3F,$D5,$87,$E1,$2B,$FC,$D4,$FE,$03,$0F
	.dc.b	$EA,$B3,$B7,$FC,$FD,$3F,$96,$37,$3F,$34,$FE,$F0,$F8,$E7,$E3,$DF
	.dc.b	$8E,$BF,$9C,$57,$8F,$0F,$6D,$FF,$49,$BF,$E0,$51,$FB,$17,$F2,$87
	.dc.b	$87,$0F,$F3,$73,$AA,$F8,$4E,$9F,$C9,$CF,$E0,$2B,$F0,$91,$F0,$8F
	.dc.b	$AC,$7F,$35,$CF,$EB,$3F,$E1,$8A,$FA,$55,$E1,$8F,$D8,$57,$AD,$FB
	.dc.b	$78,$FF,$A1,$27,$51,$F0,$2B,$CD,$1F,$E3,$58,$FC,$5C,$7F,$28,$CF
	.dc.b	$F0,$11,$FB,$3B,$F9,$53,$FD,$E2,$3E,$86,$3F,$DF,$9F,$59,$FA,$13
	.dc.b	$37,$FC,$DD,$A7,$F5,$BC,$FF,$17,$8F,$CF,$AF,$EA,$93,$F3,$CF,$CF
	.dc.b	$9E,$1B,$F9,$58,$7D,$A3,$F1,$6F,$AA,$3F,$C9,$07,$A8,$FE,$35,$8F
	.dc.b	$A9,$FC,$A1,$93,$F3,$A7,$92,$7F,$96,$BD,$CF,$CC,$BD,$E7,$EC,$47
	.dc.b	$F2,$8C,$7E,$C4,$7F,$09,$3F,$87,$DF,$E6,$27,$E1,$D7,$E7,$9E,$D3
	.dc.b	$F3,$5A,$7D,$DC,$7E,$1B,$83,$CA,$7E,$5F,$3F,$CE,$BD,$27,$F2,$EE
	.dc.b	$E3,$F7,$A7,$EC,$DF,$92,$7C,$08,$FC,$3B,$F9,$23,$F0,$EF,$C9,$AF
	.dc.b	$F3,$23,$A7,$E0,$46,$8F,$01,$F8,$C3,$FC,$0A,$7F,$91,$71,$F9,$D1
	.dc.b	$FA,$58,$F9,$33,$F8,$77,$D5,$87,$E1,$DF,$94,$7E,$1C,$CE,$3F,$87
	.dc.b	$69,$F8,$73,$EA,$1F,$67,$0F,$7E,$1E,$0B,$FE,$EB,$3E,$03,$FC,$30
	.dc.b	$72,$C3,$EE,$BF,$0F,$2F,$C8,$61,$DE,$7F,$17,$3E,$E3,$F5,$D3,$A7
	.dc.b	$FB,$49,$F3,$F1,$FB,$CF,$AA,$7E,$F3,$DE,$FE,$F3,$4F,$BC,$7A,$9F
	.dc.b	$9C,$7C,$58,$F0,$C7,$38,$FB,$AF,$13,$F0,$71,$AC,$FF,$81,$CF,$F3
	.dc.b	$7C,$BE,$1D,$79,$71,$FE,$2F,$3F,$12,$3E,$46,$1F,$CE,$32,$F9,$11
	.dc.b	$FD,$EE,$7F,$5A,$78,$CF,$CF,$E1,$DB,$2F,$E1,$53,$F8,$93,$F8,$51
	.dc.b	$F1,$E7,$C5,$7D,$63,$C4,$7E,$4B,$2F,$C2,$3E,$2C,$7F,$89,$E5,$F9
	.dc.b	$C8,$F8,$07,$E3,$4F,$E7,$D7,$E3,$7F,$24,$68,$FA,$A7,$6A,$FC,$BB
	.dc.b	$E4,$35,$1E,$4A,$EF,$7F,$91,$C7,$D9,$5F,$9A,$C1,$97,$F4,$13,$E2
	.dc.b	$C7,$F5,$0C,$7F,$90,$BF,$8C,$7E,$96,$7F,$B8,$8F,$AD,$5F,$49,$B8
	.dc.b	$7E,$1B,$0E,$F1,$EC,$39,$65,$F3,$AF,$96,$1F,$D0,$B0,$FE,$A3,$5F
	.dc.b	$D0,$8F,$F2,$69,$EC,$CF,$E9,$CF,$C0,$9F,$95,$87,$68,$F3,$D7,$ED
	.dc.b	$4F,$8D,$87,$F6,$B9,$D3,$F8,$08,$F1,$9F,$3A,$32,$FE,$AC,$E4,$7B
	.dc.b	$33,$FC,$6B,$F8,$93,$57,$F2,$9F,$F0,$FD,$77,$C7,$DA,$FF,$27,$8F
	.dc.b	$E7,$D8,$7F,$60,$C3,$FC,$45,$1F,$C0,$E3,$E1,$65,$FD,$12,$3E,$16
	.dc.b	$1F,$E3,$B7,$E1,$1F,$67,$3F,$B0,$FB,$8F,$F1,$6E,$3F,$E9,$6A,$FF
	.dc.b	$20,$46,$4F,$5D,$FD,$A7,$F6,$99,$FE,$FE,$3F,$C5,$91,$BC,$7C,$B1
	.dc.b	$FB,$E3,$CB,$9F,$F0,$9C,$FE,$D5,$72,$3B,$67,$F0,$DF,$F1,$D5,$7B
	.dc.b	$F2,$F2,$5F,$F4,$A9,$FF,$1F,$CF,$F9,$48,$D3,$FE,$A2,$AF,$F4,$5B
	.dc.b	$FA,$93,$F6,$4F,$E0,$2B,$E6,$C7,$2B,$FE,$A7,$7F,$5A,$3F,$35,$1F
	.dc.b	$7D,$F8,$2F,$A1,$FE,$48,$7A,$33,$F9,$6F,$DF,$3F,$75,$1F,$C2,$CF
	.dc.b	$E3,$2F,$E2,$0F,$D8,$63,$F8,$87,$CD,$8F,$E5,$73,$FD,$86,$3E,$69
	.dc.b	$FF,$0F,$65,$FD,$5E,$BD,$F1,$FE,$88,$3F,$05,$87,$E1,$DF,$C0,$DF
	.dc.b	$EA,$EF,$95,$FE,$98,$E5,$5F,$8F,$AF,$A4,$7A,$6F,$C5,$97,$F6,$DC
	.dc.b	$7E,$DD,$7F,$2E,$9F,$B7,$5F,$CF,$8E,$11,$CD,$F8,$A6,$0F,$81,$7F
	.dc.b	$45,$EF,$3C,$8F,$C4,$BB,$15,$E0,$7F,$8F,$47,$ED,$A7,$F0,$CF,$E0
	.dc.b	$6B,$C0,$FF,$28,$7E,$7C,$FD,$83,$F6,$D8,$7C,$2C,$7F,$4D,$7F,$0A
	.dc.b	$7F,$42,$7E,$26,$3F,$42,$F8,$E3,$BD,$7E,$1A,$3E,$E9,$FB,$B3,$EC
	.dc.b	$BF,$CC,$5F,$BB,$5D,$F1,$FD,$04,$78,$8F,$83,$3F,$5E,$BC,$51,$F0
	.dc.b	$2B,$F3,$EF,$DC,$3F,$2C,$7F,$02,$9F,$79,$F8,$DC,$FF,$64,$FF,$6A
	.dc.b	$C7,$F6,$D3,$F4,$9F,$9D,$19,$3F,$34,$FC,$79,$E3,$3F,$E4,$23,$F7
	.dc.b	$55,$E2,$9F,$DD,$57,$9F,$3F,$E7,$F8,$7F,$A2,$23,$F4,$CF,$C6,$BF
	.dc.b	$31,$F3,$0F,$CB,$57,$27,$D6,$7E,$5B,$1F,$D8,$4F,$38,$FE,$4B,$7F
	.dc.b	$AC,$AF,$59,$E7,$8F,$D8,$1F,$CA,$0E,$CF,$F9,$ED,$FE,$BE,$7D,$98
	.dc.b	$E5,$7F,$00,$F9,$11,$C8,$FE,$1E,$FC,$C7,$EE,$3F,$8A,$AF,$64,$79
	.dc.b	$CF,$C0,$9F,$C1,$8F,$39,$E3,$3E,$D1,$F8,$BB,$FE,$30,$FE,$EA,$BE
	.dc.b	$4E,$5F,$EA,$08,$E2,$78,$8F,$5D,$FF,$93,$C3,$2F,$F8,$EC,$79,$6B
	.dc.b	$DB,$3F,$8D,$3F,$01,$1F,$5D,$FB,$67,$8C,$FC,$61,$E3,$32,$7D,$67
	.dc.b	$F7,$C7,$17,$F6,$99,$FE,$CD,$F9,$B8,$FE,$31,$FC,$41,$AC,$BF,$11
	.dc.b	$5C,$A3,$EF,$E1,$F9,$EB,$E5,$3F,$E0,$F9,$FE,$6B,$8F,$F2,$07,$F4
	.dc.b	$A7,$0C,$7E,$3E,$5F,$EB,$1A,$F8,$13,$F3,$EF,$9E,$7F,$DF,$A7,$F6
	.dc.b	$58,$FF,$0F,$AF,$86,$73,$3D,$F9,$FE,$1E,$7C,$4F,$F2,$08,$FF,$35
	.dc.b	$1C,$0F,$8C,$7B,$4F,$94,$FE,$6D,$CB,$F9,$68,$FC,$6C,$7E,$44,$F3
	.dc.b	$67,$F1,$1F,$E3,$71,$FA,$8C,$BE,$CC,$7F,$08,$3E,$09,$D2,$3F,$30
	.dc.b	$FE,$53,$0F,$DB,$BF,$E5,$E8,$CB,$F9,$33,$F8,$99,$FC,$C3,$3F,$98
	.dc.b	$FA,$F1,$FE,$85,$7F,$32,$3F,$D8,$75,$F3,$1F,$C9,$1F,$9A,$3F,$08
	.dc.b	$7A,$0F,$CB,$E3,$FC,$C5,$FC,$B9,$FE,$5D,$C3,$E9,$3E,$1C,$FF,$A5
	.dc.b	$65,$F0,$63,$F1,$37,$F8,$DB,$F9,$11,$ED,$3F,$68,$F5,$7C,$4F,$D7
	.dc.b	$BE,$B1,$FC,$F4,$F9,$07,$B7,$2F,$E7,$B8,$FB,$6B,$FD,$6F,$5F,$23
	.dc.b	$2F,$F2,$84,$78,$2F,$E4,$61,$F0,$63,$C1,$9F,$E3,$27,$FC,$50,$7D
	.dc.b	$A3,$F0,$33,$F9,$D7,$EB,$1F,$B1,$9F,$E2,$CF,$E0,$4F,$97,$3C,$DF
	.dc.b	$F2,$E1,$E9,$AF,$CB,$4F,$03,$F5,$91,$F7,$4E,$95,$F8,$C8,$CC,$FA
	.dc.b	$A3,$F0,$2F,$F2,$99,$FC,$09,$D7,$1F,$C0,$D7,$5A,$FA,$07,$08,$FE
	.dc.b	$25,$3C,$2B,$F8,$F9,$FB,$7C,$BE,$C5,$7E,$98,$FD,$6D,$74,$7C,$87
	.dc.b	$4A,$F3,$9F,$6C,$F8,$57,$F3,$2B,$9E,$3F,$6B,$0E,$F9,$7D,$7C,$3F
	.dc.b	$8F,$E0,$9F,$E3,$F5,$F6,$71,$FF,$54,$4F,$EC,$CF,$CA,$9F,$8C,$3F
	.dc.b	$7C,$67,$1F,$C4,$3F,$68,$FA,$78,$F8,$9C,$47,$95,$FB,$A7,$E7,$CF
	.dc.b	$89,$8F,$ED,$B0,$FA,$F8,$FC,$09,$EF,$5F,$8B,$9F,$A5,$3F,$8B,$3F
	.dc.b	$C1,$11,$FE,$38,$3C,$73,$F8,$29,$F8,$2F,$F1,$49,$FE,$80,$FF,$3F
	.dc.b	$C3,$FC,$27,$5F,$80,$7E,$0B,$DD,$FB,$33,$DE,$7F,$B8,$9F,$A8,$79
	.dc.b	$BF,$2E,$39,$C7,$D7,$8F,$D4,$CF,$E1,$5F,$5C,$7E,$17,$3F,$E6,$19
	.dc.b	$7D,$17,$FD,$8A,$7E,$F2,$7E,$39,$F4,$4F,$D1,$9D,$27,$E3,$38,$C3
	.dc.b	$E4,$BF,$85,$BF,$65,$74,$8F,$E6,$31,$E3,$78,$61,$F8,$F7,$85,$FA
	.dc.b	$6F,$9C,$7E,$EB,$0F,$33,$FD,$33,$0F,$33,$F8,$6A,$FC,$D3,$FC,$54
	.dc.b	$EE,$F9,$AB,$F9,$2C,$7D,$33,$F8,$E1,$F8,$68,$FB,$67,$F8,$1D,$F0
	.dc.b	$1F,$C6,$1C,$BF,$C9,$0F,$F5,$89,$F7,$F1,$FC,$56,$1E,$B8,$E2,$7E
	.dc.b	$9A,$78,$BF,$64,$F0,$9C,$4F,$E2,$47,$AF,$0F,$D5,$57,$4C,$3D,$F8
	.dc.b	$FE,$38,$FF,$34,$46,$6B,$DE,$7C,$A3,$EF,$9F,$B8,$3C,$71,$F5,$1F
	.dc.b	$94,$70,$33,$7F,$47,$0F,$56,$3F,$A0,$C7,$E8,$DF,$F6,$B3,$F8,$DE
	.dc.b	$7F,$C8,$5F,$1E,$1E,$A7,$CD,$87,$F6,$38,$C3,$F9,$8B,$FE,$03,$1F
	.dc.b	$34,$FC,$ED,$73,$9E,$11,$FB,$C3,$F2,$6F,$DE,$7F,$91,$65,$F1,$23
	.dc.b	$A6,$3F,$A2,$9F,$82,$7E,$5C,$F8,$87,$E4,$27,$83,$F4,$B0,$F4,$BD
	.dc.b	$CF,$A4,$7E,$14,$F2,$E7,$FC,$35,$FB,$06,$E3,$DE,$FF,$6A,$BF,$BF
	.dc.b	$87,$AE,$30,$72,$3D,$79,$7D,$E7,$E2,$9E,$39,$F8,$35,$FC,$66,$3C
	.dc.b	$67,$81,$C1,$E0,$7F,$8D,$9C,$8F,$CA,$BE,$C3,$71,$F1,$B0,$FF,$2C
	.dc.b	$E5,$FA,$4A,$FE,$59,$39,$7F,$C9,$71,$9A,$FE,$29,$5F,$C2,$27,$EC
	.dc.b	$1F,$E6,$98,$FE,$45,$5C,$5F,$F1,$55,$71,$8F,$E2,$13,$FE,$14,$BF
	.dc.b	$C5,$C7,$F5,$0B,$FA,$73,$E8,$8F,$82,$7C,$AC,$7E,$86,$1E,$B8,$C4
	.dc.b	$FC,$A3,$EB,$1E,$37,$F7,$58,$7E,$66,$3C,$EF,$F6,$17,$81,$F2,$E3
	.dc.b	$9B,$F3,$A7,$B3,$F5,$CC,$47,$91,$C4,$7D,$18,$FE,$6D,$1E,$13,$F8
	.dc.b	$36,$2B,$FF,$64,$47,$FB,$B6,$BE,$5B,$F0,$CF,$2E,$7F,$D9,$DF,$E8
	.dc.b	$D8,$FD,$6C,$FF,$C4,$31,$FC,$64,$F8,$C7,$E3,$0F,$01,$FC,$34,$C9
	.dc.b	$F8,$43,$F2,$D8,$FD,$53,$FC,$3D,$97,$D5,$8F,$E0,$CF,$C7,$7F,$8C
	.dc.b	$1E,$E8,$E0,$FF,$69,$59,$FF,$34,$8F,$EF,$A7,$F5,$3B,$AF,$A8,$B4
	.dc.b	$B4,$EA,$78,$1A,$8C,$B6,$35,$3C,$5B,$3A,$9F,$0C,$E5,$E0,$43,$3F
	.dc.b	$00,$FC,$4F,$9B,$63,$CA,$D7,$71,$A2,$3A,$6B,$A3,$04,$30,$7F,$AF
	.dc.b	$BF,$16,$10

;--------------------------------------------------------------------------------
;	FTRAPcc
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
ftrapcc060_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#FPSP060,-(sp)
	peamsg	'FTRAPCC'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FTRAPcc',13,10
;------------------------------------------------
;	d1	actual result upper
;	d2	actual result lower
;	d3	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result upper,expected result lower,expected status,...
;	a1	expected result upper
;	a2	expected result lower
;	a3	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result upper,expected result lower,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	ftrapcc060_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FTRAPcc
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	clr.l	trapv_occurred		;actual result upper
	jsr	([ftrapcc060_execute,za0,d5.l*4])	;EXECUTE
					;actual result lower
	move.l	trapv_occurred,d2	;actual result upper
	fmove.l	fpsr,d3			;actual status
;
	movem.l	(a0),a1-a3		;expected result upper,expected result lower,expected status
;
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	test_double
	lea.l	24(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTRAP'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	' @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	output_double
	lea.l	28(sp),sp
@@:
;
	lea.l	(12,a0),a0		;expected result upper,expected result lower,expected status,...
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
ftrapcc060_execute::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	ftrapcc060_execute_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
ftrapcc060_execute_&cc::
	moveq.l	#3,d1			;actual result upper. 3=too long
	FTRAP&cc			;EXECUTE
	subq.l	#1,d1			;actual result upper. 0=ok
	subq.l	#1,d1			;actual result upper. 1=too long
	subq.l	#1,d1			;actual result upper. 2=too long
	rts
  .endm
	.cpu	68000

	.align	4
ftrapcc060_expected_compressed::
	.dc.b	$00,$00,$18,$00,$10,$84,$21,$08,$69,$AD,$29,$08,$73,$69,$86,$1B
	.dc.b	$72,$E0,$87,$89,$CB,$82,$07,$2E,$0E,$4E,$1C,$1E,$17,$0E,$0F,$11
	.dc.b	$08,$7B,$9D,$BE,$33,$27,$C4,$72,$FB,$CF,$86,$F9,$5A,$FB,$C8,$E0
	.dc.b	$87,$77,$2F,$21,$F8,$14,$7C,$06,$1F,$33,$4C,$BB,$7E,$D4,$7E,$02
	.dc.b	$31,$1E,$B6,$BE,$96,$BF,$75,$AF,$E2,$23,$D9,$18,$8F,$A5,$1D,$63
	.dc.b	$E2,$36,$3C,$2D,$7E,$E3,$5F,$C3,$35,$FC,$73,$5F,$CB,$3E,$08,$FC
	.dc.b	$8C,$FE,$38,$FC,$51,$27,$F2,$4E,$5F,$A6,$E5,$FC,$BB,$97,$EF,$39
	.dc.b	$7A,$9C,$8E,$24,$38,$99,$38,$99,$38,$9A,$32,$43,$24,$3E,$AB,$E1
	.dc.b	$7F,$5E,$F9,$5F,$C7,$39,$7F,$69,$5F,$AE,$23,$FB,$16,$BF,$B4,$6B
	.dc.b	$FB,$87,$F6,$04,$7F,$2E,$D7,$F4,$0D,$7F,$3C,$D7,$F8,$1B,$F5,$5F
	.dc.b	$E1,$4D,$7F,$86,$B5,$34,$8F,$E5,$CF,$59,$4F,$E2,$2F,$3A,$E3,$19
	.dc.b	$3F,$89,$46,$48,$E4,$FC,$3B,$8A,$FC,$B4,$F2,$6C,$79,$6F,$F8,$91
	.dc.b	$FC,$60,$B5,$FC,$6D,$AF,$F1,$E6,$BF,$C8,$9A,$FF,$26,$6B,$FC,$A5
	.dc.b	$AF,$F2,$C6,$BF,$CB,$DA,$FF,$32,$6B,$FC,$D5,$AF,$F3,$86,$A6,$91
	.dc.b	$FE,$22,$FA,$DA,$7F,$35,$C3,$F9,$81,$FC,$D8,$AE,$4F,$E3,$2E,$5C
	.dc.b	$1F,$C8,$DD,$B4,$D9,$0C,$1F,$D1,$1D,$BF,$D2,$1F,$B8,$7E,$14,$AF
	.dc.b	$D9,$3F,$AA,$38,$AF,$E1,$D0,$8E,$0E,$CF,$E8,$4E,$E3,$F9,$A3,$63
	.dc.b	$F9,$0B,$5F,$EB,$8D,$7F,$AF,$E3,$FD,$4C,$FE,$B2,$57,$FA,$E3,$5F
	.dc.b	$EB,$ED,$7F,$B1,$B5,$FE,$CE,$D7,$FB,$53,$5E,$E7,$F4,$67,$BB,$4E
	.dc.b	$E5,$3B,$94,$EE,$53,$B9,$4E,$E5,$3B,$9F,$D6,$DE,$ED,$3F,$BC,$65
	.dc.b	$FD,$CC,$FE,$F2,$57,$27,$F5,$C7,$2E,$0F,$EC,$CE,$DA,$6C,$86,$0F
	.dc.b	$F0,$3B,$B7,$FC,$16,$FE,$FC,$FE,$10,$57,$EF,$07,$15,$F8,$47,$71
	.dc.b	$F8,$3C,$74,$D8,$DC,$FF,$77,$6C,$7F,$64,$78,$C7,$F0,$67,$F8,$C6
	.dc.b	$7F,$E1,$A2,$4F,$F8,$75,$FF,$0C,$11,$FE,$FE,$D7,$FC,$46,$F0,$69
	.dc.b	$C4,$D9,$0F,$F0,$73,$C5,$FF,$16,$BC,$5A,$71,$36,$71,$3F,$C2,$4F
	.dc.b	$17,$FC,$6E,$F1,$69,$C4,$D9,$C4,$FF,$02,$BC,$5F,$F2,$06,$9F,$E3
	.dc.b	$A2,$3F,$E3,$D6,$BF,$E4,$06,$BF,$E4,$47,$FC,$78,$47,$FC,$46,$D7
	.dc.b	$FC,$58,$D7,$FC,$54,$D7,$FC,$9E,$FF,$49,$23,$F8,$26,$D7,$F0,$46
	.dc.b	$BE,$18,$FC,$14,$7F,$49,$23,$82,$47,$F4,$48,$CB,$63,$F1,$FA,$FF
	.dc.b	$96,$88,$FF,$97,$5F,$F2,$C1,$1F,$F1,$FB,$5F,$F3,$1B,$C1,$A7,$13
	.dc.b	$64,$3F,$C9,$CF,$17,$FC,$DA,$F1,$69,$C4,$D9,$C4,$FF,$29,$3C,$5F
	.dc.b	$F3,$BB,$C5,$A7,$13,$67,$13,$FC,$8A,$F1,$7F,$D0,$1B,$7F,$9E,$88
	.dc.b	$FF,$9F,$5A,$FF,$A0,$1A,$FF,$A1,$1F,$F3,$E1,$1F,$F3,$1B,$5F,$F3
	.dc.b	$63,$5F,$F3,$53,$5F,$F4,$7B,$FE,$09,$23,$FC,$09,$B3,$FD,$11,$AF
	.dc.b	$86,$3F,$81,$47,$F8,$24,$8E,$09,$1F,$E0,$48,$CB,$63,$F8,$7E,$FF
	.dc.b	$E9,$53,$FD,$36,$57,$FD,$3C,$D7,$FD,$42,$D7,$FD,$4A,$D7,$FD,$52
	.dc.b	$D7,$FD,$58,$D7,$FD,$5E,$D7,$FD,$64,$D7,$FD,$6A,$D7,$FD,$70,$D7
	.dc.b	$FD,$76,$D4,$D2,$3F,$E9,$47,$D6,$D3,$FD,$71,$C1,$FE,$B3,$3F,$D7
	.dc.b	$25,$72,$7F,$A7,$1C,$B8,$3F,$D4,$CE,$DA,$6C,$86,$0F,$F6,$3B,$B7
	.dc.b	$FD,$96,$FF,$8F,$CF,$F2,$41,$5F,$E8,$2D,$9F,$F0,$1B,$A8,$FE,$80
	.dc.b	$E6,$3F,$D6,$8D,$8F,$F5,$0B,$5F,$F6,$E3,$5F,$F6,$FF,$0F,$FB,$4C
	.dc.b	$FF,$6C,$95,$FF,$6E,$35,$FF,$6F,$B5,$FF,$71,$B5,$FF,$73,$B5,$FF
	.dc.b	$75,$35,$EE,$7F,$B1,$9E,$ED,$3B,$94,$EE,$53,$B9,$4E,$E5,$3B,$94
	.dc.b	$EE,$7F,$B6,$DE,$ED,$3F,$DF,$1C,$5F,$EF,$33,$FD,$F2,$57,$27,$FB
	.dc.b	$71,$CB,$83,$FD,$CC,$ED,$A6,$C8,$60,$10,$84,$79,$38,$23,$FE,$83
	.dc.b	$3F,$D1,$85,$7F,$C0,$AD,$AF,$F2,$3B,$A8,$FF,$02,$39,$8F,$F7,$C3
	.dc.b	$63,$FD,$CA,$D7,$C0,$79,$9E,$90

;--------------------------------------------------------------------------------
;	FTRAPcc
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
ftrapcc88x_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040,-(sp)
	peamsg	'FTRAPCC'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FTRAPcc',13,10
;------------------------------------------------
;	d1	actual result upper
;	d2	actual result lower
;	d3	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result upper,expected result lower,expected status,...
;	a1	expected result upper
;	a2	expected result lower
;	a3	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result upper,expected result lower,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	ftrapcc88x_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FTRAPcc
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	clr.l	trapv_occurred		;actual result lower
	jsr	([ftrapcc88x_execute,za0,d5.l*4])	;EXECUTE
					;actual result upper
	move.l	trapv_occurred,d2	;actual result lower
	fmove.l	fpsr,d3			;actual status
;
	movem.l	(a0),a1-a3		;expected result upper,expected result lower,expected status
;
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	test_double
	lea.l	24(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTRAP'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	' @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	output_double
	lea.l	28(sp),sp
@@:
;
	lea.l	(12,a0),a0		;expected result upper,expected result lower,expected status,...
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
ftrapcc88x_execute::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	ftrapcc88x_execute_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
ftrapcc88x_execute_&cc::
	moveq.l	#3,d1			;actual result upper. 3=too long
	FTRAP&cc			;EXECUTE
	subq.l	#1,d1			;actual result upper. 0=ok
	subq.l	#1,d1			;actual result upper. 1=too long
	subq.l	#1,d1			;actual result upper. 2=too long
	rts
  .endm
	.cpu	68000

	.align	4
ftrapcc88x_expected_compressed::
	.dc.b	$00,$00,$18,$00,$10,$84,$21,$08,$69,$AD,$29,$08,$73,$69,$86,$1B
	.dc.b	$72,$E0,$87,$89,$CB,$82,$07,$2E,$0E,$4E,$1C,$1E,$17,$0E,$0F,$11
	.dc.b	$08,$7B,$9D,$BE,$33,$27,$C4,$72,$FB,$CF,$86,$F9,$5A,$FB,$C8,$E0
	.dc.b	$87,$77,$2F,$21,$F8,$14,$7C,$06,$1F,$33,$4C,$BB,$7E,$D4,$7E,$02
	.dc.b	$31,$1E,$B6,$BE,$96,$BF,$75,$AF,$E2,$23,$D9,$18,$8F,$A5,$1D,$63
	.dc.b	$E2,$36,$3C,$2D,$7E,$E3,$5F,$C3,$35,$FC,$73,$5F,$CB,$3E,$08,$FC
	.dc.b	$8C,$FE,$38,$FC,$51,$27,$F2,$4E,$5F,$A6,$E5,$FC,$BB,$97,$EF,$39
	.dc.b	$7A,$9C,$8E,$24,$38,$99,$38,$99,$38,$9A,$32,$43,$24,$3E,$AB,$E1
	.dc.b	$7F,$5E,$F9,$5F,$C7,$39,$7F,$69,$5F,$AE,$23,$FB,$16,$BF,$B4,$6B
	.dc.b	$FB,$87,$F6,$04,$7F,$2E,$D7,$F4,$0D,$7F,$3C,$D7,$F8,$1B,$F5,$5F
	.dc.b	$E1,$4D,$7F,$86,$B5,$34,$8F,$E5,$CF,$59,$4F,$E2,$2F,$3A,$E3,$19
	.dc.b	$3F,$89,$46,$48,$E4,$FC,$3B,$8A,$FC,$B4,$F2,$6C,$79,$6F,$F8,$91
	.dc.b	$FC,$60,$B5,$FC,$6D,$AF,$F1,$E6,$BF,$C8,$9A,$FF,$26,$6B,$FC,$A5
	.dc.b	$AF,$F2,$C6,$BF,$CB,$DA,$FF,$32,$6B,$FC,$D5,$AF,$F3,$86,$A6,$91
	.dc.b	$FE,$22,$FA,$DA,$7F,$35,$C3,$F9,$81,$FC,$D8,$AE,$4F,$E3,$2E,$5A
	.dc.b	$43,$F9,$23,$C1,$A6,$4F,$E4,$8E,$5A,$75,$AF,$A7,$1F,$85,$3E,$F1
	.dc.b	$63,$ED,$1F,$C3,$DC,$47,$08,$C9,$FD,$06,$38,$B8,$B4,$62,$78,$10
	.dc.b	$F1,$96,$7F,$AE,$E3,$FD,$4C,$FE,$B2,$57,$FA,$E3,$5F,$EB,$ED,$7F
	.dc.b	$B1,$B5,$E6,$7F,$42,$79,$B4,$E6,$53,$99,$4E,$65,$39,$9F,$D4,$DE
	.dc.b	$6D,$39,$94,$E6,$53,$99,$4E,$67,$F5,$97,$FB,$BE,$5F,$DC,$CB,$1F
	.dc.b	$D6,$9A,$FF,$63,$6B,$FD,$85,$E0,$EC,$86,$CA,$6C,$A7,$F5,$47,$BC
	.dc.b	$FE,$D9,$FE,$10,$7F,$01,$2C,$7D,$D6,$C6,$E3,$04,$70,$43,$C0,$57
	.dc.b	$C0,$6D,$FF,$0D,$E7,$FE,$18,$24,$7F,$86,$DF,$F0,$B1,$1F,$EF,$AD
	.dc.b	$7F,$C4,$2F,$06,$9C,$4D,$90,$FF,$06,$3C,$5F,$09,$B3,$89,$0F,$F0
	.dc.b	$73,$B7,$FB,$DB,$5F,$F1,$B3,$E1,$79,$9C,$4D,$9E,$13,$99,$4F,$F1
	.dc.b	$EE,$9F,$E3,$82,$3F,$E3,$B6,$BF,$E3,$E6,$BF,$E4,$27,$FC,$74,$47
	.dc.b	$FC,$42,$D7,$FC,$54,$D7,$FC,$50,$D7,$C2,$7E,$F0,$3F,$D2,$48,$E5
	.dc.b	$B3,$FE,$11,$8D,$99,$3F,$81,$47,$F4,$92,$38,$24,$7F,$44,$8C,$B6
	.dc.b	$3C,$5A,$FF,$96,$88,$FF,$97,$5F,$F2,$C1,$1F,$F1,$FB,$5F,$F3,$1B
	.dc.b	$C1,$A7,$13,$64,$3F,$C9,$CF,$17,$FC,$DA,$F1,$69,$C4,$D9,$C4,$FF
	.dc.b	$29,$3C,$5F,$F3,$BB,$C5,$A7,$13,$67,$13,$FC,$8A,$F1,$7F,$D0,$1B
	.dc.b	$7F,$9E,$88,$FF,$9F,$5A,$FF,$A0,$1A,$FF,$A1,$1F,$F3,$E1,$1F,$F3
	.dc.b	$1B,$5F,$F3,$63,$5F,$F3,$53,$5F,$F4,$7B,$FE,$09,$23,$F8,$56,$CF
	.dc.b	$E0,$5A,$F8,$63,$FA,$14,$7F,$82,$48,$E0,$91,$FE,$04,$8C,$B6,$3F
	.dc.b	$25,$BF,$FA,$54,$FF,$4D,$95,$FF,$4F,$35,$FF,$50,$B5,$FF,$52,$B5
	.dc.b	$FF,$54,$B5,$FF,$56,$35,$FF,$57,$B5,$FF,$59,$35,$FF,$5A,$B5,$FF
	.dc.b	$5C,$35,$FF,$5D,$B5,$34,$8F,$FA,$51,$F5,$B4,$FF,$5C,$70,$7F,$AC
	.dc.b	$CF,$F5,$C9,$5C,$9F,$E9,$C7,$2D,$21,$FE,$A7,$78,$34,$C9,$FE,$A7
	.dc.b	$72,$D3,$AC,$FF,$8D,$A3,$FC,$90,$7F,$90,$4B,$1F,$DF,$5B,$1B,$8C
	.dc.b	$11,$C1,$0E,$C5,$7F,$D4,$71,$D5,$FE,$21,$C3,$FE,$D3,$3F,$DB,$25
	.dc.b	$7F,$DB,$8D,$7F,$DB,$ED,$7F,$DC,$6D,$7F,$DC,$ED,$7F,$DD,$4D,$7F
	.dc.b	$DD,$AD,$7F,$DE,$0D,$7F,$DE,$6D,$7F,$DE,$CD,$7F,$DF,$2D,$4D,$23
	.dc.b	$FE,$D2,$7D,$6D,$3F,$DE,$DC,$5F,$EF,$13,$FD,$EE,$57,$27,$FB,$69
	.dc.b	$CB,$48,$7F,$B9,$5E,$0D,$32,$7F,$B9,$5C,$B4,$EB,$5F,$E7,$48,$FF
	.dc.b	$43,$1F,$E7,$F2,$C7,$F8,$ED,$B1,$B8,$C1,$1C,$10,$EC,$57,$FD,$C3
	.dc.b	$1D,$5F,$F7,$5C,$00

;--------------------------------------------------------------------------------
;	FTRAPcc.L #<data>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
ftrapccl060_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#FPSP060,-(sp)
	peamsg	'FTRAPCC.L #<DATA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FTRAPcc.L #<data>',13,10
;------------------------------------------------
;	d1	actual result upper
;	d2	actual result lower
;	d3	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result upper,expected result lower,expected status,...
;	a1	expected result upper
;	a2	expected result lower
;	a3	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result upper,expected result lower,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	ftrapccl060_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FTRAPcc.L #<data>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	clr.l	trapv_occurred		;actual result upper
	jsr	([ftrapccl060_execute,za0,d5.l*4])	;EXECUTE
					;actual result lower
	move.l	trapv_occurred,d2	;actual result upper
	fmove.l	fpsr,d3			;actual status
;
	movem.l	(a0),a1-a3		;expected result upper,expected result lower,expected status
;
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	test_double
	lea.l	24(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTRAP'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.L #<data> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	output_double
	lea.l	28(sp),sp
@@:
;
	lea.l	(12,a0),a0		;expected result upper,expected result lower,expected status,...
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
ftrapccl060_execute::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	ftrapccl060_execute_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
ftrapccl060_execute_&cc::
	moveq.l	#3,d1			;actual result upper. 3=too long
	FTRAP&cc.L	#$53815381	;EXECUTE. $5381=subq.l #1,d1. -2=too short. -1=too short
	subq.l	#1,d1			;actual result upper. 0=ok
	subq.l	#1,d1			;actual result upper. 1=too long
	subq.l	#1,d1			;actual result upper. 2=too long
	rts
  .endm
	.cpu	68000

	.align	4
ftrapccl060_expected_compressed::
	.dc.b	$00,$00,$18,$00,$10,$84,$21,$08,$69,$AD,$29,$08,$73,$69,$86,$1B
	.dc.b	$72,$E0,$87,$89,$CB,$82,$07,$2E,$0E,$4E,$1C,$1E,$17,$0E,$0F,$11
	.dc.b	$08,$7B,$9D,$BE,$33,$27,$C4,$72,$FB,$CF,$86,$F9,$5A,$FB,$C8,$E0
	.dc.b	$87,$77,$2F,$21,$F8,$14,$7C,$06,$1F,$33,$4C,$BB,$7E,$D4,$7E,$02
	.dc.b	$31,$1E,$B6,$BE,$96,$BF,$75,$AF,$E2,$23,$D9,$18,$8F,$A5,$1D,$63
	.dc.b	$E2,$36,$3C,$2D,$7E,$E3,$5F,$C3,$35,$FC,$73,$5F,$CB,$3E,$08,$FC
	.dc.b	$8C,$FE,$38,$FC,$51,$27,$F2,$4E,$5F,$A6,$E5,$FC,$BB,$97,$EF,$39
	.dc.b	$7A,$9C,$8E,$24,$38,$99,$38,$99,$38,$9A,$32,$43,$24,$3E,$AB,$E1
	.dc.b	$7F,$5E,$F9,$5F,$C7,$39,$7F,$69,$5F,$AE,$23,$FB,$16,$BF,$B4,$6B
	.dc.b	$FB,$87,$F6,$04,$7F,$2E,$D7,$F4,$0D,$7F,$3C,$D7,$F8,$1B,$F5,$5F
	.dc.b	$E1,$4D,$7F,$86,$B5,$34,$8F,$E5,$CF,$59,$4F,$E2,$2F,$3A,$E3,$19
	.dc.b	$3F,$89,$46,$48,$E4,$FC,$3B,$8A,$FC,$B4,$F2,$6C,$79,$6F,$F8,$91
	.dc.b	$FC,$60,$B5,$FC,$6D,$AF,$F1,$E6,$BF,$C8,$9A,$FF,$26,$6B,$FC,$A5
	.dc.b	$AF,$F2,$C6,$BF,$CB,$DA,$FF,$32,$6B,$FC,$D5,$AF,$F3,$86,$A6,$91
	.dc.b	$FE,$22,$FA,$DA,$7F,$35,$C3,$F9,$81,$FC,$D8,$AE,$4F,$E3,$2E,$5C
	.dc.b	$1F,$C8,$DD,$B4,$D9,$0C,$1F,$D1,$1D,$BF,$D2,$1F,$B8,$7E,$14,$AF
	.dc.b	$D9,$3F,$AA,$38,$AF,$E1,$D0,$8E,$0E,$CF,$E8,$4E,$E3,$F9,$A3,$63
	.dc.b	$F9,$0B,$5F,$EB,$8D,$7F,$AF,$E3,$FD,$4C,$FE,$B2,$57,$FA,$E3,$5F
	.dc.b	$EB,$ED,$7F,$B1,$B5,$FE,$CE,$D7,$FB,$53,$5E,$E7,$F4,$67,$BB,$4E
	.dc.b	$E5,$3B,$94,$EE,$53,$B9,$4E,$E5,$3B,$9F,$D6,$DE,$ED,$3F,$BC,$65
	.dc.b	$FD,$CC,$FE,$F2,$57,$27,$F5,$C7,$2E,$0F,$EC,$CE,$DA,$6C,$86,$0F
	.dc.b	$F0,$3B,$B7,$FC,$16,$FE,$FC,$FE,$10,$57,$EF,$07,$15,$F8,$47,$71
	.dc.b	$F8,$3C,$74,$D8,$DC,$FF,$77,$6C,$7F,$64,$78,$C7,$F0,$67,$F8,$C6
	.dc.b	$7F,$E1,$A2,$4F,$F8,$75,$FF,$0C,$11,$FE,$FE,$D7,$FC,$46,$F0,$69
	.dc.b	$C4,$D9,$0F,$F0,$73,$C5,$FF,$16,$BC,$5A,$71,$36,$71,$3F,$C2,$4F
	.dc.b	$17,$FC,$6E,$F1,$69,$C4,$D9,$C4,$FF,$02,$BC,$5F,$F2,$06,$9F,$E3
	.dc.b	$A2,$3F,$E3,$D6,$BF,$E4,$06,$BF,$E4,$47,$FC,$78,$47,$FC,$46,$D7
	.dc.b	$FC,$58,$D7,$FC,$54,$D7,$FC,$9E,$FF,$49,$23,$F8,$26,$D7,$F0,$46
	.dc.b	$BE,$18,$FC,$14,$7F,$49,$23,$82,$47,$F4,$48,$CB,$63,$F1,$FA,$FF
	.dc.b	$96,$88,$FF,$97,$5F,$F2,$C1,$1F,$F1,$FB,$5F,$F3,$1B,$C1,$A7,$13
	.dc.b	$64,$3F,$C9,$CF,$17,$FC,$DA,$F1,$69,$C4,$D9,$C4,$FF,$29,$3C,$5F
	.dc.b	$F3,$BB,$C5,$A7,$13,$67,$13,$FC,$8A,$F1,$7F,$D0,$1B,$7F,$9E,$88
	.dc.b	$FF,$9F,$5A,$FF,$A0,$1A,$FF,$A1,$1F,$F3,$E1,$1F,$F3,$1B,$5F,$F3
	.dc.b	$63,$5F,$F3,$53,$5F,$F4,$7B,$FE,$09,$23,$FC,$09,$B3,$FD,$11,$AF
	.dc.b	$86,$3F,$81,$47,$F8,$24,$8E,$09,$1F,$E0,$48,$CB,$63,$F8,$7E,$FF
	.dc.b	$E9,$53,$FD,$36,$57,$FD,$3C,$D7,$FD,$42,$D7,$FD,$4A,$D7,$FD,$52
	.dc.b	$D7,$FD,$58,$D7,$FD,$5E,$D7,$FD,$64,$D7,$FD,$6A,$D7,$FD,$70,$D7
	.dc.b	$FD,$76,$D4,$D2,$3F,$E9,$47,$D6,$D3,$FD,$71,$C1,$FE,$B3,$3F,$D7
	.dc.b	$25,$72,$7F,$A7,$1C,$B8,$3F,$D4,$CE,$DA,$6C,$86,$0F,$F6,$3B,$B7
	.dc.b	$FD,$96,$FF,$8F,$CF,$F2,$41,$5F,$E8,$2D,$9F,$F0,$1B,$A8,$FE,$80
	.dc.b	$E6,$3F,$D6,$8D,$8F,$F5,$0B,$5F,$F6,$E3,$5F,$F6,$FF,$0F,$FB,$4C
	.dc.b	$FF,$6C,$95,$FF,$6E,$35,$FF,$6F,$B5,$FF,$71,$B5,$FF,$73,$B5,$FF
	.dc.b	$75,$35,$EE,$7F,$B1,$9E,$ED,$3B,$94,$EE,$53,$B9,$4E,$E5,$3B,$94
	.dc.b	$EE,$7F,$B6,$DE,$ED,$3F,$DF,$1C,$5F,$EF,$33,$FD,$F2,$57,$27,$FB
	.dc.b	$71,$CB,$83,$FD,$CC,$ED,$A6,$C8,$60,$10,$84,$79,$38,$23,$FE,$83
	.dc.b	$3F,$D1,$85,$7F,$C0,$AD,$AF,$F2,$3B,$A8,$FF,$02,$39,$8F,$F7,$C3
	.dc.b	$63,$FD,$CA,$D7,$C0,$79,$9E,$90

;--------------------------------------------------------------------------------
;	FTRAPcc.L #<data>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
ftrapccl88x_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040,-(sp)
	peamsg	'FTRAPCC.L #<DATA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FTRAPcc.L #<data>',13,10
;------------------------------------------------
;	d1	actual result upper
;	d2	actual result lower
;	d3	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result upper,expected result lower,expected status,...
;	a1	expected result upper
;	a2	expected result lower
;	a3	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result upper,expected result lower,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	ftrapccl88x_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FTRAPcc.L #<data>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	clr.l	trapv_occurred		;actual result lower
	jsr	([ftrapccl88x_execute,za0,d5.l*4])	;EXECUTE
					;actual result upper
	move.l	trapv_occurred,d2	;actual result lower
	fmove.l	fpsr,d3			;actual status
;
	movem.l	(a0),a1-a3		;expected result upper,expected result lower,expected status
;
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	test_double
	lea.l	24(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTRAP'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.L #<data> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	output_double
	lea.l	28(sp),sp
@@:
;
	lea.l	(12,a0),a0		;expected result upper,expected result lower,expected status,...
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
ftrapccl88x_execute::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	ftrapccl88x_execute_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
ftrapccl88x_execute_&cc::
	moveq.l	#3,d1			;actual result upper. 3=too long
	FTRAP&cc.L	#$53815381	;EXECUTE. $5381=subq.l #1,d1. -2=too short. -1=too short
	subq.l	#1,d1			;actual result upper. 0=ok
	subq.l	#1,d1			;actual result upper. 1=too long
	subq.l	#1,d1			;actual result upper. 2=too long
	rts
  .endm
	.cpu	68000

	.align	4
ftrapccl88x_expected_compressed::
	.dc.b	$00,$00,$18,$00,$10,$84,$21,$08,$69,$AD,$29,$08,$73,$69,$86,$1B
	.dc.b	$72,$E0,$87,$89,$CB,$82,$07,$2E,$0E,$4E,$1C,$1E,$17,$0E,$0F,$11
	.dc.b	$08,$7B,$9D,$BE,$33,$27,$C4,$72,$FB,$CF,$86,$F9,$5A,$FB,$C8,$E0
	.dc.b	$87,$77,$2F,$21,$F8,$14,$7C,$06,$1F,$33,$4C,$BB,$7E,$D4,$7E,$02
	.dc.b	$31,$1E,$B6,$BE,$96,$BF,$75,$AF,$E2,$23,$D9,$18,$8F,$A5,$1D,$63
	.dc.b	$E2,$36,$3C,$2D,$7E,$E3,$5F,$C3,$35,$FC,$73,$5F,$CB,$3E,$08,$FC
	.dc.b	$8C,$FE,$38,$FC,$51,$27,$F2,$4E,$5F,$A6,$E5,$FC,$BB,$97,$EF,$39
	.dc.b	$7A,$9C,$8E,$24,$38,$99,$38,$99,$38,$9A,$32,$43,$24,$3E,$AB,$E1
	.dc.b	$7F,$5E,$F9,$5F,$C7,$39,$7F,$69,$5F,$AE,$23,$FB,$16,$BF,$B4,$6B
	.dc.b	$FB,$87,$F6,$04,$7F,$2E,$D7,$F4,$0D,$7F,$3C,$D7,$F8,$1B,$F5,$5F
	.dc.b	$E1,$4D,$7F,$86,$B5,$34,$8F,$E5,$CF,$59,$4F,$E2,$2F,$3A,$E3,$19
	.dc.b	$3F,$89,$46,$48,$E4,$FC,$3B,$8A,$FC,$B4,$F2,$6C,$79,$6F,$F8,$91
	.dc.b	$FC,$60,$B5,$FC,$6D,$AF,$F1,$E6,$BF,$C8,$9A,$FF,$26,$6B,$FC,$A5
	.dc.b	$AF,$F2,$C6,$BF,$CB,$DA,$FF,$32,$6B,$FC,$D5,$AF,$F3,$86,$A6,$91
	.dc.b	$FE,$22,$FA,$DA,$7F,$35,$C3,$F9,$81,$FC,$D8,$AE,$4F,$E3,$2E,$5A
	.dc.b	$43,$F9,$23,$C1,$A6,$4F,$E4,$8E,$5A,$75,$AF,$A7,$1F,$85,$3E,$F1
	.dc.b	$63,$ED,$1F,$C3,$DC,$47,$08,$C9,$FD,$06,$38,$B8,$B4,$62,$78,$10
	.dc.b	$F1,$96,$7F,$AE,$E3,$FD,$4C,$FE,$B2,$57,$FA,$E3,$5F,$EB,$ED,$7F
	.dc.b	$B1,$B5,$E6,$7F,$42,$79,$B4,$E6,$53,$99,$4E,$65,$39,$9F,$D4,$DE
	.dc.b	$6D,$39,$94,$E6,$53,$99,$4E,$67,$F5,$97,$FB,$BE,$5F,$DC,$CB,$1F
	.dc.b	$D6,$9A,$FF,$63,$6B,$FD,$85,$E0,$EC,$86,$CA,$6C,$A7,$F5,$47,$BC
	.dc.b	$FE,$D9,$FE,$10,$7F,$01,$2C,$7D,$D6,$C6,$E3,$04,$70,$43,$C0,$57
	.dc.b	$C0,$6D,$FF,$0D,$E7,$FE,$18,$24,$7F,$86,$DF,$F0,$B1,$1F,$EF,$AD
	.dc.b	$7F,$C4,$2F,$06,$9C,$4D,$90,$FF,$06,$3C,$5F,$09,$B3,$89,$0F,$F0
	.dc.b	$73,$B7,$FB,$DB,$5F,$F1,$B3,$E1,$79,$9C,$4D,$9E,$13,$99,$4F,$F1
	.dc.b	$EE,$9F,$E3,$82,$3F,$E3,$B6,$BF,$E3,$E6,$BF,$E4,$27,$FC,$74,$47
	.dc.b	$FC,$42,$D7,$FC,$54,$D7,$FC,$50,$D7,$C2,$7E,$F0,$3F,$D2,$48,$E5
	.dc.b	$B3,$FE,$11,$8D,$99,$3F,$81,$47,$F4,$92,$38,$24,$7F,$44,$8C,$B6
	.dc.b	$3C,$5A,$FF,$96,$88,$FF,$97,$5F,$F2,$C1,$1F,$F1,$FB,$5F,$F3,$1B
	.dc.b	$C1,$A7,$13,$64,$3F,$C9,$CF,$17,$FC,$DA,$F1,$69,$C4,$D9,$C4,$FF
	.dc.b	$29,$3C,$5F,$F3,$BB,$C5,$A7,$13,$67,$13,$FC,$8A,$F1,$7F,$D0,$1B
	.dc.b	$7F,$9E,$88,$FF,$9F,$5A,$FF,$A0,$1A,$FF,$A1,$1F,$F3,$E1,$1F,$F3
	.dc.b	$1B,$5F,$F3,$63,$5F,$F3,$53,$5F,$F4,$7B,$FE,$09,$23,$F8,$56,$CF
	.dc.b	$E0,$5A,$F8,$63,$FA,$14,$7F,$82,$48,$E0,$91,$FE,$04,$8C,$B6,$3F
	.dc.b	$25,$BF,$FA,$54,$FF,$4D,$95,$FF,$4F,$35,$FF,$50,$B5,$FF,$52,$B5
	.dc.b	$FF,$54,$B5,$FF,$56,$35,$FF,$57,$B5,$FF,$59,$35,$FF,$5A,$B5,$FF
	.dc.b	$5C,$35,$FF,$5D,$B5,$34,$8F,$FA,$51,$F5,$B4,$FF,$5C,$70,$7F,$AC
	.dc.b	$CF,$F5,$C9,$5C,$9F,$E9,$C7,$2D,$21,$FE,$A7,$78,$34,$C9,$FE,$A7
	.dc.b	$72,$D3,$AC,$FF,$8D,$A3,$FC,$90,$7F,$90,$4B,$1F,$DF,$5B,$1B,$8C
	.dc.b	$11,$C1,$0E,$C5,$7F,$D4,$71,$D5,$FE,$21,$C3,$FE,$D3,$3F,$DB,$25
	.dc.b	$7F,$DB,$8D,$7F,$DB,$ED,$7F,$DC,$6D,$7F,$DC,$ED,$7F,$DD,$4D,$7F
	.dc.b	$DD,$AD,$7F,$DE,$0D,$7F,$DE,$6D,$7F,$DE,$CD,$7F,$DF,$2D,$4D,$23
	.dc.b	$FE,$D2,$7D,$6D,$3F,$DE,$DC,$5F,$EF,$13,$FD,$EE,$57,$27,$FB,$69
	.dc.b	$CB,$48,$7F,$B9,$5E,$0D,$32,$7F,$B9,$5C,$B4,$EB,$5F,$E7,$48,$FF
	.dc.b	$43,$1F,$E7,$F2,$C7,$F8,$ED,$B1,$B8,$C1,$1C,$10,$EC,$57,$FD,$C3
	.dc.b	$1D,$5F,$F7,$5C,$00

;--------------------------------------------------------------------------------
;	FTRAPcc.W #<data>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
ftrapccw060_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#FPSP060,-(sp)
	peamsg	'FTRAPCC.W #<DATA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FTRAPcc.W #<data>',13,10
;------------------------------------------------
;	d1	actual result upper
;	d2	actual result lower
;	d3	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result upper,expected result lower,expected status,...
;	a1	expected result upper
;	a2	expected result lower
;	a3	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result upper,expected result lower,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	ftrapccw060_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FTRAPcc.W #<data>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	clr.l	trapv_occurred		;actual result upper
	jsr	([ftrapccw060_execute,za0,d5.l*4])	;EXECUTE
					;actual result lower
	move.l	trapv_occurred,d2	;actual result upper
	fmove.l	fpsr,d3			;actual status
;
	movem.l	(a0),a1-a3		;expected result upper,expected result lower,expected status
;
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	test_double
	lea.l	24(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTRAP'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.W #<data> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	output_double
	lea.l	28(sp),sp
@@:
;
	lea.l	(12,a0),a0		;expected result upper,expected result lower,expected status,...
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
ftrapccw060_execute::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	ftrapccw060_execute_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
ftrapccw060_execute_&cc::
	moveq.l	#3,d1			;actual result upper. 3=too long
	FTRAP&cc.W	#$5381		;EXECUTE. $5381=subq.l #1,d1. -1=too short
	subq.l	#1,d1			;actual result upper. 0=ok
	subq.l	#1,d1			;actual result upper. 1=too long
	subq.l	#1,d1			;actual result upper. 2=too long
	rts
  .endm
	.cpu	68000

	.align	4
ftrapccw060_expected_compressed::
	.dc.b	$00,$00,$18,$00,$10,$84,$21,$08,$69,$AD,$29,$08,$73,$69,$86,$1B
	.dc.b	$72,$E0,$87,$89,$CB,$82,$07,$2E,$0E,$4E,$1C,$1E,$17,$0E,$0F,$11
	.dc.b	$08,$7B,$9D,$BE,$33,$27,$C4,$72,$FB,$CF,$86,$F9,$5A,$FB,$C8,$E0
	.dc.b	$87,$77,$2F,$21,$F8,$14,$7C,$06,$1F,$33,$4C,$BB,$7E,$D4,$7E,$02
	.dc.b	$31,$1E,$B6,$BE,$96,$BF,$75,$AF,$E2,$23,$D9,$18,$8F,$A5,$1D,$63
	.dc.b	$E2,$36,$3C,$2D,$7E,$E3,$5F,$C3,$35,$FC,$73,$5F,$CB,$3E,$08,$FC
	.dc.b	$8C,$FE,$38,$FC,$51,$27,$F2,$4E,$5F,$A6,$E5,$FC,$BB,$97,$EF,$39
	.dc.b	$7A,$9C,$8E,$24,$38,$99,$38,$99,$38,$9A,$32,$43,$24,$3E,$AB,$E1
	.dc.b	$7F,$5E,$F9,$5F,$C7,$39,$7F,$69,$5F,$AE,$23,$FB,$16,$BF,$B4,$6B
	.dc.b	$FB,$87,$F6,$04,$7F,$2E,$D7,$F4,$0D,$7F,$3C,$D7,$F8,$1B,$F5,$5F
	.dc.b	$E1,$4D,$7F,$86,$B5,$34,$8F,$E5,$CF,$59,$4F,$E2,$2F,$3A,$E3,$19
	.dc.b	$3F,$89,$46,$48,$E4,$FC,$3B,$8A,$FC,$B4,$F2,$6C,$79,$6F,$F8,$91
	.dc.b	$FC,$60,$B5,$FC,$6D,$AF,$F1,$E6,$BF,$C8,$9A,$FF,$26,$6B,$FC,$A5
	.dc.b	$AF,$F2,$C6,$BF,$CB,$DA,$FF,$32,$6B,$FC,$D5,$AF,$F3,$86,$A6,$91
	.dc.b	$FE,$22,$FA,$DA,$7F,$35,$C3,$F9,$81,$FC,$D8,$AE,$4F,$E3,$2E,$5C
	.dc.b	$1F,$C8,$DD,$B4,$D9,$0C,$1F,$D1,$1D,$BF,$D2,$1F,$B8,$7E,$14,$AF
	.dc.b	$D9,$3F,$AA,$38,$AF,$E1,$D0,$8E,$0E,$CF,$E8,$4E,$E3,$F9,$A3,$63
	.dc.b	$F9,$0B,$5F,$EB,$8D,$7F,$AF,$E3,$FD,$4C,$FE,$B2,$57,$FA,$E3,$5F
	.dc.b	$EB,$ED,$7F,$B1,$B5,$FE,$CE,$D7,$FB,$53,$5E,$E7,$F4,$67,$BB,$4E
	.dc.b	$E5,$3B,$94,$EE,$53,$B9,$4E,$E5,$3B,$9F,$D6,$DE,$ED,$3F,$BC,$65
	.dc.b	$FD,$CC,$FE,$F2,$57,$27,$F5,$C7,$2E,$0F,$EC,$CE,$DA,$6C,$86,$0F
	.dc.b	$F0,$3B,$B7,$FC,$16,$FE,$FC,$FE,$10,$57,$EF,$07,$15,$F8,$47,$71
	.dc.b	$F8,$3C,$74,$D8,$DC,$FF,$77,$6C,$7F,$64,$78,$C7,$F0,$67,$F8,$C6
	.dc.b	$7F,$E1,$A2,$4F,$F8,$75,$FF,$0C,$11,$FE,$FE,$D7,$FC,$46,$F0,$69
	.dc.b	$C4,$D9,$0F,$F0,$73,$C5,$FF,$16,$BC,$5A,$71,$36,$71,$3F,$C2,$4F
	.dc.b	$17,$FC,$6E,$F1,$69,$C4,$D9,$C4,$FF,$02,$BC,$5F,$F2,$06,$9F,$E3
	.dc.b	$A2,$3F,$E3,$D6,$BF,$E4,$06,$BF,$E4,$47,$FC,$78,$47,$FC,$46,$D7
	.dc.b	$FC,$58,$D7,$FC,$54,$D7,$FC,$9E,$FF,$49,$23,$F8,$26,$D7,$F0,$46
	.dc.b	$BE,$18,$FC,$14,$7F,$49,$23,$82,$47,$F4,$48,$CB,$63,$F1,$FA,$FF
	.dc.b	$96,$88,$FF,$97,$5F,$F2,$C1,$1F,$F1,$FB,$5F,$F3,$1B,$C1,$A7,$13
	.dc.b	$64,$3F,$C9,$CF,$17,$FC,$DA,$F1,$69,$C4,$D9,$C4,$FF,$29,$3C,$5F
	.dc.b	$F3,$BB,$C5,$A7,$13,$67,$13,$FC,$8A,$F1,$7F,$D0,$1B,$7F,$9E,$88
	.dc.b	$FF,$9F,$5A,$FF,$A0,$1A,$FF,$A1,$1F,$F3,$E1,$1F,$F3,$1B,$5F,$F3
	.dc.b	$63,$5F,$F3,$53,$5F,$F4,$7B,$FE,$09,$23,$FC,$09,$B3,$FD,$11,$AF
	.dc.b	$86,$3F,$81,$47,$F8,$24,$8E,$09,$1F,$E0,$48,$CB,$63,$F8,$7E,$FF
	.dc.b	$E9,$53,$FD,$36,$57,$FD,$3C,$D7,$FD,$42,$D7,$FD,$4A,$D7,$FD,$52
	.dc.b	$D7,$FD,$58,$D7,$FD,$5E,$D7,$FD,$64,$D7,$FD,$6A,$D7,$FD,$70,$D7
	.dc.b	$FD,$76,$D4,$D2,$3F,$E9,$47,$D6,$D3,$FD,$71,$C1,$FE,$B3,$3F,$D7
	.dc.b	$25,$72,$7F,$A7,$1C,$B8,$3F,$D4,$CE,$DA,$6C,$86,$0F,$F6,$3B,$B7
	.dc.b	$FD,$96,$FF,$8F,$CF,$F2,$41,$5F,$E8,$2D,$9F,$F0,$1B,$A8,$FE,$80
	.dc.b	$E6,$3F,$D6,$8D,$8F,$F5,$0B,$5F,$F6,$E3,$5F,$F6,$FF,$0F,$FB,$4C
	.dc.b	$FF,$6C,$95,$FF,$6E,$35,$FF,$6F,$B5,$FF,$71,$B5,$FF,$73,$B5,$FF
	.dc.b	$75,$35,$EE,$7F,$B1,$9E,$ED,$3B,$94,$EE,$53,$B9,$4E,$E5,$3B,$94
	.dc.b	$EE,$7F,$B6,$DE,$ED,$3F,$DF,$1C,$5F,$EF,$33,$FD,$F2,$57,$27,$FB
	.dc.b	$71,$CB,$83,$FD,$CC,$ED,$A6,$C8,$60,$10,$84,$79,$38,$23,$FE,$83
	.dc.b	$3F,$D1,$85,$7F,$C0,$AD,$AF,$F2,$3B,$A8,$FF,$02,$39,$8F,$F7,$C3
	.dc.b	$63,$FD,$CA,$D7,$C0,$79,$9E,$90

;--------------------------------------------------------------------------------
;	FTRAPcc.W #<data>
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
ftrapccw88x_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#MC68881+MC68882+MC68040+FPSP040,-(sp)
	peamsg	'FTRAPCC.W #<DATA>'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FTRAPcc.W #<data>',13,10
;------------------------------------------------
;	d1	actual result upper
;	d2	actual result lower
;	d3	actual status
;	d5	cc=0..31
;	d6	fpsr=(mzin=0..15)<<24
;	d7	0=failed,1=successful
;	a0	expected result upper,expected result lower,expected status,...
;	a1	expected result upper
;	a2	expected result lower
;	a3	expected status
;------------------------------------------------
	lea.l	push_decompressed,a0	;expected result upper,expected result lower,expected status,...
;decompress data
	move.l	a0,-(sp)
	pea.l	ftrapccw88x_expected_compressed
	jbsr	decompress
	addq.l	#8,sp
	move.l	#0<<24,d6		;fpsr=(mzin=0..15)<<24
66:
	moveq.l	#0,d5			;cc=0..31
55:
;FTRAPcc.W #<data>
	fmove.l	#0,fpcr
	fmove.l	d6,fpsr			;fpsr
	clr.l	trapv_occurred		;actual result lower
	jsr	([ftrapccw88x_execute,za0,d5.l*4])	;EXECUTE
					;actual result upper
	move.l	trapv_occurred,d2	;actual result lower
	fmove.l	fpsr,d3			;actual status
;
	movem.l	(a0),a1-a3		;expected result upper,expected result lower,expected status
;
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	test_double
	lea.l	24(sp),sp
	move.l	d0,d7			;0=failed,1=successful
;
	move.l	d7,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTRAP'
	putstr	(uppercase_cc,za0,d5.l*4)
	putmsg	'.W #<data> @'
	move.l	d6,-(sp)		;fpsr
	jbsr	printfpsr
	addq.l	#4,sp
	putcrlf
	move.l	d7,-(sp)		;0=failed,1=successful
	movem.l	d1-d3/a1-a3,-(sp)	;actual result upper,actual result lower,actual status,expected result upper,expected result lower,expected status
	jbsr	output_double
	lea.l	28(sp),sp
@@:
;
	lea.l	(12,a0),a0		;expected result upper,expected result lower,expected status,...
;
	addq.w	#1,d5			;cc++
	cmp.w	#31,d5			;cc=0..31
	bls	55b
;
	add.l	#1<<24,d6		;mzin++
	cmp.l	#15<<24,d6		;fpsr=(mzin=0..15)<<24
	bls	66b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts

	.align	4
ftrapccw88x_execute::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	ftrapccw88x_execute_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
ftrapccw88x_execute_&cc::
	moveq.l	#3,d1			;actual result upper. 3=too long
	FTRAP&cc.W	#$5381		;EXECUTE. $5381=subq.l #1,d1. -1=too short
	subq.l	#1,d1			;actual result upper. 0=ok
	subq.l	#1,d1			;actual result upper. 1=too long
	subq.l	#1,d1			;actual result upper. 2=too long
	rts
  .endm
	.cpu	68000

	.align	4
ftrapccw88x_expected_compressed::
	.dc.b	$00,$00,$18,$00,$10,$84,$21,$08,$69,$AD,$29,$08,$73,$69,$86,$1B
	.dc.b	$72,$E0,$87,$89,$CB,$82,$07,$2E,$0E,$4E,$1C,$1E,$17,$0E,$0F,$11
	.dc.b	$08,$7B,$9D,$BE,$33,$27,$C4,$72,$FB,$CF,$86,$F9,$5A,$FB,$C8,$E0
	.dc.b	$87,$77,$2F,$21,$F8,$14,$7C,$06,$1F,$33,$4C,$BB,$7E,$D4,$7E,$02
	.dc.b	$31,$1E,$B6,$BE,$96,$BF,$75,$AF,$E2,$23,$D9,$18,$8F,$A5,$1D,$63
	.dc.b	$E2,$36,$3C,$2D,$7E,$E3,$5F,$C3,$35,$FC,$73,$5F,$CB,$3E,$08,$FC
	.dc.b	$8C,$FE,$38,$FC,$51,$27,$F2,$4E,$5F,$A6,$E5,$FC,$BB,$97,$EF,$39
	.dc.b	$7A,$9C,$8E,$24,$38,$99,$38,$99,$38,$9A,$32,$43,$24,$3E,$AB,$E1
	.dc.b	$7F,$5E,$F9,$5F,$C7,$39,$7F,$69,$5F,$AE,$23,$FB,$16,$BF,$B4,$6B
	.dc.b	$FB,$87,$F6,$04,$7F,$2E,$D7,$F4,$0D,$7F,$3C,$D7,$F8,$1B,$F5,$5F
	.dc.b	$E1,$4D,$7F,$86,$B5,$34,$8F,$E5,$CF,$59,$4F,$E2,$2F,$3A,$E3,$19
	.dc.b	$3F,$89,$46,$48,$E4,$FC,$3B,$8A,$FC,$B4,$F2,$6C,$79,$6F,$F8,$91
	.dc.b	$FC,$60,$B5,$FC,$6D,$AF,$F1,$E6,$BF,$C8,$9A,$FF,$26,$6B,$FC,$A5
	.dc.b	$AF,$F2,$C6,$BF,$CB,$DA,$FF,$32,$6B,$FC,$D5,$AF,$F3,$86,$A6,$91
	.dc.b	$FE,$22,$FA,$DA,$7F,$35,$C3,$F9,$81,$FC,$D8,$AE,$4F,$E3,$2E,$5A
	.dc.b	$43,$F9,$23,$C1,$A6,$4F,$E4,$8E,$5A,$75,$AF,$A7,$1F,$85,$3E,$F1
	.dc.b	$63,$ED,$1F,$C3,$DC,$47,$08,$C9,$FD,$06,$38,$B8,$B4,$62,$78,$10
	.dc.b	$F1,$96,$7F,$AE,$E3,$FD,$4C,$FE,$B2,$57,$FA,$E3,$5F,$EB,$ED,$7F
	.dc.b	$B1,$B5,$E6,$7F,$42,$79,$B4,$E6,$53,$99,$4E,$65,$39,$9F,$D4,$DE
	.dc.b	$6D,$39,$94,$E6,$53,$99,$4E,$67,$F5,$97,$FB,$BE,$5F,$DC,$CB,$1F
	.dc.b	$D6,$9A,$FF,$63,$6B,$FD,$85,$E0,$EC,$86,$CA,$6C,$A7,$F5,$47,$BC
	.dc.b	$FE,$D9,$FE,$10,$7F,$01,$2C,$7D,$D6,$C6,$E3,$04,$70,$43,$C0,$57
	.dc.b	$C0,$6D,$FF,$0D,$E7,$FE,$18,$24,$7F,$86,$DF,$F0,$B1,$1F,$EF,$AD
	.dc.b	$7F,$C4,$2F,$06,$9C,$4D,$90,$FF,$06,$3C,$5F,$09,$B3,$89,$0F,$F0
	.dc.b	$73,$B7,$FB,$DB,$5F,$F1,$B3,$E1,$79,$9C,$4D,$9E,$13,$99,$4F,$F1
	.dc.b	$EE,$9F,$E3,$82,$3F,$E3,$B6,$BF,$E3,$E6,$BF,$E4,$27,$FC,$74,$47
	.dc.b	$FC,$42,$D7,$FC,$54,$D7,$FC,$50,$D7,$C2,$7E,$F0,$3F,$D2,$48,$E5
	.dc.b	$B3,$FE,$11,$8D,$99,$3F,$81,$47,$F4,$92,$38,$24,$7F,$44,$8C,$B6
	.dc.b	$3C,$5A,$FF,$96,$88,$FF,$97,$5F,$F2,$C1,$1F,$F1,$FB,$5F,$F3,$1B
	.dc.b	$C1,$A7,$13,$64,$3F,$C9,$CF,$17,$FC,$DA,$F1,$69,$C4,$D9,$C4,$FF
	.dc.b	$29,$3C,$5F,$F3,$BB,$C5,$A7,$13,$67,$13,$FC,$8A,$F1,$7F,$D0,$1B
	.dc.b	$7F,$9E,$88,$FF,$9F,$5A,$FF,$A0,$1A,$FF,$A1,$1F,$F3,$E1,$1F,$F3
	.dc.b	$1B,$5F,$F3,$63,$5F,$F3,$53,$5F,$F4,$7B,$FE,$09,$23,$F8,$56,$CF
	.dc.b	$E0,$5A,$F8,$63,$FA,$14,$7F,$82,$48,$E0,$91,$FE,$04,$8C,$B6,$3F
	.dc.b	$25,$BF,$FA,$54,$FF,$4D,$95,$FF,$4F,$35,$FF,$50,$B5,$FF,$52,$B5
	.dc.b	$FF,$54,$B5,$FF,$56,$35,$FF,$57,$B5,$FF,$59,$35,$FF,$5A,$B5,$FF
	.dc.b	$5C,$35,$FF,$5D,$B5,$34,$8F,$FA,$51,$F5,$B4,$FF,$5C,$70,$7F,$AC
	.dc.b	$CF,$F5,$C9,$5C,$9F,$E9,$C7,$2D,$21,$FE,$A7,$78,$34,$C9,$FE,$A7
	.dc.b	$72,$D3,$AC,$FF,$8D,$A3,$FC,$90,$7F,$90,$4B,$1F,$DF,$5B,$1B,$8C
	.dc.b	$11,$C1,$0E,$C5,$7F,$D4,$71,$D5,$FE,$21,$C3,$FE,$D3,$3F,$DB,$25
	.dc.b	$7F,$DB,$8D,$7F,$DB,$ED,$7F,$DC,$6D,$7F,$DC,$ED,$7F,$DD,$4D,$7F
	.dc.b	$DD,$AD,$7F,$DE,$0D,$7F,$DE,$6D,$7F,$DE,$CD,$7F,$DF,$2D,$4D,$23
	.dc.b	$FE,$D2,$7D,$6D,$3F,$DE,$DC,$5F,$EF,$13,$FD,$EE,$57,$27,$FB,$69
	.dc.b	$CB,$48,$7F,$B9,$5E,$0D,$32,$7F,$B9,$5C,$B4,$EB,$5F,$E7,$48,$FF
	.dc.b	$43,$1F,$E7,$F2,$C7,$F8,$ED,$B1,$B8,$C1,$1C,$10,$EC,$57,$FD,$C3
	.dc.b	$1D,$5F,$F7,$5C,$00

;--------------------------------------------------------------------------------
;	FTST.X FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
ftst_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#63,-(sp)
	peamsg	'FTST.X FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FTST.X FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected status,...
;	fp2	source
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	ftst_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
	movea.l	a0,a2			;source handle,...
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
	movea.l	a0,a4			;expected status,...
;
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	ftst.x	fp2		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	move.l	(a4),d5			;expected status
;
	move.l	#-1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	move.l	d7,-(sp)		;actual status
	jbsr	test_status
	lea.l	(12,sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTST.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	' @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	move.l	d7,-(sp)		;actual status
	jbsr	output_status
	lea.l	(12,sp),sp
@@:
;
	addq.l	#4,a4			;expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
ftst_data_compressed::
	.dc.b	$00,$01,$42,$C8,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CC,$7B,$39,$F8,$ED,$7C,$56,$FE,$1F,$87,$BF,$C7
	.dc.b	$DC,$E5,$EE,$73,$F7,$3A,$7B,$9D,$7D,$CE,$DE,$E7,$7F,$73,$C1,$EE
	.dc.b	$78,$7D,$CF,$17,$B9,$E3,$F7,$3C,$9E,$E7,$97,$DC,$F3,$7B,$9F,$DE
	.dc.b	$FE,$6B,$FA,$A0,$F8,$8C,$7C,$46,$7E,$23,$5F,$11,$BF,$88,$E1,$F1
	.dc.b	$1C,$7E,$17,$97,$C1,$73,$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F
	.dc.b	$82,$F0,$7C,$17,$87,$E0,$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E
	.dc.b	$0B,$CD,$F0,$5E,$7F,$82,$FF,$09,$7C,$DF,$E1,$3E,$17,$FC,$3C,$BE
	.dc.b	$0B,$3E,$86,$BE,$0B,$7F,$05,$C3,$E0,$B8,$FC,$17,$2F,$82,$E7,$F0
	.dc.b	$5D,$3E,$0B,$AF,$C1,$76,$F8,$2E,$FF,$05,$E0,$F8,$2F,$0F,$C1,$78
	.dc.b	$BE,$0B,$C7,$F0,$5E,$4F,$82,$F2,$FC,$17,$9B,$E0,$BC,$FF,$05,$FE
	.dc.b	$29,$F9,$CF,$C2,$F4,$7F,$8D,$5F,$F7,$89,$FE,$AB,$3F,$CC,$67,$F5
	.dc.b	$53,$F8,$69,$FC,$31,$3F,$C3,$1E,$5E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7F,$E4,$1F,$63,$FC,$34,$7E,$1F,$83,$FC,$9C
	.dc.b	$BD,$9D,$3B,$F6,$78,$7B,$3C,$7D,$9E,$5E,$CF,$3F,$67,$A7,$B3,$D7
	.dc.b	$D9,$ED,$EC,$F7,$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9
	.dc.b	$EC,$F9,$7D,$9F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$F9,$67,$E7,$7F,$11
	.dc.b	$CE,$7F,$10,$FF,$99,$D7,$B1,$BF,$63,$87,$B1,$C7,$D8,$E5,$EC,$73
	.dc.b	$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1,$EC,$78,$7D,$8F,$17
	.dc.b	$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7F
	.dc.b	$9C,$7E,$7B,$F1,$3C,$E5,$F0,$99,$F8,$4D,$7C,$26,$FE,$13,$87,$C2
	.dc.b	$71,$F8,$4E,$5F,$09,$CF,$DF,$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF
	.dc.b	$07,$BF,$E1,$F7,$FC,$5E,$FF,$8F,$DF,$F2,$7B,$FE,$5F,$7F,$CD,$EF
	.dc.b	$F9,$FD,$FF,$F4,$47,$CF,$FF,$98,$4F,$C4,$74,$1F,$C4,$3F,$E9,$77
	.dc.b	$FD,$E6,$7F,$AB,$4F,$F0,$E9,$FD,$54,$FE,$A8,$F1,$F6,$BA,$FB,$5D
	.dc.b	$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF
	.dc.b	$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$F4,$FB,$5F,$EA,$1F,$0F,$F1
	.dc.b	$27,$E2,$38,$4F,$E2,$1F,$F5,$82,$F6,$38,$7B,$1C,$7D,$8E,$5E,$C7
	.dc.b	$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E,$C7,$87,$D8,$F1
	.dc.b	$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7,$F6,$3D,$1E,$C7
	.dc.b	$FA,$E7,$E8,$3F,$12,$7E,$23,$9B,$FD,$8C,$BD,$7D,$73,$EF,$D7,$E1
	.dc.b	$EB,$F1,$F5,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF,$DF,$D7
	.dc.b	$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5,$FC,$DE
	.dc.b	$BF,$9F,$D7,$FF,$68,$7D,$0F,$FA,$A4,$FC,$49,$F8,$8E,$72,$F8,$1D
	.dc.b	$FC,$0F,$0F,$81,$E3,$F0,$3C,$BE,$07,$9F,$B9,$D3,$DC,$EB,$EE,$76
	.dc.b	$F7,$3B,$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C
	.dc.b	$BE,$E7,$9B,$DC,$F3,$FB,$9E,$8F,$73,$FD,$D1,$F4,$5F,$C4,$0F,$C4
	.dc.b	$1F,$87,$E7,$3F,$87,$7F,$DF,$0F,$FB,$CC,$FF,$56,$1F,$E6,$47,$7E
	.dc.b	$CF,$4F,$67,$AF,$B3,$DB,$D9,$EF,$EC,$F8,$3D,$9F,$0F,$B3,$E2,$F6
	.dc.b	$7C,$7E,$CF,$93,$D9,$F2,$FB,$3E,$6F,$67,$CF,$EC,$FA,$3D,$9F,$F8
	.dc.b	$01,$EE,$FF,$10,$3F,$11,$C3,$F9,$DF,$67,$3E,$CE,$BD,$9D,$FB,$3C
	.dc.b	$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB
	.dc.b	$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B
	.dc.b	$D9,$72,$CF,$A3,$FF,$78,$1F,$C4,$0F,$C4,$1F,$87,$75,$EC,$70,$51
	.dc.b	$FE,$AC,$3F,$CC,$87,$F1,$03,$FC,$C4,$F5,$F5,$38,$43,$D4,$BE,$A6
	.dc.b	$3D,$4C,$FA,$9A,$F5,$37,$EA,$70,$F5,$38,$FA,$9C,$BD,$4E,$7E,$A7
	.dc.b	$4F,$53,$AF,$A9,$E5,$FA,$4F,$EB,$27,$F1,$13,$F1,$47,$E2,$78,$4B
	.dc.b	$DF,$E1,$EF,$F1,$F7,$F9,$7A,$FC,$FD,$7E,$9E,$BF,$5F,$5F,$B7,$AF
	.dc.b	$DF,$D7,$F0,$7A,$FE,$1F,$5F,$C5,$EB,$F8,$FD,$7F,$27,$AF,$E5,$F5
	.dc.b	$FC,$DE,$BF,$9F,$D7,$F9,$9F,$4B,$FD,$60,$FC,$49,$FC,$3F,$97,$EB
	.dc.b	$BF,$CA,$9B,$EA,$F0,$F5,$78,$FA,$BC,$BD,$5E,$7E,$AF,$4F,$57,$AF
	.dc.b	$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E,$AF,$93
	.dc.b	$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FE,$0F,$D5,$FC,$27,$21,$FE,$B2
	.dc.b	$7F,$11,$3F,$14,$E5,$E1,$EE,$F1,$F7,$79,$7B,$BC,$FD,$DE,$9E,$CF
	.dc.b	$5F,$63,$B7,$B1,$DF,$D8,$F0,$7B,$1E,$1F,$63,$C5,$EC,$78,$FD,$8F
	.dc.b	$27,$B1,$E5,$F6,$3C,$DE,$C7,$9F,$D8,$F4,$7B,$1F,$9B,$FA,$6F,$F0
	.dc.b	$F9,$FD,$5C,$FE,$20,$7E,$1C,$FC,$37,$2F,$EA,$17,$AF,$C7,$D7,$E5
	.dc.b	$EB,$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D
	.dc.b	$7F,$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$F6
	.dc.b	$5F,$4F,$FE,$1E,$3F,$87,$9F,$87,$E4,$3F,$87,$6F,$AF,$C3,$D7,$E3
	.dc.b	$EB,$F2,$F5,$F9,$FA,$FD,$3D,$7E,$BE,$BF,$6F,$5F,$BF,$AF,$E0,$F5
	.dc.b	$FC,$3E,$BF,$8B,$D7,$F1,$FA,$FE,$4F,$5F,$CB,$EB,$F9,$BD,$7F,$3F
	.dc.b	$AF,$FC,$27,$EA,$3F,$0E,$7F,$87,$8F,$EA,$C7,$E1,$8F,$C5,$72,$7F
	.dc.b	$16,$5E,$BF,$1F,$5F,$97,$AF,$CF,$D7,$E9,$EB,$F5,$F5,$FB,$7A,$FD
	.dc.b	$FD,$7F,$07,$AF,$E1,$F5,$FC,$5E,$BF,$8F,$D7,$F2,$7A,$FE,$5F,$5F
	.dc.b	$CD,$EB,$F9,$FD,$7F,$47,$AF,$FC,$87,$C5,$FF,$32,$9F,$E1,$D3,$FA
	.dc.b	$B1,$FC,$34,$FC,$3F,$29,$FC,$3B,$FC,$C1,$FE,$AC,$3D,$4E,$9E,$A7
	.dc.b	$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78,$FD,$4F
	.dc.b	$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F4,$7A,$9F,$CE,$7E,$A7,$FC
	.dc.b	$CA,$7F,$56,$3F,$10,$7E,$1C,$F4,$71,$7F,$46,$5E,$A7,$1F,$53,$97
	.dc.b	$A9,$CF,$D4,$E9,$EA,$75,$F5,$3B,$7A,$9D,$FD,$4F,$07,$A9,$E1,$F5
	.dc.b	$3C,$5E,$A7,$8F,$D4,$F2,$7A,$9E,$5F,$53,$CD,$EA,$79,$FD,$4F,$EA
	.dc.b	$1F,$55,$FC,$38,$FF,$32,$9F,$E1,$D3,$F8,$69,$F8,$8E,$43,$F8,$87
	.dc.b	$FA,$FA,$F4,$F9,$FA,$7D,$3D,$3E,$BE,$9F,$6F,$4F,$BF,$A7,$E0,$F4
	.dc.b	$FC,$3E,$9F,$8B,$D3,$F1,$FA,$7E,$4F,$4F,$CB,$E9,$F9,$BD,$3F,$3F
	.dc.b	$A7,$E8,$F4,$FD,$3E,$9F,$F6,$9F,$AB,$FF,$56,$9F,$E6,$53,$FC,$3C
	.dc.b	$7F,$55,$3F,$12,$7E,$23,$90,$FF,$87,$4F,$C4,$3F,$DE,$97,$A5,$D7
	.dc.b	$D2,$ED,$E9,$77,$F4,$BC,$1E,$97,$87,$D2,$F1,$7A,$5E,$3F,$4B,$C9
	.dc.b	$E9,$79,$7D,$2F,$37,$A5,$E7,$F4,$BD,$1E,$97,$F8,$07,$DD,$FF,$56
	.dc.b	$9F,$E1,$E3,$F8,$81,$FC,$3C,$FC,$4F,$0F,$F6,$DF,$4B,$87,$A5,$C7
	.dc.b	$D2,$E5,$E9,$73,$F4,$BA,$7A,$5D,$7D,$2E,$DE,$97,$7F,$4B,$C1,$E9
	.dc.b	$78,$7D,$2F,$17,$A5,$E3,$F4,$BC,$9E,$97,$97,$D2,$F3,$7A,$5F,$E1
	.dc.b	$7F,$AC,$FE,$AA,$7F,$AB,$0F,$F3,$21,$FD,$50,$FC,$41,$F8,$7E,$33
	.dc.b	$F8,$77,$FC,$4E,$FF,$AB,$0F,$F0,$EB,$AF,$3F,$B7,$9F,$DF,$CF,$F0
	.dc.b	$79,$FE,$1F,$3F,$C5,$E7,$F8,$FC,$FF,$27,$9F,$E5,$F3,$FC,$DE,$7F
	.dc.b	$9F,$CF,$F4,$79,$FE,$9F,$3F,$FC,$61,$F5,$BF,$EF,$43,$FD,$58,$7F
	.dc.b	$99,$4F,$F0,$E1,$FC,$40,$FC,$41,$FE,$63,$3F,$0F,$C0,$7F,$0E,$FF
	.dc.b	$90,$97,$9D,$DB,$CE,$EF,$E7,$78,$3C,$EF,$0F,$9D,$E2,$F3,$BC,$7E
	.dc.b	$77,$93,$CE,$F2,$F9,$DE,$6F,$3B,$CF,$E7,$7A,$3C,$EF,$F2,$47,$D7
	.dc.b	$7F,$BD,$0F,$F3,$29,$FD,$5C,$FE,$AC,$7E,$20,$FE,$1F,$C1,$FE,$56
	.dc.b	$5E,$6F,$2F,$37,$9F,$9B,$D3,$CD,$EB,$E6,$F6,$F3,$7B,$F9,$BE,$0F
	.dc.b	$37,$C3,$E6,$F8,$BC,$DF,$1F,$9B,$E4,$F3,$7C,$BE,$6F,$9B,$CD,$F3
	.dc.b	$F9,$BF,$E5,$FF,$67,$FC,$3A,$7F,$BD,$0F,$F5,$61,$FE,$1C,$3F,$88
	.dc.b	$1F,$89,$3F,$11,$C7,$FD,$63,$D4,$E5,$EA,$73,$F3,$FA,$79,$FD,$7C
	.dc.b	$FE,$DE,$7F,$7F,$3F,$C1,$E7,$F8,$7C,$FF,$17,$9F,$E3,$F3,$FC,$9E
	.dc.b	$7F,$97,$CF,$F3,$79,$FE,$7F,$3F,$FC,$ED,$F5,$EF,$97,$BF,$27,$8F
	.dc.b	$90,$CF,$C4,$EB,$D7,$DF,$C1,$70,$F8,$2E,$3E,$FF,$2F,$6F,$9F,$B1
	.dc.b	$D3,$D8,$EB,$EC,$76,$F6,$3B,$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F
	.dc.b	$1F,$B1,$E4,$F6,$3C,$BE,$C7,$9B,$D8,$FF,$46,$7D,$83,$E6,$2F,$C8
	.dc.b	$E3,$E3,$33,$F1,$9A,$F8,$5D,$FC,$0F,$0F,$7B,$8F,$BD,$CB,$DE,$E7
	.dc.b	$EF,$74,$F7,$BA,$FB,$DD,$BD,$EE,$FE,$F7,$83,$DE,$F0,$FB,$DE,$2F
	.dc.b	$7B,$C7,$EF,$79,$3D,$EF,$2F,$BD,$E6,$F7,$BF,$D4,$5F,$63,$FE,$60
	.dc.b	$1F,$31,$8F,$96,$CF,$C8,$EB,$E3,$77,$F0,$7C,$3E,$23,$8F,$C1,$F2
	.dc.b	$F8,$3E,$7F,$07,$D3,$E0,$FA,$FC,$1F,$6F,$83,$EF,$F0,$7E,$0F,$83
	.dc.b	$F0,$FC,$1F,$8B,$E0,$FC,$7F,$07,$E4,$F8,$3F,$2F,$C1,$F9,$BE,$0F
	.dc.b	$CF,$F0,$7F,$EB,$9F,$1F,$F0,$9F,$09,$FE,$C2,$5F,$05,$9F,$82,$D7
	.dc.b	$C1,$6F,$E0,$B8,$7C,$17,$1F,$82,$E5,$F0,$5C,$FE,$0B,$A7,$C1,$75
	.dc.b	$F8,$2E,$DF,$05,$DF,$E0,$BC,$1F,$05,$E1,$F8,$2F,$17,$C1,$78,$FE
	.dc.b	$0B,$C9,$F0,$5E,$5F,$82,$F3,$7C,$17,$9F,$E0,$BF,$DA,$1F,$65,$F8
	.dc.b	$5E,$8F,$F6,$CB,$FE,$EE,$3F,$CC,$07,$F9,$7C,$FE,$A2,$7F,$0A,$4F
	.dc.b	$F0,$A7,$8F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B,$FB,$9E,$0F,$73,$C3
	.dc.b	$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7,$9B,$DC,$F3,$FB,$9E
	.dc.b	$8F,$73,$FD,$D1,$F6,$7F,$EA,$63,$F0,$FC,$3F,$EF,$25,$ED,$EB,$DB
	.dc.b	$DF,$B7,$C3,$DB,$E3,$ED,$F2,$F6,$F9,$FB,$7D,$3D,$BE,$BE,$DF,$6F
	.dc.b	$6F,$BF,$B7,$E0,$F6,$FC,$3E,$DF,$8B,$DB,$F1,$FB,$7E,$4F,$6F,$CB
	.dc.b	$ED,$F9,$BD,$BF,$3F,$B7,$E8,$F6,$FF,$E0,$07,$DA,$7E,$23,$A0,$FE
	.dc.b	$21,$18,$F6,$B7,$ED,$70,$F6,$B8,$FB,$5C,$BD,$AE,$7E,$D7,$4F,$6B
	.dc.b	$AF,$B5,$DB,$DA,$EF,$ED,$78,$3D,$AF,$0F,$B5,$E2,$F6,$BC,$7E,$D7
	.dc.b	$93,$DA,$F2,$FB,$5E,$6F,$6B,$CF,$ED,$7A,$3D,$AF,$2F,$2F,$FA,$E3
	.dc.b	$E1,$2F,$1B,$E1,$2F,$C2,$63,$E1,$33,$F0,$9A,$F8,$1D,$FC,$27,$0F
	.dc.b	$80,$E3,$F0,$1C,$BE,$03,$9F,$C0,$74,$F8,$0E,$BF,$01,$DB,$E0,$3B
	.dc.b	$FC,$07,$83,$E0,$3C,$3F,$01,$E2,$F8,$0F,$1F,$C0,$79,$3E,$03,$CB
	.dc.b	$F0,$1E,$8F,$B5,$FF,$76,$9F,$88,$E9,$3F,$88,$7E,$8B,$FE,$EE,$3F
	.dc.b	$CC,$47,$F8,$6C,$FE,$A6,$3D,$BE,$9E,$DF,$5F,$6F,$B7,$B7,$DF,$DB
	.dc.b	$F0,$7B,$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$E5,$F6,$FC,$DE
	.dc.b	$DF,$9F,$DB,$F4,$7B,$7E,$9F,$6F,$E8,$FD,$B7,$E2,$4F,$C4,$71,$97
	.dc.b	$B7,$AF,$6F,$7E,$DF,$0F,$2F,$8F,$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA
	.dc.b	$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB,$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9
	.dc.b	$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE,$DF,$E2,$3E,$DF,$F8,$87,$39,$FC
	.dc.b	$43,$FA,$07,$FD,$E0,$7F,$BB,$CF,$F3,$09,$FC,$44,$FE,$20,$F1,$F5
	.dc.b	$7A,$7A,$BD,$7D,$5E,$DE,$AF,$7F,$57,$C1,$EA,$F8,$7D,$5F,$17,$AB
	.dc.b	$E3,$F5,$7C,$9E,$AF,$97,$D5,$F3,$7A,$BE,$7F,$57,$D1,$EA,$FE,$83
	.dc.b	$DE,$FD,$0F,$0F,$FA,$87,$E1,$B1,$F0,$79,$F8,$3D,$7C,$1E,$FE,$0F
	.dc.b	$86,$1C,$7D,$AE,$5E,$D7,$3F,$6B,$A7,$B5,$D7,$DA,$ED,$ED,$77,$F6
	.dc.b	$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F,$6B,$C9,$ED,$79,$7D,$AF,$37
	.dc.b	$B5,$E7,$F6,$BF,$6F,$F7,$1F,$C3,$CF,$C3,$9F,$86,$E8,$FE,$1E,$BD
	.dc.b	$9D,$FB,$3C,$3D,$9E,$3E,$CF,$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6
	.dc.b	$F6,$7B,$FB,$3E,$0F,$67,$C3,$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C
	.dc.b	$BE,$CF,$9B,$D9,$F3,$FB,$3F,$C3,$BE,$E7,$F0,$E7,$E1,$B9,$8B,$E0
	.dc.b	$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0,$B9,$7B,$DC,$FD,$EE,$9E,$F3
	.dc.b	$F8,$06,$17,$EC,$FE,$E9,$35,$FE,$3E,$D7,$F9,$AB,$53,$48,$FF,$26
	.dc.b	$F8,$DF,$C3,$91,$C1,$0F,$E7,$EE,$5A,$64,$FE,$8C,$47,$4D,$34,$E0
	.dc.b	$8E,$08,$74,$21,$C4,$AF,$06,$9C,$0D,$11,$D3,$4D,$38,$23,$82,$1D
	.dc.b	$08,$71,$2B,$C1,$A7,$03,$44,$74,$D3,$4E,$08,$E0,$87,$42,$1C,$4A
	.dc.b	$F0,$69,$C0,$D1,$1D,$34,$D3,$82,$38,$21,$D0,$87,$12,$BC,$1A,$70
	.dc.b	$34,$47,$4D,$34,$E0,$8E,$08,$74,$21,$C4,$AF,$06,$9C,$0D,$11,$D3
	.dc.b	$4D,$38,$23,$82,$1D,$08,$71,$2B,$C1,$A7,$03,$44,$74,$D3,$4E,$08
	.dc.b	$E0,$87,$42,$1C,$4A,$F0,$69,$C0,$D1,$1D,$34,$D3,$82,$38,$21,$D0
	.dc.b	$87,$12,$BC,$1A,$70,$34,$47,$4D,$34,$E0,$8E,$08,$74,$21,$C4,$AF
	.dc.b	$06,$9C,$0D,$11,$D3,$4D,$38,$23,$82,$1D,$08,$71,$2B,$C1,$A7,$03
	.dc.b	$44,$74,$D3,$4E,$08,$E0,$87,$42,$1C,$4A,$F0,$69,$C0,$D1,$1D,$34
	.dc.b	$D3,$82,$38,$21,$D0,$87,$12,$BC,$1A,$70,$34,$47,$4D,$34,$E0,$8E
	.dc.b	$08,$74,$21,$C4,$AF,$06,$9C,$0D,$11,$D3,$4D,$38,$23,$82,$1D,$08
	.dc.b	$71,$2B,$C1,$A7,$03,$44,$74,$D3,$4E,$08,$E0,$87,$42,$1C,$4A,$F0
	.dc.b	$69,$C0,$D1,$1D,$34,$D3,$82,$38,$21,$D0,$87,$12,$BC,$1A,$70,$34
	.dc.b	$47,$4D,$34,$E0,$8E,$08,$74,$21,$C4,$AF,$06,$9C,$0D,$11,$D3,$4D
	.dc.b	$38,$23,$82,$1D,$08,$71,$2B,$C1,$A7,$03,$44,$74,$D3,$4E,$08,$E0
	.dc.b	$87,$42,$1C,$4A,$F0,$69,$C0,$D1,$1D,$34,$D3,$82,$38,$21,$D0,$87
	.dc.b	$12,$BC,$1A,$70,$34,$47,$4D,$34,$E0,$8E,$08,$74,$21,$C4,$AF,$06
	.dc.b	$9C,$0D,$11,$D3,$4D,$38,$23,$82,$1D,$08,$71,$2B,$C1,$A7,$03,$44
	.dc.b	$74,$D3,$4E,$08,$E0,$87,$42,$1C,$4A,$F0,$69,$C0,$D1,$1D,$34,$D3
	.dc.b	$82,$38,$21,$D0,$87,$12,$BC,$1A,$70,$34,$47,$4D,$34,$E0,$8E,$08
	.dc.b	$74,$21,$C4,$AF,$06,$9C,$0D,$11,$D3,$4D,$38,$23,$82,$1D,$08,$71
	.dc.b	$2B,$C1,$A7,$03,$44,$74,$D3,$4E,$08,$E0,$87,$42,$1C,$4A,$F0,$69
	.dc.b	$C0,$D1,$1D,$34,$D3,$82,$38,$21,$D0,$87,$12,$BC,$1A,$70,$34,$47
	.dc.b	$4D,$34,$E0,$8E,$08,$74,$21,$C4,$AF,$06,$9C,$0D,$11,$D3,$4D,$38
	.dc.b	$23,$82,$1D,$08,$71,$2B,$C1,$A7,$03,$44,$74,$D3,$4E,$08,$E0,$87
	.dc.b	$42,$1C,$4A,$F0,$69,$C0,$D1,$1D,$34,$D3,$82,$38,$21,$D0,$87,$12
	.dc.b	$BC,$1A,$70,$34,$47,$4D,$34,$E0,$8E,$08,$74,$21,$C4,$AF,$06,$9C
	.dc.b	$0D,$11,$D3,$FD,$93,$D6,$30,$35,$A6,$5C,$11,$C1,$0E,$6E,$5A,$64
	.dc.b	$7C,$3F,$52,$3E,$16,$BE,$46,$BC,$0D,$10,$D1,$4F,$41,$08,$42,$10
	.dc.b	$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84
	.dc.b	$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21
	.dc.b	$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08
	.dc.b	$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42
	.dc.b	$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10
	.dc.b	$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84
	.dc.b	$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21
	.dc.b	$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08
	.dc.b	$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42
	.dc.b	$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10,$84,$21,$08,$42,$10
	.dc.b	$84,$21,$08,$42,$10,$84,$3F,$A6,$FD,$CF,$F8,$A4,$8F,$F8,$B5,$AF
	.dc.b	$F8,$C1,$A9,$A4,$7F,$C6,$9F,$75,$FE,$33,$23,$82,$1F,$E3,$A7,$2D
	.dc.b	$32,$1C,$B4,$FF,$20,$80

;--------------------------------------------------------------------------------
;	FTWOTOX.X FPm,FPn
;--------------------------------------------------------------------------------
	.cpu	68030
regs		reg	d0-d7/a0-a5
cregs		reg	fpcr/fpsr/fpiar
fregs		reg	fp0-fp7
	.offsym	0,_a6
_size:
_regs:	.ds.b	.sizeof.(regs)
_fregs:	.ds.b	.sizeof.(fregs)
_cregs:	.ds.b	.sizeof.(cregs)
_a6:	.ds.l	1
_pc:	.ds.l	1
	.text
	.even
ftwotox_test::
	link.w	a6,#_size
	movem.l	regs,(_regs,a6)
	fmovem.l	cregs,(_cregs,a6)
	fmovem.x	fregs,(_fregs,a6)
;
	move.l	#43,-(sp)
	peamsg	'FTWOTOX.X FPM,FPN'
	jbsr	mnemonic_start
	addq.l	#8,sp
	beq	99f
	putmsg	'test: FTWOTOX.X FPm,FPn',13,10
;------------------------------------------------
;	d1	fpcr=(XRN..DRP)<<4
;	d3	0=failed,1=successful
;	d5	expected status
;	d7	actual status
;	a2	source handle,...
;	a4	expected result handle,expected status,...
;	fp2	source
;	fp5	expected result
;	fp7	actual result
;------------------------------------------------
	lea.l	push_decompressed,a0
;decompress data
	move.l	a0,-(sp)
	pea.l	ftwotox_data_compressed
	jbsr	decompress
	addq.l	#8,sp
;relocate decompressed handle
	move.l	#indirect_decompressed,d0
@@:
	add.l	d0,(a0)+		;source handle
	tst.l	(a0)
	bpl	@b
	addq.l	#4,a0			;-1
@@:
	add.l	d0,(a0)+		;expected result handle
	addq.l	#4,a0			;expected status
	tst.l	(a0)
	bpl	@b
;	addq.l	#4,a0			;-1
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPn,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.x	fp2,fp7			;source
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	ftwotox.x	fp7,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTWOTOX.X FPn='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	lea.l	push_decompressed,a2	;source handle,...
	lea.l	(4*1649+4,a2),a4	;expected result handle,expected status,...
22:
	move.l	#0,d1		;fpcr=(XRN..DRP)<<4
11:
;FPm,FPn
	fmove.l	#0,fpcr
	fmove.x	([a2]),fp2		;source
	fmove.s	#$7FFFFFFF,fp7		;fill=NaN
	fmove.l	d1,fpcr			;fpcr
	fmove.l	#0,fpsr
					;source
	ftwotox.x	fp2,fp7		;EXECUTE
					;actual result
	fmove.l	fpsr,d7			;actual status
	fmove.l	#0,fpcr
;
	fmove.x	([a4]),fp5		;expected result
	move.l	(4,a4),d5		;expected status
;
	move.l	d1,-(sp)	;fpcr(rp<<6,-1=strict)
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	test_extended
	lea.l	36(sp),sp
	move.l	d0,d3			;0=failed,1=successful
;
	move.l	d3,-(sp)		;0=failed,1=successful
	jbsr	statistics_update
	addq.l	#4,sp
	beq	@f			;not output
;output
	putmsg	'FTWOTOX.X FPm='
	puthex24	([a2]),(4,[a2]),(8,[a2])	;source
	putmsg	',FPn @'
	move.l	d1,-(sp)
	jbsr	printfpcrrprm
	addq.l	#4,sp
	putcrlf
	move.l	d3,-(sp)		;0=failed,1=successful
	move.l	d5,-(sp)		;expected status
	fmove.x	fp5,-(sp)		;expected result
	move.l	d7,-(sp)		;actual status
	fmove.x	fp7,-(sp)		;actual result
	jbsr	output_extended
	lea.l	36(sp),sp
@@:
;
	addq.l	#8,a4			;expected result handle,expected status,...
;
	add.w	#1<<4,d1		;rprm++
	cmp.w	#176,d1		;fpcr=(XRN..DRP)<<4
	bls	11b
;
	addq.l	#4,a2			;source handle,...
	tst.l	(a2)			;source handle,...
	bpl	22b
;
	jbsr	mnemonic_end
99:
	fmovem.x	(_fregs,a6),fregs
	fmovem.l	(_cregs,a6),cregs
	movem.l	(_regs,a6),regs
	unlk	a6
	rts
	.cpu	68000

	.align	4
ftwotox_data_compressed::
	.dc.b	$00,$02,$84,$2C,$10,$86,$36,$BD,$58,$F1,$B3,$E9,$6B,$DE,$DF,$C5
	.dc.b	$70,$F9,$6E,$3F,$45,$CB,$EB,$B9,$FD,$D7,$4F,$C1,$BA,$FE,$25,$DB
	.dc.b	$F1,$EE,$FF,$95,$78,$3F,$36,$F0,$FE,$85,$E2,$FD,$3B,$C7,$FA,$D7
	.dc.b	$93,$F6,$6F,$2F,$EE,$5E,$6F,$DF,$BD,$1D,$0E,$FB,$E5,$63,$D5,$CF
	.dc.b	$C0,$6B,$E2,$F7,$F2,$DC,$3E,$87,$8F,$D6,$72,$FB,$7E,$7F,$81,$74
	.dc.b	$FC,$37,$AF,$E3,$1D,$BF,$23,$EF,$F9,$77,$83,$F3,$9F,$0F,$E8,$9E
	.dc.b	$2F,$D3,$FC,$7F,$AD,$79,$3F,$65,$F2,$FE,$E1,$E6,$FD,$EF,$E7,$79
	.dc.b	$CF,$42,$FB,$58,$F8,$4C,$FC,$76,$BE,$63,$7F,$45,$C3,$EB,$38,$FD
	.dc.b	$B7,$2F,$C0,$39,$FE,$15,$D3,$F1,$4E,$BF,$8F,$76,$FC,$A3,$BF,$E6
	.dc.b	$5E,$0F,$CF,$3C,$3F,$A3,$78,$BF,$50,$F1,$FE,$B5,$E4,$FD,$93,$CB
	.dc.b	$FB,$77,$9B,$F7,$8F,$C2,$7C,$27,$E1,$07,$C1,$E3,$E0,$F3,$F0,$7A
	.dc.b	$F8,$3D,$FC,$1F,$0F,$83,$E3,$F0,$7C,$BE,$0F,$9F,$C1,$F4,$F8,$3E
	.dc.b	$BF,$07,$DB,$E0,$FB,$FC,$1F,$83,$E0,$FC,$3F,$07,$E2,$F8,$3F,$1F
	.dc.b	$C1,$F9,$3E,$0F,$CB,$F0,$7E,$6F,$83,$F3,$FC,$1F,$E6,$FC,$CF,$C2
	.dc.b	$7C,$27,$E7,$97,$C1,$67,$E0,$B5,$F0,$5B,$F8,$2E,$1F,$05,$C7,$E0
	.dc.b	$B9,$7C,$17,$3F,$82,$E9,$F0,$5D,$7E,$0B,$B7,$C1,$77,$F8,$2F,$07
	.dc.b	$C1,$78,$7E,$0B,$C5,$F0,$5E,$3F,$82,$F2,$7C,$17,$97,$E0,$BC,$DF
	.dc.b	$05,$E7,$F8,$2F,$D9,$7C,$2F,$E1,$7A,$3F,$6A,$BE,$0B,$3F,$05,$AF
	.dc.b	$82,$DF,$C1,$70,$F8,$2E,$3F,$05,$CB,$E0,$B9,$FC,$17,$4F,$82,$EB
	.dc.b	$F0,$5D,$BE,$0B,$BF,$C1,$78,$3E,$0B,$C3,$F0,$5E,$2F,$82,$F1,$FC
	.dc.b	$17,$93,$E0,$BC,$BF,$05,$E6,$F8,$2F,$3F,$C1,$7F,$09,$F8,$6F,$E1
	.dc.b	$27,$E1,$BA,$3F,$86,$AF,$81,$D7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81
	.dc.b	$E5,$F0,$3C,$FE,$07,$A7,$C0,$F5,$F8,$1E,$DF,$03,$DF,$E0,$7C,$1F
	.dc.b	$03,$E1,$F8,$1F,$17,$C0,$F8,$FE,$07,$C9,$F0,$3E,$5F,$81,$F3,$7C
	.dc.b	$0F,$9F,$E0,$7D,$1F,$03,$FC,$87,$E1,$FF,$0F,$D0,$7F,$0E,$FF,$26
	.dc.b	$5F,$01,$BF,$80,$E1,$F0,$1C,$7E,$03,$97,$C0,$73,$F8,$0E,$9F,$01
	.dc.b	$D7,$E0,$3B,$7C,$07,$7F,$80,$F0,$7C,$07,$87,$E0,$3C,$5F,$01,$E3
	.dc.b	$F8,$0F,$27,$C0,$79,$7E,$03,$CD,$F0,$1E,$7F,$80,$F4,$7C,$07,$F3
	.dc.b	$9F,$86,$FC,$47,$3F,$F2,$3F,$80,$C7,$C0,$67,$E0,$35,$F0,$1B,$F8
	.dc.b	$0E,$1F,$01,$C7,$E0,$39,$7C,$07,$3F,$80,$E9,$F0,$1D,$7E,$03,$B7
	.dc.b	$C0,$77,$F8,$0F,$07,$C0,$78,$7E,$03,$C5,$F0,$1E,$3F,$80,$F2,$7C
	.dc.b	$07,$97,$E0,$3C,$DF,$01,$FD,$3F,$E2,$3F,$A7,$9F,$87,$E7,$3F,$87
	.dc.b	$7F,$AA,$AF,$7F,$87,$BF,$C7,$DF,$E5,$EF,$F3,$F7,$FA,$7B,$FD,$7D
	.dc.b	$FE,$DE,$FF,$7F,$7F,$C1,$EF,$F8,$7D,$FF,$17,$BF,$E3,$F7,$FC,$9E
	.dc.b	$FF,$97,$DF,$F3,$7B,$FE,$7F,$7F,$D1,$EF,$FA,$7D,$FF,$ED,$1F,$13
	.dc.b	$F8,$83,$F0,$FC,$E7,$F0,$EF,$F6,$E5,$EF,$70,$F7,$B8,$FB,$DC,$BD
	.dc.b	$EE,$7E,$F7,$4F,$7B,$AF,$BD,$DB,$DE,$EF,$EF,$78,$3D,$EF,$0F,$BD
	.dc.b	$E2,$F7,$BC,$7E,$F7,$93,$DE,$F2,$FB,$DE,$6F,$7B,$CF,$EF,$7A,$3D
	.dc.b	$EF,$F0,$07,$C5,$7E,$20,$FE,$1D,$CD,$FE,$04,$5E,$EE,$BD,$DD,$FB
	.dc.b	$BC,$3D,$DE,$3E,$EF,$2F,$77,$9F,$BB,$D3,$DD,$EB,$EE,$F6,$F7,$7B
	.dc.b	$FB,$BE,$0F,$77,$C3,$EE,$F8,$BD,$DF,$1F,$BB,$E4,$F7,$7C,$BE,$EF
	.dc.b	$9B,$DD,$F3,$FB,$BF,$E1,$7F,$87,$FF,$0B,$9F,$89,$3F,$11,$CF,$FD
	.dc.b	$5B,$DD,$D7,$BB,$BF,$77,$87,$BB,$C7,$DD,$E5,$EE,$F3,$F7,$7A,$7B
	.dc.b	$BD,$7D,$DE,$DE,$EF,$7F,$77,$C1,$EE,$F8,$7D,$DF,$17,$BB,$E3,$F7
	.dc.b	$7C,$9E,$EF,$97,$DD,$F3,$7B,$BE,$7F,$77,$FC,$5D,$F1,$7F,$C3,$8F
	.dc.b	$C3,$9F,$86,$E6,$3F,$86,$7F,$C6,$AB,$DC,$E3,$EE,$72,$F7,$39,$FB
	.dc.b	$9D,$3D,$CE,$BE,$E7,$6F,$73,$BF,$B9,$E0,$F7,$3C,$3E,$E7,$8B,$DC
	.dc.b	$F1,$FB,$9E,$4F,$73,$CB,$EE,$79,$BD,$CF,$3F,$B9,$E8,$F7,$3F,$C8
	.dc.b	$DF,$19,$FC,$38,$FC,$3F,$29,$FC,$3B,$FE,$4C,$5E,$DF,$0F,$6F,$8F
	.dc.b	$B7,$CB,$DB,$E7,$ED,$F4,$F6,$FA,$FB,$7D,$BD,$BE,$FE,$DF,$83,$DB
	.dc.b	$F0,$FB,$7E,$2F,$6F,$C7,$ED,$F9,$3D,$BF,$2F,$B7,$E6,$F6,$FC,$FE
	.dc.b	$DF,$A3,$DB,$FF,$2F,$7C,$6F,$F9,$78,$FE,$20,$7E,$20,$FE,$1D,$CD
	.dc.b	$FE,$64,$5E,$D7,$0F,$6B,$8F,$B5,$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB
	.dc.b	$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0,$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D
	.dc.b	$AF,$2F,$B5,$E6,$F6,$BC,$FE,$D7,$A3,$DA,$FF,$3B,$7C,$17,$F5,$63
	.dc.b	$F8,$79,$F8,$93,$F1,$1C,$DF,$E0,$7F,$6B,$7E,$D7,$0F,$6B,$8F,$B5
	.dc.b	$CB,$DA,$E7,$ED,$74,$F6,$BA,$FB,$5D,$BD,$AE,$FE,$D7,$83,$DA,$F0
	.dc.b	$FB,$5E,$2F,$6B,$C7,$ED,$79,$3D,$AF,$2F,$B5,$E6,$F6,$BF,$D1,$9F
	.dc.b	$1D,$FD,$50,$FC,$39,$F8,$6E,$43,$F8,$67,$FD,$26,$BD,$9E,$3E,$CF
	.dc.b	$2F,$67,$9F,$B3,$D3,$D9,$EB,$EC,$F6,$F6,$7B,$FB,$3E,$0F,$67,$C3
	.dc.b	$EC,$F8,$BD,$9F,$1F,$B3,$E4,$F6,$7C,$BE,$CF,$9B,$D9,$F3,$FB,$3E
	.dc.b	$8F,$67,$FD,$49,$F1,$FF,$EA,$43,$FA,$B1,$FC,$38,$FC,$3F,$29,$FC
	.dc.b	$3B,$FE,$AC,$5E,$C7,$2F,$63,$9F,$B1,$D3,$D8,$EB,$EC,$76,$F6,$3B
	.dc.b	$FB,$1E,$0F,$63,$C3,$EC,$78,$BD,$8F,$1F,$B1,$E4,$F6,$3C,$BE,$C7
	.dc.b	$9B,$D8,$F3,$FB,$1E,$8F,$63,$D3,$EC,$7F,$AF,$7E,$43,$FC,$3A,$7F
	.dc.b	$56,$3F,$88,$1F,$88,$3F,$87,$72,$FF,$B1,$17,$AF,$C7,$D7,$E5,$EB
	.dc.b	$F3,$F5,$FA,$7A,$FD,$7D,$7E,$DE,$BF,$7F,$5F,$C1,$EB,$F8,$7D,$7F
	.dc.b	$17,$AF,$E3,$F5,$FC,$9E,$BF,$97,$D7,$F3,$7A,$FE,$7F,$5F,$FD,$A9
	.dc.b	$F1,$9F,$E1,$C3,$F8,$79,$F8,$93,$F1,$1C,$9F,$E7,$CF,$5F,$7E,$BF
	.dc.b	$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA,$FD,$BD,$7E,$FE
	.dc.b	$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D,$7F,$2F,$AF,$E6
	.dc.b	$F5,$FF,$DD,$5F,$23,$FE,$EA,$3F,$C3,$87,$F5,$43,$F0,$E7,$E1,$B9
	.dc.b	$0F,$E1,$9F,$F7,$9A,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$E8,$F5,$BD,$3E,$B3,$8E,$7C,$97,$F9,$90,$FF,$0E
	.dc.b	$9F,$D5,$CF,$E1,$E7,$E2,$13,$F8,$87,$8F,$AB,$CD,$CF,$AB,$65,$EA
	.dc.b	$DF,$57,$1E,$AE,$7D,$5D,$7A,$BB,$F5,$78,$7A,$BC,$7D,$5E,$5E,$AF
	.dc.b	$3F,$57,$A7,$AB,$D7,$D5,$ED,$EA,$F8,$7E,$4F,$FC,$CA,$7F,$58,$3F
	.dc.b	$89,$1F,$8A,$3F,$88,$71,$7C,$15,$EA,$71,$F5,$39,$7A,$9C,$FD,$4E
	.dc.b	$9E,$A7,$5F,$53,$B7,$A9,$DF,$D4,$F0,$7A,$9E,$1F,$53,$C5,$EA,$78
	.dc.b	$FD,$4F,$27,$A9,$E5,$F5,$3C,$DE,$A7,$9F,$D4,$F9,$1F,$1B,$F2,$79
	.dc.b	$3F,$D5,$5F,$0B,$88,$CF,$C2,$EB,$E0,$77,$EE,$F0,$F6,$38,$FB,$1C
	.dc.b	$BD,$8E,$7E,$C7,$4F,$63,$AF,$B1,$DB,$D8,$EF,$EC,$78,$3D,$8F,$0F
	.dc.b	$B1,$E2,$F6,$3C,$7E,$C7,$93,$D8,$F2,$FB,$1E,$6F,$63,$CF,$EC,$7D
	.dc.b	$FF,$94,$FF,$55,$9F,$E6,$43,$FC,$3C,$7F,$56,$3F,$10,$7E,$1F,$A0
	.dc.b	$BD,$9E,$5E,$CF,$3F,$63,$A7,$B1,$D7,$D8,$ED,$EC,$77,$F6,$3C,$1E
	.dc.b	$C7,$87,$D8,$F1,$7B,$1E,$3F,$63,$C9,$EC,$79,$7D,$8F,$37,$B1,$E7
	.dc.b	$F6,$3F,$2D,$F2,$BF,$EA,$A3,$FC,$3A,$7F,$57,$3F,$87,$9F,$86,$E3
	.dc.b	$3F,$86,$7F,$50,$BD,$7E,$7E,$BF,$4F,$5F,$AF,$AF,$DB,$D7,$EF,$EB
	.dc.b	$F8,$3D,$7F,$0F,$AF,$E2,$F5,$FC,$7E,$BF,$93,$D7,$F2,$FA,$FE,$6F
	.dc.b	$5F,$CF,$EB,$FA,$3D,$7F,$D7,$7C,$B7,$E1,$CF,$F5,$61,$FE,$64,$3F
	.dc.b	$AB,$9F,$C4,$4F,$E2,$1C,$67,$F8,$83,$FC,$05,$7A,$5D,$3D,$2E,$BE
	.dc.b	$97,$6F,$4B,$BF,$A5,$E0,$F4,$BC,$3E,$97,$8B,$D2,$F1,$FA,$5E,$4F
	.dc.b	$4B,$CB,$E9,$79,$BD,$2F,$3F,$A5,$E8,$F4,$BD,$3E,$97,$F0,$7F,$89
	.dc.b	$FF,$79,$9F,$EA,$C3,$FC,$CC,$7F,$87,$8F,$E2,$07,$E2,$4F,$C4,$71
	.dc.b	$FF,$19,$5E,$8F,$3F,$47,$A7,$A3,$D7,$D1,$ED,$E8,$F7,$F4,$7C,$1E
	.dc.b	$8F,$87,$D1,$F1,$7A,$3E,$3F,$47,$C9,$E8,$F9,$7D,$1F,$37,$A3,$E7
	.dc.b	$F4,$7F,$8F,$7C,$BF,$FB,$C8,$FF,$32,$9F,$E1,$F3,$FA,$B9,$F8,$93
	.dc.b	$F1,$1C,$47,$F1,$0F,$F3,$15,$E8,$74,$F4,$3A,$FA,$1D,$BD,$0E,$FE
	.dc.b	$87,$83,$D0,$F0,$FA,$1E,$2F,$43,$C7,$E8,$79,$3D,$0F,$2F,$A1,$E6
	.dc.b	$F4,$3C,$FE,$87,$A3,$D0,$FE,$6D,$F3,$1F,$C3,$CF,$F7,$A1,$FE,$AC
	.dc.b	$3F,$C3,$E7,$F5,$93,$F1,$27,$E2,$38,$8F,$E2,$1F,$E9,$4B,$CF,$EB
	.dc.b	$E7,$F6,$F3,$FB,$F9,$FE,$0F,$3F,$C3,$E7,$F8,$BC,$FF,$1F,$9F,$E4
	.dc.b	$F3,$FC,$BE,$7F,$9B,$CF,$F3,$F9,$FE,$8F,$3F,$D3,$E7,$FF,$4F,$F9
	.dc.b	$97,$C8,$DF,$83,$C7,$C9,$67,$E4,$35,$F1,$5B,$F8,$7E,$1E,$DF,$1F
	.dc.b	$5B,$97,$AD,$CF,$D6,$E9,$EB,$75,$F5,$BB,$7A,$DD,$FD,$6F,$07,$AD
	.dc.b	$E1,$F5,$BC,$5E,$B7,$8F,$D6,$F2,$7A,$DE,$5F,$5B,$CD,$EB,$7F,$65
	.dc.b	$E4,$7C,$B5,$F8,$CF,$ED,$1F,$8A,$7F,$87,$A7,$FC,$D0,$5F,$8A,$E1
	.dc.b	$F0,$FC,$7D,$FE,$5E,$DF,$3F,$6F,$A7,$B7,$D7,$DB,$ED,$ED,$F7,$F6
	.dc.b	$FC,$1E,$DF,$87,$DB,$F1,$7B,$7E,$3F,$6F,$C9,$ED,$F9,$7D,$BF,$1F
	.dc.b	$07,$9F,$DB,$F4,$7B,$7F,$DF,$7E,$6F,$FA,$B0,$F8,$8C,$7C,$46,$7E
	.dc.b	$23,$5F,$11,$BF,$88,$E1,$F1,$1C,$7E,$17,$97,$C0,$F3,$F8,$1E,$9F
	.dc.b	$03,$D7,$E0,$7B,$7C,$0F,$7F,$81,$F0,$7C,$0F,$87,$E0,$7C,$5F,$03
	.dc.b	$E3,$F8,$1F,$27,$C0,$F9,$7E,$07,$CD,$F0,$3E,$7F,$81,$FF,$0A,$7C
	.dc.b	$E7,$E1,$BE,$13,$FC,$40,$BE,$07,$3E,$66,$BE,$07,$7F,$03,$C3,$E0
	.dc.b	$78,$FC,$0F,$2F,$81,$E7,$F0,$3F,$63,$D2,$EB,$F0,$3D,$BE,$07,$BF
	.dc.b	$C0,$F8,$3E,$07,$C3,$F0,$3E,$2F,$81,$F1,$FC,$0F,$93,$E0,$7C,$BF
	.dc.b	$03,$E6,$F8,$1F,$3F,$C0,$FA,$3E,$07,$FC,$59,$F3,$BF,$88,$E9,$FF
	.dc.b	$1C,$7C,$07,$F0,$A5,$1F,$EF,$63,$FD,$60,$7F,$9A,$0F,$EB,$47,$F1
	.dc.b	$23,$81,$FC,$41,$FF,$21,$2F,$7B,$BF,$BD,$E0,$F7,$BC,$3E,$F7,$8B
	.dc.b	$DE,$F1,$FB,$DE,$4F,$7B,$CB,$EF,$79,$BD,$EF,$3F,$BD,$E8,$F7,$BF
	.dc.b	$3B,$EF,$FE,$7B,$5F,$CF,$EB,$FA,$0D,$7F,$43,$AF,$BB,$DE,$F7,$FB
	.dc.b	$DF,$E4,$BF,$6F,$FC,$48,$7E,$3F,$8B,$FC,$B0,$BD,$CD,$3B,$F7,$38
	.dc.b	$7B,$9C,$7D,$CE,$5E,$E7,$3F,$73,$A7,$B9,$FB,$0E,$8E,$DE,$E7,$7F
	.dc.b	$73,$C1,$EE,$7E,$D3,$AF,$F6,$BD,$7F,$B6,$EB,$FD,$BF,$5F,$EE,$3A
	.dc.b	$FF,$73,$D7,$FB,$AE,$6F,$DD,$F3,$7F,$98,$FF,$15,$FF,$1E,$1E,$9E
	.dc.b	$73,$E9,$7F,$CE,$4F,$FB,$38,$FF,$44,$1F,$DE,$0F,$E6,$07,$F2,$F7
	.dc.b	$8F,$0F,$5E,$1E,$DC,$1D,$F8,$3E,$47,$BB,$F2,$7D,$DF,$95,$EE,$FC
	.dc.b	$BF,$77,$E6,$7B,$BF,$37,$DD,$F9,$DE,$EF,$CF,$F7,$7E,$87,$BB,$FE
	.dc.b	$7B,$F9,$FF,$D4,$70,$1F,$D4,$3F,$E8,$95,$EE,$6F,$DC,$E1,$EE,$71
	.dc.b	$F7,$39,$7B,$9C,$FD,$CE,$9E,$E7,$5F,$6F,$B7,$B7,$DF,$DB,$F0,$7B
	.dc.b	$7E,$1F,$6F,$C5,$ED,$F8,$FD,$BF,$27,$B7,$F3,$7C,$9F,$9D,$D9,$F3
	.dc.b	$FB,$3E,$87,$67,$D1,$EC,$FA,$5D,$9F,$4F,$B3,$EA,$76,$7D,$5E,$CF
	.dc.b	$AD,$D9,$FE,$99,$F4,$FD,$E7,$BB,$B4,$FB,$9F,$F5,$42,$EB,$E1,$D7
	.dc.b	$C7,$AF,$97,$5F,$3E,$BE,$9D,$7D,$7A,$FE,$67,$B7,$F3,$7D,$BF,$9D
	.dc.b	$ED,$F8,$7B,$BE,$87,$B7,$FE,$AB,$FA,$0F,$E0,$1C,$05,$F0,$99,$F8
	.dc.b	$AD,$7C,$AE,$FE,$17,$87,$C0,$F1,$F8,$6E,$5F,$01,$CF,$E0,$3A,$74
	.dc.b	$75,$F8,$0F,$81,$E4,$FC,$1F,$27,$E1,$79,$3F,$0F,$C9,$F8,$9E,$4F
	.dc.b	$8F,$E0,$3C,$9E,$D7,$97,$DA,$F3,$7B,$5E,$7F,$6B,$D1,$ED,$7A,$7D
	.dc.b	$AF,$57,$B5,$EB,$F6,$BD,$9E,$D7,$FB,$1F,$E8,$7F,$D4,$07,$E9,$4F
	.dc.b	$D2,$76,$97,$C0,$6F,$E0,$38,$7C,$07,$1F,$80,$E5,$F0,$1F,$67,$D9
	.dc.b	$E9,$EF,$F5,$F7,$FB,$7B,$FD,$FD,$FF,$07,$BB,$E1,$F7,$7C,$5E,$EF
	.dc.b	$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE,$F9,$FD,$DF,$47,$BB,$F8,$7F
	.dc.b	$57,$D5,$EE,$FF,$B8,$7C,$FF,$D3,$9F,$89,$3F,$11,$D2,$7F,$10,$FF
	.dc.b	$BC,$5F,$F7,$61,$FE,$A5,$3F,$CB,$AE,$FD,$9E,$9E,$CF,$5F,$67,$B7
	.dc.b	$AF,$DF,$CC,$F0,$79,$9E,$1F,$33,$C5,$E6,$78,$FC,$CF,$27,$67,$97
	.dc.b	$7F,$37,$B3,$E7,$F6,$7D,$1E,$CF,$FB,$E7,$E8,$BF,$1E,$7E,$3B,$84
	.dc.b	$17,$B1,$BE,$3E,$1E,$C7,$1F,$63,$97,$B1,$CF,$D8,$E9,$EC,$75,$F6
	.dc.b	$3B,$7B,$1F,$AF,$DF,$C1,$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E
	.dc.b	$C7,$97,$D8,$F3,$7B,$1E,$7F,$63,$D1,$EC,$7A,$7D,$8F,$3F,$3B,$E8
	.dc.b	$FF,$DD,$87,$F3,$13,$F1,$A7,$E3,$1F,$6A,$F5,$F8,$BF,$EA,$83,$FC
	.dc.b	$C1,$C2,$EB,$CF,$AB,$AF,$57,$7E,$AF,$0F,$57,$8F,$AB,$CB,$D5,$E7
	.dc.b	$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E,$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F
	.dc.b	$57,$C7,$EA,$FB,$3E,$93,$F9,$99,$FC,$54,$FC,$69,$F8,$CE,$62,$F7
	.dc.b	$F8,$7B,$FC,$7D,$FE,$5E,$CF,$3F,$67,$A7,$B3,$D7,$D9,$ED,$EC,$F7
	.dc.b	$F6,$7C,$1E,$CF,$87,$D9,$F1,$7B,$3E,$3F,$67,$C9,$EC,$F9,$7C,$FF
	.dc.b	$37,$85,$E7,$F6,$7E,$9F,$C0,$7F,$1A,$3F,$19,$CB,$FC,$A7,$D9,$CF
	.dc.b	$B3,$AF,$67,$7E,$CF,$0F,$67,$8F,$B3,$CB,$D9,$E7,$E6,$74,$F6,$7A
	.dc.b	$FB,$3D,$BD,$9E,$FE,$CF,$83,$D9,$F0,$FB,$3E,$2F,$67,$C7,$EC,$F9
	.dc.b	$3D,$9F,$2F,$B3,$E6,$F6,$7F,$13,$F4,$BF,$8A,$E6,$3F,$DE,$4F,$E2
	.dc.b	$A7,$E2,$5C,$FB,$5C,$3D,$DE,$3E,$EF,$2F,$6F,$9F,$B7,$D3,$DB,$EB
	.dc.b	$ED,$F6,$F6,$FB,$FB,$7E,$0F,$6F,$C3,$ED,$F8,$BD,$BF,$1F,$B7,$E4
	.dc.b	$F6,$FC,$BE,$DF,$9B,$DB,$F3,$FB,$7E,$8F,$6F,$F4,$3F,$4D,$FD,$E0
	.dc.b	$FE,$B4,$7F,$12,$3F,$0C,$7E,$17,$97,$F7,$AB,$DA,$E3,$ED,$72,$F6
	.dc.b	$B9,$FB,$5D,$3D,$AE,$BE,$D7,$6F,$6B,$BF,$B5,$E0,$F6,$BC,$3E,$D7
	.dc.b	$8B,$DA,$F1,$FB,$5E,$4F,$6B,$CB,$ED,$79,$BD,$AF,$3F,$B5,$FB,$8F
	.dc.b	$A7,$FE,$BE,$7E,$18,$FE,$1F,$C9,$FC,$49,$FE,$54,$DF,$63,$87,$B1
	.dc.b	$C7,$D8,$E5,$EC,$73,$F6,$3A,$7B,$1D,$7D,$8E,$DE,$C7,$7F,$63,$C1
	.dc.b	$EC,$78,$7D,$8F,$17,$B1,$E3,$F6,$3C,$9E,$C7,$97,$D8,$F3,$7B,$1E
	.dc.b	$7F,$63,$F8,$7F,$AF,$F8,$73,$FC,$7C,$7F,$5A,$3F,$0C,$7F,$0D,$E4
	.dc.b	$FE,$56,$BD,$7E,$3E,$BF,$2F,$5F,$9F,$AF,$D3,$D7,$EB,$EB,$F6,$F5
	.dc.b	$FB,$FA,$FE,$0F,$5F,$C3,$EB,$F8,$BD,$7F,$1F,$AF,$E4,$F5,$FC,$BE
	.dc.b	$BF,$9B,$D7,$F3,$FA,$FE,$8F,$5F,$F9,$3F,$D4,$7F,$8F,$8F,$F1,$01
	.dc.b	$FD,$60,$FE,$18,$7E,$1F,$94,$FE,$1D,$FE,$8A,$FF,$58,$1E,$AF,$4F
	.dc.b	$57,$AF,$AB,$DB,$D5,$EF,$EA,$F8,$3D,$5F,$0F,$AB,$E2,$F5,$7C,$7E
	.dc.b	$AF,$93,$D5,$F2,$FA,$BE,$6F,$57,$CF,$EA,$FA,$3D,$5F,$E7,$FF,$53
	.dc.b	$FE,$26,$3F,$86,$9F,$88,$E2,$3F,$88,$6F,$AB,$C3,$D5,$E3,$EA,$F2
	.dc.b	$F5,$79,$FA,$BD,$3D,$5E,$BE,$AF,$6F,$57,$BF,$AB,$E0,$F5,$7C,$3E
	.dc.b	$AF,$8B,$D5,$F1,$FA,$BE,$4F,$57,$CB,$EA,$F9,$BD,$5F,$3F,$AB,$FD
	.dc.b	$5B,$EA,$BF,$87,$1F,$E8,$53,$FC,$42,$7F,$0D,$3F,$11,$C8,$7F,$10
	.dc.b	$FF,$74,$5E,$A7,$3F,$53,$A7,$A9,$D7,$D4,$ED,$EA,$77,$F5,$3C,$1E
	.dc.b	$A7,$87,$D4,$F1,$7A,$9E,$3F,$53,$C9,$EA,$79,$7D,$4F,$37,$A9,$E7
	.dc.b	$F5,$3D,$1E,$A7,$A7,$D4,$FE,$DF,$E2,$7F,$A1,$8F,$F3,$41,$FE,$20
	.dc.b	$3F,$AA,$1F,$89,$3F,$11,$C8,$7F,$C3,$E7,$E2,$1F,$F0,$7A,$F4,$BA
	.dc.b	$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA,$5E,$2F,$4B,$C7,$E9,$79
	.dc.b	$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$A3,$D2,$FF,$03,$FD,$5F,$F9
	.dc.b	$B8,$FE,$AA,$7E,$24,$FC,$41,$E7,$E1,$FF,$11,$AF,$4B,$8F,$A5,$CB
	.dc.b	$D2,$E7,$E9,$74,$F4,$BA,$FA,$5D,$BD,$2E,$FE,$97,$83,$D2,$F0,$FA
	.dc.b	$5E,$2F,$4B,$C7,$E9,$79,$3D,$2F,$2F,$A5,$E6,$F4,$BC,$FE,$97,$F8
	.dc.b	$7B,$EB,$3F,$AB,$1F,$EC,$D3,$FC,$D2,$7F,$55,$3F,$12,$7E,$23,$90
	.dc.b	$FE,$21,$FF,$1E,$BF,$EC,$D1,$E8,$75,$F4,$3B,$7A,$1D,$FD,$0F,$07
	.dc.b	$A1,$E1,$F4,$3C,$5E,$87,$8F,$D0,$F2,$7A,$1E,$5F,$43,$CD,$E8,$79
	.dc.b	$FD,$0F,$47,$A1,$E9,$F4,$3F,$C6,$FF,$5B,$FE,$CE,$3F,$D6,$07,$F9
	.dc.b	$A0,$FF,$0E,$1F,$C4,$4F,$C4,$9F,$E6,$63,$F1,$1C,$47,$F1,$0F,$F9
	.dc.b	$61,$79,$FD,$BC,$FE,$FE,$7F,$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7
	.dc.b	$F9,$3C,$FF,$2F,$9F,$E6,$F3,$FC,$FE,$7F,$A3,$CF,$FF,$27,$FC,$17
	.dc.b	$FA,$D8,$FF,$0E,$9F,$C4,$4F,$E2,$07,$E2,$78,$3F,$C6,$9E,$7F,$0F
	.dc.b	$3F,$8F,$9F,$CB,$CF,$E7,$E7,$F4,$F3,$FA,$F9,$FD,$BC,$FE,$FE,$7F
	.dc.b	$83,$CF,$F0,$F9,$FE,$2F,$3F,$C7,$E7,$F9,$3C,$FF,$2F,$9F,$E6,$F3
	.dc.b	$FF,$CC,$BF,$5D,$FE,$1D,$1F,$27,$8F,$8F,$CF,$C4,$EB,$DE,$DF,$BD
	.dc.b	$C3,$DA,$E3,$ED,$72,$F5,$B9,$FA,$DD,$3D,$6E,$BE,$B7,$6F,$5B,$BF
	.dc.b	$AD,$E0,$F5,$BC,$3E,$B7,$8B,$D6,$F1,$FA,$DE,$4F,$5B,$CB,$EB,$79
	.dc.b	$BD,$6F,$3F,$AD,$FE,$7C,$FA,$F7,$CB,$DF,$94,$C7,$C8,$67,$E2,$B5
	.dc.b	$E7,$6F,$E1,$B8,$7C,$37,$1F,$84,$E5,$F0,$3C,$FD,$EE,$9E,$F7,$5F
	.dc.b	$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE,$1F,$7B,$C5,$EF,$78,$FD,$EF,$27
	.dc.b	$BD,$E5,$F7,$BC,$DE,$F7,$FA,$4B,$EC,$3F,$DF,$27,$F9,$80,$FE,$A8
	.dc.b	$7F,$53,$3F,$08,$7F,$0B,$F7,$FF,$D5,$EB,$DE,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$FD,$53,$EC,$FF,$98
	.dc.b	$4F,$C2,$F1,$FF,$B0,$9F,$F7,$A9,$FE,$61,$3F,$AA,$9F,$C3,$4F,$E1
	.dc.b	$8F,$F8,$77,$D8,$E7,$EC,$74,$F6,$3A,$FB,$1D,$BD,$8E,$FE,$C7,$83
	.dc.b	$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79,$3D,$8F,$2F,$B1,$E6,$F6,$3C
	.dc.b	$FE,$C7,$A3,$D8,$FF,$60,$FD,$8F,$E1,$B8,$BF,$DB,$2B,$D7,$CF,$AF
	.dc.b	$AF,$5F,$7E,$BF,$0F,$5F,$8F,$AF,$CB,$D7,$E7,$EB,$F4,$F5,$FA,$FA
	.dc.b	$FD,$BD,$7E,$FE,$BF,$83,$D7,$F0,$FA,$FE,$2F,$5F,$C7,$EB,$F9,$3D
	.dc.b	$7F,$2F,$AF,$E6,$F5,$FC,$FE,$BF,$FB,$63,$EC,$9F,$35,$7E,$4F,$1F
	.dc.b	$1F,$9F,$8F,$D7,$C4,$EF,$E1,$B8,$7B,$BC,$7D,$DE,$5E,$EF,$3F,$77
	.dc.b	$A7,$BB,$D7,$DD,$ED,$EE,$F7,$F7,$7C,$1E,$EF,$87,$DD,$F1,$7B,$BE
	.dc.b	$3F,$77,$C9,$EE,$F9,$7D,$DF,$37,$BB,$FE,$EE,$FB,$3F,$F5,$31,$F8
	.dc.b	$6F,$38,$BE,$17,$5F,$0B,$BF,$85,$E1,$F0,$3C,$7E,$17,$97,$C1,$73
	.dc.b	$F8,$2E,$9F,$05,$D7,$E0,$BB,$7C,$17,$7F,$82,$F0,$7C,$17,$87,$E0
	.dc.b	$BC,$5F,$05,$E3,$F8,$2F,$27,$C1,$79,$7E,$0B,$CD,$F0,$5E,$7F,$82
	.dc.b	$F4,$7C,$13,$9E,$78,$FF,$87,$7F,$BE,$7C,$0E,$5F,$EF,$DF,$03,$67
	.dc.b	$FD,$DC,$7F,$98,$CF,$F0,$D8,$F7,$B5,$EF,$6F,$DE,$E1,$EF,$71,$F7
	.dc.b	$B9,$7B,$DC,$FD,$EE,$9E,$F7,$5F,$7B,$B7,$BD,$DF,$DE,$F0,$7B,$DE
	.dc.b	$1F,$7B,$C5,$EF,$78,$FD,$EF,$37,$DA,$7E,$23,$93,$D8,$FF,$BB,$8F
	.dc.b	$F3,$11,$FE,$61,$3F,$A9,$1F,$C3,$D3,$FC,$3D,$E3,$EC,$74,$F6,$3A
	.dc.b	$FB,$1D,$BD,$8E,$FE,$C7,$83,$D8,$F0,$FB,$1E,$2F,$63,$C7,$EC,$79
	.dc.b	$3D,$8F,$2F,$B1,$E6,$F6,$3C,$FE,$C7,$A3,$D8,$F9,$DF,$6B,$FE,$EF
	.dc.b	$3F,$14,$7E,$27,$87,$EB,$3F,$EE,$F3,$FC,$C6,$7F,$86,$CF,$E1,$E3
	.dc.b	$D3,$E7,$E9,$F4,$F4,$FA,$FA,$7D,$BD,$3E,$FE,$9F,$83,$D3,$F0,$FA
	.dc.b	$7E,$2F,$4F,$C7,$E9,$F9,$3D,$3F,$2F,$A7,$E6,$F4,$FC,$FE,$9F,$A3
	.dc.b	$D3,$FC,$37,$DB,$7E,$24,$FE,$1D,$C7,$F8,$D5,$E9,$EB,$E1,$37,$F0
	.dc.b	$9C,$3E,$13,$8F,$AB,$CB,$D5,$E7,$EA,$F4,$F5,$7A,$FA,$BD,$BD,$5E
	.dc.b	$FE,$AF,$83,$D5,$F0,$FA,$BE,$2F,$57,$C7,$EA,$F9,$3D,$5F,$2F,$AB
	.dc.b	$E6,$F5,$7C,$FE,$AF,$E7,$78,$7F,$11,$CD,$FF,$00,$7E,$0B,$1F,$05
	.dc.b	$9F,$82,$D7,$C1,$6F,$DA,$E1,$F0,$5C,$7D,$AE,$5E,$D7,$3F,$6B,$A7
	.dc.b	$B5,$D7,$DA,$ED,$ED,$77,$F6,$BC,$1E,$D7,$87,$DA,$F1,$7B,$5E,$3F
	.dc.b	$6B,$C9,$ED,$79,$7D,$AF,$37,$B5,$FB,$3F,$B7,$FD,$A7,$39,$FE,$20
	.dc.b	$7E,$1D,$C7,$C0,$EF,$E0,$78,$7C,$0F,$1F,$81,$E5,$EF,$73,$F7,$BA
	.dc.b	$7B,$DD,$7D,$EE,$DE,$F7,$7F,$7B,$C1,$EF,$78,$7D,$EF,$17,$BD,$E3
	.dc.b	$F7,$BC,$9E,$F7,$97,$DE,$F3,$7B,$DE,$7F,$7B,$D1,$EF,$7F,$0B,$FB
	.dc.b	$8F,$E1,$C7,$E1,$B9,$8F,$E1,$9F,$E2,$8B,$DD,$E1,$EE,$F1,$F3,$B9
	.dc.b	$7B,$BC,$FD,$DE,$9E,$EF,$5F,$77,$B7,$BB,$DF,$DD,$F0,$7B,$BE,$1F
	.dc.b	$77,$C5,$EE,$F8,$FD,$DF,$27,$BB,$E5,$F7,$7C,$DE,$EF,$9F,$DD,$F4
	.dc.b	$7B,$BF,$C8,$FE,$E7,$F0,$E3,$DE,$C7,$BD,$9F,$7B,$5E,$F6,$FD,$EE
	.dc.b	$1E,$F7,$1F,$7B,$97,$BD,$CF,$DE,$E9,$EF,$75,$F7,$BB,$7B,$DD,$FD
	.dc.b	$EF,$07,$BD,$E1,$F7,$BC,$5E,$F7,$8F,$DE,$F2,$7B,$DE,$5F,$7B,$CD
	.dc.b	$EF,$7F,$39,$F6,$BF,$0E,$7F,$55,$3F,$0C,$3D,$DD,$7B,$BB,$F7,$78
	.dc.b	$4F,$1F,$77,$97,$BB,$CF,$DD,$E9,$EE,$F5,$F7,$7B,$7B,$BD,$FD,$DF
	.dc.b	$07,$BB,$E1,$F7,$7C,$5E,$EF,$8F,$DD,$F2,$7B,$BE,$5F,$77,$CD,$EE
	.dc.b	$F9,$FD,$DF,$EA,$1F,$75,$FD,$4C,$FC,$31,$F8,$5F,$78,$FE,$15,$FE
	.dc.b	$B4,$BD,$CE,$3E,$E7,$2F,$73,$9F,$B9,$D3,$DC,$EB,$EE,$76,$F7,$3B
	.dc.b	$FB,$9E,$0F,$73,$C3,$EE,$78,$BD,$CF,$1F,$B9,$E4,$F7,$3C,$BE,$E7
	.dc.b	$9B,$DC,$F3,$FB,$9E,$8F,$73,$FB,$47,$DD,$FF,$0C,$3F,$0D,$CB,$FD
	.dc.b	$BD,$7B,$7A,$F6,$F7,$ED,$F0,$F6,$F8,$FB,$7C,$BD,$BE,$7E,$DF,$4F
	.dc.b	$6D,$FC,$03,$0B,$A7,$F7,$BF,$9F,$F8,$5F,$15,$FC,$1A,$31,$DF,$FC
	.dc.b	$79,$FE,$F9,$3A,$74,$D7,$F8,$6B,$5D,$C7,$17,$9C,$F5,$53,$B7,$27
	.dc.b	$49,$E0,$E9,$B1,$B8,$C4,$FF,$84,$FE,$D5,$B7,$82,$30,$D9,$4E,$9F
	.dc.b	$81,$E0,$7F,$85,$E3,$11,$D0,$DC,$F1,$8C,$DF,$C1,$77,$F4,$0F,$77
	.dc.b	$95,$6C,$97,$EC,$AE,$3C,$DA,$8C,$4E,$9F,$04,$F6,$C7,$51,$89,$D6
	.dc.b	$3E,$17,$FC,$5D,$8F,$86,$7D,$E7,$42,$1C,$4F,$BD,$87,$4A,$E5,$1A
	.dc.b	$6B,$F5,$2F,$59,$FF,$89,$71,$F8,$2D,$C7,$D1,$8F,$86,$FE,$39,$CE
	.dc.b	$74,$F5,$C7,$9E,$7E,$68,$F2,$13,$26,$5A,$8E,$66,$6F,$9A,$C3,$C1
	.dc.b	$97,$93,$0E,$98,$7A,$89,$8F,$39,$D5,$F4,$9D,$63,$9C,$BE,$6E,$5D
	.dc.b	$B2,$EB,$18,$CB,$AC,$F8,$AD,$97,$1C,$7E,$05,$78,$B1,$C4,$7A,$63
	.dc.b	$E0,$63,$D2,$36,$78,$F0,$E9,$7C,$30,$F5,$47,$7C,$30,$F8,$4F,$36
	.dc.b	$78,$C7,$CA,$7C,$2B,$C6,$1D,$63,$C8,$E8,$C1,$33,$FC,$DD,$70,$7E
	.dc.b	$39,$F9,$88,$E0,$E3,$3F,$C9,$E3,$B7,$F0,$75,$6B,$F0,$07,$08,$F9
	.dc.b	$4F,$D1,$9D,$3F,$82,$CF,$E2,$3F,$1C,$F2,$4F,$F1,$0C,$77,$9F,$D8
	.dc.b	$9E,$87,$CA,$C7,$D2,$E0,$E4,$FA,$5E,$66,$EB,$F8,$DE,$1D,$33,$F9
	.dc.b	$73,$C6,$FE,$8E,$7F,$2E,$31,$3F,$46,$3E,$5C,$74,$3B,$DF,$ED,$33
	.dc.b	$E9,$3C,$67,$57,$D1,$F8,$47,$48,$F4,$67,$F0,$A7,$59,$7C,$FA,$FD
	.dc.b	$C9,$B3,$9E,$5D,$9F,$D3,$D7,$1C,$BC,$D1,$E0,$CF,$57,$FC,$36,$39
	.dc.b	$3D,$5F,$D6,$65,$AC,$FE,$84,$69,$FB,$ED,$3F,$86,$E1,$DF,$3C,$47
	.dc.b	$F3,$CC,$39,$3F,$CB,$8C,$E7,$FC,$3A,$B8,$47,$C4,$CB,$F4,$2E,$9F
	.dc.b	$D0,$36,$FF,$53,$8F,$48,$EC,$64,$F9,$67,$38,$F9,$71,$BB,$FD,$BC
	.dc.b	$EF,$2F,$84,$EC,$FB,$58,$79,$23,$E2,$4F,$F0,$A9,$E1,$97,$E2,$1A
	.dc.b	$7F,$B8,$EB,$84,$7E,$7A,$BF,$20,$72,$C7,$E9,$BE,$03,$18,$F9,$AF
	.dc.b	$99,$EC,$3C,$91,$E2,$CF,$E0,$61,$DA,$7F,$34,$7E,$10,$DE,$1F,$59
	.dc.b	$A7,$D6,$CB,$E1,$67,$F0,$EF,$9C,$62,$7C,$39,$FF,$32,$7C,$78,$77
	.dc.b	$7F,$3A,$77,$C7,$D2,$FE,$48,$DD,$73,$7C,$27,$F3,$DB,$EE,$F6,$C3
	.dc.b	$C4,$F0,$30,$78,$A3,$6F,$2C,$3E,$93,$73,$FC,$2B,$73,$FC,$2B,$C8
	.dc.b	$C1,$F8,$4C,$7C,$E7,$E7,$F2,$F6,$5F,$EB,$B3,$FC,$FD,$70,$AF,$AD
	.dc.b	$1F,$6C,$FE,$3B,$1D,$63,$9C,$7C,$F6,$C7,$CF,$35,$1C,$5F,$ED,$D8
	.dc.b	$FB,$27,$71,$FD,$87,$2F,$86,$F8,$6B,$F8,$53,$C8,$FB,$E7,$CB,$3E
	.dc.b	$41,$B3,$EA,$3D,$DD,$19,$3F,$7A,$73,$3E,$D9,$A9,$F1,$DF,$ED,$73
	.dc.b	$FB,$51,$8C,$FE,$7B,$73,$F9,$F7,$D6,$FF,$5A,$DC,$BD,$D8,$F8,$E7
	.dc.b	$EC,$DF,$1A,$FA,$93,$9C,$3F,$98,$C7,$3C,$BE,$1D,$6C,$FE,$7B,$19
	.dc.b	$3F,$0C,$7C,$27,$55,$F6,$9B,$1F,$C4,$4F,$85,$9F,$E0,$2F,$F6,$75
	.dc.b	$B3,$C5,$97,$F3,$CB,$E3,$7F,$20,$F0,$1E,$D3,$C1,$3C,$1F,$F0,$66
	.dc.b	$1E,$88,$C4,$FA,$0E,$78,$FC,$66,$E5,$F1,$AF,$EE,$CF,$C6,$AE,$27
	.dc.b	$F0,$F8,$F4,$E7,$F5,$23,$31,$ED,$BF,$3E,$7E,$DC,$BF,$AD,$65,$E3
	.dc.b	$AE,$B9,$FC,$EB,$F1,$CE,$67,$F0,$79,$7E,$05,$FC,$33,$FA,$57,$F3
	.dc.b	$27,$C0,$39,$DF,$F5,$DA,$F5,$C6,$4E,$77,$FC,$F1,$F9,$07,$A8,$EA
	.dc.b	$FD,$27,$A9,$93,$E7,$C7,$47,$0F,$A1,$FC,$09,$8C,$7E,$1B,$FC,$A6
	.dc.b	$38,$1F,$13,$2F,$91,$1E,$83,$F2,$25,$8F,$AD,$3C,$B0,$FA,$CD,$CF
	.dc.b	$EB,$57,$B3,$1F,$81,$39,$8F,$81,$5F,$1D,$FD,$E1,$E6,$BF,$13,$F0
	.dc.b	$2B,$FB,$1D,$F4,$AE,$07,$E0,$67,$C9,$7F,$81,$BE,$31,$F1,$23,$C0
	.dc.b	$FC,$13,$67,$F0,$22,$9F,$E7,$B3,$CB,$8F,$CF,$7C,$A7,$D4,$9E,$A7
	.dc.b	$CD,$71,$3F,$33,$3F,$C6,$47,$F8,$2C,$F6,$46,$72,$F9,$B1,$F1,$4F
	.dc.b	$9A,$5A,$F9,$AF,$93,$0F,$9A,$DC,$7E,$16,$3E,$59,$F8,$53,$C7,$0F
	.dc.b	$84,$77,$7E,$51,$FD,$16,$B8,$E5,$FE,$0E,$CF,$F3,$97,$F3,$30,$FD
	.dc.b	$06,$1F,$76,$73,$1D,$CC,$E1,$E5,$8C,$DF,$F8,$73,$3F,$83,$97,$F8
	.dc.b	$12,$73,$5F,$D7,$A7,$38,$7F,$AD,$A7,$35,$FC,$FA,$73,$5F,$E6,$43
	.dc.b	$E6,$9F,$59,$CB,$F5,$23,$F3,$35,$EF,$8F,$C0,$1F,$1C,$FB,$77,$EF
	.dc.b	$7E,$99,$CF,$3F,$A7,$97,$C1,$9F,$E3,$95,$CE,$FE,$F1,$DA,$F8,$1F
	.dc.b	$31,$FE,$87,$9F,$CC,$8F,$EE,$59,$7B,$2B,$FA,$76,$5E,$8C,$7E,$B9
	.dc.b	$DE,$BF,$02,$DC,$3F,$03,$1F,$6C,$FC,$B1,$4F,$CB,$1F,$9B,$31,$1F
	.dc.b	$86,$6E,$1E,$39,$FE,$AF,$87,$8E,$7F,$B2,$DF,$8F,$3F,$E2,$15,$ED
	.dc.b	$9E,$D3,$9C,$BF,$73,$1F,$2C,$F0,$63,$E7,$CB,$FA,$24,$66,$FF,$4B
	.dc.b	$1F,$89,$8F,$9E,$74,$CF,$E7,$E3,$F8,$BC,$3E,$7C,$FC,$C3,$EC,$9E
	.dc.b	$A3,$EC,$9F,$30,$E7,$87,$85,$FF,$26,$57,$49,$FE,$0F,$1C,$30,$F8
	.dc.b	$2F,$0B,$F0,$3F,$E8,$2A,$E0,$FE,$94,$E1,$87,$F7,$3C,$BE,$44,$7F
	.dc.b	$80,$B2,$F8,$8F,$F5,$A3,$A6,$3F,$A1,$BE,$78,$FE,$12,$79,$9F,$D6
	.dc.b	$2F,$9E,$5F,$C8,$70,$E7,$5F,$E6,$2C,$7F,$8B,$BF,$2A,$FF,$5C,$76
	.dc.b	$3F,$55,$87,$69,$FF,$04,$67,$F7,$DF,$44,$EE,$BF,$59,$5B,$8F,$ED
	.dc.b	$35,$B7,$C8,$6E,$BF,$2F,$3B,$8F,$EB,$F3,$B7,$FB,$0E,$3E,$AC,$FF
	.dc.b	$93,$63,$EA,$C3,$F0,$75,$F6,$4F,$C3,$CF,$D0,$C7,$DD,$1F,$87,$9F
	.dc.b	$D7,$1F,$87,$7F,$7E,$7E,$1C,$FD,$73,$EE,$7E,$59,$E2,$3D,$07,$E7
	.dc.b	$31,$F4,$65,$FD,$1B,$2F,$43,$FD,$23,$2F,$46,$7F,$E8,$BB,$FC,$7C
	.dc.b	$FE,$2D,$FD,$6E,$3F,$45,$FD,$6D,$70,$7F,$D4,$38,$76,$CB,$F9,$96
	.dc.b	$3D,$9F,$E6,$95,$BB,$FF,$35,$56,$F1,$FF,$74,$5E,$DF,$F0,$8E,$1E
	.dc.b	$A3,$F1,$D3,$B7,$FD,$10,$7C,$A8,$FE,$34,$FC,$6C,$7D,$19,$FD,$18
	.dc.b	$FE,$0F,$1E,$B3,$FD,$2C,$F1,$3F,$54,$71,$CB,$EA,$67,$FE,$6A,$8F
	.dc.b	$87,$18,$7F,$C5,$11,$F5,$9F,$C9,$9C,$6F,$F4,$0F,$D6,$3F,$1A,$7D
	.dc.b	$6A,$FA,$0F,$B5,$FD,$06,$1E,$DC,$BF,$D7,$13,$BC,$3E,$4B,$B7,$F1
	.dc.b	$B3,$BC,$3F,$D1,$47,$8E,$7E,$A9,$BC,$3E,$E5,$FF,$38,$9F,$D5,$9B
	.dc.b	$C3,$E2,$67,$F5,$63,$FD,$2F,$7F,$02,$FF,$84,$BF,$1E,$FF,$84,$BE
	.dc.b	$F3,$F6,$27,$BC,$FD,$29,$E9,$8F,$B6,$EE,$3F,$62,$EE,$FF,$28,$EC
	.dc.b	$FF,$03,$1B,$AF,$D8,$BB,$C7,$F1,$AE,$EB,$FB,$63,$EF,$3F,$26,$6E
	.dc.b	$FF,$1A,$EE,$BF,$B7,$3B,$9F,$C9,$BB,$3F,$1A,$FE,$7E,$FE,$2C,$6E
	.dc.b	$7E,$2C,$FC,$B8,$FD,$C1,$FC,$DD,$C3,$FB,$F7,$95,$FE,$26,$3F,$C9
	.dc.b	$D3,$F4,$9E,$0F,$EF,$E7,$FA,$85,$7E,$6B,$1F,$BD,$5F,$E5,$E9,$E7
	.dc.b	$87,$E8,$1D,$E1,$D7,$2F,$D0,$57,$63,$CA,$7F,$09,$3E,$E1,$FC,$F4
	.dc.b	$FB,$98,$FF,$26,$7E,$81,$FC,$24,$F8,$95,$F9,$47,$67,$F2,$63,$67
	.dc.b	$ED,$8D,$C7,$E5,$1F,$69,$FA,$68,$F3,$DF,$E9,$A3,$C7,$1F,$D1,$5E
	.dc.b	$A7,$F1,$B3,$AE,$3F,$A6,$7A,$9F,$0E,$BF,$35,$87,$CD,$9F,$F0,$01
	.dc.b	$FB,$FC,$FF,$86,$57,$E3,$2F,$C2,$FE,$60,$F0,$DF,$F2,$B7,$C3,$5F
	.dc.b	$B6,$7C,$37,$FB,$07,$C2,$7F,$B8,$67,$C3,$87,$D3,$7E,$E4,$7C,$FB
	.dc.b	$FC,$1D,$F8,$E3,$F2,$47,$F8,$EE,$3C,$67,$E9,$1F,$19,$FC,$20,$F1
	.dc.b	$DF,$F0,$27,$C7,$7F,$BC,$BF,$C4,$1E,$0B,$FC,$44,$F9,$95,$7F,$BC
	.dc.b	$A7,$CF,$3F,$C8,$2F,$C3,$1F,$AE,$8F,$09,$FB,$A3,$C3,$7F,$BA,$9D
	.dc.b	$C7,$E1,$4D,$DF,$EE,$9D,$9F,$3E,$BC,$97,$F9,$5C,$BF,$37,$3F,$24
	.dc.b	$FC,$29,$FC,$60,$F6,$9F,$BA,$3A,$C7,$F0,$07,$AD,$FF,$00,$76,$FE
	.dc.b	$E9,$DE,$1F,$C0,$1D,$DF,$F0,$07,$6F,$EE,$8F,$92,$FE,$9B,$0F,$92
	.dc.b	$F2,$3F,$81,$C7,$C8,$AF,$E4,$27,$C2,$8F,$D0,$1B,$7F,$4C,$7E,$D9
	.dc.b	$F9,$D1,$F0,$AB,$F4,$13,$F8,$4B,$FC,$C6,$3F,$96,$8F,$61,$D7,$0F
	.dc.b	$D0,$61,$D7,$0F,$D6,$3F,$8D,$7D,$2F,$F0,$F7,$CC,$7E,$EE,$79,$5F
	.dc.b	$E6,$67,$91,$FA,$93,$E8,$D7,$F1,$E7,$E2,$80,$FF,$BF,$5E,$27,$D9
	.dc.b	$3C,$47,$EF,$6B,$DC,$FE,$B6,$BD,$D8,$7F,$4A,$7E,$21,$F2,$A7,$6F
	.dc.b	$E8,$4D,$CF,$F1,$57,$71,$EB,$36,$7E,$5D,$D9,$FC,$24,$DC,$7F,$15
	.dc.b	$7E,$0C,$FE,$9E,$3C,$C7,$DF,$BF,$CA,$3F,$A7,$7F,$4C,$FC,$47,$F8
	.dc.b	$66,$7F,$D9,$63,$E2,$61,$F0,$EF,$C3,$87,$F0,$97,$C3,$3F,$C6,$5F
	.dc.b	$0B,$F9,$B7,$6F,$EF,$5D,$CF,$F0,$F7,$67,$E6,$CD,$C7,$EC,$DD,$DF
	.dc.b	$F0,$37,$75,$F9,$B9,$F6,$9F,$83,$3D,$58,$7F,$15,$7E,$03,$F9,$28
	.dc.b	$FC,$4D,$FF,$15,$9F,$D3,$9F,$C5,$63,$E7,$BF,$BF,$AF,$1C,$FE,$FD
	.dc.b	$F1,$D7,$F0,$5A,$DD,$7E,$92,$77,$7F,$B9,$77,$1F,$C1,$4F,$19,$F8
	.dc.b	$79,$F1,$CF,$E9,$23,$F9,$39,$FE,$DD,$8E,$B7,$FA,$4A,$EB,$7F,$BF
	.dc.b	$7A,$9F,$AA,$9D,$DF,$E8,$63,$77,$FB,$C3,$E2,$9F,$B6,$3C,$6F,$E4
	.dc.b	$E7,$C7,$5F,$99,$7F,$5B,$3F,$8D,$7A,$C7,$E3,$63,$F8,$D4,$FD,$47
	.dc.b	$F0,$F4,$9F,$D6,$4F,$FB,$72,$FF,$60,$EE,$BF,$5F,$5B,$7F,$D6,$87
	.dc.b	$E5,$6B,$FD,$8A,$FE,$C0,$EF,$7E,$27,$F1,$CF,$A1,$FD,$53,$E8,$7F
	.dc.b	$28,$EE,$BF,$1C,$ED,$FD,$53,$BB,$FC,$A3,$B3,$F1,$D3,$E0,$3E,$3B
	.dc.b	$FB,$33,$91,$FA,$33,$95,$FE,$5D,$E5,$1F,$12,$FE,$1C,$7F,$9B,$EF
	.dc.b	$EA,$9F,$CA,$9E,$31,$F8,$D9,$EE,$7E,$4E,$3B,$BF,$AB,$76,$FE,$34
	.dc.b	$FD,$11,$F1,$EF,$E7,$C7,$8A,$3E,$7C,$F8,$9F,$F7,$49,$F3,$C5,$27
	.dc.b	$F9,$24,$7F,$A9,$1E,$F1,$F8,$F8,$F2,$9F,$A1,$8F,$2B,$FB,$97,$67
	.dc.b	$F5,$38,$FC,$05,$7F,$0B,$8D,$BF,$AD,$77,$3F,$B3,$3A,$C7,$E8,$63
	.dc.b	$77,$FA,$D7,$73,$FA,$73,$71,$FA,$18,$DD,$FE,$9D,$DD,$7F,$81,$A3
	.dc.b	$E1,$57,$F9,$24,$F9,$31,$FD,$92,$FE,$AD,$7F,$52,$8C,$3D,$27,$F5
	.dc.b	$87,$CF,$BF,$E5,$2F,$F9,$A4,$FC,$01,$F9,$C3,$FA,$5C,$F8,$5F,$9F
	.dc.b	$3F,$3A,$3F,$D6,$F5,$FB,$C8,$F9,$07,$D9,$8F,$91,$3F,$C8,$1F,$E3
	.dc.b	$53,$B8,$F9,$07,$AA,$BF,$9F,$BE,$07,$F9,$FB,$E7,$BF,$E4,$CE,$DF
	.dc.b	$E4,$CF,$58,$FE,$28,$FF,$26,$3E,$71,$B3,$F8,$D1,$B8,$FE,$10,$FF
	.dc.b	$0E,$9F,$EC,$0F,$D4,$8F,$E1,$15,$F9,$48,$C1,$FB,$A3,$91,$FB,$A3
	.dc.b	$95,$FF,$63,$77,$5F,$BA,$76,$FE,$E9,$D9,$FC,$5C,$DC,$7E,$EA,$7C
	.dc.b	$11,$E9,$9F,$E4,$0F,$29,$FD,$D3,$C8,$FE,$12,$72,$3E,$DC,$7F,$1B
	.dc.b	$BF,$BF,$1F,$E0,$A3,$F5,$A7,$DF,$8F,$C4,$BF,$E1,$13,$E8,$1F,$CC
	.dc.b	$4F,$14,$FF,$0B,$7C,$47,$EF,$4F,$15,$7F,$14,$7C,$47,$69,$F1,$5F
	.dc.b	$DA,$9F,$C1,$D7,$E7,$AF,$F9,$65,$7E,$EE,$FF,$BD,$E1,$FD,$F5,$E5
	.dc.b	$5F,$C4,$CF,$CE,$4F,$F1,$38,$F0,$D7,$F1,$B7,$C3,$7F,$92,$7F,$92
	.dc.b	$1E,$A3,$F8,$B1,$EA,$3F,$81,$1C,$B0,$FE,$38,$EC,$FE,$10,$6E,$3F
	.dc.b	$81,$3B,$3F,$8E,$3E,$C7,$15,$FC,$45,$E5,$1F,$C1,$5E,$55,$FC,$15
	.dc.b	$DD,$FE,$A9,$F4,$C7,$C4,$8F,$81,$1C,$2F,$F8,$33,$C0,$FD,$5D,$78
	.dc.b	$5F,$B3,$7F,$9C,$7C,$37,$FC,$3D,$E4,$7F,$04,$9F,$E0,$EE,$27,$E6
	.dc.b	$3C,$AF,$F3,$8F,$2B,$FD,$2D,$FE,$32,$BF,$93,$BE,$79,$FE,$6E,$7E
	.dc.b	$EE,$7F,$B2,$DF,$68,$FA,$6F,$9C,$C1,$FA,$03,$13,$FC,$12,$BF,$2B
	.dc.b	$3F,$E5,$D9,$DC,$F8,$6F,$F8,$83,$E1,$96,$7F,$EF,$0A,$FF,$68,$CF
	.dc.b	$F0,$6B,$FE,$25,$5F,$C1,$A7,$BE,$1F,$C3,$23,$CB,$5B,$C3,$F2,$D3
	.dc.b	$FC,$CA,$B9,$BF,$C2,$AB,$E3,$1F,$BB,$3E,$31,$FC,$1C,$E6,$EE,$3D
	.dc.b	$2F,$E8,$EB,$B4,$FD,$B3,$C8,$E1,$EA,$7E,$F8,$EB,$7F,$C3,$5E,$A7
	.dc.b	$F1,$33,$75,$FB,$37,$77,$FC,$36,$BE,$44,$7E,$3A,$FF,$A6,$E1,$F1
	.dc.b	$F1,$FD,$86,$1F,$A9,$3E,$A5,$F8,$DF,$4D,$FF,$05,$7E,$4B,$FC,$B8
	.dc.b	$F1,$E3,$FC,$7D,$F1,$CF,$ED,$5F,$19,$FC,$5B,$1F,$CE,$BC,$1F,$E3
	.dc.b	$AF,$0C,$7F,$67,$7D,$AF,$F9,$1C,$EC,$FE,$24,$6C,$FD,$9D,$F9,$CF
	.dc.b	$C5,$5F,$F0,$83,$FB,$33,$88,$FD,$04,$62,$FF,$05,$7F,$01,$FB,$15
	.dc.b	$BA,$F0,$1F,$C2,$0F,$05,$FE,$FE,$BC,$D7,$FC,$9E,$77,$1F,$C2,$1D
	.dc.b	$C7,$EF,$DF,$86,$7F,$1E,$3F,$53,$3F,$C8,$DF,$D3,$F8,$1F,$E1,$E8
	.dc.b	$C5,$23,$15,$FD,$D5,$B5,$FE,$A4,$7F,$DB,$51,$FC,$37,$EC,$9F,$B9
	.dc.b	$F1,$AF,$F6,$67,$DE,$3E,$4F,$FD,$9D,$5A,$C3,$FC,$5B,$39,$A4,$E5
	.dc.b	$E3,$5C,$2F,$A4,$2B,$83,$F0,$0F,$F1,$47,$D9,$BF,$01,$76,$F6,$C3
	.dc.b	$FD,$BD,$39,$8C,$52,$30,$F8,$0C,$57,$27,$07,$73,$FC,$61,$1D,$FB
	.dc.b	$79,$B6,$39,$35,$F4,$5F,$8A,$95,$E2,$9E,$9A,$FF,$89,$76,$FA,$7C
	.dc.b	$3F,$43,$EF,$5C,$E8,$F4,$C6,$1B,$19,$BF,$54,$E5,$F4,$DF,$F3,$48
	.dc.b	$C4,$7F,$8C,$83,$8B,$FE,$D0,$DE,$14,$EA,$FB,$3C,$74,$79,$3F,$E9
	.dc.b	$B8,$F7,$9F,$35,$C6,$FF,$7E,$77,$C9,$F2,$27,$3B,$FF,$90,$F9,$3F
	.dc.b	$55,$5A,$C3,$E5,$D6,$9E,$97,$C4,$E5,$F1,$12,$36,$F5,$EF,$F9,$DB
	.dc.b	$FF,$54,$C3,$CF,$AF,$87,$BE,$61,$DA,$3C,$84,$70,$43,$DA,$7E,$B4
	.dc.b	$FE,$F7,$3D,$1D,$6B,$EF,$CF,$F6,$0D,$AF,$81,$DF,$2B,$DA,$FC,$3E
	.dc.b	$FF,$3E,$DF,$0B,$0E,$3B,$FE,$69,$B7,$F3,$70,$F9,$5D,$FF,$2F,$BE
	.dc.b	$4F,$C8,$7E,$77,$17,$4E,$2F,$F2,$FE,$DF,$63,$E2,$3E,$91,$2F,$E7
	.dc.b	$FC,$47,$D5,$F8,$89,$1F,$49,$FB,$1E,$0F,$38,$FF,$31,$DF,$37,$EB
	.dc.b	$DF,$37,$F0,$71,$F8,$6B,$FC,$07,$C4,$49,$FB,$AF,$E0,$FE,$23,$A7
	.dc.b	$57,$F9,$5A,$FA,$57,$2E,$06,$3D,$79,$7E,$76,$5F,$8B,$F8,$89,$7F
	.dc.b	$88,$7F,$1D,$F1,$1D,$38,$FF,$CA,$19,$7E,$63,$E2,$3F,$24,$4B,$FC
	.dc.b	$7F,$C4,$7E,$57,$E2,$24,$7E,$49,$FD,$FF,$E0,$5F,$EC,$68,$FC,$1F
	.dc.b	$6F,$E3,$77,$5F,$39,$F4,$61,$F9,$F8,$E3,$F3,$32,$7F,$83,$E0,$8E
	.dc.b	$BF,$29,$E7,$9F,$F2,$6E,$7F,$AE,$BE,$9C,$1F,$8D,$C7,$A5,$72,$F8
	.dc.b	$BF,$2F,$C5,$C8,$C1,$FC,$0B,$BB,$F5,$BA,$3E,$2F,$D6,$49,$F1,$FC
	.dc.b	$5F,$BB,$E2,$E4,$79,$9F,$83,$F1,$7E,$58,$FD,$0D,$F3,$7D,$77,$CD
	.dc.b	$FC,$0C,$2F,$E4,$FC,$5F,$CA,$F8,$B9,$5F,$0A,$3A,$72,$7F,$97,$78
	.dc.b	$58,$71,$AE,$A4,$3A,$E1,$F5,$3E,$2E,$47,$CC,$7E,$BF,$C5,$F4,$DB
	.dc.b	$FC,$B1,$87,$4F,$8B,$FB,$7F,$17,$30,$FA,$71,$F7,$88,$E0,$87,$E0
	.dc.b	$2B,$E5,$C6,$A3,$DE,$DE,$4F,$E4,$19,$BE,$2F,$F1,$3F,$17,$2F,$F0
	.dc.b	$31,$E5,$F1,$BF,$C8,$98,$FF,$94,$7D,$87,$8D,$FD,$CB,$5F,$46,$1E
	.dc.b	$2C,$FA,$E9,$F9,$4F,$8B,$FC,$AF,$C5,$FE,$5B,$E2,$E5,$FE,$3A,$3B
	.dc.b	$3F,$ED,$1C,$3E,$6F,$C5,$FE,$70,$91,$F9,$4F,$8B,$FE,$BB,$F8,$1F
	.dc.b	$F7,$0C,$FF,$9D,$37,$D1,$FE,$DA,$DA,$FE,$D8,$BE,$57,$BF,$99,$FE
	.dc.b	$D9,$CB,$BD,$7E,$83,$2E,$EF,$E1,$B2,$EF,$F2,$5E,$0F,$90,$93,$FD
	.dc.b	$B3,$24,$F4,$F2,$BF,$A4,$72,$FF,$94,$38,$FE,$37,$43,$3E,$BC,$1F
	.dc.b	$49,$FD,$CE,$9E,$CE,$6F,$6F,$C1,$4D,$7B,$47,$57,$FC,$98,$7F,$8A
	.dc.b	$B3,$F7,$F9,$5F,$13,$E3,$A4,$F9,$63,$E3,$FC,$1C,$8F,$43,$F2,$BE
	.dc.b	$3B,$A4,$7F,$49,$EB,$FB,$AF,$F6,$EE,$5F,$F6,$C7,$2B,$93,$FB,$6C
	.dc.b	$FC,$1C,$BB,$BF,$CB,$23,$2D,$F8,$EF,$AF,$F1,$D2,$3E,$44,$78,$3C
	.dc.b	$4F,$F0,$76,$DF,$EC,$2C,$7B,$6D,$D3,$8F,$EE,$E7,$F8,$08,$FC,$0F
	.dc.b	$C7,$4C,$BE,$93,$F8,$4F,$8E,$E9,$C5,$F9,$7E,$7F,$CB,$6E,$F8,$EF
	.dc.b	$C4,$93,$0F,$B5,$F1,$DF,$8C,$F8,$E9,$1F,$75,$FC,$7F,$C7,$74,$EC
	.dc.b	$FB,$93,$F1,$78,$7F,$1F,$87,$38,$F8,$4F,$E5,$B0,$FC,$BF,$C3,$C9
	.dc.b	$FC,$43,$F9,$AF,$8A,$E9,$1F,$E2,$13,$FD,$7B,$A7,$A2,$FA,$71,$7E
	.dc.b	$79,$E9,$3F,$91,$F8,$9F,$F1,$FF,$E0,$9F,$E1,$3E,$53,$78,$3F,$C7
	.dc.b	$B5,$FC,$BF,$6F,$ED,$F2,$F8,$9E,$3F,$13,$2B,$FC,$8F,$39,$6C,$F6
	.dc.b	$23,$83,$A1,$FD,$C2,$38,$3F,$DB,$9B,$9F,$F8,$BD,$BE,$37,$F9,$89
	.dc.b	$B3,$FE,$61,$7F,$C1,$79,$7A,$BE,$27,$DA,$7B,$B8,$BE,$3F,$6F,$FA
	.dc.b	$27,$BB,$F2,$F8,$7F,$90,$70,$E7,$87,$C2,$CF,$CF,$3E,$3F,$89,$F8
	.dc.b	$FF,$13,$F2,$3E,$26,$4F,$C0,$8F,$F3,$9E,$AF,$CC,$9F,$05,$FC,$7F
	.dc.b	$37,$C5,$53,$FD,$27,$5B,$9B,$5C,$30,$F9,$7F,$15,$F5,$8E,$B1,$98
	.dc.b	$FF,$2A,$D7,$17,$FD,$00,$D7,$31,$BE,$2F,$F4,$B4,$E5,$B1,$CF,$81
	.dc.b	$C3,$FE,$66,$AE,$51,$8B,$FF,$51,$56,$EB,$4D,$8C,$E3,$E6,$23,$E4
	.dc.b	$8F,$15,$FD,$9F,$87,$FC,$8F,$C3,$CA,$FC,$4C,$70,$7F,$CF,$93,$E2
	.dc.b	$C3,$BE,$1F,$EB,$AF,$BC,$7E,$97,$FE,$77,$CF,$CD,$7E,$D3,$78,$78
	.dc.b	$6F,$F3,$FF,$19,$FA,$0E,$DF,$D0,$FC,$64,$AF,$CC,$C7,$E9,$3E,$33
	.dc.b	$EC,$12,$3F,$70,$7E,$3F,$E3,$3E,$E6,$1F,$AA,$24,$7E,$8B,$E2,$3F
	.dc.b	$5B,$F1,$32,$3F,$4A,$FE,$C3,$E3,$3B,$47,$4E,$CF,$D8,$E3,$DE,$3F
	.dc.b	$29,$8F,$67,$F5,$B8,$F6,$7F,$4D,$8F,$FA,$E3,$E2,$3F,$C7,$5C,$DF
	.dc.b	$E9,$57,$FC,$65,$3F,$EB,$A6,$BF,$E5,$A7,$7B,$FE,$AF,$1F,$07,$07
	.dc.b	$F3,$8E,$94,$7E,$67,$3F,$16,$1F,$B5,$CF,$C5,$97,$ED,$73,$F1,$7C
	.dc.b	$5F,$8F,$E2,$E5,$F8,$63,$CD,$F1,$7F,$E0,$1D,$8D,$31,$F1,$59,$F8
	.dc.b	$8F,$64,$CB,$F5,$47,$BF,$E2,$E4,$7B,$5F,$85,$F1,$5D,$A3,$E6,$57
	.dc.b	$66,$E1,$D9,$CD,$F5,$7E,$24,$FC,$B8,$F9,$9F,$17,$F3,$7E,$2E,$57
	.dc.b	$CB,$8E,$CF,$FB,$26,$BB,$36,$FB,$39,$BE,$DB,$7D,$5A,$ED,$3F,$57
	.dc.b	$E2,$FE,$D1,$1F,$B1,$F1,$1F,$77,$E2,$FE,$47,$C5,$49,$F6,$C7,$CB
	.dc.b	$F8,$BF,$C1,$12,$3F,$01,$F1,$7F,$85,$F8,$A9,$1F,$83,$7F,$11,$F1
	.dc.b	$7D,$63,$E3,$E1,$D5,$F6,$61,$D5,$FC,$56,$1D,$5F,$C5,$61,$F9,$1F
	.dc.b	$87,$91,$F8,$F7,$F2,$9F,$0F,$D6,$3E,$36,$1D,$7B,$7F,$C0,$BB,$32
	.dc.b	$ED,$5F,$95,$CB,$B6,$FF,$9B,$CF,$FD,$F3,$3F,$EB,$2F,$C1,$FF,$C6
	.dc.b	$DE,$01,$FF,$1B,$47,$FB,$FA,$BF,$DE,$D1,$88,$E1,$E0,$7F,$36,$FC
	.dc.b	$DB,$F4,$7A,$7F,$85,$AB,$FC,$67,$5F,$B3,$6B,$FC,$4A,$77,$E4,$FF
	.dc.b	$25,$7F,$57,$87,$F2,$CF,$D4,$F6,$DF,$8C,$ED,$F1,$F3,$0F,$E6,$B7
	.dc.b	$C3,$3E,$53,$99,$FE,$77,$3D,$94,$A3,$0D,$9F,$E9,$31,$83,$A4,$F0
	.dc.b	$6C,$EE,$3C,$33,$F2,$AF,$E2,$FC,$75,$A5,$35,$F9,$93,$98,$C6,$88
	.dc.b	$C4,$FD,$1F,$83,$F5,$6D,$FC,$CB,$3F,$26,$7F,$CD,$FF,$49,$7A,$BE
	.dc.b	$88,$F8,$1A,$FF,$54,$AF,$4E,$3F,$6F,$E3,$BA,$E9,$FE,$44,$AF,$B7
	.dc.b	$F1,$38,$33,$F1,$3A,$F8,$EF,$F4,$65,$1A,$63,$E3,$B3,$F1,$BD,$09
	.dc.b	$C5,$F8,$7E,$3F,$83,$A7,$E2,$FE,$3B,$C3,$F1,$B2,$B4,$7C,$27,$CB
	.dc.b	$F1,$BE,$18,$FC,$46,$5E,$07,$F1,$59,$78,$1E,$38,$7E,$5F,$2F,$0C
	.dc.b	$7E,$6F,$1F,$7F,$C4,$48,$DE,$9F,$E5,$16,$FC,$5F,$75,$CD,$F9,$9C
	.dc.b	$FC,$11,$8E,$BF,$5D,$B4,$F1,$4F,$EA,$73,$F0,$F1,$7E,$9B,$4F,$9F
	.dc.b	$F1,$BF,$43,$E3,$65,$F8,$A3,$E9,$FC,$6F,$CE,$F8,$D9,$3F,$A7,$8F
	.dc.b	$A3,$F1,$BF,$84,$FC,$2F,$FC,$98,$7F,$93,$75,$FF,$25,$D3,$3F,$CE
	.dc.b	$BF,$8F,$9F,$F1,$89,$3E,$27,$97,$C6,$78,$27,$F6,$99,$78,$1F,$71
	.dc.b	$F9,$4C,$BC,$2F,$CE,$BF,$0F,$C6,$FE,$2B,$3F,$27,$0C,$AD,$3E,$7F
	.dc.b	$8D,$ED,$1D,$39,$FF,$86,$65,$DF,$97,$F8,$36,$7D,$AB,$D5,$D5,$F8
	.dc.b	$8D,$3B,$C7,$D8,$CF,$E1,$7C,$67,$C3,$F0,$64,$F9,$63,$E3,$7C,$67
	.dc.b	$FB,$8E,$CD,$31,$F1,$99,$F8,$CF,$98,$4B,$F7,$7C,$57,$CF,$F8,$B9
	.dc.b	$1F,$05,$FA,$5F,$13,$DA,$3F,$C1,$D8,$F6,$6E,$3D,$9C,$E1,$D5,$F6
	.dc.b	$9F,$55,$FB,$98,$7D,$DF,$88,$FB,$DF,$13,$2B,$E8,$47,$77,$E5,$61
	.dc.b	$DD,$B8,$77,$73,$87,$7E,$9F,$C1,$F2,$2D,$3E,$0A,$FB,$9F,$0D,$F8
	.dc.b	$D2,$3F,$7F,$E1,$BF,$8B,$2F,$8D,$47,$F4,$8E,$6F,$F4,$05,$5E,$9F
	.dc.b	$E9,$75,$FE,$1E,$BF,$F1,$E9,$C6,$F4,$FF,$24,$8E,$6F,$F0,$2F,$C2
	.dc.b	$1F,$8D,$7B,$75,$FC,$DC,$63,$6F,$1D,$F0,$BF,$96,$75,$3F,$85,$94
	.dc.b	$CB,$BA,$FD,$2C,$65,$B1,$CF,$3F,$F4,$D6,$3E,$6A,$E5,$18,$7D,$73
	.dc.b	$CF,$A3,$DF,$5E,$42,$4F,$E9,$FE,$27,$F7,$04,$7F,$55,$F1,$3A,$7F
	.dc.b	$D9,$D7,$E1,$D7,$9C,$76,$AF,$14,$F7,$8F,$91,$7C,$72,$F8,$BF,$83
	.dc.b	$3C,$EF,$AF,$F6,$8F,$88,$A4,$C3,$E2,$C6,$5B,$19,$7E,$5C,$65,$B1
	.dc.b	$9B,$F9,$33,$96,$C6,$75,$FF,$05,$E8,$EF,$FC,$4B,$70,$FA,$AD,$7E
	.dc.b	$CB,$6F,$EA,$B6,$B9,$D7,$22,$57,$26,$BC,$9B,$5F,$7B,$ED,$3F,$DA
	.dc.b	$6D,$E8,$FC,$DC,$62,$FF,$C4,$2D,$BF,$EE,$ED,$4D,$24,$F2,$24,$7E
	.dc.b	$42,$32,$FF,$27,$F8,$FF,$E7,$1F,$1F,$23,$06,$E3,$FC,$05,$F2,$2F
	.dc.b	$CD,$7F,$DE,$2D,$F2,$7C,$F3,$2B,$FB,$3E,$6F,$DF,$DE,$AF,$F8,$2E
	.dc.b	$30,$EE,$7F,$4B,$9A,$F1,$5F,$15,$BA,$FC,$26,$D7,$C2,$6B,$F0,$9B
	.dc.b	$5C,$E3,$D4,$7F,$72,$FC,$3F,$FD,$42,$FF,$87,$DF,$F0,$89,$1E,$33
	.dc.b	$DB,$AD,$F3,$2F,$92,$CD,$74,$3F,$93,$E4,$FD,$F7,$AF,$86,$CB,$85
	.dc.b	$78,$6F,$FC,$53,$18,$8F,$E5,$F3,$BF,$94,$CF,$C8,$6B,$E5,$25,$7A
	.dc.b	$63,$8F,$C8,$F2,$F8,$A9,$3E,$D8,$EB,$D7,$23,$E0,$3E,$0F,$94,$EB
	.dc.b	$1F,$0F,$0E,$AF,$8B,$3F,$F4,$DE,$DF,$CD,$16,$1D,$A7,$58,$F8,$9F
	.dc.b	$99,$F2,$7E,$CF,$94,$91,$83,$E8,$7C,$A7,$C0,$23,$F4,$FE,$53,$E1
	.dc.b	$FC,$7F,$C4,$F9,$09,$3F,$5E,$3E,$3F,$C8,$48,$FB,$4F,$CA,$F9,$0E
	.dc.b	$71,$E9,$C2,$ED,$E9,$C7,$AC,$FE,$07,$1E,$AF,$BF,$1F,$A7,$F2,$9F
	.dc.b	$53,$E5,$24,$FE,$1E,$3E,$BF,$CA,$48,$FC,$53,$F6,$BE,$53,$FC,$C7
	.dc.b	$E8,$FF,$A5,$09,$E0,$63,$B7,$3F,$27,$30,$FF,$21,$D7,$FA,$22,$BF
	.dc.b	$CA,$2F,$F9,$B6,$3A,$90,$FA,$98,$77,$78,$CF,$80,$8F,$4B,$F1,$FC
	.dc.b	$97,$93,$E3,$E4,$F8,$23,$D4,$F4,$3B,$70,$7D,$FC,$F9,$C7,$E5,$32
	.dc.b	$F5,$BD,$4F,$4F,$C7,$FC,$0F,$3E,$46,$0F,$6F,$C6,$7C,$4F,$8C,$91
	.dc.b	$83,$E0,$FC,$67,$C8,$F8,$C9,$18,$3E,$2F,$95,$F3,$3E,$3E,$46,$0E
	.dc.b	$47,$96,$FA,$BE,$7C,$3E,$9E,$1D,$23,$D5,$86,$1F,$A3,$F1,$FF,$63
	.dc.b	$E3,$E4,$60,$FA,$BF,$1D,$F7,$08,$FD,$8F,$8F,$FB,$FE,$B7,$E0,$3E
	.dc.b	$2E,$4F,$DC,$8F,$C1,$FC,$5C,$8F,$BE,$FF,$AA,$3F,$11,$7C,$77,$49
	.dc.b	$F9,$D8,$DE,$6F,$F1,$1F,$2B,$3E,$F5,$F8,$8C,$FB,$BF,$57,$3E,$9E
	.dc.b	$B7,$5F,$8F,$93,$FE,$90,$D1,$3F,$EB,$BF,$9F,$7E,$F8,$F8,$EC,$36
	.dc.b	$FF,$D3,$F3,$A9,$CF,$C4,$FF,$B0,$FF,$0F,$FE,$77,$D1,$7A,$9D,$14
	.dc.b	$C4,$74,$F9,$19,$97,$4C,$FF,$D8,$6D,$E2,$F0,$E7,$23,$4F,$81,$FF
	.dc.b	$58,$E5,$E0,$6E,$5E,$06,$E5,$E0,$7E,$07,$C1,$7C,$F2,$3F,$0F,$E4
	.dc.b	$BB,$3F,$E5,$ED,$BD,$CF,$80,$F3,$6D,$CD,$FF,$31,$E3,$FD,$6B,$0F
	.dc.b	$74,$78,$B6,$FB,$9F,$25,$23,$E8,$3F,$7F,$7F,$C0,$6D,$F8,$1D,$A5
	.dc.b	$7D,$58,$FC,$26,$5F,$85,$F9,$09,$3F,$66,$3F,$11,$F2,$52,$3E,$E3
	.dc.b	$F8,$BF,$92,$EB,$1F,$7F,$0E,$AF,$F4,$CE,$4F,$F2,$B7,$91,$FE,$DE
	.dc.b	$7F,$CD,$B1,$F8,$5C,$7B,$BB,$C7,$F2,$DF,$0F,$F9,$7F,$92,$93,$F8
	.dc.b	$B8,$FC,$D7,$C8,$C8,$FC,$73,$F9,$DF,$8C,$FC,$F7,$C7,$7E,$7F,$E3
	.dc.b	$E5,$7E,$4E,$3F,$45,$F0,$F2,$3F,$2C,$FE,$97,$E3,$3A,$47,$A7,$0B
	.dc.b	$E1,$7F,$A1,$B6,$FF,$27,$7C,$29,$63,$DA,$FE,$06,$5E,$27,$FD,$E3
	.dc.b	$F8,$9F,$F7,$41,$F2,$C8,$CB,$5F,$EE,$71,$BF,$88,$C6,$5F,$E9,$68
	.dc.b	$7C,$46,$39,$39,$68,$9C,$E1,$E0,$C3,$4A,$75,$F1,$18,$6C,$79,$A3
	.dc.b	$14,$F8,$6B,$F0,$B2,$BD,$6E,$BE,$17,$78,$F0,$CE,$57,$84,$E8,$43
	.dc.b	$DB,$8F,$6C,$3B,$FC,$47,$83,$E2,$25,$7C,$88,$EA,$FC,$AB,$EA,$FC
	.dc.b	$6C,$3C,$F8,$48,$ED,$B7,$C7,$C7,$D7,$F0,$FE,$CF,$87,$95,$F5,$23
	.dc.b	$DF,$F0,$B2,$3E,$BB,$F0,$BE,$17,$E1,$FC,$24,$9F,$B6,$FC,$6F,$85
	.dc.b	$F8,$FF,$0F,$F2,$3E,$1E,$57,$E0,$23,$A3,$F7,$30,$BA,$7D,$BC,$7A
	.dc.b	$4F,$E0,$31,$B3,$F7,$F1,$FA,$7F,$0D,$F5,$3E,$1E,$5F,$E3,$23,$EB
	.dc.b	$FC,$34,$8F,$C7,$BF,$6B,$E0,$FE,$F7,$E2,$9F,$EB,$CC,$7F,$DB,$EF
	.dc.b	$F9,$32,$BF,$25,$18,$F8,$7D,$F6,$70,$F8,$89,$5A,$8F,$C7,$DF,$F8
	.dc.b	$26,$3F,$18,$FF,$82,$5F,$CC,$E3,$E3,$69,$FA,$2C,$7A,$BC,$CE,$38
	.dc.b	$79,$8C,$BF,$E5,$4A,$7C,$3F,$AB,$E2,$24,$F1,$EF,$56,$A3,$FD,$1D
	.dc.b	$E2,$FF,$A2,$3F,$3E,$71,$53,$FD,$3B,$12,$79,$53,$E2,$7E,$39,$98
	.dc.b	$FF,$42,$CD,$E0,$FF,$54,$37,$7F,$F5,$2B,$67,$FD,$4A,$D7,$FD,$4A
	.dc.b	$DF,$8D,$3B,$CF,$D3,$F1,$5F,$5B,$E2,$65,$F9,$23,$85,$75,$FA,$27
	.dc.b	$D8,$38,$33,$7F,$00,$F5,$FB,$3F,$80,$F6,$A4,$7B,$9F,$C1,$FC,$5F
	.dc.b	$E1,$3E,$03,$F0,$BF,$19,$2B,$E1,$C7,$E2,$3E,$33,$F1,$3F,$19,$27
	.dc.b	$E3,$C7,$E3,$3E,$2A,$47,$CA,$7F,$1F,$F1,$5D,$63,$E6,$E1,$D5,$FB
	.dc.b	$98,$75,$7E,$8E,$1D,$5F,$C0,$61,$F9,$8F,$89,$FC,$CF,$C5,$49,$FA
	.dc.b	$F1,$F9,$CF,$8C,$91,$F6,$9F,$E3,$9C,$7F,$E1,$E8,$FC,$33,$FE,$12
	.dc.b	$6F,$C5,$E8,$DF,$C5,$F0,$F8,$A9,$3C,$1E,$7F,$15,$DA,$3F,$29,$85
	.dc.b	$EC,$76,$7F,$32,$CB,$C1,$5F,$EA,$C7,$C0,$DC,$BC,$0F,$93,$2F,$3F
	.dc.b	$C4,$7A,$3E,$22,$4F,$A2,$3D,$7F,$11,$23,$D6,$FB,$BE,$1B,$DF,$F0
	.dc.b	$B2,$7D,$EF,$C2,$F8,$6F,$87,$F1,$12,$7E,$1B,$F1,$BE,$23,$A4,$79
	.dc.b	$F0,$BB,$FB,$23,$FA,$46,$DF,$E3,$57,$D9,$8F,$76,$9F,$3B,$1F,$A3
	.dc.b	$F0,$FF,$4B,$E2,$24,$FD,$28,$EB,$C3,$F1,$F2,$B3,$DB,$5E,$99,$F8
	.dc.b	$7C,$56,$9F,$E0,$8A,$EE,$7C,$DC,$FE,$FF,$C3,$7E,$03,$E2,$24,$FE
	.dc.b	$02,$3F,$07,$F0,$DF,$84,$F8,$69,$3F,$84,$8F,$C3,$FC,$47,$E2,$3C
	.dc.b	$99,$3F,$88,$8F,$EC,$DF,$8B,$7F,$BB,$0F,$EF,$31,$FC,$45,$AF,$FB
	.dc.b	$A9,$FF,$78,$E7,$FE,$04,$8F,$C8,$E1,$E0,$7F,$09,$87,$81,$FC,$AE
	.dc.b	$1E,$07,$F1,$18,$78,$3E,$07,$C3,$F0,$52,$7C,$11,$E4,$F8,$29,$1E
	.dc.b	$37,$CF,$F0,$5E,$8F,$86,$F4,$FC,$34,$AF,$44,$7B,$3E,$16,$47,$AD
	.dc.b	$F7,$FC,$37,$48,$F6,$61,$79,$7E,$A7,$0F,$F8,$16,$79,$9F,$0F,$1E
	.dc.b	$CF,$C3,$C7,$E4,$FC,$2F,$CA,$F8,$69,$3F,$26,$3F,$D0,$3C,$B2,$3F
	.dc.b	$CE,$CF,$FA,$13,$8B,$FC,$FD,$E6,$7F,$9F,$FC,$A9,$5F,$E7,$28,$FF
	.dc.b	$44,$F7,$48,$FF,$3F,$3F,$E8,$CE,$8E,$91,$FE,$6F,$F8,$C6,$38,$DF
	.dc.b	$F3,$DF,$19,$FE,$94,$CF,$BD,$FF,$9F,$B3,$EE,$FF,$9C,$BC,$E6,$9F
	.dc.b	$84,$F8,$69,$3F,$83,$7F,$0F,$F0,$DD,$7C,$2F,$F1,$F6,$5F,$EA,$0D
	.dc.b	$31,$68,$F0,$3E,$B3,$C0,$43,$C1,$87,$E4,$BD,$D9,$1F,$91,$7F,$2B
	.dc.b	$F0,$DF,$96,$F8,$7F,$CB,$FC,$44,$AF,$CB,$47,$E6,$BE,$1F,$F3,$7F
	.dc.b	$11,$27,$F3,$51,$F9,$EF,$87,$91,$F9,$D7,$F4,$3F,$0F,$D6,$3F,$41
	.dc.b	$87,$57,$F4,$18,$75,$7F,$49,$87,$57,$F4,$98,$7F,$93,$FF,$17,$FF
	.dc.b	$09,$6F,$FE,$58,$7F,$C2,$52,$E5,$FE,$01,$18,$6C,$F1,$F8,$99,$5F
	.dc.b	$E0,$1A,$FF,$09,$79,$47,$FC,$25,$7C,$BE,$23,$C1,$F0,$32,$F8,$47
	.dc.b	$8F,$E1,$A4,$73,$7C,$DE,$67,$68,$F2,$67,$E7,$CF,$18,$F9,$BE,$7B
	.dc.b	$D9,$8F,$5B,$D9,$FE,$6F,$CF,$FB,$B3,$E0,$3D,$1F,$0D,$F0,$FE,$12
	.dc.b	$46,$0F,$67,$C2,$7C,$7F,$84,$91,$83,$E0,$77,$7C,$B2,$3F,$0F,$AB
	.dc.b	$E7,$77,$C8,$F8,$CF,$D1,$F8,$2F,$A5,$87,$49,$F3,$F2,$7C,$9C,$73
	.dc.b	$1D,$88,$6B,$46,$5F,$6F,$2F,$B9,$F0,$9F,$77,$E1,$25,$FD,$58,$E6
	.dc.b	$F1,$C3,$C1,$C3,$F7,$B3,$F4,$CF,$5C,$6C,$FE,$0B,$1F,$C4,$7C,$1F
	.dc.b	$E2,$7E,$12,$5F,$E0,$63,$F1,$9F,$05,$23,$F0,$8F,$FA,$77,$F1,$8F
	.dc.b	$EE,$78,$7F,$4B,$7F,$B9,$B7,$E0,$B4,$6F,$E0,$F8,$7C,$2F,$1F,$85
	.dc.b	$98,$FF,$93,$F0,$4F,$78,$F8,$98,$DC,$7B,$7C,$87,$49,$E1,$95,$7E
	.dc.b	$7E,$0C,$BC,$DF,$0B,$E7,$F8,$69,$5D,$E3,$D5,$F0,$52,$3C,$4F,$B7
	.dc.b	$E0,$BD,$DF,$09,$EF,$F8,$39,$5E,$78,$F8,$5F,$0D,$23,$D4,$FC,$5F
	.dc.b	$86,$E9,$1E,$4C,$2E,$FF,$1F,$1E,$B3,$F0,$71,$EA,$FA,$B1,$F9,$DF
	.dc.b	$0D,$F3,$FD,$89,$3F,$1E,$3F,$DA,$2B,$F0,$1F,$F6,$F3,$FA,$6F,$89
	.dc.b	$7D,$C7,$EE,$3C,$9F,$F7,$24,$63,$C8,$E3,$7C,$35,$FF,$0A,$61,$CC
	.dc.b	$FB,$DE,$DC,$7F,$D1,$86,$4B,$3C,$70,$FF,$77,$36,$3F,$DE,$4D,$78
	.dc.b	$1F,$BA,$2D,$78,$23,$44,$8F,$29,$0E,$E5,$3E,$FF,$AF,$F9,$3F,$6A
	.dc.b	$47,$E2,$E3,$55,$D6,$2D,$FC,$2F,$8A,$78,$58,$9F,$39,$7C,$77,$53
	.dc.b	$E2,$BF,$41,$F1,$32,$BF,$30,$FE,$8F,$E2,$BF,$E0,$07,$C3,$FC,$1C
	.dc.b	$0F,$F8,$6B,$4F,$F1,$83,$74,$60,$AD,$BC,$BE,$27,$9F,$C5,$48,$FF
	.dc.b	$5A,$5A,$7B,$FC,$44,$9C,$BE,$2F,$8A,$F1,$C7,$93,$E4,$23,$C8,$EC
	.dc.b	$74,$C3,$B3,$FC,$63,$0E,$CF,$83,$E2,$3D,$BF,$13,$23,$07,$92,$7E
	.dc.b	$09,$1F,$3F,$C5,$FC,$4F,$89,$F8,$BF,$15,$27,$D7,$1F,$23,$E2,$64
	.dc.b	$7B,$9F,$97,$F1,$3C,$E3,$C9,$85,$E2,$F9,$78,$F5,$9F,$8B,$8F,$57
	.dc.b	$D5,$8F,$D4,$F8,$8F,$AB,$F1,$12,$7E,$5C,$7D,$8F,$85,$91,$F3,$9F
	.dc.b	$B7,$F0,$DF,$73,$E1,$64,$FD,$27,$EF,$FC,$37,$E0,$3E,$22,$4F,$D6
	.dc.b	$7F,$07,$F1,$1D,$23,$E6,$61,$76,$FC,$1E,$3D,$27,$E8,$63,$67,$F0
	.dc.b	$F8,$FE,$1A,$7F,$A5,$FE,$33,$FD,$F9,$FC,$43,$FD,$EF,$AF,$F1,$7D
	.dc.b	$68,$C4,$EA,$7C,$06,$B8,$99,$E2,$7F,$25,$0C,$FC,$55,$BC,$ED,$6B
	.dc.b	$2F,$26,$7F,$E8,$BC,$BF,$C0,$E7,$F0,$F9,$F2,$F8,$DA,$C3,$CB,$3A
	.dc.b	$F8,$5C,$57,$AB,$E2,$6E,$DE,$96,$D2,$AF,$C4,$E3,$E2,$66,$1F,$AA
	.dc.b	$CF,$C5,$3C,$3E,$22,$4E,$73,$F0,$C7,$4F,$88,$EB,$8F,$6C,$E5,$EC
	.dc.b	$F9,$B1,$E0,$71,$A7,$CD,$CF,$C3,$1F,$53,$2F,$3F,$8F,$23,$7A,$FF
	.dc.b	$0E,$6F,$C1,$FB,$3E,$0E,$46,$9F,$7F,$C1,$FC,$0F,$88,$F8,$3F,$11
	.dc.b	$2B,$9C,$7C,$4F,$86,$91,$D9,$F8,$FF,$0F,$D2,$3F,$0F,$85,$E4,$FA
	.dc.b	$D8,$F5,$9F,$2E,$3D,$5F,$C6,$E3,$F4,$3E,$1A,$47,$A9,$FA,$7F,$0F
	.dc.b	$F6,$3E,$29,$F1,$6F,$3F,$0D,$BF,$95,$3A,$9E,$1F,$BE,$FF,$45,$F9
	.dc.b	$4F,$F0,$FE,$7F,$82,$7F,$C4,$3C,$17,$73,$67,$AF,$A7,$32,$EB,$9F
	.dc.b	$F3,$56,$FA,$7E,$2C,$E4,$69,$F0,$BC,$CF,$BD,$97,$81,$B9,$78,$1B
	.dc.b	$8F,$E4,$BE,$26,$47,$DC,$7F,$2F,$F8,$DF,$F8,$8B,$BF,$FD,$27,$9F
	.dc.b	$C1,$AF,$F5,$26,$5E,$27,$FC,$B3,$BB,$7F,$85,$9B,$3F,$81,$BC,$BF
	.dc.b	$09,$F9,$0C,$FF,$C5,$B9,$FE,$F9,$B1,$FD,$25,$BD,$67,$FC,$45,$1C
	.dc.b	$7E,$2F,$CF,$F1,$92,$30,$75,$F8,$8F,$5F,$C4,$C8,$C1,$E1,$F8,$7F
	.dc.b	$79,$1F,$27,$C3,$FC,$2C,$3A,$47,$F9,$83,$16,$39,$8E,$66,$70,$EC
	.dc.b	$EF,$1F,$97,$F0,$D2,$3E,$03,$F3,$BE,$1F,$E7,$FC,$4C,$9F,$88,$FD
	.dc.b	$2F,$89,$FA,$7F,$11,$F5,$3E,$26,$57,$C9,$8F,$AF,$F0,$F2,$3E,$63
	.dc.b	$F6,$BE,$17,$A4,$7A,$70,$BE,$1B,$93,$E8,$65,$DA,$BE,$A6,$5D,$9F
	.dc.b	$81,$97,$E1,$3E,$1B,$F0,$BE,$04,$9F,$B5,$1F,$88,$F0,$3F,$13,$DB
	.dc.b	$27,$EF,$47,$F2,$2F,$C7,$3F,$C1,$44,$8C,$7C,$16,$7E,$16,$47,$F1
	.dc.b	$49,$FE,$EE,$FF,$A4,$17,$C1,$F7,$8F,$C5,$61,$DD,$FA,$D8,$77,$61
	.dc.b	$CC,$EE,$75,$F8,$5F,$27,$C2,$C8,$C1,$E8,$3A,$90,$EA,$73,$21,$E5
	.dc.b	$3A,$9E,$7F,$85,$F8,$1F,$0B,$23,$D9,$1F,$0F,$DB,$91,$EF,$7E,$37
	.dc.b	$C1,$FC,$7F,$82,$93,$F0,$DF,$95,$F0,$7F,$2F,$E1,$7E,$67,$C2,$CA
	.dc.b	$F9,$11,$D1,$F9,$18,$5E,$9F,$C0,$47,$F8,$D7,$85,$97,$68,$F9,$D9
	.dc.b	$5A,$DF,$8C,$B6,$7F,$6F,$E1,$3E,$E7,$C2,$CC,$7E,$C4,$7D,$FF,$83
	.dc.b	$91,$F6,$DF,$C1,$7C,$1F,$E0,$FE,$0F,$F0,$9F,$07,$2B,$F0,$11,$F8
	.dc.b	$7F,$85,$91,$F8,$37,$F1,$5F,$0B,$D2,$3F,$09,$85,$DB,$CF,$AF,$F7
	.dc.b	$F9,$E6,$7E,$2F,$1E,$CF,$E2,$F1,$FE,$EB,$E6,$FF,$84,$77,$34,$9F
	.dc.b	$29,$FE,$8C,$EE,$FE,$FB,$F1,$5F,$E9,$2F,$86,$BF,$16,$75,$97,$FA
	.dc.b	$2A,$BA,$FC,$84,$8C,$10,$FF,$44,$7C,$67,$F9,$A5,$FF,$46,$CF,$27
	.dc.b	$2D,$DF,$FD,$18,$DF,$27,$FD,$2A,$D9,$E2,$75,$9F,$B3,$A7,$F8,$67
	.dc.b	$5F,$6F,$C3,$FC,$03,$E0,$FC,$57,$1B,$FF,$58,$61,$CF,$E6,$DF,$9E
	.dc.b	$BE,$0B,$5A,$F9,$EF,$67,$F8,$2B,$37,$C5,$FC,$CF,$8A,$93,$97,$E7
	.dc.b	$FC,$5F,$D0,$E2,$FA,$3F,$19,$2B,$94,$7D,$4F,$8C,$FA,$BF,$17,$27
	.dc.b	$B4,$7D,$8F,$82,$91,$E1,$7E,$DF,$C5,$75,$8F,$26,$1D,$57,$2F,$C4
	.dc.b	$C7,$B4,$7A,$71,$EC,$EB,$9B,$FC,$A3,$B7,$F8,$7E,$30,$7F,$A7,$57
	.dc.b	$C5,$7E,$2B,$E1,$24,$FC,$38,$FC,$6F,$C2,$48,$F8,$CF,$E4,$3E,$2F
	.dc.b	$F2,$3C,$9F,$92,$F8,$B9,$5F,$2E,$3F,$2B,$F1,$52,$3E,$73,$FE,$4C
	.dc.b	$FC,$7B,$FB,$B4,$78,$F0,$EC,$F9,$BB,$7F,$37,$1D,$E3,$A1,$F5,$F0
	.dc.b	$EE,$EF,$4F,$BC,$FF,$91,$2F,$FB,$89,$FD,$C1,$E8,$DE,$0F,$E0,$AE
	.dc.b	$BE,$1F,$C8,$49,$E5,$F0,$DE,$7F,$85,$91,$D5,$F5,$7C,$37,$AF,$E2
	.dc.b	$24,$F8,$5F,$76,$5C,$E3,$C3,$85,$CB,$C3,$8F,$49,$F2,$E3,$67,$CB
	.dc.b	$8F,$C7,$F8,$7F,$91,$F0,$F2,$FD,$D1,$CF,$97,$FD,$1F,$AB,$3C,$57
	.dc.b	$FB,$6E,$3F,$C0,$2F,$A4,$87,$80,$F9,$18,$7D,$5F,$85,$FA,$DF,$0B
	.dc.b	$27,$E6,$47,$D9,$F8,$4F,$B5,$F0,$B2,$7E,$84,$7D,$DF,$87,$FB,$DF
	.dc.b	$0F,$27,$EA,$47,$E0,$7E,$1A,$47,$D7,$7F,$09,$F0,$FD,$63,$ED,$61
	.dc.b	$D5,$FB,$38,$75,$7E,$F6,$1D,$5F,$BB,$87,$E3,$7E,$12,$47,$E0,$DF
	.dc.b	$C8,$7C,$2F,$FA,$73,$F1,$FF,$F5,$D4,$FE,$3D,$BC,$9F,$D2,$1B,$C5
	.dc.b	$FE,$BA,$7F,$8F,$75,$9F,$EF,$B3,$BF,$87,$E8,$75,$F8,$8E,$DF,$11
	.dc.b	$31,$FF,$4B,$DA,$3C,$5F,$11,$E3,$98,$68,$F3,$7A,$3C,$DE,$38,$76
	.dc.b	$79,$E1,$EB,$F8,$79,$1D,$9F,$77,$C4,$7F,$9F,$AA,$4F,$FA,$0B,$E2
	.dc.b	$BF,$C2,$6D,$CB,$FC,$FD,$E3,$7F,$9F,$FC,$69,$3F,$E7,$28,$FF,$43
	.dc.b	$E7,$23,$93,$FE,$8A,$CF,$AC,$7F,$9B,$30,$EA,$DC,$7F,$D1,$F8,$F6
	.dc.b	$8F,$F3,$EE,$3D,$9F,$E2,$BF,$1A,$F3,$99,$FD,$7F,$88,$95,$F2,$9F
	.dc.b	$B5,$F1,$1E,$0E,$2F,$F0,$2F,$C6,$7F,$9E,$F4,$C7,$2B,$04,$FF,$9F
	.dc.b	$F5,$F1,$47,$83,$3F,$E1,$2D,$D7,$38,$FE,$0F,$E2,$7F,$10,$49,$FB
	.dc.b	$7F,$13,$F8,$BF,$8A,$FC,$67,$C5,$C9,$FB,$F1,$F8,$FF,$8A,$FC,$87
	.dc.b	$C5,$C9,$FC,$1C,$7E,$4F,$E2,$24,$7E,$19,$FC,$B7,$C5,$F5,$8F,$C4
	.dc.b	$E1,$D5,$FC,$C6,$1D,$5F,$C6,$E1,$D5,$FC,$E6,$1F,$9F,$F8,$79,$1F
	.dc.b	$92,$7F,$45,$F1,$1F,$A3,$F8,$B9,$3F,$96,$7F,$4D,$F1,$7F,$26,$3F
	.dc.b	$9A,$7E,$41,$FE,$0F,$1F,$17,$9E,$1F,$F1,$7C,$62,$7F,$DE,$AE,$FE
	.dc.b	$27,$5C,$FF,$4F,$C4,$4A,$73,$49,$CC,$61,$DF,$C4,$79,$A7,$F7,$19
	.dc.b	$DC,$7F,$71,$9E,$F1,$F3,$67,$4C,$67,$9F,$89,$D7,$C3,$4C,$BC,$F1
	.dc.b	$C7,$E1,$B9,$7C,$34,$9F,$5C,$75,$F8,$39,$1E,$E7,$C1,$F0,$9E,$1F
	.dc.b	$86,$93,$F0,$9F,$27,$C2,$F5,$8F,$8B,$87,$57,$E6,$E1,$D1,$F9,$F8
	.dc.b	$5D,$FE,$6E,$3E,$DE,$FF,$77,$C3,$4B,$F9,$F1,$D5,$FA,$F8,$F5,$BF
	.dc.b	$AB,$8F,$3D,$BE,$97,$07,$F9,$06,$FD,$47,$D8,$CB,$E5,$7C,$2F,$CB
	.dc.b	$F8,$69,$3F,$72,$3E,$77,$C2,$C8,$FB,$EF,$D1,$F2,$7E,$97,$89,$27
	.dc.b	$F0,$6F,$D5,$F8,$5F,$AD,$F0,$DF,$77,$8B,$FC,$07,$D7,$FE,$53,$F9
	.dc.b	$4C,$E1,$D0,$87,$E4,$B2,$EC,$F4,$21,$F9,$6C,$7B,$3F,$C0,$78,$7F
	.dc.b	$D0,$78,$7F,$BD,$0F,$F0,$17,$07,$E4,$DB,$3F,$91,$75,$F0,$3E,$52
	.dc.b	$47,$39,$E1,$F0,$BE,$9F,$85,$F5,$7C,$14,$AD,$1C,$E3,$DD,$F0,$F2
	.dc.b	$33,$B7,$F9,$C6,$3E,$17,$C3,$F6,$9F,$F1,$16,$7F,$DE,$DD,$B5,$CE
	.dc.b	$3E,$17,$38,$F8,$5E,$38,$FC,$DF,$86,$91,$C9,$FA,$1F,$0F,$FB,$08
	.dc.b	$7C,$47,$F5,$CD,$7E,$27,$C3,$7D,$53,$51,$99,$FE,$58,$DD,$BF,$D1
	.dc.b	$5F,$0F,$F6,$CD,$FB,$52,$BE,$19,$D9,$C3,$FE,$8A,$6F,$17,$FA,$FA
	.dc.b	$34,$70,$BF,$03,$83,$81,$99,$E3,$8F,$F1,$79,$F2,$6E,$F8,$4F,$C7
	.dc.b	$7C,$2C,$C3,$E5,$C7,$07,$D7,$85,$CF,$FD,$49,$F0,$99,$9F,$06,$7E
	.dc.b	$5C,$BA,$FC,$2F,$E6,$7E,$1E,$57,$D6,$7F,$39,$F0,$FF,$9D,$F8,$7F
	.dc.b	$CF,$7C,$3C,$AF,$B7,$1F,$C7,$7F,$21,$FF,$07,$C7,$F2,$17,$FC,$1E
	.dc.b	$47,$5A,$C7,$F1,$31,$F4,$4F,$ED,$F9,$FE,$7A,$BB,$4F,$E2,$AF,$C4
	.dc.b	$F0,$3C,$1F,$09,$1C,$1D,$3D,$EF,$27,$C3,$C8,$C1,$E0,$F8,$7F,$41
	.dc.b	$1F,$1F,$C3,$FA,$FE,$17,$D9,$F0,$B2,$7C,$F1,$EF,$F8,$39,$1E,$A7
	.dc.b	$E1,$7C,$1F,$38,$FF,$41,$6E,$C7,$1B,$FF,$BF,$21,$97,$7A,$F8,$79
	.dc.b	$77,$78,$E5,$F3,$7E,$03,$E7,$7C,$0C,$9F,$93,$1F,$47,$E0,$24,$7C
	.dc.b	$C7,$EA,$7C,$0F,$D5,$F8,$39,$3F,$41,$FB,$1F,$07,$F6,$7E,$0E,$4F
	.dc.b	$D4,$7E,$E7,$C1,$F4,$8F,$5E,$17,$7F,$56,$3D,$27,$DF,$8D,$9F,$76
	.dc.b	$3F,$85,$F8,$1F,$C3,$7C,$0C,$BF,$C0,$C7,$3F,$13,$FB,$67,$13,$3C
	.dc.b	$57,$E2,$F3,$E9,$5A,$CE,$CF,$E3,$F2,$FE,$DD,$F4,$DF,$B1,$D1,$7F
	.dc.b	$DD,$E3,$FB,$E5,$7F,$AD,$5B,$3A,$8F,$EF,$11,$C6,$78,$7C,$27,$4F
	.dc.b	$84,$95,$FC,$82,$BF,$C0,$0D,$F8,$2F,$0F,$C1,$C8,$C1,$BF,$7B,$CA
	.dc.b	$47,$97,$BF,$E8,$C7,$AC,$79,$71,$C3,$C8,$F4,$5F,$57,$D5,$87,$C0
	.dc.b	$F2,$E4,$79,$5F,$87,$EF,$7C,$4F,$83,$93,$E9,$7E,$3F,$C1,$FC,$8F
	.dc.b	$82,$F9,$3F,$07,$2B,$DD,$1F,$33,$E0,$64,$7C,$17,$E7,$FC,$17,$48
	.dc.b	$F9,$98,$5E,$3F,$99,$E8,$38,$D9,$F7,$BF,$95,$9F,$77,$EA,$67,$F6
	.dc.b	$BD,$E9,$1F,$3D,$FB,$BF,$05,$FE,$2E,$F8,$57,$C4,$7E,$AA,$BF,$D0
	.dc.b	$9E,$36,$63,$11,$FE,$7B,$F1,$7F,$C6,$BF,$11,$FE,$75,$F8,$5B,$9B
	.dc.b	$E1,$69,$88,$E5,$E6,$CC,$3F,$D2,$0F,$5F,$2F,$B6,$72,$7C,$27,$FA
	.dc.b	$57,$2E,$4F,$F6,$AC,$BF,$D0,$98,$F6,$8C,$64,$CB,$F3,$3F,$09,$27
	.dc.b	$F1,$CF,$E7,$23,$B4,$7F,$A5,$71,$BC,$9F,$E8,$8D,$59,$F8,$2B,$FD
	.dc.b	$09,$97,$77,$FD,$0D,$87,$E9,$A2,$47,$E6,$DF,$F2,$B7,$E4,$5F,$E1
	.dc.b	$A1,$F2,$F8,$F9,$79,$5F,$E0,$09,$CB,$7E,$1F,$8F,$C2,$C8,$C1,$BF
	.dc.b	$84,$EA,$47,$97,$C2,$F8,$31,$EB,$0D,$BF,$13,$8F,$F8,$73,$AC,$F8
	.dc.b	$27,$5E,$3F,$F9,$CE,$3F,$C1,$95,$C3,$2F,$15,$68,$FF,$3A,$4E,$AB
	.dc.b	$87,$C3,$FC,$02,$4F,$B7,$E1,$FE,$1F,$C3,$C8,$F8,$0F,$C6,$F8,$7F
	.dc.b	$8F,$F0,$FF,$23,$E1,$E5,$7C,$58,$F9,$7F,$0D,$23,$E4,$3F,$3B,$E1
	.dc.b	$FA,$47,$97,$7F,$6E,$38,$DF,$E0,$E5,$DA,$BE,$86,$5D,$9E,$1B,$33
	.dc.b	$FB,$3F,$0F,$F6,$BE,$1A,$57,$D7,$8F,$BB,$F0,$DF,$7B,$E1,$A4,$FD
	.dc.b	$B8,$FC,$0F,$C2,$C8,$FB,$CF,$E1,$3E,$17,$FD,$35,$F0,$3F,$DF,$B8
	.dc.b	$0D,$27,$C2,$F6,$BF,$C3,$63,$D9,$F3,$E3,$D5,$F4,$E3,$78,$3E,$FE
	.dc.b	$5D,$BE,$1B,$BF,$C5,$CB,$FF,$27,$E7,$FE,$37,$E0,$65,$D7,$1F,$C3
	.dc.b	$E7,$E1,$BE,$A7,$A8,$87,$C8,$CB,$D7,$F0,$FE,$CF,$26,$4F,$68,$F7
	.dc.b	$F9,$32,$3C,$2F,$C2,$F1,$BE,$1F,$8B,$27,$CA,$FC,$6D,$7E,$3F,$C3
	.dc.b	$FC,$8F,$87,$95,$EA,$8E,$8F,$E0,$67,$FC,$0C,$F6,$3D,$F7,$D5,$D5
	.dc.b	$DD,$7F,$96,$D7,$D2,$9F,$8B,$F0,$DF,$5B,$E1,$E4,$60,$F9,$3F,$05
	.dc.b	$F6,$88,$FC,$CF,$82,$FB,$BF,$05,$F7,$BE,$0A,$4F,$D0,$8F,$C0,$FC
	.dc.b	$3C,$8F,$A6,$FE,$13,$E1,$F9,$C7,$AF,$0B,$A7,$9B,$1E,$B3,$F6,$B1
	.dc.b	$EA,$FC,$1C,$7F,$1B,$F0,$D2,$3E,$FB,$F9,$0F,$87,$FF,$3C,$F5,$7F
	.dc.b	$BA,$FF,$23,$FE,$ED,$5F,$EE,$F6,$BF,$E8,$28,$DB,$FE,$96,$8F,$EF
	.dc.b	$FE,$57,$FB,$E5,$B1,$FE,$82,$9D,$BF,$C1,$CE,$47,$F9,$E5,$CE,$7F
	.dc.b	$E6,$96,$D7,$F9,$DD,$B3,$CC,$E0,$43,$A7,$C4,$FF,$97,$A7,$C5,$F0
	.dc.b	$DE,$BF,$86,$95,$E7,$8E,$6F,$88,$E3,$7F,$07,$E1,$73,$87,$39,$F0
	.dc.b	$1F,$C2,$F2,$7C,$37,$C7,$F8,$89,$3F,$0D,$F9,$5F,$11,$F2,$FE,$23
	.dc.b	$E6,$7C,$44,$AF,$91,$1F,$3F,$E1,$BE,$87,$C2,$C9,$F9,$91,$F4,$FE
	.dc.b	$02,$47,$CF,$7E,$B7,$C2,$F5,$8F,$A5,$87,$57,$CD,$87,$57,$EB,$61
	.dc.b	$D5,$F5,$61,$F7,$FE,$23,$F0,$1F,$11,$27,$EE,$47,$E0,$FE,$1E,$47
	.dc.b	$DF,$7F,$0D,$F0,$FF,$87,$F8,$5F,$C4,$7C,$24,$AF,$C2,$47,$E2,$FE
	.dc.b	$12,$47,$E1,$D3,$FE,$DE,$FC,$3F,$FD,$FB,$C8,$C7,$BC,$FC,$F7,$BB
	.dc.b	$D0,$FC,$86,$1D,$DF,$A9,$87,$47,$FA,$05,$FF,$5B,$EB,$FF,$13,$BF
	.dc.b	$D7,$C9,$F0,$3B,$8C,$FC,$25,$ED,$F2,$FA,$EB,$FE,$B9,$82,$31,$5B
	.dc.b	$8C,$4B,$E0,$B1,$5C,$7E,$12,$C9,$A6,$3E,$13,$3F,$09,$30,$F8,$8F
	.dc.b	$0F,$84,$E3,$94,$9F,$90,$F4,$CB,$9C,$7C,$BC,$2B,$F0,$B0,$F0,$FC
	.dc.b	$17,$8B,$E0,$A4,$FD,$18,$E8,$FC,$8C,$2E,$DF,$23,$1E,$6F,$CA,$C6
	.dc.b	$ED,$F1,$B5,$67,$EE,$F8,$1F,$7F,$C1,$4C,$FE,$F4,$7C,$2F,$84,$F8
	.dc.b	$7F,$09,$27,$F0,$51,$F1,$BE,$0F,$E3,$FC,$24,$9F,$C3,$47,$CA,$F8
	.dc.b	$09,$1F,$89,$7E,$6F,$C0,$75,$8F,$C6,$61,$D5,$FC,$1E,$1D,$5F,$C8
	.dc.b	$61,$D5,$FC,$3E,$1F,$5B,$E0,$E4,$7E,$55,$FB,$3F,$03,$F7,$3F,$0E
	.dc.b	$FE,$EC,$49,$C7,$C2,$E7,$E1,$75,$E8,$49,$FC,$FF,$2F,$F9,$E2,$D5
	.dc.b	$CF,$E0,$E5,$6B,$97,$F2,$34,$9F,$07,$C2,$78,$6A,$1B,$3F,$29,$3E
	.dc.b	$4C,$BC,$51,$9E,$E4,$FF,$1B,$6E,$7A,$9F,$F1,$34,$FF,$66,$6F,$C2
	.dc.b	$7B,$CF,$81,$F0,$5F,$E8,$1F,$84,$93,$E1,$F8,$9F,$F4,$3C,$7F,$A1
	.dc.b	$7E,$0B,$FD,$07,$F9,$F3,$E3,$6F,$8D,$30,$FF,$3C,$4E,$5E,$5F,$15
	.dc.b	$FD,$BD,$FF,$26,$63,$FE,$92,$F8,$6F,$F4,$81,$FE,$91,$F8,$8F,$0C
	.dc.b	$EA,$99,$E3,$9B,$FC,$F1,$A7,$8B,$4F,$05,$FF,$95,$FE,$2D,$A7,$F0
	.dc.b	$5E,$6F,$E2,$71,$8A,$FB,$DF,$13,$2F,$E5,$71,$7F,$A0,$9B,$B7,$7C
	.dc.b	$3F,$8D,$EF,$F0,$A7,$D7,$AF,$C3,$7D,$F8,$F8,$B5,$FF,$72,$67,$E2
	.dc.b	$E0,$F3,$57,$6E,$0F,$C7,$11,$F0,$E1,$F9,$1D,$FF,$25,$F1,$32,$7C
	.dc.b	$B1,$F9,$5F,$87,$FC,$B7,$C3,$C9,$F4,$C7,$E6,$7E,$1A,$47,$B1,$FC
	.dc.b	$E7,$C3,$F5,$8F,$7E,$1D,$5F,$C2,$E1,$D5,$F8,$78,$75,$7F,$13,$87
	.dc.b	$E9,$7E,$26,$47,$C8,$7F,$51,$F1,$3F,$A9,$F8,$99,$3F,$31,$FD,$67
	.dc.b	$C4,$FF,$28,$F4,$7F,$C2,$3D,$C7,$FC,$16,$DF,$8F,$D7,$3E,$FE,$26
	.dc.b	$63,$F9,$FB,$47,$3F,$8A,$EF,$3F,$9E,$CA,$F3,$FE,$CF,$3E,$73,$C7
	.dc.b	$2F,$11,$87,$FD,$15,$8B,$E2,$BC,$DF,$13,$27,$C1,$1E,$9F,$89,$F5
	.dc.b	$7C,$4C,$9F,$24,$7B,$7E,$1A,$47,$9D,$F8,$1F,$0F,$F0,$7E,$1A,$4F
	.dc.b	$AD,$F8,$9F,$0F,$D6,$3D,$D8,$75,$7C,$B8,$74,$7C,$F8,$5D,$9C,$5F
	.dc.b	$D7,$35,$67,$F3,$FE,$27,$E8,$7C,$4C,$C7,$E5,$47,$77,$E0,$E7,$DF
	.dc.b	$1E,$71,$FD,$8F,$06,$7D,$F2,$ED,$9D,$AD,$67,$F7,$7E,$23,$EF,$7C
	.dc.b	$4C,$9F,$B1,$1F,$81,$F8,$69,$1F,$6D,$FC,$27,$C4,$FE,$17,$E1,$A4
	.dc.b	$FD,$F7,$F1,$1F,$0F,$F8,$9F,$89,$FE,$E3,$F9,$2F,$F0,$DA,$FF,$61
	.dc.b	$37,$83,$D3,$DB,$F8,$DA,$EF,$5F,$7B,$2B,$5A,$E1,$67,$E0,$AF,$C1
	.dc.b	$E7,$6B,$59,$F8,$3C,$8F,$0F,$C6,$CB,$FE,$65,$AF,$F0,$F7,$FC,$14
	.dc.b	$FF,$83,$35,$34,$9F,$0D,$EA,$F8,$6F,$5F,$C2,$CB,$E7,$1E,$EF,$8A
	.dc.b	$91,$D9,$F8,$3F,$15,$F0,$BE,$26,$4F,$89,$F8,$B8,$75,$8F,$2E,$1D
	.dc.b	$5F,$8F,$87,$CA,$E1,$91,$EA,$7E,$6F,$C5,$7F,$88,$61,$F1,$9F,$D5
	.dc.b	$DF,$7F,$C4,$FD,$23,$51,$98,$FF,$3B,$FC,$07,$D6,$F8,$A9,$1F,$D6
	.dc.b	$CE,$71,$C0,$8E,$67,$97,$27,$E4,$A3,$66,$67,$6F,$63,$04,$30,$48
	.dc.b	$E5,$3D,$3B,$7F,$C8,$3C,$2F,$87,$FC,$59,$2F,$ED,$FC,$47,$6C,$38
	.dc.b	$E3,$70,$E3,$F0,$F6,$FF,$CB,$F9,$75,$6D,$7E,$5B,$E2,$3F,$2F,$F1
	.dc.b	$5F,$98,$F8,$A9,$5F,$88,$8F,$CD,$FC,$57,$E7,$3E,$2A,$4F,$E3,$23
	.dc.b	$FC,$9F,$F9,$37,$F9,$86,$1A,$FE,$7A,$1F,$13,$A3,$73,$C3,$E2,$E3
	.dc.b	$FE,$07,$9F,$F0,$AB,$7E,$2B,$C2,$FE,$6B,$1F,$0B,$FA,$CC,$79,$BC
	.dc.b	$70,$EC,$F3,$3C,$BA,$F9,$BE,$2E,$47,$78,$F4,$FC,$5C,$8F,$13,$EC
	.dc.b	$F8,$BF,$6F,$C4,$FB,$BE,$22,$57,$9E,$3E,$0F,$C2,$C8,$F5,$3F,$13
	.dc.b	$E1,$7A,$47,$C4,$E7,$FF,$3E,$F0,$7F,$87,$F8,$BE,$36,$7D,$EB,$A1
	.dc.b	$F0,$F2,$EE,$F1,$3F,$D0,$55,$FE,$03,$71,$E2,$A1,$F0,$5F,$53,$E0
	.dc.b	$E5,$7C,$D8,$FA,$FF,$05,$23,$E8,$3F,$6B,$E0,$FE,$DF,$C2,$C9,$FA
	.dc.b	$8F,$DE,$F8,$5F,$BF,$F0,$B2,$7E,$C3,$F8,$2F,$85,$E9,$1E,$4C,$D8
	.dc.b	$E3,$7F,$C2,$65,$D6,$B5,$95,$9F,$C4,$65,$F8,$CF,$83,$E9,$5F,$85
	.dc.b	$CB,$0F,$A0,$87,$7C,$7D,$4F,$A0,$87,$63,$F1,$D5,$F9,$8F,$83,$FF
	.dc.b	$53,$FE,$4F,$FE,$00,$3F,$C0,$3A,$23,$3F,$09,$AF,$86,$98,$7F,$97
	.dc.b	$6F,$FC,$0A,$3E,$1B,$9F,$C2,$C9,$D4,$76,$F8,$69,$1C,$5F,$0F,$C0
	.dc.b	$F6,$8E,$98,$76,$7E,$EE,$1D,$AF,$A4,$63,$0F,$4C,$7A,$BE,$0A,$4F
	.dc.b	$95,$F6,$FC,$17,$BB,$E0,$A4,$FA,$5F,$83,$F0,$DF,$0B,$E1,$BE,$1F
	.dc.b	$C2,$CA,$F7,$47,$C6,$F8,$69,$1F,$05,$F9,$3F,$0B,$D2,$3C,$F8,$5E
	.dc.b	$8F,$EB,$FB,$32,$ED,$5F,$27,$2E,$CF,$BB,$2F,$A7,$F0,$D2,$3E,$73
	.dc.b	$F5,$BE,$17,$FD,$C1,$F1,$2F,$8C,$77,$61,$B7,$FE,$7C,$8C,$77,$7F
	.dc.b	$A1,$3E,$0F,$FD,$0B,$F0,$32,$BF,$D0,$71,$C7,$C3,$91,$FE,$84,$7A
	.dc.b	$77,$F5,$C2,$4F,$53,$FD,$13,$87,$57,$FD,$1B,$87,$57,$FD,$01,$F1
	.dc.b	$8C,$7F,$1F,$F1,$12,$7F,$14,$FE,$4B,$E2,$3A,$C7,$FA,$37,$1B,$C9
	.dc.b	$FE,$20,$DB,$E0,$67,$DE,$BA,$E7,$DE,$BF,$CF,$B9,$F7,$9F,$CB,$7C
	.dc.b	$47,$E7,$88,$FE,$67,$E2,$4F,$FA,$DF,$9F,$FC,$2D,$E6,$63,$CA,$FF
	.dc.b	$1B,$FC,$4E,$AB,$F8,$7D,$7F,$82,$E7,$FC,$77,$F1,$79,$CF,$FC,$3B
	.dc.b	$3E,$1A,$DE,$99,$C7,$A4,$FF,$84,$4D,$57,$0F,$88,$D5,$B1,$FF,$23
	.dc.b	$C6,$1B,$1E,$7F,$87,$C5,$FA,$23,$83,$E5,$26,$5A,$E4,$DE,$72,$BE
	.dc.b	$04,$72,$F8,$89,$1F,$0D,$EB,$F1,$5D,$BE,$2A,$4F,$C7,$7C,$3F,$15
	.dc.b	$E2,$F8,$AF,$1F,$C5,$4A,$F9,$91,$E6,$F8,$A9,$1F,$3D,$F4,$FC,$4F
	.dc.b	$48,$F9,$B8,$5E,$2F,$97,$8F,$59,$FA,$D8,$F5,$7E,$9E,$3F,$0B,$E2
	.dc.b	$7E,$1F,$C4,$C9,$FB,$91,$F1,$BE,$23,$E3,$FC,$3C,$9F,$C0,$47,$CA
	.dc.b	$F8,$89,$1F,$83,$7E,$6F,$C2,$FC,$EF,$86,$93,$F8,$77,$EA,$7E,$51
	.dc.b	$FE,$01,$DB,$F0,$38,$75,$AF,$C3,$E3,$E1,$7A,$10,$FC,$5E,$3D,$9F
	.dc.b	$C9,$D7,$F9,$9A,$FF,$C1,$31,$8E,$BF,$CA,$47,$F8,$1A,$3F,$B0,$B6
	.dc.b	$78,$7C,$3F,$79,$F0,$9F,$9D,$C3,$13,$F9,$6C,$3D,$F3,$E2,$3F,$45
	.dc.b	$9F,$B3,$E2,$24,$78,$4F,$CA,$62,$8F,$83,$F0,$FF,$0B,$E2,$25,$6F
	.dc.b	$B9,$69,$F8,$FF,$09,$F2,$3E,$22,$5E,$E9,$1F,$33,$E2,$3E,$6F,$C3
	.dc.b	$7C,$EF,$88,$97,$B8,$EE,$FC,$2C,$6E,$FF,$27,$2E,$B3,$F1,$B1,$B3
	.dc.b	$F3,$71,$FB,$3F,$11,$F6,$BE,$22,$5F,$8E,$3E,$EF,$A3,$23,$CC,$FE
	.dc.b	$03,$D1,$FC,$0F,$9F,$F8,$2F,$3A,$57,$AE,$3F,$0B,$F0,$BF,$86,$F8
	.dc.b	$89,$3E,$F8,$E8,$FD,$AC,$2E,$9F,$81,$C7,$AC,$FC,$5C,$7A,$BF,$80
	.dc.b	$C7,$F2,$3F,$0F,$23,$E5,$3F,$C2,$BC,$A7,$53,$C7,$FF,$41,$F0,$FF
	.dc.b	$0E,$F9,$77,$EA,$BF,$C2,$5B,$7A,$C7,$FD,$01,$5F,$C5,$3E,$69,$F9
	.dc.b	$DF,$E5,$27,$FC,$69,$18,$F0,$38,$47,$F8,$DC,$FF,$4A,$D6,$38,$3F
	.dc.b	$CF,$ED,$F0,$7F,$96,$7C,$77,$E0,$2B,$B9,$0E,$4E,$5B,$3C,$5F,$F3
	.dc.b	$EB,$72,$7E,$C1,$E5,$C3,$FD,$CF,$18,$C7,$E2,$4E,$71,$FD,$5F,$C7
	.dc.b	$73,$C7,$CE,$78,$70,$FF,$77,$EA,$FD,$57,$FC,$9B,$39,$BF,$4E,$7E
	.dc.b	$3D,$5F,$9D,$BC,$8F,$E0,$BF,$1F,$FC,$E7,$F2,$EF,$F5,$46,$9F,$EE
	.dc.b	$26,$E3,$FB,$C9,$D7,$C8,$6F,$E3,$A6,$9F,$AB,$B4,$F3,$F8,$DE,$9F
	.dc.b	$1D,$2B,$51,$DF,$E4,$24,$71,$7C,$5F,$21,$DA,$3A,$63,$D9,$FF,$35
	.dc.b	$E3,$D9,$E6,$78,$70,$EC,$EB,$0F,$67,$C8,$7B,$7E,$42,$4F,$9E,$3E
	.dc.b	$07,$BD,$23,$D4,$FC,$3F,$8D,$F8,$9F,$19,$F1,$7E,$36,$57,$BE,$3E
	.dc.b	$47,$C8,$48,$F8,$4F,$CB,$F9,$0E,$91,$E8,$C2,$F2,$7C,$DC,$7A,$CF
	.dc.b	$C9,$C7,$AB,$EE,$C7,$EA,$7C,$7F,$D5,$F8,$F9,$3F,$3E,$3E,$C7,$C7
	.dc.b	$48,$FA,$4F,$DB,$F8,$EF,$B9,$F1,$92,$7E,$B3,$F7,$FE,$33,$FB,$C7
	.dc.b	$E5,$FF,$DD,$7D,$A3,$49,$F1,$BD,$6F,$E8,$E3,$77,$FC,$36,$5D,$67
	.dc.b	$EA,$E5,$67,$F1,$59,$76,$F8,$AE,$F8,$CB,$FE,$65,$97,$F8,$3B,$BB
	.dc.b	$FC,$39,$D2,$D3,$B4,$61,$F6,$57,$A8,$87,$80,$FC,$96,$3E,$CF,$8B
	.dc.b	$F6,$FC,$6C,$9E,$F1,$F0,$3E,$37,$E0,$FC,$6C,$9F,$1C,$7C,$4F,$43
	.dc.b	$E2,$FC,$5C,$9F,$3C,$7C,$8F,$8B,$91,$EA,$7E,$5F,$C5,$75,$8F,$6E
	.dc.b	$1D,$5F,$F6,$1E,$EC,$7B,$47,$C2,$C7,$B3,$AC,$7E,$AF,$C5,$C8,$F8
	.dc.b	$EF,$D8,$F8,$DF,$B3,$F1,$B2,$7E,$5B,$F7,$3E,$33,$EE,$FC,$6F,$DE
	.dc.b	$F8,$D9,$5F,$42,$3F,$03,$ED,$48,$FA,$6F,$E1,$3E,$37,$A4,$7A,$30
	.dc.b	$BD,$4E,$7F,$F2,$97,$03,$3E,$F7,$F7,$33,$EE,$FC,$0C,$FF,$1F,$F1
	.dc.b	$B2,$3F,$02,$FE,$4B,$D0,$FF,$43,$7B,$B2,$7F,$BE,$7C,$7F,$FA,$2E
	.dc.b	$3F,$D1,$5E,$C7,$F9,$4B,$E5,$5F,$20,$EC,$C3,$F3,$70,$FF,$02,$4E
	.dc.b	$BA,$33,$5F,$E9,$3F,$85,$95,$DB,$E4,$51,$FE,$97,$F8,$7F,$14,$FF
	.dc.b	$BF,$74,$F1,$3B,$D3,$C6,$F6,$CF,$C4,$DC,$BC,$67,$81,$FF,$08,$7C
	.dc.b	$77,$F8,$CF,$93,$FC,$F4,$D8,$FF,$36,$37,$C8,$3F,$D0,$E1,$8F,$89
	.dc.b	$F2,$68,$CB,$CF,$CD,$FE,$93,$87,$07,$86,$BF,$D3,$F9,$F9,$1F,$46
	.dc.b	$FF,$E7,$9C,$BF,$C6,$DA,$79,$3E,$53,$06,$7E,$4B,$5F,$25,$BF,$90
	.dc.b	$98,$FB,$23,$97,$C8,$73,$F8,$19,$3F,$02,$3B,$7C,$84,$8F,$86,$F8
	.dc.b	$7E,$47,$AC,$7C,$6C,$3A,$AE,$1F,$C8,$3F,$1B,$1E,$EF,$CC,$C7,$BB
	.dc.b	$BD,$3E,$6E,$5E,$DD,$64,$FD,$27,$E0,$7C,$87,$C1,$F9,$19,$3F,$59
	.dc.b	$F8,$9F,$23,$F1,$7E,$47,$E3,$7C,$84,$AF,$B7,$1F,$27,$E4,$64,$7D
	.dc.b	$E7,$E6,$76,$F4,$8F,$21,$F7,$B0,$C6,$FE,$CE,$BF,$BF,$97,$7A,$FC
	.dc.b	$3E,$5D,$DE,$33,$F8,$4C,$BE,$C7,$6F,$D9,$EB,$95,$F8,$E8,$FB,$9D
	.dc.b	$5F,$77,$E1,$E4,$FE,$4A,$3F,$D7,$1F,$98,$7F,$7E,$8F,$F5,$E3,$71
	.dc.b	$34,$FE,$E1,$1E,$25,$F0,$BC,$89,$1F,$E6,$D8,$7C,$27,$8A,$3F,$43
	.dc.b	$87,$89,$F5,$F1,$FE,$59,$E1,$CB,$E4,$23,$CF,$E2,$3C,$FF,$0D,$E8
	.dc.b	$F8,$89,$3D,$E3,$B3,$D3,$0E,$DC,$7F,$37,$1E,$D3,$E9,$C7,$AB,$4E
	.dc.b	$C7,$A7,$E2,$3E,$29,$1F,$67,$C3,$FC,$8F,$88,$91,$EF,$7E,$5F,$C3
	.dc.b	$7C,$CF,$86,$93,$F0,$DF,$9F,$F0,$BF,$43,$E1,$BE,$8F,$C4,$4A,$F9
	.dc.b	$11,$D1,$F9,$D8,$5D,$7F,$0D,$C6,$CB,$AD,$7D,$3C,$AD,$6B,$2F,$BB
	.dc.b	$F1,$1F,$7B,$E1,$E5,$FD,$58,$FC,$0F,$C4,$48,$FB,$0F,$E1,$3E,$0F
	.dc.b	$F0,$BF,$11,$F8,$6F,$88,$95,$F7,$63,$F1,$3F,$0B,$FF,$00,$7F,$31
	.dc.b	$FF,$05,$1F,$E0,$BD,$D1,$D6,$BF,$11,$8D,$D7,$D7,$97,$69,$FC,$46
	.dc.b	$5D,$9F,$C7,$65,$FE,$87,$FD,$61,$F0,$34,$93,$FE,$C1,$F9,$16,$FF
	.dc.b	$CA,$23,$FD,$88,$3E,$49,$E5,$A3,$3E,$87,$F4,$06,$F9,$79,$7F,$CF
	.dc.b	$B3,$76,$FE,$23,$1C,$E3,$1D,$9B,$B7,$8D,$8A,$ED,$18,$7C,$78,$F8
	.dc.b	$9B,$7B,$7A,$18,$21,$E2,$2E,$1E,$29,$E5,$19,$24,$66,$7A,$4E,$A3
	.dc.b	$13,$E2,$F2,$F3,$52,$30,$76,$C7,$FD,$43,$19,$8F,$F5,$46,$0F,$2B
	.dc.b	$18,$74,$9F,$0D,$FC,$3F,$33,$F0,$9F,$17,$F8,$5F,$8C,$95,$FE,$F1
	.dc.b	$E9,$4F,$E2,$7E,$2F,$F1,$5E,$BC,$AD,$47,$E3,$7D,$8F,$C7,$7C,$4C
	.dc.b	$9E,$51,$F9,$1F,$88,$EF,$1D,$70,$C3,$C8,$F0,$61,$D5,$FE,$57,$8B
	.dc.b	$1E,$D1,$E5,$C7,$B3,$AC,$7F,$3D,$F1,$3F,$9F,$F8,$B9,$3E,$B8,$FD
	.dc.b	$17,$C5,$C8,$F7,$3F,$A5,$EA,$FD,$37,$C5,$FE,$9F,$E2,$65,$7C,$38
	.dc.b	$EA,$F3,$21,$E6,$F8,$8E,$8F,$A2,$EF,$27,$F4,$9D,$BF,$4B,$3C,$CF
	.dc.b	$9D,$87,$67,$DF,$87,$ED,$FE,$1E,$47,$D3,$7F,$75,$F1,$1F,$BB,$F8
	.dc.b	$B9,$3F,$5D,$FD,$F7,$C5,$FE,$FF,$E2,$A4,$FD,$B7,$F8,$17,$C5,$72
	.dc.b	$8F,$BD,$75,$FE,$B3,$F9,$9F,$F8,$77,$4F,$F8,$03,$1D,$48,$7D,$4C
	.dc.b	$3A,$BF,$5B,$0B,$AF,$BF,$1F,$D7,$7B,$8C,$7F,$C5,$04,$BC,$BF,$E2
	.dc.b	$98,$63,$E0,$9F,$C0,$E7,$EE,$7C,$07,$AC,$87,$84,$FC,$C5,$F4,$F8
	.dc.b	$7E,$BF,$0F,$27,$F8,$66,$5F,$DE,$F2,$FF,$15,$C3,$2F,$37,$C2,$F9
	.dc.b	$3E,$1A,$4F,$0F,$03,$F8,$A5,$6A,$30,$71,$B7,$C2,$FB,$3E,$1A,$4E
	.dc.b	$AB,$E1,$4F,$96,$31,$5E,$29,$F8,$DA,$7C,$4F,$87,$F8,$B9,$F8,$AF
	.dc.b	$EA,$E7,$87,$8E,$3D,$9F,$B3,$97,$CC,$8F,$9B,$E2,$C9,$EE,$FD,$0F
	.dc.b	$77,$E8,$F8,$F2,$7C,$6F,$D4,$F2,$3E,$AF,$C3,$FD,$6F,$87,$95,$E8
	.dc.b	$8E,$4F,$AA,$EB,$F8,$AF,$2B,$A3,$F8,$CC,$2F,$1F,$D1,$C7,$AC,$FC
	.dc.b	$2C,$7A,$BF,$92,$C7,$F0,$BE,$54,$8F,$8E,$FF,$39,$F8,$97,$D9,$BD
	.dc.b	$DC,$2F,$72,$92,$FF,$D2,$F0,$F7,$3F,$D2,$7E,$27,$FA,$53,$C6,$95
	.dc.b	$FE,$87,$8E,$78,$48,$E2,$F6,$F7,$7A,$47,$FA,$47,$0F,$03,$DB,$1F
	.dc.b	$1E,$3D,$A3,$FD,$25,$8F,$67,$FC,$9F,$F1,$9F,$EA,$6C,$BA,$47,$9B
	.dc.b	$1A,$FE,$D3,$D4,$EF,$C3,$FE,$6E,$D7,$E0,$E7,$8E,$EF,$ED,$11,$F1
	.dc.b	$34,$F0,$61,$DB,$E3,$3F,$88,$B7,$5C,$E3,$FA,$3F,$57,$F6,$64,$9F
	.dc.b	$CA,$FC,$5F,$ED,$BE,$2F,$F6,$FF,$0D,$27,$F3,$31,$FB,$EE,$0F,$F2
	.dc.b	$A1,$FE,$52,$C3,$FC,$D0,$DD,$7F,$52,$FF,$96,$7C,$3F,$F5,$25,$F6
	.dc.b	$21,$FE,$6A,$8F,$F2,$CB,$D8,$FD,$3E,$1E,$17,$F8,$BE,$1E,$1B,$F1
	.dc.b	$7C,$4C,$9F,$D8,$5F,$93,$E2,$BC,$C4,$8C,$FC,$4F,$A7,$E2,$A4,$70
	.dc.b	$7D,$9F,$15,$ED,$F8,$AF,$77,$54,$AE,$B1,$F0,$6E,$47,$43,$C3,$D1
	.dc.b	$D1,$F6,$61,$78,$3E,$2E,$3D,$67,$CF,$8F,$57,$E1,$63,$F3,$79,$E4
	.dc.b	$7B,$1F,$A1,$CD,$F4,$7E,$27,$E9,$7C,$54,$AF,$83,$1F,$57,$E2,$A4
	.dc.b	$7C,$47,$EC,$7C,$17,$28,$F8,$F7,$5F,$A5,$F1,$1D,$5F,$97,$87,$57
	.dc.b	$EB,$61,$D1,$FB,$18,$5D,$7F,$03,$8F,$E1,$70,$95,$F5,$1F,$C4,$7C
	.dc.b	$47,$58,$FC,$06,$3D,$6F,$F1,$38,$F3,$E5,$FC,$4E,$56,$FF,$9C,$72
	.dc.b	$7F,$93,$1F,$BF,$F0,$FF,$93,$23,$F8,$2F,$88,$FC,$B7,$C4,$C8,$FC
	.dc.b	$2B,$FD,$5B,$F3,$47,$C4,$DF,$86,$95,$FC,$E6,$7F,$C4,$6D,$F8,$8F
	.dc.b	$D0,$61,$23,$B1,$BF,$86,$EC,$FE,$6F,$1B,$BF,$FB,$2F,$0E,$F3,$D7
	.dc.b	$1E,$EF,$FB,$2A,$3B,$B7,$1F,$37,$C3,$C8,$F2,$3E,$9F,$86,$F5,$7C
	.dc.b	$3C,$9F,$43,$ED,$F8,$8F,$77,$C3,$FB,$FE,$22,$57,$B6,$3E,$17,$17
	.dc.b	$38,$F8,$18,$61,$E6,$43,$CD,$87,$57,$91,$F1,$EF,$AB,$EB,$BF,$F2
	.dc.b	$FF,$99,$FE,$6B,$FD,$9D,$ED,$53,$35,$FE,$72,$F2,$FF,$CE,$1E,$C6
	.dc.b	$A3,$FD,$07,$3F,$E6,$89,$91,$FE,$86,$F6,$76,$61,$FF,$7D,$9F,$E8
	.dc.b	$D8,$C7,$1F,$F9,$D5,$BD,$5F,$E8,$BF,$8A,$FF,$41,$57,$06,$D6,$E3
	.dc.b	$C8,$43,$55,$C9,$FF,$4C,$4F,$0A,$E9,$39,$9F,$11,$FE,$8E,$F3,$B1
	.dc.b	$5B,$8E,$33,$F0,$3E,$03,$2F,$1C,$DE,$EF,$E5,$FD,$19,$7F,$95,$7F
	.dc.b	$35,$F1,$1F,$9B,$F8,$9F,$CE,$7C,$54,$AF,$CD,$47,$E7,$FE,$2B,$F4
	.dc.b	$1F,$15,$27,$F3,$D1,$C9,$E8,$43,$F2,$97,$D5,$FD,$25,$F5,$7F,$31
	.dc.b	$7D,$5F,$D4,$5F,$57,$F3,$97,$FA,$FF,$87,$FD,$87,$C5,$49,$FD,$74
	.dc.b	$7E,$CF,$E2,$A4,$7E,$C5,$FD,$B7,$C4,$FF,$93,$3F,$36,$FF,$61,$4B
	.dc.b	$95,$CB,$FE,$36,$9F,$F2,$61,$27,$C0,$76,$21,$C6,$38,$7C,$4F,$82
	.dc.b	$3F,$65,$85,$BF,$DC,$61,$E1,$9F,$17,$CB,$C7,$FC,$7D,$8F,$92,$BC
	.dc.b	$71,$FE,$4D,$87,$2F,$A7,$8E,$4F,$08,$F6,$74,$C8,$E6,$FB,$FA,$7E
	.dc.b	$07,$C4,$C9,$EE,$FC,$3F,$8A,$F8,$97,$27,$A1,$E4,$E8,$E8,$FB,$CF
	.dc.b	$26,$18,$CF,$CF,$F1,$6D,$3F,$CB,$75,$D8,$87,$18,$D6,$5E,$18,$EC
	.dc.b	$78,$0D,$4F,$C0,$72,$62,$FC,$1C,$3F,$5D,$F7,$47,$6C,$3C,$74,$BF
	.dc.b	$81,$3D,$33,$FC,$17,$C5,$FE,$0F,$E2,$A4,$FD,$28,$FC,$37,$C5,$7E
	.dc.b	$1F,$E2,$64,$FD,$68,$FC,$57,$C5,$7E,$2F,$E2,$E4,$FD,$A8,$EA,$F3
	.dc.b	$21,$EC,$F8,$BE,$AF,$E0,$2F,$AB,$F0,$2F,$AB,$F8,$4B,$EA,$FC,$4B
	.dc.b	$FC,$D7,$8B,$23,$F1,$2F,$E7,$7C,$2F,$CF,$78,$92,$7F,$1A,$FE,$87
	.dc.b	$E1,$7F,$45,$F0,$DF,$EA,$EF,$CD,$FF,$1F,$5F,$55,$BC,$7F,$2A,$7A
	.dc.b	$4F,$62,$1F,$5B,$0E,$CF,$D8,$C2,$D7,$D2,$C3,$BC,$FF,$A6,$F0,$EB
	.dc.b	$CB,$FE,$A4,$6C,$7E,$7F,$1F,$F3,$B7,$A6,$F3,$2F,$9F,$2B,$FD,$53
	.dc.b	$F1,$4A,$D4,$61,$79,$BC,$89,$19,$F3,$8F,$FA,$CB,$E2,$DF,$16,$F2
	.dc.b	$33,$3F,$E6,$1F,$1A,$F0,$FE,$D2,$F7,$18,$E0,$E7,$1C,$AF,$D3,$B4
	.dc.b	$C3,$C8,$FB,$3D,$2F,$1C,$7E,$AF,$6F,$56,$B8,$CF,$FC,$ED,$B3,$6E
	.dc.b	$D6,$CF,$B3,$E9,$D9,$B6,$A7,$CB,$39,$D7,$C0,$F8,$88,$78,$70,$69
	.dc.b	$E1,$DB,$FD,$07,$1F,$2A,$7E,$66,$9F,$53,$3E,$D9,$FC,$79,$F1,$67
	.dc.b	$83,$39,$7E,$37,$E2,$BF,$1D,$F2,$12,$BF,$05,$1F,$91,$F8,$BF,$C9
	.dc.b	$7C,$5C,$9F,$C3,$47,$E5,$7E,$2B,$F2,$DF,$15,$27,$F1,$51,$D5,$FC
	.dc.b	$66,$1D,$5F,$CD,$61,$D5,$FC,$86,$1D,$5F,$CF,$61,$D5,$FC,$A6,$1D
	.dc.b	$5F,$D1,$61,$FA,$5F,$89,$91,$F9,$97,$F5,$1F,$15,$FA,$9F,$8B,$93
	.dc.b	$F9,$D7,$F5,$9F,$17,$FA,$DF,$8B,$FD,$77,$C3,$CA,$FD,$14,$72,$7A
	.dc.b	$10,$FD,$95,$F4,$7F,$69,$77,$A3,$F9,$7F,$6F,$FB,$22,$7F,$71,$7D
	.dc.b	$5F,$DD,$61,$E1,$7F,$61,$F1,$1F,$BF,$F8,$89,$18,$3F,$69,$F1,$7F
	.dc.b	$C0,$FE,$2E,$46,$0F,$DC,$7C,$5F,$F0,$72,$3F,$BB,$F8,$BF,$E1,$58
	.dc.b	$48,$E8,$7F,$0B,$F8,$AF,$E1,$D8,$74,$8F,$E1,$D8,$61,$EA,$43,$F8
	.dc.b	$96,$1F,$C5,$70,$E9,$1F,$C5,$70,$BA,$FC,$2C,$79,$4F,$43,$F8,$76
	.dc.b	$18,$9F,$8F,$87,$A2,$7A,$9E,$93,$F8,$A7,$6F,$F2,$32,$3F,$C1,$A7
	.dc.b	$FA,$EF,$0F,$F9,$C2,$3F,$AD,$3F,$E6,$E8,$CF,$C4,$FE,$DB,$D5,$4E
	.dc.b	$7C,$0F,$E1,$F1,$8F,$88,$D4,$BE,$2A,$93,$1F,$05,$F0,$97,$C4,$F9
	.dc.b	$27,$F9,$B6,$77,$3F,$2F,$C5,$F4,$D2,$57,$21,$A7,$92,$3C,$3A,$77
	.dc.b	$9F,$5E,$97,$0F,$7E,$5E,$6F,$8A,$F3,$F9,$32,$FA,$65,$EE,$9F,$5F
	.dc.b	$95,$EC,$F1,$E5,$6B,$74,$FC,$0F,$1E,$4E,$5F,$87,$E4,$7C,$4F,$88
	.dc.b	$93,$C5,$F8,$FF,$11,$F2,$3E,$1E,$4F,$57,$E5,$FC,$47,$CC,$F8,$79
	.dc.b	$3E,$17,$E7,$FC,$47,$D0,$F8,$8F,$A3,$E6,$CA,$F3,$47,$D4,$F3,$A4
	.dc.b	$7A,$5F,$AF,$F0,$FF,$77,$F3,$8F,$F1,$B7,$6F,$DC,$5F,$11,$93,$57
	.dc.b	$DC,$F8,$58,$F7,$7F,$13,$8F,$67,$F1,$78,$DE,$9F,$C7,$E5,$CA,$7E
	.dc.b	$7E,$3D,$9B,$8E,$5F,$F1,$64,$3E,$27,$CC,$48,$E3,$F0,$FE,$9F,$89
	.dc.b	$91,$D1,$F6,$7C,$3F,$B7,$E2,$7D,$DF,$0F,$2B,$C3,$1F,$07,$E1,$E4
	.dc.b	$79,$1F,$89,$F1,$3F,$17,$E1,$FE,$37,$C4,$4A,$F4,$C7,$C9,$F8,$69
	.dc.b	$1E,$C7,$E6,$7C,$3F,$CD,$F8,$6F,$9D,$F0,$F2,$BE,$0C,$7D,$1F,$88
	.dc.b	$91,$F1,$1F,$A9,$F1,$1F,$57,$E2,$24,$FC,$87,$EC,$7C,$47,$D9,$F8
	.dc.b	$8F,$B5,$F1,$12,$BE,$6C,$7E,$E7,$F3,$9F,$F0,$9F,$98,$69,$3E,$17
	.dc.b	$5F,$0D,$2F,$F4,$72,$8E,$5F,$0B,$CF,$E1,$A5,$69,$ED,$F0,$BD,$FE
	.dc.b	$1A,$4F,$27,$C5,$F0,$DE,$3F,$86,$93,$D9,$F3,$7C,$37,$9F,$D3,$F4
	.dc.b	$7C,$2C,$AF,$1C,$7A,$FE,$16,$47,$99,$F7,$7C,$17,$BF,$E1,$7E,$07
	.dc.b	$C2,$CA,$F5,$C7,$C3,$F8,$3F,$89,$F0,$72,$7D,$F1,$F1,$FE,$16,$47
	.dc.b	$C2,$7E,$57,$C2,$FC,$BF,$85,$93,$F1,$9F,$E4,$DF,$9D,$7F,$2B,$F4
	.dc.b	$7F,$D1,$2F,$F2,$79,$FE,$2E,$BE,$13,$67,$0F,$85,$E3,$F0,$BC,$BE
	.dc.b	$16,$5F,$F1,$4B,$FE,$58,$DF,$84,$EF,$F0,$B2,$30,$6F,$E1,$3C,$7F
	.dc.b	$0B,$23,$07,$3F,$85,$F3,$91,$ED,$F0,$BE,$AF,$85,$F5,$FC,$2C,$9F
	.dc.b	$14,$7B,$BE,$16,$47,$95,$F8,$3F,$0B,$F0,$BE,$16,$4F,$A5,$F8,$BF
	.dc.b	$0B,$F1,$BE,$0F,$E3,$FC,$24,$AF,$74,$7C,$AF,$84,$F9,$7F,$09,$27
	.dc.b	$E1,$47,$F5,$3F,$6B,$F8,$BC,$7F,$42,$6F,$9E,$69,$FC,$36,$3F,$D6
	.dc.b	$CF,$F0,$B7,$F9,$FB,$5C,$FC,$17,$5F,$82,$ED,$F0,$52,$76,$F8,$7E
	.dc.b	$0B,$C5,$F0,$32,$79,$BE,$5F,$82,$F3,$7C,$14,$9E,$EF,$A7,$E0,$BD
	.dc.b	$5F,$05,$27,$C6,$FB,$74,$F7,$7C,$17,$BF,$E0,$A5,$7A,$23,$E1,$7C
	.dc.b	$0C,$8F,$5B,$F1,$7E,$0B,$E3,$7C,$14,$9F,$7B,$F2,$7E,$0B,$FC,$E3
	.dc.b	$F1,$4F,$DF,$DE,$B6,$1B,$7F,$E9,$46,$BF,$2B,$E1,$3E,$97,$C2,$7D
	.dc.b	$3F,$83,$93,$F3,$23,$EB,$7C,$24,$8F,$9E,$FF,$82,$FF,$3C,$FE,$26
	.dc.b	$3E,$17,$1F,$09,$2B,$FB,$9C,$E5,$BF,$09,$C4,$8E,$BE,$13,$A7,$C2
	.dc.b	$48,$E2,$F7,$F8,$4F,$07,$C2,$49,$EA,$F8,$FE,$13,$C9,$F0,$9E,$5F
	.dc.b	$84,$95,$E2,$8F,$47,$1F,$A7,$E1,$24,$F9,$A3,$D9,$F0,$9F,$E9,$CF
	.dc.b	$CD,$5E,$6F,$FB,$FE,$9E,$F3,$E2,$7F,$D6,$F5,$AF,$27,$8D,$7C,$0F
	.dc.b	$85,$F8,$C4,$7E,$1F,$C3,$7C,$9F,$85,$91,$F1,$9F,$99,$F0,$DE,$1F
	.dc.b	$D7,$9C,$57,$A6,$5F,$F8,$A7,$35,$6A,$3E,$7F,$C3,$FD,$52,$3F,$4B
	.dc.b	$E1,$FE,$C7,$C3,$C8,$FA,$AF,$F9,$17,$F3,$EF,$E4,$1C,$1E,$CF,$D7
	.dc.b	$ED,$32,$46,$1B,$F1,$BF,$1E,$1F,$0B,$CF,$E2,$65,$FF,$8D,$31,$FE
	.dc.b	$55,$EC,$1F,$E5,$51,$9F,$89,$F1,$11,$E1,$F1,$3F,$EC,$9F,$DE,$DE
	.dc.b	$25,$24,$FF,$92,$FA,$9F,$25,$FE,$C1,$8F,$AF,$C1,$FE,$5E,$CB,$CB
	.dc.b	$1E,$3D,$7E,$01,$F0,$7E,$2B,$E1,$7C,$54,$9F,$43,$F1,$7E,$2B,$FD
	.dc.b	$BF,$0F,$AC,$F4,$BD,$8F,$C2,$E5,$E1,$CF,$F0,$D1,$8F,$84,$E3,$5C
	.dc.b	$27,$3C,$3F,$0F,$E2,$FE,$97,$C5,$CC,$3E,$44,$7D,$5F,$8A,$FA,$DF
	.dc.b	$17,$27,$E6,$47,$4F,$8D,$91,$FE,$6D,$3C,$CF,$A3,$F2,$CF,$F5,$A6
	.dc.b	$5E,$0E,$27,$C6,$FE,$07,$E3,$66,$1F,$5E,$3F,$09,$F1,$BF,$85,$F2
	.dc.b	$A4,$FD,$B8,$E9,$E7,$F8,$23,$E1,$1E,$03,$C5,$97,$83,$A3,$E9,$7C
	.dc.b	$6F,$E3,$C9,$3F,$85,$E4,$FC,$97,$C6,$7E,$4F,$E3,$64,$FE,$26,$3B
	.dc.b	$BE,$DE,$6F,$95,$82,$24,$61,$C6,$3E,$17,$FD,$1F,$FA,$0F,$F4,$E3
	.dc.b	$FA,$67,$0F,$D2,$6D,$FF,$A9,$1D,$7C,$77,$FA,$A5,$FE,$05,$3B,$F4
	.dc.b	$31,$3E,$4B,$F8,$38,$F8,$5F,$84,$7C,$4C,$7C,$5D,$2C,$5F,$1D,$D4
	.dc.b	$95,$FE,$9A,$87,$C7,$78,$7E,$3A,$4E,$5F,$27,$C7,$78,$75,$FC,$7C
	.dc.b	$F8,$1B,$9F,$8F,$3F,$0E,$2F,$8E,$F6,$12,$FC,$1F,$1D,$EF,$F8,$EF
	.dc.b	$81,$F1,$D2,$7C,$91,$DF,$8B,$E8,$47,$76,$E5,$DF,$9F,$E4,$7C,$6C
	.dc.b	$AF,$63,$F2,$FE,$3B,$E6,$7C,$77,$CD,$F8,$E9,$5F,$06,$3B,$3F,$73
	.dc.b	$C1,$FC,$77,$0F,$BB,$87,$D7,$F1,$DF,$54,$95,$F2,$7E,$3B,$EC,$7C
	.dc.b	$64,$8F,$98,$FD,$BF,$8B,$EB,$1E,$93,$B1,$F9,$8D,$BF,$01,$F1,$92
	.dc.b	$3E,$A3,$F8,$3F,$8C,$FF,$76,$79,$1F,$E7,$42,$71,$E3,$E3,$3A,$F2
	.dc.b	$7E,$0C,$F8,$98,$7F,$B5,$AB,$FC,$E8,$DF,$7B,$99,$D3,$D3,$EB,$F1
	.dc.b	$72,$78,$3E,$0F,$8B,$EB,$C9,$F1,$33,$FA,$7C,$7F,$8E,$C9,$F1,$7E
	.dc.b	$6F,$8B,$99,$78,$63,$D3,$F0,$72,$3C,$8F,$B3,$E2,$FA,$C7,$EB,$F2
	.dc.b	$FA,$7E,$33,$1E,$DC,$9F,$08,$F8,$7F,$15,$F1,$3E,$2E,$61,$EF,$7E
	.dc.b	$3F,$C5,$F4,$8F,$D3,$E3,$D3,$07,$C5,$FC,$CF,$8B,$95,$F1,$E3,$E7
	.dc.b	$FC,$5F,$D0,$F8,$B9,$3F,$2E,$3A,$78,$1F,$A0,$C7,$A7,$C5,$7D,$6F
	.dc.b	$8A,$97,$F4,$63,$EC,$FC,$54,$8F,$A8,$FE,$72,$FF,$9A,$FD,$0F,$F6
	.dc.b	$29,$73,$FF,$0B,$7F,$10,$47,$F7,$B1,$FD,$78,$FE,$25,$C6,$69,$9F
	.dc.b	$87,$ED,$D3,$FC,$66,$75,$18,$B4,$61,$FE,$3D,$87,$93,$E1,$FD,$F7
	.dc.b	$FB,$0D,$7E,$AE,$3F,$D7,$AF,$FA,$A6,$1E,$1F,$85,$E8,$D8,$E7,$7E
	.dc.b	$1A,$E1,$8F,$C0,$F8,$5D,$C1,$A6,$3E,$17,$C3,$7F,$A8,$3F,$8F,$74
	.dc.b	$F0,$F8,$39,$1D,$8F,$5D,$A3,$AF,$C2,$76,$E8,$95,$BD,$3E,$44,$78
	.dc.b	$BE,$13,$BC,$FC,$FC,$7C,$BF,$05,$E6,$F8,$39,$3C,$73,$FE,$BA,$DF
	.dc.b	$82,$F5,$91,$D7,$C1,$F4,$F2,$7F,$6B,$1F,$22,$9F,$07,$F0,$BE,$0E
	.dc.b	$4F,$58,$F8,$BF,$05,$23,$C0,$FC,$8F,$83,$E9,$D1,$FA,$8F,$D0,$BF
	.dc.b	$D3,$A4,$AF,$A6,$FF,$50,$7E,$FB,$FE,$9B,$E7,$FF,$02,$B7,$A3,$F0
	.dc.b	$CD,$9D,$FC,$27,$67,$F8,$C1,$F2,$63,$FA,$6B,$7E,$0F,$B9,$E0,$F8
	.dc.b	$3F,$0F,$C1,$F8,$BE,$0E,$57,$18,$E8,$FD,$C3,$A7,$C1,$7A,$08,$F7
	.dc.b	$F8,$2F,$5F,$C0,$C8,$F1,$3E,$EF,$82,$E7,$B7,$F3,$9E,$AF,$83,$F0
	.dc.b	$5F,$0B,$E0,$A6,$1E,$98,$F8,$BF,$03,$F1,$BE,$0A,$4F,$B6,$39,$F1
	.dc.b	$FF,$20,$C3,$E5,$FC,$14,$AF,$84,$FC,$EF,$82,$F9,$FF,$05,$F4,$3E
	.dc.b	$0A,$57,$C7,$8E,$6F,$F8,$22,$FE,$AF,$C1,$7D,$6F,$82,$93,$F3,$63
	.dc.b	$F5,$7F,$A1,$FF,$70,$E0,$FF,$1D,$BF,$E9,$F9,$FD,$8B,$63,$3A,$76
	.dc.b	$9F,$E1,$78,$72,$F8,$2E,$7F,$05,$27,$F4,$58,$7F,$46,$79,$FC,$17
	.dc.b	$83,$E0,$A4,$6A,$3A,$7A,$9F,$E0,$EA,$FE,$83,$3C,$BE,$0B,$CE,$7A
	.dc.b	$3E,$0B,$D3,$F0,$52,$7B,$BE,$CF,$82,$E7,$AF,$F1,$3F,$98,$7E,$05
	.dc.b	$FE,$8F,$BF,$F3,$04,$62,$F5,$39,$E1,$FF,$30,$CF,$0C,$BE,$3E,$1F
	.dc.b	$23,$E1,$26,$3E,$E8,$EE,$FF,$9E,$FE,$5B,$F8,$9F,$1F,$F2,$06,$C7
	.dc.b	$FA,$C9,$B8,$B9,$7F,$6F,$3F,$23,$E1,$7E,$AF,$C2,$C8,$C1,$FA,$63
	.dc.b	$FA,$89,$FD,$74,$C5,$7F,$9E,$0E,$9E,$6F,$F3,$36,$C7,$FA,$3E,$FB
	.dc.b	$BF,$E4,$0F,$7B,$F0,$7F,$09,$F8,$4F,$85,$93,$F6,$A3,$DA,$EC,$E5
	.dc.b	$87,$5B,$E7,$87,$F8,$E3,$0F,$C6,$FC,$24,$BF,$C1,$BF,$90,$F8,$5E
	.dc.b	$75,$B9,$FF,$32,$57,$FA,$38,$FE,$EF,$CE,$F8,$4F,$CB,$92,$7F,$1B
	.dc.b	$F0,$BC,$B8,$7F,$55,$1C,$88,$7F,$01,$3F,$37,$F0,$BF,$A0,$23,$FC
	.dc.b	$DF,$93,$FB,$37,$5F,$F6,$5F,$63,$FC,$41,$EA,$7F,$8B,$A3,$1A,$79
	.dc.b	$6F,$E7,$E3,$FD,$87,$0F,$F6,$29,$FD,$A0,$9F,$1B,$E5,$9F,$DA,$1E
	.dc.b	$4D,$BE,$03,$E5,$7F,$D5,$72,$BF,$EB,$ED,$9F,$34,$63,$E1,$BA,$47
	.dc.b	$EC,$8E,$98,$7F,$96,$A7,$FC,$3F,$5F,$90,$F8,$12,$F8,$6E,$A7,$6F
	.dc.b	$8C,$E7,$A7,$BB,$1E,$67,$A0,$FF,$63,$62,$F7,$9E,$77,$FB,$33,$EB
	.dc.b	$B1,$5F,$EB,$96,$BC,$27,$6F,$C8,$3F,$B8,$6D,$FE,$1F,$70,$7A,$CF
	.dc.b	$C7,$11,$FE,$60,$FF,$B9,$FF,$41,$FD,$49,$EB,$71,$9F,$E7,$E3,$F5
	.dc.b	$F1,$CE,$7B,$3F,$A8,$3E,$86,$3F,$88,$7B,$3F,$25,$F3,$78,$4F,$C5
	.dc.b	$DE,$9F,$F0,$B6,$CF,$F8,$16,$75,$39,$F4,$BE,$5F,$89,$F8,$7A,$EA
	.dc.b	$7C,$BC,$7F,$D3,$FF,$B5,$B9,$7F,$DE,$AD,$9F,$F7,$13,$5F,$F6,$AE
	.dc.b	$FF,$AE,$D3,$D1,$7E,$C2,$38,$24,$FF,$B3,$39,$7C,$53,$F2,$A3,$FD
	.dc.b	$79,$E4,$FD,$9F,$06,$47,$FB,$82,$3D,$2F,$E0,$87,$57,$E4,$DF,$92
	.dc.b	$7F,$B5,$A3,$FD,$EC,$47,$ED,$F2,$FE,$AF,$AB,$FC,$FD,$1F,$07,$2F
	.dc.b	$2B,$D3,$2F,$2C,$62,$BF,$CA,$C2,$FA,$57,$80,$FB,$93,$FE,$49,$7F
	.dc.b	$91,$19,$9F,$CD,$3F,$40,$D9,$E1,$3D,$07,$BC,$F3,$ED,$E6,$7F,$C6
	.dc.b	$C7,$E5,$31,$F8,$19,$FE,$E0,$F9,$44,$BF,$F4,$CE,$1E,$68,$C5,$77
	.dc.b	$3F,$CD,$F5,$FE,$36,$8C,$C7,$C1,$3E,$76,$1F,$03,$0E,$58,$74,$9F
	.dc.b	$29,$EE,$AF,$BE,$F5,$AF,$7C,$79,$67,$E5,$57,$2A,$F7,$9F,$15,$F7
	.dc.b	$5F,$8E,$7E,$9B,$EE,$7F,$45,$3B,$3E,$31,$F8,$83,$C6,$7E,$AE,$BC
	.dc.b	$6F,$CE,$24,$F3,$7F,$24,$70,$8F,$F7,$16,$1E,$58,$E1,$87,$E1,$DF
	.dc.b	$75,$FA,$9F,$C9,$5F,$37,$78,$71,$AF,$C3,$47,$5A,$FC,$23,$E5,$79
	.dc.b	$3F,$83,$AE,$53,$F3,$E3,$DB,$87,$91,$E4,$7F,$16,$AE,$71,$F6,$27
	.dc.b	$E7,$C7,$C6,$8C,$5F,$8F,$0F,$E9,$0F,$CE,$CA,$E1,$EA,$CA,$BF,$2C
	.dc.b	$97,$C9,$F8,$27,$84,$FB,$D5,$FA,$67,$E5,$9F,$54,$E3,$87,$E1,$DC
	.dc.b	$4F,$D4,$CB,$EF,$47,$E4,$DF,$94,$7C,$2A,$FE,$BD,$8F,$CB,$7F,$80
	.dc.b	$CF,$C7,$C3,$CF,$7F,$16,$3B,$9E,$43,$CE,$7E,$A0,$FD,$B5,$F6,$9E
	.dc.b	$38,$79,$6B,$91,$EF,$AF,$AC,$7B,$CF,$BA,$7C,$73,$E9,$9F,$ED,$AC
	.dc.b	$3C,$CF,$ED,$71,$FD,$C6,$5F,$6E,$3D,$07,$07,$F0,$AF,$A0,$EA,$78
	.dc.b	$0F,$45,$7E,$D3,$1F,$44,$71,$BE,$EF,$B2,$FB,$E5,$FE,$DF,$8F,$CF
	.dc.b	$E5,$F5,$5E,$B8,$7B,$B2,$F8,$07,$E4,$9F,$A9,$7F,$12,$7E,$DC,$7E
	.dc.b	$E5,$F9,$87,$9F,$2F,$4C,$7C,$73,$8C,$7C,$4C,$BC,$EF,$E0,$2F,$BB
	.dc.b	$FC,$02,$78,$47,$E6,$89,$1D,$A3,$C0,$FF,$13,$39,$1C,$63,$9B,$EC
	.dc.b	$78,$BF,$21,$FE,$13,$87,$C1,$BF,$A5,$5F,$06,$B9,$1D,$EB,$8D,$FE
	.dc.b	$52,$3E,$49,$FC,$8B,$1F,$C9,$BF,$B4,$9F,$90,$70,$C7,$E4,$47,$9E
	.dc.b	$FD,$38,$78,$2B,$CE,$FC,$87,$8C,$FE,$37,$0F,$04,$FE,$EA,$3E,$8C
	.dc.b	$E9,$FC,$71,$E1,$3D,$07,$1B,$F0,$D7,$D0,$BF,$9B,$3E,$E3,$89,$F2
	.dc.b	$8F,$E8,$33,$C3,$1F,$8D,$8F,$BA,$7E,$F3,$FB,$0A,$FA,$77,$FB,$59
	.dc.b	$FE,$D7,$97,$E4,$A7,$F2,$D8,$7E,$4A,$7F,$AB,$61,$D1,$D9,$E9,$3E
	.dc.b	$C1,$E9,$3F,$10,$ED,$FE,$D1,$5D,$71,$FE,$87,$5D,$67,$F7,$75,$BB
	.dc.b	$FE,$45,$7B,$9F,$E1,$95,$B9,$FD,$55,$6E,$FF,$BE,$5E,$EF,$FA,$2D
	.dc.b	$EE,$7F,$13,$1F,$81,$3D,$27,$C4,$3F,$7C,$7F,$0E,$CB,$EF,$3F,$60
	.dc.b	$87,$B8,$FA,$38,$7E,$45,$C6,$3F,$E7,$C7,$E0,$9D,$CF,$82,$7C,$79
	.dc.b	$EA,$FD,$83,$67,$F8,$2D,$C9,$F8,$D7,$F0,$0F,$08,$ED,$1F,$81,$7F
	.dc.b	$0D,$5F,$8D,$9E,$13,$E6,$8F,$C8,$4F,$0A,$FC,$5C,$FE,$76,$7F,$A7
	.dc.b	$1C,$70,$F9,$57,$FD,$9E,$7F,$8D,$3E,$1C,$7E,$84,$7F,$81,$67,$79
	.dc.b	$7F,$53,$8F,$5E,$1F,$A8,$8F,$C9,$3F,$1E,$FD,$59,$7D,$5A,$FE,$91
	.dc.b	$87,$47,$FB,$BE,$1D,$1F,$E8,$39,$7E,$B3,$1F,$E4,$4F,$E6,$63,$E6
	.dc.b	$9F,$6C,$FE,$54,$7D,$B8,$FE,$55,$87,$EC,$E3,$E0,$1F,$B0,$3E,$31
	.dc.b	$F3,$A3,$F4,$D9,$7C,$48,$FC,$0C,$FD,$4B,$F2,$47,$F3,$DB,$F2,$3F
	.dc.b	$D2,$CF,$DA,$4F,$D2,$25,$F4,$8F,$C5,$CF,$A7,$2F,$CF,$C7,$E1,$67
	.dc.b	$4F,$98,$95,$E1,$74,$7E,$60,$FA,$B1,$EE,$3E,$2E,$3F,$C6,$A7,$C5
	.dc.b	$87,$F0,$CC,$BE,$85,$FF,$24,$3F,$21,$7F,$5B,$2F,$E3,$AF,$8B,$2F
	.dc.b	$C6,$CF,$C7,$C3,$9B,$FD,$16,$F9,$C7,$CD,$9F,$65,$FE,$D2,$BC,$E7
	.dc.b	$D4,$3C,$E7,$F0,$C3,$F4,$D9,$7C,$27,$ED,$3F,$C4,$9F,$B4,$6C,$F3
	.dc.b	$9F,$BA,$7F,$19,$87,$9F,$2F,$C5,$61,$FE,$7E,$CB,$EB,$D7,$F5,$CB
	.dc.b	$FD,$45,$6A,$FF,$C5,$91,$EB,$3B,$9F,$C5,$0F,$2E,$1E,$B7,$CA,$FF
	.dc.b	$1A,$77,$5F,$C5,$1D,$9F,$A5,$3E,$0B,$EB,$7E,$09,$D4,$F5,$9F,$9A
	.dc.b	$7D,$6F,$DD,$7F,$CC,$38,$7E,$C6,$3A,$9F,$A6,$3A,$CF,$8D,$EB,$3E
	.dc.b	$37,$AB,$F3,$4E,$AF,$D3,$7A,$D7,$E2,$63,$F3,$51,$FC,$1A,$BE,$FC
	.dc.b	$7F,$0F,$CB,$EF,$D7,$E3,$A3,$FC,$34,$78,$8F,$E6,$E7,$C4,$BF,$E6
	.dc.b	$2F,$89,$F9,$57,$E2,$3E,$21,$E2,$8F,$BC,$FF,$09,$7E,$F4,$7E,$51
	.dc.b	$F8,$8F,$F8,$1A,$FE,$24,$75,$3E,$81,$F5,$8F,$19,$F1,$23,$C7,$3D
	.dc.b	$63,$E8,$47,$57,$E2,$3D,$6B,$E2,$47,$F9,$00,$FD,$EE,$3F,$8C,$9F
	.dc.b	$F0,$44,$7F,$20,$7F,$84,$3F,$D1,$EF,$E4,$DF,$F3,$0C,$BF,$99,$57
	.dc.b	$F3,$4C,$BF,$25,$7F,$D1,$70,$FD,$B3,$F2,$4F,$E4,$A7,$E0,$30,$FC
	.dc.b	$E3,$F3,$CF,$EB,$E6,$CF,$DE,$61,$ED,$3F,$01,$7E,$D7,$E7,$BE,$D8
	.dc.b	$F9,$EF,$A2,$7F,$CB,$D0,$20,$58,$E2,$24,$61,$C1,$C5,$C3,$86,$BC
	.dc.b	$23,$6E,$DA,$F7,$6C,$70,$8F,$1B,$5E,$07,$27,$0F,$13,$11,$C8,$8E
	.dc.b	$08,$79,$5C,$B4,$C9,$E7,$29,$FD,$BD,$FE,$9B,$D1,$FC,$BD,$FF,$3C
	.dc.b	$64,$F9,$3C,$57,$CB,$C5,$8F,$CC,$A7,$5E,$A9,$5A,$7E,$56,$39,$8C
	.dc.b	$42,$BB,$3F,$E9,$3C,$7C,$37,$BC,$3E,$3E,$16,$BE,$E5,$F0,$7F,$AB
	.dc.b	$DF,$07,$91,$E7,$AE,$2E,$0F,$2B,$E9,$C7,$A4,$78,$F0,$B9,$31,$CE
	.dc.b	$3D,$F2,$FE,$19,$97,$7C,$3C,$27,$73,$DE,$77,$8F,$7C,$F1,$CB,$E8
	.dc.b	$4E,$CF,$01,$B7,$CE,$F4,$3C,$E7,$C0,$CB,$C7,$97,$9B,$1F,$2C,$62
	.dc.b	$79,$1E,$23,$96,$1E,$28,$D4,$62,$FB,$CF,$B2,$3C,$98,$79,$DF,$ED
	.dc.b	$18,$FB,$E7,$51,$E4,$3F,$59,$87,$89,$F0,$9E,$C3,$B1,$E3,$8E,$27
	.dc.b	$8D,$DD,$7E,$29,$E3,$97,$D1,$73,$8F,$A2,$B8,$E3,$E8,$3E,$5C,$F8
	.dc.b	$EF,$D0,$FF,$19,$C7,$D1,$3A,$AF,$C8,$E1,$D6,$BC,$A7,$E4,$F0,$ED
	.dc.b	$5B,$3B,$47,$C2,$3E,$2C,$69,$C1,$F0,$E7,$A5,$79,$A3,$EC,$1E,$EC
	.dc.b	$BE,$C3,$EC,$30,$7C,$0C,$BE,$53,$E4,$CB,$E2,$3E,$6B,$EB,$8F,$96
	.dc.b	$3A,$9B,$BE,$AF,$1C,$7C,$CF,$C3,$3F,$21,$5C,$9F,$4E,$3F,$4D,$E4
	.dc.b	$78,$CF,$31,$83,$C5,$8F,$9A,$3D,$D1,$F7,$DB,$3D,$63,$C0,$7D,$5A
	.dc.b	$F1,$C7,$FA,$1B,$2F,$E2,$33,$DA,$B9,$E5,$F0,$1E,$78,$77,$CB,$E0
	.dc.b	$57,$76,$E1,$DC,$E9,$1F,$58,$F9,$A7,$A0,$ED,$1C,$AF,$B3,$CA,$BB
	.dc.b	$4F,$C4,$8F,$E3,$D9,$7C,$CC,$3C,$B9,$7A,$B2,$FF,$2B,$61,$C3,$0F
	.dc.b	$7D,$7F,$00,$BE,$18,$77,$C3,$85,$7D,$2C,$3F,$BA,$DF,$E5,$30,$E5
	.dc.b	$97,$AA,$BF,$39,$8F,$27,$C8,$7E,$02,$BE,$3C,$66,$3C,$38,$FF,$1D
	.dc.b	$7E,$0C,$FC,$A7,$E0,$9C,$8F,$94,$5A,$F9,$58,$7F,$BE,$B1,$F3,$CF
	.dc.b	$FA,$13,$1F,$2C,$FF,$BF,$B1,$F0,$5F,$9F,$0F,$C5,$1E,$4B,$FC,$51
	.dc.b	$F9,$33,$C7,$8F,$CF,$7C,$67,$A0,$FE,$7B,$5D,$2F,$C5,$1C,$0F,$36
	.dc.b	$1E,$27,$E5,$10,$F8,$E7,$82,$FE,$BB,$C4,$EC,$7F,$18,$3C,$38,$7A
	.dc.b	$70,$FB,$F1,$C5,$FB,$0F,$18,$D3,$FC,$6E,$32,$FE,$96,$7E,$73,$EF
	.dc.b	$31,$1E,$F7,$11,$F1,$5F,$CA,$18,$8F,$CA,$61,$F0,$63,$D9,$3F,$C6
	.dc.b	$30,$FE,$EB,$8F,$D4,$8F,$F2,$F6,$3F,$0E,$BF,$DD,$97,$C5,$FF,$77
	.dc.b	$56,$F0,$FF,$77,$56,$EF,$FC,$79,$7B,$7F,$C7,$D5,$BB,$FF,$54,$56
	.dc.b	$EF,$FD,$E7,$7B,$7F,$DE,$B5,$BA,$FF,$24,$56,$DF,$F6,$B6,$5F,$AE
	.dc.b	$7E,$C9,$F4,$31,$FC,$53,$F3,$CF,$CA,$47,$F3,$AC,$BE,$7C,$7E,$58
	.dc.b	$FB,$2F,$E5,$9F,$8B,$87,$F8,$B2,$BE,$2C,$7D,$77,$F6,$AF,$73,$67
	.dc.b	$F8,$C0,$FC,$39,$F4,$B2,$FC,$3B,$F6,$0F,$39,$F5,$4F,$2E,$5F,$05
	.dc.b	$F6,$E5,$F0,$70,$FF,$23,$D7,$D7,$9F,$E5,$F9,$7C,$37,$FC,$78,$7F
	.dc.b	$47,$C3,$FA,$AB,$DA,$7C,$27,$E1,$4F,$C5,$98,$C7,$F0,$B1,$F1,$CF
	.dc.b	$CD,$E5,$F1,$EB,$E5,$C7,$AE,$7F,$2B,$5C,$DF,$C9,$3C,$E3,$D6,$73
	.dc.b	$C3,$E7,$DF,$38,$F8,$E7,$A2,$7F,$8E,$3E,$C3,$33,$F0,$2B,$FB,$3C
	.dc.b	$79,$31,$FE,$7F,$8F,$F3,$E8,$FE,$CC,$FF,$9D,$2F,$FC,$87,$E0,$3E
	.dc.b	$03,$E3,$63,$CE,$FE,$B5,$78,$1F,$F0,$87,$CF,$BD,$AE,$59,$FF,$9F
	.dc.b	$B1,$E5,$5F,$E2,$D6,$BC,$27,$AE,$BF,$CC,$2B,$F7,$F3,$FE,$2E,$F9
	.dc.b	$97,$E1,$4F,$8D,$EB,$3F,$E5,$CF,$A5,$78,$38,$D5,$18,$9F,$F2,$16
	.dc.b	$1C,$AF,$BC,$7C,$AE,$2F,$D1,$76,$BF,$47,$78,$FD,$A3,$0D,$8C,$D7
	.dc.b	$F9,$37,$C5,$FF,$2A,$7C,$8F,$48,$FC,$51,$FE,$38,$E8,$F3,$CF,$F9
	.dc.b	$67,$E4,$6F,$41,$A6,$34,$E5,$BF,$E5,$F0,$FD,$1B,$FE,$79,$BF,$F3
	.dc.b	$17,$C8,$F2,$24,$F8,$F0,$7C,$8F,$2E,$4F,$D9,$4F,$8C,$D7,$C8,$F8
	.dc.b	$BE,$46,$46,$0F,$DA,$1F,$C6,$0E,$9F,$23,$E8,$23,$DF,$E4,$7B,$3F
	.dc.b	$84,$BF,$B3,$A7,$9F,$27,$99,$F0,$3C,$49,$7E,$78,$E2,$FE,$6E,$F8
	.dc.b	$9F,$EC,$7D,$BF,$D0,$1F,$21,$F2,$3E,$42,$57,$BE,$38,$F8,$3F,$95
	.dc.b	$8F,$ED,$35,$FE,$87,$F0,$BE,$8F,$D7,$FF,$B3,$75,$F9,$4D,$CB,$B6
	.dc.b	$BF,$94,$9F,$D9,$61,$FC,$77,$1E,$87,$5E,$BF,$E5,$19,$31,$7D,$07
	.dc.b	$62,$5F,$E1,$61,$F2,$DF,$E1,$3B,$FE,$61,$3E,$F3,$E4,$1F,$DB,$8E
	.dc.b	$1F,$27,$E8,$23,$CF,$E4,$B9,$75,$FF,$16,$8F,$F0,$3E,$9F,$A0,$AF
	.dc.b	$7F,$C9,$49,$F1,$3F,$0B,$E4,$7F,$C5,$75,$FA,$68,$FD,$47,$86,$BA
	.dc.b	$62,$3E,$4F,$C9,$7C,$AF,$92,$98,$FB,$63,$FC,$6B,$C1,$F0,$FC,$9F
	.dc.b	$D4,$7E,$FA,$F8,$7A,$7D,$28,$FA,$7F,$27,$30,$F8,$CF,$D6,$F9,$3E
	.dc.b	$7C,$8F,$93,$FB,$24,$AF,$99,$EC,$FD,$CF,$66,$47,$CF,$7E,$FF,$9D
	.dc.b	$CB,$E4,$FF,$02,$49,$FA,$9F,$27,$F8,$4F,$92,$91,$F5,$DF,$C3,$FC
	.dc.b	$9F,$8B,$E4,$FF,$12,$49,$FB,$9F,$0F,$FB,$CF,$D1,$FF,$DD,$3D,$5F
	.dc.b	$B5,$6F,$87,$F9,$96,$D3,$99,$0F,$97,$F9,$2B,$E5,$7A,$7C,$AC,$BF
	.dc.b	$E1,$D3,$96,$FC,$A7,$87,$E5,$64,$60,$FF,$25,$E3,$DB,$AB,$CC,$48
	.dc.b	$E7,$F2,$7E,$9F,$95,$91,$D9,$F6,$7C,$17,$93,$DF,$F7,$79,$92,$BC
	.dc.b	$71,$F0,$7E,$4F,$E1,$7C,$9C,$9F,$3C,$7B,$3E,$4B,$E3,$7C,$9C,$9F
	.dc.b	$5C,$7C,$9F,$93,$F9,$5F,$0B,$27,$DF,$1F,$13,$E0,$BE,$77,$C9,$49
	.dc.b	$F8,$71,$F4,$7E,$4B,$F9,$6F,$E9,$1F,$EE,$5E,$33,$79,$93,$F9,$68
	.dc.b	$FD,$C9,$3D,$1E,$3F,$19,$CB,$E3,$39,$FC,$7C,$CF,$F9,$F5,$A7,$97
	.dc.b	$C9,$F8,$09,$3A,$F9,$3F,$1F,$C4,$F9,$3E,$26,$4F,$28,$F1,$7C,$4F
	.dc.b	$A3,$E2,$64,$F6,$8F,$5F,$C4,$FB,$3E,$26,$4F,$8A,$3D,$5F,$13,$F0
	.dc.b	$09,$1E,$6F,$89,$F8,$7F,$13,$F1,$3E,$22,$4F,$AA,$3E,$17,$C4,$FC
	.dc.b	$82,$47,$BB,$E2,$3E,$5F,$C4,$C8,$F8,$2F,$CE,$F8,$9F,$DF,$FB,$3F
	.dc.b	$CA,$67,$F3,$8D,$EB,$FD,$C4,$7F,$60,$8F,$E4,$AF,$F6,$77,$F9,$1B
	.dc.b	$7E,$0B,$46,$FE,$1F,$AF,$C3,$71,$24,$F1,$F8,$4E,$9F,$0D,$D7,$E1
	.dc.b	$24,$F5,$8F,$2F,$C2,$F8,$7E,$12,$4F,$86,$3C,$9F,$09,$23,$C8,$F9
	.dc.b	$FE,$13,$DB,$AF,$A4,$93,$E9,$F8,$3F,$67,$C1,$C8,$F6,$3E,$FF,$83
	.dc.b	$F8,$BF,$07,$F0,$7E,$0E,$57,$C1,$8F,$89,$F0,$7F,$17,$E0,$E4,$FC
	.dc.b	$58,$F9,$BF,$07,$F2,$7E,$0E,$4F,$C9,$8F,$F0,$5F,$E1,$FF,$C3,$3A
	.dc.b	$0F,$F0,$B9,$FF,$0A,$35,$7C,$27,$F8,$26,$3F,$89,$C7,$F9,$0C,$DF
	.dc.b	$C1,$74,$F8,$29,$18,$39,$FC,$17,$80,$8F,$6F,$82,$FE,$83,$7F,$CF
	.dc.b	$E3,$C3,$F0,$3E,$63,$CF,$F0,$5E,$8F,$82,$93,$E6,$7D,$7F,$05,$CB
	.dc.b	$E0,$7D,$BF,$05,$2B,$D7,$1F,$03,$E0,$64,$7B,$9F,$87,$F0,$5E,$2F
	.dc.b	$81,$F8,$A4,$9F,$87,$F0,$5F,$23,$E0,$64,$7C,$67,$E5,$FC,$17,$AB
	.dc.b	$E0,$7E,$6F,$C1,$4A,$F9,$71,$F4,$3E,$07,$FC,$77,$F1,$7F,$C3,$CF
	.dc.b	$E2,$1C,$7F,$C3,$AB,$F9,$8C,$7F,$81,$89,$F0,$5C,$3E,$0B,$8F,$C0
	.dc.b	$72,$F8,$29,$7F,$E4,$BB,$FE,$1C,$8F,$B3,$F0,$B5,$7C,$4F,$FA,$0D
	.dc.b	$FF,$27,$9A,$38,$3E,$5F,$83,$F3,$92,$3B,$7C,$1F,$AB,$E0,$BD,$7F
	.dc.b	$07,$27,$C5,$1F,$EB,$87,$A1,$E0,$F9,$0C,$78,$6C,$BC,$CB,$ED,$1E
	.dc.b	$47,$C6,$F8,$49,$87,$B5,$F9,$3F,$0B,$F2,$BE,$13,$E5,$FC,$2C,$AF
	.dc.b	$85,$1F,$E7,$1F,$79,$ED,$7F,$9F,$72,$EB,$5E,$0D,$3C,$31,$88,$F8
	.dc.b	$5F,$1D,$FE,$99,$AF,$AF,$F0,$DF,$63,$E1,$A5,$7C,$F8,$FB,$7F,$0B
	.dc.b	$F7,$3E,$1A,$4F,$D3,$8F,$BF,$F0,$DE,$1F,$6A,$4F,$D7,$8F,$C1,$FC
	.dc.b	$37,$E1,$3E,$1A,$4F,$DB,$8F,$C3,$EF,$FE,$82,$FC,$6F,$F9,$31,$FC
	.dc.b	$9F,$B7,$FC,$A1,$C3,$FE,$5F,$8F,$E4,$F7,$F6,$FD,$64,$63,$A3,$FD
	.dc.b	$04,$DA,$E1,$F0,$DD,$7E,$1F,$B7,$C3,$CC,$7F,$D2,$D7,$FE,$3D,$6F
	.dc.b	$C3,$F8,$FE,$1E,$46,$0D,$F3,$79,$F1,$91,$83,$9F,$C3,$7A,$C8,$F6
	.dc.b	$F8,$7F,$77,$C3,$C8,$F0,$BF,$07,$E1,$FE,$17,$C3,$C9,$F2,$BF,$17
	.dc.b	$9F,$E3,$7C,$3F,$C7,$F8,$79,$5E,$A8,$F9,$5F,$0D,$23,$DA,$FC,$DF
	.dc.b	$87,$FF,$62,$FC,$83,$E1,$D3,$86,$D7,$FB,$11,$AF,$FB,$3D,$FF,$64
	.dc.b	$3F,$47,$8F,$F1,$9C,$FF,$43,$E2,$3E,$D1,$F6,$FC,$9F,$B9,$F1,$12
	.dc.b	$FE,$7B,$F7,$F0,$FF,$60,$F9,$1F,$CD,$CF,$E6,$98,$7F,$AE,$DB,$D7
	.dc.b	$FE,$9F,$75,$F1,$1C,$3E,$22,$63,$FE,$CF,$9F,$F2,$13,$F8,$55,$F1
	.dc.b	$1D,$BE,$22,$4E,$E3,$C3,$F0,$F2,$39,$3E,$4F,$88,$F2,$FC,$44,$9E
	.dc.b	$CF,$A3,$E2,$3D,$3F,$0F,$EA,$F8,$89,$5E,$38,$F6,$FC,$47,$BB,$E2
	.dc.b	$24,$F9,$E3,$E0,$FC,$44,$8F,$53,$F1,$3E,$23,$E2,$FC,$47,$C6,$F8
	.dc.b	$89,$5E,$F8,$F9,$3F,$0F,$F2,$BE,$22,$4F,$C3,$8F,$9B,$F1,$1F,$EE
	.dc.b	$EF,$D2,$7F,$C8,$47,$F9,$0F,$97,$F9,$A5,$7E,$AE,$3F,$D7,$24,$F8
	.dc.b	$8E,$1F,$11,$C7,$E2,$39,$7C,$44,$BF,$F7,$F5,$FF,$34,$79,$7C,$3F
	.dc.b	$72,$3A,$F8,$7F,$17,$C3,$78,$FE,$1E,$4F,$28,$F3,$7C,$37,$9F,$E1
	.dc.b	$A4,$F6,$8F,$57,$C3,$7A,$F7,$93,$E2,$8F,$77,$C1,$C8,$F2,$BF,$07
	.dc.b	$E1,$3E,$17,$C1,$C9,$F4,$BF,$17,$E1,$3E,$37,$C2,$7C,$7F,$83,$95
	.dc.b	$EE,$8F,$95,$F0,$9F,$2F,$E1,$24,$FC,$28,$FE,$A5,$E6,$7F,$41,$8F
	.dc.b	$F4,$E7,$E0,$5F,$E4,$12,$3A,$F3,$7F,$DF,$51,$FC,$5E,$37,$EF,$65
	.dc.b	$E3,$F0,$77,$E1,$15,$FC,$5E,$7F,$91,$3F,$E4,$86,$FC,$26,$E1,$F0
	.dc.b	$94,$C4,$67,$E1,$35,$F0,$92,$FF,$A4,$1E,$88,$E5,$F0,$9C,$FE,$12
	.dc.b	$4E,$B2,$F1,$CF,$7F,$84,$93,$9C,$D1,$E3,$F8,$4F,$27,$C2,$4A,$D3
	.dc.b	$E7,$F8,$4F,$47,$C1,$C9,$E4,$FA,$FE,$13,$D9,$F0,$52,$7B,$3E,$FF
	.dc.b	$84,$F8,$1F,$03,$F0,$7E,$0A,$57,$8E,$3E,$27,$C1,$7C,$5F,$82,$93
	.dc.b	$E7,$8F,$9D,$F8,$C7,$F9,$24,$9D,$58,$F8,$EC,$FC,$26,$BE,$12,$61
	.dc.b	$F0,$EF,$F8,$8F,$77,$E0,$79,$D1,$FC,$4A,$75,$F0,$7D,$8E,$FF,$09
	.dc.b	$E0,$F8,$4F,$0F,$C2,$4A,$E9,$1E,$4F,$84,$91,$DD,$F3,$FC,$27,$A3
	.dc.b	$E0,$E4,$F8,$DF,$5F,$C2,$7B,$3E,$0E,$4F,$9D,$F7,$FC,$27,$C0,$F8
	.dc.b	$4F,$83,$F0,$92,$BD,$91,$F1,$3E,$0F,$E2,$FC,$24,$9F,$81,$1D,$1F
	.dc.b	$EE,$9C,$2C,$71,$C7,$F0,$E1,$97,$7A,$F9,$19,$77,$78,$E5,$DD,$F9
	.dc.b	$99,$77,$7A,$E5,$F5,$BE,$0E,$47,$D1,$7E,$CF,$C2,$7D,$AF,$84,$93
	.dc.b	$F5,$5F,$D5,$6B,$FE,$E3,$1F,$09,$82,$4F,$E7,$61,$F0,$9C,$3E,$13
	.dc.b	$8F,$C2,$4A,$D4,$74,$F8,$3E,$BF,$07,$27,$94,$78,$3E,$0B,$C3,$F0
	.dc.b	$72,$7B,$47,$93,$E0,$BC,$BF,$07,$27,$C5,$1E,$8F,$82,$91,$E5,$7D
	.dc.b	$7F,$07,$EC,$F8,$3F,$6F,$C1,$CA,$F5,$47,$57,$D9,$87,$57,$FC,$55
	.dc.b	$C0,$C7,$AC,$63,$1B,$D7,$F9,$BA,$65,$DE,$BE,$36,$5D,$DE,$79,$7C
	.dc.b	$EF,$83,$91,$F2,$DF,$A3,$F0,$7F,$4B,$E0,$A4,$FC,$F7,$EA,$FC,$1F
	.dc.b	$D6,$F8,$29,$3F,$4D,$FB,$3F,$07,$D2,$3D,$38,$5D,$7C,$D8,$F4,$9F
	.dc.b	$6E,$36,$7D,$58,$F4,$9F,$83,$8D,$9F,$76,$3C,$FC,$EF,$F4,$77,$97
	.dc.b	$F9,$BC,$F1,$82,$3F,$3D,$9F,$5A,$DE,$ED,$31,$7B,$86,$9D,$EB,$7A
	.dc.b	$5A,$D6,$7F,$CF,$FE,$AF,$FD,$39,$1F,$E5,$F7,$FA,$36,$66,$9A,$D7
	.dc.b	$F8,$83,$FD,$0D,$BF,$07,$CC,$E9,$F0,$9D,$7E,$0F,$B7,$C2,$4C,$7F
	.dc.b	$9E,$60,$9F,$14,$43,$27,$97,$E0,$F9,$BC,$30,$EC,$F2,$C3,$B3,$D3
	.dc.b	$0E,$CF,$AE,$FD,$DF,$09,$EF,$F8,$39,$3E,$28,$F8,$5F,$09,$23,$CA
	.dc.b	$FC,$5F,$83,$F8,$DF,$07,$F1,$FE,$0E,$57,$AA,$3A,$BE,$CC,$3A,$BF
	.dc.b	$2F,$0E,$8F,$CD,$C2,$EF,$F2,$B1,$EB,$3F,$13,$1E,$AF,$D3,$C7,$FC
	.dc.b	$D7,$F9,$0B,$DF,$A4,$9F,$ED,$7E,$47,$FB,$5F,$DC,$DF,$BD,$27,$FB
	.dc.b	$7F,$C4,$30,$FF,$3A,$47,$03,$FC,$14,$7F,$AE,$3F,$4B,$7C,$17,$F8
	.dc.b	$17,$E2,$D1,$FE,$72,$EC,$FF,$72,$46,$3E,$0F,$53,$C2,$FB,$E1,$E7
	.dc.b	$6E,$BE,$77,$FD,$CD,$AF,$9D,$DE,$9E,$77,$FD,$D5,$A7,$9E,$3F,$C2
	.dc.b	$D3,$FE,$F3,$CF,$BE,$9E,$0C,$E6,$1F,$E0,$CD,$3C,$F3,$C2,$3F,$CF
	.dc.b	$A3,$1F,$85,$97,$8F,$87,$FC,$55,$97,$8E,$BC,$3A,$7C,$66,$E9,$E5
	.dc.b	$AD,$C7,$C8,$7F,$C6,$59,$78,$23,$C5,$AF,$C7,$51,$F1,$DC,$E7,$FB
	.dc.b	$1F,$87,$FD,$97,$C3,$C9,$FD,$2C,$7E,$D7,$E1,$FF,$6D,$F0,$F2,$7F
	.dc.b	$53,$1F,$B9,$BF,$DD,$70,$49,$FD,$6C,$75,$79,$90,$FD,$F6,$FD,$5F
	.dc.b	$D9,$DF,$57,$F8,$15,$F5,$7F,$6F,$7D,$5F,$E0,$D7,$FC,$27,$E1,$E4
	.dc.b	$7E,$F1,$FE,$19,$B7,$F9,$2F,$F4,$BF,$8D,$C7,$FC,$79,$B5,$FC,$75
	.dc.b	$BE,$09,$FA,$ED,$F8,$8E,$5F,$11,$2B,$FC,$89,$5F,$E2,$2B,$FE,$53
	.dc.b	$E1,$FD,$7E,$4F,$E2,$B9,$F9,$2F,$F8,$E6,$7E,$3B,$91,$C9,$F3,$74
	.dc.b	$79,$F3,$ED,$3F,$F0,$07,$26,$79,$8F,$0F,$C4,$7B,$7E,$22,$46,$0F
	.dc.b	$2F,$C3,$FC,$1F,$88,$91,$83,$D3,$F1,$1F,$14,$8F,$B3,$E2,$3E,$46
	.dc.b	$12,$3A,$1E,$3F,$88,$F9,$98,$74,$8F,$36,$18,$7A,$90,$F5,$61,$F4
	.dc.b	$F0,$FA,$97,$D6,$7D,$D8,$61,$F3,$1D,$0F,$9F,$7D,$5E,$67,$5E,$1F
	.dc.b	$F2,$96,$8C,$7B,$4F,$D6,$F8,$8F,$C1,$79,$F2,$30,$7D,$AF,$87,$FC
	.dc.b	$31,$1F,$BB,$F0,$DF,$89,$F8,$79,$1F,$80,$7F,$19,$F0,$DF,$8D,$C2
	.dc.b	$4F,$43,$EC,$7C,$3F,$37,$ED,$61,$74,$F3,$E3,$D2,$7E,$F6,$36,$7D
	.dc.b	$78,$FE,$67,$E1,$E5,$7E,$41,$FC,$E7,$C3,$FE,$77,$E1,$E4,$FE,$51
	.dc.b	$FF,$55,$7D,$03,$E3,$EF,$95,$8F,$89,$97,$FE,$9D,$AF,$F2,$2B,$F9
	.dc.b	$CC,$3B,$BF,$91,$C7,$BB,$D4,$87,$E5,$71,$EF,$8F,$68,$E9,$87,$65
	.dc.b	$9F,$FA,$1F,$1E,$F1,$8E,$D3,$3A,$DE,$E7,$FB,$16,$3F,$42,$D8,$CC
	.dc.b	$7F,$AE,$8F,$D2,$FB,$99,$38,$73,$7F,$84,$A3,$4D,$7C,$07,$78,$E0
	.dc.b	$EB,$DB,$E2,$61,$D9,$D8,$F0,$57,$F8,$85,$B3,$BC,$BF,$AD,$CE,$67
	.dc.b	$8B,$63,$33,$BC,$7E,$0D,$75,$37,$19,$7C,$C7,$83,$DC,$C3,$BA,$EF
	.dc.b	$18,$BE,$D7,$D1,$B5,$F3,$B1,$F9,$2F,$63,$D2,$43,$DE,$F0,$F1,$5A
	.dc.b	$3E,$23,$F1,$BF,$13,$2F,$F1,$51,$F9,$0F,$88,$FC,$8F,$C4,$C9,$FC
	.dc.b	$74,$7E,$53,$E2,$3F,$2B,$F1,$32,$7F,$25,$1F,$98,$F8,$8F,$CC,$E1
	.dc.b	$27,$F2,$D1,$D5,$FC,$C6,$1D,$5F,$B7,$9F,$9B,$1E,$D1,$F9,$DC,$7B
	.dc.b	$3A,$CD,$97,$E9,$BE,$27,$F4,$FF,$13,$2B,$F4,$91,$FA,$AF,$89,$91
	.dc.b	$FA,$77,$F5,$BF,$11,$FA,$EF,$89,$FD,$7F,$C4,$CA,$FD,$64,$75,$79
	.dc.b	$90,$F1,$74,$F4,$7C,$97,$79,$FF,$4D,$87,$59,$FD,$B6,$1D,$5F,$56
	.dc.b	$1F,$BE,$E8,$91,$FB,$B7,$F8,$0F,$47,$F0,$2E,$99,$3F,$BF,$7F,$82
	.dc.b	$FC,$4F,$E4,$32,$FF,$7D,$7B,$3F,$E6,$DE,$BF,$F4,$1B,$6F,$FD,$04
	.dc.b	$DE,$53,$FE,$76,$E3,$FD,$65,$FE,$F7,$2F,$0D,$FE,$F2,$3A,$FC,$3F
	.dc.b	$8A,$3E,$96,$76,$7F,$82,$67,$E3,$CF,$C3,$54,$F0,$97,$3F,$0B,$73
	.dc.b	$FF,$43,$9F,$E8,$39,$F4,$E3,$93,$5F,$11,$DE,$3F,$D3,$FA,$33,$C6
	.dc.b	$1F,$E9,$F8,$65,$D6,$21,$EB,$AF,$2B,$4F,$8B,$F2,$FE,$3F,$88,$F2
	.dc.b	$7C,$4C,$9D,$6B,$F2,$27,$D1,$F1,$5E,$9F,$89,$95,$AD,$D3,$ED,$F8
	.dc.b	$AF,$77,$C5,$4A,$D4,$77,$78,$61,$5F,$D1,$DB,$E2,$BB,$C7,$5C,$7B
	.dc.b	$BA,$C7,$B3,$E1,$C7,$B3,$CB,$1F,$9D,$F1,$32,$3C,$CF,$D1,$F8,$AF
	.dc.b	$A5,$94,$9F,$53,$F5,$7E,$1B,$EB,$7C,$27,$D7,$F8,$59,$5E,$F8,$E4
	.dc.b	$FC,$1B,$AF,$AB,$E1,$3A,$3E,$CC,$2F,$4F,$E6,$31,$EB,$3F,$27,$1E
	.dc.b	$AF,$C2,$C7,$FC,$F3,$F1,$5F,$E3,$78,$7A,$5F,$E7,$3F,$81,$C9,$AF
	.dc.b	$82,$DF,$9F,$2B,$FD,$3B,$7F,$E3,$26,$FB,$DF,$E9,$5F,$65,$F0,$57
	.dc.b	$E0,$A5,$7F,$A0,$E7,$2F,$2C,$7B,$34,$EB,$5D,$DF,$F4,$AE,$1D,$DB
	.dc.b	$8F,$AB,$1F,$05,$76,$9F,$EB,$7F,$10,$C5,$96,$A7,$A9,$0E,$B5,$F1
	.dc.b	$23,$C1,$C1,$FE,$92,$CB,$C0,$DC,$BC,$17,$EB,$BF,$04,$FF,$A4,$30
	.dc.b	$EF,$39,$3F,$35,$F0,$FF,$B5,$23,$F9,$DF,$88,$FD,$C7,$C3,$FE,$E7
	.dc.b	$E1,$E4,$FE,$86,$3F,$7D,$F1,$3F,$E7,$43,$FC,$E6,$4F,$0F,$F8,$05
	.dc.b	$CA,$E8,$7E,$BA,$7E,$F7,$83,$FC,$5A,$FA,$CF,$E9,$30,$F0,$BB,$C3
	.dc.b	$BF,$14,$8C,$1C,$F6,$F1,$91,$ED,$C5,$E6,$DE,$47,$85,$F4,$F0,$FA
	.dc.b	$BE,$1B,$D7,$ED,$4A,$F3,$47,$BA,$E4,$74,$3D,$5F,$11,$D1,$FE,$BB
	.dc.b	$BF,$E2,$9B,$8F,$67,$99,$F0,$70,$EC,$F0,$C3,$E5,$FC,$44,$8F,$8C
	.dc.b	$FC,$EF,$86,$F9,$FF,$0F,$F4,$39,$65,$7C,$B8,$FA,$7F,$0D,$23,$E7
	.dc.b	$3F,$5B,$E1,$B9,$47,$D1,$BA,$FA,$3E,$1B,$AB,$F5,$70,$EA,$FB,$30
	.dc.b	$E8,$FB,$B0,$BD,$2E,$4F,$F2,$2F,$3B,$81,$A7,$E1,$B3,$98,$FE,$05
	.dc.b	$FC,$4F,$C3,$F8,$23,$E4,$E9,$E0,$CB,$A4,$31,$E9,$A7,$83,$3E,$F9
	.dc.b	$B4,$C5,$EE,$3C,$2F,$E3,$FE,$1B,$F2,$E4,$7F,$25,$F0,$FF,$9A,$F8
	.dc.b	$79,$1F,$95,$7F,$AD,$7E,$99,$FE,$75,$8F,$E8,$AF,$F9,$93,$4F,$2E
	.dc.b	$88,$C6,$7F,$09,$B5,$E9,$5F,$0F,$FA,$8C,$64,$F8,$8E,$BF,$0D,$E2
	.dc.b	$7F,$17,$95,$CB,$BE,$1E,$39,$FF,$15,$BD,$B2,$F1,$78,$9E,$83,$D3
	.dc.b	$97,$AB,$E1,$E5,$FA,$9F,$6F,$C2,$FB,$BE,$1A,$4F,$B9,$F8,$3F,$0B
	.dc.b	$F0,$BE,$1B,$E1,$FC,$3C,$AF,$87,$1F,$1B,$E1,$B9,$C7,$C6,$C3,$0F
	.dc.b	$32,$1E,$8C,$3A,$BC,$8F,$9B,$7D,$5F,$6D,$FF,$A3,$7F,$37,$3F,$E0
	.dc.b	$FF,$89,$77,$BC,$9C,$36,$FF,$D2,$17,$C1,$B8,$7F,$A5,$F0,$E3,$1F
	.dc.b	$E9,$56,$BF,$E8,$A9,$4F,$84,$E1,$E0,$C9,$E6,$43,$84,$6D,$C7,$1F
	.dc.b	$E6,$95,$74,$73,$CF,$FE,$8C,$8C,$47,$23,$27,$FA,$C1,$CC,$F3,$6C
	.dc.b	$F1,$AF,$8D,$B4,$BF,$82,$FC,$7D,$F7,$1C,$E7,$CC,$F7,$3F,$D5,$39
	.dc.b	$7F,$19,$C7,$B4,$E9,$B3,$F1,$FC,$5F,$D2,$FC,$4F,$E9,$BE,$22,$4F
	.dc.b	$E9,$A3,$F5,$3F,$11,$FA,$AF,$87,$93,$FA,$A8,$FD,$6F,$C4,$7E,$BB
	.dc.b	$E2,$64,$FE,$BA,$3F,$63,$F1,$32,$3F,$62,$FE,$D3,$E2,$7A,$C7,$ED
	.dc.b	$30,$EA,$FE,$93,$0E,$AF,$EE,$30,$EA,$FE,$A3,$0F,$DE,$FC,$3F,$EF
	.dc.b	$BE,$1A,$4F,$EF,$A3,$FC,$B5,$FA,$6F,$F9,$AC,$91,$8F,$87,$CF,$C5
	.dc.b	$6B,$E2,$A4,$FF,$92,$AB,$FC,$BC,$FF,$19,$5F,$11,$D0,$91,$BF,$89
	.dc.b	$EE,$C3,$91,$DC,$E9,$7D,$9F,$F5,$5D,$2F,$C7,$1E,$1F,$87,$F4,$FC
	.dc.b	$44,$8C,$1E,$5F,$87,$F6,$91,$F4,$7C,$47,$C0,$F8,$99,$1E,$B7,$E1
	.dc.b	$FC,$4F,$C4,$F8,$99,$3E,$F7,$E3,$FC,$47,$38,$F2,$E1,$7B,$3F,$93
	.dc.b	$EA,$CB,$AD,$7A,$B2,$B5,$AC,$BE,$8F,$C4,$74,$AF,$99,$96,$1F,$51
	.dc.b	$0E,$E7,$A8,$F4,$10,$EC,$7D,$59,$FB,$BF,$0F,$F7,$BE,$22,$4F,$D9
	.dc.b	$8F,$C0,$FC,$4F,$E0,$BE,$1A,$4F,$DD,$8F,$C2,$F9,$5F,$86,$F8,$69
	.dc.b	$3F,$81,$8F,$C4,$FC,$34,$8F,$C2,$3F,$EA,$5E,$EF,$E0,$9B,$7C,$4C
	.dc.b	$3A,$D7,$DC,$C7,$C2,$F3,$3F,$19,$8F,$67,$F0,$38,$F3,$F8,$69,$1F
	.dc.b	$E9,$AC,$51,$DF,$E2,$7C,$1F,$0F,$2B,$4F,$8F,$E1,$FC,$9F,$0D,$E5
	.dc.b	$F2,$65,$73,$8F,$47,$A3,$23,$B3,$EB,$F4,$7A,$47,$FB,$FB,$3F,$F2
	.dc.b	$AF,$A1,$FE,$04,$6C,$7F,$7B,$58,$F8,$23,$D1,$8F,$81,$E5,$8F,$C7
	.dc.b	$F8,$79,$1E,$D7,$E5,$7C,$3F,$FB,$47,$F4,$AF,$D1,$7B,$B7,$C3,$34
	.dc.b	$D7,$D0,$FF,$69,$E7,$FF,$06,$FC,$5B,$F5,$27,$6F,$FB,$6E,$31,$E2
	.dc.b	$59,$36,$7B,$7C,$1C,$DB,$FD,$9D,$A7,$EC,$E3,$C5,$F0,$7E,$19,$CE
	.dc.b	$BE,$46,$EB,$E4,$F6,$FA,$47,$FB,$5F,$4F,$0B,$74,$F0,$B7,$3F,$C5
	.dc.b	$7C,$5C,$8F,$C0,$3F,$8D,$F8,$7F,$04,$37,$FF,$15,$ED,$F1,$A3,$C2
	.dc.b	$7A,$38,$3E,$27,$3A,$19,$FC,$5C,$BE,$17,$0C,$9F,$11,$F8,$FF,$88
	.dc.b	$FC,$E1,$1F,$C9,$7C,$47,$E7,$FE,$2F,$F4,$1F,$17,$27,$F2,$D1,$FA
	.dc.b	$3F,$8B,$FD,$27,$34,$9F,$CD,$47,$E9,$F9,$A4,$7E,$75,$FD,$57,$2F
	.dc.b	$58,$FD,$06,$1D,$5F,$C9,$61,$D5,$FD,$26,$1D,$5F,$CB,$61,$EF,$7E
	.dc.b	$36,$DF,$E5,$AF,$D3,$FE,$1C,$7F,$94,$5B,$F1,$26,$9F,$E3,$13,$FC
	.dc.b	$7B,$1F,$27,$8F,$FC,$8B,$18,$8E,$A7,$3F,$88,$F0,$4E,$BE,$2E,$E1
	.dc.b	$E1,$F8,$9B,$7E,$1F,$89,$B9,$78,$5B,$92,$AF,$C4,$CB,$F1,$61,$EB
	.dc.b	$8D,$FC,$3F,$0D,$25,$68,$F2,$17,$2F,$76,$7D,$B1,$EF,$9C,$AD,$67
	.dc.b	$E1,$6F,$C3,$F9,$3E,$1E,$46,$0D,$FC,$47,$A3,$E1,$E4,$60,$E7,$F1
	.dc.b	$1E,$C2,$3D,$BC,$9F,$7F,$C3,$48,$F0,$BF,$0B,$E1,$7E,$1E,$1D,$27
	.dc.b	$E7,$EF,$F5,$F1,$CC,$76,$21,$AD,$19,$7C,$CC,$BE,$6F,$C3,$7C,$EF
	.dc.b	$88,$97,$F0,$23,$9B,$C7,$0F,$07,$37,$E0,$33,$F4,$CF,$5C,$6C,$FE
	.dc.b	$13,$1F,$B3,$F1,$12,$BE,$5B,$F7,$3E,$1F,$EE,$FC,$44,$9F,$9E,$FE
	.dc.b	$0B,$F5,$0F,$CC,$47,$B9,$7C,$4E,$49,$1F,$6E,$3F,$C2,$7C,$07,$FC
	.dc.b	$11,$19,$F8,$9E,$87,$5F,$3B,$C1,$1F,$17,$0B,$A7,$FA,$82,$FC,$2D
	.dc.b	$C7,$BD,$78,$DF,$37,$C4,$79,$FE,$27,$D1,$F1,$32,$BC,$91,$EB,$F8
	.dc.b	$99,$1E,$77,$DD,$F1,$3E,$FF,$88,$F8,$1F,$11,$2B,$D9,$1F,$0F,$CE
	.dc.b	$F8,$9E,$74,$9F,$81,$1D,$1F,$8D,$85,$D7,$D5,$8F,$59,$F8,$D8,$F5
	.dc.b	$7E,$66,$3F,$3F,$E1,$E4,$7C,$B7,$FD,$25,$F3,$BF,$DE,$BC,$8F,$E1
	.dc.b	$5E,$27,$FB,$4E,$BF,$58,$3F,$18,$FF,$46,$E7,$FE,$83,$6D,$70,$F4
	.dc.b	$3F,$7F,$5F,$BB,$6C,$BD,$EA,$6B,$0E,$D5,$FC,$30,$E0,$78,$1C,$E1
	.dc.b	$FE,$91,$6E,$1D,$E7,$58,$73,$23,$82,$1D,$67,$8B,$FD,$15,$AF,$80
	.dc.b	$D7,$9D,$E5,$23,$F9,$1F,$8A,$EA,$F3,$AB,$87,$C3,$F7,$6E,$3E,$5A
	.dc.b	$D1,$F2,$B3,$7B,$DF,$A1,$F8,$9F,$D1,$7C,$54,$AF,$CF,$47,$E9,$7E
	.dc.b	$27,$F4,$DF,$13,$27,$F4,$51,$FA,$9F,$88,$FD,$57,$C4,$C9,$FD,$34
	.dc.b	$7E,$B7,$E2,$A4,$7E,$A5,$FE,$63,$FA,$8F,$DF,$87,$8B,$FA,$BC,$3B
	.dc.b	$47,$EC,$F1,$F1,$3D,$0F,$D9,$63,$DD,$FD,$C6,$3D,$3E,$2F,$AF,$C5
	.dc.b	$C9,$FE,$65,$97,$F8,$E9,$BF,$17,$E2,$23,$AF,$8A,$F2,$F1,$F9,$B8
	.dc.b	$64,$F2,$8F,$4F,$C5,$C8,$EA,$FB,$3A,$B9,$C7,$F2,$8C,$7F,$C6,$FC
	.dc.b	$06,$9F,$E3,$88,$E8,$79,$B0,$EE,$F1,$C3,$E3,$FC,$5F,$C8,$F8,$B9
	.dc.b	$3E,$C8,$F9,$7F,$13,$23,$DE,$FC,$EF,$89,$F9,$FF,$0D,$27,$E1,$BF
	.dc.b	$4B,$B3,$E9,$F7,$49,$F8,$EF,$D6,$F8,$4E,$91,$EB,$C2,$F7,$3B,$BF
	.dc.b	$C3,$DC,$1E,$DC,$FB,$5F,$C1,$CE,$DE,$E3,$EF,$F0,$7E,$8A,$FE,$27
	.dc.b	$C1,$F7,$9F,$AF,$9F,$E1,$A3,$BE,$5D,$AD,$9E,$2F,$9E,$7D,$AB,$59
	.dc.b	$D9,$EF,$AF,$E4,$3E,$13,$F2,$3F,$0D,$27,$F0,$B1,$F9,$4F,$86,$FF
	.dc.b	$02,$DF,$F8,$58,$FF,$0B,$E7,$FE,$4F,$AF,$EC,$70,$F8,$ED,$FC,$2C
	.dc.b	$9F,$F0,$04,$E5,$BF,$0B,$D0,$8E,$BC,$7F,$03,$C3,$E3,$3C,$26,$1E
	.dc.b	$F8,$75,$7C,$38,$79,$A3,$CF,$F0,$D2,$7C,$2F,$AB,$E2,$3D,$7F,$11
	.dc.b	$27,$CA,$FB,$BE,$23,$DF,$F1,$1F,$03,$E0,$A5,$7A,$A3,$E1,$FC,$1C
	.dc.b	$8F,$6B,$F1,$BE,$0F,$A4,$7C,$4C,$2F,$67,$DF,$EC,$71,$33,$EF,$7F
	.dc.b	$1B,$3E,$EF,$CC,$CF,$E8,$FC,$1C,$8F,$96,$FF,$8E,$B7,$7E,$97,$FC
	.dc.b	$A9,$BD,$F9,$83,$4C,$77,$FF,$A4,$7C,$EF,$F4,$A7,$EA,$7F,$C4,$7E
	.dc.b	$2C,$DF,$8E,$4F,$3D,$E6,$9F,$E3,$DC,$FC,$31,$DF,$C3,$ED,$39,$D3
	.dc.b	$C2,$DD,$3C,$2F,$FA,$47,$0F,$0F,$83,$23,$07,$1E,$FF,$51,$1F,$C5
	.dc.b	$FC,$67,$67,$FC,$C3,$9F,$C0,$7B,$1E,$1D,$FD,$CF,$FA,$87,$6F,$71
	.dc.b	$ED,$C7,$F3,$BF,$19,$27,$F2,$EF,$E8,$3E,$2B,$F4,$3F,$17,$FA,$2F
	.dc.b	$8B,$95,$F9,$C8,$FD,$2F,$C5,$FE,$9B,$E2,$64,$FE,$82,$3F,$CD,$BF
	.dc.b	$A9,$FF,$87,$38,$3F,$82,$B6,$7F,$9F,$B7,$AC,$FF,$8C,$23,$B1,$FA
	.dc.b	$AC,$7C,$2F,$EC,$31,$F0,$BF,$AE,$C7,$C2,$FE,$D3,$1F,$0F,$C4,$48
	.dc.b	$FF,$39,$61,$FE,$38,$3F,$C6,$EF,$F9,$71,$7C,$67,$A0,$91,$C3,$E2
	.dc.b	$3D,$7F,$13,$EC,$F8,$79,$3D,$23,$DF,$F0,$F2,$3B,$BF,$0B,$E1,$B9
	.dc.b	$C7,$7C,$2F,$87,$FA,$5C,$99,$76,$AF,$3E,$5D,$9F,$36,$5F,$33,$E1
	.dc.b	$FE,$6F,$C4,$49,$F6,$C7,$D0,$F8,$9F,$A3,$F1,$32,$7E,$0C,$7D,$4F
	.dc.b	$88,$91,$F1,$1F,$AF,$F0,$DF,$63,$E1,$E4,$FC,$87,$ED,$FC,$3F,$58
	.dc.b	$F9,$78,$75,$7E,$56,$1D,$1F,$99,$85,$DF,$DD,$8F,$E1,$3E,$1F,$F0
	.dc.b	$BF,$0D,$2F,$EA,$C7,$57,$EA,$63,$D6,$FE,$36,$3C,$FA,$FF,$17,$95
	.dc.b	$BF,$C5,$E5,$F9,$0F,$87,$91,$F8,$07,$F2,$7D,$FF,$EC,$0F,$27,$FC
	.dc.b	$21,$DC,$69,$3E,$1F,$5F,$0F,$30,$FF,$5D,$4A,$39,$7C,$3F,$3F,$87
	.dc.b	$E9,$F0,$92,$F7,$1D,$F6,$91,$C9,$F1,$63,$D6,$3F,$D5,$B8,$F5,$7B
	.dc.b	$E3,$D9,$F0,$FC,$2F,$A7,$E1,$E4,$60,$F2,$FC,$3F,$B4,$8F,$A3,$E1
	.dc.b	$BE,$07,$C3,$FC,$1F,$87,$93,$EC,$8F,$89,$F0,$FF,$17,$D5,$93,$F0
	.dc.b	$23,$9B,$E6,$C2,$F4,$7F,$3D,$D5,$97,$6A,$F8,$F9,$76,$7D,$B9,$7D
	.dc.b	$1F,$87,$91,$F3,$1F,$A9,$EA,$7F,$9E,$3F,$13,$FF,$3B,$4F,$FB,$AF
	.dc.b	$E4,$1F,$5D,$FE,$39,$8F,$F8,$03,$18,$E3,$E1,$B4,$9F,$E4,$1E,$CF
	.dc.b	$F3,$E3,$0F,$F9,$F7,$0E,$AD,$9C,$EF,$FE,$15,$84,$EA,$30,$DE,$54
	.dc.b	$C8,$C3,$6B,$9C,$EA,$39,$C6,$1D,$1F,$E9,$C9,$E3,$D3,$F9,$3F,$57
	.dc.b	$F2,$D8,$CA,$FC,$8C,$71,$79,$9C,$6F,$B6,$3F,$E2,$DC,$FE,$3F,$56
	.dc.b	$F2,$F2,$DF,$01,$5F,$E3,$3F,$6B,$FB,$0F,$B3,$FA,$52,$4F,$E8,$1F
	.dc.b	$D4,$7C,$57,$EA,$7E,$2B,$BE,$FF,$E4,$0F,$65,$F0,$F8,$C3,$FC,$F7
	.dc.b	$18,$7F,$53,$59,$E1,$FE,$DF,$ED,$DA,$7C,$3D,$31,$19,$F8,$99,$87
	.dc.b	$D0,$DB,$FC,$19,$69,$E5,$F1,$3E,$5A,$D6,$9E,$5D,$3B,$19,$8F,$83
	.dc.b	$1E,$9C,$7C,$11,$C0,$F4,$E5,$E2,$71,$9F,$F8,$1B,$E2,$7D,$1F,$13
	.dc.b	$27,$3A,$27,$D9,$F0,$F2,$72,$FB,$FE,$1F,$E0,$7C,$37,$C1,$F8,$79
	.dc.b	$5C,$A3,$E2,$7C,$4C,$8E,$AF,$C7,$F8,$9E,$91,$EA,$3E,$7E,$18,$E3
	.dc.b	$F8,$B8,$F6,$9F,$2E,$3D,$9E,$1A,$FF,$88,$32,$FA,$5F,$13,$F4,$FE
	.dc.b	$22,$57,$B2,$3E,$B7,$C4,$48,$F7,$BF,$67,$E2,$3E,$D7,$C3,$C9,$F8
	.dc.b	$6F,$DD,$F8,$8F,$BD,$F0,$F2,$7E,$3B,$F8,$1F,$88,$E9,$1E,$4E,$46
	.dc.b	$38,$E0,$FB,$F9,$75,$AD,$65,$67,$F0,$79,$7E,$37,$D8,$FF,$06,$1F
	.dc.b	$E0,$BA,$F7,$B6,$7F,$CA,$8F,$F8,$47,$1E,$F3,$EC,$21,$E3,$C3,$B7
	.dc.b	$4F,$E7,$F1,$B3,$F9,$EC,$7C,$1F,$11,$E1,$F8,$F9,$3F,$88,$C9,$3E
	.dc.b	$5F,$87,$F3,$7C,$34,$AD,$47,$A7,$E1,$7D,$5F,$0D,$27,$94,$7B,$7E
	.dc.b	$26,$47,$57,$E0,$7C,$4F,$58,$F0,$61,$D5,$F8,$38,$75,$7C,$98,$75
	.dc.b	$7E,$2E,$1F,$27,$E1,$E4,$7A,$5F,$99,$F1,$1F,$37,$E0,$E4,$FB,$5F
	.dc.b	$A1,$F0,$9F,$47,$6F,$A5,$F0,$92,$BE,$14,$7D,$5F,$88,$91,$F1,$5F
	.dc.b	$B1,$F1,$1D,$23,$EB,$E1,$7A,$3E,$BE,$3D,$67,$E6,$63,$D5,$FB,$B8
	.dc.b	$FE,$07,$E1,$E4,$7D,$17,$F0,$9F,$11,$FE,$89,$E4,$FF,$5D,$63,$FE
	.dc.b	$D3,$F2,$BE,$AF,$9D,$93,$FD,$1D,$FA,$5F,$F1,$9C,$7F,$5D,$E9,$33
	.dc.b	$79,$9F,$C7,$E3,$1E,$4D,$93,$67,$C1,$E9,$4B,$FF,$45,$68,$8F,$27
	.dc.b	$A5,$E1,$9C,$E7,$5F,$F3,$FE,$7E,$3D,$7F,$9C,$67,$E2,$9F,$E3,$1A
	.dc.b	$78,$9B,$1E,$26,$E7,$FA,$5F,$89,$91,$F9,$57,$F5,$1F,$13,$DF,$0F
	.dc.b	$89,$AF,$C6,$F0,$BF,$AF,$F0,$7C,$6C,$3F,$D4,$5A,$78,$E7,$FB,$1E
	.dc.b	$9F,$17,$3F,$85,$C1,$2F,$C2,$7F,$2A,$FD,$53,$FC,$62,$4B,$FE,$88
	.dc.b	$FF,$7D,$CF,$FD,$90,$FF,$88,$E7,$F8,$BA,$F5,$F6,$70,$F8,$8E,$3F
	.dc.b	$13,$CB,$E2,$A5,$F7,$8E,$BF,$15,$23,$C4,$F8,$3E,$2B,$94,$79,$70
	.dc.b	$EA,$F9,$F0,$EA,$FA,$70,$EA,$F9,$30,$F4,$FB,$D2,$3D,$CF,$B3,$E1
	.dc.b	$FD,$BF,$0D,$27,$E1,$3F,$03,$E1,$FE,$0F,$C4,$FC,$2F,$8A,$95,$F1
	.dc.b	$E3,$E2,$FC,$4C,$8F,$94,$FC,$8F,$89,$E9,$1F,$1B,$0B,$C5,$F2,$B1
	.dc.b	$EB,$3F,$47,$1E,$AF,$CC,$C7,$E8,$FC,$37,$D2,$F8,$79,$3F,$5E,$3E
	.dc.b	$AF,$83,$F5,$BC,$19,$3F,$6E,$3E,$CF,$83,$23,$EF,$3F,$DF,$FF,$55
	.dc.b	$FC,$34,$3D,$D4,$3B,$F4,$49,$FE,$DF,$1F,$E0,$47,$B1,$F8,$9C,$3C
	.dc.b	$2F,$E1,$30,$F0,$3F,$86,$C2,$EC,$EB,$FF,$1B,$E5,$E0,$F8,$4F,$27
	.dc.b	$C3,$79,$7E,$1A,$63,$E0,$8F,$7B,$DC,$87,$73,$DA,$43,$C0,$7A,$27
	.dc.b	$E0,$7C,$2F,$C1,$F0,$64,$FB,$23,$E2,$78,$32,$3D,$EF,$C7,$F8,$3F
	.dc.b	$91,$F0,$52,$7E,$1B,$F2,$FE,$0F,$E6,$7C,$2F,$CD,$F8,$59,$5F,$22
	.dc.b	$3A,$3F,$DF,$B9,$3F,$81,$37,$1E,$CF,$42,$1B,$C3,$B3,$F4,$BE,$13
	.dc.b	$ED,$7C,$2C,$8C,$1F,$5B,$E0,$FE,$F1,$1F,$B3,$F0,$5F,$81,$F8,$2F
	.dc.b	$C1,$7C,$14,$9F,$BB,$1F,$85,$F8,$59,$1F,$80,$7F,$11,$F0,$BC,$E3
	.dc.b	$D5,$85,$E9,$FB,$91,$F8,$BD,$99,$77,$8F,$C4,$E5,$DD,$F8,$59,$7F
	.dc.b	$94,$7F,$56,$FF,$3A,$47,$F8,$89,$B8,$74,$E9,$46,$3A,$BF,$12,$FF
	.dc.b	$82,$5F,$F3,$E7,$C4,$7F,$A8,$2F,$FD,$2D,$F9,$FF,$F2,$75,$F1,$48
	.dc.b	$C4,$7F,$A3,$3C,$AF,$F4,$F1,$FE,$67,$97,$E7,$FF,$A6,$6F,$6E,$AB
	.dc.b	$FD,$0D,$55,$FF,$4D,$37,$4F,$F4,$AB,$72,$F2,$B6,$7F,$D2,$CD,$7C
	.dc.b	$47,$43,$FC,$BB,$9F,$F8,$59,$BF,$13,$F1,$CF,$91,$EB,$F1,$D7,$E0
	.dc.b	$4F,$88,$FF,$5B,$69,$FB,$77,$04,$CB,$CA,$EB,$C0,$FA,$5F,$0D,$23
	.dc.b	$C2,$FD,$5F,$87,$FA,$DF,$0D,$F5,$FE,$1E,$57,$9A,$3E,$D7,$C4,$FD
	.dc.b	$BF,$89,$93,$EA,$8F,$BD,$F1,$32,$3D,$AF,$E0,$7E,$27,$AC,$7C,$0C
	.dc.b	$3A,$BF,$83,$C3,$AB,$F1,$30,$EA,$FE,$1F,$0F,$C5,$7C,$3F,$E2,$FE
	.dc.b	$1E,$4F,$CA,$8F,$C7,$7C,$24,$8F,$9A,$FE,$47,$E1,$FF,$25,$F1,$3F
	.dc.b	$93,$F8,$99,$5F,$4A,$3F,$2D,$F1,$32,$3E,$AB,$F9,$9F,$89,$E9,$1F
	.dc.b	$99,$C2,$F2,$7E,$67,$A5,$97,$6A,$FB,$B9,$76,$7F,$41,$97,$E8,$FE
	.dc.b	$27,$F4,$9F,$13,$27,$F0,$71,$FE,$B7,$F2,$7F,$DC,$99,$7E,$95,$BD
	.dc.b	$DF,$CB,$1B,$5F,$CF,$5F,$EF,$51,$9F,$83,$E4,$49,$FF,$55,$CF,$F7
	.dc.b	$B6,$FC,$17,$72,$3A,$F8,$3F,$15,$C8,$E2,$F9,$7E,$33,$A4,$7F,$A7
	.dc.b	$2F,$FC,$11,$8F,$9F,$1F,$57,$C2,$FA,$FE,$1F,$D9,$F0,$F3,$0F,$1C
	.dc.b	$75,$79,$65,$70,$F8,$19,$76,$7A,$4F,$AB,$1E,$AF,$B3,$E1,$FE,$47
	.dc.b	$8D,$23,$07,$C0,$F8,$5F,$99,$F0,$B2,$30,$7C,$4F,$84,$FA,$1F,$0B
	.dc.b	$23,$07,$C8,$F8,$7F,$A8,$47,$E5,$FC,$3F,$D7,$C3,$A4,$7C,$3C,$30
	.dc.b	$F2,$3E,$35,$F5,$7E,$46,$1F,$7F,$E1,$A4,$7D,$67,$F0,$5F,$0F,$F8
	.dc.b	$3F,$85,$93,$F6,$9F,$C3,$7C,$37,$E1,$FE,$1B,$FD,$FD,$FA,$BF,$F8
	.dc.b	$03,$D6,$37,$A7,$F9,$34,$63,$E2,$36,$4C,$7F,$DD,$F5,$FC,$2F,$97
	.dc.b	$F2,$19,$78,$27,$EF,$3E,$07,$F1,$78,$F8,$3D,$59,$18,$39,$FA,$3E
	.dc.b	$42,$3D,$BE,$23,$FC,$FF,$12,$3F,$CB,$AF,$FA,$0A,$3F,$CF,$9F,$0D
	.dc.b	$FE,$7D,$F8,$29,$5F,$E6,$E8,$FF,$43,$E9,$23,$FC,$CA,$FF,$A2,$BE
	.dc.b	$37,$A4,$7F,$9E,$B0,$BA,$7F,$A4,$7E,$31,$97,$6A,$FF,$3C,$E5,$D9
	.dc.b	$FF,$48,$F9,$4C,$FE,$6F,$C4,$C9,$F8,$EF,$D0,$F8,$AE,$BE,$0F,$F9
	.dc.b	$E3,$26,$98,$BF,$F1,$84,$78,$1F,$51,$E0,$21,$E0,$C3,$ED,$FC,$3C
	.dc.b	$8F,$AA,$FD,$E8,$FB,$F9,$FE,$03,$E2,$A5,$7D,$A8,$FC,$1F,$C4,$FE
	.dc.b	$13,$E2,$A4,$FD,$E8,$FC,$3F,$C5,$48,$FC,$0B,$F8,$AE,$9E,$B1,$F8
	.dc.b	$4C,$3A,$AE,$2F,$EF,$1D,$0C,$BB,$CF,$E2,$B2,$EE,$EF,$7F,$F2,$1D
	.dc.b	$33,$FC,$BF,$C4,$4A,$FC,$8B,$F9,$AF,$87,$FC,$DF,$C3,$49,$FC,$AB
	.dc.b	$F9,$EF,$87,$FC,$FF,$C5,$7E,$83,$E2,$A5,$7E,$6A,$0C,$DB,$FC,$1D
	.dc.b	$FA,$C7,$F8,$36,$3F,$CC,$8D,$F0,$0D,$3F,$C2,$11,$FE,$69,$D5,$97
	.dc.b	$8E,$3D,$86,$5F,$23,$E1,$3F,$57,$8F,$91,$DD,$FE,$D3,$1F,$2C,$63
	.dc.b	$0F,$F0,$13,$7E,$23,$39,$7A,$1B,$BF,$F9,$B2,$B7,$3E,$CF,$85,$CD
	.dc.b	$FB,$3E,$16,$E5,$EC,$6D,$AA,$FC,$14,$BF,$8D,$8F,$58,$DF,$C1,$F0
	.dc.b	$F8,$29,$5A,$3E,$4B,$D3,$E0,$FA,$E1,$DB,$29,$5B,$CF,$E6,$CF,$78
	.dc.b	$F8,$18,$DD,$3E,$AE,$5E,$6F,$85,$F3,$FC,$2C,$BE,$5B,$27,$BC,$63
	.dc.b	$3E,$EF,$CB,$CB,$A6,$DE,$B8,$E3,$9F,$B1,$E7,$F0,$9F,$13,$E1,$64
	.dc.b	$60,$EF,$E2,$FC,$82,$3E,$2F,$82,$F9,$7F,$03,$23,$CA,$FC,$EF,$82
	.dc.b	$F9,$FF,$0B,$F4,$3E,$16,$57,$AA,$39,$BF,$81,$C2,$F1,$FE,$5B,$1E
	.dc.b	$93,$F8,$5C,$6C,$FE,$6B,$1F,$B7,$F0,$7F,$73,$E1,$25,$FC,$68,$FC
	.dc.b	$0F,$E5,$FF,$E0,$6E,$73,$49,$F0,$1A,$F8,$0D,$FC,$1C,$C3,$EC,$57
	.dc.b	$F7,$CE,$4F,$F5,$DA,$F8,$5E,$A4,$9D,$FC,$2F,$82,$A4,$72,$7C,$78
	.dc.b	$F7,$8E,$B8,$77,$7F,$4B,$87,$77,$FD,$B0,$BE,$13,$D6,$48,$F2,$FC
	.dc.b	$2F,$E8,$17,$C4,$3E,$0B,$FC,$FD,$F4,$CE,$BA,$6E,$75,$5F,$9B,$38
	.dc.b	$D7,$0F,$06,$46,$67,$6F,$F0,$03,$66,$67,$A1,$1C,$10,$E1,$1C,$B9
	.dc.b	$7F,$D5,$71,$C8,$CC,$F1,$69,$C6,$3F,$D4,$D9,$BE,$1F,$EE,$FC,$3C
	.dc.b	$AF,$AF,$1C,$6B,$AE,$1D,$F8,$3F,$D5,$38,$FC,$4F,$13,$51,$8C,$3B
	.dc.b	$3F,$CA,$3E,$2D,$B7,$EB,$E7,$F0,$77,$F8,$C2,$3F,$86,$F8,$6F,$C7
	.dc.b	$FC,$2F,$E4,$3E,$1A,$4F,$E2,$A3,$F2,$7F,$11,$F9,$4F,$88,$93,$F8
	.dc.b	$E8,$FC,$BF,$C4,$48,$FC,$8B,$F9,$AF,$88,$EB,$1F,$94,$C3,$AB,$F8
	.dc.b	$8C,$3A,$BF,$98,$C3,$AB,$F8,$CC,$3F,$47,$F0,$FF,$A4,$F8,$89,$3F
	.dc.b	$9E,$8F,$D3,$FC,$44,$8F,$D0,$BF,$CA,$AB,$FC,$58,$3E,$1F,$1F,$0F
	.dc.b	$2B,$F8,$F4,$E5,$BF,$09,$C4,$8E,$BE,$1F,$B3,$FA,$0C,$2F,$3F,$FB
	.dc.b	$8E,$FB,$E3,$D6,$BF,$A4,$C7,$F8,$D7,$97,$FA,$EF,$88,$8F,$04,$EB
	.dc.b	$E1,$FD,$1F,$0F,$23,$CB,$1E,$BF,$86,$91,$E8,$7D,$DF,$0D,$EF,$F8
	.dc.b	$39,$3E,$C7,$E1,$7C,$27,$C3,$F8,$39,$3F,$01,$F8,$DF,$09,$D2,$3C
	.dc.b	$F8,$5D,$FD,$98,$F4,$9F,$5E,$36,$7E,$06,$3F,$3F,$E1,$BE,$87,$C3
	.dc.b	$4B,$F9,$B1,$CF,$83,$E0,$EE,$CF,$14,$E0,$F8,$9A,$75,$BD,$E9,$6B
	.dc.b	$58,$33,$FB,$DF,$0D,$F7,$FE,$1A,$57,$DC,$8F,$C1,$7C,$27,$E0,$FE
	.dc.b	$1A,$4F,$E0,$23,$F0,$DF,$09,$F8,$7F,$85,$93,$F8,$48,$FE,$E7,$FA
	.dc.b	$CF,$F8,$3C,$91,$8F,$99,$EA,$FE,$2B,$0E,$AF,$E0,$30,$EA,$FE,$3B
	.dc.b	$0E,$AF,$E1,$30,$EB,$F1,$B2,$3F,$B8,$63,$FE,$06,$DF,$F0,$EB,$E2
	.dc.b	$3C,$7F,$07,$2B,$6F,$9B,$E1,$3C,$FE,$9F,$A3,$D3,$95,$D2,$3D,$7F
	.dc.b	$11,$23,$BB,$EE,$F4,$3A,$47,$F8,$83,$2F,$F0,$83,$D8,$F2,$E1,$D9
	.dc.b	$F3,$E1,$D9,$E1,$87,$C8,$F8,$69,$1E,$C7,$E5,$FC,$47,$F8,$83,$C3
	.dc.b	$7C,$57,$E0,$EB,$FC,$2D,$18,$ED,$D4,$E6,$7F,$C7,$33,$C3,$B3,$FC
	.dc.b	$53,$F2,$2F,$85,$FC,$25,$7F,$87,$E3,$1E,$26,$A7,$FC,$B5,$39,$76
	.dc.b	$F4,$C1,$E2,$F8,$09,$1E,$48,$78,$DE,$4C,$E4,$E5,$F3,$E5,$DF,$BF
	.dc.b	$C1,$94,$AF,$04,$7E,$33,$E2,$64,$7E,$01,$FC,$7F,$C4,$F4,$8F,$26
	.dc.b	$17,$C7,$FF,$51,$63,$D6,$7D,$58,$F5,$7F,$D4,$38,$FE,$63,$59,$1F
	.dc.b	$8C,$7F,$37,$F0,$FF,$9C,$F8,$7F,$CE,$FC,$3C,$AF,$C8,$C7,$E8,$3E
	.dc.b	$23,$F4,$3F,$13,$27,$F2,$B1,$FA,$4F,$88,$91,$F9,$87,$F4,$FF,$13
	.dc.b	$D6,$3F,$37,$87,$57,$F1,$98,$75,$7F,$3F,$87,$57,$F2,$18,$7F,$98
	.dc.b	$BF,$5A,$FF,$11,$13,$C2,$C7,$C3,$E7,$E2,$A5,$FF,$95,$EB,$FC,$25
	.dc.b	$E7,$97,$C4,$73,$F8,$AE,$9F,$15,$2F,$84,$77,$F8,$A9,$1C,$DF,$17
	.dc.b	$C5,$76,$8F,$F4,$1E,$1D,$89,$87,$58,$F1,$61,$D5,$DE,$1E,$BF,$88
	.dc.b	$F6,$7C,$4C,$9F,$44,$7B,$FE,$1B,$E0,$7C,$34,$9F,$64,$7C,$3F,$83
	.dc.b	$91,$EF,$7E,$37,$C2,$7C,$7F,$66,$4F,$C3,$7E,$57,$C2,$75,$8F,$8D
	.dc.b	$87,$57,$D7,$87,$47,$DB,$85,$F0,$1C,$BF,$D6,$38,$19,$FD,$5F,$86
	.dc.b	$FA,$DF,$0D,$31,$FA,$51,$DD,$F8,$F9,$F7,$C7,$9C,$7F,$3F,$C3,$FC
	.dc.b	$3C,$DC,$BC,$19,$77,$7D,$B1,$F7,$7E,$17,$F0,$9F,$0D,$23,$07,$E0
	.dc.b	$7E,$0F,$F1,$04,$7F,$D3,$5F,$AD,$BE,$3B,$FD,$5A,$FF,$70,$CF,$C7
	.dc.b	$D4,$8C,$67,$FD,$D9,$B7,$FD,$55,$FE,$F4,$DF,$87,$E7,$F0,$F2,$3B
	.dc.b	$C7,$85,$FC,$3E,$57,$1E,$F7,$E3,$9F,$2F,$A3,$E4,$23,$E8,$F8,$8F
	.dc.b	$3F,$A3,$E8,$F4,$24,$FB,$23,$D7,$E7,$48,$F7,$BE,$EF,$3B,$DF,$F0
	.dc.b	$BF,$03,$CE,$95,$F1,$23,$E1,$F9,$F2,$3E,$3B,$F1,$BE,$23,$A4,$7F
	.dc.b	$B8,$B5,$FE,$65,$B7,$AB,$2E,$D3,$CC,$F9,$F8,$F6,$78,$63,$F4,$3E
	.dc.b	$1E,$47,$D4,$7E,$9F,$C4,$7F,$A0,$FC,$47,$1F,$FA,$17,$F3,$27,$4D
	.dc.b	$35,$13,$0F,$F4,$14,$B9,$78,$19,$8F,$F4,$05,$F4,$7F,$D1,$0D,$73
	.dc.b	$5B,$D7,$FC,$FB,$39,$6C,$67,$9F,$FE,$00,$CE,$5B,$19,$A1,$FE,$AF
	.dc.b	$71,$D1,$F8,$DF,$88,$FC,$8F,$C4,$CA,$FC,$AC,$71,$78,$1E,$6A,$F8
	.dc.b	$5E,$0E,$71,$F0,$4F,$66,$AD,$1D,$BF,$9E,$F8,$79,$3F,$A2,$7F,$43
	.dc.b	$F1,$1F,$A2,$F8,$7F,$D1,$FC,$44,$AF,$D3,$C7,$A9,$70,$7E,$9B,$D8
	.dc.b	$C4,$63,$87,$FC,$B7,$0F,$89,$E1,$5B,$6C,$EA,$A9,$FE,$3F,$AF,$F0
	.dc.b	$F3,$7E,$27,$3C,$69,$CB,$7E,$27,$D1,$5F,$BC,$CF,$D0,$FE,$57,$3E
	.dc.b	$37,$E6,$CF,$C5,$3F,$51,$AF,$A4,$F1,$1C,$3E,$27,$8F,$C3,$48,$E1
	.dc.b	$A7,$BA,$7A,$FC,$2C,$9C,$9E,$B7,$C3,$F0,$DE,$2F,$89,$F1,$FC,$4C
	.dc.b	$AD,$EF,$ED,$9F,$3F,$C4,$49,$CE,$BF,$EB,$97,$CD,$5F,$0F,$E2,$3C
	.dc.b	$11,$89,$FF,$73,$3C,$32,$F0,$BC,$B0,$F0,$3B,$C7,$C8,$F6,$F8,$6F
	.dc.b	$8F,$F0,$D2,$30,$78,$BE,$1B,$E5,$91,$F2,$E3,$F3,$BE,$0E,$47,$A1
	.dc.b	$FA,$3F,$09,$F4,$BE,$0E,$4F,$B1,$FA,$BF,$09,$CE,$3D,$58,$5E,$5F
	.dc.b	$C7,$63,$D2,$7D,$D8,$D9,$FC,$96,$3F,$7B,$E1,$BE,$FF,$C3,$4B,$F9
	.dc.b	$31,$CF,$9B,$F2,$B9,$59,$FC,$86,$5C,$E7,$F3,$39,$59,$FC,$A6,$3F
	.dc.b	$92,$FC,$D7,$FD,$FF,$9F,$F4,$65,$F0,$F9,$35,$86,$FE,$1F,$87,$C2
	.dc.b	$CC,$7E,$FD,$FF,$85,$07,$8B,$D7,$E1,$64,$EA,$3C,$1F,$0F,$23,$8B
	.dc.b	$E3,$F8,$7E,$D1,$D3,$0E,$CF,$E1,$B0,$ED,$7D,$23,$18,$7A,$E3,$D9
	.dc.b	$F0,$D2,$7C,$AF,$BF,$E1,$FE,$07,$C2,$C9,$F4,$BF,$0F,$E0,$FE,$27
	.dc.b	$C1,$7C,$5F,$85,$95,$EE,$8F,$91,$F0,$F2,$3E,$0B,$F2,$FE,$1F,$A4
	.dc.b	$79,$F0,$BC,$9F,$37,$1E,$B3,$F2,$31,$EA,$FB,$71,$FA,$9F,$0F,$23
	.dc.b	$E6,$BF,$5F,$D2,$FE,$1B,$F1,$8F,$29,$C9,$86,$DF,$EE,$27,$53,$9E
	.dc.b	$0F,$F4,$37,$99,$FE,$8C,$F9,$5F,$E1,$5D,$3F,$CF,$3E,$2D,$18,$C5
	.dc.b	$1F,$E9,$1A,$ED,$CD,$27,$B6,$68,$F1,$74,$F8,$F3,$95,$A7,$C4,$FF
	.dc.b	$A5,$6F,$C4,$C3,$07,$E3,$3D,$4F,$CB,$11,$FC,$7F,$B5,$D1,$FE,$3D
	.dc.b	$B7,$B9,$EC,$79,$36,$F7,$BF,$EA,$7D,$7D,$9A,$FB,$A3,$27,$E3,$7D
	.dc.b	$AF,$D2,$11,$FE,$37,$FA,$E7,$F8,$6D,$FE,$2E,$FF,$8A,$3C,$E3,$FE
	.dc.b	$28,$6F,$C3,$FF,$3E,$8D,$1F,$D7,$23,$7F,$13,$93,$5E,$DE,$FE,$1E
	.dc.b	$61,$E0,$79,$7C,$47,$3F,$89,$E9,$94,$AF,$2C,$72,$7C,$F8,$75,$7D
	.dc.b	$38,$78,$FE,$16,$47,$B1,$F3,$7C,$4F,$9F,$E1,$64,$FC,$07,$D5,$F0
	.dc.b	$DE,$BF,$89,$F6,$7C,$4C,$AF,$8B,$1E,$FF,$88,$91,$F2,$1F,$85,$F1
	.dc.b	$1D,$23,$E4,$E1,$7C,$5F,$F2,$AF,$1F,$FB,$59,$B8,$75,$AF,$9F,$8F
	.dc.b	$85,$FA,$7F,$0F,$F3,$7E,$22,$46,$0F,$AF,$F0,$FF,$47,$E1,$E4,$60
	.dc.b	$FB,$7F,$09,$F5,$48,$FD,$EF,$03,$FB,$AF,$EB,$BF,$A1,$EA,$34,$9F
	.dc.b	$11,$F7,$71,$EB,$7F,$84,$C7,$0F,$62,$1F,$88,$C7,$F0,$98,$F6,$F8
	.dc.b	$8E,$FF,$11,$2B,$FB,$4E,$5F,$E0,$4E,$16,$5D,$B8,$7E,$3E,$7D,$AF
	.dc.b	$A1,$DA,$3E,$5E,$7D,$A3,$9F,$C4,$7B,$3C,$99,$18,$3B,$FC,$2F,$C0
	.dc.b	$23,$E2,$F8,$8F,$87,$F0,$D2,$3C,$AF,$C6,$F8,$6F,$8F,$F1,$1F,$23
	.dc.b	$E2,$25,$7A,$A3,$9B,$FD,$FB,$83,$FB,$DB,$71,$F9,$F8,$F5,$8D,$63
	.dc.b	$75,$FF,$4C,$DB,$2F,$AB,$F1,$1F,$5B,$E2,$26,$1F,$22,$3E,$CF,$C2
	.dc.b	$48,$F9,$6F,$DC,$F8,$4F,$BB,$F0,$9F,$7B,$E1,$25,$7D,$08,$FC,$0F
	.dc.b	$C4,$48,$FA,$6F,$E1,$3E,$23,$A4,$7A,$F0,$BA,$F9,$B1,$EB,$3F,$6B
	.dc.b	$1E,$AF,$C1,$C7,$F1,$BF,$0F,$23,$EF,$BF,$90,$F8,$8F,$F2,$4C,$3E
	.dc.b	$5B,$E0,$3F,$E3,$21,$C5,$FD,$05,$C3,$5F,$F2,$01,$C7,$83,$9C,$F2
	.dc.b	$8C,$CF,$D2,$CF,$FD,$AF,$1D,$38,$F7,$72,$3C,$17,$C6,$B6,$DF,$6F
	.dc.b	$FC,$D4,$D9,$FF,$42,$B5,$F1,$1C,$AB,$A3,$E6,$8D,$12,$33,$5F,$ED
	.dc.b	$8F,$0F,$FC,$2B,$7F,$E9,$0F,$09,$1F,$E1,$6C,$3F,$C9,$ED,$7F,$C9
	.dc.b	$6F,$07,$C7,$7D,$2B,$CE,$FC,$EC,$9D,$F9,$AD,$9E,$73,$E6,$F5,$FF
	.dc.b	$9E,$30,$7C,$45,$24,$F8,$BE,$23,$5F,$0F,$BF,$7A,$4F,$9A,$39,$7B
	.dc.b	$DC,$FD,$E9,$3E,$A8,$ED,$ED,$C8,$F6,$BE,$1F,$87,$EB,$1F,$03,$0E
	.dc.b	$AF,$C0,$C3,$AB,$F1,$30,$EA,$FC,$4C,$3D,$5F,$0B,$EB,$F8,$59,$3F
	.dc.b	$2A,$3D,$DF,$09,$23,$E6,$BF,$07,$E1,$7E,$17,$C3,$FC,$3F,$87,$95
	.dc.b	$F4,$A3,$E3,$7C,$3C,$8F,$AA,$FC,$9F,$87,$E9,$1F,$5B,$0B,$D5,$F7
	.dc.b	$F1,$EB,$3F,$73,$1E,$AF,$DC,$C7,$E9,$7C,$3F,$D3,$F8,$79,$3F,$82
	.dc.b	$8F,$AD,$F0,$92,$3F,$0A,$FF,$AC,$3F,$16,$FE,$F5,$93,$B7,$F1,$91
	.dc.b	$FD,$EC,$9F,$05,$BF,$83,$E1,$F0,$53,$1F,$F4,$ED,$7F,$7B,$1F,$07
	.dc.b	$DE,$3F,$23,$95,$CB,$B4,$3C,$25,$F8,$B3,$F0,$E7,$E5,$25,$76,$CF
	.dc.b	$D1,$F0,$BE,$9F,$85,$93,$E2,$8E,$D7,$FA,$5F,$01,$A6,$2B,$C1,$97
	.dc.b	$76,$E5,$D9,$FF,$58,$C7,$C1,$CB,$E3,$7C,$2F,$C7,$F8,$59,$5F,$02
	.dc.b	$3E,$57,$C1,$7C,$BF,$85,$93,$F1,$23,$E7,$7C,$1F,$CF,$F8,$39,$3F
	.dc.b	$22,$3E,$97,$C2,$FD,$3F,$85,$93,$F3,$23,$AB,$F3,$B0,$EA,$F9,$B0
	.dc.b	$EA,$FD,$2C,$3A,$BE,$AC,$3E,$F6,$FF,$7F,$E1,$64,$FD,$88,$FC,$17
	.dc.b	$C0,$C8,$FB,$6F,$E1,$7E,$07,$F0,$DF,$03,$F8,$7F,$84,$95,$F8,$08
	.dc.b	$FF,$7C,$FE,$BD,$FE,$05,$24,$63,$E1,$BA,$3F,$33,$0B,$D7,$F8,$2E
	.dc.b	$BF,$C0,$E5,$DA,$BF,$17,$97,$67,$E9,$E5,$DB,$E1,$BB,$FA,$52,$7F
	.dc.b	$DD,$BA,$25,$1F,$EF,$6E,$97,$EA,$9F,$CC,$DB,$EA,$EA,$5A,$FE,$52
	.dc.b	$73,$1C,$6B,$FD,$E9,$3F,$98,$F8,$87,$CD,$7F,$9F,$6B,$BC,$E6,$1E
	.dc.b	$77,$58,$C3,$7C,$0F,$F4,$64,$F1,$AF,$0C,$7A,$31,$F3,$FA,$FB,$94
	.dc.b	$61,$F2,$46,$63,$DB,$F0,$B7,$3F,$71,$34,$C7,$C3,$73,$D7,$D3,$EA
	.dc.b	$5D,$3C,$F9,$FA,$B4,$F2,$EB,$9A,$FB,$BE,$CC,$AF,$9E,$FE,$03,$E2
	.dc.b	$3C,$19,$C9,$F9,$0F,$81,$E8,$7F,$AA,$31,$F8,$1C,$1F,$12,$33,$9F
	.dc.b	$CC,$CB,$C1,$1F,$73,$DC,$FC,$69,$1F,$BF,$C3,$F9,$0E,$0F,$C8,$E5
	.dc.b	$27,$F0,$71,$F9,$4D,$FF,$2B,$F1,$12,$7F,$0F,$1F,$98,$F8,$69,$1F
	.dc.b	$8A,$7F,$37,$F0,$FD,$63,$F1,$B8,$75,$7E,$B9,$FC,$67,$99,$8F,$78
	.dc.b	$FC,$9E,$3D,$DD,$F5,$AB,$F1,$7C,$9F,$DE,$57,$C2,$FE,$AB,$E1,$65
	.dc.b	$FE,$72,$3F,$5B,$F0,$92,$3F,$3E,$FE,$C3,$E1,$7F,$63,$F0,$FF,$B2
	.dc.b	$F8,$79,$5F,$A4,$8F,$DA,$FC,$3C,$8F,$D3,$BF,$B8,$F8,$7E,$91,$E6
	.dc.b	$CD,$8E,$3A,$3F,$6D,$0C,$BB,$D7,$EC,$32,$EE,$F1,$CB,$F3,$1B,$3E
	.dc.b	$CF,$FC,$20,$DF,$8E,$CC,$7F,$85,$72,$FE,$04,$E8,$FF,$05,$DF,$E2
	.dc.b	$2B,$97,$C3,$F0,$38,$FC,$47,$2F,$86,$95,$BB,$47,$6F,$07,$BF,$6C
	.dc.b	$AD,$3E,$2F,$86,$E7,$1C,$71,$EC,$F3,$C7,$AB,$F6,$B1,$EA,$F7,$C7
	.dc.b	$B3,$E1,$F8,$8F,$6F,$C4,$48,$C1,$CC,$EC,$43,$A5,$F6,$D7,$E1,$DD
	.dc.b	$E8,$FF,$1F,$63,$FA,$86,$FC,$3F,$CA,$F8,$89,$8F,$C1,$8F,$9B,$F0
	.dc.b	$D2,$3E,$23,$F4,$3E,$1B,$E8,$FC,$2C,$9F,$90,$FD,$4F,$86,$FA,$BF
	.dc.b	$11,$F5,$BE,$22,$57,$CD,$8E,$8F,$C3,$C2,$ED,$F6,$B1,$E9,$3F,$1F
	.dc.b	$1B,$3F,$7B,$1F,$C0,$FC,$3F,$E0,$BE,$22,$5F,$D9,$8F,$C2,$FC,$24
	.dc.b	$8F,$B8,$FE,$32,$BF,$BC,$74,$9F,$E4,$B3,$F8,$75,$F0,$7B,$38,$7A
	.dc.b	$5C,$7E,$22,$56,$DE,$9F,$11,$DE,$3E,$F6,$37,$4F,$EA,$D8,$F8,$0E
	.dc.b	$F3,$FB,$C7,$BB,$E1,$C7,$BB,$E3,$F8,$7F,$47,$C4,$48,$C1,$FE,$99
	.dc.b	$F9,$BF,$D1,$73,$FF,$58,$FC,$65,$F6,$6E,$EF,$E8,$51,$87,$FD,$47
	.dc.b	$5C,$2F,$FA,$FB,$78,$7E,$27,$C4,$FC,$73,$7E,$24,$AF,$03,$5F,$F5
	.dc.b	$2D,$73,$4D,$24,$6F,$7F,$F5,$2C,$62,$79,$90,$F8,$9F,$15,$F5,$FE
	.dc.b	$1E,$47,$D5,$8E,$5A,$FF,$71,$AE,$53,$FC,$B3,$4F,$83,$E1,$6A,$3D
	.dc.b	$B9,$5D,$3F,$01,$F1,$1F,$83,$F8,$79,$5F,$81,$7F,$0D,$F1,$1F,$87
	.dc.b	$F8,$9F,$C4,$7C,$54,$AF,$C3,$47,$E2,$FE,$27,$F1,$9F,$15,$27,$F1
	.dc.b	$51,$F8,$FF,$89,$91,$F8,$D7,$F2,$5F,$13,$D6,$3F,$21,$87,$57,$F0
	.dc.b	$98,$75,$7F,$29,$87,$57,$F1,$18,$7E,$6F,$E2,$3F,$39,$F1,$32,$7F
	.dc.b	$35,$1F,$CA,$BF,$5F,$FF,$0B,$47,$F4,$A6,$F3,$FF,$8A,$5B,$CB,$FE
	.dc.b	$C3,$5F,$17,$C7,$E2,$B9,$7C,$5C,$C3,$F9,$65,$7F,$89,$DB,$F1,$5D
	.dc.b	$C8,$EB,$E2,$BC,$4C,$38,$9E,$5C,$3B,$3D,$2F,$B3,$E7,$C3,$D5,$F1
	.dc.b	$1E,$BF,$88,$93,$E2,$8F,$77,$C2,$48,$F2,$BF,$07,$E1,$7E,$17,$C4
	.dc.b	$49,$F4,$BF,$17,$E1,$BE,$37,$C4,$49,$F6,$BF,$27,$E2,$3A,$47,$C9
	.dc.b	$C2,$F1,$7C,$9C,$7A,$4F,$CD,$C6,$CF,$CD,$C7,$E9,$7C,$3F,$D3,$F8
	.dc.b	$89,$7F,$2A,$39,$F2,$7D,$1C,$AC,$FF,$A7,$79,$19,$F4,$AF,$AD,$9D
	.dc.b	$AD,$65,$F7,$FE,$1B,$F0,$1F,$0B,$27,$EB,$C7,$E0,$FE,$13,$F0,$9F
	.dc.b	$0B,$27,$ED,$C7,$E1,$FE,$23,$F1,$1F,$11,$27,$EF,$C7,$F7,$4F,$BB
	.dc.b	$FE,$E6,$7F,$07,$26,$9D,$67,$F0,$B8,$75,$7F,$21,$87,$57,$F1,$38
	.dc.b	$75,$7F,$29,$87,$5F,$84,$ED,$F0,$B2,$7F,$BD,$65,$FD,$F1,$FF,$11
	.dc.b	$BF,$DA,$9B,$F0,$BE,$53,$CD,$F1,$1E,$7F,$88,$F4,$7C,$44,$AE,$71
	.dc.b	$EB,$F5,$64,$76,$7D,$DE,$87,$38,$F0,$E1,$5F,$0E,$1D,$5F,$2E,$1D
	.dc.b	$5F,$2E,$1F,$1F,$E1,$BE,$47,$C3,$C9,$F5,$C7,$F9,$BF,$82,$47,$FA
	.dc.b	$1D,$FF,$39,$70,$FF,$9B,$73,$FF,$37,$54,$AF,$F2,$FF,$C5,$A7,$FC
	.dc.b	$F1,$C9,$27,$FD,$20,$FF,$9E,$F9,$7A,$C7,$F9,$43,$E0,$7F,$D3,$79
	.dc.b	$63,$BB,$FD,$31,$F1,$8D,$3C,$17,$DB,$4F,$03,$FE,$58,$F8,$CF,$F4
	.dc.b	$CE,$9F,$83,$F8,$89,$3F,$59,$FC,$37,$C4,$75,$EB,$FF,$4A,$64,$D3
	.dc.b	$17,$FE,$94,$8F,$03,$EC,$C3,$BE,$9F,$EA,$3C,$FC,$73,$F8,$0F,$88
	.dc.b	$FC,$91,$1F,$C2,$FC,$1F,$E5,$7E,$27,$F2,$DF,$15,$27,$F1,$31,$F9
	.dc.b	$9F,$89,$FC,$D7,$C5,$49,$FC,$6C,$7E,$77,$E2,$24,$7E,$41,$FF,$29
	.dc.b	$FE,$73,$FE,$08,$EF,$61,$D6,$BF,$27,$8F,$85,$E6,$7E,$63,$1E,$CF
	.dc.b	$E6,$31,$E7,$CD,$D3,$E2,$64,$FF,$95,$F2,$FF,$02,$EF,$FE,$DD,$5F
	.dc.b	$17,$E2,$24,$EF,$E2,$FC,$BF,$17,$E6,$F8,$B9,$3C,$E3,$D3,$F1,$52
	.dc.b	$3B,$3E,$CF,$89,$E7,$1F,$E8,$EC,$7B,$BE,$3C,$7E,$0E,$3D,$A3,$CD
	.dc.b	$8F,$67,$86,$3F,$1B,$E2,$BE,$3F,$C3,$49,$F6,$47,$CA,$F8,$6F,$97
	.dc.b	$F0,$D2,$7E,$04,$7C,$EF,$83,$91,$F0,$DF,$A3,$F0,$9F,$4B,$E0,$E4
	.dc.b	$FC,$77,$EA,$FC,$27,$58,$F9,$58,$75,$7D,$98,$74,$7D,$D8,$5E,$2F
	.dc.b	$9D,$8F,$DE,$F8,$6F,$BF,$F0,$D2,$FE,$A4,$75,$7E,$36,$3D,$6F,$EB
	.dc.b	$63,$CF,$C3,$53,$E1,$DC,$8D,$35,$39,$86,$9F,$8E,$F8,$6F,$C7,$FC
	.dc.b	$3C,$AF,$C2,$47,$FA,$93,$CC,$FE,$EF,$1D,$5B,$8F,$FA,$C2,$35,$F0
	.dc.b	$BB,$F8,$49,$1F,$EA,$3A,$FE,$F0,$FF,$B3,$17,$C2,$F5,$F8,$7E,$DF
	.dc.b	$0F,$2B,$84,$78,$58,$73,$3B,$9D,$6F,$BB,$D5,$F4,$57,$A7,$E1,$BD
	.dc.b	$5F,$0F,$2B,$C9,$1E,$DF,$83,$91,$E7,$7E,$07,$C1,$7C,$1F,$82,$F8
	.dc.b	$5F,$05,$2B,$D9,$1F,$17,$E1,$E4,$7B,$DF,$91,$F0,$FD,$23,$E0,$61
	.dc.b	$7A,$5D,$1F,$63,$8F,$FD,$B3,$9F,$7B,$F9,$19,$F7,$7E,$46,$7F,$4F
	.dc.b	$E1,$BE,$A7,$C3,$C9,$F9,$D1,$FE,$6E,$F8,$4F,$C1,$C7,$FC,$00,$F9
	.dc.b	$37,$E8,$FF,$EB,$38,$46,$3E,$0F,$84,$7F,$9F,$6B,$8E,$DE,$68,$2F
	.dc.b	$BD,$7B,$1A,$E0,$FC,$8C,$E6,$39,$D7,$F9,$DA,$EC,$FF,$9C,$9B,$5F
	.dc.b	$E6,$B6,$CF,$FA,$11,$AF,$F9,$F1,$BF,$1B,$FD,$CE,$B9,$4F,$E3,$7A
	.dc.b	$FF,$2D,$F0,$32,$30,$7A,$C8,$78,$CF,$3E,$1F,$A8,$D3,$E2,$FC,$27
	.dc.b	$6C,$BC,$78,$FB,$FE,$13,$DD,$87,$EB,$BB,$86,$1A,$D3,$2F,$63,$83
	.dc.b	$92,$39,$38,$95,$E2,$FD,$0C,$7E,$DC,$62,$BE,$A4,$F0,$34,$76,$39
	.dc.b	$47,$97,$F4,$77,$85,$49,$3E,$18,$CB,$63,$2F,$41,$1A,$70,$DA,$F1
	.dc.b	$B6,$FF,$22,$D9,$FE,$2A,$D7,$9D,$7C,$9F,$D1,$1E,$C5,$24,$FC,$58
	.dc.b	$CB,$63,$2F,$96,$32,$D8,$CE,$1F,$1A,$72,$D8,$CE,$1F,$B1,$DF,$FC
	.dc.b	$0D,$7F,$C5,$FE,$07,$F9,$01,$FC,$87,$E1,$7E,$F7,$E8,$8F,$56,$FA
	.dc.b	$D2,$FE,$D4,$E5,$B3,$F8,$8F,$C7,$DC,$79,$9F,$AF,$18,$9E,$A4,$8E
	.dc.b	$24,$3E,$95,$71,$D7,$F0,$2D,$D3,$F7,$ED,$9F,$E8,$6D,$4D,$25,$77
	.dc.b	$8E,$AF,$F3,$FF,$8D,$91,$FC,$E5,$FE,$85,$F1,$DB,$C3,$C3,$57,$0F
	.dc.b	$EA,$3F,$22,$FB,$47,$8B,$9C,$BC,$77,$74,$7E,$BD,$FD,$4A,$7F,$4D
	.dc.b	$E3,$3F,$60,$FC,$EE,$9F,$C4,$63,$1F,$29,$FB,$26,$D7,$0A,$DE,$BF
	.dc.b	$CA,$DF,$E0,$CD,$EC,$E1,$18,$9F,$EC,$CB,$E4,$F3,$F2,$52,$7A,$3C
	.dc.b	$3C,$4E,$3C,$B2,$7C,$0F,$4F,$92,$EB,$8C,$9F,$23,$E0,$C3,$A4,$7E
	.dc.b	$DB,$3F,$76,$38,$D7,$DD,$9F,$F3,$59,$F8,$5F,$27,$E8,$25,$FB,$FE
	.dc.b	$3B,$B3,$CE,$BD,$95,$F2,$72,$F0,$4F,$A8,$86,$F1,$F6,$3F,$27,$C6
	.dc.b	$F8,$BE,$1C,$8C,$1F,$37,$E4,$7E,$49,$1F,$A1,$F2,$3F,$33,$E4,$3E
	.dc.b	$6F,$C8,$49,$FA,$91,$F4,$3E,$3A,$47,$D7,$7E,$9F,$C7,$73,$8F,$56
	.dc.b	$17,$87,$E1,$E3,$D6,$7E,$F6,$3D,$5F,$81,$8F,$DC,$F8,$E9,$1F,$83
	.dc.b	$7E,$FF,$8D,$CF,$A3,$F0,$99,$5A,$FC,$0E,$3D,$5F,$95,$8F,$5C,$BE
	.dc.b	$9E,$5F,$88,$F9,$0F,$C8,$7E,$C2,$FA,$1A,$67,$0F,$CC,$C7,$F9,$AB
	.dc.b	$3F,$EF,$8B,$E4,$38,$9C,$BE,$47,$9F,$91,$2F,$8B,$DB,$C2,$EF,$F1
	.dc.b	$D2,$7A,$BE,$2F,$8F,$E5,$1E,$0C,$3A,$BE,$2C,$3A,$3F,$8F,$C7,$FC
	.dc.b	$DA,$DC,$7B,$3E,$8F,$91,$F6,$FC,$8C,$8C,$1C,$CE,$C4,$3A,$5F,$6E
	.dc.b	$0F,$DB,$61,$D9,$B8,$76,$AF,$17,$C8,$7C,$A2,$3F,$23,$E4,$7F,$B5
	.dc.b	$7E,$8C,$CB,$FC,$1F,$3F,$E5,$BF,$83,$FE,$3B,$B7,$F3,$23,$73,$AF
	.dc.b	$01,$3F,$E6,$2F,$95,$FF,$0F,$F1,$7F,$78,$7F,$C4,$3C,$57,$DD,$36
	.dc.b	$66,$5F,$DD,$EB,$8B,$6B,$F9,$64,$62,$3F,$B2,$67,$E7,$DF,$FC,$75
	.dc.b	$8F,$9A,$69,$FC,$CB,$4F,$F1,$9E,$7E,$68,$FF,$3B,$DF,$17,$DC,$30
	.dc.b	$ED,$87,$F8,$C5,$BC,$9E,$DC,$BC,$67,$F3,$FC,$1C,$2F,$8E,$FC,$BF
	.dc.b	$C7,$CC,$BF,$2B,$1F,$23,$2F,$2C,$7F,$44,$D3,$CB,$AF,$A6,$3C,$63
	.dc.b	$5F,$36,$5E,$9D,$7C,$73,$E7,$D3,$C7,$B7,$E9,$BE,$3F,$F4,$FF,$1D
	.dc.b	$FA,$8F,$93,$97,$FA,$68,$F1,$3D,$BD,$76,$BE,$28,$F9,$07,$9B,$4F
	.dc.b	$0E,$BF,$AC,$F9,$3F,$D9,$92,$7F,$63,$F2,$5F,$B6,$F9,$39,$1F,$B4
	.dc.b	$7F,$73,$F2,$5E,$0E,$1F,$A5,$3F,$04,$F8,$79,$F8,$9F,$07,$47,$E3
	.dc.b	$72,$F0,$47,$EF,$B2,$F0,$3F,$AD,$CB,$F8,$2F,$C0,$7F,$55,$FD,$8B
	.dc.b	$FD,$4A,$7F,$A9,$BA,$3F,$C4,$1B,$7A,$AF,$DA,$7C,$1D,$FE,$03,$F3
	.dc.b	$F3,$F0,$3F,$EB,$BA,$65,$E0,$E6,$FD,$7B,$FC,$53,$3F,$07,$C9,$F8
	.dc.b	$7E,$4A,$61,$FD,$B3,$44,$FC,$1D,$BE,$F7,$07,$C2,$E0,$F8,$17,$F8
	.dc.b	$0D,$FC,$D2,$C3,$FD,$6F,$3C,$FE,$4F,$DC,$7B,$FE,$4F,$E0,$7C,$3C
	.dc.b	$9F,$03,$F0,$F2,$F0,$47,$C0,$C3,$C1,$C5,$F0,$32,$EF,$97,$6A,$FE
	.dc.b	$CC,$B2,$EF,$1E,$BC,$BE,$77,$C9,$FC,$FF,$7E,$4F,$BE,$3B,$BF,$74
	.dc.b	$F9,$C7,$DC,$E2,$FD,$AE,$7E,$DB,$EB,$3E,$0C,$3D,$8F,$C9,$F9,$1F
	.dc.b	$B8,$47,$E6,$7C,$87,$6E,$0F,$93,$87,$B2,$7B,$1F,$BC,$7E,$57,$C8
	.dc.b	$FE,$14,$91,$F5,$7E,$4F,$F1,$1F,$27,$23,$EC,$3F,$8B,$F9,$2E,$DC
	.dc.b	$5F,$86,$C7,$BF,$1F,$FA,$FB,$8F,$E4,$E5,$D3,$1F,$92,$FE,$4F,$2F
	.dc.b	$CA,$7C,$9C,$9F,$C2,$3F,$97,$F8,$DE,$B1,$FA,$2C,$BE,$A7,$67,$E4
	.dc.b	$B6,$F4,$CF,$E2,$F1,$ED,$C7,$FE,$35,$E6,$67,$FA,$1F,$91,$FD,$17
	.dc.b	$C7,$4C,$3F,$25,$1D,$9F,$A6,$7E,$3F,$2F,$CF,$FC,$5B,$DC,$6D,$FA
	.dc.b	$BF,$8E,$97,$FE,$AB,$AD,$4F,$FA,$D1,$B1,$9A,$E2,$D9,$D4,$62,$3F
	.dc.b	$9C,$DF,$EB,$9F,$C3,$6D,$F6,$5F,$26,$BF,$CE,$B0,$EB,$1F,$EA,$E6
	.dc.b	$C7,$FB,$95,$BE,$6F,$F9,$8B,$0E,$D3,$F1,$7E,$3B,$F8,$27,$C7,$4C
	.dc.b	$3F,$69,$1F,$0F,$1F,$01,$FC,$53,$93,$F8,$36,$5D,$28,$D3,$13,$FC
	.dc.b	$13,$2F,$F3,$27,$86,$F8,$BB,$F0,$93,$0F,$F3,$6C,$E5,$FF,$68,$63
	.dc.b	$FC,$4D,$F8,$3D,$1F,$C1,$0F,$C6,$6F,$F8,$1A,$FF,$33,$4F,$2F,$84
	.dc.b	$EE,$78,$3E,$22,$47,$66,$BF,$E5,$F9,$E2,$FD,$0E,$0F,$C7,$30,$F9
	.dc.b	$E7,$E9,$E3,$71,$FE,$81,$6E,$89,$6B,$FC,$92,$FB,$DE,$FE,$2F,$E1
	.dc.b	$FC,$5C,$BF,$83,$1F,$3F,$4F,$74,$7F,$72,$D3,$DD,$87,$2C,$BB,$65
	.dc.b	$D6,$BE,$4F,$C5,$FC,$EF,$8B,$F9,$FF,$17,$27,$E6,$C7,$57,$F8,$F6
	.dc.b	$5F,$2B,$5F,$E1,$FB,$3E,$2B,$EB,$92,$BE,$AF,$C5,$7D,$AF,$87,$91
	.dc.b	$F6,$1F,$BB,$F0,$FD,$63,$EB,$9E,$9C,$3A,$DF,$38,$FA,$0F,$E1,$2F
	.dc.b	$F0,$BF,$0D,$F8,$6F,$87,$95,$F8,$48,$E9,$E3,$7F,$0C,$8E,$87,$D3
	.dc.b	$BE,$98,$7D,$37,$A4,$7E,$33,$B7,$F2,$3F,$15,$23,$07,$E3,$B9,$7F
	.dc.b	$5D,$87,$58,$F4,$9F,$D7,$F0,$F8,$BF,$15,$F9,$AF,$8A,$93,$F9,$88
	.dc.b	$FC,$EF,$C5,$48,$FC,$DB,$FE,$C3,$FC,$5B,$FC,$E9,$1D,$EF,$F1,$9C
	.dc.b	$FF,$AC,$E4,$FA,$53,$F8,$DC,$3A,$BF,$95,$C7,$F5,$31,$CF,$E2,$3A
	.dc.b	$7C,$54,$AF,$F6,$AE,$5F,$E7,$D7,$E2,$E3,$D5,$FB,$E6,$31,$FC,$4F
	.dc.b	$33,$1E,$F9,$FE,$6F,$3E,$F7,$D3,$E2,$7D,$44,$7B,$FC,$17,$D4,$F3
	.dc.b	$BF,$99,$DF,$F5,$8D,$3F,$CB,$35,$E4,$F8,$7F,$84,$7C,$3A,$F8,$9F
	.dc.b	$15,$27,$D2,$FC,$7E,$3E,$91,$FD,$5B,$7F,$F2,$A0,$C3,$AC,$7F,$B5
	.dc.b	$DB,$8F,$FB,$CD,$BD,$FF,$A6,$7F,$C1,$31,$9F,$8A,$FA,$84,$AF,$93
	.dc.b	$F1,$5E,$07,$F7,$F8,$78,$2F,$9C,$7E,$F1,$B1,$E3,$D3,$D9,$3F,$4B
	.dc.b	$E2,$7F,$01,$F1,$52,$30,$7E,$04,$FE,$9B,$5F,$55,$FC,$71,$FD,$CE
	.dc.b	$DF,$11,$F8,$9F,$87,$93,$F7,$A3,$F1,$84,$70,$43,$F8,$D1,$DB,$9B
	.dc.b	$F8,$5B,$DA,$FD,$93,$C6,$3E,$53,$75,$F9,$5F,$11,$F9,$92,$4F,$E3
	.dc.b	$B9,$FE,$35,$FF,$43,$E1,$FE,$E5,$9F,$A2,$BC,$58,$7D,$3C,$7F,$45
	.dc.b	$F0,$9F,$A3,$F8,$89,$5F,$98,$7F,$4D,$F1,$1D,$35,$FD,$1E,$7F,$47
	.dc.b	$E2,$3F,$54,$4B,$FE,$5D,$93,$97,$F9,$1F,$E2,$BF,$CC,$67,$31,$8F
	.dc.b	$87,$C5,$6E,$7F,$9A,$D7,$F9,$41,$D9,$89,$DD,$74,$F8,$8F,$24,$FE
	.dc.b	$A4,$FC,$16,$1E,$37,$F2,$4F,$07,$A9,$71,$61,$FE,$49,$C7,$C5,$18
	.dc.b	$F8,$5D,$4E,$71,$F5,$FC,$2F,$83,$0F,$B1,$A7,$E6,$F2,$F0,$65,$F9
	.dc.b	$F7,$8C,$F2,$F8,$88,$FB,$33,$F0,$C7,$6B,$E9,$EA,$FF,$9A,$B3,$7C
	.dc.b	$3F,$88,$98,$71,$87,$C4,$79,$BE,$22,$4E,$5F,$4F,$C4,$74,$E8,$FF
	.dc.b	$1A,$5F,$0A,$E5,$87,$45,$B7,$C2,$9D,$46,$38,$3E,$DF,$C2,$FC,$42
	.dc.b	$47,$8F,$E1,$3B,$5F,$F9,$9A,$F8,$BC,$F0,$EC,$F5,$D3,$E6,$FC,$47
	.dc.b	$CE,$F8,$89,$8F,$B6,$39,$BF,$E6,$3C,$1F,$11,$F5,$09,$1F,$0F,$E1
	.dc.b	$3E,$BF,$C3,$48,$F8,$CF,$DA,$F8,$7E,$7C,$DF,$C6,$30,$E7,$5F,$2F
	.dc.b	$0E,$6F,$BC,$FC,$9E,$1F,$82,$F8,$7F,$C1,$FC,$34,$9F,$A5,$1D,$39
	.dc.b	$FF,$3F,$5C,$B7,$F9,$58,$7E,$2A,$3F,$17,$F0,$F3,$0F,$B4,$FE,$4F
	.dc.b	$F6,$3F,$D1,$47,$F0,$6D,$3F,$89,$7A,$DF,$D8,$7A,$BF,$D5,$B5,$FE
	.dc.b	$33,$AF,$35,$21,$F0,$DC,$CE,$9F,$0F,$2B,$F2,$13,$C7,$4F,$CF,$B7
	.dc.b	$A7,$FC,$32,$D9,$FD,$61,$FD,$C3,$2F,$17,$57,$F9,$F6,$75,$97,$87
	.dc.b	$1E,$D7,$E4,$9C,$9F,$E3,$27,$FC,$62,$D9,$F4,$7C,$27,$C1,$F8,$7F
	.dc.b	$95,$5A,$BE,$79,$F9,$9F,$8D,$F0,$DF,$BB,$F8,$C7,$D4,$BC,$EC,$37
	.dc.b	$0F,$E1,$33,$A9,$C9,$D9,$C4,$6A,$78,$3F,$81,$8C,$B6,$33,$87,$F9
	.dc.b	$E7,$CD,$FE,$2D,$3F,$E7,$D6,$C6,$7C,$9F,$F4,$74,$63,$87,$F9,$4F
	.dc.b	$3F,$F4,$0E,$2F,$97,$7F,$DA,$E3,$F9,$30,$F2,$7F,$DF,$AD,$C7,$47
	.dc.b	$CA,$F8,$9F,$C4,$11,$FB,$AF,$CA,$7F,$8A,$9F,$E6,$18,$E4,$FF,$98
	.dc.b	$9E,$F3,$FE,$06,$8F,$F6,$87,$1B,$E2,$7F,$28,$48,$FC,$4B,$D7,$9F
	.dc.b	$FD,$05,$7D,$39,$4F,$4A,$F3,$FB,$CF,$D5,$3F,$BB,$B7,$D6,$FE,$68
	.dc.b	$D9,$FE,$7E,$D7,$F4,$55,$F1,$FD,$AF,$36,$1F,$D7,$76,$FF,$37,$CF
	.dc.b	$7A,$F6,$1E,$FD,$77,$1F,$42,$3F,$D0,$59,$BD,$AF,$EA,$98,$7F,$AA
	.dc.b	$5D,$CF,$E8,$38,$3F,$CA,$07,$D6,$7F,$C2,$37,$DE,$BD,$F1,$DC,$FA
	.dc.b	$A7,$D0,$17,$DD,$F3,$9E,$2B,$EE,$EE,$FB,$E5,$8A,$F9,$B8,$FE,$08
	.dc.b	$C7,$5F,$FB,$46,$3A,$5F,$C0,$3B,$E1,$AC,$FF,$CB,$79,$FF,$83,$AB
	.dc.b	$FD,$70,$74,$AF,$19,$F6,$30,$E0,$F3,$8E,$B9,$FC,$58,$F6,$C7,$FB
	.dc.b	$07,$3F,$3C,$63,$2F,$F7,$26,$1F,$83,$3E,$31,$F4,$63,$E2,$57,$48
	.dc.b	$FF,$02,$36,$3F,$87,$BD,$73,$FA,$B3,$E9,$AF,$80,$76,$C7,$E2,$5F
	.dc.b	$87,$2F,$5E,$1C,$E3,$E1,$99,$3F,$C6,$07,$F9,$C0,$F1,$D7,$37,$E1
	.dc.b	$DF,$A5,$F1,$C7,$CF,$CF,$C7,$1F,$64,$E9,$9F,$F3,$6C,$BE,$4B,$F8
	.dc.b	$48,$F9,$87,$4C,$B3,$87,$C0,$BE,$73,$E6,$3F,$1A,$7E,$46,$F3,$9F
	.dc.b	$E4,$70,$FD,$04,$FD,$FB,$F3,$3D,$8E,$27,$CE,$BE,$B9,$7C,$33,$F0
	.dc.b	$39,$7F,$AB,$31,$EA,$79,$2B,$F3,$47,$5C,$B0,$F9,$71,$FC,$13,$DA
	.dc.b	$FE,$B3,$F8,$E3,$E3,$E7,$EF,$61,$BC,$3E,$53,$F5,$B1,$EB,$9F,$E2
	.dc.b	$A3,$55,$C3,$0F,$95,$9F,$BB,$1F,$4C,$F1,$A8,$70,$3E,$E9,$FE,$C6
	.dc.b	$8D,$BE,$B3,$A4,$FE,$4E,$3F,$47,$87,$C6,$3C,$79,$FC,$4A,$FB,$B5
	.dc.b	$F9,$A7,$E2,$E1,$F4,$89,$1F,$6E,$3C,$71,$87,$ED,$1F,$14,$FC,$7E
	.dc.b	$1E,$27,$9C,$F8,$F2,$F8,$B1,$FB,$1C,$3C,$6F,$1C,$7D,$78,$7C,$4C
	.dc.b	$FF,$43,$3F,$06,$3E,$A9,$30,$E9,$8F,$87,$3F,$C5,$47,$EC,$CF,$C4
	.dc.b	$9F,$37,$2F,$D6,$CF,$AA,$3D,$97,$D5,$FE,$64,$FC,$43,$D2,$43,$DA
	.dc.b	$7A,$4E,$47,$E1,$0F,$AB,$5E,$27,$FC,$1B,$1E,$2C,$3F,$1F,$8F,$E5
	.dc.b	$F2,$F8,$35,$F7,$23,$E0,$D7,$D3,$8F,$D0,$65,$F3,$DF,$D0,$1E,$03
	.dc.b	$C6,$FE,$D5,$F4,$1C,$23,$94,$F4,$8F,$DE,$1E,$93,$E0,$1F,$C1,$F0
	.dc.b	$F4,$3F,$CF,$67,$D0,$F4,$9F,$43,$DA,$7F,$39,$5E,$59,$FE,$61,$3F
	.dc.b	$93,$3A,$9F,$AD,$AF,$9F,$3B,$CB,$F6,$D3,$F8,$17,$EA,$BC,$63,$F0
	.dc.b	$4F,$87,$0F,$CD,$3E,$1C,$FF,$34,$F6,$BF,$3B,$E0,$C3,$CE,$F8,$B1
	.dc.b	$F3,$CF,$E2,$23,$B3,$F1,$B1,$F0,$E3,$F8,$D9,$F8,$EF,$9F,$3F,$C7
	.dc.b	$63,$FC,$22,$FB,$CE,$6B,$E6,$3C,$5F,$AB,$1F,$8C,$7C,$D9,$FE,$72
	.dc.b	$78,$1B,$CB,$F8,$D6,$3E,$2B,$E1,$3A,$7F,$08,$7E,$6F,$1F,$EA,$F9
	.dc.b	$FC,$4C,$7D,$D9,$7C,$4B,$E1,$1F,$99,$BE,$0F,$F1,$DA,$E1,$3A,$8F
	.dc.b	$F3,$15,$7B,$72,$F3,$5F,$C7,$AF,$35,$FE,$61,$FE,$85,$9F,$E2,$33
	.dc.b	$FC,$0C,$7D,$8C,$7E,$95,$76,$BE,$47,$B0,$B1,$E5,$C3,$F4,$B3,$B8
	.dc.b	$FD,$65,$7E,$75,$F9,$27,$F0,$D3,$E6,$9F,$5A,$BF,$1A,$74,$7D,$67
	.dc.b	$E9,$1F,$AE,$66,$FF,$0B,$3F,$07,$0F,$C2,$D7,$82,$BE,$7E,$1E,$48
	.dc.b	$FD,$85,$FE,$AD,$EA,$79,$0F,$C5,$9C,$4F,$21,$0F,$39,$C2,$FE,$85
	.dc.b	$79,$23,$81,$FA,$73,$84,$F7,$BF,$2C,$72,$9E,$0F,$49,$F5,$C7,$C1
	.dc.b	$9D,$C7,$9E,$36,$FA,$4D,$CF,$82,$77,$9F,$E4,$63,$EA,$1F,$B1,$3E
	.dc.b	$0B,$97,$D5,$5F,$3C,$EB,$19,$AE,$B3,$F0,$2B,$AE,$7F,$E4,$AB,$DB
	.dc.b	$F9,$A9,$EA,$FF,$93,$A7,$71,$FE,$00,$9D,$BF,$99,$36,$7F,$81,$23
	.dc.b	$78,$7F,$27,$9D,$BF,$C5,$72,$FC,$65,$78,$F3,$FD,$35,$7A,$73,$FC
	.dc.b	$FC,$FD,$2C,$BD,$27,$D1,$AD,$C7,$D6,$AE,$A7,$F6,$FB,$EB,$5B,$9F
	.dc.b	$E3,$99,$7B,$F0,$EB,$1F,$5E,$B7,$9F,$CB,$CF,$D3,$8F,$F5,$9B,$DE
	.dc.b	$3F,$5E,$F7,$8F,$C5,$CB,$C7,$97,$E0,$EF,$79,$7D,$1B,$DC,$FC,$5A
	.dc.b	$DE,$1F,$83,$BD,$E1,$F4,$6B,$F8,$CE,$3F,$89,$BD,$E3,$F7,$E3,$DF
	.dc.b	$3F,$E5,$AC,$7A,$E7,$FD,$57,$3F,$4C,$FF,$60,$CB,$AB,$F5,$CE,$A7
	.dc.b	$E3,$F1,$EB,$97,$C5,$8F,$DE,$67,$EF,$7E,$EE,$1D,$6B,$71,$F8,$3B
	.dc.b	$EB,$3F,$76,$37,$97,$F9,$CF,$1F,$1C,$7F,$5B,$AD,$CF,$D7,$8D,$9F
	.dc.b	$E1,$A3,$FB,$96,$7E,$F9,$DC,$7B,$CD,$9F,$E9,$BC,$3C,$78,$7E,$3E
	.dc.b	$77,$3F,$DD,$8F,$79,$F8,$F3,$F0,$67,$A4,$F7,$D7,$D7,$7A,$E3,$FE
	.dc.b	$50,$7C,$79,$7C,$57,$F8,$69,$F8,$93,$73,$FD,$A9,$D9,$D4,$DC,$FC
	.dc.b	$B7,$C6,$FE,$55,$DD,$FA,$5F,$1D,$7E,$0E,$3F,$AA,$57,$E5,$70,$F7
	.dc.b	$E3,$F9,$5B,$D9,$F7,$6B,$77,$F9,$58,$F7,$C7,$57,$C7,$7E,$F7,$C6
	.dc.b	$F5,$77,$5D,$5D,$D7,$8E,$3F,$9E,$C7,$E2,$4D,$9F,$8F,$3C,$75,$F8
	.dc.b	$37,$AC,$F8,$DF,$7C,$F5,$9F,$EA,$C7,$F6,$1C,$3F,$7F,$7E,$3A,$D9
	.dc.b	$D4,$F1,$9E,$FA,$F1,$BF,$74,$D9,$F8,$F3,$C6,$FE,$3D,$F1,$9F,$B3
	.dc.b	$3D,$26,$CE,$B5,$E9,$3E,$5B,$B3,$AD,$7F,$6E,$9F,$F0,$C5,$FC,$59
	.dc.b	$FE,$8F,$7E,$9B,$FE,$EB,$87,$F7,$B9,$F1,$9F,$9B,$3C,$71,$FA,$D7
	.dc.b	$C7,$87,$57,$FC,$C0,$FE,$26,$3C,$66,$E3,$C7,$5F,$EC,$D7,$D3,$87
	.dc.b	$BE,$76,$FE,$CE,$FF,$13,$3F,$2E,$7E,$8C,$FF,$73,$CF,$F8,$5B,$B9
	.dc.b	$F1,$DF,$5A,$DC,$7F,$35,$AD,$BF,$B9,$8F,$7B,$FD,$FF,$2F,$D9,$E3
	.dc.b	$FE,$A2,$8F,$D3,$E5,$FE,$3A,$74,$7C,$43,$F7,$15,$E8,$7A,$63,$FE
	.dc.b	$F5,$9F,$D9,$4F,$8A,$7F,$64,$7C,$A9,$F4,$3F,$12,$7F,$C2,$D1,$FE
	.dc.b	$30,$9F,$B8,$FE,$22,$BF,$65,$1E,$8A,$FF,$05,$3D,$63,$E8,$4E,$F1
	.dc.b	$FB,$8F,$8E,$BF,$7D,$5B,$9F,$D6,$3E,$98,$FE,$2C,$7C,$AB,$F1,$E1
	.dc.b	$FE,$1B,$8F,$41,$EE,$07,$FB,$36,$BF,$91,$3F,$17,$0F,$94,$F0,$3A
	.dc.b	$9E,$F3,$F8,$AE,$1E,$F3,$F0,$47,$A6,$FF,$65,$87,$BC,$FF,$10,$E1
	.dc.b	$F5,$B0,$FF,$62,$C7,$A7,$0F,$E4,$4E,$EF,$FB,$B6,$1E,$3A,$DC,$75
	.dc.b	$8F,$D3,$4E,$DF,$EE,$D7,$D7,$0F,$F1,$CE,$3F,$E5,$9B,$F9,$57,$FC
	.dc.b	$54,$F9,$53,$FA,$BB,$FF,$19,$1F,$BD,$8E,$8F,$F6,$F3,$C5,$1D,$1F
	.dc.b	$F4,$A1,$F3,$CF,$87,$7F,$3E,$7D,$98,$7F,$37,$79,$E3,$FD,$CE,$B9
	.dc.b	$9F,$CD,$B1,$F3,$3F,$53,$3F,$F3,$75,$FE,$8F,$2F,$E0,$33,$FC,$4C
	.dc.b	$FE,$03,$97,$CE,$8F,$69,$FC,$7C,$DE,$3F,$92,$9D,$9F,$22,$36,$FE
	.dc.b	$D5,$D9,$F6,$9D,$C7,$F0,$13,$D3,$5F,$86,$8F,$E0,$27,$F0,$D8,$F7
	.dc.b	$D7,$EB,$E3,$F3,$B5,$D1,$D3,$F6,$67,$A3,$E2,$8F,$E4,$91,$F5,$1F
	.dc.b	$71,$F5,$31,$FD,$4C,$ED,$FE,$31,$1F,$EE,$13,$F8,$31,$FC,$00,$FF
	.dc.b	$7A,$E7,$FC,$8D,$F3,$9F,$C0,$27,$C4,$7E,$EA,$BD,$B7,$FC,$02,$3D
	.dc.b	$13,$B3,$E1,$BB,$3C,$58,$7D,$58,$FE,$52,$FE,$D2,$3A,$C7,$EB,$CF
	.dc.b	$D7,$4F,$57,$6F,$E4,$70,$FF,$61,$BF,$6E,$3F,$05,$87,$A7,$0F,$88
	.dc.b	$7D,$B8,$F8,$98,$7F,$06,$76,$7F,$1E,$3D,$F7,$FB,$A9,$FA,$B1,$F1
	.dc.b	$0D,$3D,$30,$FC,$04,$74,$C3,$C4,$FA,$0F,$E2,$31,$F1,$2B,$A3,$EE
	.dc.b	$8F,$E0,$B5,$F2,$9D,$9E,$E3,$ED,$CF,$B9,$FC,$91,$E8,$3F,$0E,$6E
	.dc.b	$3D,$D1,$B7,$F0,$F1,$B7,$F8,$5B,$FA,$A7,$FD,$AE,$FE,$D0,$FE,$2C
	.dc.b	$FF,$39,$9F,$E2,$D1,$F9,$23,$ED,$9F,$57,$1F,$D7,$DF,$F2,$88,$FF
	.dc.b	$51,$1F,$C8,$6F,$F8,$AD,$7E,$02,$75,$3F,$EE,$03,$F2,$17,$FE,$D9
	.dc.b	$3C,$AF,$D2,$C3,$F8,$CB,$E0,$3E,$3D,$73,$7F,$84,$9E,$09,$FA,$E7
	.dc.b	$38,$FE,$47,$1C,$EB,$F8,$92,$9F,$F6,$E6,$3F,$DC,$23,$F5,$75,$F5
	.dc.b	$EF,$C3,$1F,$74,$FA,$27,$DD,$AF,$E4,$93,$BB,$EB,$1D,$27,$F8,$F3
	.dc.b	$F2,$F3,$FE,$3D,$5F,$81,$C7,$E2,$D7,$D5,$8F,$3D,$7E,$36,$3E,$81
	.dc.b	$E1,$7E,$DB,$E1,$8F,$E3,$0F,$E0,$5F,$C9,$CF,$E0,$5F,$D8,$47,$53
	.dc.b	$F9,$F9,$F8,$17,$FB,$76,$7F,$CA,$A3,$A1,$FC,$74,$E8,$78,$A7,$A3
	.dc.b	$F8,$1C,$BF,$5D,$5F,$06,$3F,$9A,$1F,$8B,$BF,$CE,$1F,$19,$E6,$7F
	.dc.b	$05,$3D,$91,$E6,$7C,$27,$D2,$37,$5F,$A8,$7F,$04,$FC,$87,$6F,$F1
	.dc.b	$C8,$FC,$BC,$75,$8F,$3B,$F8,$83,$E1,$1F,$D5,$4F,$C2,$1E,$83,$F0
	.dc.b	$91,$FE,$BD,$7E,$CB,$FE,$00,$7F,$94,$12,$3F,$C2,$2F,$93,$2F,$EB
	.dc.b	$13,$F9,$5A,$FC,$04,$FE,$56,$3F,$01,$3A,$8F,$1D,$7F,$5D,$7C,$75
	.dc.b	$FB,$47,$C6,$7F,$9E,$CF,$DA,$D7,$8C,$FF,$67,$BE,$FC,$3E,$D6,$7F
	.dc.b	$B4,$8F,$D2,$9F,$AF,$3F,$69,$87,$85,$F9,$2F,$F3,$37,$D8,$FF,$33
	.dc.b	$7D,$99,$FF,$55,$C3,$F6,$A7,$E5,$63,$C0,$7D,$FC,$3E,$01,$E5,$CF
	.dc.b	$FC,$19,$1F,$36,$7F,$48,$FA,$8F,$F1,$5C,$FA,$B2,$FB,$D1,$F8,$43
	.dc.b	$F1,$C7,$F3,$08,$FC,$74,$FC,$53,$99,$F8,$E7,$F4,$71,$EA,$BF,$E8
	.dc.b	$D9,$7C,$4C,$FF,$2F,$3F,$8C,$CB,$F7,$B8,$FE,$18,$FC,$66,$7F,$27
	.dc.b	$2F,$B3,$5F,$25,$FE,$1B,$97,$E0,$E7,$FD,$D3,$5F,$91,$CF,$F7,$B7
	.dc.b	$F2,$0F,$E9,$07,$F3,$63,$C8,$7F,$3C,$CB,$C8,$FB,$4F,$26,$7F,$C2
	.dc.b	$B0,$FC,$94,$7C,$73,$F2,$2F,$E7,$AF,$E3,$9E,$9C,$BE,$E6,$3E,$99
	.dc.b	$FA,$55,$FB,$F3,$FC,$B3,$5F,$EB,$AB,$FE,$AF,$5B,$9F,$A7,$8F,$D8
	.dc.b	$CB,$E7,$61,$E2,$7E,$74,$FF,$05,$3D,$6F,$F5,$0C,$BD,$71,$EF,$9F
	.dc.b	$36,$5F,$3A,$73,$5F,$8A,$9F,$B6,$7F,$12,$AF,$C5,$1F,$A4,$AF,$F1
	.dc.b	$44,$FE,$3F,$2F,$7D,$F2,$BF,$E4,$E7,$2B,$F6,$C7,$FA,$81,$FE,$00
	.dc.b	$7E,$7B,$2F,$EA,$37,$FE,$DB,$9D,$5F,$E0,$33,$FC,$2D,$FB,$30,$FE
	.dc.b	$57,$5D,$E3,$F9,$49,$FB,$68,$F4,$3F,$B6,$3F,$CC,$4F,$E8,$9F,$96
	.dc.b	$7D,$13,$FC,$69,$1C,$DF,$E2,$57,$ED,$C7,$E3,$47,$C2,$3D,$07,$EE
	.dc.b	$A3,$D0,$FF,$68,$AF,$C9,$5F,$F5,$9C,$3F,$25,$3F,$C3,$CF,$C9,$57
	.dc.b	$F6,$37,$D2,$F7,$C7,$F1,$D1,$FC,$36,$3E,$EB,$FC,$C4,$F8,$47,$F4
	.dc.b	$1C,$7D,$15,$FC,$5E,$79,$4F,$EF,$70,$FC,$31,$F0,$5F,$CB,$E5,$FA
	.dc.b	$47,$F7,$87,$D4,$CF,$F2,$0E,$8F,$41,$AB,$FA,$8F,$E9,$4F,$19,$FC
	.dc.b	$40,$FA,$0F,$F9,$79,$FC,$46,$3F,$E4,$77,$71,$F8,$87,$E6,$4F,$82
	.dc.b	$FF,$88,$67,$F4,$27,$9B,$FE,$45,$7E,$8E,$3F,$C0,$8F,$A2,$FF,$2A
	.dc.b	$C3,$F9,$25,$7E,$4A,$3F,$72,$7C,$33,$E6,$3F,$0C,$EF,$1F,$42,$3F
	.dc.b	$0E,$7A,$0F,$F2,$09,$E8,$AF,$25,$7F,$1E,$AE,$58,$FE,$D5,$FC,$8C
	.dc.b	$66,$B9,$3F,$DB,$DE,$57,$DF,$1F,$8D,$1E,$CA,$FE,$0A,$FE,$6C,$FD
	.dc.b	$49,$F8,$D3,$FB,$3E,$3F,$1F,$0F,$04,$68,$F8,$47,$F1,$EC,$3F,$78
	.dc.b	$FC,$27,$F8,$4B,$F0,$8F,$2D,$7F,$36,$3F,$82,$E1,$E5,$C7,$E8,$61
	.dc.b	$DE,$BF,$51,$8F,$F1,$77,$FC,$63,$19,$7D,$07,$EB,$0F,$BB,$9F,$EE
	.dc.b	$B0,$EB,$1F,$4C,$F9,$E7,$C1,$3F,$6A,$72,$AE,$D8,$FF,$8B,$AF,$B3
	.dc.b	$F9,$E3,$D7,$3C,$A3,$35,$FC,$46,$3C,$47,$E2,$B1,$FE,$5B,$87,$8F
	.dc.b	$2F,$D1,$63,$FC,$DE,$7F,$13,$5F,$B4,$BE,$53,$E2,$7F,$C4,$98,$76
	.dc.b	$9F,$99,$1F,$80,$3F,$5E,$FC,$C3,$F9,$A3,$F0,$31,$F3,$D7,$E0,$2B
	.dc.b	$95,$7E,$10,$F8,$91,$9C,$7C,$D1,$FB,$28,$F3,$3F,$76,$FF,$3C,$F1
	.dc.b	$BE,$B9,$7D,$8B,$E3,$3F,$D9,$EB,$FC,$3F,$8F,$BA,$3F,$8F,$C7,$BB
	.dc.b	$3F,$E6,$37,$FD,$62,$F8,$E1,$FA,$E3,$FB,$7D,$FE,$4A,$33,$7F,$E3
	.dc.b	$3C,$7F,$C6,$33,$FC,$1E,$F9,$3F,$C1,$E7,$E1,$1F,$1D,$FC,$91,$F8
	.dc.b	$13,$F8,$81,$E7,$3F,$93,$BF,$CB,$1F,$F2,$09,$F8,$6B,$FC,$14,$73
	.dc.b	$8F,$C1,$5F,$91,$FF,$75,$9F,$A4,$9F,$E6,$0F,$C0,$3E,$C9,$FA,$97
	.dc.b	$DA,$B2,$FF,$05,$DF,$F8,$CE,$FD,$33,$B7,$E5,$BF,$4C,$FC,$41,$D0
	.dc.b	$FC,$8D,$7F,$03,$7D,$D1,$FC,$41,$FB,$F3,$F5,$DF,$8D,$9F,$E5,$32
	.dc.b	$FE,$7B,$3F,$DC,$2B,$EB,$57,$F0,$33,$CD,$8F,$F4,$F9,$F3,$5F,$19
	.dc.b	$FD,$89,$D0,$FC,$55,$F4,$CF,$F9,$A6,$1E,$C9,$F3,$BF,$9A,$7C,$F5
	.dc.b	$FC,$39,$C9,$F6,$32,$FC,$CC,$F1,$CF,$F4,$D3,$FE,$46,$3F,$C2,$15
	.dc.b	$E3,$7F,$64,$7F,$08,$CB,$EB,$29,$FC,$49,$F9,$EC,$FE,$BB,$FD,$76
	.dc.b	$FE,$0C,$FD,$B7,$F0,$8F,$F8,$CD,$F9,$93,$FA,$2C,$7F,$7F,$3F,$04
	.dc.b	$FC,$61,$F9,$57,$F4,$55,$D6,$3E,$C3,$EA,$7F,$B2,$5F,$9A,$3C,$B9
	.dc.b	$7C,$99,$FE,$8B,$87,$C1,$8F,$D4,$DF,$4C,$FE,$C5,$74,$9C,$BF,$87
	.dc.b	$3F,$24,$7E,$E6,$BF,$35,$7F,$C0,$63,$D1,$8F,$EC,$67,$F1,$EF,$C4
	.dc.b	$BE,$B3,$C4,$FC,$A1,$FA,$83,$F6,$C7,$29,$FD,$B4,$7F,$A2,$E3,$E1
	.dc.b	$1E,$3A,$F8,$4F,$33,$EF,$9F,$AC,$39,$1F,$AC,$AF,$D3,$E7,$FA,$6C
	.dc.b	$7F,$C7,$78,$76,$7F,$CF,$F5,$F9,$23,$F0,$67,$E4,$8C,$57,$F5,$67
	.dc.b	$C1,$7F,$44,$FF,$07,$BC,$0F,$B2,$FE,$16,$BF,$03,$3F,$E2,$68,$FF
	.dc.b	$17,$CE,$B1,$FA,$D3,$FD,$C6,$BE,$B4,$FE,$A7,$0F,$43,$FE,$3B,$8F
	.dc.b	$43,$FE,$83,$3C,$AF,$F2,$E3,$8B,$F9,$13,$E6,$9E,$59,$FA,$CF,$D1
	.dc.b	$3F,$6A,$7F,$18,$3F,$8E,$BA,$7E,$8B,$F6,$F0,$FD,$9E,$7F,$A7,$8F
	.dc.b	$E1,$17,$FA,$E8,$C5,$FE,$D6,$31,$7F,$2C,$FD,$6D,$7F,$67,$CF,$F3
	.dc.b	$19,$FD,$FC,$FE,$45,$7A,$2B,$E8,$61,$EC,$7E,$DD,$FF,$0B,$CB,$ED
	.dc.b	$D7,$DF,$AF,$C8,$9F,$40,$FF,$10,$BF,$E0,$4C,$FF,$3B,$7E,$B8,$FD
	.dc.b	$2E,$7E,$87,$F8,$46,$1E,$4C,$3F,$0B,$1E,$8C,$3F,$B6,$CE,$63,$E9
	.dc.b	$19,$3F,$86,$5F,$A2,$FA,$47,$E5,$4F,$A5,$5F,$6A,$7F,$D4,$51,$F6
	.dc.b	$8F,$E6,$A7,$E3,$B0,$F3,$CF,$AE,$3F,$BC,$1F,$E6,$F7,$E2,$1F,$01
	.dc.b	$F9,$0F,$E0,$EB,$C3,$8F,$ED,$E7,$C2,$7C,$73,$F0,$59,$FE,$3B,$0F
	.dc.b	$F2,$C5,$F9,$4F,$F4,$DC,$F9,$4F,$E0,$6F,$81,$E3,$8F,$D3,$7F,$2C
	.dc.b	$75,$3F,$13,$1E,$97,$F9,$B9,$F4,$CF,$CF,$47,$19,$F0,$C7,$1B,$FC
	.dc.b	$1B,$F9,$63,$F9,$56,$1F,$CF,$E3,$F6,$59,$FF,$7A,$9F,$69,$FC,$DD
	.dc.b	$FC,$66,$5F,$AE,$C7,$FC,$95,$7F,$12,$7F,$AE,$57,$0B,$FE,$52,$FB
	.dc.b	$DF,$E1,$F5,$EF,$8F,$59,$F3,$8F,$AE,$E6,$3C,$F9,$7C,$B8,$FF,$33
	.dc.b	$E3,$FE,$02,$AF,$E4,$27,$23,$D8,$FF,$05,$BF,$EC,$B1,$FA,$78,$FD
	.dc.b	$6E,$5F,$8F,$7F,$0F,$1F,$8F,$3F,$A0,$9F,$C8,$0F,$81,$9F,$E6,$23
	.dc.b	$8C,$67,$1F,$6B,$C5,$FE,$33,$1F,$97,$8F,$06,$1F,$B9,$BF,$03,$F8
	.dc.b	$28,$E1,$87,$CF,$AE,$11,$FC,$6C,$F9,$11,$FD,$02,$7E,$26,$7F,$46
	.dc.b	$7F,$C9,$F9,$FD,$1C,$BF,$B0,$C7,$D1,$BF,$AE,$F2,$8F,$E1,$EF,$90
	.dc.b	$FD,$A9,$EB,$8D,$BE,$48,$FE,$B8,$70,$7C,$86,$4F,$E9,$51,$E5,$9F
	.dc.b	$14,$7F,$7F,$3A,$4F,$C9,$BF,$85,$18,$3F,$7F,$87,$CC,$7D,$13,$B7
	.dc.b	$CA,$60,$DC,$F2,$71,$3F,$68,$E4,$7C,$A3,$F9,$86,$5F,$8A,$CB,$F4
	.dc.b	$95,$F8,$A3,$FC,$1B,$3F,$C8,$AB,$F9,$C4,$7F,$83,$0F,$EE,$B8,$FE
	.dc.b	$72,$7A,$47,$EE,$E7,$FD,$E8,$7C,$D3,$F9,$09,$F3,$4F,$E2,$D1,$C1
	.dc.b	$FD,$96,$5F,$65,$FE,$21,$97,$C8,$9F,$E7,$51,$F1,$5F,$65,$7B,$DE
	.dc.b	$51,$F9,$43,$E1,$9F,$19,$CE,$3F,$94,$7E,$1C,$FF,$19,$9F,$24,$FF
	.dc.b	$19,$9D,$E1,$F8,$07,$73,$FB,$2B,$F2,$5F,$F3,$68,$F2,$3F,$99,$9D
	.dc.b	$C7,$F1,$A9,$DB,$FC,$E2,$7C,$0F,$E2,$27,$77,$FC,$0A,$B7,$3F,$B0
	.dc.b	$3E,$DD,$75,$3E,$01,$CA,$BE,$DE,$1F,$26,$FF,$C0,$31,$8A,$FA,$4F
	.dc.b	$F8,$06,$FA,$CF,$ED,$2F,$91,$F6,$6B,$E5,$57,$F9,$25,$FC,$C1,$E2
	.dc.b	$3E,$E9,$F9,$F7,$FA,$F1,$D5,$F9,$78,$7E,$16,$7A,$C7,$DE,$3C,$6F
	.dc.b	$F2,$67,$79,$7E,$DA,$BE,$8B,$F9,$F8,$FC,$3C,$F2,$7C,$AF,$C3,$7E
	.dc.b	$F3,$9B,$FB,$CF,$C6,$8F,$B0,$76,$3D,$75,$D9,$FC,$39,$D8,$D9,$F5
	.dc.b	$63,$93,$F6,$CE,$51,$F9,$47,$E2,$1E,$43,$F9,$8C,$FE,$B5,$FD,$4E
	.dc.b	$1F,$D2,$F1,$FE,$DF,$1F,$CF,$4E,$87,$73,$D6,$78,$AB,$ED,$3F,$3E
	.dc.b	$FF,$AF,$67,$FE,$20,$8E,$07,$A4,$E7,$3C,$0F,$C6,$E5,$FC,$4E,$3D
	.dc.b	$35,$F9,$97,$D3,$8A,$59,$27,$FD,$6A,$FD,$5A,$F9,$D3,$FA,$57,$E2
	.dc.b	$D7,$E6,$6B,$A9,$F9,$F3,$D5,$1F,$8D,$8E,$D5,$D6,$FF,$1C,$7C,$B3
	.dc.b	$EA,$1F,$BE,$AF,$86,$7B,$2B,$95,$F9,$A3,$CA,$7D,$C8,$F2,$D7,$E2
	.dc.b	$6F,$F7,$91,$89,$FB,$F1,$87,$CA,$7E,$54,$F9,$C7,$D5,$7D,$8F,$CE
	.dc.b	$78,$3E,$33,$85,$FC,$EA,$FE,$88,$FD,$1A,$F8,$B7,$F4,$6B,$93,$FA
	.dc.b	$57,$C2,$7E,$24,$97,$FE,$10,$BF,$E9,$F1,$FC,$A8,$E2,$FD,$37,$8C
	.dc.b	$7E,$81,$F8,$EF,$A2,$FC,$51,$8A,$F1,$46,$23,$F1,$13,$BB,$F9,$B7
	.dc.b	$F5,$5F,$E7,$EF,$79,$F8,$07,$9A,$FE,$95,$FD,$AA,$FC,$63,$F6,$9F
	.dc.b	$DB,$4F,$69,$F0,$C7,$68,$FB,$33,$F8,$18,$FA,$C9,$FF,$49,$C7,$E4
	.dc.b	$CF,$CA,$5F,$E2,$9C,$18,$AF,$C2,$38,$8D,$36,$32,$F4,$21,$C6,$F9
	.dc.b	$B4,$B3,$C9,$D9,$FD,$3C,$ED,$ED,$3F,$3F,$E3,$1F,$EF,$B8,$C3,$C9
	.dc.b	$AF,$08,$DC,$A7,$33,$EF,$8C,$4E,$A7,$33,$F0,$9A,$F7,$8D,$11,$F0
	.dc.b	$3F,$1B,$F0,$41,$00

	.text
	.align	4
uppercase_cc::
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
	.dc.l	uppercase_cc_&cc
  .endm
  .irp cc,F,EQ,OGT,OGE,OLT,OLE,OGL,OR,UN,UEQ,UGT,UGE,ULT,ULE,NE,T,SF,SEQ,GT,GE,LT,LE,GL,GLE,NGLE,NGL,NLE,NLT,NGE,NGT,SNE,ST
uppercase_cc_&cc::
	.dc.b	'&cc',0
  .endm

;--------------------------------------------------------------------------------
;	indirect data
;--------------------------------------------------------------------------------

	.text
	.even
indirect_start::
	pea.l	indirect_decompressed
	pea.l	indirect_compressed
	jbsr	decompress
	addq.l	#8,sp
	rts

	.align	4
indirect_compressed::
	.dc.b	$00,$12,$C7,$68,$7F,$00,$C4,$42,$11,$7F,$00,$E5,$E2,$42,$39,$70
	.dc.b	$76,$64,$76,$FC,$13,$AB,$D7,$F8,$07,$85,$70,$3C,$AE,$1E,$A3,$F7
	.dc.b	$BF,$37,$E6,$AF,$B0,$7D,$23,$EE,$1F,$86,$3F,$1F,$7E,$74,$FC,$8B
	.dc.b	$E0,$E5,$CB,$1E,$38,$70,$BD,$FE,$44,$FD,$FD,$FA,$57,$C5,$FD,$7D
	.dc.b	$F8,$D3,$F4,$E7,$E7,$4F,$C4,$5F,$8B,$BD,$E7,$E3,$7C,$F2,$E5,$8F
	.dc.b	$1C,$38,$5E,$FF,$1B,$7E,$F8,$FD,$77,$EA,$7E,$3E,$FC,$7D,$C2,$F9
	.dc.b	$57,$D3,$3F,$69,$77,$BF,$6A,$E7,$97,$2C,$78,$E1,$C2,$F6,$3F,$79
	.dc.b	$F4,$78,$5F,$84,$E0,$6E,$FD,$65,$F8,$D3,$F0,$D7,$EA,$0F,$B0,$7E
	.dc.b	$95,$CF,$AB,$F1,$99,$72,$C7,$8E,$1C,$2D,$FA,$53,$F7,$5E,$3A,$FE
	.dc.b	$53,$F4,$7F,$77,$E9,$8F,$C4,$5F,$B5,$3F,$77,$7E,$C6,$F9,$CE,$5C
	.dc.b	$3F,$9B,$C7,$8E,$1C,$2F,$7F,$38,$FD,$CF,$ED,$E9,$FA,$2F,$CA,$5F
	.dc.b	$6A,$FC,$F5,$FA,$AF,$87,$F6,$97,$D2,$F9,$F1,$E3,$8B,$1E,$38,$70
	.dc.b	$D1,$87,$0F,$DD,$9F,$B8,$BF,$3E,$FD,$87,$EE,$9E,$CF,$71,$F1,$CF
	.dc.b	$C7,$7E,$A7,$E3,$3F,$8C,$C3,$86,$DF,$B8,$C3,$85,$EF,$C5,$5F,$E8
	.dc.b	$BF,$71,$7E,$ED,$FB,$DF,$DA,$9F,$20,$FC,$D7,$E8,$7E,$B3,$F0,$3C
	.dc.b	$17,$E2,$4C,$78,$ED,$FC,$1B,$1E,$38,$70,$EB,$60,$FC,$49,$FB,$8B
	.dc.b	$F2,$CF,$E1,$9F,$A3,$7E,$37,$F1,$37,$ED,$FD,$7F,$76,$7E,$C7,$F1
	.dc.b	$B0,$E1,$B7,$F1,$4C,$38,$5E,$EB,$8F,$6B,$F0,$2F,$E3,$1F,$A5,$BF
	.dc.b	$5B,$7D,$CB,$F0,$7F,$D1,$FE,$91,$F8,$7F,$2B,$E9,$E1,$C2,$F7,$5A
	.dc.b	$F2,$DF,$78,$FD,$C1,$F8,$37,$E6,$B9,$BF,$93,$FD,$FB,$EE,$5E,$47
	.dc.b	$F0,$1F,$AC,$E1,$7B,$E6,$5E,$EB,$F8,$37,$D6,$7E,$F7,$F2,$DF,$E6
	.dc.b	$1F,$15,$FC,$67,$E1,$1F,$9D,$FF,$1D,$F0,$DF,$9F,$3F,$32,$E1,$A7
	.dc.b	$F3,$8C,$38,$5E,$E9,$F9,$93,$F6,$E7,$ED,$FF,$CF,$7F,$53,$7E,$EA
	.dc.b	$FA,$F7,$E0,$8F,$C5,$DF,$90,$BC,$17,$E0,$7C,$F2,$E5,$8F,$1C,$38
	.dc.b	$5E,$FF,$03,$7E,$DE,$FC,$F3,$EA,$FD,$36,$3F,$2D,$7F,$00,$7D,$5F
	.dc.b	$D1,$75,$7B,$E7,$FE,$A7,$E5,$FF,$3B,$85,$F0,$F1,$D7,$F8,$DF,$DB
	.dc.b	$FF,$90,$E1,$FD,$6B,$F6,$D7,$E8,$3E,$1F,$A4,$F9,$3B,$3E,$FD,$F8
	.dc.b	$43,$1E,$38,$B1,$E3,$7F,$CE,$BB,$7D,$58,$3F,$08,$7E,$DF,$F1,$B0
	.dc.b	$FE,$D3,$F2,$CF,$D7,$9F,$B1,$BF,$58,$7E,$45,$F3,$B8,$BF,$86,$63
	.dc.b	$C7,$1F,$E3,$78,$70,$BD,$F5,$FB,$9F,$25,$FF,$75,$FD,$47,$FA,$5F
	.dc.b	$D1,$FF,$4C,$E7,$7E,$33,$FD,$4B,$F2,$BD,$E7,$FD,$F2,$F7,$F6,$5F
	.dc.b	$2A,$BF,$AB,$FE,$56,$FD,$B5,$F2,$7F,$E0,$1F,$D4,$BF,$AF,$FE,$EF
	.dc.b	$FB,$AF,$DD,$9F,$96,$3F,$49,$FC,$1E,$1C,$35,$FF,$04,$E1,$C3,$E2
	.dc.b	$98,$70,$B7,$91,$FD,$9F,$F4,$CF,$F0,$77,$EA,$3F,$8F,$FC,$9F,$FA
	.dc.b	$D7,$63,$E9,$9F,$9F,$3D,$5E,$1B,$7F,$85,$B0,$E1,$F6,$9F,$3A,$F7
	.dc.b	$FA,$F7,$F0,$AF,$CA,$FF,$C3,$5F,$6F,$EC,$F6,$7F,$B2,$FE,$BE,$FD
	.dc.b	$C9,$FB,$27,$E8,$7E,$51,$87,$0E,$BF,$D2,$7D,$F3,$0E,$16,$FC,$5D
	.dc.b	$FB,$83,$F2,$2F,$F1,$47,$ED,$2F,$CD,$E7,$E5,$FF,$D2,$3E,$7B,$FA
	.dc.b	$27,$E5,$5C,$2F,$75,$AF,$B6,$57,$F4,$9F,$95,$FF,$07,$FE,$2F,$FE
	.dc.b	$21,$FC,$5B,$FC,$6F,$FB,$FB,$F1,$9C,$F6,$FF,$8E,$27,$FC,$51,$3F
	.dc.b	$E3,$AF,$C3,$7F,$C6,$F3,$99,$FF,$19,$FE,$F7,$9C,$3F,$C7,$DF,$8D
	.dc.b	$7F,$8C,$AB,$FC,$81,$7F,$E0,$87,$85,$FF,$8F,$70,$E9,$F8,$67,$8E
	.dc.b	$18,$F7,$FC,$23,$FC,$91,$FB,$02,$FF,$3F,$F9,$13,$F2,$D7,$ED,$9F
	.dc.b	$D7,$FE,$AF,$FC,$57,$87,$0B,$DD,$6B,$EF,$7F,$A2,$FE,$E4,$FD,$27
	.dc.b	$F9,$EF,$AC,$FD,$D3,$1F,$61,$FD,$37,$F0,$3F,$EE,$17,$BE,$3F,$E1
	.dc.b	$37,$BC,$3F,$A7,$FD,$A7,$F7,$BF,$8C,$7E,$E6,$FD,$BB,$FC,$C1,$A3
	.dc.b	$F4,$DF,$EB,$5F,$4E,$E6,$FE,$1F,$F3,$8F,$D6,$38,$E1,$C2,$F7,$5A
	.dc.b	$FA,$7E,$BF,$9B,$7F,$9A,$FF,$7E,$7E,$AA,$F9,$F7,$EA,$2F,$D9,$3E
	.dc.b	$5F,$A0,$FD,$53,$8F,$3F,$F9,$D3,$1E,$38,$70,$BD,$FE,$A9,$FE,$30
	.dc.b	$FC,$5B,$FC,$F1,$FB,$B3,$F3,$EF,$F1,$37,$EA,$CF,$D7,$FF,$9E,$7E
	.dc.b	$BC,$FD,$D3,$86,$BF,$E8,$0C,$38,$7E,$2A,$C3,$87,$D9,$30,$7E,$E9
	.dc.b	$FE,$3E,$FC,$A7,$FC,$EF,$8F,$FA,$1F,$E1,$7F,$D1,$3F,$3B,$FE,$61
	.dc.b	$8F,$C4,$9F,$E8,$AF,$5B,$83,$83,$FD,$1B,$1B,$7F,$D1,$7E,$B7,$E4
	.dc.b	$E3,$1E,$B7,$E5,$67,$AF,$EC,$7E,$07,$8B,$FE,$8F,$AE,$1B,$7F,$A6
	.dc.b	$7F,$6C,$FF,$05,$F8,$2F,$DC,$3F,$C8,$9F,$67,$FE,$05,$FC,$2F,$FC
	.dc.b	$AF,$7B,$F8,$AF,$C1,$56,$BF,$08,$63,$FE,$5C,$FD,$47,$F4,$FA,$FE
	.dc.b	$9F,$83,$0F,$FA,$6C,$F3,$70,$3F,$51,$FD,$37,$BB,$FE,$AE,$8C,$4F
	.dc.b	$EB,$23,$FD,$4F,$18,$87,$17,$BF,$F3,$9F,$F3,$5D,$F8,$27,$FD,$4F
	.dc.b	$86,$1F,$01,$F1,$FE,$B3,$BE,$3F,$E7,$27,$C3,$C7,$67,$C1,$3F,$EB
	.dc.b	$8F,$80,$C4,$78,$0F,$4F,$EA,$3D,$DF,$F5,$AE,$9F,$E2,$97,$B9,$8F
	.dc.b	$D4,$74,$FD,$AF,$D4,$6B,$FE,$C9,$9F,$F1,$A3,$C0,$FB,$FF,$2F,$A7
	.dc.b	$F0,$FF,$A8,$D7,$FD,$13,$F8,$55,$AF,$E6,$71,$C4,$7E,$7F,$FC,$86
	.dc.b	$33,$1E,$93,$FD,$AF,$1F,$92,$F9,$6F,$1F,$E7,$FF,$45,$BC,$3F,$E9
	.dc.b	$67,$35,$4E,$DC,$7B,$6C,$E5,$EC,$7F,$B9,$63,$B3,$E5,$8E,$0E,$8F
	.dc.b	$F7,$5B,$E7,$9F,$8A,$F0,$8D,$1F,$CE,$70,$F5,$F2,$7E,$7E,$7C,$0F
	.dc.b	$E8,$67,$FD,$5B,$FA,$8F,$DE,$A3,$FC,$CF,$8F,$E2,$36,$76,$7E,$03
	.dc.b	$8F,$FD,$86,$61,$FF,$77,$2F,$E0,$1E,$31,$B7,$F7,$73,$FC,$FF,$F0
	.dc.b	$1F,$CD,$3F,$D0,$7F,$3A,$DD,$72,$76,$F7,$FE,$01,$DB,$E1,$F8,$9D
	.dc.b	$E3,$FD,$5A,$3C,$4F,$FB,$FF,$F8,$07,$71,$F1,$3C,$6E,$BB,$7F,$55
	.dc.b	$7A,$C7,$C8,$F5,$B7,$3F,$D6,$7F,$45,$DC,$88,$57,$8C,$7F,$5F,$FD
	.dc.b	$17,$8C,$E8,$E1,$CD,$C9,$FC,$59,$F1,$E7,$FB,$37,$37,$08,$E1,$CD
	.dc.b	$B7,$F0,$86,$CF,$DB,$E9,$5F,$DC,$F2,$7D,$87,$F3,$87,$F2,$8F,$A8
	.dc.b	$FB,$87,$F6,$DE,$4F,$CF,$36,$3F,$DD,$AF,$D3,$AB,$1F,$37,$9A,$BF
	.dc.b	$73,$87,$A3,$F8,$1E,$1F,$E2,$57,$FD,$E7,$D8,$EB,$3F,$85,$76,$FE
	.dc.b	$13,$F8,$07,$6F,$E2,$7E,$4F,$6F,$E2,$1D,$E4,$21,$3F,$80,$38,$CB
	.dc.b	$F7,$EF,$ED,$9E,$53,$F5,$07,$F0,$07,$FB,$B7,$EE,$1F,$BA,$FA,$AF
	.dc.b	$75,$AF,$92,$57,$F6,$9E,$47,$07,$F0,$7F,$D6,$7F,$84,$7D,$17,$F0
	.dc.b	$9F,$D6,$7F,$85,$7E,$6B,$FC,$2F,$EF,$3F,$86,$7E,$39,$FC,$37,$F3
	.dc.b	$7F,$E1,$DF,$92,$7F,$0F,$FC,$8B,$F8,$87,$E4,$BF,$C4,$7F,$59,$FE
	.dc.b	$25,$F4,$3F,$C4,$FF,$59,$FE,$29,$F1,$5F,$C5,$7E,$4B,$F8,$B7,$E1
	.dc.b	$3F,$C5,$FE,$B7,$F8,$C7,$D4,$FF,$19,$FA,$5F,$E5,$5E,$57,$E4,$BC
	.dc.b	$77,$47,$F8,$8F,$C0,$FE,$57,$3F,$D8,$AB,$F9,$87,$ED,$DF,$2B,$C3
	.dc.b	$DC,$7F,$33,$F8,$DE,$58,$71,$C3,$F8,$9C,$71,$7F,$9A,$FC,$AF,$19
	.dc.b	$D6,$3F,$9D,$F8,$EE,$38,$7F,$3D,$78,$5F,$F1,$CF,$83,$E1,$3F,$D0
	.dc.b	$75,$DC,$FF,$40,$77,$7F,$A7,$9D,$CF,$E6,$B5,$F3,$BF,$9C,$3D,$FF
	.dc.b	$AD,$7F,$17,$3F,$67,$AF,$96,$3F,$92,$E6,$9C,$CF,$E2,$A3,$FC,$0B
	.dc.b	$18,$8F,$C7,$7D,$6F,$B6,$31,$1F,$D0,$6F,$BD,$FF,$6D,$C7,$8B,$FD
	.dc.b	$BF,$7E,$0F,$F7,$1D,$7E,$5E,$1C,$9D,$C7,$D1,$C1,$AF,$08,$C1,$C0
	.dc.b	$FE,$F4,$70,$8F,$F0,$C9,$BC,$3F,$01,$1B,$3C,$C7,$67,$F8,$84,$6E
	.dc.b	$3F,$B4,$1F,$12,$FF,$B9,$67,$B8,$FE,$37,$9E,$DF,$F0,$0E,$7F,$01
	.dc.b	$FE,$AB,$86,$DF,$E4,$59,$ED,$FF,$7E,$1D,$4F,$EF,$79,$ED,$F7,$E5
	.dc.b	$F4,$72,$FE,$ED,$87,$5C,$3F,$A9,$64,$FC,$6F,$FA,$91,$F4,$A3,$F2
	.dc.b	$AD,$C3,$FC,$04,$FD,$6C,$BF,$28,$FD,$88,$D9,$F2,$8D,$9F,$E9,$0F
	.dc.b	$73,$E9,$5F,$78,$C5,$F0,$7F,$4D,$3F,$DD,$27,$51,$8B,$EE,$F3,$38
	.dc.b	$65,$FE,$4E,$7F,$CA,$38,$7E,$37,$E1,$3B,$BF,$E1,$1F,$C0,$3F,$C2
	.dc.b	$5F,$AD,$7F,$84,$FF,$5A,$FF,$0A,$7D,$9F,$F8,$57,$F1,$7F,$F0,$B7
	.dc.b	$E2,$7F,$E3,$5D,$FF,$CB,$DB,$23,$1A,$7F,$97,$9F,$F1,$D6,$DE,$C7
	.dc.b	$FC,$79,$D1,$C1,$B3,$B3,$FC,$7B,$D1,$B7,$FC,$7F,$C9,$B6,$9B,$3F
	.dc.b	$C4,$5C,$DB,$7F,$C4,$9D,$1B,$69,$BC,$FF,$B2,$F4,$6D,$FF,$08,$69
	.dc.b	$E9,$9F,$DD,$3F,$CD,$DF,$F0,$57,$27,$F9,$6F,$CD,$FD,$A4,$FC,$06
	.dc.b	$CF,$F9,$F1,$FD,$C7,$4F,$DA,$7F,$75,$7B,$9F,$F6,$6B,$FE,$4E,$8F
	.dc.b	$DF,$6F,$B9,$FF,$68,$E9,$B7,$F9,$23,$B7,$FD,$AB,$A6,$DF,$E5,$06
	.dc.b	$F2,$FE,$A6,$EC,$FD,$AB,$FC,$96,$FF,$CF,$7D,$1F,$DD,$1B,$3B,$C7
	.dc.b	$FD,$E1,$9F,$C1,$8F,$F2,$73,$E3,$8F,$DB,$C7,$F9,$7E,$7F,$CE,$D9
	.dc.b	$F8,$E7,$FC,$D5,$7F,$D0,$27,$F0,$75,$B3,$FC,$F3,$9E,$DF,$F0,$C6
	.dc.b	$7D,$5A,$7C,$B3,$FC,$FF,$9E,$DF,$41,$F5,$BF,$5A,$FF,$22,$1F,$39
	.dc.b	$FE,$8E,$D7,$FC,$E8,$FC,$ED,$3F,$A3,$3F,$EA,$B8,$FE,$66,$D7,$FD
	.dc.b	$76,$FF,$AB,$63,$F6,$6F,$FA,$BF,$2F,$CF,$B6,$7F,$D7,$6F,$FA,$CF
	.dc.b	$0D,$C7,$F8,$CE,$BF,$C0,$DD,$3F,$E2,$2A,$FF,$5C,$4F,$FA,$39,$FE
	.dc.b	$83,$D3,$FC,$C6,$BF,$B3,$C7,$FA,$0F,$F5,$AF,$F4,$27,$EB,$5F,$E8
	.dc.b	$5F,$CA,$BF,$D0,$DF,$94,$FF,$8F,$BF,$0C,$7E,$E4,$FD,$E1,$FA,$53
	.dc.b	$EF,$1F,$8F,$3F,$0E,$7E,$27,$FD,$D3,$2E,$58,$F1,$C3,$85,$EF,$F2
	.dc.b	$4D,$71,$7E,$93,$F5,$CF,$F4,$B7,$E7,$DF,$EC,$4E,$7F,$F1,$EB,$67
	.dc.b	$FD,$52,$FF,$B2,$39,$FD,$9D,$A8,$FF,$65,$F8,$5C,$2B,$FD,$94,$FF
	.dc.b	$B8,$63,$FD,$9F,$E0,$F0,$7F,$DA,$3D,$9B,$AF,$F5,$C3,$B7,$FD,$A5
	.dc.b	$D9,$B7,$FD,$A9,$E0,$EF,$3F,$F3,$23,$B8,$FF,$60,$78,$3B,$7F,$D2
	.dc.b	$67,$FA,$B7,$0F,$F5,$97,$EB,$9F,$E8,$F3,$FB,$AE,$9F,$E5,$A6,$CF
	.dc.b	$FB,$19,$FE,$F3,$E1,$7B,$9F,$EF,$78,$EE,$7F,$05,$DD,$FE,$A0,$B5
	.dc.b	$7F,$80,$37,$E1,$5A,$D7,$83,$FE,$4C,$7F,$CE,$CF,$F9,$F6,$B8,$BD
	.dc.b	$35,$E2,$FF,$00,$3F,$99,$DF,$F8,$31,$E2,$7F,$8C,$9F,$F0,$9F,$81
	.dc.b	$FE,$A2,$AE,$73,$FE,$F2,$F0,$BF,$DF,$D1,$8C,$D1,$FE,$FE,$BF,$F4
	.dc.b	$6C,$7F,$AB,$67,$55,$FE,$3E,$3C,$27,$FA,$4E,$3D,$0F,$C7,$30,$34
	.dc.b	$F4,$C7,$FB,$9E,$FC,$D3,$A8,$F4,$D7,$6C,$F7,$3F,$DF,$0F,$F2,$2E
	.dc.b	$1F,$86,$7B,$1F,$EF,$DC,$B6,$FC,$CB,$FE,$F7,$FA,$E7,$FB,$2C,$FA
	.dc.b	$F1,$FD,$69,$AF,$FA,$97,$04,$7D,$7D,$7F,$BA,$BF,$36,$78,$1F,$78
	.dc.b	$F3,$BF,$3A,$3F,$42,$FD,$0A,$FF,$51,$B6,$7E,$ED,$76,$7F,$84,$68
	.dc.b	$CB,$8C,$60,$E2,$53,$8D,$7F,$93,$BC,$2F,$B8,$FD,$68,$FC,$09,$FE
	.dc.b	$72,$FC,$8B,$0F,$CC,$FC,$F6,$BE,$2F,$C1,$AB,$F0,$7F,$5C,$AF,$FB
	.dc.b	$33,$37,$CA,$E2,$3F,$29,$F2,$BF,$84,$DB,$F1,$2D,$8F,$B2,$FE,$23
	.dc.b	$5F,$C8,$3F,$8A,$E9,$DC,$FE,$1D,$F3,$E0,$9F,$C7,$F4,$F0,$AF,$C6
	.dc.b	$73,$70,$58,$FE,$0D,$FB,$15,$F9,$6E,$3E,$11,$AE,$9E,$14,$7B,$D7
	.dc.b	$E7,$3A,$77,$3F,$E0,$48,$F7,$57,$FA,$6B,$E5,$7F,$CE,$3C,$DF,$E8
	.dc.b	$F7,$FC,$FB,$A7,$EA,$1B,$3F,$98,$7F,$D0,$3D,$3F,$AF,$7F,$D0,$78
	.dc.b	$6E,$7F,$A3,$3F,$02,$3F,$D0,$FB,$EE,$7F,$A5,$6B,$B7,$FD,$A0,$ED
	.dc.b	$FE,$9D,$AE,$DF,$F6,$A1,$BC,$BF,$CB,$36,$89,$CF,$FD,$3E,$7F,$DB
	.dc.b	$59,$AB,$54,$FB,$FF,$F1,$2C,$61,$E7,$1F,$BD,$AF,$06,$7F,$C0,$63
	.dc.b	$FC,$87,$5F,$C1,$AF,$CD,$1F,$5B,$4D,$D7,$AC,$FE,$63,$3E,$19,$FE
	.dc.b	$11,$A7,$4D,$34,$72,$AF,$F3,$BE,$3F,$E6,$6C,$7F,$CE,$F8,$71,$C3
	.dc.b	$F9,$D9,$FE,$D0,$F9,$6F,$AF,$5F,$EA,$D7,$E8,$C7,$F2,$CE,$BD,$3F
	.dc.b	$48,$FE,$1F,$1F,$4A,$7F,$AC,$BF,$EB,$A8,$FE,$4A,$D7,$67,$CC,$8F
	.dc.b	$E5,$91,$F2,$B5,$FC,$AC,$F7,$B4,$E7,$1F,$CF,$BB,$CF,$8D,$7D,$63
	.dc.b	$B9,$4E,$E3,$AB,$ED,$63,$EA,$8F,$83,$83,$E5,$BF,$32,$BE,$5A,$FC
	.dc.b	$B7,$DE,$F8,$BF,$F3,$0B,$6F,$F6,$2F,$E0,$BF,$21,$EB,$68,$FC,$2F
	.dc.b	$D4,$BF,$22,$84,$1F,$87,$7F,$C0,$09,$DC,$64,$DD,$73,$C3,$F2,$9F
	.dc.b	$50,$FD,$9A,$10,$7E,$CF,$CE,$77,$19,$37,$5C,$E1,$F8,$57,$B1,$FD
	.dc.b	$03,$FE,$BA,$C7,$F3,$BF,$BB,$7B,$1F,$CF,$FE,$B9,$E5,$CF,$D9,$A7
	.dc.b	$7C,$13,$99,$F8,$18,$F7,$FC,$D3,$C5,$9F,$A7,$4E,$11,$88,$F6,$4F
	.dc.b	$97,$F5,$AF,$26,$27,$C8,$FE,$D7,$F7,$6F,$23,$FB,$7F,$CC,$FC,$1A
	.dc.b	$78,$DF,$16,$7E,$5F,$C1,$7B,$69,$73,$F0,$67,$F0,$FF,$76,$EE,$76
	.dc.b	$CB,$FC,$8F,$D1,$FE,$01,$E0,$FD,$FF,$E0,$5D,$AF,$E5,$3D,$9F,$99
	.dc.b	$F9,$47,$66,$E3,$DA,$BF,$83,$FE,$4D,$EC,$3B,$7E,$ED,$EA,$C9,$FB
	.dc.b	$B7,$5A,$FE,$1A,$F5,$7F,$87,$FE,$B5,$D5,$B8,$75,$AF,$E3,$1F,$AE
	.dc.b	$7C,$83,$D5,$F9,$17,$5F,$A3,$E9,$3F,$C7,$5E,$8F,$F1,$FF,$77,$A3
	.dc.b	$6F,$A2,$FD,$DB,$E9,$9F,$17,$F2,$6F,$47,$E6,$BD,$24,$F4,$5F,$B1
	.dc.b	$F4,$6D,$F4,$FD,$23,$EE,$9F,$3F,$F3,$3F,$85,$FA,$97,$33,$97,$A7
	.dc.b	$C4,$FC,$35,$7E,$3F,$CE,$FC,$29,$FD,$E3,$07,$37,$F8,$C9,$FF,$69
	.dc.b	$F4,$E3,$D0,$FE,$6F,$E6,$FD,$56,$E1,$F8,$27,$F9,$DF,$DB,$FD,$D6
	.dc.b	$F8,$1F,$9B,$C7,$A6,$2F,$B5,$FF,$71,$D7,$F4,$27,$C5,$3A,$FB,$5F
	.dc.b	$E0,$73,$FD,$1B,$F5,$EF,$05,$FF,$4D,$7D,$75,$D3,$F8,$07,$C1,$3F
	.dc.b	$D4,$5F,$C6,$57,$F4,$EF,$B5,$EF,$3F,$D5,$5E,$F1,$FD,$4B,$ED,$7B
	.dc.b	$CF,$F5,$9F,$B4,$EE,$9E,$F3,$FD,$63,$F4,$FE,$F2,$7B,$BF,$D8,$7F
	.dc.b	$55,$EE,$9E,$EF,$F6,$5F,$CB,$3B,$AF,$8E,$EE,$76,$7F,$B5,$7E,$BF
	.dc.b	$D8,$EA,$BF,$4F,$EA,$74,$5F,$9B,$74,$F9,$9F,$49,$CE,$1F,$94,$7A
	.dc.b	$0E,$7F,$B0,$F9,$8E,$5F,$A5,$F9,$3F,$44,$F8,$27,$1F,$D2,$3D,$E7
	.dc.b	$8B,$F1,$9F,$69,$E1,$FD,$17,$D7,$FA,$1F,$C8,$3C,$1F,$94,$7C,$63
	.dc.b	$D3,$FB,$F7,$C4,$3D,$1F,$9F,$7C,$2F,$C1,$BE,$79,$E7,$FC,$BF,$E6
	.dc.b	$9F,$03,$B7,$E5,$9E,$FF,$90,$F9,$3F,$90,$7D,$63,$DD,$FA,$87,$D4
	.dc.b	$3E,$3F,$E4,$7F,$48,$F8,$DF,$BE,$FD,$0F,$C6,$FE,$E9,$F1,$7F,$3C
	.dc.b	$FB,$67,$CE,$FD,$F7,$EC,$9F,$37,$F4,$EF,$AF,$F9,$17,$E1,$0F,$99
	.dc.b	$E5,$FE,$08,$FA,$BF,$90,$7E,$00,$FA,$9F,$BE,$7D,$EF,$D0,$3F,$16
	.dc.b	$7D,$3F,$C4,$FF,$12,$7D,$CF,$96,$FC,$39,$F6,$FE,$5B,$F0,$BF,$2B
	.dc.b	$F9,$23,$ED,$7C,$9F,$E4,$0F,$C1,$FC,$97,$E3,$8F,$C1,$7C,$97,$E3
	.dc.b	$3E,$43,$F3,$27,$E0,$7E,$3B,$F2,$E7,$E2,$BE,$37,$F2,$A7,$E2,$7E
	.dc.b	$37,$F2,$7F,$19,$FA,$03,$F1,$1F,$13,$F9,$E3,$F2,$3F,$11,$F9,$C3
	.dc.b	$F2,$1F,$11,$F9,$AF,$87,$FD,$39,$F8,$FF,$85,$FD,$29,$F9,$8F,$84
	.dc.b	$FD,$19,$FB,$EF,$43,$FC,$4F,$F0,$9F,$E4,$47,$F8,$3F,$4F,$F7,$9F
	.dc.b	$92,$FF,$1F,$43,$E0,$FF,$A3,$3F,$C0,$FE,$13,$FB,$1F,$C8,$EB,$E0
	.dc.b	$FF,$81,$CF,$EF,$CF,$F6,$69,$4F,$F5,$E9,$FB,$FC,$7F,$DA,$8F,$EE
	.dc.b	$F1,$FE,$E9,$8F,$83,$DE,$EF,$3F,$E5,$BF,$73,$FA,$A3,$FE,$58,$7B
	.dc.b	$BF,$E5,$FF,$6F,$FD,$43,$EF,$FA,$1F,$74,$7F,$98,$FD,$CF,$F0,$EB
	.dc.b	$FE,$60,$F6,$FF,$B1,$37,$2F,$E4,$CF,$F9,$93,$D9,$EE,$FF,$9A,$DE
	.dc.b	$EB,$D8,$EF,$E8,$FF,$9A,$A3,$07,$FA,$70,$FF,$36,$FB,$1E,$07,$FC
	.dc.b	$E9,$EB,$F8,$1B,$87,$81,$FF,$3C,$7A,$BE,$04,$F8,$2F,$BC,$7F,$AD
	.dc.b	$4E,$C7,$F9,$FF,$D4,$EC,$BD,$3E,$CD,$BE,$CB,$D1,$EC,$7B,$3D,$0E
	.dc.b	$B1,$FE,$89,$7E,$4A,$F4,$3A,$D7,$43,$EA,$C3,$CD,$F4,$1F,$1B,$CC
	.dc.b	$E7,$07,$9F,$99,$E6,$F2,$FC,$A7,$2F,$27,$E0,$1F,$3B,$C8,$E2,$78
	.dc.b	$BC,$8F,$6F,$8F,$EC,$3C,$3E,$2F,$C7,$3E,$AF,$89,$E0,$3F,$B6,$71
	.dc.b	$7F,$A9,$B4,$FF,$34,$F8,$08,$C7,$83,$FE,$2C,$6F,$87,$FE,$3B,$78
	.dc.b	$78,$3F,$E3,$46,$F1,$7F,$91,$5E,$9D,$FF,$90,$6E,$3F,$EF,$27,$C1
	.dc.b	$DF,$FE,$78,$7C,$5D,$DD,$F3,$FC,$76,$5D,$DF,$F5,$9F,$67,$76,$E5
	.dc.b	$DD,$FC,$06,$3F,$E9,$BC,$3D,$9D,$7E,$E9,$FF,$5D,$C7,$77,$FD,$7B
	.dc.b	$D5,$DE,$BF,$D8,$8F,$F8,$4A,$7E,$25,$F7,$9F,$F6,$43,$FE,$50,$8F
	.dc.b	$91,$CF,$ED,$BF,$F4,$45,$7E,$7C,$FC,$83,$FE,$89,$3F,$D9,$D3,$F3
	.dc.b	$F9,$7C,$12,$7C,$13,$FE,$D0,$E3,$F0,$49,$F0,$47,$FB,$53,$8B,$C1
	.dc.b	$3F,$ED,$B7,$ED,$FC,$DF,$F1,$89,$FB,$4F,$F0,$9C,$BF,$8C,$62,$8C
	.dc.b	$47,$FA,$4E,$31,$7F,$BB,$DB,$F8,$CC,$62,$1B,$FC,$5B,$F9,$F1,$FB
	.dc.b	$DF,$07,$BB,$FE,$ED,$DF,$C6,$FC,$2C,$84,$26,$1D,$72,$FC,$56,$1F
	.dc.b	$EF,$07,$E7,$3F,$EF,$2D,$7F,$24,$DC,$FF,$24,$FC,$3C,$3E,$80,$CB
	.dc.b	$C6,$FF,$B8,$B0,$FF,$63,$3F,$E4,$86,$BE,$CA,$F7,$FC,$3F,$E4,$1F
	.dc.b	$F4,$DC,$EB,$1F,$CF,$4F,$95,$FC,$F3,$E1,$AF,$C8,$35,$FC,$14,$BF
	.dc.b	$16,$FA,$8D,$8F,$C5,$D7,$F3,$C8,$C4,$7F,$A7,$23,$F6,$53,$CE,$3F
	.dc.b	$95,$4B,$F1,$9F,$15,$7D,$C6,$BF,$6D,$E1,$3F,$A0,$F2,$2B,$FC,$8F
	.dc.b	$F1,$8A,$FE,$FF,$E2,$AE,$0F,$8A,$BF,$B3,$E3,$B3,$3F,$8C,$EB,$F1
	.dc.b	$9D,$FC,$6F,$0F,$D6,$78,$FC,$77,$2F,$8E,$E7,$FA,$9F,$4F,$8F,$EB
	.dc.b	$FA,$FF,$6F,$D7,$7B,$F8,$DE,$0F,$CF,$BC,$3F,$25,$E2,$F9,$2F,$1F
	.dc.b	$DD,$F9,$3D,$8F,$2F,$EB,$9E,$6F,$E0,$1F,$3F,$CA,$FA,$3F,$3A,$F4
	.dc.b	$F9,$FE,$AF,$3F,$D7,$FA,$7F,$B3,$E5,$FD,$BF,$91,$7B,$BF,$80,$7D
	.dc.b	$FF,$01,$F0,$3F,$41,$F8,$3F,$35,$F0,$BE,$6B,$E1,$FF,$00,$FC,$4F
	.dc.b	$4F,$E2,$FE,$D1,$F1,$BF,$63,$F8,$FF,$3B,$F2,$3F,$57,$F9,$3F,$11
	.dc.b	$F2,$BE,$23,$E5,$FE,$99,$F3,$3E,$7F,$E6,$FE,$CD,$F3,$BF,$55,$F9
	.dc.b	$FF,$05,$F4,$3F,$07,$FA,$3F,$45,$F4,$BE,$8B,$E9,$FE,$D3,$F5,$3E
	.dc.b	$43,$EA,$FE,$9B,$F5,$BF,$0D,$FA,$FF,$4B,$F6,$3F,$47,$FB,$3F,$0F
	.dc.b	$F6,$BD,$AF,$B7,$F2,$FF,$A5,$F8,$5F,$D2,$C2,$3E,$FF,$D3,$BD,$F9
	.dc.b	$F1,$3F,$B5,$7F,$33,$1C,$5A,$7F,$0F,$3F,$9D,$7C,$57,$E7,$0E,$4F
	.dc.b	$EF,$66,$7C,$0F,$F7,$19,$CC,$7F,$02,$F8,$9F,$F0,$27,$E0,$79,$BF
	.dc.b	$0C,$FF,$0F,$7C,$99,$FF,$53,$BE,$33,$99,$E0,$7F,$0B,$F8,$8F,$6C
	.dc.b	$FF,$03,$FA,$FB,$3E,$73,$E0,$FD,$D7,$2A,$F8,$7F,$BD,$FE,$83,$F5
	.dc.b	$DF,$D0,$FE,$77,$FB,$63,$BE,$5E,$AC,$1F,$9F,$B2,$FD,$6E,$BF,$26
	.dc.b	$31,$3D,$CF,$E8,$4F,$FB,$32,$BE,$21,$FC,$8B,$7E,$1B,$72,$AE,$AF
	.dc.b	$ED,$2B,$FC,$51,$5C,$0F,$E5,$0E,$1E,$7F,$07,$FC,$7E,$31,$1F,$46
	.dc.b	$3C,$95,$EA,$8F,$E6,$1E,$A6,$23,$CC,$7E,$07,$F0,$FC,$2E,$1F,$F7
	.dc.b	$87,$EC,$7F,$E0,$5B,$F1,$5F,$E1,$E1,$E2,$FF,$0A,$5E,$2D,$F1,$7F
	.dc.b	$A9,$61,$FE,$EC,$3F,$CA,$25,$C7,$F9,$B9,$DC,$D9,$FC,$5B,$5C,$3E
	.dc.b	$63,$F9,$67,$2F,$C3,$CF,$FA,$5D,$7F,$93,$63,$1C,$DC,$29,$F8,$1F
	.dc.b	$81,$FC,$AB,$E0,$75,$F8,$27,$3C,$38,$47,$93,$92,$BF,$D3,$32,$AF
	.dc.b	$12,$10,$FC,$B9,$63,$FC,$05,$5D,$63,$FC,$C5,$7B,$7C,$A7,$53,$D5
	.dc.b	$7F,$7F,$0F,$F0,$56,$3C,$67,$F9,$B9,$FE,$6B,$CB,$FC,$F6,$FE,$02
	.dc.b	$BF,$1A,$FC,$D3,$27,$A3,$27,$EB,$8D,$3F,$C0,$CF,$D8,$9F,$9C,$7F
	.dc.b	$03,$7E,$EB,$83,$8D,$7E,$2C,$FE,$F7,$C1,$FC,$27,$F3,$CF,$24,$FD
	.dc.b	$C6,$BF,$E8,$1E,$2D,$3C,$69,$58,$8F,$E0,$31,$F9,$95,$5E,$D8,$FF
	.dc.b	$04,$F1,$FF,$7D,$8C,$47,$F8,$2D,$F8,$13,$FE,$31,$6B,$FC,$77,$9B
	.dc.b	$CD,$9F,$87,$9B,$D1,$C5,$FD,$8A,$FF,$73,$7F,$80,$BF,$E4,$9C,$4E
	.dc.b	$7E,$12,$8C,$3E,$8C,$BF,$7B,$0E,$8F,$ED,$AB,$A6,$F1,$FF,$11,$BF
	.dc.b	$C2,$F1,$E8,$FE,$30,$E0,$FE,$47,$D5,$FE,$1C,$BF,$D1,$E3,$FC,$A6
	.dc.b	$74,$7F,$8A,$F9,$7F,$CA,$1C,$1E,$5C,$3F,$93,$C7,$F2,$E9,$C1,$E7
	.dc.b	$AF,$E2,$4E,$0F,$F4,$0D,$BA,$FD,$27,$F4,$18,$FF,$5D,$46,$5F,$AE
	.dc.b	$7E,$7B,$B3,$DE,$6D,$FE,$87,$1F,$E8,$B7,$F1,$4B,$BB,$E2,$F7,$FC
	.dc.b	$6F,$03,$E3,$F8,$3F,$23,$C2,$F9,$3E,$1F,$CA,$F1,$3E,$5F,$8B,$F3
	.dc.b	$3C,$6F,$9B,$E3,$FC,$EF,$23,$E7,$F9,$3F,$43,$C3,$FA,$3E,$2F,$D2
	.dc.b	$F1,$FE,$9F,$95,$F5,$3C,$BF,$AB,$E6,$7D,$6F,$37,$EB,$F9,$DF,$63
	.dc.b	$CF,$FB,$3E,$87,$DA,$F4,$7E,$DF,$A5,$F7,$3D,$3F,$BB,$EA,$7D,$EF
	.dc.b	$57,$EF,$FA,$DF,$80,$F5,$FF,$03,$EC,$7E,$0B,$D9,$FC,$1F,$B5,$F8
	.dc.b	$4F,$53,$F0,$BE,$DF,$E1,$BD,$CF,$C3,$FB,$BF,$88,$F7,$BF,$13,$EF
	.dc.b	$FE,$2B,$E0,$3F,$17,$F0,$3F,$8C,$F8,$2F,$C6,$FC,$1F,$E3,$BE,$13
	.dc.b	$F1,$FF,$0B,$F9,$0F,$86,$FC,$8F,$C3,$FE,$4B,$E2,$3F,$27,$F1,$3F
	.dc.b	$94,$F8,$3F,$CA,$FC,$57,$E5,$BE,$2F,$F2,$FF,$19,$F9,$8F,$8D,$FC
	.dc.b	$CF,$C7,$7E,$6B,$E3,$FF,$37,$F2,$1F,$9C,$F8,$AF,$CE,$FC,$8F,$E7
	.dc.b	$BE,$4B,$F3,$FF,$27,$3D,$FF,$F2,$EF,$CA,$08,$4F,$80,$FF,$2D,$46
	.dc.b	$A3,$23,$E9,$3F,$5D,$F1,$1F,$AA,$3F,$CB,$B1,$AA,$FF,$33,$BC,$0E
	.dc.b	$B3,$C2,$BF,$3D,$F0,$BF,$E8,$FA,$FF,$3F,$FC,$3F,$F4,$16,$CF,$F3
	.dc.b	$D7,$FC,$FD,$7C,$CF,$43,$87,$B5,$FF,$01,$F9,$6F,$F4,$37,$CB,$88
	.dc.b	$5A,$FF,$46,$FA,$38,$89,$F4,$BC,$71,$F2,$E7,$FE,$A4,$BF,$F5,$76
	.dc.b	$7E,$5F,$8A,$FE,$C5,$1F,$EB,$1F,$99,$E5,$8F,$38,$C3,$81,$8F,$F1
	.dc.b	$5A,$F8,$FA,$FC,$F6,$BF,$36,$BF,$D7,$9F,$33,$FD,$85,$FF,$5F,$7C
	.dc.b	$57,$87,$4F,$F5,$D1,$F1,$6F,$DF,$8F,$85,$AF,$C1,$C7,$C1,$3F,$E4
	.dc.b	$BF,$0B,$E5,$65,$F2,$AF,$FD,$9E,$FD,$6C,$7F,$D9,$9A,$F8,$4F,$E4
	.dc.b	$E7,$C0,$3F,$DC,$51,$CA,$3F,$D9,$37,$FE,$57,$F9,$AE,$BC,$3F,$CC
	.dc.b	$EF,$FD,$89,$9F,$C3,$BF,$F7,$73,$F1,$73,$D6,$1F,$E5,$FF,$A1,$F0
	.dc.b	$3F,$E6,$1F,$9C,$FC,$26,$1F,$EF,$A8,$FF,$6E,$61,$E0,$7F,$D6,$8D
	.dc.b	$CF,$C2,$E7,$E7,$7E,$96,$5F,$81,$BF,$F3,$1C,$63,$2F,$0E,$0E,$0F
	.dc.b	$0B,$FC,$8C,$FF,$11,$E1,$DF,$83,$FC,$4B,$F2,$3C,$3E,$1B,$E5,$E9
	.dc.b	$4F,$CB,$6F,$D3,$97,$C3,$B5,$3E,$06,$3E,$9F,$4F,$BF,$A7,$EE,$D3
	.dc.b	$FD,$00,$F6,$9F,$77,$2F,$66,$E5,$D9,$F8,$5E,$07,$66,$E3,$DB,$8B
	.dc.b	$FB,$E1,$FE,$FE,$F9,$BF,$F0,$0E,$5F,$E8,$43,$FD,$1C,$E1,$FF,$43
	.dc.b	$EF,$FE,$22,$7F,$D2,$5F,$3D,$EA,$9E,$0F,$C9,$79,$73,$FC,$26,$E7
	.dc.b	$F0,$9E,$DF,$3F,$FE,$3B,$7C,$1F,$41,$FE,$3F,$6E,$1F,$E3,$B7,$C9
	.dc.b	$D7,$DE,$3F,$13,$A7,$ED,$A7,$D1,$F4,$3D,$E7,$FD,$3A,$FE,$42,$3D
	.dc.b	$9D,$7D,$E7,$F1,$07,$F8,$F8,$A7,$FA,$6A,$3E,$17,$D1,$77,$8F,$F5
	.dc.b	$83,$ED,$7F,$6B,$9F,$C3,$7E,$FF,$35,$FA,$3E,$D1,$FE,$B7,$FA,$4F
	.dc.b	$63,$FE,$B0,$7F,$0F,$87,$CF,$ED,$ED,$27,$B3,$FE,$BF,$FA,$5F,$63
	.dc.b	$71,$F9,$4F,$FA,$EB,$B7,$BB,$F9,$53,$ED,$1F,$68,$FF,$5F,$F6,$F8
	.dc.b	$1F,$F6,$0F,$CA,$78,$0E,$EB,$BB,$BB,$6F,$BB,$FE,$D1,$F8,$CE,$FF
	.dc.b	$19,$EC,$8F,$F6,$A3,$D5,$FF,$6B,$78,$1E,$C3,$AF,$C2,$74,$3E,$A4
	.dc.b	$7F,$B4,$BC,$1F,$4F,$83,$CE,$3F,$C9,$E7,$E0,$4F,$F6,$BF,$B9,$F1
	.dc.b	$8F,$3F,$85,$E6,$39,$C7,$FB,$97,$D4,$F8,$7E,$A7,$33,$C9,$0F,$13
	.dc.b	$E7,$9F,$03,$CC,$F7,$9C,$A3,$FD,$1D,$E2,$7F,$A5,$3C,$4F,$F7,$3F
	.dc.b	$93,$F9,$57,$FD,$41,$D2,$7F,$DD,$7F,$33,$F9,$F6,$F2,$7E,$7D,$FF
	.dc.b	$77,$F8,$DF,$A7,$6F,$85,$FC,$7E,$F9,$65,$FE,$FD,$FA,$6F,$D2,$57
	.dc.b	$FB,$DB,$E9,$FC,$13,$FF,$00,$5F,$C4,$C7,$F3,$0C,$7F,$B1,$5F,$2A
	.dc.b	$79,$1F,$A1,$AF,$F7,$FF,$D4,$78,$64,$1F,$DC,$46,$21,$E4,$F8,$9D
	.dc.b	$F9,$3E,$24,$F8,$9E,$7F,$53,$F0,$5B,$97,$C1,$7B,$F9,$5E,$14,$F8
	.dc.b	$1F,$1F,$D5,$77,$5F,$57,$DC,$EC,$BC,$CE,$CD,$BE,$CB,$EB,$3B,$1D
	.dc.b	$57,$9B,$D7,$D0,$F5,$1D,$3E,$B7,$99,$F1,$E7,$F7,$FC,$BF,$D1,$B8
	.dc.b	$FE,$67,$9F,$CF,$EB,$BC,$DF,$5F,$F0,$CF,$2E,$9E,$43,$96,$0F,$B0
	.dc.b	$F8,$07,$8F,$D1,$F7,$7A,$3F,$2C,$F6,$FD,$8F,$B0,$E3,$E9,$7C,$73
	.dc.b	$D5,$F6,$5F,$17,$EC,$FE,$91,$F1,$3D,$4F,$86,$78,$3E,$D3,$E7,$1F
	.dc.b	$07,$D5,$F9,$9E,$AF,$D9,$3E,$5F,$D3,$FC,$A3,$CF,$EB,$7D,$53,$E4
	.dc.b	$7D,$27,$D3,$FA,$4F,$C0,$1F,$4B,$D8,$FA,$27,$BB,$E8,$7E,$E1,$F3
	.dc.b	$FD,$9F,$B5,$EC,$FE,$1C,$FB,$3F,$3F,$F6,$0F,$8B,$ED,$7E,$0C,$FA
	.dc.b	$DF,$39,$F8,$1F,$9C,$FC,$71,$F8,$0F,$73,$EF,$9F,$33,$E6,$7F,$14
	.dc.b	$7D,$DF,$77,$F1,$1E,$EF,$E5,$4F,$C3,$FC,$BF,$E1,$8F,$A7,$EF,$7E
	.dc.b	$44,$FC,$27,$CA,$7E,$3F,$E5,$3F,$38,$7E,$3B,$E0,$3F,$1A,$7D,$AF
	.dc.b	$91,$FC,$C1,$F8,$BF,$81,$FC,$B7,$C0,$FE,$8C,$FC,$AF,$C7,$FE,$50
	.dc.b	$FC,$0F,$C1,$7E,$7C,$FC,$97,$C6,$7E,$77,$E3,$3F,$54,$7E,$73,$E1
	.dc.b	$3F,$36,$7E,$23,$E2,$7F,$4C,$7E,$67,$E1,$7F,$B3,$7D,$87,$F1,$2F
	.dc.b	$37,$F9,$E9,$FD,$D9,$FE,$26,$7E,$E7,$EA,$BF,$BA,$37,$D4,$FE,$1C
	.dc.b	$FF,$14,$F4,$3F,$83,$B5,$FF,$3D,$FC,$43,$D5,$FF,$76,$FD,$AF,$83
	.dc.b	$1F,$E0,$FF,$6D,$FD,$D7,$0F,$E0,$CF,$F5,$A7,$F8,$DE,$3F,$E5,$DB
	.dc.b	$FF,$7A,$FC,$47,$F8,$0E,$7F,$86,$7D,$BF,$F8,$0D,$B9,$FF,$80,$DF
	.dc.b	$E2,$1F,$13,$DF,$27,$C4,$F7,$7F,$90,$1F,$E0,$39,$FE,$2D,$F7,$1E
	.dc.b	$E8,$FE,$36,$F7,$7F,$CA,$17,$FE,$5C,$C9,$5F,$C9,$E3,$DF,$18,$9F
	.dc.b	$E4,$7F,$73,$E1,$FB,$AF,$04,$1F,$73,$FC,$A3,$E2,$FC,$0F,$F2,$9D
	.dc.b	$BE,$6B,$FC,$B3,$EE,$FD,$CD,$C7,$DC,$FF,$30,$F8,$BF,$77,$C5,$F6
	.dc.b	$3A,$C3,$EF,$3A,$9E,$B5,$F1,$9D,$1B,$7D,$17,$E2,$7F,$27,$EF,$BD
	.dc.b	$27,$4F,$8E,$F3,$9C,$FE,$FF,$91,$FE,$25,$E2,$FE,$81,$F1,$FF,$47
	.dc.b	$E3,$FE,$09,$E4,$FC,$03,$DE,$78,$FE,$43,$C4,$71,$CF,$FA,$5F,$E0
	.dc.b	$3F,$67,$F0,$2F,$92,$7B,$3E,$4B,$E3,$9E,$BF,$C0,$FD,$47,$1A,$FE
	.dc.b	$6B,$E2,$08,$43,$F2,$9A,$7F,$8D,$A3,$F4,$07,$FA,$0C,$FF,$23,$7E
	.dc.b	$09,$FA,$57,$F4,$BD,$8F,$C1,$71,$1F,$E2,$48,$7E,$0D,$88,$FF,$30
	.dc.b	$46,$1F,$EC,$BE,$47,$F9,$56,$7F,$B2,$7C,$EF,$F9,$55,$B3,$FE,$55
	.dc.b	$73,$F7,$BE,$7C,$FD,$38,$73,$A5,$BE,$73,$F7,$31,$FA,$FC,$3A,$C7
	.dc.b	$38,$FE,$EB,$3E,$17,$FB,$BD,$F8,$5A,$6C,$F9,$18,$78,$63,$FC,$F9
	.dc.b	$7F,$DE,$DF,$04,$FC,$48,$C5,$F1,$9F,$F3,$97,$E0,$FF,$0E,$BF,$C0
	.dc.b	$58,$A3,$FC,$3E,$7F,$88,$BE,$7B,$B5,$7F,$07,$8F,$F1,$2C,$FB,$0E
	.dc.b	$15,$FE,$0E,$BF,$F3,$33,$FE,$6D,$9F,$F1,$5E,$3C,$AB,$73,$F1,$6B
	.dc.b	$95,$7F,$8C,$1F,$F4,$13,$FE,$32,$F8,$CE,$0F,$F8,$D3,$E3,$37,$5D
	.dc.b	$DD,$BE,$1C,$76,$FF,$88,$3C,$FF,$F4,$75,$7F,$27,$8D,$9F,$5A,$76
	.dc.b	$FF,$B2,$CF,$C4,$47,$CB,$77,$1F,$ED,$B3,$F0,$18,$FA,$9F,$7C,$7F
	.dc.b	$AA,$A7,$E9,$4F,$0C,$BA,$D7,$D4,$AE,$A7,$DE,$CB,$AE,$5A,$CB,$DB
	.dc.b	$3F,$A5,$7C,$35,$F2,$7F,$08,$FB,$13,$FE,$B8,$7F,$CA,$A7,$F9,$63
	.dc.b	$F0,$9D,$47,$D3,$FC,$2B,$EB,$4F,$F9,$71,$FF,$57,$3F,$76,$BF,$15
	.dc.b	$9A,$3E,$ED,$7E,$B1,$AF,$FB,$1F,$E7,$FF,$A4,$47,$F9,$9B,$F0,$BE
	.dc.b	$F3,$F9,$FF,$C3,$3B,$BF,$83,$7B,$BF,$85,$DB,$7B,$7C,$38,$FD,$9D
	.dc.b	$FE,$05,$FF,$4D,$E1,$FD,$57,$F0,$7F,$7C,$7F,$57,$FC,$36,$E6,$FC
	.dc.b	$3A,$FE,$1F,$8F,$C4,$33,$F8,$8C,$C3,$F5,$7D,$5F,$91,$8F,$D3,$75
	.dc.b	$7E,$65,$B3,$E0,$C5,$1A,$DF,$E2,$BC,$37,$FE,$1E,$DA,$F8,$95,$C1
	.dc.b	$FD,$CD,$F7,$7F,$77,$C5,$DD,$A7,$7A,$FD,$C4,$30,$EF,$3C,$30,$EE
	.dc.b	$D3,$BD,$7F,$BF,$2B,$FB,$D7,$E2,$5F,$10,$FF,$4B,$9F,$E0,$9E,$AF
	.dc.b	$BA,$FF,$82,$F2,$FE,$4E,$DB,$F4,$BF,$E0,$FC,$77,$1F,$BB,$3E,$B1
	.dc.b	$FE,$8A,$94,$FF,$85,$71,$F9,$0F,$F8,$5F,$A3,$8D,$7F,$12,$E8,$EE
	.dc.b	$F3,$76,$F5,$CB,$77,$FE,$9B,$8F,$E6,$B7,$FE,$B2,$E3,$FF,$10,$9E
	.dc.b	$E8,$F9,$B5,$D8,$FD,$C3,$FC,$6A,$7E,$5B,$D8,$FE,$9B,$9E,$E7,$F6
	.dc.b	$F9,$7A,$A7,$EA,$BB,$8F,$E5,$A7,$EF,$AB,$EA,$E1,$EA,$AF,$EA,$EE
	.dc.b	$E5,$97,$5C,$B5,$1F,$B9,$AF,$9A,$7E,$09,$FD,$7F,$E2,$7F,$9A,$3F
	.dc.b	$D9,$A7,$F8,$FF,$A7,$33,$F4,$7A,$7F,$AE,$CF,$F9,$0D,$FC,$B3,$F7
	.dc.b	$6B,$F8,$9B,$F6,$E7,$FA,$EB,$5F,$F6,$0E,$BF,$E3,$F8,$FF,$26,$74
	.dc.b	$F6,$9F,$E8,$FD,$3D,$9F,$C0,$3D,$9F,$C1,$70,$6F,$5F,$C2,$47,$63
	.dc.b	$F0,$CF,$EA,$7F,$10,$FD,$AF,$E2,$92,$7F,$DA,$3F,$1F,$FE,$18,$FC
	.dc.b	$59,$ED,$3F,$0D,$7E,$16,$FC,$59,$F2,$1F,$E5,$5C,$BF,$DA,$D8,$F1
	.dc.b	$C3,$85,$EF,$EF,$BF,$CB,$5F,$8B,$FF,$04,$FC,$63,$F8,$2F,$E3,$3F
	.dc.b	$C1,$BF,$1A,$9B,$FE,$CB,$C1,$FE,$3F,$1F,$B0,$F0,$7F,$C1,$2D,$9F
	.dc.b	$DD,$B9,$E4,$FE,$7E,$EF,$93,$FC,$18,$D9,$FF,$75,$74,$7F,$9A,$E3
	.dc.b	$A7,$4F,$09,$FC,$01,$C0,$A7,$0E,$F4,$FF,$02,$9F,$E0,$47,$F8,$8E
	.dc.b	$33,$3F,$89,$BE,$4F,$F8,$9D,$F3,$BF,$8D,$9F,$EA,$BF,$8D,$FF,$26
	.dc.b	$3F,$A3,$E1,$FE,$7F,$7F,$C5,$1E,$0E,$AB,$ED,$E1,$FE,$27,$9F,$F1
	.dc.b	$63,$FC,$71,$FF,$17,$65,$B9,$FE,$B5,$8E,$DF,$C1,$BB,$7F,$0B,$E0
	.dc.b	$6E,$11,$3C,$0E,$0F,$F6,$37,$83,$FD,$97,$3E,$0F,$F0,$33,$F8,$86
	.dc.b	$1F,$EA,$3C,$FF,$9B,$3F,$3A,$7E,$33,$F2,$67,$AC,$7F,$93,$1E,$AF
	.dc.b	$F0,$23,$6F,$EC,$F2,$DC,$7F,$94,$DD,$BF,$C0,$8F,$F1,$EC,$FD,$5C
	.dc.b	$37,$1F,$D6,$6F,$E6,$5F,$EF,$71,$F8,$D3,$FE,$5C,$C7,$67,$47,$FB
	.dc.b	$7F,$E3,$9F,$D6,$8F,$AB,$91,$FF,$2E,$78,$3F,$E0,$E7,$FC,$BD,$E0
	.dc.b	$FF,$96,$DB,$8F,$E2,$5F,$B7,$7F,$87,$7E,$CC,$7F,$97,$9A,$FF,$02
	.dc.b	$BE,$CF,$EE,$BC,$0E,$CF,$EF,$3C,$0E,$CD,$3B,$1F,$C0,$37,$DB,$F8
	.dc.b	$23,$B1,$F8,$23,$F3,$9F,$8E,$FF,$76,$FC,$7B,$FB,$BF,$E3,$FF,$DE
	.dc.b	$3F,$20,$FE,$F3,$F9,$0F,$F7,$AF,$C8,$A6,$1F,$C9,$7A,$7F,$C5,$B1
	.dc.b	$FA,$3E,$9F,$F3,$D3,$67,$F5,$0E,$73,$FD,$33,$BE,$6F,$F3,$EB,$67
	.dc.b	$E0,$3C,$F8,$77,$1F,$A7,$36,$53,$67,$EA,$EB,$6F,$F9,$FE,$B6,$D3
	.dc.b	$67,$F9,$9A,$7F,$15,$F9,$1F,$F8,$58,$FB,$A7,$FA,$53,$A7,$F7,$8F
	.dc.b	$FA,$5B,$8B,$FD,$26,$DA,$FA,$CF,$FA,$6B,$0D,$C7,$F3,$03,$FC,$51
	.dc.b	$29,$FF,$4F,$E1,$FC,$2D,$FF,$50,$F3,$F0,$AF,$F1,$77,$3F,$67,$93
	.dc.b	$B7,$A6,$5B,$BF,$E3,$71,$B3,$FC,$B1,$D1,$FE,$1E,$C3,$E5,$57,$FA
	.dc.b	$81,$FE,$30,$FF,$AB,$DB,$3D,$4F,$F5,$36,$7E,$39,$FE,$39,$1F,$E9
	.dc.b	$09,$E3,$96,$E7,$FD,$85,$9E,$DF,$ED,$F5,$B3,$F8,$F6,$3B,$8F,$F6
	.dc.b	$4E,$7B,$7F,$B8,$4F,$F8,$5F,$0F,$2E,$7B,$9F,$F2,$FF,$47,$F9,$28
	.dc.b	$FD,$89,$FE,$C8,$3F,$98,$1F,$EC,$AE,$8F,$F6,$3B,$6F,$F5,$4F,$DB
	.dc.b	$9F,$F6,$0B,$F6,$63,$FC,$4C,$DD,$5A,$23,$FD,$AB,$CF,$DA,$BE,$E7
	.dc.b	$3F,$66,$9D,$8F,$C0,$6D,$B7,$FC,$D1,$B6,$DA,$7E,$F7,$3F,$F3,$2F
	.dc.b	$E4,$9F,$E6,$6F,$C9,$7F,$CC,$FF,$93,$7F,$9A,$3F,$27,$FF,$34,$FE
	.dc.b	$51,$FE,$6A,$FC,$A7,$FC,$D7,$F9,$57,$F9,$B3,$F2,$BF,$F3,$6F,$E5
	.dc.b	$9F,$E6,$EF,$CB,$7F,$CD,$FF,$97,$7F,$9C,$3F,$2F,$FF,$38,$FE,$61
	.dc.b	$FE,$72,$FC,$C7,$FC,$E7,$F9,$97,$F9,$D3,$F3,$3F,$F3,$AF,$E6,$9F
	.dc.b	$E7,$6F,$CD,$67,$2F,$F4,$2F,$C8,$9F,$A4,$BF,$01,$79,$CF,$B0,$7E
	.dc.b	$20,$FC,$49,$FA,$37,$CA,$CF,$FA,$5E,$6C,$F9,$E5,$CB,$1E,$38,$3F
	.dc.b	$12,$E1,$7F,$D9,$7C,$60,$FB,$4F,$E8,$19,$7E,$6A,$33,$F6,$9F,$E5
	.dc.b	$F8,$C4,$21,$E4,$74,$BF,$F7,$2F,$E3,$7C,$9E,$D4,$FC,$6F,$94,$61
	.dc.b	$E3,$E5,$31,$EB,$F6,$9C,$6B,$FD,$DF,$F6,$9C,$1B,$97,$7C,$7F,$E0
	.dc.b	$07,$E3,$3B,$7D,$64,$1F,$8C,$F1,$70,$7C,$07,$C3,$8F,$FA,$27,$27
	.dc.b	$E7,$1F,$EA,$C3,$FA,$99,$F3,$BC,$BF,$F3,$8D,$FE,$6A,$31,$97,$D2
	.dc.b	$3F,$BE,$FC,$CF,$FA,$16,$33,$19,$8F,$7C,$62,$3E,$CF,$CC,$FF,$6F
	.dc.b	$8C,$47,$F8,$1B,$E6,$79,$3F,$C7,$CC,$6B,$E1,$70,$3E,$67,$93,$8F
	.dc.b	$98,$E2,$FF,$A6,$03,$E2,$78,$FC,$C7,$07,$5F,$31,$C3,$5F,$64,$7F
	.dc.b	$86,$23,$83,$F2,$B5,$FC,$4F,$1F,$17,$1C,$7E,$06,$E7,$E7,$AE,$10
	.dc.b	$FC,$6F,$83,$F5,$3E,$E3,$83,$6B,$83,$F7,$8A,$7D,$AE,$4D,$BF,$7D
	.dc.b	$F2,$BE,$5F,$7F,$AB,$F8,$3F,$7F,$6F,$E1,$9D,$BE,$4E,$57,$E7,$3F
	.dc.b	$EE,$C3,$FC,$26,$7D,$4C,$FF,$C7,$33,$FC,$6A,$31,$8F,$FB,$0D,$AE
	.dc.b	$67,$EA,$BF,$E2,$B8,$E5,$CD,$B9,$FB,$2F,$E8,$A3,$35,$C6,$7E,$C7
	.dc.b	$36,$DB,$79,$E1,$FA,$B0,$9E,$CF,$E1,$39,$78,$4E,$9F,$D7,$FE,$75
	.dc.b	$EB,$9F,$F6,$D7,$E7,$75,$FE,$6B,$F9,$E5,$D5,$FB,$05,$FC,$FB,$1F
	.dc.b	$9F,$E7,$F3,$DD,$7E,$7B,$BF,$D0,$26,$5F,$C0,$38,$5C,$BF,$D0,$27
	.dc.b	$F8,$4C,$39,$71,$5B,$8F,$E8,$BF,$A0,$F8,$4F,$F2,$19,$E7,$E6,$CC
	.dc.b	$F5,$AA,$FF,$83,$EB,$FA,$6F,$C4,$FA,$4F,$F2,$5F,$E7,$FF,$C9,$BE
	.dc.b	$A7,$F9,$77,$E4,$2F,$C8,$9E,$DB,$F4,$27,$E4,$0B,$FE,$AF,$87,$1F
	.dc.b	$49,$8F,$1C,$38,$5B,$EB,$7F,$AE,$FE,$87,$FE,$A9,$DB,$F8,$27,$4A
	.dc.b	$3F,$CE,$91,$8F,$D1,$3E,$71,$B7,$F9,$DC,$A3,$71,$86,$BF,$CA,$1C
	.dc.b	$F9,$1F,$E8,$C9,$7E,$43,$FD,$13,$F1,$97,$C4,$B7,$7C,$7B,$F4,$6F
	.dc.b	$EE,$7F,$A3,$3F,$46,$FE,$EB,$8F,$1C,$38,$5E,$F2,$5B,$F2,$37,$E9
	.dc.b	$EF,$90,$FE,$A7,$EC,$BF,$4A,$FE,$AF,$FA,$53,$F0,$4F,$E9,$1E,$BB
	.dc.b	$F4,$AE,$38,$70,$BD,$D6,$BE,$67,$AF,$67,$F5,$EF,$C9,$3F,$B3,$FE
	.dc.b	$C0,$FC,$ED,$F8,$2B,$F1,$67,$E9,$CF,$D3,$78,$F7,$FF,$AC,$31,$E3
	.dc.b	$87,$0B,$DF,$E9,$CF,$D4,$9F,$A8,$FF,$6C,$F9,$0F,$6E,$18,$FD,$0D
	.dc.b	$FA,$8F,$F7,$4F,$C1,$B8,$5E,$FA,$97,$BA,$FE,$F3,$F9,$CB,$F6,$27
	.dc.b	$D1,$FF,$78,$E3,$7C,$4B,$D8,$7E,$0F,$FE,$25,$FD,$4D,$F8,$53,$83
	.dc.b	$96,$3C,$70,$E1,$7B,$A7,$E5,$5D,$FD,$DF,$EE,$FF,$2B,$FE,$D0,$D7
	.dc.b	$F6,$7D,$3F,$DD,$31,$C7,$B9,$FC,$06,$7F,$69,$3F,$DE,$2B,$2E,$CF
	.dc.b	$54,$FF,$80,$AB,$FC,$03,$FB,$73,$ED,$BF,$C2,$7F,$AA,$FC,$BE,$A7
	.dc.b	$D1,$3E,$CD,$FA,$23,$9D,$F6,$0C,$B9,$6B,$FE,$E6,$CB,$96,$3C,$70
	.dc.b	$E1,$F6,$0F,$DB,$FF,$9B,$DF,$FB,$B3,$F2,$07,$E5,$6F,$71,$FA,$C3
	.dc.b	$F5,$87,$EA,$0F,$CA,$9E,$A7,$3F,$9A,$FF,$15,$E5,$CB,$1E,$38,$70
	.dc.b	$BE,$9F,$AE,$BF,$0D,$FF,$18,$FE,$51,$FE,$33,$F9,$77,$EB,$8F,$CF
	.dc.b	$DF,$AE,$3F,$5C,$63,$C7,$8C,$31,$E5,$87,$1B,$7E,$8A,$FD,$7B,$DB
	.dc.b	$7C,$3F,$28,$7E,$2C,$FC,$45,$F8,$A3,$F3,$8F,$DA,$FE,$BC,$FC,$33
	.dc.b	$97,$67,$F9,$23,$1E,$38,$70,$BD,$FE,$18,$FD,$C3,$F5,$9A,$79,$BF
	.dc.b	$2B,$FF,$29,$7E,$1B,$FE,$53,$FD,$81,$F8,$2B,$F6,$CF,$6E,$1C,$30
	.dc.b	$61,$C3,$EF,$98,$70,$BF,$67,$ED,$7F,$B8,$BF,$87,$FA,$AB,$F2,$FF
	.dc.b	$97,$FA,$6F,$E9,$7F,$11,$FD,$F2,$BE,$16,$BE,$1D,$6A,$E1,$F1,$3F
	.dc.b	$E6,$AF,$8C,$FC,$1D,$FD,$0F,$C7,$5F,$B2,$3F,$64,$7E,$72,$F8,$CC
	.dc.b	$7E,$52,$FC,$77,$F8,$6F,$A5,$FE,$70,$C7,$8E,$1C,$2F,$7F,$8E,$BF
	.dc.b	$74,$E9,$A7,$DE,$F9,$1F,$E9,$DF,$A3,$FC,$2F,$C2,$5F,$71,$FC,$67
	.dc.b	$E4,$78,$67,$FC,$47,$EC,$97,$BA,$FE,$3D,$2F,$DD,$5F,$A3,$FF,$A0
	.dc.b	$BC,$E7,$E5,$1F,$A1,$ED,$FF,$57,$F4,$3F,$35,$FC,$16,$1C,$38,$3F
	.dc.b	$23,$87,$0F,$69,$87,$0B,$79,$0F,$DA,$9F,$B3,$7F,$A4,$3F,$35,$FE
	.dc.b	$D1,$F3,$CF,$DA,$5F,$B8,$3F,$4C,$F9,$5F,$B5,$71,$C3,$87,$3F,$F3
	.dc.b	$9B,$DD,$7C,$FF,$A3,$7E,$EE,$FC,$27,$FD,$3B,$E6,$BE,$9B,$F4,$FF
	.dc.b	$90,$FF,$08,$FD,$5B,$FB,$37,$E6,$5C,$2F,$75,$AF,$94,$57,$F6,$BF
	.dc.b	$CC,$7F,$23,$F8,$67,$F3,$DF,$C0,$5F,$93,$FD,$2F,$A1,$F8,$DF,$AD
	.dc.b	$BF,$28,$DF,$81,$F6,$6F,$7E,$37,$F8,$73,$B9,$6F,$CA,$1F,$6F,$FD
	.dc.b	$1B,$0F,$DE,$FE,$0A,$FD,$6D,$FB,$67,$F3,$BF,$D4,$3F,$91,$FE,$15
	.dc.b	$FC,$F7,$0E,$18,$7F,$82,$EF,$7F,$5E,$BD,$FD,$67,$F7,$AF,$D3,$BF
	.dc.b	$D8,$3D,$EF,$CF,$5F,$A6,$3F,$5D,$7E,$D9,$FD,$47,$F5,$9F,$95,$F8
	.dc.b	$3B,$1E,$38,$70,$BD,$FD,$9A,$FF,$01,$FB,$CF,$F5,$2E,$0F,$E2,$FF
	.dc.b	$2B,$FA,$FF,$01,$FA,$03,$F6,$FF,$77,$ED,$EF,$D4,$98,$F1,$C7,$F9
	.dc.b	$AF,$C7,$7D,$4C,$38,$73,$B0,$FD,$8F,$A4,$FD,$47,$FD,$BB,$F6,$7F
	.dc.b	$DA,$7F,$32,$FE,$95,$F8,$9F,$F5,$EF,$CD,$DF,$B1,$7E,$42,$F7,$C9
	.dc.b	$FE,$E8,$FB,$8F,$F1,$07,$6A,$FE,$27,$EE,$2F,$DC,$7F,$DD,$3F,$58
	.dc.b	$FA,$DF,$81,$A7,$F0,$BF,$25,$FB,$57,$DD,$C3,$87,$37,$F3,$8C,$38
	.dc.b	$76,$7F,$8B,$79,$FF,$07,$7F,$E4,$CF,$DF,$3F,$BF,$DF,$F3,$EF,$65
	.dc.b	$FA,$3B,$F6,$D7,$EE,$6D,$1F,$4D,$FC,$77,$F1,$A6,$5C,$B2,$FE,$5D
	.dc.b	$8F,$1C,$38,$5F,$F2,$3E,$2F,$E3,$5F,$B6,$FF,$4C,$B7,$EB,$2F,$B9
	.dc.b	$7C,$4F,$BF,$E7,$3F,$C5,$5E,$DF,$F8,$AF,$E5,$18,$F1,$C3,$85,$EF
	.dc.b	$E2,$FF,$CF,$7F,$AF,$3F,$7E,$FE,$A3,$A7,$F5,$AF,$D8,$3E,$C7,$EE
	.dc.b	$3F,$D0,$FF,$24,$FE,$57,$F8,$7F,$C0,$BD,$DA,$F7,$D3,$FE,$5F,$FA
	.dc.b	$5F,$F1,$2F,$D6,$BF,$48,$7E,$D7,$EB,$F5,$3F,$86,$FE,$ED,$FE,$77
	.dc.b	$F5,$1F,$AC,$BF,$76,$F6,$E1,$C3,$5F,$ED,$B8,$70,$F9,$6F,$CA,$5E
	.dc.b	$FF,$5A,$ED,$D5,$FE,$94,$FC,$4A,$E1,$FE,$93,$C7,$FC,$46,$FF,$BA
	.dc.b	$4F,$EE,$EE,$3F,$30,$FF,$6D,$CF,$9C,$FF,$4E,$BE,$93,$FD,$6B,$3F
	.dc.b	$EE,$A9,$46,$BD,$1F,$B9,$F8,$CF,$F9,$0B,$F1,$9F,$F2,$0F,$E9,$CF
	.dc.b	$C6,$9E,$C7,$F8,$E7,$AB,$FC,$81,$5A,$9F,$83,$9F,$F9,$48,$FE,$9D
	.dc.b	$F8,$E7,$F9,$4B,$F1,$DF,$F2,$47,$D6,$72,$9D,$0F,$AE,$FF,$2E,$CE
	.dc.b	$BF,$30,$FF,$09,$DF,$FA,$7A,$BF,$CD,$1F,$A5,$7F,$31,$F1,$3F,$CD
	.dc.b	$1F,$25,$AF,$C8,$7F,$CE,$93,$9E,$1F,$F3,$97,$97,$CB,$F2,$5F,$F3
	.dc.b	$E6,$BF,$C9,$3F,$24,$7E,$3F,$FE,$35,$D3,$FD,$02,$7F,$88,$BF,$27
	.dc.b	$FF,$44,$56,$E7,$FD,$47,$EB,$7F,$A4,$FD,$EF,$F4,$CF,$E5,$AF,$CA
	.dc.b	$FF,$D7,$18,$FF,$A6,$0F,$F1,$5F,$E5,$DF,$EA,$0A,$DC,$FF,$8B,$CF
	.dc.b	$F5,$14,$FF,$8C,$3F,$31,$7C,$DF,$F8,$66,$3F,$50,$7F,$AA,$EB,$FD
	.dc.b	$81,$C5,$FE,$EE,$BE,$33,$FE,$C5,$AF,$F6,$1F,$37,$FB,$2B,$C5,$FF
	.dc.b	$58,$CF,$FB,$20,$ED,$3A,$FC,$C7,$FD,$9D,$FB,$5F,$FA,$DF,$F3,$3F
	.dc.b	$DB,$7E,$77,$FE,$42,$FC,$F7,$B4,$63,$0F,$19,$F1,$F3,$7E,$6F,$FE
	.dc.b	$C2,$8F,$F7,$0F,$8D,$F8,$C2,$1F,$E8,$C7,$FD,$CB,$87,$57,$9D,$72
	.dc.b	$7F,$CA,$31,$D9,$ED,$E1,$75,$9F,$F4,$A4,$7F,$A0,$5F,$F3,$7A,$9E
	.dc.b	$8F,$FB,$7B,$F1,$FF,$43,$F2,$BE,$E7,$FD,$A9,$FA,$27,$33,$FD,$27
	.dc.b	$1F,$65,$F7,$69,$F3,$5E,$31,$FC,$67,$F5,$7F,$F6,$E7,$E8,$DE,$33
	.dc.b	$F8,$E6,$9E,$E3,$F5,$7F,$B2,$EA,$BF,$CD,$7F,$A6,$3C,$3F,$BB,$1F
	.dc.b	$02,$FF,$92,$4E,$7D,$EF,$B9,$18,$AF,$C1,$FE,$9A,$FD,$37,$F6,$3F
	.dc.b	$B1,$F5,$C5,$8F,$5E,$85,$EF,$CF,$7E,$37,$FE,$F3,$F4,$7D,$B8,$7D
	.dc.b	$07,$DB,$CB,$FE,$80,$E4,$D4,$FD,$23,$F9,$C7,$A7,$FC,$1B,$F3,$2C
	.dc.b	$C7,$CF,$3F,$9E,$7D,$3F,$17,$F4,$58,$7C,$8F,$CD,$F8,$46,$0E,$87
	.dc.b	$FA,$3B,$AB,$80,$9F,$9F,$DB,$C2,$31,$3C,$0F,$F5,$A9,$F9,$19,$FA
	.dc.b	$A6,$CF,$AE,$F4,$FD,$4F,$F1,$07,$FB,$15,$FD,$B3,$F6,$3A,$3F,$2D
	.dc.b	$F1,$DB,$9F,$F5,$0C,$7E,$63,$E7,$3C,$55,$B8,$FE,$4F,$BE,$EB,$F0
	.dc.b	$1B,$69,$E3,$3F,$A1,$CB,$4F,$E8,$B3,$D4,$FE,$1C,$CC,$7F,$32,$F6
	.dc.b	$BF,$51,$BF,$C5,$CB,$EB,$D7,$F0,$2C,$11,$F9,$CE,$EF,$ED,$9C,$3B
	.dc.b	$AF,$8D,$8F,$6D,$3F,$84,$ED,$E7,$7F,$85,$ED,$9A,$FD,$E5,$F1,$7E
	.dc.b	$E1,$FD,$32,$FF,$71,$97,$82,$3E,$93,$CD,$FE,$F9,$AF,$37,$F5,$C7
	.dc.b	$F2,$3F,$C1,$3F,$70,$7D,$19,$FF,$3C,$C7,$C6,$AF,$96,$FF,$0B,$9F
	.dc.b	$9D,$1F,$D5,$F5,$FC,$34,$62,$7F,$0E,$FD,$3D,$BE,$3C,$77,$3A,$3F
	.dc.b	$95,$C7,$8C,$63,$7F,$C6,$57,$18,$E6,$78,$FD,$CE,$79,$7F,$AB,$63
	.dc.b	$F4,$5B,$FF,$2F,$FC,$47,$F8,$5E,$FF,$42,$7F,$A0,$47,$F0,$EC,$78
	.dc.b	$BF,$C4,$39,$35,$87,$3A,$F9,$BC,$9A,$7F,$47,$1A,$AF,$E2,$FC,$19
	.dc.b	$9F,$A7,$F7,$79,$7F,$53,$CB,$9C,$FF,$93,$6F,$FA,$BE,$8F,$9C,$7E
	.dc.b	$5F,$97,$C2,$FE,$67,$97,$DA,$FE,$6E,$B3,$87,$D3,$DF,$2F,$D5,$D7
	.dc.b	$2F,$E6,$79,$7F,$69,$7F,$A3,$8C,$C7,$E6,$5F,$E7,$71,$FA,$DE,$4F
	.dc.b	$E8,$3C,$BE,$59,$FE,$19,$FA,$07,$F5,$9A,$FC,$EC,$7F,$4B,$7F,$9B
	.dc.b	$DF,$8A,$3F,$27,$A7,$8B,$3F,$F2,$56,$99,$7F,$9E,$EB,$D6,$BF,$33
	.dc.b	$AE,$5F,$E1,$C7,$F0,$EC,$7E,$6E,$3E,$A8,$FE,$18,$7F,$12,$FD,$5B
	.dc.b	$F2,$C7,$C5,$70,$7C,$58,$FE,$29,$1F,$06,$32,$7F,$5F,$D7,$DE,$FF
	.dc.b	$80,$E3,$E1,$63,$F8,$27,$E2,$61,$F8,$E7,$E3,$4E,$4F,$E1,$3C,$DF
	.dc.b	$DF,$63,$F8,$B3,$D3,$77,$E0,$1F,$E6,$BF,$C8,$DF,$B0,$3F,$1C,$FF
	.dc.b	$02,$4E,$61,$FA,$D6,$32,$0B,$E3,$1F,$E8,$2F,$C9,$9F,$8C,$7F,$85
	.dc.b	$70,$E5,$39,$8F,$E8,$19,$7F,$AB,$B0,$FE,$8D,$F8,$C7,$F8,$4F,$EC
	.dc.b	$3F,$C0,$DC,$1F,$C0,$AB,$55,$FD,$1F,$83,$4F,$F4,$D7,$D9,$5F,$E0
	.dc.b	$DF,$DE,$F3,$3F,$D3,$7B,$F9,$4F,$F8,$47,$BF,$38,$FA,$9E,$35,$FD
	.dc.b	$5F,$E1,$72,$FF,$8C,$CF,$D8,$D7,$F8,$67,$BF,$27,$08,$FE,$B5,$F9
	.dc.b	$9F,$0F,$03,$BC,$7F,$61,$F0,$39,$9B,$CB,$FC,$1B,$1F,$E9,$BD,$BD
	.dc.b	$B1,$FB,$28,$FF,$21,$7E,$B9,$FE,$17,$3F,$90,$1F,$D3,$FF,$76,$FF
	.dc.b	$1C,$3F,$D4,$7C,$0F,$8C,$FF,$53,$F0,$33,$5F,$D2,$9F,$C0,$47,$F5
	.dc.b	$7E,$3C,$CF,$F8,$B7,$0E,$2F,$F5,$0E,$3C,$F7,$7F,$9F,$1E,$31,$FD
	.dc.b	$53,$87,$33,$F1,$63,$FB,$8E,$1F,$D8,$F8,$32,$70,$8F,$F1,$D7,$89
	.dc.b	$C1,$EE,$7F,$57,$AE,$FD,$FF,$E2,$1F,$0F,$FB,$E6,$1F,$E4,$AB,$1E
	.dc.b	$8C,$EE,$1C,$9F,$ED,$EF,$F9,$0A,$7F,$C9,$D8,$78,$67,$F2,$FB,$F1
	.dc.b	$AF,$F6,$99,$FC,$B2,$BF,$CA,$9B,$64,$E1,$0D,$B8,$5F,$78,$FF,$5D
	.dc.b	$1D,$CF,$C9,$61,$FD,$02,$FF,$CB,$DA,$F7,$E0,$E1,$B7,$84,$EB,$A7
	.dc.b	$8F,$0F,$CF,$7E,$BD,$FE,$55,$3F,$AF,$1F,$5D,$FF,$2E,$BF,$E0,$FC
	.dc.b	$B7,$1F,$5C,$C9,$F6,$72,$FE,$A3,$3F,$E1,$5A,$FF,$3F,$3F,$E1,$79
	.dc.b	$E2,$7D,$13,$89,$F4,$E7,$8B,$FE,$1A,$DF,$E1,$3A,$84,$7F,$7E,$8F
	.dc.b	$F3,$D4,$FF,$96,$A3,$E2,$C7,$F8,$3E,$31,$C1,$FC,$45,$D5,$F9,$27
	.dc.b	$B1,$E4,$B7,$81,$FE,$CC,$8F,$03,$F8,$57,$FD,$4D,$F9,$A7,$FA,$3B
	.dc.b	$2F,$C1,$C7,$F8,$76,$30,$FF,$09,$7F,$D2,$78,$F3,$8D,$1F,$E8,$67
	.dc.b	$A6,$89,$FF,$4C,$F4,$F4,$AF,$DB,$46,$1E,$87,$FA,$3B,$A7,$A3,$A3
	.dc.b	$19,$7F,$2F,$9F,$F5,$1D,$F3,$9F,$F2,$C4,$61,$E6,$7F,$A4,$76,$E6
	.dc.b	$E8,$FF,$4D,$BE,$68,$FF,$56,$F3,$F3,$9F,$86,$7F,$A8,$1F,$F5,$6F
	.dc.b	$89,$C6,$3F,$D3,$9D,$1D,$1D,$98,$C7,$CC,$FF,$85,$8F,$F5,$5D,$FF
	.dc.b	$98,$9F,$F5,$B7,$EA,$5F,$EA,$93,$FC,$64,$7F,$A6,$3A,$7C,$59,$A3
	.dc.b	$13,$FC,$E5,$FF,$4E,$F3,$F8,$67,$53,$FE,$66,$8F,$A3,$3F,$EA,$2E
	.dc.b	$0E,$6D,$9E,$47,$F9,$B3,$83,$94,$7F,$AA,$1E,$31,$FE,$70,$BE,$2D
	.dc.b	$38,$8C,$9B,$72,$8F,$F3,$BB,$C9,$FC,$87,$17,$FB,$97,$0F,$F5,$AC
	.dc.b	$7C,$C3,$FD,$EC,$FF,$62,$7F,$47,$5C,$DA,$FF,$6F,$8D,$73,$FF,$40
	.dc.b	$9F,$F4,$C4,$61,$E9,$87,$C4,$7F,$DA,$55,$CF,$2F,$F6,$23,$FE,$6A
	.dc.b	$AF,$F7,$6E,$9E,$EA,$FF,$6C,$3C,$A3,$FD,$E3,$A7,$13,$D9,$7F,$ED
	.dc.b	$DC,$BD,$97,$FE,$E6,$8E,$0F,$E8,$F0,$CB,$FE,$E3,$D3,$8E,$5D,$EB
	.dc.b	$FA,$04,$FF,$AC,$31,$FF,$7F,$E9,$E7,$3B,$C6,$CF,$6F,$EA,$7F,$EF
	.dc.b	$13,$FC,$F4,$7D,$FC,$BF,$DE,$D8,$23,$F0,$13,$FE,$60,$7E,$F6,$5A
	.dc.b	$6B,$FE,$6E,$7E,$AD,$66,$7F,$A1,$9F,$E0,$98,$FA,$DC,$DF,$81,$AD
	.dc.b	$4F,$D6,$79,$3F,$E1,$C8,$F8,$CF,$DD,$3F,$D9,$D3,$CB,$2F,$5C,$6F
	.dc.b	$3F,$CE,$4F,$67,$F8,$5B,$E0,$8F,$D2,$B5,$F2,$1F,$E2,$AE,$5F,$23
	.dc.b	$DC,$65,$F0,$DD,$63,$FC,$5B,$C7,$FC,$87,$99,$F8,$B9,$DC,$61,$FC
	.dc.b	$5B,$F9,$5F,$C9,$3A,$C7,$FB,$06,$31,$5C,$E3,$FC,$77,$F8,$37,$F9
	.dc.b	$07,$F3,$5F,$A7,$5F,$9D,$8F,$F2,$27,$93,$8A,$DB,$83,$FC,$91,$F0
	.dc.b	$1F,$EF,$3A,$E1,$19,$1E,$F6,$6B,$FC,$93,$F0,$6F,$D9,$64,$7C,$1D
	.dc.b	$98,$7E,$B6,$B5,$FB,$7F,$86,$91,$C6,$E0,$F9,$2E,$D3,$A8,$C8,$F9
	.dc.b	$1C,$FE,$83,$F2,$72,$EF,$5F,$E5,$C8,$FF,$19,$46,$21,$FB,$17,$CA
	.dc.b	$7F,$C7,$19,$FF,$01,$8F,$DA,$F9,$3E,$0F,$99,$F9,$55,$EB,$AF,$E0
	.dc.b	$D1,$F2,$A3,$07,$F9,$A9,$F8,$DB,$B8,$39,$3F,$27,$83,$D7,$F4,$FE
	.dc.b	$FB,$E2,$1F,$8A,$E1,$F7,$ED,$F0,$E3,$F0,$A7,$F9,$E2,$33,$B7,$C6
	.dc.b	$F9,$BF,$85,$9F,$F1,$B7,$E0,$BF,$C7,$7F,$63,$F8,$17,$F3,$1F,$7B
	.dc.b	$D0,$7E,$1D,$F0,$1F,$C7,$1F,$E8,$67,$E0,$38,$37,$9F,$E1,$74,$FA
	.dc.b	$F3,$F0,$B4,$FB,$3F,$31,$EF,$BE,$0F,$B9,$BA,$FD,$CE,$3D,$BF,$7B
	.dc.b	$5F,$C0,$FE,$C9,$EC,$AF,$F4,$BB,$EC,$7F,$D3,$3E,$CF,$B1,$F5,$3E
	.dc.b	$CE,$4F,$C0,$7C,$BF,$AE,$7E,$F9,$FE,$A0,$7D,$6E,$03,$F6,$B5,$FC
	.dc.b	$8B,$EB,$8F,$8D,$F9,$7F,$AD,$E0,$FA,$DB,$97,$AD,$DB,$EB,$7E,$77
	.dc.b	$C0,$FA,$D3,$EB,$5F,$B2,$FA,$F1,$FC,$C3,$EB,$E3,$FE,$AD,$FB,$37
	.dc.b	$AE,$BF,$4A,$FA,$9B,$C7,$F4,$DF,$B9,$1E,$9B,$FD,$17,$D7,$7A,$67
	.dc.b	$F5,$AF,$A5,$B9,$7A,$67,$6F,$A5,$FD,$57,$E0,$1E,$93,$D1,$F6,$5E
	.dc.b	$7C,$BF,$B5,$3E,$7B,$FF,$66,$FD,$57,$9D,$FC,$73,$E7,$6E,$7E,$7C
	.dc.b	$3F,$6D,$3E,$7B,$FE,$E9,$F6,$BE,$7B,$FC,$B3,$E7,$6E,$5E,$79,$DB
	.dc.b	$E7,$7F,$BE,$7C,$5F,$CF,$3E,$77,$E7,$5E,$7A,$FF,$70,$BE,$75,$F5
	.dc.b	$BE,$7C,$BF,$84,$3E,$76,$E7,$E7,$C3,$FC,$11,$3E,$7A,$FF,$75,$FD
	.dc.b	$5F,$9E,$FF,$88,$3E,$76,$E5,$E7,$9D,$BE,$77,$FD,$E7,$F8,$77,$E1
	.dc.b	$4F,$C2,$7E,$5F,$E7,$83,$E7,$FC,$17,$CD,$97,$F8,$65,$F3,$37,$3F
	.dc.b	$35,$7F,$BF,$A7,$CC,$FF,$A8,$F0,$FE,$6E,$FE,$EF,$D8,$F3,$5F,$F8
	.dc.b	$7A,$7C,$D3,$B7,$F9,$AF,$67,$F3,$4F,$4B,$F3,$47,$F2,$EF,$53,$FC
	.dc.b	$55,$E7,$7E,$65,$AF,$F9,$D1,$FF,$17,$7A,$3F,$15,$B6,$F3,$7F,$C1
	.dc.b	$31,$88,$FE,$87,$E1,$32,$FE,$D7,$39,$9F,$E9,$19,$67,$87,$FD,$13
	.dc.b	$FA,$C7,$D9,$AF,$F1,$E3,$F6,$5B,$C1,$F8,$F6,$BF,$D8,$7C,$5E,$51
	.dc.b	$FE,$43,$F2,$FF,$A6,$37,$2F,$F1,$EB,$FE,$47,$FD,$6B,$DD,$1E,$57
	.dc.b	$F8,$84,$7F,$92,$FF,$67,$F7,$4F,$A5,$FC,$14,$67,$4F,$D7,$35,$F9
	.dc.b	$B3,$EE,$3E,$07,$ED,$1E,$B6,$E4,$3E,$3B,$87,$FB,$D7,$8D,$EA,$8C
	.dc.b	$47,$F8,$03,$3C,$43,$E9,$FD,$AF,$CD,$7E,$D3,$F3,$FF,$5E,$F6,$A7
	.dc.b	$DA,$FD,$3F,$98,$F6,$A7,$DA,$E7,$2F,$6B,$5F,$6F,$07,$FA,$FB,$F1
	.dc.b	$3F,$64,$9F,$AC,$FD,$F7,$F9,$93,$F0,$CD,$65,$FD,$AD,$C7,$CE,$FA
	.dc.b	$93,$EA,$5F,$88,$FA,$8F,$4A,$FC,$B7,$D2,$7A,$1C,$63,$E8,$6B,$E8
	.dc.b	$E0,$FF,$43,$7D,$EF,$A0,$F3,$C9,$F3,$E7,$FE,$DA,$7F,$98,$4F,$FA
	.dc.b	$37,$F1,$7F,$31,$E5,$FC,$EB,$E6,$9E,$4F,$98,$F9,$67,$8C,$FF,$00
	.dc.b	$F5,$7D,$87,$FB,$FE,$3F,$0F,$F0,$EF,$92,$78,$8F,$0E,$1F,$E9,$B7
	.dc.b	$C3,$1F,$A9,$FB,$EF,$8C,$7C,$5F,$DA,$7E,$E9,$F1,$3F,$30,$FB,$67
	.dc.b	$C3,$3C,$33,$F4,$5F,$0E,$BF,$EF,$6F,$C6,$BE,$C9,$F0,$8F,$0C,$FE
	.dc.b	$DD,$F0,$AF,$B1,$FA,$C7,$D5,$FC,$C7,$F1,$A7,$D4,$FD,$AB,$F1,$67
	.dc.b	$D3,$3C,$36,$D3,$FD,$6D,$1E,$27,$FD,$71,$FA,$7F,$E2,$8F,$A6,$78
	.dc.b	$B0,$3E,$2F,$DA,$FF,$0C,$7E,$17,$F6,$CF,$CE,$9F,$84,$FA,$9F,$CD
	.dc.b	$9F,$83,$3C,$37,$FE,$C5,$7C,$17,$FC,$4F,$F6,$DF,$CC,$9F,$82,$3C
	.dc.b	$07,$C0,$F9,$4F,$CA,$1F,$93,$F9,$4F,$D5,$9F,$92,$FD,$5B,$F5,$27
	.dc.b	$E4,$4E,$F5,$FC,$7D,$EE,$BF,$14,$FD,$39,$F9,$03,$DC,$7C,$EF,$BB
	.dc.b	$FD,$11,$FA,$1F,$DB,$BF,$72,$7E,$83,$F2,$7F,$DB,$9F,$9F,$3B,$D1
	.dc.b	$EF,$F8,$BF,$ED,$0F,$CE,$9F,$2C,$FB,$7E,$47,$EB,$CF,$D7,$7E,$11
	.dc.b	$FC,$18,$FD,$6F,$DE,$FF,$04,$3F,$58,$76,$3D,$7E,$37,$F0,$03,$F5
	.dc.b	$27,$D7,$3F,$0F,$E5,$FE,$EC,$FD,$D7,$ED,$FF,$C4,$8F,$DC,$FE,$6F
	.dc.b	$FC,$40,$FD,$C1,$EA,$3E,$3F,$EC,$DF,$C3,$0F,$DA,$9F,$83,$3F,$27
	.dc.b	$F3,$5F,$C1,$CF,$E0,$DF,$35,$FC,$78,$FE,$0B,$F3,$1F,$C7,$0F,$E0
	.dc.b	$87,$C6,$3E,$9F,$CB,$7F,$18,$3F,$80,$9F,$8F,$3F,$3F,$F2,$BF,$C4
	.dc.b	$CF,$E2,$5F,$2B,$FC,$A8,$FE,$23,$F2,$9F,$CA,$0F,$E2,$07,$D2,$3E
	.dc.b	$FF,$C9,$7F,$24,$3F,$86,$9F,$9B,$3F,$57,$F2,$3F,$C7,$CF,$E3,$DF
	.dc.b	$23,$FC,$D8,$FE,$3B,$F2,$1F,$CD,$0F,$E3,$87,$DE,$3F,$17,$F1,$DF
	.dc.b	$CC,$0F,$E3,$27,$E9,$CF,$DB,$FC,$6F,$F2,$B3,$F9,$57,$C6,$FF,$42
	.dc.b	$3F,$94,$FC,$67,$F4,$03,$F9,$41,$F8,$A3,$F2,$FF,$15,$FC,$F0,$FE
	.dc.b	$4A,$7E,$CC,$FE,$07,$F1,$1F,$CD,$CF,$F5,$AF,$6F,$F8,$1B,$E2,$3F
	.dc.b	$DD,$0F,$FB,$A0,$C3,$FE,$FF,$F8,$4F,$EF,$D1,$FD,$EB,$E6,$B7,$E7
	.dc.b	$7F,$BF,$7E,$1B,$FC,$6D,$3F,$DF,$3B,$7F,$DB,$FD,$BC,$23,$FC,$15
	.dc.b	$3F,$E6,$37,$FB,$81,$FE,$D7,$3F,$BA,$7C,$47,$47,$FB,$A7,$BB,$FE
	.dc.b	$D5,$F6,$B9,$43,$E1,$BF,$91,$BF,$E1,$53,$FB,$61,$FC,$5C,$E8,$7F
	.dc.b	$6D,$34,$7F,$86,$FE,$8B,$9F,$C2,$FF,$9F,$E3,$93,$FE,$99,$7A,$7C
	.dc.b	$27,$F9,$4D,$BA,$7F,$80,$5E,$A7,$F6,$47,$FC,$02,$FC,$93,$FB,$53
	.dc.b	$FD,$9F,$5F,$0F,$C1,$7F,$95,$67,$D0,$FF,$95,$5B,$9F,$FA,$05,$AF
	.dc.b	$F7,$87,$FA,$D1,$ED,$3D,$73,$EC,$F8,$1F,$B7,$1F,$E3,$7F,$81,$F6
	.dc.b	$37,$8C,$42,$F1,$FC,$CF,$80,$F7,$4F,$F8,$FD,$FB,$6D,$CF,$FD,$0C
	.dc.b	$FF,$5D,$7D,$8F,$D1,$F7,$BD,$8F,$F9,$1D,$F6,$37,$3F,$63,$5F,$63
	.dc.b	$FE,$4D,$F7,$7F,$C9,$DE,$C7,$E0,$63,$07,$D8,$1C,$3F,$82,$F7,$7D
	.dc.b	$B0,$7E,$C3,$FE,$57,$F7,$3D,$6F,$F5,$2C,$BE,$C3,$F2,$0F,$EB,$2F
	.dc.b	$B1,$E0,$FB,$1F,$F3,$07,$B5,$EC,$4F,$B1,$B9,$7B,$1A,$FB,$17,$B3
	.dc.b	$EC,$7F,$0E,$FD,$65,$EC,$FA,$8F,$4B,$FD,$3E,$DD,$9F,$0E,$31,$17
	.dc.b	$D8,$FF,$0F,$C7,$E8,$5F,$D6,$3F,$E7,$7F,$5B,$D0,$79,$CF,$EF,$6F
	.dc.b	$E9,$DF,$D4,$BF,$E7,$DF,$57,$CC,$FC,$53,$07,$D1,$DB,$F6,$3E,$AF
	.dc.b	$D0,$3C,$F8,$7C,$E3,$F0,$F3,$FE,$65,$7E,$63,$5F,$F3,$F3,$FA,$51
	.dc.b	$E9,$FC,$D3,$E6,$1E,$67,$FD,$20,$F9,$97,$A5,$E6,$7F,$80,$75,$7E
	.dc.b	$F2,$3E,$67,$A5,$F8,$43,$E5,$FA,$3F,$82,$3C,$A7,$C8,$AF,$F4,$E3
	.dc.b	$F2,$17,$9F,$F7,$CF,$BC,$78,$DC,$F1,$7F,$BC,$DF,$E2,$1E,$77,$8D
	.dc.b	$FE,$2A,$FD,$DF,$3B,$F2,$A7,$DC,$F3,$7F,$26,$7C,$63,$C5,$27,$C5
	.dc.b	$E5,$FE,$3C,$FC,$71,$E1,$3E,$17,$95,$DE,$7E,$53,$FC,$E3,$87,$FD
	.dc.b	$6D,$F3,$FF,$BE,$9F,$F5,$C7,$83,$FB,$F7,$F5,$51,$97,$F8,$14,$4F
	.dc.b	$07,$8B,$FC,$D2,$3F,$93,$CE,$FC,$7F,$F8,$00,$DD,$47,$FB,$1A,$51
	.dc.b	$98,$FF,$55,$BD,$FC,$6F,$F5,$47,$57,$F4,$E8,$C7,$37,$ED,$2F,$F9
	.dc.b	$D4,$62,$3F,$95,$B5,$FF,$55,$47,$6C,$7F,$91,$F8,$1C,$DF,$E4,$B1
	.dc.b	$FC,$FB,$4F,$5B,$FC,$86,$34,$D7,$FA,$E3,$CC,$FE,$28,$7C,$67,$F8
	.dc.b	$E6,$5F,$EE,$58,$C1,$FC,$65,$FE,$A8,$F1,$1E,$17,$FA,$1B,$1F,$9D
	.dc.b	$E0,$7C,$09,$E2,$7F,$13,$CB,$C6,$D7,$FB,$3B,$CE,$3F,$51,$1F,$EC
	.dc.b	$FE,$BF,$5C,$FF,$0C,$3F,$A1,$1F,$23,$1F,$EC,$5D,$FF,$83,$8F,$27
	.dc.b	$7F,$E0,$5F,$B7,$3F,$D9,$E3,$F0,$2F,$F4,$28,$F5,$77,$7D,$CD,$3C
	.dc.b	$AF,$E0,$27,$F8,$44,$64,$FC,$0F,$91,$E4,$9F,$81,$3F,$DC,$F2,$F8
	.dc.b	$17,$E0,$07,$F0,$33,$E1,$BF,$EC,$ED,$3F,$28,$FC,$01,$9F,$C0,$EC
	.dc.b	$FD,$04,$FE,$65,$F7,$3F,$9B,$EC,$F7,$5F,$C3,$3B,$BE,$EC,$1C,$BF
	.dc.b	$2B,$AF,$DD,$5F,$51,$F7,$3E,$67,$FD,$63,$7F,$A7,$7D,$B1,$EA,$E9
	.dc.b	$F6,$BD,$CF,$DD,$67,$F9,$B7,$EB,$C7,$B7,$0F,$F0,$B3,$ED,$CF,$ED
	.dc.b	$F4,$7B,$27,$FC,$34,$FB,$3E,$7F,$FC,$69,$7F,$B8,$7F,$C6,$73,$F5
	.dc.b	$F2,$47,$F8,$53,$9F,$F5,$8F,$F8,$5B,$2B,$7F,$E1,$27,$F5,$71,$97
	.dc.b	$F5,$75,$CA,$7F,$42,$7F,$00,$9F,$D0,$F9,$5C,$A3,$F3,$BC,$5F,$81
	.dc.b	$1C,$5F,$C4,$2B,$F4,$F1,$FC,$01,$FC,$9C,$FE,$58,$D1,$F8,$2E,$4F
	.dc.b	$73,$F8,$4E,$4F,$B2,$E8,$FE,$CB,$A7,$E5,$B8,$71,$A7,$D0,$9A,$FC
	.dc.b	$F7,$F8,$8E,$1F,$91,$8F,$D4,$F9,$9A,$7F,$A8,$EB,$EF,$AF,$E7,$79
	.dc.b	$FB,$78,$BF,$8E,$E1,$FE,$4B,$7F,$02,$D8,$F8,$1F,$17,$E5,$E1,$FC
	.dc.b	$9C,$FD,$8B,$FE,$CC,$7F,$22,$F8,$CF,$B2,$FF,$2A,$7E,$CE,$BF,$92
	.dc.b	$8D,$1F,$40,$FE,$43,$1F,$41,$FE,$6E,$FC,$CC,$3F,$87,$E4,$8C,$47
	.dc.b	$F3,$2C,$7F,$A0,$C7,$C1,$E2,$FE,$5C,$FC,$3D,$2E,$1F,$B2,$7F,$61
	.dc.b	$19,$7F,$06,$FA,$FA,$F7,$2D,$3F,$5F,$FB,$27,$B2,$3F,$9B,$BE,$CE
	.dc.b	$0F,$EA,$2F,$86,$3F,$A7,$BE,$16,$9E,$1C,$3E,$AB,$FD,$22,$3C,$51
	.dc.b	$FC,$7E,$3F,$A5,$FE,$79,$CD,$FE,$A6,$F3,$69,$CE,$3F,$A3,$C7,$27
	.dc.b	$F8,$4C,$7F,$8B,$B7,$FA,$3B,$FE,$A3,$8B,$F0,$4F,$F0,$51,$FA,$8E
	.dc.b	$EB,$FB,$4C,$7F,$09,$FB,$8C,$46,$FE,$41,$FB,$C6,$2B,$F8,$77,$D9
	.dc.b	$BC,$3C,$47,$F6,$ED,$9F,$00,$FD,$13,$33,$FE,$73,$F4,$DF,$98,$3F
	.dc.b	$62,$CC,$FE,$FB,$F2,$D7,$E3,$AF,$43,$3F,$B3,$BF,$0C,$7D,$DB,$E3
	.dc.b	$F8,$FC,$F3,$F0,$77,$EC,$1D,$BA,$BE,$0E,$BF,$E8,$4E,$D7,$E6,$AF
	.dc.b	$8D,$7E,$AC,$F5,$FC,$FF,$00,$FC,$21,$F2,$6F,$D3,$1F,$B9,$FC,$0F
	.dc.b	$15,$FA,$6B,$F4,$37,$C6,$BF,$6E,$7E,$31,$F1,$BE,$91,$E2,$3D,$27
	.dc.b	$E0,$AF,$D2,$5E,$7F,$CC,$FD,$21,$F2,$5E,$3F,$31,$F7,$2F,$CC,$FE
	.dc.b	$A7,$ED,$2F,$C0,$7E,$27,$EE,$3F,$E2,$FF,$C5,$1F,$71,$F6,$7C,$47
	.dc.b	$E2,$0F,$DA,$DF,$58,$FD,$5D,$FB,$DF,$FA,$AB,$C9,$FE,$15,$F0,$2F
	.dc.b	$C2,$7F,$13,$FB,$DB,$F7,$57,$D2,$BF,$23,$7D,$1F,$E2,$BC,$7F,$67
	.dc.b	$B4,$FD,$C9,$F9,$A3,$F6,$57,$EA,$EF,$DC,$BF,$D0,$FE,$AF,$FB,$7E
	.dc.b	$B6,$43,$8F,$DA,$FF,$A3,$3F,$42,$FF,$72,$7E,$0E,$FB,$B7,$EA,$FF
	.dc.b	$47,$ED,$DF,$62,$F8,$36,$7F,$E9,$4F,$DE,$3F,$DD,$9F,$85,$3F,$63
	.dc.b	$FF,$1F,$FE,$27,$F2,$FF,$74,$7E,$30,$FD,$7B,$FD,$37,$7F,$EF,$3F
	.dc.b	$B3,$EF,$F2,$FF,$A0,$FD,$93,$F3,$92,$FB,$3F,$F5,$0C,$7F,$BE,$7E
	.dc.b	$F1,$FA,$A7,$C0,$FD,$73,$F0,$BF,$19,$E9,$FB,$77,$F0,$0B,$EE,$7F
	.dc.b	$E0,$07,$E1,$EE,$17,$D9,$3F,$31,$7E,$16,$FD,$CF,$EE,$FE,$95,$CB
	.dc.b	$E1,$7F,$C2,$7F,$4A,$FC,$CD,$F1,$3F,$E2,$DF,$25,$FB,$2F,$F5,$EF
	.dc.b	$8E,$FC,$FF,$DF,$F0,$F0,$7F,$59,$D8,$FD,$B5,$FB,$FB,$A5,$F9,$8C
	.dc.b	$F8,$17,$E2,$1E,$0E,$97,$E4,$2F,$CF,$5F,$81,$3F,$04,$7E,$3C,$FD
	.dc.b	$37,$F0,$7F,$56,$FD,$E3,$D3,$F9,$E7,$E2,$3E,$A9,$FA,$DB,$F7,$E7
	.dc.b	$EC,$AF,$D1,$DF,$AF,$FD,$1F,$CC,$FE,$0F,$DC,$7F,$00,$FD,$FD,$E8
	.dc.b	$BE,$5D,$FB,$67,$DE,$F4,$9F,$A6,$7F,$03,$F2,$3F,$C7,$7F,$91,$BF
	.dc.b	$6A,$FF,$5A,$FE,$F6,$FD,$47,$FA,$47,$97,$FC,$07,$E0,$DF,$9E,$7D
	.dc.b	$0F,$86,$7E,$D3,$F9,$EF,$9F,$FD,$F7,$E9,$0F,$D9,$9F,$BD,$3B,$9F
	.dc.b	$A8,$7D,$8F,$97,$FD,$27,$E7,$0F,$CC,$BF,$8B,$7E,$A2,$FC,$D3,$97
	.dc.b	$E9,$FF,$D0,$B9,$DF,$27,$FC,$BB,$F3,$C7,$EB,$8F,$D8,$1D,$6F,$D4
	.dc.b	$9F,$A3,$BF,$2F,$7E,$47,$F8,$8F,$CB,$FF,$0B,$F9,$67,$FB,$6B,$D8
	.dc.b	$FA,$3F,$8F,$FE,$5B,$F2,$27,$E5,$BF,$8E,$F8,$EF,$F2,$8F,$D4,$3F
	.dc.b	$3C,$FC,$1F,$ED,$4F,$D1,$DF,$AA,$BF,$55,$7E,$C5,$F9,$6F,$99,$FE
	.dc.b	$65,$F8,$D3,$A7,$AE,$9E,$4F,$DE,$9F,$92,$7B,$BB,$7F,$CD,$FF,$97
	.dc.b	$7E,$FB,$AF,$F6,$DF,$A3,$FE,$8F,$E7,$BF,$2B,$F9,$57,$F4,$7B,$FF
	.dc.b	$3A,$FC,$2F,$F9,$9B,$D7,$7A,$7F,$D3,$7F,$22,$7D,$DB,$BF,$FD,$F3
	.dc.b	$C7,$ED,$FE,$00,$FE,$2D,$F7,$FE,$2F,$7B,$FA,$87,$E1,$1B,$FD,$0F
	.dc.b	$F5,$FF,$B8,$7F,$5E,$FA,$AF,$D6,$FE,$A7,$FD,$3F,$F3,$57,$EF,$EF
	.dc.b	$CD,$1F,$BE,$FE,$DB,$F4,$A7,$E6,$1F,$CD,$7E,$31,$F6,$AF,$E0,$0E
	.dc.b	$1E,$CB,$F4,$BF,$E9,$3D,$9F,$E9,$0F,$BB,$FE,$73,$F1,$1F,$AF,$FC
	.dc.b	$43,$F9,$FF,$E3,$CF,$D6,$5F,$A0,$3F,$5D,$FE,$0B,$F7,$FF,$8B,$F9
	.dc.b	$E7,$F0,$07,$F3,$DF,$DA,$77,$FA,$9F,$E3,$3E,$63,$F8,$5F,$E8,$CF
	.dc.b	$D2,$9F,$B9,$3F,$78,$7E,$9C,$FD,$67,$F5,$BF,$56,$FD,$4F,$F8,$2F
	.dc.b	$3F,$D2,$FC,$B3,$F1,$FE,$8F,$F0,$CF,$D9,$7F,$29,$F9,$3B,$F4,$6F
	.dc.b	$F0,$3D,$FF,$AB,$3F,$51,$FF,$1B,$7E,$73,$FC,$8F,$ED,$7F,$2F,$FB
	.dc.b	$43,$F6,$5F,$EF,$5F,$BE,$FF,$20,$ED,$7E,$E7,$FD,$D3,$F5,$EF,$F7
	.dc.b	$CF,$E0,$4F,$D7,$DF,$61,$FE,$41,$F2,$FF,$93,$FE,$22,$FC,$4B,$FC
	.dc.b	$89,$F3,$4F,$5B,$F8,$4F,$E3,$1F,$CE,$B9,$3E,$A9,$FE,$C0,$BE,$1F
	.dc.b	$A4,$7E,$AE,$B7,$F6,$6F,$D6,$DC,$5E,$5E,$1F,$27,$E9,$6E,$1F,$27
	.dc.b	$03,$F7,$9F,$83,$AB,$A7,$E0,$FE,$29,$FC,$BB,$E8,$5D,$3F,$07,$F5
	.dc.b	$EF,$9B,$6F,$8C,$FF,$0C,$D7,$F1,$CF,$C2,$5E,$0F,$CE,$F7,$BF,$90
	.dc.b	$DF,$FB,$63,$CC,$FB,$9E,$5F,$DB,$F2,$BE,$D7,$9B,$FC,$E3,$F2,$DF
	.dc.b	$C4,$5F,$FB,$7B,$D3,$FC,$47,$A5,$F8,$7F,$11,$E8,$FF,$7F,$F2,$BF
	.dc.b	$A8,$FB,$3F,$94,$FC,$19,$EC,$FE,$53,$D8,$7E,$BD,$FA,$2D,$3F,$DD
	.dc.b	$3F,$03,$FA,$2F,$CB,$BF,$C9,$DE,$5F,$E8,$7E,$0F,$F9,$07,$DA,$BE
	.dc.b	$27,$F5,$FF,$A4,$3E,$27,$F5,$FF,$A1,$3E,$2B,$F5,$FF,$92,$BE,$47
	.dc.b	$F7,$4F,$F8,$83,$E6,$FF,$A1,$7E,$89,$FE,$2E,$FA,$1F,$E0,$DC,$3F
	.dc.b	$27,$F4,$DF,$E8,$7F,$33,$FC,$17,$E6,$3F,$82,$7E,$C7,$FC,$46,$9F
	.dc.b	$3F,$FC,$47,$F2,$27,$CF,$FF,$11,$FC,$49,$FB,$57,$F1,$EB,$1E,$3F
	.dc.b	$D1,$5F,$4F,$FC,$7F,$EB,$BF,$BE,$FA,$0F,$A6,$FF,$2A,$C7,$F2,$DE
	.dc.b	$0F,$F1,$6F,$D6,$FF,$2C,$FA,$CF,$E5,$7F,$B1,$FF,$38,$CF,$AF,$D2
	.dc.b	$FF,$38,$F7,$7F,$C1,$5F,$87,$BE,$CB,$FC,$1B,$D3,$FD,$1F,$EE,$7F
	.dc.b	$A3,$7D,$CF,$F4,$5F,$AD,$FE,$89,$F6,$EF,$DF,$BF,$A9,$E9,$E8,$FB
	.dc.b	$FF,$EA,$7F,$7D,$FD,$4B,$E9,$7F,$AD,$7E,$93,$FD,$82,$BD,$BF,$83
	.dc.b	$FF,$60,$FC,$79,$F8,$3F,$F6,$0F,$3B,$FA,$FF,$E6,$1F,$DB,$6F,$FA
	.dc.b	$9F,$E1,$BF,$DB,$3F,$0C,$FE,$D7,$F8,$F7,$E9,$BF,$10,$FE,$F3,$A7
	.dc.b	$ED,$BF,$10,$FE,$F1,$F0,$9F,$DD,$FF,$06,$FF,$03,$65,$F3,$3F,$0B
	.dc.b	$FF,$03,$75,$FF,$81,$7F,$13,$FF,$02,$7E,$24,$FD,$FB,$F4,$BF,$84
	.dc.b	$7F,$85,$3E,$0F,$FC,$27,$F8,$CF,$F8,$4B,$F1,$8F,$F2,$9F,$EC,$DF
	.dc.b	$B3,$D3,$EB,$FE,$41,$FE,$22,$FC,$7F,$FC,$43,$F8,$FF,$F8,$83,$D3
	.dc.b	$FE,$09,$F9,$1F,$F8,$BB,$F2,$3F,$F1,$6F,$93,$F7,$BB,$FF,$C5,$BF
	.dc.b	$AB,$7F,$A1,$7E,$9B,$FC,$79,$A7,$E0,$BF,$29,$FF,$1E,$7E,$19,$FE
	.dc.b	$3B,$FC,$A3,$FC,$75,$F2,$EF,$C9,$DF,$4D,$FE,$18,$C3,$F1,$3F,$95
	.dc.b	$FF,$92,$B4,$FF,$19,$7E,$8A,$FC,$B3,$FC,$95,$7F,$A3,$BF,$C7,$FD
	.dc.b	$97,$F9,$5B,$F5,$2F,$F1,$0F,$95,$FA,$7F,$B1,$EB,$D3,$FE,$62,$FC
	.dc.b	$BF,$FC,$C3,$F9,$FF,$F9,$3F,$F3,$0F,$F3,$06,$6F,$B5,$FF,$37,$FA
	.dc.b	$5F,$E6,$EF,$D1,$FF,$B4,$FC,$FF,$F9,$B7,$E9,$3F,$DB,$7C,$1F,$9E
	.dc.b	$FC,$D7,$FC,$F0,$7F,$9E,$BF,$35,$F5,$FE,$95,$FE,$54,$FC,$E7,$FD
	.dc.b	$0B,$F9,$CF,$FA,$13,$F3,$9F,$F4,$1F,$CC,$EB,$1F,$D3,$FE,$7D,$FE
	.dc.b	$91,$FD,$51,$F9,$F7,$FA,$47,$F3,$DF,$F4,$87,$E4,$BB,$BF,$85,$F9
	.dc.b	$BF,$FA,$7B,$EA,$3F,$D3,$BF,$A0,$FF,$A7,$3C,$0D,$1F,$C4,$BA,$BF
	.dc.b	$D4,$FF,$85,$7E,$5B,$F4,$2C,$CB,$F4,$6F,$F5,$8F,$E9,$8F,$D1,$FF
	.dc.b	$D6,$3F,$A3,$6B,$F3,$7F,$F5,$DF,$A9,$FE,$BA,$FD,$27,$FD,$73,$E1
	.dc.b	$7F,$B4,$3A,$FF,$D8,$BF,$11,$FE,$C4,$F0,$7F,$D8,$73,$FE,$D4,$D5
	.dc.b	$FA,$6F,$FB,$3F,$F1,$4F,$F6,$77,$E3,$BE,$29,$FF,$4B,$FC,$57,$FB
	.dc.b	$5F,$F4,$EF,$F6,$B6,$1C,$65,$FA,$7F,$FB,$87,$A7,$FD,$DB,$39,$FD
	.dc.b	$1B,$FD,$D1,$FA,$07,$FB,$9E,$38,$6F,$FD,$17,$F5,$2F,$F7,$74,$73
	.dc.b	$C7,$F8,$9F,$DB,$7F,$BD,$63,$1F,$A7,$7F,$BF,$3F,$53,$F2,$57,$E3
	.dc.b	$CF,$89,$F9,$46,$1F,$C9,$7E,$A8,$FC,$C7,$11,$F9,$5F,$D1,$EA,$F2
	.dc.b	$2E,$BF,$EF,$FF,$17,$FD,$2F,$F2,$EF,$D9,$BF,$D4,$5F,$4E,$FD,$9B
	.dc.b	$FD,$0D,$FB,$0B,$F7,$3E,$18,$7F,$2C,$C3,$87,$CA,$7F,$8E,$6F,$7F
	.dc.b	$53,$AA,$57,$5E,$DF,$F5,$2C,$7F,$A9,$BF,$58,$B3,$CA,$32,$71,$7A
	.dc.b	$E7,$E3,$F9,$4F,$E5,$93,$9D,$7C,$F9,$7F,$A6,$FF,$58,$FE,$7F,$8F
	.dc.b	$49,$FF,$5E,$7C,$8B,$CB,$7B,$2F,$87,$FF,$1E,$5E,$F8,$3F,$D7,$15
	.dc.b	$A9,$7E,$BD,$DE,$73,$7C,$74,$C6,$3D,$61,$E5,$F8,$A7,$E6,$E7,$E2
	.dc.b	$AF,$97,$C7,$C1,$F8,$FA,$78,$5F,$F6,$46,$FE,$E8,$C4,$7B,$6F,$C5
	.dc.b	$FA,$EF,$FB,$EF,$E7,$7F,$9F,$FE,$41,$8F,$77,$FD,$6F,$F4,$CF,$DD
	.dc.b	$B1,$5F,$D1,$30,$BE,$8F,$95,$FB,$5B,$7F,$B8,$7F,$0A,$FD,$2F,$CC
	.dc.b	$3F,$7D,$D5,$7D,$AA,$D7,$A5,$FD,$76,$7F,$CD,$3C,$1F,$27,$0E,$3F
	.dc.b	$81,$F0,$ED,$FC,$2C,$74,$F7,$31,$8F,$FB,$BF,$DB,$70,$FD,$3F,$29
	.dc.b	$FA,$3F,$CF,$9C,$F6,$A7,$FD,$BB,$A7,$D7,$9C,$CF,$4E,$C4,$70,$EB
	.dc.b	$C7,$2F,$F8,$FF,$D0,$FF,$58,$F3,$7D,$B8,$C7,$D8,$3F,$5F,$C7,$DA
	.dc.b	$AD,$C3,$CA,$31,$D3,$EA,$FD,$C9,$F0,$EF,$DB,$1C,$3F,$47,$F5,$7F
	.dc.b	$CB,$56,$A7,$3A,$FF,$BD,$2B,$F2,$D3,$FA,$0E,$CD,$E1,$CE,$5F,$B6
	.dc.b	$77,$8C,$78,$7F,$2B,$F4,$1F,$F4,$A7,$D2,$FF,$36,$FC,$39,$FB,$F7
	.dc.b	$E9,$67,$3F,$1C,$9F,$9B,$97,$E9,$67,$C0,$DF,$CF,$18,$7F,$40,$FD
	.dc.b	$13,$B4,$63,$07,$93,$F6,$BF,$54,$7E,$DF,$D7,$EC,$3F,$DB,$DF,$2D
	.dc.b	$FB,$59,$CE,$BF,$E0,$3F,$D3,$9F,$31,$9E,$4F,$DF,$EF,$FB,$7F,$C3
	.dc.b	$BC,$71,$FA,$8C,$38,$E4,$FC,$1B,$C1,$A7,$7F,$C8,$3F,$17,$EB,$BC
	.dc.b	$C7,$EA,$EE,$07,$D4,$7E,$27,$6F,$E0,$97,$BA,$D4,$FF,$29,$9F,$E1
	.dc.b	$1F,$B5,$78,$B6,$FD,$2D,$D7,$F4,$4F,$1C,$3F,$81,$57,$4C,$7C,$31
	.dc.b	$FC,$0A,$BF,$8B,$72,$BF,$5C,$7D,$B3,$F4,$97,$F0,$07,$F0,$8C,$BF
	.dc.b	$8C,$5E,$EB,$52,$FD,$EB,$F9,$06,$7E,$3F,$57,$F9,$C5,$DF,$57,$FD
	.dc.b	$AD,$9F,$6C,$FF,$AF,$E9,$E2,$8C,$77,$FE,$CB,$F2,$FF,$E4,$9F,$34
	.dc.b	$FC,$04,$BF,$76,$FE,$5D,$3A,$9C,$FE,$73,$FC,$C7,$F3,$AF,$04,$CE
	.dc.b	$5E,$35,$FE,$D4,$CF,$8C,$63,$ED,$FF,$82,$7E,$30,$FD,$CF,$F2,$9F
	.dc.b	$AE,$3F,$65,$7E,$EA,$FC,$07,$FA,$15,$EF,$C8,$5E,$E9,$F8,$A6,$BD
	.dc.b	$3F,$89,$7D,$75,$FC,$69,$FD,$79,$E5,$7D,$F4,$F1,$E9,$E2,$D3,$D1
	.dc.b	$DD,$FD,$53,$F6,$BF,$F5,$CF,$E6,$FF,$91,$F8,$57,$CA,$7F,$BF,$6B
	.dc.b	$5F,$9A,$7F,$B1,$3C,$AF,$E3,$72,$FC,$27,$FA,$D6,$3F,$EC,$2D,$FC
	.dc.b	$67,$F6,$95,$F4,$FF,$6D,$EF,$7F,$C9,$23,$F9,$54,$63,$27,$D3,$FF
	.dc.b	$34,$F9,$57,$CF,$BC,$07,$ED,$2C,$1F,$00,$FD,$45,$FA,$BF,$F7,$5C
	.dc.b	$78,$E1,$C2,$F7,$4F,$D7,$F5,$97,$5D,$FD,$23,$0F,$0F,$C1,$7E,$95
	.dc.b	$E7,$A7,$87,$4F,$C8,$F2,$7F,$80,$27,$39,$F9,$1F,$57,$E6,$AF,$D9
	.dc.b	$9F,$9C,$3F,$74,$FC,$37,$C1,$FF,$2C,$C7,$95,$EE,$B5,$D9,$FC,$83
	.dc.b	$F0,$1F,$64,$BF,$36,$F2,$3D,$78,$3D,$3A,$23,$96,$7F,$93,$FD,$13
	.dc.b	$C7,$AF,$8A,$3F,$2D,$FB,$E7,$CF,$FC,$6F,$F8,$6F,$E1,$DF,$BB,$FD
	.dc.b	$65,$E2,$3F,$6C,$F9,$95,$AF,$51,$5A,$F9,$4C,$FC,$0F,$C7,$A7,$E7
	.dc.b	$96,$19,$75,$CB,$C1,$18,$C7,$9F,$C1,$FF,$74,$F4,$9F,$82,$7F,$A1
	.dc.b	$FF,$4A,$7D,$73,$F1,$57,$E1,$0F,$DF,$7F,$C6,$98,$70,$FC,$31,$87
	.dc.b	$0B,$DF,$66,$B2,$FE,$07,$D7,$F8,$0C,$38,$36,$3C,$19,$F7,$F3,$7C
	.dc.b	$91,$8F,$B9,$FF,$1F,$57,$FB,$C8,$7E,$64,$FD,$C5,$F7,$8F,$D2,$1F
	.dc.b	$B6,$7E,$0B,$F6,$F7,$E1,$DF,$E4,$BC,$38,$5E,$EB,$5F,$51,$9C,$79
	.dc.b	$E5,$F9,$6B,$ED,$FA,$B7,$F3,$97,$8E,$7D,$BC,$97,$EE,$5F,$E1,$E9
	.dc.b	$CE,$5F,$92,$F9,$2F,$F1,$0F,$EC,$3E,$8F,$CF,$1F,$90,$FF,$AD,$7F
	.dc.b	$1C,$7C,$E3,$F3,$FF,$F3,$05,$EE,$B5,$2F,$CD,$33,$9B,$1F,$E9,$97
	.dc.b	$F1,$3A,$AC,$E7,$0E,$DE,$E7,$8A,$31,$DE,$D3,$F5,$BC,$7F,$37,$F8
	.dc.b	$05,$FA,$87,$FA,$5B,$99,$FC,$00,$FC,$43,$EF,$FE,$C9,$FE,$75,$BD
	.dc.b	$FE,$23,$F4,$EB,$5E,$2B,$F3,$5D,$64,$FC,$F3,$FB,$16,$1C,$34,$FE
	.dc.b	$06,$F2,$D3,$BE,$9F,$C1,$FD,$CF,$F3,$FC,$E7,$3F,$17,$DF,$FF,$93
	.dc.b	$3F,$63,$7E,$71,$F9,$6F,$8C,$7D,$8B,$F3,$47,$C8,$FF,$3E,$CF,$FD
	.dc.b	$13,$87,$0F,$47,$F9,$75,$EE,$9F,$8F,$7F,$A4,$6B,$5C,$1F,$C6,$1E
	.dc.b	$FF,$8A,$7F,$7B,$79,$67,$E0,$CF,$C3,$5C,$74,$F2,$65,$E3,$F8,$0F
	.dc.b	$F3,$1F,$E1,$7F,$CA,$E5,$FB,$F3,$F2,$97,$E6,$2F,$E0,$07,$E8,$5F
	.dc.b	$EA,$3C,$38,$5E,$EB,$5E,$7E,$7E,$4F,$E1,$EB,$FC,$BA,$FF,$9C,$7E
	.dc.b	$8D,$FE,$11,$78,$E7,$DB,$EF,$FC,$77,$F2,$B2,$E9,$F8,$0F,$F9,$E3
	.dc.b	$67,$E6,$BF,$CD,$FF,$2F,$7E,$EA,$FC,$C9,$F8,$3B,$DC,$FF,$5C,$61
	.dc.b	$C3,$E4,$98,$70,$BD,$F0,$6B,$2F,$F5,$EF,$0F,$F4,$4C,$38,$36,$3C
	.dc.b	$19,$F7,$D3,$C9,$18,$C1,$BF,$FA,$3B,$A3,$EF,$FB,$4F,$DB,$BF,$C0
	.dc.b	$7F,$7B,$FE,$3D,$7F,$D9,$93,$A7,$3F,$6C,$FC,$DB,$3E,$17,$F0,$BD
	.dc.b	$BF,$AB,$DF,$E2,$BD,$CF,$F1,$D3,$C7,$EA,$D9,$7F,$97,$2B,$8E,$7E
	.dc.b	$19,$CF,$BD,$E5,$C7,$F8,$B6,$9E,$0F,$E0,$0F,$F6,$F6,$EF,$D5,$BF
	.dc.b	$D2,$3F,$A8,$3F,$67,$7D,$D3,$F8,$03,$FC,$F9,$B7,$F2,$3C,$38,$7C
	.dc.b	$33,$0E,$17,$BF,$C5,$BA,$E5,$FC,$8B,$0F,$EE,$38,$70,$6C,$78,$32
	.dc.b	$EF,$F5,$3F,$EB,$8A,$EB,$8F,$4F,$C5,$BF,$D7,$7F,$3F,$FE,$51,$F5
	.dc.b	$3F,$D3,$BE,$13,$F0,$FF,$E6,$9F,$55,$FE,$F9,$AD,$77,$FE,$16,$7F
	.dc.b	$9C,$7D,$5E,$7E,$F7,$F8,$6F,$07,$F8,$12,$FF,$CB,$9F,$B5,$7F,$99
	.dc.b	$DE,$39,$75,$F9,$CF,$0D,$FF,$14,$CB,$A7,$EF,$0C,$5F,$28,$FC,$93
	.dc.b	$FA,$47,$D6,$3F,$4F,$7E,$3D,$FE,$AE,$F5,$9F,$98,$FF,$BF,$3B,$78
	.dc.b	$E1,$C2,$F7,$F0,$1F,$B2,$AF,$F8,$03,$F3,$3A,$D3,$C1,$9F,$5C,$FF
	.dc.b	$D6,$5A,$78,$67,$38,$BF,$23,$CE,$7D,$6F,$C5,$97,$F8,$0A,$FA,$1F
	.dc.b	$E3,$0B,$F0,$EB,$E6,$DB,$DD,$97,$5A,$7E,$4C,$FD,$F9,$F2,$BF,$DE
	.dc.b	$7E,$DD,$F5,$AF,$C1,$1F,$AB,$76,$C3,$85,$EE,$B5,$FB,$46,$71,$E7
	.dc.b	$97,$F7,$13,$C3,$C3,$FD,$CF,$0F,$07,$AF,$EC,$9F,$A5,$1D,$3E,$9B
	.dc.b	$FC,$AB,$8F,$F8,$A7,$6F,$44,$E7,$6F,$2F,$E8,$1D,$30,$FB,$5F,$B0
	.dc.b	$7F,$BE,$BF,$41,$7E,$FC,$F3,$7F,$AC,$7E,$8C,$F9,$9F,$8F,$7B,$AD
	.dc.b	$7C,$DA,$BF,$A8,$6B,$F1,$E9,$F9,$33,$F4,$0F,$46,$89,$F0,$3E,$F3
	.dc.b	$AC,$B4,$F1,$E9,$FD,$A7,$E5,$7D,$15,$C6,$B5,$A7,$8B,$E8,$33,$8F
	.dc.b	$97,$0F,$4F,$EF,$DF,$7B,$F0,$07,$E6,$FF,$E0,$AE,$BF,$F4,$2F,$EE
	.dc.b	$4E,$F7,$D8,$EE,$B5,$F8,$2A,$B5,$E4,$E7,$F4,$FF,$5E,$5D,$3E,$06
	.dc.b	$E1,$C2,$FA,$71,$73,$9F,$C3,$E3,$CF,$CF,$F5,$FC,$23,$F4,$27,$E4
	.dc.b	$CF,$DA,$AF,$DA,$3F,$80,$5F,$8A,$FE,$63,$0E,$17,$B3,$FC,$9B,$2F
	.dc.b	$C5,$75,$8F,$DB,$D7,$E7,$77,$FE,$6F,$0E,$31,$8F,$27,$FC,$81,$1D
	.dc.b	$34,$F0,$FE,$13,$E6,$C3,$CB,$1E,$4B,$FA,$7E,$17,$F8,$5F,$F7,$17
	.dc.b	$D7,$FC,$BF,$B2,$FE,$FB,$F0,$9F,$AE,$AD,$4E,$7E,$9D,$F9,$BE,$7E
	.dc.b	$AF,$FD,$E1,$97,$8C,$FE,$1D,$7D,$5C,$1F,$8F,$86,$5E,$0E,$F6,$7F
	.dc.b	$EB,$0A,$E9,$39,$8F,$F5,$CC,$7F,$00,$FC,$75,$F9,$5F,$FA,$DB,$DF
	.dc.b	$FF,$6F,$7E,$F3,$FE,$B1,$F2,$35,$39,$FC,$35,$3F,$EB,$5F,$B4,$CF
	.dc.b	$07,$E7,$31,$E7,$F1,$D7,$0E,$17,$FE,$FC,$E9,$FE,$15,$5A,$CF,$BF
	.dc.b	$D0,$FE,$73,$F2,$07,$E6,$7F,$E4,$4F,$9A,$79,$3F,$E9,$0F,$15,$F9
	.dc.b	$2F,$F1,$8B,$DF,$DB,$7F,$9C,$EB,$5F,$A5,$E7,$1F,$F2,$6F,$2F,$C8
	.dc.b	$BE,$BE,$57,$EC,$9E,$39,$F6,$FC,$2B,$E2,$F0,$7F,$1F,$9C,$E5,$FE
	.dc.b	$A0,$E3,$FD,$8F,$B0,$FC,$85,$FA,$B7,$FB,$73,$EB,$7F,$D3,$BF,$43
	.dc.b	$FE,$C4,$CB,$F9,$45,$EE,$B5,$E7,$2B,$FC,$63,$5A,$C3,$F4,$FB,$FF
	.dc.b	$21,$8C,$5F,$F0,$8C,$39,$7E,$2D,$FE,$39,$C7,$F9,$D5,$6A,$31,$F8
	.dc.b	$57,$F0,$6F,$7B,$CD,$F9,$D3,$F7,$07,$0B,$F5,$0F,$E8,$DE,$FF,$F4
	.dc.b	$2B,$DD,$6A,$7F,$A5,$7E,$91,$9E,$06,$3F,$C6,$AF,$E1,$7E,$DB,$67
	.dc.b	$38,$71,$FC,$BF,$C5,$18,$CF,$C1,$F9,$8F,$F1,$AF,$CA,$BF,$B3,$7E
	.dc.b	$6F,$FE,$EA,$FC,$CD,$F9,$73,$F1,$6F,$F5,$7F,$27,$F5,$1B,$DF,$E2
	.dc.b	$4B,$DD,$3E,$FB,$FA,$DD,$7F,$71,$E1,$FE,$5E,$F6,$EA,$FD,$4B,$CA
	.dc.b	$FB,$FA,$3E,$3F,$9A,$6B,$E3,$D5,$FA,$DF,$9F,$F2,$FC,$FE,$94,$FC
	.dc.b	$CD,$FB,$13,$F6,$B7,$E1,$BE,$2B,$DD,$6A,$5F,$A9,$67,$87,$CB,$5E
	.dc.b	$0E,$1F,$2B,$C3,$1F,$0C,$73,$D5,$9F,$86,$5F,$9F,$FC,$BF,$D6,$BD
	.dc.b	$3F,$4F,$F1,$7F,$24,$FF,$3E,$7E,$0B,$F1,$27,$3C,$BF,$EF,$EF,$C2
	.dc.b	$DE,$36,$77,$F8,$F5,$E3,$DB,$F2,$5F,$17,$FE,$E3,$E9,$4F,$2C,$FF
	.dc.b	$CC,$75,$D7,$83,$FC,$21,$9A,$5F,$5B,$F7,$FF,$61,$F2,$FD,$E7,$C8
	.dc.b	$F8,$27,$E9,$5F,$62,$73,$C8,$97,$EB,$59,$E0,$FB,$75,$E5,$D7,$FD
	.dc.b	$ED,$E1,$FE,$16,$38,$FE,$7F,$DF,$7F,$F1,$16,$9F,$5B,$F1,$1F,$C8
	.dc.b	$7E,$4D,$F1,$BF,$27,$F3,$2F,$D6,$3F,$12,$E7,$88,$FF,$8E,$33,$FC
	.dc.b	$93,$FA,$2D,$3F,$16,$F8,$DE,$B9,$7F,$8C,$70,$F8,$5F,$85,$E3,$AD
	.dc.b	$1F,$17,$DF,$FD,$17,$1F,$D8,$FD,$23,$FD,$9B,$C9,$FC,$33,$4F,$D0
	.dc.b	$4E,$7F,$0E,$F8,$9F,$1F,$F9,$E8,$F1,$6B,$F5,$8F,$F2,$A7,$5F,$E5
	.dc.b	$36,$FD,$0C,$F1,$E4,$EF,$1F,$E4,$5C,$FF,$DB,$7C,$7F,$B3,$F8,$BF
	.dc.b	$17,$E8,$DF,$03,$ED,$FF,$57,$18,$E4,$F8,$DB,$FE,$CE,$3C,$3A,$FD
	.dc.b	$2E,$FF,$C3,$47,$C3,$8E,$39,$F7,$C7,$F0,$F7,$C6,$31,$C5,$FC,$0C
	.dc.b	$F8,$5F,$B1,$7E,$07,$DF,$FD,$DC,$63,$93,$FC,$2D,$7F,$C0,$E3,$C3
	.dc.b	$8F,$1C,$3F,$7F,$1C,$B8,$31,$CC,$E1,$67,$FE,$E8,$C7,$97,$E1,$5F
	.dc.b	$0F,$89,$F2,$3F,$7F,$87,$F4,$19,$FF,$AF,$72,$E5,$5A,$EB,$F2,$FE
	.dc.b	$1D,$E0,$C6,$56,$BE,$4F,$D9,$1D,$7E,$B7,$DD,$F4,$BD,$FF,$2F,$FB
	.dc.b	$DA,$76,$8C,$76,$7C,$0F,$D4,$FD,$71,$32,$F0,$FC,$9F,$27,$C7,$E7
	.dc.b	$7F,$2B,$F7,$9F,$B9,$7A,$FE,$73,$D5,$D5,$F8,$49,$CD,$FA,$A7,$ED
	.dc.b	$E9,$FB,$D9,$7E,$3D,$FC,$A1,$E7,$7F,$CA,$1E,$3D,$2E,$5F,$E2,$33
	.dc.b	$FC,$C2,$73,$0D,$3C,$9F,$A0,$3F,$4C,$7E,$CA,$FC,$51,$EA,$BF,$48
	.dc.b	$E5,$7B,$AD,$4B,$F3,$3C,$E9,$E3,$F8,$EF,$35,$EF,$1F,$C6,$F0,$7A
	.dc.b	$E3,$1E,$AF,$C5,$FD,$3F,$E2,$78,$5E,$77,$E3,$7D,$97,$A7,$E5,$3E
	.dc.b	$04,$4E,$7F,$9B,$A6,$DF,$A2,$FB,$FE,$A7,$E7,$7E,$BB,$67,$E6,$7E
	.dc.b	$79,$F0,$FF,$7A,$F9,$55,$27,$9E,$7F,$D6,$FA,$3F,$CC,$11,$D2,$31
	.dc.b	$F7,$5F,$67,$F4,$FF,$98,$FD,$BF,$A4,$F9,$DE,$87,$D5,$89,$7D,$38
	.dc.b	$FE,$D4,$7E,$6F,$AF,$F0,$9F,$2B,$CF,$2F,$03,$CD,$F1,$7D,$77,$DF
	.dc.b	$89,$87,$3E,$1F,$C4,$47,$37,$EF,$7B,$FC,$F2,$FC,$43,$CD,$B1,$F8
	.dc.b	$78,$FD,$67,$89,$F9,$57,$96,$3F,$05,$FE,$47,$E0,$FE,$2F,$C3,$70
	.dc.b	$76,$F6,$FF,$2D,$C3,$6B,$F2,$18,$F3,$9F,$91,$9F,$E6,$7D,$3F,$F4
	.dc.b	$55,$73,$C3,$F2,$91,$8F,$13,$FC,$DF,$8E,$35,$FC,$E4,$E7,$2F,$D5
	.dc.b	$3B,$E1,$BA,$7E,$9F,$1E,$6F,$93,$6F,$B5,$5F,$E7,$AC,$FC,$B8,$E9
	.dc.b	$FD,$36,$1D,$3F,$6C,$E5,$7C,$1E,$5A,$FA,$5F,$A1,$6F,$6F,$FC,$9D
	.dc.b	$8A,$7D,$F3,$FB,$B8,$7E,$81,$D2,$FE,$13,$D2,$74,$F4,$BE,$6F,$28
	.dc.b	$ED,$CF,$FE,$20,$FD,$EB,$94,$75,$79,$3A,$C1,$1C,$E7,$86,$9E,$EA
	.dc.b	$F9,$FF,$13,$C9,$E0,$F2,$C7,$D6,$F1,$C3,$FC,$37,$5C,$74,$F9,$7F
	.dc.b	$95,$71,$BF,$82,$F1,$C7,$FC,$6A,$F1,$3D,$51,$F1,$B9,$DF,$87,$7E
	.dc.b	$06,$FE,$AC,$63,$97,$F0,$31,$89,$FA,$ED,$8F,$E3,$1A,$7E,$22,$9F
	.dc.b	$9F,$73,$E0,$FC,$74,$62,$B9,$C6,$27,$9B,$63,$9D,$F8,$BE,$FB,$96
	.dc.b	$1F,$67,$0F,$F7,$7D,$72,$8C,$4F,$26,$C7,$2B,$F0,$FD,$87,$29,$FA
	.dc.b	$96,$8F,$D5,$74,$7A,$5E,$53,$A7,$E7,$57,$AB,$F3,$8E,$5A,$FC,$97
	.dc.b	$8E,$7F,$21,$E3,$19,$CB,$C5,$87,$A7,$E9,$38,$F0,$7F,$94,$1E,$38
	.dc.b	$FC,$67,$8C,$66,$B8,$D7,$A3,$EE,$F8,$F1,$B6,$FF,$80,$35,$F0,$EB
	.dc.b	$4F,$F4,$ED,$3E,$DD,$FA,$A7,$E5,$76,$72,$BF,$F1,$6D,$F1,$D3,$F0
	.dc.b	$93,$E2,$C7,$FC,$69,$3E,$29,$FC,$47,$23,$EF,$7F,$DE,$7D,$9F,$81
	.dc.b	$8C,$76,$38,$F1,$5E,$3E,$1F,$C1,$D7,$1E,$CF,$E1,$3A,$FF,$83,$FE
	.dc.b	$7F,$9F,$0F,$E4,$A9,$1E,$5E,$2F,$F2,$03,$CF,$87,$F0,$93,$AC,$39
	.dc.b	$AF,$CD,$79,$E5,$E0,$79,$70,$FF,$8E,$DE,$57,$F8,$4A,$E1,$7C,$BF
	.dc.b	$0A,$E3,$3D,$9E,$3A,$7F,$89,$5E,$3C,$3F,$11,$F8,$DA,$7F,$95,$7E
	.dc.b	$13,$8D,$F5,$78,$E2,$78,$CE,$F1,$F4,$FC,$2F,$0C,$7E,$7B,$C0,$DE
	.dc.b	$9F,$E5,$37,$BB,$FE,$55,$FC,$B7,$85,$FF,$99,$5E,$E7,$0D,$3F,$C7
	.dc.b	$8F,$07,$FC,$7D,$F0,$DC,$2B,$FC,$BC,$F9,$8E,$19,$FF,$90,$B2,$F9
	.dc.b	$B0,$FC,$EF,$FA,$8F,$17,$D7,$8C,$72,$7E,$0D,$7A,$3C,$3F,$50,$E3
	.dc.b	$FB,$F7,$0E,$5F,$C1,$D2,$3F,$11,$87,$3E,$3E,$FF,$7F,$F7,$3A,$DE
	.dc.b	$7F,$19,$FC,$0C,$63,$4F,$C9,$2F,$82,$E1,$D7,$C7,$F7,$EE,$17,$F7
	.dc.b	$E9,$1E,$0D,$7F,$2B,$0F,$85,$F0,$7C,$3F,$87,$EB,$DF,$8D,$F6,$5F
	.dc.b	$BE,$78,$39,$BC,$DF,$5D,$DB,$BB,$F4,$F9,$23,$B6,$BE,$8F,$BE,$D7
	.dc.b	$77,$FA,$6E,$91,$A9,$F3,$7A,$7F,$07,$E2,$F7,$DD,$FE,$63,$E5,$F7
	.dc.b	$FE,$C9,$8F,$03,$FC,$DE,$F1,$D7,$C3,$E3,$BF,$65,$F8,$AB,$F7,$9B
	.dc.b	$F9,$0E,$3F,$23,$CF,$E8,$9A,$FD,$73,$7F,$A8,$70,$FD,$97,$8F,$E0
	.dc.b	$5C,$BF,$04,$E7,$FB,$27,$4F,$C6,$BA,$FC,$47,$6F,$89,$EF,$EB,$78
	.dc.b	$39,$7C,$3F,$A9,$78,$BF,$51,$F1,$FB,$BE,$4F,$C0,$FC,$BF,$A6,$F9
	.dc.b	$BE,$AB,$CF,$EF,$FA,$3F,$19,$F4,$FE,$2B,$EA,$FC,$53,$EF,$FE,$26
	.dc.b	$FD,$D2,$FE,$4F,$8E,$FC,$F7,$6B,$8B,$7C,$5F,$86,$F1,$7E,$47,$BB
	.dc.b	$FE,$26,$8F,$C4,$FD,$86,$3C,$AF,$F7,$A4,$7E,$33,$EF,$7F,$25,$FB
	.dc.b	$2E,$4F,$9D,$E6,$FD,$98,$7C,$C6,$FF,$21,$FC,$9F,$C9,$F3,$37,$E6
	.dc.b	$FF,$AA,$7F,$02,$FC,$E7,$9D,$DC,$C9,$FE,$A8,$FC,$1F,$F4,$1F,$17
	.dc.b	$FA,$1F,$DE,$7C,$86,$B8,$7F,$D1,$9E,$CF,$E9,$7E,$33,$F4,$DF,$11
	.dc.b	$E9,$35,$C7,$FB,$2F,$5B,$F4,$BF,$0F,$FC,$9B,$F2,$5F,$E0,$1A,$FF
	.dc.b	$AF,$DF,$E2,$A7,$F0,$03,$F8,$36,$4F,$DC,$B6,$DF,$5B,$C5,$1F,$C1
	.dc.b	$8D,$FE,$6F,$A3,$F8,$F9,$FC,$28,$E7,$FA,$0F,$23,$F9,$19,$FC,$38
	.dc.b	$EF,$F9,$37,$63,$F9,$39,$FC,$48,$F1,$E5,$E2,$3F,$95,$9F,$C5,$8F
	.dc.b	$3F,$E6,$7E,$63,$F9,$79,$FC,$68,$FE,$41,$F4,$9F,$D5,$BC,$0F,$F0
	.dc.b	$C3,$FD,$63,$BA,$FC,$47,$F2,$78,$FE,$AD,$C1,$AE,$1F,$8F,$1F,$DE
	.dc.b	$23,$FC,$28,$FF,$67,$E1,$3F,$DB,$E3,$FC,$1F,$F7,$BF,$DB,$CE,$0E
	.dc.b	$2B,$F9,$A4,$6F,$DE,$FF,$1A,$7E,$7B,$FD,$D4,$DB,$89,$FE,$75,$1B
	.dc.b	$F0,$5D,$BF,$E0,$3F,$AB,$FE,$F8,$70,$8C,$E7,$FE,$C8,$BF,$E8,$53
	.dc.b	$C6,$BF,$C1,$5F,$73,$FE,$03,$38,$38,$9F,$E9,$11,$BA,$7E,$F3,$FE
	.dc.b	$09,$36,$E2,$7F,$A7,$47,$5F,$9E,$7E,$D5,$FE,$18,$6C,$FF,$72,$6F
	.dc.b	$93,$FE,$8A,$73,$F1,$35,$E0,$79,$0F,$F2,$A4,$66,$7F,$B0,$C7,$F9
	.dc.b	$87,$4F,$86,$FF,$98,$73,$FF,$2E,$F6,$7F,$86,$E3,$FC,$C3,$3F,$23
	.dc.b	$C3,$E1,$47,$81,$FE,$6F,$36,$BF,$7D,$D9,$D4,$FF,$3A,$1A,$FD,$0F
	.dc.b	$99,$E1,$3F,$CF,$07,$2F,$CC,$FB,$9E,$53,$FC,$F8,$76,$F1,$7C,$67
	.dc.b	$A4,$FF,$40,$1E,$2F,$D9,$BC,$E7,$B4,$FF,$42,$1E,$6F,$9F,$F5,$9F
	.dc.b	$04,$FF,$44,$1F,$03,$F2,$9F,$F6,$97,$48,$FF,$4D,$BF,$ED,$37,$1D
	.dc.b	$0D,$53,$B9,$83,$E6,$7F,$DE,$91,$FE,$D8,$8C,$7C,$AE,$EF,$94,$FF
	.dc.b	$B6,$A3,$16,$F9,$4F,$1D,$FF,$AB,$1F,$13,$4E,$E7,$43,$CB,$1C,$7A
	.dc.b	$3F,$D5,$93,$D6,$73,$FA,$BF,$4C,$7D,$4F,$47,$D9,$F8,$07,$49,$E8
	.dc.b	$7F,$AF,$27,$19,$B2,$F3,$C7,$B6,$31,$3D,$B1,$FC,$5F,$CC,$7F,$9C
	.dc.b	$F9,$BF,$DE,$9F,$0B,$D3,$3F,$95,$3E,$1D,$5F,$93,$76,$D3,$E1,$CD
	.dc.b	$CB,$DA,$F7,$8C,$E5,$F9,$8F,$96,$ED,$D4,$F3,$5F,$8A,$F7,$CF,$FC
	.dc.b	$1D,$6D,$97,$BF,$CD,$75,$F3,$5D,$F5,$FF,$6D,$E7,$77,$FC,$03,$E1
	.dc.b	$8F,$9E,$FC,$D7,$FD,$C3,$E0,$3E,$15,$6F,$92,$7E,$78,$FC,$23,$C9
	.dc.b	$7B,$AD,$4B,$F8,$03,$38,$FD,$EE,$7F,$9A,$FF,$BB,$B4,$F2,$70,$B1
	.dc.b	$E5,$9F,$B2,$BA,$E9,$E2,$8F,$F3,$64,$7C,$3F,$80,$F8,$FF,$0C,$FB
	.dc.b	$5C,$72,$7C,$5E,$34,$FF,$91,$F5,$F0,$EB,$FA,$1F,$DC,$FC,$38,$CC
	.dc.b	$3F,$DF,$4F,$C3,$9F,$D3,$78,$9F,$53,$F4,$E7,$E1,$7F,$4E,$31,$F8
	.dc.b	$47,$D5,$FD,$07,$D7,$13,$1F,$17,$9B,$FB,$3D,$3C,$3C,$8A,$FB,$FF
	.dc.b	$0F,$9E,$FF,$D4,$FE,$11,$F8,$2F,$C6,$7D,$98,$70,$8C,$7D,$3F,$E0
	.dc.b	$FF,$0A,$F9,$31,$34,$F8,$1F,$11,$CF,$A3,$F5,$73,$F8,$CF,$88,$C4
	.dc.b	$39,$1F,$BA,$7F,$95,$7F,$0E,$F6,$E9,$F1,$A3,$1C,$9F,$9F,$5F,$79
	.dc.b	$F8,$B7,$FA,$AE,$1F,$77,$83,$F8,$24,$7F,$97,$FF,$10,$EF,$97,$6F
	.dc.b	$5B,$F3,$BF,$87,$63,$F5,$BF,$EB,$3E,$53,$F7,$97,$DB,$7C,$0C,$FF
	.dc.b	$87,$56,$A7,$3C,$DF,$B1,$8F,$1E,$BE,$13,$F8,$EB,$E4,$7F,$8D,$F0
	.dc.b	$7E,$02,$39,$7D,$BF,$87,$4F,$DE,$61,$F8,$FA,$FC,$2F,$EA,$1E,$4F
	.dc.b	$CE,$9C,$9F,$13,$4F,$E4,$13,$98,$72,$7F,$03,$8F,$1E,$9E,$0A,$DE
	.dc.b	$5F,$B1,$9E,$9C,$5F,$11,$ED,$C3,$F1,$BF,$74,$FE,$F7,$F0,$FF,$AD
	.dc.b	$E2,$F8,$59,$7F,$38,$9C,$F9,$FF,$1B,$F0,$6F,$E2,$51,$E0,$9F,$E7
	.dc.b	$5C,$1F,$9E,$8F,$85,$18,$D7,$F9,$C4,$72,$E2,$F0,$6C,$EC,$FF,$0E
	.dc.b	$E5,$F0,$EF,$FA,$27,$9B,$FC,$9E,$B5,$FB,$DB,$E6,$BE,$2F,$E1,$DF
	.dc.b	$CB,$AB,$58,$AB,$5C,$AF,$C6,$75,$97,$9F,$6F,$C9,$E1,$D5,$FB,$47
	.dc.b	$F0,$38,$FF,$0E,$DF,$7E,$5D,$C6,$33,$72,$72,$FC,$9F,$FC,$0B,$F9
	.dc.b	$0F,$F1,$5F,$C0,$1F,$BC,$F9,$67,$3C,$49,$7E,$F9,$9F,$0F,$F9,$46
	.dc.b	$EB,$F9,$3B,$F5,$E4,$7C,$5B,$4F,$E4,$5E,$A6,$63,$1A,$7F,$16,$E9
	.dc.b	$FE,$3F,$93,$F3,$DF,$F1,$DF,$EE,$FF,$0F,$F3,$7F,$F6,$41,$F2,$63
	.dc.b	$3C,$E8,$FF,$66,$72,$67,$83,$CB,$5E,$4D,$F8,$6E,$C7,$A6,$5F,$32
	.dc.b	$B9,$E3,$FE,$DF,$BF,$AD,$F9,$47,$F9,$33,$F1,$5F,$57,$E8,$9F,$D6
	.dc.b	$3F,$08,$CE,$69,$CF,$C3,$BF,$0C,$E3,$B7,$C6,$C7,$C5,$C3,$EE,$7A
	.dc.b	$72,$B3,$F0,$4E,$72,$E9,$39,$F8,$CF,$C0,$FA,$FF,$E1,$2F,$CB,$DF
	.dc.b	$80,$FF,$2E,$F9,$67,$ED,$DF,$8D,$AD,$4E,$7F,$51,$C6,$5F,$62,$7F
	.dc.b	$C3,$7C,$7F,$8A,$E6,$FF,$0A,$E5,$D3,$DF,$7A,$7E,$0F,$92,$4E,$7E
	.dc.b	$7B,$F9,$2F,$E5,$9F,$E2,$5F,$CF,$1F,$98,$BF,$49,$7E,$EE,$FD,$01
	.dc.b	$FB,$A7,$F8,$3B,$0E,$17,$BA,$D7,$D9,$67,$85,$97,$F7,$5B,$FF,$0C
	.dc.b	$F0,$FE,$81,$E3,$9F,$6C,$FC,$77,$C3,$2E,$9F,$43,$FC,$EB,$F5,$BF
	.dc.b	$F1,$D7,$E8,$EF,$CB,$5F,$B9,$3F,$7A,$7E,$DE,$FC,$0F,$FC,$47,$87
	.dc.b	$0B,$DD,$6B,$F7,$7C,$E8,$F8,$CF,$F0,$55,$F6,$E1,$FD,$AB,$C7,$3E
	.dc.b	$D9,$F8,$ED,$A7,$8E,$7C,$5F,$67,$FD,$43,$E5,$FF,$A7,$FE,$96,$E6
	.dc.b	$7D,$2B,$F3,$C7,$E0,$0F,$AC,$FF,$1C,$E1,$C2,$F7,$5A,$FD,$13,$3F
	.dc.b	$24,$FD,$77,$FC,$41,$7D,$B8,$BF,$85,$BC,$73,$ED,$F3,$1E,$3B,$6B
	.dc.b	$E3,$9F,$17,$27,$F6,$6F,$65,$F5,$AF,$D2,$5F,$97,$E3,$E4,$5F,$AB
	.dc.b	$3E,$67,$FC,$9F,$87,$0B,$DE,$0B,$DF,$3E,$A3,$A7,$07,$F2,$0B,$EF
	.dc.b	$FA,$3F,$F8,$E5,$E5,$9F,$7E,$BF,$25,$F8,$E3,$C5,$D3,$FD,$F3,$F7
	.dc.b	$47,$E2,$CE,$77,$E0,$4F,$BB,$7D,$0B,$F0,$57,$E3,$3F,$E6,$6C,$78
	.dc.b	$E1,$C2,$F7,$4F,$C6,$75,$97,$5E,$1F,$E6,$F8,$78,$3C,$0F,$F2,$83
	.dc.b	$CB,$4F,$06,$9D,$EB,$8E,$7D,$61,$F8,$3F,$F8,$53,$F4,$07,$23,$F7
	.dc.b	$77,$EC,$FF,$E1,$6F,$B5,$7E,$58,$FC,$AB,$FC,$FB,$87,$0B,$DF,$84
	.dc.b	$BD,$FE,$0F,$AC,$9F,$82,$FF,$99,$70,$E1,$BF,$F5,$07,$96,$9D,$FC
	.dc.b	$0F,$25,$F8,$E3,$C5,$E9,$7F,$8A,$BD,$9F,$F6,$67,$E1,$CF,$CD,$5F
	.dc.b	$9B,$7F,$BA,$BC,$77,$EE,$9F,$E9,$2B,$DD,$6B,$F0,$7F,$D7,$7E,$C3
	.dc.b	$9C,$99,$7F,$9E,$6F,$FD,$B1,$F9,$B5,$9C,$E1,$C7,$2F,$14,$63,$3F
	.dc.b	$F7,$CF,$9D,$FE,$3E,$DC,$BE,$A5,$FA,$63,$F5,$A7,$EB,$6F,$DF,$5F
	.dc.b	$B4,$7F,$AA,$31,$E3,$87,$0B,$DD,$3F,$7A,$D6,$5D,$7A,$3F,$BC,$E1
	.dc.b	$E0,$FC,$C7,$FD,$24,$F2,$D3,$C1,$A3,$AF,$CD,$87,$96,$73,$F6,$78
	.dc.b	$D7,$C1,$7E,$5F,$8B,$FF,$5D,$FE,$9C,$FD,$A5,$F9,$07,$ED,$7E,$C5
	.dc.b	$FA,$7A,$BF,$DB,$BF,$39,$FE,$90,$78,$56,$E7,$5F,$37,$D3,$0E,$7C
	.dc.b	$5F,$E1,$2B,$ED,$F9,$17,$FA,$AD,$E3,$EE,$FF,$91,$30,$75,$7F,$9B
	.dc.b	$27,$3A,$78,$BE,$3B,$FC,$DF,$E9,$3F,$66,$F9,$1F,$AA,$FF,$93,$3E
	.dc.b	$15,$E5,$3F,$80,$3F,$DA,$F7,$BE,$D5,$EE,$9F,$8B,$EB,$1F,$A5,$C3
	.dc.b	$FE,$28,$C3,$B7,$E8,$DF,$EC,$17,$96,$1D,$FB,$3C,$77,$E2,$DB,$F9
	.dc.b	$47,$E0,$F8,$D3,$BC,$7F,$2B,$D3,$F8,$37,$47,$FA,$E3,$F6,$1F,$C4
	.dc.b	$7B,$1F,$EA,$CF,$C4,$5F,$85,$3F,$29,$7E,$7B,$FE,$F0,$BD,$FC,$77
	.dc.b	$FB,$22,$B5,$F9,$56,$7C,$AF,$E7,$DC,$BF,$E4,$0B,$EA,$D8,$FE,$35
	.dc.b	$7C,$7F,$2B,$FF,$4C,$D7,$4C,$FC,$3E,$46,$2F,$71,$E4,$D1,$F8,$EB
	.dc.b	$77,$ED,$6D,$27,$CD,$BF,$5A,$FF,$70,$7B,$FF,$D5,$3E,$79,$FA,$BF
	.dc.b	$FA,$53,$1E,$38,$70,$BD,$FC,$AF,$F8,$1A,$BF,$DC,$B5,$AA,$FE,$E9
	.dc.b	$5F,$EE,$B7,$FD,$C5,$97,$85,$CF,$ED,$95,$FF,$3C,$C6,$73,$F2,$7E
	.dc.b	$33,$FC,$8F,$2F,$F0,$9F,$8B,$FE,$CB,$AD,$63,$E2,$D3,$F4,$98,$FB
	.dc.b	$E7,$38,$71,$9F,$5F,$E0,$39,$AF,$85,$F9,$0F,$F8,$1B,$CA,$FF,$73
	.dc.b	$FD,$2F,$FB,$7B,$C5,$FC,$A7,$07,$AE,$B5,$D4,$AD,$7C,$87,$3C,$55
	.dc.b	$FE,$10,$FD,$EF,$D9,$1E,$73,$BF,$CB,$FE,$D9,$F8,$66,$33,$F0,$FD
	.dc.b	$67,$AB,$3F,$05,$7E,$F3,$3F,$04,$FF,$8C,$6F,$C1,$8F,$93,$3F,$47
	.dc.b	$E7,$BE,$3F,$13,$F9,$DF,$C6,$7F,$43,$F8,$E7,$E4,$19,$F8,$CF,$E7
	.dc.b	$93,$9F,$C8,$31,$F1,$DE,$8D,$35,$DF,$78,$7E,$E6,$7D,$3E,$3F,$FC
	.dc.b	$01,$5A,$F8,$FF,$E8,$AD,$CB,$C1,$F2,$16,$BF,$A9,$E1,$CB,$4F,$0C
	.dc.b	$7F,$5C,$F8,$6F,$95,$FB,$17,$FA,$47,$EC,$DF,$88,$3F,$4A,$7E,$C0
	.dc.b	$FD,$17,$FD,$F3,$9F,$E3,$F0,$E1,$7B,$AD,$7D,$5F,$4F,$C2,$9F,$B7
	.dc.b	$7A,$FC,$6F,$C9,$61,$DF,$8B,$E9,$87,$9E,$9E,$0D,$3E,$D4,$E7,$06
	.dc.b	$BE,$6A,$F2,$F8,$5A,$C3,$8E,$9E,$7F,$6F,$C5,$7B,$8F,$57,$E3,$3F
	.dc.b	$B5,$FD,$DF,$F4,$9F,$6F,$FE,$2C,$FD,$6B,$FB,$4F,$C4,$FF,$8C,$3E
	.dc.b	$39,$FB,$A6,$F8,$3F,$75,$7B,$F9,$E5,$EF,$F7,$9D,$52,$BF,$C6,$FE
	.dc.b	$8F,$DB,$CF,$BF,$D7,$34,$C5,$F4,$8C,$65,$D3,$EA,$9F,$55,$FC,$27
	.dc.b	$B1,$8F,$2C,$7C,$F4,$FC,$CF,$53,$FB,$FF,$CB,$3F,$C6,$1F,$5F,$F6
	.dc.b	$BC,$BF,$D5,$7E,$E6,$FC,$83,$CF,$5A,$E7,$FF,$0F,$F9,$EA,$7E,$6B
	.dc.b	$AE,$DF,$F3,$FF,$1F,$EB,$B8,$3C,$6F,$F1,$DC,$7F,$C7,$DF,$A7,$FB
	.dc.b	$21,$1D,$74,$FD,$D7,$E3,$FE,$69,$E1,$87,$FA,$2B,$4F,$0F,$E2,$F9
	.dc.b	$9E,$33,$9B,$F4,$FE,$D3,$E3,$E0,$F6,$6F,$E7,$FD,$A7,$F9,$87,$ED
	.dc.b	$FF,$C7,$FD,$DF,$F5,$4F,$77,$F6,$3F,$AA,$7E,$61,$FE,$37,$FD,$77
	.dc.b	$7B,$7F,$35,$BD,$F8,$CB,$DF,$A0,$F8,$FD,$DF,$FA,$1F,$F0,$7F,$C0
	.dc.b	$BF,$0F,$4F,$04,$3F,$22,$FD,$D4,$63,$5F,$16,$AF,$C2,$FE,$E4,$E7
	.dc.b	$4F,$04,$FF,$B6,$BE,$73,$F9,$C7,$E5,$DF,$ED,$4F,$DA,$7F,$CB,$3B
	.dc.b	$7E,$D3,$EE,$53,$FD,$3E,$B7,$F8,$82,$F7,$D6,$B7,$E6,$1B,$E1,$FE
	.dc.b	$A9,$8F,$FB,$83,$F7,$0F,$E2,$19,$72,$C6,$D6,$B2,$E9,$F7,$5F,$E5
	.dc.b	$FC,$DF,$43,$F9,$5C,$38,$4E,$7E,$8B,$F3,$1F,$67,$FD,$47,$F5,$87
	.dc.b	$E9,$6F,$8D,$7E,$E7,$FE,$CA,$FC,$49,$F9,$AF,$FA,$DF,$CB,$FF,$22
	.dc.b	$F0,$7F,$69,$C7,$8E,$1C,$3C,$46,$1C,$3E,$F3,$FB,$86,$9D,$F4,$ED
	.dc.b	$8F,$E5,$E3,$1F,$97,$7F,$2D,$75,$8F,$3C,$3F,$DE,$F5,$C8,$F0,$5F
	.dc.b	$9F,$5F,$37,$D9,$FE,$CF,$ED,$7F,$B6,$7C,$6B,$F6,$0F,$F7,$7F,$E8
	.dc.b	$DF,$E5,$CF,$DC,$DF,$A2,$BF,$2D,$FF,$02,$61,$C2,$F7,$BA,$F7,$F9
	.dc.b	$FE,$B6,$FF,$04,$67,$D7,$E8,$FF,$A2,$63,$C7,$F1,$AB,$5A,$CB,$9E
	.dc.b	$3F,$C4,$63,$91,$FE,$9C,$8C,$FE,$C5,$FC,$2F,$6F,$F0,$27,$6B,$F2
	.dc.b	$CF,$F6,$67,$EA,$2F,$CB,$3F,$CB,$9F,$98,$BE,$9B,$FC,$39,$87,$0B
	.dc.b	$DD,$6B,$EB,$F3,$A7,$F8,$7F,$2E,$9D,$BF,$D6,$F0,$E1,$F8,$2D,$AD
	.dc.b	$63,$CB,$F4,$1F,$F0,$B7,$82,$BD,$ED,$E3,$FB,$FF,$C1,$CF,$CF,$C3
	.dc.b	$FC,$59,$FA,$4F,$FB,$43,$F2,$1F,$F6,$87,$E9,$5F,$E5,$EF,$D8,$5E
	.dc.b	$67,$F8,$CF,$9B,$FC,$45,$7B,$FC,$59,$7B,$FC,$D3,$59,$74,$FC,$55
	.dc.b	$9F,$F7,$3C,$78,$C3,$F5,$BC,$56,$EB,$C3,$F5,$0F,$89,$FE,$71,$87
	.dc.b	$09,$CF,$C4,$FF,$3B,$FB,$6F,$F1,$A7,$DF,$BF,$6C,$7C,$33,$F2,$27
	.dc.b	$E5,$3F,$AE,$FC,$B7,$F8,$67,$3B,$1E,$38,$70,$BD,$F7,$AF,$FA,$AD
	.dc.b	$EF,$83,$FC,$A5,$9F,$7F,$D8,$FF,$C1,$D9,$72,$D2,$D6,$B3,$E9,$A7
	.dc.b	$EF,$BF,$0F,$F4,$4E,$6B,$CD,$3F,$C0,$BD,$AF,$EA,$DE,$3F,$F9,$3F
	.dc.b	$F2,$67,$E5,$EF,$D2,$7F,$CC,$BF,$B9,$FF,$8E,$3B,$BF,$DA,$FC,$2C
	.dc.b	$38,$5E,$FF,$04,$5E,$FF,$7C,$FF,$36,$56,$B1,$FF,$13,$3C,$3F,$4C
	.dc.b	$BB,$FF,$9B,$F4,$F0,$D7,$83,$EF,$7F,$13,$7C,$72,$EB,$1E,$1B,$FF
	.dc.b	$3C,$7E,$C5,$FE,$67,$FD,$8D,$FA,$63,$F4,$BF,$C5,$F9,$4E,$97,$EC
	.dc.b	$5F,$DA,$F8,$D8,$F1,$C3,$87,$92,$C3,$85,$BF,$2A,$EF,$A7,$FA,$17
	.dc.b	$4E,$FA,$7F,$90,$72,$E5,$F8,$35,$AD,$67,$D3,$2F,$17,$E6,$9F,$E0
	.dc.b	$5B,$F0,$61,$E6,$D9,$F0,$3F,$E0,$9F,$88,$FF,$41,$7E,$37,$FE,$A9
	.dc.b	$FC,$AB,$EA,$FD,$03,$F6,$77,$EE,$7F,$EF,$5E,$56,$1C,$3F,$12,$61
	.dc.b	$C2,$F7,$F8,$8F,$FA,$7E,$B5,$F9,$6F,$F9,$65,$E1,$57,$8F,$FD,$45
	.dc.b	$A7,$87,$C1,$FF,$3E,$62,$FD,$7F,$CD,$5A,$CF,$BD,$6B,$E7,$FF,$C4
	.dc.b	$5F,$49,$FE,$9B,$FD,$41,$FA,$17,$D5,$FD,$17,$FC,$31,$F9,$33,$F0
	.dc.b	$E7,$EC,$DF,$EB,$3C,$38,$5E,$FE,$CD,$7B,$FD,$4F,$5B,$7F,$AD,$F3
	.dc.b	$EB,$D7,$FE,$73,$C7,$8F,$81,$6B,$59,$73,$C5,$F9,$4F,$9A,$73,$83
	.dc.b	$6F,$2F,$E9,$1F,$E3,$BF,$CC,$3F,$D6,$9F,$94,$3E,$DD,$FB,$03,$F2
	.dc.b	$2F,$F2,$37,$DC,$BF,$7A,$FF,$4F,$F8,$2C,$78,$E1,$C2,$F7,$5F,$80
	.dc.b	$AD,$6B,$FE,$D1,$CB,$B7,$87,$FE,$8D,$C7,$8E,$76,$B5,$97,$3F,$C0
	.dc.b	$3F,$C7,$3F,$AB,$79,$E7,$35,$E5,$9F,$F1,$EF,$E3,$BF,$E5,$5F,$C1
	.dc.b	$FF,$D9,$DF,$7B,$FE,$E0,$FC,$93,$F0,$FF,$A4,$3F,$71,$43,$E3,$7F
	.dc.b	$DC,$D7,$C3,$F1,$4C,$F9,$2C,$38,$5F,$F8,$7B,$8F,$FA,$E6,$CC,$FB
	.dc.b	$3F,$E5,$3C,$79,$7E,$35,$FE,$A9,$7A,$5F,$8B,$4F,$0D,$78,$0F,$F7
	.dc.b	$A5,$F9,$F2,$F0,$4F,$F8,$FE,$7F,$DF,$5F,$85,$7F,$BA,$7F,$6F,$7E
	.dc.b	$D6,$FC,$17,$F8,$17,$E2,$EF,$DF,$BF,$C9,$1F,$77,$C3,$1F,$F8,$01
	.dc.b	$87,$0B,$DD,$7F,$93,$7E,$F3,$54,$AE,$BF,$83,$FF,$B0,$73,$F0,$67
	.dc.b	$DE,$BF,$CE,$11,$8C,$79,$F0,$7F,$A1,$63,$1E,$CF,$FA,$DF,$3F,$E3
	.dc.b	$9C,$6F,$CE,$58,$CF,$CC,$9F,$97,$3F,$31,$7E,$84,$FB,$C7,$E7,$0F
	.dc.b	$C8,$9F,$26,$D3,$A6,$7C,$F2,$E5,$8F,$1C,$38,$7F,$00,$6F,$5F,$0D
	.dc.b	$B5,$F0,$FE,$6D,$FE,$DF,$9C,$C7,$93,$6C,$57,$57,$CB,$B7,$93,$B7
	.dc.b	$FD,$49,$7C,$34,$F0,$CF,$A3,$D0,$FF,$54,$7E,$41,$E0,$F2,$3F,$D6
	.dc.b	$FF,$5B,$FA,$EF,$B8,$79,$BF,$EB,$3F,$A8,$F6,$5E,$FF,$0D,$FF,$12
	.dc.b	$79,$8B,$DF,$99,$FB,$BE,$7F,$F7,$37,$17,$CC,$D3,$FD,$87,$F9,$BF
	.dc.b	$FB,$E7,$2E,$5A,$DA,$D6,$7D,$32,$F0,$FE,$75,$E6,$9C,$D7,$97,$6F
	.dc.b	$F7,$FE,$6F,$D2,$3E,$4F,$E0,$6F,$CE,$FF,$9E,$E9,$FC,$1E,$DE,$C3
	.dc.b	$F4,$AF,$AD,$87,$0B,$DF,$B1,$F0,$29,$F9,$B6,$B5,$FC,$2E,$7D,$4F
	.dc.b	$AE,$FF,$8D,$30,$E5,$F2,$18,$8F,$F6,$A4,$63,$4F,$16,$18,$9E,$79
	.dc.b	$78,$34,$FE,$4B,$1F,$89,$F9,$2F,$C1,$FE,$31,$CE,$3F,$2B,$F9,$F6
	.dc.b	$5F,$CB,$C6,$1F,$F6,$84,$F0,$7C,$57,$FE,$E5,$FC,$2F,$F1,$BE,$2B
	.dc.b	$C3,$FF,$3A,$7E,$B6,$FC,$F1,$F9,$6F,$FB,$73,$F4,$4F,$CD,$61,$C2
	.dc.b	$F7,$5A,$FC,$7B,$F2,$DD,$1F,$E9,$4D,$3F,$53,$9F,$5F,$B5,$F4,$4E
	.dc.b	$67,$F5,$38,$3E,$AB,$39,$75,$B6,$BE,$3D,$5F,$69,$E9,$C7,$AE,$3E
	.dc.b	$29,$FF,$7E,$47,$F8,$27,$F0,$4F,$D4,$82,$13,$F0,$A7,$26,$E7,$F7
	.dc.b	$F3,$A3,$4F,$FA,$E3,$F3,$4E,$71,$CE,$3E,$FC,$E5,$E3,$19,$7F,$07
	.dc.b	$F6,$DF,$CD,$5F,$AA,$7D,$93,$FD,$87,$BF,$49,$F0,$1E,$79,$7D,$6F
	.dc.b	$D9,$AF,$C6,$E7,$FE,$87,$FA,$CE,$17,$E6,$BF,$F2,$45,$76,$3F,$27
	.dc.b	$CD,$FE,$8D,$DF,$75,$F0,$FE,$B3,$FD,$21,$E6,$75,$B7,$57,$A3,$EF
	.dc.b	$7A,$C5,$3A,$6F,$F9,$BC,$FF,$8F,$C7,$C0,$C5,$F1,$19,$EE,$F4,$63
	.dc.b	$BF,$CF,$7D,$D7,$4F,$87,$8F,$E6,$F1,$7E,$3B,$F3,$B1,$FC,$F7,$EB
	.dc.b	$7F,$C4,$2F,$B6,$7F,$1A,$31,$1F,$72,$7F,$5D,$82,$73,$5F,$81,$9E
	.dc.b	$7F,$84,$FC,$28,$FC,$1F,$E3,$5F,$82,$C3,$AC,$7E,$06,$31,$1F,$52
	.dc.b	$3E,$46,$09,$CC,$67,$83,$FA,$3F,$C3,$3F,$46,$ED,$3F,$91,$DD,$FB
	.dc.b	$47,$CF,$CB,$F9,$EE,$5F,$42,$31,$1F,$BE,$8E,$35,$F5,$23,$9E,$95
	.dc.b	$F3,$C7,$E7,$7F,$20,$C4,$43,$FA,$4D,$D7,$F8,$47,$47,$E4,$5F,$E2
	.dc.b	$8F,$F1,$5C,$7F,$29,$3F,$C5,$EF,$E0,$0B,$C3,$F0,$F0,$FE,$1D,$86
	.dc.b	$A3,$F4,$95,$EA,$8F,$09,$F9,$33,$F4,$FF,$BA,$FC,$57,$D3,$3F,$B5
	.dc.b	$FD,$9B,$E5,$BF,$A2,$DB,$FB,$8C,$62,$3F,$69,$1F,$B8,$8C,$47,$EA
	.dc.b	$3B,$3F,$DE,$15,$D2,$3F,$B8,$79,$3F,$8B,$8F,$E6,$53,$F8,$49,$FC
	.dc.b	$BC,$7E,$EA,$31,$1F,$7F,$E4,$9C,$FF,$98,$AF,$E7,$B1,$FD,$21,$FD
	.dc.b	$76,$DF,$CF,$63,$F4,$71,$FC,$27,$5C,$74,$FE,$E7,$86,$CF,$E6,$8E
	.dc.b	$7D,$5F,$B1,$C3,$AF,$6F,$E5,$7C,$0F,$F0,$8E,$5F,$C0,$AF,$E5,$1F
	.dc.b	$E1,$4E,$4C,$ED,$F8,$69,$FE,$1D,$3F,$7C,$E3,$E0,$7A,$5F,$F0,$5C
	.dc.b	$23,$11,$FC,$8B,$F7,$AF,$E3,$1C,$1F,$9F,$C7,$FC,$43,$18,$8F,$E2
	.dc.b	$B1,$FC,$A1,$FD,$74,$F9,$0F,$E5,$FB,$FE,$83,$F2,$8F,$E6,$05,$93
	.dc.b	$FA,$4E,$7F,$17,$67,$E1,$EF,$F6,$C7,$33,$B7,$E9,$39,$CF,$8C,$7E
	.dc.b	$22,$BD,$53,$9B,$F0,$7D,$9F,$DC,$AF,$CC,$E7,$FB,$08,$F3,$C7,$C5
	.dc.b	$3E,$16,$7C,$0F,$E0,$9F,$8A,$FF,$47,$D7,$7C,$5F,$A7,$F0,$5F,$81
	.dc.b	$7C,$8E,$D5,$77,$E1,$FE,$0D,$87,$F3,$9A,$FE,$35,$1F,$CF,$2F,$FC
	.dc.b	$85,$97,$F9,$31,$E9,$C7,$FE,$3E,$7F,$A2,$73,$FF,$44,$9F,$F2,$8F
	.dc.b	$1F,$EB,$67,$8C,$7F,$36,$FC,$63,$FA,$54,$63,$54,$7F,$81,$6F,$95
	.dc.b	$7F,$3C,$FC,$DF,$F7,$77,$BF,$1F,$8E,$5F,$E0,$C7,$84,$73,$FA,$4D
	.dc.b	$CF,$E5,$6F,$CC,$D7,$FB,$4C,$F8,$5F,$F2,$BF,$3F,$F6,$98,$C4,$7E
	.dc.b	$68,$E2,$7F,$03,$E8,$FE,$43,$1A,$8F,$F3,$36,$DF,$DD,$A3,$11,$FC
	.dc.b	$7D,$BA,$FF,$88,$1C,$DE,$F0,$FE,$59,$87,$6B,$FE,$44,$EE,$3F,$98
	.dc.b	$1B,$8F,$C7,$71,$2D,$FA,$D7,$E6,$5B,$3F,$D1,$31,$F0,$4F,$F6,$A7
	.dc.b	$EC,$3F,$DB,$31,$F4,$3F,$DB,$A7,$67,$F4,$C3,$67,$F5,$AE,$4F,$8D
	.dc.b	$1F,$DE,$AB,$E1,$B6,$76,$7F,$81,$2B,$6F,$F0,$93,$A9,$FE,$08,$3F
	.dc.b	$C1,$F1,$FE,$0B,$AF,$C8,$3F,$E2,$2F,$07,$EE,$B6,$7F,$B8,$3F,$E2
	.dc.b	$6F,$CD,$FF,$0A,$FF,$8A,$3D,$5D,$CF,$F8,$91,$FC,$7C,$7F,$8B,$3D
	.dc.b	$0F,$E8,$BB,$7E,$92,$BF,$C4,$38,$FF,$67,$D3,$E4,$C7,$F4,$5E,$DF
	.dc.b	$9D,$87,$C7,$E0,$FF,$03,$D6,$9F,$E5,$07,$F7,$77,$0F,$F8,$E3,$5F
	.dc.b	$EF,$D1,$88,$FF,$07,$98,$9F,$53,$FE,$2A,$C3,$BE,$1F,$E3,$CC,$BF
	.dc.b	$AC,$5B,$F4,$AF,$E2,$4F,$9B,$7F,$F3,$7D,$7F,$9E,$B8,$38,$BF,$E2
	.dc.b	$4F,$07,$E1,$B4,$F9,$27,$F9,$C3,$9F,$FC,$E9,$3C,$2B,$F6,$73,$FE
	.dc.b	$14,$BE,$D5,$F2,$A3,$FC,$56,$7F,$76,$8F,$13,$FD,$B0,$FD,$6E,$9F
	.dc.b	$D2,$5F,$BA,$7F,$6B,$AE,$11,$FB,$3E,$7F,$04,$FF,$8E,$EB,$9B,$FD
	.dc.b	$DB,$0C,$BF,$DE,$26,$BF,$E1,$E2,$9F,$E0,$F8,$B7,$F8,$E3,$FC,$0D
	.dc.b	$1F,$C6,$6F,$FB,$F7,$C5,$F0,$8F,$EA,$98,$7C,$D7,$F4,$07,$FA,$35
	.dc.b	$C3,$FE,$2B,$3F,$DD,$B7,$FB,$1C,$FF,$51,$1F,$C9,$5F,$F2,$EE,$BF
	.dc.b	$EE,$2F,$07,$53,$FA,$93,$8D,$F4,$DF,$FB,$84,$7F,$2B,$BF,$F2,$CB
	.dc.b	$E4,$6B,$FE,$32,$9E,$13,$FE,$84,$8F,$F3,$BC,$FE,$A5,$F0,$4E,$A7
	.dc.b	$FA,$64,$71,$9D,$3D,$CF,$D3,$1F,$E2,$C9,$F2,$71,$FE,$11,$B1,$FE
	.dc.b	$5F,$9F,$F4,$AC,$7F,$8D,$0F,$51,$B7,$31,$FE,$BF,$E3,$F1,$5F,$FA
	.dc.b	$57,$8F,$F1,$DE,$1F,$EE,$2B,$FD,$83,$D5,$5F,$F2,$1D,$7F,$08,$7F
	.dc.b	$C8,$BC,$B7,$87,$FC,$77,$3F,$80,$8F,$E6,$19,$7F,$96,$3E,$F3,$E7
	.dc.b	$DF,$EA,$EB,$F6,$93,$FE,$90,$F0,$BF,$66,$FF,$0F,$8B,$3F,$E8,$39
	.dc.b	$FF,$11,$6B,$F5,$79,$FF,$63,$7F,$EB,$4E,$DF,$D9,$F2,$FF,$B7,$36
	.dc.b	$B8,$FA,$B2,$FD,$D6,$3F,$9A,$8E,$75,$FE,$97,$D3,$F8,$04,$7F,$89
	.dc.b	$2B,$F8,$44,$FF,$9B,$67,$CF,$3F,$E4,$EF,$C8,$FD,$75,$FE,$5C,$F6
	.dc.b	$2B,$FE,$89,$3F,$CD,$58,$7C,$ED,$7F,$D8,$FC,$3A,$9F,$E1,$E6,$B5
	.dc.b	$FF,$5C,$4F,$D3,$AF,$74,$FF,$AE,$AB,$BE,$BF,$E8,$97,$FA,$77,$A5
	.dc.b	$FE,$7E,$E7,$FF,$5D,$69,$FE,$03,$3F,$DE,$A7,$F5,$3F,$1F,$FC,$08
	.dc.b	$FF,$9D,$8C,$1B,$3F,$D5,$0F,$FB,$F7,$9F,$FD,$7F,$1A,$3F,$D1,$0F
	.dc.b	$FB,$21,$C7,$27,$F9,$25,$FF,$5D,$49,$FF,$6A,$D7,$2C,$FF,$D9,$71
	.dc.b	$FC,$0C,$FD,$91,$C8,$FF,$4D,$9C,$8F,$F6,$85,$F3,$5E,$0F,$B9,$FF
	.dc.b	$4C,$47,$29,$EB,$87,$49,$FF,$11,$61,$FE,$DF,$C3,$96,$1C,$6B,$FD
	.dc.b	$17,$4C,$30,$FF,$0A,$3F,$DD,$F7,$FE,$AA,$FC,$A7,$31,$E3,$0F,$F8
	.dc.b	$62,$3F,$D5,$FF,$A5,$FF,$BB,$DB,$F7,$7E,$DA,$E1,$2F,$D2,$B1,$1E
	.dc.b	$DF,$82,$7E,$59,$88,$F9,$50,$F8,$BC,$42,$3C,$6F,$1F,$0B,$3F,$0D
	.dc.b	$E1,$8F,$F3,$52,$CB,$38,$F2,$DF,$BF,$0F,$C8,$BF,$61,$F3,$B2,$D6
	.dc.b	$3E,$0F,$27,$A4,$FF,$B0,$5E,$95,$C6,$3E,$24,$7B,$63,$11,$F6,$7F
	.dc.b	$4A,$EF,$BF,$FB,$51,$F6,$4E,$A7,$F1,$11,$EC,$6D,$FC,$0F,$CF,$7C
	.dc.b	$3C,$7E,$69,$47,$8B,$0F,$F6,$FB,$E1,$7F,$DC,$3C,$7E,$09,$FF,$46
	.dc.b	$3E,$08,$FF,$72,$F8,$DC,$67,$FD,$20,$F1,$7F,$D2,$5C,$BC,$5F,$F7
	.dc.b	$03,$C5,$FF,$4B,$63,$C5,$FF,$72,$3C,$5F,$F5,$47,$1F,$D8,$7D,$FE
	.dc.b	$E7,$E6,$FA,$BF,$DC,$13,$FE,$47,$8C,$47,$F8,$3F,$83,$15,$FE,$F4
	.dc.b	$EB,$F9,$F5,$AA,$10,$FF,$51,$F1,$F0,$AF,$F7,$7F,$77,$DE,$AE,$15
	.dc.b	$B8,$F1,$CF,$FC,$00,$EE,$EA,$DB,$E5,$5F,$C6,$23,$F7,$9E,$47,$FA
	.dc.b	$3F,$0F,$17,$6F,$EB,$27,$75,$E4,$8F,$C5,$F2,$AB,$D1,$E8,$FE,$EF
	.dc.b	$0E,$58,$7F,$07,$C3,$D5,$3C,$EF,$84,$E6,$7D,$F3,$D2,$FE,$3E,$1F
	.dc.b	$B3,$C3,$9D,$F9,$CF,$E2,$5D,$18,$C3,$C9,$1C,$4E,$98,$71,$7D,$2F
	.dc.b	$F1,$7C,$3E,$9B,$F8,$FC,$7C,$72,$E6,$FC,$96,$3F,$EE,$CF,$03,$FA
	.dc.b	$96,$3F,$CD,$FC,$7F,$F0,$57,$83,$F5,$63,$EB,$79,$29,$FE,$EB,$8F
	.dc.b	$F7,$3F,$13,$FC,$C5,$5F,$E4,$F9,$FF,$2D,$78,$DF,$EA,$EE,$AF,$B3
	.dc.b	$3E,$AD,$7F,$95,$E5,$FD,$0F,$1F,$E5,$5D,$5F,$DE,$DF,$E5,$9D,$36
	.dc.b	$BF,$6D,$3F,$6D,$FD,$C4,$5E,$1E,$B8,$3F,$7B,$FC,$46,$1F,$46,$3D
	.dc.b	$51,$F8,$88,$F6,$47,$E4,$7F,$1B,$FC,$4C,$62,$3F,$2B,$1F,$C0,$EB
	.dc.b	$AF,$C1,$7D,$39,$E2,$FF,$B0,$23,$C3,$F0,$5F,$9C,$CB,$E7,$D7,$3F
	.dc.b	$C4,$7C,$78,$F5,$84,$78,$DF,$07,$D1,$78,$9B,$D5,$F9,$47,$E9,$7A
	.dc.b	$5D,$EF,$D9,$1F,$A9,$F4,$BC,$19,$FE,$B7,$CF,$FC,$9D,$71,$F5,$6E
	.dc.b	$9E,$4C,$15,$F7,$7B,$3D,$95,$6F,$EF,$75,$7F,$8C,$1F,$F4,$17,$47
	.dc.b	$FA,$57,$D2,$FF,$77,$F4,$7F,$5D,$E8,$F9,$18,$FF,$52,$E8,$E5,$AF
	.dc.b	$F3,$5E,$8B,$7F,$D8,$FB,$7F,$B2,$57,$92,$31,$9F,$F7,$1A,$F3,$F0
	.dc.b	$7D,$CF,$C1,$7E,$4F,$0F,$D5,$7E,$43,$F6,$BF,$03,$F9,$0D,$E5,$F9
	.dc.b	$0F,$F0,$DF,$DD,$7E,$C3,$F9,$B7,$E5,$63,$F4,$FE,$E7,$E4,$BF,$32
	.dc.b	$7F,$8A,$FD,$97,$81,$FC,$3B,$E0,$7F,$49,$F9,$37,$81,$B8,$78,$1F
	.dc.b	$E3,$BF,$4F,$E0,$7F,$8F,$3E,$07,$F1,$DF,$87,$F8,$30,$FD,$77,$4F
	.dc.b	$F2,$3E,$BF,$3F,$55,$9F,$E2,$B3,$F8,$27,$F8,$BC,$5E,$57,$EC,$BF
	.dc.b	$A2,$E9,$FD,$E5,$7F,$15,$9F,$EA,$1E,$7F,$EE,$7C,$5F,$F6,$E4,$FF
	.dc.b	$AC,$F9,$7E,$8F,$91,$FC,$2E,$FF,$83,$F5,$7F,$12,$AF,$E3,$95,$A7
	.dc.b	$F8,$91,$FE,$2C,$3F,$92,$1F,$CE,$63,$FC,$58,$F1,$9E,$6F,$F8,$5C
	.dc.b	$E3,$1F,$CC,$5F,$F6,$2C,$69,$FF,$13,$9C,$67,$F0,$AE,$0F,$F1,$64
	.dc.b	$6F,$0F,$0B,$63,$81,$FE,$32,$3F,$C4,$0F,$57,$FC,$69,$F1,$7F,$CC
	.dc.b	$2F,$FB,$A3,$F8,$47,$FB,$B7,$C1,$FD,$57,$FB,$C7,$C5,$FE,$E9,$BE
	.dc.b	$87,$BE,$FF,$C2,$3F,$33,$FC,$2A,$7F,$BF,$3F,$C2,$63,$FC,$99,$1F
	.dc.b	$E0,$1F,$9A,$F0,$47,$F8,$61,$F0,$3F,$E1,$AF,$98,$F0,$3F,$E1,$CF
	.dc.b	$CC,$7C,$0D,$C3,$C0,$FF,$8E,$FF,$28,$F0,$3F,$E2,$27,$C0,$FF,$3F
	.dc.b	$3F,$C4,$BE,$8F,$67,$FC,$85,$F0,$7D,$9F,$F2,$27,$E8,$3D,$9B,$7D
	.dc.b	$9F,$F2,$D7,$E7,$DD,$9F,$F2,$53,$DB,$4F,$F0,$E7,$C1,$7B,$27,$FC
	.dc.b	$9C,$F7,$7F,$CC,$5F,$05,$DD,$FF,$24,$7B,$7D,$5F,$F2,$3E,$15,$7C
	.dc.b	$0F,$77,$FC,$D4,$F7,$7F,$C7,$FE,$FF,$9D,$FF,$1B,$4F,$1F,$7F,$84
	.dc.b	$7F,$9E,$DE,$0F,$F9,$87,$DF,$E0,$FF,$BC,$27,$C1,$E6,$70,$83,$FE
	.dc.b	$43,$EF,$FE,$9B,$EF,$FF,$92,$1B,$E3,$BB,$3F,$63,$7F,$C2,$7C,$67
	.dc.b	$BF,$FE,$56,$BF,$EA,$AF,$F9,$21,$FE,$AF,$ED,$73,$AE,$0F,$77,$97
	.dc.b	$B1,$C5,$BA,$77,$AF,$F0,$9C,$7F,$A5,$FE,$1B,$FD,$31,$97,$8A,$3F
	.dc.b	$9D,$1F,$AC,$9D,$BF,$A8,$7F,$C6,$5E,$4F,$FA,$56,$B8,$C6,$27,$FC
	.dc.b	$4E,$77,$3F,$B5,$4F,$F8,$92,$3F,$C1,$B1,$C4,$D1,$FD,$40,$E3,$5E
	.dc.b	$3C,$FF,$C4,$31,$C4,$FF,$12,$57,$FA,$4E,$FC,$31,$83,$F7,$77,$FE
	.dc.b	$4C,$9F,$F8,$00,$FF,$1D,$8F,$EA,$A7,$F8,$C7,$DD,$FF,$66,$7A,$FF
	.dc.b	$ED,$0C,$3F,$C3,$1E,$F6,$AB,$FC,$9F,$E5,$7F,$9C,$5F,$E4,$47,$F8
	.dc.b	$E7,$DA,$FF,$00,$4E,$23,$FC,$AB,$1F,$DD,$F4,$FE,$33,$F0,$3F,$D7
	.dc.b	$3C,$AF,$EA,$FE,$C5,$BF,$DE,$9F,$EC,$99,$FE,$B2,$E0,$47,$F7,$2D
	.dc.b	$7F,$C2,$FE,$B7,$E0,$67,$FC,$08,$FE,$05,$FF,$03,$7A,$9F,$81,$6F
	.dc.b	$17,$CE,$DB,$FC,$2D,$E9,$FC,$89,$FF,$06,$3F,$98,$D3,$E8,$4F,$F8
	.dc.b	$7B,$D3,$FF,$0A,$65,$FE,$26,$FA,$4F,$04,$FF,$87,$DF,$03,$FE,$21
	.dc.b	$FD,$B3,$C0,$DC,$BC,$0F,$F8,$A3,$8B,$B5,$FF,$8C,$39,$3A,$E9,$FE
	.dc.b	$2F,$F5,$3F,$AC,$69,$E2,$C1,$C7,$8C,$3F,$D0,$F5,$FD,$65,$FF,$44
	.dc.b	$C7,$F5,$DE,$57,$BB,$FD,$5B,$1F,$F0,$8F,$AB,$FE,$91,$E8,$FB,$59
	.dc.b	$FD,$4C,$FB,$F3,$FD,$46,$DF,$D1,$E7,$AF,$F2,$33,$E9,$C7,$FA,$4A
	.dc.b	$FE,$53,$FE,$41,$FD,$C3,$E2,$BF,$E4,$67,$E2,$BF,$E4,$8F,$D5,$7E
	.dc.b	$2B,$78,$BE,$2B,$FE,$67,$F9,$EE,$6F,$F9,$39,$F8,$F8,$7F,$05,$FC
	.dc.b	$57,$E3,$BF,$E4,$3F,$CB,$3B,$D7,$F9,$BD,$EE,$FF,$9C,$7F,$01,$EE
	.dc.b	$DC,$3B,$BF,$EA,$9F,$C3,$BB,$BF,$E7,$77,$BD,$FF,$93,$FF,$05,$EF
	.dc.b	$3F,$EA,$FE,$6F,$F5,$87,$47,$0E,$6B,$7F,$7A,$7F,$D0,$13,$EA,$9B
	.dc.b	$3F,$AB,$A4,$FF,$83,$39,$9C,$BF,$56,$BF,$DE,$7E,$3F,$F2,$5F,$0B
	.dc.b	$FB,$CC,$7F,$A2,$4F,$EF,$3F,$21,$FD,$E9,$FF,$20,$1F,$DD,$9C,$3F
	.dc.b	$A9,$3F,$97,$38,$7F,$D6,$AF,$F9,$2F,$D0,$E3,$1D,$0F,$ED,$EF,$83
	.dc.b	$A5,$3B,$9F,$F5,$23,$FE,$53,$C7,$8C,$7F,$2A,$8C,$3E,$19,$E2,$FF
	.dc.b	$7A,$9F,$C8,$CF,$F8,$53,$0F,$C6,$4F,$98,$FE,$00,$FE,$3C,$C3,$7E
	.dc.b	$43,$FD,$59,$7D,$23,$5F,$38,$E2,$9E,$87,$0C,$D5,$FE,$65,$AD,$56
	.dc.b	$EF,$95,$6A,$BF,$AC,$FC,$78,$87,$37,$FC,$DC,$E3,$CC,$C4,$7F,$60
	.dc.b	$D7,$E9,$3E,$A7,$A6,$9A,$1F,$25,$EA,$DB,$E1,$7E,$8F,$E4,$DB,$4F
	.dc.b	$9B,$5E,$1E,$F7,$8B,$4A,$79,$74,$E9,$A7,$6D,$BD,$6F,$FC,$00,$6B
	.dc.b	$F0,$BC,$9E,$D1,$88,$47,$78,$C7,$5F,$76,$DF,$77,$81,$EE,$AF,$B2
	.dc.b	$F8,$74,$FC,$BF,$C5,$FB,$67,$F2,$59,$F8,$27,$F3,$6F,$77,$F3,$BF
	.dc.b	$A2,$77,$6E,$1D,$DF,$D0,$F1,$F1,$BF,$CE,$3C,$5F,$CF,$4F,$16,$BC
	.dc.b	$5F,$D1,$6D,$C5,$FC,$7B,$C5,$FC,$8E,$DC,$5B,$3E,$FF,$B2,$FC,$27
	.dc.b	$4F,$F0,$D8,$FC,$1E,$A9,$CC,$89,$B7,$E1,$67,$57,$FE,$D0,$D7,$E8
	.dc.b	$DF,$0B,$E1,$1F,$46,$31,$1F,$7F,$0F,$F1,$34,$61,$E8,$EB,$2E,$77
	.dc.b	$FE,$DC,$79,$3F,$A0,$AF,$F2,$D7,$A9,$F9,$D8,$FF,$0F,$D7,$B3,$9B
	.dc.b	$F7,$31,$D8,$E0,$FF,$46,$8F,$83,$1F,$AA,$9E,$4E,$CF,$80,$72,$CB
	.dc.b	$DE,$E0,$D9,$BA,$FD,$74,$F2,$E6,$FE,$DF,$D7,$FC,$99,$B8,$FF,$AE
	.dc.b	$5F,$EE,$7E,$27,$27,$FD,$E5,$DB,$FD,$7F,$C6,$FE,$C7,$5F,$EF,$89
	.dc.b	$FF,$7B,$7C,$6E,$FC,$8F,$9D,$EA,$FD,$5A,$FA,$DF,$25,$FC,$CD,$FB
	.dc.b	$11,$FE,$20,$BF,$ED,$3B,$7C,$99,$FB,$9E,$9F,$9F,$C2,$FB,$0E,$DB
	.dc.b	$1F,$9F,$9E,$CF,$E0,$F8,$3F,$4D,$D7,$87,$F5,$03,$C1,$F2,$38,$F0
	.dc.b	$2D,$FF,$AD,$B9,$7B,$F4,$FE,$CD,$BD,$1F,$95,$9F,$36,$4D,$3F,$84
	.dc.b	$78,$1F,$56,$3E,$04,$7F,$0B,$EA,$C4,$7B,$33,$FF,$5E,$E3,$EC,$7F
	.dc.b	$85,$3E,$ED,$3F,$97,$BF,$03,$2F,$16,$5E,$0C,$7F,$8A,$78,$1D,$31
	.dc.b	$F1,$75,$FE,$52,$BF,$C9,$6F,$E5,$1F,$F2,$69,$D8,$F4,$F1,$7C,$89
	.dc.b	$F4,$65,$F5,$E7,$CA,$7F,$3A,$C3,$E6,$BF,$9B,$E2,$FC,$2B,$FE,$EA
	.dc.b	$7F,$02,$FF,$3B,$F8,$8F,$F7,$DB,$E0,$3F,$35,$1F,$DC,$1F,$93,$7F
	.dc.b	$D9,$1F,$E8,$FC,$5D,$9F,$EC,$83,$6F,$B1,$3F,$9D,$8F,$EC,$07,$FB
	.dc.b	$5B,$8B,$D0,$FD,$93,$DB,$C5,$F2,$1F,$C2,$E8,$8F,$DD,$C7,$C2,$E1
	.dc.b	$F9,$17,$F9,$7A,$FB,$1D,$9F,$AC,$8C,$63,$F1,$2F,$9B,$F2,$DE,$BB
	.dc.b	$7D,$87,$E6,$F0,$F8,$6B,$A3,$E0,$8F,$1F,$07,$82,$7F,$46,$FD,$77
	.dc.b	$C5,$1D,$DA,$FF,$6C,$F0,$BF,$01,$CD,$F6,$72,$F3,$63,$FA,$8A,$F3
	.dc.b	$71,$7E,$49,$FE,$25,$2F,$C3,$7F,$91,$5F,$F0,$3F,$3B,$FD,$31,$5F
	.dc.b	$EA,$5E,$3F,$B6,$FF,$AB,$1F,$DA,$BF,$6C,$7C,$4F,$FA,$47,$4F,$B1
	.dc.b	$1F,$E7,$BE,$7F,$EF,$9F,$37,$FC,$42,$3E,$43,$FD,$D1,$A7,$F1,$F3
	.dc.b	$FC,$42,$3C,$1F,$E3,$FA,$7F,$1C,$E2,$FC,$14,$FF,$B7,$8F,$A4,$FF
	.dc.b	$86,$9F,$C0,$1F,$E6,$7F,$5F,$8E,$1F,$E0,$D9,$FF,$07,$47,$F3,$3E
	.dc.b	$1F,$E4,$95,$FC,$B7,$2F,$B0,$FF,$9B,$F1,$FC,$37,$1F,$F5,$0C,$79
	.dc.b	$4F,$F3,$3E,$6F,$B5,$97,$F2,$CE,$6F,$D8,$35,$FF,$0F,$E9,$E9,$AF
	.dc.b	$E7,$53,$E8,$69,$F6,$E3,$4F,$C1,$E4,$F1,$CF,$F3,$8C,$7F,$34,$D7
	.dc.b	$F8,$94,$F9,$F9,$7C,$2F,$F4,$77,$F8,$04,$FD,$18,$ED,$97,$F9,$07
	.dc.b	$4E,$B1,$F8,$BC,$FA,$B7,$0E,$B7,$FE,$2A,$CF,$AC,$B5,$FD,$2E,$3F
	.dc.b	$E4,$BA,$F6,$4E,$5F,$D2,$1F,$E3,$2C,$FB,$C7,$F8,$FB,$0F,$98,$FF
	.dc.b	$94,$1F,$96,$FE,$77,$3E,$F8,$1F,$5C,$7F,$91,$F2,$F3,$BF,$E4,$E7
	.dc.b	$8B,$FE,$50,$CB,$CE,$F9,$8F,$F0,$AD,$FF,$93,$74,$F9,$F1,$FC,$A9
	.dc.b	$B7,$C0,$FF,$1A,$72,$7F,$81,$2F,$FC,$72,$D7,$EE,$BF,$E0,$5B,$FF
	.dc.b	$0A,$B5,$75,$2F,$F6,$18,$7F,$39,$BE,$6D,$79,$BF,$CF,$6B,$FC,$85
	.dc.b	$7F,$D0,$5F,$F2,$13,$FE,$04,$8F,$D8,$9F,$BC,$71,$BF,$EA,$1D,$45
	.dc.b	$3F,$9D,$1F,$E7,$4E,$2C,$3F,$D6,$4F,$F0,$71,$CC,$D9,$FE,$01,$8F
	.dc.b	$E0,$EF,$F4,$1E,$4F,$F0,$0F,$0F,$F8,$C7,$C9,$FF,$3F,$F2,$FF,$BE
	.dc.b	$BC,$1F,$F7,$25,$7F,$8D,$BC,$0F,$F0,$57,$83,$FE,$38,$F8,$BF,$F0
	.dc.b	$56,$5F,$DF,$DF,$EF,$33,$CA,$7F,$C7,$BD,$5F,$E0,$BA,$FE,$C1,$9F
	.dc.b	$F8,$2E,$3F,$C1,$4F,$F7,$13,$FB,$7E,$5F,$C5,$9F,$E2,$73,$FC,$55
	.dc.b	$FF,$09,$73,$6B,$C1,$FF,$1F,$BF,$AE,$8F,$E9,$33,$FE,$36,$3F,$D3
	.dc.b	$3D,$1F,$D4,$23,$E3,$9F,$D1,$78,$FE,$CC,$FF,$4C,$7F,$0E,$FF,$40
	.dc.b	$E3,$FC,$1B,$5F,$C1,$F2,$7E,$0A,$3F,$A9,$9F,$3C,$A7,$F0,$3E,$2F
	.dc.b	$E6,$B7,$FE,$B0,$7F,$9B,$DF,$82,$3E,$13,$E0,$9F,$F1,$0E,$9F,$E5
	.dc.b	$9B,$ED,$5F,$1F,$0F,$F3,$26,$1F,$6A,$FB,$1F,$E1,$AD,$FF,$A5,$47
	.dc.b	$FA,$CF,$B3,$FA,$44,$FF,$AD,$70,$F4,$76,$E2,$3E,$FF,$17,$DD,$7F
	.dc.b	$C0,$5C,$3F,$61,$FF,$01,$1F,$5B,$3F,$F3,$17,$37,$D3,$7F,$C6,$D1
	.dc.b	$FC,$E3,$7E,$8F,$F9,$5F,$83,$FA,$D3,$E1,$D3,$FC,$F1,$AF,$F8,$BB
	.dc.b	$4F,$6E,$1F,$E8,$18,$FE,$EF,$3A,$DB,$F5,$37,$ED,$7E,$0B,$EF,$3E
	.dc.b	$85,$7F,$95,$6B,$FA,$25,$F4,$87,$B7,$F1,$2B,$FC,$7E,$F8,$70,$FF
	.dc.b	$45,$3E,$08,$CE,$1E,$07,$FC,$A7,$1F,$C3,$67,$71,$E0,$AF,$F4,$C4
	.dc.b	$F8,$27,$FD,$29,$F2,$BD,$EB,$F3,$94,$E1,$FD,$BF,$07,$E8,$67,$FC
	.dc.b	$65,$DB,$FD,$51,$FE,$D0,$7E,$D4,$FF,$35,$73,$7F,$A4,$7C,$CF,$D6
	.dc.b	$47,$F9,$B6,$FF,$D2,$9F,$0B,$FE,$36,$C7,$FB,$CF,$5F,$D5,$7F,$B7
	.dc.b	$D7,$F3,$FD,$DF,$19,$DF,$AB,$FB,$AF,$C6,$7E,$16,$BF,$BA,$1F,$E0
	.dc.b	$E7,$0F,$F7,$5A,$46,$23,$89,$FA,$E3,$FA,$41,$C4,$C7,$83,$E2,$7C
	.dc.b	$51,$FD,$CA,$3F,$69,$87,$FB,$06,$BA,$D7,$0F,$33,$F0,$D7,$B7,$FC
	.dc.b	$76,$F1,$E4,$FD,$33,$5F,$F6,$43,$D7,$1F,$2D,$7F,$84,$31,$47,$E1
	.dc.b	$A7,$C5,$E2,$FF,$8C,$5F,$27,$85,$E0,$BF,$4B,$E0,$7D,$7E,$67,$81
	.dc.b	$B8,$F8,$1F,$79,$D4,$F8,$87,$F9,$D7,$4F,$E9,$57,$F7,$FC,$FE,$D3
	.dc.b	$F5,$9E,$CF,$D8,$F1,$7B,$36,$FB,$3F,$73,$6F,$6F,$67,$D3,$DB,$AC
	.dc.b	$F8,$3C,$8F,$63,$E2,$F0,$BB,$D2,$3F,$B7,$40,$FF,$0F,$E7,$F9,$3F
	.dc.b	$3F,$C3,$5B,$A4,$7F,$6F,$9F,$CC,$71,$74,$9E,$8F,$AA,$B8,$67,$CA
	.dc.b	$7C,$0F,$17,$E2,$F0,$71,$7F,$C0,$D8,$F8,$CF,$26,$9F,$84,$8F,$CD
	.dc.b	$F0,$73,$4F,$F8,$07,$E3,$3C,$DC,$3F,$ED,$7E,$0F,$E5,$51,$88,$1F
	.dc.b	$E4,$4C,$87,$F3,$C8,$D4,$7F,$AF,$FC,$CF,$F0,$04,$63,$2F,$F4,$5B
	.dc.b	$AA,$EB,$8F,$FB,$29,$F1,$3F,$EC,$CC,$FF,$8F,$4F,$47,$8C,$7E,$9F
	.dc.b	$C2,$FE,$E8,$FD,$3E,$DF,$E0,$8F,$CF,$E6,$FE,$6F,$9F,$F2,$A8,$FE
	.dc.b	$EB,$DF,$FC,$F5,$FA,$87,$F0,$A3,$FD,$7D,$1B,$7F,$3F,$DD,$FD,$0E
	.dc.b	$3F,$B2,$C6,$0F,$C1,$3C,$9F,$E2,$07,$FB,$CB,$CF,$FF,$80,$1E,$4F
	.dc.b	$F4,$68,$FF,$73,$C7,$CA,$97,$9F,$99,$FF,$7C,$BC,$F1,$78,$DF,$33
	.dc.b	$CA,$E1,$3F,$C6,$AB,$FD,$FF,$1C,$FC,$3C,$61,$67,$FD,$AD,$E7,$FD
	.dc.b	$48,$6D,$FE,$09,$CB,$FC,$0D,$1F,$EB,$F3,$EE,$F9,$5F,$D5,$CF,$F6
	.dc.b	$F7,$97,$C5,$FF,$04,$61,$FE,$D8,$7D,$53,$FA,$88,$BD,$5F,$C2,$3B
	.dc.b	$BD,$CF,$FA,$6B,$97,$F1,$57,$F8,$9E,$6F,$F2,$94,$63,$A7,$F0,$38
	.dc.b	$7C,$5A,$FC,$57,$5A,$B8,$F4,$FE,$0A,$B5,$8F,$57,$E9,$E6,$C3,$F9
	.dc.b	$47,$7F,$E4,$A7,$31,$FC,$A2,$71,$1D,$F3,$FC,$9B,$5F,$CE,$9F,$9D
	.dc.b	$78,$F5,$7E,$51,$B8,$FE,$BC,$E7,$5F,$89,$C2,$BF,$40,$F8,$E7,$F2
	.dc.b	$AE,$9F,$C9,$57,$F5,$83,$F3,$1D,$3F,$E5,$DA,$FF,$6A,$47,$B2,$BF
	.dc.b	$0E,$FF,$B4,$F9,$FF,$5F,$CB,$FE,$0E,$E3,$FB,$75,$FE,$08,$EA,$FE
	.dc.b	$B4,$FF,$4B,$B7,$1F,$E6,$9F,$2F,$2F,$D3,$DE,$F2,$7D,$2B,$FD,$FC
	.dc.b	$FD,$2E,$3F,$E0,$4F,$BF,$83,$F8,$6C,$63,$8F,$E9,$63,$F9,$DA,$FA
	.dc.b	$3C,$9F,$66,$3E,$AC,$7D,$18,$C4,$6F,$2F,$A3,$8F,$81,$F1,$3F,$98
	.dc.b	$E3,$FA,$13,$D5,$FC,$14,$79,$A3,$C2,$D7,$F9,$87,$0F,$D2,$F1,$7F
	.dc.b	$A5,$61,$FC,$82,$7F,$C0,$47,$EB,$F9,$3E,$9F,$89,$FB,$1E,$5F,$F4
	.dc.b	$5C,$FE,$F3,$1F,$D8,$BF,$D0,$7C,$6F,$D7,$C7,$F2,$4E,$5F,$E7,$DE
	.dc.b	$37,$F8,$2E,$3F,$C9,$38,$7D,$2B,$FF,$6F,$70,$FF,$BA,$BC,$DF,$E1
	.dc.b	$11,$F9,$2E,$2F,$E9,$91,$FC,$2F,$7C,$4B,$9B,$D3,$7F,$EF,$27,$EC
	.dc.b	$6F,$F2,$B0,$FF,$7A,$71,$7E,$63,$CD,$E3,$5F,$67,$0F,$E5,$1C,$BF
	.dc.b	$A9,$BF,$F7,$8C,$7F,$97,$71,$E9,$1F,$E4,$27,$B7,$99,$F9,$8C,$BB
	.dc.b	$57,$8F,$2F,$D3,$DF,$F6,$3C,$7F,$4F,$1F,$D1,$32,$F0,$F2,$FD,$8A
	.dc.b	$FF,$7A,$67,$FA,$7C,$79,$E1,$EF,$7A,$E9,$EE,$CB,$FC,$43,$CF,$D9
	.dc.b	$FF,$78,$3F,$C1,$B0,$FE,$E5,$3E,$D9,$FF,$0B,$61,$ED,$7F,$C5,$67
	.dc.b	$B4,$FF,$15,$CF,$B5,$FF,$16,$E9,$ED,$D7,$FC,$45,$AF,$F0,$0C,$BC
	.dc.b	$6F,$7A,$FF,$80,$1A,$7A,$DB,$8F,$AE,$FF,$CA,$59,$F9,$E7,$F8,$F6
	.dc.b	$BF,$69,$F3,$67,$FE,$03,$BF,$34,$7C,$8C,$7F,$91,$BF,$E2,$07,$C0
	.dc.b	$FF,$7C,$8F,$F0,$77,$85,$CF,$2F,$D9,$67,$F9,$BA,$FE,$67,$5F,$E8
	.dc.b	$5F,$0F,$FB,$83,$F3,$DF,$EE,$18,$7D,$6A,$E1,$C5,$FD,$4E,$3E,$CE
	.dc.b	$7F,$EF,$7A,$FE,$5F,$7F,$E0,$BC,$B1,$8F,$F2,$A8,$F0,$CF,$F3,$CB
	.dc.b	$FF,$58,$35,$FC,$97,$5F,$6C,$7F,$CF,$57,$FA,$F3,$FB,$7B,$87,$FB
	.dc.b	$C7,$47,$F6,$F8,$FF,$7A,$46,$0F,$ED,$E7,$EA,$0F,$EE,$47,$F8,$53
	.dc.b	$83,$FB,$C5,$3A,$3F,$C3,$71,$FD,$E2,$31,$3F,$DE,$37,$E5,$0F,$37
	.dc.b	$FB,$76,$1F,$EF,$BA,$FF,$2F,$C6,$23,$F9,$8C,$EE,$7F,$9A,$74,$7F
	.dc.b	$77,$ED,$FF,$5A,$D7,$F8,$9B,$C2,$FE,$0A,$BC,$DF,$EE,$98,$7F,$7C
	.dc.b	$C3,$F9,$E7,$6F,$F7,$FF,$0B,$FB,$75,$FF,$8B,$A3,$FB,$93,$FD,$D2
	.dc.b	$BF,$9E,$BF,$E3,$2A,$FF,$0F,$78,$5F,$DF,$61,$B6,$E3,$0F,$F8,$3F
	.dc.b	$9B,$E9,$78,$1F,$DB,$39,$3F,$9D,$F2,$FD,$AC,$7F,$9A,$F2,$FD,$A6
	.dc.b	$BF,$C2,$EF,$8C,$7F,$87,$DA,$FF,$2E,$7E,$F6,$9F,$CB,$AF,$FC,$90
	.dc.b	$F7,$9F,$F5,$85,$F7,$7F,$DE,$97,$F9,$3A,$DB,$F9,$2B,$E7,$3F,$ED
	.dc.b	$49,$FF,$29,$6D,$FE,$83,$D7,$FD,$FB,$3F,$C4,$4F,$F1,$D1,$FE,$EC
	.dc.b	$EB,$FE,$D2,$FF,$85,$B8,$BF,$DE,$AF,$D0,$3F,$CF,$5E,$7F,$F3,$7E
	.dc.b	$0F,$F1,$A4,$7F,$1B,$B7,$07,$CD,$E5,$FF,$3D,$69,$F9,$BC,$FD,$97
	.dc.b	$FE,$E9,$DB,$D9,$8F,$FB,$CF,$6F,$95,$3F,$EE,$D7,$D9,$B7,$AA,$3E
	.dc.b	$53,$5F,$94,$FF,$9B,$FC,$EF,$95,$1F,$E1,$C7,$BD,$7F,$BA,$2F,$BB
	.dc.b	$FE,$F6,$BE,$EF,$FB,$B1,$EF,$5C,$4E,$E5,$3B,$9F,$E2,$8F,$3B,$E6
	.dc.b	$74,$FF,$A0,$A3,$FD,$3D,$E0,$FF,$63,$7F,$50,$7F,$85,$CF,$F2,$C5
	.dc.b	$FF,$93,$D7,$87,$FA,$0E,$4F,$EC,$71,$FE,$C8,$E7,$FC,$B3,$D3,$8B
	.dc.b	$E6,$BF,$E1,$DB,$FB,$1C,$1F,$E7,$4F,$80,$FE,$49,$AF,$FA,$3B,$F3
	.dc.b	$D4,$FF,$9E,$3C,$EF,$F0,$97,$07,$FA,$C7,$F5,$1C,$E5,$FE,$AF,$F3
	.dc.b	$3F,$83,$78,$D8,$8F,$F5,$C7,$8B,$D3,$36,$79,$97,$E7,$5D,$32,$65
	.dc.b	$9D,$39,$9F,$E8,$8F,$A6,$E5,$87,$1A,$E1,$5C,$FF,$15,$DE,$3F,$EC
	.dc.b	$1F,$4B,$78,$FF,$A4,$3F,$35,$DB,$FE,$92,$ED,$F3,$F9,$DF,$BD,$B7
	.dc.b	$17,$8F,$EC,$7F,$C7,$3C,$BF,$EC,$B8,$EF,$5B,$E2,$F6,$7E,$7F,$AC
	.dc.b	$3D,$FC,$7F,$2B,$87,$11,$EF,$3F,$DA,$9C,$98,$D7,$F9,$67,$9F,$AE
	.dc.b	$3F,$06,$7E,$37,$BE,$BE,$BC,$BE,$AE,$7D,$38,$7E,$1D,$FA,$78,$3D
	.dc.b	$77,$FE,$E1,$7A,$C7,$F3,$CD,$BE,$97,$EC,$5F,$E5,$4F,$4B,$FD,$CD
	.dc.b	$1E,$A9,$FE,$83,$9F,$D3,$FB,$2D,$70,$7F,$BC,$23,$FC,$A5,$FA,$16
	.dc.b	$21,$97,$F4,$5F,$8D,$C3,$FD,$1F,$C1,$B3,$FE,$30,$E3,$FF,$31,$F9
	.dc.b	$7F,$DC,$F1,$FC,$5F,$EC,$F8,$AE,$D1,$FE,$7D,$CF,$FC,$D3,$E1,$6E
	.dc.b	$3F,$B8,$EB,$FD,$65,$FC,$47,$E7,$7F,$E6,$EF,$D2,$FF,$83,$E1,$FD
	.dc.b	$3B,$C9,$FD,$E3,$FE,$BD,$7F,$DD,$9E,$47,$F1,$B7,$F8,$EF,$2F,$FA
	.dc.b	$1B,$C5,$FF,$32,$BF,$E5,$BA,$FF,$6E,$7A,$FF,$CB,$23,$4F,$F9,$D1
	.dc.b	$E1,$E4,$FF,$BC,$BC,$AF,$ED,$D5,$FE,$B6,$C7,$2F,$FA,$2F,$CB,$ED
	.dc.b	$EA,$7F,$BE,$38,$FF,$7D,$3F,$EB,$CE,$1C,$BF,$E9,$18,$F2,$7A,$9F
	.dc.b	$D3,$F3,$7C,$17,$F0,$27,$FD,$3B,$F0,$47,$FB,$4D,$FE,$F5,$C7,$C4
	.dc.b	$E7,$83,$FB,$14,$FF,$A9,$27,$87,$C6,$FF,$68,$7F,$DC,$DF,$1B,$BF
	.dc.b	$67,$C1,$EF,$F1,$F4,$FC,$54,$E3,$FE,$1D,$8F,$78,$F1,$71,$E2,$3B
	.dc.b	$47,$97,$8F,$FD,$4A,$FF,$A7,$B4,$FE,$27,$3A,$87,$17,$E3,$7C,$2F
	.dc.b	$E1,$1F,$85,$7E,$3F,$AF,$F6,$3C,$BE,$2A,$FC,$46,$1F,$C2,$67,$F8
	.dc.b	$4F,$67,$FB,$C2,$3F,$17,$02,$7E,$69,$F8,$C8,$CC,$67,$BB,$11,$CE
	.dc.b	$7F,$8B,$7A,$BE,$2F,$31,$3F,$A8,$F2,$FF,$93,$E9,$FC,$6F,$D3,$D4
	.dc.b	$FE,$DE,$7F,$C9,$D3,$FC,$0B,$7F,$87,$E4,$FF,$04,$F3,$FF,$90,$7E
	.dc.b	$6B,$FC,$1B,$8B,$DF,$C1,$FC,$4C,$D9,$FC,$0F,$83,$F9,$8C,$FF,$01
	.dc.b	$CB,$C3,$3F,$E3,$8C,$7E,$7F,$0F,$C9,$9F,$E7,$7F,$A2,$7C,$9C,$3F
	.dc.b	$9C,$F0,$FC,$8B,$FA,$BE,$DF,$CA,$9F,$E6,$5F,$87,$7C,$27,$F9,$A7
	.dc.b	$AC,$F8,$77,$93,$FA,$EE,$3F,$F5,$5F,$4F,$F9,$3F,$B3,$FD,$54,$FF
	.dc.b	$82,$A3,$FD,$86,$FE,$B3,$3F,$A9,$18,$F2,$3E,$5B,$F9,$98,$FF,$28
	.dc.b	$BF,$E8,$E9,$79,$1F,$AD,$8E,$6F,$09,$E1,$C3,$CA,$5E,$F7,$F9,$5A
	.dc.b	$7F,$DC,$91,$E2,$F2,$78,$47,$E9,$FA,$38,$57,$EA,$2B,$F2,$FE,$47
	.dc.b	$2C,$BF,$CA,$11,$F9,$AF,$6F,$1E,$2F,$EC,$FC,$7B,$3F,$A1,$F8,$0F
	.dc.b	$0F,$57,$F2,$FE,$AF,$E8,$5C,$9F,$C5,$F9,$BF,$B4,$F3,$7C,$BE,$2F
	.dc.b	$E3,$91,$F9,$53,$F2,$B3,$B7,$FE,$00,$79,$AD,$BE,$76,$1F,$E9,$96
	.dc.b	$BF,$E8,$0F,$37,$FC,$76,$FF,$B1,$23,$FC,$8A,$FF,$A3,$9F,$F4,$27
	.dc.b	$17,$F7,$E6,$C6,$79,$BE,$2F,$8B,$FE,$A1,$C5,$F0,$3E,$19,$CF,$C1
	.dc.b	$F9,$E7,$F5,$98,$7D,$DC,$78,$5E,$F8,$7F,$4F,$CF,$F3,$F0,$E1,$86
	.dc.b	$67,$F0,$13,$C3,$0F,$C3,$3E,$19,$F8,$1D,$3F,$CB,$74,$F6,$35,$FF
	.dc.b	$70,$57,$E1,$AF,$FC,$FB,$1A,$6B,$A9,$FC,$0D,$B2,$F9,$51,$87,$E3
	.dc.b	$74,$FF,$35,$CB,$FD,$65,$8F,$FA,$6B,$1F,$3E,$BF,$16,$31,$8F,$E6
	.dc.b	$70,$FF,$64,$69,$F9,$18,$F8,$1A,$7D,$DA,$DC,$7E,$7F,$1E,$B9,$79
	.dc.b	$23,$A7,$9F,$FC,$FF,$E2,$FE,$8D,$FE,$4A,$7F,$9C,$47,$D0,$9F,$E8
	.dc.b	$9F,$23,$A6,$E3,$F9,$BA,$FE,$7D,$F2,$99,$DB,$F9,$B3,$F6,$3A,$BF
	.dc.b	$A6,$7D,$03,$8B,$F1,$D1,$8F,$87,$B7,$FC,$F7,$E9,$75,$B7,$F4,$AC
	.dc.b	$BF,$A5,$42,$B7,$5F,$D0,$3D,$0D,$E1,$FD,$65,$D7,$4F,$E7,$75,$FE
	.dc.b	$FC,$D7,$F4,$5E,$6F,$F1,$77,$F0,$FF,$03,$FC,$47,$9F,$F4,$F1,$FA
	.dc.b	$7F,$1B,$19,$7F,$3A,$87,$91,$88,$FE,$89,$E8,$7E,$A7,$B7,$FC,$FD
	.dc.b	$5F,$99,$EA,$C5,$7F,$9F,$E3,$F2,$BC,$5F,$EA,$C7,$4D,$7F,$40,$FE
	.dc.b	$A9,$CD,$7F,$BE,$1F,$BD,$DB,$FA,$AE,$4F,$DB,$35,$FF,$7A,$BF,$EF
	.dc.b	$77,$E8,$1F,$A3,$E4,$FE,$28,$FE,$CF,$0F,$D4,$C7,$F7,$D8,$E7,$18
	.dc.b	$8E,$31,$F9,$D8,$C7,$8F,$E8,$EE,$FF,$0A,$61,$E8,$7F,$43,$E0,$66
	.dc.b	$BF,$DC,$73,$97,$A7,$7E,$5F,$DA,$BF,$13,$BB,$FC,$31,$8F,$F0,$D6
	.dc.b	$BF,$B2,$7F,$54,$E6,$79,$BE,$EE,$8F,$D2,$67,$EB,$8F,$83,$B7,$E9
	.dc.b	$23,$11,$F8,$B8,$FD,$04,$7F,$06,$1E,$67,$E4,$BB,$7F,$C4,$8E,$DF
	.dc.b	$F1,$37,$CA,$E7,$1F,$F0,$A3,$97,$FC,$2D,$E8,$E5,$FF,$0A,$BD,$FB
	.dc.b	$3F,$C5,$DE,$7F,$EC,$1B,$BF,$E7,$32,$FD,$5F,$2D,$7C,$9E,$07,$EA
	.dc.b	$F1,$FF,$11,$72,$71,$6D,$FE,$4C,$FF,$4B,$F6,$FF,$62,$7F,$D4,$1E
	.dc.b	$C7,$F9,$9A,$7F,$AD,$0F,$81,$FC,$D6,$3F,$DD,$AF,$FB,$4B,$FD,$F9
	.dc.b	$FE,$35,$3F,$E1,$4F,$73,$FD,$63,$C5,$FE,$E7,$8F,$F3,$01,$FE,$DB
	.dc.b	$AF,$F2,$F7,$57,$F9,$6D,$FE,$63,$3F,$DD,$7B,$BF,$CF,$75,$FD,$AE
	.dc.b	$31,$F1,$3F,$E6,$08,$FC,$9F,$4E,$9F,$F3,$29,$88,$FF,$2F,$BF,$CF
	.dc.b	$3D,$6F,$F3,$1E,$1F,$DF,$77,$CB,$FD,$D2,$7F,$C0,$38,$7F,$39,$7F
	.dc.b	$CC,$51,$E4,$EA,$FF,$33,$69,$E3,$7F,$CD,$18,$E6,$BF,$2F,$39,$7F
	.dc.b	$33,$1E,$CE,$7F,$F3,$47,$C9,$7F,$99,$F0,$FE,$9F,$F2,$5F,$E6,$39
	.dc.b	$FF,$36,$7B,$7F,$E5,$FC,$3F,$CB,$5F,$47,$FE,$63,$BF,$F3,$67,$D1
	.dc.b	$E6,$7F,$CB,$7F,$4F,$99,$7D,$3F,$F9,$C7,$97,$E9,$D7,$C5,$3F,$C2
	.dc.b	$9F,$07,$FE,$A7,$AF,$F3,$1F,$2F,$F8,$D7,$C1,$FD,$C8,$FE,$01,$FE
	.dc.b	$41,$D7,$FD,$B7,$DD,$FF,$1B,$5F,$F9,$DF,$A7,$FC,$64,$7F,$AF,$72
	.dc.b	$FC,$38,$F8,$5F,$E3,$6F,$F8,$AF,$AB,$FC,$67,$D7,$FD,$F2,$3F,$C3
	.dc.b	$FD,$5F,$7F,$A6,$F9,$1F,$E2,$4C,$7D,$1E,$DF,$F7,$0B,$FF,$4C,$1F
	.dc.b	$E3,$03,$FD,$28,$7F,$8A,$CF,$F4,$B5,$57,$F8,$3B,$FE,$40,$E8,$FA
	.dc.b	$33,$FD,$4F,$4F,$DD,$7B,$3F,$E2,$8B,$FF,$42,$F4,$59,$FF,$4A,$73
	.dc.b	$FF,$8A,$B9,$3F,$9F,$F1,$7F,$9E,$7E,$1B,$E0,$E1,$FC,$FF,$CD,$AF
	.dc.b	$FA,$06,$3E,$0E,$1F,$E8,$FE,$6F,$F1,$37,$17,$F9,$03,$8F,$E7,$F9
	.dc.b	$5A,$A6,$7F,$E2,$3A,$FE,$6D,$7F,$E0,$C7,$FC,$0E,$F2,$C7,$F9,$83
	.dc.b	$FC,$D5,$FF,$2C,$4F,$F8,$73,$C5,$FE,$75,$7D,$34,$FE,$97,$C9,$FD
	.dc.b	$15,$FE,$F6,$D7,$71,$C7,$97,$FC,$43,$1F,$D4,$BC,$CF,$F0,$BE,$BF
	.dc.b	$E1,$6C,$BF,$D9,$59,$FF,$2B,$BF,$D4,$61,$E2,$8F,$F1,$04,$FF,$85
	.dc.b	$9D,$5F,$F8,$35,$CC,$FF,$8C,$1E,$5E,$17,$FB,$4F,$1F,$F0,$9D,$FF
	.dc.b	$B4,$5F,$EB,$F1,$FE,$25,$73,$1F,$D1,$1C,$BF,$D4,$EB,$FB,$3C,$FF
	.dc.b	$86,$E7,$FB,$3F,$2F,$FB,$67,$E6,$3F,$C2,$57,$FE,$C4,$7F,$C3,$4F
	.dc.b	$FB,$27,$C9,$CE,$1F,$ED,$57,$31,$FE,$CA,$F1,$73,$3F,$D0,$DF,$E6
	.dc.b	$11,$FE,$CE,$F9,$BF,$F0,$7B,$FE,$D3,$7F,$9E,$3F,$D0,$33,$FE,$71
	.dc.b	$5F,$E5,$0F,$2F,$33,$FB,$49,$CC,$FD,$CE,$4F,$C6,$E9,$FC,$E3,$B3
	.dc.b	$F2,$6B,$CF,$FF,$53,$3F,$E5,$C1,$E5,$7E,$53,$CE,$FD,$EE,$3F,$E2
	.dc.b	$DF,$81,$C3,$FE,$C8,$F1,$7F,$D0,$CF,$F0,$2F,$8B,$FC,$A7,$3F,$E0
	.dc.b	$FA,$FF,$31,$97,$F8,$51,$FF,$3A,$BF,$E1,$0D,$FF,$C0,$BD,$7E,$1E
	.dc.b	$95,$7F,$97,$AB,$55,$FE,$6E,$7F,$26,$F0,$D3,$FC,$16,$FE,$69,$FC
	.dc.b	$E7,$0F,$FA,$95,$CB,$73,$FF,$0C,$CF,$F8,$46,$73,$23,$FB,$AC,$FE
	.dc.b	$42,$7F,$C2,$D9,$7F,$82,$E7,$F8,$A5,$FF,$83,$E7,$FC,$D6,$FF,$9B
	.dc.b	$BA,$7F,$24,$FF,$99,$DF,$C5,$3F,$EF,$9E,$BF,$E2,$93,$FE,$03,$8F
	.dc.b	$C8,$BF,$E7,$4C,$7F,$C2,$0D,$C7,$FC,$28,$FF,$87,$23,$38,$7F,$9D
	.dc.b	$DF,$D0,$C7,$E6,$5C,$C7,$FB,$4D,$FD,$77,$4F,$F9,$D6,$3D,$DF,$35
	.dc.b	$FE,$09,$7F,$C2,$81,$FF,$07,$57,$E4,$EF,$F9,$04,$FF,$A7,$B9,$F2
	.dc.b	$F0,$DF,$93,$FE,$1D,$7F,$26,$F4,$FA,$7F,$F0,$55,$F6,$84,$7E,$81
	.dc.b	$F0,$FD,$4E,$9B,$A7,$E6,$9F,$F6,$F6,$B5,$FF,$50,$47,$E6,$0F,$F5
	.dc.b	$9F,$C8,$FE,$53,$3F,$EA,$7F,$1C,$F8,$27,$B5,$FE,$FF,$D3,$FC,$2B
	.dc.b	$CF,$FD,$7B,$2F,$F0,$91,$FD,$4E,$3F,$C3,$3F,$1D,$D7,$3F,$F6,$37
	.dc.b	$B4,$BE,$74,$29,$FD,$69,$FF,$41,$48,$69,$FE,$E7,$9F,$F3,$F7,$5F
	.dc.b	$F8,$AE,$7F,$61,$E2,$FF,$89,$B2,$FE,$CD,$87,$F7,$2F,$2F,$FB,$2E
	.dc.b	$1F,$E2,$B8,$FE,$EB,$7F,$B7,$1E,$DF,$C1,$E0,$FF,$35,$FC,$87,$9F
	.dc.b	$1F,$66,$1E,$73,$D1,$EC,$7C,$5E,$6F,$E0,$78,$79,$67,$A6,$2D,$F9
	.dc.b	$47,$BF,$E4,$3F,$81,$E6,$F8,$BC,$CA,$73,$A7,$FB,$97,$6F,$07,$C2
	.dc.b	$7F,$A4,$BC,$CF,$9C,$FF,$A5,$BE,$67,$E8,$78,$5F,$E0,$AF,$0D,$18
	.dc.b	$D3,$F3,$B9,$FF,$84,$EF,$F9,$CE,$3E,$B8,$FF,$6D,$7C,$E7,$BB,$1F
	.dc.b	$F4,$5E,$28,$C6,$1F,$EE,$EF,$7F,$E1,$DF,$D9,$F5,$3F,$97,$BF,$E1
	.dc.b	$F3,$1E,$5B,$07,$D4,$F0,$9F,$C0,$78,$19,$8C,$47,$DF,$E9,$FA,$BC
	.dc.b	$BF,$CA,$FD,$7F,$8B,$C9,$F2,$3C,$AF,$C6,$4F,$E1,$47,$67,$D9,$DF
	.dc.b	$FD,$57,$EC,$7C,$C9,$FB,$D8,$57,$F9,$B7,$7D,$7F,$D9,$BB,$39,$FF
	.dc.b	$35,$87,$F2,$4E,$5F,$C0,$3F,$EA,$EE,$1F,$67,$3F,$D2,$C3,$C5,$C5
	.dc.b	$8E,$1F,$F3,$91,$F9,$DE,$7F,$F5,$43,$83,$F4,$18,$7F,$16,$C7,$E9
	.dc.b	$3F,$DE,$F9,$BF,$1F,$5D,$34,$FA,$57,$C4,$47,$FB,$4B,$9B,$E6,$78
	.dc.b	$1F,$C1,$B1,$FF,$21,$BF,$A9,$BF,$A3,$7C,$4F,$B1,$8F,$D1,$70,$39
	.dc.b	$FF,$DF,$5E,$17,$FA,$57,$5F,$A4,$FF,$03,$9F,$DC,$78,$5F,$AC,$7F
	.dc.b	$C6,$3E,$0F,$F9,$73,$A7,$FA,$14,$7E,$F6,$BF,$8C,$79,$DF,$9E,$BF
	.dc.b	$D9,$E9,$FC,$37,$EB,$FF,$03,$7F,$C1,$5B,$87,$78,$FE,$0A,$7F,$10
	.dc.b	$F8,$3F,$EE,$3D,$5F,$C0,$74,$FC,$11,$FC,$1F,$C2,$ED,$5F,$E6,$4F
	.dc.b	$13,$53,$FE,$6B,$C3,$CB,$3F,$D5,$B8,$3F,$CE,$3E,$17,$EC,$BC,$CF
	.dc.b	$D8,$D6,$9F,$C8,$71,$FE,$16,$FE,$D6,$1F,$B3,$6B,$E2,$8F,$D8,$C7
	.dc.b	$DF,$C3,$F6,$9D,$3F,$8C,$7F,$67,$CB,$FB,$1B,$FD,$9D,$7F,$0D,$79
	.dc.b	$74,$7F,$10,$AF,$D9,$B5,$FD,$94,$FD,$8F,$A0,$FD,$8D,$FF,$A0,$9F
	.dc.b	$F7,$47,$85,$F8,$4E,$BF,$BD,$5F,$B5,$1C,$9F,$CB,$79,$FF,$97,$E6
	.dc.b	$EF,$FE,$81,$FA,$EF,$D2,$E6,$FF,$0F,$C7,$F1,$6F,$57,$F5,$7C,$7F
	.dc.b	$29,$FE,$55,$CD,$9C,$3F,$5B,$C7,$F1,$F4,$FF,$63,$72,$7E,$AD,$FE
	.dc.b	$80,$FE,$AE,$3E,$3C,$7F,$B2,$FE,$7B,$ED,$EF,$FA,$B8,$C5,$FF,$7E
	.dc.b	$E2,$FE,$31,$87,$EA,$67,$F9,$A3,$FA,$97,$F4,$D7,$FA,$9E,$4F,$E7
	.dc.b	$2F,$E9,$E3,$F3,$FE,$EF,$E9,$DF,$F7,$7B,$FA,$7E,$3F,$E2,$DD,$0E
	.dc.b	$6F,$93,$97,$E2,$FC,$1F,$EB,$5C,$9F,$98,$E5,$7D,$07,$F4,$5E,$FF
	.dc.b	$F1,$11,$FE,$FA,$F0,$3F,$B5,$47,$F6,$58,$C9,$FE,$27,$9E,$3A,$FE
	.dc.b	$C8,$FD,$92,$FD,$BF,$FC,$21,$97,$F5,$18,$7E,$B9,$FA,$6B,$FF,$0A
	.dc.b	$1F,$E3,$1F,$0B,$BE,$1F,$EF,$68,$EE,$7F,$7B,$BD,$A9,$C4,$FF,$92
	.dc.b	$A3,$77,$E1,$8F,$F3,$E4,$7F,$47,$E9,$F0,$C7,$9C,$E2,$7C,$0A,$FF
	.dc.b	$00,$D2,$FE,$25,$BA,$FE,$24,$E6,$FF,$C0,$53,$FE,$49,$9F,$F0,$4E
	.dc.b	$1E,$19,$FF,$16,$9F,$DC,$5C,$4F,$B2,$FE,$27,$1F,$F4,$AD,$7F,$C3
	.dc.b	$1E,$27,$F0,$A9,$F0,$E3,$F4,$F0,$F0,$63,$FE,$B8,$C7,$DF,$AF,$C4
	.dc.b	$D3,$FC,$2D,$8F,$1A,$F2,$D7,$A5,$FE,$D1,$DA,$FB,$7F,$E5,$11,$F1
	.dc.b	$A5,$F2,$DF,$CA,$E7,$33,$FB,$1F,$A4,$FF,$33,$46,$35,$FF,$0A,$EC
	.dc.b	$FB,$77,$EB,$9F,$8B,$F9,$6F,$F0,$CC,$3E,$67,$F9,$2F,$07,$9F,$87
	.dc.b	$D7,$B7,$F2,$EC,$FB,$BF,$E5,$4F,$A5,$EF,$B7,$B6,$7F,$C0,$13,$FE
	.dc.b	$58,$F9,$DF,$F3,$C6,$5E,$0A,$FA,$7F,$35,$FE,$28,$FB,$7F,$04,$7D
	.dc.b	$FC,$38,$E1,$F8,$CD,$7F,$85,$C6,$23,$F3,$B7,$F0,$6B,$F2,$AF,$F1
	.dc.b	$0B,$ED,$1F,$E8,$78,$ED,$B7,$BE,$3F,$C2,$87,$F9,$17,$17,$DC,$7D
	.dc.b	$D5,$FB,$FF,$F4,$8F,$B9,$F0,$65,$FB,$0C,$7F,$AD,$46,$23,$FD,$29
	.dc.b	$AF,$BB,$2F,$BD,$9F,$2C,$FF,$35,$7B,$B6,$1F,$C9,$B8,$FE,$CF,$A3
	.dc.b	$FD,$61,$B7,$E6,$3F,$B4,$38,$70,$FE,$03,$E7,$B8,$72,$FF,$92,$5F
	.dc.b	$BF,$19,$9F,$F2,$AB,$F8,$37,$FB,$CB,$FE,$35,$AF,$F2,$CF,$87,$F9
	.dc.b	$89,$FE,$C6,$7F,$94,$FA,$71,$3D,$23,$FC,$57,$F2,$1B,$BE,$27,$F7
	.dc.b	$78,$FF,$2D,$5B,$5F,$A7,$A7,$F7,$6D,$7E,$13,$FC,$16,$FF,$33,$DF
	.dc.b	$FE,$91,$8F,$DE,$6B,$FE,$63,$FC,$FB,$15,$E7,$CB,$D6,$7F,$17,$AF
	.dc.b	$F4,$C4,$72,$7F,$D4,$FD,$DE,$B7,$FD,$39,$EF,$7B,$71,$FF,$03,$DF
	.dc.b	$9C,$FF,$37,$DF,$F9,$5F,$CD,$FA,$51,$FE,$73,$F9,$7E,$33,$FC,$57
	.dc.b	$0E,$B3,$FE,$5D,$8F,$82,$F8,$E3,$FD,$5D,$1F,$E6,$19,$FF,$56,$78
	.dc.b	$5F,$37,$B7,$C7,$7C,$4F,$BD,$C3,$F4,$71,$FA,$71,$DF,$1F,$F5,$C3
	.dc.b	$5F,$E7,$BA,$7E,$6B,$3F,$5C,$F8,$67,$F9,$9B,$F1,$CF,$E8,$C5,$3E
	.dc.b	$B7,$17,$F8,$99,$F8,$07,$F8,$F0,$D1,$F6,$8F,$54,$7F,$B4,$B2,$FD
	.dc.b	$4E,$7E,$A9,$FE,$7F,$AF,$F9,$DD,$FE,$9D,$C7,$F8,$1E,$AF,$E5,$1C
	.dc.b	$5F,$A1,$EC,$FF,$02,$BF,$93,$AF,$E5,$1E,$1F,$F2,$DE,$1F,$E1,$FA
	.dc.b	$7F,$49,$9F,$E4,$71,$FD,$69,$FF,$07,$1F,$E4,$BE,$0F,$94,$FF,$B1
	.dc.b	$17,$DC,$7F,$44,$AF,$F6,$3F,$DC,$FF,$B2,$DF,$EA,$FE,$27,$F8,$5B
	.dc.b	$7F,$F5,$D6,$1F,$AF,$E7,$FE,$0B,$8F,$D2,$8F,$99,$7F,$DC,$38,$3D
	.dc.b	$D0,$E7,$FC,$C3,$FA,$F3,$F9,$79,$FE,$D1,$EA,$B8,$7E,$D6,$3E,$73
	.dc.b	$D2,$FF,$B6,$47,$E6,$5F,$EA,$71,$FE,$E3,$8F,$B2,$7E,$D0,$F9,$CE
	.dc.b	$5F,$F6,$F7,$13,$83,$E6,$47,$FB,$2F,$C4,$5E,$6F,$FB,$DF,$85,$8B
	.dc.b	$FE,$77,$3F,$12,$7F,$DD,$5E,$3F,$F8,$2E,$3F,$C1,$7C,$5F,$45,$FF
	.dc.b	$01,$1F,$E0,$07,$0F,$F9,$83,$1F,$F5,$C6,$DF,$EC,$1F,$A7,$FF,$71
	.dc.b	$7B,$9F,$EC,$F9,$FF,$41,$79,$1F,$E5,$D7,$FB,$B6,$5F,$E9,$0F,$47
	.dc.b	$FC,$BD,$EB,$FF,$BF,$FD,$FF,$F2,$ED,$7F,$A7,$5F,$F1,$E7,$C8,$FF
	.dc.b	$B9,$FC,$2F,$F7,$46,$1C,$70,$7A,$BF,$EC,$5A,$D6,$1F,$EC,$BC,$BF
	.dc.b	$CC,$2F,$FA,$7F,$1F,$F3,$0B,$7D,$1F,$F3,$0B,$FE,$5B,$FA,$EF,$F3
	.dc.b	$0B,$FD,$21,$FF,$30,$BF,$E9,$CF,$17,$BF,$07,$FA,$8A,$3F,$CC,$51
	.dc.b	$FE,$A5,$7F,$DF,$97,$FE,$F4,$F6,$FF,$DB,$11,$8C,$3F,$C8,$44,$1F
	.dc.b	$07,$F1,$F6,$C4,$7F,$91,$F1,$FF,$30,$6F,$FE,$AE,$8E,$3E,$A7,$F9
	.dc.b	$6E,$3F,$CF,$DC,$1F,$EB,$DE,$1F,$83,$97,$FA,$6A,$3F,$D7,$1C,$18
	.dc.b	$CB,$FD,$F8,$DC,$7F,$CB,$D5,$FD,$0E,$BF,$CB,$CF,$FB,$15,$FF,$2F
	.dc.b	$3F,$E7,$FF,$B8,$FF,$2F,$3F,$EF,$F7,$FC,$BC,$FF,$A2,$67,$FC,$BA
	.dc.b	$FF,$70,$F7,$59,$FB,$73,$FD,$7F,$C0,$FD,$6C,$DE,$0F,$F9,$33,$6F
	.dc.b	$F3,$CD,$7E,$8D,$B8,$FF,$B0,$7A,$3F,$CB,$F0,$DB,$D9,$5F,$CE,$DF
	.dc.b	$75,$EB,$DA,$F8,$78,$7F,$B0,$2F,$FD,$C2,$FF,$45,$C3,$FC,$DB,$0F
	.dc.b	$81,$DC,$87,$D8,$D9,$D4,$7F,$B7,$BE,$C7,$F5,$F1,$F8,$3C,$FF,$BD
	.dc.b	$7B,$BB,$F7,$79,$DB,$87,$6F,$F0,$53,$F8,$D7,$07,$C3,$BF,$04,$7E
	.dc.b	$B3,$97,$ED,$F9,$F8,$8F,$E4,$5C,$3F,$01,$FD,$8F,$07,$BA,$7F,$4B
	.dc.b	$A7,$9A,$7F,$D6,$DB,$7F,$29,$D3,$C7,$AF,$F2,$B9,$FC,$4F,$37,$E1
	.dc.b	$FB,$BF,$24,$FE,$DF,$5F,$85,$C9,$F8,$B8,$FC,$85,$7F,$AC,$39,$BF
	.dc.b	$35,$9F,$D6,$8F,$F7,$14,$7E,$02,$3E,$B4,$7C,$BB,$FD,$CD,$7F,$B5
	.dc.b	$A1,$7F,$97,$AF,$AB,$5F,$EB,$4B,$FD,$3C,$62,$BF,$D6,$B1,$FB,$89
	.dc.b	$FF,$14,$7D,$97,$EE,$27,$E3,$3F,$B8,$7F,$DA,$D9,$7E,$E3,$3F,$83
	.dc.b	$97,$ED,$27,$EF,$7D,$4F,$ED,$F2,$F9,$CF,$EC,$B2,$FF,$5C,$75,$7E
	.dc.b	$D7,$E0,$3F,$DA,$1A,$FB,$BA,$BE,$1E,$2F,$82,$FF,$67,$E5,$F0,$DF
	.dc.b	$67,$27,$EC,$70,$FF,$67,$57,$C0,$7F,$63,$C1,$F7,$6B,$F6,$33,$FE
	.dc.b	$35,$7F,$63,$1F,$E4,$5F,$B3,$FD,$8C,$FE,$19,$FD,$8B,$FE,$83,$D3
	.dc.b	$F6,$2F,$E1,$FC,$9F,$A5,$D1,$FA,$FE,$CF,$D7,$74,$FF,$3F,$7F,$43
	.dc.b	$A7,$F0,$57,$F5,$17,$F5,$CF,$ED,$8B,$5F,$C6,$67,$FE,$57,$FB,$4F
	.dc.b	$D0,$F2,$7E,$7B,$3F,$CE,$E1,$FA,$9E,$9F,$F2,$E2,$F2,$FF,$A2,$1F
	.dc.b	$ED,$BC,$7F,$7B,$C7,$74,$FE,$33,$3F,$9F,$8F,$07,$5F,$F1,$9B,$FC
	.dc.b	$EF,$77,$EB,$3E,$CF,$F3,$91,$E7,$3F,$4D,$F6,$7F,$E0,$7F,$B3,$FD
	.dc.b	$8C,$3D,$CF,$C0,$2F,$B2,$FE,$AF,$AF,$F2,$E8,$F3,$79,$5F,$A1,$6F
	.dc.b	$77,$F5,$67,$E2,$6A,$F8,$EF,$EA,$F1,$FD,$71,$FE,$AE,$F8,$F7,$FE
	.dc.b	$AD,$E6,$7F,$55,$8C,$61,$FD,$51,$FF,$64,$DF,$3E,$AF,$CF,$F0,$FF
	.dc.b	$37,$C7,$F9,$FE,$7F,$42,$FF,$D3,$7F,$6B,$F9,$F7,$F8,$3B,$F9,$FE
	.dc.b	$8F,$F4,$57,$0F,$E7,$E7,$E2,$1F,$C3,$1F,$EA,$CF,$EE,$71,$FE,$9F
	.dc.b	$E5,$7D,$47,$FA,$8E,$1F,$0F,$6F,$EA,$7E,$07,$F0,$0D,$7F,$B8,$70
	.dc.b	$7F,$62,$EC,$FE,$75,$C1,$F9,$EC,$FE,$06,$5F,$9E,$C7,$FA,$F6,$3F
	.dc.b	$6A,$7E,$B6,$3F,$13,$2F,$DE,$47,$F8,$1B,$21,$23,$E0,$47,$F2,$1F
	.dc.b	$B6,$FD,$0B,$FC,$D5,$FD,$0E,$DF,$EA,$9F,$3B,$FB,$47,$67,$F3,$07
	.dc.b	$F8,$95,$7F,$08,$D3,$FB,$6F,$87,$F9,$2B,$FE,$A7,$3F,$9F,$D7,$FA
	.dc.b	$9F,$7F,$F4,$EE,$0F,$EA,$DA,$7F,$91,$9F,$EB,$1E,$87,$F5,$7E,$DF
	.dc.b	$CB,$E3,$FD,$47,$1F,$C0,$79,$9F,$AA,$7E,$EE,$1F,$E8,$FE,$8F,$EE
	.dc.b	$58,$7F,$5A,$EE,$FE,$8B,$F7,$1F,$D3,$5F,$E8,$FF,$71,$FA,$CC,$3E
	.dc.b	$C9,$FD,$4F,$CF,$FE,$98,$E0,$71,$7E,$C1,$FF,$67,$9F,$EC,$BE,$0F
	.dc.b	$73,$FD,$13,$7F,$5B,$FD,$1B,$7F,$4B,$FD,$28,$FA,$BA,$FE,$B9,$ED
	.dc.b	$9F,$EB,$9F,$F0,$17,$37,$EA,$BC,$7F,$D4,$F8,$DF,$E3,$0C,$BF,$C2
	.dc.b	$1D,$FF,$BE,$E9,$FE,$79,$A7,$F7,$FE,$7F,$D3,$F2,$FF,$86,$D1,$FD
	.dc.b	$1F,$6F,$E0,$35,$FD,$17,$0F,$E8,$9C,$9F,$E1,$7A,$FD,$2E,$BF,$8A
	.dc.b	$7F,$4B,$C7,$FD,$BB,$BF,$F4,$93,$FD,$E5,$FD,$23,$FD,$A4,$71,$FE
	.dc.b	$95,$FF,$0E,$63,$FC,$13,$E7,$FF,$A1,$6F,$EB,$F2,$7F,$46,$DC,$BF
	.dc.b	$67,$BF,$F8,$67,$DD,$FE,$7E,$DC,$3F,$32,$7F,$8C,$31,$FE,$05,$87
	.dc.b	$E8,$CF,$E7,$EF,$F6,$CA,$FE,$7F,$C1,$FE,$5C,$7F,$9F,$C7,$9F,$1F
	.dc.b	$CF,$A7,$F3,$EF,$AF,$D1,$FC,$FB,$FE,$13,$EF,$F5,$67,$FD,$67,$B7
	.dc.b	$FC,$E5,$E2,$FC,$1A,$FE,$63,$DD,$FD,$09,$FE,$37,$9F,$E7,$5F,$E3
	.dc.b	$BD,$DF,$9D,$7F,$96,$57,$FA,$9B,$9B,$F9,$AD,$FE,$76,$FF,$CE,$67
	.dc.b	$E7,$6B,$F7,$B7,$F9,$D7,$F2,$DF,$71,$F9,$D7,$F3,$1C,$3F,$9D,$7F
	.dc.b	$C2,$A7,$F5,$83,$FC,$43,$AF,$F8,$B3,$6F,$F3,$DE,$7F,$D1,$EB,$FC
	.dc.b	$4D,$5F,$99,$3C,$A7,$F8,$83,$9B,$FC,$48,$FF,$B1,$F9,$BF,$C6,$38
	.dc.b	$FE,$67,$3F,$85,$E8,$FF,$45,$AF,$D5,$0F,$53,$FA,$F7,$83,$FD,$23
	.dc.b	$AB,$FB,$06,$7F,$EF,$FA,$F8,$C3,$E0,$BF,$D5,$FC,$2C,$B3,$E6,$7F
	.dc.b	$60,$AF,$E8,$06,$78,$BF,$D8,$2F,$F8,$FB,$C1,$FF,$12,$B9,$BF,$F5
	.dc.b	$63,$C2,$FF,$DC,$33,$FE,$24,$CF,$BC,$FC,$F3,$C0,$F8,$A7,$07,$F5
	.dc.b	$E8,$ED,$BF,$FA,$13,$87,$CF,$F0,$1F,$4E,$7D,$07,$F9,$7E,$3D,$7E
	.dc.b	$F6,$6F,$FC,$A7,$C1,$FE,$E6,$F1,$BF,$C0,$DA,$2B,$F8,$9E,$1F,$D6
	.dc.b	$A3,$FD,$A3,$F0,$58,$C3,$FC,$B7,$C5,$FE,$EE,$F3,$BF,$CD,$BC,$1F
	.dc.b	$ED,$6F,$83,$FF,$06,$46,$30,$FF,$35,$BF,$E7,$2B,$F8,$DC,$1F,$E7
	.dc.b	$43,$FC,$EE,$7F,$9D,$B8,$7E,$57,$27,$F9,$A3,$8F,$1D,$5F,$EC,$7B
	.dc.b	$FF,$7C,$ED,$FE,$79,$E8,$FF,$65,$61,$FE,$84,$3F,$A6,$F1,$7F,$65
	.dc.b	$9F,$F1,$4E,$FF,$01,$F2,$9F,$CC,$AF,$FC,$E8,$FF,$87,$4F,$F2,$C9
	.dc.b	$FA,$53,$FC,$BC,$FF,$BD,$9F,$A7,$5E,$27,$FC,$55,$CB,$FE,$B6,$BF
	.dc.b	$F3,$FF,$A1,$F1,$E7,$80,$F3,$BF,$D2,$1B,$FC,$38,$FF,$15,$60,$DF
	.dc.b	$E1,$47,$F9,$AA,$1B,$FF,$A2,$B8,$3F,$09,$3F,$D5,$4F,$CB,$3F,$92
	.dc.b	$7F,$D3,$9B,$FF,$BB,$63,$F1,$B1,$F9,$58,$FF,$49,$FB,$7F,$9B,$BF
	.dc.b	$F3,$AF,$1E,$AB,$E6,$1F,$E7,$D7,$BB,$83,$F4,$9B,$FF,$90,$A7,$FB
	.dc.b	$DF,$A5,$F9,$29,$F8,$67,$F8,$E5,$FC,$8B,$83,$F2,$A7,$C3,$3F,$D4
	.dc.b	$51,$FE,$AF,$7F,$31,$88,$FC,$A3,$FE,$62,$E8,$FD,$AF,$37,$CD,$8F
	.dc.b	$DF,$1C,$B6,$FF,$5F,$D7,$E4,$ED,$18,$8F,$C9,$CF,$F8,$AB,$CA,$F0
	.dc.b	$EF,$F9,$79,$FE,$0F,$3F,$85,$3F,$D8,$DD,$BF,$EF,$2C,$3F,$09,$0D
	.dc.b	$BF,$D7,$B8,$7F,$53,$9F,$F4,$FE,$BF,$93,$E4,$FF,$7C,$F1,$7E,$4E
	.dc.b	$7D,$D8,$7A,$0F,$4B,$E6,$7C,$87,$FA,$D9,$FE,$17,$87,$EC,$23,$F8
	.dc.b	$1D,$FE,$46,$7F,$C5,$75,$F9,$09,$FF,$43,$F2,$77,$D7,$FD,$CF,$97
	.dc.b	$EE,$2F,$F1,$D5,$ED,$8F,$F1,$D3,$FC,$87,$4F,$C1,$38,$EC,$FE,$47
	.dc.b	$1F,$62,$37,$D1,$F8,$9D,$7F,$9D,$CF,$E0,$70,$FF,$2E,$70,$F0,$EA
	.dc.b	$FE,$21,$87,$BC,$FE,$4D,$D7,$FE,$99,$BF,$BD,$3F,$A7,$CF,$BF,$1F
	.dc.b	$EA,$7A,$7F,$AF,$65,$F1,$71,$F1,$73,$7F,$36,$EB,$FD,$DC,$63,$2F
	.dc.b	$F8,$03,$C1,$FE,$F0,$E3,$FF,$77,$1F,$7A,$7F,$CD,$3C,$DE,$CC,$7E
	.dc.b	$1C,$7B,$5F,$E6,$5C,$3C,$23,$F9,$79,$FE,$A9,$E3,$FF,$49,$BF,$EA
	.dc.b	$9E,$0F,$F1,$33,$FC,$36,$BC,$D8,$71,$8F,$DD,$75,$FF,$1B,$7E,$C7
	.dc.b	$1F,$95,$F8,$9D,$7E,$F7,$EC,$1F,$E7,$EA,$FB,$19,$FF,$1D,$8F,$F1
	.dc.b	$E1,$F5,$35,$F4,$5F,$3E,$9F,$AF,$D9,$F4,$1F,$A5,$5F,$B1,$C0,$FE
	.dc.b	$C4,$FD,$8E,$9F,$6B,$0F,$A7,$1E,$D3,$FD,$91,$CD,$F5,$1C,$1F,$D3
	.dc.b	$B6,$FA,$BC,$1E,$B8,$EB,$5F,$57,$7F,$87,$1F,$ED,$06,$BF,$CB,$8F
	.dc.b	$E8,$67,$FA,$73,$8B,$EA,$6D,$CE,$3E,$A7,$DC,$7C,$17,$F1,$7D,$3F
	.dc.b	$4A,$FF,$84,$65,$EC,$9F,$6B,$FC,$B9,$FF,$4A,$9F,$E6,$13,$FB,$2C
	.dc.b	$ED,$70,$FF,$37,$9E,$99,$FF,$B5,$67,$E3,$F2,$FF,$04,$7F,$45,$BF
	.dc.b	$F0,$0C,$5B,$F8,$7A,$7E,$A6,$FF,$C6,$27,$37,$F4,$27,$3C,$FF,$DB
	.dc.b	$B9,$BF,$99,$67,$F9,$7F,$53,$8F,$47,$E7,$34,$47,$4C,$F1,$BF,$F3
	.dc.b	$1E,$2F,$CF,$72,$E7,$6F,$CD,$E1,$F7,$76,$FB,$A7,$F5,$D3,$E0,$70
	.dc.b	$FC,$57,$FC,$3D,$87,$F9,$2F,$A7,$EE,$ED,$FE,$D0,$9E,$E7,$F9,$BF
	.dc.b	$8B,$E1,$61,$D6,$3F,$8B,$71,$7F,$55,$D7,$F2,$9C,$3D,$69,$D3,$EC
	.dc.b	$D7,$F4,$AF,$9F,$8B,$F6,$DC,$BE,$BC,$BF,$2B,$7F,$03,$1F,$97,$87
	.dc.b	$B3,$6F,$EA,$1F,$73,$F8,$5B,$FE,$E9,$3F,$B0,$C3,$D4,$D8,$F4,$3C
	.dc.b	$5F,$F4,$46,$FF,$6F,$97,$E6,$63,$F7,$78,$3F,$B2,$71,$FE,$DE,$F7
	.dc.b	$5F,$D4,$DF,$D4,$FE,$CD,$F5,$3E,$95,$FB,$1B,$F6,$67,$DF,$BF,$2C
	.dc.b	$70,$7F,$2D,$C7,$8F,$53,$1E,$38,$70,$BF,$C5,$E7,$DE,$D9,$FE,$0B
	.dc.b	$F2,$8F,$F4,$04,$E6,$3C,$53,$FE,$A3,$8C,$63,$D7,$4F,$26,$8D,$BF
	.dc.b	$25,$BF,$E9,$F1,$F5,$56,$BF,$28,$CE,$5E,$6D,$9F,$B2,$3F,$66,$7E
	.dc.b	$9C,$FC,$0D,$CB,$ED,$C3,$CF,$BF,$B7,$0E,$17,$B9,$D6,$9F,$26,$73
	.dc.b	$F8,$0F,$82,$78,$E7,$E0,$CF,$BE,$99,$C3,$87,$27,$C7,$E5,$E7,$A7
	.dc.b	$78,$F9,$5D,$D8,$C7,$D3,$1F,$53,$8B,$95,$6F,$8D,$96,$2D,$FA,$1B
	.dc.b	$E7,$9F,$93,$BF,$1B,$7E,$F9,$D6,$B5,$39,$FD,$03,$17,$C3,$F5,$0E
	.dc.b	$95,$AF,$A9,$ED,$FA,$E7,$FA,$37,$D5,$7E,$32,$FC,$B7,$D1,$39,$87
	.dc.b	$D0,$63,$C8,$FF,$21,$FE,$39,$C9,$FF,$22,$9E,$93,$D1,$F8,$EB,$F3
	.dc.b	$AF,$6F,$8B,$F8,$4E,$4F,$F4,$E7,$02,$B5,$F9,$5F,$C7,$C3,$8F,$E0
	.dc.b	$DC,$EB,$5E,$0F,$C5,$FD,$BD,$F8,$FB,$F2,$A7,$E5,$13,$5E,$D5,$A9
	.dc.b	$CF,$E5,$DF,$4E,$31,$5C,$EF,$96,$7F,$53,$F1,$C7,$3B,$F5,$C7,$C1
	.dc.b	$BF,$31,$7D,$DF,$E0,$2F,$75,$A9,$7C,$F6,$7C,$CF,$D2,$7A,$5F,$9F
	.dc.b	$F1,$D8,$70,$CB,$AD,$73,$CF,$C1,$0F,$5F,$F0,$FF,$A5,$BF,$7B,$FA
	.dc.b	$DE,$6B,$F6,$77,$E4,$9F,$92,$AD,$78,$9F,$D1,$25,$DB,$F9,$6F,$C4
	.dc.b	$BF,$73,$C1,$E3,$BB,$B7,$F0,$47,$8E,$4F,$D0,$BC,$3A,$78,$3C,$3F
	.dc.b	$CC,$FD,$93,$F1,$57,$EC,$CF,$A5,$7E,$BA,$FC,$8B,$F4,$B7,$BA,$D4
	.dc.b	$BF,$80,$7F,$53,$F9,$F7,$8B,$7F,$DC,$DD,$F6,$FF,$A8,$BC,$72,$FC
	.dc.b	$0C,$F4,$CF,$C3,$0F,$D5,$FF,$5B,$FB,$2B,$D1,$7C,$C3,$F4,$E7,$98
	.dc.b	$FD,$D3,$F7,$37,$BA,$D4,$BD,$7F,$DE,$FE,$5F,$E3,$F1,$17,$4E,$18
	.dc.b	$F4,$DB,$9E,$7E,$08,$7A,$3F,$C0,$3F,$38,$FE,$45,$F2,$8F,$A4,$7E
	.dc.b	$00,$FD,$4B,$F8,$5D,$6A,$73,$F4,$BF,$C3,$BF,$7E,$F1,$4C,$F8,$66
	.dc.b	$3C,$F1,$ED,$18,$C1,$F8,$F7,$F1,$9F,$B5,$7E,$FE,$FC,$87,$F5,$5F
	.dc.b	$B3,$FE,$7B,$E0,$FF,$92,$D6,$A7,$3E,$8B,$F2,$FC,$FE,$8B,$F5,$6B
	.dc.b	$9F,$37,$E7,$8F,$EA,$AF,$1F,$7D,$97,$E7,$BE,$03,$FA,$44,$E6,$3F
	.dc.b	$3D,$5F,$CD,$3E,$C3,$D7,$F8,$83,$F5,$97,$E3,$4F,$B7,$FB,$D9,$7F
	.dc.b	$51,$BD,$D6,$A7,$EF,$4E,$7D,$4F,$C5,$57,$C6,$C7,$CC,$7F,$5C,$78
	.dc.b	$E1,$E5,$9F,$EA,$91,$8F,$9F,$FE,$81,$5F,$B1,$F8,$8F,$EA,$1F,$A6
	.dc.b	$BE,$3F,$F9,$BE,$5F,$D8,$EB,$5F,$36,$AD,$7E,$AD,$FD,$42,$7F,$74
	.dc.b	$78,$BD,$0F,$F3,$1F,$E0,$9F,$9F,$8E,$34,$FC,$93,$C3,$EC,$FF,$27
	.dc.b	$C7,$A7,$E1,$3F,$D5,$FE,$61,$F9,$8B,$EE,$5F,$BD,$3E,$C5,$F1,$FF
	.dc.b	$DF,$FE,$33,$FB,$36,$1C,$2F,$75,$AF,$95,$FE,$3B,$C7,$FE,$01,$AE
	.dc.b	$BF,$0B,$FB,$57,$87,$0F,$CA,$79,$67,$DF,$3F,$CF,$CF,$6D,$3C,$93
	.dc.b	$9F,$2B,$D1,$F2,$4F,$CF,$BF,$A6,$FC,$13,$F5,$57,$EE,$AF,$D9,$FC
	.dc.b	$F7,$BF,$9B,$FE,$1B,$5A,$FD,$23,$3F,$97,$F9,$6B,$A6,$DE,$57,$86
	.dc.b	$5D,$72,$F0,$E7,$E0,$C5,$F5,$DF,$E0,$1F,$90,$FF,$13,$FE,$F8,$FC
	.dc.b	$65,$FB,$A3,$F0,$57,$9F,$F7,$7A,$7F,$C4,$D8,$70,$BD,$D6,$BF,$70
	.dc.b	$FF,$0A,$4F,$F8,$CB,$F2,$DF,$E3,$CF,$6E,$1F,$C9,$BC,$72,$ED,$E5
	.dc.b	$78,$FE,$B1,$AF,$8E,$7C,$5F,$77,$FE,$17,$FC,$ED,$F8,$E3,$D0,$7E
	.dc.b	$26,$FB,$6F,$47,$CD,$BD,$AF,$F1,$36,$1C,$2F,$75,$AF,$7F,$38,$FF
	.dc.b	$90,$78,$3F,$D1,$95,$FE,$9D,$B7,$E2,$FF,$E9,$99,$C9,$FE,$BC,$8E
	.dc.b	$99,$F8,$79,$3E,$DC,$F4,$BE,$F1,$FC,$DE,$3F,$C7,$1F,$BF,$BF,$49
	.dc.b	$FE,$F7,$1F,$C3,$FF,$16,$FF,$2D,$FE,$3F,$A9,$CF,$C0,$A7,$FC,$BF
	.dc.b	$F9,$66,$7C,$3F,$E9,$58,$F3,$FA,$EB,$87,$0B,$FF,$32,$E5,$B3,$B7
	.dc.b	$83,$FE,$41,$F5,$5F,$AF,$BF,$32,$FF,$18,$FE,$7A,$FD,$DB,$F3,$1F
	.dc.b	$98,$FF,$93,$EF,$75,$A9,$FC,$E7,$E8,$19,$C3,$96,$5F,$C4,$2F,$AF
	.dc.b	$EF,$9F,$D7,$DE,$3F,$74,$CF,$B7,$C7,$F8,$F0,$F1,$4E,$72,$FE,$8F
	.dc.b	$8F,$5C,$FC,$1E,$63,$F5,$6C,$5E,$E3,$F0,$5E,$5F,$F9,$EF,$F8,$05
	.dc.b	$F9,$C3,$F6,$7F,$F1,$5F,$F0,$03,$F2,$1F,$E8,$9F,$A1,$FF,$9A,$EF
	.dc.b	$75,$A9,$FE,$91,$FA,$36,$70,$E5,$B7,$F2,$CB,$EB,$F5,$3F,$DE,$DE
	.dc.b	$3F,$7A,$CF,$B7,$1F,$F9,$BE,$73,$1E,$1F,$84,$FF,$6B,$7C,$E7,$FB
	.dc.b	$AF,$F1,$0F,$EF,$9E,$D7,$F8,$1F,$F4,$FF,$EE,$9F,$56,$FA,$17,$27
	.dc.b	$F3,$DD,$D8,$70,$FB,$16,$1C,$3F,$06,$DE,$7D,$78,$7F,$2B,$87,$FB
	.dc.b	$CF,$1E,$58,$FF,$82,$FF,$01,$FF,$4F,$DE,$F5,$F1,$7D,$6F,$FB,$E3
	.dc.b	$0E,$60,$F0,$E3,$FC,$6B,$4F,$0C,$BD,$7F,$6D,$6F,$2F,$36,$BE,$CF
	.dc.b	$DD,$F9,$CF,$2C,$BB,$F0,$3E,$9B,$3F,$7E,$FD,$5D,$FB,$8B,$F3,$97
	.dc.b	$BB,$FA,$5F,$8D,$FE,$33,$CB,$F1,$70,$E1,$7B,$AD,$7E,$65,$D3,$17
	.dc.b	$EF,$FF,$03,$6F,$66,$1C,$1F,$F5,$F7,$37,$BD,$FF,$2D,$D7,$8B,$4F
	.dc.b	$0F,$A1,$F9,$6D,$3C,$13,$9C,$FF,$71,$DD,$F0,$B0,$E3,$AF,$8F,$0F
	.dc.b	$F1,$1F,$E7,$3E,$38,$C6,$FF,$CD,$B2,$E7,$18,$FC,$5B,$E3,$FA,$7F
	.dc.b	$EC,$3F,$D9,$9F,$BA,$3F,$5F,$7E,$B6,$FD,$31,$F9,$77,$0F,$A1,$61
	.dc.b	$C2,$F7,$4F,$D3,$75,$EA,$FC,$3E,$2F,$E2,$D8,$70,$7E,$19,$F8,$1D
	.dc.b	$3C,$5A,$78,$74,$FE,$E5,$3D,$F3,$ED,$0F,$C9,$BE,$15,$F0,$D3,$C3
	.dc.b	$C9,$F8,$2F,$CE,$5F,$B5,$7F,$9C,$FE,$B3,$FD,$C5,$F0,$8F,$D2,$FF
	.dc.b	$A0,$73,$7E,$86,$F7,$F8,$5F,$FA,$B7,$F1,$CF,$F1,$C7,$D8,$FC,$F9
	.dc.b	$CF,$0F,$A0,$3D,$BF,$55,$F7,$3C,$B3,$EF,$E9,$71,$D7,$C5,$8F,$F9
	.dc.b	$0F,$6C,$67,$DF,$6F,$27,$85,$E9,$BD,$C7,$9F,$F3,$DF,$CE,$FE,$A6
	.dc.b	$FB,$46,$CF,$D0,$FF,$CC,$5F,$B2,$3F,$5F,$7E,$E9,$C2,$F7,$5A,$9F
	.dc.b	$F5,$B7,$2F,$F0,$1C,$7A,$7E,$CB,$F5,$DE,$D9,$5C,$38,$FE,$0A,$CF
	.dc.b	$FA,$96,$1C,$67,$31,$FE,$8E,$8C,$57,$96,$B8,$63,$FE,$5C,$9C,$E5
	.dc.b	$E3,$FD,$93,$F8,$9E,$BF,$A5,$F8,$77,$E6,$CD,$5F,$B8,$FF,$B8,$7F
	.dc.b	$19,$60,$F1,$38,$F6,$7F,$0B,$C7,$8E,$1C,$3A,$FF,$D8,$17,$FE,$9A
	.dc.b	$D5,$A7,$7D,$3B,$63,$FD,$D6,$72,$F3,$87,$EB,$9F,$DE,$63,$B9,$FD
	.dc.b	$26,$3C,$DA,$F9,$7F,$1D,$FE,$39,$7C,$AB,$59,$F8,$F6,$F8,$D9,$78
	.dc.b	$63,$19,$BF,$17,$EB,$F8,$27,$F8,$57,$D3,$FF,$7B,$FC,$3B,$F1,$AF
	.dc.b	$C0,$FE,$DB,$AF,$9F,$FE,$39,$F3,$9F,$EF,$E9,$FC,$2F,$ED,$D9,$C3
	.dc.b	$D1,$C1,$E3,$CB,$FC,$41,$FB,$FE,$37,$FE,$29,$1D,$F1,$E5,$E2,$7A
	.dc.b	$A7,$39,$F6,$D7,$C5,$EB,$67,$8F,$F9,$D6,$3C,$67,$CD,$DD,$FB,$2D
	.dc.b	$3C,$18,$71,$DF,$F3,$3F,$39,$F2,$BF,$39,$FE,$9F,$F4,$FF,$E5,$BF
	.dc.b	$82,$7E,$64,$FC,$B1,$FA,$BE,$FE,$27,$FD,$67,$8F,$F3,$DA,$D7,$EB
	.dc.b	$19,$C7,$D9,$C3,$FD,$37,$3F,$EB,$7F,$5B,$88,$47,$83,$0E,$59,$FA
	.dc.b	$E7,$39,$76,$D3,$C7,$59,$BE,$11,$E6,$B7,$EB,$3F,$DB,$3E,$37,$F9
	.dc.b	$F7,$E8,$6F,$DE,$7F,$96,$FC,$8B,$B3,$FC,$EF,$FA,$63,$F6,$0E,$1B
	.dc.b	$FF,$63,$C3,$87,$C1,$7F,$BA,$EF,$7E,$F6,$A9,$5A,$7F,$31,$9F,$83
	.dc.b	$37,$F0,$07,$E0,$23,$18,$F3,$F9,$2F,$C5,$5F,$3C,$3B,$EB,$E5,$97
	.dc.b	$83,$F9,$AC,$38,$CE,$76,$7D,$57,$8F,$F0,$57,$E8,$BF,$EA,$9A,$7D
	.dc.b	$4F,$FA,$AF,$E0,$7F,$C1,$7E,$0F,$F8,$3B,$6F,$F0,$06,$1C,$25,$7C
	.dc.b	$1E,$55,$BA,$FF,$59,$E1,$FD,$11,$FF,$08,$67,$FD,$AF,$0F,$0F,$C5
	.dc.b	$FB,$E7,$AD,$78,$F2,$F1,$47,$9E,$31,$8F,$F4,$B8,$FF,$11,$FD,$07
	.dc.b	$F7,$5F,$C9,$FF,$94,$FE,$C7,$FE,$5B,$E7,$7E,$FE,$FD,$15,$FA,$07
	.dc.b	$0C,$7F,$C3,$58,$70,$FA,$CF,$BD,$7B,$FD,$13,$54,$AF,$E6,$BB,$7E
	.dc.b	$D3,$4F,$07,$C2,$B5,$C5,$F4,$8C,$65,$D3,$D9,$FE,$1B,$7C,$2F,$CD
	.dc.b	$9B,$F1,$DC,$E3,$FE,$2D,$FC,$81,$F6,$8F,$CB,$BF,$13,$FA,$4B,$F1
	.dc.b	$2F,$ED,$9F,$82,$F9,$AF,$75,$A9,$FE,$E1,$F9,$96,$70,$7D,$4F,$9B
	.dc.b	$26,$1C,$1F,$E5,$5B,$7F,$87,$E3,$9F,$86,$D3,$C1,$F3,$1E,$6B,$E3
	.dc.b	$8F,$96,$B5,$F9,$8F,$F8,$9F,$C0,$FF,$38,$D7,$AE,$B5,$87,$2F,$DB
	.dc.b	$3E,$6F,$81,$FE,$CF,$FA,$EF,$E0,$FF,$AB,$FF,$A4,$FF,$50,$7D,$7E
	.dc.b	$BF,$01,$56,$BA,$D4,$FC,$AF,$5E,$BF,$B7,$2E,$9B,$7B,$0F,$CC,$E3
	.dc.b	$CB,$DC,$69,$FE,$A1,$FC,$63,$D9,$7E,$39,$CE,$6F,$CC,$3F,$A2,$FE
	.dc.b	$BF,$F9,$9F,$D4,$FF,$53,$F8,$31,$FE,$85,$F8,$A3,$F2,$AF,$F3,$BD
	.dc.b	$6F,$CE,$5E,$FE,$3F,$FD,$E7,$F4,$DA,$E1,$FD,$6E,$4F,$D3,$BF,$A0
	.dc.b	$E3,$C6,$7F,$AF,$3C,$F0,$E5,$9F,$92,$78,$E5,$DB,$0F,$EA,$9F,$B4
	.dc.b	$FF,$5C,$FD,$87,$F6,$1F,$A3,$FF,$00,$E9,$F1,$FE,$E7,$FB,$AF,$6B
	.dc.b	$FD,$1B,$5A,$C7,$E7,$E8,$AF,$DF,$FE,$E5,$AD,$3F,$81,$E5,$FC,$29
	.dc.b	$FE,$E5,$87,$6D,$FF,$AC,$BC,$F3,$EF,$ED,$78,$EB,$8E,$7D,$B2,$FF
	.dc.b	$6B,$F8,$5F,$BD,$97,$47,$FA,$3E,$BF,$8B,$FE,$1C,$F6,$7F,$D0,$DE
	.dc.b	$47,$FA,$93,$C8,$FE,$19,$87,$F1,$CB,$5E,$FE,$1B,$EF,$E5,$D7,$2E
	.dc.b	$97,$FE,$09,$C3,$8E,$FF,$DC,$1E,$78,$F8,$33,$FE,$3F,$5D,$CF,$E4
	.dc.b	$67,$7C,$BC,$9F,$9F,$7F,$82,$BF,$58,$F1,$F8,$5F,$3E,$7F,$B1,$FE
	.dc.b	$5D,$FE,$BA,$9C,$E3,$F9,$C9,$FE,$CF,$FB,$8E,$71,$EB,$87,$F8,$0A
	.dc.b	$FD,$BF,$A9,$D9,$CE,$1C,$74,$F0,$46,$39,$3E,$0E,$BF,$E1,$CF,$D8
	.dc.b	$BE,$2F,$67,$9F,$A9,$F7,$BF,$EC,$C9,$CE,$5F,$3A,$7F,$39,$C8,$9E
	.dc.b	$39,$7F,$8B,$DF,$9F,$7F,$E1,$67,$8D,$F8,$27,$A1,$F1,$B0,$F8,$DF
	.dc.b	$63,$FE,$2E,$F3,$BE,$BF,$67,$D7,$D7,$EB,$E9,$FE,$DE,$9C,$E3,$F2
	.dc.b	$75,$C6,$5F,$43,$0F,$F1,$2D,$7E,$03,$E0,$2C,$E6,$FF,$3D,$7D,$E3
	.dc.b	$11,$FC,$2B,$2F,$F1,$FF,$91,$F8,$89,$FC,$44,$F8,$B7,$FF,$0C,$C7
	.dc.b	$DE,$33,$CB,$F8,$C8,$CC,$7E,$58,$67,$FE,$38,$BE,$97,$F9,$3C,$7B
	.dc.b	$D7,$38,$FC,$B6,$3F,$E5,$0F,$93,$F1,$72,$FC,$3D,$BF,$DF,$F1,$8A
	.dc.b	$FC,$56,$68,$F9,$38,$7F,$98,$5D,$4F,$E5,$A7,$53,$E9,$BF,$D2,$74
	.dc.b	$7F,$97,$BE,$3B,$E0,$71,$7C,$0C,$98,$7F,$89,$30,$F8,$13,$F0,$23
	.dc.b	$11,$EF,$DB,$FC,$AB,$7C,$B2,$FF,$05,$E3,$D6,$B8,$C7,$F8,$3B,$F6
	.dc.b	$CF,$F3,$6F,$C3,$7C,$FC,$3E,$7F,$9F,$E0,$8C,$57,$E0,$F2,$FF,$6D
	.dc.b	$61,$F7,$CF,$F3,$F4,$EA,$58,$F4,$9F,$AD,$87,$F0,$1C,$7F,$CF,$7C
	.dc.b	$1F,$76,$7E,$EF,$6F,$BA,$31,$87,$E0,$32,$FF,$0B,$77,$2F,$ED,$9F
	.dc.b	$E8,$DA,$DD,$78,$23,$91,$F8,$53,$C0,$7C,$AD,$FC,$1C,$1D,$5B,$3F
	.dc.b	$E3,$0C,$75,$5E,$D3,$C0,$7F,$A4,$A7,$7A,$58,$ED,$7F,$2D,$FA,$FC
	.dc.b	$9E,$DE,$1F,$53,$71,$ED,$A7,$F2,$2A,$EC,$7F,$75,$9D,$61,$FE,$A0
	.dc.b	$78,$6B,$FE,$A2,$FB,$5F,$91,$BF,$C5,$6E,$3D,$1F,$C8,$E1,$E8,$3F
	.dc.b	$12,$EB,$0E,$51,$F1,$6B,$D1,$8F,$FB,$EE,$3E,$77,$C0,$3F,$0E,$7E
	.dc.b	$26,$FA,$17,$E6,$CF,$D0,$1F,$87,$7E,$03,$2E,$59,$7F,$8C,$71,$E3
	.dc.b	$87,$0B,$DF,$EA,$DF,$9C,$F4,$BE,$6E,$9D,$F4,$FF,$56,$C6,$30,$7E
	.dc.b	$7D,$8C,$BC,$B1,$D3,$5F,$17,$C9,$7F,$B1,$AF,$86,$1E,$79,$F3,$7C
	.dc.b	$FE,$70,$FC,$21,$F8,$DA,$FC,$93,$FD,$57,$A7,$C8,$FD,$43,$81,$FC
	.dc.b	$2C,$FC,$1F,$27,$E3,$F8,$3D,$FD,$7F,$3E,$7E,$5F,$AF,$D3,$AB,$FB
	.dc.b	$05,$7F,$84,$67,$C3,$1F,$82,$9F,$AF,$1F,$E3,$57,$87,$37,$FB,$81
	.dc.b	$E5,$D1,$FE,$E2,$FC,$0B,$CF,$D7,$D1,$B9,$FF,$85,$72,$47,$4C,$78
	.dc.b	$1F,$02,$BD,$11,$CA,$FF,$DD,$BF,$81,$7C,$2E,$AF,$3B,$71,$E8,$F3
	.dc.b	$C3,$F8,$9B,$BC,$F8,$C7,$CC,$A5,$FF,$BE,$3C,$3F,$9B,$3F,$09,$B8
	.dc.b	$79,$DE,$99,$9D,$E1,$C6,$BE,$B9,$D0,$FE,$47,$F9,$47,$4D,$BF,$D4
	.dc.b	$F6,$8C,$70,$6A,$BE,$D1,$4F,$B4,$7E,$F3,$E3,$3F,$86,$57,$18,$EF
	.dc.b	$F9,$1E,$3F,$16,$F4,$F0,$F6,$6E,$1F,$CD,$1F,$49,$EE,$FC,$43,$9F
	.dc.b	$E8,$5F,$03,$87,$CA,$DA,$FE,$76,$FC,$03,$86,$3F,$CF,$BF,$6A,$F8
	.dc.b	$EF,$B1,$B3,$98,$F5,$E6,$9E,$2B,$F1,$6E,$3B,$FF,$48,$79,$FA,$76
	.dc.b	$F8,$C7,$0E,$EE,$FB,$7F,$4E,$7C,$3E,$95,$AE,$F1,$DB,$F6,$9F,$2E
	.dc.b	$DF,$D5,$1F,$3F,$A1,$6B,$CB,$1E,$4F,$C3,$37,$A9,$FB,$E7,$A3,$F3
	.dc.b	$0C,$9F,$82,$3D,$7F,$8F,$70,$3F,$0A,$7B,$BE,$3B,$99,$F8,$83,$E0
	.dc.b	$FD,$EF,$63,$F1,$67,$C4,$FC,$93,$C2,$7E,$38,$F8,$FF,$98,$F9,$0F
	.dc.b	$C8,$9F,$2B,$F1,$4F,$39,$F9,$43,$E6,$FE,$81,$EA,$3F,$2E,$7D,$0F
	.dc.b	$94,$F6,$9F,$9A,$3E,$9F,$CC,$FC,$03,$F3,$A7,$D6,$FC,$CF,$E1,$9F
	.dc.b	$A0,$3E,$CF,$E9,$7F,$18,$FD,$19,$F7,$3F,$7B,$F9,$27,$E9,$8F,$BF
	.dc.b	$FB,$4F,$CC,$3F,$52,$7E,$0B,$F6,$CF,$9E,$7E,$B0,$FC,$2F,$C9,$FD
	.dc.b	$23,$F5,$E7,$E2,$3F,$07,$FA,$A7,$EC,$8F,$C5,$FC,$A7,$D8,$3F,$6A
	.dc.b	$7E,$3B,$E9,$BE,$D9,$FB,$83,$F2,$3F,$9A,$FD,$E3,$F7,$67,$E5,$3F
	.dc.b	$4A,$FE,$D3,$F5,$0F,$D2,$1E,$7B,$F5,$97,$D4,$BF,$42,$7E,$3A,$F9
	.dc.b	$BF,$F1,$0E,$5C,$B1,$FF,$4E,$78,$5F,$D7,$7C,$14,$E7,$CC,$FE,$65
	.dc.b	$3C,$0F,$F0,$46,$BE,$1D,$7C,$1A,$77,$F9,$CF,$E0,$38,$75,$9C,$EB
	.dc.b	$E1,$C3,$91,$FC,$20,$FF,$01,$F8,$D8,$F6,$BF,$C2,$8F,$83,$4F,$1C
	.dc.b	$FF,$18,$F4,$3F,$86,$FD,$17,$E8,$8F,$E3,$E7,$EE,$3F,$41,$FD,$29
	.dc.b	$FC,$88,$FD,$DF,$D0,$FE,$A0,$FE,$4A,$7E,$FB,$EA,$FF,$56,$7F,$28
	.dc.b	$3F,$80,$FD,$57,$EB,$8F,$E5,$67,$F0,$4F,$A7,$FD,$89,$FC,$B8,$FE
	.dc.b	$0F,$F5,$9F,$B4,$3F,$98,$9F,$C2,$BF,$2F,$FD,$B9,$FC,$D0,$FE,$1B
	.dc.b	$FA,$BF,$EE,$8F,$E6,$E7,$F1,$0F,$C4,$7F,$7A,$7F,$3A,$3F,$89,$FE
	.dc.b	$8F,$FC,$00,$FE,$7A,$7F,$16,$DF,$F8,$19,$FD,$00,$FE,$33,$CF,$FE
	.dc.b	$2E,$F0,$3F,$C1,$FF,$66,$FC,$19,$F9,$6B,$F2,$95,$BF,$62,$FF,$32
	.dc.b	$BC,$7E,$39,$FA,$CF,$D3,$8C,$63,$7A,$3F,$C9,$59,$F6,$CF,$AE,$4F
	.dc.b	$C2,$FF,$A1,$E1,$C2,$73,$F2,$3F,$D1,$BE,$FB,$FB,$67,$67,$F9,$D1
	.dc.b	$FF,$4F,$65,$FE,$01,$E0,$47,$F7,$CE,$CE,$13,$FD,$68,$FF,$49,$CE
	.dc.b	$9F,$C3,$1F,$E7,$EF,$C1,$3F,$C9,$FF,$84,$89,$1F,$E5,$2A,$FE,$AF
	.dc.b	$3E,$2A,$3E,$29,$FD,$23,$E1,$7A,$1D,$A3,$FC,$35,$5D,$48,$3E,$9F
	.dc.b	$CB,$E0,$78,$A1,$F8,$F7,$F9,$2F,$1C,$E1,$FE,$73,$9F,$34,$FF,$93
	.dc.b	$FE,$2F,$FB,$5D,$F1,$FB,$FE,$17,$FE,$9A,$8F,$33,$FA,$8C,$BF,$C5
	.dc.b	$91,$FE,$7D,$FD,$6B,$F5,$4E,$B2,$E7,$5F,$27,$6E,$4F,$AF,$F5,$BE
	.dc.b	$CF,$F9,$7E,$3F,$D4,$F1,$88,$FF,$44,$4F,$F7,$EF,$C6,$FF,$A9,$9F
	.dc.b	$E7,$CE,$CF,$E0,$F3,$93,$E3,$7E,$99,$FD,$B3,$F6,$4F,$85,$C1,$F1
	.dc.b	$B8,$7E,$CF,$77,$FA,$13,$F4,$A7,$E5,$0F,$DD,$5F,$A0,$7F,$19,$C3
	.dc.b	$FC,$D5,$E4,$7F,$AE,$B0,$E1,$7B,$AD,$7C,$57,$F8,$4A,$73,$87,$2F
	.dc.b	$57,$F0,$5F,$33,$FE,$A1,$FD,$99,$F0,$FF,$E9,$4F,$C7,$3F,$94,$7A
	.dc.b	$4F,$B1,$FF,$15,$7E,$12,$FD,$61,$F6,$4C,$78,$F3,$FF,$AA,$F1,$E3
	.dc.b	$87,$0F,$9D,$ED,$9F,$5F,$61,$A7,$5D,$5A,$BC,$3F,$27,$D0,$7F,$A9
	.dc.b	$BF,$5C,$7D,$73,$BB,$FD,$71,$C8,$FB,$AF,$F6,$2F,$BB,$E3,$C3,$85
	.dc.b	$EF,$8B,$E5,$57,$FA,$EF,$8F,$FC,$81,$0C,$BA,$E5,$D2,$9F,$7D,$FE
	.dc.b	$C5,$FD,$65,$F7,$9F,$EB,$8F,$81,$7E,$62,$F5,$DF,$8F,$3F,$13,$FA
	.dc.b	$FF,$AF,$EB,$1E,$3B,$7F,$B3,$F1,$E3,$87,$0B,$FB,$BF,$85,$F7,$CF
	.dc.b	$B6,$7D,$73,$E9,$E1,$BF,$13,$FF,$6C,$7E,$CC,$F9,$9F,$F6,$5F,$73
	.dc.b	$D0,$7E,$70,$FC,$05,$ED,$BF,$6D,$7E,$34,$FC,$25,$9F,$3E,$5F,$F7
	.dc.b	$26,$7C,$F2,$E5,$8F,$1C,$3C,$3A,$78,$34,$EF,$A7,$6D,$1F,$8D,$F9
	.dc.b	$BE,$A3,$FC,$B5,$EC,$F6,$7F,$CF,$3E,$26,$76,$FD,$96,$FF,$EF,$27
	.dc.b	$FB,$20,$F6,$57,$E6,$38,$C7,$A0,$E6,$D7,$FC,$FB,$F2,$5F,$19,$FF
	.dc.b	$26,$FC,$4F,$F2,$07,$F8,$AF,$AD,$FC,$5F,$D4,$EB,$3F,$E6,$19,$0F
	.dc.b	$D0,$1F,$31,$FE,$D5,$FB,$D7,$E2,$AF,$DE,$1F,$44,$F8,$1F,$E9,$5F
	.dc.b	$67,$FE,$FE,$C7,$8F,$E0,$1F,$E0,$AD,$3B,$69,$D7,$47,$E7,$D9,$FA
	.dc.b	$1F,$39,$FC,$2E,$7F,$85,$FE,$6A,$F9,$CF,$2C,$7F,$AD,$3C,$37,$EF
	.dc.b	$BF,$D7,$3F,$64,$7E,$EF,$FE,$20,$FC,$EB,$FD,$DB,$FA,$CF,$0F,$83
	.dc.b	$7E,$76,$FD,$51,$F5,$EF,$AE,$7C,$8F,$EC,$3E,$01,$FA,$33,$EF,$7C
	.dc.b	$3F,$74,$FD,$B7,$F8,$B7,$90,$F8,$A7,$EF,$EF,$CC,$9F,$97,$FC,$BF
	.dc.b	$CA,$FD,$3F,$7E,$FD,$85,$F8,$0B,$F5,$B7,$3B,$F6,$C7,$E6,$CF,$D7
	.dc.b	$5F,$B1,$BF,$27,$FF,$7F,$FE,$E4,$FC,$79,$F9,$DB,$F5,$1F,$E2,$FF
	.dc.b	$91,$FE,$DF,$F1,$A7,$EB,$9F,$2B,$F6,$9F,$E1,$5F,$61,$FB,$EF,$89
	.dc.b	$7B,$5F,$C3,$FE,$B7,$B7,$E3,$7F,$43,$E8,$7F,$AE,$7E,$42,$FD,$2F
	.dc.b	$F6,$5F,$A6,$BE,$FB,$F9,$BF,$E9,$9F,$03,$D5,$FC,$5F,$E0,$3E,$CF
	.dc.b	$C5,$3C,$D9,$FF,$22,$E1,$7C,$CB,$F3,$27,$E4,$5F,$7B,$F1,$07,$EB
	.dc.b	$1F,$CC,$3F,$7B,$7E,$A9,$FC,$A3,$EA,$7F,$89,$FE,$BE,$FC,$2F,$F0
	.dc.b	$BF,$0F,$FD,$C7,$F2,$6F,$B3,$F9,$DF,$EF,$FA,$9F,$C0,$1F,$C7,$39
	.dc.b	$3D,$8F,$D3,$FC,$63,$C5,$F0,$9F,$C7,$BF,$39,$78,$4F,$CD,$7F,$23
	.dc.b	$F1,$6F,$2B,$FC,$53,$F5,$EF,$D0,$5F,$B4,$BF,$78,$7E,$A2,$FD,$49
	.dc.b	$FB,$6F,$E6,$BF,$57,$FD,$EF,$E8,$0F,$CC,$3F,$CF,$5F,$B1,$6F,$BD
	.dc.b	$FA,$8B,$F6,$8F,$D0,$C3,$F6,$0F,$CA,$7D,$8B,$F4,$1F,$EA,$DF,$95
	.dc.b	$3F,$11,$7E,$CE,$FD,$63,$F5,$3D,$7F,$8C,$FA,$47,$E4,$3F,$ED,$FF
	.dc.b	$DB,$B7,$F9,$17,$D2,$F9,$FF,$D7,$FE,$CC,$FC,$C1,$FA,$9B,$F6,$E7
	.dc.b	$E8,$2F,$CC,$FF,$63,$F0,$CF,$D0,$3F,$96,$7B,$BF,$E4,$DF,$C2,$7F
	.dc.b	$23,$E8,$FF,$47,$FE,$00,$5B,$F7,$97,$E9,$5F,$37,$0F,$E2,$7F,$9F
	.dc.b	$7E,$D3,$D5,$FF,$4E,$7E,$07,$F9,$AF,$D3,$5F,$09,$FA,$8F,$DF,$FF
	.dc.b	$82,$F6,$BF,$66,$7E,$B4,$FD,$01,$FB,$07,$F5,$CF,$D3,$9F,$BE,$7F
	.dc.b	$00,$FC,$EF,$F8,$6F,$DC,$BF,$33,$FE,$07,$F3,$1F,$77,$F2,$67,$E9
	.dc.b	$2F,$B9,$FE,$4F,$87,$0B,$DF,$B9,$FE,$6E,$AD,$7E,$86,$E2,$F2,$F0
	.dc.b	$F9,$38,$3C,$7F,$8B,$B8,$3C,$7F,$AF,$FC,$1E,$A7,$57,$C1,$FC,$97
	.dc.b	$FB,$07,$CA,$BA,$BE,$0F,$ED,$3F,$3B,$E3,$5E,$27,$CE,$FC,$6D,$E2
	.dc.b	$7C,$EF,$0F,$F2,$3E,$6F,$DC,$F3,$3E,$DF,$97,$F6,$BC,$AF,$B3,$F9
	.dc.b	$4F,$E1,$F8,$7F,$A1,$FA,$7F,$87,$F4,$BF,$0D,$E3,$3D,$17,$EA,$DF
	.dc.b	$95,$BF,$E9,$3E,$DF,$E5,$7E,$E7,$FB,$9F,$B5,$F9,$4F,$67,$FD,$17
	.dc.b	$F0,$7F,$A1,$FC,$41,$F0,$7F,$A1,$F2,$1F,$05,$FA,$1F,$B5,$FD,$86
	.dc.b	$9F,$D5,$3E,$33,$F6,$1F,$9B,$7F,$88,$7C,$BF,$D7,$FC,$57,$EB,$BE
	.dc.b	$CD,$F2,$9F,$BC,$5F,$27,$FB,$CF,$C7,$9D,$3F,$CD,$3E,$5F,$F8,$3F
	.dc.b	$E6,$0F,$80,$FE,$0F,$F8,$EB,$E6,$7F,$83,$FE,$D5,$FC,$53,$87,$FB
	.dc.b	$2F,$CF,$FF,$14,$FC,$1D,$F4,$1F,$C5,$3F,$2A,$FF,$22,$6C,$FD,$FF
	.dc.b	$F9,$16,$1E,$4F,$B8,$FC,$BF,$D9,$BE,$A3,$F9,$17,$D4,$7F,$22,$7F
	.dc.b	$97,$70,$BE,$B7,$F9,$77,$C5,$3E,$B3,$F9,$77,$E4,$2F,$DE,$3F,$9D
	.dc.b	$E9,$FD,$E3,$E9,$7F,$9D,$FC,$07,$F9,$73,$EC,$7F,$9D,$70,$FF,$47
	.dc.b	$AF,$EF,$BF,$59,$FC,$67,$83,$FA,$3F,$DA,$7F,$46,$FC,$91,$FB,$4B
	.dc.b	$F0,$0F,$EA,$FE,$17,$F5,$6F,$BE,$FE,$AB,$F7,$DF,$D5,$3E,$F7,$FC
	.dc.b	$69,$E9,$FF,$64,$FC,$17,$FB,$1F,$E0,$9F,$D8,$BC,$EF,$EC,$3F,$9E
	.dc.b	$7F,$5C,$E1,$FF,$06,$F8,$5F,$DC,$BF,$0D,$FE,$E3,$F8,$67,$F7,$0F
	.dc.b	$CB,$1F,$86,$7E,$9F,$F1,$2F,$EF,$FF,$5F,$FE,$41,$FC,$47,$FB,$F7
	.dc.b	$C3,$FF,$7D,$E8,$7A,$FF,$E0,$4F,$C3,$FF,$D2,$3E,$E7,$F6,$18,$FF
	.dc.b	$05,$6F,$FE,$08,$3F,$C1,$35,$FE,$12,$FC,$5B,$FC,$2B,$E0,$7F,$87
	.dc.b	$3F,$19,$FF,$0D,$FE,$31,$FD,$A7,$EB,$BF,$C3,$5F,$8B,$BF,$18,$E5
	.dc.b	$97,$F8,$9B,$2E,$53,$FE,$1A,$8C,$C7,$F8,$BF,$F1,$CF,$F1,$78,$FC
	.dc.b	$73,$FC,$77,$FB,$57,$8F,$83,$CE,$DB,$EB,$F8,$AE,$9F,$61,$E0,$FC
	.dc.b	$7B,$FC,$6B,$3F,$E4,$7A,$F1,$CF,$F9,$1F,$89,$39,$8E,$66,$3F,$21
	.dc.b	$E6,$FA,$A7,$0F,$F9,$02,$B9,$FE,$41,$E6,$AF,$F2,$7D,$FF,$3A,$FC
	.dc.b	$3F,$FC,$C9,$F9,$27,$F9,$7F,$F2,$6F,$F3,$54,$E7,$F2,$6F,$F3,$1F
	.dc.b	$E4,$FF,$E6,$7C,$FC,$7E,$BF,$F9,$9F,$EF,$FF,$CD,$7F,$93,$FF,$99
	.dc.b	$A5,$FB,$D7,$FA,$19,$7E,$55,$FE,$86,$FC,$AB,$FD,$03,$F9,$5F,$FA
	.dc.b	$42,$B5,$39,$FA,$3F,$F4,$4D,$F2,$F6,$BF,$D1,$3E,$BF,$FA,$37,$DD
	.dc.b	$FF,$44,$4F,$F7,$5F,$CB,$FF,$D4,$BF,$97,$7F,$A7,$FF,$50,$FC,$7F
	.dc.b	$CA,$7F,$A9,$3F,$30,$FF,$51,$DF,$FA,$A2,$7F,$CC,$73,$9F,$3F,$FD
	.dc.b	$59,$F9,$8F,$FA,$C7,$F3,$1C,$F5,$FF,$AF,$7F,$1E,$FF,$5E,$7C,$6F
	.dc.b	$FB,$1F,$D6,$FF,$02,$4E,$7F,$33,$FF,$62,$69,$83,$FD,$91,$F6,$3F
	.dc.b	$EC,$48,$FF,$67,$F8,$DF,$E7,$2F,$C0,$1F,$7B,$FE,$DA,$E4,$FF,$71
	.dc.b	$7E,$52,$FC,$DF,$FA,$DC,$E7,$F3,$67,$9B,$F8,$FD,$32,$7F,$9F,$3F
	.dc.b	$38,$FF,$76,$F1,$FF,$B8,$E5,$F9,$B7,$7F,$CE,$7F,$DE,$B1,$C8,$7C
	.dc.b	$9F,$FA,$4B,$6F,$F0,$66,$3F,$87,$3F,$47,$E6,$F9,$FD,$6F,$C6,$C7
	.dc.b	$EB,$FD,$0F,$ED,$F1,$FB,$28,$FD,$4D,$7F,$A1,$DE,$47,$ED,$6B,$E0
	.dc.b	$79,$7F,$E7,$EC,$7D,$7F,$9D,$FE,$5E,$FF,$D7,$D7,$FE,$96,$9C,$BF
	.dc.b	$E8,$4B,$F9,$1E,$8F,$8A,$3F,$D0,$F9,$7F,$80,$23,$11,$FE,$AD,$30
	.dc.b	$7A,$FD,$1E,$D1,$FB,$0F,$CE,$FF,$5B,$8F,$FA,$12,$B4,$77,$38,$9F
	.dc.b	$1F,$2F,$53,$8C,$BE,$9F,$E7,$9E,$DC,$B1,$08,$FE,$29,$D1,$E8,$AF
	.dc.b	$64,$FF,$AD,$B7,$FA,$1F,$9E,$FA,$36,$FC,$23,$FC,$13,$1F,$88,$FF
	.dc.b	$06,$DB,$CE,$E7,$0F,$36,$88,$F9,$98,$FF,$B0,$E7,$E1,$4B,$F2,$AF
	.dc.b	$8B,$1F,$7D,$F0,$3F,$E2,$87,$A3,$F3,$23,$83,$F1,$0F,$E5,$CE,$1F
	.dc.b	$91,$87,$C4,$87,$D4,$7F,$AE,$BF,$07,$7E,$82,$F2,$DF,$9F,$FE,$83
	.dc.b	$E2,$F7,$F7,$DF,$E3,$8C,$38,$5B,$F4,$1F,$AB,$18,$C2,$E9,$FE,$DA
	.dc.b	$CB,$AC,$74,$EF,$64,$FD,$07,$F5,$D8,$71,$D3,$C1,$ED,$7E,$C7,$E9
	.dc.b	$FE,$1F,$07,$F9,$21,$F8,$3A,$7F,$B4,$B8,$3F,$C0,$CF,$E6,$8F,$9F
	.dc.b	$E8,$F0,$83,$C2,$3F,$78,$FC,$C8,$7E,$85,$B3,$53,$FE,$56,$75,$F7
	.dc.b	$1C,$CC,$CF,$F8,$C6,$7C,$07,$FB,$E3,$A3,$F8,$47,$E5,$DE,$13,$67
	.dc.b	$7F,$BB,$F2,$1C,$8F,$17,$D8,$F9,$CE,$A7,$97,$C9,$F5,$1E,$03,$D1
	.dc.b	$FA,$1F,$B4,$F1,$9E,$BF,$B0,$F8,$07,$98,$F7,$7C,$9F,$C3,$3D,$27
	.dc.b	$C1,$F5,$3E,$31,$EC,$3E,$27,$BB,$F2,$4F,$79,$8D,$07,$E2,$BF,$46
	.dc.b	$FF,$03,$FD,$B3,$E2,$5E,$AB,$F4,$77,$A7,$FE,$80,$C7,$8E,$3C,$7E
	.dc.b	$5F,$F7,$B1,$8E,$C6,$6F,$D1,$F3,$7E,$09,$CD,$B4,$F0,$66,$FD,$1F
	.dc.b	$F9,$C6,$5C,$F2,$F2,$EB,$E4,$D3,$F9,$E7,$C8,$7E,$14,$FB,$67,$E0
	.dc.b	$FF,$48,$FC,$41,$F7,$8F,$C3,$7C,$BF,$E2,$CF,$C0,$9F,$89,$F1,$BF
	.dc.b	$1C,$7E,$10,$FC,$67,$E9,$3F,$91,$3F,$0E,$7E,$3F,$F4,$AF,$CA,$1F
	.dc.b	$8A,$3F,$25,$FA,$5F,$E5,$CF,$C6,$9F,$95,$F9,$EF,$CD,$1F,$90,$3F
	.dc.b	$31,$FA,$37,$E7,$4F,$C9,$9F,$9B,$FB,$7F,$D0,$1F,$96,$3F,$3D,$FA
	.dc.b	$1F,$E8,$CF,$CC,$9F,$A1,$F9,$4F,$D3,$1F,$9C,$3F,$49,$F7,$BF,$A9
	.dc.b	$3F,$3E,$7E,$9F,$F4,$CF,$D6,$1F,$A2,$3F,$55,$F0,$FF,$AF,$3F,$4A
	.dc.b	$7E,$B7,$F2,$9F,$D9,$1F,$A8,$3F,$61,$F8,$57,$ED,$4F,$D5,$9F,$B3
	.dc.b	$FC,$33,$F7,$07,$EB,$8F,$DB,$7E,$9B,$FB,$B3,$F6,$27,$EE,$7F,$21
	.dc.b	$FD,$F1,$FB,$43,$F7,$9F,$A1,$7F,$01,$3F,$95,$FB,$DF,$ED,$07,$F7
	.dc.b	$27,$F0,$1F,$8B,$FE,$5F,$F8,$E7,$FB,$51,$FE,$63,$FA,$15,$F1,$3F
	.dc.b	$B7,$7A,$BF,$D4,$A7,$F8,$47,$E9,$DF,$CD,$BF,$1C,$FF,$6D,$BF,$CE
	.dc.b	$3E,$CA,$DF,$03,$FB,$7C,$7F,$0F,$FA,$8F,$E7,$BE,$47,$FB,$89,$FE
	.dc.b	$7F,$E5,$5A,$E0,$7F,$3A,$8F,$E2,$DF,$A7,$FF,$68,$CF,$FC,$EB,$EB
	.dc.b	$23,$17,$FE,$EC,$7F,$8A,$9F,$C7,$7F,$50,$FE,$DF,$BF,$06,$CF,$FB
	.dc.b	$C9,$FE,$36,$7F,$23,$FC,$73,$FB,$AE,$DD,$9B,$3F,$EF,$87,$F9,$09
	.dc.b	$FD,$EB,$ED,$3F,$C4,$CF,$F7,$CF,$8B,$FF,$14,$7C,$9F,$F7,$ED,$FC
	.dc.b	$ED,$BF,$F8,$02,$FF,$2A,$3F,$C0,$7F,$3B,$C3,$83,$F9,$58,$F9,$FF
	.dc.b	$F1,$97,$EA,$3F,$E0,$BE,$1F,$F4,$AE,$2E,$14,$E6,$3F,$84,$46,$2F
	.dc.b	$E5,$9F,$D1,$E7,$96,$1D,$23,$F9,$CD,$F7,$7A,$7D,$9F,$F8,$67,$87
	.dc.b	$BE,$08,$C6,$3D,$9F,$E8,$66,$F0,$FF,$24,$3C,$72,$FF,$25,$7E,$A5
	.dc.b	$FE,$23,$E1,$E6,$DB,$F3,$BF,$D3,$0D,$5F,$0C,$BF,$A6,$D7,$97,$17
	.dc.b	$E4,$7F,$E2,$DC,$3C,$AD,$7E,$0B,$FD,$58,$D5,$FF,$96,$DE,$18,$DC
	.dc.b	$3F,$DA,$7D,$1F,$E0,$2F,$D4,$FF,$DE,$5F,$62,$FC,$B5,$8F,$CF,$C7
	.dc.b	$F9,$CF,$E0,$4C,$38,$7E,$1F,$FE,$B5,$B7,$DB,$6F,$3E,$BF,$27,$FE
	.dc.b	$3A,$C7,$8E,$0F,$C9,$F1,$5A,$D7,$C3,$F1,$DE,$58,$69,$DF,$15,$79
	.dc.b	$BF,$0A,$FF,$3B,$7D,$4F,$F9,$83,$A3,$F0,$BC,$7F,$85,$8C,$3F,$ED
	.dc.b	$CE,$BC,$65,$F8,$43,$89,$FE,$09,$AF,$F7,$0F,$16,$23,$FC,$A0,$7F
	.dc.b	$60,$8F,$F7,$2F,$27,$FA,$17,$F5,$4F,$F3,$7F,$57,$83,$9D,$18,$CB
	.dc.b	$C1,$18,$C3,$C0,$7E,$66,$75,$8F,$38,$FF,$0D,$5F,$67,$F3,$DF,$85
	.dc.b	$FF,$9F,$3A,$7B,$37,$1E,$AF,$F8,$50,$C5,$FF,$A6,$5D,$65,$FE,$9B
	.dc.b	$FC,$BF,$FD,$0B,$5F,$AA,$D7,$F5,$51,$88,$FC,$FB,$FE,$1E,$31,$7C
	.dc.b	$71,$FF,$17,$4F,$97,$47,$EA,$BF,$ED,$CD,$BE,$8F,$E1,$3F,$B5,$8C
	.dc.b	$57,$EE,$3F,$57,$C6,$1F,$0B,$0E,$2F,$FA,$4B,$0F,$DE,$E3,$FE,$B1
	.dc.b	$C3,$8E,$1F,$A1,$FA,$DF,$F7,$4F,$07,$DE,$DF,$F8,$14,$62,$7F,$82
	.dc.b	$F0,$E3,$2E,$D3,$99,$FC,$DC,$64,$FB,$B7,$FC,$33,$83,$FD,$81,$FA
	.dc.b	$BF,$FB,$DB,$8B,$D1,$97,$F0,$F8,$C6,$01,$BF,$E0,$37,$4F,$F9,$18
	.dc.b	$DC,$BE,$C3,$AE,$7F,$EB,$0C,$7D,$EF,$F1,$BF,$AF,$FF,$12,$FD,$4F
	.dc.b	$F8,$9A,$FD,$79,$E6,$BF,$D6,$F1,$EE,$DF,$11,$FE,$7E,$3F,$D7,$B7
	.dc.b	$C9,$FF,$1B,$CF,$F2,$8E,$4F,$F6,$E4,$E6,$7F,$4F,$E3,$7F,$B7,$FE
	.dc.b	$F3,$FD,$39,$7E,$DD,$FF,$37,$97,$F2,$F9,$CC,$FC,$0C,$99,$67,$1F
	.dc.b	$B9,$5F,$EA,$37,$FD,$9E,$F0,$C3,$71,$B3,$FD,$2B,$7D,$F1,$FF,$78
	.dc.b	$7E,$65,$E9,$E2,$F8,$3E,$8F,$F3,$F8,$C5,$7D,$1C,$DA,$2F,$58,$FF
	.dc.b	$B9,$5F,$F6,$4D,$FC,$3C,$2C,$E7,$0E,$5A,$78,$23,$11,$FB,$AC,$BF
	.dc.b	$CB,$FC,$4F,$D6,$3E,$67,$0F,$C5,$D5,$C0,$FC,$FF,$FA,$A5,$6B,$1F
	.dc.b	$8D,$2F,$B7,$CE,$7F,$EF,$9C,$7F,$DB,$B7,$F5,$FD,$4B,$39,$C3,$E7
	.dc.b	$7D,$5F,$8A,$BA,$71,$FF,$0E,$DF,$CB,$3F,$7E,$BD,$7F,$96,$3F,$59
	.dc.b	$FE,$CD,$39,$9E,$12,$FD,$6F,$39,$7F,$9F,$3A,$3F,$0E,$FF,$9D,$8E
	.dc.b	$A6,$4F,$F3,$A1,$93,$89,$FE,$F5,$AF,$97,$D3,$C2,$31,$1E,$7E,$EF
	.dc.b	$97,$F9,$17,$E7,$BC,$0F,$C7,$75,$BF,$1C,$FD,$37,$E6,$DF,$DF,$6B
	.dc.b	$58,$7E,$BA,$B5,$93,$F4,$3D,$67,$E9,$CB,$F6,$8F,$97,$A9,$8F,$7F
	.dc.b	$17,$D1,$3C,$06,$CB,$F1,$EB,$FB,$79,$D4,$E7,$4F,$14,$BF,$06,$FC
	.dc.b	$2F,$EB,$7F,$B8,$F0,$7F,$82,$6F,$F9,$3F,$92,$FF,$4A,$7A,$DF,$E1
	.dc.b	$7A,$D6,$3F,$98,$B5,$7F,$A6,$7D,$2D,$69,$F4,$E7,$E7,$BF,$EF,$CB
	.dc.b	$E2,$FB,$5E,$78,$F7,$F9,$9F,$1C,$F1,$C7,$B6,$3F,$EB,$28,$FF,$14
	.dc.b	$7E,$41,$FC,$3B,$BF,$F8,$CF,$83,$F8,$3F,$D3,$9F,$40,$FC,$77,$78
	.dc.b	$7E,$16,$F7,$D8,$BD,$FE,$49,$AA,$F9,$74,$F4,$3E,$7F,$53,$4F,$07
	.dc.b	$87,$86,$EB,$E1,$FA,$4F,$D7,$C6,$34,$ED,$F3,$18,$CF,$FD,$A3,$E2
	.dc.b	$3E,$31,$F9,$83,$F4,$B7,$A2,$F9,$57,$92,$FA,$8F,$36,$3C,$70,$E1
	.dc.b	$7B,$A7,$D4,$EB,$27,$EA,$FE,$7D,$3C,$31,$E0,$7F,$1D,$BF,$FB,$73
	.dc.b	$2E,$9A,$F8,$B5,$F0,$D7,$3D,$3B,$FC,$C7,$FB,$8B,$93,$E2,$4F,$49
	.dc.b	$F1,$D7,$A3,$EA,$BE,$67,$D0,$3F,$5C,$FF,$11,$F5,$BF,$2C,$7D,$6B
	.dc.b	$EE,$1F,$2F,$C2,$F7,$5A,$97,$E0,$F9,$FA,$DF,$66,$CC,$FB,$7C,$46
	.dc.b	$1B,$A7,$7D,$3F,$DF,$D5,$D6,$BF,$05,$97,$8B,$E5,$3F,$93,$FE,$01
	.dc.b	$FD,$6B,$C8,$FE,$8B,$EB,$34,$FE,$8F,$ED,$7E,$17,$95,$7B,$8F,$EC
	.dc.b	$D3,$AA,$FF,$3F,$4E,$73,$FD,$E3,$D7,$E1,$BF,$1A,$F1,$9E,$D7,$FD
	.dc.b	$3B,$0E,$B7,$FC,$77,$1F,$1E,$3F,$D6,$3B,$BC,$F2,$3F,$15,$F8,$53
	.dc.b	$EF,$BF,$BC,$6C,$FD,$5D,$EE,$3F,$3D,$F2,$E1,$C7,$B9,$8F,$1D,$FE
	.dc.b	$B6,$1C,$3F,$20,$DD,$F6,$CF,$D2,$F1,$B7,$EB,$B8,$8D,$61,$CB,$E3
	.dc.b	$BC,$F1,$8D,$7C,$9F,$23,$8D,$3C,$39,$7A,$34,$F3,$FE,$BD,$FE,$A7
	.dc.b	$FD,$8B,$DD,$8B,$F0,$6F,$DB,$7E,$C2,$FD,$79,$F1,$0F,$C4,$7F,$8F
	.dc.b	$7C,$A7,$1C,$3E,$36,$1C,$2F,$75,$FC,$8F,$E4,$F5,$5D,$7D,$86,$5F
	.dc.b	$C0,$B3,$F0,$E3,$FC,$17,$F6,$1C,$5F,$4B,$DE,$5D,$3E,$17,$FB,$8E
	.dc.b	$1C,$27,$3A,$F9,$6B,$37,$C7,$2F,$54,$FA,$70,$FE,$EB,$DC,$C7,$E0
	.dc.b	$FE,$08,$FC,$2B,$FD,$17,$FB,$1B,$F5,$B7,$4B,$F6,$37,$E9,$5E,$CC
	.dc.b	$38,$7C,$EB,$0E,$17,$BE,$0D,$65,$FE,$96,$E0,$FE,$A5,$A7,$82,$3F
	.dc.b	$71,$18,$3F,$21,$8F,$4C,$39,$F5,$FB,$25,$39,$D7,$CB,$F8,$56,$63
	.dc.b	$19,$F9,$FF,$48,$F0,$EB,$E5,$BF,$5E,$EF,$C6,$FE,$AF,$D9,$BF,$5A
	.dc.b	$7E,$C4,$FD,$91,$FB,$2B,$F5,$F7,$E5,$7F,$D5,$BA,$3E,$57,$17,$E9
	.dc.b	$F2,$E5,$8F,$1C,$38,$5B,$F0,$3E,$F9,$F6,$C3,$E0,$0C,$3C,$5F,$A6
	.dc.b	$FE,$0A,$31,$3E,$4D,$3C,$7A,$7F,$10,$D3,$C3,$5A,$D3,$C1,$3F,$66
	.dc.b	$3B,$C6,$33,$7C,$4A,$DF,$8C,$7E,$8A,$9F,$A4,$7D,$FF,$C4,$DF,$8E
	.dc.b	$3F,$52,$7E,$CC,$FC,$C7,$96,$3C,$70,$E1,$7B,$AF,$8F,$97,$6D,$7F
	.dc.b	$86,$67,$DB,$F6,$7F,$EC,$78,$F1,$E7,$B5,$AC,$B9,$E9,$FD,$C6,$3B
	.dc.b	$99,$96,$BE,$69,$F8,$19,$77,$AF,$B7,$7C,$B3,$FE,$91,$39,$CF,$DF
	.dc.b	$8F,$2D,$3D,$1F,$9A,$F6,$C5,$FA,$B7,$E8,$3C,$0F,$C1,$FE,$AE,$FD
	.dc.b	$A1,$F4,$7F,$2B,$4F,$05,$EF,$B1,$7B,$AF,$D8,$56,$BC,$86,$3F,$03
	.dc.b	$2E,$F1,$FA,$F7,$FC,$71,$C7,$FC,$5B,$2E,$98,$F8,$7E,$03,$FC,$E7
	.dc.b	$39,$BE,$79,$F8,$34,$7D,$8E,$B2,$FE,$79,$97,$29,$F4,$F9,$1E,$F8
	.dc.b	$C7,$DA,$EE,$9F,$A9,$7E,$67,$F6,$8F,$E8,$5E,$BB,$F6,$6F,$F6,$FF
	.dc.b	$E0,$3F,$1E,$B5,$ED,$AB,$5F,$11,$E2,$E5,$FA,$95,$FA,$3F,$D5,$3E
	.dc.b	$03,$D7,$3F,$79,$FD,$45,$E7,$7D,$F9,$3C,$91,$8C,$FF,$C1,$78,$74
	.dc.b	$F2,$5A,$F9,$A7,$FD,$E7,$F8,$17,$DB,$FC,$39,$E7,$3D,$27,$EC,$4F
	.dc.b	$CF,$DF,$78,$FC,$D7,$F4,$DA,$FE,$2F,$1E,$38,$70,$BD,$D3,$EB,$7B
	.dc.b	$64,$FD,$8F,$EE,$61,$C3,$8F,$DA,$1E,$7A,$78,$74,$F0,$4F,$2C,$FB
	.dc.b	$47,$DA,$8C,$4E,$F2,$7E,$41,$8D,$3C,$7F,$B6,$5C,$3D,$5F,$B5,$7E
	.dc.b	$6F,$E5,$3F,$CB,$5F,$13,$FE,$3A,$F8,$4F,$F7,$97,$E8,$AF,$CE,$35
	.dc.b	$F6,$3F,$DF,$7E,$A3,$FC,$B1,$F2,$0A,$FE,$C7,$CF,$FA,$DC,$B9,$F1
	.dc.b	$7B,$1F,$F3,$37,$E5,$7F,$9B,$7B,$E1,$CB,$F0,$EF,$C6,$46,$34,$64
	.dc.b	$FC,$A3,$3A,$78,$67,$CB,$FB,$5F,$93,$7F,$45,$F8,$FE,$CF,$F8,$D7
	.dc.b	$95,$FD,$EB,$F0,$EF,$F3,$67,$ED,$9F,$D6,$7F,$60,$7E,$D6,$FB,$F7
	.dc.b	$E2,$1C,$39,$3F,$C3,$D8,$70,$F8,$B6,$1C,$2D,$F8,$C6,$ED,$7D,$9F
	.dc.b	$F5,$66,$7E,$1F,$B3,$FE,$2D,$1F,$EC,$7E,$8F,$ED,$8F,$68,$6B,$E5
	.dc.b	$D7,$EE,$56,$AB,$AD,$6B,$4F,$1D,$7F,$8E,$70,$FF,$0B,$FD,$23,$CF
	.dc.b	$FF,$48,$7E,$60,$FD,$AD,$FA,$43,$F1,$9F,$E9,$9A,$F5,$C3,$87,$D6
	.dc.b	$B0,$E1,$7B,$FC,$7F,$DD,$5A,$DF,$DA,$7F,$90,$DF,$07,$E2,$BF,$E9
	.dc.b	$AD,$7C,$7A,$F8,$B5,$F0,$FB,$BF,$E4,$9D,$7C,$13,$9C,$FF,$3B,$18
	.dc.b	$D7,$C5,$7E,$8D,$BF,$D3,$53,$9D,$BC,$B0,$FC,$7F,$E9,$7E,$85,$FC
	.dc.b	$53,$E6,$FF,$CA,$DF,$01,$FB,$8F,$9E,$7E,$0F,$AF,$6F,$FC,$29,$C9
	.dc.b	$FC,$DF,$CD,$FF,$6B,$7E,$2B,$9C,$7D,$53,$FE,$5F,$CB,$F8,$FF,$D6
	.dc.b	$E2,$11,$DD,$E5,$97,$B7,$97,$FC,$93,$39,$D3,$F8,$4F,$D0,$E6,$FF
	.dc.b	$CD,$5F,$98,$3F,$41,$FF,$17,$7E,$D2,$EC,$7E,$11,$FC,$47,$EA,$3F
	.dc.b	$1D,$7B,$D1,$7B,$AF,$E2,$DF,$51,$AC,$7F,$11,$97,$FA,$A6,$FF,$CB
	.dc.b	$78,$3F,$42,$FF,$7E,$CE,$E7,$9E,$7E,$0F,$33,$CD,$2B,$F0,$67,$DF
	.dc.b	$CF,$CF,$DF,$3F,$6D,$7E,$BC,$FD,$59,$EE,$7A,$BF,$5D,$7E,$DF,$FA
	.dc.b	$7C,$38,$5E,$EB,$5F,$A1,$E7,$17,$EC,$9F,$3E,$F7,$B7,$E4,$03,$CB
	.dc.b	$3E,$FC,$AF,$6F,$CB,$87,$92,$73,$ED,$7B,$3F,$0F,$FF,$2A,$78,$AF
	.dc.b	$CC,$DF,$AA,$FF,$9F,$79,$9F,$AB,$FE,$F6,$F7,$5A,$EE,$FF,$4F,$7E
	.dc.b	$DF,$9F,$11,$F9,$5F,$E1,$6F,$AF,$17,$9D,$E3,$9F,$58,$E1,$A7,$86
	.dc.b	$3F,$D1,$9F,$1F,$E5,$FC,$BD,$F9,$93,$F6,$47,$CB,$BF,$1F,$7E,$64
	.dc.b	$FD,$8B,$CD,$87,$0B,$DD,$6B,$F6,$DC,$FE,$27,$E5,$AE,$BB,$79,$5E
	.dc.b	$19,$75,$FC,$D7,$C5,$A7,$86,$3C,$1F,$29,$F8,$4F,$D0,$3F,$80,$7D
	.dc.b	$2F,$F9,$67,$C5,$7E,$5A,$FD,$D3,$F9,$25,$6B,$E8,$7F,$03,$2F,$D2
	.dc.b	$F3,$B7,$FB,$7C,$FB,$95,$FE,$CD,$FD,$C7,$F5,$CF,$1C,$5F,$B7,$78
	.dc.b	$6B,$FD,$CD,$9F,$87,$F5,$5F,$27,$DB,$7F,$96,$3F,$05,$7E,$B0,$FA
	.dc.b	$D7,$E8,$AF,$4F,$CB,$7B,$AD,$4B,$F2,$7C,$FB,$1E,$2A,$FF,$24,$69
	.dc.b	$FA,$B3,$FD,$C5,$BF,$F9,$93,$EC,$BF,$D2,$38,$F7,$FC,$5F,$C7,$18
	.dc.b	$D3,$F2,$97,$F1,$3D,$7F,$F5,$FF,$CB,$BF,$39,$FF,$34,$60,$FD,$6F
	.dc.b	$5F,$01,$FD,$56,$73,$F0,$5C,$7E,$0B,$E2,$E3,$FE,$C8,$CB,$E4,$BD
	.dc.b	$2B,$9C,$72,$CB,$B7,$E0,$3F,$E8,$5E,$1F,$F6,$EF,$03,$F7,$2F,$F6
	.dc.b	$DF,$E9,$5F,$E8,$0F,$C9,$7F,$C3,$DF,$7D,$F6,$A7,$3C,$69,$FF,$11
	.dc.b	$7E,$C3,$9F,$A4,$FA,$F5,$FE,$E4,$D7,$ED,$BC,$2B,$F9,$EE,$7E,$08
	.dc.b	$C7,$CC,$7E,$2D,$7E,$78,$FD,$8B,$FC,$19,$D7,$FE,$94,$FD,$AF,$C5
	.dc.b	$39,$FA,$CF,$AD,$F8,$4F,$F8,$1B,$E6,$7F,$B3,$71,$FA,$FF,$4A,$F1
	.dc.b	$DC,$CB,$9D,$F7,$8E,$59,$3D,$0F,$17,$E4,$8F,$D0,$9F,$B9,$BF,$43
	.dc.b	$FE,$0B,$F9,$67,$C3,$AD,$4E,$7D,$E7,$D9,$7C,$7F,$CD,$7E,$1D,$4C
	.dc.b	$79,$E3,$FE,$92,$E6,$F8,$53,$CA,$3F,$D3,$5C,$5E,$AE,$57,$B4,$FA
	.dc.b	$77,$E9,$9F,$E5,$3F,$C4,$DF,$99,$FD,$5A,$D4,$E7,$E4,$F1,$7C,$3E
	.dc.b	$EB,$FD,$17,$DF,$8A,$FF,$2E,$7E,$37,$F4,$7E,$FD,$F9,$F3,$F0,$F7
	.dc.b	$A2,$FD,$D1,$F0,$DF,$83,$BD,$D6,$A7,$F8,$97,$17,$A6,$73,$7C,$B3
	.dc.b	$E1,$8F,$58,$FC,$94,$7E,$AB,$8B,$F8,$DF,$DE,$BE,$CF,$F4,$9F,$22
	.dc.b	$3F,$25,$F1,$A9,$D7,$A5,$9F,$A7,$CC,$62,$BF,$19,$5B,$BF,$D7,$7C
	.dc.b	$66,$07,$ED,$5F,$EF,$AF,$D2,$5F,$27,$DA,$73,$F0,$89,$FE,$9F,$C5
	.dc.b	$F5,$71,$FE,$65,$82,$BF,$15,$BF,$18,$C7,$95,$E2,$F2,$FF,$9C,$7E
	.dc.b	$CF,$FB,$DE,$AF,$5C,$63,$DE,$FE,$A5,$9F,$58,$DF,$5F,$1D,$98,$7F
	.dc.b	$41,$CF,$E1,$FC,$67,$AB,$EC,$1F,$7A,$FD,$BB,$3B,$FF,$8E,$E3,$F9
	.dc.b	$AE,$9E,$9B,$E1,$F7,$FF,$19,$E5,$D9,$F2,$3F,$5A,$FF,$6A,$FE,$D2
	.dc.b	$F9,$27,$E8,$5F,$4A,$73,$1B,$FC,$55,$4E,$6C,$3F,$33,$81,$FA,$AF
	.dc.b	$C9,$FD,$91,$FB,$AB,$ED,$5F,$AB,$BF,$5F,$DF,$A9,$FC,$1F,$BD,$87
	.dc.b	$0B,$DF,$EC,$DA,$CF,$F1,$55,$AC,$BA,$65,$FC,$EF,$3F,$04,$63,$E5
	.dc.b	$38,$5B,$E0,$7F,$91,$61,$C7,$2F,$1F,$B9,$A9,$CE,$9E,$1C,$DF,$90
	.dc.b	$67,$CC,$FE,$E7,$39,$F4,$31,$97,$3B,$7E,$7B,$8A,$E5,$F6,$17,$7F
	.dc.b	$3F,$CB,$DD,$9F,$91,$BF,$42,$C2,$FD,$12,$FD,$3E,$3D,$8C,$FD,$96
	.dc.b	$BD,$AE,$5F,$27,$FE,$35,$E9,$F8,$39,$A3,$FD,$B9,$18,$FD,$CF,$5D
	.dc.b	$9F,$3F,$87,$E7,$46,$A3,$1C,$1F,$5F,$F0,$47,$5F,$9F,$F0,$2E,$2F
	.dc.b	$D2,$C3,$EB,$47,$18,$C5,$71,$FC,$67,$DD,$6F,$C7,$FE,$07,$C7,$F2
	.dc.b	$BF,$A9,$D9,$F3,$30,$FC,$34,$F3,$9C,$F5,$FD,$7F,$C1,$FC,$7B,$FB
	.dc.b	$B5,$E5,$DF,$F6,$31,$FC,$74,$71,$8C,$71,$3E,$93,$8F,$0F,$BA,$3F
	.dc.b	$C1,$F9,$FD,$3F,$C7,$F9,$ED,$E3,$94,$73,$8C,$7A,$BC,$AB,$DF,$4F
	.dc.b	$5B,$11,$E8,$CF,$F3,$5F,$7B,$D3,$83,$F2,$0F,$37,$F2,$5F,$63,$C3
	.dc.b	$6F,$3F,$EB,$17,$4F,$89,$FB,$7F,$EA,$3B,$3F,$14,$FE,$AB,$B2,$F6
	.dc.b	$6F,$2F,$D2,$4F,$EC,$F9,$7F,$47,$FB,$37,$16,$F8,$DE,$1D,$FF,$6B
	.dc.b	$3F,$87,$E0,$FC,$B3,$C7,$97,$F0,$7F,$2B,$C5,$B7,$C7,$A1,$3C,$71
	.dc.b	$FD,$8B,$C7,$0F,$C1,$FA,$DC,$5B,$7C,$63,$86,$1F,$90,$78,$4F,$C9
	.dc.b	$C3,$FA,$BD,$7C,$9C,$BF,$85,$FA,$BF,$07,$2F,$DF,$BC,$AF,$F2,$2F
	.dc.b	$2E,$3F,$C2,$71,$F8,$67,$A3,$C7,$3F,$CF,$BC,$63,$3F,$05,$F2,$F1
	.dc.b	$FE,$33,$F0,$7C,$70,$FC,$73,$C6,$BF,$3E,$F1,$8C,$D7,$1F,$D6,$DF
	.dc.b	$39,$C7,$9F,$F8,$4C,$63,$6F,$A4,$EB,$97,$85,$7A,$6B,$7D,$7C,$79
	.dc.b	$BF,$92,$DF,$F2,$0E,$7F,$E1,$CF,$F2,$DC,$3C,$15,$C6,$BF,$95,$FE
	.dc.b	$BB,$FC,$9B,$A7,$F8,$A3,$C7,$93,$F9,$B3,$C7,$83,$37,$C7,$85,$FB
	.dc.b	$D7,$1B,$FE,$78,$F1,$E4,$FE,$14,$F1,$8C,$D7,$1F,$8E,$E1,$5F,$C3
	.dc.b	$DE,$18,$7F,$09,$78,$46,$67,$C1,$F2,$9C,$2B,$F8,$73,$C1,$FD,$63
	.dc.b	$C1,$CC,$79,$FD,$DE,$11,$FB,$17,$87,$2F,$F3,$C7,$84,$66,$7E,$37
	.dc.b	$BA,$FD,$6B,$8D,$FF,$46,$8C,$4F,$DF,$AF,$E2,$71,$8E,$1F,$D8,$3A
	.dc.b	$E7,$AF,$2E,$5F,$F5,$CA,$3F,$AB,$FE,$21,$FD,$21,$C6,$1F,$89,$9D
	.dc.b	$C7,$F6,$7A,$E7,$3E,$1C,$7C,$9D,$9F,$D8,$3F,$6F,$E4,$F4,$E1,$FF
	.dc.b	$0C,$61,$FD,$67,$DC,$CE,$1F,$DA,$A7,$FC,$39,$E4,$7F,$5D,$E7,$FE
	.dc.b	$B2,$FF,$5F,$EA,$B8,$FC,$6C,$FF,$B3,$CF,$F7,$3D,$3F,$87,$7C,$0F
	.dc.b	$1C,$7F,$C1,$D7,$F3,$B6,$FF,$6B,$CF,$F7,$9E,$0D,$CF,$AF,$87,$FB
	.dc.b	$3F,$F0,$07,$1E,$6F,$F0,$03,$E1,$C7,$FC,$60,$F1,$9D,$D7,$19,$FF
	.dc.b	$1A,$7E,$61,$C6,$7F,$C0,$6F,$1B,$FF,$05,$3C,$67,$75,$C6,$7F,$C1
	.dc.b	$DE,$5F,$1B,$FE,$9A,$F1,$7F,$C1,$AF,$18,$DC,$F1,$9F,$F0,$8F,$EE
	.dc.b	$1C,$77,$FF,$05,$BC,$63,$FA,$BB,$C5,$DC,$F1,$7F,$AD,$FE,$65,$FE
	.dc.b	$24,$FD,$95,$E6,$7F,$45,$8C,$3F,$D1,$5D,$79,$9F,$CF,$5E,$1E,$3D
	.dc.b	$C1,$69,$E5,$18,$79,$3A,$F1,$38,$4F,$F9,$C7,$1F,$F0,$CD,$78,$FF
	.dc.b	$4C,$E3,$27,$8B,$FE,$02,$9D,$DF,$83,$F0,$8F,$17,$53,$F2,$0F,$0B
	.dc.b	$7A,$B6,$BE,$FB,$7E,$4F,$0F,$2B,$67,$9A,$7F,$45,$E5,$7A,$F2,$F9
	.dc.b	$1C,$9F,$E3,$8D,$FF,$C8,$9E,$4F,$63,$87,$7F,$8B,$EE,$BC,$7D,$3E
	.dc.b	$23,$A7,$A7,$E6,$F4,$FC,$FD,$5E,$63,$C1,$E8,$7A,$B6,$F5,$F5,$FA
	.dc.b	$8F,$27,$5F,$BB,$AF,$FB,$A7,$CE,$7F,$94,$7E,$BB,$F9,$E3,$79,$3F
	.dc.b	$C0,$CE,$7C,$2D,$F0,$FF,$98,$AB,$F9,$E4,$F1,$C7,$FC,$1E,$DE,$8F
	.dc.b	$F2,$73,$D7,$2E,$3C,$AB,$67,$8E,$CF,$0E,$5F,$8B,$D3,$C2,$FE,$37
	.dc.b	$4D,$78,$BB,$D7,$BC,$FE,$3B,$5E,$EF,$E3,$BB,$6E,$9F,$8C,$CF,$C3
	.dc.b	$BF,$68,$FD,$F6,$FF,$E6,$9C,$F3,$9E,$B3,$EB,$5F,$E7,$F7,$C6,$FF
	.dc.b	$A0,$77,$E7,$8F,$B3,$D5,$E8,$7F,$1D,$F1,$99,$7F,$86,$B2,$D6,$5B
	.dc.b	$CB,$AD,$FE,$A3,$6F,$3A,$CB,$A6,$5F,$23,$D5,$F4,$1D,$72,$FF,$44
	.dc.b	$3F,$CB,$DB,$8F,$F8,$BF,$EA,$3F,$CF,$5E,$4F,$F9,$FD,$FF,$3E,$65
	.dc.b	$78,$3F,$44,$6F,$87,$F7,$CF,$F9,$6A,$7F,$D1,$97,$FE,$39,$FA,$1E
	.dc.b	$7E,$3F,$FB,$FD,$FF,$43,$79,$DF,$E8,$B7,$E5,$5D,$C3,$F9,$73,$FE
	.dc.b	$8F,$FD,$83,$FD,$19,$E7,$D9,$FF,$4A,$7E,$15,$FE,$93,$FD,$83,$C7
	.dc.b	$7F,$E4,$7B,$AE,$F9,$FF,$9D,$73,$A7,$CB,$3F,$E9,$BF,$D4,$FF,$D3
	.dc.b	$1F,$AB,$5C,$BF,$D4,$1F,$85,$7F,$A7,$FE,$7B,$9F,$03,$D0,$FE,$05
	.dc.b	$E4,$B4,$65,$A9,$CC,$63,$1F,$8B,$5B,$AF,$0D,$F3,$CF,$FD,$1B,$3C
	.dc.b	$F8,$DF,$1C,$CD,$F3,$6F,$B6,$7E,$D5,$FE,$B3,$BD,$D6,$A5,$F9,$4E
	.dc.b	$72,$F8,$B9,$F7,$7F,$D1,$F9,$F9,$DF,$0E,$CC,$79,$E7,$FE,$D7,$9E
	.dc.b	$DA,$78,$E1,$F8,$6F,$F9,$D7,$E2,$BF,$D8,$1F,$85,$7F,$B1,$5F,$F6
	.dc.b	$17,$A9,$FE,$C7,$FD,$93,$B6,$3F,$E9,$E9,$BA,$7C,$97,$F0,$3A,$F9
	.dc.b	$63,$1C,$7F,$82,$8C,$6C,$F6,$FF,$DB,$1F,$5A,$FD,$33,$FD,$A9,$18
	.dc.b	$FC,$13,$FD,$AD,$F9,$DF,$6B,$E1,$A7,$C2,$E9,$FA,$B7,$F4,$9B,$9F
	.dc.b	$D6,$A7,$ED,$7F,$EF,$2F,$83,$FF,$7C,$7D,$0B,$8F,$FD,$C5,$39,$F4
	.dc.b	$13,$FC,$F7,$83,$CF,$F9,$6F,$86,$A6,$5F,$E1,$EE,$6F,$F4,$E4,$71
	.dc.b	$C2,$CB,$F4,$54,$3F,$7B,$F0,$63,$F0,$AD,$7B,$AD,$74,$79,$23,$1E
	.dc.b	$E7,$4C,$79,$FB,$BD,$BE,$FF,$AF,$E2,$BD,$38,$FF,$2F,$9F,$49,$CE
	.dc.b	$5C,$DF,$1F,$D3,$71,$D3,$FC,$E9,$5E,$2F,$CC,$3C,$FF,$B2,$73,$E8
	.dc.b	$FC,$5C,$E6,$93,$F1,$72,$EF,$BB,$0F,$7F,$C2,$FC,$27,$E0,$E9,$F1
	.dc.b	$3C,$CF,$87,$F9,$EF,$5D,$3F,$DC,$FA,$7C,$29,$CD,$7C,$17,$E4,$FE
	.dc.b	$29,$C7,$5F,$F4,$8D,$FC,$5F,$90,$F3,$DD,$9F,$2F,$55,$D3,$E9,$D7
	.dc.b	$8A,$FE,$FF,$5F,$F9,$8B,$8F,$D5,$1E,$E8,$C5,$EB,$E4,$3E,$3C,$DC
	.dc.b	$7E,$3D,$3E,$87,$EE,$F5,$72,$AF,$5C,$E6,$BF,$01,$87,$82,$FF,$07
	.dc.b	$D3,$F8,$8A,$D7,$45,$7F,$2B,$5E,$67,$F2,$FB,$5C,$3F,$6F,$8D,$7F
	.dc.b	$DD,$DF,$85,$3E,$2F,$F1,$DC,$4F,$75,$F6,$3F,$AD,$EE,$FD,$A5,$EF
	.dc.b	$83,$FD,$CF,$2F,$15,$5A,$FC,$87,$F7,$B7,$D7,$4F,$D8,$1F,$C7,$9E
	.dc.b	$59,$77,$FB,$BF,$1E,$9E,$2D,$1F,$8E,$63,$F7,$6F,$F4,$47,$C2,$BF
	.dc.b	$57,$7E,$89,$E0,$F0,$BF,$99,$E9,$EB,$BD,$D3,$E3,$F5,$B7,$87,$45
	.dc.b	$74,$F8,$3E,$DC,$ED,$1E,$37,$87,$87,$FC,$9B,$CB,$E0,$E4,$F8,$79
	.dc.b	$72,$9C,$E9,$DB,$47,$A2,$DF,$C9,$5F,$E5,$6E,$4E,$9F,$19,$EE,$BE
	.dc.b	$17,$BC,$7F,$88,$4F,$BF,$1C,$E3,$F5,$FC,$7F,$F5,$54,$FC,$3F,$D7
	.dc.b	$F3,$7C,$38,$7E,$5D,$6B,$7B,$AF,$F8,$A7,$7F,$B9,$C9,$F2,$30,$FF
	.dc.b	$1D,$47,$DB,$E0,$F6,$E1,$FE,$37,$E0,$E3,$F1,$DE,$79,$F8,$3C,$7E
	.dc.b	$2E,$57,$E0,$7E,$28,$C7,$5D,$7F,$CE,$BB,$FD,$E7,$C9,$CB,$BD,$FE
	.dc.b	$6F,$E4,$17,$8F,$F2,$3F,$99,$5D,$B8,$57,$F9,$FB,$AF,$F0,$FE,$17
	.dc.b	$DE,$C3,$E0,$BF,$80,$D6,$E7,$F6,$F4,$DB,$F7,$79,$78,$E3,$F9,$FE
	.dc.b	$3F,$87,$B7,$E1,$71,$FC,$EE,$FD,$2B,$F2,$D3,$7B,$7E,$17,$37,$F8
	.dc.b	$0B,$F2,$5F,$A9,$4F,$D1,$FF,$A1,$F8,$9F,$72,$73,$1F,$D1,$BC,$4C
	.dc.b	$78,$5F,$D8,$A7,$F4,$D9,$7F,$8B,$3A,$3F,$B1,$47,$0F,$49,$3C,$FB
	.dc.b	$3F,$90,$63,$FD,$D6,$FF,$B4,$7E,$8D,$E2,$FA,$DF,$DC,$FB,$5F,$8C
	.dc.b	$8C,$71,$FD,$DE,$0F,$F0,$CC,$35,$FE,$E6,$F8,$B3,$FF,$4F,$63,$D3
	.dc.b	$3F,$E5,$D9,$3A,$BC,$73,$9E,$F7,$12,$BE,$2E,$1F,$E1,$CF,$D9,$FF
	.dc.b	$31,$F6,$AF,$DC,$BF,$B3,$FE,$63,$FA,$09,$CC,$7F,$6A,$FD,$6F,$1B
	.dc.b	$FF,$81,$A7,$C9,$87,$83,$83,$FC,$63,$1C,$39,$31,$3D,$F8,$3F,$B2
	.dc.b	$E7,$DF,$CD,$CC,$FF,$87,$70,$FD,$57,$C9,$79,$F9,$1F,$B3,$FE,$E7
	.dc.b	$E1,$FF,$C7,$D9,$7F,$6E,$AD,$7E,$C3,$9C,$3E,$46,$3F,$C7,$0E,$95
	.dc.b	$FE,$1E,$BE,$3C,$8C,$3E,$27,$37,$8A,$73,$1F,$1F,$C5,$C7,$E6,$9F
	.dc.b	$BD,$F5,$BE,$37,$80,$FC,$33,$C9,$39,$E7,$FE,$6D,$FA,$BE,$39,$FC
	.dc.b	$13,$F1,$B7,$DE,$FF,$E0,$6C,$BE,$14,$63,$74,$7D,$3F,$A0,$F2,$7E
	.dc.b	$07,$FD,$0F,$F0,$77,$E5,$CF,$C2,$1F,$2F,$7C,$2F,$F2,$57,$2B,$0E
	.dc.b	$17,$BA,$FC,$15,$6A,$7E,$3E,$9F,$9D,$E7,$FF,$75,$61,$C7,$6F,$91
	.dc.b	$5C,$B2,$FF,$5E,$DF,$D3,$8C,$7E,$CF,$F2,$7F,$4A,$FF,$1A,$FE,$5F
	.dc.b	$EA,$E6,$7E,$EA,$FC,$1F,$ED,$56,$B8,$FE,$54,$BF,$71,$CF,$C4,$7D
	.dc.b	$7A,$7C,$75,$86,$5D,$7E,$EB,$C1,$18,$C7,$EA,$7E,$13,$E3,$FD,$D1
	.dc.b	$F9,$37,$F8,$A7,$EF,$1F,$8B,$F8,$A7,$3D,$AF,$93,$53,$F2,$CF,$3F
	.dc.b	$E6,$FE,$3B,$98,$FF,$8C,$FE,$EB,$B3,$FE,$35,$F2,$7E,$17,$E8,$EF
	.dc.b	$DE,$9F,$A1,$BF,$23,$7C,$A7,$F5,$2F,$0F,$EB,$F4,$30,$E1,$7B,$AF
	.dc.b	$F4,$05,$6B,$F5,$AF,$89,$53,$1E,$B8,$BD,$FF,$0C,$63,$E5,$FF,$39
	.dc.b	$FB,$AB,$F6,$67,$EC,$4F,$D2,$9F,$BB,$FF,$99,$BF,$10,$FD,$75,$EE
	.dc.b	$B5,$3F,$E6,$AF,$CE,$B3,$FA,$3F,$EA,$EB,$A7,$95,$FA,$FF,$35,$1C
	.dc.b	$72,$EB,$EF,$F8,$B4,$F0,$C7,$83,$DE,$F2,$F9,$FF,$A0,$FB,$07,$ED
	.dc.b	$2F,$B6,$7E,$1F,$E4,$AD,$4E,$7F,$7F,$F3,$FB,$7E,$39,$9E,$57,$FA
	.dc.b	$D6,$BF,$C0,$FC,$1C,$63,$1F,$27,$F1,$3E,$BF,$E1,$FE,$2B,$F0,$BF
	.dc.b	$26,$7E,$53,$F0,$A7,$3D,$09,$7E,$57,$F1,$BE,$D3,$E1,$54,$EB,$63
	.dc.b	$F3,$3E,$F7,$B4,$63,$17,$E0,$9F,$53,$ED,$3F,$0D,$FA,$6F,$E1,$6B
	.dc.b	$FA,$B7,$8F,$FB,$AE,$DF,$F5,$67,$86,$AD,$7E,$F3,$F5,$BE,$DF,$E9
	.dc.b	$63,$2B,$5E,$3F,$FB,$1F,$9B,$ED,$CF,$68,$7E,$11,$F8,$4F,$8A,$FD
	.dc.b	$2F,$ED,$FF,$A1,$FA,$A7,$E5,$5F,$82,$9C,$FB,$3F,$E3,$7F,$0B,$F0
	.dc.b	$BF,$AA,$7E,$06,$67,$AA,$C7,$96,$BC,$63,$18,$3D,$4F,$C9,$FB,$DF
	.dc.b	$B7,$C1,$FB,$BB,$F6,$4F,$C8,$CE,$61,$F8,$8E,$32,$FC,$77,$3F,$F0
	.dc.b	$46,$D7,$3F,$C4,$7A,$FC,$27,$F2,$AC,$9F,$97,$7F,$3D,$FC,$6B,$FC
	.dc.b	$AD,$FB,$77,$7F,$94,$EB,$F1,$FF,$D3,$67,$3B,$A7,$C3,$F5,$59,$F2
	.dc.b	$FF,$A5,$57,$F9,$7F,$6F,$D0,$3C,$29,$AF,$7F,$B0,$67,$DF,$AB,$C5
	.dc.b	$E8,$BF,$22,$FD,$C7,$EF,$BF,$E1,$5F,$D0,$7C,$53,$9F,$80,$4F,$F8
	.dc.b	$67,$D9,$F3,$73,$78,$AA,$61,$CB,$DB,$EA,$F4,$E5,$FE,$B9,$F3,$BF
	.dc.b	$E8,$BF,$C9,$9D,$0F,$B3,$78,$4F,$8B,$FE,$DD,$7B,$AD,$4B,$F4,$BC
	.dc.b	$FB,$9F,$DA,$EB,$FA,$E7,$9F,$FD,$41,$E1,$8F,$D0,$EA,$E7,$EE,$34
	.dc.b	$F0,$CF,$C6,$9E,$FF,$26,$FC,$9F,$F9,$87,$E0,$CF,$CB,$BF,$8C,$67
	.dc.b	$D6,$B5,$F2,$3F,$EF,$39,$7C,$3F,$A3,$F6,$2F,$25,$4C,$79,$FB,$3F
	.dc.b	$0A,$38,$E4,$F6,$FF,$C0,$1F,$85,$7F,$8B,$FC,$1F,$F2,$37,$D8,$BF
	.dc.b	$42,$F8,$BB,$FF,$6C,$AD,$7D,$9A,$B5,$FA,$FF,$F8,$56,$5F,$B0,$7F
	.dc.b	$83,$DF,$ED,$BC,$7F,$E3,$8F,$A4,$FE,$DB,$1C,$AD,$F8,$4F,$8B,$AB
	.dc.b	$F0,$F9,$75,$F9,$3F,$2F,$0B,$F3,$77,$ED,$2F,$83,$7E,$6E,$FD,$1F
	.dc.b	$C5,$7B,$AD,$4F,$F8,$23,$8B,$D5,$D3,$E5,$A9,$8F,$4C,$7F,$D6,$9C
	.dc.b	$3C,$A3,$1E,$87,$F8,$97,$F0,$5F,$F6,$A7,$CA,$FE,$2B,$F0,$97,$E2
	.dc.b	$8F,$DE,$FF,$C6,$D5,$AF,$27,$FD,$87,$2F,$90,$CF,$2F,$F8,$CE,$BF
	.dc.b	$13,$D3,$FE,$10,$78,$62,$FA,$7F,$07,$91,$FD,$67,$2F,$F6,$6F,$D1
	.dc.b	$78,$FD,$17,$C9,$23,$FD,$99,$F8,$B6,$63,$1E,$CA,$1F,$88,$79,$BD
	.dc.b	$6F,$16,$12,$F7,$E4,$F1,$C7,$FD,$AD,$F9,$0F,$F9,$17,$F0,$2F,$E0
	.dc.b	$BE,$63,$E0,$1F,$79,$FE,$55,$9C,$C7,$C4,$F0,$3F,$C6,$71,$BE,$DF
	.dc.b	$EC,$CF,$F0,$97,$FC,$60,$7F,$A2,$BE,$81,$1C,$A3,$A7,$C0,$7F,$0A
	.dc.b	$F8,$4F,$D3,$7C,$07,$F9,$A3,$F1,$67,$9A,$FC,$F3,$FC,$4B,$F9,$EB
	.dc.b	$F5,$57,$EB,$4F,$C4,$3F,$DC,$58,$70,$BD,$D6,$BD,$2C,$E3,$CF,$5F
	.dc.b	$F3,$55,$F6,$FD,$3B,$FD,$3F,$1E,$81,$1C,$FE,$35,$A7,$87,$4F,$36
	.dc.b	$1E,$59,$CE,$4F,$C3,$3F,$D0,$3F,$B7,$3E,$09,$FA,$CF,$F6,$AF,$9D
	.dc.b	$FE,$E1,$FB,$2E,$FA,$1E,$57,$8B,$0E,$17,$BF,$A5,$5B,$F6,$1F,$F8
	.dc.b	$01,$5D,$7E,$E3,$D3,$C7,$FB,$33,$FD,$73,$1C,$9F,$F4,$43,$D3,$E9
	.dc.b	$BF,$C7,$5F,$6B,$B9,$4E,$75,$FF,$02,$7E,$D7,$FE,$9A,$FC,$71,$FA
	.dc.b	$E7,$F5,$BF,$92,$7E,$13,$FE,$62,$FD,$E9,$FA,$CB,$93,$D7,$A7,$FB
	.dc.b	$2B,$1E,$3E,$17,$F4,$5C,$38,$7D,$66,$F9,$BE,$8D,$76,$FA,$4F,$5E
	.dc.b	$1E,$1D,$7F,$B5,$3F,$EA,$19,$E7,$AF,$8B,$F0,$0F,$F3,$CE,$9E,$0A
	.dc.b	$56,$B3,$FF,$3B,$F9,$DF,$EB,$4F,$75,$FC,$02,$FC,$FF,$FB,$27,$E7
	.dc.b	$CF,$CE,$3F,$CB,$BF,$88,$F9,$73,$FF,$71,$E1,$C2,$F7,$5F,$E6,$2F
	.dc.b	$C0,$75,$5F,$C1,$F1,$E8,$FF,$AD,$EF,$BF,$5F,$B2,$37,$1C,$FE,$8D
	.dc.b	$9F,$87,$C3,$F3,$61,$E5,$9C,$E5,$FA,$0C,$38,$FE,$88,$FD,$59,$F3
	.dc.b	$0F,$C6,$5E,$73,$F5,$47,$E7,$3C,$2F,$75,$A9,$7E,$A5,$D3,$F1,$2F
	.dc.b	$37,$4F,$CB,$E2,$69,$DF,$5F,$63,$E1,$D7,$C1,$AB,$F0,$FF,$9D,$39
	.dc.b	$D3,$C7,$EF,$63,$6F,$0D,$F9,$7F,$1C,$F8,$7F,$B9,$BF,$5D,$7E,$41
	.dc.b	$FE,$97,$FD,$81,$F9,$17,$F8,$BF,$87,$C9,$7B,$AD,$4B,$F6,$7E,$7F
	.dc.b	$8D,$7A,$BC,$8F,$DE,$F2,$30,$ED,$C9,$F5,$5E,$5F,$52,$D3,$F7,$53
	.dc.b	$E4,$D7,$C7,$39,$FC,$B3,$F1,$57,$C2,$FC,$BF,$71,$F8,$2F,$97,$FF
	.dc.b	$4A,$7E,$47,$FD,$AF,$D1,$7C,$93,$F2,$47,$DA,$6F,$E7,$17,$BA,$D7
	.dc.b	$EC,$39,$E7,$F5,$70,$8F,$EA,$17,$C1,$FD,$3C,$7E,$87,$3F,$0F,$C3
	.dc.b	$B4,$F0,$FC,$57,$1A,$D6,$1E,$0A,$D6,$5F,$CE,$E7,$97,$E4,$AF,$C8
	.dc.b	$3F,$C6,$FF,$99,$FF,$53,$F5,$1F,$2F,$EF,$AD,$7D,$D2,$B5,$F4,$B9
	.dc.b	$C3,$D3,$C1,$FD,$9E,$DF,$A5,$63,$85,$1C,$B0,$FE,$C1,$F8,$3F,$9F
	.dc.b	$BB,$F9,$AD,$6B,$1F,$1F,$CC,$FF,$A0,$F7,$F8,$17,$FE,$96,$FD,$6B
	.dc.b	$FC,$E1,$97,$F3,$AF,$4F,$EA,$4E,$7E,$07,$BF,$C9,$7E,$97,$0E,$3E
	.dc.b	$C7,$8E,$9F,$AC,$D8,$65,$D3,$E6,$3C,$9C,$5F,$4F,$7F,$F4,$8D,$6B
	.dc.b	$3F,$5E,$3A,$AF,$F0,$37,$E6,$1F,$E4,$1F,$C6,$9F,$A9,$7F,$00,$FD
	.dc.b	$DB,$FC,$DB,$B7,$D9,$AD,$4E,$7D,$6F,$C6,$47,$F9,$27,$0F,$BF,$CB
	.dc.b	$F1,$23,$FD,$63,$18,$3C,$8F,$2F,$79,$FA,$37,$8F,$74,$E6,$B1,$F6
	.dc.b	$1F,$DC,$3F,$01,$7E,$65,$FE,$D7,$FD,$65,$FB,$CF,$E7,$6B,$53,$9F
	.dc.b	$DB,$71,$F3,$BF,$6E,$7F,$B0,$63,$E3,$3F,$C4,$CF,$0F,$C4,$7F,$D2
	.dc.b	$51,$C3,$2E,$F1,$FE,$E4,$8F,$D5,$7E,$5B,$FD,$3B,$A3,$FC,$85,$F3
	.dc.b	$FF,$E5,$9F,$C1,$73,$F1,$FF,$ED,$4F,$BA,$4F,$F9,$77,$F6,$DE,$3F
	.dc.b	$98,$7E,$37,$1F,$EA,$F9,$78,$F9,$11,$D7,$0E,$3F,$7F,$F1,$63,$18
	.dc.b	$BE,$FB,$E4,$7A,$7F,$D8,$B8,$5F,$97,$BF,$42,$7E,$4E,$FE,$01,$FE
	.dc.b	$05,$7B,$AD,$4F,$FB,$13,$C1,$FD,$B4,$FE,$B7,$7F,$23,$FD,$B9,$BF
	.dc.b	$37,$FE,$62,$BE,$DE,$C7,$8A,$31,$97,$96,$BF,$88,$78,$6E,$B7,$EF
	.dc.b	$5F,$E7,$2D,$FF,$D2,$1F,$AD,$EA,$73,$C7,$ED,$FC,$31,$FA,$BE,$72
	.dc.b	$F8,$D9,$7E,$3D,$FF,$12,$73,$FE,$63,$0E,$39,$7E,$43,$87,$F8,$85
	.dc.b	$F8,$74,$F0,$7E,$19,$F8,$FF,$D9,$DF,$A4,$3F,$65,$FF,$15,$E7,$FC
	.dc.b	$37,$E5,$FF,$94,$56,$A7,$3E,$C7,$F5,$DC,$FF,$8E,$4F,$2D,$BE,$D3
	.dc.b	$FE,$4F,$7F,$C3,$58,$23,$97,$D9,$FF,$9D,$7E,$C7,$C5,$9F,$86,$3F
	.dc.b	$CF,$1F,$41,$FB,$5F,$BB,$FF,$27,$4B,$E3,$BF,$CE,$1F,$AD,$BF,$60
	.dc.b	$7E,$9F,$FD,$FA,$F7,$5A,$97,$E8,$B9,$FB,$9F,$26,$FD,$5F,$43,$FE
	.dc.b	$4D,$BE,$3F,$04,$C9,$F2,$7B,$D3,$C5,$39,$F5,$FC,$DF,$5A,$FB,$27
	.dc.b	$EA,$5F,$D7,$FF,$6A,$7E,$AC,$FD,$27,$CD,$7B,$AD,$4F,$F9,$AF,$A7
	.dc.b	$FC,$4D,$B2,$BF,$DC,$DB,$FE,$DE,$31,$97,$F3,$86,$E1,$DF,$ED,$FC
	.dc.b	$73,$E2,$F9,$E6,$9F,$ED,$1C,$3C,$33,$98,$F2,$FC,$D7,$C9,$F9,$F7
	.dc.b	$EC,$5F,$DB,$FE,$73,$D1,$F8,$67,$FB,$83,$7F,$2D,$6A,$D5,$AF,$99
	.dc.b	$E5,$F0,$DE,$5A,$E7,$A7,$A5,$E1,$5F,$0F,$C0,$FB,$D1,$8E,$4F,$E4
	.dc.b	$75,$F6,$FF,$7A,$7E,$EC,$FB,$37,$DB,$BD,$2F,$EE,$BF,$8F,$6E,$B5
	.dc.b	$39,$FA,$07,$E1,$99,$CB,$E5,$64,$FD,$6F,$E5,$3D,$B3,$EB,$9B,$E4
	.dc.b	$BF,$25,$39,$CB,$C1,$F9,$CF,$E5,$7E,$75,$FB,$AB,$F5,$4F,$F7,$26
	.dc.b	$1F,$E3,$DF,$DD,$3F,$29,$5A,$9C,$FD,$B7,$E6,$7A,$BE,$FC,$BF,$41
	.dc.b	$FB,$8F,$4E,$0F,$C0,$3C,$7E,$27,$C9,$F6,$DE,$1D,$3C,$11,$E7,$FB
	.dc.b	$4F,$D4,$7C,$12,$FF,$29,$DA,$FD,$85,$F8,$3B,$E4,$BF,$51,$7B,$F1
	.dc.b	$BF,$D3,$F5,$AF,$A1,$CE,$3F,$8F,$A7,$E3,$96,$3A,$D7,$E3,$78,$7F
	.dc.b	$33,$5C,$F3,$F0,$7E,$01,$FB,$BF,$89,$7D,$73,$7F,$F3,$0F,$DC,$3F
	.dc.b	$68,$FF,$26,$E5,$FA,$2B,$DD,$6A,$5F,$A6,$7E,$FE,$79,$F3,$7E,$95
	.dc.b	$FF,$1C,$B7,$1E,$DF,$7D,$FE,$73,$FC,$07,$C5,$18,$CF,$FC,$F1,$9F
	.dc.b	$78,$C6,$4F,$B3,$FF,$44,$FE,$3A,$FD,$49,$FA,$B7,$FA,$AF,$4F,$E7
	.dc.b	$7F,$A4,$FF,$3D,$CB,$FD,$3D,$7B,$AD,$7C,$72,$9F,$27,$D3,$17,$E8
	.dc.b	$3F,$E3,$2C,$38,$65,$FD,$D5,$E5,$97,$F9,$37,$17,$E5,$9F,$E1,$99
	.dc.b	$CD,$7F,$4C,$FC,$8B,$18,$F9,$B0,$E1,$9F,$F8,$19,$F2,$FE,$CA,$FD
	.dc.b	$0D,$EF,$5F,$B9,$7D,$04,$BD,$D6,$84,$EB,$CE,$F8,$B8,$73,$D3,$D0
	.dc.b	$BF,$51,$F6,$3D,$F2,$ED,$F7,$BE,$98,$C6,$3D,$36,$F6,$7E,$27,$FC
	.dc.b	$E3,$F3,$97,$E9,$BF,$EF,$2F,$CF,$9F,$90,$7D,$4A,$D4,$E7,$F6,$CC
	.dc.b	$7C,$B7,$CB,$9F,$E8,$3B,$FC,$B7,$7E,$37,$E0,$0F,$F3,$C4,$39,$FF
	.dc.b	$DF,$5F,$28,$D3,$C3,$5F,$E9,$0E,$5F,$BB,$FB,$AF,$F6,$7E,$34,$FC
	.dc.b	$4F,$C4,$BF,$C3,$35,$AF,$9E,$F9,$B3,$9F,$9B,$FC,$5E,$1F,$DB,$F4
	.dc.b	$FB,$CB,$F7,$FF,$C1,$3D,$6B,$A4,$F0,$BE,$7D,$5F,$3F,$1F,$C7,$FB
	.dc.b	$2F,$D8,$7F,$A4,$7E,$CE,$F9,$E7,$E7,$BF,$ED,$0B,$DD,$6A,$73,$FB
	.dc.b	$6F,$E7,$B0,$FE,$21,$9F,$C4,$31,$5F,$95,$8E,$DF,$2E,$CB,$B7,$67
	.dc.b	$E4,$E7,$31,$E1,$EF,$FC,$B7,$E4,$EF,$C8,$1F,$BF,$FE,$83,$F6,$FF
	.dc.b	$F7,$27,$17,$EB,$AB,$53,$9F,$C6,$15,$9B,$FD,$35,$7F,$4C,$DB,$CD
	.dc.b	$C3,$FB,$08,$E3,$E6,$70,$FC,$0B,$C3,$7E,$08,$E7,$F2,$9F,$E8,$1F
	.dc.b	$8A,$7C,$B7,$FB,$73,$F2,$C7,$E1,$4F,$DA,$5F,$AD,$FE,$17,$5F,$BF
	.dc.b	$B7,$FA,$FF,$1E,$3F,$8C,$B1,$E3,$87,$09,$5F,$0B,$47,$07,$FC,$E4
	.dc.b	$F8,$BD,$FF,$F5,$9B,$D2,$F9,$F9,$DE,$7D,$3C,$DA,$79,$7F,$14,$F5
	.dc.b	$7E,$84,$FD,$8D,$F9,$33,$F2,$9F,$F3,$07,$E8,$6F,$CC,$7C,$7E,$3F
	.dc.b	$E8,$36,$63,$C7,$0E,$1F,$80,$B0,$EB,$FA,$CF,$92,$BB,$71,$7A,$1E
	.dc.b	$19,$FF,$9B,$7F,$31,$F9,$9F,$AB,$3F,$6A,$7E,$92,$FC,$55,$FB,$0B
	.dc.b	$F7,$E7,$EF,$7E,$4D,$3E,$66,$3C,$70,$E1,$7B,$C3,$55,$F2,$AB,$B3
	.dc.b	$F2,$9F,$DC,$37,$1F,$0C,$F8,$3E,$DB,$CB,$18,$BF,$D3,$57,$4E,$27
	.dc.b	$CE,$3F,$68,$FF,$1C,$7D,$5F,$F4,$1F,$C8,$7F,$96,$73,$FC,$9A,$D7
	.dc.b	$D6,$AB,$5E,$37,$39,$7E,$63,$F7,$1E,$99,$FE,$01,$E3,$5F,$D3,$7D
	.dc.b	$6F,$CA,$46,$3E,$33,$FD,$6F,$E2,$FE,$27,$67,$E1,$8F,$DF,$BF,$9D
	.dc.b	$7E,$24,$FB,$C7,$E8,$9F,$8E,$C3,$85,$EF,$C8,$5E,$E9,$FA,$CF,$C1
	.dc.b	$AE,$9B,$FE,$51,$7E,$DD,$F9,$47,$95,$F7,$F6,$FF,$55,$18,$CF,$FA
	.dc.b	$AD,$FC,$BE,$07,$EC,$9F,$DA,$3C,$8F,$EF,$39,$BF,$5C,$D4,$E7,$CF
	.dc.b	$F1,$FE,$45,$C7,$83,$F4,$78,$F2,$CB,$F4,$EF,$4F,$9A,$FC,$E5,$FE
	.dc.b	$B2,$79,$C6,$34,$FD,$F5,$6F,$4F,$0F,$ED,$BF,$8A,$D7,$FD,$95,$FB
	.dc.b	$03,$DF,$FF,$0A,$7E,$05,$FA,$4F,$D8,$B5,$F8,$92,$B5,$F4,$FF,$E7
	.dc.b	$EF,$3D,$F9,$8E,$BF,$3D,$FE,$51,$C2,$38,$DE,$1B,$9F,$6F,$4F,$5D
	.dc.b	$BF,$C1,$76,$F1,$CF,$8B,$F2,$9C,$F8,$FF,$D5,$A5,$F3,$1F,$ED,$CF
	.dc.b	$8A,$7E,$A8,$FC,$CD,$E8,$3F,$17,$F7,$5E,$EB,$52,$FD,$17,$3C,$FE
	.dc.b	$5D,$BC,$57,$FE,$E8,$E0,$F5,$C6,$3F,$1E,$FF,$4A,$63,$CA,$7A,$C6
	.dc.b	$33,$E9,$1F,$37,$E8,$FF,$C2,$FF,$6D,$F7,$F8,$3F,$D5,$9F,$98,$67
	.dc.b	$EA,$3F,$D5,$1E,$37,$F6,$AF,$C4,$BF,$97,$EB,$F0,$F1,$FC,$5E,$5F
	.dc.b	$CF,$B1,$F8,$51,$FE,$2B,$8F,$2C,$63,$0E,$6B,$E7,$3E,$4C,$FD,$3E
	.dc.b	$5F,$F4,$4D,$6B,$F5,$FC,$CF,$E8,$BE,$F5,$E0,$7F,$99,$BE,$17,$ED
	.dc.b	$7E,$99,$FE,$03,$D7,$EB,$56,$BE,$F7,$F9,$7E,$0A,$7E,$9F,$F8,$DA
	.dc.b	$E7,$3E,$63,$C4,$BF,$3E,$FA,$1B,$FD,$D8,$E9,$97,$E2,$7E,$2F,$CB
	.dc.b	$BA,$73,$87,$EB,$B4,$F1,$EB,$E7,$8F,$17,$C0,$FE,$77,$E1,$BF,$05
	.dc.b	$F0,$8F,$D7,$9F,$A0,$7F,$54,$FB,$5D,$78,$6A,$D4,$FF,$58,$FD,$B7
	.dc.b	$97,$87,$E6,$CF,$E5,$65,$F8,$D7,$D8,$7F,$0A,$7B,$E5,$DB,$F0,$7F
	.dc.b	$4C,$63,$0F,$DB,$6D,$FD,$8F,$C3,$7E,$7D,$FE,$BB,$FD,$95,$F8,$4B
	.dc.b	$CE,$7E,$38,$FD,$5F,$FA,$0E,$1C,$3F,$0D,$61,$C2,$F7,$D5,$AC,$9F
	.dc.b	$93,$7A,$78,$7F,$53,$1F,$EC,$97,$D3,$8F,$3C,$39,$74,$7E,$12,$31
	.dc.b	$AF,$8F,$F2,$8C,$7E,$AF,$F6,$7F,$5A,$7F,$00,$7E,$CB,$F2,$E7,$EA
	.dc.b	$BE,$6A,$D4,$E7,$F5,$2C,$7E,$23,$E3,$CD,$5F,$83,$DB,$C6,$F0,$F4
	.dc.b	$FF,$C7,$DE,$2F,$F9,$5A,$B7,$E3,$35,$F0,$D7,$E7,$F0,$F8,$BF,$97
	.dc.b	$BF,$55,$FF,$5F,$6E,$FB,$17,$E7,$CF,$CC,$77,$5A,$9C,$FE,$8B,$CB
	.dc.b	$F3,$5F,$36,$3C,$F2,$F8,$EF,$5E,$16,$7D,$7E,$BF,$E4,$4E,$63,$C1
	.dc.b	$F5,$DF,$2F,$F5,$9F,$F2,$47,$DF,$FF,$81,$7E,$3F,$EF,$FE,$AB,$F7
	.dc.b	$27,$3E,$87,$FA,$FB,$B5,$F6,$79,$FD,$BF,$E4,$57,$F9,$47,$7F,$B2
	.dc.b	$F0,$C5,$F9,$77,$7F,$BA,$FF,$3D,$65,$FE,$C8,$F9,$3F,$B9,$F2,$4F
	.dc.b	$D9,$DF,$0A,$FC,$3D,$F8,$5B,$F2,$4F,$F4,$8F,$A5,$F9,$7C,$38,$5E
	.dc.b	$EB,$5F,$B6,$7E,$52,$7A,$70,$7E,$31,$74,$FD,$F7,$8E,$5D,$BD,$6F
	.dc.b	$1E,$9E,$28,$F0,$FD,$1F,$F1,$4F,$C3,$5F,$20,$FD,$33,$EE,$7A,$EF
	.dc.b	$8F,$7E,$42,$F8,$CF,$F1,$2E,$1C,$2F,$75,$AF,$D1,$73,$8F,$3E,$0F
	.dc.b	$DC,$DF,$6F,$CC,$7F,$4C,$F1,$F2,$7F,$61,$87,$FA,$9E,$BF,$AE,$4E
	.dc.b	$74,$F1,$7E,$BD,$E7,$FC,$55,$FA,$AB,$F1,$07,$EF,$2F,$DF,$1F,$7C
	.dc.b	$FC,$E7,$CF,$87,$0B,$DD,$6B,$EF,$33,$8F,$92,$9F,$7B,$63,$B6,$5D
	.dc.b	$7E,$BB,$C5,$18,$C3,$F9,$37,$B7,$FE,$BE,$DF,$FC,$A3,$F8,$6B,$F1
	.dc.b	$6E,$7F,$AC,$BF,$56,$7E,$8F,$F2,$BF,$00,$5E,$F6,$5E,$FE,$9B,$53
	.dc.b	$F2,$2B,$F9,$FF,$3F,$F5,$73,$E3,$BC,$B2,$EF,$5F,$76,$31,$87,$4F
	.dc.b	$0F,$F0,$1F,$80,$BF,$10,$7C,$4B,$D9,$F3,$7C,$5B,$E8,$3E,$FD,$EE
	.dc.b	$B5,$C3,$FC,$93,$F4,$3C,$FE,$2D,$F7,$2B,$FD,$A1,$C1,$E5,$3F,$DD
	.dc.b	$6F,$1C,$BE,$16,$1E,$2C,$FC,$31,$F1,$75,$FE,$CF,$FA,$AB,$F5,$07
	.dc.b	$EC,$9F,$C4,$FE,$0F,$FD,$9F,$E6,$3F,$6B,$FF,$3B,$DE,$EB,$5F,$8C
	.dc.b	$7E,$63,$F6,$1C,$E3,$FC,$53,$7F,$E9,$77,$B6,$C7,$6F,$8D,$67,$FE
	.dc.b	$DC,$FD,$6B,$C7,$39,$8F,$F6,$FF,$93,$8C,$FB,$EA,$F8,$0F,$1F,$E0
	.dc.b	$CF,$D8,$7F,$D2,$3F,$8E,$7E,$03,$F4,$17,$E7,$FA,$9C,$F8,$89,$FF
	.dc.b	$6D,$7C,$07,$FA,$EF,$8B,$FD,$01,$5F,$D3,$5F,$F4,$17,$EC,$B7,$2E
	.dc.b	$DE,$73,$3F,$C6,$FB,$5F,$2A,$73,$1F,$90,$F0,$FE,$6F,$E2,$8F,$C4
	.dc.b	$1F,$95,$BF,$0D,$7E,$9C,$FD,$DF,$5C,$1E,$3C,$38,$5E,$EB,$5F,$A4
	.dc.b	$FD,$29,$E9,$DB,$F3,$97,$E6,$BE,$97,$8E,$5D,$BB,$7C,$7A,$78,$A3
	.dc.b	$C3,$F9,$8F,$F8,$A3,$F7,$3F,$27,$E3,$CF,$DB,$1F,$8C,$BF,$6A,$BD
	.dc.b	$9F,$EB,$8B,$DD,$6A,$77,$F8,$7E,$72,$FF,$75,$6F,$FE,$0C,$BF,$81
	.dc.b	$FB,$87,$F7,$97,$8E,$5D,$7E,$AF,$C5,$7F,$EF,$1C,$BA,$79,$D8,$D7
	.dc.b	$C9,$7E,$31,$EB,$78,$FF,$44,$FF,$1C,$72,$BD,$B4,$3F,$36,$7F,$00
	.dc.b	$6F,$F0,$15,$EE,$B5,$F6,$D9,$E2,$F4,$43,$F6,$7F,$43,$DB,$2E,$BF
	.dc.b	$3B,$F2,$63,$03,$A3,$E6,$7E,$58,$FD,$1D,$FA,$93,$E1,$5F,$BE,$7F
	.dc.b	$7A,$FD,$8F,$E6,$5E,$EB,$53,$FD,$FB,$E6,$B3,$87,$C9,$A7,$AB,$63
	.dc.b	$AE,$3D,$3E,$6F,$C3,$18,$C3,$FD,$A3,$3F,$E9,$BE,$07,$6F,$FB,$E3
	.dc.b	$F5,$C7,$D1,$FF,$68,$FC,$17,$FC,$71,$D1,$F7,$FD,$15,$6B,$E8,$3F
	.dc.b	$CB,$5F,$79,$FE,$52,$F3,$BD,$99,$FF,$A9,$F3,$E8,$FE,$07,$07,$C0
	.dc.b	$7E,$05,$E7,$A7,$7F,$CD,$7F,$25,$18,$CB,$FC,$CD,$F8,$17,$F9,$F3
	.dc.b	$F2,$D7,$B3,$FD,$A7,$E8,$FF,$D6,$83,$F6,$17,$B6,$DF,$C7,$B3,$0E
	.dc.b	$3F,$45,$FD,$F3,$0E,$1F,$98,$EF,$9B,$E2,$56,$AF,$FC,$C6,$F8,$3E
	.dc.b	$87,$FC,$A0,$F2,$BF,$0F,$E3,$DE,$5D,$7C,$97,$FE,$13,$F8,$CF,$47
	.dc.b	$E2,$1F,$ED,$3F,$98,$FF,$1B,$7E,$50,$C1,$F9,$5F,$87,$0F,$F6,$76
	.dc.b	$7F,$E9,$FF,$C6,$7F,$AA,$FD,$DB,$0E,$1F,$BA,$EE,$FD,$99,$FF,$B7
	.dc.b	$DF,$3A,$FD,$9B,$CE,$F8,$6F,$96,$7E,$4C,$7A,$E3,$FE,$CE,$C3,$A6
	.dc.b	$5E,$1D,$3C,$F7,$FE,$B1,$E4,$F9,$73,$FE,$50,$FC,$CF,$FD,$AB,$F9
	.dc.b	$DF,$E2,$3F,$07,$FD,$35,$6B,$F0,$15,$6B,$BF,$FC,$D7,$F9,$6F,$69
	.dc.b	$CE,$9F,$39,$7E,$37,$F3,$9E,$3F,$8F,$33,$FC,$AF,$BB,$F8,$39,$CC
	.dc.b	$7E,$63,$E6,$F1,$9F,$7D,$7F,$D7,$5A,$FE,$0B,$F3,$00,$FD,$59,$F9
	.dc.b	$7B,$E4,$3F,$DD,$DF,$BC,$BE,$1F,$E7,$5F,$0B,$DD,$6B,$E1,$33,$2F
	.dc.b	$C9,$7F,$D2,$8E,$F8,$FF,$D5,$6F,$1C,$BB,$7E,$33,$FE,$7D,$D3,$C9
	.dc.b	$3E,$39,$CF,$D3,$FF,$B6,$A1,$F8,$43,$6B,$F4,$CF,$DA,$3D,$1F,$A9
	.dc.b	$F9,$AB,$42,$75,$BB,$F5,$ED,$61,$E6,$AF,$F4,$76,$DF,$E0,$9C,$BB
	.dc.b	$CF,$6C,$BE,$24,$F0,$AE,$91,$FC,$C2,$3F,$25,$F8,$13,$F6,$87,$E6
	.dc.b	$EF,$98,$FF,$39,$7E,$98,$FD,$59,$F4,$3C,$2F,$75,$A9,$7E,$45,$9C
	.dc.b	$3F,$D7,$D8,$F5,$FD,$32,$E1,$C3,$F0,$56,$6F,$CB,$FF,$E0,$0D,$F8
	.dc.b	$F0,$F1,$56,$BF,$4C,$CF,$E7,$0F,$D4,$DF,$89,$3F,$10,$FF,$09,$F8
	.dc.b	$0F,$6B,$DB,$7B,$AD,$7D,$8A,$9F,$BA,$EB,$1F,$3D,$FE,$AF,$9B,$FC
	.dc.b	$69,$B2,$39,$67,$FE,$20,$D3,$C7,$7E,$28,$FF,$11,$EF,$EA,$F7,$DF
	.dc.b	$95,$7D,$9E,$DF,$F4,$E7,$E4,$AF,$DF,$FE,$8D,$6A,$7E,$F7,$1A,$BF
	.dc.b	$B0,$78,$BF,$C8,$B6,$FF,$48,$63,$F3,$30,$FB,$1B,$FF,$C0,$09,$CE
	.dc.b	$3C,$BD,$0F,$15,$F8,$72,$FE,$31,$7F,$C4,$7F,$80,$5F,$9F,$3E,$AD
	.dc.b	$FC,$01,$FD,$BB,$F0,$37,$E4,$2F,$CD,$F8,$5E,$EB,$52,$FC,$DB,$3F
	.dc.b	$69,$FE,$FD,$CB,$AE,$57,$0E,$1F,$2D,$FE,$7F,$F1,$DA,$F4,$AF,$0C
	.dc.b	$FF,$A0,$F8,$B3,$C9,$FD,$0B,$1E,$7A,$F8,$3F,$73,$F5,$7C,$2F,$FB
	.dc.b	$FB,$F6,$87,$D3,$FF,$74,$F9,$27,$E1,$DA,$F6,$D5,$AF,$9C,$57,$FA
	.dc.b	$93,$4E,$D8,$AF,$D4,$BF,$08,$B1,$CB,$2E,$DC,$3D,$72,$E9,$F3,$1E
	.dc.b	$8F,$93,$FE,$ED,$F9,$33,$F8,$03,$F1,$5F,$A2,$3D,$CF,$D5,$FD,$A7
	.dc.b	$F9,$03,$7F,$EB,$F7,$BF,$2D,$7B,$EA,$5F,$F9,$3B,$2E,$B9,$7D,$9D
	.dc.b	$7F,$C0,$87,$84,$F9,$B8,$F3,$72,$7F,$51,$7B,$61,$D6,$F8,$4A,$73
	.dc.b	$9F,$F0,$0F,$91,$CC,$3F,$10,$7E,$62,$F1,$9F,$AF,$BF,$7D,$7E,$B0
	.dc.b	$FD,$FF,$C1,$87,$0B,$DD,$6B,$2C,$E3,$E6,$D3,$7F,$A8,$DE,$4F,$8B
	.dc.b	$8F,$1D,$3B,$6F,$E4,$8C,$64,$CB,$CF,$F1,$FF,$E0,$BF,$D0,$FC,$9F
	.dc.b	$02,$FD,$75,$FB,$77,$C3,$AD,$61,$FC,$86,$7F,$D7,$FA,$7D,$49,$FF
	.dc.b	$0A,$6D,$F2,$A3,$1F,$11,$E5,$E8,$FB,$F5,$CB,$2F,$D7,$D7,$11,$9F
	.dc.b	$8B,$E7,$FC,$B1,$8C,$FC,$1F,$B6,$FF,$81,$7F,$1F,$FF,$22,$FD,$EF
	.dc.b	$F9,$3B,$F0,$8F,$97,$93,$F5,$6D,$7D,$9F,$F9,$AB,$C9,$FF,$4E,$76
	.dc.b	$7F,$39,$F9,$7C,$E3,$E5,$E3,$FD,$7E,$7F,$E5,$0F,$C1,$71,$08,$EF
	.dc.b	$87,$2F,$D9,$FF,$07,$18,$D7,$ED,$F8,$D8,$FD,$73,$FD,$7B,$FA,$DB
	.dc.b	$F0,$2F,$DD,$FE,$B3,$E9,$9C,$C7,$EF,$BC,$7C,$7E,$51,$E0,$C9,$5F
	.dc.b	$E4,$CD,$7E,$67,$5F,$F9,$76,$73,$F6,$9F,$E6,$6C,$FB,$57,$86,$31
	.dc.b	$93,$EE,$B1,$CA,$BD,$FE,$3B,$FE,$2F,$FC,$F3,$FC,$F5,$F7,$0F,$84
	.dc.b	$FF,$7D,$7B,$FF,$CD,$BF,$5C,$D7,$DE,$FF,$9B,$27,$3F,$0E,$FD,$BF
	.dc.b	$38,$7C,$8C,$7F,$D0,$9F,$6B,$70,$E1,$E2,$FF,$2C,$F9,$9F,$D5,$D7
	.dc.b	$83,$2E,$FE,$66,$38,$7F,$9F,$72,$7F,$B8,$FE,$09,$D9,$FB,$EF,$23
	.dc.b	$FC,$C3,$8F,$B7,$E3,$7F,$93,$FE,$2E,$BD,$FE,$75,$ED,$CB,$9E,$5F
	.dc.b	$C0,$B1,$E3,$1F,$EC,$38,$F9,$A7,$F3,$0C,$7A,$3E,$1C,$BC,$99,$76
	.dc.b	$C5,$E1,$62,$B5,$F9,$1F,$F9,$67,$F3,$47,$E2,$EF,$D7,$1F,$A2,$7F
	.dc.b	$75,$F3,$9F,$42,$E6,$E1,$7B,$AD,$4B,$F7,$4C,$FE,$C7,$F9,$4C,$BA
	.dc.b	$E5,$70,$E1,$F3,$7F,$E1,$5D,$BF,$8B,$D6,$B2,$F0,$70,$FF,$1A,$FD
	.dc.b	$9D,$CC,$FC,$CF,$FC,$11,$F2,$0F,$D6,$1F,$9A,$BF,$51,$FF,$3A,$61
	.dc.b	$C2,$F7,$5A,$FB,$3C,$E3,$CF,$9F,$E5,$8B,$EF,$F2,$7F,$AC,$79,$70
	.dc.b	$7F,$6C,$C3,$FC,$3D,$5C,$B1,$EB,$F8,$67,$9F,$F5,$A7,$EA,$0F,$A1
	.dc.b	$7E,$69,$FC,$6F,$E8,$7F,$4F,$E1,$FF,$A9,$35,$61,$C2,$F7,$E2,$AF
	.dc.b	$E6,$5E,$EB,$AE,$9E,$8C,$38,$FE,$83,$E7,$9E,$59,$7F,$29,$FA,$7F
	.dc.b	$26,$9E,$38,$FE,$59,$F4,$DF,$D1,$B9,$FF,$D7,$7F,$4E,$FC,$5D,$FB
	.dc.b	$5B,$C6,$7E,$3A,$FD,$93,$FD,$41,$87,$0B,$DD,$6B,$F6,$CC,$FA,$4F
	.dc.b	$D8,$7F,$8A,$DF,$6E,$3F,$9E,$F1,$CF,$B7,$C9,$F8,$EF,$FC,$85,$97
	.dc.b	$4F,$C0,$F1,$B7,$96,$FC,$9B,$3F,$80,$7C,$7E,$6B,$F3,$2F,$F3,$17
	.dc.b	$EE,$DF,$9A,$FD,$FD,$FB,$DE,$A7,$3E,$52,$5F,$A7,$F2,$FC,$DB,$CF
	.dc.b	$8F,$3E,$3F,$3B,$D7,$DB,$67,$FA,$6F,$13,$E5,$4E,$63,$F5,$3D,$5F
	.dc.b	$DC,$B8,$DF,$A4,$BF,$35,$FF,$29,$7E,$78,$F8,$F7,$ED,$AF,$D4,$FF
	.dc.b	$D9,$98,$70,$BD,$D6,$BF,$30,$CE,$2F,$9D,$FE,$83,$7B,$E0,$FC,$DB
	.dc.b	$C7,$EA,$DA,$77,$F8,$DF,$26,$1E,$39,$CE,$5F,$E1,$1F,$2F,$FD,$EF
	.dc.b	$FB,$7F,$FB,$FB,$F3,$5F,$F4,$D7,$EA,$FF,$C7,$3F,$10,$7E,$21,$FE
	.dc.b	$E1,$AD,$7E,$1C,$AD,$7E,$A5,$9F,$19,$8F,$F5,$6B,$43,$F5,$7C,$3A
	.dc.b	$C3,$97,$BF,$FE,$96,$E3,$E5,$87,$8A,$3F,$D4,$7A,$BE,$67,$FC,$A3
	.dc.b	$4F,$CC,$9F,$9C,$BF,$51,$72,$BF,$49,$FF,$80,$1F,$AE,$3E,$93,$FD
	.dc.b	$F5,$87,$0B,$DD,$6B,$F3,$9C,$E3,$F8,$7C,$BF,$83,$DE,$DB,$1D,$F2
	.dc.b	$ED,$F9,$07,$F4,$29,$E3,$8F,$48,$FF,$2B,$C7,$83,$F4,$CF,$CA,$F1
	.dc.b	$BF,$19,$FF,$30,$FD,$53,$9F,$5E,$7F,$C7,$9C,$F5,$63,$1F,$1E,$2E
	.dc.b	$2F,$F0,$1E,$E8,$EB,$7F,$E6,$CF,$AF,$FF,$22,$CE,$71,$E9,$F7,$DF
	.dc.b	$4F,$17,$37,$CD,$7F,$CF,$BF,$86,$7D,$7F,$DD,$DF,$B7,$FE,$A3,$F0
	.dc.b	$1F,$F5,$CF,$D3,$F0,$EB,$FA,$77,$BA,$D7,$E2,$4F,$D9,$B5,$97,$4E
	.dc.b	$67,$E4,$FF,$E1,$AC,$3B,$4F,$F3,$47,$97,$AD,$FE,$80,$FC,$23,$FC
	.dc.b	$BB,$18,$D7,$FC,$B1,$5D,$0F,$F5,$A6,$0D,$BC,$F8,$7F,$A2,$33,$7E
	.dc.b	$51,$BF,$DF,$3F,$AF,$FE,$16,$FC,$3D,$EC,$BE,$6D,$FB,$83,$EF,$B8
	.dc.b	$7C,$03,$0E,$17,$BA,$F9,$3D,$1F,$27,$81,$9F,$F6,$CF,$C4,$FF,$C6
	.dc.b	$EF,$82,$3F,$A8,$BE,$2D,$3C,$3A,$7B,$2F,$94,$E7,$1F,$27,$89,$FE
	.dc.b	$7E,$FD,$D5,$FA,$3F,$F2,$FF,$C0,$7F,$9B,$FB,$5F,$EC,$0F,$43,$E2
	.dc.b	$56,$BE,$37,$FC,$F9,$F1,$1F,$C2,$7F,$15,$7F,$00,$6B,$2F,$E2,$FC
	.dc.b	$3F,$D9,$30,$E0,$D8,$ED,$7F,$C4,$7F,$0F,$F1,$46,$33,$7E,$27,$FE
	.dc.b	$92,$FD,$7D,$FB,$CF,$C2,$FD,$99,$FB,$DF,$F2,$DE,$8F,$B1,$5A,$E5
	.dc.b	$56,$BE,$3F,$FD,$9B,$F8,$D6,$B1,$F9,$7B,$FF,$78,$C3,$AF,$8B,$FE
	.dc.b	$62,$79,$61,$C6,$7C,$55,$E1,$CF,$FA,$37,$DA,$E3,$3E,$F1,$FC,$FF
	.dc.b	$BB,$F1,$5F,$A0,$FF,$AD,$B4,$FF,$1B,$7D,$47,$FA,$3F,$F0,$17,$E4
	.dc.b	$5A,$F9,$AF,$F7,$47,$DB,$7F,$B5,$BF,$0D,$53,$F7,$2D,$7C,$D7,$F9
	.dc.b	$DB,$3E,$9C,$3F,$E1,$07,$B6,$1C,$7F,$10,$FF,$5C,$C6,$34,$FF,$4F
	.dc.b	$7A,$5F,$EB,$CF,$D0,$BF,$D6,$7F,$80,$FF,$AD,$32,$FF,$07,$F5,$3E
	.dc.b	$1B,$F4,$55,$AF,$C3,$BF,$B3,$7D,$3F,$FC,$01,$F8,$AC,$C3,$F0,$4F
	.dc.b	$F4,$4D,$74,$E0,$FF,$11,$BC,$72,$FF,$0D,$F3,$78,$6F,$FC,$D4,$74
	.dc.b	$FC,$3B,$FD,$9F,$F8,$9B,$E9,$5F,$29,$F0,$7F,$4E,$7E,$3D,$FC,$DF
	.dc.b	$F4,$8F,$DB,$5E,$EB,$53,$FC,$EF,$1C,$E1,$CB,$5F,$F1,$AD,$F5,$FB
	.dc.b	$9F,$F4,$FB,$C7,$A9,$9B,$F6,$2F,$E8,$7E,$B7,$EC,$6B,$53,$E2,$FA
	.dc.b	$6F,$F7,$2F,$E3,$0E,$0F,$DF,$7B,$5F,$E2,$1F,$C9,$9F,$9D,$3F,$78
	.dc.b	$FE,$19,$7B,$F7,$7F,$83,$7D,$1A,$F7,$FB,$36,$B2,$7E,$15,$FE,$B4
	.dc.b	$C3,$87,$1F,$F9,$39,$E5,$A7,$7F,$3B,$C7,$7E,$2C,$BA,$5F,$FB,$EB
	.dc.b	$E0,$9E,$37,$F9,$8B,$D9,$FE,$D3,$F4,$FF,$AC,$FD,$AF,$F9,$05,$6B
	.dc.b	$F0,$1F,$F1,$BF,$D1,$7F,$83,$FD,$57,$E5,$BA,$C9,$F9,$5F,$FB,$23
	.dc.b	$0E,$1A,$7F,$98,$9E,$5A,$76,$FC,$D3,$C5,$7E,$1C,$BF,$C3,$5E,$03
	.dc.b	$0F,$E6,$1F,$BE,$BF,$37,$7D,$3B,$F6,$CF,$F1,$07,$B6,$FC,$D9,$FB
	.dc.b	$E7,$0B,$DD,$6A,$7F,$CD,$3E,$27,$F2,$9C,$7A,$63,$FE,$DE,$7B,$7E
	.dc.b	$91,$70,$E3,$CF,$FC,$33,$F2,$FF,$26,$1C,$67,$3A,$7F,$80,$B5,$F0
	.dc.b	$65,$E5,$C3,$F8,$9E,$6F,$37,$FA,$57,$E8,$AF,$DF,$3F,$CC,$9F,$26
	.dc.b	$FD,$67,$FD,$0D,$FA,$FB,$E5,$37,$5A,$9C,$FC,$16,$3C,$1F,$F7,$AE
	.dc.b	$3C,$FF,$01,$B7,$BF,$C2,$7F,$51,$C1,$BD,$3B,$43,$E1,$7D,$7F,$45
	.dc.b	$FE,$1E,$F2,$3F,$C3,$1F,$3B,$FC,$37,$CA,$7B,$5F,$E5,$EF,$C2,$7F
	.dc.b	$B1,$56,$BD,$75,$6B,$E9,$9F,$85,$EB,$E0,$7F,$B3,$6F,$FE,$A2,$C3
	.dc.b	$FC,$49,$F0,$DF,$E2,$C8,$FD,$5E,$BE,$1A,$FE,$D9,$97,$4D,$3F,$C3
	.dc.b	$DD,$FF,$D7,$F1,$F0,$69,$E4,$FA,$3F,$4E,$5F,$87,$9C,$DF,$F9,$D3
	.dc.b	$E1,$3F,$BC,$78,$5F,$C9,$FC,$27,$E3,$EF,$D6,$5F,$93,$FF,$A8,$FF
	.dc.b	$22,$DE,$FF,$8F,$AD,$4F,$FA,$9F,$F6,$0C,$FE,$FD,$E9,$DB,$C2,$33
	.dc.b	$EC,$FF,$B0,$1F,$05,$F2,$FC,$3F,$71,$8C,$FC,$3F,$82,$F2,$F1,$DF
	.dc.b	$AF,$7F,$93,$FF,$63,$7E,$FC,$FD,$81,$FB,$63,$E4,$3F,$C2,$98,$70
	.dc.b	$FB,$56,$1C,$2F,$7E,$C6,$B2,$7E,$05,$E9,$E1,$C6,$17,$B7,$FC,$19
	.dc.b	$97,$3C,$B9,$78,$BE,$5C,$7C,$9B,$78,$FD,$8F,$83,$FB,$DB,$F6,$7F
	.dc.b	$EB,$3F,$7A,$FC,$D1,$F3,$BF,$CE,$7E,$DF,$FC,$55,$7B,$FA,$0F,$F3
	.dc.b	$FD,$6B,$17,$DC,$EB,$2F,$EC,$7C,$5F,$EE,$AC,$3B,$7C,$87,$9F,$CA
	.dc.b	$FF,$3A,$C7,$3D,$7C,$1A,$FE,$83,$8F,$FC,$4D,$39,$D3,$C7,$C3,$F6
	.dc.b	$3F,$7A,$FF,$1D,$7E,$48,$FC,$BD,$F7,$BF,$EC,$1F,$DA,$5F,$97,$7F
	.dc.b	$90,$6F,$75,$AF,$A0,$57,$FA,$43,$4F,$F1,$A5,$7F,$91,$7F,$26,$F3
	.dc.b	$71,$7F,$1B,$8E,$EF,$E2,$67,$7E,$BB,$4F,$16,$9F,$ED,$1F,$47,$CD
	.dc.b	$39,$D3,$C1,$F9,$C6,$2F,$7A,$F8,$BF,$4D,$FF,$2B,$7E,$2A,$FB,$87
	.dc.b	$E8,$BF,$E9,$7F,$C7,$58,$FF,$BC,$7C,$0D,$D6,$BF,$0C,$56,$BF,$1E
	.dc.b	$7D,$16,$B1,$F5,$F1,$FF,$B3,$30,$FF,$55,$3F,$BB,$E3,$FF,$2A,$69
	.dc.b	$E1,$AF,$06,$9F,$C3,$EB,$96,$7D,$A3,$EB,$76,$FF,$AB,$BF,$31,$7E
	.dc.b	$6C,$FC,$99,$FA,$97,$F4,$0F,$CA,$BF,$D0,$BF,$4F,$FE,$73,$BD,$D6
	.dc.b	$BE,$4D,$5F,$E8,$AF,$A3,$D6,$2F,$B3,$F4,$E1,$C3,$6F,$A5,$A7,$86
	.dc.b	$BC,$1A,$77,$F2,$BF,$7B,$7D,$A3,$C5,$E1,$E2,$B7,$A7,$83,$F0,$C7
	.dc.b	$E4,$BF,$E8,$7F,$BD,$FA,$3F,$C0,$0F,$C6,$FF,$C0,$1F,$A0,$7F,$41
	.dc.b	$FC,$9B,$5D,$0A,$D7,$9E,$AF,$E8,$DF,$A7,$EB,$F4,$9F,$A3,$97,$4E
	.dc.b	$4F,$49,$FD,$9D,$EF,$9F,$6E,$AF,$E2,$B1,$8C,$7A,$79,$1F,$C6,$3D
	.dc.b	$5F,$F2,$5F,$CA,$FF,$AD,$3F,$53,$FD,$D7,$D0,$BE,$DF,$FD,$41,$5A
	.dc.b	$F0,$D5,$AF,$AA,$CF,$DF,$FE,$8B,$7F,$19,$FE,$B3,$7F,$AF,$38,$BE
	.dc.b	$55,$FC,$1F,$F1,$6F,$15,$7F,$63,$C3,$AF,$B5,$FE,$A6,$C3,$FD,$61
	.dc.b	$F8,$8B,$F1,$27,$E5,$8F,$D8,$DF,$87,$3E,$BD,$FB,$F7,$F8,$DF,$3F
	.dc.b	$F5,$36,$3C,$70,$E1,$7B,$AF,$F2,$4D,$6B,$17,$E9,$7F,$9E,$C3,$86
	.dc.b	$7E,$C7,$96,$7E,$0C,$DF,$23,$FD,$0A,$73,$9F,$8F,$F3,$4F,$E8,$BF
	.dc.b	$B4,$3E,$C1,$F9,$B3,$F5,$67,$E9,$4F,$DA,$1F,$99,$B0,$FF,$68,$1C
	.dc.b	$6F,$85,$6E,$7F,$DA,$5C,$5F,$EC,$3A,$D7,$07,$D2,$7C,$1F,$01,$FB
	.dc.b	$87,$96,$5E,$0C,$BF,$7B,$5D,$B3,$F2,$CE,$7E,$9F,$D3,$18,$CF,$C1
	.dc.b	$EA,$7F,$B8,$FF,$43,$FF,$7B,$F9,$0F,$D5,$3E,$8F,$E7,$CF,$D7,$9F
	.dc.b	$B5,$6F,$D2,$5E,$FB,$57,$BF,$DB,$B5,$3F,$C3,$B2,$5B,$D2,$C3,$CB
	.dc.b	$3E,$FD,$7F,$DC,$AB,$C7,$1E,$2E,$9C,$56,$F3,$EF,$AB,$F7,$6F,$F7
	.dc.b	$E7,$E2,$BF,$DA,$3F,$18,$7E,$5F,$FE,$37,$F2,$7F,$88,$FE,$0B,$AF
	.dc.b	$9B,$FF,$64,$CE,$7D,$D7,$E1,$19,$C3,$F9,$2E,$3F,$DC,$BF,$43,$B8
	.dc.b	$70,$FC,$01,$FA,$27,$F8,$83,$BF,$FC,$BD,$7E,$1C,$FC,$1F,$19,$FE
	.dc.b	$0C,$8C,$5B,$6F,$E4,$7F,$48,$F1,$5F,$B5,$7F,$99,$F2,$7E,$65,$FE
	.dc.b	$13,$FC,$3D,$FB,$5B,$F5,$47,$E1,$6F,$E0,$07,$EC,$9C,$F2,$E5,$8F
	.dc.b	$1C,$38,$5E,$FF,$65,$D6,$7F,$E7,$5A,$FF,$10,$7E,$B7,$FE,$09,$C3
	.dc.b	$8F,$D9,$B5,$C5,$EE,$31,$97,$4F,$57,$FC,$D7,$7C,$2F,$CF,$B7,$9B
	.dc.b	$E4,$33,$8F,$19,$CE,$BE,$98,$F1,$9F,$03,$2E,$B7,$EE,$C5,$F2,$3A
	.dc.b	$F5,$1F,$72,$F8,$57,$D4,$BF,$5C,$FF,$4D,$FE,$28,$FD,$83,$85,$EE
	.dc.b	$B5,$E2,$A9,$FB,$C6,$B1,$F7,$EF,$FB,$AC,$3F,$C1,$6F,$F7,$FE,$3F
	.dc.b	$AD,$A7,$8A,$BC,$3A,$7F,$AA,$BE,$5B,$DF,$7C,$67,$39,$FF,$9B,$23
	.dc.b	$B6,$9E,$2D,$1F,$95,$FC,$2F,$C5,$3F,$5B,$E7,$7F,$B5,$FE,$49,$F8
	.dc.b	$BB,$EA,$FF,$D6,$FE,$0F,$E2,$33,$FF,$57,$61,$C3,$F1,$E6,$1C,$2F
	.dc.b	$FC,$A1,$7B,$BF,$F5,$6F,$AE,$CB,$DF,$8F,$EF,$3F,$3D,$B3,$9C,$B9
	.dc.b	$E6,$FD,$D3,$DB,$7D,$F4,$F2,$F9,$38,$C3,$84,$78,$7F,$60,$FB,$BF
	.dc.b	$7B,$FC,$23,$F7,$67,$E4,$7F,$EA,$9F,$5B,$FB,$37,$A7,$AC,$55,$AF
	.dc.b	$97,$FD,$F7,$CE,$3F,$6C,$D7,$B9,$F5,$B3,$E9,$CB,$F5,$4F,$F6,$83
	.dc.b	$DF,$4E,$DF,$0F,$F5,$23,$18,$FF,$09,$F5,$71,$A7,$82,$FC,$99,$BD
	.dc.b	$9F,$26,$BE,$3D,$FF,$BE,$7E,$ED,$F9,$8F,$D7,$FF,$C6,$BF,$88,$FF
	.dc.b	$9A,$7E,$2F,$D3,$EE,$3D,$6D,$7D,$12,$B5,$97,$F0,$AF,$D1,$B3,$1F
	.dc.b	$90,$C7,$FC,$7F,$FA,$05,$BE,$1F,$33,$FD,$3F,$0F,$F6,$B7,$87,$F6
	.dc.b	$EF,$C3,$39,$FC,$4F,$F2,$31,$8D,$7C,$7E,$E7,$EB,$BE,$C9,$FA,$97
	.dc.b	$FA,$4B,$EF,$17,$FE,$7E,$FC,$35,$FB,$CE,$EB,$53,$9F,$CD,$B1,$FB
	.dc.b	$BF,$E9,$B1,$E7,$DD,$6F,$7F,$81,$BF,$40,$FD,$1D,$6B,$4E,$F2,$FD
	.dc.b	$0B,$37,$C3,$57,$E4,$9F,$C1,$3F,$1A,$FE,$41,$E7,$3F,$37,$7E,$6B
	.dc.b	$C7,$DF,$7E,$C9,$BF,$84,$FD,$3D,$6B,$4F,$DE,$FC,$DE,$7F,$77,$FD
	.dc.b	$EE,$7D,$3C,$DF,$C4,$07,$BD,$F1,$FD,$0B,$F7,$71,$8D,$3D,$1F,$97
	.dc.b	$63,$3E,$F1,$E3,$FD,$97,$F8,$DF,$DF,$7F,$BA,$BF,$74,$7E,$18,$FE
	.dc.b	$01,$FF,$47,$FE,$F1,$FE,$74,$F1,$7F,$69,$7B,$AD,$7E,$1E,$AF,$F6
	.dc.b	$4D,$6B,$D6,$FF,$14,$F0,$FA,$30,$EC,$D8,$65,$E0,$CB,$FD,$3B,$EF
	.dc.b	$7F,$14,$BE,$D1,$FE,$A0,$EF,$C5,$F0,$D3,$C3,$9B,$F5,$EF,$E6,$BF
	.dc.b	$15,$FC,$4B,$E4,$DF,$AD,$FF,$5B,$FB,$B7,$E0,$4F,$9E,$DF,$81,$FD
	.dc.b	$26,$B5,$2F,$85,$CE,$1F,$CB,$B1,$FE,$3B,$F3,$57,$0E,$1F,$3C,$CF
	.dc.b	$FC,$95,$CD,$FC,$66,$FC,$59,$F8,$7E,$E7,$F9,$7C,$62,$F8,$69,$FE
	.dc.b	$61,$BF,$E4,$1F,$A0,$7F,$10,$FA,$C7,$EF,$2E,$CF,$77,$91,$FE,$55
	.dc.b	$BD,$FC,$17,$F8,$82,$B5,$C9,$F9,$4E,$EF,$EA,$F9,$73,$EE,$FC,$4B
	.dc.b	$FC,$73,$EB,$7F,$A4,$BD,$EF,$8F,$E0,$FF,$D1,$23,$1A,$7C,$DF,$D5
	.dc.b	$B1,$9F,$78,$F1,$FC,$47,$F6,$4F,$C6,$9F,$8A,$3F,$6A,$7E,$8A,$FC
	.dc.b	$ED,$F2,$0F,$CF,$1F,$B2,$71,$C3,$85,$EE,$B5,$FA,$8E,$7F,$37,$FE
	.dc.b	$BF,$9F,$6C,$EE,$1C,$33,$EB,$F9,$B7,$F5,$CA,$D6,$5C,$CC,$56,$F2
	.dc.b	$F1,$7A,$DF,$DF,$21,$FB,$87,$F9,$0B,$F7,$6F,$F3,$F7,$C5,$FD,$2C
	.dc.b	$B5,$39,$FA,$A4,$FD,$6F,$C9,$B3,$F3,$1F,$DD,$31,$E7,$F9,$F5,$C3
	.dc.b	$85,$BF,$3B,$FE,$E3,$5E,$0B,$67,$E0,$FC,$4B,$38,$70,$ED,$7E,$51
	.dc.b	$FE,$33,$FB,$AF,$F1,$97,$A1,$FD,$77,$CE,$FE,$E1,$F7,$59,$F1,$3F
	.dc.b	$B7,$FC,$EF,$F5,$E9,$FE,$E9,$F9,$5E,$67,$FB,$6F,$D3,$7F,$8E,$DF
	.dc.b	$F1,$E7,$95,$FE,$E2,$78,$FD,$36,$51,$C7,$2F,$07,$BB,$8A,$E1,$3D
	.dc.b	$3E,$9B,$FB,$9F,$D3,$67,$6F,$3E,$BE,$3F,$A7,$FE,$FF,$18,$D3,$FC
	.dc.b	$15,$B7,$F8,$27,$EA,$FB,$5E,$FE,$9F,$D1,$FA,$33,$F4,$6F,$F2,$77
	.dc.b	$E3,$4F,$DF,$5F,$AB,$3F,$34,$FB,$7C,$5E,$4C,$38,$7C,$5F,$FC,$00
	.dc.b	$BD,$D7,$0A,$D6,$2F,$4F,$E0,$72,$78,$23,$FC,$B8,$FC,$39,$DE,$1C
	.dc.b	$F2,$7B,$DE,$FA,$F8,$F9,$F8,$34,$F3,$53,$F6,$DD,$52,$B5,$AF,$93
	.dc.b	$C4,$FA,$DE,$B7,$EC,$BD,$AF,$DE,$6B,$F3,$A3,$1E,$DF,$F0,$3E,$7F
	.dc.b	$F3,$A4,$7E,$FF,$2F,$B9,$5B,$F6,$1E,$D7,$B2,$31,$EE,$7F,$0E,$F6
	.dc.b	$BF,$95,$7B,$7E,$4C,$7C,$1C,$D7,$1F,$E0,$1A,$FF,$60,$F6,$FF,$D0
	.dc.b	$9F,$B8,$7F,$A0,$FF,$63,$7C,$9B,$E6,$7F,$AB,$7E,$14,$FC,$83,$FD
	.dc.b	$FD,$FA,$17,$0E,$3F,$C5,$E1,$C2,$F7,$4F,$C7,$7F,$4D,$C6,$CF,$B6
	.dc.b	$7F,$C3,$31,$FA,$51,$8C,$3D,$8D,$F9,$6F,$F5,$D6,$2F,$BA,$FC,$B4
	.dc.b	$E7,$5F,$27,$C8,$E3,$1F,$1E,$5E,$9C,$78,$E9,$FB,$4D,$3C,$55,$AD
	.dc.b	$3C,$F2,$FD,$C7,$D9,$2D,$BD,$98,$72,$FD,$DF,$3F,$A1,$FF,$22,$E4
	.dc.b	$FF,$02,$FB,$8F,$DB,$7E,$07,$C9,$6B,$E0,$FF,$CA,$3F,$34,$AD,$6C
	.dc.b	$F9,$3D,$63,$F0,$36,$FD,$AE,$7F,$CB,$BE,$67,$10,$8F,$06,$3C,$F4
	.dc.b	$F8,$93,$CA,$31,$C9,$F0,$E3,$86,$9E,$2D,$3C,$3F,$91,$7F,$B0,$FF
	.dc.b	$09,$FF,$61,$7E,$04,$FC,$C1,$FA,$5B,$F5,$67,$E4,$9F,$E8,$3F,$D5
	.dc.b	$3F,$A1,$75,$FF,$B2,$35,$FE,$17,$8F,$1C,$38,$5E,$FF,$66,$ED,$5A
	.dc.b	$CB,$A3,$FB,$1C,$FC,$31,$E0,$8C,$1B,$BE,$7E,$E7,$F8,$F6,$39,$9F
	.dc.b	$ED,$A9,$F3,$E9,$E6,$9F,$F1,$F4,$EB,$F6,$1F,$F0,$4F,$53,$C4,$FB
	.dc.b	$DF,$6B,$EC,$FC,$EF,$5F,$3E,$AD,$78,$2A,$F8,$3F,$23,$A9,$FD,$86
	.dc.b	$FF,$EE,$9C,$BF,$C2,$BF,$1B,$8D,$3F,$96,$C7,$83,$4E,$FE,$BF,$F2
	.dc.b	$DB,$E1,$A7,$79,$F1,$FC,$9E,$70,$FE,$6B,$FA,$43,$F1,$47,$EF,$6F
	.dc.b	$DF,$9F,$1D,$FE,$D1,$FD,$F3,$D1,$D9,$F4,$F1,$63,$C7,$0E,$17,$BE
	.dc.b	$0F,$93,$39,$AF,$9E,$70,$C9,$F8,$7F,$AE,$31,$87,$2F,$A5,$7E,$13
	.dc.b	$E7,$9C,$EB,$E4,$9F,$F6,$86,$9E,$3C,$BC,$53,$9D,$5F,$BD,$2B,$F5
	.dc.b	$D8,$7F,$54,$FC,$4F,$D3,$F9,$0F,$F4,$3F,$D4,$5F,$A6,$BF,$37,$7E
	.dc.b	$D9,$C7,$D5,$FE,$B7,$8F,$FB,$E3,$66,$3C,$70,$E1,$F7,$3B,$AE,$D4
	.dc.b	$FD,$C7,$F9,$56,$1C,$B4,$F0,$D7,$F4,$58,$C5,$74,$F8,$96,$BE,$4F
	.dc.b	$D7,$FF,$B4,$56,$B4,$F0,$7C,$E7,$F6,$BF,$CF,$BA,$7E,$6B,$FA,$D9
	.dc.b	$FF,$51,$F3,$3F,$33,$7E,$69,$F0,$EF,$7E,$67,$AB,$F0,$25,$EF,$F3
	.dc.b	$4D,$70,$FC,$7C,$DF,$B9,$FF,$43,$C7,$8C,$FF,$5D,$79,$E9,$E0,$FC
	.dc.b	$7B,$C9,$3C,$73,$ED,$93,$EA,$3F,$C0,$5D,$FF,$0F,$87,$E9,$FC,$2B
	.dc.b	$F2,$4F,$E0,$9F,$03,$F6,$6B,$5D,$1F,$42,$B5,$EA,$3F,$5A,$D6,$BF
	.dc.b	$77,$2F,$F2,$43,$FD,$C7,$0E,$2D,$35,$87,$3C,$9E,$8F,$92,$73,$C1
	.dc.b	$FD,$5A,$5F,$A8,$7F,$86,$7E,$3B,$ED,$7B,$7F,$A2,$E2,$FC,$1F,$B3
	.dc.b	$FE,$0D,$F3,$FF,$21,$5A,$D3,$EA,$4F,$F8,$47,$ED,$F3,$A7,$E4,$F1
	.dc.b	$F3,$BF,$DC,$AF,$83,$FE,$09,$79,$63,$DB,$F5,$EF,$14,$F1,$C7,$AF
	.dc.b	$E0,$BF,$E2,$DF,$33,$F1,$DE,$2F,$ED,$A7,$F2,$FC,$1F,$E2,$1F,$E0
	.dc.b	$0F,$D2,$56,$A7,$F4,$73,$FE,$25,$F0,$33,$8F,$E9,$F0,$F2,$BF,$E4
	.dc.b	$0B,$E0,$FF,$87,$5E,$57,$DB,$2F,$14,$F1,$8E,$B9,$7F,$90,$7F,$2B
	.dc.b	$F2,$F8,$1F,$27,$3F,$D5,$E9,$FB,$79,$CD,$FE,$6E,$7F,$DF,$BE,$6A
	.dc.b	$7D,$BF,$17,$FE,$15,$79,$E5,$FE,$2F,$78,$EF,$FB,$6F,$CA,$7C,$13
	.dc.b	$EB,$CF,$C1,$D1,$FE,$53,$F4,$FE,$3F,$5F,$C7,$D3,$F7,$BF,$89,$7F
	.dc.b	$09,$9C,$ED,$F1,$BB,$F1,$97,$C5,$F6,$BF,$C4,$D5,$CA,$3F,$C8,$0F
	.dc.b	$0A,$F1,$47,$01,$CB,$32,$F0,$67,$FE,$65,$F5,$7E,$CF,$4F,$D9,$E0
	.dc.b	$FE,$1F,$E7,$FF,$1B,$9C,$E5,$F2,$77,$C6,$7F,$5F,$3F,$F1,$C5,$72
	.dc.b	$8F,$F2,$93,$C2,$BF,$B9,$70,$F7,$9F,$BD,$97,$7C,$FF,$CE,$DF,$90
	.dc.b	$7E,$2B,$1F,$C5,$61,$E2,$E5,$4E,$63,$E4,$67,$FC,$BE,$38,$63,$FE
	.dc.b	$61,$75,$3F,$D2,$32,$EB,$3E,$78,$FC,$B6,$3F,$E8,$5F,$DD,$3C,$1C
	.dc.b	$9F,$07,$6F,$E7,$51,$8A,$FC,$46,$68,$EF,$87,$F9,$69,$D4,$FF,$90
	.dc.b	$67,$53,$F1,$6F,$F4,$9F,$73,$FE,$93,$F8,$BF,$77,$17,$BB,$2F,$E9
	.dc.b	$11,$8C,$7F,$09,$8A,$3B,$E3,$FE,$A2,$75,$2C,$3A,$4F,$D3,$C3,$F6
	.dc.b	$5F,$75,$FE,$A1,$F5,$BE,$67,$17,$CC,$F3,$BF,$AB,$46,$31,$FC,$06
	.dc.b	$7F,$D5,$31,$FB,$87,$FA,$2E,$75,$3D,$E3,$89,$F8,$1B,$EE,$FF,$61
	.dc.b	$FD,$4F,$ED,$72,$7D,$AF,$C0,$71,$3F,$87,$8F,$BB,$78,$8F,$CA,$C7
	.dc.b	$E5,$9F,$F4,$D5,$72,$9F,$F0,$0E,$BF,$EB,$B8,$C4,$7C,$FF,$B4,$EF
	.dc.b	$C3,$FC,$47,$1F,$EE,$51,$89,$FE,$29,$A2,$3E,$5E,$3F,$EC,$57,$58
	.dc.b	$D8,$F3,$61,$FE,$10,$DB,$FD,$95,$F8,$37,$B3,$83,$B6,$28,$C6,$5D
	.dc.b	$9E,$B8,$7F,$80,$DC,$E3,$CA,$3E,$15,$F5,$74,$7F,$B6,$BF,$08,$7E
	.dc.b	$35,$FD,$9B,$F2,$07,$E3,$8F,$CD,$BF,$C1,$DE,$CF,$0C,$18,$70,$F5
	.dc.b	$98,$70,$BF,$F0,$A7,$E1,$FB,$AD,$4E,$7C,$9F,$26,$1C,$B3,$F0,$67
	.dc.b	$DF,$CA,$FF,$75,$79,$3F,$73,$A7,$E0,$B7,$8B,$E6,$BF,$03,$8F,$FD
	.dc.b	$4D,$8F,$DF,$AF,$B2,$FF,$89,$67,$87,$27,$FB,$D7,$97,$F0,$FD,$3F
	.dc.b	$31,$B8,$FD,$87,$A1,$8E,$5E,$18,$7E,$2E,$BE,$56,$4F,$C9,$F9,$DF
	.dc.b	$F9,$23,$A7,$FC,$69,$18,$C7,$FC,$97,$5B,$C3,$F3,$E7,$FB,$52,$78
	.dc.b	$E2,$8E,$27,$7C,$3F,$7D,$F4,$5E,$8E,$8E,$D8,$7F,$A6,$2B,$F3,$79
	.dc.b	$FE,$72,$7F,$23,$5F,$14,$E1,$87,$5B,$E1,$5C,$0F,$F5,$B6,$99,$FD
	.dc.b	$17,$E2,$EF,$D9,$BA,$FA,$9F,$A0,$6B,$0E,$19,$FC,$5A,$FE,$29,$F8
	.dc.b	$D7,$CE,$C7,$CE,$DC,$3E,$1B,$F6,$4D,$5F,$95,$E1,$8F,$9F,$F1,$AE
	.dc.b	$5B,$38,$3F,$CC,$11,$8C,$2E,$7F,$30,$FF,$2D,$EA,$9F,$16,$DC,$E7
	.dc.b	$ED,$69,$FE,$6E,$F9,$AF,$37,$07,$5C,$3F,$D9,$D5,$A8,$FB,$33,$F8
	.dc.b	$F3,$D9,$EC,$FC,$BF,$C1,$3D,$DC,$3C,$5B,$A7,$F9,$D5,$FC,$B1,$C3
	.dc.b	$0F,$F3,$C7,$3F,$C5,$CF,$C0,$DA,$CC,$FA,$CF,$B1,$1C,$73,$7E,$B1
	.dc.b	$C7,$9B,$FD,$04,$F3,$E1,$B7,$C4,$A7,$DF,$C7,$FB,$AE,$9F,$83,$F2
	.dc.b	$7C,$5E,$7F,$FA,$2D,$F2,$70,$5B,$E5,$9A,$7C,$5F,$A5,$7A,$71,$FF
	.dc.b	$49,$3E,$BE,$0B,$7C,$23,$D1,$E4,$EF,$0F,$F4,$44,$F4,$8F,$67,$9B
	.dc.b	$93,$F7,$87,$BF,$D2,$E0,$7E,$FC,$F8,$5E,$97,$33,$F8,$11,$F1,$7E
	.dc.b	$FB,$B1,$FC,$14,$F9,$1F,$98,$78,$4F,$E1,$07,$CB,$F4,$3C,$87,$F0
	.dc.b	$B3,$E7,$7C,$D7,$9C,$FE,$1C,$7D,$1F,$A2,$F5,$1F,$C4,$4F,$A9,$F0
	.dc.b	$5E,$D3,$F8,$A1,$F5,$FD,$EF,$80,$7F,$17,$3E,$D7,$E3,$BF,$0C,$FE
	.dc.b	$34,$7D,$DF,$DB,$BE,$31,$FC,$74,$FC,$07,$E9,$5F,$24,$FE,$40,$7E
	.dc.b	$0F,$E0,$FE,$61,$FC,$8C,$FC,$37,$EB,$FF,$3C,$FE,$4C,$7E,$27,$F6
	.dc.b	$0F,$A4,$7F,$29,$3F,$19,$FB,$5F,$D5,$3F,$96,$1F,$8F,$F9,$EF,$B0
	.dc.b	$7F,$2F,$3F,$25,$FA,$AF,$DB,$3F,$99,$18,$E6,$FE,$35,$F1,$2F,$C8
	.dc.b	$DF,$9B,$BF,$4B,$FF,$53,$7D,$1B,$D8,$7E,$13,$FE,$DB,$FC,$07,$F0
	.dc.b	$C7,$F3,$D3,$F4,$1F,$67,$F8,$93,$FA,$01,$FA,$3F,$C8,$7F,$18,$7F
	.dc.b	$43,$3F,$4D,$F8,$27,$E3,$CF,$E8,$C7,$EA,$7F,$6A,$FC,$91,$FD,$24
	.dc.b	$FD,$67,$E6,$DF,$95,$3F,$A6,$1F,$AF,$FD,$63,$F3,$07,$F4,$F3,$F6
	.dc.b	$5F,$9C,$7E,$6C,$FE,$A4,$7E,$D7,$F2,$8F,$CF,$1F,$D5,$4F,$DC,$7E
	.dc.b	$6F,$FA,$13,$FA,$C1,$FB,$BF,$81,$FD,$21,$FD,$6C,$FD,$F7,$93,$FA
	.dc.b	$73,$FA,$F1,$FC,$07,$F7,$DF,$D5,$1F,$D8,$4F,$E0,$9C,$BF,$EF,$AE
	.dc.b	$BF,$F6,$FF,$EB,$AF,$92,$7D,$5B,$F1,$7F,$F6,$97,$D3,$B2,$7E,$95
	.dc.b	$FD,$47,$B5,$F8,$B7,$F6,$69,$CC,$7F,$55,$8F,$F3,$47,$B1,$FD,$2F
	.dc.b	$F7,$AF,$E2,$87,$F9,$F3,$D6,$FF,$16,$C6,$6B,$81,$FE,$79,$7F,$A9
	.dc.b	$57,$F4,$5F,$9E,$FF,$43,$E1,$E0,$AF,$E6,$BF,$2D,$A7,$FC,$F2,$79
	.dc.b	$7C,$1D,$DF,$FB,$0B,$F0,$DF,$EC,$38,$EE,$BF,$C7,$BE,$E7,$F1,$FC
	.dc.b	$7A,$47,$F6,$53,$DC,$6F,$3F,$F4,$BC,$F2,$8F,$89,$87,$C3,$3F,$BA
	.dc.b	$7E,$63,$CD,$C4,$23,$85,$BF,$80,$3F,$8F,$1F,$CD,$CF,$F6,$BD,$F8
	.dc.b	$DF,$F6,$CE,$3E,$2B,$FE,$E9,$1A,$C7,$E4,$C7,$D5,$FC,$9B,$FA,$07
	.dc.b	$D1,$FC,$79,$FB,$0F,$F6,$DD,$BF,$BE,$72,$7D,$9F,$D8,$9F,$7E,$FD
	.dc.b	$D1,$F9,$6B,$EC,$3F,$D6,$BF,$12,$F1,$1E,$03,$DE,$FC,$17,$CC,$BC
	.dc.b	$F7,$9A,$FA,$47,$C7,$BF,$38,$7D,$DF,$FA,$27,$27,$D9,$FE,$2F,$F5
	.dc.b	$BF,$F3,$17,$97,$FE,$1D,$E2,$7E,$44,$FC,$69,$F9,$6B,$F0,$B7,$EC
	.dc.b	$FF,$93,$F7,$5F,$8B,$BD,$07,$EE,$AF,$E0,$0F,$6F,$E5,$AF,$C2,$DF
	.dc.b	$72,$F8,$EF,$CE,$76,$7D,$5F,$C8,$BF,$DD,$DF,$B8,$3F,$2E,$7D,$DF
	.dc.b	$C5,$C1,$F9,$F7,$DE,$F3,$7F,$0F,$EA,$FF,$B7,$27,$F1,$3E,$6F,$F6
	.dc.b	$BF,$9A,$C5,$7F,$87,$FE,$3B,$FC,$75,$5F,$8E,$F5,$BE,$F4,$BF,$24
	.dc.b	$FF,$72,$FD,$F7,$FA,$46,$3F,$BA,$63,$FE,$94,$8C,$3D,$43,$87,$FC
	.dc.b	$71,$DF,$FE,$47,$8F,$F0,$BC,$BF,$5A,$7C,$7B,$F7,$67,$E6,$6F,$DB
	.dc.b	$1F,$8B,$BE,$65,$EC,$BF,$56,$7C,$8F,$F0,$3F,$A6,$7E,$5A,$FD,$CB
	.dc.b	$F8,$5F,$E5,$2F,$CB,$7F,$9D,$7E,$7C,$F8,$C7,$EA,$3F,$C2,$FF,$05
	.dc.b	$FE,$7B,$FB,$1B,$E9,$5F,$49,$FC,$3B,$F0,$2F,$C0,$FD,$E3,$F1,$CF
	.dc.b	$E2,$5F,$08,$FE,$01,$FC,$D7,$E5,$4F,$CE,$1F,$BF,$BF,$33,$7E,$E9
	.dc.b	$F0,$7F,$58,$FE,$35,$F7,$2F,$C6,$1F,$97,$3F,$3F,$F2,$FE,$68,$FD
	.dc.b	$29,$F9,$5B,$F5,$17,$E0,$FF,$C8,$FF,$0E,$7E,$C6,$FC,$BD,$FA,$4B
	.dc.b	$F8,$01,$DC,$FC,$77,$E6,$FE,$C8,$FE,$00,$FE,$59,$FA,$57,$E8,$3E
	.dc.b	$B7,$F3,$BF,$0B,$F5,$3D,$9F,$D0,$6E,$FD,$37,$F9,$97,$CE,$3E,$A1
	.dc.b	$F8,$E3,$F6,$CF,$CE,$7E,$C6,$FB,$8F,$87,$F8,$FB,$F2,$6F,$E7,$5F
	.dc.b	$77,$F4,$BE,$27,$D1,$A7,$EE,$B8,$9F,$A0,$7F,$9A,$FA,$FF,$A0,$7E
	.dc.b	$10,$FD,$8B,$D7,$F9,$93,$F5,$6F,$DF,$7D,$2F,$EB,$7F,$1C,$7E,$09
	.dc.b	$FD,$1F,$E3,$3C,$DE,$DB,$D6,$7E,$C0,$FD,$73,$FA,$77,$D4,$7E,$F3
	.dc.b	$93,$DB,$FB,$0F,$D5,$FC,$43,$FC,$F9,$F0,$7F,$BE,$FC,$B9,$E2,$BF
	.dc.b	$62,$FE,$A9,$F8,$87,$D2,$FC,$1F,$F8,$4F,$EB,$AF,$D3,$DF,$B2,$3F
	.dc.b	$00,$7E,$E2,$FC,$E7,$FA,$E7,$EC,$7F,$CD,$FF,$77,$7C,$97,$F8,$4F
	.dc.b	$F3,$4D,$FB,$CF,$DC,$1F,$A1,$FF,$63,$F0,$DF,$B6,$FE,$57,$EB,$1C
	.dc.b	$DF,$C8,$3F,$1F,$FE,$13,$FB,$43,$F5,$BF,$ED,$5F,$86,$FE,$37,$F2
	.dc.b	$BF,$E5,$DF,$95,$BF,$73,$5B,$DC,$FF,$75,$FC,$E7,$F7,$1F,$CC,$9F
	.dc.b	$93,$BF,$7C,$7E,$A8,$FC,$4D,$F9,$4F,$EC,$7F,$0C,$7E,$F3,$FC,$7F
	.dc.b	$EC,$7F,$82,$7E,$0B,$F9,$1F,$4B,$F9,$37,$E8,$7F,$95,$FD,$21,$FB
	.dc.b	$7F,$E7,$2F,$FD,$A1,$F9,$BF,$F8,$03,$D5,$FE,$81,$F8,$0F,$E6,$7F
	.dc.b	$50,$F1,$FD,$5B,$F7,$9F,$F0,$37,$43,$F7,$B7,$EB,$AF,$D3,$1F,$A6
	.dc.b	$7F,$3A,$FD,$CB,$F8,$87,$E0,$3F,$E4,$1D,$FF,$DB,$DF,$6E,$FC,$A7
	.dc.b	$FC,$1F,$F3,$4F,$63,$FC,$2F,$FA,$1B,$ED,$BF,$DC,$9F,$83,$7F,$85
	.dc.b	$70,$E1,$93,$0E,$1F,$91,$30,$E1,$FC,$01,$E7,$D3,$FD,$D9,$C5,$E7
	.dc.b	$E1,$F3,$70,$79,$7F,$00,$7E,$CB,$F0,$ED,$D5,$F0,$FF,$1E,$FE,$95
	.dc.b	$F2,$AE,$AF,$AB,$FB,$F7,$CF,$BF,$E6,$BE,$17,$CF,$FC,$49,$E1,$7C
	.dc.b	$FF,$7F,$F9,$35,$7F,$BF,$7C,$CF,$BB,$E5,$FD,$CF,$2B,$ED,$F9,$BF
	.dc.b	$DB,$7F,$1D,$FC,$55,$87,$A7,$F8,$BF,$C8,$5E,$9F,$E2,$FA,$3F,$15
	.dc.b	$FB,$57,$E6,$2F,$87,$B7,$F9,$8F,$BD,$7B,$7F,$98,$CD,$ED,$7F,$60
	.dc.b	$FD,$4B,$F4,$D8,$77,$F8,$4F,$D3,$7C,$1F,$E9,$7E,$17,$F9,$47,$DC
	.dc.b	$FE,$CE,$BC,$9F,$17,$FB,$3F,$C7,$BE,$C7,$C5,$FE,$CB,$E2,$BF,$98
	.dc.b	$7C,$97,$EF,$BE,$DD,$F2,$5F,$BE,$FC,$6B,$F5,$3F,$47,$FC,$2B,$3F
	.dc.b	$EB,$5F,$53,$FD,$27,$E6,$7F,$84,$FC,$CF,$F0,$8F,$D7,$DF,$45,$FC
	.dc.b	$5B,$AB,$F8,$AF,$D0,$7F,$14,$FC,$63,$F8,$DE,$7F,$C8,$3E,$05,$F4
	.dc.b	$DF,$C8,$3E,$B3,$F9,$97,$D3,$3F,$5F,$FE,$5D,$C1,$F3,$67,$F9,$77
	.dc.b	$D6,$FF,$2D,$F8,$77,$D6,$7F,$2D,$FB,$FF,$F0,$67,$AB,$FC,$EB,$B3
	.dc.b	$FC,$CF,$F8,$F7,$F8,$9F,$F0,$FF,$DD,$7A,$BF,$E1,$6F,$B1,$FE,$91
	.dc.b	$F7,$3F,$D1,$FE,$E3,$FA,$37,$DB,$FF,$45,$FC,$3D,$FA,$9B,$F0,$2F
	.dc.b	$EA,$FE,$1F,$F5,$6F,$A4,$FE,$AB,$F7,$FF,$D5,$3F,$0D,$FE,$C9,$CF
	.dc.b	$F8,$3F,$C2,$3F,$B2,$7E,$44,$FC,$23,$FB,$27,$37,$F6,$3F,$C1,$FF
	.dc.b	$CF,$DD,$FF,$DC,$7F,$37,$7E,$1D,$FD,$C7,$F0,$DF,$EE,$1F,$91,$7E
	.dc.b	$EF,$F1,$1F,$EF,$DB,$7F,$7D,$F9,$37,$DB,$7F,$7D,$FC,$E3,$F6,$5F
	.dc.b	$83,$FF,$83,$BF,$17,$FF,$06,$FE,$2D,$FE,$0C,$F9,$4F,$F0,$5F,$E2
	.dc.b	$DF,$E0,$AA,$FF,$0B,$7E,$30,$21,$F5,$0F,$ED,$07,$FA,$3C,$C9,$AF
	.dc.b	$AF,$FD,$04,$72,$8C,$8F,$DA,$3F,$C5,$9D,$3F,$E6,$9F,$C6,$BB,$47
	.dc.b	$F8,$D2,$77,$E8,$62,$BF,$C6,$87,$1F,$C7,$7F,$C7,$8F,$FA,$2E,$7A
	.dc.b	$79,$BF,$E4,$6F,$CB,$BF,$4F,$FA,$83,$F2,$2F,$F2,$37,$E4,$3F,$E6
	.dc.b	$9F,$DB,$BF,$CA,$B8,$FE,$D7,$F2,$7F,$F2,$AF,$E4,$FF,$E5,$4F,$9C
	.dc.b	$FF,$29,$FB,$7F,$C6,$FF,$29,$FF,$31,$7E,$3F,$FC,$FF,$0F,$F3,$A7
	.dc.b	$AD,$FE,$61,$F0,$3F,$90,$7E,$2B,$FE,$70,$FA,$2F,$F3,$7F,$DC,$FF
	.dc.b	$9B,$BF,$2A,$FF,$36,$FC,$F7,$F2,$CF,$0F,$FC,$FD,$F9,$7F,$F9,$F7
	.dc.b	$F2,$FF,$F3,$E7,$4B,$D0,$FF,$3E,$7D,$4F,$F8,$F3,$7F,$E1,$53,$F2
	.dc.b	$87,$E6,$9F,$56,$75,$3F,$E5,$2C,$BF,$88,$47,$F6,$B8,$E3,$5F,$4B
	.dc.b	$E3,$73,$D9,$FE,$97,$8F,$B3,$5F,$EA,$1A,$DD,$7F,$98,$23,$11,$FE
	.dc.b	$B4,$9F,$F4,$EC,$61,$FF,$30,$9F,$D2,$2B,$F0,$8F,$79,$FE,$97,$F5
	.dc.b	$FF,$EB,$0F,$CC,$FD,$57,$FE,$BC,$8F,$F5,$AD,$74,$7C,$8F,$E4,$23
	.dc.b	$FD,$82,$FE,$4F,$F4,$5F,$24,$EB,$1F,$7B,$FE,$CB,$9E,$58,$7F,$B4
	.dc.b	$0F,$EA,$2F,$89,$F8,$6F,$E6,$33,$F9,$53,$FE,$B3,$9F,$EC,$3F,$9B
	.dc.b	$7F,$A2,$6B,$DE,$FF,$B9,$5F,$D2,$5F,$CE,$8F,$7C,$7F,$B4,$71,$FF
	.dc.b	$77,$D7,$A2,$3E,$6F,$E3,$BC,$DE,$31,$FD,$BB,$F0,$FF,$F8,$01,$FA
	.dc.b	$6F,$FA,$97,$F3,$9F,$F7,$FF,$E7,$3F,$EF,$EC,$7F,$D7,$B7,$E8,$C3
	.dc.b	$F5,$71,$88,$FD,$89,$FE,$D3,$71,$1F,$87,$CF,$F0,$7F,$9C,$79,$A3
	.dc.b	$8B,$E8,$AE,$0E,$63,$CB,$5F,$9F,$9F,$BB,$1F,$E0,$4F,$37,$84,$7F
	.dc.b	$BC,$A7,$53,$DE,$3F,$C1,$3F,$9D,$73,$8F,$B4,$F0,$8F,$EC,$86,$8F
	.dc.b	$64,$78,$3F,$3B,$E0,$FE,$75,$E1,$1F,$9C,$7A,$63,$F9,$EF,$C8,$B7
	.dc.b	$5F,$E1,$87,$75,$4E,$B1,$FE,$73,$7F,$C3,$BF,$5B,$C3,$1F,$F7,$54
	.dc.b	$7E,$10,$F9,$51,$D9,$FF,$11,$FC,$F6,$CE,$B2,$74,$BF,$3C,$E8,$68
	.dc.b	$CF,$E7,$BD,$CE,$27,$0F,$C7,$BC,$47,$43,$9F,$BB,$E5,$3B,$9D,$BE
	.dc.b	$5B,$D0,$78,$8F,$0F,$9B,$F3,$39,$9F,$9F,$BF,$3F,$7E,$04,$FD,$4D
	.dc.b	$E4,$BE,$01,$F0,$BF,$E4,$0C,$BE,$21,$EE,$3D,$BF,$9A,$FC,$73,$E0
	.dc.b	$9F,$03,$F4,$1F,$94,$7C,$43,$E1,$FD,$57,$CD,$3E,$39,$F1,$BF,$08
	.dc.b	$FA,$07,$CA,$3E,$4F,$E7,$1F,$4C,$F9,$A7,$CC,$FD,$0B,$EB,$1F,$40
	.dc.b	$F9,$FF,$A1,$FD,$93,$E9,$9F,$4B,$E2,$FE,$E1,$F5,$8F,$AB,$F0,$DF
	.dc.b	$7C,$FB,$27,$D8,$FC,$AB,$F0,$47,$DC,$3E,$DF,$E6,$FF,$85,$3E,$F9
	.dc.b	$F7,$BE,$0F,$F1,$07,$E0,$8F,$C0,$FD,$77,$E2,$CF,$C2,$9F,$84,$FD
	.dc.b	$13,$F1,$C7,$E2,$0F,$C3,$FB,$5F,$91,$3F,$16,$7E,$2B,$ED,$FF,$28
	.dc.b	$7E,$38,$FC,$6F,$E3,$1F,$97,$3F,$22,$7E,$43,$F1,$AF,$CD,$1F,$94
	.dc.b	$3F,$27,$FA,$2F,$E7,$4F,$CB,$9F,$96,$F8,$BF,$D0,$1F,$AE,$FD,$1B
	.dc.b	$FC,$FE,$FE,$70,$FC,$DF,$E8,$FF,$B2,$FC,$0F,$FD,$06,$FE,$D3,$D9
	.dc.b	$BE,$57,$E9,$CF,$E0,$51,$FA,$2F,$CE,$7F,$73,$FA,$47,$FA,$29,$FD
	.dc.b	$DF,$DE,$DA,$E0,$7F,$02,$F3,$13,$FA,$AF,$80,$FE,$03,$FA,$5F,$FA
	.dc.b	$49,$FE,$07,$FA,$55,$BE,$27,$F5,$E8,$FD,$8F,$E0,$FF,$C3,$F7,$5F
	.dc.b	$FA,$6A,$31,$76,$F8,$94,$FE,$D9,$E9,$FF,$05,$C9,$F9,$1F,$F1,$9D
	.dc.b	$F9,$C6,$2B,$F8,$5C,$2F,$FD,$4F,$5F,$C1,$0F,$E0,$7F,$A5,$FF,$21
	.dc.b	$DB,$8B,$67,$FD,$5A,$FF,$0A,$3F,$93,$7E,$8F,$FD,$E1,$FE,$A5,$8B
	.dc.b	$F4,$CF,$E5,$55,$E2,$6B,$FE,$B6,$7F,$89,$9F,$CB,$FF,$34,$FF,$00
	.dc.b	$BF,$CC,$7E,$C7,$FC,$07,$F8,$A7,$F3,$DE,$1F,$47,$0F,$FA,$FA,$31
	.dc.b	$87,$AF,$1F,$F5,$4E,$FF,$19,$AF,$F8,$46,$D9,$74,$8F,$E7,$30,$C7
	.dc.b	$FC,$23,$A7,$F4,$BD,$FB,$E0,$8C,$63,$DD,$FE,$56,$6F,$1E,$38,$FF
	.dc.b	$44,$BE,$8F,$FB,$53,$F4,$DF,$EA,$DB,$F4,$6E,$1E,$97,$F9,$A9,$AB
	.dc.b	$FF,$11,$BC,$32,$FF,$12,$D7,$F5,$FA,$F2,$B5,$F8,$4F,$F3,$E3,$55
	.dc.b	$C2,$FF,$AB,$CF,$93,$27,$2F,$F6,$DD,$3F,$09,$3F,$88,$8C,$38,$9D
	.dc.b	$D7,$CD,$9C,$DF,$F8,$E1,$DE,$3F,$E3,$AF,$C6,$7F,$B3,$FE,$01,$FE
	.dc.b	$45,$F8,$67,$E9,$CF,$C4,$DF,$A5,$BE,$1D,$FA,$7F,$FB,$EB,$1E,$3D
	.dc.b	$8C,$78,$E1,$C2,$FF,$A1,$6F,$FE,$F1,$DF,$FD,$B1,$A7,$7D,$3B,$69
	.dc.b	$FE,$61,$AD,$57,$FA,$F3,$2E,$95,$AF,$C1,$FD,$13,$FE,$17,$E5,$F8
	.dc.b	$7D,$1E,$38,$C7,$1F,$C3,$DF,$FB,$59,$FE,$15,$AF,$D6,$71,$F2,$AF
	.dc.b	$F4,$4C,$7E,$BF,$8B,$FC,$BB,$F3,$5F,$E1,$BE,$6E,$BB,$23,$19,$7C
	.dc.b	$F7,$FB,$B9,$8C,$78,$E3,$FE,$2B,$BE,$8A,$DF,$A8,$7F,$89,$F0,$EA
	.dc.b	$D7,$ED,$3F,$E0,$23,$17,$FE,$70,$75,$8F,$F9,$CB,$F0,$1F,$F2,$2F
	.dc.b	$47,$E1,$2D,$F9,$EF,$79,$CC,$7E,$23,$E3,$F1,$9F,$EC,$67,$33,$FB
	.dc.b	$08,$C9,$FE,$B0,$BF,$C7,$E1,$FE,$80,$F9,$6F,$F2,$9E,$FD,$FC,$8F
	.dc.b	$64,$63,$0F,$CA,$F9,$38,$C7,$F7,$D3,$99,$ED,$BA,$31,$8F,$FA,$33
	.dc.b	$F5,$1F,$F2,$FF,$0F,$9F,$2F,$91,$18,$C3,$F9,$26,$CC,$F3,$9F,$F0
	.dc.b	$BA,$D5,$7E,$A5,$BC,$3F,$E9,$8F,$BC,$FF,$35,$CF,$9B,$1F,$A9,$18
	.dc.b	$8F,$E6,$18,$75,$D5,$3F,$E3,$43,$52,$F2,$7A,$65,$FE,$80,$8F,$DA
	.dc.b	$FA,$1F,$EA,$6F,$D4,$BF,$D2,$FC,$5F,$47,$26,$7F,$83,$9C,$DF,$B2
	.dc.b	$7C,$11,$FE,$42,$38,$E1,$FE,$71,$77,$2D,$BC,$53,$CC,$FF,$59,$E1
	.dc.b	$F7,$F1,$FF,$5B,$FC,$F7,$FA,$A3,$8F,$EA,$E4,$F3,$FF,$29,$39,$CB
	.dc.b	$E0,$CB,$D0,$CE,$5F,$EB,$EF,$3F,$FD,$71,$7C,$E3,$FD,$08,$F1,$BF
	.dc.b	$A9,$BF,$82,$7F,$D8,$B9,$7E,$13,$3F,$F6,$4F,$D4,$7F,$AE,$B9,$7E
	.dc.b	$06,$8F,$CD,$7F,$4D,$39,$CB,$DD,$3F,$7F,$F5,$3C,$E5,$FE,$D0,$F2
	.dc.b	$7F,$D9,$57,$CE,$3F,$D2,$CF,$1E,$2F,$1C,$70,$1E,$84,$CF,$E8,$69
	.dc.b	$FE,$DC,$FD,$0B,$FD,$9B,$A7,$C2,$FC,$BF,$C9,$D2,$9C,$CF,$E4,$76
	.dc.b	$53,$6F,$DF,$57,$FB,$5B,$C3,$FF,$6F,$3D,$34,$FF,$55,$3C,$AF,$F3
	.dc.b	$5E,$27,$86,$7F,$DD,$78,$FD,$2F,$98,$FF,$77,$FC,$EF,$FA,$73,$77
	.dc.b	$57,$E1,$FD,$DF,$0F,$D5,$FF,$0D,$AD,$4E,$77,$7E,$97,$9C,$FF,$DF
	.dc.b	$F3,$FE,$1A,$7F,$D7,$95,$C3,$3F,$F7,$A3,$79,$3F,$5D,$93,$5F,$1D
	.dc.b	$F2,$BF,$AB,$3F,$CE,$7E,$67,$CD,$D3,$FA,$AE,$D7,$C8,$BE,$05,$FA
	.dc.b	$67,$F2,$3B,$DE,$3F,$BB,$BD,$D3,$F4,$0D,$66,$3F,$DE,$DC,$FF,$EB
	.dc.b	$E7,$FC,$7D,$87,$27,$FD,$AC,$F4,$E6,$FC,$D6,$6E,$AE,$5D,$5F,$EB
	.dc.b	$7B,$FF,$80,$1A,$78,$69,$F0,$FA,$FC,$DF,$E8,$76,$FE,$4B,$B7,$FA
	.dc.b	$F7,$E0,$5F,$D8,$B2,$56,$B9,$BF,$2D,$EC,$29,$FA,$A7,$F4,$1A,$E7
	.dc.b	$E8,$7A,$76,$EC,$7B,$7B,$DA,$7E,$F6,$3F,$DE,$15,$D7,$87,$F1,$D9
	.dc.b	$3F,$10,$E3,$5F,$9B,$D3,$C3,$F2,$9E,$EE,$AF,$C4,$71,$FF,$27,$F0
	.dc.b	$1F,$AA,$BD,$37,$EA,$CF,$C9,$BF,$AB,$61,$C3,$6F,$CE,$DE,$E9,$F8
	.dc.b	$4E,$B5,$F4,$67,$FC,$71,$F6,$73,$31,$E5,$18,$7A,$69,$E2,$D1,$F6
	.dc.b	$5E,$7A,$E7,$A7,$82,$5F,$AA,$7D,$8F,$43,$FD,$FB,$9B,$EC,$1E,$9F
	.dc.b	$F6,$DF,$D1,$BF,$C2,$5E,$73,$F2,$8F,$EE,$18,$70,$BD,$F0,$2F,$7F
	.dc.b	$3B,$AD,$BF,$1D,$9F,$5C,$FE,$B4,$64,$FF,$3A,$60,$F5,$73,$8B,$2E
	.dc.b	$D7,$E3,$EA,$FF,$40,$46,$27,$C1,$1F,$D8,$70,$E9,$8F,$79,$CE,$AF
	.dc.b	$CC,$FC,$1F,$36,$FC,$8B,$FD,$DB,$F9,$1B,$F2,$57,$EA,$EF,$CF,$DF
	.dc.b	$33,$C2,$F7,$5A,$9F,$D1,$67,$E7,$D7,$F5,$99,$75,$E3,$F5,$F0,$7E
	.dc.b	$BE,$53,$E1,$D7,$31,$C3,$2E,$BA,$3F,$3A,$FA,$D3,$9C,$5B,$79,$E9
	.dc.b	$EC,$EA,$BF,$8D,$F8,$6F,$57,$F8,$36,$EF,$D6,$1F,$74,$FD,$53,$E9
	.dc.b	$7B,$7F,$D8,$6F,$75,$AF,$BF,$CE,$3F,$C3,$F2,$FC,$9F,$1F,$AF,$0E
	.dc.b	$11,$83,$F8,$63,$CF,$D8,$F9,$58,$3F,$68,$F3,$4E,$75,$F2,$6D,$FC
	.dc.b	$2B,$F5,$EF,$8D,$FA,$E3,$F1,$9F,$E3,$BF,$A8,$FF,$AF,$BE,$93,$FC
	.dc.b	$F3,$FA,$E7,$F9,$06,$F7,$5A,$FA,$05,$7F,$9F,$BF,$5D,$D7,$07,$F1
	.dc.b	$FC,$BA,$7C,$1F,$E3,$B1,$E3,$F8,$FD,$AD,$65,$CF,$1F,$F5,$14,$72
	.dc.b	$18,$72,$D3,$C9,$F2,$7E,$99,$E3,$92,$7D,$1F,$16,$FD,$4B,$97,$A7
	.dc.b	$FE,$03,$FD,$79,$F6,$EF,$CE,$1F,$AB,$7F,$A5,$BC,$3D,$FC,$5A,$F7
	.dc.b	$5A,$CB,$F3,$9E,$4B,$93,$F1,$D9,$FF,$83,$FE,$63,$D9,$C1,$FC,$27
	.dc.b	$2E,$4F,$F0,$CA,$DC,$75,$D7,$C5,$AF,$C0,$C7,$A5,$FF,$47,$C7,$D1
	.dc.b	$5A,$E1,$F8,$13,$D7,$5F,$2E,$1F,$EA,$EF,$2F,$D3,$18,$D7,$F3,$39
	.dc.b	$74,$BE,$18,$F8,$25,$FA,$BF,$F9,$F7,$F5,$EF,$D4,$7E,$C1,$FD,$7B
	.dc.b	$F6,$27,$EC,$8F,$CC,$DF,$28,$FC,$75,$FB,$3F,$1D,$7F,$AF,$63,$C7
	.dc.b	$0E,$17,$BF,$D8,$F5,$4A,$EC,$FF,$30,$CF,$C3,$D8,$FD,$77,$F8,$AC
	.dc.b	$63,$5F,$1E,$BE,$2F,$D9,$3F,$B1,$CE,$74,$EF,$F9,$3E,$32,$F1,$E4
	.dc.b	$C7,$8E,$BE,$9F,$92,$F1,$FE,$0A,$FC,$E9,$F8,$FB,$F0,$9F,$F1,$2F
	.dc.b	$E6,$6F,$CF,$5F,$B3,$70,$BD,$D6,$A5,$F9,$B6,$77,$F4,$EF,$FE,$01
	.dc.b	$CF,$B6,$78,$84,$63,$0E,$FF,$B2,$7A,$67,$39,$76,$F9,$06,$BF,$C5
	.dc.b	$E7,$C5,$8F,$9E,$73,$A6,$BB,$9C,$9F,$C8,$FE,$BD,$FA,$93,$F2,$A7
	.dc.b	$E7,$6F,$CE,$9F,$82,$7F,$19,$F9,$0F,$AF,$87,$0B,$DD,$6B,$DD,$7E
	.dc.b	$87,$AC,$BA,$70,$7D,$B3,$FC,$41,$87,$83,$F1,$AF,$80,$F3,$D3,$C1
	.dc.b	$A3,$F6,$6F,$81,$39,$CF,$C7,$F8,$F6,$32,$F4,$61,$C3,$47,$E2,$3E
	.dc.b	$1D,$BC,$B9,$FF,$31,$F9,$6C,$5E,$FE,$63,$FD,$B5,$F0,$7F,$D1,$7F
	.dc.b	$48,$FB,$7F,$84,$7F,$57,$FA,$CF,$B3,$E3,$BF,$41,$DF,$2F,$F8,$E6
	.dc.b	$F7,$F2,$9F,$E3,$3F,$13,$F4,$BF,$93,$6A,$95,$D1,$FF,$05,$E7,$DF
	.dc.b	$C6,$7C,$37,$F7,$08,$C6,$BE,$1C,$7F,$67,$7C,$F0,$EF,$AF,$E3,$3E
	.dc.b	$97,$FC,$83,$87,$0D,$3C,$5B,$3E,$EF,$C5,$FB,$27,$8B,$F6,$77,$C0
	.dc.b	$FF,$51,$F8,$87,$D9,$FF,$9B,$3C,$5F,$F3,$9E,$FF,$E5,$5C,$38,$5E
	.dc.b	$EB,$FA,$FD,$6A,$95,$EE,$FC,$F7,$D5,$1C,$78,$19,$62,$37,$18,$C7
	.dc.b	$C5,$9F,$AA,$78,$57,$9B,$E5,$BF,$A9,$56,$67,$8C,$E7,$4F,$4F,$E8
	.dc.b	$7F,$3F,$D9,$FE,$1D,$EC,$7F,$14,$F7,$1E,$DE,$7D,$9F,$E0,$1C,$7F
	.dc.b	$C4,$A7,$F8,$BF,$17,$AF,$0F,$E2,$3A,$7F,$9C,$E3,$17,$FC,$46,$3E
	.dc.b	$C9,$FC,$2B,$D9,$FE,$13,$ED,$79,$6F,$C3,$ED,$3D,$9C,$61,$E0,$D7
	.dc.b	$C9,$ED,$7E,$17,$E8,$7F,$5B,$F1,$FF,$E2,$3F,$CE,$DF,$B4,$3F,$2C
	.dc.b	$7C,$1E,$FC,$C5,$EE,$B5,$F8,$2E,$63,$E7,$63,$FB,$8F,$8F,$B7,$C3
	.dc.b	$E8,$59,$F6,$FA,$BF,$99,$7E,$29,$CE,$4F,$82,$CE,$1C,$70,$F2,$EA
	.dc.b	$F6,$3F,$3D,$F2,$AF,$8D,$7E,$38,$FC,$29,$FA,$C7,$F4,$4F,$CB,$1F
	.dc.b	$7B,$C2,$F7,$5A,$F8,$D5,$7F,$4C,$F0,$3F,$1D,$97,$5F,$C3,$FF,$14
	.dc.b	$F7,$CE,$E3,$CB,$3F,$EA,$3F,$05,$F8,$7A,$D6,$7D,$7E,$7F,$17,$BD
	.dc.b	$3B,$FD,$BF,$ED,$FE,$4B,$F9,$FF,$1B,$F5,$EF,$F0,$57,$D0,$3F,$28
	.dc.b	$FE,$C1,$C3,$FA,$1B,$DD,$6A,$5F,$B4,$F3,$F9,$07,$CD,$FE,$86,$FA
	.dc.b	$F9,$7F,$8A,$3F,$C2,$33,$FC,$BB,$0E,$FF,$82,$FE,$7E,$BA,$FD,$07
	.dc.b	$F9,$3B,$F2,$5C,$5E,$E3,$FC,$A3,$9F,$FA,$6B,$4F,$DA,$7E,$39,$FE
	.dc.b	$F9,$FC,$65,$F9,$2B,$F0,$DF,$F0,$D7,$EA,$4F,$D3,$FF,$8B,$5E,$EB
	.dc.b	$5F,$42,$AF,$EF,$98,$F4,$FC,$1D,$F8,$E7,$EE,$F0,$ED,$C7,$EB,$79
	.dc.b	$67,$FE,$27,$EB,$FD,$C4,$63,$2F,$F3,$3D,$78,$30,$6B,$E6,$C3,$CB
	.dc.b	$FA,$1F,$F2,$95,$F8,$D3,$F6,$A7,$E7,$3F,$E8,$6F,$C6,$9F,$A1,$BF
	.dc.b	$33,$E1,$7B,$AD,$4B,$F3,$AC,$E1,$FC,$4F,$1E,$BF,$23,$70,$E1,$F7
	.dc.b	$8C,$DF,$9D,$7F,$11,$BF,$1E,$1E,$2A,$D7,$EC,$99,$C3,$8E,$1E,$4F
	.dc.b	$CE,$3F,$A1,$70,$7F,$6E,$87,$BF,$FC,$53,$F6,$B7,$E3,$CF,$D3,$F7
	.dc.b	$F4,$9E,$1A,$D4,$BF,$22,$CF,$CD,$FF,$2B,$CB,$A7,$4F,$E2,$DE,$D7
	.dc.b	$D7,$CF,$FE,$51,$18,$F8,$D6,$9F,$B2,$BE,$FA,$F9,$27,$39,$FF,$A0
	.dc.b	$7C,$0F,$E7,$D8,$71,$C3,$CF,$FB,$67,$F5,$DF,$6B,$FB,$A7,$A1,$FE
	.dc.b	$05,$BF,$E8,$91,$8F,$A6,$FF,$01,$C6,$38,$7F,$C0,$7A,$FF,$2D,$F9
	.dc.b	$BF,$F0,$94,$3E,$A7,$FA,$36,$1D,$3D,$8F,$E8,$0F,$6F,$A8,$B5,$FE
	.dc.b	$00,$FA,$9F,$F0,$87,$D4,$FF,$6F,$FC,$31,$F9,$AB,$F1,$57,$ED,$AF
	.dc.b	$8A,$FF,$02,$FE,$51,$FB,$BE,$1F,$EB,$58,$70,$BD,$D6,$BF,$39,$E9
	.dc.b	$8F,$3E,$2F,$DC,$5F,$6F,$D2,$BF,$AE,$3C,$73,$ED,$A7,$F9,$E7,$8F
	.dc.b	$FD,$A1,$39,$CB,$FC,$2B,$D9,$FD,$6E,$31,$4D,$9F,$9D,$7F,$82,$7F
	.dc.b	$08,$7E,$DD,$FD,$03,$F2,$1F,$F2,$9F,$ED,$4F,$CC,$BF,$B2,$56,$BE
	.dc.b	$59,$5A,$FA,$E7,$E4,$1C,$F1,$FF,$51,$71,$FF,$11,$C3,$83,$63,$B7
	.dc.b	$C0,$B3,$61,$DA,$73,$0F,$CA,$71,$9F,$82,$F8,$69,$FB,$3E,$9F,$F0
	.dc.b	$36,$7E,$5D,$7D,$33,$FB,$39,$F2,$E8,$1D,$8F,$CA,$3F,$49,$F9,$D3
	.dc.b	$E7,$3F,$AC,$FD,$EF,$F8,$FB,$0E,$38,$70,$BD,$D6,$BE,$F5,$5A,$C7
	.dc.b	$A7,$A1,$DF,$0F,$EF,$DC,$96,$7C,$1F,$55,$99,$D5,$F1,$CB,$AF,$B7
	.dc.b	$8C,$38,$47,$83,$E6,$BF,$59,$5B,$C9,$B7,$A6,$7F,$CE,$3F,$77,$E4
	.dc.b	$C3,$C5,$FB,$7B,$D4,$FF,$39,$FE,$A4,$FD,$13,$FA,$5F,$EB,$7F,$E9
	.dc.b	$7D,$3D,$F7,$BF,$2D,$7B,$F9,$45,$FE,$FA,$F7,$8F,$F9,$FB,$87,$D8
	.dc.b	$78,$4F,$A1,$8F,$87,$F0,$EC,$C6,$44,$77,$C7,$B7,$EE,$1F,$5F,$0E
	.dc.b	$B9,$F8,$B4,$FE,$C3,$F8,$16,$76,$F6,$E9,$EC,$9F,$5F,$E1,$5D,$7F
	.dc.b	$33,$7E,$AC,$FD,$81,$F0,$8F,$C9,$9F,$B8,$BC,$D6,$1C,$70,$E1,$7B
	.dc.b	$AD,$7E,$9F,$9C,$7D,$7C,$1F,$86,$C3,$BF,$EA,$78,$DF,$F0,$51,$E2
	.dc.b	$AF,$0E,$9E,$0F,$A9,$F5,$73,$7E,$56,$B5,$A7,$7F,$47,$33,$E4,$DB
	.dc.b	$C7,$87,$A3,$7F,$F4,$0E,$BF,$96,$FD,$A5,$F6,$4F,$AA,$7E,$BE,$FD
	.dc.b	$AB,$F5,$FC,$DF,$EB,$2C,$38,$5E,$EB,$5E,$9F,$E1,$7F,$66,$FA,$D9
	.dc.b	$74,$FC,$D3,$D9,$BF,$EA,$E3,$BE,$98,$84,$63,$0E,$7D,$BE,$A8,$C6
	.dc.b	$BF,$C0,$3B,$3D,$BE,$23,$F4,$BF,$F3,$5F,$E7,$CF,$C8,$BF,$B9,$FD
	.dc.b	$AB,$F3,$5F,$C4,$5E,$FE,$3D,$7B,$F9,$BF,$F2,$EF,$DD,$6B,$27,$D8
	.dc.b	$79,$70,$E1,$C3,$F9,$6E,$7F,$DE,$C7,$3C,$79,$7D,$BF,$93,$0F,$1E
	.dc.b	$DF,$E0,$4F,$B9,$FA,$7F,$AD,$FF,$82,$3F,$5B,$7E,$BA,$FC,$9B,$FB
	.dc.b	$7F,$0F,$FB,$83,$5F,$E1,$37,$BA,$D7,$CC,$7F,$A2,$3D,$1F,$D1,$4F
	.dc.b	$F8,$4B,$E4,$3E,$53,$D7,$87,$CD,$E5,$7F,$0B,$8E,$59,$FF,$7C,$FC
	.dc.b	$47,$C7,$F3,$CD,$BC,$73,$FE,$00,$F9,$2F,$C4,$FE,$D4,$FC,$79,$FB
	.dc.b	$97,$FB,$DB,$F0,$07,$DE,$FF,$05,$FD,$3B,$F6,$97,$BA,$D4,$FF,$09
	.dc.b	$FC,$93,$38,$3D,$6F,$C0,$5E,$F8,$7C,$E7,$F2,$F7,$97,$C4,$7F,$A3
	.dc.b	$F0,$FE,$79,$3C,$B1,$EB,$FA,$2F,$9B,$1F,$2D,$EF,$DB,$7E,$15,$FC
	.dc.b	$1F,$F7,$37,$EB,$3F,$E3,$8F,$CF,$1F,$6A,$E7,$7E,$91,$FE,$82,$EF
	.dc.b	$FE,$73,$87,$0E,$96,$1C,$2F,$7F,$82,$75,$CB,$FC,$87,$C3,$EA,$18
	.dc.b	$78,$3F,$43,$FC,$F3,$CF,$0E,$5C,$3F,$B6,$8C,$6B,$E3,$F9,$7F,$DC
	.dc.b	$FE,$0D,$FE,$F6,$F8,$37,$E6,$2F,$A2,$7E,$99,$F9,$6F,$BF,$FE,$49
	.dc.b	$7B,$AD,$4F,$E6,$BF,$64,$CF,$E3,$6F,$AC,$FD,$85,$F5,$E1,$F4,$3C
	.dc.b	$72,$EB,$E1,$F8,$AD,$A7,$8A,$7C,$3C,$B9,$8C,$67,$FA,$8F,$A0,$FE
	.dc.b	$6D,$FB,$A7,$F9,$93,$F6,$6F,$F0,$FF,$E9,$CF,$C0,$7F,$BB,$7E,$D3
	.dc.b	$AF,$6D,$5A,$9F,$EF,$5E,$CE,$63,$F8,$36,$1F,$E6,$EF,$32,$DF,$0F
	.dc.b	$AC,$65,$D7,$A3,$F9,$0D,$F8,$67,$39,$3A,$FF,$92,$FE,$42,$FC,$F9
	.dc.b	$F9,$3F,$F7,$DF,$09,$F4,$7F,$E8,$2F,$D7,$BF,$A8,$5E,$EB,$53,$FE
	.dc.b	$86,$FD,$BB,$38,$3F,$2A,$FE,$39,$7B,$ED,$FC,$23,$C7,$E8,$D9,$F6
	.dc.b	$D3,$C7,$87,$8A,$73,$97,$E8,$3C,$9F,$F0,$5F,$D6,$BF,$2E,$7E,$EC
	.dc.b	$FD,$5D,$FB,$6F,$FA,$97,$E4,$5F,$AD,$7F,$67,$C3,$85,$EE,$B5,$F7
	.dc.b	$19,$C5,$FB,$B7,$F3,$3B,$DF,$17,$E6,$DE,$39,$F6,$FC,$45,$F2,$9F
	.dc.b	$D3,$A7,$39,$7F,$AD,$7F,$31,$FE,$A3,$FB,$93,$F1,$87,$97,$E9,$FD
	.dc.b	$DD,$FA,$8F,$FA,$A7,$F4,$DF,$EF,$97,$BA,$D4,$FF,$AB,$7F,$57,$CE
	.dc.b	$1C,$B7,$FD,$75,$F5,$FC,$3B,$FA,$53,$C7,$89,$9F,$6F,$CE,$BC,$78
	.dc.b	$78,$A7,$39,$7F,$9A,$7A,$3F,$C6,$F8,$71,$C3,$CB,$F8,$83,$EA,$3F
	.dc.b	$C2,$1F,$B8,$BE,$51,$F9,$9B,$F6,$27,$E6,$AF,$D3,$9F,$AD,$3F,$41
	.dc.b	$E3,$87,$0B,$DD,$6B,$F4,$CC,$E3,$FD,$8F,$2E,$DF,$A3,$DC,$38,$67
	.dc.b	$D7,$D2,$FE,$FB,$5E,$3B,$F1,$4F,$87,$C9,$FF,$00,$FD,$8F,$E1,$BE
	.dc.b	$59,$FA,$9B,$F3,$67,$EF,$0F,$41,$F8,$97,$F8,$B7,$0E,$17,$BA,$D7
	.dc.b	$E0,$F9,$F3,$19,$7F,$79,$BF,$C9,$FD,$5D,$9C,$E1,$C7,$C8,$F1,$C6
	.dc.b	$34,$F0,$FE,$3D,$8C,$7C,$B1,$E4,$D5,$E2,$FF,$8F,$FF,$78,$7A,$FF
	.dc.b	$E9,$3F,$D0,$3F,$D0,$56,$FC,$61,$F8,$36,$EB,$5F,$0D,$EF,$9F,$F3
	.dc.b	$A7,$D2,$7F,$9B,$71,$E7,$F6,$1F,$E1,$17,$AF,$E2,$57,$0E,$39,$7F
	.dc.b	$A3,$BC,$EE,$92,$D3,$C3,$DB,$FE,$2E,$F1,$DF,$95,$BF,$3F,$79,$5F
	.dc.b	$86,$FC,$FD,$F4,$FF,$DF,$FC,$DF,$F4,$2E,$1C,$2F,$7E,$1A,$F7,$F8
	.dc.b	$A7,$F9,$7E,$BA,$FE,$F3,$FE,$27,$78,$6F,$FD,$3D,$E5,$9F,$7F,$A1
	.dc.b	$F2,$6B,$E3,$8F,$15,$7F,$91,$3F,$6E,$FF,$53,$7E,$B5,$FC,$2F,$F5
	.dc.b	$17,$DF,$FF,$AD,$3C,$6F,$F3,$75,$6B,$D6,$56,$BA,$DF,$75,$AE,$66
	.dc.b	$5F,$E3,$BC,$3F,$D6,$5F,$92,$59,$CE,$3C,$BF,$46,$FF,$4B,$6B,$FE
	.dc.b	$AD,$BE,$D9,$F8,$BF,$09,$FF,$2C,$FC,$7F,$6F,$CC,$BB,$FC,$D7,$E1
	.dc.b	$3F,$60,$7E,$97,$FE,$83,$AD,$70,$7F,$64,$F2,$15,$FC,$47,$BB,$FD
	.dc.b	$67,$5E,$8F,$D2,$FF,$CA,$0F,$5D,$BF,$BC,$3C,$B2,$FC,$8F,$C0,$78
	.dc.b	$B2,$69,$E2,$CF,$EF,$E1,$FE,$93,$F1,$7F,$A6,$5F,$F5,$DF,$B4,$7B
	.dc.b	$7F,$E5,$5F,$9C,$D7,$B1,$FC,$83,$E5,$FF,$C4,$53,$FE,$59,$E7,$CE
	.dc.b	$DF,$E6,$3C,$B9,$E7,$FE,$0B,$7A,$DF,$0F,$C6,$BF,$CF,$B1,$8C,$FB
	.dc.b	$FC,$3F,$FA,$03,$F3,$7F,$D2,$7E,$84,$FC,$6B,$F7,$FF,$77,$FE,$DB
	.dc.b	$E8,$FF,$55,$56,$A7,$3A,$3F,$61,$CF,$A8,$C7,$FC,$ED,$7F,$AF,$FD
	.dc.b	$B6,$CE,$70,$E3,$87,$0B,$F0,$7C,$03,$3F,$F1,$9C,$E6,$BB,$F3,$7F
	.dc.b	$AE,$FF,$09,$7E,$4E,$FD,$A5,$FA,$A7,$EB,$FF,$3A,$FF,$1C,$FE,$AD
	.dc.b	$BA,$D4,$E7,$EE,$5F,$BD,$67,$F7,$9F,$F4,$66,$5D,$32,$B8,$70,$F6
	.dc.b	$D9,$BE,$EF,$FD,$4F,$7B,$CF,$C1,$FA,$77,$FA,$AB,$8D,$F0,$1F,$17
	.dc.b	$D5,$7E,$17,$FB,$FF,$93,$FF,$45,$F9,$9F,$F0,$02,$F7,$D9,$FC,$0A
	.dc.b	$B5,$FA,$07,$FB,$66,$79,$F7,$7F,$A8,$5D,$F4,$FF,$92,$5E,$39,$76
	.dc.b	$F8,$0F,$F2,$B4,$F1,$D3,$C5,$8F,$F0,$FF,$95,$FF,$5F,$F7,$BF,$3E
	.dc.b	$FE,$DF,$F8,$13,$F7,$B7,$EF,$9F,$EB,$EF,$B6,$FF,$75,$5E,$FE,$83
	.dc.b	$FD,$FD,$5A,$FC,$6B,$FD,$67,$C8,$F7,$DB,$CC,$3B,$79,$3F,$EB,$97
	.dc.b	$94,$7F,$98,$5E,$9A,$78,$74,$FE,$DD,$7C,$A9,$5A,$CB,$CB,$F7,$3E
	.dc.b	$AD,$BD,$31,$E3,$FB,$8F,$F8,$01,$F6,$6F,$DD,$DF,$47,$EF,$FC,$53
	.dc.b	$FC,$C9,$FA,$83,$F7,$0D,$D6,$B2,$56,$BE,$E3,$38,$7F,$B4,$71,$E9
	.dc.b	$F9,$FD,$C3,$85,$FF,$9B,$FF,$47,$FF,$75,$D7,$86,$FC,$1F,$80,$7C
	.dc.b	$FE,$1F,$2F,$E6,$5F,$E8,$6F,$9E,$7E,$5D,$FE,$01,$D3,$F8,$97,$2F
	.dc.b	$C1,$C9,$7B,$AD,$7C,$3F,$2F,$DF,$7C,$93,$FE,$90,$CB,$FD,$19,$7C
	.dc.b	$3F,$0C,$67,$DB,$C0,$7D,$87,$9B,$0E,$91,$F4,$30,$FB,$3F,$7E,$FA
	.dc.b	$0F,$D3,$FD,$1B,$F3,$AE,$3F,$2B,$F2,$7C,$3F,$C0,$61,$C2,$F7,$5A
	.dc.b	$E9,$56,$AB,$FC,$C1,$5F,$B2,$7F,$D4,$58,$F7,$EA,$FF,$7F,$3E,$1F
	.dc.b	$67,$FC,$13,$8F,$A3,$D4,$FC,$6E,$3C,$EB,$5A,$79,$77,$79,$FA,$E7
	.dc.b	$FC,$57,$CF,$3D,$AF,$F0,$E7,$E3,$4F,$D0,$1F,$B6,$7F,$AC,$BF,$60
	.dc.b	$F3,$E1,$C2,$F7,$5A,$FC,$45,$FB,$66,$B9,$3F,$3F,$9F,$5F,$77,$F3
	.dc.b	$11,$FD,$60,$FF,$20,$E1,$FE,$2B,$7F,$D8,$13,$E3,$FA,$7F,$F7,$EE
	.dc.b	$3E,$2E,$CF,$45,$74,$BF,$0E,$AF,$AC,$CC,$BE,$D1,$FA,$97,$F2,$9F
	.dc.b	$9B,$7E,$F5,$FC,$4B,$C4,$FD,$0F,$4F,$D1,$BD,$FC,$5A,$F7,$ED,$AF
	.dc.b	$F8,$15,$EF,$17,$EF,$5E,$CC,$FC,$39,$72,$7F,$DB,$4F,$4C,$79,$FB
	.dc.b	$9A,$BF,$26,$9F,$C9,$3E,$F7,$8F,$86,$FC,$83,$FC,$CB,$F9,$67,$FB
	.dc.b	$B3,$E2,$5F,$B7,$7C,$DD,$7D,$D7,$BE,$95,$EF,$E6,$35,$BF,$E5,$FE
	.dc.b	$37,$CD,$86,$30,$EC,$F9,$9E,$57,$FE,$FA,$F9,$BF,$1E,$9E,$2F,$C6
	.dc.b	$BF,$D5,$5E,$0F,$F1,$1C,$BF,$45,$39,$D9,$F8,$CF,$97,$EF,$9F,$BD
	.dc.b	$7F,$94,$BF,$64,$7E,$45,$FE,$83,$FC,$35,$FB,$A6,$EB,$53,$9F,$DF
	.dc.b	$B1,$F4,$DE,$8E,$EF,$D3,$65,$D3,$F0,$8C,$70,$7F,$23,$8E,$F4,$D3
	.dc.b	$BF,$E0,$FF,$EA,$EE,$6E,$71,$8E,$BF,$C3,$DB,$EE,$DF,$91,$3F,$57
	.dc.b	$FC,$57,$EC,$6F,$D4,$9F,$6D,$ED,$BD,$D6,$A5,$F9,$36,$70,$F2,$E9
	.dc.b	$FD,$03,$2E,$DF,$53,$88,$FF,$4F,$63,$CB,$AD,$A3,$5F,$2C,$E6,$3F
	.dc.b	$B3,$D7,$E0,$7E,$65,$F9,$03,$F3,$8F,$ED,$BB,$BF,$3F,$7E,$D9,$FA
	.dc.b	$2E,$8F,$E1,$D8,$70,$BD,$D6,$BB,$3A,$62,$FC,$87,$CD,$7B,$CF,$F3
	.dc.b	$D8,$F2,$AF,$07,$A1,$F1,$F1,$7E,$65,$FB,$2C,$3B,$4E,$7D,$FF,$DA
	.dc.b	$FE,$06,$FC,$09,$FA,$17,$F6,$4F,$D0,$9F,$6D,$FC,$CB,$CB,$FC,$F5
	.dc.b	$EE,$B5,$3F,$CD,$BC,$BC,$E1,$CB,$83,$CB,$E1,$FF,$65,$C3,$B7,$E9
	.dc.b	$7F,$3D,$E5,$2C,$FB,$E9,$E4,$C3,$94,$78,$B7,$C6,$9E,$0A,$FC,$37
	.dc.b	$17,$F3,$7C,$5F,$97,$3F,$2B,$7E,$EF,$FC,$43,$F1,$D7,$EB,$BF,$E6
	.dc.b	$0C,$38,$5E,$EB,$52,$FC,$0F,$DF,$3C,$F7,$F5,$9F,$E0,$8C,$BB,$FA
	.dc.b	$9E,$B9,$E5,$ED,$33,$FE,$31,$7C,$B4,$F1,$CE,$7C,$BF,$E4,$BF,$96
	.dc.b	$BC,$F7,$E6,$BF,$D8,$BE,$B5,$FB,$F7,$F8,$5B,$E1,$7F,$87,$5E,$EB
	.dc.b	$53,$FE,$1A,$FC,$8F,$38,$3F,$45,$FC,$E5,$EF,$5F,$48,$79,$7C,$4B
	.dc.b	$3E,$F3,$E4,$BF,$1C,$E7,$1F,$F2,$FF,$37,$F6,$5C,$38,$E1,$C3,$F3
	.dc.b	$3F,$2D,$FF,$B2,$3F,$7A,$7E,$4B,$FE,$AC,$F1,$9F,$B8,$EB,$F1,$35
	.dc.b	$6B,$F0,$CF,$D3,$7D,$6E,$78,$3C,$FA,$FF,$81,$3F,$24,$F5,$E9,$D7
	.dc.b	$4F,$F5,$BF,$85,$F2,$A3,$18,$FF,$82,$30,$FE,$69,$FB,$8F,$F9,$93
	.dc.b	$F3,$B7,$D6,$FF,$9C,$B8,$3F,$D3,$7F,$98,$6B,$E6,$15,$AF,$C0,$7F
	.dc.b	$71,$F8,$73,$F7,$3D,$4F,$F0,$7C,$BF,$CD,$FF,$41,$70,$E3,$9F,$5F
	.dc.b	$A9,$FE,$C3,$5E,$18,$F0,$78,$58,$AD,$E7,$DB,$F1,$6F,$3F,$E1,$EF
	.dc.b	$C6,$5F,$74,$FD,$B7,$FD,$13,$C8,$F1,$F7,$5A,$9C,$FE,$E9,$8E,$6F
	.dc.b	$BF,$BB,$27,$E3,$1E,$97,$B6,$7D,$75,$7B,$5E,$89,$CE,$5F,$E9,$8D
	.dc.b	$3D,$7F,$9E,$BE,$A5,$E2,$3F,$7B,$FF,$48,$62,$FD,$A7,$E8,$DE,$EB
	.dc.b	$5F,$5D,$DB,$F6,$FC,$FD,$3F,$CB,$AE,$9C,$5E,$87,$86,$5D,$23,$9E
	.dc.b	$7E,$08,$FF,$4F,$69,$FE,$0E,$F8,$17,$EE,$0F,$D7,$7F,$D0,$BF,$7B
	.dc.b	$F8,$EF,$CF,$1F,$25,$FD,$DE,$F7,$5A,$F3,$54,$FD,$33,$58,$F3,$D7
	.dc.b	$F5,$38,$76,$FD,$2F,$FA,$43,$CB,$F0,$86,$9D,$EB,$8E,$7D,$7E,$0F
	.dc.b	$CD,$EE,$FF,$89,$7E,$55,$F9,$23,$E8,$5F,$8B,$7F,$BF,$BD,$AF,$F2
	.dc.b	$96,$BF,$C2,$FA,$3F,$C5,$57,$BA,$FF,$6E,$56,$B1,$7E,$1F,$E4,$7F
	.dc.b	$C6,$5A,$FA,$1F,$05,$F7,$EB,$F1,$E3,$D3,$37,$E2,$98,$D7,$CB,$1E
	.dc.b	$4F,$C6,$3F,$C3,$D3,$FD,$EF,$E2,$5F,$AA,$4F,$8B,$FE,$93,$F8,$76
	.dc.b	$BE,$8F,$F9,$7C,$E7,$E3,$9F,$82,$E7,$0F,$EE,$18,$FF,$81,$7E,$CE
	.dc.b	$E1,$C3,$0F,$E7,$3E,$67,$F9,$2A,$BC,$11,$DF,$D6,$C7,$8B,$FE,$05
	.dc.b	$FC,$47,$FC,$6D,$F9,$4B,$F4,$A7,$E5,$3E,$8A,$D7,$D8,$2B,$5F,$B4
	.dc.b	$67,$E8,$FC,$75,$FE,$D4,$EC,$FF,$32,$F7,$7F,$A0,$F6,$FD,$05,$34
	.dc.b	$EF,$F9,$C7,$8A,$31,$9F,$5F,$A7,$F4,$71,$BF,$74,$FF,$01,$74,$BF
	.dc.b	$26,$7E,$B0,$FD,$1B,$E6,$DE,$FE,$27,$FA,$CD,$6B,$C2,$CE,$3F,$22
	.dc.b	$BF,$C1,$7C,$3F,$CF,$38,$3F,$CC,$D3,$DA,$FF,$C3,$13,$C3,$0E,$7B
	.dc.b	$FF,$9D,$3F,$59,$7E,$D8,$FD,$AA,$FE,$14,$FD,$33,$FD,$2B,$EC,$7F
	.dc.b	$97,$6F,$75,$A9,$FF,$4C,$F7,$67,$0E,$59,$FF,$4E,$BE,$BE,$9F,$F8
	.dc.b	$B5,$E3,$2C,$BB,$7C,$57,$8E,$3C,$46,$71,$7C,$AF,$9E,$DF,$88,$BF
	.dc.b	$35,$FF,$54,$FE,$25,$E9,$FD,$BB,$CD,$5A,$9C,$EF,$F8,$3E,$FF,$F5
	.dc.b	$E4,$8F,$F2,$4E,$FE,$77,$AB,$71,$ED,$AB,$2F,$E1,$B3,$DA,$73,$1F
	.dc.b	$EF,$28,$FF,$41,$7E,$96,$F8,$E7,$CD,$7F,$3A,$FD,$61,$FA,$C7,$F1
	.dc.b	$5F,$DE,$B7,$5A,$F3,$15,$AF,$73,$FC,$03,$BF,$FB,$4F,$2E,$99,$7F
	.dc.b	$95,$DE,$DF,$9D,$DC,$79,$61,$DB,$E4,$FF,$D6,$35,$C3,$2E,$BA,$3F
	.dc.b	$55,$CE,$1C,$23,$F9,$47,$EF,$FE,$AE,$B7,$E4,$EF,$E0,$0F,$EA,$1F
	.dc.b	$AF,$BD,$47,$EA,$BB,$AD,$7C,$62,$B5,$86,$79,$BF,$C2,$3B,$FF,$36
	.dc.b	$FD,$5F,$D5,$3D,$B3,$EB,$9F,$A2,$31,$8F,$3E,$0C,$67,$DB,$3F,$17
	.dc.b	$EB,$5F,$43,$EC,$7F,$D1,$DF,$A0,$B6,$2F,$DC,$DF,$BA,$6F,$DD,$CD
	.dc.b	$6A,$5F,$AC,$66,$3C,$F8,$7F,$0A,$6D,$70,$C3,$E1,$47,$37,$C0,$7F
	.dc.b	$B7,$9F,$F6,$27,$DA,$3F,$7E,$7E,$F5,$F8,$2F,$D3,$5F,$8D,$3F,$75
	.dc.b	$FE,$87,$E1,$3F,$00,$F5,$65,$FE,$5B,$C7,$8F,$6B,$1E,$38,$7F,$50
	.dc.b	$C3,$85,$FC,$1C,$FC,$19,$FF,$A7,$5F,$17,$DC,$DC,$B9,$E3,$E2,$F9
	.dc.b	$4F,$F7,$BD,$70,$CF,$B6,$BF,$D9,$32,$F5,$FE,$97,$FE,$6E,$FD,$8B
	.dc.b	$CD,$F3,$9F,$9F,$FD,$AF,$3F,$7C,$9C,$F8,$AF,$DB,$33,$8F,$F9,$4B
	.dc.b	$46,$4F,$B4,$F3,$3D,$79,$7F,$C8,$BF,$81,$F9,$23,$1A,$FC,$DE,$5F
	.dc.b	$7E,$3F,$E7,$FF,$D1,$7F,$CC,$FF,$BC,$BF,$7E,$47,$F5,$0F,$9B,$FA
	.dc.b	$6E,$A3,$3F,$2B,$F2,$74,$F9,$FB,$70,$FB,$FF,$91,$6F,$56,$C3,$2E
	.dc.b	$D2,$FC,$83,$E4,$BD,$47,$E9,$5F,$33,$F4,$DF,$EA,$9F,$85,$3F,$1F
	.dc.b	$7E,$57,$7F,$02,$FC,$3D,$6A,$72,$7F,$22,$E3,$F7,$EC,$9F,$EB,$DC
	.dc.b	$7F,$2D,$7C,$1F,$DF,$79,$63,$FE,$3C,$9E,$BF,$38,$CF,$C5,$2F,$6B
	.dc.b	$CD,$39,$CF,$C3,$F9,$A3,$4F,$9F,$E5,$FF,$B6,$FE,$C1,$F9,$8F,$8F
	.dc.b	$F7,$EF,$F0,$8C,$6F,$A3,$F3,$35,$BA,$D7,$D4,$FC,$9C,$39,$FC,$8B
	.dc.b	$0F,$F7,$95,$FF,$B9,$FE,$D2,$CE,$70,$ED,$F5,$3F,$BD,$C3,$8E,$1E
	.dc.b	$2E,$36,$9F,$43,$8F,$C4,$7F,$03,$CF,$FD,$EB,$F9,$A7,$EC,$FF,$06
	.dc.b	$74,$3F,$64,$7E,$D5,$C3,$06,$1C,$3D,$1F,$F4,$DD,$EF,$D7,$D4,$F4
	.dc.b	$AF,$F5,$6A,$FC,$5B,$F8,$7C,$61,$67,$E1,$B7,$CE,$FA,$F9,$BE,$3D
	.dc.b	$6B,$2E,$FF,$1F,$9C,$7D,$5F,$A3,$BF,$32,$7E,$9F,$FE,$1B,$F6,$5F
	.dc.b	$B5,$3F,$32,$F2,$DE,$EB,$53,$FE,$EF,$E8,$FB,$73,$FD,$5F,$1F,$4B
	.dc.b	$D5,$B8,$F6,$FA,$CF,$F6,$AF,$57,$8E,$7E,$2D,$75,$9F,$F6,$3C,$78
	.dc.b	$2B,$86,$7E,$4F,$CC,$7E,$67,$9E,$FD,$A1,$FA,$DF,$E4,$7F,$0B,$7E
	.dc.b	$2E,$FB,$E7,$EB,$3C,$2F,$7E,$52,$F7,$4F,$D0,$75,$F4,$9F,$CA,$37
	.dc.b	$FE,$53,$C7,$FC,$1B,$5F,$06,$B9,$9F,$54,$63,$2E,$9A,$FE,$8E,$7C
	.dc.b	$15,$E5,$8F,$ED,$11,$EA,$FD,$79,$F8,$FF,$F9,$E7,$E6,$BF,$B3,$57
	.dc.b	$E3,$32,$F0,$56,$BE,$F3,$FB,$87,$CB,$FB,$BE,$97,$96,$3E,$3C,$BA
	.dc.b	$7E,$11,$71,$E3,$3D,$6F,$FC,$47,$5C,$30,$E7,$DD,$EA,$FC,$19,$FB
	.dc.b	$03,$F7,$5F,$F4,$D7,$E7,$6F,$DD,$DF,$99,$FD,$2B,$DD,$6A,$7F,$AB
	.dc.b	$6B,$F1,$E7,$9E,$7F,$1E,$31,$C5,$E5,$F1,$FF,$AA,$F3,$7F,$AB,$34
	.dc.b	$F0,$7E,$14,$D7,$FD,$A1,$8F,$79,$CC,$3F,$6D,$FA,$3F,$B1,$3E,$65
	.dc.b	$FA,$2B,$F1,$F7,$EC,$3F,$9C,$FC,$CB,$F1,$37,$BA,$D4,$FE,$83,$F2
	.dc.b	$5C,$E1,$E5,$D3,$FC,$05,$B3,$0F,$EF,$51,$FE,$7C,$DF,$D3,$1C,$F3
	.dc.b	$EF,$E5,$F9,$30,$F1,$C7,$8B,$E4,$BF,$41,$F8,$77,$F5,$BF,$9F,$7E
	.dc.b	$92,$FC,$0B,$FB,$A7,$EF,$FF,$E3,$AE,$8F,$EA,$B7,$BA,$D7,$D0,$29
	.dc.b	$F9,$B7,$F0,$FA,$E9,$C1,$F3,$F8,$3F,$C0,$31,$FD,$A7,$E6,$7D,$93
	.dc.b	$FE,$C6,$F9,$AF,$F0,$EE,$5D,$3C,$1F,$2C,$63,$5F,$F0,$37,$B7,$E8
	.dc.b	$FD,$9F,$FC,$A3,$FB,$DB,$F6,$67,$E5,$EF,$CC,$5F,$B6,$78,$EF,$75
	.dc.b	$A9,$FF,$50,$72,$7B,$25,$F8,$67,$93,$85,$1F,$E5,$EE,$0F,$51,$FE
	.dc.b	$33,$8C,$FD,$3B,$4F,$16,$9F,$DA,$6B,$C1,$39,$CF,$CB,$F8,$47,$D0
	.dc.b	$FD,$E7,$F0,$BE,$FB,$F4,$77,$E9,$BF,$E6,$FF,$D0,$3E,$55,$6A,$73
	.dc.b	$F5,$0F,$B6,$CF,$E0,$FE,$7A,$FC,$7E,$BE,$63,$FD,$E2,$F1,$C5,$D7
	.dc.b	$E1,$F2,$7F,$D9,$19,$78,$3E,$03,$F0,$7F,$93,$3F,$3C,$FE,$67,$FA
	.dc.b	$83,$F6,$FF,$37,$EF,$1F,$76,$B5,$F1,$AA,$D6,$2F,$C1,$B5,$C9,$F2
	.dc.b	$EF,$A7,$07,$97,$A9,$1C,$B0,$ED,$F9,$77,$8B,$5F,$0E,$BF,$E9,$6E
	.dc.b	$EF,$C8,$FE,$EE,$FD,$93,$F9,$67,$83,$FE,$4F,$FD,$D1,$F9,$DF,$E4
	.dc.b	$6B,$5F,$19,$F8,$3F,$BC,$57,$EC,$F1,$E7,$FA,$BF,$DF,$BE,$9C,$3E
	.dc.b	$57,$8E,$5F,$E5,$BF,$87,$FD,$3C,$63,$1F,$E5,$98,$7F,$8E,$3F,$19
	.dc.b	$7E,$45,$FE,$AD,$FD,$8D,$FA,$93,$F2,$A7,$E6,$CF,$D0,$3F,$BA,$61
	.dc.b	$C3,$F1,$46,$1C,$2F,$7F,$9B,$6B,$2F,$F2,$C7,$07,$EF,$CF,$F4,$26
	.dc.b	$1C,$5F,$C0,$3C,$EF,$97,$AB,$FD,$76,$31,$A7,$8F,$D3,$C7,$E0,$8E
	.dc.b	$27,$EA,$9F,$2B,$27,$DE,$FE,$77,$8B,$DD,$7B,$AD,$4F,$EC,$67,$37
	.dc.b	$FB,$CC,$BF,$D0,$7D,$97,$0E,$1D,$5F,$E9,$BF,$D5,$7C,$37,$E0,$8F
	.dc.b	$F4,$F7,$C4,$E3,$3E,$BF,$95,$7C,$6F,$8A,$FE,$5D,$EE,$BF,$76,$FF
	.dc.b	$46,$7E,$CF,$FE,$D4,$DF,$E9,$56,$A7,$3F,$8C,$F1,$CB,$C9,$2F,$D1
	.dc.b	$EC,$7F,$20,$AF,$F4,$B7,$E2,$9F,$1E,$B9,$7C,$1F,$FB,$8F,$1F,$BD
	.dc.b	$E5,$FF,$3A,$F8,$CF,$D2,$7E,$9A,$FC,$57,$FD,$D9,$5A,$F9,$1F,$E6
	.dc.b	$DE,$62,$B5,$F9,$AF,$E0,$31,$7E,$ED,$F1,$DF,$E8,$BD,$BF,$69,$ED
	.dc.b	$7C,$72,$FB,$51,$8C,$3F,$D3,$59,$7E,$47,$A1,$FA,$CF,$FB,$7B,$F2
	.dc.b	$CE,$BF,$AF,$BF,$4F,$FC,$7D,$6B,$F1,$B5,$6B,$F5,$0C,$F9,$DF,$66
	.dc.b	$B9,$EF,$F8,$C7,$85,$7A,$3F,$36,$FC,$64,$63,$87,$FB,$DE,$BF,$E9
	.dc.b	$AF,$1B,$FD,$D9,$FB,$53,$F0,$6F,$F5,$47,$C6,$3C,$8F,$B5,$7B,$FC
	.dc.b	$7D,$7B,$F9,$3F,$E7,$74,$FD,$EB,$FC,$B5,$97,$FB,$DF,$DC,$FE,$92
	.dc.b	$F6,$E0,$FE,$40,$F8,$34,$EE,$FF,$8F,$4C,$61,$FC,$CA,$79,$E5,$E0
	.dc.b	$AF,$2E,$2F,$D9,$BB,$FE,$D6,$FC,$B1,$9B,$F7,$C7,$C8,$BF,$78,$FF
	.dc.b	$30,$F0,$FC,$1C,$38,$5E,$EB,$59,$74,$C7,$D3,$4F,$CE,$6C,$76,$CB
	.dc.b	$AF,$E8,$9E,$9A,$E7,$1E,$0F,$2B,$D9,$F9,$5B,$F7,$AF,$F5,$97,$E0
	.dc.b	$7F,$EA,$BF,$C3,$7E,$86,$FF,$4E,$B5,$F8,$97,$F9,$EF,$F0,$EF,$F4
	.dc.b	$B7,$E8,$1F,$32,$5F,$95,$7C,$C7,$A7,$47,$CC,$78,$E5,$FE,$B7,$E8
	.dc.b	$F0,$E9,$E0,$C7,$FD,$B5,$E7,$FF,$BF,$BA,$FF,$8D,$FC,$EB,$F2,$DF
	.dc.b	$F3,$37,$DD,$FF,$26,$FC,$27,$ED,$FC,$27,$F7,$99,$CF,$C2,$FF,$9A
	.dc.b	$F8,$FF,$15,$3F,$DF,$78,$3E,$EB,$FD,$BD,$B8,$F5,$FC,$09,$97,$EC
	.dc.b	$E7,$AC,$E6,$1B,$FE,$0F,$CE,$7E,$27,$FD,$23,$F3,$1F,$F2,$67,$97
	.dc.b	$FE,$FD,$D3,$F3,$55,$AF,$82,$56,$B3,$FC,$9C,$BF,$38,$FC,$9B,$D2
	.dc.b	$3F,$26,$F1,$9F,$F2,$C7,$DC,$7E,$4F,$F5,$67,$EC,$6F,$C6,$9F,$A3
	.dc.b	$3F,$32,$7E,$68,$FD,$F3,$F4,$38,$70,$BD,$D6,$BA,$B3,$8F,$E9,$2B
	.dc.b	$FC,$47,$AF,$E9,$1E,$19,$75,$F3,$3C,$55,$E1,$8F,$07,$C6,$7F,$A8
	.dc.b	$F3,$FF,$14,$FE,$2E,$FC,$AD,$E2,$BF,$5A,$7E,$F7,$FE,$A0,$E7,$78
	.dc.b	$4D,$E6,$5C,$B1,$E3,$87,$03,$FD,$A9,$1D,$A7,$FC,$30,$FF,$8E,$3F
	.dc.b	$1B,$FF,$22,$BC,$72,$F0,$7D,$5F,$97,$ED,$9A,$F9,$67,$C9,$F5,$7E
	.dc.b	$9F,$AD,$7F,$00,$FF,$99,$FE,$D3,$F0,$3F,$8A,$BE,$8F,$C3,$5A,$F9
	.dc.b	$05,$6B,$F0,$DC,$FE,$F3,$E2,$AE,$7C,$3E,$67,$85,$7E,$33,$D9,$F0
	.dc.b	$67,$DF,$C8,$7C,$57,$CB,$FE,$00,$FF,$62,$FE,$69,$F1,$BF,$21,$7C
	.dc.b	$BF,$F9,$07,$B3,$E4,$EB,$F2,$2F,$7D,$8B,$DF,$EC,$3F,$22,$7F,$D9
	.dc.b	$F5,$F2,$17,$E2,$FF,$21,$E3,$5F,$E2,$9B,$FA,$95,$F8,$B9,$CF,$87
	.dc.b	$F7,$FF,$50,$FF,$57,$7D,$F3,$F3,$CF,$B7,$F8,$2B,$F0,$BF,$AF,$5A
	.dc.b	$9C,$FC,$CE,$3F,$14,$FA,$F3,$FE,$B7,$CF,$ED,$BB,$D1,$F8,$67,$7F
	.dc.b	$C4,$FF,$CD,$18,$FE,$23,$F2,$DF,$C8,$7E,$B5,$FC,$BB,$9B,$F9,$7F
	.dc.b	$13,$F1,$47,$D2,$7E,$22,$B5,$C7,$F1,$67,$5C,$FF,$2B,$7F,$D4,$57
	.dc.b	$F3,$87,$F1,$AB,$F5,$8F,$C6,$BC,$B1,$FA,$F3,$D7,$06,$5E,$2E,$DF
	.dc.b	$D2,$7C,$2F,$F5,$BF,$B6,$FC,$F7,$E1,$7F,$E0,$BF,$CD,$FF,$37,$39
	.dc.b	$C7,$FD,$55,$F5,$4F,$DF,$33,$E6,$7E,$72,$BF,$B2,$57,$E8,$1E,$18
	.dc.b	$7F,$B2,$E3,$9F,$CD,$7F,$92,$1F,$F5,$A6,$7E,$28,$C4,$FF,$86,$F8
	.dc.b	$47,$C1,$FC,$A5,$FA,$DB,$F2,$CF,$EF,$FF,$AE,$3F,$38,$FF,$29,$7B
	.dc.b	$BC,$3B,$BF,$0B,$87,$0B,$DD,$7F,$95,$FF,$5C,$D6,$DF,$BD,$AD,$3F
	.dc.b	$E8,$F7,$BF,$BB,$FE,$BC,$79,$7D,$3B,$3F,$07,$77,$F0,$5D,$BC,$DB
	.dc.b	$79,$6B,$59,$FE,$1B,$2E,$E3,$1E,$78,$FA,$7E,$FF,$FD,$7F,$8F,$BF
	.dc.b	$F5,$F7,$EE,$7F,$ED,$AE,$07,$E6,$5F,$E5,$CF,$C2,$B9,$E4,$BD,$FD
	.dc.b	$CF,$F1,$8A,$FF,$30,$FE,$B5,$A9,$E7,$AF,$B5,$7E,$99,$ED,$78,$E5
	.dc.b	$DB,$F5,$EF,$63,$F2,$A3,$38,$3F,$4D,$CE,$1F,$27,$57,$B6,$FD,$95
	.dc.b	$F4,$0F,$C1,$7F,$D8,$7F,$84,$FA,$2F,$75,$A9,$FF,$63,$FE,$C5,$9F
	.dc.b	$CC,$FE,$85,$74,$CB,$EA,$3C,$31,$E9,$F8,$2F,$8A,$7E,$64,$E7,$0F
	.dc.b	$F2,$1C,$7C,$6F,$E0,$07,$EE,$0F,$D8,$7F,$63,$F8,$97,$FB,$FF,$F6
	.dc.b	$3D,$4E,$7D,$5F,$F5,$A7,$E4,$6F,$AB,$CE,$5F,$85,$C9,$FA,$B7,$E1
	.dc.b	$5E,$B9,$F4,$D3,$C9,$DA,$9C,$E5,$FE,$BF,$F9,$AC,$FE,$86,$F9,$67
	.dc.b	$ED,$AF,$D2,$9F,$62,$FD,$55,$F9,$FF,$E0,$30,$E1,$7B,$AD,$7D,$06
	.dc.b	$7C,$2F,$CC,$57,$5E,$2F,$CE,$3C,$32,$EB,$F3,$1E,$2D,$3C,$31,$E0
	.dc.b	$F9,$BF,$4F,$CB,$BC,$9F,$F3,$EF,$ED,$6F,$D6,$1F,$AD,$F9,$BC,$EF
	.dc.b	$DB,$5E,$FF,$05,$5E,$EB,$DB,$5A,$FD,$13,$F5,$97,$D7,$7F,$D8,$3C
	.dc.b	$6F,$B7,$C2,$78,$B3,$F0,$E6,$F3,$B1,$A7,$68,$F1,$FE,$3B,$FA,$41
	.dc.b	$FB,$0B,$F7,$BF,$F4,$E7,$E2,$CF,$31,$F1,$FB,$AD,$7E,$22,$AD,$7C
	.dc.b	$9F,$3F,$CD,$BF,$81,$65,$D3,$8B,$F8,$33,$DB,$3E,$BF,$A6,$7E,$02
	.dc.b	$31,$8F,$3F,$C1,$FF,$2D,$FA,$BB,$F3,$BF,$2F,$E7,$EF,$A0,$4B,$F4
	.dc.b	$E7,$ED,$BF,$C2,$23,$87,$0A,$F8,$46,$FE,$13,$58,$F9,$77,$FF,$76
	.dc.b	$1D,$17,$07,$F6,$88,$FE,$D3,$DB,$FD,$36,$FA,$E1,$D3,$C2,$F3,$E1
	.dc.b	$E6,$D7,$CB,$F4,$BE,$AF,$C9,$5F,$2A,$F8,$F7,$E6,$3F,$EF,$1D,$1E
	.dc.b	$D7,$A6,$F7,$5A,$FC,$57,$F9,$EF,$E4,$F9,$C7,$C9,$C1,$D7,$1A,$E6
	.dc.b	$F8,$7C,$27,$86,$BC,$19,$FF,$BF,$FC,$2F,$9B,$F8,$6F,$F4,$6F,$9F
	.dc.b	$7E,$61,$FC,$97,$F1,$F7,$E7,$DF,$2E,$B5,$39,$FC,$7F,$1D,$BE,$0D
	.dc.b	$FE,$25,$7F,$50,$DF,$CB,$18,$F9,$3F,$EF,$58,$7F,$31,$FD,$0B,$C3
	.dc.b	$18,$F8,$E7,$27,$E0,$3F,$2F,$7E,$66,$F9,$3F,$E4,$BF,$C0,$0F,$DE
	.dc.b	$7F,$D6,$5A,$7D,$3B,$DD,$6A,$5F,$B6,$7E,$0E,$5F,$A3,$F9,$FD,$DF
	.dc.b	$F1,$1F,$88,$9E,$FC,$3E,$89,$CC,$74,$F9,$06,$9F,$62,$BA,$6B,$E5
	.dc.b	$9C,$ED,$F9,$8E,$87,$EB,$CF,$69,$FA,$6B,$F6,$C7,$EF,$BF,$E0,$5E
	.dc.b	$4F,$C6,$61,$C2,$F7,$5A,$F0,$3F,$39,$3D,$35,$F3,$1F,$E6,$FE,$1F
	.dc.b	$F1,$37,$93,$ED,$8C,$4F,$3C,$FC,$1F,$71,$E5,$D7,$C9,$1E,$3F,$D2
	.dc.b	$FF,$5F,$FB,$AF,$F8,$2B,$F2,$37,$93,$FE,$DE,$FC,$59,$F5,$5F,$AD
	.dc.b	$AD,$4E,$7F,$5B,$C7,$1F,$87,$6F,$DA,$D7,$F6,$17,$CE,$62,$3F,$DD
	.dc.b	$D5,$FE,$EA,$F8,$6F,$0C,$63,$D7,$FD,$7F,$67,$D6,$F9,$3F,$C1,$7E
	.dc.b	$7E,$FD,$21,$FB,$67,$FA,$A3,$F6,$9F,$E0,$15,$A9,$CF,$E1,$EF,$C1
	.dc.b	$B3,$E3,$F9,$7A,$FF,$B7,$7D,$45,$C3,$86,$4F,$D3,$BF,$88,$C6,$0F
	.dc.b	$E2,$71,$FC,$3F,$EE,$5F,$9A,$BE,$0D,$FA,$93,$F8,$07,$F8,$95,$EE
	.dc.b	$B5,$3F,$E1,$6E,$4F,$E1,$53,$FC,$67,$9F,$E2,$EB,$EC,$FC,$9B,$CD
	.dc.b	$3C,$72,$FF,$02,$57,$4F,$57,$FD,$09,$97,$87,$6F,$E4,$FF,$30,$FA
	.dc.b	$47,$DA,$3D,$8F,$F4,$9F,$ED,$DF,$E4,$3D,$FF,$8D,$5E,$EB,$5F,$85
	.dc.b	$29,$F6,$DF,$CA,$EB,$FC,$A5,$AF,$F2,$6E,$14,$77,$8F,$4B,$CF,$2F
	.dc.b	$F5,$0F,$B5,$E4,$D3,$C7,$1F,$EA,$4F,$C6,$7F,$C2,$5F,$99,$BF,$60
	.dc.b	$7E,$6F,$FE,$E4,$F9,$3F,$EF,$BF,$42,$FD,$37,$FC,$87,$7B,$AD,$7D
	.dc.b	$22,$BF,$C0,$5C,$9F,$E8,$1A,$E9,$FA,$4F,$E4,$DE,$1C,$1F,$92,$79
	.dc.b	$7D,$F7,$F5,$FC,$1F,$7B,$E4,$D7,$C7,$39,$F9,$9C,$69,$E0,$D3,$FB
	.dc.b	$BE,$9D,$FF,$4F,$FF,$36,$7A,$1F,$EE,$4F,$C5,$7F,$21,$F7,$DE,$89
	.dc.b	$CF,$E3,$9F,$ED,$1F,$29,$FB,$D6,$7E,$77,$CD,$5F,$E7,$5E,$3F,$33
	.dc.b	$C3,$17,$DC,$FC,$88,$C7,$E0,$37,$E7,$9A,$F8,$73,$7E,$F9,$F2,$3D
	.dc.b	$37,$E8,$3F,$ED,$0F,$5D,$FB,$13,$F7,$CF,$F1,$F7,$27,$C9,$BD,$FC
	.dc.b	$1A,$F7,$4F,$9E,$E9,$8F,$CD,$BF,$E6,$DC,$5F,$35,$E3,$7D,$BB,$3E
	.dc.b	$FC,$63,$3F,$F3,$F6,$7F,$EB,$7F,$4B,$FD,$2F,$EC,$BF,$5B,$FE,$83
	.dc.b	$FA,$47,$E6,$7F,$34,$FF,$41,$D6,$A7,$3E,$43,$EC,$73,$F3,$1F,$EC
	.dc.b	$6D,$FF,$87,$DF,$46,$C7,$F0,$9B,$EB,$D1,$FE,$63,$AE,$9F,$24,$CF
	.dc.b	$F8,$C6,$BF,$EC,$BF,$D6,$1F,$1A,$FD,$03,$FA,$A7,$E5,$AF,$D1,$9F
	.dc.b	$92,$7E,$2B,$0E,$17,$BF,$84,$5E,$E9,$F8,$D7,$FA,$7B,$2E,$BF,$6D
	.dc.b	$FC,$A5,$EF,$9D,$C7,$96,$1F,$E4,$1E,$2E,$11,$8D,$1F,$B4,$FF,$9F
	.dc.b	$7E,$23,$FA,$8F,$D8,$7F,$58,$FA,$C7,$E5,$BF,$EF,$2F,$D9,$FF,$D5
	.dc.b	$B5,$AF,$9F,$FF,$16,$4F,$FB,$DB,$E0,$F3,$F3,$0C,$7F,$9E,$5F,$F5
	.dc.b	$CF,$3A,$CE,$70,$E3,$F8,$EF,$86,$31,$9F,$83,$C4,$FF,$48,$7C,$33
	.dc.b	$F6,$9F,$F1,$6F,$E9,$3F,$CE,$FF,$21,$7C,$13,$F2,$AF,$F6,$0D,$EF
	.dc.b	$ED,$97,$BA,$7E,$D5,$AC,$78,$F4,$7F,$4E,$C3,$83,$63,$BF,$94,$CF
	.dc.b	$7E,$0B,$E5,$D7,$30,$F0,$67,$E5,$F8,$8F,$4F,$E0,$CF,$D7,$3F,$DA
	.dc.b	$9F,$24,$F9,$07,$E6,$3A,$9C,$C6,$3F,$44,$F5,$DF,$F4,$5C,$FC,$CF
	.dc.b	$36,$D7,$0F,$B6,$E5,$F8,$6F,$F4,$DC,$3F,$D6,$BB,$F9,$EF,$CD,$83
	.dc.b	$F7,$CF,$E6,$7F,$89,$BF,$78,$F8,$BF,$1F,$E0,$9C,$C7,$C8,$F9,$5F
	.dc.b	$1F,$EA,$DF,$DD,$F9,$3F,$D5,$15,$71,$E9,$F2,$CC,$BA,$7E,$9B,$F0
	.dc.b	$AB,$9F,$27,$DC,$F0,$7F,$DE,$3F,$99,$BF,$57,$7E,$CE,$FA,$EF,$F6
	.dc.b	$EF,$1F,$A7,$8F,$FB,$76,$1C,$2F,$75,$FE,$52,$AD,$6F,$FE,$2A,$FC
	.dc.b	$83,$CD,$7F,$EF,$4C,$3F,$CE,$F3,$FE,$2E,$FC,$6D,$9F,$87,$3F,$A9
	.dc.b	$5F,$8A,$BE,$D1,$FE,$1A,$8C,$71,$7E,$33,$F5,$1F,$EC,$FF,$77,$FE
	.dc.b	$D9,$D3,$C3,$82,$B5,$E7,$FF,$94,$7F,$4B,$E9,$FA,$BF,$92,$67,$84
	.dc.b	$CB,$FB,$8F,$E0,$9F,$16,$78,$FD,$D7,$E8,$BD,$87,$0F,$E4,$FE,$8B
	.dc.b	$F9,$CE,$DF,$DB,$0F,$83,$3A,$F2,$3F,$D6,$9F,$26,$9F,$F6,$E9,$AE
	.dc.b	$6F,$B5,$5F,$17,$8F,$FC,$52,$7E,$6D,$E5,$87,$F4,$6F,$13,$C5,$97
	.dc.b	$87,$1F,$F6,$35,$FE,$C2,$FF,$19,$F7,$DF,$81,$FD,$83,$FD,$F9,$F7
	.dc.b	$79,$EE,$FF,$36,$F8,$09,$FF,$80,$1F,$AD,$71,$E2,$F1,$D4,$C3,$97
	.dc.b	$E9,$5F,$06,$38,$DF,$D3,$E1,$FB,$FF,$BA,$7E,$B7,$E3,$1E,$7B,$EC
	.dc.b	$7E,$67,$EB,$FF,$C1,$2B,$53,$9E,$A7,$EB,$79,$FB,$3F,$CE,$57,$EB
	.dc.b	$79,$3F,$0E,$F0,$C7,$FC,$3B,$1C,$FD,$1F,$EA,$18,$F7,$DF,$F8,$BD
	.dc.b	$FF,$08,$F9,$B7,$E7,$CF,$D0,$FC,$5E,$E7,$F1,$4F,$65,$5A,$FD,$17
	.dc.b	$CD,$EE,$F8,$AA,$78,$BF,$E5,$9A,$F8,$1E,$DF,$F1,$0D,$FF,$CA,$1A
	.dc.b	$77,$F6,$3F,$93,$FE,$3B,$FE,$4B,$F8,$37,$EB,$1F,$B3,$FC,$33,$FC
	.dc.b	$C5,$D7,$FA,$6A,$D4,$E7,$8B,$F9,$4C,$7F,$0F,$CB,$F4,$EF,$E1,$5E
	.dc.b	$BC,$FF,$E9,$CA,$FD,$F7,$E0,$18,$CB,$BF,$84,$E6,$F1,$AF,$C8,$DF
	.dc.b	$9E,$3F,$4C,$FE,$45,$F8,$4F,$82,$B5,$39,$E1,$7E,$B9,$E4,$FC,$AB
	.dc.b	$F8,$87,$07,$F0,$DE,$0F,$E7,$0F,$F0,$FC,$FA,$E7,$FB,$AE,$2F,$76
	.dc.b	$7D,$23,$FC,$89,$7E,$EF,$DF,$3F,$B6,$FE,$D8,$FD,$DF,$FC,$09,$E2
	.dc.b	$67,$E5,$D3,$9F,$A3,$FE,$2B,$9F,$C9,$FD,$07,$C7,$8C,$BD,$E3,$FE
	.dc.b	$83,$CB,$96,$3D,$7F,$05,$FA,$BF,$17,$FE,$11,$FA,$07,$EC,$BF,$E0
	.dc.b	$3F,$DA,$3E,$8C,$E6,$3F,$C0,$DF,$BD,$FD,$4F,$8D,$F8,$33,$38,$3F
	.dc.b	$DB,$D5,$FB,$2E,$0F,$B9,$3C,$7D,$AF,$C1,$FA,$7F,$82,$FC,$F1,$FB
	.dc.b	$F7,$FA,$E6,$FF,$05,$39,$8F,$F6,$77,$A1,$F8,$28,$C7,$9B,$CB,$C4
	.dc.b	$61,$CB,$CD,$FC,$7F,$C7,$7F,$8A,$FD,$3F,$5F,$EF,$CF,$CC,$7F,$17
	.dc.b	$39,$E0,$4B,$F5,$BF,$C7,$FE,$65,$F7,$2A,$71,$FE,$D2,$FF,$5F,$C3
	.dc.b	$F9,$79,$E5,$FA,$1F,$F8,$13,$F0,$87,$EC,$3F,$E2,$6F,$B9,$FB,$5F
	.dc.b	$91,$7F,$BA,$30,$FE,$87,$5A,$F0,$7F,$CB,$1E,$0A,$BF,$D7,$15,$AF
	.dc.b	$C6,$3F,$AE,$5F,$4F,$0B,$FC,$08,$F1,$C7,$F0,$DE,$E7,$87,$3F,$06
	.dc.b	$4F,$8B,$F2,$7E,$61,$F9,$FF,$D5,$BF,$3D,$FB,$63,$F3,$BF,$F0,$AE
	.dc.b	$6A,$D7,$98,$A7,$C6,$EB,$E2,$BC,$75,$3D,$06,$5F,$A9,$F9,$BE,$F1
	.dc.b	$8C,$7F,$CD,$DC,$3D,$BD,$BF,$DB,$7E,$8B,$FE,$55,$FD,$F3,$1E,$E7
	.dc.b	$F0,$48,$FF,$2C,$7E,$C7,$BB,$F7,$FC,$8F,$C2,$C3,$8F,$CF,$FC,$F7
	.dc.b	$9F,$DA,$7F,$8D,$3E,$2B,$FC,$15,$F8,$43,$F2,$87,$EC,$6F,$D3,$2F
	.dc.b	$81,$FD,$D2,$F7,$5A,$9F,$F6,$84,$E7,$D0,$61,$CF,$0F,$F0,$CF,$C5
	.dc.b	$7F,$1F,$FA,$2F,$E5,$55,$AC,$7B,$FE,$17,$FE,$46,$FC,$57,$FC,$39
	.dc.b	$D3,$FC,$83,$F1,$A7,$E5,$0E,$6F,$F2,$2D,$6B,$D5,$56,$BE,$13,$39
	.dc.b	$FC,$E9,$CF,$99,$D6,$31,$5F,$CE,$3A,$3F,$C9,$FB,$F8,$3F,$57,$FF
	.dc.b	$75,$7E,$7E,$FC,$AF,$38,$7D,$89,$CF,$E0,$78,$E0,$47,$FB,$C7,$F0
	.dc.b	$DF,$F2,$FE,$3C,$FF,$1B,$FF,$44,$E7,$FD,$E6,$3F,$A0,$7E,$50,$FD
	.dc.b	$A7,$DF,$F7,$1E,$AF,$6D,$F9,$0E,$F0,$FF,$2A,$DE,$F6,$FC,$4D,$EF
	.dc.b	$F6,$6D,$52,$B5,$B7,$F9,$AF,$8B,$F6,$3A,$78,$34,$EF,$0E,$3F,$F3
	.dc.b	$0F,$77,$F8,$D7,$07,$E3,$1F,$4B,$F1,$8F,$87,$F5,$FD,$A7,$3B,$FE
	.dc.b	$6A,$7E,$3F,$E3,$B9,$BE,$1C,$7F,$9A,$C7,$A4,$63,$EA,$BB,$ED,$F0
	.dc.b	$BE,$F3,$B7,$CA,$FF,$0C,$FC,$87,$B4,$63,$9F,$FC,$07,$8F,$96,$37
	.dc.b	$F9,$0F,$2E,$5F,$8B,$1C,$B8,$B1,$1F,$0F,$EE,$BF,$CC,$DF,$80,$7F
	.dc.b	$BC,$3C,$47,$80,$E6,$F6,$56,$B9,$3F,$1F,$2E,$4C,$F2,$B1,$E5,$DB
	.dc.b	$DB,$9F,$FC,$D3,$1D,$79,$71,$97,$83,$F1,$D7,$EC,$8F,$A9,$7E,$D9
	.dc.b	$D3,$05,$6A,$5F,$B3,$66,$38,$FC,$87,$4D,$7C,$51,$D3,$E4,$B0,$B8
	.dc.b	$5F,$A9,$7B,$B3,$F4,$CE,$63,$E0,$FD,$8E,$2B,$7C,$3C,$9E,$3C,$1D
	.dc.b	$2F,$9E,$BE,$4E,$AF,$C3,$70,$FF,$A4,$67,$D3,$97,$B7,$8F,$BC,$78
	.dc.b	$B7,$E5,$2D,$BD,$5E,$1F,$9F,$F5,$CF,$A7,$CF,$F3,$A3,$1C,$FF,$47
	.dc.b	$4E,$FC,$7F,$8F,$BF,$C6,$78,$1F,$AA,$DB,$F2,$D1,$FA,$EA,$DE,$1E
	.dc.b	$4D,$FF,$D6,$D1,$8D,$BD,$7F,$86,$F7,$7C,$DD,$77,$7F,$33,$F8,$0F
	.dc.b	$CB,$7E,$46,$DF,$EA,$DC,$3C,$71,$8F,$A7,$D7,$2F,$CA,$BB,$87,$8F
	.dc.b	$7F,$37,$03,$0F,$BF,$D7,$F8,$3F,$C1,$3B,$6B,$F8,$1C,$EC,$FF,$B5
	.dc.b	$B0,$E3,$7C,$39,$3F,$05,$FB,$EF,$E6,$F7,$FC,$D6,$FF,$1A,$31,$CF
	.dc.b	$F9,$87,$F2,$FB,$DE,$0F,$E5,$7D,$1F,$99,$FC,$01,$EE,$FE,$4B,$83
	.dc.b	$F8,$27,$8A,$F1,$BF,$80,$56,$A7,$F2,$FE,$27,$F8,$0B,$E3,$F3,$87
	.dc.b	$18,$FE,$1F,$9F,$F1,$8B,$E1,$7F,$C3,$E7,$C1,$5F,$A0,$FC,$5F,$F8
	.dc.b	$96,$BF,$BE,$E3,$FE,$4B,$87,$79,$CE,$7D,$B1,$47,$F1,$9F,$C9,$B9
	.dc.b	$C6,$32,$FD,$B6,$7B,$F9,$4F,$E3,$DE,$B7,$58,$F2,$6B,$FB,$09,$CE
	.dc.b	$5B,$BF,$C0,$F1,$FF,$23,$D3,$E0,$57,$F2,$7F,$73,$D3,$DE,$FC,$27
	.dc.b	$F9,$57,$E8,$DD,$74,$FE,$65,$87,$CC,$AD,$64,$AF,$99,$97,$5D,$3F
	.dc.b	$C0,$78,$FF,$33,$F9,$2E,$AF,$F3,$5E,$4E,$FC,$FF,$CD,$FF,$51,$E3
	.dc.b	$CB,$F1,$B0,$FE,$65,$9A,$7C,$5E,$6B,$6F,$16,$1F,$CF,$FF,$03,$E7
	.dc.b	$7F,$E1,$FE,$0F,$B3,$C0,$FD,$07,$D7,$C9,$87,$DB,$B5,$BE,$1C,$3E
	.dc.b	$9B,$E1,$1F,$EB,$8C,$39,$FC,$6F,$CC,$AD,$3F,$4A,$F3,$5F,$3B,$83
	.dc.b	$D5,$87,$FA,$2E,$3F,$CE,$17,$D7,$F0,$2F,$C0,$B6,$1D,$7E,$ED,$FE
	.dc.b	$75,$7E,$0F,$3F,$5B,$FD,$77,$D1,$FA,$FD,$5F,$9C,$DF,$DE,$FE,$7B
	.dc.b	$5B,$97,$E6,$30,$FB,$3C,$9F,$E3,$6A,$FC,$C4,$7E,$8B,$EB,$BE,$7F
	.dc.b	$E2,$5F,$B0,$E0,$FC,$BC,$63,$C1,$D7,$0F,$F3,$3C,$BD,$75,$F9,$CF
	.dc.b	$A9,$FA,$FF,$B1,$FD,$AC,$BF,$4D,$18,$D7,$F5,$FD,$AF,$C4,$3F,$CE
	.dc.b	$FF,$B0,$3F,$2E,$7E,$DF,$A9,$CC,$3F,$13,$FD,$BF,$EA,$9D,$BE,$46
	.dc.b	$F0,$78,$F6,$7E,$D9,$FB,$CF,$D4,$3C,$7F,$9E,$DA,$D7,$E7,$FF,$C0
	.dc.b	$3F,$45,$F3,$FC,$6F,$F0,$2F,$E0,$0F,$F0,$07,$C2,$33,$7E,$C7,$9C
	.dc.b	$7F,$B7,$43,$F7,$6C,$79,$3F,$E3,$AF,$BD,$FE,$21,$8F,$2E,$7B,$87
	.dc.b	$F5,$9F,$65,$FA,$4F,$4F,$27,$F6,$57,$FE,$0F,$9C,$E5,$D1,$7D,$D7
	.dc.b	$2C,$FF,$B7,$FE,$55,$67,$CD,$B6,$F6,$F2,$65,$D3,$4F,$06,$DB,$DB
	.dc.b	$97,$83,$E2,$C3,$D2,$F8,$2F,$86,$7F,$E2,$9D,$37,$A7,$87,$E0,$38
	.dc.b	$56,$9E,$1E,$17,$F1,$CE,$0F,$F1,$44,$7A,$3C,$2F,$8A,$FF,$99,$9F
	.dc.b	$E5,$11,$AD,$FE,$15,$71,$DF,$FB,$D4,$7A,$A3,$19,$FF,$80,$37,$F1
	.dc.b	$6F,$F0,$FE,$0F,$C7,$5A,$7C,$79,$76,$C7,$FB,$96,$DB,$DB,$86,$7E
	.dc.b	$0A,$D6,$6C,$3F,$01,$C5,$F7,$16,$9D,$F4,$FE,$99,$F0,$2F,$CD,$BF
	.dc.b	$45,$CB,$CF,$2F,$E4,$B3,$9C,$BF,$CA,$53,$EB,$9F,$E9,$13,$9E,$3F
	.dc.b	$8D,$C7,$FC,$CE,$31,$B7,$6C,$7F,$5D,$FB,$AF,$5F,$53,$F9,$05,$71
	.dc.b	$8E,$98,$F8,$2B,$F6,$9F,$87,$71,$D3,$F7,$B9,$75,$9F,$DC,$7E,$19
	.dc.b	$CF,$D2,$65,$E3,$8E,$79,$F6,$AF,$DF,$7E,$19,$CF,$67,$C2,$DE,$7E
	.dc.b	$58,$FA,$DE,$39,$74,$C3,$F8,$37,$A7,$C7,$DF,$E5,$FA,$77,$09,$FE
	.dc.b	$14,$F0,$C3,$E2,$79,$DD,$FD,$DF,$57,$7B,$F3,$1E,$1C,$7F,$E4,$27
	.dc.b	$86,$3F,$E4,$1E,$5D,$FE,$7F,$C2,$37,$7F,$5F,$B7,$FB,$FE,$7F,$56
	.dc.b	$1F,$5D,$DB,$EA,$7B,$F6,$70,$3A,$E9,$FE,$50,$F0,$7C,$7D,$BF,$82
	.dc.b	$F6,$3F,$88,$7E,$A1,$D1,$FE,$25,$F9,$D5,$BF,$F4,$97,$89,$FE,$AA
	.dc.b	$EF,$FB,$D5,$FC,$5F,$CD,$E0,$EB,$CB,$FC,$7F,$EE,$0E,$59,$F8,$CA
	.dc.b	$DD,$3F,$EB,$9F,$CA,$DF,$9E,$BF,$04,$7E,$C3,$CB,$1E,$38,$70,$FA
	.dc.b	$86,$1C,$2D,$FA,$C6,$ED,$9F,$7C,$FC,$51,$0D,$47,$8F,$E3,$71,$1E
	.dc.b	$7F,$39,$B7,$0A,$E5,$9F,$83,$67,$E0,$FB,$AD,$6D,$E5,$DB,$D7,$F9
	.dc.b	$D6,$73,$F8,$50,$FD,$D7,$FD,$3B,$F8,$E3,$F7,$D7,$DB,$7F,$BA,$7F
	.dc.b	$07,$F5,$5E,$FE,$65,$7B,$AF,$F7,$77,$EB,$5A,$C5,$F4,$7E,$AC,$F1
	.dc.b	$85,$E3,$F5,$C7,$8A,$BC,$3A,$78,$3F,$1C,$F3,$4F,$CC,$9C,$E4,$F5
	.dc.b	$33,$F7,$2F,$C8,$7F,$D3,$5F,$A4,$BD,$C7,$E4,$AF,$D7,$DF,$55,$F4
	.dc.b	$F0,$E1,$E3,$30,$E1,$7B,$FA,$BD,$65,$FE,$A5,$DB,$E9,$F0,$FA,$16
	.dc.b	$3E,$1F,$4B,$D9,$D7,$F7,$23,$AE,$5D,$2F,$CF,$87,$9B,$6F,$2F,$67
	.dc.b	$BB,$F7,$07,$D4,$3C,$97,$EB,$C9,$7E,$06,$FC,$25,$FB,$DF,$E3,$67
	.dc.b	$8D,$F0,$77,$2F,$DF,$75,$8F,$F4,$9E,$2F,$C8,$6D,$EB,$BE,$31,$EE
	.dc.b	$E1,$FC,$AC,$74,$D3,$C5,$A7,$EF,$2B,$9E,$7D,$E3,$EF,$C7,$E7,$FF
	.dc.b	$0B,$7D,$13,$F6,$07,$C8,$BF,$1D,$7E,$D2,$FC,$95,$F2,$3C,$70,$E1
	.dc.b	$7B,$AD,$7C,$06,$7C,$EF,$9F,$A7,$EE,$F4,$EF,$A6,$21,$18,$C3,$97
	.dc.b	$E7,$7F,$8C,$8C,$6B,$E2,$FA,$9F,$C7,$FD,$C7,$F0,$3F,$49,$F8,$B3
	.dc.b	$F1,$47,$DD,$BF,$5D,$FF,$5A,$75,$7E,$0F,$0E,$17,$BA,$D7,$D6,$FF
	.dc.b	$01,$9E,$9B,$FC,$D3,$F8,$93,$FC,$47,$C3,$F4,$3C,$B2,$EF,$F8,$A7
	.dc.b	$93,$D8,$6B,$E4,$9F,$1F,$E7,$5F,$A6,$FC,$B3,$CB,$9F,$F9,$07,$EF
	.dc.b	$3F,$B5,$7D,$DB,$E5,$3F,$89,$56,$BE,$A5,$5A,$FD,$03,$3C,$3F,$E6
	.dc.b	$6E,$1F,$30,$BF,$83,$FB,$D7,$CD,$79,$5F,$F7,$7C,$7C,$57,$FE,$6E
	.dc.b	$C7,$AF,$9E,$F9,$7F,$E3,$5F,$2B,$FC,$83,$F7,$07,$D9,$7F,$30,$FD
	.dc.b	$F8,$3F,$5F,$7E,$3A,$FD,$E3,$FC,$0F,$87,$1B,$E1,$7B,$A7,$2E,$B8
	.dc.b	$BF,$9F,$65,$DB,$E2,$FF,$88,$63,$C7,$E2,$ED,$6B,$2E,$7A,$7E,$6B
	.dc.b	$BF,$FB,$24,$E7,$0F,$2E,$9E,$5F,$98,$F5,$CF,$F4,$BC,$5F,$9D,$BF
	.dc.b	$6A,$FC,$F7,$CF,$3F,$56,$7E,$64,$FC,$A3,$CD,$8B,$1E,$38,$70,$BD
	.dc.b	$FB,$FA,$9E,$BF,$13,$E8,$BE,$19,$F7,$7F,$8A,$F1,$7F,$4C,$AE,$7E
	.dc.b	$5B,$5F,$17,$E1,$7E,$8B,$E3,$A7,$7F,$C2,$31,$F1,$4F,$87,$FF,$61
	.dc.b	$79,$3F,$4F,$F1,$6F,$F1,$BF,$EB,$1E,$8A,$D7,$D3,$FF,$69,$EA,$57
	.dc.b	$F8,$F3,$DE,$D6,$3F,$EC,$BD,$3C,$A7,$F7,$1C,$3B,$7E,$15,$EB,$79
	.dc.b	$E7,$DF,$B3,$C7,$7E,$2C,$BA,$FC,$47,$A7,$F4,$2F,$E1,$5F,$A9,$3E
	.dc.b	$7F,$FC,$F7,$E1,$7F,$A5,$BF,$20,$F9,$B5,$A9,$CF,$DD,$3F,$3C,$CF
	.dc.b	$4F,$F3,$AE,$1F,$28,$BE,$AE,$1E,$5F,$4E,$C9,$F4,$FE,$3B,$69,$E3
	.dc.b	$AD,$7E,$3D,$F3,$F7,$F8,$BD,$0F,$D5,$DF,$94,$BF,$25,$7E,$55,$EB
	.dc.b	$E6,$FC,$8F,$07,$D1,$C3,$85,$EE,$9F,$67,$F8,$AC,$BA,$DF,$A9,$FA
	.dc.b	$7D,$7F,$87,$7C,$39,$F8,$33,$7C,$2F,$97,$B5,$A7,$6F,$C9,$BF,$C2
	.dc.b	$BB,$78,$BD,$07,$EE,$CF,$DF,$9F,$8D,$BF,$47,$FE,$75,$F9,$DF,$F3
	.dc.b	$AC,$38,$5E,$EB,$5F,$26,$AF,$F3,$2D,$6B,$0F,$CE,$63,$DB,$E3,$6E
	.dc.b	$3C,$72,$FE,$8D,$E4,$7E,$66,$BC,$77,$E2,$9F,$E9,$5E,$57,$E7,$7F
	.dc.b	$46,$ED,$F8,$D7,$F5,$0F,$0F,$FC,$4B,$EB,$FF,$3A,$D7,$E6,$D6,$B5
	.dc.b	$FF,$6E,$FC,$F7,$F1,$2F,$E0,$0F,$DA,$CF,$9F,$F4,$0F,$D8,$3D,$3B
	.dc.b	$3F,$0E,$F1,$CB,$FC,$63,$E0,$78,$78,$DA,$F8,$72,$FC,$3E,$1F,$EA
	.dc.b	$0F,$C6,$7F,$C8,$3F,$B8,$7F,$94,$FF,$25,$7E,$9F,$FD,$8F,$E3,$9F
	.dc.b	$92,$3F,$3A,$E1,$8F,$F9,$67,$0E,$17,$BA,$7E,$2F,$AA,$57,$F9,$7F
	.dc.b	$F0,$3F,$F1,$86,$7E,$09,$67,$87,$A1,$8C,$7C,$55,$FE,$51,$AE,$15
	.dc.b	$E6,$ED,$FE,$C9,$EB,$E7,$0F,$F3,$67,$DE,$BF,$39,$FE,$D1,$DF,$FE
	.dc.b	$4A,$EE,$FF,$0D,$FE,$7B,$E5,$AD,$71,$2B,$5F,$4A,$FC,$E7,$58,$BF
	.dc.b	$01,$F3,$66,$C7,$8B,$FE,$34,$DF,$FC,$D7,$1D,$31,$F0,$7C,$97,$96
	.dc.b	$B8,$61,$E4,$F7,$BF,$35,$E2,$E7,$47,$EE,$4F,$D3,$DF,$A7,$BF,$54
	.dc.b	$FF,$23,$7A,$FF,$E1,$7F,$1D,$C1,$FE,$7A,$C7,$8E,$1C,$2F,$7F,$B9
	.dc.b	$FD,$39,$CE,$DE,$73,$FD,$24,$F8,$3F,$38,$F6,$BC,$B3,$F0,$7D,$17
	.dc.b	$FB,$FF,$3E,$F7,$C3,$5F,$35,$6B,$DD,$FF,$43,$4F,$FA,$33,$EC,$DF
	.dc.b	$AA,$3F,$5D,$7E,$37,$FD,$AF,$07,$E7,$2F,$C5,$7F,$DE,$78,$F1,$C3
	.dc.b	$85,$EE,$9F,$94,$6A,$BF,$D5,$59,$76,$CB,$FC,$F9,$1C,$7F,$2A,$AE
	.dc.b	$B1,$E7,$F3,$BF,$33,$E5,$7C,$F3,$9B,$F2,$CF,$D0,$FC,$C7,$33,$FE
	.dc.b	$09,$FA,$6F,$E9,$BF,$BD,$BD,$37,$C7,$BF,$4B,$79,$BE,$5C,$38,$7C
	.dc.b	$13,$0E,$17,$BF,$D0,$75,$F4,$6C,$DF,$A4,$FA,$31,$E2,$7F,$A0,$0F
	.dc.b	$F5,$BC,$EF,$1E,$9A,$78,$BE,$1B,$CF,$39,$CF,$BF,$D9,$7F,$B3,$FF
	.dc.b	$04,$7E,$41,$FE,$C0,$F9,$77,$E1,$6F,$CF,$1F,$AE,$3F,$06,$F2,$7E
	.dc.b	$12,$FC,$9F,$FD,$35,$8F,$1C,$38,$5E,$F8,$FF,$D4,$9F,$B1,$EB,$8F
	.dc.b	$FD,$BD,$A7,$6D,$BF,$D6,$5B,$78,$A7,$38,$F8,$5F,$F6,$7C,$65,$FF
	.dc.b	$55,$47,$6D,$7C,$71,$D4,$CE,$DE,$88,$F9,$5F,$8F,$FB,$F5,$F3,$4E
	.dc.b	$74,$FF,$5C,$7E,$B3,$C7,$4F,$83,$39,$E0,$7E,$D3,$D7,$E1,$5F,$6E
	.dc.b	$FD,$D5,$F9,$A3,$F7,$0F,$F1,$57,$E2,$1F,$1F,$8F,$D1,$87,$0B,$DD
	.dc.b	$6B,$F1,$1E,$98,$FF,$AA,$78,$FD,$FC,$3F,$EF,$2C,$38,$46,$01,$FC
	.dc.b	$C6,$7C,$59,$F8,$7E,$C9,$ED,$7B,$EB,$77,$E1,$AD,$79,$D9,$9F,$F2
	.dc.b	$D7,$D6,$7F,$AE,$69,$FA,$5F,$F3,$8F,$D9,$3F,$DA,$DC,$1F,$47,$CA
	.dc.b	$FF,$25,$7E,$25,$FD,$06,$B5,$F9,$FF,$D1,$9F,$F2,$77,$ED,$FF,$73
	.dc.b	$87,$CF,$A7,$6C,$30,$FA,$AB,$9E,$7E,$0F,$91,$7E,$57,$E5,$9C,$E7
	.dc.b	$FD,$2B,$DC,$C6,$FE,$78,$F3,$7E,$8F,$FE,$EB,$FD,$B7,$BC,$3F,$43
	.dc.b	$FF,$57,$FF,$00,$BF,$26,$7E,$6B,$FD,$53,$EC,$38,$70,$7C,$CC,$38
	.dc.b	$7E,$20,$C3,$85,$FF,$5A,$EC,$DD,$FC,$6C,$BF,$D7,$6F,$E5,$B3,$F0
	.dc.b	$C3,$F7,$8F,$CE,$C6,$31,$E9,$A7,$8B,$E7,$7F,$65,$39,$CF,$BF,$C3
	.dc.b	$63,$2F,$26,$5E,$AC,$3D,$3C,$8F,$C2,$FC,$7B,$FA,$31,$F3,$CB,$F0
	.dc.b	$CC,$E1,$C7,$83,$FB,$9F,$DC,$E3,$7F,$1C,$7F,$76,$FB,$3F,$6F,$EA
	.dc.b	$2F,$C0,$1F,$7C,$F0,$7F,$C1,$3F,$97,$3F,$53,$FE,$D9,$0E,$5F,$BF
	.dc.b	$87,$1B,$E1,$F4,$EC,$38,$76,$F6,$C9,$F4,$5F,$1B,$9F,$F3,$B8,$F1
	.dc.b	$7E,$61,$89,$F1,$E7,$E2,$F9,$8F,$8D,$2A,$F0,$4F,$9B,$E6,$F3,$18
	.dc.b	$CF,$C5,$B7,$FB,$62,$BD,$17,$D7,$5F,$3E,$BE,$DF,$D2,$FE,$FF,$DC
	.dc.b	$7F,$A8,$FE,$0F,$FD,$B1,$FA,$5B,$F3,$57,$ED,$97,$3E,$79,$FF,$B1
	.dc.b	$FF,$06,$FE,$89,$C0,$C3,$85,$BF,$3C,$EB,$4F,$C7,$7E,$16,$1C,$73
	.dc.b	$BA,$78,$72,$E5,$A3,$F3,$4F,$83,$39,$D3,$C9,$F3,$D8,$CF,$C3,$1E
	.dc.b	$0F,$7F,$BE,$5E,$AC,$7D,$3F,$88,$7E,$73,$F1,$9F,$EA,$1F,$BA,$FF
	.dc.b	$9A,$FF,$38,$7E,$58,$FD,$DD,$E8,$EF,$E1,$D7,$BA,$D7,$E1,$79,$BF
	.dc.b	$5E,$3D,$39,$3E,$D8,$7B,$FD,$93,$3E,$FE,$1F,$DC,$9C,$C7,$8B,$7C
	.dc.b	$69,$E0,$BE,$19,$F9,$7D,$8F,$26,$BE,$3D,$3F,$CA,$3D,$7F,$E5,$AD
	.dc.b	$7F,$94,$FE,$D3,$FE,$88,$FD,$9B,$FD,$61,$FB,$EB,$E4,$5F,$AB,$BF
	.dc.b	$48,$FD,$07,$DD,$7F,$1A,$C3,$87,$E2,$2C,$38,$5B,$DF,$DE,$7D,$BF
	.dc.b	$07,$69,$FD,$7B,$2E,$50,$FD,$AB,$15,$BA,$F1,$EA,$ED,$FE,$ED,$87
	.dc.b	$09,$CF,$B9,$8B,$E1,$AF,$8F,$0F,$4F,$9D,$E2,$97,$E6,$AF,$CC,$9F
	.dc.b	$9A,$3F,$35,$7E,$6A,$FB,$97,$EA,$8F,$CD,$F9,$63,$C7,$0E,$17,$BA
	.dc.b	$7E,$3D,$AC,$BF,$0F,$9F,$F7,$8E,$1F,$E8,$38,$F1,$8C,$1F,$D6,$9E
	.dc.b	$9A,$78,$B4,$7D,$77,$B6,$BD,$38,$7A,$2B,$5F,$37,$9C,$38,$E9,$E3
	.dc.b	$FA,$0F,$81,$F6,$8F,$D9,$DF,$1E,$FC,$87,$FC,$67,$F0,$5F,$E1,$BF
	.dc.b	$D5,$F7,$5A,$9C,$FD,$23,$F5,$8C,$F0,$FD,$6E,$1F,$F2,$A5,$F4,$7D
	.dc.b	$66,$32,$ED,$F6,$1F,$EE,$2F,$CB,$7D,$5A,$F8,$E7,$3A,$BE,$9F,$38
	.dc.b	$F8,$70,$FF,$74,$E9,$FE,$7E,$F2,$3F,$31,$9F,$5D,$BC,$BF,$3B,$FC
	.dc.b	$CF,$F0,$8F,$ED,$BD,$BF,$E1,$6F,$CC,$7F,$11,$E3,$FF,$41,$FD,$BB
	.dc.b	$FD,$11,$92,$F7,$F5,$0B,$DF,$C9,$2D,$F8,$CF,$5D,$7F,$C1,$F9,$FF
	.dc.b	$88,$BC,$3F,$F1,$BE,$5C,$B1,$B5,$AC,$BA,$79,$3F,$A3,$CD,$DD,$FE
	.dc.b	$4F,$C3,$84,$E7,$A7,$18,$71,$C7,$C7,$AB,$F1,$BF,$14,$F2,$D5,$F8
	.dc.b	$03,$F5,$17,$E0,$9F,$7B,$D6,$FD,$7F,$1F,$BF,$0F,$F1,$3E,$FF,$9B
	.dc.b	$C3,$85,$EF,$F3,$7E,$B3,$F9,$FF,$33,$D7,$7D,$EE,$CF,$1C,$BB,$F7
	.dc.b	$3E,$7B,$D3,$7B,$9C,$F9,$FF,$E6,$DF,$C3,$7F,$C4,$3F,$91,$3F,$63
	.dc.b	$7E,$73,$FD,$8F,$F2,$D7,$E5,$8F,$21,$FB,$4F,$F6,$FC,$78,$E1,$C3
	.dc.b	$E5,$58,$70,$BF,$E2,$5C,$BF,$ED,$5D,$BF,$C8,$BA,$77,$D3,$FC,$DB
	.dc.b	$97,$2F,$83,$B5,$AC,$FA,$65,$CE,$3C,$06,$6F,$CD,$AB,$D7,$F6,$4F
	.dc.b	$F9,$4B,$CF,$FE,$F3,$F8,$53,$F3,$AF,$E4,$9E,$7F,$F8,$57,$F2,$2E
	.dc.b	$77,$FE,$8C,$F9,$EF,$F6,$CD,$6B,$E9,$DF,$A6,$6B,$2F,$F0,$1F,$AB
	.dc.b	$E6,$C3,$86,$76,$7B,$E1,$DB,$E5,$FC,$75,$A8,$F0,$FE,$01,$8A,$DF
	.dc.b	$C2,$FF,$B2,$F3,$7E,$F9,$E4,$F0,$3F,$9D,$7C,$33,$E7,$7F,$C3,$9F
	.dc.b	$05,$E4,$FD,$A7,$58,$FF,$44,$9C,$F0,$BF,$6A,$CE,$1E,$7C,$7F,$A2
	.dc.b	$7B,$77,$0E,$1F,$43,$FC,$9B,$E4,$BC,$F5,$E0,$CB,$BF,$D8,$FF,$AC
	.dc.b	$FF,$2B,$FF,$33,$CB,$F2,$AF,$E6,$5F,$A9,$7F,$AC,$7C,$CF,$F7,$4F
	.dc.b	$EA,$BF,$E3,$3B,$DD,$6B,$E0,$7F,$8F,$FE,$21,$FE,$D2,$F2,$55,$AF
	.dc.b	$57,$FD,$5E,$F0,$E6,$BD,$9F,$A6,$C7,$96,$7F,$E3,$48,$EC,$66,$DA
	.dc.b	$78,$FE,$5B,$FD,$95,$F6,$5E,$8F,$AB,$7D,$27,$F6,$AF,$D5,$3F,$BE
	.dc.b	$7A,$8F,$A1,$FF,$25,$5E,$EB,$5C,$FF,$BB,$F8,$0C,$EB,$FB,$8C,$BF
	.dc.b	$DA,$6F,$FB,$0A,$F8,$34,$D5,$F2,$FC,$CF,$C7,$18,$ED,$69,$FE,$0D
	.dc.b	$FC,$DB,$FD,$C1,$F8,$67,$CE,$AF,$3F,$0F,$B3,$DB,$7E,$56,$FC,$C7
	.dc.b	$FC,$C1,$7B,$D3,$E9,$D6,$BC,$0C,$E3,$FC,$38,$F3,$9F,$EA,$03,$27
	.dc.b	$F2,$23,$E2,$9F,$ED,$CA,$E9,$8F,$87,$F2,$EF,$2C,$F3,$C3,$BF,$E4
	.dc.b	$3F,$EF,$7F,$96,$F4,$78,$9F,$3B,$83,$E9,$F8,$9F,$EC,$F9,$CD,$FF
	.dc.b	$86,$BA,$BF,$39,$F8,$3F,$F9,$E6,$7E,$07,$D8,$FF,$BC,$1E,$79,$FF
	.dc.b	$BB,$9E,$38,$FD,$EF,$A8,$F0,$4F,$B7,$3F,$06,$2F,$A7,$FF,$69,$F8
	.dc.b	$6F,$C9,$7E,$67,$85,$F3,$2B,$F0,$FE,$37,$FA,$26,$73,$BF,$E7,$25
	.dc.b	$BE,$73,$F9,$5F,$94,$7C,$97,$AD,$F4,$71,$97,$77,$E8,$65,$DF,$9F
	.dc.b	$71,$8C,$BF,$53,$F8,$47,$9B,$E2,$BF,$01,$E0,$7E,$5A,$7E,$37,$77
	.dc.b	$FA,$76,$73,$1E,$EF,$92,$C6,$1F,$77,$F4,$2F,$23,$83,$EE,$61,$CF
	.dc.b	$97,$FA,$AC,$67,$CD,$FF,$7B,$CE,$73,$F8,$FF,$8C,$7E,$46,$BC,$5D
	.dc.b	$BF,$AC,$E4,$FF,$57,$4E,$74,$F9,$5B,$62,$FF,$1C,$70,$DF,$9B,$89
	.dc.b	$FC,$DE,$3F,$53,$F3,$3E,$DD,$1E,$5C,$BF,$D7,$51,$8A,$FC,$56,$38
	.dc.b	$C3,$AD,$FC,$5F,$1B,$94,$62,$BF,$7B,$F1,$7F,$80,$FC,$77,$D5,$CB
	.dc.b	$EA,$CF,$FD,$8D,$18,$AF,$D6,$E5,$8C,$3D,$37,$F4,$7F,$46,$E5,$18
	.dc.b	$AF,$E1,$3E,$DF,$E3,$3F,$45,$F8,$DC,$7F,$5E,$7F,$CE,$BD,$29,$CC
	.dc.b	$7D,$BF,$5B,$FD,$AD,$18,$85,$F3,$C3,$E4,$E1,$E1,$7F,$DB,$9F,$8A
	.dc.b	$FD,$37,$B7,$36,$1F,$9E,$78,$7D,$3C,$3F,$BB,$07,$E2,$EF,$F4,$75
	.dc.b	$CF,$0F,$E4,$7A,$7E,$9B,$F1,$4E,$DC,$9F,$91,$F7,$BF,$DD,$91,$8A
	.dc.b	$FC,$A6,$E8,$C6,$3C,$72,$F1,$E1,$D1,$FF,$7A,$7E,$2D,$E9,$E1,$E9
	.dc.b	$82,$31,$8F,$46,$E1,$FB,$87,$39,$FE,$EB,$F6,$8F,$85,$C1,$C9,$B7
	.dc.b	$E4,$6D,$EF,$2F,$8B,$5E,$3C,$5F,$90,$7C,$49,$7E,$92,$FD,$3F,$FD
	.dc.b	$27,$F1,$CF,$DA,$5F,$84,$7F,$B6,$FF,$4B,$65,$CB,$1F,$9B,$8F,$1F
	.dc.b	$AD,$FF,$10,$E1,$C3,$F0,$27,$ED,$1C,$35,$F0,$5E,$F4,$ED,$A7,$FB
	.dc.b	$9F,$2D,$1D,$35,$F1,$E5,$E2,$FC,$0F,$F8,$DE,$3F,$83,$AF,$96,$DE
	.dc.b	$2F,$AA,$DE,$1D,$CF,$E1,$8F,$85,$F9,$2F,$1E,$7F,$EC,$7D,$3E,$C8
	.dc.b	$C6,$1F,$D9,$B8,$91,$BC,$3F,$95,$BC,$79,$BF,$96,$FE,$C9,$E3,$E9
	.dc.b	$E7,$68,$C6,$1C,$DD,$61,$C3,$3F,$25,$B0,$FE,$6F,$FB,$5F,$B7,$A3
	.dc.b	$9B,$71,$F2,$39,$C3,$78,$7B,$6B,$93,$F5,$BD,$FF,$8C,$F9,$1B,$7E
	.dc.b	$D7,$35,$FD,$19,$DE,$1F,$D2,$3F,$34,$E3,$BB,$7F,$C0,$C6,$30,$B8
	.dc.b	$FC,$D3,$ED,$6C,$9F,$0F,$0F,$F5,$2F,$8D,$F1,$EF,$CE,$91,$8C,$3F
	.dc.b	$1A,$FA,$3E,$53,$FA,$CF,$C6,$FA,$B8,$38,$36,$BF,$27,$F0,$FE,$ED
	.dc.b	$3A,$36,$3F,$2F,$F7,$B9,$F7,$6D,$79,$CF,$F3,$AF,$4F,$1F,$63,$F3
	.dc.b	$AF,$3D,$BF,$3F,$F8,$3F,$5F,$59,$F9,$EA,$30,$FE,$A7,$0E,$98,$6B
	.dc.b	$0E,$52,$CB,$C4,$DA,$F1,$72,$F4,$D7,$D5,$CB,$67,$CD,$B7,$83,$7F
	.dc.b	$85,$36,$7E,$87,$07,$F8,$F3,$1F,$97,$1F,$4F,$7F,$B1,$18,$8F,$E4
	.dc.b	$87,$F0,$18,$E1,$C5,$F3,$E3,$DB,$CB,$F8,$0C,$2C,$7D,$DD,$3B,$4F
	.dc.b	$E3,$3C,$2F,$E5,$B8,$FD,$FA,$FC,$17,$36,$3A,$3F,$1F,$8F,$C8,$7F
	.dc.b	$23,$B5,$AF,$C4,$74,$ED,$FC,$57,$4F,$4E,$AF,$D0,$61,$F4,$1F,$D1
	.dc.b	$5D,$AF,$D0,$5E,$DF,$D1,$5F,$93,$AF,$F6,$38,$7D,$87,$F6,$77,$6B
	.dc.b	$F6,$37,$B7,$F6,$77,$EC,$EC,$FE,$07,$87,$E0,$1F,$E0,$B7,$6B,$F8
	.dc.b	$66,$FF,$D6,$9C,$1F,$38,$3F,$B4,$D7,$F9,$44,$66,$7F,$88,$C6,$3D
	.dc.b	$6F,$F1,$64,$7F,$88,$63,$13,$FE,$57,$8E,$D1,$88,$FE,$75,$1D,$7B
	.dc.b	$BF,$96,$F4,$7F,$28,$F0,$3F,$9C,$5F,$83,$C2,$FE,$61,$E0,$FF,$35
	.dc.b	$F0,$FF,$37,$87,$0F,$13,$F9,$CE,$1F,$CC,$F0,$E1,$8F,$3E,$5F,$BF
	.dc.b	$E3,$7F,$50,$D3,$F1,$AD,$7F,$4A,$FF,$52,$D6,$9F,$D4,$A3,$E1,$78
	.dc.b	$FF,$D8,$1F,$CA,$B5,$FD,$6B,$FD,$8A,$29,$FD,$FB,$AF,$FB,$66,$3F
	.dc.b	$DC,$63,$FB,$7D,$7C,$AF,$23,$FB,$B3,$FA,$37,$FB,$AC,$7E,$E1,$FE
	.dc.b	$F5,$CB,$F9,$8F,$27,$FC,$05,$5F,$E8,$EB,$46,$21,$E5,$7F,$00,$8D
	.dc.b	$F4,$71,$C7,$FC,$23,$5E,$4F,$33,$FC,$38,$BD,$7F,$E0,$F1,$8F,$3A
	.dc.b	$D7,$F8,$7A,$BF,$C5,$6E,$0F,$F0,$C7,$85,$FE,$21,$CD,$E8,$7F,$8D
	.dc.b	$EB,$FD,$9F,$18,$AF,$F1,$A4,$3C,$FF,$F5,$5C,$E6,$7F,$C4,$5C,$1C
	.dc.b	$EF,$96,$9F,$B8,$F5,$3F,$C9,$D8,$FF,$A4,$F0,$46,$35,$E6,$EB,$D4
	.dc.b	$E1,$97,$F9,$3E,$DE,$A7,$F0,$2F,$5F,$FC,$C9,$1F,$E8,$9C,$51,$8D
	.dc.b	$7F,$DA,$BE,$A7,$F2,$A8,$C7,$AF,$C7,$1F,$F3,$36,$1E,$BF,$5F,$F8
	.dc.b	$6F,$B3,$FE,$80,$8F,$F4,$1E,$48,$C6,$DF,$EE,$0F,$67,$F9,$AC,$7F
	.dc.b	$B2,$7D,$3F,$F4,$1D,$6F,$1F,$F6,$17,$27,$FA,$0F,$3E,$B1,$8C,$BF
	.dc.b	$D7,$19,$FF,$1C,$F6,$BF,$D4,$7B,$FF,$A0,$7D,$CF,$E8,$31,$8F,$6F
	.dc.b	$FD,$DF,$EE,$63,$1F,$F5,$29,$FE,$A5,$D3,$FD,$44,$FF,$B3,$7A,$3C
	.dc.b	$77,$C7,$4E,$D3,$FE,$A3,$CB,$FD,$F3,$9F,$F2,$CF,$7F,$FD,$7F,$9B
	.dc.b	$E0,$1A,$FF,$52,$9C,$C7,$FA,$DB,$44,$7F,$B2,$B1,$F1,$3F,$EC,$A7
	.dc.b	$FD,$C1,$1A,$7F,$D9,$1B,$7F,$B4,$FE,$03,$C1,$18,$E4,$65,$FE,$CB
	.dc.b	$8F,$EC,$3F,$07,$FE,$E4,$86,$DF,$EF,$4F,$85,$CD,$7F,$B9,$77,$FF
	.dc.b	$7C,$7C,$33,$E1,$B3,$3F,$EE,$FF,$4F,$EA,$E5,$FE,$F9,$8F,$90,$F5
	.dc.b	$C5,$F0,$DF,$D3,$A3,$1F,$0D,$FD,$43,$E2,$3F,$CB,$5E,$A3,$AB,$FD
	.dc.b	$4F,$F1,$2F,$7D,$F1,$5F,$DE,$2F,$75,$A9,$78,$99,$D7,$FD,$0F,$8F
	.dc.b	$3D,$BF,$13,$87,$0E,$6B,$5A,$C7,$BF,$B9,$E3,$8C,$4F,$F0,$1F,$11
	.dc.b	$AB,$E2,$BF,$B3,$FC,$6F,$A3,$C1,$FD,$C7,$4F,$DC,$F8,$D7,$C7,$3E
	.dc.b	$3F,$FC,$1F,$7B,$EB,$FF,$7B,$D6,$BF,$28,$CE,$DE,$8C,$BF,$81,$3F
	.dc.b	$91,$BE,$DE,$97,$E8,$9E,$59,$F7,$F8,$BF,$1D,$71,$CB,$AF,$C8,$7F
	.dc.b	$7C,$FA,$0F,$9B,$F0,$5F,$C6,$FC,$7F,$B5,$9B,$E4,$FF,$C7,$3E,$3A
	.dc.b	$F7,$E3,$7E,$5F,$C4,$FE,$53,$4F,$8A,$FF,$14,$D6,$B5,$FD,$33,$DB
	.dc.b	$D6,$FD,$BB,$CB,$3E,$FD,$1E,$3A,$FA,$BA,$F8,$F2,$FE,$6B,$F2,$79
	.dc.b	$9F,$F2,$B7,$0B,$E6,$1F,$2F,$FD,$CB,$E6,$DF,$30,$F9,$8F,$1E,$1F
	.dc.b	$EC,$6F,$11,$8F,$1C,$38,$5B,$E6,$F7,$7D,$BE,$6F,$CE,$F1,$BF,$03
	.dc.b	$EC,$31,$9F,$8B,$66,$8F,$9B,$F4,$69,$E0,$AD,$4F,$F7,$EE,$2F,$F5
	.dc.b	$B4,$E7,$5F,$1F,$2F,$8B,$E8,$5F,$40,$FA,$07,$D1,$BE,$75,$F4,$2F
	.dc.b	$A3,$7D,$27,$1C,$38,$5E,$EB,$5C,$39,$C7,$DB,$C4,$CF,$BF,$D2,$61
	.dc.b	$F0,$CF,$83,$4E,$FA,$7D,$39,$CE,$5D,$7E,$93,$FC,$8F,$F5,$BF,$B8
	.dc.b	$F1,$7F,$84,$FA,$BF,$CE,$FC,$B7,$D4,$BE,$FB,$FC,$DF,$7B,$F7,$3F
	.dc.b	$79,$F0,$CB,$DF,$D5,$6B,$8B,$F7,$39,$BE,$AB,$FA,$0E,$3C,$67,$F9
	.dc.b	$A3,$CF,$4F,$07,$D5,$79,$27,$8E,$7D,$B2,$FE,$BB,$1D,$4F,$ED,$DF
	.dc.b	$14,$FA,$DF,$E1,$3F,$64,$FB,$07,$D7,$BE,$BF,$86,$EC,$38,$7C,$13
	.dc.b	$0E,$16,$FB,$2D,$D3,$EC,$BD,$3A,$0F,$0E,$98,$6E,$3C,$F4,$7C,$27
	.dc.b	$AE,$73,$9F,$97,$EC,$B1,$AF,$8F,$1F,$4E,$1E,$8F,$CB,$FF,$D3,$BF
	.dc.b	$7B,$DF,$1F,$CB,$FA,$DF,$3F,$ED,$9F,$6E,$FB,$9F,$EA,$7E,$7F,$0B
	.dc.b	$F3,$79,$EC,$38,$5E,$FE,$4D,$F7,$1B,$BE,$BC,$1F,$DD,$B4,$F0,$63
	.dc.b	$C7,$A7,$15,$D3,$2E,$7A,$7E,$6E,$39,$56,$F5,$F2,$CB,$EE,$7F,$D2
	.dc.b	$B8,$F4,$C3,$8E,$8F,$57,$8C,$63,$EE,$BC,$9F,$52,$FB,$CF,$A5,$F3
	.dc.b	$0F,$BC,$7D,$EB,$EF,$BF,$D5,$BA,$7C,$7C,$38,$77,$7F,$72,$BD,$D7
	.dc.b	$F8,$D7,$E1,$97,$BC,$9F,$11,$EE,$E4,$65,$C9,$FF,$05,$56,$F2,$E9
	.dc.b	$9B,$9B,$DB,$5F,$85,$BF,$0E,$BE,$7A,$FC,$1C,$FB,$32,$F5,$EB,$E4
	.dc.b	$F8,$8F,$A5,$F7,$0F,$C0,$7F,$61,$F8,$07,$FA,$63,$CD,$FF,$2D,$78
	.dc.b	$7A,$9C,$FB,$3F,$AD,$F8,$B7,$A9,$9F,$4F,$D5,$97,$3F,$C0,$6E,$1C
	.dc.b	$32,$FE,$DF,$BE,$F4,$ED,$F4,$BE,$A8,$FF,$73,$56,$F3,$7E,$05,$FE
	.dc.b	$F0,$FC,$43,$F2,$9F,$83,$BD,$F7,$E0,$CF,$C1,$FF,$B7,$F6,$BF,$07
	.dc.b	$FE,$17,$DE,$C7,$8E,$1C,$2F,$74,$FC,$53,$B6,$DF,$AF,$CF,$B4,$FF
	.dc.b	$92,$B0,$E3,$9D,$AD,$63,$CF,$E4,$BE,$DF,$D3,$FF,$B9,$2F,$79,$F9
	.dc.b	$3F,$08,$C6,$1C,$31,$F4,$61,$F8,$2C,$9F,$84,$F8,$A9,$A7,$BE,$F7
	.dc.b	$BF,$B7,$F0,$AD,$7D,$0B,$F0,$76,$9F,$D3,$7F,$0C,$7E,$18,$FC,$2B
	.dc.b	$FC,$3F,$CB,$F0,$F5,$FE,$19,$8F,$1C,$38,$5E,$FE,$A7,$AE,$4F,$C3
	.dc.b	$7E,$06,$0C,$38,$3F,$E5,$FE,$4F,$F6,$9D,$78,$BE,$5D,$AF,$8B,$F0
	.dc.b	$8F,$83,$7C,$6B,$53,$E6,$FB,$FC,$E3,$E3,$C3,$8E,$9E,$93,$C7,$AF
	.dc.b	$9B,$77,$E1,$DF,$1F,$F1,$1F,$F2,$17,$E2,$0F,$C3,$DF,$88,$FF,$64
	.dc.b	$FA,$47,$E8,$FB,$FB,$15,$EE,$B5,$F4,$F9,$FA,$3F,$7E,$5D,$39,$3D
	.dc.b	$C7,$F3,$37,$BE,$7D,$BE,$63,$F4,$73,$D7,$1E,$91,$FD,$FE,$31,$3B
	.dc.b	$D1,$F9,$96,$33,$F3,$78,$9F,$81,$FB,$37,$E2,$CF,$C2,$5F,$8B,$BF
	.dc.b	$18,$7E,$2F,$F9,$9F,$C6,$38,$5E,$EB,$53,$F9,$DF,$89,$CE,$1F,$6B
	.dc.b	$1E,$BF,$8C,$DC,$38,$65,$D3,$E7,$3E,$BD,$78,$AF,$C3,$3E,$0F,$53
	.dc.b	$37,$C3,$2F,$D5,$7D,$77,$6D,$3C,$3E,$97,$F6,$5F,$27,$F2,$1C,$AD
	.dc.b	$3A,$56,$BF,$1B,$E1,$A7,$B3,$07,$E3,$6C,$5F,$8E,$FF,$55,$F8,$0F
	.dc.b	$F0,$17,$E3,$BF,$E0,$8C,$FF,$75,$F9,$03,$D0,$E1,$7B,$F9,$45,$EF
	.dc.b	$35,$BF,$20,$DF,$07,$F6,$FC,$FF,$C1,$DF,$90,$7F,$B2,$B2,$E5,$8D
	.dc.b	$AD,$65,$D3,$EA,$7D,$99,$BE,$DF,$FB,$C6,$FE,$8B,$DE,$1E,$1E,$DC
	.dc.b	$E5,$D3,$1E,$3B,$7E,$67,$0E,$18,$7F,$89,$38,$5F,$92,$3F,$22,$7E
	.dc.b	$48,$FC,$91,$F9,$2F,$FB,$1B,$1E,$39,$FC,$AC,$38,$5E,$EB,$FD,$91
	.dc.b	$A7,$F7,$FA,$F9,$FB,$7B,$74,$F0,$61,$67,$96,$7E,$0C,$DF,$86,$7A
	.dc.b	$AF,$FC,$2D,$5A,$CF,$C9,$F3,$F9,$9F,$F5,$7F,$DD,$3F,$17,$7E,$31
	.dc.b	$FE,$DC,$F8,$7F,$F0,$CF,$E4,$DF,$4A,$F7,$5A,$FA,$0F,$E8,$1F,$55
	.dc.b	$FE,$2B,$E4,$F8,$BF,$56,$CD,$F8,$C7,$A6,$31,$85,$9F,$06,$9D,$F0
	.dc.b	$F2,$56,$A3,$C5,$F4,$98,$AD,$E5,$E0,$FC,$9F,$C9,$7B,$8F,$1F,$C9
	.dc.b	$FC,$2D,$7F,$C5,$1F,$8B,$FF,$55,$FC,$A5,$F9,$4F,$F3,$4F,$C2,$9F
	.dc.b	$8C,$7F,$8A,$BA,$3F,$C5,$D8,$30,$E1,$7B,$A7,$E5,$7D,$6B,$FC,$85
	.dc.b	$5F,$D8,$7F,$2B,$E3,$1F,$E4,$5C,$39,$CF,$86,$33,$18,$E7,$67,$E4
	.dc.b	$F8,$56,$8F,$C5,$39,$61,$EA,$BD,$EB,$E9,$E7,$CC,$FF,$94,$FF,$1C
	.dc.b	$FF,$6B,$FE,$4B,$FE,$3C,$F3,$3F,$D6,$1F,$97,$7F,$A7,$B6,$F6,$D6
	.dc.b	$BB,$BF,$CC,$1F,$68,$A7,$E5,$DE,$DC,$97,$D1,$F4,$9F,$E5,$37,$FD
	.dc.b	$CB,$F9,$27,$C0,$8E,$11,$D7,$2F,$F6,$07,$D6,$F9,$A7,$9F,$E2,$AD
	.dc.b	$7C,$D9,$BF,$2E,$F9,$1F,$99,$7F,$97,$FF,$31,$7E,$62,$FB,$CF,$F6
	.dc.b	$57,$E6,$4F,$53,$F3,$FF,$91,$7E,$83,$0E,$17,$BC,$97,$BE,$57,$E1
	.dc.b	$5B,$DF,$FC,$F5,$A7,$6D,$3D,$54,$CB,$9F,$CD,$61,$EB,$3E,$2D,$7F
	.dc.b	$1D,$5B,$C3,$9E,$9E,$1C,$FC,$DF,$83,$FB,$27,$FD,$11,$F9,$9F,$F4
	.dc.b	$EF,$CD,$5F,$90,$FF,$96,$BE,$AD,$BB,$F0,$CF,$36,$0C,$38,$5E,$EB
	.dc.b	$FD,$FD,$A7,$86,$B5,$F2,$4F,$CD,$7D,$11,$E0,$86,$78,$8D,$BE,$2F
	.dc.b	$BA,$FF,$1C,$7D,$0F,$9E,$73,$97,$97,$ED,$B1,$F9,$97,$FA,$87,$F3
	.dc.b	$6F,$F3,$4F,$57,$4F,$CD,$7F,$D8,$FB,$78,$EB,$5F,$80,$AB,$5F,$11
	.dc.b	$EC,$9F,$F5,$57,$E1,$1E,$37,$A6,$9F,$A6,$EB,$FF,$50,$47,$2B,$F4
	.dc.b	$D7,$8A,$99,$F8,$B8,$1F,$43,$FB,$1F,$A3,$FF,$54,$7E,$72,$FC,$E3
	.dc.b	$FC,$23,$F9,$4B,$F3,$97,$CE,$BF,$3C,$7E,$8B,$FC,$AB,$1E,$38,$70
	.dc.b	$BD,$D3,$F3,$DD,$6D,$FE,$C3,$CF,$B7,$E7,$BF,$8A,$C7,$8F,$D2,$DA
	.dc.b	$D6,$5C,$F4,$FF,$22,$C7,$21,$87,$9E,$9A,$F9,$E9,$E1,$FB,$6F,$FD
	.dc.b	$A5,$D5,$FE,$87,$F0,$DF,$9F,$3F,$25,$FA,$1A,$3F,$24,$ED,$7B,$AD
	.dc.b	$74,$7C,$5F,$D2,$F3,$F8,$1F,$F9,$E7,$3F,$E2,$FC,$3F,$ED,$1C,$3F
	.dc.b	$D2,$9B,$FA,$E7,$FD,$1F,$5F,$E7,$FC,$79,$F8,$3E,$59,$CE,$DE,$4D
	.dc.b	$BF,$05,$F9,$86,$7E,$DB,$F9,$0F,$E8,$0F,$CB,$5F,$99,$7F,$BD,$FF
	.dc.b	$2E,$7D,$87,$C5,$BD,$D6,$BF,$20,$57,$FB,$87,$C8,$71,$FF,$B9,$EF
	.dc.b	$AE,$DE,$A3,$FC,$DB,$F1,$1F,$E1,$97,$07,$F9,$A5,$69,$E3,$D1,$97
	.dc.b	$F3,$3A,$F1,$56,$B3,$F3,$FE,$3F,$FE,$F5,$CF,$C3,$43,$FD,$13,$F9
	.dc.b	$1B,$EA,$FD,$72,$FC,$D7,$F2,$5F,$17,$F8,$1F,$C2,$17,$BE,$2F,$EF
	.dc.b	$75,$F9,$7F,$C3,$75,$2F,$D3,$BE,$95,$F0,$CA,$CF,$1C,$BB,$78,$3F
	.dc.b	$4E,$B5,$97,$4F,$8F,$FA,$DE,$AB,$F4,$2F,$F3,$8F,$E8,$7F,$0F,$ED
	.dc.b	$BE,$1E,$5E,$CE,$1F,$F7,$7D,$6B,$A3,$FA,$CD,$7F,$90,$EB,$58,$73
	.dc.b	$FB,$DF,$C3,$3D,$7D,$0B,$87,$1E,$76,$1E,$28,$C7,$CD,$7E,$27,$F0
	.dc.b	$57,$E8,$8F,$B0,$F8,$BF,$9B,$BF,$45,$7E,$45,$FE,$2F,$DF,$CD,$87
	.dc.b	$0B,$DD,$6B,$F1,$9F,$D7,$4F,$4F,$91,$FC,$D3,$BE,$3F,$C1,$BC,$7A
	.dc.b	$59,$F7,$F1,$BC,$9A,$F8,$E7,$39,$7F,$3A,$C3,$F7,$5D,$CF,$C3,$9F
	.dc.b	$9B,$FF,$0D,$F6,$7F,$D6,$DF,$96,$BE,$57,$75,$A9,$FE,$25,$D6,$AF
	.dc.b	$F7,$DF,$1F,$D1,$C7,$A6,$3F,$B1,$7B,$70,$5C,$79,$63,$FE,$C3,$FC
	.dc.b	$47,$F6,$55,$C2,$3A,$FE,$5F,$FE,$15,$FC,$EB,$F2,$DB,$3F,$49,$FE
	.dc.b	$43,$F9,$4B,$F4,$87,$E9,$2F,$D2,$5F,$83,$3E,$3F,$FC,$9F,$97,$2C
	.dc.b	$78,$E1,$C2,$F7,$F4,$DA,$CF,$F4,$95,$D9,$FF,$11,$DF,$1F,$88,$FF
	.dc.b	$46,$C6,$0D,$AD,$3C,$9A,$78,$F3,$F5,$56,$A7,$39,$7F,$8E,$63,$C0
	.dc.b	$7F,$BE,$3C,$8F,$F6,$7F,$E9,$5F,$D8,$B1,$7E,$97,$EC,$E7,$69,$F1
	.dc.b	$3F,$1C,$61,$C3,$E3,$3F,$A2,$DF,$8E,$F7,$97,$EF,$EF,$7F,$8A,$7F
	.dc.b	$82,$B1,$E5,$8F,$F8,$5E,$BD,$87,$8E,$59,$F8,$FF,$35,$F5,$CE,$63
	.dc.b	$CD,$F0,$98,$FC,$39,$FA,$6B,$F4,$DF,$E8,$BE,$23,$F1,$EF,$ED,$DF
	.dc.b	$17,$F0,$2F,$75,$A9,$FE,$E5,$F1,$59,$C1,$F9,$97,$9F,$4F,$E3,$18
	.dc.b	$70,$7C,$EF,$2F,$8B,$FE,$2B,$87,$F2,$3E,$3F,$83,$5F,$EE,$3C,$3C
	.dc.b	$95,$AF,$AF,$F5,$FE,$70,$FD,$3B,$FD,$91,$F4,$08,$7E,$73,$FE,$C2
	.dc.b	$FA,$EF,$AF,$1B,$FC,$07,$FC,$46,$EB,$F0,$87,$E8,$9A,$C7,$FC,$AD
	.dc.b	$AF,$67,$F9,$E3,$FE,$7A,$DB,$CB,$7C,$F3,$F0,$66,$F2,$38,$67,$DA
	.dc.b	$1E,$5F,$D2,$F8,$67,$B1,$FC,$97,$EA,$9E,$E7,$FA,$F3,$F1,$A7,$E8
	.dc.b	$9F,$8E,$BD,$FB,$FF,$E5,$5A,$D7,$DF,$E7,$1F,$E5,$DC,$3F,$82,$BE
	.dc.b	$BE,$57,$D1,$8F,$E7,$93,$DE,$FF,$D9,$3F,$03,$E3,$8C,$7E,$1B,$FE
	.dc.b	$18,$AF,$CE,$FE,$62,$FC,$C1,$F9,$93,$F3,$27,$E6,$4F,$8A,$7E,$66
	.dc.b	$FC,$E3,$8E,$1C,$2F,$75,$AF,$C2,$73,$CF,$E5,$CF,$B6,$77,$0E,$19
	.dc.b	$F5,$F8,$9F,$CC,$D6,$B2,$E7,$F8,$67,$E7,$3F,$4E,$FD,$5F,$C2,$3D
	.dc.b	$35,$FE,$4B,$F4,$F7,$E3,$DF,$B8,$BD,$D6,$BC,$E5,$3F,$4C,$D7,$AF
	.dc.b	$FC,$BB,$9B,$F5,$B8,$7E,$E3,$E1,$FE,$63,$CB,$3F,$CF,$FD,$4F,$8E
	.dc.b	$FF,$9A,$E7,$D7,$F4,$8F,$EF,$1D,$9F,$E6,$BF,$D7,$3F,$2B,$F5,$6F
	.dc.b	$D9,$FF,$B8,$7A,$FF,$E2,$BF,$97,$FE,$B3,$F6,$BF,$CA,$BF,$2E,$7E
	.dc.b	$F2,$FC,$8D,$D6,$FD,$27,$FC,$C5,$F5,$9F,$57,$6F,$F3,$C6,$1F,$90
	.dc.b	$E0,$7E,$36,$FD,$49,$F9,$B7,$F1,$2F,$B0,$EF,$D5,$FE,$0D,$C3,$FC
	.dc.b	$BF,$FB,$CF,$EB,$7D,$7F,$E2,$FE,$13,$F6,$ED,$DA,$DF,$49,$FE,$F4
	.dc.b	$FD,$43,$F3,$7E,$0F,$E6,$BF,$31,$DD,$6B,$F6,$D5,$3F,$7B,$7D,$73
	.dc.b	$E6,$BF,$55,$CB,$EA,$CD,$DA,$C7,$FD,$99,$F3,$2F,$CE,$1F,$8A,$76
	.dc.b	$BD,$FE,$34,$FC,$13,$F9,$9F,$EF,$6F,$C2,$B5,$9B,$F3,$47,$E1,$2F
	.dc.b	$CF,$3B,$7E,$D3,$A9,$7E,$29,$E8,$FD,$11,$E7,$67,$E1,$1F,$99,$FA
	.dc.b	$7F,$13,$E1,$F1,$3F,$D0,$BF,$76,$FF,$18,$F4,$7A,$F7,$F0,$4B,$F0
	.dc.b	$A7,$ED,$4F,$DA,$78,$F8,$2F,$DD,$5F,$B3,$F4,$FD,$11,$EC,$FF,$BE
	.dc.b	$F4,$E9,$C9,$FE,$2F,$9F,$F6,$AF,$CA,$6F,$F1,$9F,$E7,$55,$AD,$95
	.dc.b	$FE,$C1,$FC,$93,$5F,$B2,$6F,$F6,$2E,$1F,$8E,$71,$FD,$F3,$C1,$FB
	.dc.b	$53,$F3,$CF,$17,$C4,$7F,$05,$D3,$A6,$7F,$B1,$F2,$3E,$0F,$EB,$3F
	.dc.b	$E4,$BF,$37,$E1,$FD,$0F,$FB,$1B,$E8,$3E,$67,$EF,$5F,$E0,$2F,$D5
	.dc.b	$9F,$C0,$3F,$3F,$F7,$37,$E7,$1F,$5F,$CB,$FA,$FF,$87,$3F,$37,$7E
	.dc.b	$B0,$F4,$9F,$AF,$FD,$CF,$C6,$5F,$AB,$B8,$1F,$BD,$FF,$AE,$B9,$BE
	.dc.b	$97,$ED,$6F,$DC,$FE,$CF,$E8,$BF,$B3,$9F,$C8,$7E,$25,$FD,$93,$E9
	.dc.b	$3F,$C8,$FF,$B0,$BF,$3F,$FC,$77,$EA,$DF,$DA,$3F,$19,$7E,$87,$FE
	.dc.b	$D3,$F9,$9F,$CF,$FA,$4F,$D5,$7F,$63,$ED,$3F,$56,$7E,$F0,$FD,$1B
	.dc.b	$F4,$BF,$99,$3F,$4B,$F7,$78,$DE,$EF,$51,$F8,$2F,$EC,$3E,$DB,$F7
	.dc.b	$7F,$5B,$E9,$FC,$D5,$FB,$B7,$D5,$FC,$BF,$F6,$BF,$AE,$BF,$51,$7D
	.dc.b	$7A,$3C,$5F,$99,$BB,$3F,$49,$DD,$FE,$FF,$E4,$FD,$47,$CA,$FF,$B1
	.dc.b	$BD,$17,$CE,$FC,$FF,$C6,$7F,$80,$FC,$CA,$5F,$BE,$FF,$43,$FC,$A7
	.dc.b	$FD,$5F,$F0,$2F,$DB,$FF,$83,$FE,$AB,$FE,$6C,$FC,$3D,$FA,$AF,$BF
	.dc.b	$F3,$5F,$F7,$7F,$E5,$DA,$F8,$C7,$E1,$3F,$E5,$ED,$BF,$4B,$F9,$1F
	.dc.b	$C2,$E4,$79,$3F,$E4,$ED,$3E,$D6,$5F,$80,$F5,$1F,$A1,$7C,$BE,$EF
	.dc.b	$0F,$A7,$FC,$87,$D6,$FE,$2F,$F9,$B7,$E4,$3E,$FF,$3F,$06,$FD,$B5
	.dc.b	$FA,$7F,$E3,$BF,$50,$7E,$FF,$CB,$F5,$F7,$EA,$BF,$13,$F6,$EF,$F4
	.dc.b	$3F,$46,$78,$9F,$93,$FF,$04,$F1,$BF,$8F,$73,$FA,$3E,$CB,$FD,$75
	.dc.b	$CF,$F9,$1C,$FF,$84,$FC,$2B,$F7,$4F,$F6,$8F,$DA,$77,$FA,$5F,$C4
	.dc.b	$7D,$8B,$F7,$E9,$9F,$3C,$D9,$F3,$BF,$F4,$17,$E0,$BF,$E6,$DE,$EE
	.dc.b	$5F,$99,$BF,$4E,$FE,$ED,$E7,$63,$E0,$9F,$B5,$7E,$5F,$4E,$99,$F3
	.dc.b	$C3,$F8,$A7,$C6,$FD,$B9,$CF,$84,$E4,$FE,$43,$5F,$C6,$7C,$07,$EB
	.dc.b	$7D,$7F,$2D,$7C,$4D,$FC,$DB,$1D,$39,$FC,$6B,$5F,$1F,$BF,$A8,$F2
	.dc.b	$7E,$59,$FE,$6A,$DF,$BE,$DD,$B8,$DF,$BF,$BE,$47,$D5,$F9,$C7,$F1
	.dc.b	$EF,$43,$F0,$9C,$9E,$5E,$CF,$F5,$27,$E8,$2F,$D4,$FF,$C3,$1A,$74
	.dc.b	$97,$E5,$FF,$D9,$70,$FE,$0C,$72,$D7,$F9,$EF,$AB,$FC,$43,$F5,$4F
	.dc.b	$D5,$E1,$C7,$F2,$7F,$DA,$FE,$D5,$FE,$FC,$8E,$9E,$27,$F6,$7F,$CF
	.dc.b	$DF,$8B,$FE,$27,$F3,$BF,$F0,$AF,$E1,$CF,$D9,$7F,$D4,$D8,$7B,$BD
	.dc.b	$E7,$EE,$9F,$D8,$FF,$40,$FF,$2A,$CF,$D0,$F8,$E7,$EE,$1F,$DC,$FF
	.dc.b	$41,$7E,$3F,$FE,$05,$AF,$AB,$F9,$C7,$83,$F3,$1F,$AD,$F2,$7F,$EE
	.dc.b	$09,$FB,$3F,$86,$7E,$0F,$E4,$DF,$98,$3F,$21,$FE,$C7,$5F,$82,$FD
	.dc.b	$49,$FB,$6B,$F5,$FF,$F7,$2F,$BA,$FD,$81,$F4,$9F,$BE,$FC,$87,$F6
	.dc.b	$B2,$FD,$F7,$F8,$4F,$EE,$1F,$E1,$1C,$FF,$31,$F8,$DF,$F9,$7B,$F5
	.dc.b	$C7,$E7,$3F,$95,$F7,$5F,$61,$FD,$7A,$FF,$D9,$7E,$27,$F9,$3F,$C6
	.dc.b	$FE,$C3,$FA,$6B,$F4,$EF,$F3,$FF,$DE,$7F,$80,$AF,$7F,$8F,$FF,$8F
	.dc.b	$3F,$47,$F9,$DF,$82,$3F,$79,$7E,$C3,$F2,$AF,$7F,$91,$FF,$C0,$0F
	.dc.b	$C9,$7F,$92,$77,$7F,$9D,$B9,$95,$D3,$F0,$4F,$F2,$5F,$EB,$DF,$C8
	.dc.b	$7F,$7E,$7E,$A9,$FD,$37,$93,$FC,$71,$F9,$87,$C4,$EC,$FD,$C7,$C5
	.dc.b	$BF,$72,$75,$74,$BF,$F1,$F7,$E4,$DF,$36,$DF,$A6,$3F,$53,$7E,$E4
	.dc.b	$F8,$3E,$18,$7F,$8D,$F0,$FE,$83,$F9,$BF,$F8,$13,$F7,$2F,$DD,$FE
	.dc.b	$A5,$AA,$FF,$1B,$57,$F8,$AB,$F4,$3F,$ED,$BE,$B3,$EC,$3F,$B0,$57
	.dc.b	$8B,$F2,$4F,$F7,$9F,$E5,$7E,$CF,$BC,$FE,$6F,$D5,$F3,$32,$FF,$19
	.dc.b	$FE,$43,$FD,$B7,$C7,$FF,$34,$7E,$DD,$DF,$C6,$FF,$21,$E5,$FE,$2D
	.dc.b	$FC,$B1,$FB,$D7,$F7,$8F,$5B,$59,$72,$FC,$8F,$F3,$7F,$18,$FC,$DF
	.dc.b	$38,$FF,$89,$BC,$3F,$F4,$87,$C3,$E3,$76,$0F,$C9,$FF,$90,$FE,$67
	.dc.b	$98,$FF,$0E,$FE,$67,$F9,$AE,$6F,$65,$FF,$89,$7E,$81,$FB,$C6,$FD
	.dc.b	$17,$E5,$2F,$D7,$B0,$FF,$82,$BA,$7E,$07,$07,$F8,$E2,$B7,$F8,$1E
	.dc.b	$BF,$29,$DF,$EC,$9C,$FF,$28,$7E,$61,$D7,$1F,$D2,$9C,$B6,$FF,$1B
	.dc.b	$7E,$8C,$FE,$01,$F4,$7E,$3E,$F8,$4F,$F0,$F5,$EF,$95,$FB,$63,$F5
	.dc.b	$7F,$E0,$FC,$DE,$C9,$FF,$5A,$FE,$4C,$FC,$E7,$E3,$7E,$5D,$FE,$0D
	.dc.b	$F6,$9F,$5D,$F2,$7F,$08,$FD,$A7,$CA,$FF,$00,$D7,$DB,$F0,$BF,$E3
	.dc.b	$0F,$B3,$FF,$68,$FE,$E1,$F4,$BE,$DB,$FC,$C9,$FA,$4B,$F7,$4F,$D7
	.dc.b	$70,$FE,$1F,$F6,$0F,$E5,$9E,$77,$F8,$D7,$F5,$DF,$BB,$FA,$B7,$D6
	.dc.b	$F0,$5F,$B5,$7F,$9C,$FF,$0B,$FC,$0F,$E1,$9F,$EE,$7F,$9D,$FF,$01
	.dc.b	$FD,$67,$F9,$9B,$4F,$F5,$07,$DB,$BF,$27,$FE,$1F,$F6,$AF,$83,$FF
	.dc.b	$0B,$D0,$9F,$90,$FF,$8B,$FF,$4B,$FF,$25,$FD,$9B,$F0,$07,$0B,$F6
	.dc.b	$AF,$D2,$5F,$F9,$8B,$F7,$FC,$7A,$4F,$DD,$7F,$88,$FE,$9D,$FA,$9F
	.dc.b	$57,$FC,$63,$FB,$13,$F7,$96,$1F,$C2,$FF,$7D,$FF,$5D,$7D,$B7,$F8
	.dc.b	$13,$87,$C9,$B3,$F6,$2F,$EC,$77,$F2,$7F,$13,$FE,$3F,$EB,$7F,$87
	.dc.b	$BF,$17,$F5,$60,$FD,$DB,$FC,$1F,$F0,$CF,$C5,$BF,$0D,$F8,$A3,$F6
	.dc.b	$0F,$F2,$7E,$DD,$F0,$FF,$59,$67,$FE,$F3,$FA,$77,$8A,$F9,$8F,$EB
	.dc.b	$BF,$BC,$F7,$BF,$F5,$F7,$E4,$5F,$E5,$BF,$A4,$FE,$75,$F3,$7F,$D3
	.dc.b	$3F,$4D,$D5,$7F,$AF,$7E,$97,$FC,$49,$FB,$2F,$F7,$CC,$1F,$81,$FF
	.dc.b	$96,$B7,$F1,$7C,$C7,$FB,$0B,$E5,$59,$FF,$A9,$3F,$46,$F1,$F9,$1D
	.dc.b	$2B,$FD,$8F,$F9,$1F,$FA,$73,$F0,$77,$EC,$DE,$8F,$AC,$D7,$CD,$FF
	.dc.b	$B0,$3E,$8F,$FC,$17,$D0,$FD,$DD,$EC,$EB,$CC,$72,$3F,$6C,$7E,$CC
	.dc.b	$F5,$DF,$B2,$EF,$C7,$FF,$1D,$FE,$16,$FD,$1D,$F6,$D9,$FC,$4D,$F2
	.dc.b	$FF,$E4,$AF,$1F,$87,$B8,$FD,$C3,$FC,$F3,$F9,$C7,$3F,$94,$7E,$C2
	.dc.b	$F8,$7F,$F0,$5F,$87,$F1,$3A,$BF,$8F,$7E,$60,$FD,$F1,$FB,$F7,$1F
	.dc.b	$A2,$FE,$01,$F9,$2D,$FC,$D1,$F8,$F6,$7E,$01,$FB,$6D,$AC,$7E,$5B
	.dc.b	$9E,$CE,$5F,$4B,$FE,$2E,$DF,$C1,$E0,$FF,$17,$FD,$7B,$FD,$D3,$18
	.dc.b	$EE,$FF,$63,$7C,$53,$F5,$7F,$57,$D2,$BE,$5B,$7F,$05,$FC,$37,$F6
	.dc.b	$87,$CC,$BF,$62,$F9,$5E,$27,$FA,$03,$F4,$A7,$E0,$FF,$37,$F6,$CF
	.dc.b	$B7,$FB,$8B,$F5,$5F,$AD,$F5,$3F,$E4,$7F,$CF,$FE,$7F,$EE,$FF,$67
	.dc.b	$E2,$9C,$CF,$CF,$FF,$B3,$7B,$7F,$77,$E8,$7F,$BC,$7A,$EF,$89,$FC
	.dc.b	$6F,$5F,$CD,$FD,$BB,$FC,$5B,$F3,$BF,$63,$F3,$FF,$F7,$F4,$FD,$1E
	.dc.b	$67,$D3,$B7,$7E,$65,$FE,$31,$D3,$F1,$FF,$4A,$E1,$FF,$03,$70,$FF
	.dc.b	$09,$FC,$77,$F3,$1F,$B9,$7D,$9F,$AF,$7E,$E5,$F7,$FE,$67,$FD,$A7
	.dc.b	$BF,$EC,$FE,$8B,$F9,$D7,$E0,$1F,$C9,$FF,$0D,$7E,$AA,$FC,$6F,$FC
	.dc.b	$6F,$7F,$B9,$E9,$FF,$06,$F3,$79,$7F,$1C,$7E,$B5,$FE,$93,$FD,$6F
	.dc.b	$F6,$7D,$CF,$0D,$FA,$5B,$F6,$0F,$F1,$8F,$67,$F8,$DB,$3F,$F1,$DF
	.dc.b	$97,$FC,$33,$0E,$17,$BF,$C8,$BE,$1F,$47,$EA,$BF,$1B,$F1,$7D,$C3
	.dc.b	$83,$C5,$BF,$86,$7D,$9F,$80,$39,$BD,$9C,$BE,$BF,$CA,$7B,$FD,$0F
	.dc.b	$F3,$3F,$9A,$77,$7C,$7E,$DF,$8D,$FA,$BE,$AB,$FC,$27,$E1,$7C,$FF
	.dc.b	$C7,$BF,$03,$F6,$9F,$E4,$ED,$BE,$C7,$47,$D7,$F1,$BE,$B4,$7D,$FF
	.dc.b	$CB,$FF,$DF,$1F,$23,$E7,$FD,$87,$A6,$FF,$E1,$0F,$27,$F0,$FE,$77
	.dc.b	$F1,$2C,$3F,$47,$E5,$7E,$3B,$B7,$F1,$B1,$FA,$6F,$47,$F2,$9E,$B7
	.dc.b	$E1,$63,$F4,$DE,$1F,$E6,$BD,$3E,$BF,$17,$FE,$20,$AF,$D0,$7A,$D8
	.dc.b	$F7,$DE,$FF,$F8,$53,$DB,$FD,$3F,$A9,$98,$FF,$0B,$FB,$BF,$AD,$AC
	.dc.b	$7C,$73,$E2,$5F,$03,$FB,$3D,$FF,$6D,$39,$F5,$D7,$C2,$DF,$05,$FE
	.dc.b	$AC,$BD,$C7,$F8,$6B,$E0,$F5,$F0,$EF,$86,$BF,$0F,$C7,$E2,$7F,$86
	.dc.b	$E1,$C3,$CC,$D3,$E0,$F8,$CF,$E2,$DF,$17,$E6,$F8,$EF,$E3,$9F,$99
	.dc.b	$FA,$FE,$4B,$F9,$17,$EA,$6F,$92,$F7,$7C,$AF,$F2,$8F,$D2,$7F,$AB
	.dc.b	$67,$FC,$3F,$EF,$BF,$D3,$DF,$31,$FC,$BA,$3F,$8A,$FC,$EF,$F3,$7F
	.dc.b	$9C,$FE,$6D,$F3,$9F,$2F,$E8,$3F,$9F,$FC,$FF,$E5,$7E,$0F,$FC,$CB
	.dc.b	$F4,$7F,$D1,$FE,$8F,$FA,$36,$5F,$CB,$7E,$9F,$FA,$7F,$E6,$CF,$A7
	.dc.b	$FE,$9F,$3F,$D5,$3E,$AF,$FA,$CF,$E4,$6F,$AC,$FE,$B3,$CD,$F8,$87
	.dc.b	$FB,$17,$E5,$7F,$D6,$37,$7D,$77,$E3,$3E,$CB,$FB,$67,$D9,$7F,$6B
	.dc.b	$EC,$FE,$25,$A7,$F7,$1F,$98,$FE,$EF,$F6,$7F,$DD,$BE,$A9,$FA,$AF
	.dc.b	$E7,$FE,$E3,$FC,$05,$C1,$FE,$08,$FC,$0D,$A7,$F8,$0B,$C9,$FF,$4E
	.dc.b	$F9,$5F,$E0,$EF,$BB,$FF,$06,$FC,$43,$EE,$DF,$B4,$7D,$BF,$8C,$FF
	.dc.b	$03,$FD,$07,$F8,$67,$F0,$0F,$C2,$FE,$01,$FE,$18,$CF,$FD,$4F,$3F
	.dc.b	$E2,$3F,$D6,$FF,$1B,$F1,$FF,$E2,$2B,$E5,$87,$ED,$FF,$03,$FF,$17
	.dc.b	$7E,$E2,$FC,$99,$F5,$3F,$E2,$FF,$8F,$DD,$FF,$A2,$BF,$07,$FF,$1C
	.dc.b	$78,$5F,$DB,$7F,$34,$CF,$C8,$BE,$2B,$FC,$5F,$F0,$7F,$E1,$2F,$96
	.dc.b	$FF,$20,$71,$FF,$9A,$BE,$DF,$FC,$8F,$FA,$E7,$F7,$A8,$FF,$36,$FE
	.dc.b	$15,$FE,$4E,$EE,$FA,$31,$FE,$54,$FC,$B1,$F8,$67,$F9,$52,$3F,$CD
	.dc.b	$5F,$87,$7F,$96,$BC,$8F,$0E,$BF,$E9,$4F,$C4,$3F,$CC,$1F,$88,$63
	.dc.b	$E2,$FF,$6B,$C1,$FE,$66,$FC,$47,$1F,$21,$FE,$8A,$FC,$4F,$FC,$D9
	.dc.b	$F7,$18,$A7,$C9,$7F,$9C,$F8,$F1,$F1,$FF,$E0,$4F,$C5,$2F,$D5,$FF
	.dc.b	$B8,$38,$BF,$CE,$D3,$FB,$4F,$C5,$77,$E6,$BF,$16,$BF,$A7,$63,$F1
	.dc.b	$9E,$7F,$8D,$FF,$A4,$B1,$E3,$F8,$E7,$F4,$67,$FD,$01,$F8,$F7,$FA
	.dc.b	$71,$FF,$41,$7E,$3F,$FE,$A2,$7F,$D0,$9E,$FB,$E1,$7F,$D5,$11,$FE
	.dc.b	$8D,$FC,$87,$FD,$5F,$F5,$AF,$C8,$7D,$F2,$FC,$8F,$FD,$6E,$FF,$A7
	.dc.b	$3F,$13,$7E,$4B,$FE,$BF,$F3,$3E,$47,$E4,$FF,$EC,$6F,$C9,$FF,$23
	.dc.b	$7F,$EC,$1F,$8F,$FF,$26,$7E,$97,$FE,$73,$FD,$17,$FC,$7F,$3F,$EC
	.dc.b	$5F,$CD,$7F,$15,$F8,$3F,$FB,$67,$F2,$CF,$F6,$C4,$FF,$BD,$3E,$47
	.dc.b	$FD,$C9,$FA,$4B,$E2,$3F,$DC,$9F,$97,$FE,$2B,$F1,$BF,$F7,$6F,$EA
	.dc.b	$0F,$C3,$7F,$DD,$B8,$7E,$3B,$EC,$BF,$DE,$DF,$A6,$3D,$4F,$F7,$B6
	.dc.b	$9F,$F0,$03,$F5,$FF,$EF,$FF,$8E,$7F,$2D,$F8,$EF,$F7,$FF,$81,$F9
	.dc.b	$AF,$DB,$1F,$9B,$B9,$FE,$27,$E6,$EF,$D5,$DE,$F7,$E4,$BF,$44,$FF
	.dc.b	$3A,$FE,$26,$FD,$BF,$FB,$D7,$E7,$8F,$D3,$BF,$DF,$7F,$9E,$F3,$E2
	.dc.b	$61,$FE,$7C,$F8,$67,$E7,$D8,$F9,$07,$E7,$CF,$D3,$BF,$C0,$F8,$70
	.dc.b	$F9,$0F,$F4,$E6,$1F,$67,$F1,$57,$8C,$F9,$4F,$DB,$7D,$3F,$E4,$75
	.dc.b	$ED,$F2,$AF,$D0,$7C,$9F,$A0,$FC,$5F,$CD,$FF,$19,$5A,$F9,$67,$E1
	.dc.b	$EF,$CE,$3F,$15,$FA,$13,$F4,$5F,$E9,$17,$D3,$F4,$57,$87,$FC,$17
	.dc.b	$F4,$57,$E3,$3F,$E3,$8E,$6F,$47,$33,$9B,$F8,$4F,$C0,$3F,$23,$7E
	.dc.b	$8D,$FE,$2D,$D7,$C1,$87,$0B,$DF,$E6,$AF,$CF,$3F,$C1,$DE,$63,$E5
	.dc.b	$3C,$7A,$7C,$4E,$9F,$D0,$7D,$C7,$FB,$07,$C1,$7E,$55,$EC,$D3,$A5
	.dc.b	$6B,$CD,$FE,$AD,$F8,$67,$F9,$D6,$73,$C4,$C7,$FB,$8F,$C3,$BF,$47
	.dc.b	$74,$78,$35,$F6,$67,$E0,$FB,$7F,$F1,$17,$53,$D8,$D7,$2F,$C0,$CF
	.dc.b	$FC,$9B,$FA,$2B,$F4,$87,$AB,$AC,$BD,$3F,$8F,$3E,$B1,$F8,$47,$FA
	.dc.b	$F7,$8F,$CB,$97,$F9,$5A,$5F,$A5,$3F,$4A,$E5,$CB,$E9,$C7,$97,$E6
	.dc.b	$DF,$ED,$FC,$78,$C7,$5F,$B2,$7E,$97,$C3,$3E,$90,$FD,$33,$15,$BF
	.dc.b	$A9,$BA,$FF,$BA,$7F,$4D,$B6,$FD,$3A,$FC,$E6,$BE,$BF,$F7,$5A,$F8
	.dc.b	$30,$7C,$63,$F1,$4E,$7F,$24,$FD,$73,$BF,$C0,$7F,$AB,$BD,$DF,$0F
	.dc.b	$EA,$1F,$EF,$FC,$BC,$DE,$CF,$F9,$CF,$E4,$87,$A7,$85,$FA,$8F,$FB
	.dc.b	$97,$CC,$F8,$7F,$82,$FF,$6C,$E8,$7C,$7F,$F8,$B3,$2F,$BF,$F8,$67
	.dc.b	$EB,$FF,$3D,$FD,$57,$C7,$7D,$3F,$D5,$1F,$AA,$3F,$54,$7A,$2F,$BE
	.dc.b	$7E,$2D,$F6,$7F,$31,$7D,$67,$F5,$FF,$BC,$7E,$07,$FD,$5B,$7F,$C8
	.dc.b	$F1,$69,$7E,$30,$FD,$6D,$F4,$9F,$88,$E1,$FD,$37,$E8,$DF,$E4,$2F
	.dc.b	$D8,$BF,$CE,$5F,$8E,$7E,$33,$E4,$9F,$AB,$FF,$86,$71,$68,$FC,$E1
	.dc.b	$F5,$5F,$97,$FC,$9D,$F4,$EF,$C9,$3F,$23,$F9,$77,$D4,$FC,$63,$EA
	.dc.b	$71,$FE,$E3,$C6,$FE,$11,$FA,$C3,$F1,$6E,$3E,$53,$C8,$FD,$95,$FE
	.dc.b	$FB,$E5,$9F,$AC,$FE,$99,$62,$FD,$5F,$E8,$FE,$23,$FA,$9F,$C4,$DF
	.dc.b	$B4,$7F,$A6,$3E,$11,$FA,$E7,$C2,$FC,$3D,$FA,$E7,$EB,$FF,$15,$F6
	.dc.b	$75,$BF,$0B,$7E,$BB,$F5,$3E,$C7,$FC,$A1,$C5,$FC,$42,$FF,$55,$ED
	.dc.b	$76,$FC,$A7,$C7,$F9,$13,$F5,$EE,$9F,$8B,$3E,$13,$E5,$FD,$8F,$F8
	.dc.b	$DB,$9B,$F9,$BE,$7C,$E9,$FB,$07,$D0,$F6,$FF,$92,$7C,$7B,$F6,$0E
	.dc.b	$7F,$8C,$7F,$81,$BA,$3F,$85,$7C,$77,$F6,$6F,$99,$7C,$DF,$FA,$8F
	.dc.b	$3F,$F3,$CF,$C5,$7E,$0F,$F4,$BF,$E1,$1B,$77,$C9,$F9,$CF,$BF,$DF
	.dc.b	$FA,$1F,$B0,$F6,$FC,$69,$F8,$1F,$F5,$4F,$17,$F9,$DE,$BD,$71,$FF
	.dc.b	$48,$FA,$0F,$C4,$1F,$0D,$F9,$4D,$7B,$64,$FC,$67,$FD,$09,$F9,$23
	.dc.b	$F6,$2E,$D0,$FC,$C1,$F3,$3F,$43,$0E,$33,$FE,$97,$FD,$A5,$FB,$27
	.dc.b	$0F,$C6,$DF,$84,$3F,$6C,$C7,$E7,$0F,$43,$5F,$8E,$3F,$3D,$BF,$A3
	.dc.b	$E3,$F6,$0C,$FB,$FA,$FC,$CB,$7F,$98,$75,$FC,$73,$FB,$A7,$17,$8B
	.dc.b	$7F,$EF,$1F,$BA,$FA,$3F,$20,$7E,$9D,$FC,$7A,$BC,$FF,$B3,$BE,$43
	.dc.b	$F9,$74,$7A,$BE,$A9,$FB,$43,$F0,$9F,$9D,$EC,$FF,$43,$FD,$15,$F9
	.dc.b	$3F,$D1,$FD,$41,$FB,$2B,$F5,$87,$ED,$3F,$57,$F5,$0F,$F1,$4F,$C8
	.dc.b	$BF,$29,$FC,$2F,$C8,$7F,$73,$DD,$F9,$8B,$D1,$FC,$4D,$FF,$45,$F3
	.dc.b	$BE,$5F,$E7,$9F,$96,$8E,$FF,$2F,$FA,$0E,$17,$E4,$5F,$9A,$FD,$CB
	.dc.b	$C7,$F2,$4F,$C7,$FF,$C2,$9D,$BF,$A9,$FC,$11,$FB,$5F,$F8,$7B,$07
	.dc.b	$DA,$FF,$83,$FE,$CF,$F5,$DF,$AD,$BD,$2F,$DA,$FD,$6B,$F6,$CD,$FE
	.dc.b	$9B,$FB,$3F,$DB,$9F,$A6,$7F,$86,$BF,$6E,$FD,$F7,$EC,$AF,$D6,$7F
	.dc.b	$B3,$7D,$7F,$F8,$A7,$F2,$0F,$F4,$BF,$BD,$FE,$D8,$FD,$77,$FB,$37
	.dc.b	$E3,$9F,$33,$E2,$3F,$4B,$F5,$0F,$D8,$7A,$C7,$F1,$EF,$B1,$7E,$85
	.dc.b	$FE,$08,$BE,$13,$FE,$27,$AD,$FD,$97,$F8,$8E,$9F,$BE,$FF,$83,$FD
	.dc.b	$0F,$D8,$DE,$F8,$AB,$8C,$79,$E7,$E1,$FC,$CB,$33,$FB,$89,$CE,$7D
	.dc.b	$B4,$7B,$39,$CB,$AE,$3E,$9C,$BF,$7F,$AF,$F8,$D7,$C3,$E5,$F4,$7E
	.dc.b	$2F,$90,$FE,$33,$F8,$EF,$D4,$E7,$FE,$5D,$B7,$F4,$D3,$FB,$26,$FF
	.dc.b	$C2,$2B,$7E,$12,$B7,$F9,$F7,$F8,$AF,$1E,$9F,$50,$FD,$DB,$D3,$8E
	.dc.b	$4E,$51,$FE,$2B,$7D,$B9,$76,$59,$76,$C7,$FA,$07,$E7,$FE,$89,$E3
	.dc.b	$A7,$9F,$EC,$FF,$C8,$F1,$FE,$6F,$CF,$E6,$FE,$BF,$FE,$69,$FD,$01
	.dc.b	$E6,$FE,$97,$F1,$47,$EE,$8C,$F8,$63,$FC,$93,$0E,$17,$BA,$FF,$23
	.dc.b	$7E,$6F,$E5,$9F,$F3,$A7,$BF,$EC,$8C,$78,$7F,$EA,$BC,$3E,$BC,$63
	.dc.b	$3F,$0F,$BD,$FE,$A3,$FD,$73,$F9,$8C,$E7,$3F,$27,$E3,$DE,$DC,$1F
	.dc.b	$C0,$3F,$C0,$BF,$20,$FF,$52,$7D,$23,$C9,$FD,$4F,$D9,$7C,$BB,$DF
	.dc.b	$C4,$7F,$91,$3F,$41,$5E,$FB,$DF,$B5,$EF,$3E,$98,$7D,$0E,$8F,$C9
	.dc.b	$C7,$F1,$0F,$88,$FA,$FD,$DF,$4E,$3C,$59,$74,$D7,$E7,$CE,$74,$F0
	.dc.b	$69,$E5,$F8,$8F,$6F,$DD,$7E,$7B,$F5,$D7,$EB,$4F,$88,$7E,$89,$FC
	.dc.b	$7B,$C5,$7E,$DE,$E7,$FF,$54,$63,$C7,$0E,$17,$BA,$7C,$2E,$B9,$3F
	.dc.b	$AD,$E7,$DB,$F1,$CF,$A9,$8F,$1D,$2D,$6B,$2E,$7B,$FF,$78,$FD,$27
	.dc.b	$CF,$39,$AF,$2C,$FF,$7C,$F9,$6F,$A9,$CF,$FE,$63,$EB,$FF,$40,$7E
	.dc.b	$EF,$FE,$30,$FC,$FF,$F9,$D7,$37,$E6,$6B,$5F,$97,$2B,$5F,$4D,$F5
	.dc.b	$E7,$FC,$D9,$D9,$F3,$9F,$E0,$67,$5D,$7E,$A1,$FD,$D9,$E7,$5F,$E3
	.dc.b	$EF,$47,$C7,$F9,$07,$FA,$0B,$1E,$DF,$8A,$BF,$6E,$FC,$6E,$1F,$C3
	.dc.b	$BF,$0C,$FF,$36,$FE,$83,$FE,$17,$75,$E6,$BF,$51,$FD,$0E,$BF,$E3
	.dc.b	$3F,$C9,$97,$BC,$57,$FC,$EF,$CA,$FD,$B6,$3F,$E7,$8F,$37,$F4,$59
	.dc.b	$F7,$C7,$8E,$5F,$A2,$8C,$65,$CF,$2F,$2C,$FF,$84,$B0,$F2,$FE,$11
	.dc.b	$FE,$43,$97,$06,$A9,$F3,$9F,$9A,$F6,$DE,$4F,$E2,$3F,$46,$FF,$77
	.dc.b	$7A,$1E,$AA,$D4,$E7,$81,$F2,$D9,$F3,$BF,$CA,$F9,$79,$4F,$F1,$1D
	.dc.b	$FE,$9B,$C7,$F5,$C6,$27,$FC,$BB,$87,$7F,$A0,$F2,$5F,$6B,$66,$FA
	.dc.b	$6F,$46,$5F,$E8,$2F,$81,$78,$AF,$CC,$FF,$C9,$BB,$7E,$90,$FE,$41
	.dc.b	$5B,$9D,$79,$09,$7E,$EF,$F5,$AB,$A7,$07,$A4,$F6,$FE,$46,$FD,$43
	.dc.b	$D3,$18,$CF,$C3,$9F,$F9,$73,$83,$CB,$7F,$EF,$DD,$3B,$EB,$F3,$FF
	.dc.b	$57,$7B,$9F,$23,$F7,$E7,$EE,$EF,$DF,$BF,$D2,$5F,$BB,$7E,$22,$F7
	.dc.b	$C0,$BD,$D7,$FA,$5F,$F6,$9D,$62,$EA,$F2,$E1,$C3,$87,$D4,$F2,$C3
	.dc.b	$8F,$67,$8E,$FC,$5A,$78,$7B,$7F,$09,$F1,$AF,$11,$FB,$8B,$F7,$9F
	.dc.b	$F1,$1F,$EA,$FF,$E9,$4F,$D8,$7F,$3B,$7B,$AD,$79,$EA,$FF,$4C,$FB
	.dc.b	$7A,$C7,$E8,$70,$7C,$DC,$38,$36,$3B,$E5,$DB,$9F,$C7,$18,$C7,$A7
	.dc.b	$AB,$FC,$23,$F3,$B7,$ED,$8F,$DE,$3F,$D7,$1D,$7F,$B7,$FC,$DF,$F8
	.dc.b	$25,$6A,$73,$C5,$FE,$98,$F5,$FF,$86,$4F,$2D,$BF,$82,$9F,$ED,$77
	.dc.b	$FA,$EB,$8F,$3D,$1C,$FF,$2D,$FC,$A5,$F8,$F3,$F1,$47,$E5,$FA,$3F
	.dc.b	$5D,$F8,$BF,$F1,$7F,$BB,$7E,$9D,$FD,$F7,$A7,$F8,$37,$D5,$7F,$06
	.dc.b	$AD,$4E,$7D,$7F,$D0,$79,$7F,$DE,$67,$96,$FF,$95,$7F,$91,$7E,$63
	.dc.b	$F9,$47,$8F,$CB,$32,$FE,$81,$5D,$7D,$D6,$9E,$2A,$D7,$E7,$3F,$C0
	.dc.b	$FF,$29,$7F,$00,$FD,$FF,$D4,$1F,$69,$FE,$B6,$E7,$FF,$42,$EF,$FE
	.dc.b	$0B,$BD,$F8,$0B,$DF,$E4,$9F,$EF,$FF,$DA,$3F,$90,$D6,$B7,$FD,$63
	.dc.b	$DB,$F5,$3F,$D5,$BC,$AF,$BF,$E9,$1F,$E3,$AC,$BB,$6B,$E3,$D7,$FD
	.dc.b	$8F,$E2,$FF,$7B,$FA,$47,$ED,$CF,$41,$FC,$01,$8F,$4B,$FC,$2B,$FA
	.dc.b	$47,$F7,$BB,$DD,$6A,$FF,$61,$F3,$8F,$93,$D4,$73,$E2,$FC,$1D,$F6
	.dc.b	$F5,$FF,$A6,$47,$FA,$FA,$39,$E7,$DF,$0F,$25,$F8,$E3,$C5,$EE,$FF
	.dc.b	$2F,$FD,$BB,$FD,$A9,$FB,$17,$F9,$D3,$CC,$7C,$6B,$7F,$D4,$72,$7F
	.dc.b	$8D,$EF,$7E,$32,$F7,$4F,$CE,$BF,$A6,$D7,$4D,$FF,$8C,$3F,$EE,$1F
	.dc.b	$9B,$FE,$2E,$F2,$BE,$FF,$15,$E3,$FC,$A9,$AF,$8F,$57,$85,$FD,$17
	.dc.b	$E1,$1F,$AE,$BF,$46,$7D,$17,$8F,$F4,$CF,$E0,$9F,$BB,$FF,$5E,$BD
	.dc.b	$D6,$B5,$57,$F0,$9F,$DA,$75,$8F,$3E,$1F,$E5,$D8,$76,$FE,$00,$FE
	.dc.b	$5A,$F2,$CF,$B6,$9E,$3B,$F1,$47,$87,$83,$FC,$79,$E9,$FF,$B6,$7E
	.dc.b	$2D,$F9,$57,$9B,$F0,$FF,$F5,$0E,$9F,$E3,$BA,$D4,$E7,$D4,$7E,$87
	.dc.b	$9F,$C7,$BF,$9F,$5F,$F0,$2B,$E8,$D8,$63,$DB,$3F,$F6,$77,$BF,$E2
	.dc.b	$9C,$E5,$E1,$F8,$8F,$ED,$DE,$EF,$F8,$B3,$D4,$78,$EF,$CF,$1E,$EB
	.dc.b	$F3,$97,$E5,$3F,$E0,$8C,$38,$5E,$EB,$5F,$1D,$9F,$A7,$FF,$3E,$70
	.dc.b	$FE,$9F,$3F,$F5,$0E,$1F,$E3,$AE,$3F,$EA,$EF,$2D,$3B,$F9,$3E,$4C
	.dc.b	$17,$E4,$9F,$1F,$E6,$3F,$E0,$5E,$07,$EF,$EF,$DF,$3F,$AC,$FA,$EF
	.dc.b	$CB,$7F,$CF,$DF,$05,$FE,$1E,$BD,$D6,$A7,$FD,$05,$F1,$D9,$C3,$97
	.dc.b	$07,$F6,$DB,$EB,$F8,$D7,$F6,$C7,$8F,$3B,$3E,$DF,$6D,$E3,$C3,$C5
	.dc.b	$39,$CB,$F9,$AF,$C5,$BE,$6B,$FC,$69,$FA,$1B,$F5,$F7,$E2,$4F,$4B
	.dc.b	$FC,$3B,$FA,$7B,$F3,$86,$3C,$78,$7F,$C5,$18,$F1,$C3,$87,$D1,$B0
	.dc.b	$7E,$3F,$C3,$4E,$FA,$7F,$9F,$1F,$0F,$E8,$37,$2E,$73,$E2,$D7,$FC
	.dc.b	$5B,$F0,$FF,$EC,$FB,$E3,$9F,$6F,$D8,$FF,$D2,$DF,$97,$79,$3E,$91
	.dc.b	$FB,$93,$EA,$9F,$73,$FD,$4B,$DD,$FF,$4B,$5E,$EB,$53,$FD,$5B,$DF
	.dc.b	$CF,$2F,$F1,$BE,$0F,$F7,$1B,$FE,$C3,$BE,$D1,$FD,$81,$E5,$8F,$6F
	.dc.b	$B9,$F1,$D7,$F8,$B7,$1E,$BE,$BF,$FA,$AF,$E4,$7E,$27,$ED,$6F,$C4
	.dc.b	$BF,$B2,$FD,$4B,$EF,$DF,$90,$7F,$AA,$EF,$75,$A9,$7E,$23,$9E,$C6
	.dc.b	$3F,$CB,$B8,$18,$7D,$1F,$17,$0F,$2F,$A1,$69,$FE,$A8,$F3,$7F,$E0
	.dc.b	$0C,$E6,$1F,$7F,$FE,$C1,$FB,$FF,$F7,$47,$B7,$FE,$1C,$FC,$27,$C3
	.dc.b	$F8,$FB,$F1,$EF,$F6,$05,$6B,$EB,$7F,$DD,$DC,$2A,$7E,$E1,$AC,$3F
	.dc.b	$91,$F0,$7F,$82,$30,$FF,$7A,$7E,$43,$E9,$79,$4F,$BB,$F7,$1F,$13
	.dc.b	$FC,$AB,$2E,$BF,$5F,$FE,$A2,$FD,$79,$F9,$57,$E8,$7E,$C3,$FC,$C3
	.dc.b	$FB,$83,$F7,$CF,$F4,$F5,$6B,$D7,$56,$BF,$00,$CE,$1F,$27,$A1,$6F
	.dc.b	$D3,$F0,$F1,$BF,$F3,$6F,$9D,$E1,$BF,$07,$DE,$7F,$1E,$E3,$FF,$71
	.dc.b	$7E,$4C,$F9,$B7,$ED,$7F,$E9,$8C,$7F,$A1,$FE,$45,$E8,$F2,$7F,$85
	.dc.b	$E0,$18,$71,$BE,$1D,$AB,$7E,$6F,$DB,$27,$DC,$FE,$07,$0E,$18,$FF
	.dc.b	$8D,$1E,$5F,$83,$7C,$0C,$7F,$D6,$93,$C7,$2E,$BF,$A6,$7F,$BF,$FF
	.dc.b	$17,$7E,$9D,$FE,$B0,$FC,$BF,$F9,$FF,$EF,$6F,$C8,$9F,$A1,$7F,$BF
	.dc.b	$EF,$7F,$86,$AF,$75,$FE,$2B,$E3,$F6,$E8,$BD,$BF,$E4,$87,$BF,$E0
	.dc.b	$7F,$99,$79,$DF,$83,$ED,$7C,$9A,$78,$FE,$55,$AF,$FA,$D3,$1F,$E2
	.dc.b	$7D,$BF,$25,$7E,$AA,$FD,$53,$F8,$BF,$C0,$BF,$57,$7E,$F7,$FE,$8C
	.dc.b	$FD,$93,$8E,$1F,$47,$0E,$1C,$2C,$38,$5F,$FA,$48,$DD,$76,$C7,$F9
	.dc.b	$1B,$FE,$62,$BF,$07,$C8,$FE,$CD,$E7,$7C,$BE,$43,$EF,$63,$F1,$27
	.dc.b	$39,$F9,$3E,$87,$3A,$FA,$FE,$03,$F6,$FF,$00,$FD,$7F,$FD,$31,$FB
	.dc.b	$D7,$EC,$7E,$8B,$C7,$5A,$9C,$FB,$2F,$C7,$73,$EE,$3F,$1C,$FE,$0D
	.dc.b	$7D,$38,$FF,$CF,$2F,$1F,$A1,$67,$F3,$78,$7E,$CC,$E6,$3F,$13,$F6
	.dc.b	$FF,$D0,$6F,$F8,$5F,$EF,$2F,$C8,$DF,$90,$FF,$A6,$FE,$03,$E7,$FC
	.dc.b	$27,$FB,$27,$E4,$FE,$B6,$EC,$38,$5E,$FB,$5F,$47,$BC,$FF,$49,$7B
	.dc.b	$CF,$EF,$47,$F9,$7F,$1E,$78,$BF,$44,$CC,$FE,$AF,$5F,$36,$1C,$73
	.dc.b	$F2,$78,$19,$C7,$8E,$5E,$8F,$C4,$7F,$8C,$7E,$6D,$FD,$3F,$2F,$DA
	.dc.b	$7E,$DA,$FC,$D7,$FC,$95,$F8,$77,$F6,$2F,$C9,$3F,$8E,$FA,$DF,$D7
	.dc.b	$6F,$75,$AE,$85,$7F,$22,$F9,$BD,$53,$1E,$9F,$7F,$FD,$86,$31,$87
	.dc.b	$FB,$0D,$F5,$E1,$F3,$EB,$79,$78,$7E,$CF,$CD,$5C,$30,$FF,$4B,$DF
	.dc.b	$83,$EB,$BF,$1F,$F8,$2F,$E9,$7F,$1C,$7F,$00,$3F,$2A,$FF,$1D,$FA
	.dc.b	$BF,$C5,$BF,$3C,$7E,$B7,$FD,$FB,$0E,$17,$BF,$53,$F7,$5E,$EA,$DF
	.dc.b	$93,$6F,$7F,$E3,$5A,$76,$FD,$6B,$E9,$C6,$44,$78,$B5,$CE,$ED,$3B
	.dc.b	$C7,$93,$27,$DB,$7A,$B0,$EB,$AF,$8F,$7F,$E5,$BF,$86,$7E,$A2,$7F
	.dc.b	$95,$79,$CF,$D7,$1F,$AD,$3F,$44,$78,$5F,$CB,$FE,$E1,$F1,$BF,$BC
	.dc.b	$C3,$85,$EE,$B5,$F9,$4E,$71,$7E,$65,$FD,$BE,$F7,$9F,$FB,$89,$E3
	.dc.b	$9F,$68,$FE,$F6,$71,$C7,$A2,$F4,$BF,$2D,$EC,$7F,$19,$FC,$C7,$FD
	.dc.b	$6B,$EF,$3F,$10,$79,$3F,$D2,$FE,$C1,$FB,$D3,$D0,$FE,$5B,$87,$0B
	.dc.b	$DF,$C8,$2D,$FA,$E6,$F3,$ED,$F8,$A7,$F1,$DD,$BF,$4F,$97,$28,$FF
	.dc.b	$2C,$C6,$27,$7A,$F8,$F5,$FE,$B9,$F8,$1F,$A2,$BF,$AC,$6B,$E1,$9F
	.dc.b	$EC,$19,$7E,$53,$D9,$7E,$B7,$FC,$33,$D3,$7E,$CD,$F2,$BF,$7C,$FD
	.dc.b	$17,$8B,$FC,$2E,$F7,$EF,$7F,$0B,$AD,$7E,$0D,$FA,$89,$E7,$F1,$FF
	.dc.b	$A7,$77,$B7,$E9,$BD,$9F,$E8,$51,$CB,$3E,$FD,$AF,$C3,$3C,$9B,$78
	.dc.b	$F3,$FB,$7F,$85,$7A,$F2,$7E,$44,$FD,$99,$F8,$E7,$F8,$D7,$F0,$F7
	.dc.b	$E5,$6F,$C2,$7F,$CC,$DD,$BF,$D7,$B0,$E1,$7B,$AD,$7E,$B3,$F4,$72
	.dc.b	$FE,$FD,$5D,$7F,$1A,$FD,$2E,$1C,$73,$EF,$5F,$44,$DB,$E2,$FB,$A6
	.dc.b	$9E,$2F,$95,$F4,$63,$E0,$9C,$E6,$FC,$57,$D9,$6F,$CB,$9F,$BA,$7F
	.dc.b	$88,$BE,$0D,$FA,$E7,$FA,$07,$F1,$3F,$36,$DD,$AF,$7F,$55,$FC,$5E
	.dc.b	$BF,$D0,$DF,$95,$6B,$2E,$9E,$57,$AB,$C9,$FE,$E5,$1D,$F4,$C6,$5F
	.dc.b	$E1,$D8,$C6,$3C,$FD,$4F,$2E,$7D,$B5,$F1,$DF,$EE,$CF,$E9,$9F,$AA
	.dc.b	$FF,$29,$FB,$AF,$F6,$07,$C7,$FF,$B0,$3C,$7F,$F0,$6F,$84,$AD,$7C
	.dc.b	$D7,$F5,$2F,$C1,$BF,$BB,$CE,$7E,$F3,$FC,$CC,$EF,$E2,$6D,$F0,$E7
	.dc.b	$F0,$4E,$FE,$3D,$A7,$FB,$3F,$B3,$C9,$F5,$4F,$DF,$DF,$B4,$3F,$74
	.dc.b	$FE,$3B,$FB,$FF,$F9,$87,$F3,$1F,$F0,$3D,$EE,$B5,$F1,$8A,$FF,$33
	.dc.b	$7E,$71,$AC,$79,$F3,$7F,$40,$ED,$FF,$0E,$63,$DF,$CC,$FF,$3F,$3C
	.dc.b	$F4,$EF,$F8,$0F,$93,$0F,$1C,$78,$BE,$FF,$F3,$9F,$61,$FE,$31,$FC
	.dc.b	$85,$F8,$AF,$F8,$AB,$BD,$F3,$0D,$FF,$03,$7B,$AD,$4B,$8F,$3F,$1C
	.dc.b	$C7,$EE,$F0,$FF,$91,$63,$FC,$6A,$FE,$64,$C4,$78,$3E,$BF,$F2,$5F
	.dc.b	$57,$E4,$AE,$58,$F8,$F4,$FC,$07,$E1,$2F,$C6,$9F,$BE,$7F,$8E,$3F
	.dc.b	$79,$7E,$45,$FE,$3A,$FC,$DF,$FC,$67,$7B,$AD,$4F,$F8,$FB,$F7,$9C
	.dc.b	$E1,$CB,$7F,$98,$7F,$9B,$EF,$B7,$EE,$5F,$EA,$97,$97,$D6,$B3,$EF
	.dc.b	$E6,$F9,$30,$F1,$CE,$72,$7D,$4F,$F8,$2F,$2F,$F3,$D7,$E6,$4C,$FF
	.dc.b	$A6,$F8,$9F,$EB,$FF,$AF,$4B,$F3,$AF,$F0,$F5,$E2,$F7,$F8,$A7,$FA
	.dc.b	$A6,$9F,$A1,$E9,$E9,$F3,$BF,$C2,$3E,$2B,$FC,$C1,$1D,$FE,$77,$18
	.dc.b	$F8,$B1,$E7,$9B,$ED,$FC,$D3,$DF,$1F,$CE,$E7,$FE,$B7,$F8,$AF,$D8
	.dc.b	$7A,$7E,$6F,$DA,$1F,$99,$3F,$09,$76,$7F,$8F,$FF,$14,$FF,$31,$DE
	.dc.b	$EB,$53,$FE,$43,$F0,$33,$C2,$C7,$FD,$AD,$7F,$0B,$C7,$B3,$9C,$38
	.dc.b	$F9,$9E,$28,$C6,$7E,$0F,$AE,$FE,$09,$F9,$AF,$F5,$DF,$D0,$3F,$55
	.dc.b	$ED,$7F,$B7,$3E,$55,$F9,$9F,$F9,$DA,$B5,$EA,$FF,$87,$3E,$DD,$5F
	.dc.b	$C8,$FE,$03,$FD,$2B,$5C,$FA,$3F,$0A,$1F,$ED,$5F,$B9,$FF,$BB,$5E
	.dc.b	$79,$7F,$BA,$B9,$BC,$7F,$0E,$D7,$C7,$9F,$F8,$03,$E3,$FF,$D4,$3F
	.dc.b	$0E,$F0,$3F,$DC,$9D,$EF,$C7,$5F,$C0,$37,$F5,$CF,$E5,$95,$A9,$C8
	.dc.b	$FC,$AF,$37,$FB,$0C,$3A,$7E,$7F,$6F,$86,$3D,$78,$B7,$7D,$E3,$FC
	.dc.b	$21,$1F,$E9,$AF,$B8,$FF,$53,$E8,$FC,$95,$F8,$D7,$F2,$BF,$AD,$FE
	.dc.b	$8B,$CF,$FE,$BA,$BD,$D6,$BE,$2F,$FC,$5B,$F6,$5C,$FF,$08,$FF,$0A
	.dc.b	$67,$F2,$AF,$FD,$67,$ED,$7F,$8F,$BE,$1F,$FD,$6D,$1C,$B4,$FE,$71
	.dc.b	$F8,$4F,$F3,$F8,$C6,$5F,$EF,$2C,$3F,$A9,$F5,$BF,$70,$79,$AF,$C8
	.dc.b	$FF,$82,$FE,$C6,$FD,$AD,$F9,$97,$0B,$DD,$6A,$7F,$C6,$1C,$BF,$ED
	.dc.b	$4C,$7A,$7E,$67,$FC,$69,$ED,$95,$C3,$8F,$C1,$33,$F2,$61,$C6,$73
	.dc.b	$0D,$FF,$C1,$5F,$92,$7E,$F5,$FF,$6B,$FD,$77,$FB,$A7,$F6,$CD,$61
	.dc.b	$FE,$E4,$AD,$7C,$47,$F8,$DB,$E3,$14,$FD,$83,$FB,$9D,$73,$E3,$FE
	.dc.b	$58,$FF,$09,$8F,$E6,$2F,$2C,$7F,$DE,$DF,$69,$E2,$F8,$0F,$26,$5D
	.dc.b	$7F,$0A,$F9,$DD,$0F,$C7,$1F,$94,$FB,$7E,$35,$FA,$5B,$D9,$FC,$35
	.dc.b	$EE,$B5,$2F,$D2,$B3,$87,$DD,$E0,$0B,$9F,$90,$E6,$31,$1C,$F9,$BF
	.dc.b	$A9,$F6,$F9,$2F,$C7,$1F,$D5,$BD,$3F,$E2,$FE,$3B,$C3,$7C,$33,$F2
	.dc.b	$AF,$F1,$B7,$EE,$FF,$DC,$FC,$BF,$E5,$77,$BA,$D7,$D7,$2B,$FB,$C7
	.dc.b	$8B,$AC,$7A,$6F,$FE,$62,$BF,$4A,$C3,$93,$93,$FD,$EB,$1D,$2F,$F9
	.dc.b	$24,$73,$8C,$7D,$9F,$F8,$67,$F1,$77,$E5,$1F,$E9,$3F,$CE,$7F,$C2
	.dc.b	$DF,$5E,$F0,$DF,$A8,$77,$D7,$C3,$87,$0F,$C2,$18,$70,$B7,$E0,$5B
	.dc.b	$D7,$E7,$DE,$DF,$E3,$2F,$83,$E0,$3E,$83,$CE,$FC,$3F,$8F,$F9,$75
	.dc.b	$F2,$7A,$2D,$9F,$89,$FF,$BA,$7F,$47,$F8,$7F,$5D,$FC,$5F,$E5,$DF
	.dc.b	$8E,$7F,$AA,$BF,$3D,$7D,$6F,$EB,$2F,$7F,$09,$FE,$28,$AD,$4F,$F5
	.dc.b	$FF,$49,$5A,$FC,$D7,$FC,$C2,$F0,$C6,$F1,$FE,$2B,$3F,$06,$3C,$A3
	.dc.b	$FD,$41,$5B,$8F,$17,$D2,$7F,$9E,$7F,$50,$7E,$E5,$FE,$33,$FB,$6F
	.dc.b	$F3,$07,$E0,$AF,$C8,$7F,$B1,$6D,$F8,$0C,$D8,$70,$FA,$EF,$F7,$A7
	.dc.b	$C1,$B0,$E1,$FB,$EF,$F9,$92,$BA,$FB,$9F,$E2,$07,$86,$1F,$E1,$F7
	.dc.b	$96,$7E,$0E,$6F,$26,$BE,$3C,$BA,$7C,$B7,$FA,$3F,$C1,$FF,$6A,$FC
	.dc.b	$AB,$F3,$37,$EA,$EE,$67,$E5,$8F,$CB,$FF,$CE,$B8,$70,$BD,$D6,$BF
	.dc.b	$75,$CF,$A0,$FC,$8B,$FC,$6B,$7D,$B8,$BF,$C6,$8F,$1C,$FB,$6B,$E3
	.dc.b	$BF,$F2,$A6,$5D,$38,$3F,$CE,$5F,$1B,$FA,$9F,$2B,$FD,$47,$F8,$7F
	.dc.b	$F0,$9F,$9B,$7E,$E5,$F9,$3D,$57,$BF,$C2,$D7,$BF,$BE,$D4,$FE,$CF
	.dc.b	$C2,$FE,$08,$F5,$BF,$EE,$2F,$1A,$FE,$2D,$F0,$5E,$2F,$57,$F5,$18
	.dc.b	$F4,$F7,$1F,$A1,$7F,$10,$F6,$1F,$90,$7E,$EB,$F5,$4F,$F4,$EF,$E0
	.dc.b	$8F,$C9,$5F,$B1,$A3,$F7,$DF,$96,$7E,$FE,$F8,$7E,$19,$9B,$DD,$7F
	.dc.b	$03,$F8,$9D,$71,$7F,$1E,$CB,$AB,$FC,$20,$F1,$61,$CB,$DA,$C4,$60
	.dc.b	$E1,$8F,$5C,$FC,$5F,$27,$99,$61,$E8,$CF,$C3,$E1,$FF,$B5,$7E,$9F
	.dc.b	$FC,$A9,$F2,$BF,$77,$F4,$1F,$F6,$B7,$E7,$9F,$0B,$D2,$FC,$E5,$6B
	.dc.b	$C6,$56,$BF,$60,$FF,$5E,$CF,$2D,$3F,$CE,$CF,$F9,$7F,$E1,$3F,$CF
	.dc.b	$6F,$1A,$FF,$71,$7E,$09,$E1,$FB,$FF,$EE,$58,$F4,$F8,$A7,$CA,$7F
	.dc.b	$BA,$7F,$2F,$7E,$60,$FD,$2D,$FB,$2B,$F5,$4F,$F1,$67,$EC,$EF,$DA
	.dc.b	$3F,$D9,$F8,$70,$BD,$D6,$BF,$01,$CE,$3C,$F8,$BF,$D1,$97,$DB,$F3
	.dc.b	$EF,$F4,$8B,$C7,$3E,$DF,$1E,$FD,$8F,$C9,$87,$8E,$73,$F8,$A7,$E1
	.dc.b	$FE,$57,$F9,$FF,$EA,$2E,$97,$E8,$CF,$B1,$FF,$66,$7C,$33,$F1,$DF
	.dc.b	$EF,$DE,$67,$F5,$2E,$1F,$EB,$F8,$F1,$C3,$87,$C9,$B0,$E1,$F5,$3F
	.dc.b	$D2,$F4,$EF,$A7,$6C,$7F,$0B,$97,$35,$97,$8F,$F7,$1F,$C4,$C7,$13
	.dc.b	$BD,$79,$B5,$F2,$FB,$FE,$B9,$CE,$BE,$2D,$3F,$BD,$47,$13,$FA,$9F
	.dc.b	$E9,$1F,$E8,$CF,$DA,$1F,$AE,$3E,$15,$FB,$04,$DF,$C9,$F3,$4C,$38
	.dc.b	$5E,$EB,$F6,$35,$A9,$FF,$4B,$E3,$EC,$E1,$FE,$D7,$87,$1F,$AB,$F5
	.dc.b	$DE,$FB,$5A,$78,$B4,$FD,$DE,$3E,$7C,$3C,$15,$AD,$3C,$DE,$E7,$EF
	.dc.b	$FE,$7F,$FB,$BF,$D0,$7F,$89,$FE,$8D,$F9,$17,$D9,$FD,$15,$FB,$4B
	.dc.b	$F2,$4F,$E2,$D7,$BA,$D4,$BF,$74,$CF,$6B,$F0,$CF,$F0,$C5,$F5,$E0
	.dc.b	$FF,$42,$BC,$7A,$D9,$FF,$95,$F9,$3F,$A6,$4E,$61,$AF,$FA,$C3,$B3
	.dc.b	$F9,$1F,$CC,$BF,$75,$7E,$41,$FE,$CF,$FD,$31,$FB,$DF,$FB,$12,$F7
	.dc.b	$5A,$F5,$14,$FC,$AF,$5F,$19,$FE,$48,$E0,$FF,$0B,$1F,$9B,$7F,$66
	.dc.b	$E3,$0E,$79,$7F,$B7,$74,$F2,$57,$F9,$43,$2E,$DE,$F7,$CD,$C3,$FC
	.dc.b	$77,$FB,$90,$7E,$77,$FE,$3D,$F9,$97,$E7,$CF,$D2,$1F,$92,$7F,$43
	.dc.b	$BE,$17,$BA,$D7,$E6,$99,$97,$EA,$9F,$E4,$57,$79,$7F,$AC,$9E,$3F
	.dc.b	$70,$CF,$BF,$D7,$79,$6B,$C9,$5E,$4A,$D6,$2F,$53,$FC,$A7,$F0,$DF
	.dc.b	$D1,$2F,$FA,$37,$CB,$FE,$F7,$E3,$DE,$E7,$F1,$7F,$D4,$7F,$B8,$5E
	.dc.b	$EF,$FC,$85,$7B,$AD,$7E,$44,$AD,$78,$3F,$D8,$3F,$0D,$FD,$BC,$62
	.dc.b	$FF,$85,$61,$FD,$0A,$31,$E1,$B3,$F0,$FC,$13,$C4,$F3,$FC,$9F,$F9
	.dc.b	$4B,$3F,$ED,$77,$BD,$7F,$96,$7E,$2B,$F0,$7C,$0F,$77,$B5,$CF,$F5
	.dc.b	$47,$C3,$BF,$19,$7D,$0F,$F7,$5B,$DE,$EB,$DD,$7F,$9F,$F4,$FF,$12
	.dc.b	$57,$C9,$FC,$9B,$FC,$F0,$F6,$E2,$FF,$72,$BC,$AF,$BF,$E1,$DE,$3E
	.dc.b	$86,$BE,$3D,$7F,$D4,$98,$7C,$7F,$53,$FC,$DB,$D4,$FD,$89,$F9,$BB
	.dc.b	$F6,$1F,$F7,$55,$3E,$D1,$F9,$37,$1C,$7F,$D1,$B8,$F1,$C3,$87,$E3
	.dc.b	$5F,$E8,$5B,$FE,$37,$D9,$FE,$3D,$AD,$5F,$D1,$C7,$C3,$9B,$F4,$BF
	.dc.b	$AF,$18,$CB,$9A,$FC,$33,$F1,$2E,$75,$F2,$FC,$A7,$E4,$78,$DE,$4B
	.dc.b	$F4,$97,$E4,$1F,$D8,$3F,$3A,$FE,$27,$F8,$27,$F8,$96,$F7,$5A,$F9
	.dc.b	$25,$7F,$18,$FD,$9B,$58,$F3,$E1,$FE,$8B,$AF,$FA,$53,$1E,$FF,$51
	.dc.b	$FE,$AB,$79,$E9,$DF,$D4,$F2,$61,$E3,$8F,$17,$AF,$EA,$C7,$FD,$6B
	.dc.b	$DA,$FD,$61,$F9,$1B,$EA,$5F,$36,$FD,$7B,$FB,$C7,$C8,$FD,$1C,$58
	.dc.b	$F1,$C3,$85,$EF,$E7,$7F,$C6,$3D,$DF,$E6,$DA,$ED,$F2,$BF,$3B,$0F
	.dc.b	$0E,$9E,$07,$EA,$18,$C7,$9E,$6F,$A1,$F4,$4F,$82,$BC,$FA,$F9,$AB
	.dc.b	$FC,$23,$3F,$E5,$1F,$B1,$7D,$93,$F4,$6F,$D1,$F8,$AF,$DE,$DF,$8E
	.dc.b	$3F,$61,$E1,$7B,$F2,$97,$BA,$FF,$5C,$F9,$FF,$E8,$7C,$BA,$E5,$FE
	.dc.b	$DB,$7B,$FF,$00,$DC,$79,$61,$DF,$F4,$8F,$E2,$D5,$C3,$2E,$BA,$3F
	.dc.b	$14,$FE,$2B,$D3,$FE,$D0,$FA,$C7,$E5,$4F,$A9,$FC,$6F,$D8,$BF,$53
	.dc.b	$F8,$9F,$AB,$FF,$11,$C3,$85,$EF,$C2,$FF,$2F,$57,$C9,$FB,$6D,$73
	.dc.b	$FF,$BD,$33,$EB,$9F,$F1,$6C,$78,$FC,$1D,$AD,$65,$CF,$4F,$07,$67
	.dc.b	$9A,$73,$87,$93,$3F,$A3,$CF,$FE,$AB,$FA,$4F,$F4,$1F,$0F,$FB,$63
	.dc.b	$F0,$EF,$F2,$1F,$9A,$FB,$CF,$F4,$35,$6B,$F1,$35,$6B,$F5,$DC,$FE
	.dc.b	$01,$F5,$F5,$FD,$08,$17,$FE,$90,$FD,$83,$FC,$17,$E8,$7F,$BE,$23
	.dc.b	$A6,$1F,$E5,$2F,$B3,$FE,$8D,$18,$FB,$3F,$F5,$26,$1F,$9B,$F4,$7F
	.dc.b	$C2,$3F,$C0,$2F,$D9,$3F,$D6,$5F,$0A,$FD,$85,$FA,$8E,$FE,$C1,$7B
	.dc.b	$FC,$21,$7B,$FD,$6B,$53,$F9,$FC,$FF,$97,$5B,$F4,$FC,$69,$E0,$D3
	.dc.b	$BF,$E3,$3F,$9E,$8C,$67,$D3,$C9,$FA,$18,$BF,$3E,$F8,$FF,$AE,$BF
	.dc.b	$7A,$FF,$2B,$FE,$3A,$FE,$00,$7E,$49,$F8,$0C,$38,$74,$30,$E1,$F7
	.dc.b	$AC,$38,$79,$BB,$CF,$E9,$FE,$CF,$F8,$3C,$78,$E1,$67,$C3,$A7,$83
	.dc.b	$E4,$FC,$B1,$8C,$BA,$76,$7F,$75,$FA,$1F,$CD,$F8,$DF,$C9,$5F,$F1
	.dc.b	$07,$82,$FD,$E3,$FD,$91,$1A,$F8,$A3,$AF,$CC,$73,$E7,$7F,$4A,$DF
	.dc.b	$FD,$5D,$8F,$FB,$E6,$BA,$C7,$E1,$8E,$32,$9F,$07,$D5,$FE,$5B,$C5
	.dc.b	$FC,$55,$6A,$73,$F2,$1F,$E0,$BF,$AC,$F9,$3F,$B4,$FF,$86,$F8,$5F
	.dc.b	$89,$BF,$33,$7E,$91,$FE,$DF,$BD,$FD,$87,$D5,$AD,$7E,$DF,$9E,$E6
	.dc.b	$5F,$96,$D7,$F3,$58,$7D,$6F,$D8,$B1,$1F,$EE,$6A,$DE,$3C,$FE,$A3
	.dc.b	$C9,$18,$D7,$C7,$F4,$BF,$84,$FA,$57,$17,$FA,$3F,$F2,$0F,$E7,$3F
	.dc.b	$88,$BF,$26,$7E,$B3,$F8,$3B,$DF,$C9,$2F,$74,$FD,$E3,$58,$FF,$A6
	.dc.b	$77,$FF,$25,$E1,$DB,$F4,$5F,$C2,$3C,$B0,$EF,$1E,$3A,$F1,$7C,$C3
	.dc.b	$47,$9F,$FE,$37,$FD,$0F,$F7,$FF,$32,$FD,$83,$FD,$93,$FB,$F3,$F1
	.dc.b	$0F,$F3,$97,$81,$EC,$DF,$F8,$06,$1C,$2F,$74,$FD,$7B,$EF,$6B,$F7
	.dc.b	$2B,$F8,$5F,$E7,$7F,$31,$EF,$87,$E1,$5E,$5F,$4C,$CF,$FA,$74,$F2
	.dc.b	$3B,$FA,$7F,$ED,$7C,$FC,$BE,$17,$F9,$3F,$F7,$0F,$F1,$47,$D3,$FF
	.dc.b	$05,$F8,$4F,$BB,$F4,$9E,$3E,$1F,$46,$4B,$DF,$C1,$AF,$7F,$84,$F7
	.dc.b	$AD,$7B,$BF,$E5,$D7,$F3,$57,$D9,$B0,$CB,$C1,$E3,$FF,$A3,$F8,$FF
	.dc.b	$37,$7D,$B1,$FC,$7F,$E3,$DF,$E6,$0B,$FC,$FF,$E1,$4F,$CF,$DF,$9E
	.dc.b	$BF,$2D,$FF,$61,$7D,$03,$F5,$8F,$F2,$E6,$1C,$2F,$75,$AF,$43,$3C
	.dc.b	$1F,$CC,$F2,$E9,$FA,$5F,$F9,$83,$0E,$19,$DA,$D6,$3C,$BF,$15,$FF
	.dc.b	$5F,$FE,$53,$E5,$9C,$D7,$8E,$7F,$D8,$77,$FC,$83,$E5,$DB,$BE,$65
	.dc.b	$F9,$93,$F3,$EF,$CE,$FE,$42,$FD,$7F,$F9,$06,$1C,$2F,$75,$AE,$0C
	.dc.b	$E2,$F9,$8F,$E3,$D7,$BD,$BF,$D5,$2F,$1C,$FB,$7A,$5F,$E8,$B9,$E3
	.dc.b	$97,$48,$7E,$13,$FE,$16,$FC,$79,$FA,$7B,$F3,$07,$BB,$FE,$7D,$FC
	.dc.b	$B5,$F9,$43,$F4,$6F,$F1,$66,$1C,$2F,$7E,$F2,$F7,$F0,$7A,$C9,$F8
	.dc.b	$6F,$F4,$DC,$38,$6F,$FE,$2A,$3F,$DF,$33,$E1,$D3,$C0,$F9,$4C,$63
	.dc.b	$D7,$F4,$EF,$F1,$6F,$C4,$7F,$88,$FE,$9D,$FA,$9B,$F2,$9F,$F5,$87
	.dc.b	$CE,$FF,$97,$7E,$EF,$F9,$87,$DB,$2F,$75,$AF,$BE,$FF,$23,$63,$FE
	.dc.b	$2D,$FC,$3B,$FB,$13,$D7,$7F,$33,$DF,$3E,$DF,$89,$FF,$74,$CF,$AE
	.dc.b	$5D,$21,$F2,$DF,$EB,$FF,$C8,$1F,$A8,$BE,$2D,$E1,$BF,$3B,$7E,$76
	.dc.b	$FD,$9F,$FC,$1B,$ED,$FE,$D3,$1E,$38,$70,$BD,$D7,$1A,$D6,$2F,$8A
	.dc.b	$FE,$6B,$87,$0E,$1F,$E9,$CF,$2C,$FC,$19,$BF,$3A,$F3,$61,$E5,$9C
	.dc.b	$CF,$F8,$67,$17,$DA,$FF,$52,$F3,$1F,$B6,$3E,$9F,$FC,$3D,$F9,$6F
	.dc.b	$F9,$CF,$CC,$7E,$7D,$FD,$6F,$0E,$17,$BD,$D7,$BF,$C3,$75,$1F,$E2
	.dc.b	$7C,$BA,$E5,$FD,$82,$3F,$D9,$98,$72,$FC,$0F,$11,$E5,$8E,$B3,$E2
	.dc.b	$FC,$17,$F8,$84,$F4,$3F,$5D,$97,$87,$2F,$F0,$A4,$7F,$BC,$FF,$07
	.dc.b	$7E,$33,$FA,$4F,$DC,$9F,$9E,$7F,$4D,$FD,$A1,$FB,$87,$F2,$1B,$DF
	.dc.b	$A6,$BD,$FC,$E2,$FF,$D4,$BC,$BF,$CB,$2F,$79,$FF,$7F,$7B,$FA,$3F
	.dc.b	$DA,$9E,$78,$72,$E7,$F2,$6B,$E3,$D7,$FA,$F7,$57,$BF,$1F,$F2,$67
	.dc.b	$D6,$FF,$94,$BE,$E5,$F9,$4B,$EE,$5F,$77,$FE,$89,$FD,$CB,$E1,$5E
	.dc.b	$EB,$53,$FE,$8D,$F8,$6C,$FC,$B7,$F9,$23,$2F,$9B,$C5,$F8,$07,$BC
	.dc.b	$75,$73,$D7,$FE,$14,$C7,$A6,$7E,$0F,$4F,$CD,$5F,$E3,$FB,$FF,$29
	.dc.b	$E7,$E0,$FC,$DB,$FD,$4B,$F1,$8F,$C4,$5F,$85,$BF,$58,$7D,$A3,$F7
	.dc.b	$B7,$D8,$FF,$90,$7D,$4F,$EA,$F8,$F1,$C3,$85,$EE,$9F,$A5,$7F,$B0
	.dc.b	$6B,$B7,$E2,$DF,$E5,$D7,$86,$FF,$C4,$5E,$59,$F8,$33,$FD,$E7,$CC
	.dc.b	$7F,$8C,$27,$39,$78,$FD,$5F,$D2,$4F,$F8,$D3,$F1,$07,$ED,$EF,$63
	.dc.b	$FD,$67,$FA,$D7,$FB,$9B,$F6,$47,$E5,$3F,$DA,$6F,$75,$AF,$C6,$D4
	.dc.b	$FC,$CB,$58,$FF,$AA,$F8,$3F,$C8,$D8,$70,$6C,$77,$86,$5F,$EF,$4F
	.dc.b	$A8,$F2,$D7,$FA,$46,$7B,$9A,$C5,$F1,$9F,$5B,$F0,$D7,$EA,$AF,$CF
	.dc.b	$3F,$DD,$DF,$3B,$FE,$5F,$F9,$57,$EF,$3F,$AB,$BD,$D6,$BF,$02,$FE
	.dc.b	$BD,$E1,$E7,$1E,$5A,$FE,$18,$FF,$51,$5F,$6E,$9F,$DC,$C6,$F0,$E5
	.dc.b	$E7,$F8,$EF,$C5,$F8,$EB,$5F,$F3,$87,$AD,$EC,$C1,$F9,$73,$F4,$7F
	.dc.b	$5F,$98,$FC,$53,$FC,$0F,$EF,$3F,$70,$FF,$0C,$5E,$F9,$7F,$CD,$D5
	.dc.b	$AF,$95,$CE,$3F,$D7,$72,$FF,$3A,$5E,$DB,$1D,$AB,$AC,$70,$BE,$7F
	.dc.b	$49,$FE,$E5,$8F,$E2,$1F,$A8,$7F,$88,$7E,$43,$FD,$63,$FA,$E7,$FB
	.dc.b	$EB,$E4,$9F,$BF,$EB,$E9,$BF,$AC,$FE,$2A,$AD,$7E,$25,$9C,$3E,$3F
	.dc.b	$1F,$EC,$F3,$67,$FE,$2D,$FC,$C3,$3E,$D7,$FA,$1A,$3C,$19,$73,$FA
	.dc.b	$1F,$E5,$37,$FE,$D3,$9C,$E6,$F8,$CF,$E4,$FE,$B7,$5F,$DB,$3F,$CE
	.dc.b	$BD,$1F,$E6,$FE,$7F,$F3,$07,$9F,$FF,$00,$6B,$4F,$CD,$FC,$0B,$E9
	.dc.b	$FD,$CF,$F9,$F3,$F4,$8F,$F5,$D6,$3C,$B5,$FE,$6F,$87,$0E,$CF,$F5
	.dc.b	$B4,$F6,$C3,$ED,$FE,$7F,$FE,$9E,$AE,$BA,$78,$72,$F2,$F1,$7F,$A8
	.dc.b	$B0,$FE,$6F,$E8,$BE,$FB,$FB,$4F,$BB,$FE,$E0,$FC,$5F,$C3,$F1,$5F
	.dc.b	$AC,$9C,$FC,$47,$FA,$37,$4F,$74,$FE,$C3,$83,$CB,$9F,$58,$7E,$9F
	.dc.b	$E6,$79,$79,$5F,$C5,$FF,$24,$FF,$6E,$ED,$E2,$CF,$C3,$E3,$E3,$F5
	.dc.b	$5F,$EF,$BF,$86,$3F,$3F,$7E,$DA,$FC,$21,$F8,$AF,$C6,$BD,$D6,$A5
	.dc.b	$F8,$7E,$7E,$77,$E3,$D7,$F8,$23,$8B,$CE,$06,$A8,$E7,$97,$FA,$1B
	.dc.b	$8F,$EE,$46,$3D,$DE,$10,$FA,$9F,$F6,$6F,$EB,$FF,$D8,$BE,$0F,$F9
	.dc.b	$97,$C0,$3E,$3F,$FB,$BF,$B0,$C3,$86,$BE,$1C,$38,$7C,$4B,$0E,$1F
	.dc.b	$3A,$FD,$93,$86,$9D,$BA,$FF,$8E,$63,$FD,$F7,$2E,$6B,$2F,$13,$F8
	.dc.b	$8A,$E1,$AF,$93,$57,$E8,$BE,$A9,$CE,$BE,$19,$7E,$9F,$9C,$3E,$37
	.dc.b	$E0,$DF,$E3,$1E,$57,$EB,$2F,$B4,$7E,$D2,$FC,$2B,$FC,$33,$C1,$E4
	.dc.b	$C3,$85,$EE,$B5,$FB,$47,$B3,$2E,$BF,$65,$FD,$09,$F4,$18,$CF,$F8
	.dc.b	$FE,$3C,$E7,$FD,$01,$87,$3F,$C4,$BC,$D6,$CB,$BE,$9E,$4F,$0B,$3F
	.dc.b	$20,$D1,$FA,$0F,$F9,$1F,$F7,$2F,$F2,$6F,$D1,$BF,$7A,$F2,$67,$FD
	.dc.b	$52,$F7,$F6,$AB,$DF,$BD,$AC,$BA,$65,$E6,$E8,$FC,$66,$3C,$75,$C3
	.dc.b	$F8,$FA,$DE,$3C,$FE,$CF,$FC,$95,$3C,$F3,$EF,$9F,$F9,$83,$F0,$5F
	.dc.b	$F2,$57,$D9,$F4,$FD,$C1,$FA,$7B,$F2,$57,$EF,$CF,$CE,$1F,$B1,$BF
	.dc.b	$02,$FD,$4F,$C5,$7E,$23,$1E,$38,$70,$BD,$D7,$F4,$7A,$D6,$3D,$76
	.dc.b	$FF,$0D,$E7,$E1,$87,$E5,$3F,$CB,$23,$18,$F3,$CD,$F5,$5E,$8C,$3F
	.dc.b	$5D,$A7,$83,$5F,$35,$7F,$9F,$A5,$9F,$E7,$3F,$42,$7E,$9E,$FD,$6D
	.dc.b	$F9,$2F,$E3,$FF,$62,$FD,$37,$EA,$DC,$37,$F9,$F8,$70,$BD,$F5,$2F
	.dc.b	$F5,$9F,$BE,$6F,$4F,$E0,$17,$B7,$DA,$63,$3F,$F7,$D3,$81,$1D,$73
	.dc.b	$F1,$79,$BE,$D9,$CD,$79,$E7,$CD,$E5,$E7,$F1,$17,$E0,$BF,$D6,$3E
	.dc.b	$21,$FB,$CB,$EA,$BF,$B8,$7C,$BF,$97,$3F,$F6,$FD,$EE,$B5,$F3,$FF
	.dc.b	$EB,$6E,$5F,$E0,$39,$74,$F9,$CF,$A5,$87,$F1,$B8,$EF,$F3,$78,$8E
	.dc.b	$73,$FE,$26,$C3,$C1,$F5,$DE,$5C,$BB,$69,$F6,$2F,$FD,$01,$EA,$3F
	.dc.b	$51,$FF,$26,$7E,$04,$FE,$01,$7C,$8B,$EA,$BF,$DB,$FC,$1F,$DB,$70
	.dc.b	$E1,$EE,$7F,$A0,$2F,$74,$FD,$FF,$AE,$4F,$D1,$3F,$B8,$E9,$FC,$CF
	.dc.b	$1E,$2F,$F8,$84,$FE,$1B,$3E,$2C,$3C,$3D,$DF,$E6,$6A,$EF,$AF,$97
	.dc.b	$5F,$F0,$16,$1E,$AF,$D8,$DF,$87,$BB,$DE,$6B,$F7,$3F,$F7,$87,$EC
	.dc.b	$FB,$AD,$4E,$7F,$07,$7E,$69,$9C,$BC,$9A,$B3,$7E,$A3,$EA,$9F,$E4
	.dc.b	$71,$E0,$D3,$BE,$8F,$AF,$F4,$4E,$73,$F1,$7C,$CF,$FA,$03,$F5,$4F
	.dc.b	$F7,$0E,$DF,$CA,$FF,$5F,$79,$9F,$E7,$FF,$DA,$3F,$D6,$95,$AF,$89
	.dc.b	$56,$BE,$81,$FB,$66,$BC,$2F,$2F,$0F,$CB,$C3,$AB,$63,$E2,$E1,$CB
	.dc.b	$2F,$F4,$64,$F1,$CF,$C5,$0F,$CB,$3F,$C9,$5F,$B5,$7F,$67,$FC,$FB
	.dc.b	$F0,$9F,$7C,$FD,$E5,$F9,$C3,$F5,$1F,$DE,$FD,$7F,$F2,$4C,$38,$78
	.dc.b	$4C,$38,$5B,$F1,$8D,$ED,$FD,$D6,$FB,$3F,$EA,$CC,$39,$7C,$F3,$F4
	.dc.b	$BF,$F1,$EC,$63,$5F,$1E,$BF,$EF,$BB,$E6,$30,$F0,$E7,$E7,$F9,$2F
	.dc.b	$5F,$EF,$EF,$D8,$5F,$9C,$FF,$5C,$FC,$CF,$FC,$21,$F9,$27,$2C,$57
	.dc.b	$BA,$D7,$D4,$3F,$1A,$D4,$F3,$D3,$D0,$7A,$D6,$1E,$88,$C6,$5E,$0F
	.dc.b	$1D,$9B,$F4,$CF,$2C,$E6,$1F,$4F,$F4,$BF,$77,$7B,$D3,$AF,$FA,$F7
	.dc.b	$E1,$1E,$3F,$61,$CF,$F4,$6F,$77,$F6,$FE,$4F,$FD,$8D,$F3,$FE,$D9
	.dc.b	$FE,$2F,$BF,$A0,$F9,$0F,$67,$1C,$A8,$E7,$9F,$F6,$DF,$C2,$7C,$98
	.dc.b	$F8,$F1,$FE,$19,$9F,$8B,$1F,$F1,$85,$BF,$5A,$7C,$67,$F3,$FF,$CD
	.dc.b	$1F,$6E,$FD,$27,$F8,$7E,$3C,$71,$FC,$96,$1C,$3E,$3D,$87,$0B,$FC
	.dc.b	$8D,$EF,$2E,$DF,$9C,$7C,$FC,$78,$E7,$67,$C4,$F8,$73,$FE,$0B,$3F
	.dc.b	$88,$9C,$E5,$DA,$7F,$BB,$E1,$D7,$EB,$7F,$C0,$9F,$8F,$38,$3F,$39
	.dc.b	$FB,$E7,$FB,$D2,$B5,$F3,$AA,$D7,$86,$A7,$ED,$1D,$F1,$FF,$07,$6B
	.dc.b	$F4,$75,$43,$F4,$9F,$43,$E0,$D3,$F5,$5F,$4F,$FE,$BD,$AD,$E9,$DE
	.dc.b	$7F,$5D,$EC,$E7,$F8,$07,$F9,$0F,$E2,$1F,$43,$F3,$77,$CE,$FF,$89
	.dc.b	$78,$BE,$45,$6B,$95,$5A,$FC,$AB,$E2,$4B,$F3,$5F,$94,$FC,$8E,$1F
	.dc.b	$F1,$09,$AC,$FF,$DA,$58,$72,$F9,$CF,$14,$63,$1F,$F2,$9D,$7F,$BF
	.dc.b	$7E,$79,$F8,$6F,$F0,$5F,$D6,$5F,$B0,$BF,$24,$7C,$06,$EB,$53,$9F
	.dc.b	$3B,$18,$FF,$C0,$0C,$5F,$9A,$FE,$07,$EF,$3F,$C4,$B8,$70,$FC,$17
	.dc.b	$FD,$C9,$18,$EF,$6B,$E0,$FB,$4C,$FC,$7E,$3F,$70,$7E,$19,$F7,$F9
	.dc.b	$7F,$C6,$5D,$3F,$95,$CB,$FA,$3E,$2B,$DF,$67,$E7,$BE,$B3,$F8,$9D
	.dc.b	$6B,$87,$F5,$CF,$F4,$EF,$5B,$CD,$A7,$7D,$3F,$19,$5E,$2C,$BA,$47
	.dc.b	$E3,$FF,$0D,$F4,$FE,$15,$FE,$30,$FC,$A9,$F8,$C3,$F5,$07,$EB,$5F
	.dc.b	$E6,$0E,$1F,$35,$EE,$B5,$2F,$D3,$7C,$B3,$FE,$49,$D3,$CA,$FF,$81
	.dc.b	$5F,$BB,$7C,$BE,$C9,$97,$83,$2F,$ED,$95,$CA,$73,$97,$8F,$C0,$FF
	.dc.b	$58,$61,$FB,$BF,$C9,$BF,$CA,$DE,$3B,$F3,$57,$E2,$7F,$E5,$BF,$D1
	.dc.b	$BF,$CA,$9E,$F7,$F9,$B2,$F7,$5A,$FB,$A5,$3F,$2D,$FE,$7F,$5D,$36
	.dc.b	$FE,$34,$FF,$98,$E3,$F8,$13,$CB,$1F,$F3,$47,$93,$E4,$CB,$B7,$DD
	.dc.b	$B4,$F2,$57,$F9,$B5,$D6,$3F,$EC,$9C,$1F,$BF,$3F,$7B,$FE,$B9,$F6
	.dc.b	$6F,$D4,$9F,$9A,$3F,$1B,$F2,$DE,$FC,$75,$EE,$9F,$B6,$6B,$1F,$3D
	.dc.b	$FF,$50,$CF,$D6,$F1,$BE,$DF,$47,$E4,$96,$9E,$4B,$F1,$E9,$FC,$63
	.dc.b	$83,$E5,$7E,$23,$FA,$EF,$CD,$FF,$C4,$7F,$8D,$7F,$83,$BE,$63,$E4
	.dc.b	$CE,$7A,$D3,$FE,$12,$F9,$9C,$FA,$3F,$2A,$BF,$6F,$AF,$CE,$78,$53
	.dc.b	$F2,$3F,$26,$DF,$ED,$BF,$49,$A7,$F8,$CB,$F0,$2F,$F5,$3F,$81,$FC
	.dc.b	$F7,$F2,$47,$E2,$6F,$C5,$9F,$A9,$FF,$01,$FD,$65,$F9,$17,$F8,$C7
	.dc.b	$0E,$17,$BA,$D4,$FF,$07,$9C,$DF,$F4,$1C,$3A,$FD,$CD,$C3,$87,$BE
	.dc.b	$C9,$ED,$7C,$BB,$F1,$E1,$E2,$AD,$7D,$0E,$70,$7E,$AD,$FC,$63,$E8
	.dc.b	$DF,$87,$7F,$8C,$BE,$E7,$FC,$23,$97,$B2,$B5,$3F,$E3,$6F,$DF,$33
	.dc.b	$87,$83,$A7,$FB,$85,$ED,$B1,$DB,$97,$FC,$5D,$7F,$C1,$38,$7E,$0E
	.dc.b	$3D,$3D,$7F,$47,$D6,$3F,$2E,$FF,$4C,$FE,$CB,$FE,$5F,$F8,$B6,$BE
	.dc.b	$4A,$D7,$A0,$AD,$78,$99,$F8,$2F,$91,$5C,$F8,$3C,$AF,$0A,$FF,$35
	.dc.b	$78,$9E,$0C,$FB,$FD,$C7,$F0,$2D,$FE,$E7,$E9,$FF,$CE,$3F,$7D,$7E
	.dc.b	$EA,$FD,$99,$FB,$0B,$F5,$2F,$8F,$7B,$AD,$4B,$F5,$4C,$F8,$7F,$6E
	.dc.b	$BF,$A1,$F0,$7C,$87,$86,$3F,$E2,$68,$E7,$F2,$2C,$FC,$33,$FC,$9A
	.dc.b	$7E,$AF,$E6,$CE,$D7,$EF,$EF,$CB,$FF,$C7,$FF,$72,$FC,$8F,$75,$A9
	.dc.b	$CF,$CE,$E3,$1F,$B3,$8B,$EA,$7F,$20,$F5,$CB,$A7,$D9,$BC,$7F,$C7
	.dc.b	$4E,$71,$FF,$26,$FE,$8F,$F9,$1F,$CA,$FF,$D5,$1F,$B2,$FF,$3C,$F9
	.dc.b	$6F,$F4,$DF,$C3,$7E,$32,$73,$F1,$7F,$EF,$4F,$BE,$7E,$81,$9F,$B2
	.dc.b	$FD,$05,$7F,$AC,$78,$7F,$1A,$F0,$C7,$E6,$47,$3F,$C3,$D9,$F7,$FC
	.dc.b	$67,$F8,$67,$EA,$9F,$B8,$FC,$61,$F9,$DF,$F3,$0F,$C2,$5F,$95,$7E
	.dc.b	$33,$C6,$FF,$1B,$5E,$EB,$53,$FE,$52,$9C,$FB,$EC,$3F,$C6,$97,$FC
	.dc.b	$13,$EA,$EC,$E7,$0E,$DF,$A8,$F8,$A3,$1F,$2B,$FE,$A0,$FC,$57,$CB
	.dc.b	$F9,$3F,$F9,$1F,$F4,$77,$C4,$BB,$7F,$01,$FB,$97,$FA,$0B,$55,$EE
	.dc.b	$B5,$E4,$3E,$73,$53,$FE,$51,$E7,$FF,$31,$5F,$0F,$4F,$CF,$3C,$72
	.dc.b	$FF,$6D,$57,$0F,$BA,$69,$E2,$9F,$9B,$C9,$F1,$3A,$9F,$12,$FC,$91
	.dc.b	$FB,$3F,$FA,$E3,$F5,$A7,$EC,$7E,$7B,$DD,$6B,$F1,$25,$7F,$B2,$F1
	.dc.b	$E9,$8F,$CA,$B7,$EC,$36,$3B,$65,$FE,$C4,$FC,$F7,$ED,$D7,$48,$FF
	.dc.b	$63,$F9,$3F,$D2,$7C,$17,$EA,$8F,$A4,$7D,$C3,$F7,$5F,$F6,$7F,$EE
	.dc.b	$58,$DB,$FC,$E9,$87,$0B,$DD,$6B,$D3,$FF,$58,$CF,$4F,$AF,$FF,$39
	.dc.b	$3B,$E8,$FF,$47,$3C,$72,$ED,$E4,$3E,$2B,$C9,$AF,$8E,$73,$F0,$DE
	.dc.b	$7F,$B1,$7E,$65,$FE,$2F,$FA,$5F,$F7,$F7,$E4,$5F,$EA,$1F,$8B,$6B
	.dc.b	$35,$FF,$56,$5E,$FF,$1E,$5B,$F0,$CF,$3E,$5D,$30,$F2,$BF,$83,$7F
	.dc.b	$C4,$38,$F4,$58,$F4,$FB,$7F,$EF,$F8,$F4,$9E,$79,$77,$EA,$F9,$DF
	.dc.b	$69,$F9,$6F,$99,$FF,$74,$7E,$52,$FB,$87,$E5,$1F,$22,$B5,$F3,$8A
	.dc.b	$D7,$E5,$39,$F9,$EF,$97,$5F,$A0,$D3,$E5,$BC,$2B,$FC,$DB,$DD,$F5
	.dc.b	$A3,$1C,$1F,$E0,$7F,$DF,$FF,$D0,$3F,$B2,$FF,$8C,$7F,$60,$7E,$72
	.dc.b	$E3,$FF,$66,$E3,$FE,$B1,$BD,$FD,$C2,$F7,$4F,$CB,$7F,$DD,$F5,$D3
	.dc.b	$5F,$F6,$53,$FE,$8E,$FC,$AF,$FD,$64,$F2,$BE,$F3,$E3,$FC,$21,$A7
	.dc.b	$8F,$4F,$C1,$4F,$4F,$D0,$BF,$99,$7A,$AF,$E0,$17,$EE,$BF,$EC,$7D
	.dc.b	$35,$F2,$9F,$BE,$9C,$FB,$0F,$D3,$F3,$97,$93,$27,$CA,$79,$1E,$B7
	.dc.b	$D3,$DB,$F8,$D1,$8C,$FF,$C5,$57,$F8,$CE,$1F,$F7,$8F,$EB,$0F,$B8
	.dc.b	$FF,$4B,$7E,$A2,$FD,$E7,$AF,$4F,$C9,$FC,$6D,$7B,$F0,$D7,$BE,$2F
	.dc.b	$93,$5A,$CF,$E4,$AF,$CF,$7E,$4B,$CB,$3E,$FF,$23,$F6,$E3,$18,$F4
	.dc.b	$F3,$3F,$CD,$FF,$83,$FE,$5F,$ED,$FF,$B7,$7D,$BB,$F4,$37,$EF,$FF
	.dc.b	$D7,$7E,$89,$F0,$0E,$36,$3C,$7E,$A7,$FC,$9B,$87,$0E,$2F,$EF,$5C
	.dc.b	$7B,$AD,$69,$FE,$C2,$E3,$F0,$9F,$B2,$7C,$2E,$71,$EA,$E7,$1E,$BA
	.dc.b	$7F,$6F,$F1,$F8,$CF,$F9,$2E,$73,$97,$F2,$CC,$7D,$9F,$3F,$FE,$89
	.dc.b	$FD,$31,$F0,$3F,$8A,$D3,$F8,$C6,$9D,$B2,$FF,$2A,$5E,$EB,$5F,$4D
	.dc.b	$E9,$9F,$34,$FF,$A4,$9F,$36,$3D,$BC,$BF,$64,$6C,$E6,$B2,$F0,$F8
	.dc.b	$FF,$E6,$19,$D6,$7F,$DF,$EB,$59,$3C,$5F,$F3,$0E,$3F,$57,$37,$EF
	.dc.b	$0F,$D3,$1F,$B3,$FF,$98,$3F,$27,$7E,$A9,$D2,$F7,$5A,$FA,$C5,$7F
	.dc.b	$BF,$38,$BE,$06,$CB,$7C,$77,$AA,$73,$1E,$0E,$0F,$79,$F5,$DE,$B9
	.dc.b	$FF,$9C,$3E,$2B,$E9,$E1,$CF,$5F,$34,$FF,$9D,$3D,$BF,$AF,$C1,$FE
	.dc.b	$27,$FC,$8F,$E9,$F5,$BE,$F1,$F6,$CF,$BF,$F9,$57,$BE,$6F,$D1,$D6
	.dc.b	$BD,$BC,$F6,$32,$F4,$74,$7F,$31,$C3,$FC,$69,$F7,$F8,$E3,$FE,$7F
	.dc.b	$1C,$F5,$F0,$6B,$E4,$C7,$96,$9D,$BF,$0C,$F3,$FE,$64,$FC,$67,$FD
	.dc.b	$75,$C4,$FC,$FD,$FA,$1B,$F0,$8F,$3D,$EF,$F0,$BF,$AB,$5A,$FC,$3F
	.dc.b	$38,$F8,$F8,$3E,$F5,$BF,$79,$C7,$0A,$39,$69,$DF,$E8,$BC,$78,$78
	.dc.b	$B2,$FF,$3F,$72,$7B,$FE,$11,$F9,$F3,$F5,$4F,$F2,$B7,$ED,$DF,$CD
	.dc.b	$FF,$7D,$F9,$75,$A9,$CA,$FD,$4F,$37,$E2,$8F,$F0,$CF,$07,$FB,$F3
	.dc.b	$1E,$D3,$D7,$EE,$99,$7E,$5A,$BE,$6D,$F5,$9C,$FC,$D7,$FB,$F3,$D6
	.dc.b	$FE,$C9,$F8,$9B,$F3,$27,$ED,$1F,$EB,$AF,$C8,$FF,$8F,$FE,$9F,$FE
	.dc.b	$4A,$F3,$3F,$DA,$FA,$7F,$76,$C3,$85,$EF,$C2,$5B,$E7,$7F,$8A,$67
	.dc.b	$DB,$3E,$B8,$FF,$0B,$8C,$77,$FE,$75,$D6,$3C,$FC,$9F,$35,$77,$BF
	.dc.b	$2F,$E2,$6D,$BF,$99,$F9,$BA,$A7,$C8,$7A,$F9,$13,$FC,$EB,$F6,$67
	.dc.b	$EE,$2F,$C8,$7C,$B8,$7F,$84,$30,$E1,$7B,$AF,$F2,$8F,$27,$C0,$AE
	.dc.b	$BA,$3F,$1B,$F4,$47,$18,$FA,$2F,$B4,$C6,$7E,$2F,$89,$69,$FD,$6A
	.dc.b	$7C,$27,$49,$D4,$7F,$55,$8C,$FE,$55,$F5,$7F,$53,$7E,$97,$FE,$B1
	.dc.b	$FC,$C5,$F8,$3F,$F4,$FD,$7C,$D7,$BA,$D4,$FF,$44,$9C,$E8,$FD,$2F
	.dc.b	$D3,$93,$0E,$CF,$A4,$C4,$B2,$F0,$FC,$23,$3F,$F5,$A6,$3D,$27,$30
	.dc.b	$F8,$DF,$F1,$DF,$A7,$FD,$8F,$1F,$F1,$EF,$EB,$9F,$C2,$FF,$48,$7E
	.dc.b	$3A,$FC,$BD,$F3,$9F,$EF,$CF,$CB,$3F,$D7,$5F,$2B,$FE,$39,$C3,$85
	.dc.b	$EE,$BF,$E0,$0F,$93,$AC,$BA,$FC,$07,$F3,$27,$FA,$E6,$9E,$15,$87
	.dc.b	$85,$CE,$3D,$69,$A7,$93,$4F,$1F,$E4,$1E,$AA,$D4,$E7,$37,$E5,$B9
	.dc.b	$C7,$DB,$FA,$E7,$FB,$53,$57,$E4,$BF,$37,$F7,$A7,$EA,$3E,$7B,$F1
	.dc.b	$DE,$EB,$5F,$87,$67,$81,$3F,$ED,$EA,$F3,$9D,$BF,$79,$FE,$BB,$C5
	.dc.b	$ED,$AD,$FE,$2F,$FE,$43,$C7,$EE,$7C,$2F,$97,$0E,$F3,$9E,$DF,$C5
	.dc.b	$C7,$7D,$DF,$B0,$BF,$3E,$7E,$D1,$FC,$33,$F3,$AF,$F3,$4E,$BD,$EF
	.dc.b	$75,$AF,$19,$4F,$DA,$3A,$63,$E6,$E8,$FB,$58,$7F,$BE,$F3,$FF,$3A
	.dc.b	$1A,$C7,$97,$D5,$79,$F8,$7E,$6E,$1E,$3D,$7F,$8E,$FD,$4F,$E3,$3D
	.dc.b	$DF,$F3,$E7,$6B,$F8,$01,$F9,$1B,$F6,$B9,$F5,$EF,$CB,$DF,$BF,$B9
	.dc.b	$BF,$C1,$D8,$F1,$C3,$86,$1F,$ED,$0B,$7E,$43,$BC,$FB,$E7,$DB,$8B
	.dc.b	$F1,$B1,$8E,$6F,$A6,$EB,$1E,$7A,$BC,$4F,$45,$FF,$A7,$B4,$F0,$69
	.dc.b	$E6,$E5,$CC,$FE,$8F,$F7,$6F,$F5,$EF,$E7,$FF,$D6,$FE,$99,$F4,$0F
	.dc.b	$D5,$3C,$97,$CE,$F7,$F2,$DF,$E9,$EA,$7E,$79,$AA,$E9,$FB,$27,$99
	.dc.b	$E1,$F8,$45,$7C,$0B,$3F,$06,$FE,$4D,$D9,$F6,$D5,$F5,$59,$EC,$FF
	.dc.b	$11,$7E,$BF,$FB,$CF,$59,$F6,$5F,$E1,$EE,$CF,$9F,$5A,$F4,$FF,$CB
	.dc.b	$13,$FE,$E6,$9C,$FE,$6B,$E3,$D3,$F2,$7F,$53,$70,$E1,$8B,$F4,$1F
	.dc.b	$04,$B3,$F0,$7E,$A5,$FC,$43,$F0,$8F,$F6,$07,$E3,$4F,$C9,$1F,$18
	.dc.b	$FC,$C1,$FA,$9B,$C1,$FF,$42,$7E,$83,$F8,$5C,$78,$E1,$C2,$F7,$5F
	.dc.b	$E8,$9F,$D9,$35,$D9,$FE,$D8,$CF,$B6,$7F,$B4,$C7,$8F,$E9,$F6,$B5
	.dc.b	$97,$3D,$3C,$1F,$7D,$E7,$9C,$E1,$E5,$9F,$27,$A1,$99,$64,$FA,$E6
	.dc.b	$5F,$EA,$4F,$C5,$BF,$D5,$DF,$6E,$F8,$9F,$2E,$1C,$2F,$78,$AF,$7F
	.dc.b	$3F,$AC,$BA,$65,$E7,$C3,$87,$E3,$16,$73,$87,$1F,$C1,$7F,$D7,$FD
	.dc.b	$CE,$2F,$9F,$A7,$6D,$3F,$D8,$19,$7A,$7E,$EB,$FD,$33,$FB,$5F,$F4
	.dc.b	$BF,$D8,$9F,$BF,$FF,$BA,$BA,$BC,$F5,$AF,$90,$56,$BF,$11,$FE,$1B
	.dc.b	$AF,$7C,$87,$FB,$4B,$F0,$06,$7E,$9C,$7F,$D4,$FF,$2F,$67,$39,$74
	.dc.b	$D1,$FA,$4F,$96,$73,$A7,$93,$EC,$7E,$87,$E2,$6F,$DD,$9F,$46,$FC
	.dc.b	$1D,$F8,$8F,$8B,$F6,$AF,$EC,$DC,$5E,$5E,$26,$3C,$70,$E1,$7B,$F9
	.dc.b	$DF,$AB,$3D,$7F,$2C,$FA,$0E,$EF,$FA,$CF,$4F,$B9,$E7,$A7,$8B,$4F
	.dc.b	$0C,$3F,$71,$F3,$CE,$72,$F8,$BE,$4F,$E2,$7E,$AB,$FA,$5F,$88,$FC
	.dc.b	$89,$F9,$F3,$F8,$03,$E9,$FB,$5F,$88,$BD,$D6,$A7,$EB,$FC,$BE,$7E
	.dc.b	$61,$8F,$E1,$EF,$FA,$BF,$E9,$16,$73,$87,$1F,$A6,$F1,$46,$33,$F0
	.dc.b	$7C,$CF,$E7,$7E,$2D,$F0,$1F,$E1,$5E,$67,$E8,$4F,$C4,$DD,$2F,$D6
	.dc.b	$3F,$3B,$87,$0E,$BF,$9F,$7B,$A7,$E4,$7A,$C9,$F9,$5F,$E7,$70,$E1
	.dc.b	$D7,$F3,$8F,$C5,$BC,$F4,$F0,$E9,$F8,$09,$E5,$9F,$68,$FF,$26,$C7
	.dc.b	$FB,$C7,$B7,$EB,$7E,$87,$FE,$F4,$FD,$AD,$FA,$43,$F1,$BF,$F4,$C7
	.dc.b	$E3,$1F,$43,$E1,$3F,$0D,$7B,$AD,$7E,$08,$A7,$D9,$75,$CF,$F3,$F7
	.dc.b	$D5,$FF,$7A,$67,$E0,$FC,$61,$F5,$5F,$EC,$78,$C6,$BE,$2D,$7F,$D5
	.dc.b	$1C,$7F,$A8,$DB,$C3,$82,$B5,$AB,$F4,$2D,$75,$7E,$47,$F7,$27,$EF
	.dc.b	$0F,$CA,$9F,$56,$FA,$4F,$D2,$FD,$8B,$F7,$26,$DF,$8E,$E3,$FD,$BE
	.dc.b	$5C,$B1,$E3,$87,$08,$DE,$7D,$F3,$ED,$7F,$E0,$63,$27,$25,$FB,$D7
	.dc.b	$FB,$96,$78,$61,$D3,$D1,$7E,$C5,$EA,$C3,$84,$E7,$E2,$FF,$96,$7E
	.dc.b	$F7,$FD,$0B,$EE,$5F,$8A,$7C,$3F,$D4,$BF,$49,$F9,$A7,$F2,$AA,$D4
	.dc.b	$E7,$D2,$FD,$57,$B3,$FB,$F9,$FE,$97,$F9,$1F,$F2,$97,$A6,$9F,$8D
	.dc.b	$0F,$2E,$A6,$5F,$E4,$A9,$ED,$F2,$1F,$E6,$DC,$BC,$5F,$2B,$FD,$07
	.dc.b	$F5,$3F,$ED,$7F,$A3,$3F,$1F,$FE,$3F,$F4,$0F,$CC,$DF,$96,$7F,$40
	.dc.b	$BD,$D6,$A7,$FD,$CD,$C1,$FC,$56,$7F,$BA,$FE,$C9,$FC,$FD,$EB,$AF
	.dc.b	$9D,$E3,$F1,$BF,$E7,$6F,$0F,$C5,$F4,$6D,$7C,$53,$FE,$7A,$FC,$FF
	.dc.b	$F1,$9E,$57,$F1,$7E,$EF,$F4,$BF,$EA,$EF,$DC,$5F,$BC,$BF,$2F,$7E
	.dc.b	$EA,$FE,$00,$FC,$F7,$07,$F1,$4F,$81,$63,$C7,$0E,$17,$F2,$BC,$AF
	.dc.b	$E5,$D9,$F6,$C3,$F4,$67,$F4,$47,$FD,$4F,$F1,$BF,$A3,$8C,$63,$D3
	.dc.b	$4F,$1E,$8F,$D1,$BD,$35,$AD,$3C,$12,$DF,$F5,$1E,$B7,$F3,$3E,$87
	.dc.b	$EF,$1F,$47,$F0,$47,$E8,$3F,$E8,$4F,$DA,$DF,$BF,$7E,$8B,$0E,$17
	.dc.b	$BA,$D7,$8D,$9D,$7F,$AD,$E5,$D3,$D9,$FD,$04,$64,$FB,$D7,$F7,$DF
	.dc.b	$D2,$72,$7F,$06,$78,$C7,$5F,$BA,$F6,$47,$1B,$F3,$9F,$EB,$FC,$7F
	.dc.b	$D3,$3F,$5D,$FE,$0D,$F2,$DE,$CF,$F6,$CF,$CA,$BF,$DD,$DC,$98,$FC
	.dc.b	$67,$FC,$17,$5A,$C9,$5A,$B7,$EF,$9A,$7D,$33,$7F,$0E,$FF,$06,$5F
	.dc.b	$1B,$EB,$97,$86,$FC,$1F,$8D,$7E,$03,$A7,$E9,$E3,$FA,$6F,$C2,$BF
	.dc.b	$C4,$1C,$FF,$DF,$B6,$FF,$3F,$FE,$17,$F8,$8F,$45,$5A,$F1,$7F,$13
	.dc.b	$F8,$1F,$F4,$CE,$7F,$EE,$95,$AF,$C4,$7F,$09,$9F,$5F,$AA,$B3,$C9
	.dc.b	$69,$DF,$EB,$3C,$73,$98,$F1,$7C,$0F,$FA,$B7,$F6,$07,$ED,$4F,$CC
	.dc.b	$5F,$BA,$3F,$11,$FE,$77,$F1,$1F,$EC,$EF,$5B,$FD,$A9,$87,$0B,$DD
	.dc.b	$6B,$F2,$BF,$F6,$34,$F4,$FC,$D3,$FD,$55,$C1,$FE,$20,$8E,$0F,$FB
	.dc.b	$10,$3C,$F2,$F0,$78,$FF,$C8,$A7,$9E,$9E,$59,$CF,$C1,$FF,$93,$7F
	.dc.b	$64,$FE,$93,$F0,$6F,$C7,$5F,$AE,$BF,$3E,$FF,$19,$FA,$FF,$E4,$CB
	.dc.b	$DD,$6A,$7F,$C6,$FE,$E6,$7E,$49,$8F,$F9,$2E,$FF,$AA,$F9,$76,$73
	.dc.b	$87,$1F,$C7,$7C,$51,$8C,$FC,$1F,$59,$F8,$2F,$D8,$7F,$D7,$3F,$B1
	.dc.b	$7F,$B7,$BF,$6D,$7D,$AF,$F9,$6F,$C3,$FF,$18,$56,$BE,$9B,$FA,$AD
	.dc.b	$6B,$F5,$AF,$E1,$B3,$CB,$9F,$F8,$38,$7F,$D8,$BF,$41,$F8,$47,$97
	.dc.b	$CC,$AF,$E5,$63,$DB,$F1,$96,$9E,$3F,$B6,$FF,$77,$7E,$72,$FC,$EF
	.dc.b	$FC,$1F,$FB,$FB,$F2,$87,$EB,$6F,$D2,$3F,$86,$79,$1F,$98,$E1,$FF
	.dc.b	$2D,$E3,$C7,$EA,$98,$F1,$C3,$87,$E6,$7F,$ED,$9F,$53,$FC,$89,$7D
	.dc.b	$FE,$7B,$F0,$B8,$A3,$93,$F8,$57,$A6,$1C,$FC,$EF,$3E,$DE,$6D,$BC
	.dc.b	$BE,$8F,$EC,$3F,$26,$FE,$49,$F4,$4F,$D6,$BF,$35,$F6,$AF,$CD,$3F
	.dc.b	$C8,$3D,$1F,$E5,$FB,$DD,$6A,$5F,$6D,$FC,$F2,$7F,$97,$7E,$31,$FB
	.dc.b	$07,$AF,$07,$D2,$78,$FA,$8C,$BB,$4F,$1F,$7F,$FC,$23,$97,$8B,$F3
	.dc.b	$1F,$F0,$D7,$D9,$7F,$99,$7E,$C5,$F8,$A3,$F7,$0F,$F8,$01,$F8,$03
	.dc.b	$F5,$EF,$F3,$9F,$EF,$DF,$E1,$4C,$38,$5E,$FF,$27,$FF,$3F,$57,$F9
	.dc.b	$E3,$F0,$4D,$6D,$FE,$8E,$CF,$AF,$EF,$1F,$CD,$39,$99,$72,$D7,$0D
	.dc.b	$CB,$9F,$CE,$F9,$B1,$E5,$A7,$7D,$BC,$9F,$6D,$FE,$BD,$FD,$0D,$FB
	.dc.b	$A3,$F5,$AF,$F1,$DF,$B0,$FD,$33,$FD,$D7,$F3,$9F,$EB,$CB,$DD,$6A
	.dc.b	$7F,$DD,$FF,$35,$9C,$1F,$A0,$FF,$AE,$EF,$7A,$FF,$84,$DE,$3E,$DF
	.dc.b	$EF,$F0,$FF,$76,$4F,$1C,$7A,$7D,$87,$F6,$08,$FF,$57,$E2,$FC,$B1
	.dc.b	$F9,$0F,$F9,$BB,$E2,$5F,$89,$7F,$A4,$FE,$2B,$FA,$C5,$EE,$B5,$3F
	.dc.b	$E9,$7F,$A8,$CE,$1F,$ED,$CC,$7F,$AD,$E3,$FE,$9F,$C3,$84,$60,$F0
	.dc.b	$CF,$83,$F0,$3F,$F3,$2E,$1F,$EA,$19,$E7,$8F,$6F,$D4,$7D,$11,$F3
	.dc.b	$2F,$CF,$F7,$2F,$D6,$DF,$3E,$F7,$9E,$DF,$2A,$D4,$E7,$D1,$C5,$F0
	.dc.b	$FC,$3B,$AE,$FE,$DB,$DF,$AB,$8C,$FE,$2F,$D9,$BE,$79,$FA,$63,$F2
	.dc.b	$F7,$E4,$3E,$8A,$D4,$E7,$F5,$0C,$5F,$0D,$BA,$72,$2F,$79,$FE,$37
	.dc.b	$8F,$FD,$75,$FB,$EF,$F9,$D3,$F6,$2F,$EE,$7F,$9C,$7F,$98,$BF,$0D
	.dc.b	$FC,$DD,$6B,$E2,$D5,$AF,$83,$FF,$52,$FA,$1F,$9C,$AF,$F7,$6F,$D5
	.dc.b	$B2,$7E,$29,$F9,$E9,$CC,$7F,$A0,$9F,$F0,$9F,$A3,$FE,$D7,$AE,$19
	.dc.b	$76,$FA,$E6,$BF,$CA,$AB,$58,$79,$E7,$3F,$3B,$EB,$C3,$8F,$D5,$FF
	.dc.b	$B1,$38,$9F,$BA,$BF,$2A,$F5,$D6,$BD,$2F,$C9,$CB,$F3,$DC,$E7,$FE
	.dc.b	$96,$C7,$97,$CD,$F5,$8C,$5F,$E6,$AB,$1B,$BF,$1E,$9E,$AF,$E1,$BF
	.dc.b	$4D,$E7,$39,$9F,$EC,$DE,$37,$F2,$FE,$6F,$91,$CB,$F3,$67,$8C,$F4
	.dc.b	$79,$FC,$1F,$F8,$73,$F7,$07,$E9,$FF,$91,$F9,$3F,$E9,$7F,$9E,$7F
	.dc.b	$02,$F1,$3F,$CC,$75,$AF,$37,$FA,$4F,$CC,$2B,$F8,$3F,$E1,$5A,$C7
	.dc.b	$F3,$5A,$FF,$5C,$C3,$AF,$EA,$DF,$E1,$E7,$96,$7D,$B6,$F1,$5F,$87
	.dc.b	$2E,$9A,$E3,$47,$EF,$4F,$D9,$1F,$B4,$7F,$AF,$3F,$2D,$F0,$56,$A7
	.dc.b	$3C,$6F,$CC,$33,$87,$1C,$3F,$CA,$72,$FC,$29,$C3,$38,$3D,$D7,$BD
	.dc.b	$FC,$5F,$96,$7F,$BE,$3F,$34,$7E,$F1,$FE,$73,$CB,$96,$3C,$78,$BF
	.dc.b	$75,$7B,$A7,$D0,$79,$A3,$97,$E1,$1D,$3E,$89,$97,$93,$6F,$1E,$4F
	.dc.b	$DD,$FF,$99,$FC,$BF,$FB,$77,$F3,$B7,$DB,$BF,$68,$7D,$97,$FA,$62
	.dc.b	$F7,$EB,$2F,$75,$FE,$42,$E2,$F3,$E5,$D7,$DF,$FE,$D8,$FF,$84,$B0
	.dc.b	$FE,$83,$F8,$DF,$F9,$51,$E7,$87,$83,$E3,$FF,$CC,$71,$8D,$1F,$95
	.dc.b	$E3,$1F,$EE,$DF,$BD,$FF,$0D,$FB,$07,$CB,$BF,$33,$F3,$56,$BE,$11
	.dc.b	$5A,$F5,$F3,$C8,$C7,$96,$3D,$63,$17,$C3,$F2,$9C,$7E,$AE,$FC,$EF
	.dc.b	$E2,$7E,$F4,$ED,$E7,$39,$E7,$4B,$F4,$3C,$DF,$CB,$FC,$43,$9B,$F3
	.dc.b	$7F,$3E,$FF,$40,$FE,$6A,$FD,$17,$FB,$17,$EF,$DF,$E2,$AF,$DA,$FF
	.dc.b	$B3,$FE,$FF,$FE,$C8,$AD,$7C,$F2,$B5,$CF,$FE,$5A,$F0,$3E,$8D,$6B
	.dc.b	$C0,$FF,$20,$BD,$7F,$35,$FF,$41,$BC,$AF,$B7,$97,$E2,$D3,$C3,$EA
	.dc.b	$B5,$FC,$46,$3F,$9A,$C3,$C3,$F5,$4C,$9F,$9A,$BF,$2C,$FF,$70,$6B
	.dc.b	$E2,$CF,$FB,$E5,$EE,$B5,$F9,$6F,$AE,$73,$87,$2F,$8A,$F1,$70,$B0
	.dc.b	$E1,$F2,$78,$7F,$CA,$1C,$CF,$9F,$7C,$B7,$3B,$E1,$F7,$2A,$D7,$CC
	.dc.b	$67,$0E,$3F,$93,$F3,$A7,$EE,$5F,$A4,$E1,$FC,$CF,$E5,$5F,$E8,$4F
	.dc.b	$D7,$9F,$BD,$3F,$3B,$7E,$AD,$FE,$2F,$FC,$AF,$FD,$F7,$7B,$AD,$4F
	.dc.b	$F8,$CF,$F1,$2C,$FE,$1D,$FD,$6F,$B3,$FC,$E7,$7D,$5B,$0C,$BB,$E5
	.dc.b	$DB,$C4,$FF,$62,$DF,$58,$F0,$F8,$98,$BE,$19,$BF,$2B,$7E,$B4,$FD
	.dc.b	$03,$F8,$2F,$07,$B2,$F7,$5A,$9F,$DE,$CE,$74,$F7,$CF,$2E,$AE,$D1
	.dc.b	$8C,$1E,$DE,$32,$7E,$99,$FE,$0D,$FC,$BD,$FA,$EF,$C0,$C3,$E4,$CE
	.dc.b	$7F,$25,$C5,$F0,$F8,$D7,$A7,$CE,$BF,$C2,$DF,$25,$FE,$E4,$FC,$E7
	.dc.b	$F7,$9F,$8E,$BE,$85,$F3,$2F,$C8,$9F,$9E,$FE,$1F,$97,$F8,$8E,$1C
	.dc.b	$2F,$75,$AF,$C1,$55,$AF,$82,$7E,$A1,$FE,$E6,$C3,$BF,$7F,$FA,$A1
	.dc.b	$E5,$9F,$7E,$6F,$25,$B5,$F2,$4F,$8F,$EC,$B3,$1E,$7F,$D3,$FF,$A6
	.dc.b	$F8,$0F,$D0,$FC,$3F,$1D,$FC,$CB,$3F,$35,$6B,$F3,$EC,$EB,$FD,$22
	.dc.b	$73,$F2,$CF,$2F,$AF,$23,$0F,$E1,$D7,$8F,$D4,$7F,$DE,$9F,$B9,$BF
	.dc.b	$44,$E7,$2C,$BE,$25,$6B,$EC,$73,$18,$FC,$9F,$F5,$15,$8E,$4F,$76
	.dc.b	$6C,$3F,$E0,$07,$E0,$3F,$D7,$3A,$3F,$9F,$FE,$A4,$FC,$AB,$FA,$3F
	.dc.b	$E0,$1F,$D5,$3D,$27,$EF,$6F,$4B,$C7,$7B,$F9,$AF,$F3,$5F,$CF,$2F
	.dc.b	$66,$B1,$FE,$BB,$BF,$FB,$6A,$F8,$36,$3B,$D7,$1F,$53,$C9,$D5,$FD
	.dc.b	$7A,$FB,$69,$E3,$F4,$F3,$8F,$F7,$49,$CF,$E7,$1F,$CF,$7F,$38,$FF
	.dc.b	$6C,$F0,$7B,$34,$F5,$CE,$7E,$6F,$C9,$BF,$FA,$0A,$3F,$A2,$7C,$AF
	.dc.b	$4C,$99,$74,$D9,$F6,$39,$8F,$1F,$E0,$4F,$D9,$DF,$B2,$DD,$BC,$1A
	.dc.b	$7F,$A3,$AB,$52,$FC,$17,$C3,$C5,$FE,$7A,$C3,$FB,$2E,$98,$BE,$9B
	.dc.b	$F8,$FF,$25,$7E,$5E,$FD,$59,$F8,$7F,$F9,$26,$3F,$B4,$FC,$87,$E1
	.dc.b	$AB,$75,$AF,$B9,$E3,$C8,$FC,$47,$58,$3F,$76,$FF,$63,$75,$FF,$83
	.dc.b	$31,$E3,$B7,$F8,$38,$F1,$57,$87,$4F,$06,$38,$CF,$B4,$FF,$02,$AE
	.dc.b	$B0,$FD,$B9,$F9,$4F,$F6,$5F,$C5,$BF,$8F,$73,$FF,$83,$B3,$F3,$D6
	.dc.b	$BE,$51,$5F,$CA,$EB,$59,$7C,$1A,$D7,$7F,$85,$E1,$F1,$F7,$F5,$57
	.dc.b	$EC,$AF,$DD,$DF,$9B,$F9,$2D,$5A,$9F,$E2,$59,$76,$DF,$D1,$3C,$BF
	.dc.b	$78,$E9,$8F,$D9,$E1,$FF,$58,$F2,$7F,$8C,$BE,$AD,$F9,$13,$F7,$7F
	.dc.b	$E9,$BE,$4B,$F4,$A7,$EF,$9F,$EF,$DF,$03,$FC,$5B,$87,$0B,$DD,$6B
	.dc.b	$F0,$AE,$B9,$FF,$94,$2B,$4F,$D3,$32,$7F,$B7,$32,$FE,$F2,$E7,$0E
	.dc.b	$B1,$E3,$F9,$6F,$F2,$6E,$1F,$D0,$2B,$8D,$F8,$67,$3F,$25,$FE,$5E
	.dc.b	$C3,$8E,$5F,$D9,$BF,$0B,$7E,$9B,$FC,$E7,$F6,$AE,$3A,$75,$9C,$FA
	.dc.b	$EF,$9B,$CE,$1C,$70,$FE,$27,$8F,$4C,$FF,$D4,$B0,$F5,$FC,$1C,$7F
	.dc.b	$2F,$3F,$41,$F6,$7F,$74,$7F,$00,$74,$C3,$E7,$F5,$A7,$FC,$FF,$97
	.dc.b	$5B,$FC,$07,$E5,$1D,$38,$FE,$54,$74,$B7,$99,$99,$FE,$79,$FA,$1F
	.dc.b	$F9,$FB,$F1,$C7,$EA,$1F,$E0,$CF,$C2,$DF,$B0,$7D,$0F,$8B,$5E,$EB
	.dc.b	$5F,$23,$98,$7E,$8D,$F5,$F2,$C1,$77,$FF,$36,$E1,$CA,$19,$F8,$3E
	.dc.b	$93,$FC,$C7,$1C,$23,$1E,$E7,$9E,$B8,$77,$BF,$35,$F3,$FE,$8C,$FC
	.dc.b	$8B,$BC,$E7,$74,$BF,$44,$CD,$FB,$BE,$A7,$A7,$07,$FA,$7B,$0E,$1F
	.dc.b	$2D,$8D,$BF,$35,$F8,$13,$F0,$9F,$E2,$3F,$BC,$F6,$8C,$7B,$6F,$57
	.dc.b	$14,$FD,$0B,$96,$5F,$E8,$28,$E5,$8F,$E6,$BF,$39,$FD,$FF,$EF,$CF
	.dc.b	$DC,$1F,$B9,$7F,$30,$FA,$E7,$E7,$1B,$F7,$97,$BA,$D7,$E7,$79,$8F
	.dc.b	$9D,$E5,$7F,$A9,$B2,$FE,$17,$EE,$63,$4F,$F5,$DC,$78,$34,$EF,$DB
	.dc.b	$F9,$39,$E5,$18,$FB,$FF,$C3,$57,$9B,$F6,$37,$E3,$5F,$D8,$FE,$45
	.dc.b	$FC,$01,$E1,$9C,$FC,$BF,$FA,$3F,$F2,$DC,$5E,$F3,$B3,$F1,$DE,$1F
	.dc.b	$17,$EC,$F5,$DF,$AC,$E3,$8B,$F5,$9F,$92,$DF,$D1,$BC,$7C,$FF,$0B
	.dc.b	$5F,$E7,$54,$AF,$DD,$7E,$53,$FE,$C6,$F8,$EF,$E7,$DF,$A4,$7F,$5A
	.dc.b	$FB,$F7,$C0,$7F,$9C,$7F,$10,$7A,$4C,$38,$70,$30,$E1,$7F,$E7,$7F
	.dc.b	$C3,$77,$5A,$FD,$23,$F3,$8F,$0A,$BC,$7F,$EE,$8D,$3C,$3D,$8D,$7C
	.dc.b	$3F,$57,$E6,$AD,$69,$DE,$B5,$FA,$CF,$FB,$7B,$6F,$25,$6B,$2F,$F1
	.dc.b	$97,$E9,$BF,$E5,$6F,$DE,$FF,$DD,$35,$AA,$FE,$9F,$8F,$A7,$D6,$5E
	.dc.b	$FC,$8F,$4D,$73,$9C,$FD,$FB,$E0,$7C,$93,$CB,$3E,$F2,$FD,$13,$3C
	.dc.b	$3F,$33,$F4,$E7,$ED,$5F,$C0,$3F,$5F,$FE,$9B,$BF,$8E,$B5,$39,$F9
	.dc.b	$FF,$0E,$7D,$70,$E3,$F0,$0F,$D2,$FA,$F1,$7D,$19,$CF,$AD,$FC,$D7
	.dc.b	$F4,$AF,$F8,$03,$F5,$EF,$CA,$9F,$BF,$BC,$E7,$C9,$FF,$A8,$3E,$63
	.dc.b	$FD,$E3,$C4,$C7,$8E,$1C,$2F,$7F,$82,$AD,$FC,$03,$F8,$4B,$DF,$E8
	.dc.b	$1F,$C6,$5E,$39,$DE,$21,$FE,$F9,$CB,$A6,$9F,$EB,$0E,$CF,$44,$EB
	.dc.b	$6F,$45,$F9,$EB,$FD,$6B,$E8,$6A,$B9,$4F,$FB,$07,$D3,$7C,$CF,$F6
	.dc.b	$BF,$3B,$C9,$5C,$F3,$F8,$3F,$45,$FE,$FD,$AF,$EE,$1F,$93,$EA,$B9
	.dc.b	$CE,$4E,$DA,$78,$9E,$1F,$03,$8C,$3E,$5F,$E9,$CF,$C7,$BF,$C2,$7F
	.dc.b	$9C,$7F,$B9,$BA,$3E,$56,$5E,$1D,$D7,$BD,$97,$FA,$9B,$DD,$FB,$2B
	.dc.b	$5E,$2F,$7C,$38,$E3,$FE,$20,$F9,$4F,$F3,$A7,$99,$F3,$30,$7E,$8A
	.dc.b	$FC,$3F,$FC,$AB,$F9,$CF,$E0,$7F,$35,$7E,$49,$FD,$63,$EA,$BF,$D1
	.dc.b	$D8,$7D,$4C,$38,$7D,$BB,$0E,$17,$FE,$FF,$BD,$DF,$6C,$7F,$CC,$2F
	.dc.b	$F8,$3E,$33,$87,$87,$F4,$AF,$EB,$91,$8A,$F1,$F2,$33,$FF,$07,$4E
	.dc.b	$73,$F0,$4E,$7E,$43,$DB,$97,$97,$0F,$F3,$7F,$EE,$DF,$9B,$FC,$B9
	.dc.b	$FA,$B3,$F1,$1F,$F5,$FF,$0F,$C3,$DF,$D7,$7B,$AF,$F2,$6E,$BD,$E3
	.dc.b	$97,$DA,$F6,$CF,$D9,$B2,$7B,$7B,$59,$C7,$E2,$FE,$97,$FA,$BF,$CF
	.dc.b	$3F,$92,$F5,$78,$31,$F3,$62,$AD,$7C,$0F,$E1,$74,$F7,$47,$1F,$83
	.dc.b	$E9,$C5,$FE,$8A,$8E,$9E,$A7,$FA,$BF,$E0,$3E,$E7,$DF,$BD,$07,$D7
	.dc.b	$3F,$79,$FE,$05,$F9,$83,$F2,$DF,$F0,$2E,$BF,$C7,$70,$E1,$7B,$F6
	.dc.b	$57,$BE,$2F,$F7,$2E,$7D,$B3,$EB,$C5,$F6,$E3,$1F,$05,$F6,$9D,$63
	.dc.b	$CF,$0D,$E7,$DF,$0F,$2E,$BE,$4F,$C4,$FD,$59,$7F,$00,$D7,$C7,$E9
	.dc.b	$7F,$08,$E4,$7E,$A1,$FE,$8F,$CB,$96,$9F,$C5,$3D,$7F,$EA,$32,$FD
	.dc.b	$B7,$CF,$97,$3D,$7E,$1E,$5C,$EB,$BC,$B3,$EF,$87,$F5,$AF,$53,$31
	.dc.b	$F1,$3E,$43,$F9,$1F,$ED,$6F,$AC,$FF,$09,$6D,$E2,$9C,$C7,$FB,$6E
	.dc.b	$31,$E7,$2B,$FC,$63,$18,$7F,$D8,$39,$79,$87,$E2,$8F,$DB,$9F,$36
	.dc.b	$FD,$F1,$F9,$A3,$F6,$D7,$ED,$93,$3F,$EA,$D8,$F1,$C3,$85,$EE,$BC
	.dc.b	$FD,$7F,$C0,$6B,$59,$FD,$93,$F9,$7B,$E1,$FC,$4F,$ED,$46,$34,$F1
	.dc.b	$E9,$E2,$D3,$FC,$65,$3E,$9B,$F0,$4E,$7E,$07,$E6,$C7,$8F,$C0,$7E
	.dc.b	$FD,$FB,$BF,$DE,$7F,$D4,$FC,$1E,$8C,$FD,$F5,$AF,$51,$4F,$17,$B5
	.dc.b	$6B,$0F,$E0,$9F,$07,$DA,$31,$6F,$A0,$C7,$27,$83,$F7,$6F,$F7,$67
	.dc.b	$DA,$3F,$45,$E7,$39,$FC,$35,$3F,$E8,$1D,$FC,$93,$9F,$85,$E7,$AB
	.dc.b	$1F,$F7,$B7,$DF,$7F,$BF,$47,$E6,$CF,$C5,$9E,$FB,$F6,$A7,$EB,$3F
	.dc.b	$67,$F7,$B6,$BF,$7B,$AF,$FA,$FE,$5C,$B1,$E3,$87,$0E,$C7,$E8,$7C
	.dc.b	$38,$3F,$9D,$5F,$83,$F0,$8F,$F7,$7E,$3C,$B6,$F1,$3F,$CB,$EB,$71
	.dc.b	$E5,$DB,$C9,$B3,$EA,$3D,$53,$9D,$3C,$FE,$CE,$36,$F5,$FE,$D6,$FD
	.dc.b	$CF,$F6,$7F,$A4,$78,$F5,$F0,$56,$A7,$F6,$FF,$84,$66,$7F,$D6,$99
	.dc.b	$78,$70,$E5,$E8,$75,$D9,$1D,$7C,$FF,$07,$C2,$BF,$31,$F9,$BF,$05
	.dc.b	$FE,$F0,$E8,$F9,$19,$F4,$D9,$7B,$D3,$FA,$EF,$E5,$5E,$1E,$7F,$99
	.dc.b	$39,$F9,$4E,$B9,$FD,$8E,$EF,$F5,$C7,$F0,$0F,$F9,$57,$E5,$7F,$D3
	.dc.b	$BC,$5F,$EB,$4F,$AC,$FF,$48,$7C,$AF,$82,$B5,$F8,$17,$F1,$4F,$BF
	.dc.b	$57,$ED,$2B,$5F,$52,$C7,$FD,$0D,$C4,$C7,$FC,$C5,$F7,$98,$79,$E7
	.dc.b	$FE,$C2,$FA,$2F,$1E,$1C,$74,$ED,$F3,$1F,$E2,$4E,$6F,$83,$F4,$EF
	.dc.b	$8E,$FF,$59,$7C,$CF,$E4,$3E,$03,$EE,$7E,$57,$F4,$6B,$5F,$81,$FF
	.dc.b	$77,$F0,$3F,$DD,$9E,$33,$F4,$9D,$65,$FE,$BB,$E3,$FC,$41,$FE,$6F
	.dc.b	$7F,$D8,$4E,$06,$1D,$33,$F0,$5F,$97,$C9,$FF,$43,$5F,$93,$1F,$F7
	.dc.b	$D7,$CC,$FA,$7E,$8F,$FC,$25,$E9,$3F,$76,$FF,$67,$72,$BF,$45,$7E
	.dc.b	$DF,$FE,$36,$FD,$47,$F4,$F8,$70,$BD,$D6,$BE,$9D,$FB,$66,$B8,$BF
	.dc.b	$C9,$39,$F5,$FC,$2F,$FC,$1B,$8F,$1F,$9B,$B5,$AC,$B9,$E9,$FE,$05
	.dc.b	$8E,$47,$F0,$48,$CF,$8F,$FE,$FD,$FC,$31,$E8,$7F,$9B,$7D,$17,$E7
	.dc.b	$2F,$DD,$5F,$6E,$FD,$47,$FD,$47,$D3,$FE,$31,$C7,$8F,$C4,$7E,$AF
	.dc.b	$0E,$16,$FC,$33,$7B,$7F,$98,$AF,$BF,$C0,$FE,$29,$E3,$1F,$E3,$B3
	.dc.b	$C9,$AF,$8F,$0F,$17,$D1,$E3,$2E,$DA,$FF,$38,$F1,$7F,$53,$F9,$BF
	.dc.b	$F6,$7F,$DF,$3F,$D6,$FD,$7F,$EB,$8F,$B7,$FF,$1E,$56,$BE,$71,$5A
	.dc.b	$F4,$54,$FD,$B7,$F7,$D8,$F3,$DF,$F0,$CF,$F7,$38,$FE,$2B,$08,$F0
	.dc.b	$57,$DF,$3F,$CA,$98,$F7,$C3,$B7,$DD,$FE,$1E,$FF,$D0,$7F,$96,$7C
	.dc.b	$DF,$A4,$7D,$77,$F0,$0D,$7F,$C0,$A7,$D8,$9D,$46,$7D,$3F,$F0,$47
	.dc.b	$81,$FE,$61,$9E,$58,$7F,$94,$E3,$1C,$2F,$DC,$7F,$0B,$2A,$E7,$DD
	.dc.b	$FE,$0D,$C3,$C9,$39,$BF,$F1,$3E,$3E,$3F,$82,$FF,$4C,$F0,$7F,$16
	.dc.b	$ED,$FF,$23,$78,$DF,$CE,$BE,$E7,$F0,$BF,$5F,$FC,$1E,$7F,$CE,$86
	.dc.b	$BE,$2F,$FD,$61,$E0,$FF,$96,$3F,$75,$FC,$1E,$BF,$E0,$4F,$AF,$7D
	.dc.b	$BF,$F0,$CC,$3A,$CF,$E0,$CD,$BC,$F2,$FC,$9F,$D8,$F9,$23,$BD,$FF
	.dc.b	$A8,$B2,$EF,$E0,$67,$0F,$F5,$8D,$78,$FF,$14,$FF,$21,$FE,$03,$FE
	.dc.b	$D4,$E0,$F6,$63,$C6,$73,$F2,$5E,$08,$C7,$5F,$FB,$87,$1C,$71,$7F
	.dc.b	$5A,$BF,$F2,$6F,$67,$79,$F6,$4E,$7E,$19,$93,$F5,$AE,$3B,$B1,$FE
	.dc.b	$89,$2F,$DA,$73,$1F,$D1,$FE,$B7,$A7,$07,$A7,$17,$E8,$DD,$FE,$95
	.dc.b	$F0,$25,$F8,$1B,$E9,$1F,$2D,$FE,$B2,$FB,$3F,$A3,$5C,$0D,$CE,$BE
	.dc.b	$D3,$FC,$DF,$BF,$F5,$49,$D7,$1F,$D3,$8C,$65,$FD,$62,$3F,$D6,$51
	.dc.b	$89,$E7,$8B,$E4,$3C,$D7,$E5,$CF,$C9,$2F,$91,$CD,$F8,$BF,$53,$7E
	.dc.b	$53,$FE,$F8,$CB,$A6,$1E,$B9,$CF,$EE,$FE,$AC,$BA,$6C,$97,$EB,$7D
	.dc.b	$2F,$87,$0D,$FC,$11,$F9,$3B,$F0,$9F,$F6,$46,$9D,$A7,$30,$FD,$E3
	.dc.b	$B6,$FF,$32,$78,$FD,$47,$3C,$1F,$57,$F8,$1B,$7E,$4C,$FD,$1F,$FC
	.dc.b	$29,$F9,$DB,$F7,$C7,$EB,$0F,$D7,$FF,$47,$87,$0B,$DD,$6B,$F4,$9C
	.dc.b	$E3,$FE,$A1,$EC,$FA,$5E,$07,$FB,$87,$0E,$11,$80,$7A,$67,$C5,$F2
	.dc.b	$AD,$3C,$5D,$1E,$79,$CC,$79,$76,$C6,$9F,$E2,$BF,$CF,$DF,$98,$BF
	.dc.b	$36,$FF,$0E,$F3,$FF,$86,$35,$F4,$68,$BD,$D3,$F6,$4F,$4E,$3C,$F1
	.dc.b	$FC,$CE,$D8,$C3,$84,$7F,$AC,$B2,$E9,$AF,$A3,$F4,$67,$A8,$FC,$99
	.dc.b	$F4,$3C,$67,$30,$FC,$53,$14,$FA,$AF,$5E,$6B,$E7,$D7,$FB,$7F,$CA
	.dc.b	$1F,$A5,$3F,$1C,$BE,$CF,$F9,$E3,$E0,$1F,$8B,$FF,$18,$BD,$D6,$BE
	.dc.b	$FD,$4F,$B3,$D6,$2F,$CB,$BF,$C4,$F8,$70,$E6,$FA,$C1,$E7,$2C,$FC
	.dc.b	$13,$CB,$2E,$DE,$7F,$9E,$BC,$C7,$F3,$DF,$B7,$7E,$5C,$FD,$47,$84
	.dc.b	$E6,$1F,$5F,$8F,$8E,$FF,$3D,$79,$DC,$F2,$F8,$31,$FE,$7D,$F4,$31
	.dc.b	$87,$FB,$A3,$F6,$37,$E5,$BF,$EB,$9E,$FF,$F2,$BF,$0F,$A3,$E0,$D5
	.dc.b	$AF,$99,$F0,$C6,$2F,$9B,$AF,$87,$FE,$4D,$C4,$FD,$93,$EE,$7E,$0D
	.dc.b	$FE,$C8,$F8,$BF,$E8,$7F,$7F,$FD,$1A,$B5,$E2,$7F,$B3,$2B,$FA,$3D
	.dc.b	$3F,$69,$D6,$0F,$4B,$FC,$AB,$87,$0E,$AF,$9E,$F2,$C3,$B7,$E3,$3E
	.dc.b	$2B,$F0,$C7,$F5,$2F,$95,$C6,$7F,$ED,$A7,$E6,$6F,$F8,$2E,$3F,$F3
	.dc.b	$BE,$3C,$E7,$3D,$FE,$2B,$E1,$D9,$FE,$CC,$BD,$E5,$EE,$BF,$AD,$C3
	.dc.b	$DB,$3F,$F5,$6F,$ED,$FF,$D1,$BF,$51,$E5,$18,$C1,$FA,$96,$3E,$C9
	.dc.b	$7F,$A7,$F6,$3F,$7F,$C5,$FC,$AF,$4F,$97,$5A,$EF,$FE,$F3,$EF,$6B
	.dc.b	$F7,$47,$E6,$7F,$E3,$CF,$E0,$07,$E2,$1F,$D1,$BE,$DB,$FC,$F9,$BF
	.dc.b	$F4,$BD,$3F,$85,$E1,$C3,$E6,$98,$70,$F8,$B6,$0F,$CA,$FF,$AF,$E9
	.dc.b	$DF,$47,$E9,$FF,$DE,$F3,$C9,$D2,$3F,$C9,$AF,$F4,$D9,$E1,$97,$6D
	.dc.b	$7C,$93,$C0,$69,$E3,$D3,$C5,$F4,$1E,$E9,$E9,$7E,$6C,$1F,$21,$FD
	.dc.b	$8B,$F7,$2E,$58,$F1,$9C,$EE,$F8,$FC,$EB,$F1,$67,$39,$63,$2E,$99
	.dc.b	$7F,$B3,$E3,$1A,$3F,$56,$7E,$53,$FC,$07,$F1,$0C,$C6,$3D,$57,$E9
	.dc.b	$18,$A7,$E8,$5E,$9C,$7D,$BA,$FD,$5C,$39,$DF,$F8,$CB,$EA,$3E,$FF
	.dc.b	$92,$FC,$F5,$F9,$F7,$F0,$FF,$C9,$9F,$B8,$39,$FF,$31,$87,$0B,$DE
	.dc.b	$EB,$DD,$7F,$19,$ED,$5E,$F1,$FB,$11,$8D,$BF,$A6,$E7,$E1,$FC,$5F
	.dc.b	$37,$BC,$BA,$69,$FC,$DA,$FF,$C4,$33,$9C,$FC,$BF,$96,$FA,$E7,$9E
	.dc.b	$BE,$4F,$B6,$7E,$50,$FD,$3F,$F9,$C6,$DE,$1B,$FD,$F5,$EE,$B5,$F5
	.dc.b	$FD,$A7,$9F,$07,$F3,$1C,$79,$E5,$DE,$78,$CE,$6D,$F5,$59,$E1,$F9
	.dc.b	$93,$FC,$E7,$F4,$9F,$F5,$E7,$EC,$F9,$D3,$E4,$C6,$3E,$E7,$C1,$97
	.dc.b	$4E,$34,$FF,$B0,$F2,$E1,$8F,$3F,$C0,$7F,$D0,$BD,$1F,$B8,$FD,$A9
	.dc.b	$FA,$FF,$D3,$F9,$47,$E2,$2F,$51,$FB,$23,$F4,$47,$DE,$FE,$3F,$83
	.dc.b	$FB,$FE,$5C,$BA,$FF,$C2,$38,$F1,$C3,$C7,$7B,$B5,$ED,$FF,$37,$46
	.dc.b	$63,$C7,$F5,$DF,$B7,$8C,$69,$E5,$C3,$C9,$F0,$1F,$EF,$9D,$BD,$53
	.dc.b	$9D,$7F,$D4,$99,$79,$A7,$3A,$7F,$89,$3F,$1D,$FC,$C7,$E7,$6F,$DC
	.dc.b	$35,$7E,$AC,$7E,$15,$6B,$F4,$AC,$CF,$A2,$73,$D2,$FC,$9B,$C5,$3C
	.dc.b	$72,$ED,$3F,$E6,$08,$E3,$C3,$F1,$79,$3F,$B3,$7E,$AE,$F6,$BF,$C4
	.dc.b	$9C,$7E,$EC,$7F,$C7,$3F,$04,$AD,$7D,$27,$2D,$7E,$B4,$E7,$E0,$FC
	.dc.b	$3A,$79,$70,$E3,$87,$FB,$C7,$1F,$E5,$53,$FE,$0A,$FC,$85,$FB,$8F
	.dc.b	$E0,$3E,$35,$FA,$FF,$F5,$BF,$DA,$3E,$EF,$B3,$FC,$F3,$E7,$58,$70
	.dc.b	$DD,$87,$0F,$8E,$74,$70,$D3,$FC,$83,$87,$0F,$E0,$1F,$B5,$97,$3C
	.dc.b	$BC,$35,$F8,$08,$C6,$BE,$3D,$7F,$C3,$73,$CC,$F4,$46,$7E,$3B,$FC
	.dc.b	$9B,$BF,$B3,$2F,$2C,$FE,$6B,$E7,$3F,$88,$FE,$CC,$FD,$0F,$95,$6B
	.dc.b	$87,$FC,$3D,$EB,$29,$FA,$B6,$B7,$F7,$63,$FC,$6B,$EA,$3B,$F0,$79
	.dc.b	$B1,$E3,$F8,$DE,$38,$3D,$F8,$7C,$5F,$31,$FB,$C7,$F8,$D3,$F5,$6E
	.dc.b	$9C,$3F,$AE,$9C,$FD,$23,$F3,$6C,$E7,$FC,$AB,$0F,$83,$C9,$ED,$CB
	.dc.b	$A4,$7F,$A5,$F3,$FD,$8F,$C9,$7F,$70,$FD,$BB,$FA,$EF,$E8,$BF,$D7
	.dc.b	$BF,$22,$FF,$38,$FE,$F7,$FD,$93,$25,$EF,$E4,$97,$BC,$56,$FD,$AB
	.dc.b	$AC,$7F,$9E,$B2,$FE,$D3,$F8,$6F,$EC,$F1,$E5,$85,$AD,$63,$D0,$67
	.dc.b	$FE,$82,$FD,$03,$EB,$DF,$09,$CD,$3F,$08,$CB,$CB,$0F,$57,$E8,$7F
	.dc.b	$EA,$5F,$B3,$F1,$CB,$4F,$7E,$1D,$AB,$FD,$55,$FB,$BE,$A7,$3A,$25
	.dc.b	$F6,$5D,$B0,$E5,$87,$9E,$B5,$F7,$99,$C3,$AF,$A9,$F9,$EF,$D8,$9F
	.dc.b	$49,$D7,$0F,$74,$E6,$B1,$B7,$F8,$3A,$3F,$4B,$F5,$DC,$EB,$7E,$6F
	.dc.b	$FA,$43,$9F,$E9,$EB,$FE,$A6,$FD,$4B,$FC,$A5,$F1,$8F,$DE,$9B,$3F
	.dc.b	$26,$7E,$D2,$E5,$F8,$78,$70,$BD,$D6,$BF,$78,$CF,$D3,$FF,$C0,$0C
	.dc.b	$FF,$96,$F5,$7F,$8D,$B0,$EF,$18,$3C,$52,$D3,$C5,$A7,$87,$F1,$7F
	.dc.b	$91,$7D,$E3,$C9,$F8,$F6,$2F,$86,$8F,$D6,$1F,$A9,$BF,$63,$7E,$51
	.dc.b	$E4,$D3,$DD,$5A,$97,$E5,$19,$C3,$96,$0F,$0B,$B6,$1C,$23,$F1,$1B
	.dc.b	$EF,$17,$E8,$CF,$D7,$1F,$9F,$E6,$31,$F8,$ED,$FC,$6B,$9E,$9E,$79
	.dc.b	$FF,$5B,$63,$F8,$EF,$B0,$7E,$73,$FE,$F1,$FA,$57,$CE,$79,$BF,$25
	.dc.b	$FE,$C3,$DF,$F6,$B0,$FF,$2C,$DE,$EB,$5E,$3F,$38,$FF,$7D,$4F,$CF
	.dc.b	$37,$F3,$1F,$A9,$78,$FE,$21,$FE,$27,$BF,$87,$DB,$F9,$7D,$3C,$73
	.dc.b	$9E,$7F,$C9,$DF,$58,$E1,$F8,$F3,$F2,$77,$1E,$78,$3D,$33,$9F,$CB
	.dc.b	$71,$A2,$3F,$AD,$63,$87,$FC,$86,$EB,$F0,$5F,$E1,$3F,$8F,$5F,$0D
	.dc.b	$F9,$AF,$2F,$C7,$7F,$D0,$3C,$BF,$0F,$87,$CF,$AF,$C8,$C5,$F2,$5F
	.dc.b	$B1,$F9,$CF,$F1,$67,$DF,$3F,$4D,$7E,$42,$FC,$19,$F6,$FF,$B8,$D3
	.dc.b	$E7,$61,$C2,$F7,$5A,$F8,$2F,$E0,$B3,$D3,$8B,$E7,$3F,$E3,$9F,$27
	.dc.b	$E7,$3C,$72,$ED,$F6,$9E,$3F,$B2,$FF,$4D,$E5,$D3,$F5,$AC,$6D,$E5
	.dc.b	$B7,$E6,$8F,$A3,$FF,$22,$EF,$E4,$C7,$8D,$6A,$5F,$8C,$79,$27,$37
	.dc.b	$F9,$BD,$B1,$8F,$58,$FC,$DE,$3C,$E3,$C1,$F9,$43,$F3,$E7,$E3,$9C
	.dc.b	$36,$F8,$F3,$9F,$B1,$C6,$C8,$7E,$1D,$EA,$AD,$FB,$FF,$C8,$7F,$40
	.dc.b	$FE,$55,$F4,$AF,$AE,$7E,$8C,$FB,$37,$CF,$BF,$07,$FE,$63,$87,$0B
	.dc.b	$DD,$6B,$F4,$CC,$FC,$57,$FB,$FB,$9B,$F3,$D7,$D9,$B0,$CF,$C1,$9F
	.dc.b	$7F,$A3,$F2,$4E,$72,$EB,$EC,$62,$F7,$F5,$2F,$DE,$7F,$63,$E5,$FF
	.dc.b	$B4,$34,$F2,$67,$FE,$27,$97,$ED,$FE,$4C,$38,$F8,$FC,$F3,$F7,$FE
	.dc.b	$85,$F9,$8C,$B9,$6B,$E1,$FC,$C1,$FA,$47,$FA,$E3,$F0,$AE,$58,$27
	.dc.b	$3F,$85,$7F,$B3,$3C,$1F,$E0,$5C,$1F,$16,$78,$F1,$77,$79,$F1,$FF
	.dc.b	$51,$FC,$A5,$FA,$8B,$F2,$87,$E2,$AF,$E0,$07,$E5,$CB,$FF,$16,$E9
	.dc.b	$FC,$3D,$E3,$7C,$2B,$70,$FC,$E7,$FB,$3B,$DB,$F4,$4F,$E1,$CF,$0E
	.dc.b	$CF,$E2,$4F,$2C,$BC,$19,$78,$E7,$8E,$7E,$59,$CF,$C2,$FA,$63,$1F
	.dc.b	$9E,$3F,$66,$7E,$3F,$FE,$20,$DF,$CD,$5A,$9C,$FE,$E5,$D3,$83,$DF
	.dc.b	$87,$1F,$AC,$E9,$18,$AF,$F0,$F6,$1C,$A3,$7F,$95,$7F,$9B,$FF,$35
	.dc.b	$E5,$AB,$14,$FF,$9C,$FE,$FF,$31,$8F,$A2,$EB,$5B,$AF,$F3,$A6,$3F
	.dc.b	$CE,$7F,$03,$7E,$52,$FD,$2B,$FC,$CB,$E4,$BF,$51,$FD,$67,$6F,$F8
	.dc.b	$4A,$F7,$5A,$9F,$D8,$7E,$AB,$9C,$39,$73,$FF,$3F,$BE,$BD,$3F,$CF
	.dc.b	$DE,$3F,$6D,$FE,$7D,$C3,$F5,$D5,$C7,$1E,$9F,$90,$FF,$85,$EF,$7F
	.dc.b	$7C,$FD,$FB,$FA,$F7,$26,$F2,$FF,$2D,$47,$F6,$1E,$5C,$56,$F7,$E5
	.dc.b	$B7,$B6,$F7,$F6,$38,$D3,$FD,$07,$7F,$E5,$8F,$D9,$DF,$A9,$F2,$FB
	.dc.b	$84,$3E,$73,$1D,$7C,$72,$FF,$35,$53,$C1,$FF,$32,$71,$BE,$37,$FD
	.dc.b	$3D,$ED,$FE,$77,$DC,$FF,$54,$7E,$F4,$FD,$1F,$FC,$FD,$5A,$F2,$BF
	.dc.b	$8C,$F8,$2A,$7C,$D6,$BF,$0C,$FE,$91,$9F,$FA,$24,$1D,$D7,$F0,$07
	.dc.b	$FA,$42,$33,$8F,$5C,$FF,$D7,$1F,$8B,$79,$A3,$19,$7F,$5A,$FC,$0F
	.dc.b	$9D,$F3,$DB,$F6,$9F,$AA,$3F,$69,$7E,$7A,$FB,$3F,$25,$6A,$73,$F6
	.dc.b	$B8,$E2,$61,$C7,$EF,$3C,$1C,$3F,$D8,$67,$37,$FE,$12,$F2,$BF,$D6
	.dc.b	$FB,$FC,$3F,$C5,$9F,$A4,$3F,$14,$FE,$9B,$5A,$C7,$95,$6A,$5F,$94
	.dc.b	$F8,$63,$8F,$CC,$F8,$37,$FE,$1D,$1D,$3F,$26,$FF,$64,$FE,$49,$F4
	.dc.b	$FF,$7E,$FA,$9F,$AC,$7F,$94,$BF,$10,$7C,$17,$FB,$8A,$B5,$E0,$2B
	.dc.b	$5F,$A0,$E6,$3F,$BC,$F5,$7F,$BF,$3E,$83,$E5,$DF,$58,$C1,$F6,$1E
	.dc.b	$77,$FE,$60,$FC,$97,$C7,$6C,$FC,$7F,$7E,$F0,$33,$1F,$89,$FC,$A1
	.dc.b	$F9,$17,$9B,$F4,$CF,$14,$E6,$3D,$3E,$BE,$31,$5F,$0F,$C0,$B9,$C6
	.dc.b	$29,$F6,$D8,$FC,$DD,$EC,$3F,$6B,$7E,$45,$CE,$73,$1F,$E0,$5E,$F7
	.dc.b	$2F,$C0,$CB,$D4,$2B,$9F,$9F,$D7,$1F,$F4,$AF,$33,$F2,$B7,$3F,$E4
	.dc.b	$BF,$71,$7E,$11,$FD,$D3,$D4,$7E,$90,$FD,$65,$5F,$D5,$70,$E3,$8F
	.dc.b	$E9,$23,$8F,$D1,$FF,$C0,$0B,$E1,$1F,$E0,$0A,$DD,$72,$C7,$FC,$02
	.dc.b	$FE,$88,$FD,$6C,$6A,$7C,$7F,$80,$66,$73,$9F,$9E,$FC,$D8,$79,$79
	.dc.b	$33,$5E,$39,$CE,$6E,$6E,$BF,$37,$FB,$0F,$69,$FA,$F3,$E8,$FF,$53
	.dc.b	$F9,$B7,$EA,$BE,$65,$F9,$2F,$C7,$C3,$85,$EF,$D5,$5E,$FE,$5B,$5B
	.dc.b	$7F,$06,$CF,$AE,$7E,$98,$C9,$FD,$CF,$07,$EF,$39,$E5,$FF,$3F,$5F
	.dc.b	$0C,$BA,$8F,$CE,$FD,$38,$75,$C3,$C5,$9F,$F0,$0B,$CC,$BF,$55,$7D
	.dc.b	$8F,$F9,$DF,$F7,$07,$E3,$6F,$CE,$3F,$DD,$3F,$7B,$E4,$BD,$FE,$0E
	.dc.b	$BD,$D7,$FB,$BB,$CB,$D6,$3F,$E8,$1C,$BE,$A7,$07,$FA,$57,$1E,$FF
	.dc.b	$A9,$7A,$F9,$7F,$8E,$C7,$4C,$B9,$FC,$2F,$9A,$31,$B7,$8F,$D6,$F4
	.dc.b	$F7,$BD,$97,$E8,$0F,$D9,$9F,$A5,$FE,$67,$F3,$D7,$CE,$FC,$CC,$38
	.dc.b	$5E,$EB,$5F,$27,$9C,$79,$F1,$7A,$38,$3F,$83,$C3,$F7,$AF,$44,$FF
	.dc.b	$07,$8F,$0E,$9E,$0E,$3F,$A5,$3C,$F2,$ED,$1F,$EA,$61,$F1,$9F,$CF
	.dc.b	$3F,$47,$7E,$9B,$FE,$5D,$FC,$C1,$FB,$97,$F2,$8F,$DD,$DF,$79,$FB
	.dc.b	$3C,$38,$5E,$EB,$5F,$7D,$9D,$3F,$A8,$E5,$D3,$C7,$FC,$46,$1C,$3E
	.dc.b	$F6,$D6,$B1,$E5,$F1,$5F,$ED,$AD,$BC,$B3,$9A,$6D,$E4,$AF,$CD,$CF
	.dc.b	$71,$97,$AA,$B7,$A3,$F8,$07,$FC,$85,$F7,$3B,$AD,$63,$EE,$9F,$F2
	.dc.b	$37,$C7,$67,$4F,$64,$3F,$07,$FD,$FE,$DE,$9B,$E1,$CB,$FE,$D3,$FD
	.dc.b	$39,$F6,$7E,$2F,$6F,$FC,$B5,$F8,$47,$FA,$FB,$E0,$BF,$D8,$55,$AD
	.dc.b	$95,$AF,$AE,$FF,$77,$4B,$F4,$5F,$F7,$17,$E1,$C8,$EB,$CB,$FE,$E7
	.dc.b	$8C,$67,$DF,$D0,$FF,$55,$67,$DA,$31,$F9,$57,$F1,$6F,$C0,$BF,$BC
	.dc.b	$FE,$9C,$FC,$FD,$FB,$E3,$F5,$3F,$F4,$97,$E2,$BF,$E4,$8F,$D1,$7F
	.dc.b	$98,$61,$C2,$F7,$5A,$FB,$87,$D5,$EB,$83,$FC,$1F,$9F,$5C,$FF,$83
	.dc.b	$63,$C7,$F7,$DB,$5A,$CB,$9F,$E1,$1F,$EE,$8C,$BC,$D3,$9C,$3C,$93
	.dc.b	$FE,$EC,$FB,$7F,$F0,$DF,$0B,$CD,$FF,$09,$F8,$5F,$E2,$2F,$DF,$DF
	.dc.b	$A4,$3E,$73,$FD,$6D,$FB,$FB,$F5,$56,$3C,$75,$FF,$12,$E3,$C7,$0E
	.dc.b	$1E,$4B,$0F,$E6,$57,$BB,$FF,$7B,$FD,$8F,$F9,$3D,$FE,$7D,$19,$C3
	.dc.b	$C5,$18,$D3,$CB,$4D,$3C,$BA,$7F,$B2,$7A,$FD,$75,$A9,$CE,$7F,$E4
	.dc.b	$7C,$3C,$F3,$9D,$3D,$9E,$6B,$D0,$7E,$8B,$AB,$F5,$E9,$DE,$F7,$4F
	.dc.b	$D5,$75,$5E,$AC,$FC,$78,$FF,$BB,$B2,$ED,$AF,$F4,$DF,$86,$F2,$DF
	.dc.b	$4B,$DF,$E9,$5F,$DC,$B8,$7E,$06,$BD,$F0,$FE,$A9,$3F,$C3,$27,$35
	.dc.b	$DB,$EB,$F9,$E1,$E1,$DB,$D7,$F7,$58,$F1,$3F,$D7,$FF,$32,$FD,$AF
	.dc.b	$FD,$99,$F2,$CF,$CB,$7F,$55,$C3,$FE,$52,$BD,$D6,$A5,$EA,$FB,$A7
	.dc.b	$9E,$7F,$A7,$F3,$1F,$80,$7E,$B1,$FF,$0C,$E1,$DB,$CA,$F1,$46,$3E
	.dc.b	$57,$FC,$3F,$CD,$8D,$BC,$91,$FE,$22,$FC,$B5,$F9,$3F,$7B,$E1,$C0
	.dc.b	$97,$E1,$B9,$C3,$8C,$FF,$2A,$D7,$1A,$7B,$BD,$F7,$E5,$F8,$D3,$FC
	.dc.b	$5F,$1F,$E2,$CE,$8F,$C9,$71,$7F,$4A,$FB,$6D,$FB,$6F,$11,$FA,$EB
	.dc.b	$EC,$BE,$FD,$EF,$3F,$E3,$55,$AF,$A5,$CF,$C5,$FF,$9E,$77,$FC,$C1
	.dc.b	$FE,$88,$BF,$EA,$FF,$C0,$3F,$E0,$3D,$7C,$35,$FE,$7F,$C7,$9F,$E0
	.dc.b	$BF,$72,$FB,$6B,$E3,$E4,$C5,$EE,$3C,$DF,$83,$BF,$31,$F0,$F0,$FA
	.dc.b	$B7,$FF,$66,$7C,$AE,$34,$F3,$FE,$83,$CA,$B5,$FA,$15,$9F,$E6,$9A
	.dc.b	$BF,$3F,$70,$F1,$D7,$F0,$71,$FE,$95,$E1,$F4,$E3,$F2,$2D,$F8,$87
	.dc.b	$3B,$FC,$4F,$0F,$E5,$BF,$42,$FF,$2F,$FD,$AB,$DD,$7D,$EF,$F6,$4F
	.dc.b	$C0,$BF,$B1,$E9,$FC,$16,$F7,$5A,$F8,$F5,$7F,$2D,$AD,$78,$0F,$B8
	.dc.b	$F9,$9E,$5F,$FA,$B3,$1E,$FD,$9F,$48,$3D,$33,$FE,$E7,$E6,$79,$70
	.dc.b	$6B,$E5,$9F,$EE,$FE,$76,$63,$1F,$98,$7F,$AD,$3F,$7A,$FE,$03,$8F
	.dc.b	$38,$FF,$5B,$FE,$F5,$8C,$79,$F0,$FB,$A7,$FC,$E3,$BF,$A3,$1E,$7A
	.dc.b	$75,$AF,$87,$EA,$BF,$2A,$7E,$8D,$CE,$FC,$93,$9F,$94,$C5,$F0,$DF
	.dc.b	$97,$07,$FB,$6B,$ED,$FF,$82,$C7,$F6,$1F,$C5,$5F,$07,$FE,$3D,$F9
	.dc.b	$37,$E4,$3F,$C8,$78,$3F,$9A,$5E,$EB,$53,$FC,$1E,$73,$F2,$4F,$C2
	.dc.b	$7E,$7F,$47,$FB,$93,$0E,$CF,$CF,$79,$78,$EC,$BF,$99,$61,$CB,$3F
	.dc.b	$1C,$E6,$BC,$F8,$A7,$3F,$A1,$BF,$1C,$FE,$A1,$87,$1D,$7F,$CA,$D3
	.dc.b	$9F,$C1,$FF,$99,$E3,$E0,$8F,$F0,$76,$98,$CB,$A5,$F0,$F9,$9E,$F9
	.dc.b	$F8,$BF,$4E,$7E,$F5,$FE,$20,$D3,$B6,$C9,$CF,$E6,$3C,$EF,$87,$A7
	.dc.b	$CB,$2F,$F3,$45,$7F,$BB,$7C,$2F,$F7,$EF,$E8,$BF,$E5,$9F,$A1,$FE
	.dc.b	$C5,$F8,$5B,$F6,$F7,$EE,$4F,$C6,$BF,$A5,$DE,$FE,$4D,$7B,$A7,$E3
	.dc.b	$BA,$FA,$A7,$E6,$3F,$4B,$9D,$8F,$77,$F4,$0F,$3C,$3C,$1F,$79,$FD
	.dc.b	$22,$31,$A3,$F4,$5C,$63,$FE,$FC,$D3,$C3,$FB,$C7,$F2,$3F,$D3,$7C
	.dc.b	$1C,$1E,$A8,$FE,$4D,$F8,$AE,$2B,$7E,$FF,$2A,$F7,$E4,$DB,$15,$D3
	.dc.b	$E0,$5F,$9D,$71,$E0,$66,$9B,$D9,$FE,$07,$E0,$F5,$4B,$E0,$79,$DF
	.dc.b	$0F,$81,$FF,$0D,$FD,$6F,$7D,$1F,$82,$7F,$61,$F9,$4F,$17,$DB,$3E
	.dc.b	$DF,$FD,$4D,$D3,$FE,$1A,$C3,$85,$EF,$0F,$F7,$2D,$7F,$AA,$F8,$F5
	.dc.b	$1E,$B9,$D3,$FE,$D9,$DF,$FC,$DA,$F8,$3E,$3B,$FD,$DC,$FF,$87,$30
	.dc.b	$EB,$1F,$DC,$63,$AF,$E5,$3E,$87,$A4,$67,$3F,$37,$36,$67,$A6,$BE
	.dc.b	$AF,$C9,$5F,$B6,$BF,$53,$F2,$D6,$8F,$F7,$EC,$EB,$EE,$F3,$B7,$B7
	.dc.b	$24,$BF,$17,$ED,$39,$87,$E3,$18,$D7,$C9,$E0,$BF,$53,$7E,$21,$D2
	.dc.b	$73,$1F,$CD,$78,$BD,$FC,$1F,$42,$FE,$C6,$DB,$8C,$7E,$41,$FE,$52
	.dc.b	$FB,$1F,$B1,$F5,$8E,$A7,$EE,$EF,$D9,$FF,$DE,$DF,$9D,$BE,$67,$FD
	.dc.b	$B5,$D6,$C7,$8E,$1C,$2F,$74,$FC,$F7,$FC,$AF,$5A,$FD,$D3,$FB,$93
	.dc.b	$C3,$2B,$B7,$E5,$F4,$F1,$47,$87,$EB,$1A,$FC,$4A,$E6,$6F,$4F,$04
	.dc.b	$EB,$A3,$D9,$39,$BE,$5F,$BF,$78,$FF,$14,$E3,$3F,$E6,$BC,$38,$CF
	.dc.b	$9F,$F6,$8C,$E7,$E2,$E0,$66,$AF,$8D,$86,$B0,$E3,$D3,$E0,$8E,$99
	.dc.b	$FA,$7F,$4F,$FF,$73,$F4,$6E,$31,$F4,$1F,$DC,$B9,$7E,$55,$F9,$F4
	.dc.b	$EF,$87,$69,$CC,$F4,$9C,$F9,$1F,$BD,$FC,$15,$E0,$7E,$C7,$D4,$FF
	.dc.b	$44,$FE,$B7,$FD,$B3,$EA,$5F,$9E,$AD,$FA,$95,$C3,$87,$D8,$B0,$E1
	.dc.b	$7B,$F4,$75,$AF,$F0,$BC,$FF,$D2,$BF,$AE,$7E,$FE,$3F,$80,$67,$E2
	.dc.b	$75,$8F,$37,$47,$23,$F9,$5B,$DF,$E4,$DA,$F9,$BE,$C3,$D9,$AF,$93
	.dc.b	$5F,$1F,$D8,$63,$1E,$B1,$ED,$F1,$5F,$BD,$7F,$B0,$F4,$F1,$57,$6A
	.dc.b	$D4,$BF,$0F,$F1,$6F,$F1,$B5,$F0,$E3,$FE,$5B,$C7,$8C,$E7,$5F,$F3
	.dc.b	$C7,$8F,$E1,$CF,$A6,$1E,$8F,$B3,$7E,$FD,$FE,$43,$C7,$8F,$0F,$B7
	.dc.b	$1F,$F3,$BC,$BE,$2B,$C3,$39,$BF,$C5,$DF,$0D,$FE,$C4,$7E,$E3,$A7
	.dc.b	$FD,$71,$FA,$63,$F7,$3F,$F5,$97,$2B,$F2,$E7,$F0,$03,$F5,$E7,$EB
	.dc.b	$9F,$EA,$6C,$38,$7D,$E3,$0E,$17,$BF,$D7,$35,$97,$FA,$F7,$93,$FC
	.dc.b	$71,$87,$7F,$CE,$FF,$AF,$3C,$B0,$E3,$F8,$6F,$FA,$8A,$31,$AF,$8B
	.dc.b	$EC,$31,$8F,$9A,$3F,$C8,$DF,$66,$FC,$6F,$BE,$DE,$39,$CF,$E0,$78
	.dc.b	$CF,$D9,$F6,$0F,$C2,$B9,$DE,$E3,$FC,$57,$BE,$F0,$FF,$24,$FD,$67
	.dc.b	$F7,$BE,$5F,$8F,$AF,$87,$B5,$3F,$DE,$A7,$38,$F8,$67,$FC,$11,$87
	.dc.b	$F6,$7E,$0F,$5C,$7F,$92,$3D,$DF,$EB,$9F,$40,$F3,$BF,$D6,$BF,$AB
	.dc.b	$3F,$4A,$FE,$4D,$CF,$FC,$C7,$2F,$F1,$A5,$6A,$73,$F0,$5F,$D1,$B0
	.dc.b	$E5,$F1,$1F,$E9,$3C,$BA,$DF,$FB,$03,$0F,$EB,$07,$2F,$B8,$67,$DF
	.dc.b	$D2,$F2,$4E,$63,$1C,$3F,$D5,$F2,$F3,$5F,$0D,$1F,$9F,$3F,$45,$7E
	.dc.b	$41,$E3,$D3,$CD,$39,$FC,$93,$18,$71,$8F,$F0,$B6,$98,$D3,$E0,$43
	.dc.b	$ED,$7C,$18,$7B,$FF,$02,$7E,$60,$FD,$3F,$86,$DF,$EB,$C8,$7F,$00
	.dc.b	$62,$BF,$D8,$19,$EB,$F4,$5F,$F0,$57,$5F,$28,$FF,$0D,$7C,$A3,$EC
	.dc.b	$DF,$7B,$FE,$CE,$EF,$FF,$19,$FE,$38,$FC,$29,$8F,$1C,$58,$F1,$C3
	.dc.b	$87,$CD,$B0,$FD,$25,$EE,$B5,$7F,$C6,$23,$19,$3F,$43,$FF,$65,$6F
	.dc.b	$FE,$1F,$BE,$9F,$83,$FE,$D7,$27,$D8,$FF,$8C,$71,$E3,$A7,$83,$DD
	.dc.b	$C6,$3C,$73,$F5,$E3,$FB,$DF,$2D,$FA,$DB,$F7,$7F,$5D,$EF,$1F,$81
	.dc.b	$5A,$FD,$EB,$3C,$1F,$1B,$2E,$7C,$6A,$FD,$84,$E8,$F8,$73,$A8,$FE
	.dc.b	$49,$19,$E0,$F5,$D3,$EF,$9F,$BE,$7E,$2B,$83,$D7,$BF,$F1,$39,$CF
	.dc.b	$37,$1C,$FC,$31,$FB,$9C,$B1,$AF,$D7,$87,$D1,$7F,$2B,$F8,$A7,$EA
	.dc.b	$6F,$CE,$1F,$83,$3F,$43,$7E,$CA,$FD,$75,$F8,$A7,$F4,$BC,$78,$E1
	.dc.b	$C2,$F7,$4F,$D4,$35,$97,$5D,$BF,$A9,$FC,$2F,$F4,$6C,$7C,$3F,$9D
	.dc.b	$FF,$56,$9D,$E5,$CF,$4F,$CF,$46,$35,$F2,$43,$F3,$9F,$56,$5E,$99
	.dc.b	$F4,$7D,$F3,$F8,$03,$F4,$DA,$77,$AD,$6F,$FE,$AF,$97,$EE,$3D,$63
	.dc.b	$FD,$65,$F7,$5D,$35,$F6,$63,$E2,$C7,$94,$7B,$FF,$70,$FF,$4D,$FE
	.dc.b	$6B,$C5,$FF,$4E,$FE,$E3,$7E,$B7,$96,$DF,$2E,$7F,$77,$87,$F9,$A3
	.dc.b	$F0,$C6,$0F,$D2,$BF,$CA,$9D,$0F,$69,$FB,$D7,$F9,$0F,$0E,$17,$BF
	.dc.b	$AB,$FA,$75,$AF,$C1,$BF,$C1,$DF,$01,$FE,$83,$CF,$AE,$7F,$37,$1E
	.dc.b	$3F,$0D,$6B,$59,$73,$D3,$C1,$DF,$FE,$1D,$9C,$E1,$E4,$CF,$F1,$98
	.dc.b	$F8,$A3,$CF,$F0,$EF,$D8,$BF,$DF,$53,$9D,$BF,$0F,$39,$BF,$2F,$6F
	.dc.b	$FA,$97,$6F,$2D,$6B,$E3,$BC,$13,$9D,$BC,$98,$3E,$3F,$33,$FE,$FA
	.dc.b	$9F,$47,$E1,$0F,$C2,$FF,$89,$69,$E1,$E1,$4E,$7E,$3F,$C3,$87,$1E
	.dc.b	$0F,$9F,$2F,$C7,$7A,$EF,$E4,$C3,$8E,$1F,$E3,$FF,$C2,$3F,$D5,$DF
	.dc.b	$2B,$FE,$FD,$FC,$BB,$F9,$1F,$DD,$BF,$06,$FF,$46,$FE,$BB,$FE,$3F
	.dc.b	$BD,$FE,$22,$BD,$D7,$FA,$4B,$EC,$B5,$C1,$FE,$C7,$CB,$A3,$FA,$6C
	.dc.b	$38,$FE,$43,$F6,$5E,$78,$72,$F0,$BC,$95,$C7,$2E,$DE,$93,$5F,$27
	.dc.b	$8F,$AA,$FF,$0B,$D6,$A5,$F8,$4B,$F4,$8F,$F7,$AF,$3B,$4E,$F9,$73
	.dc.b	$AD,$67,$E8,$CB,$9E,$9F,$0A,$5F,$57,$DF,$7F,$8D,$87,$1E,$3C,$6D
	.dc.b	$E1,$B7,$ED,$9F,$E9,$7D,$7C,$19,$FA,$F3,$E9,$2F,$CF,$FC,$19,$F4
	.dc.b	$E0,$F4,$FE,$41,$D2,$FC,$5E,$3F,$BF,$F0,$CF,$F5,$50,$FD,$83,$F9
	.dc.b	$5F,$D0,$BE,$F7,$F8,$2F,$2F,$FB,$A6,$B7,$5A,$9F,$E0,$BB,$F9,$37
	.dc.b	$57,$FA,$D7,$F5,$2F,$A9,$1F,$EF,$F8,$EE,$FD,$93,$1F,$3B,$C3,$0F
	.dc.b	$0E,$BE,$58,$C7,$C1,$FF,$8B,$FF,$23,$C6,$1E,$78,$FF,$19,$7E,$46
	.dc.b	$FC,$9F,$E8,$4F,$9A,$73,$F8,$9E,$30,$E7,$C1,$FD,$C3,$4C,$5F,$38
	.dc.b	$FE,$51,$18,$D3,$FA,$6F,$D4,$FF,$A5,$7C,$1C,$64,$8F,$F4,$C7,$87
	.dc.b	$87,$FC,$B5,$77,$BB,$F1,$9F,$81,$3E,$D7,$F3,$DF,$98,$FF,$6A,$FC
	.dc.b	$F5,$FA,$87,$FB,$67,$9F,$E5,$5E,$FA,$D7,$BA,$7E,$55,$F8,$DA,$D6
	.dc.b	$7F,$2C,$3D,$FC,$8F,$A3,$18,$CF,$C3,$F2,$AD,$3F,$CA,$33,$CE,$73
	.dc.b	$1F,$CB,$63,$19,$F8,$AB,$87,$C2,$F9,$35,$F3,$E1,$C2,$1F,$92,$70
	.dc.b	$DB,$DD,$1F,$0A,$7A,$FC,$95,$C7,$A7,$9C,$FB,$17,$EE,$5C,$74,$FE
	.dc.b	$D7,$0F,$07,$14,$FC,$0B,$96,$9F,$E1,$48,$FF,$4B,$F8,$8E,$09,$F3
	.dc.b	$0F,$C3,$FF,$39,$7F,$D9,$7A,$9F,$B1,$3F,$46,$FD,$C5,$EF,$C8,$FA
	.dc.b	$3C,$8B,$DF,$E9,$FA,$CB,$F4,$D9,$FD,$1C,$3B,$7E,$8B,$F4,$1E,$5A
	.dc.b	$77,$FC,$87,$F8,$74,$63,$2E,$9F,$9B,$63,$1F,$2C,$77,$F9,$4F,$27
	.dc.b	$5F,$EB,$6F,$86,$A9,$FE,$09,$C3,$EF,$D7,$FC,$BB,$8F,$2A,$79,$BD
	.dc.b	$78,$3D,$70,$F3,$B1,$BF,$C3,$F7,$7B,$E5,$CF,$4E,$F1,$DB,$3F,$97
	.dc.b	$9F,$5F,$85,$E3,$BF,$C5,$F9,$9F,$F2,$66,$3E,$3F,$BE,$7E,$88,$FC
	.dc.b	$85,$F1,$EF,$CF,$3F,$CF,$F8,$7F,$8A,$7E,$83,$F8,$56,$1C,$2F,$75
	.dc.b	$AF,$D3,$3F,$CD,$39,$75,$CB,$A6,$DF,$E6,$38,$C7,$E3,$5F,$E2,$7C
	.dc.b	$B9,$D7,$87,$4F,$07,$DB,$FF,$9D,$F0,$E6,$7E,$1F,$4F,$04,$EB,$F3
	.dc.b	$AF,$ED,$7F,$7D,$FE,$9E,$F1,$5F,$87,$3F,$61,$7E,$78,$FC,$5D,$F9
	.dc.b	$87,$F6,$AC,$38,$5E,$EB,$5F,$99,$67,$EF,$5F,$93,$7F,$83,$F7,$FC
	.dc.b	$C6,$1D,$DF,$F2,$1F,$CD,$FD,$D8,$E7,$A7,$83,$C6,$F2,$E1,$FA,$9C
	.dc.b	$FB,$7C,$2F,$F5,$BF,$59,$CE,$FC,$8B,$FB,$77,$C2,$3E,$C7,$FD,$1D
	.dc.b	$F9,$9F,$F5,$7B,$DD,$6A,$7F,$D2,$5F,$4F,$9C,$39,$70,$7F,$0F,$BE
	.dc.b	$BF,$B7,$7F,$0E,$78,$FB,$2C,$FB,$7D,$B7,$8F,$0F,$14,$E7,$2F,$F3
	.dc.b	$EF,$A1,$FD,$27,$EA,$DF,$8D,$BF,$2E,$FD,$A7,$E9,$DF,$E4,$1E,$F7
	.dc.b	$EF,$5F,$DB,$6F,$75,$AF,$9F,$57,$F8,$BF,$1E,$98,$F3,$E1,$FE,$4D
	.dc.b	$87,$6F,$DF,$7F,$92,$BC,$BD,$B6,$9D,$EB,$B5,$F8,$E7,$39,$3E,$6F
	.dc.b	$FC,$11,$F8,$DB,$F2,$F7,$EE,$6F,$D4,$9F,$B5,$3F,$0D,$7E,$92,$FD
	.dc.b	$FF,$FC,$0D,$8F,$1C,$38,$5E,$E9,$FB,$E6,$B2,$7E,$9B,$FC,$F3,$0E
	.dc.b	$1C,$1F,$D0,$5E,$5A,$78,$34,$FC,$F7,$27,$F7,$59,$CE,$7E,$3F,$6B
	.dc.b	$F6,$F8,$7F,$27,$FD,$0D,$F2,$2F,$C1,$DF,$8D,$BE,$9D,$F9,$0B,$F6
	.dc.b	$27,$E7,$5F,$AA,$C9,$97,$2C,$78,$E1,$C2,$DF,$B5,$6E,$B5,$7F,$B9
	.dc.b	$3C,$59,$BF,$72,$FD,$D7,$BF,$FC,$76,$BA,$6B,$E3,$D7,$FB,$BF,$BF
	.dc.b	$FE,$11,$9C,$E9,$E6,$FC,$5B,$FC,$79,$F0,$6F,$B2,$7E,$74,$FC,$A9
	.dc.b	$FB,$2F,$F4,$1F,$D9,$FF,$4D,$D3,$FE,$1B,$C3,$85,$EE,$B5,$E9,$7F
	.dc.b	$8F,$67,$A7,$E7,$9F,$DB,$9D,$F2,$7F,$73,$78,$E5,$DB,$C6,$7B,$5E
	.dc.b	$4D,$7C,$73,$9F,$95,$C6,$9D,$ED,$F4,$FF,$E6,$5D,$BC,$98,$70,$DB
	.dc.b	$FA,$C4,$BF,$21,$F2,$4E,$6F,$FA,$C7,$E7,$58,$C7,$A7,$CA,$FE,$33
	.dc.b	$2F,$F1,$C7,$E2,$1F,$EC,$7F,$C5,$3F,$D0,$DF,$10,$FD,$CB,$FB,$A7
	.dc.b	$EE,$FF,$E5,$BA,$D7,$DF,$FF,$1E,$9F,$EF,$1F,$B5,$67,$E5,$98,$FF
	.dc.b	$84,$2F,$FD,$A5,$F2,$F6,$73,$87,$1F,$A3,$FF,$38,$57,$4C,$FC,$1F
	.dc.b	$9B,$62,$F7,$1E,$3E,$5F,$F6,$06,$DE,$89,$CE,$BF,$6B,$F2,$7F,$16
	.dc.b	$BE,$A8,$7E,$44,$FD,$BB,$3B,$7A,$63,$F6,$F8,$BE,$43,$FB,$57,$E2
	.dc.b	$FF,$CF,$3E,$CD,$F9,$DB,$47,$E8,$5F,$DC,$7E,$BB,$FD,$4F,$7B,$AD
	.dc.b	$4F,$F7,$5F,$85,$CF,$CC,$7F,$91,$FC,$5F,$E7,$67,$F8,$56,$FF,$A4
	.dc.b	$FE,$97,$FD,$06,$11,$CF,$2E,$FF,$51,$FE,$BC,$8C,$63,$FD,$E6,$BA
	.dc.b	$E9,$E6,$BD,$FE,$9D,$8D,$3C,$78,$BF,$57,$C5,$7A,$BE,$AB,$FC,$FD
	.dc.b	$C3,$F0,$72,$F2,$6D,$FE,$5D,$C7,$D7,$4F,$AC,$E3,$AB,$E6,$3F,$C5
	.dc.b	$D8,$BD,$FF,$F0,$97,$96,$FB,$D7,$E7,$7F,$E6,$7F,$63,$FD,$1D,$4F
	.dc.b	$E0,$06,$1C,$77,$FF,$1C,$63,$C7,$0E,$1F,$7B,$D7,$F7,$FD,$E3,$DB
	.dc.b	$3F,$AB,$87,$F8,$84,$F1,$C6,$6F,$FD,$F8,$DE,$6F,$E3,$7F,$AC,$FF
	.dc.b	$89,$F3,$F0,$5F,$F0,$3C,$3C,$F5,$FC,$83,$27,$E6,$3D,$2B,$CB,$5A
	.dc.b	$DB,$FD,$51,$F2,$5F,$8D,$C7,$9E,$DE,$9C,$B9,$4B,$A7,$D3,$8F,$2D
	.dc.b	$BD,$BF,$73,$D6,$3C,$51,$F0,$3C,$77,$CD,$F0,$9C,$E3,$E5,$97,$E5
	.dc.b	$19,$CF,$C1,$AF,$FA,$EF,$C8,$FF,$31,$7C,$03,$F4,$2F,$F2,$37,$E3
	.dc.b	$AF,$B4,$2F,$D6,$1F,$BC,$3F,$44,$7E,$A4,$FC,$A9,$97,$3E,$1F,$F3
	.dc.b	$76,$3C,$CE,$58,$71,$B7,$E7,$BC,$34,$F0,$E9,$FE,$59,$C3,$EC,$19
	.dc.b	$39,$BE,$47,$38,$F7,$7C,$F1,$E8,$D3,$CF,$A7,$E4,$E7,$91,$AC,$FC
	.dc.b	$71,$FE,$62,$8C,$C7,$9E,$78,$FE,$A3,$FB,$CC,$3B,$4E,$74,$F8,$52
	.dc.b	$F4,$38,$E5,$E1,$9F,$8D,$F2,$7F,$3B,$4F,$6E,$1E,$0F,$BA,$C6,$9E
	.dc.b	$C8,$F1,$E9,$FD,$AF,$4E,$F9,$73,$D7,$E9,$7E,$61,$DA,$FD,$3E,$87
	.dc.b	$2C,$7D,$98,$FC,$0F,$9F,$FB,$BE,$9F,$FA,$07,$F6,$0F,$F3,$67,$EB
	.dc.b	$DF,$E5,$4F,$AA,$7E,$EE,$FC,$DB,$F7,$1C,$1F,$BF,$CD,$87,$0F,$4B
	.dc.b	$FC,$8F,$7B,$FC,$57,$EF,$E7,$D7,$F0,$06,$9F,$E6,$D8,$C7,$EC,$17
	.dc.b	$2F,$F5,$86,$5C,$F5,$F0,$FE,$89,$E6,$9C,$E3,$CB,$4E,$FF,$37,$99
	.dc.b	$F1,$CE,$6F,$D3,$F7,$CF,$77,$C7,$5C,$B5,$F7,$6F,$FE,$8C,$F4,$33
	.dc.b	$3C,$A7,$3F,$63,$FE,$31,$E2,$F8,$1B,$F9,$70,$4B,$F0,$0F,$F0,$D6
	.dc.b	$2F,$D9,$DF,$A3,$7F,$1B,$FD,$39,$F9,$FF,$F6,$5F,$C3,$37,$5A,$F5
	.dc.b	$95,$AF,$88,$7E,$E5,$AF,$63,$FC,$37,$C5,$FE,$D5,$D3,$B6,$98,$84
	.dc.b	$63,$1E,$7A,$3F,$2A,$FF,$1B,$D6,$B4,$F0,$69,$E4,$8C,$E1,$C6,$3C
	.dc.b	$FE,$EB,$F5,$DF,$3E,$7E,$BC,$92,$FD,$E3,$3A,$78,$3E,$01,$FA,$46
	.dc.b	$32,$F8,$31,$FA,$7F,$A7,$FF,$68,$7D,$B7,$F9,$8F,$91,$F9,$F3,$D3
	.dc.b	$7E,$D0,$FC,$61,$F0,$BF,$E4,$7C,$38,$5E,$EB,$5F,$C0,$19,$F3,$3F
	.dc.b	$DE,$BC,$DF,$E4,$51,$7F,$E7,$1E,$AF,$90,$F2,$CF,$BF,$E8,$1F,$1A
	.dc.b	$31,$8B,$F4,$9C,$63,$FF,$00,$33,$F0,$7E,$28,$FC,$AF,$C1,$C1,$EA
	.dc.b	$CF,$D3,$2F,$C8,$33,$7C,$3F,$10,$B5,$AF,$85,$FF,$31,$7E,$D9,$FC
	.dc.b	$9A,$5F,$8B,$3F,$6D,$7E,$6D,$FE,$3C,$F4,$B7,$E6,$2F,$75,$AF,$AC
	.dc.b	$7E,$59,$AF,$CA,$7F,$0D,$9F,$5E,$9F,$88,$F7,$D3,$B6,$9F,$E4,$E9
	.dc.b	$EB,$97,$48,$7D,$7F,$E2,$AF,$86,$7D,$FF,$57,$DE,$3C,$B4,$7E,$B3
	.dc.b	$8A,$F4,$FD,$B7,$E2,$F7,$78,$2D,$A6,$7D,$35,$5B,$B7,$FD,$0D,$F0
	.dc.b	$8F,$CF,$5F,$8F,$3F,$30,$7E,$C1,$FE,$8F,$F4,$1F,$BA,$F0,$BD,$D6
	.dc.b	$BE,$F7,$F4,$FF,$BA,$E7,$1F,$D5,$65,$D7,$F0,$CB,$87,$0C,$FF,$D3
	.dc.b	$5F,$97,$7E,$AA,$BC,$57,$E1,$9F,$F4,$F7,$4E,$6F,$86,$5F,$B4,$F5
	.dc.b	$3B,$69,$E1,$DB,$FC,$33,$CB,$F0,$B9,$3E,$16,$9D,$2B,$5E,$3F,$0D
	.dc.b	$3D,$9F,$94,$FF,$A8,$7F,$58,$FF,$7A,$EC,$FE,$00,$7E,$CB,$F8,$6F
	.dc.b	$C0,$35,$F0,$DF,$62,$73,$D4,$FC,$93,$97,$EB,$BF,$C2,$7F,$07,$67
	.dc.b	$FF,$00,$38,$7F,$51,$18,$C7,$96,$7F,$92,$9E,$BA,$78,$61,$D1,$FC
	.dc.b	$3A,$F8,$4E,$73,$FE,$65,$C7,$FE,$AF,$CB,$D3,$AF,$93,$E8,$B9,$BE
	.dc.b	$6F,$B6,$B9,$FE,$83,$F1,$BF,$F6,$37,$D4,$FE,$07,$E1,$DF,$B7,$3F
	.dc.b	$7B,$FF,$73,$FE,$58,$FC,$A7,$7E,$07,$F3,$5A,$D7,$E2,$6A,$7E,$53
	.dc.b	$AC,$7F,$95,$65,$F8,$DE,$BB,$8F,$1C,$FF,$DE,$7F,$73,$FC,$AA,$BC
	.dc.b	$59,$F4,$FC,$B3,$17,$BD,$3F,$C4,$BF,$86,$75,$AF,$0F,$D0,$3B,$31
	.dc.b	$9F,$FB,$FA,$3F,$8E,$F3,$FE,$0F,$B1,$A7,$6A,$D7,$C1,$FF,$B9,$FE
	.dc.b	$33,$F6,$7F,$BB,$BF,$79,$78,$2F,$CA,$BF,$09,$F9,$D6,$FC,$15,$EE
	.dc.b	$B5,$DD,$FC,$BB,$6F,$06,$5D,$36,$FE,$50,$FF,$00,$FD,$F3,$FA,$73
	.dc.b	$E0,$D3,$BE,$8F,$D5,$FF,$A6,$CE,$72,$F1,$73,$E3,$5F,$0E,$1C,$3F
	.dc.b	$54,$FF,$6A,$E9,$E0,$9C,$DF,$A6,$7F,$80,$CE,$76,$F4,$66,$97,$EE
	.dc.b	$9E,$D9,$CC,$7F,$5F,$AF,$95,$FB,$4F,$F9,$7B,$EA,$DF,$97,$F1,$FD
	.dc.b	$01,$FA,$E7,$F3,$9B,$DF,$DD,$7F,$96,$6B,$52,$FD,$83,$FC,$B3,$EF
	.dc.b	$FF,$17,$E6,$67,$FE,$6A,$FD,$AB,$FB,$93,$DE,$3F,$A8,$3E,$1C,$3C
	.dc.b	$1F,$A4,$FF,$72,$8C,$6C,$D5,$FA,$5E,$75,$F1,$4E,$7B,$7F,$5B,$BF
	.dc.b	$9B,$83,$D7,$39,$F3,$BC,$BA,$FF,$2B,$CF,$AE,$7F,$A5,$FD,$DF,$33
	.dc.b	$E8,$CF,$AC,$FE,$A7,$F5,$4F,$F1,$0F,$E0,$DF,$DE,$FC,$5F,$E6,$9F
	.dc.b	$93,$7F,$9C,$BF,$5B,$7E,$2D,$AF,$A3,$FD,$8E,$6A,$D7,$EB,$99,$C3
	.dc.b	$FC,$79,$3F,$EB,$DA,$7E,$B5,$8C,$BB,$4F,$1F,$C0,$7F,$14,$F1,$31
	.dc.b	$E1,$FF,$8F,$67,$79,$7F,$A1,$3F,$6B,$7E,$2B,$B8,$C7,$DE,$DC,$5F
	.dc.b	$94,$BF,$7E,$C2,$FC,$23,$FC,$7F,$FA,$D3,$F3,$97,$E8,$EF,$CD,$5F
	.dc.b	$A7,$FF,$B8,$7E,$9F,$FD,$45,$E0,$FF,$98,$70,$E1,$7B,$AD,$7E,$5D
	.dc.b	$D3,$1E,$7C,$9F,$E9,$2F,$C0,$FF,$11,$87,$7D,$3E,$2B,$CB,$4E,$FF
	.dc.b	$8E,$7F,$81,$EF,$96,$7D,$61,$F2,$9F,$E5,$FB,$E1,$A7,$97,$F6,$9D
	.dc.b	$E3,$CB,$77,$E5,$B8,$AF,$F7,$D7,$D5,$3E,$6B,$19,$FF,$85,$7F,$38
	.dc.b	$FF,$5A,$FE,$33,$F9,$DE,$1F,$EF,$3C,$8F,$86,$FE,$DB,$FB,$0F,$F8
	.dc.b	$3F,$E9,$B7,$C9,$FD,$E6,$F7,$5A,$FA,$77,$E6,$7A,$A5,$7F,$81,$3F
	.dc.b	$0F,$FF,$5C,$C6,$63,$BB,$FE,$B8,$CF,$C5,$F4,$FF,$EC,$8C,$7C,$3F
	.dc.b	$8D,$7F,$B5,$27,$3A,$F9,$7E,$71,$B7,$DD,$C7,$A6,$3C,$67,$3F,$9F
	.dc.b	$FB,$30,$E3,$97,$B2,$1B,$F0,$CB,$A6,$4F,$B4,$F0,$67,$EF,$CF,$A4
	.dc.b	$BC,$3C,$CF,$BA,$73,$E7,$7E,$33,$06,$9E,$0D,$7C,$F3,$E3,$BE,$95
	.dc.b	$E0,$F8,$6E,$7C,$1F,$32,$FD,$BE,$57,$FB,$FF,$91,$FB,$A3,$ED,$7F
	.dc.b	$DF,$FB,$BF,$3E,$FF,$79,$7A,$1F,$F0,$06,$F7,$5A,$9F,$F7,$B7,$D1
	.dc.b	$E7,$0F,$F3,$3E,$DF,$C7,$4C,$7D,$CF,$D6,$FC,$11,$E8,$FF,$21,$CB
	.dc.b	$9F,$E3,$9F,$43,$2F,$F7,$35,$6B,$0F,$F1,$E5,$6B,$5F,$27,$E3,$FE
	.dc.b	$9C,$FD,$18,$F1,$EB,$FF,$33,$63,$C7,$1F,$F2,$E5,$6B,$0F,$65,$3F
	.dc.b	$50,$F4,$6F,$EB,$E1,$E8,$FB,$7E,$57,$FC,$8D,$97,$3D,$7C,$5B,$25
	.dc.b	$FB,$62,$BF,$DB,$9D,$7F,$CC,$FF,$0E,$7E,$83,$3D,$77,$EE,$6F,$CE
	.dc.b	$F8,$75,$FF,$43,$C3,$85,$EE,$9F,$84,$6B,$97,$FC,$BD,$5F,$CF,$FE
	.dc.b	$C3,$E6,$61,$C7,$E0,$BF,$DA,$D5,$F4,$4E,$07,$E3,$5F,$F6,$D6,$3E
	.dc.b	$3F,$AC,$FF,$34,$6D,$E8,$9C,$CF,$F2,$4F,$AB,$CF,$C8,$7F,$57,$FC
	.dc.b	$8B,$FC,$D7,$AF,$FB,$8F,$AF,$F9,$9F,$D1,$FF,$2A,$CB,$FC,$21,$5A
	.dc.b	$E3,$FC,$6F,$C1,$FF,$9B,$F8,$7E,$36,$5C,$FE,$DF,$E1,$70,$79,$F1
	.dc.b	$E3,$1F,$02,$31,$3B,$D3,$C1,$83,$F1,$6F,$2D,$7F,$A0,$73,$F0,$6B
	.dc.b	$FD,$1B,$F1,$6F,$F6,$BF,$EC,$DF,$E7,$FF,$C6,$BF,$C4,$7F,$1E,$FD
	.dc.b	$71,$F2,$6F,$D7,$3F,$AE,$DE,$FE,$9F,$FB,$AD,$6B,$F2,$CC,$F3,$3F
	.dc.b	$68,$FF,$74,$5F,$5D,$FF,$D8,$CF,$1C,$FB,$79,$1E,$2B,$6B,$E2,$C9
	.dc.b	$E5,$FF,$BB,$3F,$79,$7B,$2F,$C1,$DF,$AD,$FF,$AC,$BE,$73,$F8,$0F
	.dc.b	$A3,$FD,$CA,$F7,$5A,$FC,$23,$FC,$03,$ED,$7F,$26,$F6,$7F,$C8,$55
	.dc.b	$AE,$2F,$F6,$D3,$DB,$F2,$5F,$F8,$00,$F2,$CF,$B5,$F8,$F4,$F1,$47
	.dc.b	$FB,$0B,$F4,$DF,$F0,$C6,$3D,$27,$39,$7F,$84,$3E,$47,$FD,$F7,$39
	.dc.b	$D3,$C9,$B2,$5F,$8D,$7A,$70,$E3,$39,$FD,$B7,$17,$C2,$FF,$D6,$38
	.dc.b	$7F,$96,$FF,$80,$7F,$C1,$3F,$A3,$BF,$5F,$7E,$AE,$FD,$B5,$F8,$7F
	.dc.b	$FA,$6F,$0E,$1F,$42,$C3,$85,$EE,$BF,$C8,$55,$AF,$D3,$FF,$A5,$BC
	.dc.b	$2A,$F9,$7F,$E3,$7C,$FC,$39,$F8,$33,$FF,$6F,$5E,$E3,$1F,$65,$FE
	.dc.b	$D3,$8C,$67,$DF,$C5,$FF,$14,$E1,$CA,$3C,$5E,$27,$F8,$B3,$EC,$5F
	.dc.b	$89,$3F,$39,$9C,$CF,$D8,$FF,$0F,$7B,$AD,$57,$E2,$27,$F1,$79,$66
	.dc.b	$3D,$3D,$FF,$EE,$5C,$7A,$FE,$79,$88,$FF,$3F,$61,$CB,$2E,$DE,$17
	.dc.b	$A7,$8B,$FC,$AD,$87,$93,$1F,$1E,$7F,$DE,$F1,$ED,$1E,$1B,$7E,$EF
	.dc.b	$8C,$79,$FE,$23,$FE,$48,$C3,$FB,$6F,$E5,$BF,$E0,$9F,$C0,$DF,$60
	.dc.b	$FC,$E9,$F8,$D6,$FE,$35,$7B,$AD,$7E,$65,$98,$F5,$63,$FD,$E3,$5B
	.dc.b	$7C,$3E,$79,$9F,$6F,$43,$D9,$C2,$3E,$65,$6E,$7C,$77,$A9,$CE,$3E
	.dc.b	$1D,$5F,$A3,$F8,$72,$EB,$AF,$A7,$AF,$19,$7B,$2D,$FA,$76,$2B,$DD
	.dc.b	$D7,$FE,$6A,$F3,$1F,$27,$FE,$B1,$F9,$A7,$E7,$2F,$D2,$3F,$D0,$BF
	.dc.b	$9D,$EE,$B5,$39,$F6,$1F,$47,$9F,$D6,$3E,$E6,$5D,$3F,$04,$B8,$70
	.dc.b	$CB,$FC,$CB,$C3,$FE,$C5,$AD,$69,$E0,$97,$57,$C1,$9C,$E9,$DF,$CB
	.dc.b	$F1,$65,$FE,$1B,$8F,$17,$97,$FE,$84,$F6,$1F,$9B,$3F,$62,$7E,$8B
	.dc.b	$FC,$83,$E8,$7F,$6F,$F8,$56,$EB,$53,$9E,$6F,$D9,$F9,$3F,$EA,$5C
	.dc.b	$79,$E3,$F9,$77,$AF,$87,$70,$E3,$EF,$FF,$B2,$7E,$37,$EC,$57,$0C
	.dc.b	$BC,$1F,$4F,$8B,$DE,$BF,$C2,$3A,$BC,$59,$F9,$B2,$FF,$69,$6A,$FD
	.dc.b	$77,$FD,$3F,$F8,$CB,$E9,$1F,$76,$FB,$DF,$F5,$7F,$EC,$4F,$CE,$FF
	.dc.b	$98,$72,$7F,$AF,$70,$E1,$7B,$F8,$15,$EF,$8F,$AE,$4F,$AA,$FD,$A6
	.dc.b	$1C,$3C,$CF,$EF,$5E,$0F,$FB,$EE,$39,$EB,$E0,$FB,$7F,$CB,$46,$33
	.dc.b	$EB,$F0,$F8,$CB,$CF,$7E,$6D,$DB,$F8,$37,$F2,$6F,$E9,$CB,$DF,$F3
	.dc.b	$BF,$E9,$8F,$DD,$5F,$9B,$7D,$AF,$A9,$7C,$7E,$BE,$D1,$5A,$EE,$57
	.dc.b	$F8,$93,$8F,$D1,$8F,$FA,$8B,$93,$D7,$B7,$C5,$87,$E1,$FF,$01,$F0
	.dc.b	$69,$DE,$BB,$65,$D7,$B7,$F6,$98,$7F,$26,$FD,$3B,$FD,$E3,$F9,$D3
	.dc.b	$F5,$57,$EE,$4F,$DF,$BF,$DA,$F7,$BF,$C4,$57,$BA,$BB,$FC,$4D,$3F
	.dc.b	$82,$5F,$57,$CE,$63,$E1,$7E,$1B,$83,$E5,$AC,$FC,$79,$BF,$4C,$F3
	.dc.b	$CE,$73,$F0,$CF,$F8,$66,$FC,$B3,$9C,$FC,$9F,$BF,$63,$1E,$7A,$3F
	.dc.b	$74,$FD,$77,$EE,$4F,$D0,$5F,$9D,$7F,$81,$BE,$7D,$FA,$03,$F3,$7D
	.dc.b	$D6,$BC,$4F,$F0,$C4,$BF,$76,$CF,$07,$AF,$C9,$66,$FD,$07,$F6,$0F
	.dc.b	$7D,$3B,$61,$FA,$E8,$C6,$3F,$E2,$3F,$97,$C6,$7D,$F5,$F1,$FE,$69
	.dc.b	$E2,$C7,$8F,$ED,$F7,$0F,$1F,$E2,$96,$FD,$3F,$48,$EC,$FA,$9C,$0F
	.dc.b	$DC,$9F,$92,$3F,$5D,$7D,$07,$F3,$EF,$C3,$5F,$AD,$FF,$A9,$FC,$2F
	.dc.b	$E7,$D8,$F1,$C3,$85,$EE,$BF,$DA,$B5,$AC,$7A,$FE,$19,$FD,$95,$EF
	.dc.b	$95,$C3,$8F,$C8,$33,$7D,$C7,$F4,$7B,$DE,$7E,$3F,$BB,$C6,$1C,$2F
	.dc.b	$CB,$FA,$97,$B7,$4F,$04,$E7,$57,$E7,$99,$DF,$FA,$B7,$BE,$FC,$3D
	.dc.b	$F9,$CB,$F8,$03,$FD,$87,$F9,$C7,$EF,$3A,$F8,$61,$FC,$C6,$F7,$5A
	.dc.b	$D3,$FC,$6A,$7C,$4C,$7A,$63,$FE,$03,$7B,$7D,$05,$BE,$3F,$1E,$FD
	.dc.b	$CB,$FC,$83,$87,$19,$CE,$9F,$ED,$4C,$FC,$38,$F4,$CF,$C1,$18,$AD
	.dc.b	$DB,$F5,$AF,$3F,$E3,$BF,$EA,$9F,$BC,$FF,$1B,$7D,$DF,$FB,$47,$CC
	.dc.b	$7E,$F9,$AF,$C0,$95,$AF,$C3,$3F,$49,$FB,$F6,$63,$FC,$39,$8F,$FA
	.dc.b	$CF,$EE,$ED,$F0,$F9,$66,$7F,$ED,$DF,$8B,$FE,$FD,$7E,$19,$CE,$4F
	.dc.b	$C4,$73,$87,$1C,$3C,$9F,$83,$77,$BD,$EC,$FC,$27,$15,$E9,$E4,$B9
	.dc.b	$7E,$E7,$EF,$FE,$AF,$EC,$7F,$C9,$BE,$9B,$FC,$C9,$F3,$2F,$CA,$9F
	.dc.b	$83,$FE,$FF,$CE,$FD,$0D,$EF,$EF,$57,$BA,$7C,$07,$4F,$8C,$FF,$60
	.dc.b	$F7,$7F,$BC,$71,$E3,$1F,$E2,$E8,$FE,$31,$8F,$3C,$3C,$1E,$AF,$F8
	.dc.b	$7E,$31,$9B,$E5,$B1,$87,$F8,$DB,$4F,$0C,$F9,$BF,$6E,$F0,$6D,$EA
	.dc.b	$8F,$4F,$EC,$BF,$92,$FD,$A3,$E2,$7E,$F6,$FC,$99,$F7,$FF,$E3,$EF
	.dc.b	$C9,$FF,$6D,$C3,$FE,$33,$BD,$D6,$BE,$E5,$5F,$B6,$AD,$75,$31,$FF
	.dc.b	$2D,$61,$F4,$7E,$EA,$CE,$71,$E5,$EC,$FF,$8E,$EB,$AE,$7F,$E4,$EF
	.dc.b	$5F,$17,$B8,$F2,$7E,$5D,$E9,$8C,$6B,$E2,$FD,$FF,$F5,$BF,$8B,$3F
	.dc.b	$1F,$63,$FE,$07,$FD,$53,$F6,$DF,$8B,$B9,$B7,$5A,$FA,$75,$6B,$E5
	.dc.b	$B3,$D1,$FE,$73,$CB,$A6,$57,$0E,$17,$FC,$06,$3F,$CA,$93,$AC,$39
	.dc.b	$FA,$7F,$C4,$B9,$BB,$D7,$8E,$1A,$7D,$1F,$03,$F9,$5E,$1C,$71,$F3
	.dc.b	$F7,$F0,$CB,$9D,$BC,$AC,$57,$B3,$F4,$5F,$E1,$BC,$4F,$C8,$FF,$99
	.dc.b	$77,$7F,$AC,$F9,$BF,$CA,$7F,$27,$A9,$CF,$E1,$49,$FF,$2B,$FC,$A6
	.dc.b	$7E,$27,$FD,$25,$8F,$3F,$89,$B8,$70,$BF,$DB,$F2,$F3,$9F,$D3,$63
	.dc.b	$CF,$93,$FC,$ED,$87,$1D,$3C,$5F,$75,$FB,$9E,$5F,$E9,$33,$9B,$F1
	.dc.b	$4B,$F2,$DF,$E4,$FF,$8D,$FF,$BC,$BF,$73,$FE,$1F,$CE,$FD,$9D,$F9
	.dc.b	$8B,$F2,$8D,$FE,$0E,$BD,$D6,$BE,$07,$3F,$B9,$7F,$AB,$72,$E9,$C7
	.dc.b	$FE,$07,$7B,$7C,$17,$FB,$F2,$FF,$CE,$7F,$07,$FD,$5A,$73,$AF,$87
	.dc.b	$D1,$FF,$49,$46,$2B,$FC,$8D,$FC,$01,$E2,$D3,$C3,$F6,$8F,$C6,$7F
	.dc.b	$A0,$FE,$A6,$FD,$79,$FA,$2F,$FA,$B3,$E2,$1F,$A9,$FF,$A9,$38,$F7
	.dc.b	$5A,$EA,$FE,$5B,$3F,$EA,$8E,$3C,$FE,$91,$FE,$C7,$CB,$A6,$57,$0E
	.dc.b	$19,$74,$F8,$DF,$F5,$45,$6B,$2E,$7F,$3F,$8B,$DE,$6F,$D1,$1F,$90
	.dc.b	$78,$B3,$F3,$4E,$7F,$6B,$C6,$3E,$4B,$FE,$0F,$B3,$3E,$DA,$FB,$25
	.dc.b	$F9,$B7,$F6,$2F,$C7,$1F,$A6,$3F,$05,$FF,$13,$FE,$25,$FE,$94,$FD
	.dc.b	$0D,$FA,$5E,$EB,$5F,$3B,$FE,$D3,$FB,$C5,$77,$F2,$BF,$C2,$79,$74
	.dc.b	$CB,$FD,$D4,$F6,$FC,$22,$E3,$CB,$3F,$F5,$0F,$EC,$9F,$EC,$BA,$E1
	.dc.b	$9F,$5F,$AE,$C5,$EF,$4E,$FF,$B3,$F8,$F3,$F3,$C7,$8B,$F2,$2F,$F1
	.dc.b	$27,$DF,$FF,$87,$3F,$53,$FF,$2F,$FE,$54,$FD,$3D,$FA,$F7,$F9,$7B
	.dc.b	$F2,$87,$E4,$9C,$3A,$FF,$DC,$18,$70,$BD,$FD,$F2,$DF,$B5,$EE,$D7
	.dc.b	$FE,$6C,$FC,$EF,$FC,$57,$1F,$E2,$CE,$86,$BE,$3F,$BF,$6D,$AA,$FF
	.dc.b	$77,$56,$9F,$0C,$EB,$4F,$0F,$26,$B0,$E5,$87,$AF,$7F,$F3,$BF,$DE
	.dc.b	$E7,$1E,$33,$9D,$FD,$1F,$B7,$F9,$F6,$F6,$61,$F0,$7E,$5B,$D5,$FB
	.dc.b	$9F,$F3,$6F,$9E,$F2,$FE,$55,$FE,$B9,$FD,$4D,$F5,$1A,$F9,$1F,$ED
	.dc.b	$BA,$AB,$5F,$8E,$67,$F4,$AF,$67,$06,$F2,$BC,$DE,$D8,$69,$DF,$1E
	.dc.b	$5F,$8A,$BD,$3F,$55,$6B,$4E,$FA,$7F,$B2,$FE,$77,$38,$F1,$8F,$37
	.dc.b	$8B,$DF,$2E,$7C,$1F,$EB,$0D,$3C,$17,$BD,$3F,$C6,$DC,$5B,$C7,$FD
	.dc.b	$AD,$F6,$6F,$D5,$3E,$E7,$EC,$DF,$D1,$3E,$01,$FA,$23,$85,$F6,$5F
	.dc.b	$AD,$FB,$5D,$FC,$CF,$FA,$F2,$B5,$3F,$B1,$FC,$FB,$3F,$9C,$7C,$2F
	.dc.b	$07,$FC,$DF,$9F,$5C,$BF,$DB,$31,$8C,$79,$47,$6F,$8B,$F8,$75,$E3
	.dc.b	$CB,$AF,$E5,$D8,$BD,$E7,$FB,$FF,$98,$F1,$ED,$E8,$CB,$CF,$ED,$7F
	.dc.b	$A7,$F9,$9F,$86,$7F,$A0,$3D,$97,$86,$F8,$3F,$E8,$1F,$B5,$BF,$13
	.dc.b	$7D,$4F,$E8,$71,$E3,$9F,$C3,$C7,$8E,$1C,$2F,$7F,$AB,$FF,$A5,$73
	.dc.b	$ED,$9F,$5C,$3F,$DE,$F1,$8F,$57,$F0,$AE,$B0,$E5,$F8,$C7,$E9,$23
	.dc.b	$1F,$2C,$DB,$C9,$F9,$46,$71,$F1,$65,$E9,$AD,$68,$FD,$41,$FB,$BF
	.dc.b	$8E,$F7,$A7,$9E,$58,$E6,$5F,$99,$7E,$E3,$E3,$FF,$09,$F8,$9F,$FA
	.dc.b	$2F,$E0,$DF,$3B,$BA,$D7,$89,$FE,$C4,$FB,$C5,$7F,$A4,$B9,$FE,$3E
	.dc.b	$5D,$3C,$27,$CA,$FC,$08,$FC,$66,$3D,$DF,$8A,$62,$7F,$AC,$63,$D7
	.dc.b	$C0,$F8,$13,$CF,$5F,$2E,$6F,$0F,$38,$F9,$36,$F4,$7D,$DF,$9F,$3E
	.dc.b	$71,$EB,$BC,$61,$CB,$6F,$F3,$67,$47,$FB,$13,$1E,$55,$EC,$FB,$AF
	.dc.b	$F6,$86,$DF,$EC,$8F,$49,$FC,$00,$FD,$D5,$F9,$2B,$E2,$DF,$AD,$FF
	.dc.b	$A3,$3F,$13,$E3,$C9,$FA,$BC,$78,$E1,$C2,$F7,$F8,$13,$F1,$5D,$DF
	.dc.b	$F9,$2B,$37,$EF,$7F,$C3,$35,$F1,$47,$86,$7F,$96,$75,$7E,$ED,$ED
	.dc.b	$1E,$5D,$BC,$9B,$7F,$8A,$2B,$55,$D7,$4F,$14,$FF,$B2,$E3,$C3,$AF
	.dc.b	$9A,$F8,$E8,$FD,$43,$39,$75,$DB,$D3,$FC,$01,$FE,$45,$FC,$0B,$FD
	.dc.b	$DB,$E9,$BF,$79,$7E,$67,$F2,$3E,$11,$FA,$4B,$EE,$7F,$CC,$9F,$72
	.dc.b	$C7,$8E,$1C,$34,$61,$C2,$DF,$7B,$DF,$8B,$F8,$B6,$9D,$F4,$FE,$67
	.dc.b	$9F,$39,$CE,$4F,$97,$CE,$9D,$AF,$C9,$AF,$F0,$1B,$E8,$31,$E9,$AF
	.dc.b	$8A,$5F,$99,$FF,$25,$CF,$B5,$EF,$3F,$46,$DE,$EF,$D9,$FC,$F5,$AC
	.dc.b	$B9,$CE,$7F,$00,$FF,$36,$75,$7F,$2F,$C7,$F8,$8F,$E3,$EF,$95,$7C
	.dc.b	$8B,$F2,$4F,$F6,$FF,$CD,$7C,$4D,$3E,$DF,$B8,$C3,$85,$EF,$E8,$FF
	.dc.b	$BF,$FC,$9F,$6C,$FF,$D5,$95,$F5,$5F,$EE,$98,$72,$8F,$04,$63,$4F
	.dc.b	$26,$3D,$74,$F1,$E9,$FE,$F0,$F7,$BE,$15,$6B,$0F,$0C,$FF,$BD,$3E
	.dc.b	$67,$38,$F8,$30,$E3,$A7,$F8,$4B,$F3,$3F,$24,$FC,$09,$CE,$EF,$7F
	.dc.b	$3A,$F8,$78,$3D,$92,$FD,$8F,$3A,$F8,$33,$FF,$0E,$FC,$67,$FA,$5F
	.dc.b	$CF,$F9,$9F,$9D,$FF,$3D,$FC,$AF,$F8,$FF,$EF,$BF,$8F,$FD,$C1,$FB
	.dc.b	$86,$F8,$7F,$B0,$5E,$FD,$45,$EF,$E6,$35,$4A,$FE,$91,$FB,$DF,$E4
	.dc.b	$33,$EF,$F0,$8D,$31,$7D,$23,$19,$74,$F1,$BF,$D3,$7D,$9F,$CA,$32
	.dc.b	$F0,$69,$FA,$3F,$C0,$3F,$1D,$9F,$4C,$FD,$38,$37,$7E,$81,$E4,$C7
	.dc.b	$8C,$E7,$F7,$DF,$F5,$C7,$BD,$F8,$DF,$3D,$F9,$0F,$F7,$CF,$D7,$5F
	.dc.b	$A9,$3F,$45,$FF,$5F,$7C,$57,$0E,$5F,$F0,$16,$1C,$3E,$51,$87,$0B
	.dc.b	$FF,$62,$7D,$B6,$EF,$F9,$16,$5F,$E0,$87,$F1,$B9,$F8,$63,$FC,$A3
	.dc.b	$18,$3A,$BE,$3C,$FC,$5F,$8C,$7E,$B6,$B8,$65,$E0,$9F,$37,$26,$70
	.dc.b	$E3,$39,$CB,$FD,$AD,$F8,$9F,$8F,$5F,$46,$BE,$7E,$7F,$F4,$F9,$FE
	.dc.b	$35,$FC,$F1,$F9,$33,$F3,$5F,$98,$BF,$17,$FC,$97,$D1,$7B,$6B,$75
	.dc.b	$A9,$FC,$B7,$36,$6F,$96,$DE,$D1,$7D,$BD,$FF,$E2,$BC,$5F,$BF,$8E
	.dc.b	$7D,$5E,$2C,$7F,$DF,$1A,$79,$71,$F2,$4E,$7E,$A3,$E0,$65,$D2,$72
	.dc.b	$79,$FC,$DF,$F0,$45,$F1,$D7,$CD,$0F,$AD,$F3,$46,$3E,$A6,$E5,$E1
	.dc.b	$A7,$E6,$BF,$6F,$E7,$1F,$92,$7F,$11,$FB,$77,$E2,$5F,$A8,$FD,$0F
	.dc.b	$FD,$6F,$AF,$F9,$7E,$F7,$EB,$AF,$7E,$6A,$FF,$DA,$97,$BC,$7F,$8E
	.dc.b	$70,$FC,$7C,$7B,$FE,$6B,$F1,$A3,$FC,$A5,$1D,$31,$E7,$F6,$1F,$0A
	.dc.b	$31,$A7,$EE,$BE,$E7,$18,$F9,$E3,$F7,$BF,$18,$FD,$5F,$D9,$39,$DB
	.dc.b	$FB,$CF,$E1,$FF,$99,$FC,$37,$FC,$85,$F4,$AF,$CF,$7F,$82,$FC,$AF
	.dc.b	$C9,$FC,$53,$5E,$B7,$F9,$2A,$73,$D9,$F4,$3C,$1F,$E7,$58,$F2,$E4
	.dc.b	$FC,$13,$FE,$4D,$F7,$BD,$90,$8E,$F7,$CA,$3A,$E7,$E1,$D5,$A3,$EE
	.dc.b	$BF,$03,$87,$09,$CE,$7F,$63,$A3,$FD,$21,$39,$BF,$C0,$77,$27,$C1
	.dc.b	$97,$F9,$F7,$EE,$DF,$A5,$FE,$8B,$EC,$7F,$41,$C0,$FC,$19,$F8,$5F
	.dc.b	$1D,$3F,$AB,$61,$C3,$C5,$FF,$43,$F8,$FF,$EC,$0B,$FE,$C9,$5A,$C9
	.dc.b	$F4,$BF,$DD,$23,$17,$74,$F1,$57,$87,$0F,$F6,$4F,$C4,$F2,$9C,$D7
	.dc.b	$6D,$1F,$95,$E6,$7F,$DB,$D8,$71,$9F,$47,$EB,$DE,$79,$E3,$BF,$B3
	.dc.b	$5F,$D1,$70,$7C,$29,$E3,$97,$59,$FF,$70,$4E,$77,$F7,$5F,$1F,$B1
	.dc.b	$E9,$83,$F5,$47,$DA,$BF,$74,$7E,$5E,$FB,$D7,$E7,$EF,$AB,$FF,$11
	.dc.b	$EB,$E1,$C7,$8E,$1C,$2F,$74,$FB,$CE,$F5,$DB,$F7,$EF,$91,$A7,$87
	.dc.b	$4F,$03,$FD,$EB,$A3,$FC,$15,$5D,$35,$F1,$6A,$FA,$7F,$91,$39,$D3
	.dc.b	$CB,$F1,$D8,$CB,$D5,$87,$0D,$7F,$C1,$9B,$F9,$E7,$37,$E6,$E2,$F8
	.dc.b	$7F,$4A,$FC,$A5,$F9,$D7,$F9,$9B,$F6,$A7,$EE,$7F,$C4,$7F,$09,$DD
	.dc.b	$6B,$E9,$D5,$AF,$19,$FB,$8E,$B1,$F7,$E5,$D7,$F6,$DB,$8F,$1C,$38
	.dc.b	$73,$EF,$2E,$7A,$7E,$9E,$31,$3D,$7F,$52,$B3,$FC,$8F,$ED,$3E,$DF
	.dc.b	$B2,$FD,$17,$EC,$FD,$E3,$F1,$2F,$F7,$17,$DF,$3F,$1E,$DD,$6A,$73
	.dc.b	$FB,$0E,$3F,$51,$FA,$B8,$F3,$EC,$B7,$BD,$FF,$C9,$27,$28,$FF,$52
	.dc.b	$E1,$FE,$EF,$AF,$CE,$1C,$73,$F1,$57,$FA,$CB,$D8,$E1,$18,$D1,$FA
	.dc.b	$66,$30,$F3,$FD,$2D,$C3,$FD,$73,$FA,$EF,$E6,$3D,$BF,$AD,$FC,$00
	.dc.b	$FC,$D5,$F8,$DB,$F7,$67,$E2,$8F,$C6,$B8,$76,$30,$E1,$7B,$A7,$EB
	.dc.b	$7A,$9F,$CA,$E5,$F8,$3F,$D6,$EE,$1C,$73,$EF,$9B,$E5,$3F,$75,$7E
	.dc.b	$49,$CF,$97,$8B,$E1,$A7,$7F,$D2,$7B,$57,$2D,$7F,$DB,$3E,$5F,$EE
	.dc.b	$FE,$B1,$FA,$2F,$FA,$2F,$F6,$97,$E6,$18,$FD,$C5,$FB,$E7,$EE,$AF
	.dc.b	$7F,$3C,$BD,$E0,$B7,$E7,$1B,$87,$D0,$7E,$CB,$0E,$3D,$5F,$B1,$01
	.dc.b	$EB,$87,$4F,$43,$F8,$C4,$63,$4F,$3D,$63,$0F,$44,$79,$FF,$40,$F0
	.dc.b	$69,$E5,$8F,$57,$EB,$FF,$C4,$BF,$37,$FF,$78,$FC,$7B,$F7,$8F,$A7
	.dc.b	$F8,$D3,$F6,$47,$F0,$06,$FF,$15,$FE,$95,$5A,$97,$EC,$39,$C3,$F8
	.dc.b	$6E,$3F,$EF,$CF,$D6,$AE,$1C,$3C,$46,$6F,$C1,$7F,$99,$5F,$8B,$3F
	.dc.b	$0F,$EF,$58,$C3,$85,$F8,$FF,$56,$ED,$5E,$1F,$7E,$E5,$D3,$6E,$BD
	.dc.b	$BF,$CD,$73,$E9,$39,$FD,$2F,$F9,$8F,$E2,$0F,$DD,$BF,$A1,$FE,$F5
	.dc.b	$FD,$13,$E7,$7F,$C2,$FF,$B8,$6A,$73,$F3,$1F,$E7,$EF,$A3,$7E,$47
	.dc.b	$98,$FE,$57,$8F,$3E,$CB,$7C,$31,$7D,$4F,$F4,$FA,$F0,$5B,$3F,$07
	.dc.b	$E4,$79,$C3,$86,$6F,$C2,$3B,$57,$96,$3F,$C5,$7C,$3F,$D4,$7F,$37
	.dc.b	$FD,$77,$63,$F2,$B7,$13,$F7,$AF,$F7,$87,$E5,$5F,$D4,$2F,$75,$A9
	.dc.b	$FF,$7A,$7C,$8E,$7D,$A7,$83,$FD,$1A,$FA,$F9,$1F,$95,$0F,$2C,$BB
	.dc.b	$7B,$3F,$DA,$E3,$18,$FF,$8D,$35,$F3,$63,$FE,$11,$CF,$C1,$F8,$F7
	.dc.b	$F2,$39,$CE,$DE,$9C,$BD,$12,$F8,$7F,$EC,$FE,$7F,$F8,$2B,$F8,$03
	.dc.b	$DD,$FA,$D3,$F7,$67,$E0,$AF,$8B,$DF,$D5,$2F,$75,$AF,$8B,$CF,$EB
	.dc.b	$1F,$D8,$B2,$E9,$CB,$ED,$7B,$70,$B4,$ED,$A7,$F8,$1E,$73,$1E,$1F
	.dc.b	$D0,$B1,$A7,$7B,$E1,$9F,$93,$EB,$BC,$7A,$F8,$B2,$FF,$60,$71,$FF
	.dc.b	$82,$73,$E7,$7E,$AF,$8E,$DE,$3C,$AD,$F9,$AF,$F8,$23,$EA,$9F,$9E
	.dc.b	$3E,$AD,$F8,$F3,$F5,$97,$EF,$BF,$DA,$3F,$49,$E1,$7B,$AD,$4F,$F6
	.dc.b	$CF,$D3,$F3,$F0,$5F,$E0,$4C,$BA,$E5,$70,$E1,$97,$4E,$5F,$F1,$45
	.dc.b	$6B,$2E,$7F,$B9,$E2,$F7,$9F,$F7,$8F,$D4,$FF,$C3,$FF,$56,$F9,$AF
	.dc.b	$AD,$FA,$F3,$F3,$7F,$E4,$7F,$BE,$FF,$8B,$2B,$FC,$C3,$5A,$9C,$EF
	.dc.b	$CA,$3F,$DD,$1D,$DF,$93,$AE,$9E,$0F,$F8,$71,$B8,$F5,$EA,$7E,$1D
	.dc.b	$FE,$3F,$9C,$E5,$FC,$CB,$EB,$BF,$C8,$5F,$40,$FD,$21,$F0,$9F,$E2
	.dc.b	$9F,$DC,$FF,$9D,$75,$3F,$70,$FF,$1C,$DE,$EB,$5F,$24,$AF,$F4,$07
	.dc.b	$87,$FE,$45,$AE,$9F,$9F,$7F,$8C,$1E,$1C,$5F,$AB,$79,$7D,$AB,$3F
	.dc.b	$E9,$1F,$15,$E4,$D7,$C7,$39,$CF,$F8,$5E,$1E,$2F,$C7,$DF,$A2,$BE
	.dc.b	$A5,$1F,$E4,$7F,$41,$E1,$6E,$B5,$39,$FD,$63,$18,$78,$E1,$F6,$7E
	.dc.b	$77,$AE,$3D,$3E,$D5,$EE,$FC,$89,$C9,$FE,$55,$EE,$F9,$7C,$4F,$AB
	.dc.b	$FF,$4B,$7B,$8F,$D4,$FF,$86,$F4,$7C,$AA,$D7,$6A,$B5,$EF,$3F,$36
	.dc.b	$D7,$99,$F2,$6F,$A7,$0F,$C9,$78,$DF,$5F,$73,$C3,$A7,$83,$47,$E0
	.dc.b	$9E,$4B,$FE,$A1,$F8,$0B,$E4,$BF,$D2,$DF,$37,$E3,$9C,$C7,$FA,$67
	.dc.b	$E6,$B1,$F3,$1F,$E9,$E8,$7E,$05,$CE,$31,$F1,$BD,$6D,$FB,$AB,$F7
	.dc.b	$EE,$DF,$93,$75,$9C,$C7,$87,$F4,$AC,$56,$FE,$23,$93,$C7,$F5,$DE
	.dc.b	$96,$F9,$DF,$E3,$DF,$A1,$73,$8C,$7A,$AF,$DC,$71,$4F,$84,$E2,$FF
	.dc.b	$92,$3D,$0F,$CA,$5F,$F9,$5C,$FF,$2C,$77,$BF,$80,$1F,$17,$D6,$31
	.dc.b	$F1,$17,$5F,$1C,$3E,$03,$CB,$83,$F3,$3F,$E6,$19,$76,$8E,$C6,$7F
	.dc.b	$80,$7D,$0F,$F9,$9F,$F4,$3A,$FF,$4C,$9F,$D4,$7E,$F8,$FC,$CD,$FA
	.dc.b	$5B,$D1,$78,$1F,$92,$AF,$D9,$9B,$9D,$47,$E5,$BC,$9C,$FE,$9D,$FB
	.dc.b	$1A,$E9,$E4,$FE,$C5,$E1,$87,$4F,$3F,$F8,$1C,$63,$0F,$E3,$15,$DB
	.dc.b	$2F,$1D,$EF,$EB,$7F,$8E,$FC,$DB,$B1,$FB,$77,$EC,$FF,$4A,$7E,$E0
	.dc.b	$FD,$BB,$75,$AF,$5D,$5A,$FC,$53,$3A,$FF,$1D,$CB,$EE,$36,$F8,$57
	.dc.b	$FA,$57,$93,$FD,$A5,$D5,$FE,$0B,$9C,$FC,$8F,$F1,$FC,$3C,$9A,$F8
	.dc.b	$F5,$7E,$CB,$FD,$B3,$F7,$7F,$DF,$FE,$30,$FD,$5D,$F8,$5F,$FB,$4F
	.dc.b	$F2,$2D,$7C,$02,$B5,$3F,$ED,$7F,$C2,$39,$7E,$79,$FD,$AF,$1F,$E1
	.dc.b	$3C,$7F,$DA,$DE,$B9,$74,$F9,$CF,$ED,$51,$8C,$1F,$99,$63,$3E,$D7
	.dc.b	$E2,$FC,$67,$FC,$D3,$CB,$FD,$9F,$0E,$11,$FE,$72,$8F,$F7,$97,$88
	.dc.b	$FC,$79,$FA,$2B,$F1,$3F,$ED,$3F,$95,$7F,$82,$FB,$FF,$DE,$37,$BA
	.dc.b	$D7,$D1,$2B,$FC,$21,$F9,$DE,$B1,$FF,$57,$F1,$7F,$9D,$0F,$F3,$AE
	.dc.b	$7E,$08,$FE,$E7,$1F,$E2,$5C,$7A,$67,$E0,$F1,$B7,$DD,$C6,$73,$9F
	.dc.b	$7F,$89,$CE,$1F,$E8,$09,$CE,$9F,$E2,$1F,$DE,$BC,$5B,$79,$A3,$FC
	.dc.b	$49,$F6,$B8,$D3,$DB,$F2,$1F,$E4,$7D,$FC,$97,$B8,$FF,$23,$EB,$E9
	.dc.b	$F9,$DF,$F6,$B7,$E8,$CF,$D2,$FF,$A2,$FE,$46,$FD,$59,$F9,$73,$36
	.dc.b	$3C,$7D,$46,$3C,$7E,$59,$8F,$1F,$CA,$F8,$5E,$EB,$E4,$72,$7C,$9C
	.dc.b	$79,$BF,$E9,$ED,$7F,$DD,$1F,$39,$A8,$F8,$62,$3C,$FB,$79,$BE,$47
	.dc.b	$31,$8D,$3C,$3E,$66,$36,$F2,$DF,$B3,$83,$F9,$16,$7E,$0C,$BA,$56
	.dc.b	$B0,$F4,$D7,$F8,$BA,$B5,$F8,$13,$E9,$9E,$C3,$F3,$DF,$F3,$0F,$E0
	.dc.b	$7F,$BB,$FD,$2F,$DF,$A7,$FA,$E3,$0E,$17,$BF,$B9,$5F,$F0,$1F,$DE
	.dc.b	$B7,$7F,$EA,$DC,$9F,$AB,$FC,$18,$CE,$3C,$9F,$F3,$83,$D3,$3F,$0F
	.dc.b	$C9,$FC,$68,$C6,$5D,$BE,$4B,$18,$FA,$2F,$CF,$EE,$3F,$7D,$EB,$39
	.dc.b	$BF,$2C,$BF,$63,$F2,$FC,$1B,$F0,$47,$E7,$AF,$CE,$3F,$C4,$5F,$57
	.dc.b	$FC,$7F,$F1,$BD,$D6,$A7,$3F,$02,$FD,$B7,$38,$7B,$7A,$7E,$96,$5D
	.dc.b	$7F,$75,$C7,$89,$FA,$D8,$F0,$57,$7F,$AF,$FF,$19,$E3,$F8,$49,$CE
	.dc.b	$5E,$2F,$4B,$1A,$78,$6F,$7F,$95,$F8,$2B,$C7,$FB,$2D,$9E,$7F,$AF
	.dc.b	$DE,$1F,$E7,$1F,$85,$7D,$4F,$C2,$DF,$7E,$FD,$25,$FB,$8B,$F2,$E7
	.dc.b	$E5,$7F,$E3,$7F,$AD,$E1,$7B,$AD,$4F,$F8,$EF,$EB,$F3,$F8,$07,$BB
	.dc.b	$C5,$FF,$7E,$E7,$DB,$3C,$42,$31,$87,$2F,$B0,$FD,$1C,$63,$4F,$0F
	.dc.b	$E7,$B8,$D3,$C1,$1E,$4F,$53,$FA,$77,$6F,$E7,$EF,$86,$DF,$E4,$DF
	.dc.b	$90,$FC,$8F,$B2,$FC,$53,$FA,$A7,$E2,$DF,$E3,$CF,$C0,$1F,$A6,$BF
	.dc.b	$68,$DD,$6B,$BF,$FC,$D7,$2F,$C0,$73,$E4,$FA,$B2,$E9,$F3,$77,$0E
	.dc.b	$19,$3F,$26,$FD,$ED,$78,$6D,$A7,$87,$CA,$CE,$1C,$33,$7D,$E7,$6A
	.dc.b	$F3,$43,$F3,$8F,$D5,$FE,$58,$F1,$1F,$8E,$3F,$04,$FF,$28,$7E,$69
	.dc.b	$FE,$C6,$FC,$CF,$75,$A9,$CF,$CD,$BF,$53,$CF,$E6,$DF,$53,$2E,$99
	.dc.b	$5C,$38,$7E,$02,$CF,$FD,$A3,$E2,$74,$AF,$04,$E7,$E1,$3F,$8D,$E1
	.dc.b	$CB,$4E,$FF,$84,$6D,$F0,$F8,$D6,$FA,$6C,$FC,$6B,$1A,$7F,$B7,$BB
	.dc.b	$7F,$84,$FD,$9F,$FB,$13,$8D,$FB,$83,$F3,$4F,$F5,$77,$D8,$37,$DF
	.dc.b	$CD,$FC,$7A,$D7,$CD,$A9,$F9,$96,$B1,$FC,$9E,$5F,$ED,$0E,$CB,$8F
	.dc.b	$1C,$FF,$D7,$5F,$33,$FC,$DA,$BC,$59,$F4,$E3,$C5,$EF,$4E,$DF,$99
	.dc.b	$F5,$AF,$0E,$AF,$D9,$FF,$91,$FA,$5F,$E1,$4F,$AD,$7E,$54,$FB,$6F
	.dc.b	$D2,$79,$8F,$CF,$77,$E7,$7F,$8B,$AB,$5E,$87,$E5,$FF,$40,$CF,$EE
	.dc.b	$9F,$AC,$CF,$A7,$87,$FD,$37,$D3,$FE,$85,$1D,$F0,$E3,$E9,$7F,$50
	.dc.b	$8C,$6B,$FB,$3F,$47,$1A,$77,$8F,$1F,$EE,$FE,$2E,$0F,$3C,$79,$BF
	.dc.b	$42,$FE,$7B,$FA,$57,$FA,$43,$E7,$3F,$DE,$BF,$B7,$FF,$91,$7F,$80
	.dc.b	$3F,$B5,$EF,$FD,$4E,$B5,$F8,$F2,$B5,$F3,$DC,$BF,$12,$FF,$73,$71
	.dc.b	$FA,$7E,$3F,$FC,$25,$87,$58,$C0,$0F,$4B,$FF,$2A,$79,$FF,$DB,$23
	.dc.b	$1F,$1D,$F9,$8C,$FC,$58,$B5,$F3,$CB,$57,$C2,$F8,$E7,$30,$F9,$7C
	.dc.b	$6F,$E9,$BE,$1C,$98,$D7,$E0,$7C,$CF,$F8,$2F,$F1,$FF,$F1,$57,$8D
	.dc.b	$FC,$B3,$F2,$6F,$DD,$7E,$DE,$FC,$D9,$FA,$73,$F6,$8E,$1C,$DE,$DC
	.dc.b	$38,$7A,$0C,$38,$5B,$F7,$2D,$DF,$F0,$1F,$65,$9F,$F5,$1D,$3C,$30
	.dc.b	$FD,$7F,$15,$D6,$37,$97,$5D,$5F,$8B,$FF,$90,$30,$E3,$8F,$A2,$7F
	.dc.b	$D8,$78,$78,$35,$F2,$EB,$E4,$FD,$37,$19,$F4,$8F,$3F,$C1,$F9,$BF
	.dc.b	$27,$7E,$5E,$FC,$97,$F8,$C7,$E9,$0F,$DA,$9F,$AB,$3E,$0B,$85,$EF
	.dc.b	$55,$EE,$BF,$C7,$1E,$13,$8B,$FC,$B1,$9F,$6F,$4B,$DF,$C2,$8F,$0E
	.dc.b	$B8,$6E,$5D,$35,$FE,$71,$5E,$0A,$F2,$C3,$E3,$3E,$24,$FE,$46,$F8
	.dc.b	$E9,$E3,$FD,$BB,$86,$3C,$F5,$7E,$A7,$F5,$BF,$04,$7E,$04,$FD,$DB
	.dc.b	$F3,$3F,$9E,$FF,$B4,$FF,$24,$7B,$FB,$AD,$79,$AA,$D7,$BB,$9F,$7F
	.dc.b	$DB,$97,$4C,$AE,$1C,$2F,$FD,$BD,$F8,$6F,$A3,$89,$7B,$CB,$A7,$E3
	.dc.b	$5F,$DC,$E7,$C1,$8F,$92,$1A,$BF,$3E,$CD,$F0,$8F,$F6,$17,$57,$D1
	.dc.b	$C3,$F6,$98,$F5,$CF,$D9,$F9,$57,$07,$DD,$F9,$EF,$F9,$BF,$F2,$BF
	.dc.b	$F4,$77,$E2,$AF,$D2,$7F,$59,$CD,$DF,$F1,$37,$EA,$3B,$D3,$F8,$05
	.dc.b	$EF,$D1,$FF,$69,$5E,$FF,$2E,$D5,$7F,$B3,$B1,$FE,$0A,$FF,$23,$CB
	.dc.b	$BC,$7F,$11,$8C,$1D,$1F,$0F,$13,$97,$FD,$43,$5C,$31,$EF,$3F,$1F
	.dc.b	$85,$39,$F5,$3F,$33,$3C,$7E,$EF,$C7,$AF,$9E,$73,$8B,$F1,$1F,$F4
	.dc.b	$87,$E3,$DF,$EA,$FF,$73,$FD,$69,$FB,$BF,$F7,$5F,$AA,$7D,$8B,$F5
	.dc.b	$67,$D8,$70,$E3,$FE,$45,$87,$0F,$A3,$FF,$1D,$DE,$E3,$53,$FC,$73
	.dc.b	$F0,$FF,$F2,$E3,$FD,$17,$3F,$04,$33,$FF,$2C,$46,$1F,$15,$FF,$B6
	.dc.b	$2B,$9D,$F7,$CF,$CB,$E8,$FF,$B1,$B0,$E1,$97,$8B,$47,$EF,$1C,$34
	.dc.b	$ED,$39,$FD,$77,$CB,$FA,$63,$F7,$6D,$FE,$F0,$FD,$95,$FA,$CF,$FB
	.dc.b	$77,$CF,$DD,$6A,$73,$F8,$B3,$F3,$CC,$EB,$EA,$E3,$FE,$8D,$9F,$5F
	.dc.b	$C5,$F1,$08,$F0,$4B,$4F,$06,$8F,$8F,$F6,$5F,$2D,$3C,$12,$FC,$A7
	.dc.b	$FD,$CB,$F9,$DF,$FB,$9F,$F6,$C7,$CC,$7C,$0F,$C4,$DF,$A8,$7F,$84
	.dc.b	$7E,$B7,$F8,$E7,$8E,$C3,$85,$EE,$B5,$B7,$FB,$4E,$73,$BF,$FA,$2D
	.dc.b	$DF,$0D,$DF,$FB,$16,$9E,$0A,$FF,$6E,$61,$FE,$1F,$F0,$FC,$B5,$AC
	.dc.b	$BB,$7D,$23,$EB,$73,$2C,$79,$6F,$E8,$A7,$E6,$3F,$EC,$CD,$7C,$73
	.dc.b	$98,$FF,$65,$46,$33,$FF,$17,$47,$BA,$31,$5C,$B6,$FF,$77,$7C,$B7
	.dc.b	$F7,$9C,$BF,$C8,$1F,$8D,$7F,$B1,$3F,$00,$7E,$4E,$F9,$F7,$D6,$FF
	.dc.b	$81,$F9,$7F,$B9,$61,$FE,$01,$FB,$26,$3C,$70,$E1,$7F,$E7,$BB,$DD
	.dc.b	$F6,$FC,$4F,$FD,$69,$9F,$87,$1E,$3F,$A4,$62,$BA,$65,$CF,$3F,$0C
	.dc.b	$7A,$0C,$E0,$D3,$CF,$3F,$E7,$19,$64,$FC,$83,$DD,$F8,$9F,$F8,$FB
	.dc.b	$95,$E2,$BF,$6F,$7E,$AA,$FD,$CD,$F9,$3F,$2C,$78,$FD,$7F,$D5,$C3
	.dc.b	$85,$EF,$F1,$4D,$6B,$FE,$23,$D3,$FC,$97,$D7,$FE,$D1,$C7,$8E,$B6
	.dc.b	$B5,$97,$3D,$7C,$3E,$1F,$F8,$AF,$7F,$45,$6A,$FC,$F5,$E6,$F1,$B5
	.dc.b	$F4,$BF,$67,$B1,$ED,$3E,$A9,$FA,$83,$D3,$7E,$F0,$FC,$4D,$F9,$37
	.dc.b	$C6,$CB,$96,$3C,$70,$E1,$7B,$F0,$75,$BF,$F8,$FF,$4E,$FF,$5F,$FE
	.dc.b	$ED,$C7,$8E,$B6,$B5,$97,$3D,$7C,$3F,$33,$E8,$9C,$D7,$9A,$7C,$BE
	.dc.b	$E7,$C1,$FC,$8B,$F2,$9F,$83,$3E,$21,$FA,$8B,$E4,$DF,$B4,$3F,$7F
	.dc.b	$FF,$73,$7E,$B7,$F8,$0C,$78,$E1,$C2,$F7,$5F,$EE,$BF,$CF,$B5,$B7
	.dc.b	$F9,$67,$3E,$D9,$FE,$8E,$3F,$49,$D8,$CB,$9F,$C1,$66,$7C,$C7,$1C
	.dc.b	$FB,$6B,$E3,$DB,$32,$CB,$AE,$BE,$2E,$1F,$BD,$F9,$2F,$EC,$7E,$57
	.dc.b	$F1,$FF,$82,$FF,$98,$34,$FF,$68,$FA,$FF,$CC,$37,$5E,$FC,$AF,$F6
	.dc.b	$AF,$C4,$7F,$53,$C3,$F8,$57,$9D,$FA,$0A,$D7,$E4,$FE,$7D,$F0,$7F
	.dc.b	$9D,$F4,$EE,$FF,$A5,$67,$A5,$F8,$4F,$E2,$91,$C0,$4F,$9B,$0F,$07
	.dc.b	$E0,$7F,$97,$F8,$DF,$E1,$5F,$69,$FE,$EC,$FD,$3B,$FC,$D9,$FA,$33
	.dc.b	$F6,$1F,$F1,$77,$E9,$3F,$B4,$BD,$FE,$24,$BD,$D7,$F8,$CF,$F7,$AD
	.dc.b	$6B,$FE,$8D,$CB,$A3,$FE,$AB,$C3,$8F,$E8,$5E,$DE,$4F,$EF,$53,$BC
	.dc.b	$BA,$7D,$B3,$5F,$17,$E3,$BE,$7A,$D6,$BE,$0F,$AB,$FD,$87,$E5,$9F
	.dc.b	$C8,$E1,$FB,$BF,$FA,$57,$F3,$77,$ED,$B3,$F3,$CF,$66,$CC,$38,$5E
	.dc.b	$EB,$5C,$BF,$C4,$A7,$3F,$83,$79,$DD,$FD,$B5,$D7,$FD,$55,$A7,$82
	.dc.b	$9A,$78,$31,$FF,$08,$D7,$33,$79,$77,$87,$93,$FC,$27,$F3,$BF,$E7
	.dc.b	$9F,$B9,$7E,$23,$F7,$2F,$E3,$DE,$CB,$F1,$C7,$ED,$EF,$D2,$7F,$93
	.dc.b	$E1,$C3,$C3,$61,$C2,$F7,$F0,$DA,$DB,$FD,$81,$9F,$E5,$3F,$4A,$FA
	.dc.b	$58,$F1,$C2,$D6,$B1,$E7,$F0,$1F,$C4,$72,$FF,$0D,$73,$B0,$E1,$39
	.dc.b	$E9,$FE,$45,$D7,$FE,$33,$FC,$D5,$F0,$FF,$CE,$3F,$5C,$7D,$1B,$F2
	.dc.b	$17,$E0,$FF,$D0,$30,$E1,$7B,$AD,$7E,$0B,$9E,$0F,$F1,$BE,$4F,$C5
	.dc.b	$3F,$15,$87,$09,$FF,$7C,$3C,$BE,$3D,$A7,$83,$AB,$CB,$5C,$B3,$ED
	.dc.b	$EC,$7F,$39,$FC,$FF,$FB,$67,$D7,$3F,$75,$FF,$16,$64,$FD,$C3,$FC
	.dc.b	$27,$F8,$DB,$C8,$FE,$AF,$87,$0B,$DD,$6B,$F0,$3C,$E9,$FE,$EF,$CB
	.dc.b	$A7,$93,$FA,$1C,$38,$7D,$45,$AD,$63,$CB,$E5,$7F,$DD,$D1,$C8,$FF
	.dc.b	$36,$46,$7F,$3C,$FE,$A1,$F2,$3F,$E0,$3F,$A8,$FE,$59,$D7,$FB,$9F
	.dc.b	$C3,$7F,$C0,$3F,$A5,$BD,$EF,$ED,$B7,$BF,$8A,$FA,$5E,$DA,$F7,$FB
	.dc.b	$8E,$B5,$0C,$FF,$9C,$7E,$51,$FB,$9C,$79,$69,$6B,$75,$E2,$C7,$A7
	.dc.b	$CA,$7E,$7A,$7A,$1C,$F3,$F0,$E7,$FE,$16,$FD,$2F,$FB,$57,$EE,$1F
	.dc.b	$E1,$DF,$C1,$9F,$BC,$FF,$B6,$3D,$57,$EE,$6F,$D4,$7F,$96,$FA,$1F
	.dc.b	$E0,$8C,$38,$5E,$EB,$52,$FA,$5D,$6B,$EC,$CB,$AF,$D7,$7F,$13,$C7
	.dc.b	$8F,$E1,$F6,$B5,$97,$3C,$FF,$DF,$B1,$C8,$61,$E6,$AF,$E1,$F9,$F9
	.dc.b	$7F,$1C,$FF,$02,$7E,$4B,$FE,$37,$CB,$E1,$FE,$BD,$FD,$EF,$F5,$D7
	.dc.b	$EE,$3F,$E1,$CF,$5F,$FD,$95,$A2,$F7,$EF,$AF,$7F,$97,$7F,$89,$B2
	.dc.b	$E9,$CC,$FD,$2F,$F9,$64,$63,$A6,$E5,$F6,$B2,$E7,$8B,$F5,$4F,$2C
	.dc.b	$E7,$1F,$F2,$3E,$DE,$5F,$BB,$FF,$0E,$7E,$99,$FE,$4F,$97,$E4,$5F
	.dc.b	$AC,$F6,$BF,$6B,$F5,$FF,$EF,$0E,$25,$EE,$B5,$E1,$3D,$BF,$F1,$E7
	.dc.b	$E3,$1F,$E3,$EA,$D7,$D6,$7F,$43,$78,$7B,$57,$7F,$C8,$EB,$E0,$8F
	.dc.b	$E1,$38,$F3,$FB,$7F,$2D,$6B,$3E,$F9,$BA,$BF,$C6,$9F,$8A,$FF,$99
	.dc.b	$FF,$59,$FE,$53,$EB,$FE,$D7,$F3,$4F,$27,$0F,$FB,$DB,$2F,$CF,$D6
	.dc.b	$BD,$37,$B4,$F6,$58,$7F,$18,$C3,$86,$0C,$3F,$9A,$FC,$67,$F5,$DC
	.dc.b	$FC,$0F,$F0,$99,$FF,$09,$57,$0A,$ED,$8F,$5F,$6F,$CF,$5E,$2B,$FE
	.dc.b	$21,$9F,$8B,$3F,$EF,$B8,$FF,$1C,$AF,$DA,$FB,$0F,$D6,$1F,$B1,$FF
	.dc.b	$15,$FC,$4F,$FD,$4D,$EA,$FF,$8C,$30,$E1,$87,$FB,$1E,$F7,$A2,$F7
	.dc.b	$FB,$37,$F9,$CB,$A3,$FD,$1D,$9F,$5F,$C6,$7F,$BC,$46,$30,$EF,$8F
	.dc.b	$B2,$31,$A7,$87,$F0,$2F,$F4,$0E,$9E,$0C,$7C,$D3,$9F,$1F,$FC,$D1
	.dc.b	$F9,$D7,$FA,$6F,$85,$F2,$1F,$8A,$FC,$A7,$FC,$BF,$E3,$79,$7F,$25
	.dc.b	$FF,$42,$DE,$EB,$5C,$8A,$F5,$7C,$9E,$B8,$3F,$8C,$E5,$D3,$F0,$BF
	.dc.b	$F0,$86,$3C,$7C,$4B,$5A,$CB,$9E,$2F,$27,$CB,$39,$A6,$BE,$4F,$2B
	.dc.b	$FD,$00,$7F,$AC,$FC,$DF,$17,$A1,$FB,$7F,$45,$E8,$FE,$EF,$A3,$FD
	.dc.b	$3B,$7B,$D3,$F8,$15,$6B,$D0,$FF,$40,$F4,$7E,$D7,$5F,$E6,$39,$F4
	.dc.b	$7F,$13,$87,$1F,$DD,$FF,$C5,$4F,$3D,$3C,$1A,$7D,$08,$C7,$A3,$FC
	.dc.b	$23,$A3,$F9,$DF,$EC,$0F,$CD,$BF,$D9,$5F,$B3,$BF,$34,$7E,$40,$FC
	.dc.b	$B5,$F8,$83,$F3,$8F,$F0,$87,$D6,$B2,$E5,$C3,$FD,$1B,$2E,$58,$F1
	.dc.b	$C3,$87,$D0,$3F,$45,$E1,$AF,$83,$5F,$D4,$E3,$FE,$DB,$CB,$9A,$CB
	.dc.b	$9B,$FE,$79,$8C,$C7,$6D,$7C,$BA,$BF,$34,$F6,$E3,$D7,$5F,$14,$BE
	.dc.b	$43,$E0,$F8,$1F,$EE,$9F,$CD,$7F,$C8,$FF,$8F,$3F,$52,$7D,$67,$FA
	.dc.b	$C3,$F4,$77,$E8,$FE,$FC,$38,$7D,$4B,$0E,$17,$BF,$9D,$D7,$1F,$F6
	.dc.b	$8C,$FA,$BE,$9C,$38,$FE,$91,$FE,$7D,$E4,$47,$4C,$79,$FE,$16,$FC
	.dc.b	$33,$CF,$87,$3D,$3C,$1B,$F9,$A9,$F9,$0F,$DC,$F2,$3F,$BB,$DF,$F9
	.dc.b	$77,$A9,$FA,$DB,$F4,$E7,$27,$FA,$D7,$8F,$F8,$DF,$07,$F0,$4F,$A1
	.dc.b	$FA,$D7,$BA,$D7,$C4,$FD,$FC,$FA,$D6,$B3,$FA,$91,$8C,$DF,$8C,$7B
	.dc.b	$63,$18,$F3,$FC,$1F,$E6,$DF,$33,$FC,$1B,$87,$83,$5F,$2C,$BE,$B3
	.dc.b	$F2,$FF,$0B,$FC,$E7,$F7,$93,$FE,$6E,$FD,$41,$F8,$A3,$F5,$3F,$E6
	.dc.b	$5C,$5F,$CB,$6F,$7F,$6C,$BD,$D3,$F6,$1F,$CD,$63,$D4,$FA,$87,$7F
	.dc.b	$D7,$2E,$FF,$D1,$31,$E7,$87,$83,$3F,$2C,$E6,$FC,$90,$FA,$2F,$D7
	.dc.b	$FB,$5F,$DA,$7A,$BF,$DC,$7F,$9C,$3F,$03,$F1,$7E,$72,$FD,$57,$F5
	.dc.b	$97,$BF,$C4,$97,$BE,$05,$FB,$F8,$FF,$B7,$ED,$FD,$9B,$3F,$F7,$7B
	.dc.b	$FA,$7C,$7C,$1F,$86,$FE,$3D,$E9,$8F,$3F,$BF,$F2,$D7,$1C,$FB,$EB
	.dc.b	$F1,$7C,$BF,$E0,$FE,$B7,$F8,$67,$E5,$5F,$B5,$35,$FD,$B7,$E3,$8F
	.dc.b	$C8,$DF,$A7,$7F,$05,$C3,$85,$EF,$D7,$5E,$FF,$2A,$D6,$DF,$E1,$5C
	.dc.b	$FA,$BF,$BF,$C3,$C1,$E4,$FE,$9D,$E7,$A7,$83,$CA,$F2,$D7,$1C,$FB
	.dc.b	$65,$FA,$98,$FE,$2D,$F0,$9F,$E4,$4F,$D1,$1F,$84,$BE,$41,$FB,$63
	.dc.b	$F6,$0F,$F5,$67,$5F,$F8,$C7,$1F,$F2,$0E,$1C,$2F,$75,$BA,$D5,$7F
	.dc.b	$90,$BE,$B9,$7F,$BD,$CB,$FC,$A7,$FA,$45,$AE,$79,$F8,$73,$FE,$31
	.dc.b	$1E,$63,$39,$F9,$39,$FF,$9B,$FC,$A7,$F9,$DF,$F6,$17,$EB,$2F,$C7
	.dc.b	$3E,$9F,$DC,$FF,$94,$B8,$17,$BA,$D7,$1F,$F9,$E7,$F3,$9F,$E7,$13
	.dc.b	$CF,$4F,$E5,$8F,$FA,$0F,$F1,$FF,$E2,$8F,$1C,$BE,$FF,$C5,$78,$AB
	.dc.b	$FC,$FB,$F7,$4D,$7C,$75,$F8,$4F,$C6,$FF,$A6,$FE,$05,$FE,$9C,$F4
	.dc.b	$3F,$4F,$E7,$7F,$47,$8C,$79,$FF,$B9,$F3,$B1,$9F,$FA,$83,$E7,$FF
	.dc.b	$92,$57,$FA,$AF,$D8,$B3,$9C,$7F,$77,$F5,$DD,$A3,$18,$FF,$BC,$3E
	.dc.b	$B3,$FA,$EF,$E0,$FF,$EC,$AF,$D2,$3F,$5F,$E6,$FE,$FF,$A5,$39,$F3
	.dc.b	$BF,$3F,$F5,$BF,$D6,$E3,$FD,$B1,$F8,$4F,$F3,$57,$FD,$99,$9F,$F2
	.dc.b	$C3,$47,$FB,$27,$1F,$DD,$F3,$F8,$23,$18,$7E,$E3,$E9,$3F,$B7,$FE
	.dc.b	$1B,$FE,$E6,$F4,$3F,$B2,$FE,$47,$FD,$A6,$31,$F2,$DF,$DC,$3F,$17
	.dc.b	$C6,$3F,$EF,$1F,$A5,$FE,$8D,$5F,$EF,$58,$FE,$C4,$F0,$FC,$27,$AC
	.dc.b	$7F,$BD,$FF,$11,$7E,$21,$FD,$A7,$17,$A9,$FE,$00,$EA,$FE,$8B,$F8
	.dc.b	$9B,$E4,$1E,$E3,$E1,$9F,$AE,$FE,$9B,$8D,$FA,$07,$F7,$FC,$78,$E1
	.dc.b	$C2,$F7,$E9,$2D,$FB,$DE,$F8,$3F,$D9,$3A,$77,$D3,$FB,$06,$5C,$BF
	.dc.b	$17,$B5,$AC,$FA,$6B,$E1,$8E,$43,$0E,$5A,$78,$3A,$7F,$C3,$7C,$DF
	.dc.b	$D8,$3F,$1B,$7E,$35,$BE,$07,$8C,$FC,$B3,$F8,$8F,$E1,$7F,$E1,$9C
	.dc.b	$38,$5E,$EB,$5E,$F3,$F2,$FD,$6D,$FE,$E2,$CF,$AE,$DF,$DE,$71,$E3
	.dc.b	$F8,$ED,$AD,$65,$CF,$4F,$A3,$1C,$86,$1C,$B5,$F2,$CB,$EB,$BF,$C6
	.dc.b	$BF,$87,$7F,$7A,$FC,$7D,$F9,$13,$F2,$17,$E4,$2F,$90,$7E,$DD,$FC
	.dc.b	$70,$FF,$1A,$61,$C6,$F8,$56,$E7,$F9,$0F,$DC,$6B,$6F,$F7,$EE,$5D
	.dc.b	$BF,$18,$FF,$0A,$63,$C7,$3B,$5A,$CB,$9E,$7E,$0F,$C9,$3C,$F3,$9A
	.dc.b	$F2,$CF,$93,$F5,$4F,$F2,$7F,$E5,$1F,$E1,$4F,$CA,$1F,$3E,$F4,$1F
	.dc.b	$94,$7F,$1B,$FD,$81,$F1,$2F,$A7,$FF,$27,$63,$C7,$0E,$17,$BA,$7E
	.dc.b	$BF,$AD,$83,$3E,$F9,$FF,$8B,$F1,$E5,$AD,$AD,$D6,$B1,$F1,$78,$7E
	.dc.b	$89,$CE,$7E,$0F,$84,$6C,$F8,$4F,$F3,$5F,$E5,$BF,$E3,$0F,$CB,$DE
	.dc.b	$C3,$F2,$E7,$E2,$AF,$CB,$BF,$B4,$7A,$CF,$D9,$FF,$CD,$58,$F1,$C3
	.dc.b	$85,$EF,$E6,$FF,$CE,$9B,$E8,$F6,$E7,$DB,$F3,$0F,$F2,$56,$1C,$74
	.dc.b	$B5,$AC,$79,$E9,$FD,$DF,$D7,$F3,$CE,$6B,$CB,$3F,$DE,$FC,$1F,$F4
	.dc.b	$2F,$E3,$5F,$E4,$AF,$CA,$7F,$BB,$FE,$62,$FC,$AB,$F8,$6F,$E5,$FF
	.dc.b	$E4,$BF,$D1,$7F,$D0,$57,$BF,$51,$7B,$F8,$05,$FF,$94,$72,$DE,$BF
	.dc.b	$5B,$3F,$F0,$1F,$E2,$1F,$E6,$3C,$B9,$63,$6B,$59,$74,$D3,$C3,$F2
	.dc.b	$7E,$69,$CD,$77,$CF,$FC,$01,$F1,$8F,$90,$FF,$31,$FD,$AB,$F3,$57
	.dc.b	$BA,$FC,$CD,$F9,$B3,$F3,$6F,$F6,$57,$C4,$FF,$A7,$71,$E3,$87,$0B
	.dc.b	$DD,$7F,$B3,$FF,$41,$D6,$DF,$90,$CF,$B6,$7F,$E7,$DC,$78,$FC,$0D
	.dc.b	$AD,$65,$CF,$4F,$07,$B5,$FE,$B4,$BE,$F8,$79,$67,$C9,$AF,$FA,$E3
	.dc.b	$F1,$7F,$F3,$EF,$EF,$6F,$C7,$FF,$C9,$5E,$7F,$FB,$E7,$37,$DE,$BF
	.dc.b	$4D,$FF,$5B,$61,$C2,$F7,$F8,$F7,$C3,$A7,$DD,$EB,$83,$F5,$39,$F5
	.dc.b	$F9,$BF,$F4,$B6,$3C,$7F,$44,$B5,$AC,$B9,$E8,$F9,$5F,$34,$E7,$0F
	.dc.b	$E2,$DA,$7D,$1F,$C4,$F3,$2C,$FB,$6F,$E9,$F3,$BF,$CB,$FD,$6D,$BD
	.dc.b	$19,$72,$D3,$F5,$7F,$6B,$FE,$D1,$F2,$3F,$D3,$FF,$8D,$BE,$8D,$F9
	.dc.b	$EF,$F9,$4F,$F4,$07,$E7,$EC,$3F,$98,$E9,$FB,$3E,$0F,$E0,$F8,$F1
	.dc.b	$C3,$85,$EF,$F6,$4E,$D5,$AC,$BA,$BF,$ED,$3C,$FC,$30,$FC,$7F,$FD
	.dc.b	$79,$18,$C3,$9F,$E1,$7F,$EF,$08,$C7,$DF,$36,$F2,$F9,$3F,$EF,$4F
	.dc.b	$D1,$BF,$D7,$3F,$98,$FF,$BC,$BF,$25,$7E,$A1,$FD,$17,$F1,$17,$E0
	.dc.b	$4F,$B0,$FF,$79,$61,$C3,$F1,$76,$1C,$3C,$AF,$F6,$6D,$FF,$9D,$39
	.dc.b	$D7,$BA,$FF,$69,$3C,$7E,$6E,$F2,$FF,$2A,$D7,$C5,$AF,$87,$F2,$7F
	.dc.b	$34,$E7,$4E,$F3,$E5,$F8,$5C,$CA,$F8,$EB,$E2,$FD,$73,$CF,$3C,$74
	.dc.b	$F3,$70,$7B,$3F,$6F,$CC,$FF,$96,$BE,$8B,$FB,$CF,$C9,$3F,$1B,$7C
	.dc.b	$CF,$FA,$FF,$F4,$8D,$FD,$D7,$F7,$EB,$FE,$93,$5A,$FC,$43,$FC,$7F
	.dc.b	$CD,$F0,$32,$FE,$8B,$7F,$E0,$4F,$8B,$F7,$C7,$7C,$2C,$F3,$CF,$C1
	.dc.b	$9F,$FB,$3A,$BB,$D7,$93,$2F,$1F,$E1,$FF,$03,$0E,$11,$E1,$FD,$47
	.dc.b	$75,$E4,$C3,$FD,$AD,$D3,$8C,$BA,$FC,$93,$F4,$4C,$63,$EE,$8F,$F0
	.dc.b	$0E,$CC,$BF,$DF,$BF,$AF,$7F,$01,$FC,$9D,$F8,$4B,$F1,$17,$E9,$2F
	.dc.b	$CC,$9F,$80,$32,$E5,$C9,$FD,$9B,$2E,$58,$F1,$C3,$87,$CA,$EF,$4E
	.dc.b	$FE,$DB,$F4,$1F,$F3,$16,$3F,$DB,$B2,$E6,$B2,$F2,$3A,$8C,$C7,$6D
	.dc.b	$7C,$BF,$49,$C7,$1E,$BB,$FA,$A7,$D3,$F4,$D9,$C7,$CD,$9A,$B5,$AB
	.dc.b	$F4,$4D,$57,$86,$B5,$BF,$AF,$85,$F3,$5E,$BA,$FF,$02,$FD,$87,$1F
	.dc.b	$C1,$1E,$73,$D6,$7E,$C8,$FD,$B9,$FB,$27,$1C,$58,$F1,$C3,$85,$EF
	.dc.b	$87,$53,$D6,$BE,$17,$1F,$F8,$51,$E3,$C0,$F1,$3E,$5C,$70,$86,$BE
	.dc.b	$4D,$7C,$7C,$BE,$EC,$7C,$11,$E6,$FD,$67,$17,$C3,$5F,$1E,$EF,$D1
	.dc.b	$7A,$65,$CF,$7F,$67,$EF,$3E,$BF,$BE,$7E,$96,$FA,$77,$93,$FE,$2C
	.dc.b	$FC,$CD,$F9,$4F,$F9,$3B,$C0,$FB,$F8,$70,$BD,$FD,$8A,$F7,$F4,$5D
	.dc.b	$72,$FF,$0C,$F3,$7F,$AA,$71,$E3,$1F,$87,$8F,$8A,$7F,$8E,$31,$E9
	.dc.b	$9F,$87,$D6,$FB,$B1,$8C,$BB,$7D,$6E,$30,$F4,$5F,$9F,$5F,$C9,$65
	.dc.b	$DB,$5F,$2E,$BE,$A7,$C8,$62,$5F,$BF,$E2,$F9,$79,$9F,$A7,$FA,$8F
	.dc.b	$B1,$FA,$D3,$F5,$03,$FD,$B7,$ED,$9F,$A0,$3F,$1A,$7E,$FD,$C7,$97
	.dc.b	$FC,$A7,$7C,$4E,$15,$BC,$F5,$3D,$31,$7D,$E7,$E1,$38,$7F,$C9,$53
	.dc.b	$CA,$7C,$2F,$EB,$74,$F2,$C7,$93,$E7,$7F,$D0,$D8,$BF,$20,$FD,$B5
	.dc.b	$6B,$3F,$0F,$5E,$31,$E9,$1E,$4D,$BF,$DF,$99,$F8,$27,$C9,$39,$FD
	.dc.b	$33,$CF,$FA,$73,$A3,$FD,$7F,$FA,$77,$FA,$BB,$F2,$A4,$FE,$1F,$4F
	.dc.b	$D0,$3B,$FC,$39,$5B,$8F,$7E,$7D,$DD,$67,$ED,$3F,$CF,$6F,$7F,$C3
	.dc.b	$FE,$26,$1D,$32,$F0,$FB,$EC,$FF,$DA,$58,$79,$A7,$30,$FB,$4F,$81
	.dc.b	$3E,$69,$CE,$3F,$DC,$F2,$F1,$46,$3F,$4C,$FA,$BF,$59,$FE,$CC,$C7
	.dc.b	$FC,$61,$FA,$43,$F4,$67,$EA,$BF,$CD,$B6,$FF,$27,$FC,$97,$8F,$DA
	.dc.b	$5E,$E9,$FB,$6F,$4A,$F4,$6F,$EF,$19,$FF,$B6,$BF,$5A,$F7,$C7,$F9
	.dc.b	$36,$7A,$61,$E1,$FC,$57,$D7,$7E,$5C,$BB,$E9,$FC,$3F,$83,$ED,$4E
	.dc.b	$6F,$C5,$FB,$4F,$87,$7F,$5D,$FA,$BF,$66,$C6,$3C,$B7,$F7,$65,$CB
	.dc.b	$4F,$0D,$FC,$0F,$C9,$7F,$33,$F7,$7F,$D4,$3F,$13,$FE,$B5,$F9,$4B
	.dc.b	$F4,$FF,$8D,$F9,$D6,$BE,$19,$5A,$9F,$95,$FB,$F6,$63,$DF,$CB,$FE
	.dc.b	$8F,$A7,$E8,$18,$CB,$B7,$C8,$B3,$ED,$F9,$8F,$DC,$9C,$C7,$87,$F0
	.dc.b	$0C,$61,$E0,$BE,$1A,$7D,$9E,$3F,$04,$63,$81,$FA,$6F,$ED,$3F,$6C
	.dc.b	$71,$3E,$8F,$FB,$B7,$EE,$CF,$D3,$5F,$85,$3E,$37,$85,$EF,$35,$EE
	.dc.b	$9F,$BE,$EB,$1F,$7E,$5D,$BF,$56,$B8,$F1,$C3,$B7,$CC,$FE,$62,$BC
	.dc.b	$77,$E2,$D3,$E2,$F3,$F3,$D3,$BC,$79,$3F,$7A,$C6,$3E,$1F,$C7,$FF
	.dc.b	$87,$FE,$D2,$FC,$D5,$FC,$01,$FD,$07,$BF,$F9,$D7,$E9,$AF,$DB,$77
	.dc.b	$5A,$F8,$55,$6B,$E5,$73,$FA,$77,$D8,$CB,$A6,$57,$0E,$17,$FD,$17
	.dc.b	$EE,$3F,$63,$5A,$C7,$9F,$C4,$FC,$DE,$2E,$F8,$F8,$E1,$AB,$F4,$7C
	.dc.b	$DF,$08,$FE,$ED,$E4,$FF,$17,$C7,$A6,$7E,$AF,$87,$DB,$EC,$FC,$8F
	.dc.b	$F9,$7F,$E6,$5F,$E5,$5E,$F7,$EC,$EF,$DC,$7F,$D2,$1F,$87,$FF,$BE
	.dc.b	$B3,$FE,$45,$7B,$AD,$7E,$0B,$FE,$C8,$FC,$BB,$9F,$DB,$FE,$07,$93
	.dc.b	$F7,$57,$FE,$61,$F1,$3F,$36,$F1,$CF,$FD,$39,$CF,$FC,$4A,$31,$8B
	.dc.b	$D9,$C6,$3F,$8B,$CF,$BE,$BF,$A2,$D7,$C7,$39,$CB,$CF,$E8,$FF,$4A
	.dc.b	$3F,$C4,$DE,$6B,$F5,$C7,$DA,$3F,$80,$5E,$D7,$F8,$17,$2F,$E7,$D8
	.dc.b	$70,$BD,$D6,$BE,$6F,$A7,$D5,$7F,$B9,$F9,$3F,$8B,$5F,$F8,$CF,$F0
	.dc.b	$4F,$DB,$BC,$73,$ED,$A7,$F2,$E8,$C6,$2F,$5F,$18,$FF,$BC,$73,$EF
	.dc.b	$AB,$E8,$FB,$E3,$E9,$CB,$D1,$2F,$DA,$BF,$B3,$FA,$5F,$EC,$9F,$86
	.dc.b	$FE,$57,$CC,$F0,$1F,$96,$3F,$2F,$DF,$E1,$DF,$D5,$FE,$1F,$E7,$56
	.dc.b	$BE,$A7,$38,$FF,$10,$CB,$FD,$A9,$FB,$0D,$C3,$86,$7D,$7F,$18,$FE
	.dc.b	$93,$5E,$1C,$B9,$FB,$18,$BD,$E7,$E2,$FD,$5B,$AD,$78,$36,$FF,$39
	.dc.b	$76,$7F,$6D,$C7,$96,$7E,$99,$7E,$25,$FE,$04,$FD,$93,$FB,$3F,$A5
	.dc.b	$FD,$77,$F0,$A7,$E8,$DF,$B9,$FC,$67,$5F,$2E,$AD,$4F,$EF,$7D,$FC
	.dc.b	$C7,$F2,$CC,$3F,$B8,$FB,$76,$F8,$7C,$73,$2E,$BE,$6F,$F6,$6B,$F0
	.dc.b	$CE,$72,$7E,$A9,$9C,$38,$E1,$C3,$F1,$5F,$F0,$DA,$FD,$13,$FD,$9F
	.dc.b	$F9,$1B,$F6,$67,$E5,$8F,$D6,$FF,$C4,$FC,$FF,$E0,$FC,$38,$7E,$0A
	.dc.b	$C3,$85,$EF,$EA,$BA,$E5,$FE,$90,$F3,$FF,$AF,$7C,$DF,$F9,$B7,$1E
	.dc.b	$31,$FD,$AD,$E7,$87,$2D,$BF,$C0,$31,$8D,$7C,$7F,$AA,$63,$1F,$3C
	.dc.b	$79,$B6,$7E,$B7,$FE,$36,$CD,$E9,$BF,$16,$F7,$70,$FF,$AB,$7C,$8F
	.dc.b	$CB,$3F,$E2,$9A,$DC,$EA,$3F,$D1,$3E,$77,$3B,$E5,$C1,$FE,$01,$BE
	.dc.b	$BF,$8A,$7F,$62,$78,$EB,$F8,$8E,$FF,$F0,$CC,$63,$1F,$F5,$E7,$C3
	.dc.b	$E3,$1F,$25,$FE,$3B,$47,$E9,$1E,$9F,$8F,$FF,$4C,$FE,$68,$FC,$F5
	.dc.b	$EF,$BE,$E9,$FB,$9E,$F1,$5E,$EB,$5E,$4F,$3F,$33,$CD,$97,$FA,$93
	.dc.b	$8F,$D4,$F6,$CF,$AF,$E8,$9E,$88,$C6,$1F,$BF,$AF,$F3,$0F,$E0,$DF
	.dc.b	$EF,$CF,$D8,$7F,$11,$FA,$5B,$F7,$AF,$F1,$1F,$B3,$AF,$C4,$95,$A9
	.dc.b	$FF,$13,$F1,$E6,$3F,$C1,$38,$29,$FA,$CE,$1E,$3C,$AC,$BB,$7E,$05
	.dc.b	$FE,$44,$BF,$14,$E7,$2F,$F2,$F6,$9E,$6F,$23,$FC,$95,$F8,$01,$7C
	.dc.b	$8B,$F0,$1F,$C2,$76,$FF,$9A,$72,$D7,$D8,$FF,$AB,$AF,$74,$FD,$D3
	.dc.b	$A4,$FA,$27,$F4,$D9,$7F,$90,$EF,$8E,$5D,$F2,$6B,$E7,$BF,$19,$E2
	.dc.b	$E5,$F5,$5B,$D2,$78,$3F,$CC,$7F,$70,$FE,$BD,$F2,$1F,$42,$B5,$39
	.dc.b	$F8,$D7,$EE,$79,$F7,$3E,$5D,$73,$DF,$C8,$F0,$F8,$B6,$5F,$E0,$B9
	.dc.b	$E9,$9F,$82,$73,$EB,$79,$2F,$FC,$77,$E7,$3F,$4E,$7E,$E6,$FC,$BF
	.dc.b	$C9,$5A,$9C,$FE,$9F,$8C,$FC,$33,$9F,$86,$E9,$F5,$FF,$C2,$ED,$F1
	.dc.b	$1D,$E9,$FB,$3B,$F4,$DF,$07,$C0,$76,$9C,$C7,$8F,$F6,$0C,$56,$FD
	.dc.b	$DE,$4F,$1F,$7F,$A5,$BE,$FD,$E9,$FF,$A8,$FC,$5F,$C0,$48,$86,$E1
	.dc.b	$F1,$BE,$18,$E5,$EC,$F4,$B7,$B3,$DB,$BD,$FA,$1B,$F5,$CE,$11,$8F
	.dc.b	$D8,$2E,$BB,$FC,$4F,$CB,$9F,$F3,$B9,$EC,$FF,$3D,$FC,$3E,$F2,$EA
	.dc.b	$1F,$9F,$69,$7E,$B9,$FB,$4F,$35,$F9,$E7,$FB,$87,$ED,$DF,$76,$FC
	.dc.b	$A1,$FB,$66,$EB,$53,$9F,$B4,$E5,$F9,$DF,$ED,$31,$E7,$E1,$FE,$75
	.dc.b	$EB,$F8,$8B,$3E,$BC,$7F,$B3,$9C,$C7,$83,$8F,$FD,$73,$F8,$EF,$F9
	.dc.b	$BF,$F7,$0F,$3F,$CB,$FF,$61,$FD,$53,$FC,$55,$E0,$7F,$B6,$6B,$5C
	.dc.b	$7F,$E7,$2F,$99,$57,$F2,$1A,$D7,$D7,$FF,$AE,$F7,$FF,$59,$61,$D5
	.dc.b	$B0,$CB,$BE,$1F,$EC,$B9,$FF,$7C,$D7,$6C,$BF,$B9,$7E,$8B,$E6,$C3
	.dc.b	$86,$7E,$1C,$FF,$96,$E7,$FE,$ED,$FD,$DD,$F8,$67,$FB,$67,$E5,$1F
	.dc.b	$B0,$FF,$86,$BE,$B5,$E0,$30,$E1,$7B,$AD,$78,$19,$F0,$5C,$33,$3E
	.dc.b	$BF,$3D,$FE,$AC,$7B,$FE,$1B,$71,$E5,$ED,$FF,$A2,$B3,$EB,$7F,$D7
	.dc.b	$73,$FF,$15,$FE,$0D,$FE,$E8,$C7,$9E,$FE,$6C,$3F,$D7,$7F,$8D,$F8
	.dc.b	$74,$F0,$FB,$1F,$D2,$3F,$66,$FF,$6D,$FE,$E6,$FC,$A5,$F8,$9B,$F0
	.dc.b	$6E,$18,$7F,$02,$F9,$A6,$1C,$2F,$7F,$A8,$EB,$2F,$EA,$7F,$97,$FF
	.dc.b	$B1,$9E,$D8,$7F,$BA,$5F,$07,$E2,$8D,$3C,$1F,$9A,$7F,$C0,$09,$CC
	.dc.b	$78,$F8,$B1,$A7,$82,$F8,$67,$E6,$FD,$D7,$CB,$FA,$2B,$F4,$5F,$F1
	.dc.b	$8D,$35,$7D,$4B,$F4,$2F,$F1,$EF,$6F,$D2,$E0,$63,$C7,$E0,$BF,$DF
	.dc.b	$B8,$70,$BF,$F1,$FF,$67,$F3,$6B,$FB,$37,$DF,$E0,$7F,$DC,$6F,$18
	.dc.b	$E8,$7D,$E9,$E0,$F6,$D3,$CB,$87,$93,$ED,$B5,$9F,$8F,$0F,$16,$BE
	.dc.b	$9D,$5E,$3E,$B2,$F0,$E5,$CA,$7D,$9F,$47,$F0,$A3,$1B,$7F,$20,$D7
	.dc.b	$C3,$97,$4D,$7D,$9F,$6F,$F7,$3F,$3F,$7E,$E9,$FF,$00,$3F,$47,$7C
	.dc.b	$5F,$F2,$9F,$CB,$5C,$DB,$AD,$4E,$7B,$F1,$F1,$9F,$03,$8B,$CC,$32
	.dc.b	$EB,$E6,$66,$39,$2C,$FB,$FC,$2B,$F0,$FE,$37,$C2,$73,$E8,$E3,$4F
	.dc.b	$0D,$B6,$FE,$91,$C5,$F3,$B0,$E1,$B7,$A7,$F5,$9F,$CC,$FE,$3C,$FD
	.dc.b	$F9,$FA,$A7,$F9,$9B,$F6,$9F,$F0,$36,$DF,$E1,$2C,$BE,$BD,$EE,$B5
	.dc.b	$F3,$DE,$7E,$CE,$78,$F2,$E3,$F8,$18,$70,$87,$E9,$3E,$F8,$C6,$1C
	.dc.b	$B5,$F8,$B9,$75,$F9,$86,$BF,$E1,$7E,$6C,$E5,$E6,$9C,$EB,$F1,$FE
	.dc.b	$8B,$C3,$97,$93,$7F,$F1,$7E,$9F,$8E,$FC,$E1,$FB,$97,$F1,$3F,$A4
	.dc.b	$7E,$E8,$FD,$91,$FA,$E5,$C3,$85,$EE,$B5,$2F,$C7,$7F,$80,$E3,$D2
	.dc.b	$FD,$9E,$4F,$EA,$63,$BF,$CB,$E3,$9F,$F8,$EC,$6F,$4F,$0E,$9F,$C6
	.dc.b	$6F,$86,$9D,$E3,$F1,$F7,$DA,$31,$9B,$E2,$B1,$9F,$F8,$EB,$5F,$2F
	.dc.b	$DE,$63,$3E,$B7,$FD,$CA,$BD,$66,$F2,$F7,$FE,$65,$FC,$6A,$1E,$73
	.dc.b	$E3,$5F,$AF,$BF,$7E,$7D,$C7,$C6,$FC,$53,$85,$EE,$B5,$3F,$2B,$F6
	.dc.b	$FC,$FB,$3F,$07,$8F,$FA,$15,$F5,$7F,$0B,$C9,$FC,$3F,$4F,$0E,$9E
	.dc.b	$0D,$3B,$F7,$7E,$A2,$73,$A7,$6F,$07,$19,$78,$6F,$CB,$C3,$FD,$97
	.dc.b	$2E,$93,$9D,$BE,$EF,$ED,$3F,$CE,$3E,$BD,$FB,$CB,$F6,$07,$E3,$1F
	.dc.b	$E6,$3D,$DF,$8E,$7F,$A3,$B8,$3F,$8B,$E1,$C2,$F7,$F6,$CB,$DF,$C5
	.dc.b	$75,$CB,$FD,$29,$C9,$EE,$3F,$AE,$E1,$C5,$C0,$7A,$67,$E1,$FC,$1B
	.dc.b	$F8,$3C,$63,$1E,$DE,$9E,$30,$F4,$5F,$9F,$57,$C5,$78,$75,$F2,$EB
	.dc.b	$EA,$FB,$1F,$EA,$FC,$5F,$E2,$4F,$27,$FB,$CF,$88,$FD,$E9,$FB,$1F
	.dc.b	$F8,$A3,$93,$F9,$AD,$EF,$E4,$57,$BA,$7E,$A7,$D3,$E8,$BF,$DB,$9C
	.dc.b	$DF,$C7,$B0,$EC,$FB,$9E,$57,$DF,$D4,$D5,$7F,$A4,$F1,$EB,$9B,$EB
	.dc.b	$B3,$AF,$9A,$3C,$BF,$9D,$77,$8C,$6B,$FE,$71,$DB,$CF,$8F,$3C,$FC
	.dc.b	$DF,$A7,$EF,$3E,$B6,$FD,$5F,$FB,$CF,$EB,$8F,$D8,$9F,$BC,$7F,$AC
	.dc.b	$3D,$47,$EC,$CF,$DB,$9F,$9D,$F0,$BD,$FE,$11,$FC,$DE,$B5,$F9,$E6
	.dc.b	$71,$FD,$06,$5D,$7F,$5C,$B8,$70,$CF,$AF,$D5,$7F,$44,$AF,$15,$F8
	.dc.b	$71,$FF,$76,$F1,$FF,$66,$9C,$E7,$DF,$57,$E0,$1D,$E3,$1F,$99,$FF
	.dc.b	$8B,$FE,$AB,$FB,$D7,$E2,$EF,$99,$F9,$9F,$B7,$BF,$7C,$7E,$B0,$F9
	.dc.b	$B7,$E6,$BC,$7C,$4F,$F1,$9E,$3C,$7C,$26,$3C,$70,$E1,$E6,$EE,$D7
	.dc.b	$B7,$FA,$0E,$9E,$2D,$1F,$A2,$FF,$41,$8C,$65,$D3,$5F,$1F,$E1,$9F
	.dc.b	$E0,$59,$CE,$9D,$FC,$7C,$65,$E4,$CB,$D7,$87,$AB,$77,$CA,$78,$F5
	.dc.b	$F4,$63,$E7,$97,$E5,$FE,$7F,$45,$F8,$63,$F7,$17,$80,$FD,$8D,$F7
	.dc.b	$3F,$E6,$9F,$D9,$78,$5E,$EB,$53,$FE,$6D,$EB,$CF,$E6,$9E,$DE,$2F
	.dc.b	$F2,$DE,$7D,$B3,$C4,$7F,$10,$C7,$96,$7D,$B0,$F6,$CE,$6B,$C7,$3E
	.dc.b	$2F,$CB,$B3,$87,$3C,$FC,$3F,$77,$FD,$E7,$C1,$73,$FF,$8E,$6B,$58
	.dc.b	$7A,$A9,$F9,$97,$AB,$2E,$BB,$FF,$A1,$7F,$2B,$C6,$7D,$3E,$C3,$EE
	.dc.b	$7E,$B9,$FE,$90,$FC,$0D,$F1,$BF,$C5,$FE,$4B,$FC,$91,$F0,$BA,$FB
	.dc.b	$CF,$F5,$5C,$E7,$EA,$BF,$D2,$DC,$5F,$E4,$BC,$79,$72,$FE,$10,$FF
	.dc.b	$3F,$BF,$E9,$CF,$D4,$FE,$2B,$DE,$FB,$78,$5F,$12,$31,$F0,$DF,$DE
	.dc.b	$3E,$BB,$19,$F8,$23,$FB,$DE,$6F,$D1,$3C,$9B,$FA,$27,$3F,$B3,$FE
	.dc.b	$5F,$F4,$8F,$EB,$BF,$8E,$7F,$B1,$3F,$40,$FD,$EF,$E3,$9F,$E9,$CE
	.dc.b	$EF,$97,$C5,$FE,$94,$BD,$FE,$14,$BD,$FE,$F7,$CF,$E5,$FF,$83,$5F
	.dc.b	$B4,$5F,$67,$0F,$2A,$FE,$17,$F8,$07,$E0,$23,$1E,$FB,$F4,$CC,$60
	.dc.b	$D7,$CD,$2D,$5F,$C0,$1E,$19,$CC,$3F,$73,$FF,$58,$7C,$4F,$F9,$EF
	.dc.b	$7F,$EB,$DF,$55,$CF,$EE,$5F,$B9,$BF,$53,$7E,$77,$C3,$8F,$FD,$67
	.dc.b	$87,$0D,$3F,$D1,$57,$BF,$DE,$75,$5F,$EE,$CF,$96,$FF,$25,$3F,$E4
	.dc.b	$0D,$3C,$10,$FC,$F3,$FC,$81,$18,$D7,$C5,$83,$E2,$3F,$D0,$93,$9C
	.dc.b	$BB,$FD,$56,$32,$F1,$E5,$FE,$57,$9F,$1E,$CF,$DE,$7C,$7B,$79,$F8
	.dc.b	$3D,$B2,$FC,$8F,$38,$71,$9C,$FD,$7E,$37,$F1,$DF,$F9,$07,$DD,$FF
	.dc.b	$51,$7C,$CF,$F9,$DF,$F6,$67,$EC,$CF,$D9,$BF,$57,$FB,$3B,$F0,$BF
	.dc.b	$F6,$F6,$1C,$3E,$F1,$87,$0B,$DF,$DD,$6B,$93,$FD,$BF,$FB,$43,$D4
	.dc.b	$7E,$91,$FB,$6F,$D8,$5F,$AB,$BF,$61,$7E,$97,$C7,$16,$3C,$70,$E1
	.dc.b	$7B,$FD,$73,$53,$9A,$FF,$39,$EF,$FE,$BD,$8F,$07,$57,$F9,$CE,$31
	.dc.b	$A7,$8B,$E0,$5A,$F8,$B8,$BF,$D7,$15,$AD,$3B,$FE,$3F,$8C,$FD,$78
	.dc.b	$F1,$C3,$D3,$BF,$F8,$BB,$D0,$FE,$8B,$5A,$C3,$CF,$3E,$D9,$7E,$49
	.dc.b	$FE,$69,$FC,$D5,$F8,$C3,$F2,$97,$D3,$FF,$8E,$3B,$DF,$85,$31,$E3
	.dc.b	$A7,$F9,$BB,$1E,$38,$70,$F8,$BF,$F3,$C7,$F0,$06,$F4,$ED,$F8,$F7
	.dc.b	$FA,$4D,$F0,$6B,$71,$E4,$F8,$75,$FF,$23,$F9,$3F,$ED,$9A,$DE,$5D
	.dc.b	$7E,$6B,$18,$70,$8F,$07,$E0,$6F,$CC,$FC,$13,$9C,$7F,$CA,$32,$FC
	.dc.b	$CF,$33,$FD,$CF,$B3,$FD,$E3,$F2,$5F,$CC,$3F,$6A,$FF,$7C,$7E,$8F
	.dc.b	$AF,$9A,$FC,$0F,$E2,$EA,$D7,$E2,$EF,$D1,$7E,$27,$E5,$9F,$EF,$BF
	.dc.b	$D0,$1F,$A2,$3F,$6C,$78,$EF,$D2,$5F,$B6,$7F,$0D,$FD,$23,$8F,$10
	.dc.b	$63,$CB,$0E,$37,$C3,$E0,$DF,$A8,$70,$BD,$E7,$DF,$3E,$D9,$FF,$1D
	.dc.b	$9F,$B1,$FB,$37,$55,$FB,$4B,$F6,$D7,$E1,$9E,$BF,$39,$F5,$0F,$C0
	.dc.b	$3F,$D9,$7E,$4B,$2E,$58,$F1,$C3,$85,$EF,$F7,$EE,$F9,$FA,$EB,$5C
	.dc.b	$DF,$E1,$8A,$7A,$3E,$FC,$BA,$5E,$F5,$F1,$DE,$FE,$FB,$5F,$94,$7A
	.dc.b	$7F,$70,$7C,$BF,$FB,$AB,$F3,$A7,$ED,$7F,$E2,$8F,$C6,$9F,$B4,$7E
	.dc.b	$5F,$0E,$17,$BA,$D7,$DC,$E7,$5F,$A1,$97,$4F,$C7,$99,$F4,$FB,$57
	.dc.b	$E2,$5E,$2B,$DE,$5C,$E5,$8F,$AF,$F4,$7F,$9D,$F6,$0F,$AE,$70,$7F
	.dc.b	$A2,$3F,$23,$7E,$5B,$FD,$47,$F0,$7F,$D6,$61,$C2,$F7,$F8,$97,$FB
	.dc.b	$EA,$BF,$AB,$79,$BA,$DB,$F0,$D9,$F5,$CF,$A5,$7D,$FE,$96,$5C,$E7
	.dc.b	$3F,$49,$F3,$FF,$28,$FC,$17,$E3,$2F,$D8,$9F,$96,$3F,$5D,$FD,$37
	.dc.b	$13,$F4,$0F,$E8,$1A,$31,$E3,$87,$0B,$DD,$7F,$41,$AD,$6F,$F9,$BC
	.dc.b	$BB,$65,$D7,$D9,$67,$D6,$BF,$86,$67,$D2,$B5,$2F,$C9,$BF,$0D,$F1
	.dc.b	$FF,$99,$9F,$E6,$BF,$A7,$7F,$AA,$BF,$52,$7C,$A3,$F4,$8B,$B3,$0E
	.dc.b	$1F,$2A,$C3,$85,$EF,$F3,$7F,$E3,$95,$A9,$CF,$8F,$E0,$DF,$F6,$9A
	.dc.b	$78,$3F,$16,$6B,$E0,$F2,$FF,$35,$F5,$DF,$AD,$FB,$8F,$A3,$F9,$77
	.dc.b	$EF,$FE,$9F,$FD,$73,$F8,$8B,$F2,$AF,$E5,$57,$BE,$F5,$EF,$E8,$D6
	.dc.b	$F9,$7D,$EF,$FC,$17,$3E,$BF,$01,$FE,$17,$B7,$E3,$BE,$2A,$D6,$5D
	.dc.b	$2B,$5F,$21,$FA,$FF,$8F,$FE,$05,$E4,$BF,$64,$FE,$1F,$F8,$CB,$CB
	.dc.b	$FE,$09,$F8,$13,$F5,$6F,$E8,$18,$70,$BD,$D6,$BE,$17,$3A,$FF,$19
	.dc.b	$CB,$A7,$C4,$FF,$AF,$EB,$F8,$77,$A5,$E1,$AD,$61,$CB,$DD,$FF,$45
	.dc.b	$C7,$F4,$DF,$07,$F8,$B7,$7F,$F6,$4F,$A6,$FE,$47,$D0,$FC,$35,$FB
	.dc.b	$D7,$FA,$23,$C7,$61,$C3,$E8,$98,$70,$BD,$FE,$33,$D7,$3F,$E5,$D9
	.dc.b	$F5,$E0,$F9,$F4,$FB,$FF,$E2,$99,$73,$C7,$97,$E2,$9F,$C5,$FE,$4F
	.dc.b	$F9,$57,$ED,$0F,$DD,$7F,$DD,$5F,$14,$F8,$37,$E6,$6F,$C1,$1F,$AB
	.dc.b	$7F,$69,$C7,$8E,$1C,$2F,$75,$FC,$3B,$91,$5A,$F0,$7F,$9C,$71,$FF
	.dc.b	$39,$9C,$BF,$D1,$F4,$F1,$7E,$33,$A8,$F3,$19,$CB,$B6,$9F,$C5,$A3
	.dc.b	$C4,$67,$06,$BE,$99,$FA,$78,$FF,$88,$EB,$FA,$A7,$E5,$1F,$99,$F2
	.dc.b	$BF,$5F,$F9,$27,$F0,$EF,$A6,$7E,$55,$FE,$05,$FA,$DF,$E9,$D7,$BF
	.dc.b	$61,$7B,$FD,$53,$59,$74,$F4,$9F,$A6,$FA,$63,$23,$0F,$03,$ED,$D7
	.dc.b	$C8,$E6,$BF,$53,$8F,$6F,$39,$F8,$87,$F4,$DD,$3C,$75,$AD,$3F,$92
	.dc.b	$F5,$7F,$51,$F5,$BF,$B2,$53,$F1,$0F,$E7,$1F,$A0,$3D,$4F,$F6,$BF
	.dc.b	$EC,$0F,$DD,$3F,$C4,$58,$70,$BD,$D6,$BD,$BC,$F0,$7F,$7B,$CB,$A7
	.dc.b	$E3,$1E,$DE,$2F,$EC,$33,$9C,$3F,$A3,$3E,$DC,$7A,$7C,$7B,$5F,$17
	.dc.b	$E3,$1F,$CF,$67,$31,$8F,$DA,$7F,$B5,$7E,$57,$FD,$F7,$DA,$7E,$35
	.dc.b	$FB,$BF,$4D,$ED,$FF,$85,$3F,$09,$FD,$45,$EE,$B5,$39,$F4,$BF,$C7
	.dc.b	$D3,$9F,$95,$F4,$3B,$F6,$6F,$07,$F8,$5F,$3E,$F1,$FC,$23,$DD,$FC
	.dc.b	$B5,$6F,$4F,$0F,$B9,$FE,$48,$E4,$7E,$A5,$FE,$13,$FD,$D7,$F8,$FF
	.dc.b	$B7,$FB,$DF,$6D,$BF,$F3,$9F,$95,$FF,$28,$5E,$EB,$F6,$F5,$AF,$8A
	.dc.b	$7E,$63,$AE,$1F,$F1,$56,$5D,$1F,$C4,$61,$C7,$E1,$3E,$93,$CF,$1F
	.dc.b	$06,$6F,$CC,$7F,$C2,$93,$9F,$77,$F8,$4F,$E6,$FF,$E1,$7F,$B3,$FF
	.dc.b	$15,$7F,$00,$3F,$3E,$7E,$9D,$FD,$73,$1F,$F1,$5F,$D8,$AF,$FC,$D0
	.dc.b	$70,$AD,$FC,$B7,$7F,$AB,$FF,$67,$7D,$4E,$B8,$7F,$C8,$59,$F5,$CF
	.dc.b	$F1,$18,$F1,$F8,$1B,$5A,$CB,$9E,$9F,$E3,$AF,$9F,$F3,$4E,$70,$F2
	.dc.b	$67,$FE,$05,$FD,$CF,$FC,$85,$FB,$E0,$7E,$B3,$FC,$43,$81,$FB,$FB
	.dc.b	$F5,$9F,$E6,$1D,$CC,$38,$E1,$C3,$E1,$FF,$DC,$D7,$BF,$CA,$3F,$D0
	.dc.b	$D5,$AF,$B8,$FD,$03,$C3,$3B,$B7,$F9,$7F,$1F,$14,$78,$70,$F3,$4E
	.dc.b	$67,$B6,$9E,$4F,$33,$FC,$91,$F4,$1F,$E5,$BF,$85,$FD,$37,$AF,$FB
	.dc.b	$1F,$B3,$7E,$7B,$FE,$97,$AD,$7A,$5F,$AA,$F6,$FF,$BC,$D3,$E6,$B5
	.dc.b	$AF,$F9,$C7,$2F,$D8,$7D,$AF,$ED,$F1,$E3,$9D,$AD,$65,$CF,$DC,$7E
	.dc.b	$7D,$E4,$9C,$D7,$F0,$3F,$C1,$3F,$CF,$9A,$FF,$AA,$67,$FD,$A7,$E0
	.dc.b	$7F,$BA,$FC,$AF,$F2,$6F,$D3,$BF,$46,$FF,$44,$DE,$EF,$FD,$A3,$7B
	.dc.b	$FB,$1F,$F4,$D5,$7F,$75,$AD,$7E,$05,$F8,$FF,$C6,$7F,$95,$46,$2F
	.dc.b	$FD,$F1,$87,$F4,$58,$C7,$E3,$4C,$FC,$3F,$81,$79,$BC,$1F,$F6,$C6
	.dc.b	$1D,$F5,$7E,$8B,$EA,$EF,$FF,$19,$7E,$3A,$FC,$39,$F9,$7F,$F9,$1F
	.dc.b	$F6,$B7,$EE,$8F,$79,$C1,$F8,$9E,$1F,$F1,$5E,$3C,$70,$E1,$7F,$CB
	.dc.b	$EB,$55,$FE,$54,$AE,$CF,$A7,$2F,$0F,$AF,$EE,$7C,$79,$F8,$BE,$D5
	.dc.b	$A7,$F9,$26,$FC,$35,$AC,$FC,$DF,$4B,$F5,$38,$3F,$D4,$9F,$5A,$FC
	.dc.b	$95,$F8,$CB,$F6,$37,$E8,$EF,$DD,$FF,$CC,$9E,$DF,$B3,$1E,$38,$70
	.dc.b	$BD,$D7,$F9,$A7,$E9,$B5,$B7,$FA,$F7,$3E,$DF,$A0,$FD,$6C,$78,$E9
	.dc.b	$6B,$59,$73,$D3,$C1,$C3,$E7,$9C,$D7,$96,$7C,$9E,$3F,$E3,$BE,$03
	.dc.b	$F9,$67,$E2,$9F,$9D,$F9,$97,$E4,$2F,$CE,$1F,$A5,$BF,$2E,$FA,$B8
	.dc.b	$70,$F4,$7F,$BB,$DE,$EB,$FB,$97,$77,$F3,$8D,$BF,$96,$E7,$FC,$21
	.dc.b	$FC,$66,$1C,$5A,$6B,$0E,$79,$7E,$2A,$3B,$99,$CF,$CB,$1F,$EE,$D8
	.dc.b	$FC,$3F,$81,$FD,$6F,$F2,$D7,$ED,$DF,$E6,$6F,$CB,$9F,$AB,$BF,$19
	.dc.b	$F6,$6D,$F8,$FC,$7F,$AA,$61,$C3,$EB,$18,$70,$B7,$E7,$7D,$AF,$79
	.dc.b	$7E,$A3,$3F,$07,$E7,$B6,$73,$8F,$2D,$3F,$DD,$5A,$77,$97,$CC,$B5
	.dc.b	$F0,$6B,$E6,$FC,$0F,$F8,$0F,$CF,$FF,$81,$FD,$17,$A4,$FD,$51,$F8
	.dc.b	$6F,$FA,$7F,$EF,$3F,$4F,$7B,$AD,$4F,$FA,$8F,$F7,$5C,$EB,$FE,$09
	.dc.b	$C7,$F7,$F7,$FE,$13,$BB,$39,$78,$F7,$78,$A3,$19,$78,$3C,$AF,$E2
	.dc.b	$BF,$51,$FE,$26,$FA,$87,$ED,$FF,$E6,$FE,$67,$E4,$6F,$D9,$7F,$77
	.dc.b	$7B,$AD,$4B,$F3,$9C,$E7,$FE,$3D,$C7,$F8,$9D,$FF,$8C,$BF,$80,$6C
	.dc.b	$E7,$0E,$DF,$75,$E2,$8C,$7C,$EB,$47,$8B,$FC,$AB,$E4,$FF,$CD,$3F
	.dc.b	$BD,$7F,$24,$E5,$FE,$35,$FB,$37,$F0,$E9,$CF,$C1,$FF,$71,$9F,$E3
	.dc.b	$DF,$0B,$9C,$FF,$CC,$07,$F2,$3A,$E6,$D3,$FC,$D1,$87,$5F,$D3,$FC
	.dc.b	$11,$8C,$3A,$7E,$2D,$FC,$F3,$F0,$4F,$F4,$37,$E1,$1F,$AF,$F4,$3F
	.dc.b	$27,$18,$F3,$BF,$57,$E5,$FF,$22,$8C,$79,$CF,$CD,$F9,$67,$FE,$94
	.dc.b	$F1,$3F,$D3,$39,$7E,$E7,$8F,$FA,$4F,$E1,$7F,$EB,$1F,$C2,$5F,$BF
	.dc.b	$7E,$3B,$CD,$FE,$5B,$39,$8F,$C8,$FE,$E3,$8C,$BF,$D5,$F7,$C3,$D1
	.dc.b	$FF,$67,$F2,$74,$9C,$DF,$E9,$F8,$FF,$AB,$FB,$BF,$ED,$5F,$45,$F8
	.dc.b	$77,$EA,$3C,$8F,$ED,$1F,$95,$FF,$3C,$AD,$4E,$79,$FF,$CE,$3E,$9A
	.dc.b	$73,$D1,$F9,$F9,$EB,$A7,$FB,$5B,$2E,$B3,$FA,$3F,$BF,$FE,$CF,$F5
	.dc.b	$FF,$EF,$3F,$C4,$9F,$9F,$FE,$53,$E9,$1F,$B3,$FF,$4C,$AD,$4E,$7D
	.dc.b	$BC,$67,$FE,$F8,$FC,$5B,$FE,$00,$63,$CB,$CB,$7E,$BB,$DA,$B5,$EC
	.dc.b	$AB,$5F,$45,$FD,$E3,$F1,$95,$4F,$C6,$9F,$8E,$3F,$1A,$7E,$38,$FC
	.dc.b	$95,$E6,$7E,$F3,$F4,$CF,$EC,$39,$72,$C9,$97,$2C,$78,$E1,$C3,$E0
	.dc.b	$7F,$B5,$F6,$30,$E1,$7B,$CF,$B6,$6F,$B2,$F3,$67,$E5,$D7,$C9,$5A
	.dc.b	$FD,$07,$FC,$23,$F8,$6F,$87,$C3,$7E,$12,$FC,$01,$F8,$E3,$E6,$5F
	.dc.b	$2C,$FC,$3D,$FA,$7F,$F7,$AC,$B9,$63,$C7,$0E,$17,$BF,$C3,$35,$BF
	.dc.b	$C8,$D3,$BE,$9D,$B4,$EB,$F9,$BF,$92,$B5,$97,$39,$F0,$FE,$03,$FE
	.dc.b	$28,$FC,$9F,$E4,$7E,$4C,$FC,$97,$FD,$DD,$F9,$33,$F2,$77,$E5,$BF
	.dc.b	$E8,$8F,$C5,$7F,$C1,$F8,$70,$BD,$D6,$BF,$09,$7E,$8B,$AD,$BE,$EE
	.dc.b	$7D,$7E,$95,$A7,$5D,$1C,$5E,$3B,$DE,$7D,$25,$EC,$FF,$8F,$BE,$F3
	.dc.b	$EF,$78,$8F,$C3,$DF,$82,$BF,$0B,$7E,$58,$FC,$B1,$83,$F4,$5F,$F1
	.dc.b	$3B,$CB,$0E,$37,$C2,$B7,$F8,$16,$B6,$FC,$AE,$7D,$F3,$ED,$9F,$5F
	.dc.b	$BF,$F2,$56,$B2,$E7,$3E,$1F,$BE,$FF,$2A,$FE,$63,$F9,$5F,$86,$7E
	.dc.b	$26,$FC,$29,$F9,$87,$F0,$2F,$CC,$1F,$B5,$3F,$7A,$FF,$1E,$E3,$C7
	.dc.b	$0E,$17,$BA,$7E,$8D,$FE,$42,$CF,$F4,$5F,$9A,$FE,$9B,$F1,$B7,$E0
	.dc.b	$CF,$D2,$3F,$BE,$79,$4F,$D5,$F8,$72,$7E,$CB,$0E,$17,$BF,$84,$5F
	.dc.b	$1F,$6F,$E6,$DF,$E4,$7F,$7A,$7E,$B6,$FD,$E1,$FB,$27,$F9,$57,$D0
	.dc.b	$7E,$DD,$FE,$56,$EB,$FD,$36,$1C,$2F,$7F,$71,$FA,$5F,$DC,$B5,$9F
	.dc.b	$5C,$FF,$CD,$F8,$70,$FD,$72,$CE,$70,$E3,$F0,$FF,$E8,$2A,$EB,$7E
	.dc.b	$3D,$7F,$CC,$DE,$F6,$70,$E1,$1E,$5E,$2E,$F1,$F9,$2D,$7C,$93,$E9
	.dc.b	$FA,$DC,$F9,$C9,$CE,$FF,$BB,$F8,$BC,$C6,$38,$BE,$0F,$1F,$C6,$C9
	.dc.b	$F8,$93,$F3,$67,$E9,$0F,$D2,$BF,$CC,$7F,$A4,$3F,$4A,$FE,$79,$8B
	.dc.b	$F2,$DF,$F5,$97,$EE,$CF,$CE,$7C,$5C,$9F,$AA,$F9,$37,$CE,$FF,$A0
	.dc.b	$AF,$75,$AF,$6F,$F9,$87,$E2,$D9,$E1,$FF,$59,$65,$FE,$98,$7F,$94
	.dc.b	$19,$1E,$AA,$3C,$11,$FE,$9D,$C3,$AF,$27,$E3,$30,$7E,$53,$FE,$B2
	.dc.b	$9C,$EB,$FB,$EF,$B8,$C6,$1E,$0D,$3C,$9B,$38,$7C,$1A,$F7,$9C,$E1
	.dc.b	$E3,$F5,$5F,$94,$BE,$EF,$FB,$E7,$DC,$3E,$FD,$F9,$D3,$F0,$1E,$17
	.dc.b	$BF,$B4,$5E,$EB,$FD,$53,$D1,$F3,$78,$3F,$A8,$E7,$DB,$3F,$F5,$EF
	.dc.b	$17,$F4,$B8,$F0,$F8,$58,$AD,$E7,$D7,$2F,$17,$D3,$7B,$A7,$FD,$AF
	.dc.b	$39,$D9,$F9,$96,$71,$ED,$B7,$92,$7D,$27,$A3,$4F,$6B,$EC,$F8,$2F
	.dc.b	$F5,$EF,$EA,$5F,$D7,$7E,$B7,$FC,$DB,$F1,$FF,$D8,$BF,$6E,$7E,$19
	.dc.b	$FD,$EA,$F7,$B7,$F6,$CB,$DF,$CF,$AF,$7F,$50,$FC,$4F,$76,$BF,$F3
	.dc.b	$CF,$D8,$7F,$B7,$B4,$F0,$7D,$73,$5C,$5F,$58,$C6,$7D,$75,$FE,$77
	.dc.b	$3D,$31,$F0,$E9,$E6,$F8,$FF,$F7,$BE,$5D,$32,$F5,$61,$FE,$12,$FD
	.dc.b	$0B,$86,$7D,$30,$F3,$CE,$2B,$7B,$3F,$24,$C6,$BE,$18,$FF,$6F,$65
	.dc.b	$FE,$03,$FD,$1F,$FA,$D7,$C4,$3F,$2C,$F7,$7B,$1F,$E0,$CF,$C5,$FE
	.dc.b	$0D,$6A,$73,$F4,$DF,$E9,$CF,$05,$8F,$3F,$9C,$65,$F0,$B9,$58,$FF
	.dc.b	$67,$FD,$7B,$0F,$3E,$6F,$E4,$BF,$89,$FC,$0A,$FF,$0F,$CE,$76,$7D
	.dc.b	$FE,$67,$9E,$5D,$B6,$F1,$FD,$2F,$9A,$31,$97,$F3,$0D,$FF,$DD,$7B
	.dc.b	$F9,$B0,$FF,$15,$F5,$3F,$0F,$7E,$B1,$FE,$3E,$FD,$5B,$FB,$17,$E7
	.dc.b	$EF,$DF,$FF,$03,$7B,$AD,$7C,$FA,$BF,$D8,$59,$75,$CB,$FC,$AF,$93
	.dc.b	$F2,$DF,$85,$8F,$19,$FE,$16,$74,$9F,$0F,$4B,$4F,$EE,$5E,$57,$BE
	.dc.b	$FC,$13,$9C,$FF,$4F,$87,$1D,$3C,$7A,$7F,$A9,$74,$F1,$61,$C3,$2F
	.dc.b	$5F,$DB,$A7,$FC,$07,$87,$F9,$33,$F1,$6F,$F7,$AF,$C4,$3F,$35,$FD
	.dc.b	$9F,$BB,$FE,$B0,$F8,$8E,$1F,$41,$FE,$49,$BD,$D6,$BF,$14,$7E,$59
	.dc.b	$AC,$BA,$7E,$1A,$CF,$DF,$19,$3F,$DC,$71,$FC,$F5,$F8,$7A,$79,$1C
	.dc.b	$CF,$8F,$D3,$FE,$5D,$EF,$FF,$98,$AB,$59,$F8,$A7,$F9,$97,$E1,$59
	.dc.b	$D7,$CD,$97,$87,$5F,$37,$D7,$78,$70,$E1,$C1,$FE,$CC,$FA,$DF,$27
	.dc.b	$EB,$1F,$E6,$8F,$CB,$9F,$B0,$FE,$73,$F3,$A7,$E4,$8F,$B1,$DF,$A0
	.dc.b	$BD,$F9,$EB,$DF,$C9,$EA,$7E,$0F,$27,$B7,$3F,$F3,$AF,$C3,$E3,$8F
	.dc.b	$FC,$E9,$1E,$1D,$7C,$1D,$1E,$B3,$FD,$03,$19,$CF,$B7,$66,$65,$39
	.dc.b	$D3,$CF,$FA,$1F,$CF,$FC,$81,$FA,$BB,$F7,$77,$E7,$DF,$E8,$6F,$15
	.dc.b	$D6,$FB,$9E,$17,$BA,$D4,$FF,$70,$E0,$FF,$45,$63,$D3,$1F,$56,$BF
	.dc.b	$E9,$68,$EF,$FB,$5E,$21,$1B,$C3,$C1,$F0,$BE,$98,$C7,$3F,$EC,$BB
	.dc.b	$FD,$9F,$B9,$BF,$60,$7B,$4F,$C3,$DF,$B0,$B9,$FF,$D1,$3F,$9E,$FE
	.dc.b	$37,$0E,$17,$BA,$D7,$E0,$7F,$A1,$ED,$FC,$75,$6B,$83,$F2,$3C,$7F
	.dc.b	$EA,$58,$F0,$57,$A5,$E7,$9F,$83,$37,$D8,$79,$B5,$F2,$CE,$7E,$37
	.dc.b	$1A,$F8,$6F,$FD,$3B,$E5,$FE,$87,$DF,$FF,$20,$FE,$7D,$FE,$89,$F9
	.dc.b	$6F,$F5,$57,$E8,$9F,$E8,$BE,$1F,$D1,$56,$BD,$EF,$F0,$E7,$E1,$4A
	.dc.b	$7E,$DB,$CF,$EA,$19,$7C,$FC,$3F,$C9,$BF,$86,$59,$CE,$3C,$BF,$4D
	.dc.b	$FC,$45,$75,$D3,$FD,$75,$EA,$7E,$2F,$F5,$6F,$F2,$BF,$CF,$FE,$6F
	.dc.b	$91,$FA,$D7,$F9,$9B,$F5,$9F,$DD,$56,$BC,$BF,$A9,$3F,$E6,$AF,$DA
	.dc.b	$73,$F1,$8F,$D9,$FF,$05,$7D,$3C,$0F,$9E,$7F,$B7,$1E,$59,$76,$FA
	.dc.b	$2F,$15,$FF,$B8,$F2,$EB,$F9,$C6,$35,$F2,$69,$FE,$78,$FB,$FF,$E1
	.dc.b	$9E,$7F,$E1,$3B,$3F,$0D,$F8,$33,$F7,$AF,$F6,$76,$5A,$F1,$FF,$8A
	.dc.b	$6D,$FE,$E7,$9F,$F6,$97,$CB,$E7,$0F,$CC,$63,$F8,$EF,$AE,$B8,$70
	.dc.b	$CB,$A7,$E5,$5F,$B1,$AF,$06,$5C,$FE,$57,$F6,$99,$BE,$B5,$E3,$FF
	.dc.b	$B6,$FF,$22,$7D,$D3,$F8,$03,$F0,$1E,$7F,$FB,$D7,$0E,$17,$BA,$D7
	.dc.b	$EF,$9F,$CA,$67,$A7,$E1,$BF,$AD,$77,$D9,$F3,$0F,$81,$9F,$83,$F1
	.dc.b	$76,$9E,$0E,$4F,$2C,$E6,$3C,$7F,$79,$FC,$2F,$1F,$C5,$FE,$22,$FC
	.dc.b	$A5,$F9,$CB,$F4,$1F,$B9,$FB,$1F,$F3,$BB,$DD,$6A,$7F,$01,$FC,$01
	.dc.b	$9F,$3D,$FA,$4F,$F0,$3B,$EB,$C1,$E7,$78,$E5,$D7,$EE,$7C,$56,$D3
	.dc.b	$C5,$3E,$1F,$95,$FE,$41,$ED,$3F,$61,$7E,$C1,$FE,$DC,$F9,$CF,$E3
	.dc.b	$BF,$86,$3F,$0C,$FE,$A1,$7B,$AD,$7C,$3F,$FA,$F7,$D3,$CE,$2F,$DC
	.dc.b	$3F,$8C,$5E,$F8,$7E,$63,$C7,$F0,$86,$9F,$CA,$BE,$03,$C7,$87,$8A
	.dc.b	$73,$97,$FA,$F3,$0F,$EC,$1F,$79,$FE,$2D,$FB,$A4,$3F,$24,$7E,$EE
	.dc.b	$E2,$7E,$E9,$C3,$EC,$3F,$CF,$B7,$BA,$D7,$1E,$71,$FE,$5D,$97,$F8
	.dc.b	$E7,$F0,$8B,$87,$0C,$FA,$E7,$FD,$1E,$BC,$59,$73,$E9,$FE,$95,$E4
	.dc.b	$F9,$3D,$87,$CF,$FE,$A7,$F0,$5F,$E9,$BF,$BC,$FF,$79,$AD,$4E,$7E
	.dc.b	$69,$F9,$A6,$72,$7E,$1F,$FD,$06,$FA,$47,$E5,$5E,$3E,$6F,$FB,$1F
	.dc.b	$F2,$EF,$0C,$FF,$B9,$B1,$F0,$7D,$87,$93,$F4,$47,$E9,$FF,$ED,$6F
	.dc.b	$DC,$3F,$C6,$9E,$7F,$8E,$73,$F8,$97,$F8,$3E,$7F,$C7,$1F,$9E,$67
	.dc.b	$3F,$15,$72,$CF,$C4,$F0,$F4,$D7,$FD,$2E,$7A,$65,$DD,$FE,$DD,$FB
	.dc.b	$AB,$F5,$1F,$8B,$FB,$6F,$FA,$1F,$F4,$DF,$EA,$BF,$09,$FE,$1D,$AD
	.dc.b	$70,$AB,$5F,$33,$F8,$8F,$99,$FF,$0F,$56,$B2,$FD,$C3,$D7,$F7,$AF
	.dc.b	$ED,$0F,$2B,$ED,$E6,$78,$B4,$F0,$FE,$0C,$D7,$F8,$6E,$3E,$1F,$85
	.dc.b	$FD,$27,$7F,$F6,$9F,$25,$F9,$27,$F3,$3D,$BC,$35,$AF,$9D,$FC,$7C
	.dc.b	$F3,$9C,$FA,$9E,$2A,$FD,$76,$3E,$67,$86,$0E,$8F,$8B,$18,$EC,$F8
	.dc.b	$FF,$0F,$F8,$BF,$89,$7D,$03,$F6,$57,$E0,$DF,$1E,$B5,$39,$FB,$5C
	.dc.b	$7D,$A7,$C2,$9F,$F2,$4E,$BB,$1C,$5F,$E0,$2B,$F1,$46,$63,$1F,$75
	.dc.b	$F6,$39,$FF,$5F,$EB,$F0,$F7,$1F,$AF,$FF,$88,$39,$7E,$9D,$6B,$0F
	.dc.b	$E1,$72,$F8,$AF,$B9,$3E,$2A,$DE,$BF,$5A,$38,$61,$FE,$04,$CF,$67
	.dc.b	$D2,$F7,$7F,$7E,$FE,$17,$CB,$B9,$FE,$7F,$E8,$71,$3F,$B1,$8F,$D8
	.dc.b	$F9,$98,$8F,$C2,$D7,$F9,$49,$A6,$F7,$FC,$AC,$74,$AF,$87,$A7,$E3
	.dc.b	$BC,$17,$47,$B6,$5F,$55,$F9,$79,$CE,$3F,$88,$9F,$F1,$85,$F1,$BF
	.dc.b	$C9,$D3,$F1,$1B,$1F,$1F,$E0,$7F,$4B,$3C,$B9,$3E,$AF,$D0,$E3,$9F
	.dc.b	$C9,$D7,$DD,$F3,$67,$74,$FD,$0E,$0F,$CE,$BB,$D7,$C6,$CB,$95,$7F
	.dc.b	$15,$C7,$BF,$2F,$CC,$CF,$F0,$7E,$97,$AE,$31,$9F,$E1,$B7,$F8,$51
	.dc.b	$EA,$FD,$83,$D6,$F1,$CB,$9E,$5F,$86,$CB,$D7,$C5,$F6,$23,$AE,$1F
	.dc.b	$2E,$31,$8F,$57,$EE,$EC,$3F,$D3,$55,$F8,$0A,$E0,$FE,$0B,$0E,$AF
	.dc.b	$E1,$30,$F9,$35,$F4,$A3,$FA,$E6,$7F,$56,$31,$A7,$79,$FE,$93,$1D
	.dc.b	$CE,$1F,$07,$CB,$87,$F1,$FF,$8F,$F4,$E0,$F5,$63,$CD,$F4,$E9,$F8
	.dc.b	$5C,$3F,$D0,$18,$FB,$E3,$13,$F9,$8C,$7C,$BB,$FC,$3F,$13,$C8,$FC
	.dc.b	$5D,$2E,$1C,$77,$F8,$79,$7C,$3A,$E5,$3E,$B9,$F9,$9D,$76,$38,$1E
	.dc.b	$CA,$F9,$53,$C3,$3F,$34,$7C,$59,$F8,$8F,$C7,$AB,$B7,$F6,$CC,$FE
	.dc.b	$5C,$FB,$2A,$E9,$FB,$2B,$E2,$EE,$39,$E5,$E2,$7E,$65,$7A,$5F,$EE
	.dc.b	$F5,$A7,$FA,$36,$3E,$4F,$B5,$FF,$39,$FE,$07,$F8,$18,$C7,$D4,$A1
	.dc.b	$AF,$A2,$BF,$CF,$3D,$5E,$67,$FC,$F7,$F8,$87,$A2,$7C,$59,$F8,$9F
	.dc.b	$46,$3F,$09,$F4,$3F,$12,$3F,$10,$DD,$FC,$ED,$DF,$E1,$CF,$C3,$8E
	.dc.b	$1B,$EA,$FE,$BC,$FE,$2E,$3D,$B1,$88,$F8,$5A,$FD,$07,$F8,$F6,$D7
	.dc.b	$7D,$CF,$D9,$9F,$CA,$57,$E8,$4E,$4F,$C2,$8F,$F7,$5E,$1E,$99,$E3
	.dc.b	$1F,$42,$7F,$39,$97,$A1,$FE,$33,$87,$85,$67,$9C,$A4,$E7,$E3,$38
	.dc.b	$67,$DB,$2F,$0E,$1E,$1C,$FE,$43,$F5,$2F,$A3,$FC,$9A,$32,$FF,$1F
	.dc.b	$6E,$1A,$E9,$DE,$1C,$3A,$7F,$C4,$3F,$43,$FC,$FB,$E0,$DF,$92,$3F
	.dc.b	$52,$7E,$20,$FD,$F1,$E2,$3F,$33,$7C,$CF,$2E,$6F,$F2,$2E,$5C,$B1
	.dc.b	$E3,$87,$0B,$7E,$02,$FB,$EF,$F3,$EE,$1F,$E4,$CF,$86,$7E,$A8,$FD
	.dc.b	$BF,$FB,$87,$E6,$0F,$D1,$5F,$9C,$3F,$0E,$E5,$F1,$7F,$D5,$B1,$E3
	.dc.b	$87,$0B,$DF,$E0,$AF,$C9,$3E,$8E,$9F,$E5,$AF,$C7,$1F,$A6,$FF,$9A
	.dc.b	$7F,$44,$7E,$6C,$FC,$6B,$C1,$F0,$9C,$7C,$7F,$EC,$78,$70,$F6,$58
	.dc.b	$70,$BF,$37,$EC,$EF,$CF,$9F,$AA,$FF,$6A,$FD,$B7,$FD,$25,$F9,$B3
	.dc.b	$F5,$07,$E2,$6F,$C8,$1F,$AB,$FF,$A8,$31,$E3,$C7,$FE,$70,$C7,$8E
	.dc.b	$1C,$2F,$7F,$34,$FD,$97,$F8,$47,$90,$FC,$55,$E5,$BF,$1C,$7E,$A2
	.dc.b	$F0,$DF,$4D,$FE,$79,$C7,$8E,$1C,$2F,$75,$AF,$C2,$3B,$7A,$7F,$E0
	.dc.b	$4C,$FB,$67,$D7,$3E,$9C,$FF,$E8,$6C,$F9,$CE,$71,$7D,$E7,$FA,$2B
	.dc.b	$EF,$FE,$77,$1F,$FA,$33,$C3,$7E,$24,$F8,$57,$82,$F9,$1F,$91,$F9
	.dc.b	$EB,$BD,$97,$2C,$BF,$0D,$8F,$1C,$38,$74,$B0,$E7,$7B,$CF,$BE,$7D
	.dc.b	$B3,$FA,$9E,$F7,$97,$87,$FC,$43,$5A,$CB,$9F,$26,$75,$F4,$6B,$DF
	.dc.b	$1F,$B9,$FA,$43,$F5,$DF,$2E,$5C,$AB,$5B,$AB,$FC,$A5,$B7,$8E,$73
	.dc.b	$F5,$EF,$AF,$E3,$9F,$6E,$0F,$36,$2A,$E5,$87,$C1,$F8,$E7,$EF,$CF
	.dc.b	$C6,$5F,$A0,$BF,$0C,$7E,$0F,$FE,$9E,$FA,$1E,$BA,$FF,$95,$30,$E1
	.dc.b	$7B,$AF,$F5,$1F,$AB,$AC,$BA,$E5,$D3,$2E,$7F,$29,$F0,$F2,$E5,$39
	.dc.b	$C1,$F2,$99,$C3,$8E,$5D,$57,$E3,$9D,$74,$F0,$E7,$E6,$9F,$D5,$6D
	.dc.b	$ED,$9C,$E2,$D1,$F7,$BC,$74,$F0,$43,$F0,$6F,$6F,$EB,$0F,$CA,$5E
	.dc.b	$77,$F5,$CF,$C9,$1F,$86,$FF,$93,$3F,$4C,$DD,$6B,$E4,$D5,$AF,$9F
	.dc.b	$7C,$A6,$B1,$E5,$8F,$F9,$53,$9F,$18,$70,$87,$E3,$9F,$17,$4F,$15
	.dc.b	$F0,$F2,$5F,$25,$C3,$4F,$1C,$3E,$4B,$18,$BF,$1A,$E5,$BF,$96,$FD
	.dc.b	$3C,$DF,$EF,$CE,$7F,$D1,$69,$DB,$6F,$14,$FC,$39,$7E,$AD,$FE,$7C
	.dc.b	$FC,$E9,$FB,$47,$F3,$AF,$81,$7D,$03,$ED,$1F,$9D,$7F,$80,$B1,$E3
	.dc.b	$CC,$18,$F2,$F9,$7F,$E4,$78,$71,$BF,$F2,$65,$F0,$BD,$E5,$DE,$BB
	.dc.b	$7E,$1F,$FC,$0F,$4F,$24,$E7,$37,$E2,$F9,$C3,$C1,$A7,$7D,$7C,$F4
	.dc.b	$C5,$EE,$FF,$04,$CB,$9E,$DE,$4A,$D7,$E0,$99,$CB,$9E,$9D,$FF,$44
	.dc.b	$7E,$45,$99,$F3,$E1,$C6,$7F,$9D,$F0,$7A,$F6,$F1,$70,$7C,$7F,$3F
	.dc.b	$CD,$83,$EC,$FF,$75,$F0,$0F,$BF,$7C,$6B,$F3,$07,$E8,$8F,$C4,$3C
	.dc.b	$18,$70,$BD,$D6,$BF,$1D,$CF,$B1,$FC,$EA,$73,$F9,$77,$82,$F8,$65
	.dc.b	$D7,$F4,$0C,$5E,$E3,$C1,$4F,$CC,$B9,$EB,$E5,$9C,$FE,$69,$8C,$BC
	.dc.b	$3F,$9A,$79,$B9,$3E,$D6,$9D,$B6,$4B,$3E,$18,$FB,$7F,$37,$FE,$7D
	.dc.b	$F8,$67,$FB,$2B,$EF,$3F,$DF,$7F,$2F,$FD,$4F,$F4,$FF,$E8,$7F,$34
	.dc.b	$FC,$47,$F9,$DD,$EF,$E0,$BF,$B2,$FC,$A2,$F7,$F6,$9A,$EB,$FC,$96
	.dc.b	$7D,$2B,$5F,$22,$FC,$F3,$C3,$5A,$C7,$9E,$3F,$D4,$B6,$CC,$FF,$BB
	.dc.b	$27,$37,$FD,$E3,$D9,$ED,$39,$DF,$D1,$3F,$DF,$3D,$BC,$E9,$E1,$CB
	.dc.b	$FA,$EF,$EF,$5F,$DF,$79,$DC,$1E,$BC,$78,$CB,$F5,$8E,$BC,$6F,$D6
	.dc.b	$7F,$87,$FE,$3C,$FC,$5B,$FD,$FF,$F9,$93,$F6,$9D,$D6,$A7,$3F,$A2
	.dc.b	$E3,$DA,$61,$C7,$83,$F5,$3C,$79,$8C,$7C,$E7,$C0,$D3,$C1,$7C,$31
	.dc.b	$60,$FD,$BD,$F6,$FC,$67,$30,$FC,$CB,$14,$FD,$0F,$F8,$C7,$2F,$C0
	.dc.b	$D7,$F8,$36,$7D,$B5,$F2,$57,$F2,$0A,$D5,$3F,$17,$FF,$69,$7E,$10
	.dc.b	$FC,$7F,$FC,$81,$FA,$EB,$F0,$17,$EF,$1F,$52,$F7,$5A,$9F,$EB,$D8
	.dc.b	$F4,$9C,$FD,$1B,$0F,$D9,$65,$C7,$F0,$36,$7D,$EB,$5E,$D7,$E1,$30
	.dc.b	$E3,$87,$87,$F0,$FE,$17,$B8,$F2,$7E,$5D,$8D,$7C,$F4,$FC,$53,$19
	.dc.b	$78,$63,$F8,$0E,$AF,$C7,$FF,$95,$6C,$FD,$2F,$F3,$1F,$26,$FE,$00
	.dc.b	$FF,$33,$FE,$D1,$FE,$B2,$FD,$F7,$FC,$3F,$87,$0B,$DD,$6B,$E3,$FF
	.dc.b	$C9,$39,$75,$CB,$F9,$B5,$6B,$F0,$57,$EE,$FE,$0A,$D5,$78,$7D,$1F
	.dc.b	$EC,$B8,$F8,$3E,$63,$FC,$6D,$9F,$79,$CE,$2D,$5E,$A7,$E0,$72,$E7
	.dc.b	$A7,$79,$CF,$C8,$7A,$34,$F1,$E3,$CA,$7F,$80,$7E,$7F,$99,$F1,$70
	.dc.b	$FC,$1C,$FB,$7A,$EF,$CF,$BF,$95,$7E,$4B,$F2,$3F,$1B,$7E,$32,$F8
	.dc.b	$67,$E6,$6F,$C8,$DF,$A0,$3F,$40,$FF,$11,$FC,$33,$2E,$58,$F1,$C3
	.dc.b	$85,$EF,$F0,$1E,$FC,$1F,$D7,$F4,$EF,$A7,$6D,$3A,$FC,$3F,$F6,$3D
	.dc.b	$3A,$5E,$F3,$E9,$5E,$2F,$C8,$75,$AF,$83,$1E,$5A,$3E,$C3,$FD,$59
	.dc.b	$97,$2C,$9A,$78,$36,$F2,$D7,$EC,$6B,$53,$9C,$FD,$F2,$FD,$37,$D3
	.dc.b	$87,$2D,$7C,$79,$77,$F8,$0F,$EA,$9F,$A6,$FE,$4B,$CF,$FE,$A7,$E7
	.dc.b	$3F,$4A,$F9,$5F,$B2,$FF,$92,$3E,$37,$79,$FF,$80,$6F,$75,$AF,$AF
	.dc.b	$7E,$F7,$AA,$FD,$BF,$CB,$32,$E9,$C2,$CF,$EC,$56,$F2,$E7,$39,$FB
	.dc.b	$DF,$99,$AF,$97,$1E,$5A,$77,$38,$38,$AE,$9A,$3F,$4D,$CE,$9E,$0D
	.dc.b	$3F,$D9,$5F,$A5,$63,$1F,$6F,$A1,$FE,$2D,$E7,$69,$FB,$4F,$CC,$BF
	.dc.b	$CD,$7F,$95,$BF,$6F,$7E,$16,$F8,$7F,$F5,$BF,$B5,$EF,$BD,$D6,$A7
	.dc.b	$FD,$77,$F9,$0E,$7E,$B1,$8F,$3C,$79,$7D,$07,$BB,$A3,$FC,$59,$97
	.dc.b	$29,$ED,$F9,$16,$65,$39,$C9,$FA,$97,$FB,$97,$7F,$4D,$72,$CF,$BE
	.dc.b	$BE,$3A,$7E,$7B,$E3,$D7,$C5,$8F,$2F,$CF,$F1,$7B,$FD,$93,$ED,$7E
	.dc.b	$FB,$FE,$E9,$FD,$A5,$FB,$B3,$F3,$BF,$F0,$F7,$EC,$8F,$A0,$DF,$E3
	.dc.b	$CB,$DD,$6B,$F4,$9C,$C6,$3D,$2F,$CB,$71,$E3,$2E,$91,$FE,$D8,$8C
	.dc.b	$5F,$6F,$C1,$EC,$B3,$FF,$77,$72,$FC,$0B,$6B,$E7,$C7,$9F,$DC,$7F
	.dc.b	$8F,$2B,$53,$9C,$BF,$C6,$D3,$FE,$F6,$9F,$F4,$1D,$FF,$97,$7F,$16
	.dc.b	$7E,$D7,$FC,$EF,$2F,$E1,$FD,$9F,$D9,$3F,$55,$FF,$1E,$7E,$01,$FE
	.dc.b	$E3,$AD,$7A,$3F,$ED,$9F,$9B,$7E,$D5,$AC,$FA,$56,$B1,$FF,$0B,$46
	.dc.b	$46,$7E,$0F,$8F,$F7,$63,$FE,$E3,$F6,$BF,$87,$67,$DE,$73,$87,$F7
	.dc.b	$3F,$D0,$79,$61,$C7,$6F,$3F,$CD,$FF,$7C,$E0,$F6,$E9,$DB,$5F,$14
	.dc.b	$BF,$39,$E3,$18,$FB,$AF,$27,$E5,$0F,$93,$7E,$73,$FD,$57,$E1,$5F
	.dc.b	$97,$BE,$F9,$D3,$C2,$F7,$F2,$9F,$CE,$6B,$5C,$19,$C7,$94,$71,$DF
	.dc.b	$15,$BF,$A3,$F7,$C7,$FA,$3F,$4F,$15,$F0,$FD,$9D,$FA,$17,$0D,$BC
	.dc.b	$70,$FD,$EB,$19,$FF,$AC,$B1,$B2,$FD,$7A,$FE,$6F,$FE,$75,$FD,$3F
	.dc.b	$FB,$9F,$E0,$BF,$CA,$7F,$70,$7E,$D2,$FC,$F1,$FA,$DB,$F6,$1F,$F7
	.dc.b	$5F,$6B,$1E,$3F,$1C,$C7,$8E,$1C,$2D,$FB,$6F,$7D,$FF,$D5,$5A,$77
	.dc.b	$D3,$B6,$9F,$E3,$8D,$3A,$56,$B3,$E9,$5A,$FC,$3B,$FA,$86,$BE,$0D
	.dc.b	$7B,$E3,$E5,$EC,$ED,$8F,$FA,$EF,$1E,$3B,$78,$F8,$73,$AF,$86,$3D
	.dc.b	$BF,$A2,$BF,$69,$CD,$F0,$C9,$2F,$C9,$3A,$FC,$47,$F9,$0F,$D3,$FF
	.dc.b	$0D,$7E,$69,$F1,$7E,$09,$FA,$4E,$BC,$8F,$F7,$FF,$D2,$2B,$5F,$A6
	.dc.b	$67,$0E,$3D,$4C,$7D,$BC,$7F,$EC,$FC,$FB,$63,$C6,$9E,$6E,$AB,$FA
	.dc.b	$66,$7D,$B3,$F1,$7E,$87,$D7,$0E,$19,$7E,$07,$AF,$E6,$E1,$C3,$4F
	.dc.b	$0C,$FF,$0D,$9C,$E9,$E0,$C7,$D9,$FA,$BF,$CC,$F3,$7F,$47,$D8,$FE
	.dc.b	$01,$7E,$C6,$FC,$57,$F5,$1F,$31,$BF,$27,$FD,$AD,$5A,$9F,$D6,$7E
	.dc.b	$D9,$9C,$39,$61,$C7,$93,$D9,$97,$6C,$38,$6F,$8C,$B9,$DF,$7F,$3F
	.dc.b	$F9,$87,$07,$B3,$0E,$13,$9F,$D4,$7B,$63,$E1,$B7,$D6,$E2,$B9,$61
	.dc.b	$FE,$D7,$FD,$D3,$FC,$AB,$F8,$F7,$E6,$FE,$F7,$F7,$FF,$97,$3F,$51
	.dc.b	$7E,$78,$FD,$97,$87,$58,$61,$C7,$D1,$61,$C6,$DF,$87,$70,$BD,$F9
	.dc.b	$BF,$E9,$A7,$FC,$CA,$7F,$8B,$74,$F1,$C3,$F0,$CC,$CE,$4F,$06,$5D
	.dc.b	$F1,$F3,$7B,$19,$9E,$BB,$7A,$B6,$7E,$65,$FE,$39,$CB,$FC,$E7,$EA
	.dc.b	$FA,$3E,$F1,$FB,$73,$F7,$3F,$F6,$97,$E8,$5F,$E7,$7E,$5F,$DF,$69
	.dc.b	$EF,$C3,$85,$EF,$EB,$16,$FD,$6D,$87,$0C,$3F,$89,$70,$B3,$F5,$C6
	.dc.b	$B4,$F1,$CF,$F8,$A6,$3D,$F5,$C7,$3E,$FA,$7F,$B9,$7D,$7F,$55,$78
	.dc.b	$EF,$A6,$BE,$3E,$1F,$F2,$B7,$EF,$9F,$E9,$0F,$DA,$7F,$D6,$1F,$38
	.dc.b	$FE,$00,$F2,$F4,$FF,$01,$FB,$BF,$3D,$EF,$F0,$8F,$F0,$CD,$6B,$DA
	.dc.b	$7C,$2E,$B6,$FC,$06,$7F,$EB,$BF,$7B,$D7,$C6,$CB,$96,$B8,$6E,$5E
	.dc.b	$0F,$CF,$7C,$B8,$F2,$D3,$BE,$1F,$86,$FB,$1F,$F3,$77,$D1,$FF,$AA
	.dc.b	$3B,$FF,$C6,$7E,$DF,$F8,$03,$ED,$7F,$CC,$FF,$A8,$3F,$11,$F9,$F7
	.dc.b	$BF,$A9,$FE,$27,$F3,$9F,$E6,$FF,$B8,$5B,$F7,$AD,$EF,$F9,$9D,$3F
	.dc.b	$C7,$FF,$05,$EA,$CB,$96,$B6,$B5,$9F,$4D,$7F,$CE,$FD,$7E,$69,$CD
	.dc.b	$79,$74,$FC,$7C,$76,$3F,$47,$7C,$63,$C7,$F9,$AE,$73,$F6,$ED,$E7
	.dc.b	$DB,$FB,$8F,$0F,$F9,$03,$1E,$B3,$9F,$B2,$E1,$97,$3B,$7C,$07,$7E
	.dc.b	$BF,$EF,$3F,$5B,$F1,$75,$7E,$D5,$FE,$D3,$FA,$8D,$7B,$5F,$E4,$EE
	.dc.b	$7F,$F4,$8C,$FF,$B5,$FE,$9F,$38,$7C,$1D,$FF,$03,$6F,$CD,$B1,$9F
	.dc.b	$6C,$78,$F9,$FF,$0E,$F7,$87,$87,$27,$36,$70,$E1,$AF,$93,$F0,$9F
	.dc.b	$E1,$DC,$9E,$FC,$78,$E7,$E4,$9F,$F4,$54,$E7,$F6,$87,$E7,$3F,$C4
	.dc.b	$78,$FF,$CA,$5F,$89,$BF,$54,$FE,$81,$DB,$FE,$F0,$BD,$E7,$F8,$2F
	.dc.b	$AD,$5E,$FE,$37,$EB,$56,$AB,$F6,$37,$C7,$27,$E4,$3F,$0E,$B9,$E7
	.dc.b	$E0,$F0,$FE,$04,$63,$3E,$BE,$2E,$30,$F3,$47,$83,$5F,$F3,$6E,$DE
	.dc.b	$B9,$CE,$DE,$9F,$1B,$F1,$1F,$5A,$FD,$A9,$F8,$9F,$F9,$8F,$F1,$77
	.dc.b	$C7,$3F,$7E,$FF,$63,$71,$FF,$33,$E4,$FF,$3C,$E1,$C3,$E9,$18,$70
	.dc.b	$BF,$F6,$8D,$EF,$2E,$D9,$7D,$F7,$C1,$FB,$AD,$C7,$96,$1E,$0F,$CF
	.dc.b	$3E,$F5,$70,$CB,$AE,$8F,$C5,$73,$87,$08,$F3,$7E,$61,$EC,$8C,$6A
	.dc.b	$FD,$83,$1A,$FB,$6F,$D9,$C5,$8C,$BA,$DB,$F5,$AF,$E5,$BF,$AD,$7F
	.dc.b	$8E,$BF,$2A,$FF,$80,$1F,$8C,$BE,$AD,$F8,$C3,$E6,$5F,$8B,$F0,$E8
	.dc.b	$FF,$19,$E1,$C2,$F7,$4F,$CA,$35,$4F,$A9,$5B,$F1,$7F,$E3,$FA,$78
	.dc.b	$7F,$52,$B7,$D2,$5A,$F8,$F5,$F1,$4F,$4C,$79,$D6,$B4,$F3,$DF,$F6
	.dc.b	$BE,$F6,$5E,$DA,$D6,$3E,$BF,$DF,$B9,$69,$E9,$8F,$77,$E9,$9E,$8F
	.dc.b	$D1,$C3,$F6,$AF,$F6,$4F,$E2,$2F,$DD,$71,$EB,$EE,$B5,$E2,$7F,$BE
	.dc.b	$67,$87,$ED,$59,$C3,$DB,$C7,$D2,$FE,$F3,$FD,$5B,$7F,$C8,$47,$82
	.dc.b	$7B,$DE,$B8,$3D,$77,$FE,$07,$AD,$63,$DF,$C3,$F8,$18,$F2,$9F,$1E
	.dc.b	$9E,$7C,$9F,$A4,$F8,$F0,$F3,$47,$FB,$12,$FF,$41,$F9,$AF,$F8,$5F
	.dc.b	$CC,$7E,$F9,$FC,$5F,$F7,$67,$EC,$3F,$6B,$7F,$1E,$EC,$38,$5E,$E9
	.dc.b	$FA,$7F,$E0,$A7,$A5,$7B,$CF,$F1,$03,$FE,$25,$FD,$87,$E2,$BC,$BD
	.dc.b	$F6,$5F,$C9,$EB,$97,$CF,$34,$F2,$D6,$BF,$4F,$F8,$35,$D2,$B5,$A7
	.dc.b	$8F,$F5,$7E,$38,$70,$C7,$D7,$F5,$78,$AF,$47,$E3,$17,$3F,$0E,$EF
	.dc.b	$CD,$BF,$37,$F9,$5F,$FA,$F7,$F3,$67,$EC,$2F,$D8,$BF,$87,$7D,$2F
	.dc.b	$F6,$BE,$6F,$8B,$7B,$AD,$7A,$9F,$E1,$DF,$A2,$E7,$F6,$0C,$BF,$0B
	.dc.b	$7F,$EC,$9F,$A5,$B3,$9C,$38,$F8,$3E,$EE,$75,$F5,$D7,$F8,$F7,$2E
	.dc.b	$63,$1A,$7F,$8A,$FF,$50,$EF,$A7,$A2,$FC,$FF,$AA,$FE,$DB,$A9,$ED
	.dc.b	$3F,$6A,$7E,$B6,$F2,$3F,$C2,$9F,$8C,$BE,$47,$85,$EE,$B5,$F4,$0A
	.dc.b	$7E,$27,$AC,$7F,$43,$97,$6F,$BA,$B8,$F1,$CF,$FC,$41,$F7,$5E,$DA
	.dc.b	$F1,$DF,$8A,$7F,$C4,$97,$9B,$E1,$97,$7F,$53,$B6,$9E,$2D,$1F,$41
	.dc.b	$FC,$4B,$E5,$9F,$96,$3F,$7C,$7E,$D8,$FD,$B5,$FB,$D7,$FB,$E7,$E4
	.dc.b	$F8,$5E,$EB,$53,$FE,$FD,$FC,$47,$38,$7E,$F3,$1E,$BE,$AD,$C3,$86
	.dc.b	$5D,$38,$7E,$C5,$78,$AF,$C3,$3E,$0E,$5C,$DF,$0C,$BF,$D4,$7E,$2F
	.dc.b	$EB,$B0,$E1,$A7,$8B,$C3,$7E,$B3,$E2,$8C,$7E,$DB,$73,$F5,$D7,$FB
	.dc.b	$CB,$0F,$DB,$FE,$05,$F0,$BA,$CF,$F3,$7F,$D6,$DF,$B8,$3E,$9B,$F9
	.dc.b	$CC,$EC,$D4,$3F,$65,$CF,$07,$E0,$FC,$8F,$DE,$1D,$1F,$C0,$B8,$AE
	.dc.b	$5F,$12,$C5,$FA,$0F,$E6,$E7,$30,$FC,$4B,$18,$7E,$2A,$7C,$3E,$EF
	.dc.b	$83,$5F,$4E,$3C,$F4,$7E,$A5,$FD,$43,$E9,$1F,$1A,$FD,$5D,$FA,$7B
	.dc.b	$F4,$57,$DB,$FE,$0B,$D9,$E1,$7B,$AD,$4F,$E1,$BF,$2E,$CE,$1F,$CE
	.dc.b	$B1,$EB,$F9,$6D,$C3,$86,$5D,$3E,$B3,$F6,$D5,$E2,$BF,$0C,$F8,$3F
	.dc.b	$00,$CD,$F0,$CB,$F2,$7F,$45,$DB,$4F,$0F,$7B,$F7,$EF,$F5,$1F,$07
	.dc.b	$F8,$57,$F4,$2F,$93,$F6,$CF,$D2,$1F,$8B,$7F,$3E,$F1,$9F,$AB,$70
	.dc.b	$F0,$3F,$D9,$D8,$70,$BD,$F9,$9F,$E5,$CF,$CB,$35,$4A,$EA,$FF,$AA
	.dc.b	$B3,$F0,$78,$4F,$DC,$7F,$D2,$F1,$8D,$7C,$5A,$FF,$4E,$FC,$C3,$FB
	.dc.b	$94,$E7,$4E,$FF,$13,$8C,$BC,$79,$7F,$B7,$A7,$C7,$BB,$F5,$6F,$1E
	.dc.b	$BE,$7C,$FD,$9F,$91,$63,$0E,$11,$EF,$FD,$4B,$1B,$78,$B8,$7F,$CE
	.dc.b	$FF,$03,$EC,$F8,$67,$EB,$8F,$DB,$9F,$A1,$3F,$3F,$FF,$15,$7E,$CC
	.dc.b	$FC,$BB,$85,$EE,$B5,$F6,$2A,$7C,$86,$BF,$42,$F8,$7C,$FF,$EF,$BD
	.dc.b	$3B,$E9,$88,$FE,$B1,$97,$3D,$3F,$C7,$1F,$81,$FC,$39,$CD,$79,$27
	.dc.b	$FC,$79,$F2,$19,$C3,$9E,$BE,$2D,$1F,$A0,$F9,$B0,$E1,$B7,$97,$F2
	.dc.b	$7F,$B7,$F6,$8F,$D8,$1F,$B1,$FE,$92,$1F,$AA,$FF,$36,$FD,$57,$75
	.dc.b	$AF,$33,$FC,$CB,$3F,$CE,$BD,$CC,$FE,$F1,$EB,$C3,$A6,$37,$0E,$11
	.dc.b	$D3,$C4,$F6,$3A,$C7,$9F,$8D,$8A,$DE,$4E,$EF,$AB,$F0,$1F,$BA,$E7
	.dc.b	$1F,$E7,$6A,$DC,$FF,$99,$67,$52,$F2,$7F,$CB,$1F,$A2,$3F,$03,$7E
	.dc.b	$5D,$D7,$D0,$FE,$B7,$87,$0B,$DD,$6B,$1F,$E0,$BF,$57,$FA,$29,$FF
	.dc.b	$31,$7E,$1F,$F6,$5E,$B9,$FE,$8C,$F2,$BC,$B2,$F5,$7E,$19,$F6,$E3
	.dc.b	$18,$3F,$57,$C7,$DC,$36,$F3,$61,$E5,$C7,$C1,$39,$D3,$C9,$B7,$FA
	.dc.b	$A3,$6F,$44,$E6,$FB,$F7,$70,$C7,$9E,$6D,$5F,$A5,$7E,$F3,$CD,$C7
	.dc.b	$82,$FD,$D9,$F9,$37,$F8,$77,$F0,$A7,$E4,$7B,$C7,$FC,$5D,$5A,$F8
	.dc.b	$45,$3F,$45,$D6,$3F,$9D,$CB,$C1,$F6,$77,$1E,$39,$FF,$8A,$7F,$69
	.dc.b	$FC,$F5,$78,$B3,$E9,$F9,$CE,$2F,$7A,$76,$FB,$2E,$B5,$E1,$D5,$FA
	.dc.b	$D7,$F1,$6F,$8D,$7E,$37,$FE,$17,$FA,$07,$E2,$3F,$E8,$FF,$C7,$BC
	.dc.b	$DD,$7F,$86,$BD,$FD,$47,$F8,$56,$B5,$CD,$CF,$1F,$F1,$07,$7F,$DD
	.dc.b	$3F,$C3,$2B,$E8,$3F,$75,$87,$2C,$FB,$FE,$0A,$F8,$CF,$DD,$4E,$72
	.dc.b	$FF,$50,$78,$F8,$C7,$CD,$1E,$5F,$DE,$7C,$1A,$F8,$F1,$6E,$FD,$6B
	.dc.b	$F9,$9F,$DE,$BF,$3F,$61,$FA,$4F,$DD,$1F,$AB,$7D,$EF,$C2,$B7,$5A
	.dc.b	$E3,$56,$BC,$3F,$E3,$DC,$DF,$A8,$CB,$A6,$5F,$C5,$5E,$D9,$DC,$79
	.dc.b	$61,$DB,$F1,$2F,$E2,$D5,$C3,$2E,$BA,$3E,$E3,$38,$70,$8F,$C8,$FE
	.dc.b	$6D,$E4,$CF,$D1,$1F,$94,$FB,$1C,$61,$CB,$67,$A5,$8D,$7D,$1F,$61
	.dc.b	$FD,$53,$F0,$17,$EC,$CF,$D8,$1F,$92,$BF,$37,$7E,$9D,$FD,$67,$E2
	.dc.b	$F8,$5E,$EB,$53,$FD,$77,$97,$38,$7F,$35,$C7,$AF,$ED,$77,$0E,$19
	.dc.b	$3F,$52,$FE,$27,$C1,$C3,$3E,$79,$7F,$39,$C7,$CB,$97,$93,$0F,$1F
	.dc.b	$CB,$7C,$5E,$AF,$E2,$F5,$AC,$3C,$73,$F9,$A9,$FE,$37,$EE,$BF,$2E
	.dc.b	$FF,$0C,$7D,$7B,$F6,$EC,$79,$4F,$57,$FB,$BD,$EF,$E9,$7F,$C7,$15
	.dc.b	$AF,$D3,$73,$8F,$F8,$87,$87,$F3,$97,$D7,$EA,$7F,$AA,$3C,$73,$ED
	.dc.b	$5F,$E3,$1A,$E3,$8F,$48,$FF,$1B,$C6,$27,$79,$BF,$22,$C6,$7E,$6F
	.dc.b	$C4,$5B,$7F,$A0,$F0,$FE,$77,$E5,$BE,$B9,$F8,$33,$F6,$57,$E9,$3F
	.dc.b	$E8,$DF,$BF,$FF,$10,$63,$C7,$1F,$DF,$E1,$C2,$F7,$F8,$2A,$DF,$5B
	.dc.b	$E0,$F5,$BE,$BE,$9D,$F4,$FB,$39,$7D,$AC,$FA,$4F,$8F,$F4,$1C,$CF
	.dc.b	$F5,$09,$CE,$9D,$F5,$F2,$4F,$53,$D7,$19,$D3,$FD,$A9,$19,$3C,$F8
	.dc.b	$FB,$6B,$D9,$F6,$59,$CB,$C9,$B7,$AB,$67,$E7,$99,$CF,$A4,$7F,$BC
	.dc.b	$B4,$F1,$E5,$CF,$4F,$81,$EB,$F7,$FD,$23,$FC,$BD,$ED,$3F,$2A,$7E
	.dc.b	$AA,$EE,$FF,$5D,$FC,$16,$FB,$7F,$8C,$56,$A7,$FD,$7F,$F8,$96,$70
	.dc.b	$F8,$7C,$BB,$FD,$9A,$EF,$F4,$F4,$EF,$8F,$1F,$80,$F9,$1D,$3F,$A6
	.dc.b	$9C,$ED,$E1,$FC,$17,$39,$7F,$7C,$9C,$EB,$FE,$CE,$F2,$7C,$19,$78
	.dc.b	$FD,$0F,$F3,$BE,$9F,$81,$FD,$35,$C0,$FC,$AD,$F0,$0F,$A8,$FF,$31
	.dc.b	$78,$4F,$4B,$E9,$61,$C2,$F7,$5A,$D3,$38,$FE,$5B,$7F,$C0,$E7,$DE
	.dc.b	$3B,$46,$38,$BF,$80,$E3,$CF,$4F,$07,$D9,$FE,$46,$B9,$E7,$DA,$3F
	.dc.b	$07,$8F,$38,$C6,$AF,$DC,$77,$87,$28,$FC,$0D,$7E,$7F,$F7,$17,$E2
	.dc.b	$2F,$DE,$5F,$A2,$BF,$64,$F7,$7E,$28,$FB,$1F,$C8,$61,$C2,$F7,$5A
	.dc.b	$FC,$1B,$38,$F3,$E0,$F6,$1F,$E2,$EB,$EF,$FB,$C7,$DE,$E3,$FE,$3F
	.dc.b	$1C,$F4,$F0,$73,$B5,$FB,$78,$2B,$59,$FD,$1F,$D5,$73,$18,$BE,$1F
	.dc.b	$31,$E2,$DB,$CB,$1C,$FF,$05,$C6,$7D,$B3,$7E,$F9,$8C,$FD,$FE,$C7
	.dc.b	$EB,$7F,$2C,$FF,$0C,$FE,$C7,$FE,$F9,$E6,$78,$4F,$DC,$5F,$98,$EF
	.dc.b	$E7,$D7,$BA,$D6,$19,$8F,$81,$C9,$FE,$50,$8F,$F1,$2F,$ED,$58,$84
	.dc.b	$78,$3F,$13,$67,$E0,$FC,$0F,$E1,$CE,$63,$C7,$DB,$8C,$FC,$37,$C3
	.dc.b	$3F,$37,$D2,$F8,$63,$1B,$3F,$00,$FE,$15,$F6,$2F,$A5,$FE,$51,$F8
	.dc.b	$A7,$F6,$2F,$DE,$9F,$88,$FF,$97,$F5,$FE,$4D,$7B,$F6,$D7,$BA,$7D
	.dc.b	$47,$4C,$7E,$37,$1F,$E0,$8F,$F7,$3B,$FC,$D5,$C6,$1C,$EF,$C1,$E9
	.dc.b	$7E,$02,$31,$F2,$2D,$3F,$C4,$D7,$D7,$3F,$0C,$E7,$37,$E4,$B9,$CB
	.dc.b	$C9,$F4,$DF,$C9,$FE,$C7,$FD,$ED,$F9,$7F,$FB,$D3,$F6,$CF,$81,$EC
	.dc.b	$FF,$8B,$78,$BF,$A1,$56,$BF,$11,$56,$BD,$CE,$5E,$F3,$1F,$B1,$7F
	.dc.b	$F0,$07,$F2,$1B,$39,$C3,$8F,$2B,$F6,$4F,$CD,$DF,$5C,$BE,$87,$C2
	.dc.b	$E3,$0E,$11,$E4,$FC,$F3,$BE,$7E,$7C,$7F,$97,$FE,$4D,$8C,$FA,$EE
	.dc.b	$FD,$D7,$18,$FB,$3E,$1B,$FA,$37,$E7,$6F,$87,$FF,$1D,$FE,$08,$E3
	.dc.b	$7D,$3B,$F3,$97,$E5,$7C,$2F,$75,$A9,$7E,$07,$9F,$A9,$FC,$C6,$5D
	.dc.b	$7E,$F6,$E1,$C3,$2F,$F4,$A7,$83,$BD,$3C,$11,$FE,$89,$8C,$67,$D7
	.dc.b	$C9,$B3,$E3,$F1,$FF,$BE,$F8,$DF,$E2,$4F,$A9,$FF,$5D,$79,$8D,$5F
	.dc.b	$38,$FB,$47,$E9,$AF,$D8,$78,$ED,$FE,$BE,$C7,$8E,$1C,$2F,$7C,$5A
	.dc.b	$AF,$F4,$DE,$DF,$E7,$E7,$F9,$BE,$9E,$18,$FF,$4D,$C6,$0E,$8F,$FA
	.dc.b	$7A,$FA,$7E,$1D,$FD,$F6,$B8,$6B,$E7,$D7,$CD,$F1,$F9,$C7,$8C,$E7
	.dc.b	$3F,$4F,$E2,$1E,$3D,$7C,$F8,$7B,$77,$78,$DE,$7F,$51,$F9,$F3,$C0
	.dc.b	$7D,$37,$F8,$5B,$F4,$0F,$EE,$DF,$86,$EE,$B5,$39,$F9,$6F,$F7,$EF
	.dc.b	$9A,$E3,$FF,$28,$E5,$D3,$E6,$7F,$B9,$C6,$30,$ED,$C7,$F0,$4F,$E2
	.dc.b	$AF,$4F,$A2,$69,$F7,$71,$E9,$5F,$EE,$EC,$7C,$FA,$F9,$AF,$7F,$41
	.dc.b	$F0,$EB,$9C,$E7,$0F,$F0,$77,$6F,$A6,$31,$AB,$F6,$BC,$6F,$EB,$FB
	.dc.b	$BB,$9F,$8A,$BF,$D5,$9A,$F8,$B3,$7E,$DB,$FE,$BC,$F3,$1F,$5A,$FD
	.dc.b	$85,$FA,$D3,$EB,$7F,$C9,$B8,$F1,$C3,$85,$EE,$B5,$FA,$B7,$E2,$67
	.dc.b	$FD,$8F,$EA,$7C,$8C,$C7,$83,$EA,$1A,$66,$30,$F8,$B4,$F0,$FE,$7F
	.dc.b	$F0,$F0,$EF,$1E,$4F,$BF,$C6,$1C,$B4,$F1,$6D,$FE,$5F,$DB,$CD,$39
	.dc.b	$BF,$5C,$BF,$40,$FC,$F7,$EB,$3F,$E0,$0F,$B5,$FF,$5C,$FE,$71,$FE
	.dc.b	$3A,$FD,$3B,$F8,$E6,$5F,$1E,$B5,$F7,$8A,$D7,$87,$CB,$E8,$3F,$D4
	.dc.b	$9C,$7F,$83,$BF,$F0,$5F,$C8,$7B,$5E,$37,$FE,$46,$FC,$2B,$EF,$46
	.dc.b	$3D,$EF,$F0,$E6,$7D,$F1,$FF,$57,$67,$DE,$5A,$3F,$38,$F0,$6D,$E8
	.dc.b	$9C,$FE,$11,$FB,$9F,$9A,$7E,$AF,$FE,$E7,$FD,$69,$F9,$F7,$F8,$EF
	.dc.b	$E4,$FF,$DB,$BB,$7E,$0E,$F7,$E6,$AF,$7F,$57,$FD,$6B,$F0,$DE,$99
	.dc.b	$7F,$27,$E6,$FD,$06,$1C,$1B,$1D,$F3,$ED,$1C,$31,$E9,$1F,$E4,$BF
	.dc.b	$DA,$3F,$3D,$7E,$0C,$BB,$E8,$FD,$27,$84,$63,$EB,$EE,$1E,$9F,$1A
	.dc.b	$ED,$FD,$BB,$F7,$BF,$E3,$1F,$AA,$FF,$B9,$4F,$F1,$C7,$C5,$3F,$52
	.dc.b	$78,$6F,$C9,$37,$F8,$72,$F7,$5A,$F8,$0C,$FE,$57,$F9,$59,$E9,$CB
	.dc.b	$FB,$F7,$B7,$D3,$B3,$ED,$8F,$EE,$67,$31,$E1,$F0,$B1,$3D,$EB,$86
	.dc.b	$5E,$4F,$D6,$7C,$78,$78,$B5,$7E,$65,$FC,$CF,$ED,$7F,$8D,$FE,$C3
	.dc.b	$FE,$0C,$FC,$E5,$F9,$0B,$C6,$FE,$B5,$E0,$FE,$F6,$F7,$D8,$BD,$D3
	.dc.b	$F5,$7E,$9E,$D3,$F4,$DF,$D9,$61,$DB,$97,$F8,$E3,$CB,$0E,$FF,$01
	.dc.b	$FC,$56,$31,$A7,$F1,$DF,$83,$FF,$7B,$C6,$33,$7E,$1D,$DB,$7F,$45
	.dc.b	$F9,$FD,$AF,$E9,$DF,$67,$FD,$63,$F0,$87,$EC,$CF,$DA,$1F,$62,$BF
	.dc.b	$ED,$FC,$BF,$C6,$70,$E1,$7B,$AD,$7E,$0F,$D3,$E3,$7F,$DB,$BC,$7F
	.dc.b	$CD,$6F,$FB,$2F,$EC,$FF,$C3,$DE,$39,$F6,$F3,$BF,$98,$46,$31,$FB
	.dc.b	$97,$CF,$5F,$2D,$F9,$35,$7E,$F7,$9D,$7C,$51,$FE,$4B,$DB,$CD,$AF
	.dc.b	$FB,$BA,$73,$FA,$DE,$F0,$E5,$F9,$4F,$FB,$E3,$F2,$DF,$E4,$BA,$3F
	.dc.b	$6C,$7E,$D6,$F1,$9F,$97,$3F,$50,$7E,$9B,$FC,$AB,$F4,$3E,$58,$FF
	.dc.b	$2E,$C7,$8E,$1C,$2F,$7F,$18,$FC,$67,$77,$DF,$37,$EE,$7F,$ED,$0D
	.dc.b	$7C,$79,$72,$AF,$F5,$BC,$62,$BC,$9A,$F8,$F5,$F8,$FC,$6A,$D6,$BE
	.dc.b	$19,$74,$FF,$BF,$E7,$37,$C7,$57,$E4,$DE,$4D,$3D,$33,$9F,$D4,$FC
	.dc.b	$FE,$83,$F7,$67,$ED,$53,$F7,$D7,$E4,$CF,$C8,$3F,$D7,$1C,$3E,$3E
	.dc.b	$3F,$EA,$38,$F1,$C3,$85,$EF,$F2,$6E,$B9,$7D,$1B,$F7,$9C,$1E,$EE
	.dc.b	$05,$F3,$B6,$5E,$2F,$B9,$F7,$57,$19,$EF,$58,$8E,$37,$EA,$9E,$38
	.dc.b	$FA,$7C,$EF,$1C,$79,$B4,$79,$1F,$63,$F4,$AF,$F5,$B9,$8F,$D0,$3F
	.dc.b	$81,$FE,$25,$FE,$B3,$E0,$F8,$1A,$2B,$76,$BD,$FE,$09,$CF,$F0,$E7
	.dc.b	$E7,$1F,$81,$C3,$FB,$BE,$1D,$9F,$80,$F2,$BF,$E2,$3F,$8A,$7A,$E3
	.dc.b	$1E,$7B,$EB,$F1,$8F,$87,$3F,$04,$FF,$B6,$7E,$CB,$C1,$7E,$98,$FF
	.dc.b	$6E,$7D,$9E,$2B,$7B,$BF,$3E,$C6,$9E,$EF,$D6,$7F,$33,$FA,$2B,$F3
	.dc.b	$97,$EF,$8D,$7F,$80,$FC,$17,$F9,$8F,$97,$E1,$DE,$EB,$5E,$A7,$F8
	.dc.b	$52,$9E,$C7,$4C,$79,$F2,$FD,$FC,$3B,$7E,$93,$F1,$61,$1C,$F3,$FE
	.dc.b	$1F,$E5,$7D,$D8,$C6,$5F,$90,$AE,$99,$F8,$1F,$EA,$7E,$86,$30,$F3
	.dc.b	$FA,$5F,$B4,$FD,$17,$FC,$65,$FA,$0B,$F5,$5F,$E0,$5F,$90,$BF,$1D
	.dc.b	$7C,$8E,$FE,$A1,$7B,$E0,$5E,$FE,$A3,$5F,$09,$E9,$CF,$AE,$FF,$9B
	.dc.b	$7B,$E9,$DB,$E4,$3F,$2D,$18,$C7,$9F,$EA,$B8,$D3,$BD,$F9,$36,$7E
	.dc.b	$69,$E3,$DB,$C5,$B7,$9F,$E0,$3F,$87,$FD,$EB,$F1,$E7,$A8,$FA,$2F
	.dc.b	$CA,$7E,$C2,$FD,$49,$F9,$EF,$0B,$DD,$6A,$5F,$98,$67,$0F,$A9,$8F
	.dc.b	$5F,$C4,$6E,$1C,$3C,$66,$7F,$E2,$EF,$7F,$F9,$7D,$F8,$F0,$F1,$56
	.dc.b	$BF,$14,$FD,$B4,$E7,$4F,$07,$C8,$77,$7C,$5F,$9E,$5B,$EB,$B3,$F1
	.dc.b	$EC,$69,$EB,$FC,$A7,$F9,$7F,$D4,$FF,$A5,$3F,$26,$7E,$D7,$F9,$D9
	.dc.b	$7D,$5F,$F3,$5F,$3F,$F7,$37,$BA,$D4,$BE,$3F,$9F,$CD,$BE,$5B,$F2
	.dc.b	$97,$D7,$C5,$FE,$2C,$F1,$F5,$19,$F6,$9E,$3A,$78,$A7,$3E,$E7,$F1
	.dc.b	$5B,$E1,$7E,$4F,$C2,$77,$AF,$8A,$FF,$A4,$FE,$38,$FA,$2F,$E6,$3F
	.dc.b	$96,$7F,$9F,$BE,$19,$F9,$87,$F4,$0E,$3F,$E2,$77,$B9,$56,$EB,$FC
	.dc.b	$33,$DD,$FD,$8A,$BA,$7E,$5F,$FA,$F7,$87,$0F,$F3,$17,$95,$F1,$8F
	.dc.b	$1E,$5E,$27,$FA,$1F,$E0,$1F,$CB,$EB,$71,$FD,$1B,$1F,$F0,$66,$7E
	.dc.b	$4B,$E1,$9F,$A3,$E7,$FF,$B2,$FE,$81,$E6,$F9,$87,$DD,$3F,$43,$7E
	.dc.b	$90,$FD,$77,$FD,$09,$9F,$F3,$0C,$38,$5E,$EB,$5E,$B7,$4E,$1F,$E7
	.dc.b	$7B,$FF,$4B,$BE,$CD,$8F,$EE,$77,$C7,$AB,$FA,$45,$74,$CF,$C3,$58
	.dc.b	$36,$F9,$33,$FE,$0D,$9F,$8E,$73,$9F,$9F,$F6,$BF,$16,$BE,$CC,$38
	.dc.b	$7E,$C7,$8D,$3D,$19,$3D,$66,$6F,$C6,$7F,$C2,$FE,$1B,$F3,$A7,$E7
	.dc.b	$3F,$EB,$CF,$BE,$7E,$9B,$FE,$D5,$FD,$77,$8E,$1C,$2F,$7F,$42,$BD
	.dc.b	$FE,$20,$DF,$7C,$FF,$EC,$8D,$3B,$69,$FE,$26,$8F,$F1,$3E,$BE,$37
	.dc.b	$59,$3E,$BF,$51,$FE,$D0,$9D,$3F,$C0,$A7,$C3,$97,$9B,$BF,$5A,$F9
	.dc.b	$71,$F5,$6F,$F9,$AC,$3C,$67,$18,$CE,$AF,$AE,$CE,$BE,$6D,$FD,$98
	.dc.b	$7F,$BD,$7E,$2B,$8E,$8F,$DC,$5F,$BA,$3F,$72,$FC,$64,$FF,$8D,$7F
	.dc.b	$13,$DD,$6A,$73,$F5,$9F,$EF,$AE,$0F,$F2,$3D,$F3,$C3,$DF,$C1,$C3
	.dc.b	$F1,$0B,$C9,$F0,$2B,$58,$F6,$F4,$BD,$F7,$FE,$F4,$9C,$F6,$7D,$3F
	.dc.b	$C3,$33,$3F,$57,$5F,$14,$FD,$7D,$5F,$6B,$E6,$D7,$D5,$39,$FC,$E3
	.dc.b	$E3,$FE,$C8,$FD,$C9,$F3,$DF,$E5,$8F,$DE,$DF,$81,$29,$D7,$C2,$F7
	.dc.b	$F6,$6B,$DD,$3F,$19,$D7,$CC,$7B,$BA,$3F,$CD,$33,$DF,$3C,$71,$0E
	.dc.b	$D9,$74,$C7,$C3,$A7,$BA,$F8,$67,$DE,$1F,$B1,$66,$F8,$47,$9F,$EB
	.dc.b	$7C,$D9,$7A,$E3,$D5,$EA,$63,$0E,$58,$BF,$56,$C6,$7E,$AF,$CC,$FF
	.dc.b	$17,$FB,$D3,$F4,$C7,$E4,$BF,$E2,$9D,$BF,$53,$FB,$1B,$F6,$5D,$D6
	.dc.b	$BE,$BF,$E6,$78,$CA,$7E,$7D,$AF,$8D,$F8,$3C,$CD,$3B,$69,$86,$E3
	.dc.b	$CB,$D2,$FB,$F1,$8D,$7F,$73,$ED,$E3,$5F,$04,$79,$3F,$38,$F1,$E3
	.dc.b	$E7,$DB,$C5,$C3,$FA,$6F,$CF,$FF,$93,$FA,$EF,$DF,$1F,$A1,$7F,$94
	.dc.b	$FF,$2D,$FF,$33,$F9,$BF,$B4,$BD,$D6,$BE,$D1,$4F,$E0,$0E,$98,$FF
	.dc.b	$32,$E4,$FD,$EE,$3C,$63,$FC,$A1,$18,$F0,$BF,$D5,$79,$74,$CF,$FC
	.dc.b	$C5,$F0,$5F,$9C,$8C,$67,$FD,$4A,$FA,$E1,$E7,$BF,$37,$EF,$D8,$D7
	.dc.b	$C9,$AB,$F3,$6F,$E0,$7F,$A0,$3D,$B7,$AA,$FD,$41,$F9,$B3,$F3,$EF
	.dc.b	$F0,$F7,$D3,$F0,$BD,$D6,$A7,$FC,$45,$F4,$B9,$F3,$7E,$CE,$5D,$72
	.dc.b	$B8,$70,$CB,$A7,$E1,$1F,$B6,$AD,$65,$CF,$F2,$AC,$5E,$F3,$FF,$17
	.dc.b	$79,$9E,$1C,$BC,$D9,$79,$7E,$7D,$EF,$67,$2F,$24,$3C,$BC,$56,$FC
	.dc.b	$CF,$E4,$7F,$8A,$7F,$84,$7E,$21,$F9,$1F,$F0,$4F,$DB,$5F,$BD,$CF
	.dc.b	$0B,$F9,$8D,$EE,$B5,$2F,$DB,$B9,$FC,$FB,$1F,$CC,$DF,$F8,$77,$D2
	.dc.b	$B3,$9C,$38,$F8,$3C,$2F,$C3,$EF,$B4,$FE,$B1,$39,$AE,$FF,$B3,$DA
	.dc.b	$D6,$8F,$D8,$3F,$9E,$FE,$F6,$FC,$7F,$FC,$8F,$ED,$FE,$D3,$AD,$E8
	.dc.b	$3F,$5D,$DD,$6B,$D1,$FC,$14,$BF,$2C,$CF,$E0,$9F,$B2,$CB,$A7,$C7
	.dc.b	$DC,$38,$65,$FC,$3B,$87,$9C,$FF,$24,$CB,$F1,$9A,$7F,$32,$C3,$8B
	.dc.b	$E3,$CD,$F8,$87,$08,$C7,$E1,$BF,$DD,$BF,$53,$FE,$41,$93,$F5,$D7
	.dc.b	$EC,$0F,$C2,$9F,$8E,$3F,$6B,$7D,$33,$E0,$1F,$A0,$F3,$CB,$F9,$56
	.dc.b	$5C,$B1,$E3,$87,$0B,$7E,$41,$BB,$F0,$67,$DD,$FD,$BE,$9E,$4C,$7C
	.dc.b	$7F,$BA,$FF,$56,$8C,$57,$97,$5F,$26,$BE,$3C,$3A,$61,$E1,$9C,$FE
	.dc.b	$3B,$FD,$F7,$0E,$3A,$F9,$30,$F5,$7E,$D9,$C3,$2E,$9B,$79,$FD,$2C
	.dc.b	$5E,$F7,$7E,$11,$8D,$3C,$30,$FD,$53,$D9,$C6,$FC,$C5,$F9,$43,$F4
	.dc.b	$57,$D0,$BF,$72,$FE,$41,$AF,$0B,$DD,$6A,$7F,$92,$70,$67,$0F,$85
	.dc.b	$CB,$FD,$5F,$2E,$DF,$89,$E2,$11,$E1,$9F,$06,$7D,$FF,$07,$F8,$55
	.dc.b	$C7,$3E,$F3,$E3,$FA,$9C,$CF,$29,$CE,$7F,$CD,$7E,$3F,$C5,$AF,$97
	.dc.b	$E1,$3E,$07,$A8,$E9,$FC,$5F,$B7,$F0,$4F,$EF,$1F,$A9,$BF,$12,$FF
	.dc.b	$0E,$7C,$1F,$E0,$67,$5F,$30,$AD,$7D,$1F,$FB,$1B,$1E,$9B,$FF,$75
	.dc.b	$BF,$99,$F0,$BE,$F7,$B4,$7E,$20,$D7,$07,$F2,$EC,$7A,$67,$FE,$54
	.dc.b	$9F,$04,$E7,$2F,$27,$D2,$7C,$0C,$3C,$71,$E2,$FC,$AB,$71,$EA,$C3
	.dc.b	$FC,$69,$F6,$FF,$99,$FC,$39,$FB,$53,$F5,$17,$EE,$FF,$77,$F7,$E7
	.dc.b	$EA,$1F,$DE,$78,$BD,$7C,$1F,$EF,$7C,$38,$5E,$E9,$F9,$97,$59,$7E
	.dc.b	$7B,$F7,$73,$EF,$9D,$9E,$3D,$6D,$3F,$45,$A7,$6B,$DE,$7E,$3F,$A0
	.dc.b	$FB,$B1,$8B,$E1,$A3,$E1,$BC,$19,$FA,$A7,$3F,$1F,$8C,$7C,$DF,$BE
	.dc.b	$D9,$E3,$FA,$A7,$EE,$FF,$30,$7E,$30,$FD,$E5,$FA,$47,$F9,$6F,$F7
	.dc.b	$CF,$F6,$DF,$E9,$7B,$AD,$4E,$7F,$0A,$7E,$A9,$9C,$3F,$39,$8F,$4F
	.dc.b	$80,$B8,$70,$F1,$19,$FF,$1E,$EB,$E1,$9F,$7F,$C5,$FF,$37,$8F,$93
	.dc.b	$2F,$1E,$1C,$34,$FF,$76,$F4,$7D,$4A,$D6,$1E,$3F,$D6,$7F,$8B,$7E
	.dc.b	$77,$FE,$8F,$FD,$1D,$F8,$73,$F7,$CF,$F3,$E7,$BB,$FC,$EF,$2F,$BB
	.dc.b	$7B,$AD,$7E,$07,$FC,$A7,$E3,$F9,$FE,$0C,$CB,$F7,$17,$FE,$9B,$FC
	.dc.b	$46,$CE,$70,$E3,$F8,$B7,$ED,$2B,$A6,$9F,$E8,$BF,$C5,$71,$7B,$8F
	.dc.b	$1F,$C6,$76,$CB,$CF,$1E,$6F,$C6,$3F,$98,$7C,$23,$F1,$A7,$E5,$3F
	.dc.b	$5B,$E8,$9F,$09,$F9,$BF,$9C,$DD,$6B,$9F,$FC,$2F,$3F,$9F,$F5,$B3
	.dc.b	$87,$F1,$1C,$7A,$63,$70,$E0,$F4,$F3,$7F,$86,$CF,$86,$BC,$18,$7F
	.dc.b	$80,$BB,$F9,$E5,$DB,$CA,$FF,$41,$F4,$FE,$A3,$0E,$13,$9F,$DA,$77
	.dc.b	$9F,$A2,$FF,$AC,$5E,$F4,$F0,$E5,$FB,$5F,$C5,$FF,$A8,$7D,$A3,$F4
	.dc.b	$B7,$E7,$CF,$C3,$9F,$B1,$BF,$50,$7E,$BD,$FC,$DF,$4F,$DC,$E3,$C7
	.dc.b	$0E,$17,$BA,$7E,$3F,$DB,$2E,$BC,$7F,$C5,$46,$1E,$1F,$BA,$FE,$70
	.dc.b	$F3,$D3,$C3,$A7,$EE,$BE,$1B,$FB,$AC,$E7,$3F,$27,$3F,$F3,$68,$C5
	.dc.b	$79,$BF,$3C,$F0,$6D,$E4,$8F,$4F,$D1,$7F,$BD,$3F,$62,$FE,$FF,$F9
	.dc.b	$D3,$F7,$77,$EF,$AF,$D3,$DF,$8A,$FF,$94,$FD,$9F,$E3,$9C,$EC,$78
	.dc.b	$E1,$C2,$F7,$EF,$2F,$F8,$BD,$EF,$97,$FD,$59,$9F,$7C,$FF,$D9,$F9
	.dc.b	$72,$C6,$D6,$B2,$E9,$9F,$F9,$8F,$C7,$FF,$7A,$5F,$83,$0F,$34,$FF
	.dc.b	$99,$FF,$3F,$CE,$1C,$67,$3A,$F8,$F1,$F8,$13,$9C,$BC,$FA,$FF,$92
	.dc.b	$F5,$F3,$4B,$F6,$D7,$D1,$FF,$7B,$EC,$7E,$D3,$FC,$97,$F4,$27,$9F
	.dc.b	$D3,$0E,$17,$BA,$D7,$E7,$19,$C7,$9F,$CF,$FB,$AF,$7F,$31,$67,$39
	.dc.b	$F6,$F8,$77,$E3,$3E,$CB,$EB,$87,$E4,$B2,$FE,$6D,$F9,$26,$71,$E3
	.dc.b	$1E,$5F,$DD,$FC,$1A,$F8,$F2,$6E,$FD,$AF,$39,$73,$9C,$FE,$6B,$8D
	.dc.b	$BC,$57,$FD,$73,$8B,$E2,$F7,$BF,$38,$7D,$CB,$F7,$F7,$C6,$3F,$4D
	.dc.b	$FF,$3B,$7E,$C9,$F1,$70,$E1,$7B,$AD,$7E,$02,$FD,$A7,$59,$7F,$9B
	.dc.b	$79,$FF,$D6,$FE,$1F,$F9,$53,$1E,$31,$83,$FD,$EC,$F4,$D3,$C3,$A7
	.dc.b	$F0,$9D,$7C,$F8,$79,$A7,$31,$F2,$EB,$86,$7E,$2F,$84,$F0,$EB,$E5
	.dc.b	$C7,$FD,$1D,$F7,$7F,$77,$EA,$7F,$CC,$1E,$BB,$F5,$17,$E6,$EF,$D9
	.dc.b	$BF,$DD,$9F,$15,$BF,$8B,$FF,$01,$56,$A7,$FD,$E1,$F5,$79,$C3,$F0
	.dc.b	$58,$FF,$9A,$FE,$7A,$E1,$C3,$2E,$9E,$AF,$E0,$6B,$C3,$97,$3F,$B0
	.dc.b	$C5,$EF,$3F,$F7,$E7,$E5,$3D,$6B,$C1,$F8,$37,$FA,$5F,$97,$F3,$BC
	.dc.b	$EF,$A6,$FF,$6A,$7E,$8E,$FC,$39,$E3,$BF,$34,$FF,$05,$61,$C2,$F7
	.dc.b	$5A,$97,$ED,$DF,$A1,$C7,$A7,$EB,$BF,$9D,$7B,$7E,$29,$70,$E3,$97
	.dc.b	$FB,$C3,$2E,$92,$CF,$C5,$2F,$C0,$7F,$41,$87,$29,$CE,$7F,$EE,$BC
	.dc.b	$38,$C6,$32,$F1,$65,$DA,$9F,$6D,$8C,$FB,$6B,$E9,$FC,$FB,$F7,$1F
	.dc.b	$B1,$3F,$65,$7E,$9E,$FB,$0F,$D3,$FC,$5F,$F9,$4F,$E6,$B7,$5A,$9C
	.dc.b	$FD,$9F,$FB,$47,$87,$F8,$4E,$3C,$F1,$FD,$F3,$D7,$D7,$B8,$71,$F4
	.dc.b	$3F,$A8,$7E,$CB,$FB,$DA,$E1,$97,$83,$E7,$B1,$7B,$D7,$FC,$C9,$F9
	.dc.b	$A7,$8B,$3F,$36,$5F,$D6,$75,$FF,$5A,$EB,$FC,$6B,$F3,$D7,$DD,$FF
	.dc.b	$9C,$FF,$00,$FE,$37,$F6,$3F,$E2,$6F,$BE,$FE,$31,$5A,$F9,$35,$6B
	.dc.b	$EB,$5F,$85,$6B,$17,$E4,$3F,$C6,$70,$E1,$EC,$7E,$CC,$3C,$F0,$EF
	.dc.b	$A7,$F1,$98,$C7,$C1,$7F,$99,$3E,$8B,$18,$F9,$63,$FC,$CF,$9B,$F7
	.dc.b	$BF,$06,$BE,$49,$CF,$EE,$9F,$CB,$FF,$08,$FF,$1F,$FC,$7B,$F5,$97
	.dc.b	$D7,$BF,$20,$7E,$A1,$F9,$BE,$0F,$E7,$B8,$70,$BD,$D6,$BF,$65,$E9
	.dc.b	$B7,$F6,$3E,$3F,$65,$FF,$93,$3E,$1B,$F9,$C3,$C7,$3E,$DE,$47,$F3
	.dc.b	$58,$C6,$2E,$8C,$63,$FD,$B7,$3E,$FA,$BE,$F7,$BE,$FE,$9C,$FD,$12
	.dc.b	$FE,$00,$CD,$F0,$FC,$B2,$D6,$BF,$51,$FE,$A1,$FA,$0F,$F9,$4B,$F5
	.dc.b	$67,$97,$C7,$F5,$94,$4E,$6F,$EB,$B5,$BA,$D4,$FF,$85,$A7,$3F,$6F
	.dc.b	$C3,$9B,$EE,$5F,$F9,$7B,$EA,$BF,$AC,$3C,$76,$7E,$88,$AF,$EA,$D3
	.dc.b	$9C,$DF,$B7,$E3,$1F,$1E,$9E,$08,$F4,$C6,$31,$57,$A3,$EA,$1A,$3F
	.dc.b	$60,$ED,$9B,$F3,$8F,$DE,$FA,$8F,$CE,$1F,$92,$BF,$1B,$7E,$FC,$FB
	.dc.b	$2E,$79,$72,$C7,$8E,$1C,$2F,$7F,$2D,$AC,$FD,$95,$FC,$2B,$C2,$F0
	.dc.b	$C7,$8B,$5F,$23,$A8,$6B,$A8,$F8,$93,$A7,$D1,$3A,$C7,$C1,$F8,$2E
	.dc.b	$B3,$F2,$4E,$76,$F5,$7E,$27,$9C,$3B,$65,$EF,$9F,$76,$EF,$D0,$3D
	.dc.b	$15,$DA,$B5,$BF,$C2,$FC,$D7,$3F,$B8,$BF,$76,$7F,$00,$7F,$A6,$7F
	.dc.b	$48,$7E,$45,$F6,$FD,$3D,$D6,$BF,$06,$56,$B4,$7E,$0B,$AF,$9F,$F8
	.dc.b	$11,$C3,$2B,$CF,$F0,$74,$F0,$47,$7D,$3E,$57,$2F,$DC,$9C,$E7,$DA
	.dc.b	$5F,$95,$7C,$1C,$55,$AD,$3C,$5F,$6F,$DF,$0E,$14,$FD,$1B,$15,$E7
	.dc.b	$FC,$02,$E5,$F9,$7F,$C8,$BE,$3F,$E3,$AF,$DC,$FF,$C3,$5F,$A4,$BF
	.dc.b	$4D,$7E,$00,$FD,$C3,$FC,$89,$C9,$F6,$F0,$E1,$F5,$AC,$38,$5E,$FF
	.dc.b	$00,$EB,$93,$F2,$7F,$87,$CD,$F9,$AC,$78,$F1,$7E,$1F,$3E,$97,$E2
	.dc.b	$D7,$C3,$F9,$67,$C1,$BE,$FA,$76,$F9,$6C,$5E,$F5,$F3,$7D,$7F,$83
	.dc.b	$7F,$26,$5E,$AE,$0F,$F6,$A6,$DF,$AA,$FC,$49,$FA,$FF,$F2,$8F,$DA
	.dc.b	$1F,$9C,$7F,$BA,$BF,$3F,$FF,$32,$5E,$EB,$53,$9F,$80,$FC,$7E,$1C
	.dc.b	$BD,$DF,$6F,$2F,$F0,$38,$EB,$F3,$18,$84,$6F,$3E,$FF,$06,$F9,$BF
	.dc.b	$BD,$39,$CB,$FD,$F7,$86,$33,$F0,$DB,$5F,$F0,$E6,$BE,$5D,$19,$F4
	.dc.b	$9F,$F2,$AF,$47,$EC,$57,$6B,$E5,$1F,$BE,$BF,$34,$FB,$5F,$9A,$3F
	.dc.b	$29,$FD,$5E,$1C,$2F,$75,$AF,$47,$38,$BF,$42,$F7,$5E,$F9,$7E,$E6
	.dc.b	$FF,$C6,$23,$96,$9D,$FB,$BF,$CA,$D3,$9C,$3C,$93,$9D,$BF,$41,$7C
	.dc.b	$33,$F2,$FE,$45,$E0,$D7,$C7,$93,$77,$A3,$9C,$BB,$43,$C1,$C6,$3E
	.dc.b	$EF,$DC,$FF,$83,$FC,$C3,$F0,$87,$EE,$BE,$BF,$C8,$FF,$98,$7B,$AF
	.dc.b	$2B,$75,$AE,$65,$6B,$F1,$5C,$FE,$37,$F7,$F2,$E9,$95,$C3,$85,$F4
	.dc.b	$E8,$E7,$3D,$FA,$1F,$89,$FE,$FF,$0E,$39,$F6,$96,$BF,$E3,$6E,$8F
	.dc.b	$E5,$95,$AC,$BC,$FF,$80,$7F,$22,$FC,$21,$FA,$03,$F7,$C7,$DD,$FF
	.dc.b	$93,$7F,$16,$7E,$57,$8C,$FF,$9B,$61,$C2,$F7,$F6,$EB,$DF,$AB,$D7
	.dc.b	$2F,$F1,$3F,$1F,$DE,$3F,$B1,$61,$C5,$C0,$7A,$67,$E1,$FC,$47,$F8
	.dc.b	$E4,$63,$1E,$DF,$88,$E3,$0F,$45,$F9,$F5,$FF,$57,$6B,$E6,$9C,$DF
	.dc.b	$96,$5E,$5F,$F4,$2F,$CD,$DF,$88,$3F,$3E,$FE,$0B,$F0,$8F,$CF,$DF
	.dc.b	$9D,$EE,$B5,$EB,$3D,$9F,$F5,$ED,$3F,$3B,$D6,$3F,$BB,$CB,$AF,$ED
	.dc.b	$97,$1E,$38,$7F,$B1,$3B,$3F,$9C,$57,$8A,$FC,$3D,$F8,$BD,$EB,$E3
	.dc.b	$F9,$3E,$BA,$F8,$63,$CD,$FA,$BE,$33,$F1,$FE,$29,$67,$D5,$F8,$A7
	.dc.b	$F8,$1B,$E0,$7F,$C0,$BF,$0A,$FD,$C1,$FA,$0B,$F4,$87,$EE,$4F,$D5
	.dc.b	$BF,$9C,$7D,$D7,$1F,$0F,$FC,$35,$8F,$1C,$38,$5E,$FD,$77,$E5,$BB
	.dc.b	$BF,$EC,$F9,$F6,$7F,$9D,$69,$E2,$87,$E8,$BF,$D8,$A3,$18,$F4,$D1
	.dc.b	$F9,$3F,$F8,$1A,$73,$B7,$9A,$5E,$2E,$72,$F2,$E5,$EC,$AD,$69,$FD
	.dc.b	$93,$C4,$FE,$B5,$5A,$CF,$B4,$BF,$55,$F4,$75,$7F,$89,$BF,$4F,$7E
	.dc.b	$FA,$7F,$83,$7E,$96,$F0,$77,$F4,$19,$AD,$7B,$8A,$FC,$66,$BE,$0D
	.dc.b	$7F,$73,$6F,$A4,$F8,$11,$FE,$45,$CF,$C0,$FC,$63,$13,$D7,$3F,$E2
	.dc.b	$DE,$77,$BE,$79,$C7,$98,$FE,$35,$F4,$79,$BF,$56,$9E,$4C,$DF,$B7
	.dc.b	$7A,$2F,$86,$BE,$7F,$CA,$FE,$C7,$C4,$3F,$6D,$FF,$4D,$FB,$AF,$DC
	.dc.b	$DF,$A6,$FF,$A6,$BC,$4D,$D6,$A7,$3F,$82,$7F,$BA,$FB,$BF,$CA,$78
	.dc.b	$F3,$FA,$FF,$7F,$1F,$F8,$CF,$0F,$F5,$3B,$EF,$31,$F8,$2F,$F2,$FC
	.dc.b	$79,$C7,$68,$C6,$6F,$CA,$BE,$16,$3E,$79,$CE,$7F,$E6,$4E,$9F,$A1
	.dc.b	$39,$DB,$D5,$87,$FA,$CB,$F7,$2E,$91,$8B,$7C,$A6,$34,$F5,$FE,$DB
	.dc.b	$F9,$BF,$BE,$7E,$43,$FE,$F1,$FD,$05,$FA,$4B,$F4,$F7,$E8,$DE,$1E
	.dc.b	$6F,$F3,$9E,$1C,$3E,$01,$87,$0B,$DE,$FD,$72,$FE,$A5,$CB,$F0,$F0
	.dc.b	$E0,$D8,$F0,$67,$DF,$7F,$C1,$57,$5C,$7A,$7C,$1E,$2F,$79,$79,$38
	.dc.b	$3B,$69,$E2,$CB,$D1,$B3,$F4,$1F,$DC,$7C,$E7,$FA,$2F,$E5,$BF,$D4
	.dc.b	$BF,$1D,$FE,$AF,$FA,$37,$E8,$1A,$F9,$0F,$ED,$7F,$80,$FF,$7D,$FB
	.dc.b	$C5,$7E,$AF,$B7,$F4,$19,$7F,$A5,$7B,$7F,$02,$F5,$E7,$FC,$03,$DF
	.dc.b	$0E,$5F,$13,$F9,$D8,$C6,$9F,$EB,$FF,$A3,$C6,$9E,$08,$F2,$7C,$4F
	.dc.b	$8F,$7F,$3E,$DE,$2F,$AF,$FE,$25,$F9,$9F,$F5,$EF,$CD,$7F,$D9,$9A
	.dc.b	$3F,$1C,$7E,$D7,$FC,$7F,$8B,$F6,$D7,$BE,$EF,$E8,$F5,$AF,$D0,$F9
	.dc.b	$FC,$97,$F4,$7E,$4F,$CC,$5F,$56,$C3,$3E,$F6,$FD,$9B,$F7,$17,$D7
	.dc.b	$4F,$17,$E0,$78,$C3,$86,$7E,$0D,$7F,$19,$AF,$92,$73,$7E,$99,$FF
	.dc.b	$6D,$CE,$73,$EF,$0F,$CB,$7C,$F1,$8F,$CB,$7F,$9A,$7E,$80,$FC,$AD
	.dc.b	$F9,$A3,$F5,$7F,$C1,$70,$FF,$B5,$3A,$B7,$5A,$9C,$F0,$3F,$56,$CE
	.dc.b	$1F,$BC,$C7,$A7,$E3,$57,$0E,$1E,$17,$BB,$9B,$79,$7E,$3F,$F3,$1F
	.dc.b	$F6,$37,$9D,$FE,$D8,$C7,$C9,$97,$8F,$0F,$8B,$FB,$AF,$07,$E3,$FF
	.dc.b	$00,$7F,$4D,$F8,$D7,$E5,$FF,$C5,$FF,$65,$FF,$7E,$FE,$39,$FE,$58
	.dc.b	$FD,$F7,$53,$9F,$C4,$D3,$FE,$5C,$FD,$47,$3E,$F7,$F1,$4C,$79,$FE
	.dc.b	$1B,$70,$E1,$7F,$E1,$5E,$3D,$E7,$DA,$FF,$99,$1E,$2C,$11,$E2,$FB
	.dc.b	$57,$EB,$BC,$34,$F0,$C7,$F2,$AF,$03,$FB,$96,$23,$FD,$57,$F9,$C7
	.dc.b	$0F,$4D,$FB,$CB,$DD,$7C,$F3,$F1,$EF,$F8,$03,$F3,$1F,$F0,$03,$F0
	.dc.b	$56,$3C,$70,$E1,$7F,$EF,$0C,$FB,$E7,$DB,$F8,$07,$FD,$A3,$8F,$1C
	.dc.b	$3C,$F3,$FE,$AF,$8C,$6F,$FC,$73,$1F,$17,$DC,$7F,$BE,$27,$3B,$79
	.dc.b	$B6,$FC,$F6,$1E,$43,$D9,$AF,$93,$1E,$53,$EA,$F0,$35,$5F,$02,$73
	.dc.b	$B3,$F4,$8C,$E1,$C7,$1E,$D3,$FE,$D1,$DF,$D7,$39,$D7,$DF,$FB,$27
	.dc.b	$4F,$CC,$9F,$AB,$7F,$8D,$FF,$6E,$FE,$91,$F4,$AF,$BA,$7E,$CB,$BA
	.dc.b	$D7,$C8,$FF,$88,$A7,$FD,$DB,$A7,$C8,$C7,$9F,$D5,$FC,$8E,$2F,$B1
	.dc.b	$1D,$BE,$EB,$10,$8D,$E9,$E0,$C7,$E4,$4F,$0C,$BB,$E8,$F8,$EC,$CF
	.dc.b	$19,$CE,$8F,$3B,$CD,$A3,$2F,$5C,$BE,$4F,$E5,$FE,$3B,$F9,$0F,$D6
	.dc.b	$BF,$C7,$BF,$AD,$3F,$4A,$7E,$24,$FD,$33,$7D,$8B,$DD,$6B,$F5,$9C
	.dc.b	$C7,$B3,$1F,$CD,$F8,$56,$F8,$7D,$5B,$3E,$D9,$FA,$AF,$C5,$39,$C9
	.dc.b	$F8,$A6,$70,$E3,$87,$0F,$C0,$BB,$DE,$F1,$7E,$81,$F8,$6A,$7E,$87
	.dc.b	$FC,$47,$EF,$3E,$47,$C5,$3F,$6E,$7E,$43,$BA,$D7,$5A,$B5,$F9,$06
	.dc.b	$7E,$07,$EB,$E5,$D3,$E0,$EE,$1C,$2F,$E3,$F7,$6F,$4E,$FC,$5F,$5B
	.dc.b	$0F,$1E,$3F,$E8,$9C,$FB,$6A,$FD,$E7,$84,$63,$F6,$5B,$9F,$9F,$F5
	.dc.b	$4B,$6F,$C8,$7F,$47,$FA,$2B,$E1,$9F,$90,$FD,$9F,$CF,$BF,$05,$F1
	.dc.b	$EF,$D2,$77,$5A,$ED,$56,$BF,$4A,$CE,$5F,$94,$CB,$A6,$57,$0E,$17
	.dc.b	$F8,$DF,$96,$FC,$8D,$6B,$1E,$7E,$4B,$E3,$B3,$87,$08,$FE,$D3,$F8
	.dc.b	$B7,$8B,$2F,$34,$7F,$6C,$F8,$4F,$DE,$FE,$D6,$FC,$E3,$E2,$7E,$6D
	.dc.b	$FE,$CF,$FC,$79,$FB,$4F,$FB,$D7,$D7,$FE,$2B,$7B,$EB,$5E,$EB,$FA
	.dc.b	$C5,$6B,$0F,$9D,$E0,$7F,$17,$3F,$95,$61,$DF,$F7,$EF,$D9,$BC,$F0
	.dc.b	$F0,$78,$1F,$AE,$8C,$79,$2D,$3E,$FE,$5D,$74,$F0,$CE,$74,$7E,$59
	.dc.b	$EC,$D7,$D7,$39,$F8,$2F,$E3,$7F,$A9,$7E,$D7,$F1,$6F,$B3,$4F,$DE
	.dc.b	$5F,$9E,$3F,$7D,$DF,$B3,$FE,$EE,$E7,$F8,$D5,$AF,$B7,$CE,$3F,$C3
	.dc.b	$32,$FE,$01,$F2,$57,$0E,$19,$F5,$F9,$2F,$E1,$35,$E1,$CB,$9F,$E1
	.dc.b	$78,$BD,$E7,$E2,$F1,$FA,$D7,$83,$67,$E9,$F8,$CF,$D1,$0F,$C5,$31
	.dc.b	$87,$AB,$D6,$FE,$73,$E3,$3F,$59,$7D,$AB,$F6,$47,$EF,$CE,$EF,$DA
	.dc.b	$FE,$85,$C2,$F7,$5A,$9F,$DC,$FD,$2E,$7F,$04,$FE,$57,$97,$5C,$AE
	.dc.b	$1C,$32,$E9,$EE,$7F,$28,$AD,$65,$CF,$C1,$C5,$EF,$3F,$E0,$BF,$B4
	.dc.b	$F8,$72,$F3,$65,$CB,$F6,$AF,$EA,$7F,$BE,$FF,$2D,$D7,$F7,$5F,$29
	.dc.b	$F6,$7F,$6A,$7E,$1B,$AE,$B5,$6B,$95,$5F,$E8,$4D,$FF,$B7,$E3,$CF
	.dc.b	$CC,$FC,$B0,$7F,$9D,$7E,$55,$FD,$21,$F0,$79,$3F,$DF,$B0,$E7,$5D
	.dc.b	$F1,$ED,$DB,$FD,$1A,$F7,$AF,$97,$F1,$BD,$63,$C7,$4F,$F7,$9E,$9F
	.dc.b	$E0,$DC,$BF,$C1,$DE,$B3,$F5,$87,$E7,$CF,$CF,$5F,$85,$3E,$EF,$FD
	.dc.b	$3B,$F1,$EC,$B9,$65,$FE,$5B,$C7,$8E,$1C,$2F,$FD,$47,$F1,$FB,$AF
	.dc.b	$06,$7D,$F3,$E3,$1F,$E1,$FC,$79,$CF,$8F,$F7,$9C,$CF,$13,$CF,$97
	.dc.b	$6D,$3C,$9C,$FC,$67,$AC,$E7,$37,$D1,$E6,$78,$E7,$ED,$C3,$D9,$F8
	.dc.b	$2F,$F2,$9D,$3C,$38,$F1,$8F,$F1,$CC,$62,$B8,$63,$FE,$BE,$F7,$3C
	.dc.b	$98,$FC,$19,$CD,$FF,$93,$2F,$E2,$F1,$FC,$7F,$D7,$DF,$BF,$3A,$FF
	.dc.b	$DD,$9F,$99,$7F,$A9,$72,$FA,$97,$BA,$D7,$D6,$A9,$F7,$FE,$0A,$FA
	.dc.b	$5F,$A2,$7C,$6C,$FB,$8C,$F1,$97,$48,$C6,$5D,$3F,$3A,$F8,$92,$C3
	.dc.b	$C1,$A7,$FB,$FF,$EF,$B3,$18,$D7,$C5,$F8,$1F,$C0,$E2,$7E,$69,$FE
	.dc.b	$C7,$FC,$11,$FB,$BB,$F0,$7F,$C5,$7E,$1B,$BA,$D4,$E7,$BD,$F8,$96
	.dc.b	$7E,$23,$CF,$C1,$FE,$61,$CF,$AF,$DE,$E2,$3F,$25,$8F,$2C,$DF,$8D
	.dc.b	$FA,$AF,$7A,$F8,$A5,$FA,$5E,$6F,$86,$9F,$E0,$8E,$6F,$CC,$DB,$F0
	.dc.b	$0F,$E7,$7F,$24,$FD,$21,$F9,$7B,$EE,$9F,$80,$FE,$03,$0E,$1F,$1D
	.dc.b	$FE,$2B,$BD,$D3,$F0,$1D,$64,$FC,$67,$D5,$87,$0E,$3F,$D0,$EB,$E2
	.dc.b	$AF,$0E,$3E,$0F,$C6,$3E,$AD,$F6,$DB,$C9,$B7,$F5,$4C,$7F,$75,$39
	.dc.b	$BF,$0E,$9E,$0D,$7C,$97,$E9,$FA,$BC,$63,$CA,$BF,$D2,$18,$FF,$03
	.dc.b	$FC,$41,$FA,$23,$F7,$DF,$ED,$FF,$27,$FE,$59,$F1,$3F,$AC,$70,$FF
	.dc.b	$00,$BD,$D6,$BE,$E1,$5F,$E3,$0A,$D6,$1C,$F1,$F7,$9D,$D7,$EE,$FF
	.dc.b	$77,$1E,$7D,$DF,$D8,$F1,$E7,$7F,$AE,$9C,$E7,$FE,$6B,$F1,$7F,$33
	.dc.b	$18,$AF,$0F,$D7,$7E,$D3,$C5,$FF,$53,$FE,$BA,$FC,$83,$FC,$65,$F1
	.dc.b	$AF,$DA,$DF,$AC,$6F,$F0,$15,$EF,$DF,$FF,$32,$D3,$F3,$FD,$4F,$E4
	.dc.b	$F2,$FF,$59,$FE,$5B,$70,$E3,$9F,$6F,$D7,$FF,$5B,$5E,$28,$F0,$FE
	.dc.b	$29,$8A,$DE,$7D,$BA,$3A,$D7,$9B,$47,$EF,$3F,$C3,$BD,$DF,$F1,$DF
	.dc.b	$B4,$FD,$17,$FD,$97,$F8,$67,$E5,$7E,$63,$5F,$2A,$AD,$78,$6A,$FC
	.dc.b	$D7,$EF,$3A,$9F,$D6,$63,$FE,$46,$F1,$2E,$1C,$72,$EB,$E4,$FF,$06
	.dc.b	$AF,$0C,$78,$3A,$71,$5B,$CF,$B7,$8B,$FE,$8C,$F1,$3F,$A2,$E1,$C2
	.dc.b	$73,$FB,$AF,$87,$1F,$4D,$BF,$75,$FE,$57,$FB,$27,$EC,$FD,$A7,$E5
	.dc.b	$1F,$5B,$F7,$BF,$F0,$9F,$CC,$EB,$D1,$56,$BC,$35,$7F,$86,$3A,$35
	.dc.b	$ED,$FF,$09,$CB,$A7,$95,$F9,$23,$FD,$DC,$F7,$CF,$B7,$D6,$FF,$24
	.dc.b	$8C,$63,$D3,$97,$1A,$78,$2F,$C9,$F4,$4F,$D8,$7F,$A1,$FE,$AA,$FC
	.dc.b	$15,$F5,$8F,$CD,$DF,$BE,$7F,$B6,$FF,$30,$FF,$50,$F6,$7F,$6A,$C3
	.dc.b	$85,$EE,$B5,$FB,$9F,$4C,$79,$F1,$7A,$45,$F7,$F5,$3F,$93,$3C,$B3
	.dc.b	$EF,$F8,$BF,$F7,$CA,$E5,$97,$58,$7E,$85,$FD,$0A,$F8,$67,$E5,$FB
	.dc.b	$5F,$F3,$7F,$E6,$9F,$E0,$7F,$D4,$9F,$83,$BF,$3D,$FF,$1E,$7E,$8C
	.dc.b	$FD,$71,$F9,$83,$D8,$E3,$C0,$18,$F2,$C3,$8D,$F0,$FC,$27,$75,$AC
	.dc.b	$BF,$C5,$AF,$F8,$1B,$E0,$1A,$78,$E7,$C5,$39,$8C,$61,$D3,$E6,$FC
	.dc.b	$17,$C2,$BC,$59,$78,$7E,$C1,$B7,$A7,$EB,$B7,$97,$29,$CE,$BE,$CF
	.dc.b	$7B,$CB,$87,$1C,$3E,$07,$03,$F3,$3C,$EF,$F1,$71,$E3,$2F,$C8,$FD
	.dc.b	$BE,$AB,$F7,$67,$DE,$BF,$31,$7E,$57,$FE,$DC,$FD,$53,$FB,$FF,$27
	.dc.b	$A7,$93,$F7,$98,$70,$BD,$FE,$28,$B7,$E7,$FD,$A9,$F6,$BF,$E2,$AC
	.dc.b	$79,$69,$E1,$D3,$1C,$DF,$E4,$38,$C6,$7D,$3E,$CF,$E1,$EB,$F6,$67
	.dc.b	$3B,$7F,$BC,$3E,$C7,$33,$E4,$C7,$D5,$3E,$9F,$A2,$F1,$46,$38,$3C
	.dc.b	$DE,$17,$DE,$FD,$61,$F6,$FF,$CD,$7F,$24,$7E,$85,$FE,$4D,$FD,$BB
	.dc.b	$FC,$C9,$C5,$F0,$6F,$7F,$14,$BD,$F9,$EB,$FD,$1D,$EF,$1F,$E8,$1B
	.dc.b	$FE,$14,$FC,$96,$3E,$0F,$CF,$7E,$0C,$62,$7C,$59,$FF,$97,$7F,$05
	.dc.b	$F6,$DF,$83,$3E,$FF,$9F,$E2,$F7,$A7,$86,$FF,$33,$F5,$AF,$DA,$7F
	.dc.b	$A4,$7E,$5C,$FD,$4F,$FA,$57,$DF,$3F,$60,$DD,$6A,$73,$E8,$E3,$ED
	.dc.b	$BC,$BC,$3F,$E4,$8A,$FE,$A4,$FA,$CC,$7C,$DB,$2E,$CB,$4F,$AB,$A7
	.dc.b	$8E,$35,$97,$F5,$BF,$C3,$33,$8F,$86,$7F,$D2,$5F,$94,$79,$1D,$9F
	.dc.b	$E9,$5F,$DD,$7F,$6B,$FB,$07,$FA,$E7,$F4,$5F,$F1,$47,$EF,$47,$FD
	.dc.b	$43,$FB,$2E,$BE,$3F,$1F,$0E,$AD,$7E,$0E,$E9,$D7,$C4,$7B,$B2,$E9
	.dc.b	$C9,$ED,$7B,$61,$C7,$3F,$F6,$5C,$F5,$7C,$27,$F1,$DE,$FF,$D3,$7D
	.dc.b	$37,$F8,$DB,$E2,$3E,$77,$E2,$DF,$E3,$1E,$07,$CB,$FF,$0E,$AD,$76
	.dc.b	$7F,$BF,$FE,$C1,$5F,$E3,$FC,$79,$FD,$97,$F3,$DD,$3E,$86,$1F,$E4
	.dc.b	$2E,$3F,$AA,$F2,$CF,$B4,$F5,$AF,$E8,$99,$75,$FB,$5F,$D3,$DE,$F3
	.dc.b	$EF,$F3,$BF,$C8,$36,$7E,$4D,$FB,$BF,$E0,$0F,$DE,$71,$7F,$99,$BE
	.dc.b	$5B,$53,$9F,$6B,$FC,$B1,$F7,$7F,$D8,$B8,$BF,$83,$63,$CB,$E1,$3F
	.dc.b	$12,$F4,$AB,$7C,$71,$FF,$5A,$5F,$08,$C7,$E1,$FF,$BC,$BE,$19,$F8
	.dc.b	$B2,$FF,$3B,$DF,$F1,$3F,$DC,$7F,$3D,$E2,$BF,$70,$FF,$4E,$FE,$A5
	.dc.b	$FE,$68,$FE,$01,$FE,$77,$5A,$9C,$FD,$BF,$FA,$57,$0E,$5E,$37,$E5
	.dc.b	$FA,$7F,$39,$7F,$A9,$F9,$1F,$D0,$BC,$7C,$A6,$6F,$C3,$3F,$88,$CE
	.dc.b	$63,$FD,$7F,$7E,$0C,$3F,$45,$9F,$83,$F5,$DF,$E9,$BF,$4F,$FC,$27
	.dc.b	$EA,$FF,$25,$E4,$FF,$83,$B1,$78,$DA,$F5,$7F,$C5,$1E,$4F,$E6,$7C
	.dc.b	$9F,$0F,$EC,$D9,$C7,$F6,$B9,$7F,$13,$F8,$0B,$87,$0C,$FF,$C3,$3F
	.dc.b	$95,$FF,$26,$AF,$06,$5C,$FF,$4C,$C5,$EF,$3F,$6F,$77,$F6,$29,$7E
	.dc.b	$E0,$FD,$ED,$F9,$4F,$F9,$67,$F6,$2F,$E1,$DF,$01,$BA,$D4,$E7,$99
	.dc.b	$FA,$3E,$70,$FE,$17,$8F,$4F,$DB,$AE,$1C,$3E,$D9,$9F,$D3,$EF,$FE
	.dc.b	$A1,$7E,$2C,$3C,$35,$AF,$6B,$FA,$0C,$E7,$4F,$07,$C2,$FF,$7A,$FA
	.dc.b	$A7,$E9,$1F,$E2,$3F,$D8,$5F,$AE,$FA,$BF,$46,$74,$EE,$B5,$39,$FD
	.dc.b	$13,$1F,$BF,$7F,$22,$C7,$9F,$D5,$DB,$DE,$8F,$C6,$3F,$A9,$D6,$B4
	.dc.b	$EF,$2F,$C0,$7F,$AB,$FE,$F6,$F7,$9F,$8C,$3F,$5B,$7E,$61,$FC,$1B
	.dc.b	$C9,$7E,$8B,$FE,$14,$C3,$85,$EE,$B5,$FB,$7E,$71,$FF,$64,$79,$DF
	.dc.b	$CD,$6F,$6D,$8E,$F9,$76,$F8,$6F,$83,$3C,$71,$E9,$0E,$9F,$ED,$BF
	.dc.b	$AF,$BF,$34,$7E,$17,$FD,$FF,$EB,$1F,$A0,$7F,$8C,$7F,$4B,$FF,$14
	.dc.b	$DE,$EB,$53,$FE,$35,$FC,$63,$38,$72,$D3,$FA,$55,$F5,$F5,$7F,$A7
	.dc.b	$BC,$7E,$0F,$FD,$47,$83,$E8,$FC,$78,$78,$A7,$3F,$45,$FE,$02,$F7
	.dc.b	$5F,$61,$FC,$FF,$E3,$5F,$91,$3C,$77,$80,$FC,$F7,$FC,$7F,$87,$0F
	.dc.b	$67,$FB,$25,$EE,$BF,$3F,$E7,$7F,$88,$6B,$5C,$5F,$D8,$5E,$FF,$BB
	.dc.b	$7F,$66,$79,$67,$E0,$CD,$E6,$F9,$75,$F2,$67,$E3,$E9,$F4,$64,$FB
	.dc.b	$6F,$E8,$3F,$A7,$7F,$8A,$BF,$74,$7E,$FD,$E6,$AD,$7C,$BA,$B5,$FB
	.dc.b	$FE,$70,$F1,$57,$EE,$0F,$31,$C2,$7F,$C6,$BE,$FF,$82,$7B,$FD,$8B
	.dc.b	$F2,$3F,$F1,$5F,$C6,$BF,$51,$7E,$CB,$FE,$AC,$F0,$5F,$9F,$BF,$72
	.dc.b	$FF,$33,$DE,$EB,$52,$FE,$00,$CE,$1C,$B6,$FF,$04,$5F,$5F,$D4,$FF
	.dc.b	$C1,$AF,$1F,$8B,$7E,$24,$D3,$FC,$6B,$7C,$B3,$F1,$74,$79,$BF,$2A
	.dc.b	$7E,$E0,$FD,$11,$FB,$F7,$EB,$7F,$06,$7E,$EB,$E6,$BD,$D6,$BD,$5F
	.dc.b	$F4,$57,$D2,$E7,$ED,$FC,$75,$D3,$5F,$3B,$C3,$2F,$DC,$7E,$69,$E1
	.dc.b	$D3,$C1,$1F,$BB,$BF,$93,$F6,$9F,$A7,$EB,$7D,$A7,$FA,$7B,$F7,$4F
	.dc.b	$2F,$97,$FE,$6B,$E7,$5E,$EB,$5F,$BC,$FC,$C8,$FD,$86,$DF,$21,$7E
	.dc.b	$A7,$F2,$1E,$1F,$7D,$FE,$9D,$9E,$1E,$A7,$E2,$F0,$FF,$56,$F2,$7D
	.dc.b	$CF,$AB,$FF,$51,$6D,$FC,$F3,$EE,$DF,$6A,$FD,$A3,$ED,$D6,$BE,$5B
	.dc.b	$FC,$C7,$2F,$84,$CE,$9F,$62,$BF,$DF,$8F,$DC,$78,$61,$FE,$1A,$8D
	.dc.b	$E3,$DF,$A7,$F1,$BF,$AD,$FF,$AC,$F6,$FD,$07,$C0,$7F,$23,$F7,$FF
	.dc.b	$17,$39,$BF,$E1,$FD,$BF,$E5,$BE,$7F,$91,$3E,$FC,$A4,$63,$C4,$FC
	.dc.b	$5C,$F1,$C7,$F8,$45,$74,$3E,$A2,$F7,$3F,$3F,$5F,$27,$1F,$D3,$61
	.dc.b	$F5,$63,$11,$FE,$34,$FC,$27,$F3,$F1,$E1,$C3,$E8,$56,$E3,$F3,$67
	.dc.b	$0F,$47,$AC,$FD,$0B,$6D,$FA,$E8,$FC,$09,$F8,$78,$F7,$7E,$09,$F6
	.dc.b	$FB,$3F,$DB,$1D,$5F,$0E,$7F,$D2,$73,$9A,$E2,$7C,$27,$1F,$A4,$FE
	.dc.b	$B2,$3A,$F1,$7C,$E7,$F6,$E6,$2B,$FA,$A3,$8E,$4F,$7F,$3D,$8F,$47
	.dc.b	$CC,$57,$B1,$FE,$EB,$CF,$FD,$D9,$F0,$FD,$EF,$E5,$71,$FB,$5F,$99
	.dc.b	$F1,$DE,$FE,$5F,$CB,$69,$FC,$7F,$4F,$1C,$61,$FF,$07,$F1,$FE,$97
	.dc.b	$34,$FA,$EB,$D8,$F2,$C7,$A4,$FE,$2F,$0F,$99,$C7,$F8,$5C,$7F,$96
	.dc.b	$63,$CE,$31,$B7,$C2,$FA,$3F,$8D,$1F,$8B,$C7,$8E,$3F,$AB,$C7,$EB
	.dc.b	$3F,$3B,$3E,$7C,$3F,$46,$31,$8F,$CC,$7E,$9F,$EE,$9C,$38,$FF,$8F
	.dc.b	$C0,$6F,$E5,$D7,$DD,$18,$3F,$75,$7C,$E7,$8C,$64,$5F,$93,$7F,$D1
	.dc.b	$65,$E1,$7F,$49,$E0,$DA,$FA,$E7,$F9,$74,$F9,$67,$8E,$BD,$BB,$76
	.dc.b	$3F,$58,$7A,$CF,$E1,$8F,$C2,$8F,$C7,$9E,$F3,$F2,$31,$73,$E9,$19
	.dc.b	$C7,$C3,$C5,$FB,$77,$F8,$B5,$66,$7D,$67,$3C,$76,$FF,$1B,$77,$A7
	.dc.b	$C6,$8F,$4E,$3E,$3A,$FE,$D1,$F4,$3F,$EF,$0D,$BE,$A5,$27,$39,$79
	.dc.b	$DD,$C5,$E2,$FE,$9B,$8F,$59,$F0,$CF,$D7,$7C,$37,$D1,$F6,$CF,$8E
	.dc.b	$7F,$62,$DD,$38,$37,$2F,$91,$3F,$66,$38,$69,$AC,$3C,$F3,$F8,$4A
	.dc.b	$DE,$9F,$C3,$27,$2E,$E7,$F6,$07,$F2,$17,$E4,$DF,$8F,$1E,$13,$A3
	.dc.b	$E7,$9F,$0E,$35,$7E,$48,$FF,$58,$DF,$89,$FE,$A7,$5F,$75,$FC,$C7
	.dc.b	$CA,$6F,$19,$39,$97,$E3,$3C,$F2,$93,$93,$B1,$E1,$C3,$B5,$7F,$60
	.dc.b	$CB,$E5,$3F,$A5,$E7,$E8,$FE,$9A,$32,$FF,$BA,$1B,$7A,$E1,$7E,$02
	.dc.b	$FD,$0F,$FB,$16,$22,$61,$F9,$A3,$78,$6A,$34,$4F,$13,$F6,$D6,$9F
	.dc.b	$89,$BE,$3E,$DB,$85,$F0,$3F,$C2,$18,$78,$BF,$28,$DB,$E4,$DA,$E3
	.dc.b	$FD,$5E,$B8,$1E,$5C,$BB,$D7,$F7,$9F,$D0,$1F,$80,$42,$7E,$53,$A9
	.dc.b	$FE,$15,$39,$9D,$7E,$5D,$99,$7C,$FF,$F2,$DE,$8F,$E5,$DD,$D9,$8C
	.dc.b	$47,$F9,$3F,$BB,$86,$BF,$D6,$EF,$F6,$6F,$F6,$DF,$D0,$FC,$FB,$7B
	.dc.b	$A3,$FB,$86,$7E,$13,$CF,$AF,$85,$C0,$D3,$C9,$F3,$BE,$0C,$FF,$D0
	.dc.b	$17,$FD,$E6,$31,$9F,$A7,$B7,$C5,$C9,$FC,$9F,$4F,$B3,$AF,$8B,$5F
	.dc.b	$9F,$18,$DB,$A6,$FB,$9F,$F1,$14,$74,$A5,$72,$DB,$CB,$BE,$E7,$F9
	.dc.b	$A4,$7A,$37,$C4,$7F,$27,$3F,$11,$4F,$DF,$BF,$C5,$D5,$CA,$7A,$C6
	.dc.b	$23,$E2,$57,$A7,$D2,$C4,$7D,$7C,$31,$7E,$08,$DB,$E6,$F8,$FF,$CA
	.dc.b	$61,$FE,$1D,$D7,$FA,$47,$42,$73,$5F,$9B,$F0,$1D,$1F,$90,$7F,$31
	.dc.b	$1F,$D3,$F7,$FB,$3C,$FF,$E0,$AE,$AF,$CC,$CF,$E8,$F9,$9F,$85,$78
	.dc.b	$2F,$F4,$53,$FA,$63,$F1,$87,$F3,$C9,$EA,$FC,$C8,$FC,$DD,$FE,$A3
	.dc.b	$3F,$F0,$C7,$81,$FE,$25,$D7,$17,$FE,$07,$FC,$9F,$6F,$51,$CF,$FD
	.dc.b	$09,$F2,$5F,$F8,$87,$F6,$7D,$E7,$E0,$AF,$F2,$01,$E3,$9F,$E7,$FA
	.dc.b	$FE,$AE,$FA,$4F,$F9,$17,$6D,$47,$F2,$33,$F7,$9A,$FE,$6E,$FF,$98
	.dc.b	$D7,$E8,$1F,$E6,$77,$9A,$E0,$7F,$9D,$B3,$F5,$BF,$85,$3F,$83,$6B
	.dc.b	$FC,$15,$FA,$18,$7F,$05,$C3,$F5,$7D,$DF,$C2,$23,$F8,$A6,$1F,$AB
	.dc.b	$8C,$47,$CD,$75,$C5,$EB,$D1,$1C,$6F,$58,$76,$BE,$67,$37,$0F,$DD
	.dc.b	$9F,$E2,$F7,$F7,$2F,$9D,$7E,$17,$ED,$FE,$B3,$FC,$AA,$BF,$94,$FE
	.dc.b	$DD,$F4,$1F,$E5,$7E,$E6,$AF,$8D,$7F,$2D,$F7,$F4,$FC,$63,$F8,$4F
	.dc.b	$E5,$1E,$CC,$FF,$92,$F0,$FE,$3A,$7E,$2E,$9F,$CA,$77,$F9,$11,$F8
	.dc.b	$2F,$C9,$BB,$69,$F8,$A8,$FD,$11,$E5,$3F,$AE,$78,$9E,$D7,$EE,$70
	.dc.b	$7C,$47,$F7,$11,$63,$FC,$EC,$79,$23,$E2,$E5,$F9,$BC,$FE,$7C,$64
	.dc.b	$FE,$6A,$64,$F9,$D9,$7E,$9E,$3E,$1C,$79,$35,$47,$F5,$07,$D3,$1F
	.dc.b	$E6,$5F,$C7,$BE,$2B,$F9,$EE,$6B,$8F,$F4,$3C,$7E,$C4,$FF,$0F,$E3
	.dc.b	$FC,$06,$9F,$E0,$49,$CD,$7F,$A3,$E7,$F9,$96,$5F,$86,$9B,$7F,$80
	.dc.b	$ED,$FC,$44,$FE,$02,$6C,$FE,$1E,$7F,$05,$3F,$D9,$B8,$7F,$5A,$FF
	.dc.b	$68,$EF,$FD,$7F,$EA,$9F,$E0,$3E,$AF,$F4,$8C,$E7,$1F,$B1,$8F,$FB
	.dc.b	$63,$4E,$51,$8E,$4F,$87,$3F,$C4,$78,$3F,$11,$FB,$FF,$E0,$F9,$BF
	.dc.b	$D3,$78,$7C,$AA,$FE,$25,$3F,$E1,$1E,$5E,$F1,$FC,$67,$6F,$E0,$F8
	.dc.b	$FE,$43,$27,$3E,$34,$E5,$7F,$C5,$7E,$77,$F2,$1F,$0D,$DA,$3F,$3C
	.dc.b	$7D,$7E,$6F,$F3,$C7,$47,$F1,$08,$FE,$BF,$18,$E5,$FD,$83,$AF,$D8
	.dc.b	$3F,$AE,$CF,$F8,$3A,$31,$E1,$7F,$5D,$8C,$1F,$94,$F2,$7F,$D5,$DC
	.dc.b	$8A,$EF,$1F,$DB,$E7,$67,$F9,$CF,$1F,$F6,$0F,$77,$F8,$42,$BC,$BD
	.dc.b	$FE,$2A,$FC,$FF,$CF,$78,$B1,$E9,$18,$CB,$C5,$18,$8E,$9F,$B1,$F8
	.dc.b	$B0,$F7,$BE,$19,$D7,$6F,$F9,$12,$BE,$09,$FC,$97,$8F,$C0,$FF,$61
	.dc.b	$F3,$3C,$18,$7F,$9B,$1F,$F0,$AF,$91,$E7,$9F,$E5,$9E,$4F,$76,$F0
	.dc.b	$7F,$B3,$32,$EF,$1F,$EA,$13,$FC,$6E,$78,$0F,$ED,$DD,$FE,$0C,$FF
	.dc.b	$C1,$BA,$BC,$9E,$D5,$FD,$CE,$3C,$2F,$CA,$9F,$F0,$56,$1F,$E0,$CD
	.dc.b	$3F,$D0,$3E,$67,$BD,$FF,$41,$CF,$F5,$97,$FA,$3E,$DF,$E1,$B7,$FD
	.dc.b	$0F,$C9,$EE,$C3,$FA,$33,$DB,$4F,$4C,$7F,$BE,$7C,$BF,$1E,$1F,$EC
	.dc.b	$DC,$BD,$D8,$7F,$A4,$3A,$FB,$5F,$F4,$FE,$4F,$63,$F0,$79,$3B,$3F
	.dc.b	$E0,$D7,$B5,$7F,$A6,$39,$3E,$54,$FF,$84,$5F,$53,$FC,$B7,$97,$FC
	.dc.b	$AB,$F9,$37,$F2,$DF,$1E,$CF,$F4,$0E,$5F,$C8,$CF,$F4,$F7,$F0,$0F
	.dc.b	$F5,$1E,$3F,$C5,$35,$FC,$4B,$FD,$36,$7F,$C8,$2F,$F5,$7E,$9F,$EA
	.dc.b	$4F,$F5,$9D,$BA,$CF,$F4,$E7,$AB,$FD,$43,$4E,$AD,$BE,$AF,$F5,$43
	.dc.b	$FC,$15,$5F,$C5,$78,$FF,$A2,$F6,$7F,$9B,$3C,$5F,$F0,$8F,$0F,$EF
	.dc.b	$E7,$FC,$F5,$E3,$FF,$59,$BF,$EB,$7A,$F8,$B1,$FF,$39,$4F,$F9,$83
	.dc.b	$4F,$C8,$57,$F6,$0C,$FC,$75,$A7,$EE,$47,$F6,$4E,$3F,$1C,$FF,$3E
	.dc.b	$DB,$DE,$FF,$40,$BE,$D9,$7F,$A1,$B1,$47,$C0,$9F,$0E,$DD,$AB,$4F
	.dc.b	$68,$FE,$92,$76,$3F,$D1,$1A,$FB,$B6,$FE,$EB,$19,$AF,$F4,$9C,$FA
	.dc.b	$F9,$3E,$EC,$F0,$DB,$C1,$5A,$DB,$C0,$FF,$A7,$EF,$FD,$07,$5F,$EA
	.dc.b	$4D,$BE,$66,$DD,$6B,$5B,$7B,$AB,$C7,$AF,$BA,$79,$3D,$9E,$9A,$F6
	.dc.b	$58,$A3,$DD,$5E,$C3,$B4,$BF,$66,$FB,$7E,$27,$F8,$86,$7E,$16,$5F
	.dc.b	$C0,$FF,$40,$FF,$29,$69,$FC,$16,$3F,$83,$7B,$1F,$7E,$BF,$DB,$4F
	.dc.b	$C3,$7D,$86,$37,$FA,$53,$C3,$7F,$BE,$D7,$F2,$33,$FC,$FC,$57,$F4
	.dc.b	$1E,$5F,$9B,$5E,$0E,$0F,$88,$F8,$B8,$3F,$1C,$D7,$F1,$B7,$E4,$DF
	.dc.b	$BA,$3A,$BF,$12,$7D,$71,$F0,$0D,$78,$1E,$D7,$FA,$16,$7D,$9C,$BE
	.dc.b	$C7,$79,$F6,$53,$F7,$BA,$9A,$FF,$5C,$C3,$B5,$6B,$3F,$5B,$E0,$CF
	.dc.b	$B3,$C9,$EC,$F4,$CF,$B6,$3F,$77,$2F,$93,$87,$F4,$7D,$3E,$3D,$FF
	.dc.b	$BA,$79,$EB,$F0,$75,$EA,$FE,$CF,$D0,$CB,$FE,$DA,$3E,$C9,$FC,$EF
	.dc.b	$0E,$AF,$E0,$4F,$E2,$27,$F9,$DF,$4F,$4B,$FC,$C4,$E3,$8F,$F3,$1A
	.dc.b	$EB,$9F,$C5,$9F,$47,$81,$E0,$7F,$D3,$1F,$93,$FF,$27,$5C,$08,$FE
	.dc.b	$45,$E0,$7F,$C0,$07,$F9,$25,$5A,$FE,$3D,$C1,$FE,$D8,$9F,$E4,$4F
	.dc.b	$F5,$D7,$F3,$91,$FE,$E8,$7F,$8E,$6D,$C6,$3F,$88,$6B,$FD,$6D,$F0
	.dc.b	$CF,$F1,$58,$FE,$F2,$FF,$9D,$AB,$FC,$12,$D7,$B4,$FF,$0A,$3B,$1F
	.dc.b	$E3,$BE,$6F,$F5,$07,$5F,$F8,$C2,$3F,$91,$C6,$39,$BF,$9D,$3F,$EF
	.dc.b	$9B,$FE,$5D,$87,$F8,$2D,$AF,$8A,$BF,$C8,$D8,$78,$9A,$F8,$9F,$DF
	.dc.b	$1E,$21,$50,$FE,$00,$7F,$A6,$6B,$FC,$00,$E7,$9F,$F0,$5A,$FD,$1A
	.dc.b	$FF,$4D,$47,$E6,$EB,$75,$D2,$3F,$97,$DF,$D1,$AD,$BF,$E1,$BA,$DD
	.dc.b	$7F,$0F,$9F,$BE,$7F,$A9,$7E,$47,$F9,$73,$FD,$44,$F4,$47,$9E,$3C
	.dc.b	$73,$C9,$F1,$4F,$1A,$F8,$4F,$B6,$93,$E4,$E0,$FB,$D1,$CE,$7C,$26
	.dc.b	$FE,$3F,$FD,$61,$87,$99,$EF,$1F,$EB,$97,$BB,$BC,$3B,$D7,$CB,$9E
	.dc.b	$D3,$F8,$83,$81,$DA,$3E,$87,$EE,$FF,$CB,$A3,$FD,$83,$C7,$FE,$F7
	.dc.b	$FC,$C7,$F5,$9D,$1F,$C8,$7A,$3F,$8B,$6E,$9C,$F3,$FF,$B1,$5D,$F8
	.dc.b	$7F,$59,$BC,$5F,$E2,$37,$9E,$DF,$53,$BB,$9C,$62,$3F,$C9,$CF,$83
	.dc.b	$97,$FC,$81,$96,$B5,$FE,$4D,$AF,$F2,$2F,$0F,$C1,$18,$3F,$89,$65
	.dc.b	$FE,$5F,$C3,$D3,$D5,$FC,$7D,$F5,$F8,$5E,$1A,$FD,$DC,$63,$3F,$0C
	.dc.b	$73,$E0,$FF,$5B,$4F,$F0,$D8,$FF,$4C,$F0,$7C,$17,$C0,$3C,$6F,$80
	.dc.b	$E1,$F0,$3A,$ED,$F0,$4F,$18,$FE,$99,$1F,$5A,$31,$AF,$F9,$7D,$F8
	.dc.b	$33,$E6,$3B,$F8,$DF,$01,$FE,$A9,$AF,$81,$AF,$CF,$7C,$BE,$37,$83
	.dc.b	$2F,$53,$DD,$F6,$61,$DB,$0F,$0D,$7F,$53,$E4,$ED,$3F,$E7,$1E,$4E
	.dc.b	$CD,$C7,$B5,$7F,$65,$DB,$B3,$F1,$1E,$CF,$C6,$E3,$ED,$97,$CB,$E2
	.dc.b	$EB,$3F,$D8,$78,$BA,$B7,$0E,$B8,$FF,$A0,$78,$BA,$CF,$F7,$37,$AB
	.dc.b	$FD,$D4,$F5,$7E,$63,$FC,$DF,$63,$FB,$EE,$1F,$A2,$DD,$7F,$D3,$8F
	.dc.b	$F0,$2C,$7F,$86,$B5,$FF,$3D,$3F,$C1,$B1,$EB,$87,$F7,$7E,$7F,$E3
	.dc.b	$1C,$1F,$C6,$7B,$BF,$4B,$7F,$C3,$31,$EF,$7B,$BF,$E3,$F1,$F7,$E7
	.dc.b	$F8,$8E,$BE,$0A,$D5,$7F,$69,$CB,$FD,$6B,$5F,$C5,$A7,$C2,$7F,$92
	.dc.b	$E7,$F8,$AB,$FC,$35,$FF,$16,$8C,$3F,$DB,$51,$9F,$1B,$F9,$7C,$79
	.dc.b	$A3,$C9,$1F,$E6,$1C,$5D,$19,$9F,$F2,$C6,$BE,$77,$79,$79,$DF,$F2
	.dc.b	$E3,$FE,$0B,$CB,$F9,$8E,$1F,$C1,$E7,$FB,$3C,$63,$4F,$8D,$18,$3C
	.dc.b	$38,$FC,$67,$FC,$CC,$F7,$7F,$CC,$77,$FE,$D9,$AE,$BA,$77,$9F,$49
	.dc.b	$F1,$34,$FF,$2D,$1F,$10,$FF,$3E,$AE,$8F,$64,$7F,$97,$F3,$F0,$37
	.dc.b	$1F,$C1,$D7,$D7,$F1,$59,$7B,$EB,$FC,$CC,$F8,$67,$FC,$DF,$BF,$86
	.dc.b	$7F,$CB,$BA,$7B,$B8,$3F,$23,$5C,$F1,$EF,$5F,$E6,$27,$DC,$FF,$99
	.dc.b	$B0,$F7,$4F,$87,$4E,$D3,$FE,$75,$8E,$DD,$FF,$D2,$F8,$7E,$EC,$FE
	.dc.b	$06,$7D,$C7,$CD,$19,$FF,$93,$7C,$5F,$F7,$54,$6B,$3F,$2B,$A3,$FC
	.dc.b	$BB,$5E,$9E,$0F,$F1,$67,$8B,$F8,$5E,$3F,$C2,$63,$FE,$57,$8F,$C4
	.dc.b	$78,$18,$C3,$FC,$CB,$B7,$E1,$6B,$4F,$C7,$78,$1F,$9A,$39,$70,$77
	.dc.b	$9F,$E7,$57,$FD,$E3,$83,$F2,$71,$EB,$AF,$04,$7E,$15,$F1,$70,$77
	.dc.b	$9F,$F3,$4E,$1F,$B8,$79,$9D,$CF,$F3,$0E,$9D,$DF,$81,$DB,$ED,$7B
	.dc.b	$1F,$E6,$6E,$3E,$CF,$F9,$A2,$BB,$4F,$E1,$1E,$D1,$CB,$0E,$D2,$CF
	.dc.b	$B4,$7F,$9A,$72,$FF,$60,$EB,$F1,$EB,$DD,$AD,$D3,$A9,$F1,$F1,$FF
	.dc.b	$3B,$9E,$93,$FA,$8E,$DF,$D4,$F4,$FD,$46,$5F,$E8,$99,$FA,$67,$A8
	.dc.b	$F3,$69,$FE,$49,$DF,$EF,$CE,$E3,$B4,$FE,$DE,$72,$FF,$41,$DB,$F8
	.dc.b	$3B,$CF,$5E,$EF,$FA,$5D,$F8,$AF,$F1,$7F,$13,$F0,$4F,$F1,$9F,$12
	.dc.b	$BF,$D3,$7A,$3F,$9A,$1F,$CB,$71,$F2,$36,$7F,$C1,$AF,$F2,$9B,$F8
	.dc.b	$F5,$FE,$70,$8C,$43,$6D,$CF,$F2,$83,$B1,$FC,$73,$1F,$D9,$57,$F1
	.dc.b	$4C,$3F,$C9,$2D,$7F,$7B,$87,$2A,$FE,$23,$87,$83,$25,$FF,$1C,$8F
	.dc.b	$EE,$CF,$F1,$E9,$EE,$7F,$BC,$30,$FE,$20,$7F,$35,$E4,$FD,$AB,$FE
	.dc.b	$41,$E2,$FE,$69,$3E,$B3,$F8,$76,$5E,$46,$BE,$47,$47,$FB,$F3,$0F
	.dc.b	$F0,$4E,$1F,$E5,$EE,$6F,$C4,$CE,$B0,$FE,$0C,$FF,$B2,$F9,$FD,$91
	.dc.b	$FE,$63,$3E,$1C,$7F,$53,$E7,$FC,$2E,$FA,$9F,$F7,$86,$5F,$EC,$5C
	.dc.b	$7F,$CD,$75,$FB,$3E,$5D,$3E,$0D,$BE,$2E,$3C,$CF,$87,$87,$F9,$F3
	.dc.b	$C3,$FE,$2D,$0E,$CF,$F4,$74,$FF,$17,$D3,$C1,$5E,$2A,$F5,$4F,$F9
	.dc.b	$9B,$3F,$14,$7F,$69,$9F,$13,$A7,$C5,$9F,$FA,$01,$FE,$3B,$7E,$BF
	.dc.b	$0B,$FC,$6F,$3F,$E8,$47,$B6,$5E,$8B,$FF,$33,$61,$D9,$FF,$34,$6F
	.dc.b	$F3,$9F,$EF,$15,$D9,$A7,$BB,$27,$85,$EE,$D1,$E2,$7F,$A3,$27,$FD
	.dc.b	$3B,$CD,$FC,$BE,$FF,$6D,$FB,$0F,$F1,$DC,$BF,$1D,$18,$71,$CF,$FE
	.dc.b	$6E,$8F,$DE,$79,$FF,$CE,$1F,$B8,$7F,$B4,$8F,$E0,$1F,$A2,$FE,$FB
	.dc.b	$6F,$EA,$FC,$BC,$64,$27,$52,$E9,$D4,$BF,$10,$F2,$65,$FC,$C3,$8B
	.dc.b	$FA,$3C,$63,$3D,$FE,$C5,$E3,$BF,$61,$DB,$F0,$4E,$BA,$74,$7B,$E5
	.dc.b	$E2,$F2,$3F,$D7,$F5,$E4,$F2,$76,$A3,$2F,$F9,$AF,$5F,$65,$71,$E2
	.dc.b	$FF,$32,$4F,$C3,$E8,$C3,$EE,$F2,$7D,$3D,$98,$84,$7A,$5E,$DE,$37
	.dc.b	$F2,$7F,$07,$71,$83,$E3,$BB,$CD,$3F,$E8,$19,$FF,$39,$C7,$FB,$1D
	.dc.b	$F9,$79,$FF,$BD,$63,$0F,$C0,$8F,$D0,$76,$7F,$B5,$3A,$FF,$0C,$FE
	.dc.b	$5F,$4F,$F1,$1F,$5F,$F7,$27,$F6,$C7,$FB,$D3,$C2,$FD,$F3,$AF,$2B
	.dc.b	$FD,$AF,$77,$0F,$F6,$B4,$FF,$BE,$BC,$BF,$F6,$DF,$83,$D1,$FF,$6E
	.dc.b	$F8,$5D,$BC,$FF,$CA,$57,$FB,$C3,$CF,$FC,$2E,$8F,$4F,$32,$F0,$3F
	.dc.b	$0B,$E9,$7F,$93,$63,$F8,$CF,$7F,$F8,$2B,$F5,$FC,$79,$3F,$3F,$96
	.dc.b	$F0,$7D,$BD,$B9,$CF,$6E,$4F,$D9,$4F,$FA,$73,$5F,$D9,$65,$FE,$F7
	.dc.b	$8C,$72,$7F,$1D,$69,$FE,$3E,$3F,$43,$3F,$F0,$05,$FF,$73,$18,$E4
	.dc.b	$F8,$2E,$B8,$BE,$0B,$4F,$F3,$6F,$37,$87,$1F,$E0,$9E,$17,$FB,$0F
	.dc.b	$0F,$1E,$4F,$1F,$F9,$6C,$63,$9B,$F8,$8B,$6B,$FD,$DA,$F0,$C3,$38
	.dc.b	$7F,$47,$91,$FD,$5F,$1F,$E4,$5A,$7F,$39,$F2,$BF,$C1,$5E,$0F,$D2
	.dc.b	$8C,$75,$7F,$41,$CB,$F0,$9A,$FE,$4A,$F7,$87,$23,$FC,$F2,$FF,$84
	.dc.b	$FB,$F9,$47,$FA,$2A,$37,$5C,$AB,$FB,$A4,$FF,$33,$E0,$FF,$04,$47
	.dc.b	$F2,$0C,$FF,$C5,$F5,$F3,$63,$1E,$1F,$E1,$AB,$F1,$B5,$FD,$D2,$B5
	.dc.b	$41,$BF,$C4,$AE,$B1,$9D,$FE,$74,$62,$3F,$5B,$7F,$1B,$0F,$C9,$F0
	.dc.b	$FE,$63,$3F,$E3,$79,$FF,$49,$8C,$6B,$FC,$66,$30,$7D,$C9,$DE,$8A
	.dc.b	$ED,$9A,$7E,$64,$7F,$35,$76,$DB,$E9,$5F,$82,$E1,$FE,$A8,$D8,$F9
	.dc.b	$D8,$7E,$F0,$FE,$D4,$FC,$EE,$1F,$EC,$EE,$A7,$F0,$8E,$67,$FD,$52
	.dc.b	$FD,$DE,$0F,$EF,$5A,$7E,$45,$AF,$C3,$9F,$BC,$7F,$85,$1F,$AB,$1F
	.dc.b	$8D,$7F,$19,$5F,$75,$E5,$3F,$EC,$A7,$E9,$3F,$EC,$CC,$3F,$07,$B7
	.dc.b	$D1,$7C,$CF,$F0,$8C,$3E,$4B,$5F,$86,$FC,$A7,$33,$FA,$A7,$E5,$BF
	.dc.b	$ED,$8E,$3F,$D5,$67,$FC,$43,$0F,$56,$3F,$4E,$BF,$63,$1F,$89,$9F
	.dc.b	$AA,$E9,$FE,$D0,$E6,$3F,$DC,$7B,$FE,$BB,$4F,$EF,$4F,$F2,$87,$FC
	.dc.b	$2A,$FF,$2A,$D3,$FC,$49,$E5,$5A,$FF,$6E,$BF,$C9,$DF,$CB,$4F,$E5
	.dc.b	$F5,$C3,$8E,$1F,$E5,$13,$EB,$AF,$E5,$8F,$F0,$87,$FC,$C7,$9F,$F0
	.dc.b	$6F,$2B,$FC,$1F,$3F,$83,$3A,$EB,$89,$FD,$CB,$FE,$E3,$7F,$77,$C3
	.dc.b	$78,$3F,$45,$7F,$C8,$DF,$E2,$F7,$FE,$BA,$AF,$90,$FA,$32,$61,$FE
	.dc.b	$B7,$F0,$BF,$4F,$3F,$C7,$3C,$9F,$E4,$BB,$FF,$88,$F8,$3F,$8E,$35
	.dc.b	$FA,$6F,$E3,$9F,$DF,$B7,$93,$F9,$B5,$7E,$26,$7F,$7D,$3F,$E7,$09
	.dc.b	$FE,$1B,$3E,$EE,$3F,$E1,$BC,$5F,$E5,$37,$F4,$0F,$F9,$B5,$73,$70
	.dc.b	$D3,$F8,$14,$6A,$BF,$C2,$B1,$F5,$23,$9C,$62,$3F,$86,$47,$E5,$A3
	.dc.b	$10,$DF,$F8,$5C,$61,$FC,$FB,$F1,$F5,$CB,$73,$FC,$9E,$FF,$8E,$7F
	.dc.b	$8F,$B7,$3F,$97,$3C,$71,$FE,$0A,$FE,$75,$FC,$FB,$F4,$F8,$3F,$96
	.dc.b	$F2,$7E,$96,$7F,$84,$BF,$E4,$17,$8C,$FD,$9C,$FF,$82,$6D,$FE,$57
	.dc.b	$9F,$E0,$93,$F9,$4D,$3F,$D6,$FC,$DF,$4A,$3E,$74,$FF,$1C,$AF,$9C
	.dc.b	$FF,$45,$7F,$D5,$5B,$6F,$0F,$95,$3F,$C3,$6B,$E5,$6D,$FE,$DD,$E2
	.dc.b	$FC,$9F,$83,$FD,$63,$87,$F9,$AE,$D9,$6C,$66,$FE,$7B,$FC,$31,$FC
	.dc.b	$93,$F3,$B1,$F4,$63,$FC,$12,$7F,$2D,$AF,$D7,$7E,$54,$7E,$09,$FE
	.dc.b	$A0,$7C,$B7,$F8,$6B,$FB,$E7,$94,$FE,$7D,$F9,$EF,$EB,$2B,$F8,$24
	.dc.b	$FD,$D7,$F8,$43,$FD,$F1,$CC,$7E,$FB,$04,$6A,$7E,$93,$A9,$FE,$06
	.dc.b	$FF,$16,$C3,$FD,$D5,$C5,$FE,$EB,$E7,$FC,$9D,$A3,$18,$FF,$75,$F3
	.dc.b	$F7,$3F,$E6,$ED,$FF,$C1,$51,$88,$FF,$64,$F2,$78,$6F,$FD,$ED,$1F
	.dc.b	$D4,$EF,$E2,$6D,$FB,$E9,$CE,$3F,$E0,$1F,$4B,$FC,$A1,$97,$F6,$F6
	.dc.b	$E3,$FD,$03,$3E,$CF,$D5,$BF,$F3,$FE,$5F,$E1,$8C,$FC,$58,$FF,$80
	.dc.b	$9B,$9F,$E6,$AB,$C7,$7F,$31,$AF,$F4,$CC,$7F,$CC,$17,$FD,$B1,$B5
	.dc.b	$FD,$9B,$1F,$73,$F1,$AF,$C2,$FF,$61,$C3,$CB,$87,$F8,$1D,$AF,$F5
	.dc.b	$8E,$3F,$F3,$64,$FE,$9A,$BE,$2B,$F0,$72,$FF,$4C,$C7,$93,$C5,$FF
	.dc.b	$2B,$E5,$FE,$73,$C3,$EC,$56,$8F,$F3,$CB,$FE,$15,$CB,$E9,$6D,$FE
	.dc.b	$36,$CF,$EA,$7A,$3F,$0E,$F8,$07,$FB,$C6,$9F,$E7,$FA,$F8,$93,$FE
	.dc.b	$87,$C3,$E2,$3C,$0E,$1A,$7D,$AC,$BF,$D0,$D8,$7E,$C6,$7F,$10,$3C
	.dc.b	$EF,$C0,$57,$E0,$33,$FC,$4C,$7F,$8D,$38,$1A,$7E,$27,$C7,$FC,$5E
	.dc.b	$1E,$67,$FC,$9F,$AF,$E3,$6F,$F1,$5C,$DF,$C6,$38,$3F,$03,$1F,$8D
	.dc.b	$E7,$F8,$F8,$79,$F0,$FC,$53,$77,$FC,$53,$F2,$6B,$FC,$BB,$A7,$92
	.dc.b	$7E,$44,$FC,$6D,$7E,$43,$EC,$3F,$8B,$9F,$E3,$23,$15,$FE,$02,$7E
	.dc.b	$6B,$FE,$66,$DF,$EF,$CF,$C7,$9F,$94,$FC,$7C,$BF,$DA,$F4,$EB,$FF
	.dc.b	$52,$ED,$FA,$BE,$FF,$EE,$B8,$76,$9F,$EE,$78,$FF,$AD,$67,$FD,$51
	.dc.b	$3E,$BC,$3F,$D1,$D7,$FB,$3F,$07,$FD,$21,$C1,$F2,$67,$FD,$17,$87
	.dc.b	$FA,$F6,$7F,$D1,$BA,$FD,$4D,$3F,$39,$8F,$D9,$7F,$38,$FE,$23,$8F
	.dc.b	$EA,$BF,$0D,$FA,$AF,$C0,$8F,$CE,$3E,$C7,$F3,$8F,$C9,$3F,$8A,$ED
	.dc.b	$FE,$70,$C3,$F2,$0F,$D4,$8F,$C4,$ED,$FC,$D5,$FA,$D7,$F5,$1F,$A5
	.dc.b	$E4,$7D,$4C,$FF,$80,$BF,$9A,$8F,$A3,$3F,$9A,$7F,$24,$FF,$8D,$30
	.dc.b	$FF,$6E,$4F,$EF,$67,$F3,$8F,$D3,$7F,$31,$D1,$F4,$9F,$D3,$BF,$4A
	.dc.b	$FF,$1E,$FD,$28,$FD,$23,$F4,$9F,$D0,$F9,$FF,$E2,$3D,$FF,$3F,$1F
	.dc.b	$48,$FD,$8C,$7F,$B0,$F0,$FC,$E4,$7F,$AF,$BD,$6F,$F7,$82,$1E,$B7
	.dc.b	$F8,$4B,$1C,$5F,$EB,$39,$B1,$48,$FD,$66,$3F,$AE,$EB,$FA,$F7,$F8
	.dc.b	$E7,$EB,$D7,$E6,$EB,$F5,$6F,$E6,$E7,$F9,$5C,$7E,$AE,$1D,$0C,$FF
	.dc.b	$65,$7F,$AC,$9F,$D0,$BF,$C5,$1F,$83,$C1,$FC,$D5,$F2,$3F,$67,$0F
	.dc.b	$CF,$F5,$FD,$99,$FE,$0C,$FD,$97,$C9,$3F,$67,$1F,$D5,$47,$EB,$27
	.dc.b	$EC,$3F,$B0,$7E,$C4,$7E,$A7,$6F,$F2,$BE,$7F,$C3,$BD,$3F,$B1,$7F
	.dc.b	$23,$A3,$EC,$3F,$B8,$8F,$B1,$97,$CC,$3F,$53,$1F,$C1,$75,$FB,$0F
	.dc.b	$C7,$8F,$D4,$BF,$BE,$8F,$D4,$BF,$CC,$F1,$FF,$54,$F2,$FF,$66,$7F
	.dc.b	$83,$CF,$F1,$E9,$FE,$21,$A7,$D4,$7F,$93,$3F,$52,$FE,$8F,$2F,$EA
	.dc.b	$A7,$EE,$3F,$AA,$8F,$E3,$F3,$F4,$5A,$FE,$A5,$FC,$2F,$27,$D1,$C3
	.dc.b	$F5,$F3,$FE,$A8,$D3,$F9,$8F,$5F,$D1,$E6,$FE,$77,$8F,$F8,$EB,$2F
	.dc.b	$CA,$E3,$F4,$4F,$D4,$3F,$C7,$2B,$E8,$E5,$FC,$C1,$FA,$3B,$FE,$9E
	.dc.b	$7F,$73,$BF,$D0,$7F,$29,$1F,$42,$BF,$9B,$CF,$D0,$97,$A7,$FA,$77
	.dc.b	$F8,$D1,$FA,$1F,$0F,$F9,$F3,$FA,$38,$FD,$43,$FC,$BA,$B4,$FE,$5B
	.dc.b	$CA,$FD,$4A,$F5,$BF,$55,$C2,$F5,$3F,$5B,$93,$C9,$CF,$2F,$EC,$0C
	.dc.b	$61,$FA,$0A,$FD,$9E,$5F,$E9,$59,$FD,$AE,$1F,$E9,$79,$FD,$BE,$BC
	.dc.b	$21,$3D,$B0,$F0,$D7,$F8,$97,$0F,$EC,$87,$33,$FB,$61,$FE,$2F,$D7
	.dc.b	$6F,$F6,$FC,$BF,$D6,$37,$C4,$FF,$59,$D7,$3C,$BE,$0E,$3E,$0B,$FE
	.dc.b	$EC,$E2,$7C,$19,$72,$9E,$8F,$9E,$3F,$88,$C7,$8C,$FE,$6D,$5C,$5F
	.dc.b	$F0,$9F,$AF,$FE,$7B,$CF,$E3,$C7,$F8,$5B,$D9,$C3,$F1,$CF,$F0,$CF
	.dc.b	$67,$F9,$01,$E9,$6D,$70,$7B,$A3,$FC,$9D,$D7,$CB,$4F,$54,$7F,$94
	.dc.b	$72,$C6,$7F,$E5,$6D,$F8,$CF,$F1,$4D,$BF,$CB,$39,$E3,$0E,$3E,$0F
	.dc.b	$FA,$47,$7F,$A9,$5F,$E4,$6F,$03,$C9,$87,$8C,$7E,$29,$FD,$C5,$FF
	.dc.b	$12,$EF,$FE,$5E,$8C,$47,$F8,$06,$3F,$C9,$FF,$45,$FC,$D6,$5F,$B3
	.dc.b	$7F,$94,$B1,$F0,$71,$E7,$3C,$47,$C6,$A4,$62,$3F,$DB,$DE,$E7,$3C
	.dc.b	$FF,$33,$9F,$5F,$C6,$B5,$5F,$EF,$AA,$C4,$FF,$B1,$23,$FD,$FB,$8F
	.dc.b	$8F,$DA,$E3,$3E,$C7,$F4,$7C,$B9,$8E,$14,$1F,$A5,$E6,$3E,$44,$30
	.dc.b	$C3,$EC,$8F,$8F,$FB,$8F,$6D,$4F,$68,$F9,$7F,$A5,$7B,$7C,$EF,$2C
	.dc.b	$60,$F4,$1F,$E8,$4A,$4F,$BA,$FC,$9A,$7C,$DF,$5B,$B1,$F9,$FB,$FF
	.dc.b	$3F,$DA,$3F,$59,$D9,$C2,$7E,$DC,$7C,$DF,$47,$C3,$5B,$9F,$F7,$5E
	.dc.b	$5F,$95,$8C,$47,$C0,$C3,$EE,$7A,$FE,$D3,$F4,$B1,$83,$FB,$FE,$5F
	.dc.b	$C1,$A9,$F7,$7F,$6E,$7F,$D2,$77,$F8,$58,$C4,$7F,$00,$3F,$1E,$FB
	.dc.b	$EB,$97,$E8,$3F,$2C,$FF,$55,$74,$7E,$45,$B5,$FE,$03,$E8,$FC,$84
	.dc.b	$7F,$B6,$A3,$F9,$26,$18,$7F,$C1,$17,$F7,$73,$FF,$52,$FC,$3F,$C8
	.dc.b	$BF,$05,$FE,$67,$8F,$ED,$1F,$E1,$6C,$3F,$C1,$46,$4F,$F3,$31,$FE
	.dc.b	$56,$8F,$D0,$5F,$FA,$C3,$ED,$1F,$A0,$7E,$2B,$F1,$07,$DB,$A5,$F0
	.dc.b	$7C,$31,$FF,$0E,$47,$0C,$97,$C2,$1F,$83,$EE,$9F,$AB,$C2,$1A,$7F
	.dc.b	$73,$9F,$FB,$0A,$32,$70,$FB,$4E,$5E,$70,$84,$F5,$B2,$E0,$D6,$9E
	.dc.b	$1F,$0B,$F4,$1C,$EC,$7F,$85,$65,$FC,$28,$F0,$9F,$E3,$7F,$B0,$7D
	.dc.b	$6B,$EC,$51,$85,$F4,$FE,$D8,$C1,$9F,$CC,$FF,$C8,$1F,$61,$DF,$0E
	.dc.b	$98,$73,$AE,$57,$AC,$7F,$DC,$31,$CA,$7C,$18,$66,$FF,$C9,$59,$FE
	.dc.b	$E7,$1E,$90,$F2,$3D,$4A,$33,$B7,$C3,$3E,$46,$FC,$E7,$11,$F1,$9E
	.dc.b	$19,$FD,$1E,$1F,$A3,$C5,$B7,$E9,$8E,$2E,$5C,$7F,$81,$E4,$FC,$07
	.dc.b	$26,$39,$FF,$15,$CF,$F8,$9C,$7E,$1F,$3F,$E5,$7D,$9F,$F7,$FF,$07
	.dc.b	$F1,$8D,$BF,$2A,$7F,$B3,$FA,$7E,$4F,$57,$E7,$A3,$D9,$7F,$C9,$29
	.dc.b	$3F,$A9,$9F,$F0,$8D,$70,$7F,$53,$D9,$F1,$BB,$3B,$D7,$EC,$A7,$FD
	.dc.b	$BB,$7C,$9F,$E8,$E7,$ED,$4F,$95,$DF,$E1,$8D,$47,$FA,$84,$FF,$3E
	.dc.b	$1F,$7A,$3F,$82,$BF,$E1,$08,$FE,$16,$F8,$AF,$FD,$BA,$FE,$13,$C0
	.dc.b	$FE,$26,$FD,$77,$FB,$EC,$71,$F0,$BC,$11,$EF,$7F,$91,$47,$F2,$5A
	.dc.b	$FE,$3E,$FF,$41,$3F,$07,$5F,$90,$9F,$E5,$CD,$3F,$C0,$F9,$7E,$53
	.dc.b	$C4,$F6,$4F,$CB,$7F,$9F,$57,$57,$FA,$BD,$7B,$8F,$C6,$78,$BF,$D1
	.dc.b	$70,$FE,$8F,$1F,$E2,$03,$ED,$F9,$1F,$1C,$FE,$49,$B7,$E8,$9F,$E4
	.dc.b	$DE,$3D,$9E,$05,$3E,$61,$FD,$EE,$9E,$4F,$D1,$38,$C7,$E7,$7B,$BE
	.dc.b	$A7,$77,$E0,$DC,$D7,$F7,$A9,$FE,$CB,$A7,$D3,$F2,$3F,$01,$E3,$FE
	.dc.b	$D1,$FC,$17,$8F,$FC,$F6,$BE,$B1,$F7,$33,$FE,$64,$FC,$DE,$FF,$BF
	.dc.b	$EA,$7F,$6B,$8F,$D7,$73,$7E,$46,$31,$9F,$E0,$27,$EA,$67,$F2,$70
	.dc.b	$FF,$61,$7E,$32,$F9,$2F,$C3,$F9,$98,$7F,$13,$74,$F1,$9F,$EC,$9F
	.dc.b	$43,$FC,$5D,$87,$DF,$8F,$C8,$D7,$F8,$E9,$EE,$7F,$8D,$FD,$0F,$F1
	.dc.b	$8C,$FF,$49,$3F,$BC,$B8,$7F,$C1,$11,$C8,$FF,$62,$FD,$CF,$F8,$7B
	.dc.b	$6F,$4F,$57,$F9,$C2,$3C,$B3,$A3,$CA,$7F,$8C,$6D,$1F,$A2,$C7,$FC
	.dc.b	$EF,$FA,$EF,$3C,$7F,$C4,$17,$FE,$D2,$8C,$47,$F7,$FB,$47,$66,$CF
	.dc.b	$6C,$B9,$BE,$0B,$AF,$C0,$3F,$DD,$DF,$33,$FE,$86,$FD,$CB,$C1,$1F
	.dc.b	$99,$B4,$62,$3D,$0F,$73,$E9,$BF,$EE,$67,$FC,$79,$EE,$FC,$B7,$FC
	.dc.b	$09,$97,$9F,$F7,$3F,$8D,$B7,$CF,$6D,$FF,$9E,$9F,$51,$D3,$03,$FE
	.dc.b	$47,$F0,$F8,$4F,$F0,$5C,$7E,$AF,$EE,$DF,$C6,$E7,$F8,$6D,$A3,$1A
	.dc.b	$FF,$A8,$9F,$8C,$75,$FC,$5B,$FC,$8F,$27,$FC,$74,$DC,$3F,$91,$FA
	.dc.b	$DF,$C6,$3E,$83,$AD,$7F,$97,$23,$18,$7F,$AB,$A3,$E9,$1E,$9F,$A4
	.dc.b	$FF,$0C,$E7,$E4,$6C,$FF,$AD,$3F,$30,$FB,$7B,$FF,$AA,$5F,$81,$CD
	.dc.b	$FE,$0D,$F3,$3F,$CA,$BE,$67,$F8,$2D,$FE,$81,$B7,$E8,$F6,$FF,$05
	.dc.b	$C6,$BC,$FF,$F0,$9F,$5F,$FB,$03,$0F,$EA,$9E,$8F,$FA,$03,$0E,$18
	.dc.b	$7F,$82,$8F,$CD,$38,$F4,$BF,$C0,$B1,$FA,$3A,$FE,$6F,$1F,$E0,$9F
	.dc.b	$CF,$BF,$3B,$2A,$E1,$3F,$E0,$D8,$E7,$5F,$CF,$BC,$5F,$E5,$D5,$CE
	.dc.b	$7F,$B9,$ED,$FE,$71,$8D,$C7,$F7,$EF,$67,$CD,$7F,$DF,$3D,$77,$CD
	.dc.b	$78,$2F,$D3,$EC,$FF,$82,$F0,$F4,$F9,$1E,$E7,$F7,$4D,$7F,$D8,$AF
	.dc.b	$8F,$2F,$04,$7E,$76,$31,$1E,$B8,$D9,$FD,$8B,$B7,$C1,$E4,$EB,$0F
	.dc.b	$E0,$CD,$9F,$F6,$8F,$CF,$D7,$F2,$2D,$E7,$FF,$78,$FC,$AF,$ED,$4F
	.dc.b	$CE,$76,$7E,$12,$BF,$DE,$AF,$E1,$17,$D0,$BF,$25,$FE,$12,$D8,$FD
	.dc.b	$E0,$FD,$63,$0F,$EE,$CF,$ED,$AE,$35,$FE,$41,$03,$FD,$81,$E7,$B0
	.dc.b	$F1,$47,$F8,$8D,$F1,$43,$1F,$E3,$2D,$9F,$0A,$CF,$E0,$47,$58,$C3
	.dc.b	$F7,$AF,$E6,$3E,$0D,$36,$FF,$8C,$5F,$8C,$FF,$8C,$F7,$FD,$1F,$95
	.dc.b	$CD,$DD,$FF,$B4,$B9,$BF,$17,$5E,$BD,$F4,$FB,$7A,$7F,$CC,$CF,$FB
	.dc.b	$C6,$BF,$DB,$31,$88,$E0,$7F,$90,$79,$BF,$1D,$CF,$FE,$E4,$AF,$C6
	.dc.b	$F3,$FF,$96,$6B,$F9,$F4,$63,$9F,$F4,$2F,$EA,$B2,$F1,$CF,$FB,$B7
	.dc.b	$0F,$5C,$FE,$93,$A3,$A4,$7E,$A9,$FD,$34,$F8,$A7,$F5,$3D,$3D,$2B
	.dc.b	$F3,$13,$CC,$F6,$67,$FE,$82,$7C,$71,$FC,$17,$D2,$FF,$44,$76,$FF
	.dc.b	$90,$F0,$FD,$EF,$6F,$FA,$37,$BB,$F8,$47,$77,$D5,$EF,$FE,$6F,$47
	.dc.b	$E7,$E5,$FE,$63,$F0,$3F,$07,$27,$F0,$7E,$17,$E0,$8E,$27,$F2,$19
	.dc.b	$FC,$67,$87,$FE,$3C,$AF,$E5,$11,$FE,$2D,$EC,$FE,$4A,$FF,$9B,$EB
	.dc.b	$FD,$EB,$3A,$E9,$FF,$13,$4F,$C3,$F1,$3F,$1E,$79,$8F,$E7,$8F,$E3
	.dc.b	$5F,$E7,$6F,$F9,$E7,$C6,$FE,$88,$FF,$8B,$EB,$FA,$3B,$FD,$17,$B3
	.dc.b	$FD,$C9,$18,$8F,$D3,$1F,$97,$9F,$86,$7F,$52,$7D,$4D,$EA,$FC,$E7
	.dc.b	$8F,$FD,$72,$7B,$3F,$D7,$2D,$DB,$A9,$FE,$FA,$53,$F6,$95,$67,$FC
	.dc.b	$19,$1C,$8F,$E1,$CF,$E3,$BC,$AF,$F0,$14,$7F,$A5,$5F,$F3,$2F,$93
	.dc.b	$FA,$D7,$F6,$67,$DA,$3F,$A4,$F1,$66,$79,$D7,$C0,$F2,$FE,$F9,$D3
	.dc.b	$4F,$F0,$7C,$F9,$8F,$F5,$14,$7E,$0A,$3F,$C3,$D3,$FA,$3F,$82,$FC
	.dc.b	$39,$9F,$84,$FE,$06,$7F,$80,$3F,$48,$FE,$97,$3F,$C3,$B0,$69,$FC
	.dc.b	$3E,$73,$3F,$E2,$89,$FF,$1A,$08,$CC,$7F,$55,$E0,$FF,$4A,$79,$9F
	.dc.b	$32,$BF,$A7,$71,$FE,$3D,$7D,$26,$7C,$BF,$EB,$7D,$FF,$EE,$6F,$77
	.dc.b	$FA,$37,$81,$FC,$F3,$7F,$F1,$CF,$8F,$FE,$B5,$9F,$E9,$4E,$75,$FF
	.dc.b	$15,$74,$D7,$A7,$91,$7A,$3F,$9A,$EB,$FE,$9E,$3C,$9C,$9F,$CA,$5C
	.dc.b	$47,$F2,$19,$E1,$1F,$C7,$4F,$E4,$59,$7F,$A3,$E7,$F8,$BC,$3C,$5F
	.dc.b	$F4,$94,$E6,$3F,$91,$1F,$C0,$33,$FE,$08,$F2,$C7,$F3,$7A,$7A,$DE
	.dc.b	$19,$7E,$C1,$1E,$59,$F8,$F8,$FE,$A5,$FE,$B8,$7F,$43,$E1,$F9,$DE
	.dc.b	$8B,$2F,$F5,$AE,$BF,$EE,$48,$C5,$7F,$A5,$F9,$7F,$A5,$57,$A6,$3E
	.dc.b	$25,$63,$5F,$F4,$2C,$73,$7F,$D9,$1F,$81,$FF,$51,$E2,$7E,$D1,$FE
	.dc.b	$D5,$5F,$84,$7F,$8C,$0F,$F4,$BB,$8B,$FF,$49,$EB,$C7,$EC,$BF,$C6
	.dc.b	$A7,$9A,$3F,$84,$FA,$BD,$7E,$33,$FD,$9B,$C3,$FE,$91,$9F,$F2,$D4
	.dc.b	$62,$3E,$2F,$A9,$FE,$20,$F8,$BF,$F6,$5E,$47,$A5,$B8,$FF,$C4,$BF
	.dc.b	$B8,$F1,$C3,$FD,$A7,$3F,$E2,$18,$FF,$6C,$E9,$ED,$FC,$83,$FD,$4F
	.dc.b	$3F,$E2,$57,$E0,$B4,$FA,$23,$E1,$7F,$CA,$F1,$FE,$A3,$9F,$8D,$F6
	.dc.b	$5F,$EF,$FA,$A7,$FB,$FB,$DB,$FF,$44,$78,$7F,$D0,$7E,$21,$E7,$BD
	.dc.b	$07,$C4,$7C,$FA,$D4,$E7,$E2,$B1,$8F,$F4,$CF,$10,$41,$E1,$6A,$31
	.dc.b	$13,$0C,$7A,$C2,$12,$F2,$E0,$D7,$CE,$F3,$F0,$FF,$9F,$E1,$FE,$76
	.dc.b	$CF,$FC,$63,$C9,$FE,$AE,$7F,$A5,$47,$F4,$4F,$0B,$C5,$EF,$BD,$A7
	.dc.b	$B6,$8C,$2F,$8D,$DC,$EA,$73,$F1,$5F,$EB,$FE,$0F,$59,$FC,$1F,$1F
	.dc.b	$F2,$D7,$8F,$FE,$66,$8F,$85,$C3,$F9,$7E,$EF,$F4,$19,$E4,$AF,$F3
	.dc.b	$53,$E8,$DB,$FD,$64,$E6,$3F,$CD,$7F,$01,$5F,$F5,$53,$B9,$FE,$30
	.dc.b	$E6,$3F,$D8,$5F,$A7,$FF,$A4,$8F,$F4,$89,$F8,$17,$FC,$DC,$FE,$7F
	.dc.b	$BF,$FC,$75,$1F,$95,$EC,$FD,$64,$E6,$3F,$C7,$DD,$7F,$EC,$DF,$DC
	.dc.b	$7F,$D3,$A7,$E7,$23,$FC,$36,$FF,$72,$30,$FF,$84,$E7,$C1,$CA,$7A
	.dc.b	$42,$BF,$C9,$31,$8C,$3F,$91,$BF,$DF,$0E,$73,$FE,$4E,$C7,$DC,$FF
	.dc.b	$91,$5F,$81,$C9,$63,$FC,$8D,$9F,$F1,$AD,$37,$0C,$BF,$C9,$27,$E2
	.dc.b	$2B,$FC,$9D,$3F,$2F,$4B,$5F,$C5,$4E,$1B,$FD,$BD,$FE,$F6,$1F,$C6
	.dc.b	$A7,$F0,$31,$FA,$0D,$FF,$CB,$1C,$BF,$E4,$2E,$2F,$C5,$F2,$62,$FF
	.dc.b	$D2,$91,$F8,$A7,$C9,$3E,$6C,$0F,$B2,$32,$7B,$35,$FF,$08,$C7,$F9
	.dc.b	$07,$9F,$F4,$9D,$1D,$6B,$F4,$D1,$FA,$AE,$8C,$69,$E6,$E9,$AF,$E6
	.dc.b	$35,$E1,$D6,$F7,$5E,$F3,$BB,$F8,$1C,$FF,$02,$CF,$C6,$7F,$B2,$67
	.dc.b	$5D,$DF,$C0,$E3,$E7,$77,$FE,$00,$FB,$FD,$7F,$E8,$6F,$03,$F9,$47
	.dc.b	$89,$F9,$0F,$0B,$FC,$7D,$1F,$C5,$EF,$EB,$3F,$C8,$39,$4F,$F2,$05
	.dc.b	$E1,$FF,$29,$9F,$EA,$07,$FA,$1D,$FE,$52,$FD,$FF,$13,$F9,$8F,$81
	.dc.b	$FE,$79,$6C,$7F,$99,$FC,$5F,$E6,$59,$7F,$9F,$5B,$1F,$E6,$BF,$1F
	.dc.b	$F2,$3E,$07,$FA,$4D,$FC,$9F,$8F,$FE,$95,$F2,$3F,$9D,$76,$FF,$66
	.dc.b	$1E,$B6,$70,$FE,$9E,$7E,$46,$79,$BF,$93,$F2,$B8,$D7,$F6,$EF,$0F
	.dc.b	$FD,$59,$E6,$7F,$63,$A7,$9B,$FD,$28,$CC,$7F,$42,$34,$FF,$4E,$F8
	.dc.b	$7F,$F1,$37,$8D,$DB,$D6,$FC,$B4,$FF,$26,$BF,$ED,$19,$7F,$64,$CF
	.dc.b	$C3,$9F,$E0,$27,$FD,$E5,$C1,$FD,$AF,$93,$FB,$67,$37,$F7,$2F,$23
	.dc.b	$E9,$D7,$B1,$75,$7E,$E7,$83,$FB,$A6,$BF,$DC,$FF,$43,$EB,$1F,$EB
	.dc.b	$B6,$DF,$EC,$E6,$F4,$F2,$8F,$F6,$66,$FF,$71,$FE,$28,$7E,$10,$F9
	.dc.b	$71,$E2,$7E,$29,$C0,$FF,$6A,$70,$FE,$0D,$DE,$BD,$B7,$F7,$F6,$FF
	.dc.b	$B3,$78,$7E,$0F,$27,$FB,$6F,$D1,$FD,$8E,$1F,$C0,$5A,$F6,$3F,$A7
	.dc.b	$F2,$3E,$4B,$FB,$AB,$63,$FC,$79,$E9,$7F,$85,$7A,$BF,$C2,$98,$FF
	.dc.b	$84,$E7,$CC,$7F,$30,$1D,$3F,$E1,$BF,$D9,$BF,$39,$1F,$E0,$73,$07
	.dc.b	$FB,$BF,$A7,$FD,$E3,$DD,$FE,$64,$8F,$F0,$41,$FC,$5F,$A3,$F4,$AF
	.dc.b	$F8,$A8,$C9,$E9,$E4,$FE,$7F,$9F,$F8,$A7,$AB,$F6,$73,$FE,$FB,$EF
	.dc.b	$FE,$87,$7C,$A3,$FC,$40,$7F,$83,$75,$FF,$0A,$3D,$FA,$FF,$A6,$4F
	.dc.b	$D3,$C7,$EC,$78,$1E,$58,$7D,$6F,$E8,$1B,$63,$F8,$57,$89,$FC,$3B
	.dc.b	$C6,$FF,$1E,$57,$F8,$3B,$C5,$79,$78,$BF,$F0,$77,$91,$FE,$0D,$F2
	.dc.b	$71,$1F,$CC,$A3,$F8,$B6,$5D,$A3,$F9,$1F,$97,$FE,$0D,$BD,$B8,$F2
	.dc.b	$F8,$F8,$3F,$A5,$A5,$F1,$8F,$E8,$D1,$8C,$FF,$9E,$7A,$7F,$E0,$8F
	.dc.b	$3F,$C3,$E7,$FF,$82,$0F,$E7,$98,$7F,$A4,$EF,$EF,$9F,$ED,$47,$0F
	.dc.b	$E2,$CF,$EE,$3C,$9F,$A8,$7F,$C6,$85,$3F,$3B,$D1,$FD,$DB,$8F,$E6
	.dc.b	$78,$DF,$D9,$27,$FC,$C4,$FF,$0C,$D0,$FF,$B9,$0C,$78,$DF,$DC,$8F
	.dc.b	$E6,$78,$7D,$19,$F1,$F3,$7C,$16,$DF,$F8,$9C,$FE,$D1,$B7,$C0,$F5
	.dc.b	$3F,$C4,$91,$83,$FA,$DE,$FF,$E2,$AF,$53,$A5,$7F,$5B,$3F,$AD,$5F
	.dc.b	$D3,$E2,$FA,$7C,$7A,$C3,$F0,$51,$F4,$DE,$87,$DB,$E5,$D3,$F7,$79
	.dc.b	$EF,$3F,$F5,$7F,$17,$FA,$8B,$FD,$57,$1F,$C0,$39,$E9,$FE,$B3,$97
	.dc.b	$2D,$FE,$DD,$7E,$5F,$1F,$D7,$1F,$E1,$33,$E2,$75,$7C,$67,$F8,$81
	.dc.b	$F5,$CF,$CC,$70,$FF,$0D,$7F,$4B,$D7,$FE,$2A,$AF,$D4,$77,$7F,$A1
	.dc.b	$27,$F5,$95,$BC,$3F,$9D,$47,$F8,$4F,$B7,$FA,$E1,$F6,$7B,$FF,$80
	.dc.b	$F4,$F3,$6D,$FF,$4B,$AF,$11,$FC,$20,$FE,$94,$7F,$85,$F1,$FE,$08
	.dc.b	$FF,$53,$C6,$79,$3F,$D3,$3C,$4F,$F4,$5F,$97,$FC,$5E,$BF,$82,$1F
	.dc.b	$B3,$F0,$BF,$DD,$DE,$8F,$F4,$93,$F1,$D1,$FE,$BB,$AF,$E8,$59,$FF
	.dc.b	$6E,$EF,$F7,$E4,$F3,$BF,$C9,$7F,$15,$FE,$EB,$F2,$7F,$D6,$9D,$9F
	.dc.b	$E2,$DC,$BF,$5B,$E0,$7F,$47,$7F,$A1,$1F,$CE,$78,$BF,$CA,$67,$FB
	.dc.b	$92,$78,$70,$AF,$FA,$3F,$1F,$F8,$1E,$1A,$FF,$80,$38,$7D,$9C,$9F
	.dc.b	$D8,$F2,$FD,$DF,$27,$FB,$0B,$C8,$FE,$83,$1F,$BE,$BF,$EB,$CF,$FB
	.dc.b	$CF,$CD,$FE,$3A,$FE,$C0,$FF,$6E,$73,$7F,$B3,$FF,$4F,$FF,$67,$7E
	.dc.b	$6D,$FD,$60,$FE,$B2,$FF,$5A,$EA,$F6,$6B,$FC,$A7,$4F,$13,$E6,$C3
	.dc.b	$FA,$CE,$7E,$28,$FA,$FC,$BF,$5C,$FF,$77,$F9,$BF,$D4,$FE,$87,$FC
	.dc.b	$0F,$C4,$7E,$9B,$70,$FE,$1A,$7F,$2B,$F3,$7F,$DD,$7F,$B1,$FF,$54
	.dc.b	$3F,$D8,$FE,$7B,$F7,$2F,$F0,$8A,$FB,$1F,$E6,$97,$FE,$1C,$3F,$90
	.dc.b	$FC,$7F,$F8,$71,$FC,$F9,$83,$FA,$96,$9F,$E1,$F7,$C0,$7F,$28,$FD
	.dc.b	$53,$F9,$F1,$FE,$85,$E0,$E3,$DE,$8C,$4F,$F4,$E7,$FC,$5E,$FF,$47
	.dc.b	$6E,$9F,$EF,$DF,$A5,$FE,$90,$7F,$8E,$27,$FA,$63,$FE,$EE,$FD,$87
	.dc.b	$F4,$B1,$FC,$E1,$FD,$4B,$FD,$36,$DF,$A5,$7F,$51,$3F,$DD,$2E,$23
	.dc.b	$FB,$33,$5F,$DD,$F5,$F1,$9F,$EC,$B1,$FE,$ED,$3A,$7D,$77,$F5,$B3
	.dc.b	$F8,$35,$1F,$EC,$7F,$45,$FD,$67,$C9,$F1,$3F,$D6,$FC,$DB,$7F,$D6
	.dc.b	$BD,$8F,$F4,$43,$FD,$42,$7F,$D1,$D1,$FE,$E8,$8C,$1F,$C9,$BD,$0F
	.dc.b	$F3,$EF,$B1,$CA,$3F,$AE,$C7,$F6,$67,$66,$3E,$E3,$5D,$9F,$D6,$B4
	.dc.b	$63,$FE,$A4,$6D,$7F,$3A,$F5,$5E,$A7,$5F,$63,$FA,$E7,$AD,$FE,$E6
	.dc.b	$F5,$7E,$67,$81,$E6,$E0,$B8,$FA,$F8,$3D,$19,$BE,$DF,$FD,$34,$D9
	.dc.b	$FE,$A0,$7F,$94,$D7,$EC,$9A,$3C,$9F,$A3,$7F,$10,$83,$FE,$09,$3F
	.dc.b	$C1,$E7,$D0,$F9,$CD,$1F,$D8,$49,$5F,$A4,$3F,$9A,$47,$F3,$BE,$1E
	.dc.b	$71,$FD,$B8,$FF,$62,$47,$F6,$93,$F8,$FC,$79,$23,$CE,$7F,$8D,$FA
	.dc.b	$3F,$B4,$61,$FE,$B7,$C7,$FA,$E9,$FE,$B9,$D3,$E2,$57,$FB,$3D,$FB
	.dc.b	$7E,$D7,$F6,$FE,$3F,$ED,$5C,$56,$BE,$87,$05,$7F,$DA,$C7,$F7,$1C
	.dc.b	$FE,$AB,$83,$FC,$90,$6F,$97,$FB,$44,$69,$FF,$0F,$F2,$E5,$FF,$61
	.dc.b	$F3,$F5,$E2,$FF,$6B,$F1,$D4,$43,$FD,$BB,$9F,$E6,$71,$FF,$74,$3F
	.dc.b	$99,$7F,$DD,$5D,$1F,$2C,$FF,$7E,$9F,$E1,$1E,$9F,$D3,$FB,$1F,$E5
	.dc.b	$5D,$3F,$59,$E0,$7F,$63,$BF,$D5,$1F,$EF,$3A,$FF,$5E,$1F,$EB,$F3
	.dc.b	$FC,$35,$D9,$FB,$CA,$3F,$BC,$9F,$F7,$D7,$6F,$F0,$3E,$6F,$F0,$D3
	.dc.b	$FC,$16,$29,$FD,$8F,$BF,$F8,$6C,$FF,$87,$DF,$E1,$B0,$BF,$E2,$8F
	.dc.b	$F8,$DB,$C0,$FE,$3F,$18,$8F,$F1,$B4,$7F,$9D,$E3,$0F,$F8,$2B,$C1
	.dc.b	$FE,$4D,$3E,$F6,$F8,$3F,$88,$8F,$EE,$EF,$FA,$F3,$4F,$E5,$DE,$7F
	.dc.b	$F3,$19,$FC,$B6,$5F,$EC,$0F,$17,$FD,$8B,$2F,$1B,$FD,$7C,$7F,$82
	.dc.b	$2B,$FD,$3F,$E9,$FF,$41,$7F,$C0,$77,$FE,$6B,$8C,$67,$F8,$69,$FF
	.dc.b	$03,$46,$1F,$F6,$17,$8F,$D5,$FE,$C0,$1F,$E8,$CF,$FB,$61,$65,$FE
	.dc.b	$A6,$8F,$E8,$47,$F5,$73,$FD,$35,$E1,$7C,$78,$FF,$43,$C7,$F3,$1E
	.dc.b	$1F,$F0,$37,$17,$FB,$F6,$BF,$37,$A7,$F5,$FB,$79,$DF,$C3,$FC,$47
	.dc.b	$8C,$F1,$91,$FE,$6F,$CF,$FC,$EB,$E1,$FF,$82,$F2,$FD,$B4,$32,$E9
	.dc.b	$8F,$F4,$1F,$17,$FB,$99,$FD,$D3,$C5,$FF,$05,$E1,$FE,$0C,$EC,$FF
	.dc.b	$2E,$7B,$BF,$B1,$C3,$8E,$3F,$95,$3F,$AF,$FA,$DF,$E1,$CC,$BF,$CD
	.dc.b	$1C,$3F,$EB,$69,$F4,$9F,$8B,$EB,$FE,$B6,$FF,$9B,$7A,$BF,$0A,$FE
	.dc.b	$04,$FE,$D3,$C5,$FE,$35,$F3,$3F,$01,$18,$8F,$9E,$74,$CB,$EB,$47
	.dc.b	$F8,$D8,$7A,$3F,$ED,$7D,$7F,$DD,$7C,$CF,$DD,$AF,$67,$D7,$8F,$F1
	.dc.b	$A5,$F5,$C7,$FC,$1A,$FD,$F7,$FC,$47,$C8,$D3,$FD,$69,$3F,$E4,$9E
	.dc.b	$0F,$1F,$07,$F6,$B9,$FF,$7C,$D7,$E8,$AB,$7F,$17,$F8,$6A,$FE,$E1
	.dc.b	$C1,$CB,$1F,$5F,$CA,$72,$BF,$F0,$31,$E3,$8F,$93,$83,$C1,$F0,$F6
	.dc.b	$FF,$B9,$67,$A5,$7F,$7D,$3E,$49,$E6,$E3,$FC,$73,$CF,$2F,$EF,$0F
	.dc.b	$E7,$30,$FE,$23,$5F,$0A,$1E,$CA,$F7,$1F,$93,$9F,$1F,$89,$FE,$55
	.dc.b	$BF,$E6,$3E,$F7,$FA,$2F,$E1,$8F,$F5,$EE,$5F,$EE,$FF,$47,$FD,$CF
	.dc.b	$CB,$FF,$67,$F9,$5F,$9B,$F0,$BF,$35,$E0,$FE,$73,$E9,$3F,$B3,$ED
	.dc.b	$FE,$D7,$C2,$ED,$FA,$3D,$FF,$C3,$1F,$B8,$7C,$EF,$B7,$F9,$53,$FE
	.dc.b	$45,$E3,$FF,$5C,$46,$23,$FC,$FB,$87,$F1,$BA,$C4,$7C,$FF,$2F,$F8
	.dc.b	$33,$D6,$7F,$AA,$1F,$DD,$F1,$FE,$A1,$E3,$7F,$7E,$D5,$A7,$F4,$9D
	.dc.b	$C4,$3F,$D5,$DC,$1F,$03,$0F,$E8,$A7,$FB,$EF,$87,$E3,$1F,$E1,$9C
	.dc.b	$FF,$C4,$91,$FE,$7D,$3F,$CD,$67,$EB,$4F,$A1,$B7,$FB,$4F,$C7,$FF
	.dc.b	$17,$F3,$7D,$59,$D7,$3F,$F9,$AB,$C3,$F6,$F0,$7E,$13,$1F,$E8,$B5
	.dc.b	$F8,$40,$71,$7F,$18,$FF,$BC,$A3,$87,$5F,$27,$F7,$33,$FC,$65,$FE
	.dc.b	$3F,$D7,$F2,$67,$F2,$75,$F9,$E9,$DF,$6F,$1D,$3F,$7D,$3F,$BC,$EE
	.dc.b	$FC,$F7,$87,$FC,$A2,$3F,$3D,$E8,$79,$23,$F2,$58,$7D,$37,$F9,$7C
	.dc.b	$FE,$47,$A7,$5E,$1F,$58,$79,$3F,$E0,$2B,$FE,$2B,$EB,$FF,$81,$38
	.dc.b	$7F,$9F,$79,$7F,$C0,$EF,$FC,$41,$E6,$FC,$2F,$19,$F9,$37,$FA,$08
	.dc.b	$FE,$95,$79,$7F,$A6,$77,$6F,$4F,$F4,$29,$FE,$83,$AF,$2D,$7F,$4B
	.dc.b	$BE,$9C,$1F,$E8,$7F,$11,$F8,$87,$F6,$0C,$D1,$87,$F8,$E1,$F5,$3E
	.dc.b	$63,$F8,$53,$FD,$80,$FE,$97,$87,$F1,$04,$FF,$A2,$39,$7F,$D0,$F8
	.dc.b	$F0,$9F,$E2,$87,$FA,$42,$FF,$8C,$3F,$44,$EC,$7F,$0E,$39,$9F,$84
	.dc.b	$F8,$2F,$E3,$BB,$FF,$16,$8F,$C3,$E1,$FE,$D0,$7A,$7E,$A5,$FE,$F2
	.dc.b	$C3,$FD,$14,$FF,$A3,$B0,$FC,$7C,$FF,$70,$3F,$8A,$1F,$CD,$7E,$1B
	.dc.b	$F8,$9B,$FC,$DF,$82,$F1,$FF,$88,$7D,$4F,$F7,$C5,$F2,$8C,$67,$FC
	.dc.b	$4B,$3F,$36,$1F,$9A,$F8,$94,$E6,$1A,$FF,$36,$9E,$7F,$77,$F9,$69
	.dc.b	$FF,$2B,$1F,$49,$FA,$C7,$F8,$EF,$13,$FC,$FB,$C2,$FD,$BF,$BD,$FE
	.dc.b	$A7,$8F,$F2,$CC,$EA,$BF,$CF,$9E,$CE,$5F,$F4,$97,$8D,$FC,$0C,$7B
	.dc.b	$59,$8F,$F4,$9E,$17,$C3,$FF,$48,$63,$FD,$31,$FC,$54,$F9,$4F,$F3
	.dc.b	$09,$4F,$E4,$FE,$4C,$EF,$FF,$4F,$7B,$5F,$E5,$53,$FC,$AD,$9F,$FA
	.dc.b	$72,$BF,$CC,$7C,$FF,$CB,$30,$EB,$AF,$FA,$3E,$7F,$CD,$07,$F5,$C9
	.dc.b	$FE,$5A,$FE,$46,$1E,$67,$E6,$CF,$F5,$D9,$FC,$03,$87,$E7,$4F,$03
	.dc.b	$F1,$6F,$D1,$AF,$F6,$77,$9B,$F4,$67,$F1,$A7,$E8,$9F,$F7,$66,$07
	.dc.b	$51,$FE,$4E,$73,$7F,$E9,$53,$C1,$1F,$81,$AF,$E3,$75,$E6,$8F,$F4
	.dc.b	$C3,$5F,$C9,$37,$93,$E6,$BF,$04,$BE,$EF,$D1,$9F,$8E,$7F,$38,$3E
	.dc.b	$29,$FE,$FA,$8F,$F4,$7E,$9F,$3C,$FF,$42,$E9,$E2,$DF,$FC,$0B,$E7
	.dc.b	$7F,$47,$F1,$B1,$87,$FB,$27,$82,$BD,$06,$98,$CB,$F8,$BC,$7F,$9C
	.dc.b	$70,$FF,$07,$F2,$7F,$A6,$23,$07,$E3,$9F,$EA,$1C,$7D,$34,$FF,$4C
	.dc.b	$CE,$8F,$33,$E6,$71,$CB,$FD,$4B,$5F,$1B,$F9,$E8,$E1,$BF,$8B,$2F
	.dc.b	$CF,$47,$F9,$D8,$FE,$9E,$7E,$B8,$FD,$97,$3F,$F5,$57,$FA,$D0,$E7
	.dc.b	$F6,$BF,$E8,$BF,$2B,$FD,$43,$B7,$F9,$1A,$3F,$C9,$27,$F5,$4E,$AF
	.dc.b	$F5,$24,$FF,$9E,$FB,$35,$3F,$C1,$CE,$47,$F2,$5C,$5D,$BF,$C4,$E3
	.dc.b	$FD,$30,$FF,$A9,$7C,$9F,$F5,$14,$7F,$94,$4F,$E3,$67,$F0,$A3,$CB
	.dc.b	$3F,$C9,$9F,$E5,$26,$3B,$FF,$C1,$5E,$17,$F9,$03,$C1,$FE,$B7,$3F
	.dc.b	$EA,$03,$FD,$38,$7F,$A9,$4E,$07,$97,$C2,$FE,$81,$7F,$CF,$E3,$FD
	.dc.b	$08,$27,$FC,$C0,$FF,$32,$3F,$9F,$3F,$D8,$EB,$FD,$47,$EA,$FF,$A9
	.dc.b	$3C,$5F,$F5,$09,$FE,$A6,$F1,$BE,$35,$FF,$9F,$5F,$EA,$ED,$3D,$E7
	.dc.b	$5E,$3F,$F5,$17,$9D,$FA,$E8,$FF,$50,$B7,$93,$FB,$34,$78,$CF,$F7
	.dc.b	$5F,$8F,$FD,$99,$FF,$24,$6B,$FE,$78,$7F,$AD,$BF,$EB,$E9,$FE,$A5
	.dc.b	$7F,$D5,$CF,$F5,$0E,$DF,$E8,$0A,$FF,$52,$F8,$DF,$E7,$AD,$7F,$81
	.dc.b	$F8,$9F,$E6,$79,$FC,$61,$FE,$6E,$7E,$EB,$FE,$19,$F1,$31,$C7,$FE
	.dc.b	$8D,$F2,$FF,$A5,$E5,$FC,$F4,$FE,$97,$8F,$F9,$6F,$9F,$97,$3F,$FA
	.dc.b	$B7,$CE,$FE,$13,$97,$E5,$4F,$F1,$1E,$6F,$C8,$FF,$CF,$3E,$0F,$FB
	.dc.b	$EE,$3F,$43,$DF,$FA,$AF,$7B,$F9,$D7,$85,$FC,$F4,$FF,$20,$F8,$3F
	.dc.b	$CE,$F1,$FF,$66,$1F,$CD,$FC,$EF,$F5,$3E,$1F,$AA,$F1,$BF,$C9,$5E
	.dc.b	$CF,$F4,$37,$FC,$99,$EF,$7F,$29,$E2,$E3,$87,$F8,$1B,$CA,$FF,$37
	.dc.b	$79,$7F,$5E,$3F,$CB,$A7,$DC,$3F,$C0,$87,$F9,$74,$FD,$8F,$3F,$F8
	.dc.b	$E2,$78,$47,$F0,$3C,$3C,$2D,$F5,$3F,$D9,$71,$FE,$CA,$F3,$59,$7F
	.dc.b	$13,$AF,$5F,$89,$FC,$0B,$6F,$F6,$0B,$FC,$13,$83,$E1,$0E,$0F,$E2
	.dc.b	$5C,$5F,$C9,$78,$BF,$86,$FC,$8B,$CD,$7C,$5B,$D8,$FF,$43,$FB,$1F
	.dc.b	$CE,$BA,$7F,$C7,$9A,$3A,$BE,$BF,$9B,$FD,$97,$B3,$FC,$FF,$E0,$FF
	.dc.b	$AC,$BB,$B1,$C7,$FC,$28,$FF,$4C,$F8,$1F,$C2,$3B,$3F,$09,$1F,$C1
	.dc.b	$0F,$C5,$1F,$D2,$3C,$1F,$C5,$3F,$C1,$4F,$D9,$76,$7C,$CD,$3F,$82
	.dc.b	$78,$3F,$EF,$89,$46,$23,$FE,$00,$E5,$FD,$72,$3F,$85,$61,$FA,$1A
	.dc.b	$FE,$F9,$E2,$7E,$9E,$7F,$81,$5F,$FB,$68,$FE,$C7,$D3,$FA,$DA,$FF
	.dc.b	$6D,$7B,$9F,$B6,$7E,$F1,$FE,$41,$F5,$7F,$99,$D7,$F9,$7B,$C4,$FF
	.dc.b	$6D,$79,$3E,$EA,$7E,$2B,$FE,$5D,$FD,$A3,$FD,$BC,$7F,$B8,$BC,$CF
	.dc.b	$F7,$01,$FE,$7D,$EB,$FF,$31,$7B,$36,$DE,$62,$3F,$D0,$5E,$AF,$E9
	.dc.b	$B8,$7F,$91,$EB,$FE,$E6,$F2,$FD,$6F,$C2,$3F,$DA,$6E,$1F,$F7,$D6
	.dc.b	$1F,$E5,$A3,$FD,$0D,$CF,$FE,$57,$8F,$F7,$09,$CF,$CC,$FF,$5C,$63
	.dc.b	$FE,$E3,$BF,$9D,$3C,$9F,$E0,$D7,$9E,$8F,$F5,$F3,$FE,$DF,$3F,$CD
	.dc.b	$AE,$1F,$F4,$FE,$9F,$E5,$67,$FD,$D9,$9F,$F0,$87,$F9,$2F,$1F,$F0
	.dc.b	$8E,$7F,$F7,$EF,$B5,$FE,$AD,$C7,$FC,$A5,$97,$FB,$96,$3F,$35,$7F
	.dc.b	$E5,$C8,$FF,$64,$4F,$E5,$E3,$F6,$3A,$FF,$0C,$EE,$F9,$31,$FE,$80
	.dc.b	$7E,$74,$FF,$6C,$F5,$7F,$D8,$7E,$27,$F9,$17,$C5,$FE,$EF,$EC,$D7
	.dc.b	$FC,$1D,$EA,$FF,$10,$F2,$7F,$88,$78,$B5,$FF,$27,$5F,$F9,$37,$BB
	.dc.b	$C7,$DB,$FE,$12,$F3,$7F,$7A,$7F,$BB,$76,$FF,$25,$79,$BF,$ED,$BF
	.dc.b	$3F,$FC,$93,$A7,$69,$FF,$4A,$FA,$5F,$EC,$65,$8F,$BB,$D8,$C3,$F0
	.dc.b	$3D,$FB,$CB,$F9,$B3,$E4,$7B,$3F,$03,$90,$FF,$AD,$DF,$F1,$96,$1F
	.dc.b	$CF,$BC,$8F,$9D,$F8,$D6,$CF,$E0,$27,$C8,$F2,$0F,$F0,$4F,$53,$F9
	.dc.b	$CC,$FF,$B6,$70,$D1,$FC,$23,$2F,$1E,$5F,$90,$C1,$A6,$20,$4E,$EF
	.dc.b	$F2,$4E,$D9,$8C,$69,$FE,$41,$DF,$FC,$53,$CF,$FE,$38,$E2,$FE,$4C
	.dc.b	$7F,$27,$F7,$71,$C9,$F3,$BD,$5F,$F2,$D6,$1F,$E3,$1E,$4F,$89,$97
	.dc.b	$FB,$BF,$9B,$FC,$77,$D1,$F6,$AB,$FC,$6C,$FF,$2A,$E8,$FE,$53,$AF
	.dc.b	$F9,$EF,$AA,$A2,$6D,$FC,$B3,$C1,$FE,$57,$E8,$BB,$58,$7F,$2B,$CB
	.dc.b	$F3,$E7,$F1,$8E,$EF,$DF,$73,$7E,$5E,$3F,$CE,$E7,$FB,$07,$BF,$FC
	.dc.b	$5F,$BF,$A3,$9A,$BF,$A1,$6F,$83,$FA,$6D,$3F,$97,$5F,$EA,$8F,$E5
	.dc.b	$C7,$FB,$DC,$B5,$F4,$0F,$F6,$D5,$7F,$31,$81,$FC,$8C,$FC,$84,$FF
	.dc.b	$9F,$3D,$AF,$F4,$45,$7D,$8C,$3F,$C6,$2F,$96,$BF,$99,$F8,$DF,$C2
	.dc.b	$23,$EC,$E5,$FC,$23,$C6,$F0,$5F,$F8,$E3,$CE,$FF,$75,$7C,$07,$3C
	.dc.b	$7F,$9D,$F9,$3F,$8F,$3F,$8E,$E2,$D7,$FD,$1B,$18,$C3,$FC,$69,$5E
	.dc.b	$8A,$FB,$BE,$BE,$23,$EF,$C7,$F0,$6E,$9F,$E5,$51,$F4,$9A,$EB,$7F
	.dc.b	$83,$0F,$33,$F8,$3B,$67,$B5,$7E,$B6,$B7,$E6,$E8,$FD,$B4,$7E,$3B
	.dc.b	$1E,$7E,$7F,$F4,$3F,$3F,$E6,$FB,$DF,$D2,$2B,$F6,$D5,$8E,$EF,$E2
	.dc.b	$BE,$8B,$E4,$7F,$D6,$DC,$7F,$D3,$DB,$AF,$FB,$B8,$FE,$8B,$F7,$FF
	.dc.b	$E2,$F9,$FF,$35,$FA,$68,$C4,$BB,$FE,$EC,$BE,$BB,$9C,$E1,$EB,$EE
	.dc.b	$EB,$97,$33,$FD,$49,$FF,$64,$74,$7D,$7B,$FF,$0A,$CF,$F9,$BE,$BA
	.dc.b	$FD,$CE,$B8,$1C,$18,$AF,$F3,$A1,$4F,$F3,$C5,$79,$3F,$22,$E1,$36
	.dc.b	$7F,$C5,$98,$FF,$02,$C5,$5E,$46,$CF,$84,$FE,$CF,$1F,$1B,$E0,$BD
	.dc.b	$F8,$EC,$FF,$33,$57,$F1,$0A,$FB,$71,$5F,$99,$E9,$FF,$81,$DF,$F1
	.dc.b	$E3,$7D,$1F,$F3,$93,$78,$7F,$4F,$A7,$F9,$63,$E9,$7F,$03,$87,$F1
	.dc.b	$CB,$FF,$4E,$5A,$73,$3F,$E6,$CD,$7F,$CB,$FF,$69,$F3,$A8,$FD,$8C
	.dc.b	$73,$97,$FA,$21,$1F,$B6,$E2,$CC,$7F,$9E,$72,$FF,$5C,$38,$C7,$FC
	.dc.b	$47,$1F,$BC,$F6,$3E,$9C,$B1,$FC,$53,$FE,$A9,$3F,$CE,$3A,$7F,$BF
	.dc.b	$31,$FF,$3A,$46,$E3,$1E,$EF,$C6,$FB,$BF,$EA,$9E,$53,$E0,$7F,$C9
	.dc.b	$9F,$03,$FC,$8B,$6F,$F3,$DD,$6A,$73,$3F,$E4,$FF,$3F,$39,$7F,$10
	.dc.b	$F3,$FF,$96,$0F,$DA,$9E,$B4,$FC,$7F,$75,$AA,$D4,$EB,$F3,$9C,$FA
	.dc.b	$82,$7E,$82,$FB,$D9,$F8,$0E,$A7,$31,$9F,$C0,$FA,$FA,$5F,$DB,$F3
	.dc.b	$FF,$26,$9F,$EA,$1E,$1F,$EE,$51,$9F,$3B,$FC,$5B,$18,$99,$CD,$E6
	.dc.b	$F7,$3F,$CF,$5F,$B8,$BE,$B1,$F9,$8B,$97,$5E,$F3,$DE,$5E,$EB,$FA
	.dc.b	$4E,$5F,$5F,$CD,$FF,$01,$73,$7C,$6E,$8F,$8B,$CE,$E8,$FF,$54,$D6
	.dc.b	$AB,$FA,$B5,$6B,$A3,$38,$7F,$3E,$E3,$F7,$1E,$FE,$FF,$E2,$11,$93
	.dc.b	$FC,$17,$E2,$70,$7E,$19,$9E,$3F,$F5,$BF,$D2,$FE,$E2,$3F,$18,$FF
	.dc.b	$0A,$69,$CC,$86,$0E,$67,$F8,$73,$C4,$FC,$AD,$A7,$E6,$4F,$73,$F3
	.dc.b	$1F,$A8,$E1,$E3,$28,$C7,$47,$F9,$81,$D6,$5F,$B6,$AF,$D0,$7E,$6F
	.dc.b	$FA,$F3,$CD,$67,$F3,$B1,$FE,$D2,$7F,$3F,$97,$FB,$53,$E6,$3F,$6C
	.dc.b	$7B,$6C,$FF,$26,$6D,$7F,$13,$7A,$3F,$DB,$3C,$5F,$ED,$BE,$3D,$F1
	.dc.b	$BF,$17,$8F,$6A,$FF,$4D,$BC,$CE,$D9,$FA,$A3,$F2,$2D,$CB,$ED,$3F
	.dc.b	$03,$6F,$DE,$E5,$F2,$6B,$E7,$BF,$7B,$6E,$11,$FE,$A8,$78,$6F,$F8
	.dc.b	$37,$D8,$7D,$B9,$FC,$1C,$FD,$67,$F0,$FC,$7C,$9C,$9F,$A3,$BF,$0C
	.dc.b	$7E,$47,$8F,$F6,$BC,$BF,$95,$BF,$F5,$BC,$E7,$93,$F2,$91,$FE,$BA
	.dc.b	$ED,$FF,$33,$74,$FE,$8B,$B3,$F4,$FD,$9F,$97,$9C,$F5,$FE,$97,$E0
	.dc.b	$71,$87,$B8,$FD,$AC,$F9,$4F,$6B,$D4,$FF,$0E,$F8,$3F,$C1,$47,$87
	.dc.b	$FC,$17,$C2,$78,$7D,$2B,$55,$A9,$FE,$0D,$1C,$CF,$E2,$5A,$FE,$CB
	.dc.b	$BF,$87,$0F,$ED,$78,$7F,$8A,$56,$FC,$7F,$E4,$DE,$CF,$FB,$C3,$C6
	.dc.b	$FF,$78,$79,$1D,$E7,$31,$FE,$4A,$E2,$FD,$B7,$5F,$E5,$30,$FF,$6F
	.dc.b	$BF,$87,$E8,$FE,$73,$DF,$FD,$5E,$7F,$DC,$7E,$57,$FB,$D3,$C0,$FE
	.dc.b	$87,$18,$3F,$C3,$3F,$00,$8F,$E8,$9D,$DF,$C0,$EB,$FA,$85,$6C,$FE
	.dc.b	$9E,$7E,$66,$FF,$A8,$39,$7C,$18,$7F,$5D,$8F,$ED,$77,$F9,$E7,$F8
	.dc.b	$77,$99,$FD,$82,$3F,$D3,$58,$7F,$02,$BF,$F4,$E4,$76,$AC,$1F,$A6
	.dc.b	$8F,$DC,$5F,$B5,$FC,$74,$7F,$78,$EE,$FE,$6C,$FF,$09,$30,$7E,$FF
	.dc.b	$C5,$FE,$F9,$3D,$A3,$89,$DE,$FF,$DF,$F5,$FE,$1E,$AF,$F0,$5C,$FF
	.dc.b	$85,$0F,$F5,$7F,$9F,$FE,$0B,$69,$FE,$1A,$29,$E9,$1E,$87,$F9,$6E
	.dc.b	$79,$E1,$FE,$5E,$BF,$F2,$D5,$7F,$AE,$BD,$1F,$E4,$D4,$7C,$47,$F2
	.dc.b	$B7,$FC,$5B,$1F,$DD,$4F,$F6,$0F,$A5,$C0,$D7,$A7,$E9,$39,$FA,$9F
	.dc.b	$6A,$BF,$C8,$55,$FE,$09,$C7,$EC,$FA,$FF,$BC,$7F,$AF,$1F,$C9,$5C
	.dc.b	$47,$FB,$3E,$3F,$CA,$0F,$F1,$93,$E7,$9F,$E8,$6F,$5B,$6F,$F6,$7E
	.dc.b	$4F,$F2,$6E,$1C,$63,$FC,$30,$7F,$95,$8A,$3E,$2F,$6E,$3A,$7F,$CA
	.dc.b	$BF,$96,$7F,$89,$B7,$FF,$43,$CB,$D7,$F8,$CF,$8C,$F0,$19,$E0,$FF
	.dc.b	$7E,$C6,$A3,$FD,$03,$CB,$FE,$4B,$7E,$B1,$FD,$0E,$3E,$AB,$F8,$AF
	.dc.b	$63,$FC,$CB,$D3,$FE,$46,$3C,$51,$BC,$3F,$AE,$57,$F9,$1D,$FF,$55
	.dc.b	$4F,$E4,$A3,$11,$FE,$DC,$8F,$F7,$C6,$BF,$EA,$CF,$47,$FC,$95,$7F
	.dc.b	$DB,$CF,$8C,$7E,$CC,$FE,$AC,$FE,$AD,$FF,$59,$19,$3F,$C7,$1D,$3F
	.dc.b	$E4,$A7,$F5,$67,$F8,$EF,$AB,$9F,$17,$D1,$AF,$F2,$5C,$FC,$D3,$47
	.dc.b	$F0,$23,$F8,$2B,$A7,$D6,$7F,$04,$3F,$C8,$07,$E0,$7A,$7F,$D8,$AF
	.dc.b	$F9,$12,$3F,$88,$BE,$5F,$43,$FC,$88,$FF,$B2,$BB,$BF,$8D,$FC,$27
	.dc.b	$FB,$8F,$E3,$3F,$C7,$31,$81,$F2,$3F,$E3,$CF,$13,$FC,$78,$7F,$BB
	.dc.b	$7C,$CC,$38,$AF,$F1,$C7,$93,$FC,$D2,$AB,$FC,$EF,$E2,$BF,$C7,$39
	.dc.b	$7E,$58,$FF,$2F,$79,$BE,$8E,$FF,$F5,$A4,$60,$FF,$09,$9F,$42,$3F
	.dc.b	$A8,$FB,$2F,$67,$FC,$69,$F2,$5F,$E3,$6F,$4D,$39,$7F,$D4,$3C,$5F
	.dc.b	$D9,$B6,$F4,$9F,$E3,$87,$0F,$A7,$1F,$F1,$D1,$FE,$CE,$E9,$FF,$60
	.dc.b	$B8,$3F,$C4,$67,$E0,$BD,$5F,$C2,$BF,$8A,$3F,$C5,$1E,$4F,$B5,$FF
	.dc.b	$15,$6B,$EC,$C3,$FC,$01,$1C,$8F,$F4,$DF,$AD,$FE,$27,$7F,$C7,$CF
	.dc.b	$C3,$C3,$C3,$3F,$E2,$99,$F8,$67,$FA,$8B,$D7,$F0,$BF,$E4,$56,$B9
	.dc.b	$3F,$08,$0F,$C2,$3F,$C0,$7A,$3F,$9E,$F8,$1F,$E2,$B9,$D8,$E9,$FF
	.dc.b	$57,$FB,$1D,$23,$FC,$5C,$7E,$F4,$F8,$1A,$7F,$8B,$1F,$DC,$9F,$E7
	.dc.b	$28,$F1,$F7,$7F,$96,$F1,$F0,$4F,$8E,$FF,$C5,$AF,$81,$FF,$74,$5F
	.dc.b	$F8,$D2,$3F,$AE,$E3,$E6,$7F,$C6,$27,$84,$FF,$32,$D7,$65,$3F,$1D
	.dc.b	$F9,$C7,$F8,$C8,$F9,$37,$5F,$F2,$E6,$63,$F8,$DC,$FD,$63,$ED,$16
	.dc.b	$3F,$AD,$CF,$F8,$CD,$FC,$34,$F1,$39,$65,$98,$F8,$73,$FE,$7B,$F6
	.dc.b	$3E,$F4,$7B,$9F,$BD,$5C,$4F,$F1,$AC,$7F,$AB,$F8,$3F,$5F,$D5,$C6
	.dc.b	$31,$1F,$EA,$DE,$AF,$F1,$17,$27,$E7,$CF,$D0,$72,$E3,$9B,$FD,$31
	.dc.b	$1F,$A3,$F2,$FF,$C5,$1B,$7B,$3A,$3C,$7B,$7F,$8A,$F9,$BF,$63,$29
	.dc.b	$FE,$17,$D5,$FE,$2E,$3F,$D7,$1D,$7F,$E3,$17,$03,$AF,$FC,$5B,$B7
	.dc.b	$A3,$7F,$F1,$93,$FB,$0D,$FF,$C5,$5E,$E7,$FB,$4B,$BB,$F8,$85,$7F
	.dc.b	$8C,$63,$1D,$DF,$C4,$23,$FC,$56,$7E,$14,$FF,$17,$9F,$87,$DB,$FC
	.dc.b	$48,$FF,$27,$1F,$21,$FE,$C2,$F0,$7F,$DF,$FE,$1F,$F2,$4F,$7B,$FC
	.dc.b	$E3,$1F,$CC,$AB,$FB,$47,$0F,$F2,$DF,$0F,$F9,$FF,$93,$FE,$24,$9F
	.dc.b	$D4,$75,$7D,$D7,$FC,$50,$7E,$3B,$C5,$FE,$7C,$FE,$B7,$B7,$FC,$4A
	.dc.b	$FF,$43,$2F,$7F,$F8,$69,$E7,$5F,$D4,$5F,$F0,$FB,$FC,$F1,$6D,$FC
	.dc.b	$45,$FD,$76,$3F,$E2,$7D,$7F,$AF,$C7,$F2,$76,$CF,$EF,$7A,$3F,$B3
	.dc.b	$CF,$F7,$06,$BF,$D1,$3C,$7D,$79,$1F,$DE,$71,$FD,$07,$C2,$E2,$BF
	.dc.b	$C7,$11,$FE,$25,$9F,$D6,$5B,$CA,$E1,$2D,$7C,$1E,$67,$CB,$C3,$FC
	.dc.b	$A7,$7F,$E1,$C7,$FC,$91,$AF,$F5,$4E,$0D,$C7,$F8,$3B,$CC,$FE,$29
	.dc.b	$E7,$7F,$85,$A3,$F9,$53,$5F,$F7,$F7,$9D,$F9,$37,$FC,$92,$7C,$60
	.dc.b	$E5,$FD,$56,$7F,$DF,$F2,$FE,$F3,$E8,$7F,$99,$B2,$FF,$0D,$D7,$F9
	.dc.b	$6E,$3F,$CD,$38,$FF,$25,$F3,$DF,$7D,$F3,$AB,$0F,$CE,$7F,$C3,$DE
	.dc.b	$8E,$5A,$7D,$CC,$9C,$BF,$E2,$4B,$57,$17,$E5,$56,$E7,$5C,$3C,$E7
	.dc.b	$33,$EA,$3F,$C4,$5C,$5F,$DF,$30,$FF,$0D,$9F,$62,$3F,$C5,$87,$37
	.dc.b	$F8,$8B,$83,$91,$FE,$04,$9F,$F1,$84,$7F,$86,$39,$3E,$05,$79,$63
	.dc.b	$B1,$FC,$8F,$E1,$FF,$C5,$0E,$03,$FD,$7F,$CF,$FF,$40,$61,$F4,$DF
	.dc.b	$F1,$24,$FF,$8B,$4F,$D3,$B8,$7E,$29,$FE,$EE,$E7,$FF,$18,$74,$FC
	.dc.b	$B9,$FD,$95,$7F,$57,$E0,$FB,$0F,$F8,$90,$FB,$1C,$3F,$D6,$F8,$7F
	.dc.b	$C7,$1D,$7E,$8C,$FF,$9D,$57,$F0,$23,$27,$EC,$1C,$F6,$F9,$E7,$C1
	.dc.b	$9F,$F8,$FB,$46,$3F,$E2,$98,$C3,$FE,$B3,$E9,$FF,$12,$F7,$BD,$67
	.dc.b	$E6,$FF,$EC,$8F,$97,$CC,$62,$75,$E0,$7F,$89,$BC,$51,$FE,$A4,$E2
	.dc.b	$FC,$C3,$81,$E7,$7F,$9B,$23,$31,$9E,$1F,$F1,$11,$FE,$E6,$F2,$3F
	.dc.b	$DC,$CF,$F9,$00,$FF,$14,$1F,$EE,$4F,$B8,$E5,$E5,$FF,$3A,$9D,$CF
	.dc.b	$F9,$08,$FE,$C3,$C2,$C3,$91,$FA,$BF,$0F,$FA,$A7,$5F,$E3,$23,$13
	.dc.b	$FD,$7D,$F1,$D7,$F8,$C3,$CE,$7A,$CF,$D7,$DF,$8E,$BE,$0D,$F5,$9F
	.dc.b	$E3,$79,$CC,$3D,$AF,$F3,$97,$AD,$FE,$5A,$E2,$F8,$87,$C7,$70,$FF
	.dc.b	$94,$75,$F2,$46,$23,$E6,$E9,$FE,$76,$8F,$ED,$B3,$ED,$8F,$F3,$2E
	.dc.b	$A9,$CE,$DC,$A7,$FB,$DD,$6F,$AF,$C7,$1F,$AD,$8F,$E4,$0F,$F9,$0D
	.dc.b	$FF,$40,$C7,$FA,$27,$1E,$F8,$7C,$A7,$AC,$F0,$8F,$F2,$0F,$07,$F7
	.dc.b	$F7,$FD,$17,$ED,$70,$C3,$CA,$F0,$EB,$FF,$4A,$39,$C7,$FD,$2D,$FB
	.dc.b	$6F,$E9,$CF,$E2,$97,$FE,$27,$ED,$FE,$0C,$7F,$8F,$B0,$FF,$59,$65
	.dc.b	$F9,$CF,$3F,$C5,$F1,$7F,$AF,$3F,$8E,$75,$9F,$E3,$8D,$AF,$E5,$F9
	.dc.b	$7F,$8E,$5E,$78,$7F,$91,$5C,$1F,$A8,$3F,$B3,$BF,$A7,$7F,$C6,$79
	.dc.b	$74,$3B,$E7,$EF,$8F,$F1,$96,$BF,$0D,$FF,$20,$C7,$C1,$7F,$C7,$7B
	.dc.b	$6D,$A7,$D6,$D7,$FC,$B9,$8F,$F8,$BE,$FF,$2F,$AF,$F8,$AA,$BF,$90
	.dc.b	$CF,$F8,$A5,$FF,$17,$79,$3F,$E3,$EE,$1E,$9A,$FE,$0E,$7A,$1F,$E3
	.dc.b	$53,$F0,$7C,$3F,$CB,$23,$FD,$0F,$C5,$F8,$E9,$FF,$57,$C6,$23,$FC
	.dc.b	$C7,$5F,$93,$EC,$FF,$25,$F5,$FF,$AD,$3D,$8F,$F4,$A7,$3F,$E8,$3A
	.dc.b	$3F,$C6,$DD,$5F,$A7,$EA,$FD,$37,$1F,$F8,$CF,$87,$F5,$DB,$76,$CB
	.dc.b	$F0,$4F,$F9,$79,$F8,$DB,$7F,$61,$EC,$1F,$BE,$9F,$F1,$B3,$FC,$03
	.dc.b	$8A,$AE,$DF,$2C,$FF,$93,$38,$7F,$85,$3F,$E3,$87,$F8,$83,$FE,$74
	.dc.b	$C3,$D9,$8F,$C3,$7F,$D9,$FE,$9B,$BF,$F9,$27,$85,$FC,$8E,$FF,$C7
	.dc.b	$9E,$0F,$F2,$33,$FC,$EA,$FF,$8F,$0F,$C4,$1F,$E7,$1E,$CF,$C6,$39
	.dc.b	$7F,$37,$DD,$FE,$38,$8F,$F0,$09,$FE,$40,$3F,$CE,$1E,$2F,$F3,$C7
	.dc.b	$19,$FF,$59,$3F,$D4,$1E,$2F,$FA,$33,$C5,$FF,$01,$E1,$FE,$40,$6D
	.dc.b	$3C,$6F,$F2,$03,$FD,$A0,$B7,$FD,$6D,$FF,$7F,$71,$FF,$5B,$ED,$FE
	.dc.b	$9F,$9F,$FA,$62,$7F,$D8,$78,$FE,$51,$77,$7F,$54,$7F,$CF,$DA,$7C
	.dc.b	$E8,$C6,$3E,$A9,$FF,$1F,$1F,$E0,$53,$F9,$E4,$7F,$83,$1C,$1F,$E2
	.dc.b	$17,$FB,$DC,$7F,$95,$61,$E5,$7F,$8E,$DF,$D0,$9E,$2F,$2B,$FC,$31
	.dc.b	$5C,$2B,$FB,$98,$F4,$B8,$3F,$E0,$E3,$F8,$FC,$7F,$85,$1F,$E3,$F5
	.dc.b	$EC,$F3,$39,$9F,$E0,$5D,$3F,$D4,$5E,$3F,$F9,$4F,$CD,$EE,$7F,$8A
	.dc.b	$70,$FF,$11,$DF,$F2,$B9,$FF,$2C,$79,$DE,$43,$5E,$7F,$D4,$DF,$FC
	.dc.b	$8F,$5F,$E0,$03,$EB,$0F,$47,$FC,$81,$A7,$91,$AF,$C3,$8F,$F7,$07
	.dc.b	$A3,$FE,$45,$AF,$F7,$14,$7F,$9D,$4F,$F2,$4F,$5E,$23,$3C,$7F,$EF
	.dc.b	$BF,$21,$FB,$6E,$6F,$FD,$D5,$F6,$96,$73,$1F,$EE,$DB,$EB,$3F,$E5
	.dc.b	$F8,$4F,$F9,$31,$D1,$FE,$8A,$3F,$C9,$A6,$05,$F8,$78,$3C,$11,$E7
	.dc.b	$9F,$92,$FE,$A0,$FF,$27,$46,$A7,$D1,$C5,$C2,$7F,$C9,$A7,$58,$FD
	.dc.b	$ED,$7F,$93,$B0,$F7,$CF,$F9,$38,$FF,$7E,$69,$FE,$4A,$97,$27,$38
	.dc.b	$FF,$28,$46,$27,$A1,$FE,$11,$70,$FE,$F2,$FE,$55,$79,$60,$7E,$58
	.dc.b	$F8,$BC,$7F,$A8,$7F,$CA,$CF,$A7,$D2,$F1,$3C,$63,$27,$AF,$1F,$04
	.dc.b	$7C,$17,$F6,$18,$7E,$BD,$FB,$FE,$EF,$F9,$6A,$BE,$73,$FE,$59,$FB
	.dc.b	$5F,$F2,$E3,$FE,$56,$6C,$7F,$AE,$CF,$F1,$27,$AB,$E1,$7F,$CA,$E7
	.dc.b	$D7,$DF,$F0,$9E,$7A,$7E,$BC,$7E,$06,$3F,$83,$5F,$C0,$F5,$FF,$C0
	.dc.b	$55,$FE,$D4,$8F,$E2,$B7,$FE,$56,$3E,$41,$FE,$32,$EF,$FC,$86,$FF
	.dc.b	$84,$8F,$E4,$B8,$7F,$99,$A3,$15,$FE,$71,$FA,$0F,$35,$7F,$8E,$DF
	.dc.b	$B1,$C3,$FE,$11,$7E,$57,$93,$F4,$B0,$47,$F3,$AF,$17,$FC,$B9,$F5
	.dc.b	$0F,$CC,$5F,$B1,$7E,$C2,$F7,$97,$DA,$F2,$7F,$A0,$79,$0C,$38,$5B
	.dc.b	$E4,$BB,$67,$FE,$35,$F8,$A7,$EE,$D3,$E6,$B3,$9F,$E1,$3C,$EF,$F3
	.dc.b	$15,$6A,$9F,$3F,$AA,$FE,$BB,$E8,$FF,$5E,$FD,$37,$FC,$D5,$97,$F9
	.dc.b	$76,$1F,$23,$FE,$69,$8C,$4F,$FA,$5D,$FE,$EF,$5F,$92,$E7,$FD,$57
	.dc.b	$17,$B2,$3F,$AA,$9F,$B2,$F0,$3F,$CC,$15,$C2,$3F,$6B,$DD,$FC,$F6
	.dc.b	$B8,$7B,$2F,$80,$FF,$30,$FE,$FC,$F9,$5F,$F3,$0F,$E2,$FE,$9A,$D7
	.dc.b	$B6,$AD,$7B,$6F,$97,$FF,$43,$72,$FC,$1E,$7B,$CD,$F2,$39,$BE,$C7
	.dc.b	$9B,$E3,$E4,$F4,$B7,$93,$FC,$EC,$E7,$8B,$FC,$C5,$C7,$FB,$0F,$5F
	.dc.b	$F8,$EC,$7F,$9F,$BE,$03,$87,$87,$FE,$55,$78,$57,$FA,$EF,$DD,$AF
	.dc.b	$F9,$74,$E4,$7C,$9C,$7F,$D0,$91,$C3,$C4,$FF,$2A,$CF,$F0,$43,$FC
	.dc.b	$DF,$F0,$5F,$90,$7F,$D1,$CF,$E7,$34,$FE,$07,$E3,$6A,$7F,$92,$C7
	.dc.b	$E7,$2B,$F2,$DF,$07,$F9,$CE,$BF,$F2,$A5,$7E,$98,$4F,$B6,$7F,$43
	.dc.b	$B6,$B6,$FE,$3B,$5A,$3F,$CA,$67,$6C,$3C,$2F,$C5,$CB,$C8,$B2,$E1
	.dc.b	$69,$EA,$7A,$DF,$F4,$C4,$FC,$43,$F0,$A7,$F9,$13,$3F,$88,$FF,$96
	.dc.b	$0F,$C6,$DB,$4F,$F2,$66,$BF,$3A,$57,$FE,$53,$77,$97,$F8,$2E,$30
	.dc.b	$7F,$56,$AF,$F1,$94,$F8,$6F,$FA,$24,$B7,$F0,$B6,$75,$5C,$1F,$F5
	.dc.b	$6F,$1F,$E3,$DF,$F2,$07,$80,$8F,$D9,$71,$7E,$27,$B7,$FC,$B5,$DB
	.dc.b	$F9,$EE,$DF,$CE,$E1,$E8,$9C,$F5,$FE,$76,$7F,$D8,$3D,$3C,$6B,$CD
	.dc.b	$BF,$FB,$5F,$5F,$D8,$1F,$94,$8F,$9F,$87,$E5,$3D,$7E,$11,$FC,$03
	.dc.b	$C1,$78,$1F,$E5,$6B,$FD,$C4,$7F,$04,$8F,$EF,$BD,$1C,$23,$F8,$3F
	.dc.b	$83,$FE,$14,$BF,$31,$FC,$3F,$87,$FD,$E7,$C3,$F0,$71,$FE,$05,$1F
	.dc.b	$C4,$BB,$3F,$61,$5C,$07,$8D,$FC,$83,$C5,$FE,$3F,$EE,$79,$A3,$1E
	.dc.b	$47,$F2,$8F,$1B,$D3,$C5,$F9,$2C,$29,$FC,$44,$FB,$9C,$5F,$CD,$B8
	.dc.b	$BF,$CB,$93,$FC,$57,$AB,$F2,$B5,$F8,$4F,$2B,$F2,$BD,$FF,$E5,$B8
	.dc.b	$FD,$71,$FE,$56,$F0,$9E,$87,$FA,$D7,$CD,$FF,$6B,$F1,$FF,$52,$A7
	.dc.b	$BB,$FE,$98,$F3,$BF,$AB,$C7,$F8,$52,$7F,$AD,$3F,$D7,$4F,$EE,$FF
	.dc.b	$0D,$FE,$64,$D7,$D5,$1F,$96,$F4,$7F,$B3,$ED,$FE,$A2,$7F,$B4,$C7
	.dc.b	$F9,$23,$4F,$F0,$CE,$BF,$E5,$F6,$C7,$F9,$78,$FF,$03,$5F,$F9,$7B
	.dc.b	$5F,$EF,$51,$FE,$5D,$7F,$AD,$FB,$7F,$E6,$18,$FF,$10,$4E,$DF,$F1
	.dc.b	$41,$FA,$BA,$FF,$2B,$FA,$9F,$C3,$EB,$FC,$27,$5F,$C0,$BD,$DF,$F3
	.dc.b	$23,$EA,$8C,$DF,$F9,$12,$C7,$27,$FC,$46,$FF,$99,$5F,$F3,$01,$FC
	.dc.b	$77,$BB,$FC,$37,$3F,$E2,$B8,$FF,$1E,$1A,$3F,$C6,$D3,$D3,$8F,$FC
	.dc.b	$C9,$1F,$60,$FF,$42,$FA,$FE,$73,$F0,$4F,$FA,$23,$DC,$F8,$27,$AF
	.dc.b	$D9,$FA,$F1,$FE,$EF,$C3,$FC,$A2,$FF,$AA,$BD,$7F,$F2,$87,$B7,$FE
	.dc.b	$5F,$8F,$C7,$1F,$C8,$5C,$3F,$E6,$1F,$6E,$BF,$EC,$06,$BF,$E2,$AB
	.dc.b	$FF,$28,$EB,$FE,$5F,$6C,$7F,$8E,$BB,$BF,$C7,$7E,$D7,$8A,$BF,$CA
	.dc.b	$FC,$D8,$F6,$7E,$67,$07,$CC,$3F,$CA,$78,$7F,$96,$4F,$F1,$BF,$B7
	.dc.b	$A8,$D1,$F3,$F0,$FA,$B3,$FE,$78,$3F,$C9,$0B,$DE,$E1,$3F,$E5,$C6
	.dc.b	$CF,$F9,$0B,$9B,$FC,$99,$C9,$FE,$1F,$E8,$ED,$3F,$E3,$F3,$FC,$A7
	.dc.b	$7F,$E8,$7E,$5F,$F1,$EC,$7F,$B5,$35,$FF,$1E,$F5,$FC,$08,$6D,$AB
	.dc.b	$FC,$3E,$5F,$E8,$CE,$CF,$F4,$6F,$AB,$FE,$C5,$BF,$F2,$06,$5F,$E3
	.dc.b	$DC,$BE,$9F,$5E,$BB,$FF,$C7,$B5,$FD,$8B,$6F,$F0,$F7,$87,$FE,$2C
	.dc.b	$F9,$17,$C4,$7F,$10,$F8,$9C,$FB,$EE,$75,$6A,$DF,$17,$FE,$27,$F2
	.dc.b	$C4,$27,$C3,$E5,$C1,$AC,$7F,$C5,$DE,$7F,$F2,$F0,$74,$D7,$FC,$98
	.dc.b	$7F,$9E,$9D,$0F,$8B,$FF,$1B,$79,$1F,$5E,$7E,$C7,$27,$F8,$B7,$C3
	.dc.b	$FB,$BD,$1F,$E2,$C9,$FE,$D0,$7F,$8D,$4F,$F4,$2F,$C0,$7F,$89,$8F
	.dc.b	$F1,$3F,$C6,$FF,$8A,$7D,$77,$AE,$9F,$F1,$5F,$C6,$FF,$84,$2F,$F1
	.dc.b	$87,$F8,$2C,$BC,$F9,$EA,$FA,$31,$4F,$F3,$61,$FE,$B0,$EB,$C7,$9B
	.dc.b	$FE,$88,$AF,$F0,$8C,$FF,$0E,$3D,$7E,$FF,$FA,$B5,$FE,$E8,$D7,$FD
	.dc.b	$9B,$C1,$FE,$11,$7E,$14,$FF,$83,$5F,$F0,$E9,$FC,$F2,$3F,$C1,$8E
	.dc.b	$0F,$F0,$84,$6D,$61,$FB,$AB,$FE,$77,$EF,$F7,$9D,$4F,$CF,$7F,$B4
	.dc.b	$9F,$2B,$5F,$8F,$A7,$CA,$8F,$EB,$4F,$CD,$8F,$97,$A7,$F6,$EC,$B9
	.dc.b	$55,$3D,$27,$F6,$1E,$0F,$B3,$C1,$FD,$33,$D1,$FF,$38,$72,$7F,$9D
	.dc.b	$3A,$3F,$B8,$3F,$89,$E5,$FE,$F3,$CF,$F8,$E3,$F2,$7D,$18,$71,$D1
	.dc.b	$FD,$A7,$B3,$F3,$9D,$7F,$9B,$8C,$76,$7E,$7B,$B7,$FB,$8E,$7F,$D9
	.dc.b	$B7,$FB,$3D,$FF,$43,$BF,$F5,$71,$F2,$8F,$F1,$5F,$87,$FC,$4B,$C2
	.dc.b	$FD,$B4,$7F,$67,$7F,$DA,$6B,$D1,$FE,$CF,$3F,$BF,$E9,$FE,$C1,$3F
	.dc.b	$EF,$5F,$13,$FC,$D5,$E2,$FF,$0A,$8F,$C4,$F4,$23,$10,$F0,$3F,$A2
	.dc.b	$E5,$FC,$4A,$1D,$3F,$C6,$5F,$E2,$3D,$3F,$C9,$5F,$E3,$D1,$97,$F4
	.dc.b	$E7,$FB,$A3,$BF,$FA,$CB,$FA,$DF,$43,$FA,$8F,$93,$FC,$8A,$FF,$AC
	.dc.b	$FC,$36,$2B,$F3,$07,$F9,$87,$0F,$E1,$0F,$82,$7F,$97,$E5,$FD,$6B
	.dc.b	$CC,$F8,$13,$E6,$7F,$90,$6D,$98,$FF,$7C,$EF,$F2,$0F,$93,$E6,$7F
	.dc.b	$BF,$3C,$EF,$E9,$33,$FD,$3F,$DE,$FC,$B2,$F3,$FF,$D5,$D9,$FF,$43
	.dc.b	$E1,$FF,$1B,$CE,$7C,$2F,$E8,$51,$FC,$27,$CF,$FE,$99,$23,$FC,$D0
	.dc.b	$78,$83,$A7,$23,$2F,$EA,$BF,$07,$D2,$3F,$A5,$4F,$FA,$C5,$F8,$55
	.dc.b	$E2,$87,$A5,$B9,$FC,$2C,$61,$F9,$1E,$8F,$F9,$2E,$31,$1F,$D6,$2B
	.dc.b	$97,$7F,$F5,$BA,$D3,$CE,$7F,$B2,$1E,$03,$FB,$09,$D3,$3F,$B2,$FB
	.dc.b	$39,$79,$47,$F8,$26,$30,$71,$3C,$B1,$E7,$E6,$FE,$E6,$FA,$79,$FF
	.dc.b	$BA,$E1,$E3,$AF,$82,$D7,$FD,$49,$97,$E3,$F0,$FA,$91,$A3,$E2,$BE
	.dc.b	$18,$FF,$19,$3F,$DE,$FF,$04,$FA,$D9,$FF,$79,$7E,$DF,$A1,$FA,$3B
	.dc.b	$FC,$EC,$FF,$95,$1F,$E5,$F1,$97,$F0,$51,$FE,$F3,$C7,$FB,$B4,$7B
	.dc.b	$4F,$EE,$7C,$9F,$B6,$F8,$0F,$EE,$71,$F8,$13,$EF,$F0,$FF,$73,$CB
	.dc.b	$FC,$09,$87,$E1,$FD,$3F,$EE,$7D,$DF,$E1,$2D,$BC,$51,$FE,$C3,$D3
	.dc.b	$FC,$BF,$8F,$D9,$3F,$AB,$F8,$1F,$9D,$7F,$C0,$A7,$FB,$0C,$FE,$F4
	.dc.b	$63,$CF,$FD,$F2,$97,$D3,$7D,$F8,$C3,$F6,$B3,$DC,$EA,$7E,$D6,$FF
	.dc.b	$DF,$39,$1F,$12,$FC,$3D,$ED,$BE,$1B,$53,$98,$7A,$FF,$DE,$FC,$E7
	.dc.b	$9E,$FB,$F9,$F8,$1E,$63,$10,$FC,$6F,$FB,$FF,$A9,$FD,$FB,$E3,$9F
	.dc.b	$35,$3E,$6F,$32,$9C,$CB,$E2,$3F,$69,$C9,$FD,$F7,$97,$FC,$47,$DF
	.dc.b	$FE,$00,$F1,$BF,$96,$F9,$BF,$BF,$F0,$7F,$C0,$1D,$3F,$C8,$AF,$6F
	.dc.b	$D3,$AF,$F6,$4F,$E4,$7F,$D6,$3E,$13,$FC,$01,$F0,$08,$FE,$B7,$C9
	.dc.b	$E0,$F3,$3F,$9C,$F3,$7F,$87,$39,$7D,$F9,$7D,$39,$CC,$FF,$50,$9F
	.dc.b	$89,$CB,$C6,$BF,$CF,$3A,$78,$9C,$1E,$57,$FD,$57,$AF,$EC,$78,$3F
	.dc.b	$DD,$53,$9F,$13,$FB,$AB,$F8,$BF,$8C,$F7,$DF,$38,$FC,$24,$62,$39
	.dc.b	$9F,$82,$71,$3C,$BB,$FE,$5F,$C6,$E3,$93,$FC,$52,$EB,$2F,$13,$9C
	.dc.b	$7F,$35,$FB,$6F,$E3,$F3,$3F,$C0,$63,$F2,$9F,$05,$F9,$B7,$FD,$3A
	.dc.b	$7E,$6C,$C9,$FA,$BD,$3F,$97,$78,$9F,$96,$FB,$1E,$87,$3F,$60,$FE
	.dc.b	$75,$B5,$F9,$A0,$FF,$87,$1F,$8A,$78,$3D,$AF,$F1,$2E,$1E,$3C,$3B
	.dc.b	$FA,$FF,$CF,$6F,$D3,$7F,$C0,$4E,$79,$78,$74,$F8,$11,$B7,$E1,$6F
	.dc.b	$63,$ED,$E1,$F3,$B0,$E9,$C3,$C2,$7C,$AF,$06,$9C,$38,$7F,$D9,$53
	.dc.b	$E3,$E5,$E1,$9F,$E0,$3A,$FF,$AC,$71,$FF,$73,$C3,$DF,$5F,$57,$2F
	.dc.b	$F7,$7B,$F9,$0E,$6F,$D9,$F3,$7E,$53,$1F,$C4,$C7,$E5,$E3,$F6,$57
	.dc.b	$FE,$34,$F8,$2F,$F2,$EF,$4F,$E8,$63,$1D,$5F,$A1,$3F,$CC,$1D,$5F
	.dc.b	$1F,$B3,$CA,$7A,$F2,$FD,$96,$9F,$06,$FE,$9C,$FF,$87,$BC,$0F,$F3
	.dc.b	$D9,$FB,$BE,$FF,$DF,$4E,$8F,$F3,$E4,$7F,$03,$8F,$E0,$7D,$FF,$E3
	.dc.b	$7D,$BD,$7B,$33,$C7,$85,$FC,$33,$BF,$FC,$6F,$E2,$FF,$15,$F1,$AF
	.dc.b	$8F,$FC,$6B,$C6,$FB,$DE,$4F,$E2,$4F,$C6,$1F,$C9,$7A,$7F,$C7,$9C
	.dc.b	$1F,$87,$2F,$17,$E1,$9F,$F1,$D9,$F8,$5E,$3F,$E7,$31,$FE,$F8,$F3
	.dc.b	$3F,$9E,$CF,$E2,$FC,$DF,$E8,$1D,$9F,$D0,$FB,$BF,$C7,$73,$FE,$06
	.dc.b	$EC,$BE,$37,$F8,$6A,$3F,$33,$ED,$7E,$74,$FF,$24,$5F,$FB,$AB,$3F
	.dc.b	$EB,$19,$FF,$B0,$23,$FA,$C5,$7F,$8F,$1F,$EC,$4F,$F8,$E9,$AF,$F8
	.dc.b	$C7,$3F,$F0,$EC,$FF,$7B,$EF,$FE,$E0,$FF,$B2,$A7,$FC,$83,$3F,$DD
	.dc.b	$5F,$F1,$04,$EA,$7F,$C8,$99,$FF,$8F,$67,$FA,$C0,$F4,$F1,$3F,$DE
	.dc.b	$A3,$FC,$81,$E9,$FF,$53,$B3,$FD,$4F,$DA,$DD,$7C,$93,$C9,$E7,$FF
	.dc.b	$09,$AF,$F2,$3D,$F6,$3C,$71,$F5,$DF,$F2,$DF,$AB,$D4,$E0,$66,$7F
	.dc.b	$C5,$D1,$C4,$FF,$30,$FA,$DD,$7D,$AF,$E2,$9C,$3F,$DB,$EB,$FC,$74
	.dc.b	$FD,$3F,$67,$FC,$8F,$97,$C5,$C1,$B7,$F9,$3F,$C3,$FE,$DD,$EC,$FF
	.dc.b	$67,$9F,$F0,$27,$2F,$FB,$46,$31,$1F,$EF,$B3,$F7,$1E,$CB,$F3,$8F
	.dc.b	$F4,$3F,$A3,$FE,$3E,$7F,$D1,$33,$FD,$F2,$74,$FF,$AD,$67,$FC,$4F
	.dc.b	$F9,$3F,$F2,$BC,$3F,$CA,$CD,$9E,$BD,$B5,$FF,$24,$BE,$ED,$3F,$C5
	.dc.b	$4E,$74,$FF,$2E,$62,$FD,$73,$55,$E1,$79,$5E,$9E,$0F,$F9,$83,$F7
	.dc.b	$FD,$F7,$72,$78,$F3,$D9,$F8,$67,$F9,$82,$38,$E0,$7C,$DD,$3E,$2F
	.dc.b	$CB,$FE,$9E,$DF,$E6,$67,$EA,$F2,$7F,$90,$5F,$1E,$1F,$E7,$07,$E1
	.dc.b	$4E,$E3,$E8,$BF,$3B,$2F,$BB,$1F,$E7,$07,$FA,$97,$E0,$FF,$E6,$97
	.dc.b	$9E,$5F,$E2,$A9,$FE,$8F,$F1,$9F,$E8,$19,$FE,$B0,$FF,$9D,$76,$D9
	.dc.b	$F6,$27,$E4,$BF,$6B,$7B,$1F,$E9,$D3,$FD,$11,$EB,$7F,$9E,$A7,$FD
	.dc.b	$71,$C7,$FE,$32,$9F,$5F,$B1,$FE,$C2,$7F,$BD,$C7,$F7,$CF,$82,$FF
	.dc.b	$62,$CE,$75,$FE,$CA,$7F,$A4,$E3,$FC,$95,$F8,$C7,$FB,$2B,$E3,$5F
	.dc.b	$13,$FA,$EF,$80,$7C,$CB,$F7,$6F,$E7,$F9,$30,$E1,$7B,$AD,$7C,$17
	.dc.b	$F8,$B2,$7F,$C3,$3F,$28,$FA,$58,$3E,$9B,$5C,$1F,$EA,$5F,$5C,$2B
	.dc.b	$74,$FA,$8D,$D7,$F4,$2F,$98,$FF,$1D,$FB,$6C,$F3,$EE,$04,$EB,$DC
	.dc.b	$E3,$E9,$7F,$BA,$0F,$91,$87,$F9,$87,$6C,$3F,$DC,$BB,$36,$FF,$7D
	.dc.b	$3F,$B4,$71,$7C,$17,$FD,$79,$F2,$6F,$80,$7E,$48,$F8,$04,$E7,$E9
	.dc.b	$3D,$2E,$E7,$4F,$1F,$94,$FF,$29,$FB,$9F,$23,$93,$E3,$F3,$E3,$3F
	.dc.b	$8D,$CF,$F1,$67,$E5,$F3,$66,$7F,$C3,$27,$9C,$FC,$0D,$FB,$1F,$F2
	.dc.b	$26,$FB,$E1,$C7,$85,$FE,$62,$FB,$8F,$D9,$D7,$FA,$0D,$FD,$BF,$11
	.dc.b	$E5,$5F,$22,$31,$3D,$FD,$37,$87,$FE,$4F,$BD,$CF,$C8,$FD,$7E,$BF
	.dc.b	$94,$7C,$95,$C2,$BF,$D1,$B1,$AC,$7F,$CE,$BF,$8E,$FE,$78,$F1,$9F
	.dc.b	$E9,$DF,$17,$F3,$D8,$68,$FF,$76,$3F,$E5,$CF,$1B,$FD,$2F,$3B,$F1
	.dc.b	$BF,$41,$7F,$E8,$0F,$9E,$FD,$49,$FC,$53,$63,$FC,$19,$B8,$FF,$12
	.dc.b	$3F,$9C,$9F,$E4,$6B,$E8,$BC,$7F,$C6,$46,$1D,$CA,$37,$8F,$FA,$C2
	.dc.b	$31,$5D,$4F,$F2,$54,$F8,$F3,$F8,$F1,$FE,$4D,$7E,$23,$FE,$56,$7E
	.dc.b	$4E,$9F,$EE,$DA,$FA,$73,$F8,$22,$1F,$E5,$5D,$B8,$C6,$8F,$F3,$E7
	.dc.b	$06,$67,$DD,$A5,$7F,$CC,$3E,$67,$D5,$9F,$F2,$F7,$17,$92,$7F,$CC
	.dc.b	$EF,$43,$C7,$8F,$B5,$FC,$D7,$27,$F0,$1E,$6E,$35,$F9,$AF,$0F,$F7
	.dc.b	$BC,$DF,$E6,$9E,$E4,$FE,$B3,$BB,$F5,$7D,$DF,$AA,$5D,$DB,$9D,$77
	.dc.b	$7E,$AA,$5D,$B9,$9F,$85,$C5,$FB,$E3,$F7,$0E,$0F,$F3,$13,$83,$E1
	.dc.b	$BF,$E8,$2C,$7F,$CA,$1E,$0F,$F0,$AF,$0F,$FD,$93,$E1,$F5,$8F,$E1
	.dc.b	$DE,$1F,$F0,$DB,$E0,$7B,$78,$7F,$72,$E0,$FE,$3B,$E0,$FE,$E6,$3F
	.dc.b	$92,$70,$79,$FC,$8F,$E6,$3E,$DF,$FB,$A3,$C6,$FE,$61,$F0,$79,$F2
	.dc.b	$3F,$97,$79,$3F,$C7,$F3,$FB,$BE,$27,$F3,$C7,$F2,$47,$F0,$13,$FA
	.dc.b	$1B,$FA,$07,$67,$CE,$7F,$90,$F8,$DF,$EA,$C8,$FD,$97,$67,$D2,$7F
	.dc.b	$D5,$D1,$98,$FD,$B4,$7F,$96,$E3,$84,$FF,$AC,$22,$1F,$EE,$53,$FC
	.dc.b	$21,$DF,$FE,$D6,$7F,$3C,$27,$FB,$44,$F7,$F2,$FC,$7D,$FF,$DA,$2B
	.dc.b	$FD,$2D,$5A,$7D,$07,$F5,$F7,$0B,$C0,$E2,$E8,$79,$BF,$99,$7C,$1E
	.dc.b	$77,$F8,$01,$FF,$7E,$B7,$2F,$F2,$BB,$FE,$06,$B3,$FE,$0C,$9D,$9D
	.dc.b	$6F,$FD,$5E,$FA,$FD,$0F,$F0,$ED,$7F,$85,$63,$F8,$47,$A1,$FE,$62
	.dc.b	$9E,$87,$F2,$E9,$D4,$FE,$BB,$5F,$E5,$3E,$8F,$F9,$A3,$D2,$E6,$78
	.dc.b	$CE,$1E,$9F,$94,$EE,$73,$F5,$3D,$07,$8B,$D5,$FF,$07,$61,$F3,$DF
	.dc.b	$F0,$73,$4F,$E4,$E7,$DF,$E0,$FD,$8F,$B1,$FE,$50,$E3,$FF,$78,$70
	.dc.b	$E3,$CA,$FF,$14,$FA,$F9,$C3,$FB,$49,$FB,$87,$0F,$F8,$4E,$7F,$C8
	.dc.b	$B8,$FF,$92,$EE,$CF,$F2,$BF,$92,$AF,$F9,$74,$7C,$A7,$F9,$1D,$FE
	.dc.b	$07,$C1,$F9,$43,$E2,$BF,$E4,$93,$EF,$1F,$97,$E3,$FF,$29,$FE,$65
	.dc.b	$F9,$99,$FC,$DC,$FF,$89,$7D,$A4,$E6,$7F,$37,$3F,$CF,$27,$EF,$BD
	.dc.b	$31,$FF,$22,$27,$FC,$54,$EF,$9F,$95,$3E,$07,$FC,$17,$7C,$AB,$FD
	.dc.b	$3B,$AF,$F9,$9B,$87,$F5,$98,$FE,$B7,$D7,$FF,$23,$57,$E5,$DF,$D8
	.dc.b	$57,$F9,$B3,$0F,$F1,$FC,$D3,$FC,$88,$7F,$26,$BF,$F2,$3B,$F2,$6E
	.dc.b	$9D,$9F,$E5,$7A,$EB,$2F,$F4,$97,$CA,$7F,$92,$E7,$E6,$C7,$A5,$FF
	.dc.b	$22,$5F,$F3,$2F,$67,$DF,$9F,$F0,$3F,$5F,$FC,$97,$5F,$DC,$4F,$F2
	.dc.b	$AE,$9F,$E3,$4F,$13,$FC,$93,$F0,$2F,$99,$7C,$C2,$5F,$33,$BF,$15
	.dc.b	$7B,$AD,$7C,$CF,$F8,$FB,$CC,$FE,$4E,$7F,$28,$F8,$CC,$38,$39,$E9
	.dc.b	$FE,$39,$3F,$E0,$0F,$9F,$FE,$41,$3F,$C8,$55,$FE,$44,$70,$7F,$91
	.dc.b	$7D,$4F,$F1,$EC,$FE,$6B,$87,$FA,$2F,$8B,$F8,$FE,$47,$26,$AB,$EA
	.dc.b	$9D,$CF,$EA,$9E,$C4,$F1,$3F,$C7,$BE,$B7,$F5,$38,$7C,$D7,$F8,$D7
	.dc.b	$6F,$40,$DF,$0F,$AF,$8F,$FC,$75,$A7,$87,$3F,$F1,$B6,$7F,$E5,$6E
	.dc.b	$6F,$F1,$F6,$B9,$8F,$F1,$B1,$FA,$23,$FC,$AB,$EE,$7F,$8F,$DF,$F1
	.dc.b	$B3,$FC,$EE,$F8,$6F,$FE,$01,$9E,$81,$7B,$BC,$63,$FD,$19,$18,$7F
	.dc.b	$DF,$77,$FB,$3C,$7F,$0D,$5B,$9F,$F1,$2D,$7F,$87,$1F,$D4,$1F,$C1
	.dc.b	$8E,$DE,$F7,$E5,$5F,$E0,$C6,$0F,$E2,$63,$AB,$FD,$33,$1F,$ED,$EB
	.dc.b	$FF,$1F,$4F,$FA,$78,$FC,$89,$E9,$F8,$0E,$8F,$F9,$55,$AF,$F7,$E6
	.dc.b	$BF,$E8,$16,$E9,$FE,$30,$8F,$0D,$FF,$8E,$9B,$19,$7C,$03,$B3,$FC
	.dc.b	$75,$87,$48,$D7,$7F,$F2,$4A,$FF,$1C,$47,$F7,$33,$FD,$6D,$ED,$FC
	.dc.b	$B7,$FC,$6A,$7F,$84,$B4,$FA,$DA,$FC,$D8,$4F,$D7,$7F,$97,$B5,$FC
	.dc.b	$9B,$FE,$40,$DB,$F0,$8F,$F9,$0B,$0F,$14,$FF,$8D,$8F,$DF,$D7,$F7
	.dc.b	$ED,$7F,$C6,$71,$E2,$3F,$CA,$FE,$4F,$8A,$75,$1C,$4F,$CC,$F1,$7E
	.dc.b	$65,$C3,$F9,$CE,$2F,$F1,$87,$C0,$FF,$A8,$F9,$BF,$49,$5F,$E2,$18
	.dc.b	$C1,$9E,$6F,$F1,$B6,$9F,$85,$9F,$F1,$33,$E8,$3F,$67,$8F,$C3,$7F
	.dc.b	$C7,$3D,$5F,$BA,$3F,$7B,$3F,$E2,$D7,$07,$F9,$09,$FF,$18,$65,$FB
	.dc.b	$0E,$7F,$F2,$23,$E7,$8F,$F1,$A7,$7F,$F1,$8F,$03,$F8,$BB,$FC,$33
	.dc.b	$0F,$F2,$3B,$9E,$EF,$E2,$B5,$FE,$2A,$E6,$FE,$48,$FF,$8B,$F8,$BF
	.dc.b	$C4,$3C,$FF,$E8,$6E,$4F,$F1,$46,$DF,$CB,$69,$E0,$FD,$E7,$F9,$8F
	.dc.b	$87,$FE,$2F,$F8,$1F,$E1,$F1,$FE,$29,$3F,$9E,$17,$1F,$D4,$78,$9F
	.dc.b	$E6,$6C,$7A,$63,$FD,$6B,$3F,$DF,$63,$FD,$53,$0F,$EF,$87,$F8,$BF
	.dc.b	$A3,$F6,$13,$FE,$26,$8F,$DA,$9F,$9B,$3F,$C5,$FE,$37,$F5,$CA,$FF
	.dc.b	$22,$B6,$32,$7F,$65,$3F,$C4,$67,$23,$F7,$FD,$7E,$5F,$1F,$FB,$2E
	.dc.b	$3F,$E4,$E6,$CF,$F8,$F3,$D0,$FD,$9F,$93,$FE,$1F,$8F,$CA,$D7,$FB
	.dc.b	$0F,$D0,$F7,$79,$9F,$E0,$99,$FF,$0C,$1F,$E0,$7D,$F2,$FF,$82,$C7
	.dc.b	$9D,$FE,$12,$7F,$9B,$B5,$FF,$2E,$0F,$3B,$FD,$07,$87,$F8,$0A,$3F
	.dc.b	$D7,$B1,$88,$FF,$32,$47,$F9,$8F,$CD,$61,$FB,$F8,$43,$F4,$3F,$7D
	.dc.b	$7F,$B8,$B8,$BF,$D3,$72,$FC,$F7,$EA,$5F,$FA,$7B,$83,$3E,$8F,$F8
	.dc.b	$4F,$8F,$F5,$F6,$8D,$D7,$F8,$57,$F5,$3D,$65,$FE,$5B,$A4,$6B,$0F
	.dc.b	$F1,$45,$7F,$BA,$F0,$FF,$3D,$F0,$63,$0F,$F7,$AC,$76,$F4,$FA,$F1
	.dc.b	$66,$7D,$F8,$7F,$8A,$B0,$FC,$AC,$7F,$87,$39,$3C,$31,$E9,$3E,$0B
	.dc.b	$E6,$78,$1F,$E8,$AE,$6E,$0F,$F1,$03,$D1,$CD,$EF,$CB,$FC,$4E,$D8
	.dc.b	$E9,$AF,$F8,$A7,$0F,$A9,$1F,$2A,$7F,$C5,$FD,$7C,$23,$F8,$DE,$1F
	.dc.b	$0F,$0F,$95,$1F,$EC,$A9,$FB,$11,$88,$FE,$71,$C3,$FE,$23,$C1,$F1
	.dc.b	$7F,$E9,$FF,$8D,$FF,$12,$52,$64,$FF,$A8,$FE,$37,$F6,$BE,$77,$F9
	.dc.b	$0A,$BF,$A1,$B7,$B7,$FD,$B3,$A7,$F6,$2F,$83,$FF,$55,$78,$5F,$E2
	.dc.b	$4F,$0B,$FC,$45,$EB,$7F,$88,$31,$FF,$05,$4E,$44,$39,$FF,$C3,$FE
	.dc.b	$37,$F8,$57,$0F,$F5,$A2,$3F,$C1,$5B,$7F,$A7,$74,$FF,$6E,$C7,$FB
	.dc.b	$3B,$1F,$F0,$BC,$7F,$28,$F2,$07,$38,$FF,$77,$B8,$F9,$BF,$E7,$F3
	.dc.b	$B8,$FE,$EA,$FF,$32,$BF,$F0,$35,$7F,$81,$30,$FE,$8D,$CD,$E1,$7F
	.dc.b	$A4,$57,$81,$FF,$65,$79,$7F,$D0,$EB,$F6,$27,$F8,$A5,$7A,$6F,$C5
	.dc.b	$BC,$35,$C2,$1E,$77,$8B,$D0,$F8,$3B,$7C,$1C,$79,$E4,$9C,$F9,$DE
	.dc.b	$A7,$75,$C6,$BF,$C3,$3A,$FA,$9F,$F6,$F5,$F1,$73,$AF,$CB,$BF,$5C
	.dc.b	$7E,$7E,$74,$FE,$B6,$B4,$FF,$9D,$0F,$F1,$0F,$A9,$C1,$EA,$FC,$68
	.dc.b	$FE,$09,$1F,$E3,$83,$FD,$C4,$FE,$3E,$7E,$BF,$A5,$C6,$7F,$C5,$A7
	.dc.b	$92,$78,$CF,$F8,$27,$CC,$FC,$B4,$7A,$FC,$1F,$F3,$C7,$EE,$7F,$B3
	.dc.b	$BF,$F1,$E3,$FC,$15,$FF,$17,$1F,$D0,$3A,$BF,$DE,$F8,$7F,$8D,$DF
	.dc.b	$F1,$A7,$77,$F8,$B9,$FF,$45,$75,$FF,$8F,$3E,$13,$A5,$1E,$9E,$8D
	.dc.b	$C3,$E1,$03,$FC,$33,$1A,$F5,$7F,$C8,$18,$7E,$9A,$7F,$C4,$E7,$FB
	.dc.b	$F3,$06,$78,$8F,$EA,$B9,$73,$C5,$9F,$B5,$B3,$F4,$6B,$FC,$5E,$FC
	.dc.b	$CD,$3E,$F3,$8D,$7E,$3E,$BF,$E2,$8D,$BE,$A5,$7F,$70,$C3,$FC,$54
	.dc.b	$FD,$3B,$FF,$1B,$EF,$F6,$F1,$F1,$3C,$33,$FF,$14,$D7,$F4,$99,$AF
	.dc.b	$FB,$0C,$FA,$B1,$FE,$2E,$CB,$C3,$7F,$E4,$2E,$0E,$8F,$F8,$B6,$3E
	.dc.b	$DF,$0F,$08,$FF,$16,$F1,$7E,$7D,$E0,$7C,$D7,$F7,$CF,$F3,$3C,$FF
	.dc.b	$2B,$1F,$C1,$78,$FF,$C7,$30,$E6,$FD,$67,$2F,$F9,$29,$B7,$FE,$4A
	.dc.b	$7F,$61,$CF,$C1,$FC,$94,$7F,$94,$9F,$3D,$7F,$03,$C7,$CC,$FF,$07
	.dc.b	$9F,$DE,$B5,$F2,$4F,$E0,$DF,$F5,$A7,$57,$57,$D5,$97,$B5,$D5,$7F
	.dc.b	$1A,$8C,$6F,$FE,$38,$7F,$8F,$EF,$FA,$A9,$FF,$1A,$F6,$7F,$B2,$BB
	.dc.b	$BF,$96,$FB,$5F,$87,$CF,$F9,$8C,$E7,$BF,$F9,$6F,$7F,$F8,$DF,$4F
	.dc.b	$F4,$7E,$3F,$CE,$5F,$F3,$E6,$9F,$E9,$E7,$F3,$67,$F8,$CF,$3F,$E8
	.dc.b	$6F,$E7,$7C,$2F,$F1,$D3,$FB,$2F,$0B,$FA,$37,$37,$F8,$D7,$C3,$FE
	.dc.b	$A1,$E2,$7F,$8D,$2B,$FC,$7E,$7F,$46,$9F,$F1,$D7,$8B,$FD,$72,$3F
	.dc.b	$C2,$BD,$1F,$D8,$72,$FE,$A5,$DD,$F8,$38,$C4,$7F,$2A,$94,$6F,$1F
	.dc.b	$F2,$74,$74,$3F,$C9,$AF,$FB,$62,$1E,$37,$D3,$AF,$E1,$37,$FE,$35
	.dc.b	$BF,$EE,$D1,$FC,$5A,$B7,$1F,$09,$AF,$F2,$13,$CF,$5F,$E0,$47,$FC
	.dc.b	$C1,$E4,$7F,$83,$1F,$2B,$FC,$46,$4F,$F8,$6A,$BF,$B7,$70,$7F,$90
	.dc.b	$6B,$F9,$64,$78,$7C,$9E,$06,$DF,$F3,$2D,$FF,$89,$67,$F8,$01,$F0
	.dc.b	$BC,$AD,$6B,$FE,$65,$F2,$FC,$1E,$6F,$F8,$DE,$FF,$C6,$D7,$FE,$4E
	.dc.b	$C3,$F9,$9F,$9F,$FE,$3F,$7F,$DA,$6D,$8F,$EA,$FE,$7F,$EA,$5F,$F6
	.dc.b	$A9,$F9,$B7,$0F,$FB,$10,$FE,$32,$7F,$91,$3C,$FF,$E5,$3E,$47,$F8
	.dc.b	$10,$FF,$21,$B8,$8F,$F6,$57,$B7,$87,$F9,$14,$FF,$82,$BD,$1F,$F2
	.dc.b	$4E,$9F,$CC,$A7,$F5,$5D,$3F,$E4,$87,$F9,$75,$7D,$BC,$FF,$A6,$F8
	.dc.b	$5F,$16,$7C,$95,$FE,$52,$E1,$FF,$78,$57,$F9,$25,$E9,$3F,$E3,$0E
	.dc.b	$3C,$CF,$26,$BF,$E1,$38,$FF,$2F,$C7,$07,$DF,$7F,$71,$F8,$BE,$8F
	.dc.b	$F9,$0B,$1F,$C7,$1F,$E7,$09,$FC,$4B,$F9,$1F,$73,$CF,$CD,$CE,$F8
	.dc.b	$1F,$E9,$39,$E1,$1F,$05,$FF,$22,$F4,$FA,$A7,$A3,$FC,$36,$3F,$87
	.dc.b	$74,$FF,$93,$A7,$EB,$9F,$E4,$93,$FA,$77,$4F,$F9,$25,$4F,$F8,$E2
	.dc.b	$3E,$6C,$7F,$9F,$27,$A3,$FD,$6F,$0F,$ED,$F5,$FE,$7C,$9C,$CA,$7F
	.dc.b	$B9,$F0,$66,$3E,$D8,$BE,$EF,$53,$2A,$3F,$D0,$FA,$F8,$3A,$BF,$94
	.dc.b	$7A,$BF,$E5,$4C,$BF,$CB,$98,$7F,$6D,$8F,$C7,$F9,$8F,$C0,$7E,$54
	.dc.b	$A3,$0F,$E1,$8F,$F8,$03,$D7,$FD,$D6,$33,$1F,$86,$38,$4F,$F2,$8F
	.dc.b	$83,$7C,$1B,$E0,$9F,$B2,$BD,$0E,$57,$BA,$D4,$B9,$B3,$8F,$F3,$EF
	.dc.b	$8B,$7D,$34,$1F,$2B,$A9,$FF,$2E,$79,$CA,$D5,$3E,$A3,$5E,$DF,$F9
	.dc.b	$6B,$E0,$BF,$91,$70,$FA,$A3,$FC,$BB,$ED,$7F,$96,$B7,$FE,$A1,$1B
	.dc.b	$CF,$FC,$C7,$1F,$D3,$9F,$A9,$CB,$FE,$4A,$8C,$74,$FD,$BE,$8F,$F1
	.dc.b	$7C,$62,$78,$D7,$F8,$77,$DC,$FF,$2C,$FC,$73,$EA,$5F,$5E,$FA,$A7
	.dc.b	$D9,$FB,$6F,$75,$A9,$7B,$9F,$4F,$D7,$FF,$2E,$F2,$FC,$0E,$7C,$69
	.dc.b	$F2,$79,$BE,$47,$36,$21,$EA,$FF,$0D,$7F,$CE,$87,$F8,$9C,$FC,$87
	.dc.b	$87,$FE,$75,$71,$C9,$FF,$00,$3C,$AF,$C1,$7B,$BE,$D9,$FF,$72,$F4
	.dc.b	$FF,$97,$B0,$FF,$4D,$E1,$FE,$5E,$9E,$E7,$C8,$3F,$CC,$9D,$FC,$5F
	.dc.b	$F2,$F9,$DA,$3F,$D2,$91,$9C,$3F,$CA,$B3,$FB,$43,$FC,$AA,$7F,$2C
	.dc.b	$72,$FE,$9B,$C4,$FD,$0B,$8E,$7F,$D0,$AF,$84,$FF,$59,$FC,$07,$E9
	.dc.b	$E3,$F4,$97,$F9,$DE,$DF,$D1,$E9,$FE,$55,$9F,$CF,$69,$71,$FF,$29
	.dc.b	$47,$F8,$54,$FF,$2B,$9E,$33,$E6,$E1,$D5,$78,$DF,$E1,$68,$C3,$FD
	.dc.b	$02,$BF,$CA,$73,$E9,$3F,$CC,$59,$7F,$93,$6B,$F6,$B3,$AB,$F0,$D7
	.dc.b	$1D,$3E,$53,$FE,$4E,$3E,$BC,$7F,$14,$9F,$D8,$47,$CD,$ED,$FF,$11
	.dc.b	$6F,$C2,$3F,$06,$FC,$F6,$9B,$8F,$F2,$F7,$2E,$EB,$FA,$C9,$F8,$99
	.dc.b	$F5,$47,$E8,$B8,$3F,$C7,$AF,$E5,$38,$FF,$71,$C9,$FE,$33,$E5,$FD
	.dc.b	$AF,$2F,$F9,$D7,$B7,$FD,$71,$DB,$FA,$3E,$DF,$D1,$76,$E7,$4F,$D4
	.dc.b	$7C,$12,$B5,$D9,$FA,$29,$FC,$DE,$1E,$53,$D5,$BF,$F1,$63,$F2,$7C
	.dc.b	$1F,$E4,$18,$FD,$EF,$7B,$C0,$FF,$12,$57,$F0,$38,$C5,$FC,$9A,$DF
	.dc.b	$57,$F9,$91,$F6,$70,$FF,$14,$E1,$F3,$57,$F1,$2E,$1F,$6B,$E4,$3F
	.dc.b	$D3,$9E,$D7,$FB,$DB,$C7,$FE,$4B,$E3,$FF,$24,$E9,$E3,$39,$F1,$FF
	.dc.b	$91,$F8,$DD,$3A,$FE,$FE,$34,$FE,$88,$79,$4F,$E6,$C7,$FA,$F3,$8B
	.dc.b	$FC,$3E,$7E,$39,$FF,$2A,$1F,$8D,$3F,$27,$E0,$62,$7A,$C7,$F9,$13
	.dc.b	$CA,$FF,$07,$78,$7F,$D4,$3D,$54,$63,$CB,$FF,$2B,$57,$EC,$E7,$2F
	.dc.b	$8B,$BF,$FA,$FB,$CC,$C6,$43,$FA,$C3,$FE,$52,$F3,$3F,$D8,$F8,$FF
	.dc.b	$BF,$AD,$18,$9F,$F6,$26,$7F,$E5,$57,$FC,$50,$5C,$FF,$CA,$8F,$F9
	.dc.b	$8A,$7F,$B6,$E7,$FE,$61,$F3,$7F,$C0,$53,$FE,$8B,$C3,$FC,$16,$FF
	.dc.b	$81,$CF,$F1,$8B,$8C,$7F,$CC,$CE,$A3,$D1,$87,$F9,$61,$F9,$F9,$7F
	.dc.b	$3F,$F3,$7F,$CA,$53,$FB,$4C,$ED,$7D,$63,$FD,$B9,$B6,$1F,$AD,$56
	.dc.b	$FF,$DE,$7E,$C7,$FB,$87,$17,$A3,$FE,$27,$C7,$0F,$F8,$B2,$F5,$EC
	.dc.b	$67,$1F,$F2,$D1,$E4,$C7,$FC,$E5,$E9,$E7,$1F,$F2,$CC,$FF,$82,$B5
	.dc.b	$9A,$F7,$F5,$7E,$E4,$FF,$A1,$1F,$C0,$FA,$BF,$EF,$3F,$5F,$FC,$AF
	.dc.b	$5E,$DA,$46,$1F,$71,$FE,$5D,$F2,$7F,$C3,$AF,$F9,$89,$AF,$DA,$3F
	.dc.b	$CB,$C7,$E6,$27,$FC,$4F,$EA,$F9,$9F,$F1,$2D,$FF,$97,$67,$FC,$42
	.dc.b	$4E,$7F,$F2,$A3,$FE,$73,$D7,$FC,$9D,$7F,$E4,$B8,$D9,$FE,$C9,$E3
	.dc.b	$28,$7E,$0B,$FE,$61,$AF,$CA,$4F,$F9,$3E,$31,$1F,$E3,$C9,$FC,$14
	.dc.b	$F4,$6E,$3E,$B7,$FD,$3B,$3F,$9D,$3F,$D5,$73,$DF,$97,$B7,$37,$3E
	.dc.b	$7F,$F2,$0F,$07,$F9,$2F,$1F,$F6,$F3,$DF,$DA,$F2,$61,$FD,$3B,$24
	.dc.b	$62,$3D,$7C,$BF,$E4,$77,$E2,$DF,$CA,$C7,$FC,$9A,$FF,$BA,$76,$D6
	.dc.b	$9F,$B1,$9F,$EC,$B5,$FE,$4B,$3D,$0F,$FB,$BF,$CF,$61,$FE,$4C,$9F
	.dc.b	$F2,$16,$7F,$EF,$3F,$4B,$FC,$99,$1F,$9D,$7F,$C9,$7E,$27,$FB,$A3
	.dc.b	$DD,$7C,$33,$DF,$78,$2F,$88,$FF,$1C,$2F,$80,$5F,$0A,$DF,$C5,$FF
	.dc.b	$8F,$7D,$CF,$F1,$E9,$FE,$3E,$F6,$B0,$E0,$E9,$AF,$F9,$48,$FB,$FE
	.dc.b	$70,$84,$AF,$F2,$1B,$83,$5F,$17,$FE,$4B,$F4,$3F,$DC,$D5,$FE,$4D
	.dc.b	$F0,$BF,$96,$73,$7F,$8F,$23,$FC,$4C,$7F,$80,$7C,$6F,$F1,$B7,$A8
	.dc.b	$F8,$57,$BC,$8C,$7C,$5E,$A7,$31,$BF,$8B,$FF,$0B,$61,$E0,$F7,$FF
	.dc.b	$BF,$3F,$E0,$D3,$FC,$EF,$EF,$7F,$AC,$3A,$BF,$37,$1F,$E2,$23,$F7
	.dc.b	$87,$E7,$BA,$FF,$C3,$AF,$D8,$3F,$43,$1F,$E8,$0F,$5F,$F6,$71,$CA
	.dc.b	$FF,$C3,$8F,$7D,$7F,$4F,$C1,$88,$FE,$CB,$5C,$27,$38,$F8,$E6,$9D
	.dc.b	$C6,$1F,$E1,$D8,$FF,$63,$E3,$FE,$1B,$9F,$EA,$E7,$B7,$0F,$7B,$F2
	.dc.b	$CF,$ED,$07,$C6,$D2,$6B,$FD,$BB,$6F,$99,$19,$D7,$FB,$8D,$FD,$8B
	.dc.b	$E2,$FC,$97,$FC,$5D,$D9,$FD,$07,$83,$ED,$F6,$FF,$78,$E4,$FC,$11
	.dc.b	$F8,$7E,$5C,$38,$E5,$FE,$FB,$CD,$F8,$F3,$FC,$21,$97,$FA,$89,$C7
	.dc.b	$67,$F7,$AE,$BF,$CD,$F6,$FE,$87,$B7,$F2,$B7,$FE,$A8,$7F,$41,$DB
	.dc.b	$F9,$F5,$9F,$F7,$3D,$FF,$58,$5E,$4F,$D8,$73,$FF,$70,$8F,$DB,$9F
	.dc.b	$EB,$B3,$1C,$1F,$BD,$7F,$C9,$3C,$1B,$F0,$7F,$81,$C7,$F7,$33,$F8
	.dc.b	$39,$FC,$93,$C1,$FE,$A1,$E1,$7F,$0D,$C7,$FB,$33,$FC,$3E,$7F,$81
	.dc.b	$9F,$DD,$CF,$F0,$7C,$7F,$77,$F1,$3F,$8D,$C7,$F3,$DE,$FF,$E4,$79
	.dc.b	$78,$E3,$FB,$87,$8B,$FC,$89,$D3,$FD,$1B,$2F,$ED,$91,$FC,$E1,$E5
	.dc.b	$E3,$7F,$32,$8F,$F3,$75,$FF,$38,$7F,$9D,$DF,$F8,$7D,$F0,$7C,$17
	.dc.b	$E7,$3C,$8F,$EC,$F1,$FD,$17,$D0,$FE,$BC,$FF,$4F,$3F,$DB,$97,$FD
	.dc.b	$3D,$FF,$14,$ED,$A7,$E2,$8F,$2B,$FB,$17,$97,$FA,$18,$FE,$C2,$FE
	.dc.b	$A9,$FE,$9B,$F0,$7F,$EE,$4C,$FE,$04,$38,$F3,$DD,$FB,$89,$CC,$FF
	.dc.b	$54,$F3,$1F,$13,$99,$FF,$17,$9F,$D2,$E7,$51,$9F,$37,$F9,$BF,$3F
	.dc.b	$1F,$C8,$B1,$5F,$C1,$F0,$46,$23,$FA,$B7,$A3,$5F,$E1,$32,$8F,$07
	.dc.b	$4F,$F8,$52,$7F,$B3,$46,$F4,$F3,$78,$9A,$D3,$FA,$F6,$58,$BF,$A7
	.dc.b	$1F,$D8,$1F,$ED,$D3,$C5,$FE,$E1,$E9,$F1,$9F,$F2,$17,$27,$F7,$A9
	.dc.b	$FE,$D1,$97,$87,$3F,$F1,$0C,$74,$D7,$F7,$51,$88,$FD,$09,$FE,$2C
	.dc.b	$3A,$9B,$3F,$BC,$73,$F9,$DF,$EF,$87,$9A,$BF,$BF,$BF,$54,$FF,$02
	.dc.b	$47,$C5,$7E,$A3,$F3,$FA,$FE,$63,$F3,$B1,$FF,$2D,$69,$F6,$7A,$BE
	.dc.b	$4E,$5F,$E0,$76,$E1,$FE,$C5,$CB,$F4,$73,$EA,$9F,$B5,$E5,$7D,$77
	.dc.b	$B6,$1D,$2F,$FC,$2D,$9F,$D4,$7F,$D8,$67,$E1,$DC,$3F,$CB,$B1,$C3
	.dc.b	$F0,$4F,$04,$7E,$1E,$7F,$15,$1D,$0F,$98,$FF,$80,$E7,$FC,$0A,$BE
	.dc.b	$A7,$FA,$6F,$07,$E1,$B2,$FC,$A4,$7F,$50,$FA,$9F,$E0,$7E,$73,$EF
	.dc.b	$3E,$46,$48,$C3,$F3,$F1,$FC,$3B,$99,$FC,$24,$FF,$23,$F1,$9F,$8E
	.dc.b	$7F,$24,$F1,$DF,$33,$FE,$06,$9C,$9F,$CB,$3E,$CB,$FC,$03,$EA,$3E
	.dc.b	$92,$0F,$A4,$CB,$8F,$97,$FE,$FF,$F2,$2F,$9B,$83,$E6,$F2,$E3,$E7
	.dc.b	$7F,$C0,$1C,$5F,$EA,$0F,$33,$FB,$97,$0B,$C5,$FE,$95,$CE,$F8,$1F
	.dc.b	$6C,$E7,$9B,$FC,$F5,$87,$17,$FC,$C1,$19,$7F,$C0,$3F,$2A,$F8,$BF
	.dc.b	$F7,$07,$1F,$F8,$2A,$31,$F2,$4F,$8F,$7D,$47,$F9,$6B,$17,$9B,$FE
	.dc.b	$02,$F6,$43,$A3,$E1,$F3,$FC,$3D,$7E,$A5,$6E,$B5,$3F,$16,$72,$7A
	.dc.b	$38,$3F,$CD,$2E,$07,$A3,$FE,$49,$E1,$F7,$E1,$F0,$B8,$FF,$0F,$5F
	.dc.b	$81,$F9,$2C,$3F,$E1,$63,$07,$5A,$F8,$F1,$A8,$FD,$96,$8F,$89,$C4
	.dc.b	$7B,$DF,$9F,$5F,$39,$E1,$E1,$FC,$3F,$CC,$BF,$23,$67,$F1,$A7,$F9
	.dc.b	$EB,$9B,$F5,$91,$FE,$43,$C3,$F4,$C7,$F8,$87,$DE,$FE,$2D,$EE,$66
	.dc.b	$3F,$86,$9F,$E2,$E7,$15,$FC,$F3,$0F,$F1,$2E,$1D,$63,$FB,$2C,$F8
	.dc.b	$6B,$A9,$A3,$BD,$7F,$8B,$31,$F6,$C2,$7E,$03,$F4,$74,$F9,$71,$88
	.dc.b	$FA,$37,$FE,$F2,$CF,$F5,$72,$D3,$9B,$6B,$91,$FE,$AF,$70,$F2,$DF
	.dc.b	$F3,$17,$FC,$3E,$7F,$17,$3C,$0F,$81,$B7,$CE,$7C,$CF,$D2,$9E,$0F
	.dc.b	$E2,$F8,$BF,$1F,$DF,$FE,$48,$8C,$71,$7E,$37,$0F,$D0,$7C,$07,$F9
	.dc.b	$8F,$A3,$F3,$BC,$EE,$9F,$F1,$E7,$C2,$27,$3D,$3F,$E5,$2E,$CF,$69
	.dc.b	$E3,$C7,$F5,$5D,$3F,$93,$DF,$D5,$3F,$E6,$4E,$8F,$F4,$07,$81,$FB
	.dc.b	$EE,$FF,$DF,$71,$FF,$93,$27,$31,$FB,$F8,$FF,$27,$F6,$7E,$AB,$93
	.dc.b	$D7,$18,$8F,$D3,$F1,$7F,$0F,$F0,$7F,$C7,$DE,$2F,$F1,$7F,$12,$F8
	.dc.b	$9F,$C5,$70,$FB,$9E,$37,$F2,$4E,$1F,$E4,$71,$F8,$CE,$FF,$F3,$6C
	.dc.b	$FF,$63,$E1,$FF,$16,$F1,$3A,$FF,$95,$B6,$FF,$B6,$F9,$1F,$AE,$7F
	.dc.b	$C7,$54,$3F,$9C,$3F,$ED,$EF,$23,$9F,$95,$FD,$06,$7F,$AF,$DF,$FB
	.dc.b	$2A,$BF,$A0,$C7,$7F,$2B,$FC,$83,$1F,$98,$3F,$7F,$5F,$D3,$BE,$1B
	.dc.b	$FC,$87,$1F,$DA,$4F,$F4,$5C,$FF,$BB,$2F,$FA,$EF,$9B,$FE,$97,$7F
	.dc.b	$AF,$FA,$1F,$D0,$1F,$EC,$35,$9F,$3B,$FB,$2B,$FD,$C6,$3F,$C8,$F8
	.dc.b	$FF,$BA,$1F,$F1,$F4,$7F,$B5,$2B,$A7,$A1,$FD,$C7,$3F,$8E,$FC,$63
	.dc.b	$FB,$F4,$FF,$A9,$2B,$8C,$FC,$5B,$FE,$FF,$C1,$FE,$44,$AF,$DE,$FB
	.dc.b	$56,$BF,$79,$1F,$C7,$BD,$3F,$A2,$7D,$5F,$4F,$53,$FE,$F2,$3D,$9E
	.dc.b	$A7,$F8,$77,$23,$F6,$0F,$B9,$BF,$C5,$F5,$7E,$D9,$B3,$38,$FF,$2C
	.dc.b	$F5,$BE,$F9,$CF,$D7,$FB,$18,$FA,$DF,$B5,$CD,$FE,$72,$F8,$7F,$F2
	.dc.b	$74,$FE,$16,$C7,$F8,$6E,$33,$3F,$E2,$8F,$3F,$FD,$03,$27,$FC,$A0
	.dc.b	$63,$DB,$FF,$29,$7E,$3F,$FC,$6F,$7F,$F2,$A7,$C8,$7F,$9A,$EF,$FC
	.dc.b	$76,$7F,$8F,$B2,$F0,$3F,$E4,$13,$F2,$4F,$F9,$FF,$8F,$FC,$79,$8F
	.dc.b	$FB,$5B,$9B,$FC,$A9,$C7,$FE,$86,$BF,$F1,$87,$B7,$FE,$00,$7A,$1F
	.dc.b	$ED,$93,$FC,$C9,$E0,$FF,$A4,$F0,$FF,$33,$7C,$17,$F9,$C6,$8F,$B3
	.dc.b	$1C,$DF,$F9,$F4,$6D,$87,$FC,$8F,$F2,$BF,$E6,$0A,$FF,$32,$3E,$2F
	.dc.b	$8F,$FF,$6D,$4B,$04,$7F,$BB,$F1,$FF,$3B,$6B,$FE,$F0,$F8,$4F,$F0
	.dc.b	$B7,$C5,$97,$C3,$6B,$1F,$A7,$1A,$3F,$D7,$AF,$5A,$FE,$E5,$E5,$7F
	.dc.b	$78,$3F,$88,$46,$4F,$B6,$FF,$B3,$1C,$1F,$EE,$3F,$57,$FD,$C3,$F1
	.dc.b	$3D,$E3,$1E,$A3,$E2,$7F,$27,$1F,$E3,$33,$BD,$F7,$F3,$FF,$C1,$51
	.dc.b	$88,$FF,$04,$E7,$FE,$0D,$F1,$BF,$B0,$4F,$FB,$A4,$FF,$1E,$38,$7F
	.dc.b	$DE,$DE,$F3,$E7,$3F,$D4,$8D,$8F,$C9,$F3,$7D,$B5,$5F,$96,$7F,$C8
	.dc.b	$D5,$F3,$57,$E4,$3F,$9B,$AF,$F7,$C7,$17,$E7,$23,$13,$FE,$0B,$1E
	.dc.b	$37,$E8,$3F,$1F,$CC,$7F,$83,$EB,$1E,$07,$E6,$7C,$FF,$24,$FE,$84
	.dc.b	$FF,$76,$9F,$DC,$72,$F1,$46,$B2,$F0,$CF,$F7,$5C,$FE,$03,$E1,$9F
	.dc.b	$5B,$FB,$07,$63,$BB,$71,$FD,$C4,$F9,$DB,$FF,$8D,$5F,$F0,$01,$8D
	.dc.b	$FF,$C8,$EF,$EE,$B9,$1C,$7F,$6D,$FB,$F1,$FE,$02,$35,$19,$3F,$01
	.dc.b	$E3,$F2,$8F,$C6,$5C,$E3,$C3,$F8,$BF,$23,$FD,$43,$D1,$F9,$5C,$B8
	.dc.b	$3F,$97,$F8,$BF,$E7,$9D,$7F,$A4,$AF,$D1,$E3,$F7,$0F,$71,$FC,$43
	.dc.b	$7F,$E7,$5D,$DF,$AC,$8F,$E8,$8F,$EB,$B9,$2C,$7E,$D7,$8A,$BF,$EE
	.dc.b	$EE,$FB,$1F,$43,$0F,$E0,$93,$FE,$FD,$3B,$9F,$E5,$2F,$03,$F8,$23
	.dc.b	$FC,$DC,$F2,$1A,$3F,$9F,$4F,$F1,$17,$F9,$2D,$FF,$57,$E0,$FF,$77
	.dc.b	$76,$FF,$4F,$6F,$3F,$F9,$6B,$17,$85,$FC,$85,$78,$9F,$C9,$71,$FF
	.dc.b	$31,$E5,$FC,$45,$A7,$F1,$1D,$3F,$96,$BF,$C4,$FE,$27,$F8,$DB,$FC
	.dc.b	$78,$FE,$6F,$E5,$7F,$15,$D3,$FB,$D6,$7F,$A5,$F3,$BF,$D5,$BC,$3F
	.dc.b	$D6,$E7,$F9,$69,$FE,$5A,$EC,$FD,$56,$7F,$AD,$EB,$FF,$28,$C6,$32
	.dc.b	$F8,$F1,$FE,$54,$BE,$57,$FE,$FD,$EC,$F9,$7B,$F4,$8F,$F6,$83,$FC
	.dc.b	$B2,$FF,$D9,$59,$7F,$85,$38,$FE,$3C,$7E,$E7,$5F,$F0,$06,$9F,$D3
	.dc.b	$36,$FE,$A3,$E2,$3F,$20,$F8,$CF,$C8,$BF,$E9,$19,$FF,$46,$E6,$F8
	.dc.b	$71,$E1,$8F,$F7,$04,$F8,$63,$F6,$C4,$BE,$BC,$FE,$DE,$0F,$E9,$99
	.dc.b	$FF,$A8,$FF,$1F,$FF,$39,$3F,$EA,$5F,$8F,$B8,$7F,$9B,$5B,$C2,$AC
	.dc.b	$F0,$FE,$72,$BF,$82,$61,$D1,$FF,$4F,$EB,$ED,$7F,$20,$7A,$B5,$F8
	.dc.b	$FC,$DF,$E8,$19,$FF,$11,$6F,$F8,$5F,$D5,$3F,$C6,$97,$FD,$D3,$ED
	.dc.b	$9E,$E4,$26,$BF,$DE,$2B,$74,$AD,$56,$FB,$FE,$6F,$C3,$7F,$29,$9E
	.dc.b	$7F,$24,$F4,$7F,$DB,$9F,$95,$FF,$3F,$E7,$FF,$6A,$65,$E1,$C9,$7B
	.dc.b	$F7,$D6,$FA,$5E,$B7,$BA,$D4,$BE,$E7,$EE,$F2,$F8,$71,$F2,$9F,$D7
	.dc.b	$32,$FF,$62,$E1,$DE,$3D,$3F,$2B,$F8,$4E,$1E,$11,$EC,$F1,$B1,$5B
	.dc.b	$8F,$89,$5F,$8A,$E3,$DC,$7C,$19,$FE,$E9,$18,$8F,$97,$5B,$9E,$51
	.dc.b	$F3,$AB,$11,$F2,$E3,$CF,$F3,$BF,$E1,$E8,$DC,$7F,$78,$7F,$C0,$1C
	.dc.b	$DD,$63,$F9,$B6,$5F,$99,$9E,$51,$E9,$76,$72,$BE,$8E,$A3,$EF,$5E
	.dc.b	$38,$BE,$3C,$7F,$40,$B7,$DC,$7C,$4F,$85,$FE,$87,$87,$EA,$A3,$1A
	.dc.b	$FF,$18,$8F,$F0,$A4,$79,$74,$D4,$78,$4C,$9F,$47,$87,$EF,$7C,$9E
	.dc.b	$A3,$D3,$1F,$81,$F9,$3F,$F1,$C6,$9F,$93,$B7,$B8,$E6,$FF,$53,$EB
	.dc.b	$FE,$0A,$C7,$E0,$D6,$A5,$87,$2D,$BF,$07,$35,$FC,$0D,$7F,$92,$0F
	.dc.b	$93,$F3,$DF,$BC,$71,$97,$E8,$E3,$FB,$A7,$9B,$FC,$12,$3F,$D7,$3B
	.dc.b	$FF,$81,$FB,$71,$83,$87,$E3,$FB,$98,$8F,$A9,$B7,$F4,$8C,$3E,$7F
	.dc.b	$27,$0E,$1F,$CD,$FA,$F7,$B3,$F4,$FD,$5F,$E4,$5B,$FC,$5F,$27,$C1
	.dc.b	$FC,$DB,$FC,$8D,$C5,$FA,$9E,$3F,$81,$1F,$E5,$8E,$0F,$A5,$3F,$E4
	.dc.b	$77,$F1,$1A,$73,$D9,$C7,$FE,$0F,$F2,$BF,$4D,$5F,$A9,$8C,$57,$F8
	.dc.b	$E7,$F6,$FF,$E9,$FE,$84,$1F,$41,$97,$03,$E8,$BF,$C8,$1F,$6E,$FC
	.dc.b	$D9,$F8,$6B,$ED,$30,$BE,$4B,$73,$A9,$CF,$CC,$FF,$92,$74,$FF,$1F
	.dc.b	$70,$FF,$93,$BF,$4F,$FE,$69,$E8,$E6,$3F,$C6,$1E,$07,$F9,$3F,$F1
	.dc.b	$3F,$F7,$CF,$CF,$3E,$7B,$F8,$35,$7F,$95,$A7,$31,$F8,$68,$FF,$29
	.dc.b	$1F,$E9,$6E,$3F,$5F,$17,$BA,$32,$23,$3C,$5F,$E5,$38,$FF,$4E,$E9
	.dc.b	$E6,$EE,$FF,$18,$C3,$5F,$3B,$FE,$56,$36,$7A,$04,$7F,$41,$7F,$80
	.dc.b	$F5,$7F,$49,$7F,$55,$19,$7F,$84,$73,$7D,$99,$D4,$63,$D7,$FF,$57
	.dc.b	$F4,$B1,$FE,$A1,$3D,$0E,$FE,$AF,$E1,$1F,$F2,$EF,$C6,$FE,$23,$E1
	.dc.b	$BF,$DE,$51,$FC,$1F,$B7,$F8,$C3,$A3,$F8,$AF,$99,$FE,$61,$AF,$E2
	.dc.b	$9E,$47,$F9,$7D,$FE,$2D,$16,$7F,$8D,$CD,$7A,$57,$FA,$43,$E1,$7B
	.dc.b	$F8,$7F,$A4,$9F,$EC,$87,$9E,$99,$79,$DF,$85,$3F,$D6,$37,$F8,$54
	.dc.b	$9C,$E1,$F9,$ED,$BE,$1E,$BF,$E6,$3F,$63,$F4,$B5,$AA,$FE,$06,$FD
	.dc.b	$89,$FC,$7B,$6F,$EB,$3F,$73,$8B,$ED,$BF,$E4,$09,$FF,$36,$FB,$1F
	.dc.b	$E8,$99,$CD,$7F,$A2,$E7,$FC,$E2,$E7,$D8,$FF,$0E,$76,$7F,$9D,$63
	.dc.b	$FD,$21,$CF,$F9,$0F,$0F,$11,$F9,$3E,$9F,$D6,$F4,$FF,$9A,$7A,$FF
	.dc.b	$33,$E3,$DE,$DF,$CF,$7C,$D1,$FD,$07,$9D,$FE,$6F,$DB,$F5,$3E,$A7
	.dc.b	$F7,$A7,$FC,$B5,$BF,$8C,$FF,$1C,$EF,$FE,$75,$F1,$FF,$6B,$E2,$7E
	.dc.b	$DB,$C6,$C7,$8D,$FB,$53,$F7,$7E,$37,$F9,$EF,$93,$D1,$C7,$FE,$81
	.dc.b	$3F,$4D,$C7,$E8,$5A,$BD,$FF,$E2,$AD,$8F,$E8,$CF,$F8,$B7,$B7,$EE
	.dc.b	$76,$7E,$6B,$C4,$FF,$39,$C6,$29,$E2,$7E,$75,$FF,$11,$9F,$9D,$BE
	.dc.b	$5D,$BF,$E0,$2E,$3E,$2B,$0F,$E4,$F5,$87,$F9,$4F,$8E,$8F,$F1,$59
	.dc.b	$FD,$D7,$CD,$FE,$6A,$E3,$2F,$F0,$EF,$9B,$FC,$DA,$31,$87,$F4,$E6
	.dc.b	$BF,$D3,$B2,$FE,$80,$FF,$45,$F0,$3F,$A4,$4F,$0F,$43,$FA,$3B,$F1
	.dc.b	$BE,$07,$FC,$27,$E2,$7F,$53,$C3,$FA,$F6,$5F,$D4,$F8,$3F,$B8,$46
	.dc.b	$32,$F0,$35,$F0,$5F,$F8,$73,$6F,$EC,$9E,$9F,$FA,$37,$2F,$F1,$14
	.dc.b	$7F,$67,$EC,$F8,$18,$7F,$80,$27,$FB,$B6,$DF,$DB,$EF,$3E,$A7,$F8
	.dc.b	$9B,$83,$FB,$B5,$7D,$C8,$FB,$67,$F0,$BB,$E5,$1F,$E1,$4D,$FF,$C1
	.dc.b	$2E,$CF,$F1,$93,$FE,$CE,$E1,$F0,$FA,$BF,$C2,$27,$F0,$EF,$F0,$C6
	.dc.b	$9F,$75,$EC,$7F,$88,$FD,$AE,$31,$FE,$28,$F8,$4F,$F2,$FB,$F8,$CF
	.dc.b	$63,$A6,$3F,$E8,$18,$FC,$34,$71,$7F,$2B,$CB,$F9,$43,$D5,$3F,$96
	.dc.b	$3D,$9E,$DF,$F9,$37,$5F,$C1,$C7,$F9,$6A,$BF,$0D,$BF,$E0,$E3,$1C
	.dc.b	$BF,$86,$F6,$7F,$CB,$7E,$DF,$E3,$7E,$FF,$FC,$D7,$3F,$E3,$0C,$F1
	.dc.b	$E2,$F8,$FD,$AF,$C3,$E9,$FE,$5E,$C3,$FB,$C7,$3F,$DA,$F1,$BF,$31
	.dc.b	$ED,$7D,$57,$EC,$FC,$07,$97,$7F,$F3,$7D,$FF,$9E,$72,$FC,$E7,$07
	.dc.b	$F8,$97,$7F,$F1,$51,$FE,$2B,$F9,$7D,$47,$E8,$72,$CC,$EB,$D8,$FF
	.dc.b	$38,$CF,$F8,$B7,$DE,$7E,$55,$FE,$7E,$F0,$73,$87,$F9,$FA,$3F,$D0
	.dc.b	$3F,$05,$8C,$7F,$CF,$71,$FE,$84,$F8,$5F,$F3,$BE,$3D,$E7,$B9,$FE
	.dc.b	$3C,$E7,$F3,$E1,$88,$FE,$ED,$C9,$FE,$52,$8C,$78,$1F,$E7,$6B,$F6
	.dc.b	$57,$B3,$C1,$FF,$36,$D7,$16,$C7,$6F,$46,$BE,$EB,$AF,$C4,$E0,$FF
	.dc.b	$33,$DF,$86,$AC,$F9,$AE,$BC,$70,$FF,$37,$9E,$27,$E6,$38,$AF,$57
	.dc.b	$CF,$7F,$97,$DF,$A3,$FA,$EF,$F9,$3F,$3F,$D3,$C6,$38,$3F,$CD,$91
	.dc.b	$8C,$3E,$2E,$9C,$5F,$16,$3F,$E6,$C3,$EE,$70,$FD,$4F,$1B,$FD,$41
	.dc.b	$E2,$FC,$AA,$FA,$91,$FC,$9F,$0F,$F3,$46,$5F,$82,$FD,$07,$FC,$DD
	.dc.b	$8F,$F9,$8E,$31,$5F,$E0,$0A,$DD,$F8,$AB,$4F,$8A,$32,$BF,$33,$FC
	.dc.b	$7C,$FF,$BC,$3C,$8F,$14,$E6,$3D,$BD,$7E,$5D,$FF,$DD,$7F,$21,$FE
	.dc.b	$F0,$CF,$FC,$FD,$1F,$EE,$EF,$67,$18,$78,$73,$61,$D2,$3F,$01,$8A
	.dc.b	$72,$FE,$37,$7F,$89,$E0,$38,$B9,$C6,$3A,$3F,$A0,$5F,$FB,$F7,$4F
	.dc.b	$F4,$77,$0F,$FB,$8F,$F8,$07,$FC,$DF,$97,$FA,$36,$31,$E5,$70,$F8
	.dc.b	$5F,$16,$BF,$C9,$EB,$FC,$E0,$7E,$75,$FD,$5F,$46,$E3,$F4,$1C,$9F
	.dc.b	$B4,$F3,$FF,$D3,$77,$F5,$E3,$15,$E4,$F4,$3F,$D1,$DF,$4A,$F9,$37
	.dc.b	$C7,$3E,$75,$8B,$E1,$78,$69,$FD,$6E,$18,$71,$BE,$11,$FE,$BF,$F1
	.dc.b	$BF,$82,$FE,$AD,$FE,$EF,$FB,$53,$7E,$C3,$E2,$CE,$BC,$A5,$69,$BD
	.dc.b	$1A,$CF,$FB,$57,$CB,$BE,$59,$F2,$C8,$C3,$E7,$9F,$B1,$8F,$F1,$69
	.dc.b	$C9,$AF,$6F,$FC,$BB,$D7,$FC,$02,$BF,$CC,$FF,$13,$F4,$3A,$7F,$CF
	.dc.b	$5E,$7F,$F9,$96,$BF,$D8,$DD,$3F,$37,$C9,$E3,$5F,$91,$C3,$FC,$13
	.dc.b	$D5,$FE,$64,$7F,$71,$09,$C5,$FF,$09,$9F,$F3,$01,$A8,$FF,$30,$3F
	.dc.b	$76,$7F,$23,$F1,$BF,$E6,$8A,$FE,$31,$E6,$EA,$BA,$F9,$9F,$E6,$19
	.dc.b	$CC,$3E,$33,$86,$1F,$C8,$AB,$4F,$EB,$E3,$2F,$E2,$A7,$FC,$CA,$78
	.dc.b	$CB,$5E,$7F,$57,$E1,$4F,$BF,$C0,$F6,$57,$F9,$97,$7F,$C7,$3F,$E8
	.dc.b	$6D,$7F,$CC,$8F,$AE,$BF,$35,$7C,$9F,$03,$FE,$9C,$F6,$FF,$CB,$F3
	.dc.b	$FE,$27,$7F,$C9,$6F,$E0,$39,$7E,$FF,$1F,$DE,$EE,$FF,$06,$F2,$FD
	.dc.b	$D9,$C9,$F8,$4E,$5C,$DF,$33,$FD,$49,$E8,$AB,$F2,$1D,$5F,$8F,$1D
	.dc.b	$7C,$2B,$1D,$9F,$99,$EB,$FF,$0D,$F8,$5F,$A8,$F0,$BF,$4F,$E1,$7E
	.dc.b	$9B,$6C,$E1,$FA,$5F,$0B,$F4,$93,$FA,$AE,$2F,$F1,$1F,$83,$F1,$AB
	.dc.b	$FC,$9F,$DB,$87,$F7,$27,$C9,$3F,$CA,$27,$D3,$E0,$FE,$15,$CB,$FC
	.dc.b	$D9,$BE,$27,$F0,$B9,$E4,$7F,$0B,$F3,$C3,$C6,$78,$DF,$DF,$27,$F8
	.dc.b	$D7,$A5,$FD,$BD,$FF,$2A,$F7,$7F,$3C,$F2,$DC,$9B,$9F,$17,$99,$FA
	.dc.b	$5E,$EC,$5F,$F2,$CF,$33,$FC,$07,$E7,$E3,$81,$E7,$FF,$2F,$C7,$FC
	.dc.b	$A4,$7F,$19,$F2,$3F,$9D,$B8,$3F,$C1,$4F,$F8,$33,$6F,$F2,$64,$7C
	.dc.b	$1F,$43,$C0,$EF,$2F,$E9,$7C,$3F,$E7,$18,$C6,$3F,$CB,$A3,$FD,$9B
	.dc.b	$C7,$FD,$7E,$3F,$C0,$F1,$B8,$C7,$9F,$FD,$77,$C1,$FF,$29,$DF,$E5
	.dc.b	$EF,$F3,$B5,$C5,$FE,$86,$F3,$D3,$FC,$D5,$E8,$FF,$52,$7F,$A1,$57
	.dc.b	$CE,$3F,$C9,$A7,$F0,$2F,$2B,$FC,$96,$FE,$E0,$CE,$7F,$C4,$6D,$EE
	.dc.b	$7B,$9B,$3E,$3F,$4F,$FC,$A2,$F3,$3F,$D4,$3E,$DF,$F1,$7E,$BC,$E1
	.dc.b	$FC,$36,$FF,$CA,$95,$FC,$4A,$FF,$CE,$7D,$BB,$8F,$F2,$A9,$EF,$E6
	.dc.b	$FF,$3B,$D6,$4F,$F5,$6C,$64,$FF,$57,$FB,$7F,$E0,$1E,$1F,$EC,$3C
	.dc.b	$FF,$E7,$E8,$C7,$5F,$F8,$CE,$B6,$7E,$3B,$91,$D6,$F5,$BF,$CA,$B3
	.dc.b	$F3,$F8,$AB,$FC,$A3,$D1,$FF,$11,$F3,$7F,$91,$E3,$FB,$97,$67,$6C
	.dc.b	$79,$E1,$FD,$86,$7F,$41,$0F,$DB,$9E,$CF,$FA,$C7,$89,$F2,$BF,$E5
	.dc.b	$8F,$C1,$FF,$DD,$FF,$1D,$7C,$EF,$F7,$4E,$5F,$EC,$F9,$FE,$0D,$1F
	.dc.b	$ED,$A8,$FD,$9F,$47,$E3,$73,$ED,$97,$E7,$F9,$7F,$C0,$7A,$FD,$EA
	.dc.b	$FF,$27,$70,$7F,$98,$B0,$FF,$4B,$9F,$E1,$AF,$95,$FE,$24,$FF,$BD
	.dc.b	$0F,$F3,$5F,$C4,$7F,$99,$37,$FF,$36,$92,$3F,$C2,$5C,$D5,$47,$F9
	.dc.b	$CB,$E3,$BF,$CC,$31,$D8,$FC,$97,$BB,$A7,$F9,$3E,$FE,$B8,$C4,$6F
	.dc.b	$C4,$FF,$0B,$F7,$FF,$9F,$3B,$3E,$0D,$FC,$2E,$1C,$DF,$9C,$FF,$19
	.dc.b	$E1,$F8,$EC,$BF,$77,$1E,$0F,$03,$95,$FC,$CA,$FC,$DC,$FF,$BE,$23
	.dc.b	$A8,$F9,$5F,$F1,$3D,$7B,$67,$F8,$2D,$70,$BE,$B9,$EA,$7C,$43,$E6
	.dc.b	$7F,$C5,$FE,$27,$83,$2F,$E7,$DC,$5E,$68,$C4,$23,$FC,$C7,$1D,$EB
	.dc.b	$47,$4B,$E6,$FD,$5C,$32,$FF,$B4,$1F,$F3,$5F,$8D,$DF,$D1,$FF,$38
	.dc.b	$37,$C2,$CD,$79,$8F,$F2,$C1,$D4,$FF,$26,$47,$DB,$EB,$E3,$1F,$97
	.dc.b	$01,$FE,$1B,$87,$F9,$33,$BB,$DD,$9F,$E1,$76,$FF,$2A,$DF,$CD,$D7
	.dc.b	$DC,$F2,$E4,$E0,$F8,$CF,$D5,$9E,$39,$E3,$AF,$E3,$4D,$C7,$E1,$9F
	.dc.b	$C7,$3F,$E5,$23,$ED,$6B,$FC,$83,$D9,$FF,$2D,$F2,$7F,$6D,$8C,$69
	.dc.b	$FB,$5B,$FC,$2D,$D3,$FA,$0C,$5D,$79,$45,$DF,$E0,$E3,$F9,$AC,$3F
	.dc.b	$D9,$FF,$60,$FB,$17,$D9,$BE,$B1,$F7,$6A,$FE,$1D,$97,$F9,$E7,$0E
	.dc.b	$38,$70,$BD,$D3,$F0,$6F,$F2,$ED,$BF,$2E,$7D,$5C,$27,$C1,$EA,$31
	.dc.b	$19,$F2,$BF,$CA,$D5,$FE,$0A,$3D,$FC,$9F,$BE,$70,$7F,$A6,$B8,$B7
	.dc.b	$FA,$EB,$E7,$5F,$32,$F9,$9C,$2F,$AF,$DC,$EA,$73,$E3,$78,$35,$F1
	.dc.b	$EB,$E8,$DB,$0E,$36,$FC,$5F,$6B,$E1,$FF,$41,$BF,$FB,$EA,$30,$FE
	.dc.b	$55,$AF,$F0,$9E,$AF,$AA,$FF,$99,$8F,$CB,$3F,$79,$FD,$67,$CE,$F4
	.dc.b	$7F,$CC,$5D,$9F,$B8,$7F,$CB,$EF,$E0,$9A,$7E,$08,$A7,$FA,$FB,$C3
	.dc.b	$FD,$0F,$C4,$BE,$7B,$CF,$18,$7D,$3C,$B6,$BF,$39,$6C,$3C,$51,$FE
	.dc.b	$5C,$8F,$11,$EC,$3F,$5B,$8F,$B1,$C4,$7E,$BF,$BF,$F2,$37,$FE,$5A
	.dc.b	$8F,$E5,$85,$9E,$7E,$8F,$ED,$A3,$EE,$9F,$E1,$93,$F9,$AD,$E7,$E1
	.dc.b	$BC,$31,$4F,$F2,$9F,$89,$C7,$8F,$EC,$F1,$38,$FF,$CA,$51,$8E,$3F
	.dc.b	$F2,$A7,$4F,$E1,$39,$BF,$0B,$D1,$8E,$8F,$C2,$1F,$8A,$E8,$FF,$2D
	.dc.b	$6B,$F3,$FC,$3F,$C9,$3F,$4D,$78,$1E,$1D,$7F,$13,$AF,$F9,$6B,$7F
	.dc.b	$D1,$47,$E2,$55,$70,$7F,$8A,$F7,$7E,$CF,$C2,$FD,$97,$85,$FB,$1F
	.dc.b	$3F,$94,$E7,$C2,$FD,$7F,$85,$D7,$A3,$C5,$AF,$F3,$4E,$6F,$F3,$A7
	.dc.b	$97,$FE,$5A,$9F,$AB,$D7,$FC,$0B,$5C,$1F,$54,$78,$BF,$E2,$FD,$FF
	.dc.b	$DA,$F1,$FC,$75,$AF,$F5,$D3,$FC,$D6,$FF,$15,$9E,$91,$8C,$3E,$1C
	.dc.b	$7E,$02,$3D,$FE,$3F,$F2,$37,$F9,$A5,$7F,$26,$7F,$9B,$0F,$1F,$FC
	.dc.b	$CD,$8F,$CF,$AF,$CA,$CF,$23,$FB,$8F,$87,$FD,$EE,$FF,$9B,$9F,$E0
	.dc.b	$E7,$F9,$6A,$F2,$3F,$A2,$63,$F8,$DA,$FF,$33,$61,$F8,$FE,$EF,$E8
	.dc.b	$71,$83,$F9,$6D,$F1,$7F,$A7,$35,$FE,$89,$B7,$F9,$59,$FF,$36,$C7
	.dc.b	$F8,$7A,$39,$AF,$27,$FB,$95,$F8,$63,$15,$FD,$8C,$FA,$13,$FA,$E8
	.dc.b	$CC,$7C,$F6,$9F,$E0,$AC,$B9,$4F,$F7,$27,$FA,$EF,$83,$FE,$64,$BF
	.dc.b	$F0,$4F,$89,$70,$FF,$5B,$F9,$5D,$4F,$C3,$70,$7F,$8F,$23,$FC,$DB
	.dc.b	$97,$F8,$2A,$3F,$C1,$51,$FE,$64,$3E,$0F,$99,$E5,$38,$F1,$FF,$98
	.dc.b	$A3,$E2,$E3,$87,$F3,$B7,$F6,$07,$9B,$F8,$18,$FF,$15,$46,$1F,$C0
	.dc.b	$CF,$F6,$4F,$93,$FE,$7C,$FF,$1D,$6C,$7F,$A7,$BB,$FF,$C7,$8F,$E2
	.dc.b	$3E,$17,$FB,$6D,$FE,$06,$3F,$BF,$C7,$F9,$6E,$3F,$2D,$1F,$E5,$8C
	.dc.b	$7F,$A6,$FE,$A5,$EF,$7F,$9D,$F9,$9F,$E4,$5E,$4F,$F4,$8F,$93,$F1
	.dc.b	$74,$FE,$37,$18,$AF,$F3,$14,$FF,$17,$BF,$F3,$0E,$1F,$E0,$E7,$A3
	.dc.b	$FE,$44,$0E,$31,$FE,$B4,$DF,$84,$FF,$1E,$F3,$75,$AF,$CF,$1F,$E1
	.dc.b	$1E,$5F,$E4,$9A,$FF,$A1,$63,$E3,$9E,$E1,$E7,$FF,$A3,$9F,$F0,$C7
	.dc.b	$1F,$F9,$3D,$CF,$3F,$F8,$E2,$7F,$82,$C6,$0F,$F2,$DC,$EE,$3F,$23
	.dc.b	$BF,$E5,$E3,$F0,$D1,$FA,$1D,$3D,$AF,$EC,$23,$FC,$A9,$D5,$F7,$71
	.dc.b	$FA,$11,$FE,$57,$D3,$E6,$74,$FF,$8C,$B7,$FC,$86,$BF,$D9,$32,$F8
	.dc.b	$EF,$F4,$FD,$3C,$78,$79,$BE,$37,$FA,$C4,$62,$3F,$31,$E7,$E2,$3D
	.dc.b	$F1,$EC,$DB,$94,$62,$3E,$D7,$27,$E5,$71,$FE,$37,$5F,$C9,$2B,$F0
	.dc.b	$87,$F1,$9A,$FF,$4B,$4F,$F1,$58,$FF,$3E,$EB,$FE,$F9,$8F,$B5,$D9
	.dc.b	$CF,$87,$FC,$DD,$1F,$30,$FD,$BD,$FF,$A8,$9A,$8F,$AE,$7E,$C5,$FB
	.dc.b	$07,$E2,$3D,$3F,$F0,$F7,$AC,$21,$3E,$37,$FD,$44,$E0,$CE,$7F,$E5
	.dc.b	$EE,$4F,$D6,$69,$C3,$DB,$7B,$73,$D9,$FE,$DD,$5A,$C3,$B5,$6A,$B7
	.dc.b	$D5,$CF,$BB,$FB,$A5,$6A,$79,$78,$5F,$EA,$68,$F1,$1E,$7F,$93,$C3
	.dc.b	$8F,$82,$F2,$6B,$CA,$37,$38,$D7,$FC,$EB,$1C,$85,$7F,$A9,$A3,$10
	.dc.b	$9C,$3B,$DA,$BD,$E7,$F4,$7E,$17,$78,$F8,$DC,$BF,$E9,$F9,$E7,$56
	.dc.b	$7C,$17,$5E,$5B,$FF,$AB,$9F,$25,$FF,$A6,$AF,$DE,$7F,$A6,$A3,$E9
	.dc.b	$3F,$EA,$23,$D6,$FB,$9F,$D7,$1E,$AC,$79,$47,$D5,$D3,$EA,$FC,$0E
	.dc.b	$23,$E7,$43,$E0,$3F,$D3,$F7,$F7,$FE,$5B,$EB,$5E,$2F,$F0,$B1,$F6
	.dc.b	$6B,$11,$E1,$CB,$FA,$C3,$F3,$71,$CA,$F9,$7F,$F6,$2F,$4F,$C7,$7F
	.dc.b	$C6,$15,$FE,$A2,$9F,$C8,$E5,$FE,$35,$D7,$FD,$37,$18,$8F,$F5,$6E
	.dc.b	$18,$8F,$07,$83,$F4,$5F,$CC,$C6,$79,$7F,$1D,$CD,$F9,$2A,$FC,$24
	.dc.b	$63,$93,$F3,$5E,$0F,$E1,$F7,$F7,$E5,$83,$E6,$BF,$0A,$71,$1F,$86
	.dc.b	$CE,$E7,$EF,$8F,$F5,$73,$FA,$BD,$7F,$C9,$DE,$07,$E2,$E7,$33,$F0
	.dc.b	$DF,$E2,$DA,$FF,$A6,$BE,$2D,$F1,$8F,$93,$7C,$6B,$65,$3E,$BB,$85
	.dc.b	$EE,$B5,$3F,$E9,$1C,$7F,$D5,$47,$FB,$53,$E6,$44,$27,$CB,$7F,$AA
	.dc.b	$9C,$19,$F4,$3F,$D5,$BF,$24,$F9,$27,$CB,$23,$0F,$F8,$67,$A7,$FC
	.dc.b	$9D,$8E,$6B,$75,$AD,$BF,$D5,$FE,$F7,$F9,$83,$E1,$BF,$D5,$1E,$FE
	.dc.b	$1F,$F5,$54,$7F,$A9,$FE,$67,$FD,$6B,$AF,$90,$F4,$ED,$87,$1B,$7F
	.dc.b	$6C,$D3,$F4,$7C,$BF,$C5,$FA,$31,$1F,$A3,$3F,$D4,$47,$F1,$EE,$8F
	.dc.b	$AB,$1A,$9F,$B1,$3F,$50,$FF,$52,$9D,$1F,$E4,$B1,$FE,$B4,$F8,$6F
	.dc.b	$C8,$47,$F2,$9E,$8F,$F8,$03,$E7,$FF,$47,$8F,$DF,$DA,$31,$1F,$C2
	.dc.b	$5A,$FF,$08,$87,$9F,$88,$FE,$67,$E2,$FF,$A9,$E7,$D5,$39,$9F,$E4
	.dc.b	$50,$F2,$3F,$A9,$C7,$81,$AF,$F1,$77,$FD,$8A,$FB,$A1,$DB,$F9,$59
	.dc.b	$CC,$FE,$56,$73,$8F,$FA,$AE,$3A,$79,$FF,$01,$FE,$53,$3E,$1B,$FF
	.dc.b	$56,$7A,$3F,$E9,$7F,$3F,$FD,$5D,$C3,$F6,$47,$12,$FF,$D2,$B3,$A9
	.dc.b	$FB,$71,$83,$5C,$9F,$87,$F8,$3F,$F7,$B7,$A1,$FE,$F5,$8C,$73,$7F
	.dc.b	$A8,$7A,$BF,$26,$7E,$53,$C8,$C7,$67,$E4,$8F,$CC,$F6,$7F,$A9,$F3
	.dc.b	$F2,$6B,$5F,$0E,$5F,$9A,$EF,$FD,$9F,$87,$FB,$2B,$FD,$8F,$87,$FE
	.dc.b	$AF,$9C,$F8,$5F,$B0,$F8,$3F,$66,$1F,$E9,$AE,$5F,$9D,$CB,$FC,$06
	.dc.b	$31,$1F,$36,$3E,$4C,$62,$3F,$31,$C7,$FE,$B8,$CB,$E7,$73,$F4,$7F
	.dc.b	$8B,$79,$9F,$CE,$A7,$07,$F1,$67,$F9,$EF,$8D,$E2,$F2,$7F,$8F,$C7
	.dc.b	$F7,$39,$C1,$FD,$23,$C9,$FE,$44,$FF,$90,$B1,$FC,$BE,$5F,$CC,$DF
	.dc.b	$EB,$CD,$7F,$D3,$9E,$0F,$29,$FE,$A6,$7F,$34,$3F,$D9,$FC,$BF,$CF
	.dc.b	$5F,$F4,$93,$5F,$F6,$03,$F8,$98,$F8,$EF,$BA,$38,$9F,$BA,$F2,$7F
	.dc.b	$99,$DF,$F4,$CB,$FF,$0B,$47,$F5,$19,$FF,$6E,$FA,$5F,$DC,$71,$DB
	.dc.b	$72,$FF,$09,$F9,$9F,$52,$3C,$15,$C0,$FD,$E6,$1F,$B9,$70,$7F,$70
	.dc.b	$3F,$2C,$61,$F7,$EB,$F7,$BD,$9F,$D1,$78,$76,$B9,$D7,$DC,$9E,$FA
	.dc.b	$7F,$7D,$8F,$F0,$CF,$9F,$F8,$6F,$43,$E0,$49,$E7,$7F,$EA,$33,$FA
	.dc.b	$71,$6B,$FC,$24,$FF,$BF,$3C,$4E,$A7,$2A,$EE,$74,$F4,$7F,$03,$B3
	.dc.b	$D0,$F5,$CF,$F8,$CF,$07,$07,$E0,$A3,$0F,$F8,$D6,$BF,$D7,$1E,$CF
	.dc.b	$F9,$1A,$BF,$C6,$38,$FE,$E6,$31,$1F,$E4,$5D,$3F,$C3,$8F,$F9,$1F
	.dc.b	$C2,$FE,$E5,$7F,$E3,$FF,$4F,$FB,$3C,$B9,$3F,$C8,$5F,$21,$FE,$A1
	.dc.b	$F6,$7E,$7F,$91,$FE,$97,$E1,$FC,$E6,$2A,$FF,$2D,$75,$7B,$23,$FC
	.dc.b	$B9,$C9,$FC,$4F,$4F,$E1,$9E,$B7,$F9,$87,$0F,$E1,$DF,$31,$EB,$BF
	.dc.b	$F2,$AF,$C3,$EB,$3F,$BB,$39,$97,$B3,$FE,$53,$BE,$31,$FE,$5C,$F9
	.dc.b	$9F,$F2,$CC,$E6,$55,$FE,$5B,$9F,$F2,$EF,$C8,$FF,$99,$79,$FA,$E1
	.dc.b	$E1,$F6,$F1,$87,$48,$F9,$DD,$FF,$E5,$58,$E7,$1B,$7F,$D5,$FD,$1C
	.dc.b	$A3,$71,$FE,$50,$39,$3F,$A3,$F6,$7F,$15,$18,$8F,$E0,$D1,$88,$EB
	.dc.b	$1F,$77,$6F,$F2,$B6,$9F,$07,$0B,$3F,$1F,$3A,$FE,$0F,$2A,$FF,$9E
	.dc.b	$6F,$FC,$AD,$7F,$E7,$BF,$92,$FF,$69,$64,$F0,$7D,$11,$FD,$57,$37
	.dc.b	$83,$C3,$1F,$75,$EE,$7D,$11,$FE,$53,$7F,$CB,$78,$7E,$4F,$7F,$EB
	.dc.b	$F5,$B9,$F5,$D2,$7F,$75,$2F,$0B,$C3,$F2,$5C,$1B,$D7,$FD,$62,$75
	.dc.b	$87,$83,$6F,$F3,$0E,$5F,$0E,$3F,$DB,$F7,$F4,$7C,$07,$A7,$FC,$12
	.dc.b	$72,$7D,$B8,$CB,$EE,$3F,$DC,$FD,$5F,$82,$DF,$E1,$43,$A7,$FD,$55
	.dc.b	$39,$E5,$FE,$1F,$ED,$E2,$BC,$87,$23,$E2,$F3,$65,$FB,$7C,$3F,$94
	.dc.b	$74,$63,$83,$F7,$BC,$DF,$E5,$8D,$FF,$D6,$BA,$38,$BF,$D8,$1F,$25
	.dc.b	$FE,$B6,$F2,$3C,$D1,$8E,$6E,$1C,$DF,$5A,$7F,$24,$FE,$3D,$FC,$EF
	.dc.b	$17,$C0,$F2,$9C,$DC,$23,$1C,$1F,$EC,$7F,$7F,$F9,$5F,$9B,$FE,$56
	.dc.b	$F7,$FF,$CB,$18,$E7,$8B,$FC,$09,$39,$F4,$7F,$CC,$3F,$21,$99,$7E
	.dc.b	$14,$FB,$F8,$4F,$53,$51,$88,$CF,$C1,$71,$FA,$37,$D0,$3E,$81,$18
	.dc.b	$5F,$09,$B9,$D4,$E7,$BF,$C5,$E1,$FE,$07,$4F,$C1,$78,$9F,$E6,$08
	.dc.b	$C4,$7E,$06,$3F,$89,$3E,$8C,$FC,$87,$F9,$E5,$FF,$7A,$9F,$E2,$3D
	.dc.b	$33,$1F,$E6,$E8,$FF,$2B,$F6,$FF,$15,$E7,$FF,$2D,$BF,$95,$F2,$1F
	.dc.b	$01,$F4,$BC,$6F,$F1,$67,$5F,$D2,$C3,$8F,$67,$F4,$19,$CF,$5F,$F9
	.dc.b	$66,$7F,$1B,$83,$E0,$FE,$ED,$FF,$68,$9F,$F2,$C7,$CA,$62,$3A,$7C
	.dc.b	$22,$31,$87,$F9,$81,$D7,$97,$FE,$57,$D7,$F9,$63,$6F,$F8,$E7,$C2
	.dc.b	$71,$C7,$F3,$67,$F9,$67,$2E,$DB,$FF,$91,$79,$3C,$93,$A7,$DE,$FF
	.dc.b	$34,$F0,$7F,$21,$3B,$E1,$F8,$53,$FE,$5C,$D7,$0F,$EA,$AF,$8B,$FC
	.dc.b	$23,$2F,$74,$3D,$2F,$EF,$FC,$DF,$E5,$9A,$FF,$2B,$35,$FB,$5C,$9F
	.dc.b	$8D,$EC,$FF,$1F,$F9,$FF,$E0,$6E,$7F,$B3,$DB,$FE,$42,$F0,$F9,$E1
	.dc.b	$F8,$41,$E7,$FE,$1D,$FC,$57,$8B,$B7,$FC,$C9,$DB,$F9,$33,$F2,$9D
	.dc.b	$F9,$8F,$CB,$F7,$FE,$50,$FC,$D7,$06,$73,$FC,$4E,$B7,$6F,$C7,$EB
	.dc.b	$C7,$C3,$FD,$61,$FE,$CB,$C7,$C7,$B7,$ED,$3C,$4F,$F6,$6F,$8B,$FE
	.dc.b	$CB,$F1,$3C,$1D,$BF,$91,$BD,$9F,$63,$8B,$93,$FD,$63,$AB,$E9,$BF
	.dc.b	$E5,$93,$EB,$78,$3F,$90,$70,$7D,$1E,$8F,$CB,$57,$2B,$FE,$60,$D8
	.dc.b	$79,$1F,$C5,$FC,$A7,$93,$FE,$1A,$D7,$F8,$BC,$63,$A3,$FC,$A9,$8F
	.dc.b	$F7,$38,$FF,$0E,$F8,$98,$C7,$FB,$D4,$7F,$24,$CF,$E9,$F9,$BF,$90
	.dc.b	$69,$ED,$7F,$9A,$3F,$E9,$7F,$3B,$96,$9F,$CE,$BC,$EF,$F2,$CC,$FF
	.dc.b	$41,$DF,$FC,$BE,$D3,$FC,$BE,$7F,$3A,$DF,$FA,$7B,$F0,$7C,$0E,$2F
	.dc.b	$F7,$4B,$F7,$7B,$78,$DB,$FB,$67,$9B,$FE,$55,$C3,$FB,$04,$7F,$63
	.dc.b	$D7,$FD,$19,$E8,$7F,$34,$F4,$34,$FA,$EB,$7C,$1F,$E2,$1D,$3F,$CC
	.dc.b	$5A,$7F,$84,$74,$FE,$A6,$FF,$98,$32,$FE,$F5,$DD,$FE,$B5,$E3,$FF
	.dc.b	$2E,$4F,$D7,$F2,$6D,$7E,$0F,$27,$A5,$FC,$0A,$BF,$3A,$FF,$07,$F4
	.dc.b	$7F,$CB,$D1,$F0,$7D,$AF,$E0,$AF,$F9,$76,$FE,$FE,$5E,$73,$96,$1F
	.dc.b	$A6,$9F,$90,$E3,$DA,$FF,$11,$70,$7F,$5F,$84,$63,$87,$EE,$7B,$3F
	.dc.b	$CC,$31,$FE,$23,$82,$31,$C3,$FE,$43,$E9,$FF,$1D,$57,$E1,$FD,$AF
	.dc.b	$E6,$1E,$9B,$AF,$A4,$63,$B3,$F1,$FF,$0A,$F6,$BF,$AA,$74,$BC,$C8
	.dc.b	$48,$FE,$35,$E7,$7F,$A2,$A7,$33,$9D,$FF,$DB,$C7,$F8,$9D,$C9,$D5
	.dc.b	$FD,$79,$97,$8D,$7F,$B0,$23,$FC,$8F,$D5,$FA,$5C,$3F,$CC,$B1,$FE
	.dc.b	$4D,$E4,$71,$7F,$9F,$BC,$7F,$F6,$AF,$27,$B1,$FF,$31,$57,$FB,$CF
	.dc.b	$BF,$F8,$B2,$F6,$3F,$AC,$C7,$F5,$CF,$88,$FF,$30,$3F,$EF,$E9,$E0
	.dc.b	$38,$FE,$9B,$83,$FC,$46,$3E,$67,$FE,$00,$47,$E2,$23,$47,$FA,$BB
	.dc.b	$E2,$B3,$C3,$FE,$82,$E1,$F9,$9E,$57,$F6,$FB,$FF,$20,$47,$CE,$8F
	.dc.b	$F5,$AF,$83,$D3,$D2,$F2,$63,$FC,$EC,$FE,$EA,$E3,$97,$71,$E7,$F9
	.dc.b	$5E,$D8,$7F,$AF,$A7,$FD,$7F,$8A,$73,$43,$FA,$36,$1F,$E3,$1D,$7F
	.dc.b	$BF,$B6,$3E,$04,$57,$E0,$EB,$FE,$37,$C7,$D7,$1C,$FE,$A3,$BF,$57
	.dc.b	$BF,$5F,$F6,$7C,$7F,$82,$7A,$7E,$76,$BF,$ED,$2F,$6F,$FC,$6D,$87
	.dc.b	$D0,$8C,$69,$FE,$25,$F7,$5E,$87,$F8,$A7,$E0,$30,$FF,$8A,$FC,$BF
	.dc.b	$F1,$57,$D4,$08,$4F,$3B,$2E,$0D,$57,$F8,$C7,$E7,$9F,$40,$FA,$04
	.dc.b	$61,$7C,$F6,$E7,$53,$9E,$9F,$95,$E3,$FF,$72,$E8,$FC,$7B,$78,$7F
	.dc.b	$B1,$9F,$EB,$2F,$77,$E0,$BF,$E1,$C3,$8B,$85,$C7,$FE,$6F,$F8,$2F
	.dc.b	$F0,$64,$7E,$40,$FF,$1B,$38,$BF,$AF,$1F,$E1,$C3,$F2,$27,$DD,$EE
	.dc.b	$FB,$DF,$03,$B8,$F1,$B5,$FF,$40,$46,$BE,$0B,$F1,$F1,$FE,$43,$F3
	.dc.b	$FD,$5D,$1F,$5E,$3F,$C2,$27,$FA,$11,$FA,$EA,$3E,$03,$FE,$48,$6B
	.dc.b	$F9,$6F,$3D,$D9,$F9,$18,$F9,$FC,$3F,$30,$FA,$5C,$5F,$E0,$5E,$2F
	.dc.b	$A4,$7C,$F9,$FE,$13,$CB,$F6,$CF,$C0,$73,$7F,$82,$DC,$0E,$6F,$F0
	.dc.b	$77,$4F,$E2,$BA,$FF,$C0,$6E,$6F,$F8,$9B,$9F,$89,$7B,$68,$75,$FF
	.dc.b	$15,$E5,$FD,$47,$17,$C3,$C3,$FC,$03,$18,$7F,$CA,$FD,$BF,$AD,$E2
	.dc.b	$FE,$61,$1F,$AF,$F0,$3F,$4E,$BB,$FE,$14,$EA,$3F,$81,$1F,$AB,$AE
	.dc.b	$5E,$07,$F0,$2C,$3F,$8B,$56,$DF,$E0,$BE,$0E,$AF,$F0,$27,$FA,$DE
	.dc.b	$7F,$80,$3E,$1C,$7F,$CF,$39,$78,$3C,$2F,$E0,$73,$FC,$08,$FF,$1D
	.dc.b	$F8,$1F,$C3,$5F,$F0,$09,$FC,$82,$7E,$D7,$89,$FC,$3B,$A7,$FD,$11
	.dc.b	$8F,$F1,$AC,$7D,$BA,$FF,$6E,$D3,$FD,$13,$D9,$FC,$DB,$2F,$E1,$67
	.dc.b	$F2,$EB,$FE,$5A,$FF,$2C,$9F,$C2,$6B,$FC,$57,$D1,$FE,$6F,$3F,$E4
	.dc.b	$F6,$BF,$ED,$7F,$1B,$F9,$CB,$FC,$D7,$E5,$BF,$99,$F0,$7B,$23,$FA
	.dc.b	$87,$8F,$F3,$75,$FA,$3F,$27,$FD,$4F,$C6,$FC,$ED,$7D,$CF,$43,$F3
	.dc.b	$BA,$FD,$B8,$FF,$35,$47,$CD,$7E,$A8,$AF,$EB,$9B,$7C,$E8,$FE,$C6
	.dc.b	$7E,$76,$7E,$03,$ED,$3F,$98,$67,$FD,$23,$E9,$3F,$9B,$E5,$FD,$BB
	.dc.b	$DE,$FE,$97,$EB,$62,$3F,$B9,$77,$7B,$30,$FB,$51,$FD,$A1,$FF,$72
	.dc.b	$E9,$FD,$1B,$1F,$EA,$75,$FC,$2F,$9F,$DD,$FA,$D7,$3C,$3F,$B9,$79
	.dc.b	$2F,$B0,$C6,$DF,$C1,$9B,$9F,$EE,$EE,$EF,$FE,$64,$D3,$E2,$65,$FD
	.dc.b	$D2,$BD,$83,$4F,$D0,$71,$F0,$DB,$C1,$B7,$E8,$F8,$3C,$5C,$1F,$D4
	.dc.b	$7C,$FF,$E0,$78,$FF,$53,$FA,$07,$B3,$FE,$1B,$8F,$EF,$1D,$9F,$C1
	.dc.b	$E7,$33,$06,$DF,$DC,$BE,$AB,$50,$F8,$2F,$F3,$77,$B5,$AC,$7F,$BE
	.dc.b	$56,$A7,$3C,$1C,$67,$C5,$7C,$7E,$03,$FC,$1F,$87,$F7,$F9,$FF,$02
	.dc.b	$7E,$4B,$DA,$BC,$5F,$15,$FE,$02,$8C,$47,$9B,$E0,$BE,$F5,$74,$8F
	.dc.b	$44,$62,$1A,$FA,$E7,$E0,$FC,$06,$63,$FD,$2D,$DE,$E3,$EB,$5F,$1B
	.dc.b	$E0,$BF,$0F,$87,$58,$F8,$BF,$41,$FE,$14,$FC,$1B,$CD,$F0,$5F,$D2
	.dc.b	$71,$46,$39,$3F,$CC,$98,$7F,$87,$E3,$AE,$1F,$EE,$FA,$DC,$F5,$8F
	.dc.b	$F7,$54,$62,$3D,$31,$F6,$38,$FF,$CF,$F1,$CC,$F8,$0F,$FA,$7F,$E4
	.dc.b	$BF,$C4,$71,$F0,$FE,$9B,$FC,$77,$F0,$5F,$CB,$A3,$FC,$61,$E6,$7C
	.dc.b	$D9,$CD,$70,$9E,$F8,$3E,$47,$FC,$55,$7F,$4D,$FF,$11,$E5,$FB,$3E
	.dc.b	$1F,$C5,$ED,$FE,$38,$F6,$7F,$55,$8F,$A2,$3F,$C7,$7F,$2D,$8A,$F9
	.dc.b	$FE,$0F,$C2,$E6,$C6,$9F,$83,$DD,$D3,$F8,$BD,$3B,$4E,$63,$F0,$D1
	.dc.b	$FC,$EF,$0F,$CD,$F0,$7B,$BE,$9B,$FC,$4B,$AF,$E3,$A3,$1E,$87,$F9
	.dc.b	$2B,$D2,$FC,$CF,$0F,$F6,$5F,$33,$E9,$46,$2F,$FC,$71,$0D,$BF,$CA
	.dc.b	$3E,$F7,$F6,$BF,$17,$FA,$EF,$C1,$97,$CA,$EB,$0F,$F2,$04,$68,$FE
	.dc.b	$F0,$FF,$97,$FC,$5F,$EE,$9F,$30,$F8,$67,$CC,$A3,$0B,$E3,$7F,$CC
	.dc.b	$35,$C2,$B7,$5A,$F5,$F3,$87,$F9,$3F,$DF,$FF,$39,$F3,$7F,$00,$8F
	.dc.b	$EE,$47,$F8,$0C,$71,$F0,$7F,$C0,$9F,$11,$FE,$06,$DB,$0F,$F8,$20
	.dc.b	$FF,$04,$F8,$1F,$E3,$2F,$4B,$F3,$3E,$55,$BF,$CC,$76,$7C,$75,$EB
	.dc.b	$AB,$E6,$47,$F3,$BA,$E0,$FE,$73,$A7,$E9,$73,$FF,$8F,$BE,$DF,$AC
	.dc.b	$F2,$7F,$0C,$BD,$7F,$E5,$8F,$82,$FF,$CD,$D1,$88,$FC,$CE,$5F,$C1
	.dc.b	$5E,$1E,$B7,$EC,$3C,$BD,$BF,$B2,$F2,$FF,$9A,$4F,$AA,$6B,$CA,$BF
	.dc.b	$93,$CE,$3C,$1F,$E1,$50,$C7,$0F,$FB,$9E,$BF,$C6,$D9,$7A,$BE,$18
	.dc.b	$FA,$9F,$89,$1F,$C8,$B5,$FC,$95,$FF,$A1,$E7,$33,$F1,$4F,$C9,$EB
	.dc.b	$9D,$BA,$1F,$13,$D7,$FB,$DC,$5F,$5F,$E1,$7F,$CD,$FC,$9F,$E2,$5E
	.dc.b	$BF,$F2,$14,$E7,$E1,$9C,$D9,$CF,$F1,$07,$E1,$4F,$F1,$3D,$7F,$92
	.dc.b	$7A,$7F,$CC,$39,$7E,$2F,$AF,$F2,$A7,$E5,$BB,$79,$5F,$E6,$7B,$7F
	.dc.b	$2A,$7E,$73,$B7,$FC,$6D,$E0,$FE,$A7,$E1,$3F,$C8,$3F,$07,$FE,$47
	.dc.b	$E0,$CE,$1F,$A8,$9C,$F8,$7F,$AD,$E5,$CC,$FD,$3E,$9F,$A0,$B6,$FF
	.dc.b	$3C,$F5,$7D,$4E,$7F,$F0,$53,$FD,$7B,$A7,$E6,$D6,$FC,$5F,$C9,$F8
	.dc.b	$BF,$97,$C3,$F8,$56,$2F,$1B,$F8,$55,$F2,$8C,$79,$3F,$EE,$4F,$23
	.dc.b	$FC,$81,$E5,$7F,$5D,$F5,$7F,$AC,$FA,$5F,$DD,$30,$FE,$BB,$E8,$FF
	.dc.b	$64,$E7,$C4,$7A,$7C,$2B,$C3,$E7,$8F,$56,$3F,$DB,$3C,$FF,$E5,$51
	.dc.b	$83,$FB,$3B,$FC,$A9,$F8,$1E,$8F,$F3,$37,$F7,$9C,$08,$C4,$7F,$95
	.dc.b	$5F,$E7,$DB,$7F,$BF,$3E,$0F,$FC,$D1,$18,$8F,$C1,$CF,$CB,$F4,$FF
	.dc.b	$B8,$6B,$FD,$3B,$0F,$EA,$59,$FF,$81,$23,$FA,$75,$6C,$FD,$91,$4F
	.dc.b	$AD,$CD,$FE,$75,$BF,$EC,$51,$FE,$72,$6F,$9B,$FE,$79,$F6,$FF,$65
	.dc.b	$3F,$6F,$D5,$FD,$B3,$4E,$B9,$7F,$75,$7F,$87,$E1,$FE,$1F,$7F,$88
	.dc.b	$FA,$DF,$E8,$3F,$5F,$F8,$4D,$1F,$D3,$7B,$3F,$5F,$6F,$E9,$BC,$BF
	.dc.b	$E8,$98,$C7,$AD,$FE,$11,$B4,$63,$D9,$FF,$33,$79,$3F,$DE,$3D,$4B
	.dc.b	$A7,$F2,$0C,$7F,$C0,$AF,$E5,$63,$FA,$DF,$17,$F8,$82,$1E,$AF,$E3
	.dc.b	$BB,$BF,$B9,$FC,$7F,$F9,$A3,$1F,$F5,$B7,$D3,$BE,$02,$13,$E0,$3F
	.dc.b	$5B,$18,$8C,$63,$F9,$B3,$FA,$F7,$08,$61,$FE,$68,$ED,$F5,$4F,$E9
	.dc.b	$F9,$FD,$4F,$F9,$00,$1F,$E0,$6A,$9E,$37,$F9,$27,$3F,$F6,$54,$64
	.dc.b	$ED,$C1,$FE,$3B,$AF,$F0,$6F,$27,$E1,$5C,$4F,$F8,$F3,$C3,$FF,$25
	.dc.b	$ED,$88,$FE,$FB,$1F,$E1,$BE,$AF,$F1,$E5,$78,$A7,$B1,$FE,$17,$E1
	.dc.b	$FB,$51,$98,$F1,$7A,$1F,$E3,$D7,$7D,$F5,$F3,$7A,$15,$F6,$4D,$7F
	.dc.b	$C4,$7E,$5F,$F8,$E6,$FA,$47,$C9,$7F,$B3,$61,$FE,$31,$8F,$89,$1F
	.dc.b	$E3,$A8,$C4,$7C,$3C,$39,$4F,$A2,$3F,$57,$3A,$7D,$F9,$7F,$8F,$FB
	.dc.b	$FE,$5C,$7C,$58,$FE,$D1,$3F,$B9,$AF,$1F,$CC,$FF,$9C,$BC,$4F,$F1
	.dc.b	$A5,$FF,$8E,$A3,$FA,$E5,$6E,$BE,$C7,$C7,$FF,$8D,$9F,$5B,$FE,$36
	.dc.b	$E5,$EF,$7F,$E0,$1C,$FF,$C6,$AF,$F9,$12,$BF,$02,$7A,$27,$71,$FE
	.dc.b	$48,$9D,$BF,$42,$7F,$37,$3E,$AE,$1F,$C0,$46,$0F,$AC,$FF,$07,$E0
	.dc.b	$FF,$18,$47,$D9,$CB,$D8,$E0,$F6,$69,$FE,$94,$D3,$FC,$2B,$1F,$A8
	.dc.b	$F2,$BE,$3E,$9E,$6F,$8C,$FC,$24,$62,$7F,$33,$1F,$43,$D0,$F5,$F9
	.dc.b	$FF,$D2,$BD,$E2,$F9,$3F,$F1,$B5,$FF,$8D,$61,$F3,$58,$F9,$B1,$09
	.dc.b	$F0,$59,$70,$6B,$E6,$39,$FC,$1F,$F6,$13,$FB,$17,$C5,$E3,$DB,$FE
	.dc.b	$CF,$ED,$7F,$91,$F3,$FF,$21,$1F,$B0,$F9,$C7,$CF,$3E,$79,$18,$7F
	.dc.b	$C8,$53,$C0,$CC,$EE,$75,$F2,$5F,$E2,$CE,$CF,$D1,$D7,$F8,$CB,$E0
	.dc.b	$FE,$DC,$7F,$05,$F0,$B4,$BA,$3E,$67,$7E,$1F,$9D,$3F,$E3,$3B,$FD
	.dc.b	$2D,$7E,$0D,$FF,$1A,$18,$7A,$3F,$E2,$E3,$0F,$FA,$3F,$E2,$3F,$C6
	.dc.b	$8F,$EF,$27,$FC,$63,$3A,$F0,$FF,$C5,$F3,$FC,$01,$AF,$E3,$FE,$4F
	.dc.b	$8B,$FA,$F9,$FF,$16,$9D,$B1,$FF,$14,$7C,$37,$BA,$7F,$AD,$F9,$5F
	.dc.b	$E2,$A7,$F9,$26,$7E,$EB,$FF,$78,$1F,$C2,$DF,$7B,$83,$F7,$57,$FE
	.dc.b	$2F,$7D,$79,$7F,$87,$B6,$FA,$87,$D5,$F8,$3C,$70,$7D,$33,$ED,$70
	.dc.b	$7F,$86,$39,$7F,$09,$C9,$FE,$7F,$E5,$C3,$F8,$4E,$CF,$F0,$EF,$47
	.dc.b	$E3,$0F,$C8,$F4,$E1,$C7,$4F,$F8,$8B,$AF,$F3,$87,$E6,$F7,$FF,$11
	.dc.b	$76,$FE,$83,$E4,$9F,$08,$F8,$4F,$F4,$B5,$FE,$7A,$FF,$A4,$1F,$EC
	.dc.b	$4E,$5F,$89,$EC,$3C,$3F,$D6,$4B,$C3,$FF,$02,$CE,$63,$F7,$5E,$2F
	.dc.b	$ED,$DF,$E9,$31,$97,$F7,$71,$FE,$65,$F1,$FF,$C4,$3D,$7F,$C7,$FD
	.dc.b	$5F,$E1,$F6,$F2,$BF,$85,$5F,$F4,$3C,$3F,$C4,$07,$D2,$2D,$FE,$25
	.dc.b	$AF,$F3,$77,$FA,$CF,$6F,$68,$C4,$7A,$3D,$EF,$E4,$3A,$FF,$97,$71
	.dc.b	$FE,$51,$5F,$06,$BF,$DE,$B1,$88,$FE,$D5,$E5,$7F,$16,$8C,$42,$3F
	.dc.b	$D1,$91,$FC,$F7,$CB,$D7,$9B,$FC,$B9,$CB,$F9,$3C,$BF,$C4,$58,$7E
	.dc.b	$28,$FF,$13,$61,$FD,$31,$FF,$12,$1F,$AE,$3C,$7E,$EF,$E6,$FC,$2F
	.dc.b	$F0,$F4,$7F,$4D,$F1,$AD,$7E,$FB,$2F,$F1,$24,$FF,$51,$F3,$FF,$C4
	.dc.b	$4F,$E9,$4F,$F3,$37,$B7,$CD,$FF,$34,$7B,$7F,$43,$85,$C7,$F5,$23
	.dc.b	$0F,$F3,$2C,$3F,$C4,$75,$FC,$F2,$BF,$BD,$AF,$90,$FF,$67,$65,$FD
	.dc.b	$E3,$C8,$FF,$12,$CF,$F6,$3F,$53,$FC,$4B,$3F,$DF,$3C,$6F,$E9,$7C
	.dc.b	$6C,$79,$FA,$7F,$89,$AF,$A3,$1F,$E7,$9F,$53,$F1,$BD,$9F,$DA,$39
	.dc.b	$F8,$4B,$E1,$7F,$C3,$DA,$79,$63,$A5,$FF,$93,$63,$F8,$9E,$7F,$E1
	.dc.b	$B7,$FC,$3F,$A7,$FA,$7D,$FE,$C3,$F1,$BF,$EE,$78,$C7,$BB,$FE,$1D
	.dc.b	$7F,$B2,$92,$7F,$C7,$7C,$75,$47,$F6,$AF,$83,$FF,$0D,$C6,$0F,$F5
	.dc.b	$67,$7F,$F9,$1A,$7C,$1F,$23,$FE,$30,$E1,$7E,$0D,$FD,$CB,$DE,$FF
	.dc.b	$1F,$61,$FE,$12,$8C,$47,$F9,$12,$3F,$C2,$5D,$FE,$17,$B1,$FE,$12
	.dc.b	$C3,$FD,$FC,$7F,$85,$FF,$4A,$E3,$5F,$DF,$BE,$4B,$FB,$FE,$29,$CD
	.dc.b	$FF,$90,$A7,$FB,$8C,$FE,$D6,$3E,$39,$0F,$63,$97,$C3,$7F,$05,$FF
	.dc.b	$01,$1F,$E5,$AF,$7B,$FC,$1B,$F2,$9E,$0F,$1B,$D7,$18,$DB,$FC,$05
	.dc.b	$7E,$1B,$FE,$F6,$60,$63,$FE,$01,$9F,$EF,$3E,$57,$F1,$DF,$43,$FB
	.dc.b	$EF,$99,$FC,$83,$C3,$CF,$99,$FC,$82,$3F,$91,$F9,$1F,$E0,$0F,$95
	.dc.b	$10,$9D,$59,$70,$6A,$FF,$C0,$FE,$17,$E1,$F5,$A7,$DC,$3F,$CF,$BE
	.dc.b	$77,$F8,$AB,$E3,$DF,$1C,$F8,$F4,$61,$7C,$9F,$F8,$4A,$30,$67,$D7
	.dc.b	$FF,$39,$F8,$0F,$7F,$EF,$65,$FE,$36,$3F,$BE,$1F,$4C,$FF,$1C,$F2
	.dc.b	$70,$8F,$CB,$72,$FF,$89,$E7,$07,$DA,$7E,$53,$FD,$E8,$FC,$49,$FF
	.dc.b	$00,$69,$EE,$F8,$23,$FC,$67,$EF,$BE,$4F,$11,$FD,$E8,$E2,$7F,$A2
	.dc.b	$B1,$C7,$4F,$E7,$A1,$F0,$1C,$5F,$41,$B3,$35,$DF,$C3,$FF,$49,$CE
	.dc.b	$E7,$A9,$FE,$56,$E3,$F2,$BC,$8C,$65,$E7,$C7,$FB,$C4,$7E,$C7,$4F
	.dc.b	$84,$FF,$7D,$CB,$F1,$CF,$F5,$D9,$DC,$7D,$D8,$D1,$F7,$7C,$F7,$27
	.dc.b	$DD,$E4,$FB,$9C,$5F,$6F,$8F,$FB,$0C,$E7,$93,$EF,$D7,$E0,$79,$33
	.dc.b	$8F,$FA,$BB,$6E,$D9,$74,$8C,$1D,$F0,$FF,$39,$F3,$7E,$48,$FC,$BF
	.dc.b	$3F,$29,$C4,$7E,$55,$FF,$00,$9F,$EB,$AE,$CF,$D2,$7B,$0E,$BF,$D2
	.dc.b	$4E,$8F,$F0,$2F,$67,$E9,$69,$D9,$D7,$8B,$F9,$27,$1F,$F3,$0B,$FC
	.dc.b	$41,$FB,$43,$FC,$13,$1F,$2B,$D8,$FE,$8D,$E0,$7E,$F2,$7F,$D9,$7E
	.dc.b	$0F,$F7,$E8,$C4,$23,$6F,$F0,$BA,$FE,$59,$C1,$FC,$2D,$FE,$59,$E3
	.dc.b	$FF,$83,$3C,$4F,$E2,$B5,$FC,$08,$C1,$FE,$D8,$F1,$7F,$92,$3F,$C1
	.dc.b	$B3,$FE,$DE,$FE,$72,$FF,$51,$83,$C6,$FE,$51,$1F,$E0,$0F,$1F,$F9
	.dc.b	$2E,$1F,$E0,$56,$DF,$F8,$13,$E1,$BF,$28,$FF,$6A,$3F,$21,$8F,$F8
	.dc.b	$83,$C9,$FE,$85,$87,$A9,$B3,$FE,$EC,$D3,$FB,$1C,$FE,$37,$D2,$FE
	.dc.b	$9D,$3F,$AF,$3F,$A5,$9F,$0C,$FF,$79,$FC,$57,$F4,$5D,$BE,$0C,$3C
	.dc.b	$DF,$E9,$1C,$3F,$33,$E4,$71,$3F,$C4,$8F,$F1,$9E,$E2,$16,$BF,$81
	.dc.b	$C7,$E9,$4F,$E8,$0F,$E8,$9F,$9C,$7F,$30,$7F,$21,$C3,$CA,$79,$C7
	.dc.b	$F9,$03,$C6,$F9,$31,$D6,$BF,$C7,$2F,$87,$CC,$FF,$66,$E1,$E6,$C3
	.dc.b	$FA,$85,$7D,$9A,$F3,$3F,$ED,$2C,$3E,$63,$F0,$C6,$7F,$D9,$FF,$2C
	.dc.b	$EF,$5F,$03,$CE,$7B,$B8,$F9,$5F,$ED,$9E,$E6,$BD,$1F,$95,$97,$7C
	.dc.b	$3D,$59,$FC,$A8,$ED,$9F,$F4,$FF,$2B,$FB,$34,$FF,$32,$FD,$C7,$BB
	.dc.b	$FC,$D3,$E1,$AD,$BD,$07,$E6,$D8,$F4,$2C,$F7,$CF,$E3,$AF,$2B,$FD
	.dc.b	$7A,$E6,$3E,$44,$32,$EB,$C7,$FE,$6B,$EA,$FE,$2F,$AF,$F8,$7B,$0F
	.dc.b	$ED,$E7,$E1,$84,$FC,$C7,$EB,$61,$E0,$8F,$F2,$2B,$83,$FD,$A4,$FC
	.dc.b	$88,$F1,$43,$87,$FC,$F7,$C9,$EE,$F5,$7F,$8B,$57,$F8,$EC,$38,$F6
	.dc.b	$84,$2F,$B7,$FE,$0E,$F9,$9F,$F2,$5D,$7F,$87,$DF,$F2,$87,$B3,$F6
	.dc.b	$5F,$F2,$7F,$85,$F7,$67,$FC,$B1,$D4,$CB,$FC,$B0,$F9,$E3,$FD,$08
	.dc.b	$DC,$7C,$07,$87,$D9,$FF,$16,$75,$F0,$3F,$D8,$9E,$FF,$F8,$BF,$C6
	.dc.b	$61,$FE,$6F,$3F,$C4,$6E,$1E,$77,$FD,$B1,$61,$FE,$36,$7C,$B3,$E4
	.dc.b	$3F,$D0,$1A,$7F,$A8,$E7,$53,$E4,$D7,$FC,$AF,$F2,$5F,$E3,$FF,$23
	.dc.b	$B6,$3F,$EA,$6F,$8E,$C5,$F9,$23,$FD,$53,$58,$8F,$6F,$C1,$FF,$80
	.dc.b	$5F,$1D,$FF,$A7,$1F,$F1,$F1,$C6,$BF,$D4,$3E,$6F,$E8,$23,$FD,$1C
	.dc.b	$3E,$0F,$13,$EE,$3F,$A0,$C7,$BB,$C8,$F9,$E7,$D8,$70,$7F,$9B,$7E
	.dc.b	$0B,$FC,$8B,$E4,$FF,$9C,$F7,$FF,$61,$4F,$F9,$02,$FF,$CF,$31,$FE
	.dc.b	$C1,$8C,$43,$3F,$F7,$96,$FF,$E4,$98,$FF,$7A,$CF,$D4,$F3,$FC,$CF
	.dc.b	$D0,$F2,$BF,$DB,$9E,$87,$F5,$CF,$8E,$FF,$34,$46,$3D,$2F,$F6,$EF
	.dc.b	$C7,$62,$FE,$EF,$C3,$FF,$BA,$C7,$C4,$66,$31,$1F,$A1,$D3,$FC,$1D
	.dc.b	$F1,$5F,$EF,$6F,$8E,$7B,$BB,$8C,$7A,$1F,$E7,$7F,$77,$F0,$DE,$7B
	.dc.b	$E3,$3F,$E0,$07,$C7,$3E,$3B,$5A,$FE,$3A,$B5,$26,$FC,$07,$F0,$D9
	.dc.b	$FF,$5B,$7C,$3F,$F0,$17,$FD,$8C,$D7,$FD,$AC,$7F,$BC,$FB,$B8,$38
	.dc.b	$E7,$F8,$F3,$FC,$32,$3F,$DD,$0D,$E8,$FF,$74,$7A,$9F,$C5,$27,$FD
	.dc.b	$D8,$D8,$CF,$A9,$FA,$69,$FF,$1B,$B7,$D4,$D4,$FE,$93,$C1,$FF,$1A
	.dc.b	$CF,$FB,$76,$BF,$8E,$69,$7D,$4F,$05,$EE,$7F,$90,$7A,$9F,$A0,$8F
	.dc.b	$C1,$47,$DF,$D7,$FC,$49,$39,$EF,$FD,$96,$DF,$82,$C3,$E1,$61,$F9
	.dc.b	$7B,$DE,$1F,$EE,$BF,$4B,$F8,$AE,$5F,$38,$F9,$FC,$5F,$E2,$77,$1C
	.dc.b	$6E,$37,$1F,$F5,$9C,$3F,$81,$F3,$7E,$07,$9F,$10,$8F,$C0,$BF,$CB
	.dc.b	$BB,$BF,$83,$75,$FE,$37,$B3,$6F,$E3,$FC,$BE,$3F,$01,$FE,$31,$F0
	.dc.b	$BF,$31,$87,$13,$F8,$8C,$67,$C1,$FC,$CC,$7F,$3A,$F1,$3E,$13,$F1
	.dc.b	$4C,$1F,$BD,$E7,$F8,$BD,$5F,$C6,$B9,$3F,$55,$CC,$F8,$AF,$F7,$2F
	.dc.b	$95,$FE,$31,$C3,$F8,$24,$63,$D0,$FF,$14,$C7,$F3,$7F,$33,$FD,$D7
	.dc.b	$E4,$FF,$21,$C7,$FA,$56,$3F,$CE,$BC,$9F,$F7,$B4,$B0,$E3,$3F,$4F
	.dc.b	$0F,$EA,$3E,$57,$DD,$8F,$E6,$7A,$69,$FE,$27,$3A,$3F,$8B,$3F,$E9
	.dc.b	$B8,$FF,$1B,$BF,$9E,$8A,$FF,$04,$7C,$07,$F5,$EF,$2F,$B6,$9E,$E8
	.dc.b	$C1,$D1,$FE,$5E,$FF,$64,$F3,$7F,$9A,$E6,$DB,$E4,$46,$3C,$DB,$5E
	.dc.b	$4C,$7F,$AD,$78,$18,$8F,$E4,$D1,$FC,$53,$7F,$E7,$AF,$F2,$C8,$EC
	.dc.b	$7F,$4B,$DB,$EA,$73,$7C,$39,$FE,$C9,$B6,$2B,$B4,$7F,$68,$F3,$B1
	.dc.b	$1D,$B1,$B3,$FC,$D7,$CF,$FE,$E7,$3C,$23,$F9,$BE,$5F,$D9,$63,$F4
	.dc.b	$5E,$8F,$F3,$78,$3F,$DB,$23,$FC,$0A,$FF,$70,$F5,$FF,$B0,$FB,$9F
	.dc.b	$DE,$6F,$FA,$FF,$B1,$8F,$3B,$D1,$0E,$8F,$EC,$9E,$8B,$DA,$CC,$FE
	.dc.b	$AB,$3F,$E8,$A7,$E2,$F9,$84,$CF,$FC,$5D,$1F,$E0,$5E,$7C,$CB,$E2
	.dc.b	$BF,$C3,$97,$FD,$21,$FF,$23,$75,$79,$EF,$FC,$3B,$8F,$83,$C6,$FE
	.dc.b	$F3,$EB,$FF,$B1,$70,$FE,$EF,$AF,$F7,$98,$C4,$7F,$7A,$BF,$EF,$FA
	.dc.b	$FF,$BA,$DF,$EB,$46,$3E,$2B,$B4,$66,$3D,$5A,$3B,$BF,$23,$2F,$B4
	.dc.b	$FE,$40,$DA,$FA,$9D,$3D,$BE,$8B,$27,$C6,$E8,$C4,$E8,$FF,$05,$C7
	.dc.b	$F2,$FA,$7E,$2D,$F8,$79,$FF,$7F,$46,$1F,$C6,$1F,$E1,$28,$F0,$9F
	.dc.b	$E2,$43,$F3,$BE,$07,$B9,$FC,$E4,$7F,$86,$3C,$0F,$D3,$E9,$FE,$15
	.dc.b	$6A,$F0,$3F,$C4,$93,$FB,$4F,$0B,$FC,$1D,$8F,$E7,$B8,$3F,$65,$E0
	.dc.b	$FC,$AF,$DD,$3F,$7D,$7F,$E2,$7F,$03,$0F,$EF,$0F,$F1,$57,$06,$1F
	.dc.b	$F1,$BC,$FF,$8E,$8F,$F0,$66,$07,$F2,$CF,$F1,$0C,$3F,$8D,$67,$FE
	.dc.b	$0C,$8C,$75,$7F,$19,$BF,$C7,$D7,$F0,$5F,$4B,$FC,$33,$5C,$74,$FE
	.dc.b	$4B,$5F,$E3,$8F,$53,$FC,$0D,$EC,$3C,$97,$B7,$F8,$49,$CC,$3D,$BC
	.dc.b	$5F,$F8,$2B,$CD,$10,$BE,$76,$A3,$FC,$17,$E5,$5F,$43,$FA,$54,$78
	.dc.b	$4F,$E9,$BF,$24,$F7,$3F,$A6,$37,$5F,$F0,$A7,$93,$FE,$F0,$AF,$F1
	.dc.b	$01,$FE,$09,$F5,$7F,$C1,$BE,$BF,$F8,$67,$D7,$FE,$B7,$EB,$FE,$C7
	.dc.b	$DC,$7E,$C7,$FD,$CD,$B8,$7F,$82,$A3,$FA,$5B,$FE,$09,$11,$A7,$FC
	.dc.b	$15,$D9,$FE,$64,$87,$EF,$FC,$CF,$DF,$C7,$F9,$A3,$DE,$7E,$75,$FE
	.dc.b	$01,$6D,$FE,$75,$AF,$F2,$8F,$5B,$FC,$DF,$EF,$E3,$BB,$FC,$A1,$18
	.dc.b	$BE,$B4,$AD,$57,$F9,$CB,$C2,$FF,$3B,$FB,$DE,$4A,$FE,$45,$18,$CF
	.dc.b	$8C,$FA,$EB,$F0,$93,$C3,$5F,$83,$E0,$7F,$82,$DF,$F0,$B1,$E0,$F6
	.dc.b	$1E,$EF,$F8,$F3,$93,$E7,$E7,$FD,$DC,$FF,$1C,$3F,$E3,$EF,$4B,$FC
	.dc.b	$0B,$CD,$F5,$EF,$FC,$59,$D5,$F8,$06,$F4,$FD,$F8,$3F,$7F,$0F,$F5
	.dc.b	$BF,$81,$FE,$0A,$EE,$FC,$61,$F8,$C8,$C9,$F9,$17,$3D,$BF,$DF,$E7
	.dc.b	$F2,$FE,$0D,$AF,$F0,$5F,$C5,$BD,$E7,$BD,$FE,$0C,$D3,$FC,$73,$C9
	.dc.b	$FE,$A6,$E4,$E1,$87,$F5,$CE,$BF,$EE,$DE,$2F,$F8,$C2,$6C,$FF,$9C
	.dc.b	$E7,$FB,$D7,$8F,$FD,$BE,$3F,$BD,$1F,$E9,$9F,$2B,$FC,$25,$E5,$7F
	.dc.b	$51,$3F,$A8,$65,$FB,$D9,$FA,$E7,$FB,$5A,$3E,$81,$FC,$29,$FF,$59
	.dc.b	$E1,$C9,$FE,$14,$7F,$AD,$B0,$FE,$E5,$97,$F8,$16,$3F,$B7,$9F,$EA
	.dc.b	$43,$FD,$77,$86,$A3,$FD,$23,$E7,$57,$F8,$5F,$97,$FC,$7E,$BF,$DC
	.dc.b	$FE,$77,$F2,$86,$9F,$E1,$4F,$57,$FC,$13,$A7,$F3,$18,$FE,$E8,$68
	.dc.b	$F9,$05,$BF,$F0,$14,$6D,$FF,$69,$C7,$90,$FF,$7A,$FA,$1F,$8F,$3F
	.dc.b	$CD,$D2,$7F,$CD,$4D,$8F,$EF,$DE,$8F,$F8,$F7,$D2,$C9,$F9,$7F,$53
	.dc.b	$FB,$F1,$F3,$8F,$EC,$DE,$DF,$F6,$0C,$7F,$C7,$97,$FD,$FB,$C0,$FE
	.dc.b	$EE,$FF,$B9,$0F,$F0,$57,$7F,$F3,$59,$FF,$1F,$6D,$FD,$DE,$7F,$96
	.dc.b	$C7,$C8,$F6,$5E,$D7,$F1,$FF,$83,$10,$8F,$E3,$1C,$18,$AF,$C3,$D7
	.dc.b	$F7,$D9,$FB,$D1,$FD,$BA,$FA,$AD,$BF,$DF,$95,$FE,$F2,$C7,$FB,$CF
	.dc.b	$C2,$7F,$BD,$7C,$39,$7F,$EF,$A7,$2D,$20,$F8,$5C,$A8,$E2,$BE,$23
	.dc.b	$FC,$0B,$F9,$96,$23,$FE,$00,$7B,$2F,$DA,$F1,$1F,$DD,$24,$CD,$FD
	.dc.b	$E7,$C4,$DF,$FC,$1F,$5C,$27,$FC,$13,$FB,$A7,$07,$75,$B5,$83,$F7
	.dc.b	$DE,$11,$FE,$18,$79,$76,$58,$F0,$57,$83,$E2,$B1,$1D,$23,$E5,$DB
	.dc.b	$F7,$2E,$73,$D9,$E6,$D3,$E3,$9E,$5F,$97,$FF,$13,$FD,$27,$9E,$FE
	.dc.b	$DC,$FF,$85,$23,$11,$F8,$18,$FC,$F5,$BF,$46,$EE,$D8,$EB,$C3,$FE
	.dc.b	$1C,$9F,$9C,$F0,$79,$D3,$E4,$3F,$C2,$31,$87,$E9,$FE,$C5,$D0,$FB
	.dc.b	$3C,$BF,$E1,$66,$E1,$E2,$FA,$6E,$6F,$D6,$9F,$F2,$CC,$62,$3E,$CF
	.dc.b	$1F,$F9,$13,$F4,$AF,$39,$F8,$4C,$FF,$1B,$A2,$31,$5F,$80,$E6,$E0
	.dc.b	$FD,$2C,$FE,$E7,$CB,$7F,$99,$AB,$E2,$C7,$DF,$F9,$4F,$A7,$FA,$0F
	.dc.b	$E6,$35,$FF,$6E,$E1,$FE,$54,$8C,$3F,$8C,$F6,$7F,$C9,$FA,$7E,$1F
	.dc.b	$D0,$FE,$77,$F0,$9F,$E1,$9F,$4B,$F9,$F7,$91,$FE,$5C,$AD,$7A,$4A
	.dc.b	$D5,$FF,$89,$27,$F2,$1F,$30,$21,$3C,$2E,$59,$FE,$72,$32,$6B,$1C
	.dc.b	$C7,$F2,$2E,$3F,$D5,$3F,$E5,$B6,$AF,$73,$F3,$CF,$F9,$2B,$0F,$F6
	.dc.b	$67,$4F,$E9,$B9,$B9,$D7,$F8,$7E,$F8,$FC,$E3,$E6,$9F,$36,$8C,$2F
	.dc.b	$9C,$F1,$C6,$0C,$FC,$C7,$F9,$FB,$F1,$5F,$7E,$FE,$A3,$FD,$2D,$8A
	.dc.b	$33,$1E,$ED,$F8,$6F,$FE,$0E,$3D,$27,$FA,$82,$7D,$3B,$7F,$1C,$DB
	.dc.b	$73,$EF,$37,$E2,$7F,$A3,$BF,$4A,$FD,$F7,$07,$29,$FF,$54,$C7,$2B
	.dc.b	$FE,$49,$9A,$37,$3C,$8F,$E0,$7F,$AC,$61,$E5,$28,$E5,$5F,$C2,$63
	.dc.b	$13,$E9,$3E,$8E,$1F,$EA,$09,$FC,$6C,$3A,$BF,$97,$FC,$4B,$EB,$7F
	.dc.b	$82,$B6,$7F,$81,$9F,$D6,$5C,$3F,$78,$FE,$0C,$17,$D9,$72,$3F,$86
	.dc.b	$FC,$4E,$27,$99,$FC,$85,$C3,$CC,$FE,$CC,$F6,$E7,$FF,$31,$36,$3E
	.dc.b	$9F,$91,$FE,$4A,$C3,$F8,$DF,$A3,$FD,$4A,$7F,$A0,$65,$FC,$B4,$FE
	.dc.b	$97,$1F,$12,$3E,$B1,$F4,$DC,$70,$7F,$8E,$77,$F9,$7A,$F0,$E0,$FA
	.dc.b	$17,$F8,$2B,$FA,$CF,$CD,$3F,$9D,$6F,$FD,$0A,$33,$1F,$D6,$6B,$F3
	.dc.b	$CF,$F5,$BE,$2E,$6F,$E9,$27,$F2,$4F,$33,$FC,$0C,$E1,$E5,$3F,$8F
	.dc.b	$79,$1F,$B5,$E4,$FC,$E7,$57,$86,$FF,$37,$1F,$B7,$9F,$ED,$5C,$1F
	.dc.b	$D3,$F9,$FF,$4D,$D5,$FA,$CE,$9F,$44,$E6,$3F,$5F,$3F,$AA,$AE,$27
	.dc.b	$94,$FE,$B1,$9F,$EF,$B7,$FF,$30,$5F,$C8,$E7,$E7,$5F,$E3,$EA,$FE
	.dc.b	$07,$E8,$88,$4F,$73,$73,$FB,$5D,$3F,$8A,$6D,$87,$1D,$5A,$ED,$FE
	.dc.b	$43,$F2,$6F,$07,$F9,$0E,$1E,$63,$FC,$C1,$E0,$FF,$21,$9D,$53,$DE
	.dc.b	$FA,$DE,$27,$E8,$BC,$4F,$5B,$7A,$3E,$D7,$8F,$FC,$E3,$2F,$EF,$BE
	.dc.b	$47,$C0,$8C,$43,$C7,$FC,$6C,$7F,$44,$8F,$E4,$7C,$7F,$D1,$DF,$F0
	.dc.b	$8F,$2F,$F7,$FF,$27,$AC,$7E,$88,$FC,$47,$95,$FD,$4F,$C0,$F9,$AF
	.dc.b	$F5,$58,$FC,$CF,$97,$FD,$71,$FA,$2D,$D3,$F2,$0F,$E8,$BC,$CF,$EC
	.dc.b	$D3,$F8,$D6,$A9,$FC,$B4,$7F,$70,$8F,$B6,$D8,$CC,$79,$4F,$F1,$97
	.dc.b	$9B,$FD,$D3,$C1,$FD,$13,$63,$F2,$FE,$77,$F8,$2B,$5F,$F0,$74,$7E
	.dc.b	$FB,$C6,$FD,$24,$FE,$DB,$D0,$E5,$87,$F9,$23,$D0,$FD,$44,$62,$3F
	.dc.b	$6E,$F3,$3F,$7B,$E8,$E7,$C1,$FD,$77,$A5,$FC,$00,$F0,$7A,$79,$CF
	.dc.b	$FC,$90,$F8,$F8,$BE,$A7,$27,$ED,$5F,$AD,$C9,$6B,$EA,$9F,$E3,$AF
	.dc.b	$53,$FC,$77,$E8,$7F,$B9,$23,$FC,$5B,$AF,$EC,$61,$A7,$9F,$6F,$F1
	.dc.b	$7C,$FF,$8E,$BD,$3F,$F1,$3D,$F0,$8F,$E7,$BC,$BE,$5C,$1C,$F7,$A3
	.dc.b	$F1,$19,$FF,$80,$70,$FE,$32,$7F,$85,$F8,$7F,$DD,$DF,$41,$ED,$F4
	.dc.b	$9F,$90,$7B,$FD,$2B,$7D,$F5,$FF,$25,$42,$FE,$27,$4B,$BF,$FA,$0F
	.dc.b	$27,$E6,$30,$F8,$27,$F9,$94,$FF,$38,$6D,$C3,$83,$8F,$5F,$2E,$9F
	.dc.b	$AD,$C7,$FC,$7F,$97,$E1,$61,$FE,$77,$F8,$FF,$73,$FE,$20,$DF,$FC
	.dc.b	$88,$FF,$97,$0C,$3F,$E3,$DE,$2F,$9F,$18,$8F,$CB,$3F,$E6,$29,$FD
	.dc.b	$7B,$FE,$67,$E5,$FA,$8F,$FA,$3F,$D7,$FE,$BA,$FF,$9A,$3A,$B6,$FF
	.dc.b	$9E,$8F,$F3,$2F,$CE,$7D,$98,$E7,$DF,$F7,$B3,$FF,$2C,$CF,$DD,$78
	.dc.b	$D7,$F9,$C9,$DB,$4F,$F1,$F1,$FD,$83,$0F,$F3,$B3,$DB,$2F,$F5,$17
	.dc.b	$57,$E1,$15,$FE,$27,$6E,$B7,$C2,$FD,$47,$FA,$3A,$3D,$FD,$FC,$74
	.dc.b	$FF,$14,$FE,$03,$F8,$CA,$FF,$57,$61,$FD,$7D,$F8,$9C,$3E,$AC,$BF
	.dc.b	$D2,$EF,$63,$D3,$91,$FF,$29,$F2,$7F,$87,$23,$15,$EC,$AF,$F0,$57
	.dc.b	$1F,$C4,$CB,$FD,$42,$D9,$FF,$3C,$1F,$08,$A7,$A8,$FF,$3D,$1F,$E6
	.dc.b	$3F,$13,$FC,$F2,$FF,$99,$70,$FF,$0F,$F2,$FD,$08,$FF,$54,$F8,$7F
	.dc.b	$E7,$47,$EB,$98,$E2,$FD,$9C,$FC,$83,$D6,$7E,$1E,$32,$7F,$9C,$72
	.dc.b	$FF,$69,$4F,$F9,$D2,$3F,$DA,$9E,$E7,$F7,$67,$13,$FE,$D8,$8F,$EF
	.dc.b	$33,$88,$FF,$48,$4F,$C5,$EF,$F9,$3E,$BF,$D8,$3E,$74,$FF,$8B,$DF
	.dc.b	$9B,$E1,$7F,$9C,$3E,$97,$FA,$AF,$C4,$7F,$99,$BC,$A7,$98,$9F,$EB
	.dc.b	$3F,$31,$FE,$6A,$F4,$3F,$AF,$FD,$F0,$9C,$19,$87,$A5,$99,$D7,$47
	.dc.b	$F9,$CB,$D4,$FE,$D9,$F1,$2F,$80,$83,$E2,$35,$29,$CC,$BE,$6B,$FC
	.dc.b	$F7,$E4,$FF,$BE,$BB,$7F,$C6,$5D,$DF,$E7,$A8,$C7,$27,$FB,$88,$FF
	.dc.b	$03,$7E,$1A,$F9,$77,$CC,$3E,$51,$C9,$FE,$74,$F7,$D7,$BA,$D7,$D9
	.dc.b	$7F,$9A,$39,$7E,$0F,$27,$C0,$EE,$FF,$70,$FC,$17,$F9,$E3,$9F,$E0
	.dc.b	$4F,$E5,$B8,$3F,$C4,$9C,$1E,$4B,$FF,$46,$C6,$3B,$BF,$D9,$1C,$3F
	.dc.b	$EF,$8C,$7E,$BC,$7F,$B8,$BE,$0F,$FD,$23,$C7,$EE,$6D,$7F,$BF,$BE
	.dc.b	$12,$E3,$D1,$AF,$F1,$4C,$7F,$0A,$FE,$7C,$FE,$2F,$EB,$7F,$2F,$7F
	.dc.b	$48,$7F,$BF,$5C,$3B,$F0,$FF,$9D,$BA,$F2,$EB,$CF,$CB,$3F,$E6,$BF
	.dc.b	$5B,$F8,$9D,$FF,$3E,$8E,$43,$D6,$C6,$DF,$66,$3F,$D2,$F1,$88,$F3
	.dc.b	$C7,$F2,$C8,$F1,$C6,$E3,$BE,$1E,$09,$CC,$F0,$C3,$1E,$B7,$F2,$29
	.dc.b	$CE,$FF,$E6,$93,$F9,$2F,$0F,$F9,$B4,$FA,$98,$FF,$91,$B6,$F9,$3A
	.dc.b	$FF,$99,$78,$3E,$5C,$D3,$FC,$D1,$9F,$D8,$7F,$CC,$3C,$3F,$63,$D4
	.dc.b	$FF,$2E,$46,$B2,$FF,$43,$65,$F7,$DF,$F6,$7E,$3A,$9F,$F6,$D7,$1F
	.dc.b	$E9,$FC,$3C,$73,$7E,$3A,$3F,$8F,$E1,$FE,$CE,$EA,$FC,$AF,$6F,$E6
	.dc.b	$BB,$7F,$33,$8F,$F9,$2F,$E0,$FF,$DA,$5D,$BF,$99,$97,$67,$F9,$63
	.dc.b	$1F,$7D,$DE,$1F,$53,$FD,$6A,$3F,$CC,$3E,$17,$EC,$7C,$1F,$DA,$F8
	.dc.b	$7F,$E6,$08,$FD,$CC,$7E,$E3,$C2,$FF,$30,$70,$7F,$00,$E0,$F5,$63
	.dc.b	$FE,$5D,$8F,$57,$89,$FE,$63,$F1,$BF,$86,$1F,$E5,$FF,$23,$1E,$3F
	.dc.b	$F0,$EB,$6B,$FE,$B8,$DB,$F0,$11,$FE,$7A,$F2,$7E,$C9,$F3,$DD,$F9
	.dc.b	$5F,$D8,$AB,$FC,$B5,$1F,$90,$6D,$7F,$29,$8F,$CB,$F4,$7F,$99,$5F
	.dc.b	$CC,$C6,$5F,$F2,$84,$7F,$97,$9F,$E6,$D1,$FD,$66,$7F,$C4,$71,$88
	.dc.b	$F9,$1D,$1E,$13,$F1,$A7,$F4,$78,$FF,$38,$1C,$CF,$EC,$67,$F4,$73
	.dc.b	$DF,$E7,$FE,$A9,$FF,$03,$79,$FF,$EB,$98,$FE,$C3,$43,$FD,$8A,$FB
	.dc.b	$CF,$CD,$9F,$D7,$E7,$ED,$35,$FF,$5F,$6D,$FD,$A9,$FF,$5D,$3F,$E8
	.dc.b	$79,$E9,$D9,$A9,$FE,$06,$53,$E5,$4F,$F7,$77,$89,$FE,$09,$3F,$D8
	.dc.b	$3E,$8F,$F8,$09,$FB,$2F,$F8,$8B,$C3,$FF,$33,$C7,$43,$27,$F1,$0D
	.dc.b	$32,$FF,$81,$B0,$FF,$09,$BA,$30,$F9,$9F,$C2,$FA,$7F,$BA,$3C,$1A
	.dc.b	$FF,$97,$A3,$B3,$5E,$33,$FE,$EC,$F5,$3F,$0C,$7F,$02,$3F,$15,$3F
	.dc.b	$E3,$68,$FC,$7C,$FF,$B9,$7D,$5F,$E0,$EF,$F7,$93,$F2,$A7,$F9,$07
	.dc.b	$0F,$F7,$53,$FE,$46,$F0,$5A,$FF,$7E,$8F,$BE,$37,$CA,$F2,$BF,$16
	.dc.b	$FF,$91,$71,$FE,$B9,$5F,$DF,$7D,$1F,$C2,$7E,$17,$C2,$FF,$AF,$53
	.dc.b	$17,$2F,$E1,$63,$5E,$2F,$F4,$88,$C6,$3F,$E8,$6F,$1B,$15,$F2,$F0
	.dc.b	$FE,$44,$F3,$AD,$E7,$FB,$2F,$1F,$73,$E8,$E0,$61,$8C,$7F,$8D,$47
	.dc.b	$F7,$3C,$38,$BE,$6C,$FC,$B7,$F6,$63,$F6,$6B,$5F,$6D,$FE,$96,$3F
	.dc.b	$BE,$1F,$0B,$2F,$F3,$E9,$FE,$2E,$3E,$54,$7F,$50,$FC,$E5,$F6,$BF
	.dc.b	$EA,$89,$46,$1F,$F4,$3F,$B7,$B9,$FF,$55,$C7,$D1,$FB,$4F,$D8,$7A
	.dc.b	$3F,$B1,$8F,$F1,$87,$03,$F5,$0C,$C6,$4F,$F3,$E5,$7F,$A9,$63,$31
	.dc.b	$FE,$7C,$97,$E5,$DF,$EA,$9F,$CB,$F3,$3F,$E5,$96,$BF,$C4,$8F,$C1
	.dc.b	$3F,$E9,$49,$3F,$EB,$1F,$CD,$37,$3F,$E8,$B6,$F4,$7F,$B0,$1C,$FA
	.dc.b	$7F,$EC,$2F,$37,$FD,$BD,$5F,$E9,$16,$D7,$FB,$1B,$F4,$0E,$73,$4F
	.dc.b	$EF,$3A,$FC,$13,$FC,$2B,$1F,$EE,$D8,$FC,$B3,$FE,$F4,$78,$1F,$8F
	.dc.b	$E2,$FF,$06,$D7,$FA,$56,$3F,$DE,$59,$FE,$C6,$FF,$8E,$7B,$BF,$E7
	.dc.b	$9E,$FF,$8E,$FC,$EF,$D9,$3F,$90,$63,$FE,$B0,$8F,$F5,$44,$62,$3F
	.dc.b	$D1,$18,$7E,$87,$F6,$6F,$4B,$FA,$97,$F0,$18,$9F,$BA,$D8,$FF,$40
	.dc.b	$7C,$47,$FA,$CE,$7A,$75,$7F,$2F,$F8,$0F,$D6,$7D,$C7,$4C,$0F,$E0
	.dc.b	$33,$FE,$69,$E1,$7F,$AE,$E3,$19,$7F,$88,$23,$FC,$5D,$E5,$7D,$48
	.dc.b	$FD,$E3,$FC,$EF,$E0,$7F,$CD,$7F,$08,$F8,$B7,$C6,$25,$F1,$BF,$E8
	.dc.b	$17,$73,$A5,$F1,$BF,$E8,$5F,$50,$42,$7C,$1E,$5C,$1A,$BF,$F4,$67
	.dc.b	$B2,$21,$2B,$2E,$0D,$7C,$77,$FA,$4F,$E3,$FF,$41,$B7,$FB,$6F,$6F
	.dc.b	$EA,$DE,$4F,$E0,$7A,$3F,$85,$47,$F5,$9A,$DF,$67,$FA,$6F,$DF,$7C
	.dc.b	$0B,$E4,$D1,$85,$F2,$3B,$9D,$4E,$7E,$4F,$FC,$9B,$D4,$F8,$EF,$EC
	.dc.b	$DC,$7E,$E3,$E1,$F2,$6A,$33,$19,$E4,$C7,$0F,$F9,$5B,$6F,$5B,$83
	.dc.b	$FC,$1F,$C1,$EB,$F3,$B1,$3E,$A9,$FE,$6D,$FB,$47,$C0,$AD,$46,$0E
	.dc.b	$87,$F0,$5E,$1F,$3F,$EA,$3F,$0E,$3B,$EA,$8E,$55,$FA,$9C,$77,$5C
	.dc.b	$8F,$E2,$57,$FE,$6C,$F8,$AF,$E8,$BE,$37,$A2,$51,$87,$C0,$63,$DE
	.dc.b	$FE,$06,$EB,$F0,$EF,$D3,$C9,$FC,$B7,$C5,$59,$FE,$02,$7C,$47,$0A
	.dc.b	$FE,$14,$77,$8F,$29,$FE,$7B,$BE,$F1,$98,$FE,$4D,$D3,$EF,$7D,$86
	.dc.b	$34,$F8,$8F,$28,$D3,$F4,$B4,$F8,$8F,$F9,$0B,$6D,$BF,$EE,$39,$47
	.dc.b	$CA,$9F,$E9,$FC,$5A,$AF,$E7,$55,$F8,$77,$C2,$7F,$2F,$70,$F7,$9E
	.dc.b	$2F,$FA,$6B,$8B,$F1,$87,$FB,$69,$FF,$29,$D7,$E3,$23,$F5,$BC,$5F
	.dc.b	$EA,$0B,$FF,$27,$F2,$FE,$6C,$FC,$E7,$3E,$7A,$7F,$C7,$91,$9E,$8F
	.dc.b	$D1,$74,$67,$3F,$29,$FE,$53,$1C,$FF,$91,$C7,$FC,$B6,$7E,$BF,$5F
	.dc.b	$EE,$58,$FF,$91,$FA,$FF,$70,$7E,$F7,$B3,$A6,$18,$8F,$DD,$BF,$E6
	.dc.b	$63,$DD,$B7,$EA,$7A,$B6,$FC,$0D,$F8,$F6,$FF,$12,$EF,$FE,$2D,$18
	.dc.b	$F8,$37,$83,$FC,$6E,$7F,$8D,$67,$E0,$E9,$F4,$78,$7F,$C9,$B4,$FF
	.dc.b	$10,$1F,$1D,$C7,$87,$FE,$8F,$7F,$9D,$F0,$FF,$37,$9F,$F0,$D7,$8D
	.dc.b	$FE,$63,$F2,$3F,$87,$F9,$1F,$CE,$27,$FD,$9F,$E4,$7F,$84,$E3,$FC
	.dc.b	$01,$7C,$72,$FE,$E7,$F2,$5F,$56,$7A,$F9,$5F,$D2,$F9,$FF,$A9,$E3
	.dc.b	$F6,$63,$0F,$F9,$C7,$C5,$FB,$9E,$5F,$F5,$68,$FC,$5B,$5F,$F5,$33
	.dc.b	$E8,$F3,$3F,$B2,$47,$FA,$61,$ED,$B7,$E5,$DF,$F0,$0E,$7F,$94,$7F
	.dc.b	$B6,$68,$7E,$0F,$9B,$FD,$E2,$FF,$18,$FF,$78,$7F,$D4,$7E,$7F,$F4
	.dc.b	$DA,$3F,$BE,$6C,$7E,$E8,$FE,$6E,$FF,$82,$9F,$F4,$EE,$9F,$CD,$FE
	.dc.b	$4B,$F3,$1E,$8F,$C3,$8F,$F1,$BB,$D4,$FA,$7E,$96,$63,$FC,$71,$E9
	.dc.b	$FD,$63,$C5,$EA,$66,$3F,$C8,$8F,$F6,$1F,$B8,$FF,$18,$63,$FC,$24
	.dc.b	$F2,$6C,$51,$A8,$FA,$F1,$8F,$5B,$FC,$57,$1F,$EA,$3F,$5F,$FA,$0F
	.dc.b	$EF,$3D,$F9,$3F,$A1,$E9,$FE,$27,$D7,$C5,$DD,$FE,$21,$CB,$DF,$AF
	.dc.b	$F0,$AF,$4F,$E7,$46,$23,$FC,$BB,$0E,$7E,$93,$FE,$0B,$3F,$C8,$E5
	.dc.b	$3F,$61,$3F,$E0,$67,$F9,$3F,$AF,$FE,$1C,$CB,$F9,$B7,$E8,$BC,$BD
	.dc.b	$47,$E5,$9D,$3D,$3B,$87,$87,$1F,$F4,$D4,$2F,$C3,$CF,$FE,$10,$AE
	.dc.b	$A7,$C7,$CB,$F3,$CF,$1B,$AF,$CD,$27,$0F,$E6,$DD,$67,$FC,$57,$AB
	.dc.b	$AF,$57,$6E,$1F,$8F,$5F,$EE,$EE,$2F,$B5,$D3,$FC,$6F,$93,$E1,$FD
	.dc.b	$93,$F1,$5F,$F1,$7A,$F3,$FF,$C6,$95,$FE,$04,$8D,$61,$FD,$2B,$D8
	.dc.b	$C7,$E3,$BC,$30,$FF,$1D,$6D,$8A,$FE,$9F,$E2,$F1,$EA,$FD,$8C,$7F
	.dc.b	$9C,$71,$FE,$1A,$FD,$AF,$77,$FC,$E1,$0F,$C5,$BC,$67,$E1,$A7,$83
	.dc.b	$F8,$A7,$FD,$D5,$1B,$7F,$DA,$0D,$7F,$86,$E7,$D3,$BB,$DA,$7F,$17
	.dc.b	$8E,$2F,$F9,$07,$1F,$F4,$07,$E3,$BF,$E8,$5F,$3F,$F9,$21,$FE,$38
	.dc.b	$A4,$67,$CE,$FE,$4B,$EF,$FF,$8E,$7D,$FF,$D0,$4F,$FA,$37,$D2,$FE
	.dc.b	$5D,$7C,$0F,$E9,$8F,$F8,$EF,$EC,$B7,$1F,$CF,$7D,$5F,$E7,$B7,$88
	.dc.b	$FD,$54,$7F,$3D,$F5,$31,$1F,$D2,$0F,$F0,$EE,$1F,$D8,$BB,$BF,$80
	.dc.b	$47,$73,$FA,$7F,$17,$0F,$67,$FA,$A7,$C0,$7D,$D9,$D1,$FE,$36,$F6
	.dc.b	$BF,$C6,$67,$F8,$E7,$ED,$3D,$7E,$1F,$A3,$7F,$F1,$8D,$7C,$28,$C4
	.dc.b	$7C,$5C,$D1,$FD,$D6,$7F,$81,$1B,$3F,$B5,$67,$FE,$2C,$9D,$C7,$F6
	.dc.b	$4E,$5F,$93,$EB,$71,$8A,$70,$AF,$F0,$3D,$7F,$4E,$8F,$84,$7F,$85
	.dc.b	$97,$AB,$C6,$3E,$BB,$FE,$09,$91,$FE,$F4,$9D,$CF,$84,$FB,$BF,$09
	.dc.b	$FE,$90,$8F,$F7,$7C,$7F,$2F,$3D,$78,$7D,$E9,$FE,$BB,$14,$FD,$D9
	.dc.b	$F0,$BB,$30,$FF,$25,$CB,$FD,$1B,$5F,$86,$FC,$7F,$B1,$FE,$8B,$FA
	.dc.b	$63,$FE,$A0,$6D,$78,$4F,$F6,$53,$E4,$7A,$1F,$E9,$53,$F9,$83,$9F
	.dc.b	$43,$F8,$64,$7F,$A6,$6B,$38,$FF,$77,$C1,$8E,$2B,$B1,$F6,$30,$E4
	.dc.b	$FF,$94,$1F,$8E,$FD,$03,$FD,$CC,$E3,$DE,$FF,$35,$3A,$F2,$FF,$C3
	.dc.b	$55,$E4,$DF,$EA,$4F,$D9,$AD,$C7,$D5,$DB,$C7,$82,$BF,$0A,$F7,$3F
	.dc.b	$DF,$0E,$2B,$BE,$3C,$70,$F8,$2F,$E5,$3D,$DB,$7F,$E2,$2E,$1F,$F1
	.dc.b	$DB,$D4,$FF,$1B,$F1,$FE,$83,$93,$F4,$99,$7E,$62,$7F,$49,$13,$93
	.dc.b	$E6,$C7,$A8,$FC,$F7,$36,$1F,$CB,$6B,$F8,$E1,$E2,$E6,$3F,$C6,$DD
	.dc.b	$2E,$AC,$3F,$BD,$8F,$21,$E7,$D3,$FD,$81,$D5,$C5,$FF,$2B,$F5,$F3
	.dc.b	$EE,$FE,$2B,$37,$BB,$F8,$9F,$BB,$FE,$CF,$CB,$FB,$97,$D2,$FF,$71
	.dc.b	$CB,$E5,$9F,$B4,$71,$C8,$F0,$7E,$04,$A7,$83,$FE,$C4,$7F,$96,$6F
	.dc.b	$FA,$AF,$67,$1E,$2F,$FB,$BB,$B7,$8D,$7F,$31,$EE,$FE,$02,$62,$7F
	.dc.b	$9C,$61,$E4,$7F,$DF,$B3,$FD,$0A,$3E,$46,$1D,$FC,$6F,$F1,$D7,$2F
	.dc.b	$4F,$23,$67,$4A,$F4,$1E,$0F,$23,$FA,$8B,$4F,$CD,$9F,$09,$E1,$9F
	.dc.b	$AF,$C9,$FE,$BB,$3F,$D4,$30,$FF,$24,$74,$7F,$52,$D3,$FC,$95,$DD
	.dc.b	$FD,$13,$E6,$FF,$C1,$3E,$2B,$F0,$8F,$F0,$5F,$7D,$CB,$FB,$99,$F9
	.dc.b	$9A,$FF,$07,$F0,$FC,$8E,$1F,$E2,$38,$FC,$6D,$3B,$DF,$E3,$F2,$FE
	.dc.b	$99,$1F,$E6,$FE,$A7,$5F,$F4,$0E,$DF,$F3,$F3,$FC,$EF,$C0,$FC,$C7
	.dc.b	$85,$FC,$DB,$1F,$C4,$77,$8E,$2F,$F3,$13,$FD,$45,$A7,$F2,$2E,$2E
	.dc.b	$11,$FE,$20,$CF,$F9,$7B,$A9,$FC,$CB,$FC,$A8,$FC,$EB,$C5,$FE,$C7
	.dc.b	$BF,$F8,$31,$FE,$CB,$C3,$78,$7E,$24,$7F,$8B,$30,$FE,$2B,$59,$F1
	.dc.b	$FF,$80,$47,$F0,$59,$EB,$87,$82,$FF,$80,$BF,$EA,$3C,$7F,$77,$3E
	.dc.b	$5A,$B9,$7E,$92,$3F,$52,$7E,$E3,$C8,$6A,$F0,$FF,$B4,$65,$FB,$DE
	.dc.b	$0F,$97,$7C,$38,$7F,$BF,$F0,$FF,$6E,$CB,$FD,$8B,$D7,$F1,$9F,$F6
	.dc.b	$3F,$BD,$FC,$F2,$3F,$96,$DF,$F6,$0D,$7F,$95,$3F,$CD,$0C,$1F,$C6
	.dc.b	$FB,$FF,$B3,$D7,$EE,$EF,$EB,$12,$FF,$CD,$1E,$63,$F6,$6F,$F1,$99
	.dc.b	$FD,$6D,$C6,$1F,$E6,$DF,$17,$F6,$EF,$F8,$8A,$7F,$B7,$F6,$62,$3F
	.dc.b	$C2,$7E,$F7,$F5,$B3,$F5,$0F,$F6,$77,$FA,$CB,$FE,$1C,$7F,$A2,$75
	.dc.b	$F0,$8F,$EB,$7E,$4B,$BF,$FC,$3B,$F4,$BF,$D9,$4F,$EB,$07,$F5,$73
	.dc.b	$FA,$D5,$F4,$E4,$3F,$EC,$29,$FE,$AA,$F6,$BF,$E7,$CF,$F5,$F3,$CA
	.dc.b	$FF,$6E,$3F,$89,$65,$FD,$EF,$BF,$E0,$C7,$F8,$64,$FE,$29,$1F,$E0
	.dc.b	$ED,$3E,$04,$70,$3F,$80,$9F,$E3,$8F,$53,$F7,$3F,$2B,$FB,$83,$F7
	.dc.b	$2F,$FA,$13,$E4,$7F,$C5,$1C,$3F,$33,$3F,$F4,$D7,$83,$8C,$3F,$92
	.dc.b	$C7,$FA,$76,$9C,$5F,$E8,$CA,$B5,$FC,$0B,$EF,$FF,$C1,$EF,$F6,$CF
	.dc.b	$03,$FC,$23,$18,$8F,$F3,$E1,$FE,$E1,$1E,$87,$F7,$07,$FD,$00,$FF
	.dc.b	$86,$0C,$65,$FC,$85,$FF,$56,$C7,$FA,$1F,$C1,$D7,$A1,$FE,$B7,$5F
	.dc.b	$49,$D2,$31,$E5,$DA,$E1,$5E,$0C,$FE,$8C,$7F,$8E,$3B,$30,$FC,$AC
	.dc.b	$5F,$77,$FE,$C3,$FC,$83,$C8,$FF,$AB,$23,$33,$D1,$7E,$05,$D2,$3F
	.dc.b	$9B,$C6,$1F,$89,$8F,$C0,$9F,$F6,$77,$E9,$BE,$F7,$F8,$65,$6E,$BF
	.dc.b	$D6,$B5,$F0,$FF,$0C,$E9,$C9,$FE,$CF,$A4,$77,$F9,$8F,$91,$F8,$17
	.dc.b	$FB,$BF,$0F,$9A,$FC,$5C,$3E,$7F,$E0,$DF,$EF,$89,$FF,$2B,$E0,$8F
	.dc.b	$F7,$EC,$7F,$96,$B9,$76,$FF,$B4,$38,$7F,$C0,$DE,$06,$23,$CB,$87
	.dc.b	$DA,$F0,$BF,$5F,$C5,$AC,$BF,$31,$CD,$F3,$6F,$E7,$43,$F0,$4F,$F5
	.dc.b	$2C,$7F,$82,$63,$0F,$FB,$78,$FF,$35,$E0,$CB,$F0,$F5,$EB,$A7,$D0
	.dc.b	$7F,$A9,$EF,$FD,$E9,$18,$7F,$DE,$A7,$C7,$BF,$C7,$CF,$47,$A4,$7F
	.dc.b	$BE,$74,$F1,$4E,$70,$F8,$55,$E1,$BE,$97,$F0,$1F,$EF,$58,$FF,$24
	.dc.b	$D3,$E5,$F9,$DF,$E4,$FB,$FC,$D6,$1F,$E8,$4F,$2B,$FD,$5B,$EC,$AB
	.dc.b	$F4,$33,$98,$FF,$0D,$C9,$F6,$E3,$E6,$C9,$3C,$BD,$3E,$31,$FD,$8B
	.dc.b	$C9,$FE,$5B,$5B,$7F,$97,$93,$DB,$DC,$E2,$33,$F9,$1E,$3C,$FF,$F1
	.dc.b	$BF,$D6,$D7,$FC,$73,$87,$85,$FF,$1D,$FD,$A7,$2F,$07,$F2,$B3,$E5
	.dc.b	$7F,$C3,$7C,$1F,$E4,$3D,$FB,$F8,$3F,$C5,$AB,$D5,$EA,$7F,$AD,$07
	.dc.b	$CC,$79,$1C,$7C,$C5,$F5,$BF,$CA,$D1,$FD,$72,$B8,$C6,$3F,$57,$FF
	.dc.b	$1A,$70,$8F,$EC,$51,$FE,$BC,$CF,$7E,$BF,$F6,$5A,$DD,$6A,$18,$78
	.dc.b	$AB,$6F,$F6,$A8,$CE,$3C,$33,$FE,$5F,$5E,$F3,$F6,$BA,$7F,$9E,$DF
	.dc.b	$CB,$1F,$1F,$71,$09,$BF,$89,$FF,$32,$94,$D1,$BE,$0F,$3F,$17,$D7
	.dc.b	$3F,$B4,$D7,$DD,$71,$C5,$A7,$5C,$7E,$EE,$5F,$C3,$57,$E2,$5F,$C8
	.dc.b	$F3,$E7,$9F,$F1,$E7,$EA,$A7,$F3,$7C,$3F,$E9,$2E,$8F,$D8,$D7,$E6
	.dc.b	$E7,$FD,$97,$D5,$FC,$2D,$FD,$19,$F3,$3A,$FF,$CB,$71,$FA,$9B,$FD
	.dc.b	$64,$FC,$48,$FD,$B7,$5F,$D7,$8F,$D5,$BF,$DD,$BA,$3F,$D1,$5D,$9D
	.dc.b	$4F,$E0,$B3,$8B,$FE,$08,$71,$E3,$FF,$6E,$F7,$7F,$10,$8F,$F7,$8E
	.dc.b	$98,$7F,$09,$8F,$F1,$17,$F0,$DD,$DF,$EE,$4E,$FF,$51,$FC,$82,$3C
	.dc.b	$ED,$55,$B7,$FD,$C9,$5F,$EA,$7C,$5E,$07,$C3,$3F,$98,$3F,$CD,$27
	.dc.b	$FD,$9B,$E1,$FB,$FC,$3F,$E6,$D3,$FE,$F4,$F1,$3E,$51,$FD,$09,$C5
	.dc.b	$7C,$57,$F5,$7E,$B7,$C7,$9F,$E9,$9E,$F5,$AF,$E9,$75,$FD,$49,$FE
	.dc.b	$66,$D3,$F6,$63,$C9,$FF,$5B,$79,$3F,$C8,$63,$FD,$23,$EB,$7D,$3F
	.dc.b	$03,$FA,$C5,$B8,$54,$9C,$CE,$3C,$7F,$C1,$1F,$DC,$7C,$BC,$47,$F7
	.dc.b	$5E,$5E,$3F,$21,$F8,$7E,$8F,$C0,$CF,$F7,$D3,$F9,$A9,$FE,$02,$C5
	.dc.b	$F7,$3F,$E0,$7E,$5F,$F1,$FF,$07,$EF,$63,$15,$FE,$B8,$EB,$FF,$06
	.dc.b	$DF,$F6,$F7,$F9,$B6,$1F,$D0,$CF,$E7,$18,$FF,$80,$96,$9F,$E0,$7F
	.dc.b	$9E,$FF,$08,$F0,$FF,$7D,$D6,$E5,$E9,$BF,$EE,$59,$7C,$AF,$3B,$FA
	.dc.b	$17,$27,$3E,$8F,$95,$87,$F2,$0B,$FC,$1E,$1C,$7A,$BF,$C7,$B1,$FE
	.dc.b	$BE,$3F,$B7,$47,$F6,$B3,$FB,$1C,$79,$7B,$3F,$3F,$C7,$FD,$7A,$7C
	.dc.b	$5D,$BF,$AB,$5C,$3F,$52,$7F,$AD,$C7,$9F,$1F,$E4,$4F,$F0,$7F,$0B
	.dc.b	$FD,$49,$F9,$2F,$F3,$BF,$1B,$F8,$68,$F9,$9F,$E8,$7C,$3F,$86,$1C
	.dc.b	$7F,$C9,$9C,$3F,$C5,$67,$8C,$FF,$0F,$BF,$E9,$57,$FC,$13,$0F,$C0
	.dc.b	$61,$FC,$10,$FE,$B1,$9F,$9D,$FE,$FB,$1F,$CA,$1A,$7E,$38,$FD,$66
	.dc.b	$7F,$AF,$8D,$6D,$FB,$3E,$AC,$74,$FB,$A7,$F1,$17,$C7,$8B,$FC,$7E
	.dc.b	$FD,$28,$FF,$76,$EF,$FA,$2E,$3C,$9F,$E3,$07,$F3,$BC,$BC,$0F,$C1
	.dc.b	$57,$E1,$AB,$FC,$9D,$C3,$F9,$A7,$DB,$C9,$F8,$DB,$FC,$BD,$7E,$45
	.dc.b	$7A,$3F,$DD,$F3,$FC,$CD,$E2,$7F,$2D,$56,$3F,$8A,$9E,$B1,$D5,$FB
	.dc.b	$1C,$1E,$F7,$EA,$1B,$3F,$A8,$7F,$CC,$D1,$FC,$B3,$6F,$F3,$5F,$57
	.dc.b	$EE,$9F,$F3,$05,$7F,$9C,$0F,$F1,$8F,$57,$EF,$BC,$2F,$F3,$8F,$DD
	.dc.b	$FF,$96,$CF,$E0,$99,$FD,$3E,$FF,$17,$0F,$F0,$0D,$3F,$87,$6F,$EC
	.dc.b	$6E,$5E,$9D,$B6,$FF,$19,$C3,$EE,$1F,$EE,$C9,$FE,$28,$FF,$1C,$D3
	.dc.b	$FC,$C5,$3F,$E6,$FE,$2F,$F2,$EB,$83,$D0,$7F,$AC,$E3,$FD,$0F,$F8
	.dc.b	$27,$9A,$3F,$C9,$6F,$F9,$7E,$FF,$01,$8F,$F8,$3F,$93,$FC,$BD,$5F
	.dc.b	$2D,$AF,$F9,$90,$FD,$91,$FE,$09,$CB,$0F,$E2,$0F,$F3,$15,$7E,$E1
	.dc.b	$F1,$9F,$E7,$1E,$5F,$EA,$0F,$F8,$47,$7F,$F3,$26,$1F,$D3,$E3,$F6
	.dc.b	$F4,$FC,$1B,$F9,$84,$7F,$BA,$A3,$0F,$F5,$C3,$FC,$C5,$DD,$F3,$CF
	.dc.b	$C0,$1F,$76,$3F,$4F,$B7,$F9,$0F,$2F,$F3,$CC,$FF,$05,$E9,$FF,$2C
	.dc.b	$DF,$FA,$67,$BB,$FD,$BF,$87,$F6,$E8,$FE,$F8,$7E,$37,$A7,$F9,$21
	.dc.b	$FE,$6B,$EA,$FB,$C7,$E2,$67,$FD,$57,$3F,$E2,$DF,$27,$FC,$A1,$18
	.dc.b	$8F,$F2,$9F,$93,$FE,$0A,$F2,$7F,$D7,$77,$FE,$5B,$5F,$16,$F8,$9D
	.dc.b	$FC,$66,$B0,$FF,$2C,$78,$43,$0F,$F1,$95,$E7,$63,$C5,$FF,$13,$E9
	.dc.b	$E1,$7F,$CA,$ED,$7F,$8E,$8F,$43,$83,$8F,$BD,$EB,$E5,$7F,$94,$F6
	.dc.b	$FF,$1C,$73,$63,$3F,$D9,$1F,$B1,$8F,$E5,$7E,$A7,$F9,$5F,$D6,$F2
	.dc.b	$BF,$E5,$9F,$5E,$C7,$F9,$4F,$4F,$89,$3F,$E5,$DF,$D5,$FF,$CC,$38
	.dc.b	$74,$F0,$7F,$C9,$4D,$CB,$78,$7F,$96,$DB,$EB,$FF,$95,$20,$FF,$95
	.dc.b	$AF,$FA,$05,$7F,$97,$F0,$F2,$5F,$07,$FC,$B0,$7F,$96,$F3,$FE,$87
	.dc.b	$A7,$B2,$BF,$BE,$57,$D9,$8F,$F2,$DC,$FC,$28,$F9,$5B,$7F,$46,$DB
	.dc.b	$BD,$FF,$95,$B7,$B3,$FE,$58,$AD,$BF,$E5,$8F,$5B,$7C,$5F,$D4,$78
	.dc.b	$AB,$FD,$4E,$BF,$11,$D1,$FE,$54,$CF,$F0,$DC,$DF,$85,$8F,$F5,$07
	.dc.b	$2F,$81,$FF,$2B,$D7,$E3,$FB,$FD,$5C,$9E,$47,$FC,$E7,$D5,$FD,$00
	.dc.b	$FF,$2A,$75,$FF,$62,$8F,$F5,$A7,$07,$F9,$3E,$30,$FE,$8A,$BF,$AC
	.dc.b	$3F,$BC,$3F,$C3,$FD,$BF,$DB,$FB,$BF,$DA,$2F,$EB,$71,$FF,$29,$4B
	.dc.b	$0E,$6D,$38,$9F,$E1,$C7,$E7,$CF,$F8,$A2,$7F,$7F,$3F,$C9,$0F,$F2
	.dc.b	$99,$FE,$63,$EF,$FF,$2F,$C7,$F5,$E8,$C9,$FC,$35,$FE,$25,$5F,$C3
	.dc.b	$7B,$F5,$3F,$98,$9F,$EC,$07,$FA,$C5,$FF,$01,$E1,$FE,$3A,$F0,$BF
	.dc.b	$CB,$95,$67,$FC,$B6,$7F,$99,$2B,$F9,$4B,$EC,$34,$7C,$55,$E1,$FF
	.dc.b	$31,$8E,$0F,$F9,$83,$C4,$FE,$73,$27,$94,$FF,$AD,$7D,$4F,$C8,$C7
	.dc.b	$C3,$F1,$BF,$A7,$D7,$F9,$3B,$D3,$FF,$42,$46,$21,$1F,$98,$8F,$F3
	.dc.b	$0D,$FF,$78,$3F,$B4,$9F,$E0,$3C,$F4,$E0,$FE,$D9,$4F,$27,$A7,$93
	.dc.b	$FE,$91,$6F,$0F,$F9,$62,$BF,$9E,$B7,$7F,$D4,$79,$1F,$E0,$19,$CE
	.dc.b	$1F,$DC,$9F,$C0,$12,$7F,$D3,$9E,$67,$F7,$39,$D4,$7F,$A7,$5F,$F2
	.dc.b	$EF,$9B,$EA,$F4,$BF,$C0,$D7,$FD,$A7,$1C,$BE,$93,$FD,$53,$E3,$BF
	.dc.b	$09,$F8,$98,$7F,$9D,$21,$E7,$FF,$9C,$5F,$75,$7F,$71,$9F,$F1,$67
	.dc.b	$A9,$BD,$B8,$C7,$F9,$7A,$7F,$B6,$F3,$73,$E2,$FF,$49,$61,$F6,$67
	.dc.b	$C4,$3D,$BF,$07,$3F,$29,$FC,$7E,$1F,$EE,$CA,$FF,$53,$C7,$E5,$9F
	.dc.b	$CC,$9F,$E1,$BE,$9F,$F5,$DF,$02,$7F,$47,$5F,$E5,$13,$FC,$C1,$5F
	.dc.b	$2C,$FF,$7E,$47,$BF,$AB,$F6,$55,$FE,$60,$9F,$D8,$70,$7F,$98,$0F
	.dc.b	$F2,$55,$7F,$96,$A3,$E1,$5F,$F9,$37,$C0,$FE,$9B,$DC,$F3,$FF,$DC
	.dc.b	$3D,$DF,$D4,$3C,$2F,$F3,$47,$B1,$FD,$4B,$D2,$FF,$32,$6B,$FC,$FB
	.dc.b	$CD,$FE,$68,$3C,$C7,$B3,$FC,$99,$FE,$63,$E7,$FF,$38,$73,$B7,$F6
	.dc.b	$7C,$39,$BF,$DA,$72,$EE,$FF,$24,$D3,$C6,$FA,$77,$F8,$9B,$FF,$16
	.dc.b	$ED,$FE,$07,$C5,$FE,$2C,$E3,$F8,$77,$FE,$6C,$9F,$C7,$3F,$30,$F9
	.dc.b	$9C,$9F,$41,$FA,$98,$7F,$AF,$4F,$90,$FD,$A1,$C1,$FE,$6D,$EF,$FF
	.dc.b	$60,$F3,$7F,$B5,$AB,$F0,$7C,$FF,$EC,$A7,$E6,$DB,$D7,$C4,$7E,$32
	.dc.b	$3F,$D9,$3D,$7F,$90,$7F,$2B,$1F,$E8,$FE,$CF,$F4,$07,$6F,$FB,$26
	.dc.b	$7F,$3D,$3D,$1F,$F6,$8F,$AB,$FA,$68,$FF,$40,$77,$79,$BB,$7F,$5B
	.dc.b	$1F,$C0,$9C,$77,$FF,$AF,$F0,$FF,$5F,$47,$83,$0F,$F7,$1E,$3C,$72
	.dc.b	$FF,$62,$77,$FF,$A0,$EB,$F8,$EE,$3F,$E5,$9E,$0F,$F5,$E6,$FF,$EB
	.dc.b	$F8,$FE,$19,$3F,$EB,$BF,$1F,$FD,$28,$FC,$FC,$BF,$D3,$2D,$7F,$D2
	.dc.b	$2F,$F8,$E2,$FF,$D6,$BA,$FF,$AE,$63,$EF,$4F,$FA,$B9,$FC,$27,$47
	.dc.b	$F8,$8C,$83,$D6,$FF,$0D,$F9,$DF,$EB,$2A,$FE,$E2,$89,$EB,$FF,$70
	.dc.b	$AF,$F5,$AC,$F6,$3F,$CE,$67,$FB,$F3,$CB,$FE,$AF,$A7,$F4,$73,$FD
	.dc.b	$9B,$0E,$BF,$F8,$03,$D1,$FE,$AC,$3F,$DA,$B1,$F8,$7D,$7F,$D6,$5D
	.dc.b	$7F,$D1,$BA,$BF,$D6,$9B,$76,$E1,$FF,$59,$9F,$EE,$6B,$F0,$F1,$7F
	.dc.b	$AC,$5F,$A8,$79,$0F,$C8,$98,$3F,$D1,$3C,$98,$E5,$FF,$7C,$47,$E1
	.dc.b	$B5,$FF,$65,$78,$DF,$EB,$3F,$1B,$F0,$D1,$F3,$B5,$FF,$73,$9F,$D0
	.dc.b	$3B,$2C,$7E,$63,$C7,$6F,$88,$FF,$71,$47,$FB,$BB,$9D,$97,$F4,$66
	.dc.b	$CF,$1B,$F8,$EF,$38,$7C,$C7,$F1,$D8,$FE,$AB,$C3,$FE,$F4,$FB,$C4
	.dc.b	$6B,$8B,$FC,$13,$39,$CF,$8E,$FF,$D5,$30,$FE,$7B,$7D,$DF,$F6,$A3
	.dc.b	$FA,$9C,$FB,$F8,$5B,$3F,$AE,$E7,$89,$E1,$C5,$FD,$AF,$CB,$F7,$E1
	.dc.b	$FE,$B2,$C3,$FD,$6B,$0D,$BD,$35,$E1,$F3,$3F,$D9,$57,$FD,$0F,$0F
	.dc.b	$AD,$83,$D8,$E7,$5F,$DD,$DE,$55,$B9,$FB,$9C,$3E,$0F,$6B,$FC,$07
	.dc.b	$87,$FB,$7B,$7F,$8F,$1F,$DF,$2F,$F1,$97,$F4,$2F,$9F,$7F,$E2,$61
	.dc.b	$0D,$BE,$7C,$FC,$7F,$6B,$FC,$31,$C7,$F9,$DC,$BE,$9C,$63,$0F,$E8
	.dc.b	$98,$FF,$AE,$DF,$E8,$87,$7F,$CA,$77,$A7,$FB,$67,$B3,$FD,$C5,$F9
	.dc.b	$46,$23,$F0,$B8,$FD,$EF,$77,$FC,$4D,$87,$EA,$F8,$F1,$5F,$E1,$18
	.dc.b	$FD,$76,$DF,$84,$F2,$7F,$D8,$33,$F8,$2E,$0F,$A7,$1F,$7F,$D0,$FC
	.dc.b	$1C,$FE,$1B,$AB,$EB,$C6,$21,$1F,$8A,$DD,$3F,$EC,$1B,$DD,$7E,$77
	.dc.b	$4F,$C9,$F2,$7F,$98,$1F,$EC,$DE,$0F,$E0,$A3,$13,$F9,$49,$F8,$BC
	.dc.b	$1F,$E3,$1E,$FF,$D0,$E5,$FE,$BF,$DB,$ED,$D7,$E7,$AB,$F5,$93,$FD
	.dc.b	$22,$FF,$D8,$DE,$5F,$E1,$7D,$5F,$F3,$E7,$E5,$8F,$C8,$3F,$4B,$E0
	.dc.b	$FF,$A0,$3D,$F7,$AA,$AB,$97,$3C,$BF,$D9,$BB,$FF,$B2,$AF,$74,$FD
	.dc.b	$B3,$9D,$7E,$B3,$C9,$FE,$4E,$7F,$B2,$FF,$18,$7E,$E3,$A9,$CC,$E7
	.dc.b	$F0,$7C,$38,$A7,$ED,$5B,$F3,$FF,$D9,$7F,$B2,$BE,$EF,$11,$FC,$BC
	.dc.b	$90,$FC,$47,$31,$91,$0F,$C1,$FC,$5E,$67,$F1,$FF,$C3,$FE,$D7,$2F
	.dc.b	$F0,$4D,$93,$9E,$AD,$F5,$FF,$38,$BF,$D6,$DF,$FB,$53,$DC,$FF,$07
	.dc.b	$1F,$EC,$6F,$D3,$5F,$82,$FB,$FD,$EF,$F0,$9F,$C0,$25,$F8,$A6,$6F
	.dc.b	$8C,$E7,$93,$F8,$CE,$7E,$3E,$5F,$89,$CF,$8E,$0F,$9D,$1F,$3A,$9C
	.dc.b	$F9,$9F,$9B,$29,$F9,$7D,$9F,$CA,$27,$DB,$87,$AD,$E1,$C5,$EF,$9D
	.dc.b	$3F,$17,$C9,$E9,$5F,$DE,$0A,$7F,$5A,$F1,$7F,$8C,$7E,$E3,$F2,$70
	.dc.b	$EB,$3F,$CE,$0F,$E7,$3C,$9F,$CC,$27,$58,$7C,$D6,$CF,$CD,$C1,$1F
	.dc.b	$E2,$5B,$E7,$3C,$EF,$97,$2F,$D7,$C3,$93,$67,$95,$FE,$DA,$FF,$64
	.dc.b	$D7,$FA,$6E,$1D,$BF,$15,$FD,$8C,$7F,$52,$F1,$FC,$D1,$8C,$11,$FD
	.dc.b	$5A,$FE,$CD,$74,$E3,$FE,$B9,$FA,$F7,$F7,$DC,$BE,$01,$FC,$9B,$04
	.dc.b	$66,$3E,$26,$BF,$CF,$E3,$FD,$D9,$7F,$24,$7C,$37,$F5,$68,$CC,$7F
	.dc.b	$AB,$6D,$F8,$CB,$F0,$3F,$F5,$4B,$67,$FB,$27,$C1,$57,$DE,$DC,$BF
	.dc.b	$84,$FE,$1F,$D2,$7F,$DA,$B8,$23,$11,$FD,$9B,$2F,$EC,$F1,$E9,$AF
	.dc.b	$E8,$9C,$1B,$DF,$FD,$73,$FB,$BF,$C4,$C3,$E6,$B6,$BF,$8B,$AE,$1F
	.dc.b	$99,$5F,$E0,$08,$C5,$7F,$74,$F7,$BF,$CD,$B3,$FD,$DA,$7E,$C5,$FF
	.dc.b	$5A,$C1,$18,$8F,$C9,$67,$83,$F0,$4F,$D2,$E4,$ED,$18,$8F,$C5,$56
	.dc.b	$F1,$FF,$01,$F2,$7F,$03,$7A,$9B,$9F,$CC,$D6,$23,$71,$F9,$7C,$3F
	.dc.b	$32,$F7,$87,$D8,$7E,$BE,$3D,$0D,$7F,$C8,$3E,$87,$F8,$72,$3F,$5F
	.dc.b	$1F,$C3,$6F,$8B,$E1,$19,$7F,$81,$67,$71,$DD,$FD,$E7,$4F,$F2,$9F
	.dc.b	$1F,$8E,$3F,$C1,$A3,$FC,$B3,$D5,$FC,$CA,$B7,$1F,$C2,$7A,$FC,$FE
	.dc.b	$17,$F1,$BE,$EF,$F5,$6F,$81,$FC,$26,$FF,$A1,$78,$3F,$C6,$A3,$F8
	.dc.b	$CC,$63,$C1,$FE,$33,$1F,$C8,$23,$FA,$47,$17,$F2,$96,$ED,$FB,$D6
	.dc.b	$BE,$9C,$7F,$98,$78,$BF,$CF,$BC,$4F,$1E,$FF,$E3,$68,$FE,$83,$E3
	.dc.b	$62,$3F,$9E,$47,$F4,$0F,$1F,$1C,$1F,$07,$2F,$EB,$0F,$F2,$86,$BE
	.dc.b	$F7,$FA,$E7,$99,$FE,$D1,$F3,$3F,$AE,$3F,$D8,$7C,$AE,$CE,$7C,$EF
	.dc.b	$ED,$3E,$76,$3C,$FF,$ED,$9E,$7F,$F6,$2A,$FB,$FE,$87,$F9,$77,$2F
	.dc.b	$EE,$DC,$5F,$94,$6B,$FE,$6A,$3F,$0B,$9F,$E8,$BB,$2F,$07,$E8,$4F
	.dc.b	$E4,$0F,$FB,$A6,$BF,$C1,$B1,$FE,$0B,$6B,$FE,$87,$F5,$3F,$87,$46
	.dc.b	$3C,$2F,$EB,$47,$F9,$6D,$C4,$F5,$3F,$92,$2F,$53,$FB,$06,$3F,$E0
	.dc.b	$C8,$C4,$FF,$99,$A3,$F9,$D3,$FD,$97,$CB,$FF,$3B,$CB,$D5,$FF,$18
	.dc.b	$D7,$F2,$6F,$13,$F4,$31,$8F,$17,$FC,$8D,$5F,$E3,$FA,$F8,$9E,$C7
	.dc.b	$81,$FF,$3E,$EF,$FC,$93,$D8,$FD,$4C,$E6,$7F,$D4,$47,$F9,$4F,$D8
	.dc.b	$CE,$9F,$EC,$28,$FE,$6B,$C1,$88,$6F,$FE,$A2,$9F,$95,$ED,$73,$7F
	.dc.b	$98,$D7,$ED,$A3,$FC,$B9,$BF,$EB,$5B,$C3,$FB,$AF,$6F,$FC,$CF,$E6
	.dc.b	$FC,$06,$E5,$FB,$EF,$73,$A6,$5F,$E6,$7C,$B1,$3F,$EF,$4E,$5F,$8F
	.dc.b	$5F,$E7,$5E,$4C,$CF,$98,$FE,$80,$7C,$CF,$77,$9E,$1F,$71,$FF,$44
	.dc.b	$61,$FC,$31,$FF,$45,$57,$F0,$EF,$0F,$FD,$1F,$C9,$FE,$94,$BF,$E1
	.dc.b	$F1,$CB,$7F,$AE,$FF,$92,$E3,$F8,$AF,$BF,$F9,$73,$58,$FF,$18,$F7
	.dc.b	$BF,$CB,$DE,$F7,$DE,$6E,$8F,$C0,$3F,$8D,$61,$FD,$7B,$2F,$F4,$14
	.dc.b	$62,$3F,$AE,$E9,$B7,$AC,$7F,$92,$DF,$F4,$9F,$17,$3F,$1F,$FD,$89
	.dc.b	$39,$9F,$E7,$73,$5E,$D8,$7F,$AF,$8F,$F6,$29,$FE,$B9,$3F,$D7,$A7
	.dc.b	$FA,$F3,$E2,$3F,$CC,$FB,$7F,$9A,$70,$FF,$4A,$CF,$E6,$EB,$FD,$97
	.dc.b	$1F,$82,$F6,$FF,$29,$1F,$8E,$3F,$D3,$BC,$BF,$E8,$AA,$FA,$C7,$1E
	.dc.b	$CF,$F4,$3B,$CE,$7F,$C4,$1E,$97,$F9,$DB,$E0,$3F,$39,$8F,$F9,$2F
	.dc.b	$44,$FF,$9E,$BB,$BC,$99,$FF,$9B,$7C,$9F,$F2,$EF,$E3,$9F,$E4,$EC
	.dc.b	$BF,$C1,$3F,$AA,$FF,$A0,$32,$FE,$69,$E8,$3D,$0F,$F2,$77,$E3,$1F
	.dc.b	$E6,$BB,$FF,$7C,$68,$BF,$F1,$DC,$7F,$84,$86,$1F,$E8,$A7,$E2,$7B
	.dc.b	$FF,$D6,$1F,$F1,$B4,$7F,$BB,$A3,$FC,$61,$C7,$C6,$78,$9F,$E3,$4E
	.dc.b	$9F,$C6,$71,$7F,$8A,$4F,$E3,$1D,$5F,$E6,$91,$F0,$BC,$FE,$03,$7F
	.dc.b	$01,$FE,$09,$C3,$FD,$93,$7F,$E2,$4E,$4F,$F6,$9F,$2F,$FB,$FB,$DF
	.dc.b	$FF,$0B,$57,$F9,$36,$9E,$FE,$23,$F0,$7E,$FF,$F8,$5A,$7F,$DA,$51
	.dc.b	$FE,$13,$7E,$A3,$FE,$2E,$7F,$C0,$F9,$7A,$3D,$EF,$F0,$7C,$7E,$C7
	.dc.b	$5F,$46,$AB,$FF,$06,$ED,$B9,$FE,$E5,$C3,$9C,$3E,$45,$6E,$3F,$BD
	.dc.b	$3F,$8F,$F0,$FF,$BA,$BF,$30,$FE,$B0,$7D,$6E,$3F,$EB,$5C,$FF,$DA
	.dc.b	$F9,$BF,$AE,$72,$7C,$FC,$3F,$C0,$95,$FD,$93,$A7,$FA,$FF,$C1,$BA
	.dc.b	$BF,$A7,$F0,$7B,$B3,$F4,$74,$FE,$07,$5F,$E8,$B1,$F9,$E9,$1F,$EF
	.dc.b	$79,$FF,$08,$79,$FF,$EE,$28,$FC,$AC,$FF,$38,$7F,$DD,$0F,$F2,$CA
	.dc.b	$FD,$2F,$47,$F9,$63,$BF,$F9,$45,$7E,$87,$4F,$E5,$75,$FC,$46,$FF
	.dc.b	$C3,$B1,$FC,$2E,$7F,$6A,$3C,$0F,$9D,$E0,$FF,$07,$AF,$A1,$E1,$F9
	.dc.b	$A3,$F8,$04,$7D,$9A,$7C,$37,$9D,$78,$9E,$C8,$C7,$3D,$9F,$DA,$1F
	.dc.b	$E9,$3E,$EF,$83,$E2,$EA,$0F,$3B,$CC,$70,$3F,$71,$F5,$1E,$07,$03
	.dc.b	$E3,$BF,$0D,$19,$8F,$F3,$F9,$FE,$C9,$F9,$16,$2F,$9D,$7E,$05,$E6
	.dc.b	$6D,$74,$3D,$6B,$EE,$3F,$84,$ED,$FB,$98,$C4,$EE,$FA,$57,$C3,$D6
	.dc.b	$CF,$3D,$3C,$1F,$69,$CB,$1F,$EA,$98,$A3,$10,$E1,$F0,$C7,$94,$FC
	.dc.b	$26,$9E,$AA,$7D,$0F,$0F,$8E,$C3,$71,$F7,$3E,$D3,$F1,$9F,$5D,$C7
	.dc.b	$F5,$2F,$5E,$9F,$92,$DF,$DF,$18,$8F,$8F,$1F,$1E,$3E,$31,$FE,$01
	.dc.b	$8F,$F5,$5D,$E5,$FB,$BA,$FF,$B3,$6B,$E7,$7E,$A7,$FC,$C3,$A3,$FD
	.dc.b	$F5,$6F,$6F,$38,$7F,$41,$9F,$E4,$CF,$E9,$27,$FD,$5F,$AA,$FE,$B1
	.dc.b	$FE,$11,$C3,$FD,$B5,$7C,$E7,$57,$FD,$06,$73,$D7,$F7,$FF,$0B,$FE
	.dc.b	$83,$3D,$4F,$94,$FE,$07,$EE,$7F,$C6,$B1,$FE,$00,$AF,$C3,$47,$F8
	.dc.b	$8F,$BB,$F6,$F2,$F7,$3E,$46,$2A,$FC,$57,$E0,$9C,$AB,$F8,$D1,$F5
	.dc.b	$FE,$43,$EE,$F0,$7F,$1B,$BF,$E3,$EF,$F8,$EA,$7F,$9B,$46,$23,$F9
	.dc.b	$04,$7F,$BE,$77,$FC,$C7,$77,$8B,$E0,$7F,$CC,$DF,$04,$7F,$CD,$1F
	.dc.b	$57,$AB,$FE,$1B,$1A,$3F,$CD,$18,$FE,$95,$C7,$C2,$3D,$F7,$D6,$F1
	.dc.b	$B7,$C4,$C3,$FD,$C5,$C7,$9F,$34,$3E,$87,$FA,$A6,$7F,$B1,$8C,$9C
	.dc.b	$1C,$3C,$2F,$D9,$1F,$B7,$F4,$3F,$9F,$70,$FE,$D6,$7E,$1F,$07,$FB
	.dc.b	$22,$B6,$FD,$2F,$4B,$FB,$87,$E1,$FE,$08,$FF,$53,$57,$D6,$3F,$4C
	.dc.b	$E1,$FC,$8D,$7F,$0E,$7F,$D6,$46,$F3,$FE,$09,$E5,$77,$8F,$A6,$D7
	.dc.b	$F2,$93,$C5,$FC,$8E,$3F,$C0,$E7,$F8,$C4,$FF,$85,$32,$F5,$BD,$4C
	.dc.b	$4F,$E0,$6B,$C7,$F8,$47,$57,$F7,$8D,$E5,$FE,$2D,$F0,$3F,$C5,$4E
	.dc.b	$BD,$7F,$8D,$F8,$7F,$DC,$5F,$F1,$CF,$C2,$5F,$A2,$FF,$30,$6C,$7F
	.dc.b	$11,$F8,$1F,$F7,$A6,$7F,$69,$B1,$F8,$D6,$BF,$D5,$3C,$6F,$EA,$B1
	.dc.b	$88,$7E,$AB,$FE,$13,$F7,$3F,$85,$F1,$29,$F7,$3F,$D0,$1B,$7F,$C8
	.dc.b	$7D,$DA,$FF,$34,$3F,$81,$B8,$CB,$F4,$51,$FC,$4E,$3C,$19,$7F,$38
	.dc.b	$6C,$7F,$80,$8F,$E8,$39,$7F,$80,$72,$FE,$91,$3E,$79,$FE,$7F,$FA
	.dc.b	$AF,$C2,$7C,$07,$F0,$D7,$18,$78,$23,$FA,$2F,$D5,$B8,$3E,$74,$7F
	.dc.b	$48,$33,$C3,$FD,$56,$4F,$F5,$E5,$C7,$F6,$F7,$46,$2B,$FB,$FF,$2F
	.dc.b	$E1,$1F,$E4,$71,$F8,$4C,$7F,$41,$1F,$CC,$E3,$1C,$BF,$C0,$5F,$C8
	.dc.b	$4F,$9F,$97,$AF,$37,$F0,$C8,$B7,$D9,$FC,$73,$CE,$3F,$19,$5F,$99
	.dc.b	$EC,$FD,$E4,$FE,$86,$3F,$B4,$75,$7F,$96,$FE,$5F,$F7,$F1,$8F,$87
	.dc.b	$47,$EB,$6B,$FB,$FC,$E7,$B3,$FC,$A8,$FE,$EF,$2F,$0C,$FC,$CE,$CF
	.dc.b	$09,$FE,$29,$ED,$E5,$AF,$C6,$8C,$4F,$8D,$FF,$17,$77,$74,$7F,$98
	.dc.b	$EB,$E8,$9F,$F2,$E6,$3F,$C6,$FC,$1F,$E3,$5E,$27,$F2,$37,$F8,$E7
	.dc.b	$89,$D6,$BF,$91,$C6,$BC,$43,$FC,$87,$C3,$D4,$F1,$F1,$18,$5F,$92
	.dc.b	$FB,$7E,$37,$E0,$4F,$C0,$1F,$47,$37,$8F,$F8,$8E,$AF,$B5,$E4,$7E
	.dc.b	$14,$FE,$95,$97,$F9,$03,$C9,$CB,$FF,$00,$4F,$A5,$E5,$7E,$2C,$FE
	.dc.b	$AE,$FF,$8B,$0F,$CA,$F9,$BF,$D8,$79,$FF,$D4,$FE,$77,$E3,$A3,$11
	.dc.b	$FE,$8F,$F3,$71,$5F,$E1,$5D,$3F,$B2,$BF,$DA,$3C,$EF,$EE,$78,$6A
	.dc.b	$30,$7F,$71,$AF,$CC,$F9,$FF,$DF,$F0,$FE,$6B,$D5,$8A,$FE,$FF,$1F
	.dc.b	$E7,$5D,$B1,$1F,$AE,$3F,$57,$5F,$E0,$B8,$FC,$3B,$5F,$D2,$FA,$1F
	.dc.b	$C2,$B3,$FF,$0D,$D7,$9E,$3F,$C3,$93,$A8,$FF,$03,$5F,$F4,$6F,$47
	.dc.b	$F2,$3E,$9F,$F8,$92,$BF,$5E,$FE,$CC,$FF,$15,$FA,$9F,$BC,$6E,$7F
	.dc.b	$B4,$F5,$BF,$C6,$D9,$FF,$8E,$B2,$F4,$4E,$5F,$91,$3F,$77,$D7,$FE
	.dc.b	$00,$7F,$97,$3A,$BF,$C9,$DA,$7F,$18,$F5,$BF,$C9,$B1,$88,$FE,$F3
	.dc.b	$7F,$E7,$EF,$67,$FD,$35,$EB,$FE,$EA,$73,$3F,$E8,$2F,$61,$39,$9F
	.dc.b	$D2,$B6,$7F,$AB,$3D,$63,$F6,$53,$C3,$C3,$FF,$34,$4E,$6B,$FA,$E1
	.dc.b	$4F,$DC,$79,$5F,$E7,$75,$EC,$E5,$FF,$35,$74,$7F,$96,$1F,$EB,$B8
	.dc.b	$FF,$8B,$BC,$CF,$F2,$B4,$7F,$9C,$33,$FF,$0D,$73,$7F,$A5,$FD,$4F
	.dc.b	$F0,$E7,$47,$DF,$AF,$F2,$21,$FE,$72,$C3,$FB,$EB,$FE,$40,$3F,$1E
	.dc.b	$7F,$A2,$36,$C4,$7F,$92,$4F,$F6,$06,$48,$FF,$1D,$4F,$F9,$1C,$FF
	.dc.b	$2C,$4F,$F8,$AF,$BB,$F5,$B1,$FE,$4D,$8F,$F1,$77,$81,$FE,$30,$BF
	.dc.b	$F0,$E7,$8B,$FE,$20,$F1,$1F,$9D,$FF,$87,$FC,$8F,$F0,$FF,$D4,$7F
	.dc.b	$82,$F2,$FE,$55,$97,$FA,$04,$FF,$0A,$9F,$E2,$53,$CA,$63,$1F,$F0
	.dc.b	$4C,$7F,$AF,$2B,$9C,$FF,$81,$63,$9F,$4D,$7E,$6C,$7B,$B8,$BF,$A1
	.dc.b	$4F,$FA,$37,$D1,$FF,$27,$FA,$3F,$E2,$48,$FF,$48,$7A,$7F,$D5,$30
	.dc.b	$FF,$13,$FB,$1F,$D7,$DF,$EE,$5E,$BF,$D3,$D3,$FB,$A7,$B3,$F7,$FD
	.dc.b	$8F,$ED,$DA,$FE,$42,$FF,$DA,$31,$FD,$F7,$9D,$7F,$DB,$67,$F7,$47
	.dc.b	$F5,$AC,$BC,$17,$FE,$F3,$F0,$BF,$AE,$4B,$5F,$34,$FC,$69,$F5,$B7
	.dc.b	$B3,$E9,$47,$F8,$A6,$77,$3F,$E4,$08,$FE,$A9,$C2,$E2,$FE,$81,$87
	.dc.b	$F0,$0A,$FF,$3F,$D7,$F4,$D7,$EA,$1C,$7B,$BF,$A2,$DF,$18,$FA,$E7
	.dc.b	$F3,$CE,$6F,$C0,$72,$FF,$3A,$C3,$A4,$FE,$3B,$0F,$F3,$D4,$7F,$BD
	.dc.b	$20,$7F,$80,$DF,$F7,$BF,$91,$E7,$F2,$BF,$9A,$F8,$FE,$9F,$2B,$F9
	.dc.b	$8F,$0F,$E5,$F1,$FE,$5F,$8F,$E6,$A3,$FD,$43,$DB,$FC,$9A,$3F,$46
	.dc.b	$7E,$92,$5D,$DF,$A9,$7F,$91,$E7,$FE,$18,$E4,$FA,$1E,$3F,$FA,$C6
	.dc.b	$3F,$86,$5F,$F1,$77,$F1,$6F,$DA,$F1,$3F,$79,$E1,$7D,$CF,$13,$F8
	.dc.b	$55,$7F,$99,$5F,$E0,$0F,$FA,$73,$B7,$FD,$30,$FF,$08,$6B,$FD,$FD
	.dc.b	$AF,$FA,$A7,$BE,$BF,$EA,$46,$F0,$23,$FC,$37,$5F,$E0,$77,$FD,$79
	.dc.b	$E2,$7F,$2F,$8C,$79,$5B,$8C,$4F,$F1,$CC,$FF,$8A,$E3,$E0,$AF,$F7
	.dc.b	$CB,$F8,$F9,$ED,$7F,$EA,$FE,$BF,$AF,$39,$9F,$E5,$66,$8D,$E3,$FF
	.dc.b	$00,$7F,$4D,$FF,$5D,$E3,$F8,$63,$FD,$E1,$0C,$DF,$AA,$78,$6F,$0F
	.dc.b	$8E,$6C,$7E,$1E,$BF,$1F,$E6,$7F,$BD,$7F,$55,$F8,$FE,$57,$FB,$DF
	.dc.b	$F4,$1E,$F5,$FC,$72,$FF,$D9,$7B,$F7,$9C,$CF,$FB,$8A,$7F,$A2,$66
	.dc.b	$FC,$5F,$FD,$99,$9F,$FA,$A1,$B3,$FE,$D4,$E8,$F4,$7E,$A7,$F8,$58
	.dc.b	$FF,$48,$FE,$A3,$FC,$1F,$8D,$87,$B6,$3F,$C1,$B7,$FE,$96,$FC,$73
	.dc.b	$FD,$C3,$3F,$EF,$6B,$DE,$4F,$43,$FD,$A1,$EB,$FF,$B1,$FD,$6F,$F1
	.dc.b	$2F,$89,$FE,$2A,$FB,$8F,$91,$5A,$F4,$15,$AB,$FF,$67,$61,$CF,$1F
	.dc.b	$F1,$9D,$6B,$D8,$FF,$1C,$7D,$97,$29,$FE,$55,$E4,$FF,$B4,$49,$7C
	.dc.b	$63,$11,$9F,$B3,$E1,$1F,$ED,$0F,$76,$1F,$ED,$6A,$E0,$E0,$FF,$6C
	.dc.b	$FB,$BC,$1C,$0F,$B4,$C3,$BF,$B5,$EB,$E8,$FE,$1F,$AB,$FD,$F5,$C3
	.dc.b	$F9,$FC,$BF,$CF,$E7,$E7,$F9,$3C,$7F,$07,$FE,$6B,$27,$C1,$E0,$FF
	.dc.b	$35,$B9,$FB,$7C,$7C,$22,$38,$7D,$C6,$1E,$5F,$71,$FD,$D7,$8F,$E0
	.dc.b	$9F,$EF,$5E,$2F,$F0,$D3,$81,$C7,$FD,$EF,$6F,$47,$A3,$F9,$DA,$FD
	.dc.b	$B9,$FE,$90,$E0,$F6,$4E,$77,$FD,$D7,$A1,$FB,$97,$F0,$7F,$15,$FD
	.dc.b	$52,$7E,$15,$A3,$E1,$56,$F8,$7F,$55,$F6,$DD,$8F,$C6,$E1,$FD,$E3
	.dc.b	$0C,$BF,$B5,$9F,$C8,$1F,$B1,$FA,$AF,$E1,$7E,$FB,$F5,$4F,$E1,$F5
	.dc.b	$FE,$CF,$72,$7E,$B9,$C1,$FE,$6B,$3F,$92,$C7,$17,$03,$CA,$FD,$1C
	.dc.b	$3F,$2E,$EA,$D9,$D9,$FE,$D1,$8F,$E3,$BE,$FE,$63,$F8,$69,$FC,$75
	.dc.b	$EC,$E0,$FC,$25,$7F,$6E,$BF,$E0,$78,$AB,$7E,$47,$F2,$5F,$53,$FA
	.dc.b	$7C,$63,$0F,$C5,$35,$FE,$55,$B3,$4F,$E5,$EB,$6F,$9B,$3F,$E8,$7C
	.dc.b	$7F,$DA,$31,$87,$FB,$73,$C3,$7F,$AF,$3F,$97,$AF,$B4,$FF,$88,$37
	.dc.b	$FE,$D3,$1F,$ED,$43,$EE,$1F,$87,$7F,$44,$FF,$3E,$8E,$5E,$97,$F1
	.dc.b	$A9,$FF,$12,$4F,$E5,$39,$7F,$2F,$CB,$FD,$5D,$FC,$B1,$FE,$E6,$EC
	.dc.b	$FD,$57,$47,$E9,$BA,$BF,$45,$D5,$9C,$FF,$A6,$4E,$47,$56,$63,$F5
	.dc.b	$1D,$9F,$AC,$26,$7F,$C1,$EB,$87,$66,$E3,$D7,$97,$E6,$F5,$FF,$6C
	.dc.b	$BF,$ED,$5E,$EF,$E1,$5D,$DD,$5F,$E0,$BD,$F5,$FE,$0F,$7F,$C3,$1F
	.dc.b	$E1,$9E,$07,$8F,$83,$F6,$5C,$35,$FD,$A5,$7F,$19,$F1,$FF,$DA,$3E
	.dc.b	$27,$F2,$17,$F9,$54,$67,$C5,$FE,$5D,$E2,$FF,$24,$8C,$56,$E4,$A1
	.dc.b	$E4,$7E,$0B,$E1,$BF,$B9,$75,$E7,$CB,$FE,$7F,$C5,$FE,$39,$7F,$A2
	.dc.b	$47,$1F,$33,$FA,$44,$FF,$8F,$7C,$DF,$E9,$BD,$1F,$DE,$F0,$ED,$18
	.dc.b	$8F,$F0,$F7,$9D,$FD,$63,$CA,$FE,$45,$0F,$3F,$C9,$39,$8F,$EA,$DA
	.dc.b	$BD,$0F,$D1,$BF,$DB,$A3,$FB,$56,$5F,$E6,$EF,$47,$FA,$BC,$E6,$7F
	.dc.b	$58,$7F,$83,$70,$FE,$2B,$E9,$7F,$73,$97,$A7,$F9,$1C,$3F,$C7,$3F
	.dc.b	$0F,$FD,$83,$D0,$FF,$02,$79,$7F,$CD,$C9,$C5,$FE,$59,$8F,$F3,$67
	.dc.b	$AB,$FE,$40,$F5,$FF,$C2,$7A,$FB,$35,$FF,$0E,$46,$23,$FB,$CD,$A3
	.dc.b	$11,$FE,$BB,$CB,$FD,$77,$E9,$7F,$89,$32,$FD,$34,$FF,$AF,$23,$F8
	.dc.b	$DC,$70,$8F,$AE,$D5,$C9,$FE,$C4,$F2,$7F,$90,$C7,$FB,$5E,$FF,$9B
	.dc.b	$C7,$FB,$39,$FE,$A9,$F0,$3F,$E1,$D9,$F9,$71,$FE,$5A,$1F,$29,$FE
	.dc.b	$5C,$C7,$E3,$CE,$A7,$FD,$3B,$1F,$3F,$1F,$F1,$C7,$B1,$BE,$3F,$99
	.dc.b	$4F,$B9,$C6,$1C,$63,$B6,$1F,$76,$3F,$97,$FA,$5D,$FA,$3F,$05,$1F
	.dc.b	$8B,$94,$74,$AF,$27,$57,$E3,$78,$7F,$C3,$77,$F8,$43,$FC,$6F,$D5
	.dc.b	$87,$FA,$BD,$7F,$86,$DF,$F1,$F1,$FC,$D3,$93,$DD,$D9,$FA,$0A,$D4
	.dc.b	$F0,$CB,$F4,$EF,$F5,$13,$FA,$A3,$FA,$EF,$63,$FC,$31,$F4,$3F,$E0
	.dc.b	$CF,$13,$FB,$67,$C2,$7F,$80,$B3,$FF,$24,$63,$FD,$D7,$CE,$FE,$F7
	.dc.b	$F7,$5F,$DD,$6F,$FB,$AD,$FF,$BA,$FB,$3F,$B8,$E7,$F6,$71,$FF,$28
	.dc.b	$74,$7F,$6F,$D7,$FD,$55,$1D,$27,$F9,$9C,$62,$3F,$D3,$79,$FF,$95
	.dc.b	$7A,$FE,$44,$7F,$95,$FF,$55,$BE,$DF,$FB,$F2,$FE,$A4,$FF,$A6,$A1
	.dc.b	$F9,$8F,$0F,$66,$D6,$AB,$FA,$A7,$BF,$D3,$C4,$63,$72,$C7,$B1,$FC
	.dc.b	$9E,$FF,$9E,$EB,$5F,$F5,$AF,$83,$F6,$F7,$FF,$4A,$70,$7F,$84,$32
	.dc.b	$FE,$85,$C5,$FE,$38,$E2,$FE,$71,$EA,$FF,$3B,$E4,$FE,$75,$7B,$F5
	.dc.b	$3F,$D7,$77,$BC,$3E,$A7,$9C,$F3,$DE,$7B,$AB,$F9,$17,$9D,$FD,$7F
	.dc.b	$83,$F8,$F7,$37,$B7,$17,$6F,$E0,$DB,$1F,$88,$F1,$BF,$C7,$5D,$FF
	.dc.b	$ED,$DF,$3B,$E1,$1F,$C9,$3B,$BF,$90,$D7,$E3,$BC,$1F,$E3,$E7,$F1
	.dc.b	$E7,$B7,$87,$F9,$E7,$F7,$38,$FE,$8D,$FE,$32,$3C,$4F,$F7,$57,$8D
	.dc.b	$F3,$CF,$D6,$F8,$D8,$9F,$F0,$BF,$8F,$FE,$EF,$F3,$3E,$A1,$FB,$2F
	.dc.b	$27,$FD,$AB,$D3,$FC,$07,$8B,$F1,$7C,$3F,$8D,$7F,$D3,$98,$7F,$99
	.dc.b	$E3,$FC,$75,$E5,$7F,$AD,$7A,$BF,$D6,$F3,$93,$F7,$E7,$F1,$1C,$9E
	.dc.b	$26,$63,$F9,$17,$0E,$2B,$FD,$49,$39,$9F,$F4,$DD,$BC,$07,$D2,$FF
	.dc.b	$4E,$6C,$FC,$F3,$FD,$73,$3F,$EA,$FF,$77,$F8,$6F,$9D,$FC,$83,$C4
	.dc.b	$FF,$1B,$57,$F0,$D8,$69,$FC,$76,$3E,$59,$FE,$96,$E1,$FF,$5A,$57
	.dc.b	$F1,$C8,$F9,$FE,$77,$39,$FE,$33,$E7,$D9,$FE,$30,$F4,$AF,$E4,$11
	.dc.b	$FE,$79,$6B,$FD,$73,$1F,$F5,$1F,$2F,$FA,$F3,$3F,$F6,$37,$E8,$7F
	.dc.b	$E1,$FA,$FF,$59,$5A,$31,$1F,$C7,$FF,$45,$FF,$76,$F1,$FF,$A1,$EF
	.dc.b	$FD,$9D,$1F,$87,$F9,$1B,$5F,$D8,$BB,$3F,$B5,$FC,$4F,$FB,$16,$3F
	.dc.b	$20,$76,$ED,$FE,$E5,$C1,$FE,$D7,$FD,$57,$FD,$93,$3F,$E0,$03,$E5
	.dc.b	$38,$7F,$D8,$27,$FA,$F3,$BF,$FD,$02,$7F,$80,$B6,$3F,$EC,$7D,$3F
	.dc.b	$97,$75,$7F,$B5,$EB,$FB,$E1,$F5,$67,$FD,$B7,$9F,$FB,$6A,$3F,$D8
	.dc.b	$FE,$0F,$FA,$30,$FF,$09,$38,$3F,$97,$1F,$D7,$0F,$BD,$F0,$5F,$E9
	.dc.b	$23,$FB,$A3,$FE,$0A,$3F,$A3,$3F,$EC,$BF,$53,$FD,$95,$F0,$9F,$EF
	.dc.b	$58,$C7,$93,$FC,$0F,$C8,$7C,$2E,$70,$FF,$66,$E1,$FE,$D1,$FD,$2F
	.dc.b	$11,$FE,$50,$F6,$EF,$9B,$FE,$A7,$FC,$4A,$BF,$EA,$CF,$C7,$37,$EF
	.dc.b	$FF,$90,$39,$3F,$CF,$AF,$FB,$1B,$E0,$72,$FF,$B2,$FF,$2F,$EB,$E7
	.dc.b	$7F,$A8,$6B,$FD,$06,$FE,$C1,$B9,$FF,$6C,$7F,$D0,$FE,$A7,$F9,$6F
	.dc.b	$E1,$7C,$AF,$F9,$77,$E1,$6F,$BF,$FE,$5B,$F8,$6F,$F6,$27,$BF,$FD
	.dc.b	$4F,$7F,$F6,$0F,$05,$7F,$DC,$7C,$1F,$EE,$2D,$7C,$1E,$7F,$F9,$FB
	.dc.b	$6F,$99,$E5,$B7,$FD,$B4,$7F,$AF,$8F,$A0,$7E,$C7,$CD,$E0,$FF,$8F
	.dc.b	$27,$FA,$B5,$FA,$EB,$33,$FA,$0F,$77,$FD,$96,$7D,$A7,$ED,$78,$3F
	.dc.b	$D5,$DF,$C0,$BF,$E5,$49,$FF,$14,$46,$0F,$E0,$1E,$F7,$FB,$50,$E4
	.dc.b	$FF,$93,$25,$F0,$1F,$D7,$63,$F8,$0F,$A7,$8F,$4F,$71,$FA,$D9,$D7
	.dc.b	$99,$EC,$F3,$FF,$D7,$F5,$FE,$C7,$9C,$D7,$E3,$E7,$FC,$5F,$1F,$D9
	.dc.b	$23,$AB,$6B,$F9,$4F,$C0,$59,$F8,$F8,$FF,$AE,$F9,$3F,$D8,$B8,$FF
	.dc.b	$AF,$AB,$AD,$7F,$09,$AF,$ED,$AF,$FB,$4B,$13,$83,$F8,$DC,$5C,$3F
	.dc.b	$D5,$F9,$FB,$27,$FC,$E5,$B7,$F7,$47,$FD,$AB,$A7,$FA,$07,$6F,$D7
	.dc.b	$B6,$3E,$66,$15,$EB,$5F,$41,$F9,$2F,$FA,$F5,$F5,$B5,$FC,$3B,$F8
	.dc.b	$67,$FD,$1C,$6B,$83,$F0,$2F,$AB,$83,$FD,$58,$F2,$E1,$CB,$FC,$E2
	.dc.b	$91,$FD,$F3,$8C,$41,$C9,$FE,$A7,$7F,$37,$3F,$A6,$E6,$FD,$17,$47
	.dc.b	$F8,$6A,$31,$CF,$FE,$C0,$11,$FE,$C4,$8C,$A9,$4F,$FB,$32,$7F,$67
	.dc.b	$8F,$E5,$9A,$FE,$7D,$AF,$D6,$3F,$D2,$7E,$1F,$FB,$17,$B0,$4E,$CF
	.dc.b	$F5,$94,$66,$33,$DB,$8A,$FB,$F1,$FE,$CA,$B5,$7C,$3C,$FE,$16,$9B
	.dc.b	$BF,$F7,$2F,$5D,$9F,$F6,$5D,$F0,$9F,$E3,$BE,$1F,$FB,$17,$0F,$E4
	.dc.b	$7E,$E7,$F8,$93,$C1,$E4,$FF,$B3,$BC,$25,$7F,$27,$F0,$FB,$75,$7F
	.dc.b	$81,$32,$E9,$97,$E1,$3C,$6F,$C5,$C7,$F8,$9A,$13,$FD,$35,$FE,$75
	.dc.b	$D1,$FE,$CD,$7F,$27,$3F,$D2,$9F,$CE,$F9,$3F,$D2,$7C,$BF,$CC,$F9
	.dc.b	$5F,$D2,$27,$FD,$74,$7F,$AF,$9F,$F6,$9D,$71,$8E,$E7,$F8,$D6,$FF
	.dc.b	$DA,$CF,$6F,$2F,$FA,$E7,$6F,$E9,$1B,$1F,$EC,$8F,$33,$FD,$AD,$DB
	.dc.b	$FE,$35,$8E,$67,$F6,$23,$FD,$95,$83,$CF,$FE,$E4,$D8,$FE,$E4,$7E
	.dc.b	$07,$CE,$FF,$77,$4F,$F7,$A8,$FD,$9F,$A1,$FD,$F9,$FC,$F3,$73,$FF
	.dc.b	$7A,$7A,$3F,$ED,$E3,$FC,$11,$87,$F8,$32,$BF,$C9,$2F,$F8,$5F,$7F
	.dc.b	$F5,$B1,$FE,$B7,$F4,$BF,$C9,$3E,$9F,$F8,$63,$4F,$F7,$F3,$FE,$DF
	.dc.b	$3F,$82,$70,$EB,$0F,$A3,$E3,$FF,$8A,$B3,$65,$FE,$2A,$9E,$B9,$7F
	.dc.b	$B0,$FD,$4F,$F1,$5F,$AD,$FE,$34,$9F,$8B,$CB,$FE,$E4,$8C,$3F,$CE
	.dc.b	$CF,$F6,$9F,$91,$FE,$3F,$9F,$F7,$97,$2F,$FB,$87,$C7,$F1,$E7,$67
	.dc.b	$FD,$66,$7F,$81,$6F,$FC,$98,$71,$7F,$90,$C7,$E3,$27,$F8,$F4,$7F
	.dc.b	$8F,$A7,$4F,$F8,$62,$32,$FF,$94,$4F,$F3,$2E,$5F,$E4,$FF,$5B,$FD
	.dc.b	$F2,$FF,$92,$E3,$FC,$91,$87,$F9,$1E,$FF,$AE,$ED,$FC,$C3,$CF,$FA
	.dc.b	$17,$F1,$38,$97,$FE,$32,$C3,$E9,$FA,$BF,$F0,$06,$FF,$09,$18,$3F
	.dc.b	$C6,$47,$F8,$D0,$FF,$6E,$74,$FF,$8B,$DF,$F7,$1F,$57,$E7,$DF,$F2
	.dc.b	$CE,$3F,$EF,$1F,$3B,$FD,$C1,$87,$F7,$37,$FD,$77,$87,$EC,$BC,$0F
	.dc.b	$DC,$77,$FF,$87,$7E,$0F,$FC,$2B,$E2,$7F,$27,$3F,$94,$57,$F8,$53
	.dc.b	$C7,$FE,$26,$43,$F8,$AB,$97,$03,$E1,$3F,$C4,$7C,$FF,$84,$D3,$FD
	.dc.b	$F3,$E3,$7F,$1F,$F3,$7F,$94,$79,$9F,$CB,$5F,$E5,$58,$FF,$2C,$F3
	.dc.b	$7F,$95,$C3,$E1,$BF,$BF,$71,$7E,$E3,$97,$F9,$40,$CB,$FD,$A5,$C5
	.dc.b	$F6,$BD,$3F,$EC,$B5,$FC,$B7,$DC,$FF,$6B,$7A,$9F,$7B,$D4,$FE,$DD
	.dc.b	$3F,$E1,$47,$6F,$F6,$89,$E9,$18,$8F,$E6,$DE,$17,$F5,$DE,$DF,$E2
	.dc.b	$D9,$F8,$73,$FE,$59,$AF,$B0,$9B,$60,$FE,$9B,$B7,$F3,$1E,$2F,$89
	.dc.b	$5E,$9C,$7E,$57,$27,$CA,$6B,$F4,$39,$3F,$9B,$72,$FD,$53,$FA,$E4
	.dc.b	$FD,$BF,$61,$CE,$F3,$BF,$99,$F5,$7F,$B6,$A3,$F6,$FC,$DF,$84,$BF
	.dc.b	$F0,$F6,$7E,$23,$F9,$0F,$17,$AF,$AF,$FA,$A7,$8F,$ED,$A7,$77,$E4
	.dc.b	$6F,$D1,$E6,$7F,$8C,$F9,$7F,$C5,$55,$F9,$A9,$E5,$8F,$F5,$DF,$07
	.dc.b	$F8,$EB,$FC,$80,$FE,$F9,$3F,$A6,$9F,$E2,$BE,$0F,$E9,$A3,$FD,$5D
	.dc.b	$D1,$FA,$CA,$FE,$1F,$E3,$FF,$AE,$B1,$73,$7F,$0A,$9F,$F0,$FD,$7F
	.dc.b	$BB,$B1,$FE,$0D,$1F,$EE,$4D,$7F,$0D,$3F,$C0,$23,$FC,$4B,$E0,$7F
	.dc.b	$71,$AF,$85,$1F,$D5,$9F,$AB,$BF,$FB,$6E,$3F,$A7,$C7,$E6,$E3,$11
	.dc.b	$F2,$63,$F8,$43,$DE,$3F,$84,$3F,$ED,$31,$EE,$A3,$F8,$54,$7E,$0B
	.dc.b	$DC,$C7,$83,$FC,$AE,$FF,$51,$1F,$9D,$8C,$47,$85,$F8,$B3,$FE,$70
	.dc.b	$8F,$F6,$C1,$FA,$FF,$77,$9F,$97,$FF,$00,$23,$27,$1F,$33,$FC,$81
	.dc.b	$E1,$75,$F3,$7F,$83,$E3,$F6,$6B,$FD,$69,$8F,$FB,$9F,$CE,$CF,$47
	.dc.b	$E0,$75,$FF,$62,$4F,$C4,$AF,$F0,$D3,$FB,$3F,$82,$CB,$F8,$D9,$FF
	.dc.b	$42,$79,$5F,$CA,$B1,$FE,$E5,$7F,$EC,$1E,$3F,$DD,$57,$FB,$12,$3E
	.dc.b	$35,$FF,$90,$56,$19,$7F,$A7,$CF,$FA,$3F,$D1,$FF,$5E,$F3,$7F,$04
	.dc.b	$9F,$E4,$57,$E5,$7F,$B0,$FC,$D7,$F2,$38,$C4,$7E,$E7,$0F,$F5,$CF
	.dc.b	$7F,$F2,$41,$E0,$BE,$07,$F9,$2F,$87,$FC,$97,$E0,$EB,$FE,$4D,$FC
	.dc.b	$0F,$FD,$6F,$E3,$FF,$22,$FC,$0F,$F9,$0E,$7F,$5F,$B7,$F9,$07,$95
	.dc.b	$FE,$4E,$F8,$5F,$F5,$66,$3F,$CB,$F1,$FF,$54,$7C,$DF,$D2,$3F,$9B
	.dc.b	$FA,$DE,$CE,$1F,$F3,$DF,$1F,$C6,$E7,$7E,$AB,$F6,$0F,$CD,$6B,$FC
	.dc.b	$33,$D2,$7E,$0B,$FD,$09,$B8,$7C,$09,$FE,$31,$1F,$C7,$7B,$FF,$28
	.dc.b	$BE,$0F,$81,$D2,$7F,$9D,$7A,$79,$BE,$51,$FC,$9A,$7F,$A7,$CF,$3E
	.dc.b	$C3,$E0,$AF,$D3,$36,$BF,$5B,$E5,$7A,$A7,$9C,$7F,$83,$33,$FE,$2D
	.dc.b	$B7,$C1,$9F,$ED,$33,$FA,$9A,$6F,$E5,$8F,$F0,$8C,$62,$7C,$B8,$FF
	.dc.b	$9F,$F8,$7E,$74,$FF,$4D,$F5,$7F,$CF,$7C,$7F,$CC,$F0,$FD,$E4,$FF
	.dc.b	$30,$E7,$FF,$38,$F3,$FF,$30,$C3,$F9,$87,$45,$79,$F4,$7F,$99,$3A
	.dc.b	$BF,$0D,$87,$F2,$CC,$7F,$B0,$76,$FF,$30,$DB,$E3,$F6,$FF,$96,$DF
	.dc.b	$EB,$47,$DC,$9F,$E8,$5E,$07,$F3,$53,$FB,$E7,$81,$FA,$4C,$3F,$4F
	.dc.b	$3F,$A4,$AF,$EF,$F7,$FC,$E3,$C9,$FF,$72,$79,$3F,$D2,$FC,$3F,$E7
	.dc.b	$C7,$ED,$F0,$FE,$73,$7F,$B7,$3F,$90,$5F,$F3,$47,$F6,$D1,$FC,$07
	.dc.b	$0E,$D0,$F6,$7F,$C7,$FE,$37,$F4,$9F,$1F,$F8,$4D,$7F,$92,$2B,$E7
	.dc.b	$DF,$F3,$B8,$FE,$C9,$E4,$FD,$7F,$27,$F8,$94,$FF,$18,$AF,$F8,$01
	.dc.b	$5F,$CF,$1E,$3E,$67,$DB,$8F,$EF,$9E,$5F,$DF,$8C,$42,$3F,$C0,$1E
	.dc.b	$67,$F4,$1A,$FF,$38,$EF,$FC,$B7,$BB,$FC,$85,$C5,$FC,$BF,$8B,$FD
	.dc.b	$8B,$5F,$8E,$6B,$FC,$55,$FE,$81,$8F,$F7,$46,$BF,$C3,$A1,$F8,$9F
	.dc.b	$F9,$93,$C7,$FF,$76,$BF,$E6,$68,$A7,$F9,$50,$FF,$78,$F9,$FF,$25
	.dc.b	$7E,$29,$B6,$9B,$11,$FE,$46,$72,$7F,$A3,$0F,$ED,$5C,$8F,$4B,$C7
	.dc.b	$1C,$7C,$8F,$F1,$14,$62,$38,$01,$C3,$FE,$52,$F8,$47,$E9,$9F,$E7
	.dc.b	$19,$D7,$EB,$3B,$9C,$C7,$F5,$9F,$A6,$FF,$17,$46,$23,$FC,$F6,$70
	.dc.b	$D9,$0F,$D5,$BE,$23,$67,$DD,$AF,$87,$F4,$3F,$F7,$07,$17,$F4,$89
	.dc.b	$FE,$7D,$49,$CC,$67,$8F,$FB,$44,$E6,$39,$9E,$09,$FA,$91,$98,$F8
	.dc.b	$27,$CC,$E0,$F2,$F1,$F4,$8F,$EA,$FB,$FC,$38,$FF,$06,$7A,$1F,$D6
	.dc.b	$B2,$FF,$08,$36,$BC,$DF,$B0,$FF,$9F,$5E,$47,$FB,$E5,$FF,$3E,$C3
	.dc.b	$EA,$7F,$8A,$4F,$D1,$C3,$3F,$A3,$7E,$1E,$8F,$FA,$FF,$6C,$D7,$FA
	.dc.b	$73,$1F,$F6,$CD,$6A,$3F,$C5,$17,$F8,$68,$C4,$7F,$A1,$CF,$BB,$B2
	.dc.b	$1F,$A3,$F7,$3F,$C6,$D3,$F9,$4D,$75,$18,$8F,$C8,$D7,$1F,$D5,$FF
	.dc.b	$DC,$3C,$FF,$E1,$1F,$CF,$7D,$07,$E1,$E9,$CD,$FE,$9F,$8C,$3F,$09
	.dc.b	$FF,$4A,$6A,$AF,$F0,$75,$7D,$8C,$DE,$1F,$F3,$DD,$3F,$43,$3E,$98
	.dc.b	$C5,$FC,$CF,$CF,$7E,$7C,$FE,$6F,$4F,$F1,$3C,$7E,$76,$3F,$D3,$5F
	.dc.b	$03,$FE,$60,$FD,$B9,$FB,$1F,$C3,$AD,$7B,$DF,$DF,$65,$FA,$1E,$70
	.dc.b	$FD,$34,$E7,$EF,$B1,$1F,$E6,$FF,$44,$3F,$12,$7D,$3E,$A3,$50,$FE
	.dc.b	$01,$C4,$7F,$9D,$E1,$F5,$FD,$3D,$8F,$F5,$38,$FD,$23,$FC,$7D,$B7
	.dc.b	$AE,$33,$F5,$BF,$07,$3F,$DD,$79,$FF,$A7,$5F,$3F,$FE,$CF,$BF,$F4
	.dc.b	$AC,$7E,$AF,$E5,$79,$BC,$0F,$D7,$4F,$37,$FD,$61,$F0,$EF,$7D,$F9
	.dc.b	$03,$F0,$2F,$83,$F2,$5F,$E9,$A9,$CF,$D5,$62,$FF,$D6,$A7,$1B,$7D
	.dc.b	$9F,$E4,$7D,$4F,$AF,$93,$B7,$FC,$BB,$1F,$5F,$97,$E3,$F3,$EE,$BF
	.dc.b	$8B,$47,$CA,$E7,$C4,$7D,$18,$FA,$1C,$D8,$E3,$F8,$5C,$3F,$EE,$FC
	.dc.b	$7E,$0D,$FF,$94,$B8,$BF,$CD,$11,$88,$FC,$34,$FE,$DB,$D4,$F2,$E9
	.dc.b	$FE,$6D,$6F,$C2,$FE,$A5,$FE,$7D,$E1,$FF,$32,$CB,$BE,$19,$3E,$61
	.dc.b	$DF,$CE,$C2,$8F,$2F,$A5,$FB,$3F,$CA,$3C,$53,$FC,$CF,$3F,$E6,$98
	.dc.b	$27,$32,$D7,$FD,$01,$18,$AF,$E1,$BF,$9A,$FF,$09,$3C,$A7,$E9,$AB
	.dc.b	$F8,$03,$FA,$BD,$7F,$D1,$0F,$EA,$B1,$FE,$09,$F8,$37,$F1,$13,$FA
	.dc.b	$D7,$C6,$BF,$54,$FE,$B4,$DC,$3F,$9B,$D2,$31,$1F,$CB,$3D,$4F,$E9
	.dc.b	$91,$8F,$4B,$F2,$9F,$57,$FE,$F3,$8C,$57,$F2,$8B,$E5,$D1,$D8,$FE
	.dc.b	$D5,$F1,$DF,$AA,$9E,$ED,$7F,$9C,$F9,$9F,$D6,$63,$9C,$7F,$A3,$B2
	.dc.b	$C6,$7F,$8E,$58,$7F,$94,$23,$F2,$39,$7F,$BE,$9E,$E7,$F7,$7E,$36
	.dc.b	$BF,$DE,$DB,$87,$F4,$C5,$B7,$CD,$6C,$7F,$B3,$8F,$EE,$C7,$F8,$97
	.dc.b	$3F,$F6,$C4,$FD,$1E,$4F,$C1,$3D,$4F,$07,$17,$DB,$CF,$F0,$CF,$DD
	.dc.b	$BA,$7E,$30,$FE,$39,$C6,$BF,$C6,$72,$F1,$CB,$F2,$95,$FB,$C3,$27
	.dc.b	$13,$FC,$01,$5C,$63,$FC,$17,$C5,$FC,$27,$3F,$F7,$75,$FE,$8B,$0E
	.dc.b	$38,$7F,$B6,$6F,$FC,$14,$7E,$68,$AF,$76,$BF,$A1,$7F,$C2,$1D,$3F
	.dc.b	$C8,$BA,$7F,$45,$D3,$FB,$F8,$FE,$45,$C9,$8B,$FD,$FC,$7F,$25,$EA
	.dc.b	$C4,$78,$7C,$2F,$F5,$3F,$80,$F0,$5E,$07,$F0,$FE,$FF,$D7,$56,$A7
	.dc.b	$35,$FC,$5B,$54,$E6,$FF,$8B,$78,$5C,$31,$F8,$C7,$F2,$AE,$5C,$3F
	.dc.b	$B9,$3D,$D5,$FB,$4E,$4F,$94,$FF,$35,$F1,$04,$2F,$89,$F9,$2C,$BC
	.dc.b	$37,$FE,$31,$9F,$E7,$FE,$2F,$28,$F8,$3C,$1F,$D3,$0F,$E9,$BE,$37
	.dc.b	$F4,$6B,$E3,$3F,$C8,$63,$FA,$7C,$62,$7F,$93,$3A,$F2,$3F,$AE,$79
	.dc.b	$7E,$A7,$FA,$D4,$67,$E2,$1F,$2E,$F3,$BF,$B3,$79,$BF,$03,$BB,$FA
	.dc.b	$67,$8F,$FE,$AD,$F3,$FF,$40,$7F,$09,$C7,$FB,$BE,$DF,$A5,$F4,$BF
	.dc.b	$BE,$FA,$5F,$8D,$7F,$C0,$31,$FC,$85,$FF,$01,$F3,$58,$FD,$67,$A9
	.dc.b	$FE,$0C,$F0,$7F,$93,$7A,$BF,$E1,$1F,$13,$FD,$1D,$97,$F8,$62,$FF
	.dc.b	$5C,$7E,$BA,$3F,$A5,$46,$63,$F8,$29,$FC,$BF,$D7,$FF,$6B,$7A,$1F
	.dc.b	$E0,$3F,$5F,$FC,$D7,$5F,$E2,$6C,$7F,$54,$7F,$AA,$63,$9D,$7F,$8C
	.dc.b	$73,$E5,$87,$F8,$C4,$FF,$09,$DF,$F8,$BB,$83,$F5,$11,$FC,$F7,$D8
	.dc.b	$FF,$1E,$BF,$E6,$CA,$FC,$43,$FE,$7B,$F3,$BF,$75,$5F,$DD,$DF,$C2
	.dc.b	$47,$FA,$1F,$E0,$7F,$78,$FF,$BC,$FE,$7B,$F8,$CC,$62,$3F,$CB,$39
	.dc.b	$E3,$4E,$1E,$E7,$FA,$46,$BF,$CB,$D8,$FE,$EF,$DA,$46,$30,$FF,$36
	.dc.b	$70,$7F,$A0,$63,$11,$FE,$B0,$6B,$FE,$29,$E6,$70,$F2,$F7,$7B,$1E
	.dc.b	$7F,$7F,$EE,$DF,$FA,$07,$2F,$F3,$57,$BB,$E8,$9F,$F2,$27,$BF,$FC
	.dc.b	$5F,$E0,$7F,$D2,$55,$FE,$17,$E0,$FE,$6B,$18,$D3,$FD,$01,$D5,$FC
	.dc.b	$76,$BF,$8A,$C7,$A3,$91,$18,$8F,$F1,$F1,$FE,$7B,$70,$FF,$90,$77
	.dc.b	$FF,$4F,$47,$58,$DC,$7F,$44,$8C,$47,$FA,$A9,$AF,$FA,$D8,$FF,$5C
	.dc.b	$9F,$E8,$83,$FD,$65,$E8,$7F,$86,$9F,$F2,$73,$78,$7F,$D5,$9E,$96
	.dc.b	$66,$9F,$EB,$B3,$FD,$74,$EF,$E0,$7B,$C7,$F9,$79,$BC,$7F,$EB,$8B
	.dc.b	$FF,$4C,$EF,$FE,$6D,$F8,$2F,$F5,$7B,$FE,$6F,$CB,$FD,$71,$C7,$E3
	.dc.b	$D7,$FD,$5E,$FF,$B9,$39,$BF,$D4,$AD,$7F,$D4,$99,$F8,$E6,$BE,$DD
	.dc.b	$7F,$D4,$6F,$FB,$16,$7F,$4A,$7E,$9B,$E5,$FF,$D3,$71,$FE,$3A,$E2
	.dc.b	$FF,$4E,$F5,$7E,$82,$3F,$D2,$E7,$FB,$57,$2F,$F3,$96,$3F,$E9,$1E
	.dc.b	$CF,$D9,$CF,$D5,$F7,$BF,$65,$97,$F5,$BE,$9F,$B7,$C3,$FE,$92,$AF
	.dc.b	$DF,$E1,$F7,$B5,$C4,$7E,$08,$FF,$6D,$47,$0C,$1D,$FF,$E8,$1C,$7F
	.dc.b	$7B,$55,$FF,$51,$78,$1F,$26,$EC,$FB,$E7,$F3,$EF,$FA,$2D,$FF,$4A
	.dc.b	$BD,$E3,$FD,$19,$87,$83,$0F,$F1,$0D,$7C,$63,$F8,$5E,$5F,$E8,$6F
	.dc.b	$19,$E3,$7F,$2E,$F1,$5F,$38,$9C,$FC,$FE,$EB,$53,$97,$F9,$AE,$3F
	.dc.b	$CD,$FC,$8F,$E6,$D3,$9F,$A8,$FF,$45,$F9,$7F,$E9,$3A,$FE,$84,$FF
	.dc.b	$A5,$7E,$A3,$FD,$19,$E7,$FF,$A3,$0A,$7F,$55,$3F,$AB,$FC,$EE,$1C
	.dc.b	$7D,$65,$7F,$AD,$7D,$37,$FA,$83,$C8,$F7,$71,$7F,$28,$CB,$FA,$9F
	.dc.b	$99,$FE,$91,$E6,$FF,$45,$7B,$3F,$E8,$53,$FB,$AC,$FF,$58,$71,$F4
	.dc.b	$1F,$E8,$7F,$6B,$FB,$D7,$B5,$FE,$8F,$97,$D4,$E5,$FE,$F2,$3E,$6F
	.dc.b	$FD,$63,$9F,$88,$FF,$44,$E5,$FD,$A3,$8B,$FB,$41,$FE,$B4,$CF,$FC
	.dc.b	$22,$FC,$35,$E4,$7F,$B1,$76,$E5,$7F,$E8,$9B,$FF,$55,$76,$FF,$A3
	.dc.b	$A3,$B7,$65,$7C,$18,$7F,$0E,$AF,$CE,$AF,$2B,$A4,$E9,$FF,$3C,$CF
	.dc.b	$F9,$4A,$7F,$CF,$8F,$F3,$19,$E7,$F0,$3A,$C3,$FC,$EF,$3F,$B5,$AF
	.dc.b	$F2,$6B,$F9,$13,$F3,$F5,$FE,$7A,$7F,$CA,$13,$F9,$47,$B7,$C1,$FF
	.dc.b	$03,$F8,$7F,$F3,$2C,$7F,$26,$3F,$CF,$1F,$0B,$93,$F7,$4E,$0F,$ED
	.dc.b	$A7,$FB,$A7,$D0,$FF,$3C,$47,$EF,$EF,$FC,$EF,$3F,$C6,$CA,$7F,$25
	.dc.b	$5F,$0F,$FE,$77,$F2,$7F,$09,$18,$F8,$4F,$E4,$93,$E4,$A0,$F3,$3F
	.dc.b	$CF,$2F,$88,$CB,$FE,$7A,$8E,$10,$C7,$FC,$F4,$FF,$1B,$8C,$BF,$EE
	.dc.b	$4C,$FF,$92,$75,$7C,$AC,$BF,$83,$CF,$83,$C7,$FE,$13,$39,$8F,$F3
	.dc.b	$CE,$7F,$E0,$48,$C4,$7C,$98,$FF,$39,$BF,$DF,$0F,$EA,$2E,$38,$FF
	.dc.b	$17,$3F,$E7,$29,$E0,$FE,$15,$FC,$06,$FF,$AA,$8F,$C5,$47,$8B,$93
	.dc.b	$F9,$C7,$27,$F9,$82,$B9,$C7,$E1,$6B,$F3,$35,$CA,$7F,$88,$47,$E7
	.dc.b	$B9,$7F,$42,$3C,$0F,$F3,$4B,$FE,$E7,$3F,$CB,$73,$F9,$C9,$FD,$A7
	.dc.b	$57,$FB,$E3,$AF,$FC,$D9,$D7,$FE,$60,$EB,$FD,$D7,$5D,$AD,$47,$F9
	.dc.b	$C2,$F8,$DF,$EE,$8F,$F3,$A7,$67,$F9,$A2,$3F,$02,$7F,$96,$37,$F8
	.dc.b	$BF,$01,$FE,$67,$EF,$FE,$24,$3B,$FF,$CD,$EF,$F1,$7F,$02,$BF,$C6
	.dc.b	$6F,$FC,$C2,$FF,$1B,$F0,$7F,$CE,$BC,$1F,$00,$1F,$06,$7F,$91,$B9
	.dc.b	$7F,$95,$79,$1F,$CC,$BC,$8F,$E6,$31,$8F,$17,$F9,$63,$63,$F9,$CF
	.dc.b	$8D,$FE,$F9,$F0,$7E,$AC,$FF,$98,$39,$FF,$68,$3A,$FF,$17,$E4,$FF
	.dc.b	$4F,$EB,$FF,$15,$74,$FF,$98,$EB,$FC,$B3,$DB,$FA,$07,$FC,$37,$DA
	.dc.b	$C7,$FA,$CC,$7F,$53,$AF,$F0,$E9,$FE,$B7,$F6,$B8,$5F,$F6,$C9,$FE
	.dc.b	$A7,$5F,$E6,$18,$C4,$7E,$B3,$CC,$FF,$2C,$BF,$DE,$B1,$FF,$10,$D7
	.dc.b	$F7,$0C,$3F,$AF,$CE,$E3,$FC,$B8,$F6,$3F,$4B,$E7,$7F,$7D,$CB,$F4
	.dc.b	$18,$FF,$8E,$5F,$F2,$6D,$FF,$6D,$3F,$C0,$FE,$7F,$F9,$56,$FF,$74
	.dc.b	$3D,$4F,$F2,$DE,$5B,$AF,$F3,$8F,$B3,$FE,$5E,$F6,$7F,$C3,$5D,$9F
	.dc.b	$E5,$2C,$3F,$B7,$AF,$4F,$F4,$AD,$8F,$8D,$7F,$DC,$63,$11,$FC,$4D
	.dc.b	$FF,$2A,$3F,$43,$3F,$F3,$1F,$A9,$FE,$2B,$A7,$AF,$FE,$32,$D3,$F8
	.dc.b	$9B,$6B,$3E,$3F,$FB,$8F,$7F,$DB,$E3,$FE,$49,$F4,$3F,$C8,$11,$88
	.dc.b	$FF,$2A,$CF,$F8,$CE,$3F,$C7,$B1,$F8,$56,$BF,$C8,$3C,$1F,$85,$97
	.dc.b	$FB,$EF,$76,$39,$AF,$F4,$11,$4F,$EC,$11,$F8,$F7,$F8,$19,$83,$FC
	.dc.b	$AF,$3F,$DD,$A3,$FC,$AF,$1F,$E8,$BE,$77,$AF,$FE,$CA,$AF,$EC,$2D
	.dc.b	$7F,$BE,$F1,$3E,$37,$E4,$DF,$F8,$BC,$7D,$AF,$F3,$88,$CE,$9F,$9B
	.dc.b	$FB,$CF,$F2,$AD,$7E,$78,$FE,$7B,$5E,$CA,$7E,$AD,$F0,$31,$69,$FE
	.dc.b	$5A,$BF,$B7,$01,$F2,$3F,$E5,$A8,$CC,$60,$FF,$2F,$BE,$DB,$FC,$0C
	.dc.b	$7C,$1D,$BD,$5A,$77,$E8,$F1,$4F,$87,$BF,$FC,$B7,$87,$E5,$B9,$7F
	.dc.b	$CB,$6F,$EC,$6F,$F2,$B8,$3D,$57,$5F,$FB,$33,$1F,$B1,$C7,$FE,$5D
	.dc.b	$C3,$FB,$4F,$B9,$E9,$F9,$8F,$EB,$B8,$7E,$B2,$7F,$AF,$CE,$71,$E1
	.dc.b	$1F,$E6,$EF,$96,$FD,$1E,$98,$CB,$FD,$AB,$1F,$A6,$F8,$8F,$F3,$0F
	.dc.b	$37,$F0,$49,$67,$FC,$47,$3F,$E0,$58,$FF,$97,$E3,$FC,$21,$EF,$7E
	.dc.b	$D6,$3F,$88,$71,$FC,$DA,$78,$9F,$BF,$9F,$BD,$CF,$E8,$8C,$65,$FC
	.dc.b	$5E,$3F,$C3,$51,$FC,$76,$7F,$87,$5F,$F9,$9A,$BF,$B6,$C7,$E6,$B1
	.dc.b	$F8,$3F,$07,$FC,$2F,$7F,$F4,$1C,$7F,$88,$FC,$6F,$C1,$61,$FE,$CE
	.dc.b	$CF,$F9,$87,$C3,$7F,$99,$5B,$F8,$77,$F9,$97,$9F,$FC,$C3,$E4,$A9
	.dc.b	$F8,$7F,$F9,$8B,$CD,$FE,$82,$3F,$14,$7E,$29,$98,$FE,$8B,$0F,$C6
	.dc.b	$31,$1B,$FC,$5B,$FC,$C1,$E9,$BF,$07,$7E,$0F,$85,$F8,$A5,$FC,$57
	.dc.b	$1F,$8B,$7F,$99,$7B,$7F,$D0,$5F,$8B,$FE,$03,$CC,$FF,$21,$C7,$FA
	.dc.b	$AB,$F0,$CF,$A1,$18,$8F,$CF,$09,$FF,$34,$FC,$07,$F7,$3F,$BE,$7E
	.dc.b	$1B,$FE,$63,$87,$C0,$A5,$F8,$CE,$63,$1F,$88,$7F,$99,$B1,$FF,$4A
	.dc.b	$76,$FF,$99,$BC,$2F,$EC,$DC,$BF,$17,$A3,$F3,$9E,$A7,$F5,$38,$FE
	.dc.b	$D5,$D3,$F1,$7B,$FF,$C1,$91,$F3,$21,$D3,$C3,$0F,$F0,$96,$5F,$B7
	.dc.b	$E1,$FC,$FE,$3E,$D8,$FD,$D7,$1F,$E3,$E3,$11,$EF,$F0,$BF,$CD,$8F
	.dc.b	$FB,$3B,$5F,$33,$FE,$8E,$78,$61,$FC,$03,$D1,$FF,$35,$7A,$BC,$B2
	.dc.b	$FD,$6B,$D3,$2F,$CE,$39,$C7,$F7,$1F,$1D,$BC,$29,$E8,$C9,$7F,$A3
	.dc.b	$9E,$1B,$BF,$59,$FE,$2D,$9F,$F9,$E7,$E7,$FF,$35,$8F,$FB,$43,$C7
	.dc.b	$FD,$EC,$62,$3F,$9C,$FE,$23,$FE,$75,$7F,$C0,$BA,$7E,$D8,$FE,$15
	.dc.b	$F8,$9F,$45,$F2,$4F,$D6,$7B,$D6,$1D,$47,$F9,$83,$E0,$FF,$92,$2F
	.dc.b	$A1,$DB,$6B,$E3,$5F,$F9,$CA,$7B,$7C,$8E,$67,$FA,$33,$CF,$8B,$FC
	.dc.b	$EF,$8E,$8F,$0F,$C2,$B3,$F3,$3F,$E6,$5C,$7F,$73,$87,$F8,$5F,$5F
	.dc.b	$81,$1E,$D9,$D4,$FF,$BD,$35,$FE,$75,$1F,$2E,$3C,$AF,$FA,$89,$FF
	.dc.b	$04,$E3,$FA,$23,$F0,$91,$9F,$83,$FF,$67,$F3,$7F,$A1,$A3,$13,$FE
	.dc.b	$EF,$E5,$EB,$CD,$F7,$E7,$FC,$DD,$87,$4A,$FD,$6D,$F1,$AF,$C3,$47
	.dc.b	$F9,$B4,$FF,$7C,$71,$FF,$70,$BF,$ED,$71,$F9,$1F,$73,$4F,$E5,$4F
	.dc.b	$F1,$B7,$4F,$73,$F1,$93,$FD,$CB,$3F,$F7,$2E,$7F,$02,$7F,$59,$D9
	.dc.b	$FE,$74,$EC,$FE,$21,$7F,$AC,$8F,$E2,$7D,$9F,$E7,$67,$5D,$BF,$E7
	.dc.b	$6C,$BF,$73,$9F,$F9,$C7,$BF,$FC,$DD,$E0,$BC,$0C,$7C,$6B,$C3,$FE
	.dc.b	$0B,$3F,$C1,$34,$FF,$3A,$CE,$6F,$F8,$47,$89,$FE,$4F,$F1,$7E,$41
	.dc.b	$EA,$3E,$2D,$7F,$9B,$BC,$BF,$F3,$AF,$94,$F2,$7F,$CE,$70,$21,$FC
	.dc.b	$9F,$C9,$CB,$89,$FE,$4F,$5C,$E3,$67,$F2,$DF,$2F,$FC,$F3,$C5,$F1
	.dc.b	$7C,$CA,$FF,$BE,$B9,$7F,$87,$DF,$F9,$EA,$3F,$89,$72,$7A,$7B,$3F
	.dc.b	$D9,$3E,$7F,$F9,$DB,$E2,$7F,$CF,$9E,$83,$D2,$D5,$2B,$5F,$13,$6B
	.dc.b	$FA,$4E,$5F,$D4,$27,$FC,$FD,$5F,$E0,$EF,$3D,$D5,$FE,$75,$F5,$7F
	.dc.b	$3F,$E6,$FE,$4D,$BE,$37,$F9,$E7,$D7,$FE,$C9,$83,$CE,$FF,$0B,$67
	.dc.b	$FD,$BA,$3F,$C4,$87,$F8,$A8,$E0,$63,$D8,$FC,$97,$B3,$FD,$DE,$3F
	.dc.b	$DC,$BA,$7F,$A0,$5F,$F7,$AF,$85,$FE,$0D,$F7,$BF,$D1,$1F,$13,$FD
	.dc.b	$E6,$7F,$A5,$ED,$F9,$BF,$98,$FF,$3A,$7B,$DF,$E0,$CD,$3F,$D0,$0F
	.dc.b	$F8,$2D,$FF,$0A,$53,$E0,$3F,$C3,$0F,$FA,$23,$DE,$FF,$4B,$39,$CF
	.dc.b	$F7,$DF,$03,$FE,$21,$7F,$C7,$3C,$5F,$E9,$98,$C7,$9D,$FE,$A3,$C7
	.dc.b	$8E,$1C,$B0,$FF,$7F,$FC,$0F,$FA,$15,$FE,$01,$8F,$F6,$2F,$7F,$FD
	.dc.b	$0B,$1F,$E3,$EF,$82,$FF,$31,$FA,$3F,$E3,$CA,$FD,$BF,$12,$31,$1F
	.dc.b	$C2,$73,$DF,$C4,$FF,$83,$3A,$B4,$FF,$83,$A7,$87,$C2,$FF,$31,$BF
	.dc.b	$F2,$CE,$5F,$E5,$63,$FD,$0A,$7D,$3E,$0A,$FF,$98,$76,$FF,$2F,$5F
	.dc.b	$FA,$23,$CA,$FF,$2D,$3F,$E5,$7F,$88,$FF,$35,$3E,$CC,$FF,$D0,$77
	.dc.b	$FE,$71,$BF,$E1,$B3,$FC,$43,$0C,$CF,$F9,$E6,$78,$CF,$FA,$93,$54
	.dc.b	$62,$3F,$D0,$D5,$8C,$7F,$A3,$D9,$FC,$12,$1F,$29,$FE,$8C,$74,$7F
	.dc.b	$AD,$1F,$F6,$94,$70,$8D,$E1,$C7,$D4,$C5,$7F,$B5,$DB,$79,$AE,$55
	.dc.b	$FE,$8C,$BE,$5F,$37,$E1,$F8,$AE,$F5,$FE,$89,$3F,$DC,$1F,$47,$FE
	.dc.b	$7D,$8C,$E9,$F8,$59,$57,$FA,$1F,$F3,$CF,$F2,$8C,$7A,$FF,$5E,$FF
	.dc.b	$42,$7C,$62,$73,$2F,$C9,$7F,$D1,$DC,$9E,$6A,$FF,$61,$E3,$D7,$E6
	.dc.b	$78,$F3,$7B,$DF,$F4,$66,$1E,$53,$E4,$71,$7F,$A4,$DC,$1F,$12,$FF
	.dc.b	$DC,$F7,$E7,$EA,$FF,$51,$72,$FC,$9F,$17,$CA,$FC,$2B,$FF,$48,$9F
	.dc.b	$EE,$2D,$FF,$D7,$BF,$10,$FD,$6B,$FD,$C5,$3F,$5F,$F1,$5F,$F4,$8B
	.dc.b	$F1,$F2,$F1,$F3,$FF,$A4,$7C,$0F,$F7,$76,$5F,$76,$97,$FE,$ED,$C7
	.dc.b	$FD,$2B,$1F,$E9,$93,$A6,$7F,$E8,$EC,$78,$63,$C7,$6F,$F4,$47,$87
	.dc.b	$FE,$81,$8C,$57,$FA,$7F,$F5,$BF,$E2,$50,$FC,$37,$FD,$11,$E4,$FD
	.dc.b	$0E,$6F,$F4,$6E,$7F,$8F,$EE,$FF,$5C,$6F,$E5,$E3,$E1,$1F,$EA,$9F
	.dc.b	$2F,$FC,$4D,$DD,$F9,$4C,$FF,$D5,$9F,$23,$F4,$EA,$E3,$F9,$9E,$1F
	.dc.b	$B1,$96,$23,$E9,$F0,$FF,$A2,$5F,$E8,$3F,$40,$F1,$BF,$D0,$BF,$30
	.dc.b	$FD,$83,$87,$A3,$FD,$00,$42,$B2,$71,$F5,$5E,$A7,$F4,$79,$7E,$9F
	.dc.b	$E1,$F6,$3F,$D1,$A7,$FA,$3B,$1F,$D8,$75,$B8,$7F,$BF,$CE,$A7,$13
	.dc.b	$9D,$7F,$69,$3C,$81,$E7,$F2,$7D,$DE,$E7,$F6,$63,$FB,$77,$E2,$FF
	.dc.b	$E9,$2B,$E5,$79,$9C,$FC,$AF,$07,$85,$7F,$77,$AD,$C6,$E1,$F0,$1F
	.dc.b	$2F,$D2,$FD,$87,$8B,$FE,$D6,$D7,$F1,$91,$FB,$3F,$37,$FD,$8D,$18
	.dc.b	$CF,$F9,$D5,$7D,$15,$F0,$9F,$E0,$D3,$FC,$1D,$F4,$8F,$CF,$FE,$C4
	.dc.b	$67,$E2,$3F,$D3,$32,$FD,$9B,$31,$FE,$9C,$9E,$3F,$71,$8F,$89,$EB
	.dc.b	$3C,$3E,$E7,$FD,$E5,$EB,$7E,$E7,$C0,$FF,$5C,$F4,$7D,$28,$FA,$3D
	.dc.b	$1F,$57,$83,$3D,$1C,$38,$3F,$C9,$47,$A8,$C0,$E3,$F8,$1E,$16,$1F
	.dc.b	$8B,$E1,$FF,$A7,$E7,$F8,$1F,$A3,$FE,$AB,$9F,$93,$19,$78,$F5,$30
	.dc.b	$F9,$13,$FD,$AB,$9B,$FD,$32,$F8,$8C,$09,$FF,$2C,$D7,$2F,$8F,$E3
	.dc.b	$3F,$BF,$AF,$E1,$3F,$31,$FC,$CA,$FF,$CD,$93,$FD,$18,$FE,$7C,$3C
	.dc.b	$76,$5F,$EF,$FA,$FD,$8F,$8D,$FC,$4B,$E5,$3F,$9D,$FC,$C7,$FA,$47
	.dc.b	$0F,$E8,$87,$FA,$44,$FE,$8D,$B5,$7F,$91,$FA,$5F,$D3,$67,$84,$7F
	.dc.b	$26,$F3,$FF,$D2,$BF,$2D,$FD,$2E,$97,$FE,$9E,$F9,$8F,$E9,$77,$4E
	.dc.b	$BF,$31,$FE,$95,$F3,$BF,$A1,$69,$C2,$31,$3F,$E0,$98,$FE,$B9,$1E
	.dc.b	$09,$FE,$91,$86,$31,$F0,$63,$FE,$9C,$C3,$F8,$51,$FE,$11,$C9,$CB
	.dc.b	$98,$FE,$17,$37,$0F,$0C,$7E,$0F,$5F,$EB,$FC,$9F,$3A,$7F,$B6,$57
	.dc.b	$F3,$1E,$1F,$E3,$53,$9B,$FA,$31,$F4,$30,$FC,$67,$07,$2E,$4F,$B7
	.dc.b	$5F,$DD,$79,$3F,$D2,$11,$88,$FE,$C2,$7E,$4D,$C7,$87,$FE,$8B,$8F
	.dc.b	$C4,$1F,$83,$CB,$FD,$26,$E0,$E8,$60,$FC,$E1,$FE,$31,$8E,$4F,$E6
	.dc.b	$BA,$FF,$CC,$B9,$FE,$78,$FE,$EE,$7E,$70,$B1,$C7,$2F,$F3,$AF,$4F
	.dc.b	$F2,$1E,$AA,$FF,$1A,$EB,$FD,$05,$FF,$9D,$E3,$1D,$7F,$E7,$2E,$0F
	.dc.b	$E7,$73,$FC,$6B,$C0,$FE,$09,$DE,$F0,$9E,$0F,$F9,$CB,$5F,$E0,$33
	.dc.b	$25,$E0,$66,$F8,$C7,$F0,$FC,$3A,$78,$3F,$C2,$CF,$E1,$FE,$17,$38
	.dc.b	$F7,$69,$F3,$F9,$7F,$D9,$8F,$CE,$E1,$F0,$78,$DF,$C9,$FE,$42,$BF
	.dc.b	$CB,$3C,$6A,$FF,$A6,$BC,$7F,$F4,$ED,$FF,$73,$8F,$E5,$DF,$1F,$E9
	.dc.b	$E4,$FE,$7C,$7F,$53,$E6,$C3,$FD,$07,$9B,$E3,$DF,$F1,$88,$F8,$FA
	.dc.b	$FF,$3E,$F3,$FF,$AC,$79,$9F,$D6,$E3,$1E,$7F,$F5,$7F,$3B,$FA,$B7
	.dc.b	$9D,$F1,$EF,$73,$F2,$7A,$FF,$57,$DD,$7A,$FF,$CE,$5E,$57,$FB,$D3
	.dc.b	$5F,$E5,$59,$FF,$A5,$67,$F8,$55,$7E,$77,$C8,$FF,$49,$C7,$F7,$DE
	.dc.b	$6F,$F5,$1B,$FE,$7A,$F1,$7F,$DF,$BE,$A7,$F8,$0E,$FF,$95,$7A,$BF
	.dc.b	$E0,$89,$FF,$80,$3E,$BF,$F8,$36,$FF,$C5,$BE,$3F,$FA,$35,$FD,$59
	.dc.b	$FE,$30,$F5,$FF,$D1,$4F,$F8,$64,$FF,$0E,$FA,$D8,$7F,$73,$9F,$F8
	.dc.b	$39,$FF,$0E,$EB,$FE,$20,$7F,$C5,$07,$ED,$3D,$9F,$F1,$6B,$FE,$26
	.dc.b	$D3,$FD,$11,$5F,$C1,$BE,$07,$FC,$77,$3F,$D0,$DA,$FF,$37,$9F,$F4
	.dc.b	$77,$91,$FE,$65,$C7,$FC,$D9,$5A,$AF,$F1,$96,$5F,$C3,$23,$FC,$9F
	.dc.b	$1F,$C5,$EF,$FD,$22,$FF,$99,$A3,$2F,$FA,$B3,$8B,$87,$B9,$FD,$6E
	.dc.b	$5E,$EF,$F9,$B9,$FF,$2F,$57,$FB,$23,$DB,$FF,$51,$FB,$DF,$E6,$0F
	.dc.b	$7F,$FD,$21,$E0,$7F,$93,$FE,$07,$FC,$D7,$7F,$DD,$B4,$FF,$4B,$3F
	.dc.b	$C8,$8F,$F5,$DF,$9F,$FC,$8B,$AF,$FA,$B4,$7F,$3E,$6B,$FE,$2E,$E5
	.dc.b	$FF,$4A,$C7,$F2,$43,$AC,$7F,$B1,$3A,$BF,$9E,$C6,$23,$FC,$99,$3F
	.dc.b	$E9,$57,$FA,$59,$93,$FC,$A2,$53,$FD,$9B,$8F,$FA,$45,$FE,$B1,$F0
	.dc.b	$39,$E3,$FF,$45,$4F,$F9,$B7,$97,$FD,$1D,$5E,$B3,$FC,$9D,$9B,$E0
	.dc.b	$7E,$3C,$7F,$AD,$CF,$F5,$17,$C2,$7F,$A2,$38,$3F,$D0,$FA,$7F,$A1
	.dc.b	$A7,$91,$FE,$64,$AF,$A2,$FE,$7F,$0F,$F5,$D4,$62,$3F,$DA,$F1,$F9
	.dc.b	$23,$C1,$C9,$E9,$D7,$C5,$F0,$BF,$E7,$59,$DF,$C2,$7F,$62,$C3,$F1
	.dc.b	$18,$7F,$B8,$B4,$FF,$00,$75,$7E,$5A,$3F,$D0,$39,$FF,$BA,$1F,$E9
	.dc.b	$FE,$EF,$E7,$71,$FD,$0F,$5F,$FB,$D3,$87,$FD,$09,$5F,$A7,$1D,$7F
	.dc.b	$E7,$F7,$F5,$A7,$FB,$0F,$6C,$3F,$E0,$63,$78,$3B,$FD,$18,$FF,$66
	.dc.b	$9B,$2F,$03,$E4,$EB,$FE,$17,$77,$9F,$E8,$AF,$FB,$6D,$7F,$0B,$D3
	.dc.b	$C3,$97,$AB,$D9,$FF,$73,$DF,$F8,$7D,$FF,$74,$DF,$F9,$F6,$BF,$D8
	.dc.b	$59,$7F,$93,$7C,$4F,$F3,$C7,$89,$FD,$17,$2F,$E4,$3F,$04,$F8,$2F
	.dc.b	$F3,$BD,$7F,$9E,$E3,$FC,$F6,$67,$C8,$47,$F9,$FF,$EA,$3F,$CF,$5E
	.dc.b	$5F,$F3,$42,$9F,$E7,$3F,$A6,$FF,$35,$79,$FF,$E6,$DC,$BF,$CD,$2F
	.dc.b	$F9,$BF,$E8,$B2,$88,$7F,$A3,$23,$6F,$F9,$E7,$E8,$FF,$D3,$1E,$3F
	.dc.b	$E2,$38,$7F,$CD,$7C,$1F,$DD,$CF,$BB,$E6,$FF,$AE,$FD,$9F,$EC,$63
	.dc.b	$D8,$FF,$31,$3F,$E6,$DF,$A4,$B1,$FE,$67,$FA,$6B,$EE,$FF,$9C,$BE
	.dc.b	$9B,$FC,$95,$F0,$3F,$AB,$EE,$B9,$FF,$88,$9F,$F6,$0E,$9F,$E4,$38
	.dc.b	$C1,$F8,$53,$FC,$88,$7E,$90,$FF,$0F,$FC,$37,$F8,$BD,$FA,$AD,$7F
	.dc.b	$DA,$46,$BB,$3F,$D7,$51,$FE,$38,$8F,$F7,$E3,$5D,$D3,$E1,$71,$1F
	.dc.b	$ED,$46,$BF,$ED,$D7,$B3,$FE,$DF,$F4,$FF,$C7,$31,$88,$E1,$7F,$C2
	.dc.b	$5F,$F1,$C1,$E9,$3F,$CA,$3E,$A7,$F8,$D6,$FF,$C7,$11,$8C,$BF,$DD
	.dc.b	$5D,$FF,$A6,$9F,$E9,$B5,$FE,$3B,$AF,$69,$CB,$8B,$EB,$70,$7F,$83
	.dc.b	$77,$FF,$50,$F0,$7F,$84,$E3,$1C,$4E,$1C,$57,$F8,$6A,$3E,$E6,$3F
	.dc.b	$E1,$68,$C7,$17,$F8,$57,$2E,$E7,$73,$FA,$7F,$17,$E2,$6B,$F1,$BC
	.dc.b	$BF,$90,$1C,$BF,$E1,$47,$1E,$C5,$75,$3C,$1F,$CB,$F4,$7F,$84,$BA
	.dc.b	$BF,$41,$3F,$EC,$C7,$F4,$87,$F8,$2F,$C0,$FF,$07,$76,$FF,$B9,$FE
	.dc.b	$03,$1E,$07,$EB,$63,$F5,$97,$FB,$4F,$07,$F6,$77,$FE,$F3,$8F,$F6
	.dc.b	$BF,$3F,$EF,$39,$FF,$D5,$DD,$3F,$C0,$A7,$FD,$13,$C1,$FA,$53,$FD
	.dc.b	$65,$F1,$5F,$53,$8B,$FC,$D6,$7F,$84,$8F,$E4,$20,$FE,$3F,$1F,$E1
	.dc.b	$27,$FA,$F6,$1F,$73,$0F,$F0,$9C,$7F,$3B,$3F,$8E,$DB,$D8,$FF,$0A
	.dc.b	$47,$F2,$58,$C3,$FD,$9E,$7F,$3B,$E4,$7F,$84,$70,$F0,$CF,$87,$C9
	.dc.b	$C4,$7F,$84,$E3,$C9,$3F,$E1,$7C,$7F,$9D,$47,$E4,$BC,$AF,$E7,$D8
	.dc.b	$7F,$83,$23,$F5,$D8,$7F,$A9,$BC,$27,$9B,$FC,$CB,$2F,$E9,$DE,$6E
	.dc.b	$7C,$EF,$E9,$D5,$FD,$4E,$7F,$86,$79,$48,$C4,$7F,$0F,$7F,$AC,$69
	.dc.b	$F4,$5F,$EC,$AF,$CD,$C8,$FF,$83,$1F,$92,$7F,$68,$9F,$F0,$4B,$FE
	.dc.b	$CD,$3F,$A8,$18,$8F,$CB,$E5,$FE,$BE,$F5,$7F,$BE,$6D,$FE,$0D,$F4
	.dc.b	$BF,$89,$57,$F7,$FB,$FF,$6A,$35,$FB,$5E,$27,$BF,$1F,$F1,$EF,$C6
	.dc.b	$7F,$B2,$25,$39,$9F,$F5,$CC,$FE,$06,$72,$7F,$77,$8E,$3E,$77,$E1
	.dc.b	$70,$E1,$0A,$E1,$18,$7B,$36,$5E,$A7,$4B,$FE,$71,$F1,$FF,$EC,$88
	.dc.b	$C6,$1F,$E0,$28,$6F,$FE,$11,$F5,$3E,$66,$9E,$5F,$DC,$FF,$C2,$18
	.dc.b	$FC,$C1,$F9,$9F,$F8,$0F,$F5,$EF,$17,$3F,$F8,$67,$DD,$F3,$FD,$DF
	.dc.b	$4F,$67,$ED,$ED,$FE,$0F,$8F,$F6,$CC,$62,$3D,$ED,$7E,$2D,$7F,$84
	.dc.b	$B1,$F8,$91,$FE,$E0,$F0,$FF,$C2,$5D,$7F,$13,$83,$FB,$BF,$67,$E3
	.dc.b	$F8,$7E,$8E,$3F,$E1,$4C,$7F,$C4,$51,$F3,$3D,$4F,$F3,$77,$0F,$F4
	.dc.b	$0F,$6B,$FD,$D7,$DD,$FE,$FE,$9F,$F0,$B6,$BF,$EE,$ED,$3E,$DF,$67
	.dc.b	$F8,$5B,$2F,$F1,$44,$7D,$EF,$A7,$FF,$09,$C7,$F9,$F3,$07,$46,$30
	.dc.b	$EB,$AB,$2F,$F0,$B5,$FE,$EC,$FE,$05,$C5,$FE,$10,$F6,$BE,$8C,$7F
	.dc.b	$05,$E6,$F2,$47,$F8,$5A,$3F,$6B,$3F,$E1,$6F,$0F,$F8,$76,$1F,$8B
	.dc.b	$19,$FA,$FC,$4F,$A7,$C5,$FE,$17,$C3,$E5,$E8,$C7,$FC,$3F,$5F,$6B
	.dc.b	$4F,$E4,$5E,$D7,$F8,$43,$D9,$3F,$E1,$4F,$23,$FC,$2F,$0F,$DA,$FF
	.dc.b	$C2,$3E,$60,$9F,$A8,$7F,$81,$B4,$FF,$0B,$7E,$63,$F3,$7C,$5F,$F6
	.dc.b	$77,$A1,$FD,$04,$7E,$9D,$B7,$03,$F2,$FE,$17,$FD,$2B,$87,$FC,$B1
	.dc.b	$E6,$7F,$29,$E0,$FC,$F7,$99,$FE,$5C,$F1,$3C,$0F,$F9,$7A,$78,$DF
	.dc.b	$F5,$9F,$65,$EB,$BE,$82,$E3,$FD,$6F,$D9,$FF,$0C,$4E,$7E,$2B,$FC
	.dc.b	$2B,$4F,$D6,$EF,$E9,$DF,$6F,$8B,$E6,$E2,$F4,$3F,$DC,$5C,$CE,$6F
	.dc.b	$F2,$5C,$63,$9B,$F1,$CF,$F0,$1E,$1F,$37,$07,$E1,$63,$1C,$7F,$E7
	.dc.b	$53,$FD,$56,$66,$3E,$13,$FE,$CA,$E2,$E7,$1F,$73,$B3,$F7,$79,$FF
	.dc.b	$A8,$B0,$EA,$DB,$E9,$C9,$FE,$5E,$F9,$7E,$F1,$F3,$3C,$0F,$CB,$7C
	.dc.b	$A6,$23,$B6,$58,$3C,$28,$F2,$E0,$9E,$11,$9C,$FF,$59,$F7,$3F,$8D
	.dc.b	$F8,$2F,$CB,$E1,$FD,$87,$17,$9D,$FE,$3B,$8F,$F7,$C9,$9F,$13,$F6
	.dc.b	$9F,$93,$FF,$0E,$F8,$67,$E5,$DF,$C8,$69,$18,$7F,$3B,$EC,$31,$CD
	.dc.b	$7F,$9D,$AB,$55,$FD,$DF,$C8,$E3,$FB,$F7,$39,$3F,$C7,$3E,$1E,$D7
	.dc.b	$4C,$3F,$C8,$13,$E1,$9E,$7C,$1F,$E0,$16,$BF,$EA,$EC,$3F,$C8,$D8
	.dc.b	$F3,$3D,$1E,$EB,$4F,$E5,$2F,$F8,$F7,$2F,$81,$92,$31,$1E,$BD,$BE
	.dc.b	$84,$38,$38,$F1,$7D,$0C,$DC,$5F,$E4,$88,$C3,$FE,$0F,$AF,$AD,$87
	.dc.b	$DA,$AF,$C6,$C7,$E1,$5F,$C1,$35,$FC,$0B,$5F,$C0,$72,$F8,$BA,$3F
	.dc.b	$86,$CF,$E8,$6B,$A5,$FF,$96,$27,$C3,$1E,$B7,$CD,$5C,$63,$FC,$45
	.dc.b	$D7,$FC,$24,$83,$AB,$94,$6A,$32,$7E,$B7,$1F,$8D,$D9,$E2,$3F,$CB
	.dc.b	$BE,$0B,$C2,$FD,$DF,$83,$FB,$9D,$7F,$7D,$E1,$FF,$02,$87,$83,$FE
	.dc.b	$5E,$AD,$DD,$AD,$78,$7D,$B4,$F6,$F2,$57,$FC,$31,$A7,$ED,$F8,$7E
	.dc.b	$76,$BE,$E8,$FF,$2D,$78,$BF,$C6,$EE,$F8,$DF,$6A,$BF,$82,$47,$F9
	.dc.b	$6F,$83,$E1,$78,$99,$CB,$FC,$8D,$F0,$1F,$EC,$5F,$2B,$F9,$8F,$93
	.dc.b	$FE,$5A,$7F,$99,$F9,$7F,$E5,$4A,$FC,$B7,$2F,$F9,$53,$6F,$E1,$5E
	.dc.b	$47,$F7,$E8,$C4,$7F,$73,$3F,$CA,$F3,$FE,$AA,$3F,$51,$E2,$7F,$95
	.dc.b	$E3,$FC,$89,$EE,$FF,$47,$7F,$29,$E7,$FF,$56,$AF,$F0,$1E,$FF,$E5
	.dc.b	$67,$F8,$24,$FF,$BC,$5F,$EB,$AF,$E6,$FD,$FF,$F1,$C4,$7F,$67,$F4
	.dc.b	$B0,$FF,$98,$FD,$7F,$F5,$9D,$7F,$68,$CF,$FB,$9B,$FE,$4C,$3F,$C5
	.dc.b	$3E,$B7,$F9,$61,$C1,$FE,$25,$F5,$7F,$CB,$17,$FE,$2D,$F5,$F9,$7B
	.dc.b	$1F,$E0,$5A,$FF,$06,$E0,$8C,$47,$F8,$B7,$2F,$F2,$E5,$FF,$B7,$FF
	.dc.b	$2F,$FF,$2E,$7B,$3F,$E0,$FA,$FF,$2B,$3F,$E3,$4F,$6B,$FC,$C1,$E2
	.dc.b	$FF,$8C,$7D,$4F,$F0,$D5,$7F,$8D,$F8,$7F,$CB,$93,$FD,$CA,$73,$CB
	.dc.b	$FC,$A5,$7B,$9F,$E2,$A7,$FC,$BB,$9F,$C3,$3F,$C0,$57,$FE,$65,$F7
	.dc.b	$BE,$87,$BD,$FC,$C7,$E0,$3F,$C7,$79,$FF,$47,$F3,$B1,$E0,$FF,$89
	.dc.b	$7D,$1F,$F3,$2F,$9B,$FE,$45,$E9,$FD,$B6,$9F,$EF,$BF,$57,$F9,$34
	.dc.b	$63,$1F,$EA,$07,$F9,$3E,$7F,$CD,$B8,$71,$3F,$C7,$F3,$E0,$9F,$BE
	.dc.b	$6E,$5F,$89,$FD,$E9,$FF,$0A,$46,$0F,$61,$0F,$F5,$14,$6D,$FF,$7D
	.dc.b	$7E,$3D,$5F,$F3,$33,$77,$FF,$54,$39,$E7,$FF,$2F,$D7,$1E,$CF,$F5
	.dc.b	$67,$C1,$77,$C3,$FD,$00,$7F,$A1,$8F,$F5,$7B,$9C,$7F,$E0,$09,$FE
	.dc.b	$A3,$C5,$F0,$5F,$E1,$4A,$FF,$59,$CF,$E5,$4C,$F9,$3F,$E6,$67,$F3
	.dc.b	$1D,$D8,$9D,$9F,$E9,$4F,$C9,$79,$FE,$5F,$E9,$C1,$DD,$FA,$2B,$F9
	.dc.b	$31,$8B,$F8,$91,$FA,$AE,$4F,$1D,$71,$3E,$05,$3E,$3B,$97,$13,$E9
	.dc.b	$3F,$D9,$93,$9C,$79,$CB,$EC,$FF,$CE,$37,$F1,$E7,$DE,$FC,$9D,$FE
	.dc.b	$0F,$E3,$5D,$3A,$7F,$CE,$EF,$D5,$F9,$3F,$ED,$99,$7E,$F7,$87,$EE
	.dc.b	$E3,$FE,$DB,$E6,$F9,$3A,$7F,$A8,$E3,$FD,$3D,$18,$7F,$D3,$C7,$F9
	.dc.b	$5F,$E0,$BF,$D9,$B9,$FF,$AD,$7C,$AF,$F3,$34,$FF,$AF,$63,$FD,$71
	.dc.b	$CF,$FE,$69,$FC,$1B,$FD,$99,$1F,$EA,$9C,$7F,$C8,$51,$8A,$FF,$76
	.dc.b	$71,$FF,$85,$71,$7C,$9F,$D7,$9F,$F6,$3E,$9F,$E7,$38,$C2,$E7,$C4
	.dc.b	$FF,$BD,$71,$FF,$0F,$61,$FE,$BB,$D3,$FD,$4F,$3A,$3E,$B6,$7F,$EE
	.dc.b	$8F,$1F,$C1,$C1,$FE,$1E,$CB,$FD,$07,$3F,$E5,$78,$C6,$FF,$E1,$58
	.dc.b	$FF,$42,$7E,$2F,$F6,$B3,$FF,$1C,$BF,$EC,$5F,$4B,$FA,$07,$A3,$FE
	.dc.b	$72,$5F,$37,$C3,$4F,$E8,$BF,$1C,$F3,$17,$3D,$36,$0A,$E1,$78,$78
	.dc.b	$56,$E5,$F3,$7A,$F5,$FF,$BA,$7C,$67,$F9,$77,$C9,$9D,$0D,$BF,$DB
	.dc.b	$55,$A9,$D7,$91,$C6,$3F,$D0,$97,$C6,$5F,$15,$E2,$F6,$84,$F9,$27
	.dc.b	$D3,$C1,$F3,$BB,$9D,$46,$4C,$29,$C7,$CF,$E1,$D7,$CF,$FC,$0F,$67
	.dc.b	$FD,$E1,$C9,$FD,$E3,$A3,$F0,$DC,$1F,$EF,$39,$C9,$F9,$57,$FC,$17
	.dc.b	$FA,$5B,$F0,$A7,$D4,$BE,$ED,$F8,$3F,$CC,$F8,$9F,$F4,$4D,$EE,$B5
	.dc.b	$79,$F8,$8F,$F4,$8C,$E7,$07,$E1,$1F,$EC,$5F,$5F,$F0,$5E,$45,$DF
	.dc.b	$FC,$41,$D2,$E9,$F9,$5D,$39,$9F,$F1,$3C,$7C,$DE,$AC,$57,$D3,$AF
	.dc.b	$F4,$A7,$56,$F8,$FE,$0F,$BB,$FC,$E3,$B3,$F9,$34,$7C,$0C,$BF,$CE
	.dc.b	$90,$F6,$7F,$29,$CF,$FE,$D9,$7C,$B1,$99,$FE,$75,$E4,$7E,$EE,$7E
	.dc.b	$54,$7F,$A2,$67,$4F,$F9,$6C,$C9,$FE,$EF,$FD,$7E,$BF,$4E,$FF,$DC
	.dc.b	$D1,$8C,$3F,$83,$7E,$C1,$FD,$06,$CF,$EA,$BC,$8F,$F6,$13,$FA,$CC
	.dc.b	$3F,$80,$39,$8F,$F7,$87,$E0,$5F,$D2,$7E,$5D,$F6,$7F,$D3,$6F,$FD
	.dc.b	$12,$E4,$FE,$3C,$E0,$F6,$9F,$C2,$B3,$C3,$FD,$33,$52,$F3,$7F,$D2
	.dc.b	$16,$FC,$FB,$B3,$67,$99,$FE,$95,$F9,$8C,$3C,$CF,$F5,$23,$8B,$FE
	.dc.b	$AB,$1D,$84,$7A,$9F,$2D,$F8,$A3,$11,$CE,$3B,$63,$FD,$B7,$D3,$FF
	.dc.b	$5C,$56,$AF,$F9,$8B,$6F,$FC,$6B,$BB,$4F,$EE,$0F,$D0,$8F,$F0,$A6
	.dc.b	$BF,$E3,$7F,$43,$FD,$5D,$1F,$E9,$7D,$FE,$85,$6E,$3F,$D7,$78,$AB
	.dc.b	$F9,$64,$E6,$3F,$A0,$AE,$1F,$C0,$57,$FB,$16,$30,$FF,$A6,$4F,$F6
	.dc.b	$A3,$F8,$C3,$F3,$0F,$FA,$FD,$C3,$AE,$2F,$49,$F9,$A8,$FC,$9D,$7F
	.dc.b	$02,$F7,$7F,$B7,$47,$E8,$AB,$F2,$B9,$7F,$AC,$7A,$FF,$2B,$CB,$FE
	.dc.b	$1E,$AF,$F0,$67,$47,$EB,$7A,$3F,$E0,$04,$62,$3F,$8F,$C7,$F2,$7E
	.dc.b	$8E,$58,$FE,$EB,$AB,$F6,$D7,$FE,$11,$8F,$F2,$B7,$47,$F4,$E8,$FE
	.dc.b	$59,$1F,$8C,$F0,$BF,$85,$F8,$3F,$C4,$3B,$FF,$83,$D6,$AB,$FD,$3D
	.dc.b	$E0,$FF,$0C,$AD,$1F,$EC,$9F,$13,$F8,$CB,$C7,$C4,$C5,$7C,$B2,$70
	.dc.b	$E1,$FD,$B1,$F9,$CE,$0F,$81,$7F,$91,$F2,$3F,$9B,$9F,$CB,$FC,$9F
	.dc.b	$E6,$DE,$4F,$18,$CF,$93,$FC,$C3,$5F,$34,$62,$BF,$9C,$46,$23,$F9
	.dc.b	$DC,$7F,$40,$F2,$BC,$5C,$BF,$37,$96,$BF,$31,$FE,$37,$7F,$06,$3E
	.dc.b	$97,$8B,$E9,$F3,$3F,$B3,$F9,$FF,$D8,$3D,$1C,$7A,$1F,$D7,$BD,$0F
	.dc.b	$ED,$5E,$86,$3D,$1F,$EC,$F1,$FD,$B6,$7E,$17,$81,$F8,$5E,$CF,$A9
	.dc.b	$E4,$FF,$0F,$E7,$FF,$7A,$3F,$9E,$F2,$3F,$C2,$52,$F3,$7F,$C0,$C7
	.dc.b	$FB,$BE,$32,$FF,$82,$7D,$2F,$F7,$CC,$7F,$84,$30,$FF,$7F,$3F,$E1
	.dc.b	$2F,$53,$FE,$00,$7A,$BF,$E1,$BC,$98,$FF,$BE,$63,$18,$E3,$93,$FC
	.dc.b	$53,$8F,$F0,$5F,$57,$EC,$E1,$FE,$5F,$3F,$CF,$A7,$F1,$6F,$5B,$FD
	.dc.b	$AE,$FF,$8C,$8F,$F1,$C7,$77,$FA,$DF,$2F,$F3,$DC,$7F,$0F,$EF,$DB
	.dc.b	$CD,$FE,$D2,$63,$0F,$F2,$0C,$7E,$F6,$9E,$C7,$F9,$1E,$91,$89,$FF
	.dc.b	$29,$FB,$3F,$EA,$1F,$6B,$FC,$A5,$3F,$E5,$79,$FE,$45,$92,$F8,$D7
	.dc.b	$F9,$7B,$DC,$FC,$A3,$FE,$60,$F7,$3F,$D5,$5E,$EF,$FA,$17,$1F,$F3
	.dc.b	$56,$3F,$CA,$CF,$24,$7F,$5A,$8C,$E9,$F7,$75,$FB,$10,$F7,$B9,$7B
	.dc.b	$FF,$56,$3F,$D0,$58,$7C,$BB,$FF,$45,$1F,$E4,$78,$CB,$FC,$D4,$CF
	.dc.b	$A0,$F8,$2F,$DF,$57,$FA,$EF,$7F,$E8,$F1,$88,$FF,$4A,$47,$BE,$31
	.dc.b	$5F,$D3,$38,$BF,$02,$FF,$18,$AF,$F3,$23,$5F,$F5,$5F,$A8,$8C,$47
	.dc.b	$F5,$5F,$4B,$FD,$51,$1D,$23,$FC,$EC,$D7,$FA,$64,$7F,$BD,$1F,$F5
	.dc.b	$BB,$5F,$F1,$3B,$5F,$E7,$47,$63,$FC,$56,$53,$FB,$1F,$17,$FA,$D9
	.dc.b	$FF,$5F,$1F,$E8,$2F,$81,$F8,$0F,$E4,$CF,$F5,$C9,$A3,$FD,$6B,$3E
	.dc.b	$1C,$BF,$D1,$FA,$FC,$D9,$FF,$64,$E9,$FA,$4F,$81,$FA,$31,$FE,$CC
	.dc.b	$CB,$FD,$2B,$C3,$E7,$D3,$FD,$77,$F4,$1F,$B6,$E2,$FF,$6B,$FC,$6F
	.dc.b	$FA,$6B,$97,$B6,$3F,$EA,$23,$C5,$AE,$5F,$F4,$15,$7F,$B7,$8F,$F7
	.dc.b	$E6,$7F,$9B,$8F,$F6,$69,$FE,$02,$EF,$FF,$52,$F5,$FF,$A8,$63,$E1
	.dc.b	$CF,$FA,$9D,$FF,$4E,$3F,$AF,$AF,$F4,$6C,$7F,$7D,$BF,$F5,$33,$FE
	.dc.b	$EC,$AF,$F2,$BF,$1F,$FA,$92,$3F,$DD,$3C,$7F,$E5,$AD,$71,$1F,$B5
	.dc.b	$3F,$81,$7B,$EE,$DF,$81,$1F,$C5,$A3,$FD,$46,$FC,$73,$FC,$33,$C3
	.dc.b	$FA,$F8,$CC,$7F,$B4,$4F,$C9,$67,$FE,$9C,$7F,$C0,$27,$FA,$FC,$E1
	.dc.b	$E8,$7F,$A5,$FC,$2F,$F4,$5F,$8F,$FE,$8D,$E3,$FF,$47,$F9,$1F,$E9
	.dc.b	$1F,$1B,$FD,$1D,$5F,$E9,$7C,$BF,$AD,$4E,$7C,$6F,$E6,$5C,$BF,$E8
	.dc.b	$9F,$28,$42,$57,$F4,$17,$07,$FA,$59,$FE,$99,$E7,$7F,$56,$3F,$AB
	.dc.b	$B8,$71,$F4,$3F,$E8,$87,$5E,$85,$7F,$AE,$FD,$17,$FA,$9B,$C2,$FC
	.dc.b	$07,$17,$F3,$23,$FD,$15,$C5,$FA,$03,$E0,$9F,$DD,$FD,$57,$CA,$7F
	.dc.b	$74,$3F,$BA,$E7,$FD,$DF,$D8,$4E,$9B,$25,$F4,$7F,$E8,$78,$E5,$0F
	.dc.b	$A5,$C3,$C0,$7D,$2F,$FA,$03,$83,$E1,$9F,$1B,$C6,$FF,$5A,$B8,$3F
	.dc.b	$D1,$59,$23,$FC,$99,$AF,$BB,$5F,$F7,$44,$7F,$9D,$35,$FF,$7B,$3F
	.dc.b	$A9,$8F,$CA,$57,$EF,$E7,$F1,$1E,$F7,$29,$E5,$3F,$E3,$29,$FE,$0B
	.dc.b	$DD,$FE,$7C,$7F,$C1,$93,$AF,$7F,$B4,$7F,$A4,$3E,$03,$F8,$75,$2F
	.dc.b	$FC,$8D,$3F,$E4,$E3,$FA,$25,$FF,$9F,$9F,$C9,$9F,$C5,$27,$FC,$FE
	.dc.b	$FF,$0B,$3F,$CE,$7F,$07,$FC,$65,$FF,$3C,$F7,$FD,$97,$27,$F1,$1F
	.dc.b	$1B,$FC,$F7,$F0,$9F,$E6,$38,$FF,$63,$FA,$BF,$DC,$A3,$11,$FE,$73
	.dc.b	$9F,$F3,$FF,$C2,$F5,$9F,$F3,$C4,$FE,$17,$1F,$F3,$BC,$FB,$3E,$17
	.dc.b	$33,$F8,$97,$9B,$FC,$24,$FF,$3E,$E5,$FC,$C7,$0F,$F3,$CC,$7F,$25
	.dc.b	$C7,$FC,$F3,$3C,$CF,$67,$AD,$FE,$77,$9F,$81,$8F,$F9,$CD,$F8,$BE
	.dc.b	$3F,$F9,$AF,$7F,$E5,$53,$FD,$F7,$6F,$F0,$54,$7D,$0B,$FF,$37,$DF
	.dc.b	$F8,$28,$FF,$01,$38,$E3,$FC,$35,$23,$EF,$4F,$F8,$11,$FF,$5F,$CE
	.dc.b	$9F,$C3,$5F,$F9,$8A,$78,$F9,$36,$7E,$94,$E6,$B8,$72,$7E,$63,$97
	.dc.b	$F3,$7F,$07,$C5,$FC,$D1,$FC,$12,$FF,$59,$D9,$FA,$B3,$F5,$9D,$5F
	.dc.b	$E5,$BE,$AF,$D3,$47,$EB,$FA,$B0,$EF,$AF,$FC,$D3,$7C,$9F,$D7,$F6
	.dc.b	$7F,$98,$E3,$CA,$7F,$9B,$F4,$F6,$79,$1F,$E6,$3E,$FF,$E1,$C0,$77
	.dc.b	$FF,$98,$DC,$8F,$03,$0E,$2F,$F8,$A3,$FE,$6C,$F0,$7F,$CB,$FB,$FC
	.dc.b	$09,$FE,$3C,$FE,$E5,$FE,$4F,$E2,$FF,$2E,$F1,$4B,$C7,$FF,$2E,$78
	.dc.b	$DF,$CA,$AB,$7E,$3F,$F3,$03,$FC,$B3,$D3,$F8,$43,$FD,$51,$DD,$FE
	.dc.b	$56,$7F,$9F,$F5,$FF,$95,$1F,$E8,$C7,$F8,$37,$AB,$FC,$AA,$FF,$64
	.dc.b	$8F,$F6,$A3,$FC,$96,$FF,$DA,$EF,$ED,$CF,$F6,$03,$89,$FE,$B7,$C5
	.dc.b	$FD,$86,$BF,$CD,$15,$FD,$86,$31,$D7,$A9,$FF,$1E,$3F,$DA,$E7,$FC
	.dc.b	$89,$3F,$DC,$7C,$3F,$E6,$5E,$06,$3C,$7F,$E2,$D8,$7F,$9A,$3E,$2B
	.dc.b	$18,$7F,$7C,$3F,$C9,$67,$F7,$EC,$BF,$CA,$BB,$F0,$A7,$9D,$DB,$CF
	.dc.b	$FF,$05,$69,$FE,$0E,$C3,$FC,$85,$87,$F9,$73,$D1,$FF,$0A,$47,$F9
	.dc.b	$BF,$D1,$FF,$3B,$FA,$7F,$E1,$E8,$7A,$3F,$E4,$0F,$17,$FC,$39,$87
	.dc.b	$F9,$15,$FF,$17,$7A,$3F,$E3,$EF,$3F,$F9,$0F,$1F,$F2,$DF,$27,$FC
	.dc.b	$79,$1F,$C6,$63,$A6,$3F,$E6,$56,$CF,$F9,$06,$FF,$8F,$4F,$FB,$17
	.dc.b	$83,$FC,$F6,$FF,$8D,$0C,$3F,$D6,$DF,$F1,$CF,$83,$F3,$69,$1F,$D1
	.dc.b	$4F,$F4,$31,$FE,$46,$2E,$5A,$BE,$1E,$9F,$4C,$3F,$D2,$F5,$CB,$A7
	.dc.b	$FD,$33,$39,$9E,$B2,$9F,$F5,$47,$A7,$F3,$78,$3F,$CB,$52,$7F,$CB
	.dc.b	$91,$9D,$3C,$B1,$98,$C4,$EF,$D9,$43,$F3,$3F,$C4,$69,$8F,$5B,$FC
	.dc.b	$85,$D9,$FE,$2D,$AF,$F4,$06,$1F,$9B,$8F,$F1,$F4,$62,$3F,$C8,$07
	.dc.b	$A3,$7E,$4F,$C2,$8F,$D8,$3C,$75,$FF,$5B,$F4,$7F,$A6,$67,$33,$C8
	.dc.b	$F8,$13,$FE,$4F,$7F,$C7,$CF,$E5,$F1,$C3,$F6,$3E,$AF,$FC,$BD,$8F
	.dc.b	$A2,$BC,$57,$FA,$9F,$73,$F6,$3E,$C7,$F9,$23,$0F,$F0,$B4,$7E,$D3
	.dc.b	$E4,$FF,$C8,$BC,$3F,$B1,$F5,$FF,$6B,$8F,$FB,$17,$6F,$ED,$D8,$FF
	.dc.b	$B5,$B1,$FF,$3C,$EB,$F8,$2B,$FF,$25,$5F,$F9,$8B,$4F,$BB,$3F,$86
	.dc.b	$AF,$EE,$DC,$7F,$A2,$E3,$FB,$27,$FA,$4F,$0F,$F2,$5E,$7F,$80,$8C
	.dc.b	$F1,$F0,$CB,$FC,$DB,$B7,$E2,$3A,$7F,$D1,$B8,$7F,$94,$B2,$FF,$70
	.dc.b	$CE,$73,$FF,$29,$5F,$FB,$76,$7F,$8A,$1E,$7F,$07,$F6,$6E,$27,$FC
	.dc.b	$15,$EA,$7E,$02,$3F,$C7,$5B,$7F,$95,$3C,$67,$8A,$FC,$AB,$FC,$A5
	.dc.b	$ED,$66,$31,$3A,$F2,$1F,$2F,$FE,$55,$FC,$57,$FD,$09,$C1,$FC,$DB
	.dc.b	$CC,$FF,$72,$BF,$E5,$2F,$37,$3E,$17,$F4,$7F,$3B,$FA,$37,$1F,$FB
	.dc.b	$B6,$BF,$A6,$70,$7F,$93,$7B,$7F,$C9,$CF,$F9,$44,$FF,$28,$46,$EF
	.dc.b	$F5,$3E,$AF,$F9,$23,$F6,$07,$C5,$BE,$2D,$F2,$0F,$BE,$FF,$25,$56
	.dc.b	$A7,$35,$83,$85,$3F,$00,$FF,$18,$63,$F1,$F3,$0E,$8F,$F2,$77,$0B
	.dc.b	$8B,$E0,$F9,$5F,$DD,$E3,$E0,$BF,$E4,$3A,$FF,$5B,$1E,$A7,$0F,$FA
	.dc.b	$6F,$AF,$FC,$7D,$1D,$E3,$B7,$07,$F1,$1E,$2F,$D2,$C7,$17,$F8,$8C
	.dc.b	$7F,$83,$A3,$E4,$1F,$1B,$CB,$EB,$1F,$37,$E0,$BB,$F2,$7C,$C8,$FF
	.dc.b	$0D,$57,$CC,$BF,$F6,$BC,$3F,$22,$FE,$0E,$FF,$94,$23,$FD,$E3,$97
	.dc.b	$F1,$C8,$C5,$7F,$63,$D7,$FC,$A5,$5F,$93,$F1,$3F,$CA,$B7,$FE,$59
	.dc.b	$F9,$6F,$E2,$3F,$08,$FA,$CF,$EB,$56,$8C,$3F,$CB,$1A,$FF,$07,$F3
	.dc.b	$3F,$CA,$13,$FE,$CB,$7E,$94,$EB,$2F,$E7,$BF,$6F,$FE,$58,$FD,$2F
	.dc.b	$AC,$9F,$E4,$3F,$0D,$71,$FE,$72,$D7,$F8,$A1,$FD,$01,$E9,$7F,$D1
	.dc.b	$61,$79,$C7,$C5,$2A,$DB,$FD,$02,$98,$FE,$1E,$73,$3F,$E0,$6C,$F8
	.dc.b	$5F,$C2,$F7,$1A,$7F,$36,$7E,$6E,$BF,$E7,$EC,$BF,$CB,$47,$CD,$BF
	.dc.b	$C9,$98,$8F,$E9,$78,$7D,$33,$FC,$A9,$C1,$CF,$8B,$F0,$77,$FD,$72
	.dc.b	$FF,$CB,$11,$F8,$4A,$FF,$0C,$71,$EA,$95,$FE,$5D,$8D,$4F,$E8,$B1
	.dc.b	$FC,$A3,$FD,$B2,$B7,$18,$8F,$D3,$F0,$FF,$BE,$63,$11,$FA,$93,$F7
	.dc.b	$AF,$E7,$7D,$8F,$CA,$57,$96,$31,$3F,$9C,$3C,$1C,$BF,$AD,$9F,$F3
	.dc.b	$9B,$FE,$65,$E7,$FE,$0F,$3F,$E7,$3E,$8F,$D5,$63,$FE,$5E,$F8,$17
	.dc.b	$77,$F0,$1E,$CF,$DF,$67,$FE,$5F,$8C,$76,$ED,$D7,$7F,$3D,$7F,$D6
	.dc.b	$F9,$FF,$02,$EB,$FF,$41,$E9,$F1,$23,$FC,$B7,$E0,$FF,$A0,$63,$F8
	.dc.b	$F7,$85,$E4,$CF,$F9,$26,$1C,$0F,$F0,$6F,$7F,$FB,$EF,$83,$57,$FE
	.dc.b	$04,$F7,$BF,$CB,$7E,$3F,$F3,$01,$E3,$7F,$30,$C7,$FC,$B7,$E4,$FF
	.dc.b	$36,$5E,$57,$C0,$5E,$5F,$F4,$2E,$DF,$C5,$46,$35,$FF,$0E,$C7,$E2
	.dc.b	$0F,$DB,$C7,$F4,$FE,$07,$99,$FD,$4F,$D2,$FF,$11,$CE,$63,$FA,$94
	.dc.b	$FF,$99,$6F,$FC,$81,$7F,$E1,$43,$07,$F5,$F3,$FC,$2F,$E7,$FF,$9A
	.dc.b	$23,$FA,$8C,$FF,$6E,$70,$FE,$20,$FF,$1B,$1F,$EA,$D7,$FC,$9A,$7F
	.dc.b	$89,$0F,$C7,$98,$9F,$F0,$A7,$67,$F8,$0E,$3F,$C5,$BE,$92,$31,$1F
	.dc.b	$E3,$CF,$4F,$1D,$BF,$E0,$B8,$FF,$1F,$E9,$87,$FC,$44,$7F,$8D,$BC
	.dc.b	$0F,$F3,$3B,$FC,$4F,$2A,$7A,$87,$BB,$FE,$74,$F7,$7F,$67,$1F,$C2
	.dc.b	$7D,$5F,$E9,$39,$FF,$88,$A3,$87,$AD,$FE,$7A,$F5,$FF,$CA,$13,$FE
	.dc.b	$2F,$9F,$D6,$65,$E9,$9F,$CD,$CE,$67,$E3,$1F,$D5,$0F,$77,$B1,$FE
	.dc.b	$3E,$8F,$F2,$1F,$B0,$E1,$FF,$3B,$6B,$F8,$03,$EB,$1F,$E4,$79,$DC
	.dc.b	$BD,$BF,$F2,$B4,$FF,$96,$70,$C7,$67,$B1,$FE,$3D,$AF,$F6,$3C,$7F
	.dc.b	$7B,$1F,$EA,$2E,$0F,$F3,$95,$FF,$99,$A7,$2F,$FA,$99,$EA,$FD,$06
	.dc.b	$BF,$E7,$BB,$FE,$8A,$9F,$F4,$6E,$BF,$CA,$63,$07,$AC,$FF,$57,$1F
	.dc.b	$E7,$88,$DB,$FE,$57,$3F,$B5,$4F,$DB,$7F,$D6,$E6,$0F,$F4,$21,$FE
	.dc.b	$85,$F1,$F8,$BF,$84,$F6,$BB,$BF,$91,$BF,$F3,$CF,$B5,$FE,$86,$9C
	.dc.b	$DF,$5A,$7B,$7F,$E8,$0C,$79,$DF,$F9,$8F,$EF,$7F,$D4,$FD,$3F,$E8
	.dc.b	$23,$FD,$1F,$EB,$23,$38,$7E,$DB,$6F,$F3,$DC,$62,$15,$FE,$6D,$E3
	.dc.b	$F3,$C7,$F9,$BF,$93,$F4,$F7,$FE,$2A,$C3,$FC,$F5,$A7,$FB,$5A,$3F
	.dc.b	$C8,$31,$8A,$F8,$27,$13,$E6,$6D,$F6,$E3,$FC,$F3,$5C,$69,$F1,$5E
	.dc.b	$8E,$7E,$F0,$F8,$AF,$F1,$3D,$F0,$C7,$E6,$ED,$B7,$A9,$FE,$AE,$7F
	.dc.b	$D0,$DF,$03,$CB,$E2,$F8,$BF,$4B,$C1,$FF,$45,$76,$FF,$02,$EA,$F6
	.dc.b	$3F,$84,$DB,$FD,$06,$DC,$7F,$AF,$65,$FD,$CE,$7F,$05,$1D,$E3,$0F
	.dc.b	$F0,$03,$FD,$67,$DC,$F2,$CF,$E7,$FA,$BF,$D6,$F8,$7F,$A0,$E3,$15
	.dc.b	$FE,$C6,$F0,$7F,$A2,$78,$3F,$57,$9F,$FD,$10,$DE,$5F,$F4,$6D,$7F
	.dc.b	$A2,$71,$FF,$66,$F5,$7C,$A8,$C9,$FE,$74,$8C,$F1,$FF,$4F,$ED,$FF
	.dc.b	$63,$EF,$FC,$9E,$7F,$D0,$B1,$8A,$FE,$EF,$87,$F1,$A9,$FE,$F3,$B7
	.dc.b	$F4,$49,$FD,$31,$FE,$76,$F0,$FE,$BF,$37,$F9,$39,$FC,$4D,$FF,$A1
	.dc.b	$A7,$FD,$C3,$E3,$FF,$52,$FC,$6D,$F8,$83,$F5,$77,$EB,$2F,$D6,$78
	.dc.b	$60,$C3,$85,$EE,$9F,$B3,$EA,$73,$87,$2F,$CB,$F8,$FA,$02,$7C,$83
	.dc.b	$F0,$68,$5E,$1F,$66,$9B,$AD,$CE,$BE,$B7,$9C,$62,$39,$7E,$15,$E6
	.dc.b	$F5,$84,$F8,$EF,$89,$87,$FC,$00,$8C,$8F,$8F,$FF,$4D,$F2,$FF,$BE
	.dc.b	$3D,$3F,$E8,$9E,$1F,$F9,$AE,$31,$7F,$E2,$DE,$BF,$F4,$E4,$7F,$A5
	.dc.b	$BE,$45,$EF,$3F,$3A,$FF,$48,$FC,$5F,$FA,$5B,$8B,$D7,$F1,$8B,$DD
	.dc.b	$6B,$F0,$0F,$9B,$39,$C3,$97,$C1,$7F,$A6,$FF,$02,$CE,$1C,$23,$FD
	.dc.b	$3D,$F8,$17,$FA,$6B,$6F,$C0,$F8,$5F,$EC,$0D,$3F,$8D,$F3,$FC,$9E
	.dc.b	$9F,$9B,$D1,$F3,$7D,$66,$1F,$E1,$6A,$F9,$52,$EA,$E9,$3F,$46,$57
	.dc.b	$CB,$AB,$3C,$FF,$1B,$3F,$F5,$CE,$5F,$1B,$0F,$F2,$6F,$37,$FB,$07
	.dc.b	$D8,$FC,$9F,$9B,$FE,$A2,$9F,$F0,$79,$FE,$57,$E9,$FF,$4C,$4F,$1F
	.dc.b	$17,$F9,$97,$03,$E1,$71,$5F,$E9,$D7,$58,$FF,$B3,$7E,$1F,$F3,$B5
	.dc.b	$FE,$9D,$C7,$FB,$16,$1F,$E5,$E5,$E4,$7F,$BC,$2B,$F5,$93,$98,$FF
	.dc.b	$31,$EB,$FC,$16,$73,$3F,$AF,$3F,$D3,$D5,$FC,$70,$FF,$37,$7C,$77
	.dc.b	$FA,$6B,$1F,$E8,$91,$F3,$23,$0F,$F3,$09,$FF,$51,$7C,$57,$F1,$F8
	.dc.b	$A7,$3F,$2F,$FD,$42,$F5,$F4,$FF,$CE,$B7,$FE,$99,$C3,$AC,$FF,$AA
	.dc.b	$FE,$3B,$F7,$7F,$19,$98,$EC,$D7,$FB,$3B,$E2,$7E,$11,$FE,$45,$71
	.dc.b	$97,$E0,$A3,$FC,$96,$FF,$A6,$1F,$26,$95,$FE,$C8,$7F,$89,$E7,$8B
	.dc.b	$C4,$72,$7F,$A7,$9F,$8C,$7F,$A2,$35,$F9,$6F,$D2,$DB,$FC,$9F,$18
	.dc.b	$7E,$AE,$1F,$E9,$C9,$FB,$2F,$F8,$E0,$FF,$4A,$4F,$F1,$AA,$FC,$24
	.dc.b	$FE,$28,$FC,$79,$FE,$8D,$E0,$FC,$3E,$3F,$E2,$99,$FF,$77,$1F,$E3
	.dc.b	$FC,$FF,$D1,$B5,$C7,$07,$0F,$FA,$63,$8B,$F2,$F7,$FE,$04,$38,$57
	.dc.b	$FA,$67,$93,$AE,$3F,$EA,$77,$E5,$73,$FE,$BF,$A7,$F8,$B5,$FE,$CA
	.dc.b	$3F,$5B,$CF,$FE,$8F,$C3,$F5,$9D,$1F,$B4,$BC,$6B,$FD,$C7,$27,$47
	.dc.b	$CC,$F0,$FF,$87,$77,$FF,$04,$F8,$A7,$88,$F0,$75,$A7,$EF,$3C,$2F
	.dc.b	$E1,$5E,$27,$F1,$2F,$07,$32,$9C,$F8,$5F,$EA,$7E,$3F,$E2,$BE,$1F
	.dc.b	$F0,$DC,$FA,$F8,$9C,$27,$E0,$E1,$FE,$C2,$DB,$E0,$F0,$74,$F1,$BF
	.dc.b	$95,$F8,$DF,$EE,$1B,$FE,$73,$1F,$CB,$BC,$7F,$F5,$35,$FC,$6E,$8F
	.dc.b	$E1,$FE,$3F,$91,$F9,$1C,$FF,$CE,$B0,$FE,$83,$18,$8F,$E8,$D1,$FD
	.dc.b	$13,$CA,$FE,$9B,$E6,$7F,$55,$8C,$7C,$8B,$E1,$DE,$8F,$F5,$CF,$3B
	.dc.b	$FA,$95,$6A,$73,$E8,$7F,$59,$F4,$3E,$2F,$7F,$EE,$BC,$5B,$CB,$FE
	.dc.b	$63,$E5,$AF,$F2,$87,$FC,$E9,$E0,$7F,$A8,$A3,$FB,$A9,$F9,$D3,$13
	.dc.b	$FC,$37,$C8,$FF,$74,$FA,$9F,$E8,$8A,$FE,$27,$5F,$E4,$AF,$77,$FD
	.dc.b	$63,$39,$AF,$F0,$2E,$1F,$EA,$BC,$7F,$C9,$57,$FC,$E3,$D6,$DB,$FE
	.dc.b	$65,$7F,$C9,$D9,$7E,$CB,$D8,$FF,$0C,$CF,$E2,$8C,$1E,$9B,$1E,$27
	.dc.b	$FD,$86,$FF,$AA,$CF,$F0,$83,$91,$EE,$65,$FF,$3A,$63,$FC,$06,$BF
	.dc.b	$D5,$91,$FE,$69,$C7,$FD,$6F,$A7,$F8,$DE,$18,$EA,$7F,$84,$79,$7F
	.dc.b	$E9,$DC,$BF,$D6,$31,$FE,$6D,$F4,$77,$F0,$6F,$73,$FC,$91,$5F,$EB
	.dc.b	$FF,$7B,$FC,$DF,$95,$3F,$9A,$6D,$FE,$B5,$7F,$B1,$CF,$F8,$C6,$FF
	.dc.b	$CA,$D1,$F0,$7D,$AF,$F5,$BF,$BF,$FE,$24,$9F,$F3,$34,$FF,$26,$3F
	.dc.b	$A3,$B8,$7C,$D5,$FE,$B7,$7A,$BF,$B6,$87,$C0,$73,$F8,$1F,$F2,$35
	.dc.b	$7F,$B5,$5F,$F2,$34,$FF,$A0,$63,$EC,$1F,$E8,$1F,$83,$FF,$46,$BF
	.dc.b	$E6,$A3,$FD,$66,$7B,$38,$2B,$FC,$C8,$FF,$48,$E1,$FE,$B2,$70,$7F
	.dc.b	$70,$F5,$3F,$D0,$FC,$99,$E1,$FE,$92,$76,$9F,$75,$FF,$B2,$7D,$4D
	.dc.b	$CE,$63,$FD,$53,$3B,$3F,$D6,$93,$FE,$F6,$7F,$D5,$AD,$7F,$D5,$4D
	.dc.b	$DF,$FD,$50,$E7,$DF,$FF,$53,$4F,$F9,$58,$FF,$42,$57,$FA,$81,$FF
	.dc.b	$5F,$5F,$F5,$3E,$4F,$F4,$6E,$DF,$92,$8F,$F3,$2D,$3E,$07,$E7,$DF
	.dc.b	$E6,$F5,$FA,$BC,$D9,$E2,$F2,$4F,$F8,$02,$BF,$82,$3B,$AF,$F7,$67
	.dc.b	$C2,$7E,$52,$7F,$0B,$3F,$EC,$03,$B7,$2F,$AF,$1B,$1C,$6F,$F5,$3F
	.dc.b	$0B,$FE,$9C,$8F,$61,$FD,$AA,$DF,$07,$FC,$26,$3F,$B7,$5F,$A7,$3F
	.dc.b	$D6,$C7,$FB,$3E,$BF,$6C,$FE,$8E,$3F,$B6,$70,$EE,$3D,$D1,$FD,$BF
	.dc.b	$83,$E1,$65,$FE,$C4,$7F,$D3,$9D,$5F,$EF,$9A,$FE,$F9,$3F,$E9,$87
	.dc.b	$E6,$9F,$E4,$0D,$FF,$D1,$3D,$9F,$C4,$E3,$FD,$2F,$3F,$E9,$37,$FC
	.dc.b	$52,$7F,$AA,$73,$FF,$54,$7C,$07,$F2,$39,$FF,$27,$35,$FE,$16,$7F
	.dc.b	$A1,$AF,$F1,$35,$FE,$96,$F2,$FF,$C1,$DE,$43,$C7,$FE,$6D,$F0,$EF
	.dc.b	$92,$FF,$46,$43,$93,$FD,$14,$EA,$33,$E3,$60,$57,$1B,$E9,$E0,$FF
	.dc.b	$A1,$3C,$CF,$F2,$C1,$FE,$8F,$F8,$3F,$E9,$6E,$0F,$F2,$DB,$FD,$5B
	.dc.b	$D0,$FF,$41,$91,$C3,$81,$F1,$5F,$E8,$D7,$03,$D1,$C3,$BF,$4B,$FD
	.dc.b	$4F,$E2,$FF,$08,$E3,$F8,$B8,$7F,$A7,$3C,$DF,$8B,$CD,$FE,$7C,$F7
	.dc.b	$3F,$BA,$7A,$DF,$E8,$AF,$63,$FD,$0D,$7F,$DD,$27,$27,$F7,$9F,$95
	.dc.b	$E1,$3F,$E8,$71,$F2,$F8,$BE,$5F,$2F,$FE,$5B,$EA,$FC,$A9,$FE,$C8
	.dc.b	$F8,$3F,$F3,$0B,$FE,$82,$CF,$CB,$18,$8F,$EA,$B1,$FE,$60,$5D,$9E
	.dc.b	$8F,$86,$FF,$26,$47,$F9,$04,$FF,$2D,$9A,$3F,$84,$CE,$27,$F8,$5F
	.dc.b	$A3,$FE,$65,$EF,$FE,$11,$F0,$FF,$E6,$0B,$E0,$7F,$92,$E7,$FC,$C3
	.dc.b	$F0,$FF,$BA,$8F,$F3,$01,$EA,$C1,$F0,$FF,$E5,$F8,$FF,$29,$46,$27
	.dc.b	$84,$F9,$A7,$FD,$3F,$5F,$EA,$08,$C4,$7F,$98,$1B,$7F,$D9,$1C,$C7
	.dc.b	$F5,$86,$BF,$E5,$EF,$91,$FE,$1B,$87,$F9,$5E,$3E,$03,$FE,$6B,$F4
	.dc.b	$7E,$67,$2F,$FA,$4E,$72,$7C,$FE,$6F,$F2,$DB,$EB,$8F,$BD,$C7,$F6
	.dc.b	$F8,$FF,$C6,$EF,$DD,$E3,$FB,$73,$A7,$F0,$C6,$8B,$CB,$F8,$58,$FC
	.dc.b	$75,$FD,$27,$FD,$D7,$CF,$F9,$3C,$7F,$C0,$31,$F9,$EE,$9F,$D0,$F4
	.dc.b	$FF,$8E,$DF,$CE,$F5,$57,$F3,$F3,$B7,$FD,$D3,$D7,$FE,$39,$EC,$FD
	.dc.b	$67,$C3,$57,$F5,$D5,$FE,$2F,$F0,$3F,$6D,$E1,$FE,$F6,$31,$E1,$FF
	.dc.b	$00,$F0,$7F,$CA,$BE,$17,$F0,$5F,$0F,$F7,$B3,$FC,$02,$9E,$2F,$D0
	.dc.b	$5D,$DF,$C4,$78,$7F,$CC,$F1,$8E,$3F,$E6,$D1,$FE,$81,$EC,$FC,$E7
	.dc.b	$4F,$D4,$F1,$BF,$90,$57,$F6,$AA,$FF,$17,$3F,$D4,$BA,$FF,$C4,$EF
	.dc.b	$F8,$54,$FE,$57,$1F,$E2,$17,$FB,$7F,$91,$FE,$20,$8F,$F2,$1C,$7F
	.dc.b	$AA,$FC,$AC,$47,$F8,$BB,$CA,$FE,$71,$D5,$FC,$FB,$C3,$FF,$54,$1F
	.dc.b	$D1,$1F,$E8,$F7,$FD,$54,$C7,$1F,$F6,$DF,$0B,$FC,$39,$1F,$67,$CD
	.dc.b	$B7,$FD,$DB,$2F,$F7,$DE,$1F,$D6,$AB,$FB,$6B,$FD,$81,$FC,$07,$9D
	.dc.b	$FE,$2C,$F3,$FF,$B2,$F0,$2B,$FB,$5D,$7E,$17,$5F,$F0,$E7,$A1,$FD
	.dc.b	$C6,$3F,$A4,$70,$7F,$86,$DF,$E9,$67,$FA,$A2,$7F,$C2,$B9,$F8,$E2
	.dc.b	$9F,$C1,$73,$FF,$20,$FA,$9F,$E0,$B8,$FE,$9F,$35,$FE,$24,$D7,$FD
	.dc.b	$EE,$D4,$7E,$08,$FE,$C6,$7F,$6C,$30,$7F,$69,$3F,$8A,$73,$7D,$8B
	.dc.b	$FE,$DB,$1F,$E7,$0F,$4F,$1C,$5F,$E1,$E6,$C7,$36,$BE,$0D,$5F,$19
	.dc.b	$DA,$FE,$FC,$B9,$BF,$CD,$73,$99,$FC,$04,$FA,$27,$FD,$77,$E9,$7F
	.dc.b	$90,$63,$F8,$FF,$A9,$FE,$1F,$7F,$C4,$46,$0F,$7F,$E6,$5F,$E2,$27
	.dc.b	$FB,$C1,$F0,$A3,$FC,$4D,$C1,$FE,$23,$F8,$6F,$3C,$FF,$86,$FD,$8F
	.dc.b	$81,$EB,$FF,$82,$F1,$FB,$B9,$FF,$76,$8F,$F2,$DC,$62,$3C,$F5,$EF
	.dc.b	$8C,$47,$C8,$C3,$FC,$61,$E0,$74,$8F,$F4,$4F,$47,$F8,$A6,$3E,$3E
	.dc.b	$BF,$E8,$CB,$FE,$F5,$87,$F8,$D7,$1F,$F0,$2F,$4F,$F8,$BF,$D6,$F9
	.dc.b	$BF,$27,$B9,$F3,$D7,$DE,$1D,$DF,$E3,$F8,$CC,$79,$5F,$B3,$A7,$FA
	.dc.b	$27,$1F,$D4,$CF,$F8,$A6,$3F,$D6,$B1,$FE,$2C,$7F,$C2,$BE,$E7,$F4
	.dc.b	$D8,$C4,$7F,$A8,$A3,$C1,$3F,$6F,$D8,$7B,$3F,$EB,$7E,$8F,$EC,$DC
	.dc.b	$FF,$92,$9C,$D7,$FA,$AE,$7F,$D5,$DC,$5F,$E9,$EC,$54,$F0,$BF,$D5
	.dc.b	$7C,$DF,$7E,$31,$87,$DC,$11,$F7,$23,$31,$F5,$F0,$FD,$0F,$6F,$FA
	.dc.b	$CA,$3A,$DF,$F1,$4D,$BF,$D7,$D3,$FE,$84,$BF,$CA,$57,$5C,$7F,$BF
	.dc.b	$61,$E9,$E3,$FF,$5E,$C6,$33,$FF,$1E,$D7,$D7,$8F,$F1,$EB,$5F,$F0
	.dc.b	$CF,$89,$F8,$E9,$FF,$19,$7C,$7B,$F5,$AF,$F7,$77,$C1,$E1,$FF,$78
	.dc.b	$7E,$FF,$FE,$33,$F9,$27,$E2,$F0,$BB,$7F,$8D,$3C,$F4,$E6,$5F,$84
	.dc.b	$71,$78,$7E,$39,$FE,$2E,$F4,$DF,$4D,$07,$D3,$65,$C1,$4D,$FD,$3D
	.dc.b	$5F,$5B,$FE,$36,$E3,$FE,$6D,$D5,$FE,$D3,$DB,$FB,$3C,$7F,$39,$F2
	.dc.b	$FF,$CF,$31,$FE,$33,$FD,$8B,$FB,$7B,$7D,$CF,$F1,$CF,$DC,$7F,$8E
	.dc.b	$E7,$FB,$AF,$E0,$36,$7F,$C7,$BF,$75,$F8,$4F,$43,$F5,$59,$7F,$45
	.dc.b	$C5,$E8,$7E,$87,$98,$4E,$6F,$89,$CB,$F8,$C9,$D4,$E6,$3E,$67,$36
	.dc.b	$7B,$7F,$93,$CF,$23,$E6,$CF,$CA,$C3,$BC,$FF,$48,$C7,$E0,$9F,$E5
	.dc.b	$6C,$3F,$65,$E9,$FF,$9F,$3C,$0F,$F1,$AF,$77,$5B,$FF,$7C,$78,$5F
	.dc.b	$E9,$7C,$5F,$15,$C3,$2E,$60,$FD,$1B,$95,$87,$49,$7C,$D6,$63,$FA
	.dc.b	$64,$F1,$9C,$E1,$E7,$8E,$3C,$DF,$53,$F2,$EF,$CA,$57,$F8,$BE,$3F
	.dc.b	$D4,$34,$CF,$BB,$FE,$63,$3F,$C9,$F7,$FE,$61,$CF,$B7,$C5,$7E,$92
	.dc.b	$78,$65,$C7,$17,$E2,$BF,$A1,$F8,$B7,$DD,$7F,$30,$B4,$61,$FE,$AA
	.dc.b	$7F,$9B,$5C,$3F,$D5,$47,$A1,$F1,$63,$39,$FF,$4C,$8F,$8B,$18,$8F
	.dc.b	$E6,$BF,$49,$FD,$8A,$1F,$BD,$F8,$5B,$7E,$03,$FC,$B1,$1F,$B7,$F8
	.dc.b	$DC,$C7,$84,$FF,$31,$63,$FD,$7A,$EC,$7A,$31,$FF,$31,$4F,$F3,$6D
	.dc.b	$3F,$CD,$EF,$33,$E0,$7C,$0F,$C4,$AF,$07,$07,$F9,$8F,$87,$E8,$BF
	.dc.b	$42,$7E,$A4,$FF,$7A,$D7,$F8,$84,$62,$3F,$0C,$7D,$7E,$1F,$F2,$AF
	.dc.b	$17,$F0,$F9,$FC,$0C,$FF,$85,$EB,$F6,$8F,$F9,$BB,$8F,$F1,$1E,$0F
	.dc.b	$FB,$4A,$55,$F9,$4A,$FC,$46,$9F,$9A,$8C,$47,$39,$FC,$9D,$71,$9F
	.dc.b	$D2,$CF,$E7,$A7,$F4,$51,$8E,$6F,$D2,$69,$FE,$FC,$8C,$4F,$E8,$0F
	.dc.b	$0F,$3F,$ED,$BA,$7F,$63,$D3,$FC,$8B,$C7,$CE,$5F,$B6,$9F,$DD,$4F
	.dc.b	$F2,$5C,$3C,$31,$FB,$9E,$CF,$E4,$9D,$9E,$CE,$DF,$E1,$39,$FF,$98
	.dc.b	$1F,$ED,$1F,$16,$F8,$B7,$82,$F0,$FF,$89,$E7,$FC,$5E,$B5,$E1,$7F
	.dc.b	$A2,$6B,$47,$FB,$3A,$75,$6F,$13,$C1,$BF,$CA,$E2,$F8,$19,$FD,$0D
	.dc.b	$3F,$E0,$0E,$1F,$EE,$8F,$27,$F9,$8F,$91,$FE,$89,$7F,$99,$F8,$F7
	.dc.b	$C8,$FE,$6B,$5B,$7F,$9D,$79,$3F,$E8,$9E,$3F,$E8,$5C,$D7,$9B,$FA
	.dc.b	$25,$FF,$47,$9F,$86,$FF,$90,$3E,$17,$FD,$11,$E7,$7F,$53,$F9,$47
	.dc.b	$CA,$BC,$EF,$F4,$6F,$C4,$3D,$D5,$EE,$9E,$97,$E2,$3D,$3F,$C8,$1F
	.dc.b	$3B,$C3,$C7,$47,$F6,$3E,$6F,$F5,$4F,$A9,$FC,$1F,$C8,$FC,$F3,$6B
	.dc.b	$F8,$87,$07,$F7,$AA,$FE,$45,$E4,$7F,$76,$F5,$7F,$CB,$73,$99,$4F
	.dc.b	$F7,$AF,$6F,$FD,$6B,$EB,$FF,$82,$33,$FF,$20,$1F,$E9,$63,$FC,$0E
	.dc.b	$7F,$90,$F3,$FF,$4B,$DF,$FA,$3F,$2F,$F0,$BC,$E2,$3F,$97,$C7,$FB
	.dc.b	$FE,$7F,$9A,$79,$1F,$E2,$29,$7B,$5F,$EF,$18,$C3,$FE,$24,$D7,$FD
	.dc.b	$0D,$9F,$F8,$C2,$BF,$9E,$18,$C7,$F7,$8F,$FA,$7C,$FE,$91,$7F,$E3
	.dc.b	$47,$FC,$A1,$1F,$E3,$DC,$F8,$3F,$D2,$8F,$F7,$44,$FF,$92,$1F,$E4
	.dc.b	$DE,$77,$9B,$DC,$FE,$CF,$EE,$FF,$3B,$AF,$F4,$E6,$1F,$E7,$2C,$BD
	.dc.b	$7C,$BF,$DB,$3D,$ED,$7B,$FF,$E6,$4F,$77,$FD,$3C,$FF,$86,$70,$FF
	.dc.b	$2E,$ED,$FC,$1F,$87,$3B,$7D,$DC,$FF,$D4,$9F,$01,$BF,$84,$FA,$53
	.dc.b	$F4,$30,$FF,$3D,$79,$1F,$E7,$98,$FF,$1C,$F2,$7F,$A9,$5F,$F3,$D0
	.dc.b	$F8,$2F,$EC,$5F,$09,$FE,$85,$CB,$FA,$9C,$FF,$A2,$54,$67,$3F,$F3
	.dc.b	$17,$0F,$F5,$EF,$4F,$D5,$97,$F1,$4B,$FF,$48,$E1,$FD,$76,$31,$1F
	.dc.b	$E7,$A8,$FF,$53,$C7,$37,$FC,$D9,$C0,$8C,$47,$FA,$C3,$CB,$FE,$B4
	.dc.b	$FF,$8E,$CF,$F5,$E9,$FD,$A4,$C1,$C4,$FF,$2C,$1F,$E7,$7F,$81,$FF
	.dc.b	$63,$3F,$EA,$71,$FA,$8E,$1F,$F3,$E3,$76,$FF,$5D,$B9,$C3,$FD,$5B
	.dc.b	$5C,$BB,$7F,$D7,$FF,$0B,$E0,$CB,$FD,$08,$7E,$27,$3D,$E5,$FE,$D2
	.dc.b	$AF,$EF,$DF,$0A,$FC,$AF,$73,$E9,$3F,$2B,$19,$7F,$D5,$7A,$7F,$B2
	.dc.b	$44,$3F,$62,$F3,$2F,$D4,$3F,$CF,$38,$FF,$A3,$DA,$FF,$BF,$63,$57
	.dc.b	$F2,$2F,$9D,$7B,$65,$F9,$27,$0F,$07,$FD,$B7,$18,$BF,$F4,$17,$B9
	.dc.b	$FA,$CF,$98,$FF,$55,$5F,$C3,$9F,$2B,$F1,$71,$F3,$FE,$B5,$E2,$FA
	.dc.b	$8F,$F5,$83,$F4,$BE,$3B,$E1,$C7,$FA,$43,$1F,$DB,$FC,$DF,$BA,$31
	.dc.b	$B7,$F8,$D7,$8B,$19,$FF,$A4,$DF,$BC,$AB,$2E,$22,$6A,$F0,$BE,$7C
	.dc.b	$FE,$7E,$FF,$82,$F3,$7C,$C9,$CF,$6F,$F7,$5F,$1F,$17,$FE,$A1,$CB
	.dc.b	$FD,$81,$BB,$0C,$CF,$8A,$73,$5F,$5B,$1F,$BB,$7F,$67,$2F,$F0,$4D
	.dc.b	$FF,$80,$41,$FE,$B3,$9F,$F5,$9B,$C3,$3C,$DE,$AF,$FD,$6F,$FB,$17
	.dc.b	$DC,$BF,$F2,$AF,$C5,$A3,$FD,$A9,$E9,$BD,$17,$95,$FE,$A0,$FA,$9F
	.dc.b	$F5,$15,$EF,$4F,$0D,$EE,$B5,$DF,$DA,$7F,$C0,$1F,$2A,$FB,$78,$4C
	.dc.b	$39,$4F,$F9,$E6,$73,$3B,$FB,$7E,$53,$98,$CF,$DC,$E3,$E6,$43,$E6
	.dc.b	$BF,$D4,$7E,$FF,$FA,$9B,$2E,$B3,$A9,$CC,$F1,$8E,$0F,$F7,$A9,$DC
	.dc.b	$62,$3F,$D6,$1F,$01,$E2,$F5,$BF,$CF,$9C,$7F,$E7,$7E,$3F,$F7,$D4
	.dc.b	$4D,$3F,$D5,$31,$FE,$16,$F8,$47,$CA,$FF,$85,$BE,$23,$FC,$31,$F0
	.dc.b	$EF,$AE,$F6,$C7,$F8,$6E,$75,$5B,$DB,$E2,$CE,$6C,$F2,$FB,$CF,$F7
	.dc.b	$8F,$9D,$F6,$FD,$2B,$A7,$D0,$E4,$F9,$66,$1F,$94,$FD,$2E,$5C,$7C
	.dc.b	$6F,$FA,$CB,$9F,$7C,$FB,$E2,$F7,$F1,$DC,$7F,$BD,$9F,$EC,$6E,$8F
	.dc.b	$C5,$43,$83,$F9,$1C,$FE,$FA,$3F,$D5,$91,$88,$FD,$61,$C4,$E6,$7F
	.dc.b	$95,$63,$FD,$45,$3F,$2E,$3F,$CB,$A7,$FB,$53,$F6,$3C,$7A,$DF,$E6
	.dc.b	$1A,$F0,$46,$F0,$FD,$1D,$7E,$7F,$07,$E9,$5F,$CC,$DB,$1F,$C9,$3C
	.dc.b	$2F,$D4,$F8,$BF,$DF,$A7,$33,$FD,$07,$C2,$FE,$85,$E1,$F3,$F9,$FF
	.dc.b	$E7,$9F,$2E,$FC,$63,$FA,$05,$FF,$A9,$1C,$9F,$C5,$5C,$1F,$C3,$CF
	.dc.b	$DB,$FB,$3F,$EA,$38,$FE,$35,$87,$FA,$B3,$EB,$F9,$D9,$FE,$95,$F3
	.dc.b	$17,$0F,$E6,$0D,$7E,$59,$F0,$1F,$23,$DC,$7A,$1F,$C1,$A3,$31,$C2
	.dc.b	$3F,$D9,$11,$8D,$BF,$8E,$E5,$FE,$DB,$8C,$57,$F6,$2C,$51,$88,$FE
	.dc.b	$52,$7F,$8F,$B3,$FE,$C1,$C1,$F2,$B0,$FE,$20,$7F,$12,$9F,$F2,$3F
	.dc.b	$07,$F1,$57,$F9,$17,$17,$FB,$A6,$B8,$47,$75,$C5,$F7,$63,$FD,$24
	.dc.b	$FE,$0A,$7F,$D3,$5C,$99,$9F,$E3,$D5,$F8,$F3,$F3,$27,$E4,$70,$FE
	.dc.b	$73,$1F,$E5,$89,$FC,$CC,$E6,$3F,$95,$FA,$1F,$E2,$78,$FF,$01,$9F
	.dc.b	$93,$8F,$F2,$FF,$67,$87,$0F,$F4,$1B,$F9,$0E,$9F,$D5,$75,$FE,$C2
	.dc.b	$FF,$C1,$71,$88,$FE,$43,$D3,$FE,$B8,$AE,$12,$EA,$FD,$BF,$5E,$72
	.dc.b	$FD,$56,$41,$D9,$FE,$C0,$F0,$FF,$87,$78,$5F,$C2,$3C,$17,$81,$9C
	.dc.b	$3F,$A2,$F8,$3F,$C2,$4F,$E2,$26,$5F,$E1,$7E,$17,$F1,$6A,$A7,$F1
	.dc.b	$5E,$1F,$E6,$A7,$C1,$E9,$E0,$FC,$CC,$BF,$5F,$E3,$7F,$2D,$F1,$BF
	.dc.b	$D4,$73,$FC,$AB,$C7,$FE,$67,$E3,$F7,$AD,$57,$FA,$92,$F8,$C7,$F2
	.dc.b	$F8,$0F,$8E,$F3,$72,$7C,$9E,$7F,$E9,$39,$7F,$48,$8C,$43,$9F,$71
	.dc.b	$F2,$E3,$FA,$5F,$2F,$BF,$CA,$FE,$AD,$F1,$7F,$EA,$8F,$88,$79,$BF
	.dc.b	$E3,$AC,$BF,$A9,$D6,$BE,$27,$FD,$4F,$2E,$CF,$91,$DF,$FC,$3B,$BB
	.dc.b	$F0,$DD,$1F,$9B,$F2,$7F,$B7,$79,$9F,$DD,$BB,$FF,$D1,$33,$FD,$C5
	.dc.b	$FF,$49,$9F,$DF,$1F,$EE,$87,$F9,$FF,$D3,$FF,$06,$FA,$5F,$E0,$6F
	.dc.b	$39,$7F,$E1,$1A,$FF,$05,$1F,$DE,$F0,$CD,$FF,$84,$30,$FF,$09,$D6
	.dc.b	$EB,$FC,$3B,$E9,$FF,$84,$E1,$EA,$7F,$86,$B4,$FF,$08,$46,$2B,$FC
	.dc.b	$49,$1F,$E5,$28,$C4,$7F,$89,$B4,$1F,$E2,$E8,$CC,$FC,$38,$FF,$53
	.dc.b	$BE,$8F,$57,$F8,$EC,$9F,$F1,$D6,$3E,$CF,$5B,$FC,$81,$3F,$E4,$5C
	.dc.b	$3F,$8D,$FB,$9F,$EA,$19,$FF,$25,$46,$CF,$D1,$FB,$1F,$EB,$CF,$6B
	.dc.b	$FC,$DD,$7F,$E5,$49,$FE,$46,$7F,$8B,$5C,$65,$F5,$F1,$FF,$5D,$9F
	.dc.b	$32,$BF,$B4,$4F,$F9,$8E,$BF,$CB,$8F,$F9,$A5,$FE,$84,$7D,$6F,$0F
	.dc.b	$E9,$FB,$9F,$66,$3F,$D8,$7F,$01,$FE,$79,$7F,$D2,$67,$F8,$D3,$4F
	.dc.b	$F1,$AC,$66,$3F,$C9,$F0,$AF,$F5,$1F,$2F,$FA,$33,$8F,$FC,$AD,$E6
	.dc.b	$FF,$A2,$75,$FE,$29,$87,$F2,$EE,$5F,$F4,$E4,$FF,$50,$9E,$87,$F6
	.dc.b	$09,$FF,$53,$5F,$FA,$3E,$33,$19,$3C,$46,$F0,$FF,$24,$CF,$F2,$48
	.dc.b	$D3,$5F,$F4,$F5,$57,$FD,$3A,$7F,$A9,$0F,$F2,$D9,$E6,$3C,$57,$FE
	.dc.b	$98,$7F,$A2,$9F,$0C,$C5,$F8,$D6,$1F,$C7,$7D,$EF,$2C,$7C,$9C,$7F
	.dc.b	$D2,$FA,$7F,$AC,$38,$FF,$91,$7B,$9F,$57,$1F,$CD,$57,$FA,$73,$DE
	.dc.b	$FF,$00,$F1,$7F,$81,$4F,$F4,$C7,$17,$A7,$5F,$F5,$0F,$E9,$7A,$9F
	.dc.b	$F0,$57,$4F,$FA,$B3,$9B,$E0,$71,$FF,$A6,$1B,$97,$FA,$53,$7F,$F8
	.dc.b	$03,$7F,$8A,$F8,$7F,$F5,$9F,$47,$C7,$F6,$3F,$D6,$BF,$B1,$F5,$AF
	.dc.b	$F3,$BD,$FF,$A5,$8F,$F6,$5F,$07,$FA,$0B,$87,$F5,$B3,$FE,$DE,$D3
	.dc.b	$FD,$22,$FF,$B8,$39,$7F,$D0,$B5,$F3,$AF,$FC,$2F,$7F,$BD,$8C,$47
	.dc.b	$CA,$8F,$F7,$7D,$7F,$10,$8F,$F6,$1B,$FE,$D8,$9F,$F6,$4B,$F0,$7D
	.dc.b	$3F,$F6,$73,$FE,$88,$9F,$F6,$39,$FE,$DD,$D7,$FC,$79,$EE,$7A,$E7
	.dc.b	$F5,$35,$FE,$74,$F0,$BF,$D0,$DF,$10,$FA,$37,$D1,$A7,$FD,$09,$F0
	.dc.b	$1F,$CB,$3E,$0D,$7B,$AD,$46,$27,$53,$9A,$FF,$49,$7D,$1F,$FA,$0B
	.dc.b	$CE,$10,$86,$2B,$FA,$23,$83,$FC,$B8,$FF,$9F,$3D,$2F,$F4,$39,$0F
	.dc.b	$F4,$11,$47,$CA,$61,$D7,$AD,$5E,$3E,$FF,$FA,$53,$C7,$FC,$77,$07
	.dc.b	$FA,$0F,$8F,$FC,$B6,$7E,$B7,$1F,$F3,$CF,$B3,$FD,$D7,$DA,$3F,$E8
	.dc.b	$03,$1F,$51,$7D,$BF,$EE,$32,$FA,$5C,$3F,$DE,$3E,$A3,$FD,$47,$97
	.dc.b	$83,$9F,$FC,$F3,$C3,$EF,$3E,$17,$99,$FE,$0C,$7E,$17,$C0,$7F,$A0
	.dc.b	$77,$F3,$DF,$F9,$EF,$7F,$E0,$9D,$7F,$E8,$58,$E9,$C1,$FD,$7E,$7F
	.dc.b	$3F,$1F,$E6,$F8,$FD,$F4,$7F,$9E,$4F,$F3,$C4,$7F,$2C,$8F,$F1,$84
	.dc.b	$7E,$F6,$7F,$CE,$AF,$FB,$5A,$BC,$33,$FE,$3C,$8F,$8E,$7F,$B9,$1F
	.dc.b	$F2,$0B,$E4,$7F,$C8,$73,$FE,$E8,$AF,$F1,$14,$62,$7F,$2C,$B8,$7F
	.dc.b	$8B,$57,$F9,$C2,$7F,$88,$9F,$E8,$4F,$86,$FF,$2C,$3F,$C8,$5A,$FF
	.dc.b	$BA,$BC,$2F,$F7,$5C,$62,$1F,$0B,$FE,$77,$7F,$90,$1F,$E6,$CA,$FF
	.dc.b	$3C,$38,$38,$FC,$37,$8B,$E1,$BF,$CE,$DE,$0F,$F1,$F9,$FF,$37,$CF
	.dc.b	$F2,$B7,$FC,$D7,$E0,$FF,$09,$69,$DB,$2F,$F3,$AC,$7F,$33,$C7,$FC
	.dc.b	$E8,$F5,$39,$7B,$3F,$06,$BC,$59,$23,$E3,$43,$CC,$FF,$32,$FC,$37
	.dc.b	$F5,$9A,$FF,$26,$70,$6E,$3E,$4D,$FD,$5B,$FF,$58,$9F,$86,$71,$C3
	.dc.b	$F8,$A7,$EF,$C6,$7D,$3F,$F6,$3D,$F4,$9F,$C3,$CF,$F9,$D3,$97,$F8
	.dc.b	$54,$F8,$39,$BF,$CB,$D3,$F9,$48,$B5,$F1,$F2,$CE,$3D,$F0,$FC,$E7
	.dc.b	$37,$F0,$0E,$6E,$2F,$E8,$4F,$E0,$B6,$F8,$2E,$F8,$FF,$9A,$73,$FF
	.dc.b	$30,$74,$FF,$98,$BB,$7F,$DE,$9D,$78,$7F,$6D,$D7,$7B,$3F,$67,$16
	.dc.b	$BF,$CD,$3D,$BF,$E6,$2E,$1F,$CF,$69,$DD,$FC,$F6,$7F,$C1,$23,$F8
	.dc.b	$87,$77,$F0,$FE,$DF,$F3,$13,$FC,$3C,$77,$E1,$C4,$F0,$7F,$8B,$F8
	.dc.b	$1F,$E5,$CE,$1F,$DD,$CF,$F1,$E7,$F8,$03,$FC,$A3,$C6,$FE,$63,$E1
	.dc.b	$BC,$57,$C4,$7F,$98,$5F,$E6,$33,$A5,$E2,$E7,$C6,$FE,$61,$1F,$CD
	.dc.b	$B0,$FF,$63,$ED,$FC,$3D,$FF,$07,$78,$7F,$E5,$8E,$0F,$C6,$1F,$E1
	.dc.b	$BE,$BF,$F3,$05,$7F,$A5,$A0,$7F,$7E,$7F,$CC,$2F,$FB,$56,$78,$74
	.dc.b	$7F,$67,$8F,$EE,$95,$FD,$75,$EC,$7F,$8E,$FC,$FF,$F7,$ED,$7F,$60
	.dc.b	$F2,$BF,$CC,$BE,$C7,$F6,$FC,$3F,$C4,$5E,$66,$30,$FE,$EF,$1F,$E2
	.dc.b	$5A,$C4,$7F,$B6,$2F,$CA,$FA,$7C,$DF,$EB,$CE,$3C,$1F,$F2,$D3,$FB
	.dc.b	$37,$FC,$A4,$FF,$40,$F3,$FF,$CA,$CF,$F0,$4C,$7F,$8F,$7A,$1F,$E0
	.dc.b	$8D,$5E,$8F,$E4,$B6,$F8,$DE,$97,$F8,$4E,$7F,$CA,$18,$7F,$34,$DB
	.dc.b	$FC,$9A,$FC,$5C,$EB,$FE,$FE,$F5,$3F,$D8,$5E,$BF,$F8,$9A,$1A,$7F
	.dc.b	$93,$BC,$7F,$F0,$BC,$E7,$87,$FC,$97,$5F,$C1,$38,$7F,$66,$FF,$03
	.dc.b	$6A,$7A,$3F,$E4,$A3,$F8,$E7,$95,$9B,$FE,$31,$1F,$9F,$3F,$CD,$67
	.dc.b	$F9,$2C,$C7,$27,$F8,$45,$FF,$2A,$78,$FF,$75,$F9,$F5,$FE,$50,$7F
	.dc.b	$C5,$67,$F9,$0E,$B4,$FF,$A9,$1A,$FF,$19,$F5,$FA,$DF,$F8,$B3,$CD
	.dc.b	$FF,$24,$61,$C1,$FF,$72,$FA,$8F,$D3,$BF,$80,$D7,$F9,$21,$BC,$29
	.dc.b	$7E,$C7,$DB,$8B,$FC,$95,$9F,$E6,$30,$FE,$10,$7F,$88,$2B,$EB,$B8
	.dc.b	$3F,$D1,$6E,$1F,$F1,$FC,$7F,$2F,$8C,$47,$F6,$E7,$F0,$71,$D6,$3F
	.dc.b	$DF,$67,$F3,$2E,$5F,$F4,$24,$EA,$7E,$69,$FA,$AC,$BF,$CC,$3C,$FF
	.dc.b	$E3,$9B,$FE,$73,$D1,$FE,$59,$AF,$96,$7E,$C3,$E4,$BF,$CC,$FD,$5B
	.dc.b	$8F,$D9,$E7,$FE,$56,$7C,$5A,$B0,$FF,$3D,$E6,$BF,$DA,$47,$F4,$6B
	.dc.b	$F9,$11,$88,$E1,$7F,$E5,$85,$7F,$C0,$63,$FA,$87,$36,$63,$1C,$7F
	.dc.b	$E5,$7E,$BF,$F3,$36,$3F,$D0,$EB,$96,$1F,$EA,$53,$F9,$C7,$37,$D7
	.dc.b	$C3,$FC,$B1,$8F,$FB,$AE,$7E,$D1,$FE,$7B,$F9,$1F,$DD,$FD,$3F,$F5
	.dc.b	$CF,$63,$F8,$BD,$7F,$97,$32,$FF,$2D,$BF,$02,$FF,$CB,$73,$FE,$50
	.dc.b	$F6,$7F,$CA,$7E,$17,$F9,$6E,$31,$E3,$7F,$97,$BE,$5B,$FC,$C1,$E2
	.dc.b	$7F,$96,$AF,$7F,$59,$FE,$55,$F2,$5F,$33,$FE,$91,$C7,$F9,$B6,$9F
	.dc.b	$E4,$3E,$EF,$E1,$FE,$60,$FF,$27,$ED,$FE,$2D,$8C,$C6,$67,$81,$FD
	.dc.b	$32,$7D,$E7,$B6,$7F,$8C,$F8,$BE,$F9,$FF,$15,$47,$F1,$CF,$43,$FC
	.dc.b	$9F,$F3,$BE,$38,$C7,$A4,$F4,$FF,$C5,$D1,$F8,$BC,$7F,$BA,$67,$FE
	.dc.b	$41,$C7,$FC,$DD,$A7,$AF,$C6,$FF,$26,$47,$F9,$75,$F0,$D7,$FA,$D0
	.dc.b	$FF,$21,$38,$7F,$D1,$91,$FE,$DF,$AE,$91,$FA,$93,$FB,$A3,$87,$F4
	.dc.b	$67,$28,$FC,$D9,$F9,$88,$7A,$FD,$7B,$FF,$CA,$8F,$78,$FF,$38,$7C
	.dc.b	$86,$6F,$FC,$A8,$E6,$FF,$C9,$DF,$8F,$FE,$FE,$BF,$C9,$FE,$57,$FB
	.dc.b	$0D,$B3,$FE,$F5,$3D,$1E,$A7,$F0,$5F,$7D,$F8,$57,$F1,$49,$46,$1F
	.dc.b	$39,$FE,$FD,$70,$FE,$E8,$78,$DF,$AF,$8C,$E7,$FC,$9E,$3F,$53,$18
	.dc.b	$8F,$F6,$4F,$D3,$7F,$95,$3F,$03,$F0,$49,$FE,$31,$F0,$57,$0F,$0F
	.dc.b	$82,$8C,$47,$84,$FE,$E6,$F5,$C3,$FB,$0C,$30,$CE,$5F,$CC,$B2,$FE
	.dc.b	$09,$7B,$AF,$EC,$F1,$FB,$E8,$C4,$7F,$67,$6B,$FD,$B3,$D5,$FF,$0E
	.dc.b	$BE,$47,$E6,$E3,$FE,$52,$8F,$F0,$2E,$DF,$DF,$1F,$E4,$92,$DF,$F9
	.dc.b	$A3,$FE,$50,$CF,$FD,$03,$3F,$DD,$63,$11,$FA,$43,$F8,$D3,$F8,$68
	.dc.b	$FD,$46,$FE,$EC,$3F,$CC,$9C,$3F,$85,$BF,$F2,$B7,$1F,$63,$F4,$4F
	.dc.b	$F9,$1F,$0F,$CA,$78,$BF,$E6,$68,$DE,$3D,$0F,$15,$73,$8F,$0B,$FA
	.dc.b	$68,$C9,$FE,$22,$ED,$FF,$6C,$F3,$73,$AF,$3F,$3F,$81,$FE,$53,$8F
	.dc.b	$FB,$23,$A3,$FC,$E1,$3F,$B8,$EA,$E1,$5F,$E7,$D8,$FC,$E6,$3F,$CF
	.dc.b	$2F,$FB,$A7,$67,$F9,$F3,$AB,$B4,$7F,$B8,$67,$E1,$3F,$DE,$7B,$7F
	.dc.b	$89,$7B,$3D,$23,$F8,$BC,$FC,$DC,$3F,$C4,$BA,$FA,$32,$E7,$1F,$E2
	.dc.b	$8F,$0F,$FE,$00,$F8,$3F,$C8,$B0,$FE,$4F,$E1,$7F,$93,$9B,$E2,$3C
	.dc.b	$4F,$F4,$A7,$C0,$7B,$F9,$7F,$D1,$BA,$FF,$84,$4F,$E6,$DC,$3F,$E4
	.dc.b	$AF,$1F,$F9,$DE,$DF,$D5,$63,$FA,$63,$FE,$7B,$9F,$E8,$50,$F2,$BF
	.dc.b	$A5,$CF,$F9,$FF,$CA,$FF,$1F,$BF,$D6,$4F,$E9,$73,$BA,$FE,$C3,$87
	.dc.b	$F8,$52,$B7,$5F,$D6,$A3,$FC,$2F,$E5,$2F,$FB,$25,$FF,$86,$2B,$FB
	.dc.b	$24,$62,$3F,$C4,$71,$FE,$AC,$78,$79,$9F,$DC,$57,$67,$F9,$3D,$FE
	.dc.b	$A4,$7F,$74,$3F,$C9,$B3,$FC,$0C,$A7,$F2,$41,$E6,$FF,$9B,$FD,$7F
	.dc.b	$F1,$04,$79,$5F,$CC,$FA,$3F,$E7,$3F,$84,$FF,$14,$FA,$1E,$97,$FC
	.dc.b	$3D,$9F,$F8,$5F,$5F,$C9,$19,$CB,$FB,$26,$3F,$9F,$9F,$F1,$2C,$7F
	.dc.b	$0C,$7F,$C5,$8F,$F8,$66,$3F,$27,$18,$8F,$F2,$D7,$AD,$FE,$36,$9F
	.dc.b	$E1,$BE,$9F,$F9,$63,$AB,$FC,$6D,$87,$F9,$2E,$FF,$93,$E5,$5F,$E2
	.dc.b	$DB,$5C,$3F,$C9,$51,$FC,$1A,$FD,$7E,$37,$58,$FD,$E3,$5F,$E3,$67
	.dc.b	$F9,$55,$FF,$44,$BF,$CE,$E3,$F8,$75,$7F,$BC,$DB,$1F,$C8,$AF,$FC
	.dc.b	$A8,$FE,$4E,$7F,$DF,$1E,$1E,$A3,$FD,$08,$7F,$65,$BE,$B4,$F5,$BC
	.dc.b	$13,$FD,$24,$FF,$50,$F2,$7F,$A8,$A3,$38,$7F,$A4,$F2,$E1,$97,$5B
	.dc.b	$FF,$3D,$CF,$F9,$57,$D6,$F3,$E5,$FE,$38,$F5,$FF,$CB,$55,$FC,$A0
	.dc.b	$FC,$5C,$7F,$8F,$0F,$F2,$4F,$AF,$CE,$3F,$C9,$7C,$5F,$CB,$5F,$E1
	.dc.b	$3E,$27,$FA,$42,$BF,$D4,$39,$FF,$97,$6B,$86,$5F,$E6,$2B,$F8,$47
	.dc.b	$C5,$BF,$F2,$A5,$F0,$BF,$F2,$47,$CB,$7F,$AF,$A7,$35,$F8,$5F,$87
	.dc.b	$FF,$26,$46,$30,$FF,$5A,$CF,$F9,$C7,$9F,$E0,$70,$7F,$96,$63,$F8
	.dc.b	$11,$FC,$BF,$AF,$FC,$BE,$FB,$EB,$FA,$BF,$C2,$7F,$96,$63,$79,$B7
	.dc.b	$FF,$2B,$65,$FB,$3D,$7D,$B1,$FB,$38,$C5,$7F,$7D,$9F,$F3,$1E,$1F
	.dc.b	$E5,$C8,$EC,$7C,$09,$FE,$0C,$F9,$39,$7F,$CC,$D8,$7F,$97,$B9,$3F
	.dc.b	$D1,$59,$FC,$D9,$C9,$FE,$5D,$AF,$46,$1F,$EA,$0A,$DC,$60,$FF,$72
	.dc.b	$AE,$CF,$F2,$F4,$7D,$D3,$D9,$87,$F9,$D7,$2F,$F4,$2E,$1F,$84,$9F
	.dc.b	$C0,$E1,$FE,$62,$7F,$09,$87,$FB,$73,$5F,$F1,$47,$77,$E0,$71,$FE
	.dc.b	$A1,$E6,$23,$11,$FE,$40,$7F,$03,$19,$BF,$F7,$CF,$8A,$F1,$1E,$F3
	.dc.b	$E2,$DF,$18,$F8,$C7,$73,$F2,$BF,$F3,$7E,$4C,$B9,$63,$C7,$0E,$16
	.dc.b	$F8,$ED,$D6,$B2,$EB,$97,$4F,$CB,$79,$F9,$A1,$F1,$2F,$B2,$84,$E0
	.dc.b	$F5,$6D,$BA,$DC,$E8,$7D,$9F,$48,$CC,$72,$FB,$9F,$4F,$A7,$FE,$77
	.dc.b	$F7,$FF,$CF,$9C,$BF,$E7,$6F,$89,$FE,$1F,$D5,$EE,$F4,$FF,$6D,$D3
	.dc.b	$FE,$07,$DB,$C3,$1F,$B5,$D8,$7D,$42,$FC,$1B,$E0,$DF,$4E,$FB,$67
	.dc.b	$C8,$3E,$DF,$1F,$0B,$F3,$30,$E1,$7B,$AD,$5F,$0C,$FE,$0D,$6A,$39
	.dc.b	$E1,$D7,$F0,$75,$7F,$18,$CF,$FA,$1B,$FE,$FD,$E5,$F9,$DC,$FF,$1F
	.dc.b	$9B,$E7,$4E,$67,$E3,$C7,$D1,$E7,$FF,$51,$56,$F9,$FA,$FA,$3F,$E4
	.dc.b	$3C,$7F,$B0,$F9,$5F,$E9,$F7,$E1,$F1,$FC,$57,$FC,$FD,$DF,$F9,$4E
	.dc.b	$5F,$F4,$2B,$C7,$9B,$F4,$71,$FE,$BD,$8C,$78,$FF,$E3,$DD,$FF,$C8
	.dc.b	$16,$C3,$E9,$DF,$FB,$0F,$E0,$77,$3F,$E8,$23,$F7,$E5,$3F,$C2,$5E
	.dc.b	$3F,$D6,$7F,$69,$3F,$EE,$5B,$FF,$2A,$47,$EB,$8F,$E7,$A7,$FA,$18
	.dc.b	$E6,$7F,$BB,$BE,$1F,$FD,$03,$8F,$F5,$28,$FF,$43,$FC,$37,$F2,$E9
	.dc.b	$CC,$7F,$47,$1E,$77,$E1,$A3,$3A,$FF,$6A,$C7,$8D,$76,$AF,$F4,$77
	.dc.b	$5F,$F3,$F8,$FF,$49,$BF,$CE,$BE,$17,$FD,$03,$97,$F6,$2C,$3F,$D0
	.dc.b	$93,$B9,$EB,$5E,$FB,$FF,$40,$BD,$6B,$FD,$15,$97,$F5,$5F,$5F,$FD
	.dc.b	$06,$F5,$3C,$1B,$FF,$9F,$AB,$FA,$39,$FE,$75,$F8,$5F,$F3,$E6,$FF
	.dc.b	$C4,$E3,$FC,$43,$3F,$29,$F9,$E7,$E9,$3C,$FF,$E5,$B1,$9D,$7D,$0F
	.dc.b	$DC,$3F,$0D,$3C,$B8,$7F,$09,$C7,$FE,$5E,$E3,$FB,$B3,$FE,$63,$73
	.dc.b	$C3,$FE,$79,$BF,$EF,$13,$FE,$19,$8F,$F3,$7F,$27,$08,$F2,$BC,$23
	.dc.b	$C6,$FF,$98,$BA,$7F,$3F,$D3,$FC,$22,$FF,$45,$D1,$F9,$F6,$C7,$E9
	.dc.b	$DF,$D6,$5F,$FA,$27,$CF,$FF,$30,$E3,$FC,$03,$C1,$FD,$FF,$77,$ED
	.dc.b	$07,$6E,$A5,$DF,$FB,$9F,$03,$F7,$A6,$5C,$78,$1F,$C1,$33,$FE,$1F
	.dc.b	$5F,$E0,$43,$FD,$4D,$E1,$FA,$F4,$FF,$52,$BF,$C6,$63,$FD,$06,$FE
	.dc.b	$7F,$C5,$FE,$3D,$3F,$E8,$88,$F8,$1B,$FF,$01,$F0,$BC,$9C,$3F,$C1
	.dc.b	$38,$7F,$95,$78,$58,$8F,$8B,$1E,$9F,$27,$F9,$D7,$97,$FC,$FB,$CD
	.dc.b	$C7,$9D,$FD,$2B,$CC,$FE,$7B,$D2,$F2,$F3,$E7,$7F,$45,$F3,$7F,$A5
	.dc.b	$F9,$DF,$0F,$C0,$FC,$1F,$17,$F0,$5E,$3F,$5F,$2F,$F6,$77,$E7,$7B
	.dc.b	$47,$F2,$31,$FB,$6F,$1F,$FC,$F5,$DF,$FD,$A5,$61,$FE,$74,$B4,$FF
	.dc.b	$89,$FC,$1F,$EF,$F3,$9F,$47,$FC,$93,$7F,$DE,$EF,$FC,$0B,$3F,$E0
	.dc.b	$AC,$3F,$BD,$F8,$9F,$E8,$7B,$FF,$02,$65,$E3,$C9,$E2,$E6,$FF,$BF
	.dc.b	$47,$F8,$22,$FF,$D0,$F1,$FE,$12,$C3,$FC,$2B,$8F,$F8,$63,$2F,$F0
	.dc.b	$AB,$FE,$1A,$9F,$76,$7F,$E9,$2F,$77,$FB,$F3,$FE,$91,$97,$C5,$FF
	.dc.b	$93,$7C,$BF,$73,$FC,$B7,$6F,$F5,$67,$AB,$FE,$32,$C7,$FD,$14,$7F
	.dc.b	$1A,$F5,$3F,$D1,$CF,$CD,$E0,$FF,$1F,$BA,$3F,$C6,$2F,$CF,$F5,$BF
	.dc.b	$D6,$DE,$C7,$F9,$2B,$3F,$F2,$75,$7F,$1D,$3E,$0B,$87,$FD,$13,$1F
	.dc.b	$6F,$87,$FD,$5D,$EC,$F1,$F6,$BE,$B7,$17,$D5,$7E,$C7,$07,$FA,$2F
	.dc.b	$0F,$30,$F7,$3F,$CD,$AF,$F8,$87,$8F,$1E,$77,$F9,$B7,$97,$FC,$55
	.dc.b	$D9,$FE,$5E,$CB,$F8,$1D,$7F,$A2,$39,$FF,$D2,$51,$FC,$DC,$E7,$87
	.dc.b	$F4,$39,$FC,$3E,$1F,$E6,$DA,$B3,$FD,$00,$FE,$91,$1F,$E6,$F7,$FA
	.dc.b	$51,$FE,$91,$3F,$C7,$87,$FA,$58,$FE,$E0,$7F,$A4,$7D,$CA,$FF,$4C
	.dc.b	$6B,$FD,$49,$BC,$7F,$EC,$AF,$6F,$FD,$61,$8F,$E5,$F9,$3F,$D1,$17
	.dc.b	$FE,$49,$8F,$EB,$47,$FA,$11,$FE,$41,$1F,$E5,$69,$FA,$51,$FE,$7D
	.dc.b	$FC,$07,$FC,$F9,$3F,$E5,$E1,$EF,$77,$E7,$F6,$57,$F6,$0B,$FF,$3F
	.dc.b	$BE,$EC,$74,$F6,$BF,$F6,$BE,$AA,$FC,$9C,$FC,$23,$FC,$47,$D5,$EB
	.dc.b	$FC,$A7,$C3,$7F,$22,$FF,$CE,$95,$FE,$84,$3F,$D0,$AE,$23,$FC,$DF
	.dc.b	$1F,$E5,$FB,$FF,$79,$46,$23,$FA,$64,$7E,$D9,$FF,$77,$1F,$E7,$07
	.dc.b	$0F,$F7,$1E,$1F,$37,$C0,$FC,$28,$FF,$4F,$4F,$F0,$47,$E6,$1F,$E3
	.dc.b	$AD,$BF,$CE,$6F,$F8,$F2,$FF,$7C,$FF,$5B,$7F,$D4,$27,$FB,$27,$87
	.dc.b	$FC,$83,$8F,$F8,$C3,$D7,$F7,$CF,$FA,$23,$DC,$46,$23,$FD,$1F,$1F
	.dc.b	$E5,$AC,$3F,$11,$82,$FF,$CC,$FE,$33,$E0,$DF,$20,$F8,$5F,$F3,$84
	.dc.b	$FF,$9E,$31,$FF,$31,$F9,$9F,$E7,$BB,$DD,$69,$E6,$62,$3F,$D0,$1F
	.dc.b	$21,$FE,$65,$F3,$FF,$CD,$47,$F9,$AE,$FF,$CD,$AE,$0F,$F3,$77,$A5
	.dc.b	$FE,$68,$F4,$C4,$21,$92,$8F,$87,$C3,$AF,$84,$AF,$1F,$84,$FF,$45
	.dc.b	$F8,$FF,$95,$E3,$FF,$37,$79,$9F,$E4,$DE,$3F,$F3,$9E,$9F,$E6,$0F
	.dc.b	$75,$F1,$3F,$DA,$9B,$ED,$FF,$9B,$E7,$53,$93,$FB,$7F,$C8,$FF,$9B
	.dc.b	$6B,$67,$0F,$93,$C4,$EE,$DF,$27,$FE,$6A,$F8,$3F,$D2,$E5,$F7,$B2
	.dc.b	$FF,$0D,$3F,$5C,$FC,$0E,$9F,$E4,$B9,$FC,$24,$7F,$92,$4F,$CF,$1F
	.dc.b	$C3,$FE,$1F,$FC,$8A,$FF,$07,$6B,$FA,$56,$BF,$AA,$3F,$C8,$A7,$21
	.dc.b	$F0,$BF,$D8,$E3,$71,$FE,$4E,$EC,$FF,$23,$BF,$B3,$3F,$CC,$11,$FE
	.dc.b	$48,$7F,$8B,$47,$F9,$21,$D1,$F9,$1F,$86,$D3,$FC,$6D,$EB,$E2,$FF
	.dc.b	$92,$A3,$C4,$EF,$C0,$F8,$B8,$7F,$92,$A7,$9D,$FF,$93,$9E,$A7,$F9
	.dc.b	$E7,$9B,$E5,$BF,$E4,$8C,$BE,$3C,$FF,$9F,$5F,$E3,$B1,$F3,$4F,$D2
	.dc.b	$67,$FE,$01,$F8,$1F,$A3,$1F,$EB,$67,$EF,$EF,$FE,$AF,$E0,$FF,$16
	.dc.b	$F0,$FF,$8A,$F8,$BE,$FC,$7F,$8C,$38,$B0,$FD,$F8,$B5,$C4,$71,$FF
	.dc.b	$8D,$F0,$FD,$AE,$9F,$E3,$9E,$5F,$CB,$69,$FE,$5C,$E5,$FF,$17,$73
	.dc.b	$FE,$80,$73,$FF,$8F,$DC,$0E,$1C,$38,$C7,$FD,$B6,$EB,$A3,$FC,$61
	.dc.b	$D5,$FB,$1E,$AA,$FE,$CE,$BF,$7F,$DF,$FC,$17,$C0,$FE,$09,$DD,$FE
	.dc.b	$F2,$9C,$F7,$FE,$FE,$9D,$FF,$E3,$68,$C1,$FE,$31,$E8,$FC,$1D,$7F
	.dc.b	$16,$7F,$9F,$70,$7F,$8A,$23,$11,$F0,$23,$FC,$64,$38,$7F,$C6,$91
	.dc.b	$FE,$B0,$F0,$BF,$91,$C7,$F2,$D8,$FE,$71,$E2,$F0,$8C,$78,$7F,$7A
	.dc.b	$7F,$C6,$FC,$BC,$FC,$5F,$E6,$D1,$FC,$C7,$1F,$F1,$63,$FD,$18,$FF
	.dc.b	$5D,$63,$8B,$FE,$F5,$1F,$EB,$DC,$7F,$C5,$EF,$F9,$D4,$FF,$60,$78
	.dc.b	$F8,$BF,$E9,$50,$9D,$D7,$F7,$C3,$FC,$09,$93,$C7,$FE,$B3,$58,$F5
	.dc.b	$BF,$81,$F8,$7F,$E2,$A8,$F8,$9F,$15,$70,$FF,$01,$66,$9F,$F1,$64
	.dc.b	$63,$6F,$91,$E6,$7F,$43,$F3,$3F,$CA,$DE,$AF,$E9,$38,$3F,$C4,$F8
	.dc.b	$7D,$2F,$57,$E7,$7A,$9F,$1F,$2F,$F0,$0E,$58,$AF,$EB,$23,$D4,$FF
	.dc.b	$15,$3F,$C1,$F4,$F1,$7A,$9F,$E0,$EA,$FF,$14,$CF,$F0,$5E,$3F,$F3
	.dc.b	$66,$1F,$E1,$48,$FE,$A4,$DC,$FF,$C4,$71,$FC,$44,$FF,$75,$78,$5F
	.dc.b	$EE,$BE,$5F,$EC,$B1,$F4,$A3,$11,$F8,$18,$FF,$13,$C6,$35,$FF,$1A
	.dc.b	$43,$D3,$C3,$FE,$29,$6B,$F7,$1A,$FC,$BB,$7A,$6B,$CF,$99,$FE,$85
	.dc.b	$8C,$57,$E2,$70,$FF,$1B,$CF,$FA,$0B,$D2,$7C,$A7,$8F,$3F,$F3,$D1
	.dc.b	$FE,$8D,$8C,$DF,$18,$CD,$FF,$87,$FF,$61,$F1,$CF,$F8,$C4,$FD,$1F
	.dc.b	$67,$F5,$99,$FF,$12,$7A,$FF,$0A,$3F,$83,$E0,$8E,$3A,$EA,$FC,$C7
	.dc.b	$F8,$96,$5D,$1C,$72,$FD,$7F,$47,$F9,$16,$3D,$F1,$FE,$31,$F5,$B4
	.dc.b	$FF,$A9,$7A,$BF,$C6,$F8,$7E,$7B,$E4,$3D,$9D,$DC,$E3,$E5,$73,$FF
	.dc.b	$8E,$63,$11,$FE,$9A,$30,$79,$2F,$FD,$3D,$38,$8F,$F1,$AC,$7F,$89
	.dc.b	$33,$FF,$1A,$AD,$BF,$80,$74,$8F,$F1,$E4,$67,$83,$F6,$7D,$18,$7F
	.dc.b	$9E,$C3,$E3,$78,$61,$D2,$73,$1F,$B7,$8E,$D8,$FF,$05,$C7,$F8,$1D
	.dc.b	$FD,$97,$EE,$3F,$EC,$03,$EE,$5F,$E5,$2F,$F6,$95,$FD,$AA,$7F,$D9
	.dc.b	$9E,$BF,$F1,$E7,$FD,$9D,$7F,$6E,$31,$1E,$D3,$FC,$2C,$E2,$FF,$C2
	.dc.b	$BD,$1D,$3C,$67,$8C,$FC,$DF,$FC,$69,$E4,$7F,$36,$C3,$F9,$CC,$63
	.dc.b	$D9,$7E,$49,$FE,$40,$8C,$7E,$54,$AF,$F1,$9F,$98,$F9,$37,$D7,$42
	.dc.b	$4E,$A3,$11,$AF,$B2,$E5,$B7,$C0,$B1,$CE,$77,$F6,$1F,$E3,$DF,$49
	.dc.b	$F4,$90,$BF,$49,$97,$06,$B3,$FE,$B3,$EA,$FF,$8F,$E3,$3F,$63,$B7
	.dc.b	$97,$D3,$7F,$8F,$B9,$FF,$DE,$7E,$77,$CA,$DB,$E9,$FA,$7F,$EE,$BF
	.dc.b	$3B,$FD,$0D,$9F,$F8,$FF,$E6,$9F,$26,$F9,$67,$CA,$E3,$F0,$2F,$F1
	.dc.b	$E5,$6A,$73,$5B,$BE,$4F,$1B,$E7,$F8,$27,$C4,$F0,$BF,$15,$B7,$FA
	.dc.b	$97,$C1,$FA,$1E,$A7,$C7,$E9,$F8,$DD,$3F,$33,$3F,$E3,$18,$7F,$3D
	.dc.b	$9F,$97,$D7,$F4,$29,$D7,$CA,$B5,$5C,$A7,$EC,$F5,$E6,$7F,$B5,$67
	.dc.b	$F1,$0F,$F5,$B6,$3D,$A3,$39,$FF,$A9,$FC,$8E,$15,$F9,$63,$D1,$4F
	.dc.b	$8A,$F9,$71,$CC,$FF,$47,$B8,$AE,$52,$F9,$9C,$74,$7E,$71,$D7,$37
	.dc.b	$D2,$E1,$FE,$3F,$97,$E4,$DF,$F4,$D6,$5F,$A8,$6E,$3F,$97,$EC,$DF
	.dc.b	$C5,$BF,$24,$FE,$4F,$B2,$30,$FF,$20,$3F,$CC,$91,$FC,$FB,$CE,$FF
	.dc.b	$2E,$36,$F8,$3F,$D0,$BE,$7B,$FB,$65,$BE,$53,$AB,$70,$E8,$7F,$AD
	.dc.b	$3E,$3B,$0F,$F8,$00,$FF,$2F,$4F,$92,$3C,$71,$FE,$04,$AF,$EF,$11
	.dc.b	$8D,$3F,$AE,$CB,$24,$27,$4F,$F5,$F8,$65,$9C,$BE,$4D,$6A,$3F,$88
	.dc.b	$DF,$F1,$2F,$81,$FF,$55,$BE,$83,$FD,$84,$B5,$F9,$F8,$A3,$13,$FE
	.dc.b	$05,$6B,$ED,$DB,$5C,$1F,$6A,$3F,$8F,$C6,$1F,$F0,$A7,$7F,$AB,$1F
	.dc.b	$F0,$46,$1F,$E3,$C3,$B7,$27,$1A,$FF,$6E,$E3,$F8,$D7,$8C,$FE,$34
	.dc.b	$FF,$3A,$75,$7F,$52,$7A,$4F,$E6,$70,$FD,$E5,$78,$A3,$13,$C4,$F6
	.dc.b	$57,$EB,$FA,$BF,$55,$CF,$8B,$FD,$4C,$7E,$B7,$A3,$11,$FC,$6B,$A7
	.dc.b	$F5,$91,$FB,$0C,$7F,$C0,$33,$FD,$D3,$1F,$DE,$47,$F0,$2E,$DF,$E0
	.dc.b	$BC,$1F,$E8,$5E,$EF,$E1,$1D,$BF,$C2,$6B,$4F,$F0,$D7,$9E,$DE,$AC
	.dc.b	$7E,$1F,$81,$FA,$8C,$BE,$45,$7F,$22,$F1,$7F,$92,$F8,$6F,$13,$FC
	.dc.b	$FF,$18,$F1,$2C,$EA,$B8,$3F,$CB,$3C,$5F,$BD,$8F,$C3,$E0,$FE,$6B
	.dc.b	$5F,$CF,$5F,$E1,$CD,$7F,$A1,$78,$7C,$34,$FF,$28,$79,$9F,$D4,$BC
	.dc.b	$8F,$E9,$DE,$4E,$65,$39,$F3,$3F,$D1,$5E,$5B,$CB,$D4,$E7,$E0,$9E
	.dc.b	$6F,$FA,$3B,$5F,$EC,$18,$FE,$6B,$C0,$FC,$A7,$C7,$7E,$07,$9B,$FD
	.dc.b	$10,$D7,$FD,$0D,$E0,$FF,$71,$AF,$F3,$27,$BB,$FE,$83,$F4,$3F,$BB
	.dc.b	$CD,$9F,$EF,$87,$F8,$FE,$BF,$5F,$D5,$FE,$03,$C7,$FD,$23,$E8,$FF
	.dc.b	$83,$63,$13,$FB,$48,$FF,$3E,$47,$F8,$53,$0F,$F0,$9F,$A9,$FE,$81
	.dc.b	$AF,$F0,$96,$1F,$E7,$58,$DF,$AB,$FE,$52,$9F,$F0,$FD,$FF,$A0,$6B
	.dc.b	$FC,$45,$1F,$BC,$7F,$D4,$13,$FE,$28,$F6,$99,$62,$7F,$A5,$9F,$E5
	.dc.b	$73,$FD,$4D,$EC,$7F,$1D,$BB,$1F,$D4,$1F,$F1,$C4,$FF,$A1,$23,$F8
	.dc.b	$CE,$BF,$03,$D9,$FD,$67,$B5,$DB,$3F,$F2,$55,$7F,$07,$CB,$FD,$0E
	.dc.b	$FF,$01,$8C,$BF,$A1,$C7,$FD,$6F,$BE,$9F,$E9,$07,$EF,$E5,$EE,$7F
	.dc.b	$3E,$7F,$CC,$35,$C2,$7F,$CC,$A7,$FA,$FF,$DD,$FA,$EF,$F1,$EF,$7F
	.dc.b	$F5,$EF,$F2,$6D,$3F,$92,$C6,$23,$FD,$05,$1F,$C9,$3D,$6F,$EB,$91
	.dc.b	$FC,$B2,$D9,$FF,$A3,$A3,$1B,$57,$FD,$15,$3F,$EA,$BA,$E9,$A7,$F2
	.dc.b	$A8,$E3,$18,$7F,$28,$D7,$55,$FE,$CF,$7F,$98,$FB,$9F,$E8,$A7,$07
	.dc.b	$E0,$E3,$0F,$73,$B7,$17,$FA,$EF,$DF,$FF,$63,$46,$32,$7B,$AC,$7F
	.dc.b	$D6,$D3,$E7,$C3,$31,$88,$D5,$70,$E8,$FF,$60,$5B,$F4,$5A,$79,$2B
	.dc.b	$FD,$95,$9F,$6F,$D2,$7F,$D9,$18,$59,$FF,$73,$6D,$FC,$E7,$93,$FC
	.dc.b	$CD,$FA,$57,$1D,$FF,$2B,$9A,$31,$87,$A1,$FF,$5B,$DF,$F4,$1F,$81
	.dc.b	$FA,$75,$BC,$7A,$CF,$FC,$00,$9F,$F3,$97,$CD,$FF,$B8,$B8,$3E,$0B
	.dc.b	$FE,$E4,$C7,$E1,$FE,$01,$FE,$E8,$C3,$D3,$8F,$FA,$71,$F3,$F1,$FF
	.dc.b	$62,$FB,$6F,$93,$D5,$E2,$8F,$EC,$BF,$4D,$FE,$9D,$AF,$A4,$7E,$B8
	.dc.b	$FF,$7F,$5B,$B3,$FC,$F7,$CD,$FD,$7A,$7F,$DF,$F3,$FE,$71,$7E,$29
	.dc.b	$FE,$F3,$D7,$FD,$95,$19,$8F,$EC,$06,$4F,$8F,$E1,$7D,$16,$9F,$EE
	.dc.b	$EA,$AF,$DF,$3C,$AF,$C8,$8F,$F6,$CF,$17,$66,$CF,$D2,$F9,$BE,$98
	.dc.b	$7E,$21,$F4,$C7,$B4,$FF,$59,$46,$5F,$C9,$3F,$E9,$8D,$FD,$3C,$5F
	.dc.b	$E8,$FF,$1F,$F9,$CF,$8B,$FE,$90,$F8,$CF,$C9,$CE,$72,$FE,$73,$E4
	.dc.b	$7F,$40,$B6,$7A,$CF,$FA,$3C,$72,$1F,$86,$E6,$3F,$B7,$7B,$10,$7C
	.dc.b	$76,$5C,$1F,$E8,$CF,$BC,$C3,$8F,$C0,$39,$7C,$A3,$E5,$5F,$2C,$8C
	.dc.b	$37,$E6,$37,$3A,$9C,$8F,$98,$CD,$71,$9D,$7E,$79,$FE,$A7,$D3,$FC
	.dc.b	$F1,$97,$9F,$4C,$47,$F9,$CA,$3D,$5A,$62,$3D,$31,$FE,$A4,$D3,$1E
	.dc.b	$3F,$F0,$D9,$FD,$8F,$C7,$FD,$5E,$AC,$47,$F1,$88,$FB,$13,$8E,$8F
	.dc.b	$F5,$C7,$4F,$F8,$F6,$B9,$78,$FF,$EB,$B9,$C4,$7F,$14,$BF,$F7,$83
	.dc.b	$FE,$3D,$3C,$11,$B4,$64,$FE,$41,$3F,$EC,$5F,$AE,$FC,$5D,$1F,$CC
	.dc.b	$56,$6B,$F1,$27,$F5,$3B,$C3,$F9,$97,$F1,$8F,$AF,$CD,$E3,$FA,$57
	.dc.b	$F5,$F8,$FC,$F1,$FC,$A9,$C6,$7E,$F6,$C7,$F3,$3F,$97,$FE,$C7,$18
	.dc.b	$8F,$F2,$FC,$7F,$3E,$F9,$07,$D8,$F9,$BE,$4F,$55,$FE,$64,$CF,$FC
	.dc.b	$A8,$FF,$00,$31,$F1,$DC,$1C,$1F,$A3,$FD,$27,$8B,$F1,$B1,$FE,$E7
	.dc.b	$18,$84,$7F,$4E,$8C,$45,$F2,$3F,$67,$C5,$FD,$83,$5F,$B4,$FC,$B9
	.dc.b	$F9,$27,$ED,$1C,$3F,$24,$FE,$FD,$8F,$F9,$D7,$83,$E8,$CE,$AB,$F0
	.dc.b	$FC,$1F,$E7,$9A,$FC,$0D,$FB,$EB,$F0,$27,$E3,$38,$B5,$1F,$80,$8E
	.dc.b	$27,$E3,$9F,$C2,$B5,$FC,$1B,$F9,$57,$27,$63,$F1,$22,$3F,$14,$FE
	.dc.b	$83,$91,$CB,$87,$F4,$7C,$B5,$FD,$EF,$37,$FB,$5A,$BF,$27,$D7,$FB
	.dc.b	$4E,$9F,$D8,$FA,$BF,$EA,$49,$CC,$7E,$E4,$FD,$D7,$67,$18,$FD,$C7
	.dc.b	$64,$B7,$6E,$72,$FD,$9E,$95,$FE,$15,$97,$5C,$3D,$19,$F3,$F1,$5E
	.dc.b	$07,$F1,$9F,$17,$0F,$F1,$BF,$06,$F8,$9F,$C8,$AB,$7E,$1B,$2F,$E4
	.dc.b	$79,$7D,$EE,$5F,$E8,$D1,$88,$FB,$C7,$F8,$58,$FF,$0B,$B9,$E7,$FB
	.dc.b	$AB,$0F,$E7,$F5,$D5,$FE,$83,$CF,$6F,$FA,$53,$5F,$D6,$DF,$F5,$AA
	.dc.b	$C3,$D3,$DE,$FE,$EB,$3F,$D5,$67,$D1,$18,$8F,$EC,$95,$FE,$4C,$7F
	.dc.b	$6E,$FF,$87,$B0,$FE,$CD,$1F,$DA,$A3,$FD,$A3,$E3,$FF,$93,$E3,$FD
	.dc.b	$81,$D7,$FD,$CE,$BF,$BB,$46,$0F,$C4,$F9,$5F,$1F,$CD,$FE,$F9,$1F
	.dc.b	$E6,$EF,$3B,$FC,$05,$47,$FC,$04,$FF,$82,$63,$FD,$99,$E5,$7F,$82
	.dc.b	$74,$FA,$1E,$87,$F8,$42,$3F,$C1,$9D,$D8,$AF,$DE,$8D,$3F,$CC,$15
	.dc.b	$FE,$18,$C1,$E9,$7F,$88,$9B,$5F,$E1,$6C,$3F,$DE,$BE,$97,$FA,$0B
	.dc.b	$D3,$FE,$47,$5F,$E2,$DA,$EA,$7F,$A1,$BC,$2C,$F1,$FF,$8D,$4F,$F3
	.dc.b	$71,$75,$7A,$DF,$E4,$0A,$F9,$3C,$78,$7F,$C5,$71,$97,$F5,$FB,$7F
	.dc.b	$90,$5E,$FB,$FF,$93,$1F,$E5,$67,$ED,$16,$BF,$E4,$68,$FF,$05,$9F
	.dc.b	$D0,$3C,$8C,$FA,$1F,$D0,$67,$E5,$9F,$EF,$97,$E0,$CE,$21,$BF,$F9
	.dc.b	$03,$87,$FC,$85,$3F,$E3,$F3,$E3,$4B,$D4,$FF,$22,$FA,$7F,$E3,$19
	.dc.b	$FF,$14,$B5,$FF,$1C,$FA,$7F,$E1,$67,$FC,$2C,$7F,$91,$BA,$7F,$24
	.dc.b	$FA,$AF,$FC,$C7,$D5,$FC,$41,$FE,$77,$1F,$C4,$5F,$59,$FE,$15,$F7
	.dc.b	$1E,$A7,$F8,$56,$BF,$C1,$3E,$1F,$F8,$33,$6F,$E0,$1E,$37,$F0,$FF
	.dc.b	$1F,$FC,$01,$18,$F2,$7F,$87,$47,$F8,$2F,$E7,$7F,$BE,$5F,$F1,$DF
	.dc.b	$53,$F9,$0E,$BF,$17,$57,$85,$FD,$B6,$3F,$81,$47,$C9,$9F,$E7,$11
	.dc.b	$FE,$33,$E9,$F6,$7A,$7F,$D2,$FD,$8F,$AF,$E8,$5C,$3F,$A0,$5F,$F4
	.dc.b	$6B,$F8,$5D,$FF,$DF,$3C,$0F,$03,$E2,$AF,$AD,$9F,$F2,$DD,$3F,$C5
	.dc.b	$35,$EB,$DB,$F9,$26,$FF,$C9,$F8,$7E,$00,$E0,$FE,$49,$7C,$31,$F9
	.dc.b	$5C,$7F,$27,$1F,$E2,$5A,$FF,$82,$27,$E8,$F6,$7D,$59,$FF,$30,$47
	.dc.b	$D7,$9F,$AC,$FF,$09,$F0,$3F,$CD,$51,$FE,$10,$F0,$3F,$83,$4F,$FB
	.dc.b	$2A,$3F,$3B,$5F,$C1,$A3,$11,$FE,$96,$8F,$E0,$11,$F9,$6E,$BF,$DC
	.dc.b	$CF,$EF,$3B,$FF,$C3,$75,$F9,$8A,$FC,$DF,$77,$FB,$77,$8B,$F8,$24
	.dc.b	$7F,$B6,$27,$F9,$2D,$FF,$A0,$BB,$FF,$D9,$AF,$FA,$97,$4F,$D2,$DF
	.dc.b	$0B,$FD,$5F,$B7,$FE,$A5,$AF,$D2,$7D,$07,$EA,$BC,$1E,$15,$FA,$39
	.dc.b	$FD,$47,$E2,$DF,$AE,$F1,$BF,$D6,$1D,$7F,$B0,$F0,$FF,$45,$5F,$EA
	.dc.b	$47,$B5,$FE,$D3,$C3,$FD,$27,$67,$E3,$8F,$F6,$1F,$87,$FA,$49,$FC
	.dc.b	$7F,$83,$F4,$1F,$DB,$3F,$5B,$0F,$68,$E1,$FD,$FF,$85,$F5,$3B,$FF
	.dc.b	$83,$5F,$28,$F0,$BF,$E8,$5A,$FE,$AB,$8F,$EC,$70,$FE,$CD,$E5,$7F
	.dc.b	$64,$BF,$EC,$1E,$63,$3F,$D6,$F6,$FF,$9F,$3E,$3B,$FC,$D7,$E0,$7F
	.dc.b	$9B,$BF,$3B,$FD,$4C,$FF,$9F,$27,$FC,$CD,$D5,$F9,$8F,$1B,$F3,$5B
	.dc.b	$FF,$12,$E9,$FF,$2F,$F0,$7D,$C8,$FE,$2B,$C3,$FA,$DB,$E1,$FA,$47
	.dc.b	$A3,$07,$D1,$7A,$BC,$86,$5E,$CF,$6D,$F9,$2F,$BB,$DB,$BA,$FF,$18
	.dc.b	$D7,$F6,$37,$F0,$BC,$0E,$73,$99,$43,$4D,$78,$BF,$E2,$2D,$BF,$63
	.dc.b	$3F,$EA,$F6,$DF,$DC,$C3,$FC,$4B,$18,$8F,$F4,$47,$0F,$C7,$8F,$DB
	.dc.b	$71,$FC,$77,$FC,$29,$1F,$E8,$AE,$2F,$F0,$7F,$99,$FC,$73,$A3,$FC
	.dc.b	$2B,$CD,$FE,$0F,$F4,$BF,$D7,$FD,$7F,$E3,$43,$F6,$9D,$5F,$73,$2F
	.dc.b	$C2,$70,$FE,$E0,$FE,$2E,$FE,$1F,$C8,$FE,$65,$DB,$FD,$22,$3F,$19
	.dc.b	$3F,$D1,$1B,$BF,$F3,$1E,$AF,$DE,$3F,$ED,$D3,$FC,$3B,$DF,$DC,$FC
	.dc.b	$EF,$27,$F8,$8F,$C1,$FD,$0E,$1F,$DC,$EB,$F4,$AD,$CB,$FD,$36,$FF
	.dc.b	$8A,$7D,$0F,$D8,$78,$BE,$3E,$7F,$95,$7F,$CC,$3F,$02,$7C,$64,$20
	.dc.b	$F8,$FF,$E6,$B2,$9C,$CE,$A3,$93,$81,$F1,$DC,$3D,$87,$CE,$42,$0F
	.dc.b	$A3,$D3,$83,$7F,$49,$87,$1F,$A1,$3F,$3F,$FF,$27,$5E,$E5,$F9,$37
	.dc.b	$F7,$69,$CC,$73,$FC,$2B,$FD,$6B,$18,$8F,$0C,$3A,$DE,$6F,$EF,$76
	.dc.b	$FF,$5D,$E3,$E8,$D1,$F1,$1F,$1B,$87,$FD,$F3,$7B,$B5,$EE,$B5,$E6
	.dc.b	$75,$AF,$F2,$D5,$74,$CF,$D1,$8F,$83,$B3,$FD,$03,$8F,$AE,$5C,$1E
	.dc.b	$4E,$4C,$C7,$BF,$E5,$B1,$8F,$38,$F8,$59,$7F,$B4,$70,$E5,$96,$AB
	.dc.b	$CB,$5F,$1F,$8B,$15,$DF,$2F,$E6,$4F,$CB,$FC,$67,$F0,$18,$FF,$86
	.dc.b	$23,$1C,$BA,$C3,$D3,$39,$7E,$76,$39,$7E,$6F,$0F,$F3,$9B,$FA,$DF
	.dc.b	$0F,$FE,$E6,$BF,$F2,$84,$62,$75,$5D,$3D,$5F,$7B,$F8,$6F,$B6,$F1
	.dc.b	$C6,$23,$F1,$51,$F9,$8B,$7C,$17,$0E,$0F,$F7,$96,$1F,$72,$31,$C3
	.dc.b	$FE,$AE,$F0,$31,$AF,$C9,$CB,$F2,$7C,$58,$8F,$8F,$1F,$95,$E5,$FD
	.dc.b	$27,$56,$77,$FC,$D5,$FE,$13,$87,$F0,$9B,$7F,$9D,$27,$F3,$DF,$9F
	.dc.b	$3E,$43,$F5,$9F,$B1,$FE,$3B,$C4,$F6,$CE,$70,$FD,$8F,$ED,$D9,$DB
	.dc.b	$C9,$B7,$6C,$FA,$3F,$A4,$F1,$3F,$C8,$33,$F8,$1E,$9F,$F8,$03,$F0
	.dc.b	$FF,$D7,$23,$F1,$DC,$DF,$83,$E0,$C4,$7E,$0A,$3F,$0D,$86,$38,$3F
	.dc.b	$0F,$97,$F5,$1A,$FC,$B5,$FB,$7B,$9D,$DF,$9B,$F2,$7F,$27,$C1,$E9
	.dc.b	$E2,$61,$93,$D3,$9F,$EA,$72,$FE,$D5,$9F,$A6,$7F,$DB,$D3,$FC,$06
	.dc.b	$72,$FE,$83,$7E,$7E,$EF,$F0,$29,$FF,$21,$D7,$4A,$6D,$FB,$8F,$27
	.dc.b	$F0,$91,$FC,$37,$87,$F7,$7E,$47,$FB,$B3,$E5,$FF,$D4,$1E,$54,$3F
	.dc.b	$D4,$5F,$6D,$AC,$BD,$33,$99,$D6,$7F,$ED,$B8,$C4,$7F,$AA,$7E,$DB
	.dc.b	$85,$3C,$37,$E9,$2F,$CA,$1F,$86,$61,$B1,$C2,$71,$33,$F7,$5C,$4F
	.dc.b	$18,$CF,$EE,$7D,$BF,$3F,$FF,$32,$6B,$FE,$ED,$FC,$AF,$F9,$9E,$7F
	.dc.b	$43,$C8,$F1,$FE,$7A,$FB,$27,$D5,$3E,$A3,$0D,$8D,$9A,$8C,$8F,$C6
	.dc.b	$B3,$3B,$39,$7E,$39,$89,$DF,$E3,$9F,$84,$CB,$E7,$72,$FC,$6E,$7C
	.dc.b	$47,$CC,$8F,$92,$61,$73,$E3,$9F,$1D,$FF,$C8,$7B,$7F,$D0,$1C,$7F
	.dc.b	$E8,$3F,$0F,$FD,$F9,$83,$8F,$E0,$BF,$CE,$63,$FC,$14,$FE,$17,$C5
	.dc.b	$FC,$BF,$C2,$7E,$13,$BF,$FA,$5D,$6A,$BF,$91,$F7,$BC,$2F,$C7,$D7
	.dc.b	$F5,$6A,$D6,$9F,$97,$F0,$BF,$A1,$46,$27,$F2,$D8,$7E,$6B,$0F,$D2
	.dc.b	$FC,$4F,$79,$FE,$B9,$5F,$E2,$0A,$FE,$17,$1F,$D8,$29,$F1,$3F,$E2
	.dc.b	$C9,$F4,$35,$DC,$FE,$6E,$31,$8F,$F5,$FF,$4F,$FB,$BC,$62,$7F,$B4
	.dc.b	$43,$EA,$FA,$FC,$3F,$C1,$8F,$EF,$5F,$15,$FE,$36,$C7,$AD,$FF,$5D
	.dc.b	$8C,$63,$FD,$0F,$E0,$7A,$4F,$F7,$3E,$4F,$F3,$06,$3E,$D8,$B8,$7C
	.dc.b	$FA,$DC,$7B,$5F,$DB,$C7,$68,$FD,$BC,$62,$FF,$AF,$47,$F9,$AF,$C9
	.dc.b	$F9,$1C,$76,$FF,$7D,$ED,$57,$96,$F7,$0F,$EC,$DC,$D7,$1F,$F0,$17
	.dc.b	$EC,$DF,$ED,$A9,$CF,$97,$DE,$72,$D7,$FC,$16,$D7,$EA,$35,$FE,$7A
	.dc.b	$D7,$FC,$33,$B7,$57,$FB,$B4,$7C,$07,$F8,$83,$5F,$F0,$43,$5F,$E2
	.dc.b	$ED,$5F,$15,$F4,$1F,$F0,$D6,$38,$F8,$BF,$87,$F3,$1C,$F8,$FF,$C6
	.dc.b	$4D,$F5,$4E,$79,$6F,$27,$E4,$70,$BC,$1F,$A3,$F6,$FF,$25,$BF,$F7
	.dc.b	$F6,$E9,$FE,$12,$7F,$39,$5F,$A3,$AF,$E6,$BC,$7F,$E7,$6E,$CF,$F0
	.dc.b	$8C,$E7,$AF,$FC,$6B,$7C,$3B,$3F,$53,$8E,$F1,$FE,$6D,$75,$FC,$16
	.dc.b	$3F,$E8,$0F,$6F,$FA,$3C,$7E,$EC,$FE,$37,$E5,$65,$FE,$97,$CF,$FE
	.dc.b	$F4,$7F,$59,$CD,$FE,$88,$7F,$87,$F8,$5F,$CD,$FC,$3C,$47,$F0,$F8
	.dc.b	$FE,$75,$E1,$62,$3F,$87,$C7,$F1,$2F,$7D,$D5,$E2,$F2,$FF,$90,$F9
	.dc.b	$7F,$C8,$FC,$95,$6B,$0F,$EB,$B5,$AF,$2B,$F9,$67,$93,$AC,$79,$63
	.dc.b	$FC,$AC,$98,$71,$C1,$87,$1F,$2B,$F9,$74,$FF,$3B,$9F,$82,$EF,$7F
	.dc.b	$2E,$1F,$E8,$73,$E3,$FA,$3F,$D2,$3D,$0F,$E8,$DE,$7F,$87,$0F,$EE
	.dc.b	$33,$9F,$4B,$FA,$8F,$A5,$99,$FE,$AD,$7F,$D3,$49,$97,$F5,$DE,$1F
	.dc.b	$E7,$0D,$7E,$4B,$5F,$EC,$8F,$E8,$FA,$7F,$CA,$AD,$E5,$F8,$FE,$AF
	.dc.b	$E1,$BA,$FF,$95,$F5,$63,$2F,$F3,$07,$5F,$E2,$71,$E3,$7F,$C0,$31
	.dc.b	$FF,$00,$FA,$BF,$E0,$BB,$FF,$3A,$47,$F8,$4D,$FF,$2C,$E1,$FE,$13
	.dc.b	$F3,$B3,$F0,$FC,$4F,$F3,$2E,$9F,$E1,$FB,$E1,$E1,$7E,$E3,$C0,$FF
	.dc.b	$34,$C7,$F8,$AB,$4F,$E7,$39,$3E,$4F,$EF,$FB,$7F,$E5,$0D,$FF,$C5
	.dc.b	$77,$BB,$FF,$22,$BC,$76,$FF,$1C,$79,$0F,$63,$8E,$5F,$E7,$39,$CF
	.dc.b	$B7,$F8,$AF,$7F,$FC,$B7,$C5,$FE,$47,$C3,$FD,$4B,$8F,$F9,$63,$0F
	.dc.b	$F3,$D4,$E7,$DB,$E1,$1F,$D6,$E7,$4F,$F9,$53,$87,$FC,$BB,$18,$E1
	.dc.b	$FF,$56,$47,$E8,$BC,$7F,$F6,$24,$7F,$47,$BB,$B7,$F4,$6C,$AB,$FE
	.dc.b	$AF,$CA,$BF,$EB,$5E,$BF,$F1,$DE,$5F,$E7,$68,$FF,$5A,$47,$D4,$F8
	.dc.b	$7F,$ED,$9F,$07,$A6,$F9,$3F,$EC,$FD,$BD,$11,$FD,$C7,$E0,$FF,$D9
	.dc.b	$58,$FF,$A1,$F1,$FF,$60,$C7,$FB,$42,$3F,$CA,$71,$FE,$BE,$6B,$FD
	.dc.b	$9D,$AF,$F9,$C1,$AF,$F3,$46,$BF,$EC,$E7,$FD,$5F,$5F,$D2,$1B,$AF
	.dc.b	$FA,$91,$FF,$5C,$3F,$EB,$7D,$3F,$DB,$DD,$3F,$90,$F5,$BF,$D5,$FB
	.dc.b	$7F,$98,$7D,$DF,$E0,$FF,$83,$FF,$B2,$73,$FF,$6D,$6F,$FE,$60,$8C
	.dc.b	$47,$FB,$7A,$3F,$D9,$91,$88,$FF,$6A,$F8,$9F,$CB,$75,$FE,$BB,$DF
	.dc.b	$FC,$89,$FE,$BA,$DD,$7F,$D7,$2F,$FB,$B3,$5F,$F7,$7F,$1F,$1A,$FF
	.dc.b	$79,$47,$FB,$BF,$E0,$B1,$F0,$FF,$EF,$18,$FF,$7D,$CF,$3C,$BF,$D9
	.dc.b	$9F,$0D,$E3,$F8,$77,$B9,$FE,$E0,$E2,$F7,$FC,$37,$FA,$6A,$3F,$DA
	.dc.b	$98,$7F,$9B,$F0,$FF,$63,$9F,$EE,$FE,$EF,$F5,$F3,$FE,$FB,$E7,$FF
	.dc.b	$08,$72,$7D,$D8,$FF,$7D,$35,$FF,$80,$39,$7F,$30,$8F,$C6,$40,$83
	.dc.b	$E7,$B5,$F1,$1F,$E0,$F9,$DC,$FF,$B0,$3D,$9F,$F5,$93,$FE,$E7,$8F
	.dc.b	$D1,$3F,$6F,$7F,$F6,$27,$89,$FE,$AE,$F8,$84,$4F,$D8,$3F,$D5,$DB
	.dc.b	$7F,$85,$6E,$F3,$FF,$A2,$79,$FF,$AA,$FB,$3F,$EA,$59,$FF,$47,$F9
	.dc.b	$3F,$EF,$6F,$13,$11,$FE,$F0,$F2,$BE,$4E,$7F,$EE,$8E,$C6,$FF,$C7
	.dc.b	$36,$F0,$47,$C1,$F2,$FD,$87,$F2,$7D,$BF,$D4,$78,$FF,$A6,$DF,$F4
	.dc.b	$F7,$37,$FA,$3B,$E1,$BF,$CD,$3C,$9F,$E8,$6F,$43,$FD,$16,$7F,$A3
	.dc.b	$BE,$97,$FD,$27,$C7,$F2,$B1,$98,$7F,$9B,$F0,$E3,$E2,$7E,$0E,$74
	.dc.b	$27,$FC,$CF,$97,$DF,$9F,$EF,$8F,$FA,$DB,$87,$FC,$BA,$FE,$0F,$E1
	.dc.b	$7F,$C3,$47,$F8,$57,$F5,$B7,$C3,$7F,$86,$F1,$EA,$7F,$91,$AB,$FC
	.dc.b	$85,$8F,$F8,$DE,$3C,$2F,$F8,$AD,$BC,$1F,$E2,$78,$FF,$11,$4F,$E4
	.dc.b	$7E,$07,$B1,$FE,$29,$E6,$FF,$0C,$DF,$67,$CF,$CF,$FC,$FF,$E0,$3F
	.dc.b	$CA,$D1,$EF,$9F,$86,$FE,$BF,$7F,$97,$E5,$7E,$D6,$7E,$13,$FD,$DF
	.dc.b	$CD,$FE,$D5,$F1,$4E,$4F,$ED,$57,$C3,$C3,$FF,$3F,$6D,$FD,$77,$6F
	.dc.b	$F4,$5F,$6F,$F8,$8B,$2F,$F6,$DE,$7F,$09,$FC,$17,$63,$BB,$FB,$EC
	.dc.b	$7F,$56,$97,$99,$FC,$3F,$8F,$FC,$23,$DF,$FE,$6F,$AD,$74,$7F,$87
	.dc.b	$BA,$BF,$DC,$31,$FD,$6F,$D3,$FE,$19,$CD,$FD,$4B,$6F,$CE,$4F,$F5
	.dc.b	$7E,$FF,$D0,$ED,$E1,$8F,$D3,$EB,$E0,$F1,$B3,$3F,$D4,$2F,$8F,$7F
	.dc.b	$FA,$27,$C3,$FE,$8B,$7F,$8D,$C3,$F6,$6F,$EE,$7C,$4F,$F1,$5C,$7E
	.dc.b	$FE,$3F,$8F,$B5,$FE,$69,$E7,$7F,$43,$AF,$F1,$6F,$5F,$CA,$AF,$E6
	.dc.b	$B5,$F8,$83,$FC,$63,$E4,$7E,$B3,$CC,$FF,$7F,$79,$BF,$C6,$7A,$BF
	.dc.b	$91,$71,$FF,$A2,$B6,$FE,$55,$FA,$EF,$1E,$8F,$E5,$90,$8E,$9C,$B8
	.dc.b	$E4,$FE,$4D,$95,$E3,$FE,$55,$7F,$CC,$E7,$F9,$57,$4B,$CC,$FF,$1D
	.dc.b	$47,$C6,$AF,$E1,$7F,$90,$2F,$5D,$7F,$E5,$CA,$D5,$3F,$77,$FF,$27
	.dc.b	$E1,$CA,$39,$60,$F4,$BF,$99,$63,$CF,$0F,$E2,$F9,$BF,$5B,$FE,$4D
	.dc.b	$BB,$8F,$F9,$51,$DF,$25,$6F,$93,$F1,$7D,$FD,$7C,$DF,$EC,$79,$7C
	.dc.b	$06,$DF,$F4,$97,$FC,$D3,$1F,$D2,$7A,$7F,$8C,$F4,$F5,$AF,$5D,$69
	.dc.b	$F6,$E3,$AD,$7D,$11,$F0,$38,$F1,$1F,$E2,$68,$FC,$6F,$A5,$FC,$93
	.dc.b	$AF,$A4,$7E,$53,$F7,$8F,$D0,$E2,$9C,$CF,$F3,$89,$73,$F1,$C3,$BC
	.dc.b	$FF,$64,$BE,$11,$88,$FC,$FF,$27,$C1,$C3,$9C,$23,$EC,$53,$BB,$3F
	.dc.b	$86,$7C,$77,$75,$F3,$27,$58,$75,$8F,$E9,$7F,$AD,$7A,$2B,$AB,$E6
	.dc.b	$AF,$F1,$BF,$0F,$EC,$38,$7F,$C0,$1C,$FF,$D1,$3D,$8F,$99,$1F,$E0
	.dc.b	$3F,$63,$19,$F5,$CF,$F0,$38,$F2,$7F,$07,$D3,$FD,$9B,$83,$75,$FE
	.dc.b	$52,$E4,$FC,$07,$3B,$E3,$BF,$AE,$ED,$FC,$03,$D0,$FE,$89,$F0,$38
	.dc.b	$CF,$FB,$1C,$7F,$47,$F8,$EC,$47,$CB,$F5,$FF,$01,$D3,$FE,$15,$F7
	.dc.b	$B2,$FF,$85,$FE,$4B,$FA,$EF,$9B,$F9,$1E,$4F,$F0,$E6,$9F,$1E,$31
	.dc.b	$DF,$F9,$5E,$4F,$F3,$D4,$FE,$46,$7F,$31,$C3,$F9,$7F,$05,$7F,$97
	.dc.b	$F6,$7F,$39,$C1,$FE,$04,$E0,$FD,$77,$4F,$F6,$C9,$FA,$B8,$5E,$5F
	.dc.b	$C8,$EF,$FD,$E3,$8B,$EC,$57,$E9,$63,$07,$F6,$D9,$43,$E5,$9F,$42
	.dc.b	$F9,$77,$CE,$6A,$5F,$93,$F1,$BE,$17,$BA,$D3,$9A,$E7,$7C,$AB,$B7
	.dc.b	$E6,$BF,$E0,$4F,$2F,$FD,$C3,$FB,$33,$F6,$87,$E2,$A8,$C3,$7F,$79
	.dc.b	$E1,$B7,$F0,$F9,$C9,$2B,$31,$C5,$E7,$1A,$FD,$E3,$2E,$EF,$B7,$9F
	.dc.b	$C3,$F3,$BF,$E4,$F8,$7F,$1B,$F1,$B9,$FD,$63,$E9,$1F,$58,$8C,$36
	.dc.b	$36,$6A,$32,$3F,$1D,$CC,$EC,$E5,$F9,$1B,$E2,$7F,$91,$63,$FC,$AF
	.dc.b	$2F,$1A,$CB,$47,$F2,$F5,$1A,$8F,$F0,$CC,$7F,$0F,$EA,$FE,$BF,$D7
	.dc.b	$8F,$0B,$FC,$C9,$D9,$64,$FE,$1D,$C3,$F8,$99,$B5,$FC,$F7,$DF,$FF
	.dc.b	$24,$CF,$F8,$8A,$3B,$57,$F3,$6C,$9F,$43,$FC,$22,$9F,$14,$F9,$EF
	.dc.b	$CB,$B7,$0F,$DF,$61,$E2,$8C,$47,$4F,$4B,$FC,$07,$18,$D3,$FA,$5C
	.dc.b	$7F,$99,$E3,$15,$FD,$4E,$3F,$92,$FC,$5B,$E8,$3F,$4A,$DC,$3B,$E5
	.dc.b	$FE,$33,$C7,$F9,$F7,$B9,$CB,$1F,$EA,$ED,$7F,$C0,$7B,$ED,$FF,$47
	.dc.b	$47,$FA,$22,$FD,$3E,$67,$F5,$3F,$80,$FE,$4F,$56,$7D,$B7,$5F,$8F
	.dc.b	$1F,$E2,$1B,$E5,$18,$C7,$58,$FF,$A6,$7E,$4E,$DF,$F6,$AF,$93,$78
	.dc.b	$3C,$2B,$4D,$9D,$35,$FE,$ED,$EF,$70,$D3,$EA,$4E,$67,$DA,$D7,$F9
	.dc.b	$B3,$5F,$F1,$3B,$F0,$DF,$F1,$4E,$3E,$2E,$2F,$E2,$33,$99,$FF,$46
	.dc.b	$ED,$9D,$3E,$EC,$7F,$96,$9F,$77,$D0,$FD,$37,$E1,$EE,$F8,$FE,$5C
	.dc.b	$7F,$E4,$26,$FB,$39,$CF,$FA,$6F,$B3,$70,$FB,$75,$E1,$CB,$FC,$03
	.dc.b	$F0,$3F,$92,$9F,$BE,$D7,$BA,$F8,$FF,$E4,$91,$F9,$D9,$E3,$18,$87
	.dc.b	$5F,$E8,$2B,$84,$7F,$2E,$38,$52,$33,$1F,$6D,$BC,$7F,$E2,$E7,$F7
	.dc.b	$51,$FB,$D3,$F8,$7F,$67,$1F,$43,$F7,$9D,$DF,$BB,$F8,$7F,$BF,$5F
	.dc.b	$E2,$DF,$0F,$F8,$5F,$89,$FC,$37,$C4,$56,$AB,$F8,$65,$7F,$10,$F0
	.dc.b	$F3,$D1,$FC,$1E,$38,$F8,$9F,$C5,$EF,$86,$5D,$EB,$F9,$9F,$93,$FC
	.dc.b	$A0,$FE,$53,$E4,$F2,$CB,$F9,$84,$67,$C8,$FE,$53,$87,$87,$C9,$FE
	.dc.b	$58,$3C,$AE,$17,$FC,$CE,$A4,$FF,$3F,$AF,$D0,$F9,$9F,$D2,$BC,$DF
	.dc.b	$E8,$DE,$73,$CE,$F6,$F8,$9F,$C7,$EB,$5E,$87,$F5,$3F,$47,$3A,$7F
	.dc.b	$94,$67,$27,$F5,$4C,$BF,$AE,$C7,$D7,$3E,$DF,$5F,$F9,$1A,$33,$0E
	.dc.b	$BF,$E3,$AF,$E0,$FC,$6F,$F4,$E4,$7F,$99,$A7,$6F,$F8,$23,$C0,$FF
	.dc.b	$7C,$7A,$5F,$DD,$27,$FB,$EE,$1F,$E5,$2F,$77,$FC,$AF,$EA,$FF,$97
	.dc.b	$67,$33,$FE,$69,$F1,$FF,$D3,$B9,$FF,$92,$3C,$9F,$F0,$4C,$62,$3C
	.dc.b	$92,$9F,$F0,$6B,$FE,$9D,$CF,$F6,$73,$FD,$0F,$DA,$C4,$FA,$A3,$FC
	.dc.b	$E9,$ED,$62,$3C,$18,$BD,$BF,$F2,$E7,$B1,$FE,$24,$8C,$5F,$F9,$32
	.dc.b	$3F,$C4,$91,$FE,$31,$AF,$66,$E3,$FC,$A7,$19,$D3,$FC,$63,$EE,$FF
	.dc.b	$AA,$74,$F9,$1E,$F7,$F8,$FB,$7F,$EB,$1E,$36,$72,$FE,$0F,$1F,$D6
	.dc.b	$FC,$7F,$CD,$FB,$FF,$E4,$AF,$72,$EF,$FE,$43,$F4,$3B,$CF,$F8,$83
	.dc.b	$6F,$F2,$9C,$FF,$52,$F8,$0F,$9D,$F0,$5F,$E5,$EF,$82,$B3,$FE,$4B
	.dc.b	$7F,$D2,$D5,$FE,$66,$8F,$E2,$2D,$DF,$FD,$2F,$F0,$7F,$ED,$28,$C6
	.dc.b	$DF,$E9,$F8,$F7,$F9,$FF,$E6,$4F,$5B,$FD,$73,$87,$F4,$CB,$FF,$42
	.dc.b	$5F,$CE,$AF,$F3,$4F,$56,$AB,$FA,$0C,$FE,$E2,$BF,$B3,$1F,$E9,$27
	.dc.b	$EA,$FB,$3F,$E4,$4E,$CF,$C4,$F2,$7F,$B0,$AF,$FD,$1B,$CF,$FE,$47
	.dc.b	$AF,$E8,$7F,$0F,$FE,$AB,$C3,$FD,$B3,$BF,$FA,$AE,$31,$1F,$E8,$2B
	.dc.b	$46,$20,$7F,$AD,$7C,$CF,$F2,$8C,$FB,$79,$FF,$D8,$53,$FE,$BA,$F0
	.dc.b	$33,$3F,$D7,$27,$FD,$9F,$D7,$F8,$32,$1F,$E9,$67,$2E,$0F,$F7,$EC
	.dc.b	$61,$C3,$DA,$BF,$D8,$FF,$01,$E5,$7F,$D9,$1E,$2F,$FB,$12,$7F,$47
	.dc.b	$AF,$C1,$87,$C1,$FC,$9A,$FF,$6A,$3F,$49,$FF,$7D,$F7,$FF,$A9,$EF
	.dc.b	$FC,$2F,$CD,$F9,$9A,$FB,$B5,$FE,$BC,$E8,$FF,$52,$FC,$2F,$F6,$E9
	.dc.b	$FC,$6F,$4F,$FB,$6B,$F0,$AC,$76,$7F,$7E,$9C,$E1,$FE,$11,$DF,$FD
	.dc.b	$27,$7F,$EE,$08,$F9,$B3,$FA,$5E,$7F,$8D,$DD,$FE,$8E,$DB,$E3,$F8
	.dc.b	$3F,$E9,$0B,$F4,$F8,$5F,$E9,$0E,$0F,$F1,$F7,$AF,$FE,$8E,$7F,$83
	.dc.b	$72,$FF,$BB,$7B,$71,$1F,$8C,$D3,$F8,$8F,$2F,$E1,$7A,$3F,$D0,$93
	.dc.b	$FC,$2B,$4F,$AD,$A7,$F8,$2A,$73,$BF,$9F,$5A,$7F,$91,$2F,$FA,$EE
	.dc.b	$3F,$CD,$BC,$BF,$F3,$DF,$95,$FE,$83,$FD,$B7,$FC,$D7,$E8,$FF,$43
	.dc.b	$F3,$DF,$BC,$FF,$97,$BD,$6F,$F3,$0F,$E4,$5F,$E7,$4D,$7E,$27,$07
	.dc.b	$F9,$7F,$5F,$F6,$6B,$FE,$53,$DB,$F2,$64,$3F,$CA,$31,$FC,$E3,$8B
	.dc.b	$FC,$A1,$5E,$7F,$84,$FF,$00,$57,$F9,$3B,$DE,$FF,$68,$FC,$73,$E1
	.dc.b	$3F,$C1,$D7,$ED,$F1,$FF,$C0,$D9,$FF,$8B,$30,$EC,$1F,$F1,$36,$7F
	.dc.b	$E5,$47,$FC,$34,$FF,$8E,$1F,$F0,$D5,$7C,$7F,$80,$FF,$0B,$E3,$FE
	.dc.b	$13,$AF,$DE,$F9,$7F,$E0,$FA,$F9,$8F,$F8,$0B,$0F,$99,$3E,$B8,$C4
	.dc.b	$78,$8F,$F0,$54,$7C,$4F,$2F,$F4,$51,$E5,$7F,$BF,$70,$7F,$7A,$1C
	.dc.b	$1F,$DB,$B8,$BF,$B8,$F1,$7F,$64,$C3,$FD,$67,$EB,$FF,$6C,$E5,$FE
	.dc.b	$B7,$97,$8B,$B7,$E7,$E7,$FD,$72,$3F,$30,$DE,$0F,$C3,$F5,$7F,$94
	.dc.b	$23,$F1,$BE,$07,$F4,$1F,$01,$C9,$9F,$89,$FF,$30,$D7,$F6,$2E,$9F
	.dc.b	$F7,$19,$FD,$47,$CA,$FF,$4C,$6F,$FE,$2C,$9F,$E9,$EF,$F2,$46,$F2
	.dc.b	$FF,$BD,$FC,$3F,$CB,$C7,$EA,$5F,$03,$FD,$4F,$6F,$B2,$FE,$C9,$FC
	.dc.b	$47,$A1,$FD,$56,$31,$1A,$F1,$7F,$D5,$17,$FB,$BD,$FF,$C5,$3E,$47
	.dc.b	$F0,$06,$E5,$FE,$58,$C3,$F8,$2E,$1F,$94,$F2,$7A,$75,$7D,$59,$F7
	.dc.b	$F0,$7F,$37,$D5,$C3,$FE,$83,$F3,$3F,$8D,$79,$4F,$2F,$F9,$4E,$9F
	.dc.b	$D0,$BC,$9F,$E4,$58,$7F,$28,$8F,$8B,$FB,$47,$FB,$03,$CF,$F8,$F0
	.dc.b	$8F,$F6,$1B,$87,$FD,$81,$E7,$7F,$90,$76,$FF,$6C,$FD,$EA,$93,$97
	.dc.b	$9C,$2F,$3C,$5F,$EC,$BB,$E3,$E9,$FF,$2F,$71,$7C,$DE,$1E,$67,$E0
	.dc.b	$23,$04,$3F,$CC,$53,$FC,$D3,$8B,$8C,$7A,$B0,$C3,$B3,$F0,$73,$FC
	.dc.b	$EA,$FF,$DB,$95,$FC,$57,$08,$66,$B8,$1F,$E0,$AC,$3F,$4E,$C3,$F1
	.dc.b	$BD,$DB,$70,$7F,$83,$63,$FC,$01,$5D,$29,$FA,$57,$B3,$9F,$C3,$1E
	.dc.b	$08,$FF,$78,$57,$F8,$1F,$1F,$A8,$DC,$3E,$5E,$7F,$D3,$67,$E4,$CF
	.dc.b	$F8,$53,$1E,$DF,$9E,$7C,$8E,$AE,$B9,$7B,$63,$AC,$7C,$6C,$7F,$4B
	.dc.b	$8F,$F5,$C8,$C4,$7D,$BF,$D1,$79,$3F,$08,$3E,$6F,$5F,$AE,$1F,$CB
	.dc.b	$67,$07,$F8,$81,$FE,$81,$F1,$BF,$26,$7A,$CF,$E1,$F0,$FE,$D7,$1E
	.dc.b	$3D,$3E,$43,$67,$5A,$7E,$E3,$4F,$E9,$5C,$DF,$E0,$4F,$6B,$FB,$9F
	.dc.b	$AB,$EE,$8F,$F0,$3F,$16,$30,$F4,$1F,$8A,$9F,$C2,$E7,$C7,$1F,$E0
	.dc.b	$79,$FF,$91,$BA,$3F,$35,$6F,$6B,$FA,$FE,$9E,$A8,$FC,$A6,$1F,$D8
	.dc.b	$AF,$EA,$F8,$9F,$90,$7F,$39,$19,$9F,$CB,$47,$DC,$F7,$3E,$FE,$BF
	.dc.b	$E2,$0D,$7D,$B1,$F1,$2B,$F3,$EF,$E8,$AF,$51,$83,$F4,$2E,$39,$FF
	.dc.b	$C2,$77,$C5,$B3,$FB,$3F,$C6,$DF,$8E,$BF,$1C,$7E,$20,$FC,$71,$5A
	.dc.b	$FC,$CB,$8E,$18,$C3,$85,$EE,$B9,$E1,$A6,$C6,$65,$FA,$DF,$2F,$D7
	.dc.b	$5F,$87,$C2,$0F,$C4,$F4,$E0,$D7,$3F,$F1,$9C,$FF,$86,$56,$67,$9F
	.dc.b	$EB,$BF,$E3,$2E,$DF,$AF,$FA,$0F,$E9,$9B,$7F,$ED,$EC,$3F,$DA,$DF
	.dc.b	$9D,$BE,$CD,$F6,$0F,$B2,$C3,$63,$A5,$FF,$1B,$AC,$D4,$FC,$5B,$32
	.dc.b	$9C,$D7,$4A,$7C,$F6,$B6,$F4,$6D,$EF,$CF,$FD,$D1,$18,$8D,$1E,$DD
	.dc.b	$BF,$55,$93,$6F,$E6,$6F,$F2,$DE,$77,$89,$FC,$AE,$3F,$98,$4E,$94
	.dc.b	$E8,$F8,$6F,$F3,$C7,$8F,$CC,$FE,$3F,$AE,$CF,$E9,$A6,$BF,$E0,$48
	.dc.b	$FC,$89,$8E,$EF,$DC,$78,$5F,$CF,$FB,$FC,$99,$7C,$0B,$FD,$7D,$F9
	.dc.b	$B7,$FC,$A7,$C2,$BE,$17,$C0,$DA,$F0,$35,$FE,$52,$7F,$0E,$3F,$C5
	.dc.b	$C7,$F1,$1F,$23,$FA,$54,$6F,$3F,$8B,$1F,$03,$E1,$5F,$3D,$ED,$6D
	.dc.b	$FB,$7D,$6E,$0F,$CC,$DF,$FC,$17,$E2,$7E,$C7,$7B,$3B,$8F,$DA,$79
	.dc.b	$58,$8F,$29,$FC,$17,$B7,$F2,$CE,$23,$FC,$76,$EB,$C9,$BA,$FD,$5C
	.dc.b	$2C,$F8,$6F,$C6,$FD,$5F,$9C,$F3,$BD,$B2,$46,$2F,$FB,$B7,$C3,$F0
	.dc.b	$AF,$E2,$FB,$FF,$6D,$8C,$47,$F3,$9B,$FF,$23,$57,$C6,$8A,$7C,$19
	.dc.b	$DB,$FC,$CE,$3E,$D3,$E1,$69,$FE,$05,$1C,$DF,$C5,$79,$BF,$05,$3F
	.dc.b	$DF,$67,$2F,$F1,$68,$CD,$A7,$33,$D6,$BF,$25,$1F,$98,$8F,$CD,$9F
	.dc.b	$55,$FC,$F3,$FA,$33,$07,$E7,$0F,$EF,$7E,$5F,$EB,$3A,$BF,$54,$D7
	.dc.b	$F8,$67,$84,$8C,$47,$F0,$F7,$F8,$2F,$7F,$EE,$4F,$DD,$74,$6F,$7F
	.dc.b	$E8,$7E,$07,$EF,$BC,$2F,$E1,$1E,$16,$73,$FE,$0D,$87,$19,$FE,$15
	.dc.b	$8B,$0F,$35,$7F,$4C,$EC,$E0,$FA,$CF,$DA,$4F,$E2,$BC,$0F,$E0,$FE
	.dc.b	$07,$F9,$AD,$FE,$13,$B5,$DB,$F8,$5D,$7E,$43,$C6,$FE,$63,$E4,$FF
	.dc.b	$37,$F1,$7F,$CE,$51,$83,$F9,$E4,$EF,$C6,$FE,$77,$5F,$CF,$7C,$8F
	.dc.b	$F3,$C7,$27,$F0,$9E,$6F,$8D,$E5,$7D,$DD,$DD,$9F,$DD,$F2,$4E,$63
	.dc.b	$F8,$A4,$62,$7F,$80,$B5,$F9,$5D,$FF,$C2,$A1,$E5,$FF,$67,$9F,$ED
	.dc.b	$71,$87,$FB,$39,$FE,$41,$C3,$EC,$4F,$F1,$79,$D4,$FD,$5E,$4F,$F0
	.dc.b	$86,$5F,$E4,$D9,$FE,$F7,$3F,$C5,$8F,$F2,$A7,$81,$F8,$1F,$33,$FC
	.dc.b	$03,$DB,$FD,$FB,$C1,$C5,$7A,$27,$FC,$15,$3F,$E0,$2A,$CC,$F7,$D3
	.dc.b	$FC,$1B,$83,$CD,$FF,$0B,$4F,$F3,$08,$C3,$FE,$4F,$CB,$F0,$11,$88
	.dc.b	$FE,$13,$D3,$F9,$3A,$FF,$14,$F9,$1F,$E1,$0D,$F1,$5F,$E2,$9D,$BF
	.dc.b	$15,$3C,$BC,$EF,$F1,$4F,$87,$EF,$6D,$FE,$75,$F8,$C7,$F8,$D2,$BF
	.dc.b	$25,$E7,$EF,$0E,$91,$FE,$F3,$8F,$47,$A3,$FE,$34,$F4,$BE,$53,$70
	.dc.b	$FF,$4A,$DF,$A3,$5F,$F2,$DE,$1F,$E5,$7D,$73,$1F,$E5,$DF,$4B,$E1
	.dc.b	$6F,$FC,$AE,$3F,$CC,$EF,$E3,$5B,$7F,$D7,$2B,$F6,$93,$E7,$BF,$D4
	.dc.b	$1F,$B7,$AF,$F0,$E1,$FB,$9A,$FC,$6F,$93,$F8,$97,$F8,$65,$FE,$F6
	.dc.b	$31,$7F,$D8,$BD,$7F,$F4,$95,$7F,$B0,$B4,$FF,$49,$46,$23,$FC,$7F
	.dc.b	$1B,$C7,$E1,$47,$FA,$A3,$D5,$FF,$52,$72,$FF,$A9,$A3,$FA,$43,$5F
	.dc.b	$EE,$BE,$AF,$37,$FD,$45,$D1,$FC,$07,$D2,$FF,$4C,$47,$C0,$3F,$86
	.dc.b	$3F,$E8,$AD,$BF,$D0,$98,$FF,$AD,$E3,$E6,$BF,$E9,$3D,$FF,$D5,$91
	.dc.b	$FE,$81,$AF,$51,$FE,$77,$3E,$DC,$FD,$E7,$07,$F6,$AE,$4E,$3D,$3F
	.dc.b	$9C,$8F,$F0,$55,$7E,$99,$FF,$5A,$67,$FE,$60,$EB,$FF,$2F,$E1,$F9
	.dc.b	$09,$FF,$3D,$57,$EE,$F2,$FE,$39,$D2,$C3,$FC,$29,$7E,$F8,$FB,$4F
	.dc.b	$AD,$FF,$35,$F8,$1F,$C4,$FC,$07,$C7,$7F,$90,$3C,$3F,$F2,$07,$E2
	.dc.b	$9F,$E2,$CF,$D0,$7E,$1E,$3F,$E2,$EF,$33,$F9,$87,$AC,$FC,$77,$FC
	.dc.b	$4F,$87,$0F,$07,$FA,$5F,$5F,$EF,$FB,$BF,$13,$EA,$FF,$5C,$FD,$1B
	.dc.b	$F9,$97,$AB,$FD,$A7,$9B,$FC,$4D,$EA,$D7,$FC,$5B,$EA,$EF,$7F,$EA
	.dc.b	$DE,$AF,$F2,$1C,$BF,$27,$D1,$FD,$23,$6F,$F2,$D7,$1F,$F3,$AE,$2F
	.dc.b	$2F,$0F,$F3,$CE,$3F,$4D,$FF,$3B,$BF,$F2,$C5,$FF,$34,$E5,$FE,$63
	.dc.b	$CD,$FE,$74,$E6,$FE,$A9,$CF,$F2,$0F,$E6,$3C,$1F,$EA,$97,$E7,$3F
	.dc.b	$E3,$CD,$FF,$A6,$65,$FC,$88,$FF,$4A,$47,$F5,$77,$EE,$78,$9F,$C5
	.dc.b	$BB,$3F,$01,$3F,$C5,$7C,$7F,$E1,$2F,$F0,$73,$F1,$36,$F1,$FF,$DF
	.dc.b	$58,$7E,$3E,$7F,$4A,$7E,$FA,$7F,$DA,$CF,$EE,$0E,$2B,$1F,$D9,$47
	.dc.b	$F3,$A3,$FB,$57,$8B,$FA,$57,$FC,$50,$7F,$AB,$8F,$F4,$3C,$FE,$8B
	.dc.b	$C3,$D7,$87,$FA,$3F,$1B,$F7,$D8,$FE,$94,$FC,$6D,$7F,$BB,$DF,$D4
	.dc.b	$FA,$EF,$0B,$F5,$33,$FC,$2B,$F4,$DF,$D9,$79,$3C,$27,$D9,$9F,$ED
	.dc.b	$AB,$F4,$F8,$7F,$BD,$7F,$42,$FD,$C7,$06,$CF,$F6,$BF,$81,$FB,$BF
	.dc.b	$2F,$FD,$BF,$8F,$F1,$6C,$3F,$DB,$38,$FC,$26,$BF,$E1,$F3,$7C,$7F
	.dc.b	$4B,$93,$FD,$A9,$1F,$D6,$67,$91,$8F,$0F,$FD,$A5,$1F,$B6,$F3,$7F
	.dc.b	$DF,$3E,$2F,$FB,$53,$6F,$E0,$D1,$EB,$3E,$2D,$FC,$1F,$3B,$FD,$9B
	.dc.b	$E4,$7E,$D3,$A3,$18,$F9,$75,$FE,$E9,$9F,$E9,$F0,$FD,$C4,$FD,$AE
	.dc.b	$4F,$F5,$E7,$67,$E5,$87,$BD,$FB,$7E,$FF,$F5,$CF,$E8,$BF,$B7,$C3
	.dc.b	$D2,$7F,$19,$3F,$C4,$FE,$C7,$F8,$8F,$CB,$FF,$4D,$69,$FE,$23,$CF
	.dc.b	$B3,$FA,$BC,$3F,$D8,$B8,$7F,$A8,$3C,$3F,$E4,$7B,$7F,$A8,$B8,$BE
	.dc.b	$EF,$1F,$F9,$0A,$72,$FF,$91,$F1,$46,$3C,$7F,$C2,$CF,$F0,$49,$FE
	.dc.b	$3A,$7E,$A6,$FF,$D2,$8F,$D7,$F2,$FF,$D2,$18,$7F,$10,$7F,$84,$4B
	.dc.b	$CB,$FF,$39,$CF,$F1,$4E,$FD,$F0,$7C,$0E,$3F,$F4,$24,$FE,$4A,$3F
	.dc.b	$D0,$6D,$8F,$E4,$67,$F9,$EB,$DB,$FC,$CF,$97,$D8,$E7,$5F,$2B,$9B
	.dc.b	$FC,$E3,$7F,$C8,$9F,$57,$57,$F0,$2F,$2F,$F9,$8C,$7F,$9A,$A7,$50
	.dc.b	$DA,$70,$FC,$AA,$F9,$18,$7D,$6D,$BF,$29,$5F,$7E,$7F,$65,$5F,$43
	.dc.b	$0F,$A9,$C6,$E3,$FF,$2B,$C7,$EB,$8F,$F3,$3F,$B3,$FE,$E7,$3F,$DD
	.dc.b	$5E,$E7,$F2,$1C,$FB,$4F,$F2,$1B,$FF,$7E,$5F,$F2,$1F,$55,$EE,$BD
	.dc.b	$77,$9A,$EE,$FC,$7E,$3F,$2B,$2F,$9C,$FE,$63,$7F,$B7,$C3,$FC,$D3
	.dc.b	$C0,$FE,$07,$3F,$D5,$6B,$F8,$97,$81,$FA,$18,$FF,$23,$F1,$7C,$5C
	.dc.b	$3F,$B8,$57,$F9,$31,$FD,$C4,$FF,$2B,$8F,$EC,$11,$FC,$D2,$3F,$95
	.dc.b	$D7,$F8,$D5,$78,$5E,$03,$F7,$8F,$F8,$F6,$FF,$8E,$3F,$C1,$4F,$F2
	.dc.b	$01,$FD,$13,$C3,$FE,$13,$87,$F4,$CF,$17,$F8,$6B,$FC,$CF,$1F,$E0
	.dc.b	$8F,$18,$FF,$12,$78,$DF,$D0,$3C,$7F,$56,$3F,$CA,$6F,$F8,$D6,$3F
	.dc.b	$C8,$30,$F0,$57,$F8,$AF,$C8,$FE,$8F,$7F,$92,$AF,$F1,$55,$F6,$BF
	.dc.b	$EB,$BD,$DF,$CB,$BA,$9E,$5F,$F3,$27,$FA,$8D,$7E,$37,$D8,$FE,$EA
	.dc.b	$FF,$39,$F8,$21,$08,$3E,$53,$CC,$E0,$CC,$71,$70,$3E,$1B,$D9,$F0
	.dc.b	$42,$10,$7C,$1E,$9C,$1B,$FA,$3C,$38,$F9,$57,$A7,$FC,$47,$1D,$CF
	.dc.b	$5F,$C0,$3F,$51,$39,$8C,$FD,$47,$FA,$E6,$31,$1D,$A5,$E3,$FF,$4F
	.dc.b	$E1,$FC,$64,$F9,$F3,$F8,$3C,$5F,$E4,$19,$CC,$FA,$36,$F8,$FB,$FE
	.dc.b	$DF,$0F,$EB,$1C,$7A,$AF,$5E,$38,$D7,$F7,$91,$E8,$1B,$FF,$5F,$E4
	.dc.b	$E1,$1F,$0B,$CF,$FD,$EB,$BF,$94,$D5,$F2,$8F,$83,$86,$23,$A4,$74
	.dc.b	$AE,$2F,$C6,$E2,$FE,$D9,$D1,$DA,$3E,$A7,$E9,$9F,$76,$93,$99,$4F
	.dc.b	$E1,$27,$37,$E1,$9F,$6C,$62,$3D,$9C,$5F,$32,$1D,$7C,$FE,$8B,$0F
	.dc.b	$5B,$FC,$0C,$EA,$FA,$47,$E1,$BF,$1D,$F2,$57,$8D,$F1,$77,$7C,$6A
	.dc.b	$FF,$79,$5F,$FA,$CE,$FE,$66,$BF,$DE,$1F,$C7,$BF,$9A,$E3,$C4,$7D
	.dc.b	$DE,$4F,$CF,$74,$FF,$7A,$8F,$CA,$F0,$FE,$83,$E7,$31,$1F,$99,$8F
	.dc.b	$D1,$FD,$4E,$23,$E0,$F8,$3F,$7F,$9B,$FC,$45,$F0,$59,$7F,$C4,$9F
	.dc.b	$0F,$FE,$1B,$F6,$FF,$09,$C1,$F8,$ED,$F0,$FC,$B3,$F1,$BC,$1F,$EE
	.dc.b	$4C,$BD,$6D,$7F,$21,$BF,$EC,$E7,$E7,$78,$DF,$94,$DB,$F7,$1B,$FE
	.dc.b	$4F,$0F,$F7,$C6,$1C,$AB,$F3,$3A,$7F,$B5,$63,$E8,$C6,$30,$FE,$A1
	.dc.b	$E3,$FE,$C6,$FF,$C8,$51,$FD,$4F,$E7,$1F,$54,$F9,$77,$CE,$B1,$56
	.dc.b	$BE,$C3,$FC,$75,$E4,$7F,$37,$8E,$37,$C2,$B7,$87,$3C,$BB,$E5,$8A
	.dc.b	$D5,$F8,$3F,$1C,$E7,$3F,$D9,$7E,$E8,$42,$0F,$C0,$B7,$18,$89,$87
	.dc.b	$67,$07,$2F,$C0,$B7,$F5,$0F,$A0,$7D,$1A,$30,$DC,$B7,$3A,$9C,$9C
	.dc.b	$70,$F5,$B8,$8D,$7B,$BC,$2F,$A4,$71,$F7,$3F,$CC,$1E,$77,$F5,$DF
	.dc.b	$86,$FF,$18,$0F,$7B,$3B,$7F,$81,$63,$3F,$4B,$88,$FF,$56,$C3,$F2
	.dc.b	$AC,$70,$7B,$8F,$89,$B7,$F2,$97,$06,$8F,$87,$C0,$F8,$6F,$F3,$E7
	.dc.b	$5F,$F1,$15,$F0,$FF,$B9,$F0,$74,$FE,$EF,$BA,$E9,$FC,$55,$FC,$75
	.dc.b	$E2,$3F,$A0,$C7,$E4,$6B,$13,$FC,$FE,$7F,$22,$FF,$24,$B7,$CD,$77
	.dc.b	$6C,$FF,$06,$9E,$7F,$10,$F8,$7F,$23,$67,$53,$FD,$2E,$31,$1F,$CC
	.dc.b	$7C,$DF,$EB,$F1,$89,$FE,$97,$1F,$D7,$FE,$21,$F1,$1E,$F6,$DF,$4A
	.dc.b	$FE,$73,$EC,$FE,$61,$FE,$79,$AA,$31,$1F,$CC,$5A,$FF,$5E,$FA,$7F
	.dc.b	$F7,$8D,$73,$6B,$FC,$EB,$D7,$B3,$D3,$E4,$FC,$33,$FC,$FB,$0F,$F2
	.dc.b	$93,$E3,$E0,$7C,$57,$AE,$FF,$B5,$EB,$FD,$E6,$31,$1F,$DC,$63,$FB
	.dc.b	$FC,$62,$3F,$BA,$5F,$2D,$FE,$F7,$B5,$F4,$31,$F0,$36,$BF,$97,$BF
	.dc.b	$84,$8F,$C0,$1F,$6B,$1F,$F7,$8B,$81,$E5,$7F,$7D,$9E,$11,$F8,$43
	.dc.b	$F0,$35,$C1,$EF,$C9,$E2,$7F,$A4,$79,$8F,$6F,$F3,$5D,$1F,$98,$6B
	.dc.b	$FB,$FF,$17,$5D,$FF,$E4,$78,$FC,$FC,$FF,$82,$7B,$7F,$5B,$DB,$AC
	.dc.b	$7F,$59,$DF,$FE,$63,$F8,$B5,$7E,$C6,$F8,$4E,$61,$E0,$7F,$C0,$0B
	.dc.b	$FF,$3C,$78,$BF,$BF,$F1,$3F,$84,$78,$79,$9F,$DF,$C4,$9F,$F4,$FE
	.dc.b	$3F,$C1,$BC,$4F,$E1,$44,$71,$3B,$F1,$B9,$6D,$FB,$BF,$12,$BF,$C6
	.dc.b	$B8,$6B,$FC,$72,$FF,$95,$F8,$FF,$CC,$3C,$DF,$E6,$CF,$F2,$B8,$CF
	.dc.b	$95,$FC,$EE,$72,$BC,$DF,$45,$71,$F3,$3F,$9A,$5F,$F4,$CE,$9F,$F0
	.dc.b	$CF,$07,$DF,$8C,$77,$7F,$4D,$3E,$B5,$7F,$07,$8E,$1D,$3F,$D5,$AB
	.dc.b	$FC,$63,$1F,$D7,$7B,$7F,$CA,$1E,$B7,$F8,$AE,$3F,$B0,$63,$FE,$2A
	.dc.b	$6D,$7F,$69,$6B,$FE,$23,$7F,$CF,$6F,$F8,$57,$1F,$EE,$71,$88,$FC
	.dc.b	$E9,$FE,$89,$F4,$BF,$15,$EA,$79,$4F,$F1,$C6,$5F,$E0,$1A,$E1,$87
	.dc.b	$F8,$13,$D2,$FF,$32,$7A,$BF,$91,$F5,$FF,$C7,$DA,$FF,$83,$6F,$FC
	.dc.b	$25,$7F,$E0,$DB,$FF,$35,$FA,$FB,$D3,$F6,$D5,$AA,$FF,$32,$47,$FA
	.dc.b	$6F,$DD,$FF,$44,$47,$F1,$2D,$EB,$FB,$AC,$2B,$F0,$FD,$9F,$F1,$65
	.dc.b	$7E,$8F,$DD,$FF,$1C,$7B,$9F,$31,$FA,$07,$F8,$EF,$DD,$FF,$3D,$67
	.dc.b	$FE,$37,$C3,$FC,$D3,$1F,$E8,$08,$FF,$23,$C7,$F9,$CB,$8F,$F5,$DE
	.dc.b	$47,$F8,$DF,$CE,$BB,$7A,$A3,$FC,$75,$58,$8F,$F2,$8F,$1F,$ED,$70
	.dc.b	$FE,$88,$7F,$A5,$A7,$FC,$6A,$FF,$25,$9F,$F2,$D7,$0F,$E0,$1F,$C1
	.dc.b	$FA,$1F,$E6,$E8,$FF,$38,$F9,$3F,$E9,$A3,$DB,$34,$FE,$7B,$B7,$F9
	.dc.b	$EB,$E0,$7F,$CF,$B8,$FF,$A3,$7E,$01,$7F,$EA,$0B,$FE,$51,$F0,$9F
	.dc.b	$D7,$2B,$FC,$C8,$D7,$FC,$EE,$FF,$A6,$F9,$3F,$D3,$DC,$BC,$5F,$F4
	.dc.b	$E9,$FE,$9D,$8D,$3F,$EA,$3A,$E5,$5F,$E9,$AF,$81,$F1,$BF,$EA,$0C
	.dc.b	$5E,$EF,$FA,$3E,$BE,$0D,$FF,$AB,$F2,$CF,$B7,$F1,$F0,$FF,$43,$3F
	.dc.b	$51,$FF,$54,$7D,$AF,$F9,$EA,$FF,$AE,$F3,$3E,$87,$11,$FE,$82,$AA
	.dc.b	$7F,$A8,$B0,$61,$FE,$6F,$9F,$F4,$97,$E7,$FF,$1A,$31,$5F,$E8,$A6
	.dc.b	$BF,$D2,$AF,$FC,$D2,$FF,$AC,$23,$F3,$5C,$5F,$EC,$37,$F4,$F1,$FE
	.dc.b	$67,$CF,$FC,$DB,$18,$8E,$1D,$FF,$E6,$7C,$7F,$DF,$5E,$A7,$F9,$89
	.dc.b	$FF,$62,$1F,$EB,$2E,$7C,$57,$8F,$E0,$BF,$09,$8F,$C7,$D7,$D7,$E1
	.dc.b	$FF,$98,$63,$F5,$F1,$EB,$8F,$CA,$E7,$FC,$53,$1F,$E9,$AF,$F2,$0F
	.dc.b	$81,$FF,$26,$F8,$EF,$DB,$7F,$C8,$FE,$57,$F8,$9F,$C3,$FF,$1B,$7A
	.dc.b	$1F,$E3,$CF,$2B,$FC,$5D,$E9,$FF,$8D,$3E,$AB,$FC,$61,$C3,$FE,$BB
	.dc.b	$E9,$FF,$15,$63,$FE,$4F,$C7,$FC,$55,$B7,$FA,$7E,$7F,$C5,$9E,$3A
	.dc.b	$27,$0F,$E5,$88,$7E,$63,$AF,$FC,$4B,$8F,$FB,$1F,$E0,$7F,$B3,$BF
	.dc.b	$E2,$7A,$FE,$D6,$B9,$BF,$C7,$1F,$05,$FB,$4F,$82,$FE,$F9,$3F,$EE
	.dc.b	$F6,$AF,$7B,$F1,$1F,$03,$FE,$0C,$8F,$F6,$4F,$C0,$57,$FD,$AD,$F0
	.dc.b	$1C,$2B,$F1,$CF,$28,$FF,$7E,$79,$7F,$E0,$37,$FD,$B8,$F9,$9F,$E8
	.dc.b	$B9,$F9,$B3,$FE,$F9,$1F,$ED,$69,$FF,$78,$CF,$C2,$DB,$E0,$9F,$06
	.dc.b	$3F,$57,$22,$70,$7F,$77,$AF,$F2,$CF,$0D,$9F,$A5,$E5,$FF,$76,$F7
	.dc.b	$7F,$BC,$FB,$A7,$FB,$DF,$37,$F7,$3E,$9F,$EE,$C7,$F5,$FA,$FF,$42
	.dc.b	$73,$FF,$A1,$31,$FE,$C3,$D7,$FD,$97,$BB,$FD,$2F,$D9,$FD,$73,$5F
	.dc.b	$F0,$E7,$83,$E5,$DB,$11,$FB,$1C,$BF,$AF,$F8,$1F,$94,$E2,$FF,$0C
	.dc.b	$78,$3F,$9A,$81,$F8,$19,$FE,$BF,$7C,$CF,$80,$FF,$5E,$BF,$F6,$93
	.dc.b	$FA,$BF,$3F,$F1,$51,$FE,$B8,$8F,$D7,$4F,$13,$FA,$EC,$7F,$BD,$F9
	.dc.b	$7B,$DF,$F5,$4A,$FD,$CB,$83,$FA,$B4,$FA,$76,$FF,$78,$78,$BF,$D4
	.dc.b	$30,$E9,$7F,$C1,$1F,$37,$8D,$FC,$BA,$3F,$A6,$47,$87,$C8,$C7,$A1
	.dc.b	$FD,$1A,$73,$3E,$DF,$27,$F8,$C4,$7F,$1D,$8F,$95,$D3,$FD,$27,$3F
	.dc.b	$F3,$74,$7F,$8F,$FC,$FF,$91,$AF,$E9,$7C,$EF,$E5,$B3,$F8,$5C,$FF
	.dc.b	$96,$F9,$BF,$88,$8F,$A9,$87,$C6,$3F,$D4,$8F,$C5,$1E,$66,$63,$31
	.dc.b	$FD,$85,$E1,$F6,$FF,$D0,$3C,$FD,$3F,$D0,$BD,$2B,$9F,$C9,$3E,$DF
	.dc.b	$C2,$BD,$1F,$EA,$38,$7E,$FE,$F5,$4F,$B4,$FE,$05,$5A,$A5,$6A,$5E
	.dc.b	$6F,$F4,$79,$CD,$FE,$23,$ED,$BF,$49,$AF,$F0,$CF,$4B,$FC,$C9,$5D
	.dc.b	$EB,$BE,$BF,$D0,$39,$3F,$0F,$9F,$A6,$DE,$3E,$2F,$FA,$EC,$63,$0E
	.dc.b	$51,$ED,$E9,$FE,$93,$5B,$8F,$EC,$B5,$BA,$F2,$C7,$C3,$E0,$DC,$72
	.dc.b	$8F,$C7,$70,$7F,$56,$C7,$E1,$FE,$7F,$F4,$B0,$74,$F5,$BF,$CB,$CB
	.dc.b	$B7,$84,$7B,$2B,$78,$F5,$8F,$D5,$F1,$7C,$89,$CC,$F9,$27,$F4,$33
	.dc.b	$FC,$C6,$7F,$B1,$63,$F5,$3F,$69,$E1,$DB,$E1,$BF,$DA,$C6,$E3,$11
	.dc.b	$FB,$8E,$1F,$ED,$71,$8F,$A6,$F5,$37,$3F,$ED,$FE,$CF,$C7,$E2,$F8
	.dc.b	$DC,$D8,$DB,$E6,$C7,$C9,$F0,$F1,$1E,$7C,$FF,$CC,$7B,$30,$F8,$9C
	.dc.b	$FF,$EE,$A8,$C5,$7C,$D8,$FE,$25,$CF,$FD,$FB,$D7,$F0,$7D,$57,$F1
	.dc.b	$5F,$89,$C5,$F8,$23,$F8,$C7,$B7,$88,$F0,$67,$ED,$CB,$FC,$D1,$C9
	.dc.b	$FE,$74,$EE,$FE,$1F,$F1,$9B,$8C,$5F,$F5,$BF,$7B,$FB,$D4,$7F,$1C
	.dc.b	$E1,$FC,$22,$E6,$CE,$5F,$E1,$7E,$DF,$C8,$66,$F2,$3E,$6C,$EE,$31
	.dc.b	$97,$E4,$B8,$BF,$3F,$EC,$7D,$DF,$33,$F3,$1A,$38,$7F,$A3,$70,$FF
	.dc.b	$6D,$E7,$FF,$02,$E3,$F8,$5C,$FD,$98,$7E,$7F,$67,$17,$F7,$3E,$7C
	.dc.b	$DF,$EA,$67,$2F,$FB,$62,$DE,$07,$F8,$AB,$D9,$F3,$EB,$6B,$B5,$6B
	.dc.b	$83,$FC,$7F,$1F,$DF,$FE,$03,$FD,$5B,$EE,$96,$AF,$23,$95,$7F,$85
	.dc.b	$2B,$52,$F8,$0C,$DF,$23,$8F,$E2,$BC,$3F,$40,$7E,$6B,$08,$63,$3F
	.dc.b	$1E,$49,$CC,$EF,$F3,$6E,$4E,$07,$E9,$1E,$4F,$C0,$9F,$80,$BF,$07
	.dc.b	$46,$1B,$1B,$35,$19,$1F,$9C,$E6,$76,$72,$FC,$EB,$13,$BF,$D0,$7E
	.dc.b	$1F,$8D,$FC,$D7,$5F,$B9,$B7,$AA,$31,$0E,$AF,$F2,$EF,$B5,$FD,$93
	.dc.b	$83,$D5,$BF,$C1,$8C,$3E,$DE,$0D,$DF,$F8,$CE,$B7,$C1,$88,$FF,$64
	.dc.b	$1F,$D2,$70,$FF,$1C,$BF,$E3,$9C,$7C,$30,$F0,$1F,$29,$F9,$AE,$DC
	.dc.b	$3F,$9C,$ED,$B8,$7E,$37,$BB,$F2,$1D,$9F,$E0,$2C,$38,$CE,$4F,$C8
	.dc.b	$61,$C1,$FE,$82,$FE,$53,$1F,$F6,$6C,$BE,$23,$C0,$DA,$FC,$9E,$2F
	.dc.b	$88,$7C,$AF,$95,$B3,$A6,$BF,$C0,$BD,$0F,$EC,$B1,$8F,$27,$FA,$CC
	.dc.b	$7F,$8C,$E3,$1A,$7F,$61,$8F,$E6,$DF,$12,$F9,$6F,$84,$DC,$3C,$7E
	.dc.b	$E7,$0B,$FE,$C6,$D7,$FC,$25,$E2,$A3,$11,$FE,$18,$E0,$C6,$9F,$2E
	.dc.b	$7F,$37,$A7,$FA,$D2,$B8,$78,$DF,$B4,$9C,$C7,$F6,$26,$BF,$D5,$5A
	.dc.b	$FF,$54,$6B,$FE,$FE,$E9,$AF,$F7,$C6,$AF,$8E,$FF,$3B,$67,$F3,$A7
	.dc.b	$39,$62,$3F,$95,$D7,$F8,$FA,$78,$E7,$FE,$D4,$CF,$F0,$91,$EE,$9F
	.dc.b	$E9,$8D,$78,$47,$F0,$EF,$8B,$F9,$57,$D3,$85,$18,$D7,$FC,$2D,$F1
	.dc.b	$5C,$F8,$7F,$C3,$AD,$F5,$4E,$7D,$EB,$97,$E5,$63,$C1,$B7,$F8,$13
	.dc.b	$9B,$F3,$1E,$8F,$E2,$F5,$F5,$CF,$F8,$23,$AB,$F9,$57,$57,$E6,$A3
	.dc.b	$13,$F9,$9E,$AF,$F2,$EF,$06,$23,$F3,$35,$FA,$0E,$1C,$70,$7F,$8A
	.dc.b	$F2,$FD,$0D,$76,$7F,$D0,$AF,$EC,$8F,$F6,$3F,$BB,$FE,$27,$EC,$FD
	.dc.b	$8B,$FE,$29,$EE,$FD,$FF,$7F,$FA,$15,$FD,$DC,$67,$C0,$FE,$06,$FF
	.dc.b	$09,$BF,$E1,$9A,$A3,$11,$FE,$43,$F3,$B1,$D9,$FE,$6E,$F1,$3F,$8D
	.dc.b	$79,$1F,$C7,$3C,$6C,$DF,$FA,$EE,$73,$E3,$FF,$24,$F1,$F3,$3F,$C7
	.dc.b	$A7,$F9,$17,$91,$87,$67,$F2,$CC,$FF,$9A,$ED,$EA,$6E,$DF,$CC,$FC
	.dc.b	$7E,$73,$FC,$72,$3A,$F9,$5F,$D0,$FC,$CF,$E9,$57,$FC,$FA,$31,$E5
	.dc.b	$EA,$7F,$A5,$7C,$2F,$FA,$6F,$CD,$F2,$F4,$7F,$96,$F0,$FC,$A9,$FD
	.dc.b	$53,$6F,$B5,$D3,$FD,$E6,$31,$1F,$68,$FE,$C9,$D7,$FE,$A0,$AF,$F3
	.dc.b	$34,$E9,$FF,$01,$7A,$3F,$DD,$2B,$FC,$B7,$3F,$DF,$1F,$F2,$3F,$A5
	.dc.b	$FD,$F3,$A3,$FD,$4B,$1F,$DA,$BB,$AB,$FE,$60,$F2,$3F,$C1,$71,$8C
	.dc.b	$7F,$C1,$DD,$DF,$E5,$9C,$7F,$C2,$FA,$7F,$8D,$7B,$FF,$C2,$F1,$88
	.dc.b	$FF,$09,$47,$FA,$A3,$D4,$FF,$52,$E5,$F9,$CA,$D6,$5F,$E0,$BF,$17
	.dc.b	$FD,$A1,$EB,$6B,$3F,$F5,$4C,$E6,$5E,$A7,$FA,$A2,$7F,$D5,$5B,$FF
	.dc.b	$B0,$FC,$8C,$6B,$FE,$B0,$8F,$F1,$AE,$5F,$E3,$8C,$FC,$1E,$C7,$F9
	.dc.b	$17,$7F,$F5,$DF,$0F,$0C,$3F,$C4,$D5,$FE,$BD,$F6,$7F,$C9,$DE,$CF
	.dc.b	$E5,$DF,$F2,$76,$3F,$E5,$AA,$FF,$3F,$B5,$FF,$5E,$7B,$5F,$42,$7F
	.dc.b	$CB,$2A,$FF,$41,$EE,$FF,$97,$3D,$CF,$AA,$D8,$C9,$F4,$F0,$FF,$35
	.dc.b	$EF,$FE,$78,$CB,$33,$FE,$48,$F7,$7F,$CE,$D5,$FE,$40,$CF,$FD,$3F
	.dc.b	$D3,$FA,$DD,$BF,$C7,$BE,$F5,$C3,$A4,$7F,$90,$2B,$11,$FC,$A3,$97
	.dc.b	$FD,$84,$FF,$1F,$C3,$FD,$3A,$FD,$7F,$57,$FD,$25,$C7,$F8,$9F,$73
	.dc.b	$FD,$55,$B7,$FA,$A7,$8F,$FC,$59,$EA,$FA,$FA,$FF,$D7,$7A,$79,$1A
	.dc.b	$FF,$AE,$76,$FF,$1F,$E7,$F9,$19,$FF,$64,$7C,$B7,$FB,$1F,$AB,$FD
	.dc.b	$25,$18,$7F,$D9,$5C,$7F,$EC,$99,$D4,$7F,$B1,$30,$FF,$5F,$FC,$07
	.dc.b	$91,$FF,$63,$7B,$0C,$FF,$D7,$9A,$FE,$16,$FF,$D4,$51,$FE,$B7,$CF
	.dc.b	$FC,$97,$7F,$EA,$A7,$FA,$5F,$DB,$7F,$A8,$A3,$FD,$F1,$CC,$BC,$47
	.dc.b	$FB,$4F,$DE,$FA,$EF,$83,$0F,$F4,$BE,$1F,$E6,$3F,$D8,$FC,$55,$F8
	.dc.b	$7A,$E1,$D9,$FE,$9B,$C3,$FD,$27,$9F,$F6,$DA,$FC,$C3,$F9,$E8,$F0
	.dc.b	$1F,$E8,$3A,$FD,$FB,$E3,$EE,$C4,$7F,$B8,$2B,$F6,$5E,$17,$CE,$F4
	.dc.b	$FF,$D0,$7D,$5F,$E8,$63,$FD,$0B,$5B,$F8,$1F,$BB,$8F,$C4,$C3,$F8
	.dc.b	$64,$AB,$F6,$D8,$FC,$BD,$BF,$CF,$FB,$7F,$70,$AB,$A7,$F0,$38,$FF
	.dc.b	$3D,$47,$F9,$BA,$3F,$DE,$6B,$83,$FC,$DB,$E4,$7F,$26,$F8,$1F,$F3
	.dc.b	$8F,$E2,$3F,$E6,$4F,$37,$FC,$CF,$F9,$2F,$F9,$5B,$D6,$FF,$2E,$FE
	.dc.b	$83,$7D,$6F,$F3,$1F,$E9,$7F,$E7,$AF,$2B,$C9,$C1,$0F,$F2,$66,$FE
	.dc.b	$0D,$BF,$C9,$DE,$9F,$F9,$5B,$3F,$F2,$57,$0F,$FA,$8A,$7F,$C8,$15
	.dc.b	$FE,$1D,$F8,$5F,$EC,$AF,$AF,$A3,$FD,$5D,$1F,$DD,$63,$FC,$45,$F0
	.dc.b	$BF,$E0,$9E,$4F,$F2,$07,$C0,$FF,$B5,$7E,$2D,$F0,$BF,$E0,$79,$F8
	.dc.b	$3E,$57,$F8,$3F,$3F,$F1,$4E,$9F,$E9,$A7,$FC,$35,$97,$9F,$7D,$7C
	.dc.b	$2F,$1C,$04,$27,$C2,$7F,$84,$73,$FF,$06,$E3,$F9,$AF,$3B,$CD,$5F
	.dc.b	$41,$F5,$BF,$CD,$23,$FD,$5D,$DB,$FE,$04,$8D,$47,$C3,$F3,$BF,$BD
	.dc.b	$75,$7D,$27,$F5,$93,$E1,$3C,$7A,$FF,$70,$E6,$FA,$1A,$FE,$16,$BF
	.dc.b	$D4,$3E,$E7,$F7,$0E,$AF,$EE,$5D,$7F,$E9,$63,$FD,$2F,$D5,$FD,$BB
	.dc.b	$3F,$F5,$17,$4F,$F6,$9F,$61,$EC,$BB,$FF,$0B,$9F,$F9,$13,$6F,$EC
	.dc.b	$5B,$FA,$FB,$3E,$24,$FE,$62,$3C,$C7,$C8,$7F,$37,$C5,$F9,$AE,$0F
	.dc.b	$E0,$91,$FD,$97,$9F,$F8,$27,$99,$FE,$9C,$8C,$5F,$E6,$1B,$7F,$A9
	.dc.b	$7F,$5B,$E6,$FF,$89,$BE,$0B,$F9,$64,$7F,$5C,$F6,$BF,$AA,$3F,$BA
	.dc.b	$BF,$E2,$1E,$3F,$F0,$98,$C4,$79,$8F,$E9,$DD,$1F,$EE,$9F,$85,$D7
	.dc.b	$97,$D0,$FE,$95,$0D,$78,$79,$9E,$AF,$5B,$31,$CF,$CF,$C5,$7F,$1C
	.dc.b	$CB,$F8,$AE,$5F,$C6,$B3,$FE,$83,$7F,$A5,$9F,$E7,$EF,$F9,$F5,$FF
	.dc.b	$42,$57,$F5,$98,$F8,$BE,$67,$F2,$BF,$53,$F0,$78,$7F,$9E,$78,$3F
	.dc.b	$9E,$BF,$41,$FE,$71,$8F,$F1,$83,$F8,$CF,$A3,$FC,$CF,$1F,$EC,$93
	.dc.b	$FE,$CF,$F6,$DE,$AA,$B5,$5F,$D1,$8F,$F0,$51,$9E,$6F,$87,$7C,$79
	.dc.b	$DA,$7A,$2B,$FA,$6C,$EB,$0C,$F8,$FC,$27,$C2,$78,$AF,$FA,$13,$DC
	.dc.b	$F9,$EF,$F4,$3A,$FF,$70,$FC,$3F,$6F,$C4,$BE,$9E,$5F,$A0,$FC,$7B
	.dc.b	$FD,$C9,$9A,$73,$3F,$C4,$E5,$5E,$6E,$4F,$16,$5F,$70,$6D,$F0,$B1
	.dc.b	$4E,$A7,$FC,$55,$9F,$17,$FC,$19,$8F,$EC,$3F,$1F,$F0,$D7,$82,$3F
	.dc.b	$91,$C7,$82,$DC,$FC,$EF,$FC,$29,$F9,$3F,$78,$C4,$23,$D7,$FA,$07
	.dc.b	$3C,$3C,$DA,$7C,$E8,$C5,$7C,$6C,$3F,$C5,$DB,$FA,$CC,$79,$3E,$3F
	.dc.b	$88,$F7,$7D,$CF,$D6,$8C,$5F,$E7,$3E,$C7,$EC,$F8,$1F,$13,$9B,$BE
	.dc.b	$48,$C5,$7F,$92,$E1,$F7,$7E,$F8,$FF,$17,$57,$F8,$BF,$8F,$F0,$73
	.dc.b	$99,$FC,$69,$F3,$7D,$BF,$7B,$4F,$C1,$9F,$88,$AF,$C5,$C7,$D6,$D7
	.dc.b	$F1,$F7,$8A,$FA,$3C,$0D,$BF,$2B,$39,$9F,$EC,$D1,$F9,$8C,$3F,$B3
	.dc.b	$DF,$9A,$7C,$C7,$DF,$F6,$BF,$29,$3F,$90,$8F,$C5,$D7,$F0,$0F,$1D
	.dc.b	$F0,$EB,$FE,$D0,$BE,$07,$F4,$35,$B8,$D7,$97,$FC,$AA,$F8,$FE,$37
	.dc.b	$AC,$7F,$CA,$91,$D3,$1E,$7F,$92,$66,$5F,$8F,$42,$18,$FC,$97,$4E
	.dc.b	$0D,$7E,$25,$87,$03,$F1,$4E,$9E,$C7,$EA,$B9,$BE,$46,$1F,$E6,$5F
	.dc.b	$C4,$9F,$88,$3F,$0C,$46,$1B,$F9,$37,$96,$30,$4F,$C9,$F1,$5C,$63
	.dc.b	$3F,$8B,$3E,$17,$FD,$BF,$D3,$FA,$FC,$BF,$D0,$31,$87,$FD,$AA,$DE
	.dc.b	$9F,$A4,$71,$F9,$9F,$AB,$3F,$63,$AF,$A7,$CD,$7E,$29,$FF,$00,$C7
	.dc.b	$E2,$8F,$DE,$56,$1F,$E1,$87,$E5,$7E,$1D,$F0,$FF,$B0,$F2,$3F,$90
	.dc.b	$C6,$27,$91,$FB,$6B,$FF,$05,$FC,$DF,$7F,$43,$53,$D0,$4F,$F8,$31
	.dc.b	$C7,$95,$76,$F8,$11,$FE,$93,$F3,$BF,$CC,$9E,$17,$E8,$AA,$CF,$80
	.dc.b	$D7,$89,$DA,$F8,$C7,$6F,$87,$67,$F0,$7E,$1F,$53,$F3,$76,$FE,$77
	.dc.b	$18,$8F,$E3,$D1,$E6,$8C,$47,$F2,$1D,$3F,$C3,$2F,$AF,$81,$D6,$E2
	.dc.b	$F9,$CF,$D5,$9F,$D4,$CE,$B8,$FE,$AC,$7C,$23,$E1,$CF,$DD,$7F,$01
	.dc.b	$7B,$7D,$C7,$FB,$9B,$7F,$ED,$BC,$DF,$8C,$E7,$FC,$57,$3B,$D8,$FC
	.dc.b	$BD,$FE,$52,$7F,$1C,$7E,$3F,$A7,$58,$72,$8C,$F5,$FE,$62,$F8,$E7
	.dc.b	$FD,$29,$E7,$DD,$FA,$61,$DF,$CB,$2F,$F6,$44,$67,$BF,$11,$B8,$78
	.dc.b	$18,$F0,$FF,$75,$E2,$FE,$FF,$C6,$FE,$09,$97,$A2,$73,$F0,$1F,$E2
	.dc.b	$B9,$78,$BC,$5F,$E0,$F1,$9F,$1F,$E0,$78,$FF,$33,$9B,$F9,$6E,$3F
	.dc.b	$C4,$75,$FE,$D3,$D1,$FC,$5E,$7E,$37,$93,$FD,$83,$97,$F9,$1D,$A3
	.dc.b	$17,$F9,$13,$FB,$FC,$65,$FE,$63,$DF,$FD,$DE,$31,$1F,$DE,$8F,$EE
	.dc.b	$CE,$23,$F2,$78,$7D,$5F,$2B,$F9,$F3,$FD,$06,$7F,$A2,$D7,$F3,$6F
	.dc.b	$07,$FD,$FE,$F8,$3C,$BF,$E9,$6F,$FB,$D7,$C3,$FB,$D1,$88,$FE,$E7
	.dc.b	$4F,$37,$FA,$BC,$7F,$BE,$FC,$EF,$C2,$46,$21,$E7,$61,$FE,$05,$A7
	.dc.b	$F8,$2A,$DE,$7F,$F6,$89,$F0,$46,$1F,$36,$7F,$A3,$8F,$EE,$F1,$FB
	.dc.b	$3F,$17,$E3,$FA,$1F,$C1,$F8,$BF,$29,$E9,$7F,$75,$C3,$FD,$BD,$E9
	.dc.b	$7F,$02,$C7,$FC,$13,$8F,$49,$CF,$07,$08,$FD,$F7,$97,$F4,$B0,$FF
	.dc.b	$00,$BC,$7D,$4E,$96,$F5,$31,$7C,$E3,$18,$7E,$97,$C3,$FE,$15,$97
	.dc.b	$E9,$EB,$F5,$BA,$7E,$A7,$CE,$FE,$D5,$1F,$E4,$0E,$6C,$3F,$80,$3F
	.dc.b	$C6,$5E,$C7,$F9,$03,$8F,$FC,$89,$28,$C4,$FF,$39,$DB,$F8,$FC,$62
	.dc.b	$3F,$C5,$50,$7F,$C8,$C7,$F1,$BC,$FF,$DE,$38,$78,$A0,$FF,$BD,$32
	.dc.b	$F7,$E9,$FD,$EF,$D4,$FF,$1D,$57,$BF,$6F,$F1,$A7,$07,$C6,$8F,$F1
	.dc.b	$91,$FE,$A0,$A7,$0F,$F8,$E7,$9B,$F0,$9A,$7F,$87,$EB,$F2,$5D,$7F
	.dc.b	$E1,$CE,$9F,$C7,$71,$7F,$85,$A7,$F2,$C7,$F9,$23,$4F,$D0,$E5,$FC
	.dc.b	$D2,$3F,$C9,$28,$FE,$ED,$E0,$FE,$F0,$FE,$F5,$3F,$C3,$3C,$4F,$E1
	.dc.b	$8F,$F1,$8F,$2F,$F8,$E7,$91,$FC,$66,$31,$E4,$FF,$16,$97,$E4,$DF
	.dc.b	$D7,$F2,$FD,$E9,$FE,$2C,$CB,$C1,$E9,$7F,$38,$8F,$9D,$F9,$2F,$8F
	.dc.b	$D1,$C4,$23,$CD,$D9,$FC,$DE,$3F,$CF,$B7,$FE,$B0,$E1,$FE,$4D,$AF
	.dc.b	$F2,$B3,$F9,$67,$3F,$91,$C6,$9E,$48,$F6,$6F,$FE,$0C,$7F,$8E,$79
	.dc.b	$1F,$C6,$38,$7F,$C3,$51,$F2,$31,$FE,$6F,$D5,$FE,$D3,$F1,$3F,$C8
	.dc.b	$5A,$7F,$B5,$79,$7E,$8D,$FE,$F6,$3E,$A9,$FE,$AE,$E6,$FD,$D3,$F3
	.dc.b	$F0,$FA,$4F,$F9,$EA,$7F,$5A,$FE,$07,$A3,$F5,$4F,$EB,$0F,$E9,$1D
	.dc.b	$FF,$AA,$EB,$FE,$3F,$5F,$8E,$E1,$FD,$0D,$7F,$1D,$3F,$27,$87,$E7
	.dc.b	$CF,$F3,$9E,$9F,$95,$8F,$D8,$47,$F8,$23,$A3,$F2,$B1,$F9,$1F,$C8
	.dc.b	$BF,$09,$1E,$4F,$03,$FB,$FD,$E3,$B7,$F7,$7F,$93,$6B,$C2,$FF,$3F
	.dc.b	$F4,$F0,$EF,$FE,$35,$DF,$FE,$72,$D3,$FC,$C7,$E1,$7F,$86,$B4,$FE
	.dc.b	$57,$1F,$0F,$6F,$F4,$C1,$FA,$7E,$6F,$D4,$BF,$E6,$1D,$BF,$D3,$BE
	.dc.b	$0F,$93,$C7,$FE,$83,$7F,$9A,$E2,$FF,$31,$BF,$E7,$43,$F0,$B7,$FE
	.dc.b	$BA,$F5,$BF,$66,$7E,$7B,$AF,$EF,$1F,$D2,$F3,$FC,$1F,$67,$F9,$53
	.dc.b	$F2,$4F,$F2,$0F,$7F,$E2,$3F,$25,$FF,$20,$7E,$9F,$E4,$F1,$1F,$96
	.dc.b	$79,$B8,$3F,$BD,$EB,$FB,$9F,$CE,$3F,$5C,$DF,$1B,$7C,$1F,$42,$3F
	.dc.b	$C2,$78,$FB,$BB,$BF,$2D,$E2,$FF,$94,$B4,$FE,$EB,$BF,$F7,$BD,$FF
	.dc.b	$B6,$79,$DF,$E7,$6E,$5F,$F0,$57,$27,$B3,$1F,$ED,$5A,$39,$BF,$D0
	.dc.b	$8F,$C2,$8F,$CD,$73,$7F,$C0,$0E,$FF,$F1,$BC,$7F,$61,$9F,$F4,$BF
	.dc.b	$17,$F0,$38,$FA,$BC,$1F,$C4,$CF,$EC,$93,$FC,$30,$FF,$63,$DF,$F5
	.dc.b	$E7,$EE,$9F,$C9,$CF,$EE,$7D,$7F,$82,$AF,$EA,$31,$FE,$58,$EC,$ED
	.dc.b	$DF,$F8,$AC,$3F,$A5,$4F,$F9,$73,$2F,$17,$85,$FD,$23,$C1,$F2,$63
	.dc.b	$F6,$AB,$FD,$BD,$F4,$0F,$82,$84,$1F,$2F,$FE,$75,$8C,$46,$7E,$0B
	.dc.b	$FC,$63,$F8,$A6,$21,$F8,$57,$F3,$F9,$CC,$70,$FC,$23,$F9,$34,$62
	.dc.b	$3B,$4F,$FB,$B7,$F2,$BF,$81,$7E,$3C,$5F,$18,$F8,$87,$C7,$7E,$D3
	.dc.b	$C6,$61,$C3,$2F,$3D,$EE,$BE,$D7,$92,$B7,$EB,$7B,$BF,$5E,$59,$8C
	.dc.b	$47,$C1,$FA,$0E,$1C,$58,$8F,$8B,$F2,$18,$E3,$FE,$B9,$BF,$C1,$97
	.dc.b	$D0,$77,$E6,$E1,$0F,$C5,$BD,$3A,$FF,$06,$BF,$91,$0F,$9F,$E1,$8F
	.dc.b	$4B,$F9,$9B,$62,$17,$F3,$F9,$1C,$19,$8C,$4F,$F6,$27,$EA,$7D,$FF
	.dc.b	$F6,$68,$E9,$87,$DD,$FC,$5F,$C7,$DB,$E6,$7C,$5A,$FF,$88,$6B,$BD
	.dc.b	$7E,$17,$83,$F4,$91,$FE,$FC,$9D,$63,$F1,$3C,$2F,$CC,$7A,$5F,$3F
	.dc.b	$DF,$F8,$38,$FE,$2A,$F6,$FD,$DE,$8F,$B9,$D5,$E0,$AF,$A9,$C3,$F8
	.dc.b	$08,$FC,$07,$D3,$7F,$77,$D3,$C7,$83,$E5,$F8,$D3,$EB,$75,$B7,$E6
	.dc.b	$A7,$3F,$51,$8B,$FC,$F7,$5F,$F8,$53,$6F,$E9,$D9,$74,$8F,$D8,$3F
	.dc.b	$E4,$88,$FE,$5B,$87,$C1,$E0,$FC,$AE,$BF,$E0,$5A,$F8,$BC,$3F,$11
	.dc.b	$F8,$F7,$C3,$1F,$B2,$DA,$F6,$76,$FE,$52,$3F,$CD,$78,$FE,$C7,$4F
	.dc.b	$87,$A7,$CD,$F2,$BF,$39,$AF,$F0,$FC,$BE,$54,$63,$3F,$F3,$34,$27
	.dc.b	$F9,$B4,$7D,$E8,$FD,$D6,$98,$8F,$93,$1F,$9D,$CB,$FD,$AF,$E5,$BE
	.dc.b	$27,$FD,$2C,$DF,$C6,$F8,$39,$F8,$9F,$F4,$B3,$6B,$8D,$70,$1F,$8E
	.dc.b	$7E,$BF,$CD,$FE,$35,$F8,$0B,$F0,$17,$DB,$A3,$0D,$FC,$77,$86,$DF
	.dc.b	$E4,$79,$C9,$2B,$31,$C5,$E7,$1A,$FC,$7B,$2E,$EF,$B7,$97,$C3,$F0
	.dc.b	$6F,$F4,$AE,$BF,$EC,$87,$FD,$D3,$E0,$74,$F7,$FF,$C3,$BF,$16,$5E
	.dc.b	$DE,$4F,$F1,$BF,$E4,$16,$7F,$D9,$7F,$90,$7F,$5C,$DF,$E8,$73,$FF
	.dc.b	$2A,$EF,$C4,$7D,$08,$FE,$5B,$D5,$8E,$EF,$C9,$CF,$DA,$9C,$1F,$6D
	.dc.b	$FB,$B5,$8A,$FC,$10,$F9,$D7,$E3,$BF,$AE,$9C,$3F,$B0,$97,$C2,$7F
	.dc.b	$68,$BF,$16,$1F,$CB,$63,$18,$FF,$19,$AD,$FC,$27,$F0,$3F,$0E,$E1
	.dc.b	$EC,$C1,$E5,$7F,$33,$C7,$8C,$7F,$44,$8F,$E3,$5F,$59,$F8,$2A,$E1
	.dc.b	$3F,$D2,$A3,$11,$E2,$8F,$CC,$7B,$DC,$9F,$E9,$BE,$1F,$FA,$0B,$2F
	.dc.b	$E1,$31,$6B,$9E,$D5,$E5,$A7,$FA,$23,$DF,$FE,$1F,$EA,$D9,$F0,$FD
	.dc.b	$1D,$7C,$1F,$65,$FE,$8E,$F0,$BF,$8A,$FE,$CF,$FE,$7B,$E3,$FE,$74
	.dc.b	$FF,$9E,$F1,$FE,$33,$85,$C3,$D7,$85,$7D,$58,$FF,$A0,$75,$FA,$1F
	.dc.b	$01,$EB,$DB,$FB,$56,$D9,$BF,$87,$BD,$7E,$14,$BE,$8B,$E2,$ED,$F4
	.dc.b	$DB,$1F,$D0,$BE,$23,$A3,$FD,$DF,$93,$FB,$1C,$62,$3E,$AF,$B2,$8C
	.dc.b	$4F,$F6,$97,$53,$FC,$1D,$B9,$3E,$4F,$F1,$79,$7D,$8E,$4C,$74,$FE
	.dc.b	$36,$5F,$13,$FC,$46,$9F,$09,$FD,$DE,$73,$1F,$D6,$6F,$F2,$F1,$88
	.dc.b	$FE,$97,$1F,$C6,$F2,$FF,$1C,$C7,$79,$D6,$DF,$DA,$F2,$7B,$EE,$8F
	.dc.b	$F1,$27,$85,$FB,$61,$E0,$EA,$33,$E1,$7F,$AB,$3C,$1C,$67,$FE,$D1
	.dc.b	$BE,$13,$88,$78,$9F,$D0,$1F,$F1,$87,$B9,$98,$FE,$0B,$97,$86,$7F
	.dc.b	$8A,$79,$7F,$C4,$0F,$E2,$3E,$5F,$3F,$2F,$F8,$80,$F2,$B1,$3B,$F2
	.dc.b	$FF,$8F,$61,$CF,$0F,$E3,$51,$2F,$F9,$34,$FF,$2C,$F4,$3F,$97,$F9
	.dc.b	$FF,$CB,$23,$1F,$26,$F8,$C7,$A5,$FC,$DE,$73,$18,$A5,$FF,$97,$B2
	.dc.b	$FB,$94,$F1,$FF,$BA,$F8,$5F,$3B,$D4,$FB,$9D,$BF,$E3,$1A,$FE,$AB
	.dc.b	$D9,$FC,$4F,$0F,$BE,$FF,$4F,$8C,$FA,$DF,$D6,$27,$FC,$27,$1F,$D7
	.dc.b	$FD,$CF,$F1,$97,$B1,$8F,$8D,$FE,$2D,$3F,$D9,$73,$FE,$D1,$DF,$FE
	.dc.b	$4E,$8C,$57,$F8,$ED,$FE,$EB,$1F,$CA,$30,$FF,$61,$46,$32,$FE,$53
	.dc.b	$01,$F2,$FF,$EC,$6A,$FF,$01,$7B,$9F,$E0,$EF,$73,$17,$FD,$BE,$3F
	.dc.b	$B7,$E1,$FE,$14,$8F,$F2,$4C,$7F,$93,$FC,$9F,$C8,$FB,$DF,$E1,$4D
	.dc.b	$BF,$C2,$FE,$F6,$34,$FE,$27,$8F,$1D,$BF,$C0,$5D,$F9,$9F,$F0,$BC
	.dc.b	$7F,$81,$7D,$9F,$F3,$AE,$DF,$E6,$7F,$33,$FC,$B7,$87,$FB,$62,$31
	.dc.b	$A6,$B6,$FF,$37,$FC,$07,$CC,$F8,$3F,$F2,$5F,$C1,$5D,$3F,$9A,$3F
	.dc.b	$07,$1F,$F2,$65,$FF,$92,$1A,$FF,$9E,$FE,$0F,$BD,$FF,$95,$1F,$F2
	.dc.b	$84,$FE,$BA,$FF,$9B,$FA,$9F,$B0,$8F,$C2,$9F,$E6,$88,$FC,$4F,$C2
	.dc.b	$7D,$AF,$85,$FF,$38,$79,$DF,$E1,$CE,$9F,$C0,$C6,$39,$7F,$CE,$5E
	.dc.b	$D3,$D6,$7C,$B7,$FA,$07,$7F,$F4,$7F,$07,$F9,$6A,$31,$1F,$E9,$38
	.dc.b	$FB,$91,$88,$7C,$3F,$3C,$BF,$D3,$31,$FE,$66,$6B,$FE,$93,$F2,$F8
	.dc.b	$4F,$F0,$9F,$82,$FF,$43,$EF,$FE,$6D,$DB,$FC,$67,$A6,$77,$FE,$9F
	.dc.b	$5F,$0E,$BF,$AB,$78,$3F,$E6,$DC,$BF,$D4,$3C,$7F,$E9,$3A,$F9,$B1
	.dc.b	$FE,$A4,$6B,$FE,$AB,$CB,$FD,$FD,$7F,$DA,$7E,$FE,$BF,$DC,$BA,$79
	.dc.b	$77,$78,$0F,$C7,$77,$FF,$A6,$71,$FC,$F6,$3F,$E5,$F8,$FC,$8C,$63
	.dc.b	$D6,$FF,$2B,$F8,$9F,$E5,$3E,$DF,$F2,$B6,$5F,$E9,$4E,$FC,$63,$FE
	.dc.b	$95,$C3,$F6,$98,$FF,$AA,$B8,$7E,$A4,$7E,$EA,$3F,$CD,$FC,$F8,$F1
	.dc.b	$7F,$81,$DF,$FB,$16,$BF,$84,$63,$FD,$F3,$8B,$FC,$8B,$E4,$7F,$90
	.dc.b	$7F,$1F,$FF,$18,$F9,$FF,$C7,$FC,$E7,$E4,$1F,$E3,$8F,$BF,$FF,$71
	.dc.b	$65,$FC,$87,$D5,$FE,$75,$F9,$0F,$F8,$83,$5F,$F6,$57,$C3,$7F,$42
	.dc.b	$7F,$88,$72,$FF,$A4,$63,$FA,$AE,$FF,$DA,$3E,$17,$FA,$DD,$7F,$85
	.dc.b	$A3,$E6,$7D,$9B,$E1,$BF,$B4,$CF,$FB,$57,$9B,$FB,$0F,$1F,$F8,$06
	.dc.b	$FF,$C2,$BF,$A4,$FF,$80,$6B,$FD,$B4,$FF,$75,$E9,$FF,$5E,$57,$F7
	.dc.b	$5F,$8B,$FF,$09,$FC,$1F,$F6,$CC,$FF,$B7,$7C,$2F,$F8,$5B,$E0,$BC
	.dc.b	$B8,$7F,$B0,$F2,$FE,$BB,$9F,$F6,$03,$FC,$EF,$DB,$89,$D7,$99,$F0
	.dc.b	$FD,$F7,$C5,$7F,$94,$5B,$CC,$E6,$FE,$B1,$D1,$FA,$BE,$F8,$5C,$3F
	.dc.b	$A6,$75,$FF,$52,$EB,$FE,$81,$AF,$F4,$3F,$6B,$FA,$5F,$76,$36,$F6
	.dc.b	$E1,$FC,$FB,$C2,$FE,$E8,$FE,$0F,$C2,$7A,$DD,$7D,$9F,$DE,$47,$E2
	.dc.b	$75,$FE,$EB,$1F,$E3,$6A,$FE,$73,$9F,$E4,$5F,$CB,$3F,$CD,$4F,$F0
	.dc.b	$3B,$F9,$CE,$AF,$F5,$CE,$1F,$C9,$F2,$FF,$34,$7A,$1F,$EC,$4F,$8A
	.dc.b	$C7,$3F,$FB,$B7,$0F,$E4,$CF,$F9,$3F,$4F,$E4,$C8,$83,$C9,$FE,$21
	.dc.b	$3F,$B4,$8F,$E1,$3A,$FF,$83,$3D,$4F,$41,$FC,$8E,$FF,$90,$6B,$F9
	.dc.b	$DB,$FE,$40,$FF,$8D,$9F,$E1,$2F,$FB,$93,$CE,$FE,$13,$F0,$7F,$C7
	.dc.b	$0F,$F7,$FF,$9B,$FD,$F6,$7F,$AA,$6B,$FE,$D6,$C3,$D9,$1F,$EC,$E8
	.dc.b	$C4,$7F,$B8,$A3,$0E,$3D,$7F,$F0,$55,$FC,$98,$DF,$E1,$3F,$D7,$FD
	.dc.b	$2F,$F0,$54,$72,$F6,$FF,$97,$46,$23,$C1,$2F,$6D,$E6,$FA,$A4,$38
	.dc.b	$FE,$4F,$17,$E3,$67,$53,$98,$F9,$DB,$FF,$C0,$1F,$53,$C9,$3E,$AB
	.dc.b	$EB,$18,$8F,$81,$AF,$83,$87,$11,$FE,$F8,$F1,$B1,$C1,$FD,$E6,$3E
	.dc.b	$19,$F9,$0C,$DC,$7C,$23,$CF,$F8,$13,$4F,$5C,$E6,$53,$FC,$5E,$31
	.dc.b	$1E,$DA,$E7,$7D,$E3,$E2,$ED,$F1,$3E,$43,$97,$C2,$FC,$7B,$47,$5F
	.dc.b	$C1,$FC,$11,$8C,$39,$61,$F2,$8F,$97,$3F,$C9,$FD,$6F,$7C,$FF,$93
	.dc.b	$29,$1E,$4E,$1F,$A6,$FC,$77,$FD,$0F,$1A,$9E,$B1,$FE,$F7,$F9,$5F
	.dc.b	$99,$BF,$D6,$F7,$BE,$DC,$63,$93,$F4,$3F,$1D,$8C,$3C,$BC,$9F,$D1
	.dc.b	$B2,$5F,$D5,$8F,$DA,$4B,$E7,$BF,$0B,$E1,$7D,$98,$FC,$86,$98,$7D
	.dc.b	$E7,$E0,$30,$4F,$AD,$B1,$F3,$67,$FD,$C4,$FC,$BF,$6B,$F0,$FA,$FE
	.dc.b	$DB,$5F,$F1,$C7,$1F,$E5,$27,$F1,$79,$FF,$07,$F8,$2F,$EC,$51,$8C
	.dc.b	$3F,$C7,$D1,$FE,$35,$CB,$FA,$9C,$3D,$BF,$F2,$FF,$C3,$FF,$99,$7E
	.dc.b	$1F,$1E,$53,$F0,$9F,$D0,$FC,$3F,$F1,$3B,$DF,$C4,$7F,$25,$C3,$38
	.dc.b	$FF,$92,$E7,$38,$F2,$FC,$33,$32,$FC,$1E,$10,$C7,$E1,$7A,$70,$67
	.dc.b	$5F,$3B,$83,$97,$E1,$9D,$3E,$37,$FD,$05,$F0,$A5,$8E,$7D,$5F,$F4
	.dc.b	$6E,$1F,$02,$76,$3F,$0E,$D5,$FF,$A3,$FF,$0E,$FD,$FF,$9B,$FC,$C7
	.dc.b	$F0,$E7,$E1,$CF,$C3,$91,$86,$C7,$0E,$0F,$F3,$64,$E4,$9F,$23,$98
	.dc.b	$DD,$49,$7E,$0C,$F8,$BF,$F6,$07,$93,$FB,$CF,$0B,$1F,$2F,$FB,$4F
	.dc.b	$02,$E5,$FB,$37,$F0,$15,$FC,$52,$F8,$A8,$CD,$F0,$AF,$D8,$47,$03
	.dc.b	$C1,$F3,$1F,$AD,$BF,$F0,$3D,$FF,$0C,$9F,$F0,$33,$E3,$F8,$D7,$C4
	.dc.b	$78,$DB,$87,$EF,$27,$F7,$9E,$6F,$F4,$D9,$CC,$FF,$16,$8F,$E2,$5F
	.dc.b	$1A,$F8,$BF,$DE,$36,$79,$5F,$9B,$E3,$78,$BF,$C7,$83,$87,$FA,$1F
	.dc.b	$95,$8C,$BF,$83,$4E,$67,$F9,$5B,$FA,$07,$F9,$3F,$C0,$72,$F1,$7F
	.dc.b	$8F,$FC,$F7,$F9,$BD,$BE,$FE,$EB,$FA,$7D,$7F,$9C,$F0,$FA,$53,$D6
	.dc.b	$7F,$8D,$1F,$D0,$DC,$38,$6B,$FD,$23,$1F,$F1,$26,$5E,$E8,$FF,$54
	.dc.b	$7C,$D7,$C1,$8F,$77,$1F,$F8,$99,$D6,$FF,$CE,$7E,$3B,$9B,$FD,$E5
	.dc.b	$EB,$7F,$7D,$B4,$F8,$EF,$B7,$3F,$D9,$5A,$FF,$6D,$8F,$F1,$7B,$FE
	.dc.b	$B9,$BF,$E0,$EF,$F5,$D9,$FC,$5F,$1F,$E1,$DB,$3A,$9E,$27,$5E,$6F
	.dc.b	$E6,$DE,$E7,$FB,$DF,$A7,$F4,$5D,$9F,$A9,$7F,$3F,$19,$8C,$11,$C7
	.dc.b	$5F,$E9,$7B,$3F,$C6,$7E,$0F,$EE,$3B,$EF,$86,$F0,$7F,$80,$36,$7F
	.dc.b	$79,$87,$EF,$E3,$04,$71,$3C,$0F,$E1,$1E,$27,$F9,$03,$C6,$FE,$1C
	.dc.b	$3C,$6C,$3F,$C4,$7C,$7A,$FF,$13,$BF,$E5,$7E,$4F,$F2,$01,$E6,$FF
	.dc.b	$12,$8C,$AF,$2F,$F9,$74,$EB,$CD,$FF,$27,$BC,$67,$54,$F3,$FF,$9D
	.dc.b	$73,$3C,$FF,$F0,$8F,$52,$B5,$3F,$C0,$CF,$B3,$D5,$9B,$FE,$93,$EB
	.dc.b	$7F,$8F,$FD,$2F,$E9,$37,$FD,$42,$FF,$C9,$13,$FE,$16,$7F,$C4,$11
	.dc.b	$9E,$BF,$E2,$4D,$7F,$C4,$31,$FD,$7A,$31,$1F,$D9,$0F,$F1,$2B,$FD
	.dc.b	$A2,$3F,$B5,$9F,$E5,$BF,$90,$F4,$7A,$DF,$DA,$E3,$FC,$BD,$EB,$61
	.dc.b	$FF,$02,$F8,$7F,$DD,$EF,$94,$7F,$88,$71,$FE,$EF,$1F,$DF,$E3,$F6
	.dc.b	$3A,$FF,$8C,$BD,$DF,$89,$EC,$FF,$81,$6F,$FC,$BB,$F0,$F8,$BF,$D8
	.dc.b	$6B,$FE,$39,$C7,$93,$FB,$A8,$FF,$38,$7B,$1F,$37,$DB,$FF,$0D,$65
	.dc.b	$67,$F6,$78,$D7,$FC,$63,$EE,$7C,$8F,$77,$FC,$59,$F0,$16,$7B,$ED
	.dc.b	$FE,$69,$7F,$80,$7B,$FF,$E3,$3D,$7F,$C5,$B1,$FE,$3A,$8F,$F3,$A4
	.dc.b	$7F,$BA,$F5,$FE,$2E,$D7,$F8,$FB,$FE,$6E,$7F,$9B,$35,$FF,$26,$3F
	.dc.b	$E5,$2A,$FF,$05,$EB,$F8,$7F,$96,$FF,$28,$C7,$F9,$63,$CE,$1F,$E7
	.dc.b	$3C,$7F,$8A,$64,$70,$7F,$93,$9F,$F3,$27,$CA,$FF,$99,$F2,$FE,$43
	.dc.b	$F0,$9C,$BE,$57,$FC,$E7,$1F,$C3,$5A,$FF,$26,$EB,$46,$23,$F9,$4F
	.dc.b	$C2,$F3,$7F,$C8,$94,$F7,$BF,$CE,$3A,$FA,$38,$FF,$D2,$FA,$E7,$E0
	.dc.b	$3E,$06,$1F,$E8,$47,$E3,$3F,$E8,$9F,$C3,$7F,$CB,$FB,$FD,$0F,$93
	.dc.b	$AF,$F4,$B9,$A7,$FA,$0E,$FF,$CB,$91,$FC,$D7,$E3,$FF,$DF,$FD,$3F
	.dc.b	$E5,$BB,$FF,$28,$CF,$E2,$AE,$BF,$92,$CB,$FD,$07,$97,$E6,$27,$FC
	.dc.b	$BD,$D9,$FE,$53,$AF,$D1,$9F,$E3,$49,$FD,$4B,$FE,$97,$8F,$F2,$43
	.dc.b	$E5,$8F,$D8,$72,$7D,$7C,$FF,$C8,$13,$FE,$73,$AF,$CE,$C7,$DA,$AF
	.dc.b	$E0,$73,$FD,$F8,$FE,$6C,$FF,$12,$F1,$BF,$88,$F8,$8F,$C3,$BF,$C6
	.dc.b	$1E,$47,$F8,$B7,$EA,$3F,$C4,$7E,$7B,$CF,$FE,$59,$E5,$FF,$86,$7D
	.dc.b	$4F,$E7,$7E,$9F,$F8,$8B,$E3,$7F,$C1,$9C,$5F,$DD,$7E,$0F,$FA,$2F
	.dc.b	$07,$F1,$78,$71,$7F,$83,$A3,$E4,$7C,$1F,$F5,$89,$FF,$06,$9F,$EB
	.dc.b	$FF,$6D,$CE,$F8,$2F,$EC,$93,$FE,$01,$F8,$2F,$EB,$FD,$9F,$E0,$2E
	.dc.b	$EF,$F4,$BF,$EA,$FF,$9A,$6F,$56,$BD,$FE,$38,$7F,$A9,$BD,$FF,$ED
	.dc.b	$39,$FF,$6D,$DB,$F0,$5E,$FF,$F6,$4A,$FF,$68,$3E,$AF,$5B,$FB,$C3
	.dc.b	$FE,$2C,$8F,$F6,$8F,$97,$FD,$5B,$9B,$FC,$7D,$CB,$FD,$53,$DB,$FE
	.dc.b	$99,$D5,$FE,$69,$AF,$F2,$CF,$67,$CD,$7E,$97,$C4,$BE,$27,$FC,$DF
	.dc.b	$DF,$FC,$DF,$6F,$E7,$47,$C9,$D3,$0F,$E2,$BC,$4F,$F7,$C7,$A7,$FB
	.dc.b	$8E,$4F,$E8,$5F,$15,$FA,$59,$FF,$7E,$47,$FB,$EA,$3F,$D7,$77,$AB
	.dc.b	$FF,$49,$9F,$EC,$9E,$DF,$E6,$0F,$ED,$4F,$F4,$56,$9F,$99,$9F,$F0
	.dc.b	$C6,$1F,$CA,$A3,$F9,$3C,$7F,$52,$8F,$F7,$0C,$62,$3E,$2F,$8F,$FA
	.dc.b	$78,$FF,$60,$57,$FB,$5B,$C8,$FE,$1F,$1F,$C8,$2B,$F6,$0D,$7F,$DB
	.dc.b	$6D,$7F,$C5,$BE,$8F,$E1,$23,$F7,$B1,$FE,$30,$EC,$FE,$C5,$F1,$9E
	.dc.b	$E9,$FE,$41,$97,$F0,$6E,$CF,$E3,$37,$FE,$FF,$7F,$D7,$39,$3F,$47
	.dc.b	$AC,$30,$63,$D3,$79,$FF,$C4,$E3,$F9,$27,$A1,$C3,$D8,$56,$B0,$FE
	.dc.b	$51,$5F,$C7,$8E,$F3,$FC,$AE,$F9,$46,$4F,$F6,$5C,$7F,$7D,$F3,$9E
	.dc.b	$87,$2C,$65,$78,$31,$FE,$46,$F8,$88,$7F,$2C,$F0,$16,$DF,$8C,$FC
	.dc.b	$0A,$CC,$62,$3F,$DB,$75,$C3,$7C,$47,$F7,$7B,$7E,$A5,$EC,$E7,$FC
	.dc.b	$81,$F9,$1F,$AD,$F4,$46,$63,$22,$1E,$E7,$F3,$0D,$3E,$46,$97,$1F
	.dc.b	$8B,$9F,$F3,$4C,$38,$8F,$AE,$F8,$9E,$CF,$58,$F0,$C7,$4D,$3F,$1B
	.dc.b	$8F,$DA,$FB,$1F,$A7,$18,$C7,$EF,$C7,$CD,$FA,$2E,$58,$7F,$BF,$5B
	.dc.b	$3F,$EE,$9A,$FF,$06,$5F,$FB,$F8,$C7,$BD,$F1,$DF,$F0,$2C,$7F,$84
	.dc.b	$6F,$F6,$5C,$BF,$80,$B7,$D9,$7F,$44,$9C,$C7,$E1,$3E,$8B,$EA,$FC
	.dc.b	$0F,$D4,$E0,$E1,$C3,$8A,$FB,$71,$F8,$4C,$9F,$B1,$7C,$19,$FF,$80
	.dc.b	$27,$E1,$DC,$3F,$92,$3E,$0F,$85,$F6,$9F,$EF,$55,$FD,$33,$5F,$C1
	.dc.b	$9A,$8F,$C4,$BF,$8F,$E3,$FE,$A3,$A7,$F2,$B9,$FF,$01,$74,$FF,$57
	.dc.b	$C3,$EF,$6F,$FE,$38,$9F,$9F,$F0,$9F,$96,$CB,$F4,$38,$F8,$34,$FE
	.dc.b	$A7,$9F,$A2,$FF,$CB,$47,$E7,$FE,$3F,$CC,$D3,$FC,$21,$EE,$FF,$56
	.dc.b	$1E,$4B,$EF,$FF,$3B,$5F,$AE,$9D,$4B,$ED,$73,$7C,$A3,$F9,$77,$C4
	.dc.b	$7F,$9D,$2B,$9F,$DB,$6A,$9F,$63,$08,$63,$ED,$F4,$E0,$D7,$DC,$E1
	.dc.b	$C0,$FB,$9E,$9E,$BF,$F0,$BE,$2B,$87,$F0,$EE,$2A,$FF,$12,$E9,$F1
	.dc.b	$7D,$2B,$E9,$9F,$4C,$8C,$37,$EF,$F7,$3A,$9C,$8F,$BB,$CD,$71,$9D
	.dc.b	$7D,$E7,$FC,$01,$E8,$F2,$7C,$33,$E6,$BE,$8E,$B9,$D7,$F1,$47,$D6
	.dc.b	$DB,$07,$DC,$7F,$D2,$54,$F9,$FF,$B3,$3F,$DF,$67,$F8,$E1,$FC,$33
	.dc.b	$E2,$3F,$E0,$0E,$3F,$0D,$B1,$F0,$DA,$FE,$A7,$C9,$FE,$57,$18,$8F
	.dc.b	$E5,$71,$FE,$A1,$F8,$7F,$F7,$35,$7C,$D6,$CF,$CD,$F8,$7E,$0F,$EC
	.dc.b	$DA,$FE,$89,$AF,$FA,$AD,$77,$7E,$4B,$91,$18,$9F,$E5,$A7,$F3,$6F
	.dc.b	$17,$0F,$B8,$FF,$4E,$38,$C3,$FA,$57,$8B,$FE,$6C,$F9,$CF,$9B,$3E
	.dc.b	$0D,$71,$7D,$0F,$E1,$1F,$0F,$EA,$7F,$AA,$F0,$A3,$11,$FC,$36,$7F
	.dc.b	$C3,$18,$FF,$1F,$6C,$7F,$1E,$AA,$F2,$F8,$7F,$07,$57,$D7,$8F,$01
	.dc.b	$FE,$7D,$94,$66,$3D,$87,$F0,$87,$F0,$CF,$F4,$FA,$FC,$11,$C3,$93
	.dc.b	$F8,$04,$73,$3F,$DA,$D3,$C1,$E6,$62,$3F,$0F,$E9,$FC,$AE,$AF,$CC
	.dc.b	$F4,$7E,$5A,$30,$FE,$74,$FC,$F7,$57,$FB,$7F,$8F,$FB,$6D,$70,$EA
	.dc.b	$FD,$25,$FD,$4C,$BF,$BD,$69,$BC,$FF,$D3,$51,$FD,$F7,$93,$11,$F0
	.dc.b	$0F,$EF,$D5,$F7,$3B,$7F,$76,$3B,$B6,$E3,$BA,$BF,$C0,$3B,$F9,$F8
	.dc.b	$3F,$C4,$BC,$5B,$E2,$7F,$10,$F8,$EF,$F8,$00,$BC,$4C,$FC,$13,$C6
	.dc.b	$FE,$29,$3F,$C6,$75,$FC,$56,$DF,$C1,$DF,$E5,$2F,$ED,$38,$7F,$DF
	.dc.b	$3E,$4F,$D0,$8F,$E5,$71,$FE,$3D,$7F,$A7,$8E,$FF,$BD,$5F,$EF,$FE
	.dc.b	$EF,$E7,$93,$93,$FD,$FF,$E5,$FF,$44,$7F,$C1,$D2,$8C,$47,$F8,$36
	.dc.b	$30,$FE,$28,$FF,$10,$BF,$D5,$9F,$F6,$44,$FF,$95,$5F,$EA,$F6,$9D
	.dc.b	$47,$F8,$28,$FE,$B7,$D9,$FE,$52,$EE,$FF,$78,$C7,$F8,$AB,$9B,$0F
	.dc.b	$F4,$93,$D9,$5F,$DC,$70,$F5,$C7,$F7,$08,$C4,$7F,$8B,$A1,$E7,$7F
	.dc.b	$C0,$17,$F8,$AD,$FF,$97,$BD,$0F,$EF,$91,$FD,$9A,$DE,$8F,$F8,$12
	.dc.b	$3F,$DC,$58,$FA,$63,$10,$CF,$D9,$EF,$FF,$6F,$F1,$FE,$1B,$63,$FE
	.dc.b	$00,$4F,$F8,$73,$4F,$F0,$E4,$61,$FF,$0F,$FA,$7F,$E2,$6C,$9E,$A7
	.dc.b	$F7,$9E,$1F,$9A,$DB,$F9,$5E,$B7,$F8,$6B,$0F,$F7,$CF,$AD,$E3,$C3
	.dc.b	$FC,$6F,$97,$D2,$9F,$F1,$9C,$F5,$9F,$E3,$C7,$F9,$16,$7E,$D5,$FF
	.dc.b	$9A,$9F,$B9,$C5,$F5,$FD,$8F,$C0,$F1,$7D,$E6,$D3,$E4,$BF,$CB,$73
	.dc.b	$FE,$3C,$6D,$FF,$8F,$1A,$FF,$97,$8F,$E8,$BE,$DF,$F6,$59,$FF,$14
	.dc.b	$3F,$E7,$18,$FF,$35,$1F,$C5,$B8,$FF,$DB,$98,$7E,$4B,$CD,$FF,$6F
	.dc.b	$7B,$5F,$07,$6F,$F6,$07,$AD,$F0,$30,$F9,$DB,$FF,$90,$B8,$7E,$54
	.dc.b	$7F,$93,$6B,$FC,$AB,$C7,$FE,$46,$E6,$FB,$7C,$3F,$E3,$79,$FC,$5F
	.dc.b	$25,$7F,$15,$B7,$F9,$B3,$D8,$FF,$17,$E3,$FE,$B1,$D3,$F5,$95,$FE
	.dc.b	$7C,$AC,$7A,$1E,$58,$7B,$1F,$E0,$FF,$13,$FC,$2F,$E0,$FF,$82,$7C
	.dc.b	$6F,$F0,$4E,$3F,$D3,$3A,$7F,$0D,$87,$F7,$AF,$2F,$F9,$2C,$FF,$24
	.dc.b	$F2,$FF,$BE,$1F,$E0,$5F,$BC,$FE,$EB,$B7,$F8,$FA,$C3,$D4,$FE,$C3
	.dc.b	$F6,$F8,$F5,$7F,$A5,$D7,$F6,$5F,$BA,$F3,$70,$7F,$60,$7D,$18,$72
	.dc.b	$72,$9C,$FA,$DC,$AB,$FD,$3F,$EB,$76,$AF,$E8,$91,$FD,$1C,$F2,$3F
	.dc.b	$CC,$36,$FE,$70,$7F,$87,$F7,$F6,$3F,$E1,$AD,$7F,$9A,$C7,$C2,$3F
	.dc.b	$9A,$70,$7F,$34,$7F,$B8,$F1,$7C,$A5,$C7,$EA,$7E,$9E,$BF,$DD,$31
	.dc.b	$FE,$21,$CB,$F6,$1F,$F6,$8F,$8D,$FE,$8B,$9F,$F5,$8D,$7F,$08,$7F
	.dc.b	$CE,$6F,$F0,$2D,$3F,$55,$5F,$31,$FF,$5E,$F4,$7E,$E6,$3F,$16,$FE
	.dc.b	$FB,$C1,$FF,$62,$4F,$ED,$8E,$33,$FE,$A7,$9F,$D7,$47,$DE,$BF,$D7
	.dc.b	$3F,$E7,$20,$7F,$84,$A7,$F5,$D1,$FE,$7B,$E8,$FB,$BC,$EF,$D8,$BE
	.dc.b	$BB,$FB,$BB,$FC,$4F,$A9,$FE,$C2,$F1,$BF,$4D,$F7,$3C,$3C,$0E,$59
	.dc.b	$FE,$8C,$FF,$5D,$FD,$E7,$4F,$27,$FD,$11,$D9,$DB,$8F,$E4,$6F,$FA
	.dc.b	$1D,$3D,$78,$7F,$B0,$B5,$FE,$87,$1F,$ED,$BE,$1F,$F7,$0F,$8B,$E6
	.dc.b	$F3,$3F,$D0,$F3,$FC,$AA,$FF,$47,$7F,$D7,$38,$FE,$5C,$7B,$39,$BF
	.dc.b	$51,$C5,$FA,$7F,$3B,$FB,$57,$47,$FA,$0B,$CE,$FD,$47,$57,$F9,$DB
	.dc.b	$EE,$FF,$CF,$5D,$5F,$CB,$78,$7F,$51,$EC,$FF,$99,$7E,$F1,$EE,$BF
	.dc.b	$57,$FC,$EA,$F9,$1F,$CF,$F8,$FF,$9D,$E4,$F6,$F1,$7E,$A8,$78,$EF
	.dc.b	$BC,$FD,$2F,$46,$EB,$E7,$BF,$E4,$4A,$FE,$7D,$E4,$7E,$CB,$97,$F8
	.dc.b	$13,$FA,$BA,$EB,$3F,$E5,$0F,$23,$A4,$62,$1E,$47,$57,$F8,$1F,$BB
	.dc.b	$E7,$9D,$3E,$B7,$87,$1F,$EA,$B8,$BE,$03,$FC,$2E,$BF,$DD,$7C,$7F
	.dc.b	$E1,$8E,$5F,$F4,$D7,$2F,$F8,$4B,$D7,$FF,$08,$F3,$FF,$75,$EA,$FA
	.dc.b	$38,$FD,$5C,$FF,$72,$FE,$B1,$FB,$58,$FF,$83,$FB,$BF,$2F,$07,$F0
	.dc.b	$39,$FE,$C2,$3C,$BD,$FF,$8A,$BF,$F7,$A4,$7F,$B9,$FC,$0F,$E2,$F1
	.dc.b	$FD,$82,$3E,$8F,$85,$8F,$0F,$F2,$F3,$FC,$F7,$0F,$A8,$7D,$5F,$13
	.dc.b	$8F,$A3,$E8,$F1,$FF,$8C,$E3,$FE,$F3,$FC,$E1,$F1,$30,$86,$23,$4E
	.dc.b	$0D,$FD,$03,$E2,$7F,$CF,$FF,$14,$21,$3E,$7F,$F6,$F1,$A8,$C9,$BF
	.dc.b	$9E,$FE,$67,$19,$8C,$CB,$CA,$79,$6F,$C0,$3E,$5D,$77,$F9,$FF,$3F
	.dc.b	$0F,$F0,$09,$CC,$A7,$D9,$9F,$FB,$BF,$7C,$CF,$88,$D6,$1F,$C1,$F6
	.dc.b	$FE,$1B,$19,$3D,$9F,$41,$E2,$E1,$E3,$1E,$AF,$93,$C5,$70,$8E,$1F
	.dc.b	$53,$FC,$36,$31,$1F,$03,$E4,$D9,$E6,$7B,$4A,$7E,$9E,$5F,$DD,$E3
	.dc.b	$3A,$6D,$F2,$70,$7F,$79,$BE,$B1,$F4,$3E,$8B,$11,$D3,$E7,$BB,$E1
	.dc.b	$F8,$98,$C7,$2F,$CC,$F7,$BD,$71,$FC,$E2,$7F,$9E,$3E,$4C,$3E,$2C
	.dc.b	$FC,$8C,$BF,$0B,$A7,$1A,$F6,$C7,$E2,$3E,$5F,$EB,$FB,$5F,$3B,$83
	.dc.b	$F1,$7F,$47,$8B,$F8,$71,$F8,$EF,$AE,$C4,$7E,$13,$93,$F4,$EB,$DC
	.dc.b	$CE,$BF,$8C,$9C,$FC,$0F,$F2,$FA,$FC,$59,$F8,$4D,$31,$5F,$92,$8F
	.dc.b	$C1,$EF,$FE,$6C,$8F,$96,$D7,$F0,$DC,$1F,$9C,$9F,$C8,$4F,$F3,$89
	.dc.b	$FC,$E6,$7F,$A5,$D3,$FA,$8F,$57,$E5,$3D,$BF,$C7,$E9,$FB,$EF,$87
	.dc.b	$FC,$A6,$3C,$71,$4E,$5B,$1F,$DD,$32,$FE,$B7,$E5,$7F,$A7,$BE,$0B
	.dc.b	$FD,$41,$F1,$6F,$A2,$FD,$0F,$99,$FC,$2A,$B4,$25,$EF,$6B,$0E,$71
	.dc.b	$FE,$AC,$C3,$9F,$D3,$E6,$5F,$3B,$08,$63,$E9,$B4,$E0,$CE,$9E,$79
	.dc.b	$C4,$71,$FA,$7E,$9F,$40,$FA,$67,$D1,$23,$0D,$CB,$73,$A9,$C9,$C7
	.dc.b	$1E,$CE,$23,$5F,$5D,$C2,$FA,$47,$1F,$A9,$FE,$0B,$EF,$FF,$2E,$F8
	.dc.b	$47,$D5,$BE,$A7,$FA,$D3,$63,$39,$AB,$84,$E4,$D7,$D5,$F4,$E0,$FF
	.dc.b	$07,$DF,$4C,$39,$9F,$D9,$7E,$BB,$FB,$4F,$C5,$7F,$04,$EA,$FE,$43
	.dc.b	$E1,$63,$0F,$DA,$C7,$E6,$7E,$2B,$FA,$A4,$7E,$6C,$FF,$0D,$F7,$7F
	.dc.b	$28,$8C,$CF,$F1,$09,$CB,$FC,$6E,$32,$FE,$32,$33,$DD,$FD,$EB,$3F
	.dc.b	$14,$FF,$25,$AF,$EC,$95,$FB,$BA,$FF,$30,$7C,$36,$AB,$F9,$37,$C5
	.dc.b	$7F,$77,$7F,$89,$CE,$AB,$D3,$0F,$37,$FA,$3D,$6E,$3F,$95,$47,$F1
	.dc.b	$CF,$8A,$FE,$EF,$5F,$C2,$74,$FF,$02,$FC,$57,$7C,$7F,$CD,$53,$D3
	.dc.b	$0F,$E9,$1F,$15,$CD,$FE,$69,$E4,$6A,$79,$4E,$67,$F8,$33,$F9,$37
	.dc.b	$F9,$4F,$A8,$8C,$47,$F3,$9F,$AD,$FF,$3F,$F8,$DE,$F6,$DF,$86,$FF
	.dc.b	$C2,$0F,$C5,$F6,$7B,$5F,$DA,$6B,$FD,$71,$AB,$17,$C5,$FC,$08,$F9
	.dc.b	$2D,$7F,$A5,$6B,$FE,$18,$E1,$FA,$BC,$1F,$D0,$BE,$2F,$AB,$FD,$C9
	.dc.b	$EB,$BE,$F3,$F8,$D6,$F8,$BF,$B7,$8F,$D6,$6B,$D9,$FC,$2D,$7F,$0F
	.dc.b	$9E,$1C,$FF,$C5,$AF,$F0,$F5,$A7,$F1,$B7,$A7,$EA,$5D,$7E,$D7,$C0
	.dc.b	$FE,$9D,$FD,$1F,$67,$E9,$FA,$FF,$45,$39,$9F,$D2,$F4,$FE,$9C,$FD
	.dc.b	$9C,$65,$C5,$7E,$BA,$3F,$7F,$E0,$7F,$03,$3F,$78,$E3,$C1,$FD,$F9
	.dc.b	$FC,$2F,$C2,$FF,$19,$61,$FE,$82,$F0,$FF,$83,$4F,$F0,$4A,$E1,$1F
	.dc.b	$C3,$3C,$4D,$D7,$F1,$8F,$1B,$F9,$0F,$8D,$87,$1E,$3D,$75,$87,$F1
	.dc.b	$FF,$2F,$F9,$7F,$94,$C7,$9C,$63,$CC,$B3,$FC,$BE,$3F,$9A,$79,$8B
	.dc.b	$FE,$85,$C7,$F4,$DB,$A7,$EE,$1F,$E8,$DD,$9F,$D3,$63,$FC,$41,$1F
	.dc.b	$D5,$7A,$3F,$C3,$31,$8F,$85,$FD,$C4,$7F,$92,$63,$FB,$09,$FE,$18
	.dc.b	$71,$CF,$FD,$86,$3F,$C2,$94,$F5,$7F,$C4,$38,$FF,$AE,$9B,$3F,$E5
	.dc.b	$6F,$4F,$D1,$1F,$DE,$7D,$3F,$EC,$17,$C6,$BF,$BC,$C7,$F9,$1F,$D5
	.dc.b	$F7,$7B,$1F,$E0,$58,$FF,$17,$D7,$F8,$2B,$2F,$E4,$71,$FE,$4D,$F5
	.dc.b	$F7,$87,$E7,$2B,$55,$E7,$8F,$83,$F0,$7F,$A6,$F6,$7E,$53,$FA,$8C
	.dc.b	$2C,$FF,$14,$F6,$BE,$3F,$C7,$FF,$20,$F7,$2C,$F5,$8F,$E4,$7E,$EF
	.dc.b	$F8,$C3,$3F,$F1,$66,$7F,$E4,$18,$FF,$30,$C7,$FB,$16,$3F,$85,$B5
	.dc.b	$FF,$21,$72,$7D,$BF,$7F,$F8,$D4,$FF,$91,$F0,$FD,$2F,$C0,$7F,$94
	.dc.b	$39,$7E,$BF,$BB,$F8,$18,$C6,$DF,$E6,$DF,$34,$3E,$5F,$FC,$C1,$A7
	.dc.b	$F9,$9F,$1C,$7B,$5E,$BC,$6C,$7F,$32,$F8,$3E,$3F,$11,$FE,$74,$CB
	.dc.b	$FC,$E9,$CB,$FE,$8D,$AF,$F4,$7F,$C2,$FF,$9D,$E7,$FC,$F9,$F0,$BD
	.dc.b	$B3,$FF,$38,$7C,$4F,$F9,$B3,$7F,$2D,$7F,$A0,$B7,$FF,$68,$F1,$7F
	.dc.b	$45,$BF,$F2,$EE,$5F,$E4,$3F,$A7,$FF,$29,$47,$C9,$FA,$BA,$FF,$99
	.dc.b	$FE,$0B,$E6,$BF,$E7,$EE,$0F,$F7,$65,$FF,$4A,$FA,$9A,$FF,$95,$6F
	.dc.b	$FC,$94,$FF,$93,$EA,$9F,$88,$CF,$FD,$29,$87,$E4,$67,$FC,$E1,$1F
	.dc.b	$8B,$8C,$4F,$F6,$B9,$FE,$D7,$1F,$E3,$D8,$FF,$25,$9F,$CD,$DC,$3F
	.dc.b	$E8,$17,$F5,$D3,$F5,$73,$FF,$1E,$57,$F9,$EF,$9F,$F3,$11,$FE,$F7
	.dc.b	$8F,$F3,$F7,$85,$FE,$3B,$C3,$FC,$59,$E2,$FF,$06,$FA,$AF,$F1,$4F
	.dc.b	$95,$FE,$26,$FA,$DF,$F0,$C7,$9D,$FE,$1E,$F2,$FF,$C1,$99,$7F,$26
	.dc.b	$F4,$BF,$C2,$9F,$59,$FE,$07,$8F,$85,$F0,$7F,$CD,$57,$C2,$7F,$94
	.dc.b	$A3,$F9,$FE,$BF,$E9,$5F,$84,$FE,$A6,$FF,$77,$AF,$99,$F0,$9F,$D4
	.dc.b	$63,$FD,$35,$F0,$7F,$D6,$3A,$7F,$B3,$78,$5F,$E0,$1F,$AD,$FF,$79
	.dc.b	$61,$FD,$CD,$FE,$C7,$7F,$E9,$4A,$FE,$BB,$F0,$BF,$83,$F7,$FF,$AB
	.dc.b	$F5,$FF,$55,$C3,$EC,$7B,$DE,$4C,$3F,$D5,$8F,$A7,$7F,$F3,$4B,$FE
	.dc.b	$18,$8F,$F5,$67,$93,$FD,$0F,$8B,$FA,$57,$BA,$E3,$FE,$6F,$CB,$FC
	.dc.b	$F0,$FF,$31,$57,$F9,$47,$9F,$F9,$D7,$3F,$F3,$0F,$59,$F2,$CF,$8A
	.dc.b	$FF,$34,$F5,$FF,$31,$F8,$BF,$EC,$1C,$3F,$2B,$7F,$F6,$B6,$7F,$CC
	.dc.b	$3C,$1F,$F7,$0F,$A9,$FB,$4F,$07,$F9,$4D,$7F,$A1,$27,$F4,$91,$88
	.dc.b	$F4,$FB,$3F,$EC,$AC,$3F,$DA,$7E,$1F,$F2,$87,$FD,$7F,$E1,$FF,$A6
	.dc.b	$B9,$F1,$1F,$98,$C7,$BE,$7F,$C8,$31,$ED,$1F,$C4,$BD,$6F,$D8,$4F
	.dc.b	$FA,$DE,$7F,$DA,$5F,$06,$EA,$FE,$43,$87,$FB,$23,$C7,$FE,$17,$7F
	.dc.b	$B0,$BF,$F7,$8F,$91,$8F,$53,$99,$FC,$8B,$0F,$EA,$9F,$1D,$FE,$41
	.dc.b	$F0,$7F,$8E,$65,$FE,$33,$C3,$F8,$C3,$FE,$DB,$F5,$3F,$DD,$BE,$77
	.dc.b	$F1,$28,$FF,$6B,$E4,$8F,$F5,$BC,$3F,$5B,$FF,$6B,$BF,$EB,$A7,$FD
	.dc.b	$95,$E8,$7F,$8D,$30,$FF,$7E,$CA,$31,$1E,$8E,$6C,$7C,$57,$F2,$1F
	.dc.b	$3D,$F5,$1F,$C5,$B9,$7E,$B4,$63,$1F,$34,$7F,$AE,$7D,$64,$BF,$12
	.dc.b	$FE,$CE,$DA,$FF,$0E,$BF,$CE,$72,$FF,$78,$F5,$70,$F4,$FF,$CC,$9F
	.dc.b	$17,$CB,$F0,$DF,$E9,$3E,$C5,$C7,$FC,$C7,$ED,$FF,$BD,$A7,$A9,$E7
	.dc.b	$E9,$FF,$38,$FB,$BF,$EC,$A7,$FA,$D6,$9F,$57,$41,$FE,$6B,$DF,$CF
	.dc.b	$F7,$DF,$EF,$2C,$BF,$D9,$B4,$8F,$49,$FD,$B6,$47,$FB,$F1,$C2,$F0
	.dc.b	$FF,$94,$DF,$FB,$62,$3F,$DB,$15,$F8,$FD,$FF,$DE,$0F,$E4,$7C,$2F
	.dc.b	$F7,$7F,$C9,$F0,$BF,$F4,$66,$41,$F0,$7F,$33,$F6,$CF,$D0,$47,$F8
	.dc.b	$3F,$04,$62,$3D,$B5,$B7,$FC,$B3,$3A,$7E,$BE,$4E,$1F,$7F,$8D,$DE
	.dc.b	$78,$E9,$FE,$23,$E6,$FF,$7D,$E3,$E4,$F1,$BF,$DD,$B7,$FB,$C7,$E5
	.dc.b	$ED,$F8,$0F,$8E,$FF,$4E,$61,$71,$FB,$F1,$FE,$E9,$F8,$6F,$F1,$B1
	.dc.b	$FE,$ED,$D5,$19,$F3,$7F,$85,$F9,$9F,$EF,$27,$95,$DF,$37,$F8,$77
	.dc.b	$BD,$FE,$F1,$F8,$7C,$FA,$3F,$E4,$D2,$9F,$EE,$6F,$89,$E3,$E9,$FF
	.dc.b	$1B,$9B,$F1,$9F,$E5,$FF,$8B,$FF,$6D,$ED,$E2,$CF,$FC,$B9,$39,$B7
	.dc.b	$C6,$70,$BF,$F3,$1C,$3E,$4F,$FD,$F9,$F0,$FF,$B2,$C3,$F1,$3E,$47
	.dc.b	$FB,$73,$E3,$BF,$A7,$67,$FE,$F9,$F6,$FF,$CE,$30,$5F,$27,$FA,$8A
	.dc.b	$E1,$1F,$E7,$9F,$93,$7C,$3B,$F6,$DF,$F2,$8B,$72,$FF,$43,$B5,$FE
	.dc.b	$E5,$D9,$FE,$DB,$9F,$AD,$DD,$FE,$D9,$6F,$6F,$4F,$C9,$B9,$9F,$EC
	.dc.b	$5F,$89,$7E,$53,$FE,$5E,$6E,$3F,$E9,$46,$BF,$8E,$C2,$7B,$BF,$ED
	.dc.b	$4A,$FF,$24,$7C,$53,$C1,$FF,$33,$B6,$FF,$D3,$BF,$15,$CB,$0E,$47
	.dc.b	$37,$0F,$F9,$01,$AE,$BD,$AF,$F6,$76,$1F,$ED,$B9,$FF,$61,$FA,$5F
	.dc.b	$ED,$7A,$E6,$7E,$FB,$E5,$BF,$B5,$BF,$ED,$3F,$5F,$67,$47,$D5,$BF
	.dc.b	$F9,$13,$0F,$F3,$EC,$E6,$7D,$CD,$7F,$84,$F8,$FF,$ED,$0D,$7F,$DD
	.dc.b	$9F,$2D,$ED,$D7,$FD,$E1,$F1,$3B,$E0,$FF,$67,$65,$FE,$69,$8E,$58
	.dc.b	$B3,$46,$37,$FF,$67,$71,$7F,$BD,$FE,$24,$FF,$A4,$1F,$BD,$7F,$B0
	.dc.b	$CB,$F8,$37,$2F,$F1,$CB,$FF,$6B,$72,$FD,$EB,$DD,$FE,$E4,$FE,$42
	.dc.b	$E1,$FA,$39,$70,$C3,$FD,$09,$D3,$F9,$2F,$89,$76,$7F,$B3,$A3,$0F
	.dc.b	$E6,$63,$FD,$7B,$D7,$F9,$C3,$F3,$7F,$06,$AF,$E1,$9D,$FF,$A5,$C3
	.dc.b	$FD,$5B,$E6,$7F,$0F,$F0,$BF,$C1,$DE,$17,$FA,$DE,$31,$E2,$FE,$C7
	.dc.b	$C5,$C5,$FE,$D3,$C5,$FD,$8C,$E2,$2C,$30,$FF,$63,$67,$FB,$B9,$FE
	.dc.b	$09,$E3,$EF,$C8,$FE,$0D,$39,$F2,$7F,$D9,$BE,$4F,$37,$FD,$99,$19
	.dc.b	$EE,$FE,$2F,$E5,$F7,$E6,$FC,$69,$F6,$67,$B3,$FD,$69,$79,$BF,$B0
	.dc.b	$EA,$FF,$0D,$F4,$7F,$2D,$EB,$FE,$4F,$5A,$DB,$FC,$BD,$E8,$7F,$4C
	.dc.b	$F9,$7F,$E8,$1E,$87,$FB,$67,$D2,$FF,$42,$FA,$58,$C3,$F9,$D6,$1E
	.dc.b	$2D,$5A,$6E,$3F,$A4,$47,$FA,$EC,$CC,$F0,$70,$7F,$B0,$A3,$F4,$27
	.dc.b	$F4,$1F,$53,$11,$FE,$7B,$8F,$EA,$33,$FD,$67,$3F,$F6,$9F,$C3,$7F
	.dc.b	$B5,$5F,$EC,$7F,$29,$FE,$EF,$F6,$3F,$B1,$E7,$70,$F1,$E3,$F9,$5B
	.dc.b	$FF,$69,$63,$FD,$B6,$7F,$DA,$57,$FD,$C5,$FF,$64,$9F,$3B,$C8,$FF
	.dc.b	$64,$7B,$3F,$ED,$1A,$FE,$F5,$83,$DA,$FF,$7B,$FB,$7F,$EB,$79,$3F
	.dc.b	$EF,$66,$E1,$FE,$08,$CD,$18,$8F,$B7,$EE,$7F,$B4,$BD,$DF,$F0,$8C
	.dc.b	$FF,$85,$71,$FF,$68,$BF,$E1,$9A,$FA,$1C,$DC,$BD,$EE,$DE,$FF,$D6
	.dc.b	$97,$37,$68,$FF,$14,$7A,$5F,$ED,$0F,$92,$FF,$17,$E5,$FE,$12,$F9
	.dc.b	$3F,$EF,$D1,$88,$47,$28,$C7,$C2,$FF,$B4,$31,$FE,$ED,$F0,$9F,$92
	.dc.b	$EA,$FF,$68,$47,$F8,$1E,$3F,$C7,$FE,$67,$F7,$BC,$7F,$91,$DF,$F9
	.dc.b	$1F,$DF,$FF,$1C,$5F,$FA,$F7,$A7,$FD,$C7,$F0,$BF,$E3,$8E,$0F,$F6
	.dc.b	$85,$7C,$1C,$3F,$9A,$FC,$37,$F8,$C7,$0F,$8F,$F1,$16,$7F,$25,$5F
	.dc.b	$2D,$EF,$C1,$FE,$D4,$7F,$C2,$7F,$13,$FE,$F0,$9F,$B3,$D9,$FE,$6B
	.dc.b	$7F,$05,$87,$FB,$45,$B5,$E5,$D3,$FC,$49,$ED,$FF,$8C,$1F,$F4,$3C
	.dc.b	$FF,$8A,$6B,$F5,$8F,$E7,$6B,$FC,$7B,$1F,$E8,$48,$C4,$7F,$B9,$7C
	.dc.b	$2F,$A7,$97,$F8,$73,$0F,$F6,$6B,$FE,$1E,$E7,$FF,$66,$5F,$FA,$DC
	.dc.b	$FE,$05,$E3,$7F,$85,$3C,$8F,$EB,$9F,$17,$FE,$0E,$F2,$C4,$2F,$C6
	.dc.b	$7F,$81,$FC,$EF,$F6,$86,$1F,$C4,$F2,$FF,$4F,$10,$FF,$0E,$94,$FA
	.dc.b	$6F,$F8,$1F,$D6,$FF,$5A,$FC,$6F,$F8,$0B,$DA,$FB,$9C,$1F,$C4,$BE
	.dc.b	$13,$F9,$EE,$FF,$C3,$61,$C1,$FD,$F7,$8B,$FD,$19,$F0,$9F,$D4,$24
	.dc.b	$FF,$55,$E7,$7C,$27,$F5,$99,$FE,$19,$F0,$9F,$D5,$BA,$8F,$E5,$61
	.dc.b	$F1,$EE,$1C,$3F,$DA,$72,$FD,$FC,$FE,$23,$E0,$FA,$3F,$CA,$3E,$0B
	.dc.b	$C0,$78,$FE,$07,$FA,$ED,$7F,$A8,$5F,$5B,$FD,$72,$7D,$71,$FD,$5D
	.dc.b	$FF,$03,$F9,$3F,$17,$E0,$7F,$A9,$1F,$E2,$BE,$4F,$E8,$99,$FC,$B3
	.dc.b	$FD,$6B,$EF,$7F,$97,$5C,$78,$7F,$D2,$7A,$7F,$DB,$DF,$23,$FD,$17
	.dc.b	$B3,$ED,$1F,$6F,$B3,$FA,$27,$B5,$FE,$58,$EE,$FE,$85,$B7,$F8,$3B
	.dc.b	$97,$F9,$EF,$0F,$9E,$3F,$DB,$3D,$3F,$8C,$E1,$FC,$A7,$93,$F9,$9F
	.dc.b	$0F,$FD,$BD,$1F,$94,$AF,$5F,$87,$FE,$79,$7E,$CC,$7E,$77,$0F,$EF
	.dc.b	$AF,$F3,$1A,$FF,$6C,$FA,$9F,$E9,$07,$F9,$74,$FF,$43,$AF,$E3,$BE
	.dc.b	$A7,$E1,$9F,$F1,$AF,$6E,$27,$EA,$C7,$ED,$39,$3F,$91,$FA,$7B,$8F
	.dc.b	$F6,$16,$5F,$BB,$7F,$AF,$B5,$FF,$20,$7A,$5F,$EE,$38,$FE,$0D,$1F
	.dc.b	$E4,$3D,$FF,$92,$61,$FE,$DB,$9F,$E1,$F3,$FD,$07,$3F,$F5,$4E,$3F
	.dc.b	$ED,$9F,$3B,$F8,$EF,$3F,$FB,$4F,$F4,$BF,$F0,$3E,$3F,$ED,$CF,$8C
	.dc.b	$FF,$51,$7C,$72,$73,$3F,$EA,$8C,$38,$3F,$EA,$7F,$84,$FF,$39,$5F
	.dc.b	$F3,$2F,$2B,$F9,$9C,$E4,$87,$E7,$2B,$FD,$AF,$EA,$F4,$BF,$F0,$BB
	.dc.b	$FE,$DA,$FD,$93,$BD,$7F,$A9,$4E,$8F,$E2,$EB,$F7,$19,$EE,$7F,$D7
	.dc.b	$11,$F9,$13,$FD,$C7,$FB,$67,$28,$FF,$57,$5E,$1F,$19,$D7,$DF,$FF
	.dc.b	$74,$FE,$1D,$D6,$FF,$83,$FD,$4F,$AF,$F1,$9E,$57,$FE,$C4,$3F,$DD
	.dc.b	$EE,$36,$D4,$78,$AB,$F8,$77,$D5,$7C,$3F,$DB,$78,$FD,$77,$C5,$FC
	.dc.b	$BB,$B5,$7B,$75,$FF,$7C,$FE,$CF,$F0,$38,$BC,$FC,$68,$F2,$71,$7C
	.dc.b	$3D,$77,$AF,$7C,$53,$FE,$DC,$B7,$E0,$FF,$63,$2F,$74,$62,$FE,$9C
	.dc.b	$6F,$F3,$1F,$81,$1F,$36,$5F,$B6,$7D,$58,$FB,$D4,$9C,$C7,$BF,$0F
	.dc.b	$F0,$D6,$3F,$D3,$FF,$33,$FC,$5D,$FF,$8C,$6B,$F0,$9E,$A7,$C0,$AF
	.dc.b	$E3,$47,$F6,$5D,$BF,$DC,$5E,$E7,$F3,$FF,$28,$BE,$69,$F7,$FE,$1A
	.dc.b	$DC,$7F,$B5,$2B,$6B,$E7,$35,$97,$FB,$EF,$0E,$87,$3F,$B8,$CC,$3E
	.dc.b	$BA,$10,$BC,$1F,$DF,$9C,$1A,$FB,$9E,$3E,$2F,$F7,$2D,$3F,$DE,$35
	.dc.b	$FE,$04,$8F,$F0,$33,$5F,$F5,$2F,$BD,$E0,$F1,$FF,$2D,$3F,$E0,$AF
	.dc.b	$67,$FC,$ED,$C7,$E4,$FB,$47,$D8,$BE,$D5,$18,$6C,$6C,$D4,$64,$7D
	.dc.b	$AE,$67,$67,$2F,$BF,$C4,$EF,$EF,$7E,$E7,$96,$E2,$9C,$5F,$00,$F8
	.dc.b	$71,$98,$C4,$7C,$9E,$2E,$31,$C7,$7F,$6D,$6F,$6F,$6E,$FF,$C2,$EB
	.dc.b	$F3,$1D,$3F,$DD,$3B,$3F,$19,$FB,$6F,$5C,$7E,$2B,$C6,$BF,$DF,$DC
	.dc.b	$5F,$91,$FD,$87,$58,$7F,$4F,$C5,$1B,$AF,$E1,$07,$F1,$39,$FC,$9B
	.dc.b	$FE,$7E,$7F,$B5,$7C,$37,$F6,$E8,$FE,$06,$7F,$06,$70,$FF,$8F,$8F
	.dc.b	$81,$F6,$BC,$3B,$FF,$77,$8F,$F7,$CF,$C7,$BF,$82,$D7,$F9,$4A,$5F
	.dc.b	$77,$F9,$FC,$7F,$C7,$F1,$89,$F8,$31,$FE,$14,$8E,$FD,$7F,$EB,$79
	.dc.b	$FF,$00,$C7,$4C,$2C,$70,$AE,$73,$D5,$FF,$02,$3F,$E5,$4A,$E4,$D9
	.dc.b	$F9,$2D,$7E,$2F,$9F,$FE,$0E,$9F,$4F,$CD,$D9,$ED,$7F,$EC,$BB,$FF
	.dc.b	$2F,$BF,$DF,$7E,$BF,$FC,$2B,$C9,$F0,$2A,$9F,$28,$FF,$72,$D7,$DF
	.dc.b	$7E,$51,$FC,$1A,$3E,$3E,$3F,$E1,$B7,$C3,$7B,$C7,$E1,$3E,$03,$FC
	.dc.b	$6D,$F2,$9F,$E7,$3B,$FE,$7D,$8D,$AF,$F6,$F6,$16,$7F,$02,$32,$C3
	.dc.b	$F8,$4C,$EB,$F8,$6C,$7E,$F9,$FE,$16,$29,$FD,$2F,$3F,$A4,$FD,$68
	.dc.b	$FF,$76,$3F,$6D,$FE,$A1,$C9,$FE,$35,$D7,$EA,$4B,$E2,$5B,$FD,$B7
	.dc.b	$95,$FF,$5D,$F3,$BF,$D3,$31,$88,$FC,$7D,$71,$C7,$F0,$5C,$DF,$E4
	.dc.b	$19,$FE,$D9,$9F,$F5,$CA,$FF,$22,$4F,$E8,$A7,$F7,$5C,$FF,$C3,$23
	.dc.b	$11,$FE,$B2,$9F,$F1,$B4,$39,$FF,$51,$3F,$D4,$E3,$0F,$EE,$E7,$7C
	.dc.b	$7F,$8F,$DB,$FD,$76,$F0,$9F,$ED,$B5,$F7,$CF,$EE,$47,$E0,$7A,$7F
	.dc.b	$83,$F4,$F0,$9F,$E1,$B1,$FC,$2A,$7F,$D5,$7D,$7F,$C6,$7B,$FF,$8D
	.dc.b	$F6,$FF,$B2,$39,$BF,$88,$4E,$7B,$BF,$91,$F7,$67,$2F,$E3,$13,$93
	.dc.b	$F9,$4E,$3D,$31,$FE,$4A,$7F,$26,$EF,$E1,$4E,$1F,$8D,$97,$F0,$CE
	.dc.b	$0F,$F4,$26,$1F,$CE,$7C,$11,$0B,$E0,$F2,$7F,$D8,$91,$B7,$FC,$CF
	.dc.b	$5F,$E3,$D7,$87,$C0,$F8,$38,$3E,$1D,$7F,$4D,$7E,$15,$7F,$5C,$F1
	.dc.b	$5F,$20,$F2,$3F,$AE,$F9,$3E,$8F,$23,$FA,$DF,$92,$F2,$70,$FF,$63
	.dc.b	$8F,$ED,$18,$7F,$BC,$3B,$7E,$C6,$8F,$2B,$F9,$1F,$1B,$C4,$FE,$81
	.dc.b	$DF,$FE,$F6,$F2,$FF,$C5,$DD,$FF,$DF,$39,$B1,$7F,$E3,$AE,$2F,$F1
	.dc.b	$BE,$4F,$33,$FC,$13,$3F,$E2,$18,$FF,$07,$3F,$EE,$6C,$7F,$C5,$D1
	.dc.b	$8A,$FF,$19,$CF,$F4,$BF,$19,$3F,$E1,$F8,$FF,$0E,$63,$FE,$1F,$8C
	.dc.b	$47,$F4,$73,$83,$8A,$FF,$23,$1F,$92,$F4,$3F,$C5,$B3,$F9,$4F,$47
	.dc.b	$F9,$BD,$3D,$3F,$E6,$9E,$9B,$3F,$0C,$7F,$8E,$EE,$9F,$C6,$FD,$2D
	.dc.b	$5B,$D5,$FE,$79,$EA,$E3,$D6,$FD,$A9,$FC,$72,$5E,$BF,$D0,$3F,$81
	.dc.b	$B8,$C3,$F9,$31,$D3,$2F,$F1,$72,$DB,$0F,$CF,$3F,$BA,$73,$7F,$B0
	.dc.b	$7D,$9F,$E4,$D1,$FE,$77,$0E,$1F,$F2,$35,$63,$CB,$FF,$2C,$F1,$FF
	.dc.b	$B2,$7D,$67,$C7,$7F,$16,$7C,$97,$FC,$B1,$F7,$3F,$D3,$FE,$DB,$87
	.dc.b	$06,$F8,$B9,$F4,$7F,$98,$AB,$FA,$A7,$DB,$F5,$9E,$9C,$DF,$E6,$8F
	.dc.b	$5F,$C1,$D1,$DF,$DB,$FF,$2D,$E1,$FE,$0B,$ED,$FF,$2D,$BF,$E7,$9C
	.dc.b	$BF,$C9,$DE,$DF,$F9,$66,$3F,$CC,$7D,$BF,$8E,$7F,$DE,$EF,$A7,$BB
	.dc.b	$E4,$F3,$7B,$5F,$CC,$CF,$F8,$6F,$8F,$FD,$FB,$DF,$FA,$0E,$6F,$F2
	.dc.b	$8F,$27,$F3,$EC,$FF,$CA,$67,$FB,$B7,$D8,$FF,$22,$F8,$5F,$AE,$FB
	.dc.b	$8F,$F2,$3F,$93,$FE,$20,$FB,$9F,$F1,$D5,$FF,$A4,$B0,$FE,$1D,$E9
	.dc.b	$FF,$8B,$BD,$27,$DD,$7F,$93,$AB,$FB,$45,$7F,$23,$C3,$FC,$59,$87
	.dc.b	$F8,$A3,$7F,$E6,$95,$E2,$3F,$CF,$75,$FD,$A5,$E3,$EB,$FF,$15,$7F
	.dc.b	$C5,$11,$FC,$F3,$C2,$FF,$76,$FB,$BF,$D4,$E7,$FD,$44,$BF,$39,$FE
	.dc.b	$A7,$CB,$99,$EC,$38,$7F,$C1,$B9,$7F,$6C,$9F,$EE,$13,$E3,$F8,$FF
	.dc.b	$F0,$24,$FD,$59,$FC,$73,$FE,$99,$9A,$FF,$53,$F1,$7F,$B0,$C7,$23
	.dc.b	$E6,$D7,$F8,$5A,$7F,$C1,$B7,$FD,$9D,$BE,$DF,$9D,$FF,$08,$D7,$F6
	.dc.b	$9F,$7B,$F5,$3F,$01,$F0,$63,$FD,$13,$E7,$7F,$08,$E0,$C4,$0F,$F5
	.dc.b	$1E,$FF,$25,$F7,$E5,$FD,$47,$1F,$9F,$1F,$E5,$19,$FA,$9C,$5F,$E7
	.dc.b	$EE,$2F,$E8,$B3,$FE,$1F,$9F,$BB,$C9,$FE,$6D,$E4,$FE,$93,$1F,$2A
	.dc.b	$7B,$F0,$7F,$99,$70,$FC,$34,$FE,$EF,$1F,$F4,$97,$A1,$FC,$D4,$73
	.dc.b	$FF,$35,$8F,$87,$3F,$CD,$5F,$EF,$A7,$F9,$5F,$0F,$E6,$2F,$C0,$3F
	.dc.b	$38,$FB,$BB,$3F,$42,$FF,$2A,$ED,$FE,$1E,$7E,$9D,$F8,$7D,$DF,$AB
	.dc.b	$AF,$E7,$F7,$FC,$B2,$7F,$BB,$F8,$1F,$B4,$9F,$F4,$17,$81,$FC,$D3
	.dc.b	$C1,$FE,$33,$E1,$FE,$EE,$FF,$92,$D7,$ED,$F0,$FE,$58,$FF,$02,$8C
	.dc.b	$EF,$FE,$32,$F0,$BF,$C2,$BE,$37,$F2,$23,$FC,$C9,$E3,$FF,$86,$3C
	.dc.b	$BF,$F2,$D7,$3F,$F4,$2F,$27,$FC,$53,$F0,$3F,$E4,$1E,$BF,$F1,$36
	.dc.b	$88,$C3,$FE,$26,$6B,$FE,$99,$F7,$FF,$9E,$78,$9F,$E9,$78,$FC,$3E
	.dc.b	$DF,$85,$BF,$E0,$31,$FD,$7A,$3F,$DA,$6F,$0C,$3F,$84,$BF,$EF,$26
	.dc.b	$BF,$12,$BF,$93,$CF,$F0,$73,$F7,$0E,$1F,$F7,$9C,$7F,$BF,$63,$15
	.dc.b	$FE,$94,$D7,$FD,$D5,$9F,$4A,$FE,$E5,$5F,$CC,$1F,$F7,$2E,$7F,$C5
	.dc.b	$3C,$FF,$5E,$DF,$EB,$1C,$BC,$07,$E7,$E3,$8C,$73,$F2,$3F,$DE,$CF
	.dc.b	$A3,$5F,$87,$E5,$F1,$BF,$E0,$07,$FB,$CB,$D1,$CE,$5F,$EC,$18,$EA
	.dc.b	$23,$CB,$3D,$73,$FB,$F5,$F9,$33,$FD,$91,$FA,$47,$D0,$E6,$ED,$AF
	.dc.b	$F9,$5E,$7F,$C1,$DC,$EB,$56,$B2,$FE,$85,$AE,$63,$B1,$FC,$EA,$3E
	.dc.b	$C5,$66,$FC,$D1,$F3,$8F,$B9,$5F,$D2,$9F,$E7,$5C,$3F,$ED,$7F,$49
	.dc.b	$F7,$F9,$9F,$E4,$13,$99,$C3,$FE,$85,$CB,$F4,$92,$8F,$F0,$33,$FC
	.dc.b	$37,$E0,$F9,$CE,$A7,$32,$95,$7B,$FB,$7E,$1E,$5F,$D4,$CF,$CA,$BF
	.dc.b	$2C,$FC,$95,$7F,$9F,$72,$F8,$DD,$DF,$5F,$0F,$F7,$87,$17,$F8,$DB
	.dc.b	$34,$FD,$3F,$2B,$E5,$61,$FD,$6A,$7F,$D3,$3A,$7A,$E7,$F1,$DB,$7F
	.dc.b	$5C,$F0,$7F,$88,$8F,$57,$FC,$9F,$EC,$3E,$43,$5E,$27,$F9,$2B,$DA
	.dc.b	$FF,$29,$7A,$58,$C3,$84,$7F,$95,$7E,$9F,$FB,$2F,$90,$F2,$DF,$98
	.dc.b	$C2,$FE,$6F,$A8,$C4,$67,$87,$FC,$C5,$18,$8F,$F2,$64,$7F,$29,$E8
	.dc.b	$FF,$34,$FC,$40,$85,$8D,$7C,$07,$F9,$DB,$E0,$33,$3B,$FA,$4F,$C8
	.dc.b	$79,$6C,$25,$EE,$B7,$F6,$DF,$DF,$3A,$3F,$E0,$07,$6D,$C3,$F8,$57
	.dc.b	$47,$2C,$7F,$D1,$FF,$0C,$F9,$E7,$CC,$BE,$C1,$F3,$F8,$78,$5B,$D7
	.dc.b	$61,$C6,$F8,$5F,$2F,$C1,$77,$5A,$C7,$94,$F6,$FC,$6B,$FC,$11,$D0
	.dc.b	$3E,$87,$37,$CC,$E8,$CC,$62,$3E,$7F,$3F,$F8,$47,$A3,$E7,$78,$DF
	.dc.b	$E1,$1F,$57,$F9,$27,$1F,$F1,$7F,$85,$F8,$DC,$5F,$EC,$7E,$6F,$A4
	.dc.b	$FF,$16,$8F,$C7,$F8,$BF,$E2,$A7,$F1,$B9,$7F,$B2,$27,$85,$FF,$21
	.dc.b	$F6,$F9,$E3,$DA,$7F,$C8,$15,$FE,$19,$C3,$F5,$D7,$FC,$9F,$DC,$DC
	.dc.b	$62,$B7,$97,$F1,$13,$F3,$7E,$B3,$EF,$7F,$3B,$F0,$5F,$E1,$D9,$FF
	.dc.b	$7F,$36,$BE,$CB,$5F,$C1,$7E,$91,$FE,$07,$8C,$43,$D3,$FF,$42,$47
	.dc.b	$F8,$4E,$3F,$8F,$1F,$EA,$67,$18,$FF,$4B,$8F,$C2,$F9,$9F,$E2,$DF
	.dc.b	$4F,$FC,$5A,$F4,$8F,$F3,$CF,$C0,$7F,$35,$8C,$78,$1F,$E3,$0B,$FE
	.dc.b	$5E,$7F,$86,$63,$F7,$35,$FD,$64,$D4,$7F,$4C,$7E,$AF,$83,$8D,$3F
	.dc.b	$A6,$57,$09,$FF,$36,$35,$FD,$07,$A7,$B7,$FB,$27,$A7,$FE,$3C,$C3
	.dc.b	$C3,$9D,$7F,$A2,$9F,$E2,$EB,$FE,$DB,$5F,$EC,$66,$BF,$CB,$1B,$45
	.dc.b	$F5,$5E,$FF,$BB,$FE,$A9,$66,$C6,$39,$3F,$B1,$65,$FE,$39,$7F,$B2
	.dc.b	$61,$71,$FE,$7F,$9F,$C3,$D3,$11,$FD,$9B,$2F,$F3,$CE,$98,$7E,$5D
	.dc.b	$7F,$93,$E7,$EC,$9F,$D0,$A3,$E2,$3F,$CD,$9B,$EE,$6D,$FB,$91,$FE
	.dc.b	$95,$77,$79,$7E,$29,$FE,$88,$D7,$FB,$D4,$FF,$96,$CE,$FB,$59,$E7
	.dc.b	$1F,$EE,$E7,$F0,$C7,$F9,$CB,$7E,$3F,$07,$F8,$C8,$FB,$D3,$FE,$32
	.dc.b	$8C,$65,$F0,$5F,$CB,$F1,$7F,$B6,$5F,$CC,$F1,$FF,$98,$5F,$C6,$9F
	.dc.b	$91,$CD,$F1,$FF,$A7,$C3,$E5,$9F,$EF,$4C,$3F,$A2,$F4,$FE,$37,$2F
	.dc.b	$3C,$FF,$B8,$3A,$3F,$67,$5F,$C9,$2B,$F7,$0F,$F0,$4D,$7F,$CE,$55
	.dc.b	$FC,$B6,$7F,$7C,$FF,$93,$BA,$B8,$F1,$FF,$2A,$9F,$F3,$A5,$FE,$76
	.dc.b	$FF,$CE,$98,$FE,$88,$FF,$3B,$4F,$2C,$3F,$DE,$3D,$7C,$9F,$F6,$44
	.dc.b	$7F,$1B,$EE,$FE,$40,$D7,$FA,$87,$6F,$F9,$FF,$BF,$F9,$37,$7F,$F4
	.dc.b	$D9,$57,$F4,$DF,$13,$F9,$77,$8D,$FC,$C8,$78,$7A,$DF,$F9,$1C,$EB
	.dc.b	$C8,$FE,$7B,$4F,$1F,$5A,$FF,$44,$F1,$FF,$9B,$CF,$F4,$3C,$BF,$AD
	.dc.b	$1E,$BE,$6F,$F4,$4F,$0F,$F1,$DB,$FF,$37,$CF,$F5,$AF,$2D,$E6,$FF
	.dc.b	$5A,$F3,$1E,$67,$6B,$FE,$C7,$1F,$D8,$BC,$DC,$5F,$F5,$B8,$FE,$C3
	.dc.b	$87,$F6,$CE,$2F,$8B,$C3,$FE,$95,$E4,$FE,$89,$1F,$DD,$B9,$7A,$47
	.dc.b	$C6,$8F,$EE,$55,$FD,$C7,$D1,$7A,$3F,$E0,$8F,$57,$FC,$0F,$E8,$F8
	.dc.b	$E7,$3E,$9F,$F8,$1F,$D5,$FF,$05,$61,$C7,$0E,$1E,$A7,$F8,$2A,$3F
	.dc.b	$63,$DD,$FE,$A1,$E2,$FD,$AF,$35,$E6,$FC,$15,$7F,$87,$A7,$F2,$E7
	.dc.b	$F2,$07,$0F,$FA,$FE,$3F,$C6,$0F,$FA,$04,$4F,$FA,$8D,$FE,$13,$1F
	.dc.b	$E7,$E8,$C4,$78,$25,$C1,$FE,$36,$87,$91,$FE,$A6,$F8,$2F,$F3,$47
	.dc.b	$83,$FE,$3E,$F6,$7F,$C9,$78,$7F,$90,$71,$F1,$47,$F9,$02,$3F,$C8
	.dc.b	$98,$7F,$A4,$67,$F1,$F7,$FE,$64,$BF,$F2,$A7,$9B,$8A,$FF,$2D,$F9
	.dc.b	$9F,$CF,$7C,$DE,$5E,$D7,$F9,$87,$6F,$F2,$96,$FF,$EB,$3B,$FD,$CC
	.dc.b	$7F,$98,$34,$FF,$35,$C7,$C6,$3F,$D7,$B5,$FE,$6D,$7F,$D7,$75,$C5
	.dc.b	$FE,$50,$7F,$AF,$B4,$FF,$5D,$DF,$F1,$6F,$33,$E4,$E1,$FE,$7E,$8F
	.dc.b	$E0,$73,$A7,$F9,$0F,$B9,$AF,$73,$FC,$06,$B1,$FD,$6F,$BB,$FE,$91
	.dc.b	$BB,$97,$94,$EB,$1F,$E0,$AF,$7A,$C7,$C2,$C7,$EE,$47,$3C,$BF,$D5
	.dc.b	$33,$FE,$BB,$E3,$FD,$CF,$BF,$FC,$60,$FF,$50,$4F,$FA,$C7,$83,$F0
	.dc.b	$98,$7E,$16,$1E,$C7,$F1,$E3,$78,$7F,$9C,$27,$FD,$C6,$FF,$9C,$B8
	.dc.b	$FF,$19,$F1,$7F,$D8,$3E,$03,$F1,$6D,$93,$FD,$93,$E2,$97,$FE,$D5
	.dc.b	$F0,$FF,$D8,$F1,$FE,$89,$F7,$3F,$DA,$53,$FE,$D1,$26,$3F,$EC,$2C
	.dc.b	$3E,$04,$7E,$6E,$3F,$D7,$31,$F6,$EB,$AD,$7D,$3A,$FE,$8B,$EB,$63
	.dc.b	$9F,$EF,$5F,$D1,$3F,$D4,$C7,$32,$57,$FA,$73,$6F,$C7,$70,$7F,$A8
	.dc.b	$38,$FF,$32,$FF,$A5,$EB,$F2,$67,$F9,$9F,$93,$F5,$D7,$FA,$68,$FD
	.dc.b	$5E,$5F,$EA,$53,$CB,$F0,$3F,$E9,$78,$FF,$37,$69,$C3,$D9,$FF,$23
	.dc.b	$D7,$C3,$DF,$F8,$97,$A7,$FE,$7F,$FC,$DB,$FC,$CB,$E5,$FF,$19,$F2
	.dc.b	$FF,$CF,$DF,$33,$FE,$73,$C7,$FD,$DB,$87,$F3,$2F,$82,$FF,$34,$7A
	.dc.b	$9F,$CC,$BE,$DF,$FC,$DB,$A7,$F8,$03,$83,$FA,$17,$9D,$FE,$92,$D7
	.dc.b	$FC,$BD,$3F,$EB,$BC,$BF,$DE,$FE,$C7,$93,$A7,$FD,$A9,$EC,$FF,$5E
	.dc.b	$9F,$F7,$DF,$B5,$6F,$F1,$7F,$01,$FD,$BA,$FF,$C3,$DE,$F7,$1B,$FF
	.dc.b	$10,$73,$7D,$AB,$FC,$1D,$7F,$8A,$FA,$FF,$27,$1F,$EC,$5A,$FF,$14
	.dc.b	$D7,$F8,$BE,$BF,$C7,$3E,$FF,$F8,$2A,$7F,$C4,$D3,$FA,$73,$F1,$BF
	.dc.b	$01,$FE,$0E,$BF,$F1,$6C,$65,$FF,$19,$F6,$7F,$88,$5F,$F0,$4C,$7F
	.dc.b	$BE,$F9,$7E,$8F,$9B,$FE,$19,$CF,$C7,$9F,$F8,$56,$FF,$C3,$67,$FB
	.dc.b	$7F,$E0,$3D,$DE,$2F,$F7,$ED,$3D,$2F,$F9,$F4,$F6,$B8,$E2,$F9,$50
	.dc.b	$7E,$45,$7E,$C5,$FB,$FC,$9F,$EC,$0E,$4F,$EB,$53,$FE,$36,$AF,$EE
	.dc.b	$5D,$1F,$73,$E0,$FF,$D5,$7D,$1F,$D9,$B4,$FF,$26,$5F,$EF,$2F,$FB
	.dc.b	$37,$B7,$FD,$97,$AF,$FA,$CF,$B3,$FD,$73,$46,$9F,$D2,$F0,$F6,$E3
	.dc.b	$FE,$00,$F0,$7F,$2D,$C8,$F0,$7F,$83,$7A,$5F,$C0,$6F,$F7,$7A,$FE
	.dc.b	$7E,$BF,$A7,$72,$7F,$90,$E7,$F9,$C4,$7E,$DF,$C4,$FD,$4D,$7E,$B7
	.dc.b	$2F,$F5,$4F,$A1,$FD,$2A,$3F,$9A,$D7,$EC,$EF,$FD,$11,$18,$7F,$84
	.dc.b	$E1,$FC,$2A,$7F,$60,$7F,$2F,$BF,$F5,$9A,$F1,$FC,$BE,$6F,$F0,$EB
	.dc.b	$FE,$47,$3F,$91,$55,$FE,$1E,$8F,$63,$FC,$63,$AB,$F8,$AA,$F8,$57
	.dc.b	$C6,$FF,$B0,$27,$FD,$21,$4D,$3F,$DD,$15,$A8,$FF,$75,$4E,$BB,$7F
	.dc.b	$90,$CE,$67,$FD,$8E,$7F,$BB,$EB,$FB,$9F,$EC,$EF,$91,$ED,$3C,$8F
	.dc.b	$F7,$A7,$6F,$F2,$A9,$D4,$64,$63,$E5,$E1,$78,$CF,$C3,$B9,$D7,$FA
	.dc.b	$16,$30,$FF,$BF,$F7,$FD,$89,$FB,$48,$CC,$62,$3F,$8E,$FE,$DB,$CA
	.dc.b	$3F,$BC,$4A,$31,$1F,$ED,$8F,$CB,$7F,$DC,$B5,$D3,$D4,$FF,$07,$DB
	.dc.b	$F1,$ED,$DF,$C9,$3C,$EE,$37,$DB,$CA,$FF,$82,$FA,$DC,$BF,$63,$E3
	.dc.b	$EA,$F4,$F1,$35,$5F,$85,$C4,$78,$3F,$01,$F4,$70,$FF,$B1,$34,$47
	.dc.b	$0E,$8F,$F5,$FC,$62,$BE,$25,$FF,$A5,$39,$FF,$A7,$C7,$CE,$3A,$F1
	.dc.b	$FE,$43,$D4,$CE,$5F,$ED,$09,$F0,$9F,$E2,$6C,$1F,$8C,$F9,$A3,$D9
	.dc.b	$0F,$C7,$BE,$AD,$7D,$98,$C4,$FF,$8F,$31,$FE,$9B,$F8,$BF,$E1,$32
	.dc.b	$EB,$97,$8A,$FF,$65,$9F,$D7,$D3,$F0,$3E,$1F,$CE,$CB,$E8,$BE,$4F
	.dc.b	$84,$FF,$29,$1F,$E5,$8F,$9B,$DC,$FC,$88,$CF,$C1,$BE,$1F,$32,$9C
	.dc.b	$F7,$E3,$CA,$10,$86,$34,$FF,$29,$B8,$35,$F4,$1D,$3D,$B1,$08,$62
	.dc.b	$7F,$CB,$6E,$0D,$79,$DC,$5C,$0F,$A8,$C3,$BF,$9E,$F4,$F9,$7F,$9B
	.dc.b	$E2,$F5,$FC,$97,$EE,$B6,$F8,$A7,$E2,$3E,$45,$EA,$3E,$71,$18,$6C
	.dc.b	$7C,$48,$C1,$3E,$7F,$11,$C6,$33,$F4,$38,$AE,$27,$4F,$A1,$FD,$BF
	.dc.b	$9F,$FE,$91,$E3,$F8,$C6,$32,$FC,$DB,$81,$C7,$98,$FF,$5B,$EB,$FB
	.dc.b	$7E,$3F,$E1,$DB,$7B,$7B,$35,$3F,$85,$97,$CE,$FE,$5B,$87,$E2,$46
	.dc.b	$30,$FD,$4E,$1C,$61,$C3,$BC,$3B,$3F,$9E,$9E,$DF,$01,$FD,$0F,$C1
	.dc.b	$BE,$AF,$EF,$BD,$6F,$F2,$0C,$E7,$2F,$E6,$E3,$C8,$FF,$49,$C3,$F1
	.dc.b	$0F,$CC,$6E,$F4,$DF,$35,$FB,$ED,$BF,$CB,$3E,$26,$63,$11,$FE,$5C
	.dc.b	$9E,$D9,$7F,$36,$6B,$FE,$8F,$7F,$4F,$1F,$C8,$2B,$B6,$9C,$A3,$FD
	.dc.b	$51,$3F,$EA,$AD,$7F,$0D,$7E,$3F,$13,$CB,$39,$8F,$E8,$7E,$2E,$5F
	.dc.b	$E6,$87,$F9,$CE,$3F,$D6,$77,$FC,$AF,$DF,$D4,$FE,$4F,$D9,$F7,$74
	.dc.b	$FE,$DE,$29,$C8,$FD,$FD,$FF,$49,$CB,$F3,$1E,$4F,$C5,$FB,$0E,$12
	.dc.b	$FA,$5E,$59,$DC,$39,$1F,$CD,$2B,$38,$FF,$05,$D3,$FC,$EA,$FF,$07
	.dc.b	$C6,$D7,$03,$FA,$36,$BF,$17,$06,$BF,$E7,$C8,$C6,$B6,$B8,$63,$5F
	.dc.b	$EB,$E7,$F3,$4C,$31,$AF,$E5,$71,$F9,$53,$FD,$B1,$FA,$A7,$F8,$C8
	.dc.b	$A7,$F5,$6D,$BF,$0A,$FD,$C8,$FF,$44,$F0,$79,$78,$3F,$D1,$EF,$9E
	.dc.b	$5F,$03,$F8,$DA,$F6,$71,$63,$0F,$E9,$6D,$8F,$8F,$CD,$FE,$98,$AE
	.dc.b	$D1,$FD,$03,$AB,$F4,$6F,$E9,$8F,$F0,$1D,$23,$31,$FB,$4C,$3F,$B6
	.dc.b	$C7,$EB,$61,$D1,$FB,$58,$C1,$FC,$54,$FF,$03,$4F,$F3,$B7,$F7,$73
	.dc.b	$D9,$FF,$05,$1B,$E2,$FE,$11,$DB,$FC,$C3,$B7,$C5,$18,$8F,$E6,$72
	.dc.b	$E3,$FF,$43,$77,$FF,$12,$F0,$5E,$0F,$F9,$07,$2F,$E4,$5E,$17,$F1
	.dc.b	$93,$F8,$D7,$85,$98,$E1,$19,$C7,$A4,$70,$C3,$A7,$81,$BD,$FD,$D9
	.dc.b	$FF,$09,$DF,$E1,$BF,$D9,$BC,$4F,$E7,$23,$C4,$EC,$E3,$C5,$AE,$B3
	.dc.b	$FE,$86,$FF,$40,$F1,$BC,$9C,$3F,$C9,$78,$EB,$FC,$9E,$BC,$BB,$7F
	.dc.b	$59,$F2,$BF,$AE,$7C,$93,$E4,$F1,$E6,$FB,$7C,$CF,$EB,$B8,$70,$7F
	.dc.b	$B1,$F9,$BF,$07,$C0,$FB,$1D,$DF,$8F,$A7,$9D,$FD,$D7,$4F,$F2,$AF
	.dc.b	$6F,$E0,$FC,$FF,$F0,$D7,$85,$FE,$04,$F0,$FF,$DE,$11,$FE,$00,$AF
	.dc.b	$F2,$3D,$7E,$27,$D0,$FF,$03,$47,$F8,$4B,$17,$BB,$FE,$3A,$AF,$F0
	.dc.b	$84,$62,$7F,$C8,$51,$FF,$00,$63,$FC,$35,$1F,$C9,$8F,$F3,$7B,$8E
	.dc.b	$AF,$F7,$04,$7E,$CE,$37,$AF,$F9,$A8,$FF,$0C,$D7,$E6,$FD,$3F,$E6
	.dc.b	$15,$8A,$7A,$9F,$A8,$7F,$C7,$91,$FE,$28,$C3,$FD,$FE,$7F,$C0,$0F
	.dc.b	$55,$87,$E9,$7D,$6F,$F1,$F5,$53,$F8,$AF,$AF,$AC,$1E,$CF,$F3,$3F
	.dc.b	$67,$1E,$D7,$F4,$49,$FF,$28,$5F,$F0,$E7,$FC,$AF,$9F,$E9,$FD,$BF
	.dc.b	$E0,$11,$FE,$5A,$DF,$FB,$17,$C2,$7F,$9A,$2F,$FC,$74,$48,$F2,$C6
	.dc.b	$5F,$F3,$3F,$AA,$C1,$F1,$BF,$CC,$6F,$FA,$CE,$BF,$E4,$E8,$C4,$7F
	.dc.b	$A4,$A3,$DF,$18,$8F,$C0,$4F,$F9,$D3,$F6,$DF,$E2,$9E,$DF,$A2,$3F
	.dc.b	$CE,$3E,$AF,$F9,$E3,$83,$FA,$97,$CF,$F8,$6B,$E1,$E1,$F3,$6F,$FD
	.dc.b	$CF,$1F,$42,$3F,$96,$F5,$6F,$8F,$FA,$D7,$D2,$59,$E1,$7F,$50,$FF
	.dc.b	$42,$F3,$FF,$75,$7F,$D0,$FD,$1E,$4F,$80,$FF,$32,$57,$F9,$3B,$2F
	.dc.b	$C4,$57,$F8,$37,$AB,$F2,$53,$FE,$6B,$EB,$FC,$BB,$FE,$89,$7D,$BD
	.dc.b	$9F,$9D,$8F,$F6,$AD,$FF,$97,$1F,$C8,$EB,$FE,$8F,$9F,$87,$DD,$FA
	.dc.b	$99,$FC,$3E,$3F,$07,$BF,$FA,$6E,$1F,$B3,$F1,$7F,$73,$E2,$7E,$EF
	.dc.b	$E5,$FF,$CA,$5E,$3F,$F8,$A3,$E4,$3F,$C7,$F8,$FF,$7D,$C3,$F8,$0F
	.dc.b	$AB,$FE,$36,$F4,$FF,$C6,$9F,$03,$FE,$5C,$AF,$F4,$F7,$57,$F8,$AF
	.dc.b	$1F,$F1,$76,$3F,$E2,$DF,$07,$FC,$51,$EB,$F8,$B0,$FE,$35,$87,$F8
	.dc.b	$4B,$3F,$E7,$F5,$FE,$71,$F8,$0F,$F0,$BF,$0F,$D0,$AF,$D1,$C3,$F1
	.dc.b	$2F,$EA,$9E,$16,$AB,$FC,$04,$FF,$5E,$AF,$F7,$5B,$FE,$08,$7F,$D6
	.dc.b	$BF,$49,$FE,$08,$7F,$CB,$51,$F8,$D9,$FD,$BA,$FD,$DB,$AF,$C0,$FF
	.dc.b	$61,$A3,$63,$1A,$FF,$6A,$F5,$F4,$7F,$7D,$C7,$F4,$1C,$FF,$DF,$30
	.dc.b	$F3,$1F,$1B,$DD,$F5,$67,$ED,$8F,$EC,$DE,$FF,$EC,$3D,$BF,$8B,$1F
	.dc.b	$C0,$BD,$1F,$EC,$9C,$78,$8F,$D6,$47,$F5,$6B,$E5,$CD,$5F,$ED,$13
	.dc.b	$FE,$3E,$E3,$FF,$74,$73,$7D,$47,$E2,$47,$E6,$C5,$7F,$55,$E8,$FB
	.dc.b	$DE,$E7,$F9,$E3,$2F,$EB,$17,$FD,$AF,$83,$FB,$26,$3F,$E9,$8E,$8F
	.dc.b	$E8,$1D,$9F,$8C,$F5,$0B,$B3,$FB,$E7,$1F,$F8,$7B,$1F,$E7,$FB,$FF
	.dc.b	$BD,$7D,$2F,$E4,$FC,$1F,$C1,$E7,$F3,$CF,$F7,$BF,$03,$FA,$0D,$7F
	.dc.b	$88,$61,$C5,$FD,$16,$3F,$DF,$5E,$77,$FB,$23,$AF,$97,$87,$FC,$68
	.dc.b	$FD,$6D,$F0,$C7,$FA,$05,$FF,$1A,$CB,$FD,$9E,$FF,$29,$F1,$7F,$90
	.dc.b	$9F,$BA,$7F,$DC,$7F,$01,$FC,$ED,$FB,$17,$FE,$17,$F2,$3F,$81,$61
	.dc.b	$FC,$2E,$3F,$C2,$5E,$57,$F0,$A7,$F9,$B5,$7F,$0E,$7F,$C3,$DE,$67
	.dc.b	$F2,$E3,$F8,$BB,$FE,$21,$F4,$7E,$37,$9D,$FE,$31,$FC,$0B,$FA,$0D
	.dc.b	$FF,$3E,$0E,$1E,$8D,$E0,$FF,$42,$C3,$D1,$FE,$93,$F1,$0F,$8E,$FE
	.dc.b	$19,$5F,$EF,$E8,$C5,$7C,$A3,$FE,$00,$F1,$FF,$39,$8C,$C7,$F3,$F3
	.dc.b	$FB,$7A,$FD,$1F,$FD,$1E,$D8,$FC,$E7,$2A,$31,$1F,$EC,$9D,$71,$FB
	.dc.b	$D7,$F9,$DE,$4F,$23,$FC,$D8,$BE,$CB,$FA,$D5,$73,$7F,$AE,$73,$75
	.dc.b	$FB,$3E,$47,$F6,$4A,$2A,$FE,$CB,$3B,$7F,$86,$7B,$7F,$06,$7F,$D3
	.dc.b	$0F,$C6,$FC,$47,$84,$FF,$9C,$3E,$33,$E3,$47,$F9,$7B,$07,$E4,$9F
	.dc.b	$DF,$2B,$B6,$3E,$F8,$C4,$7C,$DF,$D2,$39,$F0,$FF,$81,$63,$76,$9C
	.dc.b	$C4,$38,$1F,$E0,$3D,$FF,$B2,$4E,$E3,$07,$F8,$8D,$7E,$A5,$FD,$32
	.dc.b	$3F,$2F,$18,$7F,$C3,$A6,$FE,$07,$FC,$4B,$97,$BA,$51,$FA,$6E,$1F
	.dc.b	$B1,$F0,$5F,$86,$BA,$7E,$14,$67,$F8,$C8,$FD,$AC,$7E,$3C,$FC,$63
	.dc.b	$8F,$6B,$FC,$37,$87,$F8,$16,$7F,$31,$DF,$E1,$F9,$0F,$EB,$0F,$C4
	.dc.b	$C7,$FD,$25,$FA,$57,$ED,$3C,$8F,$E1,$7A,$3E,$17,$FC,$A1,$1F,$C3
	.dc.b	$B4,$56,$FC,$9F,$E2,$47,$B1,$C5,$F1,$F7,$3F,$CB,$59,$F0,$D3,$F8
	.dc.b	$DF,$98,$30,$FE,$25,$FF,$32,$5F,$F0,$6F,$88,$FF,$37,$9A,$9B,$F1
	.dc.b	$16,$3F,$CE,$13,$B4,$42,$7C,$57,$1F,$23,$F9,$97,$07,$F9,$02,$BF
	.dc.b	$89,$FC,$5F,$EE,$B1,$FF,$20,$C7,$F1,$6F,$53,$F9,$A8,$F8,$0F,$13
	.dc.b	$8F,$5E,$C7,$FA,$47,$D6,$4F,$FA,$5A,$B9,$4E,$61,$F1,$9F,$2B,$AF
	.dc.b	$E1,$F0,$5C,$BE,$13,$F1,$4F,$F5,$36,$1F,$C4,$75,$F4,$EB,$FE,$3F
	.dc.b	$8F,$E2,$9A,$FA,$70,$E5,$86,$AB,$F9,$CD,$3A,$3F,$AE,$47,$17,$F6
	.dc.b	$37,$C6,$5D,$5A,$AF,$BE,$EA,$FF,$D9,$32,$F1,$FF,$55,$1F,$CF,$31
	.dc.b	$FC,$2C,$7F,$89,$5F,$9C,$FE,$42,$3E,$6F,$EA,$3C,$1F,$F3,$CF,$C1
	.dc.b	$7F,$96,$5F,$F6,$DE,$2F,$C5,$FF,$A8,$C7,$F1,$E6,$BF,$C7,$5E,$0F
	.dc.b	$FB,$E0,$FE,$24,$E3,$BF,$E2,$3A,$9F,$93,$E0,$F5,$7F,$DE,$1F,$82
	.dc.b	$FF,$07,$BF,$E7,$6D,$7D,$8D,$7F,$8B,$EE,$8C,$47,$F2,$1D,$B0,$FF
	.dc.b	$1C,$AF,$AF,$F2,$58,$8F,$F0,$4D,$7F,$49,$7F,$20,$D7,$FB,$29,$F5
	.dc.b	$DC,$3F,$DA,$2B,$C1,$F2,$BF,$26,$3F,$94,$CB,$4C,$47,$F3,$3B,$C3
	.dc.b	$FC,$D7,$4F,$CF,$28,$5F,$08,$C6,$FC,$B3,$B1,$C4,$FE,$BF,$96,$1F
	.dc.b	$A0,$73,$3E,$06,$7F,$AE,$C3,$11,$FE,$28,$7E,$3B,$F9,$36,$CE,$A3
	.dc.b	$ED,$5D,$C7,$CF,$F0,$3F,$67,$0F,$DC,$EA,$8C,$47,$F9,$40,$F8,$35
	.dc.b	$97,$D9,$1F,$88,$7F,$02,$D7,$F5,$7C,$7F,$87,$C3,$E8,$FC,$9F,$FA
	.dc.b	$AF,$F0,$27,$17,$47,$F4,$13,$FE,$2B,$3F,$DD,$2E,$1F,$E6,$E7,$F8
	.dc.b	$1F,$1F,$F6,$6C,$67,$3F,$DB,$3F,$AE,$9F,$ED,$1C,$B8,$D7,$F6,$13
	.dc.b	$FB,$79,$C9,$FE,$0E,$B4,$66,$3F,$CA,$39,$FF,$A9,$E3,$FC,$2C,$FE
	.dc.b	$79,$FE,$0B,$1C,$5F,$E1,$51,$FC,$29,$FF,$0E,$73,$7F,$0F,$27,$C9
	.dc.b	$76,$BF,$E2,$B3,$FC,$6B,$AF,$F8,$EF,$59,$FE,$3F,$F2,$9E,$4C,$BF
	.dc.b	$BD,$F6,$7F,$28,$EC,$C4,$7F,$81,$63,$F9,$5D,$E2,$39,$DC,$97,$6E
	.dc.b	$77,$F8,$F5,$F1,$0F,$F0,$3F,$7F,$F3,$D2,$77,$F3,$70,$7F,$41,$F9
	.dc.b	$4C,$38,$3F,$A2,$EB,$FD,$1D,$DF,$83,$E3,$E1,$F8,$47,$F3,$1E,$1C
	.dc.b	$38,$3F,$AC,$57,$F5,$EF,$23,$FB,$2F,$8C,$F1,$FD,$9E,$47,$F6,$A9
	.dc.b	$CF,$8D,$8A,$FE,$CF,$E4,$7F,$BD,$BA,$3F,$4E,$7F,$A1,$7B,$FF,$DF
	.dc.b	$35,$FC,$9F,$86,$9F,$8F,$3F,$09,$D3,$FE,$FE,$CF,$FC,$07,$EA,$FF
	.dc.b	$8A,$BE,$47,$C5,$5F,$E3,$3C,$7F,$D3,$2F,$F9,$ED,$E0,$FF,$83,$7B
	.dc.b	$7F,$C7,$13,$C2,$3F,$CA,$07,$F8,$43,$04,$7F,$86,$23,$A5,$7F,$30
	.dc.b	$9F,$06,$7F,$E1,$BA,$FF,$11,$57,$F9,$FC,$FE,$4E,$E2,$FF,$C7,$F5
	.dc.b	$FC,$6E,$3F,$C5,$D3,$FE,$FF,$CF,$FD,$0B,$8F,$83,$CD,$FF,$42,$4F
	.dc.b	$F5,$4A,$DE,$1F,$CD,$F0,$D4,$FF,$15,$21,$E8,$F3,$BF,$C8,$B1,$63
	.dc.b	$D3,$9F,$F4,$87,$E0,$F9,$FF,$CA,$FD,$0B,$1C,$8F,$F2,$E2,$AF,$87
	.dc.b	$E9,$7F,$96,$A6,$9F,$E5,$E8,$AF,$F0,$23,$D6,$63,$1F,$F4,$3E,$9C
	.dc.b	$F3,$FF,$34,$CF,$73,$ED,$7A,$7F,$E7,$28,$FF,$3A,$47,$FB,$D7,$E5
	.dc.b	$38,$7A,$BF,$C7,$A7,$FC,$F7,$3C,$4F,$EF,$CE,$1F,$E6,$AF,$FA,$13
	.dc.b	$5F,$EC,$2D,$7F,$C1,$6F,$FA,$23,$1F,$F0,$53,$FE,$92,$7F,$CC,$C7
	.dc.b	$E1,$BD,$7F,$F4,$DE,$7F,$84,$73,$7F,$E9,$AB,$7E,$AB,$FD,$62,$3F
	.dc.b	$D0,$9A,$7F,$A1,$63,$11,$FE,$A7,$BD,$9F,$D6,$FC,$7F,$F5,$17,$83
	.dc.b	$FA,$4C,$3F,$DD,$1E,$AF,$9D,$FF,$75,$7C,$3F,$FA,$57,$0F,$B7,$02
	.dc.b	$7C,$9F,$FA,$AF,$87,$94,$64,$FF,$57,$71,$77,$7F,$D2,$D5,$FE,$2C
	.dc.b	$F9,$3F,$F5,$97,$2F,$F6,$DF,$37,$FD,$69,$CF,$CA,$FF,$CD,$BD,$1F
	.dc.b	$94,$F4,$7F,$D6,$0E,$63,$FD,$75,$EC,$FF,$A1,$72,$FF,$61,$F6,$7F
	.dc.b	$B9,$3B,$3F,$D6,$CE,$EB,$FC,$FF,$3F,$B0,$F5,$FF,$D8,$5D,$DF,$E7
	.dc.b	$B9,$FF,$78,$F1,$FE,$D5,$FF,$1E,$3F,$E7,$DE,$6F,$F7,$DC,$7F,$B0
	.dc.b	$4F,$17,$27,$F9,$9F,$C2,$FE,$25,$EC,$97,$84,$F8,$6F,$F3,$67,$8D
	.dc.b	$FC,$74,$7C,$3F,$F9,$9F,$1F,$ED,$FE,$DF,$F4,$1F,$3F,$FC,$C5,$E8
	.dc.b	$7F,$97,$BE,$27,$FC,$E9,$5F,$E5,$4D,$7F,$CA,$38,$F3,$C7,$FC,$A7
	.dc.b	$EB,$7F,$8C,$32,$FE,$AB,$7F,$E7,$2B,$FF,$45,$F4,$7F,$95,$39,$BF
	.dc.b	$C9,$FC,$BF,$EC,$BC,$38,$7C,$F7,$C7,$C5,$F9,$5F,$C9,$F7,$BF,$BD
	.dc.b	$67,$F8,$47,$FC,$06,$FD,$F1,$9F,$F8,$C7,$1F,$F3,$27,$B5,$67,$FC
	.dc.b	$CD,$EF,$7E,$52,$7F,$7C,$BF,$69,$FF,$22,$FB,$DF,$5F,$DE,$FF,$13
	.dc.b	$D7,$E4,$2F,$B4,$6E,$31,$1F,$8F,$8E,$9D,$5F,$E1,$D8,$FF,$3A,$47
	.dc.b	$FB,$82,$7C,$78,$FA,$27,$F2,$D9,$7F,$87,$78,$3E,$DF,$A9,$F0,$2F
	.dc.b	$AD,$7F,$85,$78,$3F,$15,$3F,$E0,$B9,$FF,$01,$CF,$E8,$EB,$FC,$0F
	.dc.b	$C5,$F3,$6B,$E8,$CF,$EA,$9F,$F0,$1B,$FD,$45,$FF,$01,$F2,$7F,$5F
	.dc.b	$E4,$FE,$ED,$3F,$76,$7F,$05,$ED,$FF,$80,$7A,$1C,$FF,$DE,$F9,$FF
	.dc.b	$C2,$1A,$7F,$AF,$B1,$FE,$C1,$E0,$7F,$B0,$FB,$3F,$23,$D7,$FD,$86
	.dc.b	$BD,$7C,$3F,$D7,$9F,$F1,$4F,$27,$E7,$3D,$0F,$D0,$6F,$FC,$5B,$C0
	.dc.b	$FD,$27,$1F,$F5,$AB,$FD,$44,$FF,$56,$CF,$F5,$97,$FD,$5E,$7F,$04
	.dc.b	$7F,$9C,$EB,$FA,$DF,$85,$FB,$5C,$7F,$A4,$CF,$FA,$9F,$0A,$7F,$07
	.dc.b	$AF,$AD,$C9,$FD,$26,$75,$83,$C6,$FE,$0D,$7F,$D0,$0F,$F6,$6F,$8F
	.dc.b	$FC,$4E,$7F,$9E,$D7,$F4,$30,$7F,$A1,$E7,$FB,$6F,$93,$E8,$3F,$87
	.dc.b	$79,$5C,$7E,$07,$F7,$76,$F3,$3F,$84,$7C,$17,$E2,$70,$FF,$6D,$FA
	.dc.b	$9D,$23,$11,$FA,$11,$87,$F2,$58,$CC,$7E,$96,$7F,$AE,$79,$3F,$CF
	.dc.b	$F3,$FE,$D5,$FB,$27,$9A,$31,$1F,$D6,$E3,$78,$FF,$23,$F8,$3B,$3F
	.dc.b	$A0,$7F,$D9,$D5,$FC,$DF,$0F,$E6,$1B,$7F,$35,$8C,$47,$F3,$79,$ED
	.dc.b	$AF,$3E,$25,$6A,$95,$FE,$D4,$8C,$3D,$B7,$FE,$69,$C9,$EA,$6F,$17
	.dc.b	$F4,$5C,$FF,$DC,$F5,$EE,$F4,$3F,$9D,$79,$F6,$7F,$6E,$FA,$0F,$F1
	.dc.b	$9F,$2F,$F4,$7A,$C4,$FF,$82,$23,$1E,$8F,$9A,$BD,$19,$FF,$85,$2B
	.dc.b	$EE,$9F,$EE,$5D,$1C,$5F,$3B,$D0,$FF,$73,$FA,$1F,$D5,$E3,$E7,$47
	.dc.b	$F9,$0B,$AF,$C1,$F5,$7F,$D3,$30,$FF,$0B,$C7,$8F,$F2,$1C,$47,$B3
	.dc.b	$8F,$E3,$F6,$7F,$BC,$B3,$E9,$C5,$8A,$FF,$11,$C7,$7D,$FF,$A8,$C7
	.dc.b	$F8,$88,$F3,$F0,$FF,$56,$70,$7E,$59,$FC,$17,$27,$BB,$B7,$FB,$FC
	.dc.b	$BA,$7F,$13,$18,$7F,$10,$FF,$B0,$B9,$3E,$1E,$1F,$6B,$4F,$EB,$F9
	.dc.b	$7F,$95,$30,$4F,$DE,$73,$1F,$99,$CB,$F2,$BE,$D7,$CC,$F6,$3F,$53
	.dc.b	$87,$F9,$2B,$8F,$38,$7F,$99,$63,$E7,$BF,$D8,$CF,$E6,$7D,$5F,$91
	.dc.b	$C7,$FC,$71,$CD,$FC,$BD,$F2,$37,$27,$07,$F6,$9F,$4D,$C3,$FC,$67
	.dc.b	$DC,$7E,$33,$98,$C7,$AD,$7D,$D3,$FD,$BE,$9F,$91,$7F,$6E,$F3,$BF
	.dc.b	$D1,$BF,$8A,$FF,$92,$FD,$8F,$F4,$97,$C0,$89,$F8,$57,$F8,$6E,$33
	.dc.b	$19,$FC,$3B,$87,$27,$FA,$8A,$3F,$D3,$F1,$E8,$9F,$E9,$9E,$7F,$F0
	.dc.b	$1E,$47,$E8,$9F,$C4,$27,$31,$FC,$16,$39,$E1,$FD,$BE,$31,$C6,$AF
	.dc.b	$E7,$FE,$EF,$F1,$FF,$D6,$F1,$7F,$EB,$EF,$57,$FA,$11,$FE,$C3,$F4
	.dc.b	$3F,$D8,$97,$C1,$FF,$21,$B7,$F3,$7F,$F0,$2F,$0F,$C2,$EC,$FF,$02
	.dc.b	$E9,$E3,$C6,$BF,$DE,$75,$F8,$D3,$FA,$CD,$BD,$91,$EF,$8F,$F1,$A7
	.dc.b	$67,$C7,$7F,$8A,$6F,$FE,$0C,$C3,$D9,$87,$37,$07,$F6,$6B,$FA,$53
	.dc.b	$D6,$7E,$0E,$7F,$DA,$F1,$EF,$7F,$E1,$43,$86,$16,$3C,$77,$F1,$FD
	.dc.b	$8F,$E3,$56,$7C,$6F,$FC,$01,$BF,$F4,$51,$F9,$AE,$EF,$C1,$7B,$FC
	.dc.b	$0F,$4F,$C6,$FF,$87,$33,$FE,$25,$8B,$EC,$5F,$99,$7B,$BE,$0B,$55
	.dc.b	$FE,$61,$E0,$FF,$31,$46,$23,$F5,$8D,$F0,$BF,$D3,$4D,$C0,$FE,$8E
	.dc.b	$B8,$D7,$FA,$76,$3F,$49,$C1,$A8,$FE,$81,$E3,$5D,$BF,$A0,$61,$5F
	.dc.b	$E5,$FE,$6D,$AF,$C9,$64,$F8,$5F,$F1,$84,$7F,$55,$CF,$FC,$6B,$F0
	.dc.b	$7F,$D5,$23,$FA,$3F,$95,$A7,$FC,$75,$7F,$E3,$4C,$7F,$A8,$63,$99
	.dc.b	$FE,$6D,$8B,$0F,$F1,$95,$7F,$25,$7F,$C6,$EF,$F2,$7F,$81,$B5,$E4
	.dc.b	$D5,$1C,$23,$0F,$8D,$FE,$AD,$87,$F8,$E5,$FE,$29,$85,$7C,$B5,$5F
	.dc.b	$EB,$FE,$85,$7E,$A6,$6F,$81,$FE,$1D,$85,$9F,$ED,$58,$5C,$BF,$B5
	.dc.b	$6F,$5F,$EB,$3B,$DA,$F4,$61,$F5,$DC,$1F,$C9,$E7,$FC,$90,$7D,$48
	.dc.b	$FC,$06,$78,$8F,$F4,$F4,$D3,$EA,$BF,$E0,$DE,$1F,$D4,$F4,$7E,$17
	.dc.b	$E0,$7F,$CA,$13,$E2,$6D,$7C,$97,$F3,$97,$FC,$CC,$3F,$A1,$9F,$C2
	.dc.b	$B5,$FE,$B0,$FE,$B6,$7F,$89,$4F,$0C,$BF,$65,$1F,$C0,$72,$C5,$FE
	.dc.b	$EA,$3F,$D7,$5D,$5F,$E2,$AC,$3F,$9B,$CE,$67,$FD,$85,$3F,$EF,$79
	.dc.b	$CC,$FE,$FE,$BF,$CC,$6E,$EB,$F9,$2B,$E8,$C3,$F8,$4D,$7F,$99,$A7
	.dc.b	$FC,$32,$66,$78,$5F,$5A,$F3,$CF,$92,$F5,$5F,$C3,$4B,$D6,$3F,$90
	.dc.b	$61,$FC,$97,$E1,$7E,$3C,$FF,$2B,$F0,$5F,$05,$FE,$AE,$F0,$1E,$06
	.dc.b	$AB,$FC,$93,$3F,$EB,$23,$5E,$16,$63,$8F,$85,$FC,$D3,$0F,$F2,$B1
	.dc.b	$FE,$75,$D7,$E7,$E7,$FD,$09,$F9,$CD,$7E,$6D,$73,$F1,$3F,$AA,$10
	.dc.b	$78,$9F,$D9,$AB,$FA,$AC,$63,$83,$83,$B1,$E2,$FF,$BD,$67,$FD,$41
	.dc.b	$C1,$F0,$C8,$3C,$6F,$B1,$8F,$F8,$AB,$E0,$3F,$DB,$23,$C9,$FE,$E3
	.dc.b	$7F,$EA,$98,$CC,$7F,$6E,$8C,$79,$DF,$DE,$3C,$CF,$EE,$75,$C7,$CC
	.dc.b	$FF,$48,$61,$FA,$CF,$7F,$F8,$E9,$FA,$B3,$FC,$04,$FF,$A4,$FD,$1F
	.dc.b	$F0,$7B,$FC,$DF,$7C,$3F,$CE,$83,$FE,$10,$9F,$F4,$BF,$0F,$F0,$47
	.dc.b	$FD,$33,$E0,$FF,$88,$BC,$5D,$C6,$2F,$FD,$07,$DB,$FE,$21,$87,$A7
	.dc.b	$FE,$2B,$C7,$FD,$3D,$97,$F8,$86,$FF,$D1,$3E,$A7,$FA,$2A,$FF,$D3
	.dc.b	$D9,$FF,$1B,$6A,$3F,$D9,$D9,$78,$3D,$1F,$F2,$AC,$64,$CD,$FF,$A9
	.dc.b	$0F,$F2,$3B,$FE,$58,$EF,$C4,$7F,$A6,$63,$FC,$95,$6F,$5F,$FC,$F0
	.dc.b	$D9,$FF,$2B,$CF,$F2,$88,$C4,$7F,$BC,$FC,$7F,$F2,$D4,$7F,$99,$E3
	.dc.b	$FD,$47,$EC,$7F,$98,$F5,$FF,$59,$EB,$FE,$E5,$8F,$F3,$34,$FF,$AC
	.dc.b	$70,$F7,$7B,$3F,$E7,$67,$FD,$15,$9E,$5F,$F7,$01,$FE,$8D,$E0,$FF
	.dc.b	$5A,$3F,$EA,$39,$FA,$3E,$D7,$FA,$0A,$BF,$D6,$37,$FE,$99,$CA,$BF
	.dc.b	$E2,$9B,$FF,$20,$C6,$1F,$F1,$66,$DF,$E4,$6D,$3F,$D2,$79,$1F,$F4
	.dc.b	$AE,$17,$DC,$F8,$A7,$F8,$5A,$F1,$8F,$FA,$9C,$E1,$EE,$FF,$86,$2F
	.dc.b	$FD,$7D,$1E,$37,$FD,$63,$B7,$FA,$33,$8B,$F1,$3F,$0D,$FE,$6F,$A7
	.dc.b	$8F,$FD,$5F,$1F,$F4,$7E,$D8,$9F,$BF,$1F,$EC,$3F,$57,$FD,$A3,$5B
	.dc.b	$F8,$0F,$F6,$25,$7F,$B4,$3A,$3F,$D6,$38,$FF,$9F,$BF,$62,$FF,$5C
	.dc.b	$6D,$FE,$68,$DF,$FC,$5F,$5F,$EA,$DC,$FE,$1E,$15,$FE,$A5,$EE,$7F
	.dc.b	$B8,$1F,$F2,$97,$2F,$FB,$2C,$FB,$0F,$F9,$C1,$FF,$39,$F4,$7F,$A8
	.dc.b	$63,$FD,$89,$E8,$7F,$A7,$9F,$EA,$B4,$F7,$BF,$D3,$D1,$EE,$8F,$CF
	.dc.b	$BF,$EB,$0C,$1E,$AF,$FB,$3F,$A3,$FC,$FF,$DB,$FE,$93,$9F,$D7,$CF
	.dc.b	$FB,$4E,$3F,$D0,$75,$EC,$F6,$BF,$8A,$F8,$DF,$C5,$07,$E3,$5F,$E7
	.dc.b	$3C,$3F,$9F,$D7,$EB,$7F,$3F,$C4,$7F,$34,$F3,$7F,$98,$7A,$2F,$BE
	.dc.b	$FF,$33,$CF,$FA,$4B,$8F,$FC,$C3,$97,$F2,$0C,$BF,$CB,$13,$FD,$4F
	.dc.b	$1A,$FE,$47,$55,$FF,$96,$3D,$4F,$F3,$E5,$7F,$A1,$7C,$1F,$1E,$A8
	.dc.b	$FF,$44,$78,$7F,$C1,$FB,$BF,$DA,$63,$B3,$FC,$8B,$8F,$DA,$F6,$BF
	.dc.b	$C8,$0F,$F8,$AF,$0F,$F2,$57,$B3,$5F,$F5,$A7,$B5,$F8,$DF,$23,$FD
	.dc.b	$6E,$FF,$89,$E3,$FC,$20,$FF,$8B,$1F,$F0,$94,$D7,$FC,$73,$BF,$F8
	.dc.b	$BB,$E2,$7F,$27,$18,$9E,$9B,$7F,$89,$E3,$FD,$85,$1F,$9B,$E0,$FF
	.dc.b	$1A,$FB,$5E,$63,$FC,$B8,$FC,$4F,$3F,$FC,$23,$8F,$FB,$13,$6F,$E9
	.dc.b	$D8,$FC,$4E,$1F,$EA,$1E,$4D,$CF,$FB,$A9,$FE,$A5,$C7,$97,$0E,$72
	.dc.b	$FE,$79,$CB,$F8,$13,$F0,$35,$FD,$B4,$E1,$C3,$FD,$BF,$D7,$FE,$E5
	.dc.b	$D3,$FE,$A1,$E8,$FE,$DD,$5F,$CB,$4F,$A8,$7F,$75,$BE,$9D,$1F,$CD
	.dc.b	$7C,$AF,$A2,$7F,$48,$9F,$E6,$9C,$DF,$CF,$7B,$3F,$31,$C7,$FE,$E6
	.dc.b	$DF,$87,$6F,$CA,$BD,$CF,$F1,$18,$F9,$A8,$93,$DF,$BF,$F5,$3E,$47
	.dc.b	$F5,$2F,$0B,$96,$1F,$D1,$F2,$FE,$99,$1F,$B3,$84,$7F,$9E,$73,$78
	.dc.b	$7C,$27,$FA,$C3,$FB,$C9,$FE,$5B,$97,$86,$7F,$80,$BF,$CA,$5F,$F2
	.dc.b	$B5,$3E,$13,$96,$9F,$D6,$FC,$5F,$E1,$B3,$FC,$8A,$FF,$99,$61,$FD
	.dc.b	$7F,$C8,$FE,$29,$8F,$F1,$D9,$EC,$FC,$F3,$F3,$DC,$AF,$CB,$1E,$27
	.dc.b	$F6,$9A,$FE,$0D,$E3,$FC,$59,$CC,$7C,$7F,$33,$F8,$53,$FD,$D3,$0E
	.dc.b	$57,$C6,$3E,$61,$FE,$9B,$EB,$79,$0F,$C1,$3B,$61,$FD,$DE,$31,$5D
	.dc.b	$48,$3C,$7F,$EE,$D1,$A8,$4F,$2B,$EB,$F7,$BE,$19,$FD,$DD,$23,$E9
	.dc.b	$65,$FC,$1A,$31,$3F,$12,$3E,$21,$E6,$FC,$73,$94,$7A,$A5,$18,$8F
	.dc.b	$F0,$DC,$61,$F3,$9F,$CB,$7C,$1F,$7F,$EB,$9F,$81,$E9,$FC,$75,$23
	.dc.b	$19,$7C,$4F,$8C,$E9,$85,$39,$9F,$95,$E1,$E3,$1E,$47,$C2,$FE,$37
	.dc.b	$93,$33,$FD,$D7,$0F,$F3,$53,$C8,$F5,$77,$FB,$3E,$47,$C6,$DC,$7E
	.dc.b	$AF,$CA,$7E,$9E,$BE,$4B,$5F,$85,$F9,$07,$9B,$03,$F7,$F8,$BE,$7B
	.dc.b	$67,$67,$CD,$E7,$FC,$5D,$70,$8C,$1F,$86,$F2,$3F,$0B,$F9,$07,$53
	.dc.b	$F2,$59,$7C,$BA,$46,$2B,$84,$3F,$62,$FD,$DC,$7F,$93,$E3,$0F,$DD
	.dc.b	$3F,$BE,$BF,$D3,$DF,$88,$7D,$4F,$1B,$F2,$6F,$E4,$63,$4F,$E3,$8F
	.dc.b	$C9,$E7,$F8,$09,$FE,$1F,$C3,$F8,$57,$F8,$8E,$5F,$E4,$7B,$7C,$17
	.dc.b	$C7,$3F,$A7,$61,$FC,$3F,$97,$F1,$91,$88,$FF,$5A,$E9,$FE,$05,$E7
	.dc.b	$FD,$B6,$DF,$A8,$F6,$3F,$21,$36,$BF,$8C,$7B,$7F,$E8,$0F,$77,$FC
	.dc.b	$C7,$9A,$FF,$D0,$91,$FE,$7D,$F7,$14,$87,$89,$CF,$0C,$C6,$E7,$5F
	.dc.b	$1D,$9E,$7F,$E7,$C7,$FA,$D7,$2E,$D7,$FC,$A6,$33,$EF,$F3,$F8,$2F
	.dc.b	$F4,$E1,$FE,$A4,$AD,$D6,$20,$3E,$5B,$83,$C3,$E5,$EB,$FE,$A5,$EB
	.dc.b	$F3,$7B,$BF,$C4,$2F,$F6,$9C,$1F,$DF,$38,$FF,$82,$BF,$C4,$27,$53
	.dc.b	$FB,$68,$FF,$64,$7A,$7F,$D4,$B3,$F1,$BF,$D5,$39,$7F,$D9,$7F,$31
	.dc.b	$F0,$E3,$03,$E6,$B1,$EA,$FF,$B1,$E1,$F3,$5F,$C7,$3C,$BF,$E3,$FC
	.dc.b	$7F,$18,$C6,$9F,$15,$FE,$CD,$C7,$88,$FE,$70,$7F,$2A,$D3,$F9,$6A
	.dc.b	$F7,$FF,$CD,$9C,$1B,$E0,$F5,$F5,$71,$8F,$F3,$87,$5F,$F2,$CF,$C8
	.dc.b	$3E,$14,$5C,$3F,$AE,$5F,$0C,$5F,$B7,$69,$B3,$A3,$E6,$62,$F8,$27
	.dc.b	$D8,$7E,$41,$B3,$FC,$D6,$7F,$C4,$F8,$FF,$AA,$67,$33,$FA,$79,$FE
	.dc.b	$83,$E0,$7F,$6D,$7A,$7E,$19,$F9,$AF,$84,$7E,$51,$FC,$D3,$74,$63
	.dc.b	$0F,$E6,$99,$7F,$57,$8F,$EC,$51,$FB,$83,$7E,$C7,$33,$94,$6A,$3E
	.dc.b	$4C,$F4,$EB,$FE,$5B,$3F,$92,$F2,$3E,$5B,$6B,$F1,$9A,$6D,$FE,$0B
	.dc.b	$E3,$57,$E9,$FC,$1D,$C7,$FD,$41,$F1,$3E,$8E,$9E,$A7,$F4,$5C,$31
	.dc.b	$7C,$BE,$19,$F8,$EF,$F3,$E5,$FA,$4F,$F4,$0C,$AD,$7B,$5D,$BE,$FF
	.dc.b	$87,$FC,$2E,$DF,$9C,$C5,$18,$9F,$D2,$6B,$8C,$FF,$A5,$CF,$FA,$37
	.dc.b	$5F,$EB,$B8,$F0,$DF,$FB,$1E,$D7,$1F,$E4,$79,$57,$F9,$01,$FD,$FB
	.dc.b	$2C,$67,$F9,$2C,$BE,$44,$FE,$CD,$FA,$19,$7E,$51,$AF,$FB,$15,$FA
	.dc.b	$93,$C7,$83,$CB,$C5,$F8,$19,$FC,$05,$FD,$8F,$85,$72,$7F,$20,$7F
	.dc.b	$1B,$87,$E2,$E7,$F9,$50,$E5,$4E,$A3,$07,$E4,$72,$FF,$7F,$46,$63
	.dc.b	$FC,$33,$3F,$ED,$3E,$9F,$E1,$11,$FA,$2F,$7B,$F4,$AF,$EC,$EF,$FD
	.dc.b	$BD,$0E,$CD,$CF,$28,$C3,$FE,$4B,$ED,$FD,$CC,$7E,$F2,$BF,$6E,$FE
	.dc.b	$FB,$B7,$F7,$6D,$AF,$E2,$EF,$F0,$2A,$E3,$CB,$FC,$33,$BF,$F8,$4C
	.dc.b	$FF,$06,$7F,$87,$3F,$C6,$BC,$2F,$E2,$9E,$1F,$FB,$B7,$9F,$F8,$FC
	.dc.b	$7F,$52,$F0,$FF,$8B,$D6,$FC,$54,$7F,$9D,$23,$0A,$78,$78,$BC,$F4
	.dc.b	$FE,$0C,$7B,$F0,$F8,$2F,$F9,$0A,$FF,$9E,$E9,$F0,$27,$FC,$83,$E3
	.dc.b	$FF,$3F,$F1,$F9,$5F,$F4,$18,$79,$18,$71,$87,$F4,$67,$FA,$37,$93
	.dc.b	$C3,$7F,$E9,$B3,$FD,$29,$F8,$38,$FF,$5E,$F3,$5F,$18,$F3,$FF,$AD
	.dc.b	$FC,$2F,$83,$CF,$FE,$CD,$E7,$DF,$43,$FA,$F1,$FD,$8B,$3F,$CD,$EF
	.dc.b	$FE,$E5,$EC,$FF,$39,$75,$7F,$72,$9F,$EF,$73,$FC,$84,$FC,$88,$72
	.dc.b	$FF,$21,$3F,$2B,$39,$E7,$FF,$1E,$CF,$F8,$32,$87,$FB,$62,$BF,$C7
	.dc.b	$C7,$FA,$4E,$35,$19,$87,$87,$C2,$3C,$1E,$8F,$F8,$62,$7F,$C2,$D5
	.dc.b	$FE,$18,$F0,$3F,$CC,$8D,$8F,$C6,$FA,$7F,$E2,$6C,$BF,$CE,$B3,$FE
	.dc.b	$26,$9F,$F5,$91,$FE,$78,$C3,$FD,$61,$EA,$7F,$8B,$E3,$FC,$31,$3F
	.dc.b	$E3,$27,$FC,$EB,$5F,$E7,$E7,$CB,$19,$7F,$84,$9F,$85,$F5,$7F,$DB
	.dc.b	$7E,$7F,$FA,$32,$7F,$C8,$53,$FE,$C1,$8E,$5A,$7C,$BA,$FF,$25,$4F
	.dc.b	$EF,$6B,$79,$F7,$3C,$0E,$23,$F5,$3D,$5F,$E5,$77,$83,$5F,$AD,$CF
	.dc.b	$FE,$61,$69,$FE,$65,$F5,$BB,$47,$D6,$F6,$3F,$CD,$93,$F7,$5F,$F3
	.dc.b	$76,$96,$3F,$CD,$9E,$BF,$D7,$9F,$F3,$A4,$E6,$7E,$09,$E6,$F6,$7F
	.dc.b	$D0,$4F,$E1,$DF,$F4,$0E,$5F,$E8,$98,$D9,$FD,$9D,$C6,$7F,$CF,$DB
	.dc.b	$1F,$C1,$BD,$CF,$ED,$11,$FE,$9A,$8F,$EB,$ED,$9F,$E1,$BE,$F7,$E1
	.dc.b	$5F,$C6,$37,$0F,$CC,$3F,$89,$F7,$BF,$27,$F2,$1F,$D2,$F0,$FF,$60
	.dc.b	$87,$0F,$FA,$F3,$4F,$E6,$11,$88,$FF,$61,$7B,$4F,$C3,$FF,$D5,$FC
	.dc.b	$DF,$EB,$49,$FE,$F6,$FE,$67,$D4,$FF,$52,$C6,$21,$F1,$5F,$EB,$EE
	.dc.b	$0E,$38,$38,$7F,$D6,$7C,$DF,$EC,$27,$FD,$D9,$F1,$17,$1E,$58,$FF
	.dc.b	$B1,$B9,$3F,$D7,$3F,$43,$FE,$C8,$E7,$E7,$D1,$F9,$2B,$FF,$0A,$E7
	.dc.b	$F9,$2C,$3F,$C5,$EF,$E6,$79,$7F,$DB,$3C,$BF,$C0,$E3,$F4,$D3,$FB
	.dc.b	$1A,$FD,$37,$4F,$FB,$73,$AF,$CF,$D9,$F2,$EB,$FD,$BE,$FF,$A1,$1F
	.dc.b	$F6,$83,$EC,$ED,$FF,$40,$63,$FE,$87,$F5,$BF,$B9,$E5,$FE,$66,$F0
	.dc.b	$7F,$CF,$9F,$3F,$FE,$71,$E1,$10,$9C,$7F,$E7,$9F,$2F,$FC,$87,$7F
	.dc.b	$CE,$BC,$DF,$E8,$1F,$41,$FE,$7D,$DB,$FC,$2F,$3F,$D2,$70,$E4,$72
	.dc.b	$7F,$A8,$47,$FB,$8B,$83,$FC,$9C,$FF,$B9,$3A,$BF,$C9,$9C,$5F,$D8
	.dc.b	$E3,$FD,$2B,$5F,$D9,$B2,$FF,$29,$FB,$EF,$A9,$FC,$0A,$FC,$27,$F0
	.dc.b	$7F,$03,$FD,$FB,$0F,$F1,$4B,$FE,$06,$7F,$8E,$1F,$C9,$32,$FF,$17
	.dc.b	$3F,$EF,$4F,$6E,$BF,$EF,$69,$FC,$86,$7F,$E4,$DF,$80,$7E,$D5,$F9
	.dc.b	$07,$FC,$53,$1F,$E4,$2F,$80,$B7,$FE,$3A,$AF,$CA,$E1,$CA,$BF,$C1
	.dc.b	$4F,$F8,$9E,$7C,$1E,$97,$F2,$37,$C5,$1E,$77,$CF,$97,$F8,$CB,$DF
	.dc.b	$FF,$15,$FB,$DE,$FC,$3F,$27,$5F,$E1,$4D,$FF,$47,$E5,$FC,$78,$F9
	.dc.b	$CF,$F8,$22,$7E,$64,$7F,$23,$8F,$F0,$43,$F4,$5F,$AA,$FF,$53,$F6
	.dc.b	$FF,$C0,$9C,$BF,$68,$FD,$34,$FA,$67,$CA,$FE,$0F,$B7,$FB,$FF,$3F
	.dc.b	$F7,$2E,$5F,$F0,$9E,$FC,$B0,$FC,$1F,$4F,$F6,$5F,$57,$FB,$5F,$A9
	.dc.b	$FE,$C4,$EC,$FE,$C7,$A7,$4C,$FF,$32,$FF,$5A,$D3,$FA,$BB,$F6,$8C
	.dc.b	$78,$1F,$A1,$EC,$FE,$AF,$C1,$FE,$EA,$03,$CD,$E5,$E1,$7C,$63,$FD
	.dc.b	$F7,$1F,$D9,$A7,$FB,$77,$A3,$FD,$4E,$3F,$A2,$78,$BF,$30,$FD,$E3
	.dc.b	$FD,$5B,$C6,$F9,$FC,$3F,$C0,$B3,$FE,$68,$7F,$8D,$38,$7F,$84,$BF
	.dc.b	$CD,$FC,$6F,$F3,$4B,$FC,$D6,$FE,$CE,$CF,$2B,$F0,$53,$F9,$6F,$2F
	.dc.b	$E0,$7A,$1F,$E6,$AA,$FE,$41,$7F,$80,$3F,$DA,$27,$B8,$7A,$38,$7F
	.dc.b	$B4,$79,$7F,$E0,$27,$06,$BC,$BA,$BF,$0A,$AB,$EF,$B8,$3C,$E3,$FC
	.dc.b	$BE,$FF,$B8,$33,$FE,$F6,$FF,$AF,$0E,$75,$FE,$1D,$7F,$D9,$67,$FA
	.dc.b	$D8,$FF,$66,$C7,$F8,$37,$C5,$F4,$CF,$E7,$A7,$FD,$20,$79,$3E,$0B
	.dc.b	$93,$FE,$A8,$3F,$A2,$1A,$3E,$EF,$CE,$FC,$2B,$FE,$51,$15,$F8,$15
	.dc.b	$E2,$8F,$82,$FF,$9D,$1F,$41,$EC,$2B,$FE,$81,$7F,$CF,$31,$FE,$2A
	.dc.b	$BF,$99,$3E,$99,$E0,$7B,$4F,$9B,$7D,$5F,$F0,$C0,$7D,$2F,$A7,$D4
	.dc.b	$EC,$FE,$5B,$CF,$FF,$24,$BF,$E1,$F5,$EB,$7F,$92,$BD,$8E,$71,$8F
	.dc.b	$07,$F5,$2E,$BF,$2C,$EF,$D1,$FB,$73,$F9,$DF,$C4,$61,$C7,$DC,$57
	.dc.b	$F7,$9F,$7F,$E6,$F8,$2F,$F3,$A7,$89,$FE,$78,$F5,$BF,$CF,$9F,$07
	.dc.b	$97,$FC,$FC,$7F,$9F,$FE,$2F,$13,$B3,$FC,$FF,$F9,$6F,$C9,$C7,$C4
	.dc.b	$33,$C3,$E2,$E1,$FC,$5B,$AC,$FE,$87,$C3,$7C,$89,$FD,$CF,$BB,$FE
	.dc.b	$62,$F0,$BF,$0B,$CD,$FE,$84,$9F,$ED,$27,$CD,$3F,$D1,$27,$F9,$1D
	.dc.b	$FF,$34,$D7,$38,$FF,$2E,$90,$3F,$96,$AE,$91,$FE,$C0,$6B,$C3,$C9
	.dc.b	$FD,$75,$FE,$71,$FF,$79,$47,$F0,$D3,$FC,$F0,$78,$FA,$3F,$88,$C7
	.dc.b	$08,$FD,$4E,$1F,$E5,$B8,$C4,$FE,$96,$3F,$8B,$FD,$66,$EB,$F2,$CF
	.dc.b	$FB,$FF,$0F,$D9,$E1,$EB,$BF,$F3,$BF,$5E,$4E,$17,$FC,$0F,$C0,$FF
	.dc.b	$6C,$5F,$08,$FC,$9C,$7C,$C9,$DC,$7F,$26,$8F,$CB,$75,$BF,$3A,$CC
	.dc.b	$FF,$2A,$C1,$39,$8F,$F7,$DF,$D2,$7F,$99,$DF,$B1,$8A,$31,$1F,$A2
	.dc.b	$9C,$57,$48,$FF,$47,$CF,$F0,$CC,$BE,$B5,$7D,$4A,$D5,$DB,$FE,$77
	.dc.b	$5A,$DB,$AC,$7F,$9F,$1A,$FF,$92,$7F,$6A,$F3,$3F,$E6,$03,$F6,$4E
	.dc.b	$27,$E5,$7E,$11,$F1,$67,$FD,$1D,$A7,$FB,$FE,$31,$1F,$2F,$74,$62
	.dc.b	$3E,$41,$8D,$BD,$93,$63,$CE,$7E,$6F,$36,$0D,$F6,$D9,$D9,$FC,$22
	.dc.b	$77,$C1,$F8,$68,$FF,$5F,$CF,$E3,$DF,$F6,$11,$FA,$1B,$DB,$F9,$27
	.dc.b	$F2,$F3,$D8,$FA,$15,$E2,$D7,$F3,$B3,$6B,$F5,$27,$F4,$B8,$C3,$FA
	.dc.b	$63,$D5,$7F,$AF,$70,$7F,$1E,$9F,$69,$FA,$89,$B1,$D0,$FC,$54,$F1
	.dc.b	$3F,$69,$3F,$12,$B8,$47,$F0,$4E,$2F,$E2,$ED,$7F,$8B,$B5,$FF,$26
	.dc.b	$F1,$FF,$B0,$BA,$7F,$8C,$93,$97,$1D,$1F,$C6,$39,$D2,$E8,$CC,$7F
	.dc.b	$19,$87,$4F,$FB,$33,$1F,$E3,$3D,$55,$FE,$59,$D5,$5F,$F1,$26,$7F
	.dc.b	$ED,$2E,$CF,$F8,$03,$AF,$F3,$9E,$FF,$F8,$03,$DB,$97,$F9,$D4,$7F
	.dc.b	$5B,$E2,$FD,$59,$F8,$03,$FA,$57,$5F,$C6,$7E,$61,$FD,$2A,$38,$39
	.dc.b	$3F,$DC,$9C,$5F,$D4,$DF,$F5,$EB,$FD,$54,$AF,$F5,$B8,$FE,$BA,$A3
	.dc.b	$2F,$EA,$0F,$EB,$6F,$E9,$F9,$FF,$B5,$CF,$F7,$F7,$FB,$5B,$FD,$C6
	.dc.b	$7F,$81,$B7,$C1,$FD,$77,$2F,$9E,$7C,$F7,$FB,$13,$FC,$E3,$E0,$DC
	.dc.b	$3F,$D6,$D8,$7F,$75,$AF,$23,$6F,$84,$79,$4F,$F1,$DC,$FF,$9E,$A7
	.dc.b	$FC,$37,$E1,$7C,$99,$FE,$90,$F2,$7F,$C4,$5E,$1A,$BF,$81,$47,$F2
	.dc.b	$52,$9F,$C8,$71,$AB,$C4,$FF,$10,$BF,$DD,$75,$78,$7F,$DC,$1B,$1F
	.dc.b	$E3,$13,$FA,$FF,$17,$FA,$A9,$C7,$C4,$88,$41,$F1,$3F,$DA,$BA,$3F
	.dc.b	$33,$D7,$47,$57,$E6,$3E,$2F,$FB,$4E,$3F,$DF,$71,$FF,$19,$ED,$FE
	.dc.b	$63,$3F,$CB,$4F,$ED,$7C,$4F,$F2,$17,$1F,$F9,$17,$D6,$D5,$7F,$90
	.dc.b	$FC,$0F,$EB,$B7,$FD,$B8,$FF,$33,$CE,$A7,$FA,$CC,$7C,$D3,$FA,$C1
	.dc.b	$FD,$7B,$8B,$FD,$AD,$E1,$7F,$3B,$8F,$EA,$66,$CF,$E9,$C7,$1E,$3F
	.dc.b	$A5,$D3,$FC,$C6,$3F,$BA,$78,$DF,$27,$0F,$2F,$4F,$F0,$C9,$F1,$2F
	.dc.b	$1B,$84,$7F,$81,$7C,$5F,$E6,$4F,$D7,$AF,$E6,$0F,$DE,$EB,$FE,$58
	.dc.b	$DF,$17,$FD,$17,$3F,$C9,$A3,$5B,$FC,$2A,$FF,$02,$6F,$F0,$5F,$95
	.dc.b	$3F,$E0,$97,$F9,$04,$67,$7F,$F3,$BD,$FF,$B3,$3D,$1F,$E3,$57,$FE
	.dc.b	$90,$E3,$FC,$1E,$1F,$85,$E7,$FB,$BC,$DF,$E2,$D8,$C7,$3F,$F1,$59
	.dc.b	$73,$FF,$46,$C3,$FC,$F9,$D3,$F8,$73,$F1,$3D,$3F,$C4,$72,$FF,$5A
	.dc.b	$63,$FC,$3F,$AF,$F2,$87,$F9,$B3,$5F,$F3,$AE,$1F,$E5,$27,$FD,$50
	.dc.b	$FF,$53,$6B,$FE,$89,$F0,$3F,$C9,$9F,$14,$F5,$BF,$D6,$93,$F9,$DC
	.dc.b	$7F,$C7,$7E,$87,$E5,$78,$3F,$D7,$F3,$FC,$CE,$44,$9F,$F3,$56,$3F
	.dc.b	$90,$F7,$7F,$D6,$7D,$DF,$93,$8F,$CC,$71,$FE,$47,$0F,$F6,$AC,$7E
	.dc.b	$DC,$47,$FA,$EB,$EA,$7E,$7C,$7F,$BE,$A5,$F5,$38,$8E,$1D,$1C,$27
	.dc.b	$FD,$CB,$DC,$FA,$8D,$57,$F0,$3E,$FF,$C8,$3F,$F0,$07,$C2,$AF,$FB
	.dc.b	$37,$E0,$7F,$21,$5C,$23,$FC,$03,$A5,$E6,$F0,$E5,$FE,$F2,$5A,$E7
	.dc.b	$4F,$75,$FF,$B5,$EF,$FB,$1D,$FB,$A3,$FD,$A1,$7F,$BA,$7F,$75,$BF
	.dc.b	$E2,$A7,$F8,$5F,$07,$F5,$6E,$AF,$EB,$BE,$EF,$E0,$78,$FE,$78,$9F
	.dc.b	$E9,$DD,$1F,$C5,$7D,$CF,$E8,$79,$7F,$B9,$F3,$FE,$74,$FC,$D7,$FD
	.dc.b	$66,$FF,$AA,$32,$FF,$65,$F1,$FF,$B0,$B6,$FE,$43,$1F,$E3,$4E,$EF
	.dc.b	$F6,$65,$FF,$B1,$BC,$0F,$C6,$CF,$F3,$F3,$FD,$46,$7C,$AE,$0F,$F5
	.dc.b	$EB,$FE,$0B,$3F,$CE,$51,$E1,$7F,$DB,$2F,$8E,$BF,$DB,$75,$FE,$D2
	.dc.b	$3F,$D8,$33,$D5,$F0,$BF,$E4,$28,$FF,$5F,$57,$FB,$73,$1F,$E4,$31
	.dc.b	$FE,$BF,$E7,$FF,$5B,$D7,$E4,$0F,$F5,$BF,$E9,$0F,$D2,$1F,$A4,$3F
	.dc.b	$4A,$7E,$94,$AD,$7E,$A1,$FE,$BA,$78,$56,$E7,$57,$CB,$0F,$E8,$55
	.dc.b	$AC,$39,$BD,$FF,$72,$FF,$61,$FE,$2A,$F8,$58,$41,$F0,$DA,$70,$6F
	.dc.b	$E5,$FF,$D9,$7E,$57,$F0,$71,$F2,$98,$F1,$7F,$C3,$50,$FA,$AF,$D4
	.dc.b	$6B,$FC,$CF,$8E,$BF,$DD,$8F,$CB,$9F,$E4,$73,$FC,$08,$60,$FF,$26
	.dc.b	$F1,$FF,$5D,$F0,$7F,$CA,$51,$FE,$4F,$3F,$42,$7F,$76,$F1,$BF,$D2
	.dc.b	$DC,$FF,$E4,$EE,$AF,$D9,$4F,$F8,$11,$BD,$3F,$EB,$DF,$7B,$FD,$2B
	.dc.b	$5F,$E8,$E7,$FD,$7E,$DC,$7F,$D1,$FE,$27,$F9,$63,$B7,$F7,$58,$FF
	.dc.b	$4D,$97,$9D,$FD,$0A,$3F,$D2,$BE,$86,$9F,$F0,$9B,$72,$FF,$3F,$E3
	.dc.b	$DF,$6F,$DD,$D7,$F5,$DC,$FF,$DE,$DE,$87,$9A,$3F,$D4,$E6,$8F,$F0
	.dc.b	$EB,$97,$FC,$ED,$7F,$08,$FF,$2D,$95,$FF,$33,$EF,$F2,$33,$FF,$30
	.dc.b	$B6,$11,$FE,$84,$8C,$45,$3D,$D3,$F4,$CF,$F6,$01,$FE,$0F,$E1,$FC
	.dc.b	$17,$0F,$F4,$79,$1F,$D3,$BC,$64,$FE,$1A,$BF,$C2,$DD,$1F,$8C,$E7
	.dc.b	$FC,$64,$61,$C7,$47,$3C,$3F,$3B,$DD,$F9,$FE,$BF,$D1,$DF,$E7,$BB
	.dc.b	$7F,$53,$DB,$F9,$E5,$DB,$C6,$BF,$45,$DD,$CA,$FF,$D4,$58,$FF,$BE
	.dc.b	$CF,$F4,$F1,$FE,$9C,$3F,$E0,$07,$81,$FE,$1A,$7F,$CA,$4F,$F3,$BE
	.dc.b	$5F,$F5,$D4,$FF,$BA,$CF,$EE,$93,$FC,$0D,$AF,$FB,$35,$FF,$32,$9F
	.dc.b	$C4,$27,$FC,$84,$FF,$85,$7C,$1F,$E2,$0F,$F9,$47,$B3,$FC,$8B,$3F
	.dc.b	$D4,$7A,$3F,$C1,$57,$FE,$6F,$3F,$8F,$1F,$E4,$AB,$FF,$29,$3F,$E4
	.dc.b	$13,$FD,$BC,$7F,$92,$BA,$FF,$CA,$4E,$4F,$F5,$65,$7E,$F7,$87,$E9
	.dc.b	$47,$F9,$27,$F1,$31,$08,$3F,$0F,$FD,$F7,$37,$F8,$2A,$7F,$08,$75
	.dc.b	$3F,$7B,$A7,$FB,$DD,$FF,$7C,$61,$FE,$61,$7F,$28,$7F,$BD,$B3,$FF
	.dc.b	$1C,$69,$F9,$B8,$FC,$49,$FE,$2B,$CF,$F5,$0F,$5F,$B3,$E4,$FF,$8D
	.dc.b	$3B,$AE,$1F,$E3,$8E,$ED,$CF,$F8,$D7,$1E,$79,$7C,$1F,$57,$FC,$E5
	.dc.b	$DD,$DE,$BF,$CD,$13,$E4,$7F,$D3,$DD,$BF,$E3,$68,$F6,$C7,$C9,$7F
	.dc.b	$41,$E1,$7F,$1E,$3F,$2D,$C3,$F1,$8F,$91,$C3,$F9,$9E,$CF,$CF,$71
	.dc.b	$FF,$2F,$C3,$F0,$5B,$7F,$5F,$7F,$3B,$D5,$F9,$9F,$D4,$FE,$A5,$7F
	.dc.b	$33,$F9,$7F,$9F,$DF,$FE,$72,$E7,$FD,$5F,$C7,$7D,$FF,$33,$FB,$89
	.dc.b	$3E,$2B,$F6,$3F,$9A,$7F,$83,$B9,$AE,$BF,$E9,$8D,$BF,$CE,$2F,$D8
	.dc.b	$ED,$F9,$E7,$F9,$E3,$F1,$CE,$10,$7F,$C2,$1D,$AF,$47,$A3,$67,$A5
	.dc.b	$FF,$A0,$63,$E6,$47,$FA,$67,$07,$73,$F4,$2F,$23,$67,$8D,$FF,$8B
	.dc.b	$CF,$F5,$37,$7F,$DC,$7F,$D1,$9F,$99,$A3,$11,$FE,$25,$F0,$2B,$C6
	.dc.b	$7F,$99,$BF,$E2,$EF,$53,$FC,$F1,$8F,$F8,$A7,$C1,$F0,$63,$FE,$8D
	.dc.b	$CF,$F9,$67,$77,$E1,$57,$81,$ED,$9F,$F4,$96,$38,$DF,$E6,$5F,$C5
	.dc.b	$9F,$F1,$87,$06,$3C,$0F,$EA,$BC,$5F,$E2,$CE,$3F,$F1,$B1,$C2,$BF
	.dc.b	$3B,$CB,$F6,$0F,$F1,$4F,$2F,$F1,$FE,$8F,$ED,$9C,$FF,$89,$EA,$FE
	.dc.b	$4D,$D5,$FC,$6F,$C6,$FE,$3F,$E2,$7F,$80,$3B,$7F,$93,$67,$FA,$8F
	.dc.b	$13,$FC,$D6,$FF,$32,$F0,$BF,$28,$7E,$F3,$F0,$EF,$E0,$DF,$A6,$7F
	.dc.b	$AC,$BA,$3F,$1B,$E2,$BC,$2F,$E6,$FF,$29,$F8,$DE,$3F,$31,$FE,$9A
	.dc.b	$CB,$FD,$6B,$D5,$FE,$BF,$F8,$0F,$F3,$E3,$F5,$A3,$FC,$B2,$FD,$BE
	.dc.b	$FF,$E0,$51,$FE,$4F,$8F,$C4,$74,$7F,$95,$1F,$F7,$9C,$7F,$A7,$1F
	.dc.b	$F2,$8B,$5F,$C2,$0A,$FC,$DC,$7F,$AD,$07,$47,$F9,$B3,$B3,$F8,$8C
	.dc.b	$7F,$82,$85,$FF,$BC,$63,$FA,$AF,$5F,$F9,$CA,$ED,$7E,$4B,$0F,$F4
	.dc.b	$05,$3C,$4F,$C1,$47,$FB,$11,$AF,$F9,$99,$FC,$FF,$91,$FA,$5F,$17
	.dc.b	$F1,$B3,$FE,$CA,$6B,$FE,$6C,$F1,$78,$BE,$CF,$13,$F3,$98,$7E,$5E
	.dc.b	$7F,$C2,$F3,$FE,$E6,$9F,$D0,$71,$E6,$3F,$D1,$3F,$53,$FA,$37,$FD
	.dc.b	$17,$FA,$6F,$E8,$34,$F5,$40,$4E,$63,$FD,$D9,$3F,$E1,$8A,$FF,$49
	.dc.b	$DD,$9E,$36,$E1,$FF,$0E,$E5,$AF,$1F,$D3,$3F,$77,$C6,$F6,$BF,$5A
	.dc.b	$7F,$5F,$7C,$B8,$3F,$5A,$FF,$55,$DF,$F8,$57,$8D,$CA,$FE,$7B,$5F
	.dc.b	$45,$3C,$6F,$D7,$5F,$E0,$A7,$EA,$C6,$1F,$F7,$BB,$FC,$6E,$3F,$A7
	.dc.b	$E1,$C7,$97,$F0,$93,$F8,$69,$F2,$3F,$8D,$AF,$E2,$15,$FE,$A4,$CF
	.dc.b	$9F,$2F,$E5,$67,$BB,$5F,$2D,$7E,$79,$FF,$63,$6D,$FE,$17,$7C,$C7
	.dc.b	$C9,$8F,$EB,$78,$7E,$87,$2F,$F5,$47,$37,$FB,$5B,$E0,$FF,$80,$75
	.dc.b	$7F,$9A,$3A,$7F,$C4,$1D,$7F,$B5,$C7,$FB,$A7,$6F,$EE,$BB,$7F,$B7
	.dc.b	$67,$D3,$D1,$AF,$6D,$FF,$D4,$EF,$F7,$F8,$FF,$51,$BF,$DB,$3D,$DD
	.dc.b	$BF,$DB,$BA,$FF,$D5,$2F,$9A,$3F,$A1,$BF,$EA,$23,$FD,$30,$7F,$82
	.dc.b	$63,$FD,$C9,$EE,$F8,$1B,$7E,$4A,$EB,$5F,$E4,$27,$FD,$52,$F5,$30
	.dc.b	$6C,$D4,$7F,$AC,$B8,$7D,$31,$F2,$3C,$1F,$F7,$93,$EA,$3F,$08,$53
	.dc.b	$FC,$9F,$A7,$F8,$3E,$BF,$C4,$5F,$BA,$3D,$C7,$C0,$3E,$01,$F0,$2F
	.dc.b	$85,$D7,$C9,$70,$BD,$D6,$A7,$95,$FF,$8B,$7F,$55,$7E,$31,$08,$3F
	.dc.b	$0E,$D3,$83,$5B,$7E,$0A,$71,$1C,$BF,$0D,$FF,$1C,$7E,$03,$FE,$29
	.dc.b	$DE,$DF,$CA,$FC,$2F,$FC,$1D,$F9,$57,$F8,$F5,$FF,$08,$7E,$6B,$FE
	.dc.b	$06,$F5,$7F,$D7,$3F,$5D,$FE,$41,$EA,$FF,$19,$79,$FF,$CC,$1B,$8F
	.dc.b	$E8,$4F,$F5,$87,$91,$FE,$24,$7F,$C8,$9E,$87,$F2,$CE,$9F,$E6,$FD
	.dc.b	$DF,$ED,$7A,$FF,$43,$6D,$9B,$FE,$59,$3F,$E7,$EB,$FE,$53,$1F,$EB
	.dc.b	$BF,$47,$A3,$FE,$53,$8F,$E5,$5E,$56,$23,$F3,$11,$FD,$B7,$1F,$F2
	.dc.b	$43,$A8,$FF,$51,$5D,$7F,$AE,$8F,$84,$FE,$77,$3C,$63,$FC,$C1,$1F
	.dc.b	$CD,$F4,$FE,$FB,$7F,$E5,$3B,$FF,$08,$FB,$9E,$1B,$FF,$26,$E1,$FC
	.dc.b	$C1,$FC,$49,$F0,$BC,$2F,$F0,$CA,$F9,$27,$86,$FA,$8E,$53,$FE,$9E
	.dc.b	$C5,$C3,$98,$FF,$62,$79,$DF,$EC,$3C,$3F,$96,$47,$A8,$E9,$1E,$97
	.dc.b	$E3,$7A,$3E,$87,$E9,$B5,$FF,$3E,$5F,$F8,$45,$F8,$BE,$87,$17,$FC
	.dc.b	$CF,$C1,$FD,$5E,$31,$1F,$EB,$28,$FE,$C3,$18,$8F,$B0,$D7,$FD,$83
	.dc.b	$BF,$F2,$E3,$E9,$95,$FE,$14,$FE,$24,$FC,$2B,$C3,$93,$FD,$95,$BF
	.dc.b	$F2,$F8,$FF,$7D,$72,$57,$E0,$45,$AF,$F6,$AC,$FF,$9F,$B5,$FC,$BC
	.dc.b	$7F,$26,$6D,$7F,$A0,$BE,$03,$FC,$D3,$CD,$FE,$C3,$E9,$FD,$1D,$FF
	.dc.b	$2C,$7F,$46,$FF,$2E,$EB,$FF,$6A,$DF,$FA,$29,$FE,$34,$FF,$B5,$3B
	.dc.b	$BF,$63,$DD,$FB,$7A,$FD,$A3,$FB,$4E,$EF,$F6,$84,$EA,$3F,$DA,$5E
	.dc.b	$37,$F0,$2F,$0F,$F8,$27,$88,$F1,$72,$FF,$08,$F1,$BF,$86,$3F,$C4
	.dc.b	$32,$FF,$62,$F0,$FF,$B0,$F4,$FF,$09,$9F,$EB,$43,$F8,$F7,$91,$FE
	.dc.b	$9C,$9F,$F4,$2C,$64,$FE,$70,$7F,$28,$7F,$D5,$DD,$9F,$CE,$EB,$F1
	.dc.b	$8F,$FB,$12,$76,$7F,$B1,$CF,$E7,$D5,$FE,$03,$E3,$FC,$FC,$7F,$B1
	.dc.b	$FB,$7F,$9C,$E1,$F1,$BA,$7F,$2D,$3F,$A0,$3F,$3E,$7F,$38,$3C,$87
	.dc.b	$4B,$FF,$6D,$BF,$D4,$A3,$AB,$FB,$1C,$79,$BF,$EC,$23,$FB,$1F,$E9
	.dc.b	$DF,$D9,$AB,$FD,$B1,$F8,$35,$9F,$B9,$EC,$BD,$FF,$F5,$01,$FE,$A1
	.dc.b	$8F,$C9,$1F,$D8,$BA,$FC,$B8,$FF,$15,$3F,$D9,$39,$7F,$70,$8F,$F3
	.dc.b	$9E,$3F,$ED,$1F,$17,$D7,$7F,$C6,$34,$FE,$61,$D9,$FA,$38,$FE,$21
	.dc.b	$18,$F1,$7F,$7F,$DF,$FC,$7D,$FE,$03,$EB,$7F,$8B,$7C,$6F,$DB,$5F
	.dc.b	$FB,$C5,$AF,$FA,$B3,$C2,$A8,$FF,$29,$B8,$71,$E1,$7E,$D7,$0B,$87
	.dc.b	$6D,$3F,$80,$57,$F0,$9D,$EE,$7F,$E0,$23,$F6,$D1,$FF,$00,$63,$FC
	.dc.b	$51,$AF,$BA,$7E,$3C,$7C,$49,$FE,$CF,$B7,$FB,$74,$F9,$33,$F3,$23
	.dc.b	$FD,$BC,$7F,$BC,$BC,$0F,$F6,$F7,$17,$F7,$08,$F5,$71,$FD,$9A,$FB
	.dc.b	$79,$7F,$B3,$27,$FC,$0C,$FF,$AE,$DF,$F1,$0F,$67,$EA,$7A,$BF,$4B
	.dc.b	$F8,$2F,$F6,$6E,$EF,$EC,$DF,$8D,$FF,$89,$AD,$F8,$6B,$E5,$61,$07
	.dc.b	$CA,$FF,$63,$DB,$F0,$DE,$BF,$FA,$67,$B0,$BF,$42,$FA,$1F,$59,$FE
	.dc.b	$7F,$C8,$FF,$A1,$DF,$8A,$7D,$5E,$5F,$F3,$47,$BF,$F8,$B8,$FF,$25
	.dc.b	$8F,$C9,$B9,$1F,$E8,$FE,$FF,$F4,$A4,$7F,$97,$DA,$FE,$77,$AB,$F1
	.dc.b	$6F,$F9,$86,$AB,$FE,$44,$EF,$FF,$44,$79,$5F,$E9,$3F,$00,$FF,$A6
	.dc.b	$FC,$3E,$7D,$3F,$E8,$36,$BF,$E6,$3F,$80,$FF,$72,$7B,$FF,$E9,$BF
	.dc.b	$1D,$F6,$1F,$E9,$DF,$22,$E9,$FE,$73,$F4,$3F,$D1,$B8,$FF,$9A,$FE
	.dc.b	$0F,$FA,$AB,$FE,$58,$F8,$34,$F6,$8F,$F5,$8E,$58,$3F,$D0,$47,$FB
	.dc.b	$B7,$C9,$F3,$E1,$FE,$88,$E2,$46,$23,$F7,$D3,$F9,$87,$E1,$79,$3F
	.dc.b	$D4,$F8,$BF,$D0,$73,$FE,$B9,$8F,$06,$1F,$E9,$93,$FC,$DF,$CB,$FE
	.dc.b	$7A,$3F,$D7,$1E,$27,$D8,$C3,$FD,$7B,$2E,$5F,$E9,$D1,$FE,$C1,$E7
	.dc.b	$FF,$60,$57,$E0,$23,$CC,$D7,$CA,$FF,$10,$EB,$FE,$1F,$D7,$F8,$98
	.dc.b	$FF,$3C,$F7,$FE,$47,$BF,$F2,$7D,$FF,$D1,$BC,$1F,$F1,$DF,$83,$FD
	.dc.b	$0B,$DA,$FF,$27,$78,$9F,$9E,$C7,$F3,$9A,$7F,$01,$ED,$FF,$7A,$BF
	.dc.b	$E7,$AF,$53,$FC,$CE,$BF,$0B,$BC,$D8,$FD,$4B,$3F,$73,$AF,$DD,$7F
	.dc.b	$99,$65,$FD,$CB,$0A,$7B,$F9,$BF,$D8,$0F,$F3,$93,$FC,$CF,$CB,$FC
	.dc.b	$04,$FE,$25,$F9,$3F,$F2,$28,$3F,$CA,$BC,$6F,$F0,$DC,$7F,$3A,$3F
	.dc.b	$D3,$2E,$1F,$E2,$4D,$BF,$6F,$A1,$FE,$43,$F0,$BF,$8C,$4E,$6B,$FC
	.dc.b	$6B,$87,$FB,$AF,$F0,$EF,$E6,$32,$7F,$9C,$F0,$BF,$03,$FF,$7A,$36
	.dc.b	$FF,$18,$BF,$11,$FE,$A7,$EA,$23,$13,$E7,$3F,$CA,$9E,$26,$1F,$F1
	.dc.b	$84,$FE,$72,$BF,$AD,$E3,$F9,$EC,$3F,$9A,$57,$E5,$AF,$C0,$FE,$E7
	.dc.b	$C1,$F3,$3D,$4F,$F1,$5E,$5F,$D2,$A3,$DD,$1F,$EB,$ED,$7F,$9D,$EB
	.dc.b	$FD,$23,$55,$6A,$BB,$F2,$7F,$BD,$63,$11,$E2,$9D,$D7,$F3,$9E,$0F
	.dc.b	$E6,$A7,$F3,$BF,$23,$FC,$F1,$3F,$CE,$EB,$FD,$CD,$1F,$E7,$6E,$1F
	.dc.b	$C0,$1F,$C6,$E3,$F8,$E7,$17,$F7,$67,$FD,$8F,$CB,$FD,$D1,$FE,$A9
	.dc.b	$C9,$57,$2D,$E6,$C6,$D9,$E7,$FC,$D5,$7F,$9A,$34,$F1,$3F,$4C,$C1
	.dc.b	$E0,$CB,$6F,$E4,$8F,$F3,$41,$FE,$22,$FC,$9F,$F2,$47,$E4,$78,$FF
	.dc.b	$0A,$58,$FB,$9C,$5F,$71,$FB,$D8,$AD,$CB,$FC,$62,$77,$8E,$3F,$0A
	.dc.b	$FC,$79,$F8,$EE,$3F,$BF,$7F,$EF,$6F,$17,$F8,$FC,$62,$3F,$88,$1F
	.dc.b	$C4,$63,$EC,$CF,$F8,$3B,$94,$FE,$07,$4F,$F0,$2E,$3F,$79,$FC,$04
	.dc.b	$F8,$E3,$FD,$33,$9F,$DF,$9E,$47,$F7,$5D,$7F,$B8,$9F,$DD,$FA,$BE
	.dc.b	$ED,$7F,$BC,$39,$3E,$63,$F2,$47,$57,$F5,$78,$C9,$F7,$E3,$F9,$1E
	.dc.b	$8F,$C6,$5E,$E7,$E7,$BE,$79,$E6,$3F,$0C,$7E,$06,$F7,$3F,$B4,$FE
	.dc.b	$23,$FE,$17,$C7,$8E,$7F,$E1,$9C,$78,$E1,$C2,$FF,$B7,$FD,$37,$F8
	.dc.b	$77,$C3,$7E,$75,$FA,$CF,$DF,$1F,$94,$BE,$FB,$3F,$B3,$3E,$91,$BF
	.dc.b	$1C,$38,$5E,$D5,$6F,$F2,$5F,$F1,$57,$94,$FB,$DF,$DC,$E4,$FD,$21
	.dc.b	$F2,$CF,$C4,$5C,$7F,$B2,$FD,$47,$8E,$1C,$2F,$75,$AF,$05,$FA,$07
	.dc.b	$6F,$D6,$BF,$82,$FA,$2F,$DD,$9C,$AF,$A9,$7E,$5C,$FC,$4B,$FD,$AD
	.dc.b	$DB,$FC,$6F,$1E,$38,$70,$BD,$D7,$FA,$87,$EC,$3F,$88,$FE,$E8,$FC
	.dc.b	$C5,$F9,$33,$F3,$17,$EF,$5F,$ED,$2F,$C1,$7F,$C9,$5E,$93,$1E,$38
	.dc.b	$70,$BD,$FB,$6F,$C7,$F7,$97,$F1,$EF,$DB,$1F,$BC,$7C,$9F,$01,$F9
	.dc.b	$8B,$5E,$9C,$9F,$ED,$CC,$38,$5E,$EB,$5F,$A2,$78,$BE,$F3,$F9,$4F
	.dc.b	$D4,$7D,$1A,$7E,$C4,$FD,$FB,$EE,$FE,$E5,$BC,$FF,$CC,$57,$BA,$D7
	.dc.b	$96,$FC,$9F,$DB,$E8,$7F,$34,$F5,$FF,$9A,$E1,$F0,$CF,$9B,$D1,$F9
	.dc.b	$77,$FD,$1F,$F6,$3F,$C0,$FD,$6E,$31,$FE,$92,$FD,$37,$FD,$C4,$FF
	.dc.b	$BD,$B0,$E8,$FF,$40,$78,$3F,$EF,$AF,$D7,$3F,$81,$7B,$5F,$EA,$78
	.dc.b	$EF,$18,$8F,$F1,$29,$FA,$A8,$E0,$7F,$9F,$5F,$E1,$3D,$5E,$0C,$95
	.dc.b	$C7,$DC,$EF,$18,$BF,$F5,$3B,$B9,$F2,$C7,$FA,$1F,$DE,$DC,$62,$3F
	.dc.b	$D6,$9F,$09,$F8,$28,$C5,$F4,$FD,$73,$CD,$BF,$85,$F4,$7D,$9E,$70
	.dc.b	$F3,$69,$FE,$5E,$AD,$46,$23,$FC,$55,$3E,$6F,$BA,$F1,$F8,$3F,$D9
	.dc.b	$B3,$FC,$8C,$79,$B8,$BF,$C6,$DC,$58,$C7,$FD,$9B,$C5,$B5,$C0,$BF
	.dc.b	$6F,$17,$09,$CF,$2F,$D0,$AE,$07,$E2,$A3,$C9,$C5,$B8,$C4,$7C,$E3
	.dc.b	$E7,$4F,$D8,$E1,$D3,$F2,$A7,$9B,$F3,$2B,$35,$FE,$E2,$E3,$CB,$FE
	.dc.b	$E4,$BE,$DF,$9E,$7E,$0B,$8B,$C6,$FE,$13,$87,$35,$F2,$BE,$63,$FD
	.dc.b	$49,$C3,$FB,$4A,$FE,$95,$0A,$F4,$7E,$F1,$FE,$0E,$FB,$4F,$F4,$57
	.dc.b	$CB,$3F,$7B,$FE,$6D,$FA,$87,$ED,$F3,$F8,$F7,$BF,$9D,$FF,$4F,$D6
	.dc.b	$B5,$FE,$03,$F8,$77,$E6,$9F,$F0,$85,$FC,$E8,$FC,$EE,$7F,$D6,$B9
	.dc.b	$F7,$3F,$ED,$1A,$FF,$69,$74,$F0,$E5,$61,$FA,$4E,$9E,$17,$CC,$E0
	.dc.b	$7F,$92,$CF,$D8,$C7,$EA,$B8,$F7,$5C,$A7,$7D,$3D,$6B,$F5,$FD,$1A
	.dc.b	$7F,$B3,$D3,$F8,$03,$F5,$9D,$7F,$A4,$7F,$B9,$47,$95,$FE,$E9,$D3
	.dc.b	$D9,$FE,$ED,$AE,$72,$FC,$DE,$F9,$7F,$3B,$D3,$97,$FD,$15,$AE,$5F
	.dc.b	$F4,$61,$9C,$BF,$D1,$06,$6D,$F2,$1F,$ED,$3E,$AF,$E5,$F9,$7F,$08
	.dc.b	$AD,$5F,$F9,$8C,$FE,$27,$1F,$C3,$33,$CE,$1E,$13,$FE,$00,$67,$C3
	.dc.b	$3C,$5F,$F9,$67,$0F,$CE,$61,$F1,$3E,$57,$E1,$C7,$F0,$DD,$51,$F0
	.dc.b	$2B,$F4,$AF,$F8,$5F,$0F,$A2,$FF,$86,$63,$EE,$3F,$E1,$EE,$AF,$F0
	.dc.b	$DD,$F3,$8C,$65,$AA,$FE,$31,$5F,$97,$AF,$D7,$57,$19,$D1,$FE,$2F
	.dc.b	$FA,$1E,$7C,$BF,$D4,$BE,$43,$F9,$B6,$3F,$1E,$BF,$7F,$1F,$CE,$72
	.dc.b	$DC,$60,$FC,$97,$36,$F4,$57,$F3,$FE,$6D,$C6,$2B,$F3,$1B,$EE,$BF
	.dc.b	$A2,$71,$E9,$FE,$8D,$CD,$9A,$FC,$F7,$36,$5F,$F2,$49,$FC,$4B,$E5
	.dc.b	$7E,$6E,$DF,$DA,$9F,$F2,$1E,$5E,$37,$FC,$8B,$C9,$E2,$7F,$C8,$FB
	.dc.b	$E6,$FF,$9F,$6F,$97,$FA,$06,$B9,$7F,$CA,$9A,$E5,$FF,$59,$9F,$AF
	.dc.b	$BF,$F4,$7F,$1F,$FB,$C6,$3C,$80,$7E,$57,$FE,$AC,$72,$31,$E1,$9F
	.dc.b	$FB,$1E,$3F,$78,$F8,$63,$FA,$AF,$27,$F2,$7C,$3E,$AD,$78,$4F,$E0
	.dc.b	$59,$F9,$6B,$F7,$91,$FD,$B2,$7E,$6D,$7E,$F1,$FD,$BF,$CB,$7C,$89
	.dc.b	$FE,$76,$FC,$47,$F8,$53,$FE,$6C,$BF,$B2,$FF,$9B,$A3,$EF,$3F,$02
	.dc.b	$B3,$1E,$FC,$BD,$8F,$23,$15,$A7,$FD,$F8,$72,$3E,$37,$27,$C2,$F9
	.dc.b	$2F,$F5,$CF,$EE,$1F,$EF,$5F,$9F,$7A,$7F,$E2,$69,$CC,$7F,$BE,$7D
	.dc.b	$0C,$61,$F3,$FD,$47,$DE,$C2,$7C,$6E,$A5,$39,$9D,$7C,$46,$7F,$15
	.dc.b	$FA,$DC,$9F,$E7,$4B,$FB,$5F,$8B,$79,$5F,$B9,$F8,$B7,$97,$6E,$8F
	.dc.b	$9F,$4E,$DF,$81,$F5,$CF,$C7,$AE,$1C,$69,$E3,$F9,$5E,$FC,$1F,$ED
	.dc.b	$8F,$C2,$7B,$68,$FC,$13,$B3,$FD,$C2,$7F,$DB,$B1,$DE,$30,$7F,$9E
	.dc.b	$F0,$F0,$63,$EF,$F9,$0F,$14,$32,$F8,$5F,$71,$E0,$BF,$CB,$79,$BE
	.dc.b	$0F,$8A,$EF,$C7,$F0,$67,$FC,$C7,$CD,$F3,$DE,$E3,$F3,$FE,$BA,$7C
	.dc.b	$D7,$BE,$5F,$3F,$F2,$CE,$7B,$FF,$BD,$1F,$46,$3D,$7F,$1D,$E3,$8F
	.dc.b	$47,$8B,$DB,$F2,$7F,$32,$FA,$8F,$8B,$FA,$4F,$48,$FD,$73,$D3,$1F
	.dc.b	$F8,$03,$EA,$74,$C7,$D4,$F4,$7D,$9F,$03,$D3,$F7,$5F,$9F,$FB,$DF
	.dc.b	$3C,$4F,$37,$F7,$5F,$BF,$73,$C7,$F7,$90,$8E,$91,$8F,$B3,$F3,$FC
	.dc.b	$C7,$DA,$FD,$17,$A1,$CE,$FF,$81,$F4,$FD,$67,$E5,$F9,$BE,$5F,$6B
	.dc.b	$AF,$91,$E9,$6D,$7E,$22,$BF,$C6,$31,$E2,$FC,$3B,$EB,$B7,$AB,$EB
	.dc.b	$70,$FF,$60,$AF,$E2,$5E,$9F,$E7,$2B,$D3,$F9,$27,$F1,$07,$D7,$FB
	.dc.b	$57,$7A,$FE,$30,$F8,$DF,$E3,$5F,$BB,$76,$6E,$3D,$AB,$E1,$FC,$BF
	.dc.b	$59,$FE,$41,$F9,$77,$57,$F9,$17,$C7,$75,$7F,$D1,$5B,$7A,$F5,$F3
	.dc.b	$43,$F3,$0E,$CD,$CB,$B3,$FC,$AF,$F0,$8E,$CB,$F2,$0E,$CB,$F5,$6E
	.dc.b	$C7,$A5,$7E,$6F,$D0,$FC,$8E,$DF,$E3,$9C,$FE,$96,$7E,$2F,$8B,$E1
	.dc.b	$F8,$2F,$88,$DE,$1F,$CE,$FF,$74,$F0,$FE,$E3,$E0,$FD,$77,$D8,$77
	.dc.b	$FD,$E7,$D2,$70,$A7,$D1,$7A,$3F,$13,$F3,$FE,$47,$F0,$CF,$37,$D7
	.dc.b	$FC,$03,$87,$EA,$5E,$EF,$D4,$BD,$BF,$3F,$F2,$4F,$67,$DB,$7C,$63
	.dc.b	$B7,$E9,$DF,$34,$F8,$1F,$81,$FC,$A3,$C5,$F5,$3F,$23,$F2,$7F,$8F
	.dc.b	$F9,$1F,$D3,$3E,$37,$E9,$FF,$40,$F4,$7D,$B7,$D7,$3E,$57,$7F,$D4
	.dc.b	$3D,$9F,$0D,$F4,$BF,$4C,$FA,$3F,$A9,$7D,$D3,$E8,$7E,$45,$F6,$8F
	.dc.b	$85,$F1,$9F,$81,$3E,$A7,$C5,$7D,$E3,$E3,$7C,$47,$DC,$F8,$8F,$B7
	.dc.b	$F0,$FF,$87,$3E,$D7,$C2,$FE,$10,$F9,$9F,$03,$F8,$B3,$EF,$7B,$FF
	.dc.b	$88,$3E,$87,$BD,$F8,$6F,$7B,$F0,$BE,$EF,$E4,$4F,$C2,$7B,$5F,$8E
	.dc.b	$3F,$89,$F8,$5F,$A9,$E3,$10,$FE,$A2,$FF,$91,$7D,$8D,$BF,$DF,$7D
	.dc.b	$4F,$E8,$EF,$F0,$4F,$53,$FB,$DB,$FC,$1B,$D0,$FE,$24,$DF,$3B,$F4
	.dc.b	$0F,$F0,$AF,$33,$F9,$7B,$71,$FD,$CB,$FC,$3B,$C8,$EA,$FF,$82,$FC
	.dc.b	$8E,$AF,$F8,$37,$C4,$EA,$DC,$3A,$BF,$E1,$1F,$0B,$AB,$70,$EA,$FF
	.dc.b	$85,$7A,$F8,$A7,$F9,$33,$FC,$7B,$A7,$8A,$7E,$0B,$FC,$86,$FF,$52
	.dc.b	$FF,$24,$E1,$FE,$73,$F2,$DF,$E2,$6F,$5F,$F9,$EB,$6B,$F8,$D3,$FE
	.dc.b	$2A,$F2,$7F,$AE,$B7,$2F,$E9,$CF,$F8,$BB,$C3,$E3,$F2,$3F,$CC,$1E
	.dc.b	$AF,$F3,$2E,$9E,$2D,$AF,$EA,$DF,$2D,$FE,$0E,$D1,$33,$E4,$BF,$9C
	.dc.b	$FA,$9F,$CE,$AB,$53,$9A,$FF,$1C,$4B,$E7,$BF,$C7,$55,$FE,$C1,$9F
	.dc.b	$F5,$49,$FD,$08,$E5,$F9,$3F,$F8,$AE,$1F,$5D,$FE,$FD,$9F,$F2,$15
	.dc.b	$72,$8C,$43,$F2,$FC,$3D,$0E,$82,$9F,$8F,$FF,$A2,$BB,$7C,$6F,$7F
	.dc.b	$C9,$39,$DF,$99,$EB,$FB,$87,$A3,$F1,$0F,$F4,$8F,$E8,$57,$2F,$F2
	.dc.b	$A6,$FF,$E8,$A1,$F0,$1E,$EF,$9F,$FE,$BB,$F3,$FF,$1A,$BD,$DB,$27
	.dc.b	$31,$F1,$8F,$69,$E7,$FD,$1F,$FB,$2E,$1F,$40,$FA,$51,$F2,$A3,$31
	.dc.b	$F3,$E3,$E7,$1E,$FF,$C1,$FA,$BF,$E1,$13,$E8,$B8,$F8,$EF,$EE,$3C
	.dc.b	$7F,$53,$F2,$5E,$76,$FC,$3B,$E4,$E6,$9C,$C7,$0C,$7E,$67,$E4,$1E
	.dc.b	$AE,$8F,$B5,$B7,$DE,$FC,$93,$FD,$FF,$5E,$D3,$E9,$7E,$C1,$F0,$35
	.dc.b	$FF,$00,$9F,$89,$FE,$01,$F5,$E5,$FE,$04,$FC,$7F,$F1,$97,$FE,$06
	.dc.b	$F8,$77,$ED,$0F,$DA,$9F,$0B,$85,$F0,$DF,$8E,$BE,$15,$BA,$D7,$C6
	.dc.b	$66,$7F,$D1,$9F,$B6,$F2,$AD,$0F,$C0,$75,$29,$D6,$3B,$FC,$0B,$FD
	.dc.b	$29,$D7,$FE,$9A,$FD,$69,$F4,$0F,$D3,$51,$85,$E0,$70,$EC,$FF,$10
	.dc.b	$CE,$4D,$78,$19,$D3,$FD,$41,$FA,$07,$39,$D7,$E0,$39,$94,$E7,$2D
	.dc.b	$7C,$07,$93,$CC,$FF,$16,$F0,$7A,$39,$3E,$4F,$16,$71,$F8,$31,$F2
	.dc.b	$4F,$9B,$C3,$EC,$FC,$8B,$CB,$97,$5E,$1F,$4D,$FE,$66,$73,$1E,$E9
	.dc.b	$F8,$1B,$FC,$4F,$C6,$FF,$6B,$7F,$1B,$B3,$FD,$4D,$C9,$FE,$63,$E3
	.dc.b	$FE,$17,$D7,$F7,$EB,$47,$EB,$23,$F2,$4F,$FB,$07,$AF,$F3,$BE,$33
	.dc.b	$E5,$3A,$CA,$1F,$BD,$7F,$AC,$FF,$56,$FE,$31,$5F,$E6,$26,$FA,$FF
	.dc.b	$E8,$BF,$C4,$7F,$D9,$90,$FA,$1F,$F2,$9F,$AB,$FE,$A7,$FC,$1F,$3B
	.dc.b	$62,$3F,$AD,$1F,$D7,$23,$8F,$91,$DF,$E8,$FC,$78,$7F,$AE,$36,$FD
	.dc.b	$C6,$3F,$B9,$E0,$7D,$1F,$09,$F7,$76,$FF,$16,$57,$FC,$81,$F8,$55
	.dc.b	$C6,$BC,$FF,$AA,$FC,$0F,$3F,$F9,$44,$FF,$11,$E8,$FC,$BB,$FC,$4F
	.dc.b	$D7,$FF,$74,$79,$1A,$8F,$11,$FC,$5B,$14,$66,$3C,$D3,$FE,$EF,$FC
	.dc.b	$6B,$C5,$5F,$C7,$67,$D2,$A1,$8F,$D2,$7E,$04,$FF,$46,$D7,$D1,$5B
	.dc.b	$DF,$F8,$1B,$FC,$FE,$B7,$5E,$4D,$3C,$1B,$F3,$E2,$FA,$1C,$5F,$43
	.dc.b	$8B,$C9,$9E,$23,$FC,$19,$1E,$7E,$2F,$F4,$5F,$2F,$FA,$83,$3F,$BF
	.dc.b	$CB,$FE,$D5,$E5,$FA,$1B,$7A,$F8,$7F,$03,$EC,$FF,$43,$8F,$E6,$9F
	.dc.b	$99,$5F,$78,$E7,$1F,$87,$C5,$F3,$6B,$D8,$7D,$67,$F9,$B6,$1F,$D7
	.dc.b	$BF,$58,$FE,$57,$3F,$CA,$9A,$FF,$3E,$E8,$7C,$E6,$23,$FB,$53,$7D
	.dc.b	$7F,$EC,$EF,$F3,$EE,$0F,$97,$4F,$D6,$B5,$7F,$D7,$5E,$5E,$EF,$F6
	.dc.b	$E7,$A4,$53,$F1,$F9,$FF,$4B,$E8,$FE,$E7,$F3,$7C,$27,$FB,$33,$E3
	.dc.b	$F7,$73,$1F,$92,$3F,$A1,$C7,$83,$13,$E5,$BF,$83,$3C,$8E,$EF,$F8
	.dc.b	$22,$FF,$C0,$3F,$73,$FD,$3B,$DA,$FE,$9A,$FC,$69,$D5,$7F,$77,$8F
	.dc.b	$4B,$4F,$11,$F2,$5F,$03,$FD,$B1,$AF,$F6,$2C,$1F,$31,$FD,$73,$C7
	.dc.b	$FE,$B6,$F8,$6F,$BC,$EA,$BA,$9F,$E1,$B9,$F7,$1F,$47,$6E,$31,$FB
	.dc.b	$87,$FB,$14,$BE,$D7,$FB,$4D,$FF,$77,$AF,$F0,$FC,$62,$3F,$C5,$3C
	.dc.b	$38,$9E,$E5,$3F,$83,$1D,$27,$F4,$F9,$FF,$8B,$5B,$3B,$F0,$3F,$DD
	.dc.b	$3F,$75,$DE,$FF,$C3,$D5,$FE,$00,$8F,$F0,$45,$7F,$8C,$3E,$63,$E5
	.dc.b	$E4,$7E,$D1,$FE,$0A,$F0,$3F,$B9,$E1,$B3,$F9,$93,$FD,$DA,$BF,$C7
	.dc.b	$37,$FE,$F8,$FA,$8F,$F3,$EF,$E0,$FF,$C2,$AF,$FC,$3B,$39,$F0,$93
	.dc.b	$99,$FA,$93,$FB,$6F,$01,$F9,$CF,$F2,$79,$EA,$0F,$E7,$D1,$AD,$BE
	.dc.b	$7D,$7F,$04,$9F,$F0,$AB,$5F,$D6,$3C,$5F,$74,$F6,$F1,$F7,$1D,$75
	.dc.b	$4E,$63,$F6,$9F,$11,$E5,$C7,$C9,$1F,$E4,$98,$C6,$3F,$80,$FC,$CB
	.dc.b	$C1,$3F,$E6,$16,$ED,$FE,$5E,$FC,$BB,$33,$4F,$66,$DF,$EA,$37,$F1
	.dc.b	$27,$EA,$CF,$A4,$7C,$1E,$7F,$F1,$0B,$F4,$8C,$57,$F8,$43,$F4,$4F
	.dc.b	$E0,$3F,$0F,$FC,$AB,$9F,$FA,$53,$6B,$FC,$6C,$FF,$2E,$E8,$FE,$7B
	.dc.b	$1F,$E3,$BD,$3F,$71,$5F,$B0,$A7,$E6,$BF,$B0,$6D,$7E,$93,$1F,$A9
	.dc.b	$FA,$3E,$23,$F2,$E7,$F9,$E5,$C5,$FE,$4E,$3F,$CA,$71,$E1,$8F,$E0
	.dc.b	$AD,$7F,$7E,$F9,$E3,$96,$1F,$C7,$3C,$76,$7F,$C7,$B9,$3F,$AD,$74
	.dc.b	$7E,$5A,$FF,$18,$7F,$6E,$3E,$29,$8C,$7C,$4F,$F6,$23,$97,$07,$D4
	.dc.b	$7F,$B0,$BF,$61,$FE,$C7,$C3,$9B,$39,$5C,$5F,$82,$9F,$F4,$C4,$61
	.dc.b	$FC,$C7,$95,$F8,$CE,$4F,$F0,$16,$FF,$DC,$B9,$FF,$29,$CF,$9C,$3F
	.dc.b	$D2,$A7,$FA,$8F,$0F,$CB,$DF,$E9,$EF,$84,$7E,$23,$9B,$FD,$D6,$F0
	.dc.b	$C3,$C8,$78,$DF,$D0,$5F,$3C,$7C,$5C,$3F,$E7,$78,$ED,$D3,$FB,$87
	.dc.b	$95,$7E,$BE,$30,$FF,$A4,$A3,$F8,$33,$F9,$11,$CF,$DB,$4F,$CB,$C6
	.dc.b	$33,$FF,$07,$C7,$F3,$8A,$E3,$3F,$86,$3F,$D2,$AF,$F2,$77,$FD,$83
	.dc.b	$97,$F1,$98,$C4,$7F,$8A,$63,$F8,$CB,$FE,$16,$EB,$FE,$46,$FF,$87
	.dc.b	$1F,$E4,$6E,$4F,$E5,$65,$3C,$17,$B7,$FB,$B7,$0F,$FB,$69,$C1,$FE
	.dc.b	$1D,$8F,$F1,$3F,$6F,$F3,$F8,$FB,$AD,$7D,$2F,$B4,$FC,$A1,$FD,$35
	.dc.b	$D9,$FE,$38,$ED,$FE,$DC,$FF,$8B,$BB,$BF,$AA,$BF,$E3,$47,$FA,$BD
	.dc.b	$58,$F4,$E4,$9F,$D0,$1F,$D9,$E3,$F1,$EF,$F6,$7C,$3F,$C6,$73,$FD
	.dc.b	$CA,$7F,$B7,$C3,$C6,$FF,$20,$F8,$3F,$BE,$8D,$BF,$54,$FD,$61,$FE
	.dc.b	$01,$AE,$55,$F5,$FC,$2F,$F2,$4F,$87,$C7,$C3,$B3,$E6,$8F,$DA,$1C
	.dc.b	$7C,$5F,$C0,$1B,$8F,$F2,$6B,$FB,$9F,$1B,$F0,$87,$EE,$FC,$2F,$B7
	.dc.b	$3F,$E5,$27,$ED,$F8,$A9,$E4,$FE,$63,$A7,$FA,$5F,$81,$F3,$23,$F3
	.dc.b	$DE,$27,$F4,$AE,$6C,$1F,$E7,$7F,$17,$FC,$4E,$FE,$C2,$3F,$A7,$D7
	.dc.b	$1E,$EF,$E7,$D3,$CF,$BF,$87,$1F,$5F,$02,$BF,$E0,$6D,$BD,$9E,$A7
	.dc.b	$F1,$47,$E7,$F1,$7F,$57,$F1,$7F,$90,$D3,$87,$FC,$C3,$E2,$E1,$FF
	.dc.b	$31,$EB,$C7,$0F,$F7,$E4,$7F,$99,$B9,$BF,$CE,$CD,$CF,$C3,$87,$DE
	.dc.b	$D2,$E7,$FE,$7C,$DB,$D3,$3F,$E2,$8C,$3F,$85,$77,$7F,$11,$CF,$F2
	.dc.b	$31,$FC,$37,$8F,$FB,$BD,$63,$83,$8B,$F8,$A7,$FB,$C7,$12,$F8,$F2
	.dc.b	$7D,$2F,$C8,$FF,$D2,$07,$F6,$DF,$0B,$FD,$31,$DB,$FD,$DF,$9D,$F9
	.dc.b	$CF,$F8,$7F,$3F,$F4,$3C,$63,$93,$FB,$E5,$F0,$51,$97,$FB,$79,$FD
	.dc.b	$E5,$7E,$E1,$FE,$75,$5F,$85,$78,$66,$1E,$17,$FD,$49,$FB,$5F,$F8
	.dc.b	$CA,$3F,$C0,$47,$67,$7E,$16,$23,$07,$90,$FF,$59,$1E,$4F,$C7,$3F
	.dc.b	$C1,$59,$7D,$68,$E7,$83,$F2,$DF,$0E,$7C,$E3,$13,$CC,$FF,$5E,$2F
	.dc.b	$CA,$BF,$C6,$27,$A7,$1C,$BF,$EB,$BE,$FF,$F5,$2D,$7C,$1E,$FF,$F1
	.dc.b	$57,$37,$FB,$43,$EB,$7E,$74,$62,$04,$FC,$1F,$35,$FD,$4A,$FF,$C3
	.dc.b	$A4,$FC,$13,$31,$FE,$2C,$8C,$8F,$C0,$F3,$7F,$ED,$E3,$AE,$38,$71
	.dc.b	$87,$2E,$CE,$37,$FE,$2A,$C7,$0E,$30,$FC,$27,$BF,$FE,$3F,$8E,$F3
	.dc.b	$8C,$BC,$D1,$E2,$AF,$C5,$78,$9F,$EB,$DC,$BC,$35,$F8,$68,$C4,$32
	.dc.b	$FF,$23,$69,$FE,$4D,$E2,$1F,$B1,$BF,$BD,$19,$8C,$47,$F9,$37,$A3
	.dc.b	$8D,$FE,$2A,$37,$D9,$F9,$9F,$03,$F7,$9F,$94,$66,$7F,$17,$9F,$E6
	.dc.b	$1F,$E9,$78,$FF,$B9,$23,$11,$FE,$CD,$0B,$F1,$1E,$51,$FE,$55,$87
	.dc.b	$E5,$BD,$E1,$E3,$ED,$FC,$51,$F9,$77,$6F,$83,$1F,$F1,$9F,$3F,$FB
	.dc.b	$66,$7C,$38,$FF,$BF,$8F,$F7,$BA,$CB,$C3,$7F,$E6,$DF,$07,$FC,$16
	.dc.b	$FE,$C0,$C6,$9E,$38,$F9,$1E,$3F,$E9,$E7,$F8,$27,$89,$F9,$6E,$0F
	.dc.b	$F3,$B3,$EE,$8F,$F0,$B7,$07,$E5,$D4,$65,$FC,$64,$7E,$5D,$CF,$37
	.dc.b	$E1,$79,$7F,$2D,$18,$3F,$0A,$FE,$6B,$97,$BF,$27,$F3,$B8,$FC,$5C
	.dc.b	$F6,$8C,$46,$E3,$07,$92,$FF,$4E,$3A,$39,$C6,$63,$C9,$8A,$FC,$97
	.dc.b	$F9,$DA,$E6,$7E,$A5,$ED,$3B,$8F,$F1,$55,$7E,$A6,$FA,$C3,$D1,$F8
	.dc.b	$3C,$FF,$23,$0F,$DF,$CF,$CA,$EA,$FA,$AF,$C8,$7F,$83,$3F,$E4,$67
	.dc.b	$FC,$59,$DB,$AC,$BF,$87,$CB,$B7,$F2,$D0,$EE,$FE,$33,$CD,$FA,$4E
	.dc.b	$2F,$F2,$7F,$7F,$13,$F9,$1E,$1F,$E5,$A7,$F9,$43,$FE,$4A,$F0,$3B
	.dc.b	$1F,$A5,$C7,$FC,$BF,$E0,$FF,$3D,$F3,$7F,$03,$E1,$F8,$CF,$CE,$F3
	.dc.b	$E3,$0F,$CB,$1F,$82,$F1,$3D,$06,$9F,$C2,$F8,$DE,$BF,$1F,$F4,$5D
	.dc.b	$FF,$D0,$F7,$FC,$D6,$5F,$D3,$73,$C9,$FD,$37,$E0,$3F,$A0,$63,$FA
	.dc.b	$6C,$7F,$DF,$7B,$F2,$3E,$0E,$1F,$DC,$CC,$CF,$EB,$1F,$F8,$03,$C3
	.dc.b	$E0,$37,$EA,$FF,$83,$79,$3F,$DD,$53,$FC,$33,$C2,$FE,$BD,$CD,$E6
	.dc.b	$3F,$C2,$A7,$F5,$28,$FF,$0A,$73,$FA,$8F,$F7,$63,$FE,$42,$EC,$FE
	.dc.b	$25,$87,$F1,$68,$FF,$58,$70,$7A,$F3,$FE,$02,$DE,$FD,$CF,$F7,$AC
	.dc.b	$39,$BF,$DF,$DF,$E1,$1D,$3F,$D3,$6F,$FA,$84,$FF,$80,$32,$FF,$19
	.dc.b	$4E,$67,$F8,$24,$FF,$92,$34,$F6,$BF,$E4,$AD,$7E,$0B,$FE,$0B,$7A
	.dc.b	$ED,$FC,$96,$3F,$C1,$CF,$CC,$E9,$FF,$06,$BE,$2E,$9F,$E5,$5C,$3F
	.dc.b	$E1,$3E,$3F,$66,$7F,$DE,$B3,$FC,$1B,$F7,$71,$F8,$47,$F3,$8E,$4F
	.dc.b	$BD,$3F,$C8,$0F,$E6,$C7,$F8,$1A,$BF,$94,$FB,$1A,$8F,$E3,$07,$F8
	.dc.b	$2B,$E4,$7F,$C0,$C7,$E0,$5F,$E5,$3D,$7F,$DE,$34,$FC,$D4,$3A,$BF
	.dc.b	$C3,$11,$F9,$3E,$4F,$D4,$48,$F2,$E7,$FD,$F7,$83,$F4,$B7,$FE,$1F
	.dc.b	$7F,$53,$3F,$E1,$76,$C6,$63,$D4,$64,$AF,$F8,$9F,$F2,$6F,$95,$1F
	.dc.b	$E0,$C6,$BE,$53,$FC,$4B,$C7,$FE,$0E,$7F,$C2,$2D,$75,$87,$DD,$CB
	.dc.b	$FC,$58,$B9,$3F,$43,$3F,$E1,$56,$BF,$32,$34,$FF,$88,$7C,$BF,$F0
	.dc.b	$BC,$FF,$10,$8D,$3F,$E1,$57,$D2,$7F,$8E,$3D,$8F,$E3,$2D,$3E,$46
	.dc.b	$9F,$C3,$DF,$94,$79,$BC,$4F,$F0,$C7,$87,$FD,$F3,$14,$62,$3F,$BE
	.dc.b	$4F,$F9,$12,$3F,$9D,$E1,$FC,$C6,$B7,$3F,$57,$CC,$FF,$08,$CF,$F8
	.dc.b	$66,$3F,$C9,$7F,$35,$FC,$E2,$1E,$6F,$F9,$56,$31,$E3,$FF,$95,$FF
	.dc.b	$32,$FF,$05,$F9,$17,$0F,$E8,$5A,$78,$BC,$1F,$F3,$14,$7B,$36,$FF
	.dc.b	$04,$BE,$C8,$CB,$FE,$45,$3F,$2E,$BC,$9F,$E7,$6D,$8F,$93,$E6,$FF
	.dc.b	$81,$B8,$FF,$C9,$F1,$FE,$06,$E0,$FE,$49,$3F,$D2,$E3,$FC,$14,$DC
	.dc.b	$3F,$BE,$78,$9F,$E8,$FF,$37,$FA,$3C,$FF,$81,$2B,$FC,$D3,$D1,$F9
	.dc.b	$9F,$37,$19,$76,$6C,$FF,$98,$67,$59,$FF,$82,$4F,$F0,$94,$F9,$67
	.dc.b	$FC,$C6,$D7,$FC,$D7,$3F,$E0,$67,$F3,$A5,$3C,$27,$FA,$27,$C5,$FC
	.dc.b	$29,$F1,$B3,$FC,$24,$7E,$F3,$C1,$70,$FD,$28,$FF,$03,$BF,$E3,$EF
	.dc.b	$2F,$FC,$1B,$C9,$F6,$F9,$6D,$7F,$81,$39,$BF,$3D,$CD,$FE,$12,$E8
	.dc.b	$FC,$3C,$7E,$02,$7F,$C1,$43,$A3,$F0,$F5,$B8,$FE,$EF,$CF,$FE,$17
	.dc.b	$D3,$F1,$B1,$FD,$DE,$3F,$CB,$EF,$F8,$6B,$A7,$F3,$F2,$7F,$C4,$51
	.dc.b	$FE,$1E,$C3,$FB,$A1,$FD,$76,$7F,$5B,$1F,$DD,$47,$57,$F8,$AE,$33
	.dc.b	$C9,$FC,$3E,$3E,$6C,$F3,$3F,$21,$C1,$BB,$FD,$CC,$7F,$05,$BF,$E0
	.dc.b	$0F,$DB,$3F,$C1,$07,$FA,$46,$FF,$C1,$0D,$3E,$87,$67,$F0,$E7,$FC
	.dc.b	$72,$FF,$10,$E3,$FF,$1D,$F6,$FF,$83,$4F,$51,$FC,$7A,$72,$7F,$A7
	.dc.b	$72,$FD,$EC,$E2,$3F,$93,$9F,$EB,$0C,$3F,$95,$B8,$3D,$1D,$DF,$14
	.dc.b	$FE,$66,$FF,$92,$9F,$E6,$B3,$FE,$4C,$EF,$FF,$07,$4F,$F9,$29,$FE
	.dc.b	$BF,$E0,$FF,$88,$4F,$9C,$7F,$44,$8F,$F1,$C7,$7F,$F9,$AF,$AD,$F9
	.dc.b	$77,$E7,$0F,$F2,$E4,$7F,$83,$AF,$FB,$87,$7B,$2F,$EB,$3C,$3F,$ED
	.dc.b	$8D,$BF,$AF,$3F,$E0,$6C,$BF,$C1,$0F,$FB,$73,$7F,$F0,$B9,$FD,$A1
	.dc.b	$FF,$6F,$F8,$3F,$DB,$3C,$0F,$F7,$27,$1F,$F8,$CB,$C6,$FE,$BF,$18
	.dc.b	$C3,$FB,$BD,$FF,$83,$5F,$E5,$47,$F7,$68,$FF,$76,$72,$FF,$80,$25
	.dc.b	$1F,$EA,$0E,$EC,$3F,$D5,$F8,$7F,$83,$4F,$F5,$8F,$27,$F7,$B8,$7F
	.dc.b	$A5,$9F,$F4,$05,$F0,$8F,$CE,$1F,$C7,$7C,$9F,$2B,$83,$C4,$7E,$4C
	.dc.b	$FC,$D1,$AE,$6F,$CE,$DF,$F9,$34,$FF,$08,$47,$FB,$87,$C7,$FF,$0F
	.dc.b	$65,$F8,$E8,$FC,$DD,$E1,$FF,$42,$EF,$FE,$1A,$E0,$FF,$07,$61,$FE
	.dc.b	$13,$9F,$D0,$ED,$FE,$0E,$7F,$C7,$D1,$F0,$9F,$C2,$1E,$D1,$C4,$F9
	.dc.b	$FF,$C2,$36,$3F,$AC,$B5,$FF,$04,$AF,$9F,$FF,$28,$9F,$90,$7F,$7D
	.dc.b	$1F,$E0,$A3,$FB,$7F,$0F,$E6,$1F,$AF,$D7,$FC,$EA,$77,$1F,$C8,$97
	.dc.b	$E5,$3F,$E0,$88,$F2,$BF,$E6,$E7,$FC,$1D,$1A,$7F,$58,$7F,$75,$8F
	.dc.b	$D8,$3C,$0F,$E5,$E5,$3B,$7E,$B3,$FB,$97,$E0,$BF,$BD,$7E,$1F,$C0
	.dc.b	$FF,$81,$FB,$7B,$BF,$E0,$98,$A7,$4B,$47,$F8,$8F,$F5,$1F,$E1,$F5
	.dc.b	$E6,$7F,$88,$D5,$3F,$C1,$07,$F0,$23,$E8,$BF,$E0,$97,$FC,$03,$48
	.dc.b	$FF,$4E,$B6,$3E,$9F,$E9,$7F,$CA,$A0,$FF,$83,$17,$EA,$FF,$E0,$5F
	.dc.b	$23,$E3,$BF,$E0,$7F,$BE,$B7,$FD,$0F,$B1,$F8,$D6,$23,$FD,$6D,$1F
	.dc.b	$E0,$9A,$FE,$99,$1F,$86,$7F,$C2,$BE,$2F,$FB,$52,$F8,$3F,$ED,$6F
	.dc.b	$0B,$97,$8F,$FE,$0E,$F2,$6E,$3F,$E2,$D2,$65,$F6,$BF,$4B,$FF,$08
	.dc.b	$37,$3F,$25,$7D,$EF,$CF,$FC,$3B,$1F,$0E,$7F,$ED,$40,$34,$E1,$7F
	.dc.b	$E1,$4D,$3F,$CA,$9E,$7B,$F3,$7F,$F0,$69,$FE,$AB,$5F,$57,$FD,$51
	.dc.b	$B5,$FE,$19,$ED,$FF,$19,$AF,$53,$9D,$7F,$8E,$A9,$FA,$B7,$F8,$60
	.dc.b	$FF,$09,$FE,$7F,$FE,$0E,$3F,$C2,$BF,$AC,$7F,$2D,$3F,$9A,$FA,$A7
	.dc.b	$FC,$12,$62,$3F,$DB,$70,$7F,$D8,$1E,$4F,$6C,$BF,$36,$D7,$FC,$12
	.dc.b	$FA,$1F,$F1,$61,$FE,$63,$71,$E1,$FF,$82,$A3,$FC,$9B,$18,$8F,$F1
	.dc.b	$AE,$7F,$E7,$98,$F8,$D3,$DB,$5F,$EE,$9C,$3F,$82,$F2,$7E,$94,$FC
	.dc.b	$D0,$E1,$FA,$9C,$5F,$50,$FA,$8E,$B9,$BF,$C2,$8D,$3F,$25,$CD,$FE
	.dc.b	$89,$AF,$CC,$C7,$F9,$5F,$9F,$F1,$07,$F9,$D7,$9F,$5D,$3F,$E1,$8B
	.dc.b	$B5,$FE,$52,$EA,$AF,$8B,$2F,$C2,$F3,$E1,$FC,$C4,$7A,$2F,$C7,$97
	.dc.b	$82,$3F,$43,$97,$E9,$1D,$1F,$E2,$0A,$F6,$61,$DB,$DE,$E7,$7F,$B2
	.dc.b	$AF,$C4,$1F,$BD,$70,$FD,$43,$DD,$DB,$FE,$7E,$F0,$BF,$CB,$73,$FB
	.dc.b	$D9,$FF,$21,$B6,$3F,$C3,$D1,$CE,$7F,$83,$3C,$8F,$F0,$F9,$92,$9F
	.dc.b	$E0,$BF,$03,$B3,$F9,$A3,$FD,$1B,$5F,$C6,$1F,$F4,$AB,$5F,$EA,$F5
	.dc.b	$E9,$9F,$E4,$11,$F7,$BC,$1F,$E5,$14,$7F,$94,$3F,$CC,$E3,$FD,$9A
	.dc.b	$DA,$F0,$47,$F8,$C7,$C2,$FF,$5F,$BF,$88,$7F,$9B,$CF,$1E,$6F,$F6
	.dc.b	$EC,$FF,$97,$E7,$FA,$03,$FD,$B6,$BE,$3F,$87,$FE,$47,$3F,$C4,$31
	.dc.b	$93,$DE,$7E,$2F,$C4,$FF,$7F,$9F,$E2,$4C,$3F,$AD,$57,$D0,$F1,$7F
	.dc.b	$CA,$07,$F8,$8E,$1F,$21,$CA,$FE,$44,$7F,$61,$CB,$FD,$17,$1F,$A0
	.dc.b	$9F,$F2,$C1,$FD,$A3,$8B,$FD,$69,$C7,$FC,$AF,$E3,$7F,$CD,$35,$FC
	.dc.b	$A4,$FC,$E7,$27,$E7,$3D,$5F,$E4,$11,$F9,$07,$F8,$6B,$FE,$5A,$C7
	.dc.b	$FC,$01,$E0,$BE,$2B,$FD,$8B,$1E,$2E,$DC,$56,$E3,$FC,$49,$1F,$BD
	.dc.b	$7F,$C2,$1D,$7F,$EB,$A9,$FC,$D9,$FE,$7D,$C7,$F5,$A7,$FB,$0F,$6F
	.dc.b	$D1,$67,$E8,$7F,$49,$CF,$FC,$F6,$BF,$C0,$FC,$5F,$E4,$2E,$3F,$F2
	.dc.b	$0F,$9B,$FD,$D3,$97,$CB,$A0,$FF,$12,$F8,$FF,$75,$FF,$10,$1F,$E1
	.dc.b	$37,$F0,$D8,$70,$AF,$EF,$BC,$FC,$AB,$8B,$FE,$47,$8F,$EE,$33,$FE
	.dc.b	$2C,$F3,$BF,$C4,$97,$FA,$9B,$FA,$AD,$5E,$3F,$8B,$AF,$F0,$B1,$C3
	.dc.b	$CA,$EC,$DC,$3F,$C7,$8F,$4E,$8F,$F7,$77,$C9,$7D,$BE,$3F,$8F,$F2
	.dc.b	$5F,$D8,$3C,$0F,$E3,$F3,$F6,$B9,$7F,$DD,$B3,$FD,$4B,$1F,$E2,$54
	.dc.b	$AF,$BB,$1F,$17,$18,$7C,$7F,$7B,$2F,$C9,$F4,$7F,$C9,$BD,$7F,$7A
	.dc.b	$BF,$DB,$BE,$57,$A2,$BE,$77,$A5,$FE,$D9,$8C,$42,$3F,$9E,$76,$FF
	.dc.b	$A1,$3B,$7F,$A0,$63,$FE,$58,$7F,$0D,$5E,$3C,$BF,$99,$78,$FF,$DB
	.dc.b	$B5,$FE,$DD,$AF,$F3,$5E,$E6,$BF,$E0,$9D,$FF,$96,$57,$C2,$8F,$C6
	.dc.b	$F0,$7C,$27,$F1,$FD,$BF,$CC,$F0,$FE,$7D,$C9,$FE,$6B,$EE,$FE,$4B
	.dc.b	$AF,$EB,$1F,$E4,$17,$FE,$BA,$EA,$FE,$3C,$FC,$7E,$8F,$F3,$57,$47
	.dc.b	$BA,$7F,$A4,$78,$9F,$E8,$9A,$FE,$2C,$7F,$9A,$30,$FF,$43,$61,$FC
	.dc.b	$5E,$3F,$4D,$1F,$E6,$38,$FC,$8C,$7B,$23,$7D,$BF,$E9,$A7,$4F,$78
	.dc.b	$FE,$1C,$7C,$2E,$CE,$BD,$FF,$43,$A3,$FA,$D7,$87,$FE,$63,$FA,$DE
	.dc.b	$1D,$FD,$B0,$FF,$6E,$7A,$3F,$A7,$7F,$CC,$DE,$5F,$E8,$1F,$E0,$07
	.dc.b	$CE,$70,$FF,$B0,$E2,$F8,$FF,$4A,$3F,$DC,$A7,$D4,$8C,$BF,$D2,$0F
	.dc.b	$F0,$C1,$FE,$5E,$3F,$CF,$7F,$AC,$FE,$B7,$8F,$95,$FF,$9B,$23,$FC
	.dc.b	$C9,$87,$D2,$AF,$F4,$CD,$FF,$96,$5F,$E8,$A7,$94,$FF,$44,$F2,$7F
	.dc.b	$96,$63,$FC,$EA,$FF,$96,$EE,$CF,$FA,$2B,$F5,$BF,$F5,$E3,$E0,$3F
	.dc.b	$D2,$2F,$F9,$6F,$AF,$FC,$AD,$1F,$E9,$47,$DD,$D1,$FD,$EF,$93,$F5
	.dc.b	$DE,$1F,$F1,$8E,$0F,$E5,$3E,$9F,$F1,$A8,$FE,$F1,$AB,$0F,$8D,$3F
	.dc.b	$E5,$3C,$7F,$D3,$FF,$96,$FF,$94,$1A,$7F,$98,$F0,$FF,$46,$FE,$7F
	.dc.b	$FA,$1E,$EF,$F0,$D7,$AB,$8A,$FF,$07,$0F,$1F,$FC,$61,$CB,$FD,$7B
	.dc.b	$8F,$F7,$AF,$F6,$09,$FF,$28,$FE,$89,$F0,$1F,$91,$48,$C4,$3E,$47
	.dc.b	$6F,$F8,$28,$FB,$87,$7F,$23,$FC,$B0,$7F,$55,$7F,$CA,$BE,$D7,$F9
	.dc.b	$2D,$FD,$AE,$1E,$35,$19,$7F,$DA,$67,$F8,$0D,$FF,$10,$77,$FF,$5C
	.dc.b	$F5,$B5,$5E,$9A,$46,$A7,$F8,$B9,$FE,$CB,$77,$1F,$E0,$DC,$FF,$87
	.dc.b	$BE,$23,$03,$C9,$FE,$35,$7D,$5F,$85,$1E,$F3,$F2,$EE,$1F,$F2,$DC
	.dc.b	$FF,$95,$73,$F8,$F9,$FF,$95,$FD,$8F,$F3,$29,$F0,$EB,$FC,$3B,$C1
	.dc.b	$F6,$78,$3F,$D0,$BC,$1F,$E4,$D8,$FF,$2D,$71,$7E,$07,$9C,$FE,$AF
	.dc.b	$C7,$FC,$57,$27,$F1,$FC,$7F,$BB,$F4,$7F,$A6,$1F,$F2,$C1,$FD,$A5
	.dc.b	$FF,$2A,$FA,$BB,$8F,$F2,$4C,$DE,$AF,$F2,$2F,$4F,$E6,$B5,$FF,$32
	.dc.b	$E7,$FE,$5F,$EB,$FC,$0F,$65,$C7,$E1,$71,$FF,$26,$7F,$C6,$BD,$BF
	.dc.b	$AD,$7F,$CD,$EF,$EB,$5F,$F3,$8F,$8F,$FE,$D7,$69,$F4,$47,$3F,$EB
	.dc.b	$F1,$77,$FF,$01,$9F,$DD,$46,$1F,$E4,$5D,$5F,$DE,$7C,$0F,$F2,$D7
	.dc.b	$9F,$F3,$73,$FF,$3B,$3F,$E3,$AB,$FE,$59,$7D,$8F,$E2,$1A,$69,$FE
	.dc.b	$36,$7F,$12,$8D,$E1,$FE,$3E,$8F,$F2,$21,$E7,$3F,$92,$9F,$E4,$53
	.dc.b	$F9,$29,$FE,$8E,$F0,$BF,$C7,$6E,$CF,$4F,$89,$F8,$1E,$BF,$EA,$BE
	.dc.b	$1A,$BF,$B1,$F9,$9F,$CB,$EF,$F4,$3B,$FF,$AF,$2B,$C7,$8F,$F7,$39
	.dc.b	$C6,$9F,$E4,$AF,$80,$FF,$34,$57,$BA,$FF,$48,$FC,$7A,$FE,$F3,$BE
	.dc.b	$B8,$7F,$CC,$F3,$FE,$80,$AF,$E6,$55,$FE,$FC,$E2,$FF,$7D,$F1,$FF
	.dc.b	$BD,$F9,$7F,$C8,$F3,$FC,$7C,$FF,$47,$72,$FF,$B7,$3D,$5F,$E3,$BD
	.dc.b	$5F,$E3,$6F,$4F,$DB,$DB,$FE,$AA,$D3,$F3,$38,$7F,$BE,$E7,$F8,$B3
	.dc.b	$FE,$10,$F5,$3F,$85,$BF,$B8,$E2,$FF,$1A,$79,$B8,$7F,$D8,$91,$FE
	.dc.b	$E6,$F9,$3F,$EB,$DE,$3F,$EF,$E7,$FC,$2B,$CD,$FE,$00,$F1,$F1,$1F
	.dc.b	$E0,$7B,$FD,$E5,$FF,$65,$F1,$FF,$71,$3D,$27,$F8,$07,$8F,$FD,$A7
	.dc.b	$B5,$E2,$FE,$D7,$8D,$B7,$FB,$07,$87,$FD,$66,$7F,$95,$EB,$F9,$D6
	.dc.b	$1F,$B2,$E2,$1F,$E5,$7E,$4F,$93,$3A,$97,$85,$FE,$C1,$E5,$FD,$64
	.dc.b	$7F,$AC,$4F,$D4,$9F,$89,$3F,$D4,$7C,$FF,$EA,$C7,$F5,$27,$F5,$E3
	.dc.b	$FD,$4F,$E1,$7F,$A7,$BB,$7F,$87,$47,$E7,$4F,$D0,$1F,$DC,$8F,$F5
	.dc.b	$6B,$FE,$98,$C3,$FC,$01,$D5,$FD,$DE,$7F,$34,$7F,$84,$B9,$7F,$29
	.dc.b	$19,$F3,$3E,$AC,$BA,$BF,$C1,$11,$FE,$45,$8F,$EF,$47,$DB,$E4,$5F
	.dc.b	$F8,$19,$FF,$13,$1B,$AF,$CE,$BF,$E6,$23,$AB,$F9,$B9,$FF,$38,$1D
	.dc.b	$2F,$CD,$3C,$72,$FD,$4C,$F8,$DF,$F3,$A7,$A5,$F8,$1B,$FD,$15,$7F
	.dc.b	$96,$E3,$FC,$E7,$DB,$FE,$33,$9F,$F0,$56,$5B,$D7,$9C,$7E,$97,$1F
	.dc.b	$F1,$F1,$FE,$0A,$3F,$8C,$61,$D7,$B7,$67,$FA,$33,$1F,$C9,$CF,$F8
	.dc.b	$20,$FF,$3B,$CF,$B3,$87,$FC,$EF,$1F,$AB,$3F,$07,$CA,$21,$7F,$06
	.dc.b	$FF,$05,$F2,$7E,$30,$FC,$6E,$5F,$EB,$49,$D9,$C2,$7F,$67,$3B,$1E
	.dc.b	$97,$ED,$2F,$FC,$C8,$7F,$83,$07,$93,$FE,$15,$C3,$FD,$09,$87,$43
	.dc.b	$C5,$3F,$7F,$A7,$F4,$9E,$97,$EE,$A7,$43,$2F,$DC,$69,$FD,$CB,$92
	.dc.b	$BF,$E8,$AF,$43,$3B,$FC,$6F,$3F,$F6,$72,$8F,$F4,$5F,$7F,$DC,$93
	.dc.b	$FE,$59,$D3,$FC,$0F,$EA,$DF,$0B,$F7,$F8,$7F,$AE,$FE,$B3,$F8,$14
	.dc.b	$FF,$A3,$9B,$5F,$E6,$CC,$BF,$D0,$BE,$17,$48,$FC,$F1,$FC,$18,$D1
	.dc.b	$FC,$10,$FB,$37,$F9,$A7,$07,$E2,$DF,$F5,$37,$89,$FE,$01,$8E,$73
	.dc.b	$EC,$9F,$81,$0E,$DF,$E1,$AF,$FA,$5B,$83,$FD,$20,$E7,$8B,$F8,$87
	.dc.b	$0F,$FA,$7A,$35,$3F,$4E,$3F,$89,$47,$DA,$AF,$E5,$AF,$DA,$E6,$FF
	.dc.b	$73,$F2,$7F,$7F,$F4,$BF,$D3,$D5,$FC,$47,$0F,$AF,$87,$F8,$0F,$9B
	.dc.b	$F0,$18,$FF,$82,$3C,$1C,$B6,$BF,$D5,$16,$EA,$FC,$F6,$BF,$CB,$63
	.dc.b	$15,$F1,$B7,$FE,$37,$3F,$CE,$F9,$FF,$C0,$D3,$FC,$4A,$3E,$3B,$FC
	.dc.b	$78,$C3,$FC,$26,$BF,$D5,$EF,$E2,$63,$F8,$FF,$7F,$FA,$2F,$B7,$8D
	.dc.b	$7E,$E2,$B8,$77,$FF,$00,$6D,$FF,$AB,$5F,$F5,$54,$7F,$06,$9F,$E1
	.dc.b	$93,$FE,$0A,$C7,$F9,$36,$1F,$C3,$2F,$EF,$57,$F1,$4F,$1F,$DD,$E0
	.dc.b	$FF,$AB,$EF,$F9,$65,$7B,$70,$FF,$01,$1F,$E5,$8C,$BF,$D6,$AF,$F8
	.dc.b	$1D,$F6,$79,$3F,$93,$8F,$E5,$A7,$F2,$5E,$7F,$F0,$1F,$C2,$7F,$7B
	.dc.b	$D3,$FC,$00,$D9,$79,$DF,$E0,$8F,$D5,$FF,$D5,$95,$0F,$F0,$33,$89
	.dc.b	$FF,$36,$C7,$F9,$D3,$9D,$39,$9F,$F0,$D4,$3D,$DF,$F5,$B6,$9F,$EF
	.dc.b	$FF,$37,$FC,$17,$18,$81,$F1,$EB,$FA,$04,$77,$9E,$6D,$8F,$F1,$2D
	.dc.b	$75,$7F,$C5,$18,$7F,$86,$FD,$9E,$9C,$3F,$D4,$E7,$FC,$47,$1F,$EB
	.dc.b	$0C,$7F,$A0,$FE,$49,$FE,$53,$E7,$7E,$AC,$EC,$FE,$DB,$EA,$FF,$8A
	.dc.b	$AB,$53,$9E,$E7,$A3,$9C,$C7,$E9,$61,$F0,$6F,$D0,$3F,$98,$7A,$3B
	.dc.b	$AD,$4E,$74,$7A,$19,$7F,$9A,$E5,$FC,$CF,$4C,$47,$F3,$98,$FE,$6D
	.dc.b	$9F,$3D,$7F,$99,$78,$3D,$B2,$FE,$D1,$B7,$FB,$03,$6F,$E7,$93,$FE
	.dc.b	$2B,$9F,$ED,$9F,$98,$FF,$9F,$E7,$FC,$17,$AF,$8F,$6F,$F2,$4C,$E6
	.dc.b	$7D,$7B,$F1,$E4,$7E,$63,$FE,$F1,$DF,$CF,$5F,$23,$2F,$EF,$13,$99
	.dc.b	$F4,$4F,$F8,$63,$6F,$F2,$AF,$D9,$E6,$7E,$8C,$F1,$C1,$39,$E0,$F7
	.dc.b	$E3,$FE,$CF,$FC,$E3,$F1,$CF,$FA,$EB,$CE,$46,$3C,$0F,$F0,$4C,$3D
	.dc.b	$0F,$C1,$F6,$7F,$84,$6F,$FD,$B1,$1F,$83,$E8,$78,$5F,$6A,$DF,$77
	.dc.b	$DB,$B7,$FC,$31,$18,$C3,$F6,$17,$C1,$FF,$66,$FE,$F3,$B7,$F0,$66
	.dc.b	$0D,$E1,$F8,$6F,$41,$3E,$7A,$FE,$81,$F2,$0F,$C6,$B9,$61,$F8,$C8
	.dc.b	$C3,$FE,$37,$3F,$1F,$FA,$3E,$1F,$F3,$D9,$E0,$71,$9F,$FB,$67,$DC
	.dc.b	$EF,$3E,$23,$FD,$B7,$8F,$87,$CD,$FF,$2E,$3E,$08,$FF,$6B,$6D,$7D
	.dc.b	$2F,$81,$1F,$C4,$74,$F8,$11,$F3,$63,$FD,$4F,$E2,$FE,$BD,$FF,$6F
	.dc.b	$70,$FC,$DE,$0F,$EC,$7C,$9C,$CF,$80,$FE,$A3,$CD,$FB,$F5,$FD,$4B
	.dc.b	$AB,$F0,$7D,$4F,$13,$5D,$5F,$83,$8F,$C5,$57,$F5,$FE,$DC,$46,$E3
	.dc.b	$07,$3E,$BF,$C4,$E2,$EF,$FC,$E7,$83,$F9,$FA,$FE,$D1,$18,$F0,$3F
	.dc.b	$4D,$4F,$0F,$F5,$58,$FF,$9C,$3C,$3F,$D7,$4E,$70,$FF,$72,$C7,$0A
	.dc.b	$F1,$EB,$ED,$7C,$BE,$15,$AF,$E0,$FE,$87,$F8,$BF,$4F,$16,$3C,$E7
	.dc.b	$F8,$2E,$7C,$DF,$E1,$B9,$7F,$04,$F1,$7F,$5D,$5B,$8C,$69,$FC,$52
	.dc.b	$BF,$C0,$5E,$3F,$E8,$1F,$F3,$85,$23,$EA,$47,$F8,$1F,$C8,$FC,$04
	.dc.b	$FF,$2B,$8F,$F6,$D7,$89,$FE,$D3,$8C,$77,$FF,$1E,$8F,$E3,$EF,$F2
	.dc.b	$0F,$2B,$ED,$5F,$F3,$BE,$CF,$C1,$5F,$37,$FD,$98,$7F,$9F,$1F,$E8
	.dc.b	$F8,$7F,$38,$3F,$4E,$E1,$FF,$45,$DF,$FA,$31,$FF,$0E,$79,$5F,$D3
	.dc.b	$63,$FD,$CB,$2F,$2F,$FA,$D5,$7E,$2D,$FE,$B9,$8F,$E3,$7C,$7F,$F6
	.dc.b	$D4,$FC,$26,$BF,$D8,$0F,$F4,$76,$1F,$ED,$8C,$BD,$51,$FD,$58,$F7
	.dc.b	$9F,$9D,$3F,$D2,$9A,$7F,$4E,$7F,$DC,$BC,$9F,$ED,$AF,$27,$EA,$75
	.dc.b	$FF,$97,$7C,$BF,$E8,$1C,$DF,$E6,$7E,$2F,$E3,$D7,$FC,$B5,$FE,$3B
	.dc.b	$7F,$81,$BF,$F2,$FF,$97,$FD,$01,$FF,$38,$E1,$FC,$FE,$B9,$74,$D7
	.dc.b	$FC,$E5,$D3,$FC,$E3,$3F,$C6,$C6,$9F,$E7,$E7,$FB,$94,$FF,$56,$76
	.dc.b	$FF,$B4,$71,$FE,$67,$1F,$B1,$C7,$FD,$0B,$8F,$F2,$3E,$FF,$F5,$FC
	.dc.b	$FF,$28,$3F,$AF,$F3,$7F,$BA,$FB,$FF,$94,$78,$1F,$0E,$7F,$93,$B5
	.dc.b	$F9,$DC,$9F,$C2,$BA,$3F,$99,$47,$F2,$68,$7E,$43,$FE,$F0,$F1,$7B
	.dc.b	$F9,$6F,$CB,$3F,$DE,$67,$FA,$CB,$CB,$7F,$00,$7E,$3F,$5F,$E7,$9E
	.dc.b	$66,$33,$F2,$C7,$F8,$30,$FE,$81,$E6,$73,$8F,$F3,$9B,$5F,$F6,$56
	.dc.b	$FF,$C8,$9F,$E5,$C7,$F2,$22,$F9,$5E,$67,$AD,$75,$3F,$91,$F0,$7B
	.dc.b	$E3,$FD,$02,$FC,$1E,$FF,$E4,$31,$CA,$3D,$13,$FA,$3E,$6F,$A5,$3F
	.dc.b	$E7,$97,$FB,$27,$2F,$D2,$8F,$E2,$A7,$DC,$E7,$FF,$58,$74,$FF,$29
	.dc.b	$F1,$FF,$15,$1F,$C9,$3C,$2F,$E1,$5E,$2F,$E3,$E3,$F9,$24,$61,$FC
	.dc.b	$97,$67,$F1,$3C,$3F,$D5,$B8,$BB,$30,$FF,$AC,$3B,$35,$DF,$CF,$4B
	.dc.b	$8F,$2C,$3F,$D2,$1D,$FF,$E2,$BC,$3F,$D9,$B8,$F5,$F1,$79,$57,$F1
	.dc.b	$43,$FA,$39,$FD,$CF,$4F,$E1,$AF,$EB,$5F,$E1,$1C,$1F,$E4,$48,$D7
	.dc.b	$87,$FC,$05,$FE,$DA,$3C,$17,$CA,$FF,$BD,$23,$31,$87,$EB,$FB,$5F
	.dc.b	$DD,$39,$BF,$CB,$D1,$FB,$C1,$E8,$7F,$0F,$AF,$E2,$5E,$1F,$FA,$0E
	.dc.b	$3F,$CA,$BD,$5F,$E3,$E9,$F2,$E7,$E5,$BF,$EF,$F1,$E8,$F3,$71,$1B
	.dc.b	$F1,$FF,$CA,$75,$E1,$8F,$F2,$E9,$FE,$FE,$C7,$CF,$3F,$E6,$99,$F1
	.dc.b	$9F,$E0,$DB,$FE,$71,$5F,$90,$8F,$F0,$D4,$62,$3F,$D8,$C7,$F9,$84
	.dc.b	$FD,$0E,$7E,$5C,$FF,$CA,$1B,$FF,$C0,$0F,$97,$FF,$63,$57,$A3,$8F
	.dc.b	$FC,$C5,$18,$AD,$F1,$84,$3B,$7F,$CD,$37,$F9,$3F,$8C,$F9,$18,$7F
	.dc.b	$91,$DF,$F3,$AB,$78,$77,$3F,$EF,$F3,$EA,$C7,$E7,$9F,$F4,$53,$FE
	.dc.b	$43,$BF,$F5,$CE,$9F,$97,$E6,$61,$E8,$BF,$E4,$0F,$9D,$FF,$2F,$9F
	.dc.b	$E7,$2E,$B4,$C9,$F4,$6F,$F2,$FB,$3C,$DC,$5F,$86,$6E,$1F,$D0,$F1
	.dc.b	$FF,$03,$3E,$DD,$3F,$02,$FF,$97,$0F,$E4,$C7,$E3,$BC,$FF,$53,$C5
	.dc.b	$F5,$61,$C2,$BF,$C3,$B3,$F0,$4F,$C3,$F7,$70,$E1,$FE,$63,$1F,$3F
	.dc.b	$87,$E8,$6F,$FE,$DA,$F2,$3F,$C5,$D3,$FC,$3F,$1F,$F2,$D4,$61,$FE
	.dc.b	$3D,$CD,$BF,$03,$FD,$D3,$5F,$CC,$8F,$C1,$CF,$E4,$6B,$F8,$D4,$73
	.dc.b	$3F,$CB,$BE,$47,$FB,$45,$78,$FF,$C9,$6B,$F8,$F4,$7E,$C2,$31,$1F
	.dc.b	$EB,$8C,$3F,$DD,$F1,$88,$47,$F3,$1F,$47,$FC,$B5,$E3,$7F,$81,$63
	.dc.b	$F9,$BB,$6F,$F7,$7E,$2F,$19,$E1,$08,$C5,$FD,$27,$B3,$FE,$7B,$E9
	.dc.b	$F5,$BE,$69,$FF,$38,$C6,$23,$CC,$7D,$A7,$E4,$2E,$BF,$DB,$CF,$F9
	.dc.b	$60,$F8,$E7,$F8,$74,$C1,$F1,$CE,$BA,$FC,$7F,$03,$EE,$F0,$7D,$93
	.dc.b	$F8,$E7,$17,$E0,$B0,$FF,$43,$4E,$A7,$EF,$5F,$58,$FF,$55,$BF,$8C
	.dc.b	$3F,$C4,$7C,$9C,$7B,$BF,$CC,$CF,$E1,$E3,$FA,$19,$FE,$CF,$E9,$FF
	.dc.b	$13,$57,$E6,$23,$1C,$FF,$E2,$FE,$8F,$CE,$F3,$FF,$3F,$EB,$FD,$2F
	.dc.b	$5F,$EB,$30,$FF,$22,$F6,$7F,$97,$B2,$FD,$3E,$9B,$AF,$D5,$57,$9E
	.dc.b	$BF,$3F,$9F,$F4,$6D,$7F,$2F,$B7,$F8,$F6,$7C,$D5,$FE,$5B,$F0,$FB
	.dc.b	$E5,$E6,$9F,$E0,$95,$F1,$1F,$2E,$1F,$D2,$8F,$F2,$FE,$1F,$A2,$C3
	.dc.b	$CB,$18,$8F,$E2,$7C,$7D,$E3,$1D,$7F,$C0,$EB,$F9,$26,$03,$EF,$1F
	.dc.b	$D3,$FC,$BF,$D2,$D7,$F9,$06,$77,$3F,$E0,$78,$FF,$33,$F8,$1F,$CC
	.dc.b	$FB,$EC,$7F,$47,$9F,$E6,$BB,$FF,$98,$23,$FD,$43,$1F,$D6,$EB,$F9
	.dc.b	$FC,$7D,$D7,$FC,$2C,$FF,$9C,$BC,$2F,$F6,$2D,$FF,$B1,$B9,$3E,$CC
	.dc.b	$62,$47,$F5,$48,$FF,$34,$78,$5F,$D7,$63,$FC,$CE,$DE,$4F,$F3,$04
	.dc.b	$7F,$85,$3A,$FF,$B6,$79,$9F,$AD,$8F,$E3,$95,$FE,$60,$7F,$AA,$1F
	.dc.b	$D8,$7E,$AF,$FC,$BC,$B2,$FF,$6C,$C7,$F5,$7F,$0F,$FB,$AC,$FF,$AF
	.dc.b	$5F,$8F,$AB,$E9,$FF,$A9,$F0,$FF,$A0,$7C,$FF,$E7,$2F,$AD,$E0,$D7
	.dc.b	$A3,$E3,$BF,$EF,$5E,$EF,$FB,$A2,$FF,$86,$5F,$F5,$E8,$FF,$7C,$C7
	.dc.b	$F2,$BE,$CF,$F7,$6F,$8F,$FC,$9F,$F5,$4F,$EF,$11,$8E,$9F,$EB,$11
	.dc.b	$EB,$C7,$F9,$E4,$E6,$53,$F6,$A7,$FC,$CA,$3B,$3F,$4B,$F3,$9F,$DB
	.dc.b	$AF,$FD,$F7,$18,$9F,$E0,$9B,$7E,$BE,$3F,$D7,$59,$79,$67,$FD,$B5
	.dc.b	$87,$16,$9C,$2F,$EE,$BF,$C1,$FD,$5F,$EB,$F1,$FE,$CE,$F5,$7D,$73
	.dc.b	$FD,$6F,$C0,$FE,$B7,$1E,$BE,$7F,$F3,$F4,$BC,$1F,$F6,$57,$4F,$F4
	.dc.b	$FF,$0B,$CF,$87,$F6,$2D,$FF,$C0,$F1,$88,$47,$DD,$8F,$F7,$1E,$BF
	.dc.b	$D5,$FC,$3F,$2F,$0F,$CD,$CB,$FC,$57,$18,$8F,$9B,$1F,$0A,$31,$0D
	.dc.b	$BF,$BC,$FE,$DF,$8E,$0F,$F0,$BB,$6F,$E4,$BC,$27,$FA,$C4,$7F,$8C
	.dc.b	$F9,$FF,$AE,$F8,$BF,$A3,$8F,$F3,$DF,$81,$FA,$4F,$31,$F2,$BF,$87
	.dc.b	$D7,$FA,$7F,$57,$F6,$48,$FC,$3C,$2B,$87,$9B,$F8,$FB,$FE,$C3,$DD
	.dc.b	$FE,$4A,$AF,$EF,$BE,$6F,$FA,$0E,$3E,$AE,$9F,$E4,$EC,$FF,$B1,$FA
	.dc.b	$9E,$37,$FA,$FB,$6F,$F8,$77,$9D,$FD,$7A,$7D,$47,$F9,$5B,$D4,$FF
	.dc.b	$4A,$3F,$ED,$B3,$FC,$B3,$EA,$63,$1E,$F1,$8F,$43,$FD,$5F,$E8,$78
	.dc.b	$E3,$DD,$BF,$F0,$48,$C4,$7F,$A7,$B1,$C6,$7F,$AB,$F3,$7D,$B3,$FE
	.dc.b	$F7,$3F,$CC,$EE,$1F,$96,$7F,$8E,$F8,$7E,$57,$93,$FB,$B8,$DC,$7C
	.dc.b	$AC,$7F,$D3,$71,$EC,$3F,$B0,$9F,$E6,$93,$FD,$5C,$FF,$18,$F3,$1C
	.dc.b	$FF,$7E,$7F,$DF,$AF,$DF,$E0,$FF,$15,$CE,$A7,$F4,$A7,$F3,$3E,$AE
	.dc.b	$0F,$37,$84,$FF,$6A,$7A,$78,$1F,$D9,$27,$FD,$49,$DB,$FE,$8B,$ED
	.dc.b	$FC,$DD,$FF,$70,$7F,$41,$1F,$E7,$1C,$3F,$B3,$78,$1F,$D8,$BC,$1F
	.dc.b	$ED,$3E,$17,$EC,$7C,$2F,$ED,$19,$7E,$C7,$C2,$57,$EC,$63,$FB,$8F
	.dc.b	$85,$FA,$57,$F7,$78,$FF,$71,$8F,$F1,$D7,$83,$F1,$27,$71,$FA,$87
	.dc.b	$FA,$86,$3E,$47,$FD,$37,$5F,$C2,$F0,$FD,$FE,$BE,$29,$FE,$29,$1F
	.dc.b	$BB,$7F,$C1,$15,$FC,$6F,$C2,$F4,$C7,$F6,$57,$F8,$E5,$FF,$80,$FC
	.dc.b	$5F,$E4,$DD,$7F,$E0,$FF,$1B,$EF,$F8,$7F,$EB,$17,$EF,$61,$FE,$B5
	.dc.b	$3F,$94,$0F,$03,$FB,$43,$FD,$52,$BF,$AA,$C6,$63,$D3,$E3,$FF,$3D
	.dc.b	$7F,$C3,$8F,$ED,$75,$FC,$64,$7C,$68,$FE,$8D,$1F,$82,$F3,$3F,$A1
	.dc.b	$67,$FD,$3B,$C9,$FE,$D5,$7F,$D5,$4F,$F1,$37,$95,$A3,$F8,$26,$9F
	.dc.b	$EC,$96,$C7,$F3,$4F,$17,$FB,$03,$FD,$78,$F5,$E9,$FD,$59,$FF,$7C
	.dc.b	$94,$FE,$8D,$A7,$EF,$DF,$EA,$67,$FB,$58,$FF,$43,$1F,$E1,$1E,$2F
	.dc.b	$F6,$EB,$FD,$23,$7F,$F6,$EF,$95,$F8,$CE,$3F,$F1,$69,$FE,$92,$F4
	.dc.b	$BF,$9B,$F4,$7F,$BC,$FD,$4F,$F6,$57,$5F,$FB,$E2,$95,$8C,$7F,$DF
	.dc.b	$D1,$FE,$36,$AF,$F6,$87,$77,$F2,$EC,$FC,$39,$7E,$A6,$FF,$DF,$3E
	.dc.b	$A7,$F2,$5F,$33,$F8,$B4,$7F,$B9,$BC,$BF,$E4,$EF,$FB,$27,$9F,$FD
	.dc.b	$37,$18,$8F,$F4,$49,$FC,$A4,$F8,$03,$C0,$FE,$52,$FE,$FC,$E9,$F7
	.dc.b	$5F,$76,$BF,$91,$C6,$4F,$EB,$BC,$8F,$4B,$53,$FC,$B9,$AB,$C9,$FF
	.dc.b	$0F,$F9,$AF,$D4,$FE,$CE,$0F,$3B,$0F,$F8,$74,$FF,$43,$E1,$CA,$FF
	.dc.b	$92,$3D,$3C,$8F,$E4,$73,$E3,$3F,$C5,$7E,$5F,$F2,$17,$FD,$11,$A6
	.dc.b	$DF,$E4,$47,$C0,$37,$E5,$7F,$24,$7F,$DF,$13,$FE,$B2,$EC,$FC,$34
	.dc.b	$68,$F9,$9A,$FF,$9F,$B9,$7E,$B7,$2F,$FB,$A7,$D6,$FE,$49,$18,$87
	.dc.b	$2F,$F1,$9B,$FD,$EF,$47,$E2,$3A,$71,$E8,$FE,$1E,$7F,$CF,$FD,$9F
	.dc.b	$8C,$AF,$C6,$46,$1C,$75,$FF,$31,$F1,$FF,$C3,$D9,$7E,$5B,$C3,$FE
	.dc.b	$04,$FF,$A5,$4F,$0D,$A1,$E3,$7B,$E3,$F7,$67,$EF,$CF,$DF,$18,$3F
	.dc.b	$D5,$C7,$F2,$0F,$17,$FB,$B7,$0F,$F6,$7A,$FF,$64,$CF,$D7,$3F,$C3
	.dc.b	$FD,$FF,$E4,$AF,$2B,$FC,$9F,$5F,$E2,$2F,$A0,$FF,$23,$D6,$CF,$F3
	.dc.b	$0D,$7F,$9D,$3C,$EF,$F6,$5C,$7F,$8A,$04,$66,$3F,$C9,$71,$FE,$75
	.dc.b	$35,$0F,$27,$FC,$15,$3F,$CC,$E3,$DF,$3D,$CF,$DF,$9F,$E8,$7E,$6F
	.dc.b	$F6,$24,$FF,$94,$63,$FC,$E7,$1F,$E8,$BF,$5F,$C7,$1F,$E4,$C3,$FD
	.dc.b	$29,$CF,$FC,$DB,$97,$FC,$6B,$E3,$7F,$BE,$E3,$F9,$CF,$1F,$F8,$E3
	.dc.b	$9B,$FD,$63,$F2,$7F,$BA,$E5,$FF,$1A,$3F,$F0,$07,$A3,$FC,$7B,$CC
	.dc.b	$27,$AF,$FE,$B8,$CF,$31,$8E,$80,$CB,$F8,$7B,$8F,$61,$ED,$FE,$4F
	.dc.b	$7F,$F2,$FF,$5F,$FA,$D2,$3B,$67,$FC,$46,$7F,$27,$A7,$F0,$EF,$2B
	.dc.b	$CB,$9E,$A7,$E8,$FA,$7F,$E4,$0D,$BC,$91,$83,$FD,$33,$3E,$91,$1F
	.dc.b	$C6,$E3,$EC,$F7,$71,$D7,$F3,$7C,$FF,$C6,$EF,$85,$EF,$CA,$FE,$2D
	.dc.b	$CF,$FC,$5B,$1F,$F3,$4C,$62,$3E,$A1,$FC,$73,$C1,$FF,$22,$74,$7D
	.dc.b	$6F,$2B,$FC,$91,$18,$EA,$FF,$1E,$5B,$C3,$FE,$1E,$DC,$7E,$FE,$3D
	.dc.b	$0F,$F2,$52,$F2,$BF,$8E,$3F,$89,$6B,$F8,$98,$FF,$67,$9C,$23,$2F
	.dc.b	$E3,$72,$FC,$77,$8F,$FC,$83,$C9,$FE,$CD,$F5,$FF,$E4,$FF,$2B,$F8
	.dc.b	$FE,$3C,$E3,$FC,$A5,$EB,$72,$C7,$C9,$87,$F2,$0A,$F3,$C7,$83,$CA
	.dc.b	$C4,$78,$F6,$FE,$3E,$FE,$7F,$CA,$F7,$3E,$7D,$3F,$C2,$D1,$F2,$DF
	.dc.b	$CE,$C7,$F1,$D3,$D1,$3F,$C8,$B8,$3F,$9A,$D7,$D1,$C3,$C1,$5F,$C2
	.dc.b	$F9,$BE,$E7,$37,$F5,$C8,$C1,$F8,$4E,$5F,$CC,$47,$E5,$E3,$F8,$97
	.dc.b	$3F,$0E,$FF,$F2,$04,$6E,$BF,$13,$5F,$9E,$EF,$FE,$4A,$FE,$83,$AF
	.dc.b	$F8,$F7,$5F,$E5,$AB,$F8,$EF,$91,$FD,$CE,$9E,$07,$E8,$A5,$E1,$FF
	.dc.b	$27,$CB,$FC,$7D,$39,$EF,$C4,$7E,$66,$BC,$37,$EB,$C3,$FC,$7D,$1E
	.dc.b	$AC,$F7,$E1,$FE,$F2,$3F,$5D,$3F,$E3,$BA,$69,$FB,$DC,$3F,$98,$78
	.dc.b	$7F,$ED,$3D,$3F,$84,$46,$0F,$E1,$78,$FB,$DF,$E6,$DE,$57,$E6,$5F
	.dc.b	$F2,$3B,$5F,$EE,$C6,$8F,$E3,$98,$FF,$21,$7F,$92,$1F,$E3,$13,$F9
	.dc.b	$E0,$E5,$FF,$2B,$47,$F2,$B8,$FE,$7D,$C6,$35,$3F,$E6,$9F,$1B,$FB
	.dc.b	$AD,$7F,$3B,$AF,$F2,$CF,$95,$FE,$B7,$F2,$39,$F8,$1F,$9C,$9F,$F2
	.dc.b	$EB,$5F,$3F,$8A,$FC,$13,$FA,$4C,$7F,$38,$FC,$13,$FC,$0A,$FF,$42
	.dc.b	$7F,$DC,$FE,$5F,$F9,$40,$FF,$4E,$47,$F4,$28,$E2,$FF,$4E,$FC,$13
	.dc.b	$FD,$37,$B7,$A7,$6F,$EA,$3F,$9C,$7F,$BC,$3C,$DF,$E5,$9E,$BF,$FB
	.dc.b	$A7,$9F,$F8,$8C,$E7,$A3,$FD,$B5,$E7,$7F,$17,$DD,$3F,$EE,$CF,$63
	.dc.b	$F1,$55,$F8,$AD,$7F,$A9,$CE,$7C,$FF,$BD,$7C,$27,$F4,$19,$7F,$7D
	.dc.b	$76,$63,$2E,$D6,$8F,$F4,$86,$7F,$E3,$E8,$C4,$79,$E5,$3D,$BF,$38
	.dc.b	$F0,$65,$FC,$F7,$14,$62,$3F,$BF,$C3,$F3,$8F,$F7,$D6,$5E,$38,$C6
	.dc.b	$3E,$3C,$7F,$21,$0F,$DD,$37,$1F,$CF,$A3,$13,$B8,$FD,$D7,$9E,$FC
	.dc.b	$C3,$E9,$CF,$FB,$DD,$FF,$7D,$7A,$9F,$EA,$39,$7C,$67,$D6,$E4,$FE
	.dc.b	$D7,$18,$BF,$E6,$3C,$9F,$DC,$CF,$CE,$C3,$F1,$7E,$B2,$8C,$3F,$9C
	.dc.b	$C3,$FD,$61,$D4,$FD,$E7,$6D,$9D,$63,$FF,$00,$27,$CA,$FF,$BE,$A1
	.dc.b	$EC,$BF,$16,$EE,$D8,$FF,$26,$1F,$E5,$58,$F1,$FA,$B9,$8F,$F7,$F1
	.dc.b	$F6,$3D,$4F,$03,$FE,$AC,$CB,$FD,$89,$18,$EE,$FB,$B1,$FD,$12,$53
	.dc.b	$97,$FC,$E1,$E8,$7B,$9F,$F4,$D9,$FE,$FC,$DB,$D3,$9F,$F5,$69,$F9
	.dc.b	$53,$F2,$63,$F4,$4F,$F9,$8B,$83,$E8,$70,$7F,$08,$AF,$B1,$3F,$C1
	.dc.b	$32,$FF,$17,$28,$CB,$E9,$BF,$F7,$CF,$27,$E1,$38,$FF,$CA,$7E,$1F
	.dc.b	$E1,$2B,$FD,$C3,$1F,$DF,$E3,$F1,$B1,$88,$DF,$3F,$17,$FA,$14,$FE
	.dc.b	$62,$7F,$45,$E3,$7F,$87,$3B,$BF,$5F,$E3,$FE,$96,$B5,$5F,$EE,$49
	.dc.b	$FD,$9F,$97,$9E,$FF,$D3,$F8,$3F,$E2,$4C,$BF,$4F,$DD,$FC,$D2,$73
	.dc.b	$1F,$E2,$8F,$03,$F9,$A6,$5F,$B7,$CB,$CD,$3F,$CF,$EB,$CD,$E5,$F5
	.dc.b	$9F,$2C,$B3,$F2,$C7,$F9,$AE,$BF,$61,$9E,$E3,$C3,$E0,$7E,$E7,$CE
	.dc.b	$C6,$7F,$C4,$B5,$F8,$DE,$27,$D4,$BF,$77,$8B,$F9,$ED,$8F,$F2,$28
	.dc.b	$FF,$7A,$73,$7F,$22,$9C,$C7,$FA,$48,$F8,$DE,$37,$13,$87,$B9,$FE
	.dc.b	$69,$AF,$E5,$93,$FE,$E2,$3F,$98,$1F,$E7,$F8,$FF,$71,$78,$AC,$7F
	.dc.b	$D0,$93,$FC,$A8,$FF,$71,$71,$FF,$0F,$E0,$FF,$47,$BF,$E4,$5A,$F9
	.dc.b	$5C,$BF,$E9,$7E,$3F,$F0,$14,$61,$FF,$4B,$DF,$F2,$23,$FD,$29,$5C
	.dc.b	$DF,$F0,$03,$FC,$AA,$FE,$BC,$FF,$17,$6B,$FC,$D7,$C6,$FF,$75,$70
	.dc.b	$FF,$A2,$1F,$F7,$67,$2F,$FA,$2B,$A3,$FD,$61,$CF,$72,$EB,$97,$F9
	.dc.b	$6F,$C5,$FC,$EF,$07,$FA,$F2,$7F,$86,$76,$FF,$B1,$67,$F5,$7D,$7F
	.dc.b	$AA,$9F,$F7,$64,$FC,$4F,$67,$F9,$C7,$8D,$FB,$79,$FD,$49,$FE,$47
	.dc.b	$F9,$9F,$F6,$74,$1F,$F6,$C4,$7F,$0E,$FD,$8B,$F9,$EE,$1F,$E9,$18
	.dc.b	$FF,$21,$46,$23,$FB,$94,$7F,$90,$A7,$8F,$87,$CE,$3F,$DA,$3E,$BF
	.dc.b	$F4,$98,$F1,$47,$EE,$BC,$2F,$4D,$FF,$8F,$23,$F8,$11,$FE,$FE,$CB
	.dc.b	$FD,$CB,$E5,$7E,$36,$3D,$35,$C2,$BF,$CB,$98,$7E,$C7,$87,$E5,$D7
	.dc.b	$FB,$25,$FF,$6E,$78,$3F,$B5,$E4,$FA,$73,$FE,$CD,$7E,$9B,$FA,$53
	.dc.b	$FA,$F7,$7F,$E1,$79,$BF,$4C,$FF,$B7,$FB,$BF,$65,$CB,$F8,$4E,$6F
	.dc.b	$D5,$3F,$86,$C7,$F5,$9D,$7C,$71,$E8,$7F,$BA,$BA,$7F,$88,$75,$7E
	.dc.b	$52,$FF,$BB,$63,$FB,$DF,$03,$95,$FF,$B8,$7B,$7F,$3A,$FE,$C0,$FD
	.dc.b	$11,$FE,$DB,$F0,$3E,$97,$17,$E9,$1F,$E5,$F1,$FE,$D6,$9C,$BF,$CA
	.dc.b	$3B,$BF,$50,$FF,$A2,$BD,$7F,$EF,$3F,$01,$FA,$98,$FE,$27,$EE,$FF
	.dc.b	$BC,$5F,$F6,$BE,$BF,$E1,$18,$C4,$7F,$89,$A3,$1C,$FF,$ED,$97,$FC
	.dc.b	$07,$C1,$FE,$BD,$F0,$3C,$11,$8E,$1F,$F2,$FC,$7F,$82,$3A,$BF,$DD
	.dc.b	$73,$9F,$55,$E5,$73,$CB,$FC,$DC,$72,$29,$FE,$13,$AE,$2F,$F9,$E7
	.dc.b	$3F,$C1,$BF,$B9,$E7,$FF,$67,$72,$71,$8F,$F1,$66,$1E,$AB,$E3,$18
	.dc.b	$D7,$FD,$E5,$D3,$E3,$D3,$E4,$36,$7C,$07,$FB,$F9,$F4,$67,$FE,$F9
	.dc.b	$F2,$3F,$C4,$7C,$FF,$EC,$DA,$FA,$11,$88,$FF,$74,$47,$C1,$1C,$7F
	.dc.b	$D4,$A3,$FD,$D7,$CD,$FE,$FF,$EC,$FE,$09,$5F,$E7,$F3,$ED,$76,$73
	.dc.b	$AF,$F1,$8C,$7F,$96,$23,$B7,$0F,$C2,$F0,$FF,$4D,$3F,$EC,$CF,$0F
	.dc.b	$FC,$39,$8F,$FB,$22,$7F,$63,$D1,$F8,$2A,$FF,$1F,$46,$1F,$F2,$07
	.dc.b	$47,$F9,$EB,$C3,$FD,$E4,$7F,$B2,$CF,$EE,$6E,$2F,$FC,$B8,$7F,$87
	.dc.b	$BC,$2E,$FE,$D7,$F9,$27,$CE,$78,$BF,$9C,$C1,$E6,$E1,$FF,$16,$9F
	.dc.b	$C2,$CF,$F1,$7E,$1F,$E1,$1F,$1B,$B3,$FE,$85,$3F,$C2,$F8,$FE,$7A
	.dc.b	$7E,$1F,$47,$A5,$69,$FE,$56,$8C,$3F,$E0,$DD,$F8,$4E,$5F,$8B,$C7
	.dc.b	$FC,$2E,$3F,$1E,$7F,$97,$BA,$FF,$C1,$5C,$9F,$5B,$2F,$F0,$7C,$EB
	.dc.b	$AB,$FC,$05,$CF,$F5,$63,$FC,$2F,$7F,$76,$DC,$BF,$96,$8C,$47,$DA
	.dc.b	$BF,$E2,$D5,$FD,$EB,$AD,$D7,$CB,$AF,$F1,$F9,$FF,$A8,$74,$DD,$BB
	.dc.b	$FF,$60,$FE,$5B,$3F,$F0,$47,$6F,$E8,$A3,$F0,$F1,$8E,$FF,$F6,$8F
	.dc.b	$17,$FA,$BB,$BF,$FC,$23,$AF,$E9,$6F,$F8,$CE,$FF,$9B,$78,$CF,$E3
	.dc.b	$B4,$E1,$1F,$04,$FE,$3F,$9F,$89,$F9,$13,$E8,$96,$9F,$C9,$B2,$FE
	.dc.b	$07,$5E,$7A,$FE,$1D,$9F,$C8,$7F,$86,$DF,$BE,$7F,$95,$F8,$4C,$BE
	.dc.b	$5D,$FE,$AC,$FE,$3B,$3F,$EC,$5A,$76,$FE,$76,$3F,$96,$1F,$E0,$D3
	.dc.b	$EE,$78,$5F,$CA,$15,$7F,$86,$1F,$F3,$F1,$FE,$BC,$F0,$3A,$F1,$7F
	.dc.b	$AF,$A3,$FD,$43,$0F,$C3,$7F,$99,$E3,$FE,$D5,$F7,$33,$C9,$FE,$B9
	.dc.b	$F8,$3F,$B7,$87,$EE,$30,$FF,$23,$ED,$EB,$9C,$CF,$FB,$4A,$B8,$78
	.dc.b	$9F,$C5,$23,$FC,$7B,$F0,$9F,$E2,$29,$FE,$58,$FE,$26,$3F,$8A,$1F
	.dc.b	$E4,$2F,$3F,$F9,$1E,$3F,$E5,$4E,$1E,$D3,$FC,$70,$FF,$5E,$38,$7F
	.dc.b	$8E,$C6,$E1,$E7,$F2,$C3,$FC,$AB,$7E,$6B,$FF,$0E,$1F,$DA,$39,$3F
	.dc.b	$CA,$1C,$1F,$C7,$71,$FF,$31,$F4,$FF,$99,$F4,$FE,$4D,$7E,$A3,$F9
	.dc.b	$29,$4F,$0F,$67,$FB,$6F,$87,$F9,$14,$7F,$26,$F1,$FF,$95,$70,$7F
	.dc.b	$90,$9B,$1D,$DF,$F6,$4C,$F0,$C5,$FB,$B6,$AB,$E8,$35,$FF,$7F,$BC
	.dc.b	$27,$F1,$DE,$47,$FA,$5B,$4F,$BD,$5F,$C9,$3C,$0F,$F0,$95,$7F,$B1
	.dc.b	$FF,$1E,$FB,$EB,$9B,$FD,$C7,$82,$73,$1F,$CA,$7C,$3F,$F2,$FE,$1F
	.dc.b	$CA,$1E,$77,$F9,$DF,$27,$F9,$6E,$5F,$DF,$3C,$9F,$E5,$1E,$47,$0C
	.dc.b	$3F,$9F,$F9,$3F,$C9,$DF,$F5,$D9,$FE,$F5,$F2,$71,$7E,$88,$FE,$5B
	.dc.b	$5F,$E2,$97,$F9,$61,$F8,$EC,$7F,$94,$BF,$EF,$D9,$F0,$3F,$EB,$EC
	.dc.b	$BC,$0F,$F2,$9D,$3D,$68,$FF,$59,$3E,$EF,$1B,$FC,$73,$1F,$9E,$F4
	.dc.b	$3F,$C6,$FC,$1E,$EC,$7F,$80,$CF,$EB,$B8,$7E,$93,$FE,$A9,$7E,$94
	.dc.b	$6B,$C0,$FB,$33,$F9,$6E,$D7,$37,$E7,$E7,$F9,$9C,$62,$7F,$97,$F3
	.dc.b	$72,$F1,$3F,$9D,$47,$F0,$79,$D3,$FC,$97,$A3,$F5,$2F,$F3,$67,$F2
	.dc.b	$18,$FF,$0F,$C3,$FD,$8B,$E2,$BB,$1E,$07,$87,$BF,$F4,$7C,$9F,$CD
	.dc.b	$72,$FD,$17,$7F,$F8,$6B,$2E,$55,$A9,$FE,$68,$FC,$07,$FC,$A4,$7B
	.dc.b	$7C,$0F,$E8,$91,$F9,$5E,$4F,$E7,$B9,$F8,$72,$FF,$3C,$E9,$FE,$CE
	.dc.b	$8F,$F6,$96,$3F,$A6,$8C,$69,$E6,$CB,$F9,$26,$3F,$C4,$74,$F3,$47
	.dc.b	$FB,$43,$87,$CD,$87,$CF,$AD,$4F,$F4,$9F,$0B,$EF,$BF,$CA,$0F,$B8
	.dc.b	$7F,$5C,$F2,$FF,$D8,$1E,$57,$F2,$DF,$07,$F9,$14,$FF,$7A,$3F,$8F
	.dc.b	$5F,$F9,$7F,$CD,$FE,$0F,$1F,$EE,$0E,$4F,$A8,$3C,$DC,$F8,$BF,$E6
	.dc.b	$9C,$3F,$84,$EF,$FE,$B4,$7E,$9B,$FE,$E9,$3E,$A7,$9D,$F9,$38,$CC
	.dc.b	$67,$D0,$C7,$5F,$FA,$12,$7F,$17,$97,$FA,$F3,$2F,$03,$FE,$AC,$3F
	.dc.b	$44,$7F,$A9,$87,$8D,$FB,$B7,$FD,$36,$63,$BF,$FC,$35,$19,$8F,$E8
	.dc.b	$E7,$F9,$1B,$C5,$FD,$DB,$CD,$FF,$7D,$71,$FF,$55,$3F,$8E,$B8,$70
	.dc.b	$3C,$5F,$F1,$47,$E1,$BF,$E2,$7F,$14,$FF,$8A,$7D,$0E,$15,$CE,$BF
	.dc.b	$D5,$DE,$6F,$F1,$48,$FD,$7C,$F8,$B1,$FF,$54,$E7,$FE,$31,$F3,$FD
	.dc.b	$15,$FE,$C9,$9F,$F1,$21,$F0,$8D,$9F,$03,$B3,$E4,$ED,$FE,$A1,$8F
	.dc.b	$DD,$1F,$ED,$87,$E5,$C7,$D2,$E2,$FD,$40,$E3,$FF,$12,$D7,$F0,$9E
	.dc.b	$DF,$F5,$67,$47,$FB,$A2,$BE,$DC,$FE,$90,$73,$69,$FF,$7E,$E3,$FE
	.dc.b	$A8,$AF,$F5,$57,$3F,$F1,$3C,$BF,$23,$5F,$48,$FE,$E2,$7F,$17,$F3
	.dc.b	$3F,$C1,$1D,$3F,$E2,$7C,$BB,$F3,$FF,$72,$C9,$39,$97,$4F,$E6,$35
	.dc.b	$FE,$05,$EE,$FE,$62,$B1,$48,$FC,$DE,$9F,$9E,$C3,$FC,$2D,$E8,$FE
	.dc.b	$5F,$83,$11,$FE,$42,$BF,$8B,$97,$E8,$17,$A2,$F6,$3F,$8C,$67,$98
	.dc.b	$C5,$FF,$88,$72,$FD,$07,$8F,$FE,$AD,$D3,$C9,$C1,$FE,$8C,$8C,$57
	.dc.b	$F2,$1A,$FF,$45,$BE,$4C,$BE,$AE,$98,$8F,$24,$7F,$1F,$D3,$F5,$38
	.dc.b	$7F,$AB,$63,$6F,$FA,$63,$BB,$FD,$1F,$1E,$A8,$FD,$21,$FE,$CD,$E5
	.dc.b	$FE,$CC,$FF,$47,$7F,$C4,$8F,$F4,$9F,$C0,$FF,$B3,$F4,$7F,$4B,$8F
	.dc.b	$F1,$C7,$A1,$67,$FD,$21,$D7,$FE,$87,$EF,$FD,$37,$A7,$FE,$BE,$6E
	.dc.b	$3F,$E3,$FE,$EF,$D3,$E3,$DE,$3F,$BA,$3F,$EA,$63,$FD,$1F,$DF,$DE
	.dc.b	$3C,$71,$FE,$AA,$EF,$E7,$03,$F6,$50,$8C,$C7,$EB,$31,$FD,$56,$9C
	.dc.b	$E3,$BD,$33,$FE,$D0,$F8,$F4,$F6,$79,$1F,$DA,$7C,$0F,$D8,$DF,$F9
	.dc.b	$58,$FE,$75,$C1,$F3,$38,$3F,$47,$DD,$FB,$0E,$3F,$AB,$1F,$94,$CB
	.dc.b	$FC,$AF,$E6,$6E,$3F,$5D,$18,$3E,$C3,$F6,$DF,$F2,$EF,$3F,$E1,$2B
	.dc.b	$F4,$71,$83,$F6,$93,$F9,$B8,$FC,$44,$7E,$CF,$C1,$FE,$C5,$D5,$F9
	.dc.b	$29,$FD,$94,$FF,$90,$63,$1D,$3F,$E9,$49,$FC,$75,$FA,$BC,$1F,$37
	.dc.b	$4F,$F5,$C8,$ED,$3E,$2D,$3F,$AC,$F5,$FE,$AB,$2F,$E0,$5E,$0F,$EE
	.dc.b	$3C,$0F,$C8,$D7,$ED,$FB,$BF,$6D,$97,$F0,$1F,$1B,$E9,$73,$7E,$27
	.dc.b	$1F,$E8,$87,$E3,$7E,$FD,$F8,$2B,$F0,$6F,$E5,$6D,$F9,$BD,$D7,$F8
	.dc.b	$76,$B7,$5A,$1F,$2D,$A9,$7E,$6A,$FB,$27,$E1,$4F,$88,$C2,$F9,$1E
	.dc.b	$15,$89,$CF,$B3,$CF,$E3,$7B,$78,$FE,$0C,$9E,$77,$FB,$FA,$73,$C1
	.dc.b	$89,$FE,$AB,$E1,$7B,$63,$D9,$7E,$8E,$1F,$F7,$DC,$E7,$83,$D5,$C1
	.dc.b	$FE,$FE,$F8,$1F,$D8,$6D,$E3,$E1,$F4,$F0,$67,$4F,$F1,$F4,$E7,$85
	.dc.b	$C1,$FE,$41,$F4,$7F,$D5,$B5,$AC,$3E,$0F,$17,$C0,$C7,$5E,$6F,$69
	.dc.b	$C9,$F1,$B0,$E5,$F2,$3F,$E0,$5C,$3F,$B7,$BF,$CF,$1A,$7F,$39,$EE
	.dc.b	$FE,$6B,$D9,$FC,$4B,$35,$7F,$09,$ED,$FC,$37,$7F,$E7,$E5,$F8,$2F
	.dc.b	$56,$E3,$FD,$2B,$C2,$FF,$2D,$79,$EF,$77,$C4,$D9,$D0,$70,$FF,$86
	.dc.b	$74,$FF,$26,$7C,$8F,$2A,$F1,$47,$F8,$70,$7B,$DF,$E6,$7F,$47,$8E
	.dc.b	$B9,$95,$E6,$F8,$C7,$F8,$92,$F1,$87,$F6,$0F,$12,$BE,$19,$FE,$84
	.dc.b	$BC,$DF,$6B,$98,$FE,$B6,$E7,$6F,$99,$3F,$22,$3E,$65,$7F,$5F,$F4
	.dc.b	$FF,$8A,$70,$FA,$A3,$F5,$B3,$D6,$3E,$B5,$71,$8E,$DB,$FF,$1B,$E5
	.dc.b	$FB,$DD,$1F,$A1,$E6,$FC,$5D,$79,$67,$3D,$1F,$A0,$8F,$47,$56,$2B
	.dc.b	$78,$A3,$18,$7E,$42,$BD,$DD,$BF,$9B,$EF,$FC,$F4,$60,$FD,$C7,$81
	.dc.b	$FA,$23,$F7,$DD,$DF,$E0,$58,$FE,$9D,$E0,$FE,$AB,$C0,$C4,$7F,$A6
	.dc.b	$E3,$F5,$BE,$16,$23,$F7,$BA,$FE,$63,$3F,$DD,$43,$C7,$FE,$B0,$EA
	.dc.b	$3E,$01,$E7,$8C,$F8,$9F,$C2,$B4,$F3,$C6,$34,$F3,$67,$FC,$33,$0E
	.dc.b	$87,$F6,$3F,$C4,$39,$F0,$FF,$0D,$FC,$43,$C9,$18,$D3,$7A,$F8,$B1
	.dc.b	$E7,$9B,$F1,$0F,$E2,$D1,$8D,$FF,$81,$E7,$FE,$B9,$F8,$1F,$84,$FF
	.dc.b	$0A,$31,$7E,$7A,$FE,$1D,$E1,$7F,$63,$9F,$F6,$07,$E1,$FF,$0F,$67
	.dc.b	$E1,$FD,$25,$39,$FC,$3F,$F8,$0C,$7F,$B1,$F0,$F2,$E9,$88,$ED,$C3
	.dc.b	$FE,$5C,$D7,$F8,$1C,$7F,$52,$F1,$BF,$C5,$0F,$F0,$23,$FC,$68,$FF
	.dc.b	$83,$F4,$FC,$97,$8F,$FE,$34,$F0,$7F,$C4,$F5,$B8,$C4,$7E,$9A,$7F
	.dc.b	$4B,$1F,$E8,$9A,$FF,$27,$C6,$1C,$79,$DF,$BB,$BF,$14,$7F,$88,$3C
	.dc.b	$8F,$E0,$37,$C6,$3F,$B9,$46,$1E,$9F,$4F,$FB,$69,$D3,$63,$AD,$7F
	.dc.b	$82,$FC,$6F,$F1,$05,$73,$F1,$3F,$C0,$9E,$2E,$2F,$CD,$18,$BF,$1C
	.dc.b	$7F,$7B,$F0,$7E,$0B,$FD,$AB,$5F,$DF,$C6,$3D,$0F,$ED,$4E,$A7,$E5
	.dc.b	$B5,$FF,$1C,$3F,$CE,$AB,$E9,$F3,$FE,$25,$FD,$C1,$F6,$78,$7F,$AA
	.dc.b	$F0,$FF,$86,$7C,$1F,$F1,$C7,$37,$E0,$F9,$F1,$7F,$86,$F0,$3F,$33
	.dc.b	$D1,$88,$75,$7E,$43,$4F,$C6,$46,$34,$FC,$87,$5F,$EC,$7A,$BF,$0F
	.dc.b	$3F,$A4,$3C,$53,$FA,$BD,$7F,$6D,$A7,$86,$74,$7E,$4B,$3F,$0B,$FC
	.dc.b	$E7,$97,$E6,$70,$7F,$B2,$72,$FA,$8D,$9F,$D1,$1F,$9D,$EC,$77,$7E
	.dc.b	$77,$0F,$D8,$77,$FE,$5A,$BF,$D9,$5E,$35,$E6,$FE,$79,$E2,$D7,$FC
	.dc.b	$6F,$D1,$7D,$DF,$F6,$A5,$FF,$B3,$3A,$BF,$CC,$15,$FE,$AB,$EF,$FE
	.dc.b	$72,$F2,$9F,$41,$FE,$AD,$C7,$F2,$8F,$FB,$67,$1B,$3F,$D2,$FC,$9F
	.dc.b	$E8,$AD,$9F,$C7,$1C,$4A,$71,$3F,$A2,$19,$E6,$F0,$C7,$F5,$9A,$FE
	.dc.b	$87,$3F,$E4,$AE,$0F,$F2,$6D,$7F,$41,$AF,$F1,$EC,$7F,$94,$E3,$11
	.dc.b	$FD,$87,$0E,$35,$DE,$7E,$FE,$DF,$C6,$77,$FC,$23,$F3,$27,$F9,$D1
	.dc.b	$FC,$D7,$CF,$FC,$47,$37,$F0,$6F,$1B,$DF,$97,$F6,$DF,$2B,$DE,$DB
	.dc.b	$FF,$7B,$F9,$1F,$E8,$BE,$BF,$E3,$BC,$3F,$D2,$E7,$F8,$EF,$5F,$F4
	.dc.b	$5E,$8F,$E2,$5A,$F8,$E7,$FD,$FF,$E4,$FF,$A4,$E3,$C9,$7F,$E7,$C7
	.dc.b	$4D,$F1,$FF,$C5,$1D,$3F,$83,$F5,$3F,$75,$87,$F1,$77,$F8,$37,$8F
	.dc.b	$9E,$6F,$F5,$EF,$BB,$FC,$5B,$1F,$ED,$55,$FE,$6C,$8F,$F3,$7C,$7F
	.dc.b	$7A,$8E,$6D,$9F,$E1,$3E,$47,$F8,$4A,$7B,$B6,$33,$DF,$FE,$EB,$5E
	.dc.b	$47,$F8,$22,$BC,$79,$7E,$09,$FE,$2A,$D9,$FE,$1E,$FE,$10,$FF,$78
	.dc.b	$B8,$E9,$F1,$3F,$9A,$7D,$2F,$B3,$93,$FD,$17,$18,$F5,$7F,$D0,$75
	.dc.b	$FE,$09,$07,$4A,$FE,$27,$E6,$F9,$F6,$FF,$52,$7A,$7F,$47,$87,$FC
	.dc.b	$A9,$D9,$8C,$38,$F1,$FE,$1F,$3F,$E3,$0F,$D9,$6D,$7D,$49,$FF,$06
	.dc.b	$9F,$80,$F2,$7F,$CF,$4E,$0F,$E2,$1C,$FF,$83,$7F,$23,$F8,$87,$F1
	.dc.b	$0B,$FE,$2A,$FF,$69,$87,$27,$FA,$FA,$BF,$8A,$B6,$FF,$92,$CE,$BE
	.dc.b	$0F,$F8,$7E,$FF,$E5,$76,$BF,$E5,$57,$FD,$65,$C3,$FE,$36,$B4,$63
	.dc.b	$C1,$FF,$37,$7D,$B7,$0B,$FF,$5E,$D2,$31,$C3,$FE,$5A,$C6,$9F,$E5
	.dc.b	$73,$87,$DB,$E5,$FE,$93,$7F,$E8,$9F,$53,$FC,$33,$E8,$7E,$11,$B1
	.dc.b	$FD,$97,$BB,$FC,$B1,$5C,$8F,$F2,$76,$3E,$3F,$27,$FD,$B4,$F8,$38
	.dc.b	$BD,$33,$FE,$CA,$F4,$1C,$1F,$DD,$27,$32,$CF,$FB,$AC,$B7,$FF,$72
	.dc.b	$70,$75,$AF,$F2,$E3,$D7,$9F,$F1,$31,$FD,$5B,$D1,$FC,$23,$FD,$63
	.dc.b	$CE,$FF,$76,$78,$9F,$D5,$3A,$7F,$C5,$3E,$86,$2B,$FA,$A7,$57,$FA
	.dc.b	$C7,$D0,$FF,$4D,$7E,$39,$D3,$CA,$FF,$09,$F9,$78,$BF,$CB,$C7,$F1
	.dc.b	$17,$F0,$3F,$19,$FC,$E3,$87,$F8,$F4,$60,$FC,$EB,$FC,$E7,$B1,$3F
	.dc.b	$EF,$5E,$7F,$F5,$43,$FE,$E2,$AF,$C5,$FD,$9F,$3B,$F8,$E7,$FB,$D2
	.dc.b	$D1,$98,$FF,$79,$F3,$62,$BF,$9C,$4E,$67,$FD,$07,$3E,$09,$FE,$71
	.dc.b	$87,$EC,$23,$11,$DB,$D1,$FD,$D6,$3C,$6F,$FC,$12,$DE,$6E,$1A,$FC
	.dc.b	$5F,$03,$FC,$AF,$B6,$31,$F8,$B1,$FE,$5A,$F3,$BF,$7B,$F1,$BF,$DF
	.dc.b	$32,$FF,$4F,$74,$FF,$00,$F8,$6C,$47,$FA,$96,$3F,$D9,$5F,$29,$E1
	.dc.b	$F4,$AE,$5F,$2A,$BF,$A9,$3F,$B0,$9F,$94,$D3,$FA,$A7,$89,$FE,$D3
	.dc.b	$FB,$2E,$51,$FB,$67,$D4,$FF,$A9,$5C,$F1,$1C,$AA,$EA,$D9,$EA,$7F
	.dc.b	$AB,$63,$89,$C3,$F1,$1F,$89,$E3,$BA,$3F,$D2,$76,$9C,$C3,$83,$FD
	.dc.b	$27,$87,$F0,$AF,$8A,$FC,$DF,$B4,$3F,$AB,$47,$FA,$33,$A9,$3A,$9C
	.dc.b	$CB,$E1,$BF,$95,$FD,$1B,$E8,$BF,$4B,$C7,$FE,$94,$AF,$F4,$AC,$BE
	.dc.b	$83,$96,$1D,$CF,$07,$B3,$FC,$D3,$F0,$6E,$57,$FC,$DB,$0E,$4F,$EC
	.dc.b	$23,$93,$FC,$EA,$B9,$1C,$EB,$F1,$D9,$F9,$FC,$5E,$3C,$3E,$AB,$E1
	.dc.b	$7E,$AD,$7F,$29,$F8,$8F,$A2,$7B,$9F,$ED,$BC,$3C,$F5,$D6,$3E,$41
	.dc.b	$CB,$DB,$F0,$E5,$E0,$8C,$4E,$F6,$F3,$C7,$BF,$3F,$5C,$62,$7E,$84
	.dc.b	$7F,$AE,$B2,$F8,$5B,$7A,$6F,$E5,$6B,$BB,$F4,$47,$94,$F5,$F9,$5F
	.dc.b	$C5,$3E,$33,$F8,$AE,$FE,$F9,$F9,$D3,$9C,$7E,$BE,$3F,$6F,$5F,$4D
	.dc.b	$32,$ED,$18,$7F,$93,$47,$D3,$3E,$2A,$FC,$47,$E2,$E7,$F6,$A7,$FC
	.dc.b	$53,$F9,$8B,$F4,$9F,$CA,$7E,$08,$F9,$FF,$F2,$DF,$ED,$AF,$D6,$7F
	.dc.b	$C5,$FE,$E3,$F7,$A7,$E6,$AF,$3B,$F3,$1F,$48,$FD,$87,$F2,$1C,$DF
	.dc.b	$E7,$8E,$D7,$D8,$3F,$48,$FD,$67,$CC,$BF,$2A,$7E,$F1,$FD,$73,$0E
	.dc.b	$1E,$13,$F5,$D7,$E7,$AF,$D4,$9F,$B8,$BF,$39,$7E,$63,$FE,$44,$DF
	.dc.b	$FC,$13,$FB,$6F,$E9,$3E,$79,$F5,$2F,$C0,$9F,$95,$6F,$C1,$7E,$C0
	.dc.b	$F9,$57,$CF,$6B,$F1,$9E,$3C,$9E,$BB,$FF,$0A,$FE,$44,$F9,$17,$C9
	.dc.b	$3C,$67,$E9,$3D,$6D,$F5,$6E,$17,$E0,$2F,$C6,$5F,$99,$6F,$D9,$7E
	.dc.b	$9A,$FB,$2F,$E2,$5F,$91,$73,$CD,$F8,$1B,$EF,$39,$FB,$F9,$FD,$15
	.dc.b	$F9,$43,$F6,$57,$CB,$F1,$C5,$F9,$B3,$F6,$87,$D1,$67,$F1,$77,$EC
	.dc.b	$3E,$2E,$DF,$81,$87,$F6,$CF,$01,$F3,$4F,$DC,$BA,$4B,$F4,$F7,$E4
	.dc.b	$98,$F8,$97,$BF,$AA,$7D,$9D,$AF,$F5,$5F,$9B,$FB,$C6,$CF,$F0,$C9
	.dc.b	$AF,$F7,$C1,$7D,$3E,$3B,$CB,$1F,$E0,$0C,$7C,$31,$88,$FB,$99,$F9
	.dc.b	$E3,$1B,$FB,$BB,$3A,$D7,$F9,$43,$8F,$E2,$71,$63,$9F,$E5,$F3,$7C
	.dc.b	$AC,$7B,$F5,$7D,$1F,$03,$F9,$64,$7C,$6E,$EF,$B3,$DD,$F6,$23,$E4
	.dc.b	$F8,$5F,$81,$AF,$C0,$78,$1E,$CF,$1F,$F1,$5E,$37,$E2,$7C,$5F,$83
	.dc.b	$E5,$7E,$4F,$CA,$FC,$97,$9B,$F1,$3C,$EF,$CE,$3F,$9B,$F3,$7E,$4F
	.dc.b	$AB,$FA,$7F,$53,$F4,$DE,$9F,$E9,$7D,$2F,$9B,$ED,$7E,$C7,$E5,$1E
	.dc.b	$D7,$EC,$7D,$9F,$F2,$5F,$2F,$FB,$BB,$E0,$7F,$7D,$F0,$1F,$BD,$F9
	.dc.b	$07,$C0,$7E,$F7,$E3,$3F,$DC,$3F,$0F,$FC,$27,$E4,$7F,$41,$F0,$FF
	.dc.b	$C2,$3E,$1D,$F5,$CF,$D5,$BF,$C6,$7F,$7E,$FC,$77,$FC,$2F,$F9,$E3
	.dc.b	$F2,$27,$F0,$0B,$EA,$FF,$D0,$38,$F1,$F3,$BF,$D0,$78,$F1,$FA,$76
	.dc.b	$3C,$70,$7D,$BF,$0F,$CF,$BF,$C8,$3F,$51,$FC,$6F,$CC,$7D,$43,$F7
	.dc.b	$37,$D7,$70,$D7,$CF,$87,$0E,$4F,$D8,$5E,$E9,$F7,$5F,$E9,$4F,$49
	.dc.b	$FB,$B7,$F9,$3B,$E5,$1F,$BC,$BF,$45,$7E,$4E,$FD,$79,$FB,$CF,$F9
	.dc.b	$83,$1E,$3B,$FF,$A7,$B1,$E3,$87,$0B,$DF,$D9,$F7,$FD,$D3,$FB,$05
	.dc.b	$FF,$6E,$C3,$F8,$03,$FD,$9B,$E8,$7F,$A2,$C2,$7F,$CC,$3F,$95,$7E
	.dc.b	$9B,$E8,$9F,$93,$7F,$83,$FF,$75,$FF,$22,$5E,$F1,$FF,$57,$DE,$EB
	.dc.b	$5F,$07,$EF,$FE,$01,$FF,$35,$FE,$94,$FC,$99,$E1,$33,$7E,$8C,$FC
	.dc.b	$77,$E3,$E9,$FC,$5B,$1E,$38,$70,$BD,$D3,$ED,$FE,$67,$0F,$A3,$E3
	.dc.b	$FF,$AA,$6D,$E7,$FA,$9C,$E7,$E6,$9C,$FD,$9F,$87,$83,$32,$E1,$F5
	.dc.b	$4F,$97,$C8,$CC,$63,$57,$DE,$F0,$D3,$FC,$2A,$7F,$8B,$78,$B0,$E3
	.dc.b	$EE,$EB,$FD,$CB,$ED,$3F,$89,$7C,$BF,$F8,$4B,$44,$61,$FF,$16,$7A
	.dc.b	$BF,$E3,$07,$D9,$D3,$F4,$B6,$F4,$47,$03,$3C,$5F,$57,$7E,$11,$EF
	.dc.b	$AE,$1B,$7A,$23,$BF,$CE,$7F,$A7,$BC,$EF,$D7,$FC,$FB,$F2,$B7,$EE
	.dc.b	$AF,$CA,$7F,$3F,$87,$03,$FD,$DB,$7C,$2B,$73,$FE,$B4,$DD,$FA,$5F
	.dc.b	$FA,$AB,$E4,$9F,$BE,$7F,$84,$7D,$D7,$E7,$CF,$0F,$FA,$4E,$9F,$0B
	.dc.b	$0E,$17,$BA,$D7,$77,$FB,$F3,$E0,$DF,$BF,$7F,$AD,$BF,$69,$7E,$8C
	.dc.b	$FC,$5F,$FC,$3D,$F7,$4F,$DD,$5F,$81,$F1,$C3,$85,$EE,$BF,$C4,$94
	.dc.b	$FC,$4B,$B7,$ED,$FF,$EC,$2F,$DA,$5F,$BB,$BF,$04,$FD,$B7,$E3,$5F
	.dc.b	$E0,$DF,$13,$7C,$A1,$7C,$2B,$7F,$7D,$FE,$CC,$E2,$FE,$D3,$5B,$FD
	.dc.b	$47,$FC,$39,$E2,$3E,$C9,$FA,$7B,$F6,$1F,$65,$B0,$E2,$B0,$E3,$7C
	.dc.b	$29,$F8,$FF,$C7,$FC,$F3,$F0,$BE,$5F,$E2,$31,$46,$3C,$9F,$F5,$F4
	.dc.b	$78,$72,$FD,$37,$AF,$E1,$F6,$1F,$BE,$FF,$2D,$9F,$16,$9F,$AF,$AD
	.dc.b	$4F,$F2,$3B,$E5,$5F,$AB,$F5,$BF,$5F,$8F,$6C,$3C,$1E,$B7,$DB,$7F
	.dc.b	$71,$F8,$2F,$F2,$0F,$A1,$74,$7E,$13,$2E,$7E,$C7,$F7,$8A,$D4,$F3
	.dc.b	$CF,$FC,$E7,$EA,$FF,$A7,$2F,$C2,$FF,$A7,$FF,$09,$FF,$49,$0F,$C3
	.dc.b	$3C,$15,$FE,$86,$EA,$FF,$46,$FA,$F8,$8F,$F3,$D5,$FF,$A6,$7E,$03
	.dc.b	$0F,$F9,$E8,$FF,$4E,$7C,$4F,$E4,$FF,$1E,$FE,$5B,$F8,$A7,$FA,$7D
	.dc.b	$C5,$FF,$9A,$CF,$F7,$AD,$7F,$AC,$7F,$1B,$CC,$7C,$5F,$D5,$DF,$B4
	.dc.b	$7F,$8B,$FE,$35,$F2,$9A,$E4,$FF,$0B,$D6,$BE,$D3,$FA,$4F,$CD,$E7
	.dc.b	$AF,$E6,$F8,$B7,$E0,$7F,$7F,$D9,$7E,$0F,$F3,$1F,$B6,$BF,$1D,$C7
	.dc.b	$42,$1F,$8A,$7E,$B7,$C8,$FC,$D7,$1F,$F4,$3F,$C8,$1F,$AF,$FF,$A2
	.dc.b	$7F,$78,$7E,$EA,$FC,$77,$85,$FE,$87,$E4,$98,$70,$F6,$18,$70,$DF
	.dc.b	$F7,$58,$7E,$47,$8A,$E3,$FC,$5F,$E3,$3F,$A4,$BF,$9F,$FD,$AB,$4D
	.dc.b	$35,$E8,$7E,$6F,$F5,$CC,$F9,$07,$F1,$B3,$FA,$7F,$A8,$C9,$89,$7E
	.dc.b	$65,$C0,$C7,$E1,$DF,$03,$F3,$DF,$CF,$FE,$4B,$CF,$F0,$DF,$C9,$F8
	.dc.b	$DF,$C4,$6B,$5E,$A2,$B5,$E8,$7E,$E3,$4F,$DB,$57,$90,$F0,$73,$F9
	.dc.b	$8F,$17,$EB,$BE,$83,$C9,$CB,$EA,$3C,$DF,$80,$FB,$0F,$47,$E5,$9E
	.dc.b	$E3,$D5,$F8,$FF,$C0,$3D,$9F,$89,$FC,$23,$DD,$FA,$0F,$C4,$3E,$07
	.dc.b	$F0,$0F,$C6,$3E,$17,$BB,$F2,$0F,$89,$F3,$9F,$28,$F8,$DE,$F7,$CC
	.dc.b	$3E,$47,$9D,$F3,$8F,$95,$E5,$FD,$03,$E6,$7B,$7F,$48,$F9,$DF,$AB
	.dc.b	$FD,$42,$F1,$FF,$25,$FA,$CF,$F4,$BF,$E9,$4F,$CA,$5F,$9A,$7F,$6C
	.dc.b	$F8,$7F,$E7,$D8,$7E,$4A,$F7,$5A,$FA,$27,$6F,$F4,$3E,$DF,$C0,$9F
	.dc.b	$77,$F2,$1F,$C1,$9F,$7F,$E0,$BF,$0A,$7E,$07,$C6,$FC,$39,$F8,$3F
	.dc.b	$AF,$FC,$49,$F8,$5F,$C8,$BF,$16,$7E,$1F,$F2,$6F,$C6,$9F,$89,$F8
	.dc.b	$6F,$D2,$FE,$C6,$FD,$7F,$FD,$97,$F9,$A3,$F0,$2F,$F2,$5F,$A5,$FD
	.dc.b	$5E,$F7,$5A,$9F,$F2,$7F,$89,$9E,$8F,$E9,$1F,$46,$8F,$F6,$BF,$EF
	.dc.b	$FF,$EA,$DF,$98,$7C,$66,$EF,$FD,$83,$5A,$FC,$7F,$FA,$C7,$3E,$7E
	.dc.b	$1B,$FC,$2D,$EC,$5F,$84,$FE,$63,$F0,$9C,$7E,$09,$FB,$2B,$CD,$7C
	.dc.b	$7B,$F7,$79,$E5,$EE,$B5,$39,$9F,$36,$FF,$D4,$FE,$1B,$DB,$F8,$63
	.dc.b	$AB,$E1,$7D,$F7,$F0,$5A,$D1,$FD,$97,$B3,$E4,$56,$FA,$7E,$1C,$E8
	.dc.b	$7E,$BD,$FD,$7F,$EF,$9F,$A9,$3F,$4C,$7E,$FF,$FE,$14,$F8,$2F,$EE
	.dc.b	$98,$7F,$71,$C3,$85,$EE,$BF,$AC,$56,$AB,$FA,$A7,$A5,$FE,$24,$E8
	.dc.b	$FF,$06,$70,$7F,$05,$3F,$90,$FB,$CF,$D7,$DE,$AF,$FB,$7B,$E1,$9F
	.dc.b	$BA,$74,$BD,$D6,$BB,$15,$AF,$99,$EB,$A7,$F8,$17,$B7,$E8,$F6,$FD
	.dc.b	$0E,$AF,$9F,$3F,$06,$73,$D7,$F3,$FC,$2E,$33,$FE,$85,$FD,$75,$F8
	.dc.b	$C7,$FB,$C7,$D5,$7E,$E5,$FE,$1E,$AF,$95,$F8,$02,$F7,$E3,$AF,$7F
	.dc.b	$5F,$D3,$E1,$1F,$9F,$FE,$46,$7F,$C6,$FF,$11,$FD,$9F,$2F,$A9,$5A
	.dc.b	$D7,$97,$D9,$2B,$FC,$F7,$5F,$67,$A7,$C1,$8B,$F4,$3F,$F5,$AF,$C9
	.dc.b	$7F,$4E,$AD,$6D,$E0,$9C,$FC,$CF,$BB,$E0,$1F,$84,$3E,$1B,$FD,$4F
	.dc.b	$FA,$5F,$FA,$EF,$1F,$C3,$56,$A7,$3F,$87,$BE,$B7,$E3,$F8,$2F,$DF
	.dc.b	$1F,$A2,$FF,$AB,$3C,$A7,$E5,$DF,$9A,$F2,$BF,$09,$7B,$AD,$4F,$B2
	.dc.b	$7F,$11,$F4,$DF,$E9,$7F,$C8,$1D,$8F,$DE,$DF,$91,$3F,$2A,$FA,$F7
	.dc.b	$BA,$D4,$FF,$8E,$B0,$E7,$F0,$BF,$ED,$7F,$D6,$9F,$A7,$FF,$B3,$3C
	.dc.b	$7E,$33,$98,$FF,$67,$FC,$0F,$4C,$9F,$BE,$3E,$1D,$FA,$1F,$FB,$1F
	.dc.b	$D1,$D4,$E6,$3F,$D9,$5E,$AF,$1F,$8F,$FF,$2D,$7F,$00,$BF,$6E,$FB
	.dc.b	$DC,$1E,$EC,$FF,$C7,$FD,$0A,$F9,$9D,$FE,$19,$7E,$94,$FD,$0D,$F8
	.dc.b	$77,$F9,$42,$BF,$75,$42,$56,$E7,$A4,$EA,$7F,$C6,$DF,$9D,$FA,$BE
	.dc.b	$17,$B7,$12,$3D,$BF,$8D,$78,$69,$F2,$8F,$CB,$7F,$81,$6D,$FE,$18
	.dc.b	$9F,$01,$E1,$9D,$7E,$3F,$C6,$3F,$D9,$F9,$5F,$9F,$FD,$D6,$9F,$EE
	.dc.b	$39,$E3,$F8,$66,$DF,$EF,$9C,$5E,$B9,$FF,$1E,$1F,$ED,$C3,$FC,$E9
	.dc.b	$1F,$EE,$E7,$FC,$53,$F3,$BF,$E9,$78,$C4,$7E,$57,$EC,$F0,$F2,$3F
	.dc.b	$31,$F9,$FE,$1F,$E6,$67,$F8,$C8,$FE,$41,$F8,$06,$7C,$0C,$47,$F2
	.dc.b	$3E,$7C,$3F,$E3,$8F,$BD,$7E,$52,$FB,$8C,$CF,$FA,$56,$73,$F7,$B8
	.dc.b	$87,$A7,$FC,$7B,$6C,$3F,$C8,$38,$38,$FD,$7B,$F2,$27,$E4,$EE,$0C
	.dc.b	$C6,$38,$7B,$FE,$70,$FD,$EB,$FD,$63,$F1,$98,$7F,$D6,$7C,$7E,$4F
	.dc.b	$D8,$3F,$15,$F8,$2F,$E3,$BA,$3E,$77,$0F,$C2,$13,$A7,$DF,$F3,$DF
	.dc.b	$D6,$3E,$3D,$FC,$03,$FE,$DD,$BE,$38,$FF,$AC,$EB,$53,$FD,$AA,$FE
	.dc.b	$B9,$FE,$B7,$9F,$F5,$C7,$CE,$FF,$79,$F8,$8F,$EA,$78,$72,$AF,$EA
	.dc.b	$D1,$FD,$F1,$E2,$7F,$68,$E3,$7E,$5C,$FC,$4D,$F0,$0F,$CB,$7E,$DE
	.dc.b	$3E,$4B,$DD,$6A,$7C,$93,$E8,$F6,$BF,$3B,$0F,$DF,$1F,$BA,$7B,$7C
	.dc.b	$4F,$F3,$AE,$FF,$C3,$FC,$8E,$D5,$AF,$4F,$F0,$DF,$8E,$FF,$16,$FC
	.dc.b	$4D,$F5,$EF,$13,$FB,$1E,$5C,$AB,$53,$9E,$CF,$C9,$7C,$DB,$E9,$1D
	.dc.b	$5F,$C1,$FF,$80,$1F,$A0,$FF,$71,$CB,$FD,$0F,$7B,$AD,$4F,$6E,$9F
	.dc.b	$F7,$27,$EE,$5F,$DE,$6D,$FB,$F3,$F6,$B6,$DC,$EF,$7F,$4F,$FC,$26
	.dc.b	$B5,$3C,$E5,$FB,$03,$F6,$87,$ED,$6F,$CD,$7F,$DC,$3F,$BF,$FF,$2F
	.dc.b	$D3,$FD,$37,$7B,$AD,$7A,$CA,$FE,$71,$8F,$DC,$FD,$35,$FB,$07,$F0
	.dc.b	$DF,$8D,$FF,$7A,$7D,$77,$FA,$A6,$B5,$39,$FC,$67,$FC,$2F,$E8,$FE
	.dc.b	$E7,$EC,$BF,$C3,$BE,$67,$F8,$8B,$EE,$F1,$F7,$7F,$E4,$1F,$80,$FE
	.dc.b	$A1,$C5,$FE,$AE,$5F,$9F,$FE,$DF,$F3,$DF,$E4,$1F,$81,$BE,$F3,$F8
	.dc.b	$25,$6B,$E0,$55,$AF,$C9,$3F,$84,$7C,$57,$FA,$BB,$EF,$5F,$16,$FD
	.dc.b	$91,$F3,$9F,$89,$E6,$1F,$CB,$7E,$A1,$7C,$2B,$7F,$31,$FC,$4F,$F0
	.dc.b	$0F,$EB,$9F,$3D,$FD,$AB,$F7,$8F,$F1,$97,$D7,$67,$DA,$FE,$C5,$F4
	.dc.b	$49,$FF,$1B,$7D,$FF,$F1,$BF,$03,$FC,$67,$E5,$FF,$8E,$3D,$3C,$7B
	.dc.b	$AF,$BD,$FF,$1E,$CF,$F8,$F3,$E5,$3F,$91,$7E,$91,$FD,$73,$E6,$35
	.dc.b	$A7,$FA,$1F,$A7,$3F,$3D,$FD,$87,$1F,$EF,$BF,$83,$FC,$7E,$8F,$DA
	.dc.b	$FE,$6C,$BF,$F4,$6F,$E1,$BF,$E3,$AF,$D3,$FF,$9D,$F8,$BF,$ED,$EA
	.dc.b	$D7,$88,$AD,$7C,$AF,$F3,$3F,$B3,$7E,$C4,$FC,$DF,$FC,$41,$F8,$1B
	.dc.b	$3F,$E5,$DA,$FF,$04,$BD,$D6,$A7,$FD,$F7,$3F,$C2,$3F,$07,$FF,$3A
	.dc.b	$B7,$E0,$3F,$CE,$AF,$F0,$CF,$27,$F6,$5F,$20,$F0,$BF,$B5,$7E,$59
	.dc.b	$FD,$16,$73,$1F,$DB,$7D,$9F,$E9,$1F,$47,$FE,$D3,$F9,$BF,$F3,$2F
	.dc.b	$E3,$CD,$3F,$86,$F3,$7F,$BF,$2B,$5F,$83,$FF,$BF,$27,$F8,$AC,$FF
	.dc.b	$96,$BE,$EF,$FC,$BD,$F8,$37,$EB,$FF,$07,$FF,$31,$46,$3C,$5F,$EA
	.dc.b	$DF,$9E,$BF,$0C,$7E,$26,$FD,$19,$F8,$EF,$F3,$8F,$C0,$37,$5A,$9C
	.dc.b	$FD,$17,$F9,$0F,$CC,$FE,$7F,$FB,$7F,$EC,$FF,$7B,$FE,$9B,$F9,$93
	.dc.b	$F7,$37,$BF,$FD,$96,$B5,$F5,$6A,$D7,$A3,$FD,$A7,$E0,$DF,$AC,$7F
	.dc.b	$55,$FB,$CF,$85,$F9,$6F,$F8,$E3,$C4,$FE,$73,$5A,$EF,$56,$BC,$FF
	.dc.b	$EE,$5F,$88,$7F,$9C,$BF,$11,$FF,$3A,$F1,$7F,$54,$8C,$47,$EE,$B1
	.dc.b	$FD,$5F,$EA,$FF,$E2,$2F,$0D,$FB,$37,$FA,$B7,$F7,$AF,$D9,$D7,$F4
	.dc.b	$8A,$D4,$E7,$F3,$EF,$F0,$07,$E2,$BF,$D7,$BC,$27,$E8,$0F,$AE,$7E
	.dc.b	$CD,$FC,$E7,$2F,$EA,$37,$BA,$D4,$FF,$B9,$27,$FA,$AF,$5B,$F3,$87
	.dc.b	$DD,$BF,$75,$7E,$D9,$FE,$57,$FC,$23,$FC,$21,$7B,$AD,$4F,$F9,$6F
	.dc.b	$EB,$BF,$C2,$9F,$8C,$FF,$A2,$F5,$FF,$48,$7E,$35,$FE,$8F,$E8,$FE
	.dc.b	$EF,$39,$FC,$6B,$FD,$1D,$3F,$3E,$7F,$BD,$FC,$D7,$FA,$47,$D9,$7D
	.dc.b	$0B,$F4,$37,$E4,$9F,$C5,$38,$FF,$B6,$DE,$EB,$53,$FE,$05,$F5,$75
	.dc.b	$F9,$63,$E2,$1F,$B2,$3F,$65,$FE,$73,$3F,$1F,$5F,$F0,$5B,$B9,$D7
	.dc.b	$D2,$A3,$F7,$33,$FE,$10,$FC,$AD,$E8,$FF,$A1,$BF,$7B,$FE,$F7,$FA
	.dc.b	$8F,$E4,$B8,$3F,$C0,$35,$AF,$C5,$D5,$AF,$A1,$FF,$1D,$7E,$43,$FE
	.dc.b	$A7,$FC,$8B,$FD,$57,$F9,$3E,$3E,$37,$FD,$5D,$F8,$2F,$FA,$A7,$E7
	.dc.b	$3F,$C8,$1F,$26,$FD,$A7,$FD,$07,$D4,$FD,$A9,$FB,$6F,$F5,$1E,$4F
	.dc.b	$F1,$2D,$EF,$F1,$B5,$EE,$BF,$59,$D7,$F2,$FF,$14,$7E,$73,$FC,$8F
	.dc.b	$F5,$3F,$ED,$1F,$81,$FF,$27,$CF,$FC,$5F,$5A,$F6,$95,$A7,$FC,$A9
	.dc.b	$F1,$AF,$DB,$FF,$61,$F4,$7F,$EA,$FF,$C6,$7E,$C6,$3F,$E3,$8A,$D7
	.dc.b	$C2,$AB,$5F,$91,$FF,$97,$7F,$25,$FF,$5C,$7E,$7B,$FE,$50,$FC,$47
	.dc.b	$8F,$0F,$E0,$67,$3F,$8C,$7F,$8A,$AF,$FC,$C9,$F9,$7B,$F1,$65,$7F
	.dc.b	$80,$BF,$4B,$7E,$FF,$FC,$F3,$C7,$FE,$BB,$7B,$AD,$4F,$E9,$78,$3F
	.dc.b	$C9,$BC,$DF,$E8,$DF,$15,$F9,$CF,$F9,$A3,$E6,$3F,$93,$E5,$FD,$9A
	.dc.b	$B5,$39,$FC,$53,$FC,$ED,$F5,$EF,$B0,$FF,$11,$FE,$60,$FD,$81,$F5
	.dc.b	$FF,$A1,$E5,$FC,$E5,$EF,$F0,$F5,$EE,$BF,$DF,$1F,$3D,$FE,$7F,$8F
	.dc.b	$F5,$DF,$EB,$2F,$4B,$FD,$C5,$EF,$7E,$F6,$B5,$F6,$6A,$D7,$E5,$AA
	.dc.b	$FE,$0D,$5F,$E8,$9F,$31,$FA,$8B,$F3,$1F,$F3,$C7,$EA,$AF,$85,$FF
	.dc.b	$7A,$5E,$EB,$5F,$34,$AD,$3F,$E0,$89,$FF,$4A,$7E,$F0,$F5,$3F,$C7
	.dc.b	$7F,$6A,$FD,$19,$F1,$7D,$3B,$7F,$79,$7B,$FC,$53,$FB,$5D,$6B,$2E
	.dc.b	$95,$FE,$4A,$FC,$67,$FC,$9F,$F9,$63,$F1,$AF,$F2,$7F,$D7,$6A,$BF
	.dc.b	$C3,$3F,$8D,$7F,$93,$EB,$5D,$8F,$A7,$D5,$3F,$1A,$78,$7F,$17,$A3
	.dc.b	$E0,$7E,$DD,$F2,$EB,$F8,$85,$6B,$AB,$F6,$1D,$2A,$F9,$75,$FC,$57
	.dc.b	$E6,$5F,$3C,$FC,$ED,$FA,$8F,$F6,$33,$A5,$7F,$8B,$AB,$73,$AF,$B1
	.dc.b	$DF,$E5,$5E,$6F,$D0,$7F,$CD,$BF,$B1,$FD,$9F,$D1,$DE,$DF,$F8,$6F
	.dc.b	$C1,$FF,$61,$56,$B8,$15,$AF,$81,$F7,$FE,$3B,$FE,$5D,$FC,$7B,$FC
	.dc.b	$C3,$F8,$0E,$3F,$1F,$FF,$32,$C7,$F9,$93,$F0,$DF,$8B,$F0,$FF,$E6
	.dc.b	$8F,$CE,$7F,$0B,$FB,$FB,$F4,$CF,$4F,$37,$F0,$2F,$C8,$95,$A9,$F8
	.dc.b	$73,$FE,$4A,$FD,$4B,$F1,$9F,$05,$FD,$3F,$F0,$4F,$F0,$0F,$E6,$7F
	.dc.b	$4A,$73,$F5,$C9,$FF,$02,$FE,$0F,$F5,$BE,$F1,$F3,$AC,$3F,$D2,$1F
	.dc.b	$6E,$FD,$6F,$F9,$87,$5F,$FB,$8E,$F7,$5A,$9F,$C0,$6B,$FE,$75,$3F
	.dc.b	$CB,$BF,$55,$FE,$7A,$FC,$21,$F9,$27,$F9,$EB,$1F,$C2,$CE,$63,$FC
	.dc.b	$FB,$F9,$1F,$E2,$3C,$DF,$EC,$BF,$84,$7F,$3C,$FD,$A7,$FC,$25,$F8
	.dc.b	$E6,$7E,$D5,$39,$FB,$7F,$CF,$EB,$E1,$1F,$BD,$B8,$9F,$B6,$FF,$7B
	.dc.b	$F8,$F7,$E3,$7F,$EE,$9B,$E1,$7B,$AD,$4F,$EA,$67,$FC,$EF,$F1,$1F
	.dc.b	$27,$F2,$D7,$E8,$CF,$C0,$7F,$B4,$F4,$7E,$82,$B5,$39,$FB,$8E,$DC
	.dc.b	$FF,$DB,$71,$7E,$51,$FE,$96,$F9,$0F,$F6,$E4,$E6,$31,$F8,$6F,$9A
	.dc.b	$BD,$9F,$83,$3F,$18,$7E,$82,$FD,$A9,$FA,$23,$CC,$7D,$27,$EC,$B0
	.dc.b	$E1,$7B,$AD,$63,$FB,$6F,$CB,$FF,$C9,$1F,$A6,$3F,$13,$7E,$C9,$FE
	.dc.b	$59,$F0,$F5,$F8,$B2,$B5,$3F,$E5,$DF,$90,$FD,$FF,$E4,$5F,$C1,$FF
	.dc.b	$47,$7D,$DF,$F9,$E3,$A7,$EA,$F3,$FC,$1A,$D4,$E7,$F0,$CF,$E0,$FC
	.dc.b	$FF,$EA,$DF,$CB,$3F,$D6,$3F,$96,$E3,$F2,$EF,$F5,$A7,$E5,$CF,$C9
	.dc.b	$BF,$9B,$F2,$FF,$85,$FF,$00,$FE,$DB,$FB,$53,$F6,$37,$E4,$8F,$67
	.dc.b	$FD,$03,$C3,$F4,$6F,$75,$A9,$FF,$44,$CF,$D5,$F6,$FE,$AF,$DB,$FD
	.dc.b	$6F,$CB,$9F,$A8,$FE,$8B,$4F,$F1,$35,$6B,$C6,$FF,$5F,$4F,$E8,$B0
	.dc.b	$FB,$DC,$4F,$D8,$5F,$98,$FF,$83,$7E,$97,$E6,$F9,$BF,$C9,$AB,$53
	.dc.b	$9E,$EF,$46,$9F,$9B,$E5,$FF,$66,$7E,$17,$FE,$CB,$FC,$CB,$FD,$8F
	.dc.b	$39,$BF,$F6,$4C,$E7,$F3,$2F,$E6,$3F,$83,$FF,$9E,$7F,$4B,$7E,$67
	.dc.b	$F2,$7F,$2B,$56,$BD,$C5,$6B,$B7,$FD,$31,$F4,$9F,$CE,$BF,$60,$7E
	.dc.b	$9E,$FC,$BF,$FD,$CF,$D1,$FE,$40,$AD,$4E,$7C,$2F,$8F,$F8,$CB,$EB
	.dc.b	$3F,$A1,$FE,$5F,$FE,$DA,$F2,$BF,$DA,$DF,$7F,$F9,$DE,$EF,$D2,$4E
	.dc.b	$7F,$35,$FF,$6C,$F4,$7E,$EF,$AD,$E1,$BE,$71,$FA,$F3,$F5,$E7,$EA
	.dc.b	$2F,$C6,$3F,$A8,$E1,$C2,$F7,$5A,$FB,$AF,$EA,$BF,$78,$FD,$9D,$FB
	.dc.b	$77,$1F,$8C,$FE,$D5,$FA,$CE,$A7,$39,$FF,$0C,$FB,$BF,$C3,$F0,$7F
	.dc.b	$99,$FE,$71,$F4,$9F,$95,$F8,$07,$D6,$BF,$05,$7D,$37,$F6,$4B,$DF
	.dc.b	$88,$BD,$D3,$EA,$3F,$B4,$FE,$77,$FE,$F8,$BF,$95,$9F,$F1,$D8,$C7
	.dc.b	$CE,$7F,$6E,$FD,$41,$F8,$EB,$E1,$5F,$B5,$BF,$45,$7E,$F1,$FE,$7D
	.dc.b	$F0,$FF,$89,$61,$C2,$F7,$4F,$D2,$3E,$56,$9F,$CA,$FF,$44,$FF,$28
	.dc.b	$7E,$DC,$F8,$17,$C9,$B2,$FE,$A7,$6B,$DD,$6A,$7E,$3C,$8F,$E5,$1E
	.dc.b	$4B,$F4,$E7,$CE,$FE,$BB,$F1,$9F,$E4,$F5,$FE,$93,$F9,$EF,$F1,$F5
	.dc.b	$6B,$D4,$FF,$29,$7C,$17,$F8,$3B,$F1,$B7,$E1,$FF,$3F,$F4,$37,$A7
	.dc.b	$FE,$95,$FB,$7D,$4E,$63,$FD,$31,$F2,$1F,$E1,$6F,$D4,$9F,$A1,$3F
	.dc.b	$1A,$FF,$2E,$E3,$ED,$DF,$53,$9F,$C7,$12,$F0,$7E,$3F,$6F,$F3,$DF
	.dc.b	$91,$E7,$87,$F5,$0F,$D3,$7A,$7C,$97,$F8,$92,$73,$F5,$29,$F0,$71
	.dc.b	$FF,$8A,$3E,$47,$D7,$FA,$1B,$F6,$6F,$BB,$F9,$6F,$F1,$1C,$FE,$55
	.dc.b	$6A,$73,$F1,$FF,$E3,$0F,$C6,$9D,$1F,$C9,$FF,$2B,$7E,$4B,$EE,$F5
	.dc.b	$35,$39,$8F,$83,$FA,$37,$F8,$E3,$F6,$FF,$F6,$AF,$E4,$9F,$ED,$8F
	.dc.b	$79,$F8,$07,$87,$75,$6B,$F3,$9F,$D4,$CF,$F2,$19,$FF,$65,$7D,$63
	.dc.b	$8D,$FB,$47,$F7,$4F,$97,$7E,$93,$FC,$4A,$FF,$BC,$5E,$EB,$53,$FC
	.dc.b	$7B,$B3,$FB,$47,$C1,$FE,$FF,$E5,$1F,$5A,$E7,$FD,$8F,$27,$F9,$42
	.dc.b	$B5,$39,$F8,$2F,$F1,$D6,$9F,$ED,$DF,$A8,$7E,$3C,$FD,$27,$FB,$4F
	.dc.b	$DA,$7E,$0B,$EA,$3F,$CB,$35,$A9,$CF,$93,$FD,$2B,$0F,$F7,$37,$EC
	.dc.b	$DF,$C2,$7C,$7F,$E1,$5F,$02,$FD,$F3,$F7,$55,$FE,$13,$AD,$7C,$67
	.dc.b	$ED,$A7,$F9,$EF,$27,$F8,$0F,$F2,$37,$EE,$6E,$3F,$C2,$7E,$CC,$EB
	.dc.b	$FF,$66,$E7,$FE,$1D,$BD,$D6,$A7,$F9,$54,$FF,$BE,$7F,$65,$FE,$47
	.dc.b	$E8,$BF,$21,$FF,$02,$7D,$2F,$A6,$BF,$65,$5A,$9C,$F2,$FF,$9E,$30
	.dc.b	$FE,$7D,$CC,$FD,$39,$FA,$9F,$FB,$07,$F2,$7D,$77,$FF,$84,$E7,$3F
	.dc.b	$00,$FB,$5E,$5B,$FF,$85,$BE,$53,$FD,$69,$FA,$73,$F5,$DC,$F8,$FF
	.dc.b	$C9,$3F,$2E,$4F,$1E,$3F,$F1,$D7,$E3,$DF,$E1,$CF,$C8,$BF,$94,$F8
	.dc.b	$3F,$63,$A3,$FD,$1B,$39,$CF,$F7,$3E,$27,$F0,$EC,$1F,$A1,$BF,$56
	.dc.b	$FE,$1B,$FA,$C3,$F5,$B7,$E4,$9F,$E9,$4B,$FF,$24,$5E,$FE,$35,$7B
	.dc.b	$AF,$E6,$3E,$1F,$F8,$E7,$B5,$FA,$B7,$F5,$DC,$1F,$58,$FC,$CF,$F0
	.dc.b	$37,$BA,$D4,$E7,$CB,$FF,$52,$FE,$CD,$FE,$05,$FD,$FF,$FC,$77,$FA
	.dc.b	$0F,$F8,$EF,$AF,$3F,$06,$9C,$FE,$1B,$F1,$73,$FF,$25,$7E,$F7,$FC
	.dc.b	$7B,$E1,$F5,$F7,$8C,$BF,$A7,$F2,$FF,$98,$EB,$58,$AB,$5F,$86,$FF
	.dc.b	$AD,$BF,$56,$FF,$5A,$FE,$63,$FC,$AB,$F3,$BF,$93,$F2,$99,$FB,$C4
	.dc.b	$E7,$9D,$BF,$FA,$FB,$F1,$E7,$EC,$AF,$C3,$BF,$D1,$DF,$47,$F9,$9F
	.dc.b	$5B,$53,$9F,$AB,$FE,$93,$EA,$FD,$1C,$7F,$CC,$99,$FF,$18,$FC,$C9
	.dc.b	$F9,$23,$F0,$2F,$EB,$5E,$B7,$FB,$32,$B5,$39,$FA,$AF,$89,$D3,$FD
	.dc.b	$34,$7E,$CD,$FD,$D3,$F3,$27,$A7,$FD,$13,$EA,$7F,$CA,$75,$BA,$D4
	.dc.b	$E7,$D7,$FF,$6C,$FE,$B4,$FD,$2D,$FB,$2B,$F6,$77,$E9,$CF,$DF,$BF
	.dc.b	$86,$E7,$FD,$77,$0E,$17,$BA,$D7,$C8,$7F,$B9,$3C,$E7,$E9,$EF,$CF
	.dc.b	$1F,$B1,$FF,$91,$BE,$77,$FD,$21,$9F,$E3,$AB,$53,$9F,$03,$FD,$DB
	.dc.b	$FA,$83,$F7,$D7,$ED,$1F,$E0,$FF,$03,$FB,$AF,$DF,$EA,$73,$F7,$FF
	.dc.b	$53,$0F,$F7,$AF,$EA,$3F,$D6,$61,$F8,$C7,$F8,$BB,$EF,$7F,$DA,$33
	.dc.b	$98,$C7,$D3,$FF,$BF,$3B,$BF,$86,$7D,$3F,$EF,$EF,$F9,$F7,$E0,$19
	.dc.b	$F9,$3F,$E9,$FD,$DF,$E1,$3F,$A3,$FD,$4F,$2F,$D5,$FC,$5B,$FD,$DB
	.dc.b	$FB,$4F,$F9,$93,$D8,$FF,$4D,$7B,$7F,$E0,$5F,$37,$FD,$4D,$5A,$3F
	.dc.b	$45,$EE,$7F,$28,$C3,$EF,$72,$BF,$56,$FF,$10,$7E,$D9,$FE,$80,$FD
	.dc.b	$E7,$FD,$D3,$39,$E1,$4E,$79,$3B,$61,$F8,$7F,$CB,$FF,$98,$D3,$F6
	.dc.b	$CF,$F3,$BF,$E7,$5F,$D4,$6B,$5F,$41,$FE,$2D,$9C,$FA,$7C,$A7,$FC
	.dc.b	$39,$83,$F4,$37,$E9,$CF,$CA,$BF,$95,$7E,$C9,$FA,$8B,$FF,$6B,$5E
	.dc.b	$EB,$5F,$3C,$AF,$C0,$FE,$B2,$FD,$6D,$FB,$93,$F5,$4F,$E4,$FF,$93
	.dc.b	$FF,$A8,$B3,$FF,$16,$5E,$EB,$5F,$40,$AF,$CA,$D7,$FB,$63,$EB,$DF
	.dc.b	$AE,$BF,$69,$7E,$BC,$FD,$CB,$FB,$F7,$E6,$5F,$EB,$AA,$D4,$E7,$C8
	.dc.b	$F8,$18,$BF,$49,$FF,$3A,$78,$AF,$DD,$FF,$94,$FE,$21,$F5,$EB,$FC
	.dc.b	$81,$5A,$9C,$FE,$33,$CB,$F3,$67,$ED,$2F,$63,$FA,$1F,$EB,$6F,$CB
	.dc.b	$FF,$C2,$15,$A9,$CC,$7F,$AB,$FF,$63,$FE,$01,$F2,$1F,$D6,$3B,$7F
	.dc.b	$A9,$F8,$5F,$8B,$9C,$FD,$9F,$F5,$7F,$B4,$7D,$A7,$F2,$BE,$8F,$ED
	.dc.b	$1E,$7F,$FB,$D3,$F7,$4F,$EA,$DF,$75,$FC,$56,$73,$F9,$AA,$73,$F9
	.dc.b	$77,$E5,$E7,$FB,$87,$E9,$8F,$DA,$7F,$9C,$7E,$E6,$FC,$9B,$FC,$FD
	.dc.b	$D7,$FC,$A2,$F7,$5A,$9F,$E5,$13,$93,$FB,$AF,$E3,$2F,$D9,$7F,$C1
	.dc.b	$9D,$9F,$EE,$1F,$CF,$7C,$15,$A9,$CF,$E5,$1F,$E2,$38,$E4,$3F,$02
	.dc.b	$FC,$57,$D7,$7F,$19,$FC,$EF,$FC,$FD,$E0,$F2,$AD,$7A,$1F,$ED,$7F
	.dc.b	$99,$57,$FA,$0F,$EC,$B5,$DF,$F3,$7A,$FF,$C0,$58,$76,$8C,$75,$FC
	.dc.b	$B8,$FF,$3F,$43,$DF,$7E,$00,$F1,$1F,$5C,$FD,$29,$F0,$DD,$AF,$75
	.dc.b	$A9,$FF,$13,$F5,$7D,$EF,$DE,$C7,$FA,$9F,$F3,$07,$E0,$2F,$4F,$CF
	.dc.b	$3A,$E5,$FF,$63,$4F,$F7,$8D,$FD,$1F,$4C,$FD,$FB,$F8,$2F,$CE,$7E
	.dc.b	$BF,$1F,$7C,$E7,$E2,$3F,$9B,$74,$FF,$AF,$30,$F8,$1F,$71,$FD,$AF
	.dc.b	$E6,$DF,$07,$EF,$F8,$EF,$8F,$39,$8F,$85,$F6,$5F,$3F,$EE,$BF,$09
	.dc.b	$FC,$00,$FD,$99,$F8,$47,$FB,$67,$F3,$0D,$79,$6A,$D4,$FF,$8C,$B5
	.dc.b	$E5,$FB,$57,$F0,$EF,$09,$FB,$0B,$F6,$BF,$F1,$75,$7E,$0B,$45,$EE
	.dc.b	$B5,$F5,$1E,$3F,$AB,$FF,$3E,$FD,$17,$F8,$67,$B7,$E5,$63,$C6,$73
	.dc.b	$8F,$D0,$FA,$9F,$56,$DF,$ED,$D8,$4B,$F7,$2F,$F6,$77,$83,$FD,$CE
	.dc.b	$7D,$59,$7F,$89,$6B,$DF,$C5,$F0,$E7,$FC,$50,$47,$FA,$17,$E3,$7F
	.dc.b	$DE,$FE,$F3,$FB,$E6,$9F,$DF,$7C,$8C,$BF,$EC,$B8,$FE,$E5,$7F,$EC
	.dc.b	$DF,$98,$FE,$65,$3A,$13,$A7,$FC,$DF,$F2,$FF,$92,$7F,$C3,$78,$77
	.dc.b	$AF,$DD,$47,$F8,$EF,$F0,$CF,$F1,$DC,$E7,$D9,$FF,$6D,$63,$FE,$77
	.dc.b	$F4,$3B,$65,$FC,$B3,$F1,$B5,$BF,$73,$7E,$EC,$FB,$0D,$FC,$97,$FA
	.dc.b	$B2,$B5,$3F,$8D,$F1,$38,$ED,$ED,$FC,$A9,$F9,$CF,$F8,$1F,$8B,$AE
	.dc.b	$BE,$09,$CF,$99,$FD,$47,$2F,$F5,$66,$3C,$B1,$FF,$45,$7C,$47,$F2
	.dc.b	$1F,$61,$F1,$5F,$17,$D9,$FD,$CF,$D4,$7F,$71,$AF,$E4,$FF,$8A,$7F
	.dc.b	$A0,$7C,$7F,$DF,$E4,$9C,$FC,$6F,$FA,$02,$FF,$BC,$7D,$96,$2F,$6D
	.dc.b	$CF,$FC,$B1,$F3,$9F,$CC,$BD,$9F,$E0,$7F,$03,$FD,$D3,$E1,$7F,$C0
	.dc.b	$13,$9F,$80,$4F,$F7,$8F,$97,$CF,$17,$F0,$CD,$BE,$5F,$EA,$6F,$C4
	.dc.b	$DF,$AB,$FF,$A9,$B8,$BF,$B3,$DE,$FB,$17,$BA,$FA,$53,$FE,$5A,$D3
	.dc.b	$FA,$2E,$AF,$9D,$FF,$2D,$FE,$77,$FC,$1F,$2E,$77,$F2,$AB,$5F,$81
	.dc.b	$A9,$F8,$7F,$9A,$BE,$7F,$D6,$7F,$BC,$3F,$08,$FF,$32,$E3,$9E,$BF
	.dc.b	$95,$39,$C7,$FA,$2C,$F6,$9F,$E5,$7D,$9B,$E2,$FB,$1D,$5C,$7A,$5E
	.dc.b	$3B,$F1,$27,$D0,$7E,$43,$2D,$4E,$63,$F2,$7F,$73,$C7,$E3,$FF,$93
	.dc.b	$FE,$EB,$FE,$0D,$FB,$0C,$7C,$5A,$3F,$91,$71,$FC,$53,$F9,$DE,$8F
	.dc.b	$DA,$9E,$7B,$F7,$FE,$1A,$F9,$67,$31,$CF,$E4,$BD,$1F,$9B,$36,$7E
	.dc.b	$A6,$F0,$FF,$C4,$9C,$5F,$07,$45,$EE,$B5,$7F,$37,$D4,$FF,$57,$7C
	.dc.b	$37,$FA,$BF,$DF,$FF,$58,$7A,$79,$9F,$EE,$07,$F2,$91,$FA,$A6,$6B
	.dc.b	$84,$3D,$CE,$1F,$0B,$FE,$B9,$FB,$1F,$F5,$D4,$62,$5F,$A4,$FF,$12
	.dc.b	$DF,$F0,$71,$FC,$6C,$E5,$CD,$FE,$91,$8C,$C7,$F2,$43,$E3,$9A,$73
	.dc.b	$A7,$FB,$1F,$E3,$BA,$79,$3F,$C9,$EB,$D1,$F0,$5E,$2C,$3F,$D5,$FE
	.dc.b	$BF,$F2,$37,$F8,$FC,$F8,$E3,$1F,$82,$FF,$2A,$9F,$E3,$99,$7F,$A3
	.dc.b	$5F,$E4,$C3,$E0,$7F,$DA,$DF,$01,$C6,$33,$18,$78,$DF,$7F,$7B,$FD
	.dc.b	$B9,$DD,$B9,$FF,$74,$46,$1D,$9D,$BD,$CF,$A6,$E0,$FF,$77,$38,$7F
	.dc.b	$CF,$D5,$EB,$8D,$FA,$98,$9F,$DF,$7B,$7F,$C4,$7D,$9F,$E2,$F1,$F0
	.dc.b	$8C,$FB,$7F,$C4,$5F,$F6,$9F,$C1,$7F,$B0,$7B,$BF,$8E,$FE,$A0,$FD
	.dc.b	$55,$E6,$BE,$FF,$F1,$F7,$BF,$41,$7B,$AD,$78,$5F,$E9,$3A,$FE,$35
	.dc.b	$57,$2F,$0F,$BF,$FE,$4A,$BF,$F4,$C7,$E9,$DA,$9F,$F7,$2B,$FC,$86
	.dc.b	$3F,$D3,$9F,$6D,$FE,$52,$CD,$FB,$DB,$F5,$6F,$F6,$6F,$C3,$EB,$1F
	.dc.b	$E3,$D5,$A9,$FF,$68,$FA,$59,$9F,$E7,$73,$FC,$9B,$F2,$AF,$F4,$EF
	.dc.b	$CD,$E7,$D9,$FE,$6D,$39,$F9,$AF,$E9,$9A,$78,$7E,$F1,$FB,$7B,$F6
	.dc.b	$37,$C7,$F1,$F7,$BE,$35,$6A,$7F,$8E,$EB,$FF,$00,$30,$FE,$8B,$E4
	.dc.b	$FF,$20,$FC,$7F,$F7,$53,$93,$EC,$C6,$7E,$17,$A7,$81,$F6,$FA,$6F
	.dc.b	$7F,$D9,$F2,$3F,$D3,$98,$7E,$03,$F7,$27,$EB,$9F,$EF,$1F,$6B,$8C
	.dc.b	$63,$F1,$1F,$F3,$3D,$F0,$BF,$EB,$B9,$3F,$5E,$7D,$87,$FB,$DB,$D9
	.dc.b	$F0,$C6,$3F,$11,$FE,$8D,$C5,$FD,$7B,$F1,$E7,$E3,$0F,$C7,$1C,$7E
	.dc.b	$89,$CC,$7F,$9B,$75,$F0,$76,$FF,$63,$D7,$F8,$6F,$AE,$FD,$CF,$F8
	.dc.b	$DE,$7F,$DA,$B6,$FF,$7E,$C7,$F3,$E8,$FD,$F7,$AF,$FC,$58,$F1,$F0
	.dc.b	$FF,$83,$1F,$E3,$A0,$FE,$AF,$97,$BD,$FF,$13,$FB,$DF,$E2,$8C,$3F
	.dc.b	$09,$1F,$E2,$AF,$7B,$FC,$57,$EF,$E2,$39,$46,$F0,$FA,$C7,$F8,$C2
	.dc.b	$3F,$8B,$1F,$C5,$71,$F2,$FC,$DF,$29,$DC,$FF,$8D,$BE,$7B,$FC,$67
	.dc.b	$18,$8F,$37,$EE,$58,$D3,$CD,$1F,$CB,$8F,$E4,$1F,$93,$7F,$19,$AF
	.dc.b	$EF,$07,$F2,$B7,$1E,$F7,$F9,$06,$14,$9E,$1E,$EE,$63,$13,$FC,$70
	.dc.b	$FE,$C6,$73,$31,$EE,$6E,$BC,$0F,$AE,$78,$CF,$F9,$10,$0F,$6F,$55
	.dc.b	$E7,$7E,$44,$F0,$F9,$CD,$D6,$F2,$FE,$6D,$3D,$39,$BF,$BC,$CF,$AB
	.dc.b	$0F,$E0,$4F,$B3,$F6,$7F,$E0,$D4,$72,$BF,$7F,$C9,$8F,$BD,$FF,$17
	.dc.b	$FE,$5E,$F2,$FF,$C3,$78,$7C,$69,$CC,$7F,$90,$1F,$F1,$AF,$E9,$CF
	.dc.b	$E0,$1F,$DE,$D6,$CF,$E5,$31,$9F,$80,$F8,$39,$2F,$FC,$7B,$D0,$FD
	.dc.b	$F7,$FB,$4E,$1C,$4F,$D7,$F9,$7F,$E8,$C8,$FF,$2A,$C6,$78,$3F,$DE
	.dc.b	$5F,$A8,$3F,$29,$FF,$5E,$CE,$7D,$27,$86,$C5,$7B,$F9,$3F,$1E,$9F
	.dc.b	$18,$F8,$58,$7F,$44,$FC,$31,$EB,$FF,$5B,$FC,$E7,$73,$98,$FE,$BD
	.dc.b	$C3,$88,$FE,$0B,$4F,$D1,$3F,$A1,$F8,$19,$E0,$FD,$D4,$E7,$E0,$7F
	.dc.b	$5F,$9F,$F9,$8A,$1F,$B0,$BF,$54,$7D,$DF,$1E,$3F,$5D,$6A,$7F,$27
	.dc.b	$CF,$F2,$F5,$F4,$7E,$7E,$FC,$93,$FD,$6B,$39,$9F,$2C,$E7,$EC,$FD
	.dc.b	$1E,$17,$ED,$7C,$3F,$F0,$86,$DF,$C2,$23,$FC,$3B,$F2,$DA,$7F,$D5
	.dc.b	$33,$F9,$AF,$8B,$C4,$7F,$09,$E7,$E3,$97,$F0,$D8,$FE,$06,$7F,$84
	.dc.b	$F0,$E9,$3C,$67,$EF,$4F,$F0,$5F,$2B,$8C,$7F,$07,$8F,$F0,$E6,$3D
	.dc.b	$9F,$F7,$66,$17,$97,$71,$FC,$37,$E5,$7F,$4B,$97,$A6,$7B,$6C,$F2
	.dc.b	$FF,$82,$65,$FC,$10,$FC,$9B,$83,$F5,$A6,$BE,$47,$8B,$C3,$B3,$FC
	.dc.b	$68,$F2,$F3,$37,$5F,$C1,$4F,$E0,$BA,$74,$71,$F9,$1F,$07,$5E,$6F
	.dc.b	$07,$F8,$21,$FC,$12,$3B,$BC,$F2,$E0,$E8,$FE,$05,$BF,$57,$11,$C1
	.dc.b	$FD,$3F,$C7,$7F,$AA,$B2,$FD,$FE,$9F,$C0,$8F,$E0,$3F,$94,$FF,$04
	.dc.b	$3F,$81,$FE,$AB,$FC,$18,$FE,$0B,$FB,$CF,$F0,$83,$F8,$3F,$E0,$FF
	.dc.b	$C2,$8F,$E1,$3F,$AD,$7F,$0C,$3F,$85,$FE,$6D,$FC,$38,$FE,$1B,$F8
	.dc.b	$F7,$F1,$03,$F8,$7F,$83,$FC,$48,$BA,$FF,$22,$F9,$F7,$EB,$D3,$F3
	.dc.b	$07,$ED,$CF,$D9,$FF,$3B,$87,$F2,$2B,$DD,$AF,$75,$FA,$6C,$3F,$9D
	.dc.b	$DF,$E6,$7B,$7F,$93,$B5,$FE,$13,$C8,$FC,$8F,$F0,$75,$F8,$67,$F9
	.dc.b	$29,$FC,$A3,$D5,$91,$FE,$32,$F9,$FF,$F1,$EB,$FD,$23,$E2,$FF,$96
	.dc.b	$9F,$CB,$3F,$2F,$FE,$5E,$7F,$2E,$FC,$DF,$F9,$89,$FC,$C3,$F5,$CF
	.dc.b	$E6,$67,$F3,$2F,$33,$F9,$EF,$C6,$AD,$F9,$03,$E0,$FF,$DC,$59,$7F
	.dc.b	$9F,$EF,$75,$A9,$FF,$80,$3B,$FF,$9F,$E7,$37,$C7,$D0,$FD,$0F,$DF
	.dc.b	$7F,$90,$FE,$83,$FC,$85,$F9,$8F,$F7,$E8,$C7,$D0,$7F,$90,$34,$FE
	.dc.b	$C7,$1F,$E3,$9F,$9C,$FF,$1C,$7C,$E7,$F3,$88,$C7,$CE,$7F,$8D,$B9
	.dc.b	$D1,$FD,$38,$FE,$9A,$7F,$A3,$7E,$43,$F8,$66,$1D,$0F,$F0,$83,$FE
	.dc.b	$03,$FC,$D7,$A4,$7F,$07,$F9,$0F,$F0,$97,$0F,$FB,$B6,$FF,$C3,$31
	.dc.b	$8F,$A2,$DD,$7F,$82,$E3,$FB,$3F,$C3,$7F,$66,$FD,$CD,$FA,$CB,$F6
	.dc.b	$A7,$E4,$8F,$CD,$5E,$37,$FA,$97,$0E,$17,$BA,$D4,$FE,$1B,$F5,$CF
	.dc.b	$EE,$3C,$6F,$DB,$E7,$E5,$0F,$97,$7E,$2B,$FD,$F7,$99,$87,$0B,$DD
	.dc.b	$6B,$F2,$2F,$EC,$3F,$96,$7F,$7C,$F8,$37,$EE,$6F,$B9,$FE,$B5,$F8
	.dc.b	$C7,$F1,$7F,$CF,$F7,$D1,$FE,$08,$BD,$D6,$BD,$CF,$C9,$FE,$DA,$FD
	.dc.b	$1F,$FC,$0F,$FA,$FB,$F6,$47,$C4,$3F,$0E,$7E,$B6,$FD,$0B,$3A,$B2
	.dc.b	$E4,$FF,$85,$B1,$E5,$87,$1B,$E1,$F4,$FF,$E1,$AB,$FF,$76,$DF,$F8
	.dc.b	$73,$F3,$D7,$D4,$7F,$BE,$7F,$21,$7A,$0F,$2F,$FC,$17,$87,$0C,$18
	.dc.b	$70,$FC,$1D,$87,$0B,$FF,$59,$DF,$1B,$FD,$97,$CD,$78,$67,$FC,$4B
	.dc.b	$F3,$7F,$E6,$5F,$A0,$7C,$6B,$DE,$75,$6F,$1F,$DD,$7D,$AD,$7F,$A7
	.dc.b	$7A,$3F,$EE,$1D,$3B,$46,$23,$FC,$1F,$1F,$E6,$58,$FE,$7C,$7F,$8D
	.dc.b	$07,$E7,$3F,$E2,$D7,$FC,$7B,$FA,$DB,$F6,$A7,$E3,$7F,$3F,$E0,$3E
	.dc.b	$97,$CE,$FF,$90,$B0,$E1,$83,$0E,$17,$BE,$DF,$4C,$F6,$FD,$9F,$FC
	.dc.b	$87,$F6,$BF,$E4,$9F,$51,$FA,$57,$F0,$9D,$DF,$31,$D6,$F7,$F6,$5F
	.dc.b	$A1,$AD,$7D,$D7,$F9,$4B,$F7,$5F,$F5,$0E,$7F,$E5,$4F,$D2,$5F,$30
	.dc.b	$FD,$C5,$FC,$00,$FD,$2F,$FC,$71,$CB,$C7,$F0,$0F,$F2,$AE,$1C,$2F
	.dc.b	$75,$FE,$3D,$FC,$AF,$B7,$D9,$7F,$96,$BD,$67,$D2,$FE,$1B,$F7,$C7
	.dc.b	$D0,$FF,$88,$3C,$1D,$F0,$7F,$99,$AF,$7E,$62,$F7,$F5,$EE,$0F,$F3
	.dc.b	$5F,$D1,$BF,$2A,$FF,$33,$FE,$B7,$FD,$27,$E1,$9F,$A6,$BF,$76,$FC
	.dc.b	$97,$C3,$FF,$AE,$70,$E1,$B7,$F9,$D3,$0E,$17,$BF,$3F,$FD,$AF,$FB
	.dc.b	$8B,$B7,$FA,$A5,$FF,$9E,$BF,$51,$FF,$3D,$FC,$7B,$F4,$97,$E4,$CF
	.dc.b	$D5,$9F,$A9,$FF,$09,$C7,$8E,$2C,$78,$E1,$C2,$F7,$E9,$BE,$75,$FA
	.dc.b	$57,$FA,$0B,$CC,$7E,$16,$FC,$DD,$FB,$47,$EE,$FF,$3A,$7E,$44,$FD
	.dc.b	$6F,$96,$3C,$70,$E1,$7B,$A7,$E9,$2F,$DD,$9F,$BB,$7F,$A4,$3F,$7A
	.dc.b	$7D,$63,$F1,$DF,$CB,$67,$F9,$0F,$C2,$7F,$DF,$B8,$70,$E6,$FF,$4E
	.dc.b	$E1,$C2,$F7,$EA,$FF,$9C,$3E,$95,$FA,$E3,$EC,$3F,$D3,$BF,$83,$BF
	.dc.b	$00,$7E,$95,$E4,$FC,$A5,$FA,$C3,$F6,$CF,$E7,$78,$F1,$DB,$FD,$57
	.dc.b	$8F,$1C,$38,$5E,$FF,$17,$EF,$9F,$6C,$FA,$F2,$B4,$EB,$E9,$FF,$AD
	.dc.b	$34,$E9,$A7,$F9,$C6,$F7,$5E,$23,$FD,$73,$F5,$CF,$D0,$BF,$D6,$BF
	.dc.b	$B5,$7F,$A9,$FF,$1A,$FC,$07,$E6,$AF,$D5,$DF,$24,$FC,$AF,$FB,$3E
	.dc.b	$1C,$3F,$19,$61,$C2,$F7,$FA,$06,$B6,$FF,$63,$E7,$D7,$CA,$69,$D7
	.dc.b	$4F,$1E,$7C,$F2,$E5,$F9,$D6,$31,$E3,$A7,$6C,$78,$E8,$FD,$29,$FB
	.dc.b	$9F,$7B,$6B,$DF,$5F,$55,$7F,$01,$E0,$F8,$15,$D3,$4F,$47,$E7,$AF
	.dc.b	$DA,$B9,$63,$C7,$34,$FD,$CC,$38,$E0,$FC,$59,$F9,$DF,$F8,$2B,$E2
	.dc.b	$BF,$DA,$DE,$8B,$C5,$FD,$86,$1C,$2F,$6F,$F1,$3F,$C8,$FF,$4D,$2F
	.dc.b	$DC,$7D,$55,$AC,$39,$FE,$5F,$F0,$F3,$F0,$67,$DF,$1F,$D7,$7E,$5D
	.dc.b	$9C,$38,$47,$EC,$7F,$55,$7E,$D9,$D2,$F8,$6C,$97,$EA,$19,$DF,$C9
	.dc.b	$2D,$1F,$B2,$E7,$6F,$1F,$D9,$DD,$3F,$7D,$F8,$2F,$97,$C1,$FF,$4B
	.dc.b	$7D,$03,$F3,$37,$EF,$AF,$C7,$EF,$EB,$9F,$5E,$F7,$5A,$9D,$FE,$13
	.dc.b	$9F,$BA,$FF,$00,$4E,$7C,$7E,$D1,$8C,$7A,$7D,$86,$33,$F0,$47,$7E
	.dc.b	$6F,$F1,$5F,$27,$D8,$D3,$BC,$E7,$E4,$FB,$5F,$0D,$5E,$47,$3D,$7C
	.dc.b	$9D,$FF,$DF,$B2,$E5,$AF,$86,$FF,$0B,$FA,$FF,$F9,$2B,$17,$74,$F8
	.dc.b	$FF,$F7,$E7,$E6,$0F,$D2,$BF,$DF,$5F,$BD,$FF,$83,$A7,$85,$6F,$F2
	.dc.b	$D5,$EF,$ED,$17,$FE,$14,$FC,$BB,$77,$FD,$12,$F7,$8F,$4F,$9E,$FD
	.dc.b	$EE,$7E,$2A,$D6,$5C,$F1,$7E,$4B,$A9,$F0,$4E,$7F,$05,$7E,$AD,$DB
	.dc.b	$6F,$4E,$7E,$89,$7E,$A5,$9D,$7C,$3B,$F9,$BE,$9B,$FD,$29,$C9,$FE
	.dc.b	$37,$D3,$AE,$3F,$02,$5F,$8F,$7A,$BF,$03,$FC,$8F,$C9,$3E,$F3,$F7
	.dc.b	$9F,$1B,$FE,$21,$F6,$B5,$E6,$AB,$5E,$C7,$F4,$CF,$96,$7E,$BF,$A9
	.dc.b	$CE,$1F,$E4,$4D,$FC,$5B,$B4,$EF,$A7,$6F,$2B,$E1,$4E,$74,$ED,$3E
	.dc.b	$2F,$CA,$FF,$CD,$18,$72,$C3,$8E,$BF,$D7,$27,$E1,$CE,$6F,$8C,$FF
	.dc.b	$60,$FC,$17,$F6,$33,$9D,$3C,$B0,$FC,$C3,$D1,$97,$C0,$FC,$17,$F9
	.dc.b	$AF,$DF,$7D,$9F,$C9,$5F,$AE,$7F,$75,$F9,$BF,$E9,$DF,$6F,$FE,$61
	.dc.b	$FC,$EF,$FD,$33,$7B,$AD,$7C,$5F,$F8,$8B,$F0,$A7,$E1,$BA,$E5,$FC
	.dc.b	$AE,$7D,$3E,$8F,$FB,$9D,$FF,$8A,$3E,$07,$C3,$5A,$C7,$9E,$8F,$D2
	.dc.b	$F3,$3F,$ED,$E9,$CF,$E0,$3F,$CB,$75,$F1,$65,$D3,$04,$BF,$72,$F1
	.dc.b	$61,$C6,$B5,$2F,$DC,$73,$AF,$97,$87,$DF,$3F,$E0,$6E,$76,$7F,$9D
	.dc.b	$FA,$47,$EE,$BF,$C5,$8F,$F2,$5F,$DD,$7E,$1F,$F2,$AE,$F3,$AF,$85
	.dc.b	$DF,$D0,$FF,$8E,$BB,$7F,$AF,$65,$D3,$83,$F7,$37,$BC,$B9,$FC,$5F
	.dc.b	$C0,$8C,$61,$FE,$42,$F8,$8C,$57,$86,$3F,$C8,$BE,$1B,$F5,$1F,$0C
	.dc.b	$E7,$6F,$C9,$7E,$25,$8D,$BC,$71,$F9,$5F,$C5,$71,$5E,$AC,$3F,$8C
	.dc.b	$F0,$FE,$EF,$F0,$FF,$ED,$FF,$95,$FF,$97,$3E,$E5,$FA,$BF,$C2,$FE
	.dc.b	$00,$F9,$B5,$AF,$C1,$D5,$AC,$9F,$A2,$6B,$E6,$D9,$73,$C9,$E4,$7D
	.dc.b	$EC,$B9,$63,$DB,$CB,$C6,$1C,$33,$EB,$83,$E3,$DF,$0D,$D7,$3E,$90
	.dc.b	$FC,$DB,$17,$C2,$38,$76,$7D,$1E,$CF,$F5,$4E,$5D,$74,$F6,$D3,$E4
	.dc.b	$7F,$DE,$FF,$77,$FC,$77,$EB,$3F,$A6,$FE,$1F,$FE,$E7,$FD,$99,$F2
	.dc.b	$9F,$E8,$1F,$DC,$B7,$A7,$F3,$4B,$DF,$E4,$CB,$DF,$E1,$7F,$EE,$DD
	.dc.b	$D9,$F5,$F9,$5F,$F5,$0D,$EF,$3F,$F1,$1E,$5D,$0F,$F1,$E4,$EB,$1F
	.dc.b	$0C,$FF,$8B,$AF,$96,$9E,$58,$FE,$C7,$F9,$C7,$1D,$3C,$77,$E8,$C3
	.dc.b	$FD,$43,$9F,$8B,$2F,$44,$75,$C3,$FB,$5F,$E1,$BC,$63,$07,$FB,$EB
	.dc.b	$F1,$07,$E7,$FF,$D0,$7F,$26,$FF,$5A,$7E,$F1,$F7,$BF,$4A,$FF,$3C
	.dc.b	$7C,$6F,$FB,$AE,$F7,$9F,$F5,$8B,$DF,$5A,$F7,$E2,$2F,$76,$BD,$D7
	.dc.b	$FA,$0F,$BF,$FB,$0D,$6A,$73,$F8,$1B,$E1,$73,$AF,$97,$F1,$86,$DE
	.dc.b	$5A,$7E,$49,$FE,$9F,$C7,$95,$6B,$6F,$25,$6B,$F4,$5F,$26,$3C,$F1
	.dc.b	$E5,$F1,$7F,$E3,$4C,$78,$C6,$3E,$8F,$D5,$F1,$9F,$D9,$7F,$45,$7E
	.dc.b	$F5,$FE,$3E,$FA,$0F,$E1,$7F,$A0,$BE,$3F,$E5,$E5,$FD,$3B,$9B,$FB
	.dc.b	$EE,$1C,$2F,$FC,$8B,$E8,$30,$E1,$F8,$D7,$6B,$DE,$5D,$6B,$5C,$FF
	.dc.b	$1B,$4F,$14,$E7,$0F,$8D,$7C,$46,$5D,$F2,$F2,$E9,$F1,$B4,$F2,$56
	.dc.b	$B0,$E5,$3F,$DD,$A7,$37,$C7,$3F,$0F,$EA,$9D,$63,$19,$34,$7E,$DF
	.dc.b	$9C,$BC,$9E,$97,$9F,$4F,$F0,$EF,$E4,$BF,$E0,$2F,$D1,$9F,$8C,$BF
	.dc.b	$67,$FD,$F7,$C5,$BF,$37,$F5,$5E,$EB,$5E,$AF,$FB,$DB,$F1,$5C,$F8
	.dc.b	$AC,$B9,$E5,$FC,$1F,$F4,$3F,$06,$8C,$B9,$4F,$F0,$BF,$9C,$CC,$63
	.dc.b	$27,$E8,$6F,$C9,$BA,$E5,$D2,$73,$D9,$8D,$BC,$7F,$A1,$BF,$4B,$C6
	.dc.b	$9F,$C5,$E3,$F4,$5D,$7F,$3B,$D3,$FF,$6A,$FE,$D2,$FC,$DF,$EA,$FE
	.dc.b	$43,$FE,$A7,$F8,$7F,$B1,$5A,$9C,$FE,$00,$FD,$03,$3F,$8C,$31,$E5
	.dc.b	$E0,$B2,$7E,$19,$F0,$32,$E5,$39,$FD,$1B,$18,$71,$CF,$C3,$F2,$AF
	.dc.b	$D0,$BA,$E7,$D3,$34,$BF,$42,$CD,$FA,$3D,$D7,$E8,$D8,$D3,$C5,$0E
	.dc.b	$3F,$E0,$1F,$57,$FE,$4B,$FA,$67,$96,$FD,$C9,$FB,$4B,$E4,$3F,$7D
	.dc.b	$F3,$CF,$D6,$F8,$F8,$DF,$E6,$4C,$78,$E1,$C2,$F7,$F9,$16,$A9,$5A
	.dc.b	$C7,$A7,$BE,$FC,$B3,$F8,$6E,$BE,$3D,$7C,$55,$AF,$CA,$F3,$8F,$79
	.dc.b	$CE,$7F,$C4,$70,$FD,$DE,$1C,$74,$F1,$65,$D7,$D8,$F1,$69,$E1,$C3
	.dc.b	$F8,$C7,$C3,$F4,$CF,$CB,$9F,$4C,$3C,$9F,$87,$70,$8C,$68,$FD,$BF
	.dc.b	$D9,$A3,$F7,$5F,$F5,$9F,$DA,$BE,$A7,$FD,$6B,$FB,$8B,$F2,$9D,$D6
	.dc.b	$A7,$3F,$A7,$63,$F1,$16,$1C,$57,$E6,$BE,$5C,$BB,$46,$AF,$FD,$83
	.dc.b	$F9,$5E,$71,$E9,$2C,$FF,$C6,$BE,$6B,$9F,$FD,$11,$8F,$2C,$FB,$53
	.dc.b	$ED,$BC,$51,$E1,$FA,$7F,$F1,$E6,$1C,$79,$7F,$D4,$38,$71,$D7,$FD
	.dc.b	$A3,$4F,$DC,$FF,$A5,$7E,$FB,$FE,$95,$F4,$1F,$A6,$BF,$10,$7E,$4F
	.dc.b	$FC,$2F,$DE,$FC,$AF,$E8,$7C,$35,$FF,$56,$E1,$C2,$F7,$EB,$FF,$A3
	.dc.b	$3C,$B7,$E4,$3B,$B5,$EF,$2E,$99,$7F,$56,$D3,$C5,$A7,$86,$7C,$1F
	.dc.b	$03,$9C,$3B,$CE,$73,$ED,$F2,$7D,$72,$E9,$1E,$1F,$C4,$F1,$5B,$CD
	.dc.b	$FA,$7F,$EB,$B9,$5A,$F8,$38,$3D,$35,$AF,$3B,$D1,$F8,$1F,$F8,$7F
	.dc.b	$EC,$1F,$B7,$3F,$3B,$7E,$CB,$FC,$9F,$E4,$37,$F5,$4B,$DD,$6B,$E2
	.dc.b	$3F,$6F,$B7,$83,$97,$FD,$AB,$9F,$5C,$FA,$67,$FE,$E0,$9E,$78,$F2
	.dc.b	$8F,$A7,$18,$9D,$E1,$FE,$AB,$C3,$A4,$B0,$E9,$8F,$87,$F2,$FE,$0F
	.dc.b	$A3,$F7,$97,$EF,$D8,$CD,$0F,$C6,$3E,$C7,$E2,$3F,$D0,$FF,$40,$7E
	.dc.b	$82,$FD,$B5,$F1,$8F,$C5,$1F,$2D,$C3,$89,$87,$0B,$DD,$3E,$2F,$53
	.dc.b	$9C,$3F,$D4,$BC,$58,$CB,$B4,$7F,$70,$8C,$61,$CB,$EA,$2C,$F8,$7F
	.dc.b	$76,$BF,$8D,$BF,$4E,$B7,$FE,$27,$FC,$D3,$FD,$E5,$F8,$03,$F2,$9F
	.dc.b	$E4,$9C,$FF,$E6,$6F,$E0,$17,$DF,$7F,$96,$3B,$18,$70,$BD,$FD,$FA
	.dc.b	$F7,$D5,$D7,$84,$7F,$B7,$69,$F1,$5F,$F0,$06,$32,$6C,$D7,$C5,$E5
	.dc.b	$FF,$1B,$E3,$DE,$3F,$1D,$8F,$7F,$86,$CD,$76,$DB,$D5,$5A,$CF,$F9
	.dc.b	$A4,$FF,$91,$7F,$06,$EF,$FA,$F3,$F2,$77,$EB,$2F,$C7,$9F,$AD,$E7
	.dc.b	$F5,$07,$E4,$1C,$72,$F5,$E3,$C7,$0E,$17,$BF,$D7,$75,$4A,$ED,$F0
	.dc.b	$FE,$B8,$CC,$71,$7D,$DC,$CD,$7C,$BA,$F9,$35,$F1,$AD,$7F,$CB,$33
	.dc.b	$AD,$3C,$32,$FD,$23,$FC,$BB,$E7,$7C,$CD,$FF,$41,$F7,$7F,$D8,$7F
	.dc.b	$1F,$FF,$60,$7F,$00,$BD,$CF,$F3,$45,$EF,$D0,$FF,$2E,$FC,$A2,$F7
	.dc.b	$ED,$EB,$8B,$ED,$E7,$FA,$BF,$B9,$F6,$71,$FF,$A8,$67,$38,$FF,$99
	.dc.b	$9F,$6E,$5D,$75,$F1,$7E,$21,$F5,$2F,$85,$7C,$EB,$F0,$E7,$FE,$D5
	.dc.b	$F9,$9F,$F3,$BF,$E4,$FF,$86,$FD,$D5,$F7,$4F,$CC,$9F,$BD,$7D,$BE
	.dc.b	$B7,$17,$F7,$FD,$58,$F1,$C3,$85,$EE,$BF,$8A,$56,$B5,$FC,$BE,$5D
	.dc.b	$BF,$32,$F6,$E3,$C7,$3B,$5A,$CB,$9F,$47,$DF,$F9,$DF,$3C,$E6,$BC
	.dc.b	$B3,$F8,$3E,$BC,$CB,$2F,$5D,$6B,$6F,$4F,$DD,$FF,$A6,$F3,$F0,$6F
	.dc.b	$E7,$FD,$87,$18,$F5,$8F,$57,$E2,$78,$8E,$9B,$79,$31,$F8,$7F,$B2
	.dc.b	$67,$81,$F9,$5B,$F0,$E7,$7B,$F3,$07,$EA,$BF,$EE,$3F,$C7,$78,$5E
	.dc.b	$EB,$53,$FE,$E7,$FD,$77,$38,$7C,$4D,$59,$76,$FA,$1C,$3E,$09,$EF
	.dc.b	$9F,$6F,$93,$F8,$73,$9C,$BA,$FE,$4F,$8D,$3C,$17,$E4,$FB,$CF,$F2
	.dc.b	$E7,$67,$D7,$D3,$BD,$6B,$6F,$F7,$D7,$C6,$7C,$3F,$C1,$5F,$AE,$3F
	.dc.b	$59,$FE,$33,$FB,$D7,$D1,$FC,$93,$F8,$0E,$7C,$EF,$75,$AE,$F5,$3D
	.dc.b	$2E,$98,$BF,$15,$F8,$18,$70,$DB,$F7,$7D,$FF,$C7,$23,$9F,$B6,$D7
	.dc.b	$F0,$1F,$95,$7E,$0A,$73,$1F,$83,$FC,$AB,$19,$79,$EF,$86,$BE,$2F
	.dc.b	$C9,$3C,$3B,$F9,$77,$7E,$6B,$F8,$1C,$3F,$D3,$FE,$0F,$F1,$CD,$DF
	.dc.b	$A4,$7E,$93,$F5,$CD,$7E,$0F,$FA,$AF,$67,$F9,$AC,$FE,$A3,$F7,$4C
	.dc.b	$E1,$F6,$F1,$FF,$54,$FE,$A7,$70,$E1,$97,$4F,$3F,$F3,$75,$E0,$CB
	.dc.b	$9F,$9D,$8B,$DE,$7F,$B4,$F9,$8E,$B5,$E0,$F4,$7F,$D5,$9D,$9F,$6F
	.dc.b	$1E,$59,$FA,$25,$F9,$67,$08,$C7,$EC,$3F,$EF,$FF,$6F,$FA,$4F,$E9
	.dc.b	$FF,$EC,$5F,$A7,$7E,$85,$F9,$CE,$B7,$C9,$FF,$BB,$EF,$7D,$FF,$E3
	.dc.b	$0A,$D4,$BF,$36,$E9,$9F,$F5,$8C,$BA,$3F,$CD,$B0,$E3,$F8,$37,$F2
	.dc.b	$07,$9E,$7E,$0F,$33,$FB,$2F,$89,$FD,$83,$8B,$FA,$4E,$1C,$F5,$F0
	.dc.b	$57,$F6,$AF,$C5,$B5,$8F,$2D,$BC,$7F,$C0,$3E,$2C,$1A,$FB,$67,$CF
	.dc.b	$F3,$D9,$FA,$E7,$6B,$F5,$1F,$8F,$FB,$43,$F7,$47,$EE,$CF,$6B,$85
	.dc.b	$EF,$A9,$7B,$A7,$B5,$AF,$DE,$FD,$DB,$FF,$7A,$D3,$BE,$98,$8F,$F2
	.dc.b	$DE,$5C,$F1,$F0,$57,$B2,$78,$4E,$74,$FF,$0C,$72,$FC,$49,$E3,$87
	.dc.b	$A2,$7C,$FF,$A7,$F9,$A3,$1A,$7F,$B4,$6F,$87,$E1,$2F,$D7,$5F,$B8
	.dc.b	$3F,$37,$7F,$00,$FF,$8C,$3F,$48,$7C,$07,$CB,$C3,$85,$EE,$B5,$FA
	.dc.b	$DE,$71,$E7,$BF,$DF,$3F,$C3,$37,$DF,$DC,$F8,$0F,$2C,$FB,$FD,$6F
	.dc.b	$6A,$E5,$97,$58,$7C,$0F,$C3,$9F,$F6,$9E,$3E,$8B,$E3,$AF,$FB,$97
	.dc.b	$6F,$6C,$E7,$5F,$57,$E2,$9D,$2F,$7F,$9F,$59,$F7,$7C,$77,$E1,$FE
	.dc.b	$47,$FC,$A1,$F9,$23,$F3,$27,$E4,$FF,$EF,$9F,$D6,$9F,$BB,$6F,$ED
	.dc.b	$17,$BA,$D7,$A9,$9F,$C9,$7E,$F6,$5D,$38,$FE,$21,$FE,$6B,$C3,$8E
	.dc.b	$7D,$BE,$7B,$FC,$77,$C7,$F7,$6F,$C7,$3E,$2F,$C7,$3F,$A8,$E1,$E1
	.dc.b	$D3,$C1,$0F,$D8,$3E,$35,$F0,$D7,$D1,$1F,$A3,$F8,$47,$9A,$F4,$5F
	.dc.b	$82,$3F,$78,$7E,$B1,$FE,$B7,$E4,$E1,$7B,$AD,$4F,$FA,$EF,$93,$3F
	.dc.b	$BD,$7E,$6F,$2E,$B9,$5C,$38,$65,$D3,$E4,$7E,$D5,$6B,$2E,$7E,$BE
	.dc.b	$2F,$79,$FF,$6E,$FD,$EB,$C3,$97,$9B,$27,$EF,$BF,$1B,$D7,$FF,$30
	.dc.b	$FA,$4F,$CE,$DF,$AB,$FB,$7F,$4D,$FF,$7A,$7E,$66,$FC,$BF,$F8,$25
	.dc.b	$EE,$B5,$F8,$DA,$9F,$97,$EB,$17,$EA,$1F,$B7,$C3,$87,$77,$DA,$79
	.dc.b	$6E,$D1,$97,$CD,$AE,$59,$F6,$F8,$EF,$CD,$61,$C2,$FC,$BF,$AE,$6B
	.dc.b	$1E,$58,$F8,$FE,$EF,$15,$EA,$FC,$02,$CF,$9F,$F7,$3F,$E3,$DF,$84
	.dc.b	$7E,$4B,$F5,$3F,$E3,$1F,$0D,$FE,$D4,$FD,$0F,$FD,$5B,$E4,$FF,$07
	.dc.b	$AD,$7B,$FF,$C9,$3F,$14,$57,$F9,$96,$B5,$E5,$3F,$35,$FE,$2D,$87
	.dc.b	$5E,$5F,$CC,$BC,$B2,$FF,$70,$FE,$73,$FB,$58,$C6,$4F,$C0,$71,$8B
	.dc.b	$5F,$2C,$F8,$38,$BD,$33,$9D,$BC,$9B,$39,$3F,$9D,$7D,$67,$F7,$2F
	.dc.b	$35,$F3,$2F,$D1,$9F,$B8,$FF,$97,$3F,$2B,$DF,$CE,$7F,$B5,$EB,$53
	.dc.b	$FE,$60,$FC,$A7,$3F,$67,$FC,$C7,$2E,$9C,$DF,$B4,$7B,$5F,$0E,$5F
	.dc.b	$E2,$11,$8C,$FB,$FD,$76,$33,$ED,$1E,$2F,$A1,$FF,$5A,$F5,$7F,$35
	.dc.b	$BE,$1A,$FF,$9C,$3F,$04,$4B,$F0,$2F,$F5,$CE,$8F,$D0,$DF,$8F,$BF
	.dc.b	$58,$7E,$F7,$FE,$C8,$FC,$61,$F4,$9C,$B1,$E3,$87,$0B,$DF,$D2,$2F
	.dc.b	$FC,$03,$9F,$7E,$4F,$F6,$5E,$9D,$F4,$FF,$48,$E5,$CB,$E6,$6D,$6B
	.dc.b	$3E,$9A,$FF,$B6,$3E,$A3,$FB,$A5,$F8,$30,$F3,$4F,$FB,$73,$E7,$73
	.dc.b	$87,$19,$CE,$BE,$3F,$D1,$FD,$13,$9D,$FC,$FB,$FF,$2F,$DF,$CD,$3F
	.dc.b	$7B,$5F,$0E,$1E,$BF,$98,$E3,$39,$CB,$AE,$1F,$DC,$B0,$E3,$E8,$3F
	.dc.b	$4F,$7E,$86,$F9,$77,$EA,$8F,$39,$FA,$CB,$E4,$BA,$E3,$C7,$0E,$17
	.dc.b	$BA,$FF,$70,$65,$DB,$27,$E7,$9F,$23,$6F,$E2,$31,$E0,$7F,$DA,$BC
	.dc.b	$DF,$EF,$6C,$BA,$6B,$E2,$D5,$FA,$77,$C8,$BE,$3A,$F8,$25,$F7,$F9
	.dc.b	$C3,$8E,$DE,$3F,$CB,$FC,$FA,$7B,$B3,$E9,$BB,$F2,$BF,$8F,$F9,$D7
	.dc.b	$F1,$D1,$FA,$63,$F1,$D6,$2F,$DD,$3F,$6D,$C9,$F4,$AF,$85,$EE,$B5
	.dc.b	$FA,$9F,$4F,$84,$7E,$3B,$F0,$6F,$B7,$27,$C0,$01,$E7,$9F,$96,$3E
	.dc.b	$04,$63,$0F,$EB,$B9,$78,$70,$E1,$97,$87,$37,$E3,$3E,$1C,$3D,$79
	.dc.b	$7A,$A7,$F4,$18,$7E,$2B,$83,$FD,$21,$E3,$BF,$38,$7E,$58,$F2,$DF
	.dc.b	$B9,$3F,$2F,$FC,$6E,$1C,$2F,$75,$AF,$DC,$33,$F8,$4B,$2F,$C0,$5F
	.dc.b	$FA,$77,$F2,$0B,$39,$C3,$8F,$A5,$F7,$6B,$A6,$9E,$1F,$D6,$31,$7B
	.dc.b	$8F,$26,$BF,$B8,$F0,$BE,$4F,$3F,$E6,$EB,$5B,$7A,$3D,$1F,$1E,$DE
	.dc.b	$DC,$78,$FC,$66,$36,$F4,$E7,$FD,$93,$8B,$F5,$BF,$25,$F6,$3F,$51
	.dc.b	$FE,$57,$FB,$AB,$F1,$E7,$EB,$2F,$D1,$7F,$5F,$7B,$F6,$17,$BA,$FF
	.dc.b	$33,$63,$D3,$C1,$FE,$3B,$C7,$F8,$13,$FC,$18,$B3,$FD,$36,$3C,$F0
	.dc.b	$F0,$7E,$1B,$FA,$18,$C6,$9F,$CF,$70,$F0,$E1,$C7,$3F,$0C,$F9,$B9
	.dc.b	$F3,$97,$AA,$3D,$3F,$AB,$FF,$06,$FC,$E5,$FB,$FB,$F3,$87,$ED,$3F
	.dc.b	$ED,$4F,$A7,$7E,$B0,$FC,$9B,$85,$EE,$B5,$2F,$C6,$33,$FA,$57,$EF
	.dc.b	$F2,$EB,$95,$C3,$87,$E2,$9F,$E4,$6F,$27,$F7,$95,$AC,$BC,$1F,$82
	.dc.b	$62,$F7,$AF,$8B,$F2,$0F,$0E,$5E,$6C,$BF,$C9,$FA,$BF,$1E,$FE,$43
	.dc.b	$FA,$53,$F3,$0F,$DD,$7E,$C2,$FD,$6D,$FB,$83,$F6,$77,$C9,$F0,$BD
	.dc.b	$D6,$A5,$F2,$D9,$FC,$EB,$F8,$D6,$5D,$7D,$5B,$87,$0C,$BF,$CE,$7E
	.dc.b	$07,$39,$69,$E1,$9F,$F3,$D7,$89,$FC,$EB,$0E,$53,$9C,$DE,$1F,$18
	.dc.b	$C7,$D9,$DD,$BC,$9F,$84,$5C,$BD,$5F,$84,$7F,$40,$FC,$B7,$E1,$7E
	.dc.b	$8F,$FE,$F4,$FD,$B9,$F9,$A7,$F8,$37,$F0,$1D,$73,$AB,$53,$FE,$11
	.dc.b	$FC,$73,$31,$FC,$DF,$0F,$99,$CF,$6F,$87,$CE,$32,$EB,$F9,$2F,$F3
	.dc.b	$5B,$F0,$CE,$72,$FF,$62,$73,$F4,$8C,$7D,$43,$F2,$BC,$63,$E6,$87
	.dc.b	$DF,$FF,$5A,$FD,$D9,$F9,$AF,$F8,$6F,$E0,$BF,$D9,$7F,$86,$7F,$17
	.dc.b	$FC,$13,$53,$9F,$A2,$FF,$5B,$7B,$8F,$C2,$B3,$F8,$A7,$F5,$1C,$B9
	.dc.b	$FC,$3D,$C3,$86,$5F,$C8,$78,$79,$CB,$4E,$FF,$63,$FD,$47,$0E,$31
	.dc.b	$E2,$FD,$FD,$F9,$5F,$08,$C7,$E6,$FF,$E1,$AF,$D2,$3F,$8A,$FE,$A7
	.dc.b	$FE,$CB,$FD,$49,$FB,$D7,$F9,$AF,$8D,$F9,$1B,$F0,$2E,$1C,$1F,$D1
	.dc.b	$30,$E1,$7B,$AF,$D9,$79,$AC,$BA,$FE,$2E,$BF,$E4,$78,$7F,$A9,$A7
	.dc.b	$39,$E2,$3F,$A7,$DF,$17,$B5,$A7,$CB,$F3,$1F,$EF,$8E,$8D,$E1,$CB
	.dc.b	$3F,$47,$E6,$B8,$C7,$8E,$1C,$36,$F5,$FE,$07,$EA,$9F,$26,$7D,$70
	.dc.b	$F5,$76,$67,$6F,$6E,$AC,$7E,$1F,$EB,$99,$C3,$84,$E7,$81,$F9,$2F
	.dc.b	$5F,$63,$FD,$6D,$F9,$CB,$D5,$7E,$9C,$FD,$61,$F9,$6B,$E2,$78,$71
	.dc.b	$B0,$E1,$7B,$AF,$F1,$27,$47,$C5,$E4,$FE,$43,$6F,$CF,$3E,$54,$72
	.dc.b	$C2,$D6,$F4,$F1,$E9,$E2,$D1,$F3,$5F,$1F,$0E,$78,$78,$25,$EF,$67
	.dc.b	$0E,$3A,$F8,$FC,$3E,$D9,$73,$D9,$E3,$7C,$DF,$D4,$DF,$93,$3F,$5E
	.dc.b	$7E,$5F,$FE,$DF,$E6,$FE,$23,$F9,$36,$EB,$53,$9F,$7D,$FA,$D6,$70
	.dc.b	$F5,$F8,$7F,$DB,$32,$EB,$F9,$3E,$39,$7F,$C2,$F1,$E0,$AE,$FE,$87
	.dc.b	$E6,$BC,$7F,$74,$E7,$3F,$17,$5E,$35,$F0,$DF,$E8,$35,$7E,$B7,$E1
	.dc.b	$BF,$27,$87,$F8,$9F,$DF,$BF,$81,$ED,$F1,$BE,$61,$FB,$53,$F8,$03
	.dc.b	$F0,$1E,$07,$D7,$BD,$F6,$FF,$A8,$AB,$5F,$01,$CF,$C9,$FE,$6B,$CB
	.dc.b	$FA,$6C,$38,$47,$F4,$38,$FD,$34,$63,$3F,$05,$BF,$07,$FB,$59,$76
	.dc.b	$D3,$C7,$EA,$63,$26,$1C,$35,$72,$78,$63,$1E,$45,$DF,$CD,$4F,$C8
	.dc.b	$F1,$97,$BA,$1F,$B1,$FE,$AF,$F7,$9F,$F5,$07,$C9,$7F,$58,$FD,$B7
	.dc.b	$FD,$F7,$C2,$FD,$73,$5F,$82,$AB,$5F,$75,$F2,$3F,$46,$CF,$C4,$7D
	.dc.b	$BE,$86,$7D,$1F,$83,$E1,$FF,$8B,$23,$BF,$CA,$B5,$62,$FD,$23,$F3
	.dc.b	$95,$A9,$CF,$E0,$18,$D7,$C5,$87,$1C,$3E,$CF,$EE,$5E,$6D,$BC,$BC
	.dc.b	$1F,$D2,$FB,$7F,$61,$FA,$D3,$E4,$FF,$BC,$7E,$42,$FC,$6F,$FD,$C1
	.dc.b	$FB,$E7,$F8,$5E,$F7,$5A,$9C,$FE,$B1,$FC,$47,$0E,$5F,$31,$EF,$7A
	.dc.b	$7D,$E5,$BE,$1F,$8B,$FF,$A2,$B9,$F8,$46,$38,$FF,$71,$7C,$B5,$F1
	.dc.b	$5B,$5F,$F4,$36,$BE,$2C,$D5,$A9,$FD,$EC,$FF,$0B,$F0,$FF,$8B,$7D
	.dc.b	$5B,$CD,$7E,$7D,$FD,$3F,$F0,$1D,$F6,$7F,$97,$2F,$75,$AF,$8B,$53
	.dc.b	$F5,$5F,$E5,$D5,$FC,$87,$F2,$AF,$B8,$F6,$E3,$FB,$67,$FA,$39,$E7
	.dc.b	$97,$F5,$DF,$9A,$FE,$29,$18,$C7,$FB,$BE,$BE,$7F,$77,$FC,$C1,$A7
	.dc.b	$87,$4F,$E5,$73,$9D,$BD,$59,$7A,$7E,$0F,$C7,$87,$08,$F6,$7E,$9B
	.dc.b	$8A,$DE,$5E,$AF,$D2,$FF,$A3,$7D,$4F,$E6,$7D,$D7,$E3,$BF,$E0,$0F
	.dc.b	$A0,$FC,$AF,$E3,$3A,$F3,$3F,$8B,$CE,$7C,$97,$AB,$9C,$3F,$39,$8F
	.dc.b	$E8,$FE,$06,$E1,$C3,$E6,$7F,$C4,$1C,$FF,$CC,$EB,$C1,$97,$7F,$A6
	.dc.b	$C5,$EF,$5F,$D2,$FB,$5D,$6B,$C1,$FB,$63,$9D,$F9,$97,$FA,$4B,$F4
	.dc.b	$2F,$E2,$9E,$83,$F3,$2F,$F2,$A7,$F0,$07,$F7,$7F,$7B,$85,$EF,$D9
	.dc.b	$5E,$FD,$75,$FF,$7D,$F7,$77,$E2,$7F,$B1,$B3,$FE,$41,$F3,$FF,$D4
	.dc.b	$32,$E5,$8D,$AD,$65,$D3,$4F,$0F,$E6,$7F,$E0,$0B,$DC,$79,$3F,$34
	.dc.b	$C5,$F0,$C3,$C5,$A7,$F8,$37,$5F,$0C,$B4,$F6,$E1,$EC,$F2,$BF,$87
	.dc.b	$70,$7C,$2A,$FF,$6D,$FD,$A3,$E8,$BF,$B7,$7C,$13,$F5,$67,$ED,$3F
	.dc.b	$91,$C3,$86,$0C,$38,$7C,$BF,$F6,$4B,$DF,$8D,$FC,$5A,$F7,$5A,$BF
	.dc.b	$E6,$FC,$5F,$EC,$28,$F0,$7C,$C7,$C3,$9E,$9A,$78,$B1,$F0,$FE,$17
	.dc.b	$F0,$27,$39,$F7,$D1,$FB,$5E,$72,$F1,$ED,$E9,$9F,$F2,$27,$DD,$7C
	.dc.b	$49,$CE,$5E,$6F,$B1,$C6,$9D,$EF,$FC,$95,$F9,$8E,$33,$ED,$C2,$FC
	.dc.b	$8B,$E6,$FE,$D8,$FC,$A3,$FD,$DD,$F8,$BB,$F4,$17,$E5,$6F,$A0,$7E
	.dc.b	$F9,$C2,$F7,$5A,$97,$E8,$59,$FC,$5F,$E1,$65,$D7,$8A,$E1,$C3,$2F
	.dc.b	$F6,$57,$5F,$D2,$AF,$15,$B4,$F1,$57,$FA,$53,$A3,$A4,$63,$37,$EE
	.dc.b	$D8,$D7,$C4,$7E,$23,$E4,$3F,$DD,$1F,$A6,$BF,$03,$7E,$5D,$FB,$EF
	.dc.b	$D5,$BF,$C2,$3B,$FE,$0A,$F7,$5A,$F3,$D4,$FC,$97,$A7,$BD,$FE,$01
	.dc.b	$DF,$F0,$20,$FF,$5D,$AC,$FF,$05,$8F,$4C,$FF,$86,$FA,$BF,$81,$8C
	.dc.b	$64,$F5,$F1,$87,$1D,$3C,$5A,$3F,$19,$F0,$E7,$EB,$CF,$D5,$F2,$5F
	.dc.b	$A6,$FD,$4D,$E9,$FF,$3D,$FC,$25,$F8,$47,$FB,$A7,$F7,$C7,$E1,$BB
	.dc.b	$AD,$4E,$7F,$13,$C7,$C6,$7E,$67,$17,$EE,$3F,$E5,$0E,$6C,$E7,$D3
	.dc.b	$F1,$2F,$F3,$1C,$63,$1F,$F7,$B7,$B5,$F9,$DC,$3C,$78,$FF,$BF,$67
	.dc.b	$C5,$FA,$26,$F0,$6C,$F0,$B0,$FF,$A8,$7B,$3F,$6B,$8F,$2D,$3C,$12
	.dc.b	$FC,$77,$F8,$2F,$CB,$3F,$00,$FE,$AB,$E2,$7F,$B6,$BF,$42,$FE,$AF
	.dc.b	$FB,$8E,$A7,$3E,$0A,$7F,$AD,$FE,$79,$9F,$C2,$3F,$71,$8F,$3F,$96
	.dc.b	$B8,$70,$BF,$AD,$CD,$CE,$7F,$4B,$8F,$3F,$3F,$F7,$98,$71,$D3,$C5
	.dc.b	$5E,$5E,$1F,$E1,$F3,$9B,$F1,$4B,$F6,$AF,$E3,$FF,$9E,$3F,$23,$7E
	.dc.b	$A4,$FC,$75,$FA,$A3,$F2,$07,$EA,$CF,$CA,$38,$E1,$C2,$F7,$5A,$FD
	.dc.b	$8B,$3F,$65,$FC,$57,$3E,$D9,$DC,$38,$67,$D7,$CE,$FE,$2D,$5A,$CB
	.dc.b	$9E,$F8,$BD,$E7,$E2,$F4,$7C,$39,$79,$F2,$F3,$6D,$FA,$EF,$87,$FD
	.dc.b	$37,$33,$F5,$AF,$F7,$6F,$E6,$5F,$EF,$0F,$D9,$9F,$A7,$FF,$92,$FF
	.dc.b	$69,$FE,$83,$5A,$FA,$75,$6B,$E6,$FF,$DB,$3D,$7F,$CE,$AB,$FD,$EF
	.dc.b	$FA,$FF,$F3,$17,$AE,$7F,$B6,$3C,$1F,$3C,$D3,$C1,$A3,$E0,$3F,$9A
	.dc.b	$5F,$7C,$FC,$7E,$AE,$30,$E1,$AF,$8B,$0F,$F3,$06,$7E,$69,$CD,$F9
	.dc.b	$72,$C1,$DB,$5F,$E0,$9A,$F9,$EB,$78,$7F,$9A,$7E,$63,$FB,$E7,$CF
	.dc.b	$7C,$6F,$BA,$FF,$67,$F1,$BF,$41,$71,$BA,$9F,$1D,$FA,$7C,$38,$7E
	.dc.b	$1F,$F8,$4B,$DD,$6B,$F3,$2E,$BC,$FF,$C5,$F3,$EA,$FF,$A4,$30,$E3
	.dc.b	$F4,$DF,$EC,$C7,$9E,$9E,$1D,$3C,$1E,$4F,$FB,$EA,$73,$9F,$6F,$0B
	.dc.b	$18,$F5,$D7,$C5,$87,$9F,$67,$D1,$F8,$A9,$AF,$BB,$1F,$3D,$3E,$5F
	.dc.b	$5F,$97,$BF,$40,$FF,$46,$71,$3F,$68,$7B,$EF,$AC,$FF,$78,$F4,$B8
	.dc.b	$BF,$1F,$8F,$1F,$BB,$F9,$18,$70,$B7,$E6,$5D,$B3,$7D,$17,$C1,$E1
	.dc.b	$FE,$79,$8F,$17,$E3,$18,$9F,$1E,$1E,$2F,$CE,$7E,$0C,$AB,$C1,$B7
	.dc.b	$9F,$6F,$F7,$9C,$F7,$D3,$C7,$1E,$9F,$CA,$F1,$5B,$D5,$FA,$6F,$C7
	.dc.b	$AF,$CB,$7E,$94,$FD,$29,$F8,$E3,$F2,$EF,$F5,$BF,$C4,$EF,$CC,$5E
	.dc.b	$EB,$5F,$84,$3F,$4D,$D4,$FB,$F8,$3E,$5D,$FF,$71,$7D,$E6,$23,$96
	.dc.b	$3F,$D0,$3A,$3E,$74,$FF,$61,$D3,$CB,$A7,$92,$BF,$DA,$33,$DA,$31
	.dc.b	$97,$D5,$BE,$33,$D6,$73,$E7,$EF,$4F,$3F,$F0,$0D,$C7,$FD,$A7,$F8
	.dc.b	$F7,$E2,$7D,$77,$EB,$FF,$8A,$FD,$D5,$F9,$DB,$F1,$8F,$F7,$0F,$ED
	.dc.b	$FB,$AD,$4E,$7E,$DD,$F9,$96,$7C,$7F,$85,$97,$4F,$D6,$EE,$1C,$32
	.dc.b	$F5,$F1,$FA,$2B,$C3,$6D,$3C,$35,$F1,$7C,$5F,$E3,$F5,$A9,$CE,$6F
	.dc.b	$4F,$C1,$18,$FD,$F3,$F4,$9F,$9E,$3F,$7E,$FF,$5A,$FE,$CB,$FE,$FA
	.dc.b	$C9,$F0,$BF,$E2,$5D,$BE,$4D,$EF,$F0,$55,$EE,$BF,$03,$5A,$C3,$FD
	.dc.b	$7F,$D7,$F7,$30,$ED,$F9,$5F,$E0,$5E,$58,$77,$F0,$3F,$37,$18,$DF
	.dc.b	$F5,$9F,$01,$8C,$7C,$B1,$FA,$FC,$DF,$BC,$FA,$A7,$3B,$7A,$3F,$79
	.dc.b	$FE,$01,$E9,$3F,$2F,$FF,$23,$FE,$E0,$FC,$E5,$F8,$3B,$F5,$A7,$ED
	.dc.b	$5C,$2F,$75,$A9,$7E,$CB,$9F,$B6,$FC,$D6,$5D,$7E,$DE,$E1,$C3,$27
	.dc.b	$EA,$9F,$B6,$AF,$15,$B4,$F1,$53,$F2,$AD,$61,$C7,$37,$E0,$9D,$EF
	.dc.b	$71,$FE,$B2,$EB,$FC,$DE,$5D,$27,$3E,$46,$F7,$F4,$79,$5F,$C8,$7F
	.dc.b	$43,$7E,$22,$FA,$57,$D4,$3F,$5F,$FD,$F7,$F0,$03,$F1,$BE,$17,$BA
	.dc.b	$D7,$A8,$AF,$D1,$EF,$FA,$CC,$BA,$FE,$8B,$FB,$D7,$BE,$77,$1E,$59
	.dc.b	$FF,$0B,$F1,$BF,$8C,$D6,$B3,$EB,$F9,$3E,$2F,$7A,$77,$FC,$5F,$C7
	.dc.b	$97,$A3,$2F,$3E,$3F,$CF,$7E,$31,$FA,$8F,$F7,$8F,$C6,$9F,$98,$FF
	.dc.b	$94,$BE,$79,$F9,$36,$EB,$53,$9F,$D5,$B1,$F3,$9F,$C5,$71,$E7,$FA
	.dc.b	$15,$BD,$FD,$D7,$F9,$F3,$93,$7A,$76,$8F,$EF,$91,$8C,$BC,$1F,$AA
	.dc.b	$59,$FF,$16,$FC,$0F,$F5,$3F,$D3,$5F,$9A,$FF,$30,$FD,$6F,$FC,$69
	.dc.b	$FA,$93,$F6,$87,$DA,$EE,$B5,$F6,$0A,$D7,$D8,$E7,$E4,$FF,$96,$65
	.dc.b	$D3,$E4,$6E,$1C,$2D,$F9,$4F,$F4,$8A,$D6,$9E,$0F,$97,$C5,$EF,$57
	.dc.b	$D9,$F8,$72,$5F,$9A,$5F,$89,$66,$31,$E7,$5D,$FC,$D4,$FC,$E1,$E4
	.dc.b	$3F,$7A,$FF,$1A,$FE,$70,$F8,$E7,$E5,$6F,$D5,$5C,$AF,$CB,$1F,$AD
	.dc.b	$BF,$19,$E7,$97,$2C,$78,$E1,$C2,$F7,$F9,$06,$BC,$0F,$F5,$46,$9D
	.dc.b	$F4,$FF,$22,$C7,$F9,$1B,$5F,$23,$AC,$5F,$BB,$EA,$3F,$D6,$13,$A5
	.dc.b	$7C,$F2,$F0,$7E,$CB,$AD,$3C,$99,$7A,$F7,$F5,$7E,$67,$9C,$7B,$4E
	.dc.b	$76,$F6,$FE,$97,$E7,$9E,$BB,$7B,$30,$F8,$5C,$2F,$B9,$D7,$EB,$AF
	.dc.b	$B8,$7E,$6F,$FE,$9D,$FB,$6F,$F3,$C7,$EF,$1F,$2B,$93,$D1,$7B,$AD
	.dc.b	$7E,$00,$AF,$F0,$0D,$6B,$0F,$A5,$AF,$C1,$8C,$5F,$C3,$8F,$0C,$F8
	.dc.b	$3F,$0D,$65,$FE,$87,$F2,$BD,$D4,$CF,$C1,$5A,$CD,$F9,$26,$A7,$38
	.dc.b	$7A,$3F,$49,$EB,$9F,$9B,$57,$E6,$DF,$17,$F5,$D7,$ED,$2F,$D8,$9F
	.dc.b	$A1,$3F,$15,$FF,$4D,$7E,$B5,$FE,$EC,$E4,$FB,$18,$70,$BD,$D6,$BE
	.dc.b	$E3,$A6,$3E,$4E,$3F,$81,$DD,$FE,$8A,$C3,$84,$60,$FF,$79,$3C,$F3
	.dc.b	$F0,$7D,$3B,$F2,$5F,$83,$39,$CB,$FD,$59,$F7,$B8,$CB,$D1,$87,$0D
	.dc.b	$BF,$DE,$9B,$79,$A7,$37,$EA,$F5,$7C,$BB,$FA,$34,$61,$FE,$74,$C3
	.dc.b	$84,$E6,$1F,$9E,$7E,$2F,$5F,$9F,$F6,$8F,$7D,$F9,$97,$F4,$EF,$35
	.dc.b	$FA,$66,$FB,$3F,$CB,$F5,$AF,$96,$FD,$D7,$E3,$39,$FC,$53,$F0,$5B
	.dc.b	$FE,$8F,$4E,$BC,$FF,$14,$F8,$58,$F2,$D3,$FA,$DF,$CE,$FE,$12,$BC
	.dc.b	$79,$F5,$F9,$BC,$5E,$F4,$FE,$11,$F8,$77,$8F,$83,$D1,$97,$9F,$F1
	.dc.b	$1F,$D3,$78,$5E,$BF,$DE,$9F,$B4,$BD,$2F,$F4,$BF,$EB,$5F,$E4,$1F
	.dc.b	$07,$FD,$87,$7B,$AD,$7C,$FA,$BF,$8A,$72,$7E,$F6,$BA,$72,$7A,$DF
	.dc.b	$91,$F8,$17,$C0,$C7,$9F,$1B,$4F,$0E,$9F,$EA,$6F,$D1,$3F,$43,$87
	.dc.b	$78,$FF,$54,$FC,$36,$2F,$86,$DE,$2D,$5F,$BB,$FB,$27,$37,$EA,$FC
	.dc.b	$CF,$F8,$27,$4B,$F3,$0F,$F3,$7F,$EF,$7F,$03,$C1,$7E,$52,$FD,$97
	.dc.b	$75,$AD,$7F,$0F,$3F,$AA,$E5,$FE,$35,$8F,$3F,$C1,$BD,$9C,$C3,$FC
	.dc.b	$33,$87,$F9,$D5,$FE,$4F,$5C,$23,$A6,$4F,$DC,$FF,$7F,$87,$3D,$3C
	.dc.b	$9E,$5E,$2F,$79,$FF,$B3,$31,$E5,$9A,$FD,$53,$FE,$25,$9C,$C6,$3E
	.dc.b	$7F,$D7,$5B,$D7,$D7,$F8,$CF,$F2,$7F,$79,$EB,$FF,$B5,$FE,$F3,$E1
	.dc.b	$7C,$77,$F4,$8F,$88,$D4,$E7,$93,$FA,$EF,$BE,$FD,$4B,$3F,$AE,$7C
	.dc.b	$2E,$76,$7D,$3C,$2C,$37,$4E,$BD,$5F,$C7,$A3,$19,$7F,$68,$F8,$6C
	.dc.b	$69,$DF,$6F,$EB,$3F,$88,$F8,$B2,$65,$E6,$D9,$F8,$7F,$F4,$1F,$15
	.dc.b	$FA,$73,$F3,$EF,$F5,$AF,$3F,$F7,$4F,$79,$F9,$6E,$EB,$5F,$2B,$FE
	.dc.b	$B7,$DB,$F9,$3F,$E5,$B9,$C7,$D9,$97,$4D,$6E,$1C,$33,$FE,$FF,$F0
	.dc.b	$9F,$CE,$2B,$C3,$7E,$0C,$5F,$99,$67,$0E,$11,$FC,$D3,$F1,$6E,$DA
	.dc.b	$78,$76,$FE,$F1,$E0,$3F,$76,$FC,$67,$57,$AF,$DA,$7E,$56,$FA,$A7
	.dc.b	$EF,$CE,$FF,$F0,$1F,$E7,$7F,$E0,$8C,$78,$F0,$B1,$E3,$87,$0B,$FF
	.dc.b	$04,$FE,$5D,$BC,$FF,$35,$7B,$CB,$F8,$BE,$3C,$B1,$F0,$D7,$F5,$D8
	.dc.b	$C6,$7E,$3A,$F1,$7A,$9F,$E0,$09,$CE,$7D,$F8,$F1,$87,$6C,$7C,$78
	.dc.b	$7A,$7E,$55,$F1,$3E,$3A,$F8,$19,$F5,$DB,$FC,$3F,$F8,$F6,$76,$F6
	.dc.b	$4E,$70,$7E,$AF,$9C,$FA,$E1,$FE,$26,$FD,$83,$A7,$EC,$FF,$F0,$07
	.dc.b	$E5,$19,$BE,$D9,$FA,$0B,$F3,$87,$E9,$DC,$3F,$22,$61,$C2,$F7,$5F
	.dc.b	$E6,$AE,$AF,$B1,$C5,$FC,$0F,$36,$0F,$85,$F9,$73,$98,$E6,$FC,$B3
	.dc.b	$18,$F5,$CD,$E4,$FC,$89,$61,$E9,$D7,$D1,$4F,$C9,$B5,$39,$D7,$CB
	.dc.b	$ED,$7A,$72,$E9,$B7,$F9,$37,$6F,$3E,$7F,$E0,$1F,$59,$FA,$27,$FA
	.dc.b	$FF,$E8,$DF,$86,$7F,$93,$3F,$20,$FA,$9F,$2E,$BD,$FE,$09,$FE,$40
	.dc.b	$AF,$F2,$89,$A8,$7E,$51,$F0,$32,$ED,$95,$C3,$93,$DE,$FF,$4F,$3C
	.dc.b	$63,$1E,$6F,$C1,$AD,$E3,$DF,$E8,$DF,$83,$6E,$31,$F8,$37,$E2,$7F
	.dc.b	$3D,$FF,$61,$FE,$2E,$DD,$FB,$DB,$F4,$E7,$EA,$9F,$E7,$4E,$4F,$87
	.dc.b	$87,$0B,$DD,$6B,$D3,$E9,$F2,$EF,$D3,$BD,$97,$DB,$8F,$ED,$07,$96
	.dc.b	$7D,$FE,$6B,$ED,$46,$31,$FF,$6F,$D7,$4D,$7C,$D7,$E5,$D7,$FD,$99
	.dc.b	$AF,$93,$35,$6B,$0F,$4F,$E4,$5C,$76,$F3,$E1,$EC,$C3,$85,$EE,$1F
	.dc.b	$84,$7E,$B7,$F3,$B7,$C1,$7F,$0B,$FB,$47,$E0,$EF,$C8,$1F,$89,$FE
	.dc.b	$67,$3F,$C3,$E1,$C3,$D7,$FF,$58,$5E,$EB,$F4,$B5,$AC,$7F,$8B,$71
	.dc.b	$7C,$4C,$3B,$FE,$1B,$F8,$37,$96,$7E,$0C,$FF,$C2,$55,$E5,$BF,$26
	.dc.b	$5E,$3F,$1B,$F4,$58,$70,$8F,$07,$EE,$BB,$D7,$C7,$8F,$A3,$BB,$F8
	.dc.b	$57,$9E,$FD,$CB,$FC,$DF,$F8,$7F,$F7,$FF,$65,$FB,$0F,$F8,$4F,$8B
	.dc.b	$F4,$77,$BF,$A0,$FF,$80,$3F,$1F,$FC,$52,$BF,$B3,$D6,$B1,$FF,$3C
	.dc.b	$F1,$7E,$03,$0E,$DE,$87,$E7,$DE,$59,$F7,$F4,$7F,$75,$18,$CB,$F8
	.dc.b	$FF,$C7,$E3,$1F,$2C,$77,$FD,$0F,$D3,$18,$DB,$CF,$F9,$9F,$F2,$5F
	.dc.b	$94,$FF,$65,$FE,$83,$FE,$11,$F8,$B7,$EC,$DF,$E6,$BF,$CF,$35,$F8
	.dc.b	$2A,$B5,$3F,$E6,$FB,$CC,$7E,$D5,$FF,$6C,$F1,$5A,$E0,$FF,$6C,$AF
	.dc.b	$F3,$B3,$C0,$C7,$B9,$FC,$56,$BC,$71,$BF,$AD,$FF,$01,$F2,$FF,$0E
	.dc.b	$9C,$E7,$E6,$9F,$EF,$33,$9C,$7B,$CE,$7E,$BF,$B3,$BF,$83,$FE,$89
	.dc.b	$F5,$BF,$C0,$3E,$95,$F9,$3B,$F0,$37,$E9,$CF,$DE,$1F,$B5,$F0,$F6
	.dc.b	$98,$70,$BD,$D3,$F1,$8D,$7E,$1B,$FC,$5F,$3E,$DC,$7F,$BC,$7C,$1A
	.dc.b	$77,$D1,$DB,$FC,$D6,$73,$97,$8B,$E8,$F1,$AF,$83,$0E,$1A,$3E,$7B
	.dc.b	$C9,$B7,$8F,$3F,$47,$ED,$3F,$F0,$01,$FF,$00,$FE,$0D,$FE,$1C,$FA
	.dc.b	$87,$EA,$1F,$E2,$2F,$C2,$3F,$A2,$FE,$F5,$BE,$2F,$F2,$95,$EE,$B5
	.dc.b	$DB,$FE,$F4,$EB,$F8,$B5,$AE,$97,$EB,$5F,$E9,$4D,$3B,$C7,$FA,$12
	.dc.b	$7F,$DB,$51,$8D,$3C,$3F,$4A,$D7,$FA,$AD,$78,$46,$9E,$19,$D7,$97
	.dc.b	$FE,$65,$CF,$C3,$5A,$CB,$FD,$23,$AF,$9B,$F1,$7E,$58,$F1,$CB,$FB
	.dc.b	$0F,$CE,$F5,$F9,$5F,$F7,$3F,$A7,$F6,$BE,$09,$FA,$57,$F4,$EF,$D1
	.dc.b	$B5,$EA,$7F,$89,$7A,$15,$AF,$90,$7E,$8D,$AF,$7F,$E0,$71,$7F,$9E
	.dc.b	$74,$EA,$FC,$03,$18,$F3,$C3,$94,$76,$D3,$C5,$A7,$F3,$AE,$3F,$6E
	.dc.b	$1E,$6D,$3C,$33,$FD,$8B,$F0,$AE,$11,$8D,$5F,$87,$E3,$3F,$37,$E4
	.dc.b	$D7,$0F,$ED,$3F,$71,$F3,$7F,$03,$7E,$9A,$FD,$35,$FB,$6F,$E8,$3F
	.dc.b	$69,$7E,$29,$F1,$38,$7E,$7E,$1C,$2F,$7E,$0A,$F7,$F5,$3D,$72,$E9
	.dc.b	$CB,$F0,$46,$1E,$0C,$FE,$3C,$62,$7C,$51,$E1,$FC,$CB,$EF,$63,$FE
	.dc.b	$58,$BF,$04,$F9,$7F,$01,$CE,$1F,$EA,$79,$CD,$78,$F5,$7E,$2D,$E3
	.dc.b	$D7,$CF,$AF,$B3,$F7,$1F,$C7,$7E,$CA,$FC,$61,$FB,$F3,$F3,$57,$E9
	.dc.b	$0F,$D9,$DF,$95,$7F,$41,$E6,$FC,$8E,$3C,$70,$E1,$7B,$AF,$E0,$55
	.dc.b	$AC,$7A,$F0,$FC,$3C,$3C,$1F,$9F,$FC,$27,$96,$7E,$0C,$DF,$85,$7B
	.dc.b	$AD,$5A,$CF,$C9,$F8,$9E,$63,$17,$E6,$F9,$FF,$6E,$7E,$CC,$BA,$6A
	.dc.b	$FD,$97,$F5,$DF,$83,$3F,$30,$FD,$7F,$EA,$CF,$C3,$FF,$CB,$3F,$73
	.dc.b	$FC,$B3,$AB,$FD,$13,$7B,$F4,$57,$BF,$95,$FF,$44,$7C,$F7,$4C,$BF
	.dc.b	$73,$C9,$F8,$1C,$3B,$7E,$1D,$F8,$07,$96,$1C,$7D,$CF,$74,$63,$5F
	.dc.b	$F3,$4F,$B3,$8C,$7C,$B1,$E4,$F8,$BE,$DA,$F8,$A3,$CF,$F1,$78,$CB
	.dc.b	$C9,$C0,$F8,$7C,$57,$B3,$F1,$5F,$E1,$DF,$8A,$FE,$67,$F2,$87,$E4
	.dc.b	$3F,$BE,$F8,$A7,$E6,$AF,$D9,$77,$E3,$AF,$7E,$A2,$F7,$F2,$7A,$FC
	.dc.b	$9F,$F3,$B9,$F5,$E9,$FD,$13,$DF,$4E,$D7,$F7,$63,$18,$73,$FA,$6C
	.dc.b	$67,$DE,$FC,$9A,$FF,$83,$7B,$7F,$D7,$78,$70,$C7,$CF,$D3,$FC,$A7
	.dc.b	$DF,$FC,$DF,$A3,$FE,$FD,$E8,$7C,$CF,$FB,$63,$EC,$B5,$E7,$2B,$53
	.dc.b	$FE,$DC,$FA,$0C,$C7,$ED,$B0,$FD,$2F,$6D,$BE,$1F,$8A,$FF,$36,$C3
	.dc.b	$FC,$C5,$C9,$C2,$31,$F9,$4F,$E6,$70,$F1,$C6,$FF,$11,$FF,$44,$70
	.dc.b	$FF,$B5,$27,$3B,$79,$A5,$F8,$07,$F4,$3D,$1F,$B6,$FE,$93,$D7,$FA
	.dc.b	$1F,$58,$FD,$69,$F9,$CE,$EB,$5C,$CA,$D7,$EE,$F9,$C3,$F8,$8E,$3D
	.dc.b	$3D,$7B,$87,$0B,$7A,$DF,$B1,$AD,$5F,$82,$F1,$7B,$CF,$FA,$2F,$B3
	.dc.b	$FD,$52,$F7,$9F,$86,$5F,$A3,$79,$70,$FE,$2B,$39,$FB,$CC,$65,$EC
	.dc.b	$B7,$E3,$BF,$D8,$FF,$6B,$7F,$00,$3E,$11,$FA,$2B,$EA,$1F,$8A,$BF
	.dc.b	$32,$DF,$0F,$F6,$6C,$78,$E1,$C2,$F7,$5F,$C8,$2B,$58,$F5,$EE,$FE
	.dc.b	$71,$87,$83,$E1,$BF,$9B,$BC,$B3,$F0,$67,$FE,$99,$AF,$35,$F9,$67
	.dc.b	$3F,$AC,$7F,$19,$BE,$19,$F8,$3F,$47,$DE,$BE,$3C,$BD,$1F,$76,$D7
	.dc.b	$FD,$67,$A0,$7C,$43,$0F,$F1,$BF,$D6,$BA,$0F,$F2,$6F,$DA,$F9,$7D
	.dc.b	$4F,$1C,$38,$DF,$0B,$DF,$C4,$BF,$37,$DD,$77,$CB,$B7,$5F,$18,$E5
	.dc.b	$A7,$91,$E7,$F5,$59,$7F,$C2,$99,$F7,$BF,$34,$F9,$32,$F2,$FD,$9E
	.dc.b	$AF,$B6,$9E,$3D,$BF,$C6,$71,$E2,$3A,$BE,$D3,$FA,$0F,$E1,$F9,$C7
	.dc.b	$AC,$67,$6F,$F5,$66,$7E,$FF,$C9,$7F,$0B,$F8,$A3,$F2,$E7,$ED,$EF
	.dc.b	$DB,$DF,$8F,$3F,$15,$F0,$61,$C2,$F7,$5A,$F2,$73,$DD,$FD,$17,$83
	.dc.b	$E0,$72,$FC,$3C,$3B,$C6,$35,$F1,$53,$5F,$16,$BE,$1F,$AB,$F8,$56
	.dc.b	$C3,$C1,$3E,$5E,$4C,$C6,$34,$F3,$FD,$37,$8B,$1E,$71,$EA,$F8,$CC
	.dc.b	$6F,$E8,$E1,$FE,$6D,$BF,$97,$6F,$14,$E4,$F9,$D2,$FD,$D5,$F1,$CF
	.dc.b	$DA,$DF,$9D,$FE,$B3,$F7,$97,$AF,$C2,$F7,$5A,$F0,$D4,$FA,$AD,$7C
	.dc.b	$7F,$E1,$79,$B8,$67,$77,$FC,$56,$BE,$18,$F0,$6B,$FB,$AF,$5F,$E2
	.dc.b	$CE,$74,$ED,$F4,$F8,$DB,$C3,$7E,$5F,$DB,$BA,$D7,$8F,$67,$EF,$7F
	.dc.b	$8B,$F8,$F7,$E4,$98,$F8,$DF,$47,$E0,$4D,$35,$39,$CF,$F8,$5C,$FE
	.dc.b	$B3,$83,$F4,$18,$72,$F8,$4F,$57,$83,$FA,$D8,$EB,$9E,$38,$FF,$71
	.dc.b	$18,$D3,$E3,$75,$7E,$02,$31,$9F,$F1,$8E,$4C,$6B,$E1,$E7,$6C,$F8
	.dc.b	$7F,$2E,$9F,$E7,$2C,$3C,$BF,$11,$FA,$7C,$DE,$0B,$EC,$DF,$8B,$3F
	.dc.b	$4F,$7E,$D2,$FC,$A3,$F8,$47,$37,$EE,$F1,$E3,$87,$0B,$DD,$7F,$65
	.dc.b	$AD,$63,$C7,$93,$F3,$C7,$F1,$5C,$38,$B8,$0F,$4C,$BC,$59,$3E,$DB
	.dc.b	$F4,$33,$9C,$BC,$BF,$8D,$E3,$1F,$56,$1C,$33,$FE,$BB,$9F,$8B,$4F
	.dc.b	$36,$7E,$BF,$87,$F8,$58,$70,$CB,$D3,$F3,$78,$BD,$E2,$F9,$AF,$E1
	.dc.b	$DF,$9A,$FF,$A0,$7F,$5A,$FF,$2F,$F9,$2E,$3F,$EC,$38,$EB,$EB,$15
	.dc.b	$A9,$FE,$CB,$86,$7E,$F7,$F0,$1C,$AC,$BA,$3F,$0D,$ED,$F5,$9F,$D6
	.dc.b	$2F,$FB,$6C,$F5,$C3,$A7,$B1,$FC,$12,$F7,$A7,$8F,$E3,$BF,$B4,$F2
	.dc.b	$7F,$31,$9C,$E3,$FD,$DB,$C1,$FE,$51,$F9,$73,$F2,$5F,$E3,$FF,$BA
	.dc.b	$FF,$20,$FD,$59,$F3,$1F,$E4,$3C,$BF,$9D,$5E,$FD,$85,$EE,$9F,$AC
	.dc.b	$F4,$C7,$F9,$4F,$0F,$DD,$C3,$B6,$FF,$01,$E5,$87,$19,$E1,$8F,$41
	.dc.b	$9B,$F3,$7F,$E4,$37,$C2,$73,$9F,$F8,$2B,$CA,$FE,$B1,$39,$BF,$4F
	.dc.b	$CA,$7F,$45,$FC,$B9,$F9,$1B,$EE,$9F,$92,$FF,$9F,$7E,$51,$F8,$B7
	.dc.b	$F9,$67,$93,$FA,$C6,$1C,$2F,$7F,$65,$FE,$0E,$A7,$D5,$75,$C9,$FA
	.dc.b	$B7,$DC,$C3,$87,$1F,$E7,$5E,$5A,$77,$E1,$FE,$77,$18,$CB,$FC,$33
	.dc.b	$F6,$F8,$C7,$CD,$7E,$5C,$DF,$AA,$F7,$D7,$C7,$BF,$A3,$E7,$F1,$7B
	.dc.b	$F8,$0B,$9F,$9F,$D7,$FF,$03,$7C,$F7,$F7,$DF,$C2,$5F,$AA,$FF,$81
	.dc.b	$FC,$2F,$F2,$4F,$D9,$BF,$14,$FF,$3E,$F4,$30,$E1,$7B,$F8,$55,$EF
	.dc.b	$E3,$3A,$F4,$7F,$84,$73,$EB,$FB,$BF,$F3,$9C,$78,$CD,$35,$8F,$3C
	.dc.b	$3F,$CC,$79,$77,$9C,$9F,$A5,$CB,$BF,$DC,$FF,$87,$67,$52,$CB,$D6
	.dc.b	$EE,$FD,$5F,$A2,$79,$23,$58,$F7,$D5,$E5,$E7,$DC,$7E,$FA,$FB,$B7
	.dc.b	$E5,$3F,$E7,$FF,$D9,$9F,$91,$BF,$6E,$E1,$7B,$AD,$4B,$F0,$AC,$FC
	.dc.b	$6F,$BF,$6F,$EC,$B7,$D5,$FF,$01,$F2,$FF,$8B,$F4,$F0,$E9,$E0,$FB
	.dc.b	$46,$BE,$CE,$4F,$7D,$6B,$5F,$27,$BD,$F0,$F1,$E5,$9F,$8F,$1E,$3B
	.dc.b	$7E,$7F,$1E,$33,$9C,$BD,$7E,$47,$C5,$FC,$8B,$F8,$97,$82,$FD,$B1
	.dc.b	$F4,$DF,$9C,$F8,$47,$E7,$1B,$F0,$7F,$D1,$B5,$AC,$7F,$A3,$FE,$47
	.dc.b	$9C,$7E,$06,$5F,$C7,$7C,$9B,$87,$0C,$FF,$4F,$FB,$77,$A7,$9F,$FC
	.dc.b	$B1,$7E,$2C,$FA,$76,$E6,$53,$9B,$F2,$7E,$D3,$DE,$FC,$5B,$BF,$39
	.dc.b	$C5,$7A,$7F,$57,$BC,$1F,$D7,$3F,$3B,$FC,$27,$E7,$4F,$CB,$1F,$88
	.dc.b	$3E,$73,$F9,$27,$EC,$1F,$ED,$BF,$AB,$DD,$6A,$73,$F3,$6F,$C3,$B3
	.dc.b	$FA,$FF,$DF,$CB,$A6,$57,$0E,$1F,$0C,$CF,$F8,$4F,$2F,$4A,$F0,$4E
	.dc.b	$7D,$3F,$81,$87,$2D,$3B,$FE,$37,$B7,$C3,$FA,$B7,$E7,$7E,$B7,$D5
	.dc.b	$F2,$4F,$CC,$5F,$AB,$3D,$B7,$E6,$BF,$CF,$FC,$8F,$D4,$E1,$C2,$F7
	.dc.b	$5A,$FC,$63,$A7,$1B,$F1,$9F,$CD,$5F,$6E,$3F,$CC,$9F,$E8,$F7,$96
	.dc.b	$7D,$FF,$08,$FB,$11,$8C,$7F,$A2,$D7,$4D,$7C,$D7,$E5,$D5,$FB,$BE
	.dc.b	$75,$F1,$C3,$F2,$EF,$DD,$7D,$2B,$F5,$CF,$E0,$18,$BF,$1B,$7E,$3C
	.dc.b	$FC,$8D,$FB,$1F,$0B,$DD,$6A,$7F,$C1,$9C,$BF,$C2,$B1,$E9,$F0,$5F
	.dc.b	$B7,$7B,$7D,$FD,$C3,$8E,$5F,$E2,$5E,$1E,$96,$D3,$C5,$3F,$B7,$F6
	.dc.b	$3F,$D5,$58,$F4,$9C,$E7,$FE,$2B,$CF,$C1,$97,$48,$7E,$79,$D2,$B7
	.dc.b	$FA,$FD,$DB,$D1,$EA,$FF,$1B,$FD,$25,$FA,$83,$F6,$57,$E4,$5F,$EB
	.dc.b	$FF,$C1,$5F,$93,$FF,$36,$DF,$F9,$3E,$1C,$2F,$7F,$83,$7F,$B0,$EB
	.dc.b	$FC,$7B,$C5,$FE,$C1,$AE,$BB,$7F,$19,$78,$72,$FE,$00,$3C,$F3,$F0
	.dc.b	$7C,$A7,$ED,$E3,$18,$FF,$B3,$FE,$0F,$1B,$79,$EF,$CD,$9F,$DB,$D7
	.dc.b	$CB,$39,$BF,$27,$D5,$7F,$39,$FA,$4F,$F6,$C7,$A4,$FE,$00,$FC,$E7
	.dc.b	$DB,$7F,$86,$78,$F4,$7F,$B8,$23,$3E,$7F,$F9,$4F,$8F,$FD,$B9,$87
	.dc.b	$2E,$2F,$53,$FE,$E3,$F4,$FF,$9B,$3D,$AF,$AF,$C7,$7F,$11,$8C,$78
	.dc.b	$FF,$E6,$EF,$C1,$31,$9F,$78,$FF,$38,$E7,$EB,$CF,$BC,$E7,$4F,$F1
	.dc.b	$77,$E2,$BF,$EF,$7F,$9A,$FF,$77,$7E,$2E,$8F,$D0,$FD,$93,$F6,$AF
	.dc.b	$F7,$D7,$BA,$FC,$BF,$87,$1F,$FB,$F7,$0E,$1E,$77,$F9,$E6,$F7,$F6
	.dc.b	$5A,$A5,$69,$FF,$68,$E7,$E0,$CD,$FA,$BF,$FA,$B6,$31,$8F,$3A,$FF
	.dc.b	$13,$1C,$EF,$BE,$9E,$58,$7E,$1B,$FD,$CE,$F8,$CE,$75,$7E,$85,$C7
	.dc.b	$3F,$6E,$BE,$CF,$C7,$31,$A7,$7B,$7E,$D7,$8C,$79,$7C,$67,$AF,$F1
	.dc.b	$A7,$96,$FD,$73,$F8,$DF,$DE,$BE,$8F,$F4,$3F,$1B,$BA,$D7,$B2,$AD
	.dc.b	$79,$2F,$CC,$F5,$FA,$D7,$C1,$E7,$0D,$3B,$E9,$98,$C6,$5D,$31,$F0
	.dc.b	$79,$DF,$06,$78,$DF,$0C,$FC,$1A,$FC,$69,$F2,$63,$E8,$8F,$DB,$FD
	.dc.b	$DE,$2B,$7A,$BB,$3E,$D7,$C1,$FF,$84,$BF,$4D,$FD,$07,$EF,$8F,$C4
	.dc.b	$7F,$BA,$FE,$D5,$AF,$A1,$56,$A7,$FB,$CF,$B5,$98,$F5,$61,$FE,$1C
	.dc.b	$FD,$2E,$DF,$0F,$39,$97,$5F,$C8,$FD,$9C,$4C,$3C,$55,$AC,$DF,$95
	.dc.b	$EA,$73,$87,$97,$E5,$DE,$EF,$87,$0E,$11,$FE,$0A,$F6,$1E,$3F,$D0
	.dc.b	$CF,$FD,$A7,$FA,$2B,$F6,$97,$E5,$0F,$C7,$1F,$B4,$7F,$5C,$FC,$0B
	.dc.b	$F3,$DF,$85,$B0,$E1,$7B,$AF,$EC,$3F,$B4,$6A,$7A,$65,$FA,$23,$EF
	.dc.b	$DF,$1F,$BA,$FB,$CF,$3C,$FC,$39,$F8,$3E,$F3,$E1,$CE,$73,$ED,$F6
	.dc.b	$F8,$CB,$C3,$7E,$6F,$93,$7F,$00,$78,$72,$E7,$B7,$F7,$4F,$DA,$71
	.dc.b	$97,$9E,$DF,$AD,$63,$1E,$5F,$BA,$FE,$EF,$F2,$F7,$D0,$FD,$EF,$D7
	.dc.b	$5F,$9A,$BF,$6C,$7E,$1F,$FE,$F9,$EA,$F9,$BC,$FF,$9C,$C7,$8F,$8A
	.dc.b	$C7,$8E,$1C,$3F,$41,$ED,$9B,$C6,$FD,$06,$1C,$31,$F9,$01,$E7,$87
	.dc.b	$2F,$89,$FC,$3C,$63,$5F,$27,$D0,$63,$1F,$44,$79,$F6,$7E,$83,$E0
	.dc.b	$DB,$C9,$39,$FD,$E7,$F8,$D7,$EC,$1F,$ED,$5E,$27,$E7,$BF,$E3,$5F
	.dc.b	$DD,$1F,$BD,$3F,$7B,$DF,$C7,$AF,$7F,$56,$BD,$FE,$BB,$A9,$FD,$E6
	.dc.b	$5F,$EE,$1E,$BB,$87,$1C,$FB,$7E,$21,$FA,$2A,$F1,$47,$87,$C6,$C5
	.dc.b	$6F,$3E,$DA,$BF,$7A,$ED,$7E,$78,$7E,$95,$FC,$D7,$F2,$D7,$E5,$0F
	.dc.b	$B1,$7E,$8A,$FD,$4D,$F9,$93,$F3,$3F,$E4,$BE,$2F,$DC,$F0,$99,$72
	.dc.b	$C7,$8E,$1C,$2F,$FD,$7B,$7B,$AD,$5F,$DA,$7C,$3F,$6D,$FC,$75,$E5
	.dc.b	$97,$87,$2F,$C7,$D7,$9F,$3F,$34,$E7,$F2,$8F,$E0,$B7,$C3,$3F,$04
	.dc.b	$EF,$4F,$1E,$3E,$9F,$C2,$B1,$3D,$75,$78,$F8,$C3,$D1,$F5,$1F,$D6
	.dc.b	$7F,$13,$FE,$AB,$FA,$1F,$E7,$BF,$01,$FF,$12,$F4,$7E,$57,$8F,$E1
	.dc.b	$56,$BC,$25,$6B,$9F,$97,$CF,$3F,$57,$FE,$75,$7D,$39,$BF,$36,$F1
	.dc.b	$BF,$F1,$8F,$D3,$7F,$20,$8C,$7C,$3B,$E9,$B1,$8B,$6F,$24,$B6,$7F
	.dc.b	$00,$78,$27,$30,$F7,$FF,$BA,$7E,$AA,$FC,$63,$FA,$FF,$D9,$FF,$A3
	.dc.b	$FF,$5E,$FC,$47,$CA,$EA,$73,$E2,$A7,$F1,$9E,$16,$7E,$7B,$FA,$A6
	.dc.b	$3C,$FF,$67,$B8,$70,$B7,$E5,$FF,$CE,$AB,$C1,$7F,$E6,$7C,$7A,$7E
	.dc.b	$65,$9C,$38,$79,$2F,$C5,$BB,$6D,$E5,$8F,$17,$6B,$F4,$FF,$E7,$1F
	.dc.b	$64,$FD,$27,$F8,$97,$EB,$39,$FD,$39,$FB,$2B,$F5,$86,$3C,$7C,$1F
	.dc.b	$ED,$F8,$F1,$F8,$46,$3C,$70,$ED,$7B,$B5,$ED,$FE,$90,$7F,$78,$8D
	.dc.b	$47,$8F,$E8,$73,$39,$D3,$CF,$A7,$99,$68,$FC,$9B,$55,$E4,$AD,$4B
	.dc.b	$F4,$6C,$CF,$8A,$73,$9B,$9B,$DB,$8F,$6D,$BD,$6F,$69,$7E,$8D,$AC
	.dc.b	$3B,$56,$B2,$7E,$E7,$DB,$F2,$7F,$C8,$7A,$1F,$8F,$F7,$7F,$D6,$5F
	.dc.b	$A1,$7F,$B2,$78,$FE,$25,$6B,$C4,$FE,$41,$F8,$7A,$9F,$15,$DF,$55
	.dc.b	$FE,$6F,$AB,$E2,$C6,$63,$BB,$F2,$31,$EB,$E5,$7F,$14,$CB,$AF,$E5
	.dc.b	$9E,$58,$C6,$9F,$EC,$5F,$97,$C6,$DE,$7D,$BC,$5F,$96,$F5,$BD,$C7
	.dc.b	$97,$E2,$BC,$9F,$A0,$3F,$67,$7E,$A9,$FC,$87,$F2,$1F,$C4,$7B,$DF
	.dc.b	$E3,$9E,$0F,$C1,$5E,$EB,$5E,$1F,$F2,$6F,$53,$9E,$3C,$B8,$FD,$5D
	.dc.b	$BF,$E0,$DC,$3B,$7E,$1F,$F0,$32,$E7,$7E,$0F,$71,$AF,$E5,$FE,$0B
	.dc.b	$EF,$E1,$DE,$73,$A7,$F2,$4C,$38,$EB,$E2,$DB,$FC,$F5,$18,$AD,$C7
	.dc.b	$FA,$77,$4F,$F0,$F7,$E3,$7F,$F0,$07,$F1,$0F,$AB,$FA,$A3,$F4,$B7
	.dc.b	$E9,$3F,$A6,$CB,$F4,$17,$BE,$F5,$EE,$9E,$87,$DD,$A7,$E3,$AF,$DB
	.dc.b	$7D,$DA,$78,$21,$A6,$39,$BF,$CB,$F7,$D7,$2F,$17,$BF,$F9,$E8,$C6
	.dc.b	$CF,$CF,$31,$8F,$2D,$BC,$73,$E8,$FD,$FF,$C5,$39,$C7,$CD,$F2,$B8
	.dc.b	$D3,$DF,$7E,$EF,$94,$FC,$A7,$E9,$2F,$CA,$3F,$DF,$DF,$99,$FA,$7C
	.dc.b	$17,$E4,$CF,$C4,$77,$5A,$E5,$56,$BE,$3F,$3F,$8F,$FB,$79,$99,$F5
	.dc.b	$FC,$23,$0F,$79,$69,$DF,$F4,$3F,$DE,$CE,$63,$C5,$EC,$E3,$4F,$06
	.dc.b	$7E,$4F,$C8,$BF,$59,$FB,$6F,$F4,$0F,$95,$FF,$12,$7E,$A9,$FC,$33
	.dc.b	$F6,$EF,$F5,$8F,$85,$FC,$66,$B5,$F4,$EA,$D7,$E0,$3C,$BE,$47,$FC
	.dc.b	$03,$C7,$F9,$83,$FD,$27,$7D,$5C,$07,$9D,$7F,$1A,$FB,$CF,$E2,$B1
	.dc.b	$8F,$77,$FC,$D9,$97,$87,$06,$9E,$69,$69,$FB,$0A,$FE,$5D,$F9,$A3
	.dc.b	$F1,$DF,$F2,$CF,$C5,$3F,$3C,$7E,$9F,$FE,$AB,$F3,$BF,$86,$5E,$FE
	.dc.b	$DB,$FA,$45,$6B,$F0,$A7,$EF,$7A,$C9,$FC,$03,$F5,$70,$E1,$C9,$EC
	.dc.b	$79,$69,$DF,$47,$E7,$1F,$CC,$27,$39,$F8,$BC,$CC,$65,$E6,$8F,$07
	.dc.b	$EC,$3D,$F6,$F1,$E3,$FE,$C1,$EE,$C6,$7D,$77,$7D,$BE,$31,$F4,$7E
	.dc.b	$E3,$FC,$9F,$ED,$5F,$88,$FF,$B5,$3F,$0F,$FF,$51,$7E,$2D,$FA,$ED
	.dc.b	$B5,$39,$FC,$25,$3F,$B2,$D7,$3F,$3F,$F9,$3C,$79,$FC,$ED,$C3,$85
	.dc.b	$FF,$36,$E6,$E7,$3F,$ED,$BC,$79,$DF,$F5,$03,$8E,$7E,$2F,$57,$D3
	.dc.b	$D7,$FD,$96,$73,$5E,$29,$6D,$FD,$C3,$DF,$FF,$57,$7E,$54,$FD,$AF
	.dc.b	$FC,$D7,$8F,$E4,$FF,$1C,$7E,$59,$FD,$5A,$3F,$BC,$6D,$FE,$E4,$BE
	.dc.b	$1F,$59,$FD,$EB,$C8,$FF,$7E,$DF,$F6,$1B,$DD,$AF,$6F,$F4,$BC,$BC
	.dc.b	$39,$3F,$65,$FE,$65,$18,$CB,$A6,$7F,$EE,$7C,$79,$E3,$E0,$7C,$DF
	.dc.b	$8B,$7F,$7F,$BF,$F6,$47,$5B,$E4,$FF,$7F,$EC,$7F,$08,$FD,$71,$FB
	.dc.b	$D7,$F8,$C7,$1F,$2D,$EF,$D5,$FE,$19,$ED,$AF,$7E,$77,$C7,$E2,$1F
	.dc.b	$E2,$3C,$F4,$FA,$CC,$E1,$CD,$C1,$BC,$3A,$DB,$EC,$BD,$13,$C2,$FC
	.dc.b	$F8,$F8,$76,$CC,$F1,$D7,$D5,$AB,$F0,$CF,$24,$E6,$FD,$12,$FC,$1F
	.dc.b	$3F,$99,$3F,$1D,$FF,$65,$FE,$36,$FD,$73,$FA,$4F,$DF,$FF,$A9,$79
	.dc.b	$3E,$BD,$EF,$F0,$25,$EF,$E2,$57,$FB,$8B,$DE,$3F,$ED,$6D,$7D,$F8
	.dc.b	$FF,$A3,$9B,$1E,$0C,$BB,$C7,$92,$31,$83,$F0,$2F,$7D,$F8,$72,$F0
	.dc.b	$65,$FE,$E2,$CB,$F1,$7F,$46,$FB,$7F,$D9,$FE,$95,$FD,$13,$F3,$62
	.dc.b	$FC,$8F,$E2,$D6,$FC,$CC,$F2,$2F,$7E,$AE,$B1,$FD,$EF,$0F,$D8,$C3
	.dc.b	$B7,$C3,$7E,$52,$3F,$03,$1C,$F3,$F0,$7E,$CB,$F5,$A3,$19,$75,$FC
	.dc.b	$6F,$18,$F9,$A3,$C1,$6F,$D1,$3C,$1A,$F9,$35,$F4,$FE,$89,$F7,$3F
	.dc.b	$72,$FE,$5D,$F2,$E3,$FA,$EF,$63,$F2,$CF,$DB,$F8,$FF,$94,$E1,$5F
	.dc.b	$0B,$DD,$7F,$A9,$6B,$53,$FC,$E7,$F8,$03,$D9,$7C,$3D,$AE,$CF,$E6
	.dc.b	$27,$97,$C6,$FE,$A3,$E8,$7F,$27,$18,$CF,$FC,$41,$8F,$4C,$3C,$D9
	.dc.b	$F8,$3C,$FF,$54,$63,$1F,$D8,$E8,$FD,$E7,$37,$ED,$9C,$F9,$BF,$9C
	.dc.b	$F9,$5F,$EF,$DD,$2F,$D3,$BF,$DC,$7F,$42,$FA,$6F,$F4,$8E,$DF,$06
	.dc.b	$F7,$5A,$9F,$F0,$B7,$81,$FC,$4A,$79,$F1,$FE,$D1,$FF,$05,$7E,$EB
	.dc.b	$F0,$9E,$3F,$59,$FE,$EE,$FC,$BB,$F6,$31,$8C,$DE,$7E,$3C,$F6,$DE
	.dc.b	$5C,$3F,$DE,$BA,$7F,$A9,$F4,$FE,$43,$FB,$4B,$E5,$5F,$5E,$F9,$6F
	.dc.b	$ED,$3E,$5B,$F4,$17,$EC,$5F,$BC,$C3,$85,$EF,$E3,$7F,$67,$5F,$EA
	.dc.b	$0C,$BA,$E5,$D3,$93,$F8,$7E,$1D,$FF,$70,$FB,$AF,$2D,$3B,$FC,$EF
	.dc.b	$F0,$C8,$C6,$5F,$BE,$F3,$7F,$87,$FC,$2B,$F5,$9F,$F0,$1F,$ED,$9F
	.dc.b	$97,$FC,$BD,$F6,$AF,$D3,$FF,$88,$DE,$FE,$71,$7B,$AF,$F4,$6F,$0F
	.dc.b	$F4,$2A,$D7,$07,$F2,$47,$B6,$FF,$92,$79,$5F,$7F,$97,$F1,$E9,$E2
	.dc.b	$F4,$1A,$BE,$3B,$3A,$F8,$27,$3A,$BE,$B3,$CB,$87,$1C,$3D,$3F,$86
	.dc.b	$FF,$B4,$3E,$FB,$F5,$1C,$EF,$8B,$FE,$3F,$F9,$67,$F8,$93,$EF,$3F
	.dc.b	$D4,$FF,$A2,$EF,$93,$FA,$0D,$EF,$BD,$7B,$E2,$7E,$5D,$BB,$FC,$06
	.dc.b	$4F,$CB,$7F,$D1,$DA,$78,$23,$F9,$6C,$FF,$AB,$63,$18,$F4,$FA,$6F
	.dc.b	$F5,$AE,$3F,$ED,$B9,$CE,$5E,$5E,$5C,$61,$E4,$C7,$D3,$87,$0F,$80
	.dc.b	$F1,$7E,$20,$FC,$65,$FB,$D3,$F1,$C7,$A0,$FC,$6F,$FB,$B7,$95,$C2
	.dc.b	$F7,$5A,$9F,$EF,$5F,$2D,$9C,$3D,$7A,$FF,$64,$CB,$B7,$E8,$58,$84
	.dc.b	$78,$67,$C1,$9F,$7F,$4F,$D3,$5A,$CF,$BC,$F8,$FD,$1C,$FD,$8B,$F1
	.dc.b	$DF,$C8,$7E,$A9,$FE,$53,$F9,$67,$E9,$CF,$C1,$7E,$85,$EF,$CC,$5E
	.dc.b	$E9,$F9,$AE,$B1,$FC,$36,$BF,$44,$FE,$FB,$87,$17,$01,$E9,$7E,$1F
	.dc.b	$47,$CB,$5E,$4F,$8F,$FF,$08,$7E,$2F,$8C,$FC,$31,$FE,$13,$CD,$FB
	.dc.b	$87,$E1,$7D,$77,$E3,$EF,$05,$F5,$5F,$8C,$FD,$F1,$FC,$03,$FE,$73
	.dc.b	$9F,$E7,$D7,$C1,$DF,$88,$AD,$F8,$1D,$71,$E9,$C7,$EA,$C3,$BF,$E2
	.dc.b	$7F,$55,$E5,$9F,$7F,$7F,$E9,$C6,$32,$E9,$F2,$DF,$5B,$F0,$B7,$E7
	.dc.b	$9F,$AD,$FC,$77,$FC,$57,$F8,$97,$F9,$33,$F1,$9F,$D0,$56,$BD,$75
	.dc.b	$6B,$EB,$BF,$8E,$FB,$FE,$4E,$0F,$F7,$F6,$4F,$5F,$F1,$18,$F1,$9F
	.dc.b	$A2,$F3,$C3,$BF,$CF,$78,$EF,$C5,$F2,$8D,$5F,$99,$67,$5F,$0C,$E7
	.dc.b	$57,$E8,$5F,$B5,$FD,$5B,$F8,$BF,$77,$D4,$F3,$1C,$7F,$E7,$EF,$C1
	.dc.b	$F5,$F0,$7F,$EA,$1F,$03,$F5,$33,$FE,$82,$FC,$2B,$3F,$47,$F9,$FC
	.dc.b	$B9,$F2,$FE,$21,$EB,$7C,$3E,$23,$F3,$F1,$8C,$FB,$F9,$18,$CF,$B4
	.dc.b	$78,$BE,$81,$FB,$9F,$F0,$BF,$DC,$DF,$98,$3F,$4F,$7E,$E4,$FA,$B7
	.dc.b	$E4,$4F,$C5,$9F,$A4,$F1,$C3,$85,$EE,$B5,$F9,$E6,$7E,$47,$F7,$19
	.dc.b	$F6,$CE,$E1,$C3,$3E,$BE,$EF,$EE,$2B,$59,$73,$F6,$FF,$77,$FB,$53
	.dc.b	$F0,$57,$EA,$6F,$D6,$9F,$A1,$F9,$F7,$FF,$25,$7A,$DF,$E1,$9C,$38
	.dc.b	$5E,$EB,$5F,$90,$FF,$2B,$9C,$F6,$FE,$D1,$ED,$F6,$5F,$C3,$DE,$39
	.dc.b	$76,$E0,$FB,$75,$C7,$4F,$1C,$E7,$E6,$BF,$88,$FA,$4F,$C7,$BC,$BF
	.dc.b	$B9,$BE,$7D,$FC,$00,$FC,$C9,$F9,$6F,$F3,$DC,$38,$70,$B0,$E1,$7B
	.dc.b	$FD,$C3,$59,$3F,$4C,$FE,$49,$87,$0D,$FF,$86,$BC,$B0,$E3,$D3,$E4
	.dc.b	$BF,$1E,$BE,$2F,$C2,$FF,$93,$F0,$BA,$FF,$DE,$FF,$9C,$7E,$13,$E6
	.dc.b	$9F,$89,$FF,$03,$F7,$7F,$5D,$7B,$F0,$97,$BA,$7E,$DF,$FD,$5E,$BA
	.dc.b	$73,$FF,$39,$78,$70,$FF,$22,$79,$5F,$7F,$C5,$7C,$7A,$78,$BC,$3F
	.dc.b	$F4,$3F,$D4,$7F,$3C,$F9,$EF,$F3,$97,$9A,$FA,$97,$E6,$8F,$DD,$1F
	.dc.b	$AF,$3E,$83,$FB,$46,$1C,$2F,$75,$AF,$8A,$CF,$B0,$FC,$CB,$FA,$75
	.dc.b	$F6,$E2,$FE,$6E,$F1,$CF,$B7,$C1,$78,$EF,$FC,$2B,$97,$4F,$8B,$FE
	.dc.b	$A3,$F4,$6F,$D7,$BF,$BC,$FE,$73,$FD,$3F,$F7,$1F,$F3,$5F,$C7,$7F
	.dc.b	$7B,$AD,$7D,$42,$B5,$F8,$07,$F0,$CF,$03,$F8,$E5,$6B,$83,$FA,$4B
	.dc.b	$D7,$EA,$FF,$B2,$3C,$AF,$B7,$E0,$3E,$2D,$3C,$3F,$38,$D5,$FA,$0F
	.dc.b	$F6,$6F,$CB,$1E,$2F,$F4,$AE,$C7,$EC,$3F,$E9,$FF,$C7,$1F,$97,$FF
	.dc.b	$84,$6F,$7F,$15,$FE,$34,$AD,$7E,$DF,$9C,$7F,$A9,$F1,$7F,$5E,$BE
	.dc.b	$BF,$8C,$FF,$79,$78,$E7,$DB,$AB,$FD,$83,$5C,$72,$E9,$1F,$CC,$23
	.dc.b	$FC,$41,$EF,$FE,$98,$FF,$11,$FE,$6D,$FC,$57,$DB,$7E,$01,$AF,$6F
	.dc.b	$FB,$07,$62,$B5,$F9,$76,$7E,$7F,$FC,$15,$8F,$3C,$BF,$B9,$BD,$6F
	.dc.b	$87,$BF,$E4,$EE,$FE,$99,$39,$C7,$F9,$27,$87,$FE,$0E,$C7,$FD,$F3
	.dc.b	$F8,$63,$F5,$DF,$F2,$BF,$E0,$BF,$95,$FC,$5F,$FC,$71,$5A,$9C,$F8
	.dc.b	$FF,$98,$CF,$FA,$C4,$B6,$FF,$41,$9F,$E0,$67,$AB,$46,$3D,$F9,$59
	.dc.b	$3C,$FF,$25,$6B,$2F,$1F,$B9,$E7,$F8,$9F,$CA,$7E,$AD,$F2,$BE,$41
	.dc.b	$F9,$E7,$9A,$73,$CB,$F2,$67,$FC,$4D,$BF,$9A,$7E,$A6,$5F,$EC,$53
	.dc.b	$87,$8D,$FE,$35,$E0,$FF,$4D,$AC,$BE,$BF,$D6,$78,$63,$19,$7D,$5F
	.dc.b	$CD,$3E,$57,$93,$F9,$2C,$3E,$C7,$11,$7E,$81,$F2,$A7,$5D,$1F,$E8
	.dc.b	$49,$7E,$39,$9F,$7B,$C5,$5F,$9C,$DB,$E5,$3C,$3D,$5F,$F6,$E7,$0F
	.dc.b	$D0,$8E,$79,$FF,$A0,$6B,$EB,$F3,$7F,$34,$C9,$F9,$BF,$E4,$3E,$43
	.dc.b	$3E,$0F,$96,$3F,$27,$F1,$FC,$3E,$2F,$E1,$61,$C6,$7F,$CF,$3F,$37
	.dc.b	$F8,$78,$C7,$23,$E6,$BF,$15,$ED,$FE,$2B,$BF,$F3,$1C,$58,$DF,$EF
	.dc.b	$C7,$C4,$DF,$F1,$75,$F4,$67,$D7,$E2,$7F,$BD,$E5,$3F,$D3,$27,$F1
	.dc.b	$4F,$5E,$66,$9F,$99,$F5,$BF,$21,$D5,$F3,$FF,$69,$FD,$6F,$3F,$E5
	.dc.b	$27,$39,$27,$FD,$27,$F9,$F7,$E7,$27,$31,$FE,$C7,$AE,$19,$FF,$21
	.dc.b	$BE,$1F,$6F,$FA,$7F,$3F,$F3,$DE,$8F,$DD,$F4,$3D,$F8,$7E,$8A,$73
	.dc.b	$7F,$E5,$CD,$BF,$D8,$33,$9C,$7F,$1D,$5E,$D7,$F2,$51,$9C,$3A,$6D
	.dc.b	$DE,$7F,$19,$8F,$CF,$7F,$67,$C1,$FA,$FE,$9F,$C7,$3E,$DE,$5C,$CF
	.dc.b	$E0,$63,$E0,$7B,$58,$AD,$FD,$6F,$F0,$19,$E5,$B7,$F9,$BF,$7F,$E0
	.dc.b	$8F,$EE,$F0,$FD,$BB,$FC,$17,$7F,$E0,$6E,$7C,$AF,$CF,$79,$4E,$4C
	.dc.b	$DF,$EA,$EB,$DC,$FD,$E3,$F8,$33,$FA,$57,$F8,$8A,$F2,$BC,$31,$8F
	.dc.b	$8C,$F1,$F9,$3F,$6B,$C9,$FF,$65,$F2,$FF,$10,$C3,$F8,$7F,$4F,$EB
	.dc.b	$E7,$EF,$4E,$6B,$F0,$3D,$3C,$B1,$FE,$02,$7A,$BC,$4F,$AF,$8F,$F2
	.dc.b	$FB,$E9,$18,$D3,$55,$FC,$6E,$9F,$90,$FB,$6B,$F9,$07,$8D,$F6,$B8
	.dc.b	$FF,$35,$49,$FC,$1E,$5F,$33,$EE,$BF,$AB,$E5,$FE,$D0,$FC,$87,$FC
	.dc.b	$AF,$A7,$AE,$73,$F4,$29,$7C,$0F,$D0,$9F,$99,$FB,$87,$CD,$77,$97
	.dc.b	$5E,$FF,$A5,$83,$F6,$CF,$B7,$E3,$FD,$ED,$BF,$CA,$5A,$7C,$98,$C7
	.dc.b	$91,$F7,$1F,$E7,$5E,$4F,$AF,$93,$EB,$DF,$F3,$57,$F8,$67,$0F,$E1
	.dc.b	$BC,$6F,$C1,$4F,$F3,$AC,$7E,$A4,$63,$8F,$F0,$EF,$E0,$FF,$0C,$FC
	.dc.b	$23,$F8,$DE,$7F,$C2,$63,$FC,$9F,$0F,$E0,$2F,$F5,$4C,$7F,$26,$7E
	.dc.b	$4F,$D6,$DC,$FF,$2D,$8F,$CB,$75,$71,$CF,$F0,$91,$83,$F3,$7E,$3F
	.dc.b	$E4,$FC,$9F,$CA,$5F,$F4,$1C,$BF,$3E,$FF,$43,$F0,$3F,$9C,$61,$F8
	.dc.b	$D8,$FE,$8D,$E7,$EA,$BC,$5A,$7E,$8A,$7F,$AF,$BF,$C1,$EB,$84,$FE
	.dc.b	$A7,$E1,$BA,$F3,$FE,$AF,$B6,$DF,$F0,$6B,$ED,$3F,$8D,$C3,$E7,$4B
	.dc.b	$8F,$CB,$C9,$FB,$3F,$1B,$CB,$CB,$F8,$EB,$F8,$F7,$FD,$C7,$4F,$27
	.dc.b	$27,$AF,$3F,$53,$D4,$F1,$4F,$A7,$8B,$FB,$17,$25,$8F,$97,$5F,$07
	.dc.b	$0F,$CD,$47,$D0,$7E,$B6,$BE,$39,$D3,$FA,$29,$F8,$93,$E4,$6E,$1B
	.dc.b	$CB,$D1,$3F,$7B,$6F,$A2,$B8,$FF,$21,$7F,$83,$8F,$ED,$51,$C6,$3F
	.dc.b	$27,$18,$53,$CA,$38,$9C,$4E,$11,$5F,$36,$3E,$D7,$DB,$1E,$87,$DF
	.dc.b	$AD,$CB,$AC,$E7,$2F,$85,$5F,$97,$7F,$99,$5F,$37,$6F,$5C,$77,$3E
	.dc.b	$7C,$3B,$46,$B2,$F8,$73,$FC,$8F,$0F,$13,$D2,$32,$EE,$3E,$0B,$F4
	.dc.b	$9D,$E3,$FE,$B2,$C3,$D9,$3F,$C0,$67,$9F,$EB,$EF,$D5,$6F,$EA,$DD
	.dc.b	$B2,$F9,$17,$F4,$73,$F0,$B8,$3A,$67,$E2,$7F,$8E,$65,$D9,$FE,$46
	.dc.b	$DD,$38,$F4,$23,$D9,$C3,$FD,$FF,$5F,$E4,$F1,$87,$1F,$8A,$EF,$0F
	.dc.b	$E0,$91,$BE,$1C,$3F,$9D,$39,$7E,$65,$87,$F6,$27,$EE,$EF,$A3,$FE
	.dc.b	$0D,$FC,$BB,$58,$F8,$79,$BF,$CA,$D5,$AA,$FE,$BD,$1F,$27,$1F,$0C
	.dc.b	$7F,$91,$72,$7E,$45,$CA,$3F,$C3,$10,$FD,$5B,$11,$FA,$A2,$9F,$C5
	.dc.b	$F8,$2B,$FE,$65,$FB,$8F,$0E,$1D,$A3,$DF,$97,$8D,$FD,$D5,$70,$7F
	.dc.b	$DB,$87,$2C,$BF,$C8,$35,$C6,$3F,$83,$1E,$93,$E3,$1E,$CE,$7E,$EE
	.dc.b	$0F,$DB,$1E,$47,$F3,$31,$CD,$FC,$CB,$D7,$7E,$F3,$C4,$FF,$4E,$5F
	.dc.b	$B2,$37,$19,$D7,$6F,$D4,$3F,$D0,$73,$F0,$A3,$71,$F6,$30,$FF,$10
	.dc.b	$DF,$C3,$7F,$86,$36,$37,$3F,$E7,$7C,$37,$1E,$B7,$6D,$36,$7F,$36
	.dc.b	$3E,$09,$F8,$6C,$36,$D3,$63,$F4,$FF,$F4,$27,$D8,$FD,$C7,$FD,$0D
	.dc.b	$F0,$7F,$7A,$FF,$D1,$1E,$F5,$BF,$F1,$37,$E7,$D5,$F8,$39,$7F,$2E
	.dc.b	$7A,$E3,$FC,$E5,$FC,$3B,$FC,$EF,$5F,$C4,$E7,$FE,$2F,$CF,$CF,$7E
	.dc.b	$97,$EE,$5F,$F4,$E9,$FE,$73,$B7,$E7,$9F,$F3,$53,$FA,$0C,$BF,$C7
	.dc.b	$18,$F2,$AF,$F5,$A4,$63,$1F,$CB,$BF,$D5,$63,$F3,$38,$FF,$8F,$E3
	.dc.b	$84,$7F,$5C,$7C,$31,$FA,$4F,$80,$E9,$F9,$F7,$EE,$F1,$FD,$D6,$7E
	.dc.b	$0C,$3C,$0D,$C7,$F5,$0F,$93,$5F,$21,$DE,$7A,$9F,$CB,$86,$5F,$E1
	.dc.b	$2B,$FE,$0D,$3D,$F7,$FE,$0B,$C7,$6F,$F8,$24,$F3,$7F,$6C,$FA,$CF
	.dc.b	$85,$AF,$F9,$9E,$3F,$DA,$DF,$5F,$FB,$B8,$C4,$3F,$67,$C3,$FE,$58
	.dc.b	$C7,$F8,$B4,$F1,$BF,$99,$FB,$F3,$ED,$DF,$68,$FB,$A7,$EF,$AF,$A9
	.dc.b	$FE,$11,$7B,$AD,$4B,$F7,$BC,$E9,$F6,$B8,$3F,$90,$FE,$F3,$ED,$8F
	.dc.b	$F7,$5E,$1C,$72,$6B,$E2,$D3,$C3,$1F,$A0,$E8,$FF,$79,$72,$7F,$9E
	.dc.b	$7D,$EF,$E3,$11,$8C,$7F,$DE,$DE,$CE,$1F,$01,$FC,$B7,$C2,$F0,$75
	.dc.b	$FF,$BE,$EB,$84,$FE,$2D,$E5,$C7,$F8,$DE,$8F,$F8,$01,$AF,$F9,$23
	.dc.b	$E0,$BF,$93,$46,$2F,$FC,$35,$F0,$B8,$7D,$27,$E9,$8F,$BE,$62,$FF
	.dc.b	$31,$8F,$E6,$21,$F4,$1F,$E7,$5E,$3F,$E8,$B3,$98,$FF,$48,$74,$7F
	.dc.b	$4B,$8F,$4D,$3C,$5B,$1C,$FD,$0E,$B3,$E0,$D7,$F5,$FD,$DF,$82,$D3
	.dc.b	$FD,$1B,$9F,$F9,$56,$31,$F0,$9F,$D2,$9F,$C3,$FC,$27,$0C,$3F,$D7
	.dc.b	$5F,$09,$FB,$8F,$1B,$E7,$7E,$54,$E7,$7E,$F3,$FD,$4A,$73,$0F,$CA
	.dc.b	$71,$87,$D0,$9F,$F5,$3B,$CC,$FE,$28,$2F,$FD,$C2,$F0,$EA,$39,$F8
	.dc.b	$BF,$E1,$31,$F4,$7F,$30,$74,$E2,$BE,$8C,$63,$6F,$EC,$B5,$BD,$B5
	.dc.b	$C5,$D2,$9B,$FF,$14,$E8,$FE,$17,$87,$B7,$6F,$8F,$97,$F9,$46,$73
	.dc.b	$BF,$B7,$8F,$FB,$8C,$7F,$80,$66,$DF,$E3,$A3,$7F,$17,$D6,$7F,$80
	.dc.b	$CF,$C9,$F8,$CF,$E4,$7A,$23,$DF,$E3,$FB,$2B,$FC,$CD,$D5,$FE,$0E
	.dc.b	$AD,$47,$EE,$F8,$7E,$2F,$95,$FE,$02,$9F,$EC,$F3,$FD,$28,$75,$7E
	.dc.b	$B3,$1E,$B7,$FC,$EA,$B9,$E3,$F2,$B4,$FE,$6B,$DD,$F5,$37,$F7,$7E
	.dc.b	$DF,$FE,$0E,$8C,$70,$B8,$5F,$B1,$67,$9B,$F5,$B5,$F6,$F7,$FD,$89
	.dc.b	$E2,$78,$F3,$76,$9F,$BB,$97,$E1,$F9,$7F,$A2,$F8,$1F,$C0,$BA,$BE
	.dc.b	$16,$CF,$C9,$3F,$C3,$F3,$9E,$2F,$F6,$AF,$D2,$E3,$8F,$F7,$B2,$F4
	.dc.b	$6C,$7F,$A3,$EB,$FC,$CB,$E0,$7F,$AD,$6B,$97,$07,$EE,$3A,$7F,$C7
	.dc.b	$7D,$1F,$CD,$BC,$3F,$CD,$60,$FD,$0D,$FB,$17,$F8,$C2,$B5,$D1,$F9
	.dc.b	$29,$7C,$2E,$73,$FE,$1F,$4F,$52,$C7,$F0,$0B,$FC,$CF,$D2,$7F,$64
	.dc.b	$AE,$79,$7F,$A5,$34,$FF,$34,$79,$1F,$80,$8F,$D1,$FE,$82,$FC,$75
	.dc.b	$F8,$E7,$F9,$1E,$B5,$C2,$AD,$7E,$75,$9F,$0F,$F6,$33,$4F,$E6,$55
	.dc.b	$C2,$9F,$9A,$FF,$83,$23,$1E,$DF,$F8,$3B,$F6,$2F,$F0,$57,$99,$FB
	.dc.b	$DF,$AB,$7C,$07,$FB,$CF,$F0,$4F,$F2,$C5,$6B,$91,$5A,$F8,$F7,$E6
	.dc.b	$5A,$F1,$BF,$0D,$BF,$E5,$F0,$FF,$0B,$7E,$0D,$FD,$15,$E5,$87,$19
	.dc.b	$E1,$A7,$87,$47,$CD,$7F,$88,$BF,$08,$FE,$93,$FB,$CF,$F0,$8F,$1B
	.dc.b	$FC,$ED,$F2,$AF,$C2,$BF,$CD,$95,$AC,$BF,$19,$D8,$A7,$0E,$BC,$66
	.dc.b	$5F,$D5,$F0,$FE,$AB,$FB,$CD,$9C,$E3,$CB,$F4,$8F,$14,$63,$4F,$F4
	.dc.b	$37,$AD,$FE,$36,$FB,$BF,$C1,$73,$FE,$2F,$C8,$7E,$B6,$FC,$AD,$F8
	.dc.b	$07,$FA,$06,$F7,$E5,$FE,$42,$B5,$E5,$E7,$CB,$FE,$63,$BF,$EB,$AF
	.dc.b	$AB,$63,$F9,$95,$F1,$DF,$C5,$18,$CF,$C3,$F9,$DF,$F9,$2B,$F1,$27
	.dc.b	$CF,$3E,$0B,$F8,$A7,$EE,$4F,$DB,$9F,$68,$FC,$AF,$FD,$2D,$87,$0B
	.dc.b	$DD,$6B,$F1,$9C,$E3,$CF,$7F,$E0,$77,$DB,$F2,$BF,$F0,$13,$C7,$CE
	.dc.b	$69,$DF,$C2,$F2,$61,$E3,$9C,$E5,$FE,$C1,$F0,$7F,$D9,$9F,$AB,$3F
	.dc.b	$6A,$7E,$2A,$FD,$17,$F0,$BE,$5B,$F4,$E7,$D8,$FF,$AB,$70,$E1,$7B
	.dc.b	$D5,$7B,$FC,$BF,$59,$3F,$42,$FF,$0C,$61,$C3,$8F,$F8,$EB,$CB,$4E
	.dc.b	$FF,$A9,$7F,$9D,$23,$19,$74,$FC,$83,$FC,$EF,$F9,$13,$E7,$DF,$0F
	.dc.b	$FB,$BF,$C8,$FF,$4F,$F9,$EB,$F1,$8F,$F6,$15,$EE,$B5,$EB,$29,$FB
	.dc.b	$EE,$B1,$FE,$83,$BF,$FA,$5B,$0E,$DF,$29,$FE,$2F,$79,$7C,$1B,$4F
	.dc.b	$DA,$F8,$FE,$4C,$3C,$73,$9C,$DE,$7F,$FA,$3B,$F2,$C7,$E8,$0F,$D4
	.dc.b	$DF,$5D,$F2,$3D,$F7,$E2,$2F,$D5,$7F,$DB,$58,$70,$BD,$F7,$7F,$A4
	.dc.b	$E9,$FB,$CE,$B2,$E9,$C7,$FD,$23,$0E,$FF,$A9,$7F,$91,$DE,$5A,$77
	.dc.b	$F9,$DF,$25,$F8,$E3,$C5,$F3,$FF,$EA,$4F,$DD,$7F,$C2,$5F,$08,$FC
	.dc.b	$99,$F9,$AB,$F2,$87,$EC,$2F,$D4,$BF,$DE,$18,$70,$BD,$D6,$BF,$53
	.dc.b	$CF,$C1,$3F,$41,$FF,$2D,$5F,$6E,$2F,$EB,$EF,$1C,$FB,$7C,$97,$8E
	.dc.b	$FF,$93,$E5,$D3,$F1,$DF,$F5,$CF,$E9,$5F,$E6,$BF,$CA,$9E,$5F,$CD
	.dc.b	$F0,$9E,$53,$F3,$57,$AA,$F9,$FF,$F1,$2F,$03,$2E,$58,$F1,$C3,$FA
	.dc.b	$2D,$EE,$BF,$C5,$75,$FE,$79,$F7,$7F,$CE,$4F,$86,$3F,$BB,$BC,$BC
	.dc.b	$16,$7E,$29,$F0,$E5,$E0,$AD,$4B,$D0,$FF,$69,$FE,$16,$FD,$BF,$F8
	.dc.b	$57,$E6,$BF,$1F,$E7,$DF,$3A,$FC,$0F,$9E,$9F,$E3,$EC,$38,$7C,$23
	.dc.b	$0E,$17,$FE,$0E,$E0,$F5,$5A,$FB,$7C,$27,$FA,$2D,$E3,$1F,$E1,$07
	.dc.b	$9D,$F8,$7E,$2F,$CB,$A7,$93,$B7,$FC,$4B,$AF,$FB,$AB,$F5,$97,$DE
	.dc.b	$BE,$FD,$F1,$6F,$C0,$5F,$A8,$BF,$1A,$7E,$07,$EA,$E0,$65,$CB,$1E
	.dc.b	$38,$70,$BE,$9C,$5E,$AA,$D5,$FF,$89,$9F,$0F,$E5,$BF,$EA,$57,$96
	.dc.b	$7E,$1C,$DF,$85,$79,$F5,$F3,$4E,$7E,$27,$1A,$F8,$2F,$CD,$8F,$97
	.dc.b	$F7,$2F,$DD,$F9,$EF,$D7,$7F,$AF,$F2,$7D,$EF,$D9,$F5,$EE,$2B,$5F
	.dc.b	$43,$F6,$FC,$0F,$F0,$2F,$1F,$D2,$CB,$F8,$2F,$E3,$5F,$EB,$B7,$AF
	.dc.b	$27,$F9,$01,$EF,$A7,$6B,$FF,$3D,$DF,$1C,$BB,$4F,$8B,$87,$3E,$1F
	.dc.b	$E7,$30,$E3,$9F,$87,$F7,$AF,$9D,$FB,$A3,$F4,$07,$E6,$DF,$E9,$2F
	.dc.b	$23,$F8,$BF,$D3,$FF,$11,$DB,$F0,$F7,$BA,$D6,$DF,$DF,$FD,$FE,$78
	.dc.b	$F2,$E1,$FF,$2B,$DF,$5F,$4B,$FD,$AC,$F1,$FC,$07,$F9,$06,$1F,$A4
	.dc.b	$AE,$38,$F4,$8F,$4C,$7F,$A6,$2B,$86,$9E,$4F,$C7,$FF,$D8,$73,$FE
	.dc.b	$9C,$F2,$7F,$DD,$DE,$5F,$FB,$AF,$E4,$9F,$9B,$3E,$BF,$FC,$97,$F6
	.dc.b	$DF,$BE,$C3,$F5,$3F,$84,$7E,$A7,$F0,$EF,$EA,$97,$BF,$A3,$EB,$5A
	.dc.b	$FC,$07,$83,$FE,$CF,$3C,$17,$FB,$17,$38,$75,$D3,$C7,$8F,$4F,$C4
	.dc.b	$F8,$57,$83,$5F,$36,$6F,$BA,$CE,$1C,$36,$F4,$ED,$FE,$1A,$CF,$D1
	.dc.b	$F9,$4B,$F3,$BF,$F2,$6F,$ED,$7F,$E7,$5F,$C9,$DF,$BA,$3F,$56,$F4
	.dc.b	$DE,$FE,$15,$7B,$AF,$E1,$D8,$F4,$C7,$FC,$AF,$C1,$EA,$3F,$91,$E1
	.dc.b	$DF,$C1,$F8,$31,$D2,$7C,$3F,$81,$34,$F5,$4E,$73,$EF,$39,$FC,$17
	.dc.b	$E0,$FB,$EF,$C5,$1F,$B9,$3F,$36,$7E,$DB,$FE,$62,$F5,$9F,$B3,$FD
	.dc.b	$2C,$38,$5E,$EB,$5F,$9D,$E7,$1F,$EF,$5A,$FA,$4F,$F6,$FD,$F0,$70
	.dc.b	$7F,$2D,$8F,$0E,$5E,$0F,$94,$FF,$0D,$CF,$3C,$7B,$47,$F0,$68,$C4
	.dc.b	$EF,$3F,$F3,$17,$1F,$E2,$FE,$15,$F8,$FF,$F7,$8F,$98,$7E,$35,$FE
	.dc.b	$9F,$FB,$3F,$F1,$55,$EE,$B5,$39,$FD,$23,$F1,$F8,$72,$F9,$EF,$9F
	.dc.b	$7B,$FC,$3E,$EB,$FD,$5A,$B5,$9F,$6F,$BA,$7E,$3B,$F8,$E9,$CE,$7F
	.dc.b	$EA,$FF,$7B,$1A,$F8,$2F,$FC,$8B,$F6,$9E,$08,$C6,$EF,$D9,$BF,$43
	.dc.b	$E2,$3F,$48,$FF,$0F,$7E,$0A,$FC,$53,$FB,$47,$E4,$EF,$D9,$37,$5A
	.dc.b	$9C,$FD,$86,$3C,$DF,$4F,$20,$FD,$A5,$7F,$8A,$5F,$AB,$97,$7F,$B0
	.dc.b	$67,$DF,$E0,$FE,$AF,$47,$EE,$6B,$53,$9F,$8E,$FD,$6C,$63,$4F,$2F
	.dc.b	$EB,$BF,$BA,$FD,$DF,$FC,$01,$FA,$BB,$F7,$B7,$EC,$EF,$7F,$FA,$BF
	.dc.b	$BF,$BF,$A2,$5E,$EB,$5F,$28,$FA,$FD,$7E,$99,$EC,$E4,$69,$D9,$FC
	.dc.b	$7B,$E0,$D3,$BE,$8F,$C6,$FF,$81,$CE,$72,$F1,$7E,$1D,$8D,$7C,$38
	.dc.b	$70,$D3,$FB,$1F,$27,$F1,$9F,$D1,$5F,$BF,$FF,$AC,$FE,$A7,$FC,$07
	.dc.b	$F8,$1B,$F2,$E7,$E4,$7F,$C6,$AF,$7F,$8A,$AF,$75,$FE,$0D,$C7,$A6
	.dc.b	$3F,$EB,$DE,$2F,$D3,$E1,$DB,$E5,$7D,$CF,$2C,$3B,$FD,$27,$F1,$C8
	.dc.b	$C7,$C3,$35,$7E,$5D,$9C,$BC,$D3,$9D,$1F,$8A,$F9,$67,$37,$E9,$FC
	.dc.b	$5B,$1B,$F9,$F0,$E1,$F8,$06,$35,$F4,$FE,$DF,$FC,$DB,$EA,$9F,$9E
	.dc.b	$7D,$8F,$D2,$9F,$A8,$FF,$8A,$7E,$33,$FC,$7F,$CD,$FD,$06,$F7,$E1
	.dc.b	$2F,$7F,$4D,$FC,$AB,$E4,$3A,$65,$F8,$2E,$9F,$E1,$38,$76,$F0,$FF
	.dc.b	$00,$F2,$C3,$8F,$9D,$FC,$FA,$31,$AF,$F8,$E7,$E2,$31,$8F,$96,$3C
	.dc.b	$9F,$8A,$F6,$D7,$C5,$1E,$7D,$83,$F1,$1F,$F2,$AF,$EA,$FF,$E6,$7F
	.dc.b	$C8,$1F,$99,$3F,$45,$FE,$57,$FB,$1B,$F7,$0E,$38,$70,$DD,$87,$0B
	.dc.b	$DF,$EA,$1A,$E4,$FD,$A6,$7F,$E7,$3F,$C8,$FF,$88,$E3,$C7,$0B,$5A
	.dc.b	$C7,$9E,$9E,$1F,$8D,$79,$9F,$BD,$DF,$D1,$7B,$C3,$C1,$5F,$A1,$8F
	.dc.b	$00,$CF,$D9,$87,$3D,$DF,$A6,$6B,$F7,$DF,$F5,$2F,$B4,$FD,$BD,$FA
	.dc.b	$F7,$E7,$BF,$32,$7E,$53,$BF,$C3,$BF,$CF,$95,$AF,$45,$5F,$E0,$3E
	.dc.b	$1F,$2E,$EC,$FA,$BF,$DE,$B9,$3F,$D0,$70,$FC,$EF,$FB,$0C,$F5,$D7
	.dc.b	$C5,$AF,$EB,$BD,$8F,$81,$87,$0D,$7C,$33,$FB,$2E,$5C,$E1,$FC,$87
	.dc.b	$2F,$5E,$DE,$5D,$5F,$B0,$F5,$8C,$7D,$37,$C6,$F9,$4F,$F2,$BF,$E2
	.dc.b	$5F,$C9,$3E,$11,$FC,$01,$FE,$1E,$F1,$F5,$F0,$AA,$D4,$FF,$88,$BF
	.dc.b	$2A,$CF,$E1,$5F,$6B,$76,$5D,$1F,$80,$7F,$35,$7B,$E5,$FC,$CA,$DF
	.dc.b	$A9,$7A,$AB,$8E,$5D,$EB,$4E,$6F,$8C,$79,$BE,$AB,$F8,$87,$4F,$D0
	.dc.b	$9C,$E3,$FE,$36,$FC,$2F,$F0,$7F,$AA,$7E,$CF,$F2,$37,$EB,$BF,$DB
	.dc.b	$32,$FE,$31,$F3,$3A,$E2,$56,$BE,$5B,$FA,$3F,$B0,$FD,$CB,$53,$EF
	.dc.b	$CB,$F7,$DF,$A5,$DC,$38,$E7,$D7,$F0,$0F,$65,$78,$63,$C1,$F8,$16
	.dc.b	$2B,$79,$F6,$FB,$4E,$B5,$E5,$D1,$F8,$AF,$E7,$3D,$97,$ED,$6F,$CA
	.dc.b	$5E,$9B,$F7,$7F,$6F,$E9,$8E,$6E,$17,$BA,$D6,$6A,$7E,$5D,$AF,$82
	.dc.b	$FB,$59,$F6,$CE,$E3,$C7,$3F,$89,$DD,$D2,$7C,$31,$F3,$3F,$10,$FB
	.dc.b	$57,$C7,$4E,$F9,$FF,$AD,$BB,$7F,$4B,$39,$BF,$3F,$ED,$7C,$EF,$7F
	.dc.b	$AB,$FE,$EB,$E4,$9F,$AE,$7B,$BE,$99,$F8,$33,$F5,$9F,$E2,$3F,$8F
	.dc.b	$EE,$B5,$39,$F7,$DF,$90,$E7,$DF,$FC,$B6,$5D,$34,$B8,$70,$C9,$F3
	.dc.b	$1F,$93,$AD,$69,$E0,$97,$E3,$19,$BE,$19,$BF,$30,$F1,$64,$C3,$84
	.dc.b	$BE,$3B,$F8,$AF,$E2,$2F,$CC,$9F,$66,$FC,$C9,$F8,$03,$EA,$FE,$D7
	.dc.b	$C2,$70,$BD,$D6,$A7,$EF,$7C,$0E,$70,$FD,$96,$3D,$7F,$0D,$B8,$70
	.dc.b	$CB,$A7,$9D,$FA,$FA,$F1,$5F,$86,$7C,$1E,$6E,$6F,$86,$5F,$8B,$F6
	.dc.b	$BB,$69,$E1,$F2,$7F,$AF,$78,$3F,$C8,$3F,$76,$FA,$9E,$3F,$F9,$BB
	.dc.b	$F7,$77,$E1,$5F,$93,$FC,$3F,$F9,$AD,$6B,$F0,$45,$6B,$17,$E1,$FA
	.dc.b	$EE,$FC,$0F,$99,$FC,$28,$61,$D9,$C3,$CE,$FB,$F2,$FF,$0B,$8C,$67
	.dc.b	$FE,$C3,$D7,$CD,$87,$EA,$B3,$F0,$CF,$E9,$FF,$0B,$F0,$6D,$E9,$8F
	.dc.b	$D5,$7D,$26,$2B,$7B,$7F,$0A,$C7,$9D,$EE,$3F,$D2,$30,$FC,$C7,$FC
	.dc.b	$A9,$B7,$E3,$3E,$9F,$FD,$23,$FA,$9B,$F4,$C7,$E7,$EF,$CB,$5F,$8D
	.dc.b	$FF,$76,$C7,$8F,$95,$FE,$31,$D9,$8F,$1C,$38,$3B,$CF,$FC,$C5,$5B
	.dc.b	$7F,$D9,$37,$CB,$3F,$F1,$94,$7F,$B5,$F4,$F3,$3C,$23,$CB,$A7,$93
	.dc.b	$E4,$FF,$B8,$5F,$1C,$FC,$33,$E8,$BC,$9C,$63,$39,$FA,$FF,$67,$F2
	.dc.b	$4F,$C1,$9C,$EC,$FC,$1B,$3D,$CF,$D1,$5F,$B9,$BF,$4F,$FE,$02,$7F
	.dc.b	$31,$FC,$AF,$73,$A8,$CF,$C4,$70,$E2,$F5,$E9,$FA,$DA,$5F,$FA,$83
	.dc.b	$83,$E9,$E5,$DF,$2C,$5F,$FA,$26,$31,$97,$83,$F0,$FF,$6C,$F3,$AF
	.dc.b	$27,$A3,$E5,$FD,$1F,$38,$70,$8F,$46,$BF,$E0,$1D,$7C,$73,$9C,$BF
	.dc.b	$A5,$7E,$09,$F1,$3E,$53,$FC,$FB,$FB,$13,$F6,$D7,$C4,$BF,$31,$7E
	.dc.b	$89,$FD,$0B,$3F,$6E,$1C,$2F,$75,$AB,$E9,$F8,$03,$1F,$6F,$0F,$F7
	.dc.b	$98,$7E,$2B,$8C,$79,$5F,$7E,$DF,$7C,$63,$1F,$CC,$E7,$E0,$C3,$A6
	.dc.b	$7E,$0D,$3F,$73,$39,$D3,$D5,$97,$A6,$5F,$A1,$79,$27,$38,$FB,$3E
	.dc.b	$47,$18,$F3,$E4,$FC,$3F,$C7,$3E,$93,$F5,$5C,$0F,$C6,$DF,$A7,$3D
	.dc.b	$27,$CD,$F0,$BD,$D6,$A7,$FC,$1B,$C3,$F8,$4C,$7A,$7E,$79,$F1,$1E
	.dc.b	$DD,$D7,$0E,$39,$75,$E7,$E9,$3F,$23,$2F,$0F,$B1,$F8,$4C,$38,$EB
	.dc.b	$E5,$C3,$C9,$FA,$1F,$08,$C6,$AF,$C8,$FF,$47,$F8,$CB,$F1,$87,$D2
	.dc.b	$3F,$5B,$FE,$FB,$E7,$3F,$5A,$FF,$59,$F2,$7E,$73,$0E,$17,$BF,$53
	.dc.b	$FD,$69,$5F,$0E,$B5,$8F,$4E,$3F,$E0,$27,$FB,$97,$0F,$07,$EA,$FF
	.dc.b	$84,$79,$E7,$E0,$FC,$6F,$F3,$F1,$8C,$BF,$D8,$18,$E3,$1F,$3D,$F9
	.dc.b	$B2,$7E,$ED,$EC,$9C,$DF,$93,$F2,$CF,$E0,$3F,$9B,$3F,$48,$FF,$36
	.dc.b	$FE,$FD,$FE,$65,$F1,$1F,$90,$FD,$AD,$7F,$73,$7B,$FB,$45,$EE,$BE
	.dc.b	$DF,$6F,$F2,$CA,$E9,$CB,$F9,$E7,$87,$27,$C0,$0F,$3B,$E5,$1E,$4C
	.dc.b	$FC,$7F,$67,$FA,$EF,$AE,$FD,$FD,$EE,$3F,$63,$93,$AF,$86,$9E,$49
	.dc.b	$CE,$18,$CB,$CD,$F0,$36,$78,$FE,$57,$FC,$8B,$F3,$57,$E3,$CF,$D7
	.dc.b	$BF,$B0,$EE,$FD,$FD,$F9,$17,$F8,$F3,$93,$F8,$DE,$1C,$2F,$75,$AF
	.dc.b	$9E,$E9,$8B,$F6,$0F,$DF,$DE,$F2,$F8,$4F,$1F,$8E,$47,$7F,$03,$F8
	.dc.b	$EC,$E6,$3C,$5F,$92,$E3,$0F,$35,$F0,$CB,$CB,$F8,$D7,$83,$4F,$1E
	.dc.b	$DF,$D6,$B6,$FE,$63,$FC,$00,$FD,$8D,$FB,$D3,$F4,$AF,$F0,$E7,$E1
	.dc.b	$4F,$D8,$3F,$7D,$87,$0B,$DD,$6A,$7F,$B3,$4E,$71,$FE,$37,$8F,$5C
	.dc.b	$6E,$1C,$3E,$69,$93,$ED,$BF,$9C,$DE,$F3,$FF,$16,$7E,$0F,$FC,$73
	.dc.b	$C3,$FE,$EB,$39,$C3,$8F,$E5,$DF,$DE,$75,$F2,$69,$E9,$9C,$FA,$FF
	.dc.b	$D7,$37,$74,$3F,$43,$7E,$70,$FC,$81,$FB,$DF,$F0,$1F,$2F,$85,$EE
	.dc.b	$B5,$3F,$C1,$7C,$EC,$E1,$FC,$E3,$1E,$BF,$99,$DC,$38,$65,$D3,$F0
	.dc.b	$5F,$EA,$B5,$E2,$BF,$0C,$F8,$3E,$E7,$37,$C3,$2F,$E2,$5F,$A6,$F6
	.dc.b	$D3,$C3,$EC,$3D,$DC,$63,$E8,$87,$D2,$E3,$0F,$E3,$5F,$80,$39,$5C
	.dc.b	$7F,$E1,$EF,$B2,$7E,$96,$FC,$79,$F8,$37,$F9,$77,$D3,$7E,$99,$FD
	.dc.b	$57,$1E,$3E,$5F,$E5,$31,$E3,$87,$0F,$AB,$FD,$85,$FF,$58,$BD,$DF
	.dc.b	$F9,$CB,$27,$E3,$FF,$E2,$E8,$FE,$AF,$A7,$8E,$7C,$51,$93,$D1,$19
	.dc.b	$C7,$B6,$6F,$C7,$7F,$4B,$8F,$6D,$3C,$7A,$FA,$7E,$2F,$38,$F1,$8F
	.dc.b	$66,$AF,$A0,$EB,$A7,$AF,$5F,$85,$5A,$FC,$87,$32,$FC,$E1,$FC,$01
	.dc.b	$FD,$77,$F7,$2F,$E8,$5F,$79,$F3,$3D,$7D,$D6,$BE,$3D,$5A,$DB,$F8
	.dc.b	$57,$0F,$C4,$AD,$65,$EF,$8C,$64,$FC,$47,$E1,$4F,$48,$F0,$7C,$4F
	.dc.b	$FA,$66,$3D,$A7,$F0,$6B,$F0,$63,$E4,$F9,$BC,$CF,$A6,$F8,$EB,$F8
	.dc.b	$2F,$CB,$3A,$6B,$E6,$CB,$FD,$49,$AF,$EF,$F5,$FB,$3F,$46,$FC,$43
	.dc.b	$F0,$DE,$7F,$F6,$1E,$87,$E8,$2F,$D9,$FE,$3D,$EF,$B5,$7B,$A7,$EF
	.dc.b	$5A,$C7,$F0,$7C,$3F,$07,$4F,$F3,$8E,$3C,$63,$07,$F2,$27,$A6,$1E
	.dc.b	$1F,$B9,$F6,$C6,$36,$FF,$5F,$FE,$AF,$8C,$7C,$F1,$FE,$C2,$CD,$F4
	.dc.b	$7E,$1D,$7C,$B3,$9F,$DC,$71,$9F,$9F,$F4,$BB,$3C,$7F,$56,$FC,$1F
	.dc.b	$82,$FD,$FF,$FD,$09,$EC,$BF,$6D,$7E,$40,$FD,$B7,$F8,$67,$4F,$E8
	.dc.b	$B0,$E1,$F6,$FF,$CB,$AF,$74,$FD,$3B,$AE,$5F,$E8,$FE,$7F,$82,$7E
	.dc.b	$C9,$FF,$17,$38,$C3,$9E,$7E,$1C,$FF,$BA,$CF,$2C,$BB,$43,$F5,$9F
	.dc.b	$C1,$DF,$0C,$FC,$3F,$A9,$EF,$1E,$5A,$3F,$38,$FD,$0F,$E7,$CF,$91
	.dc.b	$FF,$1D,$FE,$CF,$FE,$5F,$F8,$E7,$EF,$CF,$85,$DD,$6B,$EB,$95,$AF
	.dc.b	$C7,$B3,$F5,$5F,$9C,$E7,$FE,$0B,$9F,$5F,$DF,$71,$E2,$FF,$AC,$A3
	.dc.b	$C1,$4D,$7C,$1E,$27,$E7,$27,$31,$E3,$FC,$97,$1A,$F8,$74,$F2,$FC
	.dc.b	$1F,$82,$BD,$1C,$1F,$CA,$75,$FE,$2B,$FA,$E7,$FA,$DB,$F4,$57,$ED
	.dc.b	$6F,$49,$FB,$1B,$F6,$A7,$A1,$FC,$13,$0E,$17,$BA,$D7,$EC,$39,$FC
	.dc.b	$27,$E2,$F8,$7F,$6A,$FB,$36,$19,$F8,$33,$EF,$F8,$97,$EE,$2F,$AC
	.dc.b	$78,$BE,$83,$17,$C3,$3F,$06,$AF,$C0,$3C,$19,$73,$D7,$D1,$F8,$1E
	.dc.b	$2F,$7F,$4B,$73,$F6,$E2,$FD,$CF,$F8,$FF,$D6,$FF,$47,$FB,$DF,$F5
	.dc.b	$9F,$9E,$FD,$B7,$EF,$7C,$96,$B9,$FF,$D3,$1F,$83,$7F,$98,$E7,$F0
	.dc.b	$5F,$8B,$E7,$EE,$FE,$DF,$47,$F8,$6B,$3E,$8F,$E7,$1E,$D7,$C7,$1F
	.dc.b	$E2,$F1,$8C,$BC,$1F,$77,$8C,$FB,$C7,$8F,$C9,$7C,$97,$8F,$6F,$3E
	.dc.b	$DF,$93,$F9,$4F,$E7,$3F,$B9,$3C,$67,$E8,$EF,$2F,$FD,$B3,$8B,$F5
	.dc.b	$17,$E4,$1C,$2F,$75,$A9,$FF,$2E,$ED,$FD,$73,$1E,$98,$FB,$B8,$D1
	.dc.b	$DF,$C9,$C3,$AC,$3B,$FE,$67,$FC,$D2,$31,$DB,$EA,$FD,$A7,$1A,$78
	.dc.b	$23,$DB,$A3,$F6,$5F,$26,$9E,$99,$CF,$E8,$3F,$DF,$7E,$57,$FC,$8F
	.dc.b	$B7,$F7,$AF,$CC,$9F,$4C,$FD,$25,$7F,$E2,$0F,$DD,$7F,$6B,$C7,$FE
	.dc.b	$02,$C3,$FC,$9F,$F3,$2C,$78,$E1,$C3,$E8,$D8,$7F,$81,$B0,$E1,$87
	.dc.b	$7C,$BF,$79,$9F,$8B,$2E,$5F,$BC,$E2,$BA,$E7,$D3,$4F,$F1,$8F,$DF
	.dc.b	$7F,$7A,$C3,$84,$E7,$47,$81,$9C,$38,$E7,$EB,$C7,$D5,$FA,$3F,$C1
	.dc.b	$9C,$EF,$E7,$DF,$FC,$BD,$9F,$4C,$38,$65,$EA,$F4,$B8,$61,$E9,$C3
	.dc.b	$87,$A7,$8F,$2D,$F2,$CF,$92,$7E,$9C,$FB,$97,$EE,$AF,$5B,$FC,$63
	.dc.b	$D1,$F1,$31,$E3,$87,$0B,$DD,$3F,$27,$ED,$93,$F4,$9F,$8D,$B3,$1E
	.dc.b	$2F,$F8,$17,$8B,$EE,$C7,$4D,$7C,$5A,$F8,$6B,$9E,$9D,$E1,$ED,$7C
	.dc.b	$99,$E9,$97,$AE,$F8,$E8,$FC,$E3,$8C,$63,$C8,$F8,$BF,$96,$3F,$59
	.dc.b	$FF,$7E,$7D,$73,$F6,$E7,$E4,$CF,$C2,$BF,$C4,$DC,$DF,$84,$C3,$87
	.dc.b	$D1,$7F,$9B,$6F,$75,$FE,$07,$AD,$63,$FE,$1F,$DF,$DC,$7F,$A8,$30
	.dc.b	$E2,$E0,$3D,$32,$F1,$64,$FC,$8F,$DD,$39,$CB,$CB,$FA,$7E,$31,$F4
	.dc.b	$C7,$8B,$E7,$7D,$B1,$8C,$3D,$5F,$47,$F8,$4F,$C8,$1F,$B6,$2F,$FA
	.dc.b	$DF,$EE,$1F,$E3,$CF,$AA,$7C,$86,$EB,$5F,$1E,$AD,$7E,$43,$9F,$A5
	.dc.b	$FB,$B9,$74,$CA,$E1,$C2,$FF,$C8,$FF,$75,$F5,$EB,$58,$F3,$CD,$F5
	.dc.b	$39,$C3,$84,$7F,$93,$BF,$1E,$F1,$65,$E6,$8F,$F2,$8F,$E7,$F8,$C3
	.dc.b	$94,$7F,$86,$7A,$3F,$73,$97,$6D,$3D,$9E,$3F,$E9,$3F,$3A,$7E,$74
	.dc.b	$F4,$1F,$5F,$FE,$71,$F4,$9F,$8B,$BF,$65,$E1,$7B,$AD,$4B,$F6,$DC
	.dc.b	$E1,$F9,$EC,$7A,$FA,$17,$0E,$1F,$75,$FC,$5B,$CB,$FC,$C5,$78,$AF
	.dc.b	$C3,$3F,$C6,$BC,$AC,$DF,$0D,$7C,$7F,$65,$DB,$5F,$0F,$D4,$7E,$FB
	.dc.b	$E2,$1F,$89,$3F,$76,$7E,$DF,$FE,$FA,$FC,$0D,$F2,$BF,$EF,$CE,$9F
	.dc.b	$E3,$B8,$70,$BD,$FE,$21,$FE,$CA,$A7,$ED,$1D,$72,$7E,$C7,$FB,$AC
	.dc.b	$38,$7A,$1F,$99,$79,$69,$DD,$FD,$A1,$8C,$7F,$DA,$BF,$90,$E3,$0F
	.dc.b	$35,$79,$72,$7E,$CD,$DF,$4F,$1E,$DE,$87,$F8,$DF,$C1,$FF,$8C,$FF
	.dc.b	$3B,$7E,$83,$FD,$0B,$E3,$1F,$BB,$7F,$44,$E3,$FE,$73,$7B,$AD,$4B
	.dc.b	$E2,$F9,$FC,$CB,$1F,$E2,$B7,$FE,$3F,$FC,$FA,$CE,$70,$ED,$C3,$FC
	.dc.b	$3E,$BA,$7C,$57,$F7,$0E,$4C,$5E,$E3,$FB,$9E,$6F,$D0,$7B,$E7,$E8
	.dc.b	$9C,$F5,$63,$3F,$2F,$EC,$76,$78,$FE,$E9,$FC,$EB,$F2,$07,$E7,$DF
	.dc.b	$EC,$EF,$CE,$DF,$99,$BF,$51,$FF,$2B,$FE,$09,$BA,$D4,$E7,$EC,$9F
	.dc.b	$6B,$9F,$C4,$7F,$99,$E5,$D3,$F0,$8B,$87,$0C,$BF,$80,$F8,$7B,$D3
	.dc.b	$BC,$72,$8C,$67,$D7,$E0,$2C,$FF,$9A,$3E,$03,$FD,$E9,$F9,$4F,$F7
	.dc.b	$8E,$56,$0F,$DB,$3F,$D8,$5F,$42,$FC,$E5,$F9,$9F,$F8,$CF,$C2,$63
	.dc.b	$C7,$0E,$1F,$88,$7F,$87,$EF,$7F,$87,$F6,$9E,$B9,$F6,$CF,$FC,$E5
	.dc.b	$87,$1F,$D4,$6D,$68,$E7,$A7,$87,$E9,$FF,$DD,$F5,$DE,$FC,$B9,$3E
	.dc.b	$A7,$38,$71,$9C,$ED,$FE,$08,$DB,$CF,$39,$DB,$CD,$2F,$D0,$BC,$D2
	.dc.b	$FA,$77,$EF,$2F,$C9,$5F,$A7,$3E,$C3,$F1,$1F,$90,$3F,$6A,$E3,$87
	.dc.b	$0B,$DD,$6B,$F7,$7C,$E3,$FD,$F3,$2F,$F7,$87,$17,$F8,$A3,$0E,$11
	.dc.b	$83,$E9,$BC,$F3,$F0,$79,$2D,$3D,$F5,$E7,$C3,$CD,$5A,$FC,$87,$38
	.dc.b	$71,$CF,$CF,$F9,$0F,$8B,$5F,$2E,$7C,$FF,$1F,$C6,$3D,$B4,$FE,$2F
	.dc.b	$C7,$F2,$35,$ED,$87,$09,$78,$7F,$23,$C5,$7E,$17,$FC,$1B,$F5,$A7
	.dc.b	$E8,$DF,$E0,$FF,$D0,$3F,$D3,$79,$F3,$BD,$FB,$6B,$DF,$D7,$2F,$FD
	.dc.b	$19,$7B,$C7,$F8,$8E,$BF,$14,$FD,$3B,$FC,$AD,$C6,$3D,$30,$E7,$F5
	.dc.b	$7F,$1A,$31,$9F,$F8,$87,$F6,$2C,$61,$E7,$8F,$37,$D6,$FA,$E3,$1A
	.dc.b	$78,$FF,$03,$FC,$2F,$E3,$2F,$9C,$7E,$50,$FD,$47,$FD,$A5,$F8,$E3
	.dc.b	$F7,$5F,$F3,$47,$17,$C9,$C3,$85,$EF,$DE,$FF,$78,$57,$F9,$92,$B5
	.dc.b	$8F,$4E,$3F,$76,$1D,$FE,$53,$F4,$B1,$FF,$00,$67,$C3,$9F,$83,$C9
	.dc.b	$FC,$1D,$F6,$8F,$1F,$97,$8A,$DE,$7E,$0F,$D8,$FD,$71,$8D,$BD,$1F
	.dc.b	$8E,$7E,$8B,$F5,$2F,$F6,$97,$C5,$FB,$7F,$7B,$7E,$A6,$FD,$2F,$FC
	.dc.b	$95,$C5,$F8,$5B,$DE,$CB,$DD,$3F,$0E,$E9,$F8,$87,$F5,$EE,$3F,$B5
	.dc.b	$87,$FB,$7F,$F2,$AF,$6B,$CB,$0E,$FF,$37,$F9,$E8,$C6,$9F,$E4,$CB
	.dc.b	$E7,$AF,$96,$3C,$9F,$A7,$63,$5F,$14,$79,$DC,$63,$E4,$FD,$82,$CF
	.dc.b	$AF,$F0,$5F,$E0,$3F,$1A,$F9,$BF,$B5,$E3,$7F,$1E,$FA,$C7,$E4,$8E
	.dc.b	$9D,$D6,$BD,$6F,$F1,$B4,$BA,$33,$E0,$7D,$8C,$BA,$7D,$4D,$C3,$86
	.dc.b	$5F,$E7,$DE,$3D,$E9,$DF,$EB,$FF,$7B,$87,$8F,$1F,$D3,$69,$DB,$F6
	.dc.b	$2D,$BE,$0F,$84,$FE,$33,$F9,$13,$E8,$9F,$A8,$7E,$AF,$EA,$3F,$C5
	.dc.b	$5F,$10,$FC,$E7,$75,$AF,$53,$FC,$EF,$F5,$3F,$DB,$FE,$57,$38,$FE
	.dc.b	$43,$2E,$9C,$37,$0E,$19,$FF,$8D,$7E,$F3,$F8,$9D,$78,$6F,$C1,$8B
	.dc.b	$EA,$F3,$87,$08,$FE,$F9,$FB,$7F,$6D,$3C,$3B,$7F,$87,$3C,$9F,$E2
	.dc.b	$1F,$A4,$BF,$63,$7E,$F0,$FC,$73,$F6,$1F,$9A,$7F,$91,$BF,$06,$FE
	.dc.b	$73,$7B,$AD,$78,$EA,$FF,$25,$FE,$11,$AC,$79,$F7,$7F,$0B,$C3,$B7
	.dc.b	$C0,$7E,$B9,$E5,$9F,$6F,$C4,$3F,$97,$C6,$32,$E9,$F8,$8E,$31,$F2
	.dc.b	$DF,$93,$ED,$3E,$8E,$FE,$A9,$CD,$F8,$E7,$F9,$04,$E7,$6F,$3E,$DF
	.dc.b	$E6,$0F,$DF,$BC,$D1,$8A,$7E,$2F,$FD,$5B,$F3,$07,$E5,$CF,$DB,$7F
	.dc.b	$21,$F8,$C3,$F2,$17,$DE,$3F,$7A,$E1,$7B,$F3,$FF,$DF,$95,$AF,$84
	.dc.b	$CF,$C2,$7F,$0E,$CF,$AE,$77,$0E,$19,$F5,$CF,$FA,$8D,$6B,$2E,$78
	.dc.b	$E2,$F7,$97,$8B,$F4,$CF,$0E,$5E,$68,$F2,$FD,$9B,$8F,$FC,$1D,$D7
	.dc.b	$FE,$12,$F0,$DF,$93,$BF,$2E,$7E,$58,$F9,$C7,$C6,$3E,$C1,$F8,$DF
	.dc.b	$3F,$A6,$FF,$0D,$E7,$F9,$1C,$B9,$63,$C7,$0E,$1F,$AF,$EE,$BC,$15
	.dc.b	$DD,$E2,$7F,$89,$35,$F2,$BC,$FE,$03,$2E,$71,$EF,$B7,$A3,$6F,$3E
	.dc.b	$DE,$6F,$96,$D5,$F6,$9C,$E9,$FE,$3A,$CB,$AE,$1C,$75,$F3,$6E,$FA
	.dc.b	$9F,$36,$3E,$BC,$78,$CF,$F2,$FC,$3C,$BF,$9B,$7F,$2F,$FC,$95,$F8
	.dc.b	$2F,$F9,$A7,$F1,$67,$E2,$CF,$CE,$39,$DE,$EB,$52,$FD,$0F,$3F,$25
	.dc.b	$FC,$DF,$5F,$87,$9F,$68,$67,$F0,$A7,$6F,$83,$D3,$FF,$73,$E1,$F0
	.dc.b	$25,$97,$79,$FF,$75,$FB,$59,$8C,$6B,$E6,$F0,$B9,$EB,$E4,$FD,$8A
	.dc.b	$E1,$FE,$F2,$FC,$CB,$86,$BE,$2A,$F4,$FE,$A1,$F1,$BE,$D3,$FD,$47
	.dc.b	$9B,$F4,$C7,$E7,$FF,$E1,$8F,$DA,$BF,$CE,$BC,$7F,$72,$F7,$5A,$FA
	.dc.b	$75,$3E,$93,$A7,$E1,$CF,$C6,$FE,$06,$3C,$63,$DB,$CB,$F8,$48,$C6
	.dc.b	$9E,$1D,$3F,$C4,$FF,$45,$F0,$F5,$EF,$1F,$E2,$AF,$47,$18,$FC,$6D
	.dc.b	$BC,$5A,$F7,$CB,$9C,$63,$2F,$C1,$FC,$23,$F5,$6F,$F7,$1E,$CF,$DA
	.dc.b	$DF,$86,$7D,$FF,$A6,$DD,$6A,$73,$E0,$3F,$53,$CF,$E4,$3E,$CF,$2F
	.dc.b	$F5,$39,$F5,$FC,$5B,$1C,$DF,$E8,$68,$C6,$BD,$F5,$FF,$33,$4F,$6A
	.dc.b	$F8,$D9,$F8,$BC,$CF,$C1,$61,$C2,$3C,$3F,$8C,$EF,$1E,$5A,$FF,$BD
	.dc.b	$F8,$BF,$5F,$FB,$27,$FA,$AB,$C7,$FF,$39,$7E,$BA,$FC,$9D,$E1,$7F
	.dc.b	$B2,$F8,$7F,$7F,$7B,$FB,$55,$EE,$9F,$90,$FE,$AA,$BF,$D6,$B9,$FB
	.dc.b	$38,$91,$DD,$FB,$58,$F4,$FC,$1F,$F4,$99,$74,$CF,$F5,$75,$CE,$73
	.dc.b	$9F,$97,$CA,$FD,$06,$3E,$38,$F1,$7E,$E3,$BD,$BD,$78,$FA,$A7,$19
	.dc.b	$75,$DB,$FB,$B6,$FF,$0B,$3E,$B7,$F0,$3E,$63,$F8,$1F,$E9,$8F,$B7
	.dc.b	$7E,$8A,$FC,$99,$FB,$43,$F1,$E7,$D8,$3F,$7C,$E3,$87,$0B,$DD,$6B
	.dc.b	$F6,$9C,$E3,$F1,$72,$ED,$E2,$DC,$38,$67,$D7,$E2,$3F,$7D,$5E,$3B
	.dc.b	$F1,$4F,$87,$E2,$B3,$7C,$32,$F1,$FD,$67,$6D,$3C,$3B,$3F,$31,$FE
	.dc.b	$3F,$FA,$2F,$B7,$F1,$17,$EB,$0F,$CC,$9F,$93,$BF,$6D,$FD,$17,$4F
	.dc.b	$F3,$6C,$38,$5E,$EB,$5E,$57,$4D,$FF,$DD,$3E,$17,$F1,$CF,$0F,$FA
	.dc.b	$3E,$1F,$2F,$D5,$FC,$23,$CB,$4E,$FF,$9E,$FF,$1B,$8C,$65,$FE,$53
	.dc.b	$D3,$C1,$93,$6F,$3E,$1E,$6D,$BC,$D5,$A9,$CD,$FA,$7F,$72,$FE,$7B
	.dc.b	$F0,$AF,$DF,$1F,$BD,$7E,$7B,$E2,$1F,$9A,$66,$37,$3A,$F1,$7F,$A3
	.dc.b	$4F,$1F,$DE,$F3,$5F,$9D,$C3,$A7,$87,$70,$E1,$8F,$4E,$DF,$E6,$D5
	.dc.b	$D7,$2E,$7E,$0E,$2F,$79,$F9,$7B,$7A,$D7,$82,$7E,$C7,$07,$F6,$6C
	.dc.b	$39,$65,$E8,$97,$E6,$FC,$23,$1F,$8C,$FF,$80,$7F,$48,$FE,$EF,$F9
	.dc.b	$EB,$F5,$1F,$F2,$CF,$8B,$FE,$C3,$FB,$17,$E4,$8F,$8D,$E1,$E0,$86
	.dc.b	$1C,$6B,$F4,$7F,$61,$FE,$62,$B7,$C7,$70,$BD,$E5,$FE,$D1,$7F,$9C
	.dc.b	$67,$E2,$8F,$F3,$4C,$60,$ED,$8F,$5C,$FC,$9D,$9F,$E0,$5B,$E1,$9F
	.dc.b	$80,$7C,$B6,$70,$E3,$8F,$B3,$4F,$27,$C6,$78,$F3,$E9,$1F,$E7,$8F
	.dc.b	$DD,$78,$5B,$D3,$FF,$35,$7D,$73,$F5,$9F,$D0,$FA,$9F,$C0,$7F,$0C
	.dc.b	$F6,$56,$A7,$3E,$3F,$E2,$3A,$DC,$3F,$E2,$4A,$FF,$3A,$79,$1F,$D6
	.dc.b	$A3,$1F,$32,$CF,$1A,$7C,$DB,$E9,$E3,$B5,$FE,$21,$3E,$0B,$E1,$A7
	.dc.b	$93,$EC,$7D,$38,$70,$BF,$E2,$DF,$B8,$6F,$2F,$56,$5C,$F6,$7E,$9D
	.dc.b	$F5,$7F,$4B,$7C,$9F,$F9,$03,$E7,$9F,$8A,$3F,$21,$7E,$6B,$FC,$73
	.dc.b	$97,$DB,$C7,$F8,$3C,$78,$FD,$0F,$F9,$1B,$0E,$16,$FD,$83,$B6,$6F
	.dc.b	$D2,$BD,$F8,$B1,$E2,$FE,$04,$C4,$F8,$F0,$F1,$7C,$87,$D6,$AE,$FA
	.dc.b	$F9,$B5,$7E,$37,$9C,$38,$47,$A3,$EC,$3C,$3B,$79,$63,$D7,$E8,$E3
	.dc.b	$0E,$5B,$BF,$27,$C6,$BE,$AF,$8C,$FC,$D7,$C8,$BF,$6E,$7E,$57,$FE
	.dc.b	$3E,$D9,$FA,$B3,$F4,$C7,$E9,$3C,$2F,$7F,$53,$F3,$AB,$5F,$86,$67
	.dc.b	$1F,$87,$97,$5F,$83,$B8,$70,$CF,$AF,$CA,$FE,$5A,$BC,$57,$E1,$C7
	.dc.b	$FB,$2F,$67,$D4,$EA,$56,$B4,$F0,$6C,$FC,$AF,$C1,$18,$FA,$2F,$DB
	.dc.b	$74,$BF,$10,$F4,$CB,$F2,$07,$E4,$0F,$D1,$BF,$C3,$FA,$F7,$C3,$87
	.dc.b	$2F,$F9,$82,$F7,$4F,$D2,$FA,$E5,$F1,$37,$FB,$D8,$77,$E6,$FC,$1B
	.dc.b	$CB,$4F,$06,$9F,$E1,$ED,$3B,$61,$E4,$CD,$D9,$FE,$90,$9C,$EB,$E1
	.dc.b	$BF,$37,$E0,$DE,$08,$C6,$DF,$9B,$AF,$8D,$FA,$F3,$F7,$DF,$D1,$FC
	.dc.b	$A3,$F6,$87,$E9,$7F,$E2,$0C,$FF,$8D,$5E,$FC,$AF,$EF,$B5,$AF,$88
	.dc.b	$E5,$97,$4C,$9E,$7F,$E7,$B0,$E1,$CB,$F9,$D7,$96,$9D,$F4,$FB,$3E
	.dc.b	$AF,$FA,$AA,$73,$9F,$8B,$BB,$F8,$AC,$62,$BF,$C6,$7E,$27,$93,$3F
	.dc.b	$4D,$F8,$FC,$3C,$6F,$E6,$BF,$F1,$CF,$DB,$63,$4F,$07,$03,$F1,$7F
	.dc.b	$E6,$DE,$B7,$F8,$EB,$F6,$07,$EA,$4F,$DD,$1F,$9A,$BF,$28,$7E,$B5
	.dc.b	$C3,$E9,$D8,$70,$BD,$D3,$F3,$AD,$7C,$27,$F0,$0C,$FB,$74,$FE,$FD
	.dc.b	$F0,$69,$DF,$47,$E5,$1F,$CD,$27,$39,$78,$BF,$1A,$C6,$BE,$0C,$38
	.dc.b	$68,$FD,$DF,$C9,$B7,$8F,$3F,$47,$EF,$3F,$D4,$FD,$57,$E3,$5F,$E0
	.dc.b	$0F,$D7,$3F,$DF,$1F,$80,$3F,$78,$7E,$F3,$BA,$D7,$9C,$AD,$7C,$D6
	.dc.b	$7E,$7F,$F8,$DE,$5D,$3E,$82,$E1,$C2,$FF,$DD,$5C,$FC,$E5,$A7,$83
	.dc.b	$EC,$7F,$A8,$E1,$C6,$3C,$7A,$B4,$E1,$18,$FD,$FB,$FD,$F5,$F9,$87
	.dc.b	$FB,$C3,$F0,$77,$95,$FE,$92,$FD,$93,$FD,$EF,$FB,$0B,$F8,$07,$F9
	.dc.b	$DF,$A0,$C3,$85,$EF,$E1,$57,$BF,$CD,$7A,$F1,$7F,$7F,$CF,$AF,$E0
	.dc.b	$1F,$EB,$2C,$78,$FB,$B6,$B5,$97,$3C,$7F,$A5,$C7,$21,$87,$2D,$7C
	.dc.b	$BF,$1F,$FE,$FF,$C7,$C7,$92,$7D,$3B,$3F,$0B,$E5,$87,$1D,$BC,$F3
	.dc.b	$FD,$2F,$7F,$66,$7E,$2C,$79,$56,$BF,$22,$E3,$B7,$97,$07,$E4,$7E
	.dc.b	$5F,$CF,$5F,$8C,$3F,$4E,$7E,$84,$FC,$87,$FC,$FD,$FA,$13,$F5,$9E
	.dc.b	$17,$BA,$D7,$DE,$A9,$F4,$1A,$C7,$E3,$70,$B3,$EF,$F3,$38,$7C,$33
	.dc.b	$E0,$D3,$FD,$19,$F3,$1F,$16,$FF,$C6,$15,$AC,$FB,$FC,$EE,$67,$A6
	.dc.b	$BE,$3C,$3C,$FF,$8E,$78,$6F,$CB,$AB,$93,$EA,$7D,$43,$F6,$E7,$EE
	.dc.b	$1F,$D1,$7F,$01,$7E,$50,$FC,$63,$E2,$F6,$7C,$ED,$FF,$C5,$F8,$70
	.dc.b	$F9,$6F,$F5,$05,$EF,$F1,$FE,$B3,$D2,$BF,$DE,$DE,$17,$F9,$25,$E3
	.dc.b	$F9,$47,$C4,$D7,$C5,$AF,$87,$17,$E9,$9E,$D9,$CE,$7D,$BF,$24,$C6
	.dc.b	$7E,$1D,$FC,$FB,$BF,$3C,$F0,$E7,$CE,$1F,$9E,$7E,$4F,$F5,$47,$EC
	.dc.b	$6F,$C7,$BF,$CB,$7F,$BF,$3C,$FF,$F5,$D7,$D0,$6E,$B5,$F1,$0A,$D6
	.dc.b	$AF,$D6,$B5,$F3,$7F,$7B,$3E,$B9,$DC,$78,$E1,$D7,$F5,$5F,$B1,$5A
	.dc.b	$CB,$A6,$9F,$C8,$3A,$3C,$18,$F9,$23,$FD,$99,$E5,$E2,$B7,$AB,$F6
	.dc.b	$AC,$67,$E4,$F4,$EE,$1F,$ED,$3F,$A6,$FD,$97,$C9,$3F,$62,$FF,$0B
	.dc.b	$71,$3C,$17,$E0,$CF,$8D,$7E,$2F,$C2,$F7,$5A,$97,$E3,$39,$FA,$9F
	.dc.b	$CE,$E5,$D7,$F6,$CB,$87,$0C,$BF,$D7,$1E,$37,$E0,$6B,$C5,$7F,$27
	.dc.b	$2F,$0F,$D7,$FE,$6E,$73,$AF,$97,$0F,$27,$E0,$7D,$E3,$1A,$FF,$A5
	.dc.b	$70,$FC,$A7,$CD,$BF,$53,$FF,$27,$7E,$E8,$F9,$D7,$E6,$7F,$D6,$7E
	.dc.b	$3B,$F8,$A5,$EF,$EE,$17,$BA,$FE,$BD,$FA,$D6,$B1,$FF,$2C,$78,$1F
	.dc.b	$9A,$3F,$D4,$58,$77,$F8,$5F,$DE,$BC,$F0,$E5,$F3,$BF,$B5,$8C,$69
	.dc.b	$E2,$DB,$18,$F9,$A3,$CB,$F3,$2F,$D9,$7C,$1B,$79,$27,$33,$FC,$AF
	.dc.b	$F6,$5F,$ED,$FF,$94,$FF,$79,$FD,$A9,$FB,$43,$F2,$DF,$F3,$EF,$2F
	.dc.b	$ED,$EF,$7F,$4C,$BD,$D3,$F0,$5E,$9F,$41,$FE,$03,$E5,$F7,$0C,$3B
	.dc.b	$B8,$79,$DF,$83,$CE,$FE,$25,$18,$CF,$F1,$95,$D3,$4F,$34,$79,$7F
	.dc.b	$52,$C6,$9E,$38,$F4,$7C,$3E,$32,$F2,$FC,$45,$9F,$67,$C4,$FF,$4B
	.dc.b	$AF,$F3,$6F,$EA,$2F,$C7,$9F,$90,$BF,$28,$7F,$00,$3C,$7E,$1E,$CB
	.dc.b	$0E,$17,$BA,$FF,$3D,$F3,$FF,$1D,$CB,$AF,$17,$C2,$7F,$CF,$7F,$B7
	.dc.b	$7F,$3A,$7C,$3A,$78,$34,$FE,$2F,$5D,$B2,$EB,$0F,$81,$FE,$6D,$7C
	.dc.b	$33,$F0,$7E,$0D,$BC,$79,$69,$FD,$43,$4F,$E5,$DE,$4B,$F5,$4F,$F6
	.dc.b	$37,$EE,$1F,$ED,$EF,$51,$C7,$F9,$FB,$DD,$6B,$E1,$BF,$D2,$72,$FD
	.dc.b	$DB,$FB,$96,$3C,$FB,$FF,$A9,$BD,$7E,$4A,$E1,$C7,$2F,$F0,$BE,$5D
	.dc.b	$25,$9F,$87,$E9,$3F,$A8,$61,$C6,$3C,$99,$FE,$BE,$F8,$46,$3F,$3F
	.dc.b	$FE,$1D,$F9,$E7,$F4,$1F,$99,$7F,$00,$BF,$3F,$FC,$77,$EA,$1F,$E2
	.dc.b	$DF,$DD,$5F,$A9,$70,$ED,$1F,$DC,$6F,$87,$8E,$BE,$1F,$A5,$6E,$B5
	.dc.b	$FA,$FF,$F4,$F7,$FD,$0D,$A7,$87,$0F,$04,$64,$E1,$A7,$93,$4F,$1F
	.dc.b	$D7,$FF,$33,$E0,$61,$C6,$73,$E9,$F9,$27,$38,$79,$B0,$E5,$AF,$97
	.dc.b	$16,$FF,$17,$2E,$53,$98,$7E,$17,$E2,$D3,$DB,$BF,$C3,$F7,$F1,$B7
	.dc.b	$92,$DF,$05,$EF,$FC,$CB,$FB,$D7,$8E,$FC,$A7,$FC,$9B,$FA,$9B,$F5
	.dc.b	$F7,$EE,$9B,$F9,$95,$EF,$EC,$57,$BF,$94,$D7,$EB,$3F,$17,$9F,$85
	.dc.b	$DE,$8F,$7C,$35,$F0,$EB,$E0,$FC,$EB,$E2,$73,$FF,$7F,$9C,$E9,$DB
	.dc.b	$0C,$E1,$F7,$27,$3A,$79,$FF,$1C,$EF,$97,$AB,$81,$FB,$37,$DD,$FC
	.dc.b	$D3,$F4,$3F,$AE,$7B,$3F,$3B,$7E,$90,$FC,$DD,$F9,$4E,$FC,$85,$EE
	.dc.b	$B5,$FB,$86,$7F,$47,$F6,$72,$7F,$85,$73,$EB,$C5,$F8,$5E,$3F,$7C
	.dc.b	$78,$3A,$9B,$78,$3F,$20,$F5,$DA,$B5,$3E,$4F,$9B,$CC,$63,$4F,$EF
	.dc.b	$1F,$87,$F9,$76,$F2,$7C,$F7,$F3,$AD,$3E,$6F,$2F,$F3,$CF,$91,$FE
	.dc.b	$31,$F9,$53,$F3,$BF,$F3,$BF,$EE,$7F,$9F,$AD,$7C,$12,$B5,$F8,$07
	.dc.b	$F2,$4C,$79,$FC,$33,$F3,$AF,$7F,$1F,$EB,$31,$EC,$FB,$F9,$BF,$C6
	.dc.b	$D1,$D3,$1F,$07,$D7,$FB,$A3,$1A,$BE,$CB,$19,$7F,$6A,$D7,$C5,$3F
	.dc.b	$E8,$CF,$C1,$FC,$D3,$9C,$7C,$BF,$84,$63,$87,$DB,$7F,$E9,$2F,$CB
	.dc.b	$31,$8F,$2E,$17,$E9,$DF,$BD,$FC,$9D,$FA,$43,$F3,$77,$C7,$BE,$C9
	.dc.b	$FB,$C3,$F2,$3F,$A9,$D3,$F9,$8C,$78,$E1,$C2,$F7,$4F,$DB,$3B,$65
	.dc.b	$D7,$A7,$E2,$1F,$C4,$B0,$F0,$FE,$27,$F2,$A1,$1D,$34,$F1,$68,$F8
	.dc.b	$7F,$C2,$CE,$73,$7E,$3F,$FC,$53,$0E,$99,$7A,$F1,$E3,$A7,$EE,$36
	.dc.b	$F4,$4E,$70,$F3,$FE,$A3,$FC,$7F,$F4,$3F,$F4,$C7,$E9,$1F,$91,$FB
	.dc.b	$0F,$F0,$47,$EF,$D9,$F9,$C4,$E7,$AB,$FC,$49,$EA,$E7,$F5,$3F,$56
	.dc.b	$3C,$B8,$3D,$E1,$EB,$E5,$FF,$B2,$6F,$F8,$BC,$F5,$C3,$A7,$E1,$DF
	.dc.b	$A2,$BD,$E9,$FE,$19,$FC,$D3,$58,$F1,$FA,$87,$DA,$7F,$37,$FC,$8D
	.dc.b	$FB,$DF,$E5,$7E,$49,$FA,$7F,$DA,$FB,$27,$C9,$EE,$B5,$39,$F8,$BC
	.dc.b	$7A,$3F,$4F,$1E,$7F,$A3,$5B,$DF,$85,$FE,$3A,$ED,$FE,$33,$5E,$0B
	.dc.b	$FE,$9F,$8F,$E2,$3F,$60,$FD,$DC,$E7,$5F,$1E,$1F,$ED,$3F,$BF,$EF
	.dc.b	$18,$DB,$FD,$17,$97,$4C,$FB,$C3,$EC,$37,$9F,$AF,$ED,$7F,$AB,$7E
	.dc.b	$A4,$FD,$45,$F9,$C3,$F7,$9F,$E6,$BF,$38,$FD,$AD,$FA,$67,$0B,$DD
	.dc.b	$6A,$5F,$99,$67,$0F,$CA,$63,$D7,$F1,$4B,$87,$0F,$94,$FF,$3F,$78
	.dc.b	$3F,$C6,$2B,$C5,$7E,$19,$FF,$41,$7E,$33,$9B,$E1,$AF,$8F,$F6,$2E
	.dc.b	$DA,$F8,$7F,$49,$FF,$45,$7E,$AD,$FE,$BD,$FD,$BD,$FB,$97,$ED,$7F
	.dc.b	$0F,$FE,$ED,$F2,$BF,$E0,$AF,$C6,$F7,$E6,$7F,$7C,$BD,$FB,$8B,$DF
	.dc.b	$D5,$FF,$B5,$FB,$99,$F5,$CF,$F6,$98,$FF,$7D,$9C,$BC,$E1,$EE,$7F
	.dc.b	$57,$AE,$18,$F5,$FA,$4F,$E8,$DE,$B7,$F7,$99,$CE,$7E,$6F,$89,$C6
	.dc.b	$5C,$F0,$E1,$BF,$F8,$8B,$E4,$BD,$32,$DF,$D3,$87,$A3,$F5,$4C,$9F
	.dc.b	$D4,$FE,$15,$F2,$DF,$DC,$FE,$21,$FC,$02,$F2,$B7,$A2,$F7,$F3,$8B
	.dc.b	$DF,$E2,$9A,$FD,$4F,$D9,$B7,$F8,$93,$4E,$CF,$F5,$7E,$2F,$F7,$74
	.dc.b	$78,$75,$F0,$7D,$8F,$B2,$93,$9C,$BB,$7E,$7B,$98,$C6,$7E,$6F,$DC
	.dc.b	$3C,$BA,$F9,$23,$D3,$F0,$B8,$CF,$D9,$C3,$FE,$1E,$EB,$FA,$7A,$76
	.dc.b	$E1,$F6,$56,$BF,$09,$F9,$3F,$A3,$7F,$36,$DD,$FB,$FB,$F2,$4F,$E2
	.dc.b	$FF,$8C,$BF,$0B,$DF,$B1,$FE,$1B,$AD,$7C,$2A,$A6,$DE,$0C,$BA,$6D
	.dc.b	$F0,$CF,$F2,$4B,$FD,$13,$F7,$1F,$93,$08,$F1,$61,$D3,$CC,$F8,$D1
	.dc.b	$8D,$3F,$94,$FA,$18,$D3,$C5,$1E,$6A,$F2,$ED,$EA,$8F,$4F,$EC,$1F
	.dc.b	$86,$FA,$97,$B5,$FE,$2E,$FD,$83,$F8,$EF,$DA,$BF,$6D,$FF,$4B,$F0
	.dc.b	$FC,$8B,$DF,$D6,$AF,$74,$FD,$0B,$A6,$2F,$CB,$BD,$78,$70,$E3,$F8
	.dc.b	$07,$F8,$A2,$7C,$1E,$FF,$F9,$C3,$F0,$9F,$C1,$57,$6C,$BC,$7E,$26
	.dc.b	$2F,$7B,$79,$7E,$3F,$BE,$DE,$3C,$BF,$CF,$3B,$7F,$A7,$77,$FB,$FF
	.dc.b	$99,$7F,$74,$FB,$9F,$F5,$77,$DA,$BF,$78,$7E,$52,$EA,$FD,$4D,$EF
	.dc.b	$E8,$D7,$BA,$7E,$63,$AF,$1D,$FA,$FF,$D8,$C3,$B7,$3F,$B9,$E5,$87
	.dc.b	$7F,$8E,$FC,$F4,$63,$4F,$E8,$1E,$7F,$FA,$2A,$D3,$9C,$DF,$A1,$F7
	.dc.b	$DF,$D3,$87,$A3,$D5,$C6,$FE,$68,$F5,$FA,$D8,$D7,$D1,$E8,$FF,$04
	.dc.b	$FC,$33,$DD,$F9,$F3,$E6,$5F,$A8,$FF,$A8,$3F,$3D,$7E,$39,$BE,$85
	.dc.b	$EE,$B5,$F8,$96,$7F,$5D,$FC,$BE,$5D,$39,$7E,$EB,$DB,$3E,$BE,$03
	.dc.b	$F2,$AF,$DF,$4E,$71,$FF,$57,$FD,$06,$34,$EF,$87,$0D,$9F,$86,$78
	.dc.b	$F6,$F1,$66,$9F,$F5,$9F,$7F,$EF,$3D,$9F,$F3,$77,$E9,$CF,$D4,$1F
	.dc.b	$74,$FD,$15,$F9,$C7,$E7,$70,$E1,$EF,$30,$E1,$7B,$A7,$E9,$3F,$CB
	.dc.b	$32,$EB,$CB,$F9,$A7,$FC,$F9,$F8,$EF,$EE,$1F,$0E,$9E,$0D,$3F,$DF
	.dc.b	$D5,$C7,$3E,$D3,$9E,$9F,$E4,$17,$C3,$3F,$07,$EF,$DB,$D7,$C7,$97
	.dc.b	$A3,$F4,$0F,$E7,$BF,$5C,$F7,$9F,$76,$FE,$01,$7D,$7B,$F1,$CF,$C4
	.dc.b	$FE,$1B,$C2,$F7,$5A,$9F,$C6,$FE,$C5,$9C,$3F,$88,$E3,$D7,$F0,$6B
	.dc.b	$87,$0C,$BA,$7C,$17,$F3,$7A,$F1,$5F,$86,$7C,$1D,$F9,$BE,$19,$7E
	.dc.b	$6B,$E6,$7B,$69,$E1,$F3,$9F,$2D,$8C,$7D,$10,$FD,$37,$18,$7E,$8B
	.dc.b	$F4,$E7,$7B,$F3,$5F,$D4,$7D,$43,$F7,$B7,$E1,$AF,$D4,$7F,$D2,$9F
	.dc.b	$30,$FD,$7F,$E2,$63,$C7,$C3,$FF,$7C,$E3,$C7,$0E,$1F,$83,$30,$E1
	.dc.b	$D5,$E0,$D3,$BE,$9F,$EB,$BC,$7F,$D6,$51,$8C,$3F,$DA,$4D,$D3,$C3
	.dc.b	$F0,$DF,$EF,$38,$C6,$5D,$7E,$6B,$18,$78,$B1,$F4,$E1,$E8,$D9,$FA
	.dc.b	$D7,$86,$B5,$9F,$9B,$5F,$67,$DD,$E6,$5F,$4A,$F9,$D7,$E5,$EF,$CB
	.dc.b	$1F,$B3,$FF,$81,$7E,$0F,$FC,$27,$CD,$F0,$F1,$E3,$87,$0B,$DF,$C1
	.dc.b	$7F,$26,$FA,$4E,$D9,$F5,$FC,$9F,$DF,$80,$E3,$A6,$1B,$87,$2F,$88
	.dc.b	$F7,$C6,$35,$FF,$0B,$7A,$18,$C7,$D1,$87,$9E,$7C,$DF,$93,$78,$35
	.dc.b	$F2,$47,$A7,$F7,$EF,$81,$D9,$F7,$FD,$E7,$E5,$8F,$D4,$DF,$BE,$BE
	.dc.b	$07,$FD,$6F,$CB,$EF,$C3,$85,$EE,$B5,$F4,$FD,$3B,$7F,$D7,$BC,$5F
	.dc.b	$78,$F7,$5F,$77,$01,$E7,$97,$83,$2F,$7C,$63,$0F,$E4,$55,$D7,$3F
	.dc.b	$3D,$EF,$F6,$1C,$67,$E3,$C5,$FB,$6E,$2B,$A6,$D7,$2F,$67,$9D,$F8
	.dc.b	$1F,$C8,$9E,$93,$ED,$1F,$8E,$FE,$03,$E6,$DF,$42,$FC,$D3,$85,$EE
	.dc.b	$B5,$3F,$E8,$8F,$2F,$FA,$A6,$3D,$3F,$68,$FB,$8F,$6C,$AE,$1C,$7A
	.dc.b	$7F,$C1,$7E,$CF,$DE,$AD,$65,$E1,$F6,$31,$7B,$D7,$C7,$F2,$BE,$2C
	.dc.b	$BC,$F9,$7F,$85,$35,$7D,$67,$E7,$7E,$9D,$E2,$3E,$8D,$F8,$C3,$F3
	.dc.b	$17,$AB,$FE,$59,$FC,$E3,$85,$EE,$B5,$3F,$E5,$DF,$B3,$CE,$1F,$97
	.dc.b	$C7,$AF,$97,$70,$E1,$97,$4E,$8F,$CC,$D7,$8A,$FC,$33,$E0,$FC,$0F
	.dc.b	$37,$C3,$2F,$F3,$57,$D8,$76,$D3,$C3,$A7,$F3,$6F,$23,$F6,$1F,$A0
	.dc.b	$BF,$73,$FF,$06,$FC,$AB,$F3,$57,$3B,$F3,$AF,$F0,$E7,$81,$FE,$07
	.dc.b	$C3,$87,$D3,$FF,$76,$BD,$D3,$EA,$BF,$82,$56,$BC,$CF,$CB,$BD,$FE
	.dc.b	$07,$F6,$CF,$2C,$FC,$19,$BF,$20,$FD,$BC,$E7,$2F,$1F,$E4,$18,$DB
	.dc.b	$CF,$1E,$0F,$94,$EF,$18,$C3,$D1,$F9,$DE,$37,$F3,$47,$AF,$B3,$1A
	.dc.b	$78,$38,$3E,$E7,$17,$F1,$3F,$DB,$1F,$82,$3F,$39,$FF,$6C,$7E,$14
	.dc.b	$FC,$83,$E2,$F8,$FF,$C8,$6F,$75,$A9,$F9,$3F,$AD,$E7,$0E,$5C,$FF
	.dc.b	$B8,$BE,$BF,$6B,$FC,$5D,$E3,$F8,$7B,$3E,$DE,$0F,$F1,$89,$CC,$78
	.dc.b	$70,$C1,$E5,$AE,$19,$7D,$8F,$C1,$7C,$1A,$78,$F8,$9F,$BF,$FF,$37
	.dc.b	$F1,$5F,$8F,$3F,$64,$7E,$3C,$FD,$B3,$FC,$43,$F9,$AF,$EF,$39,$BF
	.dc.b	$9B,$61,$C2,$F7,$5A,$FD,$5B,$A6,$2F,$CC,$BF,$99,$5E,$F9,$BF,$8B
	.dc.b	$BC,$73,$ED,$F5,$CD,$3F,$9A,$CE,$72,$FF,$18,$FA,$D8,$CB,$CD,$87
	.dc.b	$0B,$7E,$69,$E0,$D7,$C7,$97,$3F,$CC,$7F,$AA,$4F,$F9,$AB,$F0,$27
	.dc.b	$E3,$9F,$E6,$CF,$DE,$FF,$8B,$79,$BA,$FB,$A5,$69,$4F,$F1,$BF,$B3
	.dc.b	$D7,$E4,$BF,$D3,$71,$E9,$C7,$FC,$C9,$ED,$97,$5F,$B8,$FE,$9F,$18
	.dc.b	$C7,$9F,$C4,$E1,$EF,$5E,$3F,$7F,$FD,$FD,$D3,$FD,$8E,$F8,$69,$FC
	.dc.b	$AF,$D9,$F3,$6D,$E5,$8F,$E5,$DF,$A7,$63,$1E,$5F,$A3,$7F,$85,$B8
	.dc.b	$3F,$CF,$9F,$9E,$7F,$AE,$FE,$D7,$F7,$7F,$9B,$BF,$61,$7D,$DF,$F8
	.dc.b	$73,$7F,$F2,$E7,$27,$F6,$BC,$38,$78,$6C,$38,$5B,$F8,$03,$B5,$EF
	.dc.b	$F1,$0F,$F0,$A3,$FE,$8B,$32,$30,$FF,$66,$7D,$CE,$63,$FB,$C6,$5E
	.dc.b	$07,$54,$D3,$CF,$A7,$F8,$B2,$79,$E1,$E4,$CF,$D5,$D1,$CB,$6F,$7E
	.dc.b	$3C,$76,$F4,$63,$EE,$FC,$5B,$8C,$E7,$0F,$5F,$C3,$79,$BE,$55,$F8
	.dc.b	$23,$E3,$DF,$95,$3F,$71,$7E,$A5,$F8,$AF,$C2,$B8,$5E,$EB,$53,$F8
	.dc.b	$EF,$D7,$F3,$87,$BF,$8F,$7F,$29,$79,$BE,$06,$9E,$0A,$EF,$A7,$6F
	.dc.b	$8D,$F6,$61,$FE,$1D,$AD,$67,$DB,$CF,$CC,$FF,$84,$35,$F1,$61,$E6
	.dc.b	$ED,$7E,$8F,$E2,$C3,$CB,$1F,$CE,$F4,$FA,$1E,$B7,$C0,$FC,$B5,$F2
	.dc.b	$4F,$C7,$DF,$A6,$3E,$A9,$F1,$BF,$73,$0E,$17,$BA,$D7,$EA,$59,$DD
	.dc.b	$97,$BF,$BF,$FD,$9D,$87,$CE,$FC,$1B,$1C,$5F,$E3,$98,$E7,$AF,$83
	.dc.b	$E5,$BD,$B1,$8C,$FF,$83,$57,$5C,$7C,$F7,$E6,$D9,$F8,$B6,$76,$F2
	.dc.b	$47,$FB,$CF,$4F,$0E,$BE,$49,$CF,$EF,$5B,$C3,$97,$EC,$BF,$98,$FC
	.dc.b	$8F,$FD,$AD,$E9,$3F,$51,$7E,$DC,$FB,$C7,$E8,$6F,$CA,$B8,$7E,$25
	.dc.b	$FE,$21,$BD,$D6,$BC,$2C,$FD,$3F,$C0,$E5,$69,$D9,$FD,$27,$67,$E8
	.dc.b	$E3,$C3,$87,$2C,$FE,$FC,$63,$4F,$17,$E7,$78,$D3,$C1,$1E,$5A,$EB
	.dc.b	$87,$39,$CE,$9E,$8F,$D7,$FF,$6B,$F3,$BF,$C2,$B5,$7C,$93,$F3,$67
	.dc.b	$EB,$2F,$DF,$B7,$C8,$F7,$18,$70,$BD,$FE,$DD,$A9,$F7,$65,$FE,$43
	.dc.b	$FC,$A2,$E1,$C7,$0E,$1F,$15,$F9,$AA,$F1,$69,$E1,$F8,$DC,$5E,$E3
	.dc.b	$C9,$AF,$F1,$3E,$1F,$E1,$17,$C3,$5F,$E6,$FC,$FF,$C4,$3E,$8B,$FC
	.dc.b	$E5,$E7,$3F,$07,$FF,$39,$7C,$93,$F1,$17,$F0,$0E,$FE,$D7,$FC,$79
	.dc.b	$5A,$97,$EF,$B9,$C3,$EC,$63,$FE,$7D,$FA,$3B,$87,$0F,$8E,$67,$FE
	.dc.b	$7F,$D3,$86,$7E,$07,$F6,$F8,$79,$23,$B7,$99,$AB,$E3,$A3,$D0,$C6
	.dc.b	$3E,$7E,$3B,$2F,$9B,$FE,$5B,$EA,$7F,$5C,$F1,$5F,$B6,$BF,$19,$FE
	.dc.b	$55,$EA,$DE,$8F,$E5,$D7,$BA,$D7,$C4,$29,$F8,$BF,$4A,$7E,$B3,$F9
	.dc.b	$77,$B7,$27,$E5,$BD,$3F,$F5,$83,$FD,$B2,$3F,$4D,$8F,$4F,$AC,$FE
	.dc.b	$2B,$1D,$F4,$FE,$79,$DB,$88,$DC,$7F,$3F,$FD,$17,$1B,$FA,$A3,$D3
	.dc.b	$F0,$DF,$D2,$A1,$F8,$CB,$F3,$0F,$C8,$FE,$02,$FC,$B5,$FB,$77,$FB
	.dc.b	$BF,$97,$FA,$66,$1C,$2F,$75,$AF,$DB,$BA,$63,$CF,$87,$D8,$7F,$B2
	.dc.b	$EF,$BF,$C6,$FE,$DD,$E5,$E5,$FF,$AB,$F0,$FF,$41,$D7,$2C,$7A,$F9
	.dc.b	$7F,$CD,$2F,$7A,$F9,$75,$FC,$2E,$BE,$4D,$3D,$59,$F3,$9F,$F1,$77
	.dc.b	$E6,$DF,$33,$F1,$9F,$CB,$7E,$7C,$FC,$1D,$FA,$83,$F5,$DF,$87,$F0
	.dc.b	$8D,$18,$F1,$F0,$98,$F1,$C3,$86,$BF,$1A,$F7,$C5,$FE,$E9,$D3,$BB
	.dc.b	$FC,$4B,$1E,$5F,$5B,$FD,$5D,$E9,$8F,$3D,$1F,$8D,$FF,$7A,$9C,$ED
	.dc.b	$E6,$97,$E3,$19,$C7,$B6,$DE,$4A,$D6,$BF,$C3,$B5,$F1,$D3,$5F,$83
	.dc.b	$5A,$FD,$53,$DD,$9F,$6D,$FD,$7F,$9B,$E3,$0E,$BB,$F9,$73,$7D,$F6
	.dc.b	$65,$F9,$67,$F8,$85,$7E,$5A,$C3,$F1,$BF,$46,$FD,$7B,$85,$EE,$B5
	.dc.b	$8F,$F8,$A3,$F1,$AC,$FD,$07,$F7,$DC,$AF,$E8,$5F,$1D,$1E,$08,$B7
	.dc.b	$CB,$F1,$EF,$89,$39,$3C,$7A,$7E,$7B,$E9,$33,$9F,$86,$3C,$BF,$6F
	.dc.b	$E0,$AD,$E9,$FE,$3A,$F0,$7D,$DF,$B7,$FF,$B6,$3E,$81,$F8,$4B,$F4
	.dc.b	$9F,$F6,$AF,$C5,$3E,$CF,$7F,$86,$7F,$BD,$AB,$5F,$49,$FC,$FF,$F3
	.dc.b	$3C,$E3,$EB,$D5,$9F,$FB,$93,$F2,$7C,$3D,$F0,$E3,$F2,$7E,$C8,$C6
	.dc.b	$BF,$EE,$4F,$95,$C6,$9D,$E3,$C7,$FA,$F7,$6C,$3C,$3B,$FF,$95,$3C
	.dc.b	$4F,$C1,$FD,$BF,$F9,$5B,$E0,$BF,$CE,$7A,$BE,$C9,$FA,$C3,$F5,$8F
	.dc.b	$BD,$7B,$FB,$55,$EE,$BF,$63,$E3,$FF,$18,$AE,$9C,$7F,$9F,$3E,$BB
	.dc.b	$FE,$64,$FC,$53,$E0,$3C,$EF,$FB,$04,$FD,$88,$C6,$4E,$2C,$78,$BF
	.dc.b	$D0,$B3,$F0,$CF,$9A,$7C,$B3,$9C,$3C,$9F,$7F,$8D,$BD,$B7,$FE,$9C
	.dc.b	$FC,$B7,$18,$F2,$FB,$BF,$EB,$1C,$2F,$D3,$3F,$98,$FE,$2A,$FD,$F9
	.dc.b	$FB,$FB,$F2,$7F,$E0,$9E,$5B,$F2,$B7,$DD,$FF,$07,$C7,$8E,$1C,$2F
	.dc.b	$74,$FC,$5F,$5C,$DF,$9A,$CF,$B7,$E6,$5F,$E5,$9C,$78,$E9,$6B,$59
	.dc.b	$73,$D5,$AF,$F4,$67,$D1,$3A,$AF,$34,$6B,$E5,$7F,$D0,$7F,$B7,$7F
	.dc.b	$4C,$FC,$F9,$F9,$63,$F0,$87,$E6,$EF,$D3,$1F,$93,$7E,$EB,$F5,$9F
	.dc.b	$E4,$58,$F1,$C3,$85,$EE,$BF,$7F,$E2,$6B,$6F,$DA,$E7,$DB,$3F,$F3
	.dc.b	$A6,$3C,$7E,$3E,$D6,$B2,$E7,$A7,$83,$F3,$8F,$3C,$E7,$0F,$2C,$F9
	.dc.b	$3F,$2F,$FF,$4E,$BF,$D9,$AD,$F5,$9F,$D5,$FF,$3C,$78,$CF,$A5,$7E
	.dc.b	$36,$F0,$BF,$9E,$63,$C7,$0E,$17,$BA,$FB,$9C,$DF,$0F,$5F,$E1,$F9
	.dc.b	$F6,$F9,$3F,$F4,$76,$3C,$7F,$5F,$B5,$AC,$B9,$E8,$DB,$CF,$39,$C1
	.dc.b	$B7,$9A,$BF,$AF,$E6,$FD,$4B,$FC,$03,$F8,$3B,$D4,$7C,$0F,$F8,$A3
	.dc.b	$C6,$FC,$B7,$D8,$7F,$B1,$F0,$E1,$A7,$FA,$6E,$F7,$F5,$BF,$94,$F0
	.dc.b	$D7,$FE,$D3,$CF,$AD,$6B,$1F,$D0,$7C,$1F,$FB,$22,$73,$9F,$7C,$3F
	.dc.b	$C6,$19,$F6,$CD,$A7,$96,$B5,$9B,$F5,$9F,$ED,$18,$F2,$D7,$C5,$2F
	.dc.b	$D9,$73,$8F,$9F,$0E,$3F,$8E,$E2,$3A,$EF,$E5,$D7,$D3,$F8,$56,$7F
	.dc.b	$55,$7E,$19,$FE,$51,$F1,$9F,$BB,$BF,$5A,$79,$2F,$D0,$38,$5E,$EB
	.dc.b	$52,$FD,$FB,$3F,$68,$C7,$9E,$3F,$03,$93,$FC,$67,$9F,$82,$B5,$8F
	.dc.b	$7E,$8F,$91,$C1,$FD,$52,$B5,$8F,$2C,$FF,$DB,$BD,$1F,$26,$73,$87
	.dc.b	$8B,$C9,$F0,$DE,$FD,$8B,$3E,$6F,$77,$E2,$7D,$13,$F3,$3F,$CF,$FE
	.dc.b	$52,$FD,$1B,$FC,$31,$F8,$E3,$F6,$7D,$D6,$A7,$3F,$A4,$E3,$C8,$61
	.dc.b	$C7,$E8,$9F,$AF,$7A,$B9,$3F,$D6,$98,$F1,$9F,$C8,$C7,$1A,$E1,$D0
	.dc.b	$F4,$B1,$9F,$F8,$A6,$1E,$DE,$30,$FA,$9F,$B8,$59,$7E,$61,$FE,$2D
	.dc.b	$FD,$0F,$FC,$BF,$F2,$1F,$96,$F1,$3F,$93,$7E,$63,$FA,$6C,$9D,$5B
	.dc.b	$DF,$FC,$03,$7B,$F8,$25,$EF,$F7,$3D,$57,$F6,$CF,$2F,$F5,$55,$AF
	.dc.b	$47,$D1,$3A,$C3,$96,$1F,$B4,$FC,$B3,$FC,$1B,$E0,$B6,$F2,$E5,$CA
	.dc.b	$BC,$9F,$65,$C6,$31,$97,$C0,$FD,$0F,$1C,$1E,$A8,$F8,$DE,$FE,$36
	.dc.b	$F1,$73,$FE,$57,$4E,$DB,$FB,$32,$E7,$EB,$F8,$B6,$FF,$41,$FE,$93
	.dc.b	$FE,$91,$FC,$4B,$F3,$9F,$A5,$BC,$AF,$D1,$FA,$FF,$02,$B5,$E8,$2B
	.dc.b	$5E,$83,$F3,$ED,$7D,$C3,$2E,$79,$3F,$46,$F0,$6F,$FE,$2E,$CF,$9E
	.dc.b	$5D,$FE,$F3,$32,$9C,$DF,$ED,$3F,$74,$EB,$8F,$4D,$3F,$D5,$3C,$5E
	.dc.b	$EC,$38,$69,$E1,$DB,$CD,$8B,$F1,$EF,$F5,$45,$EF,$3E,$B0,$FC,$87
	.dc.b	$FC,$E9,$BF,$FA,$9B,$EB,$9F,$8E,$FF,$11,$FB,$47,$EC,$1F,$E6,$6E
	.dc.b	$B7,$E6,$FC,$38,$3F,$C9,$38,$70,$F8,$3F,$F3,$25,$EF,$F2,$5D,$57
	.dc.b	$F9,$7B,$2E,$93,$FE,$70,$F3,$FF,$0B,$A7,$86,$73,$F7,$2F,$D0,$B3
	.dc.b	$87,$69,$CF,$27,$FB,$9F,$8F,$E7,$4E,$74,$F0,$E1,$FC,$5F,$F3,$EF
	.dc.b	$0E,$DE,$5E,$1F,$5C,$FF,$1B,$FC,$1B,$39,$F4,$FD,$43,$CF,$F5,$8F
	.dc.b	$DF,$3F,$C2,$FF,$97,$BE,$99,$FB,$1B,$F0,$C7,$E9,$5C,$2F,$75,$A9
	.dc.b	$7E,$4D,$9F,$A0,$63,$CF,$1F,$C8,$71,$E6,$31,$F1,$3F,$02,$1A,$78
	.dc.b	$EF,$8E,$5D,$7F,$27,$E0,$F7,$FC,$9E,$E5,$FE,$AE,$FC,$FE,$FD,$7B
	.dc.b	$F2,$0F,$A7,$F2,$8F,$D0,$1F,$46,$FD,$B9,$F9,$E3,$E4,$DB,$3F,$69
	.dc.b	$E3,$87,$0B,$DD,$6B,$F3,$4C,$E3,$CF,$1E,$5F,$94,$FA,$F3,$F0,$4E
	.dc.b	$6D,$F7,$D9,$CB,$A6,$1E,$1F,$8A,$F2,$78,$BF,$7B,$9D,$97,$2D,$7C
	.dc.b	$34,$FD,$9F,$C3,$18,$FD,$7A,$E3,$CF,$F2,$AF,$F6,$97,$EE,$1F,$EA
	.dc.b	$0C,$9F,$9C,$FE,$F7,$85,$FA,$6F,$C2,$FC,$CD,$FB,$3F,$F4,$7E,$06
	.dc.b	$3C,$70,$E1,$7B,$AF,$F1,$FD,$6B,$1F,$F5,$6D,$6B,$36,$3F,$ED,$49
	.dc.b	$F1,$9A,$F7,$D9,$F8,$AB,$7E,$57,$F2,$AD,$3C,$15,$AD,$3C,$1F,$9D
	.dc.b	$72,$C7,$D7,$39,$C3,$BE,$BF,$DF,$B5,$F2,$56,$B3,$ED,$A3,$F1,$D7
	.dc.b	$E5,$1C,$F7,$F4,$4E,$7D,$07,$E7,$9F,$ED,$1C,$7F,$D9,$3F,$46,$F8
	.dc.b	$47,$EC,$FF,$E7,$5F,$D6,$9F,$AD,$7F,$A8,$BF,$72,$E3,$F8,$C7,$D1
	.dc.b	$C3,$85,$EE,$BF,$D4,$FF,$2B,$AC,$BA,$E5,$D3,$E4,$59,$BF,$71,$FB
	.dc.b	$5A,$F8,$EF,$FC,$21,$7B,$AF,$0F,$E1,$3F,$ED,$EA,$D6,$9E,$0C,$79
	.dc.b	$E6,$FC,$15,$FB,$B7,$87,$3E,$7B,$79,$67,$FB,$EE,$3C,$B4,$ED,$87
	.dc.b	$FA,$07,$E5,$38,$67,$D7,$5F,$83,$F7,$CF,$3F,$E4,$FE,$35,$F9,$1F
	.dc.b	$C1,$3F,$63,$E9,$3F,$1B,$FD,$7F,$EE,$0F,$D3,$BF,$A6,$6E,$C3,$87
	.dc.b	$A2,$C3,$87,$A8,$C3,$87,$E7,$DD,$B9,$7F,$DF,$BA,$76,$D3,$AF,$87
	.dc.b	$F0,$F3,$E7,$97,$2C,$5F,$A3,$67,$1E,$3A,$77,$C7,$CB,$FA,$EF,$97
	.dc.b	$6F,$25,$6B,$1E,$52,$FD,$FF,$C5,$3F,$EC,$1C,$39,$6F,$E7,$AF,$DD
	.dc.b	$E1,$CA,$B5,$87,$2F,$DB,$78,$FC,$8B,$F3,$E7,$ED,$4F,$C1,$3F,$D0
	.dc.b	$9F,$99,$7D,$8F,$C5,$77,$5A,$9C,$F8,$3F,$EE,$1D,$3B,$63,$CF,$1E
	.dc.b	$5F,$48,$CB,$FC,$73,$AF,$29,$CC,$7F,$53,$8C,$63,$E0,$FD,$4F,$FB
	.dc.b	$76,$3D,$33,$F0,$63,$CE,$7F,$CD,$B3,$9C,$79,$FE,$C6,$FD,$97,$9D
	.dc.b	$F8,$63,$FA,$6D,$7C,$3C,$DF,$91,$BF,$58,$7E,$B3,$FE,$CC,$FC,$95
	.dc.b	$F8,$73,$F6,$9F,$9F,$87,$0B,$DF,$E2,$CB,$DF,$ED,$5A,$CB,$A6,$5C
	.dc.b	$FC,$BF,$C0,$72,$7C,$39,$CE,$1C,$7E,$6B,$39,$F5,$C3,$C5,$CD,$FE
	.dc.b	$EA,$DF,$D1,$87,$0D,$13,$FE,$FD,$9C,$EB,$E4,$9C,$FE,$41,$E3,$8C
	.dc.b	$57,$EC,$E9,$FA,$6F,$F3,$7F,$83,$7E,$C8,$FD,$E9,$E4,$3C,$A7,$C3
	.dc.b	$BF,$4E,$FE,$BD,$97,$2E,$40,$CB,$9E,$3C,$B0,$E3,$6F,$D1,$7F,$8F
	.dc.b	$E9,$E1,$D3,$C1,$A7,$7D,$35,$9E,$8E,$B9,$F4,$9D,$72,$7F,$11,$C3
	.dc.b	$96,$9D,$F0,$ED,$F9,$C7,$1C,$7C,$99,$76,$D7,$D7,$A3,$F6,$0D,$6B
	.dc.b	$D6,$7D,$FF,$09,$D7,$0F,$0E,$1C,$72,$FE,$FD,$FA,$6F,$08,$C7,$C8
	.dc.b	$BF,$6C,$F0,$7F,$00,$7F,$96,$FD,$17,$D7,$BF,$30,$7E,$20,$FC,$63
	.dc.b	$FC,$83,$D3,$F2,$F6,$63,$C7,$0E,$17,$BE,$AE,$BF,$77,$FC,$32,$73
	.dc.b	$F8,$CF,$9F,$0E,$33,$9C,$FB,$FD,$C6,$72,$E9,$1E,$2B,$7F,$00,$74
	.dc.b	$DB,$CF,$39,$FC,$03,$19,$F8,$BF,$0F,$E5,$9F,$5D,$9F,$A3,$BE,$0B
	.dc.b	$78,$FB,$A3,$FC,$47,$C3,$F3,$F8,$3F,$0F,$FB,$53,$F2,$2F,$EB,$B8
	.dc.b	$7F,$12,$FC,$6F,$FC,$8D,$F8,$C7,$F6,$4E,$FF,$F0,$65,$EF,$E3,$3B
	.dc.b	$7A,$CB,$FF,$26,$7E,$67,$BC,$FA,$E7,$D2,$B5,$FA,$57,$F5,$7C,$B9
	.dc.b	$CE,$70,$7D,$3E,$70,$E3,$97,$69,$7E,$C9,$FC,$7F,$0E,$33,$98,$7B
	.dc.b	$FE,$1B,$6B,$E7,$DF,$D9,$A7,$ED,$F1,$E3,$18,$FC,$AF,$85,$6F,$6F
	.dc.b	$67,$E1,$FE,$3C,$5F,$9D,$FF,$9A,$7F,$4B,$FF,$05,$FC,$23,$EA,$5F
	.dc.b	$99,$BF,$4A,$F1,$5E,$FE,$85,$7B,$A7,$3E,$BE,$E9,$97,$4C,$B9,$E7
	.dc.b	$98,$C7,$EA,$9E,$3C,$FA,$65,$E1,$5F,$92,$6F,$4E,$F8,$FA,$3E,$8B
	.dc.b	$17,$E7,$FD,$67,$F9,$8F,$77,$E0,$F4,$F6,$EB,$E2,$C7,$8D,$7F,$91
	.dc.b	$AB,$5F,$20,$FC,$BB,$FD,$99,$FB,$53,$F7,$8F,$E6,$DF,$A6,$FF,$1A
	.dc.b	$BD,$D6,$A7,$3F,$92,$FD,$5C,$39,$7A,$0C,$79,$7C,$5F,$F9,$A3,$6D
	.dc.b	$63,$C6,$1F,$A9,$7D,$3C,$38,$DF,$0C,$5F,$A6,$70,$CB,$A6,$7F,$D2
	.dc.b	$65,$FA,$4E,$6F,$BF,$EC,$1F,$8C,$DB,$CF,$87,$0D,$13,$E5,$FD,$55
	.dc.b	$FB,$97,$FB,$BB,$F6,$7F,$F5,$D7,$C7,$7F,$37,$FD,$A5,$FA,$83,$D7
	.dc.b	$61,$C3,$97,$FD,$CD,$87,$0F,$88,$61,$C3,$DA,$DE,$7D,$7F,$13,$FE
	.dc.b	$C7,$5A,$F8,$AF,$E7,$B9,$73,$CB,$FD,$D9,$97,$3C,$BF,$B1,$43,$BF
	.dc.b	$FD,$B6,$FF,$44,$F4,$7F,$C1,$9F,$9D,$7F,$6B,$F8,$F7,$ED,$3F,$E3
	.dc.b	$9F,$DF,$B8,$7E,$35,$FE,$89,$BD,$F3,$7F,$75,$AF,$F1,$F7,$E1,$5F
	.dc.b	$EE,$AF,$B9,$FF,$47,$7E,$AA,$FD,$3B,$FA,$B7,$C2,$3F,$36,$FE,$BB
	.dc.b	$E2,$7D,$DF,$C5,$BF,$D2,$7C,$3F,$D5,$B0,$E1,$F2,$5F,$CD,$2F,$7E
	.dc.b	$F3,$F6,$BF,$F6,$A5,$FF,$5C,$FC,$01,$FB,$AB,$F4,$C7,$E5,$1F,$2B
	.dc.b	$5F,$F4,$CF,$E5,$DC,$78,$BF,$D4,$58,$70,$BD,$F5,$FF,$B0,$3F,$09
	.dc.b	$7E,$AE,$FC,$0F,$E9,$EB,$FD,$AB,$F2,$57,$ED,$4F,$CB,$5F,$94,$3F
	.dc.b	$60,$7E,$BA,$FC,$3F,$FC,$9B,$97,$2C,$99,$72,$C7,$8E,$1C,$2F,$FC
	.dc.b	$FF,$F8,$B7,$F3,$7B,$FE,$F3,$F7,$7F,$ED,$82,$7D,$5F,$F9,$97,$D1
	.dc.b	$FF,$6C,$7C,$77,$FA,$F6,$74,$7F,$B7,$A7,$FC,$D5,$FB,$2E,$AF,$9C
	.dc.b	$EA,$7F,$C0,$D8,$77,$9E,$33,$C7,$ED,$7F,$C1,$33,$98,$FF,$74,$E0
	.dc.b	$FC,$FB,$F9,$0F,$C1,$FF,$16,$F9,$FE,$D8,$F9,$EB,$5F,$13,$FE,$AE
	.dc.b	$F8,$BF,$F3,$C6,$5C,$F2,$FE,$A9,$97,$2F,$94,$EF,$87,$AE,$73,$87
	.dc.b	$FA,$DB,$E4,$33,$97,$C4,$FC,$25,$F6,$AC,$5F,$92,$3F,$65,$FF,$15
	.dc.b	$FE,$F1,$E4,$CF,$FD,$41,$7B,$AD,$7B,$6F,$DE,$35,$97,$4C,$FF,$DD
	.dc.b	$07,$80,$61,$FE,$A9,$F5,$33,$19,$8E,$BA,$78,$B4,$F9,$BB,$FD,$5C
	.dc.b	$3C,$F3,$9F,$B5,$E1,$CE,$F5,$FF,$DF,$9E,$EB,$F3,$2F,$F7,$AF,$C6
	.dc.b	$FA,$6B,$5F,$8D,$FF,$0A,$9F,$F7,$CF,$ED,$99,$C3,$C7,$BE,$FF,$31
	.dc.b	$BD,$5F,$67,$1E,$39,$F5,$FB,$8F,$0E,$1E,$0C,$BF,$D4,$78,$F7,$FD
	.dc.b	$47,$FC,$4B,$F4,$EF,$0B,$FA,$9F,$B1,$9F,$73,$FC,$1B,$1F,$D5,$BB
	.dc.b	$38,$61,$F1,$36,$F0,$7E,$B5,$E2,$E7,$FC,$B7,$03,$4E,$B8,$FD,$9C
	.dc.b	$3A,$69,$DF,$BF,$E2,$E9,$FE,$97,$E9,$FE,$19,$D4,$F7,$FE,$24,$E7
	.dc.b	$DF,$FA,$BF,$B7,$7C,$7F,$CC,$7C,$19,$FE,$AB,$6F,$F1,$FB,$E5,$DF
	.dc.b	$C7,$87,$5F,$DE,$3F,$CC,$F3,$D3,$F0,$0F,$C5,$64,$FC,$63,$EF,$7E
	.dc.b	$70,$FB,$C2,$FC,$91,$FA,$83,$F1,$BF,$E0,$7D,$3F,$A9,$AE,$06,$E7
	.dc.b	$5E,$1F,$E5,$61,$F9,$57,$A3,$5F,$C7,$C7,$77,$D0,$7F,$04,$79,$E3
	.dc.b	$E0,$FB,$1F,$2E,$7E,$48,$F1,$F8,$FE,$8E,$CF,$F5,$6F,$A8,$FD,$D1
	.dc.b	$FC,$01,$FE,$CF,$DF,$CD,$5A,$9C,$FE,$91,$E5,$F0,$7C,$9D,$BF,$02
	.dc.b	$AE,$DF,$6F,$F0,$9F,$EB,$F8,$3D,$8F,$04,$63,$D1,$F9,$3E,$BB,$F0
	.dc.b	$AF,$91,$FB,$73,$F2,$57,$E6,$3F,$0E,$B5,$C8,$AD,$7A,$B9,$E8,$F8
	.dc.b	$67,$C6,$F2,$7C,$77,$C2,$BF,$D4,$13,$CF,$2E,$FF,$4B,$FB,$7F,$05
	.dc.b	$F5,$6F,$D9,$3F,$AC,$7D,$9B,$F0,$6F,$57,$57,$F2,$1B,$DD,$6A,$5F
	.dc.b	$3B,$FC,$36,$79,$E7,$E4,$E9,$FF,$77,$7D,$6F,$C8,$9E,$3E,$3F,$F9
	.dc.b	$13,$F0,$BF,$16,$BE,$18,$FF,$23,$F8,$5E,$5F,$C9,$BF,$D8,$7F,$79
	.dc.b	$FE,$E0,$FD,$C9,$F6,$FD,$E7,$3F,$89,$25,$FB,$5E,$77,$F1,$6A,$FC
	.dc.b	$A3,$C5,$87,$0F,$AD,$ED,$E0,$7F,$99,$B2,$7B,$3F,$C6,$F5,$7E,$1C
	.dc.b	$FD,$87,$FC,$4F,$EB,$3F,$76,$FF,$5B,$75,$FF,$45,$BD,$D6,$A5,$F8
	.dc.b	$9F,$F3,$19,$FF,$59,$70,$7C,$57,$AB,$71,$ED,$DD,$FE,$97,$F2,$3C
	.dc.b	$51,$8C,$FF,$C5,$55,$CF,$F7,$8F,$F5,$2F,$94,$FC,$6D,$FA,$37,$F8
	.dc.b	$7F,$7F,$25,$6A,$73,$F2,$7C,$70,$F1,$7D,$97,$E7,$36,$5E,$FF,$33
	.dc.b	$F8,$4F,$1E,$DF,$E7,$DE,$E3,$F4,$97,$EF,$9F,$8A,$FD,$17,$F8,$E7
	.dc.b	$E3,$3F,$D3,$EB,$53,$9F,$59,$F9,$AE,$70,$FD,$5D,$3F,$4A,$B1,$D3
	.dc.b	$C7,$64,$FC,$E7,$C3,$39,$8F,$F1,$EF,$27,$F7,$5F,$DA,$BF,$C2,$3F
	.dc.b	$A1,$7F,$12,$F9,$9F,$F4,$6F,$E6,$BF,$D8,$67,$3F,$0D,$FD,$7F,$F0
	.dc.b	$A7,$DF,$67,$F3,$8F,$DE,$D7,$F8,$6B,$AB,$F3,$4F,$0C,$7F,$89,$67
	.dc.b	$DF,$EA,$1A,$77,$F4,$BF,$B1,$5F,$E3,$FC,$47,$ED,$AF,$BE,$7E,$80
	.dc.b	$FC,$BB,$C9,$E8,$FF,$9C,$AB,$52,$FC,$EF,$D1,$1E,$49,$98,$74,$C1
	.dc.b	$F0,$1E,$09,$E5,$0F,$C7,$BF,$BA,$7E,$3B,$F9,$2F,$9B,$7E,$60,$FD
	.dc.b	$FD,$F9,$B3,$F4,$8F,$F0,$35,$EE,$B5,$3F,$C5,$F9,$7F,$C4,$53,$F9
	.dc.b	$CE,$7F,$DF,$3F,$E7,$E6,$FA,$7F,$E1,$4B,$ED,$F8,$77,$8A,$31,$97
	.dc.b	$F1,$EA,$FF,$0A,$79,$1F,$E5,$0B,$FF,$25,$FE,$DE,$FC,$CD,$FB,$36
	.dc.b	$BE,$EB,$D3,$F7,$2A,$D7,$E5,$39,$CB,$F8,$86,$4F,$C0,$7F,$9B,$BD
	.dc.b	$6F,$87,$93,$FE,$17,$8C,$60,$FA,$4F,$F0,$D7,$9D,$FE,$F9,$F1,$1F
	.dc.b	$9B,$3F,$16,$7E,$05,$F2,$B9,$FF,$C8,$97,$BA,$D4,$FF,$81,$E7,$3F
	.dc.b	$91,$FF,$48,$A7,$E9,$B6,$3F,$C6,$17,$D7,$C1,$F0,$C6,$31,$FE,$73
	.dc.b	$87,$F9,$47,$F6,$57,$E9,$DF,$F0,$03,$E1,$3F,$CE,$1E,$0B,$F4,$7F
	.dc.b	$F1,$F5,$6B,$F0,$DF,$E2,$DF,$80,$FE,$0F,$EC,$F3,$8F,$F5,$AA,$FF
	.dc.b	$39,$F4,$FF,$37,$78,$65,$D2,$37,$87,$2F,$81,$FF,$1F,$7A,$FF,$F0
	.dc.b	$03,$F4,$77,$EC,$7F,$E9,$FF,$D4,$9F,$BF,$7F,$93,$EB,$53,$9F,$96
	.dc.b	$C7,$E0,$5F,$DA,$A7,$FA,$3E,$7F,$E4,$EE,$3F,$F3,$74,$70,$F9,$8F
	.dc.b	$E3,$F1,$8A,$67,$E0,$9F,$E6,$F7,$FE,$5D,$FC,$9F,$F1,$1F,$BB,$FF
	.dc.b	$A6,$76,$FE,$89,$FA,$46,$7C,$94,$E7,$E6,$DF,$BA,$F1,$FD,$43,$FB
	.dc.b	$EE,$3F,$91,$CB,$FA,$D3,$D3,$F0,$D6,$7E,$4F,$94,$FF,$48,$4E,$63
	.dc.b	$C7,$C1,$FE,$87,$FC,$AD,$FA,$13,$E4,$DE,$5B,$C3,$7E,$B4,$FD,$EF
	.dc.b	$FC,$ED,$87,$0B,$DD,$6B,$F6,$4C,$E3,$FE,$12,$A7,$E0,$B6,$3B,$65
	.dc.b	$D7,$E5,$BF,$CE,$95,$CE,$3C,$1F,$2B,$FE,$77,$FD,$79,$4F,$D0,$1F
	.dc.b	$9C,$3F,$00,$FE,$0B,$D5,$FE,$89,$BD,$D6,$A7,$F8,$4E,$D9,$FC,$9F
	.dc.b	$FC,$4B,$5D,$3A,$3F,$C0,$4F,$0C,$7A,$7C,$07,$87,$3F,$04,$77,$FB
	.dc.b	$5F,$F4,$57,$CA,$7F,$BA,$7D,$1F,$F7,$C7,$E3,$FF,$EC,$DF,$6B,$FD
	.dc.b	$35,$39,$FC,$9B,$FD,$C9,$E9,$7F,$9A,$FB,$3F,$D6,$D3,$E9,$DB,$FC
	.dc.b	$24,$FF,$BC,$DB,$E3,$7E,$EE,$FF,$DD,$9D,$DE,$08,$C6,$5F,$CA,$6B
	.dc.b	$FD,$51,$F4,$4F,$C3,$7F,$B6,$FC,$FB,$F3,$FF,$E2,$BF,$95,$6A,$73
	.dc.b	$1F,$C6,$36,$E3,$FB,$97,$F9,$23,$0E,$58,$FF,$89,$5E,$9F,$4D,$FD
	.dc.b	$AA,$FF,$9D,$CF,$38,$C7,$37,$FA,$B7,$E6,$DF,$B3,$3F,$5B,$FF,$38
	.dc.b	$FE,$79,$FE,$73,$FC,$3F,$FD,$7D,$5A,$9C,$F8,$4F,$D7,$F3,$FB,$8F
	.dc.b	$F9,$6A,$B9,$E9,$FE,$37,$78,$7E,$08,$CB,$FA,$FF,$9F,$FE,$ED,$9C
	.dc.b	$C7,$DD,$EB,$FF,$72,$7E,$19,$FD,$73,$F4,$27,$E6,$AF,$D4,$DF,$02
	.dc.b	$FC,$6F,$FD,$A5,$7B,$AD,$4B,$F7,$5C,$FE,$F5,$FE,$6E,$AF,$EC,$FB
	.dc.b	$FF,$92,$DE,$18,$FF,$A9,$F3,$F0,$FD,$23,$4F,$0C,$BF,$11,$FF,$6A
	.dc.b	$7E,$6C,$FC,$29,$FB,$9F,$F3,$2F,$B1,$7E,$E9,$FD,$D3,$3F,$F7,$F5
	.dc.b	$EE,$B5,$2F,$C3,$3F,$DD,$53,$CF,$6F,$F2,$F3,$FD,$6D,$FE,$35,$CD
	.dc.b	$FE,$AA,$FC,$19,$3C,$FE,$5B,$FC,$51,$CD,$E4,$D7,$C7,$1F,$E2,$BF
	.dc.b	$C0,$7F,$DD,$5F,$A4,$FE,$3F,$F7,$0F,$DA,$FE,$17,$FE,$3D,$FC,$07
	.dc.b	$FD,$F5,$39,$F5,$FF,$CF,$1F,$80,$3F,$29,$CF,$9D,$FE,$96,$A7,$9B
	.dc.b	$61,$97,$4F,$4F,$BC,$63,$1F,$F4,$17,$D6,$FF,$BE,$FD,$C7,$E5,$0F
	.dc.b	$C6,$1F,$AE,$3F,$55,$FB,$FF,$A6,$BB,$9D,$5F,$07,$1E,$38,$70,$BF
	.dc.b	$C2,$FE,$53,$BA,$D6,$5F,$EA,$CB,$7E,$C5,$63,$C1,$9F,$7F,$A9,$F2
	.dc.b	$46,$31,$FF,$24,$77,$73,$03,$BB,$F4,$DF,$45,$FD,$03,$F7,$77,$DB
	.dc.b	$3F,$03,$FD,$5F,$0F,$A3,$0E,$38,$70,$F8,$C6,$1C,$2D,$EA,$77,$BE
	.dc.b	$DE,$BF,$F7,$17,$FD,$40,$B3,$FF,$63,$B7,$0E,$79,$BE,$FF,$CF,$A7
	.dc.b	$87,$0E,$78,$79,$EB,$F6,$31,$E4,$FD,$15,$FA,$83,$A3,$FC,$17,$CB
	.dc.b	$FE,$C9,$F8,$4F,$27,$0A,$F7,$5A,$FC,$6F,$F0,$3A,$77,$AD,$7E,$A9
	.dc.b	$FE,$DC,$BE,$B9,$FF,$AE,$5E,$3E,$43,$3F,$F0,$9F,$57,$E7,$67,$30
	.dc.b	$FA,$DF,$5C,$FE,$87,$DE,$7E,$3A,$FD,$0B,$FC,$F7,$EF,$BF,$45,$FE
	.dc.b	$29,$BF,$8E,$F7,$5A,$97,$C1,$7B,$27,$9E,$DF,$EC,$D7,$FC,$66,$DC
	.dc.b	$7B,$7D,$3B,$2F,$24,$F1,$9C,$C7,$AA,$3E,$FF,$E1,$4F,$C9,$BF,$6F
	.dc.b	$D4,$FE,$00,$FD,$CF,$85,$F3,$2B,$5E,$B7,$D9,$9F,$DE,$F8,$D9,$FC
	.dc.b	$A7,$F2,$39,$7F,$02,$3A,$E5,$87,$3C,$5F,$47,$1E,$79,$F7,$F8,$0F
	.dc.b	$93,$5C,$33,$ED,$E6,$62,$3F,$39,$F8,$F7,$F9,$EB,$E7,$DF,$AC,$BF
	.dc.b	$63,$FE,$F3,$F9,$77,$E0,$AB,$53,$9F,$9B,$FD,$0E,$DF,$81,$9F,$F4
	.dc.b	$1E,$3F,$EF,$77,$F9,$1B,$7F,$03,$FF,$06,$DF,$5F,$9C,$F1,$78,$6B
	.dc.b	$EB,$9B,$E4,$73,$FC,$02,$FC,$8B,$FC,$19,$DF,$FC,$2B,$DC,$FF,$5F
	.dc.b	$6D,$F9,$1A,$D7,$D1,$7E,$0F,$B1,$5F,$EC,$DA,$D7,$E7,$3F,$27,$87
	.dc.b	$AE,$55,$FC,$FE,$3C,$B2,$FE,$2F,$CB,$AC,$FA,$73,$FA,$3D,$8F,$DC
	.dc.b	$FC,$E3,$F0,$4F,$F7,$AF,$7F,$F1,$DD,$BF,$49,$5A,$9C,$F0,$7E,$B2
	.dc.b	$3F,$80,$EB,$E3,$AD,$F4,$F8,$DE,$3E,$AB,$2E,$DF,$84,$7F,$57,$F2
	.dc.b	$7F,$A2,$63,$D6,$3F,$C8,$F1,$FD,$1F,$F3,$1F,$F6,$27,$C0,$3F,$42
	.dc.b	$FF,$23,$7E,$4B,$F1,$33,$FD,$C5,$6A,$73,$FA,$57,$F0,$08,$FF,$66
	.dc.b	$6D,$E5,$C7,$A7,$BD,$E5,$7B,$7E,$29,$FE,$5A,$C3,$FC,$A5,$F7,$7F
	.dc.b	$73,$0E,$9F,$54,$FC,$9F,$EF,$E8,$FD,$D3,$FC,$41,$F8,$5B,$F7,$E7
	.dc.b	$EB,$0F,$C3,$FF,$88,$DE,$FE,$9D,$7B,$AF,$F5,$A4,$FF,$11,$1F,$B2
	.dc.b	$79,$78,$BF,$98,$C7,$77,$D4,$63,$2F,$0E,$5F,$E3,$59,$E5,$18,$E5
	.dc.b	$FC,$AF,$E8,$0F,$DF,$3F,$9C,$E4,$FD,$59,$F9,$3F,$E5,$71,$FE,$45
	.dc.b	$7B,$F8,$DF,$5D,$6B,$E5,$FF,$95,$CF,$F6,$DD,$7C,$67,$F4,$D7,$B3
	.dc.b	$8F,$3D,$1C,$F2,$F0,$73,$79,$34,$F1,$E3,$FB,$7C,$3F,$9B,$E8,$FD
	.dc.b	$EB,$F8,$5F,$E3,$AF,$D8,$DF,$A6,$EA,$73,$0F,$DC,$38,$FE,$01,$F2
	.dc.b	$71,$FE,$C1,$8F,$37,$C8,$F5,$F6,$FF,$B6,$E5,$FD,$46,$7A,$46,$3E
	.dc.b	$13,$F7,$BE,$27,$F3,$BE,$57,$EB,$2F,$C9,$3F,$D7,$BE,$BF,$F4,$EA
	.dc.b	$D4,$E7,$F0,$F7,$D0,$E7,$D5,$F2,$74,$AF,$F9,$9E,$BF,$C9,$4F,$EE
	.dc.b	$CF,$2C,$9C,$FE,$2F,$8B,$FF,$06,$67,$E1,$F1,$3F,$88,$FE,$9C,$FD
	.dc.b	$F9,$F9,$2F,$F6,$BD,$9F,$BB,$BF,$58,$FE,$C7,$7B,$AD,$4B,$EF,$F3
	.dc.b	$87,$9A,$BF,$D6,$5B,$FD,$B7,$87,$C4,$7B,$3E,$7F,$C3,$5E,$08,$F7
	.dc.b	$FE,$3F,$FC,$93,$CD,$FF,$2F,$7D,$AB,$E3,$1F,$8A,$7F,$78,$FC,$73
	.dc.b	$FB,$AD,$6A,$73,$F3,$DF,$C4,$FB,$7F,$BA,$4F,$F9,$9F,$6F,$C6,$3F
	.dc.b	$E6,$56,$FD,$77,$F9,$9E,$F8,$C7,$48,$C6,$3F,$E0,$58,$FF,$01,$FE
	.dc.b	$28,$F8,$47,$E7,$CF,$D3,$9F,$89,$7F,$8C,$3F,$60,$DD,$6A,$73,$F1
	.dc.b	$0F,$DA,$73,$97,$E7,$F2,$7E,$5B,$F7,$9E,$D9,$F5,$CF,$F9,$05,$7F
	.dc.b	$8B,$27,$38,$F8,$3F,$36,$FE,$97,$FB,$5F,$F8,$0F,$F0,$C7,$E2,$9F
	.dc.b	$E9,$7F,$3F,$FC,$B3,$CB,$FE,$17,$AD,$4E,$7E,$0F,$FC,$3F,$1F,$E0
	.dc.b	$4C,$FF,$19,$E5,$23,$FC,$1D,$1F,$B5,$78,$FE,$0F,$F8,$E8,$E3,$E6
	.dc.b	$33,$F0,$CF,$ED,$72,$FF,$13,$7E,$36,$FC,$D3,$FD,$ED,$F9,$47,$FA
	.dc.b	$AF,$ED,$5F,$B8,$7F,$8C,$2B,$5F,$8C,$AB,$5E,$B6,$70,$FD,$1D,$3F
	.dc.b	$41,$B1,$D3,$3F,$8F,$B7,$F9,$66,$B9,$E1,$FE,$B9,$E4,$FE,$E3,$F8
	.dc.b	$17,$1F,$D4,$1F,$9D,$3F,$63,$7E,$C6,$FD,$A3,$FC,$87,$7B,$AD,$4F
	.dc.b	$BF,$8F,$FC,$85,$3F,$77,$4F,$E3,$4F,$95,$BB,$B2,$EF,$97,$F0,$9B
	.dc.b	$E3,$39,$C7,$C5,$F6,$FF,$E0,$6F,$A2,$FF,$61,$FE,$C2,$FD,$D9,$F5
	.dc.b	$8F,$DD,$7F,$99,$7C,$CF,$FA,$9E,$F7,$5A,$97,$C7,$FF,$97,$27,$EE
	.dc.b	$6F,$FC,$BD,$FF,$67,$B7,$F1,$36,$5D,$F2,$FF,$61,$4F,$19,$CE,$3E
	.dc.b	$2E,$CF,$F0,$DF,$E4,$FF,$CB,$BF,$0A,$7E,$F3,$FE,$57,$F6,$7F,$D4
	.dc.b	$3C,$7F,$E6,$7A,$D4,$E7,$CE,$FF,$37,$C7,$F3,$6C,$FF,$A1,$3F,$E8
	.dc.b	$86,$FC,$7B,$1E,$BF,$73,$FE,$00,$9F,$F4,$7D,$74,$87,$E8,$7F,$E2
	.dc.b	$EF,$D8,$9F,$B5,$7F,$5B,$F6,$DF,$AF,$7F,$68,$FB,$8F,$F3,$FD,$6B
	.dc.b	$E5,$D5,$AF,$9E,$7E,$E9,$AC,$7F,$99,$DF,$8F,$5F,$EB,$0F,$1B,$EB
	.dc.b	$F2,$5E,$1A,$F0,$66,$FB,$5F,$F2,$0F,$4F,$6F,$C5,$5F,$96,$BF,$1C
	.dc.b	$7E,$7A,$FC,$1B,$FD,$27,$7B,$AD,$4B,$F4,$7C,$EB,$FD,$1A,$9F,$A0
	.dc.b	$58,$FF,$7D,$5F,$5F,$DA,$7C,$31,$8C,$BF,$17,$C3,$FE,$A0,$E2,$FF
	.dc.b	$28,$7E,$FA,$FD,$EF,$CB,$F6,$FF,$AF,$F0,$3F,$D4,$15,$A9,$CF,$AD
	.dc.b	$C7,$C7,$FE,$AF,$3F,$E5,$9C,$7F,$BE,$3B,$D7,$F6,$BE,$EF,$CC,$FB
	.dc.b	$C6,$9D,$E7,$EE,$5F,$FB,$47,$F7,$5F,$CF,$7E,$E4,$F9,$17,$E9,$EF
	.dc.b	$D1,$BF,$6D,$C1,$FE,$B3,$BD,$D6,$A7,$F6,$DE,$8F,$FA,$33,$55,$7E
	.dc.b	$A7,$2F,$F0,$82,$FD,$5B,$FB,$7B,$CB,$27,$37,$F9,$E2,$31,$F8,$CF
	.dc.b	$F9,$EB,$F7,$47,$EC,$AF,$DF,$1F,$8C,$BF,$80,$7C,$D9,$7F,$B3,$2F
	.dc.b	$75,$A9,$F5,$7D,$3E,$70,$FE,$FF,$4F,$DC,$2C,$75,$C7,$A6,$9E,$18
	.dc.b	$C6,$0C,$FF,$D3,$DC,$0F,$D1,$9F,$92,$3E,$2B,$FD,$87,$F0,$DF,$E2
	.dc.b	$4E,$6F,$F7,$95,$EE,$B5,$F8,$B2,$BF,$C5,$3F,$39,$AC,$79,$F3,$7F
	.dc.b	$8E,$B0,$ED,$FA,$CF,$F8,$A5,$E5,$9F,$6F,$D4,$7C,$77,$E2,$8F,$0F
	.dc.b	$D2,$79,$BF,$77,$7E,$73,$FC,$C3,$F5,$9F,$D1,$7E,$F6,$AF,$2D,$6B
	.dc.b	$E3,$95,$AF,$D7,$F3,$F2,$BE,$1A,$E7,$BF,$95,$E1,$5F,$AD,$FA,$7F
	.dc.b	$06,$7D,$FC,$EF,$EA,$DC,$1F,$ED,$01,$F9,$8F,$FB,$CF,$E0,$9F,$AA
	.dc.b	$BE,$4D,$FA,$87,$F2,$CF,$29,$87,$1C,$38,$5E,$EB,$F9,$AE,$C3,$73
	.dc.b	$FE,$3A,$7F,$DB,$75,$C7,$E6,$FE,$54,$F2,$C7,$F7,$75,$F5,$A3,$1E
	.dc.b	$FF,$DE,$FB,$37,$E6,$1F,$D1,$3D,$87,$EE,$8F,$D4,$9F,$97,$FD,$EB
	.dc.b	$DF,$BB,$FC,$7A,$B5,$F9,$4E,$71,$FB,$54,$FC,$DA,$C7,$5A,$FF,$00
	.dc.b	$71,$FC,$0D,$BD,$9C,$3F,$E9,$FA,$D6,$1F,$92,$F9,$47,$E0,$0F,$1F
	.dc.b	$F4,$F5,$FE,$B5,$FD,$8F,$F2,$35,$A9,$CF,$D3,$3E,$93,$38,$7E,$36
	.dc.b	$9F,$BD,$D8,$E9,$E8,$7C,$2F,$9D,$F0,$46,$33,$78,$1F,$A4,$F8,$07
	.dc.b	$ED,$2F,$A5,$FF,$4E,$7C,$E3,$F0,$27,$E6,$7F,$A4,$BD,$D6,$A5,$F8
	.dc.b	$9E,$7F,$51,$FD,$05,$74,$E7,$FD,$23,$C3,$F0,$BF,$EF,$1E,$FF,$87
	.dc.b	$4F,$04,$7F,$7B,$E8,$FD,$BF,$E0,$2F,$D5,$5F,$BA,$3E,$1D,$FB,$EB
	.dc.b	$9B,$FD,$83,$DD,$F9,$7C,$38,$5E,$EB,$5F,$B8,$7E,$F2,$7A,$6F,$FB
	.dc.b	$75,$FB,$5F,$ED,$DE,$39,$76,$CB,$C7,$9F,$8A,$3C,$3F,$81,$7F,$0C
	.dc.b	$FD,$63,$FD,$9F,$F9,$B3,$F4,$E7,$ED,$3F,$E9,$8F,$D9,$7F,$85,$D6
	.dc.b	$A7,$3F,$86,$7F,$9E,$FC,$8F,$C4,$6F,$FA,$8A,$FF,$6C,$BF,$C2,$D7
	.dc.b	$EB,$BF,$C2,$DE,$58,$BD,$1F,$17,$D2,$B4,$F1,$4F,$F6,$DC,$18,$F6
	.dc.b	$FD,$AD,$FA,$F7,$E8,$BF,$74,$FF,$24,$79,$CF,$CC,$9F,$B6,$70,$EC
	.dc.b	$FD,$B6,$4C,$78,$FA,$AC,$78,$E0,$FD,$E7,$85,$FE,$F6,$F6,$FF,$B7
	.dc.b	$9F,$0F,$AB,$FE,$CD,$79,$DF,$8B,$E4,$BF,$82,$6D,$E7,$DB,$CD,$F7
	.dc.b	$4D,$FF,$BA,$CF,$AF,$F2,$AF,$E1,$5F,$29,$FF,$00,$3F,$69,$FF,$2B
	.dc.b	$7E,$85,$E2,$DB,$D1,$5A,$E6,$57,$F9,$6F,$F4,$FD,$6E,$AF,$C7,$DF
	.dc.b	$9D,$F8,$2F,$FB,$36,$34,$FF,$5B,$BE,$7D,$9E,$8E,$5F,$8F,$5A,$CF
	.dc.b	$C9,$F8,$76,$70,$F8,$9F,$BC,$7E,$C3,$EC,$FF,$C6,$BF,$B7,$FF,$B5
	.dc.b	$BE,$7F,$C7,$39,$FB,$64,$FF,$B6,$BE,$73,$3E,$87,$C9,$A7,$A7,$61
	.dc.b	$8F,$4F,$4B,$C1,$9F,$D9,$BE,$98,$FF,$3E,$8F,$AD,$EC,$7F,$64,$FD
	.dc.b	$5D,$F9,$EB,$F6,$CF,$E0,$3F,$A4,$7C,$6A,$D4,$E7,$DC,$7E,$CF,$9F
	.dc.b	$9E,$FB,$95,$FD,$93,$2F,$90,$F0,$C7,$FD,$95,$DD,$F9,$38,$C7,$BF
	.dc.b	$E6,$E8,$7E,$D2,$FD,$71,$FB,$47,$93,$EB,$7F,$D3,$DF,$6B,$F6,$2B
	.dc.b	$53,$9F,$1B,$FA,$0E,$7F,$9F,$9F,$CF,$EB,$F9,$17,$A3,$71,$EB,$C2
	.dc.b	$C9,$F8,$EF,$8E,$FA,$CE,$7E,$97,$E5,$FE,$DD,$FE,$E1,$E9,$F3,$7C
	.dc.b	$13,$F4,$EF,$D7,$F8,$FF,$91,$AD,$68,$AD,$6C,$FC,$CF,$5E,$27,$F8
	.dc.b	$EB,$83,$F1,$E6,$33,$AD,$C3,$96,$7F,$CE,$A3,$18,$F8,$7F,$42,$F2
	.dc.b	$FE,$8D,$FE,$B7,$FD,$8D,$F9,$2F,$F8,$87,$ED,$DF,$9B,$79,$6B,$53
	.dc.b	$9F,$C5,$F1,$F8,$B7,$83,$87,$F8,$AF,$12,$FF,$DB,$9A,$F9,$A7,$31
	.dc.b	$CB,$EB,$DF,$61,$E2,$C1,$AF,$8A,$5F,$95,$7C,$AF,$C4,$3F,$1F,$FB
	.dc.b	$0B,$F4,$D7,$F0,$0F,$F8,$33,$E5,$BE,$55,$6A,$73,$F5,$3F,$D5,$36
	.dc.b	$F2,$CB,$F0,$EF,$47,$B1,$FC,$63,$89,$3F,$9C,$E2,$F4,$4E,$63,$A6
	.dc.b	$7F,$E0,$CC,$39,$F8,$0D,$7C,$93,$FC,$BF,$0F,$B9,$FA,$A7,$F3,$FF
	.dc.b	$C9,$FF,$43,$F8,$9B,$F2,$4F,$F1,$B6,$5F,$22,$B5,$E8,$2B,$5F,$BF
	.dc.b	$72,$F7,$BC,$F9,$7F,$88,$F8,$18,$3F,$64,$CC,$A7,$34,$D3,$C3,$F6
	.dc.b	$5F,$94,$C3,$BE,$5F,$ED,$EC,$3F,$91,$7E,$9D,$FD,$EF,$B7,$FD,$55
	.dc.b	$E3,$7F,$A2,$3F,$80,$7F,$C7,$BC,$1F,$D4,$AB,$5F,$88,$AB,$5D,$7C
	.dc.b	$BE,$8F,$FC,$F5,$97,$CA,$3F,$C7,$57,$D5,$C0,$79,$D7,$FA,$58,$FE
	.dc.b	$26,$E3,$F0,$6F,$E2,$53,$BF,$C5,$9F,$96,$7F,$AB,$FD,$5F,$17,$E2
	.dc.b	$4F,$D0,$FD,$75,$AF,$C4,$BF,$82,$E7,$FE,$C1,$FD,$33,$3F,$79,$F2
	.dc.b	$35,$5F,$4D,$3C,$B1,$8D,$3B,$61,$FA,$BE,$1F,$85,$7D,$75,$F8,$5F
	.dc.b	$49,$E8,$8C,$57,$FB,$3F,$F5,$2F,$E9,$DF,$B8,$7F,$55,$FC,$23,$FD
	.dc.b	$4B,$F0,$CF,$C5,$1F,$B8,$FF,$B2,$74,$FE,$FF,$7B,$F9,$95,$EE,$9F
	.dc.b	$AE,$74,$F9,$9F,$F4,$B7,$17,$CC,$C3,$FB,$17,$E5,$9F,$84,$79,$61
	.dc.b	$DF,$F3,$BF,$E7,$F1,$8D,$1F,$A7,$E3,$F5,$27,$EE,$0F,$D1,$9F,$A6
	.dc.b	$F9,$FF,$75,$7E,$AB,$EC,$BD,$D6,$B1,$53,$F1,$AD,$63,$E3,$B6,$76
	.dc.b	$3B,$65,$E5,$DF,$79,$73,$D3,$FA,$AD,$70,$F4,$9E,$3B,$F4,$27,$E8
	.dc.b	$1F,$EC,$AF,$E0,$1F,$EF,$97,$BA,$D4,$FF,$B3,$BE,$1F,$3A,$BF,$18
	.dc.b	$FD,$4D,$F5,$D7,$F2,$AF,$1C,$BA,$FC,$47,$8A,$FF,$88,$E5,$D3,$BF
	.dc.b	$CB,$FB,$A7,$FB,$D3,$F8,$05,$FC,$02,$DF,$FD,$47,$F4,$7E,$4A,$D4
	.dc.b	$E7,$C9,$7C,$96,$7F,$58,$F1,$D7,$FB,$FB,$83,$C6,$F0,$C7,$FA,$34
	.dc.b	$73,$FC,$95,$9F,$82,$7F,$D1,$FF,$07,$FE,$34,$FD,$D5,$E7,$FF,$A0
	.dc.b	$BF,$5D,$FE,$B5,$F9,$1F,$F2,$3E,$0F,$F1,$15,$6B,$F0,$C5,$6B,$E2
	.dc.b	$3C,$BE,$67,$F9,$4A,$B9,$F9,$9F,$C4,$1E,$1B,$FE,$F1,$E5,$7D,$BF
	.dc.b	$23,$FF,$06,$C6,$3E,$73,$FD,$FF,$E1,$63,$F4,$AF,$E0,$5D,$CF,$D8
	.dc.b	$3F,$D2,$3F,$B1,$3F,$6F,$F2,$D6,$A7,$3F,$9E,$63,$E2,$3C,$53,$FC
	.dc.b	$2F,$6F,$13,$BF,$A5,$FC,$D4,$72,$F7,$3F,$3B,$87,$FA,$7B,$F0,$1F
	.dc.b	$F1,$0F,$B9,$FD,$43,$F0,$67,$EC,$1F,$C2,$FE,$35,$F9,$BF,$C4,$E8
	.dc.b	$FF,$39,$DE,$EB,$52,$FD,$5F,$FC,$99,$2F,$C8,$7F,$BA,$EB,$8F,$99
	.dc.b	$FF,$4B,$3F,$C8,$A7,$97,$C3,$7F,$3B,$FB,$6F,$1D,$7F,$A6,$F2,$F1
	.dc.b	$7E,$55,$FE,$36,$FB,$E7,$DF,$3F,$78,$78,$AF,$DF,$1F,$A5,$BF,$38
	.dc.b	$FF,$51,$71,$FF,$A2,$B1,$E3,$87,$0B,$DD,$7F,$97,$6B,$58,$BF,$6E
	.dc.b	$FE,$9B,$87,$0E,$1F,$E7,$2F,$2C,$FC,$19,$FF,$A2,$F0,$E3,$9F,$58
	.dc.b	$FE,$C9,$18,$9D,$E4,$FC,$7F,$FC,$E5,$87,$7F,$CA,$7F,$BE,$FE,$77
	.dc.b	$FE,$D8,$FD,$33,$FC,$9D,$8F,$A7,$55,$6B,$CC,$57,$FB,$0E,$B5,$F9
	.dc.b	$77,$A2,$F7,$BF,$A1,$E3,$F0,$2C,$BF,$DC,$BF,$65,$F3,$67,$31,$FE
	.dc.b	$E8,$FA,$0C,$69,$DF,$3F,$E7,$19,$FF,$AD,$3E,$CB,$FD,$EF,$F8,$9F
	.dc.b	$EC,$BF,$02,$75,$F9,$BE,$47,$FD,$1D,$5A,$F6,$7F,$CE,$33,$EA,$DB
	.dc.b	$3F,$90,$7F,$82,$78,$7F,$D1,$79,$F5,$86,$7F,$DF,$B1,$E5,$9F,$6C
	.dc.b	$3F,$CF,$31,$8C,$7F,$9D,$E3,$FE,$9D,$FA,$D7,$E3,$1F,$CC,$FE,$55
	.dc.b	$F9,$6B,$F0,$9F,$3F,$8F,$BA,$D4,$E7,$1F,$D0,$7B,$DF,$EB,$EC,$79
	.dc.b	$FC,$A7,$F8,$41,$EB,$F2,$57,$0E,$39,$7C,$2D,$3A,$4F,$F5,$8C,$BC
	.dc.b	$1F,$A1,$7F,$A4,$3F,$75,$FC,$C7,$DE,$FE,$9F,$CA,$7E,$ED,$FE,$63
	.dc.b	$FD,$9B,$FD,$6D,$5A,$F3,$FF,$C8,$B3,$FE,$67,$FD,$DF,$3E,$67,$EF
	.dc.b	$B5,$F9,$3C,$5F,$ED,$4C,$3A,$C7,$F8,$41,$E5,$97,$6F,$94,$F1,$5F
	.dc.b	$FA,$33,$2E,$BF,$1F,$FE,$A4,$F4,$1F,$B8,$FE,$EF,$57,$EC,$DF,$AF
	.dc.b	$E1,$FF,$63,$F8,$1F,$EF,$6B,$DE,$3E,$AA,$D7,$F0,$07,$FB,$3E,$79
	.dc.b	$FE,$D3,$FE,$38,$77,$C1,$FE,$24,$78,$E5,$DB,$5F,$F1,$B6,$5D,$74
	.dc.b	$F1,$65,$FB,$FF,$B1,$FF,$5C,$F9,$1F,$E4,$3F,$DA,$FF,$CB,$7F,$6A
	.dc.b	$FD,$D1,$FB,$53,$F4,$9F,$F7,$05,$EF,$EF,$3F,$D6,$B5,$AF,$C3,$73
	.dc.b	$F0,$0C,$BF,$C9,$B7,$FE,$4F,$E6,$B3,$9C,$38,$FE,$19,$E2,$8C,$69
	.dc.b	$E1,$FC,$8B,$FD,$9B,$F9,$D7,$F5,$6F,$C9,$DF,$9D,$3F,$28,$7E,$EE
	.dc.b	$A7,$EC,$1F,$EF,$57,$85,$6E,$75,$F6,$19,$F5,$98,$FF,$98,$AF,$FB
	.dc.b	$1F,$E5,$76,$73,$87,$1E,$1F,$1C,$63,$3F,$0F,$A9,$8C,$7C,$B1,$E4
	.dc.b	$C5,$FB,$63,$B1,$E4,$3C,$19,$F8,$D3,$D9,$FF,$54,$FE,$70,$EB,$7E
	.dc.b	$BD,$ED,$FC,$4F,$96,$3C,$71,$63,$C7,$F1,$36,$3C,$70,$F8,$9E,$D7
	.dc.b	$0B,$EF,$9F,$FA,$D9,$FF,$35,$E1,$E1,$FC,$A7,$FD,$06,$F3,$C3,$97
	.dc.b	$27,$FB,$FA,$31,$AF,$93,$8F,$FE,$00,$FE,$23,$FE,$6C,$F3,$5F,$AC
	.dc.b	$4F,$9B,$7E,$52,$FD,$43,$C5,$B3,$0E,$17,$BA,$FB,$BC,$BF,$1A,$BE
	.dc.b	$35,$7F,$9E,$3D,$0F,$F4,$BB,$DE,$3F,$D0,$6F,$2B,$62,$F9,$8F,$2F
	.dc.b	$E0,$7F,$DA,$E7,$C9,$E3,$7D,$1C,$BB,$62,$FD,$C7,$FD,$45,$F9,$33
	.dc.b	$F7,$67,$E4,$2F,$DA,$3F,$D9,$5E,$67,$AA,$F7,$5A,$9F,$F6,$77,$CD
	.dc.b	$67,$E8,$D8,$FF,$AB,$AF,$FD,$5D,$FB,$B5,$9C,$E1,$C7,$F0,$3F,$27
	.dc.b	$85,$F7,$6F,$AE,$9E,$1F,$BF,$CE,$DF,$89,$F9,$1F,$F3,$77,$D4,$FF
	.dc.b	$91,$FF,$14,$7E,$79,$FD,$EF,$F0,$CF,$9F,$5A,$FB,$65,$6B,$E0,$3F
	.dc.b	$D0,$7C,$DF,$0A,$BF,$CF,$1F,$7D,$FE,$C4,$7A,$E7,$FE,$B0,$79,$5F
	.dc.b	$6F,$D0,$7C,$73,$FD,$AB,$EC,$3F,$D0,$39,$76,$D1,$FA,$96,$B1,$E3
	.dc.b	$1F,$E0,$FF,$C7,$BF,$07,$F3,$6F,$C9,$DC,$DF,$EE,$CF,$82,$7E,$28
	.dc.b	$FD,$09,$DD,$C2,$F7,$E0,$7F,$17,$AD,$7E,$65,$9C,$7F,$DB,$B9,$75
	.dc.b	$FC,$3E,$E1,$C3,$3E,$BE,$1F,$AA,$BC,$57,$E1,$C7,$F1,$5C,$7F,$98
	.dc.b	$FA,$57,$ED,$CF,$DA,$9F,$B2,$BC,$57,$CA,$7F,$8A,$7F,$09,$E1,$7B
	.dc.b	$AD,$4F,$F8,$B7,$F1,$8C,$E1,$FE,$F2,$C7,$AF,$EE,$37,$0E,$19,$74
	.dc.b	$FA,$FF,$9F,$5E,$2B,$F0,$CF,$83,$E2,$FE,$9F,$DA,$7F,$A7,$7F,$27
	.dc.b	$FA,$5F,$12,$FD,$A3,$FD,$5B,$E3,$7E,$82,$B5,$E2,$FE,$22,$B5,$D5
	.dc.b	$FB,$69,$FF,$53,$7D,$27,$F0,$FE,$5F,$E2,$18,$71,$FC,$3D,$9E,$67
	.dc.b	$C3,$18,$CB,$A7,$83,$8A,$F1,$FE,$9D,$F8,$8F,$35,$F9,$D7,$F2,$6F
	.dc.b	$8A,$7E,$00,$FC,$37,$FD,$47,$F8,$4F,$F0,$8B,$DD,$7F,$2C,$AD,$7E
	.dc.b	$20,$FD,$53,$58,$FF,$42,$DF,$C9,$C1,$FB,$EC,$78,$C6,$3D,$7F,$E5
	.dc.b	$51,$D3,$2F,$0E,$9F,$92,$AE,$71,$DE,$3F,$A8,$46,$23,$79,$3F,$27
	.dc.b	$FE,$3B,$F9,$3B,$F1,$C7,$C0,$3F,$15,$7E,$5E,$F7,$DF,$3E,$FC,$8F
	.dc.b	$8E,$1C,$2F,$75,$AF,$C2,$B3,$F8,$BF,$CF,$DF,$79,$DE,$3F,$E3,$7A
	.dc.b	$F8,$63,$C1,$AF,$7D,$7F,$6F,$39,$CF,$AF,$D3,$7E,$EB,$F1,$C7,$AA
	.dc.b	$F0,$5F,$C0,$2F,$C4,$9F,$B9,$3F,$64,$7E,$BD,$FC,$F3,$1E,$38,$70
	.dc.b	$BD,$D7,$F3,$4E,$0F,$E5,$F5,$DB,$F1,$9F,$47,$6F,$F5,$B8,$E3,$C7
	.dc.b	$F5,$4F,$EC,$CF,$4D,$3C,$5A,$3D,$9F,$46,$DE,$79,$CF,$99,$FC,$53
	.dc.b	$B3,$FD,$25,$F2,$6F,$C3,$9F,$BE,$7D,$2F,$25,$FB,$9F,$F5,$4B,$DD
	.dc.b	$6B,$A9,$4F,$D0,$F5,$F8,$4F,$E6,$38,$7C,$C7,$FB,$E1,$FE,$BE,$E3
	.dc.b	$0E,$79,$77,$8E,$39,$F8,$E3,$F0,$7F,$73,$E7,$BF,$EC,$F9,$F8,$72
	.dc.b	$FE,$1F,$CB,$FD,$5B,$EB,$3F,$AE,$FE,$58,$FD,$D9,$F8,$BF,$F8,$37
	.dc.b	$E2,$5F,$47,$FD,$BA,$F7,$5A,$C3,$F9,$57,$D9,$E7,$E7,$59,$7D,$2E
	.dc.b	$1F,$F0,$A6,$1F,$DA,$BD,$DC,$42,$39,$E9,$FE,$1B,$EA,$FE,$83,$18
	.dc.b	$CB,$F9,$39,$E7,$C7,$AE,$7E,$1D,$7F,$C6,$D3,$FD,$C3,$ED,$DF,$B6
	.dc.b	$3D,$CF,$F1,$17,$E4,$0F,$D4,$1F,$AC,$3F,$61,$FF,$04,$E1,$C2,$F7
	.dc.b	$5A,$FD,$7F,$38,$F3,$E1,$FA,$B7,$DB,$C4,$F5,$BC,$7E,$B9,$A7,$7F
	.dc.b	$99,$FE,$B7,$39,$8F,$17,$95,$FD,$7B,$C5,$7E,$E8,$FD,$C3,$FD,$BD
	.dc.b	$FB,$0B,$F6,$7F,$F7,$4F,$BF,$FE,$1E,$BD,$D6,$A7,$FD,$DB,$F5,$39
	.dc.b	$C3,$96,$FF,$8A,$BE,$BF,$BC,$7D,$07,$8F,$88,$CF,$B7,$95,$E3,$C3
	.dc.b	$C5,$39,$C9,$F9,$5F,$F7,$4F,$C8,$DF,$06,$F1,$5F,$BB,$3F,$56,$7E
	.dc.b	$50,$FD,$19,$F9,$67,$F8,$DB,$1E,$38,$70,$BD,$D3,$F2,$0D,$65,$D7
	.dc.b	$8F,$F4,$78,$78,$3F,$0D,$FC,$4B,$CB,$4F,$06,$8F,$CD,$FC,$D8,$79
	.dc.b	$67,$3F,$0B,$8D,$7C,$17,$E5,$EE,$FF,$24,$FC,$47,$F8,$DF,$A9,$FB
	.dc.b	$67,$E2,$7F,$08,$FF,$20,$FE,$5D,$AF,$B3,$FA,$D3,$9F,$3B,$F9,$97
	.dc.b	$99,$FE,$40,$C7,$FC,$81,$DF,$F9,$E7,$A7,$27,$EF,$1E,$D7,$D7,$B7
	.dc.b	$FC,$3F,$18,$F2,$BF,$D2,$3F,$27,$8C,$FB,$C7,$FA,$53,$37,$E3,$BF
	.dc.b	$E6,$0E,$EF,$E4,$FF,$39,$F5,$3F,$2F,$FF,$12,$FE,$64,$FD,$5B,$5F
	.dc.b	$14,$AD,$7D,$82,$BF,$C5,$F8,$7F,$30,$F4,$BF,$6A,$F4,$E2,$FE,$26
	.dc.b	$F6,$ED,$67,$FE,$32,$FB,$2E,$39,$F5,$9C,$E5,$F8,$0C,$3F,$CE,$BA
	.dc.b	$FF,$BA,$BF,$55,$FC,$F7,$E0,$AF,$DA,$DF,$AF,$FF,$B1,$7A,$BF,$CD
	.dc.b	$37,$BE,$D5,$EE,$9F,$BD,$74,$FC,$61,$97,$F0,$CC,$3F,$DE,$FF,$BA
	.dc.b	$D9,$CE,$3C,$B3,$78,$9F,$E4,$EB,$ED,$9F,$8F,$C6,$C6,$1C,$23,$C1
	.dc.b	$F9,$CF,$FA,$3B,$F1,$CF,$CE,$F1,$7F,$0D,$FC,$EF,$FA,$17,$E0,$3F
	.dc.b	$D7,$78,$BF,$D0,$35,$AE,$E5,$6B,$F3,$DE,$5E,$49,$FE,$47,$5D,$39
	.dc.b	$BF,$A0,$BC,$6B,$FA,$77,$CE,$7F,$99,$23,$1F,$18,$F9,$AF,$F3,$3F
	.dc.b	$DF,$7F,$81,$7F,$48,$7E,$77,$F2,$3F,$4D,$BF,$AB,$FF,$A9,$AB,$53
	.dc.b	$91,$F8,$26,$7E,$99,$87,$F3,$6A,$FF,$08,$7E,$2B,$67,$37,$D7,$DD
	.dc.b	$F0,$C6,$3A,$19,$F9,$78,$7A,$65,$E4,$9C,$E0,$F1,$7F,$D7,$3F,$AD
	.dc.b	$FF,$BB,$BA,$DF,$93,$BF,$6A,$7E,$E9,$FE,$B1,$F3,$37,$F9,$02,$F7
	.dc.b	$5A,$F8,$77,$E0,$7A,$9F,$E9,$F9,$7F,$BE,$BE,$FE,$E1,$C7,$3E,$D9
	.dc.b	$FF,$4E,$FA,$6F,$F7,$15,$F8,$E7,$3F,$1F,$FE,$94,$8C,$69,$DF,$D4
	.dc.b	$ED,$5C,$B5,$7E,$67,$FE,$D4,$FC,$55,$FA,$3B,$F7,$1F,$ED,$BF,$72
	.dc.b	$F4,$5F,$BB,$7F,$26,$F2,$BF,$D9,$38,$70,$BD,$D6,$BE,$2B,$A6,$3F
	.dc.b	$7B,$B3,$FC,$0F,$7B,$6C,$77,$F9,$FF,$F3,$5D,$FC,$F9,$E3,$87,$4F
	.dc.b	$8E,$FF,$57,$FD,$1B,$F4,$4F,$EC,$B8,$3F,$04,$FE,$D5,$E0,$3F,$7E
	.dc.b	$FF,$77,$DE,$FE,$7F,$FC,$0F,$5A,$F8,$2C,$E3,$FE,$B1,$D3,$FC,$37
	.dc.b	$7B,$6C,$76,$AF,$EE,$DF,$87,$78,$A3,$1E,$A3,$F6,$FF,$F6,$1F,$E0
	.dc.b	$0F,$97,$FF,$2F,$47,$E2,$7F,$53,$FF,$15,$7E,$7E,$E8,$FF,$78,$DE
	.dc.b	$FE,$F3,$97,$94,$BF,$F1,$87,$EE,$3B,$F1,$5C,$13,$0E,$0F,$F8,$C9
	.dc.b	$F0,$79,$DF,$E1,$87,$A6,$9E,$1F,$C2,$7C,$BB,$79,$33,$ED,$86,$34
	.dc.b	$F0,$E9,$FE,$3F,$FC,$6F,$CD,$9F,$82,$9F,$8C,$3F,$22,$7E,$26,$FD
	.dc.b	$77,$FC,$A9,$E7,$3E,$DB,$FD,$87,$C9,$FA,$EC,$7D,$38,$F1,$C3,$87
	.dc.b	$D8,$B0,$E1,$ED,$78,$2F,$79,$F6,$7F,$D3,$7A,$78,$F1,$F1,$7C,$6F
	.dc.b	$FA,$06,$31,$5E,$4D,$7C,$7E,$67,$C8,$D7,$C5,$8F,$AA,$B5,$A7,$84
	.dc.b	$D6,$1C,$B5,$F2,$EB,$EB,$F5,$3C,$98,$7F,$A1,$7F,$71,$FF,$01,$7D
	.dc.b	$6F,$F8,$AB,$A1,$F4,$6F,$D0,$77,$F8,$0A,$F7,$F5,$BF,$87,$A7,$E7
	.dc.b	$BA,$9F,$EA,$59,$7F,$B3,$F8,$7C,$F8,$71,$F4,$7D,$D3,$F8,$28,$C6
	.dc.b	$9E,$1F,$B8,$F4,$CE,$70,$EF,$AB,$F4,$AC,$CF,$A7,$1F,$F2,$8E,$BE
	.dc.b	$3F,$CF,$3D,$BF,$AB,$FF,$34,$FC,$B9,$FB,$07,$FA,$07,$E6,$9F,$A0
	.dc.b	$3E,$EB,$7F,$0E,$BD,$FE,$05,$FD,$7A,$9F,$B2,$EA,$7D,$3B,$33,$FE
	.dc.b	$83,$F8,$7E,$1F,$06,$9D,$F0,$F3,$C6,$31,$FE,$D3,$F2,$98,$CF,$C1
	.dc.b	$7B,$EF,$EC,$F1,$FD,$A7,$EB,$FE,$D3,$FE,$44,$FC,$03,$D1,$F7,$8F
	.dc.b	$CB,$3F,$21,$F3,$1A,$F9,$4F,$F4,$67,$27,$FA,$FE,$7F,$29,$D3,$9C
	.dc.b	$3D,$38,$FF,$93,$BF,$1C,$B8,$70,$CB,$A7,$2F,$CE,$AF,$06,$5C,$FF
	.dc.b	$21,$C5,$EF,$3F,$D0,$78,$9F,$8C,$FC,$9D,$FC,$00,$FD,$B5,$F8,$A3
	.dc.b	$F6,$67,$EB,$6F,$C9,$5F,$B9,$F1,$C3,$85,$EE,$B5,$F9,$86,$7F,$09
	.dc.b	$FA,$39,$F6,$CE,$E1,$C3,$3E,$BF,$4B,$F8,$6A,$D6,$5C,$FF,$19,$C5
	.dc.b	$EF,$3F,$17,$EF,$7F,$C6,$FD,$FF,$E0,$DE,$7B,$F5,$47,$EC,$CF,$DA
	.dc.b	$9F,$A7,$FF,$39,$FC,$FF,$F4,$3E,$AB,$1E,$38,$70,$BD,$FC,$6A,$DE
	.dc.b	$87,$7F,$3B,$F4,$57,$D9,$FE,$2C,$7E,$FF,$0E,$7C,$98,$8C,$1C,$34
	.dc.b	$F2,$69,$FD,$2F,$E6,$33,$3F,$F0,$07,$0F,$14,$FF,$4F,$FA,$4C,$EF
	.dc.b	$ED,$C3,$8E,$BE,$5F,$D1,$FC,$98,$F1,$9C,$ED,$FC,$13,$83,$DD,$3F
	.dc.b	$CE,$BF,$77,$7E,$C7,$FE,$5F,$FC,$E5,$FB,$43,$F7,$07,$E8,$7F,$C9
	.dc.b	$71,$E3,$8B,$1E,$3F,$4E,$C7,$8E,$1C,$3F,$1B,$F1,$D6,$AF,$E1,$61
	.dc.b	$CA,$3C,$3F,$5D,$F1,$67,$7A,$78,$F4,$F1,$7C,$0F,$C4,$BE,$F9,$F6
	.dc.b	$FC,$FB,$17,$BD,$3C,$FE,$E7,$F5,$DF,$A2,$FE,$91,$8F,$FB,$7B,$F3
	.dc.b	$17,$E6,$DF,$E6,$2F,$37,$FC,$D9,$F2,$7B,$DF,$FA,$ED,$EE,$B5,$F8
	.dc.b	$0B,$F1,$BD,$57,$FB,$FB,$E2,$7F,$74,$FB,$33,$EF,$0C,$FD,$9C,$89
	.dc.b	$CC,$78,$BE,$17,$F8,$17,$83,$FD,$16,$73,$9F,$97,$F2,$1C,$65,$FD
	.dc.b	$C3,$F5,$FF,$F1,$3F,$D6,$3F,$6E,$7D,$17,$F6,$1F,$C5,$7F,$A0,$6D
	.dc.b	$DF,$0E,$1F,$62,$C3,$85,$EF,$E4,$7D,$79,$7F,$8C,$BF,$40,$F3,$ED
	.dc.b	$FD,$BE,$3B,$E3,$ED,$39,$E1,$CB,$E6,$BF,$99,$C7,$2C,$BB,$6B,$E4
	.dc.b	$97,$D3,$FA,$AB,$FC,$7B,$AB,$F1,$27,$EF,$9F,$E9,$6F,$AB,$7D,$53
	.dc.b	$F5,$BF,$9F,$F6,$7F,$E1,$7B,$DF,$D1,$2F,$7F,$B6,$6B,$2E,$99,$F9
	.dc.b	$CE,$FC,$3E,$C7,$FC,$33,$8F,$3D,$3C,$1F,$0F,$E5,$8C,$65,$D7,$F0
	.dc.b	$7C,$63,$E7,$BF,$36,$DF,$C9,$21,$FA,$0F,$F8,$57,$E4,$3F,$87,$53
	.dc.b	$F6,$A7,$EB,$8F,$8E,$7C,$B3,$CA,$FF,$19,$63,$C7,$E0,$3E,$1E,$1C
	.dc.b	$2F,$7F,$21,$AE,$2F,$F1,$8E,$9D,$9F,$D5,$E1,$C7,$E4,$FF,$1E,$F3
	.dc.b	$D7,$C5,$AF,$87,$5F,$F0,$94,$E7,$3E,$DF,$4F,$FE,$15,$FA,$8F,$F1
	.dc.b	$77,$C1,$BF,$58,$7E,$87,$FD,$5F,$F7,$47,$E0,$2F,$DA,$3F,$23,$CD
	.dc.b	$FE,$34,$E1,$FF,$22,$E3,$C7,$0E,$17,$BF,$B6,$FF,$28,$67,$DB,$3E
	.dc.b	$B8,$7F,$04,$8C,$7E,$E7,$FA,$87,$58,$72,$FC,$2F,$CD,$5D,$AF,$CB
	.dc.b	$EC,$36,$F2,$FD,$17,$F8,$BF,$9F,$FC,$91,$F7,$1F,$8E,$F8,$AF,$C7
	.dc.b	$F1,$BF,$12,$7E,$D4,$FA,$6F,$F3,$26,$1C,$3A,$D8,$70,$BD,$FE,$67
	.dc.b	$AD,$FF,$CC,$79,$F5,$7F,$8E,$E1,$C7,$ED,$7F,$80,$BC,$F0,$E5,$E3
	.dc.b	$F9,$6B,$8E,$5D,$86,$BE,$5F,$17,$FC,$91,$F4,$BF,$E6,$1F,$D8,$FF
	.dc.b	$CE,$3F,$9D,$3F,$0E,$FF,$05,$7E,$A3,$FE,$23,$FC,$7F,$FC,$F9,$7B
	.dc.b	$CD,$7B,$E7,$5F,$F8,$A6,$76,$7F,$9F,$31,$EA,$FF,$36,$BE,$5F,$9E
	.dc.b	$7F,$1A,$7A,$61,$CE,$FF,$37,$5C,$85,$F9,$B2,$F2,$E9,$FE,$61,$FA
	.dc.b	$7F,$57,$E4,$9F,$DC,$FE,$0B,$F8,$0F,$DE,$3F,$61,$7E,$FD,$FE,$95
	.dc.b	$BD,$FA,$1F,$EC,$1E,$4F,$F4,$05,$7F,$7B,$F2,$3F,$CB,$DB,$7A,$73
	.dc.b	$7E,$0F,$FC,$C3,$1E,$33,$FD,$41,$E7,$A7,$83,$F1,$CF,$24,$F1,$CF
	.dc.b	$B6,$AF,$C7,$7F,$CF,$1F,$09,$F4,$5F,$D2,$76,$FC,$BF,$B5,$7C,$57
	.dc.b	$FA,$AE,$B5,$E3,$AB,$5A,$67,$3F,$9D,$8F,$F6,$77,$FB,$1D,$F0,$69
	.dc.b	$AB,$E5,$CB,$FE,$47,$FB,$0F,$1C,$E7,$1F,$B9,$E4,$FF,$A3,$7E,$93
	.dc.b	$F0,$DE,$07,$F2,$CE,$C7,$2F,$F2,$EF,$A7,$FE,$ED,$CF,$FE,$1A,$CB
	.dc.b	$F0,$17,$BA,$D7,$CF,$FF,$B1,$71,$E7,$EC,$7F,$5A,$7A,$F5,$DC,$38
	.dc.b	$BF,$8E,$BF,$D6,$C7,$13,$39,$F8,$A7,$F8,$4C,$7F,$B4,$7E,$83,$F3
	.dc.b	$9F,$29,$F9,$CF,$92,$FC,$B7,$8B,$9A,$FC,$A4,$E7,$8F,$FC,$D1,$F2
	.dc.b	$59,$9F,$C7,$CB,$C7,$FE,$DF,$8F,$49,$A7,$23,$F4,$78,$FE,$A7,$97
	.dc.b	$FD,$69,$3B,$D1,$3F,$AD,$E4,$FF,$5C,$7D,$EF,$EC,$7C,$BF,$F7,$0F
	.dc.b	$7F,$EB,$32,$F9,$9C,$DF,$EB,$07,$F5,$53,$AC,$7F,$DC,$FC,$2F,$C7
	.dc.b	$BF,$DE,$18,$F3,$BF,$F0,$2B,$D7,$6B,$1C,$8F,$F2,$C5,$FF,$AC,$E7
	.dc.b	$86,$5E,$1F,$C1,$FF,$D9,$9F,$95,$FF,$05,$F6,$FF,$DE,$BD,$BF,$C0
	.dc.b	$3D,$7F,$E0,$1B,$AB,$59,$FF,$3D,$CF,$F9,$EF,$33,$91,$7B,$AF,$E0
	.dc.b	$99,$7F,$8D,$5C,$E1,$DF,$6F,$F0,$EB,$D3,$3F,$07,$C6,$79,$2B,$89
	.dc.b	$DF,$47,$DF,$FF,$B8,$F9,$FF,$8E,$7A,$7F,$1B,$7F,$E5,$F7,$FE,$04
	.dc.b	$D5,$5A,$8F,$CE,$7A,$65,$DD,$E8,$9D,$47,$F9,$39,$EB,$B7,$F8,$C9
	.dc.b	$E5,$CD,$FC,$92,$F9,$57,$13,$F9,$16,$5E,$38,$71,$FF,$BE,$7E,$93
	.dc.b	$F9,$97,$A1,$C3,$5F,$E5,$BC,$BF,$CC,$F6,$FF,$33,$E5,$FC,$AE,$B5
	.dc.b	$5F,$CD,$FF,$0D,$F9,$B8,$72,$AF,$F2,$2B,$D3,$6B,$1C,$4F,$E7,$F5
	.dc.b	$FE,$2A,$78,$99,$3F,$9E,$BF,$31,$FB,$3F,$A8,$79,$BC,$1F,$C9,$79
	.dc.b	$3F,$1F,$34,$E6,$BF,$1B,$3F,$A9,$9C,$C7,$A6,$BF,$CB,$35,$F0,$3F
	.dc.b	$72,$B3,$93,$F2,$BB,$F5,$FB,$7F,$8B,$E5,$FE,$4B,$3F,$E3,$39,$7F
	.dc.b	$B1,$AF,$E2,$4E,$7C,$A6,$FF,$90,$C3,$8D,$7F,$A1,$9F,$9F,$F2,$3F
	.dc.b	$E6,$A7,$3D,$7F,$66,$78,$1D,$0F,$CE,$7D,$A7,$BF,$ED,$FE,$BF,$99
	.dc.b	$F8,$DD,$7F,$91,$EA,$9C,$DF,$F2,$79,$6F,$FA,$29,$CC,$FF,$AF,$2B
	.dc.b	$83,$F7,$7F,$1E,$FF,$3E,$CF,$2E,$AF,$25,$6A,$BF,$B1,$F3,$B2,$FC
	.dc.b	$F6,$7E,$59,$FC,$D7,$EA,$8F,$AF,$7E,$7F,$FD,$DB,$D2,$FF,$78,$7E
	.dc.b	$72,$FA,$5F,$26,$1C,$2F,$7E,$F2,$F7,$EB,$6A,$99,$75,$CB,$CF,$1C
	.dc.b	$7F,$08,$AE,$B1,$E7,$87,$FB,$FF,$C6,$F3,$4E,$6F,$C9,$F6,$BF,$93
	.dc.b	$E1,$7D,$B7,$E9,$3C,$7F,$CF,$FA,$7F,$C2,$A3,$18,$7E,$6B,$0C,$3F
	.dc.b	$A5,$AF,$F5,$2C,$FE,$B3,$CC,$B3,$93,$FC,$3F,$E0,$FE,$66,$79,$C7
	.dc.b	$F8,$73,$93,$F3,$9F,$95,$FE,$D3,$D3,$FD,$DF,$A3,$FA,$4D,$53,$9B
	.dc.b	$FE,$69,$F0,$9F,$C7,$A3,$11,$F7,$5F,$DE,$FC,$67,$FA,$E1,$CD,$FF
	.dc.b	$14,$BE,$75,$FB,$DC,$3B,$E3,$FA,$DF,$D3,$FF,$82,$FA,$7F,$C3,$3C
	.dc.b	$3F,$F1,$8E,$9F,$CB,$27,$38,$7E,$97,$C3,$C3,$FC,$1E,$FF,$D8,$F3
	.dc.b	$FC,$3B,$F0,$DB,$39,$3F,$9E,$F0,$77,$70,$7F,$9D,$F2,$FD,$F7,$D7
	.dc.b	$FF,$16,$F4,$BF,$8E,$F8,$7E,$2D,$D3,$9B,$FD,$24,$61,$FE,$31,$5F
	.dc.b	$ED,$89,$D4,$7F,$97,$67,$4F,$E7,$27,$AD,$FF,$0D,$FC,$7B,$F9,$2F
	.dc.b	$89,$FC,$B2,$BC,$11,$87,$F9,$1D,$A3,$11,$FC,$EF,$2E,$4F,$F2,$9A
	.dc.b	$E5,$7F,$C6,$7E,$D3,$AF,$9F,$F7,$B8,$7F,$AB,$C6,$2F,$FB,$86,$48
	.dc.b	$C4,$7F,$84,$23,$0F,$5A,$FE,$83,$F8,$1F,$F2,$5F,$B7,$F4,$FA,$1F
	.dc.b	$67,$1F,$EC,$71,$8B,$FE,$39,$82,$31,$1F,$D1,$8D,$9E,$9A,$FE,$9B
	.dc.b	$F9,$FF,$F3,$1F,$A9,$F8,$BE,$5F,$D4,$C7,$FB,$7C,$62,$FF,$B1,$62
	.dc.b	$8C,$47,$F5,$68,$C3,$F1,$6B,$FA,$CF,$B1,$FC,$F7,$F4,$1F,$A3,$E0
	.dc.b	$FD,$1C,$7F,$BD,$C6,$2F,$FA,$C6,$08,$C4,$75,$6D,$FF,$47,$F1,$FE
	.dc.b	$DF,$6F,$DB,$F8,$BF,$F0,$2C,$62,$FF,$5F,$87,$F1,$97,$EC,$57,$E0
	.dc.b	$5D,$98,$3F,$BC,$5F,$F5,$3F,$43,$F0,$FD,$BF,$87,$FB,$6F,$F0,$8C
	.dc.b	$62,$BF,$55,$82,$31,$1F,$0D,$B7,$FD,$77,$EA,$7F,$23,$D1,$FE,$60
	.dc.b	$AF,$91,$18,$E6,$FE,$FF,$ED,$7F,$87,$A3,$13,$F6,$71,$7E,$01,$CB
	.dc.b	$9B,$FA,$03,$67,$FC,$21,$7F,$E2,$68,$C6,$5F,$DB,$7F,$74,$F0,$F2
	.dc.b	$FF,$47,$6D,$F0,$C5,$18,$AF,$EE,$9F,$31,$E8,$E6,$FE,$9C,$DB,$E0
	.dc.b	$D9,$FE,$F5,$F9,$E7,$BB,$8F,$FA,$A3,$6B,$B3,$67,$FC,$01,$F4,$FF
	.dc.b	$13,$93,$FA,$D3,$6B,$C8,$D9,$FF,$03,$7C,$BF,$CA,$F0,$BF,$B0,$36
	.dc.b	$BD,$4D,$9F,$F0,$67,$E8,$3F,$43,$97,$FB,$33,$6B,$E0,$36,$7F,$C2
	.dc.b	$5F,$B9,$EF,$9B,$F3,$8D,$AF,$F0,$93,$99,$FF,$0C,$7E,$F3,$D7,$9B
	.dc.b	$F4,$4D,$AF,$F0,$C3,$99,$FF,$0F,$7E,$B3,$E3,$E5,$FD,$43,$6B,$FC
	.dc.b	$3C,$E6,$7F,$C4,$9F,$80,$7A,$79,$3F,$5C,$DA,$FF,$12,$39,$9F,$F1
	.dc.b	$57,$E2,$BE,$FE,$4F,$DA,$36,$BF,$C5,$4E,$67,$FC,$61,$F5,$BF,$17
	.dc.b	$93,$F7,$4D,$AF,$F1,$83,$99,$FF,$1B,$7D,$67,$CB,$E4,$FE,$00,$DA
	.dc.b	$FF,$1B,$39,$9F,$F1,$E7,$E1,$5F,$47,$93,$F8,$33,$6B,$FC,$78,$E6
	.dc.b	$7F,$C8,$5F,$85,$6F,$92,$D7,$F2,$FB,$FF,$0B,$C7,$F9,$0E,$7A,$7E
	.dc.b	$6B,$FE,$49,$FD,$23,$B7,$2D,$C3,$86,$1F,$F0,$02,$3F,$C9,$53,$E3
	.dc.b	$F9,$EF,$F2,$87,$E2,$BE,$5E,$6B,$87,$0C,$3F,$C4,$9E,$0F,$FA,$42
	.dc.b	$BD,$5E,$8F,$F9,$5F,$F4,$1F,$F3,$AF,$EF,$2F,$C6,$DF,$26,$FD,$BD
	.dc.b	$F4,$EF,$D2,$5F,$92,$7F,$07,$F2,$59,$72,$C7,$8E,$1C,$2F,$7E,$77
	.dc.b	$FA,$82,$B5,$97,$5C,$BA,$7E,$4B,$FE,$69,$FC,$6B,$E7,$77,$DE,$3F
	.dc.b	$53,$FE,$69,$E5,$FA,$5D,$DF,$E6,$F6,$F2,$7F,$AB,$3E,$AB,$EB,$73
	.dc.b	$7F,$9C,$DB,$1F,$EB,$0F,$D0,$33,$DB,$72,$F8,$D7,$FE,$54,$E4,$7E
	.dc.b	$83,$CB,$B3,$FD,$6E,$F4,$F9,$7F,$F5,$CF,$E1,$5D,$BB,$6E,$3C,$BD
	.dc.b	$4F,$F5,$EF,$E6,$FE,$2F,$2A,$CF,$6F,$8B,$FF,$60,$FE,$03,$E6,$F2
	.dc.b	$EC,$F8,$BE,$2D,$FA,$B3,$1F,$F5,$4F,$17,$FB,$17,$F1,$CC,$C6,$2F
	.dc.b	$FA,$10,$FB,$4E,$3B,$E6,$3F,$A3,$79,$78,$7F,$D7,$98,$7F,$B3,$FF
	.dc.b	$64,$EF,$C7,$B9,$FF,$06,$7C,$16,$1D,$9D,$BF,$69,$F2,$F4,$EE,$3F
	.dc.b	$D7,$BE,$0E,$1D,$9F,$D5,$7E,$CF,$D7,$DB,$B8,$FF,$66,$7E,$4F,$87
	.dc.b	$67,$F8,$6F,$F3,$5F,$83,$E0,$6E,$1B,$61,$D9,$FE,$CC,$FA,$1F,$8F
	.dc.b	$E2,$75,$F0,$BF,$DB,$1C,$E8,$C4,$7C,$CF,$CA,$3C,$5F,$1F,$A6,$C7
	.dc.b	$D0,$F2,$7F,$2B,$E6,$7E,$67,$F5,$27,$EC,$B9,$D3,$FD,$C9,$E8,$FE
	.dc.b	$9F,$CD,$FC,$D7,$9E,$B7,$EA,$7F,$6B,$D5,$F8,$1D,$7E,$46,$EB,$9C
	.dc.b	$BD,$63,$F0,$7F,$4F,$B1,$F1,$FA,$EC,$E9,$57,$E1,$A7,$E6,$7E,$09
	.dc.b	$F0,$3D,$AF,$9D,$D7,$70,$EA,$F0,$E7,$4F,$D3,$FC,$FB,$E3,$FB,$9F
	.dc.b	$57,$AE,$E1,$E4,$78,$47,$D9,$F7,$3E,$6F,$BB,$F8,$1D,$7F,$DF,$B9
	.dc.b	$23,$18,$77,$75,$E3,$7D,$4F,$7B,$F1,$1D,$BC,$1B,$5E,$57,$33,$F6
	.dc.b	$BD,$FF,$C7,$79,$9D,$9B,$5E,$B7,$33,$F8,$0F,$80,$FC,$A3,$E4,$6C
	.dc.b	$FC,$17,$31,$F8,$6F,$81,$FC,$D7,$47,$A9,$B5,$F1,$DC,$CF,$E3,$3E
	.dc.b	$0B,$F4,$1C,$FF,$01,$B5,$F3,$5C,$CF,$E4,$BE,$0F,$F4,$DC,$DF,$19
	.dc.b	$B5,$F4,$DC,$CF,$E6,$3E,$13,$F5,$9C,$BF,$31,$B5,$F6,$5C,$CF,$E7
	.dc.b	$BE,$17,$F7,$1C,$9F,$E8,$3E,$44,$62,$FF,$C5,$71,$87,$F5,$F3,$FE
	.dc.b	$46,$7F,$87,$FC,$37,$F0,$1E,$4E,$4D,$8E,$2D,$C3,$F5,$BF,$0F,$FC
	.dc.b	$1F,$93,$C0,$DA,$EE,$D9,$FD,$9F,$C4,$7F,$0D,$7C,$CD,$9F,$2B,$63
	.dc.b	$F7,$3F,$13,$FC,$4F,$93,$D8,$DA,$F5,$B6,$7F,$7F,$F1,$5F,$C6,$79
	.dc.b	$3E,$13,$6B,$E0,$B6,$7F,$82,$FE,$1B,$FC,$7F,$93,$E4,$36,$BE,$3B
	.dc.b	$67,$F8,$5F,$C3,$FF,$2B,$DF,$FD,$61,$EC,$7F,$36,$8C,$58,$9F,$81
	.dc.b	$65,$FE,$47,$5F,$E7,$7E,$3F,$F3,$D4,$7F,$9A,$44,$7F,$A3,$9F,$E8
	.dc.b	$3F,$1F,$FC,$E7,$8B,$FD,$7D,$F9,$0E,$27,$F9,$6C,$7F,$B0,$7F,$61
	.dc.b	$C4,$73,$8E,$D5,$C6,$3E,$F6,$1F,$E9,$C7,$C1,$A7,$FB,$27,$CC,$FE
	.dc.b	$9F,$F2,$7F,$D1,$E3,$FD,$99,$F5,$D8,$AF,$E7,$71,$E9,$9D,$47,$4F
	.dc.b	$95,$FE,$9D,$CD,$FE,$80,$C3,$FA,$CC,$62,$7F,$CF,$78,$61,$FE,$8B
	.dc.b	$5F,$0E,$75,$1F,$EB,$DC,$7F,$D9,$2F,$C5,$E5,$F9,$8F,$F5,$DC,$3F
	.dc.b	$B2,$FC,$BF,$F5,$D9,$CF,$1B,$2C,$DF,$F5,$29,$F9,$98,$74,$EE,$FE
	.dc.b	$AF,$3F,$A7,$F9,$AF,$EC,$DC,$FD,$B8,$FF,$C4,$37,$FE,$E9,$AF,$EB
	.dc.b	$73,$9C,$7F,$BA,$DF,$09,$FC,$07,$9B,$FD,$42,$31,$0D,$FF,$A9,$FC
	.dc.b	$9F,$F7,$3C,$FD,$7B,$F6,$8C,$57,$E2,$63,$D5,$9A,$73,$2F,$B5,$E6
	.dc.b	$FF,$6E,$BF,$8B,$F4,$DF,$D8,$3E,$7B,$FC,$09,$E1,$7D,$3C,$BF,$C1
	.dc.b	$71,$8B,$FF,$56,$C7,$68,$FF,$05,$75,$7F,$82,$B3,$46,$EB,$FA,$7E
	.dc.b	$5D,$5F,$F0,$1E,$1F,$EB,$EC,$7F,$B7,$FD,$17,$29,$FD,$17,$27,$F9
	.dc.b	$2B,$3F,$F0,$DC,$E6,$7E,$D7,$0E,$1F,$F0,$6E,$9E,$19,$FF,$0E,$7D
	.dc.b	$A5,$9C,$9D,$33,$EE,$E0,$FF,$10,$65,$FE,$00,$FA,$4F,$F1,$27,$67
	.dc.b	$DB,$CF,$E5,$C6,$2B,$F6,$DC,$DE,$5E,$04,$E6,$7F,$95,$3F,$E2,$DD
	.dc.b	$BE,$23,$9A,$FA,$53,$FE,$36,$8C,$63,$FE,$10,$FA,$FF,$F1,$A7,$67
	.dc.b	$DE,$CB,$EC,$64,$9C,$E1,$F7,$E5,$FA,$D7,$F9,$02,$73,$3F,$CE,$DF
	.dc.b	$F1,$F7,$C4,$FD,$57,$35,$E2,$BE,$95,$FE,$36,$C3,$F1,$B8,$F8,$E3
	.dc.b	$F8,$A7,$D6,$3C,$37,$E3,$EF,$85,$7E,$E4,$E0,$7C,$23,$E2,$FC,$99
	.dc.b	$72,$C7,$8E,$1C,$2F,$7F,$98,$6A,$99,$F7,$FD,$13,$D1,$1C,$74,$AE
	.dc.b	$B1,$E7,$A7,$87,$2F,$44,$E6,$BC,$D3,$E5,$F6,$FF,$C7,$BF,$6B,$FE
	.dc.b	$61,$F0,$BF,$37,$C6,$FD,$07,$FC,$D3,$39,$C7,$FC,$4F,$3F,$E9,$CF
	.dc.b	$D0,$33,$1F,$E5,$CB,$FD,$05,$7F,$9A,$F8,$6C,$E4,$E3,$D9,$E0,$71
	.dc.b	$E0,$FF,$A9,$B9,$7F,$C9,$FF,$73,$FE,$73,$F2,$7F,$43,$A7,$F1,$3B
	.dc.b	$FF,$3D,$CE,$5F,$D0,$DE,$1F,$F3,$6D,$7E,$C6,$7F,$CF,$9F,$21,$67
	.dc.b	$27,$F2,$1C,$7B,$B8,$EF,$F1,$3F,$E8,$8F,$BB,$FF,$42,$79,$7F,$9E
	.dc.b	$CD,$F9,$FB,$EA,$B5,$7F,$E7,$CA,$D4,$BE,$BF,$33,$FE,$70,$13,$FC
	.dc.b	$17,$0F,$F4,$87,$C9,$D9,$E4,$7E,$27,$1F,$CA,$4F,$8A,$31,$F9,$87
	.dc.b	$F9,$D3,$E7,$BF,$D3,$1E,$77,$E8,$4C,$4F,$FA,$39,$C7,$17,$FA,$8A
	.dc.b	$77,$3F,$DA,$5F,$F5,$07,$CA,$FF,$17,$73,$8F,$FB,$63,$E2,$3F,$D0
	.dc.b	$3F,$3D,$FE,$A8,$C3,$F8,$FF,$A1,$ED,$F8,$FF,$F5,$6C,$E7,$2F,$C2
	.dc.b	$CF,$C1,$E2,$CC,$7F,$A9,$AF,$FD,$5A,$E0,$FF,$49,$BF,$E7,$27,$FB
	.dc.b	$E5,$73,$8F,$F7,$4D,$3E,$BF,$C7,$5C,$CE,$FF,$5F,$FE,$96,$FC,$03
	.dc.b	$FD,$7B,$EA,$7E,$23,$87,$FC,$15,$DB,$FE,$C2,$9C,$DF,$F6,$19,$FF
	.dc.b	$07,$79,$39,$8F,$F5,$DD,$F9,$BC,$D5,$F4,$D3,$F9,$E3,$CB,$AB,$FD
	.dc.b	$FB,$8F,$F9,$E2,$31,$C4,$8F,$EE,$5F,$81,$FF,$B4,$3D,$3F,$C5,$75
	.dc.b	$FE,$27,$CB,$FF,$6A,$CE,$75,$FC,$04,$FE,$37,$C9,$CC,$7F,$B3,$AF
	.dc.b	$C9,$5C,$F4,$FE,$9C,$F1,$F3,$59,$FE,$47,$8F,$FD,$73,$39,$8F,$CA
	.dc.b	$71,$FF,$AF,$3F,$06,$FF,$71,$FA,$1F,$D9,$FC,$2F,$8F,$CD,$FE,$E8
	.dc.b	$9C,$EB,$D2,$7E,$5F,$36,$63,$FD,$C3,$7F,$1E,$B9,$EB,$FD,$85,$E3
	.dc.b	$97,$F3,$EF,$23,$C1,$7C,$0E,$87,$FA,$DC,$FF,$7B,$FD,$CF,$FB,$DA
	.dc.b	$3F,$87,$F9,$5F,$0F,$E7,$73,$5F,$EF,$09,$CF,$6F,$F2,$AE,$A6,$1C
	.dc.b	$73,$FE,$E4,$F3,$EB,$B8,$71,$9F,$E3,$56,$FC,$AF,$FD,$BF,$7C,$74
	.dc.b	$69,$FB,$FF,$A9,$FF,$71,$FE,$17,$FE,$2A,$F1,$5E,$63,$C0,$FE,$2D
	.dc.b	$CC,$E5,$FE,$87,$C2,$C3,$85,$EF,$C1,$F9,$7F,$8B,$2B,$5C,$5F,$E6
	.dc.b	$1C,$FA,$E7,$FE,$0B,$C7,$8F,$C3,$5A,$D6,$5C,$DF,$E9,$DC,$BE,$67
	.dc.b	$37,$E4,$F8,$1F,$F7,$DF,$E1,$DE,$9F,$37,$FA,$B7,$9D,$F8,$FF,$80
	.dc.b	$7E,$75,$E3,$AD,$78,$7F,$C8,$3A,$BF,$AF,$D7,$FB,$D7,$85,$5A,$D3
	.dc.b	$FC,$E8,$F5,$F8,$8F,$F1,$13,$CB,$2F,$CE,$7C,$1F,$8A,$BD,$DA,$78
	.dc.b	$B2,$FE,$D7,$E7,$FF,$A0,$72,$7D,$4F,$CF,$E5,$FC,$E9,$FD,$AF,$47
	.dc.b	$D4,$3E,$FB,$E5,$56,$FB,$97,$BA,$7E,$19,$AD,$7E,$74,$FF,$4D,$7F
	.dc.b	$D8,$17,$C5,$FF,$47,$3C,$E3,$C1,$DD,$E5,$8C,$77,$33,$FF,$01,$5F
	.dc.b	$DB,$FA,$BF,$F3,$7F,$AA,$F1,$3F,$CD,$9C,$2F,$41,$FA,$63,$F5,$47
	.dc.b	$EA,$7C,$72,$FF,$51,$63,$C7,$C4,$FE,$11,$7B,$F3,$75,$4A,$ED,$7F
	.dc.b	$E5,$1D,$3C,$3A,$7B,$6B,$FD,$03,$CC,$9C,$E5,$D3,$EF,$7E,$84,$63
	.dc.b	$6F,$36,$8F,$A3,$FA,$9F,$8D,$7F,$A2,$BF,$2F,$FD,$07,$B1,$FE,$1D
	.dc.b	$FC,$FB,$FC,$69,$F9,$43,$ED,$7F,$39,$7B,$F2,$3F,$92,$74,$2F,$7F
	.dc.b	$6D,$AE,$2F,$F6,$06,$7D,$1F,$F5,$F6,$1C,$7F,$62,$FF,$36,$3C,$F4
	.dc.b	$F0,$7C,$7B,$5F,$2D,$F2,$D3,$BE,$7F,$E4,$7F,$C3,$BF,$19,$F9,$37
	.dc.b	$FA,$8B,$F4,$7F,$F1,$B7,$1B,$F3,$17,$E5,$2F,$C5,$F1,$F9,$1F,$ED
	.dc.b	$70,$E1,$E8,$7F,$80,$6F,$75,$C7,$ED,$F5,$B7,$FB,$67,$3F,$F2,$17
	.dc.b	$C4,$FF,$A0,$F1,$E3,$A5,$AD,$65,$CF,$4F,$07,$8F,$E6,$9C,$D7,$96
	.dc.b	$7C,$9E,$EE,$65,$FA,$D3,$99,$83,$ED,$FF,$C2,$DF,$A1,$3E,$F5,$F1
	.dc.b	$FE,$6C,$78,$E1,$C2,$F7,$4F,$92,$D6,$5D,$72,$F4,$61,$C3,$F0,$1B
	.dc.b	$39,$C3,$BF,$CC,$F9,$63,$1B,$B5,$F2,$E1,$D3,$4F,$04,$E7,$47,$AD
	.dc.b	$E5,$FA,$F7,$E9,$0F,$AF,$74,$32,$7E,$A0,$FB,$B7,$EA,$3C,$70,$E1
	.dc.b	$7B,$AD,$7C,$FE,$7E,$CB,$D7,$9F,$6C,$EE,$1C,$33,$EB,$F4,$1F,$3E
	.dc.b	$B5,$97,$3E,$3F,$E2,$DF,$84,$7F,$BB,$3E,$73,$FD,$99,$E6,$BF,$61
	.dc.b	$7E,$D4,$FD,$19,$FB,$E3,$F2,$4F,$E8,$18,$F1,$F7,$7F,$D0,$D8,$70
	.dc.b	$BD,$9A,$DF,$E0,$67,$D9,$FF,$7D,$9E,$23,$35,$FD,$45,$C9,$C4,$47
	.dc.b	$9A,$BB,$7C,$3F,$19,$CD,$78,$B3,$F4,$7E,$1D,$9F,$DF,$1F,$93,$BF
	.dc.b	$51,$7E,$08,$F7,$9F,$93,$BF,$3C,$FF,$70,$FC,$13,$8B,$EC,$6C,$CF
	.dc.b	$9E,$5C,$B1,$E3,$83,$ED,$7D,$15,$AB,$F5,$E3,$CD,$FC,$14,$79,$3F
	.dc.b	$5D,$FF,$78,$65,$DA,$3A,$E9,$E5,$D1,$EB,$FA,$F6,$F5,$4E,$7E,$97
	.dc.b	$1B,$F8,$EF,$D5,$F6,$BF,$DD,$7F,$1A,$F4,$FE,$7F,$FE,$A5,$FA,$47
	.dc.b	$E9,$0F,$0B,$E1,$7E,$35,$FE,$D5,$E3,$61,$C3,$ED,$18,$70,$BD,$FD
	.dc.b	$A7,$5E,$5F,$CE,$67,$FE,$B1,$F0,$FE,$CC,$64,$47,$85,$F8,$78,$F5
	.dc.b	$C3,$A7,$4F,$8E,$78,$57,$9F,$5F,$34,$FF,$56,$D3,$CB,$5E,$0D,$3C
	.dc.b	$B5,$AD,$7F,$81,$56,$AB,$C3,$5A,$D7,$DF,$EB,$F9,$7E,$DD,$FA,$CF
	.dc.b	$E1,$BF,$47,$FF,$05,$F6,$BF,$62,$FE,$29,$9F,$3B,$DF,$4A,$F7,$5F
	.dc.b	$EE,$9A,$D6,$1F,$90,$D3,$DE,$7F,$01,$C3,$BF,$EA,$DF,$1A,$B7,$8F
	.dc.b	$3F,$A3,$FC,$97,$C5,$FF,$82,$F9,$7F,$23,$87,$83,$4F,$2F,$E6,$39
	.dc.b	$CF,$A6,$9E,$9C,$7F,$35,$AF,$F7,$DD,$1F,$8E,$FF,$49,$FD,$E1,$F8
	.dc.b	$03,$F4,$37,$E0,$BF,$DF,$FF,$6D,$7A,$FF,$E3,$DC,$38,$5E,$EB,$5F
	.dc.b	$86,$E7,$6F,$E9,$19,$74,$7F,$3F,$7D,$FE,$A7,$DC,$F2,$CF,$BF,$CE
	.dc.b	$BE,$B7,$FC,$51,$39,$CB,$FC,$21,$F6,$F8,$C7,$F9,$47,$E4,$FF,$E5
	.dc.b	$5F,$97,$FE,$ED,$F3,$7F,$CA,$BE,$91,$F9,$B7,$8E,$D7,$BF,$9E,$5E
	.dc.b	$FF,$76,$D4,$FF,$97,$FF,$40,$F2,$DF,$6F,$B7,$FE,$E8,$FA,$8C,$67
	.dc.b	$E0,$F8,$7F,$E8,$7D,$3F,$32,$73,$9F,$69,$FF,$3A,$47,$6B,$FE,$FD
	.dc.b	$39,$D7,$FC,$01,$7C,$6B,$CF,$5A,$D3,$D9,$97,$AF,$BF,$FB,$2F,$EA
	.dc.b	$4F,$C9,$9F,$9F,$BF,$32,$FD,$D7,$EE,$DB,$F9,$75,$EE,$B5,$E6,$3F
	.dc.b	$44,$D7,$99,$F0,$37,$69,$D9,$F8,$06,$31,$E7,$A7,$83,$4F,$DB,$70
	.dc.b	$7A,$6F,$CB,$39,$FC,$63,$EC,$61,$C7,$5F,$16,$1F,$C3,$FC,$BF,$34
	.dc.b	$63,$57,$D3,$FD,$9F,$C2,$1F,$9A,$3F,$4A,$7A,$3F,$CD,$78,$1E,$CB
	.dc.b	$F7,$7E,$17,$BA,$D4,$BF,$0A,$CF,$0F,$BF,$2E,$B9,$5C,$38,$72,$7F
	.dc.b	$84,$B0,$F4,$D6,$B1,$F0,$6B,$8B,$DE,$9E,$2E,$BF,$0E,$5E,$6C,$B9
	.dc.b	$7C,$47,$FA,$97,$F2,$DF,$E9,$3F,$AF,$FF,$96,$7C,$D7,$E4,$CF,$98
	.dc.b	$7E,$4A,$FD,$91,$F9,$1F,$1E,$BF,$F0,$46,$3C,$70,$E1,$7B,$FC,$93
	.dc.b	$54,$AE,$CF,$F5,$0C,$FC,$3F,$6E,$FC,$8B,$F8,$84,$63,$5F,$1E,$BE
	.dc.b	$2F,$CE,$FF,$D3,$53,$9D,$3B,$FE,$01,$8C,$BC,$79,$31,$E3,$AF,$A7
	.dc.b	$F2,$5F,$1E,$BE,$7D,$3D,$BC,$0F,$C9,$7C,$FE,$7B,$F7,$C7,$86,$FC
	.dc.b	$BF,$FB,$9F,$CE,$BE,$23,$FD,$9D,$E5,$39,$3F,$B9,$E3,$C7,$0E,$1F
	.dc.b	$21,$FE,$17,$BF,$F0,$BD,$EF,$8F,$F8,$4E,$7F,$8E,$7E,$21,$8B,$F8
	.dc.b	$6D,$CF,$C3,$F0,$BE,$F8,$FE,$17,$5E,$69,$F2,$FE,$7D,$9C,$FC,$51
	.dc.b	$E8,$C9,$F3,$9F,$53,$C1,$7E,$97,$FE,$3F,$FD,$AF,$FD,$A3,$F9,$3F
	.dc.b	$F9,$8B,$E7,$35,$39,$FC,$39,$3F,$E6,$6F,$98,$CF,$D5,$F9,$B8,$3F
	.dc.b	$CC,$F6,$FC,$B7,$19,$F6,$A6,$7D,$BF,$1A,$F5,$4E,$63,$C3,$F6,$F8
	.dc.b	$C3,$C1,$9F,$F9,$D3,$E4,$7B,$D7,$9B,$ED,$3F,$19,$D1,$F2,$F5,$FC
	.dc.b	$07,$D8,$3E,$7F,$FC,$F7,$F2,$9F,$E9,$1F,$CF,$DF,$BF,$7E,$53,$C7
	.dc.b	$FF,$32,$D6,$BE,$FD,$4F,$DF,$35,$8B,$F4,$1F,$76,$1C,$33,$FA,$E7
	.dc.b	$83,$1E,$5F,$17,$F5,$67,$AE,$CD,$7F,$D4,$5F,$0B,$9C,$BC,$D3,$9D
	.dc.b	$7C,$3F,$B1,$78,$32,$F2,$6C,$FA,$BC,$5E,$E1,$FA,$AF,$EB,$BE,$E9
	.dc.b	$FB,$33,$F5,$47,$CA,$3F,$27,$FC,$B7,$C1,$3F,$28,$E1,$7B,$AD,$79
	.dc.b	$FF,$C7,$3F,$1D,$CE,$3F,$77,$2E,$BE,$65,$C3,$86,$7F,$A1,$F0,$7F
	.dc.b	$29,$5E,$2B,$F0,$CF,$E9,$7C,$3C,$DF,$0C,$BF,$94,$FE,$C7,$DB,$4F
	.dc.b	$0E,$CF,$D1,$FF,$2F,$D9,$FE,$5E,$F0,$9F,$B6,$FF,$9F,$FF,$2D,$7F
	.dc.b	$00,$7F,$BC,$7F,$30,$7E,$73,$C3,$0F,$F4,$E6,$1C,$3E,$F5,$87,$0F
	.dc.b	$80,$60,$FC,$E7,$86,$1F,$E3,$0C,$DF,$BB,$FF,$92,$3D,$4F,$F4,$2E
	.dc.b	$7C,$EB,$FD,$07,$18,$AF,$26,$AC,$BE,$8D,$74,$BF,$16,$7E,$1F,$A4
	.dc.b	$FF,$58,$63,$C7,$5F,$2E,$3E,$BF,$88,$F1,$FE,$93,$FD,$4B,$ED,$BF
	.dc.b	$D4,$5F,$68,$FC,$11,$FB,$DB,$F4,$8D,$FC,$DF,$F9,$93,$F0,$2F,$F3
	.dc.b	$5D,$6B,$F4,$DC,$E3,$EB,$E2,$FF,$44,$E7,$FD,$7B,$E6,$B1,$1F,$EC
	.dc.b	$38,$F0,$61,$CB,$E3,$FC,$F3,$C3,$2E,$DA,$F8,$FE,$47,$33,$C6,$73
	.dc.b	$A7,$9B,$F6,$EF,$0E,$3E,$4D,$DF,$A4,$E3,$2F,$64,$7E,$6B,$93,$FD
	.dc.b	$67,$AF,$8A,$73,$FB,$6F,$D6,$F9,$97,$E8,$9F,$EC,$1F,$D3,$BF,$8A
	.dc.b	$F2,$3F,$58,$7C,$06,$EB,$5F,$04,$AD,$7E,$17,$9E,$CF,$83,$CB,$FD
	.dc.b	$F6,$FF,$02,$FD,$E3,$FD,$5F,$9F,$82,$59,$F8,$3D,$0F,$83,$5C,$67
	.dc.b	$39,$77,$F3,$73,$87,$1C,$3C,$FA,$FE,$D7,$87,$FD,$A3,$39,$BF,$34
	.dc.b	$FF,$38,$9F,$F6,$AF,$5F,$F9,$93,$F3,$E7,$E5,$5F,$EC,$8F,$C0,$9F
	.dc.b	$AE,$7F,$07,$BD,$FD,$B2,$F7,$F8,$42,$F7,$ED,$7E,$83,$2F,$F3,$7F
	.dc.b	$E7,$7F,$09,$ED,$C5,$F0,$41,$FE,$B0,$C7,$A6,$9E,$1F,$CC,$FE,$0D
	.dc.b	$79,$63,$C9,$ED,$E2,$B7,$9F,$87,$67,$EF,$DE,$19,$CE,$BE,$AF,$D3
	.dc.b	$7F,$67,$FA,$33,$CA,$FF,$43,$FE,$C3,$FE,$E1,$FD,$99,$F9,$17,$F8
	.dc.b	$57,$97,$FD,$F3,$7B,$FC,$01,$7B,$A7,$E3,$7D,$31,$FF,$2F,$72,$7F
	.dc.b	$BA,$B1,$E3,$1F,$D2,$E3,$07,$FB,$BB,$1E,$98,$78,$79,$3E,$FC,$63
	.dc.b	$F1,$46,$9F,$E8,$9B,$ED,$A7,$8A,$73,$A3,$F3,$BC,$E5,$E5,$FA,$7B
	.dc.b	$5A,$F4,$2E,$5E,$EF,$2B,$F8,$AF,$CE,$3F,$21,$7E,$65,$FC,$6B,$DE
	.dc.b	$FF,$35,$7C,$EB,$EE,$F7,$5A,$9F,$E3,$FF,$7B,$FD,$D7,$D5,$CE,$1F
	.dc.b	$A0,$C7,$A7,$E2,$57,$0E,$18,$7F,$9D,$7E,$DF,$F3,$B5,$E1,$BF,$06
	.dc.b	$3F,$E6,$0E,$5E,$77,$DB,$4F,$F5,$BF,$4F,$E2,$27,$F1,$38,$71,$AD
	.dc.b	$7C,$CF,$F8,$53,$F0,$7D,$7E,$BD,$FA,$CF,$98,$7E,$B5,$F9,$EF,$CA
	.dc.b	$5E,$E3,$EE,$78,$69,$F9,$4C,$38,$5E,$EB,$FD,$1F,$AB,$2E,$BE,$93
	.dc.b	$3F,$E7,$91,$8F,$D5,$AE,$5C,$EB,$58,$F8,$7E,$5B,$F9,$BD,$F2,$C3
	.dc.b	$CB,$E6,$B6,$7D,$76,$B2,$E5,$5A,$9F,$4F,$EB,$7E,$89,$6F,$E8,$C3
	.dc.b	$96,$EF,$DD,$B5,$F4,$6F,$DC,$DF,$AD,$BF,$34,$7E,$A1,$F1,$BF,$45
	.dc.b	$7E,$B9,$C2,$F7,$5A,$E4,$57,$F8,$17,$9F,$E2,$6F,$FB,$CC,$FB,$7B
	.dc.b	$9F,$0A,$11,$E1,$D3,$0D,$C7,$A6,$FE,$FA,$DE,$7E,$0D,$7E,$CF,$D3
	.dc.b	$E7,$0E,$11,$E7,$E7,$F3,$65,$EB,$8F,$57,$EE,$38,$CF,$D1,$B3,$F3
	.dc.b	$AF,$91,$F5,$0F,$DE,$7E,$17,$EE,$EF,$C6,$5F,$7C,$FC,$FD,$FB,$4F
	.dc.b	$0B,$DD,$6A,$5E,$E6,$7F,$6F,$F8,$19,$75,$D6,$E1,$C3,$27,$A7,$EF
	.dc.b	$AD,$69,$E1,$97,$E5,$B9,$BE,$19,$BF,$6B,$F1,$64,$C3,$84,$BE,$43
	.dc.b	$F0,$9F,$96,$3E,$F5,$F9,$73,$F6,$67,$AD,$FE,$D0,$F0,$DF,$A6,$F0
	.dc.b	$BD,$D6,$BE,$AF,$F0,$7F,$8F,$E7,$1F,$B9,$97,$5F,$C9,$2E,$1C,$33
	.dc.b	$FF,$6D,$F7,$FD,$BA,$F1,$5F,$86,$7F,$DB,$FE,$76,$6F,$86,$5F,$90
	.dc.b	$FC,$13,$B6,$9E,$1D,$BF,$B2,$F9,$1F,$8E,$F1,$BF,$A8,$74,$7E,$6F
	.dc.b	$F3,$57,$E0,$EF,$CB,$BF,$33,$F8,$57,$E9,$7E,$3B,$ED,$F9,$6B,$DD
	.dc.b	$7E,$7B,$DD,$D6,$3F,$83,$CB,$FC,$67,$EB,$7F,$0F,$C7,$8C,$60,$FE
	.dc.b	$B2,$F4,$C3,$9F,$37,$E6,$63,$1A,$78,$F8,$F1,$8F,$9E,$3C,$DB,$BF
	.dc.b	$29,$F0,$ED,$E5,$9C,$FD,$36,$33,$F3,$DF,$F1,$AE,$1F,$85,$9F,$6E
	.dc.b	$0F,$DB,$FE,$F1,$FC,$0F,$F2,$2F,$F6,$57,$E1,$6F,$D8,$9F,$96,$BF
	.dc.b	$48,$9F,$03,$BF,$BC,$5E,$EB,$54,$F7,$B5,$F6,$9F,$BB,$CB,$AF,$47
	.dc.b	$C3,$0F,$83,$3E,$F7,$FD,$67,$7F,$B1,$39,$C7,$C5,$F7,$7F,$BF,$8C
	.dc.b	$57,$97,$93,$C9,$A7,$8E,$3D,$1F,$B1,$FF,$1F,$F8,$87,$E9,$6F,$DA
	.dc.b	$9F,$81,$BF,$7D,$7E,$84,$FD,$BB,$FC,$37,$C1,$F8,$9C,$78,$E1,$C2
	.dc.b	$F7,$5C,$EB,$58,$FF,$88,$B8,$BD,$D8,$70,$6C,$78,$72,$F0,$65,$FE
	.dc.b	$8E,$BE,$39,$75,$87,$D2,$7F,$1D,$C7,$C3,$7C,$32,$FF,$74,$E1,$C2
	.dc.b	$31,$E3,$7F,$BB,$BF,$54,$FE,$C1,$F6,$8F,$C5,$5F,$8B,$3F,$6C,$FA
	.dc.b	$DF,$AB,$3F,$4B,$FE,$81,$9B,$1E,$38,$70,$BD,$D7,$DD,$AD,$71,$7F
	.dc.b	$8D,$32,$ED,$97,$FA,$5F,$1E,$3F,$B1,$DA,$D6,$5C,$FB,$9A,$7F,$A5
	.dc.b	$B2,$E7,$4C,$B9,$E9,$E6,$8F,$F4,$BE,$5E,$4C,$FD,$58,$F2,$D3,$F7
	.dc.b	$78,$F2,$C7,$8C,$E7,$9B,$E2,$63,$E2,$C3,$8E,$7F,$07,$F1,$5C,$CF
	.dc.b	$87,$5F,$34,$FC,$5F,$80,$CF,$EC,$2F,$0B,$E3,$70,$BF,$6A,$7E,$E2
	.dc.b	$F4,$DF,$B0,$70,$BD,$D6,$A5,$FA,$C6,$7E,$DF,$E3,$6C,$CF,$B7,$36
	.dc.b	$1B,$A7,$7D,$3B,$57,$5A,$4E,$7D,$3F,$8F,$3F,$E6,$8B,$E3,$9F,$87
	.dc.b	$F4,$4E,$17,$CF,$47,$EB,$3F,$0B,$9B,$F4,$FF,$90,$BE,$AF,$FC,$05
	.dc.b	$F5,$2F,$CA,$5E,$B6,$FC,$75,$EF,$D7,$FF,$62,$D7,$C3,$EF,$FB,$B9
	.dc.b	$74,$E1,$FF,$6C,$F1,$FF,$8D,$63,$F6,$BF,$51,$F0,$9F,$0E,$BE,$0F
	.dc.b	$1F,$D7,$7F,$C3,$74,$EF,$3E,$4C,$73,$7C,$33,$F0,$FE,$91,$E5,$D7
	.dc.b	$C9,$BF,$A7,$EB,$BE,$FF,$DF,$3F,$45,$FF,$32,$7E,$CA,$FD,$BF,$F4
	.dc.b	$9F,$61,$F6,$B5,$FC,$AD,$EF,$EC,$D7,$BF,$2B,$FC,$C1,$E0,$F4,$CB
	.dc.b	$FD,$CB,$C7,$EE,$3F,$D3,$38,$71,$70,$1E,$97,$CF,$2F,$7C,$63,$3F
	.dc.b	$DB,$7A,$D8,$C3,$CF,$1E,$6E,$DF,$06,$9E,$48,$F4,$CE,$31,$E9,$EE
	.dc.b	$D9,$F6,$FE,$9F,$FA,$0F,$D6,$9F,$5C,$FD,$77,$FB,$2E,$FF,$EB,$CF
	.dc.b	$DA,$FF,$C8,$1C,$7F,$B6,$BD,$D6,$BF,$00,$53,$E7,$7A,$62,$F6,$3F
	.dc.b	$3B,$87,$0C,$BE,$0B,$CB,$DE,$FF,$5E,$E3,$FC,$82,$78,$E5,$D7,$F6
	.dc.b	$2F,$C0,$DE,$E7,$C9,$F8,$C6,$B0,$E3,$A3,$F2,$CF,$DD,$FE,$5C,$FC
	.dc.b	$A5,$FB,$7B,$F3,$37,$CA,$BF,$7E,$FE,$87,$EB,$70,$BD,$D6,$A7,$FA
	.dc.b	$3F,$E0,$59,$CF,$F6,$F9,$75,$CA,$E1,$C3,$2E,$9F,$63,$F9,$8A,$D6
	.dc.b	$5E,$0F,$CC,$3F,$D8,$1E,$7F,$FA,$56,$73,$7E,$3F,$86,$FE,$65,$D5
	.dc.b	$FC,$92,$73,$B7,$F5,$6F,$6B,$F8,$BF,$85,$FD,$1B,$F2,$B7,$E8,$2F
	.dc.b	$DA,$1F,$B9,$7E,$F3,$CC,$DF,$A8,$BD,$D6,$BD,$87,$A3,$A9,$FE,$27
	.dc.b	$97,$F5,$2F,$C4,$6E,$1C,$73,$ED,$9F,$F1,$AE,$EE,$11,$8F,$B7,$FD
	.dc.b	$BE,$1E,$58,$DE,$6F,$CC,$F7,$AF,$87,$E0,$2E,$5F,$C4,$BE,$82,$DF
	.dc.b	$F8,$EB,$87,$FD,$75,$F5,$EF,$C4,$9F,$95,$7F,$A8,$7E,$3D,$B3,$F5
	.dc.b	$BF,$EB,$FD,$2C,$78,$E1,$C3,$F0,$1F,$F2,$E5,$EF,$E1,$57,$BE,$41
	.dc.b	$FE,$70,$B7,$E3,$9F,$EC,$B8,$FF,$7C,$63,$D3,$4C,$47,$0B,$E4,$F1
	.dc.b	$8F,$36,$2F,$C5,$F5,$7C,$73,$F1,$FC,$66,$30,$F0,$EF,$EF,$C7,$8E
	.dc.b	$DF,$E5,$3C,$FA,$CE,$70,$F5,$CB,$F1,$6F,$5E,$5F,$EE,$1F,$A9,$78
	.dc.b	$1F,$D5,$FF,$4E,$FE,$79,$F6,$BF,$17,$D5,$DE,$DF,$02,$F7,$F1,$AB
	.dc.b	$DE,$3F,$C1,$76,$F6,$D6,$AF,$E0,$C6,$32,$7E,$F3,$F1,$27,$A9,$FE
	.dc.b	$8D,$8F,$17,$C4,$FF,$A5,$63,$E0,$19,$CB,$C1,$1F,$E9,$9F,$9F,$CE
	.dc.b	$3E,$4C,$FC,$7F,$94,$74,$8C,$57,$FA,$7F,$4F,$F1,$EE,$9F,$4B,$F0
	.dc.b	$27,$EF,$DF,$88,$F2,$FF,$AE,$7D,$DB,$F0,$EF,$EE,$17,$BA,$D6,$3F
	.dc.b	$EA,$C9,$7E,$65,$EA,$C7,$9F,$6F,$B7,$0E,$1D,$57,$4F,$C4,$F1,$7F
	.dc.b	$B6,$B3,$E7,$A7,$FA,$AA,$39,$46,$3F,$60,$F8,$39,$79,$B0,$FF,$29
	.dc.b	$EB,$E1,$FD,$33,$71,$8C,$7D,$1F,$A6,$E3,$83,$D7,$C0,$FD,$2F,$18
	.dc.b	$FA,$7F,$5C,$FA,$3E,$FF,$F9,$F7,$DD,$FE,$47,$FB,$A3,$F2,$D7,$E8
	.dc.b	$7F,$C2,$79,$7F,$33,$7B,$F7,$D7,$BA,$7E,$3B,$D3,$EC,$19,$7C,$1E
	.dc.b	$7F,$D4,$E3,$FE,$8B,$F8,$3C,$47,$FA,$C2,$B7,$97,$4D,$1F,$29,$F8
	.dc.b	$2C,$3C,$1A,$79,$7F,$19,$C6,$1C,$23,$C5,$FA,$7F,$87,$4F,$5E,$3E
	.dc.b	$AF,$CC,$BF,$1F,$F8,$93,$F2,$1F,$F1,$AF,$ED,$BF,$E5,$0F,$8C,$7E
	.dc.b	$90,$F8,$4D,$D6,$BE,$A9,$5A,$FC,$5B,$3F,$55,$EA,$CB,$A7,$EF,$37
	.dc.b	$0E,$16,$FD,$67,$F3,$75,$E1,$B6,$9E,$1F,$17,$38,$70,$D7,$F9,$27
	.dc.b	$67,$F0,$8C,$38,$5F,$9E,$5F,$03,$FA,$AF,$C0,$1F,$A8,$3F,$75,$7E
	.dc.b	$98,$F4,$DF,$BA,$BF,$67,$79,$7C,$70,$E1,$7B,$AD,$7E,$41,$9D,$3E
	.dc.b	$BE,$7D,$B3,$B8,$70,$CF,$AF,$E1,$BF,$B8,$AD,$65,$CF,$E7,$31,$7B
	.dc.b	$CF,$C5,$EB,$78,$72,$F3,$E5,$E6,$DB,$FD,$97,$C5,$FC,$A3,$3E,$73
	.dc.b	$9F,$A2,$E1,$A7,$9E,$DF,$A5,$FF,$0C,$F1,$BF,$9C,$FD,$CB,$F0,$4F
	.dc.b	$F2,$C7,$E9,$9F,$77,$9B,$5A,$FF,$A8,$A7,$3E,$0B,$F1,$2C,$FC,$67
	.dc.b	$E5,$B2,$E7,$E6,$FF,$1C,$F6,$FF,$77,$1D,$B0,$EB,$97,$F1,$88,$C7
	.dc.b	$A6,$D3,$FC,$EB,$A7,$F4,$29,$CC,$7E,$3F,$F2,$4F,$1E,$DE,$7B,$E1
	.dc.b	$B3,$E4,$3F,$93,$FD,$5B,$F4,$E7,$E8,$0F,$27,$FD,$E3,$93,$F5,$D7
	.dc.b	$E7,$DC,$2F,$75,$A9,$FF,$2B,$EF,$FD,$53,$1E,$98,$FE,$D1,$ED,$E3
	.dc.b	$DC,$38,$FE,$14,$CF,$FD,$E9,$C9,$FC,$72,$F8,$E7,$E2,$FC,$C3,$F9
	.dc.b	$CC,$62,$F8,$68,$FC,$3F,$C9,$A7,$A6,$73,$F9,$C7,$F7,$0F,$97,$FE
	.dc.b	$FB,$F9,$53,$E8,$5F,$96,$2B,$FC,$3B,$FA,$CF,$ED,$38,$FF,$C0,$5D
	.dc.b	$BF,$D9,$B0,$E1,$7B,$FA,$15,$BE,$63,$B5,$EF,$2F,$E7,$19,$F8,$3F
	.dc.b	$35,$B3,$9C,$79,$77,$7F,$7B,$AF,$2D,$F9,$35,$FF,$17,$FD,$CE,$70
	.dc.b	$E1,$1E,$6F,$C2,$FB,$C3,$E7,$1B,$F9,$AB,$D7,$F6,$3A,$9C,$E3,$D7
	.dc.b	$83,$D3,$EA,$67,$0E,$11,$EC,$F5,$7A,$FA,$0F,$95,$7C,$93,$F7,$A7
	.dc.b	$C4,$3F,$10,$7B,$1F,$8D,$EA,$F9,$3C,$4C,$B9,$63,$C7,$0E,$16,$FC
	.dc.b	$7F,$BE,$6F,$CE,$FE,$3E,$3C,$63,$C2,$FF,$80,$B9,$FF,$C2,$11,$E4
	.dc.b	$D7,$C7,$AF,$8A,$B9,$E7,$DE,$1E,$DF,$CB,$9E,$93,$9C,$DF,$97,$F0
	.dc.b	$C3,$9C,$7E,$57,$7F,$5F,$C0,$3F,$36,$FF,$7A,$7D,$53,$F6,$87,$E3
	.dc.b	$8F,$C1,$9F,$B7,$FD,$6C,$38,$7D,$17,$F9,$A6,$F7,$5F,$E0,$4E,$3F
	.dc.b	$EF,$75,$D7,$E9,$3E,$0B,$C3,$87,$E0,$1F,$E4,$87,$9E,$7E,$1C,$FF
	.dc.b	$05,$3D,$F4,$F2,$E5,$E4,$FD,$0B,$E0,$61,$C2,$3C,$3F,$3D,$BD,$7C
	.dc.b	$98,$FA,$7E,$8F,$F1,$BD,$4F,$CE,$57,$FD,$63,$F8,$03,$FC,$71,$F4
	.dc.b	$DF,$D4,$F9,$3E,$FD,$EE,$B5,$3F,$E3,$DF,$D6,$79,$E0,$FA,$5F,$BB
	.dc.b	$7B,$E1,$F6,$9F,$8C,$9E,$FF,$26,$CF,$FC,$8D,$F6,$DF,$66,$FB,$4E
	.dc.b	$72,$FF,$10,$69,$E2,$C3,$8E,$1E,$6F,$C3,$BC,$3A,$F9,$2D,$F8,$FE
	.dc.b	$31,$E5,$0F,$9E,$C6,$1E,$EC,$FF,$5F,$F9,$AB,$F3,$57,$A0,$FA,$FF
	.dc.b	$F3,$87,$A2,$FC,$49,$FB,$3F,$0B,$DD,$6A,$5F,$B2,$E7,$0F,$CF,$E3
	.dc.b	$D7,$CE,$B8,$70,$FB,$6F,$E2,$DE,$57,$E6,$6B,$C5,$7E,$19,$FE,$35
	.dc.b	$E3,$E6,$F8,$6B,$E3,$FB,$0E,$DA,$F8,$7E,$D7,$F8,$4F,$C6,$BF,$0C
	.dc.b	$7F,$00,$3F,$63,$FF,$77,$FE,$00,$F9,$5F,$F7,$8F,$7F,$EF,$30,$E1
	.dc.b	$7B,$FC,$25,$7B,$FD,$EF,$AE,$4F,$E0,$0F,$DD,$E1,$C3,$C4,$FB,$EF
	.dc.b	$2D,$3B,$FC,$97,$ED,$63,$19,$74,$FC,$5F,$18,$F9,$AF,$CB,$B3,$F2
	.dc.b	$5E,$FB,$78,$F6,$F4,$3F,$C9,$BE,$F3,$FC,$65,$F9,$93,$F3,$BF,$E8
	.dc.b	$5F,$1C,$FC,$43,$FA,$27,$27,$F1,$8B,$DD,$6A,$5F,$19,$CF,$E6,$98
	.dc.b	$FF,$16,$BF,$F1,$F7,$E6,$B6,$73,$87,$6E,$1F,$E2,$15,$D3,$E2,$7F
	.dc.b	$B7,$F2,$E2,$F7,$1F,$DC,$B3,$7E,$73,$DF,$3F,$44,$E7,$AF,$19,$F9
	.dc.b	$7F,$29,$B3,$C7,$F7,$DF,$E8,$BF,$84,$BF,$32,$FF,$66,$FE,$46,$FD
	.dc.b	$5D,$F9,$7F,$F9,$5B,$F0,$4D,$D6,$A7,$3F,$66,$F8,$1C,$FE,$11,$FC
	.dc.b	$D3,$2E,$9F,$83,$5C,$38,$65,$FE,$55,$F1,$37,$A7,$78,$E5,$18,$CF
	.dc.b	$AF,$DC,$D9,$FF,$33,$FC,$07,$F5,$CF,$BE,$FF,$72,$FD,$1B,$EE,$FF
	.dc.b	$C2,$9F,$97,$7F,$AA,$FD,$7D,$66,$AD,$7D,$32,$BF,$D6,$1F,$B9,$EA
	.dc.b	$7F,$A5,$E3,$FE,$F0,$F9,$5B,$87,$1C,$BA,$FE,$6B,$FD,$16,$BC,$31
	.dc.b	$E0,$F9,$6C,$56,$F3,$ED,$C6,$FD,$7F,$B5,$F9,$A1,$ED,$FF,$C0,$0F
	.dc.b	$C9,$BF,$D6,$FF,$78,$FD,$69,$F8,$EB,$F7,$EF,$F1,$A7,$8C,$FC,$A1
	.dc.b	$FA,$47,$1D,$C7,$F9,$C3,$0E,$37,$C2,$9F,$BB,$EE,$B5,$F4,$1F,$EC
	.dc.b	$9D,$3C,$38,$78,$27,$FD,$D7,$18,$CF,$C5,$F6,$DF,$90,$C7,$BF,$A6
	.dc.b	$AD,$67,$E6,$FC,$6F,$31,$8B,$E1,$E1,$F8,$78,$3D,$D9,$74,$E0,$F6
	.dc.b	$7E,$31,$9D,$BD,$38,$77,$F7,$31,$7B,$D9,$F9,$3F,$AF,$E4,$FF,$05
	.dc.b	$EE,$3F,$4C,$7E,$E3,$FE,$09,$FD,$07,$FD,$27,$C9,$E7,$E5,$F8,$F8
	.dc.b	$70,$BD,$FD,$4A,$FF,$C2,$97,$BB,$FE,$BB,$D9,$FC,$2D,$F8,$A7,$27
	.dc.b	$F9,$2B,$83,$79,$78,$F2,$FF,$0D,$77,$FC,$49,$E1,$39,$C5,$FA,$66
	.dc.b	$67,$8E,$7E,$4C,$DF,$83,$7B,$F3,$FE,$39,$39,$FC,$13,$EE,$7C,$D8
	.dc.b	$7E,$46,$FD,$6B,$FD,$9D,$F8,$C3,$F6,$6F,$F3,$27,$17,$B6,$F8,$5E
	.dc.b	$FE,$03,$FD,$DB,$5F,$E6,$1A,$D5,$74,$DB,$DC,$F7,$FD,$43,$F1,$51
	.dc.b	$F9,$28,$E7,$97,$83,$F5,$6F,$64,$63,$1F,$F7,$CF,$9D,$88,$F3,$D7
	.dc.b	$9B,$27,$EB,$1E,$C9,$CD,$F9,$3F,$1A,$FC,$D7,$E6,$7F,$EC,$FF,$8E
	.dc.b	$F7,$7F,$00,$BF,$4C,$7E,$FF,$FE,$49,$E1,$FB,$57,$BD,$57,$BA,$FF
	.dc.b	$59,$56,$BF,$13,$FE,$B9,$C3,$F6,$F0,$FF,$6E,$7E,$29,$EE,$84,$73
	.dc.b	$C3,$C1,$F3,$BF,$66,$31,$9F,$F9,$32,$BC,$F8,$7F,$67,$CF,$C3,$3E
	.dc.b	$6F,$D1,$73,$AF,$AA,$3D,$33,$83,$7A,$BF,$5A,$C6,$9E,$9F,$BF,$FD
	.dc.b	$E7,$E7,$4F,$9E,$FB,$DD,$BF,$B5,$F1,$9F,$8F,$FC,$CE,$4F,$CE,$5E
	.dc.b	$FA,$7F,$51,$E9,$7F,$8E,$A9,$CF,$D3,$2F,$C4,$72,$7E,$77,$0E,$DF
	.dc.b	$55,$F0,$DE,$5A,$77,$F2,$3F,$35,$18,$CB,$FC,$8B,$EF,$63,$1F,$2C
	.dc.b	$77,$FD,$4B,$B6,$BE,$2D,$FC,$FF,$2D,$FC,$5F,$E6,$9F,$44,$FD,$1F
	.dc.b	$F6,$DF,$51,$FE,$2E,$F9,$87,$E6,$BB,$AD,$79,$DF,$E7,$CF,$A9,$FE
	.dc.b	$E7,$F2,$D9,$C7,$ED,$65,$D3,$92,$E1,$C3,$3F,$F1,$CF,$E0,$DF,$B3
	.dc.b	$AF,$0D,$F8,$31,$7D,$5E,$70,$E1,$1F,$E0,$0F,$DC,$3B,$69,$E1,$D9
	.dc.b	$F5,$3F,$CC,$7E,$1D,$FA,$CB,$F7,$37,$E3,$BF,$B7,$FC,$C3,$FC,$95
	.dc.b	$F8,$BE,$EB,$53,$9F,$B0,$FF,$49,$F8,$9F,$C5,$71,$E7,$8F,$E5,$5E
	.dc.b	$BE,$05,$C3,$8F,$92,$CD,$F8,$77,$F0,$FB,$E3,$9F,$87,$F0,$FC,$61
	.dc.b	$C2,$FF,$CB,$9C,$BE,$3C,$FC,$F9,$73,$D7,$F9,$86,$BE,$3C,$38,$43
	.dc.b	$F6,$5E,$19,$FA,$7E,$53,$FA,$97,$C8,$3F,$29,$7E,$DB,$F9,$0F,$C5
	.dc.b	$DF,$90,$3E,$E1,$FA,$77,$0B,$DF,$9F,$FE,$FD,$AD,$7C,$0E,$7E,$07
	.dc.b	$F6,$D9,$F5,$CE,$E1,$C3,$3E,$B9,$FF,$2F,$AD,$65,$CF,$1C,$5E,$F2
	.dc.b	$F1,$79,$FE,$1C,$BC,$D1,$E5,$FB,$07,$1B,$F7,$6F,$EE,$7E,$9F,$FA
	.dc.b	$0B,$F1,$37,$E5,$4F,$CD,$DF,$34,$F8,$A7,$D7,$3F,$19,$E7,$97,$2F
	.dc.b	$2B,$FD,$CD,$8F,$1C,$38,$5B,$F7,$DD,$E9,$FE,$02,$BD,$BE,$23,$FB
	.dc.b	$BE,$1C,$F4,$FE,$4D,$19,$38,$9E,$7C,$7B,$67,$FE,$34,$D7,$D9,$3D
	.dc.b	$6B,$D7,$AF,$AA,$BF,$C7,$95,$AA,$E5,$39,$CD,$FB,$BF,$6C,$FD,$73
	.dc.b	$9F,$7F,$C3,$F8,$AB,$F2,$77,$E0,$0F,$C0,$BF,$CD,$7F,$AA,$3F,$12
	.dc.b	$7E,$6F,$C2,$F7,$5A,$97,$E8,$19,$F8,$BF,$77,$06,$F2,$BC,$3E,$7D
	.dc.b	$3C,$11,$DF,$D4,$6B,$DE,$F8,$67,$E2,$F6,$BE,$5C,$7F,$8C,$30,$E3
	.dc.b	$3E,$1F,$CD,$F7,$87,$2D,$5F,$AE,$63,$27,$E6,$5C,$B2,$65,$CB,$0E
	.dc.b	$1F,$9B,$7D,$4F,$B7,$FF,$53,$7E,$94,$FD,$ED,$F9,$F7,$F8,$8B,$F7
	.dc.b	$8F,$F3,$C7,$1F,$D1,$BD,$D6,$BE,$99,$5F,$E0,$AA,$D7,$E1,$CF,$C6
	.dc.b	$7E,$1F,$87,$F9,$8C,$7B,$EF,$F8,$5B,$DE,$9E,$1D,$3F,$C6,$3F,$41
	.dc.b	$E9,$BE,$F1,$FE,$34,$F4,$B1,$9A,$F8,$6B,$E2,$D3,$D9,$39,$DB,$CD
	.dc.b	$B6,$F4,$FC,$57,$C8,$BF,$5E,$FF,$74,$60,$FD,$D1,$F1,$5F,$82,$FC
	.dc.b	$57,$F3,$37,$BF,$C4,$FD,$95,$AF,$0D,$FA,$EE,$B2,$7E,$3F,$F0,$39
	.dc.b	$BF,$CD,$38,$F1,$7E,$07,$3F,$EE,$A3,$A6,$BE,$1D,$7F,$CC,$DB,$79
	.dc.b	$B1,$F2,$E9,$E4,$F3,$FE,$E6,$1C,$23,$C5,$F7,$FE,$1D,$FC,$B8,$FE
	.dc.b	$6B,$F0,$1F,$D8,$FE,$63,$FE,$B1,$F2,$3F,$D0,$3F,$AD,$3F,$27,$FC
	.dc.b	$3F,$E1,$BA,$FC,$23,$F3,$BF,$6E,$AD,$78,$AF,$C7,$F5,$D3,$ED,$E7
	.dc.b	$69,$D5,$F6,$9F,$E8,$87,$C1,$87,$3C,$FF,$41,$CB,$FB,$B9,$CE,$5E
	.dc.b	$3F,$33,$F3,$11,$8A,$FC,$DF,$EC,$3E,$5D,$FD,$31,$F9,$EB,$C6,$3D
	.dc.b	$75,$FE,$EF,$97,$5D,$BD,$75,$EA,$F9,$BF,$E2,$7F,$B0,$7F,$99,$3C
	.dc.b	$A7,$E4,$6F,$DF,$9F,$8E,$FF,$AA,$BF,$66,$DF,$DD,$7E,$AA,$B5,$3F
	.dc.b	$EA,$EF,$D9,$F3,$87,$C3,$C7,$FC,$DF,$E5,$5C,$38,$65,$D3,$E2,$3F
	.dc.b	$69,$5E,$1C,$B9,$FC,$5E,$2F,$79,$FF,$AE,$BE,$AF,$AD,$78,$3F,$09
	.dc.b	$7E,$5D,$FC,$B3,$E7,$3F,$CF,$BE,$9B,$F5,$47,$E1,$EF,$C8,$9F,$B3
	.dc.b	$FE,$7F,$A7,$F9,$56,$1C,$2F,$75,$AF,$1F,$A6,$FF,$EE,$BF,$0B,$F9
	.dc.b	$F6,$1C,$21,$F8,$9F,$E0,$71,$E5,$9F,$7F,$D5,$BF,$88,$46,$32,$FF
	.dc.b	$2B,$61,$C7,$5F,$35,$F9,$75,$F2,$CE,$75,$F1,$C3,$F8,$07,$FA,$4F
	.dc.b	$EF,$6F,$DB,$1F,$B7,$FF,$1E,$F8,$C7,$E6,$39,$8D,$CE,$BC,$4F,$E8
	.dc.b	$B3,$C7,$F6,$DC,$D7,$E6,$70,$E9,$E1,$5C,$38,$63,$D3,$B3,$F9,$55
	.dc.b	$75,$CB,$9F,$76,$2F,$79,$F9,$7B,$3A,$D7,$82,$7E,$BF,$07,$F5,$9C
	.dc.b	$39,$65,$E8,$97,$E6,$3C,$23,$1F,$99,$FF,$80,$3F,$47,$FE,$FF,$F9
	.dc.b	$DB,$E8,$BF,$CA,$DE,$37,$FB,$2F,$EA,$5F,$92,$3E,$3B,$87,$83,$FE
	.dc.b	$F8,$C3,$85,$7F,$42,$FB,$1F,$F2,$F5,$FF,$48,$E0,$67,$DB,$3F,$F6
	.dc.b	$F6,$3F,$CD,$63,$18,$7F,$5C,$6E,$7E,$1E,$DF,$F7,$BC,$63,$2F,$F3
	.dc.b	$57,$E4,$98,$C3,$C5,$8F,$A2,$7C,$3F,$19,$E6,$96,$BE,$6C,$BD,$9F
	.dc.b	$B5,$F1,$97,$A2,$FC,$1B,$FC,$9B,$FA,$C7,$E7,$7D,$2F,$F6,$EF,$E1
	.dc.b	$FC,$EB,$5E,$3A,$B5,$E5,$7E,$1B,$93,$DF,$90,$EB,$93,$F4,$CF,$74
	.dc.b	$67,$1E,$6F,$C3,$31,$3E,$2F,$25,$A7,$FB,$BA,$7C,$53,$98,$7D,$AF
	.dc.b	$AE,$FC,$B9,$F9,$3F,$6B,$DE,$BE,$7C,$B9,$EA,$FD,$37,$E3,$7C,$B3
	.dc.b	$9F,$FC,$7D,$F4,$2F,$C4,$DF,$BC,$BF,$1A,$7A,$DC,$2F,$75,$A9,$7E
	.dc.b	$49,$9F,$D2,$BD,$FC,$BF,$E1,$4B,$FF,$25,$3E,$F3,$1F,$49,$FE,$4B
	.dc.b	$C3,$BF,$C7,$FD,$5D,$7C,$93,$9D,$1F,$8E,$67,$1F,$0C,$79,$7E,$D3
	.dc.b	$C9,$5B,$D5,$E8,$63,$83,$D9,$7E,$BF,$88,$C6,$BE,$9F,$DE,$FF,$11
	.dc.b	$F2,$0F,$8A,$7E,$57,$FE,$39,$FA,$07,$EA,$AF,$D2,$DF,$A4,$70,$BD
	.dc.b	$FD,$32,$F7,$4F,$C3,$75,$8F,$C4,$CB,$B7,$C2,$5C,$78,$E1,$DB,$E4
	.dc.b	$BF,$2D,$5E,$3B,$F1,$69,$FD,$7B,$D1,$FF,$3B,$CE,$74,$F0,$4F,$97
	.dc.b	$F6,$4E,$F1,$8D,$7F,$C5,$7E,$5F,$E5,$BB,$7F,$D2,$BF,$43,$EA,$A7
	.dc.b	$E4,$0F,$C7,$DF,$79,$FE,$18,$C3,$87,$6F,$F8,$D3,$97,$FC,$B9,$7B
	.dc.b	$A7,$E9,$9F,$B1,$AD,$67,$F7,$DF,$F1,$DF,$37,$E1,$1E,$59,$F8,$33
	.dc.b	$FF,$0A,$F0,$FF,$A4,$A7,$39,$7D,$AC,$3F,$71,$18,$AF,$2F,$E0,$DD
	.dc.b	$F6,$F5,$5F,$8F,$AF,$F8,$17,$EB,$CF,$DC,$3F,$41,$E4,$3F,$4A,$7E
	.dc.b	$8D,$FE,$5C,$CF,$75,$A9,$CF,$C3,$3F,$79,$CF,$A3,$F9,$CC,$BA,$79
	.dc.b	$D7,$0E,$19,$7D,$5E,$1F,$E0,$55,$AD,$3C,$12,$EE,$FE,$25,$39,$D3
	.dc.b	$BF,$89,$E2,$CB,$F5,$F1,$E2,$F0,$F1,$8F,$2D,$BF,$D1,$FD,$7F,$6F
	.dc.b	$1E,$59,$FF,$9D,$BF,$3A,$FE,$45,$FA,$63,$17,$EC,$0F,$CC,$1F,$BA
	.dc.b	$BF,$37,$7E,$46,$FD,$6B,$8E,$1C,$2F,$75,$AF,$CE,$F3,$F0,$9F,$BE
	.dc.b	$CF,$B6,$77,$0E,$19,$F5,$FC,$A7,$F9,$85,$6B,$2E,$7F,$8E,$62,$F7
	.dc.b	$9F,$8B,$F7,$3F,$0E,$5E,$7C,$BC,$DB,$3F,$78,$FE,$7D,$F7,$6F,$C4
	.dc.b	$7F,$BC,$FE,$B9,$FE,$F0,$87,$EF,$8F,$CB,$77,$5A,$F2,$15,$AF,$98
	.dc.b	$CF,$CE,$FF,$18,$8E,$9F,$3D,$6F,$84,$7F,$B8,$BC,$3E,$72,$CF,$C1
	.dc.b	$F5,$DF,$D3,$6B,$8C,$78,$F4,$67,$C2,$31,$FB,$C7,$FB,$CB,$E3,$BF
	.dc.b	$DE,$3F,$81,$FF,$BD,$3E,$DD,$FB,$1F,$FB,$BF,$F6,$07,$F0,$07,$F3
	.dc.b	$5F,$17,$FA,$C7,$07,$F4,$FC,$38,$5E,$E9,$F9,$9F,$5A,$D5,$FF,$A4
	.dc.b	$9F,$F5,$36,$5E,$08,$FD,$CC,$60,$DE,$5E,$2F,$84,$67,$FC,$76,$7A
	.dc.b	$0C,$FC,$33,$AF,$8E,$FE,$C5,$87,$9A,$B4,$FF,$A8,$B4,$F3,$7E,$19
	.dc.b	$CB,$0E,$38,$FB,$BE,$3B,$17,$BC,$BC,$3B,$3F,$55,$CE,$BE,$5C,$3E
	.dc.b	$2F,$E4,$7E,$4F,$D6,$3C,$9F,$A7,$3F,$45,$7E,$41,$FE,$74,$FC,$99
	.dc.b	$FA,$C6,$F8,$17,$BA,$D7,$E5,$79,$8F,$F7,$8F,$37,$F7,$DE,$56,$7F
	.dc.b	$07,$E5,$F3,$1F,$E0,$18,$F0,$E9,$E0,$FB,$9F,$EB,$33,$CE,$31,$F3
	.dc.b	$7F,$1B,$0E,$97,$C3,$3F,$37,$E3,$DB,$7D,$1C,$BF,$43,$E6,$3F,$D4
	.dc.b	$BF,$B7,$BE,$3F,$FD,$1B,$F9,$33,$F1,$67,$E4,$5B,$F8,$25,$EF,$EF
	.dc.b	$D7,$BF,$C8,$75,$3E,$AE,$2F,$F0,$25,$BF,$27,$C6,$9E,$0D,$3B,$FE
	.dc.b	$99,$E9,$BF,$F1,$04,$E7,$3E,$BF,$93,$66,$79,$E5,$E1,$D3,$CD,$F9
	.dc.b	$D7,$83,$2F,$4E,$EF,$C7,$BF,$1B,$FA,$8B,$F2,$67,$E3,$BF,$E3,$CF
	.dc.b	$E0,$0F,$CE,$ED,$FD,$73,$A7,$EA,$5E,$EB,$5E,$6F,$FA,$C3,$F5,$6E
	.dc.b	$78,$FF,$78,$E0,$FC,$F0,$BE,$0E,$1E,$5F,$0A,$CF,$F5,$7F,$A9,$FA
	.dc.b	$E7,$31,$FA,$EF,$86,$C6,$1E,$6B,$E1,$97,$FB,$03,$CC,$F0,$E9,$E4
	.dc.b	$D9,$FB,$46,$2F,$70,$F4,$B1,$9F,$BB,$E8,$7F,$5F,$F1,$EF,$D6,$7F
	.dc.b	$C2,$5C,$4F,$49,$F8,$1B,$E2,$5F,$8A,$70,$BD,$D6,$A5,$F8,$AE,$7E
	.dc.b	$97,$F0,$19,$75,$FD,$A6,$E1,$C3,$2F,$F5,$7F,$57,$E8,$6B,$C5,$6D
	.dc.b	$3C,$54,$FA,$BF,$C0,$56,$A7,$39,$FF,$A1,$70,$E5,$8F,$58,$FF,$47
	.dc.b	$69,$FC,$1F,$F2,$6F,$F4,$CF,$B6,$FD,$D1,$F3,$0F,$CC,$7F,$AB,$FE
	.dc.b	$6B,$FC,$4A,$F7,$5A,$9F,$EB,$7F,$9A,$E7,$F0,$A6,$3F,$9D,$BF,$F5
	.dc.b	$2F,$C1,$59,$CE,$1C,$7C,$1F,$CC,$57,$4C,$FC,$1A,$E2,$F7,$1E,$3F
	.dc.b	$90,$7E,$C7,$DF,$3F,$44,$E6,$7F,$93,$FE,$C2,$FD,$51,$F9,$53,$E7
	.dc.b	$9F,$BF,$BF,$62,$7E,$5D,$FE,$77,$E6,$FE,$55,$8F,$1C,$38,$5E,$EB
	.dc.b	$FD,$1D,$5A,$C7,$A7,$9B,$F9,$EC,$38,$36,$3C,$39,$78,$32,$79,$BF
	.dc.b	$B7,$BE,$D3,$9C,$31,$7C,$32,$F0,$7E,$A3,$EC,$8C,$63,$E8,$FB,$EC
	.dc.b	$6B,$E6,$8F,$5F,$C2,$E3,$2F,$06,$CF,$85,$FE,$97,$FB,$6B,$F4,$B7
	.dc.b	$EA,$2F,$C7,$DF,$91,$3F,$2D,$7E,$F9,$F8,$FE,$3F,$E9,$D8,$F1,$C3
	.dc.b	$85,$EE,$BF,$CF,$75,$AC,$7A,$F3,$7F,$02,$C3,$C1,$FB,$27,$F2,$17
	.dc.b	$96,$7E,$0C,$DF,$05,$FC,$6E,$73,$9F,$8F,$E0,$71,$97,$A3,$0E,$1A
	.dc.b	$3F,$06,$F6,$CE,$6F,$C9,$F1,$BF,$DA,$BF,$18,$7E,$6A,$F8,$F7,$EE
	.dc.b	$1F,$EB,$5F,$5F,$E3,$7E,$B7,$BA,$D4,$E7,$99,$FB,$BE,$7F,$13,$FE
	.dc.b	$75,$97,$4C,$AE,$1C,$3E,$2F,$FD,$37,$87,$F3,$5A,$D6,$3D,$FE,$8F
	.dc.b	$17,$BD,$3E,$B7,$8F,$E1,$CB,$CB,$97,$2F,$C6,$7F,$89,$7C,$37,$F6
	.dc.b	$5F,$97,$7E,$7A,$FD,$01,$E5,$3F,$4D,$FF,$18,$FE,$FA,$FD,$3F,$8F
	.dc.b	$85,$FE,$E9,$C7,$8E,$1C,$2F,$7F,$A4,$6A,$95,$DB,$CC,$FE,$E9,$1F
	.dc.b	$DD,$7C,$5F,$E9,$0F,$3D,$3C,$4E,$8E,$55,$DB,$4F,$27,$D5,$3D,$7F
	.dc.b	$F0,$CE,$FE,$BC,$3D,$55,$AF,$5B,$38,$F5,$DB,$CD,$8B,$83,$F8,$BE
	.dc.b	$9E,$2A,$F3,$E3,$CB,$7F,$83,$F8,$47,$8A,$31,$87,$1F,$84,$C6,$5D
	.dc.b	$32,$7C,$2F,$C1,$FC,$C3,$F9,$17,$98,$FC,$A3,$FC,$A5,$E3,$3F,$35
	.dc.b	$7E,$E7,$BF,$82,$5E,$FE,$B7,$F5,$D5,$FD,$87,$8F,$CD,$97,$F2,$DF
	.dc.b	$D5,$BE,$37,$27,$1C,$2D,$6B,$1E,$7A,$78,$3F,$37,$F8,$76,$C7,$CB
	.dc.b	$3E,$4B,$CC,$63,$3F,$0F,$E3,$9D,$6B,$D3,$A3,$F3,$8F,$BD,$F9,$9B
	.dc.b	$D3,$7E,$B5,$EC,$FC,$E5,$FA,$2B,$F3,$67,$E5,$5C,$2F,$7C,$AB,$DD
	.dc.b	$3F,$5D,$D7,$E4,$7E,$CF,$0F,$FC,$79,$A7,$7D,$31,$C8,$11,$9C,$BA
	.dc.b	$6B,$FE,$63,$E4,$FB,$D3,$9D,$3C,$BF,$35,$F1,$23,$15,$FE,$00,$FC
	.dc.b	$2F,$CD,$97,$AE,$3F,$C0,$7F,$09,$F9,$DF,$D3,$1F,$89,$7F,$61,$F8
	.dc.b	$37,$E5,$6F,$CE,$3F,$CE,$7F,$B7,$EE,$B5,$39,$FB,$BF,$E4,$FC,$1F
	.dc.b	$91,$C7,$9F,$E7,$1E,$CE,$34,$76,$F9,$6C,$37,$4E,$FA,$7F,$20,$AE
	.dc.b	$D5,$FC,$D3,$2F,$17,$D9,$7E,$03,$0E,$11,$E1,$FC,$0F,$78,$F2,$D1
	.dc.b	$F8,$2E,$2B,$CD,$F9,$65,$D3,$FD,$1B,$FA,$67,$EF,$7F,$80,$1F,$AF
	.dc.b	$BF,$37,$7E,$4C,$FC,$85,$FB,$6B,$F2,$7F,$AB,$CD,$FA,$FC,$78,$E1
	.dc.b	$C2,$F7,$4F,$DA,$3B,$65,$D7,$9B,$F8,$26,$3C,$63,$FD,$9D,$18,$F0
	.dc.b	$BF,$D0,$19,$74,$D3,$C5,$A3,$E2,$3F,$43,$39,$D1,$F9,$0F,$F1,$3C
	.dc.b	$38,$E5,$EB,$C7,$8E,$BF,$B6,$D3,$C1,$B7,$9F,$4F,$67,$EA,$3F,$C7
	.dc.b	$FF,$72,$0F,$D2,$7F,$27,$F5,$9F,$E0,$EF,$DF,$F5,$39,$F1,$93,$FE
	.dc.b	$12,$F4,$B3,$FA,$9F,$E0,$6B,$9E,$15,$E1,$5F,$E2,$FF,$81,$FD,$DC
	.dc.b	$6B,$0E,$7E,$13,$F2,$5C,$CF,$08,$FF,$1A,$FE,$6B,$E2,$C7,$CB,$1F
	.dc.b	$E3,$8F,$B2,$FE,$71,$F2,$C1,$E3,$FF,$41,$FD,$7B,$ED,$FD,$7F,$F8
	.dc.b	$97,$97,$F9,$45,$6E,$B5,$C8,$AF,$E9,$95,$AA,$E7,$DD,$F5,$CF,$DC
	.dc.b	$3D,$FF,$44,$FD,$03,$CF,$E4,$19,$7E,$2B,$EB,$7F,$8C,$CE,$63,$F1
	.dc.b	$BF,$AE,$E2,$3C,$F5,$C3,$3F,$17,$DD,$FB,$27,$37,$EA,$FD,$C3,$1B
	.dc.b	$7A,$36,$7D,$86,$32,$E5,$1F,$D3,$F7,$FE,$8D,$FA,$F3,$F6,$B7,$E7
	.dc.b	$2F,$DE,$7F,$9C,$FC,$A3,$F6,$97,$EA,$5F,$D8,$F0,$E1,$7B,$F8,$DF
	.dc.b	$F4,$15,$3F,$36,$D6,$5D,$3A,$3E,$30,$C3,$C1,$F5,$1F,$C1,$5E,$7A
	.dc.b	$78,$3B,$FF,$9E,$C6,$32,$FF,$47,$7C,$DE,$31,$F3,$DF,$9B,$37,$EA
	.dc.b	$DE,$0D,$7C,$9B,$FA,$7F,$49,$FF,$4C,$7E,$B3,$FE,$EB,$C7,$FD,$9D
	.dc.b	$E5,$BF,$0A,$FF,$03,$7C,$5F,$F8,$8B,$F1,$8D,$F1,$FF,$33,$BD,$D6
	.dc.b	$BE,$A5,$F8,$BE,$AB,$E5,$7E,$2B,$FB,$D7,$FC,$0B,$D2,$D3,$C1,$3F
	.dc.b	$BE,$9C,$C6,$34,$F1,$7D,$0F,$F4,$DF,$4B,$FA,$16,$1C,$6F,$C5,$87
	.dc.b	$F1,$5C,$FC,$FF,$2B,$AC,$B9,$47,$97,$E3,$BC,$9B,$7A,$33,$FE,$B1
	.dc.b	$B3,$F5,$2F,$41,$FE,$DF,$F8,$17,$C8,$FF,$7B,$F8,$47,$EF,$CF,$2B
	.dc.b	$7E,$EA,$F7,$F2,$FF,$CB,$A9,$E5,$6B,$F5,$0F,$EF,$3C,$3F,$D6,$F9
	.dc.b	$3C,$18,$F1,$7F,$90,$C6,$75,$F1,$6B,$E1,$FA,$BF,$65,$7F,$63,$9C
	.dc.b	$E7,$FC,$F3,$F3,$CC,$CF,$4D,$7C,$7B,$FA,$3F,$6C,$F3,$C6,$37,$7E
	.dc.b	$B3,$8D,$3D,$D1,$FE,$29,$E9,$F9,$D8,$F3,$9C,$FE,$0B,$F7,$7F,$72
	.dc.b	$7E,$B5,$FE,$1F,$FD,$F5,$F9,$07,$F2,$5F,$C5,$5F,$99,$EE,$B5,$39
	.dc.b	$9C,$71,$7C,$4C,$FF,$BF,$63,$D3,$F6,$EC,$42,$3B,$CB,$2E,$FF,$01
	.dc.b	$FE,$52,$AE,$71,$8F,$47,$E1,$DF,$08,$DE,$3F,$E5,$8D,$FD,$B3,$9C
	.dc.b	$FD,$1F,$AE,$FE,$67,$E9,$7F,$80,$7D,$5B,$F0,$3F,$E2,$FF,$C0,$0F
	.dc.b	$DB,$7F,$D4,$39,$7C,$6B,$DD,$6A,$5F,$9F,$F3,$F5,$5F,$97,$7B,$2F
	.dc.b	$AF,$0F,$E0,$0F,$F0,$F3,$CB,$E0,$59,$BF,$0E,$FB,$B3,$98,$FF,$4B
	.dc.b	$57,$5D,$7C,$D7,$C3,$3F,$EA,$FA,$F9,$73,$FF,$5E,$6B,$E5,$FA,$BF
	.dc.b	$DB,$7E,$62,$FD,$2F,$FC,$D1,$F4,$2F,$B1,$7E,$EC,$F5,$3F,$A8,$67
	.dc.b	$F8,$4C,$38,$7D,$B7,$FB,$1A,$F7,$4F,$CB,$BA,$E4,$FD,$7F,$ED,$E1
	.dc.b	$C3,$8B,$E0,$07,$9E,$9E,$1D,$1F,$1D,$F9,$C9,$CE,$7E,$4F,$8F,$C6
	.dc.b	$5E,$88,$F0,$FE,$87,$E0,$DB,$C9,$8F,$A7,$D7,$C6,$7D,$77,$FF,$2A
	.dc.b	$70,$7A,$F2,$E5,$7E,$FF,$3F,$F8,$D7,$E1,$BE,$2F,$D7,$9F,$2C,$FD
	.dc.b	$5B,$EA,$FE,$80,$E1,$DF,$1A,$F7,$5A,$FC,$43,$3A,$7E,$73,$2E,$9C
	.dc.b	$BF,$09,$ED,$9F,$5E,$E7,$E4,$FF,$BC,$9C,$E3,$F8,$4F,$9E,$C6,$9D
	.dc.b	$F0,$E1,$B3,$F0,$7F,$1E,$DE,$2C,$D3,$FE,$BF,$E2,$FD,$EF,$A6,$F6
	.dc.b	$DF,$A7,$3F,$54,$7E,$52,$FD,$19,$F9,$B7,$E8,$71,$E3,$87,$0B,$DD
	.dc.b	$6B,$ED,$FF,$A1,$65,$D7,$2F,$DD,$BD,$FF,$27,$B8,$71,$CF,$B4,$7F
	.dc.b	$1A,$9E,$18,$F4,$E9,$C5,$EF,$2F,$1F,$F0,$07,$68,$C6,$AF,$CF,$BF
	.dc.b	$A9,$7E,$3E,$FC,$A9,$F7,$CF,$E0,$17,$D6,$BF,$14,$79,$0F,$C2,$38
	.dc.b	$E1,$C2,$F7,$5A,$FD,$8F,$38,$FF,$16,$CB,$B7,$E0,$B7,$0E,$19,$F5
	.dc.b	$FA,$0F,$E6,$B5,$E3,$BF,$14,$F8,$7E,$03,$37,$C3,$2F,$1F,$C9,$76
	.dc.b	$D3,$C3,$B3,$E5,$F1,$8F,$A6,$1F,$A6,$E3,$0F,$5F,$EA,$8E,$E7,$E6
	.dc.b	$3F,$D4,$7D,$67,$EF,$8F,$D5,$5F,$B4,$7F,$A7,$7E,$5D,$FB,$07,$C5
	.dc.b	$C7,$8F,$A1,$FE,$B3,$C7,$8E,$1C,$3F,$04,$61,$C3,$AB,$C1,$A7,$7D
	.dc.b	$3F,$D8,$D8,$FF,$9F,$A3,$18,$7F,$B7,$1B,$A7,$87,$E6,$3F,$BF,$C6
	.dc.b	$32,$EB,$F3,$18,$C3,$C5,$8F,$A7,$0F,$46,$CF,$D7,$7C,$35,$AC,$FC
	.dc.b	$DA,$FB,$3E,$E7,$32,$FA,$4F,$F5,$E7,$E4,$8F,$CB,$1F,$B3,$FF,$86
	.dc.b	$BF,$4C,$DD,$6A,$73,$F0,$3F,$E2,$8E,$6F,$8B,$8F,$3E,$27,$E4,$FE
	.dc.b	$D8,$FF,$6D,$61,$D9,$F8,$46,$25,$9F,$87,$3F,$F2,$97,$B1,$EF,$AE
	.dc.b	$93,$9C,$BF,$88,$47,$92,$F8,$E7,$E4,$F8,$17,$EE,$39,$C3,$84,$3F
	.dc.b	$7F,$F8,$3D,$5F,$E3,$8F,$75,$F9,$6B,$F5,$17,$EF,$9F,$E9,$AF,$C3
	.dc.b	$F7,$F5,$FF,$C8,$6B,$53,$FE,$9E,$FA,$5C,$D7,$BF,$81,$97,$57,$DE
	.dc.b	$63,$0E,$58,$F6,$F8,$3F,$75,$78,$F2,$EB,$F0,$B8,$BD,$E7,$FE,$AC
	.dc.b	$FC,$D7,$C7,$AF,$A3,$2E,$5F,$B7,$63,$1F,$2E,$7E,$FE,$AF,$AB,$A7
	.dc.b	$5B,$F0,$7D,$17,$E0,$BF,$3F,$FE,$FD,$F6,$EF,$C7,$3F,$03,$F7,$EF
	.dc.b	$A0,$7E,$6F,$BF,$99,$FF,$36,$D6,$A7,$FC,$7F,$A7,$E0,$B1,$FF,$04
	.dc.b	$7E,$D7,$F0,$DE,$BC,$DF,$09,$EF,$7D,$BD,$CF,$BD,$18,$F0,$BF,$D5
	.dc.b	$DE,$D6,$33,$EF,$1F,$EB,$1C,$FF,$C6,$DE,$7F,$F9,$2E,$F8,$6D,$E7
	.dc.b	$FA,$BF,$CE,$FE,$0D,$FD,$57,$E8,$9F,$8C,$3F,$32,$7A,$BF,$E5,$BF
	.dc.b	$CE,$B7,$F1,$5F,$E0,$FA,$D4,$FF,$97,$FE,$DB,$38,$7D,$DC,$7F,$B0
	.dc.b	$79,$57,$0E,$19,$74,$E8,$FC,$C5,$78,$72,$E7,$F9,$66,$2F,$79,$FF
	.dc.b	$9A,$FE,$BB,$AD,$78,$34,$FE,$6B,$D5,$FA,$CF,$D0,$5F,$AB,$FF,$83
	.dc.b	$7E,$61,$F4,$AF,$A2,$7E,$79,$8C,$38,$5E,$EB,$52,$FA,$AF,$E0,$38
	.dc.b	$F4,$F8,$2F,$CC,$3D,$BF,$04,$B8,$71,$C9,$F9,$C7,$EC,$EB,$86,$9E
	.dc.b	$29,$7E,$43,$9B,$E1,$9F,$FA,$23,$5F,$1E,$7F,$E0,$5D,$7C,$7F,$94
	.dc.b	$F8,$AF,$7D,$16,$B5,$97,$B3,$A7,$F8,$BF,$ED,$8E,$77,$E7,$BF,$ED
	.dc.b	$CF,$C3,$1F,$90,$FC,$3F,$03,$75,$AF,$C4,$55,$AC,$9F,$AE,$6B,$1F
	.dc.b	$DC,$65,$D7,$C7,$B8,$F1,$C3,$AB,$FC,$4E,$7C,$55,$E1,$CD,$F8,$66
	.dc.b	$6B,$84,$7D,$5F,$C1,$BB,$69,$E2,$8F,$B1,$FC,$01,$FC,$CB,$7F,$EF
	.dc.b	$1F,$B2,$BF,$1D,$7E,$AB,$FE,$1B,$FC,$75,$F3,$DB,$FA,$15,$EE,$B5
	.dc.b	$FB,$16,$7F,$30,$FE,$25,$97,$4F,$5B,$F8,$8B,$DB,$3E,$BF,$60,$D3
	.dc.b	$F9,$74,$E7,$1F,$F1,$57,$AB,$8D,$3B,$E1,$C2,$DF,$BF,$F8,$F5,$F1
	.dc.b	$65,$CF,$F3,$0F,$E9,$B2,$F8,$17,$E0,$0F,$D7,$BF,$CD,$DF,$A3,$7F
	.dc.b	$BD,$BD,$9D,$D6,$A7,$3E,$FF,$F8,$E3,$C8,$FE,$9B,$8F,$3F,$C9,$3F
	.dc.b	$97,$3D,$72,$B8,$71,$FB,$E6,$7F,$E4,$AE,$5E,$B5,$E1,$9C,$FC,$2F
	.dc.b	$F4,$DC,$39,$69,$E0,$FC,$53,$6F,$8B,$DB,$B7,$D3,$67,$E9,$D8,$D1
	.dc.b	$1F,$E2,$FD,$FF,$C9,$FA,$87,$EA,$7F,$D1,$3F,$19,$FD,$C7,$E6,$AF
	.dc.b	$D8,$FF,$D3,$F7,$FE,$17,$FB,$CF,$F7,$35,$EF,$CD,$5E,$FE,$F5,$FC
	.dc.b	$03,$BC,$FB,$5F,$57,$FD,$DC,$78,$71,$F1,$BE,$2F,$8E,$FF,$2E,$BA
	.dc.b	$39,$19,$D3,$CB,$A7,$F8,$96,$33,$7E,$3C,$BD,$1F,$29,$D7,$5F,$3E
	.dc.b	$2C,$78,$E9,$FD,$12,$73,$18,$F8,$5E,$FF,$20,$F9,$47,$C5,$3F,$41
	.dc.b	$7E,$E1,$9E,$9F,$F5,$37,$17,$AB,$0E,$17,$BA,$D7,$EB,$DD,$31,$E7
	.dc.b	$C5,$ED,$C3,$84,$7F,$5D,$8C,$6B,$F5,$5F,$F0,$0C,$78,$B3,$F0,$BF
	.dc.b	$EC,$EA,$E3,$8F,$78,$FF,$6A,$57,$78,$C6,$6F,$D2,$B1,$87,$28,$FE
	.dc.b	$6B,$87,$D2,$F4,$3E,$07,$E8,$6F,$8E,$7D,$2B,$F7,$DF,$F1,$4F,$C3
	.dc.b	$FD,$4B,$DD,$6A,$7F,$C5,$BF,$AA,$67,$76,$3F,$6F,$7C,$7E,$01,$74
	.dc.b	$F0,$56,$B1,$E5,$F4,$DF,$66,$FF,$2D,$9F,$6D,$7C,$5E,$7E,$67,$F9
	.dc.b	$6E,$1C,$67,$CD,$F5,$1F,$EE,$EE,$4F,$95,$87,$1D,$FF,$C8,$3F,$11
	.dc.b	$E5,$DF,$DD,$87,$0F,$D8,$B1,$B7,$AF,$27,$EC,$9F,$92,$FC,$91,$F9
	.dc.b	$E7,$EB,$3F,$53,$7E,$DE,$FB,$67,$E9,$CF,$C8,$38,$5E,$EB,$53,$F8
	.dc.b	$CE,$7F,$D0,$E3,$D3,$E8,$FE,$27,$27,$E0,$63,$BF,$9B,$8E,$7F,$DC
	.dc.b	$46,$35,$F0,$EB,$E0,$9E,$F4,$9C,$FE,$7B,$F1,$EF,$86,$9E,$27,$73
	.dc.b	$CB,$47,$EC,$3F,$B0,$F9,$67,$E5,$1F,$C3,$36,$FE,$BD,$FA,$3B,$F5
	.dc.b	$B7,$ED,$BB,$AD,$7C,$72,$B5,$FA,$66,$7D,$8F,$57,$17,$F7,$4B,$7E
	.dc.b	$4D,$8D,$3B,$D3,$4E,$FE,$1F,$DF,$9C,$C7,$8B,$E2,$31,$87,$83,$3F
	.dc.b	$27,$AB,$DE,$BC,$FE,$D7,$E1,$FC,$37,$E3,$AF,$8B,$7D,$C3,$E0,$9F
	.dc.b	$1C,$FC,$31,$FB,$37,$F1,$2C,$78,$E1,$C2,$F7,$4F,$DE,$B5,$97,$EC
	.dc.b	$B9,$7F,$8F,$F8,$7F,$C7,$B1,$F0,$FC,$F7,$C2,$79,$EB,$E1,$D7,$C5
	.dc.b	$7C,$B4,$ED,$1F,$C0,$F0,$E5,$8F,$A6,$F8,$67,$FA,$E9,$CD,$F0,$87
	.dc.b	$A3,$CE,$B7,$D5,$76,$F4,$FC,$AF,$F2,$EF,$55,$FB,$67,$E2,$7F,$05
	.dc.b	$7E,$27,$FC,$A7,$D7,$BB,$7F,$38,$BD,$F8,$BF,$ED,$8A,$D7,$CA,$73
	.dc.b	$C5,$FA,$D7,$E0,$2F,$7C,$9F,$C7,$8F,$EA,$4F,$2C,$FB,$FC,$17,$F8
	.dc.b	$F2,$79,$65,$D6,$3F,$C7,$71,$89,$DF,$E9,$1F,$C0,$EF,$96,$7E,$7E
	.dc.b	$DF,$E9,$38,$FF,$BA,$BF,$2E,$FC,$67,$C3,$3F,$2A,$7E,$DB,$FE,$EC
	.dc.b	$E1,$FE,$A7,$7B,$F0,$7F,$D4,$F5,$AE,$1E,$7F,$85,$B2,$FC,$DD,$FF
	.dc.b	$BD,$BF,$2C,$B3,$9C,$38,$FC,$0F,$F1,$7A,$E9,$A7,$87,$C9,$C5,$EE
	.dc.b	$3C,$9A,$FE,$4B,$5F,$1C,$E7,$5F,$3F,$CF,$25,$F9,$B7,$F4,$7F,$C4
	.dc.b	$5F,$BF,$BF,$3D,$7E,$02,$FD,$45,$FA,$F3,$F1,$CF,$F7,$1E,$5C,$BB
	.dc.b	$3F,$C0,$99,$72,$C7,$8E,$1C,$2F,$E3,$DE,$ED,$7B,$7F,$AE,$67,$E3
	.dc.b	$CD,$F5,$1F,$CF,$E3,$19,$74,$CD,$F8,$AF,$F8,$0E,$73,$AF,$9E,$5F
	.dc.b	$89,$E7,$2F,$2E,$5E,$DA,$D6,$BF,$C4,$F5,$F2,$6B,$E9,$9C,$FE,$AB
	.dc.b	$F1,$F3,$EB,$BF,$AF,$F3,$5C,$67,$D3,$7F,$26,$6F,$BA,$CC,$BF,$2B
	.dc.b	$7E,$37,$FE,$29,$FC,$B1,$8F,$CB,$F9,$C7,$EC,$1C,$2F,$78,$2F,$79
	.dc.b	$2D,$F8,$96,$F2,$FE,$CD,$77,$F8,$07,$E4,$3F,$E3,$7C,$79,$E3,$CB
	.dc.b	$F1,$AF,$89,$3E,$5B,$F2,$69,$F7,$7E,$87,$38,$70,$8F,$37,$DA,$77
	.dc.b	$BF,$24,$7A,$7C,$1F,$7F,$EE,$1F,$E9,$BF,$C3,$9F,$83,$BF,$4B,$FF
	.dc.b	$6A,$FC,$53,$EC,$B7,$EC,$FF,$BD,$6B,$5F,$84,$FF,$44,$F8,$7C,$E3
	.dc.b	$EA,$D5,$9F,$FA,$9F,$87,$0F,$7C,$38,$FC,$7F,$B2,$31,$AF,$FB,$93
	.dc.b	$E4,$B1,$A7,$78,$F1,$FE,$C1,$DB,$0F,$0E,$FF,$E5,$8E,$FF,$F1,$57
	.dc.b	$DA,$3F,$6D,$FE,$7B,$F1,$7F,$6B,$F2,$57,$EB,$8F,$D6,$BE,$FD,$EF
	.dc.b	$E2,$BF,$CE,$FC,$4B,$DC,$6B,$1F,$E8,$FC,$7F,$A0,$3F,$CD,$97,$DF
	.dc.b	$D0,$F8,$0F,$3C,$FF,$13,$F2,$5F,$CE,$F5,$F2,$E1,$E4,$CF,$B7,$1F
	.dc.b	$DC,$C3,$84,$78,$76,$4F,$87,$4F,$2E,$BE,$AF,$D3,$F1,$A7,$B7,$ED
	.dc.b	$DF,$94,$E3,$2F,$54,$7F,$5D,$F3,$DF,$A9,$7F,$93,$FF,$15,$7E,$FA
	.dc.b	$FC,$A9,$F9,$1F,$F0,$FF,$2B,$F8,$6F,$E8,$9F,$C4,$70,$E1,$7B,$AD
	.dc.b	$7A,$CF,$C5,$75,$D5,$FE,$8F,$CF,$AE,$7F,$E4,$1C,$78,$BD,$B4,$69
	.dc.b	$FD,$7E,$73,$97,$87,$E4,$3F,$D1,$1F,$88,$7F,$99,$7F,$3D,$7E,$4E
	.dc.b	$FC,$15,$F9,$73,$F4,$C7,$EA,$5F,$B1,$FD,$6B,$F9,$46,$3C,$70,$E1
	.dc.b	$7B,$AF,$DB,$78,$9A,$E2,$FF,$53,$67,$DB,$3E,$B9,$F4,$FC,$D7,$C7
	.dc.b	$5A,$CB,$9C,$F8,$3F,$1E,$FF,$2B,$7C,$8F,$E3,$AB,$FC,$42,$FD,$A6
	.dc.b	$FE,$6B,$FC,$03,$93,$FD,$47,$FA,$5F,$FB,$16,$7F,$B3,$1F,$E9,$0F
	.dc.b	$CF,$F5,$F6,$5F,$E8,$5F,$C7,$7F,$A9,$4F,$F8,$B3,$F6,$CF,$F0,$83
	.dc.b	$FE,$10,$9F,$F3,$0F,$ED,$BC,$23,$FD,$81,$FA,$5B,$F1,$3C,$47,$F1
	.dc.b	$AF,$33,$C5,$ED,$FF,$A4,$74,$FE,$33,$3F,$E9,$78,$FF,$4A,$7E,$47
	.dc.b	$FD,$D6,$30,$7F,$B2,$FF,$6D,$7D,$B7,$28,$FF,$67,$7E,$F0,$FC,$6F
	.dc.b	$11,$FE,$AD,$9F,$81,$EF,$FF,$A7,$FE,$95,$D5,$8A,$FF,$6B,$7F,$00
	.dc.b	$3F,$73,$C4,$3F,$57,$7C,$26,$3E,$63,$FB,$24,$FF,$A3,$7F,$3E,$FF
	.dc.b	$55,$7A,$5F,$CB,$23,$FB,$44,$7F,$AB,$7F,$59,$7E,$C1,$F4,$63,$FD
	.dc.b	$61,$F8,$87,$F5,$0A,$E4,$7F,$B9,$FE,$2F,$AB,$FE,$B9,$F3,$7F,$D6
	.dc.b	$DE,$1F,$F8,$52,$30,$7F,$BB,$57,$DD,$7D,$D9,$FF,$5E,$7E,$9B,$BA
	.dc.b	$FF,$22,$C7,$FA,$FF,$E8,$7A,$C7,$F9,$23,$F5,$6F,$F6,$17,$E2,$9E
	.dc.b	$9F,$1D,$1F,$EC,$5F,$0B,$1F,$AB,$FF,$B1,$FB,$B5,$FB,$C7,$FB,$27
	.dc.b	$E1,$6E,$5F,$28,$FC,$77,$DC,$65,$A0,$FC,$A1,$FE,$B5,$7F,$DA,$5F
	.dc.b	$81,$78,$5F,$F6,$A7,$C9,$59,$E1,$EC,$57,$F2,$27,$F9,$82,$7F,$DB
	.dc.b	$5F,$A7,$E7,$B3,$5E,$86,$FF,$08,$E1,$E0,$F1,$FB,$EE,$5F,$92,$F3
	.dc.b	$FC,$C3,$A7,$A7,$D7,$F0,$5E,$DF,$89,$F7,$FD,$0B,$C1,$E3,$F8,$7C
	.dc.b	$CF,$17,$A5,$E3,$F8,$AF,$27,$EF,$1E,$5F,$AD,$F3,$7D,$BF,$9F,$CD
	.dc.b	$F4,$7E,$53,$E9,$F3,$FD,$5F,$B7,$7A,$FF,$38,$F6,$7E,$A7,$F4,$7E
	.dc.b	$4B,$FD,$E9,$F2,$F7,$BB,$E4,$78,$15,$F8,$27,$F9,$DE,$7F,$DF,$5D
	.dc.b	$B9,$EC,$D7,$5E,$FA,$B8,$74,$F1,$E8,$E5,$CF,$CF,$9B,$A7,$2F,$5E
	.dc.b	$4E,$DC,$7D,$F8,$BF,$09,$F2,$3F,$53,$83,$F8,$53,$E0,$69,$F7,$75
	.dc.b	$FD,$16,$9F,$E4,$68,$C1,$FE,$10,$70,$FE,$B0,$C4,$F5,$7F,$DF,$51
	.dc.b	$FD,$31,$FC,$45,$74,$E0,$FC,$54,$7E,$DC,$E8,$E3,$1D,$47,$E3,$F2
	.dc.b	$AF,$F0,$E3,$89,$F9,$3B,$EC,$FF,$0C,$39,$9F,$97,$D7,$8C,$72,$DB
	.dc.b	$9E,$7D,$37,$EB,$C1,$DB,$F0,$7E,$FC,$5E,$0E,$3F,$87,$E0,$FA,$39
	.dc.b	$6E,$BE,$C3,$E2,$47,$B3,$9B,$31,$ED,$3E,$1C,$7C,$0E,$7C,$F0,$EB
	.dc.b	$A3,$7C,$5C,$3A,$78,$F4,$72,$EA,$E7,$D5,$D3,$AF,$AF,$3F,$6E,$CE
	.dc.b	$FD,$1E,$0E,$DF,$0F,$4F,$8B,$BB,$C7,$D5,$E4,$EF,$F2,$F5,$F9,$BC
	.dc.b	$0F,$3F,$67,$A3,$C1,$F4,$F6,$FA,$BC,$2F,$5F,$77,$B3,$C3,$F6,$F7
	.dc.b	$FB,$BC,$4F,$C3,$F8,$BF,$73,$F2,$EB,$E0,$7F,$07,$FC,$0A,$BF,$D7
	.dc.b	$4F,$C2,$F9,$3F,$0D,$FC,$63,$F8,$4F,$C8,$7F,$22,$FF,$1F,$8C,$9F
	.dc.b	$C4,$7F,$0A,$FC,$AF,$9B,$FC,$77,$D3,$FC,$A3,$FC,$89,$FC,$B7,$A5
	.dc.b	$FC,$9D,$B5,$FC,$67,$F4,$EF,$CE,$CF,$F3,$4F,$47,$32,$F5,$7F,$99
	.dc.b	$BF,$9A,$9F,$E7,$3D,$DF,$A1,$F6,$3F,$33,$EB,$FF,$52,$AF,$CF,$C7
	.dc.b	$FA,$B7,$D9,$D4,$7F,$A9,$B1,$AF,$FA,$8B,$DC,$BE,$D7,$F3,$D8,$FD
	.dc.b	$1E,$2C,$3F,$A8,$5F,$0F,$73,$FA,$8D,$7E,$B3,$DE,$FD,$37,$D3,$7E
	.dc.b	$CD,$FD,$5F,$92,$EC,$C5,$7F,$64,$F3,$AB,$FE,$EE,$FB,$FF,$DD,$7C
	.dc.b	$27,$EC,$BF,$1E,$79,$B8,$AF,$EE,$7C,$35,$7B,$9A,$7F,$7B,$7F,$DF
	.dc.b	$BC,$AD,$7C,$37,$EC,$67,$F6,$A3,$8D,$F1,$3F,$23,$D1,$FE,$14,$FF
	.dc.b	$55,$FA,$BB,$5F,$D3,$7F,$1E,$7E,$51,$88,$7C,$6B,$CB,$C7,$A9,$FD
	.dc.b	$C6,$7B,$E7,$D9,$FE,$AD,$F0,$9F,$E0,$B7,$F8,$9E,$7F,$E1,$7F,$B4
	.dc.b	$E9,$E1,$7F,$19,$9E,$33,$FD,$DB,$CD,$CF,$9B,$FC,$7F,$E9,$BF,$8B
	.dc.b	$F8,$FF,$C8,$2F,$F8,$8E,$7F,$ED,$D8,$7D,$D6,$E7,$FB,$34,$3E,$8F
	.dc.b	$FD,$6F,$FA,$FF,$E3,$67,$FB,$57,$E3,$2F,$95,$FF,$29,$77,$FF,$9A
	.dc.b	$BE,$0B,$FC,$31,$F7,$0C,$38,$60,$C3,$87,$CB,$B0,$E1,$E2,$BF,$03
	.dc.b	$E1,$C9,$FC,$BB,$DD,$7C,$43,$F0,$D7,$EF,$0F,$D9,$7F,$DE,$BF,$80
	.dc.b	$BF,$74,$E3,$BF,$F7,$FC,$78,$E1,$C2,$F7,$F9,$4E,$A9,$FA,$43,$F6
	.dc.b	$5F,$17,$E3,$0F,$CA,$DF,$B0,$3D,$67,$E2,$7C,$70,$E1,$C6,$C3,$85
	.dc.b	$EC,$ED,$C2,$D1,$F1,$8F,$C7,$9F,$B9,$7E,$3F,$F2,$6F,$CB,$F6,$7C
	.dc.b	$8C,$38,$5E,$EB,$5E,$B7,$A7,$CD,$F0,$7C,$23,$F3,$B7,$E2,$CF,$BB
	.dc.b	$7E,$EA,$FB,$EE,$1C,$4C,$38,$5E,$EB,$FD,$85,$B7,$D8,$FC,$35,$F9
	.dc.b	$5B,$E0,$3E,$E6,$8F,$CA,$9F,$A9,$7E,$77,$0E,$17,$B7,$C9,$3A,$F7
	.dc.b	$7E,$87,$E7,$9E,$6F,$2D,$F8,$D3,$F5,$87,$E4,$1F,$E9,$EF,$C5,$F7
	.dc.b	$C3,$F8,$9A,$D4,$FF,$A8,$BE,$F7,$ED,$7E,$B7,$FE,$F7,$FA,$8F,$D7
	.dc.b	$7E,$67,$E8,$E6,$7D,$C6,$BE,$25,$5A,$E3,$F8,$5F,$9B,$7E,$0F,$F1
	.dc.b	$0F,$0F,$C8,$BF,$2F,$FF,$38,$FE,$50,$FC,$CB,$E0,$74,$FE,$72,$F7
	.dc.b	$5A,$9E,$13,$FA,$0F,$BE,$7E,$88,$FD,$31,$FA,$6F,$E2,$FE,$E7,$EB
	.dc.b	$7E,$21,$F9,$3B,$DD,$6B,$D4,$57,$DB,$FD,$43,$F1,$7C,$1F,$E5,$BF
	.dc.b	$CC,$3F,$CF,$BD,$FF,$A3,$FC,$4F,$3E,$D2,$73,$ED,$BF,$18,$FD,$07
	.dc.b	$B1,$FB,$1F,$63,$F7,$DF,$10,$F0,$1F,$A2,$3F,$21,$DF,$C3,$FE,$8B
	.dc.b	$E2,$FF,$80,$56,$BF,$74,$FD,$57,$E6,$CF,$DD,$1F,$52,$FD,$25,$F4
	.dc.b	$CF,$DE,$BD,$7F,$B5,$70,$BD,$D6,$A7,$DF,$FA,$0F,$C3,$D5,$F8,$4B
	.dc.b	$F6,$87,$ED,$DF,$2B,$EC,$1F,$95,$EE,$B5,$39,$FC,$47,$F7,$DF,$86
	.dc.b	$BF,$2E,$DF,$B5,$EE,$C3,$F9,$EF,$E5,$5A,$9C,$F4,$F4,$FA,$47,$EE
	.dc.b	$9F,$C1,$FD,$D7,$E8,$6F,$CF,$5F,$B6,$3F,$7E,$F8,$1C,$5F,$A1,$F9
	.dc.b	$FF,$F2,$3E,$1C,$2F,$75,$AB,$FE,$25,$F9,$77,$F1,$9F,$A7,$FE,$35
	.dc.b	$CC,$FC,$83,$F8,$FF,$EE,$5A,$FA,$1F,$EA,$87,$C3,$8F,$E4,$7F,$9E
	.dc.b	$FF,$1B,$FC,$5D,$FA,$33,$F4,$D7,$EA,$0F,$D6,$DF,$A5,$3F,$48,$78
	.dc.b	$FC,$70,$E1,$7B,$AD,$78,$9F,$C9,$FF,$2C,$FB,$7F,$B5,$7F,$1F,$E6
	.dc.b	$FE,$BB,$FB,$33,$CD,$D7,$98,$AD,$7C,$FA,$9F,$8D,$7F,$31,$FA,$B7
	.dc.b	$CB,$FE,$23,$EA,$DF,$84,$3F,$7A,$7E,$9F,$F8,$CF,$8C,$FE,$97,$87
	.dc.b	$0F,$5B,$F9,$B5,$EE,$BF,$D9,$35,$FD,$43,$F6,$57,$E9,$73,$F5,$1F
	.dc.b	$EA,$9F,$26,$FA,$47,$EF,$5F,$D0,$EF,$76,$BD,$D7,$FB,$AB,$0F,$EB
	.dc.b	$7F,$89,$BF,$00,$FC,$07,$E1,$CF,$DC,$DF,$B8,$BF,$23,$FE,$0F,$E8
	.dc.b	$FF,$91,$6F,$75,$A9,$FE,$0F,$3F,$E8,$CF,$95,$FA,$BF,$8A,$3F,$2B
	.dc.b	$FE,$25,$EF,$FF,$31,$DB,$F8,$37,$88,$AD,$7C,$E2,$BF,$9B,$7B,$7F
	.dc.b	$D7,$65,$FB,$8B,$E5,$7F,$AD,$F1,$3D,$AF,$DA,$7E,$89,$BA,$D7,$C7
	.dc.b	$AB,$5F,$11,$FC,$0F,$9F,$FB,$DF,$E3,$AF,$D1,$1C,$6F,$CE,$5F,$97
	.dc.b	$7F,$8E,$FF,$1C,$7E,$D1,$FD,$BF,$0E,$17,$BA,$D7,$F0,$07,$F7,$3F
	.dc.b	$C6,$1F,$A9,$FF,$5B,$FC,$6D,$FA,$B7,$F0,$AF,$D3,$1F,$BC,$EE,$B5
	.dc.b	$39,$FC,$B7,$FB,$E7,$81,$FE,$63,$3F,$D5,$FE,$6F,$FC,$01,$E3,$FF
	.dc.b	$09,$C6,$7E,$8B,$FD,$67,$F9,$FF,$08,$7E,$75,$FE,$42,$9F,$F0,$37
	.dc.b	$E9,$FF,$D0,$7F,$5A,$7E,$67,$FD,$AF,$E7,$3F,$9E,$FD,$96,$BE,$C9
	.dc.b	$5A,$FC,$17,$FC,$3D,$F6,$6C,$7F,$C2,$1F,$8C,$75,$FC,$DB,$FC,$F1
	.dc.b	$C1,$FE,$26,$8C,$70,$6B,$6F,$F2,$9D,$7A,$78,$3F,$C7,$71,$FE,$56
	.dc.b	$FC,$E3,$8C,$62,$3F,$D9,$B0,$AE,$51,$82,$9F,$ED,$A7,$8B,$FE,$40
	.dc.b	$3F,$CC,$2F,$1C,$3C,$91,$83,$89,$FE,$E2,$AF,$90,$7C,$F3,$E4,$CE
	.dc.b	$1F,$A5,$C3,$E9,$91,$B8,$FA,$DF,$67,$E2,$8C,$47,$FA,$6F,$F1,$DF
	.dc.b	$13,$DC,$E0,$F3,$8D,$C7,$F4,$43,$C0,$7F,$A8,$FF,$2C,$E3,$1E,$08
	.dc.b	$EB,$1C,$5E,$FF,$A2,$F8,$5F,$0F,$E7,$5E,$3D,$BF,$05,$F9,$77,$2C
	.dc.b	$3F,$D5,$A7,$FB,$E9,$F1,$47,$E1,$FF,$30,$E5,$7E,$DF,$13,$83,$EC
	.dc.b	$7C,$0F,$C1,$FD,$1F,$6F,$C9,$FD,$1F,$E1,$7E,$6D,$FB,$8E,$97,$E8
	.dc.b	$8F,$D0,$BF,$C6,$5F,$3C,$FC,$B7,$FC,$B5,$7B,$AD,$7E,$25,$FE,$63
	.dc.b	$F3,$3E,$FF,$F0,$07,$D5,$EA,$FB,$7C,$7F,$4B,$CA,$FC,$14,$62,$3F
	.dc.b	$D8,$87,$B0,$FA,$E7,$EE,$B7,$FD,$64,$6F,$E5,$38,$C7,$E4,$67,$FB
	.dc.b	$F9,$C2,$7A,$7E,$DB,$E1,$9D,$7E,$6D,$C2,$BF,$20,$7E,$11,$EE,$FE
	.dc.b	$5B,$E1,$38,$3F,$66,$7F,$79,$F4,$DC,$63,$F1,$4F,$0C,$3F,$7B,$3E
	.dc.b	$62,$4F,$17,$FD,$CD,$D1,$C5,$FC,$C3,$C5,$FC,$71,$FE,$EB,$D3,$C3
	.dc.b	$3F,$97,$CF,$84,$FF,$0D,$78,$47,$58,$67,$C1,$FD,$09,$CA,$FD,$AD
	.dc.b	$9F,$F7,$49,$FA,$4F,$03,$F6,$B9,$74,$8F,$C7,$9F,$EA,$49,$FC,$3F
	.dc.b	$81,$F9,$2A,$FD,$8B,$AC,$FE,$0C,$FF,$1A,$CF,$C4,$8F,$E3,$B7,$EA
	.dc.b	$7F,$D3,$87,$8D,$EF,$7F,$C2,$27,$F6,$D5,$F8,$8D,$BF,$93,$B6,$BC
	.dc.b	$D9,$7E,$1B,$D3,$DE,$3F,$C4,$F5,$DB,$6B,$67,$F0,$2D,$76,$FE,$0B
	.dc.b	$7D,$B4,$D9,$FD,$B6,$7D,$67,$FB,$D5,$F5,$6D,$FE,$F8,$CB,$E0,$9D
	.dc.b	$23,$EE,$9E,$FC,$B5,$5F,$7B,$4D,$3F,$6E,$7E,$26,$5A,$7F,$11,$96
	.dc.b	$A7,$FA,$29,$D6,$FE,$C4,$FA,$0F,$E0,$B1,$97,$F8,$FE,$39,$9F,$BF
	.dc.b	$1F,$A5,$9F,$E7,$18,$F5,$DB,$E8,$BF,$9A,$7F,$03,$3F,$D5,$56,$5B
	.dc.b	$8F,$05,$7D,$DF,$07,$F9,$4D,$FF,$BC,$3E,$71,$FA,$E3,$F6,$BF,$F6
	.dc.b	$97,$CF,$BE,$D1,$F9,$9F,$F9,$1F,$0E,$17,$BF,$C8,$BF,$C9,$75,$AF
	.dc.b	$BA,$FE,$A9,$C1,$FD,$1A,$FF,$A2,$BF,$D9,$DF,$E8,$FB,$F8,$B6,$FE
	.dc.b	$AF,$CD,$B6,$F0,$7C,$D1,$7F,$D1,$39,$B8,$47,$F5,$3E,$5E,$0D,$38
	.dc.b	$4F,$E2,$6B,$E7,$73,$6A,$79,$87,$C9,$1F,$DA,$3A,$3A,$CF,$F0,$A3
	.dc.b	$E4,$C7,$E2,$DF,$D8,$3F,$DB,$0F,$D4,$D7,$57,$F8,$4C,$7A,$5F,$E0
	.dc.b	$C7,$E1,$6F,$4F,$F2,$28,$E6,$FF,$04,$3F,$14,$E8,$F2,$9F,$D5,$79
	.dc.b	$7F,$AD,$E9,$FE,$03,$E8,$FF,$09,$1F,$E0,$83,$FC,$29,$3F,$E0,$5A
	.dc.b	$F5,$E1,$F8,$E8,$FC,$71,$EB,$3F,$1C,$7F,$7D,$AF,$83,$3F,$E1,$E9
	.dc.b	$F6,$5F,$E3,$AB,$FC,$52,$FE,$3A,$7F,$4A,$F8,$AB,$FC,$52,$F8,$9F
	.dc.b	$F1,$5F,$37,$A2,$7F,$A3,$DF,$3C,$3F,$C3,$EE,$E3,$3B,$6F,$1F,$F0
	.dc.b	$1C,$FF,$50,$7F,$9E,$35,$F4,$9F,$86,$D7,$83,$F8,$8C,$FC,$97,$E0
	.dc.b	$7B,$47,$4E,$8D,$DF,$E0,$E3,$F0,$CD,$7F,$17,$AF,$4B,$F9,$11,$FC
	.dc.b	$CF,$4F,$F2,$3F,$47,$48,$FC,$24,$FF,$3D,$BF,$0C,$7F,$83,$5D,$47
	.dc.b	$C9,$3F,$86,$E3,$A9,$FE,$64,$EB,$1F,$04,$66,$7E,$EA,$C3,$F9,$C3
	.dc.b	$FE,$28,$DB,$FC,$3F,$39,$3F,$BA,$FC,$3F,$E1,$36,$C5,$7F,$8B,$A3
	.dc.b	$9D,$FF,$93,$6B,$E1,$E7,$E3,$CF,$DF,$7F,$E1,$AA,$F9,$B8,$FD,$97
	.dc.b	$C3,$1F,$87,$3F,$2E,$7F,$4D,$BF,$D5,$CF,$86,$76,$F8,$23,$FC,$13
	.dc.b	$BF,$A2,$79,$FA,$3F,$EE,$AF,$B1,$87,$FB,$B3,$E5,$FD,$38,$A7,$33
	.dc.b	$AA,$F1,$FA,$0F,$DD,$E1,$07,$E9,$BF,$E7,$87,$06,$A7,$D7,$F2,$3E
	.dc.b	$7D,$3F,$C3,$AF,$9F,$2F,$F2,$AF,$EE,$DE,$79,$FF,$10,$FE,$5D,$E2
	.dc.b	$CF,$CD,$87,$C4,$8C,$40,$F4,$FE,$09,$E0,$CB,$CD,$9E,$1C,$1E,$EF
	.dc.b	$CB,$FC,$59,$FF,$99,$1F,$12,$FD,$37,$C4,$FF,$9A,$3F,$0D,$EF,$C3
	.dc.b	$F1,$9F,$0E,$5F,$23,$F4,$4E,$74,$79,$F8,$7C,$A3,$FD,$6B,$3E,$BB
	.dc.b	$F9,$FF,$11,$CB,$7F,$BF,$F6,$FE,$6D,$BE,$9B,$E8,$DB,$EA,$FE,$77
	.dc.b	$E2,$AF,$C2,$BC,$78,$BF,$C8,$FF,$95,$F1,$9F,$C4,$BC,$5F,$C5,$FE
	.dc.b	$61,$E6,$3D,$FF,$4D,$E5,$F2,$78,$5F,$F9,$41,$E0,$FF,$94,$BF,$46
	.dc.b	$F6,$9F,$1B,$EC,$7D,$9F,$82,$78,$E7,$FC,$AE,$F8,$F8,$BF,$D1,$FF
	.dc.b	$8E,$F4,$6E,$DF,$4B,$4F,$CE,$FE,$33,$F2,$0F,$A3,$F3,$7F,$1F,$E8
	.dc.b	$7A,$5F,$FA,$61,$E8,$BC,$6E,$8D,$C3,$A6,$7F,$A0,$7F,$05,$C9,$F9
	.dc.b	$CF,$0F,$D7,$5F,$EA,$2F,$84,$EA,$9E,$BF,$47,$F6,$8F,$C1,$FC,$27
	.dc.b	$D9,$FD,$5F,$67,$8B,$F3,$EF,$BE,$7E,$23,$E0,$BE,$F7,$C0,$F9,$0E
	.dc.b	$F8,$BF,$41,$FC,$29,$F8,$EF,$B1,$FC,$27,$B1,$EA,$37,$F9,$8F,$90
	.dc.b	$F7,$FB,$DE,$E3,$9F,$E4,$DE,$83,$E1,$FE,$67,$F9,$03,$F3,$3F,$9D
	.dc.b	$7E,$3F,$F1,$2F,$69,$F2,$3E,$77,$F2,$A7,$E7,$BF,$10,$FC,$A7,$F0
	.dc.b	$0F,$CA,$3C,$FF,$A3,$7E,$68,$FD,$1F,$EB,$3F,$99,$F8,$EF,$9E,$7B
	.dc.b	$3F,$79,$F9,$27,$D4,$FA,$1F,$A6,$7C,$0F,$C9,$FE,$69,$F6,$3F,$32
	.dc.b	$FD,$29,$FA,$EF,$DB,$BF,$49,$F8,$0F,$D2,$3E,$EF,$CF,$FD,$C3,$E4
	.dc.b	$FD,$0F,$EB,$0F,$DA,$FE,$AB,$FA,$BF,$C1,$3F,$02,$7C,$EF,$97,$FB
	.dc.b	$67,$E1,$BD,$A7,$4B,$F0,$57,$EE,$0D,$5E,$BB,$E6,$DF,$8C,$FE,$EF
	.dc.b	$EF,$FF,$8F,$7D,$77,$F8,$6F,$F4,$17,$DD,$BF,$11,$FE,$4F,$DB,$D7
	.dc.b	$F7,$C7,$2B,$F6,$97,$EE,$9F,$C0,$FA,$BF,$C9,$5E,$57,$FA,$F3,$EB
	.dc.b	$FF,$DA,$7F,$55,$FC,$F3,$E5,$3F,$DE,$BF,$57,$FC,$FB,$E4,$DF,$90
	.dc.b	$BF,$58,$7E,$3A,$FC,$33,$FD,$19,$DA,$F1,$FF,$9C,$FC,$2B,$9D,$F8
	.dc.b	$0B,$EB,$7F,$25,$F6,$2F,$C7,$1F,$99,$7F,$40,$FA,$CF,$F6,$F7,$D0
	.dc.b	$39,$BC,$1F,$86,$BF,$07,$7E,$63,$FD,$1B,$F5,$D7,$EA,$AF,$01,$F7
	.dc.b	$EF,$D6,$FC,$FF,$B2,$8F,$E9,$BE,$FF,$F9,$AB,$F4,$77,$EF,$2F,$C7
	.dc.b	$9F,$46,$FC,$AF,$E0,$7C,$C7,$FA,$B3,$E5,$3F,$D5,$43,$F5,$3F,$34
	.dc.b	$7F,$66,$FC,$49,$FA,$1B,$F2,$E7,$EE,$6F,$DE,$5F,$90,$3F,$56,$7E
	.dc.b	$D7,$FD,$83,$F7,$67,$EA,$6F,$DB,$3E,$D7,$EA,$EF,$DA,$7F,$03,$E8
	.dc.b	$7F,$66,$F9,$C7,$E5,$8F,$C2,$1F,$AC,$BF,$4A,$FF,$36,$FC,$F7,$C8
	.dc.b	$E6,$FE,$F7,$E9,$BF,$30,$FB,$9F,$B2,$3F,$36,$7E,$46,$FC,$E1,$FB
	.dc.b	$9F,$F7,$4F,$C8,$9F,$B9,$BF,$17,$7D,$C3,$ED,$5F,$9E,$BE,$EB,$97
	.dc.b	$07,$F8,$2B,$E6,$1F,$A0,$BF,$0E,$7E,$8E,$FD,$15,$D6,$FD,$A5,$F9
	.dc.b	$B7,$F8,$13,$E7,$DE,$EB,$EE,$DF,$42,$F5,$FE,$7F,$C0,$BF,$72,$FF
	.dc.b	$05,$FC,$43,$ED,$DF,$6A,$FC,$29,$F6,$2F,$CF,$DF,$8D,$3F,$40,$FF
	.dc.b	$09,$FC,$97,$B7,$F3,$BF,$47,$E0,$BF,$9C,$FC,$8D,$FB,$2F,$F8,$63
	.dc.b	$F7,$AF,$DD,$F4,$FE,$D7,$F5,$C7,$CA,$F8,$FE,$2B,$FC,$3B,$FA,$6F
	.dc.b	$8F,$F6,$0F,$E0,$5F,$96,$3F,$68,$FF,$20,$77,$7A,$3F,$65,$F9,$FF
	.dc.b	$80,$3F,$13,$7B,$FF,$A0,$FC,$23,$FC,$4F,$F9,$F7,$CC,$FC,$5B,$E2
	.dc.b	$FE,$E2,$FB,$B7,$ED,$0F,$DF,$FF,$C5,$BF,$8F,$3F,$21,$7E,$EA,$F8
	.dc.b	$C7,$DC,$BF,$7C,$7E,$95,$F6,$FD,$8F,$EF,$7F,$BA,$BF,$2F,$7D,$03
	.dc.b	$F3,$3F,$DE,$7E,$FC,$FD,$73,$FC,$77,$F8,$17,$E0,$FF,$6A,$7E,$D0
	.dc.b	$FD,$E9,$FA,$CF,$F8,$7F,$F7,$2F,$F2,$0F,$E1,$EF,$C5,$DF,$A8,$3F
	.dc.b	$42,$7E,$3A,$FC,$EB,$F4,$9F,$5F,$F3,$3F,$15,$7A,$8F,$37,$F8,$17
	.dc.b	$EC,$3F,$8B,$FE,$01,$FF,$26,$FE,$3F,$FE,$D0,$F4,$3F,$8B,$7B,$DC
	.dc.b	$3F,$7F,$7E,$AF,$FE,$52,$FC,$C1,$FB,$B3,$F2,$C7,$EE,$8F,$CE,$3F
	.dc.b	$23,$FA,$5F,$ED,$FC,$5F,$F3,$3F,$E2,$4F,$C5,$BF,$DB,$BF,$90,$3D
	.dc.b	$3F,$E6,$9F,$B7,$FF,$2C,$C7,$75,$A9,$CF,$D7,$FF,$B1,$FF,$07,$FE
	.dc.b	$45,$BF,$8B,$6F,$0E,$1E,$0F,$B2,$D5,$39,$7D,$9F,$86,$7E,$73,$A9
	.dc.b	$CB,$EC,$C3,$FC,$F7,$F0,$4E,$CF,$8F,$D9,$F1,$BE,$69,$D9,$D3,$C4
	.dc.b	$FA,$5E,$1F,$D1,$F0,$BE,$87,$D9,$FF,$28,$CF,$FD,$0B,$E5,$FD,$DF
	.dc.b	$2B,$EE,$79,$3F,$6E,$7F,$0F,$E8,$7E,$1B,$CF,$FC,$2F,$9D,$E8,$F6
	.dc.b	$3F,$25,$EB,$FE,$47,$D6,$FC,$87,$AB,$FC,$27,$1F,$F4,$8F,$BB,$F9
	.dc.b	$DF,$9A,$F9,$5E,$EF,$E7,$3D,$CF,$83,$F2,$AF,$84,$FD,$3F,$C1,$FE
	.dc.b	$9B,$F0,$5F,$F1,$66,$5F,$03,$E5,$BF,$09,$F8,$F3,$D4,$FD,$87,$9E
	.dc.b	$F8,$8F,$97,$F1,$DF,$BA,$FC,$69,$F1,$DF,$BA,$F9,$9F,$E9,$59,$FF
	.dc.b	$A9,$3E,$53,$F8,$27,$D0,$BE,$53,$F8,$27,$CA,$79,$37,$FE,$1F,$F3
	.dc.b	$3F,$C3,$BE,$67,$F8,$6F,$CC,$7D,$8D,$FF,$8D,$FC,$F7,$F1,$AF,$9D
	.dc.b	$FE,$33,$B7,$DC,$FA,$3F,$E4,$FA,$FF,$26,$FA,$1F,$E4,$BE,$77,$F2
	.dc.b	$4E,$1F,$F5,$AF,$D5,$7F,$34,$FA,$9F,$E6,$7F,$51,$FC,$CB,$E9,$FF
	.dc.b	$9D,$57,$F4,$0F,$B0,$FE,$7F,$F5,$FF,$CF,$BE,$BB,$F9,$EC,$FF,$AF
	.dc.b	$BE,$8B,$FA,$5F,$DA,$FF,$4A,$FB,$4F,$E9,$3F,$84,$FE,$3B,$E7,$3F
	.dc.b	$AA,$FC,$1F,$F5,$4F,$5F,$FA,$9E,$3F,$D8,$7E,$FF,$FB,$07,$DF,$7F
	.dc.b	$5F,$FB,$DF,$EB,$DF,$79,$F9,$6F,$C0,$BF,$B4,$9F,$DA,$BF,$02,$FE
	.dc.b	$FD,$87,$FB,$3B,$B3,$FB,$A7,$CC,$7F,$73,$FC,$13,$FB,$97,$E2,$5F
	.dc.b	$9E,$FC,$2B,$FB,$FF,$E1,$5F,$DF,$BF,$09,$FE,$FB,$F5,$FF,$A2,$FC
	.dc.b	$33,$FC,$13,$F2,$3F,$E0,$8F,$C3,$3F,$C0,$FA,$FF,$B6,$BF,$10,$FF
	.dc.b	$0A,$7B,$5F,$E1,$3F,$CA,$7F,$9A,$FE,$21,$FE,$12,$AF,$F6,$6F,$CC
	.dc.b	$7F,$87,$FF,$13,$FF,$0F,$72,$FF,$8E,$BD,$5F,$52,$FC,$5F,$FC,$2D
	.dc.b	$AC,$33,$F9,$D5,$7D,$B3,$A0,$23,$5F,$63,$C2,$7F,$C6,$87,$EF,$3E
	.dc.b	$7B,$FC,$79,$F8,$E7,$F8,$EF,$F1,$BF,$F1,$D6,$BF,$EF,$4F,$B5,$7E
	.dc.b	$3D,$FE,$46,$FC,$FB,$FC,$61,$F8,$FF,$F9,$16,$BF,$B8,$7E,$19,$FE
	.dc.b	$51,$FC,$8B,$FC,$A1,$F9,$17,$F9,$3F,$E6,$7F,$84,$FA,$9F,$EB,$DB
	.dc.b	$FF,$7F,$BF,$EC,$17,$FD,$BD,$F9,$BF,$FB,$79,$AF,$FB,$D3,$E3,$3F
	.dc.b	$88,$FE,$0D,$C7,$97,$E1,$BC,$5F,$8B,$F5,$9F,$ED,$6B,$FE,$4B,$7F
	.dc.b	$ED,$17,$F5,$5F,$98,$7F,$B9,$1A,$FF,$BE,$3F,$30,$F8,$57,$F8,$83
	.dc.b	$FC,$63,$3F,$2B,$9B,$FC,$8D,$F9,$67,$CF,$7E,$9C,$65,$46,$5F,$EC
	.dc.b	$11,$F8,$FC,$3D,$95,$F6,$BC,$BF,$F7,$6C,$F1,$FC,$BB,$D9,$7F,$E5
	.dc.b	$37,$D8,$FF,$95,$7F,$09,$F2,$5F,$90,$FC,$1F,$D1,$77,$9F,$E6,$AF
	.dc.b	$87,$B3,$F9,$BF,$E0,$7F,$0E,$BF,$12,$F8,$1F,$C5,$FC,$3E,$6F,$FD
	.dc.b	$A1,$A7,$C7,$9F,$C7,$FD,$77,$81,$FC,$5E,$9E,$FF,$8F,$F0,$36,$FE
	.dc.b	$26,$5F,$D1,$B9,$FD,$71,$FE,$90,$E2,$FF,$6A,$D7,$C7,$FC,$DF,$C1
	.dc.b	$27,$C8,$FF,$52,$C7,$C1,$3F,$37,$F3,$8E,$F1,$FA,$27,$DD,$1F,$47
	.dc.b	$F2,$4E,$E7,$C8,$AF,$E5,$DA,$FD,$07,$FC,$FD,$F1,$BD,$9F,$F0,$C6
	.dc.b	$1D,$8F,$D6,$FE,$57,$DA,$3F,$D0,$8F,$68,$FD,$8F,$89,$EC,$3B,$7D
	.dc.b	$B7,$A8,$E9,$F5,$7F,$1E,$7F,$B3,$FC,$DF,$3C,$3F,$D6,$47,$C6,$D3
	.dc.b	$F1,$1F,$5D,$F1,$8F,$47,$E5,$9F,$10,$F3,$7E,$5B,$CE,$7F,$1D,$F9
	.dc.b	$77,$33,$C9,$7F,$E9,$9F,$CB,$FE,$79,$F0,$7F,$30,$F9,$A7,$BF,$F0
	.dc.b	$0E,$5A,$7F,$0A,$FC,$3B,$C8,$72,$BF,$F1,$17,$3F,$E9,$67,$F8,$7F
	.dc.b	$E0,$BF,$2F,$CB,$FA,$E7,$C9,$FB,$BE,$77,$FE,$A9,$F6,$7D,$C7,$3F
	.dc.b	$C7,$7F,$02,$7D,$3F,$91,$FB,$E7,$D1,$FC,$C7,$8D,$7F,$99,$CF,$D4
	.dc.b	$70,$FE,$AF,$E8,$7C,$9F,$99,$7E,$28,$FB,$9F,$99,$FE,$30,$FB,$DF
	.dc.b	$9A,$72,$FC,$D7,$E7,$1E,$BF,$83,$FC,$69,$F7,$FF,$36,$FC,$99,$F8
	.dc.b	$5F,$C7,$FC,$3F,$3B,$F5,$4F,$85,$F9,$BF,$E5,$0F,$C3,$7E,$71,$F9
	.dc.b	$A3,$F1,$9F,$4B,$E8,$F8,$0F,$B8,$7C,$9F,$91,$FC,$D9,$F8,$DF,$A9
	.dc.b	$FD,$09,$F9,$3F,$CE,$BD,$FF,$75,$F8,$33,$E8,$7E,$77,$FA,$23,$F2
	.dc.b	$9C,$5F,$13,$E2,$3F,$10,$7D,$51,$F9,$F7,$E5,$F9,$FF,$DB,$10,$FC
	.dc.b	$DF,$E4,$F8,$1F,$D1,$A3,$13,$FD,$03,$A7,$F8,$B4,$7D,$09,$FC,$A1
	.dc.b	$F8,$13,$A9,$FC,$3A,$3F,$8E,$3C,$A7,$FA,$74,$62,$BF,$35,$3F,$D5
	.dc.b	$1F,$F2,$B3,$C4,$C7,$E7,$BE,$4A,$FF,$77,$4A,$38,$C6,$30,$F6,$3E
	.dc.b	$B9,$E0,$D3,$D5,$5D,$FE,$D3,$72,$FC,$8B,$6D,$AD,$9E,$0A,$EB,$D7
	.dc.b	$FB,$33,$E0,$65,$E0,$8F,$24,$FF,$3D,$8F,$83,$97,$9A,$3F,$B5,$E3
	.dc.b	$BC,$7C,$B7,$DE,$7F,$BA,$7B,$1F,$E6,$6C,$9F,$9F,$79,$BE,$2F,$CA
	.dc.b	$73,$8F,$F6,$A5,$FC,$7B,$F8,$BF,$9F,$FA,$8F,$04,$38,$3C,$DC,$1E
	.dc.b	$03,$27,$D8,$8F,$F0,$FE,$DF,$62,$3E,$C5,$FF,$56,$9F,$F0,$24,$62
	.dc.b	$3F,$AA,$C6,$0F,$13,$FE,$08,$7D,$A7,$88,$EC,$7F,$51,$46,$F0,$F9
	.dc.b	$11,$FE,$11,$3B,$67,$F0,$E7,$F0,$B1,$E6,$3F,$10,$FD,$03,$FC,$3B
	.dc.b	$1B,$E0,$F1,$D7,$E2,$27,$E6,$E1,$FD,$BE,$3C,$3B,$6A,$7E,$91,$EF
	.dc.b	$38,$6D,$F8,$27,$F4,$78,$F4,$C7,$D9,$87,$C0,$8F,$F1,$3E,$FE,$AA
	.dc.b	$F6,$C6,$1F,$0B,$ED,$C3,$83,$E0,$8F,$F1,$89,$C1,$C5,$FC,$1D,$FC
	.dc.b	$DB,$F8,$F3,$F9,$3F,$61,$E9,$E0,$F0,$63,$FB,$87,$D1,$E4,$3A,$5F
	.dc.b	$A1,$3B,$DF,$92,$BF,$41,$7E,$84,$F8,$7F,$F5,$BE,$3C,$70,$E1,$7B
	.dc.b	$A7,$EB,$1F,$E4,$EE,$9F,$5F,$4F,$DE,$E4,$FF,$26,$F0,$FF,$00,$7E
	.dc.b	$E4,$7E,$33,$93,$FC,$A3,$1F,$E2,$7E,$7F,$E0,$EF,$F8,$8F,$8B,$F0
	.dc.b	$B3,$FE,$5A,$8E,$93,$A7,$F2,$4D,$BF,$9B,$5F,$7C,$FC,$3C,$FF,$84
	.dc.b	$0F,$C3,$BF,$87,$3F,$32,$6F,$A7,$C5,$1A,$9E,$73,$FC,$67,$9F,$2F
	.dc.b	$5E,$9F,$63,$4F,$C8,$E1,$F8,$D7,$ED,$5F,$E7,$DF,$E5,$78,$FE,$2E
	.dc.b	$3F,$C6,$46,$4F,$D4,$74,$F4,$9F,$F4,$8B,$E2,$8F,$D6,$61,$E2,$9F
	.dc.b	$F4,$AE,$1B,$9F,$D8,$3E,$8C,$7F,$CF,$FD,$3B,$7F,$D0,$3D,$39,$BF
	.dc.b	$C4,$DF,$EA,$DF,$91,$D1,$E1,$AF,$F4,$3B,$E1,$9F,$F4,$1E,$FE,$17
	.dc.b	$F7,$DB,$E5,$FF,$43,$6F,$CF,$0F,$F5,$4C,$FF,$0A,$DF,$EC,$3F,$D0
	.dc.b	$F7,$FB,$8F,$3B,$FE,$11,$7C,$5F,$F5,$8E,$5F,$E9,$7B,$FC,$04,$71
	.dc.b	$7E,$76,$1F,$E9,$88,$FB,$CF,$E0,$4F,$E1,$F1,$A5,$B7,$96,$BF,$D4
	.dc.b	$5B,$75,$BF,$C2,$D7,$F3,$3C,$BF,$88,$3F,$6A,$3F,$24,$7E,$33,$0F
	.dc.b	$1E,$FC,$67,$4F,$E0,$A7,$F8,$D1,$FD,$00,$FB,$C7,$D6,$8F,$E1,$C7
	.dc.b	$FA,$1E,$F8,$3E,$9D,$FC,$CF,$FA,$E3,$A6,$BF,$CA,$37,$F0,$3F,$CD
	.dc.b	$9E,$2B,$F1,$B7,$E2,$EF,$99,$7E,$16,$F3,$DF,$62,$FD,$11,$F4,$5F
	.dc.b	$DC,$B2,$FE,$6B,$97,$2C,$78,$E1,$C2,$DF,$09,$F1,$38,$3F,$D8,$FF
	.dc.b	$1F,$F4,$B1,$FF,$40,$7E,$85,$67,$FD,$23,$F9,$F7,$F1,$E9,$FF,$42
	.dc.b	$3F,$B2,$7F,$D0,$DE,$AE,$E7,$FD,$55,$F3,$35,$FF,$79,$FC,$FF,$17
	.dc.b	$FD,$36,$F1,$EA,$FF,$44,$C7,$D8,$BF,$F6,$FF,$E5,$5E,$39,$FF,$58
	.dc.b	$7E,$8B,$67,$FD,$C1,$F7,$5E,$39,$FF,$5A,$3E,$3B,$FF,$53,$FE,$21
	.dc.b	$E1,$7F,$AE,$7C,$B7,$85,$FF,$5F,$3B,$7F,$D8,$1E,$1E,$DF,$F7,$4F
	.dc.b	$87,$B7,$F4,$7E,$0F,$97,$2F,$F7,$77,$83,$F9,$57,$FB,$1F,$E8,$3E
	.dc.b	$57,$FD,$ED,$B7,$95,$FF,$7B,$3F,$1F,$E4,$7E,$E7,$C7,$BE,$41,$FA
	.dc.b	$33,$F4,$77,$E3,$6B,$FB,$98,$DE,$E6,$3C,$70,$E1,$6F,$5B,$F9,$85
	.dc.b	$FF,$BF,$B5,$FF,$6B,$7C,$EB,$F4,$97,$E2,$AF,$21,$F8,$67,$F8,$D7
	.dc.b	$F4,$27,$E9,$3E,$0C,$38,$5E,$EB,$5F,$2D,$E4,$F0,$9F,$93,$FE,$4B
	.dc.b	$F4,$BF,$F5,$E7,$D1,$BF,$18,$7E,$97,$BA,$D7,$D0,$AB,$5E,$9F,$BB
	.dc.b	$ED,$3F,$43,$FA,$7B,$F0,$A7,$E9,$CF,$93,$7E,$9C,$FB,$7E,$1E,$D3
	.dc.b	$0E,$17,$BA,$7E,$9F,$F2,$BD,$97,$DA,$34,$7D,$D3,$F5,$07,$E8,$1F
	.dc.b	$AE,$F8,$FE,$17,$BA,$D4,$FE,$CB,$EC,$3E,$AE,$8F,$9C,$7E,$B4,$FD
	.dc.b	$4B,$FC,$87,$FA,$53,$E5,$7B,$78,$EF,$53,$FC,$57,$E4,$FB,$B1,$E3
	.dc.b	$87,$0B,$7D,$A7,$E0,$BE,$ED,$FA,$A4,$FD,$53,$F9,$2F,$EA,$CE,$07
	.dc.b	$C8,$EE,$B5,$F2,$0A,$D7,$EA,$DF,$8C,$FA,$37,$D4,$3F,$53,$7E,$A6
	.dc.b	$FC,$99,$F9,$3B,$F0,$1F,$B9,$CF,$E9,$E6,$FF,$21,$E5,$CB,$1E,$38
	.dc.b	$70,$BE,$97,$BB,$7D,$F7,$FA,$0F,$F5,$97,$E7,$2F,$D5,$9F,$7A,$FD
	.dc.b	$65,$FA,$CF,$E8,$B0,$E1,$7B,$AD,$7E,$0D,$FA,$6F,$D5,$BF,$C4,$BF
	.dc.b	$46,$FA,$77,$E9,$1F,$80,$FC,$CD,$F8,$C6,$FE,$A7,$FC,$47,$5A,$F3
	.dc.b	$14,$FD,$27,$F6,$3F,$69,$FC,$BF,$EA,$3F,$98,$7E,$46,$FD,$69,$FB
	.dc.b	$4B,$F5,$3D,$FD,$3F,$FA,$54,$56,$E5,$F9,$77,$EF,$3F,$42,$7E,$19
	.dc.b	$F0,$7F,$35,$FC,$4F,$E5,$1F,$9F,$75,$19,$F5,$93,$FA,$3F,$AF,$FE
	.dc.b	$0B,$F6,$EF,$C6,$1F,$96,$BF,$02,$7B,$F8,$FC,$B9,$F8,$DF,$0B,$DD
	.dc.b	$6A,$D4,$FC,$AF,$F5,$1F,$46,$FC,$73,$D7,$F6,$2F,$CB,$7F,$69,$F0
	.dc.b	$BA,$9C,$C7,$ED,$F1,$FE,$33,$E3,$7F,$95,$7D,$2F,$4F,$DD,$FE,$EB
	.dc.b	$F0,$4F,$A7,$E9,$7E,$16,$B5,$CB,$FE,$C7,$F8,$D5,$7D,$3A,$FC,$57
	.dc.b	$E0,$9F,$C9,$FE,$4D,$FA,$EB,$F5,$97,$E2,$3F,$03,$F5,$DF,$E5,$37
	.dc.b	$BA,$D4,$FC,$4F,$96,$FE,$5B,$E1,$FF,$09,$F6,$9F,$AF,$7E,$AB,$F6
	.dc.b	$07,$EA,$8F,$D7,$37,$F3,$3F,$EA,$2A,$D4,$BF,$60,$FE,$1D,$F5,$CF
	.dc.b	$D8,$7E,$37,$DC,$FA,$BF,$68,$7E,$B1,$A9,$CF,$3A,$5F,$A1,$7F,$40
	.dc.b	$FB,$37,$EC,$0F,$D8,$FF,$8B,$7E,$65,$FB,$BF,$B9,$FE,$17,$F3,$DF
	.dc.b	$C2,$AF,$75,$AF,$94,$FF,$05,$7C,$DF,$F4,$AE,$A7,$EB,$4F,$C6,$5F
	.dc.b	$A7,$3F,$4A,$7E,$C8,$FC,$B7,$F9,$67,$8D,$FB,$9C,$78,$E1,$C2,$F7
	.dc.b	$4F,$CF,$7F,$AB,$FC,$DB,$DF,$78,$2F,$D9,$5F,$B3,$7B,$3F,$74,$7E
	.dc.b	$65,$C2,$F7,$5A,$E9,$57,$E6,$F7,$FE,$15,$F4,$9E,$6F,$81,$FF,$22
	.dc.b	$FD,$EF,$F2,$6F,$DA,$5F,$B4,$FF,$68,$AD,$7D,$2F,$F6,$0E,$95,$7C
	.dc.b	$0C,$7F,$8A,$7C,$CF,$F1,$2E,$4F,$EC,$BB,$FD,$3F,$D0,$FE,$9D,$6B
	.dc.b	$E6,$FE,$26,$BE,$DF,$11,$5F,$DE,$AB,$F8,$F7,$A1,$FE,$B7,$FD,$01
	.dc.b	$FB,$63,$F6,$B7,$CA,$3F,$6C,$7E,$69,$F9,$2F,$C9,$58,$70,$BD,$D3
	.dc.b	$F5,$3F,$F0,$1F,$E7,$0F,$AB,$7B,$BF,$CA,$BE,$B7,$FB,$47,$AB,$FD
	.dc.b	$A7,$83,$F9,$85,$EE,$B5,$F0,$2A,$7E,$5F,$FE,$0E,$FC,$C5,$F9,$73
	.dc.b	$F6,$DF,$E3,$BE,$EF,$F0,$BF,$DB,$BF,$DA,$5A,$7F,$3B,$BD,$D6,$BE
	.dc.b	$51,$5F,$E5,$5A,$FE,$81,$E5,$FE,$D7,$F6,$0F,$F0,$F7,$96,$FC,$ED
	.dc.b	$FB,$7F,$AA,$F7,$F1,$4B,$DD,$6B,$F6,$FF,$93,$F6,$7F,$DA,$3F,$2A
	.dc.b	$FF,$2F,$7E,$57,$FE,$65,$F3,$DF,$92,$3E,$33,$E7,$FC,$67,$F6,$EF
	.dc.b	$AD,$FF,$76,$57,$F7,$09,$FE,$E3,$E2,$FE,$F5,$E1,$3F,$EF,$18,$D4
	.dc.b	$73,$7F,$34,$F5,$B7,$E4,$1A,$8D,$7E,$F1,$FD,$F3,$E1,$5F,$91,$66
	.dc.b	$7F,$03,$3F,$E0,$0B,$7E,$45,$9C,$5F,$A8,$F1,$FC,$A3,$1F,$8B,$7F
	.dc.b	$81,$BE,$9F,$96,$5F,$8A,$FC,$9F,$FC,$13,$A6,$31,$47,$F8,$33,$DC
	.dc.b	$DF,$CE,$E2,$FE,$24,$62,$78,$C7,$96,$7F,$C7,$0F,$C9,$C3,$8E,$7E
	.dc.b	$38,$F7,$4B,$6F,$56,$BC,$5F,$81,$9F,$3F,$81,$F7,$70,$F5,$E4,$FA
	.dc.b	$5F,$99,$FF,$A6,$B2,$F9,$B9,$7D,$17,$FB,$C4,$7E,$99,$CF,$47,$DF
	.dc.b	$FC,$4B,$FD,$49,$D1,$F7,$F9,$7E,$A7,$C6,$7F,$8A,$BA,$FF,$CD,$3F
	.dc.b	$A7,$3F,$43,$7E,$E6,$FD,$57,$FC,$45,$FB,$67,$EC,$B0,$E1,$F3,$6C
	.dc.b	$38,$5E,$FE,$4F,$FA,$07,$67,$F2,$8E,$0F,$C6,$69,$D9,$FC,$2C,$65
	.dc.b	$47,$F9,$EB,$E2,$7F,$09,$39,$8F,$2E,$BE,$4D,$5F,$9E,$FA,$B0,$EB
	.dc.b	$A7,$8E,$B5,$F8,$DE,$67,$AF,$E1,$FF,$DF,$3C,$7F,$F2,$1F,$ED,$0F
	.dc.b	$CB,$DF,$BA,$FB,$6F,$CF,$7B,$FC,$41,$7B,$FC,$9B,$55,$D3,$C6,$F3
	.dc.b	$F3,$7C,$BC,$3F,$C1,$33,$EE,$8C,$78,$BF,$B9,$C7,$9F,$E2,$DE,$5E
	.dc.b	$36,$1D,$FF,$0A,$FD,$6E,$1F,$57,$F1,$3F,$8B,$F3,$6F,$B5,$FF,$06
	.dc.b	$7E,$65,$FC,$6B,$A7,$E3,$56,$A7,$3C,$9E,$6C,$FF,$CB,$BF,$9E,$79
	.dc.b	$B8,$3E,$9D,$FD,$07,$CC,$62,$7F,$AF,$5F,$F8,$67,$6F,$35,$75,$FB
	.dc.b	$6F,$F4,$95,$F5,$FD,$31,$E2,$FE,$D3,$F0,$7F,$F5,$1F,$DB,$7F,$94
	.dc.b	$F6,$FB,$55,$AF,$3D,$5A,$FC,$73,$96,$1E,$5C,$7F,$07,$6F,$D2,$F1
	.dc.b	$BA,$39,$61,$DA,$B8,$E3,$D3,$F2,$B7,$91,$F3,$7F,$79,$7E,$07,$FC
	.dc.b	$7B,$F6,$EF,$F6,$57,$EF,$03,$E3,$4E,$BC,$D4,$EB,$F0,$2C,$FE,$47
	.dc.b	$E2,$AE,$7A,$7C,$83,$F3,$2F,$1A,$FF,$6B,$57,$87,$1F,$07,$E6,$BF
	.dc.b	$E8,$EB,$FA,$9F,$82,$FF,$10,$FA,$5F,$3F,$EF,$8F,$D9,$BF,$C1,$FD
	.dc.b	$1F,$A1,$AD,$70,$2B,$51,$CB,$DF,$F8,$AF,$F1,$2C,$7F,$9B,$5F,$19
	.dc.b	$FE,$F3,$7C,$74,$FC,$85,$F4,$C3,$E3,$E1,$F9,$CF,$96,$FF,$47,$FE
	.dc.b	$58,$F9,$EF,$F4,$27,$C5,$7F,$45,$FA,$AD,$7E,$35,$FE,$DA,$9C,$E0
	.dc.b	$FD,$6B,$3F,$B1,$FD,$BC,$B9,$CF,$C7,$7A,$D7,$08,$E7,$97,$7F,$8B
	.dc.b	$FF,$2A,$78,$7F,$B9,$F4,$3F,$D3,$7F,$82,$FF,$B8,$7F,$01,$7E,$BC
	.dc.b	$FD,$97,$FC,$69,$CF,$FC,$2E,$F7,$ED,$7F,$95,$2B,$5F,$2B,$FB,$89
	.dc.b	$FF,$52,$FE,$FD,$F8,$F7,$AE,$DF,$6D,$E3,$97,$6F,$47,$C5,$F9,$07
	.dc.b	$FB,$AB,$2E,$9E,$B7,$F5,$DF,$CA,$7F,$DA,$7F,$7A,$E9,$FF,$15,$FC
	.dc.b	$7F,$F2,$5F,$D5,$1F,$9A,$FF,$95,$B8,$BF,$DB,$DB,$FF,$6F,$C3,$87
	.dc.b	$C3,$FF,$60,$F8,$A6,$1C,$3F,$3B,$FF,$10,$E9,$DB,$4F,$F2,$8E,$3F
	.dc.b	$CA,$A3,$18,$7F,$28,$6E,$9E,$1F,$67,$CD,$5D,$EF,$CB,$97,$6D,$F3
	.dc.b	$2F,$D3,$9F,$86,$7F,$72,$F4,$9F,$96,$7F,$45,$F5,$DF,$AE,$F9,$30
	.dc.b	$FF,$04,$61,$C3,$56,$1C,$3E,$2B,$F9,$EF,$DF,$EE,$FB,$67,$FE,$DD
	.dc.b	$7F,$C1,$7C,$1F,$EF,$CC,$79,$47,$B4,$EB,$3E,$3D,$3F,$AB,$74,$FA
	.dc.b	$2B,$C1,$97,$7F,$9D,$F5,$FE,$49,$FE,$81,$C7,$E2,$1F,$E1,$D7,$0F
	.dc.b	$F8,$6F,$E6,$3C,$91,$FE,$87,$77,$08,$DB,$FE,$25,$F1,$3E,$3F,$EF
	.dc.b	$0F,$96,$FF,$0B,$FD,$1B,$F6,$EF,$F4,$9F,$E8,$EF,$DD,$BE,$05,$EF
	.dc.b	$97,$FC,$73,$5A,$E6,$CE,$3F,$0F,$3E,$AF,$F8,$89,$F8,$42,$F9,$65
	.dc.b	$DF,$F0,$EF,$F0,$27,$83,$F8,$CA,$D5,$F9,$31,$FF,$52,$FB,$7F,$7B
	.dc.b	$F0,$CF,$EB,$7E,$07,$F8,$A8,$7E,$7B,$FC,$9B,$F7,$6F,$AD,$3A,$F8
	.dc.b	$5F,$F4,$2C,$FF,$2A,$FB,$FC,$FC,$BF,$FA,$23,$4F,$2D,$FF,$66,$F2
	.dc.b	$3E,$4B,$C7,$2E,$BF,$AA,$F8,$6F,$FD,$1F,$97,$4F,$98,$FC,$8F,$B1
	.dc.b	$FE,$6C,$FD,$45,$EC,$23,$FD,$D5,$EB,$BF,$3B,$FC,$54,$6E,$75,$F2
	.dc.b	$2A,$7E,$41,$AF,$B7,$63,$F7,$30,$FF,$3C,$FE,$73,$67,$38,$F2,$FC
	.dc.b	$4F,$C7,$18,$AF,$F7,$D7,$DE,$FF,$96,$FF,$72,$FF,$33,$72,$BE,$C1
	.dc.b	$FA,$3B,$DE,$7E,$66,$FC,$D9,$F1,$6F,$9E,$7E,$A3,$FE,$9C,$CF,$9E
	.dc.b	$5C,$B1,$E3,$87,$0B,$7E,$55,$BD,$3F,$CE,$D7,$E0,$FD,$FB,$FC,$4F
	.dc.b	$8F,$2D,$7C,$55,$FE,$2E,$8C,$65,$D3,$4F,$F4,$54,$F3,$3D,$55,$E9
	.dc.b	$D7,$D1,$3F,$8C,$BF,$7F,$E6,$DE,$9F,$D8,$9F,$25,$FE,$C8,$FC,$D7
	.dc.b	$F8,$2F,$E3,$1C,$EB,$53,$9F,$69,$FA,$B6,$78,$98,$FA,$36,$FF,$48
	.dc.b	$61,$F1,$7E,$93,$10,$8E,$7E,$2B,$4F,$E2,$73,$CE,$FC,$73,$9F,$DD
	.dc.b	$FE,$B4,$FF,$A6,$FF,$6B,$7E,$33,$FE,$9D,$F9,$FF,$F7,$DF,$91,$F4
	.dc.b	$FF,$6A,$F8,$55,$AB,$FF,$00,$FD,$6F,$F9,$03,$77,$E7,$DA,$CB,$FD
	.dc.b	$21,$CB,$F8,$2C,$3A,$FB,$FE,$87,$96,$9D,$BE,$0F,$C7,$9A,$73,$9F
	.dc.b	$F9,$33,$E0,$73,$E0,$7F,$B9,$33,$7E,$39,$FD,$0B,$E8,$BF,$C2,$DF
	.dc.b	$B2,$FD,$7A,$D4,$E7,$EA,$1F,$8D,$67,$D7,$FF,$25,$EB,$F8,$13,$FD
	.dc.b	$79,$7D,$5C,$1F,$95,$8F,$8B,$7D,$FB,$BC,$71,$8F,$93,$FF,$80,$3F
	.dc.b	$5D,$EA,$F4,$3E,$DF,$E3,$0F,$CC,$DF,$42,$F7,$BE,$3F,$EE,$7F,$8D
	.dc.b	$BD,$D6,$A7,$E5,$E9,$9F,$17,$FD,$FF,$B7,$92,$FF,$0F,$F9,$6F,$E0
	.dc.b	$9E,$39,$75,$F8,$1F,$15,$B4,$F1,$4F,$87,$9F,$FD,$A1,$F1,$9F,$EC
	.dc.b	$BF,$C2,$04,$20,$F8,$64,$6E,$3F,$D2,$2F,$EF,$BF,$45,$FA,$7F,$B3
	.dc.b	$3F,$43,$7D,$FB,$E4,$5D,$AD,$FA,$63,$C7,$8D,$8F,$1C,$38,$5B,$C8
	.dc.b	$DF,$27,$FB,$C7,$4F,$C0,$FE,$D1,$E2,$8C,$63,$C9,$FF,$54,$E5,$D7
	.dc.b	$2E,$9A,$3E,$B3,$E0,$D7,$09,$CF,$C1,$62,$1F,$B1,$7E,$CB,$F4,$4F
	.dc.b	$E1,$BF,$16,$FD,$29,$F9,$EF,$F1,$AE,$6F,$8B,$80,$E3,$EA,$30,$E3
	.dc.b	$7C,$3F,$2A,$F5,$D6,$BF,$3F,$F5,$3C,$2A,$E7,$E1,$F8,$26,$9E,$1E
	.dc.b	$BF,$34,$E6,$3C,$9E,$7F,$A4,$7E,$BE,$FA,$A7,$D6,$3F,$39,$7E,$FE
	.dc.b	$FC,$51,$F0,$1E,$4C,$78,$E1,$C2,$F7,$5F,$E9,$7E,$AF,$E7,$B5,$DB
	.dc.b	$F3,$BF,$4B,$C3,$87,$E8,$9F,$0D,$E7,$9F,$87,$3F,$F3,$5F,$37,$C8
	.dc.b	$A5,$6B,$2F,$2E,$3F,$2E,$BF,$CF,$1F,$85,$FE,$2B,$E6,$9F,$5D,$FE
	.dc.b	$F3,$FD,$21,$F4,$5F,$EE,$CF,$C5,$FF,$A8,$DE,$EB,$52,$FC,$EB,$F3
	.dc.b	$95,$AF,$1F,$FC,$29,$B7,$FB,$C2,$31,$F9,$DF,$FB,$B9,$D6,$7E,$0F
	.dc.b	$AD,$FE,$45,$5C,$8F,$C0,$67,$DE,$1A,$3E,$CB,$D5,$87,$F6,$4F,$A0
	.dc.b	$FC,$E7,$CC,$3F,$5D,$7D,$B7,$C2,$C9,$F1,$BD,$74,$61,$C2,$F7,$CD
	.dc.b	$FE,$F5,$FB,$3D,$4F,$4A,$F5,$1F,$93,$7F,$55,$FB,$9F,$AE,$31,$E8
	.dc.b	$B4,$F1,$69,$F4,$B6,$F3,$D6,$A7,$3F,$94,$7F,$7C,$E8,$FF,$55,$7E
	.dc.b	$9D,$FB,$5F,$C3,$3F,$D9,$7F,$BD,$3F,$08,$FD,$2F,$CB,$FA,$6B,$5F
	.dc.b	$82,$AB,$5E,$8B,$F6,$CD,$70,$7F,$B9,$B6,$F4,$8C,$3F,$81,$7E,$1D
	.dc.b	$E9,$8C,$4F,$FB,$B3,$1E,$99,$FF,$8F,$BA,$7F,$C2,$38,$78,$33,$F2
	.dc.b	$73,$7B,$31,$E9,$F8,$67,$F8,$3B,$F1,$87,$E4,$EF,$C8,$7F,$C2,$9C
	.dc.b	$5E,$9A,$D4,$E7,$E3,$DF,$B7,$E7,$D4,$F2,$6A,$BF,$F0,$CE,$7E,$A8
	.dc.b	$C7,$D3,$B4,$EF,$A7,$F2,$3A,$ED,$39,$CF,$C5,$F6,$FE,$7F,$51,$FA
	.dc.b	$67,$FB,$C3,$EC,$7F,$A5,$7E,$71,$FD,$9F,$F2,$2F,$C3,$56,$BE,$77
	.dc.b	$FC,$85,$F2,$2A,$FE,$D9,$D3,$AC,$7F,$DF,$3A,$FC,$EC,$3A,$FE,$83
	.dc.b	$F3,$5E,$59,$F6,$F1,$BC,$57,$E1,$CB,$A7,$83,$F2,$FC,$27,$EB,$9F
	.dc.b	$AD,$FD,$55,$F0,$2F,$C5,$DF,$AE,$7F,$43,$C3,$FD,$E5,$87,$0F,$55
	.dc.b	$87,$0B,$DF,$57,$E8,$6B,$5C,$3F,$84,$7B,$FF,$00,$7E,$0D,$E5,$7C
	.dc.b	$67,$C9,$A7,$8F,$4F,$17,$B3,$8C,$FB,$C7,$97,$5F,$F2,$86,$BE,$4B
	.dc.b	$E1,$AF,$A3,$F0,$9F,$14,$62,$BF,$CC,$F5,$BC,$39,$DF,$BB,$F2,$6F
	.dc.b	$47,$CF,$3F,$4E,$7E,$7A,$FC,$DF,$FD,$45,$F9,$5F,$F3,$5C,$BE,$1D
	.dc.b	$EE,$B5,$EE,$29,$F9,$47,$4C,$7E,$05,$BF,$37,$B1,$DB,$2F,$F5,$77
	.dc.b	$77,$D5,$AE,$91,$FE,$B1,$E3,$C5,$6F,$2E,$DF,$19,$D7,$4F,$0E,$9F
	.dc.b	$B1,$D7,$CB,$7B,$8F,$F3,$DC,$62,$77,$FA,$A7,$1C,$BD,$BD,$17,$2F
	.dc.b	$57,$2F,$DB,$FD,$75,$F8,$2B,$C5,$FF,$33,$F6,$7F,$09,$ED,$D4,$E7
	.dc.b	$EA,$13,$FC,$37,$87,$3D,$BF,$83,$C7,$F1,$AD,$BE,$15,$FE,$72,$8E
	.dc.b	$78,$F6,$FA,$7F,$CC,$5F,$8A,$3C,$3E,$AF,$F7,$CE,$2F,$CA,$CE,$73
	.dc.b	$F0,$CF,$F7,$29,$CE,$5D,$A3,$F1,$3E,$AF,$FA,$83,$81,$87,$1D,$3F
	.dc.b	$92,$71,$7F,$8A,$FF,$1F,$FF,$52,$FE,$BF,$FE,$A6,$DE,$19,$3F,$C1
	.dc.b	$5F,$4B,$FE,$32,$FA,$3F,$F1,$9E,$CB,$FF,$56,$CF,$F8,$DE,$B8,$CE
	.dc.b	$5E,$35,$C7,$E0,$3F,$B5,$5F,$FA,$D3,$E8,$5F,$BE,$32,$7B,$CF,$88
	.dc.b	$7E,$85,$FE,$58,$FC,$A7,$C9,$A7,$F8,$F3,$0E,$17,$BA,$FF,$2E,$FE
	.dc.b	$39,$AC,$BA,$E5,$EC,$C3,$18,$3F,$1B,$C6,$FE,$79,$E9,$5E,$2D,$3C
	.dc.b	$3F,$45,$E8,$9F,$F1,$66,$3D,$34,$F0,$7E,$BD,$99,$FF,$0E,$FC,$D7
	.dc.b	$FB,$77,$F1,$BF,$ED,$7F,$8F,$3F,$1F,$71,$7A,$BE,$6F,$FC,$49,$F8
	.dc.b	$72,$F7,$4E,$3D,$63,$FE,$C6,$CB,$EB,$70,$F9,$F1,$E3,$1F,$43,$97
	.dc.b	$FD,$BF,$1D,$31,$F0,$FC,$4F,$96,$31,$F2,$0D,$9F,$8F,$FF,$87,$7F
	.dc.b	$60,$FF,$73,$FC,$1B,$F1,$5F,$F0,$DF,$E2,$EF,$35,$F8,$BF,$F9,$AF
	.dc.b	$F1,$9F,$F3,$A6,$1C,$2F,$75,$AF,$AE,$7E,$33,$AE,$2F,$F2,$A6,$7D
	.dc.b	$7C,$3F,$A3,$8F,$1F,$A1,$B5,$AC,$B9,$E8,$F4,$3C,$D3,$9A,$6D,$E5
	.dc.b	$AF,$EB,$5C,$DF,$8C,$AF,$F6,$BF,$E0,$BF,$ED,$AF,$3F,$FD,$BD,$F8
	.dc.b	$57,$FA,$2F,$C6,$FF,$73,$CE,$7E,$7F,$FD,$B1,$F3,$7F,$EE,$2C,$79
	.dc.b	$63,$FE,$DD,$AF,$F7,$1B,$F9,$EB,$E3,$F5,$3F,$8A,$79,$FD,$7F,$FB
	.dc.b	$7B,$F0,$9F,$F7,$0E,$1C,$BF,$0A,$FF,$72,$C3,$F0,$AF,$F2,$65,$FF
	.dc.b	$9D,$F1,$7E,$59,$FE,$C6,$F3,$7F,$43,$FB,$3B,$DB,$7E,$66,$FD,$F3
	.dc.b	$D5,$87,$F9,$FF,$0E,$1E,$9B,$0E,$16,$FD,$5F,$77,$DB,$F0,$6F,$D1
	.dc.b	$3E,$B3,$93,$FB,$37,$06,$F3,$F1,$E7,$FC,$6F,$83,$E2,$E7,$E0,$CB
	.dc.b	$CD,$F1,$1E,$CC,$39,$E1,$C3,$3F,$81,$87,$0C,$BF,$80,$69,$E7,$CF
	.dc.b	$DB,$F0,$9E,$6F,$C2,$5F,$C0,$25,$F3,$AF,$C8,$1F,$36,$FD,$59,$F8
	.dc.b	$CF,$1C,$38,$5E,$EB,$5F,$B8,$67,$93,$EC,$71,$7F,$A9,$F3,$EE,$61
	.dc.b	$F0,$61,$CB,$3E,$FF,$B4,$7B,$67,$35,$E4,$9F,$1F,$E0,$19,$F5,$1F
	.dc.b	$91,$7F,$AC,$7F,$0E,$7E,$F3,$FE,$BC,$F2,$1F,$83,$FD,$4B,$DF,$E0
	.dc.b	$4B,$DF,$E0,$5F,$B0,$FA,$9D,$65,$FE,$BB,$DB,$EA,$E1,$DB,$E8,$7D
	.dc.b	$0F,$2C,$38,$F2,$78,$EF,$C5,$AF,$FB,$33,$CA,$C6,$8F,$C7,$1E,$6F
	.dc.b	$F9,$4B,$CA,$F9,$FE,$07,$E8,$B4,$F2,$D6,$B7,$53,$EB,$BC,$B8,$F3
	.dc.b	$C7,$FC,$B7,$E6,$76,$C3,$87,$E3,$FE,$0E,$EF,$F5,$3F,$C5,$FE,$F7
	.dc.b	$CA,$E3,$18,$F5,$3F,$31,$BF,$87,$8B,$FD,$B5,$39,$E6,$DF,$A9,$FD
	.dc.b	$1F,$F0,$AF,$A9,$F0,$CF,$D2,$9E,$B3,$F5,$F7,$E1,$6F,$CB,$3F,$C4
	.dc.b	$9F,$7D,$FA,$4C,$38,$5E,$EB,$5F,$43,$FE,$67,$DF,$F8,$1D,$6B,$5F
	.dc.b	$D2,$BD,$FF,$15,$FC,$43,$CB,$3E,$F9,$BF,$6E,$F2,$EB,$E4,$9C,$FC
	.dc.b	$3E,$35,$7E,$6C,$FD,$D5,$D1,$FE,$70,$EF,$F3,$63,$EC,$AD,$4F,$F9
	.dc.b	$D3,$C7,$CE,$1C,$B0,$E3,$E9,$F5,$BD,$CD,$9E,$BF,$62,$FD,$CB,$8D
	.dc.b	$74,$87,$E7,$1D,$3F,$07,$7E,$20,$FD,$5D,$E7,$7F,$B1,$3A,$3E,$2F
	.dc.b	$07,$D0,$BD,$D6,$BF,$39,$E5,$87,$1A,$F8,$15,$AF,$8C,$B3,$D7,$1F
	.dc.b	$F3,$EF,$C1,$FF,$25,$E9,$FA,$B7,$D2,$5F,$AA,$7C,$0F,$C7,$1F,$99
	.dc.b	$3F,$80,$5E,$5B,$F2,$6D,$4E,$61,$FB,$7E,$2B,$7E,$15,$FA,$0F,$DF
	.dc.b	$7E,$B4,$FC,$69,$FB,$F7,$C2,$FB,$3F,$EA,$9C,$CF,$CB,$BF,$96,$DE
	.dc.b	$EB,$5E,$82,$9F,$9F,$EB,$17,$E6,$FF,$C3,$F0,$E1,$A7,$F3,$07,$97
	.dc.b	$43,$4F,$EC,$5F,$82,$79,$30,$F1,$CE,$73,$7E,$6B,$9D,$5F,$B8,$3F
	.dc.b	$3B,$FF,$60,$73,$FD,$FF,$03,$FA,$EE,$BE,$7A,$D7,$E1,$AA,$FC,$0D
	.dc.b	$6B,$4F,$83,$5A,$F1,$FB,$3C,$3E,$53,$C1,$F8,$B3,$F2,$57,$ED,$7F
	.dc.b	$B3,$E0,$F8,$1B,$FF,$BF,$E7,$3E,$2F,$F9,$7F,$C4,$FE,$C5,$C1,$F2
	.dc.b	$71,$E5,$8F,$EE,$F9,$B8,$C6,$37,$FE,$65,$F8,$8F,$F8,$F6,$5F,$97
	.dc.b	$7F,$5E,$84,$60,$FE,$B6,$FF,$90,$FF,$03,$CD,$7F,$9E,$27,$E7,$79
	.dc.b	$FF,$3F,$EC,$BE,$BF,$E7,$0F,$D0,$9F,$7C,$C3,$87,$77,$F7,$53,$FC
	.dc.b	$7B,$E0,$2F,$87,$C5,$6F,$F0,$CF,$EF,$D5,$FD,$EF,$E9,$7F,$CA,$31
	.dc.b	$8D,$3E,$F6,$BF,$E6,$A8,$C5,$74,$C1,$F7,$7E,$89,$CE,$DE,$6F,$B3
	.dc.b	$C6,$1F,$03,$D1,$FD,$3F,$89,$FA,$BF,$C4,$BC,$53,$9E,$AF,$DB,$4F
	.dc.b	$EB,$FD,$0C,$CF,$F9,$5E,$7F,$D3,$0F,$F9,$43,$2F,$D9,$43,$E0,$7D
	.dc.b	$33,$CF,$C2,$7A,$7F,$C0,$2B,$C7,$3C,$E7,$38,$FF,$09,$F5,$33,$F9
	.dc.b	$CF,$F5,$10,$BF,$09,$FE,$A7,$FB,$A3,$F1,$6F,$2D,$6E,$B5,$F2,$0A
	.dc.b	$7E,$5F,$AF,$91,$78,$BE,$6D,$58,$F7,$7E,$71,$F8,$67,$A6,$7F,$D8
	.dc.b	$BE,$7B,$CB,$7C,$33,$EF,$F3,$9E,$CF,$DA,$FF,$91,$FE,$0B,$FD,$5B
	.dc.b	$F0,$9F,$F7,$6F,$EA,$9F,$2A,$73,$F5,$DF,$87,$9F,$F7,$8F,$CA,$67
	.dc.b	$93,$FD,$97,$BF,$88,$5F,$F2,$EF,$DE,$3D,$0F,$2F,$1F,$FD,$AF,$F6
	.dc.b	$1E,$1B,$FF,$69,$E5,$D7,$E7,$BE,$7F,$C8,$3F,$54,$FF,$3E,$F5,$3F
	.dc.b	$4C,$7E,$2D,$F8,$AF,$DC,$BE,$FD,$EC,$F9,$D3,$AF,$23,$FC,$A1,$AF
	.dc.b	$E9,$67,$A7,$E2,$FF,$2D,$E1,$C3,$E9,$79,$65,$DF,$2F,$CB,$CF,$6C
	.dc.b	$FC,$79,$7C,$08,$FC,$7F,$C4,$7F,$97,$3F,$4F,$7D,$BB,$D1,$7E,$5B
	.dc.b	$FD,$EF,$F6,$AD,$FD,$BA,$F7,$5A,$F2,$1F,$A8,$6A,$7F,$01,$97,$F9
	.dc.b	$A3,$F1,$8B,$87,$1C,$FB,$67,$FC,$37,$8F,$F8,$15,$F8,$E7,$31,$F9
	.dc.b	$AF,$C6,$7F,$DD,$7F,$AB,$BD,$CF,$F5,$DF,$E1,$0F,$1D,$FA,$F7,$EA
	.dc.b	$AF,$75,$A9,$7E,$8B,$9F,$C5,$D8,$FE,$56,$FF,$DE,$FF,$A2,$D9,$CE
	.dc.b	$1C,$79,$B8,$5F,$87,$DA,$FF,$5D,$7C,$17,$EB,$BF,$4A,$FF,$42,$7E
	.dc.b	$82,$FC,$99,$E7,$7F,$A1,$FC,$8F,$F6,$8F,$1F,$F1,$FB,$DD,$6B,$EC
	.dc.b	$3F,$D2,$9F,$21,$FC,$0A,$7F,$D1,$DF,$A4,$FE,$A5,$EB,$B7,$E5,$DE
	.dc.b	$39,$75,$8E,$16,$CF,$C5,$3F,$E9,$7E,$DF,$E0,$7F,$AB,$FE,$EF,$F4
	.dc.b	$DF,$F6,$97,$E1,$8F,$CE,$FF,$8E,$FE,$75,$FC,$4A,$B5,$D9,$FE,$D3
	.dc.b	$9F,$E4,$3F,$A2,$E7,$CA,$63,$FB,$DB,$FE,$0D,$F6,$16,$73,$87,$1F
	.dc.b	$3B,$C3,$18,$CF,$FC,$DD,$C1,$FC,$5B,$A1,$FB,$3B,$E5,$3F,$C4,$3F
	.dc.b	$6F,$FE,$DA,$F2,$9F,$B9,$FF,$26,$BD,$D6,$BF,$14,$FF,$28,$7D,$0E
	.dc.b	$71,$7E,$27,$FC,$46,$F7,$9F,$EF,$DE,$3F,$5B,$FE,$98,$C3,$F2,$5F
	.dc.b	$19,$FE,$C3,$9C,$EB,$FE,$EA,$F0,$3F,$95,$FE,$A6,$FD,$1C,$BE,$A7
	.dc.b	$FD,$F9,$F1,$7F,$93,$FA,$4F,$E7,$73,$B3,$5C,$FF,$C9,$FC,$E7,$EA
	.dc.b	$BA,$C7,$9F,$07,$F1,$3B,$ED,$E3,$7F,$29,$79,$67,$DB,$0F,$1D,$F8
	.dc.b	$A3,$C3,$E4,$FF,$6C,$F8,$5F,$F6,$E7,$E2,$FF,$ED,$73,$FD,$47,$ED
	.dc.b	$3E,$93,$F7,$FF,$A9,$FF,$25,$D7,$FB,$BD,$6F,$F0,$05,$EF,$D7,$7C
	.dc.b	$46,$ED,$7F,$ED,$D7,$FD,$D4,$7F,$BA,$F9,$59,$F8,$BC,$DD,$CE,$8E
	.dc.b	$55,$DF,$1F,$27,$99,$E9,$9F,$1D,$76,$CF,$C7,$AF,$EF,$E7,$C5,$F9
	.dc.b	$7F,$E2,$3E,$7F,$FD,$1F,$F2,$EF,$CD,$1F,$94,$7F,$94,$B7,$F1,$E1
	.dc.b	$FE,$14,$C3,$87,$CC,$7F,$6B,$BD,$FE,$0F,$D3,$C5,$FF,$23,$DF,$9F
	.dc.b	$0E,$2B,$0F,$CB,$FD,$8F,$AE,$38,$69,$E2,$C5,$F3,$FE,$B9,$E1,$8F
	.dc.b	$F9,$4B,$3F,$07,$9D,$9F,$C1,$5F,$9F,$BF,$1E,$7E,$81,$FD,$F3,$F3
	.dc.b	$F7,$CB,$BF,$2F,$FA,$58,$70,$BD,$FD,$12,$F7,$F2,$5A,$CB,$A6,$DE
	.dc.b	$BE,$EF,$F1,$AA,$C7,$C3,$FA,$07,$AF,$9B,$DF,$1D,$76,$FF,$06,$FE
	.dc.b	$29,$F7,$70,$E3,$39,$D3,$C1,$C1,$EF,$FD,$0D,$F9,$A3,$F4,$47,$EC
	.dc.b	$7F,$ED,$ED,$7F,$A7,$FD,$7F,$E2,$EF,$75,$AF,$A9,$57,$F5,$4F,$CC
	.dc.b	$B5,$8F,$3E,$1F,$51,$FE,$4E,$C3,$BF,$D9,$7A,$77,$F6,$C7,$4D,$3C
	.dc.b	$1E,$37,$97,$0F,$24,$78,$FC,$FF,$AD,$DA,$FD,$8F,$FD,$E7,$F9,$DF
	.dc.b	$FA,$27,$BF,$F1,$7F,$91,$7E,$86,$B5,$C2,$AD,$78,$CF,$C9,$B5,$8F
	.dc.b	$FB,$F3,$87,$CE,$7E,$11,$FC,$AB,$8C,$39,$DF,$77,$C7,$3E,$2F,$C0
	.dc.b	$BF,$92,$F3,$FE,$2F,$CB,$FF,$37,$8E,$87,$E6,$FF,$E3,$0F,$DC,$1F
	.dc.b	$A3,$7E,$DE,$B7,$5A,$FB,$25,$7F,$BD,$7C,$1F,$CE,$D7,$F9,$EB,$F7
	.dc.b	$3F,$95,$BF,$F9,$EA,$3B,$BE,$A7,$9E,$5F,$E3,$EE,$8F,$27,$DE,$27
	.dc.b	$C9,$3F,$E4,$2E,$7F,$D1,$7D,$6B,$EB,$3F,$DB,$7F,$17,$FE,$96,$FC
	.dc.b	$2D,$F9,$4B,$F2,$EF,$E1,$D7,$B3,$FA,$0F,$E0,$3F,$EE,$09,$FF,$4F
	.dc.b	$70,$7E,$D2,$7A,$7C,$27,$A3,$83,$FD,$CF,$9F,$FB,$A1,$FA,$33,$CF
	.dc.b	$3F,$07,$4F,$92,$BC,$79,$7F,$BB,$39,$FF,$71,$F8,$0B,$F7,$03,$E6
	.dc.b	$FD,$E9,$F8,$AF,$CA,$FC,$77,$F9,$24,$EC,$D4,$7C,$EF,$CA,$73,$6F
	.dc.b	$D2,$3C,$D5,$BD,$FF,$0A,$F1,$E2,$65,$DB,$A7,$C7,$87,$8A,$73,$97
	.dc.b	$F2,$3E,$3F,$E2,$BF,$86,$FF,$BB,$3C,$67,$DC,$3F,$69,$FE,$31,$83
	.dc.b	$F3,$0F,$E7,$57,$BA,$D7,$C0,$FD,$2E,$EC,$FD,$8B,$2F,$99,$7F,$EF
	.dc.b	$8F,$26,$CE,$70,$E3,$D3,$FC,$7E,$BA,$69,$FC,$9F,$9F,$F9,$17,$E3
	.dc.b	$8F,$D5,$BE,$4F,$EC,$7F,$D1,$BF,$70,$FF,$00,$FE,$97,$FD,$36,$B5
	.dc.b	$C8,$AD,$7A,$2F,$D4,$F5,$8F,$D2,$EB,$FD,$6E,$1D,$7F,$25,$FC,$13
	.dc.b	$CB,$0E,$DE,$3F,$8A,$FC,$3E,$BB,$5F,$F1,$BE,$3F,$D7,$7F,$08,$FC
	.dc.b	$D7,$E8,$EF,$CC,$DA,$3F,$6E,$7E,$87,$FE,$68,$EF,$FE,$8B,$87,$0B
	.dc.b	$DD,$6B,$F1,$DE,$9D,$AC,$BF,$2D,$7F,$A7,$FD,$E2,$CE,$70,$E3,$CF
	.dc.b	$FD,$26,$BA,$69,$E1,$EA,$FE,$99,$F8,$A3,$F4,$6F,$F7,$07,$E7,$8F
	.dc.b	$88,$FC,$C7,$53,$F4,$4F,$EF,$17,$B1,$5B,$F1,$FF,$C6,$3F,$B9,$6B
	.dc.b	$17,$E4,$3F,$AB,$BE,$1C,$FF,$C6,$9E,$58,$71,$EB,$F1,$DF,$8B,$4F
	.dc.b	$D4,$79,$5F,$D8,$7E,$B1,$FB,$E9,$7E,$A8,$FC,$C1,$F4,$9F,$33,$9B
	.dc.b	$FC,$1D,$5C,$0D,$CE,$BB,$BF,$AC,$77,$7F,$74,$9E,$BF,$96,$7F,$01
	.dc.b	$78,$70,$FF,$30,$79,$65,$DF,$2F,$F5,$2D,$F1,$CF,$C9,$39,$E7,$FE
	.dc.b	$F1,$FB,$73,$D7,$3F,$E0,$1F,$C2,$BE,$4F,$EE,$0F,$DC,$7F,$C4,$93
	.dc.b	$B3,$5C,$C9,$7E,$C7,$AC,$1F,$B5,$7F,$17,$BE,$1C,$1F,$D1,$9E,$5E
	.dc.b	$AF,$FA,$73,$1F,$DF,$79,$3F,$DE,$27,$3A,$7D,$AE,$7F,$F0,$77,$0B
	.dc.b	$DC,$7E,$F8,$FD,$AD,$FC,$02,$FD,$05,$F2,$AF,$D6,$FF,$C7,$98,$F1
	.dc.b	$C3,$85,$EE,$BF,$DB,$93,$FE,$22,$35,$C1,$FD,$71,$F0,$7E,$F3,$FC
	.dc.b	$C5,$E5,$97,$83,$2F,$9F,$BF,$F9,$17,$4F,$2C,$E7,$9F,$FC,$4D,$F9
	.dc.b	$33,$F3,$17,$EC,$DF,$E2,$3F,$A3,$7E,$63,$F2,$F5,$FF,$29,$5E,$EB
	.dc.b	$53,$F3,$F8,$73,$87,$2D,$FF,$B8,$DF,$5F,$5F,$FA,$33,$C7,$E9,$BF
	.dc.b	$DE,$B8,$7C,$EA,$09,$D6,$7E,$2E,$5F,$F1,$E7,$C7,$BF,$4E,$2F,$E0
	.dc.b	$0F,$E5,$D1,$FE,$21,$FC,$0F,$FC,$D1,$3B,$35,$F1,$1C,$7E,$93,$97
	.dc.b	$47,$F9,$3A,$75,$A7,$F8,$21,$ED,$F8,$37,$F5,$D7,$96,$5D,$BC,$DF
	.dc.b	$1E,$7E,$28,$F0,$F3,$7F,$94,$7C,$E7,$E8,$FF,$C1,$7E,$43,$F0,$5F
	.dc.b	$A2,$7F,$1D,$FC,$0F,$FC,$F9,$5A,$13,$AF,$75,$FA,$4E,$B0,$7E,$93
	.dc.b	$FD,$BE,$F8,$70,$7F,$87,$9E,$58,$76,$F5,$FC,$57,$E1,$F0,$DA,$7F
	.dc.b	$30,$E3,$FF,$36,$FE,$36,$FC,$D4,$BE,$0F,$FC,$3D,$B7,$F9,$2F,$F1
	.dc.b	$BF,$F4,$94,$EC,$D7,$D1,$79,$BA,$5F,$8F,$6B,$17,$EE,$5F,$E0,$5B
	.dc.b	$E1,$C5,$FE,$34,$79,$67,$DB,$A7,$FC,$F1,$18,$CB,$E0,$F3,$FF,$9E
	.dc.b	$FC,$97,$EB,$1F,$1B,$E8,$BF,$C5,$7F,$BE,$FF,$47,$F8,$2F,$F5,$35
	.dc.b	$6B,$A1,$5A,$EB,$FE,$43,$CB,$FE,$8C,$AD,$6F,$FE,$46,$7A,$FE,$91
	.dc.b	$FE,$1B,$79,$5F,$6E,$AF,$16,$9E,$1F,$A6,$6B,$F2,$FD,$0F,$F3,$EF
	.dc.b	$D7,$BD,$07,$EA,$6F,$DE,$3F,$B6,$76,$3E,$C9,$E6,$7F,$AF,$70,$E1
	.dc.b	$7B,$F9,$55,$EF,$F3,$2D,$65,$D3,$8F,$FC,$BB,$87,$7F,$C7,$7F,$C6
	.dc.b	$4F,$2D,$3B,$F6,$7F,$A8,$E3,$19,$74,$EA,$FF,$53,$7E,$44,$FB,$47
	.dc.b	$EC,$2F,$D2,$DF,$94,$FD,$AF,$C5,$5F,$90,$FF,$B4,$F0,$E1,$7B,$AD
	.dc.b	$7E,$9B,$9C,$79,$F0,$7F,$9C,$AF,$B7,$E8,$1F,$E4,$77,$8E,$7D,$BC
	.dc.b	$7F,$E0,$54,$9C,$E3,$F8,$AF,$63,$FD,$79,$F9,$EB,$F3,$E2,$FC,$6D
	.dc.b	$F8,$5B,$C9,$F9,$DF,$9F,$FF,$BA,$6B,$81,$B9,$D7,$7B,$F3,$2D,$63
	.dc.b	$FE,$7D,$D7,$FD,$13,$7C,$1B,$1E,$0C,$BB,$E5,$FC,$BA,$7F,$8C,$D7
	.dc.b	$69,$CF,$37,$FB,$4F,$F3,$C7,$E7,$85,$F9,$E3,$E9,$3E,$97,$53,$F2
	.dc.b	$6F,$F8,$01,$5C,$0D,$CE,$BF,$3D,$CE,$0F,$C0,$7F,$CE,$55,$BD,$7F
	.dc.b	$D4,$2F,$1C,$BB,$75,$7C,$E8,$C6,$3F,$81,$E5,$FF,$74,$7E,$72,$FC
	.dc.b	$E4,$F8,$FF,$39,$FC,$BF,$87,$FE,$5E,$F4,$FF,$DE,$B3,$B3,$5E,$5A
	.dc.b	$7E,$B4,$EA,$DF,$9D,$7F,$A1,$EF,$86,$DF,$EB,$67,$97,$13,$2E,$F2
	.dc.b	$C3,$95,$78,$FC,$AF,$F7,$24,$FF,$48,$F6,$FF,$8C,$F9,$8F,$CD,$7F
	.dc.b	$09,$F9,$33,$D4,$7C,$BF,$3F,$4B,$FD,$E7,$87,$0B,$DD,$3F,$37,$D6
	.dc.b	$5D,$7D,$DF,$D9,$BF,$EC,$CB,$FE,$4F,$F9,$BF,$FA,$85,$E5,$E0,$34
	.dc.b	$FC,$DC,$61,$FF,$17,$D7,$41,$F9,$87,$DF,$FC,$C7,$87,$E6,$1F,$AE
	.dc.b	$E8,$7A,$DB,$AD,$60,$AD,$7E,$65,$9E,$86,$3F,$EB,$3B,$FC,$57,$E6
	.dc.b	$56,$73,$87,$1F,$57,$F3,$1A,$78,$A7,$39,$3D,$4F,$9B,$2F,$CB,$1F
	.dc.b	$96,$5F,$CB,$3D,$DE,$57,$CD,$FC,$BB,$D9,$5A,$13,$AE,$87,$E1,$FA
	.dc.b	$C3,$A6,$3F,$EF,$6B,$EB,$F9,$77,$FB,$2D,$E5,$87,$19,$E1,$1D,$37
	.dc.b	$67,$F3,$3C,$1F,$E7,$F5,$FE,$0A,$FA,$77,$DC,$FF,$04,$FC,$8D,$F9
	.dc.b	$3B,$F2,$57,$E3,$6F,$D0,$BF,$1D,$F1,$7F,$E1,$2C,$99,$72,$F7,$19
	.dc.b	$72,$C7,$8F,$E5,$1C,$2F,$C1,$A7,$F1,$27,$91,$FE,$1C,$D7,$CA,$F4
	.dc.b	$87,$E5,$1A,$8D,$67,$E1,$CF,$C1,$97,$7F,$67,$35,$DB,$4F,$1E,$FE
	.dc.b	$BF,$0B,$32,$A7,$E3,$BF,$CF,$7F,$1E,$7E,$3D,$F1,$38,$FF,$D9,$DC
	.dc.b	$1E,$4B,$DF,$B0,$BD,$F0,$FC,$6F,$C4,$BD,$D9,$F5,$CF,$FA,$43,$E9
	.dc.b	$C7,$94,$7C,$C8,$C7,$0F,$CB,$CB,$AE,$3D,$1F,$39,$8D,$3E,$BE,$BE
	.dc.b	$AF,$C6,$9F,$8C,$97,$E3,$5F,$E8,$5F,$A6,$7E,$2E,$FC,$6F,$CB,$5C
	.dc.b	$0D,$FD,$D7,$F9,$1A,$7F,$C3,$DC,$1F,$4A,$7A,$F8,$7F,$47,$83,$8E
	.dc.b	$7D,$DF,$49,$FE,$54,$9E,$99,$F8,$76,$F3,$57,$96,$3C,$9A,$FD,$2F
	.dc.b	$C5,$7F,$DE,$7D,$5E,$6F,$AC,$F9,$31,$8D,$5F,$86,$63,$F1,$7F,$F8
	.dc.b	$01,$AF,$80,$57,$FC,$01,$70,$F4,$D3,$B4,$BF,$18,$7E,$30,$FC,$63
	.dc.b	$F3,$3A,$FE,$13,$E1,$BF,$DB,$7F,$89,$FF,$BA,$BF,$2E,$FB,$71,$8F
	.dc.b	$C2,$DF,$8A,$63,$7F,$F7,$46,$FE,$3A,$FF,$74,$36,$3F,$DD,$1F,$89
	.dc.b	$FF,$BD,$72,$ED,$18,$CB,$F2,$55,$F8,$8F,$C3,$3F,$D8,$7F,$87,$7F
	.dc.b	$B2,$EF,$1F,$86,$7F,$B0,$A3,$FD,$9D,$F9,$26,$3F,$0F,$FA,$9D,$BF
	.dc.b	$A8,$C3,$FD,$9B,$8F,$EB,$23,$18,$71,$FC,$3F,$FD,$9B,$E2,$7E,$E6
	.dc.b	$FB,$C7,$8A,$BF,$62,$7F,$A9,$BF,$07,$FF,$57,$7E,$11,$F9,$C8,$C7
	.dc.b	$DB,$FF,$AD,$BE,$4B,$1F,$5D,$FE,$BC,$D3,$EA,$57,$FA,$C1,$B1,$FE
	.dc.b	$AC,$C3,$FD,$7D,$C1,$FB,$2A,$E7,$D9,$FB,$4F,$1F,$FD,$13,$E0,$7F
	.dc.b	$A4,$7E,$FF,$F5,$71,$8F,$BE,$7D,$BF,$FA,$53,$8B,$F8,$94,$FF,$A5
	.dc.b	$35,$F8,$11,$8A,$7D,$FF,$C6,$8C,$57,$FA,$7A,$7E,$1D,$7F,$AB,$23
	.dc.b	$B6,$5F,$C1,$BF,$14,$7E,$67,$FA,$AF,$D6,$1D,$3F,$8B,$FB,$BF,$F1
	.dc.b	$67,$07,$F1,$8B,$DF,$90,$BD,$F8,$DF,$E8,$9F,$D5,$FF,$82,$D6,$B5
	.dc.b	$FC,$D3,$DB,$C4,$F3,$47,$FA,$FA,$39,$DF,$2F,$D6,$7C,$9A,$F8,$F5
	.dc.b	$FC,$E7,$C1,$FF,$1C,$F9,$4F,$F4,$27,$DF,$7F,$A4,$F7,$FE,$57,$18
	.dc.b	$E6,$FF,$48,$C6,$38,$7F,$D2,$3A,$FE,$AE,$BF,$D2,$2F,$CA,$3F,$D9
	.dc.b	$EF,$4E,$9E,$D3,$FE,$A2,$FB,$EF,$F5,$0F,$27,$F2,$CF,$B4,$FF,$34
	.dc.b	$FC,$E7,$F9,$E7,$E8,$7F,$91,$C6,$3E,$B1,$F3,$F8,$FB,$6F,$F3,$B6
	.dc.b	$BD,$C5,$7F,$9D,$DF,$DF,$3D,$34,$ED,$3F,$E8,$8F,$B6,$FF,$43,$E7
	.dc.b	$FC,$F3,$EA,$FF,$C9,$9F,$0D,$FE,$58,$FA,$FF,$E6,$71,$8F,$AD,$FF
	.dc.b	$46,$7D,$7E,$3E,$BF,$FC,$B5,$AF,$86,$BF,$CB,$4D,$8F,$F2,$D6,$1F
	.dc.b	$E6,$7D,$3B,$46,$30,$FD,$47,$89,$FD,$8F,$E9,$FF,$C6,$BF,$51,$FE
	.dc.b	$42,$F6,$FF,$A1,$C6,$39,$3F,$C8,$18,$63,$CD,$FF,$1F,$E9,$FC,$6E
	.dc.b	$BF,$C7,$EF,$C3,$78,$67,$FD,$61,$EB,$C9,$FD,$C7,$AB,$FB,$3F,$CF
	.dc.b	$7F,$88,$3E,$6F,$FC,$49,$F4,$3F,$D4,$A3,$1F,$3D,$FE,$4F,$FA,$1C
	.dc.b	$7C,$FF,$F8,$8B,$4F,$E5,$55,$FE,$22,$6C,$7F,$87,$AB,$F4,$B1,$FE
	.dc.b	$21,$FA,$1F,$F1,$8F,$D0,$78,$63,$FB,$97,$C8,$FF,$82,$7E,$53,$FC
	.dc.b	$11,$F2,$9F,$E0,$7D,$BF,$BE,$FC,$4F,$F7,$69,$CF,$CB,$FF,$95,$F0
	.dc.b	$E3,$D7,$FC,$69,$E7,$F2,$F5,$FE,$F5,$F2,$FF,$DF,$5E,$06,$2F,$FB
	.dc.b	$E3,$83,$FB,$17,$C7,$7F,$4F,$F9,$1F,$EB,$9F,$25,$9F,$91,$FE,$AB
	.dc.b	$1F,$D7,$FC,$CC,$57,$F8,$C2,$FF,$BB,$1F,$E1,$DA,$FE,$CC,$E0,$FE
	.dc.b	$BD,$1D,$7D,$AF,$E5,$11,$B8,$C7,$67,$F8,$6E,$3F,$C4,$DF,$17,$FD
	.dc.b	$27,$E3,$7C,$CF,$F4,$BF,$8D,$BF,$19,$FD,$2A,$BF,$9C,$1C,$17,$A3
	.dc.b	$F0,$E3,$FD,$45,$1F,$D2,$70,$ED,$03,$3F,$8C,$BE,$F5,$F1,$EF,$8F
	.dc.b	$F3,$D7,$F8,$BF,$E2,$3F,$9A,$7C,$4B,$F0,$E7,$C5,$FC,$BE,$DF,$E7
	.dc.b	$F1,$FE,$9F,$7F,$A6,$57,$F3,$FB,$F3,$9C,$FD,$6F,$34,$FF,$46,$D7
	.dc.b	$CB,$3F,$E9,$5A,$EF,$EA,$E3,$DA,$FE,$29,$ED,$7D,$97,$F8,$B7,$D2
	.dc.b	$DE,$AF,$E2,$B8,$76,$13,$FC,$59,$FC,$91,$FC,$5A,$3F,$8A,$7E,$00
	.dc.b	$F7,$BF,$D5,$D3,$CE,$BE,$5C,$E7,$1F,$2F,$B1,$FC,$17,$DD,$7B,$7F
	.dc.b	$8B,$8C,$7B,$77,$D6,$FE,$0F,$97,$80,$FF,$76,$BF,$EE,$E7,$F0,$C6
	.dc.b	$27,$FD,$D5,$ED,$F8,$B8,$BF,$CB,$F1,$FC,$4F,$2F,$8B,$E8,$FE,$DB
	.dc.b	$8F,$FD,$37,$5F,$E6,$48,$C7,$A5,$F9,$D7,$FD,$45,$8F,$67,$F7,$0F
	.dc.b	$C5,$79,$3B,$F4,$BF,$D2,$FE,$87,$F9,$DF,$F4,$E7,$E6,$5F,$97,$F2
	.dc.b	$3F,$C8,$3E,$47,$F9,$1B,$CA,$56,$B1,$FC,$FD,$6B,$E2,$9D,$3A,$C7
	.dc.b	$F4,$1F,$33,$FE,$45,$C3,$86,$1F,$B2,$C3,$15,$DF,$E6,$7F,$C8,$93
	.dc.b	$C4,$EA,$FF,$91,$1C,$1F,$87,$FC,$39,$F8,$83,$E4,$FF,$C2,$78,$F6
	.dc.b	$E3,$FC,$AF,$BF,$C1,$78,$5F,$95,$C7,$E0,$67,$FC,$9E,$1E,$1F,$92
	.dc.b	$7F,$45,$9F,$6C,$BF,$C2,$B1,$CE,$BF,$2E,$F9,$3A,$7F,$0D,$DB,$7B
	.dc.b	$3F,$A1,$77,$3B,$B1,$DF,$D7,$0F,$C1,$1F,$E9,$9E,$0F,$8B,$8F,$F5
	.dc.b	$3E,$DF,$DC,$46,$31,$FE,$99,$DD,$8E,$0F,$F5,$77,$CC,$3F,$26,$7E
	.dc.b	$51,$FE,$91,$FD,$29,$F8,$2F,$EE,$7E,$53,$E4,$DE,$EB,$53,$FB,$DF
	.dc.b	$A3,$CE,$1C,$B7,$F2,$65,$DA,$1F,$B4,$FA,$71,$E5,$F4,$6D,$3B,$F0
	.dc.b	$79,$71,$FE,$21,$5A,$9F,$1E,$D9,$94,$E7,$4F,$E1,$DF,$45,$F8,$EF
	.dc.b	$CC,$7F,$09,$EE,$7F,$36,$E3,$46,$3C,$6F,$F4,$CF,$89,$BC,$BF,$98
	.dc.b	$F8,$9F,$8F,$AF,$CD,$47,$CD,$78,$69,$F3,$1E,$5B,$59,$FE,$4D,$E1
	.dc.b	$BE,$A9,$F3,$CE,$B7,$E2,$6F,$DC,$1F,$BD,$FF,$B1,$F8,$3C,$B9,$72
	.dc.b	$C7,$8E,$1C,$2F,$77,$E7,$AF,$F5,$9D,$7F,$A6,$9F,$31,$C7,$1F,$53
	.dc.b	$6F,$96,$BE,$78,$E3,$3E,$5C,$BC,$9B,$67,$BB,$F2,$7A,$79,$2F,$F8
	.dc.b	$87,$DA,$3E,$19,$E9,$7F,$88,$2B,$56,$F0,$7F,$8D,$E1,$CD,$FE,$34
	.dc.b	$FF,$BD,$27,$71,$FE,$1F,$E2,$D5,$7F,$1C,$D6,$E5,$DF,$17,$86,$FD
	.dc.b	$23,$F8,$5F,$E8,$FF,$D5,$74,$7D,$07,$F3,$5F,$C7,$5F,$89,$3C,$F7
	.dc.b	$EE,$4E,$37,$EF,$3C,$B1,$FE,$BD,$8F,$1C,$38,$5E,$FF,$48,$D5,$76
	.dc.b	$F7,$3F,$B0,$3F,$B1,$D3,$C5,$87,$86,$30,$74,$96,$9E,$4D,$3C,$7A
	.dc.b	$7E,$DB,$0E,$35,$AC,$FC,$3C,$39,$C7,$96,$3C,$74,$F5,$7E,$75,$E4
	.dc.b	$F4,$DF,$B1,$BF,$6C,$7E,$7A,$F8,$EF,$F6,$AF,$33,$F3,$DE,$17,$BA
	.dc.b	$D7,$CF,$FD,$CF,$CF,$73,$8F,$F3,$1E,$2F,$9D,$9F,$6F,$DA,$31,$C1
	.dc.b	$FE,$05,$8F,$0D,$78,$35,$FF,$71,$FE,$59,$E9,$AF,$E4,$BA,$77,$9F
	.dc.b	$F7,$3F,$E8,$79,$C3,$8E,$BE,$2D,$DF,$A5,$7D,$0E,$FF,$C9,$ED,$F7
	.dc.b	$DF,$ED,$86,$7C,$1F,$C7,$E9,$63,$CA,$FE,$63,$C3,$FF,$37,$39,$F0
	.dc.b	$7F,$19,$D1,$E5,$AF,$EE,$7E,$1B,$1E,$D8,$FF,$32,$F0,$78,$65,$FA
	.dc.b	$0F,$09,$F8,$7F,$F5,$CF,$9A,$FC,$97,$94,$FC,$81,$F8,$13,$F4,$67
	.dc.b	$CD,$3F,$4F,$7D,$FF,$F4,$8E,$06,$5C,$B1,$E3,$87,$0B,$DF,$C3,$77
	.dc.b	$E5,$FF,$1A,$69,$DF,$4F,$EC,$D8,$F1,$FC,$16,$D6,$B2,$E7,$AF,$87
	.dc.b	$AF,$FA,$2D,$F7,$C3,$CD,$3E,$5E,$EC,$E1,$C6,$73,$A7,$A3,$6F,$14
	.dc.b	$FF,$1A,$CB,$AC,$FB,$3F,$70,$CF,$EB,$EF,$CE,$9F,$BE,$BF,$74,$7D
	.dc.b	$FF,$F8,$13,$F7,$57,$EF,$7C,$2F,$75,$AD,$54,$FB,$0D,$63,$EE,$D9
	.dc.b	$9F,$7F,$37,$0F,$86,$7C,$1A,$7F,$84,$BF,$58,$F6,$DF,$0D,$3B,$CF
	.dc.b	$F8,$5B,$D3,$CE,$1C,$75,$F1,$68,$FD,$7F,$C3,$18,$3F,$C3,$BD,$FF
	.dc.b	$C4,$FE,$2D,$F1,$9E,$8E,$27,$E8,$4F,$94,$7E,$74,$FD,$2B,$8F,$0F
	.dc.b	$F9,$BF,$1E,$3A,$FC,$4C,$38,$5B,$F1,$ED,$DF,$F8,$C3,$3F,$8A,$FF
	.dc.b	$80,$F4,$F1,$43,$F4,$3F,$F1,$54,$63,$1E,$9A,$3E,$C3,$FC,$0F,$39
	.dc.b	$DB,$CF,$3F,$CE,$F1,$EF,$AF,$92,$73,$A3,$EE,$F3,$A7,$BF,$6F,$3F
	.dc.b	$D1,$E3,$5E,$F6,$FD,$B7,$18,$F2,$F9,$BF,$5F,$E7,$2F,$C3,$DF,$9D
	.dc.b	$3F,$13,$7E,$87,$FD,$A7,$07,$CF,$70,$BD,$D6,$BE,$47,$F1,$7C,$19
	.dc.b	$F9,$2F,$83,$C9,$FE,$9B,$D3,$B6,$98,$84,$63,$0E,$5F,$83,$7C,$49
	.dc.b	$E1,$97,$6D,$7F,$BD,$7E,$2D,$99,$E3,$39,$D3,$F3,$7F,$65,$E5,$CF
	.dc.b	$D5,$87,$A7,$F2,$5F,$F3,$1F,$DE,$7F,$59,$F9,$47,$A6,$FC,$5D,$F9
	.dc.b	$DF,$CD,$FC,$FD,$EA,$7E,$AF,$D0,$63,$C7,$0E,$17,$BA,$7E,$75,$DB
	.dc.b	$87,$FD,$75,$9F,$6F,$DB,$3F,$CD,$98,$F1,$D2,$D6,$B2,$E7,$E0,$B5
	.dc.b	$F0,$CF,$33,$FC,$37,$3A,$87,$E8,$7F,$E5,$0C,$3F,$9B,$4B,$F0,$C7
	.dc.b	$EA,$EF,$D2,$3F,$A9,$47,$AB,$87,$D7,$97,$FB,$3F,$E1,$FE,$1C,$F0
	.dc.b	$AD,$FE,$23,$E1,$9F,$8F,$FA,$37,$A2,$D1,$C1,$FF,$32,$67,$E2,$C3
	.dc.b	$97,$CB,$3F,$05,$F3,$57,$7D,$7C,$DA,$F9,$69,$F9,$D6,$AB,$AD,$6B
	.dc.b	$5F,$4F,$E1,$3E,$38,$C5,$FB,$FE,$DF,$FD,$3F,$F1,$7F,$E1,$5C,$9F
	.dc.b	$54,$FD,$E7,$FB,$AF,$E4,$EF,$C5,$5F,$9C,$7F,$47,$E6,$FF,$0F,$EF
	.dc.b	$FE,$F1,$C7,$8E,$1C,$2F,$7F,$95,$76,$AD,$65,$D5,$FF,$42,$67,$E1
	.dc.b	$8F,$F2,$7C,$60,$DD,$F3,$FC,$67,$FD,$0F,$5B,$8E,$DF,$97,$7F,$A2
	.dc.b	$B6,$FF,$17,$7E,$19,$FE,$09,$FD,$7F,$F6,$3F,$B4,$3F,$22,$7E,$5E
	.dc.b	$FC,$2B,$FD,$7F,$87,$0E,$56,$1C,$2F,$7F,$85,$6B,$70,$CF,$FC,$2D
	.dc.b	$F1,$5F,$EC,$1C,$79,$61,$6B,$75,$E3,$F5,$DA,$F8,$FD,$3F,$45,$6B
	.dc.b	$4F,$0D,$6B,$47,$17,$FA,$9B,$E7,$BF,$DB,$5F,$A9,$FF,$10,$FC,$89
	.dc.b	$FB,$AB,$F4,$CD,$FE,$FF,$FE,$DA,$BD,$FA,$AB,$DD,$74,$FC,$F7,$5B
	.dc.b	$7F,$AC,$B2,$7B,$BF,$ED,$BC,$78,$CF,$FA,$E5,$E7,$87,$2D,$3C,$93
	.dc.b	$C7,$2E,$DA,$78,$FA,$7F,$D7,$BE,$2F,$F8,$8F,$F3,$CF,$F5,$CF,$E4
	.dc.b	$6F,$D2,$DF,$B2,$FF,$86,$FF,$5D,$FF,$78,$DE,$FF,$0B,$5E,$EB,$FC
	.dc.b	$41,$F6,$7A,$D7,$F6,$59,$7F,$AE,$DF,$F3,$C6,$1D,$FF,$0A,$FF,$7A
	.dc.b	$3C,$F0,$F0,$7E,$93,$FE,$73,$F2,$BF,$7D,$7C,$B3,$EF,$AF,$93,$F0
	.dc.b	$0F,$F6,$AF,$E6,$DF,$E7,$9F,$53,$FB,$67,$D6,$7F,$95,$FF,$75,$7E
	.dc.b	$A5,$F4,$BA,$DA,$30,$E1,$F3,$9F,$E3,$2F,$B7,$61,$C3,$A1,$F9,$67
	.dc.b	$0B,$DE,$7F,$E8,$7E,$0F,$34,$72,$FB,$1C,$47,$90,$FC,$9E,$DE,$4C
	.dc.b	$BA,$ED,$F8,$89,$E8,$6B,$4F,$16,$9E,$8F,$1B,$FD,$E2,$FF,$A5,$7F
	.dc.b	$31,$FE,$CD,$F7,$9F,$DB,$FA,$DC,$7F,$E9,$3E,$9F,$6D,$EF,$E0,$D7
	.dc.b	$BA,$FF,$4C,$7E,$81,$AD,$3F,$4F,$97,$47,$D3,$CB,$F0,$9C,$E0,$FC
	.dc.b	$B7,$DB,$1C,$45,$79,$3E,$7F,$EA,$7D,$2F,$A2,$78,$E7,$E1,$F1,$BF
	.dc.b	$6B,$97,$F8,$6A,$7F,$6F,$FA,$27,$5F,$59,$AB,$F2,$67,$EA,$2F,$96
	.dc.b	$7E,$BD,$F6,$BE,$87,$FD,$71,$8F,$1C,$38,$5E,$FE,$C3,$59,$F6,$E2
	.dc.b	$7E,$7D,$EA,$DB,$C5,$2C,$7C,$4F,$C2,$CB,$AD,$FF,$14,$CB,$AE,$9E
	.dc.b	$A9,$CE,$DE,$8D,$BC,$FD,$3E,$EF,$AD,$FF,$66,$7D,$EB,$F3,$AF,$F7
	.dc.b	$CF,$EE,$81,$FA,$87,$F1,$CF,$9C,$FF,$26,$6F,$FC,$6A,$F8,$7C,$C6
	.dc.b	$EF,$7F,$00,$F9,$DD,$DA,$F6,$FD,$AC,$FC,$39,$BE,$D7,$E0,$46,$32
	.dc.b	$E9,$A7,$F2,$CB,$E7,$87,$83,$3F,$34,$FD,$19,$FD,$4F,$E4,$7F,$EE
	.dc.b	$7F,$D4,$1E,$B7,$F9,$C3,$F3,$8F,$F5,$07,$E7,$FF,$CC,$7F,$0E,$DE
	.dc.b	$1F,$CD,$2F,$7F,$52,$BD,$FC,$47,$F9,$67,$B5,$9F,$5C,$DF,$B8,$7D
	.dc.b	$78,$C6,$17,$2E,$95,$E1,$FC,$19,$AB,$C4,$FC,$7E,$1C,$34,$F2,$F8
	.dc.b	$DF,$92,$E9,$FF,$80,$1F,$A4,$7F,$96,$7C,$17,$E6,$CF,$C1,$1F,$AB
	.dc.b	$BF,$65,$FF,$46,$F1,$FF,$92,$B8,$7F,$A7,$E3,$C7,$0E,$17,$BF,$C0
	.dc.b	$7F,$87,$E7,$DB,$CD,$FD,$56,$1F,$AE,$8C,$7E,$75,$F9,$37,$5A,$F8
	.dc.b	$B5,$F0,$FE,$17,$E7,$BE,$F8,$7E,$CF,$3E,$EF,$EB,$3F,$63,$FE,$C3
	.dc.b	$3F,$DA,$6F,$FC,$C3,$FA,$D7,$F6,$CC,$1F,$94,$FF,$24,$E0,$61,$C3
	.dc.b	$E6,$7E,$07,$C6,$30,$E1,$6F,$DD,$FF,$94,$DE,$F2,$FF,$36,$E7,$FC
	.dc.b	$22,$31,$F2,$1F,$AD,$75,$8F,$3F,$35,$AB,$F2,$AF,$46,$1F,$DE,$35
	.dc.b	$F0,$EB,$E7,$F8,$DF,$E0,$FE,$5F,$B3,$DE,$FF,$18,$7E,$10,$FB,$E7
	.dc.b	$EE,$EF,$C8,$9F,$8A,$FF,$59,$E6,$FD,$86,$BF,$E0,$4C,$78,$E1,$C2
	.dc.b	$F7,$FB,$77,$F4,$4C,$FB,$7D,$6B,$F2,$0F,$E0,$D1,$8D,$6E,$5C,$E9
	.dc.b	$1E,$3D,$7C,$5F,$B1,$7A,$2B,$53,$9C,$9F,$AB,$FF,$22,$FC,$67,$EB
	.dc.b	$FE,$8C,$FD,$99,$F5,$0F,$D8,$7F,$E0,$0F,$E8,$FD,$BF,$30,$FE,$B3
	.dc.b	$87,$0B,$DD,$6B,$3F,$CB,$72,$AB,$58,$FF,$1F,$78,$74,$DD,$BF,$C4
	.dc.b	$FA,$78,$6B,$C1,$F8,$BB,$5F,$3D,$78,$0D,$E9,$E0,$9D,$70,$7F,$3C
	.dc.b	$FD,$EB,$F2,$3F,$A9,$7A,$BE,$87,$FD,$BD,$F5,$EF,$DF,$7F,$DB,$1F
	.dc.b	$B8,$7F,$72,$BD,$EB,$FE,$7B,$F9,$7F,$F2,$E5,$7F,$B7,$7E,$9F,$5B
	.dc.b	$7F,$8F,$33,$E8,$FF,$56,$C3,$8E,$5F,$CE,$DE,$79,$F8,$30,$FF,$73
	.dc.b	$57,$21,$7E,$5C,$7B,$FD,$CF,$F5,$43,$F5,$3F,$75,$FB,$FD,$DF,$B7
	.dc.b	$B8,$3E,$67,$EE,$4D,$7F,$C1,$78,$70,$F6,$18,$70,$BD,$FE,$A1,$AD
	.dc.b	$7F,$CA,$B9,$F5,$7F,$B7,$E1,$C7,$F2,$1F,$EA,$8F,$3C,$39,$74,$FF
	.dc.b	$35,$EA,$FF,$5E,$DF,$2C,$FB,$ED,$E5,$AF,$AF,$5F,$EC,$0A,$FF,$31
	.dc.b	$FE,$F2,$DD,$FB,$47,$E6,$3C,$E7,$E3,$8F,$C7,$9F,$5B,$FE,$26,$FC
	.dc.b	$3B,$F8,$5E,$3C,$70,$E1,$7B,$FA,$4D,$67,$DB,$3E,$AF,$F8,$2B,$0E
	.dc.b	$31,$FD,$E2,$30,$6E,$F9,$F8,$3F,$E0,$4A,$DC,$76,$FC,$77,$FC,$0F
	.dc.b	$F9,$47,$F1,$CF,$D7,$3F,$B5,$7D,$83,$F7,$54,$3F,$50,$FF,$0F,$FE
	.dc.b	$BF,$FE,$3E,$BE,$1E,$8B,$0E,$17,$BF,$A7,$7E,$5B,$BD,$FF,$D1,$33
	.dc.b	$FD,$DF,$EC,$FF,$C0,$D8,$F2,$C2,$D6,$B1,$E9,$A7,$8B,$47,$E0,$1E
	.dc.b	$8A,$D5,$F8,$25,$F0,$BF,$E2,$2F,$9E,$FE,$6F,$F8,$03,$F2,$4F,$F7
	.dc.b	$0F,$EC,$0F,$DC,$FF,$D9,$3F,$4C,$FB,$8F,$F2,$CE,$1C,$2F,$75,$AF
	.dc.b	$B7,$CE,$BF,$EA,$2C,$BA,$7C,$87,$F8,$7F,$0E,$1F,$89,$DA,$D6,$3C
	.dc.b	$BF,$0F,$7E,$01,$E5,$9C,$D3,$6F,$25,$3F,$5A,$FF,$1D,$FD,$47,$49
	.dc.b	$CF,$07,$9F,$C8,$4E,$78,$3D,$7E,$A7,$F9,$CA,$3D,$5B,$FF,$8C,$DF
	.dc.b	$2E,$9F,$E5,$27,$86,$7E,$48,$E0,$79,$32,$F1,$E7,$FE,$C0,$FC,$6F
	.dc.b	$FC,$3D,$39,$CB,$E4,$FE,$F2,$9C,$CB,$8B,$E8,$6F,$ED,$BF,$2D,$6A
	.dc.b	$7E,$4D,$7F,$9F,$23,$1C,$2B,$FF,$31,$C6,$1E,$59,$F9,$E3,$19,$F9
	.dc.b	$AF,$C5,$D7,$FD,$DB,$E3,$9F,$90,$7F,$25,$FC,$E3,$FC,$23,$F9,$D7
	.dc.b	$FA,$C7,$B7,$78,$7F,$B7,$2F,$7E,$DA,$F7,$F8,$1F,$F9,$FB,$46,$7D
	.dc.b	$73,$7F,$00,$79,$A3,$18,$5C,$BA,$57,$87,$E6,$DA,$BF,$1F,$F9,$98
	.dc.b	$70,$D3,$CB,$EC,$FC,$FF,$37,$F6,$FF,$11,$FE,$F8,$F4,$9F,$97,$7F
	.dc.b	$8A,$78,$3F,$89,$6C,$F5,$3F,$85,$5E,$EB,$FC,$6D,$5A,$9F,$E1,$D5
	.dc.b	$F3,$F0,$FE,$13,$F7,$BF,$39,$EF,$EC,$7F,$18,$F4,$BC,$55,$FC,$2F
	.dc.b	$2F,$17,$A9,$F8,$3D,$7F,$95,$7B,$5F,$C9,$65,$FB,$97,$E0,$E7,$3F
	.dc.b	$03,$FD,$9E,$5F,$AE,$67,$1F,$E5,$DE,$FF,$DE,$BF,$E6,$D1,$F8,$37
	.dc.b	$8F,$BD,$FD,$9B,$87,$F9,$AC,$BE,$07,$FA,$56,$9F,$CD,$73,$FC,$B7
	.dc.b	$C1,$FF,$56,$F8,$4F,$EB,$FA,$3F,$38,$FC,$B4,$E7,$E1,$FF,$B4,$4F
	.dc.b	$F9,$47,$3F,$EA,$19,$FF,$63,$F8,$9F,$F7,$2B,$FE,$D3,$8C,$BF,$D7
	.dc.b	$A1,$5F,$96,$BE,$9E,$03,$F5,$9F,$1D,$F3,$F8,$9F,$ED,$1F,$AF,$7E
	.dc.b	$AF,$87,$FC,$1D,$FB,$63,$F6,$77,$C8,$BF,$5E,$7E,$C4,$FD,$63,$F5
	.dc.b	$D8,$70,$BD,$D6,$BF,$61,$CE,$9F,$E0,$AC,$BF,$53,$7F,$E1,$BF,$DF
	.dc.b	$6C,$E7,$0E,$3B,$78,$E3,$1A,$78,$7C,$EF,$E0,$1F,$0B,$FE,$38,$FE
	.dc.b	$00,$FF,$15,$7E,$9E,$FC,$7D,$F7,$CF,$D2,$7F,$81,$5E,$FE,$69,$7B
	.dc.b	$A7,$E5,$3A,$D3,$FC,$7D,$D9,$FB,$EC,$3B,$47,$F0,$37,$95,$F7,$F2
	.dc.b	$3C,$73,$FE,$45,$D3,$C7,$A3,$C8,$FE,$2D,$F5,$9F,$E6,$4F,$AA,$7E
	.dc.b	$EA,$FD,$FF,$FD,$31,$FA,$77,$F1,$7A,$D4,$E7,$EF,$1F,$97,$E7,$4F
	.dc.b	$F3,$47,$D9,$FF,$9C,$9F,$C1,$5F,$F9,$C9,$FE,$30,$F2,$C7,$FD,$49
	.dc.b	$1C,$4F,$F4,$37,$D9,$7F,$9F,$F1,$F1,$7E,$2B,$FC,$AF,$E2,$FF,$D2
	.dc.b	$7E,$F3,$F1,$6F,$F1,$4F,$EA,$6F,$D5,$7F,$96,$56,$A7,$26,$32,$FF
	.dc.b	$4B,$FD,$D7,$F2,$69,$FF,$4F,$C7,$F2,$B7,$87,$C8,$3F,$75,$F0,$4F
	.dc.b	$0C,$BC,$13,$FE,$31,$8F,$E7,$5F,$81,$7F,$B0,$B9,$3E,$F7,$BA,$F9
	.dc.b	$47,$E3,$DA,$BF,$F6,$2F,$E0,$EB,$DD,$3F,$73,$D5,$3F,$0B,$FF,$6D
	.dc.b	$FE,$15,$FC,$F3,$3F,$F6,$7C,$FF,$41,$7C,$1D,$CD,$3F,$9F,$46,$1F
	.dc.b	$EC,$93,$FD,$3F,$F0,$FF,$F7,$77,$E5,$7F,$27,$E5,$DF,$8F,$BF,$6F
	.dc.b	$D5,$FF,$BF,$FC,$7F,$EC,$F7,$BA,$FD,$67,$90,$C7,$FD,$F1,$7F,$D4
	.dc.b	$DF,$F7,$EF,$CF,$FF,$52,$7B,$F5,$7F,$82,$7B,$7F,$B2,$46,$3A,$7F
	.dc.b	$D8,$B8,$3F,$16,$FF,$63,$FE,$32,$ED,$7C,$5B,$F2,$97,$E7,$3F,$C1
	.dc.b	$FE,$77,$FD,$71,$8F,$1D,$C3,$1E,$58,$71,$BE,$1F,$84,$FF,$71,$CF
	.dc.b	$BE,$6F,$C7,$3F,$B4,$C6,$2E,$E3,$D3,$1E,$7A,$78,$BD,$87,$E3,$7F
	.dc.b	$DF,$F0,$E1,$39,$FC,$6F,$FC,$05,$C7,$FE,$E2,$FC,$7D,$F9,$03,$F4
	.dc.b	$B7,$E7,$DF,$E1,$EF,$DA,$5F,$B3,$7F,$BF,$B9,$7F,$3D,$C1,$EC,$C7
	.dc.b	$8E,$1C,$3E,$B5,$87,$0F,$D0,$7F,$C1,$9A,$77,$D3,$FC,$83,$8F,$F8
	.dc.b	$1E,$31,$87,$F8,$19,$BA,$78,$7E,$B3,$CF,$5D,$EF,$CD,$3E,$5C,$BF
	.dc.b	$C3,$9F,$94,$7F,$BF,$3B,$DF,$2A,$FC,$85,$FB,$1B,$E4,$5F,$A9,$BF
	.dc.b	$78,$7C,$1F,$F8,$87,$2E,$58,$F1,$C3,$85,$EF,$F5,$2D,$6F,$F7,$34
	.dc.b	$EF,$F9,$87,$F8,$7B,$1E,$3A,$DA,$D6,$5C,$F5,$F0,$FE,$3F,$E8,$9C
	.dc.b	$D7,$9A,$7C,$BF,$07,$FE,$37,$7F,$13,$3F,$CC,$3F,$18,$7E,$26,$FD
	.dc.b	$F5,$F8,$D3,$F5,$17,$E8,$DF,$D4,$FB,$FF,$B0,$FC,$73,$1E,$38,$70
	.dc.b	$BD,$F2,$7F,$90,$74,$EF,$7B,$CF,$FC,$73,$18,$C1,$FA,$16,$35,$FC
	.dc.b	$ED,$EF,$2E,$95,$E9,$9C,$BE,$0D,$7C,$DF,$A3,$FF,$94,$F7,$F6,$FA
	.dc.b	$6F,$D5,$FF,$D2,$DF,$50,$FD,$19,$FB,$93,$F7,$E7,$E5,$9F,$E5,$7C
	.dc.b	$78,$E1,$C2,$F7,$4F,$D5,$B5,$B7,$EC,$B3,$ED,$FB,$1F,$F9,$53,$1E
	.dc.b	$3F,$84,$DA,$D6,$5C,$F4,$FF,$1F,$47,$21,$87,$9E,$BF,$C3,$59,$F9
	.dc.b	$BF,$02,$FF,$37,$7E,$81,$F6,$3E,$F7,$D1,$F3,$DF,$CE,$7F,$25,$7C
	.dc.b	$B3,$F2,$4F,$F0,$8F,$7B,$0E,$1C,$6C,$38,$5E,$FB,$3F,$CE,$D5,$A9
	.dc.b	$FF,$36,$BC,$3F,$73,$BC,$1F,$C4,$34,$F0,$FC,$13,$5F,$0F,$E0,$1E
	.dc.b	$6A,$D6,$9D,$EB,$5A,$7F,$BA,$B5,$7E,$85,$F9,$2F,$75,$FA,$8F,$F5
	.dc.b	$FF,$C6,$DF,$8E,$BF,$0A,$FA,$1F,$0B,$FE,$8B,$C3,$85,$EE,$B5,$DC
	.dc.b	$FC,$B3,$5B,$FF,$26,$CF,$AF,$CE,$FF,$A2,$31,$E3,$ED,$DA,$D6,$5C
	.dc.b	$F4,$FE,$17,$B7,$9A,$73,$5F,$EF,$7D,$3C,$9F,$A2,$7F,$A8,$7D,$4F
	.dc.b	$D5,$FD,$D7,$F3,$AF,$CC,$7F,$D2,$DF,$BA,$7F,$9E,$FF,$37,$FE,$97
	.dc.b	$5A,$F8,$B5,$6B,$A5,$5F,$D5,$75,$FE,$AF,$AF,$F3,$EC,$BF,$A2,$FD
	.dc.b	$47,$FA,$83,$2E,$58,$DA,$D6,$5D,$3C,$26,$9F,$E8,$BF,$89,$F3,$56
	.dc.b	$AF,$C1,$F0,$78,$8F,$F3,$8E,$1C,$6F,$D1,$C1,$FE,$B9,$CB,$BE,$5D
	.dc.b	$37,$F3,$FE,$05,$9C,$7F,$A9,$FE,$37,$FE,$D5,$F0,$DF,$9A,$BF,$7E
	.dc.b	$7C,$B7,$F3,$AC,$F8,$60,$C3,$85,$EE,$BF,$9F,$FE,$9B,$A9,$FF,$70
	.dc.b	$7E,$1B,$EF,$BE,$1C,$3F,$9D,$7F,$D8,$D8,$FF,$1C,$AE,$79,$78,$27
	.dc.b	$DD,$18,$C5,$D1,$FE,$D9,$EB,$FE,$47,$FB,$C3,$E3,$7F,$C2,$D8,$3F
	.dc.b	$5D,$7E,$28,$FD,$C7,$FD,$93,$DB,$F9,$8A,$FE,$E7,$87,$1F,$A6,$FE
	.dc.b	$EB,$87,$0B,$FF,$70,$5E,$ED,$7B,$7F,$DB,$D9,$78,$B2,$FF,$15,$C6
	.dc.b	$0E,$AF,$93,$0F,$1F,$E5,$FE,$99,$CE,$5E,$0C,$DF,$9F,$66,$7B,$E1
	.dc.b	$C6,$7D,$5F,$A2,$7F,$BD,$72,$F0,$61,$C7,$5F,$DE,$6B,$E7,$F3,$1C
	.dc.b	$8F,$D7,$3F,$D9,$9F,$A9,$3F,$0E,$7E,$C1,$FD,$93,$0E,$17,$BF,$C5
	.dc.b	$57,$BA,$7E,$B9,$E1,$E1,$FF,$0D,$E7,$DB,$F1,$9F,$F7,$EE,$FF,$D5
	.dc.b	$A3,$C3,$AE,$2B,$79,$F5,$CB,$C5,$F9,$8F,$BE,$F7,$A7,$83,$67,$D0
	.dc.b	$67,$0E,$11,$E8,$FB,$8F,$16,$5F,$CA,$A7,$3B,$BF,$43,$CE,$5D,$34
	.dc.b	$F5,$7E,$39,$8B,$DF,$E4,$1C,$BF,$58,$4B,$F5,$87,$DD,$7F,$90,$7F
	.dc.b	$59,$FC,$77,$E4,$7B,$AD,$4E,$7B,$5F,$21,$9C,$3E,$07,$03,$2E,$BF
	.dc.b	$AD,$61,$EF,$3D,$BE,$AD,$A3,$F2,$AF,$C1,$56,$B3,$F1,$7C,$A6,$35
	.dc.b	$F0,$E1,$C3,$57,$E8,$9E,$1B,$F2,$7E,$7E,$F0,$1F,$AE,$FF,$7C,$F1
	.dc.b	$1F,$46,$EB,$FF,$79,$74,$BF,$3D,$FE,$3B,$87,$0E,$5F,$EF,$B8,$70
	.dc.b	$BD,$D7,$F8,$37,$1E,$9F,$53,$FE,$16,$BF,$F3,$4C,$77,$CB,$11,$FE
	.dc.b	$62,$C7,$A3,$FE,$CA,$C7,$A7,$CD,$7F,$64,$E2,$D6,$1C,$73,$F2,$F2
	.dc.b	$63,$0F,$1E,$5E,$9C,$3F,$B5,$E8,$FD,$AB,$E0,$63,$C6,$73,$F4,$38
	.dc.b	$C3,$FC,$E7,$F1,$9F,$EE,$1F,$D7,$9F,$70,$FD,$79,$F9,$7F,$E0,$FF
	.dc.b	$74,$E1,$83,$0E,$17,$BA,$FC,$47,$DB,$EA,$7A,$57,$C0,$E4,$F7,$C7
	.dc.b	$FB,$BF,$F2,$8F,$87,$3B,$FC,$11,$A7,$8B,$4F,$0F,$DF,$7B,$EB,$C1
	.dc.b	$5A,$D3,$C1,$F5,$D9,$CB,$C7,$87,$A3,$C3,$F5,$E9,$E0,$C3,$84,$E7
	.dc.b	$F4,$8F,$06,$9D,$F7,$FC,$DD,$EE,$31,$83,$F0,$6F,$83,$F9,$23,$07
	.dc.b	$E2,$EF,$C8,$1F,$9A,$7F,$A7,$7F,$13,$7E,$BD,$C2,$F7,$5A,$FC,$0F
	.dc.b	$FD,$F5,$F8,$66,$71,$F8,$7B,$7E,$13,$3E,$DF,$97,$63,$B7,$FC,$C3
	.dc.b	$1E,$1A,$F0,$6B,$FE,$AB,$FB,$FF,$8D,$39,$CF,$B7,$B7,$8D,$7C,$37
	.dc.b	$E5,$F5,$3C,$15,$E3,$E0,$7E,$C7,$F7,$BC,$F7,$EF,$8E,$97,$E4,$EF
	.dc.b	$D3,$7F,$D7,$7E,$EB,$F4,$CE,$17,$BA,$D7,$83,$FC,$BF,$F4,$BC,$FD
	.dc.b	$8F,$AE,$3F,$D8,$79,$F6,$CF,$11,$FE,$CF,$C3,$96,$7D,$A3,$AC,$F8
	.dc.b	$A3,$FD,$97,$F1,$7F,$85,$BE,$59,$78,$34,$FC,$5F,$4F,$EB,$A7,$37
	.dc.b	$E8,$F1,$BF,$37,$F1,$EE,$3F,$F1,$EF,$8C,$FC,$C1,$A7,$DD,$FD,$8B
	.dc.b	$75,$A9,$CF,$33,$F0,$CC,$FD,$3F,$AB,$2E,$85,$BE,$18,$FF,$7A,$E2
	.dc.b	$DE,$7D,$E3,$F8,$3C,$7F,$B9,$AB,$C9,$7F,$B4,$CB,$BF,$E8,$DC,$1F
	.dc.b	$0F,$6D,$C3,$F2,$5F,$6B,$75,$7E,$71,$FC,$87,$E7,$3F,$D5,$3F,$B5
	.dc.b	$BF,$4E,$7E,$76,$FC,$5D,$FA,$93,$F6,$17,$EC,$7F,$C9,$7A,$FF,$C4
	.dc.b	$BC,$9F,$EF,$AC,$B9,$63,$C7,$0E,$17,$FD,$46,$F7,$6B,$F0,$7E,$45
	.dc.b	$F9,$ED,$3C,$7A,$78,$AB,$F8,$34,$63,$2E,$99,$BF,$03,$FE,$51,$39
	.dc.b	$D7,$CF,$3F,$E2,$7C,$FB,$E7,$E4,$9C,$E9,$FB,$29,$F1,$FD,$AB,$E4
	.dc.b	$1E,$F3,$F7,$1F,$F1,$6F,$B8,$FD,$6F,$FD,$F3,$87,$0B,$DD,$6A,$5F
	.dc.b	$AD,$FC,$DE,$1F,$F3,$C5,$74,$7D,$66,$3E,$AB,$E0,$3A,$CB,$C1,$F4
	.dc.b	$AC,$FF,$CF,$3C,$3F,$E8,$DC,$3B,$CE,$4F,$83,$7C,$73,$F1,$7A,$1E
	.dc.b	$18,$7E,$17,$FE,$E2,$3F,$E0,$07,$CF,$3F,$34,$FF,$4C,$FE,$63,$B9
	.dc.b	$D7,$E2,$5F,$E9,$A9,$FF,$4E,$FE,$87,$9F,$8F,$7E,$65,$E8,$CB,$AC
	.dc.b	$76,$7D,$A6,$30,$EF,$7F,$4F,$7F,$F7,$4D,$79,$31,$ED,$D7,$EA,$C3
	.dc.b	$84,$78,$7C,$77,$E4,$1E,$1C,$B9,$C7,$B3,$A3,$F1,$71,$E8,$DF,$D9
	.dc.b	$3E,$69,$FC,$7F,$EE,$6F,$D4,$DF,$1A,$FE,$00,$7E,$C7,$FE,$64,$7F
	.dc.b	$01,$7C,$2B,$71,$AF,$AD,$E1,$F2,$1F,$37,$1E,$98,$FC,$37,$B7,$E3
	.dc.b	$57,$0E,$39,$7F,$9C,$3F,$08,$F7,$57,$0C,$BA,$7B,$58,$BD,$E7,$E5
	.dc.b	$CB,$ED,$7D,$9B,$F5,$9F,$D8,$DB,$F7,$EF,$E1,$C3,$F7,$FD,$CE,$A3
	.dc.b	$3F,$83,$E3,$C9,$FB,$18,$73,$F1,$6D,$EF,$DB,$FE,$25,$C9,$FB,$6A
	.dc.b	$F0,$5F,$F9,$97,$1F,$E3,$9D,$5F,$5E,$73,$A7,$8F,$0F,$27,$E5,$1F
	.dc.b	$91,$F9,$7F,$EE,$3F,$2D,$FE,$49,$E7,$C7,$C2,$7F,$70,$8F,$F0,$56
	.dc.b	$BF,$C1,$2B,$FC,$0B,$9F,$F8,$9E,$31,$4F,$81,$FC,$85,$74,$F2,$FF
	.dc.b	$10,$F6,$F9,$8B,$5D,$7E,$5B,$FB,$FF,$8A,$CB,$FC,$3B,$39,$F8,$DF
	.dc.b	$1F,$CB,$FF,$7C,$FA,$3F,$D4,$7E,$B8,$FC,$C5,$F9,$4F,$F8,$87,$D6
	.dc.b	$7E,$B2,$FD,$F7,$FD,$2B,$C7,$FC,$73,$0E,$17,$BA,$D6,$9D,$31,$E7
	.dc.b	$AF,$E9,$AF,$B7,$E5,$DF,$AB,$78,$FC,$FF,$F0,$FC,$3E,$9F,$07,$A6
	.dc.b	$D5,$AD,$7C,$7E,$1F,$EA,$63,$17,$E5,$F9,$2F,$E0,$5F,$6D,$FA,$1F
	.dc.b	$9B,$7C,$D3,$F6,$6F,$E7,$9F,$AD,$7B,$F8,$FF,$9A,$DE,$EB,$5F,$08
	.dc.b	$A7,$DE,$F4,$F3,$3F,$CA,$FC,$1E,$5C,$78,$C7,$F3,$E8,$C7,$0F,$A7
	.dc.b	$2E,$99,$FF,$4B,$F9,$DF,$E0,$11,$8C,$FC,$B7,$D7,$5F,$3D,$F9,$AF
	.dc.b	$1A,$F9,$35,$F2,$69,$FD,$33,$E4,$5F,$8E,$7F,$71,$C3,$FD,$CB,$F7
	.dc.b	$5F,$EE,$0F,$90,$FE,$49,$5A,$F2,$3E,$17,$C2,$7F,$AB,$39,$FF,$B2
	.dc.b	$E5,$CF,$25,$FF,$1F,$38,$6F,$FC,$79,$E5,$9F,$6B,$FE,$35,$18,$C3
	.dc.b	$FD,$6F,$F5,$58,$C3,$C9,$1D,$FE,$03,$C7,$18,$D7,$FD,$EB,$F8,$3E
	.dc.b	$35,$F2,$C7,$FB,$E3,$F1,$6F,$E7,$9F,$7E,$FC,$97,$FD,$CB,$F2,$8F
	.dc.b	$D4,$1F,$96,$BF,$48,$FF,$2A,$72,$7F,$70,$C3,$87,$E0,$BF,$E0,$2B
	.dc.b	$DD,$7E,$9B,$97,$F4,$D5,$D7,$C7,$FE,$7C,$F0,$8F,$E7,$AF,$2C,$BC
	.dc.b	$19,$7F,$73,$9F,$2E,$7E,$4C,$7C,$7E,$77,$F3,$DC,$38,$47,$83,$CB
	.dc.b	$DE,$9E,$3C,$7D,$1F,$5D,$FD,$67,$F0,$1F,$F6,$07,$C1,$BE,$B7,$F6
	.dc.b	$7E,$B3,$F2,$27,$E1,$FB,$EC,$FE,$77,$5A,$97,$E1,$D9,$FD,$B3,$FA
	.dc.b	$C6,$5D,$38,$FF,$AB,$BD,$AF,$F7,$FF,$80,$FE,$62,$7F,$84,$4E,$7D
	.dc.b	$4F,$F5,$0F,$D2,$FF,$57,$8C,$6B,$E3,$F9,$0F,$16,$FE,$7B,$FF,$53
	.dc.b	$EA,$FC,$AF,$FB,$A7,$BB,$FE,$19,$FC,$D3,$F8,$A7,$EF,$EF,$CA,$1F
	.dc.b	$83,$7F,$89,$7C,$1F,$F1,$3D,$EF,$81,$7B,$A7,$EF,$1D,$3E,$5D,$97
	.dc.b	$F7,$5C,$3F,$C4,$DF,$25,$67,$38,$F2,$CF,$FC,$1C,$F1,$BF,$1C,$3C
	.dc.b	$0F,$EE,$98,$78,$73,$F0,$7E,$95,$BC,$39,$68,$FC,$47,$FC,$21,$F8
	.dc.b	$FB,$A9,$FA,$7B,$F3,$B7,$EE,$8F,$D7,$FF,$D4,$DF,$49,$C2,$F7,$5A
	.dc.b	$9F,$F5,$57,$EE,$99,$C3,$FC,$21,$8F,$5F,$A9,$B8,$70,$CB,$A7,$AB
	.dc.b	$FE,$0B,$AF,$15,$F8,$67,$C1,$B6,$6F,$86,$5F,$EB,$7F,$9D,$FF,$14
	.dc.b	$7A,$1F,$ED,$FF,$C0,$5F,$A7,$3F,$63,$43,$F7,$AF,$AD,$F0,$BF,$CD
	.dc.b	$6F,$85,$EE,$B5,$DD,$D3,$E3,$38,$EB,$FE,$26,$BF,$F7,$67,$E6,$1F
	.dc.b	$E2,$B7,$8E,$7D,$BE,$9F,$FC,$49,$18,$C5,$F3,$F8,$C7,$C7,$9F,$7D
	.dc.b	$5D,$7F,$E4,$1F,$C7,$9F,$8E,$3F,$70,$FF,$6A,$FE,$14,$FC,$E7,$EF
	.dc.b	$7C,$E6,$EB,$5F,$84,$AB,$5E,$2F,$FA,$A3,$9B,$FA,$96,$5D,$3F,$77
	.dc.b	$FF,$21,$3D,$B3,$B8,$F2,$C3,$8F,$37,$59,$F0,$E9,$FC,$1F,$CB,$FF
	.dc.b	$41,$E1,$C7,$4F,$04,$FE,$47,$F6,$6E,$11,$8D,$5F,$B9,$63,$3F,$2F
	.dc.b	$DF,$FF,$97,$FF,$23,$7C,$3B,$F3,$1B,$FA,$3B,$DB,$FF,$5D,$78,$3B
	.dc.b	$AD,$09,$D7,$E0,$9F,$E4,$6E,$1F,$F4,$BE,$3D,$31,$FF,$30,$3D,$BF
	.dc.b	$16,$B8,$72,$C3,$B7,$6F,$F9,$6E,$B8,$4F,$5C,$DF,$9E,$66,$38,$47
	.dc.b	$FB,$37,$EA,$19,$FF,$1D,$E7,$FF,$0C,$7E,$59,$FE,$3D,$FD,$11,$FB
	.dc.b	$BF,$F2,$1F,$90,$7E,$D2,$FD,$AF,$8F,$BF,$FE,$07,$C3,$85,$EE,$BF
	.dc.b	$DC,$33,$9F,$B5,$7E,$57,$FE,$0D,$CF,$FB,$D4,$E7,$3C,$47,$F8,$A7
	.dc.b	$4F,$22,$CB,$AD,$34,$7C,$56,$E3,$96,$9E,$7F,$9A,$C6,$1F,$E3,$3F
	.dc.b	$8C,$FF,$15,$7E,$1E,$FC,$C5,$FA,$43,$F5,$57,$DF,$BF,$21,$F6,$E3
	.dc.b	$C7,$EC,$58,$F1,$C3,$85,$BF,$3A,$DD,$FF,$46,$BF,$F2,$CB,$FE,$4D
	.dc.b	$DB,$FC,$59,$A7,$8A,$3D,$FC,$5E,$E8,$F2,$EB,$E4,$D7,$FD,$99,$7C
	.dc.b	$30,$F0,$CF,$A3,$BB,$E4,$E3,$E5,$D7,$C9,$87,$AB,$EF,$3C,$7F,$1A
	.dc.b	$FC,$79,$FA,$6B,$F4,$0F,$CB,$7E,$56,$FC,$65,$F8,$FF,$0B,$DD,$6A
	.dc.b	$7F,$B2,$71,$7D,$4C,$7A,$7E,$6B,$F6,$36,$E1,$9D,$AD,$63,$CB,$C3
	.dc.b	$69,$FE,$80,$E3,$FF,$27,$65,$E5,$C3,$C9,$5A,$E9,$FB,$39,$7F,$5B
	.dc.b	$C7,$96,$3E,$7F,$C6,$FE,$7F,$B5,$FE,$D0,$FD,$CD,$F9,$03,$CC,$7E
	.dc.b	$D0,$FC,$43,$FA,$BF,$0F,$E3,$30,$E1,$7B,$AD,$7C,$CF,$4F,$C3,$99
	.dc.b	$7A,$F8,$FF,$CC,$98,$7F,$B7,$BF,$6F,$C7,$07,$F9,$FE,$39,$EB,$E0
	.dc.b	$FC,$4B,$D3,$18,$CD,$F2,$5E,$CF,$D7,$1E,$C7,$D1,$FD,$E6,$F9,$AF
	.dc.b	$84,$7E,$51,$F8,$CB,$DF,$72,$F7,$5F,$CA,$79,$7F,$4F,$5A,$DF,$F1
	.dc.b	$BC,$3F,$E7,$B8,$EF,$FA,$07,$A2,$51,$D3,$0F,$1F,$A3,$F8,$C9,$CE
	.dc.b	$BE,$51,$AF,$E5,$23,$FD,$3D,$7C,$67,$39,$FE,$DF,$3F,$DC,$FE,$E9
	.dc.b	$F4,$FE,$FD,$F1,$1F,$E0,$AF,$97,$7E,$E9,$FA,$FB,$DF,$3A,$F7,$F2
	.dc.b	$8B,$DF,$A1,$FA,$EC,$BA,$6D,$EE,$3F,$D6,$0F,$E0,$7D,$2F,$73,$E1
	.dc.b	$D3,$C1,$F7,$DF,$8C,$8C,$63,$D7,$EF,$BF,$1F,$D3,$FE,$B2,$F6,$DF
	.dc.b	$87,$BF,$24,$FF,$21,$7C,$6B,$F6,$6F,$DF,$DE,$EB,$5D,$8A,$7E,$B3
	.dc.b	$AF,$AA,$FE,$2F,$C3,$E7,$3F,$06,$FF,$2C,$71,$87,$3C,$BF,$CA,$1F
	.dc.b	$5F,$E4,$AF,$E6,$39,$76,$FA,$BC,$6B,$E6,$BF,$2F,$E8,$FF,$C4,$BE
	.dc.b	$B5,$F8,$FF,$F1,$FF,$CA,$BF,$8F,$FE,$95,$FE,$D5,$FC,$07,$53,$9E
	.dc.b	$94,$FF,$B6,$7F,$66,$CF,$EA,$DF,$87,$C7,$9F,$C1,$5C,$38,$5F,$E1
	.dc.b	$E7,$39,$76,$FA,$AF,$DC,$C6,$32,$EB,$C8,$FB,$5F,$E4,$FF,$94,$BF
	.dc.b	$78,$7C,$EB,$F7,$F7,$EA,$CF,$9B,$BF,$85,$F0,$BD,$D6,$A7,$7F,$82
	.dc.b	$E7,$0F,$CF,$63,$D7,$E7,$EE,$1C,$32,$E9,$A7,$F0,$EA,$F1,$5F,$86
	.dc.b	$7C,$1B,$FF,$12,$F7,$FF,$A0,$FC,$3B,$F6,$B7,$E0,$4F,$DD,$9F,$AE
	.dc.b	$FF,$94,$3D,$2F,$EB,$38,$70,$BD,$D6,$BC,$7F,$E8,$F3,$D3,$6F,$C2
	.dc.b	$3F,$D3,$BF,$5A,$FD,$B3,$C7,$2E,$DE,$1F,$8F,$E4,$5A,$F8,$E7,$C5
	.dc.b	$F4,$B9,$8C,$67,$E4,$FC,$3F,$FA,$AF,$C3,$3E,$C4,$BF,$01,$FE,$C3
	.dc.b	$F8,$73,$F2,$77,$C4,$70,$BD,$D6,$A5,$F9,$46,$7F,$29,$FE,$19,$8F
	.dc.b	$52,$DF,$0F,$9D,$65,$FC,$27,$93,$C3,$39,$8F,$F1,$B7,$AA,$99,$7C
	.dc.b	$F3,$F0,$3F,$DB,$BF,$70,$FE,$13,$FB,$2B,$F3,$D7,$E2,$49,$7E,$6B
	.dc.b	$FD,$87,$B7,$FC,$17,$1C,$2B,$6E,$BF,$0B,$E9,$F0,$CF,$D1,$3F,$93
	.dc.b	$4F,$6D,$7F,$70,$F1,$CB,$B7,$85,$FD,$5A,$30,$7F,$29,$2E,$7F,$DA
	.dc.b	$37,$F6,$5B,$F3,$67,$ED,$DF,$CB,$3F,$44,$FD,$0F,$81,$FE,$1C,$D3
	.dc.b	$FC,$17,$5A,$F8,$E5,$6B,$EA,$9C,$7A,$C7,$FC,$5B,$97,$F0,$DC,$38
	.dc.b	$36,$3B,$79,$2C,$FF,$89,$F2,$7F,$8B,$AF,$B6,$5E,$2E,$8F,$EE,$31
	.dc.b	$8B,$E1,$9F,$F4,$2B,$FF,$1A,$FA,$AF,$9B,$FF,$19,$F5,$3F,$22,$7E
	.dc.b	$30,$FD,$5F,$F4,$D8,$70,$BD,$D6,$A7,$FD,$D3,$39,$F9,$5F,$EA,$D8
	.dc.b	$F5,$CA,$E1,$C3,$1F,$AD,$C3,$FE,$30,$AD,$67,$E1,$9F,$F2,$44,$E6
	.dc.b	$B8,$65,$FE,$FA,$CB,$C5,$8F,$E5,$33,$F1,$77,$7F,$8E,$BF,$32,$FC
	.dc.b	$4F,$B3,$FE,$E6,$F1,$FF,$A1,$7E,$8D,$FE,$80,$C7,$FC,$AB,$5A,$FC
	.dc.b	$43,$F8,$17,$C2,$29,$FA,$77,$3E,$E7,$E8,$9F,$DC,$70,$EB,$C3,$FD
	.dc.b	$0D,$E5,$9B,$DD,$FE,$E5,$18,$CB,$F5,$0D,$CB,$FC,$77,$A7,$F9,$9B
	.dc.b	$F0,$A7,$EC,$FF,$E5,$3F,$D8,$5F,$92,$3F,$53,$7E,$EC,$FC,$E7,$FC
	.dc.b	$AF,$87,$0F,$9A,$61,$C2,$F7,$FB,$16,$B2,$7D,$A7,$F8,$27,$0E,$1C
	.dc.b	$5F,$D6,$DE,$58,$71,$F6,$BF,$C7,$31,$8D,$7C,$5F,$43,$8C,$7C,$D1
	.dc.b	$E5,$D9,$FA,$DF,$F9,$AB,$F6,$07,$D6,$FF,$2C,$F9,$CF,$E0,$9F,$A1
	.dc.b	$F8,$BC,$FD,$4E,$77,$4F,$9B,$E9,$F3,$F0,$FF,$E1,$FC,$79,$F9,$F7
	.dc.b	$0E,$17,$FC,$43,$0F,$E2,$15,$E0,$B6,$5E,$0F,$82,$FF,$29,$CE,$63
	.dc.b	$C7,$C9,$F9,$DE,$9F,$F4,$BF,$E7,$FF,$EA,$5F,$AE,$FF,$62,$7E,$D1
	.dc.b	$FE,$CA,$FD,$07,$FD,$1D,$5A,$FC,$11,$5A,$F9,$05,$2F,$FD,$17,$87
	.dc.b	$FA,$BF,$F0,$7F,$F1,$C3,$D7,$7F,$F0,$53,$DF,$F0,$AE,$18,$7F,$8E
	.dc.b	$63,$17,$FE,$DA,$F7,$BF,$CD,$71,$89,$EF,$F7,$FF,$E9,$8F,$A1,$7E
	.dc.b	$6E,$FC,$F5,$F8,$9B,$F6,$E7,$E8,$1F,$CD,$7E,$57,$85,$EE,$B5,$3F
	.dc.b	$CE,$7F,$1E,$CE,$1F,$E4,$FC,$7A,$EB,$70,$E1,$97,$4E,$0F,$F3,$FD
	.dc.b	$78,$AF,$C3,$3E,$0B,$CD,$F0,$C7,$FA,$37,$C9,$7F,$AC,$3F,$4E,$FC
	.dc.b	$4F,$E6,$0F,$8D,$FF,$79,$7D,$F3,$F6,$3F,$EB,$FD,$7F,$EC,$CB,$DD
	.dc.b	$6A,$7F,$AD,$CE,$7C,$67,$E6,$7F,$E6,$7B,$EB,$C3,$FE,$39,$78,$FE
	.dc.b	$09,$FE,$90,$FC,$13,$FD,$2F,$18,$CF,$FD,$3D,$5F,$D6,$E3,$19,$3E
	.dc.b	$5B,$FD,$8F,$F8,$C3,$F1,$6E,$FE,$5B,$EC,$9F,$26,$FD,$8F,$FC,$D8
	.dc.b	$FF,$4A,$BE,$1B,$FF,$BB,$2B,$75,$FE,$88,$E1,$FF,$74,$D6,$B5,$FF
	.dc.b	$28,$3D,$FD,$DF,$F4,$13,$CB,$2F,$06,$5D,$EB,$CB,$9F,$93,$2F,$1F
	.dc.b	$DA,$FF,$AD,$30,$E1,$1E,$0F,$80,$FF,$70,$FE,$DB,$FE,$25,$F0,$9F
	.dc.b	$BD,$3F,$35,$7E,$BC,$FD,$1B,$F8,$CE,$DF,$D9,$30,$E1,$7B,$AD,$7D
	.dc.b	$27,$4F,$A8,$7E,$55,$FE,$99,$BE,$DC,$3F,$E6,$B7,$8E,$7D,$BD,$9F
	.dc.b	$F6,$4C,$63,$1F,$F0,$83,$74,$7D,$4F,$FB,$E3,$F1,$FF,$F1,$C8,$FB
	.dc.b	$77,$E6,$6F,$94,$FD,$17,$E5,$3B,$F7,$57,$BA,$D7,$98,$FC,$EF,$5F
	.dc.b	$A8,$FF,$AB,$EB,$AF,$0F,$FA,$05,$EF,$9F,$6C,$FF,$0F,$3D,$32,$EB
	.dc.b	$39,$8F,$F7,$1D,$70,$CB,$BF,$EF,$AF,$03,$F0,$F5,$FE,$29,$FD,$1B
	.dc.b	$FA,$7F,$E4,$0F,$D1,$1F,$7D,$FE,$E3,$FC,$3F,$FD,$83,$87,$0C,$18
	.dc.b	$70,$BD,$FD,$AA,$DF,$BC,$F7,$EF,$FF,$0D,$DF,$6C,$3F,$D3,$8F,$F5
	.dc.b	$EF,$C9,$BF,$D9,$CF,$3C,$FF,$DE,$5E,$7F,$FB,$F2,$31,$97,$C0,$6E
	.dc.b	$4F,$DB,$3B,$63,$FE,$43,$C1,$FA,$7B,$DC,$7C,$03,$EE,$5F,$B5,$BF
	.dc.b	$78,$7E,$55,$CB,$2F,$83,$8F,$1C,$38,$5E,$FE,$E7,$59,$F6,$F7,$7F
	.dc.b	$DC,$CF,$83,$4B,$87,$1D,$3B,$FC,$6F,$F9,$5F,$9F,$E5,$5E,$F3,$EB
	.dc.b	$F6,$99,$E6,$FF,$20,$FA,$4F,$D2,$A7,$C5,$FF,$89,$BF,$58,$F6,$7A
	.dc.b	$3E,$7A,$D5,$2B,$5B,$FF,$76,$D3,$EC,$56,$B8,$FF,$D9,$8F,$5F,$D0
	.dc.b	$BF,$DF,$CF,$2B,$ED,$EA,$F8,$EB,$FC,$DD,$39,$FA,$E6,$9F,$D8,$E7
	.dc.b	$FD,$B3,$9F,$9A,$73,$9B,$F6,$BF,$99,$F8,$6B,$F7,$87,$E8,$BF,$C5
	.dc.b	$F3,$7C,$83,$A9,$FA,$67,$0B,$DF,$C0,$FF,$B9,$AB,$5F,$A1,$67,$C4
	.dc.b	$FF,$33,$F0,$7F,$9A,$39,$7F,$2D,$AF,$7D,$73,$3E,$B1,$19,$C7,$A7
	.dc.b	$1E,$63,$1B,$3E,$87,$A7,$47,$F9,$D2,$73,$AF,$9F,$E1,$BF,$11,$FB
	.dc.b	$47,$F1,$AF,$B5,$7E,$BA,$F9,$27,$EE,$AF,$C8,$BD,$DC,$5F,$AB,$C3
	.dc.b	$85,$EE,$B5,$F4,$BD,$3C,$0F,$F3,$47,$1F,$AB,$0E,$11,$FE,$7E,$8C
	.dc.b	$6F,$FE,$88,$C7,$9E,$7E,$03,$E0,$38,$C5,$E7,$63,$0F,$F3,$A6,$7E
	.dc.b	$1D,$1F,$90,$FE,$8F,$D7,$7E,$4A,$FC,$79,$FB,$13,$F6,$7F,$F7,$97
	.dc.b	$ED,$5F,$DA,$78,$BF,$83,$61,$C2,$F7,$5A,$D7,$A6,$2F,$D3,$BE,$AF
	.dc.b	$1E,$2E,$E3,$CA,$BC,$19,$F7,$FC,$5B,$F2,$19,$F5,$8C,$7D,$4F,$DD
	.dc.b	$8C,$69,$E4,$F8,$DF,$DE,$7E,$2F,$FC,$BF,$F7,$F7,$E4,$4F,$C3,$FF
	.dc.b	$B6,$7E,$38,$FD,$7B,$7F,$14,$BD,$D6,$BF,$6F,$CB,$E6,$EE,$FF,$5B
	.dc.b	$E5,$D7,$93,$E9,$9F,$ED,$57,$C1,$9F,$7F,$99,$FD,$44,$62,$FF,$BE
	.dc.b	$78,$FF,$95,$8C,$4B,$57,$EA,$5F,$C5,$7E,$87,$FC,$31,$FA,$AF,$E9
	.dc.b	$FD,$97,$CD,$BF,$3F,$FF,$02,$75,$FF,$37,$BD,$FA,$9F,$ED,$7A,$D7
	.dc.b	$E3,$3C,$FE,$9B,$F2,$BC,$5E,$81,$7D,$9C,$3C,$B2,$EF,$F9,$EF,$EA
	.dc.b	$E3,$18,$7F,$AB,$B2,$F0,$60,$D3,$CD,$3E,$5F,$DA,$3F,$97,$77,$3E
	.dc.b	$E5,$EB,$3F,$0B,$7E,$7B,$FE,$76,$E1,$7E,$87,$C2,$F7,$5A,$FB,$BF
	.dc.b	$E1,$1F,$7D,$9F,$C7,$BE,$A7,$20,$FC,$BD,$F6,$7D,$DA,$78,$74,$F0
	.dc.b	$69,$FE,$84,$EB,$FE,$0D,$39,$CF,$C7,$FB,$9F,$F0,$2C,$3A,$4E,$6F
	.dc.b	$CD,$FB,$AF,$F4,$9F,$CC,$5F,$B6,$BF,$74,$7E,$E8,$FC,$D9,$F8,$53
	.dc.b	$E1,$5F,$A7,$F1,$C3,$85,$EE,$B5,$F9,$96,$71,$F6,$EC,$CF,$BF,$E3
	.dc.b	$F8,$7C,$33,$E0,$D3,$BF,$AF,$FC,$86,$73,$97,$5F,$7F,$1A,$78,$2F
	.dc.b	$CB,$B3,$F2,$EF,$EC,$9F,$8C,$BF,$3F,$7E,$DC,$FC,$9D,$F3,$8F,$D0
	.dc.b	$7F,$9D,$FE,$BB,$C2,$F7,$5A,$9F,$E7,$FF,$9B,$67,$0F,$5E,$3D,$7F
	.dc.b	$2B,$B8,$70,$CB,$A7,$C7,$FF,$39,$AF,$15,$F8,$67,$C1,$FB,$06,$6F
	.dc.b	$86,$5F,$D3,$3F,$4D,$FE,$F9,$E3,$7F,$8B,$FF,$1A,$FF,$36,$7C,$0F
	.dc.b	$F8,$8F,$EC,$46,$BD,$F4,$EB,$D1,$CF,$EA,$19,$FC,$53,$E8,$E3,$CF
	.dc.b	$C0,$FD,$43,$D7,$F0,$5F,$F2,$25,$FF,$65,$8C,$67,$FE,$31,$F8,$6F
	.dc.b	$EA,$51,$89,$F0,$FD,$EF,$F8,$47,$CA,$FF,$26,$7A,$9F,$ED,$EF,$D5
	.dc.b	$DF,$BC,$3F,$48,$7C,$BE,$F4,$FE,$03,$F8,$5A,$F7,$5F,$4F,$3F,$F1
	.dc.b	$C6,$5D,$37,$FD,$F3,$FE,$53,$6E,$1C,$AF,$BF,$E7,$1F,$D9,$EB,$C7
	.dc.b	$8F,$5C,$DF,$8E,$E7,$0E,$11,$E5,$FC,$BF,$FC,$55,$F8,$7B,$B1,$F1
	.dc.b	$1F,$E2,$FF,$D1,$59,$FF,$52,$F6,$37,$5A,$FB,$45,$6B,$D0,$7C,$06
	.dc.b	$B1,$FD,$0E,$5D,$7E,$0E,$E3,$C7,$0E,$1C,$7B,$CB,$9E,$9F,$E7,$58
	.dc.b	$FF,$1F,$7E,$1C,$FC,$F1,$97,$E6,$BD,$27,$E9,$DF,$D9,$BF,$3B,$DD
	.dc.b	$6B,$C9,$FE,$F9,$3F,$DA,$BF,$2F,$CE,$3F,$B4,$3A,$63,$6F,$86,$3D
	.dc.b	$38,$BF,$C2,$55,$AC,$79,$FE,$09,$8B,$DE,$5F,$DD,$BF,$5D,$FF,$2A
	.dc.b	$7E,$54,$FD,$69,$E2,$3F,$2D,$7E,$52,$FD,$07,$F8,$8F,$D5,$F0,$BD
	.dc.b	$D6,$A7,$F8,$AE,$F9,$C3,$F8,$5E,$3D,$79,$6E,$1C,$32,$E9,$7F,$E2
	.dc.b	$9A,$F1,$5F,$86,$7C,$1B,$7F,$8B,$79,$1F,$8F,$7F,$AF,$3F,$0A,$FE
	.dc.b	$7B,$F0,$9F,$C1,$BF,$80,$3F,$CD,$D5,$AF,$AB,$FF,$54,$FC,$6A,$BF
	.dc.b	$85,$FC,$E6,$B1,$7E,$57,$FC,$83,$0E,$1D,$BF,$D5,$DE,$59,$F6,$F1
	.dc.b	$3C,$57,$E1,$CB,$A7,$A1,$8D,$3B,$E9,$FC,$63,$F6,$AF,$F4,$07,$4F
	.dc.b	$F8,$E3,$F7,$7F,$C7,$FE,$52,$FD,$51,$F3,$AF,$D1,$37,$F6,$2B,$DD
	.dc.b	$6B,$F0,$4C,$FE,$8B,$FC,$DB,$2E,$9C,$9F,$DB,$5E,$DE,$73,$4E,$DC
	.dc.b	$7F,$E5,$59,$CC,$78,$78,$71,$A7,$7B,$E1,$9F,$93,$EE,$7F,$D3,$1F
	.dc.b	$A9,$BF,$50,$7E,$7F,$FE,$5E,$F8,$67,$D8,$FA,$BF,$36,$DD,$6B,$EF
	.dc.b	$7F,$9C,$4F,$BB,$F7,$5C,$F6,$7F,$4E,$CB,$A6,$57,$0E,$19,$74,$E2
	.dc.b	$FF,$36,$56,$B2,$E7,$C1,$8B,$DE,$7F,$2B,$F5,$8F,$F5,$77,$ED,$1F
	.dc.b	$CB,$BF,$6E,$FF,$45,$FE,$23,$F8,$1F,$CB,$BF,$D6,$FF,$09,$FE,$CF
	.dc.b	$AD,$7E,$19,$FE,$B8,$F0,$55,$EE,$AD,$7C,$6B,$F2,$8F,$EC,$B8,$75
	.dc.b	$ED,$FF,$0E,$3C,$B2,$FC,$77,$B3,$FE,$83,$8C,$65,$FD,$F6,$9F,$BB
	.dc.b	$3F,$79,$FE,$35,$FB,$67,$F9,$83,$D4,$7E,$6A,$F0,$FF,$D9,$17,$BF
	.dc.b	$82,$5E,$E9,$D3,$AC,$5F,$98,$7F,$7A,$C3,$86,$7F,$E3,$17,$96,$1D
	.dc.b	$FE,$43,$FD,$31,$18,$FA,$BF,$E0,$5F,$1D,$8C,$7C,$B1,$FC,$17,$37
	.dc.b	$07,$FB,$6F,$CF,$FC,$1F,$E8,$5F,$E2,$DF,$C9,$FF,$0D,$FA,$FB,$F2
	.dc.b	$7D,$78,$AA,$D7,$82,$A7,$ED,$9A,$F7,$3F,$C2,$39,$74,$E3,$FF,$24
	.dc.b	$3D,$BE,$C5,$A7,$E4,$FA,$BF,$D6,$13,$98,$FC,$B7,$D9,$E3,$5F,$05
	.dc.b	$F0,$D3,$C1,$FA,$FF,$FB,$CF,$F1,$5F,$0F,$E6,$4F,$75,$F8,$1F,$E7
	.dc.b	$FF,$46,$7E,$9D,$BE,$25,$EE,$B5,$19,$E4,$FF,$14,$E3,$D3,$8B,$FC
	.dc.b	$C0,$F6,$CB,$AE,$8C,$FF,$D9,$13,$9C,$7F,$5D,$F2,$7F,$EC,$BF,$A6
	.dc.b	$7C,$E4,$FD,$B5,$FA,$2B,$F5,$BF,$EB,$3F,$BD,$BC,$EF,$4E,$FF,$E5
	.dc.b	$5C,$78,$E2,$C7,$8E,$1F,$D8,$3F,$75,$E1,$A7,$F9,$6F,$0F,$F1,$8D
	.dc.b	$F8,$1F,$F3,$C6,$1E,$2F,$DB,$FF,$C8,$AF,$4C,$79,$FB,$1E,$7B,$F3
	.dc.b	$6A,$FC,$E7,$F4,$17,$B8,$F4,$7C,$03,$F4,$6F,$0E,$7F,$E6,$BF,$B1
	.dc.b	$7E,$4D,$FE,$94,$FC,$D7,$E9,$FE,$53,$FE,$5D,$F1,$B7,$6B,$DF,$E1
	.dc.b	$AB,$DF,$DD,$BF,$56,$DD,$6A,$FF,$D2,$4F,$6F,$D7,$7F,$CB,$EF,$2B
	.dc.b	$EF,$F8,$C7,$FB,$EE,$31,$CC,$D3,$FA,$56,$1D,$35,$F3,$4E,$73,$7E
	.dc.b	$59,$E6,$FC,$3B,$FC,$B3,$F8,$C3,$F1,$07,$90,$FA,$97,$E6,$7F,$D0
	.dc.b	$7C,$0F,$B3,$AB,$1E,$38,$70,$BD,$FC,$A7,$5F,$A8,$7E,$EF,$FE,$78
	.dc.b	$BE,$F8,$7F,$AB,$DE,$39,$F7,$FC,$73,$FD,$31,$9F,$69,$CE,$5F,$E1
	.dc.b	$4B,$E9,$AF,$9F,$0F,$36,$AF,$93,$FA,$3F,$B5,$7E,$3F,$F2,$8F,$F3
	.dc.b	$D7,$C8,$FF,$42,$F2,$BF,$B5,$EB,$EA,$AD,$7E,$0C,$AD,$79,$9C,$A9
	.dc.b	$FB,$9F,$FA,$42,$FA,$6D,$FE,$C7,$78,$DF,$58,$E1,$97,$83,$E4,$9F
	.dc.b	$94,$7B,$3A,$BF,$35,$87,$09,$FF,$1F,$F8,$FF,$88,$F9,$5F,$F6,$17
	.dc.b	$EE,$AF,$C0,$1E,$83,$F3,$E7,$CC,$7F,$5F,$E1,$FA,$78,$70,$BD,$D6
	.dc.b	$BE,$2F,$A7,$BA,$FD,$6F,$FD,$53,$7D,$B8,$BF,$DC,$0F,$1C,$FB,$67
	.dc.b	$E7,$9F,$F6,$54,$E7,$17,$A7,$99,$FF,$39,$FA,$3F,$EC,$FF,$DA,$9F
	.dc.b	$86,$3F,$74,$7E,$FE,$FD,$AB,$F4,$BF,$82,$7F,$06,$C5,$87,$0B,$DF
	.dc.b	$E1,$BA,$CB,$F8,$05,$4E,$4F,$C8,$E1,$FC,$3B,$F4,$EF,$F6,$1B,$CF
	.dc.b	$1E,$5F,$9C,$79,$63,$1A,$F8,$FF,$21,$C6,$5E,$78,$F3,$6D,$F3,$B2
	.dc.b	$FC,$EF,$CA,$7F,$45,$F8,$7F,$EC,$7F,$52,$FC,$95,$F9,$83,$F3,$EF
	.dc.b	$DF,$DE,$FE,$67,$FC,$39,$5A,$F5,$F3,$F2,$BF,$D3,$F9,$3D,$66,$33
	.dc.b	$FF,$6D,$E1,$C7,$37,$B9,$FD,$CE,$31,$8F,$D0,$BF,$E1,$BF,$B9,$BE
	.dc.b	$89,$FB,$4F,$F9,$87,$E1,$5F,$AB,$BA,$BF,$DE,$7C,$7F,$B8,$C3,$85
	.dc.b	$EE,$B5,$ED,$74,$C5,$E2,$FF,$BB,$AF,$79,$FA,$31,$E5,$5E,$0F,$BF
	.dc.b	$FE,$4D,$8F,$F8,$26,$FA,$C6,$3E,$97,$F0,$D1,$8D,$7C,$BA,$BF,$27
	.dc.b	$FE,$4F,$FA,$C7,$E2,$3E,$2B,$F4,$FF,$05,$FD,$03,$F6,$0C,$F7,$7F
	.dc.b	$73,$F5,$3F,$8B,$70,$7F,$62,$FD,$83,$3F,$0B,$F3,$38,$F7,$95,$AF
	.dc.b	$E2,$F9,$F5,$C3,$F8,$F7,$D3,$FF,$87,$E3,$1A,$7F,$55,$F2,$FF,$39
	.dc.b	$FB,$AF,$F3,$EF,$D7,$9F,$86,$FD,$FF,$B1,$7E,$01,$FF,$00,$74,$FE
	.dc.b	$47,$7B,$AD,$4F,$FB,$1A,$73,$F1,$8F,$CE,$BC,$9E,$F7,$F5,$BC,$3B
	.dc.b	$70,$FC,$D8,$C5,$33,$F0,$F9,$2D,$1F,$17,$E6,$AD,$6B,$E5,$F4,$31
	.dc.b	$87,$0B,$F3,$FE,$E9,$FD,$53,$EA,$7F,$AA,$7D,$8B,$F5,$7F,$F0,$67
	.dc.b	$E6,$FF,$F0,$03,$F7,$7D,$7C,$AF,$F9,$2E,$73,$F6,$AF,$AF,$CE,$1E
	.dc.b	$AE,$31,$FE,$44,$BF,$E3,$EF,$AC,$CE,$5E,$0C,$3B,$FD,$8F,$F0,$98
	.dc.b	$C7,$D3,$34,$FE,$3B,$7F,$DA,$BE,$03,$FD,$7D,$F8,$A7,$F8,$27,$D5
	.dc.b	$FC,$E7,$E5,$1F,$EA,$EE,$CF,$EA,$75,$A7,$FC,$6F,$E4,$A7,$FC,$CD
	.dc.b	$D7,$FE,$09,$AE,$78,$F9,$4F,$E5,$9F,$87,$3D,$5F,$64,$63,$3F,$06
	.dc.b	$2F,$CA,$7F,$90,$61,$DF,$4F,$D7,$7E,$17,$8C,$98,$70,$D1,$FB,$A7
	.dc.b	$F8,$13,$F6,$87,$E8,$A7,$FD,$B7,$F9,$8F,$FA,$B7,$F3,$67,$C8,$EE
	.dc.b	$B5,$F8,$8A,$B5,$F2,$19,$E9,$F3,$E3,$D0,$B7,$C2,$BF,$D7,$1F,$1F
	.dc.b	$FC,$DE,$75,$87,$3F,$C2,$FF,$86,$74,$7F,$7C,$9C,$DF,$8F,$F7,$5F
	.dc.b	$F0,$DE,$5F,$CD,$BB,$BF,$C4,$3F,$8B,$3F,$59,$7E,$92,$F4,$F7,$F2
	.dc.b	$1F,$E2,$1F,$35,$7B,$A7,$A1,$AF,$A6,$F9,$F9,$F5,$F0,$3E,$A3,$DF
	.dc.b	$0E,$39,$F9,$BC,$2F,$F2,$14,$E7,$27,$EB,$5F,$E9,$98,$C5,$79,$3C
	.dc.b	$6F,$F1,$9F,$E1,$1F,$E7,$DF,$CB,$1F,$87,$FF,$87,$7E,$23,$FC,$D3
	.dc.b	$FA,$EE,$BE,$F1,$5A,$FA,$CF,$E4,$7F,$7C,$FC,$B7,$53,$F8,$6C,$BF
	.dc.b	$D1,$1F,$31,$70,$E3,$9F,$5E,$3F,$EC,$D5,$E1,$8F,$07,$C5,$E2,$B7
	.dc.b	$9F,$6F,$AA,$FF,$24,$FE,$DE,$FD,$9D,$F1,$FF,$E4,$BF,$DF,$DF,$3C
	.dc.b	$FD,$B1,$E1,$70,$BD,$F8,$4B,$DD,$7F,$80,$7C,$5F,$F1,$CE,$5D,$7E
	.dc.b	$AF,$F3,$EF,$7C,$EE,$3C,$B0,$EF,$D7,$FD,$FE,$B5,$97,$5D,$3F,$51
	.dc.b	$C7,$E1,$C7,$CD,$1E,$5F,$B0,$C5,$6F,$5F,$E3,$3A,$FF,$92,$BD,$EF
	.dc.b	$F1,$87,$94,$FD,$D5,$F9,$C3,$EA,$FF,$A3,$78,$3F,$E6,$CB,$DD,$6A
	.dc.b	$7F,$A5,$78,$59,$E4,$63,$FB,$AB,$FF,$1E,$7E,$45,$67,$38,$71,$F6
	.dc.b	$3F,$C3,$55,$D3,$3F,$07,$E4,$5F,$E1,$DF,$CC,$1F,$86,$3F,$3F,$FE
	.dc.b	$87,$F9,$13,$F5,$B7,$EF,$6C,$BF,$D0,$38,$70,$BD,$D6,$BA,$B3,$8F
	.dc.b	$F8,$2B,$AB,$F8,$95,$ED,$B1,$DF,$E3,$D9,$F7,$F8,$FF,$24,$E6,$3C
	.dc.b	$5F,$21,$8C,$FB,$EA,$CB,$FD,$27,$FA,$47,$F9,$2F,$C2,$FF,$39,$FE
	.dc.b	$CC,$FD,$71,$FB,$5B,$F7,$DD,$FD,$C7,$E3,$3E,$A1,$7B,$A7,$E3,$3A
	.dc.b	$C7,$F9,$3E,$5F,$E5,$7F,$12,$E3,$C7,$0E,$DF,$67,$FE,$47,$AF,$16
	.dc.b	$5D,$34,$71,$E7,$0E,$11,$E4,$F2,$7B,$56,$F5,$FE,$D9,$EC,$FF,$B0
	.dc.b	$FF,$3A,$FF,$7D,$7D,$3B,$E3,$5F,$AC,$3F,$6E,$FE,$29,$FA,$57,$FA
	.dc.b	$D2,$F7,$5A,$9F,$E3,$1F,$BC,$E7,$ED,$3F,$C2,$9D,$FF,$D1,$6D,$0F
	.dc.b	$D4,$BF,$A2,$C7,$2C,$BB,$7C,$C7,$F9,$8E,$31,$8B,$D4,$C6,$1F,$E1
	.dc.b	$FC,$FC,$1F,$12,$FD,$37,$FD,$8F,$F1,$CF,$CA,$DF,$A2,$BF,$4B,$7E
	.dc.b	$D6,$FD,$73,$F2,$5F,$A8,$F0,$BD,$D6,$A7,$F2,$DF,$8B,$67,$0F,$EB
	.dc.b	$78,$F5,$E7,$B8,$70,$CB,$A7,$D6,$FF,$9E,$EB,$C5,$7E,$19,$7E,$F7
	.dc.b	$FE,$F1,$F7,$3F,$D8,$35,$AD,$3C,$1E,$5B,$F1,$6F,$F7,$0F,$82,$FD
	.dc.b	$7F,$EE,$FD,$AF,$FB,$AF,$F2,$3F,$F6,$6F,$D2,$6A,$73,$F7,$49,$FF
	.dc.b	$68,$FC,$1E,$7F,$78,$FE,$E7,$8F,$3F,$2A,$E1,$C2,$DF,$19,$FE,$93
	.dc.b	$AD,$67,$DF,$C4,$C5,$EF,$F1,$4F,$EE,$1C,$5F,$ED,$AB,$DE,$BE,$19
	.dc.b	$FB,$13,$FE,$DE,$FD,$A5,$FB,$17,$E8,$BF,$6F,$7E,$4F,$FD,$DB,$F0
	.dc.b	$CF,$C8,$5E,$EB,$53,$9C,$13,$9F,$53,$FC,$1B,$8F,$4F,$D4,$6E,$1C
	.dc.b	$3E,$A1,$97,$FA,$B6,$BC,$36,$CF,$C3,$4F,$D1,$F5,$87,$1C,$1F,$99
	.dc.b	$F8,$3D,$2F,$F1,$AF,$13,$EA,$DF,$7F,$FE,$0B,$FC,$61,$FA,$F3,$EB
	.dc.b	$1F,$AC,$FF,$8C,$F8,$FF,$DF,$F8,$70,$FA,$CF,$F7,$1D,$EE,$BF,$9D
	.dc.b	$E7,$DA,$B5,$9F,$F8,$B5,$EF,$CD,$FE,$2C,$79,$67,$E0,$CD,$F9,$7F
	.dc.b	$FB,$2E,$73,$97,$8F,$F0,$EC,$6D,$E7,$8F,$07,$EB,$3E,$DC,$FF,$CA
	.dc.b	$3F,$B5,$FF,$B3,$7F,$3C,$7E,$CA,$FD,$6D,$F8,$07,$EC,$FE,$73,$86
	.dc.b	$0C,$38,$5E,$EB,$F7,$1F,$82,$EA,$7F,$DB,$1F,$B9,$7F,$91,$1E,$D7
	.dc.b	$FE,$49,$78,$E3,$DB,$DD,$FF,$72,$46,$30,$FF,$2D,$F4,$39,$BE,$9C
	.dc.b	$E6,$BF,$D0,$93,$9C,$3F,$CD,$7B,$FF,$79,$FD,$65,$F9,$5B,$F4,$3F
	.dc.b	$67,$27,$E0,$F0,$E1,$87,$DA,$F8,$A6,$1C,$2F,$7F,$3B,$F0,$32,$FF
	.dc.b	$02,$74,$7F,$97,$9E,$D8,$7F,$98,$DF,$06,$7D,$F8,$3F,$DA,$73,$C7
	.dc.b	$2E,$B3,$9B,$FF,$40,$E7,$E2,$C3,$8E,$5E,$6F,$DE,$3E,$7F,$E5,$5F
	.dc.b	$E8,$FF,$D6,$9F,$BF,$BF,$3E,$FE,$11,$FA,$FF,$F9,$B7,$2F,$5D,$EE
	.dc.b	$B5,$EC,$2B,$FC,$0D,$5A,$FC,$2D,$8F,$F9,$E3,$0F,$F4,$E7,$0D,$9C
	.dc.b	$E3,$DF,$F6,$FF,$E3,$5D,$1F,$A4,$BE,$DA,$7F,$19,$F8,$1F,$5C,$FE
	.dc.b	$A3,$F4,$87,$ED,$AF,$A6,$7E,$F4,$F1,$1F,$B8,$BF,$68,$FF,$18,$71
	.dc.b	$7D,$4C,$78,$E1,$C2,$F7,$5F,$DB,$EB,$58,$BF,$34,$FF,$48,$E1,$C3
	.dc.b	$B3,$FD,$2C,$F2,$CF,$C1,$9F,$F2,$09,$FC,$DC,$E7,$2F,$1E,$7E,$A9
	.dc.b	$FF,$5A,$CE,$6B,$CD,$E9,$70,$CB,$D9,$8F,$4D,$3F,$DB,$9A,$FF,$08
	.dc.b	$D9,$FA,$D4,$FC,$95,$FA,$7B,$ED,$5F,$95,$3F,$18,$FD,$6E,$1C,$30
	.dc.b	$61,$C2,$F7,$F5,$BA,$9F,$9B,$C3,$FE,$B3,$BE,$0D,$8F,$06,$4F,$C2
	.dc.b	$7E,$CC,$FF,$B4,$2B,$B6,$3D,$5F,$81,$07,$35,$E7,$CF,$FD,$B5,$CD
	.dc.b	$FA,$8F,$CE,$9F,$B0,$7F,$B8,$FF,$4B,$7E,$10,$F6,$1F,$84,$FF,$AE
	.dc.b	$30,$E1,$7B,$AD,$4B,$F7,$DF,$E1,$B8,$F4,$E0,$FF,$65,$BD,$BC,$2B
	.dc.b	$87,$1C,$BF,$5B,$97,$49,$67,$E2,$97,$E4,$5E,$CC,$39,$4E,$73,$7E
	.dc.b	$FB,$FC,$73,$F1,$17,$C5,$3D,$1F,$57,$E5,$6F,$A9,$7E,$CF,$FE,$88
	.dc.b	$DF,$F8,$4E,$1C,$2F,$75,$AF,$C5,$FA,$63,$CF,$A3,$FD,$D1,$7D,$BE
	.dc.b	$2B,$FD,$CE,$F1,$E1,$69,$DF,$E4,$BE,$9C,$E6,$3C,$5E,$3E,$31,$F3
	.dc.b	$5F,$0C,$FC,$BF,$65,$E0,$D7,$C7,$BF,$F4,$ED,$47,$F9,$CF,$F0,$47
	.dc.b	$E9,$EF,$C3,$7F,$C8,$5E,$EF,$BB,$F6,$1F,$E7,$FF,$3B,$FE,$D7,$AD
	.dc.b	$7D,$52,$BE,$1F,$C8,$EB,$1F,$F7,$BE,$5F,$DA,$0F,$C5,$E1,$C5,$C1
	.dc.b	$BC,$39,$DF,$EC,$F9,$FF,$99,$D7,$7C,$FC,$7C,$5C,$B1,$F3,$C7,$9B
	.dc.b	$57,$F0,$0F,$F5,$4F,$0D,$FB,$07,$D9,$FD,$81,$FB,$1B,$E8,$9F,$8F
	.dc.b	$7D,$ED,$BF,$A0,$E1,$C3,$95,$87,$0B,$DF,$DE,$F5,$CB,$FC,$15,$C7
	.dc.b	$EA,$C7,$8C,$7D,$D8,$F8,$11,$8C,$39,$E7,$E1,$FC,$B3,$F5,$B9,$F7
	.dc.b	$CF,$B7,$E2,$3F,$B0,$F8,$27,$BE,$FC,$0F,$F9,$47,$DA,$7A,$3E,$A9
	.dc.b	$FB,$2F,$F7,$2B,$DD,$6B,$8D,$4F,$DA,$F5,$8F,$3E,$1F,$47,$B5,$FE
	.dc.b	$1C,$C7,$BF,$E4,$FF,$52,$77,$97,$3C,$7C,$98,$78,$FA,$FF,$CF,$79
	.dc.b	$FF,$05,$BD,$C6,$29,$F8,$EF,$F8,$1B,$3F,$F6,$A7,$EB,$1F,$AB,$F4
	.dc.b	$DF,$94,$BF,$04,$FF,$30,$74,$FF,$72,$BD,$F4,$FF,$0C,$AD,$7D,$BF
	.dc.b	$3F,$C4,$19,$79,$F0,$E1,$1F,$C0,$36,$F6,$9F,$9E,$C7,$9E,$7E,$0D
	.dc.b	$7F,$8E,$C6,$32,$FF,$47,$57,$5C,$7C,$D1,$E5,$E0,$FF,$0F,$FD,$AF
	.dc.b	$F4,$DF,$D5,$03,$F3,$8F,$F4,$7F,$B3,$FE,$71,$DB,$FC,$17,$5B,$AD
	.dc.b	$7D,$C7,$F1,$2F,$2B,$9F,$C6,$FF,$B8,$75,$F3,$9F,$E9,$67,$AB,$8B
	.dc.b	$E5,$97,$68,$E3,$87,$8A,$3F,$D3,$FE,$5F,$F8,$2A,$BF,$CA,$FF,$99
	.dc.b	$BF,$2B,$3F,$CC,$BC,$D7,$E3,$8F,$8F,$7E,$6B,$FE,$6E,$C3,$8E,$1C
	.dc.b	$3D,$EF,$F7,$C5,$EF,$F5,$2D,$6D,$FC,$2B,$3E,$DF,$B2,$FD,$3C,$78
	.dc.b	$C5,$9D,$63,$CF,$4F,$0F,$C4,$79,$E7,$35,$E5,$CB,$FC,$7F,$87,$F9
	.dc.b	$C3,$E2,$DE,$57,$F6,$5F,$96,$7D,$A3,$F6,$8F,$F3,$8F,$EF,$3E,$4C
	.dc.b	$D8,$70,$F9,$1F,$F1,$3D,$EF,$E7,$FF,$CB,$9D,$9F,$EC,$4B,$FE,$47
	.dc.b	$7B,$7F,$79,$A7,$87,$47,$E6,$5E,$B9,$EB,$1E,$3C,$7C,$5C,$1E,$B9
	.dc.b	$CE,$7E,$0D,$7D,$71,$FC,$C3,$F2,$D7,$E5,$C5,$F9,$03,$ED,$7F,$D1
	.dc.b	$5F,$64,$F8,$DF,$65,$70,$37,$3A,$FB,$8C,$E1,$F9,$AC,$7F,$88,$F0
	.dc.b	$7F,$A6,$2F,$84,$7B,$0F,$14,$F8,$72,$F0,$7D,$A7,$9A,$7C,$93,$9C
	.dc.b	$7F,$D3,$9B,$7F,$48,$FB,$4F,$E7,$9F,$A2,$7F,$6E,$FD,$29,$FB,$6B
	.dc.b	$F7,$57,$E7,$EF,$D4,$DF,$9A,$FF,$AA,$35,$FF,$55,$E3,$C7,$0E,$17
	.dc.b	$BF,$0B,$55,$FD,$5A,$BB,$3F,$D2,$33,$F0,$FC,$CB,$F2,$AF,$E3,$11
	.dc.b	$8D,$7C,$7A,$F8,$BF,$52,$F4,$61,$E0,$C1,$5A,$CD,$FB,$16,$AB,$C1
	.dc.b	$F7,$8F,$D4,$1F,$B0,$3E,$11,$FA,$27,$F9,$67,$E0,$FC,$F8,$38,$7F
	.dc.b	$D5,$38,$70,$BD,$FD,$7F,$F9,$57,$87,$E4,$E7,$FE,$C7,$C3,$D8,$7C
	.dc.b	$47,$C0,$E0,$F1,$DF,$3C,$BF,$9D,$57,$28,$C7,$E1,$BE,$9A,$F1,$78
	.dc.b	$FF,$4F,$F6,$5F,$F0,$5F,$E5,$EF,$D5,$BF,$BF,$63,$F6,$AB,$5F,$42
	.dc.b	$AD,$7E,$57,$EC,$E0,$FD,$5F,$E8,$3F,$3B,$3F,$B3,$87,$DF,$7C,$C7
	.dc.b	$83,$A5,$A7,$83,$A7,$C9,$87,$ED,$27,$39,$F7,$E6,$CF,$C1,$3F,$7A
	.dc.b	$FF,$31,$FE,$1B,$FD,$E3,$A1,$FA,$37,$8E,$B5,$F2,$CA,$D7,$E2,$79
	.dc.b	$FD,$E7,$CD,$C1,$F8,$9B,$E8,$F9,$B6,$FE,$A3,$9F,$6F,$9A,$FF,$4A
	.dc.b	$4F,$58,$C7,$C8,$7A,$3F,$52,$FF,$38,$FD,$2F,$E7,$BE,$E7,$FC,$C9
	.dc.b	$FB,$C7,$C5,$9C,$F9,$3F,$E9,$EF,$31,$F9,$EE,$7E,$DF,$C7,$A7,$E1
	.dc.b	$6F,$FD,$86,$FF,$89,$7B,$FF,$57,$1C,$B2,$FE,$D9,$18,$A6,$7F,$AE
	.dc.b	$FD,$DF,$FC,$73,$87,$D5,$EF,$FC,$A7,$A0,$FD,$FF,$FD,$BF,$F4,$6F
	.dc.b	$D1,$F5,$E5,$FF,$0D,$9C,$E1,$83,$E2,$61,$FE,$B3,$6C,$EF,$F0,$9F
	.dc.b	$F5,$AC,$72,$C7,$B4,$7E,$EF,$5F,$E0,$77,$E2,$C3,$C3,$FA,$0F,$E2
	.dc.b	$78,$9F,$59,$FD,$A3,$E6,$BF,$1D,$F7,$6F,$27,$53,$9F,$2B,$FD,$13
	.dc.b	$E3,$F1,$FC,$4B,$E4,$F7,$FF,$A4,$EA,$E5,$D3,$E5,$7F,$C3,$F8,$7F
	.dc.b	$46,$E2,$E7,$18,$FD,$7B,$F2,$35,$BD,$7C,$3F,$88,$FE,$E3,$A9,$F6
	.dc.b	$4F,$B9,$7E,$96,$FD,$6D,$F2,$1F,$E4,$AC,$F8,$5E,$EB,$53,$FE,$4E
	.dc.b	$FE,$01,$CE,$1F,$80,$C7,$AF,$E1,$D7,$0E,$19,$74,$FC,$D3,$F7,$15
	.dc.b	$E2,$BF,$0C,$F8,$3D,$8C,$FD,$0F,$F9,$67,$F5,$67,$77,$F7,$8E,$37
	.dc.b	$E3,$DE,$BA,$D4,$E7,$E3,$F1,$F1,$9E,$39,$FF,$31,$73,$FE,$E3,$54
	.dc.b	$70,$F9,$E7,$EC,$9E,$0F,$B7,$6B,$E0,$9F,$E0,$D8,$7F,$20,$F9,$C7
	.dc.b	$DA,$7F,$B6,$79,$5F,$A1,$3F,$4E,$7E,$FE,$F4,$B8,$5E,$EB,$53,$DB
	.dc.b	$8B,$FD,$11,$8F,$4F,$99,$FD,$4B,$DB,$E7,$2E,$1C,$72,$7C,$27,$F2
	.dc.b	$1A,$E1,$A7,$8A,$7F,$9E,$EB,$E7,$FB,$C7,$E8,$7F,$03,$C8,$7E,$A6
	.dc.b	$FE,$00,$7E,$A1,$BA,$D4,$E7,$8B,$18,$FF,$20,$D3,$FD,$87,$FB,$AF
	.dc.b	$9B,$3E,$BD,$3F,$D0,$F0,$F0,$F1,$FB,$27,$3A,$F8,$38,$3D,$11,$8A
	.dc.b	$F1,$7E,$71,$F3,$7F,$3D,$7E,$F7,$FE,$D4,$F3,$BF,$CF,$BF,$77,$FD
	.dc.b	$3F,$3F,$8F,$5A,$FC,$2B,$F9,$D7,$9F,$FE,$72,$F0,$79,$63,$F2,$A9
	.dc.b	$F2,$B6,$3A,$57,$F2,$2C,$39,$C6,$3F,$0C,$F9,$9E,$57,$F1,$CF,$A2
	.dc.b	$7E,$65,$FD,$93,$B1,$F9,$A7,$E0,$EB,$53,$9F,$CF,$31,$ED,$FD,$A9
	.dc.b	$FE,$4B,$D7,$F2,$5D,$FC,$E3,$F1,$AE,$FF,$13,$FE,$24,$C7,$FB,$8F
	.dc.b	$57,$DD,$FA,$2F,$F2,$57,$D8,$7F,$94,$F8,$BF,$0F,$18,$FB,$5F,$F2
	.dc.b	$DC,$63,$EC,$FF,$13,$3F,$E6,$2D,$7E,$C3,$CB,$DE,$E9,$E1,$FF,$94
	.dc.b	$79,$FF,$15,$F9,$83,$F7,$97,$EC,$CF,$D5,$9F,$AE,$BE,$C3,$FD,$5F
	.dc.b	$B7,$E6,$30,$E1,$7B,$AD,$7E,$69,$FA,$79,$7E,$7D,$F9,$C7,$B6,$FF
	.dc.b	$8B,$78,$E5,$DB,$F0,$AF,$1E,$9E,$28,$F0,$FD,$1F,$E7,$FF,$27,$7E
	.dc.b	$D1,$FD,$C3,$EE,$9F,$A0,$FF,$65,$FC,$BF,$F6,$D5,$AF,$AA,$56,$BE
	.dc.b	$99,$F0,$9A,$FC,$AB,$F5,$F7,$D3,$83,$F3,$CF,$1B,$EB,$F9,$3F,$87
	.dc.b	$4F,$06,$9F,$D0,$EF,$FD,$23,$7B,$8F,$EE,$DF,$7F,$FC,$1F,$F1,$97
	.dc.b	$EB,$2F,$C8,$7F,$D2,$5F,$9A,$7F,$A4,$FE,$37,$53,$9F,$0D,$3F,$E9
	.dc.b	$7F,$D3,$33,$A7,$EC,$71,$7E,$DB,$FC,$19,$EB,$97,$4F,$0F,$F8,$24
	.dc.b	$63,$1E,$7E,$17,$F0,$7F,$CD,$BF,$84,$FE,$EE,$FC,$CB,$F8,$C7,$98
	.dc.b	$F9,$7F,$E3,$D5,$A9,$CF,$ED,$B8,$F3,$7F,$8A,$CF,$F9,$7F,$7F,$E0
	.dc.b	$8E,$FD,$5F,$EA,$9E,$4F,$F1,$C8,$C7,$CD,$FF,$8D,$FE,$7F,$F1,$43
	.dc.b	$F4,$F7,$A1,$F8,$EF,$CC,$BF,$25,$87,$F8,$03,$2F,$F3,$77,$9F,$FE
	.dc.b	$6F,$BD,$F9,$CB,$7E,$BD,$FE,$54,$CB,$F2,$DF,$47,$FE,$12,$7B,$E3
	.dc.b	$FE,$38,$7C,$38,$72,$E8,$F2,$65,$D7,$4F,$CF,$F3,$7F,$93,$7E,$EF
	.dc.b	$FD,$59,$F8,$6F,$FA,$CB,$F0,$BF,$F5,$8F,$07,$96,$73,$F8,$67,$FA
	.dc.b	$BF,$F3,$4F,$F3,$2C,$7F,$AB,$76,$FF,$10,$1F,$EA,$DF,$96,$FF,$12
	.dc.b	$C7,$0C,$3F,$C4,$33,$FE,$B1,$F9,$7F,$F5,$56,$7F,$E5,$DF,$09,$FA
	.dc.b	$73,$F5,$F7,$DD,$3F,$32,$7E,$76,$F5,$5F,$B9,$7F,$9D,$B1,$E3,$87
	.dc.b	$0B,$DD,$3F,$43,$D6,$5D,$75,$FF,$2E,$E1,$E0,$E8,$FF,$1D,$BC,$B4
	.dc.b	$F0,$69,$E8,$A0,$9D,$65,$E4,$D7,$FC,$ED,$F4,$BF,$EB,$BF,$8B,$FF
	.dc.b	$58,$FA,$FF,$EB,$FE,$7F,$BD,$39,$FB,$4F,$F5,$F4,$63,$F1,$0F,$F5
	.dc.b	$F6,$7F,$E4,$28,$C5,$7F,$93,$8E,$1F,$88,$B2,$EF,$7F,$EB,$B8,$F3
	.dc.b	$5B,$F3,$07,$EE,$CF,$C5,$DF,$81,$7F,$88,$FF,$35,$FF,$06,$F9,$9F
	.dc.b	$E9,$EB,$DD,$6B,$E4,$15,$FE,$12,$FD,$7F,$58,$F3,$CB,$FD,$0B,$87
	.dc.b	$6F,$C2,$7F,$CC,$6F,$2C,$FB,$6D,$FE,$99,$8C,$65,$D3,$C7,$FF,$68
	.dc.b	$7D,$FF,$FB,$23,$EB,$3F,$CC,$D0,$FC,$17,$FC,$47,$39,$F9,$DF,$F3
	.dc.b	$3F,$E0,$BF,$E7,$3F,$63,$3B,$7F,$B2,$F1,$57,$FB,$22,$FA,$7E,$39
	.dc.b	$6B,$53,$CB,$F0,$6F,$F3,$9D,$75,$FC,$1B,$FC,$E5,$5E,$43,$EF,$FB
	.dc.b	$4F,$35,$FB,$0B,$F1,$F7,$CF,$3F,$35,$F8,$EF,$93,$0F,$E2,$17,$C2
	.dc.b	$B7,$5F,$33,$F7,$0D,$63,$D7,$1F,$36,$1C,$3C,$8B,$39,$C3,$8F,$81
	.dc.b	$E5,$AE,$B7,$E3,$CF,$C5,$F1,$B9,$C3,$D1,$F1,$BF,$E3,$AB,$7C,$DB
	.dc.b	$F7,$8F,$4F,$8D,$FE,$8B,$9F,$7F,$07,$F5,$0F,$B5,$FC,$FD,$EF,$E8
	.dc.b	$FC,$F3,$FD,$26,$BF,$D3,$DC,$9F,$E9,$8C,$FE,$37,$C9,$FA,$2B,$95
	.dc.b	$78,$3F,$20,$F2,$46,$31,$FE,$69,$3D,$3F,$3A,$FE,$3D,$E5,$BF,$7E
	.dc.b	$7E,$F0,$FC,$23,$5F,$0C,$AD,$4B,$F4,$5F,$2F,$D9,$78,$71,$97,$BC
	.dc.b	$5F,$94,$78,$A7,$BC,$3C,$2F,$2F,$E8,$0F,$CE,$5F,$BB,$7F,$49,$C7
	.dc.b	$E4,$F4,$7C,$7E,$0F,$BB,$7B,$AD,$7E,$61,$F1,$BE,$8F,$E1,$4C,$C7
	.dc.b	$9F,$3F,$E4,$30,$F9,$F3,$C6,$3F,$79,$1F,$39,$FF,$36,$7E,$66,$FB
	.dc.b	$47,$9C,$FC,$9B,$5E,$62,$B5,$2F,$DE,$3E,$BF,$ED,$DF,$4F,$19,$7B
	.dc.b	$C7,$C5,$97,$E5,$A7,$BF,$DF,$7F,$11,$F0,$3E,$E7,$77,$F6,$9C,$38
	.dc.b	$67,$F0,$A7,$3F,$68,$FB,$27,$03,$F6,$2F,$F1,$67,$E0,$1F,$88,$AD
	.dc.b	$4E,$79,$1F,$BF,$7E,$33,$C4,$FC,$45,$4C,$79,$E3,$F5,$78,$39,$46
	.dc.b	$38,$3F,$17,$F8,$F3,$E4,$DF,$B8,$BF,$50,$FF,$36,$FE,$DD,$F9,$8A
	.dc.b	$D4,$E7,$EF,$9F,$AF,$FE,$73,$F1,$BF,$CC,$54,$C7,$9E,$2F,$D0,$3B
	.dc.b	$C6,$3F,$2D,$FC,$DF,$C0,$FF,$9A,$3D,$57,$CD,$F4,$D3,$F4,$F3,$98
	.dc.b	$EF,$F8,$1F,$EA,$BE,$0B,$F4,$D3,$3F,$00,$61,$CB,$F1,$5E,$B1,$8C
	.dc.b	$1F,$79,$FA,$AF,$19,$F9,$9B,$F0,$A7,$C4,$3F,$63,$7E,$75,$FB,$8B
	.dc.b	$DD,$6A,$5F,$9F,$7E,$F3,$EA,$7F,$71,$53,$1E,$98,$F3,$EE,$E5,$18
	.dc.b	$FC,$63,$C4,$EB,$17,$E6,$BF,$C8,$7F,$08,$FE,$1F,$F8,$23,$B5,$FB
	.dc.b	$B7,$FB,$27,$DC,$7E,$F3,$BF,$83,$FF,$3A,$56,$BE,$D7,$F6,$FF,$BC
	.dc.b	$E7,$F0,$9F,$E4,$B9,$F4,$E0,$FF,$14,$1C,$74,$ED,$A7,$58,$C6,$3C
	.dc.b	$FD,$FF,$E5,$17,$FC,$8B,$C9,$FE,$C7,$FB,$67,$8B,$F2,$9F,$1F,$C0
	.dc.b	$7E,$F6,$B5,$F3,$6A,$D7,$77,$9F,$0F,$1F,$0B,$DC,$B1,$D3,$8F,$EB
	.dc.b	$ED,$BC,$79,$43,$C1,$F1,$E0,$FC,$BD,$F9,$57,$EA,$7F,$46,$7E,$0E
	.dc.b	$DF,$E3,$56,$BD,$45,$6B,$F1,$FF,$95,$F9,$87,$92,$A6,$3C,$FD,$EE
	.dc.b	$F1,$C6,$FF,$87,$F2,$7B,$F4,$7B,$1F,$8B,$FB,$CF,$C4,$FE,$B2,$FD
	.dc.b	$1F,$EB,$4E,$7C,$B4,$BC,$0C,$F3,$7D,$4C,$BA,$7B,$9F,$1B,$0E,$1F
	.dc.b	$BE,$7C,$07,$96,$9F,$0B,$91,$F9,$FF,$F8,$5F,$DE,$7E,$0F,$FE,$C0
	.dc.b	$F8,$4F,$F1,$8E,$9F,$0F,$0F,$F6,$26,$7F,$89,$FB,$FF,$A7,$94,$CB
	.dc.b	$97,$2F,$57,$A7,$D4,$7D,$0F,$CC,$FF,$4B,$F9,$D3,$EF,$DF,$0B,$FE
	.dc.b	$76,$E3,$F8,$95,$A9,$CF,$ED,$3F,$97,$F8,$4F,$C0,$CC,$E8,$63,$CF
	.dc.b	$C2,$ED,$18,$C3,$F0,$78,$FC,$FF,$C2,$5F,$B1,$BF,$20,$9F,$16,$FC
	.dc.b	$37,$F4,$95,$A9,$CF,$D6,$63,$A3,$E7,$74,$FC,$4E,$6F,$EF,$BF,$86
	.dc.b	$7E,$53,$0E,$18,$7F,$29,$D3,$C1,$A7,$78,$FE,$E3,$1F,$81,$FC,$F9
	.dc.b	$F8,$DF,$F6,$BF,$C8,$5F,$B9,$7F,$9C,$EB,$53,$98,$FF,$31,$47,$E2
	.dc.b	$7E,$23,$F4,$77,$2B,$97,$EB,$7F,$9E,$8E,$BF,$46,$C5,$F9,$07,$E8
	.dc.b	$BE,$0B,$E4,$FC,$6B,$F4,$EF,$E6,$1F,$AC,$FF,$05,$9C,$C7,$F3,$3F
	.dc.b	$A8,$C7,$37,$C2,$3F,$CE,$BC,$BF,$E1,$EF,$73,$F6,$B7,$C3,$E5,$3B
	.dc.b	$77,$B3,$ED,$F2,$7F,$E2,$BF,$D2,$7F,$A5,$70,$BF,$4B,$7E,$EE,$F8
	.dc.b	$6F,$C9,$FD,$0F,$F8,$F6,$F7,$5A,$9F,$CB,$FD,$0E,$7E,$65,$8F,$F8
	.dc.b	$F2,$FF,$AB,$75,$59,$CE,$1C,$7C,$EF,$14,$63,$3F,$07,$9B,$FE,$3F
	.dc.b	$F9,$2F,$F5,$DF,$A1,$FE,$C9,$D7,$FC,$9D,$18,$FC,$57,$FD,$8D,$ED
	.dc.b	$E3,$F1,$47,$E2,$BF,$E4,$CA,$FF,$62,$F5,$7F,$8E,$0F,$F4,$DB,$C7
	.dc.b	$5E,$D5,$FE,$C5,$C3,$F8,$2F,$93,$FB,$0F,$99,$7D,$93,$F0,$87,$E6
	.dc.b	$8F,$D4,$DF,$96,$B2,$FD,$DF,$95,$FE,$27,$C7,$8E,$1C,$2F,$74,$FC
	.dc.b	$A3,$FC,$D1,$5A,$DF,$EE,$07,$C3,$F8,$A7,$F9,$9B,$F0,$C7,$E6,$E5
	.dc.b	$F7,$2F,$DA,$B3,$F6,$AF,$D7,$BF,$CE,$95,$C0,$DC,$EA,$F2,$73,$CB
	.dc.b	$C3,$3D,$BC,$4F,$04,$F7,$C7,$B3,$FE,$2A,$9E,$91,$8E,$CF,$F3,$B7
	.dc.b	$CB,$FF,$B8,$7E,$4D,$F9,$D7,$FA,$16,$31,$F6,$57,$CC,$7E,$37,$FE
	.dc.b	$82,$9F,$EF,$BE,$25,$9C,$FE,$3B,$FE,$F5,$EF,$EB,$18,$FC,$77,$F8
	.dc.b	$D6,$9F,$E8,$5F,$CA,$BF,$72,$3F,$42,$7C,$1F,$F9,$8B,$A3,$FC,$97
	.dc.b	$DD,$FE,$5B,$AD,$D6,$BE,$97,$FD,$DD,$DD,$FE,$9B,$9F,$E0,$7F,$79
	.dc.b	$FE,$96,$7A,$F4,$FD,$67,$8E,$5F,$E6,$ED,$BC,$5E,$FA,$7C,$53,$FE
	.dc.b	$72,$D7,$FD,$35,$F8,$CF,$FB,$AB,$F1,$AF,$F7,$E7,$E3,$FF,$EA,$98
	.dc.b	$C7,$07,$FB,$E7,$83,$FC,$E1,$1F,$EF,$ED,$3F,$0A,$FF,$BE,$5B,$F8
	.dc.b	$EB,$F1,$F7,$E1,$DF,$EF,$AD,$FB,$CE,$63,$FD,$FB,$A7,$FA,$AF,$F3
	.dc.b	$87,$EB,$BF,$9A,$FD,$C5,$F9,$C7,$D8,$EB,$FC,$F6,$DF,$E8,$1B,$DF
	.dc.b	$98,$BD,$F8,$1F,$E5,$3D,$7F,$D8,$55,$AD,$FF,$36,$F6,$FC,$E3,$FD
	.dc.b	$7E,$F2,$BE,$3D,$FE,$3D,$7C,$5A,$FE,$23,$EC,$F1,$A3,$F8,$01,$FA
	.dc.b	$97,$9B,$EA,$1F,$95,$F8,$31,$FD,$9D,$6B,$15,$3F,$2D,$D6,$3C,$F1
	.dc.b	$F7,$FC,$FF,$6C,$38,$7D,$7F,$83,$F1,$77,$E8,$4E,$FF,$F3,$7F,$E9
	.dc.b	$FC,$A7,$31,$FE,$72,$F5,$F1,$5B,$FC,$D7,$93,$C7,$F0,$1B,$5F,$E7
	.dc.b	$8F,$DA,$BC,$7F,$6B,$FE,$46,$FD,$D5,$FB,$33,$F6,$4C,$FD,$EA,$73
	.dc.b	$FB,$96,$17,$E3,$FF,$1B,$0E,$5F,$65,$F2,$3F,$13,$7E,$B8,$FC,$9B
	.dc.b	$F8,$D7,$C0,$FC,$29,$CC,$7F,$1D,$F3,$31,$5B,$F2,$F9,$3C,$7C,$8F
	.dc.b	$A7,$ED,$78,$FF,$5F,$FE,$25,$FA,$47,$CC,$8C,$7C,$0B,$F2,$FC,$70
	.dc.b	$AD,$F0,$9C,$A3,$1F,$39,$F7,$BE,$57,$FB,$34,$FF,$81,$BE,$4F,$EB
	.dc.b	$C6,$3E,$A1,$F2,$98,$FA,$4E,$7B,$71,$7A,$FD,$7F,$E2,$3E,$B1,$FA
	.dc.b	$5F,$F8,$F9,$70,$FE,$09,$CF,$DC,$97,$EF,$99,$AF,$F2,$2F,$E2,$1C
	.dc.b	$DF,$F2,$3F,$75,$BE,$1F,$A0,$7E,$33,$E2,$FF,$CB,$3F,$93,$BF,$5B
	.dc.b	$7E,$99,$9F,$92,$4E,$7E,$BF,$0F,$F5,$0F,$81,$0F,$80,$FD,$7E,$3D
	.dc.b	$3E,$BB,$3F,$09,$FE,$F3,$FA,$DF,$C0,$8F,$D5,$FF,$CD,$9E,$3B,$F7
	.dc.b	$9F,$EA,$B5,$BE,$25,$EF,$F0,$6F,$C2,$53,$E1,$D7,$BC,$BF,$D1,$FA
	.dc.b	$E3,$E8,$BF,$CA,$78,$F8,$23,$FA,$BB,$E2,$D3,$C3,$EC,$79,$AB,$7A
	.dc.b	$78,$27,$CB,$EB,$E7,$0F,$7E,$1E,$AF,$C1,$5F,$AF,$3F,$0D,$7E,$71
	.dc.b	$FC,$5B,$6F,$55,$FF,$9D,$6B,$52,$FC,$47,$CD,$8F,$4D,$3D,$F9,$74
	.dc.b	$CB,$FD,$79,$97,$28,$C7,$C6,$78,$EF,$E2,$DF,$C6,$87,$E9,$5F,$99
	.dc.b	$F7,$3F,$92,$E1,$E2,$9C,$F9,$FF,$B8,$FD,$67,$C3,$8F,$2D,$FF,$DC
	.dc.b	$B8,$FE,$87,$2E,$D1,$F6,$30,$E3,$EC,$E2,$7E,$AF,$EE,$3F,$82,$FB
	.dc.b	$B7,$C3,$7F,$20,$D7,$E0,$CE,$63,$FC,$97,$18,$CD,$5F,$E5,$EE,$6F
	.dc.b	$A3,$9B,$0E,$9F,$11,$DE,$3E,$6F,$AD,$FE,$0B,$FA,$97,$C4,$7A,$A7
	.dc.b	$F1,$BC,$FF,$BB,$F6,$7F,$C4,$5F,$0F,$9E,$0F,$C5,$CE,$7D,$DE,$8E
	.dc.b	$FD,$CE,$DF,$37,$FE,$29,$F3,$5F,$99,$FF,$AF,$79,$3E,$DC,$E6,$3F
	.dc.b	$D2,$5C,$5E,$88,$E1,$F9,$DF,$2E,$06,$1F,$E6,$AE,$5F,$17,$07,$F7
	.dc.b	$FF,$93,$7E,$60,$FD,$65,$FA,$96,$BE,$97,$FA,$3C,$E7,$F0,$1C,$5F
	.dc.b	$0F,$83,$F8,$95,$AF,$84,$F8,$DF,$B2,$3F,$53,$FF,$18,$FE,$D3,$FD
	.dc.b	$2F,$97,$F2,$73,$9F,$AE,$4F,$FA,$92,$73,$5F,$E3,$8F,$8E,$E6,$FF
	.dc.b	$8E,$BF,$59,$FA,$5F,$A1,$BC,$9F,$E0,$BF,$13,$FE,$39,$C3,$F3,$F3
	.dc.b	$9F,$6F,$FD,$53,$9F,$C0,$8C,$7C,$47,$2C,$3F,$86,$FC,$3F,$E6,$FC
	.dc.b	$17,$E9,$5F,$E9,$EF,$CA,$9F,$BC,$BF,$41,$F8,$BF,$3D,$F9,$8B,$DF
	.dc.b	$4A,$F7,$5F,$2B,$DD,$D6,$3F,$EA,$8E,$FF,$F2,$BE,$1D,$BB,$7F,$83
	.dc.b	$F1,$3C,$4F,$F6,$CC,$F4,$CB,$9B,$E5,$C3,$C9,$AF,$8F,$C6,$FD,$97
	.dc.b	$C0,$79,$7F,$44,$75,$3F,$41,$7E,$89,$F7,$3C,$8F,$D7,$DE,$EB,$53
	.dc.b	$F8,$0F,$D1,$33,$C6,$C7,$CB,$BF,$FB,$B3,$0F,$85,$FA,$26,$38,$FF
	.dc.b	$98,$C7,$3D,$3B,$F0,$79,$31,$E5,$A7,$6D,$BF,$99,$FC,$6F,$FB,$DB
	.dc.b	$F3,$B7,$E5,$E5,$F0,$BF,$D3,$7C,$17,$73,$F3,$9F,$E5,$F8,$71,$C3
	.dc.b	$85,$EE,$9F,$0F,$AE,$0F,$E8,$B9,$76,$3E,$7E,$1C,$7F,$3C,$B3,$AC
	.dc.b	$79,$FC,$A7,$FB,$C3,$1F,$3C,$E7,$0F,$2C,$FF,$BD,$3E,$E7,$33,$FC
	.dc.b	$83,$F3,$97,$E8,$9F,$D9,$7F,$31,$FD,$CF,$0B,$F0,$CF,$25,$EE,$B5
	.dc.b	$EC,$A9,$F9,$A6,$BE,$1D,$97,$4C,$3F,$AF,$93,$C4,$FE,$EE,$FF,$62
	.dc.b	$9D,$FE,$5D,$EC,$8E,$DB,$35,$F1,$C7,$5C,$3C,$F3,$9F,$D6,$BA,$72
	.dc.b	$76,$F9,$F7,$E1,$9F,$ED,$DF,$D4,$9F,$92,$FF,$80,$79,$9F,$9C,$FB
	.dc.b	$2F,$7F,$4A,$BD,$E7,$F6,$BF,$7A,$D6,$5F,$EE,$7D,$BD,$67,$CA,$C3
	.dc.b	$BF,$CA,$FA,$5E,$78,$72,$CF,$57,$E3,$D3,$FC,$25,$D3,$F4,$BE,$3D
	.dc.b	$F8,$47,$F4,$2E,$27,$E6,$EF,$C6,$3E,$77,$D6,$FD,$2B,$D9,$F3,$A7
	.dc.b	$5D,$6F,$CA,$75,$8B,$F3,$3F,$99,$7C,$38,$BD,$26,$9E,$59,$7D,$B8
	.dc.b	$C6,$7E,$2F,$C6,$BE,$6C,$7F,$86,$3F,$40,$7E,$D2,$5F,$0C,$FD,$D7
	.dc.b	$F4,$BF,$AC,$7F,$5B,$ED,$FC,$0D,$70,$37,$3A,$FC,$E7,$F1,$B1,$D3
	.dc.b	$E3,$7F,$04,$EF,$2F,$3F,$89,$F8,$87,$C1,$9F,$87,$3F,$07,$95,$FE
	.dc.b	$50,$9E,$73,$9C,$BF,$6F,$F3,$BF,$A0,$C7,$C5,$87,$F2,$6F,$C5,$FF
	.dc.b	$CD,$3F,$B6,$F0,$9F,$5F,$BD,$FE,$F9,$FB,$67,$EF,$D9,$C3,$87,$D0
	.dc.b	$F3,$D3,$C5,$7F,$EF,$FF,$9A,$C6,$3F,$13,$F7,$F7,$E5,$CF,$CE,$5F
	.dc.b	$A6,$73,$9C,$C7,$F4,$CE,$4F,$75,$F9,$A7,$8F,$D1,$F4,$D5,$1D,$3F
	.dc.b	$50,$C7,$E8,$1F,$C4,$3C,$AF,$F3,$A6,$3C,$B7,$F8,$D3,$9F,$D5,$3D
	.dc.b	$4E,$FA,$2F,$96,$F9,$6E,$DF,$8A,$3E,$45,$F0,$5F,$BC,$CF,$A7,$0F
	.dc.b	$E0,$2B,$52,$FE,$00,$F8,$D1,$C7,$1E,$78,$3E,$53,$D5,$FB,$EB,$F2
	.dc.b	$CF,$F2,$BF,$EC,$7F,$36,$31,$F4,$6F,$D6,$B1,$5F,$C9,$B8,$2F,$EB
	.dc.b	$DF,$0F,$F5,$57,$E1,$2F,$80,$7E,$6D,$F5,$E7,$30,$FC,$DF,$14,$F9
	.dc.b	$9E,$4F,$F8,$3F,$9B,$DB,$DC,$F9,$E7,$F0,$07,$E7,$7F,$2C,$FC,$14
	.dc.b	$63,$CE,$FF,$73,$5F,$08,$DF,$E3,$5F,$5D,$E3,$E7,$FD,$9F,$D9,$9C
	.dc.b	$CF,$CE,$9F,$89,$FE,$36,$73,$0F,$CE,$31,$4F,$C2,$F9,$3F,$EC,$4E
	.dc.b	$7F,$E6,$BF,$9C,$7F,$01,$3F,$8F,$FE,$49,$FE,$8D,$F9,$B7,$85,$AF
	.dc.b	$83,$56,$BF,$08,$FE,$E3,$F9,$26,$7F,$33,$FE,$65,$3C,$F2,$FD,$8F
	.dc.b	$BF,$FC,$EA,$3B,$7C,$5B,$4E,$D5,$D6,$7A,$65,$FD,$39,$DF,$E5,$FF
	.dc.b	$A1,$F3,$9E,$F6,$F1,$4E,$5C,$3B,$C2,$FA,$CF,$CE,$BA,$65,$EA,$BF
	.dc.b	$EA,$5E,$1B,$E5,$9F,$B6,$75,$F8,$CF,$C2,$47,$CF,$EF,$FE,$15,$BF
	.dc.b	$AA,$73,$A6,$F0,$FF,$39,$BE,$5F,$A5,$7C,$47,$E4,$FC,$EC,$3F,$95
	.dc.b	$FA,$0B,$97,$93,$2F,$E9,$F7,$C3,$B3,$1B,$7F,$85,$7F,$42,$7E,$85
	.dc.b	$FC,$0F,$0F,$1F,$0F,$AF,$EA,$95,$AF,$80,$F8,$D1,$8E,$1E,$6E,$BE
	.dc.b	$9F,$AF,$E9,$6F,$9C,$7E,$C8,$FC,$BF,$E7,$4E,$61,$F9,$C6,$2B,$F9
	.dc.b	$CF,$35,$FC,$CF,$CF,$FA,$66,$BC,$CF,$F1,$67,$0F,$9B,$9B,$FB,$5C
	.dc.b	$E7,$EB,$FE,$D3,$D3,$F1,$1E,$39,$FF,$89,$32,$FC,$27,$85,$FE,$47
	.dc.b	$FB,$A7,$EA,$5F,$D7,$B3,$E9,$39,$8F,$EF,$31,$8F,$B3,$57,$F5,$DB
	.dc.b	$E7,$1F,$E6,$F7,$AF,$90,$FC,$C1,$FB,$CF,$FA,$EA,$3F,$31,$D7,$FA
	.dc.b	$0C,$D1,$B9,$FD,$EE,$5E,$28,$E5,$E3,$7D,$BA,$DE,$0F,$27,$FC,$53
	.dc.b	$FA,$2F,$F2,$5F,$D3,$9F,$B4,$BF,$5E,$7E,$81,$FC,$A3,$EC,$BF,$C5
	.dc.b	$37,$BA,$D4,$FF,$2C,$F2,$F3,$F1,$6C,$7F,$C4,$57,$FC,$C7,$E5,$2C
	.dc.b	$E7,$0E,$3F,$1B,$E2,$8C,$67,$E0,$F8,$DF,$F1,$EF,$E8,$5F,$0F,$F4
	.dc.b	$37,$E5,$0F,$D0,$FF,$0D,$E6,$FF,$83,$79,$FF,$B6,$5E,$EB,$5E,$2F
	.dc.b	$E5,$BD,$2F,$F2,$0C,$F3,$EC,$FE,$06,$03,$F4,$7E,$E3,$FC,$74,$F3
	.dc.b	$CB,$F2,$9E,$6F,$93,$A5,$AF,$92,$7F,$31,$E6,$7F,$95,$3C,$97,$E4
	.dc.b	$0F,$D1,$7F,$D1,$DF,$BE,$BF,$36,$7E,$79,$FA,$6E,$3F,$F0,$26,$1C
	.dc.b	$2F,$75,$AF,$C4,$3F,$CB,$33,$9E,$1F,$48,$7B,$FE,$A7,$FE,$51,$79
	.dc.b	$79,$0C,$FC,$1D,$7E,$5D,$7C,$93,$9C,$BF,$AE,$63,$FE,$66,$FB,$67
	.dc.b	$E3,$3F,$E5,$2F,$C3,$9F,$09,$FB,$0F,$83,$FD,$B7,$0F,$F8,$6E,$F6
	.dc.b	$2B,$7E,$87,$F2,$4F,$B9,$E9,$8F,$E3,$32,$F4,$DF,$06,$C7,$7C,$BF
	.dc.b	$7B,$D3,$FE,$7E,$AE,$B8,$F4,$E6,$FF,$40,$FE,$8F,$FD,$44,$F9,$DF
	.dc.b	$A4,$BE,$B3,$E6,$FE,$41,$FE,$80,$9D,$74,$AB,$5D,$5F,$C2,$B8,$7F
	.dc.b	$C6,$95,$FD,$6F,$EB,$BF,$CF,$4F,$5E,$0F,$9E,$F2,$BE,$DC,$FE,$2F
	.dc.b	$91,$4F,$8B,$4F,$97,$8F,$F8,$FF,$F1,$DF,$E0,$23,$F4,$57,$E6,$9F
	.dc.b	$C1,$3A,$3F,$DD,$5E,$B7,$FA,$3A,$B7,$5A,$F6,$7F,$A6,$9F,$E9,$A8
	.dc.b	$E7,$B7,$E0,$9F,$E1,$BF,$8C,$FF,$A4,$5E,$38,$FF,$10,$E7,$FF,$53
	.dc.b	$C6,$2D,$CB,$FE,$AA,$FD,$5B,$FB,$DF,$EA,$AF,$B9,$FF,$3C,$7E,$70
	.dc.b	$FC,$67,$F8,$A7,$0F,$F9,$66,$F7,$5A,$97,$E8,$FF,$EA,$E9,$E7,$BF
	.dc.b	$E5,$9F,$F0,$3F,$E5,$3F,$EA,$57,$8F,$E0,$0C,$BF,$D9,$F5,$C7,$E7
	.dc.b	$1A,$78,$EB,$5E,$AF,$FB,$03,$E7,$FF,$41,$9B,$F3,$57,$E5,$FF,$77
	.dc.b	$F3,$17,$E8,$7F,$EC,$0B,$DD,$6B,$C0,$57,$FA,$63,$8F,$FC,$E9,$5D
	.dc.b	$36,$FD,$6B,$FA,$DF,$53,$FD,$78,$F2,$CB,$F1,$5D,$1E,$3F,$43,$E0
	.dc.b	$E7,$D7,$9F,$FD,$AB,$FB,$E3,$F2,$B2,$F9,$9F,$F5,$CF,$D7,$7C,$DF
	.dc.b	$DA,$BF,$DA,$B3,$B3,$5F,$88,$FF,$8F,$FB,$1F,$B0,$EB,$1F,$F4,$B7
	.dc.b	$0F,$F0,$2B,$E0,$D8,$EF,$97,$6F,$27,$C7,$18,$C7,$FC,$95,$D1,$FE
	.dc.b	$EA,$FC,$AF,$FB,$70,$F6,$9F,$A4,$3F,$11,$FF,$13,$72,$7F,$BA,$AB
	.dc.b	$75,$A9,$FF,$15,$7E,$95,$9F,$9C,$7E,$95,$FE,$E3,$BE,$BB,$FF,$18
	.dc.b	$78,$E5,$D7,$2F,$15,$A7,$C5,$3E,$1E,$5F,$F7,$FF,$D0,$3F,$56,$FF
	.dc.b	$06,$F3,$FE,$C3,$F2,$87,$E4,$2F,$D3,$7F,$DF,$F7,$B3,$FC,$D1,$EA
	.dc.b	$2B,$7F,$97,$6B,$1F,$EC,$9B,$FF,$2C,$BE,$0D,$8E,$F7,$C7,$9F,$C7
	.dc.b	$18,$C3,$F7,$FE,$9F,$F4,$BB,$FF,$2B,$FE,$82,$FA,$FF,$E7,$FF,$7A
	.dc.b	$FD,$0F,$FC,$A1,$FA,$63,$C8,$FF,$62,$7D,$A3,$0E,$1E,$F3,$0E,$1F
	.dc.b	$66,$FD,$87,$86,$BD,$F0,$E1,$A7,$FB,$93,$5C,$9F,$D7,$63,$A7,$C1
	.dc.b	$E7,$4E,$FA,$76,$CF,$C9,$D9,$FD,$8E,$73,$8F,$A3,$7F,$E0,$F9,$74
	.dc.b	$3F,$CB,$5E,$BB,$F3,$8F,$EE,$9A,$BF,$44,$FA,$7F,$B7,$6F,$E9,$17
	.dc.b	$BF,$92,$FF,$0C,$D7,$D8,$FD,$97,$53,$CF,$2F,$EB,$B7,$C3,$7F,$46
	.dc.b	$5F,$36,$39,$E7,$E0,$C7,$F8,$4E,$DF,$EE,$D9,$CE,$7D,$BC,$8F,$66
	.dc.b	$1F,$57,$EE,$5F,$A7,$7F,$A9,$7F,$21,$7E,$31,$FB,$5F,$CE,$7F,$9A
	.dc.b	$E7,$FE,$F7,$BD,$FE,$02,$BD,$FA,$2B,$FB,$17,$BC,$7F,$D6,$1C,$3F
	.dc.b	$DD,$B1,$EF,$F9,$A7,$AA,$3F,$C2,$11,$D3,$1E,$7E,$4F,$9A,$73,$7E
	.dc.b	$5D,$3F,$84,$74,$E6,$5F,$BB,$FE,$6C,$F1,$7E,$9B,$F4,$FE,$6F,$FC
	.dc.b	$09,$AF,$F9,$1A,$BF,$C3,$D5,$BD,$97,$BF,$92,$FB,$93,$FA,$5F,$DF
	.dc.b	$7C,$D7,$DB,$C6,$EE,$FF,$84,$27,$95,$FE,$07,$97,$C9,$18,$E9,$FA
	.dc.b	$7C,$DF,$A3,$FB,$F7,$E3,$85,$F9,$EF,$F2,$5F,$8F,$F9,$5F,$9F,$7E
	.dc.b	$36,$76,$6B,$DC,$4F,$CF,$F9,$BD,$61,$FE,$EB,$DB,$F6,$06,$2F,$D2
	.dc.b	$72,$C7,$B7,$47,$C9,$8C,$63,$D3,$9B,$E6,$78,$EF,$9F,$FD,$97,$CE
	.dc.b	$7F,$BC,$BF,$22,$FF,$15,$7F,$00,$FE,$86,$B4,$7F,$89,$3F,$12,$4F
	.dc.b	$F8,$BF,$F6,$0D,$60,$FD,$27,$E5,$DF,$0D,$FF,$C7,$2F,$2C,$BB,$73
	.dc.b	$F8,$AF,$C3,$97,$4F,$5B,$F0,$1C,$0F,$CE,$9F,$60,$FC,$ED,$F9,$2F
	.dc.b	$F2,$2F,$CF,$3F,$E0,$07,$87,$F8,$1D,$FF,$BF,$E1,$C2,$F7,$F0,$0B
	.dc.b	$79,$5F,$BD,$CF,$B6,$7F,$82,$7C,$13,$70,$E5,$9F,$F3,$09,$F2,$C6
	.dc.b	$30,$73,$7E,$5F,$E4,$FF,$CD,$BF,$42,$FD,$C3,$D3,$FB,$43,$DF,$FF
	.dc.b	$26,$F0,$F9,$EF,$75,$A9,$7E,$B5,$FC,$46,$79,$F1,$7F,$98,$9F,$F3
	.dc.b	$CF,$DA,$FE,$61,$E3,$C2,$CB,$F9,$AD,$7F,$9F,$E7,$31,$F9,$68,$FF
	.dc.b	$41,$7C,$63,$F1,$DF,$F1,$17,$EF,$2F,$DB,$9F,$B1,$BF,$2C,$7E,$AB
	.dc.b	$FC,$8F,$0E,$1F,$4E,$C3,$85,$EF,$E8,$F5,$97,$F8,$B7,$3F,$F3,$CE
	.dc.b	$1D,$F4,$FD,$7B,$CB,$0F,$07,$C6,$7F,$4B,$E4,$FB,$53,$9D,$BC,$7C
	.dc.b	$DF,$C0,$FD,$CF,$EC,$3F,$88,$3F,$18,$7E,$82,$FD,$0F,$F0,$FF,$A1
	.dc.b	$7F,$39,$BD,$D6,$A7,$F1,$7F,$8C,$67,$B9,$8F,$F0,$2B,$FE,$DB,$F8
	.dc.b	$8D,$9C,$E1,$C7,$C1,$F1,$46,$33,$F0,$69,$8C,$7C,$91,$E3,$F0,$DF
	.dc.b	$90,$77,$9C,$EB,$F9,$3F,$D6,$B1,$9F,$96,$DF,$88,$63,$1E,$5E,$85
	.dc.b	$C3,$F3,$5F,$47,$6F,$FB,$77,$2F,$E8,$78,$7D,$F8,$78,$7F,$25,$D7
	.dc.b	$E8,$9B,$E1,$E1,$F0,$BC,$7F,$78,$E5,$F9,$C7,$3F,$DC,$FA,$7E,$79
	.dc.b	$D7,$DD,$ED,$F9,$37,$F6,$3F,$C6,$9F,$B4,$1F,$EF,$5F,$C0,$0F,$CC
	.dc.b	$FF,$BC,$7C,$96,$EB,$5E,$F2,$B5,$F3,$FF,$B8,$F8,$9F,$F0,$56,$2F
	.dc.b	$D5,$3F,$BE,$79,$7C,$63,$F9,$A9,$E0,$3B,$E7,$F9,$E8,$F1,$65,$DB
	.dc.b	$1F,$17,$BB,$FD,$26,$F8,$67,$E1,$C3,$FC,$19,$F7,$FD,$63,$1A,$3F
	.dc.b	$73,$C6,$3E,$69,$B7,$FE,$13,$F2,$ED,$FF,$00,$EC,$FE,$21,$97,$6C
	.dc.b	$3E,$17,$B2,$FD,$EB,$8A,$FC,$06,$FE,$DD,$8F,$DB,$73,$F8,$B6,$BE
	.dc.b	$2F,$7F,$1D,$C3,$C7,$E3,$F9,$AF,$2F,$CD,$79,$FE,$E7,$D3,$F6,$AF
	.dc.b	$F1,$67,$EC,$9F,$E3,$3F,$D3,$9F,$A1,$97,$EA,$0F,$D2,$DF,$9F,$3B
	.dc.b	$9E,$E7,$2F,$04,$71,$C7,$96,$1C,$6F,$87,$B9,$BA,$D6,$5F,$E2,$FC
	.dc.b	$FC,$58,$3F,$50,$C5,$74,$AF,$29,$E4,$D3,$C7,$A7,$F8,$CF,$0E,$33
	.dc.b	$9C,$BF,$CE,$97,$DF,$F3,$3F,$F3,$0F,$B7,$FE,$69,$FC,$59,$F9,$83
	.dc.b	$F1,$3F,$ED,$FE,$0B,$3F,$F2,$B6,$1C,$3E,$FD,$87,$0B,$79,$BD,$7D
	.dc.b	$47,$BD,$E8,$D5,$8F,$81,$FE,$E2,$6F,$5F,$16,$BE,$18,$C6,$7D,$A1
	.dc.b	$F8,$5F,$B6,$7E,$6F,$E9,$CF,$D2,$5F,$A8,$7D,$5F,$D3,$DF,$97,$7F
	.dc.b	$4C,$FC,$97,$E7,$5E,$EB,$53,$FD,$43,$F5,$0C,$E1,$CB,$1F,$2F,$2F
	.dc.b	$C1,$7B,$C3,$F3,$AF,$66,$3C,$FA,$9A,$78,$3E,$27,$CD,$82,$B5,$2F
	.dc.b	$CB,$FF,$9B,$74,$FF,$A0,$7F,$6D,$FE,$07,$FB,$CB,$F3,$07,$D3,$7F
	.dc.b	$06,$FC,$F3,$E4,$F0,$78,$F4,$61,$C2,$F7,$EE,$2F,$F1,$17,$BA,$FE
	.dc.b	$1F,$FB,$27,$D3,$7B,$E1,$EE,$CF,$C5,$83,$3F,$16,$7E,$1A,$FC,$B5
	.dc.b	$F7,$8F,$E3,$38,$FB,$BF,$27,$7E,$4E,$FC,$BD,$FA,$5F,$E0,$39,$3F
	.dc.b	$4D,$F8,$BF,$E7,$EF,$75,$AF,$03,$F5,$9E,$43,$F5,$3D,$65,$CF,$6F
	.dc.b	$31,$FA,$CC,$3B,$FE,$A7,$F8,$D7,$9E,$9D,$F0,$F2,$5F,$8E,$3C,$5F
	.dc.b	$8C,$7D,$4F,$C0,$5F,$A7,$BF,$77,$FF,$06,$7E,$84,$F7,$3F,$45,$F0
	.dc.b	$5F,$B5,$BD,$D6,$A7,$F4,$DC,$99,$C3,$96,$FE,$63,$FD,$0E,$BF,$63
	.dc.b	$FD,$16,$1C,$BE,$09,$9F,$7F,$43,$C9,$7E,$39,$CE,$5F,$A0,$C7,$F8
	.dc.b	$37,$C8,$BF,$55,$3F,$A1,$F9,$17,$D3,$7F,$8A,$BF,$07,$DD,$6B,$CA
	.dc.b	$56,$BC,$9F,$EC,$BC,$3F,$8B,$C7,$A7,$EA,$BF,$B7,$7B,$17,$0E,$57
	.dc.b	$C7,$3E,$B3,$E1,$FC,$E3,$F3,$71,$FC,$83,$F6,$EF,$F3,$C4,$7F,$53
	.dc.b	$F4,$9F,$28,$FD,$95,$FA,$3F,$F2,$1B,$DF,$CB,$7E,$FA,$B5,$F6,$19
	.dc.b	$FC,$05,$97,$F0,$BB,$FF,$42,$7C,$75,$9C,$E1,$C7,$D0,$FD,$8D,$74
	.dc.b	$D3,$C3,$F1,$BF,$B4,$FD,$49,$FA,$A1,$7E,$A0,$FC,$6F,$FC,$CF,$F0
	.dc.b	$DF,$C3,$7A,$3F,$61,$5C,$0D,$CE,$BF,$53,$FE,$75,$1D,$3F,$05,$FE
	.dc.b	$40,$EF,$9B,$F0,$8F,$1C,$7B,$7D,$CF,$FB,$4B,$D8,$FE,$8F,$5E,$3D
	.dc.b	$3C,$95,$AF,$57,$F8,$5F,$EC,$9F,$CD,$7E,$BB,$FC,$9D,$CA,$FD,$53
	.dc.b	$FD,$3F,$FA,$C7,$F4,$BA,$D7,$DA,$7F,$AA,$67,$FD,$47,$F9,$C6,$7C
	.dc.b	$BF,$F2,$06,$BF,$97,$BF,$E8,$1F,$5B,$FC,$DD,$E3,$97,$5A,$F0,$9F
	.dc.b	$E4,$7C,$7A,$6B,$FC,$7B,$F4,$DF,$F1,$79,$FC,$FB,$ED,$1F,$92,$BF
	.dc.b	$00,$7E,$A1,$FD,$7E,$B7,$EF,$7F,$92,$2B,$5E,$9E,$7E,$DD,$F8,$5F
	.dc.b	$F4,$9B,$EB,$C3,$FA,$D7,$8E,$7D,$BC,$FF,$15,$A7,$C5,$8B,$EE,$7F
	.dc.b	$99,$FD,$47,$F8,$BC,$85,$F0,$7F,$96,$39,$CB,$F7,$4E,$8F,$F1,$B6
	.dc.b	$3F,$DF,$27,$FC,$A7,$3F,$DF,$BD,$CF,$F1,$7F,$E1,$69,$CC,$FF,$70
	.dc.b	$8F,$EB,$EE,$7E,$5B,$EF,$D6,$3D,$FE,$91,$F8,$2E,$07,$D6,$BF,$46
	.dc.b	$7D,$27,$F8,$3F,$F4,$EF,$F5,$DF,$C3,$73,$CF,$FC,$1F,$87,$0B,$DF
	.dc.b	$C3,$2F,$FD,$83,$F1,$3B,$BF,$E2,$39,$75,$7F,$AF,$E7,$E1,$87,$EE
	.dc.b	$DE,$8E,$0F,$5E,$7D,$75,$F1,$71,$FA,$31,$F0,$E1,$E7,$9F,$37,$16
	.dc.b	$67,$BE,$CF,$DF,$9F,$A0,$FF,$9D,$BE,$6D,$F8,$97,$97,$4F,$55,$6B
	.dc.b	$A9,$5F,$E2,$FD,$BC,$7A,$FC,$5A,$D7,$E6,$FD,$F0,$E3,$97,$FA,$0B
	.dc.b	$F4,$AC,$5F,$CB,$F3,$7F,$DC,$7F,$91,$3E,$F9,$F8,$F7,$86,$B5,$E9
	.dc.b	$2B,$5F,$A3,$E7,$8B,$F8,$4C,$E7,$F0,$1E,$B1,$8A,$7E,$D9,$F0,$7C
	.dc.b	$4F,$D2,$FE,$E5,$FE,$42,$F8,$A7,$EB,$5F,$4E,$73,$F5,$BF,$E4,$E9
	.dc.b	$7E,$FF,$9E,$9F,$D1,$F5,$FD,$63,$FC,$9B,$7F,$A4,$7F,$84,$9B,$58
	.dc.b	$FF,$94,$3E,$2F,$C5,$5F,$A5,$8F,$F2,$9F,$3E,$2B,$FC,$9D,$F6,$BF
	.dc.b	$EC,$3F,$A0,$FC,$47,$67,$E6,$A7,$F7,$BF,$60,$AD,$7E,$AF,$DE,$31
	.dc.b	$EE,$FF,$91,$35,$C5,$7F,$B3,$70,$ED,$5A,$C5,$F9,$96,$B0,$FF,$51
	.dc.b	$FE,$79,$F9,$5F,$C6,$FF,$80,$F6,$FF,$B5,$7F,$04,$FF,$80,$1F,$12
	.dc.b	$EC,$D7,$AD,$F6,$71,$56,$BF,$0B,$CE,$1F,$99,$F3,$7F,$96,$E5,$FC
	.dc.b	$33,$BF,$1C,$5F,$E7,$68,$EF,$8F,$2F,$9B,$6B,$F9,$99,$CE,$69,$FF
	.dc.b	$17,$5F,$EF,$FE,$CF,$F9,$CF,$BB,$FE,$42,$F9,$37,$EB,$AF,$C5,$5F
	.dc.b	$BF,$3F,$39,$7E,$49,$C7,$2F,$F4,$3E,$3C,$70,$E1,$7B,$FB,$7D,$57
	.dc.b	$F5,$6F,$B2,$FF,$64,$3F,$E5,$6D,$3C,$30,$D3,$F7,$11,$87,$C7,$A7
	.dc.b	$8B,$CF,$FE,$35,$39,$CB,$BF,$A3,$FC,$77,$F0,$5F,$EA,$7E,$23,$F4
	.dc.b	$F7,$E0,$5F,$CA,$FF,$65,$FF,$51,$7C,$6F,$F9,$52,$F7,$5A,$F7,$15
	.dc.b	$AF,$D5,$BF,$92,$56,$BD,$9F,$E1,$8F,$0C,$6F,$17,$FA,$A3,$3F,$04
	.dc.b	$77,$FB,$2F,$F1,$3C,$F2,$35,$9F,$7C,$BF,$D6,$9E,$77,$A6,$7F,$D6
	.dc.b	$3F,$9E,$3C,$17,$E5,$CF,$DB,$DF,$87,$3F,$31,$7E,$55,$EB,$C7,$8E
	.dc.b	$1C,$2F,$75,$AE,$2F,$15,$32,$ED,$97,$AA,$31,$83,$F3,$BC,$4F,$FB
	.dc.b	$D2,$78,$F9,$3F,$D3,$7E,$3B,$FA,$97,$27,$F8,$0A,$3F,$A9,$4F,$EF
	.dc.b	$3D,$1F,$F7,$03,$FE,$CC,$8E,$1F,$0B,$EF,$8F,$69,$FD,$3B,$E3,$BC
	.dc.b	$B8,$FF,$B4,$9F,$0C,$7F,$AC,$3E,$37,$FA,$D7,$E2,$1F,$E1,$0F,$93
	.dc.b	$7E,$33,$FE,$4A,$FD,$BF,$FA,$0F,$BF,$FE,$C1,$F0,$3F,$AE,$5E,$EB
	.dc.b	$5F,$1A,$AF,$F6,$37,$BD,$AC,$FF,$DC,$39,$74,$FB,$2F,$E9,$58,$F1
	.dc.b	$FA,$BB,$5A,$CB,$9E,$3F,$E2,$2E,$2F,$F8,$01,$39,$C1,$AF,$93,$6F
	.dc.b	$4C,$E6,$3C,$FF,$43,$FE,$0E,$BE,$33,$FE,$10,$C7,$AC,$FF,$83,$27
	.dc.b	$3A,$79,$6F,$8F,$C0,$FA,$70,$FF,$05,$7C,$17,$F8,$2F,$E6,$3F,$C2
	.dc.b	$DF,$33,$FE,$47,$C7,$AD,$FF,$84,$BE,$63,$FC,$3D,$F4,$7C,$61,$87
	.dc.b	$F3,$8C,$7E,$D7,$17,$F8,$FA,$7F,$9C,$47,$BE,$31,$3D,$FE,$6B,$FC
	.dc.b	$39,$3C,$E3,$18,$7F,$86,$E3,$17,$FE,$1E,$C9,$CD,$E9,$FC,$01,$FB
	.dc.b	$DB,$F6,$77,$E5,$8F,$CF,$5F,$4C,$FD,$A1,$F9,$47,$1C,$38,$5E,$EB
	.dc.b	$5F,$8D,$67,$1F,$B3,$B7,$FB,$5E,$DF,$A9,$63,$4F,$0D,$78,$34,$EF
	.dc.b	$F8,$8F,$D8,$C3,$84,$E7,$2F,$F4,$F7,$37,$D8,$A7,$EF,$4F,$CD,$5F
	.dc.b	$A7,$7F,$0C,$F0,$FF,$73,$F1,$1F,$5E,$F7,$5A,$F5,$7F,$D5,$BE,$A3
	.dc.b	$F1,$FD,$65,$CF,$6F,$F7,$EF,$6F,$F9,$C3,$1E,$FF,$9C,$FB,$1E,$7A
	.dc.b	$F7,$F3,$3D,$11,$8C,$FF,$CB,$78,$7A,$F3,$7E,$59,$FC,$FF,$F2,$B7
	.dc.b	$D6,$BF,$2D,$FE,$69,$D9,$F8,$DB,$DF,$C3,$2F,$75,$FC,$E7,$CC,$D6
	.dc.b	$3F,$D2,$72,$F3,$83,$F8,$CA,$FC,$B7,$E9,$E3,$D3,$0F,$0F,$E5,$BF
	.dc.b	$8D,$BE,$13,$9D,$3C,$9B,$FF,$96,$FC,$EF,$F2,$87,$E3,$1F,$CF,$FE
	.dc.b	$0D,$F9,$1F,$FB,$8F,$F6,$B7,$CB,$FF,$97,$6F,$7F,$09,$FE,$0E,$AD
	.dc.b	$7D,$36,$75,$FF,$29,$E5,$FE,$38,$7F,$C2,$A6,$4F,$F7,$0B,$FD,$35
	.dc.b	$CF,$80,$C3,$AE,$7E,$2F,$C2,$9E,$47,$9E,$FA,$E7,$E1,$9E,$11,$D7
	.dc.b	$F3,$7F,$F1,$E6,$6F,$A9,$79,$0F,$A2,$7E,$B6,$FD,$9F,$87,$CF,$FF
	.dc.b	$9B,$EF,$75,$AF,$B9,$CF,$C0,$7F,$98,$F8,$7F,$D6,$FA,$76,$D7,$D3
	.dc.b	$18,$C7,$96,$9D,$FE,$57,$D1,$4B,$F2,$E7,$DB,$CC,$CF,$E7,$4F,$D0
	.dc.b	$1F,$9A,$3F,$4B,$FE,$9D,$F4,$1F,$8E,$F9,$8F,$95,$7B,$AD,$7C,$7F
	.dc.b	$F8,$37,$C6,$FD,$4F,$67,$C4,$D9,$7D,$7F,$38,$F5,$47,$E2,$E3,$93
	.dc.b	$EA,$31,$A7,$8B,$4F,$0F,$DD,$79,$A3,$19,$7F,$86,$FF,$02,$F6,$FC
	.dc.b	$87,$F5,$B0,$7D,$F3,$FA,$DE,$1F,$CF,$7F,$1D,$FF,$50,$AF,$8E,$FF
	.dc.b	$47,$47,$FB,$8B,$F7,$17,$C8,$FF,$98,$BF,$69,$FE,$EB,$EF,$BF,$24
	.dc.b	$7E,$11,$FE,$AF,$C3,$85,$EF,$E6,$FF,$8F,$57,$CB,$E4,$56,$BE,$D3
	.dc.b	$BE,$BF,$AB,$D3,$BF,$EE,$F8,$84,$78,$A5,$AF,$8B,$A7,$FD,$85,$87
	.dc.b	$38,$F2,$F7,$63,$0F,$DB,$7E,$68,$FC,$2D,$EA,$FD,$DF,$C9,$1F,$A3
	.dc.b	$3F,$0A,$FE,$13,$C1,$E6,$C7,$8E,$1C,$3A,$98,$70,$B7,$ED,$BF,$06
	.dc.b	$F7,$FA,$DF,$A7,$4F,$06,$97,$2E,$58,$F8,$3E,$D7,$CD,$82,$F7,$9F
	.dc.b	$6D,$7F,$AC,$CF,$53,$FA,$36,$3F,$C4,$7E,$31,$F9,$73,$F3,$8F,$F7
	.dc.b	$BF,$C5,$6F,$E4,$BD,$F8,$7F,$51,$AD,$7D,$C7,$D9,$E3,$F7,$4F,$43
	.dc.b	$E8,$57,$C6,$F0,$2E,$7D,$F0,$ED,$E5,$FA,$2B,$AE,$9E,$2D,$1C,$DE
	.dc.b	$CF,$B7,$FF,$3B,$FD,$BF,$F9,$EB,$EE,$7D,$51,$8F,$B9,$FC,$6E,$FE
	.dc.b	$28,$FF,$47,$EF,$FC,$4B,$B3,$FD,$0F,$C3,$FC,$82,$30,$7F,$A1,$DF
	.dc.b	$F4,$87,$C9,$FF,$A5,$23,$AC,$63,$2F,$F4,$2E,$1E,$EF,$CF,$1F,$B9
	.dc.b	$FF,$B8,$7D,$BF,$A1,$FB,$CB,$0F,$F6,$8D,$EE,$B5,$D4,$AD,$7A,$BF
	.dc.b	$4F,$1E,$7F,$84,$7E,$5D,$EB,$F6,$97,$0E,$37,$FE,$68,$C7,$A4,$FF
	.dc.b	$1F,$C7,$A7,$1F,$F4,$B9,$7E,$59,$FD,$37,$77,$E5,$3F,$DC,$7C,$97
	.dc.b	$E5,$58,$FA,$3F,$F4,$1F,$D0,$31,$E3,$87,$F5,$5C,$38,$5E,$FE,$DD
	.dc.b	$EA,$EF,$3E,$D5,$D6,$FF,$A0,$C6,$3E,$79,$F8,$BF,$F4,$89,$E0,$75
	.dc.b	$8F,$DA,$E3,$E3,$FC,$27,$D5,$7C,$76,$F3,$E2,$DB,$FD,$A9,$AF,$9B
	.dc.b	$2F,$E8,$7D,$FF,$EC,$8F,$C7,$BF,$D9,$3F,$A5,$3F,$43,$FF,$13,$5E
	.dc.b	$FE,$17,$F3,$DD,$EB,$DD,$7E,$F6,$B5,$87,$FB,$4B,$1F,$59,$FD,$DF
	.dc.b	$0F,$D3,$7E,$1D,$EB,$AD,$E3,$CC,$64,$FA,$0F,$5E,$1C,$71,$F3,$56
	.dc.b	$BF,$64,$CC,$FF,$B0,$FF,$02,$FF,$5E,$FE,$0B,$FE,$CD,$FC,$1F,$E3
	.dc.b	$CE,$7F,$03,$FF,$65,$FC,$5B,$F0,$6C,$FE,$0E,$FC,$1F,$FD,$97,$7F
	.dc.b	$E8,$FE,$8F,$F6,$43,$EA,$39,$7D,$37,$FB,$25,$E3,$3F,$EC,$FC,$7F
	.dc.b	$D9,$31,$E0,$FD,$65,$D7,$F6,$63,$1E,$47,$F4,$5E,$5F,$A5,$48,$F0
	.dc.b	$CF,$FA,$07,$2F,$C9,$4E,$7E,$27,$FA,$59,$FD,$33,$E8,$7E,$0C,$76
	.dc.b	$AF,$F0,$6F,$B1,$E1,$CB,$FA,$64,$7A,$67,$4F,$E7,$7E,$53,$E4,$FE
	.dc.b	$B8,$FD,$D7,$FD,$D1,$FA,$EF,$FC,$00,$FA,$57,$DB,$7C,$7E,$0F,$C4
	.dc.b	$5E,$FA,$57,$BA,$7E,$21,$F4,$2B,$A7,$07,$DF,$E1,$FD,$64,$71,$D7
	.dc.b	$E6,$3C,$F0,$F0,$7B,$BE,$4D,$7C,$7E,$D7,$F9,$2B,$CF,$FB,$FE,$1B
	.dc.b	$F7,$87,$83,$FE,$FF,$FB,$3F,$E8,$BF,$82,$BC,$2F,$C0,$DE,$EB,$5F
	.dc.b	$0F,$FE,$D3,$FA,$CC,$E2,$AF,$37,$07,$F9,$35,$E0,$FE,$05,$E5,$CB
	.dc.b	$FE,$6B,$B7,$67,$92,$FC,$73,$9F,$C3,$3F,$86,$7C,$27,$FB,$F5,$7D
	.dc.b	$93,$F0,$CF,$F4,$BF,$DF,$FF,$09,$AD,$4E,$7E,$ED,$F8,$4E,$76,$FF
	.dc.b	$7F,$FB,$BF,$A9,$BF,$E2,$DF,$5A,$FA,$C7,$BB,$9B,$FF,$33,$4E,$61
	.dc.b	$AF,$8B,$5F,$E4,$98,$78,$17,$E1,$9E,$8C,$3F,$8D,$7E,$6F,$FC,$83
	.dc.b	$F3,$57,$E8,$4F,$A3,$FF,$39,$72,$79,$6F,$F9,$3D,$EE,$B5,$0F,$DD
	.dc.b	$F5,$EF,$29,$F4,$1E,$7A,$E3,$F0,$36,$B4,$B3,$F0,$E7,$FE,$E5,$DB
	.dc.b	$F9,$4E,$3D,$E3,$FC,$C3,$E5,$FF,$74,$C7,$FD,$0D,$E5,$FF,$49,$FC
	.dc.b	$37,$FC,$1F,$F6,$6F,$D1,$9F,$A1,$FF,$7A,$C3,$86,$1F,$A0,$F8,$B6
	.dc.b	$1C,$2F,$71,$EF,$C7,$A5,$7F,$9E,$34,$FF,$6E,$4F,$FB,$C2,$3C,$C7
	.dc.b	$89,$E7,$F3,$3F,$E7,$AB,$7C,$DF,$CF,$AF,$06,$3D,$F4,$C6,$4F,$DA
	.dc.b	$5F,$BC,$FF,$31,$FC,$3F,$F6,$3E,$BB,$ED,$FC,$95,$AF,$8A,$56,$BE
	.dc.b	$EF,$38,$7C,$FC,$7A,$7E,$29,$70,$E1,$7D,$35,$DE,$3C,$BD,$6F,$F4
	.dc.b	$8C,$7C,$1F,$C2,$BF,$DE,$7F,$85,$3F,$0E,$C6,$7F,$EF,$3F,$67,$5F
	.dc.b	$35,$DE,$FF,$85,$E1,$F1,$CF,$AA,$FF,$BF,$3D,$0F,$7B,$FE,$FE,$F1
	.dc.b	$FF,$D7,$DF,$95,$3F,$01,$FF,$33,$DF,$F1,$E8,$C3,$FE,$C3,$FC,$9B
	.dc.b	$FD,$89,$F0,$BF,$D9,$E7,$32,$9F,$ED,$5C,$FD,$3B,$7F,$BF,$7C,$0F
	.dc.b	$F9,$DC,$5F,$F8,$3E,$FF,$D9,$D1,$FE,$78,$78,$18,$7F,$99,$9F,$CC
	.dc.b	$DE,$33,$D3,$E7,$3F,$BC,$63,$89,$E7,$1F,$CD,$1F,$F0,$71,$93,$FD
	.dc.b	$0E,$FF,$B1,$B0,$EF,$1F,$E8,$AF,$C3,$BF,$D1,$98,$F9,$63,$18,$FB
	.dc.b	$F1,$F5,$D7,$92,$3F,$8A,$CF,$56,$BE,$0C,$71,$B7,$8E,$3F,$C4,$18
	.dc.b	$75,$AD,$C7,$FB,$BF,$1F,$1C,$38,$FC,$79,$7C,$A7,$FA,$CD,$69,$F1
	.dc.b	$61,$C3,$CB,$7E,$E4,$FC,$79,$FB,$D3,$DA,$7E,$77,$FE,$DE,$DB,$C5
	.dc.b	$87,$0B,$DD,$6B,$F3,$AF,$64,$F4,$D3,$FD,$DB,$7C,$3F,$4E,$F8,$91
	.dc.b	$F6,$A7,$96,$5D,$FC,$AF,$26,$9E,$38,$F1,$7E,$87,$F4,$3E,$09,$F8
	.dc.b	$D7,$FB,$CF,$F6,$27,$ED,$9F,$EB,$7F,$CE,$9F,$AA,$7D,$9B,$DF,$42
	.dc.b	$F7,$D2,$B7,$E3,$9B,$CB,$FD,$03,$C1,$F4,$3A,$7F,$CA,$19,$72,$8F
	.dc.b	$5C,$62,$7C,$5A,$7F,$B1,$FB,$7C,$D1,$8C,$9F,$97,$FE,$23,$E4,$1F
	.dc.b	$BE,$7F,$AE,$BC,$67,$EE,$5F,$E5,$7F,$C7,$7F,$D9,$5D,$5F,$8B,$BD
	.dc.b	$FD,$47,$F8,$02,$B5,$F7,$1F,$95,$9E,$7F,$57,$F3,$38,$51,$C1,$F3
	.dc.b	$83,$39,$F8,$6F,$C1,$1C,$E3,$1D,$6F,$95,$FC,$FF,$DB,$FF,$A1,$BF
	.dc.b	$47,$7E,$26,$FD,$45,$F7,$2A,$7F,$00,$FE,$A5,$E1,$5B,$9D,$7E,$17
	.dc.b	$9F,$B6,$63,$E9,$BF,$F4,$9F,$EC,$56,$73,$87,$1F,$AF,$F1,$C6,$33
	.dc.b	$F0,$F2,$FE,$D3,$EF,$9F,$A7,$3F,$4D,$7D,$03,$EF,$1F,$9E,$BF,$00
	.dc.b	$7E,$E3,$FB,$FC,$78,$E1,$C2,$F7,$4F,$CD,$75,$97,$6E,$9F,$9A,$7F
	.dc.b	$BA,$9F,$03,$8C,$39,$E7,$E1,$CD,$F9,$EF,$9E,$FC,$D3,$9F,$CD,$7F
	.dc.b	$86,$FD,$B7,$F8,$4F,$F0,$A7,$E3,$7F,$88,$FA,$8F,$D0,$FE,$1D,$FC
	.dc.b	$56,$B5,$39,$F2,$9F,$87,$E7,$E8,$5F,$65,$F8,$5B,$E9,$C3,$E8,$78
	.dc.b	$F8,$9F,$E2,$FF,$93,$F0,$DF,$FC,$01,$CB,$C1,$F2,$BF,$C8,$38,$DF
	.dc.b	$9F,$3F,$7E,$FF,$05,$7D,$CB,$F3,$37,$E4,$9B,$AD,$7D,$2F,$F9,$EE
	.dc.b	$73,$EA,$FF,$2D,$C7,$9F,$DA,$FE,$65,$EB,$95,$C3,$8E,$5D,$7E,$E7
	.dc.b	$A4,$F8,$31,$FF,$71,$78,$7F,$CC,$7F,$20,$7E,$E2,$FC,$61,$F7,$DF
	.dc.b	$E3,$DF,$9F,$7E,$71,$F9,$9C,$38,$5E,$EB,$52,$FC,$97,$FA,$16,$3D
	.dc.b	$3F,$47,$FD,$73,$DB,$2B,$87,$1E,$76,$7F,$D7,$7A,$F8,$CE,$61,$FA
	.dc.b	$7F,$F6,$AF,$A8,$FE,$1B,$F8,$8B,$F2,$C7,$E4,$9E,$9F,$DC,$DF,$8F
	.dc.b	$3F,$39,$FE,$C7,$87,$0B,$DE,$CB,$DF,$2E,$B6,$FE,$01,$F1,$6D,$3C
	.dc.b	$91,$FD,$7B,$9B,$FC,$5B,$39,$C7,$FB,$B6,$5D,$75,$F1,$7E,$3F,$E7
	.dc.b	$C3,$9E,$3E,$79,$F3,$7E,$0B,$99,$FE,$97,$EE,$BD,$27,$E9,$CF,$CE
	.dc.b	$BF,$43,$F9,$9B,$EC,$FC,$D8,$70,$BD,$D6,$BF,$72,$CF,$CD,$7F,$06
	.dc.b	$FC,$87,$CB,$7F,$C6,$E3,$F8,$4D,$E2,$39,$65,$DF,$D9,$7E,$89,$E5
	.dc.b	$BD,$CE,$79,$7D,$1F,$BC,$32,$FF,$37,$F5,$BF,$50,$FC,$E7,$E5,$5F
	.dc.b	$9F,$BF,$F2,$9D,$EF,$F0,$2F,$AB,$5A,$FD,$E3,$D7,$3C,$FE,$D7,$C8
	.dc.b	$EF,$7F,$EC,$7D,$5F,$CC,$D4,$73,$CF,$C1,$DA,$F7,$BC,$BB,$79,$33
	.dc.b	$FD,$77,$17,$CF,$FD,$D1,$F9,$8B,$F2,$A7,$B7,$FC,$03,$E1,$5F,$AF
	.dc.b	$FF,$66,$E4,$FF,$35,$E1,$C2,$F7,$EF,$7F,$23,$A7,$EC,$DF,$52,$BA
	.dc.b	$FB,$9F,$35,$E1,$C7,$EA,$3F,$A6,$3C,$F3,$F0,$7D,$A7,$97,$5F,$24
	.dc.b	$78,$FE,$2B,$F0,$DF,$AA,$7E,$17,$EB,$7F,$D6,$1E,$9F,$FA,$07,$F6
	.dc.b	$B7,$DA,$7E,$12,$B5,$EA,$AB,$5F,$6B,$9F,$17,$FC,$2D,$BF,$8C,$5F
	.dc.b	$E4,$7E,$D7,$F0,$8F,$2B,$FF,$47,$7C,$A7,$8A,$FF,$C3,$D8,$F5,$FC
	.dc.b	$0B,$FC,$9F,$8F,$F9,$43,$F1,$7F,$AD,$F6,$EF,$C7,$9F,$BF,$3F,$40
	.dc.b	$FF,$07,$FE,$CB,$FE,$D0,$F7,$38,$63,$FD,$BF,$0E,$17,$BF,$B3,$5F
	.dc.b	$FB,$5B,$E2,$37,$7F,$E3,$9C,$9F,$11,$FE,$67,$D3,$C3,$1E,$09,$FF
	.dc.b	$15,$C6,$31,$E9,$87,$F9,$33,$C2,$4E,$75,$FF,$11,$FE,$99,$FE,$4A
	.dc.b	$8F,$C8,$7E,$68,$FA,$DF,$F0,$AE,$1F,$DE,$3F,$69,$7E,$D7,$FE,$41
	.dc.b	$CB,$9E,$9F,$E0,$9C,$38,$7D,$87,$AB,$EA,$18,$70,$EB,$F5,$D6,$AF
	.dc.b	$CE,$71,$CD,$FA,$77,$B6,$31,$8F,$3C,$3F,$92,$C7,$2A,$F2,$E3,$DB
	.dc.b	$D6,$FF,$48,$ED,$FA,$9F,$8E,$7C,$E3,$F6,$1F,$E4,$FF,$B0,$7F,$42
	.dc.b	$FD,$8B,$FC,$0B,$7B,$AD,$7A,$EA,$D7,$E7,$5F,$4B,$1E,$7E,$17,$97
	.dc.b	$6F,$E9,$38,$75,$7D,$66,$33,$EF,$F2,$0F,$CC,$BC,$B3,$FE,$2F,$C3
	.dc.b	$CB,$5A,$E2,$CC,$BF,$09,$7F,$00,$BF,$3D,$FF,$0B,$F8,$2F,$81,$7E
	.dc.b	$F8,$FC,$E3,$F0,$F8,$70,$BD,$D6,$B9,$73,$8F,$FA,$0F,$2F,$3D,$ED
	.dc.b	$B1,$DF,$EA,$3F,$BD,$DB,$F0,$3F,$2D,$F5,$9C,$FD,$B7,$FB,$03,$F3
	.dc.b	$9F,$E3,$1F,$AD,$7F,$3B,$E4,$FE,$99,$E2,$BF,$19,$7E,$4E,$FD,$EB
	.dc.b	$FD,$73,$87,$0F,$91,$FE,$EB,$F2,$EC,$38,$5B,$F4,$AD,$F1,$7F,$95
	.dc.b	$F4,$FE,$91,$FB,$CF,$FB,$27,$2E,$5A,$DA,$D6,$7D,$32,$F1,$7C,$FF
	.dc.b	$9E,$73,$5E,$6D,$BF,$BA,$79,$6C,$7F,$99,$7E,$75,$FD,$FF,$F5,$87
	.dc.b	$B0,$FC,$DF,$FB,$BF,$ED,$EF,$C2,$BC,$BA,$30,$E1,$7B,$D5,$6F,$94
	.dc.b	$DD,$75,$BF,$F5,$1E,$7E,$0F,$A7,$7E,$B3,$EB,$9E,$90,$D7,$C7,$AF
	.dc.b	$F8,$23,$E6,$FD,$95,$A9,$CE,$8F,$CE,$33,$E7,$3F,$2C,$FF,$20,$FC
	.dc.b	$DF,$F7,$2F,$85,$FF,$10,$FC,$F7,$AA,$B5,$F2,$6A,$D7,$D5,$BF,$30
	.dc.b	$D6,$3F,$BE,$D7,$D2,$7F,$A2,$F0,$E2,$E0,$FF,$0B,$C7,$83,$E7,$FF
	.dc.b	$99,$FE,$1B,$E4,$8C,$69,$FB,$EE,$AF,$A5,$7F,$E9,$5F,$D8,$DE,$37
	.dc.b	$FB,$DA,$73,$29,$CF,$D9,$FC,$77,$71,$FE,$F8,$FC,$4F,$E2,$FC,$73
	.dc.b	$F3,$8F,$9B,$F8,$9F,$FB,$7B,$F7,$DF,$F7,$37,$D1,$7D,$AA,$D7,$6F
	.dc.b	$FB,$0F,$E9,$D5,$FE,$EB,$F6,$F5,$8F,$D6,$CF,$DD,$87,$5F,$CB,$3E
	.dc.b	$C3,$CB,$3E,$DF,$8E,$F8,$AF,$C3,$97,$4F,$9D,$FF,$59,$70,$B7,$FF
	.dc.b	$01,$7C,$5F,$F5,$0D,$3F,$C0,$87,$E3,$E7,$4F,$FA,$E6,$34,$FF,$82
	.dc.b	$7C,$64,$EA,$13,$8E,$3E,$71,$E0,$BE,$F5,$87,$11,$D7,$F2,$EE,$37
	.dc.b	$0F,$F5,$4D,$75,$51,$97,$BF,$E9,$3E,$2B,$F5,$CF,$8A,$FF,$E0,$0C
	.dc.b	$62,$BA,$6B,$E0,$9F,$3F,$4F,$99,$F8,$15,$F1,$23,$19,$F5,$9F,$85
	.dc.b	$0C,$F8,$67,$E5,$8F,$64,$EC,$EA,$7C,$2F,$CA,$B1,$97,$F9,$8A,$3F
	.dc.b	$C6,$5E,$86,$2B,$F2,$19,$7D,$4F,$B3,$F7,$5B,$F5,$4E,$31,$8F,$9D
	.dc.b	$B3,$E2,$AE,$0F,$BB,$1E,$FD,$9E,$DF,$D5,$2C,$7D,$4F,$CA,$33,$BF
	.dc.b	$18,$FB,$D5,$EA,$DF,$E8,$7E,$B3,$E2,$D7,$D2,$F8,$63,$39,$71,$BF
	.dc.b	$C0,$CF,$D5,$E2,$F9,$DF,$A1,$BF,$14,$7D,$BB,$F6,$F7,$EA,$EF,$D5
	.dc.b	$1F,$BF,$BF,$80,$3E,$56,$3C,$70,$E1,$7B,$A7,$6E,$B2,$EB,$AF,$E0
	.dc.b	$00,$C3,$C5,$FB,$77,$9D,$E9,$A7,$8B,$47,$E4,$5E,$8C,$3C,$F3,$9F
	.dc.b	$CB,$7F,$11,$F4,$4F,$D5,$7F,$D9,$FF,$B6,$3F,$6E,$7E,$CF,$FE,$AE
	.dc.b	$E3,$FC,$15,$EF,$F1,$05,$EE,$BF,$D6,$7C,$3A,$C7,$FD,$B3,$C3,$E9
	.dc.b	$3F,$C7,$CB,$F1,$3F,$9D,$8F,$3C,$39,$7B,$9E,$4A,$F1,$E9,$E2,$FA
	.dc.b	$1F,$DF,$FE,$B9,$FE,$7F,$FC,$97,$97,$E5,$1F,$E3,$1F,$C9,$BF,$8F
	.dc.b	$F2,$FF,$AF,$BB,$D7,$BD,$97,$BF,$07,$E2,$FE,$51,$FC,$02,$B5,$F9
	.dc.b	$47,$E8,$9E,$15,$78,$BE,$8E,$9E,$0F,$A9,$6B,$E0,$9E,$46,$B3,$EF
	.dc.b	$3A,$CD,$F9,$47,$AA,$BE,$DF,$E2,$DF,$EF,$6F,$C6,$3F,$DF,$9D,$39
	.dc.b	$94,$E7,$ED,$1F,$84,$FF,$C0,$0E,$2F,$6D,$72,$FA,$57,$E3,$1E,$78
	.dc.b	$E3,$0F,$C5,$F1,$8F,$3C,$1F,$8C,$77,$8C,$7E,$2F,$FC,$B3,$D7,$FF
	.dc.b	$56,$7E,$1D,$FE,$AB,$FC,$37,$FD,$73,$8F,$F2,$69,$CF,$C2,$FF,$AA
	.dc.b	$F2,$FE,$C1,$F4,$D9,$D7,$F2,$F8,$FF,$AE,$5F,$E5,$3B,$B0,$E3,$DD
	.dc.b	$FC,$41,$BF,$17,$FE,$B8,$FC,$3B,$C5,$7C,$B3,$ED,$F8,$76,$21,$DF
	.dc.b	$FC,$7D,$FE,$7D,$87,$FB,$0E,$31,$EF,$F8,$E7,$67,$F1,$F3,$C7,$3B
	.dc.b	$7F,$83,$35,$FE,$7F,$EC,$7F,$44,$F7,$FF,$D9,$DF,$82,$FF,$47,$FB
	.dc.b	$B7,$E1,$3F,$B3,$FD,$47,$58,$7E,$DA,$B5,$3F,$B7,$F3,$F3,$3F,$EA
	.dc.b	$EC,$7F,$97,$E3,$D2,$3F,$7B,$9F,$FA,$22,$DF,$6D,$FD,$7A,$B5,$5F
	.dc.b	$E2,$3C,$FC,$3F,$37,$98,$FF,$1F,$7B,$DF,$D6,$3E,$AF,$F0,$FE,$0B
	.dc.b	$F3,$F7,$C3,$F8,$BE,$A3,$FA,$AD,$6A,$5F,$11,$9C,$3F,$A8,$63,$FD
	.dc.b	$AB,$AD,$AF,$F9,$62,$7B,$77,$7F,$AB,$BE,$97,$C5,$6C,$FA,$CF,$FA
	.dc.b	$CB,$F6,$CC,$FE,$11,$F1,$3E,$01,$D3,$F1,$23,$1D,$17,$1F,$AC,$78
	.dc.b	$74,$FF,$7F,$4F,$CC,$E5,$E7,$3F,$23,$97,$E1,$BC,$4F,$3F,$E1,$2F
	.dc.b	$DB,$DF,$89,$3F,$1E,$D6,$4A,$D4,$BF,$5E,$CC,$71,$F0,$7B,$E3,$D3
	.dc.b	$17,$87,$8C,$79,$43,$F6,$17,$EB,$DC,$A3,$1F,$1D,$E5,$C3,$FC,$2B
	.dc.b	$F8,$B5,$7F,$9E,$8F,$C5,$30,$FF,$86,$3A,$FE,$8B,$FD,$0B,$8F,$F4
	.dc.b	$1E,$27,$E3,$3E,$53,$FC,$1F,$8B,$F7,$67,$D9,$7F,$94,$7F,$38,$D7
	.dc.b	$81,$FE,$1C,$AD,$4F,$F9,$57,$F2,$6C,$CF,$E5,$33,$FE,$9F,$8F,$4C
	.dc.b	$79,$FB,$7F,$7A,$31,$87,$F4,$EA,$FC,$EE,$3E,$AF,$96,$7E,$AC,$FC
	.dc.b	$81,$FC,$02,$E3,$DE,$1F,$D7,$2F,$75,$AF,$DB,$73,$C3,$E3,$AF,$F0
	.dc.b	$41,$BC,$7A,$7C,$7F,$E4,$33,$F0,$47,$7F,$C6,$7F,$27,$D1,$E6,$FA
	.dc.b	$C7,$2F,$EC,$A3,$1C,$B7,$97,$F3,$F9,$FD,$2E,$7F,$9E,$FD,$A8,$F9
	.dc.b	$1C,$DF,$0A,$38,$47,$F8,$6E,$3E,$7F,$C6,$7E,$8B,$F1,$1F,$F7,$EF
	.dc.b	$E2,$4F,$C4,$3F,$DF,$BE,$33,$D7,$FE,$07,$5A,$9C,$E3,$8F,$89,$7E
	.dc.b	$15,$F0,$69,$1F,$F0,$03,$4F,$E5,$7E,$1F,$E2,$27,$97,$E2,$3F,$E8
	.dc.b	$38,$C5,$7F,$C0,$08,$7B,$FF,$E6,$8F,$D9,$3F,$D5,$1F,$71,$FE,$A0
	.dc.b	$FB,$DF,$F3,$34,$63,$EF,$BF,$D6,$51,$FE,$B1,$E4,$4F,$0D,$3F,$9D
	.dc.b	$BF,$EB,$1F,$B1,$FF,$24,$BC,$7D,$EF,$F5,$74,$FF,$9B,$23,$1F,$33
	.dc.b	$FE,$AD,$8F,$F3,$7F,$D3,$FE,$83,$E7,$3F,$C0,$5F,$35,$FE,$10,$AF
	.dc.b	$F1,$46,$4A,$D7,$CD,$FF,$81,$3D,$8F,$F0,$7C,$E7,$1F,$F1,$2E,$3F
	.dc.b	$E1,$57,$FA,$7D,$F0,$F9,$6F,$F2,$EB,$CB,$E5,$3F,$C3,$39,$F7,$AF
	.dc.b	$19,$CB,$2E,$DC,$5F,$E8,$E8,$FD,$AF,$EF,$2F,$C3,$FF,$AD,$7D,$AB
	.dc.b	$EC,$1F,$5C,$FA,$77,$93,$FE,$8E,$C3,$87,$C7,$30,$E1,$7B,$FC,$B7
	.dc.b	$59,$7F,$7C,$CF,$F9,$81,$E0,$7F,$B4,$3F,$DA,$8C,$5F,$3C,$FC,$3F
	.dc.b	$57,$E7,$9F,$F2,$3E,$5E,$0C,$BB,$FA,$59,$94,$E6,$FD,$3A,$BF,$2A
	.dc.b	$FF,$52,$F8,$5F,$8C,$F1,$FF,$31,$D1,$FE,$AC,$8C,$78,$BA,$F0,$FF
	.dc.b	$2F,$9F,$94,$FF,$2B,$CF,$E6,$3F,$16,$FF,$46,$C7,$E5,$3C,$67,$93
	.dc.b	$FA,$5F,$1B,$FD,$B5,$39,$8C,$74,$79,$FF,$68,$F3,$74,$7C,$DE,$9F
	.dc.b	$F5,$CC,$63,$85,$D3,$8E,$9F,$95,$97,$EE,$CF,$E5,$B5,$F4,$9F,$30
	.dc.b	$7A,$F4,$77,$9F,$BB,$CB,$F7,$39,$F1,$D3,$F3,$F4,$FB,$DC,$FF,$EC
	.dc.b	$AF,$90,$FF,$3A,$FA,$BF,$EB,$1F,$B2,$FE,$D3,$F9,$33,$E1,$1F,$A0
	.dc.b	$3F,$74,$FE,$CD,$F9,$57,$F9,$BF,$ED,$FF,$D8,$7D,$DF,$D0,$B0,$E1
	.dc.b	$7B,$F9,$25,$FF,$9D,$7C,$DD,$F3,$7F,$A1,$2F,$6F,$FB,$5B,$4F,$0E
	.dc.b	$8F,$A0,$F2,$67,$D6,$3C,$7A,$F8,$BE,$9F,$FD,$1B,$7C,$35,$F0,$68
	.dc.b	$FC,$47,$38,$F1,$9C,$F9,$2F,$C2,$7C,$7A,$FE,$87,$E4,$FD,$DF,$A8
	.dc.b	$3F,$4D,$7D,$FB,$F0,$77,$93,$C3,$C3,$FF,$01,$F1,$30,$E1,$7B,$FD
	.dc.b	$83,$59,$3F,$67,$F5,$E9,$FE,$A4,$C7,$8B,$F0,$78,$3F,$B5,$C7,$4C
	.dc.b	$79,$FC,$7F,$AA,$B8,$61,$E6,$DB,$CB,$F0,$39,$9F,$56,$BE,$39,$F4
	.dc.b	$EE,$FD,$0F,$FD,$F3,$F0,$5F,$DE,$3F,$1C,$FF,$04,$7E,$EC,$FC,$0D
	.dc.b	$F0,$FF,$E6,$EF,$C7,$BF,$CE,$5D,$9F,$D0,$75,$FF,$5D,$E1,$C2,$F7
	.dc.b	$F6,$2B,$FF,$3D,$DE,$ED,$7D,$9F,$BD,$97,$87,$E7,$3F,$57,$18,$3A
	.dc.b	$AC,$FC,$99,$FF,$A0,$3E,$43,$EF,$DF,$19,$CE,$7F,$C6,$30,$F6,$9C
	.dc.b	$A3,$39,$BF,$7C,$CC,$F9,$32,$ED,$AF,$AA,$9F,$B3,$6B,$EB,$1E,$E3
	.dc.b	$F0,$F7,$E8,$AF,$9B,$7E,$86,$FD,$47,$F9,$57,$2F,$C1,$C7,$8E,$1C
	.dc.b	$2F,$74,$FC,$BB,$B6,$5D,$77,$F8,$47,$FA,$19,$7B,$FF,$B5,$E4,$FF
	.dc.b	$43,$47,$8F,$5F,$16,$AF,$8C,$F8,$B8,$7F,$BC,$35,$F0,$D6,$BD,$4C
	.dc.b	$CF,$5D,$BC,$98,$7A,$7F,$68,$F1,$C6,$36,$7E,$C4,$E1,$FF,$33,$61
	.dc.b	$FC,$9B,$EB,$9F,$A2,$3E,$49,$F8,$73,$F5,$F7,$77,$F9,$5F,$1E,$38
	.dc.b	$FE,$AB,$1E,$38,$70,$BD,$FE,$8D,$DF,$DA,$FF,$03,$53,$F7,$FF,$E2
	.dc.b	$D1,$8D,$2E,$3E,$37,$C5,$A7,$87,$4F,$E0,$B5,$BC,$FB,$7E,$21,$8C
	.dc.b	$38,$47,$86,$9F,$BD,$F8,$67,$38,$73,$E8,$C4,$7A,$72,$EF,$B7,$BB
	.dc.b	$F6,$4C,$FA,$9F,$E9,$49,$CE,$4F,$D4,$B8,$FE,$30,$F9,$A7,$E5,$8F
	.dc.b	$D5,$FF,$B7,$7C,$1B,$F3,$F7,$EB,$7C,$2F,$75,$A9,$FD,$C7,$4F,$CC
	.dc.b	$C7,$A6,$3F,$13,$81,$1D,$FF,$2E,$C3,$AC,$3B,$FE,$47,$F0,$EB,$79
	.dc.b	$76,$F9,$B6,$BF,$E2,$1C,$FB,$4E,$63,$CD,$F2,$DE,$5D,$3D,$57,$C3
	.dc.b	$5F,$DD,$60,$FB,$2F,$F4,$3F,$E9,$6F,$DD,$DF,$94,$BE,$0F,$E2,$7D
	.dc.b	$83,$E1,$DF,$9C,$7E,$47,$1E,$38,$70,$BD,$D3,$F4,$FD,$71,$06,$7D
	.dc.b	$FF,$5B,$FE,$BD,$8F,$2D,$2D,$6E,$BC,$91,$E3,$E3,$6A,$F8,$EF,$E9
	.dc.b	$F8,$71,$D3,$CF,$F1,$B8,$CB,$C1,$E7,$BE,$B1,$FB,$03,$F3,$2F,$E1
	.dc.b	$7F,$87,$3F,$44,$7A,$7E,$8C,$7F,$DB,$F8,$70,$BD,$D7,$EC,$B9,$FE
	.dc.b	$55,$76,$A7,$EC,$1E,$A8,$E3,$95,$CF,$C4,$F8,$7C,$96,$8F,$8C,$F4
	.dc.b	$5F,$0D,$3C,$BF,$17,$8C,$78,$C7,$8B,$F3,$5F,$34,$63,$1E,$7B,$3F
	.dc.b	$58,$FE,$0D,$E8,$FF,$B1,$3F,$26,$FE,$75,$F6,$68,$7E,$72,$FC,$D9
	.dc.b	$F8,$5B,$F5,$8E,$3C,$1F,$67,$1E,$38,$70,$BD,$F8,$7A,$A5,$7F,$47
	.dc.b	$F1,$7F,$C3,$57,$E1,$F8,$C6,$98,$BE,$91,$89,$E9,$F8,$67,$F8,$29
	.dc.b	$E1,$5E,$7D,$7C,$DE,$5F,$F8,$3B,$E9,$3F,$DB,$FE,$F3,$F1,$4F,$0D
	.dc.b	$BF,$76,$7E,$C4,$FC,$C5,$FC,$01,$FC,$A7,$1E,$38,$70,$BD,$D3,$F1
	.dc.b	$2D,$70,$7E,$67,$3E,$DF,$8B,$7F,$8D,$71,$E3,$C7,$6B,$59,$73,$D3
	.dc.b	$F6,$D5,$AD,$7C,$91,$FE,$DB,$AF,$E7,$5F,$9F,$7F,$BE,$3E,$1F,$FA
	.dc.b	$97,$EE,$EF,$C8,$BF,$AE,$FD,$77,$DB,$FC,$7B,$78,$7E,$AE,$F7,$5A
	.dc.b	$EE,$FE,$D9,$E5,$32,$E9,$F2,$8C,$FF,$C9,$91,$8F,$A0,$B9,$73,$AD
	.dc.b	$63,$E0,$F1,$7F,$C7,$77,$CB,$0F,$27,$C8,$FF,$9F,$7E,$D3,$FC,$7F
	.dc.b	$F9,$57,$EF,$72,$79,$1F,$DE,$3E,$13,$FD,$A3,$F3,$2F,$DF,$7F,$61
	.dc.b	$F9,$AF,$F5,$9C,$38,$5E,$FE,$F3,$FB,$FD,$7E,$D7,$EE,$75,$97,$EE
	.dc.b	$B1,$FE,$62,$FF,$95,$70,$E3,$1F,$E2,$78,$C1,$BC,$BC,$37,$FE,$13
	.dc.b	$E3,$5F,$09,$CF,$D8,$FF,$96,$3E,$C7,$D5,$F6,$1F,$88,$FD,$D9,$FA
	.dc.b	$73,$F3,$47,$C7,$FF,$71,$FD,$BB,$FB,$96,$1C,$3C,$06,$1C,$2F,$7A
	.dc.b	$FF,$86,$8F,$F3,$8E,$FF,$C8,$F3,$FC,$AF,$C7,$FF,$9F,$F1,$E5,$85
	.dc.b	$AD,$63,$D3,$4F,$16,$9F,$E2,$29,$F0,$9A,$AF,$04,$3E,$3F,$FC,$E7
	.dc.b	$F7,$FF,$57,$EF,$1F,$56,$F4,$FF,$D6,$DF,$A7,$BF,$6A,$7E,$B6,$FC
	.dc.b	$83,$FC,$1D,$8F,$1C,$38,$79,$4C,$38,$5B,$F1,$ED,$EF,$FC,$FF,$4E
	.dc.b	$FE,$F7,$FA,$6F,$2E,$5A,$DA,$D6,$7D,$32,$F1,$7E,$69,$E8,$9C,$D7
	.dc.b	$9B,$67,$C1,$FF,$A3,$BF,$0A,$FC,$87,$D5,$BF,$45,$FF,$21,$FC,$77
	.dc.b	$F7,$DF,$BF,$FF,$53,$F9,$DF,$E2,$7B,$DD,$6B,$E7,$BF,$D5,$74,$FD
	.dc.b	$B3,$5A,$FF,$57,$CB,$A7,$E2,$DF,$EB,$3C,$78,$F9,$D6,$B5,$97,$3C
	.dc.b	$5F,$92,$79,$67,$35,$FD,$AF,$3F,$F5,$CF,$D3,$FF,$A9,$BF,$53,$FD
	.dc.b	$47,$EA,$BF,$D9,$3F,$14,$7E,$B7,$FB,$7F,$C5,$3F,$C5,$DF,$9F,$7F
	.dc.b	$8F,$EF,$7F,$40,$BD,$FB,$0B,$FF,$1A,$7C,$86,$F6,$FE,$E1,$9F,$57
	.dc.b	$FD,$79,$8F,$2F,$C5,$7F,$D9,$CF,$4C,$79,$E1,$FE,$0E,$AE,$42,$FC
	.dc.b	$DA,$79,$7A,$FF,$D7,$FE,$17,$F0,$1F,$D4,$9E,$EF,$FA,$D3,$F2,$F7
	.dc.b	$E2,$0F,$C8,$10,$DF,$FC,$B5,$87,$1B,$E1,$7B,$AF,$77,$A7,$F8,$DD
	.dc.b	$7F,$C1,$59,$76,$FD,$53,$FD,$CD,$8F,$1F,$B1,$B5,$AC,$B9,$E9,$FE
	.dc.b	$27,$AD,$65,$E4,$8F,$D3,$57,$F9,$AF,$F5,$6F,$E3,$5F,$8B,$BF,$3A
	.dc.b	$FF,$46,$7E,$33,$EA,$D1,$FB,$C2,$F8,$61,$FE,$1D,$C3,$87,$E0,$8C
	.dc.b	$38,$5F,$C9,$E0,$E1,$7B,$CB,$FD,$30,$FF,$BB,$F1,$F0,$C7,$F8,$32
	.dc.b	$3F,$96,$1D,$86,$3D,$B0,$F2,$79,$7F,$EF,$0B,$E3,$97,$87,$89,$A3
	.dc.b	$F2,$BF,$F7,$AF,$E7,$7F,$CD,$FD,$A7,$E5,$4F,$D6,$DF,$BE,$FF,$95
	.dc.b	$B0,$FF,$77,$6B,$FE,$8F,$C3,$85,$EE,$B5,$F8,$8B,$F2,$9D,$70,$FF
	.dc.b	$8F,$F3,$EB,$9F,$AA,$3C,$26,$70,$FC,$FB,$9E,$36,$5D,$AB,$C7,$F7
	.dc.b	$9E,$AF,$AA,$E3,$7F,$DF,$30,$F1,$4F,$BB,$D4,$FF,$27,$62,$EC,$FE
	.dc.b	$A9,$F7,$EF,$DA,$5F,$BA,$BF,$57,$7C,$A3,$E4,$5C,$6F,$D1,$FF,$D5
	.dc.b	$D9,$72,$C7,$8E,$1C,$2F,$7F,$A4,$EB,$83,$FC,$BD,$A7,$7F,$D4,$BE
	.dc.b	$04,$64,$47,$8F,$5C,$ED,$E7,$9E,$39,$76,$DB,$C9,$C5,$EC,$D7,$C5
	.dc.b	$87,$A6,$7D,$1C,$1F,$03,$F2,$2F,$EE,$5F,$38,$F9,$1F,$F0,$17,$ED
	.dc.b	$0F,$C8,$DF,$99,$BF,$58,$7E,$B3,$FE,$CB,$C7,$8E,$1C,$2F,$74,$FD
	.dc.b	$13,$5C,$3F,$E7,$BC,$FB,$7D,$07,$AB,$1E,$3F,$51,$6B,$59,$73,$D1
	.dc.b	$F9,$EF,$9E,$73,$83,$6F,$35,$7F,$8C,$67,$FD,$BD,$F9,$A7,$F8,$37
	.dc.b	$EC,$DF,$9F,$3E,$21,$F9,$8B,$F2,$FF,$B9,$F8,$AB,$F2,$2E,$39,$7F
	.dc.b	$A4,$71,$E3,$87,$0B,$DF,$95,$AA,$FF,$32,$65,$D5,$FC,$A6,$7E,$18
	.dc.b	$7E,$FB,$F6,$23,$18,$73,$C7,$FC,$45,$DE,$9C,$EB,$E5,$F1,$3F,$CE
	.dc.b	$19,$7F,$BE,$7A,$7F,$C5,$1F,$2C,$FD,$7D,$FA,$DB,$E8,$5F,$BF,$3F
	.dc.b	$4B,$FF,$06,$FE,$39,$C7,$2F,$F5,$66,$3C,$70,$E1,$7B,$FA,$3F,$16
	.dc.b	$FF,$81,$BD,$E6,$FD,$13,$F2,$3A,$F8,$A3,$C3,$3F,$B3,$8C,$63,$D3
	.dc.b	$4F,$E1,$53,$E1,$3F,$D0,$F5,$E8,$D7,$CF,$3F,$D8,$A5,$8B,$F2,$9F
	.dc.b	$F2,$37,$E8,$BF,$D1,$3F,$0E,$FB,$5F,$24,$FC,$51,$FA,$AB,$F5,$0E
	.dc.b	$38,$70,$F8,$06,$1C,$2F,$7D,$1A,$D3,$FD,$9D,$9F,$57,$F8,$66,$1C
	.dc.b	$79,$3F,$5E,$F3,$C3,$97,$C1,$FF,$08,$8C,$75,$B6,$F2,$79,$BF,$C3
	.dc.b	$3E,$FF,$FC,$BD,$F2,$9F,$E4,$4F,$D7,$9F,$98,$FF,$91,$BF,$5C,$FE
	.dc.b	$55,$F3,$DF,$02,$F7,$F5,$EB,$DA,$AF,$E6,$1F,$AC,$EF,$87,$FD,$CD
	.dc.b	$97,$F9,$43,$B3,$F8,$2E,$3C,$B1,$B5,$AC,$7A,$67,$E1,$FD,$E7,$CD
	.dc.b	$39,$AE,$F9,$FF,$B3,$63,$E9,$FE,$21,$FE,$77,$FB,$17,$BF,$F9,$CF
	.dc.b	$DF,$7F,$09,$F0,$CF,$C8,$DF,$47,$C3,$1F,$F7,$D6,$1C,$3C,$D6,$1C
	.dc.b	$2F,$F8,$EF,$85,$FE,$55,$BF,$F2,$96,$5F,$A8,$7F,$9F,$67,$E1,$87
	.dc.b	$EC,$1F,$C8,$A3,$18,$F4,$F2,$3F,$AA,$E3,$FC,$F2,$73,$97,$9B,$CE
	.dc.b	$FE,$7F,$9E,$2F,$7A,$FE,$42,$31,$B7,$AF,$E1,$31,$B7,$AB,$6F,$23
	.dc.b	$FD,$BA,$BF,$2E,$E7,$6F,$E6,$B1,$CB,$5E,$F3,$C7,$6F,$36,$3F,$BC
	.dc.b	$FB,$8F,$81,$C1,$F1,$77,$F8,$9F,$BD,$7F,$A2,$78,$79,$6F,$72,$E4
	.dc.b	$7F,$65,$C3,$8E,$DF,$12,$31,$5E,$17,$81,$F3,$33,$F0,$F0,$A3,$FD
	.dc.b	$2D,$FB,$9F,$FA,$CF,$D7,$FF,$70,$FE,$68,$FD,$4B,$FC,$ED,$F0,$BF
	.dc.b	$EE,$4F,$57,$78,$71,$BD,$D6,$BE,$45,$96,$A9,$5D,$1F,$46,$5D,$E9
	.dc.b	$FB,$CF,$F6,$0D,$D3,$9D,$3C,$5F,$3D,$FD,$C3,$ED,$BD,$15,$AC,$7C
	.dc.b	$58,$AF,$7B,$7E,$4E,$7F,$D6,$5E,$4F,$ED,$BC,$8F,$E1,$1E,$93,$F7
	.dc.b	$DC,$D7,$F0,$2F,$4B,$F8,$3F,$99,$FB,$BF,$4B,$F8,$6F,$EF,$DA,$A7
	.dc.b	$A5,$FC,$1D,$FF,$17,$EF,$FC,$5D,$E5,$E9,$7A,$F5,$F0,$60,$F4,$7D
	.dc.b	$31,$8F,$47,$D7,$EC,$7F,$28,$F5,$FF,$91,$FA,$6F,$C3,$7F,$D9,$53
	.dc.b	$9C,$BF,$9E,$4B,$F0,$FC,$E9,$FC,$DB,$CC,$F9,$57,$FC,$CA,$3F,$0C
	.dc.b	$F1,$F6,$7F,$97,$6F,$FC,$F6,$7F,$99,$FA,$9F,$CD,$FE,$BF,$E6,$FB
	.dc.b	$3F,$D4,$BE,$0B,$FC,$41,$F1,$FD,$F0,$FE,$AD,$87,$F6,$5F,$82,$FE
	.dc.b	$D9,$F0,$5F,$D5,$E9,$F1,$5F,$EE,$1C,$7F,$AF,$DF,$E5,$DF,$EC,$9E
	.dc.b	$57,$DF,$7B,$79,$4F,$BB,$F0,$D7,$5E,$DF,$EE,$12,$FD,$83,$F0,$3F
	.dc.b	$1B,$FE,$10,$F7,$9C,$3F,$DB,$BE,$AF,$FD,$57,$E0,$7F,$BC,$6B,$53
	.dc.b	$9F,$B3,$7E,$17,$9D,$3F,$C1,$F8,$FE,$46,$FF,$C3,$7F,$A9,$D9,$CE
	.dc.b	$1C,$5E,$03,$E4,$BF,$C5,$17,$E2,$E1,$FF,$59,$FD,$1F,$E5,$BE,$5F
	.dc.b	$FC,$85,$F8,$67,$F8,$5F,$F0,$7F,$F4,$DF,$EF,$FF,$97,$EB,$FF,$1D
	.dc.b	$67,$FE,$3E,$F9,$C5,$EF,$EB,$57,$FA,$6B,$DF,$C2,$FF,$95,$6B,$F5
	.dc.b	$19,$7F,$93,$FE,$7E,$E5,$FE,$44,$9E,$5F,$55,$E3,$8C,$7C,$C3,$5F
	.dc.b	$F5,$6F,$D7,$7E,$B3,$E9,$BF,$CD,$5F,$4F,$FD,$5B,$F2,$77,$DF,$3F
	.dc.b	$12,$70,$7F,$9C,$33,$FF,$3A,$7D,$57,$FA,$0B,$E4,$D8,$70,$FC,$6F
	.dc.b	$79,$FF,$9D,$6F,$F8,$C3,$FE,$84,$F0,$7F,$80,$3E,$0B,$E3,$FA,$3F
	.dc.b	$8E,$BA,$EB,$E3,$FA,$3F,$E0,$3F,$61,$FE,$98,$FC,$41,$FA,$13,$F3
	.dc.b	$37,$E2,$0F,$CD,$3C,$F7,$BA,$D4,$BF,$27,$CE,$7F,$EA,$DF,$B6,$FE
	.dc.b	$25,$7F,$EA,$CF,$CA,$6C,$E7,$3E,$F9,$FF,$0D,$E5,$FF,$5A,$CF,$FA
	.dc.b	$BB,$EE,$18,$AF,$7F,$1B,$FC,$63,$E3,$BF,$DA,$1F,$A2,$3F,$6B,$7E
	.dc.b	$1E,$FC,$FB,$FB,$56,$9F,$ED,$9D,$1F,$7F,$FE,$E6,$F1,$98,$F1,$C3
	.dc.b	$87,$ED,$9F,$53,$3F,$F7,$27,$DF,$7F,$2B,$7F,$DB,$D8,$FF,$47,$7C
	.dc.b	$3A,$78,$3E,$8B,$C9,$3F,$EE,$1C,$BA,$CF,$93,$6F,$E5,$FF,$83,$7F
	.dc.b	$BF,$FC,$77,$E2,$BF,$E6,$9F,$A4,$7E,$B5,$F8,$4A,$D4,$E7,$F1,$2C
	.dc.b	$67,$FD,$3F,$F0,$77,$E1,$6F,$C2,$BF,$A0,$61,$FD,$4E,$95,$AF,$C2
	.dc.b	$33,$1C,$6F,$86,$9E,$2F,$8A,$FF,$26,$F4,$31,$F0,$C7,$F1,$EF,$C3
	.dc.b	$BA,$7E,$1E,$FC,$3D,$F8,$7B,$ED,$1F,$A6,$BF,$3C,$FF,$7B,$7E,$61
	.dc.b	$F9,$EC,$78,$E1,$C2,$F7,$5F,$EF,$BF,$C8,$35,$C1,$FD,$AF,$3E,$DF
	.dc.b	$1B,$FD,$6F,$1E,$3A,$5A,$D6,$5C,$F4,$F0,$7E,$AF,$FD,$72,$F7,$A7
	.dc.b	$6F,$C7,$7F,$AF,$F5,$FC,$5F,$90,$7E,$33,$FE,$65,$F1,$7F,$D8,$BF
	.dc.b	$95,$FF,$6A,$FC,$6F,$F6,$B7,$BA,$D7,$C2,$2B,$FB,$7C,$EB,$7F,$F0
	.dc.b	$16,$3D,$3F,$16,$FE,$E5,$3C,$7E,$92,$D6,$B1,$E7,$87,$F4,$AA,$D6
	.dc.b	$7E,$2F,$8A,$FE,$EB,$F4,$7F,$6F,$F1,$E7,$D7,$BE,$A9,$4F,$C1,$FF
	.dc.b	$CF,$3F,$B8,$BF,$33,$FE,$1B,$8F,$1C,$38,$5E,$FE,$BD,$7F,$C4,$BC
	.dc.b	$3F,$F5,$07,$07,$F8,$6F,$4E,$FA,$7F,$82,$F2,$E5,$F4,$96,$B5,$9F
	.dc.b	$4D,$7F,$D1,$3F,$59,$E8,$9C,$E1,$E6,$9F,$F4,$6E,$BF,$E1,$0F,$7F
	.dc.b	$F2,$DF,$91,$FF,$B9,$FE,$B1,$F8,$D7,$F8,$AF,$F1,$6F,$D2,$F0,$FF
	.dc.b	$26,$BD,$FE,$23,$FD,$EE,$B5,$E9,$3F,$1A,$D6,$DF,$E3,$3C,$FA,$3F
	.dc.b	$E3,$DC,$38,$FE,$D1,$FE,$25,$79,$E9,$E0,$96,$9F,$A2,$9C,$E5,$E3
	.dc.b	$FA,$3F,$F1,$3F,$E2,$7F,$AD,$F3,$07,$F8,$4F,$EB,$DF,$A3,$7F,$B2
	.dc.b	$3F,$2A,$7E,$AD,$FC,$FA,$F8,$5E,$EB,$5E,$E6,$75,$FF,$25,$4F,$4E
	.dc.b	$8F,$F1,$D5,$F0,$F7,$6D,$6A,$B9,$7D,$AF,$F7,$B8,$E4,$79,$23,$3E
	.dc.b	$47,$F8,$FB,$8B,$F8,$27,$DE,$BE,$55,$FA,$AF,$FA,$8F,$F7,$EF,$F5
	.dc.b	$37,$D8,$3F,$5A,$FE,$AB,$87,$0B,$DF,$E1,$AB,$DF,$9B,$AD,$7F,$CC
	.dc.b	$59,$F5,$CF,$FC,$A5,$8F,$1F,$CE,$AD,$6B,$2E,$78,$FF,$AD,$7F,$32
	.dc.b	$F3,$4E,$70,$F2,$7E,$20,$F4,$BF,$CA,$FF,$83,$FF,$1D,$F8,$B6,$0F
	.dc.b	$C8,$7F,$B6,$FE,$7F,$FC,$93,$F3,$A7,$DC,$FF,$6E,$C3,$85,$EF,$EA
	.dc.b	$17,$BF,$1B,$5B,$7F,$9E,$F3,$EB,$9F,$F9,$AF,$1E,$3F,$9C,$5A,$D6
	.dc.b	$5C,$F1,$FE,$67,$F8,$1F,$9A,$73,$87,$93,$E3,$3F,$D8,$5E,$B7,$F5
	.dc.b	$C9,$FF,$44,$7C,$C3,$F4,$4F,$A1,$FB,$7B,$F0,$07,$EA,$7F,$E0,$7F
	.dc.b	$DC,$BF,$C0,$FF,$3F,$FC,$F7,$0E,$1E,$0B,$0E,$17,$FE,$0D,$F7,$F7
	.dc.b	$9F,$6C,$DF,$A3,$FF,$A1,$72,$E5,$1F,$7E,$7F,$D4,$51,$8C,$7A,$69
	.dc.b	$E2,$FC,$5F,$FD,$0F,$39,$CF,$BF,$E0,$58,$C3,$B6,$1C,$34,$FF,$0F
	.dc.b	$7E,$EF,$E2,$9F,$EE,$99,$FB,$B3,$F6,$FD,$9B,$E6,$7F,$C4,$7F,$4F
	.dc.b	$FD,$67,$9D,$FB,$C3,$F5,$37,$E1,$0E,$A7,$DF,$3C,$07,$9D,$CB,$7F
	.dc.b	$F5,$A6,$5C,$B1,$E3,$87,$0B,$7E,$43,$BB,$FC,$1E,$7D,$DF,$F6,$26
	.dc.b	$9E,$38,$FF,$2F,$C6,$0E,$AF,$97,$4F,$26,$9F,$ED,$3A,$E1,$97,$82
	.dc.b	$3F,$C3,$D1,$FD,$D7,$F6,$DF,$ED,$1F,$A6,$FD,$9F,$C5,$BF,$DF,$FF
	.dc.b	$7A,$FD,$2F,$C9,$5A,$EA,$FF,$5E,$FE,$39,$FE,$4F,$FD,$0B,$3F,$35
	.dc.b	$FE,$49,$E0,$FF,$54,$E9,$D6,$1A,$7F,$B0,$E3,$0F,$7B,$FE,$C3,$5C
	.dc.b	$A3,$1F,$92,$FF,$B1,$2B,$78,$F8,$34,$FA,$DE,$2A,$73,$B7,$A3,$EA
	.dc.b	$7F,$CB,$7F,$8D,$FF,$84,$7F,$04,$7E,$6D,$F9,$5F,$D2,$1F,$7B,$FE
	.dc.b	$14,$FD,$F5,$FA,$FF,$0E,$1F,$F7,$56,$1C,$37,$61,$C3,$EC,$18,$3F
	.dc.b	$70,$E1,$87,$F9,$FB,$3F,$D1,$3F,$EF,$BD,$3C,$51,$FE,$DD,$8F,$F6
	.dc.b	$6C,$63,$1E,$BA,$7F,$89,$BF,$21,$FF,$71,$CE,$73,$F0,$7C,$16,$30
	.dc.b	$F2,$E3,$EB,$C3,$D5,$F1,$1E,$2D,$BD,$18,$79,$F5,$FF,$5C,$F1,$FF
	.dc.b	$BE,$31,$E3,$AF,$83,$7F,$5F,$B5,$C7,$4E,$91,$EE,$F0,$BD,$BF,$6B
	.dc.b	$FC,$17,$F4,$FF,$F3,$97,$E2,$5F,$BA,$F7,$1F,$B2,$EB,$CE,$56,$BD
	.dc.b	$55,$3E,$A3,$5F,$3D,$FE,$45,$E0,$FF,$23,$70,$F9,$56,$9D,$DF,$94
	.dc.b	$0C,$E5,$DB,$F0,$4F,$F6,$3E,$5F,$E7,$9E,$9F,$F2,$85,$F9,$E7,$3F
	.dc.b	$01,$F2,$70,$E3,$AF,$97,$5F,$E3,$19,$F8,$70,$E1,$BF,$B7,$E2,$BF
	.dc.b	$D4,$BF,$A9,$FF,$81,$BF,$18,$FF,$2D,$FE,$4B,$FB,$EF,$CA,$BF,$9C
	.dc.b	$F3,$6B,$5F,$F0,$05,$6B,$D5,$57,$F3,$EF,$DD,$35,$4F,$9F,$5A,$7E
	.dc.b	$24,$FB,$F6,$67,$E1,$CB,$33,$CB,$0E,$30,$CF,$C9,$F8,$97,$1B,$E1
	.dc.b	$1E,$7F,$B8,$C6,$1E,$6B,$E1,$AF,$9B,$D9,$7E,$81,$D7,$6F,$56,$FF
	.dc.b	$D5,$BF,$01,$F4,$FE,$A9,$DF,$F5,$4F,$F3,$8F,$C2,$FE,$FB,$EF,$7F
	.dc.b	$C2,$7C,$7E,$8D,$BE,$C5,$EF,$EF,$3F,$D5,$7E,$AF,$F9,$D7,$E3,$39
	.dc.b	$F1,$B3,$F6,$E3,$C7,$16,$3E,$DF,$D5,$73,$CB,$FE,$8D,$8E,$9A,$FF
	.dc.b	$0E,$E5,$F6,$DF,$0C,$BA,$6B,$E0,$E7,$CE,$1C,$36,$FF,$43,$7E,$55
	.dc.b	$E2,$DB,$D7,$1F,$E8,$9F,$C0,$31,$87,$A3,$85,$F9,$06,$36,$F8,$1F
	.dc.b	$03,$F7,$3F,$22,$7C,$73,$F3,$37,$E6,$2F,$0B,$FC,$3F,$E3,$BF,$65
	.dc.b	$E1,$7B,$AD,$7C,$E7,$FB,$7F,$F6,$8C,$E3,$F0,$39,$FF,$45,$9F,$6F
	.dc.b	$CD,$B1,$08,$F0,$CF,$83,$4F,$F1,$6F,$8B,$F1,$27,$39,$76,$EA,$C6
	.dc.b	$9E,$1B,$F2,$FD,$27,$82,$BC,$7B,$F5,$EE,$FA,$9F,$0B,$DF,$F8,$01
	.dc.b	$F4,$9E,$EF,$DE,$9E,$EB,$F7,$9F,$CF,$DE,$EB,$52,$F8,$2C,$F0,$3F
	.dc.b	$03,$F5,$DF,$5E,$4F,$D1,$E9,$E0,$A6,$9E,$0D,$9A,$B7,$FC,$0E,$3D
	.dc.b	$E7,$3E,$0E,$30,$E3,$B7,$8B,$F2,$BF,$0E,$5C,$F2,$F4,$FD,$2F,$ED
	.dc.b	$BE,$51,$FB,$43,$F0,$1F,$F5,$0F,$90,$FC,$79,$FA,$A3,$F1,$1E,$17
	.dc.b	$BF,$C0,$BF,$7D,$5A,$FA,$FC,$FD,$47,$D8,$CF,$AE,$77,$0E,$19,$F5
	.dc.b	$FC,$93,$F3,$75,$AC,$B9,$F1,$E2,$F7,$9F,$8B,$F7,$9F,$0E,$5E,$68
	.dc.b	$F2,$FE,$CF,$8C,$3C,$7B,$BF,$26,$FE,$27,$ED,$7F,$11,$7F,$CD,$98
	.dc.b	$3F,$2E,$7E,$17,$F5,$FB,$7F,$69,$7B,$FB,$3F,$89,$5A,$F0,$F9,$FC
	.dc.b	$0F,$5F,$07,$EC,$FB,$3F,$CE,$5F,$E1,$B0,$FC,$97,$F3,$38,$F3,$D3
	.dc.b	$C1,$E4,$FE,$EA,$31,$97,$FA,$37,$4F,$0E,$1E,$9D,$3C,$3A,$BB,$3F
	.dc.b	$96,$FC,$2B,$E7,$5F,$99,$7F,$B5,$7F,$7E,$FE,$57,$FB,$17,$FA,$BB
	.dc.b	$3F,$E1,$D7,$BA,$D7,$12,$9F,$8F,$F4,$C5,$FB,$27,$EC,$30,$E1,$C9
	.dc.b	$EE,$3F,$CF,$13,$E0,$F9,$16,$9F,$CE,$BC,$CF,$E2,$B7,$DE,$73,$9F
	.dc.b	$B4,$FF,$46,$5F,$1B,$F3,$FE,$93,$E1,$D7,$CB,$AB,$F2,$96,$EF,$CA
	.dc.b	$7E,$77,$BE,$FD,$EF,$FB,$5F,$D2,$7C,$FF,$DC,$DF,$8B,$BF,$40,$E3
	.dc.b	$87,$0B,$DF,$7A,$F7,$F6,$3A,$E6,$FE,$E5,$9F,$5F,$95,$FE,$6F,$8F
	.dc.b	$1F,$C9,$6D,$6B,$2E,$78,$F8,$23,$90,$C3,$96,$BE,$5E,$7F,$EC,$18
	.dc.b	$F8,$F2,$4F,$A7,$67,$E0,$1C,$B0,$E3,$B7,$9E,$5F,$AF,$76,$F9,$5F
	.dc.b	$E1,$9E,$C3,$F6,$57,$E1,$2F,$E0,$07,$E3,$7F,$E7,$CE,$7F,$95,$C5
	.dc.b	$FE,$78,$F8,$0F,$E3,$B8,$70,$BD,$FE,$99,$D6,$7F,$90,$D7,$F1,$6E
	.dc.b	$3F,$F4,$44,$7F,$A0,$E3,$07,$3C,$FC,$59,$F8,$76,$F7,$CE,$72,$ED
	.dc.b	$C9,$88,$FF,$62,$E1,$E2,$BF,$46,$AF,$D5,$3E,$5F,$CA,$FF,$5B,$E3
	.dc.b	$FE,$E3,$AF,$F6,$5F,$96,$FE,$C3,$F2,$3F,$DD,$7E,$53,$71,$FD,$BF
	.dc.b	$4F,$29,$FD,$9F,$E5,$FE,$CB,$FD,$A7,$E5,$FF,$B1,$FC,$6F,$8F,$3F
	.dc.b	$ED,$1F,$01,$70,$EB,$F2,$FF,$DA,$BA,$3E,$4E,$3C,$7E,$5F,$FB,$47
	.dc.b	$CB,$7E,$03,$F1,$F7,$E9,$BF,$B7,$FC,$91,$F9,$37,$F9,$B3,$C6,$7E
	.dc.b	$95,$BA,$D4,$E7,$F3,$FC,$7C,$7F,$9B,$1E,$7E,$C5,$BD,$FD,$83,$93
	.dc.b	$E5,$56,$B4,$EF,$2F,$98,$CD,$F0,$D7,$FC,$B3,$C3,$EA,$F9,$E7,$D9
	.dc.b	$3F,$80,$3F,$13,$F8,$2B,$F2,$77,$B0,$FD,$C7,$F3,$B8,$70,$BD,$D6
	.dc.b	$BF,$1B,$CE,$3C,$FA,$7E,$8D,$F6,$F8,$3F,$A2,$F1,$F1,$DA,$77,$F5
	.dc.b	$FF,$09,$39,$8F,$14,$E3,$0F,$33,$C3,$2F,$2F,$DB,$FE,$C3,$E6,$9F
	.dc.b	$1F,$FE,$74,$F9,$97,$ED,$7F,$A3,$FC,$89,$E5,$6E,$B5,$39,$FB,$FC
	.dc.b	$7E,$63,$F8,$6C,$79,$FA,$56,$F7,$EB,$BF,$26,$FC,$D5,$6B,$4E,$F2
	.dc.b	$FC,$83,$37,$C3,$57,$DA,$F8,$72,$FE,$FB,$AF,$97,$E2,$7F,$84,$FE
	.dc.b	$F0,$FD,$9B,$F9,$EF,$E2,$4F,$CE,$7F,$D2,$7F,$AA,$BC,$7D,$D6,$A7
	.dc.b	$3F,$BB,$E3,$E4,$BF,$3F,$8F,$3F,$8E,$B7,$BF,$BF,$FC,$57,$85,$CA
	.dc.b	$5A,$77,$9F,$F3,$D4,$E6,$31,$A3,$F8,$03,$F8,$FF,$C8,$BF,$2A,$FC
	.dc.b	$5F,$E1,$CF,$D5,$5F,$68,$FA,$F7,$EE,$DC,$2F,$75,$A9,$7E,$B1,$9F
	.dc.b	$D2,$BF,$6F,$97,$5F,$22,$E1,$C3,$2F,$F4,$26,$FC,$E5,$A7,$86,$7F
	.dc.b	$CA,$F3,$98,$C6,$4F,$AF,$FE,$6F,$F8,$8B,$EB,$9F,$73,$FE,$8F,$F9
	.dc.b	$2F,$F5,$BF,$ED,$DF,$88,$F5,$FF,$80,$DE,$EB,$5F,$32,$AF,$F7,$C5
	.dc.b	$6B,$0F,$F1,$C6,$9F,$C4,$B0,$E0,$D8,$EF,$F0,$EC,$BF,$D8,$BE,$8F
	.dc.b	$F2,$5B,$ED,$39,$CB,$F2,$18,$7A,$30,$E3,$87,$9B,$C5,$C6,$9E,$4D
	.dc.b	$3F,$D5,$BA,$7F,$18,$C9,$F8,$E3,$F1,$17,$EC,$9F,$CB,$FF,$27,$7E
	.dc.b	$CE,$FC,$97,$FA,$DE,$1C,$2F,$7F,$2A,$BD,$FE,$BF,$AC,$BA,$70,$FF
	.dc.b	$30,$C3,$BF,$E6,$3F,$CC,$1E,$5A,$77,$FA,$7F,$E8,$51,$8C,$BA,$7E
	.dc.b	$47,$8C,$7C,$D7,$E5,$D9,$FA,$2F,$F6,$EF,$D3,$9F,$A4,$3F,$1F,$6E
	.dc.b	$FC,$3D,$F4,$8F,$9D,$FF,$36,$F3,$7F,$3A,$C7,$8E,$1C,$2F,$74,$9E
	.dc.b	$D8,$F5,$E3,$FE,$97,$1E,$0F,$DA,$BF,$A5,$BC,$B3,$F0,$66,$FB,$FF
	.dc.b	$EB,$D3,$9C,$BC,$7F,$80,$63,$2F,$46,$1C,$33,$FF,$43,$6B,$FD,$3B
	.dc.b	$E5,$FF,$99,$7A,$5F,$DF,$7B,$BF,$BD,$FA,$7F,$E7,$7A,$D7,$C3,$FF
	.dc.b	$90,$BE,$97,$F8,$6F,$D2,$7F,$09,$F1,$3F,$C2,$79,$72,$E4,$FE,$C6
	.dc.b	$FF,$3D,$F6,$3F,$B1,$BD,$AF,$8F,$5F,$F7,$58,$C6,$9F,$E0,$AE,$EC
	.dc.b	$67,$DE,$3C,$7F,$80,$F6,$8C,$69,$FC,$67,$ED,$FF,$C4,$7F,$74,$F0
	.dc.b	$BF,$23,$F0,$8F,$D2,$5F,$B7,$BF,$1E,$FF,$0C,$F2,$7F,$6E,$C3,$87
	.dc.b	$89,$FC,$AE,$F7,$5F,$C3,$EB,$58,$FE,$73,$93,$FB,$FE,$1D,$FF,$17
	.dc.b	$FE,$FE,$F2,$CF,$C1,$9F,$7A,$F2,$DF,$93,$2F,$1F,$E1,$3F,$E1,$5C
	.dc.b	$38,$47,$83,$F3,$2F,$F1,$EF,$A1,$F8,$DF,$DE,$5F,$40,$FD,$C9,$FB
	.dc.b	$DF,$FA,$EB,$F6,$3D,$F8,$AB,$DD,$6B,$F0,$D7,$E4,$BA,$FD,$EF,$FC
	.dc.b	$2B,$9F,$5E,$1F,$F0,$CB,$DF,$4E,$DA,$3F,$34,$FF,$15,$4E,$72,$F0
	.dc.b	$FA,$78,$D7,$C1,$87,$0D,$1F,$96,$FF,$95,$BE,$37,$FD,$99,$F0,$EF
	.dc.b	$AF,$7E,$7F,$FE,$84,$F8,$AF,$ED,$7D,$BF,$E2,$0B,$DD,$6B,$F0,$6F
	.dc.b	$E7,$7F,$88,$73,$FC,$1F,$FA,$07,$27,$F8,$C2,$FF,$DB,$1F,$98,$7F
	.dc.b	$8C,$1E,$39,$FF,$A4,$3E,$3B,$FC,$85,$18,$C5,$F4,$B8,$C7,$FA,$56
	.dc.b	$7D,$F5,$FF,$5E,$6B,$E3,$9C,$E5,$E7,$FC,$03,$FC,$E3,$F8,$E7,$CC
	.dc.b	$F8,$C7,$EC,$3F,$5B,$DF,$7E,$E6,$FC,$FF,$7E,$07,$F4,$8A,$D4,$BF
	.dc.b	$37,$CE,$1F,$E4,$8C,$7E,$BF,$ED,$F7,$0E,$1D,$7F,$C4,$7D,$2F,$F2
	.dc.b	$E5,$78,$72,$F0,$7C,$EE,$2F,$7A,$F8,$BF,$39,$FF,$43,$FE,$EC,$CF
	.dc.b	$FD,$91,$F9,$C3,$F2,$55,$3F,$50,$7E,$2D,$C2,$F7,$5A,$96,$B9,$E4
	.dc.b	$FF,$2F,$E5,$D7,$EA,$6E,$1C,$32,$E9,$E0,$6F,$4F,$04,$7F,$80,$3A
	.dc.b	$7F,$CB,$F8,$79,$71,$56,$B3,$FF,$23,$57,$F9,$93,$F4,$87,$DE,$3F
	.dc.b	$39,$7E,$1A,$F8,$2F,$F3,$1F,$E5,$4F,$D6,$7F,$D4,$F8,$70,$BD,$D6
	.dc.b	$BE,$37,$38,$BF,$35,$FF,$3C,$DE,$F8,$7F,$CF,$AF,$1C,$FB,$7E,$0A
	.dc.b	$EC,$FF,$47,$4E,$72,$FF,$39,$FA,$78,$CB,$CD,$87,$0D,$BF,$C8,$FB
	.dc.b	$FA,$E7,$37,$E8,$FC,$7F,$FD,$7B,$F0,$8F,$AC,$7C,$4B,$E1,$9F,$B4
	.dc.b	$BF,$7C,$FF,$1F,$F3,$F0,$BD,$D6,$A7,$FC,$87,$FB,$4E,$7F,$31,$FF
	.dc.b	$4A,$E5,$D7,$2B,$87,$0C,$BF,$D3,$9C,$BC,$E7,$C1,$1D,$FF,$32,$FF
	.dc.b	$56,$7C,$37,$F7,$DF,$B4,$7E,$63,$F4,$FF,$17,$FF,$1C,$7A,$1F,$ED
	.dc.b	$6A,$D4,$E7,$EC,$7F,$41,$CD,$FE,$9E,$9F,$F0,$37,$27,$FA,$A1,$E9
	.dc.b	$AF,$FA,$B1,$E3,$DF,$FE,$F0,$F3,$3C,$3F,$49,$FE,$70,$CB,$C1,$F2
	.dc.b	$3F,$EB,$FF,$D7,$7F,$C8,$7F,$26,$FC,$89,$E2,$BF,$4D,$7E,$4C,$FC
	.dc.b	$93,$FD,$D7,$87,$0B,$DD,$6B,$AF,$3F,$64,$FB,$BF,$F5,$ED,$F6,$E2
	.dc.b	$FF,$60,$BC,$73,$ED,$F2,$9E,$3B,$FF,$39,$E5,$D3,$EA,$71,$B7,$96
	.dc.b	$FC,$9B,$3C,$6E,$F3,$9D,$BC,$FF,$3D,$FF,$00,$3B,$5F,$A5,$3F,$4C
	.dc.b	$7E,$FA,$FC,$51,$F8,$EF,$F3,$2F,$D5,$78,$5E,$EB,$53,$FC,$DB,$F5
	.dc.b	$5C,$FE,$A1,$FE,$D8,$CB,$AE,$57,$0E,$19,$74,$E4,$FF,$76,$56,$B2
	.dc.b	$E7,$F4,$58,$BD,$E7,$FD,$0B,$F5,$B7,$87,$FE,$DC,$FC,$31,$F4,$FF
	.dc.b	$DD,$7F,$14,$FA,$7F,$0C,$FC,$09,$FB,$37,$0C,$9D,$F3,$1E,$3F,$49
	.dc.b	$FE,$C7,$EB,$FE,$BB,$83,$F0,$6E,$17,$BD,$3F,$C0,$4F,$FB,$DB,$0F
	.dc.b	$0F,$E5,$DF,$EF,$67,$9E,$BE,$2F,$C2,$7B,$EF,$E7,$C3,$CD,$9F,$F6
	.dc.b	$AE,$4C,$ED,$E4,$CD,$97,$2A,$F1,$FD,$07,$F1,$39,$D7,$E8,$CF,$80
	.dc.b	$7E,$E4,$FB,$C7,$EB,$6F,$D8,$5F,$95,$3F,$7C,$E5,$97,$C2,$CB,$96
	.dc.b	$3C,$70,$E1,$6F,$D7,$37,$6B,$DB,$E7,$3E,$5F,$C2,$58,$F2,$7E,$DB
	.dc.b	$99,$C9,$E8,$D7,$CF,$AF,$9B,$5F,$ED,$33,$DA,$FC,$73,$9F,$A3,$E3
	.dc.b	$3F,$63,$F6,$67,$EC,$8F,$CE,$7F,$C2,$DF,$97,$3F,$52,$7E,$E7,$FE
	.dc.b	$BA,$DF,$CD,$87,$0B,$DD,$6B,$F1,$8F,$6C,$E7,$2F,$45,$F0,$C9,$F1
	.dc.b	$3E,$B9,$DC,$78,$7E,$D5,$A7,$87,$E3,$BC,$F4,$C7,$C1,$5A,$CF,$FD
	.dc.b	$07,$3C,$B4,$F1,$EA,$FD,$67,$E7,$FE,$7B,$FC,$3B,$D6,$FC,$97,$3B
	.dc.b	$F8,$07,$FC,$FD,$F8,$26,$BD,$AF,$F7,$AF,$93,$F5,$E7,$FD,$05,$F8
	.dc.b	$6E,$7E,$03,$D1,$CD,$F5,$F3,$E9,$BF,$B8,$FC,$83,$DF,$0E,$5F,$2D
	.dc.b	$EA,$8C,$69,$E1,$F0,$F1,$A7,$82,$3C,$9F,$84,$7F,$A2,$72,$FC,$27
	.dc.b	$E6,$8F,$DC,$9F,$AC,$7F,$A7,$3F,$03,$F8,$FF,$37,$F8,$DB,$DD,$6B
	.dc.b	$F0,$D5,$6B,$F6,$6F,$C8,$E3,$CF,$F2,$5F,$5F,$17,$E9,$E3,$B6,$78
	.dc.b	$8F,$DE,$C6,$30,$E7,$DE,$F4,$3E,$AC,$E7,$2F,$B3,$E9,$FD,$8F,$CB
	.dc.b	$5F,$9F,$7F,$93,$7E,$D3,$FD,$CD,$F3,$EF,$DE,$9F,$A5,$FE,$92,$F7
	.dc.b	$F7,$3F,$E5,$2F,$C6,$3F,$B5,$53,$F6,$7D,$65,$FE,$57,$E1,$FA,$07
	.dc.b	$F1,$2C,$3B,$FE,$C7,$E9,$84,$74,$D3,$C3,$FA,$5F,$96,$FC,$99,$7F
	.dc.b	$77,$BC,$67,$E1,$D7,$CD,$F8,$E7,$EE,$3F,$7A,$FF,$51,$FA,$0F,$C5
	.dc.b	$5F,$88,$67,$E7,$3F,$D4,$31,$FE,$11,$EE,$FD,$E5,$FF,$AC,$73,$FD
	.dc.b	$F4,$62,$9F,$8A,$F9,$EB,$AC,$FE,$81,$EF,$F8,$AD,$AE,$BF,$5B,$FC
	.dc.b	$2B,$F4,$C7,$CB,$7F,$B5,$FF,$4F,$FE,$19,$FB,$FF,$D5,$F9,$FD,$4E
	.dc.b	$7D,$F4,$FD,$9F,$77,$3F,$87,$7C,$7C,$79,$F9,$77,$0E,$17,$F7,$F8
	.dc.b	$79,$CB,$3F,$E3,$5E,$17,$F2,$FE,$1F,$E3,$5E,$33,$F6,$A7,$EC,$6F
	.dc.b	$DD,$9F,$8C,$7F,$2A,$F8,$27,$EC,$DF,$C9,$30,$E1,$7B,$AD,$7E,$01
	.dc.b	$9C,$79,$ED,$F8,$03,$FD,$FA,$BF,$4F,$FC,$06,$1C,$B3,$EF,$F1,$7F
	.dc.b	$EA,$7B,$E5,$97,$58,$7A,$FF,$C3,$AF,$86,$7E,$5F,$9D,$FE,$6F,$FC
	.dc.b	$01,$FE,$6C,$FD,$E1,$F9,$E7,$F3,$9F,$CD,$3F,$35,$FB,$86,$BF,$08
	.dc.b	$56,$BB,$95,$F9,$FF,$D3,$35,$3F,$95,$C7,$FC,$EF,$FB,$A5,$C3,$8E
	.dc.b	$5D,$7C,$FF,$E4,$35,$E1,$8F,$07,$16,$2B,$79,$F6,$F5,$1F,$AB,$FF
	.dc.b	$4E,$F0,$1F,$80,$FF,$8B,$7F,$44,$FE,$B3,$F9,$63,$83,$E4,$77,$FF
	.dc.b	$51,$BD,$FD,$82,$F7,$5F,$E5,$FA,$D6,$0F,$C6,$7F,$5D,$87,$0D,$BE
	.dc.b	$A3,$CB,$0E,$FE,$17,$F3,$A8,$C7,$5B,$47,$E7,$9F,$CF,$BE,$A9,$F9
	.dc.b	$4B,$F5,$57,$D5,$7F,$37,$FC,$51,$D2,$FD,$6B,$FB,$5E,$1C,$2F,$7F
	.dc.b	$26,$BD,$FE,$1F,$AC,$9F,$90,$7F,$05,$C3,$87,$17,$E2,$DE,$5A,$77
	.dc.b	$D3,$C9,$7E,$38,$F1,$70,$E3,$4E,$F9,$F9,$3F,$4C,$FE,$FB,$FB,$73
	.dc.b	$F0,$AF,$F5,$C4,$FF,$92,$7E,$31,$ED,$FF,$95,$FA,$BF,$C0,$17,$BF
	.dc.b	$A1,$F4,$7D,$E7,$F6,$2A,$7C,$E7,$4C,$BF,$D6,$FC,$5F,$C2,$8F,$F0
	.dc.b	$F6,$1D,$FB,$7F,$90,$3C,$F4,$F0,$7E,$93,$FD,$AA,$31,$97,$F6,$EE
	.dc.b	$DC,$63,$E6,$8F,$07,$D7,$7F,$85,$BF,$6C,$FD,$BF,$E1,$7F,$A9,$F9
	.dc.b	$87,$E2,$AF,$D5,$35,$E9,$7F,$BA,$FE,$7B,$FC,$87,$5A,$FC,$AB,$3F
	.dc.b	$90,$7F,$33,$CB,$9F,$1F,$EE,$5E,$B7,$C3,$F5,$9F,$EF,$F1,$89,$FF
	.dc.b	$64,$FD,$66,$32,$ED,$1F,$EC,$CD,$1F,$A8,$7F,$8B,$3F,$44,$7E,$A0
	.dc.b	$FD,$0D,$F8,$83,$F4,$CF,$F7,$4F,$C9,$7F,$13,$F0,$3F,$C6,$18,$70
	.dc.b	$BD,$D6,$BF,$7E,$E9,$8B,$E8,$7F,$A6,$5E,$F5,$FE,$24,$F1,$CF,$B7
	.dc.b	$E4,$9F,$E4,$DC,$FA,$E1,$E3,$9C,$FC,$A7,$F8,$73,$F2,$0F,$F0,$FF
	.dc.b	$E7,$FF,$D5,$BF,$44,$7E,$D7,$F9,$EF,$C4,$BF,$C8,$F5,$AF,$A3,$56
	.dc.b	$BD,$17,$E8,$DA,$F3,$5F,$BA,$7F,$61,$C3,$AF,$17,$F2,$77,$96,$1C
	.dc.b	$63,$86,$9E,$1D,$3F,$C0,$F1,$8C,$FB,$C7,$EC,$3F,$69,$FF,$2C,$F8
	.dc.b	$EF,$86,$7E,$BE,$FD,$FF,$FC,$B9,$F9,$DB,$F4,$FF,$DE,$F3,$FF,$97
	.dc.b	$70,$E1,$7B,$F8,$85,$EF,$3E,$B9,$74,$D7,$F9,$F6,$1D,$FE,$C3,$FB
	.dc.b	$E3,$CB,$4E,$FF,$8C,$FF,$91,$A3,$19,$74,$FC,$53,$18,$F9,$AF,$CB
	.dc.b	$B7,$FB,$D7,$6F,$F3,$87,$E4,$3F,$EC,$9F,$DF,$5F,$9B,$BF,$5F,$7E
	.dc.b	$60,$F7,$BF,$D0,$18,$70,$FB,$F6,$1C,$2F,$74,$FD,$3B,$FC,$FF,$97
	.dc.b	$FB,$57,$F2,$9F,$F0,$A3,$DF,$93,$FA,$D3,$E1,$D3,$C1,$A7,$79,$E3
	.dc.b	$9F,$69,$CF,$B9,$FE,$64,$BE,$19,$F8,$3E,$FB,$FD,$1B,$F4,$AF,$C9
	.dc.b	$9F,$A0,$3E,$BF,$F9,$67,$E1,$1F,$E1,$0D,$37,$5A,$9C,$FD,$8F,$F1
	.dc.b	$0E,$4F,$F4,$A6,$3C,$FF,$1B,$FF,$16,$3D,$72,$B8,$71,$F8,$AF,$F4
	.dc.b	$57,$C5,$FF,$9E,$2B,$59,$78,$3C,$BC,$5E,$F5,$FF,$10,$FC,$1F,$FA
	.dc.b	$93,$F0,$67,$E5,$6F,$D6,$DE,$63,$F7,$0F,$F3,$A7,$D8,$6E,$B5,$39
	.dc.b	$E5,$FF,$3E,$7D,$86,$7E,$13,$FC,$85,$97,$4C,$AE,$1C,$32,$FF,$5A
	.dc.b	$7A,$BF,$E8,$EA,$D6,$5C,$F9,$B1,$7B,$CF,$FD,$11,$F6,$7F,$EB,$AF
	.dc.b	$D2,$1F,$9E,$BF,$67,$7C,$37,$F4,$EF,$DF,$BF,$B3,$7C,$BE,$EB,$53
	.dc.b	$9F,$A2,$7E,$41,$9C,$3F,$CB,$18,$F4,$FC,$22,$E1,$C3,$E7,$59,$BF
	.dc.b	$4F,$FF,$52,$DF,$8B,$0F,$0D,$6A,$FF,$D5,$3F,$06,$FD,$83,$F5,$FF
	.dc.b	$A4,$3F,$4C,$7E,$FE,$FD,$3B,$F3,$BE,$7F,$FB,$A7,$0E,$1E,$F3,$0E
	.dc.b	$17,$BF,$4F,$FD,$B7,$5A,$BF,$F1,$A3,$DF,$EC,$BF,$CE,$4F,$2B,$E3
	.dc.b	$F9,$57,$93,$4F,$1E,$9E,$2E,$EC,$69,$DE,$3C,$BA,$FF,$A7,$73,$FF
	.dc.b	$7E,$7E,$B3,$FD,$03,$F5,$BF,$F0,$2F,$E6,$CF,$D4,$7E,$2F,$E8,$BF
	.dc.b	$EE,$EA,$D7,$6A,$B5,$D2,$FD,$07,$5F,$31,$FE,$96,$CF,$FC,$99,$87
	.dc.b	$56,$C3,$3E,$F9,$B8,$7F,$DA,$B7,$DB,$2F,$17,$06,$30,$E1,$A7,$87
	.dc.b	$0F,$B3,$F8,$1B,$C4,$FF,$80,$1F,$B2,$F5,$FD,$4D,$F9,$3B,$F6,$77
	.dc.b	$E5,$4F,$DC,$BF,$1F,$87,$0E,$16,$1C,$2F,$74,$FE,$01,$ED,$C6,$26
	.dc.b	$7D,$DF,$F3,$53,$ED,$FD,$5B,$FD,$50,$F8,$F4,$F1,$68,$FD,$1F,$FD
	.dc.b	$E7,$39,$CB,$CB,$F7,$58,$D7,$C7,$87,$0D,$1F,$B3,$78,$B4,$F1,$7E
	.dc.b	$3B,$FE,$41,$FC,$E9,$F8,$53,$EE,$DF,$5F,$FD,$EB,$F2,$FE,$18,$7F
	.dc.b	$B0,$75,$61,$C2,$F7,$E3,$3F,$44,$DE,$7F,$E4,$DF,$CF,$3F,$D8,$4F
	.dc.b	$7C,$7F,$D1,$0F,$87,$4F,$06,$8F,$3F,$E3,$74,$FC,$7B,$F3,$EB,$E0
	.dc.b	$BD,$FE,$9D,$A9,$D6,$3C,$F5,$F1,$FE,$E9,$E8,$FC,$91,$F4,$1F,$EB
	.dc.b	$0F,$CD,$9F,$95,$3F,$1A,$FE,$9B,$F5,$BB,$AD,$4E,$7E,$5D,$F6,$B9
	.dc.b	$FC,$6B,$FD,$B9,$97,$4F,$90,$B8,$70,$C9,$F6,$FE,$5E,$0F,$C0,$5F
	.dc.b	$8B,$06,$BE,$2B,$FE,$2F,$5D,$63,$19,$F9,$B9,$3E,$14,$BF,$20,$FE
	.dc.b	$3D,$C6,$F9,$A7,$EC,$7F,$D9,$61,$BF,$D8,$AD,$D6,$BE,$79,$5F,$ED
	.dc.b	$2C,$7A,$D7,$F2,$9F,$D7,$FF,$DF,$0F,$6C,$BF,$D7,$CF,$2C,$BB,$47
	.dc.b	$5F,$81,$E7,$97,$5F,$7B,$E0,$4F,$9E,$73,$5E,$5F,$C5,$BF,$17,$8F
	.dc.b	$FA,$D3,$F2,$97,$47,$F9,$DF,$81,$F8,$FB,$F3,$9D,$FE,$32,$BD,$FD
	.dc.b	$5F,$F8,$12,$BF,$B3,$6B,$F8,$7C,$BF,$D7,$FF,$B0,$7C,$88,$F1,$C7
	.dc.b	$77,$FD,$B0,$F8,$73,$F0,$7E,$23,$AC,$BB,$47,$8F,$9B,$19,$F8,$35
	.dc.b	$F2,$FD,$1F,$E9,$3F,$79,$FE,$03,$FA,$CB,$F4,$57,$EA,$EF,$DE,$5F
	.dc.b	$99,$BF,$69,$E1,$EE,$B0,$E1,$7B,$AF,$DB,$F6,$FE,$03,$2F,$E1,$BF
	.dc.b	$8F,$7A,$F9,$38,$E1,$6B,$58,$F3,$D3,$C3,$A3,$ED,$BE,$9E,$5E,$69
	.dc.b	$CF,$B9,$8B,$E1,$AF,$87,$F1,$AF,$DF,$FE,$D4,$FD,$37,$FA,$87,$C7
	.dc.b	$3E,$BB,$EA,$FE,$DC,$F9,$0D,$D6,$A7,$3F,$1F,$8F,$94,$F3,$72,$7E
	.dc.b	$EF,$2E,$9A,$62,$3F,$92,$47,$79,$FE,$C5,$7F,$82,$FD,$4F,$F1,$73
	.dc.b	$9D,$3C,$5E,$17,$FA,$33,$0E,$D1,$8D,$5F,$15,$FC,$63,$E4,$DF,$88
	.dc.b	$7F,$69,$FD,$17,$FC,$D1,$E4,$FE,$2B,$AB,$F5,$B5,$AD,$D5,$AF,$0F
	.dc.b	$FC,$43,$5A,$C3,$FB,$96,$9F,$8D,$8C,$70,$FA,$E3,$FB,$E4,$F7,$FB
	.dc.b	$B6,$7F,$C4,$27,$C9,$39,$8F,$F7,$F4,$63,$2F,$0D,$70,$C5,$F6,$DF
	.dc.b	$CC,$BF,$2B,$FF,$01,$7E,$6D,$FD,$67,$E3,$3F,$D5,$BF,$B9,$3C,$DD
	.dc.b	$64,$AD,$7E,$16,$A7,$EC,$BA,$9F,$46,$3F,$E0,$DF,$C2,$2E,$1C,$78
	.dc.b	$BF,$B8,$62,$F9,$9F,$DE,$DF,$8A,$73,$F3,$9F,$C0,$3F,$54,$FC,$F7
	.dc.b	$E0,$5F,$DA,$7F,$72,$FE,$F1,$F0,$EF,$D0,$7F,$A5,$D6,$BE,$5D,$5A
	.dc.b	$FC,$23,$3E,$97,$F9,$F3,$A7,$E7,$1F,$D0,$EF,$AB,$83,$FB,$63,$CE
	.dc.b	$BF,$C0,$DF,$25,$E3,$AF,$F5,$26,$1D,$B9,$FF,$85,$61,$FD,$73,$DF
	.dc.b	$FF,$05,$7E,$54,$F1,$9F,$B0,$FF,$9D,$FF,$34,$7D,$0E,$F8,$D7,$BA
	.dc.b	$D7,$D7,$67,$F3,$5F,$C1,$65,$D3,$3F,$31,$FE,$93,$C3,$8E,$7D,$BF
	.dc.b	$00,$FE,$D1,$3D,$72,$E9,$1F,$E8,$78,$C4,$EF,$47,$EF,$FF,$DD,$3F
	.dc.b	$4B,$FF,$50,$FE,$26,$FD,$33,$F0,$1C,$EF,$D6,$DF,$4F,$BF,$C3,$57
	.dc.b	$BF,$13,$FD,$BF,$4F,$9D,$D4,$FE,$6F,$2F,$F5,$6F,$9F,$70,$E3,$9F
	.dc.b	$6F,$07,$F9,$FD,$78,$A3,$C3,$F2,$FF,$D0,$BE,$E9,$F9,$23,$EF,$BF
	.dc.b	$67,$E7,$BF,$40,$53,$F2,$7F,$F0,$7B,$C2,$B7,$3A,$F3,$F3,$83,$F4
	.dc.b	$6F,$D8,$5E,$FA,$7E,$7B,$C7,$D5,$FE,$9B,$87,$FB,$0E,$78,$E3,$D3
	.dc.b	$C0,$F3,$63,$E5,$BD,$E1,$FE,$73,$E0,$FE,$E3,$FA,$27,$E1,$47,$E2
	.dc.b	$DF,$E1,$7F,$C9,$1F,$A5,$BF,$56,$FF,$13,$D6,$EB,$52,$FC,$AB,$3E
	.dc.b	$0B,$F2,$0F,$E1,$17,$D7,$83,$F0,$EF,$1F,$AB,$67,$FE,$41,$9E,$33
	.dc.b	$E2,$9C,$FC,$77,$F7,$08,$E1,$5E,$4F,$8D,$FF,$1C,$E0,$FD,$C3,$F0
	.dc.b	$7E,$67,$F9,$4F,$F7,$F7,$B0,$FC,$8B,$75,$AD,$55,$AF,$A9,$CF,$91
	.dc.b	$FC,$7F,$2E,$9E,$35,$C3,$85,$FC,$FE,$6F,$F0,$CD,$78,$6F,$F5,$B8
	.dc.b	$F4,$FA,$4F,$F0,$3E,$6F,$DD,$DE,$6F,$F9,$7F,$CD,$FF,$1B,$7C,$CB
	.dc.b	$F6,$AF,$F2,$9D,$EE,$B5,$F5,$2A,$FF,$3C,$6B,$FE,$28,$AD,$67,$FB
	.dc.b	$27,$B7,$E7,$DF,$CD,$DE,$5F,$2F,$FA,$CF,$E0,$1F,$F2,$85,$F1,$D7
	.dc.b	$C7,$3F,$B0,$C9,$F8,$57,$F8,$7B,$95,$F0,$5F,$E6,$EF,$51,$F3,$4F
	.dc.b	$DA,$BE,$F7,$D8,$7F,$9A,$AF,$7F,$75,$FC,$D6,$B5,$D1,$FE,$22,$DB
	.dc.b	$FC,$7F,$5A,$E2,$FE,$12,$F6,$F8,$2F,$E9,$CF,$2C,$FB,$E7,$FE,$DA
	.dc.b	$AE,$DA,$78,$F2,$F1,$7B,$DF,$E3,$4F,$80,$FF,$73,$7E,$A6,$FD,$A9
	.dc.b	$E2,$BF,$5C,$7E,$5B,$FC,$DF,$83,$FC,$A3,$87,$0B,$DD,$6B,$E0,$3F
	.dc.b	$D0,$B3,$D3,$5F,$E3,$EF,$FB,$CF,$E2,$3F,$B2,$3C,$72,$ED,$E2,$F8
	.dc.b	$FE,$BD,$AF,$8E,$7C,$5D,$5F,$E4,$AF,$0B,$FC,$57,$FB,$BF,$F3,$08
	.dc.b	$7F,$00,$FF,$BF,$BF,$5C,$FF,$4A,$D6,$BE,$1F,$FD,$4D,$2F,$C5,$3F
	.dc.b	$D9,$1C,$7F,$E6,$AA,$E7,$C1,$FC,$D1,$FF,$1B,$F4,$7F,$79,$79,$65
	.dc.b	$DA,$FC,$5F,$34,$D3,$C5,$97,$F9,$FB,$1F,$F5,$2F,$E5,$AF,$C3,$BF
	.dc.b	$D5,$5F,$B0,$FF,$BE,$FE,$75,$F9,$C3,$CA,$DD,$6B,$E8,$BF,$C0,$B3
	.dc.b	$FA,$0E,$1F,$F3,$FE,$3C,$FF,$41,$FF,$06,$3D,$7F,$00,$B8,$71,$CB
	.dc.b	$FD,$13,$7C,$23,$1F,$8A,$FF,$9C,$AF,$86,$9E,$3C,$9F,$84,$7F,$AF
	.dc.b	$3F,$31,$7E,$03,$FC,$0B,$EF,$9E,$2B,$F7,$D7,$E8,$EE,$1E,$17,$BA
	.dc.b	$D4,$FF,$B2,$77,$FF,$4B,$E3,$D3,$E7,$BF,$C4,$8F,$6F,$76,$E1,$C7
	.dc.b	$2F,$F2,$B7,$37,$49,$FF,$4E,$65,$E1,$FC,$C3,$FD,$17,$87,$1D,$7C
	.dc.b	$B8,$79,$3F,$03,$FF,$6A,$FB,$6F,$D1,$5F,$B5,$3F,$31,$7E,$79,$E9
	.dc.b	$F4,$5F,$B9,$F0,$BD,$D6,$B9,$7F,$73,$F9,$3E,$71,$FF,$1E,$65,$D7
	.dc.b	$EF,$6E,$1C,$33,$F8,$9F,$07,$FE,$A0,$AF,$15,$F8,$67,$E4,$71,$66
	.dc.b	$F8,$65,$FC,$1B,$F2,$4F,$F7,$6F,$D3,$FF,$91,$BE,$0B,$FA,$5F,$E5
	.dc.b	$1F,$E5,$8F,$86,$FF,$38,$76,$7F,$B0,$AB,$5F,$22,$AD,$7E,$61,$CB
	.dc.b	$C6,$63,$FE,$53,$BF,$F2,$8F,$E2,$36,$73,$87,$1F,$97,$FF,$08,$E5
	.dc.b	$D6,$73,$97,$F9,$A3,$E9,$FF,$D7,$78,$F8,$23,$C9,$FB,$62,$FF,$DA
	.dc.b	$BF,$AC,$BF,$33,$FF,$69,$F8,$4F,$B9,$FE,$47,$FA,$13,$F3,$BE,$1E
	.dc.b	$37,$FB,$F7,$0E,$1F,$88,$FF,$52,$BD,$ED,$AA,$0A,$ED,$E5,$7F,$9B
	.dc.b	$DE,$31,$FE,$2E,$79,$E5,$E1,$F0,$DF,$A2,$FF,$B5,$27,$38,$FF,$36
	.dc.b	$F1,$7F,$DA,$FF,$B0,$BF,$35,$7F,$00,$BF,$56,$7E,$AB,$FE,$F8,$FC
	.dc.b	$A1,$EB,$79,$77,$63,$C7,$0E,$17,$BF,$93,$D4,$E6,$BF,$C8,$6F,$7F
	.dc.b	$67,$FD,$10,$F1,$CB,$BF,$E3,$5F,$E5,$6E,$9F,$F8,$03,$39,$C5,$3F
	.dc.b	$63,$0F,$8A,$3F,$51,$FE,$75,$F8,$13,$F7,$37,$E7,$2F,$D1,$BF,$CB
	.dc.b	$DF,$8D,$FA,$6F,$75,$A9,$FF,$31,$7A,$39,$F9,$3F,$F4,$3E,$5F,$E5
	.dc.b	$AB,$FE,$8D,$FB,$1F,$FA,$7D,$E3,$97,$5B,$FF,$80,$31,$8C,$3F,$C0
	.dc.b	$57,$E6,$8C,$7D,$57,$F6,$6E,$0F,$4C,$BF,$62,$FE,$2B,$F9,$5F,$F8
	.dc.b	$B1,$FE,$8B,$F8,$AB,$EF,$3E,$84,$6B,$CD,$E1,$F2,$8A,$FF,$25,$E3
	.dc.b	$D2,$B9,$E3,$FE,$72,$7F,$90,$FE,$55,$FE,$B7,$79,$1F,$D4,$BB,$3F
	.dc.b	$1F,$5C,$7E,$15,$A7,$8F,$2F,$E5,$F8,$FD,$0F,$DE,$7F,$C3,$1F,$B2
	.dc.b	$BF,$19,$F3,$79,$9F,$CA,$BD,$7F,$C3,$56,$A7,$3E,$DF,$F4,$3D,$3E
	.dc.b	$0C,$F2,$C7,$FD,$0E,$FF,$88,$FC,$EF,$F6,$63,$C7,$8F,$F5,$57,$E6
	.dc.b	$AF,$17,$D4,$34,$F1,$57,$EB,$FE,$E3,$55,$FA,$AC,$39,$6B,$FC,$FB
	.dc.b	$83,$F3,$5F,$5D,$FC,$2F,$AD,$FB,$1F,$FA,$1B,$F6,$37,$EC,$6F,$C2
	.dc.b	$77,$EC,$FE,$32,$B5,$2F,$C5,$F3,$87,$FB,$8F,$1F,$E2,$DF,$AF,$5C
	.dc.b	$38,$7E,$12,$CF,$F6,$BB,$F0,$CF,$C1,$F4,$9E,$88,$FF,$2C,$5F,$0A
	.dc.b	$F1,$FE,$1D,$FB,$18,$FE,$AF,$F8,$57,$F5,$5F,$D9,$BF,$AB,$FE,$6E
	.dc.b	$FC,$6B,$5D,$5F,$EF,$CF,$8A,$56,$BE,$1B,$38,$7F,$BE,$71,$FE,$C3
	.dc.b	$FB,$0D,$C3,$85,$FF,$62,$FC,$43,$C9,$5E,$0C,$79,$FC,$17,$FB,$97
	.dc.b	$7E,$71,$8D,$5F,$A9,$7F,$07,$FD,$11,$F9,$7B,$F5,$4F,$E3,$9F,$84
	.dc.b	$BF,$58,$57,$FA,$77,$C7,$FD,$2B,$C2,$B7,$3A,$FC,$33,$A6,$1F,$E6
	.dc.b	$1D,$3F,$D9,$F7,$B6,$C7,$7F,$86,$FF,$57,$DB,$E6,$7E,$75,$F5,$9C
	.dc.b	$FA,$B8,$BE,$1A,$79,$69,$F8,$47,$F2,$1F,$D0,$FF,$C0,$9F,$A2,$7F
	.dc.b	$14,$F9,$27,$4B,$E8,$5F,$A8,$6F,$99,$7B,$AD,$7E,$AD,$98,$FE,$07
	.dc.b	$C5,$F8,$1C,$BF,$C2,$1F,$99,$62,$35,$F0,$AD,$3B,$FA,$1B,$CF,$AC
	.dc.b	$E7,$3F,$E4,$97,$C2,$FC,$BA,$BF,$73,$FE,$75,$FA,$4B,$F2,$FF,$F5
	.dc.b	$EF,$CD,$BF,$32,$FF,$73,$FE,$73,$FC,$33,$2F,$E5,$F7,$BF,$C5,$17
	.dc.b	$BF,$39,$7F,$E6,$CF,$03,$F8,$ED,$F5,$FD,$93,$D5,$18,$CF,$FD,$97
	.dc.b	$97,$C0,$8C,$65,$D3,$3F,$F7,$BF,$E1,$7F,$A3,$C7,$C1,$A7,$7F,$CD
	.dc.b	$3F,$4D,$E0,$7F,$3F,$FB,$1F,$E8,$1F,$83,$7F,$C0,$1F,$CA,$5F,$4F
	.dc.b	$FD,$5E,$B5,$F2,$3F,$E7,$EF,$C8,$94,$FD,$A3,$53,$FE,$58,$E1,$F2
	.dc.b	$1F,$D3,$AF,$FA,$67,$D5,$FA,$A3,$13,$FE,$5E,$3A,$69,$E4,$71,$9B
	.dc.b	$2F,$F5,$4D,$76,$D3,$CF,$1E,$2F,$C6,$BF,$B6,$7D,$8B,$F1,$C7,$EC
	.dc.b	$8F,$C3,$BF,$1B,$F3,$CF,$CD,$5F,$9D,$F0,$BD,$D6,$A5,$F9,$E6,$70
	.dc.b	$F4,$6B,$FE,$0C,$CB,$B7,$EC,$B8,$8F,$F2,$2E,$3C,$BC,$4F,$F0,$17
	.dc.b	$D3,$FF,$13,$BF,$26,$1E,$39,$FF,$03,$5E,$6F,$86,$BE,$5F,$CF,$7F
	.dc.b	$C0,$DD,$2F,$B8,$FE,$AD,$E0,$BE,$1F,$F1,$D8,$7F,$59,$F0,$BF,$B1
	.dc.b	$DE,$FE,$6F,$F2,$95,$AE,$7E,$78,$FF,$5F,$D3,$EA,$9F,$E2,$6B,$ED
	.dc.b	$3E,$B7,$96,$5D,$FC,$2F,$F0,$CD,$72,$C7,$AC,$3F,$4B,$FE,$61,$7C
	.dc.b	$32,$F2,$F6,$FA,$BE,$C1,$FA,$E3,$E1,$7F,$8D,$57,$F8,$37,$E9,$F5
	.dc.b	$39,$F7,$FD,$DF,$44,$FC,$DB,$3B,$FA,$0F,$CA,$E5,$FE,$49,$71,$A2
	.dc.b	$3B,$E7,$DB,$8F,$CB,$18,$C7,$E5,$7E,$61,$8C,$FC,$1A,$FF,$87,$FF
	.dc.b	$65,$FF,$1B,$79,$0F,$C3,$9F,$98,$3F,$41,$FD,$CF,$E3,$5F,$D7,$BD
	.dc.b	$DD,$D6,$A7,$3F,$31,$FC,$E3,$7F,$F0,$BE,$3C,$F1,$FC,$43,$D7,$1B
	.dc.b	$87,$1F,$41,$1F,$D7,$AF,$8C,$E6,$1F,$29,$FD,$7E,$FA,$65,$E0,$F7
	.dc.b	$BF,$C9,$DE,$63,$F5,$FF,$E8,$BF,$A9,$BF,$58,$FF,$5C,$FA,$FF,$DD
	.dc.b	$75,$FF,$15,$DE,$FE,$19,$7B,$FC,$25,$7F,$DD,$EF,$78,$BE,$13,$F4
	.dc.b	$58,$F1,$ED,$FA,$CF,$3C,$79,$61,$FD,$D6,$31,$9F,$F2,$7A,$FE,$F1
	.dc.b	$F9,$5B,$F5,$47,$E1,$8F,$9A,$7E,$1A,$FE,$01,$7E,$FA,$FC,$77,$FC
	.dc.b	$D9,$8F,$1C,$38,$5E,$EB,$D3,$C3,$FE,$48,$AD,$67,$F8,$B7,$C1,$F3
	.dc.b	$9F,$BB,$79,$67,$E0,$CF,$FB,$B5,$F1,$D3,$CB,$39,$F9,$BF,$4C,$63
	.dc.b	$3F,$07,$EB,$7F,$E7,$FF,$CB,$5F,$B3,$BF,$66,$7E,$CE,$FC,$C7,$FC
	.dc.b	$A5,$F4,$3F,$EF,$BE,$1F,$F3,$AE,$1C,$2F,$75,$AF,$D0,$BA,$63,$CE
	.dc.b	$FF,$49,$7D,$BF,$0A,$FE,$28,$F1,$CB,$B7,$AF,$FD,$E6,$B8,$E5,$D2
	.dc.b	$3F,$AF,$F0,$FF,$9B,$2F,$86,$7E,$4F,$C6,$FF,$D3,$1E,$AF,$EE,$BE
	.dc.b	$05,$F8,$C7,$E1,$BE,$11,$FA,$0F,$E9,$F2,$FF,$3A,$DE,$EB,$52,$F9
	.dc.b	$AE,$7E,$9F,$FB,$B3,$7F,$DE,$5F,$F0,$6F,$63,$F9,$63,$C7,$C4,$67
	.dc.b	$FE,$22,$AE,$3A,$78,$A7,$3F,$3B,$FE,$41,$BE,$17,$E4,$FC,$6B,$D3
	.dc.b	$FB,$33,$F5,$67,$D4,$FE,$8F,$F4,$BF,$E8,$9F,$8A,$EA,$73,$E4,$A7
	.dc.b	$FA,$47,$E3,$19,$DF,$CF,$8B,$F6,$7F,$3B,$D7,$2E,$9F,$1F,$E3,$8C
	.dc.b	$63,$CE,$F1,$97,$6C,$BF,$AA,$7E,$4F,$FE,$C5,$F8,$4F,$F7,$17,$F0
	.dc.b	$0F,$FB,$67,$E7,$7E,$6F,$EF,$6F,$9E,$D7,$E2,$AA,$D7,$62,$BF,$C6
	.dc.b	$DD,$DF,$EA,$6C,$7F,$DD,$9F,$9C,$FF,$57,$7A,$ED,$FC,$AD,$EF,$F5
	.dc.b	$CD,$3E,$DF,$E8,$1F,$E6,$B9,$CC,$7D,$EF,$CA,$7F,$CD,$DF,$10,$FC
	.dc.b	$6D,$FB,$4B,$F3,$DF,$F5,$67,$C2,$BB,$1E,$77,$FB,$87,$0E,$17,$BF
	.dc.b	$33,$F8,$DD,$7F,$00,$E1,$FF,$5E,$D7,$5F,$D7,$7F,$B7,$3C,$38,$7F
	.dc.b	$A1,$3C,$B3,$EF,$F2,$BE,$4D,$7C,$71,$E2,$FD,$47,$FD,$09,$FB,$9F
	.dc.b	$F9,$9F,$E2,$1F,$89,$BF,$21,$FE,$79,$F4,$2F,$D7,$3F,$DE,$D7,$BA
	.dc.b	$D5,$FF,$83,$3C,$3C,$FD,$D7,$FC,$01,$DF,$FA,$A5,$F5,$6C,$7F,$47
	.dc.b	$BE,$3D,$5E,$28,$C6,$7F,$D2,$F8,$B1,$97,$92,$3C,$7F,$B6,$27,$FD
	.dc.b	$BF,$F8,$0F,$F6,$AF,$D8,$9F,$9B,$7F,$A0,$7F,$46,$79,$6F,$BA,$E1
	.dc.b	$BF,$FB,$FB,$0E,$17,$BA,$7E,$41,$AA,$1F,$EA,$1F,$61,$FB,$07,$F8
	.dc.b	$77,$0F,$04,$FF,$71,$79,$F0,$33,$F8,$5A,$F9,$B0,$6B,$E6,$AD,$75
	.dc.b	$7F,$AD,$70,$E3,$87,$A3,$F2,$5F,$16,$5F,$EB,$1F,$C1,$FE,$AF,$E7
	.dc.b	$BF,$EB,$3F,$11,$F8,$03,$F8,$07,$79,$2F,$6F,$F6,$49,$72,$EA,$7E
	.dc.b	$FE,$3E,$95,$E7,$FB,$5E,$38,$BF,$31,$F3,$46,$3C,$1B,$D8,$D7,$C9
	.dc.b	$9B,$F6,$DF,$4D,$BE,$71,$F9,$57,$C5,$F9,$37,$E0,$1E,$AF,$D1,$37
	.dc.b	$5A,$E5,$56,$B2,$FF,$5F,$75,$7D,$4E,$2F,$AB,$9F,$5C,$FF,$CA,$0F
	.dc.b	$7F,$22,$E5,$CF,$1E,$FC,$3F,$EE,$3A,$E1,$9F,$6D,$7F,$6D,$97,$8B
	.dc.b	$2F,$34,$7C,$EF,$D9,$3F,$21,$FA,$8F,$87,$F1,$77,$EA,$CF,$D0,$1F
	.dc.b	$06,$FD,$E9,$F8,$DF,$E2,$3F,$26,$E3,$AF,$E3,$31,$E3,$87,$0B,$DF
	.dc.b	$8B,$FE,$A9,$E5,$69,$DF,$4E,$D8,$FF,$A7,$63,$1F,$8A,$7F,$A8,$9D
	.dc.b	$63,$CF,$47,$B3,$F9,$AC,$79,$E1,$E1,$C7,$D1,$B7,$9E,$DE,$76,$F2
	.dc.b	$F7,$7C,$43,$F4,$7F,$F4,$37,$D4,$FF,$BB,$78,$7F,$65,$A7,$4D,$10
	.dc.b	$BE,$1F,$80,$FF,$BC,$BC,$AC,$FE,$9F,$F0,$AF,$75,$D5,$F6,$1F,$15
	.dc.b	$7E,$6B,$ED,$8C,$67,$E2,$BF,$F2,$E4,$73,$C7,$C1,$39,$FD,$4F,$D5
	.dc.b	$3F,$3F,$DB,$7C,$1B,$F3,$DF,$E0,$9E,$3B,$E8,$1F,$77,$CE,$F7,$5A
	.dc.b	$97,$EE,$59,$C3,$C9,$8F,$C0,$B7,$E6,$78,$D5,$1C,$BD,$AF,$D3,$7D
	.dc.b	$77,$92,$F8,$61,$E3,$9F,$F4,$94,$7A,$FE,$D9,$F4,$9F,$EC,$CF,$C9
	.dc.b	$F9,$FC,$D7,$FA,$EF,$A7,$E8,$69,$FE,$72,$AD,$68,$A7,$EA,$7C,$FF
	.dc.b	$2A,$F2,$65,$D3,$3F,$9E,$7D,$87,$BF,$E2,$4D,$3B,$CF,$8F,$2E,$BF
	.dc.b	$98,$7A,$74,$F0,$5E,$F4,$F0,$7E,$C3,$F7,$7F,$02,$7E,$90,$FB,$AF
	.dc.b	$E2,$3E,$73,$F4,$17,$EC,$AF,$D1,$38,$5E,$FE,$05,$7B,$A7,$E8,$FA
	.dc.b	$F8,$6F,$A3,$9F,$6C,$EE,$3C,$70,$ED,$EB,$7C,$FA,$D6,$5D,$34,$FF
	.dc.b	$43,$F1,$F8,$31,$F2,$C7,$93,$F3,$6F,$57,$EE,$6F,$A1,$FE,$45,$F6
	.dc.b	$AF,$C7,$7F,$C7,$FF,$9F,$6A,$73,$1F,$E4,$2F,$DC,$B8,$FE,$61,$E6
	.dc.b	$C3,$96,$9F,$43,$64,$75,$F8,$1F,$F3,$DE,$1F,$E4,$AA,$E9,$18,$FD
	.dc.b	$4F,$D5,$FB,$6F,$F8,$B7,$F7,$CF,$F7,$17,$EB,$0F,$C3,$7F,$73,$E3
	.dc.b	$FE,$CF,$45,$EF,$F1,$75,$EF,$F6,$1F,$AD,$3F,$E3,$7D,$3C,$8F,$FB
	.dc.b	$51,$BF,$64,$CF,$BF,$4F,$8A,$31,$8F,$4F,$92,$C5,$79,$35,$7D,$0F
	.dc.b	$E0,$FC,$8F,$D7,$7A,$2F,$E0,$17,$E9,$5F,$E7,$FF,$C9,$35,$EA,$FE
	.dc.b	$BA,$73,$F7,$DE,$AE,$5F,$66,$3F,$B7,$E1,$F9,$4F,$F8,$D5,$B7,$C7
	.dc.b	$1F,$F7,$94,$74,$CB,$C1,$1F,$EF,$8B,$E7,$5B,$C1,$FB,$47,$F1,$5F
	.dc.b	$13,$F7,$3E,$AF,$ED,$7D,$AF,$E0,$DE,$D6,$7C,$EF,$DE,$47,$B3,$2F
	.dc.b	$E4,$57,$C3,$5F,$D6,$BF,$C0,$BD,$9F,$E0,$EF,$4F,$6B,$F8,$8F,$B5
	.dc.b	$FC,$0E,$31,$ED,$7F,$02,$C3,$9F,$ED,$8F,$D2,$FF,$C9,$1F,$9F,$FF
	.dc.b	$B7,$BE,$BF,$FC,$2F,$5A,$9C,$F6,$7F,$93,$30,$F5,$61,$C7,$5F,$1A
	.dc.b	$FD,$87,$C6,$F5,$FA,$E6,$7F,$EE,$AF,$C3,$7C,$93,$98,$FF,$76,$FB
	.dc.b	$9E,$7F,$7B,$F8,$3F,$67,$F1,$4F,$C8,$7E,$44,$63,$9B,$F8,$97,$C0
	.dc.b	$E3,$E0,$7E,$0C,$FF,$11,$DB,$FB,$D1,$C3,$E0,$7F,$8C,$FC,$17,$57
	.dc.b	$F8,$7C,$67,$E0,$7F,$87,$4B,$E0,$7E,$B7,$E8,$CF,$D3,$9F,$9E,$BF
	.dc.b	$31,$7E,$83,$FE,$33,$FC,$A7,$75,$A9,$CF,$CE,$BF,$2B,$E7,$E7,$FD
	.dc.b	$6C,$BA,$6D,$F6,$9E,$D9,$F5,$CF,$FD,$C3,$F2,$7F,$E1,$3F,$09,$5A
	.dc.b	$CF,$C3,$F8,$9F,$F7,$1F,$D4,$7F,$C3,$1F,$5A,$F9,$57,$E2,$6F,$C8
	.dc.b	$5F,$A2,$3F,$00,$7E,$4A,$FD,$07,$FC,$C1,$97,$2C,$78,$E1,$C2,$F7
	.dc.b	$F0,$FA,$E3,$FF,$2C,$E9,$FD,$E1,$F9,$E6,$43,$E3,$D3,$31,$98,$C5
	.dc.b	$F9,$3C,$0F,$F1,$87,$98,$CF,$C5,$87,$A7,$6F,$47,$9D,$EF,$CB,$C3
	.dc.b	$AF,$9A,$7D,$9B,$BF,$19,$ED,$F8,$53,$F6,$F7,$E9,$1F,$D0,$3E,$CB
	.dc.b	$FD,$BB,$F9,$DE,$A7,$3D,$3F,$E4,$DF,$C4,$1F,$AA,$F2,$F8,$DF,$4F
	.dc.b	$1F,$97,$83,$E0,$E9,$D5,$F8,$31,$97,$E0,$C7,$87,$4F,$07,$D2,$FA
	.dc.b	$67,$39,$F8,$35,$FF,$44,$E1,$C7,$9F,$FC,$7F,$F8,$FB,$F7,$67,$E4
	.dc.b	$6F,$C8,$DF,$C0,$2F,$2B,$E6,$7D,$63,$0E,$17,$BA,$7E,$5B,$A9,$F5
	.dc.b	$F6,$FE,$22,$DF,$B0,$63,$83,$EE,$C7,$3D,$3C,$3A,$7F,$89,$36,$F3
	.dc.b	$4E,$74,$F2,$7A,$5E,$DD,$FD,$38,$70,$DB,$F4,$58,$7B,$7E,$85,$FA
	.dc.b	$37,$E5,$BF,$39,$F7,$7E,$03,$FE,$5D,$AD,$4E,$70,$FF,$44,$FC,$AF
	.dc.b	$E2,$70,$FE,$FD,$BF,$D0,$E1,$FD,$0C,$74,$8C,$1F,$96,$7B,$F8,$8C
	.dc.b	$FE,$67,$BF,$E7,$9C,$C7,$CF,$F7,$7E,$B7,$CE,$FF,$81,$BC,$87,$E8
	.dc.b	$9F,$E3,$AF,$A3,$7E,$F4,$FD,$AB,$F4,$17,$BA,$D4,$BF,$6F,$CF,$8E
	.dc.b	$FC,$3F,$E7,$70,$30,$EC,$FA,$0C,$4F,$F8,$FB,$0F,$07,$D0,$79,$23
	.dc.b	$1F,$23,$FE,$1F,$FA,$FC,$7C,$4B,$F5,$97,$E2,$EF,$CF,$DF,$8F,$3F
	.dc.b	$64,$F1,$DE,$EB,$52,$FD,$97,$3B,$78,$F7,$FF,$13,$FE,$43,$E8,$7F
	.dc.b	$C5,$38,$71,$CB,$FC,$FB,$5D,$33,$F0,$C7,$FA,$1A,$3E,$0F,$89,$FC
	.dc.b	$AB,$F2,$B7,$D9,$BF,$56,$7E,$7F,$AF,$87,$56,$A5,$F8,$E7,$CA,$FC
	.dc.b	$DB,$C5,$CC,$CA,$9D,$72,$FE,$E5,$E2,$7F,$AE,$F8,$3F,$C7,$F3,$9F
	.dc.b	$C1,$3E,$57,$C4,$7F,$91,$BF,$47,$7E,$C8,$FD,$2D,$F9,$4F,$D8,$AD
	.dc.b	$4E,$7E,$6B,$1C,$9F,$17,$6F,$F1,$C3,$6B,$78,$76,$FB,$E6,$5D,$BF
	.dc.b	$46,$FB,$58,$7F,$94,$88,$7E,$18,$F6,$98,$F8,$9F,$E1,$1D,$1F,$CA
	.dc.b	$B2,$FF,$5C,$FA,$5F,$C3,$E1,$1F,$EA,$73,$F9,$5F,$17,$FB,$6E,$3F
	.dc.b	$98,$FE,$3B,$CA,$5F,$05,$FE,$FE,$F9,$87,$E5,$6F,$E0,$07,$E7,$7F
	.dc.b	$CE,$F1,$EB,$7B,$AD,$7E,$61,$9C,$7F,$C7,$C7,$F7,$0D,$3C,$8F,$63
	.dc.b	$F3,$97,$CB,$F1,$9F,$F4,$55,$FE,$56,$79,$61,$D7,$A7,$CC,$F7,$CF
	.dc.b	$F9,$0E,$3D,$3D,$5F,$EB,$1F,$9B,$3F,$02,$7E,$D0,$FC,$A3,$C9,$5A
	.dc.b	$9C,$FE,$51,$8D,$3C,$67,$E3,$0B,$3B,$FC,$83,$E3,$3C,$7E,$3E,$FC
	.dc.b	$D3,$3F,$06,$6F,$C2,$7E,$17,$E5,$6F,$D9,$FF,$D6,$BF,$B4,$7F,$82
	.dc.b	$3C,$DC,$C6,$17,$E8,$3F,$27,$F2,$DF,$E0,$7D,$3F,$6B,$0B,$C0,$7A
	.dc.b	$7C,$0B,$2F,$D0,$4F,$48,$C5,$7D,$DF,$C2,$9F,$72,$FC,$27,$FD,$71
	.dc.b	$F4,$0F,$CD,$7E,$C5,$6A,$73,$E0,$63,$2F,$8B,$FB,$DF,$8A,$B5,$E1
	.dc.b	$BC,$4E,$D9,$F5,$8F,$C7,$DF,$E4,$BC,$7F,$D7,$F1,$FE,$EB,$D2,$FD
	.dc.b	$F7,$AF,$FE,$05,$9C,$FA,$FF,$BD,$F5,$B1,$0F,$5B,$F9,$2B,$FB,$8C
	.dc.b	$7F,$92,$3C,$30,$FD,$EC,$77,$8F,$E0,$58,$F7,$F5,$7F,$B7,$47,$FA
	.dc.b	$FB,$F0,$E7,$C5,$BF,$60,$7F,$00,$3F,$47,$FF,$64,$79,$5F,$DE,$70
	.dc.b	$E1,$7B,$AD,$79,$9E,$6F,$6F,$CB,$4C,$A9,$DB,$2E,$B9,$78,$A7,$58
	.dc.b	$F3,$FD,$6F,$1E,$CF,$F0,$6F,$73,$E0,$BF,$C2,$3D,$BB,$C3,$F0,$30
	.dc.b	$FE,$1A,$FE,$B7,$D6,$FF,$22,$47,$F9,$13,$DA,$FE,$1B,$EE,$7F,$19
	.dc.b	$AE,$71,$8F,$6B,$C5,$F0,$6F,$A3,$FE,$5B,$F7,$3F,$EE,$7F,$C7,$7E
	.dc.b	$8C,$E7,$B7,$FC,$EF,$F2,$0F,$D0,$F3,$9F,$8F,$F4,$FF,$83,$7B,$C7
	.dc.b	$F0,$93,$CB,$2E,$BD,$DF,$07,$7F,$E3,$A7,$FB,$C3,$B3,$E7,$3F,$EF
	.dc.b	$2A,$FF,$14,$F4,$7F,$96,$9F,$F2,$3F,$D8,$AB,$EC,$47,$F9,$40,$FE
	.dc.b	$3C,$7F,$20,$72,$FF,$94,$0F,$F2,$57,$ED,$6F,$95,$7B,$6F,$B1,$7E
	.dc.b	$4F,$D7,$2F,$07,$43,$0E,$17,$BF,$D2,$F5,$97,$86,$FF,$C8,$67,$07
	.dc.b	$D2,$79,$6F,$96,$5E,$0C,$BC,$D5,$DB,$1E,$BF,$A0,$63,$0F,$67,$A4
	.dc.b	$E8,$F3,$7E,$89,$F9,$6F,$BF,$F0,$D7,$B6,$B5,$92,$BF,$37,$FA,$16
	.dc.b	$BE,$93,$FA,$1F,$ED,$FE,$6C,$3A,$9E,$63,$96,$5D,$BE,$37,$C5,$18
	.dc.b	$C7,$A7,$B3,$8C,$3F,$A6,$FA,$8F,$D3,$BF,$CA,$5F,$BB,$3F,$5F,$6B
	.dc.b	$E1,$CB,$F8,$35,$6A,$5F,$AE,$67,$E9,$FC,$95,$F5,$F1,$B9,$76,$C3
	.dc.b	$AE,$7E,$1A,$E7,$F5,$8D,$1B,$67,$D2,$7C,$93,$F4,$6F,$EB,$5F,$89
	.dc.b	$7F,$A0,$3C,$0F,$C2,$EB,$FD,$46,$73,$F2,$FE,$7F,$C7,$3C,$3A,$FF
	.dc.b	$57,$3F,$AD,$76,$FE,$4F,$3E,$B7,$FD,$97,$F4,$FF,$06,$9D,$FE,$57
	.dc.b	$FC,$F3,$7B,$8C,$6A,$FC,$B7,$EC,$FE,$D6,$FD,$97,$FD,$E1,$F4,$2F
	.dc.b	$A5,$7D,$83,$E5,$FF,$C8,$9C,$9F,$9C,$C3,$87,$E2,$7F,$E3,$5B,$DD
	.dc.b	$7F,$8F,$6B,$58,$FF,$BC,$F8,$BD,$27,$F8,$13,$0F,$07,$E2,$3F,$41
	.dc.b	$E7,$9F,$87,$3F,$05,$79,$AF,$CB,$97,$93,$CC,$FC,$2F,$E8,$3F,$E1
	.dc.b	$7F,$9F,$FF,$71,$FE,$95,$FD,$87,$F1,$8F,$CA,$CE,$7E,$25,$3F,$D9
	.dc.b	$7A,$F3,$EA,$7E,$FF,$4F,$28,$BE,$15,$FB,$9E,$1F,$DD,$C6,$38,$BC
	.dc.b	$37,$FC,$D3,$DB,$FD,$07,$CB,$F4,$8F,$E5,$B1,$FA,$3F,$DA,$31,$F1
	.dc.b	$1F,$D0,$21,$FB,$47,$2A,$DF,$CA,$DF,$86,$7C,$27,$F9,$03,$B5,$FA
	.dc.b	$B7,$59,$CF,$C4,$25,$FB,$96,$6F,$FC,$8F,$E2,$73,$7F,$C9,$3C,$5D
	.dc.b	$FE,$0F,$F9,$87,$E3,$6F,$C5,$DF,$99,$7A,$67,$31,$F8,$3A,$F8,$3B
	.dc.b	$8F,$F3,$DF,$C2,$B0,$7C,$6F,$58,$C7,$C7,$FB,$6B,$F9,$7D,$F2,$7F
	.dc.b	$C9,$5D,$7F,$C2,$BA,$FF,$C9,$5F,$0D,$63,$F9,$59,$4F,$F2,$59,$FC
	.dc.b	$00,$C7,$BF,$FC,$57,$F6,$A7,$EC,$8F,$DD,$73,$F0,$09,$CF,$D5,$61
	.dc.b	$7D,$AF,$F0,$EC,$FA,$E3,$CF,$F1,$2C,$67,$FE,$45,$FC,$E7,$F4,$71
	.dc.b	$E3,$F5,$FF,$7B,$1F,$A6,$FA,$3C,$7A,$A8,$FE,$FB,$78,$7F,$A0,$3E
	.dc.b	$BF,$C1,$1F,$9D,$BF,$30,$7E,$92,$F8,$CD,$4E,$61,$F8,$FE,$2B,$F2
	.dc.b	$9E,$E7,$C4,$E0,$F8,$B8,$F4,$97,$E0,$BE,$AC,$DF,$64,$FB,$2F,$C4
	.dc.b	$7E,$78,$FC,$53,$E7,$4E,$7C,$94,$F0,$EF,$FE,$B7,$A7,$D2,$AD,$7C
	.dc.b	$87,$57,$87,$93,$FC,$E3,$D8,$FF,$45,$79,$BF,$96,$CB,$F3,$38,$FF
	.dc.b	$64,$9C,$F9,$7F,$9D,$F3,$3F,$A5,$C7,$FA,$47,$7F,$E8,$8E,$BC,$D6
	.dc.b	$5D,$7C,$6F,$C9,$67,$FA,$2C,$B1,$0F,$89,$7C,$9B,$C1,$7E,$EE,$F0
	.dc.b	$3B,$DF,$E6,$EF,$75,$AF,$0B,$33,$CF,$0E,$5F,$89,$76,$BD,$FD,$37
	.dc.b	$83,$D1,$FD,$3F,$A5,$C1,$FD,$4F,$C0,$5F,$43,$F4,$DE,$8E,$DF,$D2
	.dc.b	$F2,$F6,$FC,$59,$F8,$8B,$F7,$B7,$E4,$8F,$D1,$B5,$39,$8F,$F0,$C7
	.dc.b	$8F,$FA,$4E,$BF,$F5,$C6,$1C,$B0,$FF,$6D,$E3,$C1,$FF,$46,$74,$7D
	.dc.b	$1E,$4F,$F1,$09,$CA,$7F,$D6,$F7,$8F,$91,$FE,$E3,$F0,$7F,$B4,$BF
	.dc.b	$F0,$DB,$FE,$BC,$F5,$FF,$B3,$BF,$EB,$AB,$FF,$01,$BF,$B6,$F9,$EF
	.dc.b	$F0,$7F,$F0,$0B,$F1,$4C,$D7,$FB,$12,$73,$F8,$A6,$23,$FC,$4B,$F4
	.dc.b	$3F,$EB,$1C,$BB,$C6,$FE,$87,$FC,$51,$F1,$7E,$48,$FF,$66,$F8,$1F
	.dc.b	$C0,$7D,$EF,$E0,$17,$C2,$7F,$01,$EE,$FF,$01,$9F,$F5,$A4,$E6,$BF
	.dc.b	$84,$7B,$7C,$EF,$7E,$E3,$C4,$C7,$CD,$FF,$83,$7C,$0F,$ED,$1F,$B4
	.dc.b	$79,$E3,$1F,$2A,$FC,$6F,$1F,$3A,$BF,$F3,$06,$BE,$3B,$E5,$F5,$7D
	.dc.b	$FE,$0B,$F8,$4F,$CA,$77,$AF,$44,$63,$CD,$F8,$8F,$F0,$EF,$AE,$E1
	.dc.b	$F0,$9F,$C3,$7A,$3F,$C4,$BE,$6F,$FB,$DB,$A3,$FB,$0F,$EA,$7F,$E3
	.dc.b	$48,$C7,$CA,$7F,$60,$FD,$23,$1E,$2F,$F5,$FE,$8F,$E5,$7B,$7F,$95
	.dc.b	$AF,$7E,$E7,$4A,$FE,$D1,$E1,$D7,$FA,$F7,$E4,$7C,$9F,$EC,$1E,$D5
	.dc.b	$7F,$D1,$7C,$BB,$7F,$D1,$BF,$57,$7D,$8F,$EA,$FF,$8F,$ED,$FE,$B3
	.dc.b	$F0,$17,$0F,$F6,$26,$D7,$D9,$E5,$D7,$FD,$5B,$F2,$7E,$CF,$F5,$8F
	.dc.b	$CA,$6F,$AD,$FC,$BB,$8F,$D5,$B7,$F5,$53,$FA,$AF,$47,$8D,$C1,$E7
	.dc.b	$3F,$AD,$79,$FF,$E4,$E3,$F3,$DE,$9F,$F4,$E7,$F3,$DA,$7F,$95,$5F
	.dc.b	$CF,$5F,$F9,$F1,$FE,$A7,$1F,$88,$F8,$EF,$F3,$1C,$63,$C9,$FE,$A9
	.dc.b	$F1,$D8,$AF,$E9,$BE,$67,$F9,$FD,$C1,$F9,$A7,$F8,$BB,$81,$E0,$FF
	.dc.b	$4E,$F1,$B0,$FF,$50,$FC,$86,$F8,$4F,$BB,$F5,$5F,$F6,$6A,$FF,$28
	.dc.b	$F0,$62,$03,$E2,$FF,$A3,$7E,$81,$DD,$FE,$91,$FB,$0D,$F9,$0F,$F2
	.dc.b	$DD,$5F,$AA,$E4,$FF,$4A,$F8,$DF,$F0,$87,$E7,$FE,$7F,$8C,$FE,$91
	.dc.b	$FA,$05,$9F,$E6,$F5,$79,$F8,$D7,$F4,$DF,$CF,$FF,$D1,$3D,$5F,$53
	.dc.b	$C2,$FC,$2D,$7F,$AA,$23,$1E,$0B,$E3,$3F,$A9,$CE,$67,$FA,$4F,$CD
	.dc.b	$FF,$A1,$9D,$F8,$75,$FC,$17,$97,$FC,$F3,$C1,$EF,$18,$9F,$E8,$9F
	.dc.b	$86,$62,$3F,$A1,$47,$FB,$23,$0E,$D1,$8D,$BA,$3F,$E2,$1F,$13,$F2
	.dc.b	$93,$A8,$FF,$11,$FC,$D3,$C5,$E9,$1F,$E2,$6F,$4B,$8B,$F9,$4C,$33
	.dc.b	$F0,$BF,$CD,$FF,$3F,$BF,$11,$FD,$07,$F2,$9B,$AF,$0A,$FE,$8B,$F2
	.dc.b	$9F,$EC,$DF,$CF,$3F,$DA,$98,$FF,$33,$FA,$97,$D5,$BF,$59,$FF,$6B
	.dc.b	$56,$A9,$F1,$6B,$DF,$E8,$FA,$9F,$F6,$D1,$CF,$3E,$CE,$0E,$33,$F9
	.dc.b	$9E,$36,$1D,$7E,$2B,$FD,$9B,$F5,$FC,$EF,$FD,$C0,$7F,$39,$F2,$1F
	.dc.b	$1B,$9F,$B2,$FD,$4D,$FF,$B5,$2B,$07,$F4,$9F,$8C,$FD,$E6,$3E,$91
	.dc.b	$F6,$0E,$37,$DD,$70,$BF,$CE,$D6,$A7,$FD,$D7,$AF,$9A,$7D,$53,$F1
	.dc.b	$9F,$69,$E6,$3D,$77,$FA,$09,$E1,$19,$9E,$9E,$3F,$78,$FE,$01,$3A
	.dc.b	$CF,$F8,$07,$A7,$FE,$4F,$F5,$35,$EA,$FF,$B9,$2E,$CF,$EF,$AF,$AF
	.dc.b	$E2,$8F,$C3,$9E,$83,$F1,$BF,$EA,$5F,$60,$FD,$BD,$F6,$FF,$37,$0E
	.dc.b	$17,$BF,$91,$FF,$5B,$57,$F0,$8C,$BA,$E5,$D3,$4F,$E2,$E7,$15,$EB
	.dc.b	$E2,$31,$87,$3C,$FF,$D4,$F1,$8D,$3C,$7E,$37,$C8,$D3,$FC,$13,$E3
	.dc.b	$7F,$A4,$33,$F5,$BF,$E0,$CF,$AA,$F3,$3F,$D3,$F0,$E0,$FF,$32,$3E
	.dc.b	$77,$2F,$F9,$F2,$F9,$FE,$5D,$C2,$75,$FA,$35,$7F,$9C,$FE,$45,$5F
	.dc.b	$E7,$DF,$21,$5F,$F4,$A7,$E1,$75,$7C,$8D,$7F,$29,$CB,$FE,$1B,$CF
	.dc.b	$FD,$16,$FF,$3E,$D3,$F9,$DE,$DF,$80,$8F,$E7,$F3,$A9,$F8,$DE,$CF
	.dc.b	$27,$E4,$7E,$19,$5F,$9B,$F8,$7D,$7F,$A2,$FE,$21,$5F,$F4,$CF,$E8
	.dc.b	$35,$FE,$67,$F9,$FD,$7F,$D6,$B8,$F7,$7E,$BF,$EF,$9F,$AB,$FD,$3F
	.dc.b	$E7,$F1,$5D,$FF,$C7,$3C,$55,$FB,$5C,$1F,$98,$AF,$F1,$D6,$19,$9F
	.dc.b	$EC,$3F,$91,$57,$FB,$37,$E4,$75,$F9,$37,$C5,$FF,$22,$F1,$E5,$7C
	.dc.b	$0D,$FB,$4C,$47,$EF,$FF,$4C,$F9,$99,$7F,$4E,$7E,$63,$FE,$0E,$DF
	.dc.b	$FD,$73,$1F,$32,$34,$7E,$47,$E2,$B9,$3F,$93,$F1,$2B,$F9,$8E,$7A
	.dc.b	$FF,$70,$E7,$AF,$FB,$13,$F0,$0A,$FF,$5E,$F1,$2A,$E9,$BD,$3F,$C5
	.dc.b	$7F,$0F,$FC,$CE,$D4,$FC,$B5,$57,$FD,$AB,$C5,$FC,$0B,$F7,$6F,$6C
	.dc.b	$9F,$6B,$FC,$29,$F6,$3D,$27,$F9,$11,$FE,$9B,$7F,$1B,$C1,$FE,$DA
	.dc.b	$8F,$AB,$5F,$E9,$59,$FE,$5C,$7F,$B7,$BD,$84,$66,$1E,$67,$81,$CF
	.dc.b	$97,$FC,$F5,$FF,$72,$4E,$F8,$DF,$3B,$FE,$C2,$FC,$E1,$FA,$73,$F4
	.dc.b	$E7,$C9,$FF,$A3,$70,$E1,$7B,$AD,$4F,$FB,$FE,$73,$F7,$7F,$8E,$CF
	.dc.b	$ED,$DB,$F4,$FC,$69,$F8,$18,$E5,$93,$CE,$FF,$08,$E3,$CB,$4F,$1C
	.dc.b	$BF,$08,$CE,$5F,$EF,$EF,$5F,$FC,$75,$F1,$2F,$8B,$7C,$53,$F0,$17
	.dc.b	$CA,$7C,$2D,$FF,$C3,$B8,$70,$BD,$D3,$F1,$9D,$57,$F4,$9E,$8F,$EF
	.dc.b	$55,$FE,$B2,$DF,$E7,$61,$C7,$27,$DB,$79,$30,$EB,$80,$FE,$A5,$F1
	.dc.b	$9F,$EB,$7F,$98,$7A,$7D,$B1,$FF,$59,$CE,$7E,$9F,$17,$C6,$7F,$D6
	.dc.b	$DF,$47,$D7,$1F,$24,$F4,$FB,$4F,$F5,$BF,$1F,$E1,$E3,$CF,$AF,$C1
	.dc.b	$E2,$F5,$C7,$9F,$8B,$DD,$C1,$F8,$78,$C6,$3A,$78,$F0,$77,$7C,$B9
	.dc.b	$7F,$90,$F7,$D4,$78,$78,$1B,$E3,$7F,$3E,$37,$83,$FD,$AD,$E3,$F3
	.dc.b	$C3,$D9,$1F,$ED,$59,$D6,$5F,$ED,$AF,$03,$F3,$7F,$89,$7F,$94,$A5
	.dc.b	$FA,$FB,$E7,$FF,$05,$FB,$1B,$F5,$67,$CE,$3E,$91,$F9,$9F,$2C,$7F
	.dc.b	$CB,$78,$F1,$EE,$63,$C7,$0E,$1F,$AF,$6E,$FB,$E6,$FD,$83,$F9,$3E
	.dc.b	$BE,$3C,$B9,$57,$F6,$6E,$84,$E6,$FC,$BB,$79,$3E,$DF,$F3,$73,$9D
	.dc.b	$7C,$33,$E8,$FC,$C7,$F3,$FF,$AC,$FB,$FF,$69,$7C,$D7,$AB,$F4,$8F
	.dc.b	$F2,$E7,$0F,$F8,$73,$6F,$DF,$DE,$EB,$5F,$4B,$F0,$BE,$53,$FC,$73
	.dc.b	$C0,$AD,$7E,$3F,$FD,$01,$E1,$CD,$71,$FF,$3F,$E9,$E0,$86,$9E,$0F
	.dc.b	$A0,$F2,$D6,$B4,$EF,$AF,$F8,$57,$1F,$AD,$F7,$BF,$E4,$CF,$DE,$FF
	.dc.b	$85,$FC,$4F,$F1,$4F,$17,$E7,$EF,$E3,$CB,$FD,$1B,$7B,$F4,$57,$BF
	.dc.b	$BB,$F6,$D6,$BE,$8D,$FC,$03,$E7,$86,$1D,$DF,$A5,$A7,$8A,$9A,$78
	.dc.b	$B3,$FD,$AD,$74,$38,$CE,$B3,$F0,$CB,$F0,$7F,$E2,$3F,$6F,$F9,$1F
	.dc.b	$C6,$5F,$8C,$FF,$93,$3E,$99,$F8,$43,$EF,$FF,$8C,$7A,$FF,$CC,$B0
	.dc.b	$E1,$7B,$AD,$7B,$7F,$C4,$B3,$56,$BB,$7F,$B9,$BC,$36,$BB,$FF,$AC
	.dc.b	$B5,$F0,$D7,$F9,$CF,$1E,$73,$DC,$D6,$5D,$F4,$FE,$4D,$19,$1F,$95
	.dc.b	$3F,$62,$FF,$61,$7C,$3B,$F2,$07,$E0,$AF,$C6,$FF,$82,$EF,$FB,$1C
	.dc.b	$38,$7D,$D3,$0E,$17,$BF,$6F,$B5,$6B,$3F,$D2,$61,$C7,$37,$E7,$7E
	.dc.b	$C9,$F1,$C7,$8B,$4F,$0F,$7F,$D0,$9F,$F6,$8E,$1E,$0C,$FB,$FB,$99
	.dc.b	$9F,$E5,$9F,$85,$FF,$91,$7D,$87,$C1,$7B,$D4,$67,$E3,$70,$FF,$78
	.dc.b	$E9,$FA,$37,$FE,$DE,$AD,$9F,$DE,$B7,$FE,$E1,$5D,$7D,$CE,$33,$FE
	.dc.b	$53,$FC,$37,$B3,$E1,$CA,$BF,$E5,$7C,$AB,$FD,$04,$FE,$75,$93,$F2
	.dc.b	$0F,$E3,$9F,$98,$3E,$FF,$FC,$DB,$F8,$C3,$E7,$DF,$8D,$3E,$F7,$F9
	.dc.b	$06,$3C,$78,$7F,$DE,$B8,$F1,$FA,$D6,$3C,$70,$E1,$FB,$A7,$F8,$13
	.dc.b	$4E,$FA,$7F,$9E,$B1,$FE,$31,$18,$FD,$63,$FC,$86,$EB,$1E,$7A,$F8
	.dc.b	$BF,$05,$FE,$E7,$39,$CF,$B7,$6E,$31,$1F,$CB,$3F,$18,$7E,$A9,$FC
	.dc.b	$CF,$F2,$A7,$EC,$AF,$CD,$DF,$B3,$79,$70,$E3,$F1,$2C,$78,$FC,$53
	.dc.b	$1E,$38,$3F,$67,$E1,$87,$7F,$97,$F4,$BC,$B0,$F0,$BE,$E3,$1A,$78
	.dc.b	$F4,$FF,$12,$C7,$2C,$FC,$13,$9F,$47,$D9,$2F,$B3,$FF,$54,$FD,$B7
	.dc.b	$FA,$BB,$EE,$3F,$D5,$BD,$9F,$4E,$B5,$F4,$BF,$EB,$4A,$D7,$D4,$F9
	.dc.b	$B0,$FF,$58,$D7,$8C,$DF,$4F,$A1,$E3,$F7,$1F,$EA,$BF,$C4,$FC,$35
	.dc.b	$E0,$FB,$5F,$F1,$27,$DC,$F9,$3C,$87,$CF,$7F,$87,$BF,$73,$7E,$A4
	.dc.b	$FC,$99,$FB,$03,$F2,$5F,$8F,$87,$0F,$A3,$61,$C2,$F7,$FB,$BE,$B2
	.dc.b	$7E,$D3,$E5,$C3,$87,$0F,$A5,$E5,$87,$1E,$8F,$25,$F8,$F5,$F1,$7D
	.dc.b	$77,$F8,$EB,$F2,$AF,$EF,$5F,$B6,$BF,$5A,$7E,$71,$FE,$86,$F9,$47
	.dc.b	$D1,$BF,$2A,$7E,$8B,$FE,$5C,$C7,$8E,$1C,$3E,$F3,$FB,$2D,$EF,$F5
	.dc.b	$4D,$71,$7E,$4B,$4E,$DE,$C7,$FA,$7B,$1E,$3A,$DA,$D6,$5C,$F5,$F0
	.dc.b	$FE,$59,$E7,$9C,$D7,$97,$3F,$F3,$6C,$76,$1E,$BB,$F0,$F7,$ED,$5F
	.dc.b	$E0,$CF,$E0,$0F,$F2,$8E,$EF,$D5,$F8,$5E,$EB,$5D,$AA,$7E,$E9,$AC
	.dc.b	$5F,$94,$79,$F3,$EF,$1D,$9F,$F0,$87,$0F,$F8,$97,$2E,$9E,$73,$5F
	.dc.b	$F3,$0F,$EB,$5E,$BB,$E3,$AF,$83,$EE,$F1,$F7,$DF,$E8,$05,$F7,$7F
	.dc.b	$E6,$EF,$BD,$FF,$3F,$69,$FE,$78,$9C,$FD,$F7,$F9,$FB,$E1,$BE,$16
	.dc.b	$1C,$72,$F1,$9F,$E7,$BF,$BE,$F4,$47,$5C,$BF,$CF,$3E,$1F,$83,$E8
	.dc.b	$3F,$CF,$FF,$61,$E5,$EE,$7E,$42,$FC,$FD,$F3,$EF,$D2,$FC,$BF,$10
	.dc.b	$FC,$73,$E9,$61,$C2,$F7,$5A,$F0,$F3,$8B,$EC,$BE,$55,$EF,$87,$CE
	.dc.b	$7F,$2D,$9F,$06,$7D,$F7,$69,$E5,$9C,$E5,$FA,$EB,$FF,$4A,$FB,$5F
	.dc.b	$E5,$5F,$80,$7E,$48,$F9,$37,$BC,$F9,$17,$ED,$AF,$C5,$9F,$29,$FE
	.dc.b	$B8,$CB,$96,$3C,$70,$E1,$7B,$FC,$1F,$5A,$7F,$2E,$D3,$BE,$9F,$EF
	.dc.b	$AC,$78,$F6,$5A,$D6,$5C,$F5,$F0,$F8,$FE,$89,$CE,$1E,$69,$F2,$FE
	.dc.b	$57,$99,$7E,$2E,$FD,$65,$F8,$6B,$F4,$FF,$E7,$BF,$B2,$3F,$21,$7E
	.dc.b	$F7,$E6,$C7,$8E,$1C,$2F,$74,$FD,$47,$59,$75,$FC,$2B,$D1,$87,$0C
	.dc.b	$EC,$F8,$5F,$07,$C8,$B4,$FF,$5D,$D7,$2C,$FB,$4E,$7F,$15,$F5,$7E
	.dc.b	$0F,$FE,$7F,$F4,$BF,$CF,$DF,$33,$FE,$84,$F9,$37,$47,$F9,$CE,$73
	.dc.b	$ED,$FA,$6F,$FD,$1D,$F8,$47,$91,$FF,$3F,$E9,$FE,$93,$35,$F8,$47
	.dc.b	$F9,$FB,$EF,$DF,$84,$78,$23,$1F,$83,$7C,$BF,$A3,$7E,$93,$F8,$6F
	.dc.b	$A0,$FE,$73,$2F,$C7,$5F,$8A,$7C,$9B,$DF,$85,$FD,$87,$E3,$D7,$BF
	.dc.b	$A3,$D6,$5F,$EA,$3D,$BF,$D2,$5C,$7F,$EA,$7C,$78,$C6,$0F,$EE,$AF
	.dc.b	$4D,$3C,$3F,$37,$E5,$BF,$26,$5D,$BC,$2F,$C1,$7E,$97,$FD,$1B,$F5
	.dc.b	$D7,$E8,$2F,$D8,$5F,$AB,$FF,$AA,$BD,$BF,$C0,$5E,$EB,$53,$FE,$AE
	.dc.b	$E7,$CF,$CB,$FF,$70,$E0,$F9,$5D,$BF,$49,$FF,$04,$E1,$DD,$F4,$07
	.dc.b	$A6,$5E,$0F,$CA,$7C,$B6,$D3,$CB,$3E,$4F,$BB,$FF,$09,$E2,$FC,$73
	.dc.b	$FC,$5F,$F2,$1E,$9F,$8D,$FD,$AF,$E5,$5F,$E1,$3F,$D7,$D7,$BD,$BF
	.dc.b	$C5,$B7,$BF,$61,$7F,$E1,$9F,$A8,$DE,$5D,$3F,$56,$FC,$26,$1E,$08
	.dc.b	$EF,$AE,$23,$67,$47,$FA,$4E,$1D,$3E,$77,$84,$E7,$5F,$2F,$C4,$E2
	.dc.b	$1F,$89,$7F,$AD,$7F,$11,$FF,$61,$FE,$0D,$98,$C7,$C4,$7F,$B0,$7F
	.dc.b	$16,$EF,$7C,$3E,$29,$F8,$A7,$8F,$93,$DD,$39,$FC,$53,$1C,$3F,$EC
	.dc.b	$4D,$3F,$D8,$13,$CB,$3E,$BD,$9F,$0A,$3A,$FB,$9F,$C7,$BF,$4E,$7E
	.dc.b	$92,$FC,$33,$FB,$27,$E6,$0F,$DB,$B7,$CE,$BD,$D6,$BF,$6D,$CF,$B7
	.dc.b	$E2,$CF,$FC,$97,$9F,$57,$FD,$DD,$BF,$F9,$39,$FE,$25,$8F,$3D,$3C
	.dc.b	$1F,$3E,$FC,$87,$E4,$61,$E5,$9C,$FC,$17,$CB,$F5,$1F,$90,$BD,$F7
	.dc.b	$D4,$BF,$4C,$F9,$FF,$06,$FD,$1B,$E5,$E1,$C2,$F7,$5A,$FD,$BB,$38
	.dc.b	$F4,$E0,$F3,$9F,$E6,$4B,$E0,$E0,$FE,$8F,$1E,$1C,$BC,$1D,$3F,$BE
	.dc.b	$9E,$78,$F6,$8F,$E1,$B1,$F7,$3F,$5A,$7E,$9B,$FC,$33,$F2,$97,$9B
	.dc.b	$FE,$0F,$97,$D1,$6E,$B5,$39,$FC,$47,$1E,$D7,$CF,$C5,$87,$4F,$D6
	.dc.b	$30,$F6,$96,$5D,$BE,$B7,$FC,$2B,$7C,$A3,$1E,$DF,$E5,$FF,$1E,$FF
	.dc.b	$7E,$7E,$43,$FF,$00,$7E,$33,$F2,$71,$8F,$25,$F8,$9B,$F2,$1C,$F8
	.dc.b	$BF,$EF,$EC,$FE,$35,$FF,$BF,$5F,$E0,$FE,$1F,$FB,$F7,$F1,$FF,$F7
	.dc.b	$E6,$1D,$EF,$FD,$F5,$9F,$EE,$BF,$1F,$FD,$47,$BE,$F2,$FF,$DD,$FF
	.dc.b	$A2,$7F,$56,$F4,$BF,$D4,$FF,$45,$FA,$6A,$D7,$E2,$8A,$D7,$E1,$CF
	.dc.b	$DF,$75,$8F,$F9,$FF,$83,$F2,$5C,$7F,$AA,$C7,$8C,$7D,$D7,$9E,$1D
	.dc.b	$FD,$BF,$1D,$F8,$BE,$71,$AB,$EC,$3F,$7D,$F4,$2F,$CB,$FF,$D8,$5F
	.dc.b	$AE,$3F,$66,$FF,$06,$7E,$E7,$F9,$DD,$FF,$73,$7B,$F9,$25,$EF,$E9
	.dc.b	$16,$FD,$AB,$F8,$1D,$EF,$8B,$F2,$4F,$7F,$CD,$7F,$04,$F3,$C3,$96
	.dc.b	$5E,$4D,$3C,$7A,$7F,$87,$7E,$FF,$CF,$F5,$9F,$EF,$1C,$3A,$BF,$EE
	.dc.b	$DE,$FF,$DB,$47,$F8,$78,$FF,$71,$1F,$E2,$CC,$FC,$6F,$1C,$7D,$6F
	.dc.b	$18,$FF,$7A,$3C,$0E,$1E,$1F,$F1,$8F,$45,$F9,$77,$F8,$BF,$E4,$3F
	.dc.b	$DD,$9F,$9E,$7F,$6B,$F9,$25,$EE,$B5,$F8,$C2,$BF,$B8,$7E,$ED,$AD
	.dc.b	$7F,$86,$65,$D3,$9B,$F9,$C6,$3C,$74,$B5,$AC,$B9,$E2,$F9,$AF,$F7
	.dc.b	$0D,$F7,$C3,$FA,$A6,$7D,$E3,$38,$76,$FC,$59,$FA,$83,$F1,$6F,$F5
	.dc.b	$9F,$E8,$DF,$E8,$3F,$A9,$F2,$71,$FF,$09,$BD,$D6,$BF,$06,$7E,$F9
	.dc.b	$AA,$FF,$28,$D7,$E1,$DF,$EA,$79,$FC,$78,$EF,$18,$CF,$C5,$9F,$87
	.dc.b	$91,$A3,$F0,$5F,$3D,$F8,$6B,$53,$F9,$78,$F0,$7E,$4F,$FE,$B6,$FB
	.dc.b	$07,$DF,$3C,$3F,$F7,$47,$E0,$6F,$CA,$77,$F8,$6F,$F9,$2A,$B5,$F8
	.dc.b	$C2,$9F,$B7,$EB,$E3,$7E,$7E,$81,$A7,$77,$D4,$65,$F1,$61,$D3,$CA
	.dc.b	$F4,$CF,$0C,$BC,$1A,$7F,$BE,$FC,$EC,$E0,$FD,$ED,$FC,$00,$FC,$A1
	.dc.b	$F9,$83,$F2,$37,$EE,$5F,$DA,$BA,$FF,$B4,$E3,$C7,$0E,$17,$BA,$7C
	.dc.b	$FF,$C6,$CB,$AE,$DE,$83,$D6,$BE,$8F,$DF,$18,$C3,$9E,$6F,$77,$CF
	.dc.b	$7E,$6D,$7C,$B3,$F7,$35,$F1,$F8,$4F,$C8,$5F,$AD,$BF,$55,$FE,$C9
	.dc.b	$EE,$73,$BD,$D6,$A7,$38,$FD,$CC,$39,$6B,$E7,$E2,$FF,$7A,$FE,$4F
	.dc.b	$E6,$7F,$DE,$F8,$71,$F9,$CF,$ED,$55,$C6,$31,$AF,$CE,$FC,$01,$FA
	.dc.b	$63,$F0,$07,$EF,$CF,$CF,$5F,$94,$BF,$34,$FC,$6E,$1C,$2F,$75,$AF
	.dc.b	$E0,$1C,$F2,$F8,$EB,$AE,$BE,$77,$86,$5D,$79,$FC,$5A,$78,$63,$C1
	.dc.b	$F1,$BF,$86,$FA,$4F,$BD,$F8,$4B,$F1,$9F,$E0,$C7,$F9,$03,$F6,$AF
	.dc.b	$D7,$4E,$A3,$3F,$70,$FD,$7B,$3C,$8F,$D7,$BE,$DD,$F4,$DF,$F0,$0F
	.dc.b	$1F,$73,$B6,$1E,$1B,$FF,$55,$63,$E0,$FD,$47,$F3,$3F,$9E,$FF,$3A
	.dc.b	$F2,$BE,$47,$E7,$0F,$9B,$7E,$17,$FE,$5D,$F1,$3F,$C7,$77,$BE,$B5
	.dc.b	$EE,$BF,$D3,$F5,$AF,$8F,$7B,$DF,$90,$C3,$B7,$0F,$E5,$5E,$58,$77
	.dc.b	$FC,$BB,$C7,$6D,$3C,$7A,$7F,$21,$C3,$F8,$97,$DD,$BE,$31,$F8,$77
	.dc.b	$F9,$43,$F2,$F7,$E0,$8F,$CE,$5F,$79,$C2,$F7,$F7,$AB,$DD,$3F,$77
	.dc.b	$D6,$3F,$A6,$CB,$B7,$E2,$17,$1E,$38,$76,$FC,$1F,$F8,$0D,$78,$EF
	.dc.b	$C5,$A7,$8B,$5F,$E8,$BF,$86,$FF,$92,$FF,$1C,$7E,$83,$FE,$DB,$FC
	.dc.b	$A1,$E9,$FC,$6F,$BE,$FE,$03,$D7,$FE,$97,$C5,$87,$0F,$1D,$87,$0E
	.dc.b	$C6,$0F,$DD,$7B,$E1,$FC,$6F,$0E,$FE,$37,$F4,$ED,$7C,$5A,$FF,$BA
	.dc.b	$EB,$F9,$CC,$63,$3E,$BA,$FD,$8F,$0F,$FA,$7C,$E7,$5F,$07,$DC,$63
	.dc.b	$0F,$E5,$1F,$89,$BF,$0B,$72,$BF,$2B,$7E,$14,$FD,$B5,$FB,$3B,$F2
	.dc.b	$4F,$3E,$5F,$ED,$0C,$78,$E1,$C2,$F7,$FB,$A6,$B3,$ED,$E0,$7A,$F0
	.dc.b	$F0,$E1,$E0,$7D,$66,$34,$F1,$69,$E1,$7D,$11,$DD,$CE,$3E,$78,$F1
	.dc.b	$FD,$DF,$FB,$E3,$CA,$FE,$3D,$F5,$7F,$A3,$E8,$7E,$47,$AE,$4F,$D2
	.dc.b	$FB,$FF,$E2,$EF,$23,$EB,$FE,$CB,$9F,$C4,$7C,$DA,$34,$E8,$F9,$8C
	.dc.b	$61,$CB,$3E,$D1,$D7,$4F,$0F,$DD,$7F,$A4,$FC,$2F,$C2,$7D,$07,$E2
	.dc.b	$3E,$0F,$F0,$1F,$AB,$FD,$0F,$D6,$9F,$95,$FC,$CA,$D7,$8A,$AD,$7E
	.dc.b	$BB,$9F,$0F,$F9,$1E,$BE,$3B,$FC,$9F,$E3,$DE,$87,$8D,$FD,$CF,$D2
	.dc.b	$3E,$4C,$63,$AB,$F6,$57,$F6,$BF,$23,$7E,$D6,$FD,$31,$F9,$6F,$F0
	.dc.b	$2F,$2D,$FA,$84,$E4,$FC,$F6,$1C,$2F,$7E,$7F,$E9,$A9,$F8,$A7,$5C
	.dc.b	$BA,$67,$F2,$70,$EF,$F2,$BF,$39,$E5,$A7,$7F,$1F,$EF,$C6,$32,$FD
	.dc.b	$8F,$E4,$DF,$82,$F8,$17,$EC,$EF,$D9,$BF,$B9,$FE,$06,$FC,$9F,$FB
	.dc.b	$AF,$17,$E5,$AF,$75,$A9,$FE,$F3,$FA,$3E,$70,$E5,$BF,$DD,$BE,$BE
	.dc.b	$3F,$E0,$9E,$3F,$42,$CF,$B7,$D7,$78,$F0,$F1,$4E,$72,$FF,$11,$69
	.dc.b	$FC,$0F,$E8,$DF,$93,$7C,$DF,$41,$FA,$43,$F5,$95,$7C,$8B,$E1,$7B
	.dc.b	$AD,$4B,$97,$F6,$2F,$4E,$BF,$CB,$BD,$BA,$2D,$F1,$C7,$AF,$17,$08
	.dc.b	$C7,$C0,$FF,$95,$3F,$37,$FF,$74,$7C,$8B,$F6,$0F,$F0,$F7,$ED,$6F
	.dc.b	$BF,$7E,$B7,$FC,$5F,$E6,$3F,$CD,$38,$70,$BD,$D6,$BC,$C7,$E8,$DA
	.dc.b	$D7,$FC,$6D,$9F,$5F,$9D,$FF,$34,$E3,$C7,$E9,$AD,$6B,$2E,$7A,$7F
	.dc.b	$87,$63,$91,$FE,$75,$8C,$FD,$BF,$A6,$3F,$C8,$5E,$E3,$F0,$87,$EF
	.dc.b	$DF,$EC,$9F,$CA,$5F,$AA,$FB,$7F,$44,$E7,$7B,$AD,$4F,$C1,$FC,$0B
	.dc.b	$38,$7F,$9C,$71,$FF,$33,$ED,$FE,$4F,$7F,$85,$C7,$67,$27,$89,$F0
	.dc.b	$FB,$6C,$FC,$3F,$6B,$E7,$AE,$39,$78,$3C,$5C,$7D,$C3,$F2,$B7,$D5
	.dc.b	$FF,$30,$F4,$9F,$4A,$FB,$D7,$E2,$FF,$13,$0E,$17,$BA,$D7,$F0,$06
	.dc.b	$71,$7E,$AD,$F3,$B5,$FF,$35,$C7,$67,$D2,$62,$F9,$7D,$EF,$92,$7C
	.dc.b	$7F,$29,$FC,$1B,$17,$D9,$FA,$F2,$7E,$31,$FD,$93,$F1,$07,$E0,$9F
	.dc.b	$DF,$3F,$03,$F4,$7A,$2E,$AF,$95,$87,$0F,$87,$FF,$6E,$5E,$FF,$68
	.dc.b	$FB,$F3,$D3,$F3,$BF,$43,$BE,$2F,$3B,$C7,$3E,$FF,$44,$FA,$CF,$26
	.dc.b	$DE,$3C,$BF,$C1,$B3,$FE,$B2,$FC,$2B,$ED,$7E,$A3,$FE,$0B,$B7,$E2
	.dc.b	$CF,$CD,$BD,$BF,$98,$3F,$76,$FF,$63,$E1,$C3,$EA,$FB,$5E,$F9,$97
	.dc.b	$FE,$B8,$E5,$5E,$FF,$00,$FF,$63,$BC,$73,$BC,$5F,$E9,$9D,$7C,$59
	.dc.b	$73,$F6,$FC,$D3,$9D,$3B,$ED,$F4,$3C,$5C,$CB,$F5,$B7,$EC,$2F,$D9
	.dc.b	$9F,$A9,$7E,$8F,$B1,$F4,$0F,$C5,$FD,$78,$FF,$AA,$71,$E3,$87,$0F
	.dc.b	$3B,$EB,$5F,$FA,$B3,$93,$D7,$7B,$CF,$FD,$2E,$FE,$8C,$E6,$78,$DF
	.dc.b	$13,$FE,$CE,$E3,$FD,$46,$3D,$B4,$F2,$7E,$2B,$EA,$8C,$65,$F8,$5E
	.dc.b	$4F,$6F,$E9,$CF,$DF,$7F,$D7,$BF,$8E,$3F,$76,$7E,$67,$FE,$CB,$FC
	.dc.b	$63,$C3,$7B,$EB,$5E,$EB,$FD,$A1,$FB,$CE,$B1,$F8,$5C,$1E,$53,$FD
	.dc.b	$95,$87,$17,$01,$E9,$7C,$FA,$7C,$B5,$E4,$CF,$C7,$F2,$DF,$53,$F7
	.dc.b	$17,$8D,$FE,$3D,$F9,$DF,$E7,$FF,$A9,$7F,$24,$ED,$F8,$F5,$AF,$86
	.dc.b	$FF,$26,$FC,$82,$BF,$94,$FE,$7B,$AC,$7F,$C8,$DC,$1E,$4C,$3A,$FE
	.dc.b	$77,$F8,$A8,$47,$3C,$FB,$FC,$6F,$8E,$FC,$59,$75,$F1,$3F,$DD,$1D
	.dc.b	$EF,$CA,$BF,$8E,$7E,$77,$FE,$FE,$F1,$9F,$BF,$7E,$DF,$F5,$8F,$F2
	.dc.b	$0F,$EE,$8C,$38,$6F,$FE,$F1,$C3,$85,$EF,$D8,$7E,$73,$BC,$FA,$FE
	.dc.b	$47,$FA,$DC,$3F,$10,$64,$E4,$B9,$33,$19,$18,$F6,$F3,$1A,$FF,$92
	.dc.b	$74,$F1,$D6,$B0,$FD,$D6,$9E,$9F,$BC,$F7,$E7,$FC,$E7,$F0,$67,$EE
	.dc.b	$AF,$DC,$DF,$9A,$BF,$28,$7B,$8F,$CA,$1F,$94,$F8,$F1,$E3,$87,$0B
	.dc.b	$DD,$74,$DF,$C7,$5A,$DF,$DF,$18,$CD,$FB,$DF,$BA,$31,$8F,$3C,$DF
	.dc.b	$0B,$E8,$C0,$31,$F0,$EB,$E7,$AF,$F0,$5C,$FC,$7F,$C3,$1F,$B1,$7D
	.dc.b	$2F,$CF,$5F,$93,$BF,$78,$7C,$A3,$F8,$03,$85,$EE,$B5,$2F,$DF,$33
	.dc.b	$F9,$FF,$9F,$9F,$FB,$16,$7D,$BB,$F1,$AF,$9A,$31,$AF,$83,$5F,$F5
	.dc.b	$67,$1F,$D6,$9C,$E9,$E3,$FB,$9F,$76,$0F,$BA,$FE,$53,$F8,$27,$F8
	.dc.b	$A3,$F6,$87,$EC,$6F,$C9,$BF,$C4,$F9,$FE,$02,$F7,$F5,$8B,$DD,$3E
	.dc.b	$B3,$F1,$55,$FC,$D3,$DE,$F3,$E7,$DF,$E1,$3F,$D7,$0F,$F9,$DF,$8F
	.dc.b	$FB,$F5,$74,$C7,$C3,$FB,$07,$9A,$31,$AB,$F2,$0F,$4F,$C0,$BF,$6B
	.dc.b	$7E,$D8,$FB,$7F,$E6,$DF,$A6,$BF,$51,$7E,$A9,$F9,$DC,$38,$5E,$FE
	.dc.b	$29,$7B,$F9,$1D,$65,$D3,$87,$F1,$7C,$3F,$E1,$3C,$7C,$1F,$BE,$7B
	.dc.b	$1E,$7A,$F8,$27,$CB,$7E,$48,$F1,$DE,$33,$F0,$65,$E5,$EF,$EF,$18
	.dc.b	$D9,$E0,$63,$0F,$5C,$7F,$AB,$3A,$BF,$63,$A7,$83,$7F,$4C,$BF,$11
	.dc.b	$F7,$7E,$A0,$FD,$0D,$F7,$5F,$0B,$F4,$E7,$E0,$4F,$CE,$B7,$5A,$E7
	.dc.b	$56,$BF,$1E,$E7,$FA,$2F,$BF,$2E,$9C,$BF,$10,$FD,$3E,$E8,$F0,$69
	.dc.b	$DF,$F5,$CF,$AF,$18,$CB,$AF,$B3,$8D,$3C,$1A,$79,$3F,$2B,$F1,$E3
	.dc.b	$E7,$DD,$FB,$06,$31,$F2,$C7,$EC,$39,$3E,$1E,$7D,$71,$F6,$FE,$4B
	.dc.b	$B7,$DF,$8D,$BF,$F5,$85,$FF,$1A,$F9,$87,$E4,$AF,$DB,$DF,$AE,$BF
	.dc.b	$4B,$F0,$FE,$F8,$FD,$2B,$F9,$36,$1C,$2F,$75,$AF,$CF,$73,$8F,$3E
	.dc.b	$7F,$8C,$7F,$89,$D7,$C3,$7E,$93,$0E,$59,$F7,$E3,$FE,$1B,$AF,$96
	.dc.b	$FC,$93,$9F,$7F,$F7,$37,$C3,$3F,$2F,$E7,$7E,$0D,$7C,$79,$37,$FF
	.dc.b	$24,$E7,$CE,$73,$0F,$9F,$E1,$9F,$86,$1E,$8F,$F4,$3F,$D9,$DF,$AC
	.dc.b	$FF,$4E,$FC,$2D,$E9,$3F,$01,$FE,$71,$F9,$EE,$EB,$53,$9F,$8A,$FF
	.dc.b	$6F,$ED,$FD,$23,$1E,$7E,$5F,$F0,$47,$AF,$9B,$70,$E3,$97,$F3,$4E
	.dc.b	$8E,$93,$FE,$A5,$CB,$C1,$F2,$DF,$DB,$3F,$6C,$FF,$43,$7D,$1F,$F1
	.dc.b	$7F,$C1,$9F,$95,$3F,$4F,$7E,$20,$FC,$2B,$D7,$C2,$C7,$8F,$C0,$B1
	.dc.b	$E3,$87,$0B,$FF,$61,$5E,$F6,$FE,$8B,$9F,$F2,$5D,$FF,$A6,$65,$CB
	.dc.b	$1B,$5A,$CB,$A6,$9E,$3D,$3C,$53,$FD,$F6,$F8,$4E,$7F,$4B,$F6,$4F
	.dc.b	$F5,$2F,$CE,$9F,$9D,$3F,$3A,$FD,$37,$E3,$8F,$CA,$FE,$BF,$07,$97
	.dc.b	$06,$1C,$2F,$74,$F8,$9F,$05,$75,$C5,$F9,$07,$A2,$38,$C7,$77,$FB
	.dc.b	$5E,$BF,$EB,$3C,$7A,$79,$8F,$49,$AF,$FA,$0E,$FA,$9C,$B4,$F3,$FD
	.dc.b	$A7,$F8,$8B,$EA,$3F,$D7,$9F,$96,$F8,$F7,$7E,$5C,$FC,$BB,$9F,$E5
	.dc.b	$FF,$C7,$EF,$7C,$5E,$FA,$D6,$8A,$FE,$41,$C6,$BD,$FD,$DF,$F7,$F7
	.dc.b	$8E,$57,$6F,$EF,$BA,$78,$72,$E7,$8F,$96,$73,$9F,$7D,$3F,$BC,$C6
	.dc.b	$47,$D5,$BE,$4B,$FE,$00,$2F,$C9,$FF,$C7,$1A,$7F,$B6,$63,$5F,$8F
	.dc.b	$3F,$20,$78,$6F,$C7,$F9,$FE,$41,$C3,$3F,$F0,$8E,$1C,$1F,$F0,$DE
	.dc.b	$3C,$F3,$FE,$77,$18,$36,$F8,$9E,$47,$99,$65,$E1,$C7,$C1,$F6,$5F
	.dc.b	$E5,$4F,$8A,$FF,$77,$79,$9F,$E3,$2F,$D2,$7F,$A0,$FE,$9A,$FC,$B1
	.dc.b	$FA,$EF,$FA,$6F,$AB,$FB,$66,$BF,$E2,$6C,$38,$7C,$1B,$0E,$17,$FE
	.dc.b	$9B,$CB,$AF,$D3,$3C,$CF,$F1,$94,$63,$4B,$97,$3A,$47,$8B,$4F,$9F
	.dc.b	$3F,$E6,$3A,$E1,$97,$79,$6B,$EB,$AE,$57,$C7,$2F,$F7,$DE,$BE,$89
	.dc.b	$FE,$DB,$97,$6C,$79,$53,$F3,$4E,$D5,$FE,$42,$F4,$7F,$C9,$3F,$0B
	.dc.b	$FE,$27,$FC,$F9,$F9,$B3,$F6,$8F,$F4,$C6,$9D,$6F,$7F,$3C,$BD,$D3
	.dc.b	$F5,$CE,$B9,$7F,$9E,$3E,$BD,$F9,$E7,$C2,$8C,$69,$72,$E7,$8F,$83
	.dc.b	$F3,$BF,$75,$6B,$3E,$DA,$BB,$B3,$F1,$0F,$D4,$5F,$81,$7C,$CF,$CA
	.dc.b	$FF,$CC,$7D,$EF,$3B,$C7,$7B,$AD,$70,$FE,$23,$EA,$F3,$8B,$F0,$2F
	.dc.b	$35,$EF,$6F,$F3,$7F,$77,$F9,$A1,$47,$3E,$C6,$BF,$E7,$2E,$1F,$2E
	.dc.b	$3E,$49,$CE,$6F,$CF,$73,$AB,$F6,$4C,$FE,$E1,$E7,$EB,$FB,$5A,$79
	.dc.b	$E7,$3C,$1E,$6B,$E1,$AF,$FA,$03,$F0,$8E,$75,$DB,$93,$AF,$E6,$7E
	.dc.b	$0E,$5E,$9C,$7E,$9C,$FE,$14,$E7,$3E,$8F,$9B,$8B,$8E,$7F,$E8,$CD
	.dc.b	$FE,$9F,$63,$F1,$E7,$ED,$4F,$C1,$7F,$C3,$3E,$2B,$F3,$7F,$F1,$D6
	.dc.b	$DF,$1F,$0E,$17,$BF,$94,$FC,$DD,$7F,$1C,$AD,$62,$FC,$E7,$F0,$D8
	.dc.b	$70,$EB,$FA,$27,$FA,$61,$E7,$9F,$83,$F4,$1F,$2D,$F9,$23,$C7,$78
	.dc.b	$CD,$F9,$E7,$9B,$EE,$FC,$B9,$73,$C3,$D7,$39,$FA,$AF,$CC,$E7,$E8
	.dc.b	$8F,$6F,$C3,$F3,$7D,$FF,$1B,$78,$3F,$6B,$93,$F2,$DE,$39,$74,$9C
	.dc.b	$F8,$D8,$D7,$F5,$B1,$FB,$CF,$43,$FC,$F3,$E0,$7E,$E7,$5F,$8F,$E3
	.dc.b	$FC,$C9,$7D,$9F,$18,$F0,$F0,$DC,$3F,$82,$F0,$7B,$E7,$87,$E6,$19
	.dc.b	$E8,$FE,$3D,$C7,$E1,$FC,$3B,$E1,$4D,$C3,$D3,$36,$98,$7C,$9B,$EF
	.dc.b	$AF,$F0,$DF,$CD,$9F,$BB,$7F,$B7,$7F,$38,$FF,$26,$7A,$4F,$C0,$5F
	.dc.b	$8D,$7F,$02,$BD,$FE,$14,$BD,$D3,$F5,$1D,$63,$FE,$E7,$E2,$FB,$E3
	.dc.b	$0E,$FF,$A6,$FF,$09,$79,$E1,$E0,$FC,$0F,$F8,$3C,$63,$E9,$BF,$8C
	.dc.b	$7D,$96,$31,$FF,$79,$7C,$F7,$9B,$F8,$07,$CB,$C3,$EB,$9C,$DF,$F1
	.dc.b	$F9,$F2,$4E,$77,$FF,$21,$CF,$F2,$3C,$BA,$46,$38,$59,$63,$F3,$8F
	.dc.b	$27,$17,$B2,$91,$E7,$E1,$C3,$E9,$E1,$B9,$7F,$2F,$FC,$9F,$F9,$0F
	.dc.b	$23,$F2,$97,$E5,$8F,$CA,$DF,$95,$AF,$F6,$9F,$99,$FF,$30,$C3,$85
	.dc.b	$EE,$B5,$E2,$BF,$2D,$D6,$4F,$88,$FE,$69,$87,$0D,$7E,$8B,$CB,$4E
	.dc.b	$FA,$7F,$88,$79,$7F,$A7,$4E,$73,$F1,$4A,$D5,$FE,$E9,$7F,$D5,$3F
	.dc.b	$36,$DB,$FD,$69,$C3,$FD,$8B,$F5,$3F,$E1,$97,$BB,$7D,$F7,$F9,$0E
	.dc.b	$F7,$F0,$2B,$7E,$67,$DB,$F0,$CF,$F6,$C6,$7F,$E6,$BC,$7F,$9B,$19
	.dc.b	$3C,$38,$61,$F3,$65,$DF,$2E,$DF,$81,$FE,$4F,$37,$E4,$5A,$CB,$AE
	.dc.b	$3E,$2F,$96,$C6,$5F,$C2,$7C,$FF,$DD,$F9,$1F,$E0,$BF,$2F,$C1,$3F
	.dc.b	$D8,$FA,$7F,$93,$F9,$9F,$BB,$F2,$F3,$E7,$FF,$2B,$F5,$7C,$97,$8B
	.dc.b	$FE,$5A,$F9,$F8,$BE,$25,$7F,$04,$F3,$BF,$C0,$7E,$A7,$8A,$BB,$6B
	.dc.b	$E2,$F3,$F1,$CF,$FE,$DE,$FA,$8F,$F4,$27,$DE,$FE,$F3,$EB,$1F,$A0
	.dc.b	$79,$AB,$59,$7F,$B0,$70,$FF,$43,$7C,$EE,$7E,$57,$FD,$1F,$A7,$8F
	.dc.b	$A3,$FB,$16,$1F,$EE,$AE,$6F,$4F,$0F,$D9,$8E,$7A,$FF,$9F,$2F,$C7
	.dc.b	$87,$FA,$63,$3E,$DF,$41,$E9,$FB,$4F,$F2,$3F,$E9,$2F,$8C,$FE,$F7
	.dc.b	$F9,$2B,$E2,$3F,$CC,$5A,$FE,$16,$F7,$5A,$9F,$F7,$F4,$E7,$E7,$78
	.dc.b	$69,$E7,$3F,$29,$7D,$9C,$07,$9F,$C7,$B1,$FC,$CC,$FD,$A9,$CC,$7F
	.dc.b	$82,$E3,$EE,$FD,$43,$ED,$9F,$83,$FE,$23,$ED,$1F,$96,$BF,$1F,$7D
	.dc.b	$F7,$0B,$DF,$91,$FE,$B9,$AD,$7E,$07,$9C,$7E,$5F,$57,$F8,$3F,$3E
	.dc.b	$DE,$5E,$35,$FD,$74,$78,$6B,$C1,$87,$F9,$2B,$8B,$F5,$53,$9C,$BB
	.dc.b	$75,$FE,$1F,$F4,$1F,$F5,$4F,$EC,$4F,$C0,$DF,$9D,$BD,$E7,$EF,$0F
	.dc.b	$C8,$3F,$37,$87,$0B,$DD,$6B,$E7,$F3,$F7,$0C,$BC,$DC,$2C,$3F,$D6
	.dc.b	$FF,$9F,$E1,$E5,$A7,$7F,$96,$F2,$61,$C7,$3E,$BE,$47,$E7,$7E,$23
	.dc.b	$FC,$CF,$F8,$2F,$F8,$EF,$F4,$97,$ED,$FF,$EA,$3E,$2F,$D7,$D6,$BE
	.dc.b	$9D,$5A,$FB,$27,$EE,$1A,$F7,$5F,$A9,$79,$B0,$EB,$C7,$F8,$47,$96
	.dc.b	$1D,$BF,$19,$F1,$5B,$5F,$16,$BF,$E4,$3E,$CF,$E3,$1F,$BF,$BF,$1E
	.dc.b	$FE,$C5,$FA,$83,$E7,$5F,$A7,$3F,$68,$FE,$79,$87,$0B,$DD,$6A,$7F
	.dc.b	$D5,$53,$9F,$9D,$F9,$78,$F5,$2D,$F0,$C3,$A7,$07,$F1,$09,$F1,$53
	.dc.b	$2F,$15,$7F,$A5,$6B,$F8,$AF,$D1,$7C,$4F,$DB,$3F,$D9,$DF,$89,$7F
	.dc.b	$BB,$3E,$CD,$F9,$96,$B9,$D5,$AE,$DF,$F4,$07,$E6,$39,$FC,$67,$EF
	.dc.b	$E5,$CF,$3F,$D6,$BD,$7D,$B6,$9F,$EF,$8F,$27,$F8,$C4,$E6,$3F,$DF
	.dc.b	$7F,$8D,$FF,$1B,$F9,$E7,$E9,$6F,$C5,$7F,$B6,$E2,$FD,$3D,$F9,$B3
	.dc.b	$0F,$E5,$F8,$70,$BD,$D6,$BF,$36,$CE,$2F,$C6,$BF,$2B,$7B,$DB,$F8
	.dc.b	$1B,$C4,$FF,$78,$5F,$F9,$4E,$38,$E1,$D3,$CA,$FE,$61,$C5,$FE,$8D
	.dc.b	$F2,$1F,$AE,$BF,$1F,$7E,$26,$FD,$CB,$F8,$77,$83,$FD,$77,$0E,$17
	.dc.b	$BA,$D7,$CF,$7F,$48,$9E,$9A,$7F,$16,$7F,$D3,$5F,$93,$FE,$B1,$E3
	.dc.b	$97,$6F,$D0,$BC,$7E,$7F,$EE,$32,$E9,$E5,$FF,$45,$FD,$91,$F2,$AF
	.dc.b	$D1,$9F,$B5,$7A,$3E,$33,$FD,$B1,$FB,$0F,$F5,$8B,$DD,$6B,$1F,$E6
	.dc.b	$5F,$86,$BF,$80,$35,$93,$F4,$7F,$DF,$E1,$C3,$8B,$F9,$6B,$CB,$4E
	.dc.b	$D8,$78,$EF,$C5,$1E,$1F,$91,$FE,$B5,$E0,$3F,$56,$7E,$35,$FD,$FF
	.dc.b	$E3,$3F,$B9,$E5,$FA,$AE,$AF,$F0,$35,$EE,$B5,$F0,$CA,$7E,$5F,$FD
	.dc.b	$CA,$B5,$B7,$F4,$07,$B7,$E3,$9F,$C5,$9E,$5F,$55,$FC,$EF,$0F,$C6
	.dc.b	$57,$1D,$3C,$73,$9F,$13,$FB,$87,$8E,$FA,$DF,$F6,$8F,$C0,$BE,$5F
	.dc.b	$F2,$BF,$A9,$FF,$82,$B8,$7F,$C3,$97,$BF,$98,$5E,$FC,$D5,$BD,$3F
	.dc.b	$F0,$4D,$EF,$6F,$EA,$EF,$7F,$C3,$FF,$96,$BC,$F0,$E5,$E9,$F9,$35
	.dc.b	$F1,$EB,$FA,$6F,$AA,$FF,$03,$7D,$A7,$F2,$0F,$C4,$7F,$2D,$E1,$3F
	.dc.b	$6F,$7C,$DB,$F7,$4F,$F0,$CD,$EF,$CB,$FC,$45,$6B,$F6,$9C,$FC,$63
	.dc.b	$2F,$E7,$F7,$FC,$B3,$F1,$5B,$39,$C3,$8F,$C1,$F8,$A3,$1A,$78,$7E
	.dc.b	$B7,$FC,$35,$F0,$6F,$C5,$9F,$B6,$BF,$10,$FF,$5C,$FE,$25,$FE,$49
	.dc.b	$F6,$BF,$C5,$B7,$BA,$D7,$C1,$AB,$FC,$9D,$F8,$B6,$B1,$7E,$7D,$FD
	.dc.b	$53,$0E,$1D,$DF,$E0,$27,$96,$7D,$BE,$5B,$C7,$7E,$28,$F0,$FB,$5F
	.dc.b	$E2,$FF,$B0,$FC,$6F,$A0,$FD,$4D,$F9,$BB,$F5,$0F,$D6,$FE,$BD,$FE
	.dc.b	$42,$BD,$D6,$A7,$F6,$3F,$BC,$67,$D3,$FF,$05,$6F,$FE,$14,$BF,$CC
	.dc.b	$FE,$E3,$FD,$A9,$E3,$97,$5F,$5F,$C5,$7F,$E1,$1C,$BA,$7E,$37,$FE
	.dc.b	$43,$FC,$B5,$FB,$C3,$F2,$CF,$F0,$E7,$E0,$CD,$3F,$DC,$7F,$96,$FF
	.dc.b	$95,$AF,$75,$A9,$FF,$73,$FD,$1E,$70,$FD,$66,$FF,$E2,$8B,$DB,$63
	.dc.b	$B7,$D9,$FF,$A5,$6F,$FD,$47,$3C,$70,$E9,$F4,$9F,$E5,$7F,$B3,$7D
	.dc.b	$EB,$F7,$67,$E8,$4F,$CC,$9F,$A9,$7F,$A8,$7A,$FF,$CD,$78,$70,$BD
	.dc.b	$D6,$BF,$0D,$F6,$7B,$FF,$E8,$9A,$EB,$FB,$67,$F8,$49,$E1,$B7,$F9
	.dc.b	$01,$E5,$9F,$7C,$DF,$A0,$F9,$75,$F2,$4E,$7C,$EF,$F4,$3F,$E7,$7F
	.dc.b	$E9,$DC,$5F,$B5,$79,$BF,$13,$7E,$BA,$FE,$01,$7E,$C4,$FD,$2D,$FA
	.dc.b	$8F,$FA,$93,$2E,$58,$F1,$C3,$85,$EF,$F0,$7D,$67,$FE,$AB,$AE,$CF
	.dc.b	$FA,$0E,$F8,$F0,$1F,$F3,$89,$8D,$3C,$7A,$78,$BB,$BD,$37,$CE,$D5
	.dc.b	$AC,$9F,$83,$EA,$BF,$CC,$5F,$08,$FD,$19,$F2,$4F,$DB,$1F,$9C,$FF
	.dc.b	$43,$FD,$CB,$CF,$87,$0B,$DD,$6B,$E7,$3F,$B2,$76,$79,$EB,$FD,$8D
	.dc.b	$E2,$33,$7E,$31,$EA,$9C,$C7,$85,$F8,$06,$31,$EB,$9F,$F9,$DB,$0E
	.dc.b	$93,$E1,$D3,$CD,$3F,$E2,$49,$CF,$E7,$AF,$DC,$DE,$63,$E1,$FF,$DA
	.dc.b	$FF,$A9,$BF,$76,$FF,$06,$E1,$C2,$F7,$5A,$97,$EF,$5F,$03,$1E,$9F
	.dc.b	$A2,$7A,$75,$47,$7C,$F0,$DC,$3B,$FD,$77,$CF,$8C,$66,$D1,$D5,$F4
	.dc.b	$FF,$49,$FA,$5F,$0E,$FD,$6D,$FA,$12,$5F,$A2,$7D,$5C,$7E,$1C,$70
	.dc.b	$AD,$BA,$FB,$6F,$AE,$F4,$DF,$E6,$9C,$F9,$9F,$AE,$FA,$E7,$96,$5D
	.dc.b	$FF,$02,$F2,$53,$4F,$24,$F8,$FE,$AB,$F1,$5F,$9D,$3F,$7C,$FF,$5D
	.dc.b	$7E,$15,$FE,$72,$F4,$7F,$80,$FE,$4B,$F8,$8A,$D7,$D9,$7F,$A4,$BF
	.dc.b	$01,$FD,$9F,$A9,$FE,$AA,$E4,$FB,$B5,$AE,$2F,$3B,$D6,$3E,$83,$CB
	.dc.b	$2E,$DB,$78,$B4,$F0,$E3,$FB,$EE,$9F,$CF,$7E,$F3,$FE,$B9,$FC,$3D
	.dc.b	$F9,$B7,$F7,$2F,$CA,$1F,$A8,$3F,$2F,$FC,$ED,$EE,$B5,$3E,$1F,$4F
	.dc.b	$F8,$24,$FF,$B0,$FF,$13,$FC,$13,$D7,$87,$E6,$BC,$7E,$B1,$97,$F9
	.dc.b	$CF,$8B,$F3,$33,$98,$FF,$71,$57,$EB,$3E,$25,$FA,$23,$F0,$CF,$EE
	.dc.b	$DF,$94,$7F,$A7,$3D,$F7,$E5,$7B,$AD,$7D,$42,$B5,$F9,$76,$7F,$10
	.dc.b	$FC,$B6,$5D,$32,$B8,$70,$BE,$99,$F3,$9E,$FC,$FF,$C9,$A3,$F7,$FF
	.dc.b	$27,$FE,$AC,$FC,$29,$FB,$BF,$F9,$83,$D6,$FF,$4F,$FE,$67,$AC,$3F
	.dc.b	$CE,$13,$9F,$C4,$9F,$BE,$67,$0F,$D4,$63,$FE,$B6,$FA,$8B,$87,$0F
	.dc.b	$C2,$FF,$88,$FA,$DF,$C5,$6B,$C1,$97,$77,$F8,$C7,$E6,$6F,$D1,$FF
	.dc.b	$CA,$9F,$8A,$B6,$7E,$F9,$FC,$FB,$F0,$8F,$E2,$F7,$B7,$FD,$3F,$3A
	.dc.b	$F3,$BF,$CC,$DD,$FF,$C3,$2B,$5C,$9F,$A6,$7B,$7B,$3F,$BF,$79,$65
	.dc.b	$DF,$2F,$8D,$7C,$73,$F1,$E5,$E2,$FC,$DF,$F9,$77,$E8,$CF,$C8,$3F
	.dc.b	$C5,$5F,$8D,$BE,$B5,$F9,$9F,$E4,$7F,$61,$FE,$5B,$7B,$FB,$AF,$F6
	.dc.b	$65,$6B,$CE,$7E,$67,$AC,$9F,$75,$FC,$4F,$0E,$1C,$9F,$BE,$79,$69
	.dc.b	$DF,$47,$AD,$E4,$C3,$C7,$9F,$8B,$AB,$FA,$27,$E8,$CF,$A3,$7D,$1B
	.dc.b	$F2,$97,$EB,$FF,$EE,$1E,$47,$C5,$FF,$44,$C3,$85,$EE,$B5,$FA,$36
	.dc.b	$71,$FF,$0A,$70,$FF,$12,$BD,$B6,$3B,$E5,$DB,$F2,$26,$7E,$49,$CE
	.dc.b	$3F,$EF,$1F,$DA,$3F,$AD,$7E,$2F,$FE,$0A,$FA,$67,$EC,$BF,$EC,$1F
	.dc.b	$D3,$1F,$AA,$FF,$83,$BC,$0F,$E9,$37,$BA,$D4,$BF,$36,$FE,$BB,$3C
	.dc.b	$F7,$FE,$4E,$FF,$85,$BE,$47,$FA,$03,$C7,$CE,$64,$FD,$2B,$C7,$F4
	.dc.b	$6D,$7C,$75,$AE,$BF,$EE,$1F,$6C,$FE,$01,$FF,$65,$FE,$23,$FE,$DF
	.dc.b	$F0,$BF,$B9,$7C,$DF,$F6,$EA,$D7,$E3,$BF,$E8,$7F,$C9,$7E,$FF,$CA
	.dc.b	$3F,$7D,$D6,$5F,$ED,$4E,$3F,$E7,$B8,$75,$FD,$AF,$FA,$B3,$CB,$4E
	.dc.b	$DD,$3E,$2B,$F0,$E5,$F8,$AF,$C6,$FF,$C0,$9E,$43,$F3,$E7,$E4,$DF
	.dc.b	$AF,$FC,$79,$F6,$FF,$E8,$9F,$C7,$7F,$C0,$77,$BA,$D4,$FF,$A3,$7A
	.dc.b	$B3,$83,$F3,$AF,$ED,$17,$BE,$1F,$EA,$6F,$1F,$57,$FD,$FD,$83,$F2
	.dc.b	$AF,$1E,$1E,$29,$CF,$6F,$F8,$67,$F3,$C7,$DB,$3F,$05,$FF,$0E,$F8
	.dc.b	$7F,$DB,$30,$7E,$7D,$FE,$18,$BD,$D6,$BE,$A3,$DD,$F9,$F6,$71,$E5
	.dc.b	$BF,$F6,$6B,$EB,$F5,$FF,$DF,$DE,$3F,$6B,$F8,$7F,$C1,$7F,$C8,$DE
	.dc.b	$67,$F8,$52,$73,$B7,$F7,$CE,$1F,$F1,$77,$E7,$1F,$9C,$EA,$FF,$58
	.dc.b	$FE,$72,$FC,$05,$F9,$EF,$F8,$B2,$B5,$F8,$E2,$B5,$F9,$EE,$7D,$AF
	.dc.b	$D9,$70,$FF,$7A,$BF,$DA,$FE,$79,$FE,$13,$78,$DB,$F1,$FF,$0D,$FE
	.dc.b	$E7,$4F,$0F,$1F,$F9,$07,$D7,$7E,$70,$FC,$E1,$F9,$83,$F3,$97,$E7
	.dc.b	$3F,$ED,$AF,$CD,$7F,$C8,$38,$70,$BD,$D6,$BF,$28,$7C,$C6,$B2,$7E
	.dc.b	$47,$FE,$27,$C3,$87,$1F,$F8,$49,$E5,$A7,$7D,$1F,$99,$F9,$70,$F2
	.dc.b	$4E,$6F,$FC,$B1,$F9,$17,$6F,$CE,$7F,$0D,$F9,$C3,$F3,$6F,$E5,$9F
	.dc.b	$2F,$FE,$56,$AD,$77,$2B,$5F,$1D,$FE,$F1,$F0,$BF,$C9,$75,$E8,$FC
	.dc.b	$E7,$FC,$78,$F5,$CF,$FC,$4E,$F2,$BE,$DC,$FE,$2E,$36,$BE,$2D,$5E
	.dc.b	$A7,$F9,$B5,$F9,$7D,$0F,$75,$F9,$4B,$E8,$1F,$43,$FA,$5E,$0F,$F2
	.dc.b	$EE,$1C,$2F,$75,$AF,$5F,$FC,$E3,$3D,$38,$3F,$C7,$6F,$D8,$FC,$E3
	.dc.b	$FC,$A8,$F1,$CB,$B7,$E4,$5E,$3E,$76,$BE,$39,$F1,$74,$FF,$A0,$F7
	.dc.b	$7E,$68,$FC,$D9,$EE,$7F,$42,$F0,$DF,$9B,$3E,$A3,$FD,$07,$87,$0B
	.dc.b	$DF,$8C,$BD,$FE,$45,$AC,$BA,$71,$7F,$94,$B0,$EF,$F5,$5F,$E6,$B7
	.dc.b	$96,$9D,$DF,$25,$78,$E3,$C5,$F9,$B7,$FA,$67,$C0,$FF,$7D,$F7,$BF
	.dc.b	$2C,$FF,$12,$F6,$7F,$A5,$A7,$31,$FE,$28,$FC,$B3,$1F,$95,$3F,$2C
	.dc.b	$FF,$3D,$52,$1F,$96,$E1,$BF,$96,$32,$ED,$F9,$6F,$86,$73,$1E,$1E
	.dc.b	$CF,$F5,$67,$E5,$3F,$EE,$EF,$C6,$7F,$DC,$D2,$F7,$7F,$D5,$33,$9F
	.dc.b	$B8,$FF,$7B,$4F,$FA,$0B,$C6,$FF,$4E,$CF,$FB,$F3,$B7,$FD,$1C,$FF
	.dc.b	$BC,$BA,$3F,$DA,$87,$2F,$CA,$7F,$DE,$78,$7F,$BF,$F9,$7C,$31,$8F
	.dc.b	$CA,$7F,$BD,$53,$1F,$F6,$1B,$FE,$C5,$FC,$83,$FD,$89,$F9,$2F,$FB
	.dc.b	$62,$73,$3F,$EE,$0F,$9D,$FF,$64,$FE,$49,$FE,$D8,$AF,$E5,$F5,$AF
	.dc.b	$C9,$3F,$DA,$F7,$E4,$8C,$5F,$FA,$DA,$3F,$92,$E1,$FE,$D7,$C3,$FD
	.dc.b	$C3,$F1,$29,$F2,$4E,$7F,$24,$FF,$75,$FC,$46,$71,$F1,$FE,$37,$FE
	.dc.b	$A8,$FB,$8F,$F5,$C7,$E6,$BC,$31,$E5,$39,$F9,$FF,$F5,$BD,$F0,$FC
	.dc.b	$7F,$FC,$2B,$5F,$20,$E0,$7F,$AE,$1F,$F6,$2B,$98,$E1,$F6,$1F,$EC
	.dc.b	$77,$FD,$6D,$F8,$87,$8F,$5F,$45,$71,$AD,$67,$FE,$BA,$8E,$BF,$49
	.dc.b	$FE,$95,$F6,$BF,$D3,$BF,$8B,$63,$F1,$2F,$F4,$AF,$DF,$3F,$15,$FF
	.dc.b	$50,$67,$D3,$0F,$F5,$06,$3F,$2E,$31,$7F,$66,$3C,$D1,$8F,$C5,$BF
	.dc.b	$D5,$58,$7F,$AD,$67,$B4,$63,$5F,$44,$7C,$FE,$DF,$F4,$17,$E1,$5F
	.dc.b	$E8,$7D,$5F,$86,$7F,$A3,$38,$7F,$03,$5A,$FC,$23,$FD,$17,$18,$FC
	.dc.b	$25,$F8,$57,$CB,$CB,$E5,$DF,$FA,$35,$F2,$9D,$E7,$FD,$1D,$8F,$68
	.dc.b	$DC,$63,$5F,$2C,$3F,$05,$FF,$36,$FE,$09,$FE,$75,$FC,$1B,$EB,$C6
	.dc.b	$3F,$06,$5F,$F9,$E3,$8B,$E9,$CF,$F9,$E3,$0F,$90,$7E,$35,$FF,$3C
	.dc.b	$BE,$43,$B7,$D2,$FF,$9D,$AF,$FC,$F3,$5C,$A3,$1F,$75,$E4,$FC,$5D
	.dc.b	$FA,$AF,$F8,$8F,$E9,$5F,$3E,$F3,$5E,$27,$F9,$8B,$97,$F5,$38,$70
	.dc.b	$FA,$7F,$F0,$A5,$EE,$9E,$0F,$E4,$AB,$AF,$E5,$3F,$2F,$0E,$39,$FF
	.dc.b	$8C,$78,$3D,$15,$CF,$4F,$0E,$9E,$09,$E5,$5E,$5C,$FC,$9E,$47,$CC
	.dc.b	$FC,$03,$FC,$E7,$F8,$07,$F9,$AF,$83,$CD,$C5,$F9,$A9,$CF,$DF,$FF
	.dc.b	$9D,$B6,$FE,$03,$1F,$E8,$1F,$43,$C9,$5B,$F9,$BF,$F4,$1E,$BF,$2E
	.dc.b	$B8,$FE,$01,$FE,$76,$8F,$D6,$C6,$3E,$D3,$EE,$FD,$B7,$F9,$47,$ED
	.dc.b	$3F,$CB,$3E,$77,$EC,$A3,$1F,$74,$FB,$AF,$F2,$E6,$DF,$C4,$A7,$FC
	.dc.b	$C3,$97,$D9,$7F,$CB,$8D,$FB,$7F,$F2,$CD,$B2,$EF,$18,$F8,$EF,$C6
	.dc.b	$FA,$0F,$C8,$FF,$DD,$3F,$57,$FE,$F4,$FA,$3F,$EB,$BF,$A0,$7F,$04
	.dc.b	$AD,$7E,$22,$AD,$7C,$7B,$EC,$B5,$8B,$E2,$7E,$2E,$FF,$C6,$F1,$E2
	.dc.b	$FE,$3D,$E7,$87,$7D,$3C,$77,$E2,$FC,$4B,$FC,$63,$F5,$BF,$A2,$FA
	.dc.b	$FF,$F1,$CF,$D9,$7F,$91,$39,$7F,$99,$C6,$3E,$C7,$F8,$CB,$FE,$A6
	.dc.b	$D3,$F3,$8F,$F9,$09,$BF,$35,$FE,$3F,$F1,$3F,$C9,$58,$F5,$8C,$7D
	.dc.b	$8B,$EA,$3F,$63,$F1,$BF,$E1,$DF,$A3,$FF,$14,$E9,$FD,$0A,$31,$F2
	.dc.b	$BF,$E2,$7F,$13,$F9,$0C,$7F,$89,$F4,$FD,$63,$FE,$27,$7F,$62,$7F
	.dc.b	$67,$F6,$7F,$C4,$BF,$57,$FE,$31,$C7,$C1,$3F,$CC,$AB,$C1,$F3,$9F
	.dc.b	$C0,$8F,$F0,$0E,$BF,$E0,$9F,$91,$FF,$07,$69,$FD,$52,$73,$CD,$FE
	.dc.b	$0C,$8C,$7C,$E7,$F8,$33,$1F,$0D,$7F,$83,$1F,$E0,$0F,$03,$B4,$7F
	.dc.b	$82,$F0,$FF,$0B,$7C,$97,$F1,$2F,$47,$FB,$A7,$99,$FD,$FF,$E4,$FF
	.dc.b	$A3,$C6,$35,$FE,$ED,$C3,$FD,$0A,$3F,$BB,$67,$FC,$35,$FE,$EC,$FF
	.dc.b	$60,$3F,$AF,$3D,$30,$ED,$F2,$7F,$DD,$7C,$EF,$EB,$DF,$8E,$7F,$4C
	.dc.b	$F8,$EF,$EB,$5F,$1F,$FD,$DE,$31,$F1,$FF,$C2,$3F,$22,$FE,$F5,$F2
	.dc.b	$0F,$47,$FA,$FD,$6B,$F0,$8F,$E1,$4F,$0F,$7A,$EF,$FD,$9B,$1E,$5C
	.dc.b	$FF,$D7,$23,$B1,$99,$FE,$B7,$ED,$79,$23,$F9,$D7,$E2,$6F,$8D,$FE
	.dc.b	$91,$F1,$BF,$D1,$F2,$FE,$C7,$F1,$7F,$D1,$FC,$1F,$E9,$7F,$8A,$74
	.dc.b	$C5,$2F,$8C,$FE,$93,$1F,$DB,$EF,$8C,$32,$F3,$CF,$08,$FE,$93,$8F
	.dc.b	$8E,$31,$3E,$3F,$8E,$FF,$47,$FC,$5E,$7D,$EF,$E2,$7F,$03,$FC,$CF
	.dc.b	$2F,$E3,$1E,$0B,$E0,$BF,$91,$7E,$15,$E5,$BF,$E3,$D9,$F8,$CF,$F0
	.dc.b	$A2,$EC,$FE,$3F,$1F,$E1,$DC,$39,$DF,$F2,$B8,$E5,$2C,$3B,$7C,$1B
	.dc.b	$E0,$FF,$8A,$57,$C3,$F5,$FF,$81,$FA,$BF,$EB,$0F,$69,$EE,$67,$D8
	.dc.b	$FE,$09,$EE,$FF,$9C,$3D,$CC,$7B,$9F,$13,$1F,$E1,$B9,$78,$CF,$E2
	.dc.b	$17,$C3,$D0,$F8,$D3,$DB,$0F,$F6,$15,$7D,$3F,$33,$F7,$7F,$65,$7B
	.dc.b	$BF,$6B,$F6,$3F,$EA,$4F,$C0,$B1,$F6,$7F,$4A,$FF,$79,$87,$FB,$7B
	.dc.b	$9F,$F8,$0E,$1C,$3D,$6F,$DD,$F8,$C9,$CF,$D9,$FF,$A8,$FD,$6F,$B1
	.dc.b	$E9,$7E,$A7,$D4,$FC,$23,$FA,$BF,$26,$FA,$5F,$E8,$CD,$3E,$91,$FD
	.dc.b	$B5,$7E,$59,$F8,$4A,$FD,$54,$F0,$87,$D0,$7E,$3B,$D7,$FF,$2F,$EB
	.dc.b	$FE,$66,$F2,$FF,$29,$E5,$7F,$87,$70,$FC,$7F,$95,$F8,$9E,$2F,$CB
	.dc.b	$DE,$FC,$BF,$CA,$D7,$F9,$6C,$D9,$FE,$45,$38,$F9,$BF,$A2,$7F,$C6
	.dc.b	$87,$E9,$9F,$F1,$67,$97,$FE,$59,$8F,$F2,$3F,$67,$D3,$F0,$BE,$FE
	.dc.b	$1F,$E7,$A8,$C7,$83,$CA,$91,$FD,$5F,$C1,$FF,$02,$47,$2F,$05,$F0
	.dc.b	$98,$8F,$C0,$BF,$D5,$FC,$4F,$F7,$DF,$87,$E1,$9C,$FC,$27,$F5,$FF
	.dc.b	$80,$C4,$3F,$4C,$FF,$77,$FE,$38,$FB,$E7,$E8,$CF,$71,$F9,$63,$F3
	.dc.b	$9F,$36,$1C,$2F,$75,$AF,$CE,$B3,$F6,$8F,$C1,$FC,$DA,$77,$86,$98
	.dc.b	$C7,$96,$9D,$FE,$8F,$CB,$18,$CB,$FC,$EB,$CF,$FE,$A0,$FC,$E5,$F8
	.dc.b	$43,$F3,$7F,$E1,$BF,$A2,$3E,$FB,$F5,$7D,$1F,$2E,$F7,$5A,$9F,$D7
	.dc.b	$FC,$06,$70,$7E,$8B,$E6,$BD,$F0,$FD,$F3,$FD,$2C,$F2,$F1,$99,$F7
	.dc.b	$FD,$3F,$E5,$CE,$63,$C5,$97,$CE,$F2,$FF,$41,$B7,$FA,$E1,$FD,$03
	.dc.b	$7C,$CF,$D0,$65,$F8,$F3,$FA,$3B,$FA,$07,$C4,$7E,$8F,$CC,$FC,$FE
	.dc.b	$7F,$EB,$F8,$FC,$45,$FF,$58,$F7,$7F,$D5,$9F,$52,$FA,$FF,$F4,$BF
	.dc.b	$D5,$FF,$9F,$BF,$5A,$78,$8F,$CB,$3F,$D6,$5F,$85,$7F,$BF,$30,$E1
	.dc.b	$9F,$FA,$D7,$0E,$17,$B8,$57,$ED,$2B,$70,$AF,$F4,$5F,$99,$FE,$85
	.dc.b	$9C,$C7,$86,$7F,$D6,$51,$8F,$45,$97,$93,$2F,$F6,$27,$DA,$7A,$B4
	.dc.b	$F4,$D6,$B4,$F1,$79,$79,$CB,$CD,$39,$C3,$D7,$EB,$B3,$F2,$F9,$BF
	.dc.b	$B1,$F8,$1C,$67,$FA,$BE,$FF,$D4,$7C,$0B,$D1,$7E,$93,$AD,$7F,$96
	.dc.b	$ED,$EB,$C3,$84,$7A,$9E,$5E,$7F,$F8,$F6,$38,$1F,$BA,$F4,$7F,$9D
	.dc.b	$FA,$1E,$78,$C6,$7E,$0C,$BF,$C9,$9B,$79,$AF,$FA,$0C,$7E,$3A,$3E
	.dc.b	$6F,$EA,$BF,$B7,$FD,$D9,$F0,$2F,$8B,$FF,$29,$7C,$E3,$E8,$FC,$37
	.dc.b	$BA,$D7,$07,$FA,$63,$F5,$CC,$F9,$DF,$E9,$AD,$FD,$47,$FB,$CE,$FB
	.dc.b	$38,$0F,$3C,$BF,$CC,$1D,$3E,$B8,$C6,$1F,$ED,$1D,$5F,$55,$FE,$E6
	.dc.b	$F9,$17,$EB,$6F,$C1,$DF,$AD,$BF,$3A,$7C,$B3,$F0,$E7,$EF,$FF,$67
	.dc.b	$2E,$58,$F1,$C3,$85,$EE,$35,$C0,$3F,$D5,$AF,$77,$F1,$F5,$CB,$F0
	.dc.b	$0F,$CA,$3D,$33,$F1,$E7,$E2,$E1,$FD,$4C,$E7,$2E,$FE,$46,$31,$ED
	.dc.b	$B7,$B3,$1E,$3A,$7A,$7F,$65,$F1,$D7,$DE,$EE,$FF,$44,$F8,$18,$F0
	.dc.b	$FF,$25,$C7,$F1,$AB,$F1,$A7,$98,$FE,$76,$53,$97,$8D,$FC,$F1,$E5
	.dc.b	$DF,$F8,$C6,$F8,$1F,$8B,$F8,$4B,$97,$FA,$63,$D3,$B8,$7F,$83,$FF
	.dc.b	$18,$F2,$FD,$63,$F6,$8F,$F0,$8F,$DB,$BE,$C9,$FA,$6B,$F4,$2F,$F4
	.dc.b	$37,$07,$9F,$76,$3C,$7C,$D6,$3C,$70,$E1,$F6,$DD,$B3,$FD,$5E,$1E
	.dc.b	$BC,$38,$36,$3C,$39,$F8,$3F,$44,$F9,$D3,$DA,$73,$83,$F4,$7C,$E1
	.dc.b	$E1,$CB,$C1,$AB,$F7,$3F,$E0,$1F,$79,$F8,$3C,$9E,$93,$F4,$3F,$C6
	.dc.b	$FE,$37,$F9,$0F,$C9,$7F,$88,$FD,$7F,$0B,$FC,$55,$EF,$C9,$5E,$FD
	.dc.b	$A7,$E5,$5B,$BE,$B9,$7E,$61,$FE,$3B,$9F,$87,$1F,$07,$07,$F2,$38
	.dc.b	$C5,$78,$BC,$26,$AF,$17,$F9,$CD,$6B,$5F,$0D,$6B,$CF,$CE,$7E,$6C
	.dc.b	$79,$63,$FC,$AB,$E8,$3C,$9A,$7A,$77,$F4,$7E,$DD,$E7,$F8,$77,$EB
	.dc.b	$4F,$7F,$ED,$FD,$D3,$F1,$75,$3F,$53,$E1,$7B,$EB,$5E,$E9,$FB,$D6
	.dc.b	$BE,$17,$DD,$C3,$FD,$A3,$0E,$CF,$C4,$31,$9F,$83,$37,$C1,$FB,$27
	.dc.b	$C3,$5D,$F2,$F2,$75,$E7,$3F,$4E,$3C,$5F,$93,$E2,$BA,$3F,$01,$1A
	.dc.b	$DB,$CF,$A7,$B3,$F8,$07,$D7,$87,$1C,$2D,$6A,$DF,$19,$FD,$27,$A7
	.dc.b	$F1,$5F,$B4,$FF,$BA,$BF,$4F,$7E,$44,$FC,$77,$FC,$6F,$F6,$3B,$E3
	.dc.b	$FE,$25,$7B,$AD,$7A,$8F,$CB,$B5,$4A,$7E,$06,$FD,$9B,$FB,$46,$9E
	.dc.b	$08,$FF,$3C,$D7,$E6,$CE,$35,$D7,$37,$CB,$7F,$80,$2F,$87,$C6,$B6
	.dc.b	$F3,$D7,$F2,$3E,$2F,$E5,$F8,$F9,$B2,$5E,$F6,$F3,$FD,$BF,$69,$CE
	.dc.b	$3C,$B7,$7E,$87,$FD,$B7,$D0,$FD,$77,$EB,$EF,$D1,$9F,$8A,$3F,$01
	.dc.b	$FF,$1B,$FE,$A1,$FE,$58,$FC,$73,$87,$07,$F3,$7C,$38,$5E,$FE,$81
	.dc.b	$7F,$E5,$EF,$9A,$DD,$AF,$B7,$C8,$FF,$83,$74,$F0,$E9,$FE,$78,$AF
	.dc.b	$E0,$1C,$9F,$C7,$5E,$D1,$E4,$D7,$C7,$F2,$1F,$E2,$6B,$E3,$AF,$87
	.dc.b	$4F,$CB,$E3,$C7,$1F,$AB,$F5,$2F,$D7,$DF,$2E,$FC,$45,$FA,$93,$F4
	.dc.b	$67,$CC,$7A,$3A,$58,$F1,$C3,$85,$EF,$E6,$B5,$3D,$73,$FF,$12,$EF
	.dc.b	$FC,$67,$0E,$3F,$96,$7B,$E7,$FA,$F4,$74,$D7,$C7,$AF,$8B,$F7,$1F
	.dc.b	$65,$6B,$1F,$04,$F9,$BE,$17,$33,$EC,$BE,$3A,$7A,$71,$F1,$C6,$37
	.dc.b	$FC,$F7,$03,$F0,$EF,$E5,$9F,$4E,$FD,$5D,$F5,$5F,$E6,$AF,$D9,$9F
	.dc.b	$A6,$7F,$B7,$3F,$17,$FF,$19,$61,$C2,$F7,$F5,$AB,$DF,$C8,$3F,$3B
	.dc.b	$DF,$0F,$F7,$BD,$3B,$7C,$BF,$F9,$2F,$2E,$5A,$DA,$D6,$7D,$32,$F0
	.dc.b	$F9,$DF,$E6,$2B,$DE,$9D,$F6,$FF,$43,$61,$FE,$45,$FB,$AF,$E8,$DF
	.dc.b	$26,$FD,$1B,$D9,$C7,$FD,$7B,$AB,$F4,$BD,$BB,$C3,$FC,$1D,$7B,$DB
	.dc.b	$F0,$1E,$C2,$FF,$57,$FB,$A6,$ED,$7B,$7F,$CC,$39,$F8,$33,$7E,$DF
	.dc.b	$FE,$67,$8C,$65,$D3,$1F,$F7,$CF,$42,$73,$AF,$97,$F1,$8F,$F3,$FF
	.dc.b	$EF,$7F,$D7,$FF,$30,$7E,$FA,$FD,$6D,$F8,$FF,$F5,$AF,$D4,$DF,$AD
	.dc.b	$F8,$34,$63,$C7,$0E,$17,$BA,$7E,$EB,$FE,$6B,$AD,$63,$FE,$7D,$78
	.dc.b	$7D,$ED,$E3,$FF,$19,$69,$E2,$AF,$0F,$C8,$B5,$7E,$99,$E8,$BD,$EB
	.dc.b	$E0,$AD,$7D,$5F,$FA,$6B,$AF,$F4,$5E,$33,$F1,$27,$EF,$DF,$E0,$3F
	.dc.b	$D9,$BF,$C2,$3F,$81,$7F,$9F,$EF,$75,$AF,$9F,$57,$F8,$5B,$F0,$9D
	.dc.b	$6B,$FA,$AC,$BA,$3F,$E9,$4C,$3B,$FE,$15,$FE,$99,$79,$E7,$DF,$EC
	.dc.b	$7C,$95,$C7,$3E,$D9,$7F,$32,$8F,$F4,$87,$E9,$BF,$E1,$1F,$98,$7D
	.dc.b	$93,$F2,$67,$E3,$3F,$0F,$D3,$7C,$27,$F8,$C7,$C0,$63,$C7,$0E,$17
	.dc.b	$BA,$F9,$55,$AC,$BF,$CC,$19,$76,$CB,$FD,$61,$8F,$1F,$DA,$2D,$6B
	.dc.b	$2E,$7D,$4D,$3C,$33,$E8,$35,$87,$9A,$75,$E9,$FF,$B3,$39,$7F,$C5
	.dc.b	$9D,$4F,$DE,$9E,$63,$F2,$1F,$F2,$2F,$8F,$FE,$1C,$FD,$1F,$FD,$65
	.dc.b	$87,$0B,$DD,$6B,$E9,$9F,$A0,$EB,$6F,$F3,$D6,$7D,$7F,$64,$FF,$65
	.dc.b	$63,$C7,$4E,$DE,$97,$F2,$FF,$9A,$F2,$61,$C3,$4E,$B3,$FC,$CF,$E3
	.dc.b	$FF,$DC,$7F,$A8,$7F,$90,$FD,$6F,$F6,$B7,$E2,$2F,$C8,$3F,$D1,$1F
	.dc.b	$98,$3F,$80,$5F,$89,$FF,$B2,$70,$E1,$F8,$73,$0E,$1A,$B0,$E1,$F8
	.dc.b	$D6,$E3,$FD,$25,$9F,$6C,$FA,$FC,$47,$8A,$73,$8F,$39,$F0,$FB,$5F
	.dc.b	$EF,$8F,$C9,$FF,$CB,$1F,$53,$F8,$2F,$D0,$5F,$A4,$BF,$50,$FF,$23
	.dc.b	$7E,$BF,$FA,$AF,$25,$87,$0F,$21,$87,$0B,$DF,$BB,$FE,$E3,$AD,$4E
	.dc.b	$7E,$E5,$F3,$1E,$1E,$0F,$F5,$56,$BE,$1D,$7C,$15,$FE,$71,$9F,$F7
	.dc.b	$67,$E2,$DF,$E6,$DF,$C1,$5F,$A4,$7F,$1E,$FD,$4B,$F5,$FF,$63,$FE
	.dc.b	$01,$FB,$AD,$E1,$FE,$B5,$BD,$FC,$62,$F7,$F3,$FC,$B9,$19,$F5,$CF
	.dc.b	$A7,$AF,$FE,$55,$F8,$FF,$F5,$F5,$6B,$3E,$95,$AD,$7E,$3C,$7F,$C0
	.dc.b	$1F,$A2,$FF,$41,$78,$8F,$DE,$BF,$9F,$7D,$53,$F6,$67,$98,$FD,$9B
	.dc.b	$F0,$3A,$7F,$83,$F2,$FF,$66,$63,$C7,$E4,$FF,$C4,$F8,$70,$B7,$8B
	.dc.b	$DE,$F7,$9F,$6A,$EB,$F8,$87,$A3,$5F,$1C,$E7,$4F,$E6,$B6,$D1,$F9
	.dc.b	$67,$FA,$67,$F6,$57,$F0,$07,$FA,$16,$BF,$DB,$DF,$8B,$FF,$2A,$FD
	.dc.b	$3F,$85,$EE,$B5,$5F,$EE,$BF,$8F,$7E,$75,$A8,$FF,$73,$63,$D3,$BF
	.dc.b	$FD,$4D,$97,$F3,$49,$D0,$9D,$3E,$0F,$8F,$FA,$3F,$97,$7F,$AB,$7E
	.dc.b	$15,$F9,$5B,$F7,$5F,$F5,$87,$A8,$FD,$AD,$F7,$BE,$DC,$38,$5E,$F2
	.dc.b	$FF,$7C,$5E,$FE,$F7,$51,$FE,$FB,$CB,$AE,$5F,$E6,$18,$F0,$99,$BF
	.dc.b	$E5,$AE,$79,$18,$F6,$AF,$1F,$CE,$70,$A6,$1D,$1F,$0F,$4F,$E3,$3F
	.dc.b	$3E,$FF,$64,$60,$FC,$65,$F9,$87,$FB,$2F,$F3,$27,$EE,$EF,$C4,$BE
	.dc.b	$C6,$1C,$2F,$79,$05,$F0,$FA,$9D,$C6,$72,$ED,$97,$B6,$32,$7A,$AD
	.dc.b	$F9,$9E,$78,$19,$77,$BF,$27,$4F,$A7,$0E,$98,$78,$9F,$0E,$DF,$A8
	.dc.b	$F6,$3F,$DC,$9F,$5C,$FC,$F3,$E6,$72,$3E,$93,$F2,$55,$A9,$CF,$D1
	.dc.b	$E3,$1F,$66,$1C,$5F,$25,$6F,$8A,$C6,$AF,$EC,$7D,$0F,$86,$30,$7F
	.dc.b	$A6,$FD,$AF,$DD,$FE,$B1,$FE,$F7,$FD,$CB,$F3,$33,$F2,$7F,$3E,$F9
	.dc.b	$DC,$8A,$D7,$17,$E0,$E7,$FD,$D1,$39,$AF,$A1,$7C,$9F,$F4,$1D,$F0
	.dc.b	$CF,$E5,$3C,$A3,$FD,$59,$DD,$FC,$33,$F3,$CF,$B1,$0F,$CF,$3E,$A2
	.dc.b	$FD,$B3,$F6,$D3,$A9,$CF,$EE,$18,$AF,$C0,$47,$07,$07,$33,$F0,$67
	.dc.b	$E5,$E7,$93,$F9,$DB,$FE,$3B,$FB,$4F,$8B,$E9,$7C,$31,$FC,$16,$31
	.dc.b	$3F,$87,$FA,$4C,$3F,$8F,$AF,$06,$CA,$DE,$BF,$89,$78,$D7,$C0,$1F
	.dc.b	$B8,$6D,$FC,$C7,$EC,$0F,$CC,$A1,$0F,$F4,$14,$EE,$53,$99,$DF,$E7
	.dc.b	$DD,$2F,$E5,$FD,$C7,$C3,$C3,$85,$F4,$C3,$F7,$91,$8C,$39,$7C,$07
	.dc.b	$EA,$61,$BF,$D4,$F8,$BE,$D5,$E5,$F9,$6F,$4C,$E6,$33,$EE,$63,$1F
	.dc.b	$A6,$7D,$AC,$39,$BF,$B5,$C3,$C4,$FF,$BB,$63,$2F,$F6,$CA,$F1,$BF
	.dc.b	$1B,$9B,$F2,$DE,$4F,$BF,$F1,$57,$E1,$59,$AF,$C8,$CE,$7E,$AF,$10
	.dc.b	$EB,$FC,$A3,$ED,$C3,$9E,$3F,$C0,$72,$FC,$D4,$EF,$0F,$0C,$7F,$80
	.dc.b	$63,$F6,$3F,$B0,$7C,$FF,$82,$7E,$BF,$F6,$BC,$4F,$DE,$CE,$7F,$57
	.dc.b	$C5,$FE,$C0,$F9,$F3,$AC,$BC,$0D,$7E,$FF,$B1,$F4,$63,$11,$F8,$3F
	.dc.b	$B4,$FA,$9E,$7F,$F2,$6F,$27,$F0,$9F,$97,$FF,$C0,$0F,$6F,$F1,$73
	.dc.b	$9B,$FC,$4C,$BF,$0B,$FF,$64,$70,$7F,$1A,$D3,$F8,$B9,$F0,$6F,$83
	.dc.b	$F8,$47,$95,$FE,$4E,$FF,$CA,$F5,$C6,$BF,$9F,$E3,$E2,$F9,$4F,$C4
	.dc.b	$7A,$BF,$D3,$BF,$01,$FC,$CF,$C9,$BF,$17,$7E,$D3,$FA,$0A,$D6,$9F
	.dc.b	$9F,$AD,$7E,$81,$9C,$FF,$A7,$7A,$3F,$56,$FA,$47,$E6,$1E,$36,$B7
	.dc.b	$DA,$F8,$AB,$FA,$B6,$7E,$2D,$3F,$39,$F1,$1F,$DD,$FA,$FF,$5B,$F9
	.dc.b	$1B,$F2,$8F,$F7,$17,$17,$F8,$0B,$1F,$EE,$FF,$15,$FD,$3F,$5F,$DA
	.dc.b	$5E,$FE,$EB,$F7,$39,$73,$9C,$FD,$27,$6C,$38,$E7,$DB,$D9,$FD,$67
	.dc.b	$C8,$7F,$84,$BC,$27,$ED,$6F,$C3,$BF,$C8,$7F,$92,$7F,$1F,$E0,$FF
	.dc.b	$11,$64,$F9,$6F,$F1,$15,$EF,$A1,$6F,$6F,$F8,$5E,$7D,$6B,$5C,$BE
	.dc.b	$0C,$7F,$C4,$BA,$78,$34,$FF,$28,$F3,$FE,$FF,$E6,$FF,$C8,$1F,$3D
	.dc.b	$FC,$33,$F2,$37,$EA,$EF,$DA,$DF,$B7,$FF,$18,$BD,$FB,$0B,$DD,$3F
	.dc.b	$7C,$D6,$BF,$E4,$6C,$BA,$65,$FE,$7B,$8C,$9F,$67,$E8,$3F,$DE,$F8
	.dc.b	$F3,$97,$E4,$7F,$C4,$FD,$7F,$F3,$4F,$E2,$0F,$D4,$5F,$8D,$BE,$C1
	.dc.b	$FB,$FF,$F2,$CB,$DD,$6A,$5F,$8A,$E7,$5F,$F3,$7F,$D7,$FF,$9E,$32
	.dc.b	$E9,$93,$D4,$F0,$56,$BE,$B9,$5A,$F5,$FF,$93,$7D,$AF,$FA,$5B,$E3
	.dc.b	$1F,$9F,$7C,$8F,$B1,$78,$3F,$E6,$FF,$19,$7B,$AD,$4F,$F8,$7E,$73
	.dc.b	$9F,$FA,$5B,$0E,$7C,$AC,$7F,$C8,$D5,$AA,$7D,$AF,$FA,$B3,$1F,$06
	.dc.b	$7F,$CE,$7E,$A3,$FD,$81,$F8,$DF,$F0,$1F,$C0,$DF,$B6,$7F,$97,$3F
	.dc.b	$24,$FE,$A3,$5A,$F7,$D5,$AF,$55,$F8,$CE,$B5,$FF,$5F,$65,$FC,$9F
	.dc.b	$0E,$17,$D3,$E8,$3C,$15,$AC,$79,$E6,$FD,$5B,$FA,$67,$57,$FB,$B7
	.dc.b	$F0,$DF,$B1,$ED,$7C,$CF,$CD,$1F,$5B,$FD,$92,$B5,$E0,$7E,$D7,$B1
	.dc.b	$4F,$C4,$B5,$AF,$FB,$B7,$F1,$1F,$E7,$58,$75,$8F,$EB,$AF,$2C,$BE
	.dc.b	$CF,$B5,$E2,$9F,$F7,$A6,$9E,$2C,$9F,$7D,$FD,$87,$DE,$FF,$09,$F3
	.dc.b	$7F,$A5,$FD,$37,$D6,$3E,$E3,$FC,$29,$6F,$A2,$FE,$F7,$87,$0F,$B6
	.dc.b	$FD,$E5,$EF,$E8,$FA,$FC,$FE,$B6,$FF,$2B,$69,$D5,$FE,$E7,$87,$1F
	.dc.b	$C9,$FF,$AA,$3C,$F5,$F0,$EB,$E0,$AE,$47,$D6,$CF,$BC,$7F,$89,$FF
	.dc.b	$54,$FE,$F5,$F8,$B7,$F8,$B3,$F1,$6F,$C8,$FB,$3F,$4F,$F8,$03,$FD
	.dc.b	$53,$F9,$43,$F5,$FF,$F0,$9D,$EF,$D0,$FE,$01,$E0,$2F,$7D,$7A,$DB
	.dc.b	$FC,$DD,$9F,$47,$FC,$17,$87,$1F,$DC,$BF,$B6,$3C,$F4,$F0,$7D,$0F
	.dc.b	$F1,$BA,$E4,$2F,$CB,$97,$79,$7E,$19,$FE,$13,$FC,$77,$FC,$8B,$EF
	.dc.b	$FF,$AD,$FF,$50,$7E,$9E,$FC,$9B,$FD,$B3,$FB,$77,$F1,$1F,$55,$87
	.dc.b	$0F,$AE,$61,$C2,$F7,$F1,$1F,$E3,$2A,$D7,$D5,$7F,$80,$DE,$15,$78
	.dc.b	$3F,$D1,$9A,$78,$74,$F0,$7C,$4F,$F9,$63,$4E,$F5,$AD,$3B,$D6,$BD
	.dc.b	$CF,$F1,$67,$D5,$FF,$97,$BF,$32,$74,$7F,$B3,$BF,$3D,$7E,$58,$FC
	.dc.b	$C1,$F9,$AB,$EA,$3F,$C8,$F8,$F1,$C3,$85,$EE,$BF,$C2,$7B,$2B,$5F
	.dc.b	$A8,$FF,$87,$5E,$1E,$F5,$E0,$FF,$53,$6B,$E2,$86,$BE,$2D,$7F,$98
	.dc.b	$D7,$80,$DE,$9E,$08,$7C,$C7,$F9,$1F,$C8,$FF,$3C,$FE,$51,$FC,$9B
	.dc.b	$F0,$B7,$E6,$1F,$CA,$3E,$93,$7E,$1F,$F9,$82,$F7,$EA,$AF,$6F,$FB
	.dc.b	$2B,$17,$EF,$5B,$D7,$FD,$79,$9F,$57,$FC,$B3,$87,$2F,$DF,$FF,$C7
	.dc.b	$4F,$4C,$79,$E7,$E6,$AE,$42,$FC,$DA,$79,$7E,$83,$D5,$3D,$B4,$F2
	.dc.b	$4F,$A7,$47,$E2,$DC,$B0,$E3,$5E,$7F,$8F,$EB,$F9,$CF,$F6,$6F,$CA
	.dc.b	$5F,$B1,$3F,$4F,$79,$6E,$87,$EC,$7C,$3C,$96,$1C,$2F,$75,$F8,$1E
	.dc.b	$EF,$F0,$37,$0F,$FB,$77,$3E,$CF,$C2,$31,$83,$F6,$0F,$87,$18,$C7
	.dc.b	$A6,$7F,$EE,$6C,$39,$CB,$0F,$46,$9E,$7A,$FE,$71,$3F,$E9,$6F,$6B
	.dc.b	$FD,$6B,$1F,$D4,$FF,$0A,$FE,$41,$E2,$3F,$1B,$FE,$7D,$F4,$7B,$C3
	.dc.b	$FD,$E5,$7B,$F8,$2F,$CD,$57,$F4,$3F,$CE,$75,$5F,$13,$E5,$FF,$E0
	.dc.b	$03,$FE,$94,$CF,$BC,$3F,$1C,$FF,$3B,$46,$34,$F0,$E1,$E6,$DD,$39
	.dc.b	$CB,$BF,$EE,$3F,$E9,$B8,$C6,$DE,$7F,$5D,$FA,$17,$8B,$7F,$34,$E7
	.dc.b	$E9,$7C,$BE,$F7,$F8,$DB,$F2,$6F,$AB,$E1,$BF,$46,$FF,$40,$FE,$2D
	.dc.b	$AF,$AD,$56,$A7,$FD,$0B,$FA,$C6,$63,$D3,$CC,$3F,$C0,$74,$F5,$F3
	.dc.b	$18,$9E,$5F,$53,$E9,$8E,$19,$F8,$FB,$5A,$78,$FD,$7D,$57,$19,$CE
	.dc.b	$9F,$E9,$6F,$C5,$BC,$51,$8F,$C0,$9F,$C0,$38,$C7,$D9,$0F,$C5,$71
	.dc.b	$87,$FA,$7B,$F3,$AF,$F6,$BF,$E1,$5F,$EF,$6F,$C0,$1F,$A1,$7F,$8D
	.dc.b	$BF,$29,$7B,$9F,$D9,$67,$F9,$CF,$80,$BE,$17,$BA,$D7,$2F,$4E,$1F
	.dc.b	$89,$87,$4F,$7F,$FD,$69,$87,$0E,$5B,$5A,$C7,$97,$C8,$FF,$A0,$23
	.dc.b	$91,$FE,$8F,$8C,$FD,$BF,$FB,$6E,$F9,$6B,$E6,$BF,$F4,$A7,$E1,$FC
	.dc.b	$33,$F5,$E4,$AD,$43,$F2,$EE,$BF,$2E,$FC,$AD,$F8,$DB,$F6,$17,$E2
	.dc.b	$BF,$EF,$CF,$CC,$7F,$80,$F2,$FC,$1E,$2F,$EF,$F8,$70,$BD,$CA,$BF
	.dc.b	$CD,$EE,$EB,$FC,$63,$5D,$9F,$F7,$BF,$07,$F8,$52,$3F,$85,$46,$0E
	.dc.b	$B8,$F8,$F1,$5B,$E6,$BE,$15,$6B,$2F,$0F,$D5,$E2,$3B,$E1,$C6,$FC
	.dc.b	$9F,$BC,$FC,$0C,$BD,$19,$7B,$7F,$2D,$F9,$FF,$A2,$BF,$08,$7E,$8F
	.dc.b	$FE,$35,$F4,$3E,$47,$E0,$FF,$EC,$1C,$FD,$97,$BA,$D7,$22,$BF,$D3
	.dc.b	$75,$AC,$39,$ED,$ED,$18,$77,$F8,$7F,$C1,$C6,$27,$C3,$F5,$9F,$C0
	.dc.b	$B1,$7E,$3B,$ED,$C3,$C1,$39,$FD,$1F,$17,$C3,$5F,$3F,$A3,$EC,$8C
	.dc.b	$63,$E5,$F6,$31,$BF,$A3,$77,$81,$8C,$FC,$11,$F0,$F8,$3E,$D7,$C8
	.dc.b	$BF,$35,$FF,$4C,$7E,$BC,$FD,$A5,$F8,$C3,$F2,$19,$87,$0B,$DD,$6A
	.dc.b	$5F,$83,$7E,$7B,$1E,$9F,$83,$7C,$27,$B7,$DC,$DC,$38,$E4,$FB,$FF
	.dc.b	$BD,$5C,$34,$F1,$4F,$FA,$7F,$A7,$A5,$F0,$CF,$FD,$59,$AF,$8F,$36
	.dc.b	$1C,$27,$FD,$37,$2F,$C5,$BF,$8E,$60,$FC,$EB,$FA,$67,$DB,$7F,$16
	.dc.b	$F6,$9F,$8C,$BF,$5A,$7E,$73,$F9,$6C,$78,$E1,$C3,$DA,$61,$C2,$FF
	.dc.b	$35,$9F,$7C,$FF,$95,$5F,$F5,$9F,$BC,$F0,$C7,$F8,$CF,$1E,$7F,$20
	.dc.b	$D7,$33,$FC,$C7,$0E,$4F,$9E,$75,$9F,$83,$5F,$E1,$95,$AC,$BC,$9A
	.dc.b	$7A,$BF,$40,$E5,$9F,$6C,$FD,$D8,$7B,$7C,$FF,$66,$9E,$6D,$3A,$E3
	.dc.b	$EB,$FC,$D3,$32,$F9,$0F,$F0,$27,$EF,$AF,$CA,$BE,$57,$D8,$3F,$4A
	.dc.b	$7D,$DF,$0B,$DD,$6A,$5F,$BE,$E7,$E6,$FF,$BE,$E5,$F0,$CF,$EA,$17
	.dc.b	$D9,$F8,$7C,$9F,$0A,$39,$F2,$7F,$9A,$FF,$37,$F6,$D7,$0B,$6B,$E5
	.dc.b	$AF,$F3,$87,$2E,$AB,$D5,$5A,$D3,$C9,$F9,$CF,$8F,$0E,$1B,$FA,$FE
	.dc.b	$E7,$E0,$FD,$A7,$E8,$BF,$4E,$7E,$80,$FD,$AD,$FA,$07,$FB,$B3,$B7
	.dc.b	$7E,$2A,$F7,$5A,$FC,$45,$FA,$CE,$A7,$EF,$F0,$F0,$F6,$EF,$3F,$E0
	.dc.b	$35,$F0,$6B,$DF,$57,$CE,$FA,$E7,$3A,$78,$BD,$8C,$6D,$E1,$C3,$86
	.dc.b	$CF,$DE,$7E,$DF,$C7,$3F,$5A,$7E,$31,$FC,$93,$D9,$FF,$0D,$7E,$12
	.dc.b	$FD,$37,$75,$AF,$6B,$FA,$77,$D2,$AB,$FA,$77,$1F,$E5,$B2,$E9,$97
	.dc.b	$AB,$CC,$FF,$45,$63,$F1,$B9,$BE,$DC,$62,$7C,$18,$FF,$7B,$9E,$F1
	.dc.b	$8F,$81,$FB,$18,$74,$CF,$C3,$9B,$F7,$EF,$CA,$FE,$A0,$FD,$5D,$F9
	.dc.b	$4B,$F2,$AF,$F1,$17,$C3,$3F,$5C,$7E,$B1,$C2,$F7,$5A,$97,$EB,$19
	.dc.b	$FB,$9F,$57,$0F,$F8,$8F,$3E,$D9,$E2,$11,$8C,$3B,$FB,$1F,$8F,$8C
	.dc.b	$7C,$F7,$F4,$4F,$3F,$1A,$78,$23,$FA,$3E,$8F,$CC,$3C,$99,$FA,$67
	.dc.b	$3C,$D8,$CF,$CD,$F5,$FF,$B4,$FC,$79,$F8,$D3,$E8,$9F,$95,$7F,$9B
	.dc.b	$7C,$17,$E6,$3F,$E7,$DF,$03,$F8,$5E,$1C,$2F,$7F,$6F,$F8,$5A,$7D
	.dc.b	$B7,$5C,$BA,$73,$7D,$83,$FC,$89,$87,$83,$E9,$FE,$13,$CF,$4F,$06
	.dc.b	$1F,$AF,$8C,$63,$F9,$1F,$D1,$31,$8F,$9E,$F7,$F4,$3D,$F4,$F1,$ED
	.dc.b	$E8,$FD,$97,$F8,$8F,$C1,$FF,$08,$1F,$4B,$FC,$4F,$E9,$3F,$CD,$7C
	.dc.b	$7A,$F8,$1F,$F4,$34,$E7,$E6,$BF,$CE,$9C,$BF,$CA,$B0,$FE,$23,$F9
	.dc.b	$CF,$D9,$7A,$72,$FB,$4F,$F4,$76,$1C,$BE,$0D,$FB,$37,$F1,$59,$CD
	.dc.b	$78,$A7,$FC,$65,$F8,$5F,$F0,$D8,$F3,$67,$E1,$FC,$B3,$F9,$4F,$91
	.dc.b	$FD,$F3,$5F,$F0,$7F,$E6,$18,$F9,$EF,$EF,$3F,$37,$FE,$11,$F4,$BF
	.dc.b	$AA,$DF,$F8,$47,$5F,$23,$FE,$0B,$6D,$7F,$80,$3D,$1F,$E3,$93,$D7
	.dc.b	$D5,$7B,$38,$AE,$BF,$3F,$DF,$D3,$FF,$02,$7C,$FD,$F9,$EF,$F0,$1F
	.dc.b	$DD,$5E,$0F,$F1,$1F,$D0,$7F,$40,$F0,$FF,$82,$FE,$85,$FE,$35,$FA
	.dc.b	$47,$ED,$4F,$D1,$3F,$1B,$DF,$FD,$5A,$F7,$F5,$7F,$E5,$1A,$D7,$1F
	.dc.b	$3F,$79,$97,$EA,$6F,$F8,$8F,$EE,$D6,$73,$87,$1F,$CD,$7F,$9C,$57
	.dc.b	$4D,$3C,$3E,$76,$2F,$71,$E4,$D7,$FD,$41,$AF,$8E,$73,$AF,$9F,$E8
	.dc.b	$7F,$AD,$FD,$33,$EE,$7F,$CC,$BF,$B8,$7F,$A1,$BF,$4C,$FF,$2F,$7E
	.dc.b	$E3,$A9,$CF,$DB,$27,$FC,$C5,$F8,$CE,$7E,$6F,$F8,$16,$3C,$FE,$C6
	.dc.b	$E1,$C2,$FF,$CB,$9E,$1F,$39,$67,$DF,$EC,$BF,$AA,$61,$C6,$3C,$5F
	.dc.b	$49,$AE,$1F,$F0,$0C,$E7,$5F,$14,$FF,$89,$E7,$FC,$0B,$F9,$13,$D4
	.dc.b	$7E,$E2,$FC,$72,$FE,$3E,$FD,$45,$F8,$57,$F7,$CC,$38,$5E,$C5,$6C
	.dc.b	$D6,$1F,$E1,$EE,$0F,$D3,$57,$06,$C7,$83,$1E,$FF,$90,$FF,$6F,$AE
	.dc.b	$B1,$E2,$F9,$DC,$56,$F1,$EF,$9B,$F2,$6E,$F1,$E3,$CF,$D1,$FA,$AE
	.dc.b	$31,$F5,$DB,$E9,$3F,$C2,$BF,$A4,$BF,$6A,$7D,$37,$F8,$AF,$E4,$DF
	.dc.b	$42,$FD,$1B,$FA,$E7,$2F,$F8,$B3,$0E,$17,$BA,$D7,$8B,$D3,$1E,$7C
	.dc.b	$BF,$C1,$6F,$B1,$FD,$04,$E3,$F6,$5F,$D0,$AD,$FA,$37,$F8,$3E,$73
	.dc.b	$A7,$8B,$F5,$4C,$63,$E6,$C3,$86,$BF,$D6,$F5,$FF,$1F,$FE,$3D,$FD
	.dc.b	$6B,$27,$E1,$6F,$D5,$DF,$21,$FC,$8B,$E8,$FF,$C6,$B7,$BA,$D4,$FF
	.dc.b	$26,$FD,$23,$3F,$33,$E4,$E2,$FE,$3D,$7F,$D9,$B4,$FE,$B4,$F1,$CB
	.dc.b	$AF,$7F,$F8,$9E,$31,$8B,$D2,$C6,$3E,$EC,$FB,$FC,$37,$F9,$9F,$7F
	.dc.b	$4C,$E7,$2F,$3C,$BF,$59,$FF,$5F,$FE,$89,$FE,$72,$FB,$67,$ED,$AF
	.dc.b	$CF,$9F,$44,$FC,$C7,$E0,$79,$FF,$EA,$CD,$58,$F1,$C3,$85,$EF,$4F
	.dc.b	$F6,$FF,$B3,$DB,$94,$7F,$95,$69,$F9,$4F,$F8,$BB,$0E,$5A,$5A,$DE
	.dc.b	$BE,$48,$F1,$EB,$F4,$7D,$1F,$F3,$A5,$F0,$D3,$C1,$F8,$DE,$31,$E3
	.dc.b	$9F,$93,$0F,$4F,$D1,$78,$E5,$FB,$3B,$F6,$27,$EF,$77,$EB,$7F,$C3
	.dc.b	$1D,$CF,$D4,$F8,$5E,$EB,$5F,$4C,$AF,$F6,$A7,$37,$C4,$CB,$DF,$7B
	.dc.b	$7F,$CD,$3C,$1F,$EA,$08,$F0,$7E,$53,$EF,$9F,$1E,$9E,$23,$FC,$59
	.dc.b	$8B,$EF,$FD,$93,$AC,$FC,$3F,$51,$8C,$7C,$98,$70,$D3,$FD,$B7,$A7
	.dc.b	$8F,$07,$D0,$FE,$F7,$F0,$6F,$F0,$AF,$CE,$7F,$93,$7D,$3F,$F7,$5F
	.dc.b	$07,$FA,$D2,$B5,$F4,$CA,$D7,$D3,$7C,$99,$E5,$8F,$B0,$FB,$AF,$F0
	.dc.b	$9F,$7B,$D8,$F2,$AF,$F2,$D7,$0F,$AA,$31,$F6,$9F,$8E,$BE,$9E,$DF
	.dc.b	$FA,$FB,$2F,$04,$FF,$29,$F9,$5F,$B1,$E7,$3F,$44,$7D,$B7,$F9,$EB
	.dc.b	$EB,$1F,$65,$FE,$67,$FA,$DD,$D6,$BE,$F5,$5A,$C9,$FB,$0E,$9F,$A5
	.dc.b	$97,$5C,$AE,$1C,$6F,$AF,$D0,$7D,$2A,$D6,$3D,$33,$FF,$41,$EF,$E0
	.dc.b	$C7,$C9,$1F,$E7,$9F,$D1,$3F,$21,$DB,$FA,$BF,$25,$FA,$FF,$F9,$9F
	.dc.b	$E9,$DF,$0A,$FD,$63,$7E,$4F,$EC,$EB,$52,$FD,$67,$38,$7E,$27,$1F
	.dc.b	$DB,$7A,$17,$0E,$1F,$6E,$CF,$FA,$47,$87,$F8,$3B,$F1,$67,$E1,$E0
	.dc.b	$FC,$6F,$81,$F9,$B7,$F8,$37,$E0,$8F,$D2,$9F,$B5,$7F,$09,$C7,$F1
	.dc.b	$97,$BF,$63,$FD,$7F,$5A,$FB,$0F,$D7,$4F,$E9,$7F,$3F,$FC,$F3,$D7
	.dc.b	$2F,$C4,$07,$96,$5D,$FE,$47,$C7,$E5,$FF,$08,$CB,$AF,$EA,$98,$C3
	.dc.b	$CB,$9F,$93,$7F,$DE,$7D,$5B,$E7,$5F,$A2,$BC,$57,$EE,$8F,$CF,$97
	.dc.b	$FC,$F3,$8F,$F4,$CF,$1B,$E1,$5B,$8F,$E7,$F3,$AC,$1F,$83,$7E,$E6
	.dc.b	$78,$69,$EC,$79,$65,$E0,$C9,$F5,$7F,$BA,$9C,$E5,$E3,$D3,$19,$7A
	.dc.b	$30,$E1,$4F,$DC,$BF,$8C,$7E,$D0,$FD,$A1,$FA,$2F,$FA,$43,$F3,$87
	.dc.b	$B1,$FE,$08,$EF,$DD,$6B,$F0,$25,$6B,$E7,$5F,$BE,$6B,$E3,$FF,$89
	.dc.b	$67,$D7,$3B,$8F,$1C,$3A,$F2,$7F,$12,$AD,$65,$D3,$4F,$C6,$FB,$7F
	.dc.b	$CF,$A9,$F9,$9F,$FB,$D7,$EB,$7F,$93,$76,$3E,$AD,$FA,$37,$EE,$BB
	.dc.b	$3F,$8B,$5E,$FC,$05,$EE,$BF,$C0,$F9,$75,$AF,$F7,$96,$5F,$8A,$7B
	.dc.b	$37,$F0,$FF,$F7,$5E,$1C,$B2,$7E,$53,$FC,$AA,$FB,$65,$E3,$FC,$6B
	.dc.b	$F9,$6F,$CD,$BF,$32,$FE,$4F,$FB,$D3,$F4,$27,$D5,$BF,$2B,$7E,$B3
	.dc.b	$FD,$2F,$0E,$1F,$0C,$C3,$85,$EF,$F4,$6D,$65,$FC,$CB,$8B,$F4,$38
	.dc.b	$77,$FC,$27,$FA,$13,$CB,$0F,$07,$E4,$DF,$2F,$9F,$F9,$D4,$E7,$6F
	.dc.b	$1F,$E1,$5E,$8C,$BC,$F1,$E6,$DF,$FB,$B7,$3F,$D3,$FC,$FD,$F2,$8F
	.dc.b	$8A,$FE,$09,$FB,$A3,$F3,$B7,$DD,$FF,$7C,$DB,$FA,$46,$1C,$2F,$75
	.dc.b	$AF,$C2,$FF,$B5,$4F,$4F,$33,$FA,$D3,$BE,$4F,$DE,$BC,$7D,$E6,$7D
	.dc.b	$FE,$7F,$FA,$DC,$E6,$3C,$5F,$3D,$8D,$BC,$D7,$C3,$37,$E8,$DF,$D8
	.dc.b	$BF,$1B,$FE,$F1,$FA,$8B,$F2,$7F,$EE,$5F,$3A,$FD,$3D,$F5,$DF,$DF
	.dc.b	$EF,$75,$A9,$7E,$D5,$9F,$9F,$79,$BF,$DC,$EF,$AF,$1F,$F1,$67,$8F
	.dc.b	$97,$FE,$06,$F9,$6F,$15,$B5,$F1,$4F,$F8,$2B,$F1,$1F,$EE,$FD,$CF
	.dc.b	$DE,$9F,$80,$7E,$93,$F0,$07,$E6,$CF,$D9,$5F,$8D,$FF,$85,$70,$E1
	.dc.b	$7B,$AD,$71,$67,$17,$E5,$FF,$E0,$9B,$DF,$0F,$F2,$C7,$8F,$A5,$FE
	.dc.b	$8E,$C1,$F9,$07,$93,$0F,$1C,$E7,$F4,$FC,$67,$DF,$5F,$EC,$FE,$2F
	.dc.b	$F8,$27,$2E,$7B,$FA,$25,$FA,$C7,$F8,$AF,$F2,$4F,$F2,$1F,$D9,$3F
	.dc.b	$4D,$7E,$BD,$FC,$67,$F1,$FF,$F0,$CD,$6A,$73,$F1,$2F,$CC,$3F,$21
	.dc.b	$E7,$FF,$09,$AE,$7C,$BF,$D1,$9F,$F2,$7F,$DB,$FF,$89,$DE,$59,$7F
	.dc.b	$25,$FC,$BB,$AD,$FF,$96,$32,$EB,$E0,$E3,$5F,$25,$F8,$FE,$6B,$FC
	.dc.b	$85,$F9,$2F,$E1,$BF,$78,$FE,$77,$F5,$3F,$E6,$1F,$C7,$5F,$AD,$6B
	.dc.b	$CA,$56,$BE,$95,$5F,$D9,$7A,$BF,$C5,$D8,$F3,$E3,$FE,$BA,$FE,$4B
	.dc.b	$F1,$AF,$F1,$F3,$DF,$E6,$DA,$7F,$9B,$FE,$6B,$FC,$7B,$39,$8F,$F3
	.dc.b	$97,$CC,$FF,$90,$3F,$0A,$7C,$C3,$F0,$37,$DD,$BE,$F9,$F8,$EB,$F2
	.dc.b	$97,$6F,$F9,$7F,$1E,$38,$70,$BD,$D3,$F2,$9D,$65,$FE,$2A,$E1,$FE
	.dc.b	$EB,$87,$06,$C7,$87,$3F,$06,$7F,$8C,$9E,$39,$75,$8F,$F1,$E7,$17
	.dc.b	$F9,$47,$0F,$0D,$F0,$C9,$F9,$47,$F9,$CF,$F7,$77,$EE,$4F,$D0,$AF
	.dc.b	$C6,$FF,$B9,$FD,$AF,$F7,$97,$87,$FD,$36,$F7,$5A,$FC,$5D,$5F,$EF
	.dc.b	$3A,$D6,$1C,$F8,$3F,$C1,$F8,$76,$EE,$FF,$38,$3C,$A5,$97,$FB,$D7
	.dc.b	$F1,$2F,$F3,$7C,$E6,$3F,$DF,$1F,$1F,$8A,$F3,$1C,$33,$F0,$FE,$35
	.dc.b	$FE,$8D,$FD,$CF,$E8,$ED,$EE,$FC,$19,$F9,$E3,$F4,$07,$EE,$5B,$ED
	.dc.b	$F8,$1C,$8B,$DD,$3F,$2A,$D7,$17,$FA,$1F,$3E,$BC,$7F,$E2,$A7,$BE
	.dc.b	$1C,$73,$7E,$B3,$FE,$88,$9C,$E5,$E2,$FD,$07,$1A,$78,$23,$BC,$7F
	.dc.b	$A9,$7F,$72,$7E,$C6,$FD,$BD,$FA,$EB,$F3,$6E,$BE,$6F,$F3,$4F,$07
	.dc.b	$FD,$1D,$87,$0B,$DD,$6B,$EA,$3A,$63,$CF,$87,$FC,$83,$7D,$BF,$17
	.dc.b	$FF,$51,$3C,$73,$ED,$C0,$D7,$FD,$43,$39,$CB,$DB,$38,$C7,$CD,$7C
	.dc.b	$35,$7E,$8B,$E0,$D7,$C7,$7F,$ED,$DF,$B7,$FF,$61,$7D,$B7,$F8,$3B
	.dc.b	$DC,$FF,$0A,$7D,$57,$E8,$E1,$FB,$46,$A3,$F8,$B7,$D6,$FF,$85,$3C
	.dc.b	$EF,$F2,$67,$F0,$06,$7E,$0B,$FD,$75,$87,$3E,$5F,$F2,$EB,$D6,$F8
	.dc.b	$7D,$F7,$FA,$EA,$31,$8F,$EC,$7B,$B0,$F6,$3C,$5F,$B0,$FF,$B5,$FF
	.dc.b	$50,$FF,$79,$7E,$B2,$F2,$DF,$A0,$BF,$48,$7E,$9A,$F2,$B8,$7E,$14
	.dc.b	$C3,$85,$EE,$BF,$4D,$C1,$FE,$F4,$CB,$FD,$FD,$F7,$9F,$ED,$37,$BC
	.dc.b	$7F,$9D,$9F,$0E,$7E,$0C,$DF,$A5,$FF,$B5,$27,$38,$F8,$FF,$37,$C6
	.dc.b	$9E,$1C,$38,$67,$E7,$F1,$7F,$DA,$9C,$3F,$EF,$2F,$CB,$1F,$01,$FD
	.dc.b	$AF,$EA,$1F,$B7,$BD,$5F,$F7,$C5,$EE,$B5,$3F,$EC,$3C,$39,$FE,$40
	.dc.b	$FD,$7F,$FD,$D9,$7D,$78,$FF,$D1,$EF,$1F,$A2,$66,$F9,$1F,$F7,$6C
	.dc.b	$E6,$3F,$C0,$75,$E0,$C1,$AF,$9B,$0E,$39,$BF,$2B,$D4,$E7,$0F,$4F
	.dc.b	$D3,$B5,$FF,$79,$7E,$7F,$FE,$95,$FC,$D5,$FA,$C7,$F0,$3F,$CC,$3F
	.dc.b	$DF,$DF,$B6,$EF,$E0,$FF,$CD,$57,$BA,$D7,$E4,$AF,$DF,$B5,$C4,$15
	.dc.b	$D5,$FF,$56,$BF,$EA,$3E,$EF,$F3,$A4,$7F,$92,$F0,$EB,$F8,$7B,$4F
	.dc.b	$1F,$C0,$35,$67,$D7,$6F,$4D,$EF,$5F,$47,$C7,$E7,$2F,$F4,$07,$E4
	.dc.b	$EF,$35,$FA,$47,$FA,$83,$F1,$CF,$F1,$B7,$EB,$4F,$1B,$D3,$C2,$C3
	.dc.b	$85,$EF,$99,$6F,$88,$DD,$75,$FB,$1F,$5E,$1E,$13,$E1,$47,$27,$FD
	.dc.b	$9A,$F4,$F9,$6F,$F7,$46,$1F,$C2,$E7,$9E,$9E,$69,$CF,$B1,$CE,$7C
	.dc.b	$93,$9D,$7D,$3F,$15,$E8,$FC,$41,$FB,$C7,$E2,$BE,$D3,$FB,$87,$E5
	.dc.b	$3F,$03,$F1,$4D,$4E,$7E,$51,$3F,$13,$F3,$FC,$F8,$7E,$AD,$77,$FA
	.dc.b	$A5,$EC,$74,$FF,$AA,$F5,$EF,$6D,$7B,$FE,$99,$E7,$9C,$C7,$8A,$71
	.dc.b	$AF,$87,$4F,$A7,$F8,$57,$82,$BC,$FD,$5F,$E5,$1C,$FB,$EB,$F8,$BF
	.dc.b	$DF,$7F,$C1,$3F,$05,$FD,$93,$F1,$DF,$E1,$9F,$B4,$7F,$2C,$BD,$DA
	.dc.b	$F7,$F3,$2B,$DF,$C8,$FD,$E9,$FF,$09,$F4,$FB,$78,$FF,$D8,$D1,$DB
	.dc.b	$0F,$B7,$8F,$48,$E7,$87,$F1,$BF,$6B,$EA,$46,$3D,$FF,$E2,$37,$D7
	.dc.b	$E4,$FF,$9B,$E9,$E1,$96,$BF,$F0,$03,$2F,$3F,$E8,$8F,$D9,$5F,$95
	.dc.b	$BF,$71,$FF,$13,$FE,$AF,$FC,$13,$D3,$FC,$AD,$EE,$B5,$A2,$BF,$84
	.dc.b	$FE,$C7,$AC,$79,$ED,$EA,$E2,$FF,$71,$63,$DF,$F3,$2F,$81,$3F,$B7
	.dc.b	$8E,$9A,$78,$3B,$3F,$21,$18,$CF,$B7,$77,$E4,$FF,$16,$7D,$CB,$F6
	.dc.b	$2F,$F1,$37,$B3,$FE,$72,$F6,$9F,$C0,$1F,$AC,$BD,$D6,$BE,$EB,$F8
	.dc.b	$CF,$E6,$79,$C7,$FD,$C7,$BF,$9C,$FE,$15,$7C,$1C,$07,$9F,$96,$CF
	.dc.b	$FD,$03,$E7,$79,$6F,$C9,$39,$C7,$F9,$C6,$1D,$F2,$F1,$68,$FD,$FF
	.dc.b	$F8,$0F,$E7,$9F,$E1,$DF,$AC,$FF,$6A,$7A,$0F,$D4,$3F,$DD,$BB,$EB
	.dc.b	$EA,$15,$A9,$FF,$78,$FD,$76,$7F,$06,$F9,$BA,$B2,$E8,$FA,$9E,$DF
	.dc.b	$8B,$B3,$ED,$1F,$BF,$32,$F8,$7F,$27,$C6,$7E,$0A,$E1,$97,$FC,$00
	.dc.b	$BF,$E3,$3F,$A6,$7F,$89,$BF,$4D,$7E,$AD,$FE,$CC,$FC,$C9,$FA,$DB
	.dc.b	$F0,$2D,$FD,$52,$F7,$F8,$9A,$F7,$FB,$0E,$BE,$4F,$D7,$9F,$5E,$2F
	.dc.b	$A6,$F7,$D3,$B7,$C4,$7F,$17,$8C,$63,$CF,$E2,$F1,$A7,$7B,$F2,$6C
	.dc.b	$FC,$1F,$F9,$8F,$EB,$2F,$C7,$7F,$D1,$5F,$BE,$3F,$75,$7D,$1B,$9F
	.dc.b	$FD,$AF,$EA,$FF,$00,$C3,$87,$E0,$2C,$38,$5E,$FF,$6D,$EB,$97,$FA
	.dc.b	$57,$8F,$F1,$78,$77,$E2,$FA,$EF,$2C,$38,$F9,$1F,$CC,$63,$1A,$F8
	.dc.b	$B9,$31,$8F,$9A,$3C,$BB,$7F,$88,$F6,$FE,$27,$F6,$8F,$80,$FF,$11
	.dc.b	$FD,$3B,$F5,$F7,$EF,$2F,$DB,$BF,$C0,$D8,$70,$BD,$D6,$A7,$FB,$FC
	.dc.b	$E7,$E9,$BF,$1B,$8F,$5F,$AF,$B8,$70,$C7,$F3,$F7,$BC,$BC,$10,$E0
	.dc.b	$FE,$99,$87,$97,$1F,$F1,$A6,$7D,$FF,$15,$FE,$C7,$F0,$BF,$4F,$F7
	.dc.b	$B7,$DE,$3F,$15,$7D,$BF,$F8,$FB,$F2,$0D,$F2,$AF,$75,$AC,$D8,$EB
	.dc.b	$EA,$3F,$4A,$F5,$EB,$FD,$EB,$DF,$3E,$D9,$FF,$2C,$E2,$FE,$7B,$39
	.dc.b	$C7,$C2,$FF,$65,$31,$3F,$E5,$0F,$88,$FE,$F6,$3F,$73,$7E,$32,$FD
	.dc.b	$2D,$FB,$73,$F4,$67,$E5,$5E,$ED,$3F,$A5,$E3,$C7,$0E,$17,$BA,$FF
	.dc.b	$5E,$D6,$B1,$EB,$B7,$F1,$6C,$3C,$1F,$B8,$7F,$02,$79,$67,$E0,$F2
	.dc.b	$3F,$C4,$15,$FE,$27,$9C,$E5,$E3,$FC,$BF,$FC,$01,$18,$D7,$CB,$FC
	.dc.b	$01,$FE,$16,$E4,$7D,$C3,$F5,$5F,$E5,$FE,$CB,$F5,$37,$87,$FE,$DC
	.dc.b	$E2,$FE,$CB,$87,$0B,$DD,$6B,$EE,$3A,$63,$CF,$7F,$E5,$F7,$DB,$F6
	.dc.b	$5F,$E3,$6F,$1F,$8A,$69,$DF,$EC,$FF,$C2,$D3,$98,$F1,$75,$63,$1F
	.dc.b	$35,$F0,$CF,$CB,$FB,$17,$F8,$C3,$EC,$5F,$AD,$FF,$82,$FF,$80,$3F
	.dc.b	$CA,$DF,$6D,$FE,$64,$F6,$75,$39,$E3,$FF,$23,$7C,$A7,$DD,$7F,$C0
	.dc.b	$38,$72,$FA,$4F,$E6,$2F,$4F,$8E,$B7,$C7,$1F,$F3,$7F,$27,$4B,$67
	.dc.b	$E0,$F9,$3F,$F1,$86,$3C,$E3,$C7,$9B,$F8,$03,$FC,$8F,$8B,$F4,$07
	.dc.b	$D5,$5F,$AB,$7E,$E5,$FB,$E9,$D9,$A1,$1A,$F5,$1F,$B0,$EB,$A7,$FA
	.dc.b	$4E,$3D,$71,$B8,$71,$C3,$AF,$C2,$7F,$91,$6B,$53,$D3,$2F,$E7,$DC
	.dc.b	$3E,$0C,$3C,$91,$FC,$43,$D0,$C5,$6F,$47,$A1,$FE,$62,$FA,$4F,$F2
	.dc.b	$F7,$E1,$AF,$D9,$7F,$D4,$7F,$AD,$FF,$1B,$FB,$BD,$7D,$CF,$FB,$72
	.dc.b	$73,$EF,$3B,$33,$87,$F6,$2C,$7F,$CD,$1E,$AD,$C3,$87,$90,$CF,$E9
	.dc.b	$F7,$FF,$8E,$AF,$C3,$9F,$83,$D8,$FF,$31,$46,$2F,$86,$9F,$E2,$7B
	.dc.b	$FF,$20,$FE,$EB,$FE,$3F,$FD,$4B,$FB,$0F,$E6,$6F,$D8,$1E,$17,$F4
	.dc.b	$5B,$DF,$D8,$2F,$75,$AF,$BD,$FF,$41,$E3,$FE,$49,$F9,$AF,$EF,$6F
	.dc.b	$5E,$CF,$F0,$9B,$DF,$E4,$5A,$77,$D3,$FC,$F1,$39,$8F,$17,$D2,$E3
	.dc.b	$4F,$05,$F0,$CF,$CB,$87,$FA,$4B,$AD,$7F,$47,$E4,$5F,$BF,$7F,$AB
	.dc.b	$7F,$6D,$7E,$E7,$BA,$D4,$E7,$E3,$F1,$AF,$F8,$4B,$1E,$7D,$16,$F7
	.dc.b	$C1,$FE,$30,$E4,$FF,$34,$D6,$B4,$EF,$2F,$D6,$3F,$D2,$53,$9C,$FC
	.dc.b	$5F,$B2,$FF,$AA,$3C,$17,$EF,$4F,$D1,$5F,$96,$BF,$4C,$7E,$84,$F8
	.dc.b	$4F,$F2,$4E,$FF,$E7,$8C,$78,$E1,$C2,$F7,$4F,$D4,$3B,$64,$FD,$CF
	.dc.b	$FC,$5B,$87,$0E,$0F,$F2,$2B,$CB,$4F,$06,$9D,$E7,$8E,$7D,$7F,$29
	.dc.b	$FE,$63,$1F,$A0,$C7,$D1,$7C,$33,$7E,$BD,$FE,$C4,$FC,$DF,$FC,$F3
	.dc.b	$F2,$1F,$87,$D5,$FB,$F3,$E8,$D5,$BF,$8C,$FE,$9F,$E3,$79,$EB,$5F
	.dc.b	$8D,$67,$1F,$F2,$2D,$FF,$A1,$3E,$46,$E1,$C3,$2E,$BF,$8B,$7F,$B0
	.dc.b	$6B,$C3,$97,$3F,$6F,$FD,$89,$FA,$63,$AB,$F9,$4F,$E5,$3F,$CA,$BA
	.dc.b	$7F,$DA,$C7,$FB,$72,$75,$F0,$89,$D7,$A7,$EE,$ED,$FF,$54,$D7,$3F
	.dc.b	$D3,$FF,$CB,$2F,$0E,$9F,$F3,$8B,$CA,$FB,$6D,$E2,$CF,$C3,$97,$D7
	.dc.b	$FC,$8B,$0F,$73,$ED,$66,$FD,$73,$FD,$DF,$F8,$DF,$F8,$97,$C1,$7E
	.dc.b	$9C,$F4,$5F,$46,$FC,$D5,$FB,$7F,$0F,$5F,$F1,$97,$BA,$D7,$ED,$B9
	.dc.b	$FD,$CB,$FC,$E3,$9F,$5E,$4F,$F4,$4B,$DE,$F8,$7E,$89,$FE,$ED,$8C
	.dc.b	$69,$E0,$FD,$17,$FD,$E1,$FA,$BB,$F2,$46,$7F,$9E,$FD,$1F,$F9,$27
	.dc.b	$CF,$B9,$9D,$9F,$ED,$2C,$38,$5E,$FD,$55,$FF,$9D,$2B,$58,$F4,$DF
	.dc.b	$FD,$3B,$87,$7F,$D2,$3F,$D1,$2F,$2F,$35,$A7,$F3,$3F,$CE,$BC,$B8
	.dc.b	$79,$27,$39,$FF,$90,$3F,$0E,$65,$31,$E5,$8F,$A3,$F3,$5F,$3D,$6A
	.dc.b	$9F,$98,$7E,$E3,$E5,$7F,$D1,$3E,$67,$F8,$2B,$F1,$4F,$E4,$DC,$1F
	.dc.b	$EE,$EA,$D7,$C1,$FF,$8B,$3E,$97,$FD,$B3,$F0,$5D,$32,$F0,$D7,$F0
	.dc.b	$2F,$D0,$7F,$D4,$6F,$58,$FF,$5D,$BC,$B2,$FF,$0B,$FC,$27,$A3,$2E
	.dc.b	$DE,$C7,$F6,$FC,$FB,$7C,$0E,$77,$F4,$63,$C7,$5F,$F7,$4F,$C3,$F8
	.dc.b	$74,$F5,$E3,$C7,$4F,$57,$BB,$9F,$86,$7E,$83,$FE,$5C,$F9,$D7,$EC
	.dc.b	$1F,$E5,$CF,$89,$7E,$3D,$BA,$D4,$E7,$F2,$1C,$7D,$97,$FA,$FB,$1E
	.dc.b	$7F,$9F,$5B,$DF,$CA,$3E,$3B,$D1,$C1,$F5,$2F,$C3,$83,$5F,$0D,$FF
	.dc.b	$9D,$6B,$AC,$63,$57,$E9,$1E,$DF,$AD,$7D,$EB,$F3,$D7,$CB,$3E,$E9
	.dc.b	$F9,$E3,$DE,$7D,$37,$1C,$38,$5E,$EB,$5F,$4B,$99,$FF,$6C,$E5,$DB
	.dc.b	$2B,$1C,$32,$EB,$97,$B3,$6F,$F2,$94,$E7,$1F,$F2,$2C,$7D,$3F,$DB
	.dc.b	$1F,$8D,$7E,$F7,$E9,$DE,$8B,$F2,$7F,$F6,$97,$D2,$6E,$B5,$39,$FC
	.dc.b	$2D,$F8,$4E,$7E,$AF,$FD,$DB,$97,$4F,$6E,$E1,$C3,$26,$7E,$5A,$D3
	.dc.b	$E0,$8F,$F1,$2F,$96,$F2,$7F,$45,$5A,$D3,$C1,$F1,$3F,$8B,$FD,$69
	.dc.b	$FA,$AA,$BF,$95,$7E,$20,$FD,$E1,$F9,$2F,$FB,$5F,$3E,$9C,$9F,$B2
	.dc.b	$C7,$8E,$1C,$2F,$7F,$AE,$75,$C9,$5F,$B3,$BF,$EA,$AF,$07,$06,$EF
	.dc.b	$97,$C9,$FD,$19,$EB,$A3,$4F,$27,$E2,$7F,$57,$F4,$57,$EA,$0F,$CD
	.dc.b	$5F,$6A,$FC,$CD,$F0,$BF,$C7,$BE,$73,$F5,$58,$70,$BD,$D6,$BE,$23
	.dc.b	$FB,$F7,$27,$BB,$85,$7D,$BF,$73,$F3,$C7,$2D,$3C,$0F,$AA,$B8,$57
	.dc.b	$5D,$3C,$5A,$7E,$8E,$B8,$67,$E0,$8F,$F1,$6C,$62,$77,$93,$F5,$2F
	.dc.b	$E0,$1F,$BD,$3D,$B7,$E2,$2F,$CC,$BF,$B3,$7E,$2A,$FD,$99,$FA,$C7
	.dc.b	$0B,$DD,$6A,$5F,$9B,$E7,$0F,$5F,$0F,$BF,$2E,$DF,$AC,$E3,$C2,$FE
	.dc.b	$55,$1E,$1A,$F0,$7D,$0B,$5F,$E8,$DC,$3C,$F4,$F1,$F9,$DF,$A9,$C3
	.dc.b	$84,$78,$7F,$47,$D6,$1C,$76,$FF,$6A,$67,$FC,$DF,$F6,$1F,$F1,$6F
	.dc.b	$DB,$BE,$35,$F5,$5F,$AB,$FB,$E7,$E1,$9F,$C8,$EF,$75,$AF,$33,$FC
	.dc.b	$ED,$ED,$67,$E2,$BE,$6F,$17,$E4,$79,$BE,$E6,$1F,$E3,$AF,$DD,$BD
	.dc.b	$D3,$FD,$32,$39,$E9,$FB,$9F,$80,$FE,$05,$18,$C9,$F3,$DF,$C1,$7F
	.dc.b	$72,$7E,$CC,$FD,$09,$FB,$37,$EB,$3A,$BF,$65,$EF,$7F,$39,$BD,$D6
	.dc.b	$BD,$2F,$E3,$5C,$EF,$D4,$F5,$97,$D6,$E0,$FC,$64,$FF,$5B,$C3,$8C
	.dc.b	$60,$FE,$C6,$F4,$CF,$F7,$7F,$6D,$FE,$42,$8C,$63,$FC,$9F,$EE,$3F
	.dc.b	$8E,$3F,$E6,$EF,$8F,$7E,$82,$FD,$DF,$FD,$11,$F5,$AF,$6B,$FA,$85
	.dc.b	$EE,$B5,$F6,$BF,$E1,$9F,$D5,$73,$F1,$DF,$D3,$37,$F2,$1F,$C3,$16
	.dc.b	$7F,$3A,$30,$F8,$33,$FF,$4B,$73,$79,$2B,$78,$F6,$FA,$9C,$5E,$F2
	.dc.b	$FF,$14,$7E,$7F,$FD,$8F,$8F,$FD,$DB,$F8,$BB,$F2,$3E,$FF,$62,$FD
	.dc.b	$41,$E8,$6F,$ED,$D7,$BC,$D7,$BF,$47,$5F,$09,$E7,$CF,$AF,$1F,$B1
	.dc.b	$EF,$A7,$6F,$8F,$FE,$8D,$18,$C7,$9F,$E8,$98,$D3,$BD,$F9,$36,$FF
	.dc.b	$44,$FE,$17,$FC,$8F,$F2,$D7,$E2,$EF,$CA,$9F,$AF,$3F,$24,$7E,$9E
	.dc.b	$FC,$8F,$FD,$53,$8F,$1C,$38,$5E,$EB,$5F,$8E,$AB,$5F,$6D,$FC,$61
	.dc.b	$E1,$95,$E3,$FF,$13,$69,$E2,$8F,$0E,$9E,$08,$FF,$17,$D7,$0C,$BB
	.dc.b	$75,$E3,$0E,$3A,$78,$B0,$F3,$EA,$FC,$D7,$C5,$5D,$B9,$9E,$0B,$F7
	.dc.b	$CF,$F0,$9F,$EC,$6F,$DE,$7E,$D6,$DF,$16,$F7,$5A,$97,$ED,$7D,$31
	.dc.b	$FF,$21,$F0,$7F,$59,$F0,$3D,$91,$8C,$3B,$BF,$D3,$B1,$E9,$9F,$87
	.dc.b	$EB,$5A,$7F,$B0,$EB,$A0,$C7,$FB,$25,$6E,$7F,$C2,$33,$AF,$CC,$9F
	.dc.b	$B0,$3E,$E7,$DB,$F8,$03,$F7,$B7,$E5,$BF,$EF,$3D,$7F,$B1,$61,$C2
	.dc.b	$F7,$5A,$F8,$CE,$98,$F3,$CB,$D0,$7F,$96,$EF,$BF,$DD,$FB,$A1,$1C
	.dc.b	$F5,$FF,$3C,$7E,$51,$FE,$61,$CF,$BE,$1E,$59,$FF,$3E,$65,$FE,$C9
	.dc.b	$9F,$EE,$9F,$BC,$FF,$4F,$FB,$0F,$F6,$A7,$E5,$EF,$DB,$1F,$BA,$BF
	.dc.b	$4B,$F8,$B7,$BF,$C4,$D7,$BA,$7E,$25,$AF,$98,$65,$E7,$C3,$FA,$E7
	.dc.b	$E9,$56,$73,$8F,$2C,$DF,$81,$7E,$37,$8F,$FC,$FF,$87,$7D,$3C,$9F
	.dc.b	$82,$66,$53,$9B,$F3,$FE,$19,$F6,$BF,$77,$7E,$1F,$FE,$34,$FB,$B7
	.dc.b	$EA,$AF,$D4,$5E,$EB,$F3,$CE,$17,$BA,$D4,$BF,$40,$CF,$DA,$7D,$1C
	.dc.b	$BA,$FE,$CD,$70,$E1,$93,$F6,$BF,$35,$78,$AF,$FC,$AF,$97,$87,$F2
	.dc.b	$6C,$E1,$C3,$37,$EB,$1E,$9F,$C2,$BF,$A6,$70,$FF,$87,$3F,$1B,$7E
	.dc.b	$F3,$FE,$FA,$F0,$3F,$C5,$9A,$FF,$9F,$BE,$57,$EC,$7D,$6F,$FB,$4E
	.dc.b	$9D,$9C,$FE,$AF,$D1,$9F,$F5,$7E,$AF,$C6,$9F,$EA,$CC,$79,$69,$DF
	.dc.b	$EB,$BD,$35,$E2,$CB,$FD,$BD,$FA,$E7,$D1,$FC,$09,$F9,$83,$F4,$C7
	.dc.b	$E8,$4F,$DA,$3F,$BC,$FE,$6E,$FC,$93,$F5,$B8,$70,$BD,$D6,$BF,$7C
	.dc.b	$CE,$3C,$F8,$3C,$A2,$FB,$F1,$7E,$85,$E5,$9F,$7F,$A7,$FF,$41,$D7
	.dc.b	$2C,$BA,$C3,$C6,$FC,$47,$E1,$2F,$CA,$DF,$32,$FB,$67,$E3,$0F,$C0
	.dc.b	$1F,$A6,$3F,$75,$FE,$0B,$8F,$1C,$38,$5E,$EB,$F9,$87,$7F,$F9,$7E
	.dc.b	$BB,$7D,$C7,$ED,$DE,$1B,$FB,$1E,$59,$F8,$33,$FF,$53,$5F,$1D,$3C
	.dc.b	$B3,$9F,$1B,$D3,$18,$CF,$C1,$FB,$6F,$93,$F5,$07,$E6,$1F,$EF,$4F
	.dc.b	$C7,$DF,$A9,$BF,$49,$7E,$AB,$BA,$D4,$E7,$F4,$4E,$5B,$F9,$F1,$E7
	.dc.b	$C3,$E7,$7A,$FE,$54,$CF,$AE,$7F,$32,$73,$1E,$0D,$FF,$6D,$F8,$BB
	.dc.b	$D4,$7E,$0F,$FE,$9A,$F8,$07,$F0,$0B,$F5,$1F,$2F,$3F,$FA,$1F,$0E
	.dc.b	$17,$BA,$D7,$77,$F3,$69,$E9,$EB,$7F,$23,$77,$C1,$F8,$87,$8F,$E2
	.dc.b	$DF,$C5,$2D,$E5,$79,$35,$F1,$CE,$7E,$3F,$19,$F7,$D5,$CB,$FD,$13
	.dc.b	$F1,$2F,$E1,$7F,$45,$FB,$2F,$C5,$5F,$A1,$FC,$EF,$D0,$35,$EE,$7F
	.dc.b	$78,$F6,$3F,$D5,$53,$F4,$BE,$27,$3E,$AF,$F3,$9C,$B9,$F2,$7E,$79
	.dc.b	$EB,$7C,$3F,$17,$FE,$3D,$18,$CF,$BF,$E2,$B8,$CF,$B4,$78,$BC,$17
	.dc.b	$C5,$F9,$FE,$13,$E7,$FD,$FB,$95,$FB,$77,$F9,$6B,$F0,$3D,$73,$FF
	.dc.b	$BC,$27,$3F,$6E,$FD,$E3,$97,$7F,$9F,$2F,$B3,$C7,$E7,$7A,$DF,$08
	.dc.b	$E7,$9F,$7C,$DF,$51,$FC,$DF,$B3,$FA,$5F,$EA,$DF,$EB,$8F,$DA,$5F
	.dc.b	$A4,$BF,$74,$FA,$BE,$37,$FB,$1E,$F7,$F8,$CA,$F7,$5C,$2B,$5F,$2E
	.dc.b	$FD,$73,$FB,$26,$1D,$B6,$FE,$12,$F2,$C3,$BF,$DB,$78,$ED,$A7,$8F
	.dc.b	$4F,$EF,$B9,$74,$8C,$67,$F9,$58,$F2,$7E,$C2,$F9,$FF,$F4,$B7,$C4
	.dc.b	$BB,$DF,$9B,$7F,$50,$BD,$D6,$A7,$3F,$9D,$7F,$83,$B7,$63,$CF,$83
	.dc.b	$CE,$FF,$6C,$6D,$F1,$C7,$E5,$4F,$87,$F1,$3F,$F6,$FE,$3F,$DE,$EC
	.dc.b	$4C,$3A,$E1,$C7,$F0,$3F,$F0,$EF,$EE,$3F,$E6,$4F,$3D,$FA,$F7,$FC
	.dc.b	$01,$F8,$FF,$F0,$C7,$ED,$BA,$FC,$17,$F4,$F3,$9F,$A5,$7E,$E9,$9C
	.dc.b	$3F,$C2,$58,$FF,$9B,$BD,$4B,$87,$0F,$B1,$67,$ED,$E5,$F2,$F0,$79
	.dc.b	$B0,$F1,$69,$E1,$F3,$7F,$BE,$7E,$78,$FD,$41,$F8,$57,$D6,$FB,$87
	.dc.b	$EE,$0F,$D3,$9F,$B8,$7F,$8E,$30,$E1,$7B,$AD,$7C,$A6,$71,$E7,$AF
	.dc.b	$F4,$CB,$ED,$F6,$BF,$E2,$B7,$8F,$85,$FE,$55,$C3,$FC,$B3,$C1,$E8
	.dc.b	$B5,$6B,$5F,$1F,$E6,$FE,$8F,$D3,$1F,$BD,$BF,$69,$FF,$48,$FE,$DF
	.dc.b	$F3,$BF,$6F,$F3,$56,$A7,$3C,$AF,$2F,$3F,$AB,$78,$EB,$9E,$BE,$77
	.dc.b	$87,$5B,$2F,$F6,$9C,$F4,$CF,$C1,$39,$FC,$7F,$C9,$18,$AF,$BF,$F3
	.dc.b	$DF,$E5,$EF,$47,$F5,$1F,$79,$FE,$30,$FC,$97,$F9,$AF,$E8,$9F,$CC
	.dc.b	$B3,$FA,$F5,$AF,$B8,$56,$BE,$2F,$97,$AC,$C7,$FC,$A9,$7F,$B3,$FD
	.dc.b	$02,$CE,$70,$E3,$F0,$1F,$DC,$F2,$E3,$87,$86,$3F,$BC,$F5,$7C,$FF
	.dc.b	$D6,$1F,$BD,$FF,$93,$3F,$1A,$7E,$23,$FE,$60,$FD,$8F,$FC,$A9,$7B
	.dc.b	$AD,$4E,$7E,$5B,$FC,$F7,$87,$2F,$C0,$3F,$CD,$0F,$48,$B5,$C3,$F0
	.dc.b	$37,$97,$FE,$4D,$AD,$67,$E0,$97,$C7,$67,$F0,$A7,$E8,$DF,$EF,$9E
	.dc.b	$C7,$DD,$BF,$67,$FF,$5C,$F0,$7A,$EF,$7F,$92,$AF,$74,$FC,$BB,$CF
	.dc.b	$5C,$79,$BF,$C4,$3B,$7F,$A8,$33,$F4,$CF,$2C,$3B,$F3,$F8,$E7,$C5
	.dc.b	$A7,$CD,$D3,$C1,$18,$CD,$F2,$DF,$E9,$EF,$D7,$7F,$D3,$1F,$10,$FD
	.dc.b	$07,$F9,$BF,$E1,$0F,$E0,$0F,$DC,$E5,$FA,$1B,$DD,$6A,$78,$4E,$7F
	.dc.b	$0D,$FF,$43,$F0,$FF,$8E,$2F,$FD,$49,$FB,$D7,$FA,$5D,$E3,$F1,$EC
	.dc.b	$BF,$D5,$D5,$E3,$BF,$F4,$7E,$5E,$2A,$FF,$3F,$5F,$0B,$F2,$7E,$2D
	.dc.b	$E9,$F9,$A7,$E7,$9F,$E7,$9F,$C7,$9F,$AE,$BF,$3C,$7E,$0D,$BA,$D4
	.dc.b	$E7,$EA,$F9,$7C,$8F,$97,$1E,$7C,$1E,$97,$AF,$E1,$8C,$FA,$FE,$47
	.dc.b	$E4,$9C,$C7,$83,$D8,$C6,$7D,$B5,$7E,$11,$FE,$CC,$FD,$69,$F9,$B3
	.dc.b	$F6,$27,$E3,$4F,$D3,$DD,$4F,$C7,$1F,$AF,$71,$C3,$85,$EE,$B5,$FA
	.dc.b	$D6,$71,$FF,$62,$65,$DB,$E3,$EE,$1C,$33,$EB,$FA,$BF,$FA,$A6,$BC
	.dc.b	$77,$E2,$9F,$0F,$06,$6F,$86,$5E,$3F,$D5,$FD,$5F,$97,$79,$7F,$45
	.dc.b	$FF,$6E,$FE,$5E,$FD,$4B,$F9,$AF,$8B,$FC,$9F,$55,$EF,$E4,$57,$BC
	.dc.b	$39,$FE,$D5,$E8,$AF,$81,$DB,$FE,$80,$C3,$8F,$32,$FB,$FD,$27,$A6
	.dc.b	$FA,$62,$D7,$17,$BC,$FB,$EB,$FE,$62,$D7,$F9,$D7,$E6,$FF,$E8,$BF
	.dc.b	$7B,$F4,$7F,$86,$7F,$9F,$7E,$B7,$F7,$15,$AF,$C4,$15,$AF,$C3,$D5
	.dc.b	$FC,$02,$B5,$EF,$7F,$BC,$F1,$E9,$C7,$FE,$98,$7B,$7A,$5F,$E8,$AC
	.dc.b	$3F,$B1,$4F,$F6,$29,$CE,$7F,$E8,$DB,$FF,$6E,$46,$2B,$BF,$A4,$C7
	.dc.b	$FA,$97,$EA,$3F,$E9,$9F,$CD,$3F,$B9,$FD,$33,$F4,$57,$E2,$16,$F7
	.dc.b	$F8,$7A,$F7,$5A,$FB,$1E,$9B,$FF,$74,$CB,$A3,$FE,$B0,$7F,$D4,$DF
	.dc.b	$92,$FF,$61,$8F,$F7,$84,$78,$7E,$A9,$A7,$87,$EB,$B3,$9F,$79,$CE
	.dc.b	$5F,$BE,$D3,$C7,$A7,$8B,$57,$E8,$FE,$2D,$3F,$C3,$3A,$3F,$43,$FF
	.dc.b	$58,$79,$4F,$D4,$9F,$A2,$3E,$F1,$F9,$EF,$1C,$58,$F1,$E6,$63,$C7
	.dc.b	$0E,$1F,$9F,$EE,$BB,$7E,$2B,$ED,$C3,$1A,$7F,$AF,$DF,$84,$F8,$B4
	.dc.b	$F0,$FC,$4F,$C2,$8C,$63,$FB,$CA,$E3,$9F,$82,$FC,$FA,$BF,$6E,$F3
	.dc.b	$7E,$35,$FE,$2A,$FC,$DD,$F9,$4B,$F0,$E7,$E9,$BF,$EB,$8F,$87,$DF
	.dc.b	$52,$F7,$5A,$F7,$DE,$8E,$A7,$D5,$D7,$FE,$17,$CF,$FC,$6F,$FB,$FE
	.dc.b	$38,$7F,$C6,$31,$E1,$D7,$C1,$AF,$F5,$C9,$E5,$18,$F1,$7F,$C9,$78
	.dc.b	$73,$FD,$BD,$F8,$BB,$F2,$DF,$EE,$7F,$8B,$7F,$32,$FC,$DF,$DB,$5A
	.dc.b	$9C,$FC,$23,$F5,$DC,$E1,$E3,$D3,$FC,$AF,$B3,$07,$7E,$67,$FD,$EB
	.dc.b	$1C,$FE,$85,$A7,$F9,$7E,$78,$CE,$63,$FC,$E3,$1D,$B3,$F0,$FC,$9F
	.dc.b	$E1,$7F,$22,$FF,$29,$FE,$AE,$F8,$CF,$F2,$5F,$EE,$AF,$CC,$FF,$9B
	.dc.b	$F1,$FF,$8F,$EF,$75,$A9,$FE,$97,$39,$FC,$05,$F1,$1E,$7B,$EB,$A7
	.dc.b	$E9,$EB,$59,$77,$FB,$B6,$7F,$A3,$AF,$F3,$1D,$F6,$9C,$FA,$3F,$42
	.dc.b	$73,$EC,$BF,$43,$7E,$E3,$FE,$7E,$FD,$37,$F9,$5F,$DB,$7B,$6B,$53
	.dc.b	$9F,$84,$7E,$8B,$9F,$03,$CD,$5C,$F2,$F3,$1D,$B2,$EA,$B2,$FF,$19
	.dc.b	$B9,$C7,$C1,$C1,$E5,$8F,$F4,$7E,$DF,$C9,$2F,$8D,$FF,$3E,$FD,$33
	.dc.b	$F6,$9F,$99,$7F,$B4,$3C,$A7,$E5,$4F,$C7,$9F,$99,$3F,$73,$F8,$BA
	.dc.b	$FF,$9B,$F0,$E1,$7B,$AD,$7E,$95,$D3,$F1,$9F,$D0,$EF,$EA,$3F,$D4
	.dc.b	$37,$DD,$C0,$79,$E5,$E0,$FC,$43,$D5,$39,$A6,$9E,$6A,$F2,$E3,$AC
	.dc.b	$38,$BE,$7F,$AA,$FE,$21,$F8,$87,$D5,$EB,$7E,$A2,$FC,$0F,$FA,$4F
	.dc.b	$E9,$9E,$DE,$1F,$F4,$55,$EE,$B5,$F1,$CA,$7E,$E1,$D3,$9F,$FC,$97
	.dc.b	$C3,$EA,$C3,$B3,$63,$F1,$78,$72,$FA,$BF,$5D,$75,$CF,$FA,$CF,$E8
	.dc.b	$5E,$EE,$67,$ED,$EF,$DA,$FF,$C5,$3F,$84,$3F,$77,$7D,$AF,$D2,$DB
	.dc.b	$FD,$6F,$87,$0B,$DD,$6B,$AF,$F9,$84,$F4,$FD,$33,$F7,$0E,$EF,$E8
	.dc.b	$BC,$7D,$7F,$F5,$2D,$F6,$9E,$39,$F8,$E7,$3F,$45,$FC,$43,$F1,$6F
	.dc.b	$F6,$3F,$C7,$BE,$49,$EC,$BF,$36,$FC,$5F,$E6,$9F,$DB,$7B,$3F,$D8
	.dc.b	$7A,$7F,$B8,$B0,$E1,$7B,$D9,$6F,$9A,$FE,$D9,$87,$0C,$3F,$B4,$E4
	.dc.b	$F2,$7F,$95,$E9,$E2,$8F,$0C,$FE,$F6,$31,$97,$5C,$5F,$99,$FA,$6B
	.dc.b	$C5,$7E,$8D,$7C,$FF,$65,$99,$F1,$7E,$70,$FD,$29,$FA,$AF,$FB,$73
	.dc.b	$C0,$7E,$86,$FC,$AB,$FC,$05,$BF,$FB,$AB,$89,$8F,$1C,$38,$5B,$C0
	.dc.b	$DE,$63,$BF,$A9,$EC,$71,$87,$2D,$31,$1E,$47,$C7,$97,$F1,$1C,$3D
	.dc.b	$15,$BC,$7C,$DE,$46,$3F,$0B,$79,$FF,$E2,$6F,$DA,$7F,$AD,$FB,$FF
	.dc.b	$90,$E9,$F9,$D8,$FF,$B1,$EB,$5F,$15,$FC,$CF,$CC,$F3,$EF,$F5,$72
	.dc.b	$F5,$3F,$3C,$F3,$BD,$9C,$1B,$AE,$67,$8C,$EB,$F9,$5F,$A3,$F4,$E7
	.dc.b	$E6,$2F,$DD,$5F,$4B,$FD,$C3,$F5,$07,$EE,$FE,$2D,$98,$70,$BD,$EC
	.dc.b	$B7,$E5,$BB,$A7,$E1,$FF,$2B,$0E,$39,$D9,$E5,$9F,$F7,$EF,$4B,$C9
	.dc.b	$3E,$38,$FF,$00,$FE,$65,$F3,$BF,$80,$3F,$DA,$7E,$AB,$F1,$1F,$EC
	.dc.b	$BD,$8F,$CB,$7C,$D5,$A9,$CF,$E9,$78,$F6,$3C,$53,$FE,$DC,$E7,$F8
	.dc.b	$FB,$A3,$87,$CA,$7E,$16,$79,$FC,$C7,$F1,$AC,$7F,$B9,$FE,$8D,$F8
	.dc.b	$0F,$57,$FC,$C3,$FA,$A3,$F2,$FF,$E9,$BF,$94,$BF,$60,$F9,$15,$A9
	.dc.b	$CF,$EF,$D8,$FC,$63,$E2,$4F,$F9,$AB,$4F,$1B,$BF,$2D,$F2,$5D,$FE
	.dc.b	$3D,$A7,$79,$FF,$1E,$DF,$D5,$FC,$57,$FD,$EF,$F9,$CF,$FA,$9F,$F5
	.dc.b	$D7,$E9,$CF,$DA,$75,$C8,$AD,$4B,$F1,$FE,$5D,$3F,$5F,$1F,$8D,$97
	.dc.b	$C3,$38,$DF,$5F,$4B,$F4,$33,$E0,$FC,$09,$A7,$F3,$6F,$3F,$F3,$BF
	.dc.b	$BB,$FE,$0B,$D7,$7F,$00,$BF,$41,$7C,$E3,$F1,$CF,$C7,$5E,$EB,$52
	.dc.b	$FD,$53,$3E,$6F,$E2,$E9,$F1,$F6,$3F,$80,$DF,$5F,$83,$FD,$AD,$73
	.dc.b	$CB,$E3,$E9,$FB,$0F,$9D,$FE,$09,$F5,$AF,$CF,$9F,$A2,$7F,$99,$BE
	.dc.b	$4B,$F4,$15,$A9,$CF,$DE,$BE,$4B,$3F,$0B,$F9,$DA,$FE,$29,$D5,$F8
	.dc.b	$C7,$86,$3F,$EB,$38,$DE,$5D,$E3,$FC,$F3,$1F,$AD,$FD,$43,$FC,$2B
	.dc.b	$FB,$C3,$F7,$A7,$E6,$4F,$D6,$7F,$07,$C1,$FC,$22,$F7,$5A,$97,$E7
	.dc.b	$DC,$FC,$6F,$D8,$57,$CE,$CB,$F3,$EF,$0C,$7F,$A8,$63,$CE,$99,$78
	.dc.b	$67,$FC,$F5,$3F,$D0,$F6,$7E,$95,$FF,$00,$7F,$18,$7E,$F0,$FD,$19
	.dc.b	$D0,$FC,$0B,$F8,$CE,$1C,$3F,$21,$FF,$7E,$5E,$EB,$FA,$7F,$BB,$FA
	.dc.b	$FA,$EB,$EB,$FF,$30,$78,$70,$FF,$1C,$79,$67,$E0,$CF,$CB,$7C,$74
	.dc.b	$F2,$65,$E3,$FC,$1B,$FA,$66,$0F,$D3,$DF,$A2,$7F,$BC,$BF,$76,$FB
	.dc.b	$FF,$8C,$BF,$32,$F3,$D6,$A1,$3A,$FD,$63,$3F,$BC,$F9,$6B,$9F,$07
	.dc.b	$95,$E1,$5F,$8A,$FC,$77,$C3,$18,$EE,$FD,$CD,$7C,$4F,$E0,$07,$CA
	.dc.b	$BF,$2A,$FF,$42,$7D,$03,$F6,$CF,$3F,$67,$D1,$BD,$D6,$A5,$FA,$7F
	.dc.b	$3F,$CE,$BE,$45,$74,$CB,$E4,$3C,$3F,$08,$FF,$03,$6D,$F1,$23,$19
	.dc.b	$F9,$33,$ED,$A7,$8E,$FF,$C1,$99,$B8,$3E,$DF,$EA,$EF,$AF,$FF,$1E
	.dc.b	$FE,$AB,$FE,$B3,$FD,$75,$FA,$A6,$A7,$3E,$62,$5D,$9C,$BE,$DF,$EF
	.dc.b	$63,$CF,$8B,$EF,$3D,$7E,$55,$9F,$FA,$FF,$E1,$7E,$DC,$E6,$3F,$D8
	.dc.b	$5F,$9D,$FD,$EF,$AA,$FF,$20,$FC,$2F,$F9,$3B,$EF,$BF,$23,$18,$FA
	.dc.b	$5F,$F2,$67,$DC,$63,$EA,$7F,$1D,$3F,$E9,$2D,$7F,$1C,$EF,$E7,$FA
	.dc.b	$6F,$FE,$5A,$CB,$FC,$93,$5D,$23,$18,$FF,$95,$FD,$7F,$CF,$FB,$FE
	.dc.b	$9F,$D8,$BF,$D2,$DF,$8F,$BF,$26,$CF,$47,$F9,$92,$73,$E0,$F0,$CB
	.dc.b	$F4,$18,$7C,$3C,$FF,$98,$77,$FE,$C6,$3A,$6E,$FC,$B3,$F3,$B1,$8F
	.dc.b	$CC,$FF,$3F,$F8,$67,$E3,$3C,$37,$E4,$2F,$DB,$1F,$A7,$7E,$E3,$3F
	.dc.b	$DE,$5E,$EB,$53,$FB,$B9,$CF,$91,$FB,$3A,$FC,$66,$BE,$4F,$57,$F7
	.dc.b	$11,$C7,$1F,$ED,$17,$D3,$C8,$FF,$2A,$65,$E1,$F6,$3C,$B1,$8A,$F1
	.dc.b	$EB,$FC,$1B,$E7,$FF,$6B,$C9,$FB,$0F,$C4,$5F,$A7,$FF,$B7,$B3,$FE
	.dc.b	$25,$5A,$F4,$BF,$DE,$12,$FB,$EE,$5E,$67,$F0,$9A,$FD,$AF,$0F,$9C
	.dc.b	$FF,$20,$7B,$08,$E5,$F9,$DF,$A3,$8F,$F8,$94,$75,$D3,$F4,$91,$8F
	.dc.b	$65,$B7,$97,$47,$EA,$3F,$C7,$BE,$FD,$C9,$F3,$BE,$C5,$F8,$2F,$97
	.dc.b	$F4,$FD,$4E,$63,$D1,$F3,$38,$E1,$FE,$3D,$83,$F5,$BF,$E4,$4F,$4E
	.dc.b	$DF,$F1,$75,$FF,$81,$2B,$9C,$62,$3F,$8F,$CE,$F2,$F0,$F2,$BF,$45
	.dc.b	$FF,$25,$FE,$96,$FC,$07,$F9,$27,$BC,$FD,$43,$C1,$EC,$3F,$66,$DD
	.dc.b	$6A,$73,$F8,$A6,$3F,$06,$FF,$32,$E3,$CF,$E1,$2D,$EF,$8B,$F7,$FC
	.dc.b	$5C,$A7,$F8,$F6,$3E,$DF,$B7,$FF,$40,$7E,$11,$FE,$A3,$F9,$FF,$F5
	.dc.b	$87,$E1,$9F,$EE,$0E,$EF,$F5,$7F,$E1,$9F,$EA,$39,$CF,$A7,$FE,$64
	.dc.b	$BF,$F5,$76,$7F,$E6,$07,$FD,$5C,$FD,$F3,$FD,$DB,$5C,$7E,$C3,$FA
	.dc.b	$FE,$1D,$3E,$7F,$FD,$73,$F8,$5F,$FA,$47,$E3,$BE,$EF,$E5,$0F,$D4
	.dc.b	$DF,$72,$FC,$71,$FA,$BB,$F4,$EF,$F3,$AE,$1C,$2F,$75,$AF,$9E,$CF
	.dc.b	$95,$F2,$F8,$3F,$CE,$B7,$D9,$B0,$CF,$C1,$9F,$7F,$33,$C9,$39,$CB
	.dc.b	$AF,$A5,$FE,$99,$3F,$77,$87,$FA,$B3,$F1,$0F,$F5,$87,$DF,$FF,$71
	.dc.b	$9C,$CF,$FA,$EA,$73,$03,$3E,$2F,$FB,$2F,$C7,$FF,$1F,$57,$5F,$27
	.dc.b	$FD,$76,$E2,$B7,$F8,$87,$FA,$E7,$A7,$FD,$93,$F5,$7B,$AF,$17,$BB
	.dc.b	$FE,$BC,$7B,$61,$FE,$6B,$FC,$B9,$F6,$4F,$C3,$7F,$DF,$DE,$23,$E2
	.dc.b	$FF,$D9,$DF,$A6,$F0,$C7,$FD,$6D,$87,$0B,$DD,$7F,$B4,$FF,$5B,$D5
	.dc.b	$2B,$4F,$9F,$2F,$06,$4F,$D7,$FD,$B3,$C2,$BA,$7E,$2D,$FE,$05,$C7
	.dc.b	$FC,$8B,$87,$0C,$BB,$ED,$ED,$C3,$FD,$CD,$E1,$FF,$9D,$7D,$BF,$F3
	.dc.b	$F7,$1F,$CE,$E7,$4E,$7F,$02,$F2,$C6,$3F,$02,$F0,$D6,$EB,$FD,$01
	.dc.b	$87,$F3,$EC,$FF,$DB,$51,$FE,$81,$FC,$0B,$C1,$87,$4F,$C0,$BF,$D0
	.dc.b	$3F,$81,$BE,$BF,$CF,$8F,$FB,$6F,$F2,$FF,$F0,$F7,$ED,$5F,$2F,$F5
	.dc.b	$7F,$F7,$37,$E5,$0F,$CC,$7E,$3D,$6B,$95,$5A,$FA,$5C,$F6,$B1,$F3
	.dc.b	$F0,$FF,$A8,$F0,$FF,$15,$6B,$8D,$3C,$17,$BC,$B9,$FE,$35,$FE,$4C
	.dc.b	$C3,$84,$E7,$3F,$F7,$C7,$EB,$DF,$2F,$2F,$F1,$A7,$EB,$9F,$81,$FC
	.dc.b	$D7,$FD,$F9,$EC,$BF,$5A,$F1,$CE,$73,$4B,$F0,$8C,$ED,$F7,$CF,$1F
	.dc.b	$EB,$7E,$55,$97,$6D,$D9,$F6,$9E,$B1,$8C,$7F,$25,$18,$9F,$1F,$E1
	.dc.b	$9F,$13,$F4,$8F,$CB,$7A,$CB,$7E,$62,$FC,$FF,$59,$7F,$05,$9C,$FE
	.dc.b	$6D,$C7,$3F,$31,$E4,$E4,$7E,$0D,$E7,$DB,$FC,$AD,$A7,$6C,$3F,$A4
	.dc.b	$DF,$4D,$3C,$19,$7C,$9C,$7F,$DF,$0F,$F9,$D2,$BF,$DC,$9D,$5F,$6B
	.dc.b	$F1,$DF,$D6,$7D,$E7,$9F,$F1,$5F,$F1,$AF,$89,$FC,$13,$5F,$F3,$25
	.dc.b	$78,$72,$FF,$0B,$69,$E6,$9C,$E7,$D7,$E4,$7C,$17,$D3,$4F,$EA,$58
	.dc.b	$F3,$C7,$FC,$B5,$6F,$D3,$FE,$37,$D0,$FD,$BA,$D7,$C8,$FE,$FE,$73
	.dc.b	$D3,$F9,$08,$F1,$7D,$A7,$8D,$D6,$5F,$0A,$F8,$E3,$D6,$3C,$3E,$07
	.dc.b	$E3,$AB,$79,$7F,$B8,$2F,$C9,$FA,$17,$9E,$3E,$B4,$63,$0F,$91,$0E
	.dc.b	$7F,$D0,$D7,$C3,$FD,$97,$E2,$3E,$FE,$0F,$A0,$FC,$1F,$2B,$F3,$5F
	.dc.b	$22,$FD,$53,$EC,$7C,$DF,$EB,$F4,$F9,$D3,$9D,$3F,$09,$3F,$4B,$6F
	.dc.b	$D6,$4F,$1E,$6F,$A0,$BF,$05,$FC,$23,$C7,$AF,$F2,$19,$F8,$33,$FC
	.dc.b	$36,$BE,$2F,$9E,$FD,$37,$E4,$FF,$3F,$3E,$B1,$8E,$6F,$E0,$39,$7F
	.dc.b	$89,$34,$FF,$1B,$D7,$8B,$E8,$BB,$DF,$EB,$FD,$6E,$9C,$1F,$A5,$F6
	.dc.b	$6F,$8D,$FA,$1D,$3E,$76,$3E,$AF,$91,$FE,$21,$D4,$FB,$5F,$99,$CD
	.dc.b	$F0,$A7,$31,$F4,$AB,$86,$DF,$03,$EB,$7D,$F5,$CB,$F2,$5F,$9A,$F4
	.dc.b	$FC,$A7,$F9,$97,$D3,$3E,$BF,$E3,$78,$2F,$D1,$9F,$6A,$FA,$47,$E9
	.dc.b	$1F,$D1,$B0,$E1,$DE,$C3,$85,$EF,$DB,$D6,$5F,$E3,$AD,$FF,$6F,$87
	.dc.b	$06,$C7,$83,$3E,$FC,$FE,$48,$C6,$1F,$C5,$AB,$9F,$E4,$FD,$FE,$FF
	.dc.b	$FD,$97,$F0,$6F,$AA,$CE,$BF,$EA,$5F,$AD,$4B,$EA,$BC,$7F,$93,$7F
	.dc.b	$1B,$DF,$E0,$65,$75,$3D,$70,$FE,$33,$D5,$FE,$00,$AD,$E5,$D7,$DB
	.dc.b	$F3,$78,$5F,$07,$83,$E3,$FC,$53,$E8,$3E,$0C,$E6,$53,$FE,$88,$C7
	.dc.b	$A6,$9F,$CC,$78,$BE,$1D,$F8,$CE,$1C,$7D,$B3,$FA,$78,$FE,$E7,$B7
	.dc.b	$FD,$FD,$87,$FB,$82,$5F,$75,$FE,$EE,$CD,$F8,$C7,$D2,$FB,$3F,$F5
	.dc.b	$87,$0F,$8B,$F3,$0F,$AB,$F6,$DF,$06,$66,$3C,$AB,$E0,$E7,$DA,$78
	.dc.b	$F9,$9F,$57,$C3,$FF,$5B,$E7,$F9,$5E,$2F,$5C,$63,$EF,$BF,$03,$F7
	.dc.b	$DF,$36,$26,$7E,$6E,$1E,$BC,$9F,$E8,$AE,$2F,$F6,$66,$7F,$81,$FD
	.dc.b	$97,$CF,$A3,$F4,$1F,$C5,$C6,$3E,$8B,$F1,$BF,$82,$FD,$98,$99,$FA
	.dc.b	$79,$F9,$3E,$BC,$6F,$3F,$C5,$FD,$03,$F2,$BF,$87,$7D,$5F,$BE,$FC
	.dc.b	$EF,$27,$D4,$8C,$78,$DF,$9A,$FB,$CF,$C3,$44,$CB,$E9,$E7,$F9,$78
	.dc.b	$FB,$D9,$FE,$67,$F2,$AF,$C5,$7E,$14,$FC,$3B,$F4,$71,$8F,$C3,$FF
	.dc.b	$4B,$F9,$4F,$E4,$62,$65,$F9,$9C,$B9,$3F,$9B,$C7,$F4,$9F,$AE,$7E
	.dc.b	$BF,$CD,$FD,$AF,$EF,$3F,$8F,$E0,$FC,$5C,$E7,$AD,$3F,$AB,$FC,$1B
	.dc.b	$F3,$D3,$31,$FD,$BE,$3D,$1F,$DC,$E5,$72,$EB,$E0,$FF,$91,$3E,$45
	.dc.b	$F8,$CB,$F7,$CF,$D6,$FE,$EF,$ED,$FC,$E3,$FC,$F7,$C1,$FE,$66,$BD
	.dc.b	$D6,$B9,$14,$FC,$87,$A6,$3C,$FC,$3F,$EF,$78,$76,$FD,$FB,$FC,$20
	.dc.b	$F2,$F5,$1A,$7C,$BF,$43,$FC,$95,$39,$8F,$9D,$C1,$FE,$4D,$FC,$0F
	.dc.b	$FD,$55,$F1,$FF,$EB,$3F,$C0,$9F,$83,$7F,$AC,$F8,$BF,$CF,$75,$AF
	.dc.b	$C0,$FF,$D6,$71,$8F,$C1,$1F,$15,$FE,$21,$AF,$F5,$A7,$07,$F8,$41
	.dc.b	$E1,$8F,$FA,$D2,$F8,$7C,$AF,$FA,$D2,$3F,$8C,$47,$FA,$23,$F1,$47
	.dc.b	$D3,$FF,$BC,$BE,$4B,$FB,$E7,$DB,$BF,$61,$F2,$76,$34,$98,$70,$FC
	.dc.b	$2F,$FD,$CB,$F3,$5F,$EF,$5B,$FF,$4F,$67,$D7,$37,$CB,$FF,$8D,$F0
	.dc.b	$E1,$8F,$F8,$99,$E5,$A7,$83,$0F,$F3,$94,$63,$1F,$F8,$01,$8F,$F9
	.dc.b	$D7,$F6,$C7,$EC,$BF,$E3,$4F,$CB,$DF,$87,$BF,$11,$7E,$11,$FE,$38
	.dc.b	$C3,$FD,$4B,$87,$0F,$A8,$61,$C2,$F7,$E8,$FF,$A3,$AB,$5B,$FF,$8E
	.dc.b	$9E,$FF,$83,$FF,$93,$DE,$57,$C7,$2F,$26,$9E,$3D,$3C,$5D,$9F,$E8
	.dc.b	$BF,$BE,$FF,$70,$7D,$B3,$F0,$C7,$C1,$7F,$A7,$27,$30,$F9,$6C,$7E
	.dc.b	$16,$FC,$97,$FC,$BD,$5F,$EE,$0E,$0F,$EF,$67,$D3,$78,$BC,$3F,$0C
	.dc.b	$FF,$6F,$BF,$EA,$8F,$D4,$1F,$BC,$3F,$46,$7E,$AB,$FE,$9A,$E2,$7C
	.dc.b	$73,$E5,$38,$5E,$EB,$53,$FE,$63,$EB,$FF,$67,$E3,$D3,$1F,$F3,$93
	.dc.b	$DB,$E2,$6E,$1C,$78,$FE,$9F,$DC,$7F,$AB,$6B,$86,$5E,$1F,$C3,$3F
	.dc.b	$D6,$1F,$87,$7F,$BB,$71,$FF,$79,$7E,$21,$FE,$F1,$D3,$FD,$BB,$39
	.dc.b	$ED,$FF,$78,$46,$3F,$0E,$FF,$78,$71,$7F,$AB,$4F,$CB,$57,$FB,$C5
	.dc.b	$C0,$79,$63,$E0,$9F,$F7,$86,$3F,$F0,$06,$FF,$D8,$5F,$5E,$F9,$67
	.dc.b	$97,$FE,$5F,$FD,$AD,$FB,$F3,$F2,$4F,$F3,$EE,$1F,$EF,$2C,$38,$5E
	.dc.b	$EB,$5F,$75,$FE,$DA,$9C,$ED,$FE,$86,$7B,$7E,$61,$FE,$9B,$78,$F9
	.dc.b	$4C,$FB,$FC,$7F,$93,$5F,$1C,$E7,$27,$CC,$E7,$4F,$06,$CE,$6F,$06
	.dc.b	$1E,$AF,$D7,$78,$7E,$19,$E8,$BF,$DE,$6E,$DB,$D3,$5A,$E0,$E3,$48
	.dc.b	$F0,$7E,$8F,$DD,$F1,$7C,$6D,$E1,$F8,$B9,$78,$78,$FE,$26,$9F,$2B
	.dc.b	$F7,$6F,$47,$E6,$7E,$1D,$BF,$86,$6D,$F2,$27,$3B,$78,$1F,$77,$A9
	.dc.b	$ED,$7C,$35,$9C,$FE,$9E,$1D,$F7,$F5,$E3,$F4,$BE,$A3,$E3,$7E,$B3
	.dc.b	$DF,$0F,$76,$9F,$72,$73,$A7,$B9,$F9,$9A,$FC,$B7,$C3,$59,$CF,$F0
	.dc.b	$38,$77,$8F,$4E,$1F,$86,$FA,$DF,$A9,$FB,$57,$7C,$FF,$07,$18,$CF
	.dc.b	$E5,$BD,$AE,$E3,$F8,$CD,$FE,$D4,$F2,$E4,$FA,$3E,$BF,$AB,$F0,$3B
	.dc.b	$7E,$2C,$6F,$4B,$83,$F0,$57,$E8,$7E,$8F,$0F,$FB,$6A,$FD,$38,$FD
	.dc.b	$CF,$D6,$3E,$9E,$56,$7E,$5E,$57,$2F,$E8,$5A,$DE,$8F,$C7,$61,$F6
	.dc.b	$B1,$FC,$2F,$EC,$3F,$8C,$CA,$E5,$F8,$2C,$AE,$4C,$7F,$2D,$7F,$88
	.dc.b	$CB,$3C,$FF,$2D,$FA,$BA,$3E,$DB,$F3,$DD,$7F,$2F,$97,$F6,$71,$8C
	.dc.b	$FF,$2A,$FE,$8F,$BB,$F4,$4F,$E9,$7B,$6E,$DF,$B8,$C3,$F2,$9D,$FF
	.dc.b	$A0,$F8,$17,$E2,$EF,$C6,$5F,$95,$3D,$3F,$ED,$FF,$58,$FC,$C3,$E7
	.dc.b	$63,$C7,$0E,$17,$BF,$8A,$5F,$D3,$EB,$5E,$FF,$1B,$FD,$7F,$2F,$EC
	.dc.b	$27,$2F,$F6,$DD,$7C,$7A,$EA,$3F,$B8,$4E,$94,$EB,$2F,$07,$DF,$6B
	.dc.b	$1F,$26,$1E,$AD,$BF,$C0,$FF,$7D,$9C,$7B,$6D,$EF,$F3,$BD,$1F,$B8
	.dc.b	$7F,$8D,$F1,$E3,$18,$AF,$79,$C3,$1F,$E5,$58,$FF,$84,$B2,$E9,$3E
	.dc.b	$4C,$7A,$4F,$F8,$5B,$E3,$F3,$AF,$B3,$F4,$6F,$8D,$DF,$F2,$F8,$3D
	.dc.b	$93,$9B,$FF,$0E,$CF,$C9,$D7,$D7,$7E,$EC,$BC,$67,$7D,$11,$F4,$72
	.dc.b	$EF,$9F,$DB,$F9,$FF,$A9,$F8,$B7,$B7,$0F,$54,$61,$E7,$A7,$D0,$8E
	.dc.b	$DA,$F9,$2F,$96,$BE,$BB,$DC,$7E,$1F,$F1,$DE,$BC,$F8,$D5,$1E,$0C
	.dc.b	$30,$F4,$DB,$EA,$BE,$7C,$38,$7E,$D5,$E6,$FB,$8B,$9F,$F8,$FB,$6F
	.dc.b	$5E,$4A,$E5,$7F,$4D,$E5,$97,$9E,$5F,$25,$F3,$B4,$F2,$3F,$42,$6E
	.dc.b	$3F,$67,$0B,$D4,$E1,$EF,$E6,$7E,$13,$5F,$63,$F8,$3C,$3B,$3E,$2B
	.dc.b	$FA,$4F,$8B,$7F,$CF,$57,$E8,$FA,$FE,$F4,$63,$2F,$D5,$E1,$F2,$BF
	.dc.b	$48,$FC,$F6,$3F,$33,$B3,$FD,$AD,$5E,$5A,$FA,$51,$8D,$1D,$7F,$90
	.dc.b	$FC,$6D,$FB,$0B,$F6,$07,$EE,$8E,$67,$E8,$3F,$E3,$0F,$C2,$BF,$13
	.dc.b	$87,$0B,$DD,$6B,$E3,$9F,$7F,$AC,$9F,$61,$E6,$DF,$F6,$58,$F1,$E1
	.dc.b	$FD,$B4,$62,$BC,$5A,$F8,$75,$FE,$EB,$5C,$F4,$EF,$1F,$EE,$DB,$FD
	.dc.b	$E6,$CF,$39,$F9,$07,$FB,$D3,$E7,$9F,$44,$E4,$7E,$47,$FA,$2C,$38
	.dc.b	$5E,$EB,$5F,$91,$E7,$1E,$7A,$F9,$BA,$7F,$BF,$E1,$DF,$F2,$3F,$64
	.dc.b	$62,$BC,$3F,$7E,$D7,$C3,$7F,$BE,$3B,$C6,$72,$79,$5F,$C0,$BF,$21
	.dc.b	$7E,$92,$FD,$3F,$FD,$69,$FB,$7B,$F0,$C7,$73,$F6,$FF,$DD,$E1,$C2
	.dc.b	$F7,$5A,$FC,$3F,$38,$F3,$E2,$F5,$1F,$E9,$3B,$EF,$F9,$D7,$AA,$31
	.dc.b	$3E,$1F,$B8,$69,$E1,$FC,$57,$CD,$39,$8F,$27,$E6,$1F,$C6,$BC,$47
	.dc.b	$CC,$3F,$28,$7E,$8E,$FC,$A1,$B3,$C5,$FF,$45,$F8,$3F,$CD,$71,$E3
	.dc.b	$87,$0B,$DD,$3E,$E3,$F8,$CD,$76,$FC,$AB,$D1,$C3,$F9,$7D,$3C,$0F
	.dc.b	$B8,$C6,$3C,$F3,$7E,$39,$E7,$AF,$36,$BE,$59,$FD,$2D,$FF,$26,$F4
	.dc.b	$9F,$82,$3F,$68,$FA,$BF,$89,$BF,$07,$7E,$3B,$F5,$F8,$FF,$A7,$61
	.dc.b	$C2,$F7,$5A,$D3,$A6,$2F,$C9,$7D,$37,$BC,$FD,$0F,$1E,$76,$9D,$F9
	.dc.b	$BF,$A3,$CE,$63,$C5,$5F,$D2,$BF,$23,$FC,$87,$DF,$7E,$83,$E7,$DF
	.dc.b	$8F,$3F,$22,$7E,$53,$FD,$0A,$F7,$E3,$AF,$74,$FC,$9F,$5E,$5B,$2F
	.dc.b	$9F,$87,$E7,$7E,$1A,$CE,$71,$E5,$9B,$F0,$CF,$24,$E7,$3F,$1F,$5F
	.dc.b	$F6,$0F,$C6,$5F,$28,$FB,$FF,$F7,$47,$A5,$FE,$33,$F8,$57,$E5,$3F
	.dc.b	$D6,$2F,$75,$AE,$EF,$E6,$FF,$94,$E7,$1E,$5C,$DF,$81,$BE,$B9,$FE
	.dc.b	$19,$E3,$F4,$3F,$E0,$CF,$89,$FF,$05,$6B,$FE,$7A,$9C,$BF,$E4,$3F
	.dc.b	$13,$CD,$8F,$96,$FF,$65,$9F,$F4,$6D,$7C,$1A,$FF,$BF,$FD,$0F,$46
	.dc.b	$D8,$BF,$DC,$69,$DB,$60,$D7,$C7,$7D,$E3,$87,$E3,$BF,$07,$2B,$9F
	.dc.b	$B8,$FF,$0D,$5F,$AA,$7F,$C5,$95,$E6,$CD,$92,$7A,$5F,$0D,$3D,$5F
	.dc.b	$07,$D6,$31,$7C,$23,$D3,$18,$BF,$E7,$7A,$7A,$71,$F4,$61,$CA,$BF
	.dc.b	$7B,$1E,$FF,$C7,$7A,$70,$7C,$BA,$F9,$11,$D2,$31,$97,$9A,$3D,$35
	.dc.b	$C6,$F9,$9E,$1F,$85,$E7,$1F,$23,$5F,$B3,$1C,$E3,$18,$FD,$58,$E7
	.dc.b	$5C,$4E,$78,$F8,$33,$7E,$35,$D3,$0F,$5E,$7F,$22,$3A,$46,$27,$7A
	.dc.b	$7C,$1C,$FF,$13,$5F,$33,$0E,$9F,$8B,$73,$9F,$C6,$BC,$9B,$5C,$74
	.dc.b	$FC,$54,$F1,$C7,$D3,$97,$D8,$F8,$7E,$55,$F8,$F7,$93,$6F,$94,$EF
	.dc.b	$4F,$6D,$FE,$33,$ED,$DF,$8E,$B7,$77,$3F,$14,$FE,$37,$EF,$BF,$23
	.dc.b	$FF,$21,$E1,$C2,$F7,$5A,$E0,$CE,$3C,$F1,$E5,$E6,$7F,$6C,$CF,$38
	.dc.b	$70,$8F,$E0,$11,$89,$7E,$B5,$E0,$CF,$0F,$5D,$7F,$CE,$9A,$78,$6B
	.dc.b	$78,$75,$9C,$E5,$E1,$BF,$E6,$7E,$0F,$F9,$A3,$0E,$1F,$91,$75,$BD
	.dc.b	$E9,$F0,$23,$B7,$2F,$FB,$DA,$B7,$97,$58,$C6,$5E,$DA,$EB,$AF,$A6
	.dc.b	$BA,$56,$B0,$F9,$71,$D2,$7E,$85,$FC,$58,$FE,$8F,$E2,$79,$63,$11
	.dc.b	$F1,$32,$F2,$EA,$FD,$CF,$AE,$1C,$B1,$FA,$F1,$D6,$1A,$7C,$58,$EB
	.dc.b	$1F,$35,$F3,$DF,$4F,$5B,$9C,$F0,$CB,$EC,$47,$3B,$F9,$CF,$2A,$E1
	.dc.b	$97,$97,$4F,$76,$7F,$DD,$E1,$C7,$F8,$7A,$F4,$FB,$5C,$F2,$FC,$23
	.dc.b	$CF,$0E,$2F,$91,$E7,$FA,$67,$93,$3F,$BC,$F1,$DB,$F2,$AF,$1D,$7E
	.dc.b	$16,$88,$E5,$1F,$CA,$3F,$42,$FF,$3A,$FC,$0B,$F0,$8F,$F2,$DF,$EA
	.dc.b	$9F,$E9,$AD,$FF,$D3,$B5,$A9,$CF,$CF,$3F,$20,$CF,$97,$FE,$29,$CB
	.dc.b	$FC,$0D,$7D,$1B,$0C,$BB,$7C,$53,$3F,$D7,$D2,$FB,$4E,$7F,$39,$FF
	.dc.b	$5D,$FE,$2E,$FC,$69,$F1,$1F,$E1,$9F,$C6,$1F,$8E,$BF,$1D,$7E,$07
	.dc.b	$FE,$B4,$C3,$85,$EE,$B5,$F8,$EE,$71,$E7,$AF,$F8,$76,$FB,$7E,$3B
	.dc.b	$FE,$22,$78,$FC,$5B,$4E,$FF,$8E,$F9,$30,$F1,$CE,$72,$FD,$16,$3F
	.dc.b	$D5,$3F,$22,$7D,$A3,$F5,$57,$E9,$4F,$D0,$FE,$57,$E8,$9A,$F1,$7F
	.dc.b	$DD,$98,$70,$BD,$D6,$BD,$7F,$F6,$7C,$E7,$2F,$F1,$A3,$DB,$CD,$FF
	.dc.b	$1C,$BC,$72,$ED,$DA,$E8,$FF,$72,$CE,$71,$FB,$5F,$7B,$FE,$E8,$FC
	.dc.b	$87,$FD,$61,$F0,$5F,$E1,$5F,$C8,$9F,$91,$3F,$1A,$7E,$77,$FE,$E0
	.dc.b	$BD,$FC,$4A,$F7,$4F,$CE,$F5,$F1,$EF,$AC,$FF,$26,$E1,$DB,$8F,$FC
	.dc.b	$98,$F2,$C3,$8C,$70,$D3,$C5,$A3,$B3,$FD,$FB,$EB,$BF,$60,$7E,$51
	.dc.b	$F8,$9F,$9D,$7D,$8B,$F5,$27,$E9,$7F,$EF,$6C,$38,$5E,$EB,$5F,$A9
	.dc.b	$E7,$1F,$C5,$EB,$FE,$5F,$BD,$B6,$3B,$F9,$BF,$E6,$FB,$73,$F9,$27
	.dc.b	$3A,$78,$BF,$3D,$FF,$0A,$F8,$2F,$67,$FC,$2F,$93,$F2,$5F,$D8,$7E
	.dc.b	$18,$FC,$95,$F9,$2B,$F2,$77,$E4,$2A,$65,$CF,$1F,$F1,$0E,$3C,$B1
	.dc.b	$E3,$87,$0F,$C9,$F7,$9F,$82,$BB,$DF,$F9,$ED,$FE,$01,$F8,$EF,$FA
	.dc.b	$01,$E5,$A7,$87,$4F,$4D,$F1,$F7,$3C,$59,$F6,$F0,$73,$97,$F8,$CB
	.dc.b	$F5,$2F,$F4,$BF,$8A,$FC,$63,$FD,$5D,$F6,$BF,$EC,$6D,$3C,$77,$BF
	.dc.b	$A7,$FD,$15,$6B,$EE,$1F,$AC,$6A,$99,$74,$7F,$D2,$4F,$FA,$93,$ED
	.dc.b	$7F,$D2,$EF,$82,$FB,$F3,$F9,$2B,$8E,$3D,$BE,$75,$A7,$49,$ED,$F7
	.dc.b	$EF,$C9,$DF,$93,$3E,$1D,$FA,$0F,$FA,$CB,$F2,$AF,$F0,$17,$6F,$D1
	.dc.b	$C5,$8F,$1C,$38,$5E,$FF,$2A,$EB,$93,$F2,$AF,$F5,$6D,$EF,$0F,$F5
	.dc.b	$63,$C7,$3E,$FB,$34,$F5,$4E,$72,$FF,$61,$F5,$FB,$7F,$73,$FF,$72
	.dc.b	$7E,$FC,$FD,$F9,$F8,$F3,$F3,$4F,$EC,$9E,$7F,$CB,$BD,$D6,$A7,$FB
	.dc.b	$47,$E5,$99,$F4,$DF,$AD,$FF,$B0,$AF,$AF,$0F,$FB,$05,$E3,$97,$5F
	.dc.b	$CC,$3C,$56,$D3,$C5,$3E,$1E,$CF,$A5,$E6,$B3,$7E,$5E,$FC,$BB,$D7
	.dc.b	$F9,$87,$F7,$4E,$CF,$99,$7B,$AD,$73,$FE,$8A,$BF,$D6,$35,$AC,$39
	.dc.b	$EF,$FE,$D5,$C3,$B7,$E5,$BF,$ED,$A7,$96,$5F,$23,$F3,$0F,$1D,$F8
	.dc.b	$A3,$E5,$FE,$61,$F8,$7F,$CF,$FF,$A5,$53,$F7,$E7,$E2,$E9,$FF,$18
	.dc.b	$FE,$6D,$F9,$57,$73,$A7,$FC,$6B,$FC,$03,$9F,$9A,$61,$FE,$EC,$9F
	.dc.b	$EA,$DF,$BE,$D9,$CD,$F1,$FD,$3B,$C5,$18,$CB,$C1,$D3,$F9,$BF,$C8
	.dc.b	$DF,$9B,$3C,$C7,$E9,$0E,$47,$B3,$FD,$8B,$F4,$3F,$D5,$61,$C2,$F7
	.dc.b	$5A,$FA,$07,$E9,$3A,$CB,$F9,$EF,$4F,$FB,$FF,$0E,$0D,$8F,$06,$7D
	.dc.b	$F3,$73,$79,$67,$39,$78,$BA,$BF,$65,$F9,$23,$F0,$2F,$EE,$DE,$93
	.dc.b	$AB,$FD,$59,$FC,$02,$FD,$F3,$F8,$05,$EF,$9B,$F6,$F5,$AF,$C5,$73
	.dc.b	$8F,$F9,$D3,$83,$F9,$59,$D9,$FF,$01,$38,$37,$5C,$BF,$02,$F8,$D1
	.dc.b	$8C,$FC,$50,$E9,$FE,$11,$F9,$27,$F7,$8F,$CC,$5F,$9A,$BF,$2F,$7E
	.dc.b	$58,$FC,$D9,$F9,$B7,$F1,$7C,$38,$5E,$EB,$5F,$82,$67,$E9,$D9,$79
	.dc.b	$76,$C7,$C5,$5E,$7F,$E1,$99,$72,$D7,$BC,$79,$30,$E3,$9F,$5F,$C2
	.dc.b	$3F,$A6,$FE,$FD,$FD,$33,$F1,$27,$9F,$FE,$95,$FC,$C1,$F9,$EF,$F8
	.dc.b	$4B,$ED,$9E,$C7,$F4,$AC,$38,$5E,$EB,$5F,$43,$9E,$1F,$F7,$6E,$5D
	.dc.b	$3E,$B3,$CD,$87,$0F,$C1,$ED,$6B,$1E,$5B,$FF,$58,$8E,$43,$0F,$2D
	.dc.b	$35,$F2,$D7,$F1,$FA,$DF,$E6,$2F,$CC,$5F,$98,$3C,$3F,$E5,$FB,$3F
	.dc.b	$22,$7E,$1B,$C2,$F7,$5A,$9F,$EC,$FC,$CE,$0F,$EE,$39,$75,$FC,$AF
	.dc.b	$D3,$A2,$3C,$1A,$61,$B8,$F8,$3F,$28,$F9,$F4,$38,$67,$E0,$F8,$AF
	.dc.b	$F0,$F7,$D5,$EA,$5F,$87,$BF,$2A,$FA,$3F,$AB,$BE,$F5,$FB,$77,$F9
	.dc.b	$D7,$F3,$EF,$8F,$7B,$ED,$5E,$EB,$FC,$F5,$FA,$0E,$B1,$FB,$79,$7E
	.dc.b	$7F,$93,$DC,$F9,$71,$F0,$73,$FB,$9E,$98,$F3,$F8,$9F,$34,$63,$5F
	.dc.b	$1F,$B9,$E9,$FA,$77,$E6,$6F,$CC,$9F,$99,$FD,$1F,$CC,$FF,$DC,$9F
	.dc.b	$41,$F7,$AF,$75,$AE,$E5,$7F,$BA,$7F,$1C,$D6,$2F,$C6,$BE,$8E,$1C
	.dc.b	$38,$7D,$5E,$07,$F8,$9A,$39,$E9,$DF,$F3,$8F,$26,$1E,$38,$F1,$7E
	.dc.b	$5F,$F3,$FF,$62,$7E,$57,$FE,$44,$F7,$5F,$5E,$FD,$A9,$F9,$2F,$F7
	.dc.b	$DE,$BF,$8F,$BB,$1E,$3F,$72,$C7,$8E,$1C,$3E,$27,$F3,$75,$AB,$F3
	.dc.b	$9F,$83,$7C,$3F,$84,$FE,$31,$E7,$7C,$BE,$93,$CD,$AF,$97,$5F,$27
	.dc.b	$CA,$FE,$1F,$F6,$C7,$E7,$0F,$D8,$5F,$99,$3F,$3E,$7C,$AF,$F7,$EF
	.dc.b	$C8,$BF,$5B,$87,$0B,$DD,$6B,$E8,$9F,$88,$6B,$2E,$9C,$5E,$81,$87
	.dc.b	$83,$F5,$1F,$D1,$BC,$F4,$F0,$69,$FC,$E6,$B9,$67,$DA,$3F,$DE,$5D
	.dc.b	$DF,$E5,$CF,$C5,$DF,$9D,$3F,$3B,$FF,$6F,$FA,$2E,$17,$B4,$FD,$DF
	.dc.b	$F8,$2E,$1C,$2F,$75,$AF,$23,$38,$F3,$D7,$D1,$7D,$BE,$47,$F7,$2F
	.dc.b	$1F,$C7,$7F,$8E,$E0,$8F,$D9,$19,$D3,$C5,$ED,$63,$1F,$35,$F0,$D7
	.dc.b	$FD,$E9,$B7,$AE,$73,$7E,$99,$7E,$53,$E4,$DB,$CF,$1F,$BB,$F0,$FF
	.dc.b	$17,$9F,$7D,$3D,$FF,$A8,$6F,$4E,$DF,$86,$DC,$FA,$7E,$2D,$6F,$E2
	.dc.b	$FD,$F5,$CF,$F8,$46,$7D,$36,$F1,$D7,$CA,$F1,$35,$E3,$EF,$E7,$B8
	.dc.b	$7E,$7D,$CF,$B9,$FA,$65,$F9,$AC,$7E,$97,$9F,$CB,$FF,$9A,$EC,$FD
	.dc.b	$F3,$F4,$FF,$92,$FF,$BB,$BE,$D1,$F5,$7F,$C5,$B8,$BF,$4F,$7B,$F2
	.dc.b	$D7,$BA,$7E,$CB,$D3,$1F,$F4,$C7,$83,$F8,$FC,$38,$36,$3B,$FE,$49
	.dc.b	$FE,$44,$BF,$E5,$15,$D7,$0F,$17,$C1,$62,$F7,$A7,$93,$F6,$6E,$DA
	.dc.b	$78,$B2,$E5,$E0,$E3,$1F,$27,$E6,$D6,$7B,$78,$57,$1F,$6F,$EE,$97
	.dc.b	$67,$85,$7F,$1E,$C7,$D5,$E7,$E6,$B5,$E1,$6F,$F1,$0E,$1F,$98,$F1
	.dc.b	$FD,$E7,$97,$C7,$73,$FC,$7B,$A7,$E0,$1F,$D8,$BF,$41,$7E,$7E,$FD
	.dc.b	$11,$C9,$FC,$3F,$F0,$F7,$C7,$BF,$45,$E1,$7B,$AD,$4B,$F2,$EC,$E1
	.dc.b	$FD,$6B,$1E,$BF,$A2,$DC,$38,$7C,$2B,$37,$C5,$7F,$55,$BF,$1E,$1E
	.dc.b	$2A,$D7,$E8,$5F,$D6,$7D,$97,$E0,$7F,$8A,$5F,$9F,$3F,$2F,$FE,$27
	.dc.b	$ED,$FF,$7E,$9D,$FA,$07,$73,$FC,$63,$E0,$F5,$83,$F1,$9F,$EE,$35
	.dc.b	$C3,$CD,$FE,$28,$F2,$38,$FC,$4F,$8E,$BC,$59,$F8,$7E,$FB,$FB,$77
	.dc.b	$C1,$3F,$0A,$7E,$8B,$E6,$FC,$DD,$C8,$F9,$7F,$F7,$77,$AD,$FD,$03
	.dc.b	$0E,$17,$BA,$D7,$C6,$7F,$86,$27,$A7,$E7,$5F,$E0,$67,$7C,$1F,$CA
	.dc.b	$5E,$3A,$B3,$EF,$F7,$FE,$4D,$7C,$73,$9C,$BF,$8A,$61,$FE,$24,$FC
	.dc.b	$59,$F9,$4F,$FB,$43,$E5,$1F,$92,$BE,$23,$FC,$A9,$F6,$DB,$AD,$4E
	.dc.b	$7F,$01,$7E,$9D,$9F,$C7,$3F,$C3,$79,$74,$2D,$F0,$C7,$FC,$BD,$9E
	.dc.b	$F3,$EF,$1F,$BE,$8F,$F1,$BF,$D6,$BF,$3B,$7E,$8D,$F9,$8F,$D1,$5F
	.dc.b	$A3,$7F,$54,$F1,$F7,$5A,$F2,$95,$AF,$9A,$FA,$9D,$BF,$DB,$B2,$E9
	.dc.b	$F3,$DF,$E3,$07,$B6,$77,$1E,$58,$76,$F2,$7F,$C5,$55,$AC,$BA,$E9
	.dc.b	$FE,$31,$D7,$FC,$9B,$F1,$3F,$7B,$F1,$77,$D3,$7F,$B3,$7F,$52,$FD
	.dc.b	$77,$ED,$FA,$EC,$FF,$6B,$4E,$78,$DF,$95,$E7,$0F,$F2,$1E,$3F,$88
	.dc.b	$FC,$6E,$E1,$C3,$F1,$96,$7F,$EA,$0D,$38,$67,$DF,$97,$FC,$81,$FA
	.dc.b	$37,$F9,$BB,$F3,$3F,$EC,$1F,$56,$F8,$CF,$F3,$17,$E9,$1F,$E6,$4A
	.dc.b	$D7,$CE,$FF,$85,$27,$FC,$CD,$FA,$46,$7E,$81,$8F,$F9,$5A,$FF,$CF
	.dc.b	$3F,$A0,$59,$CE,$1C,$7F,$44,$F0,$C6,$33,$F0,$79,$9F,$E5,$3F,$83
	.dc.b	$FF,$05,$7E,$D6,$F9,$9F,$F0,$77,$E0,$3F,$EF,$4F,$3F,$FC,$EF,$5A
	.dc.b	$9C,$FC,$A3,$5C,$FD,$77,$FB,$C3,$9B,$FD,$63,$CB,$FE,$2A,$C3,$AC
	.dc.b	$7F,$83,$5E,$5F,$59,$FE,$79,$FC,$63,$C5,$6D,$7C,$53,$FE,$7D,$F0
	.dc.b	$FF,$CD,$1F,$34,$FC,$4B,$FC,$D3,$F8,$6B,$F3,$D6,$AF,$D0,$9F,$A4
	.dc.b	$7F,$A3,$B0,$E1,$F8,$03,$0E,$17,$BF,$D2,$75,$97,$F9,$D3,$87,$FC
	.dc.b	$45,$87,$7E,$7F,$F4,$1B,$CB,$0F,$07,$E1,$7F,$8F,$C7,$FC,$8F,$39
	.dc.b	$D7,$C7,$CF,$FE,$7F,$FC,$E5,$FA,$47,$FA,$CB,$F3,$9F,$ED,$99,$BF
	.dc.b	$72,$7E,$CF,$FE,$A4,$BD,$FE,$2A,$BD,$D3,$F3,$0D,$62,$F8,$DF,$F4
	.dc.b	$BE,$1C,$38,$7F,$C7,$4F,$2C,$3B,$F0,$F8,$EF,$C5,$F5,$9F,$EF,$1E
	.dc.b	$1F,$F4,$A7,$E9,$2F,$D2,$9F,$A5,$7D,$DF,$CC,$3F,$80,$7D,$E7,$F8
	.dc.b	$8B,$5F,$F4,$F5,$EE,$B5,$EF,$2B,$FC,$47,$8F,$F9,$AA,$BA,$7A,$3F
	.dc.b	$EA,$C7,$86,$1F,$E4,$F7,$97,$8A,$CB,$F8,$77,$99,$E4,$D3,$C7,$39
	.dc.b	$CD,$E6,$7F,$AB,$F3,$7E,$F5,$FC,$07,$ED,$DF,$19,$FE,$79,$FC,$51
	.dc.b	$F6,$9F,$D9,$79,$3D,$B8,$70,$E0,$FF,$5C,$61,$C3,$F0,$56,$1C,$3E
	.dc.b	$8F,$EC,$FE,$6F,$F8,$ED,$3B,$68,$FD,$BF,$FD,$ED,$18,$C2,$E5,$D2
	.dc.b	$BC,$5A,$FF,$75,$FD,$6B,$D1,$7B,$CF,$B6,$78,$C3,$FD,$95,$4F,$C3
	.dc.b	$BF,$D0,$1F,$64,$FA,$CF,$E2,$BF,$AF,$7F,$82,$38,$7D,$58,$30,$E1
	.dc.b	$E8,$B0,$E1,$F0,$CC,$3F,$D6,$F8,$70,$BE,$DF,$84,$FA,$31,$E5,$9F
	.dc.b	$FA,$11,$FD,$DE,$DF,$CB,$6B,$AE,$5D,$3D,$9F,$45,$70,$D3,$C1,$AF
	.dc.b	$F2,$AF,$B7,$CE,$1E,$4F,$2B,$C7,$F7,$5E,$9C,$FA,$69,$FE,$9D,$9C
	.dc.b	$FE,$1B,$F3,$F3,$F8,$73,$9C,$BF,$9A,$E9,$D2,$31,$C2,$FC,$4B,$1B
	.dc.b	$3F,$46,$F8,$3F,$6B,$DF,$2E,$73,$9C,$BF,$9F,$CF,$C4,$D7,$C7,$39
	.dc.b	$C7,$F9,$25,$F8,$CF,$76,$1F,$33,$1F,$83,$E2,$7E,$8F,$F2,$EF,$CB
	.dc.b	$7C,$57,$F0,$0E,$6F,$AB,$39,$F1,$BF,$B5,$4F,$E9,$B4,$FE,$B1,$3F
	.dc.b	$A6,$CB,$FD,$F1,$18,$F1,$7F,$67,$18,$3F,$62,$FE,$DB,$1F,$D2,$78
	.dc.b	$DE,$28,$C6,$3F,$DD,$FF,$46,$C7,$94,$FC,$EF,$7E,$37,$AF,$4F,$DD
	.dc.b	$47,$2F,$5F,$18,$FF,$02,$8E,$BA,$63,$0E,$59,$7F,$80,$FC,$8E,$B5
	.dc.b	$E2,$FC,$5F,$4F,$93,$EF,$FF,$9A,$FC,$DF,$F7,$5A,$7E,$84,$FC,$5F
	.dc.b	$FC,$2F,$F0,$AF,$D9,$FF,$7B,$F0,$3C,$36,$1F,$C0,$2F,$87,$D2,$7F
	.dc.b	$B8,$6B,$F8,$1F,$56,$EB,$58,$FF,$80,$DF,$E6,$D9,$F8,$70,$7E,$19
	.dc.b	$99,$FE,$15,$39,$AF,$1E,$1F,$C3,$A7,$A6,$9E,$89,$CF,$BF,$C2,$7C
	.dc.b	$BF,$80,$BF,$0A,$FF,$6F,$7E,$A8,$FC,$D9,$F2,$7F,$E6,$6F,$2B,$CD
	.dc.b	$7B,$FC,$21,$7B,$AF,$F3,$5F,$C1,$EB,$17,$E2,$7E,$7C,$F1,$85,$F1
	.dc.b	$5C,$3F,$E2,$1C,$FA,$67,$CF,$E1,$FC,$D8,$27,$3B,$3F,$65,$FE,$B5
	.dc.b	$F8,$07,$F8,$A7,$E7,$5F,$19,$FD,$0F,$F2,$47,$93,$EA,$DF,$FD,$93
	.dc.b	$BF,$FA,$27,$4F,$AB,$87,$0F,$11,$87,$0E,$4F,$F8,$01,$C7,$FD,$DB
	.dc.b	$4E,$DA,$3F,$69,$F5,$46,$30,$B9,$74,$AF,$16,$BF,$2B,$D6,$FE,$81
	.dc.b	$7B,$CF,$B7,$C9,$E3,$0F,$C3,$FB,$0F,$D6,$FF,$C0,$9F,$24,$F9,$A7
	.dc.b	$E4,$DF,$E0,$3F,$D3,$3C,$98,$70,$BD,$D6,$BE,$3B,$F8,$67,$2F,$BE
	.dc.b	$BF,$DC,$34,$FC,$B3,$CF,$18,$F9,$8B,$A7,$85,$69,$E1,$D3,$FC,$25
	.dc.b	$5E,$03,$84,$67,$1F,$EB,$9F,$57,$FE,$5B,$FD,$DD,$B3,$F4,$CF,$F6
	.dc.b	$17,$E9,$3E,$1F,$C5,$9F,$47,$FE,$22,$C3,$85,$EF,$E8,$57,$BF,$C9
	.dc.b	$B5,$AF,$EB,$33,$EB,$FA,$37,$E2,$F1,$E3,$A5,$AD,$65,$CF,$1F,$81
	.dc.b	$F2,$FE,$69,$CD,$79,$38,$DE,$AE,$65,$EE,$3F,$0D,$FF,$75,$FE,$72
	.dc.b	$FC,$4D,$F8,$9F,$F7,$AF,$CC,$3C,$B8,$70,$BD,$D6,$BE,$15,$F8,$B6
	.dc.b	$B2,$E9,$EF,$79,$F0,$E1,$F0,$16,$73,$A7,$7D,$1C,$7E,$59,$CE,$7E
	.dc.b	$2C,$31,$A7,$82,$FF,$C1,$BE,$87,$78,$C6,$AF,$C8,$31,$AB,$4E,$D3
	.dc.b	$F5,$FB,$3B,$E5,$FE,$28,$C3,$8F,$3F,$A7,$CA,$FF,$0E,$FC,$17,$F8
	.dc.b	$BB,$E5,$71,$F0,$9F,$E2,$CF,$9F,$FD,$B6,$5C,$B2,$F9,$78,$7F,$A7
	.dc.b	$1F,$E3,$51,$9E,$2F,$F1,$6C,$7F,$9C,$B8,$FF,$A5,$61,$FE,$36,$F8
	.dc.b	$9F,$81,$87,$83,$E7,$BF,$8F,$63,$C6,$31,$97,$F8,$A2,$31,$7E,$3F
	.dc.b	$07,$EA,$FE,$05,$FE,$D6,$F4,$FF,$9D,$FE,$42,$FD,$9B,$F6,$1F,$07
	.dc.b	$AF,$C0,$7F,$92,$7C,$3A,$B5,$E7,$BF,$62,$D7,$E2,$7F,$67,$3E,$9C
	.dc.b	$9E,$CE,$4F,$F5,$E4,$77,$C7,$96,$9F,$9C,$AE,$DA,$F8,$A1,$FA,$47
	.dc.b	$E4,$7F,$3B,$7E,$72,$FD,$31,$FA,$03,$F8,$01,$F9,$CB,$F7,$8F,$B9
	.dc.b	$D1,$F3,$F1,$E3,$87,$0B,$DD,$7D,$DE,$0F,$F6,$85,$6B,$6F,$41,$FD
	.dc.b	$61,$F0,$FE,$27,$F9,$57,$9E,$7E,$1C,$DF,$83,$F9,$F5,$F3,$4E,$7F
	.dc.b	$18,$FF,$2F,$74,$7F,$B7,$7F,$1A,$7D,$83,$E8,$7F,$A0,$FA,$4F,$CB
	.dc.b	$5F,$97,$3C,$5F,$F5,$2E,$3C,$70,$E1,$F2,$5F,$B6,$BD,$F0,$EB,$8B
	.dc.b	$FB,$7E,$9D,$B4,$FE,$9D,$8F,$1F,$AB,$B5,$AC,$B9,$EB,$E1,$F9,$7F
	.dc.b	$3C,$E7,$0F,$2E,$7F,$8E,$C7,$FD,$ED,$FA,$8B,$F7,$CF,$E0,$BF,$96
	.dc.b	$3F,$03,$7E,$8F,$FE,$83,$F1,$3C,$99,$B0,$E1,$E7,$30,$E1,$7F,$E8
	.dc.b	$9F,$C8,$37,$5D,$7F,$08,$F3,$E1,$C7,$3F,$E2,$6F,$EA,$34,$F1,$C7
	.dc.b	$8B,$4F,$0F,$C5,$FA,$AB,$86,$9D,$E7,$CB,$F5,$F9,$F7,$5F,$9E,$3F
	.dc.b	$48,$7E,$EA,$FC,$AF,$F1,$DF,$43,$FE,$FF,$F2,$9B,$FF,$87,$35,$FE
	.dc.b	$45,$8F,$1C,$38,$79,$3F,$E0,$0D,$BE,$0E,$9D,$BF,$67,$F5,$BE,$0D
	.dc.b	$6E,$1C,$75,$FC,$E7,$CA,$F9,$63,$18,$FF,$A1,$73,$F9,$7F,$B8,$38
	.dc.b	$BF,$D1,$1F,$73,$FA,$5F,$CA,$9F,$AA,$BF,$58,$DD,$6B,$D1,$56,$BE
	.dc.b	$67,$3C,$5F,$3B,$2E,$9F,$3B,$70,$E1,$7F,$81,$D3,$FC,$77,$5A,$D3
	.dc.b	$C1,$E5,$7F,$AE,$F6,$FF,$52,$7D,$F7,$FA,$DF,$DF,$FF,$7E,$C6,$3E
	.dc.b	$0F,$FD,$67,$D5,$FB,$9C,$BF,$D5,$DF,$80,$FF,$88,$EB,$FD,$65,$1F
	.dc.b	$C9,$1E,$1B,$75,$8F,$F5,$A7,$E0,$3F,$EB,$EC,$98,$C9,$EE,$7F,$94
	.dc.b	$3E,$D1,$F9,$93,$F6,$7F,$E0,$DF,$36,$FD,$07,$FD,$CF,$FB,$0F,$0C
	.dc.b	$18,$70,$BD,$D7,$FB,$B3,$D1,$D4,$F4,$FD,$9B,$D1,$7D,$F2,$EC,$FA
	.dc.b	$4C,$61,$CB,$E6,$FD,$51,$DA,$7C,$9F,$00,$D3,$C9,$D1,$AF,$B2,$FC
	.dc.b	$97,$EA,$9F,$E3,$FF,$BD,$7E,$BC,$FD,$25,$F5,$1E,$9B,$DF,$D9,$2F
	.dc.b	$75,$FE,$1F,$EB,$FA,$1B,$2F,$F3,$DF,$19,$EA,$8C,$79,$2D,$31,$5D
	.dc.b	$31,$F0,$F7,$F9,$6B,$59,$F7,$D5,$F7,$FE,$EF,$C0,$FC,$DF,$8E,$7F
	.dc.b	$47,$F9,$47,$EC,$DF,$C8,$7E,$B7,$E8,$56,$BE,$4D,$5A,$F8,$1F,$E4
	.dc.b	$7C,$7E,$8A,$F8,$1F,$B2,$79,$DE,$BC,$7F,$3C,$FF,$77,$72,$B4,$F0
	.dc.b	$68,$BF,$24,$E7,$3E,$F3,$FE,$3C,$C7,$FD,$05,$87,$FA,$63,$E5,$FF
	.dc.b	$D1,$1F,$80,$7F,$A5,$D7,$E0,$2A,$DD,$7F,$A6,$2B,$5F,$80,$E7,$E5
	.dc.b	$FF,$D2,$93,$FD,$F9,$FF,$4D,$7E,$03,$FE,$81,$7F,$D3,$15,$D6,$78
	.dc.b	$1E,$1F,$99,$FF,$53,$FC,$5F,$FA,$B7,$E2,$FF,$D5,$79,$78,$3E,$D1
	.dc.b	$FB,$6B,$F6,$BF,$F0,$D7,$E6,$AF,$DB,$7F,$C3,$9F,$8E,$F4,$D3,$FC
	.dc.b	$FB,$7B,$FA,$95,$ED,$7E,$3D,$BC,$BA,$E5,$E7,$C3,$FD,$77,$1E,$0F
	.dc.b	$BF,$C4,$74,$9F,$0E,$7F,$82,$8E,$59,$77,$CB,$C9,$E1,$FA,$B8,$BF
	.dc.b	$95,$7E,$47,$FA,$1F,$C1,$DF,$BC,$FF,$A3,$7E,$6D,$F8,$FC,$E3,$F9
	.dc.b	$F7,$BA,$D4,$FD,$49,$CF,$9F,$FE,$D5,$D7,$E8,$1F,$ED,$2B,$FD,$77
	.dc.b	$4F,$A1,$E5,$F8,$1F,$FA,$CB,$AF,$C9,$5B,$BF,$F6,$F6,$5E,$3F,$91
	.dc.b	$CC,$BC,$F7,$BA,$FD,$13,$FB,$E7,$E3,$BF,$E0,$4F,$CE,$5F,$A9,$FD
	.dc.b	$1B,$DD,$6B,$E9,$94,$FC,$DB,$58,$F3,$CB,$D2,$77,$5F,$BC,$7D,$0C
	.dc.b	$79,$F9,$CD,$3F,$C2,$7F,$90,$F9,$A7,$31,$FE,$16,$FC,$EF,$D3,$F7
	.dc.b	$7F,$EC,$AF,$07,$FD,$B9,$D1,$F9,$A8,$C7,$E1,$EF,$C4,$7F,$11,$1F
	.dc.b	$EE,$5F,$9D,$FB,$EF,$FB,$87,$7F,$2B,$C1,$DF,$7F,$FB,$AB,$F1,$0F
	.dc.b	$F6,$B4,$7C,$8E,$A7,$E6,$5F,$E2,$8F,$AD,$79,$FF,$EA,$EF,$83,$7E
	.dc.b	$61,$F9,$7B,$DF,$52,$F7,$F8,$67,$F8,$FF,$F2,$3D,$64,$FD,$2B,$F2
	.dc.b	$78,$70,$D3,$E4,$BC,$B0,$E3,$8F,$8E,$FC,$5A,$7F,$AF,$A7,$EE,$F1
	.dc.b	$7F,$BD,$3F,$64,$7E,$57,$F7,$BC,$57,$E9,$0F,$DD,$FF,$59,$7B,$AD
	.dc.b	$4F,$F8,$F7,$A3,$F8,$2C,$FF,$BF,$BF,$5C,$FD,$3B,$D7,$7F,$BA,$F1
	.dc.b	$F0,$7F,$DD,$9F,$B2,$78,$BC,$D6,$BE,$29,$FF,$78,$7E,$BF,$FC,$0B
	.dc.b	$EC,$FF,$C2,$5E,$D3,$C1,$7E,$48,$FD,$F3,$FC,$0D,$F3,$5F,$17,$F3
	.dc.b	$1F,$E3,$18,$70,$BD,$D6,$BA,$DF,$BF,$EB,$84,$33,$ED,$9F,$F8,$86
	.dc.b	$3F,$C4,$5C,$6C,$BA,$7E,$97,$99,$F3,$1C,$AF,$8E,$5E,$4E,$CC,$CF
	.dc.b	$5D,$3C,$7F,$4B,$FD,$F3,$7C,$FE,$3E,$FC,$7D,$F8,$53,$E0,$5F,$8F
	.dc.b	$BC,$D7,$E4,$1E,$3E,$2F,$26,$BE,$AC,$B9,$63,$C7,$0E,$16,$F2,$FE
	.dc.b	$05,$AF,$C1,$E1,$7A,$E3,$31,$C9,$FF,$5C,$F2,$7F,$90,$E3,$AE,$9E
	.dc.b	$4D,$3F,$C7,$1A,$78,$AF,$A6,$BE,$AA,$D7,$D8,$FC,$2F,$A3,$77,$FE
	.dc.b	$03,$F4,$57,$C9,$97,$E7,$CF,$96,$F3,$DF,$0F,$0B,$57,$BA,$7E,$F1
	.dc.b	$AC,$5E,$37,$AF,$7F,$26,$3C,$77,$F5,$9F,$80,$7A,$69,$E2,$3C,$31
	.dc.b	$CF,$3E,$F1,$FE,$B8,$8F,$93,$F5,$EF,$C8,$1F,$9F,$FD,$7F,$DF,$1F
	.dc.b	$9D,$7F,$27,$FD,$EF,$75,$AE,$45,$6B,$E0,$FF,$C7,$1B,$7C,$9C,$BA
	.dc.b	$7E,$25,$E9,$EA,$FF,$39,$47,$7D,$31,$AF,$E6,$23,$19,$74,$D7,$FC
	.dc.b	$05,$3E,$0A,$F2,$43,$F5,$9F,$C9,$DB,$F7,$C7,$DF,$FF,$82,$FE,$5B
	.dc.b	$EA,$FA,$BF,$AD,$DB,$C9,$5A,$FA,$05,$6B,$F7,$5F,$87,$3C,$B5,$F9
	.dc.b	$A3,$2E,$DF,$BC,$FA,$23,$13,$CE,$FF,$07,$F4,$5E,$48,$C7,$5F,$F8
	.dc.b	$5F,$8B,$E0,$F9,$1F,$F0,$03,$F2,$FF,$A2,$BE,$E5,$F9,$CB,$F1,$C7
	.dc.b	$E3,$8F,$C7,$3F,$CB,$FC,$7F,$37,$07,$DA,$78,$F8,$BF,$5F,$95,$9E
	.dc.b	$99,$73,$C7,$FD,$AD,$C3,$E0,$AE,$F9,$F6,$FA,$6F,$E1,$3F,$46,$FC
	.dc.b	$73,$FD,$51,$F8,$43,$F1,$9F,$CD,$FC,$1B,$E0,$5F,$98,$3E,$B1,$F6
	.dc.b	$2C,$B9,$6D,$FC,$53,$2E,$5E,$7B,$2E,$58,$F1,$F6,$B8,$6B,$E0,$D7
	.dc.b	$FC,$DD,$8F,$F1,$5C,$B9,$BF,$EC,$08,$FC,$EB,$FC,$76,$33,$97,$6C
	.dc.b	$7A,$ED,$EC,$C3,$A6,$7E,$1D,$BD,$3E,$FE,$72,$FE,$53,$F3,$4F,$CD
	.dc.b	$9F,$8B,$BF,$5F,$FF,$17,$E4,$F9,$07,$E2,$7E,$8C,$78,$E1,$C2,$F7
	.dc.b	$5F,$CB,$F8,$3D,$39,$75,$CB,$F8,$B4,$62,$FE,$AE,$1E,$BC,$BF,$DB
	.dc.b	$17,$BC,$79,$CF,$73,$FC,$25,$5D,$FE,$B7,$EC,$6B,$EB,$97,$EA,$9F
	.dc.b	$E1,$7F,$D9,$5F,$85,$FF,$0C,$FC,$1F,$FA,$8E,$9C,$EB,$53,$9F,$8C
	.dc.b	$7E,$3B,$9D,$98,$FA,$30,$FF,$76,$61,$FE,$23,$FA,$5C,$67,$FD,$4A
	.dc.b	$39,$FB,$DF,$EE,$68,$E5,$7E,$28,$69,$F8,$79,$F6,$7E,$DD,$FB,$4F
	.dc.b	$C2,$9F,$93,$7F,$BF,$FE,$8D,$F9,$97,$F1,$0E,$6F,$C1,$DE,$EB,$52
	.dc.b	$FC,$BF,$9F,$A6,$FC,$7F,$CB,$9B,$0E,$CF,$E3,$78,$7F,$B4,$B9,$A6
	.dc.b	$9E,$2F,$C5,$7F,$C5,$DD,$FF,$37,$0F,$04,$7F,$8C,$7E,$AB,$E8,$78
	.dc.b	$2F,$D9,$DF,$95,$FF,$AE,$7C,$EB,$FB,$A3,$F8,$07,$ED,$5E,$EB,$42
	.dc.b	$5F,$C0,$1A,$C3,$FC,$67,$BF,$9C,$FF,$50,$3D,$9C,$0B,$E9,$F8,$4F
	.dc.b	$F0,$CC,$3F,$CF,$DC,$DF,$AA,$9C,$E7,$E6,$E4,$FE,$03,$DD,$F0,$BA
	.dc.b	$9F,$BB,$3F,$45,$F9,$5F,$0E,$FD,$71,$FB,$13,$F7,$1F,$E0,$18,$F1
	.dc.b	$C3,$87,$67,$F8,$06,$F7,$FB,$2E,$B8,$BF,$C2,$FA,$76,$E7,$FD,$1E
	.dc.b	$3C,$75,$B5,$AC,$B9,$EB,$E1,$FC,$1B,$CF,$39,$AF,$2E,$7F,$D1,$63
	.dc.b	$B1,$F4,$BC,$9F,$E3,$5F,$AE,$3F,$7F,$7E,$AB,$FE,$B4,$FD,$8F,$7D
	.dc.b	$1F,$C7,$7D,$75,$EE,$BF,$D7,$3F,$47,$AC,$7F,$92,$65,$FC,$57,$83
	.dc.b	$F1,$38,$FF,$25,$9F,$4C,$F8,$71,$E7,$9F,$83,$9F,$D3,$39,$CF,$B7
	.dc.b	$E2,$18,$FC,$EF,$FC,$39,$F8,$4F,$F1,$2F,$B1,$7E,$EE,$FC,$D5,$FA
	.dc.b	$AF,$BA,$F7,$ED,$7F,$98,$2B,$5F,$A3,$E7,$EA,$3F,$D9,$3A,$F9,$CF
	.dc.b	$F2,$3D,$FF,$8A,$7F,$6D,$F3,$BC,$B3,$EF,$9F,$8E,$FF,$DA,$39,$75
	.dc.b	$DF,$D3,$F8,$93,$F7,$97,$ED,$4F,$D1,$1F,$9C,$3E,$65,$FB,$33,$F5
	.dc.b	$CF,$AB,$8F,$1C,$38,$5E,$EB,$EA,$F9,$BF,$C4,$B8,$BF,$CB,$B7,$DF
	.dc.b	$F5,$2F,$9A,$F1,$8F,$60,$7A,$67,$E2,$F6,$BF,$7B,$7C,$EB,$5A,$FA
	.dc.b	$2B,$CF,$AF,$F5,$59,$FC,$4F,$C5,$FF,$8C,$BF,$18,$7E,$E7,$FE,$40
	.dc.b	$FD,$17,$FC,$7F,$FA,$C6,$BC,$EF,$F7,$D4,$E7,$EC,$7F,$C5,$3A,$F7
	.dc.b	$C7,$96,$5E,$67,$FC,$7B,$FA,$B7,$E1,$9E,$D7,$D7,$E3,$BF,$0B,$18
	.dc.b	$FA,$06,$9F,$DC,$70,$FC,$EF,$C4,$7F,$B4,$FF,$5B,$7E,$1E,$FC,$83
	.dc.b	$F9,$BF,$7F,$F8,$27,$6F,$81,$7B,$AD,$7C,$5F,$E2,$FF,$75,$E7,$F1
	.dc.b	$CF,$D1,$FF,$33,$7D,$74,$F9,$AF,$1C,$FA,$C7,$0D,$3C,$31,$FD,$1B
	.dc.b	$C1,$FC,$EF,$C0,$7F,$23,$FC,$31,$FB,$FB,$F6,$29,$F8,$37,$EC,$F8
	.dc.b	$7E,$A5,$EE,$B5,$4A,$FC,$FD,$6B,$E1,$BF,$91,$E9,$FE,$01,$F5,$FF
	.dc.b	$9C,$E3,$FC,$C7,$F5,$8F,$D8,$BC,$F1,$F0,$7D,$F7,$93,$0F,$F6,$EE
	.dc.b	$7D,$B2,$FF,$10,$FE,$E1,$FB,$AF,$67,$F8,$3F,$E2,$6F,$CC,$DF,$9A
	.dc.b	$7C,$DB,$5E,$FD,$CF,$E0,$F7,$BA,$7E,$C3,$FE,$23,$AD,$70,$FF,$86
	.dc.b	$DE,$D9,$7F,$80,$9E,$5E,$E7,$FB,$B3,$3F,$F7,$55,$71,$3F,$D7,$BE
	.dc.b	$F7,$F1,$59,$D7,$EC,$DE,$89,$FD,$2F,$9B,$F9,$BF,$D2,$5D,$FF,$E9
	.dc.b	$6F,$B6,$E5,$5F,$A0,$F5,$7F,$6D,$5A,$F1,$9F,$6B,$AF,$4F,$F5,$D7
	.dc.b	$E7,$3A,$DF,$9D,$BE,$93,$0E,$51,$C7,$3F,$16,$7F,$AB,$E6,$F8,$F3
	.dc.b	$FE,$05,$F1,$5E,$1F,$E2,$3F,$7C,$7E,$03,$F3,$79,$7F,$C1,$B5,$AF
	.dc.b	$27,$F3,$95,$AF,$D6,$3D,$7E,$42,$BF,$31,$97,$9C,$FF,$10,$BD,$9C
	.dc.b	$47,$E6,$6B,$FD,$4B,$9F,$92,$77,$9F,$8F,$E3,$BF,$CC,$7F,$8E,$7D
	.dc.b	$4E,$EF,$C1,$7D,$77,$F6,$0C,$13,$9E,$CF,$F3,$17,$F0,$0F,$F9,$86
	.dc.b	$3E,$FE,$5F,$E5,$57,$EF,$BF,$E5,$B7,$87,$8D,$DA,$3E,$EF,$7F,$E4
	.dc.b	$2F,$FD,$F5,$1C,$CF,$91,$A7,$C7,$FC,$E3,$F5,$79,$78,$A7,$3C,$5F
	.dc.b	$1A,$7F,$A3,$CE,$67,$E1,$F9,$DF,$E5,$0C,$FE,$6E,$16,$78,$F9,$CE
	.dc.b	$1F,$8B,$5A,$CB,$E8,$F0,$FF,$A1,$AF,$FC,$4F,$F8,$F3,$F4,$E7,$E6
	.dc.b	$CF,$D8,$5F,$B6,$FF,$C0,$0F,$29,$FA,$67,$F8,$FF,$10,$C7,$96,$1C
	.dc.b	$6F,$87,$E0,$7B,$AD,$65,$DB,$2E,$BF,$25,$FE,$95,$FD,$BB,$5D,$9F
	.dc.b	$77,$99,$FB,$7E,$3B,$BF,$15,$CB,$ED,$9F,$B5,$9F,$E3,$B9,$FF,$CE
	.dc.b	$6F,$C4,$3F,$D2,$07,$C1,$9F,$89,$F4,$FF,$06,$3E,$3F,$3B,$97,$D7
	.dc.b	$F6,$2F,$DC,$BF,$C9,$7F,$BA,$B0,$FF,$1D,$FD,$E3,$EA,$BE,$0F,$E1
	.dc.b	$5F,$C3,$70,$E1,$9F,$C4,$C3,$87,$D4,$FF,$B0,$EF,$7E,$BA,$F7,$7F
	.dc.b	$ED,$5F,$C0,$3F,$D7,$F9,$7F,$B2,$3F,$04,$7C,$4E,$F6,$FF,$68,$FE
	.dc.b	$01,$FE,$C0,$FC,$07,$FD,$99,$F9,$D6,$72,$FA,$7F,$AF,$FD,$3F,$92
	.dc.b	$7E,$D9,$FE,$6B,$FC,$1B,$FC,$E3,$F1,$1E,$1F,$D7,$3F,$CC,$FF,$21
	.dc.b	$FE,$14,$F8,$77,$EB,$EF,$CC,$BF,$BF,$FE,$22,$FA,$DE,$1B,$7F,$BF
	.dc.b	$F9,$BC,$FC,$3F,$ED,$DE,$4F,$67,$E8,$5F,$E7,$28,$7E,$E6,$FD,$EB
	.dc.b	$F9,$5F,$CF,$BF,$73,$FF,$32,$FB,$DC,$33,$FC,$5F,$B4,$FC,$E1,$FA
	.dc.b	$37,$F9,$FF,$F2,$27,$CB,$FF,$3B,$FC,$27,$E5,$67,$FE,$80,$E9,$7E
	.dc.b	$AD,$FE,$69,$FB,$A7,$ED,$AF,$C9,$DF,$8A,$3F,$61,$E3,$C1,$F9,$AC
	.dc.b	$78,$FC,$3F,$FA,$3F,$0E,$16,$FB,$0D,$DB,$F1,$1F,$EF,$7F,$A3,$3D
	.dc.b	$27,$F0,$0F,$E6,$3F,$42,$7E,$D7,$C2,$F7,$5A,$F7,$54,$FD,$F3,$5F
	.dc.b	$52,$CB,$FD,$37,$9F,$FA,$73,$D4,$FF,$18,$BF,$B0,$CB,$FC,$17,$FA
	.dc.b	$66,$A3,$F3,$3C,$FF,$EA,$29,$EF,$B7,$EA,$BC,$BC,$D7,$1C,$BB,$6D
	.dc.b	$E3,$FD,$0B,$F4,$3F,$AE,$FF,$AA,$BE,$9D,$FB,$43,$F2,$FF,$E0,$DF
	.dc.b	$92,$FF,$B0,$3F,$47,$7E,$63,$F9,$BC,$38,$5E,$FE,$35,$7B,$E4,$D6
	.dc.b	$DF,$C2,$33,$EB,$9F,$A6,$32,$7F,$83,$70,$FF,$2A,$3E,$DE,$46,$5D
	.dc.b	$AB,$C7,$F8,$CF,$F8,$AE,$73,$5D,$6F,$C5,$97,$FB,$5B,$E9,$BD,$B3
	.dc.b	$FC,$63,$C3,$7E,$88,$FD,$93,$FD,$AB,$F9,$5B,$E4,$1F,$B7,$BF,$46
	.dc.b	$F2,$E1,$C2,$F7,$5A,$F9,$7C,$E3,$CF,$EF,$99,$73,$CB,$FC,$7B,$BF
	.dc.b	$9B,$4F,$F3,$6F,$E5,$CF,$C5,$DF,$95,$7E,$0F,$F3,$77,$EC,$5F,$E3
	.dc.b	$BC,$38,$7E,$0A,$C3,$85,$EE,$9F,$BE,$7C,$AA,$D7,$E4,$7F,$37,$0E
	.dc.b	$38,$77,$CF,$E6,$75,$7F,$BA,$2F,$A6,$9E,$2F,$21,$AB,$F5,$AF,$93
	.dc.b	$87,$A2,$B5,$3F,$E1,$08,$FE,$39,$FA,$D7,$FB,$CB,$F6,$47,$6B,$F0
	.dc.b	$EF,$F6,$97,$EA,$9F,$ED,$0F,$3D,$F5,$FC,$31,$FE,$97,$87,$0B,$DF
	.dc.b	$E3,$FF,$D0,$BD,$6D,$52,$B4,$FE,$17,$3F,$06,$6F,$CE,$7E,$04,$63
	.dc.b	$1E,$7F,$15,$FC,$FA,$31,$F8,$E3,$6F,$F7,$1F,$E6,$DF,$D0,$BF,$61
	.dc.b	$FE,$BB,$9F,$9B,$ED,$BF,$86,$FF,$00,$FF,$25,$F8,$9E,$1F,$47,$FC
	.dc.b	$EE,$F7,$C7,$FE,$4C,$F2,$17,$BC,$FF,$DE,$5D,$AB,$DE,$5F,$C5,$73
	.dc.b	$FC,$F4,$63,$0F,$B0,$DC,$3C,$1D,$3E,$5A,$FE,$E3,$97,$79,$F2,$7E
	.dc.b	$6B,$FD,$6B,$F3,$2F,$F3,$5F,$3B,$F5,$AF,$C6,$F8,$4F,$AD,$7E,$55
	.dc.b	$F8,$8E,$0F,$F5,$46,$8C,$26,$1C,$2F,$7F,$85,$7F,$59,$CB,$AF,$8F
	.dc.b	$F8,$D7,$BD,$5C,$3F,$C0,$B8,$72,$AF,$CA,$E1,$FE,$24,$C7,$BC,$E7
	.dc.b	$3F,$27,$E0,$DF,$DC,$BF,$28,$FF,$41,$ED,$FE,$E6,$F8,$9F,$97,$FA
	.dc.b	$8F,$AF,$2F,$ED,$55,$AF,$C7,$FF,$9E,$71,$2B,$E0,$7C,$3E,$B7,$F1
	.dc.b	$E4,$FC,$7B,$F4,$98,$F1,$9F,$E1,$CF,$3C,$FB,$FE,$6B,$E3,$9E,$39
	.dc.b	$F6,$C9,$FA,$E7,$F8,$23,$F4,$BF,$F4,$CF,$87,$FA,$0F,$9D,$FD,$DF
	.dc.b	$E0,$0F,$DF,$FF,$55,$CC,$BD,$F9,$3F,$DF,$BA,$3F,$CD,$74,$FD,$DF
	.dc.b	$FC,$0D,$5D,$3F,$45,$FD,$D3,$C3,$7F,$E4,$8F,$2C,$FB,$FB,$FE,$39
	.dc.b	$FA,$3A,$F8,$F2,$FF,$3B,$FE,$0B,$FE,$1E,$FC,$EF,$FD,$5F,$F3,$9F
	.dc.b	$EA,$2E,$F7,$27,$E6,$3F,$7E,$FF,$3A,$71,$2F,$7E,$47,$E2,$EB,$5F
	.dc.b	$4F,$FE,$1B,$9E,$7F,$8C,$7F,$0F,$77,$B7,$F3,$97,$8E,$5D,$BB,$DF
	.dc.b	$A3,$F8,$EB,$F0,$FA,$F8,$F2,$FD,$4F,$9B,$FE,$34,$FC,$B7,$FD,$8B
	.dc.b	$F1,$BF,$E5,$0F,$D7,$FF,$D9,$9C,$7F,$E9,$2F,$D0,$7F,$C5,$95,$AC
	.dc.b	$95,$AE,$D7,$E9,$5A,$9F,$CA,$E3,$FE,$6D,$7F,$B7,$5F,$17,$FA,$83
	.dc.b	$CE,$FB,$EB,$E3,$9E,$27,$6F,$9F,$FE,$73,$E3,$7F,$92,$BC,$8F,$F6
	.dc.b	$DF,$E7,$5F,$E6,$6E,$0F,$E7,$B8,$FF,$04,$F2,$FF,$C8,$55,$AC,$7E
	.dc.b	$8F,$07,$A3,$8F,$FA,$04,$E6,$7F,$B1,$3D,$3F,$0D,$FE,$78,$F1,$C7
	.dc.b	$AF,$EE,$BE,$1A,$FD,$8E,$7E,$1C,$7E,$1F,$E1,$FF,$E5,$EF,$37,$FD
	.dc.b	$DF,$D9,$FE,$7A,$F1,$FE,$3F,$EB,$3F,$98,$EF,$FF,$0C,$E5,$FB,$AA
	.dc.b	$D7,$0A,$BF,$AB,$FE,$0D,$FE,$5B,$C7,$9D,$FF,$4F,$7A,$F0,$58,$E4
	.dc.b	$7F,$44,$BE,$5E,$57,$8E,$33,$9F,$8B,$CC,$FF,$39,$FC,$2F,$F3,$2F
	.dc.b	$4B,$FD,$4D,$AF,$DC,$F6,$7E,$36,$FF,$E2,$9C,$7F,$87,$E9,$FE,$94
	.dc.b	$AD,$69,$FD,$B6,$B5,$5F,$C7,$6B,$FC,$20,$FF,$9C,$F1,$ED,$1F,$83
	.dc.b	$8F,$09,$D0,$5F,$87,$2C,$C7,$33,$38,$7F,$94,$B4,$FF,$45,$F9,$3E
	.dc.b	$99,$FE,$58,$7E,$53,$C3,$FC,$AF,$1F,$FA,$0E,$75,$B7,$E7,$67,$F3
	.dc.b	$3F,$94,$E6,$FF,$95,$9C,$8F,$F2,$2E,$CB,$ED,$EC,$7F,$7F,$79,$CF
	.dc.b	$7D,$3C,$6F,$2C,$7B,$67,$FE,$A0,$F4,$3C,$DD,$BF,$97,$DB,$FC,$CD
	.dc.b	$AF,$FA,$5E,$73,$87,$F0,$69,$FF,$35,$19,$7C,$B5,$FE,$0F,$9F,$81
	.dc.b	$F2,$D6,$72,$71,$EA,$F0,$38,$EB,$FD,$BF,$C3,$7F,$AC,$BF,$28,$F9
	.dc.b	$1E,$7F,$E4,$3D,$0F,$EB,$5C,$69,$CE,$1F,$8B,$9F,$DF,$CE,$63,$E3
	.dc.b	$D7,$F8,$86,$BE,$97,$E2,$76,$72,$7E,$4F,$83,$C0,$E3,$AF,$F8,$4F
	.dc.b	$1F,$FB,$13,$E1,$7E,$D7,$9B,$F8,$DE,$3F,$ED,$9E,$CF,$FA,$EE,$73
	.dc.b	$87,$E0,$B9,$7F,$95,$ED,$FE,$13,$8F,$B3,$87,$F8,$CE,$BF,$09,$F3
	.dc.b	$D6,$72,$7E,$2B,$2F,$03,$8E,$CF,$E3,$7A,$FF,$B5,$BE,$1B,$FC,$2B
	.dc.b	$F5,$EF,$CA,$1F,$9E,$BA,$1F,$8D,$7F,$AC,$BF,$0C,$7E,$95,$FE,$D1
	.dc.b	$C7,$8E,$1C,$2F,$7F,$6E,$B7,$E9,$7B,$DB,$FD,$6B,$6F,$A7,$FF,$26
	.dc.b	$63,$CB,$4B,$5A,$7C,$5A,$7F,$B0,$BF,$27,$F4,$4E,$74,$EF,$F9,$4E
	.dc.b	$23,$F4,$BE,$BF,$F8,$C3,$AB,$F9,$CE,$3E,$19,$C9,$FD,$06,$1E,$AE
	.dc.b	$7C,$64,$AF,$C6,$FE,$C3,$F6,$6B,$3A,$9E,$DC,$FE,$18,$C5,$7F,$4D
	.dc.b	$E6,$E7,$D6,$FA,$9F,$DD,$7A,$9F,$E4,$4F,$1B,$F6,$B3,$9B,$FD,$AC
	.dc.b	$E7,$5F,$F3,$DC,$E6,$7F,$80,$CF,$F9,$CE,$FF,$82,$7D,$1D,$9E,$27
	.dc.b	$F4,$7C,$FC,$33,$E0,$8F,$E9,$59,$79,$3D,$4F,$E1,$9D,$BF,$E5,$6F
	.dc.b	$0F,$F8,$66,$7F,$CD,$6F,$F8,$3C,$E7,$4F,$D2,$DF,$09,$FF,$56,$3F
	.dc.b	$C5,$7E,$03,$FD,$0C,$E6,$3F,$CF,$39,$F7,$79,$9F,$D1,$F2,$F8,$BF
	.dc.b	$C0,$3F,$C7,$7D,$2F,$F3,$3E,$5E,$28,$C5,$7E,$9F,$26,$69,$ED,$3F
	.dc.b	$ED,$D7,$6F,$F2,$EC,$7A,$BF,$C8,$30,$EB,$8F,$D4,$F8,$4F,$E5,$93
	.dc.b	$FC,$8F,$1F,$E0,$B1,$8E,$2F,$C5,$47,$82,$31,$3F,$86,$C3,$93,$FC
	.dc.b	$B0,$FE,$59,$7F,$80,$FC,$6F,$AF,$A1,$F7,$FC,$8F,$4C,$62,$BF,$93
	.dc.b	$6A,$8C,$43,$E5,$39,$3D,$6F,$FD,$17,$F9,$27,$E3,$3F,$18,$F4,$F9
	.dc.b	$FF,$67,$2F,$8B,$18,$BF,$ED,$F9,$23,$11,$F9,$D3,$67,$A6,$BF,$D2
	.dc.b	$FE,$67,$E6,$3F,$13,$F8,$AF,$D4,$C3,$E8,$C6,$2B,$F8,$D6,$08,$C4
	.dc.b	$7F,$A8,$23,$0F,$C5,$AF,$F5,$1F,$67,$E9,$3F,$7E,$FA,$2F,$FA,$9A
	.dc.b	$7F,$9A,$C6,$1F,$DD,$52,$31,$1D,$5B,$3F,$AE,$FD,$EF,$ED,$F6,$F3
	.dc.b	$F4,$3F,$0F,$18,$BF,$D8,$60,$8C,$47,$9D,$B7,$FB,$8F,$27,$F0,$BD
	.dc.b	$5E,$5F,$8D,$FC,$8C,$62,$FF,$51,$87,$F2,$97,$F0,$55,$EB,$76,$60
	.dc.b	$FE,$0B,$0F,$D8,$7F,$23,$0E,$3F,$EE,$71,$87,$FD,$93,$D7,$FA,$08
	.dc.b	$C4,$FD,$9A,$7D,$C7,$2E,$3F,$EF,$6D,$9F,$E3,$77,$FA,$88,$C6,$9F
	.dc.b	$C5,$3F,$2D,$F0,$F1,$7F,$80,$9B,$7C,$31,$46,$2B,$F8,$DF,$E0,$1E
	.dc.b	$87,$FC,$12,$DA,$E0,$D8,$FE,$43,$F9,$5F,$BB,$8F,$FC,$1E,$DA,$EC
	.dc.b	$D9,$FE,$4F,$FB,$9F,$C4,$E0,$FF,$0A,$B6,$BC,$8D,$9F,$E5,$BF,$BA
	.dc.b	$FC,$AD,$BF,$C3,$6D,$AF,$53,$67,$F9,$9F,$DC,$7D,$09,$FF,$11,$6A
	.dc.b	$8C,$78,$5F,$05,$B5,$FC,$EB,$ED,$B8,$72,$7E,$75,$B8,$32,$FE,$2B
	.dc.b	$1A,$AF,$E8,$3F,$B3,$F7,$E4,$FD,$23,$6F,$86,$68,$D5,$7F,$48,$FC
	.dc.b	$47,$CD,$C9,$FA,$B6,$DF,$07,$33,$FD,$3B,$F7,$AF,$63,$FB,$16,$CF
	.dc.b	$67,$31,$FD,$53,$F6,$CF,$85,$C9,$FB,$76,$D7,$91,$CC,$FF,$5A,$FD
	.dc.b	$9F,$E4,$72,$7E,$F5,$B5,$EA,$73,$3F,$D8,$3F,$1B,$F9,$DC,$1F,$C0
	.dc.b	$DB,$5F,$01,$CC,$FF,$66,$F8,$4F,$A9,$3F,$E5,$ED,$D1,$8E,$4F,$8E
	.dc.b	$EB,$F0,$1F,$ED,$BF,$87,$70,$E5,$B8,$7F,$33,$3F,$BD,$9F,$AD,$CF
	.dc.b	$FA,$4C,$F6,$F9,$5F,$EE,$BF,$B8,$F8,$39,$AD,$F1,$3F,$C0,$6F,$D6
	.dc.b	$8F,$2F,$93,$FD,$F3,$F7,$3F,$3F,$3D,$BE,$01,$FB,$B1,$EC,$FC,$5F
	.dc.b	$FC,$07,$FB,$1F,$93,$F1,$F7,$E1,$AF,$CB,$5F,$BA,$B5,$7E,$DE,$F9
	.dc.b	$E7,$CB,$7F,$62,$CB,$96,$3C,$70,$E1,$7B,$F9,$7F,$EC,$FF,$55,$F3
	.dc.b	$3B,$BF,$C2,$6D,$E6,$FE,$D7,$F8,$77,$D0,$7F,$C2,$ED,$7F,$B7,$FC
	.dc.b	$3F,$D4,$E4,$76,$61,$F7,$8F,$99,$FE,$EB,$F8,$46,$FB,$2C,$EB,$E6
	.dc.b	$3F,$BC,$FD,$FF,$3E,$5B,$3C,$BE,$5F,$FB,$EF,$DE,$77,$EC,$B3,$DB
	.dc.b	$E5,$FF,$C0,$3F,$47,$E3,$F1,$AC,$F8,$BE,$33,$FC,$0B,$F8,$97,$9F
	.dc.b	$CC,$B3,$E6,$F8,$C7,$E1,$AF,$B9,$FC,$2F,$E4,$22,$10,$7D,$AF,$08
	.dc.b	$DC,$64,$E1,$F7,$3D,$32,$CC,$67,$E8,$F1,$97,$F9,$32,$38,$1D,$FE
	.dc.b	$93,$C7,$BF,$08,$F0,$7E,$1D,$8A,$E1,$1C,$3F,$32,$F4,$F2,$6E,$3C
	.dc.b	$FF,$55,$8A,$DC,$79,$7F,$80,$3D,$FD,$3B,$8F,$77,$76,$2B,$71,$EF
	.dc.b	$EE,$F8,$BD,$DB,$8F,$F3,$1F,$22,$31,$E0,$7C,$AF,$03,$A7,$E3,$FA
	.dc.b	$6C,$7C,$EF,$0B,$C1,$F8,$97,$26,$C7,$D2,$F2,$BF,$33,$E6,$FE,$83
	.dc.b	$C3,$FC,$FF,$EC,$EE,$5D,$70,$7F,$9D,$3C,$FF,$D3,$DE,$FC,$B5,$FF
	.dc.b	$A6,$78,$19,$74,$F3,$3F,$47,$F7,$FE,$AF,$DD,$7E,$17,$5D,$D3,$E0
	.dc.b	$F2,$FE,$06,$3F,$9D,$62,$AF,$91,$F2,$DF,$0B,$D4,$F9,$5D,$97,$1E
	.dc.b	$38,$A3,$8C,$FC,$FE,$6F,$93,$EB,$7D,$1E,$BB,$87,$07,$BC,$7D,$5F
	.dc.b	$67,$E7,$FB,$3F,$5E,$A9,$D8,$F2,$BF,$6F,$CB,$FA,$BE,$D7,$E0,$FB
	.dc.b	$7F,$8F,$F0,$7F,$28,$BF,$6C,$7F,$80,$A3,$11,$F9,$28,$FF,$21,$43
	.dc.b	$DC,$FC,$67,$77,$26,$D7,$C4,$73,$87,$E1,$3D,$DF,$C9,$79,$7E,$06
	.dc.b	$D7,$CA,$73,$3F,$8A,$F7,$BF,$31,$E4,$F9,$9B,$5F,$41,$CC,$FE,$43
	.dc.b	$DF,$FC,$F7,$47,$B1,$B5,$F5,$9C,$CF,$E5,$BE,$03,$F4,$9C,$DF,$09
	.dc.b	$B5,$F7,$1C,$CF,$E7,$3E,$07,$F5,$5C,$BF,$21,$B5,$F8,$27,$33,$FA
	.dc.b	$2F,$82,$FD,$84,$FC,$E6,$BF,$88,$73,$1F,$A8,$F8,$3F,$DE,$71,$FF
	.dc.b	$97,$78,$FF,$C1,$D7,$FB,$08,$FF,$1E,$C6,$23,$F8,$64,$7F,$B2,$A1
	.dc.b	$F0,$9F,$C1,$79,$39,$36,$BE,$E3,$70,$FD,$BF,$C2,$FF,$0B,$E4,$F0
	.dc.b	$36,$BF,$04,$D9,$FD,$EF,$C3,$7F,$11,$E4,$F3,$36,$BF,$10,$D9,$FE
	.dc.b	$07,$F0,$FF,$C5,$F9,$3D,$8D,$AF,$C7,$36,$7F,$84,$FC,$47,$F1,$DE
	.dc.b	$4F,$84,$DA,$FC,$A3,$67,$F8,$7F,$C4,$FF,$23,$E4,$F9,$0D,$AF,$CD
	.dc.b	$36,$7F,$8A,$F9,$5F,$CC,$38,$3F,$D1,$5F,$9B,$7F,$3E,$8C,$5F,$F8
	.dc.b	$0B,$F4,$0C,$3F,$C9,$AB,$FC,$ED,$C7,$FC,$3E,$30,$7D,$1F,$8D,$FE
	.dc.b	$77,$C7,$FE,$95,$F3,$3F,$A4,$C6,$27,$FC,$1B,$F9,$6E,$1F,$E6,$35
	.dc.b	$D2,$75,$1F,$AE,$F9,$0F,$E8,$BC,$BF,$E7,$CC,$3F,$A9,$C6,$27,$FC
	.dc.b	$EF,$86,$1F,$E7,$75,$E6,$9D,$47,$EE,$3E,$4B,$FA,$7F,$2F,$F8,$7B
	.dc.b	$DE,$FF,$05,$69,$FE,$A7,$AF,$E8,$73,$9C,$7F,$B0,$DF,$09,$FC,$73
	.dc.b	$5F,$E1,$3F,$27,$FD,$6B,$9B,$FC,$53,$C5,$D2,$31,$21,$0B,$FE,$DB
	.dc.b	$3A,$97,$C6,$FF,$4A,$8C,$7C,$27,$F4,$CF,$95,$FE,$CF,$D1,$FE,$36
	.dc.b	$E1,$F4,$C6,$30,$ED,$1D,$76,$FF,$6D,$47,$F5,$EA,$FF,$1E,$FE,$59
	.dc.b	$CE,$30,$79,$7E,$EB,$FA,$CF,$CC,$FF,$74,$8F,$F5,$1E,$1F,$1A,$31
	.dc.b	$1F,$E7,$98,$ED,$18,$F6,$97,$CA,$BF,$B6,$9F,$EA,$CB,$FE,$CB,$E9
	.dc.b	$7F,$81,$3C,$3F,$A5,$8F,$D2,$8C,$57,$C7,$E5,$FF,$1F,$70,$FF,$84
	.dc.b	$A7,$33,$FC,$15,$B3,$E0,$8D,$1F,$E0,$2B,$FB,$D8,$7F,$74,$F9,$DF
	.dc.b	$F0,$9F,$5F,$D6,$CB,$F0,$31,$8B,$FA,$DC,$9F,$E4,$DF,$C4,$7F,$C4
	.dc.b	$33,$99,$FE,$28,$FF,$87,$3F,$21,$FF,$6C,$46,$8F,$F0,$8E,$1F,$82
	.dc.b	$DB,$C1,$18,$3A,$E5,$FE,$06,$FA,$0F,$F1,$2F,$77,$DB,$D1,$F9,$5F
	.dc.b	$F8,$C2,$73,$7F,$72,$5F,$8C,$66,$3F,$C3,$F7,$E2,$E6,$5F,$F8,$C3
	.dc.b	$E2,$B0,$F2,$AF,$F3,$46,$7E,$17,$89,$F8,$8C,$FF,$C3,$3F,$4B,$FE
	.dc.b	$38,$F0,$7F,$03,$A3,$F3,$3F,$F2,$1C,$E6,$FF,$05,$2F,$E0,$0C,$C7
	.dc.b	$F8,$CA,$FC,$95,$FE,$42,$F8,$CB,$39,$3F,$15,$97,$81,$C7,$57,$F9
	.dc.b	$F7,$3F,$F1,$5F,$D7,$7F,$A5,$3E,$D1,$E3,$BF,$21,$7E,$A0,$FC,$0D
	.dc.b	$F8,$53,$07,$97,$FE,$5A,$79,$61,$C6,$F8,$56,$FC,$AD,$6B,$FB,$7C
	.dc.b	$FB,$E7,$F3,$71,$E3,$FA,$CD,$AD,$65,$CF,$4F,$0F,$17,$A2,$73,$87
	.dc.b	$9A,$7C,$BE,$1E,$67,$FC,$D3,$E6,$7E,$7F,$99,$E0,$FF,$14,$D3,$FC
	.dc.b	$D3,$39,$8F,$DE,$46,$2B,$F9,$9B,$FE,$6E,$F3,$BC,$91,$D0,$78,$BF
	.dc.b	$80,$7D,$A6,$5F,$9C,$FF,$9F,$BE,$EB,$FC,$EF,$E7,$7E,$8B,$27,$E8
	.dc.b	$1F,$C9,$A7,$35,$FC,$6A,$73,$E4,$F1,$8F,$F3,$8D,$FE,$4E,$BF,$D0
	.dc.b	$3F,$27,$67,$27,$1E,$9F,$F3,$3C,$F4,$F2,$7F,$D7,$9F,$4B,$FE,$6A
	.dc.b	$FB,$CF,$F4,$4F,$9D,$FA,$19,$FE,$39,$18,$C3,$FD,$85,$A7,$8F,$C6
	.dc.b	$4E,$67,$FC,$14,$FF,$A4,$BE,$53,$F4,$CF,$1A,$FF,$65,$7A,$7D,$EB
	.dc.b	$FD,$13,$8F,$EB,$FD,$2F,$F3,$DF,$DF,$7F,$A6,$A7,$F9,$07,$E5,$9F
	.dc.b	$ED,$1F,$C3,$3F,$D4,$B3,$9F,$17,$FD,$A9,$3F,$AE,$D5,$3C,$6F,$FB
	.dc.b	$4B,$FE,$A4,$AF,$DC,$3C,$73,$5F,$F1,$BB,$E3,$7F,$E9,$C3,$C3,$C3
	.dc.b	$FE,$8E,$EE,$FF,$59,$FA,$7F,$89,$FC,$0F,$FD,$C9,$F8,$7F,$FA,$E2
	.dc.b	$73,$7F,$82,$C9,$5F,$EE,$9F,$8B,$D4,$FF,$AB,$F0,$FE,$36,$E0,$FF
	.dc.b	$5F,$2D,$7F,$87,$61,$D3,$AF,$FD,$E5,$87,$9E,$31,$8F,$6C,$FF,$D4
	.dc.b	$7E,$5F,$FB,$23,$D4,$FE,$5B,$E0,$3E,$17,$FA,$3E,$DF,$EC,$EA,$D6
	.dc.b	$DF,$D9,$AB,$5E,$6B,$83,$53,$FE,$C5,$C7,$CF,$C2,$C3,$B7,$07,$F2
	.dc.b	$37,$9F,$57,$EB,$34,$F9,$91,$8C,$FC,$5F,$75,$FE,$B8,$FC,$13,$FD
	.dc.b	$BB,$EA,$FF,$6A,$F5,$FE,$37,$E0,$5F,$EE,$29,$CE,$CE,$3F,$E6,$E3
	.dc.b	$89,$F8,$06,$E7,$FD,$B5,$8F,$9A,$FA,$EF,$FC,$DD,$E7,$87,$2C,$3C
	.dc.b	$57,$C0,$ED,$97,$7F,$47,$FD,$9B,$F8,$2F,$FB,$CB,$9F,$FB,$97,$AF
	.dc.b	$F2,$BF,$1C,$FF,$7B,$CE,$74,$F3,$4F,$CE,$FC,$6B,$39,$7F,$BB,$70
	.dc.b	$FB,$8F,$F8,$F2,$FA,$7D,$6F,$F4,$B7,$96,$BF,$D0,$7A,$BD,$D5,$C8
	.dc.b	$7C,$BB,$0F,$1C,$FF,$32,$97,$4B,$F0,$8F,$F1,$14,$7F,$B2,$7B,$FF
	.dc.b	$8F,$E3,$F2,$7F,$03,$7A,$5C,$2F,$79,$E2,$F7,$C8,$BF,$F7,$06,$5D
	.dc.b	$72,$FF,$17,$5E,$DF,$F2,$BE,$1C,$B1,$7E,$CB,$FD,$9E,$31,$97,$4F
	.dc.b	$23,$FD,$F3,$18,$EC,$FD,$37,$ED,$9F,$EF,$CF,$5B,$C7,$E8,$FF,$8E
	.dc.b	$BD,$3F,$F0,$FF,$ED,$4E,$A7,$D4,$7C,$0B,$DF,$8B,$FC,$AB,$99,$7B
	.dc.b	$FA,$8D,$71,$7A,$33,$FF,$22,$BF,$E6,$EC,$3B,$FC,$5F,$F7,$D7,$9E
	.dc.b	$9E,$0F,$BD,$F2,$57,$1C,$FB,$64,$FA,$1F,$F2,$87,$3B,$F1,$1F,$97
	.dc.b	$C9,$F9,$9B,$FC,$C7,$EA,$AE,$F7,$9D,$F6,$2F,$7D,$DF,$E5,$8E,$E5
	.dc.b	$EF,$C5,$D6,$DF,$43,$3F,$D2,$3F,$E8,$6C,$3B,$FB,$7F,$E1,$57,$9E
	.dc.b	$9E,$09,$F2,$C6,$31,$EB,$F2,$1E,$8F,$C5,$3F,$03,$E3,$FF,$24,$E9
	.dc.b	$7F,$00,$BF,$54,$7E,$2F,$FE,$04,$E4,$61,$C3,$D7,$FF,$29,$DE,$EB
	.dc.b	$F0,$F3,$9D,$7F,$D3,$6F,$6F,$07,$FC,$54,$F1,$CB,$BE,$4F,$7B,$C9
	.dc.b	$7F,$8E,$D3,$C9,$97,$8F,$EE,$3E,$7F,$A1,$F9,$EF,$67,$FB,$8F,$B3
	.dc.b	$F8,$0F,$29,$F9,$73,$F4,$EF,$CD,$71,$B0,$E1,$E5,$FF,$36,$F0,$7F
	.dc.b	$47,$7B,$FD,$CF,$F5,$75,$AF,$C0,$3F,$C8,$0F,$0C,$EF,$07,$FA,$3F
	.dc.b	$5F,$0E,$3C,$BE,$CB,$CB,$39,$CF,$B6,$DE,$4F,$82,$FC,$3F,$CA,$7E
	.dc.b	$DB,$DF,$7D,$8B,$E8,$9F,$95,$3E,$29,$FB,$57,$E2,$7F,$11,$FD,$FE
	.dc.b	$3C,$70,$E1,$7B,$AF,$C7,$7E,$4F,$AD,$BF,$D4,$99,$F6,$CF,$FC,$B5
	.dc.b	$8F,$1F,$A6,$B5,$AC,$B9,$E9,$E0,$8F,$39,$9B,$F2,$C7,$93,$C4,$FD
	.dc.b	$0F,$E4,$1F,$C4,$39,$BF,$CD,$BF,$0B,$FE,$B3,$FD,$E3,$FD,$67,$ED
	.dc.b	$3F,$61,$FE,$2F,$7B,$F3,$D7,$BF,$37,$F7,$DF,$85,$EB,$6F,$F5,$E6
	.dc.b	$7D,$1F,$F7,$2E,$1C,$7A,$FF,$CE,$8F,$3C,$39,$7A,$DE,$4A,$E3,$97
	.dc.b	$65,$AF,$FB,$1B,$D6,$FD,$CF,$E4,$FF,$CA,$3F,$6F,$7C,$A3,$F1,$D7
	.dc.b	$EF,$58,$FB,$1F,$F6,$4F,$47,$F3,$0C,$38,$5E,$EB,$5E,$07,$FB,$4B
	.dc.b	$C9,$56,$B1,$E7,$6C,$7F,$DA,$7E,$FF,$3D,$BF,$DB,$5F,$81,$3F,$2D
	.dc.b	$E3,$F8,$0F,$F9,$D3,$E9,$9F,$1C,$FC,$43,$86,$0C,$38,$6A,$C3,$85
	.dc.b	$FF,$40,$DB,$C7,$7D,$B2,$78,$9F,$E9,$DC,$3F,$A4,$E3,$CA,$BF,$A5
	.dc.b	$9F,$18,$E1,$1D,$7E,$87,$FD,$29,$CB,$99,$F1,$61,$E8,$9F,$F4,$C7
	.dc.b	$46,$7E,$5B,$F9,$47,$E0,$2F,$D7,$9F,$BA,$3F,$70,$7E,$9A,$FC,$AB
	.dc.b	$DB,$87,$0B,$DD,$6B,$E2,$B3,$B3,$2F,$56,$C3,$FC,$31,$1F,$C6,$DF
	.dc.b	$59,$C2,$3C,$59,$F8,$74,$F3,$57,$3C,$7B,$FE,$29,$FC,$FB,$8B,$FB
	.dc.b	$6F,$EA,$8F,$39,$F8,$AF,$F3,$DF,$D3,$DF,$AB,$7F,$1B,$FD,$23,$FB
	.dc.b	$AE,$1C,$2F,$7F,$22,$BD,$FC,$2B,$F5,$DD,$F1,$7A,$74,$ED,$A7,$5D
	.dc.b	$3F,$93,$7D,$86,$A9,$F6,$2F,$D6,$1F,$B7,$7F,$39,$F1,$9F,$83,$BF
	.dc.b	$66,$7D,$4F,$1C,$38,$5E,$EB,$5F,$A8,$FF,$83,$3E,$F3,$FD,$89,$FA
	.dc.b	$0B,$EC,$7F,$C8,$DF,$6E,$FC,$11,$C1,$FC,$5F,$CA,$61,$FE,$12,$FC
	.dc.b	$BB,$FC,$75,$F8,$E3,$F3,$27,$E8,$DF,$E1,$8F,$CD,$5F,$A0,$F7,$E7
	.dc.b	$FB,$1F,$89,$3F,$31,$FE,$77,$FA,$AF,$E5,$BF,$1E,$7E,$99,$F8,$7D
	.dc.b	$FB,$E1,$F7,$FF,$07,$FE,$AF,$F8,$AB,$F7,$8F,$F3,$9F,$E5,$3F,$ED
	.dc.b	$CF,$C2,$BF,$C4,$FF,$37,$FE,$26,$FD,$61,$FB,$0F,$C4,$FC,$B1,$BF
	.dc.b	$E4,$B8,$1F,$97,$F0,$DB,$F1,$98,$70,$F0,$D8,$70,$F6,$FD,$1F,$A6
	.dc.b	$EE,$D7,$DB,$F7,$7F,$C8,$C6,$63,$93,$F9,$4E,$36,$DE,$5D,$BC,$9B
	.dc.b	$7E,$93,$E7,$3F,$C1,$F3,$9D,$7C,$33,$FA,$8F,$DB,$B3,$8F,$E8,$7F
	.dc.b	$31,$7E,$5A,$FD,$E3,$FB,$CF,$4B,$F2,$6F,$F5,$4F,$ED,$BE,$7B,$DD
	.dc.b	$6A,$7F,$D5,$BF,$BF,$E7,$05,$7A,$B8,$3C,$F9,$77,$9E,$CF,$AF,$6F
	.dc.b	$F2,$9D,$74,$FB,$3F,$B3,$8F,$F9,$6E,$BA,$65,$DF,$F1,$5F,$4C,$3F
	.dc.b	$4C,$FF,$13,$E5,$FE,$28,$F9,$FF,$E0,$BD,$5F,$E6,$4D,$7C,$37,$BF
	.dc.b	$86,$F7,$8A,$DE,$DF,$4F,$1E,$75,$E5,$3F,$83,$3F,$E3,$4F,$D1,$3D
	.dc.b	$93,$FE,$6A,$8F,$0D,$F3,$EB,$7E,$6F,$E6,$9C,$E3,$FC,$63,$87,$1F
	.dc.b	$2D,$AF,$C0,$3F,$DA,$BF,$80,$3F,$6A,$7E,$F5,$E3,$AD,$7E,$2B,$FE
	.dc.b	$40,$9F,$D1,$F0,$FB,$33,$FE,$A5,$A7,$93,$87,$CF,$83,$C2,$C4,$7C
	.dc.b	$78,$E7,$9F,$EB,$A3,$B5,$7F,$AA,$F2,$FF,$26,$FE,$0D,$E9,$F8,$BF
	.dc.b	$E0,$FF,$83,$79,$39,$BF,$79,$D5,$F1,$E7,$3C,$1F,$A2,$F5,$DF,$97
	.dc.b	$E7,$F3,$9F,$0F,$0F,$F3,$BF,$7E,$E1,$C3,$2F,$F4,$7E,$FC,$E5,$A7
	.dc.b	$7F,$AF,$F9,$5F,$A8,$FF,$23,$87,$D8,$BF,$2A,$7E,$9E,$FB,$6F,$B5
	.dc.b	$7B,$AD,$4B,$ED,$73,$FB,$DF,$92,$BF,$97,$69,$F7,$CF,$C6,$3C,$71
	.dc.b	$7E,$1F,$E2,$F9,$2F,$2E,$5E,$1F,$D9,$BC,$BF,$AB,$FF,$37,$FC,$01
	.dc.b	$FB,$23,$F5,$D7,$EB,$5E,$4A,$D4,$E7,$F6,$0C,$64,$C3,$8F,$A9,$D2
	.dc.b	$31,$5F,$E2,$FF,$17,$F0,$1B,$B1,$FF,$5B,$FE,$38,$FD,$4D,$F9,$4B
	.dc.b	$F4,$4F,$07,$E0,$8A,$D4,$FF,$A3,$F1,$E7,$87,$2C,$1C,$3E,$CC,$38
	.dc.b	$47,$CD,$E5,$F5,$75,$7F,$88,$7F,$51,$7E,$84,$F8,$8C,$F9,$89,$CF
	.dc.b	$E5,$78,$7F,$A3,$61,$7E,$E3,$E4,$FB,$7F,$D1,$BF,$2F,$FB,$9E,$6F
	.dc.b	$C7,$8C,$77,$FF,$49,$F6,$3F,$61,$CF,$F0,$67,$87,$8B,$CB,$E5,$58
	.dc.b	$BF,$39,$EB,$E6,$BF,$5B,$FF,$57,$7E,$4D,$D6,$31,$F4,$EF,$D6,$31
	.dc.b	$5F,$EE,$6E,$CF,$8B,$A7,$E7,$21,$F8,$57,$A3,$E8,$1F,$BB,$BF,$7B
	.dc.b	$FE,$99,$87,$09,$CC,$3E,$87,$D9,$1C,$3E,$C3,$95,$BE,$C7,$A7,$B1
	.dc.b	$FE,$01,$FA,$DD,$BF,$E0,$3F,$9F,$BF,$49,$AD,$7C,$5F,$A0,$7B,$FF
	.dc.b	$06,$EA,$FC,$1F,$CD,$6F,$3E,$BA,$BD,$1F,$40,$FD,$B9,$FB,$B7,$92
	.dc.b	$31,$E9,$DF,$8E,$DF,$07,$0F,$75,$AE,$DF,$F6,$BF,$47,$BF,$0F,$C4
	.dc.b	$F8,$DC,$FD,$8F,$F1,$84,$E5,$7D,$26,$73,$D6,$8F,$C6,$7F,$C9,$7E
	.dc.b	$23,$F3,$A7,$E6,$1F,$E2,$1F,$C9,$5F,$93,$7F,$99,$78,$78,$6D,$FE
	.dc.b	$3B,$3F,$C1,$F7,$C3,$ED,$D7,$C3,$EC,$FF,$C1,$D9,$F6,$AF,$F4,$05
	.dc.b	$FF,$96,$A3,$19,$7F,$72,$6E,$1C,$B2,$FF,$59,$57,$80,$FF,$25,$67
	.dc.b	$E7,$BF,$34,$FE,$46,$74,$3F,$44,$FD,$AF,$C8,$BF,$43,$7E,$D5,$F0
	.dc.b	$7F,$2C,$76,$F0,$BD,$D6,$B5,$53,$F2,$8D,$7C,$07,$F8,$83,$3F,$F2
	.dc.b	$CD,$7F,$9D,$30,$EF,$18,$3E,$33,$D3,$2F,$A1,$8F,$F9,$8E,$B8,$5B
	.dc.b	$3F,$35,$7D,$5F,$57,$58,$F2,$FC,$CF,$FD,$D5,$E3,$FE,$B7,$E1,$FF
	.dc.b	$C0,$3F,$8B,$FA,$67,$3E,$7A,$7F,$C0,$BF,$95,$67,$F2,$4F,$9D,$8F
	.dc.b	$3F,$C9,$2B,$FE,$75,$E1,$63,$CB,$0F,$F0,$6C,$F1,$CF,$C1,$F6,$DE
	.dc.b	$48,$FF,$3E,$7D,$FB,$F7,$0F,$9D,$F9,$BB,$F3,$6F,$D0,$7E,$F4,$FB
	.dc.b	$7F,$AB,$7B,$E8,$5E,$FC,$D5,$BF,$50,$DE,$5F,$67,$3F,$98,$78,$1F
	.dc.b	$F5,$63,$83,$FC,$FE,$F8,$B3,$FD,$87,$D1,$79,$E7,$39,$78,$27,$F6
	.dc.b	$31,$DA,$31,$8F,$8B,$F5,$CF,$3F,$E4,$0F,$B7,$FF,$33,$F8,$7F,$E5
	.dc.b	$2F,$CF,$3F,$90,$78,$1F,$EB,$4E,$15,$EF,$8F,$F5,$7F,$40,$BF,$F4
	.dc.b	$0D,$EE,$9F,$B1,$FD,$2D,$BF,$D2,$F1,$DD,$F6,$63,$D3,$3F,$0E,$3F
	.dc.b	$B4,$9F,$2C,$63,$B3,$DD,$FB,$0F,$F5,$DF,$CF,$BE,$5F,$E7,$CF,$CC
	.dc.b	$7F,$C2,$DF,$AC,$FE,$9A,$B5,$D2,$AD,$7D,$6B,$F3,$0D,$7D,$0B,$EB
	.dc.b	$BC,$D8,$75,$DB,$E8,$3C,$B0,$ED,$D7,$E2,$B6,$BE,$2D,$7F,$17,$C9
	.dc.b	$FE,$DA,$C3,$FC,$DB,$F9,$77,$FA,$2B,$EF,$1F,$9F,$7F,$18,$F0,$FC
	.dc.b	$DF,$2A,$AD,$4F,$F1,$CF,$B6,$CC,$79,$25,$FB,$0D,$8D,$FC,$6F,$C4
	.dc.b	$9E,$5F,$7D,$FE,$8C,$C3,$F8,$DE,$9E,$39,$CE,$7E,$2F,$80,$FF,$80
	.dc.b	$13,$FE,$F0,$FC,$6B,$F8,$4F,$E7,$8F,$CB,$BF,$C2,$3F,$29,$FD,$53
	.dc.b	$EE,$3F,$88,$D6,$A7,$3F,$27,$FE,$AE,$C3,$97,$B0,$C7,$F0,$17,$FC
	.dc.b	$53,$E9,$2C,$E7,$0E,$BC,$9F,$99,$AE,$9F,$57,$FD,$87,$F1,$AC,$5E
	.dc.b	$E3,$FB,$26,$6F,$D5,$BC,$7C,$3F,$13,$F6,$7F,$F6,$9F,$D9,$7F,$92
	.dc.b	$3F,$49,$CF,$4A,$73,$F4,$1F,$EC,$BE,$2F,$81,$87,$F9,$03,$C2,$FD
	.dc.b	$76,$FF,$EA,$4C,$3D,$33,$D7,$ED,$FF,$C7,$3E,$A7,$9E,$31,$A7,$F3
	.dc.b	$5C,$79,$DF,$7B,$FF,$1F,$E4,$FD,$27,$F9,$CF,$E1,$1F,$E3,$5F,$B6
	.dc.b	$7C,$C3,$F5,$DF,$F1,$97,$E1,$FF,$E5,$3C,$FF,$A9,$70,$30,$E1,$7B
	.dc.b	$FB,$4F,$F6,$87,$CE,$74,$F2,$BF,$CB,$77,$FC,$1E,$FF,$C8,$5E,$B7
	.dc.b	$EC,$5E,$39,$FF,$8F,$F1,$FE,$11,$18,$C5,$F2,$B8,$BF,$EC,$FF,$AC
	.dc.b	$FF,$86,$FF,$20,$7C,$DB,$F1,$9F,$5D,$6B,$EA,$55,$AF,$D6,$F3,$F9
	.dc.b	$A7,$86,$BF,$B8,$EF,$E5,$78,$57,$E5,$A3,$FB,$86,$BF,$C3,$BE,$09
	.dc.b	$A7,$87,$CF,$FE,$3D,$F8,$9B,$F6,$C7,$ED,$7A,$E9,$FC,$27,$E6,$6E
	.dc.b	$DF,$EC,$F7,$BA,$D7,$8A,$A7,$EE,$FA,$C5,$F2,$7F,$C5,$F0,$E1,$BF
	.dc.b	$F3,$17,$96,$BF,$E8,$8C,$5F,$8A,$79,$30,$F1,$CE,$7F,$41,$C6,$9E
	.dc.b	$0D,$1F,$5F,$FD,$EB,$F6,$4F,$EE,$1F,$36,$FC,$19,$FA,$A3,$F3,$07
	.dc.b	$EA,$BF,$E8,$8F,$1B,$F0,$78,$70,$BD,$D6,$B6,$E9,$F4,$3F,$DB,$F8
	.dc.b	$FF,$A4,$5F,$F7,$DF,$7F,$F9,$93,$C7,$3E,$DF,$61,$FD,$26,$31,$8F
	.dc.b	$F5,$3C,$FF,$23,$F9,$AB,$F4,$97,$EA,$EF,$C9,$DF,$95,$E7,$E4,$9F
	.dc.b	$B8,$7F,$85,$B0,$E1,$7B,$AD,$7E,$D1,$9C,$7F,$B0,$70,$7F,$89,$CF
	.dc.b	$F1,$85,$F0,$7F,$B0,$3C,$B2,$EF,$2C,$BF,$B0,$4E,$70,$F3,$7E,$01
	.dc.b	$FE,$11,$F8,$DF,$F1,$D7,$8B,$FE,$27,$FA,$5F,$F5,$57,$D9,$FF,$9A
	.dc.b	$7E,$67,$FA,$ED,$6A,$7F,$C5,$9F,$88,$7F,$96,$BE,$FD,$5F,$D7,$2B
	.dc.b	$5F,$39,$FE,$3B,$78,$5D,$E0,$1F,$E4,$E7,$9E,$3F,$E7,$7F,$95,$F2
	.dc.b	$4E,$B3,$F2,$67,$DF,$F1,$9C,$CB,$F6,$97,$D9,$BF,$15,$7E,$E0,$F8
	.dc.b	$07,$D2,$BF,$70,$7E,$B5,$FE,$61,$E1,$FF,$2D,$65,$CB,$1E,$38,$70
	.dc.b	$B7,$E6,$7B,$D3,$D3,$7E,$0F,$C2,$BD,$91,$E3,$D7,$C4,$FF,$91,$32
	.dc.b	$EB,$1E,$5D,$7C,$9A,$BE,$4B,$D7,$8F,$86,$73,$F2,$78,$BE,$1A,$F8
	.dc.b	$FF,$5A,$F1,$67,$F0,$7F,$43,$7E,$96,$FD,$09,$F9,$13,$D9,$FE,$79
	.dc.b	$F7,$BC,$36,$61,$C2,$F7,$5F,$D0,$7F,$07,$D4,$F4,$CB,$D4,$7C,$05
	.dc.b	$FA,$7F,$C0,$C3,$9E,$7E,$1C,$FC,$1F,$63,$EC,$BE,$D1,$E3,$EA,$C5
	.dc.b	$6F,$3F,$EA,$1E,$97,$97,$F6,$2F,$F7,$DF,$C0,$7F,$84,$FF,$6B,$7E
	.dc.b	$FB,$FE,$44,$F4,$7F,$C3,$5C,$1F,$4A,$F7,$F5,$8B,$DF,$9F,$CF,$F7
	.dc.b	$DF,$35,$3E,$82,$9F,$C7,$2F,$87,$E4,$7F,$4A,$79,$E1,$FE,$AC,$CB
	.dc.b	$BE,$5F,$8F,$FC,$D1,$F8,$BF,$E4,$79,$BF,$D4,$FF,$2E,$FD,$3F,$E7
	.dc.b	$56,$A7,$F2,$DE,$5F,$F5,$9F,$D5,$B3,$F3,$7F,$26,$B9,$E5,$F3,$8E
	.dc.b	$D9,$75,$BF,$EB,$91,$D2,$31,$F8,$77,$E8,$BF,$27,$FB,$DF,$AB,$BF
	.dc.b	$4F,$FD,$1F,$63,$F7,$0F,$E7,$DC,$3E,$AB,$DD,$6A,$7F,$97,$CE,$7B
	.dc.b	$18,$7E,$72,$FF,$B8,$67,$67,$38,$76,$BF,$14,$63,$C7,$65,$FD,$9A
	.dc.b	$FF,$2F,$87,$FB,$83,$EB,$9E,$FB,$EE,$9F,$A0,$3F,$60,$FE,$7B,$A7
	.dc.b	$9F,$D9,$FF,$13,$5E,$EB,$5F,$9D,$F4,$9F,$F1,$56,$7F,$27,$3F,$F7
	.dc.b	$6F,$D9,$7A,$67,$8E,$7D,$B9,$FE,$44,$63,$1F,$F0,$A7,$A9,$FE,$DB
	.dc.b	$C3,$CD,$9F,$7D,$5F,$8A,$7A,$B2,$7E,$BE,$F4,$D8,$7F,$0C,$FA,$F7
	.dc.b	$EE,$BB,$AD,$4E,$7F,$80,$31,$C3,$E7,$C5,$F9,$BF,$CF,$7A,$E5,$D3
	.dc.b	$DF,$FC,$0F,$2F,$D9,$9C,$E3,$FC,$6B,$E1,$BF,$04,$B9,$3F,$90,$7E
	.dc.b	$C1,$FE,$6A,$FA,$AF,$C0,$F2,$7B,$6B,$5E,$DA,$B5,$E8,$FD,$69,$FF
	.dc.b	$77,$EF,$FE,$F5,$BE,$1F,$01,$F8,$39,$E3,$7D,$6B,$86,$5E,$0F,$BD
	.dc.b	$FF,$1B,$C6,$32,$ED,$1F,$E2,$8F,$99,$FE,$33,$F1,$FF,$86,$FC,$69
	.dc.b	$FA,$AB,$F5,$8F,$F2,$17,$E9,$8F,$D6,$F7,$E2,$AF,$75,$AF,$77,$3F
	.dc.b	$A6,$7F,$32,$CB,$A6,$DF,$AB,$7B,$67,$D7,$ED,$7F,$D1,$B3,$D3,$1E
	.dc.b	$71,$E2,$8F,$E5,$BF,$79,$F9,$3F,$8E,$79,$1F,$E9,$4F,$D8,$9F,$B2
	.dc.b	$BF,$41,$DF,$90,$BD,$F2,$AF,$7F,$A1,$6B,$F4,$DF,$E8,$F9,$F5,$D3
	.dc.b	$F8,$03,$DF,$4E,$DF,$94,$FF,$5E,$8C,$63,$CE,$BF,$B0,$FA,$1F,$DA
	.dc.b	$BF,$37,$FF,$3C,$7E,$46,$F4,$BF,$C1,$BF,$B5,$FF,$46,$AD,$7E,$0A
	.dc.b	$AD,$7D,$13,$F4,$4D,$64,$FD,$8F,$FA,$DE,$1D,$63,$F8,$A3,$CA,$FB
	.dc.b	$76,$F8,$A5,$A7,$8B,$47,$E6,$9F,$DF,$BC,$D7,$E6,$7F,$E0,$EF,$95
	.dc.b	$7E,$54,$F1,$5C,$9F,$C2,$3F,$36,$7E,$2A,$C7,$8E,$BF,$E0,$3C,$78
	.dc.b	$E1,$C3,$C6,$60,$F6,$78,$69,$DF,$EF,$BF,$D0,$B8,$F8,$E3,$C5,$AE
	.dc.b	$23,$67,$57,$C9,$AF,$F1,$AE,$9E,$13,$9C,$FC,$1E,$BE,$21,$EA,$3F
	.dc.b	$0D,$7E,$B9,$FE,$71,$F0,$5F,$AF,$7F,$04,$F9,$8E,$F8,$70,$BD,$FD
	.dc.b	$9A,$F7,$EF,$FF,$6E,$E6,$F8,$37,$BC,$BD,$11,$8C,$FC,$1F,$AC,$63
	.dc.b	$8B,$FC,$5B,$7B,$CB,$A6,$BF,$ED,$E9,$CD,$F9,$F1,$F3,$4B,$F0,$1F
	.dc.b	$67,$C8,$3F,$1D,$7E,$B9,$FB,$6F,$53,$EA,$FD,$83,$F1,$2F,$5D,$EE
	.dc.b	$B5,$DE,$AF,$F3,$1F,$0F,$AA,$BA,$7E,$8F,$E7,$E1,$FF,$15,$E7,$D9
	.dc.b	$FA,$F5,$B8,$F0,$FA,$CD,$3F,$0B,$E2,$79,$A7,$31,$F8,$8F,$73,$EA
	.dc.b	$7C,$F7,$ED,$7F,$03,$7E,$24,$FD,$21,$FA,$83,$F2,$6F,$F0,$46,$BF
	.dc.b	$3B,$0E,$17,$BA,$D7,$E3,$9F,$62,$7A,$6F,$E6,$EA,$F9,$F1,$FC,$0F
	.dc.b	$F2,$DF,$64,$FF,$A0,$7D,$CF,$F5,$3E,$3C,$F6,$F2,$C6,$35,$F1,$FD
	.dc.b	$F6,$3F,$75,$F3,$FC,$C3,$E8,$FF,$C3,$1F,$9C,$F8,$70,$56,$A7,$FC
	.dc.b	$39,$F8,$1E,$63,$C3,$B7,$F1,$5F,$D3,$2D,$F0,$D7,$FC,$0D,$83,$1F
	.dc.b	$04,$63,$D4,$F2,$FE,$87,$FE,$C9,$FC,$C7,$C5,$F5,$0F,$C8,$3C,$53
	.dc.b	$9C,$7F,$C6,$1F,$87,$E3,$93,$E2,$1F,$E7,$ED,$BF,$93,$7E,$49,$E2
	.dc.b	$BE,$1F,$59,$DB,$CC,$69,$EF,$FC,$43,$E2,$FE,$B9,$FE,$D9,$FA,$0F
	.dc.b	$85,$F5,$2F,$CA,$FE,$14,$E7,$7F,$F1,$9F,$E3,$58,$DF,$C2,$7D,$1F
	.dc.b	$C9,$3E,$0D,$6F,$BB,$AE,$BF,$DF,$30,$76,$FE,$EF,$E6,$DF,$80,$3F
	.dc.b	$31,$7E,$7F,$FC,$A7,$F5,$87,$EE,$1F,$E6,$FF,$27,$F6,$B8,$70,$BD
	.dc.b	$FC,$5A,$F7,$F9,$0F,$F0,$0A,$EB,$F8,$2F,$E1,$9E,$1C,$5E,$87,$96
	.dc.b	$7D,$FF,$0E,$F2,$6B,$E3,$8F,$17,$E7,$58,$ED,$F2,$FD,$13,$F3,$D7
	.dc.b	$EB,$EF,$DA,$FC,$18,$AF,$75,$AF,$D4,$F3,$F5,$FE,$3A,$D6,$DE,$37
	.dc.b	$86,$9F,$EF,$4B,$FF,$66,$4F,$4C,$39,$FC,$27,$9B,$E1,$3E,$3F,$91
	.dc.b	$F0,$7E,$01,$FA,$B7,$8A,$73,$CD,$FC,$FB,$A7,$CB,$F0,$DE,$19,$9A
	.dc.b	$FE,$3E,$BF,$C4,$3E,$BF,$F8,$D7,$7F,$E5,$B3,$9D,$3C,$0E,$F3,$FF
	.dc.b	$30,$65,$F5,$29,$E1,$7D,$AE,$8F,$8D,$FA,$A7,$F8,$2E,$B5,$39,$E2
	.dc.b	$4F,$F0,$F9,$E5,$F0,$DF,$16,$A6,$5F,$E4,$0B,$7E,$21,$F5,$A7,$97
	.dc.b	$DE,$7F,$45,$FD,$B5,$F8,$1F,$C2,$F1,$BF,$DA,$9F,$2F,$FE,$D7,$F9
	.dc.b	$2F,$E8,$75,$AE,$75,$6B,$F1,$1F,$D2,$EF,$FD,$0A,$B5,$D7,$F9,$E7
	.dc.b	$AF,$DF,$FF,$10,$79,$5F,$6F,$D3,$7C,$5A,$78,$7E,$C7,$FB,$3F,$B7
	.dc.b	$FD,$67,$EB,$DC,$0F,$C7,$9F,$B2,$3F,$2E,$F4,$FB,$5F,$A2,$E0,$FE
	.dc.b	$A7,$87,$0B,$DD,$6A,$FF,$AE,$CF,$4F,$C7,$FF,$91,$BB,$E1,$FD,$AB
	.dc.b	$C6,$7B,$6A,$CB,$C9,$9F,$8E,$73,$B7,$F6,$FF,$49,$FA,$97,$FA,$5B
	.dc.b	$EC,$7F,$A7,$7C,$77,$F8,$0B,$F7,$4F,$ED,$B5,$AF,$C1,$BF,$D1,$9F
	.dc.b	$2E,$AF,$F0,$3F,$EE,$5A,$D7,$FC,$05,$C1,$C3,$3F,$F5,$0B,$FC,$2B
	.dc.b	$1E,$59,$F6,$FA,$1F,$15,$78,$72,$E9,$F4,$38,$D3,$BE,$9F,$E1,$4F
	.dc.b	$C4,$BF,$BF,$FD,$4B,$77,$E6,$8F,$CF,$DF,$A5,$BF,$0E,$7C,$4F,$FA
	.dc.b	$93,$B7,$FC,$03,$8F,$1C,$38,$5E,$EB,$9D,$6B,$1E,$BC,$7F,$C8,$B0
	.dc.b	$F0,$7E,$13,$FD,$51,$E5,$9F,$83,$3E,$F5,$E6,$BF,$2C,$E7,$97,$FC
	.dc.b	$3D,$F5,$2F,$C9,$9F,$8D,$FF,$99,$3F,$23,$7E,$BE,$FC,$A9,$F0,$DF
	.dc.b	$E1,$EC,$38,$5E,$EB,$5F,$2D,$9C,$5F,$BA,$7F,$68,$BD,$FD,$EB,$3E
	.dc.b	$B9,$F8,$AF,$C3,$39,$C7,$D3,$C7,$FE,$26,$EA,$7E,$94,$F8,$7F,$F3
	.dc.b	$CF,$EF,$FF,$9D,$F8,$0F,$F0,$0D,$EE,$B5,$8A,$B5,$EE,$7F,$8D,$B1
	.dc.b	$E7,$8F,$2F,$37,$FD,$3B,$8F,$18,$C7,$E1,$9F,$E4,$3F,$DA,$DF,$B1
	.dc.b	$7F,$9A,$FE,$15,$FA,$63,$F3,$B7,$A2,$F5,$7F,$C8,$78,$70,$FB,$F6
	.dc.b	$1C,$2F,$7F,$B5,$6B,$27,$EC,$9F,$E0,$EC,$38,$5E,$FE,$D7,$C5,$5E
	.dc.b	$1D,$3C,$07,$F9,$5F,$CC,$FF,$1D,$FC,$57,$FA,$5F,$ED,$DF,$8F,$3E
	.dc.b	$AF,$FD,$BB,$C9,$FE,$50,$BD,$FE,$15,$FE,$DC,$AD,$7E,$15,$FE,$5A
	.dc.b	$9C,$E1,$CB,$EA,$7B,$7D,$CF,$FA,$03,$0E,$1F,$84,$E3,$4F,$06,$5D
	.dc.b	$FA,$F9,$5F,$FA,$1B,$E3,$FF,$DD,$3F,$8D,$FF,$12,$FC,$63,$38,$7F
	.dc.b	$34,$D1,$3F,$C5,$B6,$CD,$F0,$FB,$AE,$CF,$2F,$72,$DF,$0F,$81,$FF
	.dc.b	$77,$F1,$7B,$74,$FF,$4A,$FC,$DF,$83,$F5,$CF,$83,$AF,$F9,$AB,$93
	.dc.b	$E3,$CE,$75,$F1,$BE,$6F,$BB,$F6,$3E,$8F,$A6,$BA,$79,$B1,$EF,$8F
	.dc.b	$3C,$BF,$D4,$7F,$85,$FB,$BF,$36,$EF,$EC,$86,$BF,$5A,$75,$AF,$C2
	.dc.b	$7E,$2F,$5F,$CB,$7E,$3F,$C0,$5D,$3E,$2E,$3E,$EF,$D6,$FF,$D0,$7F
	.dc.b	$8F,$BF,$00,$7E,$0E,$9F,$F5,$8F,$DB,$BF,$2E,$7E,$33,$FE,$B1,$9E
	.dc.b	$37,$C1,$DC,$FF,$AF,$5C,$DF,$3F,$6B,$FD,$C9,$F9,$CE,$78,$BD,$99
	.dc.b	$F8,$6B,$59,$FF,$AA,$27,$B1,$AC,$BC,$BF,$91,$F9,$2B,$FC,$13,$D5
	.dc.b	$E7,$AD,$42,$75,$3C,$A5,$87,$79,$F6,$FD,$27,$8E,$73,$87,$6C,$FC
	.dc.b	$F1,$C6,$FC,$DF,$B6,$FC,$4F,$CF,$BE,$1E,$5E,$1A,$E3,$9A,$B5,$1C
	.dc.b	$71,$F9,$B1,$E0,$8C,$4E,$F1,$69,$FE,$2A,$C3,$A6,$5F,$17,$F6,$FF
	.dc.b	$A7,$F5,$BF,$03,$83,$E6,$C6,$2B,$9B,$F6,$3E,$BF,$D9,$7C,$71,$EB
	.dc.b	$BF,$BF,$F7,$A7,$E4,$BF,$EA,$1F,$C8,$5F,$86,$3F,$6F,$7E,$2A,$FD
	.dc.b	$83,$E5,$61,$C3,$F0,$96,$1C,$2F,$7F,$84,$EB,$2E,$99,$73,$F8,$DF
	.dc.b	$04,$62,$BF,$D1,$DC,$DF,$57,$57,$DC,$FF,$A7,$7F,$7F,$FF,$70,$7D
	.dc.b	$37,$E5,$3D,$7D,$4E,$7F,$12,$FF,$5C,$F5,$FA,$3A,$3E,$96,$3C,$BE
	.dc.b	$FD,$97,$28,$FB,$C6,$5E,$BE,$0F,$E0,$7C,$27,$8D,$FE,$DD,$FD,$25
	.dc.b	$FC,$01,$F6,$FA,$7F,$86,$E7,$F7,$EF,$7F,$89,$AF,$7E,$1F,$C7,$CB
	.dc.b	$F2,$15,$A9,$CF,$BB,$DF,$3E,$D9,$FE,$1F,$DD,$FC,$97,$4F,$F4,$FF
	.dc.b	$C1,$DF,$A1,$7F,$B7,$72,$77,$FF,$A4,$B7,$FC,$85,$EE,$B5,$2F,$D2
	.dc.b	$FF,$41,$39,$FC,$57,$F7,$5E,$2F,$6F,$91,$67,$DA,$7F,$7D,$CD,$FA
	.dc.b	$3F,$AC,$FF,$44,$7E,$27,$F1,$3E,$A9,$FB,$3F,$F8,$4F,$1F,$D6,$56
	.dc.b	$B8,$D5,$AF,$A3,$FF,$37,$F9,$BF,$C6,$7C,$3F,$F2,$5F,$DD,$B3,$E9
	.dc.b	$9B,$9B,$C1,$39,$CB,$9C,$F1,$E5,$FD,$77,$E1,$EF,$0B,$FC,$A7,$E2
	.dc.b	$BF,$4A,$FF,$7C,$FE,$EE,$FD,$8F,$F8,$05,$EF,$E1,$7F,$DF,$F5,$AE
	.dc.b	$FC,$E3,$CA,$3F,$E0,$0F,$E1,$3F,$C0,$EF,$86,$5D,$7F,$18,$FD,$A7
	.dc.b	$1F,$AF,$F0,$A7,$C0,$BF,$5A,$7E,$0A,$FC,$5D,$C4,$FD,$A5,$F9,$4F
	.dc.b	$F1,$3C,$78,$E1,$C2,$F7,$4F,$B7,$D6,$5D,$72,$E9,$F8,$1F,$C3,$D3
	.dc.b	$C5,$39,$C3,$BF,$C9,$7F,$1B,$CF,$AE,$1F,$ED,$7D,$7C,$5A,$F4,$D7
	.dc.b	$CF,$F8,$87,$83,$5C,$5F,$FB,$77,$5E,$DD,$DF,$D4,$72,$E7,$7D,$A3
	.dc.b	$7F,$87,$7A,$32,$B9,$FB,$F2,$CE,$3C,$AD,$8F,$93,$0E,$57,$FC,$A6
	.dc.b	$3C,$FF,$86,$F3,$8C,$6F,$F0,$63,$CD,$18,$D7,$E7,$C7,$FB,$2B,$0F
	.dc.b	$81,$5B,$A6,$DF,$CF,$7E,$A3,$FD,$F3,$F6,$5F,$EF,$8F,$CF,$7F,$DC
	.dc.b	$9F,$AE,$FF,$B8,$EF,$F9,$D5,$6B,$F0,$D5,$6B,$5F,$E9,$53,$98,$FF
	.dc.b	$76,$74,$75,$FC,$3B,$FC,$BB,$7C,$3C,$47,$66,$74,$4B,$E1,$BC,$3F
	.dc.b	$31,$D3,$0E,$31,$DF,$5F,$1C,$63,$16,$5C,$F0,$E5,$1D,$B8,$79,$65
	.dc.b	$DB,$9B,$AE,$3C,$A3,$14,$E0,$FA,$73,$9C,$3F,$21,$3D,$B0,$E7,$18
	.dc.b	$D7,$C1,$F5,$1D,$3B,$3F,$DE,$19,$7B,$23,$D1,$18,$A4,$7C,$6B,$E6
	.dc.b	$B0,$F4,$63,$FD,$A7,$F0,$2F,$D3,$7E,$1D,$FE,$A3,$FD,$05,$FA,$43
	.dc.b	$EC,$BD,$17,$BF,$0B,$FD,$DB,$F7,$AB,$DD,$3F,$5B,$FF,$00,$65,$D2
	.dc.b	$B5,$8F,$E1,$35,$CE,$3C,$B1,$EF,$F4,$5F,$E0,$3F,$C3,$3F,$DE,$BF
	.dc.b	$BB,$FF,$BC,$3F,$51,$7E,$24,$FC,$45,$FA,$0F,$F8,$36,$F7,$F8,$8A
	.dc.b	$F7,$4F,$C9,$F5,$F8,$83,$2E,$99,$3F,$10,$F0,$4E,$71,$E5,$3F,$4F
	.dc.b	$97,$FC,$1D,$F9,$E3,$EA,$5F,$BA,$BF,$17,$FE,$ED,$D0,$FD,$C7,$58
	.dc.b	$70,$BD,$D6,$A5,$FA,$07,$F8,$97,$1E,$98,$F3,$F9,$4F,$95,$CB,$FE
	.dc.b	$25,$9C,$E1,$DB,$ED,$7F,$C5,$1F,$80,$3E,$1F,$F9,$77,$ED,$4E,$2F
	.dc.b	$F7,$C7,$87,$FE,$31,$AD,$7B,$CF,$80,$5E,$FF,$0D,$FF,$4C,$F1,$FF
	.dc.b	$8A,$EF,$75,$AC,$55,$E0,$CB,$BC,$7C,$9F,$03,$FC,$7D,$FB,$1B,$DC
	.dc.b	$7E,$D0,$F1,$BF,$B4,$FE,$4A,$FD,$B5,$F9,$AF,$0B,$DD,$6A,$5F,$9B
	.dc.b	$E7,$0E,$5F,$2E,$C5,$CB,$FE,$49,$CF,$C1,$5A,$97,$DD,$FF,$93,$7F
	.dc.b	$15,$FF,$7F,$7E,$11,$FE,$F6,$F1,$BF,$DF,$6F,$F9,$56,$33,$F8,$AF
	.dc.b	$FC,$00,$9F,$F7,$FF,$D1,$E7,$D0,$FF,$63,$4E,$7F,$17,$E9,$18,$FC
	.dc.b	$63,$F1,$DD,$1F,$E5,$BF,$C0,$3F,$DB,$1F,$B7,$3F,$27,$FF,$2A,$7E
	.dc.b	$EA,$FC,$0D,$87,$F9,$96,$F7,$5A,$97,$EF,$39,$F9,$E6,$3C,$FE,$A9
	.dc.b	$97,$BB,$BB,$53,$98,$65,$FE,$6D,$F8,$67,$E2,$EF,$CE,$FF,$79,$F8
	.dc.b	$D3,$F5,$07,$E3,$6F,$86,$FF,$3C,$61,$C2,$F7,$5A,$FB,$BC,$E3,$CF
	.dc.b	$E0,$D9,$73,$CB,$C1,$39,$8E,$DC,$FF,$E7,$FF,$DC,$3F,$C1,$5F,$98
	.dc.b	$3F,$54,$7E,$5C,$FC,$D9,$FA,$AF,$E1,$FD,$2F,$F3,$F6,$1C,$2F,$75
	.dc.b	$AF,$C7,$FF,$D1,$D3,$9C,$39,$7E,$51,$DF,$D6,$67,$DE,$7B,$73,$7F
	.dc.b	$A4,$7C,$B6,$2F,$C7,$DF,$B4,$BB,$9F,$99,$BF,$23,$7C,$DF,$FA,$77
	.dc.b	$1E,$38,$70,$BD,$D3,$E9,$B5,$97,$5C,$BA,$65,$FC,$5F,$C1,$CE,$3C
	.dc.b	$61,$F9,$17,$FA,$97,$F5,$FF,$E4,$1F,$1C,$FD,$BB,$FD,$87,$FB,$0B
	.dc.b	$F2,$F7,$E5,$7F,$EA,$FB,$DD,$6A,$5F,$96,$67,$E4,$FF,$93,$4F,$2F
	.dc.b	$C8,$FB,$46,$3F,$17,$64,$EB,$FF,$5B,$7E,$4A,$FC,$93,$F5,$BF,$A1
	.dc.b	$3F,$1E,$7E,$EC,$F8,$B7,$E5,$5F,$EC,$1C,$38,$7C,$03,$0E,$17,$BF
	.dc.b	$CA,$B5,$97,$4C,$B9,$EF,$E0,$8C,$57,$7E,$9F,$F6,$67,$E1,$0F,$C1
	.dc.b	$BF,$A2,$7E,$43,$F0,$BF,$28,$FF,$75,$7F,$00,$EA,$73,$D4,$9F,$F7
	.dc.b	$77,$E1,$B9,$E7,$61,$C7,$2F,$F6,$A6,$5D,$67,$38,$71,$FC,$13,$FD
	.dc.b	$B1,$F9,$27,$F4,$3F,$C9,$7F,$CC,$5F,$92,$BF,$29,$7E,$58,$FC,$AB
	.dc.b	$FD,$C5,$7B,$FC,$15,$7B,$AF,$F2,$DF,$AD,$FE,$DB,$AD,$4F,$3F,$63
	.dc.b	$BF,$D2,$B3,$EF,$9F,$EE,$38,$3F,$CB,$FF,$A3,$7F,$9F,$FF,$09,$FC
	.dc.b	$B7,$EE,$AF,$CB,$1F,$B1,$FF,$59,$FC,$3F,$FC,$F5,$BB,$0E,$1E,$A3
	.dc.b	$0E,$17,$BF,$CA,$3A,$E6,$19,$F6,$CF,$AE,$DF,$D1,$A7,$67,$FB,$F3
	.dc.b	$F4,$4F,$F5,$5F,$C4,$FF,$BB,$BF,$32,$7E,$15,$AF,$CA,$1F,$95,$AF
	.dc.b	$7A,$2F,$FC,$F7,$C3,$E7,$CB,$C5,$F8,$5F,$F9,$1B,$E0,$DF,$B5,$BE
	.dc.b	$E5,$FB,$F7,$F9,$2B,$F5,$BF,$1D,$EE,$B5,$3F,$E4,$EF,$C4,$BC,$FF
	.dc.b	$4D,$FE,$ED,$FC,$F7,$FD,$D9,$F9,$57,$FB,$C7,$5F,$06,$BF,$EC,$2F
	.dc.b	$CA,$9F,$96,$AF,$7F,$96,$72,$9C,$C7,$FB,$D7,$F4,$AF,$91,$87,$2F
	.dc.b	$DF,$DF,$A2,$FF,$66,$F4,$3F,$DC,$FF,$BB,$3F,$4F,$FE,$23,$7B,$AD
	.dc.b	$7D,$52,$B5,$E6,$FC,$EF,$CD,$5F,$AB,$BF,$1B,$7E,$48,$DB,$F6,$7F
	.dc.b	$2E,$FC,$03,$85,$EE,$B5,$F0,$BF,$CC,$BF,$78,$FB,$3F,$93,$FF,$BE
	.dc.b	$FF,$16,$FF,$60,$7E,$31,$DB,$F3,$2C,$8F,$CC,$7F,$CB,$9E,$2F,$F9
	.dc.b	$9B,$3F,$F5,$07,$A7,$FE,$D1,$C7,$94,$E7,$F7,$9E,$37,$C3,$1F,$4F
	.dc.b	$E7,$D8,$CF,$BE,$5F,$C2,$BA,$3E,$8E,$5C,$AF,$DF,$F5,$7B,$CF,$C1
	.dc.b	$FC,$01,$67,$E1,$F9,$9F,$92,$FC,$D5,$FA,$53,$B1,$FA,$B7,$F9,$0F
	.dc.b	$E7,$BF,$8F,$7E,$73,$BA,$D4,$E7,$E5,$BF,$C3,$7B,$FE,$0B,$1E,$78
	.dc.b	$F2,$FC,$5F,$FA,$45,$76,$AD,$61,$CA,$73,$F2,$DD,$6F,$86,$7F,$CB
	.dc.b	$7F,$67,$C6,$7D,$FF,$65,$B8,$7F,$53,$FD,$D3,$F8,$06,$7D,$31,$E3
	.dc.b	$0F,$C8,$BF,$5F,$F9,$C3,$F3,$37,$E7,$7F,$E5,$FF,$31,$F7,$AF,$C5
	.dc.b	$FF,$31,$9F,$E7,$30,$E1,$7B,$AD,$61,$FB,$AF,$CE,$FF,$A4,$FC,$B9
	.dc.b	$FF,$5E,$FE,$E3,$FC,$F9,$D7,$D3,$7F,$2D,$73,$E4,$3F,$55,$FE,$07
	.dc.b	$F9,$DF,$F6,$BF,$CE,$FF,$D8,$BF,$87,$36,$7B,$DF,$E8,$9E,$0F,$D7
	.dc.b	$DE,$FF,$31,$FF,$20,$56,$B9,$BF,$86,$FE,$8E,$FD,$07,$FC,$39,$F8
	.dc.b	$0F,$F9,$3F,$F0,$97,$D3,$BF,$4A,$DD,$6B,$F1,$7F,$F1,$94,$FF,$36
	.dc.b	$D7,$F7,$BF,$8D,$FF,$A9,$7E,$6B,$E5,$7E,$9A,$FC,$F7,$F5,$7F,$B7
	.dc.b	$3F,$7A,$FE,$57,$F1,$7F,$C8,$FF,$0E,$FE,$1F,$FB,$B3,$E4,$DF,$8B
	.dc.b	$3F,$49,$7E,$38,$FD,$77,$97,$DB,$FF,$2B,$C9,$F5,$5F,$E3,$B7,$07
	.dc.b	$F2,$FF,$B6,$7E,$39,$BE,$BF,$F1,$BF,$B0,$FD,$25,$F2,$BF,$ED,$6F
	.dc.b	$94,$FE,$ED,$87,$F8,$03,$1F,$E7,$19,$47,$A9,$FD,$E7,$F3,$EF,$14
	.dc.b	$78,$89,$3E,$53,$F8,$64,$FF,$41,$FC,$A7,$8F,$A9,$FE,$BB,$97,$EE
	.dc.b	$9C,$AF,$91,$EA,$87,$EF,$BD,$B0,$F0,$E1,$ED,$F5,$7B,$47,$90,$F8
	.dc.b	$1E,$97,$C5,$F9,$1C,$47,$C3,$FA,$4F,$3F,$C8,$7D,$28,$F8,$FE,$FF
	.dc.b	$CA,$7D,$FA,$78,$8F,$9B,$F0,$7F,$E3,$DF,$CD,$7E,$87,$A5,$FE,$BA
	.dc.b	$F8,$1F,$F1,$7F,$EB,$0F,$D6,$FF,$A5,$F6,$7F,$90,$F1,$FA,$DF,$49
	.dc.b	$FE,$DD,$CB,$FA,$07,$43,$F4,$67,$C4,$BE,$57,$86,$FF,$D9,$FF,$23
	.dc.b	$FF,$36,$75,$3F,$06,$FE,$C3,$F4,$DF,$EA,$3F,$D8,$B7,$6F,$93,$FF
	.dc.b	$42,$F8,$FF,$ED,$BF,$C7,$DF,$2F,$FC,$E7,$F2,$CF,$55,$FF,$73,$FC
	.dc.b	$2F,$FD,$CA,$3F,$41,$7E,$CD,$E3,$F9,$5F,$F1,$87,$77,$C5,$B7,$ED
	.dc.b	$EF,$98,$FF,$7D,$7D,$CB,$83,$E1,$7B,$EF,$D6,$FC,$BF,$B6,$08,$43
	.dc.b	$FB,$B6,$4F,$13,$F3,$9F,$97,$BF,$7D,$FF,$3D,$FE,$80,$FC,$33,$FC
	.dc.b	$21,$FB,$AF,$E9,$F0,$E1,$C1,$FE,$0D,$F1,$7F,$DA,$DF,$84,$BC,$8F
	.dc.b	$F1,$07,$9D,$FE,$21,$FC,$37,$FC,$ED,$F8,$17,$E3,$FC,$AF,$BD,$F9
	.dc.b	$86,$32,$FF,$0B,$7C,$BF,$FA,$8A,$9F,$B7,$FD,$DF,$55,$FA,$97,$5D
	.dc.b	$FC,$38,$3E,$27,$F5,$FF,$2D,$FC,$57,$F1,$EF,$E5,$7F,$22,$FC,$FB
	.dc.b	$B5,$FC,$1E,$BF,$85,$F5,$7F,$52,$3F,$21,$C3,$FA,$BF,$33,$15,$B3
	.dc.b	$E5,$7C,$B7,$FA,$3F,$F1,$4F,$E0,$51,$93,$FD,$69,$B3,$F4,$7F,$E0
	.dc.b	$DA,$FC,$9F,$CB,$DF,$8B,$BF,$1D,$FB,$5F,$A7,$BF,$08,$FF,$4D,$7C
	.dc.b	$6F,$EC,$FE,$1F,$F5,$9F,$B8,$FE,$07,$B5,$FB,$FF,$FB,$BB,$F4,$E0
	.dc.b	$FC,$37,$FC,$25,$E2,$AF,$FB,$4F,$3F,$F0,$AE,$EF,$D9,$F1,$7F,$16
	.dc.b	$E8,$F7,$8F,$D7,$BD,$15,$FC,$77,$C6,$CF,$E1,$D8,$8F,$DD,$9F,$E2
	.dc.b	$4B,$BB,$FF,$20,$7F,$90,$FE,$E9,$FC,$4F,$C4,$FF,$40,$57,$EA,$7E
	.dc.b	$6F,$F8,$07,$DB,$7F,$AD,$BD,$47,$EC,$AF,$CE,$9F,$BC,$FF,$41,$BF
	.dc.b	$E5,$F8,$7F,$22,$D3,$F8,$59,$FC,$95,$FF,$3B,$7A,$CF,$D2,$5F,$8A
	.dc.b	$7F,$A6,$BD,$87,$EB,$5F,$E9,$EC,$38,$6F,$FC,$AF,$1F,$E3,$DF,$B1
	.dc.b	$FE,$4F,$9B,$F9,$DF,$E4,$BF,$CA,$7F,$09,$FE,$7C,$FF,$28,$FB,$0F
	.dc.b	$E7,$9F,$A1,$6E,$BE,$26,$5F,$C9,$B1,$FF,$37,$65,$FC,$8F,$AF,$FA
	.dc.b	$17,$8F,$FE,$7D,$9F,$4F,$CC,$7F,$38,$FD,$E3,$F9,$14,$7F,$4B,$F2
	.dc.b	$FF,$7D,$97,$D4,$FD,$87,$F2,$FE,$0B,$F5,$F7,$77,$FB,$6F,$D9,$7C
	.dc.b	$AF,$FA,$8B,$E1,$7F,$DC,$5F,$33,$FD,$57,$F3,$CF,$F6,$17,$E4,$CF
	.dc.b	$D1,$7F,$09,$FB,$97,$F5,$0E,$0F,$D1,$FE,$9D,$FD,$97,$E3,$9F,$A3
	.dc.b	$3F,$51,$FE,$C7,$8B,$F1,$6F,$E3,$FE,$97,$F0,$5F,$C3,$3F,$E0,$07
	.dc.b	$C0,$3F,$45,$FF,$2C,$F9,$5F,$AB,$FD,$01,$F8,$07,$DC,$F8,$EF,$F7
	.dc.b	$B7,$E9,$4F,$D3,$3F,$8D,$77,$FF,$43,$FD,$67,$D7,$F3,$2F,$D6,$7F
	.dc.b	$C4,$3F,$A5,$FF,$0B,$FD,$4F,$F8,$DF,$C4,$7E,$CF,$F0,$8F,$DC,$7D
	.dc.b	$6B,$ED,$9F,$B2,$FE,$87,$4F,$F3,$47,$DE,$3F,$26,$FE,$03,$F8,$93
	.dc.b	$F7,$4F,$F6,$67,$D4,$FF,$A6,$7E,$53,$FD,$7B,$FB,$5F,$F0,$6F,$43
	.dc.b	$F4,$BF,$90,$33,$7E,$2F,$FE,$01,$EC,$FE,$4D,$F9,$E3,$ED,$BF,$D4
	.dc.b	$5E,$97,$F8,$BF,$D8,$FE,$1F,$EB,$3E,$B3,$F2,$7F,$93,$38,$BF,$9C
	.dc.b	$7E,$2B,$FE,$85,$D7,$FB,$27,$57,$F0,$BF,$8A,$FE,$29,$C3,$FB,$6F
	.dc.b	$CC,$3F,$89,$CF,$F1,$3F,$B7,$FE,$25,$E9,$7F,$AF,$63,$F8,$97,$EE
	.dc.b	$F8,$8F,$E2,$90,$F9,$AF,$57,$E2,$DF,$E1,$9E,$73,$FE,$BA,$FD,$29
	.dc.b	$FA,$9B,$F3,$47,$EE,$AF,$DD,$FF,$8A,$E7,$FD,$DB,$E5,$FF,$C0,$1C
	.dc.b	$4F,$E0,$0F,$D2,$7E,$BF,$FE,$13,$F8,$6F,$EB,$1F,$AF,$3F,$3C,$FE
	.dc.b	$89,$FB,$73,$F5,$97,$EE,$EF,$D9,$7F,$D4,$38,$7F,$18,$D7,$F6,$67
	.dc.b	$F2,$2E,$1F,$F0,$FF,$8D,$FC,$63,$7F,$E3,$5D,$7F,$CE,$3F,$40,$FE
	.dc.b	$2D,$6F,$A2,$FE,$3B,$F8,$26,$3E,$3F,$F8,$A6,$1F,$E3,$98,$FF,$22
	.dc.b	$7C,$9F,$F8,$EF,$F0,$1F,$E2,$5F,$98,$FF,$AF,$BB,$5F,$71,$FE,$F3
	.dc.b	$FD,$5B,$F8,$9E,$9D,$8F,$F1,$37,$D6,$FE,$BF,$F5,$BF,$F6,$CF,$D0
	.dc.b	$BE,$31,$FA,$9F,$FB,$17,$37,$CE,$3F,$7C,$FF,$25,$70,$FF,$2A,$FA
	.dc.b	$1F,$E1,$9C,$3C,$75,$DF,$E2,$7F,$C3,$F9,$FF,$C1,$D1,$FB,$E3,$FD
	.dc.b	$F9,$F0,$FF,$E2,$1F,$82,$7E,$3B,$FC,$AB,$F0,$5F,$F0,$E6,$BF,$E1
	.dc.b	$6F,$C4,$FF,$84,$F9,$FF,$D9,$FE,$D1,$FB,$E7,$F2,$8F,$D8,$FF,$11
	.dc.b	$EF,$7F,$4B,$FD,$D3,$FC,$C9,$F0,$EF,$4F,$FA,$67,$EE,$7F,$E2,$0F
	.dc.b	$DE,$DF,$9D,$7F,$4F,$FC,$F5,$F5,$8F,$B7,$FC,$26,$DF,$DC,$21,$3F
	.dc.b	$E8,$8F,$4D,$F9,$D7,$FA,$57,$2F,$F3,$C7,$AF,$FE,$45,$FD,$21,$FB
	.dc.b	$47,$F8,$93,$F2,$9E,$BF,$6F,$FE,$C2,$FE,$00,$FD,$DF,$2B,$E1,$1F
	.dc.b	$B4,$7F,$B2,$3F,$43,$FD,$BF,$E2,$9F,$B9,$F5,$FF,$CF,$1C,$2F,$D5
	.dc.b	$5F,$7E,$FC,$4D,$F9,$17,$F9,$FF,$EE,$1C,$9F,$C0,$FC,$9F,$F0,$2F
	.dc.b	$E0,$5F,$E0,$7F,$C8,$5F,$AB,$BF,$5B,$FF,$52,$7E,$C8,$FD,$09,$F8
	.dc.b	$83,$F2,$CF,$F0,$8F,$95,$FE,$07,$FC,$B7,$F7,$BA,$7F,$84,$3F,$01
	.dc.b	$FD,$EC,$FF,$85,$7E,$2B,$F7,$4F,$F8,$37,$EC,$3F,$73,$EA,$8F,$C9
	.dc.b	$1F,$D2,$7F,$57,$FE,$81,$F2,$7F,$E4,$5E,$D7,$83,$FE,$B7,$FA,$FF
	.dc.b	$C8,$F8,$DF,$E9,$4C,$39,$7B,$9F,$DF,$23,$FC,$95,$18,$3F,$CF,$3F
	.dc.b	$A7,$7F,$A7,$3E,$CD,$FA,$9F,$FA,$93,$EA,$5D,$5F,$D5,$7F,$6A,$7E
	.dc.b	$99,$FE,$11,$ED,$FC,$EE,$5F,$E8,$2F,$93,$FF,$56,$3D,$4F,$F4,$46
	.dc.b	$BF,$BA,$E7,$FF,$40,$7C,$6F,$EE,$BE,$47,$FC,$2B,$E1,$FE,$E6,$7F
	.dc.b	$D3,$D5,$FB,$9F,$A9,$FF,$45,$75,$BF,$00,$FE,$AD,$CF,$FE,$70,$FC
	.dc.b	$69,$FA,$DB,$F4,$0F,$F6,$17,$E9,$1F,$EA,$8F,$C3,$7F,$C1,$7F,$47
	.dc.b	$FE,$BB,$FB,$E7,$CC,$7F,$13,$FA,$AF,$EC,$1F,$99,$FF,$15,$F8,$AF
	.dc.b	$DC,$DF,$FA,$C3,$87,$F8,$77,$37,$FA,$D2,$5F,$B0,$7F,$81,$5F,$F0
	.dc.b	$54,$7D,$7F,$81,$FF,$5C,$FC,$E7,$FA,$C7,$ED,$7F,$C1,$53,$FE,$B7
	.dc.b	$F9,$2E,$8F,$FB,$2B,$EE,$BF,$C1,$2F,$FB,$2C,$FF,$10,$FE,$C9,$FC
	.dc.b	$23,$3B,$1F,$6B,$B3,$FD,$0C,$FF,$B0,$7E,$B3,$FD,$64,$FF,$84,$0F
	.dc.b	$F6,$5F,$1F,$F8,$41,$FC,$57,$91,$FE,$10,$ED,$FF,$30,$FD,$FF,$FA
	.dc.b	$73,$F2,$E7,$E8,$AF,$D6,$FF,$35,$EB,$BF,$79,$FC,$97,$81,$FE,$6C
	.dc.b	$F9,$4F,$F5,$77,$93,$FE,$14,$FC,$D7,$FD,$5B,$F9,$AF,$FA,$9B,$E9
	.dc.b	$9E,$EF,$F8,$6B,$F0,$97,$E9,$BF,$EE,$6F,$BF,$7E,$D1,$FE,$AC,$FC
	.dc.b	$99,$F2,$AF,$D6,$3F,$A9,$FB,$9F,$B0,$FC,$F5,$F8,$1F,$F9,$F3,$D7
	.dc.b	$FE,$EB,$E1,$BE,$B7,$FC,$87,$E6,$BF,$72,$FF,$53,$77,$FF,$AA,$FE
	.dc.b	$17,$EB,$FE,$1D,$FD,$77,$F5,$97,$91,$FE,$11,$FD,$57,$FC,$1F,$F1
	.dc.b	$1F,$E8,$7F,$A1,$FF,$44,$FE,$30,$FD,$77,$FC,$41,$DB,$FE,$1E,$FD
	.dc.b	$BB,$FC,$EB,$FA,$47,$F7,$1F,$CA,$7F,$D4,$7F,$9B,$7F,$A8,$BE,$EF
	.dc.b	$FA,$5F,$0F,$FB,$87,$CD,$FF,$38,$FC,$47,$FA,$53,$F3,$FF,$F6,$7F
	.dc.b	$AF,$FE,$CE,$F9,$EF,$3F,$D9,$FE,$FF,$D3,$FF,$67,$7E,$28,$F8,$DF
	.dc.b	$F6,$77,$E7,$BF,$EC,$DD,$3F,$C1,$3C,$DF,$E9,$7E,$9F,$E1,$BB,$7F
	.dc.b	$AB,$A7,$FD,$4F,$3F,$E9,$7F,$83,$FE,$1A,$FF,$A7,$7F,$32,$FF,$0F
	.dc.b	$7B,$9D,$FF,$54,$7E,$4D,$FE,$51,$FC,$2D,$F5,$1F,$9E,$FC,$83,$F9
	.dc.b	$06,$6C,$7F,$D3,$1C,$3E,$73,$F8,$B7,$C9,$FF,$BF,$FE,$BF,$FD,$33
	.dc.b	$F9,$17,$FA,$63,$D5,$FF,$4C,$7E,$22,$EE,$FF,$4C,$7E,$31,$EF,$E4
	.dc.b	$7E,$2B,$FE,$98,$CF,$FB,$14,$3F,$17,$FE,$0C,$FF,$AE,$3F,$44,$FE
	.dc.b	$47,$3E,$11,$FA,$BE,$1F,$F0,$91,$FC,$50,$FE,$55,$F8,$52,$A1,$A8
	.dc.b	$FF,$4C,$77,$7D,$1F,$D7,$DF,$B8,$3C,$1F,$E8,$FF,$0F,$F9,$CF,$CA
	.dc.b	$DF,$39,$FC,$3B,$1F,$F4,$C1,$E3,$3F,$84,$9F,$CE,$7E,$63,$FC,$BD
	.dc.b	$E1,$BF,$80,$3F,$DD,$1E,$93,$D1,$FF,$3F,$FC,$AF,$FA,$7F,$83,$FA
	.dc.b	$07,$C9,$73,$87,$CC,$7F,$A4,$FE,$4B,$F8,$7E,$9F,$20,$FA,$5E,$6F
	.dc.b	$F7,$47,$FC,$23,$FA,$2F,$FA,$67,$C9,$FE,$21,$1F,$EA,$33,$F1,$DF
	.dc.b	$2B,$FD,$37,$17,$E7,$FF,$E9,$99,$F0,$7D,$8F,$D5,$9F,$E1,$51,$FE
	.dc.b	$65,$D7,$EA,$FC,$17,$FB,$AF,$C6,$FF,$4C,$76,$7F,$A5,$FF,$2D,$FF
	.dc.b	$4B,$79,$2E,$DF,$F4,$B7,$E8,$9F,$E2,$7F,$23,$FD,$2B,$E1,$7F,$14
	.dc.b	$CF,$F4,$1E,$6F,$F0,$FE,$77,$E2,$3F,$E1,$0F,$C4,$9E,$EB,$F4,$FF
	.dc.b	$17,$17,$A7,$E8,$DF,$BA,$B9,$BF,$D3,$5F,$8E,$FF,$A8,$FA,$3F,$D4
	.dc.b	$3F,$7D,$FE,$8F,$FD,$6F,$FC,$A9,$F8,$6B,$F2,$17,$EA,$9F,$E8,$18
	.dc.b	$FD,$97,$E2,$5F,$E9,$6F,$CA,$7F,$D2,$BF,$0C,$F0,$7F,$83,$FB,$E2
	.dc.b	$0E,$9F,$F4,$57,$A5,$FE,$4F,$F4,$BF,$CC,$1F,$89,$3F,$50,$9F,$90
	.dc.b	$7F,$A9,$35,$FE,$01,$F5,$DF,$E9,$7E,$3F,$E0,$1F,$33,$FA,$1F,$D8
	.dc.b	$5F,$3D,$FE,$15,$F0,$BF,$32,$7E,$6F,$D6,$F4,$76,$FF,$81,$FF,$32
	.dc.b	$FF,$5F,$7E,$65,$FE,$BD,$FD,$43,$FC,$AD,$F9,$97,$92,$BF,$C1,$9F
	.dc.b	$09,$FE,$BD,$FC,$55,$F1,$9F,$EB,$DE,$F7,$DC,$7F,$86,$BA,$BF,$D6
	.dc.b	$5E,$7B,$97,$F6,$BF,$75,$FE,$1A,$9F,$82,$7F,$BF,$A7,$FC,$47,$9F
	.dc.b	$FA,$5E,$3F,$D3,$50,$F5,$BF,$8F,$74,$7A,$B8,$FD,$73,$9B,$FF,$10
	.dc.b	$7E,$6D,$FE,$95,$FB,$B7,$EE,$FF,$E1,$0F,$CB,$3F,$C2,$7F,$A1,$FF
	.dc.b	$AD,$FE,$3F,$F8,$56,$5F,$E2,$6C,$BF,$D2,$D9,$7F,$8A,$63,$B7,$D2
	.dc.b	$FF,$00,$FC,$0B,$F8,$C6,$BE,$19,$C5,$7F,$31,$FD,$75,$FB,$4B,$F5
	.dc.b	$7F,$F6,$07,$D4,$3F,$69,$64,$FD,$0F,$F8,$DE,$3F,$DC,$0F,$F4,$D1
	.dc.b	$E1,$E7,$FF,$4D,$3F,$C7,$3F,$17,$FF,$4C,$EB,$E1,$D7,$F8,$A7,$37
	.dc.b	$F2,$BF,$73,$1F,$99,$FF,$88,$E7,$F8,$C7,$EB,$1F,$E9,$BF,$87,$FF
	.dc.b	$5D,$E5,$FE,$45,$FB,$1F,$EA,$DF,$96,$FF,$AE,$7F,$80,$5E,$BF,$E4
	.dc.b	$FF,$65,$FF,$49,$FE,$89,$FE,$92,$F8,$4F,$E2,$D8,$FF,$A5,$BF,$52
	.dc.b	$FF,$48,$7E,$0E,$F9,$0F,$F4,$CF,$EC,$1F,$E9,$8F,$BE,$DE,$EF,$53
	.dc.b	$E7,$F1,$7F,$2B,$C1,$F5,$BF,$CB,$7D,$CF,$F5,$0F,$E5,$9A,$A7,$CA
	.dc.b	$EB,$F3,$0F,$E2,$1F,$73,$FD,$06,$FF,$09,$A7,$F6,$13,$F8,$A7,$EC
	.dc.b	$7E,$1E,$1F,$E4,$5F,$BB,$FD,$ED,$5F,$AD,$7F,$8C,$E3,$1C,$5F,$E3
	.dc.b	$5F,$DF,$BF,$CA,$7F,$B1,$7E,$4B,$0F,$ED,$9C,$7F,$AA,$7F,$09,$C1
	.dc.b	$FD,$43,$7F,$F1,$E7,$EB,$9F,$DC,$74,$F8,$55,$FC,$88,$FF,$39,$7E
	.dc.b	$3E,$FC,$E9,$EB,$FF,$35,$E1,$FE,$F1,$F9,$FF,$F3,$1B,$FE,$A9,$E8
	.dc.b	$FF,$8B,$F4,$FE,$A1,$F9,$4F,$DE,$E9,$FE,$33,$5F,$EA,$6A,$7E,$15
	.dc.b	$FE,$A6,$FD,$01,$FB,$7F,$EC,$BF,$13,$DF,$C6,$FF,$A9,$BE,$4D,$F8
	.dc.b	$E3,$9D,$FB,$57,$FB,$EB,$E8,$3A,$FC,$77,$FB,$57,$F4,$F7,$CC,$BF
	.dc.b	$69,$FF,$14,$7D,$77,$FA,$CB,$97,$F4,$1F,$B1,$FF,$8F,$FF,$00,$79
	.dc.b	$DF,$D8,$B9,$5F,$79,$FC,$3B,$9F,$B9,$E0,$FD,$DB,$FB,$E7,$E9,$8F
	.dc.b	$CE,$BF,$DA,$B6,$F2,$FE,$67,$EE,$EF,$2B,$FC,$1F,$F6,$0F,$D9,$97
	.dc.b	$FE,$AA,$1F,$A8,$7F,$81,$7E,$FB,$F2,$F1,$FE,$20,$F9,$2F,$E1,$FF
	.dc.b	$7C,$FD,$3B,$E5,$FE,$B6,$FC,$45,$FA,$D7,$FA,$FB,$EA,$9F,$AE,$FF
	.dc.b	$B5,$3F,$5D,$F5,$7C,$67,$FB,$43,$D8,$7E,$EC,$FA,$0F,$F6,$9F,$E3
	.dc.b	$4F,$D3,$FF,$98,$78,$3A,$3F,$D6,$9E,$8F,$C3,$FD,$03,$E3,$FD,$0B
	.dc.b	$F6,$E7,$67,$E9,$3F,$43,$FF,$05,$74,$7C,$3F,$0B,$FD,$79,$FA,$9F
	.dc.b	$FA,$63,$EA,$FF,$A2,$65,$FE,$1A,$FC,$B3,$FD,$3D,$7F,$CB,$EF,$F2
	.dc.b	$FA,$FF,$86,$7E,$7F,$96,$DF,$E4,$08,$FF,$0D,$F7,$7E,$92,$FF,$C8
	.dc.b	$3D,$DF,$CD,$07,$C7,$3D,$1F,$F2,$1C,$7F,$38,$EE,$7E,$43,$F4,$BE
	.dc.b	$5B,$F7,$19,$7F,$89,$FE,$0B,$FD,$13,$1F,$E5,$5F,$17,$F8,$17,$EB
	.dc.b	$0F,$C6,$FF,$B5,$7D,$6B,$9F,$9E,$FF,$CE,$F8,$FE,$07,$CC,$7F,$89
	.dc.b	$63,$F9,$AC,$7F,$41,$1F,$82,$7F,$8D,$3F,$05,$FD,$46,$DF,$E3,$6D
	.dc.b	$BF,$C9,$BE,$1F,$F8,$1B,$F4,$47,$EE,$FF,$E7,$5F,$D4,$5F,$B0,$7F
	.dc.b	$BF,$7E,$F1,$97,$E9,$B4,$FF,$70,$FC,$5F,$FB,$57,$F5,$2F,$F3,$77
	.dc.b	$CB,$FE,$A7,$6F,$E8,$5F,$45,$FE,$D5,$5F,$47,$FD,$6B,$C0,$FF,$28
	.dc.b	$C7,$F7,$0F,$63,$FD,$B9,$93,$AF,$FB,$9D,$FF,$60,$F3,$B7,$7F,$C8
	.dc.b	$3F,$32,$FF,$6D,$DF,$F9,$F7,$F3,$4F,$E5,$DC,$FF,$D8,$72,$F8,$D3
	.dc.b	$FC,$B6,$DD,$3F,$EC,$AD,$7F,$CD,$DF,$3F,$C2,$7F,$B4,$7C,$C7,$F4
	.dc.b	$2B,$F7,$FE,$B6,$F5,$3E,$97,$ED,$9F,$93,$FC,$39,$F7,$2C,$1F,$A8
	.dc.b	$FF,$1F,$C7,$F9,$EF,$D2,$7F,$A4,$75,$FF,$35,$C7,$F9,$47,$E9,$3F
	.dc.b	$9D,$7C,$EF,$F4,$5D,$BF,$79,$8B,$E9,$3F,$9F,$75,$30,$7D,$07,$F3
	.dc.b	$E7,$FB,$97,$27,$BB,$B7,$FC,$A1,$F0,$DF,$CE,$FB,$7F,$9D,$57,$F3
	.dc.b	$4D,$3F,$CC,$1F,$43,$98,$FF,$47,$68,$FA,$1F,$44,$7F,$99,$7F,$14
	.dc.b	$FF,$54,$67,$FD,$72,$3F,$96,$EF,$FE,$26,$F9,$FF,$F4,$04,$7F,$56
	.dc.b	$F2,$EA,$F4,$9D,$8F,$CA,$5F,$A2,$FA,$BE,$2D,$F9,$3F,$E3,$31,$E3
	.dc.b	$FA,$83,$F2,$5F,$F6,$DF,$E2,$EF,$CE,$5F,$A2,$FF,$A3,$FE,$FF,$FC
	.dc.b	$61,$BB,$E9,$1F,$AF,$FF,$B3,$7F,$2E,$7E,$0F,$F9,$8F,$D7,$3F,$B3
	.dc.b	$ED,$E6,$FD,$ED,$F9,$97,$E3,$FF,$77,$7E,$96,$FB,$97,$95,$E3,$C3
	.dc.b	$F8,$76,$CF,$DE,$9F,$A6,$7E,$07,$F2,$77,$E9,$EF,$D1,$1F,$BA,$7E
	.dc.b	$17,$1E,$3F,$87,$FF,$78,$F6,$7F,$D6,$BF,$84,$3E,$B3,$FA,$7F,$D8
	.dc.b	$7F,$38,$F1,$BF,$11,$E8,$FF,$36,$F2,$7F,$DB,$5F,$BA,$BD,$97,$EB
	.dc.b	$AF,$D4,$DF,$AB,$3E,$F9,$FB,$C7,$87,$1F,$F7,$17,$CD,$3F,$5B,$FC
	.dc.b	$BF,$E8,$0F,$C9,$9F,$3E,$FB,$F7,$D3,$7F,$36,$FA,$AF,$E4,$BD,$3F
	.dc.b	$CA,$7C,$07,$CC,$77,$FD,$33,$FD,$99,$FB,$E7,$F9,$6F,$E5,$1F,$9C
	.dc.b	$FF,$96,$FE,$8B,$FC,$F9,$F4,$9F,$D3,$72,$E5,$EF,$FF,$9C,$3E,$97
	.dc.b	$F8,$24,$62,$3F,$CB,$5C,$9F,$E2,$98,$FF,$08,$5F,$F3,$13,$F1,$DF
	.dc.b	$4B,$FD,$7B,$E5,$3F,$C5,$FB,$FF,$BC,$8F,$D4,$7E,$36,$F9,$D7,$EA
	.dc.b	$2F,$D4,$DD,$3F,$07,$F5,$97,$EE,$FC,$71,$FF,$5A,$9F,$B1,$D7,$F9
	.dc.b	$6E,$FF,$E8,$68,$FF,$0C,$3F,$EE,$1E,$3F,$C2,$BF,$E6,$1F,$8D,$FD
	.dc.b	$C7,$E2,$FF,$EF,$BE,$4E,$FF,$B5,$7F,$A5,$FE,$F1,$FA,$E3,$F6,$A7
	.dc.b	$EE,$3F,$CB,$BF,$0B,$FF,$77,$6B,$FE,$F3,$D5,$F4,$DF,$CB,$BF,$12
	.dc.b	$FF,$30,$CF,$F5,$AD,$FF,$96,$EF,$FD,$AB,$E7,$7F,$84,$4F,$FB,$DA
	.dc.b	$7F,$92,$3F,$CA,$BE,$9B,$FD,$B9,$F4,$DF,$CB,$3E,$77,$F9,$C5,$7F
	.dc.b	$28,$EC,$E8,$FF,$83,$F0,$E7,$87,$F2,$B3,$F9,$51,$DB,$E5,$77,$3F
	.dc.b	$C8,$F8,$7D,$33,$FD,$5F,$E6,$7F,$CB,$CF,$F9,$1B,$C6,$E6,$FF,$BE
	.dc.b	$61,$F8,$A7,$F2,$78,$FE,$41,$BF,$F9,$E9,$FF,$76,$7C,$FF,$F9,$41
	.dc.b	$FF,$03,$9F,$D7,$B8,$3F,$DE,$1F,$19,$F0,$33,$78,$DF,$C5,$2B,$FC
	.dc.b	$AF,$FB,$7F,$F3,$3F,$C3,$BA,$47,$F9,$6B,$E8,$3F,$CE,$3C,$9E,$4D
	.dc.b	$9F,$3B,$F2,$BF,$1A,$FE,$93,$7F,$CA,$8F,$F2,$FF,$96,$FC,$BB,$FC
	.dc.b	$BF,$1F,$CD,$FD,$6F,$E5,$18,$3E,$8F,$E9,$FE,$83,$8F,$C6,$FE,$AF
	.dc.b	$E3,$BF,$CB,$7C,$8F,$E5,$3E,$97,$F1,$AF,$D2,$7F,$B8,$F0,$7E,$FB
	.dc.b	$F6,$A7,$E5,$5F,$C8,$39,$DF,$A5,$FF,$AE,$BF,$66,$FF,$09,$7E,$80
	.dc.b	$FD,$B7,$F0,$FA,$7F,$18,$E6,$7C,$9B,$F7,$7F,$F2,$9F,$E2,$B3,$F1
	.dc.b	$3F,$E5,$7F,$A2,$BF,$66,$79,$6F,$C9,$21,$33,$FE,$7F,$F9,$CB,$E4
	.dc.b	$3F,$8A,$7E,$92,$20,$FD,$F3,$FC,$EB,$F2,$1F,$E0,$EF,$C3,$3F,$C5
	.dc.b	$5F,$8B,$FF,$93,$BE,$1F,$FE,$00,$7D,$AF,$FB,$83,$F1,$3F,$F7,$37
	.dc.b	$D4,$6B,$5F,$F4,$87,$EC,$2F,$B1,$FF,$2C,$FE,$11,$F9,$EF,$AB,$7E
	.dc.b	$68,$FC,$CD,$F8,$B3,$F1,$87,$ED,$F8,$FA,$4F,$95,$E2,$BE,$8F,$FC
	.dc.b	$11,$B7,$EE,$A3,$F1,$1E,$97,$F1,$1F,$BB,$FD,$5F,$E4,$FF,$E9,$6F
	.dc.b	$DF,$FF,$67,$F1,$13,$F4,$6F,$F3,$D7,$E0,$DF,$07,$E3,$FF,$C0,$BD
	.dc.b	$FF,$0B,$BF,$FC,$FD,$A0,$FF,$30,$14,$FF,$3A,$EB,$FC,$87,$D9,$FD
	.dc.b	$2F,$C2,$7F,$43,$DF,$FE,$00,$7E,$9D,$F0,$FA,$7F,$7B,$DE,$11,$FE
	.dc.b	$55,$F9,$BF,$F4,$0F,$C4,$6A,$5F,$A3,$FF,$96,$7F,$18,$FF,$2A,$63
	.dc.b	$FE,$62,$F8,$DF,$F5,$69,$FC,$23,$F5,$EF,$C6,$FC,$0B,$F4,$77,$EE
	.dc.b	$2F,$5B,$FB,$16,$1F,$E8,$EF,$C4,$FF,$36,$FF,$A3,$3E,$B7,$FC,$86
	.dc.b	$FF,$08,$FB,$2F,$F2,$97,$37,$FB,$7B,$4F,$E0,$3F,$39,$FC,$33,$F1
	.dc.b	$E7,$EC,$4F,$03,$F2,$3E,$5F,$E2,$7F,$46,$FE,$BF,$F3,$5F,$E9,$AD
	.dc.b	$3F,$C4,$99,$BE,$37,$F8,$FF,$C3,$7F,$16,$AF,$F5,$0D,$7B,$4F,$E3
	.dc.b	$3A,$FF,$1A,$BF,$F4,$01,$F4,$47,$E4,$FF,$E5,$4E,$5F,$E2,$B3,$FC
	.dc.b	$83,$F1,$7D,$D7,$D7,$FD,$47,$FC,$9B,$1F,$6F,$3F,$E3,$7A,$FF,$94
	.dc.b	$3F,$36,$FE,$3B,$F8,$1F,$FB,$E7,$1F,$E5,$3F,$96,$F7,$C9,$F2,$3F
	.dc.b	$E3,$DE,$0F,$F2,$BF,$DB,$F0,$9F,$C3,$6D,$FD,$1B,$83,$FC,$CB,$5F
	.dc.b	$E3,$8F,$C5,$3F,$8F,$57,$FB,$CE,$7F,$C8,$DF,$65,$FE,$53,$F9,$FF
	.dc.b	$E5,$51,$FD,$FC,$FF,$1F,$FC,$FF,$F2,$97,$FC,$A1,$4F,$9F,$FF,$24
	.dc.b	$C7,$FA,$90,$FE,$45,$F3,$5F,$C8,$5F,$E6,$E7,$F9,$4F,$37,$E2,$7E
	.dc.b	$2E,$4F,$F6,$94,$FF,$93,$3E,$83,$F9,$97,$9D,$FF,$00,$23,$FC,$9F
	.dc.b	$FA,$47,$F3,$5F,$C4,$5F,$8A,$FF,$A6,$75,$F5,$FA,$1F,$B2,$C3,$FA
	.dc.b	$CF,$D5,$BA,$9F,$BD,$3C,$87,$F0,$0B,$F5,$57,$EB,$0F,$D0,$7F,$C8
	.dc.b	$BF,$85,$FF,$BC,$3F,$15,$7C,$9F,$F8,$E7,$E6,$7F,$85,$7E,$3F,$FE
	.dc.b	$F1,$FD,$F1,$EE,$BF,$74,$7E,$A3,$09,$F0,$1F,$E5,$EF,$D8,$7F,$D4
	.dc.b	$BF,$B3,$FF,$9D,$FC,$38,$7F,$BE,$FE,$67,$FD,$57,$F6,$BF,$7F,$E0
	.dc.b	$B1,$F9,$3B,$ED,$BF,$B0,$FE,$11,$FB,$6F,$0F,$E8,$1F,$DB,$B4,$FF
	.dc.b	$17,$7A,$CF,$AF,$7E,$A4,$FD,$A5,$F7,$FF,$CE,$D7,$D3,$F3,$C7,$FC
	.dc.b	$E1,$FA,$4B,$F6,$8D,$FD,$57,$1C,$1F,$A4,$FF,$22,$F9,$9F,$CA,$7C
	.dc.b	$CB,$7F,$F1,$F7,$D0,$FF,$AA,$3F,$08,$FE,$83,$F5,$3F,$D0,$7C,$EF
	.dc.b	$E1,$9F,$2B,$FC,$4F,$7F,$D9,$FA,$1F,$C7,$0F,$E3,$DC,$BF,$C8,$1F
	.dc.b	$F3,$4C,$7E,$7F,$2F,$E6,$0F,$F1,$CF,$94,$FF,$76,$47,$F8,$F9,$FE
	.dc.b	$2E,$7F,$1D,$3F,$8F,$7C,$AF,$F8,$FE,$7E,$39,$FE,$1E,$9F,$F5,$C7
	.dc.b	$C5,$7F,$51,$FC,$9B,$FC,$B9,$F7,$EF,$BB,$7C,$DF,$F9,$62,$FF,$DC
	.dc.b	$FD,$DF,$E1,$2F,$D1,$BF,$DC,$DE,$C7,$EB,$36,$FC,$FF,$C9,$7F,$88
	.dc.b	$3F,$63,$FE,$D3,$F5,$BF,$D9,$3F,$64,$7E,$98,$FD,$E1,$F1,$CC,$3F
	.dc.b	$87,$63,$FE,$C5,$E3,$FF,$0D,$FC,$27,$FA,$EF,$E5,$1F,$BC,$77,$9F
	.dc.b	$E1,$E7,$EF,$3F,$04,$FE,$0D,$AF,$FB,$6B,$E6,$BF,$A3,$7C,$56,$AB
	.dc.b	$FD,$89,$F2,$DF,$C5,$33,$FE,$99,$EA,$7F,$14,$7F,$92,$4F,$F8,$E3
	.dc.b	$93,$F8,$A3,$FE,$3A,$FD,$EF,$F8,$A3,$FE,$46,$E0,$FF,$37,$AF,$92
	.dc.b	$FE,$26,$EE,$FF,$89,$FA,$DF,$15,$FE,$2B,$F8,$CF,$FB,$4E,$7F,$8E
	.dc.b	$47,$FB,$5F,$F0,$3F,$F3,$47,$4F,$F9,$4B,$E4,$7F,$D8,$79,$FF,$A5
	.dc.b	$7A,$FF,$88,$77,$FF,$89,$7B,$7F,$D1,$79,$7F,$B7,$FC,$C7,$E5,$1F
	.dc.b	$EC,$1F,$C3,$9F,$A0,$FE,$62,$1F,$9C,$FE,$7F,$EB,$7F,$B9,$93,$EB
	.dc.b	$7F,$45,$FA,$F3,$F4,$FF,$F5,$2B,$7E,$74,$A7,$FD,$97,$FA,$4F,$EB
	.dc.b	$FD,$E8,$5D,$DD,$BF,$E5,$CE,$0F,$C8,$7D,$07,$F9,$BB,$E5,$3F,$D5
	.dc.b	$1F,$2A,$FD,$4D,$F0,$CF,$6F,$FC,$31,$F7,$1F,$D9,$F1,$FF,$7E,$63
	.dc.b	$FE,$41,$F1,$BA,$47,$E1,$7D,$DF,$F3,$CC,$BE,$6D,$4F,$B3,$FC,$8F
	.dc.b	$C7,$BF,$58,$79,$6F,$3F,$F5,$B9,$B2,$FF,$7D,$47,$F6,$48,$C1,$FC
	.dc.b	$F0,$A7,$FB,$FC,$FF,$2C,$45,$F8,$9F,$F4,$5D,$3C,$9F,$C1,$79,$1F
	.dc.b	$9D,$BF,$0F,$A9,$FE,$71,$F1,$7A,$5F,$F9,$F2,$7F,$D1,$67,$2F,$A0
	.dc.b	$FF,$3B,$7A,$7A,$F2,$FF,$AD,$D7,$F4,$3A,$0F,$D4,$5F,$B6,$BE,$C9
	.dc.b	$F9,$27,$FA,$EF,$F1,$0C,$FC,$67,$F1,$9D,$FF,$D4,$3E,$E7,$F9,$13
	.dc.b	$5F,$E9,$5E,$B3,$F1,$1C,$78,$9F,$EF,$CF,$CC,$04,$21,$FC,$DD,$FF
	.dc.b	$1C,$6F,$E4,$FC,$91,$FA,$C7,$E4,$BF,$1F,$BE,$81,$7C,$2F,$F2,$BF
	.dc.b	$8B,$FF,$59,$FA,$3F,$92,$27,$A3,$FC,$BF,$C8,$FF,$5B,$FC,$CF,$F4
	.dc.b	$F2,$7D,$47,$F3,$3F,$83,$FE,$F1,$B7,$F8,$D8,$F8,$BF,$8E,$FD,$DF
	.dc.b	$D5,$5F,$BA,$FE,$43,$F2,$3F,$F8,$03,$E7,$FF,$B6,$FA,$BF,$C4,$DF
	.dc.b	$86,$FF,$7A,$F2,$BD,$79,$7F,$96,$FD,$37,$E7,$AF,$C8,$3F,$A6,$FE
	.dc.b	$27,$FE,$99,$BF,$F1,$85,$FF,$87,$CF,$F0,$1F,$9D,$F8,$77,$EB,$FE
	.dc.b	$EB,$F8,$A7,$F1,$83,$E1,$F0,$FF,$97,$0F,$D0,$9A,$FD,$53,$F1,$BE
	.dc.b	$87,$E3,$3C,$FF,$F5,$44,$FE,$30,$FD,$3F,$EC,$7F,$8C,$FB,$6F,$C8
	.dc.b	$E4,$8F,$F0,$A7,$A3,$FA,$8F,$47,$F1,$97,$DB,$EA,$7F,$18,$FF,$30
	.dc.b	$F8,$9F,$E2,$5F,$15,$FC,$53,$F3,$BF,$ED,$1E,$9F,$E3,$7D,$2F,$C6
	.dc.b	$3F,$8F,$FB,$3F,$B5,$87,$E5,$3C,$2F,$AC,$7D,$7F,$33,$F1,$5E,$6F
	.dc.b	$E2,$0F,$C5,$FE,$45,$C2,$9F,$BC,$3A,$1F,$9B,$E1,$B1,$9C,$3E,$AF
	.dc.b	$E6,$9F,$86,$F8,$AF,$E3,$5F,$7B,$F8,$7C,$7C,$5A,$78,$87,$E5,$9F
	.dc.b	$E7,$C8,$FD,$A1,$FE,$EE,$C3,$FD,$76,$FE,$B3,$D8,$F0,$FC,$A3,$F2
	.dc.b	$5F,$EA,$DF,$0B,$FC,$1E,$DF,$94,$FF,$9F,$30,$FF,$20,$C7,$F1,$27
	.dc.b	$FC,$84,$7F,$09,$F9,$BF,$DB,$BF,$B8,$FD,$FB,$F9,$CF,$C1,$FF,$B2
	.dc.b	$FF,$04,$7E,$D8,$F5,$7F,$B0,$F9,$3F,$8C,$F1,$3E,$FE,$3C,$8F,$EC
	.dc.b	$D9,$FE,$81,$FF,$3C,$9F,$8C,$DF,$F0,$EF,$F8,$43,$CF,$FF,$55,$78
	.dc.b	$FF,$8F,$8F,$F3,$11,$98,$FF,$58,$5F,$E3,$2D,$EC,$FF,$55,$F9,$97
	.dc.b	$89,$F8,$47,$FA,$A5,$FB,$3D,$FF,$F1,$C7,$8B,$FD,$69,$B8,$3E,$6B
	.dc.b	$E0,$FE,$9A,$21,$0F,$F4,$B4,$7E,$C3,$F3,$D7,$EF,$90,$86,$23,$F9
	.dc.b	$CF,$7F,$F5,$72,$7C,$C7,$FB,$0F,$83,$FA,$FF,$CC,$FF,$5F,$AF,$E7
	.dc.b	$3F,$96,$BF,$63,$78,$6F,$DA,$5F,$49,$F0,$3C,$27,$D1,$FF,$8F,$FF
	.dc.b	$12,$FE,$A3,$A7,$F4,$9F,$85,$FE,$A3,$F3,$1F,$D4,$3E,$63,$FA,$7F
	.dc.b	$CB,$FF,$4E,$F0,$3E,$C5,$FF,$49,$EE,$FE,$91,$F8,$3F,$D4,$8F,$F0
	.dc.b	$1F,$81,$FD,$23,$DB,$FD,$BC,$7F,$A2,$7C,$0F,$F4,$CD,$FE,$D3,$EB
	.dc.b	$FF,$BF,$FE,$02,$FA,$5F,$EF,$FF,$4D,$FE,$02,$F2,$7F,$CA,$9F,$B2
	.dc.b	$FD,$7C,$7E,$B6,$5F,$E5,$7F,$0B,$F9,$4F,$A5,$FA,$33,$E3,$7D,$87
	.dc.b	$F9,$24,$9F,$A2,$6E,$BF,$A6,$FE,$77,$FB,$0F,$62,$7E,$A5,$99,$78
	.dc.b	$FF,$53,$C7,$FA,$DF,$A1,$FF,$99,$79,$3F,$3F,$F9,$4F,$FA,$33,$E5
	.dc.b	$5F,$91,$FC,$EE,$F7,$CE,$FC,$EF,$27,$FB,$6F,$A1,$FE,$D7,$F4,$7F
	.dc.b	$99,$3F,$A8,$F9,$7F,$81,$F9,$97,$F3,$2F,$B8,$FF,$78,$7E,$98,$FB
	.dc.b	$3F,$EF,$9D,$9F,$DB,$79,$3F,$D6,$7E,$DF,$FA,$C1,$65,$FE,$8D,$7F
	.dc.b	$D5,$A7,$EA,$E1,$F9,$97,$F8,$32,$3F,$D8,$86,$7F,$13,$FD,$CD,$7F
	.dc.b	$81,$3E,$AB,$F6,$53,$FE,$0C,$F3,$38,$79,$3F,$C0,$BC,$1D,$E5,$FE
	.dc.b	$09,$E2,$FE,$97,$D1,$FB,$AC,$3F,$D3,$59,$7F,$04,$FC,$89,$F8,$CF
	.dc.b	$EB,$7C,$47,$C7,$FF,$77,$BF,$EF,$F4,$FD,$E3,$FC,$8B,$5F,$CD,$B2
	.dc.b	$FE,$AF,$E2,$BE,$BB,$F9,$77,$C7,$3E,$3B,$96,$4A,$78,$3F,$E3,$AF
	.dc.b	$B4,$FF,$1D,$6F,$FE,$87,$FC,$4B,$F0,$95,$BF,$CA,$BF,$D4,$1F,$8C
	.dc.b	$FF,$1A,$8C,$8F,$33,$FC,$B7,$97,$F8,$53,$0F,$F3,$6B,$FC,$B7,$E1
	.dc.b	$B8,$7C,$97,$F3,$B3,$F9,$E7,$17,$F4,$1C,$3F,$75,$F4,$FF,$E5,$2F
	.dc.b	$84,$FE,$89,$8F,$EE,$BE,$10,$4E,$4F,$E1,$99,$FF,$91,$BE,$AF,$5F
	.dc.b	$AC,$F3,$FC,$B3,$CB,$9F,$5E,$2F,$3F,$E5,$BD,$3E,$EB,$FC,$E7,$E0
	.dc.b	$7F,$B0,$E9,$FA,$2F,$B3,$F2,$EF,$E8,$DE,$0F,$D7,$D7,$FD,$55,$87
	.dc.b	$F9,$43,$C7,$7E,$84,$FA,$C7,$CA,$FF,$28,$F9,$07,$E0,$DF,$EC,$7F
	.dc.b	$AB,$FE,$83,$F9,$2F,$F9,$6F,$3F,$F2,$EF,$D2,$7F,$93,$7E,$27,$F0
	.dc.b	$AD,$EA,$FE,$8F,$F8,$2F,$F9,$6F,$F7,$77,$E8,$5F,$D1,$BE,$A1,$FA
	.dc.b	$B3,$EA,$7F,$CB,$9F,$75,$FD,$0F,$4F,$E9,$FA,$7F,$96,$F9,$BD,$51
	.dc.b	$FE,$0C,$3F,$C9,$3D,$5F,$E9,$4D,$7F,$CE,$53,$FD,$37,$0F,$E8,$D5
	.dc.b	$FE,$AB,$1F,$65,$BF,$67,$FC,$35,$F3,$FF,$E6,$5F,$AF,$FF,$0C,$D7
	.dc.b	$F5,$58,$FF,$3C,$E1,$FD,$5F,$E9,$7F,$D4,$32,$E4,$FF,$39,$FE,$15
	.dc.b	$FB,$7B,$FD,$5F,$B2,$F7,$7F,$5F,$8F,$F9,$7B,$DB,$FE,$ED,$BF,$EF
	.dc.b	$72,$FE,$C5,$1F,$C4,$FC,$37,$E9,$5F,$E5,$9F,$B5,$72,$FF,$97,$38
	.dc.b	$35,$18,$EC,$FE,$23,$F7,$9F,$E5,$EF,$4B,$F3,$9F,$8E,$08,$4F,$C5
	.dc.b	$3F,$2D,$F1,$FF,$C7,$FE,$3F,$FA,$95,$7F,$04,$D7,$F1,$BE,$8F,$E5
	.dc.b	$38,$3F,$D1,$9F,$B4,$7E,$4A,$3F,$6B,$F8,$5F,$EA,$3D,$BF,$CE,$FE
	.dc.b	$66,$FC,$CD,$F9,$9A,$30,$BF,$21,$FE,$61,$D2,$FD,$FF,$FA,$C7,$EE
	.dc.b	$EF,$D6,$69,$FC,$D7,$B3,$FC,$3B,$BF,$C9,$DB,$C9,$C1,$FE,$59,$3F
	.dc.b	$7B,$B7,$EB,$B2,$FE,$37,$4E,$CF,$D5,$56,$AF,$FD,$A3,$3F,$C2,$7D
	.dc.b	$6F,$54,$3F,$3E,$FE,$CA,$BF,$04,$FE,$4F,$FB,$A7,$FA,$D3,$E6,$3F
	.dc.b	$8E,$FB,$2F,$A1,$FF,$59,$DB,$F7,$0F,$ED,$BE,$7B,$E2,$3F,$B7,$FD
	.dc.b	$BF,$F8,$7F,$F1,$5F,$CF,$71,$7F,$AF,$FC,$ED,$77,$FF,$B2,$70,$FE
	.dc.b	$1F,$3F,$EB,$AF,$CF,$7F,$BB,$BF,$E2,$7E,$7F,$F5,$D7,$37,$F9,$BB
	.dc.b	$C8,$FE,$E9,$CD,$FE,$34,$E1,$FF,$5F,$79,$BF,$9C,$8C,$9F,$EB,$FC
	.dc.b	$3F,$2E,$FF,$1D,$FA,$AF,$F5,$CF,$E6,$1F,$D1,$32,$7D,$C7,$F5,$0F
	.dc.b	$3F,$F3,$37,$FD,$4B,$E0,$B5,$F9,$2F,$F4,$58,$FE,$4D,$E9,$7D,$0F
	.dc.b	$9A,$FE,$6B,$A7,$F8,$BB,$2F,$EA,$E4,$3F,$9C,$F4,$FF,$B9,$FE,$AF
	.dc.b	$FC,$05,$E5,$7F,$BF,$FE,$AB,$F9,$EE,$2F,$D5,$7F,$C0,$BD,$4B,$DE
	.dc.b	$1F,$95,$C7,$FD,$91,$FB,$08,$85,$FC,$AB,$C7,$F9,$96,$22,$57,$1F
	.dc.b	$9D,$7E,$77,$09,$FB,$1F,$AB,$E8,$B1,$19,$FC,$87,$F1,$7E,$67,$F8
	.dc.b	$3F,$E4,$7F,$9A,$3F,$D3,$BE,$CF,$F7,$5F,$83,$FF,$A0,$7E,$73,$FA
	.dc.b	$37,$D9,$FE,$F7,$C9,$7E,$01,$FD,$1B,$EC,$3F,$9C,$7B,$BF,$B3,$FD
	.dc.b	$43,$F9,$D7,$D2,$FF,$60,$FC,$FB,$F3,$9E,$0F,$F7,$6E,$7F,$E7,$DE
	.dc.b	$BF,$F5,$8F,$CD,$7F,$C2,$D3,$FE,$EA,$EF,$FF,$70,$76,$7C,$F3,$FC
	.dc.b	$0F,$D5,$FF,$00,$3E,$89,$F3,$AF,$BC,$7E,$62,$FB,$17,$E5,$5F,$EE
	.dc.b	$9F,$D4,$98,$FE,$8F,$4F,$E5,$E7,$F8,$0F,$F3,$BF,$F5,$D6,$1F,$E1
	.dc.b	$4E,$7F,$F0,$97,$C6,$FF,$33,$F6,$3F,$CE,$37,$FE,$FE,$97,$E3,$8F
	.dc.b	$D5,$3F,$C5,$3F,$BF,$3F,$20,$FE,$05,$F8,$03,$1F,$F2,$DE,$3F,$CF
	.dc.b	$CF,$F0,$BB,$FD,$A9,$FE,$ED,$E9,$97,$B5,$F0,$7D,$3F,$D4,$47,$F0
	.dc.b	$4F,$9E,$FF,$6C,$74,$FC,$6E,$9D,$4F,$EA,$4F,$E1,$3F,$91,$FF,$B6
	.dc.b	$A3,$E5,$1F,$A4,$BF,$F4,$C6,$1F,$D0,$BE,$73,$FD,$C5,$F0,$5F,$D1
	.dc.b	$23,$F5,$B8,$7F,$BA,$7E,$03,$FA,$1F,$BB,$FE,$33,$F8,$0F,$D6,$4F
	.dc.b	$EA,$0F,$F1,$A9,$F6,$7F,$19,$FF,$59,$3F,$EB,$BF,$9F,$FE,$8C,$FF
	.dc.b	$8B,$8F,$76,$7F,$D7,$1C,$1F,$E2,$3F,$CB,$9F,$35,$FD,$17,$EF,$DD
	.dc.b	$6F,$D9,$BF,$C5,$3E,$9F,$F8,$7F,$E7,$3F,$A4,$DF,$F4,$BF,$07,$EB
	.dc.b	$E5,$FE,$2D,$FC,$AB,$FC,$59,$F6,$CF,$A9,$FF,$0F,$7E,$E3,$FD,$33
	.dc.b	$E8,$3F,$A6,$61,$C3,$A5,$EF,$FE,$A3,$DD,$FD,$6F,$BB,$FD,$03,$DE
	.dc.b	$C4,$3F,$43,$FE,$7D,$F1,$7F,$E4,$DE,$8F,$F1,$A7,$8B,$FE,$77,$F7
	.dc.b	$7F,$DD,$1A,$7F,$40,$FC,$9B,$F5,$1F,$23,$FE,$2F,$FC,$87,$FA,$A7
	.dc.b	$C9,$7F,$8E,$FF,$72,$FE,$B2,$FF,$97,$AF,$FC,$4A,$7F,$40,$EA,$7D
	.dc.b	$AF,$F8,$96,$3F,$0F,$D3,$F3,$3C,$2F,$F2,$F7,$8F,$FE,$DE,$FA,$6F
	.dc.b	$E7,$9F,$99,$FF,$AB,$3F,$4A,$FD,$3F,$DF,$FF,$89,$31,$FD,$26,$2F
	.dc.b	$D3,$3F,$C4,$8B,$C7,$FA,$CF,$F8,$6B,$B3,$FD,$D3,$18,$8F,$D2,$1F
	.dc.b	$E7,$4F,$CD,$5F,$43,$FA,$1F,$CB,$9E,$97,$F9,$7F,$EE,$3F,$49,$9F
	.dc.b	$2E,$1F,$E7,$B8,$FF,$A1,$3D,$0F,$CB,$FD,$07,$86,$3F,$DE,$FF,$65
	.dc.b	$F8,$E7,$FD,$08,$7E,$9E,$7F,$4F,$B7,$F0,$7F,$B2,$E1,$3A,$3F,$DF
	.dc.b	$F8,$7F,$89,$1F,$F8,$03,$F3,$FF,$AF,$C3,$F4,$37,$FC,$3E,$5F,$AE
	.dc.b	$FE,$8A,$3F,$C3,$30,$AF,$F1,$2F,$CF,$FF,$A3,$23,$FD,$32,$7E,$E4
	.dc.b	$FF,$3E,$CF,$F9,$C1,$C7,$AD,$FE,$6D,$FB,$1F,$F1,$27,$4F,$F8,$8F
	.dc.b	$7F,$CB,$F6,$7E,$82,$1E,$8F,$E7,$1F,$F0,$57,$C9,$3F,$27,$84,$FA
	.dc.b	$4F,$EA,$FF,$0B,$88,$98,$FF,$4D,$F8,$BF,$F1,$1E,$DF,$63,$A3,$FC
	.dc.b	$44,$FF,$A2,$7F,$47,$FF,$04,$F9,$7F,$EA,$1F,$D6,$5F,$80,$FF,$1E
	.dc.b	$FC,$BE,$FC,$4F,$FB,$3B,$3F,$F1,$67,$AA,$FC,$E7,$F9,$7F,$90,$FC
	.dc.b	$EA,$9F,$E2,$1F,$A9,$FF,$0D,$F4,$FF,$9F,$7D,$9F,$D8,$1F,$E3,$DF
	.dc.b	$B8,$FF,$40,$7D,$C7,$F8,$EF,$E7,$3F,$C9,$5F,$53,$FE,$1C,$F4,$BE
	.dc.b	$77,$81,$FE,$4E,$DF,$E3,$FA,$9F,$E2,$0F,$8C,$FF,$51,$7A,$9F,$E2
	.dc.b	$2F,$BB,$FF,$10,$FE,$BB,$FE,$20,$F8,$C7,$9D,$FE,$20,$FA,$FF,$F0
	.dc.b	$FF,$D8,$FF,$89,$BC,$2F,$F0,$FF,$8B,$FE,$20,$F0,$7E,$D7,$D5,$FF
	.dc.b	$87,$BE,$0D,$97,$F8,$7B,$2F,$E7,$9E,$F7,$F8,$73,$8B,$FC,$1A,$7F
	.dc.b	$98,$BD,$8F,$F6,$9F,$D7,$7E,$73,$D5,$FF,$39,$FA,$3F,$94,$AF,$E8
	.dc.b	$47,$E1,$FE,$4B,$FD,$C9,$AF,$E1,$E3,$1F,$6B,$FE,$1B,$F3,$FF,$97
	.dc.b	$C7,$FB,$A2,$BF,$87,$3F,$E2,$13,$BF,$4F,$FB,$73,$E3,$BF,$D1,$4F
	.dc.b	$FB,$BA,$3F,$A6,$7E,$C0,$FD,$4D,$F9,$73,$F3,$3F,$EB,$1F,$97,$BF
	.dc.b	$35,$FF,$30,$E3,$C7,$16,$2F,$D4,$37,$5C,$B0,$FE,$C1,$7B,$F3,$5F
	.dc.b	$72,$FC,$22,$13,$83,$FB,$9F,$E9,$1F,$A7,$FC,$3B,$FD,$99,$C5,$FD
	.dc.b	$87,$1F,$00,$FC,$8F,$F9,$7F,$EC,$9F,$DE,$BC,$0F,$F7,$07,$BF,$FE
	.dc.b	$E4,$FC,$B7,$87,$97,$FE,$E3,$F2,$FF,$94,$F2,$FF,$6B,$A7,$C7,$FF
	.dc.b	$B9,$BE,$DB,$FC,$E9,$FB,$5F,$F9,$F7,$F0,$7F,$F3,$FF,$E0,$FF,$E8
	.dc.b	$1F,$CD,$3F,$C8,$3E,$27,$F9,$EF,$F2,$B7,$E0,$F4,$FF,$05,$FC,$17
	.dc.b	$EF,$FA,$DF,$72,$FB,$B8,$4F,$57,$FD,$D3,$D3,$FE,$7C,$FB,$8F,$F6
	.dc.b	$F4,$7F,$88,$7A,$BF,$8B,$CF,$F8,$6B,$C9,$FE,$EF,$1F,$C6,$BF,$4F
	.dc.b	$71,$3F,$31,$FF,$0F,$FE,$60,$FC,$FF,$FB,$2F,$CB,$7F,$87,$BD,$AF
	.dc.b	$E2,$5E,$E7,$FA,$2B,$B3,$E6,$7D,$AF,$F9,$97,$5F,$F2,$DF,$C2,$7F
	.dc.b	$1B,$5F,$6B,$FC,$66,$9C,$1E,$39,$FF,$19,$4F,$FA,$07,$F4,$07,$EE
	.dc.b	$CF,$DA,$1F,$09,$FE,$34,$F8,$F7,$93,$FE,$21,$F2,$FF,$92,$64,$DF
	.dc.b	$FC,$3D,$EF,$7F,$86,$36,$FF,$10,$4F,$FB,$63,$E4,$7F,$C3,$BF,$27
	.dc.b	$FE,$E3,$F7,$BF,$C2,$E7,$F8,$70,$F3,$57,$F4,$0A,$FF,$3A,$7D,$37
	.dc.b	$EB,$7D,$FF,$F6,$EF,$E5,$1F,$CF,$27,$FC,$F1,$5F,$BA,$E4,$7E,$9B
	.dc.b	$FE,$87,$F8,$EF,$F5,$7C,$3E,$17,$CF,$BF,$F4,$FB,$F8,$7F,$3B,$FD
	.dc.b	$83,$E7,$7E,$67,$87,$FA,$5F,$48,$43,$F9,$CF,$D9,$7F,$86,$3E,$CB
	.dc.b	$FC,$2F,$1A,$FA,$2F,$E8,$3F,$3B,$FE,$54,$F6,$3F,$37,$2F,$C7,$1F
	.dc.b	$63,$0F,$F0,$C6,$5F,$D1,$A7,$97,$CB,$7F,$86,$2F,$EE,$1F,$E7,$EF
	.dc.b	$AF,$FF,$6A,$EF,$FE,$4B,$F8,$DF,$F4,$0F,$D7,$7F,$B4,$F9,$FF,$D3
	.dc.b	$7D,$7F,$7F,$3E,$99,$BD,$C7,$D6,$BF,$06,$7E,$5A,$FC,$83,$D1,$AF
	.dc.b	$EB,$70,$F6,$7E,$06,$FC,$23,$FD,$0B,$F9,$06,$AB,$E2,$57,$B4,$BA
	.dc.b	$78,$3F,$0A,$FF,$49,$3F,$E3,$1D,$C3,$EB,$BF,$D6,$1E,$FF,$F9,$EB
	.dc.b	$BF,$E3,$F0,$F4,$E9,$FF,$3E,$F8,$BF,$DF,$F9,$3D,$8F,$F7,$3F,$C8
	.dc.b	$5F,$B9,$3F,$71,$7E,$42,$FC,$F1,$F9,$1F,$EB,$BF,$58,$FF,$43,$FD
	.dc.b	$D7,$DF,$E5,$FF,$79,$F1,$BF,$60,$ED,$1F,$E2,$33,$FC,$F9,$B7,$EA
	.dc.b	$7D,$AF,$D5,$78,$5F,$C4,$35,$7E,$77,$C7,$F0,$17,$6B,$F5,$5F,$F4
	.dc.b	$37,$E7,$5F,$E6,$DC,$FF,$D0,$BA,$F6,$7F,$D1,$1D,$7F,$ED,$2E,$3F
	.dc.b	$F6,$97,$D3,$7F,$4B,$F3,$BF,$70,$7E,$E7,$83,$FC,$FB,$5F,$2A,$BF
	.dc.b	$DB,$1F,$77,$FE,$21,$F4,$3F,$7C,$7F,$53,$C7,$FC,$38,$7F,$89,$0F
	.dc.b	$F4,$B7,$B2,$EF,$FF,$12,$79,$5F,$C1,$3D,$5F,$F7,$73,$FA,$8F,$4F
	.dc.b	$F4,$FE,$AF,$C6,$E7,$FC,$DF,$83,$FE,$79,$CF,$F0,$5E,$37,$E9,$7C
	.dc.b	$6F,$F0,$E7,$27,$FB,$87,$87,$FD,$DB,$F5,$3F,$E1,$B8,$FE,$A1,$C7
	.dc.b	$FC,$4E,$3F,$C3,$7B,$FF,$19,$3F,$0F,$5F,$E3,$DC,$3F,$16,$FE,$A9
	.dc.b	$7E,$27,$FB,$3F,$53,$F2,$53,$FE,$FA,$D7,$FA,$07,$E4,$2F,$C8,$7F
	.dc.b	$4F,$CB,$FE,$B1,$FC,$85,$F2,$BF,$D0,$1F,$C8,$47,$E7,$F3,$7E,$15
	.dc.b	$FE,$89,$FB,$4C,$CF,$FB,$AF,$E1,$7F,$91,$9F,$C9,$38,$BE,$54,$71
	.dc.b	$B7,$E4,$5E,$59,$FD,$06,$DF,$C9,$6B,$FC,$8F,$1F,$B4,$E6,$FD,$4F
	.dc.b	$07,$6C,$3F,$87,$79,$F8,$AF,$CA,$5F,$F8,$30,$FF,$29,$2F,$27,$FC
	.dc.b	$EF,$F9,$D7,$F8,$6B,$27,$BF,$FE,$E4,$F0,$DF,$B4,$7F,$B5,$FF,$47
	.dc.b	$FF,$11,$65,$FE,$D3,$F9,$E7,$E8,$4F,$E0,$07,$C9,$BC,$1F,$F3,$27
	.dc.b	$CE,$7F,$90,$7F,$14,$7D,$17,$F9,$C7,$2F,$E9,$90,$FC,$17,$FC,$51
	.dc.b	$1F,$D0,$38,$3F,$7F,$F1,$9F,$E2,$98,$FF,$35,$7B,$9F,$C0,$A5,$3F
	.dc.b	$D5,$7C,$AF,$CD,$57,$E7,$CC,$1F,$D1,$0F,$F3,$D9,$FD,$19,$FC,$C3
	.dc.b	$DC,$FC,$C1,$FE,$F0,$3F,$A0,$FD,$A7,$F8,$E9,$FF,$69,$FD,$8B,$F0
	.dc.b	$E8,$7F,$88,$3C,$57,$57,$F8,$96,$73,$3F,$E2,$9F,$B4,$FF,$16,$7E
	.dc.b	$2B,$AD,$3F,$C0,$D3,$FE,$CA,$FB,$0F,$F1,$16,$0F,$C7,$BF,$C4,$5F
	.dc.b	$2C,$26,$9F,$EE,$49,$F8,$58,$76,$F9,$7F,$D5,$F8,$5F,$D4,$3C,$4F
	.dc.b	$E5,$3E,$67,$F4,$AF,$47,$F9,$7F,$8B,$FE,$16,$CB,$F3,$15,$FC,$EF
	.dc.b	$E2,$FF,$9E,$FD,$36,$67,$F1,$BF,$16,$FA,$EF,$E2,$AE,$7E,$A3,$E0
	.dc.b	$FC,$97,$F8,$72,$FF,$DA,$3C,$1F,$98,$F8,$CF,$F0,$47,$D4,$7E,$57
	.dc.b	$83,$FB,$1E,$DF,$D0,$3F,$1D,$F8,$FE,$D7,$F9,$7F,$6F,$F5,$37,$BD
	.dc.b	$EC,$CB,$FD,$6F,$F8,$17,$F8,$FF,$F2,$E7,$CE,$BF,$6C,$FF,$10,$EB
	.dc.b	$FE,$52,$B7,$EF,$1F,$5F,$2F,$CF,$FE,$0D,$FE,$1D,$EB,$FF,$02,$FC
	.dc.b	$77,$F8,$6F,$E7,$FF,$C3,$9B,$7C,$A8,$FF,$0F,$7C,$07,$DD,$FB,$2F
	.dc.b	$CE,$F1,$BE,$DD,$F9,$27,$F7,$3B,$FF,$12,$65,$FE,$71,$D3,$FC,$6D
	.dc.b	$47,$FC,$3C,$7E,$2C,$C7,$0F,$F6,$E7,$FC,$61,$E6,$7F,$8B,$7E,$93
	.dc.b	$FD,$05,$B7,$93,$0D,$57,$FB,$6D,$FD,$1B,$FE,$CB,$FC,$6F,$FD,$B1
	.dc.b	$E6,$FF,$87,$BA,$3F,$3D,$5F,$E1,$CA,$FC,$F9,$FD,$FC,$FF,$6B,$5F
	.dc.b	$F8,$79,$FD,$77,$E2,$5F,$A0,$7F,$B6,$FC,$47,$F8,$77,$9B,$F3,$5E
	.dc.b	$5F,$E7,$BE,$99,$F9,$9F,$F5,$8F,$D0,$3F,$35,$87,$FA,$2A,$BF,$47
	.dc.b	$F2,$EF,$AD,$FD,$AE,$1F,$8B,$F0,$BF,$BF,$FC,$47,$F2,$23,$0F,$F8
	.dc.b	$97,$D0,$FD,$00,$FC,$57,$AE,$5F,$A7,$E1,$FF,$00,$D7,$F2,$5D,$FE
	.dc.b	$E7,$0F,$F8,$5D,$FE,$7D,$E6,$7F,$85,$BE,$93,$FC,$2B,$C5,$FC,$BB
	.dc.b	$1F,$F0,$7E,$1F,$9B,$F9,$9F,$F0,$A7,$E2,$FF,$A8,$8F,$F1,$BF,$C6
	.dc.b	$7F,$84,$BD,$3F,$F1,$8F,$C6,$7E,$AB,$0F,$E2,$9B,$27,$FC,$23,$97
	.dc.b	$F3,$59,$FF,$08,$61,$FE,$3F,$F4,$FF,$C2,$31,$F8,$3E,$AF,$EE,$6F
	.dc.b	$F3,$2F,$77,$F0,$CB,$F4,$57,$F0,$0B,$F7,$3F,$F0,$9F,$9D,$FE,$9F
	.dc.b	$F6,$DF,$AA,$7F,$0E,$F3,$F8,$E7,$FD,$DF,$2F,$F5,$37,$9F,$87,$F0
	.dc.b	$07,$F8,$4C,$FF,$02,$17,$F0,$07,$EC,$0F,$DD,$9F,$8C,$BF,$0E,$7E
	.dc.b	$B4,$FC,$BD,$FA,$6F,$F8,$53,$2F,$61,$FE,$63,$F8,$BF,$E4,$2E,$0F
	.dc.b	$D1,$7D,$27,$F3,$37,$E4,$FD,$7F,$F8,$5B,$7F,$87,$1F,$A7,$F9,$3D
	.dc.b	$3F,$E7,$99,$FF,$0A,$CF,$FA,$5F,$F0,$5F,$F0,$AB,$FC,$EC,$FF,$0A
	.dc.b	$4F,$FB,$A7,$DC,$FF,$22,$F5,$7F,$9F,$3E,$3B,$F9,$EC,$FF,$47,$FC
	.dc.b	$DB,$F3,$DE,$0F,$F8,$67,$E8,$FF,$C7,$2F,$F0,$5C,$7F,$B6,$FC,$27
	.dc.b	$F8,$5B,$DB,$FF,$3B,$FC,$A7,$F1,$9F,$0D,$F1,$7F,$E1,$6C,$3F,$C1
	.dc.b	$9F,$6C,$27,$CF,$FF,$9F,$2B,$FD,$03,$F8,$BF,$E7,$FC,$8F,$EF,$7E
	.dc.b	$87,$F9,$70,$FF,$0B,$4F,$FA,$16,$3F,$D1,$0F,$FA,$1F,$F0,$6F,$CF
	.dc.b	$70,$FF,$0B,$F8,$DF,$E5,$FF,$8D,$3F,$1B,$7E,$36,$CF,$F9,$A7,$BB
	.dc.b	$FD,$53,$F1,$B7,$E5,$6F,$8F,$47,$F9,$FF,$8B,$FD,$F5,$A3,$A3,$FC
	.dc.b	$47,$1F,$E3,$0F,$AB,$1F,$84,$E1,$FD,$57,$4F,$F8,$C6,$32,$7E,$EB
	.dc.b	$8B,$FD,$3F,$F5,$7F,$E1,$9F,$57,$FC,$69,$97,$F8,$61,$FF,$19,$7A
	.dc.b	$FF,$E1,$DE,$CF,$F2,$1D,$FF,$87,$FC,$EF,$F0,$25,$FF,$85,$BC,$FF
	.dc.b	$F0,$AF,$85,$FE,$80,$FA,$6F,$F0,$A7,$D2,$FE,$86,$BF,$C3,$87,$F2
	.dc.b	$FF,$8D,$FF,$6B,$7C,$A7,$FB,$BF,$F7,$D7,$E1,$3E,$DA,$7E,$E0,$FB
	.dc.b	$87,$9E,$FC,$8F,$C3,$5F,$CE,$63,$E3,$BD,$87,$37,$F4,$0E,$8F,$F4
	.dc.b	$9F,$1F,$F8,$56,$7F,$C1,$DE,$7F,$E6,$BC,$DF,$F0,$B4,$FF,$85,$6F
	.dc.b	$F3,$55,$F9,$E3,$FC,$68,$FF,$3F,$3F,$C2,$9F,$0B,$FC,$D6,$3F,$A0
	.dc.b	$9F,$C1,$9F,$F6,$93,$FE,$E4,$F1,$FF,$D9,$E7,$F5,$D7,$F8,$5F,$9F
	.dc.b	$F7,$72,$FE,$15,$F9,$BB,$F4,$FF,$E4,$71,$FC,$16,$BF,$AB,$7E,$6A
	.dc.b	$FD,$25,$F9,$1B,$F2,$2F,$E4,$9F,$0D,$FC,$D3,$E3,$7F,$9E,$78,$BF
	.dc.b	$EC,$4F,$88,$FF,$69,$7E,$11,$FE,$D2,$E0,$FE,$DD,$E6,$7F,$89,$3F
	.dc.b	$35,$FE,$73,$C9,$FC,$7B,$7F,$F6,$96,$3F,$EC,$CF,$17,$FC,$BE,$7F
	.dc.b	$B9,$BD,$6F,$E6,$59,$FF,$85,$A7,$FD,$05,$C5,$FE,$C0,$FA,$B7,$D8
	.dc.b	$FF,$92,$BF,$17,$EB,$97,$F2,$AF,$C5,$DF,$73,$FE,$76,$7F,$37,$87
	.dc.b	$F4,$19,$7D,$2F,$EE,$F9,$3F,$C2,$59,$FD,$C9,$FE,$F1,$D9,$E7,$F2
	.dc.b	$7F,$C2,$5C,$3F,$CD,$BF,$3C,$7E,$8B,$FE,$C4,$FA,$37,$EC,$0F,$C7
	.dc.b	$1F,$44,$FD,$E3,$FC,$87,$F8,$8F,$F3,$EE,$1F,$F2,$96,$DF,$E0,$5D
	.dc.b	$FE,$A3,$FD,$33,$B3,$FC,$2A,$FF,$88,$CF,$F0,$56,$FF,$E3,$EF,$27
	.dc.b	$FA,$77,$57,$E5,$FE,$C1,$F8,$C3,$F3,$87,$C5,$FD,$9F,$C2,$5E,$AF
	.dc.b	$E5,$FA,$3F,$E0,$06,$FF,$E1,$BD,$7F,$C2,$A7,$E6,$8F,$EA,$77,$F9
	.dc.b	$08,$FC,$CC,$7F,$56,$FA,$AF,$F0,$AB,$FD,$63,$6F,$F2,$8F,$4F,$FB
	.dc.b	$32,$BF,$B1,$4B,$EA,$BF,$AA,$3F,$E3,$88,$FF,$3C,$7D,$5F,$F5,$7F
	.dc.b	$47,$FC,$2D,$7F,$A1,$AF,$F0,$B6,$5F,$99,$E3,$FC,$47,$27,$FB,$02
	.dc.b	$7F,$A5,$3F,$EA,$4C,$43,$F4,$DF,$CF,$C7,$F1,$69,$FC,$87,$CB,$3C
	.dc.b	$F7,$E8,$5F,$DF,$FC,$97,$84,$F3,$27,$36,$7C,$AF,$E7,$9F,$1E,$F8
	.dc.b	$FF,$C5,$EF,$FE,$01,$F3,$BF,$D0,$5F,$69,$FE,$C6,$DF,$E8,$C7,$FA
	.dc.b	$17,$F1,$5F,$D9,$57,$E3,$67,$F9,$AC,$F0,$F8,$D7,$C7,$FF,$3B,$E3
	.dc.b	$FF,$0B,$7C,$87,$FA,$22,$BF,$9C,$7C,$67,$DE,$9F,$D4,$70,$FF,$85
	.dc.b	$71,$FF,$64,$78,$9F,$EE,$99,$F8,$9F,$2B,$FE,$A5,$E9,$FE,$B1,$CA
	.dc.b	$FB,$1F,$F0,$A5,$3B,$7F,$D6,$5D,$BF,$EC,$43,$FC,$27,$F7,$FF,$D3
	.dc.b	$3F,$35,$FF,$40,$71,$BE,$D3,$FA,$53,$FE,$9E,$FD,$BF,$F2,$DB,$FE
	.dc.b	$F4,$FF,$50,$FD,$D7,$F0,$DC,$71,$E6,$7D,$99,$FE,$20,$FD,$87,$FC
	.dc.b	$CD,$AF,$FB,$83,$F0,$C7,$EF,$0F,$DE,$BF,$9B,$7E,$D6,$CF,$FC,$21
	.dc.b	$87,$CB,$F9,$DE,$BC,$5E,$A8,$FF,$08,$1F,$EA,$4D,$99,$7F,$B1,$FF
	.dc.b	$08,$FF,$10,$CF,$F8,$C3,$E8,$7F,$91,$3F,$0E,$BF,$D9,$95,$FE,$CD
	.dc.b	$FC,$3F,$F3,$33,$FE,$DB,$3F,$CF,$98,$FF,$83,$E9,$CF,$FE,$0F,$CB
	.dc.b	$D9,$5E,$19,$FF,$7B,$9F,$E3,$EE,$7F,$F2,$8F,$AB,$FE,$EE,$C3,$FC
	.dc.b	$0D,$0F,$17,$F8,$4F,$E7,$3F,$E8,$FC,$7F,$90,$E1,$F9,$EF,$9B,$FE
	.dc.b	$E7,$F4,$FF,$C8,$33,$FF,$24,$C7,$F7,$2A,$DE,$9F,$83,$BF,$F7,$76
	.dc.b	$5F,$E0,$D7,$C4,$88,$3E,$73,$19,$FE,$C7,$F0,$77,$C3,$BE,$FF,$C1
	.dc.b	$A7,$FA,$77,$F0,$E7,$C3,$7F,$30,$D7,$C3,$3F,$EA,$1F,$A0,$FC,$8F
	.dc.b	$E1,$DF,$AA,$F3,$FF,$27,$E3,$FF,$88,$F8,$7C,$7C,$1E,$2F,$A0,$F0
	.dc.b	$E5,$FB,$DF,$95,$FF,$04,$67,$FE,$AB,$AF,$F3,$69,$F1,$E7,$FC,$E0
	.dc.b	$FF,$B7,$0E,$3E,$27,$F7,$A9,$FF,$38,$8F,$D8,$5D,$BF,$C5,$7F,$12
	.dc.b	$FE,$13,$F8,$BB,$F0,$FF,$B7,$97,$E3,$F0,$FF,$00,$78,$EF,$8D,$E9
	.dc.b	$F0,$9F,$D6,$FF,$48,$7E,$5A,$F6,$DF,$BD,$BF,$41,$FF,$6B,$FC,$73
	.dc.b	$F1,$0F,$F0,$7F,$EB,$9E,$5D,$1F,$AC,$F9,$FD,$3F,$F0,$67,$8D,$FE
	.dc.b	$BE,$E0,$FF,$1A,$47,$F4,$87,$FC,$D6,$7E,$7B,$D7,$FF,$80,$3F,$11
	.dc.b	$FC,$E2,$75,$E6,$7F,$9A,$BE,$A7,$F9,$2E,$1F,$B5,$F8,$9F,$E5,$98
	.dc.b	$FF,$B5,$7E,$87,$ED,$BF,$A1,$9F,$E4,$9E,$EF,$FB,$7F,$D1,$7D,$17
	.dc.b	$F2,$7F,$15,$F1,$BF,$D7,$BE,$3F,$FD,$6B,$F0,$6F,$83,$83,$E1,$DF
	.dc.b	$77,$3E,$CD,$E1,$2F,$FD,$43,$F8,$80,$9C,$1F,$C2,$E7,$F5,$67,$EE
	.dc.b	$F1,$ED,$E4,$F2,$F1,$3F,$13,$5B,$FA,$FF,$C5,$F9,$7F,$CF,$0F,$F0
	.dc.b	$66,$1F,$E0,$FE,$0F,$CF,$FC,$1F,$FB,$33,$C0,$F3,$E7,$E7,$F8,$9F
	.dc.b	$F3,$67,$EA,$BE,$6F,$DC,$FF,$D6,$BF,$3F,$FE,$0B,$D7,$C9,$E9,$FF
	.dc.b	$AE,$38,$B8,$F2,$BF,$11,$9F,$6F,$FD,$1B,$D4,$FF,$5E,$FD,$0F,$F3
	.dc.b	$2E,$0F,$D7,$FD,$87,$C2,$F1,$7F,$94,$E8,$F1,$3E,$BF,$2B,$F1,$9F
	.dc.b	$F6,$2F,$E4,$5F,$98,$FC,$41,$EB,$FF,$60,$FD,$0B,$F3,$1B,$7C,$CF
	.dc.b	$D6,$7F,$41,$8B,$E5,$FF,$C2,$1E,$4F,$F9,$FB,$5F,$F3,$31,$E0,$D7
	.dc.b	$F0,$BE,$87,$FB,$67,$F3,$D7,$D0,$3F,$1B,$7E,$8D,$FE,$0D,$FC,$67
	.dc.b	$FD,$13,$F8,$A7,$F8,$23,$3F,$EF,$F9,$FF,$7D,$E5,$FC,$CC,$3D,$4F
	.dc.b	$F0,$44,$7F,$90,$7C,$DF,$E8,$DB,$7F,$97,$3D,$5F,$CC,$E0,$3F,$C1
	.dc.b	$15,$F9,$FF,$A9,$FF,$03,$BF,$E4,$9A,$FC,$DC,$FF,$81,$B8,$BF,$92
	.dc.b	$E5,$FE,$06,$3F,$9D,$C7,$EF,$8F,$F1,$73,$F5,$FC,$6F,$BF,$F8,$A3
	.dc.b	$A5,$FA,$87,$E6,$A7,$F0,$7F,$8B,$38,$3D,$59,$7F,$3E,$CB,$F8,$97
	.dc.b	$99,$FC,$A0,$FE,$34,$E1,$FE,$57,$F6,$61,$F8,$AF,$E8,$35,$FF,$7F
	.dc.b	$DF,$F5,$5A,$FE,$3F,$E5,$7F,$34,$F3,$BF,$B5,$63,$FE,$08,$E4,$DF
	.dc.b	$9F,$F9,$CF,$61,$F8,$A3,$F1,$57,$E5,$1F,$E1,$7E,$B7,$D8,$3C,$7F
	.dc.b	$F0,$F6,$DF,$E6,$E3,$FC,$38,$7E,$93,$2F,$C2,$F0,$7F,$8A,$9F,$F0
	.dc.b	$47,$07,$F2,$C9,$FE,$8B,$A7,$F8,$23,$B3,$E6,$FC,$5F,$F8,$1F,$83
	.dc.b	$EC,$7B,$5F,$C1,$BE,$43,$FD,$C5,$9B,$F5,$CF,$F2,$76,$7F,$93,$BF
	.dc.b	$BD,$DF,$F7,$FE,$07,$DF,$E3,$FF,$81,$67,$F8,$E9,$FE,$B4,$C3,$F0
	.dc.b	$8B,$F2,$47,$E3,$1F,$ED,$BF,$D1,$3F,$61,$F2,$AF,$DF,$9F,$AA,$7F
	.dc.b	$81,$B4,$FF,$51,$E3,$FE,$91,$CF,$F8,$8B,$FE,$06,$3F,$0C,$7F,$8D
	.dc.b	$4F,$F0,$47,$D5,$FF,$81,$E7,$FD,$31,$C5,$FE,$1D,$F8,$6F,$F0,$3F
	.dc.b	$27,$F2,$CD,$D7,$FE,$07,$BF,$E6,$DF,$11,$E8,$F8,$9F,$F0,$35,$7F
	.dc.b	$2D,$CB,$FC,$EB,$C7,$FE,$06,$8F,$F0,$B5,$FF,$67,$8F,$C7,$4F,$E3
	.dc.b	$7A,$3F,$BB,$7C,$4F,$F8,$1E,$73,$1F,$91,$F9,$DF,$F0,$3F,$C4,$FF
	.dc.b	$82,$A3,$FB,$74,$FF,$84,$9E,$5F,$86,$BF,$1C,$93,$FE,$69,$F8,$4F
	.dc.b	$F2,$41,$FE,$06,$AF,$F0,$24,$71,$7F,$3D,$F0,$3F,$DD,$3C,$0F,$C3
	.dc.b	$ED,$FE,$20,$F9,$FF,$F0,$2F,$CF,$7F,$98,$77,$FF,$58,$7C,$3F,$F9
	.dc.b	$81,$F6,$5F,$0B,$7D,$0F,$F8,$6B,$3F,$F3,$B5,$7F,$86,$A3,$31,$EE
	.dc.b	$CF,$F5,$DF,$43,$E1,$AF,$E0,$D7,$F4,$75,$FF,$02,$C7,$C7,$AF,$F6
	.dc.b	$5F,$0F,$F9,$83,$C5,$FF,$0E,$BF,$D4,$3D,$EF,$E4,$BC,$BF,$8C,$FB
	.dc.b	$EF,$EF,$1E,$23,$F4,$1F,$F4,$5F,$DE,$7F,$19,$CD,$E0,$3D,$2F,$F0
	.dc.b	$0C,$7E,$53,$E7,$DF,$11,$FD,$07,$F0,$A7,$E1,$5F,$EB,$2F,$C0,$1F
	.dc.b	$9C,$72,$F9,$CF,$F0,$07,$EE,$9F,$DF,$F2,$F9,$7F,$17,$F9,$EF,$4B
	.dc.b	$FC,$9F,$1F,$DC,$0F,$EF,$DD,$3F,$EB,$27,$FD,$7B,$C9,$F8,$E7,$FC
	.dc.b	$B3,$F0,$9F,$DF,$A3,$FD,$DD,$C9,$F4,$65,$E4,$7F,$7E,$AF,$E9,$B7
	.dc.b	$FC,$48,$FC,$D7,$E0,$AF,$BD,$7E,$9D,$FC,$5F,$0F,$F7,$19,$FE,$F9
	.dc.b	$FB,$E7,$E1,$0F,$C2,$3F,$C0,$FF,$09,$FC,$7F,$E1,$7F,$D5,$1F,$7F
	.dc.b	$FE,$AA,$F1,$BF,$9A,$FC,$47,$F8,$17,$F0,$8F,$F0,$7B,$FE,$A8,$87
	.dc.b	$95,$FD,$E3,$C7,$7D,$6F,$F1,$5D,$BF,$96,$47,$F9,$94,$FF,$2C,$7C
	.dc.b	$C7,$F8,$2B,$CA,$E1,$F0,$CF,$CC,$7F,$B9,$65,$F8,$9B,$FF,$2E,$FC
	.dc.b	$2F,$F0,$8F,$86,$FF,$5C,$E1,$FD,$D2,$BF,$8F,$57,$F9,$8F,$CE,$FE
	.dc.b	$EB,$1F,$EB,$7F,$89,$FF,$4E,$7D,$27,$F7,$4F,$77,$F8,$9F,$85,$FD
	.dc.b	$B3,$F3,$2F,$E7,$7C,$7F,$DB,$B3,$FE,$73,$F5,$BF,$D7,$F8,$BF,$8C
	.dc.b	$9F,$E2,$8E,$FF,$F6,$BF,$9D,$FB,$EF,$A4,$FF,$10,$7E,$D8,$FD,$19
	.dc.b	$FB,$83,$F2,$17,$EF,$4F,$C9,$7F,$DF,$79,$FF,$7C,$E3,$FF,$7B,$FC
	.dc.b	$1F,$F6,$9D,$3F,$C5,$33,$FE,$FB,$3F,$DF,$7A,$FF,$1D,$3F,$8C,$71
	.dc.b	$7F,$74,$8F,$F0,$8F,$27,$F7,$39,$FC,$77,$E1,$7F,$EA,$0F,$C3,$FF
	.dc.b	$B8,$D7,$F3,$2F,$2F,$F1,$CB,$CD,$FE,$75,$D3,$FE,$CB,$8F,$EF,$5A
	.dc.b	$7F,$86,$FE,$20,$FF,$7D,$C3,$FB,$67,$A1,$FE,$42,$F8,$7F,$E7,$BF
	.dc.b	$94,$7F,$97,$71,$7E,$95,$FE,$24,$CB,$FA,$0F,$0F,$F2,$4F,$C2,$BF
	.dc.b	$DF,$5F,$AB,$FF,$A6,$7C,$FF,$F3,$A7,$D2,$FF,$85,$EB,$BD,$FF,$4E
	.dc.b	$FB,$7F,$ED,$6F,$EE,$BF,$01,$E8,$FF,$B1,$B8,$FF,$D6,$CF,$F3,$4E
	.dc.b	$7F,$C9,$78,$DF,$EA,$5F,$BE,$FF,$44,$FA,$3F,$E8,$8E,$4E,$BF,$90
	.dc.b	$FF,$81,$70,$FF,$2D,$FC,$9F,$F9,$EF,$AF,$FA,$BF,$C7,$FF,$02,$FD
	.dc.b	$8D,$FB,$43,$F3,$0F,$F4,$8F,$E1,$8F,$D0,$FF,$86,$F8,$7F,$C0,$F3
	.dc.b	$FE,$3B,$97,$FB,$27,$E1,$7F,$AA,$7A,$1F,$EC,$B8,$7E,$89,$FD,$31
	.dc.b	$1F,$E1,$5F,$13,$FC,$DB,$EC,$FF,$2F,$8C,$FC,$FF,$F8,$0B,$9B,$FD
	.dc.b	$81,$FA,$77,$F9,$CB,$A3,$DF,$0F,$CD,$3F,$CA,$9A,$FF,$8C,$FF,$80
	.dc.b	$7F,$BE,$7D,$7F,$F8,$D7,$EB,$FF,$BD,$FC,$4F,$E8,$BC,$EF,$EA,$F7
	.dc.b	$FD,$DB,$D6,$FC,$FF,$E8,$2F,$D0,$DF,$8D,$BF,$26,$FF,$40,$FE,$D7
	.dc.b	$FE,$A3,$F9,$BF,$F7,$DF,$27,$D8,$CB,$F4,$E7,$E5,$DF,$F7,$57,$95
	.dc.b	$F0,$9F,$F7,$67,$C9,$FF,$26,$E8,$6F,$F6,$E7,$FB,$A7,$46,$27,$E0
	.dc.b	$72,$FE,$B7,$D0,$FE,$1F,$97,$F0,$C8,$FF,$70,$FE,$E1,$FE,$94,$EE
	.dc.b	$FF,$07,$FE,$63,$F9,$3C,$FF,$C2,$3F,$84,$FF,$A1,$BE,$0B,$FD,$0F
	.dc.b	$97,$F9,$D7,$2F,$F2,$C4,$BF,$19,$7E,$84,$FD,$6F,$FD,$49,$F9,$CB
	.dc.b	$F2,$5F,$F2,$F6,$3E,$8C,$7F,$CC,$1F,$6E,$FD,$4B,$FD,$51,$1F,$EA
	.dc.b	$1E,$BF,$E4,$9F,$8D,$FF,$BD,$BB,$BF,$85,$ED,$E3,$DF,$FD,$5F,$EC
	.dc.b	$7E,$0B,$0F,$1F,$E4,$FF,$E4,$1F,$87,$7E,$65,$FE,$D9,$FB,$77,$E4
	.dc.b	$FF,$E2,$DB,$FF,$76,$78,$1F,$ED,$5F,$C8,$5D,$93,$AB,$F8,$67,$D1
	.dc.b	$FF,$8C,$A1,$FA,$FF,$F6,$F7,$FC,$51,$BF,$8C,$7E,$5B,$FE,$2B,$E1
	.dc.b	$C4,$68,$F4,$7E,$B3,$5F,$F6,$3F,$E2,$9F,$EC,$9F,$C4,$3F,$C6,$5F
	.dc.b	$A0,$FC,$FF,$DE,$3C,$7B,$E2,$FF,$D6,$F7,$FE,$CE,$EC,$FB,$7C,$3E
	.dc.b	$3D,$7F,$AA,$47,$F3,$C7,$FC,$ED,$F4,$7F,$ED,$0F,$63,$FC,$97,$E5
	.dc.b	$EE,$BE,$95,$BF,$5E,$7E,$C1,$FD,$F3,$F7,$EF,$F6,$E7,$E0,$1F,$CC
	.dc.b	$AB,$F4,$A3,$E8,$FE,$C7,$17,$F9,$E3,$1F,$D0,$4F,$E3,$7E,$AB,$E5
	.dc.b	$ED,$F3,$E1,$A7,$9F,$C0,$7B,$7F,$E5,$EF,$80,$7E,$D9,$FC,$B7,$F3
	.dc.b	$6F,$CE,$FE,$26,$FD,$D9,$E0,$73,$D1,$D2,$FC,$73,$E8,$C7,$D9,$DB
	.dc.b	$FC,$33,$E0,$FD,$9F,$89,$FB,$1E,$9F,$F5,$4E,$DF,$B1,$FB,$DF,$D3
	.dc.b	$7F,$DC,$7E,$E5,$7F,$D9,$5F,$8D,$FF,$BE,$EB,$FA,$DF,$85,$F5,$7E
	.dc.b	$B3,$C3,$18,$AF,$F7,$5F,$DE,$FF,$BF,$5F,$E1,$DA,$7E,$BF,$5F,$C6
	.dc.b	$FD,$EF,$D3,$8F,$F6,$06,$5F,$89,$BF,$E8,$91,$FA,$4B,$7E,$CF,$F2
	.dc.b	$30,$7E,$E7,$FA,$9E,$DF,$EA,$17,$FB,$0F,$C8,$FF,$52,$FF,$70,$87
	.dc.b	$B3,$F9,$9B,$FF,$21,$FE,$03,$F5,$3B,$BF,$51,$7E,$7C,$FF,$7D,$A7
	.dc.b	$F6,$E8,$FF,$6A,$7B,$EF,$B8,$FF,$6E,$7C,$3F,$FA,$77,$F5,$E7,$96
	.dc.b	$FE,$00,$E1,$87,$E6,$39,$DE,$A7,$E0,$BB,$FF,$BB,$7E,$47,$FA,$BE
	.dc.b	$8F,$A7,$7F,$CC,$BF,$33,$FF,$49,$75,$FE,$B7,$F6,$2B,$CB,$FE,$5D
	.dc.b	$FD,$CE,$BF,$8C,$F8,$2F,$C6,$F7,$7F,$83,$AF,$FD,$11,$F6,$AE,$77
	.dc.b	$3F,$D2,$D3,$F4,$DE,$ED,$CB,$F7,$5F,$29,$F9,$CF,$3B,$F4,$99,$FF
	.dc.b	$AC,$3F,$17,$FE,$B3,$EC,$3F,$5A,$FE,$0B,$F1,$2F,$C7,$FF,$8D,$F5
	.dc.b	$FF,$AD,$BD,$67,$8B,$F4,$B7,$FF,$6F,$F8,$9F,$E3,$DD,$7F,$8D,$47
	.dc.b	$EC,$B7,$FE,$15,$E1,$FF,$A8,$3E,$BD,$F9,$A3,$F5,$59,$F5,$0F,$CA
	.dc.b	$3F,$D8,$5F,$9F,$FC,$EC,$7F,$43,$E2,$3F,$5E,$F9,$DA,$7F,$9E,$3C
	.dc.b	$3F,$D0,$57,$FA,$17,$A3,$FC,$66,$FF,$19,$EA,$F9,$6F,$FB,$EF,$F5
	.dc.b	$CA,$FE,$E3,$E6,$6F,$83,$FE,$54,$8C,$F3,$3D,$1F,$9B,$E7,$BE,$57
	.dc.b	$FD,$AF,$F5,$5F,$E1,$B8,$FE,$97,$B7,$F8,$49,$FE,$1F,$1F,$D3,$60
	.dc.b	$3F,$0D,$7E,$A4,$FB,$FF,$E6,$F2,$FB,$77,$E8,$3F,$EC,$AF,$C4,$BF
	.dc.b	$DC,$BF,$91,$FF,$8A,$7E,$BF,$C3,$7F,$E2,$0F,$98,$FF,$28,$E9,$FE
	.dc.b	$22,$8F,$CA,$FE,$AC,$F8,$D7,$DE,$FF,$57,$FA,$37,$D9,$BE,$CF,$FA
	.dc.b	$A7,$8C,$FC,$AF,$FD,$E7,$CC,$FB,$7F,$F1,$6C,$FC,$1D,$3D,$7A,$F7
	.dc.b	$EF,$F0,$C7,$F8,$C9,$F1,$3F,$EF,$CF,$D9,$7F,$81,$3F,$CA,$FF,$3F
	.dc.b	$F7,$7E,$A1,$F1,$3F,$1F,$FD,$84,$BB,$1F,$97,$7F,$8E,$A3,$F9,$67
	.dc.b	$C2,$FF,$A6,$3C,$C6,$4F,$C4,$19,$BF,$1C,$69,$FE,$BC,$E2,$E5,$B7
	.dc.b	$F9,$DB,$97,$FC,$5D,$F6,$DF,$E0,$F7,$FD,$FC,$3E,$87,$FC,$60,$FD
	.dc.b	$59,$FF,$7D,$9F,$6B,$67,$E4,$EF,$C4,$7E,$C7,$D3,$7D,$EF,$CC,$3E
	.dc.b	$CD,$7F,$4A,$F9,$DF,$BF,$E3,$7F,$88,$23,$FD,$9F,$F5,$7C,$FF,$32
	.dc.b	$7E,$DD,$FE,$9D,$FC,$B7,$FD,$63,$F8,$AF,$8F,$C0,$F5,$61,$FE,$0C
	.dc.b	$FC,$33,$E7,$3F,$E4,$1E,$17,$E5,$7E,$B7,$FC,$8E,$7F,$92,$3F,$5E
	.dc.b	$FE,$B7,$E8,$FF,$76,$F0,$3F,$C3,$96,$FD,$3F,$FC,$07,$F3,$DF,$D3
	.dc.b	$7E,$35,$FA,$17,$F3,$1E,$9F,$81,$EC,$3D,$EF,$E0,$18,$FF,$87,$BE
	.dc.b	$D7,$F8,$07,$B5,$FE,$0C,$F4,$BF,$D4,$7B,$7F,$81,$1F,$F5,$2F,$C8
	.dc.b	$3F,$76,$7E,$C1,$FC,$07,$F2,$C7,$E4,$FF,$AF,$FA,$0F,$EC,$1F,$4F
	.dc.b	$FC,$77,$E7,$7F,$7F,$EA,$7F,$BB,$B7,$FD,$5E,$DF,$ED,$B3,$F8,$03
	.dc.b	$F5,$79,$3F,$C2,$DF,$27,$FD,$75,$FE,$4D,$CB,$FE,$0D,$7F,$94,$0F
	.dc.b	$AE,$FF,$69,$E5,$FE,$4A,$FC,$37,$F9,$46,$3F,$DF,$0F,$EF,$BF,$39
	.dc.b	$F9,$A7,$F9,$4E,$5F,$BC,$FC,$41,$B3,$A1,$F3,$5E,$9F,$A0,$FF,$21
	.dc.b	$F5,$FF,$29,$EB,$F9,$38,$FF,$7D,$CB,$F9,$2F,$93,$F8,$DF,$D5,$7F
	.dc.b	$C0,$5F,$21,$FE,$08,$EB,$FE,$C7,$FB,$57,$FB,$1B,$9F,$FA,$DE,$1F
	.dc.b	$E0,$4E,$5F,$F0,$1B,$FE,$08,$CB,$FC,$06,$7F,$73,$7F,$C7,$FA,$7F
	.dc.b	$80,$E7,$FC,$99,$FB,$67,$C5,$7F,$AC,$C6,$5F,$8B,$F3,$FF,$E3,$78
	.dc.b	$FF,$00,$46,$0F,$EA,$1E,$27,$F7,$4F,$AD,$FE,$F8,$FF,$8B,$BC,$87
	.dc.b	$D1,$3E,$E3,$F4,$F9,$FC,$5F,$95,$FF,$66,$5F,$F7,$CF,$A0,$7E,$AF
	.dc.b	$F1,$23,$FB,$DE,$FF,$7A,$FF,$C5,$BD,$DE,$EE,$1F,$EF,$7A,$FF,$82
	.dc.b	$FC,$7F,$F7,$8F,$E6,$CF,$C5,$5F,$90,$7F,$8C,$3F,$7A,$7E,$4E,$F5
	.dc.b	$7F,$BB,$E1,$FD,$1B,$4F,$E0,$1E,$4F,$E3,$B8,$BF,$CC,$5F,$B5,$FF
	.dc.b	$78,$E0,$F9,$5F,$35,$FD,$22,$FF,$D9,$1D,$3F,$DC,$DF,$C3,$1F,$E6
	.dc.b	$CF,$BE,$FE,$9B,$DD,$FD,$27,$B3,$FC,$97,$F8,$4F,$F9,$4F,$EA,$3F
	.dc.b	$85,$7A,$7F,$E7,$6F,$CB,$5F,$AF,$7F,$7B,$FC,$21,$F9,$BF,$F7,$0D
	.dc.b	$FF,$A4,$EF,$F1,$BD,$1F,$F2,$97,$D5,$FF,$87,$F8,$3F,$CA,$3A,$FF
	.dc.b	$70,$BF,$E9,$EF,$F9,$47,$DD,$FE,$31,$F9,$9F,$F5,$2F,$A4,$FF,$7B
	.dc.b	$FB,$CF,$DF,$5F,$8F,$3F,$7B,$FE,$E1,$9F,$F2,$3F,$AA,$F8,$9A,$7E
	.dc.b	$C3,$D6,$F8,$8F,$F9,$CF,$D5,$FF,$36,$BF,$EA,$5A,$CF,$DA,$7F,$A2
	.dc.b	$7F,$19,$F9,$10,$FD,$17,$FB,$8F,$97,$FE,$68,$F3,$3F,$CE,$BE,$23
	.dc.b	$F3,$D7,$E2,$BF,$E6,$8C,$FF,$B4,$FE,$35,$F8,$8F,$A8,$FE,$89,$E8
	.dc.b	$7F,$A1,$3F,$22,$7E,$60,$F8,$1F,$F4,$EF,$A2,$FC,$B3,$FC,$E7,$F7
	.dc.b	$5F,$CA,$B3,$FF,$65,$FD,$3F,$F5,$CF,$91,$FE,$8B,$1F,$CF,$7A,$FF
	.dc.b	$A1,$F8,$1F,$EC,$DE,$3F,$E4,$FF,$4E,$E0,$FE,$D3,$F2,$5F,$DF,$3E
	.dc.b	$A1,$3F,$DF,$B9,$BF,$A3,$64,$FC,$41,$F8,$73,$F5,$FF,$85,$FA,$37
	.dc.b	$F6,$DF,$D8,$9F,$1F,$FC,$E7,$7F,$E7,$3F,$A7,$7F,$68,$FC,$8B,$FD
	.dc.b	$EB,$9B,$F2,$2F,$F7,$AB,$FE,$0C,$F8,$EF,$90,$7F,$4D,$DB,$C3,$F8
	.dc.b	$0B,$F4,$E7,$E1,$2F,$C2,$DF,$B1,$3F,$56,$FD,$3F,$E7,$8F,$13,$FA
	.dc.b	$5E,$9F,$E1,$8E,$0F,$F7,$C3,$EC,$F8,$2F,$EF,$AF,$F8,$0D,$F7,$F6
	.dc.b	$FF,$6A,$DF,$FC,$4D,$F9,$1F,$FA,$AB,$D5,$FE,$A1,$F7,$DF,$D2,$BE
	.dc.b	$4F,$FC,$FF,$C5,$FE,$D2,$FC,$43,$FA,$76,$7F,$EE,$6F,$D7,$7F,$81
	.dc.b	$7D,$27,$E7,$7D,$2F,$F6,$BF,$17,$E7,$69,$9F,$F4,$D7,$FB,$A7,$93
	.dc.b	$FE,$00,$FC,$C3,$F9,$9F,$ED,$4F,$03,$E0,$FE,$99,$FD,$33,$E7,$BF
	.dc.b	$CE,$F9,$FF,$AA,$79,$FF,$A6,$65,$FE,$98,$8F,$E9,$6F,$F8,$E3,$87
	.dc.b	$FD,$C1,$ED,$FE,$7B,$5F,$F6,$FF,$3F,$F2,$49,$FF,$6E,$5F,$FA,$DB
	.dc.b	$CB,$FF,$4B,$FE,$21,$FE,$6C,$FA,$4F,$F3,$5F,$A1,$FD,$0B,$3F,$EB
	.dc.b	$3C,$BF,$D0,$B5,$FF,$6E,$FE,$1D,$FE,$64,$FB,$DF,$F3,$1F,$E1,$9F
	.dc.b	$E6,$29,$FE,$7F,$ED,$FF,$71,$E1,$FF,$30,$F0,$7F,$37,$E0,$FF,$41
	.dc.b	$7C,$7F,$F7,$5F,$88,$FE,$BD,$1F,$EE,$FF,$A5,$FF,$5C,$ED,$FE,$19
	.dc.b	$E5,$FF,$50,$9F,$CE,$BF,$1B,$FF,$0E,$7C,$3F,$F8,$0B,$E7,$3F,$B3
	.dc.b	$E7,$FC,$46,$FF,$AE,$5F,$9F,$C6,$FF,$11,$7C,$9F,$F7,$BF,$AB,$FE
	.dc.b	$F5,$F5,$5F,$DE,$7C,$FF,$E5,$9F,$29,$FE,$94,$ED,$FF,$6B,$67,$FC
	.dc.b	$9D,$FF,$5A,$9F,$CA,$F2,$FF,$36,$FC,$AF,$F8,$B7,$E0,$7F,$B5,$FC
	.dc.b	$5F,$F6,$AE,$3F,$F5,$6F,$C5,$F8,$FE,$47,$DD,$C3,$FA,$E3,$FC,$91
	.dc.b	$F2,$3F,$87,$CF,$FB,$C6,$DF,$B2,$FC,$37,$FC,$C5,$F4,$0F,$CA,$BF
	.dc.b	$DD,$5F,$9C,$3F,$5D,$FF,$67,$DF,$A3,$BB,$E6,$E7,$FE,$3F,$F9,$2F
	.dc.b	$E7,$B9,$87,$CD,$7F,$1A,$8C,$79,$4E,$6F,$F2,$37,$D1,$FE,$DF,$E4
	.dc.b	$7F,$CF,$D8,$FF,$A0,$BD,$1F,$EC,$3F,$73,$FE,$4B,$FC,$D3,$FC,$F1
	.dc.b	$B7,$F0,$8F,$98,$7E,$8A,$FC,$B7,$F6,$DD,$1F,$DB,$3F,$15,$FF,$06
	.dc.b	$F9,$9F,$B6,$D3,$F8,$D9,$FE,$4D,$DF,$F6,$7E,$2F,$F9,$73,$E4,$BF
	.dc.b	$B7,$F9,$3F,$E9,$8F,$92,$FE,$33,$7F,$E8,$1F,$90,$FE,$30,$FF,$97
	.dc.b	$7F,$60,$7D,$07,$FA,$FB,$F1,$4F,$7F,$E9,$CF,$5D,$F8,$EF,$F0,$7F
	.dc.b	$B0,$FF,$32,$6F,$FE,$83,$CB,$FD,$02,$FF,$A8,$3B,$3F,$C9,$FF,$21
	.dc.b	$FC,$6B,$E4,$3F,$8C,$D3,$8B,$F8,$E7,$C5,$7F,$99,$79,$3F,$DE,$FF
	.dc.b	$11,$FC,$67,$E3,$FF,$8C,$7C,$77,$F1,$7F,$AD,$7C,$7F,$F1,$7F,$BB
	.dc.b	$FF,$3A,$65,$FC,$5B,$3F,$19,$FE,$77,$F5,$3F,$9D,$C3,$E0,$BF,$AB
	.dc.b	$CF,$F9,$CF,$F4,$AF,$F5,$AF,$17,$C8,$AF,$F3,$48,$8F,$F0,$BF,$8D
	.dc.b	$FA,$6F,$2D,$F7,$9F,$E6,$7F,$BA,$FE,$33,$8B,$E0,$DF,$AB,$7F,$8F
	.dc.b	$BC,$9F,$F6,$9F,$C1,$BF,$36,$7E,$74,$FD,$C5,$F9,$4B,$93,$F8,$C7
	.dc.b	$57,$F8,$FB,$7F,$F0,$51,$FA,$AD,$7F,$CB,$DE,$EF,$F9,$68,$FF,$2D
	.dc.b	$FE,$83,$FE,$5F,$F8,$FF,$F2,$E7,$37,$F1,$59,$0F,$D3,$FF,$81,$DF
	.dc.b	$FB,$53,$E0,$F9,$4E,$AF,$FC,$7B,$FA,$87,$EC,$FE,$2F,$FC,$B9,$F8
	.dc.b	$0F,$F9,$6F,$0F,$F5,$B6,$5F,$EA,$3D,$3D,$FC,$7F,$67,$AF,$FD,$4D
	.dc.b	$F0,$5F,$B0,$F8,$1F,$D7,$FC,$03,$E6,$7F,$CC,$1F,$11,$FE,$FB,$C7
	.dc.b	$D9,$9F,$F2,$FE,$0F,$1F,$C6,$7A,$A3,$F8,$C4,$7B,$F4,$79,$1E,$A9
	.dc.b	$FF,$4A,$4F,$03,$CA,$FF,$BC,$3E,$8F,$F7,$38,$7E,$DB,$E6,$FE,$D7
	.dc.b	$85,$F6,$78,$7F,$8F,$78,$3E,$AC,$3F,$86,$79,$DF,$E5,$27,$F8,$0E
	.dc.b	$3F,$D4,$4E,$78,$BC,$1F,$F4,$FF,$C1,$BF,$68,$7C,$A3,$F2,$87,$E8
	.dc.b	$1F,$E2,$DD,$BF,$C5,$BD,$BF,$53,$2F,$EA,$5F,$73,$F2,$F7,$FF,$16
	.dc.b	$EB,$DB,$83,$F9,$45,$7F,$11,$3F,$BC,$1F,$E5,$3F,$94,$7C,$AF,$FB
	.dc.b	$A7,$E0,$3E,$C7,$C8,$FF,$29,$FB,$AF,$F4,$27,$2B,$D2,$FB,$1E,$77
	.dc.b	$F5,$1F,$5F,$EB,$FE,$3B,$FE,$72,$F0,$BE,$BF,$83,$E6,$EF,$FE,$47
	.dc.b	$D8,$1D,$7F,$C9,$3E,$DF,$FD,$1B,$1F,$C9,$F8,$BF,$CC,$7F,$8D,$79
	.dc.b	$3B,$BF,$C4,$7F,$67,$FC,$33,$EA,$9F,$A8,$BF,$21,$7E,$2B,$FC,$37
	.dc.b	$C3,$FF,$09,$7E,$EF,$FB,$AD,$3F,$8E,$1F,$EE,$93,$FD,$E5,$F1,$7F
	.dc.b	$C6,$1F,$27,$EA,$2F,$DA,$5F,$AB,$3F,$5E,$FE,$69,$D6,$F7,$FF,$C9
	.dc.b	$BE,$CB,$E3,$3F,$84,$FD,$BF,$FB,$B3,$3E,$C3,$F1,$EF,$E0,$AF,$F9
	.dc.b	$17,$D6,$FE,$0A,$3F,$09,$FF,$4D,$FE,$8D,$FC,$4A,$3F,$8E,$4F,$F5
	.dc.b	$5E,$8F,$F2,$1C,$7F,$1F,$F0,$7F,$AC,$7D,$97,$E2,$6B,$E6,$0F,$C1
	.dc.b	$DE,$33,$F7,$A7,$D1,$7F,$50,$FA,$1F,$F7,$97,$E5,$DC,$70,$FF,$1B
	.dc.b	$7C,$CF,$2C,$BF,$CA,$3C,$6F,$CB,$7C,$9B,$7C,$AB,$FF,$1E,$2F,$D7
	.dc.b	$3F,$C7,$1E,$97,$9A,$3F,$AB,$CF,$38,$FF,$2D,$F9,$DF,$CD,$EB,$E0
	.dc.b	$FD,$27,$FA,$02,$3F,$9D,$78,$DF,$DC,$FE,$FF,$1A,$FF,$75,$FC,$49
	.dc.b	$CB,$FE,$86,$FB,$A7,$EB,$3F,$2B,$F2,$07,$16,$70,$E1,$C1,$ED,$E0
	.dc.b	$FE,$8C,$F9,$1F,$8F,$D3,$FE,$6E,$FC,$13,$F9,$4F,$E4,$BF,$A2,$FC
	.dc.b	$3F,$FC,$D7,$ED,$BF,$71,$E5,$8F,$F9,$DB,$1F,$BE,$7F,$97,$FE,$9B
	.dc.b	$EC,$70,$7C,$58,$EF,$CD,$F7,$BF,$1F,$FF,$4E,$DF,$D9,$9F,$F0,$87
	.dc.b	$47,$C4,$8F,$94,$7F,$48,$FC,$13,$E2,$2E,$7F,$F0,$B7,$D0,$FC,$4B
	.dc.b	$FF,$33,$E5,$F0,$F9,$7F,$BB,$72,$7C,$19,$FE,$F3,$D9,$F8,$A7,$FC
	.dc.b	$ED,$E5,$3F,$05,$7E,$32,$FA,$BF,$F5,$07,$E1,$3F,$97,$C3,$F5,$78
	.dc.b	$7E,$60,$FE,$D9,$ED,$7E,$AD,$F4,$9F,$E2,$9F,$CA,$7E,$02,$FC,$73
	.dc.b	$F5,$DD,$9F,$E9,$AC,$9F,$9E,$FF,$A4,$7F,$0E,$7E,$BD,$F3,$7E,$BD
	.dc.b	$F5,$9F,$E8,$9F,$C9,$BE,$6E,$7D,$8F,$F5,$1F,$EC,$9D,$97,$EA,$5E
	.dc.b	$0F,$CA,$3F,$61,$1F,$C3,$7F,$21,$AB,$8E,$AF,$07,$CF,$D1,$FE,$B3
	.dc.b	$8F,$F2,$17,$2F,$E7,$DF,$CE,$F6,$FF,$7A,$ED,$F8,$51,$FE,$35,$E6
	.dc.b	$EF,$3F,$EB,$EF,$63,$F2,$3E,$07,$D6,$8F,$F6,$0B,$FE,$4D,$FC,$CF
	.dc.b	$FC,$12,$FF,$AF,$3D,$BF,$CB,$F9,$9F,$A8,$C7,$F7,$3C,$3F,$C9,$76
	.dc.b	$FC,$57,$DF,$EB,$2F,$E0,$F8,$FF,$AA,$78,$3C,$31,$FC,$2B,$4F,$2E
	.dc.b	$3F,$EE,$18,$F0,$BF,$C1,$33,$F5,$C7,$F0,$9B,$FF,$66,$E5,$FA,$8A
	.dc.b	$EE,$79,$8F,$19,$F3,$31,$FE,$51,$C1,$FE,$F6,$FC,$BF,$FC,$6D,$3F
	.dc.b	$CE,$FE,$03,$38,$7F,$88,$71,$FF,$10,$FE,$53,$FE,$6D,$DF,$FC,$52
	.dc.b	$FF,$9B,$F0,$FF,$60,$E1,$C4,$FF,$39,$1E,$33,$C2,$7F,$B1,$B2,$F0
	.dc.b	$BF,$04,$FF,$64,$DF,$D1,$78,$FE,$31,$FE,$36,$7E,$71,$FE,$F2,$F1
	.dc.b	$BF,$CF,$DA,$F5,$8F,$E6,$B7,$F1,$2F,$FC,$95,$A7,$E0,$67,$FC,$7F
	.dc.b	$C9,$FE,$63,$E4,$F3,$D7,$DF,$DF,$CF,$E3,$F7,$BF,$F7,$BF,$2D,$7F
	.dc.b	$CE,$D2,$E0,$FF,$6E,$47,$D0,$3F,$24,$7F,$46,$19,$FD,$FC,$3F,$C9
	.dc.b	$31,$FE,$95,$E8,$F8,$EB,$2A,$F2,$39,$61,$F9,$B7,$A7,$2F,$08,$FE
	.dc.b	$C1,$87,$EC,$FB,$FF,$65,$DD,$FB,$1E,$DF,$EC,$1D,$BF,$B0,$C7,$FD
	.dc.b	$EB,$DB,$E0,$E3,$FA,$FD,$BF,$DA,$79,$3F,$86,$7E,$25,$FD,$97,$C6
	.dc.b	$FE,$17,$E3,$7F,$0A,$F8,$0F,$ED,$5F,$11,$FC,$1F,$3F,$EA,$A7,$F6
	.dc.b	$DF,$23,$0F,$C1,$E4,$FC,$1F,$91,$E6,$8F,$E5,$67,$FA,$D7,$8F,$F9
	.dc.b	$B7,$D6,$BD,$2F,$E6,$DF,$18,$F4,$BF,$9B,$57,$F6,$6F,$57,$FB,$3F
	.dc.b	$8F,$F9,$5F,$23,$9E,$5F,$DD,$3B,$BF,$A9,$1F,$E7,$6F,$47,$FD,$84
	.dc.b	$FF,$35,$8F,$EF,$3E,$CE,$A3,$FA,$4F,$A5,$FE,$C7,$8F,$EA,$F3,$EC
	.dc.b	$F1,$31,$97,$E7,$FC,$6F,$D0,$7A,$5F,$CD,$2B,$E0,$F8,$BF,$D2,$5F
	.dc.b	$F0,$12,$F8,$2F,$ED,$B1,$EF,$3C,$E7,$F8,$9F,$C4,$FC,$2B,$FD,$FB
	.dc.b	$E0,$BF,$BE,$FC,$0F,$F7,$CF,$C0,$DF,$03,$FD,$F3,$93,$E1,$5F,$13
	.dc.b	$FC,$0B,$E0,$FE,$77,$8B,$FB,$1C,$7E,$45,$F7,$1F,$05,$FE,$FB,$B7
	.dc.b	$F8,$39,$FE,$AD,$EC,$7F,$78,$7F,$AC,$79,$BF,$DD,$DF,$EF,$BC,$3F
	.dc.b	$DD,$EB,$E0,$DF,$FA,$8B,$DC,$FD,$85,$7E,$8E,$BF,$BA,$E5,$F8,$4F
	.dc.b	$88,$FF,$18,$7C,$47,$F8,$BF,$D4,$7C,$47,$F8,$BF,$E0,$BF,$C6,$9F
	.dc.b	$11,$F9,$5C,$BF,$C7,$5A,$7F,$AD,$3C,$6F,$A8,$FF,$8B,$07,$9D,$FC
	.dc.b	$3B,$C8,$FE,$1B,$D5,$9F,$88,$FF,$03,$EF,$FE,$4C,$FB,$C7,$8B,$FE
	.dc.b	$4C,$F8,$DF,$F2,$5F,$DE,$7F,$9C,$FA,$5F,$1B,$F9,$8D,$FF,$8B,$F8
	.dc.b	$7F,$EC,$4F,$13,$F6,$BE,$67,$F1,$CF,$8D,$EF,$C5,$FD,$D3,$81,$F3
	.dc.b	$3F,$E6,$CF,$AA,$FF,$51,$47,$F9,$AF,$2F,$F3,$27,$A1,$FE,$89,$7F
	.dc.b	$0F,$F8,$2F,$FA,$96,$3F,$D2,$3B,$7F,$8B,$3E,$AF,$FD,$49,$F5,$3F
	.dc.b	$C2,$3E,$AF,$FD,$49,$CD,$FA,$EB,$0F,$A3,$7C,$D7,$FA,$07,$F0,$6F
	.dc.b	$E5,$FF,$37,$FE,$71,$AF,$DA,$EB,$FE,$74,$7F,$CE,$3F,$43,$FC,$67
	.dc.b	$C3,$FF,$26,$FD,$17,$FA,$63,$85,$F5,$3F,$EA,$2D,$FF,$D8,$BF,$6C
	.dc.b	$BF,$F3,$F5,$7B,$72,$FF,$6E,$7D,$1F,$89,$FE,$89,$C9,$ED,$D3,$FD
	.dc.b	$29,$9F,$FA,$57,$87,$F8,$3B,$FE,$A6,$7F,$D4,$BF,$3B,$FE,$A4,$FA
	.dc.b	$9F,$F5,$1F,$E3,$7F,$EE,$A9,$F8,$17,$EF,$FA,$CF,$76,$5F,$E7,$BF
	.dc.b	$A9,$FF,$52,$4F,$FA,$84,$FF,$51,$47,$67,$FD,$7A,$FF,$A9,$9F,$93
	.dc.b	$F0,$BF,$2A,$31,$1F,$C3,$F8,$FF,$9E,$78,$3E,$98,$E8,$7F,$3F,$FB
	.dc.b	$0F,$BD,$8F,$F9,$CB,$9F,$E1,$F1,$FF,$AA,$3A,$3F,$AF,$F6,$7F,$BD
	.dc.b	$70,$E2,$7F,$A4,$7B,$FF,$D5,$7F,$83,$BE,$A3,$FD,$57,$5F,$EF,$BF
	.dc.b	$B0,$FF,$70,$5B,$2F,$EE,$5F,$9D,$F8,$E3,$DD,$C1,$FC,$C3,$5F,$F7
	.dc.b	$A9,$FE,$FA,$FA,$5F,$BE,$E3,$B7,$EA,$4F,$FB,$9A,$3F,$DF,$E7,$E2
	.dc.b	$3E,$C5,$F5,$DF,$86,$FB,$3E,$13,$F0,$AB,$FD,$79,$5F,$85,$F9,$BF
	.dc.b	$C2,$4B,$C1,$FC,$25,$7E,$2C,$FF,$61,$61,$F9,$F7,$FD,$33,$E0,$BF
	.dc.b	$15,$F5,$78,$1F,$ED,$7E,$2F,$F2,$D6,$7F,$83,$BF,$F4,$FE,$FF,$9C
	.dc.b	$FB,$A7,$DD,$3E,$E9,$F2,$0F,$BA,$7E,$27,$F3,$71,$FD,$7E,$2E,$8F
	.dc.b	$F7,$0B,$FA,$8E,$8F,$97,$3F,$B4,$FB,$6F,$C2,$BF,$EA,$EE,$4F,$96
	.dc.b	$3E,$EF,$E5,$BF,$20,$1F,$B1,$EE,$F9,$0E,$47,$DD,$FC,$CA,$FE,$CB
	.dc.b	$D3,$F5,$BC,$EE,$3C,$EF,$0B,$FD,$7B,$F3,$9F,$E6,$DF,$CD,$3F,$D0
	.dc.b	$98,$3E,$49,$F2,$BF,$8E,$8F,$F7,$77,$8B,$9F,$8A,$FC,$7B,$FE,$C6
	.dc.b	$EB,$7E,$30,$FC,$0D,$F8,$C3,$F0,$47,$9C,$FC,$11,$F8,$27,$FB,$97
	.dc.b	$BB,$F9,$96,$BF,$C8,$B4,$FE,$61,$F8,$27,$F8,$F3,$F0,$4F,$E4,$31
	.dc.b	$FE,$B0,$0F,$F9,$83,$93,$F9,$55,$7F,$89,$7D,$FF,$F6,$CF,$C8,$BE
	.dc.b	$05,$F5,$9E,$CA,$7D,$C7,$FA,$C7,$AB,$FE,$00,$47,$F9,$80,$FC,$4F
	.dc.b	$EB,$7D,$3D,$AF,$F0,$7F,$E1,$0F,$8C,$7E,$6F,$FE,$F7,$F9,$B7,$D9
	.dc.b	$BE,$35,$F8,$4F,$FC,$01,$FC,$27,$CB,$EA,$3F,$43,$F3,$ED,$FE,$37
	.dc.b	$3F,$36,$7F,$97,$F9,$C7,$F2,$1F,$A2,$FF,$20,$4F,$FB,$D3,$E9,$BF
	.dc.b	$CB,$B5,$FC,$93,$E0,$F0,$E3,$8F,$9F,$D0,$FD,$BF,$9B,$FF,$74,$FB
	.dc.b	$FF,$E8,$6F,$C4,$3F,$DF,$BF,$88,$72,$C3,$FC,$99,$87,$27,$FC,$D3
	.dc.b	$CF,$FE,$2E,$FC,$41,$F8,$83,$F1,$0F,$B7,$F8,$83,$F2,$8F,$F6,$07
	.dc.b	$CC,$7B,$F1,$1F,$CE,$7E,$E3,$FC,$7D,$F4,$9D,$63,$FC,$9E,$7F,$A5
	.dc.b	$4F,$F2,$E7,$DE,$7F,$0F,$F8,$0F,$E9,$FF,$41,$FC,$0E,$5F,$89,$3F
	.dc.b	$12,$7E,$20,$CF,$FB,$0F,$47,$82,$DB,$3D,$EF,$E2,$71,$9E,$3F,$F2
	.dc.b	$4F,$CF,$FF,$60,$5F,$3F,$FD,$82,$7F,$B3,$7C,$3F,$F2,$4F,$87,$FF
	.dc.b	$45,$F0,$7F,$2C,$1F,$9C,$FF,$5E,$FA,$0C,$7E,$57,$67,$F9,$8F,$95
	.dc.b	$5F,$ED,$27,$F2,$09,$FE,$85,$8F,$9D,$FE,$7F,$F8,$BF,$F0,$9F,$9B
	.dc.b	$FE,$39,$2F,$81,$FE,$CD,$EB,$7F,$22,$F9,$5F,$E4,$3E,$0F,$F6,$3E
	.dc.b	$7F,$77,$C5,$FF,$85,$3E,$83,$F9,$C6,$1F,$D8,$71,$FD,$B6,$BF,$E0
	.dc.b	$0F,$33,$E3,$6A,$FC,$B1,$F8,$C3,$F1,$AF,$F7,$87,$E3,$3F,$C6,$FB
	.dc.b	$FF,$C8,$F9,$7F,$1E,$E8,$FE,$3B,$9F,$F4,$78,$FE,$C8,$7F,$1D,$7F
	.dc.b	$A8,$BF,$DA,$9F,$EB,$7D,$BF,$0D,$FE,$D1,$E0,$D7,$E2,$FE,$47,$7E
	.dc.b	$6F,$EF,$46,$4F,$B5,$CF,$F8,$3F,$C7,$5F,$A6,$BF,$14,$7E,$3D,$FE
	.dc.b	$16,$FB,$07,$E3,$BF,$C6,$76,$FF,$0C,$FB,$9F,$82,$7F,$D9,$DE,$AF
	.dc.b	$E1,$FE,$21,$F8,$1F,$C1,$E6,$7E,$2C,$FC,$79,$F6,$EF,$C1,$FF,$A0
	.dc.b	$65,$FE,$24,$E4,$FC,$77,$D4,$7F,$81,$A3,$FB,$37,$CF,$7F,$66,$E1
	.dc.b	$FE,$76,$FF,$37,$FC,$62,$BF,$CF,$39,$BE,$CD,$7F,$38,$8F,$E8,$1E
	.dc.b	$27,$E7,$F1,$FE,$7F,$1F,$DA,$AB,$F3,$8F,$E1,$BE,$7B,$FA,$4B,$FE
	.dc.b	$C5,$FB,$3F,$D1,$78,$FF,$D2,$3E,$C7,$F1,$5D,$8F,$C7,$3F,$13,$F6
	.dc.b	$1A,$E9,$F8,$9B,$7F,$B3,$7E,$B3,$FC,$01,$F3,$5F,$89,$8F,$DD,$EB
	.dc.b	$DB,$D6,$FF,$58,$FE,$42,$FC,$CD,$F0,$FF,$B5,$FC,$8B,$FB,$2F,$D0
	.dc.b	$3F,$22,$F8,$FC,$1F,$E0,$0F,$79,$D3,$FE,$2C,$F8,$F7,$D9,$7F,$72
	.dc.b	$FC,$EF,$FB,$2F,$07,$83,$A7,$FB,$2B,$F3,$1F,$EC,$1F,$9E,$FC,$A5
	.dc.b	$D7,$FD,$9B,$B7,$E3,$C5,$D3,$FB,$25,$FF,$84,$E3,$E8,$1F,$D9,$3E
	.dc.b	$1B,$F8,$84,$FF,$0B,$F6,$BF,$C9,$DF,$55,$FD,$97,$F0,$5F,$E3,$7F
	.dc.b	$55,$FD,$26,$7F,$9B,$FD,$4B,$F1,$63,$FC,$A3,$3F,$EC,$9F,$55,$FD
	.dc.b	$8F,$EA,$BF,$B1,$77,$7F,$61,$FA,$8F,$E4,$D5,$FC,$FB,$EA,$FF,$A1
	.dc.b	$3F,$EF,$DF,$81,$FE,$BD,$8B,$E6,$5F,$37,$FC,$8F,$C6,$FE,$71,$B7
	.dc.b	$F7,$BC,$3F,$97,$FD,$37,$F5,$FF,$A5,$FE,$BD,$F4,$BF,$D7,$7F,$21
	.dc.b	$FE,$CB,$63,$FA,$E5,$FF,$27,$FC,$99,$F6,$EF,$C9,$9F,$05,$FE,$1E
	.dc.b	$F9,$A7,$E2,$7F,$E9,$CF,$11,$F9,$AF,$F5,$CF,$9C,$FE,$B7,$9F,$F4
	.dc.b	$BE,$2F,$EA,$5C,$BF,$DD,$CF,$F6,$86,$DF,$D7,$BF,$1E,$F9,$1F,$04
	.dc.b	$F0,$BF,$9C,$78,$1E,$EA,$FF,$41,$E3,$FD,$2B,$E8,$BD,$FF,$5B,$FD
	.dc.b	$26,$3F,$E0,$0E,$CC,$7F,$A0,$7E,$3A,$FC,$A3,$FA,$3F,$8C,$FC,$8D
	.dc.b	$EA,$FF,$2B,$F8,$86,$3E,$AC,$BF,$AB,$F7,$B9,$FF,$B6,$CF,$D1,$9F
	.dc.b	$EA,$C7,$F5,$83,$FA,$74,$7F,$79,$E4,$FE,$67,$C5,$F8,$AF,$D4,$3F
	.dc.b	$9C,$FE,$4F,$FE,$9D,$FA,$2F,$EA,$9F,$24,$FA,$97,$CF,$3F,$2D,$FF
	.dc.b	$1E,$FD,$8B,$F2,$DF,$EF,$9B,$7F,$6A,$F9,$77,$E0,$BF,$B6,$EC,$7C
	.dc.b	$FF,$F9,$D7,$6F,$DB,$59,$FE,$AD,$D1,$FC,$5B,$F2,$DF,$DC,$F9,$EF
	.dc.b	$C6,$DF,$5F,$F8,$3F,$C2,$7F,$9E,$E3,$FD,$5B,$F1,$1F,$EC,$0F,$F3
	.dc.b	$C3,$FD,$83,$F1,$DF,$B4,$7A,$FD,$57,$FA,$FB,$4E,$F1,$FE,$9C,$D1
	.dc.b	$F9,$6F,$F3,$B7,$FA,$FC,$7F,$81,$3E,$A7,$FA,$FF,$BF,$FB,$8E,$1F
	.dc.b	$9D,$3F,$D2,$CF,$F7,$6F,$57,$F7,$BF,$C2,$7F,$CC,$BF,$05,$FD,$93
	.dc.b	$ED,$BF,$C2,$1F,$98,$BF,$17,$F2,$FE,$1B,$F8,$9C,$3F,$C0,$5F,$45
	.dc.b	$F2,$3E,$1F,$FB,$26,$7F,$E7,$6D,$FF,$9C,$7E,$8B,$FE,$83,$F5,$3B
	.dc.b	$70,$7F,$A7,$36,$F8,$7F,$73,$F8,$9F,$C0,$FF,$2D,$F9,$93,$F3,$27
	.dc.b	$E6,$5F,$C6,$BE,$63,$FD,$8F,$D9,$FC,$73,$5F,$F4,$06,$DF,$25,$7E
	.dc.b	$3D,$F2,$7A,$7F,$C1,$5F,$97,$7F,$61,$EE,$FF,$73,$75,$FE,$6F,$0F
	.dc.b	$36,$1F,$D9,$A3,$F4,$D1,$88,$FF,$05,$FD,$BF,$F8,$1B,$EF,$3F,$C4
	.dc.b	$97,$FE,$23,$E3,$FF,$63,$5F,$F8,$2B,$C8,$FE,$9A,$FE,$A3,$E2,$FF
	.dc.b	$C4,$DF,$96,$3F,$41,$FE,$B9,$E9,$BE,$DF,$FA,$37,$E6,$0F,$CC,$FF
	.dc.b	$9C,$FD,$47,$FA,$4B,$BB,$FA,$F6,$7F,$E9,$9D,$7F,$D8,$7F,$9A,$7F
	.dc.b	$30,$5E,$2F,$F6,$5E,$4F,$F2,$EF,$AF,$FD,$59,$FE,$31,$7F,$93,$E5
	.dc.b	$7E,$24,$F2,$7F,$E0,$0F,$BB,$FD,$E7,$9B,$D3,$7F,$EB,$CF,$A6,$FF
	.dc.b	$0E,$E0,$3F,$33,$97,$F5,$59,$FF,$0E,$FD,$DF,$F8,$73,$EC,$BF,$C3
	.dc.b	$73,$FE,$49,$F8,$0F,$F3,$6D,$7F,$95,$AF,$FD,$A7,$BF,$D9,$7F,$C3
	.dc.b	$93,$FD,$52,$75,$F2,$1F,$A1,$AF,$F5,$BF,$DE,$FF,$51,$F2,$7F,$D8
	.dc.b	$1E,$5F,$C1,$1E,$17,$F5,$0C,$3F,$CC,$7F,$47,$FE,$B7,$AF,$F6,$36
	.dc.b	$1F,$E3,$EF,$C9,$BF,$93,$79,$BF,$E8,$03,$FC,$83,$E8,$BF,$2E,$7E
	.dc.b	$25,$FC,$27,$27,$E7,$1F,$E9,$8E,$AF,$F4,$4F,$84,$FC,$C3,$FA,$9D
	.dc.b	$7F,$98,$CF,$F3,$9C,$FF,$B8,$DF,$F5,$DE,$5F,$EA,$BF,$94,$F0,$79
	.dc.b	$DB,$D7,$C7,$3F,$E9,$2E,$57,$E5,$5F,$E6,$DF,$CE,$1F,$9C,$7E,$8B
	.dc.b	$F3,$8F,$E5,$DE,$9B,$D3,$FE,$B1,$87,$F2,$DD,$5F,$31,$FE,$B1,$7C
	.dc.b	$87,$F0,$DC,$FF,$9D,$1F,$09,$F4,$3E,$2F,$73,$7A,$F9,$23,$8F,$B7
	.dc.b	$E6,$FC,$5F,$9F,$17,$F4,$FF,$4B,$FA,$7C,$FF,$A0,$FD,$37,$E4,$FF
	.dc.b	$2B,$EC,$FF,$D2,$BF,$92,$3F,$3B,$F8,$BF,$45,$FD,$37,$7F,$F4,$57
	.dc.b	$0F,$E1,$9D,$7B,$BF,$E6,$5F,$CE,$DF,$98,$74,$FC,$EF,$FC,$45,$7E
	.dc.b	$FF,$A9,$FE,$F7,$87,$F7,$BE,$3F,$B9,$C3,$FE,$BB,$E2,$FE,$3F,$F5
	.dc.b	$3F,$DE,$B9,$FF,$D3,$1F,$15,$FD,$F3,$D4,$FF,$6C,$E3,$FE,$BB,$7F
	.dc.b	$D3,$9F,$96,$7D,$DC,$7F,$05,$E3,$3F,$2C,$FE,$D1,$F7,$FF,$EA,$1F
	.dc.b	$CF,$BF,$B3,$E3,$FE,$04,$3F,$D8,$FE,$2F,$E0,$F5,$FF,$60,$3F,$85
	.dc.b	$FC,$F9,$F9,$EB,$F3,$E7,$E9,$5F,$ED,$2F,$B1,$FF,$73,$FE,$93,$FD
	.dc.b	$A3,$CC,$FF,$69,$7D,$1F,$F6,$6E,$8F,$3E,$BF,$ED,$BD,$7F,$B3,$3F
	.dc.b	$91,$EC,$FF,$6F,$71,$7F,$AD,$7A,$3F,$8D,$CF,$F5,$6F,$C6,$7F,$D3
	.dc.b	$FF,$88,$FF,$9F,$3F,$10,$FF,$3D,$FB,$AF,$95,$FF,$3D,$DF,$FB,$0F
	.dc.b	$C8,$FF,$78,$FB,$9F,$8A,$8F,$EF,$7D,$7F,$E7,$BF,$C3,$FF,$CF,$5F
	.dc.b	$87,$FF,$9E,$7F,$0E,$FF,$3C,$75,$FE,$8F,$3F,$F2,$97,$E1,$BF,$C6
	.dc.b	$B1,$FE,$59,$1F,$C2,$F7,$FF,$48,$2F,$D0,$6F,$E1,$5F,$E9,$7D,$BF
	.dc.b	$D6,$DC,$5E,$AA,$FF,$62,$7E,$09,$FE,$D7,$F8,$6F,$EA,$F3,$FE,$64
	.dc.b	$CB,$FD,$C7,$F1,$5F,$E7,$3F,$C5,$FF,$8C,$7E,$1B,$FE,$BA,$F5,$7F
	.dc.b	$AB,$FE,$45,$FD,$F7,$EA,$3F,$DB,$5A,$7F,$B2,$07,$CA,$7F,$A9,$1F
	.dc.b	$F4,$8E,$9F,$D9,$38,$FF,$7D,$F8,$17,$E6,$FE,$E9,$F9,$0B,$F4,$87
	.dc.b	$E1,$AD,$5F,$A1,$FF,$80,$33,$FF,$3A,$7E,$1B,$FE,$73,$FA,$8F,$F3
	.dc.b	$94,$7F,$AE,$B9,$3F,$DA,$CF,$8C,$FE,$61,$F9,$07,$F5,$97,$F9,$97
	.dc.b	$37,$F5,$99,$FF,$08,$FE,$0B,$FE,$FC,$7E,$6C,$A3,$FD,$BB,$F4,$CF
	.dc.b	$D2,$3F,$C4,$7E,$93,$E4,$DF,$A4,$7F,$51,$C7,$FA,$D7,$A3,$FE,$96
	.dc.b	$F7,$BF,$13,$FA,$43,$F3,$0F,$F3,$2F,$E4,$2F,$79,$FA,$47,$F5,$DF
	.dc.b	$80,$FE,$B3,$F4,$2F,$A5,$FC,$06,$9F,$EA,$8F,$D2,$3F,$AC,$7D,$2F
	.dc.b	$F3,$8F,$0F,$FC,$CD,$F8,$37,$F8,$EF,$A3,$D9,$D1,$FF,$00,$7C,$2F
	.dc.b	$F7,$AF,$C9,$FF,$5B,$FC,$B5,$9C,$8F,$F1,$97,$AF,$FA,$AE,$8F,$F1
	.dc.b	$BF,$E9,$5F,$D6,$FA,$3F,$C7,$5F,$4F,$F7,$B3,$FB,$DF,$37,$FE,$3A
	.dc.b	$C7,$FD,$77,$D5,$FC,$45,$7D,$F7,$F8,$7F,$F4,$A7,$AD,$FE,$5C,$F2
	.dc.b	$7F,$C9,$97,$FC,$5F,$1F,$C3,$F4,$7F,$5B,$E9,$F2,$CF,$F5,$BE,$1F
	.dc.b	$ED,$1F,$96,$7F,$5B,$E1,$FA,$04,$3E,$34,$7F,$49,$D7,$FD,$9A,$FF
	.dc.b	$AC,$63,$5F,$A3,$7F,$AC,$7D,$FF,$F5,$84,$FF,$B6,$67,$9C,$74,$E7
	.dc.b	$FC,$27,$C8,$7F,$4C,$F7,$3E,$6D,$FE,$13,$F2,$BF,$C3,$3F,$E5,$8F
	.dc.b	$BA,$FF,$4B,$78,$5F,$32,$1F,$8C,$FF,$AC,$D7,$E3,$3F,$EB,$39,$D7
	.dc.b	$D2,$BF,$39,$FE,$C3,$5F,$D8,$7D,$7F,$EC,$5E,$B3,$B5,$FA,$5B,$F1
	.dc.b	$CF,$EA,$BF,$05,$FE,$11,$F1,$FF,$8F,$63,$F7,$38,$BF,$AC,$3F,$C3
	.dc.b	$CF,$ED,$2F,$FA,$77,$DF,$FE,$A6,$FF,$7A,$FC,$BB,$FA,$9B,$FD,$EC
	.dc.b	$FF,$4B,$6F,$FE,$55,$AF,$F5,$BF,$E2,$6F,$92,$FF,$5B,$FC,$7F,$F7
	.dc.b	$6F,$8A,$FF,$5B,$CF,$C1,$BF,$05,$FF,$A6,$30,$F8,$9F,$A7,$B2,$FF
	.dc.b	$14,$FE,$52,$FC,$ED,$8B,$F4,$F7,$B1,$FE,$A0,$FC,$2B,$A6,$7F,$CE
	.dc.b	$4F,$F5,$09,$FE,$D2,$CF,$FC,$03,$FA,$7F,$B1,$FF,$5D,$7E,$9F,$FE
	.dc.b	$C9,$F8,$5F,$A9,$7F,$EC,$BF,$CB,$4F,$E8,$B8,$DC,$7F,$E1,$8F,$0F
	.dc.b	$F9,$16,$6F,$C6,$FF,$DB,$3F,$A3,$FF,$6F,$8F,$CA,$FC,$AF,$FA,$EB
	.dc.b	$F1,$BF,$F5,$CF,$E3,$7F,$EB,$8E,$A7,$E3,$5E,$9B,$FF,$5B,$FC,$F7
	.dc.b	$FA,$DB,$E6,$79,$47,$F7,$0F,$D2,$3E,$3B,$E3,$3F,$C4,$39,$7E,$2F
	.dc.b	$93,$FD,$A9,$1F,$B8,$FD,$27,$FD,$E0,$FC,$23,$FA,$B4,$0C,$BF,$E1
	.dc.b	$A7,$EE,$7D,$77,$FB,$A3,$EB,$73,$F4,$5F,$E1,$CD,$FF,$DA,$1F,$1F
	.dc.b	$F9,$EC,$3F,$94,$7E,$10,$FD,$4E,$73,$67,$DC,$FE,$7D,$E3,$FF,$83
	.dc.b	$3E,$BF,$FD,$A5,$D3,$F9,$AD,$FF,$DD,$5E,$BF,$F2,$5F,$CF,$FC,$7E
	.dc.b	$4F,$F8,$AA,$3F,$94,$7D,$1F,$3C,$BF,$A4,$7E,$05,$FE,$B8,$C7,$F8
	.dc.b	$CC,$7F,$99,$CF,$E9,$67,$F9,$A7,$E8,$91,$FA,$88,$FE,$7D,$F7,$FF
	.dc.b	$E9,$EF,$CE,$B3,$D3,$FC,$5E,$79,$7E,$A6,$FD,$25,$EB,$FB,$F8,$3F
	.dc.b	$41,$7F,$B9,$FC,$0B,$F7,$B8,$7F,$5B,$E6,$F1,$7E,$3D,$FB,$EE,$8F
	.dc.b	$F7,$69,$FC,$AB,$DA,$FF,$31,$6F,$FD,$5E,$3F,$96,$FB,$7F,$67,$37
	.dc.b	$E2,$BF,$E0,$BE,$5C,$6B,$FE,$03,$DB,$FC,$C1,$DD,$FE,$4E,$CF,$FC
	.dc.b	$D9,$F0,$DF,$C1,$3F,$22,$7D,$03,$F5,$37,$E4,$4F,$8F,$7E,$A7,$FE
	.dc.b	$A1,$FD,$1B,$FD,$CB,$C1,$F8,$3D,$3F,$8A,$7E,$4B,$FA,$0E,$3F,$F2
	.dc.b	$D7,$1F,$F9,$72,$3F,$C9,$11,$FC,$1A,$BF,$C3,$FF,$A2,$7F,$98,$2F
	.dc.b	$C9,$1A,$9F,$EE,$8F,$FA,$06,$FF,$90,$7D,$C7,$E6,$BD,$5F,$F4,$27
	.dc.b	$D5,$FF,$0C,$AF,$EB,$D6,$CB,$FA,$89,$F9,$68,$FE,$20,$6B,$ED,$7E
	.dc.b	$CC,$FF,$B2,$7F,$34,$FE,$4D,$F8,$16,$23,$FB,$E7,$95,$FC,$B3,$D3
	.dc.b	$FD,$DF,$6F,$F4,$BF,$03,$FC,$25,$FA,$78,$7E,$BB,$FA,$48,$FF,$02
	.dc.b	$7E,$B6,$FD,$6E,$5F,$F4,$0D,$BF,$CC,$35,$F9,$3F,$03,$F8,$8E,$9F
	.dc.b	$8B,$AE,$D3,$F9,$5F,$89,$FF,$64,$7D,$27,$DF,$C3,$F6,$58,$3F,$27
	.dc.b	$7E,$BE,$FD,$73,$D5,$E9,$7F,$A3,$3F,$5F,$7E,$BC,$FC,$2B,$DD,$97
	.dc.b	$F0,$0D,$FF,$8A,$69,$D6,$3F,$7D,$FA,$F0,$7E,$73,$D2,$BF,$D5,$51
	.dc.b	$93,$E4,$70,$7F,$88,$FF,$44,$F3,$CF,$DF,$E0,$FE,$B1,$D7,$FB,$3A
	.dc.b	$FD,$87,$AF,$FE,$E3,$F0,$7F,$55,$1F,$EE,$F3,$FD,$CD,$F8,$D7,$C4
	.dc.b	$C3,$FD,$37,$C3,$FD,$87,$F0,$27,$C2,$FD,$FF,$C2,$FF,$8A,$7E,$8A
	.dc.b	$F1,$3F,$BE,$FE,$75,$C7,$1F,$EC,$07,$FA,$23,$D6,$7B,$6F,$D7,$BF
	.dc.b	$A4,$79,$0F,$93,$7E,$7C,$F5,$9F,$AF,$73,$CB,$FD,$47,$FA,$F5,$7F
	.dc.b	$CE,$1D,$9F,$D7,$79,$BD,$F1,$FC,$C3,$E3,$7F,$B3,$7C,$AE,$63,$F2
	.dc.b	$A7,$F5,$AF,$8E,$FE,$B7,$E4,$7F,$B7,$FA,$3F,$DA,$DB,$FF,$2E,$FC
	.dc.b	$17,$FC,$25,$EF,$BF,$4E,$7E,$82,$FC,$7B,$53,$FE,$17,$FC,$53,$FA
	.dc.b	$7F,$57,$D4,$F0,$3F,$CD,$1E,$7F,$E4,$CC,$3F,$C8,$7E,$3F,$F9,$97
	.dc.b	$E6,$5F,$CF,$3F,$18,$FD,$3C,$8F,$F3,$7F,$03,$BF,$FC,$6D,$F8,$7F
	.dc.b	$FA,$3E,$3F,$95,$CE,$FE,$25,$F8,$97,$FB,$DF,$F4,$FF,$C0,$4F,$F9
	.dc.b	$E3,$E2,$9F,$1C,$FC,$93,$32,$F3,$3F,$9A,$65,$D2,$3F,$9A,$8C,$FF
	.dc.b	$96,$77,$FF,$53,$3F,$DC,$E3,$CE,$FE,$5B,$3F,$CF,$75,$97,$FC,$A7
	.dc.b	$B7,$F8,$39,$FE,$2B,$FA,$4F,$F5,$97,$6F,$F8,$B7,$F2,$07,$82,$F4
	.dc.b	$3F,$7D,$A7,$FA,$27,$E8,$7F,$CB,$B9,$3D,$1F,$F0,$4F,$85,$FE,$B1
	.dc.b	$F2,$3F,$DC,$D9,$FF,$BF,$34,$FF,$74,$E3,$FE,$39,$FB,$3F,$E8,$DF
	.dc.b	$B1,$BF,$3A,$FC,$9F,$B5,$FE,$7E,$FD,$3F,$FC,$5F,$C5,$E6,$F2,$7F
	.dc.b	$C8,$0F,$F9,$63,$BF,$F3,$99,$FF,$8C,$73,$FF,$16,$7D,$37,$C3,$FC
	.dc.b	$03,$C1,$7C,$39,$7F,$23,$5F,$1E,$BE,$47,$EA,$DF,$E0,$4C,$BF,$C9
	.dc.b	$9F,$13,$FC,$AF,$CE,$F9,$98,$FF,$2B,$E6,$FE,$E1,$8F,$FB,$E2,$3F
	.dc.b	$C0,$D1,$EE,$7C,$B3,$FD,$5B,$DE,$FD,$DF,$5F,$F8,$8B,$E2,$BE,$5D
	.dc.b	$FD,$BE,$FF,$F2,$A7,$DA,$BE,$3B,$F3,$3E,$77,$F7,$8F,$00,$7F,$30
	.dc.b	$F4,$BF,$CC,$1F,$2B,$FD,$58,$FC,$86,$FF,$E6,$3E,$7F,$F0,$1D,$BE
	.dc.b	$E7,$FB,$CF,$CB,$FF,$75,$E2,$F9,$27,$E4,$B8,$3F,$C0,$1B,$7E,$50
	.dc.b	$FF,$01,$7C,$EF,$F8,$73,$EF,$BF,$D0,$1F,$3E,$FC,$9D,$FA,$DB,$F4
	.dc.b	$DF,$F0,$2F,$A7,$FE,$73,$F8,$B7,$C6,$7F,$2E,$D7,$BF,$47,$F9,$C3
	.dc.b	$97,$9F,$C2,$7F,$10,$F9,$0F,$CC,$F3,$7F,$8A,$FE,$03,$FC,$09,$7F
	.dc.b	$EE,$9F,$47,$C6,$7E,$C7,$E1,$BF,$C0,$99,$7C,$73,$F0,$73,$FE,$04
	.dc.b	$FC,$23,$FD,$09,$EB,$7F,$97,$BE,$05,$F8,$F7,$F8,$DF,$F0,$E7,$E4
	.dc.b	$ED,$3F,$C3,$BE,$B7,$F4,$BC,$FE,$2F,$B9,$FE,$06,$7A,$EF,$FE,$3F
	.dc.b	$F5,$B8,$47,$FA,$F3,$EB,$DF,$3B,$FE,$52,$FD,$7F,$FD,$13,$83,$F4
	.dc.b	$2F,$F2,$FB,$F9,$8C,$7F,$C0,$D8,$7F,$98,$FD,$CF,$F5,$8F,$91,$FE
	.dc.b	$51,$D7,$EA,$FC,$F7,$F8,$06,$FF,$D0,$DF,$0F,$FE,$9F,$D3,$FC,$0D
	.dc.b	$8F,$92,$7F,$CE,$47,$F3,$F7,$FD,$9F,$F3,$FF,$E5,$EF,$3B,$FC,$0B
	.dc.b	$1B,$3F,$97,$FB,$FF,$E0,$3A,$FF,$49,$47,$FB,$5B,$E0,$3F,$BC,$76
	.dc.b	$7F,$90,$C7,$A3,$FE,$04,$EC,$FE,$E7,$0F,$9F,$FF,$02,$70,$FF,$6A
	.dc.b	$C7,$FD,$39,$F2,$FF,$D9,$FE,$63,$FC,$A7,$9F,$F6,$68,$FE,$D5,$8F
	.dc.b	$EA,$AB,$FB,$CE,$3F,$E5,$37,$FA,$B3,$FC,$81,$FE,$C6,$BE,$63,$F9
	.dc.b	$96,$5D,$E7,$FC,$03,$F3,$7F,$E0,$0F,$9A,$EF,$97,$FB,$17,$C7,$FE
	.dc.b	$FF,$E4,$6E,$FF,$D8,$3C,$1F,$C9,$33,$FE,$79,$FA,$7F,$F8,$67,$E7
	.dc.b	$FF,$C3,$3B,$7E,$46,$7F,$A0,$7A,$5F,$D1,$FE,$43,$FA,$3C,$7F,$43
	.dc.b	$F8,$BF,$EF,$BA,$FF,$8B,$3B,$3F,$BE,$E1,$FC,$DF,$D5,$FF,$36,$71
	.dc.b	$77,$C1,$EA,$E2,$BF,$6D,$87,$FA,$34,$FD,$3F,$BB,$FE,$DD,$FC,$9B
	.dc.b	$FD,$6F,$DB,$F9,$AE,$27,$EB,$5F,$9A,$F8,$DF,$EE,$9E,$D7,$F7,$1C
	.dc.b	$7F,$CD,$E7,$FB,$97,$F4,$77,$EA,$DF,$E0,$6F,$C8,$BF,$BC,$F8,$CF
	.dc.b	$C2,$DE,$F7,$F7,$DF,$D8,$39,$79,$EF,$D3,$7A,$67,$F9,$9D,$FF,$16
	.dc.b	$FF,$43,$3F,$D5,$7F,$1B,$FC,$B2,$3F,$D4,$9E,$67,$F9,$8A,$7F,$DA
	.dc.b	$BF,$11,$FD,$F2,$7F,$CB,$9E,$5F,$F0,$07,$FD,$55,$FB,$13,$F5,$77
	.dc.b	$E5,$BF,$E2,$DC,$3F,$DF,$DF,$9B,$BE,$ED,$FB,$17,$F1,$9F,$BA,$1F
	.dc.b	$EA,$4F,$8F,$FF,$14,$F8,$5F,$6F,$EB,$BB,$43,$E3,$7F,$D4,$7E,$17
	.dc.b	$F6,$5C,$FE,$87,$06,$67,$FA,$D4,$FF,$17,$71,$F1,$8F,$CC,$BF,$91
	.dc.b	$7C,$67,$F9,$97,$E2,$BF,$C2,$7F,$35,$FB,$FF,$D2,$FF,$96,$79,$5F
	.dc.b	$DE,$35,$FF,$1D,$F9,$FF,$CB,$B4,$FF,$30,$74,$7F,$B2,$7E,$5B,$FB
	.dc.b	$C7,$8F,$F8,$5F,$33,$FB,$AF,$99,$FE,$37,$F9,$DF,$EE,$7C,$81,$FA
	.dc.b	$BF,$8F,$F5,$CF,$F0,$4E,$4F,$D2,$1C,$3F,$ED,$5D,$5F,$87,$7F,$B8
	.dc.b	$9F,$C7,$E3,$FE,$00,$CB,$FA,$1F,$8D,$FD,$66,$3F,$CE,$2F,$FB,$75
	.dc.b	$F0,$6B,$FE,$13,$FD,$79,$F8,$63,$F2,$C7,$EB,$0F,$CF,$BB,$E3,$FC
	.dc.b	$C3,$1F,$F7,$47,$91,$DB,$E7,$3D,$91,$FE,$46,$E1,$FF,$0C,$67,$FD
	.dc.b	$DE,$3F,$BE,$F9,$7F,$DE,$1F,$E3,$B1,$FD,$D9,$FC,$59,$F0,$EF,$FA
	.dc.b	$6D,$7F,$87,$CF,$EE,$FE,$0A,$DF,$B4,$FF,$B1,$78,$3F,$B0,$E1,$9F
	.dc.b	$CF,$FF,$E0,$07,$4E,$23,$F8,$31,$F3,$C7,$D0,$FF,$09,$F3,$3F,$2B
	.dc.b	$F9,$13,$F0,$7F,$E1,$F5,$F9,$5F,$C3,$9F,$85,$BF,$0D,$FF,$3F,$7C
	.dc.b	$07,$FB,$F7,$EC,$BF,$DF,$BD,$9F,$E4,$6F,$3B,$FD,$F9,$E7,$FF,$BF
	.dc.b	$37,$FF,$48,$79,$DF,$93,$FC,$2F,$AB,$F9,$2B,$FF,$21,$7C,$27,$E2
	.dc.b	$FC,$AF,$F7,$86,$3F,$CD,$FC,$2F,$E3,$DF,$80,$3F,$00,$FF,$52,$7E
	.dc.b	$49,$E6,$FB,$0F,$F5,$07,$C7,$7F,$A8,$B3,$FC,$9E,$DF,$E2,$CF,$C2
	.dc.b	$FF,$92,$71,$7F,$41,$FA,$5F,$ED,$FF,$5B,$FE,$30,$FC,$2B,$FC,$9D
	.dc.b	$E1,$FF,$72,$F8,$9F,$F4,$37,$E7,$2F,$87,$7E,$05,$FC,$37,$E7,$1F
	.dc.b	$B1,$37,$FE,$97,$E0,$7E,$6F,$2F,$F3,$FF,$85,$FE,$BB,$E1,$FE,$E3
	.dc.b	$3F,$47,$6F,$05,$7F,$BD,$BE,$47,$EA,$7B,$4F,$D5,$BF,$CD,$3E,$EB
	.dc.b	$F1,$CF,$F7,$07,$89,$FE,$A4,$FA,$2F,$E3,$19,$7F,$B8,$B2,$FF,$7A
	.dc.b	$7A,$7F,$EF,$ED,$3C,$5A,$7F,$A3,$A7,$FB,$81,$FE,$13,$8F,$EE,$07
	.dc.b	$F8,$0C,$FD,$97,$37,$F6,$87,$F0,$DE,$EF,$FB,$D5,$FF,$7A,$9F,$8B
	.dc.b	$E5,$FE,$F2,$FE,$42,$38,$3F,$E0,$CF,$3B,$F9,$37,$A3,$F9,$7F,$43
	.dc.b	$F2,$18,$87,$EB,$9F,$98,$F4,$9E,$97,$E6,$3E,$67,$FC,$83,$97,$CB
	.dc.b	$BF,$F5,$43,$FD,$EF,$A7,$FA,$D7,$81,$FE,$51,$DD,$EA,$7F,$61,$E9
	.dc.b	$FF,$28,$FE,$15,$FE,$FC,$C7,$FD,$A3,$F3,$3F,$CA,$71,$FE,$FF,$F8
	.dc.b	$57,$F6,$17,$F0,$FF,$09,$FC,$6F,$BB,$FD,$A8,$FF,$81,$E3,$FC,$0B
	.dc.b	$F6,$1F,$D7,$67,$F5,$5F,$0D,$FC,$AB,$ED,$3F,$C9,$3F,$12,$C7,$F9
	.dc.b	$57,$A9,$FC,$C7,$EA,$1F,$B2,$FF,$1A,$7F,$67,$C9,$EB,$F5,$7F,$DE
	.dc.b	$9F,$49,$FE,$EF,$FA,$5F,$E8,$D3,$FE,$F1,$8F,$F0,$6E,$BF,$EE,$AF
	.dc.b	$A5,$FE,$DD,$7F,$8D,$AF,$EE,$AF,$FB,$BB,$D9,$FE,$E5,$E1,$FE,$FF
	.dc.b	$8F,$FD,$4C,$7F,$75,$FC,$7B,$FB,$67,$23,$F2,$FF,$E0,$9E,$1B,$F4
	.dc.b	$3F,$CE,$74,$BB,$5F,$0C,$FB,$D7,$EC,$BF,$A1,$C7,$FB,$6F,$E0,$BF
	.dc.b	$DD,$FC,$DF,$F4,$DF,$B1,$FE,$F2,$FC,$35,$F9,$07,$F6,$DF,$CC,$7F
	.dc.b	$96,$FA,$1F,$9F,$E7,$FF,$03,$1F,$EA,$27,$FC,$9C,$E0,$FF,$51,$6F
	.dc.b	$F9,$97,$FC,$78,$7F,$1D,$3F,$D7,$67,$F7,$7E,$5F,$F7,$6B,$FC,$87
	.dc.b	$E2,$F4,$FF,$80,$0F,$6F,$1F,$FA,$23,$CF,$FC,$B4,$7E,$3F,$F1,$DF
	.dc.b	$CA,$79,$BF,$CA,$AF,$F2,$73,$FE,$0C,$F0,$BF,$B6,$7E,$19,$FE,$F6
	.dc.b	$FC,$2D,$FA,$67,$FA,$36,$7F,$B8,$47,$FB,$DA,$7F,$C5,$BF,$5D,$F8
	.dc.b	$FF,$0B,$EA,$77,$FF,$BD,$7E,$33,$FD,$EB,$DF,$FD,$57,$3F,$9F,$C5
	.dc.b	$FE,$E1,$AF,$C8,$47,$F6,$77,$FC,$B5,$F1,$EF,$D1,$3F,$DE,$6F,$F8
	.dc.b	$EB,$F0,$B7,$E3,$6E,$C7,$E8,$7F,$EF,$6F,$1B,$F3,$9E,$3F,$F2,$EC
	.dc.b	$BF,$98,$67,$FA,$FD,$3F,$DD,$E7,$FB,$BF,$5F,$D7,$CF,$F7,$A9,$FF
	.dc.b	$1E,$7B,$1F,$E9,$6F,$6F,$FA,$67,$E2,$EF,$D5,$3F,$C4,$1F,$2E,$FC
	.dc.b	$67,$FC,$69,$F8,$5B,$F6,$7F,$C4,$65,$FE,$01,$F0,$7F,$DE,$EF,$FA
	.dc.b	$83,$CF,$FC,$9B,$FE,$05,$F0,$BF,$BD,$F0,$62,$3F,$C1,$1B,$7E,$1E
	.dc.b	$3F,$AE,$7E,$5E,$FD,$55,$FA,$E3,$D5,$7D,$6B,$F6,$77,$EB,$8F,$CD
	.dc.b	$7F,$DE,$7B,$FF,$84,$FC,$8F,$F7,$F7,$AB,$DF,$7F,$EE,$07,$E4,$B6
	.dc.b	$7D,$5F,$F4,$8F,$5B,$FB,$E4,$FE,$48,$FE,$49,$3F,$E2,$33,$FB,$F3
	.dc.b	$FD,$DA,$FF,$DF,$FF,$0B,$FD,$D7,$1F,$F2,$64,$FF,$80,$3A,$FF,$D3
	.dc.b	$3C,$9F,$94,$8F,$F7,$B9,$FE,$DE,$E4,$FE,$4E,$FF,$3E,$3F,$95,$7C
	.dc.b	$97,$F2,$EB,$FF,$00,$77,$7E,$2F,$ED,$BF,$C8,$5E,$6B,$F6,$97,$ED
	.dc.b	$8F,$D8,$9F,$3F,$F9,$9A,$FF,$26,$43,$F2,$E7,$E9,$1C,$72,$FC,$B7
	.dc.b	$C4,$FF,$92,$BC,$2F,$F7,$D8,$FC,$1F,$FB,$27,$A3,$FE,$FC,$FC,$2B
	.dc.b	$FD,$F9,$C5,$FE,$D0,$3F,$C8,$1D,$7F,$DC,$35,$FF,$2B,$7E,$2F,$FE
	.dc.b	$0C,$F2,$FF,$27,$5F,$EB,$7F,$90,$FF,$01,$6B,$FE,$5D,$8F,$F5,$D7
	.dc.b	$C8,$FE,$83,$D5,$7E,$2B,$FD,$CF,$D7,$FF,$04,$FA,$8F,$5B,$F4,$6F
	.dc.b	$F3,$4D,$5F,$6F,$C2,$FF,$DF,$19,$FF,$96,$BD,$FF,$F0,$43,$FE,$A7
	.dc.b	$FA,$7F,$F0,$42,$EE,$FE,$F5,$B7,$F8,$AE,$3F,$C1,$26,$5F,$F2,$E7
	.dc.b	$D9,$FF,$A4,$FF,$6A,$FF,$05,$7A,$8F,$D3,$9F,$4D,$FE,$96,$C3,$F9
	.dc.b	$96,$1F,$E0,$9C,$7E,$CF,$AB,$FD,$A3,$CB,$FC,$0C,$FF,$99,$F8,$7F
	.dc.b	$BC,$3F,$CD,$A3,$FB,$64,$7F,$8F,$61,$F8,$5F,$FA,$17,$ED,$DF,$8D
	.dc.b	$FF,$AD,$BB,$DF,$A4,$BA,$BF,$D5,$B8,$7F,$6F,$FA,$1F,$F3,$A7,$D0
	.dc.b	$FF,$77,$BF,$F7,$DC,$FE,$3C,$CE,$9F,$E0,$C3,$FD,$EC,$FF,$9E,$FF
	.dc.b	$10,$FF,$7D,$BF,$E7,$DE,$0F,$F0,$2A,$F4,$3F,$D5,$EB,$F0,$BF,$E7
	.dc.b	$5E,$7F,$F8,$32,$3D,$47,$F8,$D3,$0F,$CD,$9F,$CF,$3D,$1F,$D2,$3F
	.dc.b	$E0,$B7,$F9,$BC,$7F,$88,$3E,$87,$FC,$C7,$8B,$E0,$FF,$97,$F8,$BE
	.dc.b	$FD,$7F,$E0,$0F,$CB,$7F,$82,$2B,$58,$7F,$2D,$AF,$F7,$2C,$B3,$7D
	.dc.b	$0F,$E7,$6F,$FD,$FD,$A7,$F7,$5F,$7B,$FC,$CF,$F3,$5F,$DD,$E7,$F2
	.dc.b	$D3,$FE,$A2,$C7,$FD,$4C,$FF,$83,$11,$23,$FA,$03,$83,$3F,$35,$FD
	.dc.b	$FB,$3F,$E6,$9A,$FF,$98,$7F,$0F,$FD,$BF,$57,$F8,$1B,$F0,$F7,$D7
	.dc.b	$BE,$C1,$1F,$E5,$5F,$1F,$E0,$FE,$1E,$FA,$6F,$F1,$14,$7C,$4F,$41
	.dc.b	$F2,$5F,$D5,$75,$F8,$5B,$F9,$BE,$3F,$C5,$1F,$E6,$6B,$FC,$77,$4F
	.dc.b	$ED,$EB,$D7,$F6,$9D,$EB,$8F,$9B,$EB,$94,$FF,$99,$FF,$64,$FF,$08
	.dc.b	$FE,$35,$FE,$10,$FA,$3F,$F0,$7F,$6F,$F8,$3B,$E1,$FF,$C1,$BF,$38
	.dc.b	$FC,$5B,$FD,$F3,$D9,$FE,$99,$FB,$7F,$CD,$FE,$C3,$FE,$C8,$F8,$A7
	.dc.b	$8A,$FC,$C5,$F5,$0C,$9D,$FF,$CE,$F9,$3F,$C1,$DF,$9D,$7F,$80,$DF
	.dc.b	$EC,$1F,$86,$FB,$1F,$D5,$79,$7F,$96,$CB,$FD,$6B,$1F,$92,$3F,$C1
	.dc.b	$1F,$1D,$FE,$B2,$7F,$CA,$DD,$3F,$92,$7F,$CB,$DD,$5F,$EA,$E7,$F9
	.dc.b	$C6,$DF,$E2,$EA,$E5,$1F,$17,$A7,$F2,$7E,$53,$F3,$0F,$F1,$8C,$FD
	.dc.b	$2E,$BF,$C9,$79,$7F,$CB,$7F,$17,$FF,$58,$FD,$F8,$9F,$96,$FF,$A4
	.dc.b	$FC,$AF,$F5,$4F,$8B,$FE,$5F,$7F,$85,$75,$FE,$8F,$E0,$BF,$D6,$1F
	.dc.b	$09,$FD,$AF,$93,$FA,$2F,$E3,$7C,$76,$FB,$3F,$0D,$FE,$0C,$FC,$9B
	.dc.b	$11,$F6,$F8,$3F,$95,$FC,$3B,$F2,$0F,$C1,$72,$FE,$6F,$6F,$C6,$3F
	.dc.b	$CE,$B6,$FF,$40,$1F,$DA,$0B,$AF,$E1,$F1,$FF,$5F,$4F,$F8,$35,$FF
	.dc.b	$01,$FC,$0F,$F8,$35,$0F,$D0,$1F,$B0,$BF,$60,$7D,$1B,$F1,$D7,$E0
	.dc.b	$9F,$CB,$FF,$0C,$FF,$5B,$79,$FF,$47,$2F,$E0,$BE,$87,$F7,$DD,$3F
	.dc.b	$C1,$51,$FE,$13,$3F,$C0,$1F,$0F,$FC,$AF,$85,$F0,$FD,$FE,$B5,$83
	.dc.b	$B7,$FC,$5F,$F9,$4F,$F8,$2B,$1F,$F4,$99,$FC,$6B,$CC,$FF,$34,$BD
	.dc.b	$4F,$F2,$E1,$FE,$2F,$E0,$FF,$05,$3F,$CE,$7C,$58,$FF,$BE,$7C,$9F
	.dc.b	$93,$3F,$ED,$6F,$C9,$FF,$27,$3D,$8F,$F2,$F9,$FE,$27,$3F,$90,$F4
	.dc.b	$7F,$29,$7F,$4B,$6F,$DA,$BF,$D6,$7F,$09,$FC,$BA,$7F,$C5,$7F,$B6
	.dc.b	$FF,$4F,$8F,$F5,$A7,$E8,$9F,$EB,$8F,$8A,$D5,$30,$7E,$4D,$FD,$57
	.dc.b	$D1,$EF,$3F,$EE,$1F,$CD,$FF,$C3,$DF,$0F,$FE,$5B,$E9,$FE,$55,$E1
	.dc.b	$7F,$3B,$3F,$3B,$4F,$A4,$7E,$A3,$FE,$0F,$FA,$0F,$F0,$B7,$D2,$7F
	.dc.b	$85,$7E,$93,$74,$FC,$93,$3F,$4B,$BF,$A6,$E7,$F4,$AF,$D9,$BF,$C2
	.dc.b	$7F,$4D,$FE,$14,$E7,$FF,$7B,$7E,$79,$FE,$13,$FA,$61,$07,$47,$F5
	.dc.b	$7E,$2F,$C7,$F1,$7F,$A8,$3F,$25,$FF,$14,$FB,$01,$F4,$DF,$F0,$07
	.dc.b	$D4,$FE,$9B,$E9,$FF,$39,$FB,$2F,$EC,$1E,$67,$FA,$C7,$F3,$DF,$DF
	.dc.b	$4B,$E4,$BF,$56,$FE,$30,$FE,$05,$F2,$1F,$E1,$9D,$3F,$3F,$AF,$F8
	.dc.b	$1F,$E3,$3F,$8C,$F5,$7F,$41,$FC,$53,$E4,$F7,$BE,$B7,$F9,$FF,$CC
	.dc.b	$7F,$41,$F9,$7F,$E8,$7F,$31,$CF,$37,$67,$F2,$5F,$A9,$FF,$10,$1F
	.dc.b	$E2,$0E,$2F,$BD,$3F,$DC,$3E,$97,$F6,$9C,$2F,$C2,$BD,$3F,$81,$BC
	.dc.b	$8F,$F0,$F5,$FC,$8F,$A3,$FB,$9C,$99,$E1,$F9,$3F,$9B,$7F,$87,$FF
	.dc.b	$68,$57,$FB,$0B,$D7,$FE,$D9,$F9,$07,$3F,$C4,$BE,$7F,$CA,$7F,$AE
	.dc.b	$F4,$7E,$83,$FC,$63,$40,$F4,$7F,$9F,$BF,$E1,$8D,$BC,$4F,$E3,$7E
	.dc.b	$07,$FC,$45,$F7,$2F,$DB,$5F,$A1,$BF,$6D,$FE,$BF,$F9,$8B,$F5,$AF
	.dc.b	$B7,$C3,$F8,$4C,$FF,$CD,$9B,$7E,$A5,$FF,$36,$1F,$E1,$4F,$AE,$FF
	.dc.b	$0A,$F9,$BE,$CD,$BF,$39,$3F,$D2,$39,$FF,$59,$F5,$DF,$E8,$29,$FF
	.dc.b	$0E,$4F,$F1,$AF,$84,$DF,$E0,$FF,$ED,$48,$FE,$B2,$7F,$A3,$8F,$F2
	.dc.b	$64,$FE,$E7,$2F,$D8,$7C,$17,$EE,$27,$67,$E5,$BE,$01,$F0,$3F,$6B
	.dc.b	$E7,$FF,$C5,$37,$FE,$CB,$C3,$FC,$E9,$7F,$E2,$C9,$FC,$8F,$D0,$7F
	.dc.b	$4C,$F6,$3F,$C4,$7F,$41,$F9,$4F,$9E,$FE,$9D,$F4,$F9,$CB,$FC,$37
	.dc.b	$EB,$3E,$BA,$13,$EB,$FF,$27,$E2,$62,$3F,$C4,$5F,$15,$FA,$AD,$DD
	.dc.b	$7F,$A2,$F5,$FF,$33,$D5,$FA,$4F,$2F,$F8,$18,$FD,$03,$F4,$FF,$37
	.dc.b	$F9,$9F,$47,$EB,$7E,$50,$FC,$55,$F8,$B2,$31,$F8,$B6,$FE,$E1,$F8
	.dc.b	$B3,$EF,$7F,$C0,$6D,$FA,$DF,$D2,$70,$FE,$2F,$BB,$E2,$74,$3D,$0F
	.dc.b	$E0,$3C,$3F,$E2,$7F,$A8,$87,$F8,$A7,$F2,$9F,$D2,$EB,$FE,$28,$FB
	.dc.b	$58,$7F,$8A,$B0,$FF,$4B,$5B,$F2,$2F,$5C,$7F,$93,$B2,$FC,$DF,$97
	.dc.b	$FC,$C7,$C4,$FE,$07,$F0,$DA,$D7,$C3,$ED,$F0,$FD,$A3,$FC,$31,$87
	.dc.b	$FA,$3F,$D8,$F9,$9F,$1F,$FB,$3F,$B5,$7E,$C1,$FE,$6D,$FC,$77,$F8
	.dc.b	$DE,$CF,$D0,$78,$E3,$FE,$73,$F1,$6B,$FE,$25,$F2,$3F,$C5,$AF,$F6
	.dc.b	$EF,$6B,$F9,$F7,$AA,$FC,$57,$F8,$07,$E2,$5F,$E6,$BD,$7F,$C5,$9F
	.dc.b	$B7,$7F,$8A,$70,$FD,$AD,$FA,$DF,$F3,$CF,$C5,$FE,$75,$FF,$40,$47
	.dc.b	$F4,$FF,$B3,$FE,$A0,$E6,$3F,$A6,$FA,$DF,$9E,$7F,$90,$FA,$9F,$E0
	.dc.b	$0E,$9F,$F7,$07,$D0,$7F,$34,$BF,$F0,$1F,$AB,$FD,$5A,$FF,$93,$F6
	.dc.b	$7F,$BE,$5F,$ED,$5F,$AD,$6D,$FF,$19,$1F,$DB,$7F,$2E,$FE,$7B,$F8
	.dc.b	$CF,$F0,$FF,$1B,$EF,$FE,$03,$FE,$7E,$D7,$FC,$67,$F7,$5F,$94,$FB
	.dc.b	$4F,$F1,$AF,$B7,$FE,$31,$EC,$FB,$5E,$33,$F6,$FF,$F1,$97,$DD,$43
	.dc.b	$FC,$09,$3A,$FD,$6B,$F1,$3E,$6F,$E2,$63,$F0,$5A,$FF,$34,$F9,$DF
	.dc.b	$E9,$BF,$86,$7F,$37,$E6,$FA,$11,$FD,$43,$EF,$3F,$9D,$68,$FD,$01
	.dc.b	$F7,$BF,$EE,$1F,$D0,$5F,$8E,$6B,$C1,$FF,$1A,$FE,$38,$F9,$27,$8B
	.dc.b	$F8,$EC,$BF,$A9,$F0,$7A,$BE,$E3,$F6,$8F,$93,$F1,$1F,$EE,$DF,$0D
	.dc.b	$FD,$0B,$D3,$FE,$B3,$F8,$67,$F4,$1F,$8F,$FA,$5E,$2F,$F8,$13,$07
	.dc.b	$DD,$FF,$68,$F0,$1F,$17,$FB,$1F,$CF,$5F,$A6,$FF,$42,$FC,$8D,$F9
	.dc.b	$AF,$F4,$BC,$79,$F9,$3F,$E1,$3F,$96,$FF,$04,$69,$FE,$10,$E5,$FF
	.dc.b	$6E,$78,$0F,$CA,$BF,$D1,$7F,$B7,$BF,$1A,$7C,$87,$FA,$B3,$E8,$3F
	.dc.b	$D0,$1D,$1D,$34,$FF,$0C,$9F,$DB,$39,$7F,$C2,$5E,$BF,$FB,$3B,$D5
	.dc.b	$FC,$2C,$EA,$7F,$DA,$DA,$7F,$B4,$7E,$F7,$F2,$F3,$FE,$D4,$DF,$F9
	.dc.b	$6C,$FF,$8B,$7F,$4B,$FC,$C4,$FF,$56,$FA,$1F,$CD,$FC,$D7,$F4,$88
	.dc.b	$ED,$EB,$7F,$BE,$FE,$1B,$FC,$4F,$1F,$C0,$8F,$D3,$F3,$7F,$89,$FF
	.dc.b	$41,$FE,$D6,$FF,$60,$8F,$F3,$0F,$03,$81,$F9,$CB,$F3,$4F,$E1,$DF
	.dc.b	$0A,$F8,$27,$E9,$7F,$E2,$EF,$79,$F9,$0E,$A7,$8D,$FF,$15,$D3,$FD
	.dc.b	$AF,$FA,$8B,$F5,$08,$4F,$C7,$3F,$C1,$FF,$5D,$88,$95,$FE,$DE,$F9
	.dc.b	$CF,$F1,$2F,$B0,$FC,$DF,$FC,$2B,$F1,$58,$F7,$3F,$C3,$FE,$87,$E3
	.dc.b	$BE,$C3,$F9,$6F,$D1,$7F,$89,$3E,$E3,$F3,$BE,$DF,$E6,$BD,$C3,$7E
	.dc.b	$CB,$F4,$3F,$B5,$BE,$43,$F3,$BE,$7F,$F8,$5B,$E6,$7F,$98,$9F,$ED
	.dc.b	$AF,$C6,$3F,$C3,$3F,$91,$88,$4F,$DC,$7F,$D3,$7E,$57,$F8,$6B,$C4
	.dc.b	$FF,$1F,$F4,$FF,$88,$3F,$20,$FF,$0F,$F4,$FF,$87,$BC,$0F,$F2,$CD
	.dc.b	$FF,$A6,$8F,$F0,$9C,$7F,$88,$BE,$0B,$FA,$EF,$E6,$BF,$E1,$EF,$D3
	.dc.b	$3F,$A8,$FE,$38,$FD,$1F,$FC,$51,$F3,$7F,$E4,$CF,$84,$FF,$0F,$7D
	.dc.b	$C7,$F8,$77,$3F,$F0,$E7,$D7,$FF,$87,$B9,$7F,$CA,$3C,$3F,$E5,$4C
	.dc.b	$BF,$C5,$BE,$F7,$F4,$9E,$8F,$EC,$5F,$B4,$3E,$C5,$FB,$8B,$F0,$AF
	.dc.b	$F8,$03,$F2,$1F,$EE,$1F,$C2,$3F,$B9,$7A,$FF,$E1,$1D,$FD,$5A,$FE
	.dc.b	$B0,$FF,$41,$7C,$0F,$F8,$73,$6F,$F1,$DC,$7F,$B8,$8F,$F0,$D9,$E4
	.dc.b	$3F,$C7,$FF,$88,$DF,$80,$FF,$10,$FC,$67,$F8,$6F,$E8,$7F,$D3,$D5
	.dc.b	$FE,$6B,$C7,$BD,$7E,$99,$FE,$E1,$3F,$B1,$F7,$BF,$C5,$BF,$43,$FA
	.dc.b	$CB,$FF,$45,$1F,$DD,$3A,$3F,$B9,$FD,$B7,$F8,$5F,$F1,$FF,$81,$F0
	.dc.b	$9F,$C9,$3E,$77,$FC,$2F,$BF,$F3,$9F,$0B,$F9,$AF,$CD,$7F,$30,$1F
	.dc.b	$5B,$FE,$70,$FA,$DF,$C3,$F9,$9F,$DD,$FD,$3F,$F0,$B7,$CB,$FF,$22
	.dc.b	$DB,$FA,$77,$E4,$9F,$E1,$5C,$BF,$65,$C0,$AF,$EF,$87,$FB,$E7,$ED
	.dc.b	$BF,$17,$F8,$47,$F9,$D8,$FF,$3B,$FE,$4B,$FC,$EF,$F1,$5F,$F3,$9F
	.dc.b	$E0,$F3,$16,$7F,$E7,$0C,$BE,$FF,$37,$F3,$EF,$B2,$FF,$03,$ED,$FC
	.dc.b	$26,$7F,$D2,$73,$FA,$AD,$7C,$18,$7F,$85,$E7,$FD,$CB,$0F,$47,$FC
	.dc.b	$1F,$E6,$7F,$81,$AB,$F8,$47,$D7,$FF,$81,$7D,$2F,$F3,$D7,$57,$F8
	.dc.b	$0A,$7F,$9B,$E3,$FE,$D4,$1E,$5F,$F4,$87,$1E,$67,$F9,$B3,$E4,$DF
	.dc.b	$49,$FE,$05,$BF,$E9,$BB,$FF,$15,$C1,$FB,$57,$F8,$17,$C3,$FF,$64
	.dc.b	$F0,$FE,$1F,$3F,$F0,$11,$FE,$D8,$3F,$C0,$47,$E8,$31,$FC,$37,$C9
	.dc.b	$FF,$80,$63,$FD,$C9,$E1,$FF,$9C,$BE,$BF,$F9,$C7,$C2,$7F,$7D,$C3
	.dc.b	$F9,$5E,$3F,$EB,$AB,$FF,$6E,$6B,$F6,$3E,$1F,$FC,$4B,$F8,$DF,$FA
	.dc.b	$27,$CB,$7E,$00,$F9,$57,$E9,$DF,$EC,$AF,$C8,$D5,$A9,$5F,$1C,$7F
	.dc.b	$07,$D8,$FD,$AB,$F8,$6F,$C1,$7F,$0E,$FA,$1F,$F2,$DF,$E9,$7F,$E0
	.dc.b	$3F,$AE,$FF,$02,$E1,$F8,$AF,$9B,$FF,$31,$E3,$F6,$BE,$AB,$FD,$92
	.dc.b	$FE,$37,$EC,$3F,$D9,$3C,$7F,$73,$E3,$7F,$BF,$7D,$6B,$8B,$FD,$93
	.dc.b	$F1,$1F,$DF,$7E,$30,$DF,$C9,$7F,$8E,$FB,$FF,$ED,$D2,$6D,$F9,$0C
	.dc.b	$87,$DA,$FC,$03,$53,$FE,$E4,$FC,$B7,$F1,$DB,$7E,$7C,$97,$FE,$20
	.dc.b	$F9,$3F,$E7,$58,$7F,$6B,$ED,$FE,$48,$7E,$72,$7F,$29,$1F,$A7,$F3
	.dc.b	$7F,$C1,$30,$F7,$FF,$C7,$5F,$90,$7F,$82,$5F,$F1,$E7,$D1,$BF,$23
	.dc.b	$7E,$40,$FA,$87,$9A,$FC,$6B,$FC,$13,$93,$D1,$FF,$7B,$FE,$79,$FE
	.dc.b	$C9,$8E,$BB,$7F,$82,$F7,$AF,$F1,$9F,$43,$F7,$7F,$AF,$BF,$1B,$7E
	.dc.b	$84,$FD,$8D,$F8,$AF,$F8,$33,$F0,$67,$AF,$FA,$4F,$B8,$F9,$39,$FE
	.dc.b	$3C,$FF,$0C,$78,$DD,$9F,$F0,$7F,$BD,$FE,$E4,$FA,$FF,$F0,$77,$C6
	.dc.b	$7F,$98,$FE,$3F,$FD,$CF,$F6,$3F,$CA,$FE,$43,$F9,$66,$1F,$E5,$C3
	.dc.b	$FA,$0F,$27,$FB,$33,$97,$C3,$F8,$27,$F4,$7F,$77,$FD,$81,$FA,$87
	.dc.b	$F4,$9F,$7F,$FD,$85,$F8,$97,$FA,$BE,$7F,$CD,$7E,$C7,$FA,$C7,$AF
	.dc.b	$FA,$F6,$7F,$80,$CF,$F8,$DF,$D3,$7F,$81,$FE,$CF,$FC,$15,$F8,$4B
	.dc.b	$ED,$3F,$C1,$5F,$69,$FE,$FF,$E2,$F7,$FB,$DF,$A3,$F7,$BF,$55,$F0
	.dc.b	$1C,$FA,$7E,$1F,$D4,$88,$43,$1A,$FF,$8D,$F6,$FE,$2F,$E8,$7E,$34
	.dc.b	$FC,$77,$B4,$9F,$CB,$4F,$F9,$7F,$DA,$FD,$0F,$ED,$D9,$C7,$E8,$FC
	.dc.b	$AF,$F3,$AF,$95,$7E,$D3,$C7,$AB,$EA,$FC,$0B,$E4,$FF,$CC,$5F,$B9
	.dc.b	$3E,$FB,$FC,$55,$F4,$6F,$CF,$B8,$6F,$FE,$90,$E4,$FC,$43,$FA,$6F
	.dc.b	$D7,$3F,$C5,$DF,$56,$FC,$E5,$FA,$BB,$F7,$97,$EC,$8F,$D6,$3F,$94
	.dc.b	$F7,$7D,$3C,$BE,$EF,$A3,$FE,$E9,$7F,$63,$E2,$FE,$A3,$C7,$79,$DF
	.dc.b	$B4,$A7,$95,$F7,$7E,$07,$F8,$1E,$1F,$C5,$38,$7F,$8C,$78,$7F,$4A
	.dc.b	$7E,$46,$5F,$94,$8F,$AB,$E5,$7F,$18,$CF,$F8,$DF,$83,$F3,$3D,$BF
	.dc.b	$D5,$10,$FD,$67,$A7,$CB,$3F,$E0,$BE,$07,$CD,$F0,$3E,$A7,$E0,$3F
	.dc.b	$A0,$F8,$AF,$F1,$88,$F8,$9C,$DF,$F0,$CF,$88,$FE,$16,$0D,$67,$FD
	.dc.b	$22,$3F,$88,$FE,$8A,$FD,$C5,$FB,$A3,$F2,$7F,$AD,$F0,$8F,$CF,$3E
	.dc.b	$B6,$2F,$B7,$FF,$6C,$FC,$2F,$F9,$8F,$3F,$E3,$7E,$5F,$F5,$3F,$84
	.dc.b	$7B,$5F,$CD,$7F,$55,$7D,$0F,$F5,$CF,$81,$FF,$70,$F8,$3F,$4F,$1F
	.dc.b	$E3,$3D,$CE,$C7,$7F,$D3,$CB,$FD,$A1,$F9,$F7,$FB,$5B,$C2,$FD,$83
	.dc.b	$F5,$FE,$C3,$F8,$06,$3F,$C1,$32,$E2,$F9,$0F,$E1,$27,$F8,$74,$79
	.dc.b	$9F,$6B,$2F,$B9,$E1,$7F,$52,$E4,$FE,$53,$F9,$9B,$F3,$17,$E8,$2F
	.dc.b	$CD,$7F,$C7,$D3,$9F,$CB,$FF,$51,$D7,$FB,$2F,$7B,$F6,$07,$EC,$B1
	.dc.b	$FD,$5F,$CF,$FF,$56,$3F,$AB,$3F,$CB,$36,$FE,$87,$F8,$17,$F9,$C4
	.dc.b	$9F,$3B,$F8,$38,$7E,$75,$FD,$3F,$E7,$67,$87,$FB,$7E,$BF,$06,$7E
	.dc.b	$0F,$D1,$FF,$75,$F1,$DF,$C9,$BF,$75,$FF,$6B,$FC,$7F,$DB,$F0,$3F
	.dc.b	$89,$78,$7F,$BF,$FD,$AF,$FD,$E7,$B7,$FA,$FC,$FE,$5B,$F3,$1F,$E9
	.dc.b	$BA,$7C,$1F,$FA,$73,$E5,$1F,$9A,$FC,$AF,$27,$F0,$58,$7F,$85,$3C
	.dc.b	$6F,$E5,$6F,$F2,$83,$F5,$FE,$3F,$F9,$E7,$1F,$E5,$31,$88,$1F,$E3
	.dc.b	$E7,$F9,$FF,$89,$FC,$FA,$1E,$AF,$4F,$17,$EE,$D7,$09,$FE,$84,$FF
	.dc.b	$44,$EB,$FF,$23,$FE,$1C,$F0,$3E,$D7,$85,$F6,$7E,$FF,$1F,$8D,$7C
	.dc.b	$DE,$77,$DE,$FF,$96,$8E,$F8,$FD,$8F,$0B,$EC,$1F,$95,$7F,$CA,$7E
	.dc.b	$DB,$E9,$5F,$95,$3F,$2A,$C3,$7F,$6C,$FC,$2F,$07,$F8,$03,$C4,$FB
	.dc.b	$5F,$6B,$FE,$AB,$4F,$DB,$8F,$F0,$2F,$87,$F8,$A3,$F9,$E7,$5F,$F2
	.dc.b	$68,$FC,$97,$80,$FB,$77,$EB,$2F,$D1,$3F,$BC,$FE,$BF,$EC,$E1,$FE
	.dc.b	$01,$87,$CE,$EC,$78,$DF,$CB,$A1,$F9,$23,$F1,$DF,$EF,$9F,$1C,$FB
	.dc.b	$D7,$E5,$8F,$D4,$BF,$6B,$E3,$FE,$03,$C3,$FA,$B9,$7F,$92,$8F,$F5
	.dc.b	$4F,$A3,$FF,$00,$1F,$F0,$FF,$E5,$3F,$B5,$FA,$5F,$F5,$87,$93,$FC
	.dc.b	$C2,$7F,$BE,$FB,$BF,$EB,$37,$F7,$3C,$5F,$D4,$BE,$83,$FB,$07,$B9
	.dc.b	$FC,$C7,$CD,$FE,$AD,$F8,$57,$F8,$0B,$E8,$FF,$B5,$7E,$37,$FE,$ED
	.dc.b	$F0,$F3,$97,$F2,$3F,$8A,$FE,$47,$F1,$78,$8F,$A1,$E1,$CF,$8C,$FE
	.dc.b	$4B,$F3,$FF,$E0,$CF,$C9,$FF,$C9,$1F,$73,$FE,$48,$F1,$7F,$C1,$FF
	.dc.b	$AD,$BD,$DF,$F1,$07,$EE,$DF,$3F,$DF,$FF,$7F,$FA,$5F,$DA,$7E,$13
	.dc.b	$F9,$2F,$C9,$7F,$85,$B1,$FE,$49,$F1,$7F,$C8,$FB,$3F,$CF,$B3,$9E
	.dc.b	$1F,$90,$7F,$91,$3F,$02,$FE,$35,$FA,$26,$63,$FC,$8F,$F0,$9F,$ED
	.dc.b	$4F,$7F,$F9,$BC,$7E,$B3,$AF,$FD,$89,$F8,$BF,$CA,$8F,$E4,$03,$0F
	.dc.b	$F6,$45,$7F,$2F,$CB,$FC,$53,$87,$EA,$7B,$FF,$95,$5F,$F9,$53,$EB
	.dc.b	$FF,$72,$FF,$8C,$3F,$05,$FF,$0D,$7A,$0F,$CD,$79,$FE,$A9,$FB,$5E
	.dc.b	$B6,$5F,$DB,$BF,$07,$FD,$C7,$B0,$FC,$1F,$FD,$07,$E6,$FF,$88,$FE
	.dc.b	$3F,$FC,$AF,$87,$F8,$2F,$5F,$F3,$1F,$E7,$9F,$E4,$BE,$F7,$5B,$F4
	.dc.b	$C7,$EE,$AF,$C0,$DF,$AF,$F1,$C5,$FB,$1B,$F1,$57,$DD,$7F,$84,$63
	.dc.b	$1F,$73,$FE,$4A,$F4,$FF,$98,$5F,$FA,$27,$F1,$C7,$E4,$DF,$AF,$9F
	.dc.b	$F6,$7F,$EA,$4F,$CD,$FF,$C2,$DE,$C8,$85,$9E,$7F,$BC,$BF,$6A,$FD
	.dc.b	$97,$47,$F1,$51,$E3,$FF,$02,$D1,$5B,$F6,$BF,$DF,$9E,$3E,$7E,$D1
	.dc.b	$E9,$3F,$79,$7E,$77,$FB,$2F,$CF,$2F,$E3,$DF,$9B,$FC,$8D,$F3,$0F
	.dc.b	$CF,$75,$7F,$B8,$F8,$1F,$8D,$C3,$FC,$BB,$D4,$FF,$43,$FB,$1E,$3F
	.dc.b	$CE,$BF,$CD,$BF,$9A,$3F,$79,$84,$E8,$F8,$FF,$84,$7F,$BC,$7F,$0A
	.dc.b	$FF,$24,$2F,$D0,$3D,$99,$FF,$3A,$FC,$4B,$FD,$27,$CD,$88,$FF,$25
	.dc.b	$F5,$7E,$6B,$1F,$F2,$93,$FE,$9D,$C7,$9E,$3F,$B5,$3F,$DF,$DF,$A0
	.dc.b	$76,$FD,$CF,$FA,$DD,$7F,$47,$3F,$D0,$27,$F9,$AF,$E5,$BF,$D0,$7F
	.dc.b	$2A,$FC,$CD,$F3,$CF,$D8,$3F,$D9,$FE,$DB,$F3,$3F,$A7,$B3,$F2,$FF
	.dc.b	$F4,$5F,$87,$D3,$47,$E1,$5F,$D3,$7E,$5F,$FC,$C7,$C7,$FD,$3E,$75
	.dc.b	$F1,$FE,$2F,$C5,$FF,$DD,$FF,$83,$FF,$2C,$FA,$9F,$F1,$0F,$E8,$DF
	.dc.b	$EF,$AC,$7F,$C1,$F8,$7A,$4F,$D6,$E3,$FD,$83,$F0,$DF,$F5,$CE,$1E
	.dc.b	$1D,$BF,$AE,$7C,$57,$F2,$9F,$C7,$35,$7F,$8B,$FC,$3B,$CD,$1F,$CE
	.dc.b	$F1,$F8,$7B,$FA,$7F,$1A,$F5,$7D,$EF,$F6,$CF,$84,$FF,$58,$63,$F9
	.dc.b	$3F,$8B,$FE,$17,$87,$CA,$E3,$DC,$FF,$72,$F8,$F7,$C8,$FF,$11,$AF
	.dc.b	$E1,$71,$FC,$41,$FE,$E1,$EB,$7C,$7F,$B4,$FE,$21,$1F,$D3,$BD,$DF
	.dc.b	$EF,$7F,$29,$FC,$FE,$3F,$DB,$FC,$5F,$EB,$8F,$8B,$FF,$51,$7C,$97
	.dc.b	$F9,$BF,$E4,$3F,$BF,$74,$7F,$80,$6F,$FC,$E5,$F8,$1F,$F2,$9F,$CB
	.dc.b	$39,$C7,$F2,$1F,$0B,$FD,$59,$C3,$FC,$42,$3F,$92,$6D,$FE,$D2,$E0
	.dc.b	$FF,$03,$7C,$9F,$F1,$8F,$B3,$FE,$43,$E3,$7F,$B2,$39,$FF,$8E,$D7
	.dc.b	$F8,$2B,$F2,$1B,$EF,$FF,$88,$B2,$FE,$2E,$7F,$BD,$BF,$11,$7D,$AB
	.dc.b	$F6,$8F,$E0,$9F,$31,$F8,$9F,$C7,$DF,$B2,$FE,$BB,$F3,$D7,$E7,$4F
	.dc.b	$D3,$9E,$AB,$F7,$AF,$ED,$5F,$51,$FC,$67,$E8,$1F,$96,$3E,$2F,$F8
	.dc.b	$D7,$CB,$CF,$8C,$FE,$3D,$FB,$9F,$F8,$CB,$DD,$FD,$AF,$C2,$7E,$D7
	.dc.b	$BF,$FD,$C7,$F8,$E3,$F6,$3F,$F7,$57,$BF,$FA,$33,$F4,$9F,$A2,$FF
	.dc.b	$92,$FE,$4F,$F8,$B7,$B3,$FB,$AF,$88,$FF,$27,$7E,$7A,$F8,$68,$FF
	.dc.b	$B1,$7F,$35,$FE,$5D,$19,$FD,$13,$87,$BD,$FC,$73,$C3,$0F,$8D,$7C
	.dc.b	$E7,$F9,$0A,$7F,$97,$FE,$9B,$FE,$2E,$FC,$89,$F9,$BB,$F5,$87,$EA
	.dc.b	$18,$FD,$03,$F8,$BF,$3B,$E5,$BF,$A3,$E5,$FB,$28,$FC,$37,$93,$FE
	.dc.b	$27,$D9,$F8,$D7,$F8,$6B,$E4,$7F,$D4,$1E,$24,$FB,$7F,$E6,$1F,$9B
	.dc.b	$FF,$05,$F9,$23,$F1,$7B,$5D,$BF,$C5,$FE,$5F,$F9,$2E,$DF,$D1,$B6
	.dc.b	$FD,$56,$BF,$C4,$3E,$13,$FD,$5E,$FF,$19,$EA,$FF,$75,$76,$FF,$3D
	.dc.b	$7F,$93,$1F,$EE,$DF,$2F,$F8,$E2,$F9,$5F,$E5,$3F,$8C,$7F,$BB,$FE
	.dc.b	$3B,$F3,$7F,$0F,$FC,$04,$F9,$78,$BF,$55,$7A,$7F,$EA,$AF,$CA,$3F
	.dc.b	$D6,$5E,$8F,$F6,$0A,$FF,$08,$1F,$4B,$F4,$BF,$EF,$3C,$3F,$C7,$63
	.dc.b	$EA,$FC,$BF,$F0,$DE,$2F,$F3,$7F,$CA,$7F,$9C,$27,$FC,$E5,$A7,$F1
	.dc.b	$C7,$C3,$FB,$93,$F6,$CF,$F2,$27,$A3,$F7,$3A,$1F,$9E,$7F,$29,$C7
	.dc.b	$FC,$85,$F2,$7F,$E0,$CC,$7D,$30,$FC,$CB,$DB,$1F,$C8,$34,$7C,$6F
	.dc.b	$F4,$68,$FE,$77,$CB,$FC,$81,$FF,$08,$BF,$C8,$27,$FC,$FD,$FA,$AF
	.dc.b	$F2,$07,$FC,$9F,$3F,$1B,$CD,$FE,$40,$FF,$44,$FD,$3F,$EB,$BE,$29
	.dc.b	$FE,$E5,$F0,$DF,$C8,$23,$E6,$FC,$B7,$F3,$FA,$FF,$45,$7C,$47,$F9
	.dc.b	$92,$7F,$37,$FA,$63,$F0,$D7,$ED,$8F,$D3,$5F,$BB,$BF,$2F,$7E,$75
	.dc.b	$F8,$CC,$9F,$BC,$7D,$AF,$9A,$FE,$1F,$FB,$C7,$FB,$6B,$5F,$E9,$C2
	.dc.b	$BF,$96,$70,$FF,$81,$EB,$FC,$56,$BE,$4F,$FD,$55,$5F,$E0,$B8,$FF
	.dc.b	$73,$3F,$E6,$9A,$7E,$1F,$FE,$33,$F0,$3F,$B2,$FE,$1F,$FB,$49,$FF
	.dc.b	$1C,$57,$FA,$CB,$1F,$F1,$A3,$FC,$AC,$C1,$FE,$46,$F9,$8F,$F5,$86
	.dc.b	$29,$DC,$FF,$BB,$7E,$5F,$D8,$FF,$AC,$FE,$67,$F9,$63,$FB,$FA,$FE
	.dc.b	$6A,$7E,$E4,$FE,$0B,$F3,$DF,$C3,$BF,$33,$7D,$A3,$F3,$A7,$E6,$AF
	.dc.b	$B2,$FF,$66,$F5,$FF,$B8,$7F,$44,$7E,$90,$FD,$4B,$F8,$6F,$E0,$4F
	.dc.b	$D5,$7F,$D9,$3F,$3B,$FE,$43,$FC,$E7,$F9,$BF,$D0,$BE,$A5,$F9,$E3
	.dc.b	$47,$E9,$DF,$EB,$BE,$4F,$D7,$FC,$A3,$E4,$DE,$73,$CA,$7E,$7E,$FD
	.dc.b	$11,$E2,$F9,$3D,$BF,$AD,$DB,$FC,$8B,$F0,$87,$E9,$EF,$B0,$FE,$7B
	.dc.b	$F9,$A3,$F5,$C7,$E8,$D9,$C7,$8E,$2F,$D3,$BF,$D9,$DE,$E3,$F7,$B7
	.dc.b	$CC,$BF,$62,$7E,$65,$FC,$4B,$D0,$FF,$55,$63,$FE,$09,$F4,$BF,$84
	.dc.b	$FD,$2F,$E5,$3F,$01,$FF,$27,$FB,$8F,$CC,$9F,$95,$F8,$6F,$F8,$7E
	.dc.b	$DF,$E8,$7F,$61,$F9,$2F,$F8,$A7,$F7,$BF,$F5,$07,$E3,$FF,$CD,$3C
	.dc.b	$2F,$F1,$F6,$1F,$73,$97,$FC,$D7,$F7,$DD,$7F,$79,$7E,$92,$FC,$31
	.dc.b	$F8,$FF,$D1,$FB,$DF,$E4,$DF,$37,$FE,$4A,$FC,$B7,$FC,$95,$97,$F2
	.dc.b	$FD,$3F,$DF,$9F,$1B,$FD,$37,$DD,$FF,$25,$47,$F8,$67,$EA,$3F,$95
	.dc.b	$E1,$FD,$5C,$FE,$AF,$F3,$DF,$E4,$6E,$BF,$D3,$50,$FC,$D3,$FE,$E9
	.dc.b	$FC,$D1,$FB,$A3,$F6,$17,$E8,$1F,$03,$F1,$0F,$F6,$FF,$E7,$DF,$EE
	.dc.b	$FE,$7F,$E5,$7E,$07,$E7,$32,$FF,$22,$E7,$FC,$AA,$7F,$3E,$F7,$F8
	.dc.b	$8F,$EC,$7E,$0F,$F9,$5B,$93,$E5,$FE,$4A,$FD,$0F,$FC,$A5,$F8,$63
	.dc.b	$F6,$77,$E9,$0F,$9C,$7D,$8F,$F9,$53,$EB,$3D,$7A,$FF,$95,$4F,$F4
	.dc.b	$67,$E6,$FE,$CE,$BF,$E7,$91,$FE,$8F,$E0,$FE,$1B,$EE,$6B,$87,$F9
	.dc.b	$07,$17,$F7,$68,$FF,$4A,$7D,$2B,$F0,$EF,$F2,$E4,$FF,$50,$FA,$4F
	.dc.b	$EF,$78,$FF,$AD,$A3,$FC,$DF,$FA,$17,$F8,$D5,$FF,$18,$1F,$D5,$BE
	.dc.b	$4F,$73,$FC,$57,$E5,$FF,$95,$4B,$F3,$CF,$F1,$47,$E1,$CF,$D3,$3F
	.dc.b	$55,$FA,$03,$F6,$37,$67,$F2,$3F,$C2,$BF,$DD,$1F,$8B,$7F,$26,$CB
	.dc.b	$F9,$36,$BF,$3B,$5F,$E2,$9E,$4F,$F2,$9F,$86,$FF,$77,$7C,$FE,$F8
	.dc.b	$3F,$95,$D7,$FB,$77,$A3,$FC,$C5,$1F,$EB,$9C,$7C,$F3,$F8,$5E,$5F
	.dc.b	$F6,$B7,$CD,$7F,$27,$FA,$0F,$E5,$58,$FF,$72,$3F,$95,$7E,$13,$DB
	.dc.b	$0F,$E7,$38,$FF,$2C,$7F,$DD,$5D,$BF,$EE,$EF,$C5,$3F,$94,$F6,$7C
	.dc.b	$69,$FE,$65,$F4,$5F,$EF,$FE,$CF,$F1,$44,$7F,$7E,$C1,$FB,$D7,$FB
	.dc.b	$BB,$5F,$F2,$BC,$FF,$8B,$30,$E4,$EF,$F3,$1F,$F0,$2F,$B5,$FE,$96
	.dc.b	$FC,$F7,$C7,$BF,$F8,$23,$E0,$BF,$3F,$7F,$DF,$CF,$7F,$1F,$F3,$3F
	.dc.b	$84,$FF,$7A,$7B,$1F,$E0,$19,$7C,$67,$F7,$7F,$DF,$5F,$9F,$3F,$3C
	.dc.b	$FF,$16,$FB,$2F,$CE,$58,$F1,$F9,$E7,$E9,$EF,$D6,$DF,$AA,$FE,$37
	.dc.b	$E3,$FF,$5D,$F0,$1C,$38,$7F,$C3,$BF,$9E,$FF,$40,$F9,$07,$EB,$8F
	.dc.b	$D8,$7F,$90,$7B,$0F,$D8,$F8,$73,$7F,$93,$BF,$2F,$FE,$47,$FB,$FB
	.dc.b	$F5,$8F,$F8,$01,$FB,$FF,$F8,$3F,$F7,$9F,$F7,$77,$07,$F9,$50,$7E
	.dc.b	$CE,$FB,$D7,$E4,$1F,$EC,$3E,$7F,$5F,$BF,$FD,$DF,$F0,$57,$9A,$FD
	.dc.b	$B1,$F9,$93,$F6,$9F,$F0,$E7,$93,$FE,$51,$EA,$FF,$09,$F2,$7F,$B5
	.dc.b	$3B,$FF,$9D,$FC,$97,$FB,$73,$E5,$FF,$8E,$D7,$F8,$F3,$F1,$F7,$ED
	.dc.b	$CF,$DC,$5F,$A4,$7F,$98,$FF,$5E,$FA,$7F,$83,$7C,$EF,$9C,$FF,$21
	.dc.b	$9F,$3B,$E0,$1F,$8B,$3F,$23,$FA,$3F,$90,$FE,$B7,$83,$FC,$CD,$B7
	.dc.b	$FB,$CC,$FE,$6D,$87,$F9,$33,$F7,$F7,$C3,$BE,$8F,$FC,$35,$D4,$F8
	.dc.b	$5F,$F0,$F7,$E5,$1F,$13,$9F,$FC,$FB,$FB,$1F,$FB,$AF,$F2,$7F,$E3
	.dc.b	$59,$3F,$5B,$7D,$43,$EA,$3F,$9B,$EF,$FE,$3E,$F0,$BC,$39,$7E,$4A
	.dc.b	$3F,$9A,$7D,$5B,$E2,$7F,$A1,$F7,$FF,$3B,$3F,$81,$FA,$BF,$CF,$9F
	.dc.b	$E7,$DD,$EF,$C6,$FF,$CD,$18,$F3,$C7,$FD,$25,$CF,$FD,$E2,$3F,$D2
	.dc.b	$7E,$1B,$E5,$DF,$A7,$BF,$79,$7E,$58,$FD,$E3,$FC,$79,$9F,$F3,$FC
	.dc.b	$7F,$C6,$87,$FB,$CF,$E9,$BF,$A0,$67,$FD,$2F,$4F,$E6,$C7,$F1,$3F
	.dc.b	$27,$FD,$43,$EC,$7F,$5E,$BF,$F5,$1F,$E4,$1F,$A4,$FD,$0F,$FD,$91
	.dc.b	$F5,$AF,$D1,$7F,$C1,$7F,$A6,$7F,$3F,$F8,$2F,$E1,$98,$7F,$A0,$0E
	.dc.b	$7F,$4B,$FE,$82,$F9,$9F,$E6,$DA,$FF,$5C,$9D,$3F,$D0,$B5,$FF,$16
	.dc.b	$C7,$F9,$AA,$3F,$9C,$35,$FE,$C5,$F4,$BF,$EB,$7F,$37,$FC,$D5,$1F
	.dc.b	$C8,$7F,$6A,$FE,$83,$5F,$EB,$AF,$A5,$57,$ED,$EB,$E0,$78,$1E,$5A
	.dc.b	$F9,$91,$F1,$9F,$F4,$01,$FD,$0E,$78,$47,$C7,$FC,$2F,$FC,$3B,$FB
	.dc.b	$1B,$E0,$5E,$57,$E8,$3F,$34,$7E,$B6,$FD,$BD,$E6,$7F,$8F,$7F,$55
	.dc.b	$FF,$51,$E7,$FE,$2B,$D7,$F9,$8E,$FF,$EA,$69,$FF,$65,$75,$BF,$18
	.dc.b	$7E,$7B,$FE,$EF,$FD,$E3,$F8,$BF,$D0,$FF,$54,$CB,$FB,$16,$5E,$B8
	.dc.b	$FF,$6B,$BF,$95,$E8,$FE,$69,$FA,$67,$F8,$92,$3E,$A7,$2F,$F8,$2A
	.dc.b	$9F,$B5,$7F,$80,$3E,$73,$F9,$D7,$D3,$F9,$AF,$E3,$77,$FF,$9C,$7F
	.dc.b	$5D,$FF,$5D,$FE,$55,$FD,$6E,$7F,$D0,$1F,$47,$EC,$FC,$2B,$FC,$83
	.dc.b	$1F,$EC,$0F,$23,$F9,$A7,$07,$F8,$03,$E2,$7F,$81,$4F,$FA,$53,$E7
	.dc.b	$BF,$98,$FE,$51,$FE,$47,$FC,$85,$DC,$FD,$F3,$F9,$AF,$85,$FE,$CA
	.dc.b	$E8,$7E,$69,$FE,$6E,$F6,$7F,$CC,$5E,$E7,$F7,$1D,$3F,$96,$9F,$ED
	.dc.b	$5F,$98,$FF,$30,$FD,$8B,$EF,$9F,$AF,$3F,$2F,$7E,$CD,$C3,$99,$F8
	.dc.b	$1B,$F5,$07,$4B,$F4,$3F,$F6,$3E,$EE,$CF,$E1,$9B,$BF,$73,$FE,$83
	.dc.b	$F9,$F7,$F1,$3F,$67,$FC,$8F,$EF,$FF,$4B,$EF,$FD,$65,$BE,$F7,$FD
	.dc.b	$5F,$F9,$BB,$0F,$F0,$D7,$EC,$FF,$80,$F1,$3D,$7F,$11,$FC,$2C,$FF
	.dc.b	$23,$7D,$FF,$83,$F1,$DF,$C9,$7E,$04,$FA,$2F,$F1,$A7,$E1,$0F,$7F
	.dc.b	$E5,$6F,$E0,$FD,$E3,$F8,$EF,$27,$F4,$FF,$B0,$FF,$2E,$FC,$6F,$F7
	.dc.b	$8A,$FE,$5F,$E6,$FF,$05,$3F,$32,$FA,$39,$DF,$68,$FA,$BF,$F4,$9F
	.dc.b	$E4,$1F,$C4,$7E,$F7,$E3,$5F,$FA,$0B,$E8,$1F,$B4,$7F,$56,$FC,$F1
	.dc.b	$FA,$FF,$F4,$8F,$A0,$E7,$7F,$D3,$BA,$3D,$C3,$F6,$2F,$E6,$3B,$FF
	.dc.b	$17,$FC,$03,$F9,$6C,$FF,$2D,$FB,$FF,$E5,$8F,$F9,$A7,$F4,$9F,$EF
	.dc.b	$7E,$EB,$C3,$9F,$8E,$FF,$BF,$BE,$B1,$F9,$53,$F5,$27,$EF,$EF,$D7
	.dc.b	$1F,$A4,$7F,$81,$3F,$64,$FF,$64,$E7,$FE,$DE,$F5,$FF,$8E,$7D,$EB
	.dc.b	$F5,$D7,$EC,$4F,$D5,$FF,$D3,$DF,$AC,$BF,$70,$7E,$FB,$FE,$63,$FD
	.dc.b	$0F,$F8,$0F,$C9,$7E,$A7,$97,$FC,$85,$D4,$F8,$DF,$E3,$6F,$D1,$FD
	.dc.b	$1F,$FC,$C6,$FF,$15,$F9,$2F,$F6,$37,$CF,$FB,$6F,$F8,$6F,$C3,$FE
	.dc.b	$57,$F5,$FF,$F3,$87,$E5,$DF,$9A,$FD,$D7,$F8,$85,$7F,$98,$B4,$FF
	.dc.b	$26,$7C,$E7,$1F,$1F,$FD,$8D,$EB,$BD,$4F,$CF,$7E,$8C,$8F,$F6,$07
	.dc.b	$CC,$7F,$0F,$EC,$FE,$A9,$97,$F3,$D8,$7E,$77,$FE,$88,$F8,$9F,$E1
	.dc.b	$F3,$FE,$3F,$3F,$87,$FE,$2D,$FE,$F9,$FB,$6F,$F4,$6C,$7F,$0E,$3F
	.dc.b	$CE,$B8,$7F,$0E,$58,$7F,$85,$DF,$B5,$7F,$C3,$E3,$FD,$C5,$1F,$76
	.dc.b	$47,$F9,$9F,$F3,$AF,$88,$FF,$B8,$39,$BF,$CB,$BF,$51,$F1,$FF,$3E
	.dc.b	$E5,$F1,$9F,$4B,$A3,$FA,$16,$3F,$C2,$2F,$FD,$2B,$FA,$5F,$F0,$AF
	.dc.b	$C5,$5F,$BA,$FF,$14,$FD,$7D,$FB,$DB,$E6,$3F,$1D,$DB,$FC,$0B,$C6
	.dc.b	$FE,$C9,$E6,$FE,$37,$E7,$9F,$8E,$3D,$5F,$E2,$1E,$D7,$F9,$5F,$F1
	.dc.b	$BF,$E8,$EF,$F8,$87,$F7,$B7,$CE,$7F,$9A,$BF,$42,$FE,$37,$F7,$5F
	.dc.b	$E9,$1F,$8A,$FF,$5F,$7E,$13,$FE,$F4,$FC,$CD,$F9,$8F,$F1,$2F,$A9
	.dc.b	$7E,$A8,$FA,$BF,$F7,$87,$EB,$4F,$96,$7E,$EC,$FC,$A7,$F9,$3F,$E9
	.dc.b	$3F,$6B,$F1,$5F,$F4,$E7,$E8,$EF,$27,$F8,$A7,$EB,$7F,$C4,$BE,$2F
	.dc.b	$F6,$BF,$A9,$FF,$20,$FA,$E7,$E9,$3F,$E4,$2F,$AD,$FF,$68,$FE,$12
	.dc.b	$FD,$61,$FB,$4F,$F9,$BB,$F0,$4F,$CC,$7E,$40,$F3,$5F,$9A,$7F,$5E
	.dc.b	$FC,$5B,$F9,$61,$FC,$DB,$C3,$FE,$27,$E8,$7F,$85,$FC,$9F,$E5,$FF
	.dc.b	$BB,$BF,$5F,$FE,$B3,$F0,$EF,$D4,$FF,$DE,$FF,$84,$7B,$71,$7A,$FF
	.dc.b	$A7,$F5,$9F,$5D,$FE,$6F,$FC,$83,$97,$E0,$AF,$9F,$FE,$07,$87,$F1
	.dc.b	$3E,$CF,$E6,$1D,$FF,$C4,$B8,$7F,$DE,$5F,$9F,$FF,$12,$9F,$F7,$C7
	.dc.b	$C4,$FF,$12,$43,$E6,$7F,$89,$FB,$DF,$C4,$FF,$68,$FF,$39,$7E,$06
	.dc.b	$FC,$49,$FB,$2F,$EA,$BE,$B1,$F8,$7F,$F1,$1C,$3F,$DA,$7F,$15,$F9
	.dc.b	$6E,$2F,$5C,$7F,$1D,$DB,$E7,$7D,$E7,$F9,$A7,$EB,$1F,$B0,$BF,$5B
	.dc.b	$FF,$43,$60,$F9,$CF,$F0,$8F,$E3,$9F,$C4,$72,$FF,$28,$72,$BF,$26
	.dc.b	$FE,$22,$FF,$91,$7E,$EB,$F8,$B7,$EE,$9F,$DF,$FA,$BF,$A2,$3F,$C4
	.dc.b	$7C,$2F,$E2,$78,$7F,$2F,$F6,$1F,$B4,$BF,$69,$FF,$51,$F7,$FF,$70
	.dc.b	$FC,$89,$F1,$1E,$EF,$3B,$F6,$F8,$FF,$B7,$FE,$1F,$FB,$F7,$E9,$9E
	.dc.b	$3F,$E0,$07,$DC,$3F,$70,$F6,$69,$FC,$1B,$F4,$BE,$38,$FF,$B6,$BF
	.dc.b	$14,$FF,$16,$F2,$FF,$59,$F3,$78,$CF,$F0,$CF,$85,$FF,$78,$79,$DF
	.dc.b	$D1,$79,$DF,$AC,$7F,$2B,$FC,$21,$FC,$00,$FD,$DB,$FB,$67,$E7,$4F
	.dc.b	$CB,$BF,$C1,$1E,$47,$F8,$3B,$E7,$FF,$D9,$3E,$8B,$F3,$07,$96,$FE
	.dc.b	$00,$7C,$77,$DE,$DF,$FD,$CF,$F3,$7F,$5B,$F6,$47,$E9,$73,$F0,$47
	.dc.b	$EB,$6F,$D7,$7F,$C3,$7F,$1D,$FE,$19,$8F,$F6,$EF,$D5,$FE,$7F,$F2
	.dc.b	$BE,$F9,$FE,$13,$AD,$3F,$EE,$2F,$CB,$DF,$A5,$FF,$B8,$BF,$06,$E9
	.dc.b	$F9,$F7,$F8,$63,$CC,$FF,$6F,$FB,$9F,$ED,$5C,$1F,$3E,$F8,$17,$ED
	.dc.b	$1F,$E1,$EF,$B5,$FC,$BF,$DE,$3E,$AB,$FD,$97,$FA,$0F,$F1,$0F,$8C
	.dc.b	$FF,$75,$FD,$0F,$F5,$5F,$D1,$5F,$BF,$BF,$74,$7E,$BB,$FE,$C6,$FD
	.dc.b	$AB,$FC,$5B,$D9,$FE,$2A,$F4,$7F,$C8,$DD,$5F,$E2,$6F,$96,$7E,$B9
	.dc.b	$FA,$4F,$CE,$7F,$CD,$BF,$68,$F9,$D7,$DA,$FF,$8C,$BC,$CF,$F0,$48
	.dc.b	$D3,$F7,$19,$FF,$86,$BC,$FF,$ED,$13,$AF,$A8,$FD,$C2,$FD,$59,$B3
	.dc.b	$F7,$FF,$F2,$87,$EB,$1F,$E5,$3E,$9F,$8B,$BB,$F6,$F7,$A9,$FE,$4D
	.dc.b	$F2,$3F,$BA,$7E,$D4,$FD,$C7,$FD,$F9,$FA,$67,$F8,$37,$E6,$DF,$9E
	.dc.b	$7F,$84,$7E,$57,$FD,$E7,$87,$F5,$B3,$F6,$87,$F8,$C7,$93,$FC,$65
	.dc.b	$F1,$BF,$B4,$47,$FB,$BA,$3F,$DE,$CF,$FA,$EF,$F0,$1F,$DA,$7C,$DF
	.dc.b	$E0,$FE,$D1,$FA,$DF,$D1,$FC,$6B,$F6,$BF,$02,$FC,$33,$FC,$77,$6F
	.dc.b	$92,$FF,$3D,$67,$FE,$94,$3F,$C5,$5F,$B7,$7E,$13,$F4,$7F,$CD,$FF
	.dc.b	$00,$BE,$4D,$FB,$5F,$C7,$FC,$25,$F8,$77,$F8,$57,$E4,$7F,$C6,$5C
	.dc.b	$7F,$DB,$A7,$FD,$B3,$9F,$FA,$54,$FF,$1B,$2F,$92,$FF,$63,$47,$FB
	.dc.b	$27,$F4,$6F,$F4,$B7,$E7,$FF,$E9,$58,$DF,$CA,$75,$7F,$D7,$5F,$45
	.dc.b	$CB,$F4,$CF,$F4,$AC,$7F,$7C,$3F,$D2,$A7,$8B,$F1,$5F,$F4,$AB,$FC
	.dc.b	$07,$3F,$F1,$8F,$E8,$1F,$E7,$AF,$89,$78,$1F,$E9,$4F,$86,$7E,$17
	.dc.b	$FE,$94,$FC,$EB,$F0,$BF,$33,$FE,$93,$FC,$E7,$FD,$25,$83,$F1,$8F
	.dc.b	$F4,$97,$A1,$FB,$3F,$CA,$BF,$CC,$FF,$15,$FE,$66,$FD,$1B,$F8,$DF
	.dc.b	$E6,$DF,$E9,$6F,$B0,$FF,$32,$FE,$9C,$F9,$5F,$F3,$2F,$E1,$0F,$CA
	.dc.b	$BF,$CC,$BE,$EF,$E3,$3D,$B7,$96,$FE,$00,$7E,$DB,$FD,$AF,$D9,$FF
	.dc.b	$3F,$F8,$BF,$C6,$FA,$1F,$9B,$FF,$A3,$7F,$58,$FE,$C3,$FA,$1B,$F3
	.dc.b	$8F,$EF,$9F,$05,$FE,$8D,$FC,$C3,$FD,$6B,$E1,$FF,$21,$FD,$37,$FA
	.dc.b	$47,$6F,$6F,$6B,$F8,$BF,$EE,$DF,$C0,$FF,$21,$FF,$51,$7B,$EF,$C6
	.dc.b	$3F,$D1,$7F,$AC,$FE,$EF,$CA,$FC,$D1,$FD,$B7,$C3,$FF,$0A,$E9,$FE
	.dc.b	$75,$FC,$DB,$FD,$0B,$F8,$CF,$F9,$A7,$0F,$A3,$E5,$3C,$FF,$F5,$6F
	.dc.b	$9D,$FE,$76,$F5,$BF,$AE,$7C,$DF,$F8,$67,$F3,$D7,$E2,$FF,$C0,$79
	.dc.b	$FF,$D1,$3F,$03,$FC,$03,$E7,$7F,$BB,$47,$F7,$4F,$C7,$FF,$7F,$F9
	.dc.b	$27,$EF,$BD,$DF,$F3,$57,$4F,$E6,$3F,$24,$FA,$5F,$A5,$BF,$46,$7C
	.dc.b	$93,$F5,$EF,$E0,$BF,$7F,$FE,$84,$E9,$FF,$39,$1F,$D2,$FE,$B7,$FA
	.dc.b	$8F,$67,$ED,$8F,$F5,$91,$FB,$8F,$8F,$FD,$CC,$FF,$A5,$7E,$B1,$F8
	.dc.b	$27,$FB,$FB,$F7,$37,$E8,$4F,$9B,$FF,$39,$7A,$CC,$7F,$83,$E3,$FC
	.dc.b	$0B,$F2,$DF,$CF,$43,$87,$FD,$09,$A7,$F9,$08,$FA,$35,$E6,$7E,$98
	.dc.b	$FC,$63,$F8,$62,$FC,$23,$D7,$3F,$C8,$A3,$F8,$7F,$A7,$FE,$6E,$8F
	.dc.b	$F5,$64,$FF,$80,$1F,$F6,$87,$E1,$4F,$D8,$3F,$AD,$FC,$0B,$F7,$A7
	.dc.b	$9B,$FC,$CF,$E5,$BF,$87,$61,$FE,$93,$3F,$82,$73,$FF,$A6,$FE,$8D
	.dc.b	$FA,$DB,$D4,$7E,$4E,$FC,$39,$F0,$BF,$B4,$FA,$7F,$F4,$3F,$89,$F0
	.dc.b	$33,$F9,$A7,$FA,$57,$F3,$0F,$F4,$A3,$FC,$3B,$E5,$BF,$D5,$1E,$8F
	.dc.b	$BB,$E5,$BF,$B3,$CF,$FA,$43,$F4,$CF,$F4,$AB,$FC,$38,$FE,$32,$7F
	.dc.b	$0E,$3F,$D2,$BF,$B7,$7F,$AC,$7E,$9B,$FD,$29,$FA,$1F,$FA,$36,$7F
	.dc.b	$88,$7C,$DB,$CC,$FA,$7F,$A0,$7F,$10,$9F,$F4,$77,$7F,$E9,$3E,$BF
	.dc.b	$F5,$FF,$84,$3F,$2E,$FF,$09,$FD,$8B,$E2,$3F,$C2,$7F,$AE,$7F,$35
	.dc.b	$F8,$3F,$6F,$E0,$BF,$E8,$EF,$D0,$3F,$8A,$FE,$65,$FE,$8D,$FC,$AF
	.dc.b	$EC,$6D,$FE,$E1,$FC,$9B,$FC,$CD,$CD,$F7,$7F,$25,$FF,$32,$FE,$5F
	.dc.b	$FE,$9D,$FC,$9B,$FC,$C5,$F9,$3F,$E3,$7D,$EF,$F3,$0F,$CD,$FF,$B3
	.dc.b	$7F,$1D,$7F,$00,$FF,$A2,$7E,$5F,$FE,$00,$6C,$FA,$5F,$27,$57,$FB
	.dc.b	$4F,$F7,$BF,$EC,$DF,$80,$FF,$A3,$7F,$16,$FF,$05,$FD,$87,$F0,$EF
	.dc.b	$8A,$FF,$32,$9F,$E6,$6F,$CA,$3F,$87,$FE,$0E,$FD,$11,$FB,$CB,$F7
	.dc.b	$AF,$E3,$DF,$8F,$7F,$82,$3F,$15,$FE,$2B,$F7,$7F,$E8,$9F,$CB,$7F
	.dc.b	$D1,$1E,$67,$E6,$FF,$50,$FE,$21,$C5,$FD,$A7,$F1,$BF,$EF,$7F,$0B
	.dc.b	$FE,$E7,$F9,$DF,$F7,$C2,$FD,$A3,$FC,$7F,$EA,$BF,$5E,$FA,$1E,$57
	.dc.b	$F8,$FF,$C7,$1F,$E3,$DF,$81,$7E,$32,$FC,$25,$F9,$CB,$F7,$CF,$ED
	.dc.b	$9F,$AD,$7F,$13,$EC,$F7,$9F,$E8,$4F,$CC,$BF,$C2,$7F,$1B,$FC,$52
	.dc.b	$3F,$D0,$DA,$7F,$86,$F4,$EF,$1F,$E1,$AF,$CF,$3F,$D1,$31,$DB,$ED
	.dc.b	$1F,$81,$3F,$49,$7E,$6A,$FC,$0B,$FD,$19,$FB,$4F,$F1,$6C,$3F,$D1
	.dc.b	$27,$EF,$3E,$63,$F4,$BD,$5F,$E8,$8F,$BE,$1F,$E8,$8F,$CC,$7F,$D0
	.dc.b	$FF,$87,$7F,$A1,$AF,$FD,$0D,$DD,$FC,$23,$F4,$4F,$E0,$FE,$CF,$F6
	.dc.b	$D9,$FF,$7A,$9F,$DB,$F8,$FF,$90,$77,$FF,$89,$23,$F8,$3F,$C3,$FF
	.dc.b	$6C,$7E,$5F,$E5,$3F,$EF,$BF,$C2,$3F,$11,$1F,$A8,$7E,$7A,$F1,$34
	.dc.b	$FD,$2E,$FA,$BF,$77,$BF,$90,$63,$F2,$CC,$FC,$BE,$BF,$18,$DF,$E5
	.dc.b	$7C,$3C,$AE,$3F,$B5,$72,$FB,$7E,$7F,$3B,$D3,$F3,$BE,$BE,$2F,$6F
	.dc.b	$DF,$3B,$FE,$75,$E0,$FC,$EB,$C3,$E4,$78,$BE,$0B,$C7,$F9,$D7,$93
	.dc.b	$F2,$3F,$2F,$9F,$E6,$FD,$6B,$CF,$F9,$5F,$A3,$EF,$BD,$3F,$A8,$7A
	.dc.b	$BF,$4A,$F5,$FD,$5F,$B3,$E7,$FD,$BF,$1F,$EE,$FD,$57,$DF,$F9,$DF
	.dc.b	$C0,$FD,$67,$E0,$FC,$F7,$C2,$FB,$3F,$87,$FA,$2F,$C4,$F9,$1F,$8B
	.dc.b	$E9,$FC,$6F,$D1,$3E,$3F,$E9,$FF,$23,$F6,$CF,$93,$FB,$7F,$CA,$FC
	.dc.b	$AB,$E5,$FD,$D7,$CC,$FC,$8B,$E6,$FC,$9F,$CE,$FC,$F7,$E7,$FE,$A1
	.dc.b	$F4,$3E,$AF,$E8,$FC,$EF,$D2,$FD,$DB,$E9,$FC,$7F,$D4,$F7,$FE,$AF
	.dc.b	$C9,$FD,$6F,$D2,$3E,$BF,$E7,$5F,$63,$F2,$7F,$C4,$77,$7E,$0F,$D9
	.dc.b	$FE,$62,$7F,$50,$3F,$CD,$7E,$E7,$E0,$AF,$6F,$E1,$30,$B3,$FB,$4C
	.dc.b	$6B,$FE,$07,$3F,$C8,$17,$D7,$3C,$E9,$AD,$77,$B7,$0D,$F8,$F0,$72
	.dc.b	$E1,$E7,$C5,$D3,$8F,$AF,$27,$6E,$5E,$FC,$DE,$0F,$D4,$7C,$3D,$3E
	.dc.b	$2E,$AF,$1F,$5F,$93,$B3,$CB,$DB,$E6,$EE,$F3,$F7,$FA,$3C,$0F,$4F
	.dc.b	$83,$EA,$E8,$F5,$F8,$5E,$CF,$0F,$DB,$E2,$7B,$BC,$5F,$7F,$8D,$F0
	.dc.b	$3C,$7F,$83,$E4,$7C,$2F,$27,$E1,$F9,$5F,$13,$CB,$F8,$BE,$67,$C6
	.dc.b	$F3,$7E,$3F,$9D,$F2,$3C,$FF,$93,$E8,$7C,$AF,$47,$E5,$FA,$5F,$33
	.dc.b	$D3,$F9,$BE,$A7,$CE,$F5,$7E,$7F,$AD,$F4,$3D,$7F,$A3,$EC,$7D,$2F
	.dc.b	$67,$E9,$FB,$5F,$53,$DB,$FA,$BE,$E7,$D6,$F7,$7E,$BF,$D0,$FD,$8F
	.dc.b	$7F,$F8,$57,$C2,$FD,$9F,$63,$FC,$22,$BF,$15,$FF,$0B,$B4,$FF,$52
	.dc.b	$7C,$1F,$F8,$86,$7F,$11,$1F,$E2,$E3,$FC,$9F,$FA,$C7,$F9,$42,$5E
	.dc.b	$E3,$F5,$CF,$F3,$3F,$E3,$FF,$E0,$BF,$C0,$5F,$04,$F5,$37,$5A,$FB
	.dc.b	$F5,$6B,$E8,$9F,$B8,$B8,$3F,$DF,$3F,$BB,$BE,$61,$FA,$93,$F5,$4F
	.dc.b	$F7,$35,$EF,$E9,$17,$BA,$D7,$CF,$BD,$FF,$F7,$86,$3F,$D3,$BE,$81
	.dc.b	$F9,$7B,$CF,$7E,$E9,$BF,$D0,$BC,$1D,$EB,$DD,$3F,$14,$7C,$1B,$ED
	.dc.b	$7F,$94,$7E,$5C,$FC,$CB,$FA,$CF,$E7,$3F,$E4,$1A,$D7,$8C,$AD,$7B
	.dc.b	$1F,$ED,$BF,$11,$F6,$BF,$D8,$BE,$BB,$FD,$4B,$F6,$AF,$DE,$9D,$4F
	.dc.b	$C1,$77,$C6,$BD,$D6,$87,$ED,$BF,$77,$CA,$7C,$8F,$F4,$2F,$C4,$B3
	.dc.b	$E6,$FF,$9A,$E3,$FC,$DF,$5B,$F8,$7B,$1F,$EB,$F8,$FC,$13,$AF,$3B
	.dc.b	$FC,$E5,$3F,$E7,$6F,$8A,$7D,$6B,$E1,$9F,$9E,$4F,$9D,$7E,$96,$FB
	.dc.b	$D6,$1C,$2F,$77,$FE,$C6,$A7,$EA,$AF,$CA,$3F,$35,$D3,$FE,$84,$6B
	.dc.b	$FE,$68,$7D,$DC,$5F,$E8,$53,$86,$FF,$ED,$78,$F8,$33,$FE,$72,$E2
	.dc.b	$E9,$C1,$FE,$78,$9F,$F1,$84,$FF,$27,$7F,$B7,$3F,$E7,$87,$F0,$CE
	.dc.b	$7E,$0F,$A7,$B9,$D7,$F3,$EE,$DF,$97,$F7,$FC,$7B,$C1,$F6,$5E,$1F
	.dc.b	$AA,$FE,$65,$A7,$AD,$F2,$7D,$5F,$84,$FD,$66,$9F,$EF,$AD,$3F,$D4
	.dc.b	$35,$F0,$7F,$68,$FE,$75,$F8,$DD,$9F,$37,$E2,$7E,$93,$89,$FE,$C4
	.dc.b	$FC,$17,$E4,$79,$7F,$E7,$EF,$8A,$7C,$73,$F5,$37,$CB,$3F,$59,$DF
	.dc.b	$E3,$2B,$DD,$6B,$F7,$EF,$A5,$FB,$57,$CB,$FC,$3B,$E6,$6D,$FD,$34
	.dc.b	$7D,$07,$D4,$7F,$C7,$AF,$D0,$7F,$D4,$2F,$CE,$6F,$8F,$C6,$3E,$6C
	.dc.b	$6B,$B3,$1D,$99,$EC,$D7,$66,$FB,$38,$76,$71,$EC,$E5,$D3,$CF,$6E
	.dc.b	$9B,$75,$D3,$CD,$F8,$AF,$E2,$A3,$FC,$46,$D8,$FF,$5B,$3F,$8E,$F2
	.dc.b	$FF,$D9,$CF,$E4,$3F,$05,$DF,$DB,$7F,$B5,$1F,$F5,$BC,$7E,$53,$F2
	.dc.b	$8D,$CF,$FB,$63,$F5,$3D,$E1,$FE,$30,$77,$1F,$ED,$CF,$90,$DC,$FF
	.dc.b	$8D,$3D,$1D,$BF,$E6,$47,$6F,$F8,$E3,$F1,$9D,$BF,$EC,$AC,$7E,$77
	.dc.b	$86,$7F,$43,$E4,$7A,$3D,$2F,$D1,$7D,$CF,$B9,$B7,$EE,$7F,$4D,$FB
	.dc.b	$66,$EB,$FD,$D9,$A6,$DB,$5B,$7F,$DE,$15,$B4,$3E,$D7,$83,$87,$E2
	.dc.b	$C7,$EC,$FF,$43,$E1,$3F,$83,$CB,$F6,$FE,$6E,$A3,$FD,$F9,$E6,$FD
	.dc.b	$08,$FF,$80,$3C,$DF,$59,$B9,$7D,$67,$FC,$B6,$BF,$27,$E0,$FF,$97
	.dc.b	$BF,$35,$E0,$D3,$81,$FE,$62,$F5,$76,$FF,$A4,$BA,$76,$D3,$67,$FA
	.dc.b	$5B,$A7,$6B,$9F,$6D,$36,$39,$F6,$E8,$FC,$3E,$9F,$C5,$75,$FC,$6B
	.dc.b	$FC,$5F,$8B,$F2,$6D,$9F,$C9,$BF,$C6,$F8,$77,$5F,$CD,$E7,$FD,$53
	.dc.b	$C1,$FD,$D6,$3F,$8E,$F2,$FE,$0E,$FF,$91,$1C,$23,$EB,$FD,$97,$67
	.dc.b	$EC,$9F,$C8,$27,$83,$8F,$CA,$78,$3F,$C8,$AF,$B7,$8B,$EA,$8E,$3F
	.dc.b	$9D,$7E,$9B,$0F,$D6,$CF,$EB,$BE,$23,$F5,$FE,$37,$EC,$3C,$6F,$D8
	.dc.b	$F8,$DF,$B2,$F1,$BF,$D2,$7C,$BF,$6B,$44,$63,$83,$F5,$AF,$FA,$63
	.dc.b	$97,$F9,$E4,$EA,$7F,$D9,$3E,$17,$1A,$FF,$4E,$BF,$D7,$67,$52,$F0
	.dc.b	$B8,$D7,$EC,$E3,$F8,$0B,$5F,$E0,$2F,$FA,$9B,$9B,$71,$FB,$33,$65
	.dc.b	$36,$7F,$B6,$BC,$1D,$BF,$B5,$9F,$E7,$F7,$FE,$F1,$E2,$FC,$01,$FB
	.dc.b	$7F,$C4,$7E,$8F,$85,$F0,$5B,$7F,$C3,$1F,$AB,$E1,$7C,$07,$F6,$71
	.dc.b	$F8,$06,$F3,$F3,$BF,$D9,$F4,$ED,$FD,$95,$7F,$1E,$AF,$BF,$E1,$7C
	.dc.b	$57,$F0,$3E,$16,$EB,$F9,$B5,$73,$7F,$9C,$61,$7A,$FE,$8F,$8F,$FE
	.dc.b	$CA,$F0,$FE,$C3,$6F,$F9,$1B,$FE,$CE,$F0,$B7,$1F,$EF,$EF,$0B,$6D
	.dc.b	$36,$0F,$F0,$26,$3F,$CC,$2B,$F0,$51,$F9,$E6,$E3,$FC,$7D,$FC,$24
	.dc.b	$EE,$79,$78,$3D,$9F,$DA,$1F,$CF,$27,$EE,$4E,$E3,$F9,$54,$7D,$6F
	.dc.b	$D1,$3E,$AC,$FF,$2B,$7E,$7C,$D8,$FE,$3F,$E1,$7E,$36,$7F,$91,$C7
	.dc.b	$F4,$67,$F6,$93,$FB,$46,$E1,$C6,$7F,$6D,$C6,$9F,$D8,$F8,$1F,$A1
	.dc.b	$6D,$FF,$4C,$7F,$65,$5D,$4F,$E5,$FE,$06,$DB,$3B,$7F,$99,$F8,$3F
	.dc.b	$A6,$F2,$3F,$AF,$F9,$1F,$D8,$3C,$8F,$EC,$3C,$5F,$D8,$BE,$49,$F0
	.dc.b	$AF,$D1,$5F,$99,$BF,$6B,$7E,$E0,$F6,$7F,$A7,$61,$C2,$F7,$5A,$CB
	.dc.b	$FA,$97,$8F,$FD,$BB,$83,$FD,$8F,$CD,$FC,$AD,$FB,$C3,$8B,$FD,$99
	.dc.b	$3F,$E1,$E7,$FD,$9D,$E3,$FF,$25,$7F,$DA,$1E,$3F,$F3,$EA,$E1,$7F
	.dc.b	$C0,$1F,$F2,$D4,$FF,$B5,$7C,$7E,$13,$F8,$29,$FE,$8B,$3F,$ED,$9E
	.dc.b	$BE,$D3,$F8,$FF,$1F,$75,$FE,$69,$37,$1F,$BD,$5D,$7E,$05,$E3,$F0
	.dc.b	$7C,$98,$7F,$72,$3F,$B8,$F4,$FD,$6F,$13,$FB,$2F,$8D,$A6,$D7,$A1
	.dc.b	$FE,$D3,$E3,$7B,$DF,$C6,$C7,$A1,$AF,$E9,$71,$F9,$11,$FB,$EB,$FF
	.dc.b	$2F,$61,$FE,$B9,$F1,$BB,$47,$F7,$3B,$EC,$BC,$5E,$CF,$EF,$A3,$9B
	.dc.b	$FC,$03,$0B,$E2,$FD,$1F,$3B,$F6,$9D,$BE,$E6,$DF,$D0,$7F,$71,$E3
	.dc.b	$6E,$1E,$36,$DA,$6C,$FD,$EE,$3F,$E8,$8A,$FC,$24,$7F,$01,$D3,$FC
	.dc.b	$EB,$CF,$F8,$39,$FE,$E1,$5E,$E7,$EE,$57,$E2,$AB,$FC,$F7,$C9,$B8
	.dc.b	$FE,$E1,$5B,$7F,$10,$7E,$3E,$BF,$CE,$35,$FD,$9E,$7F,$D3,$39,$FC
	.dc.b	$EA,$F1,$37,$4F,$CB,$3F,$3E,$76,$7C,$9A,$47,$FB,$9E,$BE,$B4,$62
	.dc.b	$7B,$37,$C4,$FC,$3D,$7F,$72,$AF,$CC,$BF,$E2,$AF,$37,$F4,$0F,$EF
	.dc.b	$23,$DC,$D7,$F2,$75,$D7,$C4,$FD,$35,$FF,$9C,$63,$FD,$2F,$3F,$A0
	.dc.b	$F3,$71,$E6,$E7,$CD,$D7,$9B,$BF,$35,$C3,$F1,$B1,$FE,$E2,$FF,$A9
	.dc.b	$5F,$EE,$7D,$FF,$D6,$1F,$EE,$BD,$FB,$A3,$B7,$FB,$77,$07,$EE,$5F
	.dc.b	$E6,$3A,$7D,$76,$BE,$A7,$EB,$C7,$E6,$9F,$D9,$47,$EE,$1A,$FC,$C7
	.dc.b	$F6,$31,$D6,$3C,$47,$9F,$CD,$FE,$73,$65,$D9,$ED,$6C,$FF,$0F,$5D
	.dc.b	$9E,$D7,$FD,$31,$5C,$7B,$BA,$76,$E7,$B7,$91,$DF,$B7,$87,$4F,$53
	.dc.b	$C5,$CF,$F1,$FC,$D7,$77,$CD,$6E,$BF,$C7,$DF,$F2,$C6,$FF,$C7,$DF
	.dc.b	$D3,$C7,$F1,$86,$BF,$8C,$7F,$53,$1F,$EF,$07,$EC,$C7,$F1,$26,$BF
	.dc.b	$C9,$17,$17,$8F,$B3,$C4,$7D,$D8,$FA,$35,$CE,$BC,$C7,$E0,$7C,$CF
	.dc.b	$96,$7C,$43,$FD,$DF,$DB,$FC,$E5,$FE,$3F,$C1,$FD,$6F,$4F,$C5,$47
	.dc.b	$F0,$D9,$D6,$9F,$89,$8F,$F6,$05,$FE,$E5,$AF,$E7,$17,$6E,$F0,$E6
	.dc.b	$78,$FB,$7F,$25,$E6,$7E,$7F,$CC,$FD,$07,$8B,$FA,$1F,$D4,$66,$5F
	.dc.b	$99,$FD,$57,$F0,$63,$F4,$47,$E1,$90,$99,$A7,$51,$91,$FB,$D7,$F5
	.dc.b	$78,$7D,$A4,$27,$E1,$D9,$70,$66,$FA,$78,$7F,$AD,$A3,$E5,$FC,$67
	.dc.b	$F5,$91,$DF,$1F,$1B,$5F,$0D,$79,$7F,$20,$EB,$83,$1F,$24,$62,$3C
	.dc.b	$DA,$77,$3F,$75,$FB,$9F,$E9,$0F,$6E,$1E,$DC,$11,$88,$F8,$BA,$7F
	.dc.b	$75,$8E,$98,$DC,$3B,$5F,$0D,$3F,$84,$57,$B3,$1F,$EF,$4B,$4F,$E1
	.dc.b	$0F,$AF,$2E,$87,$BF,$2E,$D9,$FB,$33,$B9,$FB,$B3,$F2,$E5,$D8,$FE
	.dc.b	$29,$FB,$EF,$EF,$8F,$46,$3F,$72,$51,$8C,$7F,$C1,$CF,$AA,$33,$7E
	.dc.b	$A8,$F4,$5F,$F8,$19,$AF,$F3,$A7,$D5,$D3,$FE,$13,$7D,$55,$DB,$0F
	.dc.b	$53,$E0,$CB,$D4,$DC,$3D,$55,$FC,$D4,$FE,$A3,$DB,$FC,$58,$F4,$DF
	.dc.b	$F8,$8D,$F4,$57,$F0,$8A,$E7,$1F,$75,$D6,$DF,$CD,$E3,$CF,$97,$F4
	.dc.b	$47,$CF,$97,$F8,$11,$F3,$9F,$8B,$CF,$B6,$7D,$0F,$B7,$1F,$D2,$4F
	.dc.b	$EC,$BE,$6F,$F2,$A3,$E4,$1E,$1C,$BF,$DD,$D9,$FE,$55,$AF,$E9,$EF
	.dc.b	$FC,$79,$3F,$D4,$CF,$ED,$FF,$B8,$FF,$34,$3E,$91,$CC,$F9,$67,$96
	.dc.b	$8F,$F5,$B3,$FB,$BF,$EE,$9F,$CF,$0F,$B4,$78,$CF,$A6,$71,$3F,$B1
	.dc.b	$1F,$DF,$BF,$41,$FE,$86,$7E,$04,$F5,$1F,$68,$F0,$1F,$DA,$4F,$F0
	.dc.b	$2F,$AD,$FD,$28,$FC,$41,$F0,$4F,$C0,$9E,$73,$FB,$81,$FE,$0C,$FC
	.dc.b	$6F,$FA,$89,$F8,$F3,$E4,$1F,$88,$3D,$C7,$F7,$73,$FC,$27,$FB,$AF
	.dc.b	$F5,$83,$F2,$C7,$CF,$3F,$1E,$7C,$53,$FB,$F1,$FE,$1A,$F4,$FF,$AF
	.dc.b	$9F,$9D,$3E,$B1,$F9,$63,$E6,$1F,$E0,$53,$FC,$43,$EF,$7F,$66,$3F
	.dc.b	$48,$7D,$D3,$F3,$A7,$D3,$3F,$C1,$87,$F8,$A3,$E3,$7F,$B6,$9F,$AB
	.dc.b	$3F,$08,$7E,$90,$FB,$47,$F8,$4C,$FF,$17,$FE,$5D,$FD,$D0,$FD,$91
	.dc.b	$F8,$B3,$F5,$67,$E0,$4F,$F0,$D1,$FE,$36,$FD,$6F,$FB,$D9,$FB,$93
	.dc.b	$F2,$47,$EC,$8F,$C4,$1F,$E2,$13,$FC,$7B,$FB,$CF,$F8,$08,$FE,$00
	.dc.b	$7E,$64,$FD,$C9,$F8,$F3,$FC,$50,$7F,$91,$3E,$17,$FC,$12,$7F,$07
	.dc.b	$3F,$40,$7F,$00,$3F,$2C,$7F,$8B,$CF,$F2,$5F,$C4,$FF,$84,$0F,$E1
	.dc.b	$C7,$E9,$CF,$E0,$E7,$E7,$4F,$F1,$B1,$FE,$52,$F5,$BF,$C2,$E7,$F1
	.dc.b	$53,$F5,$C7,$F0,$E3,$F4,$87,$F8,$F4,$FF,$2C,$F9,$BF,$E1,$E3,$F8
	.dc.b	$99,$FA,$9F,$99,$FE,$3C,$7E,$DC,$FF,$22,$1F,$E6,$0F,$C7,$FF,$C4
	.dc.b	$A7,$F1,$A3,$F6,$1E,$C7,$F2,$53,$F7,$C7,$F9,$2C,$FF,$33,$FB,$FF
	.dc.b	$E2,$C3,$F9,$09,$FB,$7F,$DC,$3F,$96,$1F,$C1,$4F,$F2,$91,$FE,$6E
	.dc.b	$FD,$9F,$FC,$66,$7F,$28,$3F,$7D,$E9,$FF,$33,$3F,$86,$1F,$E5,$93
	.dc.b	$FC,$EB,$F0,$3F,$E3,$A3,$F9,$79,$FC,$17,$F3,$BF,$E7,$47,$F1,$33
	.dc.b	$FC,$C0,$7F,$9F,$3E,$6B,$FC,$82,$7F,$36,$3F,$86,$7C,$5F,$F4,$13
	.dc.b	$F8,$D1,$FE,$67,$3F,$D0,$7F,$09,$FE,$48,$3F,$9E,$9F,$C4,$FE,$53
	.dc.b	$FA,$41,$FC,$84,$FF,$37,$1F,$E8,$AF,$5F,$FC,$9E,$7F,$44,$3F,$8D
	.dc.b	$78,$FF,$D3,$CF,$E5,$07,$F9,$D4,$FF,$49,$AF,$51,$E4,$42,$1F,$EF
	.dc.b	$AF,$B9,$FE,$E3,$07,$FA,$12,$F9,$5F,$ED,$8D,$F2,$3F,$B6,$2F,$CD
	.dc.b	$BF,$D4,$38,$F7,$7C,$30,$FD,$6B,$BE,$3E,$18,$7E,$3B,$E0,$3B,$FE
	.dc.b	$E1,$D8,$FE,$95,$1F,$E6,$AC,$3F,$DE,$8D,$7F,$DF,$BA,$7B,$7E,$17
	.dc.b	$AE,$1E,$EA,$7D,$E7,$B0,$F5,$FE,$F9,$EA,$3A,$FE,$3F,$E8,$93,$E8
	.dc.b	$FD,$CB,$9D,$FC,$9F,$C0,$3E,$29,$F1,$3F,$65,$F8,$67,$9B,$E2,$B9
	.dc.b	$1F,$03,$E2,$B8,$E2,$FD,$57,$E4,$9E,$CF,$C8,$3E,$79,$F3,$BE,$33
	.dc.b	$C4,$7C,$8F,$8C,$E3,$F7,$9F,$40,$F8,$7F,$8F,$7D,$63,$EA,$FC,$6F
	.dc.b	$A4,$F9,$FF,$1B,$E0,$FB,$BF,$B6,$7D,$AF,$88,$FB,$27,$CC,$F9,$6F
	.dc.b	$80,$7D,$6F,$8E,$F3,$FE,$F5,$F7,$8F,$A3,$F0,$1F,$17,$EA,$7F,$02
	.dc.b	$7D,$4F,$98,$F9,$1F,$31,$F2,$7E,$63,$E5,$7C,$4F,$CB,$F9,$8F,$99
	.dc.b	$F2,$DF,$37,$E5,$7E,$77,$CA,$FC,$FF,$95,$FA,$1F,$1B,$F4,$7D,$DF
	.dc.b	$A5,$E4,$7E,$2B,$F1,$77,$81,$7A,$31,$CB,$9E,$3D,$71,$6F,$8B,$87
	.dc.b	$17,$1E,$2E,$5C,$5C,$F8,$BA,$71,$75,$E2,$ED,$C5,$DF,$8B,$C1,$C5
	.dc.b	$E1,$E2,$FC,$DF,$EE,$DE,$27,$FD,$FA,$B8,$73,$C3,$AE,$1D,$F0,$F0
	.dc.b	$E1,$E3,$C3,$CB,$87,$9F,$0F,$4E,$1E,$BC,$3D,$B8,$7B,$F0,$F8,$38
	.dc.b	$7C,$3C,$3E,$2E,$1F,$D6,$FE,$31,$E3,$44,$2B,$E5,$70,$65,$FF,$50
	.dc.b	$BA,$DF,$9E,$FD,$37,$EB,$BF,$6D,$FB,$EF,$E0,$DF,$C3,$BF,$8B,$7F
	.dc.b	$1E,$FE,$4D,$FC,$BB,$F9,$B6,$FE,$07,$FA,$87,$9D,$F1,$FD,$77,$FB
	.dc.b	$51,$81,$AF,$AD,$FF,$6B,$F8,$FE,$B8,$F7,$C7,$FB,$42,$F9,$DF,$F1
	.dc.b	$08,$FF,$68,$D7,$F0,$26,$AE,$1F,$E0,$F1,$87,$C7,$F0,$5F,$C8,$F7
	.dc.b	$F2,$E3,$F2,$F4,$AF,$CA,$3E,$7F,$E6,$FF,$CA,$27,$E6,$FE,$32,$E4
	.dc.b	$FE,$42,$B9,$2F,$E6,$7F,$C2,$AB,$DB,$83,$D1,$FF,$5A,$F0,$7F,$28
	.dc.b	$9F,$7E,$FD,$78,$3F,$9C,$C7,$FB,$BA,$FE,$DB,$EC,$9F,$16,$DF,$C8
	.dc.b	$3D,$8F,$F5,$FF,$BF,$C0,$E9,$EF,$F8,$70,$F5,$0F,$6F,$FC,$F6,$FE
	.dc.b	$2B,$4A,$FF,$3A,$21,$93,$F9,$EC,$7E,$42,$7F,$92,$C2,$B8,$1F,$D4
	.dc.b	$1F,$36,$3F,$D4,$30,$FC,$BE,$3C,$33,$EC,$6F,$E5,$39,$EB,$E1,$3F
	.dc.b	$AC,$FE,$EA,$DF,$F9,$9B,$F3,$72,$F5,$BF,$45,$F6,$FE,$F3,$F5,$5F
	.dc.b	$D2,$7D,$1F,$AC,$FD,$1F,$37,$E9,$1F,$C2,$F1,$5D,$3F,$B2,$73,$DD
	.dc.b	$FF,$B3,$BF,$D3,$AB,$D7,$1F,$DA,$9F,$A9,$87,$6D,$BF,$65,$AF,$5D
	.dc.b	$BF,$B7,$78,$3D,$E3,$FB,$2D,$FE,$85,$FD,$AF,$E5,$BF,$A5,$C3,$F7
	.dc.b	$5F,$B8,$7D,$13,$AB,$FB,$E7,$F1,$29,$76,$FF,$70,$5B,$DE,$EC,$77
	.dc.b	$E7,$C0,$D7,$16,$FC,$1E,$1D,$9C,$7B,$B9,$72,$73,$F0,$FA,$71,$75
	.dc.b	$F1,$7F,$BE,$FE,$E7,$F8,$53,$CB,$C2,$7C,$A7,$91,$FC,$7B,$E4,$6E
	.dc.b	$DC,$B6,$E7,$B7,$4D,$BA,$ED,$DB,$6E,$FB,$78,$36,$F0,$ED,$E2,$DB
	.dc.b	$C7,$B7,$F8,$43,$F0,$2F,$CC,$1E,$43,$E3,$9E,$3D,$BF,$74,$F8,$DB
	.dc.b	$B7,$1D,$B9,$6D,$CF,$6E,$9B,$75,$DB,$B6,$DD,$F6,$F0,$6D,$E1,$DB
	.dc.b	$C5,$B7,$F8,$8B,$ED,$BF,$54,$7C,$63,$EC,$9E,$3D,$BF,$9A,$BE,$36
	.dc.b	$ED,$C7,$6E,$5B,$73,$DB,$A6,$DD,$76,$ED,$B7,$7F,$29,$F0,$70,$83
	.dc.b	$44,$7F,$BD,$5F,$E4,$9A,$7F,$04,$5E,$67,$0A,$E5,$18,$71,$F3,$BC
	.dc.b	$F8,$7D,$DE,$77,$F8,$9A,$BF,$A6,$F8,$FD,$FC,$FE,$58,$79,$25,$1E
	.dc.b	$1F,$41,$F3,$F0,$9F,$43,$E7,$8C,$44,$AF,$E8,$5A,$FC,$8F,$4F,$C1
	.dc.b	$D1,$FE,$5E,$97,$A9,$C2,$FA,$E3,$F0,$63,$11,$E0,$97,$AB,$E6,$D3
	.dc.b	$E3,$3F,$17,$D7,$FB,$5E,$BF,$B5,$ED,$EC,$72,$CB,$E6,$3D,$31,$F9
	.dc.b	$DE,$CF,$3C,$BF,$C8,$19,$7C,$87,$F7,$DB,$7F,$78,$5E,$DF,$C9,$EF
	.dc.b	$F5,$7B,$9C,$4F,$AD,$EE,$F7,$A3,$DD,$FF,$2B,$DF,$D9,$F7,$BE,$97
	.dc.b	$BF,$F0,$FE,$03,$C4,$7D,$FF,$81,$E4,$71,$FA,$8F,$CB,$7C,$1F,$D9
	.dc.b	$7E,$57,$C2,$7A,$4F,$C3,$7C,$2F,$84,$F0,$7C,$3F,$E7,$7E,$1F,$F0
	.dc.b	$0F,$D1,$F8,$8F,$80,$7E,$37,$DD,$F4,$1F,$17,$D1,$FD,$1F,$B3,$FE
	.dc.b	$6C,$7F,$D0,$FE,$1F,$F0,$5E,$9F,$B1,$D7,$5F,$B5,$3F,$D9,$FA,$7F
	.dc.b	$16,$D7,$F3,$BE,$37,$F2,$0F,$80,$FF,$48,$CF,$F6,$DC,$7F,$D0,$6F
	.dc.b	$F5,$FE,$6F,$EF,$59,$6F,$1E,$F3,$FD,$8D,$FC,$2A,$F8,$9F,$C2,$37
	.dc.b	$2F,$D0,$AF,$8A,$EF,$3E,$B7,$FD,$5F,$19,$AE,$B1,$F0,$7E,$2F,$99
	.dc.b	$FA,$4D,$3F,$9B,$FC,$67,$A2,$3E,$23,$C8,$FC,$75,$9F,$95,$F0,$9E
	.dc.b	$43,$94,$7F,$71,$D3,$FB,$EA,$8F,$9D,$F0,$9E,$43,$99,$F4,$FE,$37
	.dc.b	$DA,$72,$F8,$EF,$92,$FD,$57,$D6,$78,$8F,$B5,$F1,$FF,$10,$F0,$FC
	.dc.b	$87,$D0,$7E,$E3,$F0,$8F,$49,$F8,$1F,$91,$F9,$67,$A3,$E4,$BE,$BB
	.dc.b	$F8,$37,$E4,$9F,$00,$FC,$47,$C3,$FD,$23,$DF,$F0,$FF,$79,$FC,$53
	.dc.b	$F4,$0F,$AE,$6B,$8B,$EE,$BF,$91,$F9,$3F,$C2,$BF,$91,$7E,$B9,$F7
	.dc.b	$8F,$96,$7D,$03,$F3,$1F,$29,$F8,$C7,$F3,$0F,$DE,$3F,$0A,$7D,$23
	.dc.b	$EB,$9F,$9F,$F9,$5F,$C9,$BF,$9F,$7F,$0A,$7E,$30,$FB,$27,$DE,$3F
	.dc.b	$4D,$F2,$DF,$9A,$7F,$4C,$FE,$30,$FC,$99,$F8,$03,$F0,$A7,$EB,$7E
	.dc.b	$5F,$F4,$2F,$EB,$5F,$C9,$9F,$9A,$3F,$0E,$7E,$30,$FD,$A7,$CC,$7E
	.dc.b	$A1,$FD,$A3,$F9,$A3,$F4,$27,$E3,$8F,$C9,$9F,$BB,$F9,$9F,$D7,$BF
	.dc.b	$BB,$7F,$42,$7E,$A0,$FC,$A9,$F9,$A3,$F8,$17,$CD,$7E,$D9,$FE,$04
	.dc.b	$FE,$A0,$FD,$79,$F9,$C3,$F4,$27,$F0,$9F,$9B,$FD,$EB,$FC,$25,$FD
	.dc.b	$79,$FB,$63,$F4,$67,$EA,$0F,$E2,$1F,$39,$FC,$11,$FE,$20,$FE,$D8
	.dc.b	$FD,$E9,$FA,$A3,$F5,$E7,$F1,$7F,$9D,$FE,$16,$FF,$17,$7F,$7A,$7F
	.dc.b	$04,$3F,$62,$7E,$D8,$FE,$3D,$F3,$DF,$C4,$9F,$E3,$CF,$F0,$43,$F8
	.dc.b	$59,$FB,$83,$F7,$A7,$F2,$5F,$9F,$FE,$32,$FF,$25,$7F,$85,$9F,$C4
	.dc.b	$8F,$DF,$9F,$C1,$0F,$E5,$9F,$41,$FC,$81,$FE,$58,$FF,$12,$3F,$8C
	.dc.b	$9F,$C1,$8F,$E1,$67,$F3,$3F,$95,$FE,$4E,$FF,$33,$7F,$8C,$9F,$C8
	.dc.b	$0F,$E1,$A7,$F1,$23,$F9,$D7,$CA,$7F,$2E,$7F,$9D,$3F,$C8,$07,$D0
	.dc.b	$FF,$2C,$6F,$27,$F3,$A7,$F9,$59,$FC,$60,$FF,$48,$F7,$17,$D2,$7F
	.dc.b	$38,$5F,$1B,$FC,$E2,$75,$5D,$4F,$E6,$F3,$FE,$0C,$F1,$FF,$C3,$13
	.dc.b	$CF,$E1,$BF,$9B,$FD,$1F,$F4,$4A,$FE,$A0,$FF,$36,$3C,$B7,$FE,$10
	.dc.b	$A4,$62,$3F,$C2,$91,$8B,$FE,$A3,$3F,$CC,$0F,$49,$E8,$F2,$C4,$2C
	.dc.b	$F4,$C7,$D5,$3F,$CF,$0F,$EC,$87,$A6,$BF,$C6,$F3,$E2,$C7,$FA,$7F
	.dc.b	$D2,$7F,$5C,$AF,$ED,$0F,$F4,$E3,$CD,$8B,$CE,$FF,$40,$C6,$22,$E1
	.dc.b	$FD,$A2,$3F,$A3,$9F,$40,$F4,$65,$E7,$38,$E5,$F3,$E3,$FA,$89,$FD
	.dc.b	$E4,$F3,$4F,$F9,$37,$24,$62,$1E,$57,$FA,$4B,$E9,$7F,$B9,$CF,$F7
	.dc.b	$F7,$FB,$29,$E8,$CB,$5B,$FF,$70,$8F,$ED,$E7,$F5,$C3,$F0,$A7,$9B
	.dc.b	$2F,$99,$8A,$3F,$0B,$E4,$88,$7F,$BB,$63,$5A,$7F,$7E,$9F,$F0,$91
	.dc.b	$E7,$CF,$CD,$B7,$F7,$CF,$A6,$FF,$04,$CF,$F8,$55,$FE,$F7,$F3,$3F
	.dc.b	$E0,$CC,$3F,$C1,$BF,$2B,$FE,$0E,$FA,$7F,$F3,$4E,$7F,$EC,$C5,$F4
	.dc.b	$FF,$E1,$07,$31,$FE,$ED,$A4,$3E,$43,$FB,$BA,$F9,$0E,$11,$F9,$F8
	.dc.b	$FF,$2F,$D7,$B6,$B8,$FC,$7F,$5F,$8F,$D7,$C4,$72,$F8,$9F,$03,$FA
	.dc.b	$7F,$97,$F3,$72,$7F,$88,$9F,$F0,$A9,$FE,$07,$C7,$FD,$C6,$FF,$81
	.dc.b	$E3,$F0,$47,$FB,$29,$D4,$7A,$5F,$F6,$9E,$27,$FD,$C6,$DF,$33,$C1
	.dc.b	$E2,$FF,$84,$A5,$E3,$78,$63,$C9,$DD,$9F,$03,$C8,$FE,$F7,$EA,$3D
	.dc.b	$F9,$7F,$B6,$96,$FF,$E3,$A7,$E1,$FB,$79,$F8,$EF,$F0,$FC,$3E,$3F
	.dc.b	$51,$CF,$83,$3C,$3C,$5E,$DF,$0B,$C3,$E1,$BA,$3F,$C3,$7E,$63,$FC
	.dc.b	$93,$8F,$F9,$55,$FF,$1D,$9F,$E3,$FD,$3F,$1A,$7D,$37,$FC,$8D,$C9
	.dc.b	$F2,$39,$11,$B9,$FC,$9E,$BF,$EF,$48,$F9,$71,$C7,$DC,$FF,$2A,$C6
	.dc.b	$27,$5E,$C7,$F8,$EA,$1E,$CE,$A3,$CD,$CF,$9F,$5F,$8B,$FC,$97,$C9
	.dc.b	$FF,$30,$FC,$C7,$F9,$8B,$E6,$3F,$CC,$7F,$31,$FE,$64,$F9,$9F,$F3
	.dc.b	$2F,$81,$FE,$66,$E9,$FE,$2F,$1E,$CE,$07,$0F,$D0,$E2,$FC,$D9,$FC
	.dc.b	$15,$C4,$FC,$2E,$1D,$DF,$F2,$0C,$39,$57,$1F,$07,$E2,$4E,$71,$F8
	.dc.b	$73,$D7,$C0,$DC,$FF,$28,$37,$97,$FA,$FB,$1E,$AB,$A3,$6F,$FA,$A0
	.dc.b	$F2,$4F,$F3,$08,$F2,$5F,$E2,$B0,$E3,$5F,$57,$0F,$F3,$A9,$FC,$1F
	.dc.b	$E0,$3C,$EF,$F0,$9E,$6D,$57,$E8,$32,$F8,$73,$FC,$37,$8B,$72,$E3
	.dc.b	$DC,$78,$27,$F8,$97,$7E,$BC,$0E,$6E,$6F,$FA,$F1,$F8,$AD,$15,$FC
	.dc.b	$67,$C0,$F0,$F8,$1E,$AE,$EF,$E9,$73,$F3,$1F,$8D,$BF,$D3,$73,$87
	.dc.b	$CC,$7F,$A8,$F7,$EE,$7F,$48,$6C,$3D,$5F,$99,$3E,$39,$FE,$B2,$7E
	.dc.b	$6C,$F9,$8F,$58,$77,$75,$EF,$D4,$62,$7E,$EF,$1F,$FA,$CA,$FF,$D7
	.dc.b	$CF,$FA,$90,$FE,$6D,$F0,$7F,$86,$7F,$9C,$7C,$27,$EA,$5C,$D7,$23
	.dc.b	$F4,$FB,$79,$B9,$35,$1F,$56,$3D,$3D,$DC,$E3,$F4,$F9,$7E,$5E,$FF
	.dc.b	$A2,$4F,$E1,$1F,$B5,$3F,$45,$FE,$91,$C9,$D3,$93,$F3,$9F,$97,$7E
	.dc.b	$8E,$BF,$A5,$A8,$C4,$21,$EB,$BE,$B9,$F5,$6F,$AE,$7D,$13,$EB,$9F
	.dc.b	$5D,$C7,$0E,$17,$BA,$D7,$9F,$D7,$AB,$BE,$68,$BC,$58,$F3,$F3,$E7
	.dc.b	$7F,$83,$FC,$5F,$E2,$B7,$FD,$3B,$97,$FC,$27,$E7,$FF,$7D,$8C,$47
	.dc.b	$F5,$8F,$0B,$D5,$7F,$E1,$7F,$1B,$8B,$FE,$19,$F1,$B8,$39,$E5,$FE
	.dc.b	$DD,$5C,$2F,$F2,$31,$FD,$C7,$0E,$BE,$37,$09,$CD,$6D,$FC,$7D,$6E
	.dc.b	$FF,$C4,$DD,$DB,$9C,$D7,$CD,$F3,$FD,$B1,$FD,$50,$C5,$FF,$BE,$B1
	.dc.b	$FE,$A4,$FE,$EC,$FA,$F3,$FE,$02,$ED,$F8,$0F,$F8,$13,$C7,$FE,$67
	.dc.b	$5C,$2B,$EE,$1C,$0F,$C1,$DF,$F8,$E6,$35,$D7,$C2,$FF,$C1,$BE,$37
	.dc.b	$CE,$7F,$C1,$FE,$3F,$6C,$3F,$80,$9F,$1A,$FB,$73,$6E,$3F,$7D,$8F
	.dc.b	$5B,$FF,$0B,$CF,$93,$CE,$FF,$24,$78,$FF,$69,$FF,$25,$70,$FD,$97
	.dc.b	$38,$75,$7F,$C9,$DE,$3E,$F1,$FF,$28,$F9,$FB,$7F,$81,$D6,$DF,$DD
	.dc.b	$BF,$5E,$BE,$8C,$7B,$DF,$F2,$C7,$87,$E3,$8F,$F2,$DF,$91,$D5,$79
	.dc.b	$1D,$5F,$E0,$16,$E4,$FD,$27,$9F,$F9,$8C,$7E,$C5,$62,$BE,$C4,$FF
	.dc.b	$8D,$BC,$9F,$CB,$BF,$E3,$8F,$27,$F2,$CE,$67,$E5,$3F,$65,$FD,$9B
	.dc.b	$F7,$6B,$EB,$39,$8F,$17,$17,$F7,$27,$F8,$09,$C0,$FF,$23,$79,$FD
	.dc.b	$9C,$9C,$27,$FC,$93,$CD,$E4,$8F,$C1,$DF,$F9,$F2,$FA,$47,$E2,$0E
	.dc.b	$FE,$7F,$ED,$BC,$FF,$DB,$F9,$FF,$B8,$F3,$FF,$73,$E1,$7E,$EB,$C1
	.dc.b	$FD,$DF,$3F,$F9,$BE,$3F,$86,$BF,$E7,$1F,$0B,$F2,$0E,$72,$F5,$BF
	.dc.b	$E7,$5A,$DE,$9F,$B5,$3C,$91,$99,$DC,$7E,$19,$FE,$2D,$1F,$E7,$D3
	.dc.b	$AC,$7E,$2B,$8F,$2F,$E0,$27,$F7,$53,$FE,$83,$F0,$37,$3F,$D0,$AB
	.dc.b	$F9,$07,$57,$AA,$3F,$69,$BF,$F8,$E2,$73,$87,$D3,$8F,$EB,$AF,$D7
	.dc.b	$9C,$BF,$B7,$9A,$FC,$B9,$DB,$F4,$1A,$FE,$DA,$37,$DD,$FC,$D2,$FF
	.dc.b	$C6,$7D,$DF,$B7,$7F,$C6,$B1,$FB,$17,$FC,$6F,$DB,$BB,$FF,$42,$F9
	.dc.b	$FB,$7F,$62,$7E,$23,$3F,$F5,$15,$7F,$40,$7E,$46,$7F,$81,$73,$3B
	.dc.b	$8F,$D8,$BF,$C8,$63,$FD,$55,$DB,$FC,$25,$FF,$56,$77,$7F,$03,$7F
	.dc.b	$D5,$DD,$DD,$6F,$F6,$27,$F4,$B8,$CD,$F5,$8F,$D8,$BD,$5F,$F6,$67
	.dc.b	$6F,$57,$FD,$9D,$DD,$FC,$3E,$FF,$A3,$DF,$EC,$38,$3F,$97,$E1,$FC
	.dc.b	$82,$7F,$CB,$3D,$BF,$D7,$1F,$F2,$DF,$6F,$F1,$F7,$33,$FA,$A7,$FC
	.dc.b	$BF,$E0,$6E,$3E,$D5,$F4,$8F,$04,$7F,$AC,$3C,$FD,$3F,$AE,$CF,$F5
	.dc.b	$17,$FE,$68,$9D,$CF,$F7,$23,$FB,$54,$7F,$9B,$3C,$FF,$14,$7E,$C8
	.dc.b	$FE,$19,$F9,$7F,$F7,$04,$FE,$C5,$FF,$39,$BF,$E2,$C7,$FC,$B4,$FF
	.dc.b	$67,$7F,$BD,$3F,$E5,$C6,$F8,$FF,$DE,$A3,$FA,$14,$72,$F0,$71,$E0
	.dc.b	$67,$C0,$D7,$81,$BF,$03,$87,$81,$C7,$BF,$97,$4F,$3D,$BF,$C0,$BF
	.dc.b	$88,$79,$CF,$36,$9F,$ED,$47,$CC,$FF,$A2,$97,$DC,$3E,$D9,$F6,$CF
	.dc.b	$B8,$7D,$BB,$EE,$1F,$1A,$C7,$8E,$1C,$2F,$74,$F8,$DD,$7A,$DF,$DA
	.dc.b	$7E,$56,$F5,$FF,$AF,$BE,$5A,$BF,$EC,$0C,$BE,$46,$FF,$E8,$36,$BF
	.dc.b	$E9,$A6,$BF,$EE,$FF,$96,$F1,$7C,$FF,$F3,$E9,$3D,$DF,$A1,$3C,$FD
	.dc.b	$6F,$F1,$67,$9B,$E9,$EE,$F8,$FE,$9F,$C8,$EE,$F9,$3D,$DF,$2B,$BB
	.dc.b	$E5,$F7,$7C,$CE,$EF,$9B,$DD,$F3,$BB,$BE,$7F,$77,$D0,$F3,$FC,$4F
	.dc.b	$6E,$0F,$D0,$F0,$7D,$1F,$1F,$CD,$E0,$78,$BF,$7C,$7D,$EB,$ED,$1F
	.dc.b	$57,$3E,$6D,$F7,$9F,$ED,$AB,$DD,$6A,$95,$AF,$94,$7E,$E9,$FE,$A9
	.dc.b	$ED,$FF,$80,$2D,$DF,$FD,$92,$FF,$A5,$4F,$F4,$D4,$6B,$5F,$E7,$1B
	.dc.b	$72,$7F,$9D,$75,$3F,$0D,$7E,$02,$F4,$5F,$AC,$3F,$04,$FF,$4A,$7E
	.dc.b	$08,$F9,$EE,$17,$BA,$D7,$A2,$AF,$F4,$D7,$E9,$EF,$B4,$7E,$10,$FC
	.dc.b	$23,$F2,$BF,$1E,$FC,$2B,$FB,$F5,$EE,$B5,$EB,$7F,$98,$65,$C6,$FD
	.dc.b	$8D,$F8,$77,$8B,$F0,$E7,$E2,$0F,$1D,$F8,$86,$EB,$53,$9F,$C4,$1F
	.dc.b	$BE,$BF,$15,$F4,$7E,$29,$F0,$FE,$5D,$F8,$AB,$F1,$ED,$D6,$BB,$BF
	.dc.b	$CE,$B2,$FC,$59,$FB,$DB,$F1,$A7,$C9,$BE,$5D,$F8,$D7,$EB,$7E,$ED
	.dc.b	$F8,$DF,$0B,$DD,$6B,$D4,$53,$F1,$B7,$D6,$FE,$07,$CC,$13,$E1,$33
	.dc.b	$EC,$7F,$89,$27,$FD,$AC,$7E,$0A,$76,$27,$9F,$9D,$BF,$63,$FD,$23
	.dc.b	$5F,$E2,$CA,$7B,$4E,$AF,$A3,$E0,$BF,$7E,$7E,$12,$FC,$A7,$FD,$39
	.dc.b	$7B,$AF,$AF,$5A,$FC,$FB,$3F,$BA,$72,$F8,$2F,$F3,$FD,$74,$3E,$ED
	.dc.b	$BF,$4B,$EF,$A7,$A2,$74,$79,$23,$93,$FE,$45,$33,$3F,$82,$7F,$0F
	.dc.b	$3C,$5E,$D1,$95,$F9,$A3,$F3,$CF,$5C,$72,$F1,$7D,$F3,$9F,$C7,$FD
	.dc.b	$BF,$3F,$E0,$F6,$FD,$FF,$6F,$70,$F0,$FD,$C7,$A3,$4F,$AB,$1E,$4F
	.dc.b	$96,$F9,$3F,$2F,$FC,$13,$EC,$FE,$3C,$63,$E2,$5F,$BF,$7C,$AF,$93
	.dc.b	$F9,$FF,$95,$AD,$F6,$F9,$CB,$F4,$73,$99,$C4,$7D,$5F,$B6,$18,$EA
	.dc.b	$BF,$73,$AF,$CA,$BA,$63,$70,$ED,$87,$CA,$CB,$EF,$FE,$DB,$CB,$2E
	.dc.b	$AF,$29,$D1,$FB,$4F,$CA,$F9,$B8,$0F,$95,$C7,$A1,$CD,$B4,$CF,$D3
	.dc.b	$F9,$5F,$29,$FA,$F8,$FE,$42,$39,$46,$0F,$E0,$7F,$AE,$F3,$70,$1F
	.dc.b	$33,$8F,$CB,$3A,$3F,$9A,$EE,$5F,$E7,$27,$AC,$E6,$3F,$3F,$F1,$3D
	.dc.b	$63,$07,$F1,$27,$AB,$83,$F9,$4E,$DF,$59,$F8,$5F,$15,$D1,$FC,$ED
	.dc.b	$7E,$AB,$F7,$8E,$91,$83,$F8,$EB,$D1,$C0,$FC,$B3,$D2,$E1,$E8,$FE
	.dc.b	$DA,$7F,$6B,$FA,$07,$48,$C1,$FC,$A1,$E8,$E0,$7E,$B7,$D1,$C3,$D3
	.dc.b	$4F,$8B,$1F,$CC,$3F,$5E,$E9,$18,$84,$74,$8F,$D9,$43,$E8,$F7,$E0
	.dc.b	$FF,$08,$74,$FF,$0A,$FC,$4B,$C1,$5C,$1F,$04,$7F,$0E,$FC,$53,$8D
	.dc.b	$FF,$40,$78,$BF,$D0,$BF,$17,$E3,$3F,$C5,$1E,$31,$FD,$1B,$F1,$9E
	.dc.b	$29,$E2,$BF,$1A,$E2,$FF,$4C,$78,$AF,$C6,$F8,$9C,$3F,$1C,$DC,$1D
	.dc.b	$F8,$BD,$8E,$BF,$8E,$E8,$E7,$F8,$57,$88,$F0,$FE,$5F,$C8,$EF,$FA
	.dc.b	$AF,$98,$F2,$FE,$75,$D8,$F1,$FB,$1E,$A3,$D3,$F9,$77,$88,$F3,$FE
	.dc.b	$59,$EE,$3D,$BF,$9C,$79,$8F,$5F,$D3,$FC,$23,$E0,$FC,$E7,$A8,$F7
	.dc.b	$FB,$9F,$18,$F8,$BE,$0F,$B8,$F8,$7F,$B7,$FC,$A3,$E4,$FD,$BF,$C2
	.dc.b	$3E,$3F,$9F,$F3,$8F,$9B,$E7,$FC,$63,$E5,$F9,$9F,$48,$FA,$3E,$57
	.dc.b	$CA,$3E,$7F,$91,$F5,$8F,$AB,$E4,$7C,$E3,$E9,$F8,$BF,$68,$FB,$3E
	.dc.b	$2F,$D2,$3F,$2D,$D7,$F8,$BD,$7F,$60,$FF,$3E,$30,$FF,$54,$7F,$19
	.dc.b	$E6,$7F,$24,$7F,$A3,$98,$F3,$34,$FE,$0E,$79,$BF,$C1,$CF,$EC,$3D
	.dc.b	$FE,$07,$FC,$27,$DF,$FA,$47,$FB,$2B,$FA,$37,$FB,$3F,$6F,$E7,$9B
	.dc.b	$5F,$CB,$9F,$31,$FE,$1C,$ED,$E4,$DC,$7F,$92,$3E,$63,$FC,$43,$D7
	.dc.b	$CD,$B5,$CD,$F3,$1F,$E2,$6E,$BE,$6F,$F7,$6A,$E6,$F8,$07,$4F,$36
	.dc.b	$D7,$37,$FB,$99,$FE,$00,$E9,$E6,$7B,$5F,$F1,$9C,$71,$3E,$22,$E7
	.dc.b	$F0,$9B,$5C,$BD,$CD,$F2,$F9,$8D,$71,$F8,$CE,$7C,$7E,$A3,$F8,$CF
	.dc.b	$C6,$FF,$6C,$B7,$89,$FC,$7E,$30,$EB,$D5,$FA,$F9,$FD,$AA,$FF,$24
	.dc.b	$D7,$1E,$4F,$E5,$2F,$3E,$2E,$57,$FE,$1E,$79,$57,$CE,$F4,$7F,$C2
	.dc.b	$2F,$F8,$F2,$7F,$C4,$7C,$7C,$9F,$8D,$C5,$F8,$D9,$FF,$0E,$61,$E7
	.dc.b	$E1,$E7,$1F,$DD,$F0,$E7,$3E,$BC,$F9,$4F,$F9,$7B,$7F,$1B,$70,$FB
	.dc.b	$2F,$F8,$CA,$BE,$16,$FF,$CE,$5B,$7F,$66,$7F,$CA,$3F,$0B,$F0,$B1
	.dc.b	$EE,$FC,$49,$B3,$F8,$FC,$2A,$D7,$E7,$D7,$F9,$02,$BB,$35,$ED,$87
	.dc.b	$F9,$2E,$BC,$D7,$CB,$2F,$AF,$E5,$7C,$B8,$F8,$93,$E9,$8C,$74,$7C
	.dc.b	$27,$A6,$76,$7F,$1B,$8F,$6F,$C7,$F3,$1F,$E4,$2C,$91,$E3,$9F,$97
	.dc.b	$F9,$06,$AB,$E8,$B9,$7F,$C9,$3F,$6A,$FC,$83,$FC,$B7,$7F,$F0,$06
	.dc.b	$31,$3F,$E8,$EC,$B1,$F4,$FF,$E5,$28,$FC,$08,$9F,$F5,$13,$FE,$4E
	.dc.b	$F3,$73,$18,$8F,$F3,$AF,$A3,$8A,$FF,$2D,$FE,$03,$67,$FC,$AB,$8D
	.dc.b	$7F,$CF,$78,$F2,$FA,$DF,$04,$7E,$43,$14,$78,$F2,$C3,$E5,$CA,$D7
	.dc.b	$F9,$80,$63,$87,$FC,$A7,$8D,$E7,$F7,$47,$C4,$BF,$77,$26,$63,$F5
	.dc.b	$C7,$FA,$03,$2F,$0B,$FE,$6C,$AE,$AF,$13,$17,$E7,$FC,$63,$E6,$C7
	.dc.b	$C4,$6B,$FE,$CD,$3F,$D0,$14,$8D,$E7,$C6,$7F,$D7,$F5,$FE,$90,$FC
	.dc.b	$07,$EC,$D7,$2F,$C0,$6D,$68,$F7,$FC,$76,$1F,$C2,$9D,$CE,$F3,$E4
	.dc.b	$D7,$B4,$69,$7E,$45,$C7,$1F,$F6,$7C,$3F,$04,$C5,$7C,$C9,$D4,$F6
	.dc.b	$8E,$A7,$A1,$FC,$AF,$D8,$6B,$8B,$BD,$71,$D3,$8A,$D3,$9E,$ED,$7C
	.dc.b	$58,$7C,$9C,$BA,$4F,$2C,$BF,$D8,$3A,$70,$D3,$8E,$5F,$06,$BC,$97
	.dc.b	$E4,$3F,$D8,$F1,$FE,$EA,$C3,$4F,$F8,$E2,$3E,$F6,$BF,$EB,$08,$FF
	.dc.b	$77,$E9,$A9,$F3,$1C,$EF,$F6,$39,$79,$E3,$15,$F3,$DE,$67,$E7,$A3
	.dc.b	$E7,$CF,$0C,$BC,$31,$F3,$67,$8B,$E8,$7F,$5D,$87,$FA,$AE,$F8,$DF
	.dc.b	$CC,$9F,$13,$FE,$4F,$7E,$03,$FC,$13,$0B,$3E,$89,$F1,$9F,$EE,$29
	.dc.b	$F4,$38,$7F,$D8,$CD,$7E,$6B,$D9,$F3,$39,$7C,$DB,$BE,$17,$FD,$A7
	.dc.b	$3F,$E6,$50,$7F,$07,$7F,$D8,$CF,$E8,$1D,$E7,$FC,$51,$E3,$CB,$67
	.dc.b	$8F,$C6,$F2,$E2,$F0,$C7,$4C,$3C,$0E,$8F,$01,$9B,$FF,$6B,$7B,$5C
	.dc.b	$E1,$7C,$F3,$C3,$D7,$92,$E3,$FC,$EE,$BC,$C7,$89,$47,$6C,$7C,$33
	.dc.b	$D2,$7F,$83,$BE,$13,$84,$FE,$0D,$F8,$46,$32,$EF,$A7,$B2,$32,$79
	.dc.b	$B9,$BF,$95,$E9,$F2,$A3,$B4,$B4,$E3,$A7,$2C,$7D,$DE,$E7,$A6,$FD
	.dc.b	$27,$FA,$5A,$3E,$9E,$3A,$7F,$41,$1E,$73,$47,$9F,$4F,$6D,$7F,$5D
	.dc.b	$CF,$86,$9E,$79,$CD,$FD,$0B,$F4,$DF,$C6,$AF,$34,$78,$A3,$18,$F2
	.dc.b	$AF,$3E,$7E,$7A,$F1,$52,$3E,$7E,$5A,$7F,$A1,$69,$F0,$E7,$57,$E7
	.dc.b	$8C,$4F,$D6,$CF,$87,$1F,$E0,$7C,$DF,$EB,$D5,$AA,$F2,$E5,$FC,$F6
	.dc.b	$31,$9F,$7C,$7F,$10,$7E,$46,$69,$F9,$CC,$6E,$9F,$0E,$FE,$B5,$7C
	.dc.b	$EA,$CE,$DA,$C3,$73,$D2,$FE,$05,$7E,$02,$11,$F0,$23,$EC,$C6,$31
	.dc.b	$F8,$31,$FA,$C8,$D9,$F0,$5C,$4F,$C1,$78,$BF,$2F,$3F,$5E,$BF,$8C
	.dc.b	$9D,$DF,$C2,$7F,$05,$7D,$F0,$F3,$4F,$D5,$DB,$F3,$EF,$A2,$BF,$2E
	.dc.b	$FF,$88,$6F,$F5,$CF,$E0,$62,$DF,$F2,$F7,$F2,$71,$F8,$B6,$C7,$C0
	.dc.b	$C3,$75,$EE,$9F,$BB,$8F,$95,$EF,$5F,$1B,$1E,$B3,$E5,$CF,$F4,$D8
	.dc.b	$FE,$12,$FC,$6F,$CB,$7E,$F1,$FD,$E6,$FE,$43,$F8,$F8,$F9,$0E,$8F
	.dc.b	$E4,$D8,$FE,$C5,$FE,$37,$EE,$7C,$29,$FE,$07,$C5,$99,$4E,$6F,$FC
	.dc.b	$CE,$DD,$B4,$E7,$8A,$E5,$E8,$C7,$CD,$87,$68,$F5,$E1,$FA,$DB,$FA
	.dc.b	$2F,$E0,$27,$C9,$18,$CF,$BD,$70,$7F,$A5,$46,$77,$EF,$5F,$CD,$30
	.dc.b	$FF,$2C,$67,$88,$F9,$05,$8F,$AC,$FF,$11,$AF,$0B,$FC,$6D,$BB,$71
	.dc.b	$CB,$F8,$14,$FC,$2C,$BF,$47,$97,$48,$F6,$1F,$90,$38,$3F,$DD,$1F
	.dc.b	$5C,$70,$C3,$F0,$18,$7F,$2E,$BF,$CF,$E3,$C1,$F7,$DF,$E8,$1F,$F2
	.dc.b	$CB,$E7,$BF,$0C,$79,$DB,$8F,$9E,$7E,$64,$7E,$E1,$F9,$87,$13,$F5
	.dc.b	$C7,$F2,$63,$D8,$63,$1F,$9B,$1C,$5F,$3B,$CE,$17,$DD,$F2,$9F,$08
	.dc.b	$70,$7F,$9B,$1E,$A3,$87,$F2,$98,$FD,$57,$FB,$AC,$71,$7F,$31,$C1
	.dc.b	$9D,$3F,$94,$6B,$6F,$C3,$7F,$30,$FE,$E7,$7F,$4D,$E5,$5F,$A0,$6C
	.dc.b	$FC,$77,$71,$F9,$47,$CB,$35,$B3,$F2,$1F,$23,$F1,$F5,$FE,$8B,$A7
	.dc.b	$C6,$9F,$0C,$FF,$1E,$C3,$D8,$FF,$8B,$A3,$CC,$7F,$52,$32,$7C,$83
	.dc.b	$B9,$E5,$3F,$AB,$EB,$F9,$9E,$2F,$C8,$E1,$FD,$D1,$FE,$51,$59,$C3
	.dc.b	$FB,$AD,$FC,$E9,$FE,$3C,$F8,$1F,$E5,$D8,$6E,$BE,$4C,$FE,$0C,$FE
	.dc.b	$FC,$FE,$41,$BC,$1B,$E0,$E1,$97,$8E,$7F,$1A,$7F,$85,$71,$FE,$BD
	.dc.b	$97,$96,$3E,$37,$07,$F7,$DC,$FF,$C2,$91,$E3,$3E,$31,$4F,$19,$F8
	.dc.b	$97,$FC,$09,$8F,$78,$F8,$91,$FE,$07,$C3,$F8,$7F,$1F,$F0,$DD,$BF
	.dc.b	$81,$B6,$B2,$7F,$A6,$E3,$FC,$25,$8F,$FA,$4E,$31,$CB,$A7,$F1,$FF
	.dc.b	$85,$DA,$FD,$C4,$7F,$12,$BF,$E3,$5D,$3F,$C8,$1F,$F2,$C7,$4E,$EB
	.dc.b	$FC,$A9,$5B,$3F,$D4,$FB,$F3,$E3,$FF,$13,$47,$F2,$AB,$FF,$39,$74
	.dc.b	$7A,$23,$FC,$D4,$6D,$F3,$C6,$B3,$E5,$2C,$EB,$D7,$87,$C8,$FF,$B4
	.dc.b	$0F,$F3,$21,$FC,$F2,$BF,$D0,$0F,$CE,$3C,$46,$F1,$F8,$58,$F9,$A3
	.dc.b	$27,$FA,$0B,$2F,$33,$4D,$E5,$FD,$C5,$F8,$06,$F2,$CD,$79,$E3,$2F
	.dc.b	$FA,$B5,$E3,$A7,$A1,$FF,$47,$C7,$A1,$F0,$BF,$EA,$B7,$8D,$7F,$A1
	.dc.b	$5F,$E7,$18,$FF,$B4,$B1,$E5,$5F,$EB,$C3,$D7,$3F,$ED,$A3,$C2,$7F
	.dc.b	$BA,$39,$F3,$7F,$EC,$48,$C9,$E7,$3A,$9F,$ED,$13,$A9,$FC,$06,$FF
	.dc.b	$DA,$CF,$0E,$7E,$EF,$F4,$9E,$AF,$EC,$18,$FF,$66,$7F,$B1,$5D,$9F
	.dc.b	$EA,$DA,$FF,$6D,$7F,$B8,$CF,$F7,$E7,$F7,$7C,$1F,$EF,$47,$FC,$90
	.dc.b	$7F,$8B,$1C,$6D,$FE,$FE,$8F,$EC,$F3,$C7,$0E,$D1,$C3,$0B,$5F,$E3
	.dc.b	$98,$D9,$FD,$9C,$FE,$39,$D7,$FE,$0E,$DB,$C9,$1F,$E0,$08,$FF,$08
	.dc.b	$DE,$31,$47,$F8,$53,$0C,$3F,$E1,$6E,$2B,$7E,$E3,$FC,$0F,$C9,$FE
	.dc.b	$7F,$C7,$FA,$26,$5F,$E9,$3F,$D4,$DF,$64,$F5,$DF,$0E,$F8,$F7,$EA
	.dc.b	$8F,$99,$7B,$DF,$DF,$31,$E3,$87,$0B,$DD,$3F,$3E,$9D,$1F,$E8,$FF
	.dc.b	$C9,$9E,$D3,$F7,$17,$E4,$EF,$84,$7E,$52,$FC,$F5,$CB,$CB,$9B,$FD
	.dc.b	$83,$97,$2C,$78,$E1,$C2,$DF,$79,$BB,$5E,$F2,$7C,$BB,$F3,$17,$CC
	.dc.b	$3D,$BE,$17,$BA,$D4,$BF,$59,$7E,$72,$FD,$6F,$FD,$55,$F4,$4F,$C0
	.dc.b	$5F,$6C,$F8,$67,$BE,$F8,$C7,$E1,$0F,$D8,$39,$E5,$CB,$1E,$38,$70
	.dc.b	$BD,$FE,$0D,$E0,$E6,$FF,$5C,$64,$FC,$A5,$FA,$0B,$E9,$9F,$B4,$BE
	.dc.b	$91,$F3,$DF,$0B,$8F,$E6,$75,$FF,$BA,$B2,$E5,$8F,$1C,$38,$5B,$E7
	.dc.b	$1F,$9D,$BF,$77,$FF,$64,$7E,$82,$FD,$91,$EC,$3F,$05,$7E,$F8,$FB
	.dc.b	$17,$EC,$2F,$D4,$B9,$E5,$CB,$1E,$38,$70,$BD,$FD,$AF,$83,$EF,$3F
	.dc.b	$DA,$DF,$40,$FD,$45,$F6,$BF,$98,$FD,$1D,$F8,$0B,$F2,$17,$EB,$7C
	.dc.b	$7A,$83,$1E,$58,$71,$BE,$1F,$AF,$EE,$B5,$97,$6F,$80,$7D,$7B,$F0
	.dc.b	$97,$E9,$AF,$C6,$B8,$69,$D2,$F7,$5A,$FD,$8B,$E4,$FE,$B3,$FE,$EF
	.dc.b	$F7,$5F,$B7,$BF,$4A,$FD,$7F,$6B,$F3,$CF,$3F,$EA,$DC,$38,$7D,$38
	.dc.b	$70,$BD,$D7,$D9,$F0,$9F,$B3,$7F,$BE,$FE,$89,$CA,$FD,$89,$FA,$F3
	.dc.b	$F6,$A7,$EA,$BF,$5B,$D9,$FB,$B8,$F1,$C3,$85,$EE,$BE,$DF,$E6,$BF
	.dc.b	$E9,$2F,$C9,$7E,$5F,$07,$2F,$45,$FB,$8B,$F7,$F7,$D0,$3F,$33,$7E
	.dc.b	$47,$F8,$4F,$9D,$7E,$7F,$CB,$F7,$17,$ED,$1E,$8B,$FF,$52,$7D,$17
	.dc.b	$FA,$82,$B5,$39,$D7,$FD,$2B,$18,$E2,$7E,$1D,$FE,$AF,$F9,$EF,$F7
	.dc.b	$F6,$7E,$1B,$FC,$0F,$DB,$3E,$87,$F8,$7C,$76,$9D,$29,$D7,$C9,$7F
	.dc.b	$AB,$B5,$F7,$C7,$1F,$98,$ED,$8F,$69,$F8,$7E,$63,$E9,$9F,$94,$BE
	.dc.b	$AF,$B5,$6A,$73,$1F,$EB,$88,$C4,$BF,$13,$EB,$AF,$F8,$9F,$57,$EB
	.dc.b	$CF,$E0,$07,$E5,$3F,$EC,$5E,$6F,$F0,$3C,$E6,$3F,$BE,$3B,$EA,$F3
	.dc.b	$EF,$FE,$CA,$FC,$C7,$EC,$F5,$BE,$B3,$FD,$9B,$9F,$C3,$9C,$C6,$23
	.dc.b	$FD,$87,$18,$FB,$9E,$7A,$FD,$C9,$FD,$6F,$96,$F4,$78,$47,$EC,$36
	.dc.b	$FF,$6C,$F7,$FE,$22,$B5,$7F,$ED,$A8,$C5,$B0,$EB,$F6,$3F,$8E,$FD
	.dc.b	$65,$F9,$8B,$7F,$E5,$BF,$87,$BF,$11,$7E,$EE,$FD,$8B,$FD,$CF,$8F
	.dc.b	$1C,$38,$7C,$EB,$0E,$16,$FC,$87,$FD,$DB,$F6,$DF,$9C,$F0,$FF,$DF
	.dc.b	$1F,$69,$FC,$2F,$F4,$97,$E7,$FF,$D1,$79,$3F,$CF,$35,$AF,$88,$56
	.dc.b	$BE,$93,$FD,$EF,$F5,$5F,$A6,$FD,$6B,$F2,$9E,$DB,$F3,$87,$EE,$7F
	.dc.b	$D7,$33,$FE,$AF,$DD,$FC,$5A,$F7,$5A,$C3,$9C,$FF,$1A,$8F,$F0,$2F
	.dc.b	$76,$21,$FB,$DF,$DE,$EA,$FD,$0F,$9F,$FB,$6F,$D6,$3F,$2F,$FA,$07
	.dc.b	$F4,$5F,$A2,$6B,$FA,$3C,$7F,$92,$D6,$A5,$F9,$1F,$D3,$FC,$0F,$FC
	.dc.b	$FD,$97,$D1,$1F,$BE,$FF,$78,$F0,$D3,$A9,$FA,$D2,$F5,$3E,$E7,$27
	.dc.b	$F0,$8F,$B8,$FF,$02,$10,$FF,$03,$7E,$59,$AD,$7F,$9B,$C6,$4D,$FE
	.dc.b	$DF,$F8,$5F,$D8,$FF,$88,$FE,$2C,$FC,$93,$F4,$9F,$0E,$FD,$9D,$FA
	.dc.b	$0F,$F5,$2F,$9E,$E1,$9F,$F4,$2C,$38,$7B,$1F,$CA,$2F,$7F,$25,$FC
	.dc.b	$97,$91,$FB,$87,$F1,$EF,$C9,$5F,$A0,$BE,$31,$FA,$2B,$F5,$DF,$B9
	.dc.b	$F8,$FB,$F4,$5F,$CD,$63,$C7,$0E,$17,$BF,$15,$7F,$DA,$73,$7E,$85
	.dc.b	$FC,$B3,$C3,$7E,$D2,$FD,$99,$FB,$2F,$F1,$5F,$C7,$5F,$2A,$FD,$D7
	.dc.b	$F5,$18,$F1,$C3,$85,$EE,$9F,$B8,$7E,$B7,$F0,$AF,$E7,$3E,$2B,$F6
	.dc.b	$9F,$B3,$F3,$4F,$CB,$7F,$4D,$F7,$BF,$B3,$E1,$FF,$1A,$77,$FF,$6A
	.dc.b	$C3,$87,$97,$FD,$2A,$F7,$FB,$38,$7E,$29,$FD,$E3,$F2,$2F,$7F,$E3
	.dc.b	$1F,$D8,$3F,$1F,$FE,$ED,$DF,$BA,$D7,$57,$F9,$43,$EA,$95,$FD,$EF
	.dc.b	$F7,$6E,$9F,$9E,$7F,$50,$FC,$89,$FA,$2F,$F9,$37,$F5,$97,$ED,$5F
	.dc.b	$C0,$7E,$B1,$FA,$B7,$0E,$5F,$F0,$0E,$1C,$3D,$56,$1C,$3E,$47,$FB
	.dc.b	$37,$EA,$1E,$9F,$DB,$BF,$AE,$FE,$7E,$FC,$A5,$F9,$AF,$F7,$5F,$C1
	.dc.b	$1F,$2E,$FD,$B9,$FA,$27,$1D,$FF,$C1,$F8,$F1,$C3,$85,$EF,$F5,$2F
	.dc.b	$8D,$F9,$9F,$F7,$7F,$84,$7E,$0C,$F9,$7F,$D5,$FE,$8C,$C9,$F9,$FF
	.dc.b	$0D,$3F,$C3,$7F,$11,$FA,$DC,$38,$78,$EC,$38,$7D,$87,$D3,$FC,$A7
	.dc.b	$FB,$BF,$E1,$7F,$E6,$EF,$5B,$E0,$FB,$9F,$E0,$8F,$CA,$FE,$C7,$99
	.dc.b	$FE,$81,$D7,$FC,$4F,$7B,$ED,$FD,$0F,$C4,$7F,$A0,$BC,$EF,$E3,$5F
	.dc.b	$BC,$7F,$80,$BF,$3D,$FF,$16,$FE,$1A,$FD,$2B,$FC,$EF,$FB,$CB,$F3
	.dc.b	$DF,$C7,$7C,$1B,$0E,$1A,$FF,$8D,$AF,$75,$F3,$A9,$FB,$C7,$F8,$3B
	.dc.b	$F4,$47,$CF,$7F,$61,$FD,$5F,$FD,$07,$FB,$99,$FF,$49,$DF,$0C,$BF
	.dc.b	$C7,$F7,$C3,$E4,$35,$F9,$83,$0E,$1E,$77,$F2,$8F,$A7,$FF,$0D,$FE
	.dc.b	$8A,$FD,$75,$F3,$5E,$BF,$DC,$DF,$9C,$BF,$7B,$7E,$D7,$C7,$7F,$F2
	.dc.b	$6E,$3C,$70,$E1,$7B,$FD,$03,$F9,$AF,$D4,$FF,$81,$BF,$43,$FE,$B1
	.dc.b	$FB,$97,$EF,$3E,$87,$FC,$81,$F6,$7A,$CF,$FC,$AF,$F8,$3A,$F7,$F2
	.dc.b	$9F,$C7,$FF,$13,$7E,$FD,$FE,$0D,$BF,$F1,$9F,$E5,$3F,$9C,$FC,$43
	.dc.b	$FB,$27,$B7,$EF,$FC,$CB,$F8,$DF,$E1,$7F,$E3,$8B,$FF,$42,$F9,$7F
	.dc.b	$E3,$8F,$BE,$FC,$EF,$6F,$FA,$6F,$E7,$75,$5A,$CB,$9C,$E6,$39,$67
	.dc.b	$DE,$F8,$E5,$FC,$A3,$EB,$BF,$AD,$D7,$F9,$A3,$83,$D3,$C7,$FE,$5D
	.dc.b	$DF,$FA,$6F,$2F,$C3,$C5,$CB,$F0,$F8,$BF,$BD,$E5,$FD,$C3,$2F,$9B
	.dc.b	$DB,$F5,$3C,$87,$77,$D4,$EE,$75,$7F,$58,$EF,$FB,$77,$BA,$D7,$7F
	.dc.b	$FA,$97,$C7,$FC,$67,$8F,$F8,$BF,$0F,$F0,$9E,$3F,$FB,$5F,$EB,$FF
	.dc.b	$15,$B3,$C7,$FC,$54,$FE,$40,$FF,$70,$CF,$E6,$E7,$FD,$1B,$E6,$FE
	.dc.b	$8B,$83,$F3,$7E,$07,$FA,$CF,$CB,$FC,$FF,$9B,$FA,$5C,$DE,$6F,$E9
	.dc.b	$7E,$6D,$FA,$DF,$EC,$73,$FF,$54,$7D,$D7,$F3,$DF,$17,$F6,$3E,$AF
	.dc.b	$F3,$5A,$D7,$A9,$FB,$5F,$5B,$F7,$5E,$B7,$2F,$82,$FE,$19,$EE,$FF
	.dc.b	$04,$F7,$3F,$81,$E0,$9C,$FC,$0F,$F0,$EF,$81,$FE,$1B,$2F,$89,$FE
	.dc.b	$43,$F1,$3F,$C8,$3C,$5F,$F5,$E7,$C3,$7F,$19,$AD,$71,$7F,$1E,$AD
	.dc.b	$7C,$4F,$F2,$4F,$89,$FE,$47,$4F,$91,$FE,$71,$F2,$3F,$CD,$FE,$47
	.dc.b	$F9,$B7,$C6,$FC,$09,$CF,$BF,$FC,$D2,$7F,$9D,$7C,$87,$C4,$F9,$6F
	.dc.b	$E8,$BF,$31,$FD,$27,$EE,$9F,$2B,$9E,$0F,$F5,$EF,$E3,$AF,$9B,$FE
	.dc.b	$9F,$F3,$3F,$D2,$AF,$FB,$B7,$E0,$1F,$D6,$AF,$FC,$33,$E1,$FF,$60
	.dc.b	$F8,$CF,$EB,$33,$9F,$9D,$FE,$BB,$F4,$1F,$D8,$FE,$FF,$FD,$F3,$F4
	.dc.b	$1F,$E5,$2F,$86,$FE,$DD,$F4,$FF,$DD,$3E,$55,$F4,$FF,$DD,$3E,$97
	.dc.b	$5B,$7F,$B4,$3E,$A3,$FB,$DD,$EE,$BF,$BC,$FC,$BF,$F9,$73,$EA,$BF
	.dc.b	$C0,$DF,$59,$FE,$0C,$FA,$6F,$F0,$5F,$E4,$99,$CB,$FD,$B3,$F5,$DF
	.dc.b	$E1,$2A,$D7,$CD,$7F,$82,$BE,$BB,$FC,$CB,$F6,$9F,$E2,$4F,$8F,$FF
	.dc.b	$11,$FD,$8F,$F8,$73,$F1,$8C,$E5,$FE,$21,$9C,$FD,$AF,$F8,$97,$A9
	.dc.b	$F4,$9F,$E3,$5F,$1F,$FC,$60,$7F,$22,$8C,$DF,$F8,$D2,$33,$F6,$DF
	.dc.b	$E7,$4F,$A5,$FF,$20,$7D,$BF,$F9,$13,$EA,$31,$7F,$EF,$08,$FF,$48
	.dc.b	$F8,$0F,$80,$7E,$89,$FE,$51,$A7,$DC,$FF,$94,$7E,$57,$3F,$77,$FE
	.dc.b	$59,$C7,$FD,$79,$DF,$FE,$59,$C3,$F8,$6F,$DE,$7F,$97,$E7,$9F,$CD
	.dc.b	$7F,$B5,$3A,$7F,$CC,$B5,$FC,$4B,$E4,$7F,$CD,$7F,$7C,$AD,$EB,$FE
	.dc.b	$6F,$C3,$FC,$B9,$CB,$FE,$75,$A7,$C5,$79,$BE,$07,$FD,$09,$5F,$E9
	.dc.b	$1F,$C0,$7F,$D0,$9F,$30,$FC,$03,$59,$0F,$75,$6F,$D1,$FF,$42,$D3
	.dc.b	$F3,$AF,$83,$F0,$5F,$E9,$5A,$FF,$4D,$FC,$F7,$FA,$6B,$F0,$3C,$E3
	.dc.b	$CF,$F0,$5F,$F5,$0D,$6B,$EA,$BF,$D3,$DC,$9F,$EB,$1A,$7E,$11,$FE
	.dc.b	$B1,$F5,$3F,$D7,$F3,$98,$FF,$57,$CF,$F3,$C8,$C7,$E0,$FF,$EB,$68
	.dc.b	$7E,$0B,$F5,$3F,$09,$FF,$5F,$65,$FE,$DE,$B4,$7F,$B1,$3F,$06,$FE
	.dc.b	$8B,$7B,$3E,$07,$E1,$7F,$EC,$B8,$FF,$6A,$57,$FB,$67,$F0,$DF,$F6
	.dc.b	$96,$1C,$FF,$12,$F8,$D8,$7F,$B7,$3E,$DB,$FD,$B6,$F9,$63,$35,$FE
	.dc.b	$DB,$8F,$EB,$19,$FE,$23,$F1,$2F,$F7,$3F,$E2,$3E,$57,$FD,$D3,$F6
	.dc.b	$7E,$D6,$D7,$6A,$E0,$FD,$5F,$7B,$FD,$F5,$9F,$18,$C7,$8D,$F9,$57
	.dc.b	$91,$8F,$27,$D7,$F8,$AF,$A3,$F1,$77,$E2,$EF,$C5,$DF,$8B,$FC,$7A
	.dc.b	$D4,$E7,$5F,$A1,$18,$A4,$7B,$38,$FF,$CF,$79,$7E,$27,$E1,$9F,$8C
	.dc.b	$3E,$E5,$F8,$B3,$A7,$8F,$E2,$AF,$C4,$7F,$D2,$7F,$81,$FC,$CF,$3B
	.dc.b	$C9,$F8,$53,$F1,$97,$C0,$35,$E1,$F0,$9C,$37,$FF,$25,$F1,$30,$E1
	.dc.b	$7B,$CD,$F8,$D7,$FA,$83,$F1,$EF,$B3,$F8,$EB,$E5,$3F,$23,$F3,$1E
	.dc.b	$3F,$A3,$FD,$17,$E2,$3B,$D7,$F4,$97,$BF,$2D,$7B,$97,$E3,$DF,$EA
	.dc.b	$CF,$C6,$3F,$0B,$F3,$6F,$C8,$DF,$92,$BF,$22,$7E,$14,$FC,$89,$F9
	.dc.b	$23,$D4,$E5,$A7,$F4,$4C,$78,$E1,$C2,$F7,$F9,$26,$A9,$9F,$7C,$FB
	.dc.b	$7D,$93,$F2,$57,$C0,$7D,$FF,$C7,$9C,$1E,$6D,$7A,$DE,$E9,$F7,$7F
	.dc.b	$EC,$6F,$C9,$FF,$3F,$F9,$5B,$E7,$5F,$95,$BE,$CF,$ED,$7D,$DB,$F2
	.dc.b	$B7,$E5,$5C,$78,$FF,$86,$63,$C7,$0E,$1E,$1B,$0E,$1F,$95,$F9,$3F
	.dc.b	$31,$FD,$6F,$82,$FC,$C5,$F9,$83,$F0,$97,$E2,$CF,$CC,$5F,$98,$BF
	.dc.b	$31,$E5,$C1,$F6,$B1,$E3,$87,$0B,$DF,$E7,$5F,$0B,$F1,$AF,$DD,$FE
	.dc.b	$70,$FC,$DD,$F9,$83,$F3,$77,$03,$F3,$7F,$C5,$FE,$6F,$C7,$6F,$E5
	.dc.b	$18,$F1,$C3,$85,$EF,$D4,$7E,$71,$E8,$FA,$AF,$E1,$3B,$3D,$37,$E7
	.dc.b	$9F,$D3,$7F,$3B,$7E,$78,$FC,$F3,$F9,$B7,$4B,$1E,$38,$70,$F8,$E6
	.dc.b	$1C,$2F,$F9,$B5,$EF,$83,$F9,$E6,$7D,$F3,$FE,$B5,$F9,$A3,$F3,$D7
	.dc.b	$77,$F4,$DC,$38,$60,$C3,$86,$6C,$38,$5F,$F4,$EF,$96,$7D,$8F,$F2
	.dc.b	$4F,$C1,$FF,$3D,$F9,$77,$E9,$BF,$3F,$7E,$30,$FC,$FD,$F7,$0C,$78
	.dc.b	$ED,$FD,$47,$1E,$3F,$02,$C7,$8E,$0F,$D0,$1F,$80,$3F,$50,$FE,$E7
	.dc.b	$F3,$EF,$D7,$FF,$2D,$FB,$2F,$F0,$9F,$DE,$9F,$BF,$BE,$77,$58,$F1
	.dc.b	$F9,$CF,$E9,$3C,$1F,$D1,$0F,$F0,$BE,$7D,$3E,$77,$11,$E0,$8E,$FF
	.dc.b	$B5,$E2,$F9,$6B,$E1,$F7,$39,$46,$36,$FD,$47,$E2,$7F,$D4,$23,$12
	.dc.b	$DF,$18,$FA,$AF,$D3,$F1,$1F,$E9,$AB,$DD,$7F,$23,$8F,$F5,$14,$63
	.dc.b	$E6,$3F,$CF,$5A,$7A,$36,$FE,$C7,$87,$C2,$97,$0F,$F2,$2A,$D6,$BE
	.dc.b	$7C,$BE,$37,$0F,$F5,$CC,$BA,$5F,$C8,$E0,$DE,$5C,$E3,$E5,$F0,$E3
	.dc.b	$7F,$56,$9D,$B1,$F9,$D8,$7C,$7D,$7C,$75,$AC,$38,$FE,$C9,$9D,$5F
	.dc.b	$A5,$7C,$BE,$1F,$85,$C1,$F2,$9F,$89,$FA,$0D,$CB,$9F,$0F,$F8,$CA
	.dc.b	$BE,$1D,$7C,$9C,$FC,$71,$8A,$DE,$DF,$0A,$3D,$D5,$BB,$FC,$37,$16
	.dc.b	$2F,$7A,$FB,$B3,$F5,$7D,$37,$E0,$B1,$F4,$71,$3F,$0A,$F2,$6F,$F8
	.dc.b	$1C,$7E,$D9,$F0,$34,$B1,$F3,$F5,$FA,$58,$7C,$9D,$7F,$2D,$F3,$7E
	.dc.b	$1A,$D6,$BF,$98,$E6,$D6,$1F,$37,$F2,$4F,$B3,$8F,$E5,$EF,$E5,$1F
	.dc.b	$27,$3F,$CD,$EF,$E4,$9F,$CE,$9F,$8B,$D3,$F4,$EF,$F8,$FE,$3C,$AF
	.dc.b	$F9,$06,$AB,$93,$E5,$3F,$E1,$A6,$9F,$EF,$38,$AF,$F8,$5C,$F5,$1F
	.dc.b	$EF,$26,$BC,$4F,$C0,$1F,$E3,$27,$D6,$FF,$BE,$22,$BC,$79,$FF,$4D
	.dc.b	$6F,$DB,$FF,$DB,$9F,$96,$42,$4E,$A3,$EB,$37,$97,$F6,$DC,$1F,$E7
	.dc.b	$89,$FF,$72,$61,$FB,$68,$C5,$FF,$86,$A3,$F6,$11,$88,$D6,$1E,$17
	.dc.b	$FD,$D9,$7F,$BB,$9D,$45,$9D,$3A,$AF,$DB,$CE,$A3,$AC,$E9,$D4,$FF
	.dc.b	$A9,$A7,$51,$C2,$74,$EB,$0F,$73,$F0,$4F,$5D,$7F,$24,$AE,$31,$FE
	.dc.b	$34,$9D,$BB,$3B,$57,$47,$E5,$DF,$18,$FF,$47,$7E,$B7,$FE,$16,$E9
	.dc.b	$7E,$C6,$FC,$DB,$FC,$33,$F0,$0F,$CD,$B8,$7E,$05,$F4,$35,$FF,$4D
	.dc.b	$E7,$F4,$33,$FF,$4C,$7D,$8D,$FC,$11,$AF,$CF,$E0,$F7,$46,$3F,$33
	.dc.b	$F9,$EE,$BE,$7F,$F4,$F3,$FC,$57,$2F,$F5,$3D,$F9,$63,$E5,$D6,$F8
	.dc.b	$1D,$F8,$AF,$57,$EF,$17,$1F,$E8,$7C,$4F,$CB,$3C,$35,$EF,$87,$DE
	.dc.b	$E2,$B7,$C9,$F8,$5E,$6F,$C9,$E9,$DE,$FE,$25,$79,$78,$7E,$4B,$C0
	.dc.b	$FF,$33,$ED,$F3,$6B,$86,$1F,$E4,$2F,$90,$FA,$71,$8A,$FF,$10,$71
	.dc.b	$EE,$FE,$46,$93,$ED,$FE,$2B,$FE,$4A,$8F,$C8,$5F,$C7,$E0,$B3,$F6
	.dc.b	$73,$F8,$1C,$3F,$6A,$FD,$8F,$DC,$F8,$1B,$B7,$DD,$9F,$56,$3E,$6C
	.dc.b	$AE,$3F,$17,$6F,$F2,$CF,$67,$A2,$31,$B7,$F8,$DF,$0F,$74,$63,$A5
	.dc.b	$86,$2B,$E2,$47,$F7,$AE,$1F,$97,$5F,$E6,$4F,$0B,$D3,$5F,$43,$9F
	.dc.b	$FA,$A7,$D7,$7F,$BC,$3F,$3B,$FE,$FD,$F2,$AF,$CE,$9E,$5F,$F8,$0F
	.dc.b	$F4,$87,$E3,$AF,$D2,$FF,$C0,$F8,$F1,$E1,$FE,$F0,$F3,$F3,$EF,$4B
	.dc.b	$4F,$F0,$55,$7F,$7A,$AF,$07,$9B,$FD,$3E,$3F,$BA,$46,$35,$D5,$F1
	.dc.b	$AF,$F3,$0F,$97,$FD,$FE,$B5,$B5,$AD,$5F,$A7,$2E,$71,$89,$F5,$E3
	.dc.b	$FD,$FA,$F8,$EC,$C7,$17,$B8,$FF,$11,$57,$C6,$8C,$79,$3F,$90,$8F
	.dc.b	$83,$18,$FD,$B3,$5A,$FC,$AC,$BE,$4D,$B8,$3E,$BE,$1C,$2B,$58,$FC
	.dc.b	$DE,$04,$E7,$7F,$F3,$FE,$3C,$B0,$FA,$3D,$D9,$E0,$FA,$D8,$F1,$FD
	.dc.b	$FF,$A7,$57,$D0,$D3,$ED,$F1,$7C,$57,$C9,$1F,$11,$F0,$ED,$E5,$D7
	.dc.b	$C3,$D0,$E3,$F8,$B3,$E9,$F0,$F3,$18,$DB,$F0,$7C,$7F,$0B,$F4,$CF
	.dc.b	$E8,$B3,$F8,$BF,$53,$C5,$C7,$FC,$23,$3F,$12,$F9,$9F,$BB,$AF,$E4
	.dc.b	$3E,$E7,$3C,$7F,$91,$AF,$31,$F0,$B8,$FF,$0D,$87,$C1,$E5,$FC,$AF
	.dc.b	$E2,$9F,$63,$1F,$5F,$4F,$D2,$C3,$87,$17,$D1,$FC,$7F,$E6,$FE,$AC
	.dc.b	$FD,$39,$C6,$FC,$49,$F9,$F7,$FB,$33,$F8,$07,$FD,$C1,$D1,$FE,$C2
	.dc.b	$FD,$B5,$EB,$B2,$E5,$8F,$1E,$2F,$F2,$DC,$63,$2F,$3C,$7D,$C3,$C3
	.dc.b	$9F,$82,$1F,$B1,$70,$8C,$69,$FC,$A6,$B8,$1C,$6F,$A4,$7F,$B8,$23
	.dc.b	$27,$4C,$3D,$51,$FE,$E4,$E0,$CC,$F9,$EF,$AF,$D8,$FE,$F6,$38,$E1
	.dc.b	$D2,$1B,$7E,$5A,$7C,$15,$C2,$1A,$F0,$8C,$7D,$EF,$9E,$BC,$5A,$FC
	.dc.b	$AC,$39,$CB,$F8,$03,$D7,$97,$7C,$BE,$6F,$A5,$DB,$6F,$4E,$1C,$38
	.dc.b	$5F,$AF,$E7,$3E,$98,$7D,$4F,$B5,$FD,$5E,$5C,$F4,$FE,$A1,$1E,$6D
	.dc.b	$7E,$06,$7F,$77,$83,$E7,$3F,$1A,$3F,$D2,$97,$F3,$3D,$3B,$C1,$F1
	.dc.b	$BC,$5F,$93,$3E,$7A,$F8,$17,$E6,$CF,$E0,$47,$C0,$E3,$FC,$46,$9A
	.dc.b	$D7,$CE,$F9,$27,$5C,$1E,$DD,$BE,$F6,$5F,$8A,$F7,$7F,$CC,$4F,$AF
	.dc.b	$D1,$F8,$FB,$7C,$EE,$2F,$C9,$78,$EF,$D4,$1E,$CF,$F3,$9F,$2B,$FD
	.dc.b	$6D,$87,$F8,$33,$89,$F9,$3F,$FA,$D7,$C2,$E1,$7F,$CD,$F8,$3F,$D6
	.dc.b	$98,$F2,$BF,$E6,$DE,$67,$FA,$D2,$31,$E1,$33,$C6,$1D,$F9,$F8,$BE
	.dc.b	$1F,$22,$E3,$FC,$FE,$BC,$79,$7E,$5B,$1E,$71,$E6,$CB,$18,$7A,$7C
	.dc.b	$3F,$EB,$7E,$BE,$31,$E5,$18,$7F,$47,$87,$28,$C4,$7E,$B6,$31,$BF
	.dc.b	$C1,$DF,$18,$75,$8C,$6B,$FC,$BA,$B7,$8F,$F0,$2B,$FF,$6D,$69,$F1
	.dc.b	$B1,$F0,$E1,$CF,$E1,$BF,$A3,$61,$F1,$27,$E1,$9E,$FE,$1F,$83,$C1
	.dc.b	$F1,$F5,$BB,$FD,$4A,$F6,$4E,$71,$FA,$DD,$BE,$69,$EF,$AF,$8F,$36
	.dc.b	$DE,$3C,$38,$4E,$7C,$9F,$37,$99,$FE,$A9,$DB,$EB,$FA,$3E,$4B,$F1
	.dc.b	$78,$5F,$C0,$A7,$E4,$17,$8F,$E2,$1E,$4D,$FF,$13,$97,$F3,$7E,$2F
	.dc.b	$8D,$3F,$26,$1F,$6D,$F4,$A3,$1B,$BF,$76,$F5,$67,$E6,$8F,$C8,$F0
	.dc.b	$F2,$7D,$A7,$E5,$3E,$0F,$FA,$FF,$AA,$3F,$B9,$77,$7F,$98,$76,$7E
	.dc.b	$DF,$F0,$BE,$1D,$E0,$BE,$19,$E2,$71,$F8,$5F,$EF,$5E,$77,$FB,$2F
	.dc.b	$3E,$9B,$7F,$A7,$79,$BF,$91,$F2,$F8,$27,$BF,$AF,$88,$F1,$47,$8B
	.dc.b	$0F,$E4,$51,$FC,$CA,$39,$EF,$FE,$75,$FD,$4B,$9C,$62,$BF,$69,$DD
	.dc.b	$FC,$C2,$31,$C3,$EA,$C3,$F9,$BC,$63,$87,$FC,$15,$1E,$3D,$3F,$D6
	.dc.b	$58,$7E,$F6,$7F,$BA,$4F,$F4,$9A,$E1,$DB,$FE,$94,$8F,$3D,$F0,$C3
	.dc.b	$E3,$CF,$FB,$A7,$5E,$17,$EB,$9C,$F1,$7C,$AC,$3C,$1A,$FC,$FC,$BE
	.dc.b	$77,$17,$F7,$7C,$3B,$DF,$D0,$E2,$FF,$27,$C6,$27,$E9,$7A,$3E,$2C
	.dc.b	$3E,$84,$FD,$AD,$FE,$63,$F1,$72,$F8,$EF,$8D,$F8,$CF,$97,$C6,$BB
	.dc.b	$FA,$F4,$FD,$0E,$DF,$46,$BD,$51,$F8,$3B,$F6,$D6,$F7,$6D,$F8,$6E
	.dc.b	$6F,$6F,$0F,$CB,$9C,$EF,$F8,$9C,$3D,$13,$9B,$F7,$69,$8D,$3C,$F9
	.dc.b	$7C,$DF,$67,$CB,$BF,$C7,$8F,$B7,$B7,$93,$4F,$0F,$9F,$FD,$FF,$F2
	.dc.b	$5F,$F3,$35,$7F,$83,$7F,$43,$7D,$6F,$E7,$FC,$D7,$E9,$EF,$1D,$D4
	.dc.b	$F7,$D9,$72,$DB,$F2,$DC,$98,$F1,$BF,$CB,$1E,$8F,$F5,$28,$FF,$03
	.dc.b	$D7,$82,$1A,$FE,$76,$30,$FF,$97,$E7,$FC,$0B,$18,$FC,$D7,$F4,$AE
	.dc.b	$B6,$FF,$65,$4F,$F9,$52,$31,$CF,$C3,$1F,$0E,$1E,$CE,$3D,$D7,$17
	.dc.b	$FD,$55,$7F,$0E,$B7,$87,$C1,$DF,$CF,$18,$AF,$F1,$64,$7F,$9B,$6F
	.dc.b	$97,$07,$C6,$F0,$B9,$71,$7C,$BC,$79,$71,$7F,$9B,$2F,$78,$7C,$BF
	.dc.b	$D3,$7D,$19,$F5,$CB,$E7,$DF,$CE,$E7,$F9,$59,$7F,$8E,$75,$F8,$37
	.dc.b	$F3,$3D,$6F,$90,$FA,$23,$E0,$4F,$87,$BA,$F1,$FC,$ED,$BF,$8D,$57
	.dc.b	$F8,$D3,$6F,$85,$57,$0F,$14,$FC,$0C,$3D,$5C,$1F,$E4,$BC,$F1,$97
	.dc.b	$38,$F8,$7A,$7C,$08,$C6,$BE,$AE,$6F,$6E,$1C,$A3,$19,$FB,$2B,$7C
	.dc.b	$1F,$8D,$BF,$E5,$DA,$FE,$17,$F7,$FF,$6E,$3F,$95,$EE,$F8,$2F,$3F
	.dc.b	$91,$F1,$F2,$FD,$DD,$3F,$2F,$3E,$38,$C6,$5F,$99,$DF,$FB,$67,$B1
	.dc.b	$FC,$CB,$87,$FC,$14,$D7,$FC,$A7,$E3,$7F,$86,$63,$FD,$61,$F2,$3F
	.dc.b	$D6,$5F,$E7,$3E,$97,$0C,$BF,$CD,$F5,$4F,$F5,$2C,$7F,$86,$27,$B4
	.dc.b	$59,$FF,$3A,$DE,$8F,$F0,$84,$F8,$67,$FB,$13,$E3,$85,$6A,$B8,$C7
	.dc.b	$FA,$CA,$B0,$7F,$9B,$A7,$FC,$23,$5C,$67,$67,$AE,$AB,$E1,$31,$5F
	.dc.b	$01,$F8,$B1,$A7,$CD,$E3,$FF,$9C,$23,$11,$EA,$8C,$4F,$F5,$5B,$FF
	.dc.b	$20,$75,$7F,$66,$F9,$01,$08,$62,$91,$FD,$B2,$FF,$C7,$8F,$F6,$EC
	.dc.b	$EE,$2E,$EF,$E9,$95,$FE,$2E,$AE,$D5,$FE,$46,$7C,$18,$5B,$E0,$F8
	.dc.b	$A7,$FC,$1D,$3A,$8B,$3A,$75,$5D,$63,$11,$E0,$C3,$CF,$18,$85,$EF
	.dc.b	$0F,$6C,$7F,$61,$AD,$C6,$23,$77,$FE,$3D,$C3,$C5,$1F,$16,$31,$87
	.dc.b	$58,$F1,$DF,$27,$91,$AB,$AF,$8B,$B7,$FD,$F5,$E0,$BE,$43,$FC,$E3
	.dc.b	$9A,$5F,$00,$FD,$71,$F0,$2F,$D9,$65,$72,$C3,$F9,$DB,$7D,$AF,$F0
	.dc.b	$93,$9F,$3A,$FC,$3F,$F9,$0A,$BB,$F0,$FF,$AF,$2B,$97,$A8,$FC,$83
	.dc.b	$FC,$3B,$18,$DB,$F9,$CD,$79,$70,$E5,$7E,$B8,$FF,$64,$D6,$CF,$E8
	.dc.b	$8F,$FB,$5D,$C7,$8D,$AA,$E3,$7E,$FD,$DE,$3F,$FB,$5A,$F9,$47,$C1
	.dc.b	$C3,$95,$F0,$9F,$8B,$C1,$D9,$DB,$F1,$A4,$79,$F8,$7E,$76,$7F,$47
	.dc.b	$0F,$99,$C3,$FB,$8A,$D6,$DF,$4B,$1E,$B3,$F3,$E7,$37,$C7,$1F,$A7
	.dc.b	$C3,$FC,$02,$FE,$97,$7F,$CE,$CF,$BE,$1F,$2B,$4F,$7F,$89,$F3,$DF
	.dc.b	$56,$BF,$0E,$BC,$BB,$FC,$3E,$3F,$85,$5A,$D3,$BF,$89,$EB,$D7,$EF
	.dc.b	$CF,$CA,$F4,$BA,$63,$CB,$66,$DD,$78,$7D,$9A,$BF,$5D,$EB,$7B,$8F
	.dc.b	$C7,$57,$CE,$8C,$69,$F9,$2E,$0F,$97,$8F,$CE,$E2,$6D,$F3,$BC,$2F
	.dc.b	$F6,$2E,$1F,$E5,$0F,$6D,$F1,$EF,$03,$C7,$FB,$26,$3C,$66,$1F,$E7
	.dc.b	$FC,$3F,$8B,$E9,$CF,$DF,$FF,$4B,$F7,$FE,$42,$31,$3F,$E5,$C7,$FD
	.dc.b	$DF,$3A,$DB,$F1,$F1,$F8,$03,$75,$FA,$BD,$37,$7E,$5F,$0F,$FD,$E7
	.dc.b	$87,$19,$F3,$DF,$13,$8E,$08,$F5,$E9,$99,$F4,$D7,$1D,$7D,$DC,$3C
	.dc.b	$74,$F8,$7A,$78,$67,$39,$F9,$67,$39,$F5,$9E,$3B,$67,$1E,$99,$FC
	.dc.b	$8C,$3C,$7A,$78,$B3,$5F,$F0,$9D,$3C,$7A,$FF,$31,$7F,$D2,$5A,$78
	.dc.b	$BA,$AE,$7F,$1E,$BE,$1E,$DE,$5D,$3E,$AE,$BD,$8F,$19,$FC,$A3,$0F
	.dc.b	$3E,$9E,$1C,$FE,$EF,$A9,$E0,$C1,$CF,$DB,$2F,$F4,$8F,$C1,$76,$E1
	.dc.b	$F9,$5D,$1F,$42,$7F,$0D,$A7,$EC,$78,$3E,$04,$63,$C7,$F9,$DB,$78
	.dc.b	$2B,$F1,$5D,$9F,$37,$87,$E2,$FB,$97,$0E,$7C,$5F,$A0,$79,$EF,$ED
	.dc.b	$E4,$7E,$3F,$F8,$5F,$39,$EB,$FF,$B9,$30,$7C,$5B,$E5,$9F,$2D,$FE
	.dc.b	$6A,$F8,$8F,$E6,$7F,$0E,$FD,$67,$FC,$DB,$8F,$1E,$5F,$F6,$06,$7D
	.dc.b	$37,$FC,$1C,$7F,$B1,$A3,$1A,$3E,$17,$77,$E1,$D7,$C1,$7F,$8D,$3A
	.dc.b	$E1,$FA,$98,$F2,$7A,$19,$C7,$9C,$79,$FC,$3C,$61,$CF,$7F,$4F,$81
	.dc.b	$F9,$7B,$DC,$7B,$1C,$4E,$F8,$3F,$D0,$FC,$5F,$A1,$9D,$C7,$C1,$D3
	.dc.b	$E4,$56,$E5,$C3,$FC,$F9,$D9,$8F,$17,$7B,$7A,$72,$F0,$57,$F5,$CD
	.dc.b	$BF,$9A,$DF,$08,$FE,$BF,$E0,$E3,$1E,$D1,$F4,$3C,$BF,$87,$C5,$F5
	.dc.b	$B1,$E7,$9B,$91,$FB,$36,$B8,$FE,$6E,$DF,$27,$CB,$FA,$5A,$7B,$F8
	.dc.b	$3E,$0C,$F9,$78,$7F,$D4,$FC,$3F,$06,$BC,$39,$AB,$C8,$6F,$87,$DD
	.dc.b	$A6,$3B,$BF,$D5,$71,$F7,$BB,$3C,$3C,$1F,$8C,$F5,$3C,$38,$7C,$3E
	.dc.b	$5F,$AE,$F8,$77,$F1,$D7,$B7,$67,$B7,$8E,$1F,$CA,$47,$FA,$F3,$E3
	.dc.b	$7C,$D8,$78,$F9,$BF,$9F,$F1,$F9,$B1,$E5,$B7,$D6,$F4,$3E,$9C,$62
	.dc.b	$BF,$37,$C6,$E9,$F7,$63,$FC,$F3,$DA,$FF,$01,$75,$FF,$21,$E9,$FD
	.dc.b	$1F,$0B,$E2,$38,$5F,$C5,$F3,$7F,$D9,$78,$F2,$7E,$46,$FF,$9B,$37
	.dc.b	$3F,$D2,$3D,$7C,$46,$27,$FD,$00,$FE,$86,$B9,$67,$FE,$13,$8C,$5F
	.dc.b	$78,$FF,$0B,$75,$E2,$3D,$1E,$A7,$F8,$06,$BF,$9D,$5F,$28,$FF,$14
	.dc.b	$D7,$F3,$98,$C6,$AF,$2B,$18,$75,$BF,$F1,$7F,$81,$FB,$4C,$7F,$C1
	.dc.b	$78,$70,$8F,$D9,$C6,$2B,$86,$1F,$1C,$FE,$A5,$5B,$DF,$E4,$D7,$BF
	.dc.b	$3F,$83,$3E,$2B,$F0,$9E,$AB,$FE,$A1,$5E,$A8,$EB,$87,$F9,$52,$3C
	.dc.b	$11,$B8,$FC,$2C,$63,$46,$1F,$6A,$7F,$13,$BF,$F7,$4C,$7E,$FF,$89
	.dc.b	$F6,$31,$FB,$5B,$79,$1F,$9D,$F0,$5F,$66,$BE,$57,$5F,$CF,$CF,$AE
	.dc.b	$FF,$E0,$5D,$7C,$D8,$78,$FB,$2D,$6B,$7F,$C5,$61,$F1,$2B,$D7,$C1
	.dc.b	$FC,$BB,$1F,$47,$07,$9E,$3F,$1F,$E7,$BD,$DF,$F3,$23,$FE,$61,$F9
	.dc.b	$4F,$E7,$7E,$03,$F0,$0F,$E6,$FF,$22,$F8,$F7,$A7,$C7,$D2,$FF,$5A
	.dc.b	$FB,$3F,$90,$E4,$C6,$7F,$E6,$BE,$0F,$C9,$EB,$DA,$9A,$E3,$0E,$2F
	.dc.b	$F5,$7C,$3F,$C0,$D5,$B8,$7E,$E7,$F9,$F8,$C7,$9E,$DF,$F9,$E4,$63
	.dc.b	$F5,$0B,$8F,$4A,$D7,$9D,$FE,$0D,$AE,$37,$ED,$CB,$BD,$EE,$3E,$07
	.dc.b	$5E,$36,$F8,$98,$F7,$E2,$FE,$B3,$87,$F8,$8F,$1E,$71,$FD,$73,$9B
	.dc.b	$19,$73,$8F,$EB,$FE,$4E,$31,$E7,$9F,$CD,$F2,$5C,$BF,$DA,$33,$F9
	.dc.b	$3C,$5F,$1B,$5F,$05,$6B,$F2,$6F,$F3,$8E,$DE,$08,$E3,$CC,$F6,$78
	.dc.b	$E9,$F3,$4F,$4E,$5F,$11,$F0,$ED,$F7,$23,$9E,$FE,$7C,$39,$C7,$8B
	.dc.b	$97,$F0,$75,$F8,$2C,$7F,$D7,$F9,$7E,$23,$B3,$F0,$0F,$DD,$F9,$BF
	.dc.b	$06,$FF,$7B,$A3,$EC,$69,$F4,$F9,$BC,$53,$ED,$E3,$F8,$1A,$FC,$58
	.dc.b	$C7,$C1,$FD,$7B,$F9,$5C,$DF,$4E,$78,$6D,$FC,$4E,$FC,$DC,$9F,$1B
	.dc.b	$1E,$5C,$1E,$8E,$1F,$9B,$E0,$7F,$9A,$FF,$5A,$72,$3E,$3D,$BB,$E4
	.dc.b	$BE,$57,$9E,$F9,$96,$1F,$5B,$E4,$FF,$CD,$AF,$2F,$37,$F4,$1E,$EE
	.dc.b	$B3,$E9,$E7,$FF,$91,$AB,$BE,$9D,$B8,$BF,$9C,$5E,$E3,$C3,$D1,$8A
	.dc.b	$DE,$CF,$94,$C6,$5E,$78,$F3,$78,$BF,$E1,$BC,$79,$47,$A7,$BF,$FA
	.dc.b	$86,$7F,$E6,$7C,$79,$F0,$FE,$17,$D5,$FF,$20,$C6,$29,$C5,$FE,$C6
	.dc.b	$BF,$F3,$54,$E6,$1F,$83,$F3,$D3,$C5,$8F,$C6,$E3,$FB,$3A,$78,$A3
	.dc.b	$18,$72,$AD,$E5,$DA,$FE,$5F,$5F,$D3,$9C,$E5,$F3,$EF,$F4,$79,$FC
	.dc.b	$E9,$CD,$FD,$2F,$0F,$F8,$06,$7F,$26,$BE,$75,$7D,$BE,$EB,$C7,$F7
	.dc.b	$23,$E4,$BF,$7B,$5B,$C1,$F7,$FC,$FF,$95,$A7,$C3,$C7,$E2,$E1,$E4
	.dc.b	$7E,$06,$1C,$77,$F8,$FD,$78,$C7,$E8,$6F,$F8,$7B,$F8,$17,$B8,$F6
	.dc.b	$6D,$FE,$C8,$DB,$E3,$63,$F0,$F3,$FE,$03,$F0,$DE,$FC,$1F,$17,$EA
	.dc.b	$CF,$EE,$6D,$E4,$9E,$9C,$3F,$7B,$5E,$15,$E2,$87,$ED,$BE,$28,$C6
	.dc.b	$5F,$C9,$EB,$7C,$3F,$5B,$C8,$FF,$7D,$F3,$70,$E3,$FF,$09,$78,$1F
	.dc.b	$A4,$F0,$BF,$CA,$7F,$1C,$F4,$9F,$05,$FE,$D2,$C3,$87,$81,$FD,$05
	.dc.b	$B8,$FF,$84,$9C,$F9,$D7,$83,$F5,$19,$FF,$00,$DB,$FD,$89,$5C,$A3
	.dc.b	$F4,$D1,$FE,$C3,$8C,$70,$7F,$45,$8F,$3D,$72,$86,$BF,$E7,$DA,$DB
	.dc.b	$FE,$38,$AF,$F6,$C4,$63,$D8,$E9,$7C,$6B,$FC,$DF,$D7,$E2,$8C,$4B
	.dc.b	$C3,$C6,$9F,$13,$2E,$F2,$E0,$FF,$04,$65,$D5,$F8,$D9,$F9,$AF,$86
	.dc.b	$7F,$2B,$83,$FC,$97,$7B,$7E,$5D,$7F,$95,$33,$F0,$E1,$F3,$F8,$3F
	.dc.b	$8B,$DF,$C5,$AF,$F2,$03,$F1,$FB,$6D,$FA,$FC,$EF,$6E,$7E,$0D,$3E
	.dc.b	$16,$1D,$63,$19,$7B,$38,$BF,$D7,$3B,$7B,$23,$EF,$61,$E9,$9C,$ED
	.dc.b	$E9,$D5,$F0,$9C,$FE,$13,$F0,$1A,$FF,$23,$F1,$7E,$F6,$9F,$8A,$EC
	.dc.b	$F2,$BF,$27,$CE,$FA,$D5,$F4,$7D,$DF,$3E,$9D,$B8,$7E,$67,$C8,$FD
	.dc.b	$8A,$F8,$7C,$AF,$96,$72,$3E,$39,$EC,$7F,$2E,$F9,$67,$C1,$7F,$B3
	.dc.b	$FC,$67,$CC,$FE,$63,$D3,$FF,$7E,$E1,$C3,$8B,$FC,$E1,$97,$F9,$A2
	.dc.b	$31,$E3,$7E,$B6,$3B,$92,$33,$96,$F3,$F0,$57,$F9,$CE,$73,$9B,$DD
	.dc.b	$FF,$68,$46,$3D,$BF,$EE,$D8,$F4,$3D,$58,$7A,$6F,$D1,$CF,$9B,$E1
	.dc.b	$B7,$AA,$FC,$57,$BC,$3F,$C0,$1B,$7E,$7E,$F6,$FC,$7A,$F8,$F1,$8E
	.dc.b	$AF,$E7,$CD,$DF,$38,$73,$C7,$E3,$D7,$F9,$07,$7F,$29,$FD,$9F,$0F
	.dc.b	$80,$FF,$5E,$7E,$05,$7A,$5F,$7B,$EB,$F2,$6C,$FF,$6F,$E0,$FA,$BA
	.dc.b	$63,$81,$18,$AF,$89,$87,$82,$73,$BF,$DB,$C3,$E2,$63,$E7,$D9,$DB
	.dc.b	$F1,$27,$B5,$7E,$06,$7F,$5B,$87,$3F,$99,$F4,$E3,$E0,$BF,$49,$C3
	.dc.b	$4F,$F2,$ED,$76,$8C,$6D,$FE,$05,$8E,$D5,$E4,$AF,$A3,$B7,$F8,$E7
	.dc.b	$83,$E7,$78,$1F,$33,$1E,$39,$FE,$47,$8B,$FD,$89,$F0,$5F,$E1,$2F
	.dc.b	$83,$FF,$5E,$B7,$4F,$F5,$E4,$6A,$FF,$37,$1F,$E9,$C9,$E1,$14,$4C
	.dc.b	$9F,$F4,$64,$F3,$9A,$66,$69,$F9,$C7,$FD,$89,$1E,$19,$E4,$7F,$AF
	.dc.b	$9C,$3F,$CC,$7D,$64,$62,$BF,$D8,$0E,$A7,$FC,$F5,$3A,$AE,$71,$FB
	.dc.b	$8A,$E9,$18,$8F,$DD,$BF,$EB,$CA,$E5,$1E,$2A,$E5,$5F,$ED,$FA,$FF
	.dc.b	$63,$D6,$AB,$E0,$56,$9D,$3F,$EB,$69,$7C,$0B,$E4,$61,$0B,$3B,$8F
	.dc.b	$F7,$7F,$C0,$FF,$91,$9F,$F6,$25,$FF,$B7,$63,$15,$FE,$CE,$6F,$37
	.dc.b	$FA,$ED,$CC,$5C,$3C,$2F,$86,$7F,$DE,$53,$A8,$B3,$A7,$55,$FE,$EE
	.dc.b	$9D,$47,$86,$74,$EA,$7F,$DD,$33,$A8,$E1,$3A,$75,$87,$B5,$F8,$07
	.dc.b	$C3,$AE,$33,$B8,$F0,$CE,$DD,$9A,$C2,$BE,$B8,$FB,$39,$7F,$B0,$5B
	.dc.b	$7F,$9A,$7E,$EF,$26,$63,$F3,$27,$FB,$84,$FD,$14,$FF,$BE,$1F,$CC
	.dc.b	$1F,$C0,$8F,$F2,$DE,$1F,$99,$71,$5D,$9F,$F0,$21,$A9,$FE,$6E,$69
	.dc.b	$FC,$CC,$FF,$0C,$9E,$11,$F9,$16,$BF,$98,$3F,$35,$3F,$CD,$63,$31
	.dc.b	$FE,$88,$3F,$A8,$B8,$8F,$E8,$AE,$A3,$4D,$8D,$C7,$E5,$CF,$F4,$8F
	.dc.b	$C6,$7D,$1C,$7E,$5E,$3F,$95,$8C,$4F,$C1,$3F,$AF,$BB,$CF,$10,$8E
	.dc.b	$53,$86,$CF,$17,$31,$4F,$ED,$35,$FA,$29,$F9,$D3,$A9,$A6,$8D,$57
	.dc.b	$F5,$78,$D4,$7F,$36,$6C,$E6,$38,$1F,$AD,$9F,$ED,$33,$C2,$13,$87
	.dc.b	$66,$E7,$D8,$E0,$F0,$CF,$17,$F3,$0D,$79,$BD,$78,$BA,$4F,$E6,$DF
	.dc.b	$E2,$11,$E1,$EA,$CE,$1F,$9C,$9F,$B9,$3C,$71,$F9,$F8,$78,$44,$FF
	.dc.b	$16,$9F,$57,$CE,$7F,$0A,$6E,$3C,$0F,$D5,$47,$B9,$F5,$98,$AD,$61
	.dc.b	$FA,$C9,$FE,$09,$5F,$E0,$E8,$FB,$B5,$F7,$4F,$AE,$7F,$44,$8F,$D1
	.dc.b	$3F,$74,$FB,$7A,$7E,$05,$FB,$B8,$D7,$F9,$09,$FA,$63,$A1,$87,$93
	.dc.b	$A3,$F5,$73,$E3,$7F,$9C,$1F,$7C,$F1,$9F,$CE,$0F,$C2,$18,$3F,$9E
	.dc.b	$3C,$5F,$E9,$C7,$E0,$8D,$1F,$D1,$1D,$3A,$3C,$B5,$FC,$06,$78,$47
	.dc.b	$94,$FF,$14,$57,$E9,$1E,$67,$9A,$B9,$BA,$7F,$C7,$3E,$1F,$EB,$B1
	.dc.b	$F8,$2F,$CE,$8C,$E9,$F1,$E3,$F9,$25,$7E,$90,$FF,$27,$BF,$CB,$DF
	.dc.b	$78,$7F,$A5,$BB,$31,$5B,$7C,$C6,$1E,$AF,$B2,$3F,$A9,$4F,$CD,$8D
	.dc.b	$4E,$9F,$3B,$A7,$83,$FC,$E9,$AE,$5D,$B9,$6B,$C6,$BE,$1B,$FC,$EA
	.dc.b	$32,$F8,$A3,$F9,$4C,$7E,$27,$0F,$E1,$E7,$E2,$36,$F9,$11,$FD,$EF
	.dc.b	$1D,$4E,$E3,$F1,$D5,$CA,$3F,$1A,$7D,$D8,$FB,$6F,$67,$E5,$BF,$71
	.dc.b	$A6,$4F,$92,$FF,$91,$27,$31,$99,$D3,$F9,$33,$F3,$B3,$F8,$97,$89
	.dc.b	$FD,$2E,$6B,$FE,$4E,$8A,$F5,$F7,$BF,$86,$60,$7F,$99,$C7,$F3,$37
	.dc.b	$F9,$93,$63,$F9,$33,$4F,$EF,$F8,$DC,$FF,$0B,$1F,$E5,$CA,$FE,$67
	.dc.b	$3F,$D7,$5F,$E6,$11,$9A,$FE,$47,$36,$3F,$BD,$1F,$E8,$79,$FE,$5E
	.dc.b	$FF,$A5,$1F,$E5,$8D,$9F,$E3,$13,$63,$FB,$E4,$5A,$FE,$F3,$3F,$EA
	.dc.b	$AA,$FE,$55,$5E,$FA,$FE,$40,$7F,$85,$E1,$F0,$3F,$EB,$0C,$DF,$19
	.dc.b	$FE,$78,$F9,$DF,$F2,$8F,$C7,$FF,$94,$BF,$15,$FF,$22,$E1,$C3,$2F
	.dc.b	$F4,$41,$F2,$71,$E5,$E7,$FF,$5A,$9D,$64,$F0,$F1,$5D,$F3,$FF,$59
	.dc.b	$61,$EF,$8C,$78,$1C,$E3,$FD,$7F,$3F,$E8,$AF,$81,$F9,$95,$C7,$C4
	.dc.b	$6D,$8B,$DC,$63,$8B,$EA,$56,$EB,$FC,$D7,$C7,$F3,$E3,$12,$E0,$C6
	.dc.b	$1C,$AF,$E0,$EB,$E5,$8C,$56,$C6,$FF,$3F,$0E,$78,$7C,$7A,$FC,$74
	.dc.b	$62,$9C,$0E,$CC,$D7,$4B,$F9,$DD,$78,$CB,$FA,$3F,$17,$CA,$E0,$F9
	.dc.b	$37,$F1,$38,$3E,$35,$7D,$8E,$0B,$5F,$EB,$0D,$BE,$24,$E7,$7F,$B7
	.dc.b	$B7,$AE,$18,$74,$79,$7D,$07,$0C,$3F,$05,$C3,$F2,$B3,$F8,$95,$4E
	.dc.b	$98,$7A,$E3,$A5,$7E,$26,$1E,$AF,$CC,$AD,$EF,$FE,$2F,$BF,$75,$EE
	.dc.b	$3F,$D3,$F9,$7C,$B9,$F8,$98,$3E,$3F,$18,$72,$8F,$97,$E3,$3E,$1D
	.dc.b	$F1,$FF,$9B,$F8,$67,$8A,$F8,$27,$CE,$BE,$0D,$F0,$8F,$4D,$F4,$1C
	.dc.b	$F2,$FD,$27,$07,$F8,$D3,$4E,$AB,$7F,$F6,$AB,$FE,$3B,$8C,$C3,$E3
	.dc.b	$78,$57,$8F,$8B,$4F,$97,$D8,$B8,$73,$E0,$F3,$F9,$7F,$E6,$5C,$B9
	.dc.b	$C7,$AB,$DD,$C6,$1E,$DE,$2F,$66,$5F,$EE,$CB,$DC,$7B,$F8,$71,$5B
	.dc.b	$E2,$F8,$3C,$2D,$21,$EF,$3C,$57,$C7,$8B,$31,$9E,$06,$BF,$33,$0F
	.dc.b	$97,$5F,$2A,$BA,$67,$E2,$9C,$E3,$98,$C6,$7F,$46,$1E,$CF,$0D,$3C
	.dc.b	$D3,$9E,$3F,$D6,$DF,$C2,$C7,$E3,$E9,$E2,$ED,$BA,$FD,$E9,$F8,$A7
	.dc.b	$AB,$4B,$5E,$8D,$BF,$05,$7F,$12,$B5,$97,$8F,$97,$C5,$18,$DF,$F0
	.dc.b	$BA,$63,$1E,$59,$7B,$21,$EF,$FC,$5D,$3E,$26,$5E,$9F,$84,$F9,$9E
	.dc.b	$A7,$F9,$03,$7F,$C2,$F1,$7E,$13,$5F,$9F,$E5,$79,$B1,$E3,$97,$D0
	.dc.b	$BF,$8F,$F0,$57,$8F,$E2,$F0,$78,$31,$F2,$61,$E0,$87,$C9,$FD,$CF
	.dc.b	$9B,$FF,$22,$F1,$3D,$57,$C4,$7E,$83,$E0,$9F,$42,$E3,$FC,$27,$CC
	.dc.b	$31,$E3,$AF,$FA,$51,$E7,$E6,$DF,$4F,$FA,$47,$C4,$5A,$FD,$26,$9F
	.dc.b	$EE,$C7,$C5,$3F,$D3,$63,$C3,$1F,$EC,$1B,$F2,$C7,$93,$9B,$15,$BD
	.dc.b	$7F,$6B,$E5,$7E,$4B,$0F,$24,$E6,$BF,$C6,$D1,$F9,$6C,$39,$DF,$B7
	.dc.b	$9B,$17,$BD,$DF,$A7,$E3,$0F,$3D,$FC,$3E,$0F,$E7,$71,$FA,$5C,$79
	.dc.b	$E4,$C3,$E3,$F0,$7B,$2B,$8C,$E6,$FE,$54,$7F,$3D,$BF,$A3,$A7,$C5
	.dc.b	$E1,$FA,$53,$FE,$C1,$9F,$83,$E1,$DD,$7C,$D3,$E5,$AF,$B1,$A7,$96
	.dc.b	$3C,$11,$8C,$BC,$1B,$63,$7F,$95,$B7,$9A,$DF,$45,$F2,$1F,$EE,$E7
	.dc.b	$CE,$3E,$BE,$9E,$17,$E1,$DE,$71,$F0,$F2,$5C,$7A,$F8,$1E,$57,$BF
	.dc.b	$03,$E3,$7C,$5B,$7C,$BE,$2F,$F7,$66,$9F,$2B,$6F,$93,$E4,$FC,$DE
	.dc.b	$1F,$A9,$8F,$C8,$F0,$9F,$3A,$F8,$FF,$E2,$DF,$14,$FA,$13,$AE,$97
	.dc.b	$C1,$FF,$0A,$F8,$A7,$C2,$F2,$C3,$F8,$F6,$DF,$EB,$0C,$BA,$74,$72
	.dc.b	$EB,$FF,$77,$C6,$33,$ED,$3F,$EF,$58,$C6,$9F,$EA,$D6,$DE,$70,$F2
	.dc.b	$CF,$3A,$FC,$B5,$70,$BF,$CF,$C7,$F2,$A8,$C1,$FB,$97,$F9,$8B,$8D
	.dc.b	$38,$DF,$4A,$F7,$75,$F1,$8C,$57,$E6,$B5,$C6,$1F,$13,$1F,$8B,$39
	.dc.b	$BE,$B7,$C3,$1F,$8F,$E5,$E3,$0E,$19,$7C,$BC,$BF,$99,$FC,$5F,$F4
	.dc.b	$5C,$BE,$25,$7C,$6C,$7E,$9F,$9B,$63,$89,$FB,$53,$E8,$1D,$27,$07
	.dc.b	$B7,$87,$CF,$8F,$DB,$3C,$0F,$8D,$C4,$7B,$72,$70,$F8,$F1,$F5,$67
	.dc.b	$F7,$FC,$8E,$F8,$70,$D7,$DB,$7F,$89,$C7,$EE,$7B,$1E,$7A,$F0,$5F
	.dc.b	$93,$1F,$97,$D9,$F4,$35,$F4,$7C,$DD,$BE,$9F,$13,$E8,$E3,$F9,$BA
	.dc.b	$FC,$AF,$0F,$E3,$75,$7F,$9A,$77,$7C,$53,$E5,$7F,$C4,$3E,$A7,$F9
	.dc.b	$18,$7C,$77,$E7,$AF,$FC,$1B,$B7,$E3,$DF,$F3,$97,$47,$E5,$1E,$9C
	.dc.b	$59,$9F,$F0,$86,$4F,$8E,$C7,$8D,$F1,$B4,$FF,$1D,$D7,$E4,$B0,$F2
	.dc.b	$47,$8E,$30,$72,$7C,$DA,$E2,$B9,$6A,$F9,$0C,$65,$EC,$8F,$F1,$DF
	.dc.b	$5F,$E6,$F2,$EB,$1F,$E3,$EF,$7F,$19,$F5,$8F,$83,$CD,$F9,$EA,$DF
	.dc.b	$BC,$E2,$FF,$31,$DF,$CA,$38,$46,$6B,$D7,$D4,$BE,$37,$F2,$EF,$E5
	.dc.b	$1C,$1F,$9D,$E5,$63,$6F,$5E,$3C,$E5,$F3,$79,$DB,$E7,$61,$F0,$CF
	.dc.b	$CB,$E3,$E6,$7D,$78,$72,$CE,$D7,$C3,$C7,$D5,$5E,$D3,$C5,$3F,$D2
	.dc.b	$B3,$F4,$46,$32,$FF,$27,$C6,$3D,$5F,$E2,$99,$7F,$AC,$78,$7C,$79
	.dc.b	$FD,$BF,$0F,$C7,$8F,$D7,$F9,$2B,$3F,$5B,$D0,$FA,$BC,$BF,$8F,$DB
	.dc.b	$E6,$72,$78,$EB,$E0,$ED,$FE,$96,$E4,$FF,$20,$7B,$6F,$96,$FF,$21
	.dc.b	$6E,$F8,$27,$C6,$BC,$77,$ED,$FF,$98,$E4,$E1,$F3,$6C,$B9,$71,$FF
	.dc.b	$98,$CF,$E0,$47,$6C,$FA,$EF,$88,$E1,$1E,$0F,$77,$15,$CB,$5F,$16
	.dc.b	$9C,$A3,$1A,$F9,$35,$FE,$1D,$18,$9F,$37,$97,$FE,$05,$8C,$4F,$A3
	.dc.b	$2F,$F2,$CC,$63,$6F,$E6,$11,$E2,$3F,$19,$87,$58,$FD,$AC,$7F,$99
	.dc.b	$27,$87,$85,$FE,$DE,$6E,$FF,$E9,$F6,$EB,$EF,$AE,$79,$FC,$9C,$3E
	.dc.b	$44,$FC,$BA,$DF,$B7,$E1,$D7,$E2,$61,$7A,$FE,$1D,$FD,$4D,$2C,$F9
	.dc.b	$BC,$2F,$89,$39,$CF,$B6,$FE,$E8,$C6,$FF,$63,$76,$EE,$BF,$36,$BF
	.dc.b	$7F,$E4,$39,$E9,$E8,$CB,$E6,$E5,$D7,$DC,$B3,$EC,$CF,$E4,$BE,$9D
	.dc.b	$BE,$A7,$B9,$F4,$6F,$C3,$C6,$F9,$5F,$25,$F1,$8F,$E4,$3A,$FC,$DA
	.dc.b	$DE,$FF,$CF,$AB,$CB,$8F,$D3,$8F,$4F,$B3,$8E,$1F,$99,$1E,$CB,$FF
	.dc.b	$5C,$F2,$FF,$90,$B8,$5F,$18,$FA,$1F,$E4,$DE,$43,$E5,$FF,$D9,$5C
	.dc.b	$7F,$EB,$EC,$3F,$C9,$3D,$FF,$49,$E5,$E2,$7E,$53,$E4,$BF,$4B,$92
	.dc.b	$3F,$29,$18,$F3,$AF,$17,$F1,$5A,$F0,$75,$FE,$5A,$B8,$EB,$FC,$DE
	.dc.b	$BF,$2D,$18,$F3,$3F,$DC,$51,$FB,$48,$C6,$DA,$8E,$95,$EA,$F2,$FF
	.dc.b	$37,$5A,$F9,$FB,$5A,$C3,$F9,$C7,$07,$F9,$6A,$3F,$6D,$5D,$23,$FD
	.dc.b	$D5,$1D,$CE,$18,$FC,$5F,$8A,$FF,$6D,$EF,$F2,$F5,$F9,$58,$F4,$9F
	.dc.b	$F6,$E5,$F4,$38,$E1,$F3,$3C,$0C,$71,$FC,$BC,$FF,$80,$D7,$C6,$8F
	.dc.b	$AA,$DC,$7F,$AB,$BE,$F8,$FB,$3B,$64,$FB,$3D,$DE,$E8,$CE,$7E,$EA
	.dc.b	$FD,$B6,$1E,$DC,$FF,$DA,$B1,$E2,$C7,$D3,$AF,$B7,$93,$FD,$6F,$B7
	.dc.b	$B7,$4F,$C2,$ED,$F8,$4F,$77,$84,$3E,$53,$ED,$65,$F3,$7C,$BF,$36
	.dc.b	$7F,$13,$F1,$7B,$7C,$BC,$7F,$1D,$FC,$4F,$8E,$BC,$5E,$8D,$FE,$77
	.dc.b	$1F,$D1,$8C,$6D,$F9,$4F,$5F,$C3,$C9,$F3,$F8,$9F,$41,$FC,$BF,$DD
	.dc.b	$FF,$7C,$73,$7F,$87,$7E,$31,$E7,$3C,$87,$C7,$BE,$3F,$F9,$DF,$57
	.dc.b	$FA,$67,$8D,$F0,$7F,$9F,$CF,$9F,$CA,$FF,$32,$7A,$FB,$FF,$9F,$AF
	.dc.b	$E5,$F1,$8F,$86,$FE,$68,$EB,$D9,$FB,$33,$FC,$9E,$31,$EB,$70,$87
	.dc.b	$07,$9F,$AB,$F9,$B5,$71,$E1,$F4,$D7,$F3,$68,$C7,$0B,$A3,$13,$CB
	.dc.b	$4F,$F4,$2D,$FC,$0E,$2C,$E5,$F0,$F0,$F8,$5E,$77,$8F,$0E,$1A,$B0
	.dc.b	$E1,$EB,$E7,$0E,$18,$FC,$8F,$13,$FB,$2E,$BE,$28,$F9,$7E,$1E,$32
	.dc.b	$FF,$67,$9F,$08,$F6,$F8,$FF,$25,$F3,$19,$C3,$FC,$A4,$FC,$37,$FC
	.dc.b	$A9,$85,$DB,$D7,$3F,$10,$F3,$E7,$FB,$97,$C1,$5E,$6E,$7F,$45,$7C
	.dc.b	$2D,$FF,$01,$0F,$1F,$73,$9D,$3F,$07,$D1,$8E,$0F,$85,$87,$08,$6B
	.dc.b	$F8,$DC,$FC,$33,$9F,$4F,$1A,$7F,$AC,$EB,$D9,$97,$D2,$F4,$3D,$13
	.dc.b	$F4,$74,$FA,$73,$E5,$F8,$CB,$A7,$83,$57,$8F,$EA,$8C,$6F,$F9,$7E
	.dc.b	$9E,$F8,$7A,$63,$F3,$5E,$66,$2B,$7C,$2F,$87,$C7,$1F,$E0,$EF,$F4
	.dc.b	$1F,$47,$8E,$3F,$77,$5D,$DB,$E1,$F9,$0F,$90,$FD,$47,$3F,$F8,$D8
	.dc.b	$FF,$0C,$B8,$F1,$FF,$C3,$06,$EB,$FC,$EF,$3F,$E9,$F9,$D7,$6F,$F8
	.dc.b	$75,$E1,$56,$BF,$C5,$B5,$FC,$F8,$12,$7B,$3E,$13,$26,$63,$8B,$C4
	.dc.b	$F1,$1E,$79,$F1,$46,$A3,$B4,$F4,$70,$70,$9E,$AF,$8A,$AC,$EE,$39
	.dc.b	$35,$D1,$E3,$3C,$C7,$91,$D3,$A6,$BF,$E6,$06,$BF,$E1,$17,$FC,$69
	.dc.b	$E0,$BC,$18,$42,$C6,$A7,$E6,$7A,$BF,$22,$BF,$2F,$1F,$5A,$31,$5A
	.dc.b	$8F,$E3,$AE,$9F,$79,$F7,$E3,$71,$C6,$36,$ED,$FA,$35,$E6,$8F,$2C
	.dc.b	$72,$8C,$47,$A7,$AF,$15,$D2,$3D,$95,$F3,$23,$11,$D6,$B8,$BD,$27
	.dc.b	$F0,$53,$A8,$F4,$4E,$9D,$35,$F0,$4F,$AC,$F4,$BF,$24,$E6,$E1,$F9
	.dc.b	$94,$F9,$99,$3F,$EA,$BF,$7D,$E4,$BE,$83,$FC,$51,$3B,$F5,$BF,$A6
	.dc.b	$DF,$0D,$BF,$CA,$ED,$F8,$3F,$F3,$E3,$9C,$7F,$8C,$5F,$E6,$A3,$18
	.dc.b	$FF,$A1,$1B,$ED,$FD,$49,$F2,$47,$FA,$4A,$BC,$7F,$49,$98,$E3,$9F
	.dc.b	$F5,$3A,$E9,$1B,$C3,$D1,$C7,$F7,$A3,$6F,$EA,$DB,$3C,$5A,$72,$7E
	.dc.b	$14,$F2,$7E,$37,$83,$FB,$69,$DF,$B9,$5F,$06,$1F,$19,$FF,$7D,$F3
	.dc.b	$7A,$AF,$9C,$E7,$7F,$F6,$F5,$7C,$59,$F8,$4F,$BF,$0B,$1E,$2F,$07
	.dc.b	$E0,$4F,$68,$F6,$3E,$0C,$2C,$76,$CF,$CD,$E5,$79,$63,$07,$D9,$3C
	.dc.b	$9C,$2D,$FE,$FE,$9E,$5F,$1F,$D1,$EC,$FD,$CC,$FF,$D8,$7D,$3F,$71
	.dc.b	$FA,$9E,$77,$9B,$1E,$5C,$5F,$87,$F6,$7C,$71,$8C,$BF,$CE,$18,$7F
	.dc.b	$0E,$C3,$E8,$F3,$FC,$FA,$D6,$FF,$C8,$76,$F9,$DC,$BF,$C9,$37,$7C
	.dc.b	$9B,$E0,$5E,$97,$FB,$57,$D0,$71,$7F,$9F,$70,$FC,$E1,$C7,$1F,$C9
	.dc.b	$BC,$F9,$2C,$7C,$1C,$6E,$5D,$BA,$BF,$29,$7E,$19,$F0,$57,$E4,$63
	.dc.b	$19,$FF,$B8,$E3,$F1,$D1,$8F,$94,$D4,$75,$9F,$3F,$3E,$A3,$1A,$FF
	.dc.b	$BB,$5F,$89,$7D,$23,$FD,$E1,$C5,$89,$DF,$9C,$DF,$1A,$FC,$4C,$BE
	.dc.b	$1C,$E7,$E0,$3F,$5D,$97,$38,$E1,$C1,$FE,$39,$E4,$4E,$71,$F9,$3C
	.dc.b	$5F,$43,$3E,$F1,$8E,$1F,$97,$18,$C3,$FD,$55,$CB,$8D,$7E,$14,$FC
	.dc.b	$07,$C9,$BD,$E2,$FB,$0F,$F0,$2D,$3C,$35,$CC,$F0,$57,$0D,$7C,$FC
	.dc.b	$38,$C3,$8C,$7A,$B5,$FF,$03,$E1,$CB,$BB,$F8,$8E,$FF,$87,$F8,$CF
	.dc.b	$B7,$87,$F9,$1F,$3F,$17,$D2,$7D,$C7,$EA,$73,$79,$79,$3F,$17,$F0
	.dc.b	$9F,$57,$7F,$17,$77,$CE,$C7,$C3,$ED,$5B,$F9,$FC,$DF,$EE,$9C,$BC
	.dc.b	$D1,$8E,$57,$67,$F0,$79,$FE,$AC,$7F,$24,$29,$FC,$7F,$AF,$F9,$91
	.dc.b	$FD,$D1,$FD,$31,$FD,$77,$AB,$83,$FE,$76,$8B,$E3,$7F,$72,$8F,$E3
	.dc.b	$C7,$68,$FE,$58,$F6,$9F,$E5,$0F,$82,$2B,$FE,$80,$3F,$C1,$B1,$C9
	.dc.b	$C1,$FE,$06,$E4,$FF,$08,$D7,$F6,$89,$F2,$CE,$8F,$F0,$74,$E9,$D1
	.dc.b	$F0,$4D,$1F,$E1,$87,$4E,$9A,$FF,$84,$AB,$85,$7F,$43,$8F,$F0,$A4
	.dc.b	$FF,$88,$BD,$3F,$EB,$9E,$2F,$F8,$16,$31,$3F,$1A,$7F,$BC,$3E,$67
	.dc.b	$FB,$8B,$F0,$1F,$EC,$EF,$BD,$FE,$E0,$EE,$72,$FF,$57,$3F,$D4,$6F
	.dc.b	$27,$07,$FA,$69,$FF,$0B,$F7,$F3,$7F,$C3,$11,$CD,$CB,$C2,$7D,$2D
	.dc.b	$8F,$EE,$D3,$5F,$EF,$11,$F2,$67,$DF,$1F,$E1,$EA,$F8,$4E,$27,$FB
	.dc.b	$C9,$E3,$7E,$2D,$63,$0F,$EE,$B3,$F8,$D3,$FC,$9E,$F0,$CF,$30,$8F
	.dc.b	$F2,$63,$F6,$5F,$F1,$EB,$F0,$1F,$C8,$3F,$82,$7A,$5F,$F9,$59,$FC
	.dc.b	$5C,$7E,$2D,$C3,$4C,$9C,$CC,$94,$C9,$E9,$3F,$08,$64,$E3,$1C,$1E
	.dc.b	$2D,$7C,$2D,$7F,$CD,$ED,$7F,$CB,$DE,$87,$F8,$D3,$2F,$9B,$8F,$F9
	.dc.b	$46,$31,$3F,$91,$0E,$F1,$C3,$C6,$69,$FE,$50,$8F,$CC,$46,$62,$9F
	.dc.b	$E4,$E8,$FF,$0F,$46,$62,$9F,$E4,$78,$FF,$2A,$C6,$62,$9F,$E4,$88
	.dc.b	$FF,$3B,$C6,$62,$9F,$E4,$78,$FF,$43,$C6,$62,$9F,$E4,$B3,$FC,$CE
	.dc.b	$E5,$FF,$48,$B6,$3D,$F1,$FE,$56,$F8,$2E,$B5,$FE,$49,$7E,$F5,$79
	.dc.b	$7D,$5C,$E1,$FE,$4F,$9F,$E4,$95,$C8,$E1,$97,$89,$E4,$7F,$29,$BF
	.dc.b	$74,$FF,$A9,$1B,$1F,$5B,$0F,$F3,$3F,$BF,$FE,$83,$9F,$88,$F8,$23
	.dc.b	$FD,$0C,$FF,$9F,$63,$31,$DF,$0F,$F3,$14,$FF,$99,$A7,$33,$F0,$0E
	.dc.b	$A7,$F9,$E5,$FF,$45,$CF,$E0,$6B,$F0,$39,$7F,$38,$7F,$D1,$38,$FE
	.dc.b	$13,$E7,$7F,$59,$1F,$2A,$7F,$9C,$3F,$82,$7F,$D2,$D3,$F8,$87,$FA
	.dc.b	$1E,$F9,$9F,$E6,$53,$6F,$F4,$31,$CC,$FE,$67,$1F,$88,$AF,$15,$7E
	.dc.b	$19,$F2,$45,$9F,$E5,$F3,$F8,$69,$F4,$47,$FB,$2F,$0F,$6B,$FC,$D8
	.dc.b	$F6,$9F,$CD,$8F,$F7,$03,$A7,$F9,$DB,$5F,$D9,$B5,$FF,$64,$CF,$2D
	.dc.b	$FF,$57,$95,$9F,$80,$23,$F9,$AB,$F9,$6B,$FF,$70,$C7,$F2,$D8,$F4
	.dc.b	$1F,$CB,$63,$E3,$47,$F3,$68,$E0,$7F,$1F,$8F,$88,$FF,$21,$9B,$3F
	.dc.b	$E9,$E8,$E6,$FF,$2C,$78,$BF,$CC,$1B,$19,$7D,$4E,$5A,$F1,$BF,$F6
	.dc.b	$7C,$FF,$34,$AF,$AC,$FF,$0A,$AD,$D7,$F3,$CA,$FF,$6A,$E1,$F8,$B6
	.dc.b	$CF,$F5,$D8,$FA,$3B,$E7,$0F,$F3,$7C,$FE,$16,$7F,$98,$BF,$83,$7F
	.dc.b	$98,$BF,$09,$FE,$79,$5D,$23,$F6,$26,$4F,$F0,$63,$97,$27,$E0,$23
	.dc.b	$F9,$6B,$B3,$F9,$8C,$6D,$FC,$44,$FF,$85,$71,$D4,$7F,$38,$9D,$23
	.dc.b	$F7,$93,$FD,$3E,$BD,$11,$FD,$EE,$BF,$95,$FC,$E7,$F1,$3C,$7F,$98
	.dc.b	$3F,$96,$9F,$E3,$AF,$FB,$F6,$FF,$97,$BF,$EF,$C8,$FE,$5A,$FF,$2F
	.dc.b	$7F,$95,$BF,$EF,$A6,$9F,$EF,$A8,$B9,$FF,$74,$8E,$A7,$FB,$9A,$3F
	.dc.b	$97,$CF,$FB,$D2,$33,$3F,$81,$7C,$E7,$FB,$A6,$B2,$7E,$17,$15,$7F
	.dc.b	$2D,$9F,$E5,$6F,$F2,$FB,$B1,$C6,$86,$2F,$FA,$8C,$FF,$4E,$F9,$8F
	.dc.b	$EE,$BF,$42,$F8,$9F,$F7,$9F,$CD,$7F,$AD,$FE,$29,$ED,$BB,$38,$69
	.dc.b	$FE,$A2,$3F,$AA,$E3,$FD,$FB,$CF,$B1,$FD,$F7,$F0,$BB,$C4,$D3,$F6
	.dc.b	$73,$CA,$31,$F0,$7F,$D6,$B0,$F1,$EB,$FE,$FA,$AD,$5F,$97,$CF,$FE
	.dc.b	$B7,$5A,$7C,$93,$FD,$5A,$31,$B7,$F5,$C6,$F8,$39,$9F,$F0,$85,$F3
	.dc.b	$AF,$F6,$A7,$9B,$FD,$6A,$F7,$DF,$6F,$79,$7C,$4C,$5C,$3F,$12,$7D
	.dc.b	$47,$BB,$29,$5E,$B8,$E3,$BF,$A1,$E7,$A3,$4F,$9D,$1F,$D9,$33,$F3
	.dc.b	$E3,$FD,$A1,$BC,$2F,$9A,$F9,$1E,$97,$6F,$4B,$A7,$17,$CD,$D5,$A7
	.dc.b	$FA,$9B,$3E,$87,$83,$E7,$FE,$23,$E2,$D7,$F7,$3D,$4D,$7E,$4F,$99
	.dc.b	$F1,$F2,$E9,$8F,$9F,$8B,$FD,$3D,$8F,$D4,$CF,$E2,$70,$7E,$3A,$31
	.dc.b	$5B,$8F,$C4,$F2,$7F,$87,$23,$F7,$1F,$06,$F9,$67,$C1,$BE,$91,$DB
	.dc.b	$EA,$F8,$AB,$F3,$5C,$74,$FE,$78,$7F,$89,$32,$E7,$8D,$86,$5D,$7C
	.dc.b	$FF,$F1,$2D,$6B,$3B,$5A,$BF,$0E,$9F,$E2,$6A,$D6,$96,$B5,$7E,$5F
	.dc.b	$0F,$FC,$4F,$5A,$D5,$D1,$FE,$26,$8C,$16,$F8,$E1,$EC,$8F,$F0,$BF
	.dc.b	$2F,$D6,$9C,$DF,$BD,$FF,$12,$CF,$2D,$FF,$C3,$0D,$9C,$CF,$08,$4F
	.dc.b	$C8,$D7,$E3,$6D,$FD,$06,$7E,$7F,$83,$F1,$63,$E8,$6B,$FA,$8C,$3C
	.dc.b	$D1,$8C,$FF,$D5,$39,$F9,$63,$19,$BB,$37,$D9,$CF,$87,$18,$FC,$7B
	.dc.b	$FB,$9F,$25,$F3,$9E,$FF,$49,$9A,$FB,$F9,$FD,$EE,$DF,$0C,$63,$3F
	.dc.b	$CC,$F8,$7F,$E4,$CE,$0F,$97,$F4,$97,$4F,$81,$C7,$F8,$6D,$FF,$CC
	.dc.b	$FF,$03,$F8,$CC,$1F,$0E,$F9,$C7,$33,$E6,$3F,$8F,$7C,$8B,$E6,$F8
	.dc.b	$E9,$F8,$83,$FC,$DF,$97,$37,$FC,$DA,$FE,$63,$C6,$CC,$7E,$3B,$3F
	.dc.b	$E2,$CF,$2E,$5F,$E4,$DF,$1D,$C2,$35,$AF,$F1,$BE,$96,$3E,$78,$F3
	.dc.b	$4F,$F9,$CA,$31,$B7,$EB,$A3,$FC,$CF,$18,$DB,$17,$CA,$7A,$D7,$BB
	.dc.b	$D8,$FF,$3B,$DE,$F7,$FF,$2B,$37,$CB,$FF,$3B,$B9,$F0,$FF,$D1,$F8
	.dc.b	$7A,$76,$F9,$38,$FC,$8C,$3E,$3C,$E7,$4F,$99,$A7,$CB,$C1,$C3,$F3
	.dc.b	$B5,$F2,$6F,$FD,$1F,$8B,$E1,$3F,$D2,$70,$BA,$7D,$69,$FF,$4A,$EB
	.dc.b	$ED,$F0,$7D,$3A,$7B,$34,$FB,$0F,$A0,$C6,$BF,$D5,$DF,$8A,$63,$57
	.dc.b	$BF,$8D,$3C,$1B,$FA,$E7,$CF,$8F,$E0,$3D,$1F,$F1,$36,$7F,$C5,$23
	.dc.b	$EE,$DF,$0D,$3E,$A6,$5F,$3B,$C4,$FF,$53,$F1,$FF,$84,$7E,$0B,$F1
	.dc.b	$BF,$05,$FC,$EF,$E1,$1F,$24,$F8,$F4,$BC,$7F,$F5,$EE,$1C,$72,$FF
	.dc.b	$5E,$BC,$F9,$2C,$32,$FC,$CE,$DF,$EB,$EA,$D6,$16,$B5,$7D,$B9,$7F
	.dc.b	$D7,$F5,$AF,$3E,$D6,$AF,$FA,$86,$FE,$A8,$C4,$BE,$1F,$19,$FB,$21
	.dc.b	$BF,$F1,$EA,$FF,$5F,$E3,$D6,$1F,$45,$FE,$C6,$8C,$70,$7F,$A9,$5B
	.dc.b	$C1,$C2,$FB,$DF,$C6,$86,$FF,$EC,$4A,$DF,$C2,$5B,$DE,$1F,$2F,$6F
	.dc.b	$F1,$0F,$5A,$FE,$47,$07,$C3,$DB,$D1,$C3,$EA,$3E,$8E,$FE,$AC,$BC
	.dc.b	$C7,$43,$C9,$3E,$FD,$BE,$CE,$38,$BE,$31,$E7,$C3,$C9,$5B,$CF,$C9
	.dc.b	$87,$DC,$3C,$FB,$7E,$03,$1C,$79,$FF,$E7,$3D,$3F,$11,$C5,$F7,$B1
	.dc.b	$72,$5D,$5C,$1E,$B7,$F1,$3C,$3F,$6A,$BF,$1F,$E9,$D9,$F2,$F5,$FB
	.dc.b	$27,$C9,$F1,$FE,$9E,$4F,$CA,$70,$FC,$98,$F3,$4F,$A7,$83,$F2,$A6
	.dc.b	$30,$F8,$FE,$53,$E1,$1F,$42,$FA,$4F,$EE,$DC,$9F,$47,$97,$F1,$3C
	.dc.b	$1F,$3F,$6F,$F0,$E6,$37,$A3,$FB,$E6,$7E,$53,$AF,$8F,$6B,$7A,$7E
	.dc.b	$37,$1F,$14,$63,$4F,$56,$1F,$E0,$5A,$F1,$F4,$7D,$2C,$7C,$F1,$E6
	.dc.b	$D7,$18,$71,$7F,$1B,$AF,$9F,$7F,$CE,$6B,$ED,$CF,$D9,$F4,$BE,$0D
	.dc.b	$3F,$B2,$6B,$C9,$FF,$0A,$6F,$E7,$7C,$51,$E4,$BE,$5E,$A6,$DE,$99
	.dc.b	$FF,$99,$F3,$F0,$6D,$F9,$2F,$43,$E2,$57,$CD,$E5,$E5,$C0,$F9,$7F
	.dc.b	$D3,$7A,$FF,$C6,$BC,$D7,$C9,$3E,$5B,$F6,$7E,$03,$E2,$BF,$C9,$5B
	.dc.b	$7D,$3F,$82,$FC,$77,$13,$2F,$E4,$6F,$F0,$77,$F9,$2F,$15,$DB,$FC
	.dc.b	$13,$3F,$C1,$A7,$C1,$36,$7F,$C1,$BF,$13,$75,$F1,$E3,$FC,$26,$FC
	.dc.b	$DA,$F9,$71,$FD,$6C,$63,$5F,$F0,$2C,$7E,$EE,$31,$E2,$6B,$2F,$F0
	.dc.b	$CD,$FB,$61,$EE,$FC,$29,$CE,$1F,$01,$E4,$62,$5B,$FC,$3C,$7F,$65
	.dc.b	$8F,$68,$F8,$DD,$98,$9F,$E5,$B3,$E8,$7D,$57,$DB,$97,$C9,$AF,$D0
	.dc.b	$CB,$E7,$D7,$82,$31,$97,$83,$DC,$C6,$BE,$9D,$3E,$A7,$85,$E2,$DF
	.dc.b	$BF,$D2,$78,$35,$FC,$6E,$7E,$DF,$5B,$E9,$69,$88,$F4,$CF,$C4,$F2
	.dc.b	$FC,$B3,$F8,$2F,$96,$BA,$7E,$0F,$CC,$F3,$BF,$27,$9B,$E4,$57,$9B
	.dc.b	$6F,$C3,$77,$F4,$8F,$C5,$FC,$E3,$E4,$7F,$D8,$1E,$77,$F8,$53,$87
	.dc.b	$F7,$9C,$FF,$73,$7E,$F7,$31,$EE,$BE,$4A,$E3,$FC,$FF,$87,$F2,$17
	.dc.b	$F7,$2B,$B7,$B1,$E4,$DD,$1E,$1F,$53,$FA,$66,$9E,$0B,$E1,$8F,$93
	.dc.b	$1F,$46,$1C,$23,$CF,$D5,$8C,$79,$6C,$F9,$4C,$57,$93,$3F,$07,$D2
	.dc.b	$FA,$B0,$C6,$7F,$63,$1F,$17,$CC,$5C,$3E,$1F,$A7,$E0,$F4,$5F,$3A
	.dc.b	$E0,$FF,$26,$47,$93,$BB,$E9,$B3,$C1,$FA,$6B,$E9,$2D,$FF,$8B,$B8
	.dc.b	$FA,$2F,$92,$F8,$95,$DB,$1F,$87,$EB,$AF,$E5,$F5,$7F,$02,$AD,$7B
	.dc.b	$7F,$E9,$0B,$F9,$13,$9C,$7C,$D3,$FC,$2E,$73,$CF,$FE,$14,$ED,$F1
	.dc.b	$E3,$E0,$BF,$66,$FF,$9A,$CB,$AC,$FD,$0E,$8F,$6D,$7D,$DE,$CF,$DF
	.dc.b	$D7,$8B,$97,$E2,$75,$F9,$E7,$D1,$84,$26,$9F,$87,$7F,$8A,$19,$37
	.dc.b	$1C,$97,$91,$97,$8D,$75,$D3,$11,$F1,$6F,$A5,$F4,$8E,$11,$85,$3C
	.dc.b	$23,$13,$6B,$D1,$5F,$3E,$B8,$C6,$23,$A1,$EC,$AE,$8E,$CF,$29,$AA
	.dc.b	$F6,$46,$A3,$47,$9E,$7A,$32,$34,$E9,$AF,$AD,$AF,$A8,$F8,$D3,$F0
	.dc.b	$8E,$8F,$98,$FA,$15,$F9,$7E,$5F,$C5,$7C,$13,$E3,$61,$0B,$87,$F0
	.dc.b	$3B,$FF,$78,$9F,$38,$E4,$7F,$BD,$3C,$3C,$46,$EB,$FE,$00,$4F,$D0
	.dc.b	$BE,$F7,$FE,$5F,$8C,$1F,$EF,$C7,$6F,$E1,$23,$71,$B8,$C1,$B7,$6F
	.dc.b	$DF,$8D,$C6,$E3,$06,$DD,$D7,$95,$F7,$5F,$4B,$F8,$11,$E2,$9D,$3A
	.dc.b	$AE,$73,$A8,$F6,$CE,$9D,$35,$F9,$53,$F1,$A7,$84,$7C,$FA,$FE,$11
	.dc.b	$E8,$7E,$97,$C4,$FF,$35,$F7,$7F,$B5,$3B,$7F,$D2,$FF,$22,$F8,$ED
	.dc.b	$E7,$FA,$23,$E7,$E1,$FA,$DD,$2C,$7F,$7D,$C6,$EE,$7F,$B7,$C7,$EC
	.dc.b	$2B,$C1,$E3,$7E,$85,$F1,$79,$5F,$E0,$49,$D1,$CD,$69,$FA,$A9,$DB
	.dc.b	$FE,$D3,$9F,$2E,$1D,$EB,$7E,$E7,$E8,$2F,$7E,$1F,$5B,$DE,$5F,$02
	.dc.b	$3A,$C7,$EA,$E7,$7B,$78,$F4,$FF,$27,$4F,$8F,$3F,$8D,$E5,$6E,$31
	.dc.b	$87,$F4,$AE,$77,$C9,$7F,$03,$CB,$9C,$38,$5D,$9E,$2D,$FD,$B8,$71
	.dc.b	$97,$C7,$7B,$32,$F9,$7B,$FB,$DE,$87,$C5,$3E,$21,$E2,$DF,$E4,$F0
	.dc.b	$FC,$3D,$FE,$E6,$9F,$A1,$E0,$F5,$CE,$73,$F0,$61,$F2,$B9,$3E,$CF
	.dc.b	$B9,$E0,$DB,$FA,$34,$FB,$EF,$E7,$7C,$D7,$81,$ED,$C1,$F3,$E7,$E1
	.dc.b	$E5,$E2,$DB,$E7,$6D,$FD,$3E,$3C,$31,$8D,$7E,$27,$8F,$FD,$6F,$C6
	.dc.b	$FE,$0B,$F4,$8F,$8F,$7D,$07,$FB,$47,$E6,$9F,$42,$F9,$87,$C6,$71
	.dc.b	$DF,$F8,$39,$FC,$B7,$2F,$E1,$FE,$77,$F0,$67,$AC,$64,$78,$4D,$3F
	.dc.b	$D8,$31,$FC,$E2,$7C,$7F,$2B,$BA,$D6,$BF,$8B,$AF,$E8,$15,$E8,$8F
	.dc.b	$3C,$7F,$30,$8C,$6B,$FE,$44,$8F,$EA,$31,$8F,$83,$FC,$A7,$17,$F7
	.dc.b	$59,$F1,$E5,$F0,$6F,$FD,$E1,$EE,$70,$9C,$E3,$F1,$30,$70,$7F,$97
	.dc.b	$31,$E5,$9F,$9B,$1F,$91,$F1,$FF,$D1,$F4,$F0,$5F,$CA,$BF,$4E,$5D
	.dc.b	$2F,$E7,$70,$BE,$8F,$39,$73,$CF,$BC,$FD,$1E,$3F,$A1,$3F,$13,$0F
	.dc.b	$90,$73,$AF,$84,$D9,$F5,$4F,$3C,$AE,$FE,$FB,$EB,$7E,$08,$C4,$FE
	.dc.b	$03,$93,$1B,$FC,$18,$FC,$1F,$26,$35,$F4,$BF,$AE,$DB,$E1,$E1,$CB
	.dc.b	$5F,$C4,$7A,$1E,$8F,$63,$F0,$9A,$7E,$0B,$E0,$EE,$FF,$83,$F0,$3F
	.dc.b	$01,$8F,$E4,$F2,$F6,$CF,$E0,$BA,$AC,$FD,$7F,$33,$DF,$3E,$DE,$3F
	.dc.b	$A1,$97,$FA,$13,$E1,$5F,$44,$F8,$9F,$E5,$1F,$0C,$F6,$1F,$21,$F9
	.dc.b	$BF,$83,$FE,$C1,$8F,$E7,$7E,$03,$FD,$A8,$FC,$C7,$FD,$C6,$FE,$6D
	.dc.b	$FF,$69,$BF,$49,$FF,$73,$FB,$19,$CF,$C3,$E5,$65,$F1,$E7,$74,$F2
	.dc.b	$7A,$FD,$67,$3E,$47,$F0,$2E,$4F,$CB,$65,$FE,$F2,$BF,$54,$FF,$3F
	.dc.b	$CF,$DF,$39,$C7,$FA,$27,$3E,$31,$E7,$E3,$B8,$71,$97,$28,$C7,$3F
	.dc.b	$E5,$31,$E5,$9F,$F4,$BB,$7C,$E6,$71,$E7,$1F,$27,$5F,$8D,$AF,$DC
	.dc.b	$C1,$5F,$E1,$78,$F6,$67,$F0,$CF,$E3,$AF,$37,$D6,$FE,$BE,$7C,$39
	.dc.b	$BC,$7F,$3E,$DF,$63,$6E,$0F,$DA,$E4,$BB,$79,$D7,$83,$F6,$4F,$27
	.dc.b	$C3,$79,$75,$FC,$1F,$D2,$FD,$1C,$3C,$DA,$F2,$E4,$F1,$57,$D2,$F3
	.dc.b	$AF,$07,$C7,$F8,$BF,$C6,$F3,$37,$FE,$47,$F1,$FF,$E7,$7F,$83,$7C
	.dc.b	$EF,$FB,$D7,$A7,$F9,$35,$FE,$33,$9F,$F8,$83,$FC,$83,$6F,$DC,$BF
	.dc.b	$E3,$16,$E3,$F9,$07,$BB,$FC,$C1,$CF,$9F,$FD,$56,$B9,$CF,$8A,$7F
	.dc.b	$80,$D7,$96,$99,$7F,$91,$6F,$71,$8A,$FF,$15,$46,$27,$D5,$1E,$09
	.dc.b	$D6,$5F,$6E,$B5,$28,$7C,$16,$EF,$E1,$67,$F4,$AB,$E0,$E5,$E5,$BF
	.dc.b	$8B,$E7,$FF,$22,$CF,$CD,$87,$08,$FC,$7E,$5E,$CF,$47,$FA,$45,$6B
	.dc.b	$1F,$06,$5D,$DB,$E0,$F3,$CF,$B4,$FF,$B8,$30,$FE,$7D,$E6,$FD,$1F
	.dc.b	$3F,$FD,$59,$8F,$F7,$DC,$78,$79,$3D,$B1,$FE,$BB,$9F,$CE,$F9,$9F
	.dc.b	$9F,$B7,$E1,$AF,$D7,$C3,$F2,$7E,$1F,$FA,$36,$5E,$5E,$1F,$CC,$57
	.dc.b	$F3,$EE,$E7,$C4,$3E,$41,$F2,$BF,$E1,$4F,$61,$F4,$9F,$E3,$9F,$0B
	.dc.b	$FB,$66,$1E,$7C,$4F,$47,$F9,$FB,$D1,$FE,$D8,$F6,$E1,$B9,$77,$9D
	.dc.b	$BE,$1F,$12,$E7,$FD,$C6,$7F,$A3,$DF,$9A,$3C,$BC,$58,$BF,$46,$9F
	.dc.b	$E2,$A8,$FE,$C9,$5B,$8F,$55,$FF,$30,$8C,$4B,$4F,$C4,$72,$7F,$B1
	.dc.b	$70,$E7,$39,$F9,$3E,$11,$8A,$F8,$7C,$FF,$26,$31,$BB,$BF,$1A,$F9
	.dc.b	$AF,$E4,$F8,$CE,$CF,$93,$E7,$7F,$4E,$DB,$E8,$F0,$FC,$4C,$3E,$8F
	.dc.b	$07,$B5,$F7,$F7,$FB,$35,$EF,$3D,$B8,$BE,$B6,$DF,$CB,$F7,$FF,$04
	.dc.b	$65,$EE,$DB,$C3,$DD,$F6,$3C,$1F,$06,$5F,$7F,$97,$E7,$E5,$FB,$BC
	.dc.b	$78,$CF,$CC,$F7,$3C,$53,$FE,$56,$F8,$2F,$16,$FE,$DD,$7F,$C9,$5E
	.dc.b	$6B,$77,$9C,$F9,$FF,$95,$F0,$9F,$C4,$FD,$D7,$AC,$FA,$47,$03,$2E
	.dc.b	$59,$FF,$13,$7A,$65,$FC,$D3,$3F,$F6,$F4,$63,$3F,$F5,$FB,$7B,$7F
	.dc.b	$DD,$4E,$79,$6D,$79,$2B,$F0,$1D,$3F,$EF,$2A,$D7,$CB,$5A,$D5,$BB
	.dc.b	$BF,$D2,$F1,$E6,$AE,$91,$FE,$A3,$8F,$80,$7F,$BF,$E3,$3C,$1F,$CA
	.dc.b	$63,$81,$D2,$7F,$D5,$5C,$DF,$CB,$76,$F8,$13,$9A,$F8,$CF,$33,$12
	.dc.b	$E1,$FD,$CD,$70,$DB,$E7,$65,$E7,$C3,$E7,$70,$F1,$9C,$E5,$E0,$C3
	.dc.b	$E8,$EF,$E4,$9C,$DF,$0F,$A4,$C6,$9F,$30,$FB,$7E,$0D,$AF,$B9,$D9
	.dc.b	$F3,$1E,$B3,$F0,$8F,$07,$17,$E0,$37,$F4,$6A,$BF,$07,$B7,$CE,$9C
	.dc.b	$FC,$DF,$D8,$76,$73,$C7,$CD,$1F,$B3,$FA,$5C,$71,$FC,$18,$FC,$6E
	.dc.b	$3E,$18,$C7,$1F,$C3,$F8,$AC,$72,$FD,$4D,$BC,$59,$79,$AB,$E4,$43
	.dc.b	$E5,$7F,$1B,$3F,$43,$97,$FD,$03,$8F,$4E,$6F,$95,$F2,$7F,$57,$C8
	.dc.b	$7C,$FF,$E7,$3C,$FF,$D0,$78,$5F,$9B,$F9,$1F,$F6,$BE,$3F,$7A,$7F
	.dc.b	$39,$9B,$07,$AB,$F3,$DE,$71,$F9,$87,$AF,$8F,$72,$FD,$4C,$FE,$5E
	.dc.b	$7C,$3D,$3F,$E8,$B8,$F7,$9D,$67,$C9,$8F,$C8,$E2,$FE,$9B,$1E,$7E
	.dc.b	$5C,$56,$ED,$9F,$EE,$72,$E7,$97,$68,$F6,$D7,$C6,$8C,$5F,$F1,$A8
	.dc.b	$F8,$98,$75,$D7,$F3,$B1,$FA,$2A,$DC,$7E,$FF,$C0,$DE,$5E,$0C,$3F
	.dc.b	$A9,$E6,$F9,$6C,$E7,$F3,$21,$BF,$F8,$AB,$6F,$D4,$E7,$F3,$0F,$E0
	.dc.b	$DA,$70,$3D,$AF,$A7,$B6,$ED,$F0,$75,$F6,$BE,$77,$D8,$FA,$71,$EF
	.dc.b	$7E,$98,$C6,$2F,$88,$F6,$7D,$07,$D9,$9F,$E0,$15,$CB,$8F,$EC,$70
	.dc.b	$7F,$75,$8E,$4F,$CE,$EE,$F1,$BF,$07,$C9,$F1,$BF,$0F,$8F,$FD,$85
	.dc.b	$5C,$38,$7E,$17,$27,$C4,$E4,$F8,$FB,$7F,$2A,$E3,$FE,$17,$E5,$3D
	.dc.b	$3F,$8F,$E5,$FE,$5B,$A7,$FD,$CB,$F1,$2F,$6F,$F8,$AD,$FE,$50,$FF
	.dc.b	$08,$6B,$8C,$3F,$81,$47,$3E,$7F,$E5,$6F,$5E,$AB,$3F,$53,$C8,$FE
	.dc.b	$55,$3E,$1D,$7F,$93,$61,$F9,$48,$C6,$9F,$A0,$8C,$5F,$9B,$9B,$F2
	.dc.b	$75,$A7,$FC,$C3,$B7,$B2,$75,$E4,$F0,$C5,$8F,$B7,$C5,$FD,$CE,$3C
	.dc.b	$78,$1E,$B6,$31,$ED,$0F,$55,$7F,$CB,$EF,$84,$3E,$7B,$17,$C2,$1C
	.dc.b	$7F,$B8,$D7,$C9,$A7,$C2,$3C,$7B,$FA,$AF,$E8,$CD,$9F,$3E,$FF,$D3
	.dc.b	$6B,$C9,$18,$D3,$FD,$F1,$DD,$FD,$AF,$4F,$1D,$78,$AB,$59,$B8,$71
	.dc.b	$C1,$F7,$63,$CB,$5F,$66,$B7,$AF,$E1,$71,$E9,$AF,$AE,$3F,$8C,$EB
	.dc.b	$F8,$8C,$BE,$FF,$CF,$FE,$02,$BF,$C7,$98,$FC,$FE,$0F,$BB,$1F,$8C
	.dc.b	$EE,$FE,$67,$D9,$FE,$01,$E0,$FF,$52,$7C,$F7,$D0,$F2,$5F,$3D,$FE
	.dc.b	$73,$F8,$57,$CF,$7F,$80,$30,$E1,$A7,$F3,$F3,$FB,$5E,$FF,$EE,$47
	.dc.b	$FB,$97,$76,$5F,$F0,$74,$67,$3F,$F1,$3C,$71,$8C,$4F,$FA,$56,$37
	.dc.b	$5E,$58,$FF,$4B,$47,$A0,$EB,$9E,$EB,$8B,$F8,$88,$D5,$FB,$3A,$7F
	.dc.b	$A4,$C7,$9B,$0E,$95,$EF,$D3,$FA,$94,$62,$BF,$B0,$7C,$1E,$2F,$BC
	.dc.b	$7F,$63,$D7,$F4,$51,$F0,$CC,$DF,$C8,$F7,$3B,$46,$27,$FD,$39,$9F
	.dc.b	$B2,$31,$EE,$37,$C6,$BE,$7C,$7E,$8C,$E5,$ED,$1F,$ED,$9E,$4F,$99
	.dc.b	$87,$C0,$C3,$E5,$6B,$EC,$9B,$1E,$4F,$37,$D5,$7D,$73,$BC,$0E,$0F
	.dc.b	$C0,$47,$FA,$C6,$FC,$73,$9D,$FE,$06,$BF,$EB,$1D,$3E,$7C,$63,$6F
	.dc.b	$F6,$C4,$7E,$07,$BB,$FD,$A9,$B7,$E3,$75,$FC,$16,$5E,$AC,$BA,$6D
	.dc.b	$F5,$9E,$DF,$29,$F7,$7C,$4F,$DE,$FC,$0F,$C5,$F2,$DE,$F3,$E4,$DF
	.dc.b	$10,$F8,$EF,$A5,$F1,$1E,$0D,$7F,$B1,$F7,$78,$38,$BF,$29,$A7,$F0
	.dc.b	$8E,$0F,$93,$C3,$FE,$61,$8F,$E1,$7E,$F6,$30,$E3,$BF,$F9,$67,$0E
	.dc.b	$38,$A3,$C9,$3F,$85,$BF,$3F,$23,$83,$F6,$38,$70,$BF,$57,$C9,$EF
	.dc.b	$1E,$57,$BE,$1F,$66,$5F,$85,$CB,$9E,$5F,$03,$87,$FC,$43,$9F,$3C
	.dc.b	$38,$47,$C2,$9F,$C9,$57,$C1,$3F,$92,$F1,$FA,$DF,$97,$87,$E1,$1F
	.dc.b	$9B,$37,$2F,$1C,$FD,$17,$CB,$1F,$CE,$B6,$F8,$D1,$8C,$FD,$5A,$FD
	.dc.b	$57,$F4,$7A,$79,$6F,$D1,$BF,$AE,$F8,$69,$F6,$32,$F3,$C6,$38,$BF
	.dc.b	$9D,$E1,$F5,$B0,$FB,$19,$FD,$5F,$87,$B8,$7D,$4D,$FF,$D2,$2F,$89
	.dc.b	$7C,$A7,$A7,$6F,$27,$CF,$FE,$CB,$CA,$FF,$4B,$FC,$4F,$E2,$3E,$1D
	.dc.b	$F1,$AF,$82,$73,$BE,$23,$F7,$9E,$DF,$E2,$F1,$FF,$0A,$ED,$FB,$F7
	.dc.b	$FD,$43,$EA,$7E,$1D,$EB,$1F,$B0,$7B,$F8,$D7,$91,$AF,$E6,$3D,$CF
	.dc.b	$E3,$55,$E4,$E6,$FF,$69,$61,$FC,$3E,$31,$BF,$D1,$8F,$F7,$BE,$1C
	.dc.b	$AB,$FD,$B1,$BE,$E3,$15,$EC,$F0,$7E,$04,$62,$7D,$D7,$FC,$47,$0E
	.dc.b	$B1,$F0,$7B,$71,$5B,$E1,$F8,$6F,$BE,$FE,$8D,$FC,$AE,$BF,$66,$BF
	.dc.b	$6E,$7C,$4F,$ED,$2B,$DA,$78,$76,$F1,$C7,$D1,$CB,$E8,$6F,$E0,$8C
	.dc.b	$65,$F4,$B0,$F9,$DB,$F8,$3E,$6B,$C3,$97,$F5,$C9,$F6,$FB,$3F,$46
	.dc.b	$3E,$27,$C6,$78,$31,$ED,$F1,$1E,$0E,$3F,$8F,$9F,$DE,$CF,$FD,$B3
	.dc.b	$8F,$FB,$87,$5F,$F7,$17,$4B,$E5,$DF,$3F,$FE,$9E,$E1,$C7,$27,$0E
	.dc.b	$0F,$C4,$9F,$C3,$71,$E5,$C9,$63,$3D,$FF,$CD,$A7,$B4,$3E,$53,$0F
	.dc.b	$2C,$FC,$0F,$FA,$9A,$77,$1E,$3E,$6C,$4E,$F4,$F2,$BF,$EA,$9A,$DC
	.dc.b	$7A,$39,$F1,$3B,$D7,$FD,$6E,$FF,$76,$AD,$C7,$B6,$3F,$BA,$C6,$3E
	.dc.b	$21,$E4,$EE,$79,$47,$FB,$0F,$0D,$C7,$A4,$FE,$F5,$5E,$8E,$1B,$1E
	.dc.b	$7D,$FD,$3C,$3F,$3A,$FE,$6F,$0F,$9A,$31,$9F,$CF,$C3,$AC,$7F,$9C
	.dc.b	$70,$A7,$FA,$1B,$9D,$97,$C0,$F9,$FE,$B3,$FC,$0B,$2F,$B3,$C9,$D6
	.dc.b	$31,$97,$FB,$63,$7F,$67,$93,$FE,$33,$F8,$6F,$C0,$7C,$8B,$E5,$3F
	.dc.b	$9D,$75,$7F,$14,$D5,$93,$E3,$3F,$AD,$BF,$83,$C7,$FC,$78,$FD,$C7
	.dc.b	$FC,$7D,$C5,$73,$FF,$14,$CF,$C4,$9F,$06,$DF,$E5,$A8,$FA,$F1,$8F
	.dc.b	$9B,$FB,$EE,$B4,$FE,$99,$3E,$8B,$FF,$70,$5A,$5F,$33,$CA,$73,$1E
	.dc.b	$BE,$BE,$1B,$FB,$B7,$FE,$AD,$5F,$6B,$2F,$5F,$1F,$CC,$CD,$1F,$E5
	.dc.b	$7E,$9F,$56,$13,$9B,$D5,$87,$8A,$FC,$EF,$81,$F1,$4F,$77,$C3,$AF
	.dc.b	$9F,$2F,$EC,$DB,$F8,$6F,$CF,$1F,$42,$BE,$67,$97,$FA,$BC,$FF,$CE
	.dc.b	$B1,$E0,$C7,$F6,$DA,$F8,$1F,$B3,$AF,$83,$3E,$CF,$83,$6F,$B9,$D7
	.dc.b	$F9,$BF,$3B,$FD,$55,$F4,$BF,$E8,$7F,$1B,$E4,$FC,$0F,$DD,$FA,$5F
	.dc.b	$DF,$56,$BE,$17,$FE,$00,$73,$7F,$A4,$7A,$3F,$DB,$5D,$3F,$E0,$9E
	.dc.b	$7F,$E0,$16,$F8,$5B,$F0,$FF,$63,$93,$F7,$33,$FE,$B9,$CF,$F8,$0B
	.dc.b	$E2,$E6,$FD,$3B,$DE,$77,$4D,$FF,$D3,$93,$87,$D1,$CB,$FB,$E8,$F3
	.dc.b	$70,$7F,$B7,$6F,$DD,$7C,$77,$F4,$E3,$8C,$7E,$06,$FF,$E6,$CE,$3F
	.dc.b	$C0,$65,$CA,$3F,$D5,$DE,$57,$EF,$A3,$1C,$4F,$92,$C6,$7D,$75,$F9
	.dc.b	$3C,$7E,$7C,$B9,$CE,$72,$F9,$1A,$7C,$CD,$BE,$3D,$7C,$67,$E1,$D7
	.dc.b	$FA,$05,$FA,$2F,$C1,$7E,$2B,$F0,$FA,$AE,$DF,$66,$7E,$14,$F7,$7D
	.dc.b	$DB,$FD,$8A,$B9,$F9,$A7,$AD,$FF,$98,$5F,$99,$4D,$ED,$7F,$83,$A7
	.dc.b	$DF,$5F,$23,$D9,$FE,$49,$3F,$D1,$BE,$17,$FA,$47,$9B,$FE,$A6,$F3
	.dc.b	$9F,$19,$FE,$FD,$E9,$FE,$13,$7F,$AE,$6F,$83,$FD,$1E,$3F,$9D,$E7
	.dc.b	$FC,$A1,$FE,$A3,$E4,$7F,$46,$7F,$A9,$FB,$17,$C6,$69,$FD,$5F,$8F
	.dc.b	$FA,$2D,$78,$FC,$06,$D8,$BD,$C6,$3C,$4F,$E8,$B5,$BA,$FF,$5F,$E7
	.dc.b	$EC,$BD,$C7,$FB,$03,$BF,$19,$7B,$BE,$0B,$FC,$2F,$7F,$D8,$F0,$E5
	.dc.b	$0E,$16,$DF,$D1,$A7,$31,$FD,$63,$5F,$E1,$D1,$FD,$EE,$FC,$1C,$9E
	.dc.b	$DC,$66,$5F,$0F,$4F,$4D,$7A,$E7,$E7,$F0,$7A,$A7,$CF,$A5,$9F,$36
	.dc.b	$5E,$8D,$FE,$1C,$F9,$78,$BC,$9E,$4F,$D1,$C3,$E2,$E7,$D6,$DF,$4B
	.dc.b	$78,$7F,$DC,$3E,$97,$37,$EF,$6B,$FD,$8B,$D0,$7C,$53,$C1,$FF,$05
	.dc.b	$65,$FE,$95,$FA,$1F,$E6,$DC,$FF,$E0,$9B,$FC,$17,$19,$FE,$BC,$E7
	.dc.b	$E6,$3F,$C1,$51,$D3,$C0,$FF,$03,$BD,$BB,$2E,$7D,$F8,$3F,$DA,$78
	.dc.b	$78,$E3,$1F,$49,$FE,$08,$BF,$2E,$BE,$4F,$1B,$FC,$0B,$18,$C1,$B7
	.dc.b	$FA,$AF,$1F,$F0,$04,$E6,$3F,$D5,$DA,$E3,$2F,$04,$7F,$AC,$7E,$67
	.dc.b	$19,$7C,$0F,$81,$70,$7F,$5D,$E9,$FE,$99,$97,$39,$CF,$99,$E7,$D3
	.dc.b	$D7,$E2,$7B,$31,$99,$FF,$82,$B0,$F3,$57,$AF,$4F,$0F,$17,$A9,$F1
	.dc.b	$E5,$E0,$9F,$A9,$E0,$7C,$A9,$F3,$FC,$CE,$F2,$F3,$67,$E8,$E1,$F8
	.dc.b	$38,$FA,$38,$BE,$55,$7A,$F8,$1E,$8F,$F8,$0B,$87,$FC,$F7,$AF,$F8
	.dc.b	$0F,$85,$F4,$BF,$E4,$9F,$98,$F8,$3D,$1F,$84,$AF,$F4,$2F,$99,$FD
	.dc.b	$BB,$DB,$FF,$05,$F9,$BF,$C7,$E7,$FD,$12,$F6,$7F,$DF,$B1,$AC,$7E
	.dc.b	$EC,$7F,$84,$8D,$AF,$A0,$DC,$78,$F2,$FF,$2D,$71,$6E,$3D,$17,$CE
	.dc.b	$7C,$D1,$9D,$3D,$78,$FF,$82,$E7,$D5,$D5,$9B,$E1,$C6,$DB,$E7,$63
	.dc.b	$D0,$F0,$46,$78,$3F,$95,$46,$71,$E9,$39,$DF,$D5,$9F,$83,$6F,$93
	.dc.b	$39,$F1,$B8,$61,$FE,$F9,$CF,$E1,$3E,$EC,$7C,$CF,$B4,$ED,$7F,$4B
	.dc.b	$AB,$11,$F1,$B6,$FF,$36,$BE,$B3,$FC,$75,$C9,$D5,$F4,$F4,$F3,$D3
	.dc.b	$C9,$5F,$E8,$3F,$9E,$C6,$7F,$0B,$83,$FC,$7F,$87,$82,$BF,$01,$D9
	.dc.b	$D7,$5F,$7D,$7E,$8B,$AF,$DF,$87,$1E,$2F,$C4,$E3,$E5,$E1,$F8,$95
	.dc.b	$F2,$FC,$8F,$E7,$3E,$D3,$D8,$7C,$A7,$FB,$7B,$A9,$E8,$7D,$1F,$4F
	.dc.b	$F1,$99,$B1,$FF,$6B,$F1,$FE,$59,$FE,$87,$1F,$86,$7F,$A2,$F5,$DF
	.dc.b	$1B,$FD,$0B,$3E,$E3,$93,$FE,$67,$71,$3E,$4C,$FF,$B2,$47,$9E,$5A
	.dc.b	$79,$FE,$5B,$71,$8A,$F4,$F4,$7D,$1C,$39,$6D,$FE,$F8,$C7,$E2,$9E
	.dc.b	$0C,$7E,$04,$7F,$BF,$3A,$F3,$5C,$3D,$77,$C6,$E3,$3E,$D7,$F1,$7A
	.dc.b	$FE,$56,$5F,$77,$0F,$EC,$0F,$E1,$37,$F8,$97,$FE,$A1,$8F,$23,$E7
	.dc.b	$7D,$39,$FD,$1D,$2E,$BF,$4A,$77,$3E,$6D,$3F,$C4,$91,$F0,$E3,$1C
	.dc.b	$5F,$E8,$4C,$7E,$97,$67,$8F,$CB,$F2,$67,$FE,$F4,$C3,$AE,$DF,$32
	.dc.b	$FC,$B8,$FC,$2F,$7F,$F3,$1E,$43,$C2,$FF,$4B,$7A,$FF,$E4,$FF,$6B
	.dc.b	$F6,$9F,$09,$FB,$3A,$FC,$19,$BC,$FF,$82,$BF,$EB,$C6,$DF,$F9,$55
	.dc.b	$EA,$FF,$A2,$9F,$8D,$DD,$FE,$A0,$9F,$07,$BB,$B9,$ED,$5F,$EA,$7F
	.dc.b	$97,$DC,$7F,$5B,$8F,$2F,$0F,$E2,$EF,$8E,$7F,$E6,$5A,$EF,$87,$FA
	.dc.b	$63,$DD,$F3,$47,$48,$C7,$AD,$AA,$F7,$76,$FE,$1B,$0E,$78,$36,$FE
	.dc.b	$29,$8F,$1B,$F8,$9E,$1B,$D2,$E9,$3E,$6C,$3C,$6F,$97,$5F,$97,$C7
	.dc.b	$75,$F2,$5F,$87,$3E,$D7,$E1,$BF,$A3,$1E,$1C,$7F,$C3,$F2,$C3,$FA
	.dc.b	$2E,$3F,$C1,$A3,$B4,$7D,$0C,$7F,$53,$1E,$7F,$91,$F0,$67,$F6,$FB
	.dc.b	$7B,$57,$C5,$D5,$E4,$FF,$28,$F2,$FF,$DA,$7E,$03,$C3,$79,$FF,$EC
	.dc.b	$2F,$82,$78,$BF,$E8,$9E,$3F,$EB,$18,$70,$DF,$F9,$29,$83,$F5,$7F
	.dc.b	$01,$87,$D5,$1D,$B1,$FC,$04,$6F,$2F,$07,$A5,$CB,$89,$7B,$D3,$71
	.dc.b	$8B,$FD,$BE,$1E,$6E,$1F,$D9,$5E,$DF,$C0,$CF,$59,$D7,$97,$FC,$CE
	.dc.b	$B5,$B7,$E9,$5B,$F4,$1F,$D6,$5C,$EF,$FE,$AF,$CB,$C1,$AF,$C5,$D3
	.dc.b	$E2,$4F,$FA,$2F,$3F,$69,$E4,$3D,$57,$EF,$E3,$B1,$E8,$AB,$7C,$F5
	.dc.b	$FD,$54,$FD,$33,$4B,$3F,$D6,$65,$E2,$D7,$E0,$47,$D4,$AE,$17,$CE
	.dc.b	$3E,$BE,$3C,$FE,$37,$FD,$E7,$9F,$DE,$F2,$BE,$A6,$5C,$F4,$FC,$4E
	.dc.b	$5E,$AC,$BD,$DD,$DF,$32,$3C,$7D,$9F,$25,$F2,$E9,$F8,$4F,$82,$F4
	.dc.b	$6C,$F9,$77,$C3,$BE,$43,$FC,$F3,$DE,$FA,$4F,$CE,$EF,$F6,$FA,$7C
	.dc.b	$BA,$7E,$47,$5F,$2F,$12,$3F,$29,$F3,$B8,$7F,$2D,$A7,$A1,$FA,$3A
	.dc.b	$F8,$75,$F9,$ED,$EB,$F5,$39,$DA,$DF,$99,$FD,$7D,$FA,$A3,$1F,$3D
	.dc.b	$EF,$AE,$1E,$9F,$F9,$93,$6F,$C9,$46,$3C,$1F,$C4,$61,$E4,$CF,$E0
	.dc.b	$5F,$FB,$E6,$FA,$46,$38,$7E,$15,$7E,$9F,$0F,$82,$7C,$97,$D3,$97
	.dc.b	$CA,$F1,$3C,$F7,$D3,$0E,$F7,$E4,$C2,$F5,$7C,$58,$FA,$5E,$37,$2F
	.dc.b	$84,$E7,$83,$E5,$B9,$F0,$7F,$1B,$D7,$9E,$BF,$0F,$66,$9F,$E7,$A9
	.dc.b	$F3,$4E,$63,$EE,$C7,$EF,$FC,$67,$53,$E5,$1F,$0E,$E0,$7C,$BF,$E6
	.dc.b	$DF,$E0,$DE,$BF,$E0,$30,$FF,$06,$F4,$FF,$45,$F4,$7E,$F4,$74,$E3
	.dc.b	$FD,$4B,$DB,$2B,$9F,$7D,$FF,$2D,$39,$F8,$FB,$3E,$3D,$7C,$5B,$7D
	.dc.b	$6A,$E5,$1E,$5E,$AC,$56,$F6,$F3,$DF,$EC,$EF,$71,$EA,$AC,$56,$F7
	.dc.b	$FF,$2C,$F3,$FF,$15,$BF,$67,$B1,$EA,$A9,$AF,$F9,$6B,$2E,$EF,$C6
	.dc.b	$7F,$59,$5D,$B0,$FE,$07,$87,$3B,$7C,$BF,$C0,$C3,$E6,$E3,$C2,$7D
	.dc.b	$DF,$27,$67,$E8,$E9,$FD,$1F,$0E,$51,$8F,$98,$F0,$ED,$F5,$73,$FF
	.dc.b	$28,$D7,$FA,$37,$C8,$7C,$A7,$D9,$D7,$FC,$85,$F4,$09,$FF,$37,$FC
	.dc.b	$3F,$ED,$7E,$4F,$F9,$F7,$17,$F5,$78,$FD,$C4,$FF,$92,$1B,$87,$D2
	.dc.b	$9F,$F3,$24,$63,$7F,$E8,$0E,$B7,$FF,$4A,$CF,$33,$C1,$5E,$37,$89
	.dc.b	$88,$F8,$BD,$38,$C3,$D3,$18,$E6,$FF,$4A,$61,$D7,$4F,$5E,$1E,$A3
	.dc.b	$93,$EE,$F3,$31,$87,$B3,$6F,$89,$C9,$FC,$A3,$0F,$5E,$1E,$83,$CB
	.dc.b	$16,$BC,$3F,$09,$FE,$45,$AF,$06,$1E,$5D,$7E,$67,$85,$75,$F4,$F8
	.dc.b	$7E,$6C,$FC,$7B,$BE,$73,$1E,$4B,$E5,$7F,$C0,$9F,$3D,$FC,$BF,$E7
	.dc.b	$53,$EB,$3D,$1B,$7F,$29,$7F,$B1,$E3,$CA,$FF,$94,$71,$7F,$B2,$E7
	.dc.b	$A3,$FE,$0D,$9F,$F6,$A5,$71,$AF,$F0,$97,$BF,$EB,$9E,$F1,$E2,$F3
	.dc.b	$7E,$9E,$1C,$B2,$FE,$35,$5E,$3A,$F3,$F8,$DF,$A1,$8F,$B9,$5B,$86
	.dc.b	$7E,$ED,$BE,$75,$76,$3E,$75,$79,$BC,$1F,$13,$DE,$3F,$07,$F0,$FF
	.dc.b	$01,$F2,$F8,$9C,$32,$FF,$4B,$C7,$99,$F9,$3C,$57,$81,$AB,$CE,$C3
	.dc.b	$FD,$6F,$AF,$E4,$4E,$B7,$FF,$79,$63,$F1,$8F,$87,$19,$DF,$F7,$CF
	.dc.b	$FB,$DF,$91,$E2,$B5,$FB,$9F,$20,$F9,$DF,$06,$3C,$EF,$D4,$7F,$54
	.dc.b	$7F,$3E,$F2,$AF,$C8,$37,$3F,$A8,$E7,$9E,$E3,$DD,$FE,$7D,$7F,$9F
	.dc.b	$9D,$7C,$97,$4A,$D6,$5F,$B5,$AF,$D8,$57,$38,$FF,$62,$C7,$19,$DE
	.dc.b	$5F,$C2,$3A,$3F,$BB,$5F,$DD,$AF,$CD,$3F,$01,$F3,$D7,$8F,$2F,$06
	.dc.b	$7F,$0A,$77,$3E,$2E,$7B,$3F,$1B,$1F,$D2,$74,$FC,$4A,$7C,$87,$4A
	.dc.b	$F6,$E1,$E5,$5F,$27,$F0,$9E,$5C,$BF,$BF,$F2,$9E,$23,$E3,$FF,$23
	.dc.b	$F1,$FF,$E5,$EF,$37,$59,$7E,$4A,$FD,$8F,$1B,$E0,$FE,$8E,$39,$DE
	.dc.b	$B9,$9C,$7F,$E9,$FB,$EC,$ED,$F0,$73,$FC,$FC,$3F,$8F,$61,$CA,$1F
	.dc.b	$21,$FC,$7A,$31,$9F,$FA,$EF,$0E,$99,$7A,$3D,$EF,$E1,$15,$AF,$7B
	.dc.b	$FD,$11,$97,$E5,$63,$DE,$FF,$17,$7E,$0E,$F7,$4F,$1C,$F7,$7C,$B9
	.dc.b	$5C,$7E,$2F,$91,$DA,$73,$8D,$9F,$92,$FF,$99,$F3,$F8,$1F,$37,$D6
	.dc.b	$7F,$35,$8F,$8E,$FE,$0E,$1F,$E3,$D8,$9B,$FF,$02,$5E,$0B,$E3,$BF
	.dc.b	$C1,$BF,$25,$FE,$84,$F5,$7D,$B5,$FE,$A5,$E3,$FF,$23,$5F,$F3,$76
	.dc.b	$F5,$FF,$97,$9C,$E1,$FC,$81,$BC,$BF,$E8,$F7,$3E,$A7,$D3,$C3,$A1
	.dc.b	$E2,$AF,$0B,$E7,$3F,$CA,$99,$F9,$F1,$F3,$53,$3F,$D1,$4E,$6F,$D3
	.dc.b	$F1,$3C,$70,$EB,$87,$0B,$FF,$4F,$F2,$75,$C3,$86,$3C,$FD,$67,$B7
	.dc.b	$9D,$3D,$FE,$57,$B7,$3F,$31,$EB,$3D,$1C,$37,$4F,$2C,$78,$4F,$0E
	.dc.b	$DF,$E3,$8E,$0C,$69,$E0,$DB,$FD,$5D,$5F,$53,$2F,$8B,$1F,$EB,$49
	.dc.b	$F1,$4F,$C5,$D7,$CD,$C3,$8F,$8E,$FA,$53,$9F,$93,$EB,$8F,$2F,$0F
	.dc.b	$E1,$63,$F7,$FC,$47,$C2,$3E,$63,$FC,$23,$93,$D3,$FF,$7D,$DF,$E1
	.dc.b	$7C,$6E,$B8,$7F,$B2,$B4,$FA,$79,$FF,$12,$DB,$E2,$4F,$4E,$7F,$88
	.dc.b	$F6,$DE,$EB,$DF,$2F,$2E,$2A,$F1,$EB,$FE,$22,$8F,$3E,$5F,$C7,$AF
	.dc.b	$CD,$E9,$7D,$E8,$C5,$7F,$8A,$3E,$13,$77,$CE,$F7,$BF,$FA,$26,$F7
	.dc.b	$82,$3D,$D5,$F3,$CF,$11,$EA,$A9,$97,$A3,$19,$CD,$E8,$CF,$CF,$8F
	.dc.b	$3C,$3E,$AC,$7F,$A9,$5F,$CD,$BA,$C7,$F8,$CC,$7C,$81,$E1,$7F,$AD
	.dc.b	$FE,$53,$F1,$FF,$18,$F5,$7F,$D4,$BA,$7E,$D6,$BE,$E9,$FE,$0C,$BF
	.dc.b	$43,$FE,$BF,$7F,$24,$FF,$B0,$7A,$EF,$BF,$FE,$6C,$9F,$8B,$3D,$F8
	.dc.b	$FE,$8B,$E1,$ED,$FF,$09,$61,$DE,$D9,$F9,$30,$FC,$8D,$EF,$3F,$F2
	.dc.b	$F7,$17,$FB,$1E,$31,$A6,$BD,$6F,$D3,$47,$B6,$FB,$E3,$D3,$2B,$97
	.dc.b	$EB,$A7,$CB,$3E,$0C,$D9,$B5,$F0,$FC,$07,$4A,$FF,$14,$5F,$CB,$CB
	.dc.b	$D5,$E5,$F2,$CB,$F7,$3F,$15,$73,$E7,$F3,$3F,$BF,$F6,$BF,$CD,$58
	.dc.b	$FF,$BD,$1A,$FE,$DF,$16,$C9,$CC,$FF,$0B,$72,$61,$FE,$59,$3F,$C5
	.dc.b	$A3,$11,$FE,$EB,$70,$FF,$44,$3F,$8E,$47,$F3,$28,$D9,$63,$47,$CD
	.dc.b	$31,$E3,$FF,$64,$79,$BF,$8A,$7F,$87,$1E,$43,$FA,$8C,$62,$BD,$51
	.dc.b	$FD,$52,$B1,$3F,$00,$FE,$26,$7C,$26,$C7,$8A,$3F,$84,$9F,$0A,$2C
	.dc.b	$79,$8F,$F2,$74,$F6,$F6,$DF,$2B,$08,$58,$D3,$83,$5C,$BE,$99,$C4
	.dc.b	$71,$AC,$38,$9F,$97,$3A,$8F,$90,$E6,$73,$1D,$5C,$9F,$7C,$FF,$56
	.dc.b	$4D,$9F,$3C,$7C,$43,$FD,$61,$3F,$EA,$47,$C2,$E6,$3C,$F1,$F0,$9F
	.dc.b	$6B,$63,$2F,$91,$CB,$5C,$BF,$5F,$9B,$F9,$97,$A7,$FC,$07,$E2,$1F
	.dc.b	$1F,$F9,$7A,$77,$FF,$A9,$75,$47,$F1,$BC,$FD,$98,$FF,$3B,$8F,$53
	.dc.b	$FC,$F7,$C4,$BD,$FF,$E4,$59,$FB,$33,$E0,$F1,$FE,$FB,$E2,$E2,$B5
	.dc.b	$FE,$6A,$EC,$78,$1C,$23,$1B,$5D,$FD,$1D,$7E,$7C,$7C,$D9,$FF,$2F
	.dc.b	$E3,$F2,$E1,$F9,$97,$F2,$73,$F0,$2B,$BD,$FF,$21,$BE,$99,$7F,$8D
	.dc.b	$EF,$78,$78,$3B,$7B,$78,$5D,$30,$FE,$51,$F1,$7F,$94,$F1,$7F,$8C
	.dc.b	$F8,$EF,$8F,$7C,$5A,$BE,$65,$79,$8F,$E5,$9A,$FE,$D9,$FE,$50,$F3
	.dc.b	$EF,$FF,$0B,$3F,$AD,$8E,$BC,$5F,$A6,$7B,$F8,$F6,$99,$78,$77,$E4
	.dc.b	$F8,$FE,$26,$E9,$E4,$C7,$C7,$7F,$95,$C7,$F9,$8F,$37,$E4,$DF,$5E
	.dc.b	$7D,$71,$F6,$3C,$B5,$9E,$17,$E1,$FE,$0B,$D9,$87,$3C,$7C,$7A,$F1
	.dc.b	$D3,$FD,$49,$8F,$E3,$7C,$07,$CA,$3C,$A7,$CB,$7F,$A0,$38,$7F,$25
	.dc.b	$E5,$FD,$CB,$FF,$01,$61,$FC,$3D,$7C,$9F,$EB,$5E,$79,$FE,$AD,$EB
	.dc.b	$8F,$82,$BC,$19,$77,$E9,$FE,$1F,$5A,$CF,$55,$FC,$A2,$31,$F1,$9F
	.dc.b	$C4,$1D,$7B,$DF,$E5,$49,$F1,$1E,$1C,$27,$37,$85,$FF,$06,$47,$17
	.dc.b	$DD,$9F,$81,$F8,$18,$DC,$FE,$0E,$1D,$EB,$A4,$63,$C2,$F9,$18,$7F
	.dc.b	$30,$CB,$C2,$FF,$8F,$FC,$3A,$FF,$3E,$29,$FE,$71,$7D,$8F,$F9,$56
	.dc.b	$3F,$A0,$1C,$5C,$9F,$D3,$0C,$C7,$CD,$8F,$9D,$1F,$AB,$8F,$E8,$D1
	.dc.b	$F3,$E3,$FA,$81,$F1,$CF,$E9,$47,$A0,$C9,$E0,$3B,$14,$F3,$1F,$E8
	.dc.b	$6F,$23,$F8,$79,$FC,$BF,$C7,$FE,$7B,$8F,$FA,$7A,$31,$87,$F6,$87
	.dc.b	$73,$FC,$D6,$B0,$7F,$B0,$9F,$E6,$D5,$C5,$FE,$7F,$1F,$E7,$E9,$E3
	.dc.b	$1F,$D2,$2B,$C5,$3C,$63,$F9,$5B,$D2,$7F,$A5,$4F,$F9,$ED,$FE,$E3
	.dc.b	$1C,$1F,$E7,$4E,$E7,$FB,$3B,$67,$FB,$34,$F8,$23,$FC,$EA,$F8,$A2
	.dc.b	$C7,$86,$6C,$F9,$63,$C0,$66,$3F,$CD,$F1,$E2,$8F,$F1,$FD,$7C,$77
	.dc.b	$FB,$BE,$1E,$97,$FC,$20,$53,$FC,$1C,$53,$EA,$C7,$17,$FD,$49,$1E
	.dc.b	$D7,$F1,$C7,$C9,$3F,$BA,$3E,$49,$FF,$04,$47,$C7,$C3,$E0,$D7,$8E
	.dc.b	$BE,$0C,$FF,$81,$4F,$69,$F8,$63,$B1,$FE,$0F,$3A,$1F,$86,$32,$64
	.dc.b	$F1,$79,$1F,$87,$C3,$F0,$2F,$E1,$AB,$FD,$CA,$FE,$2E,$73,$3F,$84
	.dc.b	$C3,$FA,$EC,$FE,$26,$73,$3F,$3D,$AF,$F8,$15,$EE,$E5,$EE,$7B,$A7
	.dc.b	$9E,$36,$3F,$BE,$9F,$02,$BF,$C2,$F1,$6B,$FC,$07,$3F,$68,$FF,$01
	.dc.b	$BF,$C5,$1E,$EE,$9F,$CA,$CF,$85,$F2,$C7,$18,$FF,$6B,$2F,$9E,$FE
	.dc.b	$8B,$E3,$FE,$22,$7F,$A6,$D7,$C6,$EC,$F3,$47,$08,$FC,$8B,$CB,$C2
	.dc.b	$BE,$97,$F8,$BB,$E4,$EE,$5D,$7B,$BC,$13,$FE,$32,$CF,$FA,$9E,$3D
	.dc.b	$FB,$7A,$CC,$D7,$E1,$3B,$9E,$38,$79,$EA,$F8,$DF,$A2,$EF,$7C,$EB
	.dc.b	$D5,$79,$58,$D7,$F4,$75,$FA,$EC,$BF,$80,$AD,$FE,$93,$CA,$3F,$22
	.dc.b	$F4,$CA,$E3,$D7,$AF,$C0,$F7,$C6,$E5,$E0,$C7,$EB,$60,$E3,$EB,$7F
	.dc.b	$DF,$39,$5C,$BF,$E2,$8A,$F4,$E1,$D8,$6B,$D3,$2E,$5A,$FA,$B1,$F0
	.dc.b	$1F,$83,$F5,$BD,$99,$74,$96,$3D,$3C,$7F,$E0,$FE,$0B,$D8,$FF,$77
	.dc.b	$F8,$CF,$A3,$FE,$B1,$8F,$1A,$E7,$BF,$F8,$F1,$FD,$BC,$67,$E6,$FF
	.dc.b	$31,$1C,$F8,$FF,$6E,$F5,$F6,$2E,$5F,$EF,$CB,$F0,$BF,$EF,$EF,$1E
	.dc.b	$F9,$7F,$E2,$2E,$3F,$CE,$61,$E5,$8F,$F1,$2F,$D2,$62,$FB,$88,$EA
	.dc.b	$7F,$8A,$B8,$74,$FF,$8B,$36,$FD,$37,$99,$EA,$9F,$F7,$E7,$3F,$2D
	.dc.b	$38,$F6,$F8,$B8,$BF,$C9,$75,$FC,$A3,$C3,$FF,$2E,$FA,$CF,$95,$7C
	.dc.b	$D7,$FB,$06,$BF,$A0,$D7,$53,$E2,$F3,$7F,$B4,$7D,$5F,$DE,$47,$F9
	.dc.b	$9B,$C0,$FE,$36,$F4,$AF,$E7,$2F,$68,$B2,$C5,$F2,$F8,$7C,$5A,$FF
	.dc.b	$01,$BF,$2E,$7E,$4C,$7D,$18,$7E,$6D,$E1,$1D,$BC,$6F,$1C,$BD,$AF
	.dc.b	$1B,$DF,$C3,$DF,$07,$58,$F1,$5F,$CD,$C3,$DE,$7F,$83,$72,$FE,$68
	.dc.b	$FF,$83,$5A,$FF,$83,$5A,$FF,$78,$6B,$FC,$DD,$BE,$4F,$48,$FE,$FE
	.dc.b	$7F,$74,$CF,$FA,$7B,$FD,$25,$FE,$74,$FD,$53,$FC,$F1,$BF,$F8,$39
	.dc.b	$F9,$67,$F4,$53,$81,$FE,$0E,$3B,$1F,$ED,$02,$C7,$FA,$61,$94,$D3
	.dc.b	$FB,$FF,$8B,$FE,$C0,$8C,$4F,$F6,$B8,$FF,$62,$C6,$2F,$FA,$43,$63
	.dc.b	$FC,$17,$5C,$9F,$E8,$91,$97,$FA,$43,$5F,$E9,$15,$FE,$FE,$9E,$19
	.dc.b	$F9,$63,$07,$F7,$B3,$15,$E1,$7F,$B4,$CF,$F8,$92,$6C,$7F,$87,$4F
	.dc.b	$F7,$84,$7F,$88,$A2,$C7,$F8,$64,$EC,$7F,$AE,$A7,$CD,$18,$8F,$0D
	.dc.b	$FF,$87,$9D,$57,$C6,$9E,$35,$F1,$BC,$1F,$F6,$3F,$9F,$FC,$7B,$E0
	.dc.b	$B3,$7F,$EC,$69,$FE,$2E,$DE,$3F,$3C,$7E,$87,$1F,$2B,$FE,$D0,$C6
	.dc.b	$FB,$3F,$EC,$B9,$E6,$27,$BE,$DE,$7C,$FC,$54,$CF,$F9,$25,$FC,$8C
	.dc.b	$BC,$53,$9D,$3F,$09,$3E,$C3,$A7,$91,$E4,$99,$87,$C8,$F6,$3D,$35
	.dc.b	$E1,$97,$AF,$E6,$CB,$F9,$EF,$57,$9B,$C0,$FF,$5F,$7A,$CF,$6D,$F3
	.dc.b	$7E,$AE,$FE,$75,$FE,$9A,$D7,$F1,$EF,$FB,$17,$4F,$AE,$FF,$B1,$F9
	.dc.b	$2F,$93,$FE,$C4,$9F,$CC,$F1,$7F,$37,$8F,$F5,$BC,$78,$56,$6F,$89
	.dc.b	$F8,$D5,$B8,$F2,$78,$DE,$7B,$FC,$4E,$3F,$E0,$2C,$BC,$78,$F2,$9E
	.dc.b	$22,$7D,$BA,$E1,$BD,$BF,$07,$E6,$79,$DA,$3D,$39,$7F,$5E,$C3,$ED
	.dc.b	$F6,$7F,$84,$FE,$65,$F0,$EF,$9A,$CD,$7E,$72,$BF,$8F,$3E,$27,$FC
	.dc.b	$34,$FE,$71,$FF,$0E,$72,$5F,$51,$F2,$D7,$1F,$F5,$B5,$FD,$FB,$F0
	.dc.b	$63,$CE,$D5,$E2,$EB,$F2,$C6,$2F,$53,$FD,$21,$FC,$60,$BE,$78,$FB
	.dc.b	$39,$7D,$1E,$B7,$A7,$0F,$4D,$F0,$BF,$81,$F1,$7F,$99,$F1,$52,$EE
	.dc.b	$FF,$71,$F4,$7F,$B8,$B2,$FD,$7D,$68,$FF,$00,$9C,$27,$F1,$4D,$F7
	.dc.b	$3F,$CD,$AD,$F0,$7F,$79,$83,$A5,$3B,$F0,$73,$18,$9F,$15,$FF,$25
	.dc.b	$F0,$FC,$D1,$8C,$FF,$DE,$AF,$90,$F4,$1E,$03,$47,$0C,$BD,$77,$4D
	.dc.b	$3F,$EF,$88,$F0,$E3,$E9,$AF,$E0,$11,$FC,$DA,$3E,$2F,$73,$83,$F5
	.dc.b	$9C,$FF,$E2,$6A,$D0,$9F,$8A,$DD,$BC,$90,$F0,$AD,$FE,$56,$BD,$B3
	.dc.b	$D7,$B7,$C7,$1C,$E7,$F7,$57,$A7,$C3,$1F,$06,$BE,$EB,$FC,$DB,$1F
	.dc.b	$3C,$FC,$3F,$53,$C7,$5E,$9E,$FE,$AF,$AF,$C0,$B5,$FC,$1F,$B7,$B6
	.dc.b	$7F,$BF,$CB,$EC,$65,$F8,$6E,$6F,$F2,$FF,$5F,$E6,$33,$FA,$5E,$07
	.dc.b	$C0,$9F,$C1,$DF,$E5,$7B,$FE,$A4,$7F,$99,$B9,$3D,$AF,$29,$FB,$AF
	.dc.b	$4E,$BB,$EB,$7E,$D3,$BF,$35,$CB,$2E,$73,$3D,$4E,$6F,$EE,$BB,$F8
	.dc.b	$F7,$63,$6F,$8B,$87,$E4,$7A,$BF,$D5,$BE,$53,$DC,$FE,$DF,$3F,$D7
	.dc.b	$A7,$FD,$62,$FF,$B1,$9F,$F5,$9F,$4D,$8F,$EF,$1F,$29,$7B,$7F,$C3
	.dc.b	$DD,$DF,$9A,$BE,$B8,$5B,$FE,$FD,$8F,$7F,$07,$FA,$45,$7F,$AF,$BC
	.dc.b	$3E,$53,$39,$78,$78,$5F,$E6,$EA,$F3,$F8,$FC,$FC,$07,$C0,$FF,$2E
	.dc.b	$79,$E7,$FE,$72,$3C,$CF,$E5,$5F,$EC,$F5,$FE,$81,$9C,$9F,$E8,$57
	.dc.b	$1E,$1F,$5D,$3D,$DE,$57,$86,$FF,$CF,$B7,$F9,$77,$99,$8F,$97,$F2
	.dc.b	$47,$F4,$0A,$F1,$F7,$7F,$A4,$3C,$67,$CA,$E3,$0F,$95,$3F,$E7,$EC
	.dc.b	$FF,$D0,$1B,$7F,$1D,$9F,$A4,$7F,$3F,$7F,$6C,$F8,$27,$A7,$87,$DA
	.dc.b	$79,$8C,$3E,$87,$8D,$E9,$AF,$F5,$0F,$0F,$D3,$7C,$9D,$DF,$23,$C1
	.dc.b	$FF,$4C,$FC,$D3,$E3,$31,$A7,$DF,$9F,$F0,$03,$F6,$DF,$F4,$FF,$4D
	.dc.b	$E5,$FE,$27,$3F,$12,$79,$E5,$D6,$79,$1F,$D0,$A7,$FC,$41,$BF,$EA
	.dc.b	$E1,$8F,$33,$E1,$4E,$8F,$E7,$DC,$3F,$CA,$9F,$E5,$5C,$9F,$CE,$38
	.dc.b	$1E,$87,$D5,$9F,$F1,$BC,$E5,$E1,$97,$AA,$BF,$DF,$51,$FE,$52,$9E
	.dc.b	$55,$A6,$F1,$7F,$8C,$F4,$EF,$5F,$EC,$AA,$EC,$FF,$48,$C7,$38,$78
	.dc.b	$AF,$FC,$7B,$7F,$E3,$DC,$39,$DF,$F9,$36,$F0,$F3,$AF,$F2,$85,$78
	.dc.b	$E3,$07,$F9,$9E,$71,$7D,$63,$FC,$DD,$8F,$AF,$1F,$CA,$57,$C0,$9E
	.dc.b	$37,$FE,$46,$C7,$C3,$36,$7C,$13,$FE,$39,$9F,$7C,$D9,$F7,$6F,$FD
	.dc.b	$52,$7E,$2F,$CA,$D9,$F8,$5E,$FF,$D0,$C6,$1F,$4B,$2F,$24,$62,$3E
	.dc.b	$AF,$07,$C6,$51,$97,$E1,$E9,$F6,$A3,$84,$E2,$76,$76,$BC,$4F,$F9
	.dc.b	$F3,$1F,$B2,$E6,$33,$3E,$07,$E4,$BC,$8F,$85,$7D,$D5,$E1,$EF,$36
	.dc.b	$BF,$0F,$87,$CF,$EF,$F8,$58,$79,$EB,$E4,$62,$F9,$2F,$F1,$55,$E2
	.dc.b	$7F,$18,$7F,$9A,$7C,$7B,$5B,$8F,$98,$F1,$EE,$B6,$F8,$0B,$78,$F4
	.dc.b	$3E,$97,$37,$E8,$F1,$FC,$AE,$3F,$CF,$9F,$C6,$AF,$93,$F0,$F3,$BE
	.dc.b	$31,$E3,$BE,$1D,$F4,$BF,$E2,$1F,$55,$F1,$0F,$A4,$FD,$0E,$3C,$70
	.dc.b	$E1,$7B,$A7,$16,$B2,$EB,$97,$4D,$99,$F4,$DB,$59,$72,$9C,$F4,$7E
	.dc.b	$93,$4E,$D8,$71,$F3,$FF,$49,$18,$D7,$FC,$4D,$8E,$F3,$EF,$1F,$E8
	.dc.b	$CE,$5F,$E2,$F5,$BA,$FD,$23,$E9,$F9,$9F,$CF,$3E,$BC,$AF,$7F,$C6
	.dc.b	$F5,$3E,$3E,$5F,$C5,$F0,$F6,$7C,$57,$E8,$70,$F1,$6D,$EC,$E2,$7D
	.dc.b	$2F,$B3,$83,$E0,$71,$7A,$32,$F2,$DF,$F9,$C7,$2D,$70,$7F,$8E,$1F
	.dc.b	$57,$17,$D3,$DF,$F6,$DA,$7C,$29,$A7,$90,$F1,$61,$FC,$99,$E2,$7B
	.dc.b	$1E,$F8,$FD,$AE,$CE,$F9,$7D,$2C,$FB,$9F,$CA,$2F,$E2,$F8,$1F,$96
	.dc.b	$F1,$5F,$2C,$F8,$66,$3F,$C1,$7C,$8F,$C8,$73,$FF,$44,$BF,$F4,$C7
	.dc.b	$67,$F9,$A9,$E3,$97,$F0,$77,$9F,$8B,$72,$E9,$3F,$C3,$E7,$9D,$71
	.dc.b	$EA,$F1,$7C,$BF,$8E,$FF,$BD,$5F,$F2,$5B,$F2,$FC,$5F,$F1,$3F,$21
	.dc.b	$F4,$9F,$E6,$CF,$0F,$F8,$D4,$74,$BF,$F4,$37,$81,$FB,$28,$DC,$71
	.dc.b	$AF,$F6,$9B,$FC,$C6,$39,$FB,$7D,$78,$78,$76,$71,$BF,$F1,$3D,$FF
	.dc.b	$A2,$0F,$F7,$D7,$4F,$F4,$AF,$27,$FA,$2C,$7F,$9F,$E3,$0F,$FB,$9D
	.dc.b	$FE,$A8,$FF,$A0,$1F,$EB,$5D,$7F,$D5,$9F,$EA,$71,$93,$FA,$84,$5F
	.dc.b	$23,$F9,$87,$85,$FD,$3B,$9B,$FC,$41,$18,$AF,$E9,$D1,$8C,$3F,$A4
	.dc.b	$47,$F9,$64,$F2,$1F,$D6,$DF,$ED,$1C,$DE,$A8,$C5,$FB,$76,$F7,$5F
	.dc.b	$3B,$FE,$9F,$5B,$C3,$C1,$5F,$0E,$3A,$C7,$8F,$0C,$5F,$1C,$3F,$D7
	.dc.b	$15,$8C,$BE,$44,$7D,$29,$ED,$E5,$7C,$7A,$F3,$75,$5C,$7F,$A8,$79
	.dc.b	$5F,$D2,$6F,$C8,$FF,$48,$6E,$9F,$D1,$DA,$FF,$43,$6B,$EC,$CB,$EC
	.dc.b	$4F,$F5,$BD,$71,$8F,$C5,$8F,$EB,$CE,$30,$F0,$DF,$F4,$88,$C3,$FD
	.dc.b	$2D,$F6,$1C,$6B,$E2,$3F,$12,$3F,$A4,$3A,$9F,$ED,$5C,$1E,$FE,$D7
	.dc.b	$BB,$FE,$A7,$AF,$85,$C7,$FB,$0C,$FF,$DA,$1A,$FE,$1E,$7F,$DB,$15
	.dc.b	$6F,$8E,$1F,$89,$78,$4F,$FB,$77,$7D,$57,$F4,$7D,$BF,$CD,$14,$F8
	.dc.b	$0E,$9D,$AF,$A5,$7B,$DF,$93,$8E,$67,$F4,$FC,$BF,$16,$BC,$CE,$35
	.dc.b	$FC,$56,$73,$D5,$87,$94,$7F,$23,$7E,$3E,$FE,$6E,$77,$73,$E3,$3F
	.dc.b	$45,$DD,$9E,$3F,$81,$4F,$8A,$B7,$FC,$DF,$87,$FD,$05,$3C,$E3,$7E
	.dc.b	$0E,$BA,$7F,$DD,$FC,$1F,$B7,$E9,$FF,$1D,$E1,$FE,$D9,$DB,$F4,$91
	.dc.b	$F6,$1F,$F7,$C4,$FE,$F6,$B1,$5B,$E9,$FC,$2D,$7D,$E8,$C5,$FF,$93
	.dc.b	$A3,$88,$9F,$E5,$37,$FC,$02,$3C,$67,$F9,$34,$B1,$FE,$2C,$E7,$FF
	.dc.b	$22,$DF,$4B,$FD,$DD,$3B,$3A,$DE,$E3,$F7,$D5,$BA,$FD,$BB,$6F,$F8
	.dc.b	$14,$F0,$9F,$DB,$23,$0F,$F9,$6F,$4F,$DC,$E3,$F1,$EB,$FB,$73,$F2
	.dc.b	$7B,$3E,$03,$F2,$F7,$B8,$7C,$AE,$CF,$F2,$16,$5E,$37,$FC,$8F,$AF
	.dc.b	$6C,$FF,$C6,$87,$86,$74,$FE,$DF,$0F,$AD,$5F,$3A,$73,$5F,$C5,$27
	.dc.b	$35,$E6,$BF,$04,$DD,$3F,$CA,$B8,$79,$1F,$F2,$BF,$5F,$9B,$DB,$FF
	.dc.b	$65,$F3,$FD,$27,$EA,$E0,$7F,$A8,$BD,$CC,$7A,$DF,$E8,$CC,$FE,$44
	.dc.b	$FF,$4A,$9D,$73,$FD,$FD,$BE,$EB,$D0,$D4,$7B,$EB,$E8,$47,$E1,$63
	.dc.b	$F0,$7E,$27,$8B,$9D,$3F,$9D,$AF,$ED,$D5,$E1,$13,$FE,$B6,$7F,$2B
	.dc.b	$1F,$DC,$E7,$4F,$3E,$4E,$1C,$9C,$35,$FC,$8B,$F0,$0E,$77,$89,$DD
	.dc.b	$7F,$AA,$B0,$FD,$47,$0F,$FA,$BF,$8B,$7C,$1F,$E7,$FC,$3F,$9F,$D7
	.dc.b	$EB,$78,$BF,$48,$FE,$BE,$29,$FD,$46,$3F,$57,$1F,$D9,$A3,$33,$FE
	.dc.b	$79,$6D,$F8,$5F,$8C,$7E,$D4,$5F,$8A,$F5,$5F,$E9,$1E,$4F,$EB,$53
	.dc.b	$E6,$BF,$D3,$5F,$EA,$63,$11,$FD,$6A,$39,$E1,$EE,$AF,$F5,$F4,$60
	.dc.b	$FE,$0B,$28,$FE,$B7,$19,$BF,$06,$5F,$DA,$2F,$E3,$F2,$7C,$4A,$E3
	.dc.b	$C1,$6F,$FD,$4D,$C9,$FE,$AA,$CF,$C7,$9F,$FA,$54,$FF,$4A,$B8,$7F
	.dc.b	$D3,$57,$DD,$FC,$E9,$FD,$4E,$35,$3F,$E3,$D9,$D4,$FF,$68,$D3,$C9
	.dc.b	$1E,$A9,$F1,$4F,$E9,$23,$3C,$2F,$85,$FA,$0F,$F9,$03,$4F,$E0,$CF
	.dc.b	$E6,$A7,$E2,$77,$65,$FF,$4F,$1F,$57,$2F,$C5,$D7,$F4,$D8,$C3,$FC
	.dc.b	$3F,$83,$83,$FD,$0E,$3E,$33,$8E,$0F,$C0,$70,$FF,$4D,$9E,$27,$D0
	.dc.b	$3F,$12,$7F,$8F,$8C,$BF,$B8,$7F,$86,$3F,$98,$CF,$EC,$46,$0F,$B6
	.dc.b	$FF,$85,$A3,$F3,$BA,$D9,$FB,$E5,$DB,$F8,$CB,$6B,$FA,$0B,$FE,$26
	.dc.b	$AF,$E8,$CF,$F4,$37,$F3,$8E,$EB,$E2,$47,$F8,$FA,$3F,$A5,$BF,$D0
	.dc.b	$DA,$7E,$78,$B5,$FE,$E7,$BF,$E6,$D3,$FA,$03,$D8,$FF,$39,$E1,$C5
	.dc.b	$7B,$23,$FC,$5F,$9F,$F8,$C2,$F9,$57,$C1,$7F,$DE,$5C,$3E,$9B,$F8
	.dc.b	$7C,$36,$7F,$DE,$DC,$3F,$EF,$FC,$39,$69,$F9,$83,$F2,$B9,$FC,$D9
	.dc.b	$FF,$46,$4E,$A7,$F9,$8E,$9D,$DF,$CF,$E3,$5F,$C8,$3E,$8C,$3F,$78
	.dc.b	$B0,$C3,$FB,$FC,$EB,$FE,$2C,$8F,$53,$FD,$DE,$2B,$FE,$53,$8F,$53
	.dc.b	$FE,$32,$8A,$FF,$97,$4A,$7E,$F8,$E2,$7E,$34,$EE,$7F,$95,$63,$FC
	.dc.b	$F2,$7D,$47,$81,$D9,$FB,$AF,$F3,$B8,$B3,$CE,$3E,$64,$7B,$B0,$FF
	.dc.b	$31,$F2,$30,$87,$F9,$77,$0F,$E9,$AE,$0F,$ED,$33,$C1,$C1,$FE,$56
	.dc.b	$70,$E3,$1F,$CF,$56,$A3,$F8,$24,$E9,$FF,$30,$61,$A7,$F9,$D3,$5C
	.dc.b	$B5,$FE,$18,$DA,$FC,$F3,$FD,$9E,$3F,$46,$F1,$8F,$CD,$D6,$23,$B4
	.dc.b	$7C,$68,$E7,$1F,$AF,$7F,$C5,$0F,$E6,$9F,$F6,$7D,$FF,$85,$78,$3F
	.dc.b	$DA,$73,$F3,$9C,$AF,$02,$F0,$FF,$01,$8F,$D1,$BA,$E5,$DD,$FF,$0F
	.dc.b	$9F,$D0,$71,$61,$FA,$91,$FE,$26,$7F,$D5,$CD,$7F,$D5,$AD,$7F,$C5
	.dc.b	$ED,$7F,$C6,$87,$EC,$9F,$F2,$51,$FA,$D3,$F2,$CF,$F1,$17,$F3,$4E
	.dc.b	$4F,$F5,$56,$BF,$C4,$63,$F3,$51,$4F,$F6,$84,$7D,$08,$FC,$D4,$64
	.dc.b	$FF,$09,$6B,$FE,$AE,$8F,$23,$63,$FC,$22,$E6,$3D,$F1,$FA,$F9,$FC
	.dc.b	$A3,$FC,$9F,$5F,$E0,$0D,$7F,$A9,$B4,$FE,$E2,$6F,$5F,$F1,$D3,$C4
	.dc.b	$A7,$F2,$E7,$FD,$96,$FF,$33,$29,$FD,$B1,$FF,$63,$BF,$E0,$E2,$9F
	.dc.b	$E7,$47,$FC,$DC,$FF,$82,$CC,$ED,$F9,$67,$9B,$F8,$F7,$CA,$7F,$BA
	.dc.b	$8F,$F3,$C9,$FC,$C4,$E6,$7F,$9D,$63,$E0,$BF,$CC,$4C,$C7,$EA,$67
	.dc.b	$FA,$4E,$1F,$F0,$06,$31,$C5,$FC,$5E,$31,$3F,$CB,$9A,$FE,$DB,$8B
	.dc.b	$E1,$D7,$3E,$2B,$3F,$CB,$8F,$3C,$77,$E2,$F9,$F8,$FF,$56,$94,$70
	.dc.b	$7F,$80,$F2,$F4,$F3,$1F,$45,$08,$3E,$07,$AC,$62,$35,$D7,$87,$13
	.dc.b	$CD,$D4,$57,$A4,$7C,$E9,$CE,$39,$9E,$CE,$5E,$AD,$72,$DA,$F9,$B3
	.dc.b	$E8,$AF,$AB,$3F,$EA,$78,$F5,$CE,$32,$47,$29,$FC,$0B,$FE,$CF,$D7
	.dc.b	$EA,$79,$5F,$6A,$B7,$8F,$BC,$F8,$B3,$F6,$6B,$10,$D7,$E1,$3F,$17
	.dc.b	$0A,$FC,$B8,$F7,$3F,$3A,$2C,$7B,$63,$11,$F0,$E3,$CB,$1F,$4E,$2B
	.dc.b	$FE,$1F,$AF,$13,$C2,$3E,$44,$FB,$EB,$11,$F3,$E3,$FC,$58,$FF,$AD
	.dc.b	$EF,$E1,$B8,$3F,$D9,$DF,$16,$F4,$E1,$0C,$63,$F9,$98,$C4,$6B,$C9
	.dc.b	$FC,$8C,$62,$39,$56,$1D,$CF,$E6,$A7,$53,$67,$B4,$FE,$CA,$7A,$CD
	.dc.b	$9E,$CD,$7A,$B6,$7F,$5F,$3E,$88,$B2,$9F,$C1,$4F,$09,$FE,$02,$E5
	.dc.b	$C9,$F0,$EF,$F8,$0B,$CA,$BD,$FF,$0D,$EB,$7E,$0E,$F5,$F8,$F1,$EB
	.dc.b	$7E,$44,$57,$E6,$47,$A5,$F9,$D1,$63,$C7,$1F,$42,$2A,$8F,$14,$7B
	.dc.b	$1F,$5C,$7F,$5C,$9F,$5C,$7A,$6F,$11,$FC,$7B,$C3,$FC,$E7,$C7,$7E
	.dc.b	$77,$82,$CF,$F4,$ED,$3F,$8A,$4F,$F8,$9D,$B1,$FD,$39,$AF,$F4,$E3
	.dc.b	$F8,$7C,$7F,$52,$69,$92,$9F,$D8,$E3,$8B,$FD,$78,$FE,$37,$1F,$D4
	.dc.b	$9A,$7F,$89,$0E,$7D,$1F,$B9,$8C,$47,$82,$3E,$B4,$7F,$54,$8F,$CE
	.dc.b	$7B,$5F,$B9,$F8,$DF,$E6,$1C,$BF,$C9,$1F,$E6,$5D,$36,$3F,$C3,$95
	.dc.b	$5F,$F3,$AE,$3F,$CD,$6B,$33,$99,$FF,$66,$4E,$5F,$F1,$64,$7F,$48
	.dc.b	$9F,$F3,$EB,$FC,$D9,$CB,$98,$FF,$6A,$39,$7F,$C3,$2F,$8E,$BF,$8B
	.dc.b	$4E,$23,$D1,$1F,$93,$8F,$E2,$9C,$9F,$C4,$39,$BF,$9F,$BF,$C4,$B4
	.dc.b	$AF,$F3,$2B,$AF,$F3,$A8,$FE,$92,$E6,$32,$FF,$26,$72,$FF,$36,$7F
	.dc.b	$B1,$BC,$5A,$FF,$2F,$7F,$B4,$BF,$CD,$9C,$CF,$F2,$F8,$E9,$1A,$8F
	.dc.b	$D6,$47,$F2,$F8,$FE,$25,$1E,$C8,$FE,$57,$8F,$F6,$57,$F9,$6F,$27
	.dc.b	$99,$FE,$45,$19,$7F,$A9,$47,$A1,$FE,$AF,$55,$FE,$B7,$1E,$87,$F9
	.dc.b	$54,$66,$3D,$31,$88,$E5,$1F,$DE,$23,$11,$F1,$23,$D9,$1E,$A7,$D7
	.dc.b	$1F,$C1,$DF,$E9,$1F,$1F,$FC,$87,$8B,$F9,$3C,$FF,$2F,$8C,$72,$FF
	.dc.b	$76,$75,$16,$7F,$92,$B5,$FE,$60,$FF,$80,$23,$2E,$63,$F8,$D6,$1F
	.dc.b	$E0,$B9,$FE,$53,$3F,$E1,$A7,$8B,$67,$F9,$E5,$7F,$85,$27,$F9,$34
	.dc.b	$FF,$88,$9C,$B9,$9F,$E4,$51,$FC,$96,$38,$1F,$C9,$7A,$BD,$F3,$FC
	.dc.b	$FF,$93,$C8,$FF,$19,$C3,$C8,$FF,$28,$AC,$BF,$E7,$28,$CB,$FC,$A2
	.dc.b	$3D,$CF,$F2,$86,$BF,$CA,$1F,$53,$FC,$A1,$B3,$E7,$8F,$E3,$F1,$F2
	.dc.b	$E3,$FC,$01,$7B,$8F,$74,$7B,$39,$7F,$9F,$7C,$87,$F7,$FC,$55,$FC
	.dc.b	$8C,$FE,$CC,$FF,$23,$69,$FC,$8C,$A7,$F2,$3C,$7F,$C7,$91,$FC,$8D
	.dc.b	$B1,$FC,$8D,$AF,$F2,$33,$89,$FC,$8C,$FF,$1C,$C7,$F2,$36,$9F,$C8
	.dc.b	$E3,$9F,$27,$F8,$6E,$7F,$91,$C7,$82,$3E,$A7,$67,$F5,$9E,$CF,$F0
	.dc.b	$D7,$67,$F2,$2A,$FF,$28,$3F,$C8,$9B,$CB,$FC,$89,$AF,$F2,$2C,$7F
	.dc.b	$C0,$D3,$FC,$8A,$73,$3F,$C8,$A7,$2F,$F1,$E7,$FC,$A3,$3F,$C8,$9B
	.dc.b	$1F,$EB,$C9,$EB,$CB,$FE,$64,$AF,$F5,$E4,$78,$A3,$FC,$6D,$1F,$C8
	.dc.b	$63,$CF,$1F,$DC,$E3,$FC,$69,$A7,$F8,$D7,$E3,$3F,$91,$57,$F9,$AD
	.dc.b	$FE,$44,$DE,$3F,$E4,$4D,$7F,$8D,$35,$FE,$35,$8F,$F9,$A2,$7F,$CA
	.dc.b	$6D,$8F,$E4,$4F,$F9,$C2,$3F,$91,$36,$38,$B5,$72,$FF,$A0,$27,$33
	.dc.b	$B8,$FF,$29,$47,$F9,$46,$3F,$D1,$EF,$F2,$35,$D6,$F8,$EF,$75,$FF
	.dc.b	$24,$9F,$F1,$C5,$7F,$24,$C3,$D5,$3F,$C9,$27,$2F,$F2,$47,$D4,$FF
	.dc.b	$24,$6B,$FC,$91,$F4,$3F,$E4,$98,$F8,$9C,$DF,$EA,$4A,$E1,$5E,$D8
	.dc.b	$F6,$57,$F9,$32,$3E,$34,$7B,$23,$FC,$51,$3F,$EE,$59,$F9,$87,$F8
	.dc.b	$AF,$97,$F9,$21,$FE,$96,$3F,$91,$DF,$FA,$8D,$FE,$46,$D7,$F9,$1B
	.dc.b	$5F,$E3,$6D,$7F,$8B,$C7,$37,$F8,$FB,$5C,$B5,$FF,$59,$9F,$E4,$A7
	.dc.b	$FD,$7E,$FF,$22,$6B,$FC,$7D,$A7,$FA,$25,$C3,$C0,$FF,$23,$73,$7F
	.dc.b	$A1,$E3,$FA,$07,$2F,$77,$F9,$14,$FF,$9B,$DF,$E4,$CE,$5F,$E4,$4E
	.dc.b	$5F,$F3,$74,$78,$9F,$F2,$14,$7F,$B0,$1C,$B9,$7F,$C7,$B5,$F1,$67
	.dc.b	$DB,$1F,$C7,$E7,$83,$FE,$49,$73,$AF,$AE,$7E,$A3,$FE,$5F,$7F,$D4
	.dc.b	$1E,$E7,$FA,$87,$C9,$B5,$FC,$8C,$FC,$AC,$7F,$23,$69,$FC,$8C,$A7
	.dc.b	$F2,$33,$FD,$1C,$FF,$23,$69,$FC,$8C,$A7,$F2,$33,$89,$FC,$8F,$0F
	.dc.b	$F7,$34,$7F,$23,$6C,$7F,$23,$39,$F2,$FF,$AF,$67,$F9,$1C,$78,$23
	.dc.b	$FC,$F5,$1F,$E4,$58,$FD,$6F,$CD,$FF,$AF,$7D,$FF,$E4,$91,$FE,$A3
	.dc.b	$7F,$92,$37,$97,$F9,$23,$5F,$E4,$8F,$ED,$A3,$F9,$23,$98,$FE,$48
	.dc.b	$E5,$FE,$49,$87,$FB,$FE,$7F,$92,$36,$3F,$92,$4F,$5E,$5F,$EA,$2C
	.dc.b	$9D,$4F,$FB,$66,$BF,$C9,$33,$FD,$5A,$79,$39,$AF,$F4,$C4,$FF,$B6
	.dc.b	$7E,$93,$F9,$2C,$FF,$AF,$DF,$E4,$AD,$E1,$FE,$4A,$D7,$F9,$2B,$5F
	.dc.b	$E4,$B7,$E4,$E4,$57,$F2,$68,$C2,$9F,$E4,$F4,$9C,$C7,$F9,$16,$30
	.dc.b	$77,$EC,$F7,$BE,$B3,$C5,$CD,$87,$67,$F2,$4B,$FE,$56,$E3,$0F,$1D
	.dc.b	$FF,$30,$8E,$18,$79,$30,$FF,$5D,$5F,$F2,$DB,$7D,$1F,$C5,$C3,$F9
	.dc.b	$66,$1D,$5F,$F7,$F7,$37,$5B,$FF,$7E,$CD,$9F,$E5,$98,$FB,$1F,$E5
	.dc.b	$8D,$7F,$96,$46,$DF,$F2,$64,$66,$BD,$EE,$E6,$9F,$46,$3D,$2F,$D3
	.dc.b	$29,$E5,$7E,$23,$E3,$2C,$7D,$28,$F7,$C7,$FB,$67,$AB,$6F,$CD,$3F
	.dc.b	$A6,$F9,$5F,$CC,$AB,$EE,$3F,$CC,$A3,$FC,$43,$92,$3F,$99,$C6,$1F
	.dc.b	$F7,$E7,$3F,$39,$D7,$3F,$FB,$DA,$BE,$A5,$7F,$95,$AF,$AC,$FF,$34
	.dc.b	$6B,$FC,$D1,$AF,$BA,$B9,$BF,$CB,$A3,$75,$9A,$FE,$76,$F8,$5F,$F7
	.dc.b	$D3,$FA,$73,$CD,$7F,$EF,$98,$F8,$31,$E4,$AF,$F2,$B5,$78,$1F,$E6
	.dc.b	$B1,$F1,$1F,$F2,$B4,$65,$FE,$6B,$19,$7F,$CA,$CF,$37,$F9,$AB,$5F
	.dc.b	$E5,$CF,$95,$CB,$BE,$6F,$54,$7F,$35,$8F,$24,$7D,$A7,$E7,$47,$F9
	.dc.b	$69,$F9,$71,$EB,$E7,$FB,$75,$FB,$3C,$7D,$AF,$ED,$6B,$EF,$F3,$7E
	.dc.b	$95,$FE,$6D,$38,$70,$7E,$B7,$3F,$B4,$FF,$35,$3F,$C2,$26,$4C,$9F
	.dc.b	$EF,$5A,$EC,$FF,$16,$8F,$CC,$BC,$0F,$F3,$4D,$7D,$B7,$46,$AB,$FC
	.dc.b	$B8,$F8,$A7,$E0,$3E,$09,$F6,$BF,$92,$3F,$9C,$4A,$3D,$F1,$FE,$76
	.dc.b	$CB,$FD,$FB,$C9,$FC,$19,$FF,$7F,$F2,$58,$1F,$CE,$A7,$F3,$8E,$0F
	.dc.b	$E7,$73,$5F,$F7,$BE,$3F,$CC,$A3,$11,$FC,$F1,$AF,$C0,$73,$56,$3F
	.dc.b	$CC,$55,$E1,$7F,$9B,$D7,$37,$FC,$BF,$1E,$97,$F9,$DB,$B3,$E1,$D6
	.dc.b	$1F,$51,$F1,$0F,$DA,$57,$C2,$7F,$6C,$58,$78,$9F,$D0,$23,$F7,$4F
	.dc.b	$F4,$06,$F0,$7F,$40,$6B,$FD,$01,$AF,$F4,$07,$E7,$47,$F3,$F6,$C7
	.dc.b	$F3,$EC,$3E,$34,$FF,$99,$27,$E9,$3F,$D0,$9B,$3F,$D1,$79,$3F,$8D
	.dc.b	$DF,$9A,$3F,$CC,$53,$A9,$F8,$35,$88,$E7,$1F,$06,$9E,$2F,$BE,$BF
	.dc.b	$A8,$5C,$2E,$1E,$47,$F9,$DB,$5D,$1F,$C8,$EB,$C4,$F1,$E2,$C5,$6A
	.dc.b	$3F,$86,$63,$F1,$14,$E2,$B8,$C7,$F9,$A6,$7E,$2B,$FD,$09,$F8,$2F
	.dc.b	$F6,$57,$F1,$15,$E8,$AF,$86,$FA,$A2,$CF,$66,$9F,$12,$7E,$14,$FD
	.dc.b	$FF,$17,$F9,$D4,$FE,$8D,$FE,$74,$FF,$39,$7F,$9C,$B7,$8B,$FC,$D6
	.dc.b	$ED,$FE,$71,$8F,$E7,$E7,$F9,$C4,$E6,$7F,$CC,$93,$97,$FC,$D1,$1F
	.dc.b	$AF,$9F,$F3,$7B,$63,$F9,$AC,$ED,$FC,$DC,$FF,$38,$9D,$18,$E3,$FB
	.dc.b	$91,$F9,$9A,$F6,$57,$F3,$6A,$F1,$23,$F9,$BE,$4A,$FF,$35,$D7,$79
	.dc.b	$FF,$33,$B5,$FE,$66,$DE,$2F,$B1,$3F,$E6,$AB,$FC,$13,$D6,$BF,$CD
	.dc.b	$15,$E9,$E3,$FC,$BC,$FF,$3A,$AF,$74,$7B,$2B,$F9,$B5,$7C,$98,$F5
	.dc.b	$DF,$E5,$FC,$5F,$DA,$78,$BF,$B5,$E0,$B8,$FF,$9A,$F3,$4F,$F9,$AD
	.dc.b	$B1,$FE,$6B,$6B,$FE,$6B,$3F,$BA,$47,$F3,$78,$CC,$07,$2F,$E5,$67
	.dc.b	$FC,$5B,$C9,$FC,$5D,$FF,$17,$C5,$9F,$E6,$E5,$38,$99,$9F,$E2,$73
	.dc.b	$FE,$6C,$8F,$0C,$7E,$CE,$3F,$CD,$94,$F1,$7F,$89,$78,$BF,$E6,$CB
	.dc.b	$FF,$05,$3F,$E6,$C6,$F3,$7F,$33,$A4,$FF,$2C,$E6,$FD,$A3,$FE,$6C
	.dc.b	$AD,$D7,$F3,$79,$FD,$84,$FF,$34,$9B,$39,$6B,$FD,$01,$FE,$6D,$19
	.dc.b	$72,$7F,$12,$7F,$98,$EF,$FC,$FD,$E1,$19,$AF,$E2,$11,$FC,$B7,$97
	.dc.b	$F9,$66,$1F,$C1,$5F,$E3,$97,$FD,$1D,$FE,$11,$D5,$FC,$35,$FE,$15
	.dc.b	$B5,$7F,$8B,$47,$F0,$97,$F9,$66,$3F,$E2,$F7,$2E,$5F,$F3,$45,$75
	.dc.b	$7F,$CD,$D3,$FC,$A9,$FF,$36,$4F,$F6,$17,$FC,$CB,$3B,$AF,$E9,$71
	.dc.b	$FC,$AE,$3C,$D1,$FC,$AE,$BD,$93,$FE,$62,$BF,$63,$FC,$7A,$BC,$0F
	.dc.b	$F2,$8B,$EC,$FF,$98,$9F,$EE,$4F,$F2,$77,$DC,$FF,$1D,$6B,$FC,$95
	.dc.b	$F7,$39,$72,$7F,$6B,$70,$F2,$3D,$CF,$F2,$37,$4F,$F4,$33,$0F,$13
	.dc.b	$FA,$37,$37,$F6,$3F,$17,$F9,$2C,$7F,$96,$DF,$E4,$B6,$7F,$92,$B7
	.dc.b	$93,$F9,$2C,$ED,$FE,$4A,$FF,$83,$E7,$F9,$2B,$98,$FE,$4A,$E5,$FE
	.dc.b	$4B,$87,$F8,$26,$7F,$92,$B6,$3F,$92,$CE,$DF,$EF,$F3,$FC,$96,$75
	.dc.b	$38,$E6,$FE,$C1,$3F,$DF,$7C,$4F,$65,$FF,$25,$AF,$13,$FC,$64,$72
	.dc.b	$FF,$88,$EB,$F9,$2C,$F7,$9F,$E4,$AD,$7F,$92,$B6,$7F,$AE,$4F,$F2
	.dc.b	$58,$E5,$3F,$E5,$AC,$3D,$4F,$F9,$7A,$3F,$B5,$4F,$F2,$59,$F7,$73
	.dc.b	$7B,$2B,$F9,$2D,$7C,$98,$F5,$DF,$F7,$BF,$17,$D9,$3F,$C8,$F1,$FE
	.dc.b	$44,$7F,$A7,$B9,$7F,$BB,$3F,$EA,$0E,$6B,$3F,$C7,$0A,$7F,$93,$8F
	.dc.b	$F2,$14,$65,$C9,$FC,$70,$DC,$FF,$84,$A3,$84,$74,$8F,$F1,$8C,$7F
	.dc.b	$1F,$D4,$FF,$94,$67,$35,$FE,$13,$F1,$7F,$C6,$1E,$2F,$F1,$FB,$3F
	.dc.b	$C7,$DB,$CD,$FC,$7E,$3F,$8F,$5F,$F9,$5D,$FE,$3D,$3A,$9F,$E2,$DB
	.dc.b	$7F,$94,$67,$F8,$F5,$7F,$85,$DF,$F2,$B7,$37,$F9,$45,$FF,$27,$CE
	.dc.b	$EB,$FC,$9F,$5F,$E6,$59,$FF,$27,$BF,$EB,$B7,$F8,$FB,$98,$FF,$16
	.dc.b	$42,$AC,$FF,$8B,$F5,$FF,$26,$78,$9F,$C8,$0F,$F2,$56,$27,$F8,$FC
	.dc.b	$FF,$9C,$5F,$E3,$ED,$E4,$FE,$36,$D7,$FC,$6F,$CD,$FE,$29,$9C,$CE
	.dc.b	$67,$FD,$F4,$FF,$8D,$63,$FC,$95,$9F,$F9,$CD,$FE,$3E,$D9,$FF,$24
	.dc.b	$B6,$BF,$CD,$33,$FC,$7E,$BF,$D0,$CF,$47,$FC,$F1,$E2,$FB,$2F,$F9
	.dc.b	$04,$FF,$3B,$E6,$F1,$47,$F8,$CF,$0E,$CF,$F9,$16,$F2,$FA,$4C,$3C
	.dc.b	$0C,$3F,$E4,$18,$E2,$FF,$98,$2B,$F8,$EC,$FC,$33,$11,$FC,$69,$F8
	.dc.b	$0F,$F9,$7E,$31,$1F,$0C,$FF,$31,$72,$7F,$1C,$CB,$F8,$03,$FC,$07
	.dc.b	$9B,$FD,$44,$FF,$1C,$3F,$DE,$F3,$87,$F8,$DB,$FE,$EE,$7F,$C4,$11
	.dc.b	$FE,$5B,$7F,$DF,$51,$B7,$FD,$EC,$FF,$AB,$DF,$E3,$6D,$7F,$8D,$BC
	.dc.b	$CF,$F3,$EB,$FC,$6D,$D1,$FE,$9C,$51,$EA,$8F,$F3,$E3,$FE,$99,$8F
	.dc.b	$59,$FE,$3F,$E5,$F1,$3F,$C7,$75,$AE,$6A,$BF,$EF,$6B,$E3,$5E,$B8
	.dc.b	$E1,$3F,$E7,$F7,$F8,$E6,$5F,$EB,$D7,$31,$97,$A3,$C5,$FE,$56,$FF
	.dc.b	$AF,$DC,$BD,$6F,$D9,$5F,$51,$FF,$4B,$3F,$EB,$3F,$13,$FD,$69,$E2
	.dc.b	$DA,$FF,$20,$47,$F8,$FC,$FE,$3C,$FF,$B6,$5F,$E3,$CD,$3F,$8D,$61
	.dc.b	$FE,$D3,$8F,$F7,$A4,$64,$FF,$1C,$9F,$E9,$B7,$FC,$7A,$FF,$A5,$5F
	.dc.b	$F2,$23,$FC,$64,$C1,$FE,$47,$E4,$FF,$59,$C7,$28,$CC,$7F,$B4,$A3
	.dc.b	$84,$7F,$B7,$DF,$F1,$FF,$8B,$FE,$CF,$F1,$7F,$90,$1F,$C7,$CF,$F7
	.dc.b	$2F,$2F,$FA,$B5,$FE,$3F,$87,$FB,$AD,$FE,$3E,$7F,$BA,$2F,$53,$FE
	.dc.b	$F1,$7F,$D5,$91,$FC,$79,$B1,$FE,$F7,$7F,$DE,$11,$FE,$3B,$78,$C7
	.dc.b	$FB,$A2,$3F,$8E,$C3,$54,$7F,$8F,$23,$8F,$27,$FB,$2E,$3F,$DD,$1E
	.dc.b	$27,$F8,$F0,$FE,$3A,$7F,$1C,$C3,$E4,$DA,$3F,$8E,$C6,$1F,$F6,$5F
	.dc.b	$67,$FB,$0D,$FF,$1D,$CE,$F8,$FF,$DC,$D8,$FF,$9D,$63,$11,$FE,$E3
	.dc.b	$8E,$0F,$F8,$F1,$FA,$E2,$FB,$46,$E3,$77,$A7,$FD,$B5,$79,$7C,$51
	.dc.b	$FE,$E7,$8F,$E4,$11,$FE,$D7,$F1,$7D,$F5,$FE,$3F,$9F,$E4,$17,$DA
	.dc.b	$3F,$C7,$CD,$7F,$C7,$4D,$7F,$DB,$55,$E5,$3D,$93,$FE,$EC,$E5,$F5
	.dc.b	$CF,$FB,$6E,$7C,$AF,$F8,$F2,$BA,$47,$B9,$F0,$4F,$AD,$F1,$45,$3A
	.dc.b	$CF,$F2,$28,$F8,$71,$F0,$4F,$11,$F7,$EB,$F9,$3B,$8C,$7E,$C3,$B3
	.dc.b	$1B,$D7,$51,$FC,$AC,$FF,$22,$9F,$E4,$43,$F9,$4D,$7E,$09,$FE,$2B
	.dc.b	$1F,$EA,$F8,$C7,$1F,$F3,$18,$C1,$F7,$E7,$F9,$5C,$EA,$7F,$95,$CF
	.dc.b	$E5,$67,$F9,$5B,$63,$F9,$5C,$7E,$65,$FF,$70,$47,$C3,$35,$C9,$F0
	.dc.b	$6B,$C3,$14,$F0,$47,$BE,$3D,$11,$4F,$24,$FF,$92,$6B,$FA,$13,$FE
	.dc.b	$EC,$7F,$C8,$F1,$E3,$8F,$E5,$33,$E9,$7F,$92,$B5,$FE,$4B,$1F,$29
	.dc.b	$FF,$75,$71,$FD,$D8,$EF,$1F,$EE,$C6,$C7,$AE,$BD,$D3,$CA,$7E,$49
	.dc.b	$F0,$E7,$D8,$FC,$5E,$4B,$5F,$BD,$9F,$64,$FC,$B9,$B2,$F1,$7E,$F6
	.dc.b	$1F,$BA,$AF,$C1,$76,$7E,$15,$FF,$78,$1F,$CC,$A3,$4F,$EB,$73,$F1
	.dc.b	$DF,$F3,$68,$E0,$7F,$2E,$29,$FC,$A6,$78,$BF,$CB,$4F,$E6,$2E,$2B
	.dc.b	$E0,$47,$F9,$40,$E6,$FB,$1F,$F2,$81,$FC,$C0,$E2,$7F,$09,$3F,$DD
	.dc.b	$EF,$B1,$FF,$80,$1E,$4F,$F3,$0E,$41,$FA,$47,$1C,$96,$3F,$9A,$CD
	.dc.b	$7F,$99,$5F,$F4,$6A,$CC,$E6,$7F,$CA,$6F,$67,$FC,$A6,$E5,$FF,$24
	.dc.b	$3E,$61,$1A,$79,$57,$E6,$63,$F9,$F4,$7E,$14,$DC,$FF,$38,$8F,$E7
	.dc.b	$D1,$B8,$EA,$D3,$EF,$C6,$1E,$87,$C3,$3C,$3E,$2F,$FB,$D7,$93,$E1
	.dc.b	$3F,$EF,$7E,$BB,$C7,$FC,$E2,$AA,$3F,$CA,$AE,$1F,$E6,$8D,$7F,$CA
	.dc.b	$B9,$7C,$7A,$FF,$2C,$B6,$3F,$DE,$F3,$F5,$27,$FC,$B2,$D8,$FE,$6C
	.dc.b	$F8,$AF,$F3,$B8,$2B,$33,$E8,$9F,$7C,$7F,$3A,$8F,$D2,$CB,$C6,$F8
	.dc.b	$0F,$1B,$F5,$10,$F6,$61,$E5,$7F,$9B,$35,$D5,$7B,$57,$2F,$6C,$3F
	.dc.b	$DF,$93,$F0,$63,$EF,$5F,$7A,$FE,$7D,$3F,$1E,$7E,$E4,$E6,$7F,$01
	.dc.b	$5F,$E6,$0B,$F1,$5E,$27,$F9,$E3,$F1,$4F,$87,$3F,$CE,$A7,$E7,$47
	.dc.b	$E0,$3C,$6F,$F7,$DD,$9F,$F2,$3F,$27,$F9,$32,$BF,$35,$18,$AF,$D1
	.dc.b	$3F,$CE,$9F,$D3,$BF,$CE,$9C,$CF,$F3,$96,$BF,$EF,$C7,$F3,$71,$97
	.dc.b	$31,$F9,$B8,$FF,$34,$C7,$C0,$51,$FC,$E2,$30,$FF,$98,$1E,$0F,$AF
	.dc.b	$97,$D9,$1F,$9C,$AF,$B7,$5E,$CE,$5E,$F7,$D0,$FF,$7E,$CF,$B9,$FF
	.dc.b	$40,$C7,$E1,$27,$FC,$D3,$CB,$67,$81,$FE,$FE,$BF,$5B,$FE,$60,$BE
	.dc.b	$4F,$F9,$71,$AF,$F3,$37,$37,$E8,$8F,$E6,$D3,$F2,$23,$F3,$37,$F0
	.dc.b	$2E,$C7,$D7,$7F,$6F,$E3,$7E,$5B,$C6,$FC,$BF,$0D,$C3,$FD,$FD,$97
	.dc.b	$F1,$19,$FE,$ED,$CD,$FC,$65,$FE,$69,$3F,$C7,$9F,$F3,$2C,$EE,$3F
	.dc.b	$CC,$A6,$4F,$E6,$8E,$5F,$E6,$86,$4F,$E6,$87,$EA,$A3,$FC,$CB,$CB
	.dc.b	$FC,$55,$FE,$69,$3D,$63,$F6,$B1,$C2,$7F,$42,$FF,$35,$79,$57,$F1
	.dc.b	$4F,$1B,$F6,$7E,$37,$F3,$3C,$3F,$96,$3F,$E6,$4E,$7F,$E6,$8F,$F3
	.dc.b	$28,$CF,$3F,$F3,$2C,$3F,$9D,$CF,$F3,$A8,$FE,$4F,$3A,$7F,$99,$CF
	.dc.b	$F2,$E9,$FE,$67,$39,$9F,$E4,$8D,$9F,$D9,$CF,$F9,$9D,$F0,$CF,$F9
	.dc.b	$8D,$FE,$17,$1F,$CC,$4C,$38,$AF,$E5,$F1,$FC,$3B,$93,$F9,$65,$FF
	.dc.b	$47,$7F,$96,$35,$FE,$58,$D7,$F8,$96,$9F,$C6,$1F,$E6,$17,$FD,$55
	.dc.b	$FE,$2C,$7F,$2C,$7F,$AD,$BF,$CB,$1A,$FF,$2C,$7F,$A6,$BF,$CB,$1A
	.dc.b	$FF,$2C,$9F,$17,$3F,$F2,$C8,$FE,$57,$1E,$78,$FE,$93,$1F,$CA,$E7
	.dc.b	$FB,$34,$7B,$63,$C6,$7F,$28,$8F,$0B,$FC,$A1,$AF,$F2,$89,$F7,$3F
	.dc.b	$E6,$68,$CB,$FE,$66,$8C,$BF,$CB,$23,$E1,$3F,$CC,$1A,$FF,$51,$E6
	.dc.b	$FE,$52,$E6,$33,$1F,$DA,$63,$F9,$7C,$72,$8F,$74,$7B,$1F,$6C,$7F
	.dc.b	$66,$7F,$A2,$F8,$DF,$CA,$0F,$E4,$FC,$BF,$E0,$57,$F9,$35,$FF,$80
	.dc.b	$5F,$E4,$D5,$FD,$E9,$FE,$4D,$19,$E5,$FE,$4C,$D7,$F9,$37,$32,$7F
	.dc.b	$93,$4E,$67,$FC,$01,$1F,$C9,$AB,$FA,$4B,$FC,$6D,$CC,$7F,$25,$7F
	.dc.b	$A4,$C7,$F2,$96,$DF,$F8,$02,$3F,$AF,$F8,$BE,$CA,$FF,$7F,$DF,$77
	.dc.b	$F9,$37,$2F,$B5,$FE,$4B,$85,$7F,$CB,$6F,$F8,$A2,$F2,$E6,$3F,$8D
	.dc.b	$3F,$DB,$A3,$F9,$4F,$37,$92,$7A,$CF,$A9,$FE,$7F,$1E,$A7,$F9,$2D
	.dc.b	$FC,$78,$FE,$FB,$5D,$2B,$DB,$1E,$BB,$FF,$0D,$F8,$DE,$C9,$FE,$46
	.dc.b	$7F,$8F,$31,$FE,$47,$87,$F8,$D5,$FE,$47,$1F,$E2,$E7,$F9,$1B,$B3
	.dc.b	$F9,$19,$93,$F9,$19,$93,$F9,$19,$93,$F8,$E4,$75,$47,$F9,$45,$C2
	.dc.b	$E4,$FF,$1D,$4E,$B9,$BF,$C3,$2F,$F2,$59,$FF,$0A,$BF,$C9,$5F,$F2
	.dc.b	$93,$C5,$FF,$0F,$F8,$DF,$E1,$8F,$1B,$F9,$34,$7F,$96,$9F,$E4,$D8
	.dc.b	$7F,$94,$DF,$E4,$D1,$9E,$6F,$E4,$CF,$F9,$3A,$3F,$93,$47,$F9,$22
	.dc.b	$78,$38,$E6,$FF,$0B,$CF,$F2,$6A,$E5,$3F,$E6,$69,$FF,$0B,$CF,$F9
	.dc.b	$52,$7F,$C5,$AF,$F2,$7A,$E7,$5F,$E3,$68,$FF,$2B,$1F,$C9,$A7,$FC
	.dc.b	$DC,$FF,$94,$DE,$E7,$F9,$5B,$4F,$F1,$AF,$8B,$FC,$A2,$3F,$CE,$EF
	.dc.b	$F2,$86,$F2,$7F,$28,$6B,$FC,$A1,$FF,$3D,$47,$F2,$8C,$3F,$CE,$CF
	.dc.b	$F2,$8E,$6F,$F1,$9B,$FC,$9E,$7F,$CE,$2F,$17,$FC,$8E,$FF,$27,$E4
	.dc.b	$3F,$E5,$E8,$E3,$6E,$6F,$F2,$9C,$FF,$28,$9F,$3C,$7F,$9C,$23,$F9
	.dc.b	$46,$1F,$E8,$7F,$1B,$DB,$7F,$CA,$0F,$39,$FC,$AB,$9E,$BF,$CF,$B0
	.dc.b	$AF,$F2,$77,$CE,$E5,$CB,$FC,$9D,$A7,$99,$FE,$4C,$D7,$FD,$19,$7F
	.dc.b	$E4,$D8,$F9,$26,$23,$FC,$9A,$F6,$3D,$B3,$EB,$7F,$D1,$7C,$DF,$E4
	.dc.b	$C3,$F9,$2E,$1F,$EA,$37,$F9,$2C,$7F,$AB,$9F,$E4,$B9,$7F,$93,$63
	.dc.b	$FD,$5D,$1F,$E2,$E3,$F8,$D4,$F1,$7F,$C6,$31,$FE,$68,$7F,$92,$BC
	.dc.b	$DF,$E4,$4E,$5F,$F2,$4F,$3F,$FA,$9D,$CC,$66,$3F,$D0,$F1,$FC,$96
	.dc.b	$3A,$47,$A9,$FF,$52,$BF,$E8,$48,$FF,$25,$4F,$8D,$F3,$6B,$E8,$7F
	.dc.b	$92,$C6,$5F,$E3,$AD,$7F,$C9,$17,$FE,$4C,$BF,$F4,$8C,$7F,$24,$7D
	.dc.b	$0F,$FB,$F5,$FF,$60,$BF,$E5,$7E,$6F,$53,$FC,$86,$38,$A8,$F5,$BF
	.dc.b	$E4,$C8,$F9,$71,$F5,$2F,$FD,$7E,$FF,$A7,$BC,$5F,$F4,$FF,$8D,$6B
	.dc.b	$F9,$39,$FE,$D0,$8F,$E4,$EF,$FB,$91,$FE,$4F,$81,$FE,$4F,$1B,$3F
	.dc.b	$93,$C6,$5F,$E4,$E6,$4F,$E4,$EE,$5F,$E4,$E7,$53,$F9,$11,$FE,$A0
	.dc.b	$8F,$F2,$4B,$9E,$6F,$F5,$EC,$7F,$25,$8F,$F7,$A3,$FC,$97,$2F,$F7
	.dc.b	$83,$FC,$96,$3F,$DC,$DE,$37,$FA,$F3,$C6,$FE,$4B,$1F,$EE,$C7,$F9
	.dc.b	$2B,$FF,$00,$1F,$E4,$AE,$79,$7F,$92,$E2,$9F,$E4,$B0,$3F,$D7,$2E
	.dc.b	$39,$AC,$7F,$21,$C7,$FA,$0D,$F3,$8F,$E4,$CE,$9F,$F2,$44,$6A,$FB
	.dc.b	$46,$A7,$47,$FB,$FE,$FF,$DF,$87,$F2,$7A,$E0,$FF,$27,$8F,$F6,$DB
	.dc.b	$FE,$DA,$F1,$7F,$93,$42,$D1,$FC,$9E,$30,$FF,$8D,$B9,$7F,$DC,$6F
	.dc.b	$F8,$E3,$9A,$F2,$7F,$B8,$67,$FD,$AF,$3F,$C9,$EB,$33,$FE,$FC,$58
	.dc.b	$FA,$E7,$F9,$44,$60,$FE,$C7,$87,$8A,$B8,$46,$4F,$1B,$FE,$51,$77
	.dc.b	$3A,$75,$3C,$70,$FF,$7F,$CF,$F2,$F9,$B1,$F4,$1F,$F7,$4F,$8D,$F0
	.dc.b	$EF,$F9,$66,$1E,$57,$F9,$63,$5F,$E5,$8D,$7F,$96,$61,$E7,$7F,$C9
	.dc.b	$D1,$97,$F9,$63,$6B,$E2,$C0,$FE,$5A,$FC,$67,$C8,$FA,$B0,$FE,$5C
	.dc.b	$FC,$08,$F5,$9F,$CB,$AB,$E3,$47,$BB,$97,$8B,$F5,$8A,$7C,$4E,$BA
	.dc.b	$FC,$6F,$17,$FD,$FA,$7F,$32,$7E,$8B,$FC,$C9,$FC,$23,$FC,$C8,$FF
	.dc.b	$2A,$BA,$E6,$FF,$2A,$B5,$FF,$2A,$C7,$F3,$08,$E3,$1F,$CB,$1F,$03
	.dc.b	$FC,$C1,$FA,$59,$7E,$76,$75,$1A,$E7,$F5,$CE,$6B,$D6,$EE,$2C,$FA
	.dc.b	$0A,$78,$27,$FD,$EF,$3C,$5F,$E6,$6D,$7F,$94,$47,$91,$FF,$2B,$BC
	.dc.b	$5F,$F2,$7B,$FE,$36,$8E,$10,$E7,$FD,$CC,$73,$E7,$F5,$D7,$C3,$AF
	.dc.b	$F7,$E3,$F1,$63,$EE,$CF,$8E,$AC,$FC,$B9,$F7,$BF,$36,$2B,$F4,$A3
	.dc.b	$DB,$9B,$1F,$E7,$53,$FA,$0F,$1B,$EE,$E1,$FC,$1E,$3F,$03,$CF,$FA
	.dc.b	$47,$F9,$B9,$FB,$F3,$F9,$B9,$FC,$DA,$FF,$13,$1E,$C8,$FD,$99,$F9
	.dc.b	$4D,$30,$F1,$3F,$DF,$66,$27,$4D,$32,$53,$FC,$B1,$5E,$57,$FD,$F8
	.dc.b	$D9,$F7,$CF,$F3,$E9,$F8,$A7,$9E,$3E,$0D,$7B,$A3,$FC,$D3,$E5,$7F
	.dc.b	$30,$E5,$FF,$2A,$E1,$FC,$B2,$31,$1F,$E5,$18,$C3,$FC,$EA,$33,$CD
	.dc.b	$F7,$6B,$9C,$F3,$70,$27,$B4,$FC,$27,$51,$A7,$F9,$F3,$99,$FF,$2F
	.dc.b	$3F,$C6,$63,$FC,$CC,$FE,$18,$E9,$1F,$D1,$CF,$BF,$16,$6C,$7F,$3B
	.dc.b	$9F,$2C,$7E,$7A,$7F,$CC,$91,$D6,$3E,$23,$E7,$2C,$3C,$6F,$E7,$ED
	.dc.b	$E2,$FE,$75,$87,$C7,$9F,$E7,$D5,$E7,$7F,$9F,$36,$7F,$9F,$3E,$F8
	.dc.b	$FE,$7C,$E6,$3F,$9F,$4F,$37,$FC,$C1,$19,$E6,$FB,$77,$E1,$97,$36
	.dc.b	$30,$FE,$75,$87,$18,$F8,$71,$FC,$E7,$0D,$47,$C1,$F1,$FD,$B3,$FE
	.dc.b	$63,$BF,$11,$FC,$B4,$FF,$7F,$5F,$8D,$FE,$73,$7F,$88,$9F,$F3,$43
	.dc.b	$C2,$3F,$9A,$3E,$D7,$91,$FC,$E9,$E0,$E1,$FC,$14,$6B,$97,$EE,$63
	.dc.b	$ED,$AF,$61,$FC,$FA,$BA,$3E,$C3,$FC,$CF,$1F,$4E,$3D,$F3,$F7,$6B
	.dc.b	$EC,$F8,$FF,$E5,$FE,$51,$31,$FC,$F4,$7F,$39,$BF,$CB,$3F,$CE,$70
	.dc.b	$3C,$79,$BF,$2C,$FF,$3A,$9D,$DF,$F3,$A6,$BF,$E6,$2B,$EA,$F8,$05
	.dc.b	$7B,$AB,$FC,$DF,$5F,$A3,$75,$1F,$79,$FE,$77,$5C,$A3,$F3,$B1,$FE
	.dc.b	$6B,$9F,$1C,$7E,$6E,$3F,$2F,$3F,$02,$BE,$07,$2F,$EA,$67,$E1,$47
	.dc.b	$F3,$19,$F2,$BF,$E6,$56,$BF,$E6,$CB,$FE,$0B,$3F,$E8,$7A,$F3,$4F
	.dc.b	$AC,$FE,$75,$3F,$11,$FE,$86,$D7,$FC,$DD,$87,$CA,$7F,$A1,$C7,$27
	.dc.b	$8B,$D2,$7D,$B1,$FE,$6D,$BF,$AB,$1E,$DE,$6F,$DC,$78,$FF,$B4,$F1
	.dc.b	$FF,$6B,$C1,$71,$FF,$36,$E4,$9F,$E6,$ED,$8F,$F0,$2F,$27,$F0,$F9
	.dc.b	$FF,$36,$CE,$67,$F9,$7E,$5F,$C8,$67,$F9,$9C,$7F,$11,$7F,$C4,$15
	.dc.b	$F9,$67,$FC,$45,$CB,$67,$F9,$81,$D4,$FE,$6E,$7F,$2E,$8C,$B9,$8F
	.dc.b	$E2,$F1,$FE,$6D,$8D,$47,$F1,$5F,$1F,$F8,$8F,$8F,$FE,$6C,$C4,$FF
	.dc.b	$9B,$2B,$F9,$63,$FC,$CC,$FE,$5A,$FF,$27,$7F,$96,$E3,$FC,$ED,$FE
	.dc.b	$5B,$19,$E7,$FF,$35,$3F,$D0,$A3,$F9,$6B,$FC,$CD,$FE,$63,$3C,$67
	.dc.b	$FC,$D1,$39,$7F,$CE,$CF,$F1,$49,$CB,$93,$F9,$9B,$FE,$69,$7C,$87
	.dc.b	$F3,$1E,$4F,$E5,$67,$F9,$9E,$FF,$DE,$9C,$DF,$D0,$E3,$F9,$4E,$1F
	.dc.b	$D4,$DF,$E1,$67,$F1,$B7,$0F,$F0,$ED,$7F,$A6,$3F,$CB,$AF,$93,$FC
	.dc.b	$AA,$76,$FF,$23,$8C,$BF,$CA,$9F,$E2,$2F,$F2,$A7,$FB,$2B,$FE,$4F
	.dc.b	$9F,$E9,$AF,$18,$EF,$1F,$D2,$EB,$FA,$3C,$FF,$2B,$AE,$EF,$F1,$D6
	.dc.b	$BF,$CB,$6B,$FA,$E3,$FC,$AA,$32,$FA,$8F,$E4,$91,$E6,$7F,$96,$B7
	.dc.b	$97,$CF,$1C,$E3,$82,$8F,$E8,$EE,$63,$A4,$7A,$A3,$E9,$3F,$4E,$3F
	.dc.b	$B2,$3F,$D8,$3C,$7F,$E0,$FC,$7F,$DC,$5F,$E4,$D9,$1F,$E4,$C7,$F2
	.dc.b	$5A,$D4,$FF,$05,$3F,$BF,$45,$E5,$B1,$FC,$82,$76,$FF,$96,$DF,$F0
	.dc.b	$14,$E5,$CC,$FF,$60,$8F,$E4,$71,$FE,$18,$78,$3D,$0F,$F0,$0B,$FC
	.dc.b	$B9,$A7,$F7,$FF,$1B,$D9,$5F,$C9,$32,$FF,$0F,$57,$F2,$4E,$4F,$1C
	.dc.b	$FF,$97,$27,$2F,$AE,$3C,$AF,$F2,$1E,$5F,$5B,$FC,$F6,$2A,$8F,$EB
	.dc.b	$75,$FC,$9E,$6C,$FF,$22,$C3,$9B,$FC,$89,$EB,$5E,$A8,$FE,$45,$1E
	.dc.b	$F8,$FF,$0C,$57,$F9,$62,$FF,$C6,$DE,$3F,$F8,$5B,$1F,$F2,$BC,$7F
	.dc.b	$8E,$5F,$E4,$27,$F9,$52,$FF,$C5,$4F,$F2,$03,$F9,$1B,$FE,$31,$7F
	.dc.b	$8F,$C7,$18,$FF,$28,$9C,$0F,$F7,$7F,$1F,$F7,$69,$CC,$64,$47,$F8
	.dc.b	$5A,$3F,$8F,$C7,$18,$FF,$1A,$C7,$F1,$F8,$FF,$22,$3C,$1F,$F1,$CF
	.dc.b	$8F,$FE,$30,$F1,$FF,$90,$1F,$C7,$F6,$3F,$E5,$13,$F8,$FE,$1F,$E5
	.dc.b	$77,$F8,$FD,$7F,$97,$1F,$E3,$F1,$9E,$4F,$E3,$F3,$FE,$53,$9F,$E3
	.dc.b	$EF,$F9,$35,$FE,$3E,$F1,$8F,$E3,$73,$C5,$FF,$26,$CF,$F1,$E9,$E1
	.dc.b	$1F,$E4,$B8,$FE,$3D,$3E,$39,$FF,$27,$6D,$FE,$4A,$F1,$BF,$C9,$27
	.dc.b	$F1,$F3,$FC,$8F,$3F,$E6,$F7,$F8,$FC,$FF,$9D,$1F,$E3,$FC,$7F,$E6
	.dc.b	$77,$FC,$8F,$3B,$E3,$FF,$1B,$4F,$F8,$A2,$7F,$91,$56,$E7,$FE,$00
	.dc.b	$D6,$DF,$E3,$D1,$9A,$FF,$33,$D7,$F1,$D8,$FF,$26,$BF,$C7,$5F,$F3
	.dc.b	$A3,$C1,$FF,$1A,$F8,$FE,$BA,$FF,$7F,$C7,$9D,$FF,$7F,$B5,$FF,$23
	.dc.b	$BF,$E8,$19,$FE,$3C,$E6,$3F,$8C,$5F,$95,$FF,$7D,$B5,$F6,$1F,$E5
	.dc.b	$3E,$4A,$FF,$22,$9F,$2B,$FE,$45,$7F,$D0,$D5,$FC,$7D,$CC,$FB,$A7
	.dc.b	$FD,$01,$3F,$C9,$67,$E0,$9F,$E8,$3E,$3F,$E3,$F7,$FE,$A1,$7F,$8F
	.dc.b	$9F,$E4,$33,$F8,$F3,$97,$F8,$F4,$7F,$A6,$5F,$E3,$CD,$7F,$8F,$46
	.dc.b	$DF,$E2,$ED,$7F,$CF,$67,$F9,$02,$3F,$CC,$2F,$F1,$E7,$FD,$50,$FF
	.dc.b	$1E,$7A,$1F,$E9,$E7,$F8,$F3,$C8,$FF,$4E,$72,$7F,$A6,$23,$FC,$7F
	.dc.b	$C9,$FE,$A9,$7F,$8F,$57,$85,$FF,$7A,$B5,$FF,$1D,$B9,$8F,$F5,$B4
	.dc.b	$7F,$21,$AF,$63,$FC,$66,$7D,$5A,$7A,$23,$F8,$E4,$F7,$7F,$CD,$F1
	.dc.b	$EE,$7F,$92,$45,$8F,$55,$7D,$17,$FD,$68,$FF,$AC,$3C,$6F,$F5,$8F
	.dc.b	$8F,$6B,$FC,$7B,$1F,$E3,$C3,$F8,$E6,$1F,$ED,$77,$FC,$76,$7F,$1C
	.dc.b	$8F,$F6,$43,$FC,$71,$FF,$6F,$BF,$E3,$A7,$FD,$A4,$FF,$8E,$9A,$7F
	.dc.b	$8D,$A3,$A3,$FE,$3A,$3F,$DB,$51,$FC,$72,$33,$C9,$FE,$D0,$8F,$E3
	.dc.b	$9C,$47,$F8,$E5,$7F,$AE,$3C,$7F,$F6,$5F,$8F,$FE,$3B,$3F,$8E,$1F
	.dc.b	$EE,$B3,$F8,$DC,$7F,$BD,$1F,$E3,$6F,$FB,$A5,$FE,$36,$E7,$93,$F8
	.dc.b	$DD,$A7,$F8,$DE,$3F,$EF,$97,$F8,$DD,$71,$AF,$E3,$6F,$17,$FD,$CF
	.dc.b	$1F,$C6,$E7,$84,$7F,$B9,$63,$F8,$DC,$F8,$CF,$F7,$47,$8D,$FE,$E2
	.dc.b	$3F,$C7,$07,$F1,$A4,$7F,$BF,$DF,$E3,$6D,$E1,$FF,$1A,$5E,$A3,$11
	.dc.b	$FE,$CF,$EA,$C4,$E7,$15,$FF,$8E,$ED,$39,$8F,$E3,$F1,$E0,$8C,$5F
	.dc.b	$F2,$08,$F1,$D7,$28,$FE,$3D,$1D,$8C,$D7,$F8,$FD,$F3,$5F,$95,$D1
	.dc.b	$EB,$E5,$F8,$31,$98,$CD,$BC,$7C,$46,$2F,$FC,$84,$F6,$C3,$F9,$2B
	.dc.b	$F0,$1F,$F7,$16,$3E,$87,$FC,$7E,$D7,$F9,$2C,$71,$7F,$91,$3E,$F3
	.dc.b	$0F,$F2,$38,$E1,$CB,$FE,$EA,$AF,$F7,$75,$78,$EB,$F9,$35,$77,$8F
	.dc.b	$53,$F2,$2B,$D2,$FC,$A8,$A7,$C6,$8F,$C0,$BE,$E3,$E9,$9F,$3B,$7C
	.dc.b	$38,$18,$BC,$7F,$E5,$B1,$F5,$1F,$F2,$59,$FE,$EF,$3F,$96,$B9,$E4
	.dc.b	$FE,$5A,$F3,$9E,$51,$FE,$EF,$6B,$FE,$EF,$8E,$0F,$F2,$E6,$BE,$E7
	.dc.b	$3C,$B5,$F2,$D7,$F2,$FA,$EB,$3F,$02,$73,$5E,$FC,$7F,$29,$7F,$CC
	.dc.b	$27,$F0,$D1,$87,$47,$82,$7F,$E0,$07,$27,$A6,$BF,$98,$CE,$67,$F9
	.dc.b	$1D,$74,$7F,$97,$35,$FF,$75,$B5,$FE,$57,$1F,$81,$3C,$AF,$F3,$37
	.dc.b	$DF,$CB,$E4,$7E,$1D,$FB,$F1,$F8,$F3,$70,$FE,$6F,$5E,$5B,$FF,$27
	.dc.b	$BF,$12,$FD,$6F,$C6,$AB,$6F,$1F,$EF,$E3,$FB,$7A,$FF,$28,$9F,$B0
	.dc.b	$3F,$DE,$27,$F3,$63,$07,$EA,$73,$F5,$57,$F9,$56,$3E,$F1,$FC,$DC
	.dc.b	$FC,$49,$FC,$DC,$D9,$FC,$DC,$A7,$F3,$73,$A9,$FC,$C0,$CF,$83,$F6
	.dc.b	$E3,$FC,$B3,$1F,$0C,$FE,$6E,$6C,$FD,$DC,$7B,$1F,$E3,$7E,$5B,$96
	.dc.b	$10,$B8,$7F,$1B,$8C,$46,$7C,$7F,$D1,$46,$23,$8D,$61,$FE,$7E,$A3
	.dc.b	$51,$FC,$41,$FF,$7D,$C6,$E3,$FC,$B3,$1F,$EF,$A8,$C4,$7F,$BC,$9E
	.dc.b	$0F,$9A,$7F,$3D,$3E,$79,$B3,$FC,$FD,$F6,$BC,$5F,$C4,$1E,$37,$FA
	.dc.b	$0B,$4F,$C1,$C7,$FB,$ED,$F6,$9F,$1C,$F5,$78,$DF,$EF,$AB,$F8,$8F
	.dc.b	$FB,$E3,$8B,$FD,$D5,$18,$7F,$CB,$B3,$F1,$A7,$FA,$3D,$1C,$C6,$67
	.dc.b	$F9,$FD,$71,$47,$FB,$ED,$C3,$FE,$57,$75,$CD,$F0,$AB,$FA,$75,$7B
	.dc.b	$AF,$B5,$FF,$97,$EF,$C9,$CD,$FC,$06,$7C,$58,$7C,$0C,$3F,$CC,$58
	.dc.b	$7D,$5F,$1F,$E0,$57,$2C,$7D,$D8,$F8,$9F,$F2,$C3,$5F,$3D,$7A,$DF
	.dc.b	$E7,$EF,$E3,$AF,$E3,$3F,$BD,$AE,$93,$FE,$64,$8D,$3F,$CF,$1E,$4F
	.dc.b	$F9,$6D,$E1,$39,$7B,$72,$FC,$19,$FF,$34,$46,$9E,$0F,$E3,$E7,$2F
	.dc.b	$29,$F8,$0F,$E0,$3C,$7F,$E7,$2F,$E5,$DF,$F7,$F7,$20,$73,$7E,$82
	.dc.b	$33,$1E,$B7,$F9,$EC,$F1,$8F,$E7,$B9,$7E,$A6,$7F,$CD,$33,$C6,$7F
	.dc.b	$CD,$2F,$EC,$A3,$B4,$7F,$3D,$8E,$4F,$85,$FB,$91,$FE,$6B,$7F,$54
	.dc.b	$FF,$9A,$F2,$3F,$E6,$B8,$EC,$7D,$F7,$F9,$E3,$C0,$FC,$FD,$7D,$E8
	.dc.b	$FF,$35,$BE,$7A,$F8,$3C,$9E,$77,$CA,$7F,$3C,$AF,$43,$FC,$F1,$AF
	.dc.b	$F3,$C6,$BF,$CF,$1E,$4F,$F9,$99,$F9,$0F,$C0,$8F,$6B,$FE,$6E,$9F
	.dc.b	$12,$8F,$DF,$D7,$F3,$C9,$F3,$47,$B7,$97,$ED,$3F,$BA,$F1,$FF,$33
	.dc.b	$E3,$FE,$6B,$86,$E5,$FE,$1A,$E4,$FE,$33,$3F,$DF,$F9,$AC,$FF,$88
	.dc.b	$67,$F2,$D3,$FE,$20,$9B,$3F,$CD,$F2,$FE,$49,$3F,$CE,$F5,$3F,$CE
	.dc.b	$E7,$99,$FE,$48,$AF,$DA,$47,$F9,$16,$FF,$8C,$BF,$CE,$27,$F9,$23
	.dc.b	$FE,$71,$7F,$9A,$BC,$5D,$C7,$F1,$68,$FE,$79,$1F,$AB,$F1,$FF,$6F
	.dc.b	$E3,$FF,$2E,$E6,$FE,$6A,$FF,$30,$E7,$B1,$FC,$DE,$AB,$FC,$ED,$AF
	.dc.b	$F3,$A6,$A8,$FE,$71,$3F,$D0,$A6,$CF,$F3,$36,$BF,$CB,$DE,$B3,$FC
	.dc.b	$DA,$7A,$47,$08,$FE,$25,$1B,$8F,$E0,$8F,$F3,$4C,$8F,$F1,$2E,$5F
	.dc.b	$E6,$13,$FD,$61,$FE,$60,$35,$FE,$A6,$FF,$9D,$E3,$FA,$83,$FC,$C0
	.dc.b	$FE,$5F,$AF,$F5,$B7,$F9,$69,$FC,$A2,$38,$B8,$E7,$FE,$54,$FF,$9C
	.dc.b	$A3,$2A,$3F,$A8,$BC,$63,$F8,$E3,$FE,$6F,$C3,$11,$89,$FE,$9B,$3F
	.dc.b	$CB,$A3,$F9,$14,$7B,$63,$D7,$5E,$57,$FC,$D7,$5E,$E7,$A9,$E8,$3F
	.dc.b	$A2,$F2,$E1,$FF,$31,$CF,$17,$FC,$D7,$7E,$87,$FC,$D4,$FC,$97,$2E
	.dc.b	$4F,$7B,$FE,$6A,$74,$FF,$47,$3B,$BF,$01,$F5,$CF,$B2,$3F,$B4,$BF
	.dc.b	$D2,$3C,$7F,$E5,$76,$7F,$93,$F3,$7F,$82,$1F,$E5,$35,$79,$BF,$86
	.dc.b	$9F,$CB,$9C,$3F,$C4,$1F,$E9,$11,$FC,$AB,$1F,$F0,$A3,$FE,$6A,$8D
	.dc.b	$CF,$F9,$66,$72,$FF,$2C,$8F,$EC,$93,$FC,$A5,$E7,$1F,$CB,$5C,$BF
	.dc.b	$C9,$9F,$21,$FD,$9E,$30,$FA,$4F,$F0,$3F,$8D,$EF,$9F,$E8,$FC,$DE
	.dc.b	$27,$F8,$E5,$F8,$9F,$E5,$31,$E0,$7F,$95,$F3,$57,$FA,$74,$7F,$89
	.dc.b	$70,$F8,$73,$67,$FC,$DF,$85,$7F,$96,$3F,$1A,$BF,$C4,$75,$FC,$AE
	.dc.b	$3D,$71,$E9,$9F,$F3,$0E,$5F,$4E,$3F,$C3,$F5,$FC,$B6,$BF,$C0,$9E
	.dc.b	$3F,$F8,$03,$2F,$F2,$FC,$7F,$8E,$5F,$E4,$D9,$7F,$90,$DF,$F3,$04
	.dc.b	$64,$FF,$30,$1F,$E4,$47,$F9,$53,$FD,$FD,$FE,$4C,$7F,$25,$8E,$51
	.dc.b	$FC,$84,$E0,$7F,$97,$CE,$87,$F9,$65,$DF,$2F,$F8,$66,$7F,$92,$4F
	.dc.b	$F8,$7D,$47,$D4,$E6,$FF,$1D,$F8,$FF,$E1,$AF,$1F,$F9,$26,$1F,$E5
	.dc.b	$B7,$F9,$23,$79,$BF,$92,$35,$FE,$48,$D7,$F9,$23,$5F,$E4,$98,$FF
	.dc.b	$86,$E7,$F9,$23,$63,$F9,$23,$5F,$E4,$93,$D7,$9B,$FC,$B3,$5F,$C9
	.dc.b	$2B,$84,$7F,$8D,$63,$F9,$24,$7F,$9B,$9F,$E4,$87,$F8,$D3,$83,$FC
	.dc.b	$6B,$E3,$7F,$24,$9F,$F3,$73,$FC,$93,$9B,$FC,$EA,$FF,$23,$E1,$3F
	.dc.b	$E5,$23,$F9,$1C,$FF,$A0,$9F,$E1,$9C,$BF,$CA,$E3,$FC,$6D,$1F,$C3
	.dc.b	$F9,$BF,$C9,$8F,$F2,$3C,$FF,$D0,$8F,$F2,$79,$CD,$7F,$96,$30,$FF
	.dc.b	$3A,$CF,$F2,$33,$13,$8B,$FF,$37,$C7,$F2,$38,$FF,$45,$F8,$FE,$CB
	.dc.b	$FE,$47,$1E,$57,$F9,$1C,$FF,$40,$8F,$6C,$7A,$39,$FC,$EF,$AA,$2B
	.dc.b	$FC,$8B,$1F,$3B,$FE,$4D,$6B,$FC,$72,$36,$FA,$0F,$E4,$4F,$CC,$7F
	.dc.b	$C9,$31,$CA,$7F,$CD,$55,$FE,$4C,$AF,$9F,$1E,$C7,$FD,$17,$CD,$FC
	.dc.b	$88,$FE,$43,$5F,$EA,$B7,$F9,$0B,$5F,$E4,$3C,$FF,$E6,$57,$FC,$A3
	.dc.b	$35,$FF,$18,$C7,$FA,$A5,$FE,$41,$3C,$5F,$F2,$46,$9F,$C8,$EB,$FD
	.dc.b	$6B,$1F,$C7,$DA,$FF,$1F,$79,$3F,$C7,$E7,$C5,$1F,$E8,$58,$FE,$3F
	.dc.b	$1E,$78,$FF,$52,$D7,$FA,$12,$7F,$C8,$F1,$FE,$CA,$7F,$E0,$0B,$FE
	.dc.b	$BF,$7F,$91,$BE,$07,$F8,$FB,$5F,$E3,$EF,$B5,$FE,$30,$E7,$9F,$FD
	.dc.b	$35,$1C,$23,$51,$FE,$C0,$8D,$C7,$C5,$8F,$44,$7F,$1E,$3F,$C9,$D1
	.dc.b	$CB,$6F,$F5,$EC,$FF,$A7,$3C,$6F,$F4,$EF,$8F,$6B,$FC,$89,$8F,$FB
	.dc.b	$7E,$7F,$8F,$C7,$FB,$99,$FE,$3F,$91,$FF,$22,$47,$FB,$89,$FE,$3E
	.dc.b	$D3,$FD,$F5,$3C,$DF,$F2,$01,$FE,$E1,$8F,$F7,$F3,$B3,$F8,$C9,$FE
	.dc.b	$B9,$7F,$92,$3C,$F9,$BF,$D7,$11,$FC,$82,$39,$47,$FB,$62,$3F,$C8
	.dc.b	$71,$FE,$F4,$F1,$FF,$D7,$3E,$3F,$F2,$08,$FF,$7F,$BF,$C8,$1B,$CD
	.dc.b	$FC,$81,$AF,$F2,$06,$BF,$C8,$1A,$FF,$20,$7F,$E0,$0C,$7F,$20,$6C
	.dc.b	$7F,$20,$6B,$FC,$81,$EB,$CD,$FE,$F8,$9F,$E4,$11,$C2,$3F,$DA,$91
	.dc.b	$FC,$83,$14,$FA,$27,$FD,$EB,$3F,$ED,$4F,$1B,$F9,$06,$67,$F9,$05
	.dc.b	$F0,$B4,$E7,$8F,$FD,$E7,$39,$8C,$47,$F8,$AF,$9B,$FD,$A9,$7F,$C9
	.dc.b	$2F,$8B,$FC,$92,$3A,$5F,$18,$FE,$49,$38,$3C,$F7,$D2,$FF,$DE,$D1
	.dc.b	$89,$FE,$4C,$EC,$EB,$1F,$C9,$1C,$4F,$9F,$9B,$FD,$ED,$59,$9C,$BF
	.dc.b	$E5,$17,$38,$7B,$DF,$F8,$01,$E3,$FC,$5C,$3F,$94,$C7,$67,$F9,$36
	.dc.b	$3E,$97,$FD,$E6,$D7,$FD,$EF,$1B,$7F,$CC,$AF,$C4,$7F,$CA,$1C,$BE
	.dc.b	$27,$F9,$54,$7B,$5F,$F2,$5C,$F3,$7C,$90,$8E,$D3,$87,$C1,$14,$E5
	.dc.b	$1F,$02,$78,$C7,$D4,$8A,$F3,$3E,$61,$89,$FB,$9D,$5A,$8D,$1F,$82
	.dc.b	$F1,$FF,$CA,$67,$F3,$27,$EB,$BF,$CC,$9B,$CD,$FC,$C9,$E3,$AA,$7F
	.dc.b	$99,$C6,$1F,$F2,$93,$C2,$7F,$C9,$CF,$E6,$2F,$F1,$D3,$FC,$E2,$3E
	.dc.b	$EB,$FE,$5B,$6C,$8F,$F7,$AC,$F0,$7F,$9B,$47,$33,$CB,$CF,$F1,$27
	.dc.b	$51,$4F,$5C,$7C,$37,$DA,$53,$E1,$4F,$F9,$5A,$7B,$3F,$E4,$E6,$BF
	.dc.b	$E4,$E6,$BE,$5E,$7F,$14,$F2,$8F,$F7,$F3,$5F,$F7,$F4,$F2,$E7,$FC
	.dc.b	$A3,$D2,$AD,$7C,$79,$FC,$03,$CE,$2E,$3F,$3E,$FF,$9C,$CF,$82,$31
	.dc.b	$3E,$F8,$FB,$95,$ED,$9F,$AD,$37,$C7,$FC,$16,$1F,$BF,$BF,$F2,$DD
	.dc.b	$7E,$17,$3F,$BB,$1F,$CF,$0F,$E7,$73,$B7,$FD,$FC,$7E,$F0,$FF,$7F
	.dc.b	$1F,$E5,$93,$07,$F9,$65,$E0,$FF,$3A,$3F,$76,$7F,$3A,$7A,$73,$FA
	.dc.b	$6B,$F9,$EC,$F8,$A3,$D4,$FF,$3A,$9F,$E2,$D1,$EA,$8F,$E4,$7E,$57
	.dc.b	$FB,$F7,$9B,$F3,$AF,$FB,$FF,$9A,$C0,$FE,$79,$29,$C4,$43,$53,$F7
	.dc.b	$23,$11,$FC,$FE,$78,$38,$9F,$E3,$13,$A8,$FB,$73,$A7,$FA,$05,$FF
	.dc.b	$31,$9F,$E8,$93,$E4,$9E,$EF,$08,$F1,$3F,$82,$7F,$A4,$3E,$38,$FE
	.dc.b	$84,$FC,$07,$FA,$83,$C5,$F8,$8F,$B6,$AC,$FF,$01,$8F,$8B,$E3,$7F
	.dc.b	$99,$60,$FF,$46,$8F,$96,$FF,$98,$38,$BF,$96,$61,$E9,$8C,$72,$7C
	.dc.b	$57,$FA,$25,$F2,$AF,$E9,$17,$97,$F9,$EB,$E0,$47,$F3,$9B,$E8,$E3
	.dc.b	$9B,$E0,$5E,$A3,$FC,$E1,$5F,$3A,$34,$EA,$32,$E6,$7F,$79,$7E,$CC
	.dc.b	$2F,$8F,$F0,$B0,$F6,$61,$E2,$C7,$C2,$FF,$42,$C7,$E0,$BF,$D0,$9A
	.dc.b	$FF,$42,$C3,$DC,$F9,$4F,$21,$FD,$0A,$3E,$03,$FD,$19,$AB,$97,$EF
	.dc.b	$5F,$19,$B3,$FE,$6C,$AE,$B1,$ED,$AF,$DE,$47,$CD,$8B,$5C,$DA,$7C
	.dc.b	$0F,$1F,$F9,$47,$24,$2E,$3F,$91,$8F,$E7,$D3,$FA,$17,$F9,$F3,$67
	.dc.b	$F9,$F4,$ED,$E4,$7E,$7A,$7F,$9E,$CE,$E3,$F9,$EB,$FB,$08,$FE,$78
	.dc.b	$FE,$A5,$FE,$78,$D8,$FE,$79,$CB,$F9,$D9,$E3,$39,$8F,$7C,$7F,$3E
	.dc.b	$8E,$51,$ED,$7E,$ED,$7E,$72,$7F,$CD,$F1,$E3,$AF,$81,$C9,$E3,$7D
	.dc.b	$3C,$BE,$17,$FC,$DF,$7E,$D7,$F9,$BE,$4C,$3D,$F7,$EF,$9F,$97,$7F
	.dc.b	$B5,$9F,$F3,$86,$1D,$67,$F9,$9E,$1C,$EF,$D9,$1F,$CD,$A7,$C5,$1E
	.dc.b	$CB,$FA,$CF,$E5,$BC,$7F,$DA,$F8,$FF,$B6,$E0,$B7,$FE,$70,$3F,$8E
	.dc.b	$C7,$F9,$C1,$FE,$30,$FF,$80,$B9,$3F,$8B,$3F,$CD,$A3,$67,$F3,$63
	.dc.b	$67,$F9,$C0,$C9,$FE,$4A,$8F,$E2,$11,$FE,$70,$CB,$F8,$B3,$FE,$74
	.dc.b	$9F,$E3,$AF,$F3,$77,$AC,$FE,$52,$3F,$CE,$11,$E2,$8F,$D8,$C7,$F3
	.dc.b	$78,$FE,$6B,$E3,$FF,$12,$F1,$FF,$CE,$11,$FC,$C5,$FF,$38,$37,$9B
	.dc.b	$F9,$AE,$E9,$FE,$59,$CD,$FC,$CD,$FE,$5D,$56,$BF,$9C,$4F,$F3,$29
	.dc.b	$FF,$38,$65,$FC,$BD,$CD,$66,$BF,$9F,$BF,$CB,$E3,$F9,$6C,$EE,$7F
	.dc.b	$A3,$4F,$F3,$19,$FE,$6A,$EE,$3F,$86,$C7,$F3,$08,$F0,$1F,$C4,$39
	.dc.b	$3F,$96,$1F,$CA,$DF,$E9,$CF,$F2,$BE,$6F,$E9,$8F,$F2,$A8,$FE,$AE
	.dc.b	$FF,$2A,$3F,$CD,$8E,$9F,$E5,$37,$FD,$25,$FE,$5B,$19,$7F,$CC,$0F
	.dc.b	$F5,$27,$F9,$46,$BF,$D2,$5F,$E4,$F1,$FD,$35,$FF,$34,$3E,$2A,$F4
	.dc.b	$4F,$F2,$87,$FA,$3C,$FF,$26,$3C,$A7,$F4,$3E,$5A,$FF,$21,$3C,$47
	.dc.b	$F3,$76,$BF,$CF,$AA,$C7,$96,$3F,$CC,$B1,$F1,$1F,$F3,$2B,$FD,$E9
	.dc.b	$FE,$48,$ED,$FE,$3C,$D3,$D8,$E1,$E8,$7F,$26,$70,$FC,$33,$FA,$1C
	.dc.b	$7F,$62,$F1,$FF,$8D,$F1,$FF,$24,$9F,$F0,$4C,$7F,$24,$D4,$FF,$24
	.dc.b	$6C,$FF,$24,$9D,$BF,$C9,$1F,$F0,$34,$FF,$24,$9D,$CF,$F2,$4C,$BF
	.dc.b	$C0,$13,$FC,$92,$BF,$C2,$AF,$F2,$46,$CF,$F2,$47,$FC,$05,$1F,$C9
	.dc.b	$1C,$F2,$FF,$42,$8F,$E4,$91,$CA,$3F,$AF,$C7,$F7,$EF,$1B,$DD,$5F
	.dc.b	$E5,$BA,$F3,$BF,$E5,$B6,$BF,$E5,$B6,$BF,$E5,$B7,$C6,$FF,$24,$9F
	.dc.b	$E4,$71,$EE,$8F,$E4,$6F,$F8,$82,$3F,$91,$DF,$BA,$7F,$CF,$1C,$BF
	.dc.b	$E1,$99,$FE,$A7,$CB,$67,$F9,$25,$6E,$3D,$95,$EA,$7F,$BC,$F8,$FE
	.dc.b	$BC,$3F,$91,$1F,$E3,$7C,$7F,$91,$66,$7F,$91,$47,$F9,$01,$FE,$44
	.dc.b	$EC,$FE,$45,$1B,$7F,$91,$19,$3F,$91,$1F,$E3,$A8,$FE,$44,$FF,$93
	.dc.b	$1F,$E4,$4F,$F9,$0D,$FE,$44,$F5,$E5,$FE,$EB,$1F,$C8,$A3,$C5,$1F
	.dc.b	$E1,$28,$FE,$45,$1F,$E4,$FF,$1F,$FC,$67,$E3,$FF,$22,$B3,$FC,$89
	.dc.b	$BC,$9F,$C8,$A3,$F9,$0D,$7F,$97,$1F,$E4,$2D,$8F,$E3,$AD,$7F,$8E
	.dc.b	$BF,$E5,$98,$CB,$98,$FF,$26,$47,$F2,$08,$FF,$34,$BC,$1F,$F2,$DB
	.dc.b	$FC,$81,$B1,$FE,$2C,$8F,$E4,$11,$D4,$FF,$17,$E5,$FE,$4E,$F1,$BF
	.dc.b	$C9,$87,$F2,$1B,$3F,$C8,$67,$FC,$EA,$FF,$21,$D7,$FC,$C8,$FF,$92
	.dc.b	$8F,$E4,$31,$AE,$2F,$E4,$2F,$F9,$FA,$3F,$90,$B9,$8F,$E4,$2F,$F9
	.dc.b	$DE,$3F,$90,$BF,$E6,$B7,$F9,$0B,$FE,$76,$7F,$90,$BE,$2E,$3F,$F2
	.dc.b	$3C,$7F,$92,$A3,$FC,$F7,$E3,$FA,$EB,$FC,$8F,$7E,$67,$F9,$13,$5F
	.dc.b	$E4,$53,$E3,$7F,$91,$35,$FE,$44,$DE,$4F,$14,$7F,$22,$C7,$E2,$BF
	.dc.b	$C8,$A7,$FD,$29,$87,$F2,$27,$D9,$1F,$E4,$58,$FF,$26,$47,$2A,$FF
	.dc.b	$22,$C7,$68,$F8,$11,$EC,$7F,$CC,$7C,$5F,$C8,$0F,$F7,$66,$1F,$E9
	.dc.b	$37,$F9,$01,$FC,$7C,$FE,$4B,$C9,$FE,$A1,$7F,$DF,$B1,$B7,$FC,$7A
	.dc.b	$D7,$F8,$F4,$FF,$AC,$9F,$E3,$D1,$97,$FD,$34,$7F,$95,$26,$BF,$E3
	.dc.b	$D8,$E9,$1F,$E5,$E8,$FE,$3B,$1E,$58,$F4,$CF,$F9,$F1,$FF,$4D,$CF
	.dc.b	$F9,$03,$1F,$62,$D3,$FD,$82,$FF,$3E,$E3,$F5,$3F,$C6,$B0,$F2,$28
	.dc.b	$FF,$5C,$BF,$C7,$B0,$E2,$FF,$1E,$78,$BF,$C9,$5B,$1E,$A8,$F0,$47
	.dc.b	$FA,$E5,$FE,$42,$FB,$AB,$D9,$3F,$49,$FF,$4A,$BF,$EB,$4F,$1B,$FD
	.dc.b	$6B,$E3,$DA,$FE,$3F,$8A,$7F,$8F,$9F,$C7,$B1,$FF,$65,$3F,$C7,$A3
	.dc.b	$53,$FC,$7A,$0F,$F1,$E8,$CB,$FC,$78,$FF,$6E,$47,$F1,$E7,$FD,$C4
	.dc.b	$FF,$1E,$7F,$DD,$2F,$F1,$E8,$EB,$C9,$FE,$92,$8F,$E3,$D5,$E2,$8F
	.dc.b	$F5,$7C,$7F,$1E,$97,$8F,$FE,$D0,$F1,$FF,$8F,$1F,$C7,$4F,$F7,$37
	.dc.b	$27,$FB,$B5,$FE,$3B,$87,$FB,$95,$FE,$3A,$D8,$FE,$3A,$D7,$F8,$B3
	.dc.b	$5F,$F7,$0B,$71,$FF,$75,$4F,$F1,$D7,$FD,$FC,$FF,$1D,$7C,$13,$FC
	.dc.b	$6A,$7B,$47,$FA,$BA,$7F,$8E,$A8,$F2,$3F,$EC,$CF,$1F,$FD,$D0,$7F
	.dc.b	$1D,$3F,$8E,$47,$FB,$F5,$FE,$39,$1F,$4B,$04,$7F,$B9,$8F,$E3,$B3
	.dc.b	$BE,$2F,$E3,$B4,$AD,$C7,$F1,$A8,$FB,$73,$B8,$FE,$3F,$1E,$53,$B5
	.dc.b	$6D,$D9,$88,$DB,$B3,$B4,$7F,$22,$9F,$47,$27,$B7,$AB,$FD,$D3,$E3
	.dc.b	$FE,$2A,$FF,$DA,$37,$FE,$43,$BF,$16,$1F,$EE,$1E,$4E,$99,$7A,$23
	.dc.b	$F9,$1C,$E9,$FF,$7F,$CF,$92,$FC,$31,$FC,$A1,$E3,$7F,$E4,$B9,$F0
	.dc.b	$57,$19,$A7,$AE,$B9,$B8,$39,$BF,$C8,$5E,$6F,$FB,$AE,$BF,$93,$C7
	.dc.b	$86,$FE,$6D,$FB,$2B,$EE,$3F,$76,$7E,$01,$2F,$E1,$53,$C5,$CC,$E4
	.dc.b	$FB,$DE,$3F,$F1,$2A,$DB,$FE,$4B,$3F,$C9,$47,$F2,$C8,$F8,$EF,$F2
	.dc.b	$C7,$8F,$2F,$F2,$C6,$BF,$CB,$1F,$C2,$47,$F2,$C7,$31,$E2,$9E,$EF
	.dc.b	$F9,$26,$3E,$25,$AB,$FC,$95,$5A,$E6,$F4,$D7,$8A,$EB,$E4,$8B,$3E
	.dc.b	$A8,$A7,$82,$7F,$DC,$71,$E0,$7F,$C9,$91,$E6,$7F,$DD,$EF,$57,$FD
	.dc.b	$DE,$F1,$E5,$F2,$DF,$F2,$FA,$E5,$5F,$CC,$EB,$2F,$F9,$21,$AF,$FB
	.dc.b	$B9,$F8,$BC,$DF,$05,$F9,$37,$EC,$AF,$87,$36,$7E,$AC,$FA,$9F,$97
	.dc.b	$17,$17,$8F,$F0,$A7,$F3,$67,$C4,$E6,$AF,$C8,$D2,$BF,$86,$8F,$C6
	.dc.b	$3F,$CC,$CF,$C2,$BF,$49,$FC,$3C,$57,$FC,$97,$8F,$E3,$1F,$E6,$57
	.dc.b	$CD,$FC,$61,$FC,$0C,$FE,$64,$7F,$39,$39,$4F,$B5,$D5,$7A,$DF,$E6
	.dc.b	$4F,$AE,$BF,$75,$E5,$FF,$30,$E5,$1F,$C0,$DC,$78,$36,$3F,$9A,$DD
	.dc.b	$76,$48,$FD,$D4,$EA,$75,$39,$72,$FF,$BC,$9F,$EA,$71,$CA,$3F,$C9
	.dc.b	$95,$E4,$71,$3F,$02,$7F,$CA,$EF,$43,$EF,$07,$B3,$B8,$FF,$7B,$BA
	.dc.b	$8F,$F2,$C3,$FC,$15,$CB,$9A,$F8,$8F,$90,$A7,$5F,$1B,$FD,$B9,$C4
	.dc.b	$1C,$DF,$1E,$3F,$DF,$96,$7F,$A0,$D7,$B9,$FE,$82,$F3,$AE,$47,$FB
	.dc.b	$CB,$1F,$A7,$3F,$D0,$A7,$84,$FF,$42,$AE,$8F,$F3,$BA,$DB,$FE,$5A
	.dc.b	$9F,$13,$FE,$FA,$8C,$DF,$E0,$EB,$8F,$35,$C3,$FA,$0E,$1E,$98,$FD
	.dc.b	$14,$BC,$7F,$82,$F8,$31,$F8,$38,$F8,$1F,$E6,$8D,$7D,$67,$E0,$B0
	.dc.b	$F8,$FC,$B6,$3F,$9E,$4E,$DF,$F7,$CB,$71,$FA,$95,$FD,$25,$B3,$FC
	.dc.b	$DA,$7C,$15,$F6,$EB,$F9,$CD,$7C,$08,$F6,$57,$AA,$7F,$DF,$F3,$E3
	.dc.b	$3D,$BE,$3F,$F3,$37,$F2,$2F,$F9,$87,$9B,$F4,$0F,$F9,$8E,$2F,$2F
	.dc.b	$FB,$F5,$FC,$74,$7F,$35,$6C,$7F,$35,$6B,$FC,$D7,$24,$FF,$35,$9C
	.dc.b	$CF,$E4,$E3,$FC,$C1,$1F,$9D,$7F,$99,$BE,$48,$57,$D9,$9F,$F3,$0C
	.dc.b	$FA,$23,$D6,$FD,$8C,$BE,$BC,$FF,$9E,$F9,$7D,$37,$FC,$CB,$0F,$1B
	.dc.b	$FC,$CA,$BF,$80,$D2,$3F,$CC,$57,$E1,$7F,$CB,$D3,$A7,$47,$F9,$7D
	.dc.b	$C3,$FE,$5C,$8F,$E0,$33,$FF,$00,$1E,$57,$EF,$9F,$F3,$34,$69,$FF
	.dc.b	$7F,$BE,$FE,$5F,$74,$FD,$67,$F2,$7E,$3F,$E4,$BC,$7F,$C9,$F0,$5B
	.dc.b	$FF,$80,$39,$A7,$FA,$BF,$2F,$F1,$97,$FC,$C9,$3F,$C3,$9F,$F8,$03
	.dc.b	$3B,$8F,$E5,$86,$CF,$F8,$02,$E5,$FE,$59,$A7,$F2,$69,$FF,$32,$4E
	.dc.b	$79,$7F,$86,$47,$FC,$00,$8E,$91,$FA,$F8,$FF,$31,$D7,$86,$3F,$5B
	.dc.b	$1F,$F0,$02,$5E,$3F,$EB,$BC,$7F,$E6,$58,$7F,$34,$7F,$CC,$7C,$DF
	.dc.b	$CC,$DF,$E6,$31,$9E,$7F,$E5,$02,$7F,$98,$CF,$F9,$9B,$2F,$E5,$AE
	.dc.b	$A3,$F9,$03,$FC,$C6,$3F,$94,$3F,$CC,$63,$F9,$73,$FC,$C2,$37,$3F
	.dc.b	$CA,$2B,$C0,$FF,$0B,$9F,$E5,$73,$E5,$8F,$E1,$31,$FE,$66,$9F,$54
	.dc.b	$FF,$0C,$E4,$FE,$52,$7F,$28,$8F,$EA,$2F,$F2,$83,$F9,$3C,$7F,$3B
	.dc.b	$7F,$93,$E3,$FC,$C3,$9F,$F9,$C4,$7F,$25,$9D,$3F,$C9,$4F,$F8,$03
	.dc.b	$5D,$1F,$E4,$8D,$7F,$DF,$4E,$67,$F9,$24,$7F,$97,$E7,$FA,$9B,$FC
	.dc.b	$92,$7C,$BB,$7F,$24,$AF,$E4,$73,$FE,$5F,$AD,$BF,$F0,$01,$AF,$F9
	.dc.b	$79,$AF,$FC,$00,$AF,$13,$FD,$02,$3F,$B3,$BF,$E5,$D6,$BF,$EF,$D8
	.dc.b	$DF,$37,$F6,$18,$FF,$80,$13,$FD,$D5,$FE,$93,$1F,$D7,$5F,$E4,$73
	.dc.b	$F0,$63,$DB,$1F,$45,$FA,$53,$FC,$F9,$FE,$7B,$E3,$FF,$23,$8F,$ED
	.dc.b	$4F,$F2,$3A,$FF,$02,$BF,$C8,$DB,$C9,$FC,$8F,$24,$FF,$23,$6C,$7F
	.dc.b	$23,$6B,$FC,$8E,$3F,$91,$46,$23,$FB,$A4,$7F,$22,$AF,$EE,$8F,$F2
	.dc.b	$27,$C7,$1F,$C8,$9D,$BF,$D5,$A3,$11,$C8,$FE,$B3,$E3,$7B,$27,$FD
	.dc.b	$FD,$7E,$87,$F9,$15,$7F,$C0,$1A,$FF,$0F,$3F,$C7,$23,$FC,$99,$3B
	.dc.b	$7F,$8E,$35,$FF,$7D,$CF,$07,$F9,$1F,$2F,$F8,$56,$BF,$90,$4F,$A2
	.dc.b	$7F,$CA,$8D,$E5,$F4,$D7,$F7,$3A,$FE,$47,$3F,$E2,$87,$FC,$BC,$FC
	.dc.b	$49,$FE,$E7,$E3,$FF,$72,$C7,$FC,$A1,$8F,$F8,$E1,$FE,$3E,$DA,$FF
	.dc.b	$26,$9F,$E3,$98,$FE,$3E,$61,$C1,$FC,$79,$FE,$42,$E0,$1F,$E9,$B3
	.dc.b	$FC,$4C,$FF,$C0,$18,$ED,$19,$32,$39,$3F,$C1,$D3,$FC,$82,$3F,$C6
	.dc.b	$8E,$5C,$F2,$FF,$83,$BC,$7F,$F0,$6F,$8F,$FC,$82,$7F,$C8,$6F,$F2
	.dc.b	$0C,$3F,$CA,$AF,$F2,$08,$CF,$27,$F1,$B7,$FC,$AD,$1F,$C7,$DF,$F2
	.dc.b	$13,$97,$31,$FC,$6D,$AF,$F9,$39,$E6,$FF,$90,$9C,$F2,$FF,$91,$E7
	.dc.b	$FC,$BB,$16,$7F,$97,$34,$FF,$16,$47,$F1,$F7,$D2,$7F,$8A,$B4,$FF
	.dc.b	$14,$F8,$DF,$C8,$39,$3F,$CD,$8F,$F9,$18,$FE,$3F,$C4,$7F,$8F,$9F
	.dc.b	$E2,$AE,$4F,$F2,$6D,$FF,$9B,$23,$F9,$04,$EE,$7F,$C8,$91,$FC,$82
	.dc.b	$7F,$CD,$6F,$F9,$32,$BF,$CC,$0F,$FC,$01,$8F,$0C,$FF,$C0,$0C,$FF
	.dc.b	$D0,$33,$FC,$86,$7F,$CD,$4F,$F2,$17,$9F,$2F,$F9,$13,$C7,$F5,$DF
	.dc.b	$F3,$AB,$E2,$FF,$21,$AE,$2F,$F9,$99,$EE,$FF,$8E,$E3,$6F,$F9,$1A
	.dc.b	$3D,$8F,$F8,$FD,$DB,$FE,$6D,$7F,$CE,$D7,$97,$27,$FA,$0D,$FE,$42
	.dc.b	$D8,$FF,$7F,$3E,$EE,$6F,$F3,$04,$E2,$3E,$8C,$7F,$97,$A3,$FC,$BB
	.dc.b	$CD,$FC,$87,$3F,$E0,$4F,$F0,$3E,$CF,$F5,$13,$FC,$86,$FF,$D2,$EF
	.dc.b	$FB,$DF,$1F,$F5,$4B,$FC,$84,$E3,$3F,$E8,$D7,$9C,$D7,$F8,$F3,$D1
	.dc.b	$FF,$15,$BF,$E9,$D7,$2E,$5F,$F3,$C1,$FE,$F9,$79,$18,$8F,$F3,$BC
	.dc.b	$7F,$9D,$9F,$5B,$FC,$82,$3C,$EF,$B0,$6B,$D5,$FE,$33,$15,$FF,$21
	.dc.b	$47,$FA,$E9,$FF,$7B,$C6,$63,$D5,$1F,$E3,$F8,$E5,$1E,$98,$C4,$7A
	.dc.b	$63,$FD,$61,$18,$8F,$81,$1E,$CE,$7F,$4B,$FE,$90,$7F,$D1,$FE,$37
	.dc.b	$FA,$43,$C7,$B5,$FE,$3E,$8F,$E3,$EF,$FB,$21,$FF,$1E,$BF,$EC,$17
	.dc.b	$FC,$7A,$7F,$1F,$C7,$FD,$BA,$FF,$BB,$A3,$FD,$AA,$FF,$20,$8C,$C7
	.dc.b	$F2,$03,$F8,$E1,$FE,$E1,$8F,$E4,$0F,$73,$F9,$2F,$2F,$FB,$2E,$3F
	.dc.b	$DE,$11,$D0,$E2,$76,$E7,$FF,$55,$C7,$F8,$FF,$C7,$FF,$55,$F8,$FF
	.dc.b	$C7,$EB,$FD,$B0,$FF,$1F,$7F,$DC,$6F,$F1,$F7,$3C,$DF,$C7,$F1,$FF
	.dc.b	$7B,$CF,$F1,$FD,$CF,$F1,$F8,$CC,$FF,$1F,$6B,$FC,$7E,$B9,$BF,$C7
	.dc.b	$E3,$3C,$FF,$EE,$5A,$FE,$3F,$5F,$EE,$F7,$FC,$7B,$3D,$23,$FD,$95
	.dc.b	$1F,$C7,$DF,$49,$FE,$C8,$F1,$BF,$8F,$4F,$CB,$C1,$1F,$EE,$43,$F8
	.dc.b	$F9,$FE,$3C,$3F,$8F,$73,$FF,$BC,$DF,$E3,$A7,$FB,$29,$FF,$80,$0F
	.dc.b	$F1,$DA,$FF,$7F,$3F,$C7,$5C,$F2,$FF,$B7,$DF,$EB,$35,$E0,$8F,$E3
	.dc.b	$D1,$89,$FE,$3D,$3E,$4A,$F4,$73,$FF,$B8,$3C,$7F,$F6,$FD,$7F,$8F
	.dc.b	$6F,$BE,$1F,$C7,$70,$E8,$FF,$8F,$63,$2F,$F1,$D7,$B3,$FE,$DE,$6B
	.dc.b	$FC,$59,$63,$F1,$6B,$F9,$F4,$7C,$19,$D3,$A3,$8B,$F9,$57,$FB,$AC
	.dc.b	$F1,$80,$E7,$D4,$F1,$3C,$A6,$27,$84,$6A,$35,$7F,$1D,$C4,$F9,$67
	.dc.b	$EE,$46,$CD,$F0,$D7,$8F,$8F,$FC,$A4,$FF,$24,$3F,$49,$FE,$50,$AD
	.dc.b	$1F,$CA,$63,$5C,$DF,$C9,$9E,$6F,$F9,$26,$31,$18,$8F,$E4,$58,$7E
	.dc.b	$36,$7F,$C7,$13,$F1,$63,$13,$FC,$9E,$38,$1F,$04,$EF,$CF,$EE,$7C
	.dc.b	$25,$3C,$D5,$FE,$4C,$7A,$1E,$B3,$E1,$57,$6A,$FE,$8F,$CD,$E5,$7F
	.dc.b	$C9,$5C,$F5,$FF,$76,$47,$D8,$AF,$7C,$69,$D4,$7F,$11,$AE,$93,$FE
	.dc.b	$ED,$9D,$BF,$E4,$47,$B5,$79,$1F,$3C,$7F,$30,$9F,$24,$FB,$E7,$35
	.dc.b	$ED,$C7,$E8,$DF,$F9,$6A,$7C,$07,$D7,$9F,$6B,$F6,$6A,$BF,$67,$C7
	.dc.b	$FC,$06,$1F,$BD,$8F,$F2,$99,$FB,$03,$F9,$A9,$F7,$CF,$E6,$B8,$7F
	.dc.b	$BD,$1F,$C7,$9F,$CD,$4E,$07,$FB,$C0,$FE,$68,$60,$FE,$5E,$79,$8E
	.dc.b	$87,$F3,$08,$F1,$3A,$35,$1F,$CE,$F9,$FE,$04,$7F,$94,$E3,$F9,$BB
	.dc.b	$F4,$CF,$36,$9E,$E8,$FF,$2B,$47,$E0,$FC,$BF,$F7,$6F,$30,$87,$EB
	.dc.b	$5C,$B9,$E6,$C3,$FC,$E6,$3F,$DF,$33,$88,$FF,$7B,$BE,$18,$EB,$3E
	.dc.b	$D7,$BC,$59,$FF,$7A,$35,$FE,$6A,$F1,$7C,$73,$F7,$E7,$FC,$BD,$39
	.dc.b	$9F,$F2,$BB,$F8,$78,$E2,$F5,$9F,$BB,$1F,$E6,$18,$F1,$47,$DE,$7F
	.dc.b	$D0,$71,$FB,$F7,$E0,$78,$DF,$CF,$9C,$BF,$EF,$CE,$5C,$F1,$FF,$3C
	.dc.b	$9A,$FF,$16,$3F,$8B,$C7,$C1,$9D,$4E,$A7,$2E,$5F,$E7,$CD,$72,$DE
	.dc.b	$6F,$81,$3F,$E5,$DD,$7F,$BD,$60,$8F,$E4,$B3,$F6,$DF,$F2,$F4,$66
	.dc.b	$3E,$0C,$30,$F8,$7E,$3F,$C1,$C1,$7C,$F1,$FF,$30,$47,$C0,$7F,$CC
	.dc.b	$11,$97,$FA,$23,$5F,$F8,$01,$1B,$7F,$CC,$0F,$BD,$FE,$70,$FB,$5F
	.dc.b	$F3,$03,$79,$7E,$C6,$1F,$F0,$03,$2F,$6B,$FD,$1A,$7E,$69,$F0,$AB
	.dc.b	$9B,$C1,$F6,$D7,$BA,$3E,$FF,$8F,$FE,$FF,$C0,$FF,$38,$8F,$7B,$FC
	.dc.b	$E2,$33,$C9,$FE,$5C,$9F,$D0,$4F,$F9,$2D,$FC,$D3,$97,$31,$FC,$D1
	.dc.b	$CB,$FE,$60,$C7,$EA,$CE,$67,$33,$FC,$BD,$E2,$FE,$C6,$3F,$9D,$47
	.dc.b	$43,$1C,$DE,$A7,$F3,$35,$F5,$EF,$CF,$7F,$CD,$23,$BB,$FE,$61,$E4
	.dc.b	$EA,$FF,$98,$23,$2F,$F3,$36,$BF,$F0,$02,$BF,$43,$87,$FA,$12,$BF
	.dc.b	$6C,$A7,$FC,$D3,$3E,$A7,$F9,$65,$74,$E5,$FD,$A5,$6E,$7F,$98,$57
	.dc.b	$C6,$9F,$74,$62,$3E,$6C,$7D,$6B,$FD,$AF,$8F,$ED,$F1,$FD,$DC,$16
	.dc.b	$FF,$BD,$73,$7F,$13,$9F,$E6,$3A,$1F,$F3,$2B,$67,$FC,$C5,$A7,$E9
	.dc.b	$A7,$F9,$94,$7E,$79,$FF,$33,$3C,$CF,$E5,$47,$F1,$57,$FC,$25,$CD
	.dc.b	$FC,$2D,$FF,$32,$C7,$13,$F9,$91,$FC,$75,$FE,$5F,$3D,$E3,$FC,$CA
	.dc.b	$79,$EF,$F8,$8C,$E2,$3F,$96,$78,$FF,$6F,$C7,$FF,$33,$62,$7F,$CC
	.dc.b	$D3,$FB,$47,$F9,$71,$FC,$A9,$FE,$68,$FF,$2A,$6F,$3F,$F2,$A6,$BF
	.dc.b	$CA,$B1,$FE,$0F,$39,$9C,$CF,$F2,$B8,$FE,$5B,$1F,$CA,$5F,$E5,$2F
	.dc.b	$4E,$7F,$E8,$D1,$FC,$A6,$3C,$31,$FC,$B2,$3F,$29,$C9,$FC,$9E,$7F
	.dc.b	$85,$BF,$C9,$EF,$F8,$BB,$FE,$FF,$3F,$DF,$EE,$1F,$E5,$07,$F2,$47
	.dc.b	$0F,$F2,$46,$BF,$C9,$1B,$CF,$FC,$FE,$3F,$CD,$18,$FF,$9A,$EB,$11
	.dc.b	$FE,$67,$7C,$6F,$F9,$99,$BB,$7F,$61,$85,$7F,$46,$7F,$93,$4F,$F6
	.dc.b	$58,$F6,$C7,$F2,$48,$E6,$FA,$0F,$F2,$A5,$F9,$9F,$E8,$51,$FD,$A5
	.dc.b	$F8,$1C,$BF,$D8,$DF,$85,$15,$F8,$31,$FD,$95,$FF,$33,$C6,$5F,$F3
	.dc.b	$0C,$7B,$1F,$E5,$31,$98,$FE,$C9,$1E,$08,$F8,$31,$ED,$8F,$63,$ED
	.dc.b	$8F,$EC,$6F,$F0,$BF,$1F,$F9,$31,$FC,$97,$13,$FC,$94,$FF,$2D,$4F
	.dc.b	$F4,$67,$F9,$23,$FE,$05,$7F,$92,$5F,$F5,$47,$FC,$B6,$DE,$4F,$E4
	.dc.b	$93,$FD,$BA,$7F,$92,$3C,$63,$FC,$B5,$3B,$7F,$C2,$93,$FC,$92,$39
	.dc.b	$4E,$39,$BF,$BD,$CF,$F2,$3F,$1B,$D5,$7F,$F0,$07,$9B,$A3,$FC,$6F
	.dc.b	$0E,$8F,$F2,$3C,$3C,$4F,$F1,$F7,$6F,$F2,$37,$DC,$FF,$96,$5C,$CF
	.dc.b	$F8,$36,$F1,$1D,$A3,$FC,$29,$1F,$C8,$79,$AD,$7F,$96,$EB,$FC,$45
	.dc.b	$3F,$CA,$E7,$DB,$5E,$A7,$FC,$2B,$E3,$FA,$EB,$F8,$FE,$1F,$DF,$B1
	.dc.b	$FE,$3E,$7F,$1E,$3F,$D3,$9C,$DF,$E2,$D7,$F8,$EE,$87,$F8,$EC,$F1
	.dc.b	$AF,$E3,$A6,$C1,$FC,$69,$D3,$FE,$44,$34,$7F,$BF,$9E,$5C,$DF,$E4
	.dc.b	$79,$FE,$43,$3E,$18,$FF,$16,$C7,$F1,$D9,$FF,$0A,$F8,$FF,$D5,$FC
	.dc.b	$7F,$E3,$A7,$F1,$CC,$0F,$F9,$24,$FE,$39,$3F,$E1,$C7,$F8,$E3,$79
	.dc.b	$BF,$8E,$35,$FE,$39,$8F,$F8,$E2,$7F,$8E,$4E,$67,$FC,$8D,$1F,$C7
	.dc.b	$27,$FC,$A0,$FF,$1C,$9E,$9C,$DF,$E6,$18,$FE,$39,$1E,$18,$FF,$23
	.dc.b	$C7,$F6,$FD,$7F,$B8,$78,$DF,$C7,$71,$3F,$C7,$70,$FF,$24,$BF,$C7
	.dc.b	$79,$7F,$CC,$0F,$F1,$FE,$6B,$CD,$FC,$56,$BF,$CD,$93,$FC,$26,$7F
	.dc.b	$CC,$8E,$63,$33,$FE,$4B,$9F,$F2,$E4,$FF,$91,$A7,$9C,$FF,$16,$78
	.dc.b	$BF,$C7,$6F,$CA,$E6,$3A,$DF,$F9,$8A,$B1,$1F,$E7,$3F,$1F,$DB,$5F
	.dc.b	$C7,$6F,$8B,$FC,$77,$9F,$C6,$FF,$90,$2F,$F8,$DD,$E2,$3F,$88,$6B
	.dc.b	$FE,$54,$9E,$13,$E6,$9F,$E2,$F3,$97,$FE,$00,$B5,$FE,$37,$1D,$AB
	.dc.b	$FC,$FD,$5F,$C6,$E3,$DB,$1F,$E9,$78,$DC,$7C,$E8,$F6,$3F,$E7,$EE
	.dc.b	$6F,$F2,$01,$FC,$6C,$FF,$1F,$1F,$C6,$B0,$FF,$0B,$3F,$E3,$9C,$3F
	.dc.b	$CD,$6F,$F1,$A9,$FF,$41,$BF,$C6,$9A,$FF,$1A,$7F,$CC,$EF,$F1,$A8
	.dc.b	$E2,$FF,$1A,$8D,$BF,$EA,$D3,$F8,$D3,$C8,$C7,$3F,$9D,$FF,$4A,$BF
	.dc.b	$E3,$D8,$F2,$9F,$E2,$BE,$6F,$F5,$93,$83,$F8,$55,$78,$1F,$5C,$7F
	.dc.b	$AB,$DF,$F2,$15,$70,$72,$E5,$FE,$39,$3C,$5F,$F7,$83,$76,$FF,$55
	.dc.b	$47,$F9,$02,$78,$3F,$E3,$C9,$F7,$D7,$B6,$3F,$C7,$91,$F2,$A3,$E9
	.dc.b	$57,$FA,$C1,$FF,$2C,$78,$DF,$E5,$9F,$1E,$E1,$FC,$78,$FF,$58,$C7
	.dc.b	$F8,$F0,$FF,$1D,$9F,$C7,$9F,$F6,$93,$FC,$78,$FE,$3A,$F0,$3F,$8E
	.dc.b	$9A,$3F,$8E,$8E,$7F,$F6,$24,$7F,$8E,$B3,$3A,$8F,$F6,$63,$FE,$3A
	.dc.b	$9E,$91,$FE,$E5,$8F,$E4,$91,$FE,$D9,$7F,$8F,$3F,$EC,$8F,$1F,$FC
	.dc.b	$E9,$E3,$FF,$8E,$8F,$E3,$C7,$FB,$94,$FE,$3B,$1F,$EC,$E7,$F8,$EB
	.dc.b	$79,$BF,$8E,$B5,$FE,$3A,$FF,$AD,$23,$F8,$EB,$98,$FF,$71,$47,$F1
	.dc.b	$D8,$FF,$77,$3F,$C7,$5E,$9C,$DF,$EF,$D8,$FE,$3B,$1E,$18,$FF,$71
	.dc.b	$C7,$FA,$2F,$C6,$FE,$38,$7F,$1B,$9F,$F6,$1B,$FC,$6D,$6D,$FE,$D0
	.dc.b	$8F,$E3,$91,$83,$FD,$D2,$FF,$B7,$A3,$5C,$7F,$E3,$CE,$6F,$F8,$01
	.dc.b	$3F,$C7,$67,$33,$FE,$DE,$AE,$47,$81,$FF,$6D,$BF,$EF,$B9,$FF,$70
	.dc.b	$3D,$A3,$F8,$F4,$7F,$B7,$8F,$F1,$CC,$7E,$00,$EF,$5F,$F0,$03,$C7
	.dc.b	$F7,$57,$F1,$DA,$FF,$29,$3F,$E3,$8A,$E8,$FF,$B7,$23,$F8,$ED,$F9
	.dc.b	$A3,$F8,$EC,$E9,$FF,$19,$BF,$DA,$A3,$51,$E7,$3F,$8E,$C6,$CC,$3F
	.dc.b	$ED,$87,$81,$E0,$3D,$07,$C2,$D9,$5D,$A3,$A1,$A3,$4F,$37,$E6,$99
	.dc.b	$39,$6D,$FE,$C7,$F1,$FF,$DC,$47,$F8,$F4,$FF,$70,$1F,$C8,$CF,$F1
	.dc.b	$DE,$1F,$A8,$B4,$7F,$23,$A4,$63,$8F,$FD,$B9,$5B,$AF,$E5,$7C,$9F
	.dc.b	$46,$31,$3F,$EC,$3B,$C1,$FE,$10,$BE,$51,$FE,$E3,$9D,$BE,$D7,$8D
	.dc.b	$5A,$E1,$7F,$ED,$37,$07,$91,$FF,$76,$3F,$30,$C1,$F0,$E7,$AD,$7F
	.dc.b	$26,$6B,$FE,$37,$6A,$E3,$E5,$87,$9A,$3F,$C7,$95,$F5,$6F,$F9,$AD
	.dc.b	$7F,$0E,$BE,$7C,$96,$BC,$F7,$FC,$A6,$7C,$33,$F0,$A7,$D1,$3F,$EE
	.dc.b	$DA,$FC,$0D,$7C,$2A,$B5,$CA,$7D,$F8,$7D,$19,$B8,$7F,$93,$6B,$CE
	.dc.b	$FF,$BB,$5F,$C5,$78,$FF,$83,$7F,$18,$7E,$17,$B6,$BF,$E4,$AC,$7F
	.dc.b	$00,$FF,$BA,$8F,$E6,$0E,$1F,$D6,$47,$B5,$FC,$19,$FC,$C8,$A7,$F3
	.dc.b	$28,$E0,$FF,$93,$4F,$C0,$1F,$CC,$4F,$E6,$07,$C3,$C3,$F9,$CB,$F4
	.dc.b	$0F,$F2,$69,$E8,$9F,$B4,$F2,$D2,$BF,$94,$8F,$75,$7F,$40,$5C,$70
	.dc.b	$85,$8D,$3F,$EE,$E8,$CB,$A9,$FE,$27,$18,$8E,$7C,$7F,$8D,$8C,$47
	.dc.b	$C3,$7F,$DE,$93,$C2,$39,$B9,$7F,$9C,$A8,$CB,$98,$EA,$FE,$16,$3C
	.dc.b	$4D,$8C,$B5,$F8,$31,$88,$D1,$89,$F8,$6F,$F9,$FA,$3F,$04,$FB,$3C
	.dc.b	$5F,$F7,$B4,$1F,$E8,$0D,$E1,$FF,$6E,$75,$FB,$E7,$F9,$FE,$3F,$35
	.dc.b	$FE,$7D,$7D,$5F,$F2,$D3,$BA,$FF,$2B,$4E,$DF,$F2,$D4,$7C,$79,$FF
	.dc.b	$2D,$3E,$38,$FF,$7B,$5F,$86,$FF,$7A,$F8,$F9,$2D,$FF,$3E,$BF,$D2
	.dc.b	$9F,$2D,$FF,$2D,$BF,$57,$C6,$F8,$2E,$70,$FF,$7C,$5F,$B5,$FF,$2E
	.dc.b	$63,$E5,$7F,$DF,$53,$E3,$7F,$DE,$6D,$7F,$A0,$BC,$D1,$FC,$D6,$39
	.dc.b	$3F,$E5,$E8,$F8,$8F,$F4,$27,$83,$E8,$9F,$17,$55,$C3,$FC,$BD,$5F
	.dc.b	$26,$3D,$F5,$F7,$A7,$F9,$F4,$FD,$33,$F0,$3E,$37,$F9,$0F,$90,$7E
	.dc.b	$61,$C7,$2D,$8F,$F2,$44,$D7,$F9,$BB,$5F,$E6,$CD,$51,$F9,$D9,$FF
	.dc.b	$31,$4F,$E6,$5E,$2F,$E8,$DF,$F3,$13,$F8,$F7,$C7,$3F,$CF,$63,$0F
	.dc.b	$F3,$97,$F3,$D1,$FE,$64,$7A,$47,$B2,$7D,$B1,$F9,$EC,$BE,$DE,$4B
	.dc.b	$F8,$1C,$BE,$7C,$3F,$CC,$77,$D9,$FF,$2F,$DF,$F0,$3A,$FF,$31,$B6
	.dc.b	$3F,$CC,$73,$EF,$E5,$F7,$47,$F3,$78,$E4,$A3,$CE,$FF,$37,$8A,$FF
	.dc.b	$98,$DF,$6D,$FE,$62,$FF,$DF,$D1,$F7,$DF,$E6,$CF,$D5,$AF,$DC,$F8
	.dc.b	$DE,$CF,$1B,$DB,$C1,$6B,$FC,$C8,$7F,$07,$8F,$EF,$9C,$9F,$C5,$1F
	.dc.b	$F0,$16,$9F,$C4,$DF,$E6,$F3,$FB,$F7,$FC,$17,$1F,$C5,$5F,$E6,$C9
	.dc.b	$FE,$77,$19,$3F,$99,$9C,$CF,$E1,$07,$F1,$37,$2E,$63,$F6,$B1,$FC
	.dc.b	$DB,$4F,$E6,$0F,$F9,$9A,$7C,$95,$FC,$5B,$C6,$FB,$3E,$37,$F2,$8E
	.dc.b	$6F,$E5,$AF,$F3,$13,$F9,$5D,$62,$3F,$99,$F3,$7F,$2F,$9F,$F3,$3C
	.dc.b	$FF,$29,$72,$E6,$7F,$93,$BC,$5F,$E4,$98,$FF,$43,$9C,$CE,$6B,$F6
	.dc.b	$51,$FE,$67,$8E,$94,$9F,$E2,$33,$FE,$68,$9F,$04,$7F,$30,$8F,$C9
	.dc.b	$F1,$FF,$2A,$3F,$94,$9F,$E6,$7B,$FE,$A4,$FF,$28,$AF,$E6,$4F,$F2
	.dc.b	$87,$FA,$13,$FC,$A1,$FE,$9E,$FF,$99,$5D,$BF,$CA,$1F,$EA,$0F,$F2
	.dc.b	$88,$EA,$FF,$28,$3F,$CC,$F1,$FD,$25,$FF,$36,$3C,$39,$BF,$A3,$C7
	.dc.b	$F2,$68,$FE,$BB,$1E,$A8,$FE,$4A,$7F,$99,$39,$7C,$6F,$F1,$E2,$1E
	.dc.b	$23,$FD,$FB,$5E,$27,$35,$97,$84,$FF,$65,$7F,$96,$C6,$5F,$F3,$14
	.dc.b	$F1,$7F,$92,$1C,$23,$D4,$FF,$BF,$A3,$DF,$3F,$D1,$23,$FE,$00,$CF
	.dc.b	$C8,$8F,$6B,$FD,$91,$FE,$0F,$E3,$7F,$25,$E4,$FE,$F6,$FF,$24,$BF
	.dc.b	$F0,$33,$FC,$A6,$B3,$3F,$EF,$0E,$5F,$EF,$D3,$FE,$64,$9F,$ED,$4F
	.dc.b	$F9,$8F,$80,$FF,$24,$6D,$7F,$20,$AF,$EF,$53,$FC,$92,$BF,$C0,$0E
	.dc.b	$5C,$CF,$F2,$37,$FC,$15,$1F,$C8,$EB,$9C,$7F,$62,$8F,$F2,$EB,$EA
	.dc.b	$3F,$C0,$3E,$2F,$BA,$BF,$97,$9F,$CC,$39,$BD,$6E,$0F,$E3,$58,$FC
	.dc.b	$07,$FC,$B3,$5A,$7F,$CB,$B1,$97,$94,$7C,$27,$AF,$2F,$99,$FE,$4D
	.dc.b	$1E,$C7,$FC,$CA,$ED,$FF,$35,$3E,$E7,$F9,$33,$BA,$F6,$47,$F2,$D8
	.dc.b	$FE,$C9,$87,$F2,$6B,$FA,$B7,$FE,$1E,$F1,$BF,$A0,$E3,$FE,$59,$9F
	.dc.b	$F1,$F3,$FC,$91,$B7,$FC,$80,$FF,$1C,$C7,$F2,$47,$FC,$60,$E5,$C9
	.dc.b	$FE,$C7,$E3,$FF,$1D,$47,$F2,$48,$FF,$1F,$39,$72,$7F,$24,$28,$8F
	.dc.b	$EF,$91,$98,$DC,$7F,$86,$A3,$F9,$24,$79,$23,$FC,$73,$E3,$7F,$5C
	.dc.b	$F1,$BF,$92,$5F,$F9,$65,$FF,$2A,$1F,$C9,$1C,$72,$7F,$20,$6B,$FE
	.dc.b	$46,$6B,$FE,$46,$6B,$FE,$4B,$8F,$F2,$8E,$C7,$F9,$1C,$F2,$E4,$FF
	.dc.b	$0A,$47,$F2,$39,$D5,$7F,$23,$C7,$2F,$F9,$42,$71,$1C,$0F,$F2,$9E
	.dc.b	$9F,$DD,$BC,$5F,$F2,$69,$FE,$4C,$3F,$92,$47,$F9,$4D,$FE,$48,$FF
	.dc.b	$98,$DF,$E4,$96,$7F,$92,$63,$FE,$7B,$7F,$92,$4E,$F8,$FF,$92,$3F
	.dc.b	$E7,$68,$FE,$48,$F5,$8F,$F2,$5C,$7F,$24,$8F,$F3,$B3,$FC,$91,$E1
	.dc.b	$C9,$FE,$6A,$8F,$E4,$95,$FE,$89,$F1,$BD,$57,$FE,$48,$9F,$E4,$93
	.dc.b	$E3,$8F,$E4,$67,$F9,$0A,$FC,$4F,$F2,$47,$C4,$FF,$24,$72,$FF,$1D
	.dc.b	$6B,$FE,$47,$7D,$8F,$FB,$FD,$C9,$FE,$83,$9E,$0F,$FA,$45,$FF,$23
	.dc.b	$BE,$0E,$5F,$F3,$3C,$7F,$24,$AF,$91,$1E,$D9,$FF,$43,$F1,$FF,$23
	.dc.b	$9F,$F5,$33,$FC,$8E,$3F,$D0,$CF,$F2,$38,$CB,$FC,$8E,$3F,$D4,$6F
	.dc.b	$F2,$37,$FD,$34,$FF,$23,$3F,$91,$1F,$E4,$47,$6F,$F9,$31,$FF,$54
	.dc.b	$3F,$C8,$5F,$F5,$43,$97,$2F,$FB,$EA,$3C,$31,$FE,$82,$8F,$E4,$31
	.dc.b	$E8,$8F,$F5,$0F,$2F,$F9,$12,$7F,$C8,$73,$E8,$7F,$90,$69,$E2,$7F
	.dc.b	$C8,$75,$EE,$72,$E5,$FF,$1E,$35,$FE,$28,$E7,$97,$FD,$69,$1F,$C8
	.dc.b	$A3,$D2,$F0,$7D,$2F,$F1,$ED,$7F,$94,$C7,$08,$E6,$FC,$09,$FA,$75
	.dc.b	$FE,$BA,$7F,$CB,$BE,$2F,$F9,$7B,$C6,$B5,$FE,$42,$C1,$3F,$C7,$E3
	.dc.b	$FD,$78,$FF,$90,$9F,$F6,$BB,$FC,$7E,$37,$1F,$C6,$F1,$FF,$72,$4F
	.dc.b	$F9,$08,$FE,$3D,$1C,$0F,$F1,$EC,$69,$FE,$34,$64,$FE,$3C,$7F,$B5
	.dc.b	$63,$2F,$1E,$4F,$F5,$BC,$7F,$1E,$9E,$F1,$FE,$D6,$F1,$BF,$CF,$3E
	.dc.b	$37,$F1,$EB,$FF,$6A,$3F,$EF,$13,$F8,$F3,$8E,$4F,$E3,$CD,$7F,$8F
	.dc.b	$35,$FE,$3C,$D7,$FD,$D1,$1F,$C7,$8F,$E3,$B1,$C7,$97,$FD,$61,$1F
	.dc.b	$C7,$4F,$F7,$44,$7F,$BE,$DF,$F1,$F6,$47,$F8,$EC,$74,$3F,$DD,$FE
	.dc.b	$2F,$FB,$A0,$FF,$73,$9F,$C7,$1F,$F7,$E3,$FC,$72,$4F,$F1,$C3,$F8
	.dc.b	$DC,$7F,$BC,$DF,$E3,$71,$AE,$3F,$E3,$79,$AB,$51,$FC,$65,$46,$5E
	.dc.b	$F1,$FE,$E8,$AD,$46,$8E,$3D,$DF,$EE,$A8,$FF,$74,$47,$13,$47,$A2
	.dc.b	$30,$F5,$AE,$D6,$F1,$BF,$D6,$37,$FC,$83,$0F,$F2,$25,$6A,$3F,$92
	.dc.b	$DF,$A2,$7F,$DD,$73,$97,$F9,$24,$7A,$1F,$E4,$91,$97,$F9,$23,$5F
	.dc.b	$E4,$97,$F1,$5F,$E4,$91,$99,$F5,$3C,$79,$3F,$DD,$97,$FC,$92,$3C
	.dc.b	$11,$F0,$6B,$EC,$BF,$6A,$BD,$E7,$CF,$9F,$7E,$DF,$EC,$EF,$1B,$F9
	.dc.b	$34,$FD,$3C,$51,$FC,$9E,$11,$8E,$4F,$F7,$6D,$FC,$8A,$FF,$18,$47
	.dc.b	$FB,$AA,$BE,$7D,$6E,$3F,$C6,$F1,$E1,$3B,$4F,$F9,$34,$FF,$26,$1F
	.dc.b	$CA,$EB,$A4,$FF,$27,$6B,$FE,$ED,$9F,$1C,$FF,$2B,$7B,$BF,$E4,$B3
	.dc.b	$FC,$9C,$FD,$D3,$BF,$2F,$C2,$76,$5A,$EB,$5F,$E4,$CE,$4E,$6F,$B3
	.dc.b	$96,$BF,$CB,$2B,$C2,$FF,$2C,$AE,$2F,$F9,$0A,$36,$FF,$B9,$1B,$5F
	.dc.b	$69,$F8,$31,$FE,$4A,$7A,$C7,$E5,$27,$DF,$56,$77,$3F,$6A,$7E,$2C
	.dc.b	$D9,$F9,$73,$EB,$7E,$6C,$57,$EC,$F8,$DF,$05,$FA,$A7,$F9,$34,$FE
	.dc.b	$3A,$7E,$CF,$3C,$BF,$E4,$DB,$FD,$2B,$FC,$BC,$FF,$26,$38,$7F,$C9
	.dc.b	$07,$03,$FC,$88,$E5,$E8,$7D,$33,$F1,$66,$4E,$11,$EF,$72,$7F,$BB
	.dc.b	$63,$B4,$7B,$E3,$8E,$7E,$D3,$AB,$C5,$FF,$28,$AF,$CA,$3B,$DF,$FB
	.dc.b	$CF,$93,$15,$FC,$46,$3F,$DE,$DC,$98,$B4,$7F,$BE,$5C,$2A,$84,$3F
	.dc.b	$C9,$17,$D9,$C1,$B9,$FE,$3D,$19,$8E,$15,$FC,$76,$31,$1D,$27,$17
	.dc.b	$E5,$8D,$C6,$0F,$2B,$FE,$FE,$9E,$35,$FC,$71,$E7,$36,$B9,$57,$F3
	.dc.b	$E8,$F1,$C6,$1F,$E0,$B3,$E9,$AF,$D3,$3E,$B8,$B3,$F1,$E7,$F1,$53
	.dc.b	$F3,$26,$CF,$6A,$F8,$3D,$BF,$21,$F8,$67,$C9,$ED,$C3,$8E,$1A,$FC
	.dc.b	$8E,$DA,$FC,$DD,$BE,$8F,$27,$F3,$3E,$47,$6E,$1F,$E6,$DC,$F5,$FF
	.dc.b	$7C,$C7,$BB,$5F,$B7,$5A,$75,$23,$9B,$BE,$0F,$E6,$51,$9B,$FA,$11
	.dc.b	$8B,$FE,$99,$BF,$E3,$30,$E5,$7B,$C3,$71,$8A,$FA,$51,$E2,$9D,$3A
	.dc.b	$7F,$AB,$BE,$19,$FF,$3C,$57,$5A,$FA,$33,$EA,$C2,$E1,$FD,$5E,$2F
	.dc.b	$5F,$D2,$7B,$63,$FD,$67,$0F,$33,$FD,$6F,$5F,$82,$FA,$26,$A8,$F1
	.dc.b	$3F,$E8,$08,$AF,$F9,$EE,$79,$BC,$63,$2F,$17,$D8,$FF,$A1,$DB,$3F
	.dc.b	$0A,$FF,$AF,$45,$CB,$FD,$04,$F8,$0F,$C9,$F0,$7E,$D6,$3A,$46,$CF
	.dc.b	$83,$C5,$EF,$AF,$E3,$95,$FB,$69,$FE,$7D,$35,$FF,$73,$63,$F9,$77
	.dc.b	$FC,$03,$1F,$7D,$FF,$73,$46,$5F,$F0,$84,$7E,$01,$FF,$24,$C6,$5F
	.dc.b	$F2,$29,$C4,$FF,$24,$BC,$63,$DD,$1F,$E4,$88,$F0,$47,$B2,$3F,$86
	.dc.b	$9F,$EE,$68,$FD,$54,$7D,$C9,$FB,$5C,$FF,$C5,$E7,$F9,$3B,$FC,$59
	.dc.b	$BB,$FE,$B0,$FC,$F4,$5E,$3B,$1F,$B0,$C2,$BF,$C2,$9F,$E3,$D3,$97
	.dc.b	$31,$FC,$76,$7F,$3B,$7F,$E3,$5A,$FE,$37,$58,$3F,$C0,$B5,$FC,$8A
	.dc.b	$BF,$3B,$3F,$C8,$E6,$CF,$F9,$41,$EE,$FF,$23,$73,$3F,$9C,$9C,$47
	.dc.b	$28,$F8,$2F,$C2,$E6,$FF,$06,$83,$D2,$63,$9E,$BF,$E1,$28,$AE,$CE
	.dc.b	$11,$F1,$9D,$46,$9C,$B9,$7F,$C2,$ED,$72,$D7,$DE,$7F,$AF,$5D,$18
	.dc.b	$7F,$C4,$4F,$C7,$AF,$F0,$B3,$E8,$9F,$81,$3F,$79,$F7,$3F,$E2,$AE
	.dc.b	$7F,$23,$E9,$E3,$AA,$3C,$4F,$C3,$8B,$1F,$D1,$6F,$FC,$31,$7E,$78
	.dc.b	$F1,$57,$FA,$EE,$3D,$91,$E7,$8F,$F5,$BC,$7C,$A8,$FC,$B4,$FF,$85
	.dc.b	$63,$E4,$47,$A2,$BE,$C3,$F9,$67,$F8,$87,$98,$E7,$B5,$87,$FC,$61
	.dc.b	$A7,$F6,$D9,$D4,$FF,$95,$36,$C5,$7F,$AF,$76,$FF,$02,$D6,$AB,$FC
	.dc.b	$F5,$5F,$CE,$E3,$11,$FE,$57,$31,$5F,$BC,$69,$FE,$11,$AE,$71,$FE
	.dc.b	$16,$8E,$E7,$EE,$2F,$FD,$25,$5D,$EF,$F9,$BB,$E6,$8F,$DC,$3E,$88
	.dc.b	$B5,$FB,$99,$E2,$FF,$04,$AF,$D3,$53,$9F,$E2,$F1,$7C,$6E,$7F,$DC
	.dc.b	$70,$FC,$37,$F7,$19,$FC,$07,$F7,$98,$66,$7F,$C1,$7C,$3F,$06,$7F
	.dc.b	$71,$5B,$9F,$DE,$35,$FF,$05,$56,$DF,$DB,$3F,$E0,$9C,$55,$FE,$09
	.dc.b	$CF,$E0,$BA,$9D,$3C,$9B,$5F,$E0,$97,$C8,$7C,$48,$C3,$EB,$3F,$AB
	.dc.b	$D7,$E3,$1F,$F0,$C7,$99,$FD,$E7,$97,$FB,$CE,$3F,$E4,$28,$C7,$3F
	.dc.b	$F8,$C5,$FE,$FD,$D1,$FD,$E5,$FF,$00,$45,$AF,$ED,$E7,$F7,$99,$EA
	.dc.b	$E0,$46,$1F,$EE,$57,$5F,$EE,$2F,$77,$2E,$63,$FB,$BD,$62,$39,$47
	.dc.b	$F3,$98,$FF,$68,$C7,$18,$FE,$EB,$1F,$DE,$A3,$DF,$1F,$DC,$E3,$E0
	.dc.b	$4F,$73,$FB,$95,$FC,$17,$FC,$39,$59,$7D,$BA,$FB,$DF,$EE,$51,$97
	.dc.b	$FC,$31,$1E,$B7,$FB,$74,$66,$3F,$9B,$57,$08,$E5,$3F,$DA,$9A,$7B
	.dc.b	$9F,$EE,$2F,$94,$F7,$3B,$7E,$69,$EE,$9F,$63,$FE,$0F,$F3,$1E,$3F
	.dc.b	$F3,$6E,$6F,$F3,$54,$61,$FF,$62,$F3,$E2,$7F,$9D,$D5,$9F,$E6,$C7
	.dc.b	$F3,$2C,$BF,$D0,$B7,$C6,$3F,$99,$C6,$2B,$71,$FC,$BA,$30,$FF,$38
	.dc.b	$AE,$8F,$F3,$86,$BF,$E3,$2A,$E9,$76,$FF,$D9,$57,$CA,$3F,$BB,$4F
	.dc.b	$A6,$ED,$7F,$3C,$7F,$B1,$47,$C0,$7F,$C6,$AB,$AF,$E2,$47,$C5,$9F
	.dc.b	$EB,$3C,$FE,$2C,$7F,$9E,$63,$E8,$7F,$DA,$15,$95,$CF,$FE,$59,$C7
	.dc.b	$E1,$47,$F6,$1B,$FF,$41,$46,$23,$75,$BC,$39,$B6,$3F,$D9,$D8,$77
	.dc.b	$BF,$81,$1F,$ED,$59,$D4,$7B,$A3,$F7,$13,$F3,$E3,$DB,$5F,$BB,$AF
	.dc.b	$BB,$E3,$FE,$07,$97,$DF,$02,$13,$97,$F0,$6F,$F8,$D3,$0F,$BA,$FF
	.dc.b	$3F,$8E,$19,$7F,$8E,$23,$67,$23,$FD,$87,$3C,$1F,$F6,$7B,$5F,$E8
	.dc.b	$47,$C7,$3F,$9F,$39,$9F,$C0,$47,$F8,$F0,$FB,$27,$F8,$D5,$F0,$BF
	.dc.b	$EC,$A3,$E6,$9F,$CF,$44,$E5,$FA,$31,$FA,$1E,$8F,$F3,$87,$99,$FE
	.dc.b	$37,$8F,$84,$FF,$8D,$DB,$CF,$FC,$F3,$0F,$F7,$0C,$FF,$8D,$EF,$FD
	.dc.b	$A0,$FF,$40,$6C,$FF,$40,$7F,$D3,$51,$FE,$37,$73,$1F,$CD,$DA,$FF
	.dc.b	$8A,$9B,$8F,$FA,$42,$7F,$C6,$91,$FE,$39,$78,$3D,$27,$83,$67,$FD
	.dc.b	$1F,$38,$8F,$0C,$7B,$9F,$7F,$67,$F3,$6B,$F5,$BF,$CD,$9B,$C7,$FC
	.dc.b	$CF,$9B,$D3,$3F,$EB,$19,$CC,$FF,$30,$6B,$FE,$A4,$6C,$F9,$E7,$F9
	.dc.b	$FF,$35,$AF,$E8,$17,$5F,$E7,$EE,$6B,$FC,$DD,$8A,$FF,$D5,$79,$FD
	.dc.b	$47,$F9,$B5,$7D,$07,$8B,$F1,$FB,$BE,$DF,$77,$BB,$1F,$F6,$86,$7E
	.dc.b	$67,$FD,$7F,$CF,$C5,$FE,$A0,$7B,$8F,$EA,$4D,$7F,$AB,$DD,$7F,$9C
	.dc.b	$B5,$F5,$1F,$CD,$67,$E1,$C7,$B6,$7E,$D6,$1F,$EA,$F8,$FD,$6C,$FC
	.dc.b	$E9,$F7,$CF,$F4,$29,$CE,$3E,$F9,$FB,$73,$F9,$6D,$FF,$D8,$DE,$67
	.dc.b	$F3,$7E,$56,$2C,$3F,$75,$3F,$CE,$FA,$33,$18,$AF,$E7,$35,$BC,$97
	.dc.b	$FE,$73,$76,$62,$7F,$CD,$F3,$B9,$FE,$87,$3F,$EF,$57,$FC,$E6,$F0
	.dc.b	$7F,$9E,$DF,$FA,$3D,$FE,$7F,$16,$BF,$CE,$0E,$DF,$E7,$37,$E5,$7F
	.dc.b	$9B,$47,$57,$FD,$10,$FF,$AC,$67,$FA,$23,$F1,$8F,$31,$FE,$EE,$78
	.dc.b	$B9,$7F,$89,$3F,$EF,$A7,$FA,$F7,$37,$86,$7F,$A9,$9F,$CE,$30,$F7
	.dc.b	$3F,$E7,$BA,$D3,$FD,$7B,$9F,$C8,$FF,$36,$9D,$BF,$E7,$56,$BF,$CF
	.dc.b	$E7,$C4,$FF,$9E,$5E,$75,$E3,$8F,$F3,$B4,$F2,$8F,$74,$7F,$3D,$9F
	.dc.b	$85,$1E,$D8,$FE,$75,$3F,$4E,$3F,$07,$5E,$EA,$F8,$9E,$67,$F9,$B9
	.dc.b	$FF,$60,$39,$73,$DD,$FE,$7C,$BF,$C1,$4E,$E7,$FA,$16,$9F,$82,$8C
	.dc.b	$67,$FD,$16,$B7,$7F,$E8,$1B,$D3,$FE,$F2,$9F,$E9,$55,$E1,$BE,$F5
	.dc.b	$FE,$7B,$9C,$D7,$F3,$E6,$C7,$F9,$EC,$57,$F9,$ED,$F4,$CF,$F3,$F7
	.dc.b	$2F,$F3,$B7,$C3,$3E,$DB,$F2,$DD,$8F,$1F,$77,$C1,$AF,$66,$F6,$FE
	.dc.b	$0F,$77,$F9,$F1,$BD,$03,$FC,$F2,$E1,$FF,$75,$B5,$FF,$3A,$B5,$FF
	.dc.b	$73,$37,$A3,$E9,$CF,$FB,$CA,$FE,$56,$3F,$1A,$3F,$DF,$91,$F3,$DF
	.dc.b	$F3,$E3,$F0,$9F,$E7,$F5,$E1,$8F,$CD,$BE,$6A,$FC,$03,$E8,$8B,$5D
	.dc.b	$67,$E1,$47,$32,$E9,$FE,$03,$7F,$7F,$3F,$E7,$DE,$7B,$C9,$FD,$0E
	.dc.b	$3F,$C0,$87,$F9,$F2,$73,$1F,$E7,$7E,$8F,$5C,$7F,$11,$8F,$E9,$0E
	.dc.b	$27,$F9,$6C,$FE,$DD,$CC,$66,$7F,$A2,$B9,$7F,$CF,$0F,$67,$F9,$EB
	.dc.b	$B7,$FC,$ED,$8F,$F1,$99,$47,$F4,$69,$F1,$5F,$C0,$9F,$F4,$3C,$78
	.dc.b	$63,$F0,$11,$EC,$BB,$3E,$9A,$FE,$8F,$3F,$01,$FE,$8A,$34,$F2,$3F
	.dc.b	$D1,$30,$F1,$3F,$D2,$A3,$CC,$FF,$48,$AD,$BC,$5C,$BF,$E8,$09,$E2
	.dc.b	$FF,$54,$8F,$84,$FF,$47,$7E,$14,$FE,$02,$BC,$93,$FD,$1E,$7E,$9C
	.dc.b	$FE,$01,$F5,$16,$7F,$D0,$AF,$DC,$3E,$24,$7C,$08,$FD,$24,$7E,$0B
	.dc.b	$C3,$C7,$3E,$79,$B5,$3F,$C1,$EB,$FC,$6E,$FF,$2C,$E1,$FC,$4B,$FE
	.dc.b	$43,$3F,$A6,$1F,$EF,$BC,$BE,$B3,$FC,$14,$C9,$FE,$3B,$38,$9F,$C1
	.dc.b	$4C,$9F,$EF,$37,$8B,$FC,$16,$7F,$C9,$93,$F0,$E3,$BC,$7B,$DF,$E0
	.dc.b	$93,$E0,$8F,$C1,$47,$FB,$F4,$FC,$93,$F0,$67,$F8,$16,$BF,$E6,$1F
	.dc.b	$07,$FD,$FB,$C3,$FD,$B2,$30,$FF,$BF,$9A,$FF,$05,$21,$FC,$95,$CB
	.dc.b	$8E,$0C,$60,$AF,$F2,$36,$3F,$C7,$A3,$15,$FE,$47,$7C,$06,$1F,$F2
	.dc.b	$0B,$89,$D3,$63,$2D,$7F,$8C,$57,$F0,$6B,$F1,$CD,$3A,$47,$F0,$88
	.dc.b	$F5,$C5,$3C,$71,$FC,$AE,$3E,$0C,$53,$D5,$1F,$21,$F6,$17,$C1,$FF
	.dc.b	$0A,$61,$ED,$7F,$C2,$4D,$E7,$FF,$08,$DF,$FA,$3E,$7F,$C2,$13,$99
	.dc.b	$FF,$03,$D7,$C2,$9F,$F0,$7B,$63,$2D,$47,$FB,$1E,$79,$3F,$E0,$88
	.dc.b	$FF,$17,$D7,$78,$FF,$08,$CE,$A7,$E2,$57,$F8,$51,$E8,$7C,$37,$FC
	.dc.b	$1E,$FD,$A7,$A3,$C3,$4F,$80,$FC,$1A,$FF,$63,$D7,$89,$FF,$04,$35
	.dc.b	$FF,$02,$35,$FF,$60,$B9,$E7,$F3,$CA,$FF,$C2,$BC,$F8,$7F,$DD,$33
	.dc.b	$FE,$AD,$AE,$D3,$71,$F7,$61,$FE,$11,$C3,$83,$FE,$0E,$8C,$DF,$B6
	.dc.b	$3F,$C1,$D3,$F3,$63,$EF,$61,$88,$FA,$51,$F7,$6F,$F0,$1E,$03,$86
	.dc.b	$F0,$E2,$F3,$A7,$E1,$7C,$1F,$E1,$BA,$FE,$7F,$A9,$E6,$FF,$83,$EF
	.dc.b	$F3,$F1,$87,$F1,$CF,$ED,$DF,$86,$F2,$DF,$FC,$BB,$F3,$9F,$91,$86
	.dc.b	$5F,$93,$EF,$D5,$FF,$98,$FD,$5F,$33,$F9,$DE,$74,$70,$8C,$3F,$9B
	.dc.b	$8F,$E6,$D1,$FA,$3C,$79,$C7,$9E,$3F,$6A,$E2,$7F,$C9,$F1,$8F,$A6
	.dc.b	$FF,$06,$63,$FE,$EB,$D5,$F4,$BE,$2A,$FE,$F3,$39,$8F,$14,$60,$4F
	.dc.b	$07,$FC,$39,$3C,$FC,$6E,$B1,$F8,$52,$CF,$CE,$9E,$9C,$1C,$27,$9F
	.dc.b	$CB,$6B,$A3,$8D,$17,$BB,$FE,$41,$E3,$FF,$41,$DF,$FA,$12,$72,$FF
	.dc.b	$83,$AB,$FD,$1B,$1F,$1A,$3F,$25,$1F,$E9,$0F,$4B,$AC,$F9,$B8,$7C
	.dc.b	$95,$D0,$F2,$57,$87,$D5,$B5,$FD,$FA,$7F,$D4,$B3,$FE,$10,$F9,$F7
	.dc.b	$B9,$F2,$67,$27,$FA,$7E,$33,$CB,$E5,$8C,$4F,$C1,$E4,$57,$D6,$D3
	.dc.b	$DB,$18,$AF,$8D,$CB,$E6,$8F,$89,$1C,$4F,$9E,$F2,$BF,$F4,$6D,$72
	.dc.b	$B7,$BB,$DD,$FF,$3A,$E5,$FE,$28,$FA,$0F,$F2,$76,$DF,$D3,$7D,$6F
	.dc.b	$C8,$F5,$FF,$8E,$23,$1D,$3F,$82,$D7,$FD,$8B,$18,$74,$7F,$49,$AF
	.dc.b	$A0,$E8,$FE,$99,$F2,$5D,$9F,$E9,$DD,$17,$C3,$57,$EB,$F6,$61,$E3
	.dc.b	$BF,$F3,$43,$FD,$87,$C7,$E1,$1C,$6B,$F4,$FD,$3E,$2C,$39,$F6,$70
	.dc.b	$AE,$7F,$2E,$F7,$BF,$7B,$E6,$FF,$95,$B3,$56,$A7,$F7,$CF,$F1,$98
	.dc.b	$C8,$AE,$5E,$47,$C0,$8C,$57,$77,$59,$5A,$F8,$11,$F6,$2F,$E3,$FA
	.dc.b	$1F,$C4,$23,$AC,$FC,$D9,$F9,$10,$DF,$8C,$72,$9F,$57,$85,$E1,$7D
	.dc.b	$9D,$77,$E2,$BC,$5F,$3D,$F3,$FB,$5F,$1E,$F8,$57,$91,$F9,$7C,$FF
	.dc.b	$DE,$BA,$FF,$A9,$2B,$53,$E3,$E7,$FA,$75,$FE,$A0,$9C,$BF,$24,$C5
	.dc.b	$FF,$23,$9D,$E1,$E0,$8F,$A4,$F8,$B8,$AD,$F9,$7C,$9F,$36,$1E,$78
	.dc.b	$F0,$F1,$75,$BE,$5E,$3F,$E3,$9F,$5F,$C2,$F2,$E0,$EB,$F4,$BE,$FE
	.dc.b	$7F,$C7,$78,$1F,$37,$37,$88,$F9,$7F,$E3,$33,$F8,$29,$FC,$B4,$E7
	.dc.b	$34,$FE,$0A,$7F,$D9,$9C,$7F,$E8,$EB,$E3,$C1,$F0,$70,$E1,$7F,$A2
	.dc.b	$79,$5D,$AF,$E8,$CF,$CE,$3C,$6F,$E8,$E3,$F0,$11,$E2,$31,$1E,$6E
	.dc.b	$7E,$CF,$A3,$4B,$97,$D4,$AF,$CB,$1D,$1F,$A5,$97,$9F,$CE,$FD,$16
	.dc.b	$2F,$A2,$F0,$70,$FF,$B3,$7A,$FF,$43,$CE,$DD,$F3,$2E,$3F,$E6,$3C
	.dc.b	$5F,$DE,$EF,$75,$A9,$FE,$63,$B7,$EA,$87,$42,$B7,$87,$3F,$4F,$FD
	.dc.b	$8B,$18,$F6,$3F,$D8,$AE,$B8,$AE,$1F,$96,$CB,$C6,$E3,$4F,$37,$4F
	.dc.b	$FB,$AE,$FC,$73,$4F,$CE,$9E,$D3,$96,$5E,$B7,$C2,$62,$FA,$1F,$CC
	.dc.b	$67,$CF,$A3,$CC,$DE,$3F,$17,$0F,$89,$E2,$63,$07,$CD,$78,$69,$F4
	.dc.b	$3E,$17,$EE,$E7,$FC,$3B,$83,$E2,$3F,$DA,$8F,$37,$57,$C7,$F6,$DE
	.dc.b	$2F,$F7,$4F,$2B,$E8,$57,$E5,$38,$15,$AF,$4F,$F9,$A6,$3D,$31,$E5
	.dc.b	$8F,$F7,$5C,$3E,$A4,$E7,$0E,$33,$FC,$36,$7A,$D7,$87,$E0,$FE,$8B
	.dc.b	$E3,$E6,$B6,$FA,$5F,$9D,$3F,$77,$9A,$CF,$AB,$87,$D7,$8F,$B2,$3B
	.dc.b	$BF,$42,$BD,$DD,$1D,$36,$E3,$7C,$9F,$67,$4F,$93,$0F,$3F,$AB,$C5
	.dc.b	$F4,$FA,$7C,$F8,$FE,$F7,$7F,$F8,$EA,$7F,$9D,$F7,$3E,$59,$C4,$F9
	.dc.b	$0F,$F0,$67,$AE,$DF,$E2,$E1,$C2,$F7,$5A,$E1,$C9,$FC,$2A,$33,$A7
	.dc.b	$F8,$37,$9B,$ED,$E7,$DF,$3E,$DE,$BF,$E3,$A3,$18,$FF,$41,$C7,$F0
	.dc.b	$59,$77,$B7,$C9,$7E,$82,$F9,$79,$F7,$47,$C3,$FE,$76,$7D,$9D,$37
	.dc.b	$E4,$7F,$3D,$1F,$E4,$1D,$BF,$1B,$E6,$7C,$3D,$7D,$1A,$7C,$17,$C9
	.dc.b	$97,$C4,$F0,$7B,$4F,$58,$FE,$E5,$B7,$C3,$AF,$01,$DA,$3E,$0F,$87
	.dc.b	$F9,$4C,$BD,$BC,$4C,$71,$AF,$F0,$3C,$21,$FC,$13,$9B,$FC,$A1,$CA
	.dc.b	$F9,$47,$D1,$B8,$7F,$9A,$5E,$CF,$F2,$7D,$6E,$75,$EE,$FF,$A3,$70
	.dc.b	$E7,$87,$2C,$3F,$DE,$91,$8A,$E9,$86,$CE,$75,$E0,$E8,$F2,$F1,$FF
	.dc.b	$08,$F3,$3F,$C9,$F3,$FE,$31,$AE,$11,$8D,$BD,$71,$E5,$8F,$F2,$AE
	.dc.b	$18,$8C,$69,$F7,$DF,$D8,$E5,$FE,$6F,$E0,$D4,$71,$DA,$E7,$FC,$63
	.dc.b	$CC,$FF,$49,$DF,$A6,$BC,$AF,$B9,$F0,$CF,$9F,$0E,$33,$FE,$7B,$9A
	.dc.b	$7B,$B1,$F9,$9E,$77,$AA,$BE,$56,$FF,$C2,$2F,$E3,$7C,$AF,$6C,$38
	.dc.b	$67,$75,$FF,$30,$47,$A8,$FE,$CB,$E3,$3E,$91,$F4,$CF,$8B,$FF,$24
	.dc.b	$D7,$F1,$FC,$BF,$C5,$BE,$37,$FA,$3A,$F7,$5F,$E9,$EF,$1F,$A4,$E7
	.dc.b	$0E,$5E,$EF,$D7,$8C,$57,$EB,$F0,$E7,$97,$86,$DF,$27,$E9,$BE,$5E
	.dc.b	$3D,$AD,$69,$7E,$93,$EB,$72,$FF,$46,$BF,$F5,$0F,$56,$33,$EB,$74
	.dc.b	$FF,$01,$F8,$7F,$33,$C7,$FA,$B1,$F0,$B8,$7D,$77,$E7,$CF,$C9,$BF
	.dc.b	$86,$7E,$2F,$0D,$EB,$FB,$39,$FA,$A7,$B6,$9F,$EB,$2F,$4A,$E7,$F6
	.dc.b	$A3,$E0,$F1,$B8,$7E,$3E,$9D,$BC,$DF,$06,$2F,$98,$EB,$8F,$F8,$6B
	.dc.b	$CB,$E5,$AF,$E1,$3C,$3F,$F5,$67,$D2,$3E,$97,$FA,$F7,$BF,$FE,$60
	.dc.b	$E7,$FE,$D3,$5A,$9C,$FA,$1F,$81,$8C,$78,$9F,$E6,$BC,$F1,$93,$2E
	.dc.b	$93,$E7,$9C,$C6,$1F,$D9,$57,$5E,$EF,$C8,$3E,$4F,$8C,$F4,$BE,$6E
	.dc.b	$9B,$8F,$5B,$BE,$47,$D5,$A7,$A1,$ED,$AF,$D0,$F9,$BF,$5A,$7C,$FC
	.dc.b	$1F,$E3,$FE,$0F,$6D,$79,$F8,$3F,$BF,$75,$F9,$23,$19,$F9,$1F,$67
	.dc.b	$57,$92,$3B,$70,$7B,$B1,$78,$DE,$EC,$7E,$5E,$37,$89,$EF,$F5,$C3
	.dc.b	$E8,$7C,$BF,$57,$E9,$70,$F7,$F9,$4F,$66,$DF,$DD,$75,$FE,$2B,$E3
	.dc.b	$7F,$A5,$BE,$0F,$F9,$37,$C7,$3E,$75,$EF,$38,$DD,$5F,$E0,$0C,$38
	.dc.b	$79,$8C,$38,$5B,$E5,$3F,$9B,$73,$FF,$84,$F3,$EB,$9F,$4C,$F3,$97
	.dc.b	$2C,$78,$E9,$81,$7C,$39,$3C,$F1,$8D,$3C,$18,$FF,$18,$C3,$96,$36
	.dc.b	$B5,$8F,$ED,$27,$CF,$C1,$67,$AF,$9B,$FC,$9A,$7E,$17,$65,$9E,$38
	.dc.b	$F9,$FC,$EF,$81,$18,$E0,$F8,$91,$FD,$FE,$3F,$CA,$B7,$EB,$AF,$F3
	.dc.b	$04,$7A,$F6,$FE,$A5,$C3,$E1,$7F,$A4,$C7,$9D,$FE,$72,$F2,$3B,$6B
	.dc.b	$FD,$85,$FF,$1F,$5F,$5C,$7C,$11,$83,$CF,$EC,$FC,$6B,$FB,$7F,$07
	.dc.b	$D7,$83,$EB,$6B,$7E,$67,$C5,$A7,$BE,$FE,$AE,$5F,$7F,$AB,$DF,$E6
	.dc.b	$FF,$BA,$BE,$51,$F3,$BF,$E2,$0E,$6F,$D9,$6F,$E7,$AD,$4E,$63,$8F
	.dc.b	$C5,$2F,$8F,$B7,$F6,$AF,$7A,$5A,$D7,$D2,$FD,$99,$F0,$74,$59,$7C
	.dc.b	$B7,$D9,$9F,$F7,$EF,$A9,$7E,$37,$EC,$47,$E7,$74,$FC,$3F,$BD,$F3
	.dc.b	$3B,$BF,$CC,$93,$E6,$D5,$C0,$F9,$1F,$3F,$07,$9A,$30,$7C,$DA,$E0
	.dc.b	$EF,$E1,$31,$B7,$B7,$8B,$E9,$E9,$8C,$3E,$3B,$ED,$7D,$13,$E6,$E0
	.dc.b	$F3,$CF,$A2,$39,$67,$E4,$EC,$F1,$E1,$FE,$4E,$FA,$3F,$D2,$69,$E9
	.dc.b	$F5,$6E,$1E,$1F,$82,$ED,$C7,$F1,$FC,$5E,$FB,$7A,$FE,$4B,$9E,$BF
	.dc.b	$03,$8B,$F9,$57,$1F,$FA,$1F,$CF,$7D,$0F,$EE,$7E,$83,$FC,$93,$EA
	.dc.b	$7C,$6E,$4F,$C7,$56,$BC,$A5,$6A,$7F,$4B,$39,$8F,$F2,$8C,$7E,$EE
	.dc.b	$31,$B7,$EE,$9E,$5F,$45,$FB,$77,$A6,$F7,$27,$CE,$7E,$DE,$7C,$BC
	.dc.b	$F6,$7F,$6D,$1F,$C5,$71,$D1,$88,$F5,$7C,$47,$FB,$47,$C2,$F2,$C6
	.dc.b	$36,$FF,$64,$57,$EB,$EF,$D1,$7F,$ED,$7F,$8E,$79,$9F,$03,$0E,$11
	.dc.b	$FE,$D8,$E1,$F9,$11,$CF,$1E,$BC,$3F,$13,$B3,$DD,$3F,$2F,$3B,$3F
	.dc.b	$95,$7A,$3E,$DE,$9F,$43,$F3,$33,$F1,$4E,$B8,$FF,$CF,$7F,$19,$F2
	.dc.b	$F0,$F6,$F7,$DF,$23,$E3,$7D,$1F,$F1,$9F,$81,$FD,$07,$9C,$C9,$C7
	.dc.b	$EE,$E0,$FF,$67,$C7,$C1,$CB,$DB,$E4,$FF,$0B,$F6,$5F,$41,$FA,$6F
	.dc.b	$A0,$7C,$B3,$E8,$7F,$8E,$E3,$FC,$5E,$F7,$C0,$BD,$D7,$F3,$B9,$CC
	.dc.b	$72,$F3,$B1,$8F,$5E,$EF,$E7,$B5,$CB,$E3,$7F,$9D,$3D,$22,$CB,$C9
	.dc.b	$FE,$75,$3E,$0C,$F2,$63,$3F,$E7,$51,$F0,$A0,$F8,$B4,$F5,$7C,$6F
	.dc.b	$F3,$C8,$C6,$AE,$2F,$F6,$D7,$67,$FB,$AE,$73,$0F,$86,$F8,$DA,$7C
	.dc.b	$5C,$7E,$24,$E7,$3F,$89,$39,$C7,$A5,$BC,$DC,$D7,$C7,$C3,$F8,$4E
	.dc.b	$5D,$F0,$E9,$A7,$C6,$AF,$99,$E2,$FC,$69,$FA,$93,$F2,$39,$7D,$F1
	.dc.b	$DF,$4F,$63,$E5,$F8,$7F,$9F,$DF,$F2,$23,$5C,$5E,$AC,$BC,$5B,$FC
	.dc.b	$1F,$84,$C7,$37,$DA,$CB,$E0,$62,$D9,$C7,$F8,$DF,$41,$F3,$2F,$41
	.dc.b	$EF,$BE,$2D,$F4,$3F,$E0,$4C,$38,$5E,$EB,$5C,$B9,$E5,$65,$D3,$2E
	.dc.b	$7F,$07,$FE,$AC,$C7,$8C,$76,$F0,$F1,$7C,$32,$E8,$E3,$0E,$5D,$DF
	.dc.b	$E9,$EC,$3C,$31,$FE,$78,$7C,$73,$47,$C5,$FF,$49,$8E,$FA,$7A,$F5
	.dc.b	$FE,$21,$5F,$0E,$73,$B7,$BB,$C4,$FD,$84,$E4,$F2,$6B,$F1,$27,$5D
	.dc.b	$5F,$B5,$E1,$F9,$75,$AC,$BB,$61,$E2,$8E,$5A,$7C,$39,$CE,$9F,$D6
	.dc.b	$AF,$8C,$E7,$1F,$14,$FE,$3E,$73,$B7,$CF,$9F,$F0,$66,$1E,$1D,$3D
	.dc.b	$2F,$D1,$C3,$B7,$9B,$F2,$6F,$EE,$4F,$C9,$3E,$AF,$4F,$FB,$3A,$FF
	.dc.b	$05,$14,$F5,$61,$F8,$0E,$4F,$87,$97,$8B,$7F,$85,$EA,$F7,$CB,$F1
	.dc.b	$1F,$4B,$D7,$5F,$C0,$79,$96,$7F,$87,$E7,$E5,$F8,$FE,$D5,$EF,$E4
	.dc.b	$71,$FE,$E6,$BE,$3D,$FB,$E7,$3C,$5F,$E9,$FD,$7E,$7F,$91,$FE,$9D
	.dc.b	$1E,$AB,$E5,$DF,$37,$F9,$39,$D4,$E7,$97,$1F,$1B,$FC,$B2,$1F,$45
	.dc.b	$F9,$2A,$DF,$87,$6B,$5B,$59,$FF,$14,$BF,$13,$A3,$F9,$B4,$78,$E3
	.dc.b	$E7,$3F,$EB,$17,$F2,$9E,$77,$E4,$DF,$E7,$75,$E1,$D1,$F4,$9F,$73
	.dc.b	$6F,$6C,$E6,$1F,$49,$FE,$57,$9C,$E3,$F0,$2F,$8C,$BE,$13,$F0,$98
	.dc.b	$F6,$AD,$61,$CA,$5D,$FF,$9D,$9F,$8B,$87,$29,$7C,$B6,$73,$F1,$63
	.dc.b	$EF,$3A,$70,$7C,$FF,$13,$E5,$E3,$E3,$9F,$A1,$C5,$F8,$B8,$F8,$07
	.dc.b	$6D,$3F,$D2,$73,$E6,$8C,$6B,$D5,$FA,$FD,$DF,$3F,$E8,$7E,$5E,$7F
	.dc.b	$D8,$BC,$1B,$C5,$FB,$FB,$FF,$6F,$70,$DF,$A2,$F0,$5F,$C9,$CB,$BF
	.dc.b	$1F,$C8,$E0,$FC,$36,$DF,$33,$E2,$FC,$57,$BE,$3F,$DA,$72,$FD,$C8
	.dc.b	$C7,$C8,$7F,$03,$F4,$9F,$27,$FD,$53,$DE,$7B,$2F,$98,$FE,$45,$5A
	.dc.b	$3F,$D6,$70,$E8,$CD,$F1,$EA,$FE,$43,$5B,$C3,$F4,$FD,$3D,$EB,$B3
	.dc.b	$A7,$FB,$1F,$67,$EE,$EB,$B4,$D9,$FD,$D4,$7F,$97,$23,$C6,$62,$3D
	.dc.b	$18,$BD,$8F,$F6,$67,$1F,$F1,$AB,$F1,$4E,$7D,$BF,$C0,$46,$2B,$FC
	.dc.b	$F7,$DB,$F8,$E8,$C7,$C4,$3E,$23,$19,$76,$BF,$F4,$0F,$97,$FE,$D2
	.dc.b	$CF,$B4,$E6,$D2,$F5,$FE,$06,$3F,$23,$0F,$8E,$FC,$AD,$6F,$67,$C1
	.dc.b	$EF,$F6,$4E,$F1,$F3,$E5,$D7,$CB,$F1,$CE,$73,$BC,$5F,$63,$3F,$2F
	.dc.b	$A3,$E0,$8F,$F3,$46,$9C,$AD,$F1,$BF,$37,$C1,$FE,$9D,$1F,$82,$DA
	.dc.b	$E9,$EF,$E7,$E1,$8F,$5D,$BD,$BF,$2D,$FD,$C7,$E3,$3F,$DF,$1A,$7F
	.dc.b	$4E,$21,$93,$FD,$D2,$67,$1F,$F7,$9C,$70,$8F,$F2,$26,$5C,$38,$F7
	.dc.b	$3B,$9D,$9C,$8F,$F1,$49,$B7,$95,$78,$23,$C1,$1D,$1C,$4F,$47,$F2
	.dc.b	$67,$F5,$F8,$FF,$2A,$1D,$1D,$B8,$9E,$91,$B3,$C4,$75,$31,$1E,$18
	.dc.b	$FE,$9A,$FB,$63,$FB,$8D,$61,$E2,$7F,$74,$9E,$2F,$A6,$2B,$CE,$BD
	.dc.b	$0E,$5F,$97,$DA,$F0,$E1,$74,$F8,$F1,$88,$D5,$61,$DC,$FC,$98,$FF
	.dc.b	$14,$C7,$27,$13,$A6,$C7,$26,$CF,$C4,$AF,$1C,$FC,$69,$FC,$05,$7F
	.dc.b	$05,$75,$8F,$FA,$06,$3F,$DC,$4D,$9F,$31,$E5,$3D,$17,$F1,$E7,$31
	.dc.b	$E1,$BE,$EF,$09,$FA,$93,$A8,$F7,$CE,$9D,$35,$EE,$FE,$53,$D6,$7C
	.dc.b	$FF,$F8,$FB,$A5,$F1,$7F,$E1,$4C,$FF,$35,$5A,$F4,$55,$FD,$02,$7E
	.dc.b	$E4,$E7,$DF,$FC,$4C,$7F,$9E,$70,$E8,$FC,$F9,$EF,$E7,$FC,$48,$D7
	.dc.b	$05,$9F,$1C,$5E,$7F,$99,$A7,$9E,$2F,$07,$FA,$AE,$BC,$99,$FA,$3C
	.dc.b	$9F,$D3,$C6,$36,$FE,$DF,$E0,$F9,$B5,$F7,$47,$F7,$2F,$3B,$1C,$1E
	.dc.b	$C8,$FE,$EB,$E7,$7F,$A4,$33,$ED,$7C,$32,$FF,$0A,$7C,$5F,$E8,$F3
	.dc.b	$ED,$1F,$E1,$7F,$73,$E2,$EF,$F3,$6F,$86,$29,$FF,$4C,$4E,$73,$F0
	.dc.b	$7B,$5E,$8B,$F2,$F0,$59,$F9,$FC,$3F,$26,$7F,$CE,$73,$F3,$B3,$FF
	.dc.b	$30,$B9,$AF,$56,$7F,$E2,$98,$F3,$57,$8B,$2B,$5F,$95,$9F,$8F,$F2
	.dc.b	$FF,$31,$FB,$11,$CB,$D2,$F8,$BC,$7F,$37,$6F,$CE,$6D,$E9,$8C,$63
	.dc.b	$FD,$8E,$31,$C3,$F2,$A3,$E0,$47,$F2,$2F,$45,$F1,$CF,$9A,$FE,$D5
	.dc.b	$F2,$DF,$8C,$D7,$F8,$35,$6A,$7F,$B7,$69,$FB,$78,$C7,$6F,$E3,$9E
	.dc.b	$7E,$8F,$E2,$DE,$BC,$97,$E5,$BD,$91,$F9,$BC,$7F,$95,$FB,$BF,$B2
	.dc.b	$51,$D3,$37,$9D,$FC,$C2,$31,$E2,$7E,$7F,$7F,$56,$FE,$9D,$58,$29
	.dc.b	$F1,$9E,$AA,$D4,$E7,$E9,$BE,$4F,$6B,$4F,$0D,$6B,$05,$3E,$67,$F9
	.dc.b	$86,$9E,$28,$C7,$9F,$FD,$36,$31,$2F,$8D,$C6,$1E,$0C,$7D,$4F,$9F
	.dc.b	$C2,$F9,$4F,$D1,$8F,$94,$FC,$67,$E3,$BE,$37,$D5,$CD,$F5,$B3,$F2
	.dc.b	$C7,$92,$7D,$BB,$F6,$AF,$44,$F6,$F0,$3E,$A7,$57,$49,$E5,$E8,$74
	.dc.b	$BF,$93,$8F,$C0,$F8,$5E,$95,$DF,$83,$E1,$74,$7F,$6B,$F2,$9F,$20
	.dc.b	$F9,$AF,$E3,$3F,$43,$F9,$BD,$F5,$39,$F1,$1F,$23,$FC,$FF,$0E,$38
	.dc.b	$7F,$20,$F2,$B1,$7B,$C2,$CF,$F8,$BD,$E6,$FF,$17,$AF,$0F,$83,$6B
	.dc.b	$F5,$11,$E1,$AF,$EE,$FE,$97,$E3,$63,$BE,$4C,$DB,$FF,$7C,$9C,$C6
	.dc.b	$31,$FE,$CD,$18,$97,$C9,$E3,$2F,$65,$EF,$E1,$3F,$BA,$5E,$F0,$7D
	.dc.b	$13,$97,$D3,$9F,$87,$5F,$06,$5F,$27,$B7,$F9,$BE,$3F,$33,$AB,$DD
	.dc.b	$D1,$EC,$BF,$6D,$F8,$37,$69,$F9,$A9,$61,$E1,$D3,$C7,$7F,$47,$1F
	.dc.b	$97,$E5,$76,$87,$CE,$FC,$BF,$82,$EB,$C9,$F6,$BD,$9F,$2B,$F7,$35
	.dc.b	$FF,$7E,$F8,$BF,$E2,$AD,$3B,$6F,$EC,$E4,$F9,$1B,$3E,$5F,$F0,$DC
	.dc.b	$1F,$E2,$93,$E2,$9F,$E1,$23,$E8,$1F,$E1,$12,$9F,$E2,$13,$E8,$9F
	.dc.b	$B5,$E8,$F8,$AE,$4C,$BF,$83,$C3,$F0,$2F,$C7,$8E,$C7,$80,$F1,$63
	.dc.b	$E9,$7F,$C8,$B3,$D9,$FF,$23,$CE,$DF,$F2,$54,$57,$FC,$91,$3C,$5F
	.dc.b	$F1,$14,$FF,$8E,$3D,$5F,$F0,$F4,$7F,$94,$67,$FD,$86,$BC,$5F,$F0
	.dc.b	$5B,$9A,$E1,$5F,$E1,$B8,$FE,$24,$2B,$8D,$7F,$88,$63,$81,$CE,$78
	.dc.b	$3C,$27,$71,$FA,$18,$FE,$39,$18,$7F,$C4,$D5,$FE,$4E,$F0,$3F,$BD
	.dc.b	$BF,$E4,$58,$FF,$30,$3E,$4A,$FF,$1D,$47,$58,$C4,$79,$EB,$BB,$EC
	.dc.b	$9F,$44,$EA,$3F,$C4,$CD,$7F,$AB,$B5,$F2,$B6,$B8,$CF,$B6,$7F,$C5
	.dc.b	$55,$F0,$BA,$9F,$08,$F9,$17,$C4,$7A,$30,$F9,$D5,$A9,$FE,$9D,$7C
	.dc.b	$6F,$F8,$B5,$6F,$3F,$DD,$C7,$EA,$A3,$17,$D9,$FF,$3B,$E7,$FE,$74
	.dc.b	$8D,$46,$8F,$96,$7D,$DE,$9F,$96,$F8,$AB,$CD,$F3,$9F,$E3,$58,$F6
	.dc.b	$E1,$E0,$C9,$3F,$D5,$A7,$E7,$1C,$63,$3E,$13,$C2,$CD,$F1,$BF,$EA
	.dc.b	$BF,$21,$8C,$FE,$04,$7F,$58,$F8,$8F,$EF,$5B,$7C,$1C,$B9,$C3,$E5
	.dc.b	$BE,$8E,$9D,$EF,$87,$CF,$77,$C3,$84,$63,$C7,$F4,$E5,$DA,$FF,$27
	.dc.b	$A7,$AD,$F9,$3B,$FF,$AF,$9F,$23,$F0,$B7,$DD,$79,$96,$9A,$D7,$CB
	.dc.b	$E2,$79,$31,$F7,$56,$B5,$C6,$1C,$70,$43,$E9,$7C,$B9,$7E,$17,$AF
	.dc.b	$ED,$FC,$67,$D6,$CB,$F9,$CF,$8B,$6B,$EB,$C7,$F8,$E3,$6F,$2C,$7D
	.dc.b	$8D,$6E,$DF,$4B,$CB,$F3,$70,$FC,$67,$ED,$DF,$F0,$E3,$D0,$7F,$1B
	.dc.b	$3E,$B9,$FE,$B5,$29,$FE,$C0,$E2,$FD,$EB,$FC,$6C,$E3,$CF,$E7,$EB
	.dc.b	$F2,$BE,$68,$FE,$54,$78,$0F,$04,$F2,$7F,$8E,$C7,$EE,$DF,$F6,$54
	.dc.b	$65,$FF,$6D,$35,$FE,$64,$D7,$FD,$84,$D7,$FD,$C6,$FF,$1A,$F8,$0F
	.dc.b	$F5,$2C,$7F,$12,$7F,$80,$3F,$8A,$7F,$90,$B9,$EB,$FE,$12,$D7,$F4
	.dc.b	$F1,$88,$D1,$8E,$CF,$D4,$C7,$FA,$E6,$3F,$81,$BF,$EA,$D7,$F4,$AF
	.dc.b	$FA,$C9,$B3,$97,$6E,$5A,$F1,$AF,$53,$FE,$B5,$8C,$BF,$EB,$4B,$FF
	.dc.b	$61,$3F,$C4,$1F,$F2,$27,$A8,$F9,$07,$D2,$BE,$2D,$E3,$FF,$11,$AD
	.dc.b	$4E,$74,$C5,$F0,$F8,$DF,$E1,$79,$AF,$79,$61,$E3,$16,$BF,$CC,$CF
	.dc.b	$ED,$A3,$C2,$FF,$10,$72,$FE,$EB,$DE,$F9,$1A,$FF,$50,$9F,$1E,$9E
	.dc.b	$6F,$0B,$FA,$4C,$63,$5F,$F0,$DE,$1F,$C4,$23,$13,$FE,$81,$8F,$E6
	.dc.b	$71,$8F,$8B,$D6,$5F,$0B,$0E,$90,$E8,$F3,$4E,$71,$FF,$0A,$CB,$E4
	.dc.b	$33,$C1,$F2,$67,$31,$FE,$1A,$D3,$D1,$7C,$74,$61,$C7,$CA,$F1,$CF
	.dc.b	$87,$B3,$E6,$E5,$F4,$F6,$F9,$FC,$DF,$16,$BF,$99,$EF,$FC,$C6,$7E
	.dc.b	$09,$88,$ED,$D3,$EC,$BF,$26,$7E,$03,$CF,$C1,$EE,$D9,$E7,$F8,$AB
	.dc.b	$7B,$7B,$7A,$B1,$6F,$90,$F2,$DF,$DB,$F9,$FF,$26,$1F,$07,$93,$E9
	.dc.b	$EB,$EF,$F0,$7F,$A0,$FA,$9F,$D7,$BE,$15,$F0,$EF,$9E,$FE,$BD,$39
	.dc.b	$8F,$E3,$1D,$1F,$E0,$E8,$C7,$26,$BC,$26,$1F,$C3,$FB,$3F,$D3,$D5
	.dc.b	$D6,$BF,$26,$FF,$6C,$E9,$FE,$B6,$BC,$4F,$EB,$6F,$F6,$09,$F0,$98
	.dc.b	$C3,$AE,$4F,$9F,$FF,$2F,$4F,$AA,$73,$1F,$A9,$8C,$65,$EC,$8F,$D6
	.dc.b	$6F,$8B,$E5,$BF,$E1,$EF,$FA,$AE,$3E,$FB,$FD,$97,$CE,$76,$DB,$E0
	.dc.b	$67,$DB,$04,$BE,$21,$DB,$FE,$0C,$C7,$95,$6A,$5F,$39,$9C,$3C,$F9
	.dc.b	$FF,$8B,$AF,$FC,$04,$FC,$68,$F5,$E1,$97,$C1,$D9,$E9,$D3,$CB,$5E
	.dc.b	$89,$F2,$67,$E2,$8E,$DB,$FC,$1D,$3C,$51,$8D,$BE,$0D,$7A,$7E,$0F
	.dc.b	$1E,$8F,$DB,$C1,$F1,$5E,$3F,$7F,$EB,$E7,$FE,$3F,$CB,$DD,$C7,$FE
	.dc.b	$12,$F8,$3F,$E0,$7C,$FF,$80,$AF,$A7,$D5,$E9,$CB,$C5,$AF,$C7,$F0
	.dc.b	$BF,$C6,$3E,$5B,$E3,$BF,$DB,$5F,$10,$FA,$4F,$F1,$74,$E7,$E1,$53
	.dc.b	$FD,$FF,$7E,$37,$C1,$DB,$FE,$E3,$E2,$FE,$43,$7D,$7E,$7B,$FC,$52
	.dc.b	$F7,$E1,$B5,$FE,$2A,$8E,$97,$FE,$24,$7D,$6F,$FA,$7A,$3C,$BF,$3D
	.dc.b	$FC,$5E,$3C,$39,$79,$23,$F4,$31,$C6,$77,$F1,$9F,$96,$8D,$E6,$A4
	.dc.b	$FE,$26,$7F,$C4,$B8,$7B,$EB,$53,$F8,$DE,$AC,$E1,$D3,$B9,$F4,$98
	.dc.b	$CF,$C5,$7F,$93,$FA,$57,$87,$E0,$D7,$BE,$09,$7C,$96,$76,$FF,$3F
	.dc.b	$ED,$F3,$DF,$93,$D1,$F1,$75,$F5,$C7,$B7,$4E,$FE,$05,$E0,$F3,$E7
	.dc.b	$D6,$BD,$FC,$1F,$52,$FF,$C6,$31,$6F,$EA,$E1,$FE,$EF,$8F,$C1,$FC
	.dc.b	$AF,$7A,$F9,$9E,$A7,$92,$BE,$07,$1F,$BF,$3E,$FB,$AB,$DD,$C5,$F0
	.dc.b	$4F,$C6,$E5,$FE,$27,$29,$FE,$BD,$3D,$E7,$F8,$9C,$FA,$26,$4C,$EF
	.dc.b	$F8,$87,$FD,$5A,$7E,$51,$FC,$43,$FE,$26,$F1,$7F,$D6,$31,$FE,$BE
	.dc.b	$8F,$C5,$1F,$EB,$E3,$A9,$FE,$BE,$29,$FE,$26,$E0,$F1,$23,$FD,$DA
	.dc.b	$7A,$0F,$F6,$3C,$F6,$7F,$D7,$27,$FB,$21,$7C,$C7,$FA,$AF,$0C,$CF
	.dc.b	$FA,$F6,$3F,$D5,$CE,$27,$6F,$FA,$86,$76,$EC,$FF,$52,$18,$8D,$FB
	.dc.b	$3F,$E2,$B8,$FF,$61,$57,$FB,$4D,$E2,$71,$8F,$E0,$CD,$32,$67,$D7
	.dc.b	$FF,$1C,$BF,$EC,$89,$FF,$5F,$DF,$FB,$6D,$FF,$5E,$47,$63,$FD,$6D
	.dc.b	$3F,$ED,$5B,$F1,$BF,$E3,$16,$BF,$E3,$19,$FF,$60,$4F,$78,$10,$95
	.dc.b	$97,$06,$6B,$E2,$FA,$DF,$B6,$F0,$FF,$C6,$79,$7C,$9C,$7E,$B4,$E7
	.dc.b	$C0,$F8,$D1,$8D,$7F,$94,$F7,$63,$45,$FF,$00,$F6,$3E,$3E,$7D,$AF
	.dc.b	$F1,$B5,$F0,$4D,$BF,$22,$3C,$7D,$77,$1F,$C3,$78,$DF,$E3,$27,$C1
	.dc.b	$AF,$F9,$23,$83,$CD,$8F,$28,$C6,$DD,$74,$F4,$53,$D8,$C6,$9F,$D0
	.dc.b	$37,$4E,$7C,$0F,$A5,$7C,$2D,$F3,$98,$C3,$85,$F3,$F8,$AF,$ED,$1A
	.dc.b	$F9,$32,$E7,$82,$7F,$DA,$7A,$FE,$3F,$2E,$91,$FE,$D6,$D3,$19,$74
	.dc.b	$BE,$1E,$DF,$F5,$4E,$1F,$91,$BF,$F8,$BF,$2D,$BF,$33,$4F,$AB,$1F
	.dc.b	$1C,$F8,$5A,$78,$BA,$7D,$15,$F1,$F1,$F2,$57,$E4,$CF,$91,$E9,$7D
	.dc.b	$9D,$BF,$9E,$C5,$E7,$E9,$1F,$2B,$D3,$F9,$38,$7D,$1F,$86,$B9,$FC
	.dc.b	$0E,$1F,$F7,$2E,$BF,$0F,$8B,$E3,$7C,$27,$7C,$BE,$0F,$0B,$A3,$F6
	.dc.b	$5E,$8B,$E3,$DE,$77,$8B,$87,$F4,$93,$98,$FE,$A9,$1E,$98,$E1,$EE
	.dc.b	$7C,$38,$47,$37,$0F,$ED,$B9,$BF,$81,$D7,$E2,$E7,$77,$E0,$F3,$BF
	.dc.b	$33,$1F,$95,$BF,$04,$7F,$A3,$7C,$6C,$56,$F1,$7C,$B6,$31,$FF,$22
	.dc.b	$C3,$E6,$7F,$B2,$67,$FC,$4B,$2E,$73,$9F,$94,$F6,$E3,$CE,$D1,$FE
	.dc.b	$81,$D3,$E2,$CE,$61,$F4,$DF,$06,$30,$FF,$A8,$71,$ED,$1F,$2A,$BF
	.dc.b	$69,$8F,$BC,$F0,$BE,$7D,$7F,$D5,$1E,$05,$CF,$D1,$3E,$4D,$3D,$7A
	.dc.b	$79,$3C,$BF,$04,$63,$5F,$1B,$E6,$3F,$CC,$BE,$07,$ED,$33,$FB,$1E
	.dc.b	$A7,$8B,$5F,$BB,$DB,$6B,$EA,$FC,$67,$E6,$6F,$E2,$E7,$FE,$A7,$E0
	.dc.b	$F5,$E5,$E2,$E5,$F2,$70,$FF,$AF,$EB,$E3,$FA,$9E,$48,$C6,$DF,$1B
	.dc.b	$3F,$E7,$DE,$4B,$E4,$1F,$0B,$F8,$FE,$5F,$E1,$33,$98,$6F,$FC,$6E
	.dc.b	$31,$C1,$6D,$F1,$3F,$8F,$9E,$98,$5E,$FF,$D9,$47,$F7,$FE,$6F,$09
	.dc.b	$8B,$F0,$F2,$72,$8F,$E4,$53,$E1,$8E,$BC,$D8,$AD,$E2,$FA,$4C,$6B
	.dc.b	$FA,$78,$C7,$8F,$FD,$27,$1E,$51,$BE,$AC,$61,$CB,$07,$D3,$7F,$18
	.dc.b	$D1,$87,$09,$CF,$A9,$FC,$16,$31,$4F,$A5,$70,$7C,$CC,$7A,$4E,$73
	.dc.b	$C6,$BF,$B1,$CF,$CE,$7C,$93,$E4,$6B,$F0,$E3,$D5,$E4,$7B,$38,$3E
	.dc.b	$6E,$CE,$0F,$4D,$7B,$E3,$D3,$87,$97,$3F,$0F,$81,$6B,$5A,$79,$EB
	.dc.b	$C7,$8F,$48,$F5,$4E,$2B,$F0,$37,$FC,$B6,$F7,$E9,$62,$BE,$77,$CB
	.dc.b	$F9,$76,$FA,$58,$FC,$3C,$5D,$DF,$EE,$7F,$29,$F2,$6F,$99,$7D,$2F
	.dc.b	$F8,$37,$E2,$3F,$C4,$15,$A9,$FE,$79,$7C,$6F,$C4,$F3,$E1,$FE,$C9
	.dc.b	$7C,$38,$AF,$07,$FA,$60,$B8,$F3,$F1,$FF,$DE,$98,$73,$CB,$F8,$AF
	.dc.b	$C8,$7F,$88,$F1,$F0,$E7,$E5,$B4,$FF,$0F,$C7,$FC,$69,$87,$19,$CF
	.dc.b	$BF,$FE,$32,$8C,$53,$DB,$FD,$B6,$3D,$32,$EF,$1F,$D2,$AB,$6F,$FB
	.dc.b	$A3,$A3,$AD,$6F,$1F,$0C,$FF,$BB,$3E,$13,$39,$FF,$AD,$34,$F3,$67
	.dc.b	$DF,$C4,$6B,$FE,$D2,$D3,$AC,$63,$5F,$A7,$A7,$5C,$7E,$01,$CE,$3F
	.dc.b	$32,$E3,$6F,$A4,$FD,$53,$C1,$D7,$FE,$18,$AF,$8F,$14,$FF,$63,$F5
	.dc.b	$FD,$0D,$3C,$BB,$7C,$5A,$F6,$6B,$E7,$D7,$D7,$97,$D5,$8C,$65,$E6
	.dc.b	$CB,$FB,$11,$CF,$3F,$E9,$7E,$CF,$97,$3F,$AD,$F0,$76,$7E,$DD,$F7
	.dc.b	$BF,$67,$0F,$CA,$F0,$FF,$CD,$91,$D7,$3F,$06,$CF,$33,$3C,$5F,$23
	.dc.b	$1F,$9F,$E5,$FF,$BE,$3E,$44,$7C,$37,$1E,$DD,$F8,$6F,$EE,$39,$F0
	.dc.b	$FA,$2F,$C9,$D7,$EA,$63,$D9,$E8,$FF,$98,$9F,$74,$75,$F8,$AF,$EF
	.dc.b	$39,$B9,$FF,$D0,$53,$9F,$83,$FA,$EE,$F3,$FF,$66,$61,$CB,$3F,$2E
	.dc.b	$3F,$EC,$0F,$94,$F8,$57,$BE,$7F,$F5,$35,$FB,$2F,$78,$27,$F9,$EC
	.dc.b	$E7,$66,$3F,$EC,$A9,$FF,$09,$19,$71,$E4,$32,$C6,$1C,$E3,$1D,$FE
	.dc.b	$5C,$39,$E3,$EA,$CB,$E5,$E5,$F2,$67,$F1,$59,$A7,$DB,$5F,$53,$5F
	.dc.b	$87,$1F,$32,$3E,$6B,$EA,$BF,$E5,$9B,$7C,$99,$FF,$13,$CE,$72,$F3
	.dc.b	$76,$78,$F6,$F4,$EC,$F1,$7E,$9E,$5D,$35,$F8,$3C,$FE,$77,$E1,$F0
	.dc.b	$7F,$C0,$1D,$7B,$6F,$F3,$74,$FF,$77,$ED,$D6,$BE,$17,$CD,$7E,$E3
	.dc.b	$CB,$FF,$77,$FD,$28,$FD,$63,$8C,$EF,$27,$D6,$C3,$F6,$B2,$F8,$0F
	.dc.b	$E3,$AF,$E3,$A3,$F8,$1F,$AF,$F8,$BD,$7E,$04,$F6,$C7,$F8,$C5,$78
	.dc.b	$70,$F0,$43,$C5,$CD,$F0,$9C,$C3,$E2,$3F,$55,$87,$38,$C7,$5F,$D3
	.dc.b	$C7,$FB,$2D,$F0,$BF,$F3,$4C,$7E,$E6,$F8,$47,$F8,$37,$4F,$16,$5F
	.dc.b	$07,$1E,$DE,$73,$E6,$73,$3E,$5B,$FF,$0A,$7C,$E7,$F1,$BE,$BF,$F7
	.dc.b	$26,$5D,$30,$4B,$E3,$7C,$1A,$FC,$DE,$3F,$97,$97,$F1,$5C,$13,$F2
	.dc.b	$78,$3E,$06,$F9,$3E,$6C,$7C,$2E,$DF,$82,$FB,$77,$F8,$17,$E1,$AF
	.dc.b	$3C,$7A,$B1,$BB,$7A,$27,$C7,$97,$BE,$3D,$7D,$FF,$17,$87,$BC,$FE
	.dc.b	$0F,$8B,$EF,$E7,$F5,$BC,$1B,$3F,$1F,$CE,$F1,$BC,$F8,$BE,$0F,$5F
	.dc.b	$F6,$EF,$84,$6B,$FE,$B8,$F8,$07,$CF,$FF,$3C,$8F,$EA,$D3,$A9,$FF
	.dc.b	$5E,$CF,$FB,$4A,$31,$1B,$9D,$9C,$CD,$1F,$EC,$4F,$6B,$FD,$A9,$3C
	.dc.b	$E2,$9E,$1A,$FF,$62,$BF,$CC,$0F,$ED,$31,$DA,$3F,$D9,$3D,$59,$9F
	.dc.b	$51,$C7,$E1,$BF,$C6,$3C,$7E,$2C,$7C,$F1,$FE,$34,$F9,$3C,$47,$BA
	.dc.b	$31,$CF,$FC,$EB,$2E,$D1,$BF,$0B,$18,$F5,$E8,$63,$FC,$07,$0F,$2C
	.dc.b	$E6,$3F,$82,$7A,$58,$CB,$C3,$7C,$30,$FF,$28,$78,$DE,$89,$CD,$F0
	.dc.b	$F9,$FC,$47,$D2,$EB,$F7,$65,$87,$E5,$C7,$48,$F5,$F0,$7F,$BB,$32
	.dc.b	$FE,$7B,$D9,$FE,$F5,$9F,$26,$5E,$AB,$EF,$9A,$3C,$D1,$F0,$EF,$CD
	.dc.b	$A3,$E7,$FC,$79,$3E,$93,$C7,$9F,$5C,$3E,$6F,$27,$D3,$8C,$47,$3F
	.dc.b	$27,$CB,$C5,$F2,$F5,$EF,$8F,$9F,$87,$E3,$F1,$7D,$AF,$2D,$F3,$4F
	.dc.b	$84,$FE,$77,$C7,$98,$7C,$3F,$E3,$2B,$7D,$57,$E7,$3F,$17,$1F,$E8
	.dc.b	$7E,$0F,$85,$F1,$1F,$C7,$FA,$BF,$3F,$3D,$B3,$ED,$9F,$A6,$73,$9F
	.dc.b	$87,$E1,$7F,$39,$3E,$1A,$FF,$65,$FB,$FB,$86,$1C,$EB,$FC,$F7,$E2
	.dc.b	$FE,$8A,$31,$E2,$FF,$39,$AF,$C5,$C6,$3E,$5D,$EF,$E3,$3F,$05,$A7
	.dc.b	$F8,$B5,$FE,$3A,$F8,$79,$1C,$23,$C9,$97,$FA,$2A,$93,$27,$37,$CE
	.dc.b	$33,$D5,$8A,$DC,$7F,$82,$78,$3F,$11,$8F,$C5,$C3,$EF,$EF,$F1,$1F
	.dc.b	$A0,$7C,$EC,$FE,$33,$F0,$E7,$DF,$E1,$7C,$19,$F5,$E3,$78,$3F,$DE
	.dc.b	$7A,$78,$F0,$F6,$56,$B3,$F4,$FC,$57,$CF,$C7,$C3,$5F,$3A,$BF,$99
	.dc.b	$67,$F6,$B7,$F1,$65,$FE,$FE,$7C,$53,$E0,$E4,$F9,$93,$E0,$D3,$E2
	.dc.b	$C7,$82,$31,$9F,$CE,$DB,$E0,$DF,$8E,$3F,$4D,$1F,$AB,$F1,$3F,$D2
	.dc.b	$7F,$0B,$FD,$9B,$4F,$E0,$31,$87,$CE,$FF,$A6,$3C,$4F,$E4,$2F,$FA
	.dc.b	$6A,$2F,$4F,$F2,$08,$FC,$A4,$1E,$B5,$CB,$DC,$3F,$EC,$EE,$0F,$C8
	.dc.b	$4F,$8B,$C9,$FD,$1D,$FF,$30,$8C,$75,$FF,$76,$E2,$FE,$7B,$18,$F9
	.dc.b	$7D,$65,$D3,$0F,$D5,$C7,$FB,$17,$0F,$5C,$E7,$04,$BE,$5B,$F9,$1E
	.dc.b	$1C,$DE,$1B,$7C,$5B,$E1,$5B,$9E,$7A,$7F,$B9,$B0,$E7,$1C,$FA,$3F
	.dc.b	$9C,$E3,$CE,$1F,$21,$8C,$7E,$0E,$FF,$0C,$F9,$7E,$4F,$C9,$8F,$56
	.dc.b	$BF,$11,$F6,$72,$DD,$3C,$1A,$F8,$FB,$BA,$7C,$5F,$D4,$F2,$7F,$A8
	.dc.b	$FB,$1F,$4B,$7F,$99,$F4,$3C,$79,$3F,$87,$F7,$7C,$8C,$FD,$5E,$75
	.dc.b	$9F,$93,$D3,$F1,$37,$F8,$3F,$31,$6F,$C3,$CB,$FE,$29,$E2,$F6,$46
	.dc.b	$27,$FA,$4F,$9C,$EA,$7C,$57,$F2,$0E,$BF,$E7,$53,$9F,$8B,$FE,$D6
	.dc.b	$EF,$9E,$FC,$8F,$F6,$A8,$76,$5D,$31,$97,$4C,$BF,$57,$5D,$DE,$FB
	.dc.b	$FF,$1B,$8C,$43,$9F,$15,$C7,$6F,$E9,$AF,$F8,$B0,$C7,$B9,$FE,$00
	.dc.b	$BE,$32,$8F,$E9,$D8,$FB,$AB,$7F,$2B,$FD,$B2,$F7,$48,$DE,$FF,$02
	.dc.b	$B7,$1C,$72,$F8,$B8,$F5,$8E,$1D,$1F,$E0,$2C,$7D,$9E,$2F,$BB,$0F
	.dc.b	$97,$9D,$8F,$93,$1F,$13,$1F,$9F,$DD,$EE,$7A,$D7,$AE,$7D,$BB,$F9
	.dc.b	$DF,$4B,$E3,$9F,$DA,$EB,$E9,$CF,$F7,$5A,$FD,$8E,$EE,$DC,$7E,$3B
	.dc.b	$F4,$F8,$3D,$5F,$5F,$91,$5E,$7F,$2D,$DD,$F7,$E5,$F1,$63,$C1,$C5
	.dc.b	$F0,$73,$FB,$F9,$B9,$7F,$C7,$7E,$07,$E2,$BE,$2F,$FC,$13,$C2,$AD
	.dc.b	$29,$FC,$0B,$FD,$F2,$B8,$E7,$C2,$38,$47,$4E,$8C,$5F,$2F,$0F,$FD
	.dc.b	$15,$3C,$31,$FE,$49,$7C,$0E,$D8,$21,$F1,$5D,$E7,$30,$F6,$7F,$D3
	.dc.b	$51,$FE,$54,$C7,$CF,$9F,$AE,$7F,$B5,$4F,$FB,$C3,$1F,$04,$7F,$67
	.dc.b	$E7,$FF,$24,$5F,$0C,$BC,$13,$FD,$AF,$E2,$B3,$7C,$3C,$BF,$EE,$3C
	.dc.b	$9C,$F2,$F1,$DF,$3F,$6F,$1A,$78,$6F,$77,$E5,$C3,$96,$88,$9D,$1F
	.dc.b	$06,$F8,$4E,$65,$F4,$39,$8F,$97,$7F,$63,$3A,$78,$8F,$F1,$63,$EE
	.dc.b	$AB,$1E,$6C,$6E,$1D,$A7,$C5,$5E,$1D,$7F,$9C,$DF,$CF,$AF,$E1,$BE
	.dc.b	$2F,$E1,$37,$F9,$BD,$3E,$09,$FC,$44,$EE,$FE,$4E,$5F,$E4,$CC,$7D
	.dc.b	$FE,$87,$C4,$9C,$E9,$F0,$63,$E2,$67,$E5,$8F,$89,$BE,$32,$F3,$6B
	.dc.b	$FD,$B3,$1F,$EA,$BF,$0E,$F9,$9F,$24,$63,$B6,$E5,$FD,$96,$BE,$2F
	.dc.b	$95,$FE,$DB,$8F,$C7,$78,$5F,$92,$75,$F3,$3F,$CA,$70,$EA,$BE,$23
	.dc.b	$F2,$16,$C1,$2E,$CF,$C8,$E1,$F9,$9A,$D4,$E7,$E8,$BE,$8C,$62,$91
	.dc.b	$FD,$A6,$BE,$0C,$E6,$1E,$27,$F4,$19,$E9,$87,$82,$F8,$CF,$FB,$B7
	.dc.b	$C7,$E3,$39,$BE,$31,$FE,$F1,$EA,$E3,$97,$8E,$F8,$7C,$5E,$32,$F6
	.dc.b	$9F,$1B,$1F,$23,$FA,$03,$FC,$1D,$F0,$BF,$2B,$9B,$58,$7E,$66,$7D
	.dc.b	$B5,$E5,$FA,$0B,$3F,$07,$3E,$5B,$7B,$7E,$82,$DF,$9F,$4F,$F6,$47
	.dc.b	$6F,$F0,$EA,$F8,$51,$FB,$DC,$BD,$B8,$FA,$71,$76,$63,$87,$E1,$C6
	.dc.b	$F1,$E5,$97,$58,$0F,$98,$FE,$1D,$E5,$FF,$0C,$F6,$7F,$0E,$FF,$0E
	.dc.b	$E3,$BE,$2F,$EB,$23,$F8,$AE,$BF,$96,$7F,$10,$6F,$8B,$F8,$A3,$C7
	.dc.b	$8E,$D7,$F6,$C7,$FD,$07,$A7,$F2,$A9,$D7,$C0,$7E,$D1,$BD,$5F,$3D
	.dc.b	$CF,$87,$70,$E5,$5F,$C0,$AB,$F8,$85,$69,$FE,$C3,$D5,$FE,$F4,$8C
	.dc.b	$47,$FA,$3B,$0F,$05,$FF,$7E,$ED,$F2,$EA,$CF,$FC,$A9,$5A,$8C,$65
	.dc.b	$FE,$04,$D7,$BE,$1E,$2D,$3F,$D8,$59,$3B,$3E,$5C,$FF,$78,$AF,$67
	.dc.b	$8D,$F2,$79,$7E,$54,$FF,$83,$BE,$2B,$C3,$AF,$FB,$16,$FD,$DA,$78
	.dc.b	$3C,$9B,$2D,$7F,$90,$F4,$F8,$AB,$CB,$B7,$EC,$B0,$FA,$99,$79,$76
	.dc.b	$FF,$26,$F1,$7C,$38,$C7,$4E,$B0,$E3,$5F,$ED,$B7,$FA,$F6,$0C,$1D
	.dc.b	$BF,$E2,$4D,$FF,$90,$4E,$61,$CD,$FD,$72,$31,$76,$F8,$96,$BF,$C5
	.dc.b	$BE,$D7,$28,$EB,$3F,$ED,$8D,$7F,$DC,$B7,$CA,$BF,$D0,$15,$FD,$52
	.dc.b	$31,$E9,$FD,$66,$FA,$9F,$EE,$97,$3F,$11,$FB,$3A,$FE,$C5,$18,$F5
	.dc.b	$F9,$47,$83,$06,$0B,$53,$AB,$E1,$5F,$C4,$38,$56,$E7,$51,$FE,$9E
	.dc.b	$C3,$B5,$7C,$9F,$A5,$E1,$AF,$C0,$E0,$F6,$DF,$2D,$7D,$0F,$4C,$ED
	.dc.b	$7C,$6C,$FC,$9E,$07,$A2,$31,$3E,$6C,$BE,$B6,$FD,$8F,$F8,$03,$7E
	.dc.b	$58,$FB,$5F,$4D,$9F,$2B,$E8,$61,$F9,$5C,$BC,$3E,$07,$F8,$5D,$E5
	.dc.b	$F0,$3A,$C9,$C4,$F6,$FD,$38,$78,$BE,$0B,$C9,$5E,$0D,$BE,$4B,$F2
	.dc.b	$3C,$1F,$EE,$F9,$7B,$79,$7F,$D6,$CF,$D4,$31,$BF,$EB,$67,$78,$FB
	.dc.b	$AB,$E9,$4E,$FC,$16,$BF,$CD,$30,$ED,$CD,$F9,$79,$CC,$7E,$35,$FF
	.dc.b	$7F,$3F,$43,$E0,$7D,$D9,$79,$27,$9F,$1E,$F0,$FC,$1E,$9F,$5A,$73
	.dc.b	$DB,$67,$F0,$FE,$CF,$F5,$FC,$BF,$17,$5C,$A3,$FC,$CF,$1E,$43,$F8
	.dc.b	$24,$F0,$F9,$CF,$F1,$AE,$5E,$18,$C7,$A5,$C9,$BB,$71,$C7,$A5,$AB
	.dc.b	$F6,$9D,$DF,$0E,$F8,$E3,$FD,$0B,$4F,$CB,$61,$C2,$F7,$F0,$F8,$BF
	.dc.b	$C0,$5F,$5C,$3F,$6D,$7F,$13,$83,$E5,$57,$BA,$BF,$2B,$3E,$DE,$8B
	.dc.b	$3E,$B9,$F4,$70,$7F,$AC,$74,$FD,$84,$78,$BE,$6F,$AD,$7E,$26,$3E
	.dc.b	$55,$78,$9F,$23,$C7,$E3,$39,$F0,$B0,$F8,$1B,$FC,$3C,$3D,$F9,$7C
	.dc.b	$1C,$3C,$1B,$FF,$17,$F8,$6F,$C6,$70,$BA,$3F,$03,$AF,$F0,$08,$7C
	.dc.b	$4F,$F1,$57,$FC,$45,$C9,$F9,$6F,$3F,$E1,$57,$F0,$1B,$7C,$37,$F1
	.dc.b	$7B,$E9,$F2,$5D,$2B,$BD,$21,$F2,$7F,$BE,$AD,$F2,$7C,$4B,$FE,$15
	.dc.b	$18,$F4,$BF,$80,$BA,$F9,$6F,$C2,$F4,$7F,$59,$CB,$A6,$09,$FF,$13
	.dc.b	$E1,$F8,$DB,$E1,$EB,$7F,$37,$F1,$BC,$7A,$30,$E1,$3F,$CE,$FE,$7F
	.dc.b	$31,$8C,$11,$FE,$9B,$E0,$FD,$BE,$3F,$26,$7E,$07,$8F,$F1,$DF,$91
	.dc.b	$A7,$BD,$F7,$3F,$07,$B6,$E7,$E1,$9F,$6C,$FA,$B5,$B3,$E5,$7D,$7E
	.dc.b	$47,$91,$F6,$E7,$FD,$DF,$7F,$85,$97,$D1,$C3,$E1,$FA,$9C,$B6,$F8
	.dc.b	$9E,$BF,$DD,$9E,$DC,$1F,$DD,$F5,$F9,$F1,$8D,$FE,$43,$A8,$7D,$0F
	.dc.b	$86,$BE,$37,$6F,$FA,$FF,$D3,$FE,$C5,$E5,$FE,$C6,$FF,$D0,$11,$8D
	.dc.b	$3F,$D7,$6F,$F6,$7F,$13,$F8,$6B,$C3,$B7,$F3,$D5,$CC,$FF,$18,$2F
	.dc.b	$33,$9C,$63,$E0,$FF,$0C,$EF,$1F,$E0,$11,$FE,$6F,$8F,$EA,$67,$FA
	.dc.b	$F6,$FB,$47,$F9,$CF,$97,$33,$C7,$04,$FF,$9D,$FC,$EF,$E3,$13,$C4
	.dc.b	$F0,$75,$A3,$FC,$F5,$C3,$FC,$E7,$1E,$B3,$9E,$4F,$EC,$9D,$7C,$74
	.dc.b	$F2,$CE,$6F,$F1,$78,$F8,$72,$F0,$61,$FD,$AF,$9D,$AE,$70,$E1,$0F
	.dc.b	$90,$F8,$1E,$C7,$CC,$C7,$FC,$AC,$5C,$FF,$81,$BE,$E3,$B1,$7B,$5B
	.dc.b	$7A,$39,$3E,$1C,$62,$7D,$FC,$18,$AF,$6B,$F0,$34,$F6,$79,$1D,$F8
	.dc.b	$BE,$4F,$2F,$5B,$F9,$FE,$B5,$C7,$9E,$BF,$EE,$89,$E7,$ED,$D9,$EF
	.dc.b	$C5,$F9,$3D,$7C,$5B,$FC,$8D,$BF,$D5,$1E,$1F,$CF,$BD,$F2,$FE,$51
	.dc.b	$F8,$D3,$F5,$38,$FF,$D4,$78,$72,$DF,$EE,$F2,$7C,$AF,$91,$FC,$77
	.dc.b	$9D,$FE,$D8,$C7,$CC,$FF,$9E,$1C,$9F,$35,$7C,$67,$E9,$E3,$D1,$18
	.dc.b	$EB,$FC,$AB,$AD,$BF,$D7,$B3,$DA,$BF,$D1,$5F,$3B,$D9,$BD,$1F,$36
	.dc.b	$57,$C6,$98,$7F,$62,$E3,$FF,$27,$4E,$63,$FC,$E5,$5F,$0E,$31,$CB
	.dc.b	$FE,$2E,$8F,$06,$4C,$39,$CF,$F9,$5F,$CB,$FC,$E5,$F0,$BF,$E2,$DB
	.dc.b	$FE,$06,$31,$F3,$2A,$E7,$87,$2B,$4C,$BE,$1A,$FE,$4A,$7D,$98,$79
	.dc.b	$DF,$8F,$1F,$0B,$3F,$EC,$35,$C3,$3F,$34,$7B,$30,$F2,$D6,$F3,$FF
	.dc.b	$37,$43,$E1,$BE,$CF,$C7,$7D,$5F,$27,$F4,$5F,$07,$FC,$33,$4F,$97
	.dc.b	$8F,$FA,$E7,$E8,$FE,$86,$9D,$BE,$03,$C6,$F8,$34,$FC,$87,$17,$CE
	.dc.b	$BF,$06,$1F,$E1,$9D,$FC,$DC,$3F,$07,$7F,$F7,$D4,$7E,$9F,$A9,$F0
	.dc.b	$9F,$B3,$7F,$8D,$DF,$F1,$B7,$C8,$6B,$8F,$F8,$DB,$78,$3F,$4C,$E7
	.dc.b	$E7,$AD,$FE,$11,$70,$FE,$1E,$75,$7F,$85,$C1,$18,$E1,$FD,$63,$AD
	.dc.b	$3F,$DC,$FB,$7F,$A5,$E3,$19,$BC,$5F,$E3,$D1,$8F,$85,$BA,$F8,$73
	.dc.b	$FC,$55,$A7,$F5,$B8,$73,$AD,$63,$FE,$EE,$9D,$6D,$F1,$67,$F3,$BF
	.dc.b	$0B,$EC,$C5,$BF,$E8,$36,$F3,$D7,$27,$F7,$DA,$7F,$1F,$7E,$3E,$BF
	.dc.b	$C7,$FD,$9F,$97,$5F,$B3,$9D,$BF,$19,$ED,$F0,$9D,$F0,$D6,$7E,$3C
	.dc.b	$BC,$27,$E9,$7A,$B8,$B8,$D9,$EB,$62,$FF,$6B,$0E,$16,$FF,$1A,$F8
	.dc.b	$43,$E4,$31,$7C,$23,$E4,$57,$F5,$DF,$0F,$FD,$87,$AF,$F5,$C7,$D4
	.dc.b	$FF,$1D,$7F,$34,$7F,$B2,$38,$75,$F0,$7F,$7B,$CA,$FE,$59,$39,$F2
	.dc.b	$EC,$FE,$09,$FF,$44,$DF,$38,$C5,$FA,$63,$0F,$F4,$B3,$E9,$B8,$9F
	.dc.b	$C5,$E5,$F8,$58,$FE,$C5,$7D,$EB,$EC,$F8,$BB,$6E,$3A,$C7,$BE,$1F
	.dc.b	$E2,$8E,$7C,$57,$78,$FF,$15,$FA,$7F,$AA,$9F,$F7,$56,$3D,$67,$3F
	.dc.b	$2D,$F1,$F0,$F4,$EB,$FD,$1E,$B5,$97,$86,$FF,$C6,$BE,$27,$6D,$BD
	.dc.b	$3A,$FF,$AB,$39,$7F,$AE,$61,$D6,$1A,$FF,$1D,$9C,$FC,$BF,$F6,$BA
	.dc.b	$FD,$CC,$70,$D3,$EA,$F0,$78,$FA,$7E,$9F,$0F,$FA,$B6,$3C,$CB,$C6
	.dc.b	$FA,$6F,$F7,$7D,$7F,$9F,$63,$F5,$EE,$E1,$F5,$3A,$7D,$8F,$C8,$F4
	.dc.b	$BD,$9B,$79,$BE,$7D,$AF,$F8,$5B,$5F,$E0,$7D,$7F,$DF,$51,$AF,$46
	.dc.b	$FC,$77,$FA,$E9,$CF,$85,$FE,$C0,$9F,$F0,$14,$63,$DE,$F6,$E1,$D6
	.dc.b	$BF,$D9,$3A,$75,$8C,$4A,$3F,$D2,$91,$87,$FD,$B3,$7F,$E3,$08,$C7
	.dc.b	$27,$EC,$30,$ED,$5F,$43,$93,$FC,$1F,$18,$94,$7E,$3E,$FD,$77,$CA
	.dc.b	$3F,$88,$47,$FB,$9E,$B8,$47,$F1,$3F,$07,$13,$FC,$9B,$1F,$36,$1D
	.dc.b	$3C,$2F,$F0,$34,$FC,$53,$59,$78,$E3,$FC,$FB,$E8,$F8,$32,$EB,$1F
	.dc.b	$E8,$0F,$77,$17,$C3,$0E,$9C,$3F,$EA,$6C,$FA,$6D,$F2,$27,$36,$F7
	.dc.b	$B3,$E0,$7D,$7C,$76,$5E,$0F,$95,$D1,$F5,$23,$B7,$0F,$F2,$FE,$DF
	.dc.b	$6E,$7D,$63,$F9,$96,$B8,$E2,$F7,$C6,$3D,$2F,$3D,$6F,$83,$D5,$87
	.dc.b	$9B,$DA,$F0,$71,$7E,$06,$3E,$F3,$F8,$3C,$2E,$7F,$1B,$C2,$F2,$63
	.dc.b	$ED,$BF,$B3,$3F,$0F,$E0,$2C,$F9,$7C,$2F,$C1,$63,$F5,$BC,$4B,$7E
	.dc.b	$CD,$3F,$D5,$7F,$37,$E6,$DF,$D3,$1F,$3F,$E1,$BE,$B7,$85,$FE,$D0
	.dc.b	$C7,$F3,$BE,$0F,$E6,$F1,$FE,$BF,$D1,$EA,$7F,$B0,$FB,$9F,$EC,$F7
	.dc.b	$EB,$DF,$FB,$77,$8F,$D6,$FF,$B7,$F3,$D7,$3F,$EA,$E3,$1F,$3B,$F4
	.dc.b	$5D,$7C,$0F,$FB,$66,$75,$8A,$FF,$DB,$38,$7C,$2B,$DB,$FE,$78,$C7
	.dc.b	$F2,$71,$8E,$5F,$82,$EB,$B3,$FC,$59,$3E,$6C,$3A,$65,$FE,$D3,$9F
	.dc.b	$F6,$44,$FC,$29,$CE,$1F,$C6,$3C,$F7,$B7,$C7,$3F,$1E,$1F,$E3,$4F
	.dc.b	$73,$19,$73,$B4,$3B,$7A,$5F,$09,$CF,$C2,$7D,$0E,$DF,$99,$B7,$CF
	.dc.b	$3F,$DB,$DB,$7C,$5D,$BF,$A5,$3F,$05,$FE,$18,$FB,$9F,$8D,$DF,$BC
	.dc.b	$7F,$DF,$F3,$B7,$FE,$00,$C5,$D3,$FD,$E5,$9F,$19,$F5,$D7,$CA,$BF
	.dc.b	$83,$AF,$F9,$42,$3D,$BB,$FC,$3F,$0A,$CF,$C1,$ED,$F6,$E5,$E0,$BE
	.dc.b	$58,$FC,$6D,$6C,$FC,$5E,$0F,$F5,$7F,$1F,$F2,$DF,$27,$FD,$04,$DC
	.dc.b	$3F,$1D,$0F,$A2,$BF,$2B,$FE,$D2,$9F,$E1,$92,$87,$47,$EA,$A3,$1C
	.dc.b	$1F,$ED,$8C,$3B,$52,$1D,$BF,$C0,$6B,$76,$E0,$FD,$7C,$62,$2D,$F1
	.dc.b	$7B,$F1,$7E,$BA,$75,$EB,$DA,$D5,$3E,$23,$A6,$9F,$C7,$2F,$A4,$3E
	.dc.b	$3F,$F9,$5C,$61,$FF,$0F,$CF,$BA,$F9,$D6,$FD,$BE,$51,$8A,$FF,$53
	.dc.b	$F9,$9B,$DB,$E1,$E5,$E1,$9F,$F5,$5F,$17,$C8,$F2,$BE,$5E,$7F,$1F
	.dc.b	$1F,$F7,$8B,$E9,$C3,$06,$1E,$EF,$CE,$7E,$1B,$F4,$39,$2E,$3F,$4B
	.dc.b	$4E,$4D,$CB,$FD,$77,$7E,$58,$C7,$B3,$FE,$3D,$E3,$FA,$B8,$FA,$30
	.dc.b	$F8,$92,$E9,$E1,$C9,$DB,$0F,$99,$DD,$F2,$35,$F8,$BF,$39,$72,$E5
	.dc.b	$AF,$F8,$C7,$47,$C1,$FF,$78,$D7,$FA,$26,$DF,$DF,$A3,$FB,$CE,$5F
	.dc.b	$D1,$5F,$EF,$7F,$09,$65,$E8,$3D,$CF,$F6,$7C,$71,$32,$FF,$00,$7F
	.dc.b	$E0,$05,$7F,$81,$21,$F0,$5F,$ED,$6C,$BC,$75,$C2,$18,$70,$8C,$57
	.dc.b	$F1,$07,$FB,$F5,$F3,$AF,$16,$3F,$86,$9F,$EF,$33,$9A,$FE,$2F,$D9
	.dc.b	$C2,$79,$D7,$EA,$FA,$FF,$C3,$31,$E9,$CF,$D9,$8F,$4B,$FC,$C4,$F4
	.dc.b	$38,$BF,$9B,$F3,$B1,$5C,$31,$EF,$AF,$FA,$27,$5F,$6C,$E7,$3F,$DD
	.dc.b	$4F,$FA,$E3,$B7,$C1,$7F,$10,$E3,$EE,$BE,$0B,$C9,$A7,$D1,$C7,$E8
	.dc.b	$3F,$E2,$5D,$FE,$66,$16,$3E,$51,$F2,$67,$C8,$FB,$8F,$3E,$1D,$F2
	.dc.b	$FF,$10,$47,$8B,$6F,$81,$9F,$FB,$32,$31,$5E,$4F,$86,$BF,$3B,$CE
	.dc.b	$73,$87,$F1,$5F,$07,$E5,$EF,$F0,$30,$ED,$C3,$FC,$07,$B7,$C3,$A7
	.dc.b	$6E,$1F,$E9,$77,$F8,$2F,$33,$F2,$6D,$F1,$FF,$21,$0F,$66,$FB,$9F
	.dc.b	$90,$AE,$0F,$E2,$36,$FC,$45,$7F,$AF,$BA,$3F,$A3,$C7,$93,$97,$FB
	.dc.b	$DD,$7E,$3B,$0D,$BF,$90,$1F,$0F,$ED,$C7,$AE,$1F,$EC,$88,$FE,$DD
	.dc.b	$87,$E0,$F1,$ED,$7C,$3E,$1B,$FA,$AE,$1C,$23,$1F,$1B,$E1,$B5,$F0
	.dc.b	$C1,$83,$E8,$3F,$BA,$E3,$C6,$73,$0F,$87,$F0,$65,$E7,$9C,$C7,$FB
	.dc.b	$12,$B8,$5F,$38,$7C,$8F,$ED,$71,$FE,$9B,$B7,$C2,$CF,$F1,$11,$F1
	.dc.b	$FD,$9F,$87,$B7,$FA,$D2,$7F,$51,$3E,$EE,$7F,$89,$7F,$2A,$BE,$27
	.dc.b	$7E,$B6,$F2,$69,$67,$E4,$3E,$DC,$7D,$31,$8D,$FD,$9F,$2B,$FC,$27
	.dc.b	$E1,$7F,$03,$97,$D6,$E1,$F2,$E5,$F5,$7E,$93,$C3,$97,$CA,$8E,$FE
	.dc.b	$6F,$CD,$DF,$E6,$6B,$F2,$FC,$CE,$F5,$F5,$75,$FE,$F5,$C9,$F1,$9F
	.dc.b	$99,$CF,$FC,$67,$D4,$F0,$E6,$F4,$3F,$A0,$3F,$65,$7C,$CF,$F2,$36
	.dc.b	$D7,$EC,$5C,$F5,$59,$FC,$C7,$8F,$FC,$6E,$73,$F0,$D6,$7F,$38,$FE
	.dc.b	$5B,$0F,$E5,$53,$AF,$9F,$F7,$BC,$30,$FF,$1C,$CF,$F1,$38,$C7,$D0
	.dc.b	$7F,$6C,$8E,$38,$74,$A4,$FF,$8A,$B5,$FE,$5F,$7C,$3E,$4F,$FB,$9C
	.dc.b	$62,$F8,$52,$7F,$BA,$61,$D3,$06,$5F,$B3,$AD,$79,$4E,$1D,$4E,$70
	.dc.b	$FF,$05,$7C,$87,$B3,$1F,$91,$97,$F3,$77,$F9,$A6,$9F,$13,$BB,$DF
	.dc.b	$3E,$99,$CB,$EB,$E0,$BA,$F8,$8F,$6C,$FD,$59,$F4,$E7,$67,$5A,$FA
	.dc.b	$EB,$E4,$6B,$F6,$7E,$7B,$C7,$C1,$F5,$EF,$FD,$D9,$C1,$F0,$DF,$F7
	.dc.b	$65,$F9,$1E,$7B,$FF,$53,$9D,$70,$7B,$F1,$FF,$14,$47,$EF,$2D,$F2
	.dc.b	$3F,$C7,$39,$BF,$DC,$FE,$4F,$E3,$27,$FD,$9F,$18,$F9,$4F,$F5,$A3
	.dc.b	$78,$BF,$DC,$4E,$7E,$5A,$D2,$D6,$9F,$F7,$2D,$6F,$E0,$FF,$9B,$5F
	.dc.b	$FB,$EA,$31,$87,$F1,$D7,$5E,$AD,$9E,$98,$73,$B7,$CD,$BE,$8B,$FD
	.dc.b	$DD,$87,$58,$7D,$2F,$E4,$76,$F8,$1A,$F9,$35,$63,$CA,$BF,$AE,$67
	.dc.b	$F0,$0E,$F3,$AC,$39,$FA,$AF,$17,$58,$F3,$C7,$94,$FF,$2A,$EA,$F6
	.dc.b	$63,$CA,$73,$E0,$FF,$A9,$F4,$F1,$CF,$CC,$3E,$47,$85,$71,$F9,$3E
	.dc.b	$7F,$CC,$7D,$1E,$4F,$C8,$BF,$5F,$17,$C3,$AF,$8D,$5E,$3D,$AC,$FC
	.dc.b	$77,$FC,$2D,$8F,$A2,$31,$C1,$F2,$23,$CD,$3E,$EC,$F5,$C1,$F0,$6B
	.dc.b	$FC,$01,$0F,$37,$DB,$8B,$E7,$3D,$19,$7D,$4C,$BC,$DC,$7F,$85,$F2
	.dc.b	$6C,$76,$CB,$C7,$B7,$AB,$4E,$DF,$2D,$FD,$DB,$0F,$B5,$5F,$E0,$FE
	.dc.b	$0F,$1F,$17,$CA,$E2,$F9,$BC,$BF,$99,$CF,$EA,$71,$FC,$B8,$7D,$25
	.dc.b	$FA,$0F,$CE,$6D,$FA,$19,$47,$A6,$FC,$B1,$8E,$1F,$F6,$24,$7E,$46
	.dc.b	$31,$D7,$AA,$F8,$1C,$7F,$E7,$1C,$3A,$46,$3E,$7F,$83,$AC,$7F,$2D
	.dc.b	$8F,$B4,$FC,$54,$67,$AB,$FC,$8D,$7C,$AB,$F5,$47,$E7,$7D,$1E,$19
	.dc.b	$F9,$72,$FA,$17,$FD,$AB,$CA,$FC,$C4,$E6,$F8,$79,$1F,$C0,$39,$1A
	.dc.b	$78,$27,$AE,$1F,$DC,$AB,$FD,$61,$9F,$AA,$BF,$77,$87,$C6,$E1,$BA
	.dc.b	$FC,$09,$F6,$E3,$F9,$87,$F1,$55,$E1,$6C,$FC,$1F,$46,$D7,$83,$6F
	.dc.b	$EB,$55,$F6,$AB,$E1,$47,$F5,$DF,$23,$F5,$3F,$1F,$F7,$71,$FF,$68
	.dc.b	$ED,$EA,$7E,$CF,$0F,$93,$2F,$5F,$7F,$C6,$CB,$E6,$7C,$FD,$9F,$07
	.dc.b	$47,$C6,$9E,$FC,$3F,$ED,$4E,$1F,$E4,$9C,$7F,$8B,$3F,$C5,$7C,$1F
	.dc.b	$C4,$5F,$E0,$EF,$F1,$E5,$F1,$9F,$ED,$F9,$E2,$FF,$0A,$9E,$0F,$F9
	.dc.b	$9F,$5F,$D0,$D7,$0F,$0B,$F2,$F5,$FB,$18,$C4,$FE,$62,$31,$5C,$7A
	.dc.b	$3F,$C8,$F8,$76,$C3,$F8,$94,$3E,$8B,$FA,$F5,$FF,$19,$C7,$9C,$E7
	.dc.b	$CE,$FC,$84,$62,$91,$F4,$B0,$E9,$97,$EA,$27,$38,$7F,$98,$B8,$BF
	.dc.b	$13,$87,$1C,$3F,$17,$F1,$1F,$C7,$A3,$F4,$F5,$EE,$D3,$F8,$E4,$78
	.dc.b	$BC,$7F,$5E,$BE,$8C,$AE,$7D,$F2,$FC,$AF,$91,$CF,$1F,$81,$5E,$2F
	.dc.b	$67,$F9,$5F,$8F,$F5,$67,$FC,$DD,$B7,$91,$78,$7F,$55,$FE,$11,$B7
	.dc.b	$D2,$D1,$D3,$E5,$7F,$9D,$6D,$9D,$9F,$3F,$78,$FE,$66,$FF,$23,$2F
	.dc.b	$1F,$D0,$DC,$7D,$5C,$3F,$C1,$B9,$3F,$B6,$72,$FF,$13,$3E,$1E,$DF
	.dc.b	$D6,$5F,$EC,$8F,$F6,$67,$FB,$36,$B7,$C1,$FF,$31,$CF,$F2,$69,$FE
	.dc.b	$D5,$DC,$E6,$C5,$7E,$FF,$D3,$E8,$FF,$01,$F1,$6D,$6D,$FC,$CE,$5E
	.dc.b	$49,$D7,$CA,$FF,$3F,$C7,$AE,$3F,$D8,$BE,$6B,$F9,$9F,$3F,$F0,$69
	.dc.b	$CE,$29,$7C,$F7,$F1,$99,$CC,$7D,$0F,$07,$FB,$44,$63,$A5,$1F,$E8
	.dc.b	$1C,$3A,$67,$E7,$BE,$1F,$17,$FD,$9B,$0F,$1F,$81,$EA,$D3,$C9,$C3
	.dc.b	$DF,$0E,$1C,$F7,$87,$DD,$97,$E0,$2B,$C3,$E4,$7F,$6B,$CB,$FC,$25
	.dc.b	$AF,$CD,$F0,$FE,$76,$1F,$52,$FC,$F9,$7F,$A4,$BB,$BE,$6C,$FC,$FF
	.dc.b	$33,$C5,$C7,$FE,$D2,$E5,$F9,$53,$E0,$E3,$FF,$51,$F1,$FB,$23,$1F
	.dc.b	$3D,$D1,$D6,$BF,$0A,$F8,$F1,$7C,$FC,$3E,$2F,$C7,$25,$D1,$FC,$DB
	.dc.b	$C1,$FF,$67,$B7,$AB,$FC,$4F,$C3,$FC,$EA,$B6,$FE,$73,$3C,$BC,$EB
	.dc.b	$94,$7B,$E3,$84,$63,$9F,$15,$CB,$0F,$F5,$9C,$3E,$6B,$FC,$1D,$5B
	.dc.b	$8F,$F0,$4D,$F2,$3C,$08,$FC,$2B,$FE,$FB,$BE,$98,$66,$7F,$13,$59
	.dc.b	$3A,$75,$7F,$96,$AF,$A6,$3E,$2A,$FC,$7F,$43,$CD,$FF,$7C,$E1,$C2
	.dc.b	$1F,$2D,$8C,$3C,$77,$C3,$E6,$BF,$23,$1F,$0C,$CE,$5F,$9B,$87,$D2
	.dc.b	$78,$34,$F3,$4E,$63,$E9,$EB,$EA,$C3,$DD,$39,$F8,$5E,$9B,$79,$FB
	.dc.b	$BF,$DE,$9B,$79,$4F,$24,$7A,$EB,$C1,$A7,$7C,$39,$EC,$D7,$C7,$F4
	.dc.b	$5E,$7F,$8C,$FC,$0D,$FF,$18,$E7,$FB,$BB,$7D,$7E,$EB,$AF,$DE,$3F
	.dc.b	$0B,$1F,$43,$92,$C7,$DF,$3E,$DD,$FC,$6F,$A5,$F3,$BF,$13,$97,$EA
	.dc.b	$3F,$1B,$C3,$BC,$3E,$9D,$FC,$DE,$9F,$A7,$87,$E7,$7C,$15,$E3,$F9
	.dc.b	$19,$7E,$6F,$DB,$FE,$8F,$9F,$A1,$FE,$93,$D5,$63,$F8,$E7,$99,$7B
	.dc.b	$BF,$98,$E5,$EA,$BF,$E9,$F3,$FE,$96,$F5,$7F,$CC,$11,$8F,$1D,$E8
	.dc.b	$E3,$1E,$70,$F4,$3F,$D0,$31,$F9,$3A,$E1,$1F,$E8,$4E,$4C,$4F,$08
	.dc.b	$FF,$43,$F8,$98,$D3,$FA,$1E,$1C,$FD,$74,$FF,$A0,$6F,$F6,$58,$F4
	.dc.b	$9C,$FA,$3F,$97,$9F,$84,$75,$8C,$F7,$70,$8C,$5F,$FA,$2F,$E7,$FF
	.dc.b	$D6,$B8,$73,$8C,$7C,$9F,$2C,$FA,$5F,$BD,$F2,$B7,$83,$FD,$77,$E0
	.dc.b	$DD,$7E,$14,$FF,$90,$09,$DA,$CB,$E1,$F1,$66,$7C,$98,$FC,$0E,$0F
	.dc.b	$F5,$4F,$1F,$C3,$CF,$FD,$8F,$1F,$1B,$87,$F8,$8E,$5C,$F0,$F5,$C3
	.dc.b	$7E,$77,$C3,$87,$E2,$C7,$FB,$CE,$F8,$65,$DE,$73,$F2,$7F,$23,$CC
	.dc.b	$FE,$37,$97,$C1,$F6,$FF,$0D,$97,$8B,$DB,$F4,$4F,$C5,$8E,$DC,$9F
	.dc.b	$E1,$2C,$FA,$C6,$3B,$BE,$0F,$17,$E0,$BD,$BE,$35,$AE,$4F,$C4,$47
	.dc.b	$F6,$89,$D7,$AD,$F0,$67,$2F,$F9,$DA,$33,$E6,$74,$CF,$F1,$57,$FE
	.dc.b	$44,$F4,$DE,$0F,$C2,$C7,$8C,$3E,$6F,$1C,$EB,$E1,$F2,$DF,$D1,$F3
	.dc.b	$F3,$74,$F8,$6B,$5A,$B1,$F0,$73,$FD,$7C,$7C,$1E,$4F,$47,$9E,$DF
	.dc.b	$17,$5F,$D2,$76,$7C,$0D,$3F,$47,$EB,$5A,$F4,$FA,$7D,$F0,$FF,$2E
	.dc.b	$E3,$DB,$97,$F0,$31,$E6,$E2,$E9,$BF,$93,$67,$0B,$D3,$CD,$75,$E0
	.dc.b	$FC,$67,$CB,$5C,$BF,$16,$FC,$FD,$7F,$64,$BE,$67,$E0,$37,$93,$C5
	.dc.b	$C7,$FB,$D8,$D4,$E3,$54,$71,$F9,$AF,$F3,$1F,$47,$DD,$9C,$D2,$5F
	.dc.b	$15,$F4,$AF,$87,$B8,$F1,$3D,$B1,$8F,$2F,$DF,$CB,$E9,$D3,$C5,$AF
	.dc.b	$9B,$3E,$9B,$73,$7B,$65,$74,$3F,$BE,$9F,$7C,$EB,$D8,$F9,$7C,$9D
	.dc.b	$77,$F5,$F5,$F5,$F2,$7D,$DB,$FB,$7B,$3F,$85,$7B,$7F,$A6,$E4,$FD
	.dc.b	$93,$FE,$98,$D3,$F8,$EA,$F6,$BF,$98,$2E,$CF,$EC,$B3,$F8,$18,$C7
	.dc.b	$CC,$FF,$31,$84,$6F,$E1,$3F,$C0,$31,$E0,$A4,$3E,$43,$F9,$95,$6D
	.dc.b	$EB,$5B,$9D,$7C,$0F,$7B,$E3,$5F,$E9,$9E,$5F,$3C,$62,$5C,$DE,$6D
	.dc.b	$3D,$B3,$98,$FE,$F3,$8F,$5C,$3F,$D4,$51,$FD,$F3,$0C,$61,$CB,$DC
	.dc.b	$7C,$D6,$2B,$FC,$02,$BD,$3F,$E6,$FA,$FA,$6F,$84,$BE,$4F,$38,$FC
	.dc.b	$AD,$26,$DF,$1F,$0F,$81,$8F,$B6,$3D,$F9,$7B,$1B,$9F,$8B,$D5,$EF
	.dc.b	$87,$F0,$6D,$BE,$4B,$F3,$3E,$46,$DF,$2F,$0F,$F8,$76,$7E,$5C,$9E
	.dc.b	$BE,$30,$F8,$99,$FF,$34,$F2,$7F,$CF,$51,$8E,$FF,$8D,$E2,$79,$79
	.dc.b	$3F,$B7,$ED,$C7,$5F,$B5,$3F,$E9,$68,$DF,$C1,$7F,$93,$9F,$3C,$EA
	.dc.b	$1F,$01,$C6,$FF,$D5,$31,$8E,$1E,$32,$FA,$5B,$F0,$7F,$A1,$E1,$FC
	.dc.b	$36,$6C,$39,$CB,$E9,$7B,$E5,$FD,$83,$4E,$7D,$FC,$9C,$E1,$C7,$3F
	.dc.b	$F2,$1F,$9B,$C3,$D4,$F7,$4F,$E2,$AB,$CC,$F7,$DA,$E8,$F8,$CB,$8F
	.dc.b	$F9,$33,$E3,$7A,$CF,$93,$E4,$2C,$FF,$94,$3C,$EF,$2E,$BF,$2F,$E1
	.dc.b	$A7,$77,$DF,$E0,$43,$E6,$70,$BE,$1F,$F8,$77,$07,$E6,$63,$1F,$47
	.dc.b	$F2,$DB,$E0,$E6,$93,$FD,$06,$BE,$2D,$70,$F5,$7F,$CE,$5E,$37,$E9
	.dc.b	$BA,$FF,$69,$39,$FA,$1D,$E3,$E5,$E0,$F4,$2F,$99,$F0,$3D,$BD,$3F
	.dc.b	$06,$8F,$6A,$E9,$FE,$64,$F0,$BF,$1B,$A7,$6C,$3C,$B3,$FA,$ED,$7C
	.dc.b	$5C,$3F,$0A,$73,$F1,$3F,$73,$8F,$F2,$5C,$71,$F9,$0F,$C5,$E2,$FC
	.dc.b	$63,$FE,$F8,$F1,$EF,$B8,$FA,$4B,$F4,$1F,$A9,$D3,$BD,$FF,$BF,$A1
	.dc.b	$F2,$FF,$C1,$63,$0F,$F6,$DC,$7F,$8E,$F9,$1F,$C4,$A7,$3F,$31,$F9
	.dc.b	$05,$97,$90,$61,$E0,$E5,$E9,$18,$C7,$F8,$56,$5F,$03,$BB,$DB,$3F
	.dc.b	$DF,$73,$DE,$1E,$4F,$97,$FF,$60,$79,$E9,$F8,$07,$F9,$E9,$FF,$61
	.dc.b	$7B,$BF,$E8,$07,$FD,$89,$EC,$59,$F2,$FC,$77,$57,$FC,$37,$3F,$EC
	.dc.b	$98,$FD,$54,$7F,$89,$8E,$E7,$0F,$0F,$83,$FE,$44,$9E,$11,$FD,$92
	.dc.b	$39,$9B,$F0,$FF,$A6,$6B,$FC,$8E,$31,$EC,$EB,$56,$2C,$3F,$B6,$FB
	.dc.b	$9F,$D0,$B0,$E1,$5A,$9C,$F6,$7F,$B1,$F3,$F0,$E7,$FE,$01,$8F,$F1
	.dc.b	$D4,$63,$0F,$E9,$FE,$9F,$EA,$31,$E7,$97,$7C,$38,$4F,$F7,$9F,$1B
	.dc.b	$DF,$9F,$79,$CE,$FF,$E7,$AC,$3C,$7B,$7C,$BA,$7B,$37,$29,$B7,$D3
	.dc.b	$AF,$85,$D7,$EF,$9F,$06,$B6,$BD,$7A,$FC,$BB,$F9,$13,$E8,$CF,$F6
	.dc.b	$BB,$7A,$F7,$FF,$4E,$7B,$3E,$6C,$BC,$38,$7D,$7D,$BD,$B9,$E2,$3E
	.dc.b	$31,$F6,$5D,$61,$F1,$7B,$33,$BF,$BF,$47,$B3,$8C,$BC,$FA,$7F,$6C
	.dc.b	$F9,$CF,$C8,$69,$DB,$DE,$E7,$AF,$E1,$7E,$0B,$C0,$EB,$5F,$C2,$C7
	.dc.b	$E3,$23,$DD,$A7,$DB,$8F,$E7,$1E,$67,$7D,$D1,$FC,$EF,$E6,$F0,$DE
	.dc.b	$4E,$3B,$7E,$1E,$35,$E1,$E2,$F8,$FA,$8C,$13,$FD,$DF,$3F,$91,$5A
	.dc.b	$9C,$F8,$FE,$79,$F5,$5F,$1B,$7C,$8F,$DE,$E0,$F9,$B9,$78,$23,$FA
	.dc.b	$84,$62,$FE,$E4,$FE,$99,$EE,$ED,$B7,$D6,$BF,$D3,$9B,$FC,$6F,$43
	.dc.b	$8C,$FE,$86,$3E,$4A,$F9,$2F,$86,$FB,$7B,$7E,$06,$3F,$1F,$D0,$B3
	.dc.b	$E9,$F1,$7D,$9C,$9F,$3E,$3C,$19,$7C,$2F,$5B,$E2,$6B,$F2,$57,$AB
	.dc.b	$F9,$56,$FA,$BF,$55,$CF,$AF,$64,$FE,$C2,$BF,$83,$4E,$BD,$EE,$95
	.dc.b	$AD,$1F,$4D,$E4,$E8,$FE,$01,$7D,$27,$39,$79,$6F,$85,$BE,$8F,$D1
	.dc.b	$3E,$59,$FF,$07,$65,$BC,$D8,$78,$1B,$E1,$F7,$EA,$FF,$54,$7B,$5C
	.dc.b	$7D,$CF,$55,$7F,$85,$FA,$7F,$53,$B7,$8F,$E3,$78,$3E,$5D,$FF,$9B
	.dc.b	$F4,$FF,$20,$8C,$7C,$BA,$FF,$B0,$6C,$F8,$DF,$F3,$77,$B3,$FC,$DF
	.dc.b	$E2,$7F,$4B,$3F,$EE,$9A,$DC,$7F,$0A,$E3,$FE,$61,$3C,$EB,$87,$99
	.dc.b	$FE,$E5,$8F,$DF,$46,$3E,$57,$E5,$C7,$FB,$52,$31,$F3,$BF,$E7,$0C
	.dc.b	$7A,$DA,$D3,$FD,$E6,$FC,$57,$C3,$DD,$FF,$06,$C7,$EF,$A3,$1F,$17
	.dc.b	$A8,$67,$FC,$83,$0E,$76,$AF,$F3,$85,$F0,$F8,$8F,$ED,$CD,$F7,$3A
	.dc.b	$63,$CB,$3F,$26,$7F,$E1,$6C,$18,$7F,$77,$F2,$BE,$15,$EF,$3E,$F5
	.dc.b	$AC,$7F,$7D,$AF,$CA,$AF,$92,$70,$D7,$F9,$AF,$89,$F3,$27,$E1,$E1
	.dc.b	$75,$F9,$5D,$5F,$33,$0E,$DA,$DC,$7D,$CF,$F8,$AA,$3E,$14,$63,$7F
	.dc.b	$E3,$F8,$F4,$AF,$56,$BF,$DD,$6F,$C5,$BF,$BA,$FE,$36,$F9,$C3,$F0
	.dc.b	$39,$7A,$1F,$C1,$FD,$05,$D3,$EA,$77,$F8,$31,$E5,$C3,$F1,$B9,$7E
	.dc.b	$5E,$FE,$38,$F9,$3F,$4D,$8A,$DE,$9F,$E2,$1F,$17,$C7,$C5,$F2,$A3
	.dc.b	$FC,$49,$E3,$FC,$EF,$86,$FC,$BF,$17,$FA,$EB,$8F,$C4,$DF,$37,$E4
	.dc.b	$39,$F1,$6D,$7F,$A8,$FC,$37,$7F,$CB,$8C,$3F,$EB,$CA,$ED,$87,$ED
	.dc.b	$A1,$EF,$7E,$66,$7C,$17,$CA,$3F,$8A,$C7,$08,$C7,$97,$F9,$3C,$3A
	.dc.b	$D7,$F3,$6E,$BF,$99,$CD,$FE,$25,$9C,$D7,$E8,$70,$FC,$8C,$63,$E6
	.dc.b	$3F,$B4,$F4,$FE,$E2,$31,$F4,$7A,$D7,$CD,$9F,$F4,$2B,$75,$3A,$7C
	.dc.b	$9B,$FA,$B4,$56,$BC,$DF,$46,$BF,$0F,$2F,$E2,$6F,$BF,$4F,$8B,$14
	.dc.b	$EF,$7C,$A7,$C5,$C1,$F0,$30,$F0,$C7,$D7,$E6,$FA,$B8,$7F,$5F,$8F
	.dc.b	$A7,$87,$FB,$ED,$E3,$F2,$7E,$1C,$3E,$AF,$BB,$DD,$F9,$D8,$7C,$BB
	.dc.b	$F7,$FA,$36,$7E,$7F,$87,$F3,$31,$FA,$BB,$7B,$F4,$FF,$16,$72,$7C
	.dc.b	$FC,$7D,$31,$F1,$BA,$FE,$2F,$1F,$C9,$E3,$7C,$EE,$38,$BE,$3C,$3D
	.dc.b	$9F,$E5,$5E,$17,$F9,$77,$DB,$FD,$DE,$DF,$CB,$7E,$6B,$EE,$3F,$A7
	.dc.b	$9D,$76,$D8,$FF,$80,$3C,$F6,$FF,$CC,$D3,$FC,$06,$BA,$52,$B0,$FF
	.dc.b	$AF,$C4,$FF,$8A,$E7,$71,$B3,$F7,$87,$F8,$89,$CB,$C0,$FE,$78,$E2
	.dc.b	$77,$1E,$C3,$B1,$E4,$3F,$81,$47,$27,$B1,$D4,$D9,$B3,$95,$7F,$9D
	.dc.b	$67,$8C,$68,$E9,$D7,$FE,$C6,$9F,$E9,$73,$CD,$FE,$4B,$7C,$AB,$35
	.dc.b	$FE,$CC,$F8,$1F,$9E,$DC,$3E,$64,$3A,$69,$FD,$1A,$BD,$E7,$FB,$CD
	.dc.b	$D3,$F2,$4C,$3B,$7D,$91,$CE,$37,$1C,$A3,$11,$FD,$66,$7F,$D3,$D1
	.dc.b	$88,$F9,$75,$E7,$77,$3C,$5D,$57,$C4,$AE,$11,$FD,$72,$B8,$3A,$9F
	.dc.b	$A9,$3A,$8F,$F4,$35,$78,$1C,$4E,$9D,$45,$78,$57,$FB,$D3,$1F,$93
	.dc.b	$3F,$DF,$EB,$F2,$91,$F8,$57,$E2,$3F,$79,$F5,$C7,$E7,$30,$FC,$9C
	.dc.b	$FE,$7A,$72,$7E,$12,$38,$B9,$3F,$03,$1F,$94,$8D,$BF,$96,$2C,$64
	.dc.b	$D9,$92,$99,$3F,$2C,$FE,$35,$FC,$93,$FE,$14,$8B,$19,$AF,$80,$E5
	.dc.b	$AF,$E5,$57,$B3,$F9,$D8,$FF,$10,$3F,$BD,$BF,$EC,$0F,$EF,$F0,$B2
	.dc.b	$CB,$F2,$15,$FD,$E6,$B5,$56,$B4,$E9,$FC,$2C,$6A,$3F,$D4,$8D,$9C
	.dc.b	$CF,$09,$FC,$93,$E4,$30,$F3,$7F,$D3,$F3,$CE,$3F,$2F,$1F,$E3,$F8
	.dc.b	$C4,$7E,$35,$AF,$F7,$D6,$CF,$29,$EB,$3F,$9E,$79,$18,$7F,$49,$18
	.dc.b	$AF,$CA,$47,$EC,$6F,$FB,$6D,$7E,$24,$FD,$1C,$79,$7E,$3F,$FC,$98
	.dc.b	$FC,$F8,$FF,$07,$3F,$3D,$FF,$11,$3F,$3D,$A7,$CA,$BB,$7F,$12,$3A
	.dc.b	$9F,$1E,$3F,$D9,$D3,$F3,$63,$33,$FB,$27,$C2,$7B,$8A,$FE,$E6,$6C
	.dc.b	$FE,$FB,$C9,$AF,$F0,$3B,$FD,$36,$1F,$C2,$AF,$F8,$04,$7F,$24,$8C
	.dc.b	$4F,$F0,$68,$FE,$3E,$FC,$1C,$3F,$9B,$4E,$A3,$A4,$7F,$42,$8C,$47
	.dc.b	$F1,$B9,$FF,$1B,$67,$FE,$79,$8F,$BD,$F0,$FE,$47,$F9,$03,$FC,$25
	.dc.b	$C9,$F3,$2F,$ED,$C7,$E4,$6F,$AC,$A3,$F1,$B1,$87,$EE,$B6,$3F,$91
	.dc.b	$7E,$BF,$FE,$47,$C7,$F9,$8C,$7E,$F6,$FF,$A4,$CE,$75,$FF,$2E,$BB
	.dc.b	$C6,$CF,$F5,$C9,$FF,$44,$CF,$38,$FC,$30,$0E,$5F,$E2,$87,$F9,$5D
	.dc.b	$C4,$FF,$58,$8D,$9F,$D7,$A7,$C6,$FF,$5D,$3F,$B1,$CF,$37,$FA,$B1
	.dc.b	$C0,$FE,$B3,$3A,$7F,$81,$4F,$77,$F8,$24,$57,$FA,$D1,$FD,$AE,$7A
	.dc.b	$3F,$58,$E9,$7F,$D6,$A3,$31,$F6,$9F,$F3,$86,$DF,$E0,$86,$EB,$F1
	.dc.b	$A7,$FA,$B1,$F1,$8F,$E7,$CD,$7F,$AB,$9F,$06,$39,$E3,$FC,$0E,$3C
	.dc.b	$D5,$FB,$97,$D8,$7F,$5E,$8F,$EB,$D3,$C1,$FD,$54,$7D,$18,$C4,$73
	.dc.b	$6B,$FC,$31,$AF,$FA,$C9,$AF,$80,$FE,$B7,$5C,$E7,$FA,$B4,$7F,$5E
	.dc.b	$7E,$85,$7F,$08,$9B,$3F,$C2,$AB,$E5,$57,$BF,$C9,$F8,$CF,$F9,$9F
	.dc.b	$2F,$8D,$97,$E2,$23,$B1,$FE,$21,$8F,$8D,$1F,$95,$8C,$BF,$E9,$F8
	.dc.b	$F9,$11,$FD,$66,$2B,$FD,$7E,$3F,$12,$FF,$61,$8A,$7F,$89,$A3,$E5
	.dc.b	$BF,$E2,$82,$C6,$4E,$E6,$4A,$66,$BD,$AF,$F5,$D6,$BF,$D9,$5A,$FE
	.dc.b	$1E,$3F,$93,$3F,$55,$FE,$4C,$FE,$05,$FE,$4C,$D3,$F9,$31,$72,$FB
	.dc.b	$AF,$EA,$63,$FA,$AE,$39,$C7,$33,$FB,$09,$FE,$A4,$FE,$A1,$FE,$98
	.dc.b	$D3,$FA,$2C,$DA,$FB,$91,$E5,$3F,$A3,$47,$77,$FA,$45,$59,$FF,$0A
	.dc.b	$C7,$73,$31,$C6,$3F,$09,$7F,$E5,$B7,$FA,$6C,$3F,$19,$67,$FC,$1F
	.dc.b	$E6,$BF,$CE,$31,$8C,$11,$FE,$BB,$7F,$DB,$D1,$98,$75,$FF,$BC,$A3
	.dc.b	$0F,$E2,$DB,$9F,$16,$FD,$17,$F0,$F7,$3E,$17,$FA,$B7,$0F,$F3,$7C
	.dc.b	$63,$DF,$FF,$35,$BA,$F1,$35,$3F,$C7,$63,$1E,$2F,$F8,$BD,$BF,$23
	.dc.b	$E6,$D7,$D9,$A7,$FA,$07,$26,$3F,$EB,$BA,$78,$3C,$EF,$7A,$AF,$79
	.dc.b	$2F,$77,$F2,$6F,$75,$A9,$E7,$F1,$B9,$D3,$E4,$DF,$C5,$BE,$AF,$C1
	.dc.b	$78,$C7,$B9,$E7,$95,$8F,$A3,$9F,$A6,$3D,$14,$F1,$3B,$70,$7D,$2D
	.dc.b	$7D,$15,$AF,$A2,$F1,$7C,$07,$DC,$CF,$EF,$78,$7E,$3D,$3A,$F4,$7F
	.dc.b	$41,$79,$EA,$E9,$C7,$1F,$D6,$F1,$3C,$9B,$77,$E3,$F9,$71,$F6,$33
	.dc.b	$F4,$ED,$72,$F4,$EF,$E4,$E4,$79,$7F,$56,$73,$8F,$CE,$F8,$5C,$67
	.dc.b	$D7,$1F,$AD,$E5,$FD,$5B,$DC,$7D,$9F,$83,$C5,$6F,$2F,$AD,$CC,$F6
	.dc.b	$FD,$7A,$F8,$7A,$7E,$FE,$D8,$DB,$EC,$5B,$E7,$B1,$A7,$CE,$F9,$3B
	.dc.b	$9F,$4F,$92,$FB,$9C,$BF,$2F,$83,$E3,$67,$F2,$E1,$76,$BF,$51,$5C
	.dc.b	$17,$CC,$FC,$CE,$4F,$EE,$D6,$97,$1F,$E7,$AF,$E3,$4E,$61,$F0,$5F
	.dc.b	$89,$9F,$AD,$5C,$3E,$2F,$FD,$61,$D1,$FB,$9C,$38,$43,$DA,$FC,$64
	.dc.b	$63,$E6,$7F,$D7,$95,$BB,$E5,$03,$ED,$4E,$DF,$D7,$67,$E8,$9D,$C3
	.dc.b	$E4,$7C,$F1,$87,$EF,$E5,$F0,$74,$67,$D6,$F8,$6F,$F8,$4C,$FC,$55
	.dc.b	$A9,$F1,$61,$F8,$8F,$9B,$F9,$58,$FC,$93,$E0,$D7,$47,$DB,$97,$C3
	.dc.b	$EC,$B3,$CE,$7D,$73,$EE,$9B,$3D,$F0,$F1,$F7,$79,$F4,$F3,$65,$F9
	.dc.b	$8F,$8E,$FD,$E7,$C1,$7E,$07,$83,$F6,$5A,$78,$DF,$B5,$E9,$5D,$7E
	.dc.b	$DE,$BF,$5E,$7E,$87,$17,$83,$3F,$85,$C5,$F2,$37,$F1,$6B,$D3,$87
	.dc.b	$E1,$D7,$8B,$E5,$BF,$45,$8F,$D3,$AF,$D5,$F6,$79,$F8,$BE,$3F,$23
	.dc.b	$DC,$C4,$7C,$EF,$32,$DF,$EC,$B5,$B7,$FC,$FB,$4E,$F7,$CB,$8F,$F8
	.dc.b	$D7,$7F,$E7,$CF,$F2,$4F,$27,$EF,$57,$C6,$FF,$0B,$6F,$93,$F5,$5C
	.dc.b	$FC,$05,$93,$FC,$C2,$BF,$8F,$4E,$BC,$AF,$CE,$3C,$3E,$27,$F3,$7A
	.dc.b	$FF,$96,$E3,$1F,$2D,$C3,$1E,$74,$F8,$5F,$CD,$DE,$E7,$30,$AF,$44
	.dc.b	$62,$3F,$5F,$8F,$F0,$DC,$3C,$11,$CF,$E3,$FF,$8B,$46,$21,$F3,$0E
	.dc.b	$DE,$18,$73,$8C,$77,$FC,$88,$C4,$BE,$07,$1C,$9F,$0F,$83,$F1,$3A
	.dc.b	$79,$16,$FF,$0B,$0F,$4C,$E6,$3D,$D9,$D3,$E0,$67,$E6,$AE,$E7,$D8
	.dc.b	$C3,$EA,$63,$F5,$61,$F3,$BE,$1C,$39,$E7,$C7,$3F,$0E,$5D,$3C,$6F
	.dc.b	$6C,$62,$3E,$0F,$D1,$E2,$FC,$7A,$7C,$3D,$3E,$05,$EE,$3E,$37,$46
	.dc.b	$2B,$79,$3E,$67,$19,$7A,$A3,$FA,$87,$4F,$F8,$47,$8B,$E8,$63,$F3
	.dc.b	$DF,$F0,$C7,$91,$FE,$53,$E2,$FF,$09,$FC,$1F,$F8,$E9,$FF,$03,$BF
	.dc.b	$EA,$1E,$C7,$CB,$7F,$85,$63,$0F,$E6,$A7,$FC,$15,$18,$EE,$FE,$BE
	.dc.b	$DE,$7F,$F0,$B3,$98,$B2,$BE,$B1,$F7,$7D,$BF,$0C,$E7,$0F,$D9,$74
	.dc.b	$7E,$56,$3F,$C3,$B8,$77,$87,$C0,$7F,$3F,$D9,$39,$87,$D1,$70,$BE
	.dc.b	$99,$7E,$2E,$73,$CD,$F0,$A7,$38,$F4,$BF,$D7,$C7,$B8,$CD,$74,$C1
	.dc.b	$9E,$74,$F5,$E1,$C6,$7F,$51,$7F,$E4,$77,$C9,$87,$9B,$B1,$B6,$6B
	.dc.b	$AC,$E7,$0C,$1F,$E4,$CB,$FA,$9E,$47,$D3,$AF,$39,$F2,$27,$C3,$9D
	.dc.b	$DF,$E4,$BF,$AF,$9F,$74,$EB,$0F,$8F,$1E,$9A,$DC,$79,$7B,$BC,$3F
	.dc.b	$05,$DE,$FF,$07,$BF,$DE,$C3,$EB,$78,$96,$7E,$DC,$FD,$E7,$C7,$F0
	.dc.b	$BF,$45,$F0,$78,$7E,$2C,$FB,$ED,$FF,$00,$71,$F8,$5E,$D7,$BD,$FF
	.dc.b	$34,$F1,$FB,$FD,$6D,$47,$E2,$97,$C0,$FE,$4E,$3E,$54,$63,$BF,$CC
	.dc.b	$EB,$BB,$E1,$4F,$E0,$E3,$1F,$39,$CE,$FF,$58,$FF,$1C,$9F,$C2,$4E
	.dc.b	$BD,$CF,$CA,$BC,$3D,$C7,$3F,$3C,$38,$C3,$E8,$71,$93,$2F,$F3,$FC
	.dc.b	$7F,$9B,$F6,$FF,$44,$5F,$08,$FE,$3D,$9F,$E3,$F2,$ED,$1F,$C8,$7D
	.dc.b	$0C,$56,$FB,$3F,$CD,$D7,$F0,$23,$1A,$FF,$88,$BB,$BE,$EE,$3C,$AD
	.dc.b	$2F,$2B,$E3,$E1,$F3,$70,$FE,$18,$FC,$3D,$7F,$D2,$1A,$79,$9F,$57
	.dc.b	$B3,$F0,$AB,$DF,$5E,$ED,$BD,$B9,$7F,$55,$D3,$C3,$5E,$4A,$FA,$BF
	.dc.b	$09,$DF,$17,$CF,$77,$C7,$FD,$47,$C9,$CF,$8B,$FC,$55,$87,$D2,$C7
	.dc.b	$E3,$75,$5C,$7E,$65,$7C,$FF,$0B,$18,$F8,$B8,$7F,$A7,$6B,$E2,$BD
	.dc.b	$F2,$7F,$4B,$97,$CC,$BF,$4A,$6B,$FE,$DE,$1F,$A3,$3E,$1B,$F9,$44
	.dc.b	$6A,$76,$FF,$1B,$F1,$7F,$D0,$4F,$33,$F9,$57,$91,$89,$DF,$81,$DA
	.dc.b	$BF,$D1,$11,$8A,$56,$1F,$E6,$46,$CE,$E7,$84,$C9,$C0,$C9,$E1,$3F
	.dc.b	$97,$1E,$69,$CC,$F7,$8F,$F0,$21,$D1,$C3,$DC,$EC,$72,$3D,$F3,$B7
	.dc.b	$C2,$75,$37,$5E,$27,$C8,$78,$DF,$F5,$35,$6A,$35,$35,$E1,$1B,$7F
	.dc.b	$D5,$91,$EB,$F1,$3E,$6C,$5E,$4F,$98,$FD,$88,$FA,$CF,$C9,$3F,$DE
	.dc.b	$B8,$FF,$36,$7C,$98,$FD,$89,$FB,$93,$B6,$D6,$AB,$A4,$E9,$F9,$26
	.dc.b	$1F,$C2,$46,$23,$BC,$6D,$FC,$2D,$77,$8E,$E7,$8E,$3E,$6D,$62,$3D
	.dc.b	$31,$87,$15,$CA,$3C,$11,$C9,$D3,$A7,$E3,$3C,$CF,$87,$5E,$17,$13
	.dc.b	$A7,$47,$E1,$EF,$E6,$CF,$E7,$3D,$5B,$7F,$E4,$9C,$7E,$7B,$F9,$6A
	.dc.b	$A7,$E1,$DF,$C2,$1F,$93,$7E,$91,$F8,$57,$2F,$E2,$9F,$90,$FE,$72
	.dc.b	$2C,$64,$F8,$07,$E2,$8E,$D5,$97,$F3,$27,$E5,$A7,$F1,$2D,$3F,$46
	.dc.b	$F2,$3F,$22,$FB,$CF,$C8,$94,$FC,$8C,$75,$7F,$5F,$1F,$C1,$4F,$E0
	.dc.b	$91,$FA,$FC,$BE,$D4,$FE,$B2,$7F,$C5,$18,$7E,$FD,$B8,$FE,$F9,$CC
	.dc.b	$59,$FF,$11,$3F,$8A,$79,$18,$7F,$06,$FE,$2A,$38,$C7,$82,$3C,$86
	.dc.b	$DD,$B8,$AF,$C2,$3F,$C5,$E7,$8C,$F4,$BF,$53,$A6,$BF,$93,$9E,$07
	.dc.b	$57,$F4,$AF,$57,$F1,$2D,$7F,$1B,$1C,$9F,$C9,$2F,$8C,$FE,$53,$3F
	.dc.b	$CB,$E7,$C6,$3A,$3F,$4E,$FC,$E3,$FC,$A9,$1F,$C5,$5F,$9A,$FF,$28
	.dc.b	$7E,$6B,$FB,$C7,$E6,$B4,$F9,$58,$DC,$7F,$51,$1C,$CF,$91,$1F,$C9
	.dc.b	$27,$E2,$C6,$67,$E4,$BE,$73,$E1,$46,$4F,$90,$7A,$BE,$2B,$D3,$1A
	.dc.b	$BF,$F1,$EB,$C6,$BF,$69,$1F,$CD,$F1,$FE,$2F,$18,$9F,$F2,$1C,$62
	.dc.b	$FF,$82,$57,$FA,$E6,$31,$1F,$C2,$BA,$3B,$DF,$EF,$DF,$E3,$D3,$A9
	.dc.b	$FE,$3B,$1F,$C0,$9F,$9E,$FF,$18,$8D,$47,$F5,$D8,$FE,$47,$F1,$5F
	.dc.b	$D5,$74,$FE,$5E,$FF,$57,$DA,$E5,$FE,$58,$9F,$F1,$82,$9E,$11,$FD
	.dc.b	$89,$E5,$1F,$E3,$56,$C7,$F4,$58,$E0,$73,$3F,$85,$1F,$DC,$A3,$8B
	.dc.b	$FC,$04,$C1,$B3,$6E,$9D,$35,$FA,$27,$0D,$B9,$4F,$F6,$A9,$D4,$E9
	.dc.b	$B1,$FD,$F1,$CC,$7D,$A6,$C7,$33,$FB,$FC,$FA,$8F,$34,$FF,$66,$9F
	.dc.b	$EF,$71,$FD,$A6,$3F,$B2,$46,$32,$FD,$D4,$77,$AF,$F4,$1C,$7E,$EA
	.dc.b	$31,$5F,$C1,$B5,$F9,$17,$C7,$2D,$57,$D1,$C3,$8C,$7F,$0B,$8F,$74
	.dc.b	$62,$3A,$4F,$17,$53,$F3,$E7,$51,$FC,$2A,$74,$E9,$AF,$F1,$46,$BF
	.dc.b	$D6,$DA,$FA,$30,$FE,$9B,$7D,$70,$FD,$4C,$FF,$76,$AD,$BF,$EA,$A6
	.dc.b	$BF,$E3,$3F,$99,$76,$BD,$8F,$EA,$B1,$87,$39,$FF,$8A,$A3,$1E,$97
	.dc.b	$F8,$BD,$BD,$B9,$BE,$52,$F6,$B1,$82,$30,$FF,$81,$AF,$9C,$F8,$73
	.dc.b	$FF,$14,$C6,$25,$E5,$E3,$0E,$B7,$E7,$87,$95,$FE,$2D,$BE,$10,$FA
	.dc.b	$8C,$5F,$0C,$39,$C7,$E6,$39,$17,$C2,$3F,$D0,$BA,$FC,$5B,$E1,$87
	.dc.b	$3F,$0F,$F8,$7E,$1C,$27,$36,$9F,$F2,$1C,$E7,$1F,$81,$9F,$9E,$B5
	.dc.b	$C5,$F4,$2B,$5B,$7C,$4C,$BB,$4B,$E6,$BC,$58,$FB,$B4,$F9,$77,$BC
	.dc.b	$79,$ED,$E7,$F1,$FE,$CE,$DF,$6F,$DB,$FA,$CF,$D4,$9E,$3E,$FF,$2D
	.dc.b	$FF,$73,$97,$F8,$AB,$87,$E9,$E3,$F5,$70,$F9,$1C,$7F,$47,$5F,$58
	.dc.b	$F8,$9C,$6D,$E9,$AF,$9F,$3A,$E1,$F9,$B9,$75,$9F,$8F,$85,$AD,$6E
	.dc.b	$F1,$B1,$8F,$B7,$91,$C7,$F8,$9E,$4F,$A1,$D5,$EE,$EA,$7B,$7E,$E5
	.dc.b	$E9,$D3,$FD,$B1,$5F,$5E,$33,$EE,$FE,$5D,$DF,$57,$FB,$4F,$87,$CD
	.dc.b	$18,$FA,$1F,$84,$EB,$E1,$7F,$DB,$B3,$FE,$95,$E7,$FD,$4C,$E7,$CD
	.dc.b	$FC,$B3,$79,$3F,$22,$E7,$E7,$AE,$AC,$D9,$7F,$AD,$BA,$3F,$DD,$3B
	.dc.b	$79,$EB,$53,$9F,$37,$CB,$18,$CF,$E9,$43,$D4,$F2,$74,$FF,$A7,$AB
	.dc.b	$58,$FF,$B9,$7A,$FE,$3E,$5D,$73,$F0,$E5,$CA,$7E,$4C,$FC,$7E,$0F
	.dc.b	$99,$EE,$7C,$BB,$F3,$C7,$BB,$1F,$2F,$A7,$67,$99,$EE,$3A,$E5,$76
	.dc.b	$F7,$4F,$AA,$73,$B3,$3F,$4F,$07,$CB,$C3,$D1,$9F,$EC,$E3,$15,$BF
	.dc.b	$A6,$C7,$B1,$F7,$F8,$FE,$E6,$BF,$37,$0F,$B2,$FC,$AF,$A3,$F3,$6D
	.dc.b	$F3,$F8,$7C,$4F,$CD,$DB,$C2,$7E,$2F,$4F,$C6,$70,$7E,$2B,$97,$EF
	.dc.b	$BF,$D4,$8F,$21,$FD,$B0,$A7,$F5,$6E,$2F,$94,$FF,$70,$6B,$FD,$BD
	.dc.b	$AF,$EC,$9B,$3F,$88,$8F,$F6,$7D,$7E,$51,$FF,$37,$9F,$18,$FF,$39
	.dc.b	$CF,$27,$FD,$9E,$D7,$FD,$C4,$D7,$FD,$C2,$BE,$6F,$FA,$0F,$5F,$FA
	.dc.b	$51,$BE,$4F,$FB,$AD,$CF,$91,$79,$FF,$B3,$61,$FE,$BE,$9C,$CF,$FB
	.dc.b	$B9,$B5,$9A,$FF,$34,$57,$FB,$92,$77,$5F,$E6,$DC,$3C,$55,$FE,$61
	.dc.b	$AF,$F7,$35,$6A,$BF,$D0,$35,$AC,$3F,$D6,$15,$FE,$C0,$AD,$57,$FA
	.dc.b	$FE,$B4,$EA,$51,$88,$FF,$5D,$35,$FF,$39,$35,$FF,$64,$E1,$C2,$BA
	.dc.b	$C7,$FB,$6A,$7C,$FE,$77,$C7,$C4,$FF,$66,$C3,$E0,$2F,$8E,$F9,$0F
	.dc.b	$F6,$EE,$9F,$8D,$5C,$07,$C0,$62,$7F,$B9,$F6,$FF,$0E,$C0,$78,$EB
	.dc.b	$84,$38,$FB,$C2,$1C,$7F,$37,$5F,$EF,$B8,$74,$87,$C3,$FD,$3B,$E7
	.dc.b	$0F,$86,$7B,$9F,$4A,$73,$0F,$A4,$C6,$1C,$E1,$F5,$1C,$35,$F3,$F0
	.dc.b	$FC,$6C,$3C,$99,$75,$AE,$BD,$7A,$D7,$CB,$5A,$C7,$9E,$8F,$84,$D4
	.dc.b	$F8,$B0,$F2,$C7,$F6,$9B,$E3,$BF,$C0,$C3,$8C,$BE,$47,$E4,$6D,$F3
	.dc.b	$36,$F8,$F7,$FC,$FD,$EB,$1F,$15,$F8,$8F,$C7,$CE,$E7,$E5,$9F,$82
	.dc.b	$F9,$FC,$6B,$7E,$1D,$2E,$DE,$FC,$3D,$5E,$87,$8B,$0F,$75,$7A,$BD
	.dc.b	$1F,$2F,$37,$E1,$A7,$F0,$19,$F9,$1D,$F4,$FE,$7D,$F8,$5F,$2B,$76
	.dc.b	$FA,$58,$FD,$8C,$B8,$6B,$E3,$D1,$F0,$1F,$53,$C0,$FF,$1F,$F8,$FF
	.dc.b	$E8,$5D,$71,$A7,$F9,$07,$C3,$C7,$47,$F5,$48,$FF,$22,$7D,$56,$30
	.dc.b	$FF,$22,$7C,$E7,$C9,$C3,$8F,$A4,$8F,$C9,$47,$E6,$08,$7E,$CA,$78
	.dc.b	$8F,$07,$E9,$63,$DE,$F9,$4E,$7C,$0E,$37,$C7,$2F,$F2,$B6,$3F,$DA
	.dc.b	$31,$FD,$4F,$6F,$6C,$B9,$4E,$70,$7C,$7E,$72,$F0,$E1,$FB,$0F,$8F
	.dc.b	$FC,$C6,$7D,$74,$F0,$CE,$7E,$6B,$D5,$9F,$4C,$3C,$98,$25,$D3,$F0
	.dc.b	$B4,$F0,$4E,$7E,$53,$17,$C2,$D1,$C3,$0E,$98,$F7,$87,$C1,$78,$F4
	.dc.b	$FA,$1C,$1F,$01,$FF,$5E,$5F,$8D,$F8,$A7,$A7,$3F,$F0,$5F,$5E,$27
	.dc.b	$AC,$FC,$A7,$BF,$07,$F4,$FD,$3E,$1A,$D7,$E3,$63,$F3,$A7,$D0,$FA
	.dc.b	$B4,$F3,$3F,$1F,$BB,$CB,$C1,$F8,$1E,$4F,$2D,$FC,$E7,$C0,$FD,$0F
	.dc.b	$87,$B7,$D7,$0F,$A5,$AF,$CE,$AF,$1E,$4F,$A9,$C5,$EF,$4F,$37,$13
	.dc.b	$E6,$3C,$91,$8E,$3F,$9B,$C1,$F9,$9F,$5B,$F3,$DC,$5F,$89,$F3,$7E
	.dc.b	$FB,$E2,$7E,$D3,$E8,$C9,$1F,$77,$BB,$0F,$DF,$F9,$2B,$A3,$E4,$7F
	.dc.b	$35,$DF,$CE,$31,$C7,$FC,$93,$2F,$97,$87,$E1,$EF,$F4,$1F,$47,$FE
	.dc.b	$4A,$D7,$F5,$D3,$9B,$4B,$E9,$B8,$E3,$D2,$31,$AF,$C4,$BE,$11,$BF
	.dc.b	$3F,$15,$BB,$FE,$A5,$DF,$FB,$EB,$DC,$7F,$55,$F0,$B1,$5B,$BF,$E6
	.dc.b	$1D,$9F,$1E,$31,$E5,$3E,$57,$E5,$70,$7E,$53,$2F,$D1,$67,$F1,$38
	.dc.b	$27,$5F,$E4,$7A,$FF,$CC,$B9,$7C,$4C,$3E,$35,$7F,$01,$CE,$F0,$FC
	.dc.b	$09,$F7,$CF,$C6,$F0,$6C,$FC,$4E,$0F,$ED,$9B,$7A,$F3,$F1,$75,$F8
	.dc.b	$F3,$F8,$18,$7C,$2C,$7E,$7E,$0F,$9E,$C6,$5C,$F8,$7F,$D5,$9C,$BF
	.dc.b	$4B,$1F,$7F,$37,$C1,$C7,$8D,$7F,$53,$F9,$77,$CF,$FE,$37,$83,$F9
	.dc.b	$53,$F6,$AE,$F1,$7C,$09,$FD,$A9,$C1,$F8,$7D,$58,$AF,$07,$55,$9F
	.dc.b	$8F,$97,$E3,$2F,$F9,$4D,$79,$63,$F9,$47,$93,$FC,$92,$31,$B7,$E2
	.dc.b	$3C,$2D,$E1,$CA,$3E,$46,$1F,$CC,$EB,$7D,$7F,$8A,$8F,$F0,$6C,$63
	.dc.b	$8B,$FC,$31,$18,$BE,$55,$F8,$DE,$8C,$76,$7E,$9B,$8B,$F1,$1D,$98
	.dc.b	$EB,$FC,$6F,$27,$E2,$B4,$E8,$FF,$7B,$EC,$F8,$58,$7F,$9A,$74,$E1
	.dc.b	$8B,$2F,$85,$3D,$78,$6E,$5F,$A9,$EA,$FE,$3F,$9F,$C0,$C7,$E1,$6D
	.dc.b	$EF,$EC,$B8,$F1,$E1,$FC,$65,$7C,$1A,$F5,$F5,$7C,$6A,$D7,$0F,$EA
	.dc.b	$7B,$3D,$59,$FD,$A9,$FE,$07,$D9,$75,$FD,$4E,$1C,$E7,$8F,$27,$E3
	.dc.b	$73,$F5,$EF,$F8,$8E,$8B,$B7,$E3,$F4,$FF,$16,$F0,$7C,$45,$D5,$F1
	.dc.b	$3C,$8F,$E2,$F3,$FD,$8F,$A3,$C5,$E2,$3C,$8B,$E3,$7F,$71,$D7,$E5
	.dc.b	$CF,$F7,$1F,$27,$E4,$BF,$D6,$B8,$7E,$1B,$FD,$73,$C4,$B7,$FD,$97
	.dc.b	$A7,$E1,$4E,$7C,$9B,$3F,$D6,$BC,$1F,$74,$F1,$F1,$FF,$93,$E7,$E0
	.dc.b	$AF,$EB,$FD,$BF,$D8,$A3,$1D,$9F,$D7,$FC,$6F,$C3,$E1,$C6,$BF,$B0
	.dc.b	$EB,$AE,$5F,$EC,$1E,$4F,$AB,$1F,$69,$EA,$CB,$FD,$9E,$FA,$32,$76
	.dc.b	$FE,$09,$FE,$F7,$19,$BF,$F3,$5F,$AB,$EA,$9F,$EE,$F3,$EA,$BF,$3F
	.dc.b	$67,$91,$F4,$EA,$8F,$97,$E0,$7C,$EE,$6F,$16,$1F,$E1,$77,$B7,$95
	.dc.b	$DF,$97,$FB,$46,$BD,$6B,$FC,$0B,$E0,$5D,$BF,$C1,$5E,$57,$79,$F8
	.dc.b	$9A,$7F,$0A,$F4,$7F,$CE,$BB,$FF,$0A,$7F,$CC,$6F,$F0,$57,$FC,$F3
	.dc.b	$E2,$E7,$7F,$F4,$47,$1E,$5F,$F4,$56,$B7,$CC,$FF,$44,$F0,$7E,$9B
	.dc.b	$0F,$F3,$2F,$A9,$6F,$FC,$FF,$C1,$FC,$26,$7F,$CD,$3E,$87,$FA,$13
	.dc.b	$4F,$E1,$F1,$8F,$4B,$FB,$3F,$81,$FE,$7B,$EF,$FF,$37,$7A,$7A,$CB
	.dc.b	$FC,$E9,$DD,$FC,$37,$0E,$1E,$67,$F9,$FB,$2F,$75,$6B,$C5,$FE,$1F
	.dc.b	$5A,$BF,$F3,$F4,$EB,$83,$FC,$B7,$1F,$E8,$1F,$4F,$13,$FE,$4C,$C3
	.dc.b	$D0,$FF,$A8,$B8,$FD,$97,$FE,$79,$9F,$23,$FE,$79,$6C,$FF,$9E,$5F
	.dc.b	$33,$FE,$9C,$EB,$CE,$BF,$E7,$99,$CD,$78,$B3,$78,$FE,$2A,$FF,$36
	.dc.b	$43,$D3,$F6,$FA,$BF,$B8,$DB,$FD,$5F,$A7,$D7,$C7,$F1,$9C,$DF,$57
	.dc.b	$5F,$F3,$DE,$9F,$09,$FF,$3D,$B6,$7E,$0E,$96,$7F,$D0,$FB,$FF,$9E
	.dc.b	$AB,$FA,$8F,$AD,$F9,$7F,$37,$C1,$4D,$3F,$4B,$5F,$85,$7F,$CD,$CD
	.dc.b	$EE,$CB,$F2,$A3,$FC,$E5,$19,$33,$1F,$1A,$33,$1F,$89,$7F,$DC,$B1
	.dc.b	$F8,$17,$31,$CC,$CC,$7B,$63,$FD,$E4,$FC,$68,$DF,$47,$57,$FD,$ED
	.dc.b	$CD,$FE,$F1,$7F,$D0,$91,$C5,$EA,$7A,$A3,$4E,$9A,$FF,$A2,$D6,$3F
	.dc.b	$BB,$F5,$3F,$95,$BF,$1F,$C3,$FE,$EE,$FC,$9C,$EE,$7F,$E4,$0C,$95
	.dc.b	$FD,$D6,$BE,$74,$64,$FE,$E0,$E1,$D9,$89,$D5,$7F,$56,$9D,$4E,$8F
	.dc.b	$EE,$4E,$23,$81,$FE,$9E,$AF,$EF,$F1,$C2,$7C,$07,$E0,$2A,$A3,$A6
	.dc.b	$1F,$02,$33,$1E,$78,$C5,$FF,$84,$AF,$FB,$FD,$F5,$AE,$53,$8B,$FE
	.dc.b	$E9,$5D,$E3,$FD,$8C,$5E,$5F,$8E,$BD,$7A,$7B,$6F,$D3,$7F,$E6,$A8
	.dc.b	$7B,$DF,$2E,$31,$87,$FB,$27,$2F,$2D,$7D,$0E,$7F,$91,$5A,$F0,$5C
	.dc.b	$9F,$1A,$BE,$AC,$3B,$71,$8A,$31,$E5,$F8,$B5,$F3,$4E,$69,$EF,$67
	.dc.b	$0E,$70,$F7,$7B,$EF,$F0,$F0,$E3,$97,$79,$7B,$BF,$6B,$5F,$36,$1C
	.dc.b	$61,$EF,$71,$CB,$AD,$F0,$F4,$5C,$7F,$36,$73,$69,$F9,$9C,$5F,$63
	.dc.b	$87,$E6,$F0,$7F,$62,$DF,$C1,$BF,$93,$B3,$E0,$F0,$7C,$3A,$F6,$1E
	.dc.b	$FD,$6C,$7A,$67,$8E,$9E,$0C,$FC,$D7,$BD,$3C,$59,$F9,$63,$12,$F7
	.dc.b	$BF,$09,$EF,$7E,$03,$93,$EE,$6B,$F3,$B0,$FB,$2F,$C8,$F1,$BC,$BC
	.dc.b	$7F,$34,$F0,$E5,$DC,$FC,$87,$A1,$FB,$6E,$53,$FD,$CF,$8B,$F1,$0B
	.dc.b	$E0,$3F,$6E,$DC,$F3,$0F,$84,$A7,$2A,$FD,$A5,$71,$8E,$B8,$62,$B8
	.dc.b	$4B,$E1,$31,$5E,$27,$FA,$0D,$FD,$F9,$E4,$6F,$E1,$3F,$3E,$E2,$51
	.dc.b	$FB,$4E,$DF,$F0,$64,$E6,$1F,$07,$CF,$1F,$0D,$F0,$8F,$E1,$7D,$3F
	.dc.b	$BB,$BE,$11,$FC,$F7,$1F,$E0,$51,$8F,$0B,$FC,$31,$5B,$C3,$9E,$3D
	.dc.b	$A5,$F0,$DC,$3C,$5F,$93,$5A,$CB,$FA,$44,$7F,$0D,$E7,$FA,$59,$7F
	.dc.b	$5D,$CF,$E6,$CF,$CE,$F8,$6F,$9B,$C1,$FB,$6A,$F9,$97,$71,$F9,$7C
	.dc.b	$3F,$2A,$7F,$59,$8F,$C6,$9F,$7F,$15,$E2,$F9,$DC,$56,$FE,$66,$1F
	.dc.b	$16,$7E,$97,$03,$E1,$B1,$5F,$47,$7F,$EC,$19,$7E,$1F,$5F,$F4,$66
	.dc.b	$9F,$CF,$F8,$7F,$E0,$0E,$BF,$EF,$C9,$FE,$CD,$E3,$7F,$AD,$2B,$5E
	.dc.b	$35,$AD,$5F,$F5,$29,$FE,$5C,$43,$73,$FE,$BC,$9D,$BB,$8D,$46,$23
	.dc.b	$FB,$8C,$63,$A3,$FA,$6D,$7F,$66,$8F,$E7,$55,$BA,$DC,$62,$B7,$1B
	.dc.b	$9D,$3A,$6B,$FD,$F1,$AF,$A1,$B7,$CF,$0F,$F0,$34,$FF,$6C,$31,$85
	.dc.b	$7F,$CD,$B5,$C9,$1F,$37,$C6,$F9,$78,$FF,$66,$C7,$FA,$03,$F0,$0C
	.dc.b	$3F,$2D,$D1,$A3,$DE,$61,$F9,$4E,$8D,$57,$F5,$DC,$3F,$B5,$C7,$D1
	.dc.b	$3E,$F4,$FD,$38,$C4,$EE,$3E,$B1,$89,$E8,$73,$3E,$21,$83,$FB,$6C
	.dc.b	$71,$7F,$B9,$1E,$69,$FE,$D7,$3C,$1F,$ED,$B1,$FE,$00,$9E,$0F,$96
	.dc.b	$78,$1F,$8D,$9F,$3C,$7A,$4F,$ED,$FC,$CF,$8B,$FE,$E3,$1F,$93,$8F
	.dc.b	$F1,$BE,$5F,$E2,$0C,$FF,$DF,$19,$FF,$7F,$8F,$F1,$1F,$C5,$62,$FF
	.dc.b	$82,$1D,$45,$F2,$BE,$97,$C4,$B3,$F8,$28,$E1,$3E,$09,$FC,$8C,$F5
	.dc.b	$BF,$F2,$2D,$7F,$BC,$27,$91,$A9,$EE,$E0,$FF,$24,$B8,$3F,$1E,$FF
	.dc.b	$93,$9F,$D5,$57,$78,$EB,$1D,$DF,$F2,$5B,$5F,$F2,$8F,$C5,$57,$FC
	.dc.b	$AE,$FF,$97,$A7,$94,$60,$FF,$12,$38,$7F,$33,$1F,$E4,$77,$CC,$60
	.dc.b	$FF,$24,$18,$1F,$17,$DD,$FD,$0E,$3F,$CD,$1D,$3F,$E4,$7A,$ED,$08
	.dc.b	$FC,$C4,$61,$FB,$F7,$F7,$8F,$9F,$A7,$DB,$7F,$00,$FD,$DC,$6C,$7E
	.dc.b	$D8,$D9,$F5,$CF,$C5,$57,$F9,$5F,$2F,$99,$1F,$A5,$8C,$61,$FC,$4E
	.dc.b	$3F,$8C,$C6,$23,$EC,$35,$FA,$ED,$79,$5F,$C6,$BF,$AE,$FC,$5A,$E6
	.dc.b	$FC,$78,$B1,$F7,$A7,$FD,$0E,$FD,$29,$F9,$91,$F5,$DA,$E6,$38,$3F
	.dc.b	$E8,$E9,$FC,$CC,$EA,$3F,$37,$3F,$E5,$FD,$BF,$9C,$38,$3F,$9E,$1F
	.dc.b	$9A,$3F,$CB,$73,$FA,$B7,$F5,$BA,$E5,$FF,$3F,$45,$3F,$9E,$C7,$CC
	.dc.b	$8F,$AD,$15,$FF,$38,$C7,$CD,$7E,$D1,$FE,$72,$C3,$FA,$2C,$FF,$08
	.dc.b	$8F,$C8,$9A,$3F,$CF,$31,$A7,$4D,$7F,$8B,$9C,$0F,$E1,$B3,$67,$F7
	.dc.b	$85,$8F,$F3,$D4,$F1,$7F,$88,$3F,$D5,$B1,$F8,$AF,$F5,$8F,$32,$CF
	.dc.b	$FA,$1D,$FD,$AB,$F8,$43,$18,$FC,$75,$1F,$89,$74,$62,$3F,$80,$3F
	.dc.b	$46,$3F,$72,$FE,$F0,$E7,$1F,$8E,$6B,$F8,$E6,$BF,$2D,$E8,$F2,$7A
	.dc.b	$3F,$C0,$9F,$89,$1F,$C8,$DB,$1F,$EC,$87,$9B,$FD,$DE,$7E,$03,$FD
	.dc.b	$76,$39,$BF,$CA,$E3,$2F,$F6,$1F,$5F,$FD,$CB,$87,$C6,$AF,$F7,$56
	.dc.b	$3F,$0F,$1F,$E7,$F1,$F8,$F7,$FA,$53,$E9,$7E,$F3,$F6,$5F,$C5,$B9
	.dc.b	$33,$87,$D5,$72,$7C,$C8,$FC,$0B,$4F,$EB,$85,$3E,$F4,$71,$72,$78
	.dc.b	$CF,$EA,$07,$84,$FE,$9D,$35,$CC,$71,$7F,$D9,$CD,$7F,$B3,$FC,$F3
	.dc.b	$E1,$BF,$D9,$9E,$5F,$EB,$A3,$13,$FE,$95,$DB,$FC,$51,$18,$EC,$71
	.dc.b	$63,$0F,$F6,$A4,$63,$2F,$EA,$95,$BA,$7C,$9E,$2F,$70,$F9,$1F,$F5
	.dc.b	$8E,$5F,$E2,$FB,$E1,$1F,$EB,$08,$FF,$01,$46,$3C,$87,$95,$8C,$39
	.dc.b	$DF,$FA,$DB,$DF,$C6,$1F,$EA,$DC,$7B,$60,$F9,$37,$A7,$DF,$1E,$33
	.dc.b	$9F,$92,$C6,$1C,$72,$FF,$60,$E0,$9F,$F7,$E6,$9E,$38,$C7,$77,$F9
	.dc.b	$22,$31,$3F,$E8,$0D,$7F,$DF,$11,$89,$7B,$D8,$CF,$C3,$87,$2D,$BD
	.dc.b	$B7,$B5,$F0,$08,$FA,$5A,$7D,$13,$CC,$7F,$83,$DF,$93,$95,$C3,$DB
	.dc.b	$DF,$F1,$38,$FE,$CE,$DE,$CC,$3D,$F1,$8D,$FE,$16,$DF,$EB,$3C,$BC
	.dc.b	$55,$E1,$C7,$EF,$F6,$F8,$70,$E7,$E5,$F7,$AF,$91,$8F,$CF,$E6,$EB
	.dc.b	$C9,$F4,$B8,$5E,$EE,$36,$F9,$BC,$8F,$1F,$E5,$6D,$F1,$23,$FD,$05
	.dc.b	$9F,$C7,$3A,$E1,$F1,$FC,$EF,$24,$62,$7F,$D5,$FC,$AF,$92,$F9,$1E
	.dc.b	$DF,$EE,$2B,$A9,$FE,$84,$3F,$28,$3B,$9F,$29,$F3,$5B,$DB,$A7,$3E
	.dc.b	$4D,$9F,$F4,$B4,$FC,$19,$CC,$D9,$FF,$4D,$74,$7F,$89,$BA,$3F,$5B
	.dc.b	$5C,$A1,$F3,$1F,$98,$8C,$7C,$BB,$D9,$FC,$9C,$63,$D7,$B8,$F4,$A4
	.dc.b	$7C,$9D,$3E,$35,$6F,$E2,$FD,$EF,$1E,$47,$4E,$B1,$E5,$87,$F0,$3F
	.dc.b	$98,$C6,$FF,$ED,$8C,$FA,$E0,$97,$07,$D0,$CB,$A4,$7C,$9E,$0F,$6E
	.dc.b	$BF,$27,$C3,$F9,$5A,$7C,$28,$F7,$3E,$97,$D7,$9F,$7E,$6B,$C4,$D7
	.dc.b	$C3,$5E,$48,$C6,$DE,$7F,$2B,$C5,$F3,$1D,$BA,$BE,$A7,$AF,$F5,$73
	.dc.b	$F8,$98,$F4,$C9,$EC,$DC,$BF,$DF,$F3,$EB,$D7,$E1,$F1,$FF,$1E,$8F
	.dc.b	$D8,$69,$F8,$97,$F7,$69,$FF,$09,$9F,$B1,$71,$E5,$BD,$FF,$E5,$31
	.dc.b	$87,$BC,$FF,$1F,$8C,$73,$FF,$28,$6F,$8D,$CE,$57,$CE,$BF,$DE,$D1
	.dc.b	$C4,$CD,$A1,$E9,$7E,$3F,$0E,$AF,$D9,$AE,$17,$CE,$3F,$23,$D3,$8C
	.dc.b	$39,$47,$E5,$39,$F1,$8F,$4E,$BF,$CB,$F9,$1F,$C7,$F3,$EB,$1F,$9A
	.dc.b	$F2,$71,$A7,$92,$F8,$4B,$E1,$5D,$3F,$2A,$B5,$9F,$E8,$AB,$52,$F4
	.dc.b	$7E,$8D,$7B,$BA,$3E,$33,$D7,$D9,$F8,$DC,$3E,$B8,$FE,$8F,$AF,$91
	.dc.b	$F4,$E5,$D5,$F2,$CF,$C1,$CB,$F6,$39,$E3,$83,$E0,$C6,$34,$FA,$3D
	.dc.b	$BF,$E2,$6C,$3C,$87,$DC,$3E,$AE,$97,$17,$65,$C5,$F2,$3E,$89,$F0
	.dc.b	$65,$FD,$83,$1E,$91,$8C,$FF,$80,$65,$F0,$23,$1F,$33,$F1,$DD,$69
	.dc.b	$FB,$2B,$E3,$97,$A3,$1F,$93,$E8,$71,$8C,$65,$F4,$3C,$9C,$71,$FD
	.dc.b	$48,$FA,$7B,$63,$83,$E9,$47,$D6,$BC,$63,$F0,$F8,$3E,$9F,$1F,$EB
	.dc.b	$38,$3E,$79,$F5,$FC,$5F,$89,$0F,$6F,$FB,$BC,$FE,$6E,$31,$EF,$FD
	.dc.b	$A7,$5D,$1F,$02,$7F,$39,$18,$F8,$FE,$17,$FE,$00,$F9,$2F,$CE,$57
	.dc.b	$E1,$63,$1F,$2D,$E2,$BE,$75,$FD,$3B,$B7,$9D,$71,$F3,$FF,$A8,$E1
	.dc.b	$F9,$6B,$DC,$7F,$54,$F2,$B1,$5B,$8F,$F0,$CD,$7E,$5E,$B7,$1F,$E1
	.dc.b	$CD,$B1,$5B,$E6,$FE,$B5,$E9,$FC,$68,$F8,$67,$A7,$0F,$33,$EB,$F2
	.dc.b	$AE,$ED,$7C,$B5,$E2,$8C,$65,$FE,$18,$8C,$4F,$F6,$A8,$F1,$7B,$9F
	.dc.b	$A6,$3F,$C7,$5D,$FF,$EA,$9F,$03,$C3,$DB,$F3,$F3,$FF,$14,$E7,$F3
	.dc.b	$33,$EB,$C5,$F1,$9E,$F9,$7E,$7F,$4F,$54,$63,$66,$5F,$02,$31,$EF
	.dc.b	$5D,$BC,$95,$FD,$EF,$8B,$8C,$63,$87,$FA,$36,$FF,$EB,$DD,$BE,$44
	.dc.b	$7F,$A4,$BB,$BF,$5D,$C7,$F3,$B1,$FF,$02,$79,$7C,$B7,$FF,$04,$C7
	.dc.b	$F8,$1F,$CB,$C4,$7D,$3C,$2C,$FF,$82,$FE,$42,$C3,$E4,$3F,$85,$F3
	.dc.b	$FC,$6E,$33,$F9,$67,$F9,$83,$FE,$47,$7F,$87,$BF,$E4,$9F,$0E,$FA
	.dc.b	$9F,$CD,$39,$36,$FF,$36,$F7,$2C,$FF,$8F,$BD,$2F,$D2,$DF,$1F,$69
	.dc.b	$F3,$B8,$C3,$F9,$EC,$3E,$6F,$F4,$59,$F9,$A7,$31,$FE,$03,$E8,$FE
	.dc.b	$5B,$87,$38,$FF,$03,$71,$E2,$B7,$92,$3F,$C7,$5C,$FF,$E7,$69,$CC
	.dc.b	$3C,$EF,$F1,$87,$3F,$F1,$DC,$7A,$61,$C6,$5E,$1F,$C8,$7F,$90,$4F
	.dc.b	$C3,$7D,$FA,$CC,$FD,$9B,$FF,$97,$9F,$2C,$FF,$03,$C7,$F9,$57,$BB
	.dc.b	$F4,$70,$7B,$1D,$35,$FF,$06,$F0,$7C,$CE,$AF,$9F,$9F,$F9,$A3,$C0
	.dc.b	$EB,$AF,$C7,$7F,$B1,$63,$F0,$F5,$B8,$FC,$2E,$7F,$89,$87,$8F,$81
	.dc.b	$F0,$F8,$CB,$E8,$47,$C7,$E7,$FE,$A9,$97,$C2,$E3,$7C,$7E,$33,$F9
	.dc.b	$3C,$7F,$A2,$E3,$F9,$11,$8F,$9D,$FB,$1A,$F7,$C3,$E7,$7C,$EE,$35
	.dc.b	$EF,$1F,$47,$1C,$70,$7C,$6F,$8F,$BB,$FD,$1F,$8F,$F0,$F2,$FF,$84
	.dc.b	$4F,$1E,$7F,$E1,$1F,$77,$FC,$25,$10,$9C,$1F,$E0,$87,$FC,$BC,$64
	.dc.b	$FF,$0C,$39,$8F,$F7,$EC,$74,$3F,$CC,$5F,$0B,$FE,$46,$78,$19,$84
	.dc.b	$48,$CC,$F3,$8C,$F9,$78,$BF,$34,$70,$31,$5D,$9D,$BE,$88,$DC,$6E
	.dc.b	$30,$6D,$DB,$EA,$8D,$C6,$E3,$06,$DD,$D7,$95,$F1,$1E,$98,$F6,$38
	.dc.b	$3C,$71,$B7,$11,$A7,$47,$F9,$AA,$34,$F1,$7E,$9F,$47,$C8,$9F,$AB
	.dc.b	$1F,$5A,$BF,$3B,$1F,$20,$FF,$23,$BB,$7E,$F9,$F6,$7B,$79,$4F,$18
	.dc.b	$FB,$87,$DB,$70,$F4,$3B,$9F,$E5,$22,$1F,$84,$29,$B3,$BC,$FE,$24
	.dc.b	$E8,$E8,$E9,$1D,$9E,$2D,$7F,$CC,$53,$E0,$7C,$A7,$92,$34,$E8,$EE
	.dc.b	$7E,$32,$7F,$26,$FF,$9B,$67,$2E,$67,$F3,$B8,$FC,$58,$F8,$BD,$B6
	.dc.b	$7F,$42,$FE,$5E,$3E,$60,$7F,$3B,$1A,$C7,$53,$F0,$CD,$3F,$83,$71
	.dc.b	$3F,$97,$CB,$9B,$F8,$88,$F8,$A7,$09,$FC,$9B,$EA,$30,$F3,$7C,$EF
	.dc.b	$E4,$27,$AB,$F8,$C8,$E1,$3F,$8C,$9F,$94,$FE,$3A,$69,$99,$E6,$FE
	.dc.b	$49,$AF,$E6,$66,$BF,$9B,$DF,$F6,$D9,$7C,$7A,$FE,$11,$97,$E3,$F1
	.dc.b	$F9,$B3,$67,$F8,$61,$F8,$53,$D9,$97,$FB,$A6,$7E,$9C,$7D,$B3,$E5
	.dc.b	$47,$DE,$8C,$C6,$4E,$27,$E1,$DC,$B9,$6B,$F8,$86,$BF,$A6,$6C,$FE
	.dc.b	$36,$7F,$59,$5D,$2B,$F8,$E4,$D7,$F8,$F4,$7E,$AA,$7F,$59,$3F,$D7
	.dc.b	$FE,$1F,$F9,$56,$1F,$C6,$67,$E4,$5F,$F2,$A8,$C6,$3F,$C1,$BE,$17
	.dc.b	$F9,$64,$6E,$7F,$6C,$F1,$CE,$D7,$F1,$E9,$FE,$01,$3C,$A6,$CF,$F1
	.dc.b	$C9,$FE,$57,$3D,$A3,$F2,$71,$FE,$CB,$8C,$46,$8F,$E3,$B3,$FC,$E6
	.dc.b	$33,$14,$FE,$42,$7F,$B5,$E7,$93,$8F,$83,$FD,$24,$F0,$7F,$90,$36
	.dc.b	$3F,$9F,$35,$E3,$3F,$CA,$67,$F4,$53,$99,$FF,$04,$1E,$0F,$87,$F8
	.dc.b	$AF,$8B,$0F,$E3,$B3,$E6,$7F,$8B,$B6,$FF,$BE,$B7,$3E,$F9,$7F,$31
	.dc.b	$AF,$EB,$7A,$76,$BF,$E7,$51,$FC,$A6,$FF,$AA,$5F,$F0,$35,$F0,$3F
	.dc.b	$98,$AF,$A2,$FF,$65,$D3,$EB,$BF,$56,$3E,$8B,$FD,$52,$37,$1F,$D4
	.dc.b	$CC,$9F,$D5,$5B,$8F,$F0,$18,$FC,$1C,$FD,$46,$9F,$D3,$CB,$1F,$8C
	.dc.b	$7E,$E4,$7D,$6C,$B2,$66,$FD,$33,$FD,$5E,$3E,$AC,$ED,$FA,$AD,$79
	.dc.b	$C6,$DF,$ED,$F3,$CA,$BF,$84,$47,$F7,$68,$FE,$F5,$C1,$F1,$F0,$5B
	.dc.b	$1F,$DF,$3F,$91,$8F,$EA,$6F,$E6,$9F,$E9,$59,$D8,$FF,$06,$4D,$3E
	.dc.b	$BB,$F9,$59,$FE,$9F,$1F,$9E,$7F,$A7,$CE,$67,$FA,$2C,$D3,$F6,$EF
	.dc.b	$EC,$67,$FA,$84,$7E,$C1,$FE,$A5,$1F,$8C,$6C,$6E,$3F,$A7,$9E,$2B
	.dc.b	$FC,$2B,$FD,$34,$FF,$2F,$4F,$F4,$D8,$CC,$7F,$51,$7F,$C3,$D8,$FC
	.dc.b	$58,$FF,$50,$1E,$93,$FA,$54,$FF,$00,$3E,$4F,$07,$F8,$A2,$BF,$C4
	.dc.b	$CF,$F4,$29,$F2,$3F,$E1,$37,$D4,$7F,$47,$9E,$91,$F8,$B8,$DE,$5F
	.dc.b	$D3,$DD,$C7,$F4,$57,$EA,$1F,$D3,$23,$FA,$CD,$78,$1D,$35,$FE,$9D
	.dc.b	$1C,$27,$FA,$53,$63,$FA,$54,$ED,$FF,$0E,$B5,$FE,$AC,$D7,$FC,$4E
	.dc.b	$BE,$0B,$F9,$4D,$FF,$4A,$AF,$C2,$61,$F4,$9F,$A8,$FF,$41,$69,$93
	.dc.b	$E9,$87,$F1,$2F,$C6,$7F,$C0,$F5,$F5,$23,$27,$C2,$8F,$F0,$B3,$B3
	.dc.b	$FC,$C1,$95,$7F,$C2,$D3,$C5,$C9,$C4,$FE,$8E,$53,$F2,$2E,$DE,$AD
	.dc.b	$7A,$B5,$FE,$25,$D5,$FD,$28,$FD,$5E,$7F,$EA,$B8,$FE,$83,$18,$9F
	.dc.b	$D6,$9F,$EC,$B9,$FF,$58,$79,$5A,$AF,$D9,$4B,$D8,$71,$FF,$46,$C3
	.dc.b	$94,$63,$BF,$FA,$14,$62,$7F,$D5,$37,$FD,$5A,$31,$E1,$F8,$E3,$FA
	.dc.b	$64,$63,$BF,$51,$E0,$C3,$BD,$25,$E1,$7F,$63,$BE,$1E,$CD,$C3,$86
	.dc.b	$3F,$EB,$8B,$7B,$EF,$17,$BD,$6B,$2E,$92,$F9,$4C,$F0,$7F,$B4,$75
	.dc.b	$F0,$E3,$CB,$E6,$9D,$FF,$ED,$5D,$BD,$18,$A5,$F0,$99,$D3,$D4,$7C
	.dc.b	$73,$D9,$CD,$73,$ED,$BF,$C3,$9F,$0F,$1F,$CF,$8F,$56,$3E,$DA,$DE
	.dc.b	$FE,$8C,$B7,$B7,$9A,$BC,$5E,$B7,$87,$AF,$FA,$96,$DF,$82,$E5,$FF
	.dc.b	$02,$BF,$84,$F0,$2E,$DF,$7B,$9B,$F0,$1A,$FC,$1E,$BF,$37,$07,$8F
	.dc.b	$C5,$FB,$B7,$F2,$72,$FA,$F7,$E4,$F9,$FB,$7E,$9E,$1F,$B1,$EB,$7A
	.dc.b	$63,$1C,$FF,$53,$97,$D3,$CB,$EE,$E5,$78,$DF,$27,$3F,$7D,$FD,$1E
	.dc.b	$AF,$99,$EC,$7B,$37,$72,$FD,$D7,$F3,$0F,$DF,$D6,$D3,$E0,$6F,$6B
	.dc.b	$E8,$FE,$A5,$FE,$16,$1E,$37,$F8,$83,$85,$D3,$FD,$52,$B5,$2F,$84
	.dc.b	$F8,$71,$8F,$7B,$F9,$0D,$6F,$36,$3D,$25,$EE,$7E,$7A,$F8,$2E,$8F
	.dc.b	$CD,$46,$3E,$2D,$DD,$F9,$88,$C7,$D1,$5C,$BA,$EA,$CB,$AE,$0A,$FF
	.dc.b	$7D,$69,$E3,$D3,$C5,$A2,$B5,$DF,$F8,$7C,$BF,$5B,$C7,$F1,$1F,$4B
	.dc.b	$6F,$FC,$86,$ED,$F8,$EF,$C2,$8F,$1E,$56,$7E,$04,$7B,$CC,$E1,$FC
	.dc.b	$07,$83,$E6,$DF,$DA,$9C,$E3,$F0,$BA,$BF,$2B,$B7,$EB,$6B,$EC,$ED
	.dc.b	$EF,$FA,$2E,$99,$FB,$30,$FD,$AD,$79,$6B,$E2,$67,$EF,$F9,$9F,$91
	.dc.b	$7B,$E0,$FA,$1A,$F8,$B3,$F9,$BF,$3C,$87,$D1,$FF,$85,$3E,$23,$FC
	.dc.b	$78,$FF,$85,$BD,$9F,$87,$AF,$F9,$96,$31,$F1,$7F,$E7,$27,$5C,$1E
	.dc.b	$49,$FF,$33,$46,$3E,$97,$FC,$E6,$EB,$E1,$B5,$3F,$E7,$F8,$C7,$99
	.dc.b	$FC,$16,$3C,$91,$8F,$94,$E7,$0C,$98,$B0,$FE,$2F,$4E,$FF,$F4,$3E
	.dc.b	$1C,$7D,$BB,$8F,$1C,$38,$64,$C1,$5F,$E6,$8B,$E1,$E8,$79,$70,$E1
	.dc.b	$97,$6A,$FE,$8D,$E0,$7C,$39,$CE,$3C,$E7,$39,$FA,$A7,$39,$76,$F3
	.dc.b	$BF,$69,$F2,$99,$DB,$E7,$D7,$B7,$B3,$E6,$3E,$73,$E4,$9E,$23,$FD
	.dc.b	$33,$B7,$C8,$C7,$E2,$D7,$C4,$D7,$E1,$F9,$DF,$02,$31,$AF,$97,$3F
	.dc.b	$64,$63,$4F,$27,$6F,$AE,$3E,$6E,$9E,$07,$A3,$E0,$D3,$E1,$F9,$16
	.dc.b	$7A,$FC,$67,$8B,$8F,$E6,$E2,$FA,$5F,$14,$63,$8B,$F6,$38,$FC,$18
	.dc.b	$C6,$9F,$DC,$23,$1B,$FC,$F8,$FE,$35,$D1,$8D,$FD,$5C,$8F,$9A,$DF
	.dc.b	$C1,$F8,$F2,$FC,$27,$07,$F2,$7C,$FD,$F1,$8E,$8F,$36,$9F,$CC,$9F
	.dc.b	$EA,$99,$FE,$8A,$BD,$B0,$F1,$3D,$35,$B7,$F0,$D3,$A5,$F1,$BD,$A7
	.dc.b	$4F,$F8,$4E,$FA,$71,$FD,$5C,$98,$F5,$A7,$A3,$F4,$73,$F3,$65,$D3
	.dc.b	$CD,$FE,$F9,$97,$29,$CD,$FF,$81,$F0,$F8,$F8,$FE,$B3,$4B,$1D,$70
	.dc.b	$E9,$A7,$91,$FE,$FB,$8F,$38,$FE,$A7,$F4,$1F,$13,$0F,$2F,$97,$F0
	.dc.b	$1F,$3F,$83,$71,$F8,$9D,$7E,$F9,$F5,$ED,$FE,$29,$D3,$BC,$63,$5F
	.dc.b	$E8,$11,$E4,$9F,$66,$5F,$7B,$E3,$3E,$C6,$87,$F0,$B5,$E8,$5F,$35
	.dc.b	$F6,$17,$C3,$74,$9D,$2F,$92,$FA,$F3,$A7,$EB,$F2,$7E,$96,$31,$EE
	.dc.b	$FF,$BC,$3C,$2F,$E2,$91,$8F,$07,$FC,$63,$C9,$F1,$63,$1F,$17,$E7
	.dc.b	$F7,$BC,$59,$CC,$FF,$CD,$5A,$78,$9F,$F3,$66,$2A,$F7,$57,$97,$34
	.dc.b	$BE,$2F,$F3,$FB,$FC,$2C,$FD,$BE,$9F,$AB,$07,$CA,$5C,$FF,$CF,$1F
	.dc.b	$07,$FE,$CA,$DD,$F0,$FF,$EB,$6D,$FF,$D9,$A7,$F3,$D7,$FC,$CB,$D1
	.dc.b	$88,$86,$4F,$F6,$96,$B8,$78,$D7,$47,$FD,$AD,$F4,$FF,$3D,$E5,$19
	.dc.b	$F6,$D3,$E6,$9E,$53,$98,$FA,$9C,$18,$C7,$FC,$EF,$1F,$5F,$E1,$F1
	.dc.b	$7D,$B4,$47,$C9,$8E,$F5,$C2,$73,$F1,$DE,$AC,$7B,$46,$39,$37,$7C
	.dc.b	$F0,$5B,$D1,$CE,$9E,$8C,$39,$F4,$7F,$84,$B0,$F9,$99,$F7,$C7,$8C
	.dc.b	$FF,$83,$E7,$27,$5C,$7F,$AA,$7D,$3F,$1D,$3C,$77,$C2,$1F,$45,$C3
	.dc.b	$3F,$04,$E6,$3F,$BA,$E5,$DF,$0E,$70,$F8,$2F,$AD,$7F,$0F,$2F,$8E
	.dc.b	$FD,$07,$EC,$1A,$DF,$F4,$4E,$9F,$9D,$7F,$DA,$63,$EE,$E1,$DF,$46
	.dc.b	$78,$9F,$64,$63,$C3,$F4,$61,$CA,$37,$9F,$F8,$27,$5F,$C3,$7A,$BE
	.dc.b	$2C,$37,$D1,$E4,$9F,$AB,$F1,$5E,$3E,$1F,$BB,$F1,$5F,$67,$6F,$17
	.dc.b	$9B,$E1,$E1,$F8,$BC,$3F,$0B,$E1,$7D,$98,$7E,$13,$A7,$86,$BF,$77
	.dc.b	$E4,$EE,$1F,$31,$F8,$D1,$C3,$99,$1F,$82,$8E,$DC,$7F,$91,$8F,$C2
	.dc.b	$74,$E3,$1F,$C8,$C7,$E1,$FE,$03,$18,$72,$E9,$FF,$29,$71,$7C,$4B
	.dc.b	$DC,$7F,$95,$7C,$DC,$56,$EF,$F9,$EF,$5F,$8F,$1E,$95,$FD,$23,$1F
	.dc.b	$C3,$C7,$E5,$74,$EF,$EB,$FB,$35,$FD,$37,$C9,$DD,$7B,$FC,$1F,$93
	.dc.b	$0F,$37,$67,$81,$F4,$7C,$1D,$E0,$F2,$71,$F8,$5F,$37,$AB,$78,$7F
	.dc.b	$A8,$FB,$9F,$1F,$83,$B7,$1F,$F8,$97,$87,$F4,$FC,$1F,$B7,$F6,$7E
	.dc.b	$6D,$7E,$79,$75,$EB,$87,$F7,$31,$8D,$7E,$4B,$AE,$2B,$47,$FC,$71
	.dc.b	$8F,$D7,$9D,$7B,$3F,$29,$E1,$EC,$FF,$58,$ED,$FD,$A4,$63,$CA,$E1
	.dc.b	$8F,$3C,$B8,$3D,$6E,$E5,$FE,$3E,$F2,$BF,$09,$ED,$7B,$B1,$FE,$73
	.dc.b	$A7,$0E,$9F,$2F,$C8,$F9,$B0,$FF,$22,$FB,$9D,$9F,$F2,$3F,$AD,$73
	.dc.b	$7A,$7F,$8F,$EC,$F9,$3F,$09,$3C,$0F,$89,$D1,$FE,$93,$F4,$7F,$AC
	.dc.b	$74,$7F,$06,$A4,$3E,$7B,$F1,$F1,$87,$F9,$14,$EE,$75,$D7,$F8,$EA
	.dc.b	$D3,$FB,$3E,$8F,$E3,$11,$8F,$53,$F4,$4F,$BB,$8E,$DF,$2E,$3B,$E5
	.dc.b	$78,$21,$F2,$D7,$4F,$F5,$3E,$9E,$15,$F2,$3F,$C8,$BC,$DF,$F3,$C6
	.dc.b	$5F,$C2,$5F,$C4,$3F,$CF,$1F,$F3,$77,$07,$F0,$3D,$BF,$D9,$DD,$1F
	.dc.b	$D9,$AB,$F7,$75,$FE,$73,$CD,$F2,$1C,$6B,$E2,$FC,$1F,$F3,$5B,$ED
	.dc.b	$18,$F9,$AE,$2D,$EF,$FE,$83,$8F,$3B,$57,$F8,$22,$F8,$E2,$9C,$FB
	.dc.b	$1F,$CB,$77,$FE,$37,$97,$58,$E1,$7F,$D9,$F1,$E9,$8F,$F8,$4B,$6F
	.dc.b	$F5,$9E,$1C,$32,$E9,$39,$EF,$F8,$D9,$73,$B7,$C0,$FF,$80,$B0,$E1
	.dc.b	$A7,$8A,$73,$F1,$1D,$F9,$7F,$96,$57,$5D,$7E,$26,$9F,$E5,$D7,$D7
	.dc.b	$86,$0F,$EE,$CF,$9F,$2F,$8D,$C5,$F1,$1F,$14,$79,$2B,$CF,$BF,$DC
	.dc.b	$D3,$18,$78,$DF,$F2,$8F,$17,$D8,$AD,$EB,$EA,$F9,$6F,$F0,$57,$C6
	.dc.b	$7F,$41,$CB,$E3,$78,$7E,$8C,$38,$64,$F9,$BF,$AF,$97,$C6,$F7,$FC
	.dc.b	$8F,$8B,$9B,$D7,$CB,$F3,$7C,$4F,$1F,$AD,$C3,$13,$FE,$46,$E4,$F9
	.dc.b	$2F,$F9,$1B,$5F,$86,$FF,$92,$22,$E5,$FE,$F2,$9F,$89,$97,$E0,$6B
	.dc.b	$FC,$9B,$E9,$7F,$79,$D7,$CF,$18,$F9,$DF,$0E,$1F,$E7,$6D,$BF,$BB
	.dc.b	$E2,$EA,$FC,$75,$FF,$24,$C3,$8D,$FF,$9F,$2B,$8E,$3F,$E6,$2C,$FC
	.dc.b	$F3,$CE,$7C,$7A,$7D,$B6,$FC,$1F,$BF,$1F,$EF,$D1,$E2,$F1,$3D,$78
	.dc.b	$FC,$BF,$03,$8D,$7A,$B5,$7C,$CF,$4F,$91,$F9,$07,$C7,$DB,$E9,$AF
	.dc.b	$83,$FA,$3A,$A3,$1F,$31,$F8,$C6,$F7,$FE,$11,$CF,$CC,$5B,$53,$E4
	.dc.b	$BF,$0F,$C1,$F1,$7C,$0F,$DC,$E1,$D2,$5F,$51,$F0,$3D,$CF,$2F,$87
	.dc.b	$E2,$AF,$0F,$6F,$5C,$BA,$65,$EB,$CF,$BE,$5F,$D0,$E3,$B7,$5F,$EE
	.dc.b	$F0,$ED,$F4,$3F,$33,$E5,$BE,$0E,$1F,$BD,$E2,$F8,$2F,$F8,$8F,$07
	.dc.b	$5B,$44,$64,$FF,$2D,$B8,$F8,$4F,$E3,$F1,$FE,$71,$8C,$7C,$67,$ED
	.dc.b	$9B,$BF,$F8,$85,$AF,$29,$52,$BF,$D0,$DF,$19,$98,$C5,$7F,$A2,$78
	.dc.b	$7F,$CB,$D8,$72,$F2,$6E,$3C,$B1,$FF,$47,$5A,$7F,$49,$5B,$F9,$7F
	.dc.b	$F3,$BB,$79,$FF,$CF,$AE,$7D,$4B,$8F,$3C,$3F,$D3,$18,$30,$74,$7E
	.dc.b	$67,$0E,$8F,$F5,$7B,$F9,$31,$8F,$3B,$FD,$52,$DF,$5B,$33,$D7,$1F
	.dc.b	$CF,$70,$7C,$F8,$C6,$5D,$EF,$C7,$BF,$C2,$BF,$66,$7E,$2B,$FA,$07
	.dc.b	$07,$CB,$97,$F8,$FB,$1E,$9E,$FF,$F1,$37,$B6,$1F,$C9,$2F,$AF,$C7
	.dc.b	$7D,$77,$E5,$65,$F5,$DF,$99,$C5,$6B,$E9,$CF,$D5,$E1,$F9,$FF,$1F
	.dc.b	$F4,$A7,$CD,$B7,$F8,$6B,$0F,$2C,$63,$E5,$3E,$8C,$79,$F8,$7F,$D0
	.dc.b	$D8,$FC,$AF,$53,$79,$7D,$AC,$7E,$85,$7F,$80,$F9,$3F,$2B,$CD,$FE
	.dc.b	$BD,$C7,$FD,$80,$FE,$5B,$E1,$BF,$CB,$F1,$FE,$B4,$8C,$7B,$BF,$D2
	.dc.b	$1B,$C3,$FE,$73,$73,$95,$A5,$FF,$8E,$A3,$F8,$E4,$7F,$AB,$67,$7E
	.dc.b	$5F,$FB,$D2,$3C,$67,$0C,$3F,$35,$1F,$C5,$A3,$CA,$66,$BF,$93,$7C
	.dc.b	$0F,$F3,$7B,$E6,$B7,$F4,$56,$EF,$FD,$97,$CD,$FE,$E6,$8E,$79,$79
	.dc.b	$67,$3B,$7F,$B5,$B1,$E7,$6F,$4B,$F9,$74,$61,$79,$7F,$E9,$78,$C2
	.dc.b	$F2,$FD,$78,$F5,$EA,$F8,$19,$F8,$F5,$FD,$13,$DC,$C6,$4F,$3F,$17
	.dc.b	$E5,$AF,$D4,$67,$BD,$7F,$25,$A7,$D7,$F0,$FE,$A6,$5F,$27,$5F,$A5
	.dc.b	$8F,$F9,$AB,$AA,$F1,$7F,$02,$AF,$1C,$FD,$5F,$4B,$FD,$DB,$8F,$D4
	.dc.b	$F8,$DC,$DF,$D2,$D7,$F2,$58,$7B,$23,$1B,$FC,$5F,$3B,$E5,$C6,$37
	.dc.b	$FE,$8F,$5B,$E5,$FA,$F7,$F2,$F6,$F9,$3C,$DF,$D1,$EB,$E0,$DF,$CC
	.dc.b	$F8,$5F,$C1,$F1,$7E,$16,$7E,$06,$9E,$E8,$7A,$3F,$D7,$67,$F0,$B1
	.dc.b	$8F,$9E,$FC,$93,$7C,$BF,$6C,$FF,$6D,$A5,$7F,$4E,$EB,$FA,$B7,$BF
	.dc.b	$A9,$B5,$AC,$11,$F8,$0C,$3E,$94,$E6,$1E,$E7,$F8,$2A,$7C,$73,$98
	.dc.b	$7A,$9C,$32,$FE,$BB,$7C,$23,$F4,$F7,$F0,$2F,$84,$7F,$48,$F2,$71
	.dc.b	$7C,$30,$E7,$7F,$E5,$CD,$FE,$56,$7F,$B8,$8F,$29,$F0,$B1,$F3,$4F
	.dc.b	$E9,$27,$DD,$56,$7B,$D7,$E4,$6B,$9C,$63,$D1,$FD,$AF,$47,$E4,$2B
	.dc.b	$E5,$65,$D7,$87,$B6,$7E,$1C,$78,$BF,$D7,$EB,$E1,$BF,$DF,$23,$E6
	.dc.b	$BF,$DF,$BD,$1B,$BF,$CE,$D3,$F4,$57,$C2,$31,$EC,$FC,$97,$5B,$FF
	.dc.b	$A6,$BA,$BF,$B6,$C7,$94,$FC,$BD,$7F,$02,$7E,$CB,$FC,$55,$7C,$9F
	.dc.b	$D7,$8F,$DA,$C6,$3E,$7F,$52,$7D,$95,$DA,$31,$DF,$F8,$CA,$E1,$D7
	.dc.b	$CF,$0F,$E2,$F1,$8F,$9B,$EB,$7C,$EB,$A7,$A3,$E5,$8C,$64,$8F,$F0
	.dc.b	$4E,$1C,$A7,$31,$FD,$F6,$31,$3F,$EE,$63,$87,$6F,$F2,$EE,$1F,$D6
	.dc.b	$65,$E4,$9C,$F8,$1F,$77,$5F,$54,$71,$C7,$FC,$11,$2F,$A2,$FE,$67
	.dc.b	$7F,$1B,$2F,$87,$E0,$7C,$4C,$3F,$DE,$B1,$EC,$7F,$DE,$F1,$75,$F9
	.dc.b	$13,$ED,$9F,$37,$35,$9F,$83,$AF,$E8,$35,$F3,$ED,$EE,$8F,$E9,$3E
	.dc.b	$CF,$F4,$4F,$33,$EE,$57,$F1,$ED,$3E,$CF,$0F,$8F,$1F,$97,$E0,$FF
	.dc.b	$2B,$9F,$81,$C2,$F1,$31,$A7,$5F,$1B,$E4,$E1,$E2,$E3,$7A,$9F,$D4
	.dc.b	$B9,$BF,$3D,$3F,$6F,$8F,$FA,$5A,$FA,$2F,$E5,$ED,$DF,$FB,$03,$9F
	.dc.b	$76,$C9,$FE,$35,$3F,$C7,$E7,$5E,$2F,$F1,$46,$C7,$F4,$67,$3E,$B5
	.dc.b	$BE,$54,$F9,$FF,$C3,$63,$FD,$26,$F9,$C7,$F9,$BF,$CB,$C5,$F0,$8F
	.dc.b	$F3,$97,$83,$8C,$D8,$F3,$F6,$3F,$8F,$65,$FD,$CB,$2E,$71,$FC,$9B
	.dc.b	$C4,$FE,$6F,$DF,$FE,$5A,$9C,$E2,$97,$CB,$F1,$C3,$E2,$F7,$FC,$6D
	.dc.b	$7F,$67,$1F,$DE,$B0,$F8,$DC,$9F,$06,$7D,$BE,$57,$BF,$6F,$2E,$F7
	.dc.b	$7F,$81,$DF,$F4,$AF,$D9,$C1,$6F,$6F,$9F,$3F,$54,$EB,$7F,$37,$C6
	.dc.b	$7B,$7C,$1F,$26,$7F,$5B,$17,$9D,$E3,$79,$E7,$F1,$F4,$F8,$DE,$45
	.dc.b	$BF,$97,$BF,$AF,$4E,$DC,$6E,$2F,$F4,$0E,$7F,$E8,$0F,$2F,$FC,$90
	.dc.b	$FF,$8B,$7C,$3F,$F7,$8B,$FE,$2F,$F8,$3B,$C5,$FE,$28,$7F,$C9,$8F
	.dc.b	$23,$1F,$23,$FE,$8A,$61,$FE,$8C,$EB,$F4,$BA,$F5,$B2,$E6,$FB,$53
	.dc.b	$E2,$F3,$3D,$7C,$06,$2A,$D7,$8B,$CB,$2E,$B5,$A9,$E5,$F4,$19,$CB
	.dc.b	$AE,$5C,$76,$F2,$E1,$C3,$2E,$93,$D7,$E5,$7F,$9E,$E9,$E1,$9C,$E4
	.dc.b	$9F,$E2,$53,$E9,$3E,$0D,$71,$F8,$CF,$E8,$B8,$75,$BF,$61,$C6,$1F
	.dc.b	$3F,$C7,$5F,$45,$F1,$F0,$BF,$DF,$F8,$71,$8C,$79,$DF,$0F,$2F,$A7
	.dc.b	$7F,$12,$BE,$3C,$79,$F2,$F9,$51,$F0,$63,$CB,$16,$7B,$BF,$E0,$78
	.dc.b	$F1,$69,$EE,$D3,$E2,$ED,$F2,$2B,$E0,$71,$7D,$DC,$7C,$91,$EB,$C3
	.dc.b	$D5,$A7,$93,$E3,$FC,$B9,$38,$7C,$B8,$7C,$3F,$07,$EE,$BF,$17,$D7
	.dc.b	$BA,$77,$D5,$C5,$FE,$63,$AE,$F1,$8C,$FF,$C0,$55,$FE,$D8,$E7,$FC
	.dc.b	$CE,$BF,$BF,$E5,$FC,$BC,$7E,$1E,$3C,$33,$BF,$98,$FF,$37,$47,$C5
	.dc.b	$94,$3E,$3F,$73,$FC,$26,$31,$7E,$68,$C4,$F5,$7E,$39,$89,$7C,$BF
	.dc.b	$57,$A1,$8E,$9D,$E1,$E0,$BF,$0D,$3C,$8F,$A9,$AF,$EA,$2B,$58,$74
	.dc.b	$97,$89,$F4,$F4,$F6,$63,$CA,$5E,$FF,$EF,$F4,$FD,$B6,$3C,$B3,$ED
	.dc.b	$2F,$A1,$FC,$D4,$E6,$F8,$C3,$4E,$3A,$7A,$B0,$E3,$3F,$EF,$B9,$F3
	.dc.b	$1F,$42,$66,$5F,$E0,$6C,$BD,$F1,$F0,$9F,$89,$1F,$43,$C4,$B3,$F2
	.dc.b	$3C,$1D,$CF,$AE,$BD,$C7,$CD,$C7,$D2,$FA,$2B,$C9,$97,$A3,$E6,$FF
	.dc.b	$8A,$7A,$5F,$84,$D7,$E6,$DF,$D9,$7E,$5F,$D0,$78,$31,$FA,$1E,$17
	.dc.b	$69,$F8,$5C,$1F,$E6,$7D,$FE,$96,$9D,$A1,$F3,$7E,$1D,$BE,$84,$7C
	.dc.b	$BE,$5F,$E4,$9F,$37,$F9,$4F,$87,$FC,$7E,$DF,$B3,$7F,$1C,$FE,$39
	.dc.b	$7C,$DE,$A3,$DF,$18,$F9,$4D,$4B,$E0,$35,$5F,$CC,$6B,$8C,$74,$8F
	.dc.b	$C9,$C6,$27,$A6,$BF,$CC,$A3,$1C,$DF,$56,$B9,$57,$51,$0F,$95,$E7
	.dc.b	$19,$8C,$73,$F9,$A3,$12,$CF,$19,$7B,$2F,$94,$3A,$FF,$94,$65,$E2
	.dc.b	$BE,$1E,$46,$31,$E7,$6F,$8C,$FD,$FE,$8C,$38,$4E,$7D,$2F,$85,$3F
	.dc.b	$D5,$A7,$E5,$BF,$23,$BF,$E4,$EB,$31,$FD,$D5,$7C,$2F,$0F,$DB,$3E
	.dc.b	$7D,$3D,$59,$F8,$E3,$19,$F8,$FE,$8F,$E3,$C7,$D8,$F9,$DF,$B9,$97
	.dc.b	$C2,$F5,$BE,$0B,$F2,$B1,$F8,$99,$73,$8C,$63,$ED,$E1,$FD,$77,$A6
	.dc.b	$DD,$E6,$FD,$57,$FA,$BC,$66,$3F,$9F,$C7,$F6,$D8,$C6,$38,$E5,$FF
	.dc.b	$14,$CF,$2F,$63,$FD,$A4,$F6,$31,$ED,$7F,$34,$BE,$93,$FE,$6D,$E3
	.dc.b	$FE,$6B,$18,$F8,$17,$C8,$E3,$0E,$91,$FE,$54,$AF,$EE,$55,$BC,$3F
	.dc.b	$CA,$F5,$E8,$36,$FF,$96,$BA,$71,$A3,$2F,$F3,$E4,$FF,$97,$B8,$73
	.dc.b	$18,$8F,$C6,$63,$FD,$47,$1E,$D1,$F8,$FF,$4F,$19,$76,$8C,$6B,$F1
	.dc.b	$63,$1C,$1F,$DC,$3E,$23,$1A,$78,$76,$FF,$36,$E1,$CF,$E2,$7F,$C4
	.dc.b	$D8,$FC,$AA,$D4,$FF,$9C,$B0,$FE,$F3,$4F,$93,$F9,$95,$FE,$4A,$CB
	.dc.b	$AE,$DF,$4F,$5F,$9B,$EA,$FC,$F9,$F9,$27,$CB,$AF,$14,$5A,$F8,$73
	.dc.b	$F3,$0F,$93,$16,$7C,$79,$7E,$BE,$7E,$0C,$63,$2F,$2C,$79,$2F,$E5
	.dc.b	$47,$A3,$C0,$C6,$BF,$02,$3D,$7F,$2B,$F8,$5F,$63,$F1,$B8,$6B,$2F
	.dc.b	$C0,$BF,$52,$F8,$73,$F8,$32,$EB,$F4,$56,$7A,$71,$B4,$FA,$59,$7E
	.dc.b	$4B,$1F,$C9,$79,$1D,$B8,$FF,$2F,$3F,$88,$9F,$C7,$3F,$EB,$FF,$27
	.dc.b	$E4,$3F,$EC,$1D,$AE,$5F,$E7,$C9,$DB,$FE,$7E,$F9,$6B,$F0,$3F,$56
	.dc.b	$7E,$55,$FF,$B2,$63,$EC,$5E,$2F,$FD,$0B,$CA,$8F,$E2,$D8,$FE,$32
	.dc.b	$F8,$4E,$7E,$33,$F3,$91,$8A,$7C,$C7,$F6,$FB,$F9,$58,$74,$9C,$FD
	.dc.b	$0F,$38,$C5,$7F,$4A,$E3,$F8,$B1,$8F,$35,$E5,$E3,$5F,$3D,$FF,$50
	.dc.b	$87,$7F,$69,$CD,$70,$F9,$9F,$B5,$D7,$F9,$1D,$5B,$7F,$96,$1F,$8F
	.dc.b	$CB,$F2,$B0,$8F,$C8,$E1,$8F,$C5,$8F,$8F,$97,$C0,$D7,$D3,$7E,$B9
	.dc.b	$F0,$BE,$7B,$F9,$5A,$FF,$66,$EF,$C7,$0F,$C6,$DB,$DB,$F3,$5E,$9B
	.dc.b	$FD,$56,$1D,$71,$F7,$ED,$D6,$79,$DF,$C4,$F3,$7F,$8D,$7C,$67,$CF
	.dc.b	$C3,$F4,$DE,$67,$E2,$FC,$1F,$E5,$53,$FE,$C0,$9F,$D8,$3F,$EC,$2E
	.dc.b	$0B,$9B,$E5,$A9,$FE,$E2,$BE,$31,$9E,$0F,$E4,$17,$FE,$CA,$AF,$BF
	.dc.b	$7F,$B5,$8C,$4B,$E7,$B1,$7C,$57,$95,$FC,$8E,$B7,$0F,$A3,$C5,$6D
	.dc.b	$DE,$7F,$C6,$6B,$70,$F9,$BF,$E3,$F1,$8F,$97,$FF,$47,$EB,$E8,$D7
	.dc.b	$F2,$79,$75,$9F,$E6,$7E,$07,$F4,$4B,$E1,$E2,$F9,$23,$1A,$78,$69
	.dc.b	$2F,$98,$F9,$FD,$5F,$8B,$C3,$E7,$6C,$7E,$27,$0F,$C6,$C7,$D5,$C7
	.dc.b	$E6,$7A,$57,$AD,$ED,$CD,$76,$F3,$F0,$B5,$F0,$E1,$F7,$BA,$3C,$18
	.dc.b	$7E,$D3,$1F,$9B,$C1,$F6,$3C,$0F,$06,$5F,$B2,$C3,$E8,$BF,$B4,$E9
	.dc.b	$BC,$DF,$57,$8F,$E7,$CF,$C6,$FA,$2B,$3F,$67,$6F,$17,$9B,$E5,$8C
	.dc.b	$6F,$F1,$3E,$73,$FC,$C9,$A7,$F7,$3C,$7F,$B8,$3F,$DD,$78,$AE,$FF
	.dc.b	$DE,$8B,$E5,$FE,$3F,$C0,$FF,$51,$D7,$17,$D9,$3F,$EA,$D8,$C7,$AD
	.dc.b	$FE,$A6,$6C,$FF,$B7,$DB,$CF,$FE,$DA,$73,$EC,$DB,$FF,$04,$E5,$FC
	.dc.b	$03,$0F,$E0,$33,$F8,$C9,$F4,$9A,$C3,$FD,$E0,$7F,$A8,$B4,$F1,$46
	.dc.b	$6B,$8C,$7F,$24,$E1,$63,$E0,$8B,$5D,$33,$F7,$5F,$0D,$BF,$DB,$77
	.dc.b	$B5,$F4,$9B,$D3,$CD,$9F,$F9,$4F,$3E,$78,$F2,$8F,$F5,$AF,$27,$92
	.dc.b	$31,$AF,$E1,$F7,$FF,$14,$5F,$0C,$39,$CF,$E2,$F2,$CE,$7E,$1F,$23
	.dc.b	$CD,$CD,$F5,$33,$FA,$FA,$7F,$53,$AF,$B9,$9F,$D6,$DF,$E5,$F9,$57
	.dc.b	$6F,$B5,$E0,$7D,$59,$F9,$5F,$0B,$67,$E0,$DF,$D7,$9F,$95,$76,$7D
	.dc.b	$9F,$13,$F5,$DF,$5F,$85,$ED,$E1,$F4,$F2,$FF,$4C,$FA,$AF,$AF,$B7
	.dc.b	$F9,$8F,$2F,$23,$F8,$87,$F0,$0F,$F9,$9F,$E0,$33,$ED,$9F,$F7,$F4
	.dc.b	$FD,$C9,$D7,$2F,$A5,$BE,$17,$9D,$CD,$7F,$A6,$39,$FE,$EC,$63,$D8
	.dc.b	$EF,$8F,$5C,$3B,$7D,$17,$FB,$63,$97,$FC,$B5,$7C,$A3,$FC,$05,$CB
	.dc.b	$F4,$B4,$FE,$A7,$87,$1C,$D5,$FE,$07,$EF,$D6,$1C,$7B,$7F,$93,$63
	.dc.b	$F0,$B0,$E1,$7B,$BC,$61,$FE,$67,$C6,$E7,$F0,$5F,$3C,$7F,$BF,$8E
	.dc.b	$EF,$C3,$ED,$FC,$0E,$3F,$B8,$E6,$FC,$14,$F9,$B9,$FE,$5F,$0F,$CC
	.dc.b	$C3,$FE,$00,$4F,$CB,$D3,$E1,$57,$C9,$7C,$BD,$79,$AF,$06,$AF,$A0
	.dc.b	$C6,$BF,$23,$7F,$F7,$76,$5E,$8B,$F8,$70,$F0,$7E,$1E,$5F,$02,$3F
	.dc.b	$A4,$72,$7E,$EA,$9E,$BF,$F1,$07,$E1,$61,$F3,$DF,$59,$AD,$7F,$38
	.dc.b	$F0,$F9,$4B,$4F,$6B,$FD,$9B,$87,$EF,$2B,$AC,$7C,$8C,$3F,$87,$46
	.dc.b	$36,$FF,$09,$C6,$2F,$94,$30,$FC,$57,$37,$FA,$76,$F8,$43,$E4,$78
	.dc.b	$63,$F8,$28,$C7,$6F,$F1,$2A,$DD,$7F,$A9,$78,$BD,$99,$76,$BE,$59
	.dc.b	$BD,$5C,$E5,$CF,$0F,$F5,$67,$9F,$FE,$7C,$D3,$C5,$8F,$19,$CC,$FC
	.dc.b	$B8,$F8,$58,$7E,$F7,$3F,$3B,$E6,$D7,$D9,$E8,$DD,$3C,$38,$78,$27
	.dc.b	$E0,$63,$E0,$F8,$7F,$AD,$EE,$7D,$5E,$3F,$E8,$BD,$1F,$53,$5F,$A1
	.dc.b	$DD,$DD,$F6,$7C,$F7,$5C,$7E,$6E,$DE,$5E,$1F,$62,$F3,$FC,$1B,$F8
	.dc.b	$9F,$E7,$B2,$F8,$0F,$EA,$6F,$FB,$F2,$BF,$8E,$57,$E7,$A3,$5E,$47
	.dc.b	$F1,$F7,$87,$C4,$DB,$7A,$8E,$4F,$E4,$15,$D6,$1F,$07,$8A,$DB,$FC
	.dc.b	$57,$AF,$FA,$2D,$6E,$1F,$13,$E9,$D3,$FA,$7C,$E6,$3D,$9D,$98,$C3
	.dc.b	$9C,$63,$C3,$FE,$15,$5B,$DF,$FD,$CF,$E4,$62,$F7,$18,$DF,$E1,$C6
	.dc.b	$31,$FF,$07,$78,$BF,$1B,$7F,$8B,$A7,$C6,$AF,$9E,$FF,$6B,$CB,$DC
	.dc.b	$FB,$77,$F7,$F8,$1F,$CF,$67,$D1,$3B,$AE,$D9,$79,$BD,$BE,$98,$7E
	.dc.b	$B3,$BF,$FD,$FD,$87,$C9,$E5,$E9,$C5,$F1,$34,$E5,$7F,$33,$89,$F2
	.dc.b	$18,$AF,$93,$C3,$F8,$4E,$4F,$9B,$B7,$C0,$F2,$7E,$46,$3C,$69,$C7
	.dc.b	$FE,$49,$E1,$F8,$73,$98,$FF,$18,$72,$E3,$2E,$B1,$FE,$33,$F3,$99
	.dc.b	$FF,$A7,$B0,$E7,$ED,$FF,$9C,$51,$FD,$CB,$D1,$B7,$FE,$BB,$8C,$92
	.dc.b	$B9,$47,$FB,$0D,$E9,$A7,$FB,$0E,$B8,$DB,$E9,$78,$E1,$2F,$8C,$FE
	.dc.b	$FF,$C7,$CC,$F8,$4F,$2F,$8F,$FE,$8C,$78,$61,$D2,$3F,$D7,$DC,$7C
	.dc.b	$30,$E7,$1F,$EC,$1D,$71,$5B,$F6,$1F,$0D,$FD,$8E,$30,$BE,$23,$AE
	.dc.b	$5E,$0A,$DC,$7C,$4E,$1F,$27,$03,$6F,$E2,$D5,$AB,$FF,$30,$1F,$3B
	.dc.b	$1E,$B8,$7C,$CE,$EF,$F3,$76,$BF,$02,$BD,$E7,$4E,$4A,$F6,$29,$CC
	.dc.b	$F8,$67,$08,$F7,$F8,$B7,$0F,$74,$FC,$03,$D5,$AD,$D3,$D1,$BF,$FB
	.dc.b	$96,$3E,$04,$FF,$90,$EB,$CE,$FF,$9E,$7C,$AF,$05,$FB,$F2,$F2,$E5
	.dc.b	$F3,$B5,$FE,$25,$7F,$83,$C3,$E5,$FC,$4F,$CA,$C3,$BD,$7C,$AF,$03
	.dc.b	$57,$E4,$DF,$FC,$D8,$F6,$E0,$F9,$7A,$F9,$3D,$4F,$9F,$18,$AF,$95
	.dc.b	$7F,$0F,$D1,$7B,$7F,$E6,$3C,$7E,$CC,$7F,$B1,$DF,$88,$1F,$CF,$3E
	.dc.b	$2E,$0F,$CB,$39,$9F,$F5,$33,$C2,$7A,$FB,$9F,$C2,$A3,$F2,$91,$8F
	.dc.b	$80,$D5,$7F,$AB,$7D,$DF,$E4,$55,$EE,$AE,$30,$F9,$5F,$A9,$18,$F9
	.dc.b	$3F,$C9,$57,$87,$15,$A7,$FA,$04,$F3,$C3,$B4,$7F,$43,$E1,$C5,$F0
	.dc.b	$8D,$F3,$F9,$AB,$7E,$0F,$EF,$21,$B7,$C9,$9C,$E2,$9F,$EF,$DC,$DF
	.dc.b	$C1,$74,$F9,$BE,$47,$E0,$39,$FF,$82,$C7,$F9,$4A,$FE,$79,$F9,$6A
	.dc.b	$E1,$87,$C3,$8F,$7B,$F1,$7B,$AE,$BE,$79,$FA,$F3,$EC,$EE,$FB,$3D
	.dc.b	$1C,$F2,$EF,$2F,$7B,$C5,$7F,$1F,$4F,$A3,$8F,$C1,$F2,$79,$BF,$0F
	.dc.b	$83,$F9,$66,$3C,$A3,$16,$D3,$E3,$46,$3D,$8F,$E2,$1F,$11,$F1,$B8
	.dc.b	$7B,$AF,$73,$C9,$82,$3F,$08,$FE,$71,$FC,$59,$F3,$BE,$3B,$8F,$21
	.dc.b	$FC,$B4,$F6,$8C,$63,$F9,$6A,$E0,$FF,$92,$A7,$A4,$EB,$E5,$7F,$79
	.dc.b	$5A,$F9,$EF,$EF,$55,$AC,$3F,$DF,$D0,$F9,$2F,$E3,$75,$B7,$F5,$FC
	.dc.b	$DF,$D0,$63,$1D,$3F,$D0,$38,$FF,$87,$46,$31,$FE,$FB,$5B,$CF,$D5
	.dc.b	$5E,$77,$D3,$E3,$79,$1F,$5E,$57,$3F,$1F,$89,$E2,$9C,$E9,$FE,$EB
	.dc.b	$7B,$98,$D3,$FA,$8F,$A5,$8E,$3F,$A3,$9F,$D0,$F4,$3E,$6D,$FE,$F7
	.dc.b	$6E,$99,$FE,$15,$F0,$70,$B5,$FE,$6D,$1F,$16,$31,$A7,$E1,$23,$ED
	.dc.b	$5F,$93,$CB,$D5,$FC,$9D,$7F,$CC,$7C,$1F,$07,$2F,$9B,$8F,$F7,$29
	.dc.b	$FF,$6B,$F8,$3F,$D9,$BE,$73,$F0,$D9,$FF,$35,$8C,$67,$FC,$A1,$D4
	.dc.b	$53,$FB,$DE,$3F,$A9,$8F,$ED,$F1,$4F,$EF,$F5,$FB,$8F,$89,$E9,$18
	.dc.b	$BF,$F3,$3B,$FD,$AE,$BB,$FA,$5F,$B5,$C1,$18,$F8,$3F,$EC,$CD,$E1
	.dc.b	$E1,$7C,$ED,$5F,$E0,$58,$F6,$99,$5D,$6A,$FF,$01,$5F,$58,$C4,$70
	.dc.b	$D1,$A7,$F1,$0C,$7A,$D6,$BC,$6F,$F0,$9E,$1F,$1B,$C2,$F6,$E1,$C3
	.dc.b	$5F,$F4,$66,$BE,$8D,$7C,$51,$E8,$9D,$65,$D3,$4F,$27,$87,$D2,$FE
	.dc.b	$4C,$7C,$FA,$7C,$D7,$CD,$7F,$8D,$4F,$A8,$72,$7F,$70,$79,$47,$8F
	.dc.b	$D8,$F2,$F0,$7A,$FE,$57,$C9,$3E,$2E,$27,$AF,$E5,$AF,$81,$1F,$23
	.dc.b	$B7,$B5,$6F,$91,$F4,$98,$BD,$C7,$CD,$AF,$95,$18,$DB,$7A,$79,$23
	.dc.b	$11,$FC,$F3,$8B,$EF,$CF,$AB,$D3,$7C,$17,$E3,$F8,$3F,$CF,$B7,$67
	.dc.b	$F1,$D1,$FE,$39,$E6,$FE,$7B,$78,$9F,$2F,$81,$F9,$0A,$FF,$42,$EB
	.dc.b	$FE,$5A,$AF,$A9,$18,$9F,$F0,$1C,$7E,$22,$31,$E7,$6A,$F9,$BF,$DD
	.dc.b	$F3,$F3,$CE,$B5,$FE,$41,$18,$AD,$C7,$F9,$62,$3A,$46,$1F,$F2,$DF
	.dc.b	$3F,$DF,$AE,$51,$FE,$5E,$F0,$B1,$5F,$23,$1F,$16,$1F,$0F,$A3,$E0
	.dc.b	$F5,$FB,$F5,$FE,$57,$5F,$A4,$CB,$D9,$E9,$7A,$67,$FD,$6D,$AF,$F3
	.dc.b	$5A,$EB,$18,$F8,$5E,$19,$FA,$3C,$7F,$C6,$63,$ED,$F9,$5B,$3E,$99
	.dc.b	$F5,$DF,$4D,$7F,$CE,$B5,$8E,$1F,$F2,$A4,$7C,$1F,$13,$F3,$9F,$01
	.dc.b	$FE,$0B,$EE,$FE,$14,$FF,$83,$78,$EF,$A5,$FC,$26,$3F,$C2,$7B,$FE
	.dc.b	$A7,$E3,$3F,$39,$5F,$86,$9E,$4F,$F0,$C7,$A7,$55,$BF,$D3,$E9,$F7
	.dc.b	$30,$F0,$C6,$3C,$EF,$DA,$5F,$8E,$BF,$DA,$DE,$4F,$F0,$C8,$C7,$9C
	.dc.b	$F8,$3C,$61,$D2,$3F,$B9,$78,$DF,$B1,$AD,$DB,$E3,$F1,$7B,$8F,$F6
	.dc.b	$C5,$7F,$14,$AD,$BF,$ED,$BE,$DF,$2F,$03,$3F,$1C,$FF,$B7,$FC,$1F
	.dc.b	$93,$87,$C5,$DB,$E3,$65,$EE,$7D,$6F,$C0,$AB,$87,$3D,$2E,$3D,$27
	.dc.b	$F9,$6F,$07,$CB,$B7,$C1,$7D,$2E,$7F,$AB,$F3,$BD,$B2,$FE,$1F,$7B
	.dc.b	$7A,$73,$FC,$47,$B6,$BF,$E6,$C9,$D6,$7E,$AE,$EF,$81,$9F,$C3,$D3
	.dc.b	$E3,$6D,$F0,$2B,$C9,$E1,$7F,$AF,$BD,$7F,$F5,$C9,$FE,$51,$D3,$FD
	.dc.b	$10,$E3,$5F,$F3,$DB,$A8,$B7,$FE,$CE,$F0,$FF,$C7,$93,$FE,$C2,$E3
	.dc.b	$B3,$FE,$78,$39,$9F,$ED,$48,$FF,$2F,$67,$8F,$17,$FC,$B6,$FE,$56
	.dc.b	$7F,$D8,$13,$E4,$8F,$F3,$27,$C9,$7F,$AE,$63,$10,$F9,$AF,$CC,$F0
	.dc.b	$7F,$99,$2F,$B4,$FF,$9A,$E5,$F3,$DD,$27,$33,$C6,$3F,$AB,$47,$FB
	.dc.b	$72,$B9,$63,$E4,$AD,$79,$5F,$ED,$BC,$7C,$38,$F4,$8E,$37,$C2,$7D
	.dc.b	$B3,$99,$E9,$1C,$31,$FE,$83,$1F,$E8,$FA,$E1,$7F,$E9,$06,$E3,$F1
	.dc.b	$6F,$FD,$26,$FD,$73,$E5,$1F,$4B,$1F,$37,$81,$F1,$63,$DB,$95,$9F
	.dc.b	$53,$F0,$B4,$F5,$4E,$B3,$FE,$A7,$1E,$68,$C6,$7F,$D5,$BC,$3F,$06
	.dc.b	$1F,$1A,$7E,$DC,$7F,$06,$CB,$EB,$70,$79,$F5,$FA,$1C,$7F,$3F,$0F
	.dc.b	$0F,$CD,$5B,$F8,$BA,$3E,$0F,$C7,$5E,$68,$FE,$CB,$1E,$38,$C6,$9E
	.dc.b	$6E,$0F,$BD,$A3,$D4,$FC,$14,$79,$1F,$C2,$6D,$63,$F2,$F3,$FE,$B8
	.dc.b	$8F,$4F,$0F,$E5,$A3,$97,$0F,$09,$FC,$34,$63,$D6,$FE,$7F,$1F,$16
	.dc.b	$31,$F4,$7F,$3D,$B8,$70,$C7,$F2,$96,$AF,$DF,$DF,$D0,$BE,$3D,$AC
	.dc.b	$B1,$87,$0B,$FE,$0B,$F1,$B8,$CD,$97,$3C,$13,$C7,$2F,$C2,$65,$D2
	.dc.b	$33,$18,$C7,$A5,$70,$9F,$F0,$17,$02,$BF,$C3,$75,$A9,$EB,$E3,$71
	.dc.b	$CB,$FD,$27,$5C,$F4,$F9,$7D,$5F,$27,$0F,$7C,$FC,$37,$E0,$F5,$5B
	.dc.b	$F4,$ED,$EC,$C9,$E1,$79,$7A,$BE,$AF,$B7,$FC,$23,$D4,$FC,$66,$7E
	.dc.b	$4C,$7E,$0F,$AD,$CF,$8F,$E6,$78,$7F,$3E,$7A,$FC,$5D,$9F,$8D,$CD
	.dc.b	$F2,$5F,$91,$F3,$D7,$7F,$9B,$3F,$6F,$87,$E0,$4F,$E8,$7D,$8F,$F5
	.dc.b	$97,$CE,$FF,$64,$D7,$1E,$D7,$DB,$7F,$8D,$CF,$FA,$DF,$93,$F9,$1B
	.dc.b	$FE,$8F,$8F,$E1,$CF,$F9,$21,$FD,$67,$1F,$B2,$3A,$DF,$2B,$2E,$FF
	.dc.b	$F3,$A7,$4F,$D5,$AF,$1C,$3D,$EF,$DF,$46,$38,$BF,$C5,$7B,$7F,$21
	.dc.b	$8C,$79,$DE,$2C,$BA,$E1,$FE,$F9,$F2,$FF,$DB,$9A,$FB,$34,$F0,$64
	.dc.b	$97,$C9,$FF,$06,$CB,$B4,$3E,$23,$8E,$9D,$F3,$EB,$87,$19,$D6,$5F
	.dc.b	$C5,$F0,$E3,$1C,$7C,$8C,$5F,$0C,$7A,$4B,$E4,$F3,$A7,$CC,$AF,$E9
	.dc.b	$78,$7C,$97,$D7,$8F,$C2,$7C,$3E,$3F,$BE,$BD,$78,$DD,$BE,$06,$5E
	.dc.b	$BB,$F0,$3E,$BF,$81,$EF,$87,$D3,$3E,$DF,$9D,$E0,$AF,$92,$F7,$C7
	.dc.b	$AF,$2F,$37,$BE,$08,$F9,$98,$F8,$38,$6F,$07,$C5,$AF,$F5,$27,$0F
	.dc.b	$F5,$7F,$8E,$0F,$A0,$FF,$00,$F3,$7E,$D6,$33,$1F,$DC,$1D,$DD,$7F
	.dc.b	$B7,$8E,$4B,$C5,$88,$E9,$EF,$FF,$72,$79,$BF,$E9,$A8,$FE,$C9,$18
	.dc.b	$F2,$BF,$8F,$E5,$E5,$8E,$31,$FE,$E6,$9F,$E8,$D1,$BF,$87,$FE,$25
	.dc.b	$1F,$D5,$63,$1F,$49,$FD,$39,$BE,$47,$07,$9D,$27,$FD,$E9,$87,$7D
	.dc.b	$3F,$D9,$B8,$AB,$5E,$2F,$C8,$DB,$E2,$63,$CB,$14,$BE,$37,$C5,$3F
	.dc.b	$2F,$2F,$80,$FF,$9C,$F4,$F2,$3E,$C7,$E0,$BE,$EE,$3B,$A7,$8B,$3F
	.dc.b	$EC,$17,$E5,$D9,$AF,$97,$4F,$F0,$0E,$BE,$4C,$38,$47,$93,$E4,$18
	.dc.b	$7E,$77,$4F,$F1,$21,$FD,$DF,$4F,$BB,$F0,$9E,$6C,$3F,$DD,$BC,$39
	.dc.b	$9F,$A3,$F2,$FE,$4D,$7C,$3B,$7C,$9D,$3E,$66,$1E,$0E,$1F,$E9,$9A
	.dc.b	$77,$DF,$FC,$CD,$DE,$3F,$A0,$6D,$0F,$2F,$7F,$F4,$39,$DF,$AF,$99
	.dc.b	$FF,$12,$4E,$7D,$AF,$F3,$F4,$60,$F8,$7F,$35,$B7,$AE,$3E,$1E,$FC
	.dc.b	$46,$FD,$2F,$F4,$6E,$1C,$F3,$F2,$5F,$73,$D1,$8F,$9E,$3F,$D7,$97
	.dc.b	$D2,$B8,$47,$CC,$F8,$8F,$55,$F0,$8E,$70,$EB,$E1,$7F,$63,$0E,$96
	.dc.b	$A7,$CB,$6B,$6F,$8F,$8F,$5F,$31,$BF,$BE,$73,$87,$19,$7C,$4E,$72
	.dc.b	$EB,$92,$5F,$0E,$F1,$3E,$7E,$BE,$3D,$3C,$F5,$AF,$0B,$3A,$7D,$8E
	.dc.b	$2F,$99,$AF,$B7,$0B,$C3,$F4,$E7,$E6,$CF,$C8,$9B,$3F,$1A,$3F,$DD
	.dc.b	$4D,$D3,$FD,$D2,$DC,$7E,$06,$BE,$85,$87,$AF,$3F,$2E,$B5,$F8,$8F
	.dc.b	$F4,$3C,$7F,$1D,$BF,$DE,$CF,$D3,$8F,$D9,$E1,$FB,$99,$FD,$8F,$76
	.dc.b	$D7,$C1,$CB,$F4,$BE,$7F,$DA,$AF,$26,$7F,$CA,$F3,$FD,$47,$A4,$F9
	.dc.b	$07,$CC,$4C,$B9,$CE,$63,$FC,$6F,$1F,$95,$8E,$1B,$FC,$48,$47,$F5
	.dc.b	$67,$E5,$98,$E4,$F5,$61,$FB,$7F,$37,$F2,$F3,$C3,$1F,$0F,$7F,$6A
	.dc.b	$F2,$C6,$38,$7C,$6E,$B2,$FF,$4A,$EB,$FE,$2F,$CF,$CD,$87,$CE,$F9
	.dc.b	$3F,$81,$97,$5D,$53,$F4,$BC,$6F,$F5,$F6,$1C,$72,$E9,$0F,$96,$FC
	.dc.b	$EE,$5D,$33,$FE,$A5,$F3,$9F,$D4,$F3,$FC,$7C,$E6,$D2,$D3,$E3,$63
	.dc.b	$D2,$31,$E5,$70,$8F,$7E,$1E,$CD,$7F,$C4,$8E,$5F,$99,$8F,$D5,$C6
	.dc.b	$D7,$CA,$D3,$E0,$5D,$3D,$DA,$FB,$72,$F4,$D7,$87,$3F,$EC,$B7,$EF
	.dc.b	$CB,$E1,$E9,$EC,$9F,$E9,$1B,$FE,$13,$3E,$9A,$7C,$CE,$7F,$36,$9F
	.dc.b	$9A,$C3,$E6,$CF,$C1,$D3,$F3,$BA,$FA,$6B,$C5,$EF,$FF,$1B,$F8,$DF
	.dc.b	$F4,$67,$CF,$BE,$73,$F0,$B1,$8E,$6B,$EC,$7F,$AD,$37,$FE,$79,$3F
	.dc.b	$D4,$B2,$FC,$73,$FE,$68,$7A,$1F,$05,$EB,$F3,$9F,$E3,$28,$FE,$2B
	.dc.b	$18,$F8,$9F,$D2,$B6,$BF,$0A,$E7,$E2,$AE,$0C,$D8,$B0,$77,$7B,$B2
	.dc.b	$FE,$1F,$8F,$58,$7C,$BF,$83,$1E,$58,$71,$9C,$E7,$F0,$E7,$38,$F4
	.dc.b	$B7,$C5,$67,$3F,$0C,$E7,$4F,$E3,$D8,$7C,$4C,$38,$E3,$E0,$9F,$44
	.dc.b	$E6,$FD,$3A,$7C,$0B,$F6,$78,$3F,$3A,$7E,$5C,$7F,$9A,$EB,$A5,$7C
	.dc.b	$EF,$83,$F9,$93,$DA,$3D,$95,$E4,$C7,$C1,$1F,$0A,$7C,$7A,$32,$FE
	.dc.b	$8D,$8F,$F7,$5F,$8C,$FD,$FE,$FF,$5B,$A3,$C9,$9F,$CE,$F9,$BB,$C3
	.dc.b	$F3,$F7,$FA,$73,$DF,$A7,$C3,$3F,$2F,$E3,$EF,$17,$C9,$BF,$F5,$86
	.dc.b	$8F,$77,$FB,$87,$8F,$FE,$7D,$EA,$FE,$B1,$C7,$FD,$72,$73,$EF,$FE
	.dc.b	$FE,$3F,$B3,$46,$34,$FE,$8A,$F2,$FA,$4F,$E8,$6F,$4C,$AD,$7F,$86
	.dc.b	$3C,$EF,$E3,$B3,$FD,$FE,$6C,$FF,$85,$0F,$01,$FE,$1E,$CB,$C9,$7E
	.dc.b	$08,$C6,$3F,$C2,$A3,$FC,$83,$1F,$E1,$EB,$F3,$E4,$9F,$D2,$4F,$C0
	.dc.b	$9C,$C7,$F8,$E3,$3F,$F0,$3C,$63,$D1,$FF,$79,$47,$C5,$AE,$F7,$CE
	.dc.b	$7F,$DC,$7C,$5C,$E7,$38,$7F,$90,$BC,$A7,$C6,$E7,$CE,$FE,$E5,$A7
	.dc.b	$9F,$1F,$F7,$5F,$C6,$FE,$03,$5F,$06,$08,$F4,$6B,$E1,$C3,$8E,$BF
	.dc.b	$3B,$4F,$82,$58,$FE,$FC,$FC,$E7,$FC,$9C,$FB,$0F,$2F,$6F,$CC,$8E
	.dc.b	$7A,$FF,$96,$3C,$9D,$46,$34,$FF,$7F,$DF,$A7,$83,$E8,$C3,$4F,$F5
	.dc.b	$8E,$BE,$8D,$75,$B7,$B6,$7D,$4F,$D3,$C3,$E4,$F8,$1F,$67,$2F,$F0
	.dc.b	$A3,$F8,$4F,$4B,$FB,$C4,$EE,$7D,$FF,$3F,$BD,$3F,$03,$EB,$7A,$5F
	.dc.b	$1E,$5F,$1F,$1E,$D8,$FF,$4C,$F1,$FF,$86,$E5,$F8,$78,$C7,$91,$F2
	.dc.b	$5F,$EA,$3E,$7F,$87,$93,$F2,$31,$FE,$4C,$AF,$B9,$7E,$4E,$5F,$C1
	.dc.b	$4F,$5F,$0D,$F0,$9D,$70,$FE,$F7,$0C,$7F,$96,$63,$E4,$9C,$C7,$F0
	.dc.b	$E8,$FC,$45,$F3,$87,$CC,$7F,$67,$9E,$19,$7F,$26,$9C,$FC,$17,$DB
	.dc.b	$BE,$17,$FE,$F7,$CF,$9E,$3D,$A3,$FD,$F5,$F1,$18,$C7,$94,$63,$1E
	.dc.b	$B9,$7C,$3B,$F2,$BF,$B0,$C7,$C9,$DB,$67,$A4,$F8,$5E,$D9,$BA,$7B
	.dc.b	$6B,$F0,$72,$F0,$E7,$EA,$8F,$F6,$26,$1C,$E1,$E0,$F8,$5F,$F3,$DE
	.dc.b	$DF,$3B,$2D,$D7,$D9,$F4,$BE,$AC,$7B,$78,$7C,$57,$EC,$F8,$4B,$97
	.dc.b	$87,$57,$C5,$79,$38,$7E,$4F,$17,$FB,$57,$5F,$DE,$71,$BD,$B7,$92
	.dc.b	$F9,$EF,$0F,$0F,$EF,$6B,$53,$98,$FE,$25,$18,$F1,$3F,$68,$F3,$F2
	.dc.b	$FF,$8F,$BD,$7E,$06,$E3,$FC,$7A,$3F,$C8,$B7,$FB,$5F,$8D,$FC,$77
	.dc.b	$86,$8F,$C6,$D7,$93,$3F,$F7,$AC,$7F,$24,$BF,$1C,$7F,$BE,$30,$C5
	.dc.b	$6F,$E0,$9F,$0B,$8D,$7D,$91,$8E,$CF,$E0,$98,$F2,$8D,$FC,$1E,$30
	.dc.b	$E5,$87,$FA,$53,$7F,$85,$18,$F2,$DC,$F8,$CF,$AD,$FF,$A6,$A3,$E7
	.dc.b	$6F,$EB,$CF,$B4,$E6,$BF,$82,$47,$CC,$F0,$3E,$4E,$3F,$4B,$83,$E5
	.dc.b	$9F,$1D,$F1,$F1,$F9,$6F,$BE,$77,$E6,$BE,$8F,$4F,$CD,$F3,$78,$57
	.dc.b	$C0,$F3,$3E,$17,$2B,$8D,$F2,$58,$7E,$3E,$5F,$E3,$4C,$FF,$0B,$A7
	.dc.b	$BF,$E3,$7D,$D8,$71,$E1,$F7,$74,$FF,$BB,$3A,$3F,$DB,$1C,$3F,$D6
	.dc.b	$B3,$7D,$14,$3F,$D7,$31,$FD,$72,$31,$1F,$CE,$DD,$70,$58,$E3,$36
	.dc.b	$3F,$B4,$CE,$23,$84,$77,$7C,$13,$D1,$EC,$F6,$8A,$E6,$3F,$AC,$4F
	.dc.b	$8A,$3C,$F3,$FD,$1E,$31,$1D,$F4,$FE,$C5,$18,$8F,$5C,$61,$C5,$7F
	.dc.b	$3F,$9D,$47,$FB,$62,$74,$F1,$7F,$81,$F9,$7F,$F0,$04,$FF,$78,$61
	.dc.b	$F2,$AF,$F4,$4E,$14,$7E,$95,$FF,$03,$4F,$27,$66,$E3,$93,$FE,$F4
	.dc.b	$9F,$AA,$7F,$B5,$8F,$D7,$47,$F8,$3E,$3F,$C4,$71,$83,$FD,$62,$7E
	.dc.b	$CF,$C7,$FE,$84,$F4,$8F,$E0,$B5,$E1,$7C,$95,$DD,$FF,$0D,$4F,$9E
	.dc.b	$75,$1A,$3B,$9E,$18,$B3,$ED,$8F,$88,$7E,$34,$F3,$1F,$C6,$63,$C4
	.dc.b	$F8,$27,$E5,$FC,$97,$F9,$C3,$CC,$7C,$CF,$C3,$8C,$72,$5F,$65,$5F
	.dc.b	$EF,$CC,$7E,$1D,$7F,$9E,$3B,$F1,$97,$D0,$8F,$F2,$B3,$8C,$D1,$FE
	.dc.b	$0E,$8F,$34,$61,$F9,$C7,$D8,$C7,$E5,$BC,$F1,$F0,$FC,$B7,$FA,$16
	.dc.b	$38,$DF,$A6,$3F,$BC,$C7,$C4,$8C,$72,$FF,$65,$D7,$DF,$7C,$A3,$FC
	.dc.b	$5F,$5B,$8C,$7A,$9F,$D5,$63,$AC,$63,$E1,$B5,$B3,$2F,$E8,$56,$FA
	.dc.b	$37,$95,$C2,$B5,$97,$F8,$F2,$5F,$0B,$9C,$3E,$66,$7F,$17,$5F,$3B
	.dc.b	$F8,$1E,$1F,$9B,$AF,$9F,$C0,$C4,$FC,$3C,$FA,$EB,$E1,$8F,$07,$A1
	.dc.b	$F5,$FE,$43,$A6,$1F,$E0,$C9,$F8,$9C,$9F,$3B,$91,$F2,$DF,$39,$FA
	.dc.b	$B5,$F3,$BA,$BF,$29,$18,$CF,$BF,$2F,$C2,$E0,$F8,$3C,$5F,$B7,$E0
	.dc.b	$F3,$D7,$92,$3F,$B3,$57,$F7,$AE,$87,$CA,$FF,$AE,$BC,$A7,$CB,$7E
	.dc.b	$DE,$73,$E2,$A5,$79,$BE,$1C,$36,$BF,$B7,$BF,$A4,$BE,$2F,$E4,$9E
	.dc.b	$FD,$3F,$8C,$5F,$2B,$F8,$C7,$FD,$BD,$3F,$A7,$7C,$38,$FF,$B1,$BC
	.dc.b	$BF,$DA,$46,$3E,$0D,$F1,$B8,$CF,$D5,$18,$D3,$EA,$D6,$E9,$F4,$38
	.dc.b	$BD,$C7,$F8,$3B,$0F,$44,$63,$D2,$FF,$40,$F6,$78,$E3,$1D,$8F,$8C
	.dc.b	$C7,$0F,$C5,$D7,$E2,$6F,$ED,$BF,$1E,$1E,$5A,$D6,$7F,$9B,$AE,$3B
	.dc.b	$FF,$B8,$A3,$FC,$3F,$1E,$29,$DE,$6B,$C5,$EE,$3F,$4F,$BF,$DB,$C7
	.dc.b	$CB,$F0,$9D,$3B,$3E,$A7,$9B,$F4,$67,$FC,$6F,$9D,$EB,$FA,$58,$7E
	.dc.b	$F3,$D6,$FE,$4B,$2E,$4F,$9D,$E5,$FA,$75,$F0,$71,$B9,$3F,$8E,$7A
	.dc.b	$BF,$4A,$9F,$18,$FA,$4F,$E8,$13,$98,$FE,$D3,$97,$EA,$23,$87,$9D
	.dc.b	$FC,$59,$E5,$DF,$79,$FF,$C6,$D5,$C9,$FF,$36,$3F,$CD,$67,$2F,$6C
	.dc.b	$BF,$8A,$FA,$7E,$A9,$F0,$65,$FE,$CC,$E0,$FF,$2C,$C6,$3E,$6B,$96
	.dc.b	$3F,$E4,$7D,$3D,$1E,$3F,$F0,$8C,$FF,$9E,$DF,$28,$7C,$9F,$A6,$F8
	.dc.b	$65,$FE,$69,$8F,$D2,$DF,$C4,$9C,$D7,$0E,$EE,$71,$8A,$FC,$9C,$EF
	.dc.b	$83,$E4,$E1,$CE,$DF,$3F,$9C,$FF,$97,$65,$4F,$89,$D9,$FC,$DB,$4F
	.dc.b	$D1,$C5,$8F,$E6,$F1,$F0,$F4,$F1,$F0,$7A,$73,$EF,$A7,$B0,$F2,$B8
	.dc.b	$D7,$F4,$FE,$77,$87,$6F,$27,$D0,$78,$FE,$6B,$EA,$5A,$BE,$C7,$B1
	.dc.b	$E3,$CF,$E2,$F1,$FC,$EA,$EF,$F2,$1F,$CA,$3D,$5F,$C7,$FA,$DF,$E1
	.dc.b	$9E,$2F,$EE,$0F,$F8,$7D,$BE,$B7,$F8,$C5,$FF,$05,$47,$F8,$B9,$B1
	.dc.b	$FE,$2B,$77,$DF,$FD,$62,$71,$1D,$23,$FA,$CC,$78,$27,$B4,$FF,$7A
	.dc.b	$8F,$F1,$3B,$FC,$F1,$FF,$14,$CD,$9C,$CF,$F7,$B9,$FF,$15,$4F,$19
	.dc.b	$FF,$14,$E3,$FE,$A6,$9F,$F1,$4C,$EE,$7F,$C5,$73,$FE,$AA,$BF,$F0
	.dc.b	$2D,$FC,$33,$FC,$49,$D5,$FE,$27,$8F,$F0,$74,$7F,$84,$DC,$B9,$3F
	.dc.b	$C4,$13,$CD,$FE,$62,$7F,$85,$23,$FC,$1F,$1D,$E7,$FC,$29,$3F,$E1
	.dc.b	$58,$FF,$49,$F9,$DF,$C0,$2B,$35,$DE,$BC,$A7,$F8,$66,$38,$BF,$EB
	.dc.b	$CA,$F0,$3E,$A6,$BF,$E2,$19,$F1,$3C,$2A,$BC,$A3,$A3,$D0,$FF,$22
	.dc.b	$CF,$C5,$EA,$FE,$AB,$C2,$F5,$FE,$24,$63,$9A,$F8,$5F,$A6,$C2,$F3
	.dc.b	$FF,$88,$34,$5F,$F5,$E9,$C3,$FC,$C5,$F9,$8F,$F8,$72,$3F,$A8,$E5
	.dc.b	$E1,$B6,$5F,$EA,$6E,$9D,$C6,$3D,$B7,$8D,$8A,$FF,$55,$FC,$53,$C7
	.dc.b	$FC,$F6,$7E,$AB,$4B,$E8,$33,$8F,$38,$7A,$BF,$E9,$9E,$2F,$F6,$56
	.dc.b	$DE,$AA,$D4,$BE,$4F,$EF,$65,$D7,$0E,$3D,$1E,$1E,$0F,$56,$DE,$4C
	.dc.b	$12,$F8,$CF,$B9,$87,$97,$8B,$D4,$7C,$CE,$DF,$8F,$A7,$96,$3C,$F8
	.dc.b	$F3,$7C,$31,$E4,$E5,$B3,$DF,$1B,$C3,$F0,$33,$F0,$FC,$57,$83,$0F
	.dc.b	$9B,$E9,$7D,$CF,$4B,$FC,$9F,$C5,$F6,$38,$2D,$7D,$57,$EB,$6F,$F2
	.dc.b	$F5,$FA,$19,$FC,$CE,$7F,$2B,$DF,$5F,$ED,$17,$FA,$FE,$BF,$F5,$47
	.dc.b	$07,$E9,$5F,$E1,$4F,$F1,$37,$F8,$73,$FA,$F7,$FD,$56,$FE,$CD,$FE
	.dc.b	$16,$F1,$EC,$F3,$C7,$E2,$5F,$C5,$3F,$C3,$23,$F5,$0F,$F1,$D7,$C0
	.dc.b	$FF,$10,$9F,$23,$96,$BF,$C8,$5A,$FF,$B0,$DA,$FF,$B3,$9A,$FF,$B5
	.dc.b	$5F,$DD,$C7,$EC,$4F,$F4,$C6,$9F,$EA,$AE,$CF,$E0,$33,$A9,$FF,$57
	.dc.b	$9A,$3F,$D6,$A7,$FA,$51,$D7,$3F,$F0,$07,$AC,$FE,$DD,$EF,$16,$BF
	.dc.b	$D6,$73,$FB,$BA,$FF,$23,$C7,$FA,$CD,$D1,$FE,$0D,$35,$5E,$37,$FD
	.dc.b	$7E,$D7,$CA,$D7,$CA,$74,$3F,$D6,$05,$9F,$E1,$41,$C1,$FC,$36,$AB
	.dc.b	$FC,$4E,$7F,$90,$FC,$B7,$F6,$EF,$8A,$7C,$0F,$F8,$5A,$BA,$4E,$7E
	.dc.b	$77,$E4,$E5,$FE,$D0,$8E,$18,$E2,$BF,$DB,$B1,$CD,$C6,$48,$FE,$F7
	.dc.b	$DB,$F2,$EF,$97,$95,$F1,$2B,$5E,$05,$BF,$8B,$5F,$ED,$ED,$EE,$3D
	.dc.b	$B0,$6B,$E8,$F8,$8F,$E7,$97,$D2,$31,$8F,$DD,$CB,$C5,$7C,$36,$61
	.dc.b	$FC,$9B,$1F,$91,$BF,$C4,$AD,$4B,$07,$D1,$EB,$3E,$D8,$F2,$97,$D0
	.dc.b	$7F,$91,$71,$E5,$A7,$CC,$CF,$CA,$7A,$63,$E1,$3F,$23,$BB,$E1,$75
	.dc.b	$FB,$B6,$F3,$DF,$AB,$7F,$05,$FC,$18,$F6,$E3,$8A,$DE,$1F,$43,$83
	.dc.b	$FC,$53,$F0,$BD,$A7,$E9,$7C,$A7,$CF,$F3,$3B,$57,$E0,$77,$7D,$1F
	.dc.b	$9A,$FC,$31,$F1,$B7,$F9,$55,$BE,$3F,$89,$A7,$FB,$3A,$7D,$F5,$F0
	.dc.b	$E1,$F3,$1F,$97,$F6,$1F,$31,$FC,$C7,$89,$F2,$5F,$EE,$9C,$5C,$DF
	.dc.b	$99,$BD,$F8,$EB,$FE,$B1,$CB,$CF,$1E,$4F,$E4,$6B,$AF,$C1,$7E,$B9
	.dc.b	$EF,$D7,$7E,$3B,$F1,$31,$FE,$6A,$AF,$A7,$EA,$FE,$DD,$E9,$97,$7D
	.dc.b	$5F,$33,$F6,$2F,$C7,$A7,$A3,$E4,$7F,$D2,$98,$7D,$49,$CD,$A7,$FB
	.dc.b	$2C,$E7,$5F,$81,$39,$F3,$7D,$93,$E0,$C3,$A4,$3E,$47,$86,$7E,$3C
	.dc.b	$7F,$B4,$C7,$FB,$4B,$87,$E3,$E0,$CF,$C3,$5A,$F8,$8E,$99,$FC,$5D
	.dc.b	$3E,$87,$6D,$CF,$E3,$47,$C4,$C7,$E2,$BE,$6C,$3E,$03,$74,$FE,$4D
	.dc.b	$97,$BB,$5C,$E1,$E7,$C3,$B7,$91,$F4,$7C,$0E,$73,$E9,$F2,$78,$E5
	.dc.b	$CF,$AB,$DB,$8F,$C0,$E1,$FE,$F1,$CF,$FC,$67,$E0,$7F,$7B,$EE,$B3
	.dc.b	$7C,$BF,$FA,$8B,$97,$E0,$D6,$A7,$E9,$EF,$FC,$EE,$38,$FC,$67,$EB
	.dc.b	$9D,$47,$F1,$77,$F8,$4F,$C1,$7F,$34,$7F,$85,$F9,$57,$C7,$FE,$1F
	.dc.b	$1F,$E1,$BA,$D1,$88,$67,$E5,$CD,$F0,$5F,$BE,$BF,$24,$63,$A3,$C9
	.dc.b	$7C,$34,$F5,$DB,$E7,$B3,$8F,$38,$7C,$F7,$FA,$CF,$7F,$7E,$1C,$72
	.dc.b	$EF,$2F,$87,$FE,$0B,$A7,$83,$0E,$3F,$1F,$FB,$5E,$2F,$99,$8F,$19
	.dc.b	$CF,$17,$F4,$BB,$F3,$F4,$FA,$6B,$B7,$0D,$DB,$E9,$78,$5F,$26,$7E
	.dc.b	$0C,$FC,$DD,$FE,$25,$7A,$76,$F5,$F0,$FB,$96,$DE,$6C,$3E,$4F,$37
	.dc.b	$D5,$F3,$3F,$9E,$FB,$DF,$B3,$CB,$DF,$97,$C3,$F3,$B9,$F2,$7C,$0C
	.dc.b	$39,$4F,$7C,$BF,$83,$5F,$0D,$3C,$11,$E0,$8F,$D3,$F1,$F5,$3F,$B3
	.dc.b	$64,$F4,$BF,$C3,$13,$FA,$C8,$FF,$0C,$6F,$FC,$C1,$FF,$0E,$3F,$DC
	.dc.b	$5F,$F1,$5B,$99,$FF,$13,$BF,$CF,$A3,$FC,$56,$D8,$FE,$D8,$FF,$83
	.dc.b	$E3,$FC,$56,$F5,$CB,$E0,$CF,$82,$BC,$4F,$94,$FF,$1B,$B8,$7F,$C5
	.dc.b	$17,$EA,$7F,$D6,$F5,$E5,$7F,$D6,$71,$97,$FD,$77,$3C,$5E,$87,$F8
	.dc.b	$A6,$3F,$CD,$F3,$FE,$29,$3F,$A7,$BF,$EA,$27,$FC,$20,$E5,$C9,$FE
	.dc.b	$27,$78,$BF,$EB,$29,$F9,$EF,$FA,$DA,$3A,$F4,$7F,$88,$9F,$F5,$A9
	.dc.b	$FE,$7E,$7F,$C4,$78,$7F,$86,$7C,$CF,$AD,$5F,$C7,$8F,$F6,$19,$E2
	.dc.b	$3F,$D6,$E7,$88,$FF,$59,$57,$89,$FF,$21,$5F,$89,$FF,$60,$CE,$DF
	.dc.b	$F5,$FC,$65,$CA,$AF,$17,$CE,$A9,$F2,$0F,$93,$7D,$0F,$FB,$2E,$F7
	.dc.b	$5A,$9F,$F1,$56,$1C,$F1,$E5,$8F,$3C,$5F,$11,$FE,$54,$C7,$8C,$34
	.dc.b	$C5,$F0,$CB,$C2,$E1,$61,$E5,$EE,$FF,$2B,$61,$E2,$AA,$78,$0F,$F2
	.dc.b	$C3,$F3,$7A,$3E,$7C,$75,$C7,$D9,$E9,$7F,$2F,$8F,$F1,$1D,$79,$35
	.dc.b	$FF,$02,$F9,$7F,$E4,$D9,$6D,$ED,$D3,$C6,$7C,$1C,$3F,$B1,$72,$72
	.dc.b	$AE,$D8,$7F,$83,$BC,$97,$C4,$67,$0E,$3A,$78,$A5,$F4,$3F,$E9,$AD
	.dc.b	$7C,$59,$FF,$6D,$87,$CC,$78,$A7,$39,$F9,$75,$F2,$70,$FC,$FD,$7E
	.dc.b	$43,$F6,$F8,$BE,$2E,$BE,$AB,$F8,$A7,$9B,$3F,$EF,$4F,$96,$B9,$47
	.dc.b	$F7,$DF,$27,$15,$EF,$D7,$CF,$7E,$33,$93,$EA,$E2,$C4,$FB,$B8,$FC
	.dc.b	$6B,$E5,$3C,$79,$7F,$A1,$70,$F3,$E7,$E4,$F8,$EF,$E4,$9B,$FC,$2E
	.dc.b	$AF,$73,$F1,$34,$F8,$FF,$39,$E1,$A7,$D2,$17,$C1,$7F,$4C,$F8,$EF
	.dc.b	$CC,$67,$9A,$D4,$FF,$50,$CB,$37,$C8,$E3,$E5,$FE,$51,$C4,$B8,$31
	.dc.b	$7C,$B0,$B5,$FE,$38,$7C,$91,$ED,$7A,$BF,$92,$F4,$7D,$B1,$E3,$9F
	.dc.b	$3F,$8E,$F7,$1C,$BF,$02,$B5,$97,$F6,$69,$FE,$E3,$CB,$F8,$FC,$3C
	.dc.b	$B8,$71,$F3,$1E,$C6,$74,$F7,$CE,$63,$F6,$1D,$5F,$B7,$BE,$38,$FF
	.dc.b	$7C,$9F,$DB,$CE,$6F,$8C,$FA,$AB,$CF,$DF,$73,$FE,$F3,$A7,$CA,$E1
	.dc.b	$FD,$7D,$74,$F2,$FD,$4F,$6D,$6E,$5F,$4B,$C9,$F9,$3F,$09,$FB,$AA
	.dc.b	$FA,$5A,$DA,$FF,$20,$65,$7A,$7F,$0D,$E3,$F6,$9F,$2E,$5F,$E5,$6B
	.dc.b	$F4,$61,$F3,$37,$FF,$0B,$FC,$37,$C3,$8C,$57,$C6,$E6,$FA,$F1,$8E
	.dc.b	$27,$C8,$7E,$77,$CB,$FF,$51,$FB,$4F,$A2,$7C,$D3,$E3,$BF,$1B,$5A
	.dc.b	$9C,$F4,$63,$E1,$18,$71,$FA,$2F,$E3,$F7,$BE,$CB,$5A,$C3,$F8,$DC
	.dc.b	$F8,$3A,$6C,$FF,$84,$B9,$F6,$7E,$EB,$D8,$FC,$5D,$78,$75,$F2,$72
	.dc.b	$FE,$26,$31,$B3,$E4,$F1,$8F,$7B,$F5,$47,$F3,$3C,$7E,$04,$E6,$B8
	.dc.b	$63,$FC,$77,$0E,$76,$8F,$D2,$63,$FD,$06,$73,$0F,$8D,$E7,$BF,$C9
	.dc.b	$CF,$FC,$B9,$E3,$7F,$BE,$6F,$D1,$A7,$C7,$AF,$9D,$8F,$C4,$7F,$8B
	.dc.b	$CF,$F8,$32,$7D,$87,$F3,$C7,$DB,$DB,$EC,$9E,$DE,$6F,$3F,$92,$F9
	.dc.b	$31,$FE,$A3,$BF,$6F,$77,$4F,$A5,$E5,$C3,$EF,$E0,$7B,$5E,$F8,$C7
	.dc.b	$0F,$FB,$97,$6C,$E3,$FD,$9F,$E9,$7F,$82,$7B,$EF,$A4,$7C,$9F,$CB
	.dc.b	$AF,$E9,$39,$7F,$7F,$BD,$D6,$A7,$F9,$D4,$E7,$BF,$FD,$EF,$3F,$E7
	.dc.b	$E9,$E2,$77,$8F,$F7,$DF,$8D,$FD,$56,$3F,$DF,$75,$C7,$D2,$C4,$60
	.dc.b	$E1,$38,$79,$FA,$4C,$5F,$39,$99,$CC,$7F,$1F,$78,$6B,$F0,$2B,$84
	.dc.b	$6E,$31,$97,$18,$EA,$58,$E1,$5F,$03,$4F,$F0,$34,$F6,$AF,$7C,$7D
	.dc.b	$0F,$03,$D1,$87,$96,$73,$A3,$E0,$33,$8F,$38,$C4,$F3,$DF,$E9,$63
	.dc.b	$F6,$AF,$F2,$59,$F8,$AB,$E6,$F8,$1F,$26,$31,$7F,$67,$4C,$63,$F4
	.dc.b	$E3,$FC,$0D,$1F,$58,$FA,$4F,$C4,$8F,$F1,$D7,$2F,$BD,$E1,$C3,$F8
	.dc.b	$27,$86,$BE,$BA,$FC,$37,$AD,$DB,$0F,$C5,$75,$7C,$BB,$47,$F4,$93
	.dc.b	$ED,$E5,$E6,$C7,$A7,$17,$C2,$D7,$E2,$FB,$B7,$3F,$83,$8F,$C7,$CB
	.dc.b	$F1,$3E,$33,$E6,$1F,$11,$FF,$00,$3E,$57,$FC,$65,$97,$B6,$B5,$39
	.dc.b	$E1,$F3,$46,$3D,$D5,$F9,$32,$F3,$CE,$79,$F1,$C3,$FD,$CF,$3E,$FF
	.dc.b	$1D,$F1,$E3,$19,$FE,$EF,$8B,$F5,$78,$79,$7E,$8F,$E1,$57,$19,$B3
	.dc.b	$E9,$8B,$1F,$3D,$E7,$A7,$93,$5F,$F2,$4C,$7C,$0C,$BD,$F3,$9F,$92
	.dc.b	$FE,$8D,$AF,$A2,$73,$B7,$F9,$4E,$5F,$2F,$FA,$9C,$FE,$04,$E6,$1D
	.dc.b	$1D,$76,$F8,$DA,$7C,$9C,$7B,$FA,$FF,$0B,$1F,$91,$AD,$9F,$A3,$C3
	.dc.b	$F0,$E7,$E7,$CD,$E3,$F6,$47,$83,$06,$BD,$72,$FB,$1E,$FF,$D4,$CB
	.dc.b	$D3,$C9,$75,$F8,$FE,$17,$C5,$F6,$7F,$D9,$DC,$5E,$FC,$7E,$E6,$5F
	.dc.b	$E0,$ED,$BD,$5C,$7F,$17,$E6,$FF,$84,$78,$DF,$E5,$4F,$89,$7C,$83
	.dc.b	$E9,$3F,$C4,$5C,$9F,$A3,$AD,$4E,$78,$3F,$39,$18,$AF,$C1,$BA,$F4
	.dc.b	$6F,$99,$FE,$FB,$F8,$DF,$C2,$4F,$7E,$9B,$3F,$E5,$4D,$B7,$D3,$F8
	.dc.b	$4A,$FF,$35,$C5,$8F,$C1,$79,$DE,$E8,$FE,$DF,$4D,$BD,$39,$FF,$8E
	.dc.b	$7E,$63,$D1,$39,$DB,$D1,$1F,$C7,$FB,$BF,$9E,$E7,$E9,$DD,$3F,$D6
	.dc.b	$F2,$FE,$67,$3F,$03,$A3,$D7,$9F,$C6,$9A,$7A,$8F,$46,$DF,$97,$AF
	.dc.b	$31,$D7,$1F,$1D,$79,$BC,$AF,$93,$A7,$97,$57,$16,$2B,$E9,$FC,$5F
	.dc.b	$83,$87,$E3,$F8,$BF,$17,$0E,$1C,$1F,$E6,$2D,$BE,$DC,$63,$83,$9C
	.dc.b	$7F,$4C,$F8,$BF,$F2,$3F,$CD,$BE,$73,$FD,$83,$C0,$E7,$FE,$41,$5A
	.dc.b	$9F,$F6,$2F,$5F,$F1,$98,$C7,$C5,$7F,$A3,$71,$FE,$A1,$18,$75,$EE
	.dc.b	$32,$ED,$AF,$F0,$6A,$E9,$56,$B5,$9F,$F0,$79,$F2,$CD,$9F,$EF,$4F
	.dc.b	$A7,$DE,$FE,$05,$3F,$B4,$8F,$EA,$9E,$9F,$FB,$B5,$CD,$78,$74,$F6
	.dc.b	$F8,$FE,$C8,$C6,$9F,$EB,$E8,$E5,$97,$F7,$EC,$7E,$07,$8D,$E5,$DF
	.dc.b	$F8,$AC,$7F,$9B,$B2,$A7,$6D,$BE,$56,$3F,$EC,$8E,$AE,$D1,$8E,$BF
	.dc.b	$86,$FC,$FF,$8A,$F8,$0B,$E7,$7D,$CF,$AF,$0F,$93,$F4,$7E,$0C,$FB
	.dc.b	$70,$7D,$6F,$0B,$FD,$85,$C5,$FD,$23,$D2,$74,$BE,$59,$F0,$0B,$7D
	.dc.b	$27,$FB,$D3,$4F,$E9,$FC,$BF,$E3,$B5,$8F,$2C,$38,$DF,$0E,$5E,$F5
	.dc.b	$AC,$BB,$65,$D7,$2F,$ED,$99,$F8,$E7,$38,$7E,$42,$FB,$CE,$71,$EB
	.dc.b	$1C,$0E,$8F,$A3,$5F,$EC,$33,$BF,$8E,$FF,$09,$BC,$78,$ED,$F0,$2E
	.dc.b	$BF,$0B,$93,$F0,$71,$EC,$F9,$4F,$F0,$A5,$F9,$63,$3E,$87,$F8,$65
	.dc.b	$FF,$7F,$C7,$97,$7F,$67,$C8,$FE,$12,$31,$B7,$F8,$96,$BD,$67,$A2
	.dc.b	$BC,$3B,$FD,$1E,$2B,$C1,$F2,$B8,$7F,$D1,$B7,$C3,$6F,$9D,$87,$C9
	.dc.b	$ED,$F9,$B1,$8C,$9F,$2F,$F1,$63,$7F,$0B,$F1,$30,$F0,$70,$D9,$FC
	.dc.b	$0F,$CE,$FC,$07,$F0,$7E,$27,$C2,$F4,$7F,$C5,$5F,$2A,$F8,$27,$CE
	.dc.b	$3D,$EF,$E3,$D9,$FC,$21,$7C,$2B,$73,$F0,$78,$FF,$11,$39,$BE,$57
	.dc.b	$F8,$68,$C7,$5F,$DD,$79,$7C,$E7,$A5,$E9,$E6,$DF,$80,$69,$E5,$C3
	.dc.b	$1F,$0F,$E9,$9F,$D2,$47,$37,$ED,$BE,$6D,$AF,$A5,$F6,$F5,$F2,$D7
	.dc.b	$4D,$FD,$FD,$FC,$72,$F8,$5B,$F8,$5F,$89,$CF,$7A,$FF,$1B,$BF,$C8
	.dc.b	$D6,$E5,$FC,$7D,$F6,$1E,$28,$F8,$5C,$5F,$9F,$D3,$C3,$8F,$6F,$94
	.dc.b	$F0,$7C,$77,$4F,$82,$E1,$9F,$AB,$DB,$FF,$08,$FC,$33,$E5,$FF,$DF
	.dc.b	$7E,$1F,$F0,$3C,$FF,$6D,$5A,$9C,$ED,$FB,$48,$C7,$A1,$FE,$91,$8F
	.dc.b	$B5,$18,$7F,$5C,$F1,$E3,$FC,$53,$E0,$F9,$BF,$D5,$BE,$2C,$EF,$A3
	.dc.b	$E3,$BB,$1F,$AC,$9F,$E1,$91,$7B,$3E,$9F,$C5,$FD,$1F,$1F,$FC,$B7
	.dc.b	$5E,$7A,$EF,$83,$7E,$4F,$C1,$8F,$84,$E7,$56,$BF,$A8,$9F,$D7,$DF
	.dc.b	$BB,$9F,$D7,$7B,$F4,$FA,$E3,$F2,$74,$F6,$F8,$5F,$E9,$3F,$73,$F9
	.dc.b	$37,$CD,$7F,$86,$FE,$77,$F9,$3E,$3F,$36,$73,$EE,$27,$3A,$FF,$9D
	.dc.b	$07,$2A,$B7,$3F,$CC,$32,$FF,$46,$46,$3C,$EF,$F3,$CB,$AE,$4B,$87
	.dc.b	$F2,$EC,$FF,$C5,$9F,$0B,$FE,$AE,$AF,$0F,$87,$6B,$5E,$E7,$FA,$BE
	.dc.b	$7D,$39,$59,$FE,$6F,$C9,$ED,$7F,$9C,$F4,$DF,$43,$F4,$57,$E6,$9F
	.dc.b	$0F,$92,$D3,$FC,$79,$C7,$DE,$31,$C5,$D5,$D6,$36,$BB,$47,$4B,$F5
	.dc.b	$57,$C7,$DF,$F0,$71,$FD,$33,$1F,$81,$8F,$C4,$8F,$E9,$BA,$79,$FC
	.dc.b	$2F,$95,$EB,$BC,$1F,$F5,$F7,$C7,$6E,$73,$1F,$EB,$EB,$FA,$5C,$DF
	.dc.b	$45,$FE,$2B,$93,$0E,$5E,$2F,$CE,$C3,$8F,$25,$AD,$6D,$F3,$67,$C3
	.dc.b	$A5,$97,$D2,$FC,$D9,$FB,$DD,$96,$7D,$1E,$AF,$A7,$4F,$54,$78,$76
	.dc.b	$E5,$AF,$F1,$69,$F8,$47,$99,$E5,$5C,$DF,$3F,$99,$EF,$BE,$5E,$2F
	.dc.b	$9E,$7E,$27,$1F,$AF,$C2,$FB,$DE,$5F,$D1,$58,$BE,$8B,$FD,$0F,$E0
	.dc.b	$7F,$26,$E5,$F8,$D5,$A9,$CE,$BF,$4E,$31,$F0,$BF,$E3,$F8,$C6,$1C
	.dc.b	$FD,$CF,$C9,$57,$1E,$1B,$3D,$EA,$E3,$F9,$3A,$FE,$21,$1D,$4F,$CA
	.dc.b	$3D,$9F,$D0,$BE,$39,$E7,$C1,$C5,$F5,$63,$F8,$AA,$F9,$35,$E6,$E9
	.dc.b	$E3,$4F,$A5,$F0,$7C,$4F,$6C,$FF,$45,$E3,$FF,$41,$E0,$FF,$12,$7D
	.dc.b	$22,$3F,$AD,$F9,$9F,$16,$73,$18,$D3,$F7,$C7,$EF,$A3,$3D,$1E,$CE
	.dc.b	$04,$62,$7F,$DE,$AF,$27,$89,$FE,$2C,$7C,$3A,$62,$30,$7F,$8B,$B1
	.dc.b	$C3,$8E,$FF,$F0,$63,$D0,$FF,$19,$9A,$3C,$87,$1C,$7F,$BA,$FC,$A6
	.dc.b	$E5,$5E,$FC,$B9,$DF,$09,$CD,$FA,$A3,$85,$7C,$5C,$3C,$D9,$66,$F9
	.dc.b	$F9,$9E,$68,$C7,$CD,$78,$DD,$72,$5D,$59,$3C,$5F,$A4,$DF,$81,$7D
	.dc.b	$0B,$9B,$FC,$1F,$3F,$12,$BF,$CC,$53,$9E,$3F,$85,$18,$AF,$2E,$1D
	.dc.b	$A3,$7B,$63,$0E,$D8,$79,$EB,$58,$FE,$52,$7B,$3E,$77,$CD,$9D,$C3
	.dc.b	$B6,$7E,$0A,$F3,$F0,$75,$9F,$F3,$B7,$3F,$5C,$3D,$9A,$D9,$7D,$07
	.dc.b	$3F,$23,$B7,$27,$C2,$D5,$F3,$DF,$E7,$2E,$7F,$E0,$DB,$FD,$39,$CF
	.dc.b	$BF,$FE,$75,$D3,$F2,$73,$9B,$FF,$39,$47,$E1,$6B,$6F,$C1,$7A,$F0
	.dc.b	$79,$5E,$F1,$7B,$FE,$07,$85,$E3,$BF,$C2,$CF,$E5,$29,$E3,$FC,$57
	.dc.b	$59,$7F,$71,$9F,$E6,$31,$F7,$8F,$3F,$07,$86,$FF,$CA,$F3,$E3,$F9
	.dc.b	$3F,$9D,$F0,$CF,$89,$FF,$27,$79,$9F,$5E,$31,$EC,$F8,$B8,$3F,$29
	.dc.b	$18,$C3,$57,$FC,$55,$FD,$24,$3E,$83,$AE,$DF,$93,$E3,$FF,$60,$57
	.dc.b	$D5,$AF,$07,$25,$C3,$C7,$E2,$79,$2B,$CB,$1F,$E6,$1E,$4E,$FE,$57
	.dc.b	$E7,$F1,$FF,$56,$7B,$0F,$8A,$FC,$BE,$1F,$EA,$89,$CC,$7A,$5B,$1F
	.dc.b	$AA,$78,$DF,$38,$FF,$56,$D7,$9D,$FD,$31,$7C,$3F,$F5,$CB,$B5,$1E
	.dc.b	$19,$E9,$3C,$BE,$7B,$53,$FE,$53,$8F,$EC,$B7,$FE,$C3,$F9,$8E,$CF
	.dc.b	$F8,$7B,$5E,$2F,$F8,$72,$7F,$8B,$F0,$FF,$9F,$9F,$F3,$FF,$57,$CF
	.dc.b	$7F,$23,$3F,$BC,$1E,$BF,$49,$F5,$74,$FF,$9B,$E7,$55,$9F,$85,$F8
	.dc.b	$11,$FE,$B3,$8F,$4F,$83,$FD,$F9,$BE,$A7,$E4,$9F,$F1,$6C,$FF,$92
	.dc.b	$23,$E3,$4F,$E5,$27,$E5,$79,$1D,$DF,$E7,$86,$A7,$FD,$9C,$FA,$0F
	.dc.b	$E8,$1E,$8F,$2C,$BF,$8C,$78,$7F,$EE,$7E,$BF,$B7,$5F,$EE,$28,$C7
	.dc.b	$7F,$F1,$F7,$51,$4F,$F0,$C9,$5F,$F4,$71,$FE,$F1,$E0,$EB,$1F,$CE
	.dc.b	$0E,$1C,$BF,$8D,$7F,$96,$CF,$FB,$DF,$9F,$7B,$3C,$2F,$F0,$45,$78
	.dc.b	$CF,$EA,$87,$03,$FD,$67,$E3,$F0,$78,$F9,$7C,$B3,$F0,$F1,$F9,$27
	.dc.b	$F9,$77,$81,$E4,$F2,$7E,$74,$7F,$96,$B4,$FE,$6F,$1B,$3E,$4C,$7F
	.dc.b	$10,$8F,$F3,$CB,$EC,$E3,$7C,$F4,$21,$7A,$3F,$37,$18,$8C,$F6,$77
	.dc.b	$A4,$E6,$79,$46,$63,$10,$9F,$85,$18,$9B,$3D,$67,$FC,$87,$7F,$D4
	.dc.b	$2B,$A5,$60,$EF,$5E,$27,$11,$E5,$7D,$C6,$A3,$DB,$5F,$A4,$AE,$91
	.dc.b	$E6,$9D,$D7,$C1,$AF,$A8,$E6,$33,$3F,$29,$F3,$9F,$89,$8C,$3E,$C3
	.dc.b	$E6,$47,$C0,$FA,$4F,$84,$FC,$2C,$2B,$F2,$E3,$E0,$BF,$36,$2B,$F4
	.dc.b	$E3,$DA,$E6,$72,$A3,$CA,$FD,$78,$DB,$F7,$23,$E0,$47,$C3,$8F,$E1
	.dc.b	$93,$F7,$1D,$46,$AB,$93,$F0,$2B,$DD,$3F,$C4,$E3,$EC,$C7,$E2,$6B
	.dc.b	$F4,$BE,$0F,$E5,$47,$CF,$C2,$16,$3F,$C4,$3C,$9B,$8C,$9B,$E7,$FC
	.dc.b	$65,$62,$39,$61,$87,$13,$FA,$B9,$D4,$7E,$15,$CC,$E6,$3A,$B9,$7C
	.dc.b	$73,$F8,$69,$CC,$E6,$7D,$33,$F2,$27,$D3,$36,$72,$D7,$F8,$54,$62
	.dc.b	$34,$62,$7F,$C5,$95,$FD,$6A,$3F,$C9,$6F,$FB,$40,$C3,$FE,$56,$E2
	.dc.b	$C3,$FC,$3C,$FF,$2C,$C7,$F8,$CB,$0F,$93,$1F,$E5,$19,$C4,$F2,$8F
	.dc.b	$E2,$D3,$A7,$A1,$A3,$A1,$D5,$D3,$FE,$3E,$9F,$F0,$AC,$F9,$CF,$E9
	.dc.b	$67,$A5,$B1,$FC,$8E,$3F,$C9,$A7,$F7,$68,$B1,$D0,$FF,$5E,$C7,$F9
	.dc.b	$28,$EC,$E9,$F1,$BA,$0F,$F8,$F5,$F4,$6B,$FE,$67,$7E,$2C,$7F,$3E
	.dc.b	$7D,$8F,$F2,$2A,$FC,$6E,$BF,$42,$79,$FD,$20,$84,$1F,$1F,$D1,$C1
	.dc.b	$AF,$13,$0E,$23,$BB,$A1,$18,$7A,$47,$AF,$0D,$4D,$9C,$D5,$7B,$35
	.dc.b	$F1,$BF,$48,$F4,$BE,$98,$FA,$33,$EA,$AF,$95,$39,$AC,$BC,$9B,$5F
	.dc.b	$E7,$FA,$FA,$55,$F1,$5F,$4C,$E4,$EF,$5E,$DC,$3F,$6E,$FB,$C7,$B1
	.dc.b	$F1,$1F,$8F,$35,$F9,$B1,$F0,$DF,$9D,$15,$FA,$51,$EF,$7E,$A4,$58
	.dc.b	$F8,$15,$D2,$3E,$24,$7B,$27,$71,$A7,$9B,$93,$E1,$C7,$AE,$77,$1F
	.dc.b	$22,$3F,$09,$97,$F7,$48,$FC,$6F,$C6,$FF,$0A,$DF,$F6,$32,$8F,$E1
	.dc.b	$BB,$61,$FE,$1F,$56,$BF,$8C,$CF,$27,$F8,$5E,$5F,$EB,$EB,$E5,$1F
	.dc.b	$C4,$2B,$4F,$F1,$C6,$BF,$E4,$66,$9F,$81,$8E,$31,$98,$E2,$7E,$CA
	.dc.b	$3A,$D7,$EC,$1E,$F1,$6F,$83,$FA,$D8,$F6,$3F,$7D,$FE,$78,$FF,$05
	.dc.b	$E9,$FD,$76,$1E,$9E,$BF,$F6,$DE,$7F,$11,$FF,$6E,$F0,$5B,$FE,$36
	.dc.b	$53,$F8,$7D,$71,$7F,$8C,$9F,$BA,$9F,$F7,$5C,$F8,$DC,$C6,$4F,$F6
	.dc.b	$78,$8F,$85,$18,$AF,$80,$EA,$2C,$FF,$0A,$AF,$04,$7C,$39,$FE,$1B
	.dc.b	$3E,$98,$F8,$1F,$49,$F8,$2D,$7F,$83,$63,$FE,$16,$E4,$FF,$11,$E3
	.dc.b	$FE,$F6,$8C,$71,$7F,$34,$75,$8D,$AF,$F1,$04,$FC,$97,$FC,$2F,$19
	.dc.b	$7F,$C2,$CD,$7F,$C1,$95,$D5,$CC,$65,$FE,$5B,$58,$8E,$06,$23,$FC
	.dc.b	$36,$D7,$F7,$11,$88,$D1,$8A,$F6,$CF,$F8,$4E,$BC,$0F,$F8,$DD,$F0
	.dc.b	$3F,$E0,$27,$2F,$F9,$09,$F6,$3F,$E4,$4E,$4F,$E6,$D7,$C6,$F3,$3F
	.dc.b	$E0,$A6,$AA,$FE,$35,$3C,$66,$CF,$F8,$39,$EB,$5F,$D2,$27,$11,$F2
	.dc.b	$E3,$DB,$3F,$96,$AF,$A9,$1F,$61,$FE,$7D,$D1,$F9,$8D,$7F,$DB,$5E
	.dc.b	$B7,$F1,$7E,$4F,$ED,$72,$8F,$E3,$5C,$B8,$7F,$9D,$4F,$F6,$59,$D7
	.dc.b	$2F,$F4,$2A,$C5,$7F,$41,$CE,$CF,$F4,$99,$FE,$B1,$5C,$23,$FA,$75
	.dc.b	$FF,$39,$8C,$47,$F2,$EA,$E2,$FF,$35,$6D,$7F,$6C,$9F,$0D,$7F,$6A
	.dc.b	$7C,$71,$6F,$85,$78,$1E,$8F,$7F,$67,$E1,$F5,$FC,$49,$FE,$71,$8F
	.dc.b	$BC,$FE,$B1,$C1,$5F,$E7,$3A,$7F,$70,$C1,$97,$A2,$7F,$96,$CE,$63
	.dc.b	$11,$E1,$3F,$B5,$55,$7F,$B6,$D5,$7F,$DA,$15,$C6,$FF,$A1,$CE,$EB
	.dc.b	$E0,$47,$08,$F9,$0F,$F3,$C6,$D7,$F5,$5B,$FB,$55,$6B,$F0,$3C,$5F
	.dc.b	$DF,$FC,$FF,$C1,$61,$EA,$E8,$FE,$79,$97,$A5,$FE,$76,$D7,$FC,$9D
	.dc.b	$1F,$86,$7F,$C9,$31,$97,$FC,$83,$19,$7F,$9D,$35,$FE,$74,$D7,$F9
	.dc.b	$B4,$71,$7E,$E1,$FE,$5D,$8F,$C2,$3C,$1F,$C8,$3F,$E4,$97,$F1,$6F
	.dc.b	$F9,$A9,$FC,$13,$FC,$E9,$D9,$F0,$FA,$3E,$03,$FC,$DA,$7E,$D5,$7F
	.dc.b	$2A,$8F,$E5,$F1,$87,$07,$F8,$CC,$FF,$27,$C7,$F3,$C6,$9F,$CF,$0A
	.dc.b	$7F,$92,$8A,$7F,$96,$E3,$A3,$FC,$CC,$FF,$1D,$47,$F3,$C7,$FC,$A0
	.dc.b	$F1,$69,$FE,$57,$8E,$37,$FE,$48,$9F,$F2,$D4,$F8,$A3,$FC,$29,$1F
	.dc.b	$CE,$A3,$CF,$1F,$E3,$DE,$BF,$BF,$2F,$92,$FE,$75,$7F,$09,$FF,$68
	.dc.b	$F3,$7B,$9F,$F6,$84,$67,$97,$FD,$A1,$8E,$5F,$F6,$53,$5F,$F6,$23
	.dc.b	$EA,$9C,$B9,$8F,$F6,$E3,$F0,$23,$FD,$D2,$FF,$8E,$6F,$C9,$3F,$CD
	.dc.b	$63,$93,$FE,$F0,$7F,$CD,$71,$AB,$F8,$55,$99,$CB,$FE,$CE,$BF,$64
	.dc.b	$7C,$4F,$33,$F4,$13,$FC,$F7,$95,$2C,$3C,$38,$FF,$B3,$70,$F8,$0F
	.dc.b	$F3,$36,$BF,$EC,$96,$BF,$EC,$D9,$F3,$BF,$EC,$48,$DF,$37,$B7,$0F
	.dc.b	$E8,$D5,$F0,$6B,$FA,$13,$C2,$BD,$31,$CE,$38,$3F,$EC,$B8,$FA,$D5
	.dc.b	$F7,$AB,$F9,$F6,$1E,$AC,$1F,$47,$FB,$DD,$BF,$01,$C7,$FC,$F7,$1F
	.dc.b	$E3,$1B,$7F,$1E,$E6,$FF,$54,$BF,$D0,$75,$B5,$FE,$72,$6B,$FE,$72
	.dc.b	$6B,$FE,$72,$6B,$FC,$D6,$BA,$BF,$CF,$E3,$FD,$2D,$3C,$5F,$F5,$63
	.dc.b	$97,$31,$FD,$09,$FF,$53,$C7,$57,$3C,$FF,$E6,$F8,$FF,$3C,$46,$A3
	.dc.b	$F3,$97,$FE,$99,$F1,$BF,$8D,$D7,$F3,$AB,$F2,$9F,$E6,$EC,$3C,$AF
	.dc.b	$FA,$4F,$0C,$BF,$E7,$66,$9E,$67,$FD,$3B,$CB,$E6,$7F,$99,$C7,$C0
	.dc.b	$8F,$F4,$85,$7F,$9C,$63,$84,$FF,$33,$6B,$FE,$BE,$8F,$F3,$FC,$7B
	.dc.b	$A3,$FC,$DE,$FC,$29,$F7,$47,$ED,$DF,$67,$5F,$B7,$B7,$F9,$AE,$FF
	.dc.b	$98,$CB,$FC,$BF,$CD,$FE,$F6,$CF,$FA,$A4,$62,$07,$F2,$58,$CD,$7F
	.dc.b	$38,$AE,$4F,$F3,$CC,$BF,$DD,$77,$C7,$6F,$F6,$73,$CE,$2C,$FF,$3C
	.dc.b	$76,$FF,$40,$7F,$DD,$11,$FE,$72,$9F,$F7,$0B,$FE,$74,$8B,$3F,$CE
	.dc.b	$5D,$BF,$CE,$5C,$CF,$FA,$B6,$7F,$CE,$31,$FE,$E6,$7F,$CE,$2F,$FC
	.dc.b	$00,$EF,$FE,$B3,$1F,$E9,$5B,$F4,$E5,$FE,$70,$C7,$E1,$BF,$CE,$AF
	.dc.b	$FB,$C4,$61,$FF,$36,$9C,$0F,$F3,$6B,$97,$F9,$B1,$93,$F9,$D4,$FD
	.dc.b	$19,$FE,$6D,$3B,$8F,$7C,$70,$9F,$86,$FF,$39,$8E,$11,$FE,$77,$36
	.dc.b	$FB,$E3,$FD,$F5,$3F,$6B,$B7,$F3,$9A,$7F,$BD,$71,$FF,$38,$72,$FC
	.dc.b	$13,$F9,$BD,$7F,$84,$E3,$FC,$DE,$D8,$FF,$37,$E3,$B7,$F9,$BC,$65
	.dc.b	$FF,$37,$B5,$FF,$37,$E3,$FC,$92,$BF,$9F,$D7,$F2,$6A,$D5,$7F,$64
	.dc.b	$78,$73,$DB,$E1,$23,$F9,$A5,$7F,$A5,$AB,$C1,$5F,$4D,$F1,$45,$AE
	.dc.b	$33,$FE,$ED,$BE,$33,$FE,$FD,$7F,$A4,$BB,$0F,$FA,$4F,$9B,$DD,$87
	.dc.b	$F3,$9B,$F1,$BE,$32,$47,$AE,$33,$1F,$CD,$30,$F6,$BF,$CC,$5A,$7A
	.dc.b	$63,$71,$FC,$D6,$FB,$BF,$E7,$97,$47,$91,$EE,$FA,$5F,$F3,$BC,$FD
	.dc.b	$1A,$F5,$5F,$F9,$EA,$BE,$E4,$7C,$1A,$FE,$7B,$1E,$E8,$FC,$15,$7E
	.dc.b	$6E,$3D,$43,$BF,$FC,$77,$7F,$EF,$5E,$0F,$D3,$1F,$E3,$AC,$7E,$E3
	.dc.b	$FE,$F4,$BF,$C4,$3F,$C0,$EF,$E5,$BF,$C0,$E3,$2F,$F8,$E9,$AF,$F8
	.dc.b	$E9,$AF,$F8,$EA,$7A,$BF,$E3,$78,$DE,$FF,$85,$AE,$B3,$FB,$F7,$E3
	.dc.b	$DF,$F8,$FD,$FB,$D1,$97,$2F,$F8,$ED,$F5,$9F,$85,$9F,$84,$FF,$AA
	.dc.b	$6F,$FB,$DE,$18,$71,$1F,$14,$FE,$08,$E9,$FE,$08,$FC,$A8,$FF,$1E
	.dc.b	$61,$FE,$7D,$73,$39,$3F,$83,$9D,$1F,$F2,$03,$4F,$E0,$73,$D5,$FF
	.dc.b	$24,$1F,$84,$8F,$F1,$E3,$B3,$FD,$E5,$1B,$E9,$FC,$E4,$FF,$8F,$2B
	.dc.b	$C5,$5F,$C1,$2B,$DA,$7C,$08,$F3,$57,$C2,$9F,$F8,$00,$66,$7E,$1F
	.dc.b	$57,$F8,$05,$BC,$63,$FD,$71,$7E,$F9,$FF,$5E,$D7,$B5,$FF,$01,$46
	.dc.b	$67,$FC,$04,$D7,$FC,$04,$D7,$FC,$05,$5D,$9F,$EE,$F5,$CD,$FE,$FC
	.dc.b	$FC,$3A,$FE,$FC,$E7,$9B,$E0,$C7,$08,$FA,$8F,$FA,$E9,$BC,$DF,$A1
	.dc.b	$C3,$EB,$76,$FB,$E7,$F4,$5D,$B7,$4F,$F5,$8E,$5E,$FB,$F3,$3F,$E0
	.dc.b	$3C,$FC,$CF,$C3,$3E,$25,$F9,$DF,$F5,$BB,$5F,$F5,$84,$F2,$E3,$F5
	.dc.b	$47,$F8,$12,$3E,$14,$7F,$A2,$DF,$89,$36,$BE,$24,$FE,$8E,$FF,$DC
	.dc.b	$53,$73,$FF,$04,$3F,$03,$0F,$75,$FD,$AB,$B3,$FB,$FA,$F7,$6F,$FD
	.dc.b	$F8,$87,$F8,$03,$97,$F2,$CF,$EF,$EF,$F4,$AE,$8E,$1C,$FA,$70,$33
	.dc.b	$C3,$D3,$87,$F4,$CF,$F0,$5A,$E6,$FF,$05,$6B,$FC,$15,$AF,$F0,$5C
	.dc.b	$7F,$8B,$3F,$E0,$77,$AC,$FC,$38,$FF,$5F,$47,$8E,$3F,$47,$1F,$C0
	.dc.b	$E3,$C3,$1F,$0B,$4C,$47,$EE,$63,$E1,$CF,$AB,$9B,$D2,$FC,$4D,$2B
	.dc.b	$F1,$B1,$FD,$F3,$F0,$05,$7F,$AC,$EF,$FD,$69,$8F,$F0,$EE,$2C,$47
	.dc.b	$EF,$0F,$F5,$8B,$FC,$09,$C1,$0F,$EF,$C7,$38,$FE,$70,$7F,$0A,$7A
	.dc.b	$98,$3F,$C1,$07,$FB,$4D,$FD,$DB,$93,$99,$FC,$51,$FE,$00,$9F,$F4
	.dc.b	$DC,$61,$FF,$61,$47,$FA,$78,$D4,$70,$5D,$F5,$E5,$C9,$FD,$9F,$9B
	.dc.b	$FD,$73,$7F,$D6,$DF,$F0,$1C,$E9,$D1,$FD,$DA,$34,$FF,$B0,$A3,$C1
	.dc.b	$18,$8F,$EF,$71,$C5,$D3,$A9,$FF,$63,$B6,$32,$D7,$8B,$FE,$B8,$9F
	.dc.b	$27,$3F,$F3,$FA,$FE,$FD,$36,$BF,$C0,$EF,$F3,$E8,$E2,$F7,$9F,$8D
	.dc.b	$1F,$0E,$BC,$D5,$FE,$B4,$6B,$FD,$F6,$BC,$13,$F0,$23,$FC,$05,$1E
	.dc.b	$B7,$FC,$1D,$9F,$CF,$8D,$C7,$CB,$E7,$F1,$3F,$EC,$28,$FB,$D5,$97
	.dc.b	$31,$FE,$29,$AF,$C3,$CF,$F8,$2A,$73,$1E,$DA,$FE,$81,$7F,$E0,$F8
	.dc.b	$B5,$E4,$AF,$11,$F7,$23,$F0,$9B,$FB,$F8,$BF,$D8,$13,$F2,$DF,$E4
	.dc.b	$5C,$7F,$EB,$58,$FC,$64,$7F,$44,$7E,$DB,$FE,$C5,$6C,$7F,$AF,$5E
	.dc.b	$4F,$16,$BF,$D0,$1F,$CE,$46,$5C,$F3,$FF,$77,$8F,$E8,$71,$D2,$3D
	.dc.b	$B1,$FD,$0E,$38,$C7,$E7,$61,$CF,$F9,$FE,$7F,$B5,$CF,$ED,$C3,$D6
	.dc.b	$7F,$3D,$C3,$CE,$FF,$AF,$1A,$FF,$AF,$1F,$7B,$97,$2F,$FA,$D9,$CB
	.dc.b	$FE,$BA,$7D,$6F,$17,$DE,$FF,$B2,$DF,$A3,$A3,$0D,$47,$C4,$E4,$F8
	.dc.b	$15,$D6,$3E,$5C,$7C,$08,$F7,$3E,$F8,$FD,$E7,$81,$F9,$7F,$A3,$FE
	.dc.b	$31,$CD,$FC,$1D,$FE,$35,$CD,$63,$FD,$79,$5F,$A7,$9F,$E7,$B3,$E4
	.dc.b	$73,$19,$9F,$F5,$E3,$5F,$E6,$6F,$F0,$D8,$CB,$98,$FF,$17,$47,$F8
	.dc.b	$E6,$3A,$47,$F3,$47,$A4,$FE,$5A,$71,$1E,$38,$FE,$2D,$1F,$B1,$E4
	.dc.b	$FD,$95,$FF,$1F,$AF,$E3,$8F,$F9,$7F,$9B,$F9,$23,$FE,$34,$8F,$E1
	.dc.b	$AF,$FA,$F6,$3E,$7B,$FC,$CE,$38,$9F,$E3,$02,$9F,$CD,$4A,$7F,$2C
	.dc.b	$39,$99,$37,$1F,$E4,$B8,$C4,$74,$8F,$D8,$C7,$F2,$E8,$F3,$C7,$F3
	.dc.b	$1F,$89,$FE,$19,$CD,$FC,$1C,$FF,$51,$ED,$5F,$E2,$AD,$AF,$E8,$11
	.dc.b	$FE,$B8,$7F,$94,$BF,$C3,$DF,$C5,$BF,$CC,$9F,$E4,$4F,$F2,$86,$BF
	.dc.b	$CA,$1A,$FF,$20,$8F,$E7,$98,$7F,$99,$63,$F9,$3C,$FF,$54,$7F,$D7
	.dc.b	$11,$FC,$31,$FE,$4F,$1D,$A3,$FA,$64,$7B,$0F,$E4,$5C,$DE,$87,$F9
	.dc.b	$27,$2D,$7F,$8E,$57,$A9,$FE,$8F,$1E,$A7,$31,$97,$FD,$45,$7F,$D3
	.dc.b	$2B,$C9,$1F,$EA,$59,$FE,$BC,$FF,$23,$79,$3F,$CD,$5E,$91,$E9,$8C
	.dc.b	$47,$C8,$8F,$4D,$7F,$65,$7F,$A0,$3F,$D0,$7E,$53,$F9,$1C,$FF,$6B
	.dc.b	$7F,$84,$F0,$7F,$96,$B0,$47,$F2,$39,$FE,$A4,$FF,$23,$6C,$FF,$23
	.dc.b	$79,$3F,$C7,$E7,$8B,$FE,$4F,$E5,$FF,$01,$D7,$F2,$3A,$DC,$FF,$97
	.dc.b	$DE,$6F,$F1,$E7,$6E,$5E,$FC,$BF,$D0,$A7,$11,$E4,$8F,$F0,$17,$CD
	.dc.b	$FB,$6B,$D6,$7F,$83,$6A,$F1,$D8,$F7,$DD,$7F,$CC,$2F,$F7,$A9,$CB
	.dc.b	$98,$FF,$2B,$BE,$87,$F8,$D6,$3E,$E7,$31,$93,$D2,$FF,$25,$7F,$92
	.dc.b	$CF,$F3,$57,$DB,$5E,$B9,$FF,$0D,$4E,$E7,$FD,$6E,$FF,$77,$EC,$FE
	.dc.b	$F1,$D9,$FC,$8A,$7F,$C1,$EF,$F2,$26,$F2,$FF,$1E,$C5,$3F,$C8,$A7
	.dc.b	$FC,$5C,$FF,$22,$7F,$C5,$4F,$F9,$97,$97,$FD,$8A,$FF,$99,$E2,$D7
	.dc.b	$F1,$F7,$AB,$FC,$7E,$BA,$A3,$FD,$B1,$1D,$1F,$E4,$31,$AA,$FE,$F3
	.dc.b	$5F,$C9,$2B,$C3,$1F,$E3,$98,$F7,$47,$F1,$F3,$FC,$7D,$7F,$C8,$DF
	.dc.b	$F1,$6B,$97,$27,$F9,$34,$A7,$F9,$28,$D9,$FC,$8E,$91,$FE,$5C,$8D
	.dc.b	$9F,$C8,$CF,$F1,$73,$C5,$FF,$2A,$3F,$E5,$67,$9C,$7F,$84,$A3,$FC
	.dc.b	$A3,$1F,$EF,$17,$F9,$24,$76,$E3,$FF,$2A,$76,$7F,$8B,$F4,$FF,$18
	.dc.b	$7C,$0F,$F2,$1C,$3F,$C7,$8F,$F2,$7E,$4F,$F3,$13,$FC,$A6,$2F,$27
	.dc.b	$F2,$1C,$53,$FC,$96,$73,$3F,$C8,$67,$2F,$F9,$6E,$3F,$DF,$F3,$FC
	.dc.b	$89,$B1,$FC,$89,$FE,$EF,$1F,$CA,$5F,$1C,$7F,$22,$9E,$75,$FE,$30
	.dc.b	$AF,$E4,$B5,$FE,$71,$7F,$C9,$EF,$63,$FC,$E4,$FF,$41,$E4,$F2,$DF
	.dc.b	$F9,$06,$78,$3F,$C7,$DC,$BF,$E4,$16,$C7,$9A,$3F,$92,$CF,$F2,$89
	.dc.b	$FF,$36,$BC,$DB,$03,$FC,$84,$FA,$9F,$E4,$6E,$A7,$D5,$18,$8E,$71
	.dc.b	$E8,$E4,$FF,$45,$7C,$27,$F9,$97,$8B,$F9,$2E,$5F,$C2,$9F,$E2,$97
	.dc.b	$FE,$73,$7F,$92,$3F,$E9,$37,$F9,$23,$5F,$E4,$91,$88,$C4,$7F,$91
	.dc.b	$5F,$F4,$3B,$FC,$8E,$3F,$95,$47,$58,$FF,$21,$37,$97,$FC,$C3,$18
	.dc.b	$8F,$14,$7F,$A8,$63,$CB,$1F,$EB,$08,$FF,$3F,$72,$FA,$E7,$F9,$1C
	.dc.b	$F8,$1F,$E7,$FA,$FF,$AC,$5C,$C6,$5F,$E4,$6F,$8D,$FF,$20,$5F,$08
	.dc.b	$FF,$1F,$C7,$F9,$41,$FF,$38,$9E,$E7,$F8,$D3,$C5,$CB,$CE,$6B,$F1
	.dc.b	$0B,$3F,$E4,$F7,$9B,$FC,$9D,$FF,$65,$CF,$47,$D0,$7F,$B0,$7E,$87
	.dc.b	$FD,$39,$F2,$1F,$C9,$63,$FD,$50,$FF,$25,$6F,$2F,$F2,$58,$FE,$49
	.dc.b	$1F,$EB,$87,$F9,$25,$9F,$E4,$91,$FE,$5C,$7F,$92,$36,$7F,$92,$3D
	.dc.b	$1F,$E4,$91,$88,$C1,$EC,$7F,$93,$72,$FF,$A6,$9F,$E4,$55,$DE,$3F
	.dc.b	$DB,$51,$FE,$B5,$F1,$BF,$D6,$D8,$7F,$91,$68,$FF,$23,$BF,$F6,$23
	.dc.b	$FC,$8E,$33,$1F,$C8,$DA,$FF,$1B,$3F,$DD,$11,$FC,$9D,$A6,$4A,$7F
	.dc.b	$1C,$3F,$CF,$6E,$5C,$F2,$7D,$68,$FF,$22,$CF,$82,$3F,$D6,$B1,$FC
	.dc.b	$88,$FE,$59,$3F,$EF,$DF,$57,$FD,$A1,$E3,$FF,$21,$8F,$F7,$D3,$FC
	.dc.b	$85,$FF,$74,$BF,$C8,$5B,$C7,$FC,$86,$3F,$90,$46,$23,$F9,$04,$7F
	.dc.b	$90,$E3,$9C,$7F,$1B,$C1,$3F,$C8,$E7,$F1,$6F,$F1,$F7,$73,$FE,$CC
	.dc.b	$8F,$F2,$1C,$7E,$49,$FE,$3F,$3C,$B9,$7F,$2F,$3F,$EF,$39,$F5,$D7
	.dc.b	$F9,$12,$3C,$8F,$F1,$FA,$F0,$BF,$C7,$E3,$2F,$F1,$F6,$F2,$79,$23
	.dc.b	$F8,$E4,$79,$1F,$E3,$73,$C9,$E1,$1F,$EE,$F8,$C4,$F9,$E7,$33,$9A
	.dc.b	$FE,$49,$3E,$D9,$FE,$41,$3E,$98,$F5,$BF,$4B,$92,$DB,$C6,$CC,$A3
	.dc.b	$FB,$EA,$22,$F4,$7F,$93,$9F,$E3,$0E,$4F,$33,$FC,$9A,$35,$C7,$FE
	.dc.b	$D6,$9F,$F6,$3C,$FF,$B6,$39,$7F,$7E,$E6,$F3,$5F,$E4,$97,$2F,$FB
	.dc.b	$A1,$F2,$46,$5C,$C7,$F2,$27,$C2,$FF,$91,$AF,$C3,$7F,$69,$F4,$5F
	.dc.b	$C0,$BE,$17,$FE,$DA,$F9,$EF,$83,$5E,$8C,$3F,$C8,$73,$E0,$7F,$91
	.dc.b	$3F,$0A,$BF,$94,$B6,$3F,$DD,$8D,$7F,$94,$C6,$23,$11,$FE,$4A,$8F
	.dc.b	$E5,$F1,$F3,$4E,$D1,$FE,$E7,$6F,$37,$F2,$69,$FE,$49,$1F,$1A,$3D
	.dc.b	$91,$B8,$F2,$47,$B5,$FB,$BC,$DF,$C9,$5F,$B8,$FF,$25,$6B,$FE,$49
	.dc.b	$3F,$92,$BF,$79,$FE,$4A,$7F,$24,$7F,$24,$FF,$24,$6B,$FC,$86,$F1
	.dc.b	$18,$8F,$F2,$2C,$7F,$91,$DF,$F2,$29,$FC,$92,$3F,$A0,$BF,$C9,$0F
	.dc.b	$C7,$BE,$63,$F2,$FE,$37,$F4,$67,$E3,$3F,$D2,$B9,$3E,$44,$23,$FD
	.dc.b	$D8,$D3,$FD,$D0,$69,$FF,$25,$39,$3F,$DD,$0F,$FB,$AD,$C1,$FC,$8C
	.dc.b	$E0,$7F,$23,$7F,$DD,$CF,$E8,$DF,$E4,$CF,$38,$FE,$0F,$1F,$EE,$A8
	.dc.b	$F1,$47,$AA,$BF,$5B,$3F,$C9,$A7,$D3,$E5,$FE,$C1,$F5,$78,$FF,$C9
	.dc.b	$E4,$FF,$27,$8F,$03,$FC,$9D,$BC,$5F,$EE,$79,$DB,$FE,$E7,$6B,$FE
	.dc.b	$E7,$6B,$FE,$29,$8F,$EF,$31,$8E,$4F,$EE,$EE,$A2,$D7,$F2,$98,$E7
	.dc.b	$1F,$DE,$9E,$F3,$F0,$27,$FC,$8F,$1E,$88,$FB,$31,$F0,$0F,$F2,$34
	.dc.b	$FD,$A7,$FC,$75,$5F,$EE,$88,$C3,$FC,$A5,$F6,$BF,$E4,$1A,$F4,$AE
	.dc.b	$3F,$4D,$71,$AF,$BC,$FF,$92,$5E,$33,$FE,$41,$AF,$87,$5E,$9A,$FF
	.dc.b	$78,$57,$D0,$AF,$F2,$5B,$F3,$38,$FD,$95,$EA,$7F,$27,$E3,$FD,$AF
	.dc.b	$17,$FD,$D7,$89,$FE,$45,$7F,$96,$7F,$C8,$F1,$9E,$1F,$E3,$D3,$97
	.dc.b	$FD,$D5,$19,$44,$E4,$FA,$B5,$FE,$46,$9D,$4F,$F1,$8A,$FD,$04,$FF
	.dc.b	$23,$7F,$46,$62,$3F,$C9,$13,$E1,$E4,$FC,$AD,$7F,$92,$EB,$8C,$7E
	.dc.b	$BA,$3A,$47,$EB,$BC,$7F,$CB,$F8,$FE,$FC,$3F,$8F,$E1,$E2,$7F,$C7
	.dc.b	$0D,$7D,$35,$FA,$FB,$FE,$4E,$FE,$9D,$FE,$56,$D8,$FE,$42,$F9,$DF
	.dc.b	$E3,$AE,$5F,$F7,$6C,$FC,$08,$FC,$AC,$FF,$91,$DE,$D3,$FC,$86,$7E
	.dc.b	$77,$27,$BA,$3D,$8F,$B6,$7F,$85,$F8,$DF,$B2,$F1,$FF,$92,$59,$FE
	.dc.b	$48,$DE,$6F,$E3,$07,$F1,$9A,$FE,$15,$3F,$C9,$9F,$E2,$0F,$FB,$B6
	.dc.b	$78,$4F,$F9,$19,$AF,$FB,$9E,$7F,$5F,$3D,$3B,$FF,$5E,$F6,$8B,$3F
	.dc.b	$C9,$5D,$BC,$5C,$CF,$F0,$E9,$FF,$76,$57,$A6,$3F,$93,$C7,$F0,$EE
	.dc.b	$4F,$E1,$F8,$7F,$9C,$B9,$BF,$91,$3F,$E4,$BC,$3F,$94,$3F,$EE,$DC
	.dc.b	$0F,$F2,$88,$FE,$5E,$FE,$32,$7F,$83,$BF,$E4,$B9,$DC,$FF,$92,$DA
	.dc.b	$FF,$29,$38,$9F,$E0,$83,$F8,$6C,$7F,$BB,$5F,$E5,$4E,$5C,$9F,$C9
	.dc.b	$4F,$2C,$FD,$8E,$7F,$E8,$7E,$3F,$F2,$AE,$6F,$E4,$C7,$F2,$5A,$FE
	.dc.b	$84,$FF,$25,$6B,$FC,$94,$FF,$25,$BF,$CE,$5F,$E2,$27,$F9,$09,$FE
	.dc.b	$8C,$FF,$93,$70,$CB,$FE,$E0,$8E,$0F,$FB,$AD,$DB,$C5,$FD,$8C,$7F
	.dc.b	$94,$5F,$E5,$B1,$97,$2F,$F2,$63,$A3,$E0,$3F,$AF,$BF,$C9,$9E,$EF
	.dc.b	$F1,$D7,$D6,$FF,$B9,$63,$CA,$FF,$90,$9F,$F2,$A4,$79,$23,$F9,$6F
	.dc.b	$2F,$89,$FE,$5F,$CD,$FC,$F9,$FF,$28,$CF,$48,$FE,$79,$1F,$C9,$2B
	.dc.b	$73,$88,$FF,$25,$BF,$3A,$7C,$F5,$FD,$D9,$FE,$BA,$FF,$5E,$F1,$BF
	.dc.b	$92,$9F,$C9,$27,$FB,$43,$FC,$90,$FF,$74,$D9,$FF,$22,$47,$CA,$7F
	.dc.b	$93,$BF,$E0,$07,$F9,$11,$FC,$9B,$87,$FD,$CF,$3C,$5F,$F7,$74,$ED
	.dc.b	$FE,$43,$38,$8C,$73,$7F,$76,$87,$37,$F8,$39,$FE,$43,$3F,$DF,$7C
	.dc.b	$7F,$55,$FF,$20,$9F,$03,$FE,$E6,$C3,$C0,$FF,$B8,$EB,$C0,$FF,$8F
	.dc.b	$23,$2F,$F1,$C6,$BF,$EE,$66,$C7,$F7,$29,$FE,$46,$FF,$86,$DE,$EF
	.dc.b	$F7,$47,$F9,$03,$D8,$FE,$E4,$FF,$48,$E4,$F6,$57,$AD,$F6,$4F,$F8
	.dc.b	$AB,$C6,$FF,$08,$F8,$FF,$C8,$0F,$E3,$E7,$F1,$EE,$6F,$E3,$D3,$FE
	.dc.b	$2E,$8F,$E3,$D8,$9F,$E3,$C7,$F9,$0E,$35,$3F,$C6,$5F,$D0,$C7,$F1
	.dc.b	$D7,$FC,$1A,$E5,$CC,$7F,$1D,$6B,$FE,$28,$8C,$46,$8C,$73,$7F,$90
	.dc.b	$A3,$FC,$53,$7F,$C6,$E7,$FC,$86,$FF,$1B,$B3,$FC,$6C,$FE,$35,$1F
	.dc.b	$E3,$77,$F8,$D3,$FB,$67,$F8,$D1,$FE,$AD,$7F,$C4,$6F,$F8,$FA,$39
	.dc.b	$C7,$F1,$53,$81,$FC,$74,$F0,$9F,$EA,$73,$FC,$49,$19,$7C,$BC,$7F
	.dc.b	$47,$8F,$FC,$B7,$E3,$FF,$90,$B3,$FF,$21,$F8,$FF,$E3,$F3,$F8,$D6
	.dc.b	$27,$F8,$D3,$79,$3F,$C7,$D1,$FC,$6B,$1F,$F3,$03,$FC,$5B,$9F,$FC
	.dc.b	$76,$FF,$19,$3F,$8E,$BF,$E3,$C7,$F8,$A4,$7F,$98,$DF,$F1,$E1,$FC
	.dc.b	$6D,$C4,$FF,$14,$7F,$CB,$D1,$FE,$3C,$30,$E0,$FF,$3A,$BF,$C5,$DF
	.dc.b	$1C,$7F,$14,$E4,$E6,$FF,$17,$9C,$4E,$23,$FD,$AE,$FF,$28,$8F,$E3
	.dc.b	$CF,$82,$3F,$88,$B9,$7F,$8D,$BE,$E7,$8B,$5F,$F6,$FD,$FC,$7E,$5F
	.dc.b	$F3,$9D,$7F,$B9,$AB,$E0,$47,$A2,$FF,$D1,$BE,$3F,$F9,$D3,$8B,$FC
	.dc.b	$76,$7F,$17,$AF,$F4,$63,$FE,$3A,$3F,$8B,$9F,$C5,$9F,$F4,$63,$FC
	.dc.b	$59,$FF,$42,$3F,$E3,$73,$F8,$3F,$37,$F9,$B1,$FF,$1B,$47,$07,$FC
	.dc.b	$6A,$7F,$15,$78,$BF,$C4,$B1,$C4,$74,$9F,$F5,$1F,$37,$FA,$29,$F4
	.dc.b	$57,$F8,$D0,$FE,$2A,$7F,$8D,$0F,$F6,$AC,$FA,$9F,$F6,$95,$62,$31
	.dc.b	$1F,$C3,$23,$FC,$6B,$1E,$28,$FF,$18,$9F,$C4,$9F,$F4,$3B,$FC,$49
	.dc.b	$D1,$FE,$83,$7F,$89,$35,$FF,$5C,$8D,$6C,$7F,$AF,$3C,$5F,$F5,$1F
	.dc.b	$8F,$FE,$32,$3F,$C6,$27,$F1,$3E,$6F,$F1,$04,$FF,$AB,$E3,$F8,$99
	.dc.b	$FE,$2F,$3F,$89,$39,$8F,$E1,$CF,$FB,$1A,$3F,$89,$3F,$DD,$1C,$B9
	.dc.b	$8F,$E1,$ED,$7F,$D7,$31,$88,$D1,$8E,$6F,$F6,$9C,$7F,$AE,$BC,$7F
	.dc.b	$F5,$DD,$FF,$13,$AF,$F6,$23,$FC,$4C,$FF,$16,$1F,$C4,$9F,$F6,$B3
	.dc.b	$FC,$49,$FF,$6A,$3F,$E2,$B3,$F8,$93,$FE,$0A,$7F,$89,$1F,$ED,$13
	.dc.b	$FC,$50,$F2,$3F,$88,$1F,$EB,$87,$F8,$83,$E2,$E5,$F9,$DC,$BF,$EE
	.dc.b	$AF,$1F,$FD,$A3,$E3,$FF,$B3,$4F,$E1,$E7,$F0,$E3,$FC,$17,$CD,$FE
	.dc.b	$1A,$7F,$87,$6E,$7F,$87,$61,$9E,$6F,$E1,$6F,$FB,$8E,$3F,$87,$BF
	.dc.b	$EC,$C7,$2E,$63,$FC,$44,$FF,$BE,$A3,$F8,$83,$63,$F8,$6C,$F7,$7F
	.dc.b	$E0,$0C,$FF,$10,$9F,$F7,$F3,$EA,$8F,$F0,$FC,$61,$FE,$1A,$F7,$7F
	.dc.b	$C4,$D1,$88,$C4,$7F,$08,$7C,$0F,$F0,$C4,$61,$FF,$67,$3E,$B7,$4E
	.dc.b	$8E,$DC,$BF,$EE,$39,$FF,$14,$1D,$63,$E3,$72,$FF,$AD,$1E,$07,$AA
	.dc.b	$7D,$B1,$98,$CD,$BC,$7C,$46,$38,$2B,$E1,$97,$8D,$FE,$D2,$3F,$8C
	.dc.b	$38,$E4,$FE,$2D,$3E,$C8,$C4,$7F,$14,$8E,$F3,$A7,$4F,$F1,$B5,$3C
	.dc.b	$27,$FC,$4F,$1E,$B8,$C7,$2F,$A5,$EB,$CD,$E8,$7B,$C5,$BF,$E3,$71
	.dc.b	$E5,$8F,$5B,$EA,$9F,$6D,$78,$2B,$FC,$66,$F8,$BC,$7C,$B9,$C3,$FD
	.dc.b	$A7,$3F,$C8,$1F,$E2,$31,$EA,$7F,$D9,$F7,$E9,$73,$19,$7F,$8E,$3F
	.dc.b	$2E,$7F,$94,$F3,$FB,$5F,$F1,$E4,$7C,$69,$FE,$41,$5F,$0E,$3E,$04
	.dc.b	$FF,$23,$7C,$93,$EC,$8E,$F1,$FE,$84,$7F,$DC,$AF,$C3,$AF,$7D,$7D
	.dc.b	$A8,$BC,$FF,$ED,$43,$FD,$A6,$7F,$23,$3F,$D1,$A7,$F8,$D0,$FF,$68
	.dc.b	$1F,$C8,$9F,$AE,$FF,$8E,$DF,$AC,$FF,$22,$7F,$1A,$FF,$22,$7F,$08
	.dc.b	$8F,$E4,$0F,$27,$F9,$8B,$E0,$59,$FD,$87,$FD,$A2,$62,$30,$7D,$AF
	.dc.b	$1F,$F3,$3C,$BF,$A0,$7F,$37,$CB,$67,$F4,$93,$FA,$45,$3F,$ED,$56
	.dc.b	$9F,$C9,$CF,$D8,$3F,$CA,$DC,$9F,$B3,$8F,$D5,$47,$39,$B3,$92,$9F
	.dc.b	$C9,$A3,$0E,$23,$F5,$B1,$FE,$3F,$8F,$DC,$3F,$CB,$1F,$E0,$AF,$F3
	.dc.b	$07,$F6,$AF,$FB,$5D,$FE,$13,$E5,$FE,$59,$F5,$F9,$1F,$EB,$6E,$51
	.dc.b	$F0,$DF,$F6,$C5,$1F,$F6,$C3,$63,$F9,$7C,$FC,$09,$FF,$20,$4E,$67
	.dc.b	$99,$FE,$D5,$70,$F6,$9F,$4C,$FF,$34,$7E,$74,$E9,$F8,$F1,$CB,$9B
	.dc.b	$D6,$FF,$34,$8F,$7B,$FC,$E1,$F1,$57,$DD,$8F,$E6,$95,$E9,$9C,$4F
	.dc.b	$F3,$48,$F1,$BF,$CA,$1A,$FF,$2C,$8E,$6F,$FB,$CB,$93,$CB,$5F,$EF
	.dc.b	$0E,$5F,$4C,$E6,$B2,$FF,$91,$1F,$94,$FF,$34,$79,$C7,$AA,$3F,$C8
	.dc.b	$11,$EA,$7C,$0F,$4A,$FB,$F5,$88,$F8,$F1,$F9,$9F,$23,$F8,$7E,$0F
	.dc.b	$1B,$F8,$77,$1F,$F2,$2B,$FC,$1C,$7F,$91,$E7,$53,$FC,$AD,$AF,$F2
	.dc.b	$F9,$FC,$CC,$FF,$2F,$7F,$3E,$FF,$2F,$9D,$CF,$59,$FD,$44,$F1,$6C
	.dc.b	$7F,$91,$E7,$AF,$37,$E4,$AB,$FC,$8F,$1E,$28,$FC,$DC,$73,$AF,$CD
	.dc.b	$BF,$CB,$25,$E4,$7B,$BC,$8F,$7C,$FB,$B0,$EE,$7F,$24,$BF,$4B,$FC
	.dc.b	$BD,$CB,$FE,$F3,$E5,$FC,$6C,$FF,$92,$B0,$F5,$CF,$F9,$06,$B2,$FF
	.dc.b	$30,$79,$BF,$C9,$9F,$96,$FF,$BF,$5F,$DD,$CF,$F2,$D9,$F9,$71,$EF
	.dc.b	$9E,$8E,$9F,$E2,$13,$88,$E2,$7F,$13,$F1,$FE,$DF,$91,$FC,$AF,$97
	.dc.b	$F8,$A3,$FC,$B7,$9A,$C7,$F9,$46,$AB,$FC,$A5,$FE,$2B,$39,$73,$1F
	.dc.b	$E4,$AC,$53,$FC,$C6,$7F,$5E,$F1,$7F,$84,$BF,$CC,$5E,$73,$FC,$C2
	.dc.b	$78,$BF,$CC,$23,$C7,$5F,$C5,$2B,$11,$E8,$8F,$E5,$D1,$F9,$6E,$3F
	.dc.b	$CB,$E1,$FE,$53,$AF,$E5,$6F,$F8,$37,$97,$F9,$93,$FC,$C2,$29,$FC
	.dc.b	$9C,$A7,$F2,$43,$F9,$7B,$97,$27,$F2,$13,$27,$F8,$8E,$3F,$98,$C7
	.dc.b	$18,$E7,$1F,$C9,$A3,$F9,$64,$78,$23,$F9,$7C,$73,$8F,$E2,$AF,$F2
	.dc.b	$BC,$5E,$47,$EC,$B9,$3F,$95,$1F,$D2,$CF,$E5,$27,$F2,$83,$FC,$B3
	.dc.b	$59,$7F,$92,$9F,$C8,$DC,$3F,$C7,$DF,$E4,$6E,$8F,$E9,$27,$F2,$77
	.dc.b	$6F,$F3,$E3,$83,$FD,$11,$FE,$4C,$FF,$58,$7F,$93,$3F,$D1,$9F,$03
	.dc.b	$D3,$0F,$EC,$91,$EB,$3F,$93,$1F,$D1,$78,$FB,$BF,$E4,$9B,$C5,$63
	.dc.b	$97,$C1,$3F,$C8,$E3,$CB,$1E,$88,$FE,$45,$1E,$17,$FC,$A4,$FF,$68
	.dc.b	$72,$E6,$3F,$AF,$C7,$F2,$68,$F4,$47,$9A,$BF,$BD,$3F,$C8,$9F,$E4
	.dc.b	$7E,$3F,$F2,$2E,$CF,$EF,$8F,$F2,$13,$FC,$64,$FF,$4E,$7F,$C6,$9C
	.dc.b	$B7,$88,$7F,$B8,$B0,$FE,$FF,$3F,$C8,$4C,$C6,$AB,$FB,$DC,$7F,$21
	.dc.b	$9F,$F0,$0B,$C5,$FE,$E6,$F0,$AF,$EE,$6F,$F2,$23,$BB,$FD,$F1,$FE
	.dc.b	$45,$1E,$58,$FF,$0B,$79,$1E,$CB,$FE,$7B,$3D,$9F,$F2,$67,$27,$B9
	.dc.b	$FE,$41,$19,$7C,$E7,$F0,$F7,$D6,$FF,$92,$1F,$2B,$A8,$FF,$05,$DE
	.dc.b	$67,$33,$FC,$85,$E2,$FF,$91,$9F,$90,$F5,$7C,$07,$B9,$C3,$E7,$3D
	.dc.b	$D3,$FE,$31,$9F,$EB,$3E,$3F,$F5,$AF,$23,$F9,$05,$7F,$8A,$9F,$E4
	.dc.b	$0D,$E5,$FE,$38,$FF,$87,$A3,$FD,$E5,$CB,$FE,$08,$73,$19,$9F,$E3
	.dc.b	$9C,$09,$FE,$41,$3F,$E3,$F7,$F9,$03,$FE,$3C,$7F,$90,$46,$E7,$F9
	.dc.b	$04,$7F,$94,$5F,$F1,$8B,$FE,$52,$9E,$11,$FE,$4B,$8F,$F2,$74,$7F
	.dc.b	$8C,$1F,$6B,$FC,$6C,$FF,$1A,$DF,$F1,$F7,$FC,$8E,$E5,$C9,$FC,$6C
	.dc.b	$A7,$F1,$73,$FC,$A2,$E5,$C9,$FC,$54,$C9,$FC,$7C,$FF,$22,$BC,$5E
	.dc.b	$7C,$9F,$E4,$48,$FE,$3F,$1E,$21,$1F,$E4,$E8,$E9,$1E,$68,$FF,$34
	.dc.b	$79,$1F,$E1,$0C,$FF,$C2,$3E,$47,$F2,$0A,$FF,$30,$BF,$E4,$83,$FC
	.dc.b	$8E,$7F,$20,$73,$C7,$FC,$6D,$FF,$1F,$C7,$F9,$29,$CC,$7F,$BC,$33
	.dc.b	$4F,$F1,$F9,$CC,$FF,$97,$23,$F8,$FC,$72,$8F,$E3,$6D,$E4,$FF,$33
	.dc.b	$CE,$23,$BC,$7F,$A0,$63,$D3,$3F,$E4,$37,$F8,$DC,$FF,$9A,$9F,$F7
	.dc.b	$CB,$FE,$80,$7F,$DE,$CD,$8F,$E7,$3C,$9F,$E7,$29,$FE,$6D,$C9,$67
	.dc.b	$FC,$85,$5E,$88,$FF,$40,$CE,$23,$DF,$1E,$B9,$FF,$1F,$47,$FA,$4E
	.dc.b	$7F,$90,$3E,$1A,$F4,$57,$FA,$67,$C8,$FF,$20,$F1,$7F,$1F,$AF,$F4
	.dc.b	$8B,$FE,$3F,$3F,$8F,$BF,$E9,$D7,$F8,$FB,$5F,$E0,$BC,$BF,$E8,$A7
	.dc.b	$FC,$7C,$62,$30,$FF,$19,$6B,$FC,$59,$FF,$47,$B9,$77,$1F,$E9,$68
	.dc.b	$FE,$3B,$1F,$E9,$27,$83,$E6,$E5,$FF,$59,$4F,$F9,$69,$F5,$CF,$F1
	.dc.b	$E9,$EA,$FF,$1E,$7D,$AF,$F1,$E7,$2F,$F1,$97,$D8,$FF,$8E,$5F,$F5
	.dc.b	$BB,$FE,$45,$72,$FF,$19,$79,$BF,$C7,$74,$FE,$41,$1D,$35,$F4,$4E
	.dc.b	$E3,$E4,$47,$B6,$3F,$D9,$47,$F1,$E8,$FA,$27,$FB,$37,$C6,$FF,$3A
	.dc.b	$F9,$1F,$C7,$23,$FD,$6A,$FF,$1C,$6F,$27,$F1,$8C,$13,$FC,$72,$7F
	.dc.b	$DA,$8F,$F1,$C7,$FD,$A2,$FF,$BA,$E3,$73,$FE,$E7,$6B,$FE,$E7,$E5
	.dc.b	$FF,$67,$CF,$F1,$C3,$F8,$D3,$FE,$C4,$7F,$8E,$55,$9F,$E3,$2F,$FB
	.dc.b	$32,$3F,$C7,$6F,$FB,$71,$CB,$93,$FD,$D5,$E4,$7F,$A4,$B0,$FF,$1E
	.dc.b	$60,$7F,$8F,$4F,$FB,$51,$FE,$3C,$FF,$BA,$9F,$F7,$63,$FC,$40,$FF
	.dc.b	$79,$3F,$EE,$B7,$80,$3F,$8E,$1A,$3F,$8E,$47,$77,$26,$4F,$E3,$2E
	.dc.b	$5E,$C7,$FB,$A2,$3F,$90,$BD,$B9,$3F,$DD,$B1,$FC,$82,$3D,$11,$E6
	.dc.b	$7C,$47,$FB,$BB,$C8,$FF,$58,$F9,$1F,$C8,$63,$FD,$E2,$FF,$BA,$0F
	.dc.b	$F7,$39,$FC,$85,$CF,$27,$F2,$17,$98,$C1,$3F,$C8,$E3,$73,$FC,$8E
	.dc.b	$3F,$91,$46,$23,$FD,$FB,$1F,$C8,$A3,$9C,$F2,$9F,$E4,$11,$E8,$8C
	.dc.b	$72,$FA,$E3,$C6,$53,$C5,$5F,$EE,$83,$DB,$5F,$EE,$77,$F9,$2C,$6A
	.dc.b	$3F,$DC,$4F,$82,$3F,$92,$D7,$86,$7F,$8F,$35,$0E,$3F,$C0,$C6,$79
	.dc.b	$3E,$73,$B8,$B5,$FC,$9E,$FE,$1C,$7C,$0B,$FE,$4F,$1E,$D9,$FE,$4F
	.dc.b	$5F,$2A,$FD,$93,$EC,$BB,$2F,$23,$FD,$D3,$C1,$FE,$EA,$F1,$FF,$95
	.dc.b	$5F,$E3,$E9,$1F,$EE,$93,$F9,$5C,$7C,$B7,$F9,$5B,$78,$BF,$94,$3F
	.dc.b	$4A,$3F,$DD,$06,$1C,$47,$F2,$76,$BF,$C9,$27,$BB,$FC,$A2,$3E,$2C
	.dc.b	$FF,$BA,$0C,$38,$E5,$F5,$C7,$FB,$9F,$C8,$F5,$57,$F2,$9A,$FC,$64
	.dc.b	$FF,$90,$A3,$FD,$C5,$1E,$77,$FC,$85,$7E,$07,$F9,$54,$65,$FF,$6F
	.dc.b	$3E,$67,$FD,$C5,$1B,$7F,$92,$47,$67,$AB,$F9,$C3,$0E,$9B,$CB,$F9
	.dc.b	$19,$F0,$45,$AF,$F2,$73,$F4,$5E,$0D,$9F,$7B,$F6,$B9,$7F,$96,$3F
	.dc.b	$A9,$7F,$96,$35,$FE,$58,$7F,$2B,$7F,$44,$FF,$2B,$AC,$46,$23,$F9
	.dc.b	$51,$FE,$44,$8F,$F2,$3C,$7E,$B2,$3F,$94,$BF,$96,$7F,$94,$B5,$FE
	.dc.b	$4A,$ED,$F3,$65,$E7,$8F,$C9,$F8,$DF,$B1,$F1,$FF,$80,$BF,$B3,$E4
	.dc.b	$B3,$FC,$9A,$3F,$94,$BF,$C8,$5F,$E5,$29,$FE,$53,$1F,$CA,$9F,$E5
	.dc.b	$11,$B3,$F8,$6C,$7F,$19,$8F,$E1,$F3,$FC,$05,$CC,$E4,$FE,$46,$75
	.dc.b	$3F,$98,$4F,$F0,$58,$CC,$66,$3D,$73,$FA,$E8,$F2,$45,$F2,$F1,$1E
	.dc.b	$7F,$23,$F9,$33,$79,$3F,$92,$60,$9F,$F6,$F4,$7F,$8E,$A3,$0F,$F2
	.dc.b	$36,$BF,$EE,$89,$FC,$74,$62,$3F,$86,$F2,$FC,$3A,$FE,$5D,$5F,$39
	.dc.b	$FE,$52,$F8,$67,$F9,$2B,$67,$E2,$4F,$FB,$A2,$BC,$31,$F0,$63,$73
	.dc.b	$F4,$D4,$7D,$C7,$E1,$4F,$8E,$BF,$8F,$8E,$4F,$13,$EE,$E4,$AF,$F9
	.dc.b	$2B,$0A,$F0,$8F,$6D,$FF,$2E,$9C,$CF,$F2,$97,$2F,$F2,$5B,$F8,$55
	.dc.b	$F7,$EB,$F9,$45,$7B,$E3,$DB,$5F,$CA,$23,$D4,$F3,$78,$5F,$AE,$7F
	.dc.b	$2F,$E4,$7B,$3C,$6F,$E4,$B5,$F9,$17,$F8,$67,$1F,$F1,$4C,$91,$FC
	.dc.b	$9A,$7F,$30,$FF,$26,$6C,$FF,$23,$6B,$FC,$8D,$AF,$F9,$36,$79,$B9
	.dc.b	$8D,$F2,$FE,$66,$B1,$1D,$23,$F2,$D1,$FC,$9A,$3F,$0A,$FF,$94,$1F
	.dc.b	$1F,$67,$EC,$BC,$8F,$AF,$E4,$7B,$70,$FE,$51,$8F,$81,$FF,$24,$5F
	.dc.b	$BD,$FF,$28,$35,$FE,$4E,$FA,$1F,$F2,$34,$71,$47,$FB,$95,$C3,$FB
	.dc.b	$3C,$38,$CF,$F8,$32,$78,$39,$9F,$E4,$71,$E3,$8F,$81,$1F,$E4,$F7
	.dc.b	$C8,$7C,$0E,$6F,$E1,$B5,$F9,$2F,$1F,$F2,$7E,$47,$F2,$6C,$0F,$F2
	.dc.b	$66,$F3,$FF,$BA,$1A,$FF,$B7,$CF,$F1,$FB,$87,$F8,$FE,$1E,$79,$D4
	.dc.b	$7F,$20,$9D,$3F,$C3,$E3,$FD,$DB,$1C,$67,$13,$FE,$55,$6B,$FC,$36
	.dc.b	$7F,$DD,$93,$E4,$9C,$73,$FF,$28,$9F,$D6,$F2,$7B,$67,$FD,$D1,$4C
	.dc.b	$3F,$93,$B6,$3F,$C9,$4D,$7F,$8F,$9F,$3E,$3F,$93,$C6,$63,$F8,$F9
	.dc.b	$FC,$32,$3F,$93,$BF,$CB,$5E,$2F,$F3,$87,$FD,$D8,$FF,$33,$7F,$CA
	.dc.b	$2E,$4F,$F2,$54,$78,$F9,$BF,$95,$4F,$F9,$36,$78,$47,$F4,$4F,$23
	.dc.b	$F8,$4F,$2F,$F2,$6B,$FE,$46,$FF,$1C,$AF,$C3,$38,$7F,$A4,$BF,$E4
	.dc.b	$D7,$FA,$23,$FC,$A9,$FE,$4A,$FF,$92,$DC,$BF,$CA,$5C,$BF,$E4,$97
	.dc.b	$FA,$73,$97,$2F,$F1,$E7,$FA,$53,$FC,$94,$F2,$38,$7F,$AE,$9B,$7F
	.dc.b	$A7,$3E,$C3,$F9,$EF,$2F,$A1,$FE,$45,$1E,$C7,$F8,$F3,$E2,$7F,$DD
	.dc.b	$6E,$DF,$F7,$44,$FF,$BC,$A3,$D7,$1F,$E4,$C9,$F1,$3F,$E4,$EE,$5B
	.dc.b	$1E,$28,$C4,$F9,$E3,$FA,$E4,$7F,$BB,$AB,$94,$E2,$BF,$BB,$D7,$F3
	.dc.b	$C7,$F9,$EF,$8F,$FC,$8A,$8F,$F2,$2E,$17,$27,$F7,$C8,$FE,$45,$87
	.dc.b	$F7,$77,$F9,$13,$67,$F9,$13,$5F,$E3,$D5,$D1,$FE,$12,$7F,$23,$9A
	.dc.b	$F6,$9F,$F0,$04,$6B,$97,$FB,$E3,$FC,$92,$3B,$47,$F7,$98,$FE,$49
	.dc.b	$1E,$58,$FF,$09,$C7,$38,$FF,$02,$79,$1E,$EA,$FE,$47,$3E,$DE,$4F
	.dc.b	$64,$7F,$22,$C3,$CC,$FF,$1A,$6B,$FE,$48,$BF,$E3,$95,$FC,$A9,$FF
	.dc.b	$0D,$3F,$EE,$E9,$DD,$7F,$85,$63,$FD,$F1,$3A,$AF,$E4,$31,$CE,$3F
	.dc.b	$BF,$BF,$E4,$67,$E5,$1E,$D9,$F5,$BF,$E2,$EF,$1F,$FB,$9F,$91,$FC
	.dc.b	$84,$FE,$40,$7F,$18,$E5,$1F,$E0,$A7,$FC,$87,$3F,$E2,$D7,$FD,$E1
	.dc.b	$89,$FE,$3F,$CB,$FE,$2F,$7F,$C8,$B3,$FE,$35,$7F,$8F,$9F,$CA,$5F
	.dc.b	$F1,$83,$FE,$48,$78,$57,$FB,$B1,$EC,$FF,$92,$1F,$E8,$51,$97,$C5
	.dc.b	$3F,$E4,$88,$FF,$06,$DF,$FB,$4F,$B3,$FC,$92,$FF,$20,$3F,$C9,$07
	.dc.b	$F9,$12,$B2,$7F,$1A,$29,$FC,$58,$FF,$25,$39,$72,$7F,$B9,$8D,$C7
	.dc.b	$F9,$1E,$38,$47,$F9,$4D,$FF,$23,$E0,$7F,$94,$C7,$F9,$3D,$FE,$3D
	.dc.b	$19,$E3,$FF,$30,$F9,$1F,$E2,$6C,$FF,$C4,$FE,$47,$F1,$F3,$F8,$FF
	.dc.b	$3F,$F9,$91,$FE,$3D,$8F,$F6,$57,$F8,$F6,$07,$F8,$F4,$6F,$97,$FD
	.dc.b	$CB,$39,$7F,$8D,$47,$17,$F8,$F3,$FE,$69,$9C,$B9,$8F,$E3,$0F,$F9
	.dc.b	$32,$3F,$C8,$47,$F1,$E7,$FC,$DC,$FF,$90,$9E,$55,$FE,$7C,$8F,$5C
	.dc.b	$FF,$1E,$9E,$8F,$F9,$08,$FF,$17,$3E,$17,$F8,$FB,$5D,$72,$78,$5F
	.dc.b	$F2,$24,$7B,$5F,$F2,$23,$97,$F8,$BB,$5E,$AF,$AF,$97,$FC,$FD,$5C
	.dc.b	$23,$D3,$3F,$E4,$17,$C5,$5F,$E7,$D9,$C4,$7C,$58,$FF,$4B,$F9,$1F
	.dc.b	$E5,$CE,$3F,$E4,$07,$F1,$FE,$6F,$F4,$CB,$FC,$7B,$3C,$5E,$23,$F8
	.dc.b	$EC,$7F,$20,$8F,$F4,$C4,$7F,$20,$6B,$FE,$3A,$8C,$46,$23,$F8,$A3
	.dc.b	$FE,$7E,$7F,$C5,$7C,$DF,$E9,$37,$F8,$DC,$7F,$8A,$5F,$F1,$F3,$DA
	.dc.b	$BF,$D5,$F3,$FE,$74,$7D,$53,$FE,$3E,$3F,$8E,$4F,$17,$F8,$E1,$FE
	.dc.b	$5C,$D7,$C0,$FF,$BA,$62,$BF,$ED,$E7,$83,$FC,$55,$FF,$5B,$3F,$E3
	.dc.b	$C6,$B9,$6A,$8F,$F5,$93,$98,$DC,$7B,$63,$FD,$8E,$7F,$8F,$E3,$E4
	.dc.b	$9F,$EC,$AF,$1B,$FD,$1F,$E4,$7F,$8F,$4F,$F1,$E1,$FC,$77,$9B,$FD
	.dc.b	$4A,$FF,$1C,$C8,$FF,$1C,$3F,$8D,$D6,$A7,$F8,$C3,$5F,$F1,$B3,$5F
	.dc.b	$F6,$44,$7F,$1B,$8D,$18,$8F,$E3,$4F,$FB,$5E,$3F,$8D,$4F,$83,$9B
	.dc.b	$FD,$C3,$1F,$EA,$6F,$23,$FD,$4F,$87,$F1,$BB,$FF,$31,$3F,$E3,$A3
	.dc.b	$FD,$C4,$7F,$1B,$72,$7F,$17,$3F,$DC,$89,$FF,$1A,$98,$30,$39,$BF
	.dc.b	$DC,$51,$A8,$D1,$93,$27,$F1,$A3,$FD,$9A,$FF,$1E,$30,$E2,$3E,$97
	.dc.b	$2F,$FB,$F3,$C8,$FF,$60,$F9,$1F,$E3,$93,$F8,$DD,$7F,$9F,$5F,$E3
	.dc.b	$6F,$FB,$ED,$FE,$36,$7F,$1A,$75,$C9,$FC,$6A,$3F,$DB,$31,$88,$FE
	.dc.b	$24,$D7,$FD,$F5,$1F,$C7,$21,$A2,$3F,$DB,$47,$08,$F8,$66,$9F,$E3
	.dc.b	$51,$BE,$6C,$CC,$34,$53,$8C,$75,$AF,$E3,$D1,$EB,$7F,$C7,$6D,$7F
	.dc.b	$8C,$B5,$FE,$24,$D7,$FD,$C8,$FF,$BE,$AB,$FC,$7E,$F6,$78,$BE,$FE
	.dc.b	$4F,$15,$78,$6B,$FD,$B4,$D8,$F8,$71,$FC,$76,$3C,$D1,$E9,$9F,$A1
	.dc.b	$CD,$5F,$A5,$15,$F9,$3E,$47,$FB,$EB,$83,$FD,$F7,$E3,$FF,$8F,$CF
	.dc.b	$E3,$B3,$F8,$FC,$11,$FC,$84,$FF,$1E,$B8,$E3,$FE,$43,$1F,$E3,$87
	.dc.b	$F8,$DC,$F1,$8F,$E4,$53,$B9,$FE,$36,$DE,$6F,$43,$9E,$6B,$5F,$C8
	.dc.b	$AE,$DF,$47,$EA,$C7,$F2,$2A,$F2,$D7,$B5,$79,$1E,$CB,$FF,$6F,$D7
	.dc.b	$F2,$2B,$F2,$47,$F8,$F6,$3F,$91,$47,$48,$FE,$44,$D7,$F9,$13,$C1
	.dc.b	$FE,$3D,$5E,$B7,$F8,$F4,$7E,$2E,$30,$FD,$A9,$E2,$FF,$B8,$5F,$E4
	.dc.b	$4F,$E6,$9F,$F1,$F4,$7C,$B8,$FB,$0F,$D0,$E7,$F6,$4F,$D9,$9F,$D0
	.dc.b	$F3,$FF,$B7,$CF,$F6,$F1,$FC,$8D,$FC,$CB,$FC,$8C,$FF,$1E,$9F,$C8
	.dc.b	$9C,$BF,$C8,$9A,$FF,$22,$6B,$F5,$0F,$E4,$4E,$8C,$3F,$C8,$9F,$1B
	.dc.b	$F4,$4C,$3E,$63,$F2,$1E,$3F,$E5,$79,$7F,$6A,$FE,$5F,$96,$CF,$EC
	.dc.b	$E7,$F6,$B3,$FB,$39,$B3,$FB,$89,$FD,$C4,$FF,$27,$8F,$E4,$2E,$5C
	.dc.b	$9F,$C8,$43,$FC,$AD,$D9,$FE,$3E,$32,$7F,$2D,$8F,$D4,$46,$63,$31
	.dc.b	$EA,$7F,$C9,$07,$89,$FA,$91,$88,$FE,$19,$E5,$FD,$77,$C3,$E4,$7F
	.dc.b	$B7,$B9,$3E,$0B,$FC,$A2,$3F,$9E,$52,$3F,$94,$C7,$09,$FF,$16,$1F
	.dc.b	$E3,$87,$0F,$F8,$F2,$78,$3F,$E3,$1E,$6F,$8B,$5F,$CB,$2B,$AB,$97
	.dc.b	$32,$23,$FC,$86,$B9,$C7,$78,$F9,$B3,$B8,$F8,$AF,$F2,$F8,$CC,$7D
	.dc.b	$E8,$E1,$1E,$E3,$11,$FC,$A9,$F4,$BF,$E4,$27,$8B,$FC,$8A,$7C,$2F
	.dc.b	$F2,$A8,$DB,$FC,$A9,$F9,$2F,$16,$BF,$E4,$66,$BF,$E4,$67,$E0,$B9
	.dc.b	$77,$C9,$FC,$9E,$7E,$4F,$2F,$B2,$B8,$57,$C0,$8F,$65,$7E,$6B,$C8
	.dc.b	$FB,$FE,$37,$FB,$8F,$97,$F3,$6F,$F2,$8B,$FC,$7B,$FE,$E4,$C4,$FF
	.dc.b	$B9,$1B,$C7,$FE,$E4,$6B,$FE,$E2,$E5,$FC,$EC,$FF,$92,$67,$33,$FC
	.dc.b	$86,$BA,$BF,$C9,$5D,$BF,$E4,$67,$37,$F9,$AA,$C4,$79,$23,$F6,$71
	.dc.b	$CE,$3F,$53,$E4,$7E,$6B,$C8,$F6,$DF,$FB,$FE,$7F,$5B,$5F,$CA,$5F
	.dc.b	$0C,$7F,$26,$7F,$33,$1F,$CA,$9D,$C6,$5C,$BF,$E4,$7C,$3D,$6F,$E3
	.dc.b	$E7,$FC,$95,$3D,$23,$EA,$47,$F2,$88,$E3,$1E,$B3,$F9,$7B,$C5,$F5
	.dc.b	$F2,$FB,$23,$F8,$7F,$8F,$FB,$4F,$23,$FC,$8D,$CD,$FC,$41,$FF,$24
	.dc.b	$75,$58,$FF,$73,$55,$7F,$DC,$E7,$F2,$37,$F8,$4C,$E9,$D4,$65,$CB
	.dc.b	$FC,$AD,$FE,$27,$1C,$A3,$FD,$C1,$92,$35,$3A,$9E,$4F,$F0,$28,$F0
	.dc.b	$BF,$95,$72,$E6,$3F,$95,$C7,$F2,$E9,$E9,$3F,$CB,$B9,$3F,$89,$E1
	.dc.b	$FE,$4F,$8F,$E1,$AF,$F2,$B7,$F9,$3B,$FC,$BD,$FE,$66,$FF,$2B,$7F
	.dc.b	$93,$BF,$CA,$DC,$9F,$E4,$B9,$47,$F2,$B8,$D9,$FE,$E8,$8C,$BF,$CA
	.dc.b	$CF,$E1,$91,$FC,$BE,$33,$CD,$FC,$BA,$3F,$C9,$F3,$E2,$8F,$B3,$CF
	.dc.b	$FD,$23,$C8,$FE,$5B,$CD,$FC,$A6,$BF,$9F,$BF,$CA,$5F,$E7,$AF,$F2
	.dc.b	$8C,$31,$18,$8F,$E4,$51,$FC,$A9,$FE,$38,$FF,$94,$E3,$9C,$7F,$BA
	.dc.b	$23,$F9,$44,$7F,$34,$8F,$F2,$84,$7F,$BA,$A3,$78,$FF,$46,$9F,$F7
	.dc.b	$34,$FF,$23,$9F,$29,$FD,$82,$3C,$07,$F1,$97,$C8,$FF,$90,$9A,$FF
	.dc.b	$2D,$E6,$FE,$C2,$FF,$93,$A3,$C4,$FF,$91,$9D,$BF,$C7,$67,$FD,$DB
	.dc.b	$1A,$8F,$E7,$9C,$DF,$CE,$DE,$91,$5C,$B6,$BD,$51,$FC,$82,$3E,$04
	.dc.b	$7A,$67,$FB,$CB,$FD,$79,$FE,$BF,$E3,$FF,$21,$9F,$EE,$2F,$F2,$1B
	.dc.b	$3F,$C8,$4F,$E4,$07,$F9,$17,$97,$FB,$CB,$FE,$47,$E6,$BC,$63,$F8
	.dc.b	$E5,$FF,$81,$A7,$11,$FE,$E2,$BE,$AE,$A7,$4F,$F2,$18,$CD,$FF,$7A
	.dc.b	$AF,$E4,$31,$E4,$8F,$F0,$9C,$7F,$21,$AF,$EE,$3E,$47,$B2,$FF,$90
	.dc.b	$47,$F7,$49,$FE,$41,$7E,$09,$FE,$38,$D7,$FC,$8A,$D7,$D5,$3F,$E1
	.dc.b	$A9,$FF,$32,$72,$7B,$A7,$FD,$CD,$5C,$63,$FB,$84,$FB,$A7,$FA,$A3
	.dc.b	$FC,$86,$BF,$C3,$2F,$F9,$22,$BE,$0D,$7B,$6B,$D6,$FB,$2B,$FC,$5F
	.dc.b	$E3,$FF,$85,$3C,$8F,$E4,$31,$FE,$26,$7F,$90,$B7,$97,$F8,$E3,$FE
	.dc.b	$12,$8F,$E3,$11,$FC,$87,$04,$6A,$75,$3F,$C7,$79,$BF,$C6,$73,$C2
	.dc.b	$6C,$FF,$92,$1D,$A3,$FB,$84,$66,$32,$62,$3F,$C9,$95,$E0,$7F,$C9
	.dc.b	$33,$FC,$7C,$FF,$17,$DF,$F1,$DA,$FF,$1E,$BF,$C7,$5F,$F1,$5B,$FC
	.dc.b	$75,$FF,$22,$BF,$C7,$68,$FF,$1D,$8C,$9F,$C7,$5F,$E3,$91,$A8,$FE
	.dc.b	$39,$1F,$EE,$EE,$4F,$F0,$43,$FC,$6C,$FE,$3D,$1F,$E2,$97,$F8,$DB
	.dc.b	$E0,$8F,$A7,$C9,$FE,$62,$F2,$3F,$C8,$F9,$FF,$92,$3C,$8F,$E3,$97
	.dc.b	$FE,$42,$7F,$8E,$62,$7F,$8E,$1F,$C6,$79,$7F,$CC,$AF,$F1,$99,$FF
	.dc.b	$2E,$BF,$E4,$68,$FF,$32,$BF,$C6,$A7,$97,$2F,$F1,$D8,$FE,$35,$5C
	.dc.b	$23,$F8,$C4,$E9,$FF,$30,$CF,$F1,$B9,$D1,$8A,$FF,$3D,$C7,$A6,$7F
	.dc.b	$C7,$95,$C5,$FF,$1A,$57,$57,$FC,$79,$3F,$C6,$E3,$11,$FE,$33,$6B
	.dc.b	$FC,$6E,$3F,$8F,$46,$B9,$3F,$CA,$91,$FC,$6A,$3D,$B3,$FC,$69,$BC
	.dc.b	$BE,$88,$FE,$35,$1F,$12,$3C,$F5,$FE,$92,$F2,$3F,$CE,$FC,$5F,$C6
	.dc.b	$6B,$FC,$E2,$FF,$19,$3F,$C7,$07,$F8,$DC,$FE,$30,$FF,$A5,$1F,$E3
	.dc.b	$0D,$7F,$8A,$F2,$FF,$9A,$5F,$DD,$73,$57,$F8,$B4,$FF,$A5,$9F,$F1
	.dc.b	$F3,$99,$FF,$47,$C7,$F1,$60,$FF,$AA,$1F,$E2,$D1,$FE,$9A,$8F,$F4
	.dc.b	$7C,$FA,$A7,$F8,$BC,$FF,$A7,$9F,$E2,$E6,$1C,$3F,$C4,$9F,$F5,$8B
	.dc.b	$FE,$F7,$3F,$D5,$66,$5C,$BF,$C4,$DC,$BF,$E3,$3C,$FF,$C7,$91,$EB
	.dc.b	$8F,$E7,$1A,$FF,$9A,$DF,$E2,$C6,$23,$0F,$AE,$3D,$91,$FE,$C4,$F1
	.dc.b	$BF,$D5,$3E,$47,$F1,$68,$FF,$51,$BF,$C5,$9B,$CB,$FC,$47,$04,$FF
	.dc.b	$16,$87,$37,$FB,$22,$3F,$8B,$57,$FB,$01,$FF,$70,$3F,$EC,$67,$F8
	.dc.b	$B3,$B9,$FE,$25,$5D,$1F,$F1,$8C,$7F,$B3,$A7,$F8,$B3,$CE,$3F,$8D
	.dc.b	$3F,$E8,$98,$FF,$1A,$B9,$3F,$D9,$2F,$F8,$D4,$FF,$62,$F9,$1F,$EC
	.dc.b	$6C,$3F,$8C,$59,$FE,$31,$5F,$ED,$67,$F8,$C3,$FE,$C4,$7F,$8C,$1F
	.dc.b	$C5,$DC,$1F,$E3,$37,$F8,$B9,$FE,$DF,$8E,$91,$FC,$50,$FF,$75,$C7
	.dc.b	$F8,$C4,$C3,$83,$FD,$B0,$7F,$BA,$97,$2F,$D5,$E5,$FF,$7C,$F9,$1F
	.dc.b	$ED,$EF,$23,$F8,$BD,$FF,$BB,$1F,$E2,$E7,$F8,$C4,$FE,$2C,$FF,$BF
	.dc.b	$9F,$E2,$D8,$1F,$E2,$D1,$FE,$FE,$7F,$8B,$1F,$ED,$57,$1C,$9F,$EC
	.dc.b	$F7,$FD,$A9,$1F,$E3,$23,$0E,$23,$FD,$9E,$FF,$BA,$A3,$FC,$66,$61
	.dc.b	$C0,$3F,$DF,$AF,$F8,$D3,$93,$CD,$3F,$E2,$FA,$F3,$3F,$EF,$08,$F1
	.dc.b	$BF,$C7,$2B,$E0,$3F,$E3,$39,$DB,$FE,$D7,$8C,$BF,$C7,$5A,$FF,$B3
	.dc.b	$23,$73,$FE,$D1,$AE,$15,$E9,$AF,$F2,$33,$F0,$B9,$3C,$F5,$E9,$8E
	.dc.b	$11,$5F,$F7,$C7,$91,$5F,$F7,$D7,$05,$66,$2F,$1B,$F9,$04,$23,$1C
	.dc.b	$7F,$C0,$F9,$3A,$46,$23,$FC,$29,$CB,$D6,$31,$1F,$C6,$AB,$C9,$18
	.dc.b	$8F,$E3,$31,$BB,$FE,$4E,$F4,$7F,$DB,$87,$F2,$67,$CA,$FF,$2A,$7D
	.dc.b	$07,$95,$EC,$F3,$8E,$AE,$A7,$D3,$7C,$23,$EF,$47,$BE,$F1,$18,$BF
	.dc.b	$E5,$72,$F2,$33,$39,$C3,$F9,$5C,$FA,$DF,$E3,$51,$97,$FC,$83,$5F
	.dc.b	$ED,$F8,$C4,$7F,$95,$9F,$33,$FE,$58,$E6,$F6,$BF,$CA,$E3,$2F,$FB
	.dc.b	$69,$F8,$8F,$17,$6E,$5C,$BE,$AA,$FA,$F3,$6B,$FC,$9F,$5F,$06,$7E
	.dc.b	$F4,$FB,$1F,$C0,$45,$E6,$FE,$5C,$FE,$05,$FE,$5C,$D7,$F9,$71,$FC
	.dc.b	$B7,$3F,$C0,$3F,$CB,$27,$13,$88,$FF,$6D,$BC,$5F,$F6,$DC,$7F,$2E
	.dc.b	$8D,$47,$F3,$67,$EC,$3F,$CB,$9F,$CA,$BF,$CA,$A3,$11,$89,$F4,$3F
	.dc.b	$5B,$C7,$FB,$DC,$9F,$A1,$7F,$01,$C9,$67,$F9,$48,$8F,$E5,$AD,$3F
	.dc.b	$95,$14,$FD,$54,$7E,$9E,$33,$19,$3F,$C7,$6F,$FB,$7D,$C4,$7E,$3E
	.dc.b	$3F,$C8,$91,$4F,$F2,$09,$FC,$3D,$FE,$5E,$FE,$D5,$CB,$99,$FC,$8B
	.dc.b	$E7,$9B,$E5,$E2,$3D,$5E,$47,$F3,$0B,$3F,$EE,$16,$F2,$7F,$B6,$DA
	.dc.b	$FF,$2B,$9F,$14,$FF,$91,$E7,$32,$1D,$9F,$06,$7F,$DC,$33,$A8,$FD
	.dc.b	$9B,$C2,$AC,$F1,$AF,$E6,$33,$E0,$9F,$89,$3F,$E4,$98,$F2,$C7,$EC
	.dc.b	$9F,$4D,$7D,$B9,$F8,$31,$E4,$AF,$F1,$F5,$7C,$07,$FD,$C5,$19,$7F
	.dc.b	$92,$B7,$93,$E6,$57,$FB,$8A,$39,$47,$AE,$79,$D7,$F2,$99,$F8,$33
	.dc.b	$EA,$9F,$E6,$D3,$DD,$FF,$71,$4F,$CC,$E5,$F4,$5F,$F9,$3E,$BE,$34
	.dc.b	$7A,$6F,$F2,$9E,$47,$AB,$C6,$FE,$5B,$81,$FF,$17,$71,$B9,$BF,$31
	.dc.b	$1F,$E3,$1E,$6F,$C2,$3F,$CC,$72,$CD,$7F,$30,$72,$FF,$27,$72,$FF
	.dc.b	$93,$1F,$CF,$46,$5C,$CF,$E5,$63,$F9,$4C,$74,$8F,$A5,$1F,$EE,$FA
	.dc.b	$B1,$FA,$D8,$FF,$27,$DB,$C8,$FA,$BE,$47,$B7,$1F,$F2,$7C,$FB,$5F
	.dc.b	$E4,$4D,$7F,$DB,$47,$A0,$FF,$6E,$73,$57,$F8,$F6,$5F,$4F,$0F,$F2
	.dc.b	$83,$63,$F9,$33,$EE,$7F,$91,$3F,$AB,$8C,$BF,$0A,$7D,$D3,$FC,$8A
	.dc.b	$3F,$75,$3F,$CE,$1B,$1E,$D9,$F7,$47,$F0,$BF,$1F,$F2,$3E,$47,$F2
	.dc.b	$4C,$0F,$F2,$46,$F4,$7F,$B8,$5A,$FF,$B8,$6B,$F8,$94,$FF,$94,$27
	.dc.b	$33,$FE,$DF,$E7,$FE,$19,$3F,$C8,$EB,$F2,$4F,$17,$F8,$23,$FC,$A9
	.dc.b	$FE,$30,$FF,$94,$2B,$BC,$7F,$0F,$8F,$F2,$7C,$79,$23,$F5,$91,$FC
	.dc.b	$B5,$D1,$FC,$9F,$93,$F5,$B8,$7F,$B9,$E8,$FF,$94,$1B,$3F,$E4,$D6
	.dc.b	$BF,$C9,$0F,$E5,$D1,$FC,$A2,$33,$1F,$E2,$BE,$6F,$E5,$13,$FE,$13
	.dc.b	$E7,$FE,$0E,$F1,$9F,$E4,$0F,$F2,$67,$67,$F2,$6A,$F0,$CF,$F0,$59
	.dc.b	$FE,$4B,$1E,$68,$FA,$D5,$FD,$07,$C8,$FE,$0F,$CF,$FC,$8C,$FE,$44
	.dc.b	$7F,$13,$26,$1F,$D1,$5F,$F2,$9B,$FC,$05,$FE,$44,$7F,$94,$9D,$3C
	.dc.b	$1F,$E6,$CF,$F2,$97,$83,$FE,$50,$72,$FF,$29,$C7,$FA,$13,$98,$CC
	.dc.b	$7F,$1F,$8F,$E5,$D1,$63,$FA,$D4,$7A,$8F,$41,$FC,$CF,$9B,$C6,$E3
	.dc.b	$9A,$BF,$C6,$B2,$DB,$FC,$4E,$FF,$AF,$3F,$C8,$4E,$0F,$99,$FF,$29
	.dc.b	$46,$23,$15,$E8,$9E,$51,$EE,$8F,$44,$7F,$BA,$A7,$F9,$2C,$FF,$5B
	.dc.b	$8F,$F7,$6B,$FD,$B6,$7F,$9C,$D7,$F3,$AF,$1F,$F9,$09,$FC,$83,$9B
	.dc.b	$FB,$B3,$FC,$7E,$AF,$17,$F9,$11,$FE,$CD,$1F,$EE,$E7,$31,$FC,$61
	.dc.b	$FE,$F5,$1F,$C8,$DF,$ED,$4F,$F1,$F7,$F9,$E3,$FE,$49,$7F,$BD,$3C
	.dc.b	$5E,$B1,$FE,$48,$72,$FF,$33,$7F,$BD,$46,$5E,$33,$FE,$0F,$F2,$3D
	.dc.b	$95,$FC,$82,$BF,$9D,$F1,$F9,$A3,$F8,$9E,$69,$FF,$22,$61,$E8,$9F
	.dc.b	$F7,$3C,$E5,$FE,$47,$1E,$A7,$F8,$DB,$E6,$7F,$8C,$3E,$C7,$F9,$3B
	.dc.b	$F1,$2B,$CF,$1F,$C7,$E3,$94,$7F,$6D,$AC,$47,$C6,$8F,$F1,$44,$7F
	.dc.b	$6E,$F1,$FF,$B7,$F9,$1F,$C7,$CF,$E3,$C7,$FB,$AF,$97,$FC,$2C,$FF
	.dc.b	$8F,$AF,$FB,$43,$FC,$74,$FF,$75,$BF,$E1,$D7,$F8,$E3,$60,$7F,$21
	.dc.b	$7F,$C5,$91,$FC,$77,$13,$FC,$74,$FF,$1F,$C7,$F8,$ED,$FE,$3A,$7F
	.dc.b	$BB,$9F,$F1,$73,$FC,$75,$FF,$09,$BE,$B7,$FD,$CF,$0A,$FE,$38,$7F
	.dc.b	$90,$9C,$47,$FA,$97,$87,$FC,$65,$3F,$C6,$CC,$46,$0F,$E2,$C7,$F9
	.dc.b	$35,$FE,$34,$D3,$FC,$76,$7F,$8F,$5F,$F5,$A4,$7F,$90,$DC,$C6,$4F
	.dc.b	$F7,$69,$BE,$2F,$F2,$D4,$62,$3F,$C8,$DE,$47,$F8,$8B,$3F,$F1,$1F
	.dc.b	$91,$FE,$3F,$3F,$C7,$C7,$F0,$0E,$2F,$F2,$53,$FC,$72,$BF,$C9,$CF
	.dc.b	$F8,$F4,$FE,$38,$FF,$87,$DF,$F1,$D9,$FC,$6D,$DC,$7F,$17,$74,$FF
	.dc.b	$1B,$7F,$C7,$F1,$97,$3C,$7F,$E3,$98,$FE,$37,$1F,$E6,$17,$FD,$DC
	.dc.b	$FF,$9B,$DF,$4C,$FF,$16,$E3,$FF,$3B,$47,$A7,$8E,$CF,$FB,$B6,$AB
	.dc.b	$FE,$ED,$D5,$5F,$E3,$B3,$F8,$C4,$F9,$A7,$F8,$DC,$7F,$BC,$E3,$D7
	.dc.b	$1F,$EE,$87,$55,$E5,$8C,$47,$BA,$3D,$57,$FE,$8B,$AF,$E3,$76,$F2
	.dc.b	$3F,$CA,$7C,$5F,$E3,$A3,$F8,$D6,$1F,$E4,$97,$F8,$D3,$5F,$E3,$4F
	.dc.b	$FA,$3D,$FE,$34,$E5,$FE,$35,$38,$8C,$47,$F1,$17,$FD,$08,$FF,$19
	.dc.b	$6B,$FE,$35,$7F,$CE,$2F,$F8,$FD,$FF,$46,$39,$73,$D9,$FE,$A5,$8F
	.dc.b	$F1,$CB,$FE,$6C,$E5,$F5,$C7,$F8,$E0,$FE,$33,$3E,$27,$F8,$6D,$62
	.dc.b	$31,$1F,$EA,$33,$FC,$6A,$F8,$1F,$E2,$EF,$BD,$FF,$6F,$3B,$7F,$8B
	.dc.b	$BF,$07,$5F,$F5,$34,$62,$3E,$44,$7F,$A0,$E3,$F8,$BC,$7F,$AE,$DF
	.dc.b	$F1,$FB,$4F,$F6,$07,$8D,$FE,$86,$F2,$3F,$C6,$C7,$F8,$D4,$FE,$31
	.dc.b	$1F,$EC,$37,$F8,$C3,$FE,$B1,$7F,$C6,$87,$F1,$87,$FD,$88,$FF,$18
	.dc.b	$6F,$2F,$F1,$83,$F8,$73,$FE,$87,$8F,$E3,$07,$F8,$CC,$FE,$2E,$E6
	.dc.b	$36,$F8,$5F,$F6,$B4,$62,$3C,$87,$FB,$67,$C8,$FF,$4F,$D7,$F8,$CC
	.dc.b	$FF,$19,$1F,$C5,$C4,$7F,$14,$3F,$D3,$0F,$F8,$C4,$C3,$83,$F8,$B9
	.dc.b	$4F,$F6,$BB,$87,$1C,$BF,$ED,$78,$FE,$2B,$1F,$EE,$57,$FC,$5E,$39
	.dc.b	$7E,$7C,$FF,$BB,$FC,$8F,$F5,$AF,$91,$FE,$D1,$3F,$DA,$07,$F1,$3B
	.dc.b	$FF,$79,$3F,$C4,$DF,$F7,$83,$FE,$CF,$3F,$89,$BF,$EE,$37,$FD,$9A
	.dc.b	$7F,$13,$77,$C9,$FC,$3D,$AF,$FB,$32,$31,$1A,$E5,$FF,$65,$47,$F1
	.dc.b	$28,$FF,$6A,$BF,$C4,$B7,$3E,$8A,$FE,$25,$3F,$EC,$18,$FE,$24,$D8
	.dc.b	$FE,$24,$D7,$F8,$94,$7F,$B2,$CF,$E2,$31,$E9,$8F,$F6,$5B,$5F,$F1
	.dc.b	$4B,$E8,$E4,$FB,$2C,$3A,$72,$D7,$E3,$CF,$F1,$47,$A4,$7B,$27,$51
	.dc.b	$BB,$F5,$CE,$23,$A4,$3C,$8F,$F7,$4F,$7F,$FB,$AB,$C7,$FF,$68,$9F
	.dc.b	$C6,$2B,$F1,$74,$8F,$E3,$31,$87,$FD,$A0,$FF,$B9,$5F,$E2,$CA,$38
	.dc.b	$71,$FF,$8B,$E3,$13,$FE,$33,$7C,$6E,$5C,$C7,$F1,$79,$E2,$FF,$1F
	.dc.b	$8F,$64,$FF,$1B,$77,$CB,$ED,$7C,$9C,$D5,$F3,$45,$7C,$9E,$47,$BE
	.dc.b	$FF,$DA,$77,$FC,$66,$7C,$31,$FE,$D2,$7C,$AF,$F1,$AC,$3C,$6F,$F1
	.dc.b	$66,$BF,$EC,$E6,$C7,$D2,$77,$CF,$F4,$5E,$31,$6B,$F9,$0B,$B7,$F8
	.dc.b	$F3,$F3,$6B,$F8,$DB,$4F,$64,$FD,$3A,$F6,$3F,$56,$2F,$3F,$FB,$4C
	.dc.b	$FF,$69,$1F,$C7,$9F,$C6,$C7,$F8,$DF,$3F,$C3,$4F,$FB,$44,$FE,$3D
	.dc.b	$1F,$C6,$DF,$E2,$EF,$F2,$19,$FB,$91,$FE,$33,$3F,$D9,$C7,$F1,$C7
	.dc.b	$F0,$EF,$FB,$38,$FF,$19,$3F,$05,$FE,$38,$F8,$9F,$A8,$BC,$7F,$B5
	.dc.b	$CB,$FA,$17,$EE,$72,$D9,$FC,$EC,$FE,$8E,$7F,$39,$36,$7F,$4D,$3F
	.dc.b	$A2,$9F,$D1,$CD,$9F,$F6,$71,$4F,$F1,$89,$4F,$E3,$47,$E9,$E3,$FC
	.dc.b	$72,$F8,$4C,$99,$3F,$DB,$47,$E4,$5F,$E4,$AF,$AC,$FF,$1B,$1E,$D9
	.dc.b	$F7,$C6,$21,$E3,$FC,$27,$E1,$3F,$E3,$7E,$3F,$84,$FF,$8E,$79,$6C
	.dc.b	$7F,$AE,$A6,$BF,$EC,$FA,$E6,$FF,$B1,$23,$E3,$57,$F2,$A7,$AC,$7F
	.dc.b	$B5,$DE,$2A,$BE,$04,$E6,$7A,$C7,$B6,$3F,$95,$D7,$68,$FD,$73,$E0
	.dc.b	$9B,$3F,$E3,$C8,$F6,$C7,$EC,$5F,$2D,$7F,$28,$8F,$1B,$FE,$36,$8C
	.dc.b	$A3,$FD,$B9,$1E,$67,$FD,$B1,$1A,$7F,$C7,$0D,$E3,$FC,$05,$7F,$BE
	.dc.b	$F9,$3D,$75,$CC,$4E,$1F,$E5,$55,$E4,$8F,$74,$7F,$2D,$BF,$8F,$1E
	.dc.b	$CA,$F5,$5F,$FB,$6A,$DE,$3F,$DC,$F1,$5C,$3F,$C9,$A5,$1F,$CA,$F1
	.dc.b	$FC,$F3,$FC,$A9,$B3,$FE,$42,$9D,$BF,$C9,$63,$FB,$DC,$62,$3F,$DB
	.dc.b	$8D,$7F,$DA,$11,$A9,$FE,$55,$3B,$E5,$FC,$A4,$7F,$B6,$E7,$BC,$7D
	.dc.b	$88,$FF,$21,$47,$E9,$DE,$6F,$EC,$5F,$E6,$0F,$EA,$FC,$7F,$CC,$F8
	.dc.b	$FE,$FC,$3D,$E7,$F2,$79,$F1,$38,$3F,$8E,$BE,$E7,$4E,$9F,$F6,$DB
	.dc.b	$FD,$83,$0F,$64,$7F,$B4,$9B,$C9,$E5,$8F,$F2,$4C,$7C,$18,$FD,$B6
	.dc.b	$1F,$E4,$CA,$ED,$1F,$02,$BF,$C8,$B1,$E4,$8F,$81,$1F,$93,$9F,$E5
	.dc.b	$B3,$F2,$4F,$CB,$78,$DF,$B6,$F1,$FF,$DB,$9C,$9F,$C4,$DF,$E5,$13
	.dc.b	$E7,$7F,$96,$3F,$59,$FF,$24,$BF,$C6,$DF,$E5,$51,$FC,$8A,$34,$FF
	.dc.b	$2D,$E6,$FE,$2D,$3F,$E4,$BC,$8E,$6B,$35,$FC,$B6,$72,$FF,$B9,$63
	.dc.b	$2F,$EC,$A7,$11,$C0,$C5,$7E,$C6,$3F,$88,$F1,$FF,$12,$C3,$FC,$09
	.dc.b	$C9,$FC,$BD,$FE,$57,$7F,$4D,$FE,$5E,$E4,$FF,$24,$9F,$CD,$DF,$E5
	.dc.b	$EF,$F3,$37,$2E,$4F,$F2,$51,$4F,$E4,$E6,$CF,$F6,$FE,$28,$FE,$67
	.dc.b	$1F,$CE,$5F,$F2,$7C,$66,$3F,$DC,$0E,$E7,$EB,$F2,$FF,$0D,$F1,$FF
	.dc.b	$96,$F3,$7F,$29,$BF,$C6,$3F,$CA,$5A,$FF,$29,$6B,$FC,$A4,$C3,$87
	.dc.b	$FD,$B8,$FF,$53,$7F,$96,$3F,$D4,$5F,$F2,$95,$E2,$31,$1C,$1A,$FF
	.dc.b	$26,$30,$E8,$C6,$3F,$C9,$63,$D2,$7F,$94,$23,$F8,$0B,$FE,$50,$6B
	.dc.b	$FC,$A5,$FE,$D0,$FF,$1D,$3F,$8F,$73,$78,$1F,$E5,$31,$E9,$7F,$DC
	.dc.b	$15,$C6,$39,$C7,$F9,$4A,$3F,$94,$47,$F6,$18,$E6,$D7,$F9,$24,$62
	.dc.b	$35,$CD,$E9,$AF,$E7,$CF,$F5,$E7,$FA,$FF,$8D,$FC,$0F,$87,$FD,$AD
	.dc.b	$68,$FE,$48,$D8,$FF,$6F,$35,$FE,$15,$D7,$FD,$FE,$7F,$C9,$B3,$99
	.dc.b	$FE,$33,$87,$F2,$69,$FE,$50,$D8,$FE,$2B,$5F,$DE,$E7,$FD,$CD,$3F
	.dc.b	$E0,$D7,$F9,$25,$78,$79,$BF,$AD,$C7,$49,$D4,$7F,$59,$8F,$E4,$71
	.dc.b	$FE,$11,$F1,$FD,$B8,$7B,$4F,$E8,$31,$FC,$6E,$7F,$C9,$97,$E7,$9F
	.dc.b	$E4,$F3,$E1,$7F,$DC,$B1,$97,$FC,$31,$3F,$EE,$89,$FF,$12,$3F,$CA
	.dc.b	$1E,$91,$FC,$B1,$AF,$F1,$E9,$F9,$1C,$BF,$E1,$3A,$FF,$25,$D7,$2A
	.dc.b	$FE,$3F,$5F,$3E,$3D,$F5,$ED,$7F,$BB,$78,$DF,$E1,$6F,$1F,$F9,$0C
	.dc.b	$7F,$35,$7F,$90,$BF,$D5,$1F,$E4,$2F,$F8,$E5,$FE,$43,$47,$F9,$0F
	.dc.b	$37,$F8,$CD,$FF,$75,$46,$63,$FC,$9D,$39,$7F,$DD,$5C,$DF,$E3,$1A
	.dc.b	$CC,$E6,$7F,$C9,$35,$C5,$FF,$08,$CF,$F2,$09,$FF,$22,$3F,$CB,$DB
	.dc.b	$5F,$E1,$18,$FF,$18,$DF,$F2,$09,$FE,$A0,$FF,$20,$7F,$B7,$BF,$C8
	.dc.b	$1C,$9F,$C8,$2D,$1F,$C8,$23,$FC,$AC,$FF,$20,$8C,$C7,$F2,$06,$BF
	.dc.b	$C8,$0D,$9F,$C8,$23,$F8,$FB,$FE,$5B,$7F,$8F,$9B,$71,$C9,$F5,$39
	.dc.b	$3F,$C5,$5E,$3F,$F9,$2B,$3F,$F2,$5F,$8F,$FC,$7E,$FF,$C0,$EF,$F1
	.dc.b	$F6,$F2,$7F,$1C,$7F,$CD,$D1,$FE,$EF,$C8,$E6,$33,$3F,$C2,$F9,$7F
	.dc.b	$CD,$53,$FC,$76,$7F,$BB,$3F,$C8,$0F,$F2,$33,$89,$FE,$3A,$D7,$FC
	.dc.b	$8B,$18,$8D,$18,$9F,$F2,$24,$7A,$A7,$F9,$04,$F8,$9F,$F2,$24,$62
	.dc.b	$31,$1F,$C4,$5F,$6B,$FE,$EC,$AF,$5B,$FC,$7A,$7F,$8D,$CF,$FA,$1D
	.dc.b	$FE,$4D,$C9,$FE,$7D,$7F,$DE,$71,$CA,$3F,$CF,$51,$FC,$A2,$3E,$0C
	.dc.b	$FF,$92,$1C,$F2,$FA,$2B,$FC,$B9,$E3,$FF,$9F,$38,$BF,$81,$0E,$5F
	.dc.b	$F4,$D3,$FC,$78,$FF,$20,$65,$FE,$43,$C3,$FD,$41,$1F,$C7,$0F,$E4
	.dc.b	$DC,$DF,$E9,$97,$F8,$DC,$69,$FE,$36,$E5,$FF,$27,$D7,$FA,$45,$FE
	.dc.b	$36,$7F,$B9,$6B,$FC,$81,$3D,$A7,$FC,$E5,$3F,$C6,$A3,$FD,$5D,$1F
	.dc.b	$E9,$4A,$F5,$4F,$AA,$7F,$C6,$8F,$F1,$A3,$0E,$1F,$E3,$27,$F2,$17
	.dc.b	$FD,$68,$FF,$19,$AC,$46,$23,$F8,$C3,$FE,$5C,$7F,$C7,$11,$FE,$42
	.dc.b	$8D,$47,$FA,$30,$FF,$35,$EB,$5F,$4B,$EA,$8F,$F4,$67,$8B,$FE,$AC
	.dc.b	$F1,$FF,$8B,$C7,$F9,$A1,$FE,$2E,$FF,$B3,$5F,$E2,$F6,$7F,$8B,$9F
	.dc.b	$C5,$A3,$FD,$98,$FF,$16,$3F,$C6,$87,$F8,$CD,$CF,$37,$F8,$CE,$3F
	.dc.b	$8B,$C7,$FB,$1D,$FF,$71,$3D,$E3,$FD,$BE,$D3,$FD,$4F,$18,$7C,$87
	.dc.b	$FA,$97,$C7,$FF,$62,$DF,$F1,$58,$FF,$40,$BF,$C5,$5F,$F7,$33,$FC
	.dc.b	$55,$C9,$FC,$55,$FE,$28,$FF,$9A,$DF,$E2,$87,$FB,$81,$FF,$74,$3F
	.dc.b	$ED,$C7,$89,$FE,$DC,$32,$7F,$11,$39,$19,$33,$C7,$FE,$C0,$8C,$47
	.dc.b	$48,$FF,$5F,$78,$FF,$ED,$EF,$1F,$F8,$A4,$FF,$BF,$5F,$E2,$8D,$E2
	.dc.b	$FE,$29,$9A,$7F,$8A,$1F,$C4,$DC,$47,$F1,$39,$FF,$80,$13,$FC,$4D
	.dc.b	$FF,$52,$BF,$E3,$23,$F8,$9B,$88,$FE,$20,$F7,$7F,$DA,$F1,$88,$F1
	.dc.b	$18,$E4,$FF,$6B,$47,$A6,$7F,$89,$D7,$89,$FF,$6B,$47,$F1,$38,$C4
	.dc.b	$7F,$13,$AE,$CF,$F0,$97,$C4,$E5,$CF,$19,$84,$38,$F1,$D7,$FD,$49
	.dc.b	$7F,$C5,$9F,$54,$65,$CD,$7F,$8D,$27,$FD,$D1,$3E,$EA,$F5,$BF,$EF
	.dc.b	$1F,$1F,$85,$F5,$C3,$F2,$49,$78,$DF,$C6,$B9,$3C,$91,$8E,$3F,$F0
	.dc.b	$F7,$26,$2B,$F8,$D9,$8B,$C4,$FF,$8C,$AB,$83,$FE,$D8,$9F,$E3,$71
	.dc.b	$88,$FE,$24,$EA,$FC,$B5,$E3,$AF,$F1,$A9,$FC,$6C,$FF,$1A,$3F,$EC
	.dc.b	$77,$FD,$AE,$DB,$E2,$FF,$BA,$3C,$7A,$FF,$BA,$A7,$F9,$AE,$1E,$63
	.dc.b	$FC,$A9,$CB,$5F,$F6,$C4,$D7,$F8,$8B,$5F,$A1,$8F,$A5,$4E,$2B,$FC
	.dc.b	$6D,$5D,$E7,$F8,$FB,$EB,$7F,$C8,$2E,$DE,$AD,$7D,$35,$F3,$EA,$D7
	.dc.b	$FB,$6A,$BD,$51,$EC,$9F,$AF,$17,$97,$F9,$0B,$F6,$DF,$E4,$31,$88
	.dc.b	$C4,$7F,$B5,$3B,$3F,$04,$FF,$21,$8F,$BE,$FF,$8D,$8F,$F6,$A1,$FC
	.dc.b	$85,$CB,$FE,$D3,$3F,$90,$BF,$11,$FE,$43,$38,$8C,$47,$96,$7C,$D1
	.dc.b	$F5,$3C,$6F,$BF,$D7,$FA,$17,$F0,$3C,$B6,$7F,$8F,$08,$FE,$44,$D3
	.dc.b	$F8,$F9,$FA,$B7,$F4,$91,$FA,$57,$31,$93,$F9,$01,$C4,$FD,$94,$FE
	.dc.b	$42,$33,$19,$8F,$C5,$C7,$F8,$DA,$3F,$68,$F0,$72,$27,$D9,$3F,$90
	.dc.b	$8F,$44,$5F,$1B,$11,$EB,$AF,$E4,$A7,$B7,$93,$FD,$74,$7F,$B2,$5C
	.dc.b	$3F,$E3,$16,$A3,$FD,$96,$E1,$C4,$79,$A3,$9F,$2F,$C3,$7F,$94,$47
	.dc.b	$86,$79,$09,$FE,$0B,$3D,$A6,$BA,$8F,$8C,$F0,$8B,$3C,$5A,$7C,$88
	.dc.b	$C3,$C0,$FC,$03,$EE,$8F,$F1,$73,$E9,$7F,$8C,$C7,$F2,$48,$C7,$27
	.dc.b	$AE,$7E,$2F,$25,$7D,$31,$ED,$7F,$C7,$B5,$95,$1E,$A7,$F9,$94,$7C
	.dc.b	$17,$2E,$5F,$F1,$F3,$CE,$3D,$51,$88,$F9,$51,$EA,$9F,$CD,$61,$F7
	.dc.b	$ED,$E2,$7F,$29,$BF,$53,$FE,$44,$E3,$FC,$F3,$FE,$C7,$E4,$FC,$5B
	.dc.b	$FC,$A6,$77,$C7,$FE,$DA,$BE,$6F,$FB,$8A,$36,$FF,$B6,$9F,$D0,$4E
	.dc.b	$5C,$C7,$F2,$49,$E2,$FF,$8F,$E3,$F5,$53,$FE,$48,$7F,$40,$FF,$2C
	.dc.b	$79,$D7,$ED,$63,$84,$7E,$B7,$C6,$F7,$78,$DE,$F9,$FE,$57,$3E,$8B
	.dc.b	$FE,$56,$D7,$F9,$5B,$5F,$E3,$EF,$A1,$FE,$4F,$87,$99,$CC,$65,$FE
	.dc.b	$4E,$FC,$87,$F9,$5C,$6E,$3F,$2F,$3F,$CB,$9F,$6C,$FF,$27,$9F,$93
	.dc.b	$C9,$FB,$AA,$FE,$55,$5F,$0A,$3D,$CF,$F1,$3F,$17,$ED,$78,$DF,$CA
	.dc.b	$AC,$FF,$92,$9B,$CB,$FC,$9E,$BF,$88,$4F,$FB,$7B,$97,$F8,$2B,$98
	.dc.b	$CC,$FF,$29,$6A,$3F,$DA,$CF,$F1,$D8,$D3,$A8,$FE,$29,$1F,$CA,$E3
	.dc.b	$F9,$33,$FE,$4E,$AF,$1D,$7F,$16,$8E,$73,$FA,$97,$FC,$9D,$3E,$A8
	.dc.b	$FE,$61,$C7,$F9,$7C,$3F,$C9,$D1,$FC,$B5,$FF,$27,$49,$FE,$62,$D8
	.dc.b	$FF,$0A,$72,$7F,$22,$9F,$E5,$B3,$99,$FE,$4C,$73,$3F,$C3,$BC,$BF
	.dc.b	$C3,$27,$F9,$8C,$EE,$3F,$98,$47,$09,$FE,$74,$FF,$2D,$78,$1F,$CB
	.dc.b	$4D,$C7,$D7,$AF,$E9,$5E,$37,$EC,$79,$BF,$95,$9F,$E4,$EC,$3F,$CA
	.dc.b	$37,$FC,$36,$3F,$89,$BF,$D1,$DF,$F7,$04,$7F,$13,$C7,$FA,$44,$7F
	.dc.b	$26,$8F,$E5,$33,$88,$FE,$40,$D7,$9B,$FD,$65,$CB,$B9,$FE,$93,$1F
	.dc.b	$CD,$23,$FA,$73,$C1,$FE,$96,$FA,$4F,$F7,$0C,$FF,$25,$7F,$91,$C7
	.dc.b	$9D,$FE,$8B,$CD,$C1,$FE,$47,$1E,$17,$FD,$BE,$FF,$29,$72,$E5,$FF
	.dc.b	$2A,$3C,$DE,$AE,$63,$CB,$18,$8E,$91,$EA,$8F,$E4,$51,$E8,$8F,$EF
	.dc.b	$51,$88,$FE,$CB,$1F,$C8,$F9,$BF,$92,$78,$BF,$C8,$E7,$FB,$B3,$FC
	.dc.b	$8D,$FE,$7C,$FF,$23,$7F,$C0,$4F,$F2,$37,$7C,$5F,$C8,$E3,$FD,$CB
	.dc.b	$1F,$DC,$9F,$F7,$2E,$07,$F9,$14,$72,$9F,$E3,$CF,$F8,$0E,$3F,$DD
	.dc.b	$2D,$8F,$F2,$4B,$FE,$0D,$8F,$E4,$4D,$E5,$FF,$0B,$47,$F2,$28,$FF
	.dc.b	$09,$78,$DE,$CA,$FF,$71,$C7,$14,$7F,$22,$7D,$2E,$0F,$E3,$6E,$DF
	.dc.b	$F7,$03,$5C,$B5,$FF,$26,$BE,$F7,$F9,$5B,$ED,$7F,$93,$BD,$9F,$F0
	.dc.b	$D4,$FF,$23,$BF,$89,$1F,$C8,$E3,$73,$F4,$23,$DB,$C7,$EE,$8F,$F1
	.dc.b	$A7,$8B,$FD,$6F,$C6,$FE,$44,$7F,$21,$3F,$8C,$72,$C3,$F8,$D4,$FF
	.dc.b	$8E,$63,$F9,$13,$63,$47,$F8,$CE,$3F,$CA,$11,$FE,$2A,$72,$E6,$3F
	.dc.b	$8F,$3F,$E3,$B8,$EC,$FF,$85,$1C,$B9,$8F,$E3,$EF,$F9,$16,$3F,$96
	.dc.b	$72,$FF,$8E,$1C,$C6,$63,$FC,$A1,$1F,$DD,$EB,$F9,$0E,$47,$F9,$09
	.dc.b	$FC,$80,$FF,$25,$D7,$F9,$55,$FE,$3E,$7F,$22,$7F,$C8,$6F,$FB,$B5
	.dc.b	$D1,$FC,$6E,$3A,$3F,$C6,$CD,$9F,$EE,$D3,$FC,$B1,$19,$73,$C5,$FE
	.dc.b	$50,$8F,$F7,$6C,$78,$63,$FC,$D1,$E3,$7F,$83,$F2,$FF,$08,$78,$DF
	.dc.b	$E4,$A3,$F9,$01,$FC,$26,$BF,$CD,$8F,$F1,$D6,$F1,$7F,$BA,$71,$4F
	.dc.b	$F1,$E9,$CC,$FF,$1A,$7F,$C6,$51,$FE,$F1,$7F,$CC,$2F,$FB,$B9,$FF
	.dc.b	$3A,$3F,$E4,$57,$FC,$92,$F1,$7B,$F1,$FF,$9A,$A3,$F9,$14,$6A,$3F
	.dc.b	$D0,$51,$EA,$9F,$E4,$2F,$F2,$1E,$2F,$1C,$7F,$8F,$AA,$BF,$C5,$6F
	.dc.b	$D2,$FF,$20,$8F,$F4,$15,$7F,$20,$7D,$31,$C5,$F4,$BF,$E4,$29,$E2
	.dc.b	$FF,$90,$A3,$FD,$11,$3F,$E4,$67,$DB,$19,$72,$FF,$1F,$7E,$2F,$1F
	.dc.b	$AA,$FF,$D3,$9E,$37,$F9,$0F,$87,$F9,$05,$7F,$A6,$5F,$E4,$0F,$FA
	.dc.b	$5D,$FE,$40,$FF,$97,$DF,$E4,$0E,$DF,$F2,$11,$FC,$81,$FF,$53,$3F
	.dc.b	$C8,$0F,$E3,$EF,$17,$F8,$FB,$FE,$96,$7F,$8F,$B5,$FE,$3F,$8E,$23
	.dc.b	$5C,$9F,$EB,$39,$FE,$3D,$3F,$C9,$27,$FD,$59,$D9,$EB,$9F,$E3,$F3
	.dc.b	$FE,$EA,$D7,$C7,$1F,$C7,$4F,$E3,$B1,$5F,$E4,$11,$E1,$7F,$8C,$D7
	.dc.b	$C1,$7F,$90,$BE,$F7,$F9,$0B,$5F,$E4,$2F,$17,$FD,$6C,$7F,$1D,$36
	.dc.b	$F3,$8F,$F5,$BC,$7F,$8F,$A7,$F8,$FC,$FC,$D3,$FD,$A1,$E2,$7F,$9D
	.dc.b	$FC,$6F,$F1,$F1,$FE,$3D,$3F,$8B,$F2,$FF,$1C,$9F,$F4,$94,$7F,$1C
	.dc.b	$6C,$7F,$18,$7F,$DB,$11,$FC,$71,$B1,$FC,$61,$FF,$63,$C7,$F8,$F1
	.dc.b	$FF,$6D,$39,$73,$1F,$C6,$1F,$F5,$8C,$7F,$8F,$4C,$38,$8F,$F7,$44
	.dc.b	$7F,$8F,$0F,$F4,$87,$8D,$FE,$91,$AF,$E3,$67,$F8,$E8,$FF,$1C,$9F
	.dc.b	$C6,$AB,$FD,$58,$FF,$8E,$0F,$E3,$4F,$FB,$C5,$FE,$34,$E8,$FE,$34
	.dc.b	$FF,$8E,$9F,$F6,$2B,$FE,$37,$7F,$DE,$8F,$F1,$97,$91,$FE,$37,$37
	.dc.b	$D5,$F4,$39,$3F,$E0,$07,$8D,$FE,$A7,$F1,$BF,$DC,$27,$F8,$DC,$FE
	.dc.b	$31,$5F,$EC,$17,$F8,$C3,$78,$FF,$8C,$47,$F1,$78,$C4,$7F,$0F,$5A
	.dc.b	$27,$F8,$C4,$0D,$CF,$FB,$FE,$3F,$8C,$46,$48,$72,$7A,$3C,$67,$A4
	.dc.b	$FF,$8E,$63,$C8,$70,$24,$78,$79,$71,$18,$AF,$81,$5F,$E3,$E7,$F8
	.dc.b	$F4,$79,$23,$FD,$F9,$C9,$B7,$F8,$D4,$78,$DF,$E3,$CD,$7F,$DF,$3C
	.dc.b	$9D,$B0,$F2,$C7,$F8,$FA,$74,$FF,$17,$8C,$D7,$C1,$8F,$F1,$F4,$FC
	.dc.b	$D8,$F3,$BF,$3E,$AB,$F0,$6F,$F9,$13,$DD,$FE,$44,$FF,$BD,$FC,$6F
	.dc.b	$F7,$17,$57,$FB,$8F,$C5,$FE,$43,$91,$FE,$43,$5F,$59,$FE,$42,$FC
	.dc.b	$D7,$F9,$0C,$6F,$8F,$FD,$C5,$1F,$C8,$67,$EB,$4E,$67,$F8,$FC,$62
	.dc.b	$7F,$91,$BC,$E3,$2E,$5F,$F6,$F4,$F5,$E6,$F6,$CF,$8E,$7F,$DB,$F3
	.dc.b	$E7,$9C,$CF,$9F,$9A,$CB,$C6,$F7,$5F,$F2,$4C,$3F,$90,$3F,$ED,$E9
	.dc.b	$F3,$3F,$C8,$1A,$FF,$B7,$9A,$FD,$98,$FF,$24,$5F,$9E,$3F,$91,$CF
	.dc.b	$2E,$7F,$AA,$F4,$9B,$5F,$E3,$F3,$F9,$1B,$EB,$9F,$6C,$FC,$B8,$B3
	.dc.b	$FC,$B1,$E2,$7B,$39,$FF,$DC,$07,$FB,$7C,$FE,$46,$7E,$70,$FE,$44
	.dc.b	$FE,$8D,$FF,$6F,$3F,$B0,$7F,$91,$3F,$59,$FE,$4C,$FE,$B2,$72,$E6
	.dc.b	$3F,$C9,$6F,$B1,$FF,$1E,$BC,$DF,$F1,$EC,$62,$30,$7B,$23,$FD,$BA
	.dc.b	$7D,$65,$E2,$FE,$63,$97,$F7,$6F,$E6,$B9,$6C,$FE,$D2,$7F,$71,$3F
	.dc.b	$A0,$9B,$3F,$BB,$9F,$DC,$CF,$F2,$98,$FE,$0C,$E5,$C9,$FC,$22,$7F
	.dc.b	$53,$1F,$CA,$A3,$67,$F2,$77,$F9,$53,$A3,$FC,$84,$7F,$15,$7F,$95
	.dc.b	$3E,$68,$FC,$8C,$7F,$90,$A3,$D7,$1F,$98,$9D,$C7,$AA,$FF,$D9,$1C
	.dc.b	$7F,$01,$FF,$65,$72,$58,$10,$FF,$1F,$61,$FC,$A1,$FE,$59,$5F,$2A
	.dc.b	$B5,$1F,$E3,$B9,$D3,$FC,$A2,$7E,$1C,$EA,$3E,$0B,$FC,$C2,$33,$1F
	.dc.b	$CA,$A7,$8D,$7F,$0C,$79,$CD,$AF,$2C,$FE,$D2,$7C,$93,$67,$F9,$83
	.dc.b	$E0,$3E,$2B,$FE,$E7,$E4,$F7,$5F,$F9,$06,$BC,$EF,$F9,$06,$BE,$75
	.dc.b	$79,$E3,$FD,$B7,$3A,$7F,$8E,$CF,$C1,$7E,$0F,$27,$E0,$AB,$F9,$8C
	.dc.b	$E6,$7F,$94,$B6,$FD,$51,$88,$E5,$1E,$98,$FE,$4F,$1F,$36,$3D,$35
	.dc.b	$F7,$AF,$F3,$43,$C3,$FF,$24,$B7,$7F,$F7,$23,$5F,$F1,$87,$17,$E6
	.dc.b	$A7,$FC,$93,$39,$9F,$E4,$2D,$5C,$5F,$9C,$9E,$38,$D9,$FF,$22,$3F
	.dc.b	$93,$8F,$F6,$C5,$FD,$77,$31,$99,$FF,$74,$DF,$96,$FE,$C5,$7F,$92
	.dc.b	$63,$D5,$1F,$9A,$F1,$7D,$BE,$2F,$BA,$BF,$93,$DF,$FB,$A3,$0F,$D6
	.dc.b	$C6,$38,$FF,$72,$FF,$BB,$A3,$F7,$AF,$F2,$E8,$F8,$13,$FE,$4F,$9A
	.dc.b	$FF,$94,$1A,$F5,$6C,$7E,$DA,$7F,$CA,$71,$E0,$9F,$F2,$3E,$3F,$36
	.dc.b	$3E,$04,$7F,$27,$7B,$9E,$EA,$FD,$AD,$7D,$8F,$13,$EC,$F8,$BF,$CA
	.dc.b	$39,$3F,$88,$BF,$CA,$B9,$EC,$7F,$93,$B0,$AF,$FB,$AE,$3F,$89,$CF
	.dc.b	$F2,$CC,$8F,$F2,$8A,$FE,$3A,$FF,$2C,$7F,$89,$BF,$E4,$D8,$DD,$7F
	.dc.b	$2C,$9F,$E3,$33,$FC,$B2,$73,$3F,$AD,$8F,$F2,$6C,$EA,$3F,$87,$C2
	.dc.b	$AC,$FF,$12,$E3,$F7,$3F,$CA,$4F,$E4,$78,$7F,$29,$69,$FC,$94,$A7
	.dc.b	$F2,$53,$F9,$9B,$97,$27,$F8,$D7,$97,$F9,$7C,$7F,$29,$8F,$E7,$6F
	.dc.b	$17,$F9,$BB,$FC,$A9,$A7,$F2,$AC,$51,$FC,$92,$38,$9F,$E5,$33,$AF
	.dc.b	$2F,$F2,$C8,$C4,$7F,$3E,$F1,$7F,$63,$CB,$FC,$AA,$3F,$9F,$BF,$CA
	.dc.b	$9A,$FF,$09,$7F,$A6,$BF,$E5,$67,$F9,$D3,$97,$2F,$F2,$37,$FA,$AB
	.dc.b	$FC,$9D,$FE,$9A,$E5,$CB,$FC,$9C,$FE,$5C,$F2,$7F,$92,$1F,$C9,$9F
	.dc.b	$E5,$4F,$FB,$B1,$E7,$8F,$F4,$88,$F5,$9F,$C7,$F9,$79,$3F,$C7,$27
	.dc.b	$D6,$FF,$BA,$71,$FF,$29,$C6,$23,$FC,$A6,$FF,$66,$7F,$CA,$6F,$99
	.dc.b	$FF,$27,$46,$23,$13,$E4,$9F,$E4,$53,$FC,$C6,$79,$47,$F2,$28,$FF
	.dc.b	$2B,$47,$CE,$E5,$FE,$BF,$5F,$C8,$9F,$E4,$7E,$27,$F2,$23,$FC,$93
	.dc.b	$47,$FC,$93,$5F,$E0,$07,$F9,$09,$FE,$D1,$E6,$FE,$F2,$FF,$92,$63
	.dc.b	$FB,$73,$FC,$81,$FF,$00,$BF,$E4,$87,$97,$1F,$F9,$21,$DA,$3F,$94
	.dc.b	$BF,$DF,$A3,$4E,$B9,$BF,$AC,$47,$F2,$13,$71,$FD,$7F,$C5,$F5,$D7
	.dc.b	$F2,$08,$FF,$05,$CF,$F1,$F6,$C7,$F9,$29,$FF,$0C,$C7,$F1,$F6,$C7
	.dc.b	$F1,$F6,$BF,$C7,$DA,$FF,$1F,$6F,$2F,$9E,$3F,$DC,$F8,$FC,$97,$F8
	.dc.b	$FD,$7F,$92,$23,$13,$EA,$9F,$F0,$8C,$FF,$21,$AF,$8E,$7F,$85,$3C
	.dc.b	$4F,$EA,$FE,$2F,$F1,$F8,$FF,$12,$BF,$C7,$DB,$CD,$FC,$7D,$AF,$F1
	.dc.b	$F7,$FC,$6B,$1F,$C7,$CF,$E3,$D8,$9F,$E3,$D1,$FE,$25,$7F,$8F,$1F
	.dc.b	$EE,$97,$8C,$FF,$1A,$6B,$FE,$09,$8C,$46,$8C,$72,$FF,$8B,$23,$FB
	.dc.b	$7D,$FF,$A8,$39,$3F,$C8,$EF,$F1,$BA,$FF,$1A,$3F,$C7,$DF,$F2,$0B
	.dc.b	$FE,$43,$7F,$C9,$EF,$F1,$B7,$67,$F1,$73,$67,$F1,$72,$9F,$C7,$CF
	.dc.b	$F2,$6B,$97,$3C,$BF,$E2,$58,$FE,$35,$89,$FF,$23,$CF,$82,$BE,$95
	.dc.b	$7F,$91,$BC,$5F,$F0,$4E,$BF,$E0,$AF,$17,$F8,$D5,$FF,$93,$5F,$E3
	.dc.b	$4D,$E4,$FE,$2C,$D7,$FC,$71,$92,$73,$39,$9F,$E2,$0F,$F9,$B2,$3F
	.dc.b	$8D,$3F,$E6,$A7,$F8,$D4,$F1,$9F,$F2,$04,$7F,$1A,$9F,$F3,$4B,$FE
	.dc.b	$EC,$7F,$CE,$EF,$47,$FC,$DE,$FF,$1B,$7A,$72,$FF,$98,$A3,$D9,$3F
	.dc.b	$C6,$E7,$9B,$FE,$3B,$E4,$F2,$3F,$E5,$3E,$5F,$23,$FE,$EA,$C3,$C8
	.dc.b	$FF,$13,$AF,$0B,$FC,$8D,$E2,$FF,$1C,$BC,$46,$2B,$FC,$B1,$1F,$EF
	.dc.b	$08,$FF,$3F,$BF,$C6,$9E,$77,$FE,$59,$8F,$F2,$5C,$7C,$78,$FF,$3D
	.dc.b	$F8,$BF,$E3,$EE,$3F,$F1,$E9,$FC,$70,$FE,$37,$8F,$FB,$9A,$BF,$D3
	.dc.b	$71,$FC,$6D,$FF,$4B,$BF,$C6,$DC,$BF,$E3,$67,$FD,$2E,$FF,$1B,$8F
	.dc.b	$E3,$D1,$C2,$3F,$8D,$72,$FF,$9C,$5F,$E3,$31,$CA,$3F,$CE,$11,$FE
	.dc.b	$39,$3F,$8C,$C7,$FA,$73,$9B,$FC,$AB,$3E,$99,$FE,$35,$3F,$E9,$97
	.dc.b	$F8,$D3,$5F,$E3,$4F,$FA,$A1,$FE,$34,$D7,$F8,$D3,$5F,$E3,$4D,$7F
	.dc.b	$8D,$37,$5F,$F5,$6C,$7F,$8E,$23,$B3,$FC,$55,$A7,$B5,$C3,$E3,$3D
	.dc.b	$B1,$FE,$AC,$8F,$F3,$7F,$87,$FE,$70,$F1,$7F,$8D,$57,$FA,$CD,$FE
	.dc.b	$34,$DE,$6F,$E3,$4D,$7F,$8D,$3F,$EC,$48,$FF,$1C,$1F,$C6,$8F,$E3
	.dc.b	$2F,$FB,$51,$FF,$1B,$1F,$C6,$5E,$11,$FC,$59,$E8,$7F,$A9,$63,$0F
	.dc.b	$84,$FF,$52,$3F,$E3,$56,$9F,$EC,$6F,$17,$D6,$FF,$17,$3F,$DC,$75
	.dc.b	$FC,$6D,$A7,$F1,$7C,$D1,$FC,$6D,$B1,$FE,$D9,$6A,$3F,$8C,$1F,$EE
	.dc.b	$C8,$FF,$1B,$46,$A3,$FC,$56,$7F,$B6,$E3,$F8,$E3,$4F,$F1,$B3,$FC
	.dc.b	$70,$FE,$37,$1E,$2E,$4F,$F6,$FC,$72,$9F,$F7,$CF,$8B,$FE,$A4,$F1
	.dc.b	$7F,$8D,$C7,$FB,$F9,$FE,$36,$DE,$3F,$E2,$F8,$27,$FC,$69,$39,$9F
	.dc.b	$F6,$DB,$FF,$00,$23,$FD,$BE,$E6,$3F,$88,$37,$93,$FD,$F7,$3C,$23
	.dc.b	$E1,$10,$E9,$1F,$ED,$B8,$FE,$39,$5E,$1A,$FF,$6B,$D7,$94,$FF,$6E
	.dc.b	$4E,$1F,$51,$FE,$FE,$7F,$DB,$4F,$69,$E5,$3F,$ED,$B9,$FE,$41,$39
	.dc.b	$7F,$90,$3E,$39,$DC,$FF,$8E,$23,$13,$FE,$38,$9C,$F2,$7B,$2B,$F8
	.dc.b	$D5,$7C,$37,$FC,$67,$3E,$0E,$3F,$64,$7F,$1A,$9F,$C1,$4F,$B6,$7D
	.dc.b	$75,$E9,$7E,$24,$C1,$2F,$17,$FD,$BB,$D5,$FE,$DE,$F1,$3F,$DB,$A7
	.dc.b	$F8,$F0,$FE,$42,$7F,$B6,$8F,$E4,$15,$ED,$7F,$90,$51,$FE,$3B,$1F
	.dc.b	$1D,$FE,$40,$7F,$23,$7C,$51,$BE,$2F,$E4,$31,$E5,$3C,$B3,$FC,$8A
	.dc.b	$75,$C9,$99,$F7,$CF,$FB,$6C,$FE,$46,$F4,$F1,$7D,$57,$FC,$8A,$39
	.dc.b	$BF,$C7,$9A,$FF,$1E,$6B,$FC,$79,$AF,$F1,$E6,$F2,$FA,$2B,$84,$78
	.dc.b	$A7,$83,$5E,$0D,$9F,$46,$1E,$F9,$F5,$4F,$92,$6C,$FC,$38,$B5,$FC
	.dc.b	$99,$F7,$9C,$0F,$B3,$CA,$21,$3A,$FE,$FD,$62,$3F,$8E,$35,$DF,$67
	.dc.b	$1A,$FC,$E4,$64,$FF,$1F,$69,$F9,$97,$13,$D1,$D6,$D5,$FE,$5F,$5C
	.dc.b	$9F,$E5,$4F,$99,$F2,$D7,$DD,$BF,$E6,$33,$99,$FE,$50,$FE,$11,$FF
	.dc.b	$73,$35,$CB,$4F,$8E,$E1,$E0,$7C,$58,$F5,$47,$F8,$D7,$97,$D4,$FC
	.dc.b	$1C,$7F,$DA,$13,$AE,$BF,$54,$FC,$7D,$2B,$F1,$F0,$F6,$3F,$27,$1E
	.dc.b	$6F,$FB,$FE,$FF,$A9,$E3,$FC,$CC,$FC,$B1,$68,$FE,$67,$B3,$0F,$F9
	.dc.b	$76,$29,$FB,$58,$FE,$37,$1C,$63,$F8,$E3,$1E,$67,$EF,$23,$11,$FE
	.dc.b	$5F,$38,$47,$F4,$43,$A9,$82,$9F,$D1,$1E,$C7,$8C,$FE,$57,$16,$3B
	.dc.b	$9F,$C2,$63,$D9,$1E,$E8,$F6,$C7,$E7,$0F,$85,$16,$1E,$5F,$E6,$67
	.dc.b	$E3,$63,$9F,$2D,$E2,$DB,$E1,$E5,$ED,$79,$4E,$0E,$8E,$7C,$AD,$6B
	.dc.b	$BF,$2B,$F0,$3B,$FE,$77,$5F,$EC,$D1,$8E,$0F,$A8,$FF,$83,$23,$33
	.dc.b	$FA,$19,$FA,$39,$2B,$F1,$6E,$1D,$36,$7F,$0C,$D7,$F3,$EF,$D7,$8F
	.dc.b	$F5,$3D,$76,$9F,$F5,$0B,$5F,$C1,$47,$09,$FE,$6F,$3B,$73,$3F,$9F
	.dc.b	$7C,$07,$D1,$9C,$3E,$53,$E8,$5F,$D5,$AF,$E8,$B1,$FC,$B7,$83,$F2
	.dc.b	$93,$F9,$18,$FC,$5B,$FE,$AD,$9D,$C7,$E3,$CD,$9F,$8F,$3E,$2C,$7E
	.dc.b	$3D,$FE,$96,$7F,$AB,$4E,$67,$FA,$A6,$39,$BF,$8F,$36,$7E,$30,$FC
	.dc.b	$E4,$7E,$41,$FC,$71,$D2,$3F,$1D,$1F,$9B,$8D,$C7,$BA,$7F,$95,$F0
	.dc.b	$7E,$76,$7E,$49,$FC,$37,$0F,$E4,$DC,$5D,$9F,$AF,$1F,$09,$FE,$49
	.dc.b	$C7,$FC,$AA,$3E,$1C,$60,$FF,$1E,$61,$FD,$B9,$C4,$71,$7F,$97,$4F
	.dc.b	$F1,$A9,$D4,$62,$7E,$3B,$FC,$C7,$6F,$13,$F7,$E3,$93,$FC,$CD,$CB
	.dc.b	$F7,$E7,$9B,$FC,$1F,$2E,$13,$D2,$3E,$29,$F8,$27,$D2,$7C,$8D,$3F
	.dc.b	$07,$1F,$C8,$63,$F8,$4D,$74,$8F,$8B,$0C,$FC,$71,$F2,$27,$B3,$FD
	.dc.b	$36,$2B,$FD,$47,$8F,$CE,$FF,$53,$E3,$FF,$19,$6B,$FC,$8F,$DA,$FA
	.dc.b	$B8,$FF,$22,$ED,$B9,$7F,$00,$8F,$EF,$95,$FE,$5B,$3F,$AF,$17,$D0
	.dc.b	$E3,$CD,$89,$FE,$17,$8F,$4A,$FF,$79,$47,$21,$E6,$BC,$FF,$0B,$C4
	.dc.b	$FF,$2D,$47,$73,$F8,$69,$AE,$8F,$3F,$0F,$C7,$8F,$F2,$6D,$FC,$77
	.dc.b	$FC,$57,$8F,$E3,$DF,$79,$F8,$FE,$3F,$EE,$0F,$F2,$03,$1B,$FF,$77
	.dc.b	$E2,$F0,$61,$F8,$D8,$EB,$3A,$9B,$3F,$8E,$8D,$CF,$E3,$A7,$BC,$FE
	.dc.b	$62,$6C,$FF,$7F,$E3,$FF,$1F,$E3,$F8,$8B,$FF,$20,$3F,$8A,$AA,$7F
	.dc.b	$75,$8F,$F1,$84,$7E,$62,$29,$FE,$0E,$8F,$99,$7F,$95,$8E,$2F,$D3
	.dc.b	$3F,$20,$FB,$DF,$91,$BF,$A3,$0F,$F0,$A4,$FF,$A5,$F8,$F1,$1F,$90
	.dc.b	$9F,$7B,$FE,$17,$E4,$FF,$4D,$E3,$F9,$57,$F8,$33,$6B,$DF,$3F,$E2
	.dc.b	$0E,$4F,$7E,$1F,$BF,$E1,$FA,$AF,$F7,$0B,$FE,$90,$FF,$93,$76,$FE
	.dc.b	$ED,$1F,$E3,$27,$F0,$51,$FE,$4D,$7F,$AC,$B4,$FE,$F2,$ED,$FE,$B7
	.dc.b	$D0,$F0,$F5,$5F,$90,$9B,$3B,$AF,$F1,$C1,$F9,$DC,$B9,$47,$F9,$CB
	.dc.b	$0E,$CE,$A8,$FE,$59,$57,$FA,$FE,$3F,$94,$C7,$F9,$EE,$3F,$D8,$5C
	.dc.b	$5C,$79,$DF,$41,$FA,$DA,$D1,$FD,$6C,$F5,$46,$5C,$BC,$1F,$80,$F8
	.dc.b	$6B,$F9,$CC,$72,$E0,$FA,$CF,$CE,$7F,$C7,$F8,$7E,$8E,$3F,$D6,$73
	.dc.b	$FC,$63,$8F,$FC,$54,$DD,$FF,$B2,$B5,$FE,$40,$DE,$2E,$35,$FE,$45
	.dc.b	$AF,$F2,$1C,$F1,$9F,$F5,$F3,$CE,$69,$B8,$F2,$57,$F9,$7A,$2B,$FE
	.dc.b	$67,$E3,$F5,$3F,$EF,$DA,$FD,$04,$E2,$BF,$01,$8F,$E7,$27,$FA,$04
	.dc.b	$E2,$BF,$36,$FF,$8F,$63,$F9,$AF,$17,$FB,$47,$7F,$F7,$A4,$62,$7F
	.dc.b	$C7,$4D,$7F,$97,$1A,$3F,$CC,$53,$FE,$38,$8E,$13,$EC,$BF,$D2,$71
	.dc.b	$FA,$B0,$5F,$2C,$3E,$3D,$7F,$9F,$F8,$FD,$B5,$FC,$CE,$7F,$D8,$3A
	.dc.b	$FF,$95,$7D,$4F,$F6,$04,$7F,$AB,$E3,$F4,$2D,$3F,$CA,$FC,$5F,$E9
	.dc.b	$68,$FF,$2E,$7C,$9D,$9F,$F2,$DD,$7E,$85,$E1,$9A,$34,$FF,$AD,$23
	.dc.b	$FD,$E9,$5E,$1D,$BF,$D7,$4F,$C6,$F9,$37,$CC,$62,$BF,$3E,$77,$E0
	.dc.b	$FF,$7B,$D7,$CC,$EC,$C5,$F5,$AF,$D2,$1F,$A6,$E9,$CB,$CE,$D2,$D7
	.dc.b	$B6,$1F,$EA,$B1,$5F,$E4,$EF,$A0,$7C,$3F,$8F,$BF,$53,$FE,$52,$BF
	.dc.b	$85,$E0,$2E,$9F,$C3,$A7,$FD,$DD,$87,$49,$E2,$7F,$56,$8F,$9C,$F1
	.dc.b	$F0,$3E,$02,$F0,$3E,$77,$37,$E3,$AB,$E4,$9F,$EA,$B3,$FD,$CD,$DB
	.dc.b	$FA,$F7,$E0,$C7,$FA,$F3,$9B,$FD,$31,$8F,$F8,$4B,$AF,$FD,$31,$5F
	.dc.b	$C3,$0F,$F5,$A3,$FE,$58,$8D,$4F,$F9,$45,$AF,$F3,$F6,$F2,$78,$AF
	.dc.b	$B4,$FE,$32,$F9,$BF,$EB,$3A,$E7,$19,$9F,$F6,$7D,$7C,$48,$F0,$96
	.dc.b	$3F,$DB,$3D,$7F,$91,$7E,$79,$FD,$32,$BF,$63,$39,$BF,$F0,$F4,$7F
	.dc.b	$8B,$BA,$BF,$C7,$ED,$9F,$D9,$1E,$FA,$F9,$5C,$FF,$7A,$31,$7F,$E2
	.dc.b	$07,$FA,$A1,$F2,$9C,$47,$BD,$D7,$5D,$7F,$C1,$F1,$ED,$7F,$C5,$31
	.dc.b	$5F,$F6,$84,$DC,$3F,$6C,$F8,$70,$FF,$22,$F5,$FA,$EB,$F2,$93,$E5
	.dc.b	$3C,$DA,$FF,$B2,$FD,$D7,$27,$F3,$79,$C9,$FE,$2B,$38,$1C,$67,$83
	.dc.b	$83,$79,$AB,$FC,$43,$3D,$6B,$55,$FD,$28,$FF,$0C,$18,$3F,$AE,$57
	.dc.b	$E7,$5E,$93,$FE,$DF,$F7,$75,$1A,$BF,$E9,$CE,$8F,$F6,$0F,$C4,$BE
	.dc.b	$17,$94,$7A,$BA,$B8,$BF,$DA,$7C,$2F,$86,$FE,$78,$E2,$78,$F1,$FF
	.dc.b	$23,$5F,$AA,$3A,$DE,$5F,$F6,$7F,$91,$E4,$AF,$F2,$67,$81,$67,$C7
	.dc.b	$DD,$BC,$58,$BE,$13,$97,$13,$E3,$1F,$1C,$F2,$DF,$25,$FE,$D4,$FA
	.dc.b	$27,$98,$F4,$3E,$56,$3C,$70,$E1,$7B,$A7,$85,$AC,$BA,$E5,$D3,$C2
	.dc.b	$67,$D3,$83,$59,$72,$9C,$F8,$1F,$82,$D3,$B6,$1C,$79,$3F,$05,$87
	.dc.b	$F8,$2B,$2E,$BA,$FE,$2A,$33,$7C,$23,$F2,$B1,$FE,$19,$8C,$3F,$8C
	.dc.b	$E4,$FF,$08,$57,$C2,$F9,$1F,$A9,$5E,$DC,$BE,$B5,$7D,$A3,$FC,$2D
	.dc.b	$DD,$F3,$E5,$F2,$57,$0F,$E9,$18,$FE,$16,$7E,$77,$C9,$7D,$09,$F3
	.dc.b	$ED,$F0,$B8,$9D,$9F,$0B,$83,$E3,$71,$7A,$AF,$7E,$3D,$E3,$F8,$DA
	.dc.b	$5B,$FF,$74,$6F,$EA,$9F,$B5,$3E,$13,$C8,$78,$78,$FC,$51,$2B,$BE
	.dc.b	$7F,$E4,$AB,$E5,$1E,$23,$E9,$63,$F6,$BC,$0F,$66,$5F,$8F,$A7,$BE
	.dc.b	$F2,$FE,$0E,$9F,$5A,$7F,$78,$D9,$E0,$7E,$19,$F9,$AB,$E5,$FF,$3D
	.dc.b	$3D,$3B,$6C,$F5,$AF,$D4,$70,$FD,$37,$FD,$47,$D3,$E3,$F8,$0F,$D1
	.dc.b	$78,$FF,$C8,$67,$F8,$4D,$7F,$4B,$7F,$CB,$51,$FE,$7F,$7F,$04,$16
	.dc.b	$AF,$26,$B0,$CC,$F2,$74,$7F,$52,$CF,$FD,$D5,$3B,$8F,$F7,$04,$61
	.dc.b	$E5,$D5,$FC,$7A,$FB,$D7,$FB,$E2,$B5,$5F,$EA,$BB,$D3,$D8,$F4,$46
	.dc.b	$AF,$D3,$7C,$A7,$D9,$5E,$5B,$E4,$FE,$A2,$74,$FF,$19,$7C,$35,$FC
	.dc.b	$4D,$F9,$87,$FA,$82,$FF,$C1,$2F,$F9,$36,$75,$7F,$15,$BD,$5F,$CC
	.dc.b	$FC,$5F,$E5,$D5,$F4,$67,$FA,$8D,$F2,$BF,$F0,$34,$7F,$85,$2B,$DB
	.dc.b	$7F,$CF,$BC,$8B,$3F,$D6,$E7,$D1,$5C,$26,$D6,$6A,$9F,$D7,$8E,$35
	.dc.b	$ED,$AF,$F7,$4D,$5A,$4F,$A2,$3D,$7C,$5F,$CB,$71,$FF,$80,$18,$7F
	.dc.b	$48,$D3,$CD,$D6,$F9,$E7,$B3,$9A,$E7,$AF,$F4,$6B,$DE,$DF,$D3,$5E
	.dc.b	$3F,$43,$6B,$C3,$8F,$3F,$57,$96,$DF,$E1,$99,$B3,$FA,$A7,$FB,$45
	.dc.b	$F8,$B3,$E5,$5F,$EA,$A7,$FC,$8D,$1F,$E5,$E9,$FE,$C8,$FF,$AF,$9F
	.dc.b	$37,$8D,$F9,$DF,$9E,$7C,$B6,$38,$7D,$14,$AF,$D1,$77,$F9,$DF,$D2
	.dc.b	$7C,$FD,$E2,$63,$F6,$AF,$FC,$21,$97,$F9,$3A,$FF,$D7,$31,$FB,$03
	.dc.b	$A5,$70,$7E,$47,$99,$FD,$92,$B8,$4F,$BE,$79,$9C,$2F,$FB,$3B,$F0
	.dc.b	$8F,$F6,$4F,$B0,$F9,$EF,$91,$19,$E0,$E8,$FF,$B3,$BB,$7D,$B1,$FE
	.dc.b	$C1,$8F,$B0,$FF,$B2,$BD,$2F,$A8,$FF,$B3,$2A,$EF,$F1,$39,$B9,$CF
	.dc.b	$FB,$43,$A7,$79,$3E,$4E,$E1,$FD,$3B,$2F,$F3,$20,$F6,$9F,$42,$E9
	.dc.b	$FF,$04,$CF,$AE,$7C,$E7,$F8,$21,$C7,$D0,$FD,$48,$E3,$D7,$F8,$97
	.dc.b	$9D,$5B,$E9,$B7,$E4,$E7,$F9,$66,$16,$79,$1F,$35,$F2,$F1,$78,$21
	.dc.b	$F4,$7F,$E2,$3F,$5B,$F6,$9F,$0A,$F9,$37,$D1,$7D,$2C,$7F,$D1,$75
	.dc.b	$F8,$F7,$FC,$6A,$FE,$EF,$B3,$FC,$4E,$FE,$F7,$DF,$B3,$FB,$0C,$3C
	.dc.b	$4F,$F7,$7F,$01,$1D,$E3,$FC,$6F,$5E,$2C,$7F,$A4,$C6,$E7,$C9,$F3
	.dc.b	$3A,$8F,$DD,$9D,$AF,$CC,$FF,$93,$EF,$D6,$FB,$4F,$16,$36,$7C,$38
	.dc.b	$7B,$A7,$FC,$99,$5F,$0B,$CF,$D6,$3F,$0F,$C9,$F0,$77,$BE,$6F,$FC
	.dc.b	$03,$CF,$FD,$7B,$0F,$F7,$06,$3D,$EB,$EE,$3F,$E7,$58,$FF,$74,$FC
	.dc.b	$07,$C4,$9E,$5D,$7B,$C3,$FC,$17,$7D,$9F,$F0,$6F,$75,$F5,$7F,$41
	.dc.b	$3E,$61,$7F,$80,$70,$F9,$5F,$81,$DD,$FE,$52,$CB,$FD,$F3,$87,$C6
	.dc.b	$9F,$B6,$72,$7F,$D8,$0F,$7C,$7C,$1D,$BF,$3F,$C2,$FE,$03,$D5,$F9
	.dc.b	$BD,$3C,$58,$F5,$C1,$5F,$ED,$97,$AB,$FC,$05,$E5,$E1,$7E,$9D,$CB
	.dc.b	$FC,$21,$F8,$C7,$27,$F5,$5D,$DC,$30,$FD,$54,$FC,$1A,$F0,$1E,$47
	.dc.b	$5E,$3F,$9C,$FC,$8C,$3C,$5B,$87,$ED,$7B,$3F,$D0,$B8,$7F,$53,$F8
	.dc.b	$6F,$CF,$F4,$7F,$83,$F2,$FF,$4C,$4F,$F8,$24,$FE,$B1,$CD,$FE,$84
	.dc.b	$78,$47,$FA,$2D,$E5,$E2,$D8,$FE,$9F,$6E,$CC,$75,$A3,$F5,$BB,$78
	.dc.b	$AB,$7D,$DF,$A3,$7F,$D4,$38,$F9,$63,$F2,$51,$83,$67,$1B,$E9,$0E
	.dc.b	$DC,$61,$EC,$CB,$FD,$51,$87,$EE,$70,$F6,$9F,$D9,$38,$FD,$8F,$F6
	.dc.b	$6E,$6B,$97,$9F,$5F,$4E,$3F,$C1,$3B,$7D,$1E,$33,$BF,$F9,$A7,$67
	.dc.b	$FB,$FF,$2F,$37,$67,$FA,$8B,$5F,$E8,$DB,$F6,$C3,$85,$7C,$77,$96
	.dc.b	$D7,$1F,$E5,$D3,$D9,$FE,$81,$D7,$7B,$BF,$2F,$3F,$0E,$FC,$51,$FE
	.dc.b	$74,$F2,$BD,$F3,$FD,$75,$F0,$5F,$EA,$A3,$35,$C3,$4E,$87,$F5,$D8
	.dc.b	$FE,$74,$7F,$65,$F5,$38,$E3,$E5,$C7,$F5,$5A,$29,$E4,$3C,$1F,$E5
	.dc.b	$BE,$23,$E1,$3C,$73,$F0,$EB,$7B,$FD,$08,$7D,$2D,$C1,$E0,$FE,$2E
	.dc.b	$7A,$77,$59,$FE,$73,$E3,$FE,$26,$73,$9D,$9F,$0C,$F8,$3A,$3A,$67
	.dc.b	$C7,$B7,$F2,$DE,$17,$8F,$1F,$47,$CC,$F5,$CF,$FC,$C9,$DD,$FE,$EF
	.dc.b	$D3,$F4,$58,$FF,$99,$FC,$FF,$F2,$F1,$FE,$66,$7E,$3C,$62,$3F,$CB
	.dc.b	$5C,$5F,$E4,$17,$73,$F1,$E3,$11,$FE,$64,$8C,$1D,$5F,$F3,$53,$F3
	.dc.b	$4E,$AF,$F9,$B7,$7F,$CC,$D7,$49,$D5,$74,$AF,$93,$1A,$7F,$CF,$4F
	.dc.b	$EA,$67,$E9,$1A,$7F,$64,$FF,$9A,$9E,$67,$BE,$BF,$79,$1E,$F8,$FC
	.dc.b	$6B,$FE,$86,$8E,$33,$99,$F0,$3F,$E7,$27,$D5,$3F,$E6,$17,$D8,$4E
	.dc.b	$EF,$F3,$7B,$FD,$70,$C9,$F8,$DD,$F3,$18,$8E,$47,$FA,$64,$E0,$74
	.dc.b	$8E,$EE,$0D,$9B,$8E,$51,$E2,$9C,$CE,$DB,$1E,$27,$F8,$44,$74,$9F
	.dc.b	$E7,$4F,$91,$F2,$39,$33,$1E,$F8,$CC,$78,$4F,$F3,$C1,$E1,$3E,$03
	.dc.b	$CA,$3C,$F3,$F9,$19,$F6,$CD,$1E,$2F,$EA,$3E,$39,$F2,$AE,$5F,$45
	.dc.b	$FF,$A0,$2B,$C2,$FC,$17,$FB,$AB,$79,$7C,$B9,$7F,$A5,$67,$F6,$75
	.dc.b	$E3,$AE,$54,$F9,$0C,$65,$E2,$86,$5F,$E5,$BC,$79,$46,$32,$F4,$65
	.dc.b	$FD,$DF,$1F,$81,$87,$1C,$39,$ED,$E6,$C7,$C5,$E6,$79,$2F,$C3,$D5
	.dc.b	$E0,$C7,$E0,$70,$74,$D3,$C3,$F1,$9F,$1F,$B9,$F0,$8F,$8E,$4B,$E2
	.dc.b	$FF,$D1,$DA,$FF,$9E,$79,$3C,$1A,$FF,$A0,$FE,$12,$E5,$CB,$C5,$F1
	.dc.b	$3D,$3E,$3A,$E9,$D7,$BB,$E3,$4F,$83,$57,$C7,$62,$BC,$5B,$7F,$8D
	.dc.b	$B5,$FF,$44,$46,$3E,$93,$D0,$F8,$0F,$46,$BC,$70,$F4,$BA,$7D,$7B
	.dc.b	$FE,$17,$C9,$FB,$B3,$FA,$3C,$FF,$C9,$2F,$EB,$1F,$87,$A7,$A2,$FF
	.dc.b	$80,$4F,$E4,$67,$D1,$9F,$F6,$98,$FC,$4F,$A6,$F7,$9E,$E7,$F9,$61
	.dc.b	$FF,$32,$6D,$FD,$06,$7C,$2D,$F1,$7E,$5C,$71,$BF,$B4,$F3,$E8,$B7
	.dc.b	$D3,$8F,$F1,$D3,$DF,$1F,$F3,$2D,$FE,$86,$31,$EB,$7E,$26,$3F,$80
	.dc.b	$47,$B9,$B7,$F9,$CC,$BC,$8F,$FB,$DE,$39,$3E,$AF,$37,$BB,$EC,$B7
	.dc.b	$87,$AC,$7D,$D7,$FA,$AF,$1B,$DA,$B2,$F3,$79,$BF,$A5,$EF,$7C,$8B
	.dc.b	$C6,$FF,$0F,$7C,$8F,$4C,$3F,$7D,$BF,$2B,$FF,$29,$47,$F2,$05,$3F
	.dc.b	$92,$79,$D7,$E8,$5B,$C5,$DF,$0E,$DE,$95,$AF,$F1,$67,$0F,$F0,$7B
	.dc.b	$E3,$E9,$74,$CB,$F8,$1C,$78,$79,$3B,$63,$3C,$2E,$CF,$A6,$FA,$1F
	.dc.b	$E0,$3D,$7F,$AC,$47,$B7,$C9,$F8,$19,$FF,$5F,$3D,$47,$F6,$1B,$AF
	.dc.b	$F5,$5A,$FE,$C0,$FF,$68,$8F,$E6,$B8,$FF,$67,$CA,$CF,$F5,$D6,$BF
	.dc.b	$D6,$1A,$FF,$4B,$7F,$A0,$47,$E6,$9F,$6C,$7C,$93,$0F,$23,$CE,$E1
	.dc.b	$E2,$FF,$57,$7A,$CF,$C9,$8F,$04,$7F,$4A,$AF,$E7,$3E,$3F,$F5,$67
	.dc.b	$F9,$DF,$87,$67,$FA,$A4,$ED,$FE,$9A,$64,$FE,$BC,$64,$FE,$AC,$53
	.dc.b	$FA,$8C,$FF,$62,$7F,$C0,$13,$FD,$88,$FE,$AA,$FF,$3C,$8F,$EA,$2F
	.dc.b	$F6,$17,$FA,$F3,$F8,$E3,$FA,$89,$FD,$3C,$FE,$94,$F4,$69,$D0,$E4
	.dc.b	$64,$E4,$78,$1F,$F0,$65,$7F,$82,$9A,$FE,$7E,$FE,$23,$F6,$E7,$C8
	.dc.b	$FD,$F7,$2F,$F8,$71,$AF,$F8,$25,$F9,$CE,$5C,$BF,$76,$7F,$C9,$C7
	.dc.b	$D9,$8E,$E7,$F4,$57,$F0,$8F,$28,$C1,$87,$C6,$FF,$87,$63,$FC,$63
	.dc.b	$3F,$12,$3E,$FC,$FF,$83,$CF,$96,$7F,$85,$EB,$E3,$27,$57,$FE,$00
	.dc.b	$3E,$61,$FE,$02,$6C,$7F,$86,$A3,$EF,$9F,$84,$78,$9F,$84,$3E,$A7
	.dc.b	$57,$F3,$A9,$FF,$36,$4F,$DE,$6B,$FD,$E9,$F2,$4F,$71,$1E,$68,$FF
	.dc.b	$12,$99,$7F,$C9,$45,$8F,$BC,$79,$8F,$C6,$C7,$F8,$81,$F3,$FA,$CF
	.dc.b	$86,$7B,$4F,$85,$FE,$BB,$5E,$CE,$DF,$23,$E2,$F6,$AE,$4C,$5E,$0F
	.dc.b	$EA,$2B,$CF,$1D,$76,$F3,$3D,$FC,$FB,$A7,$82,$7F,$57,$87,$38,$F1
	.dc.b	$FC,$2E,$2F,$CD,$AF,$97,$2F,$D4,$65,$D8,$F0,$61,$E3,$D2,$DF,$B6
	.dc.b	$7B,$D7,$82,$E6,$BE,$08,$F5,$D7,$68,$FE,$21,$B7,$FA,$87,$C4,$7A
	.dc.b	$BF,$CE,$BE,$93,$FC,$1F,$DD,$E2,$9F,$B9,$99,$FF,$18,$BF,$8B,$31
	.dc.b	$EE,$FC,$A8,$E7,$D9,$F1,$DE,$BD,$FF,$DD,$6B,$A6,$3F,$D2,$3C,$7F
	.dc.b	$C1,$56,$B2,$F9,$F5,$AB,$47,$D8,$8F,$E2,$33,$BF,$1F,$A4,$7E,$86
	.dc.b	$3F,$D8,$CF,$F1,$A7,$FD,$A2,$F6,$C3,$FA,$63,$97,$E0,$36,$FB,$D5
	.dc.b	$BF,$F1,$8E,$3F,$53,$C2,$F8,$78,$FF,$B3,$F0,$FE,$28,$FB,$B8,$3B
	.dc.b	$63,$E8,$E6,$FE,$57,$EA,$BA,$1E,$FB,$E8,$DE,$4F,$FB,$47,$3F,$E5
	.dc.b	$BE,$97,$D0,$C3,$F8,$44,$7C,$F7,$93,$F9,$37,$A7,$95,$71,$EB,$3F
	.dc.b	$AD,$9F,$06,$3E,$7E,$6F,$F7,$04,$63,$E9,$35,$7E,$4C,$FE,$5E,$1F
	.dc.b	$C2,$EB,$8C,$7F,$87,$23,$FA,$0C,$63,$3F,$F3,$C4,$60,$FD,$5C,$7D
	.dc.b	$A7,$AE,$1E,$5C,$77,$1F,$9E,$9E,$AF,$C6,$C3,$C5,$C3,$6B,$E3,$61
	.dc.b	$FE,$71,$C3,$D3,$EC,$76,$BF,$47,$47,$F5,$8F,$0F,$F2,$FE,$27,$F7
	.dc.b	$DF,$13,$F9,$23,$FD,$FF,$8A,$C7,$F3,$FC,$7F,$BE,$BF,$DE,$75,$FC
	.dc.b	$E5,$7F,$43,$6C,$65,$AF,$F3,$09,$FF,$70,$C7,$F0,$59,$FA,$F1,$FE
	.dc.b	$08,$8C,$4F,$F4,$18,$EF,$1F,$C8,$67,$FA,$53,$FC,$01,$FE,$76,$F9
	.dc.b	$A7,$FA,$2C,$F9,$9F,$F1,$B4,$6D,$FF,$79,$3F,$D9,$6F,$FB,$FB,$FB
	.dc.b	$57,$F8,$C4,$FF,$75,$8F,$E7,$0D,$7F,$B4,$35,$FF,$69,$B4,$FE,$FF
	.dc.b	$1F,$D0,$E3,$FD,$C3,$F4,$1F,$D3,$1F,$EA,$11,$FC,$FD,$3F,$D4,$9C
	.dc.b	$9F,$ED,$09,$FF,$71,$BF,$D0,$63,$F9,$DB,$E9,$3F,$BE,$47,$9C,$FB
	.dc.b	$8F,$F4,$23,$F9,$2C,$7F,$26,$44,$9F,$EA,$3E,$AF,$F1,$CF,$9E,$7C
	.dc.b	$57,$EB,$F9,$7D,$37,$FD,$43,$C9,$FF,$66,$E1,$C2,$3F,$DB,$EF,$28
	.dc.b	$B1,$FD,$5B,$07,$81,$8F,$15,$1F,$CA,$B7,$FF,$75,$E1,$E2,$F0,$1A
	.dc.b	$3C,$CF,$3E,$5E,$2C,$78,$D6,$B5,$CC,$EB,$0E,$58,$7A,$B0,$CC,$A7
	.dc.b	$35,$FC,$2E,$3A,$79,$DE,$C7,$D1,$87,$9F,$3F,$E7,$FD,$DE,$69,$CE
	.dc.b	$57,$3F,$1B,$F2,$36,$EF,$9F,$CE,$F2,$FF,$D6,$EF,$F7,$EF,$63,$B5
	.dc.b	$FC,$2D,$BB,$69,$F0,$3A,$FC,$F9,$F9,$CE,$FE,$FE,$31,$F4,$F7,$F6
	.dc.b	$96,$7F,$EE,$1E,$AF,$EF,$5D,$CF,$A2,$FE,$13,$5C,$FC,$3F,$1B,$C3
	.dc.b	$CB,$F8,$18,$3D,$9B,$E4,$FF,$8C,$74,$F5,$2F,$07,$EE,$D7,$7E,$6F
	.dc.b	$88,$F8,$7C,$EB,$5F,$CE,$5F,$EC,$19,$7E,$2A,$75,$E3,$7A,$E3,$D3
	.dc.b	$5B,$DB,$FC,$FD,$97,$83,$7F,$5E,$FE,$FC,$7F,$DC,$F8,$76,$CF,$C7
	.dc.b	$8F,$F3,$E8,$E9,$3F,$0F,$6F,$1E,$1F,$12,$79,$5F,$BA,$3E,$3F,$CF
	.dc.b	$FC,$BF,$53,$FD,$C3,$C1,$FE,$BE,$F9,$CF,$81,$7F,$A2,$9F,$92,$7F
	.dc.b	$5B,$C3,$F5,$41,$FF,$74,$C7,$E8,$9F,$F7,$5F,$C5,$5F,$5F,$E7,$73
	.dc.b	$FE,$2E,$7B,$FA,$B6,$7E,$93,$F2,$30,$FC,$E1,$D6,$33,$ED,$7C,$9A
	.dc.b	$E3,$97,$C8,$AD,$67,$E9,$EF,$F6,$77,$FE,$76,$FF,$DC,$D5,$C7,$1F
	.dc.b	$EB,$2F,$68,$F8,$2F,$8E,$FF,$BC,$72,$7F,$08,$9E,$F3,$F1,$EB,$CF
	.dc.b	$8F,$F5,$98,$F8,$73,$E4,$E8,$FC,$B7,$37,$C8,$BF,$C7,$79,$DD,$76
	.dc.b	$FF,$74,$61,$FB,$AF,$82,$FF,$31,$F9,$0E,$4F,$DC,$7C,$93,$E2,$1C
	.dc.b	$BB,$C3,$F9,$41,$FC,$73,$0F,$F3,$5F,$97,$FA,$A8,$FE,$13,$1D,$30
	.dc.b	$C4,$23,$13,$B7,$45,$8F,$F3,$7E,$3E,$3E,$17,$37,$ED,$AB,$A4,$3E
	.dc.b	$4F,$17,$C3,$1F,$4F,$47,$F1,$2B,$DE,$AF,$87,$C5,$EF,$14,$FB,$EF
	.dc.b	$DD,$E3,$7F,$46,$AF,$4E,$1F,$1B,$CB,$B1,$C4,$F0,$CF,$F4,$CC,$BE
	.dc.b	$4C,$76,$8C,$63,$DB,$B7,$1B,$7B,$73,$F9,$D7,$E0,$8C,$70,$FF,$92
	.dc.b	$33,$FE,$39,$3F,$28,$F9,$47,$C6,$F9,$FF,$1E,$5E,$1F,$3B,$E5,$3E
	.dc.b	$0C,$BE,$2F,$81,$FE,$3C,$D9,$F4,$2F,$8F,$7C,$CB,$E3,$FE,$F5,$FE
	.dc.b	$1B,$4E,$3E,$BF,$FB,$17,$6F,$E5,$CB,$2D,$4F,$E0,$0F,$E4,$B1,$9C
	.dc.b	$BE,$E5,$F8,$57,$A3,$D2,$FC,$7F,$39,$67,$C9,$D3,$F5,$AF,$86,$AF
	.dc.b	$9A,$C6,$5D,$63,$D3,$E8,$FA,$3B,$3E,$36,$DE,$6D,$BE,$BE,$3E,$1E
	.dc.b	$3E,$D8,$4C,$39,$ED,$F1,$23,$E1,$DF,$E7,$FC,$EF,$A0,$FD,$0C,$3F
	.dc.b	$CC,$BE,$F7,$4D,$3E,$17,$0F,$B9,$F8,$9F,$27,$72,$E9,$C9,$F9,$AC
	.dc.b	$7E,$37,$BA,$F9,$2F,$F5,$CF,$D1,$BE,$6F,$FB,$44,$FE,$BC,$F0,$F1
	.dc.b	$7F,$82,$9F,$F4,$EB,$6B,$F8,$15,$ED,$FF,$61,$F0,$FE,$AE,$BB,$78
	.dc.b	$3F,$DC,$6B,$F4,$31,$8C,$7F,$C0,$11,$8B,$F1,$E3,$FA,$CA,$D3,$FC
	.dc.b	$96,$7F,$23,$18,$F0,$39,$EF,$F4,$25,$3E,$D7,$F4,$EF,$8B,$CD,$F2
	.dc.b	$61,$F0,$27,$F7,$3E,$5F,$B9,$E3,$1F,$8F,$6F,$D0,$F2,$8F,$8F,$D5
	.dc.b	$E5,$7E,$4F,$6D,$D9,$B7,$CB,$E9,$F6,$46,$37,$FD,$4E,$5F,$4B,$0F
	.dc.b	$7E,$FF,$D0,$78,$7F,$C6,$FE,$F7,$FB,$D7,$47,$CC,$FF,$80,$3E,$59
	.dc.b	$3F,$C8,$FC,$5F,$43,$F4,$B2,$FE,$4C,$F1,$7F,$8E,$3F,$EF,$FF,$A5
	.dc.b	$FE,$12,$F5,$E8,$CF,$B3,$FC,$7B,$1B,$97,$83,$B7,$F8,$ED,$F8,$E7
	.dc.b	$FC,$23,$C3,$FC,$6A,$31,$2D,$31,$87,$38,$C7,$81,$E4,$D3,$EB,$5F
	.dc.b	$F5,$27,$F7,$0F,$FB,$DD,$F1,$F8,$9E,$6B,$FE,$63,$EC,$79,$33,$FC
	.dc.b	$15,$7F,$12,$BF,$27,$83,$E6,$BF,$07,$67,$8B,$6F,$57,$89,$FE,$44
	.dc.b	$F4,$FF,$CC,$7E,$3F,$F8,$87,$E2,$5F,$49,$FC,$FB,$E0,$9F,$27,$FA
	.dc.b	$7D,$98,$7F,$22,$7F,$A5,$BC,$BB,$7F,$92,$3D,$3D,$2B,$2C,$BF,$C5
	.dc.b	$DD,$1C,$23,$1E,$C3,$4C,$57,$8F,$A7,$FA,$D6,$3E,$6A,$D7,$93,$8C
	.dc.b	$38,$E1,$E8,$E0,$FF,$24,$DE,$E3,$1E,$2F,$BE,$B7,$C1,$ED,$7F,$CD
	.dc.b	$35,$B8,$F8,$13,$89,$DD,$FF,$9E,$BE,$57,$1A,$7F,$12,$C3,$E4,$F4
	.dc.b	$FB,$A3,$FC,$95,$7E,$AD,$3D,$9D,$5E,$79,$F9,$D9,$D9,$F8,$0F,$CF
	.dc.b	$E2,$EB,$1F,$9E,$BD,$6B,$F5,$31,$F6,$C7,$AB,$B6,$E3,$D3,$5F,$E1
	.dc.b	$33,$FE,$88,$D3,$D9,$0F,$A4,$F8,$55,$BE,$5D,$DF,$DC,$1D,$3F,$C8
	.dc.b	$7E,$53,$FD,$2F,$C9,$FE,$61,$F8,$CF,$F0,$BE,$7F,$C8,$6F,$EE,$DF
	.dc.b	$F9,$0D,$E4,$FF,$59,$7E,$F3,$FC,$AF,$A2,$F8,$AC,$69,$FE,$76,$AF
	.dc.b	$41,$E2,$8D,$7C,$37,$DF,$BF,$23,$FD,$7A,$7B,$CE,$BE,$67,$E2,$D6
	.dc.b	$B3,$FE,$C9,$5F,$6E,$31,$EF,$FE,$0D,$D7,$BF,$FE,$39,$D3,$FD,$8B
	.dc.b	$18,$D7,$85,$FB,$B2,$9A,$7F,$68,$8E,$F1,$E6,$CB,$FA,$0E,$1C,$A3
	.dc.b	$18,$F7,$E1,$E5,$D9,$F0,$BE,$6B,$E0,$FC,$47,$B6,$FD,$FE,$0F,$4D
	.dc.b	$3E,$8E,$3E,$AD,$3F,$DB,$39,$F5,$9C,$9F,$97,$F8,$5F,$B3,$F1,$6F
	.dc.b	$86,$FD,$CF,$B8,$FA,$17,$87,$F3,$AF,$F3,$59,$1F,$E8,$EF,$ED,$5F
	.dc.b	$C1,$F6,$DE,$57,$CD,$DC,$FF,$8C,$6B,$FE,$EC,$BF,$0E,$9F,$B5,$6E
	.dc.b	$5F,$98,$73,$E2,$5C,$58,$79,$FC,$2F,$93,$A7,$8E,$73,$8F,$F1,$2F
	.dc.b	$81,$FD,$D6,$3D,$AF,$FC,$C9,$AF,$C1,$BD,$C7,$F9,$9F,$6C,$63,$E9
	.dc.b	$E7,$F4,$3E,$5A,$DB,$D1,$E3,$77,$3F,$27,$05,$D3,$E5,$E1,$FC,$17
	.dc.b	$CA,$F9,$3A,$7F,$32,$CB,$E6,$65,$EC,$EE,$ED,$97,$F0,$49,$E5,$EC
	.dc.b	$7C,$18,$EB,$F1,$DF,$EF,$AF,$88,$FE,$67,$C6,$F9,$0F,$B5,$D0,$F9
	.dc.b	$57,$C7,$BE,$71,$A4,$C7,$F9,$DB,$FE,$55,$79,$BF,$E8,$07,$F9,$FF
	.dc.b	$36,$63,$F9,$A6,$5F,$D4,$54,$F8,$B8,$F8,$46,$B3,$FE,$AD,$3E,$67
	.dc.b	$D9,$1E,$93,$53,$F9,$77,$CA,$74,$33,$9F,$FB,$FA,$B9,$4F,$FB,$B6
	.dc.b	$F7,$5E,$B3,$7F,$39,$87,$97,$91,$F9,$AC,$7E,$2E,$BF,$C6,$B1,$E7
	.dc.b	$97,$C7,$D7,$C9,$7B,$CF,$FD,$F5,$7F,$2B,$BB,$CB,$87,$AF,$2F,$F5
	.dc.b	$43,$E4,$7C,$C7,$B3,$8A,$E1,$CF,$2F,$E0,$53,$C8,$F6,$E1,$E7,$BF
	.dc.b	$1B,$F6,$3C,$3D,$7A,$BF,$E5,$A9,$FE,$E3,$C9,$F4,$F1,$FD,$AE,$5F
	.dc.b	$41,$F7,$3E,$4C,$7E,$6F,$8D,$E2,$E1,$FC,$36,$DE,$CE,$9F,$7C,$63
	.dc.b	$6F,$C4,$7C,$87,$D5,$D5,$EF,$3E,$4F,$FD,$EB,$F2,$8F,$90,$FF,$3E
	.dc.b	$6B,$F4,$7B,$3C,$CB,$0F,$F1,$63,$EA,$C3,$F1,$51,$D3,$87,$F0,$EF
	.dc.b	$6C,$6C,$F7,$9F,$54,$E7,$E4,$3F,$C6,$DC,$7F,$9A,$8C,$78,$1F,$41
	.dc.b	$D7,$AE,$F0,$BE,$66,$7E,$3C,$BD,$58,$FF,$B9,$36,$CE,$1C,$17,$AD
	.dc.b	$F0,$6B,$78,$A7,$F9,$06,$1E,$BD,$7E,$1E,$FF,$E1,$FD,$7C,$98,$7F
	.dc.b	$55,$AD,$71,$7A,$A3,$C3,$1E,$77,$C7,$AD,$C1,$AF,$FB,$E2,$BC,$B5
	.dc.b	$E2,$CB,$E8,$47,$86,$7D,$59,$7F,$7A,$CB,$E2,$79,$BF,$3B,$E7,$3E
	.dc.b	$7E,$2F,$93,$F9,$B9,$F6,$DB,$FD,$23,$87,$A2,$31,$F2,$FF,$23,$E1
	.dc.b	$BF,$DD,$BE,$FB,$E5,$BF,$85,$F9,$BF,$E3,$CF,$83,$FC,$D6,$FF,$80
	.dc.b	$BF,$F2,$07,$0F,$E8,$9F,$F7,$F3,$FC,$11,$FF,$7B,$F0,$5F,$90,$7C
	.dc.b	$8D,$CF,$B7,$57,$19,$47,$87,$C6,$C3,$7A,$3C,$D3,$AD,$BC,$B9,$FF
	.dc.b	$0D,$CF,$D3,$1E,$8E,$9C,$56,$F1,$6F,$FE,$1B,$C3,$F8,$56,$1C,$27
	.dc.b	$3E,$8F,$E1,$27,$27,$F2,$39,$5D,$3C,$F1,$FB,$18,$F9,$3B,$59,$F3
	.dc.b	$4F,$C5,$EA,$E1,$AF,$CD,$F0,$3D,$BA,$BB,$EE,$9F,$E3,$0C,$3F,$92
	.dc.b	$67,$E5,$F9,$6F,$83,$8F,$C5,$DB,$FE,$00,$73,$FF,$3E,$AF,$54,$7F
	.dc.b	$2B,$BF,$6E,$BF,$06,$3F,$CE,$BE,$37,$F3,$0F,$9C,$FF,$3E,$F3,$BE
	.dc.b	$57,$F0,$1C,$5F,$DC,$2F,$DF,$CB,$FD,$2A,$FF,$A4,$E5,$FC,$BC,$E4
	.dc.b	$7F,$9E,$79,$70,$E3,$C8,$F9,$CE,$E7,$BF,$26,$A1,$1E,$2F,$96,$C3
	.dc.b	$E4,$CF,$C7,$3F,$CF,$63,$1C,$AF,$91,$E9,$3F,$E8,$9D,$BD,$52,$DB
	.dc.b	$F8,$24,$FF,$48,$BE,$31,$FE,$0C,$8C,$1C,$70,$F8,$3E,$97,$F8,$8A
	.dc.b	$3D,$58,$74,$AF,$F0,$9F,$2F,$0D,$53,$9B,$FC,$6E,$BC,$31,$E7,$C1
	.dc.b	$F8,$8C,$BE,$51,$E7,$3D,$E7,$A7,$AA,$E3,$D2,$AE,$3F,$57,$3F,$9B
	.dc.b	$A7,$C2,$C3,$E0,$DE,$F5,$F8,$19,$FC,$79,$F6,$47,$C3,$D3,$F4,$31
	.dc.b	$F6,$3C,$5F,$A4,$F9,$2F,$FC,$D0,$FA,$6B,$E5,$7A,$FC,$E7,$B6,$5F
	.dc.b	$E3,$4D,$7E,$16,$9F,$07,$6F,$9F,$8F,$DC,$3C,$35,$F8,$BF,$96,$F0
	.dc.b	$7C,$9B,$E8,$FF,$87,$78,$CF,$4D,$EF,$3E,$51,$E9,$7C,$8E,$2F,$ED
	.dc.b	$7A,$7E,$01,$65,$EA,$E5,$FE,$8C,$FA,$A7,$C5,$3A,$F8,$8F,$9E,$F0
	.dc.b	$D3,$F8,$F6,$1C,$EF,$C9,$C9,$FD,$8B,$0F,$95,$7B,$8F,$EC,$BD,$98
	.dc.b	$AD,$EC,$F8,$3C,$65,$EC,$DF,$FC,$3D,$AF,$92,$31,$BF,$E4,$33,$E9
	.dc.b	$87,$C1,$87,$77,$C6,$AD,$E5,$F5,$E3,$DD,$AF,$F6,$F8,$FF,$26,$3E
	.dc.b	$E7,$E6,$3F,$86,$35,$3F,$E2,$F7,$A6,$B9,$C7,$C3,$EB,$7A,$23,$FC
	.dc.b	$03,$A5,$CF,$EA,$6F,$E7,$CB,$DF,$3F,$5B,$87,$F8,$0F,$A5,$F8,$AC
	.dc.b	$FE,$7F,$CB,$79,$B4,$F9,$1C,$BF,$37,$6F,$E9,$9E,$85,$DB,$BF,$CC
	.dc.b	$F5,$9F,$8B,$C1,$D3,$E6,$7F,$D2,$9F,$31,$FE,$0D,$F6,$3F,$8B,$79
	.dc.b	$0F,$7B,$F5,$7C,$0B,$DE,$DF,$E9,$DF,$11,$7F,$C0,$BA,$7F,$DA,$9C
	.dc.b	$5F,$B0,$C3,$E2,$3F,$C0,$97,$B3,$F3,$9E,$9E,$7E,$F2,$F0,$57,$CD
	.dc.b	$9F,$07,$C0,$FF,$62,$8F,$43,$FE,$26,$7F,$65,$1A,$DB,$FD,$D3,$18
	.dc.b	$AF,$3D,$DB,$57,$F8,$AF,$E3,$30,$FC,$1F,$23,$CD,$7E,$EE,$0F,$E7
	.dc.b	$18,$F8,$2B,$E3,$6D,$FD,$66,$3F,$65,$9F,$87,$7F,$89,$EA,$6D,$F8
	.dc.b	$DC,$1F,$EC,$1A,$DE,$5D,$70,$F9,$39,$FC,$EE,$CF,$89,$32,$3E,$26
	.dc.b	$D3,$C6,$F8,$93,$F4,$B0,$F4,$E5,$FE,$CF,$7D,$35,$EE,$8F,$F6,$95
	.dc.b	$63,$5F,$4E,$7F,$2B,$E9,$39,$DF,$C4,$F3,$BF,$9F,$EF,$EB,$D7,$AF
	.dc.b	$03,$CE,$FE,$77,$F3,$0D,$3F,$D4,$9D,$7F,$DE,$3E,$03,$FC,$8D,$EE
	.dc.b	$38,$FF,$99,$EF,$F9,$9E,$9F,$F5,$3E,$3F,$DC,$78,$F9,$E5,$FE,$3D
	.dc.b	$7F,$9C,$C7,$5F,$0B,$F9,$CB,$DF,$E0,$3F,$BE,$4F,$F3,$18,$C6,$9F
	.dc.b	$E2,$48,$E7,$3F,$E2,$DF,$4F,$F9,$3C,$74,$C3,$CF,$1F,$1E,$3F,$B6
	.dc.b	$DF,$28,$FF,$60,$75,$F2,$C3,$D8,$FB,$EB,$97,$77,$93,$0F,$F6,$24
	.dc.b	$7F,$03,$C3,$CD,$E1,$59,$F8,$B3,$FD,$AA,$7B,$3E,$98,$F1,$F0,$5C
	.dc.b	$FD,$0F,$F9,$BF,$CA,$F2,$CE,$B2,$FF,$2D,$DF,$C6,$E0,$FE,$2F,$8B
	.dc.b	$E1,$ED,$FD,$4B,$F0,$3F,$03,$97,$B6,$DF,$5D,$F7,$6F,$FE,$DF,$F7
	.dc.b	$FF,$BB,$FC,$A7,$FB,$71,$FF,$2B,$79,$3F,$E9,$7D,$BF,$CB,$1E,$5F
	.dc.b	$F6,$EA,$FB,$9B,$97,$A5,$FC,$A9,$FF,$72,$EB,$FD,$C9,$FF,$73,$FB
	.dc.b	$B6,$BF,$D0,$5A,$FF,$2B,$9F,$03,$FD,$4A,$3B,$1A,$CB,$FC,$27,$3A
	.dc.b	$7F,$C7,$5C,$FF,$EF,$28,$C7,$BD,$FD,$D5,$BA,$7B,$2F,$94,$BE,$0F
	.dc.b	$1B,$7B,$A1,$5F,$E7,$5A,$F0,$CF,$9C,$FE,$36,$FF,$71,$BF,$3E,$DE
	.dc.b	$48,$F3,$57,$85,$EE,$FF,$89,$D0,$FA,$4F,$F1,$FB,$F0,$7C,$0F,$90
	.dc.b	$FF,$AF,$8E,$C7,$C6,$F3,$BB,$63,$F5,$3E,$0B,$B3,$E9,$E5,$FF,$7C
	.dc.b	$E1,$DB,$8F,$FB,$EF,$B8,$F9,$7F,$F0,$0E,$DF,$E9,$9F,$92,$7C,$5F
	.dc.b	$93,$0F,$F6,$16,$BF,$E1,$87,$97,$4D,$8F,$BA,$FF,$64,$8C,$E3,$FE
	.dc.b	$4F,$6F,$07,$F9,$59,$CE,$5F,$D0,$9B,$EB,$7F,$97,$5B,$C3,$F1,$31
	.dc.b	$91,$AD,$31,$1E,$53,$C3,$8F,$AE,$FF,$21,$B7,$C0,$9C,$E3,$EE,$E0
	.dc.b	$C6,$FF,$0A,$31,$C5,$DA,$B7,$87,$E6,$6D,$C1,$FE,$24,$9F,$AC,$78
	.dc.b	$3C,$1F,$63,$D8,$F5,$1E,$0E,$0B,$97,$7C,$3F,$49,$C1,$E2,$96,$BE
	.dc.b	$9D,$BF,$D9,$5C,$1E,$3C,$38,$47,$D5,$C7,$E0,$47,$E4,$1B,$EB,$7D
	.dc.b	$38,$FF,$1C,$ED,$F4,$67,$EA,$F3,$D9,$F8,$1E,$A7,$93,$3F,$0F,$0F
	.dc.b	$E0,$BC,$6F,$26,$BE,$3E,$2F,$83,$E2,$7C,$FF,$27,$F9,$D7,$C4,$3E
	.dc.b	$51,$E0,$BD,$A7,$CD,$FE,$37,$E0,$3E,$B7,$8F,$FE,$17,$ED,$FC,$FE
	.dc.b	$5F,$DE,$79,$3E,$B3,$FD,$17,$AF,$D0,$FF,$47,$F9,$DB,$E4,$B6,$FA
	.dc.b	$57,$F8,$23,$E9,$D7,$93,$6F,$E7,$B1,$E7,$AE,$11,$E6,$E8,$E1,$97
	.dc.b	$8A,$73,$BF,$DE,$BF,$1C,$63,$7F,$5C,$7E,$17,$1E,$D5,$EF,$DF,$FD
	.dc.b	$31,$3E,$2B,$E1,$1F,$AD,$9F,$91,$3F,$A7,$AF,$EC,$58,$FB,$33,$95
	.dc.b	$F8,$A9,$F5,$57,$E2,$9E,$FD,$7E,$88,$47,$8B,$2C,$37,$87,$D3,$87
	.dc.b	$D5,$C3,$CD,$BF,$A7,$4F,$2F,$2D,$AD,$63,$F6,$23,$FB,$B6,$5F,$33
	.dc.b	$0F,$47,$CA,$7A,$70,$F9,$7C,$FE,$57,$E6,$F8,$57,$2E,$7F,$4B,$FB
	.dc.b	$1F,$8E,$FF,$36,$FA,$9F,$E0,$7F,$07,$FC,$F1,$F2,$DF,$89,$F4,$7F
	.dc.b	$75,$5F,$85,$D8,$FF,$A2,$DF,$D7,$AF,$94,$FE,$20,$FF,$9E,$34,$FC
	.dc.b	$F3,$FE,$7A,$F6,$AF,$3B,$3F,$C7,$61,$CA,$31,$EA,$FE,$E2,$FC,$AF
	.dc.b	$DA,$9F,$E0,$51,$8D,$31,$1C,$6F,$D5,$F2,$7F,$4E,$B5,$F0,$7F,$CE
	.dc.b	$EB,$D7,$8F,$6C,$3F,$D2,$DF,$31,$F0,$E3,$14,$DF,$F3,$7D,$BF,$D0
	.dc.b	$B1,$E7,$1F,$AD,$C7,$9E,$7F,$26,$F8,$77,$7F,$B9,$F0,$E1,$8F,$F8
	.dc.b	$A6,$F6,$7B,$5F,$85,$E6,$5C,$3E,$97,$5F,$A6,$7C,$75,$E5,$9E,$D9
	.dc.b	$5D,$FF,$B0,$E7,$E7,$E9,$F8,$B8,$7A,$AB,$F8,$16,$3F,$CC,$FE,$4F
	.dc.b	$ED,$6D,$FD,$DF,$3E,$5A,$7F,$24,$8F,$9F,$E8,$FE,$E6,$7F,$D5,$3E
	.dc.b	$4B,$E5,$9F,$35,$FE,$19,$F8,$3F,$C1,$F3,$BE,$77,$FD,$21,$87,$03
	.dc.b	$FD,$D7,$F0,$DF,$D7,$7A,$2C,$7F,$84,$31,$B9,$75,$F1,$BC,$13,$FD
	.dc.b	$C2,$FF,$B2,$53,$E2,$F3,$3E,$48,$F1,$EF,$C2,$B7,$AB,$E7,$71,$7D
	.dc.b	$E3,$FC,$61,$E3,$FB,$6B,$7E,$EF,$FA,$7A,$BF,$8C,$46,$38,$3F,$D1
	.dc.b	$11,$E3,$39,$4F,$FA,$93,$8F,$71,$89,$FD,$2F,$7F,$F6,$2B,$E5,$C9
	.dc.b	$FE,$73,$AF,$ED,$11,$8E,$3F,$DD,$C7,$AF,$3F,$99,$49,$FF,$54,$56
	.dc.b	$F8,$FF,$C2,$8F,$BA,$FD,$BC,$5F,$57,$0F,$F2,$64,$7B,$4F,$3F,$8B
	.dc.b	$67,$5A,$FF,$AF,$32,$F4,$C6,$34,$FD,$EF,$07,$A7,$0F,$65,$7F,$AB
	.dc.b	$1F,$85,$5B,$8F,$34,$FB,$27,$7A,$7E,$0D,$DF,$0F,$E2,$31,$F0,$4F
	.dc.b	$E1,$F2,$CE,$7F,$02,$3F,$15,$9F,$CF,$F8,$4F,$F7,$56,$BF,$E2,$5C
	.dc.b	$FF,$0F,$F3,$5F,$88,$CD,$D5,$F8,$4A,$FC,$0F,$0F,$C4,$9F,$07,$CC
	.dc.b	$7C,$3C,$7F,$DB,$7F,$37,$FC,$DB,$E6,$1E,$BB,$1F,$CE,$FC,$37,$DF
	.dc.b	$F4,$3C,$3C,$8C,$7F,$BE,$3B,$31,$C1,$F9,$C8,$FF,$73,$ED,$E3,$9C
	.dc.b	$E7,$67,$C3,$E9,$FE,$51,$F1,$E1,$FD,$A2,$BE,$54,$63,$4F,$46,$2C
	.dc.b	$3D,$18,$FE,$42,$B5,$A5,$AD,$62,$C3,$DB,$7F,$6E,$F7,$A5,$BD,$D7
	.dc.b	$C2,$8F,$F7,$31,$FE,$60,$9D,$F6,$79,$2B,$78,$FC,$7E,$4F,$F0,$6E
	.dc.b	$9F,$0E,$FF,$CC,$4D,$AF,$1B,$EA,$D3,$FB,$2F,$85,$75,$F3,$77,$7A
	.dc.b	$67,$EA,$65,$FB,$0D,$3C,$58,$F8,$1F,$E2,$38,$FE,$16,$31,$9B,$E2
	.dc.b	$FF,$11,$C5,$F4,$E7,$F8,$F5,$6F,$E3,$BE,$94,$7C,$BC,$FE,$93,$F2
	.dc.b	$3E,$67,$C9,$C3,$FE,$4A,$BF,$81,$AF,$93,$D4,$F9,$0F,$9B,$5F,$C8
	.dc.b	$CF,$8E,$31,$AF,$F9,$4E,$7F,$D1,$3F,$13,$F4,$FE,$77,$FB,$4F,$88
	.dc.b	$F1,$DF,$3D,$F7,$7D,$DF,$E1,$DF,$1F,$F9,$0C,$3E,$5E,$3C,$B6,$FE
	.dc.b	$53,$1F,$77,$C0,$FC,$73,$D7,$6F,$E3,$EF,$7E,$2B,$4D,$5E,$3E,$1F
	.dc.b	$DF,$DF,$F1,$AB,$F3,$78,$BF,$CA,$63,$17,$C6,$3F,$96,$6D,$8B,$F5
	.dc.b	$BE,$BA,$FE,$05,$7C,$B7,$F7,$6F,$FC,$2E,$31,$5F,$03,$B7,$79,$75
	.dc.b	$BD,$F1,$63,$2E,$B8,$FE,$87,$6F,$E6,$59,$72,$8F,$D2,$7B,$B8,$D3
	.dc.b	$E2,$F6,$7C,$6C,$FF,$C0,$B1,$CF,$6F,$80,$7A,$AF,$F5,$9E,$57,$4E
	.dc.b	$7E,$1F,$0D,$F3,$30,$F8,$1E,$27,$F2,$BD,$3E,$2F,$CC,$59,$F4,$61
	.dc.b	$F0,$F2,$FE,$A1,$F4,$97,$8B,$EF,$4F,$C1,$AF,$77,$23,$E3,$B1,$C7
	.dc.b	$EE,$8F,$35,$FC,$8E,$EF,$F5,$77,$6F,$FB,$17,$C1,$7C,$07,$F8,$BF
	.dc.b	$9F,$EE,$76,$FF,$B1,$39,$FF,$DA,$F7,$FE,$D6,$E7,$FE,$F4,$FF,$B3
	.dc.b	$7A,$BF,$B4,$C7,$FA,$0C,$FF,$80,$38,$E2,$7B,$69,$FD,$B9,$F0,$65
	.dc.b	$6B,$6F,$E8,$1F,$F3,$F4,$6B,$E7,$FF,$BC,$BC,$3D,$A6,$9E,$2A,$E3
	.dc.b	$18,$8E,$31,$87,$F2,$CF,$FB,$2A,$79,$47,$E6,$7B,$B1,$3B,$E0,$FF
	.dc.b	$29,$BF,$18,$C6,$DF,$E9,$2B,$E3,$28,$7C,$3E,$FD,$0F,$E8,$3A,$7C
	.dc.b	$BA,$D7,$C4,$BE,$5F,$E5,$E4,$BF,$F0,$1B,$E8,$7F,$96,$BF,$00,$F6
	.dc.b	$9E,$F7,$AE,$BF,$0B,$9F,$87,$0F,$DA,$8F,$8B,$8F,$96,$FC,$93,$9D
	.dc.b	$FE,$EE,$58,$CF,$CD,$AA,$3F,$D0,$7B,$7F,$37,$C3,$ED,$1E,$7F,$6A
	.dc.b	$EB,$EE,$E2,$F3,$3E,$4E,$DF,$36,$9E,$3F,$4E,$E9,$F8,$DE,$FF,$3B
	.dc.b	$E0,$F5,$7E,$77,$9B,$FC,$67,$D3,$7C,$C3,$6F,$F3,$47,$CE,$7C,$9E
	.dc.b	$6F,$93,$7F,$CC,$A6,$F8,$3E,$68,$FE,$45,$B7,$E3,$5E,$93,$F2,$DE
	.dc.b	$D5,$78,$3F,$C4,$98,$DC,$BC,$3C,$1F,$8D,$BF,$24,$F8,$F4,$FC,$6C
	.dc.b	$63,$1F,$F3,$E1,$F8,$A7,$1E,$CF,$F8,$FB,$8F,$FB,$06,$5D,$AB,$FD
	.dc.b	$05,$F1,$1B,$8C,$53,$7F,$EC,$35,$FA,$BC,$39,$C7,$E2,$63,$E0,$D7
	.dc.b	$08,$FC,$67,$C3,$63,$A3,$E0,$4E,$6B,$E6,$BB,$E0,$BB,$FC,$E9,$F9
	.dc.b	$BD,$7F,$CB,$F5,$FF,$7B,$C7,$CB,$E1,$F9,$98,$7F,$93,$E3,$E2,$3F
	.dc.b	$E3,$66,$DF,$F9,$ED,$BC,$1F,$9C,$7D,$53,$F0,$7E,$03,$E4,$BF,$03
	.dc.b	$C0,$FD,$8F,$AC,$F8,$67,$5B,$E2,$7F,$4D,$DA,$FA,$57,$C2,$7F,$A9
	.dc.b	$F8,$3F,$B1,$EB,$F9,$6C,$BF,$D8,$2F,$EB,$9E,$8F,$F0,$F7,$B6,$77
	.dc.b	$3F,$F1,$84,$FF,$0B,$9F,$16,$7F,$EB,$D8,$F2,$CF,$93,$3F,$DA,$56
	.dc.b	$9F,$F1,$D4,$FF,$10,$8C,$75,$7F,$8C,$63,$F9,$24,$63,$4D,$5F,$3A
	.dc.b	$FD,$86,$BD,$B1,$F8,$11,$FB,$3D,$71,$97,$C2,$5C,$3F,$C5,$72,$E9
	.dc.b	$0E,$5C,$65,$D7,$3F,$91,$BB,$DE,$F8,$9B,$7C,$7A,$FA,$CF,$C7,$7F
	.dc.b	$C2,$CF,$9B,$5F,$53,$97,$CA,$E7,$4F,$AB,$BF,$9B,$6F,$46,$FE,$2B
	.dc.b	$F5,$E7,$E2,$C3,$96,$1E,$BC,$FF,$C7,$FC,$1F,$BF,$DF,$F4,$35,$F4
	.dc.b	$FE,$4B,$CD,$8F,$FB,$0A,$3C,$1E,$37,$D7,$DB,$E5,$76,$F8,$F6,$FE
	.dc.b	$67,$5E,$3D,$5F,$11,$FC,$FB,$E3,$5E,$4B,$E9,$5F,$26,$F7,$DF,$1C
	.dc.b	$F9,$E7,$AF,$FE,$A9,$D3,$FA,$F6,$5F,$D8,$1B,$D1,$FC,$9A,$3A,$C7
	.dc.b	$F5,$D7,$BF,$45,$D3,$FC,$6B,$3F,$CF,$67,$C7,$A7,$C0,$8F,$EA,$11
	.dc.b	$8D,$35,$5E,$75,$CB,$FD,$86,$75,$AF,$1E,$6F,$E5,$11,$8F,$03,$FB
	.dc.b	$1B,$AD,$FF,$B4,$E5,$DB,$2E,$98,$7C,$0F,$86,$F8,$D1,$8A,$F8,$5D
	.dc.b	$7B,$D7,$E3,$DF,$C6,$E1,$7D,$06,$77,$F9,$58,$7C,$9E,$2F,$E9,$7A
	.dc.b	$35,$F7,$BE,$CC,$B9,$BE,$6D,$3C,$7E,$1D,$9E,$F8,$5C,$7A,$E7,$E6
	.dc.b	$9C,$BF,$CA,$38,$3E,$0E,$1E,$98,$FB,$38,$FF,$70,$CB,$E1,$47,$DC
	.dc.b	$F1,$31,$5B,$D3,$FD,$D9,$F2,$5F,$E1,$FC,$3E,$4F,$27,$DD,$F7,$7E
	.dc.b	$F5,$FF,$90,$A7,$8E,$9F,$A6,$7F,$0B,$1F,$8D,$F9,$6F,$76,$DF,$C8
	.dc.b	$F6,$F2,$DF,$B3,$83,$FA,$F6,$1F,$D7,$F3,$FE,$C1,$D5,$FD,$CF,$D2
	.dc.b	$7C,$3F,$F0,$0C,$DE,$B3,$E8,$3F,$CD,$98,$7F,$77,$C7,$99,$03,$D3
	.dc.b	$A3,$FC,$6E,$F6,$EE,$B3,$FC,$BB,$93,$83,$E4,$9D,$7A,$58,$8F,$31
	.dc.b	$FC,$D3,$4F,$CC,$38,$D7,$D1,$AE,$EF,$B5,$7A,$BC,$9E,$B1,$89,$7C
	.dc.b	$5E,$30,$E9,$1F,$C7,$B6,$F5,$56,$EF,$E1,$D7,$BA,$31,$BF,$F0,$86
	.dc.b	$FC,$96,$67,$9D,$2B,$E4,$FC,$9F,$B6,$73,$1F,$33,$87,$F4,$B8,$74
	.dc.b	$C7,$E7,$7A,$DE,$DE,$1F,$C0,$61,$F3,$0F,$84,$7B,$CE,$67,$87,$7F
	.dc.b	$87,$9F,$86,$DE,$2F,$86,$73,$9F,$86,$FE,$54,$FD,$49,$FE,$37,$1D
	.dc.b	$5F,$C2,$71,$DC,$7F,$D6,$5F,$15,$78,$BF,$0F,$C9,$F2,$EF,$C3,$7F
	.dc.b	$2B,$0F,$7B,$E5,$C3,$FD,$75,$C5,$F8,$DD,$3F,$19,$CB,$FE,$31,$FA
	.dc.b	$4F,$D1,$61,$F8,$CF,$5F,$F8,$6F,$CC,$08,$43,$10,$8F,$E2,$6F,$FA
	.dc.b	$19,$FE,$2B,$BD,$8F,$E0,$93,$7C,$1E,$93,$F8,$F9,$E4,$6A,$71,$5E
	.dc.b	$27,$FB,$B4,$7E,$AA,$31,$1F,$97,$6C,$FE,$71,$CC,$59,$E8,$F8,$1F
	.dc.b	$71,$87,$D6,$F5,$30,$FF,$81,$63,$9B,$A8,$F3,$D7,$C3,$8F,$8B,$5C
	.dc.b	$63,$11,$F2,$27,$FC,$BD,$5C,$A3,$C5,$7F,$42,$7F,$D0,$C1,$FD,$F3
	.dc.b	$FB,$37,$F8,$3F,$97,$FE,$B2,$7F,$84,$ED,$67,$8C,$FC,$B9,$FE,$1B
	.dc.b	$1D,$0F,$DF,$B8,$7F,$C0,$6D,$9E,$D1,$B3,$FC,$79,$1B,$70,$7E,$1C
	.dc.b	$B3,$FC,$46,$3F,$D9,$13,$67,$62,$7F,$8B,$CF,$F2,$F9,$B3,$A3,$47
	.dc.b	$57,$F8,$91,$FC,$A2,$3E,$34,$FF,$A1,$FC,$EF,$E5,$3A,$3D,$87,$C5
	.dc.b	$3E,$79,$F1,$5F,$E2,$3F,$92,$C6,$9F,$E4,$C3,$FC,$76,$7E,$47,$1E
	.dc.b	$6D,$3F,$D7,$18,$FF,$8A,$1E,$1A,$7F,$34,$CB,$4D,$CF,$F9,$45,$F9
	.dc.b	$25,$F1,$F8,$BE,$71,$E6,$E2,$FE,$63,$5B,$D7,$CB,$5F,$CF,$23,$1A
	.dc.b	$FF,$B9,$E3,$E0,$1C,$A7,$DD,$E1,$FF,$41,$AD,$ED,$FE,$41,$AF,$E9
	.dc.b	$F1,$8F,$97,$E9,$1C,$70,$F8,$BC,$EC,$B7,$39,$C3,$E4,$47,$83,$6F
	.dc.b	$81,$9F,$CD,$9F,$99,$EA,$E6,$FF,$74,$7C,$6E,$0F,$A9,$5F,$06,$F8
	.dc.b	$E5,$FE,$E2,$D3,$CF,$7E,$49,$CE,$3F,$E5,$9E,$9F,$1C,$63,$4F,$F4
	.dc.b	$A5,$FF,$5D,$BF,$A3,$E1,$78,$23,$F0,$7E,$97,$D4,$D1,$F0,$17,$8B
	.dc.b	$E5,$61,$D2,$7C,$A7,$D2,$7E,$67,$D1,$DC,$FC,$7D,$7E,$5A,$F5,$EF
	.dc.b	$F2,$39,$FF,$BA,$E1,$EF,$C7,$F2,$3E,$07,$F2,$3F,$80,$67,$FE,$B1
	.dc.b	$DB,$EF,$4F,$5C,$BE,$2A,$F3,$7E,$D6,$3F,$76,$3F,$81,$3E,$1C,$7A
	.dc.b	$F6,$FF,$A5,$F0,$E6,$BD,$CD,$73,$7F,$61,$86,$9F,$77,$0F,$B5,$9F
	.dc.b	$82,$73,$A7,$C3,$CB,$AD,$FA,$3C,$4D,$C6,$25,$F4,$18,$CB,$B6,$5E
	.dc.b	$0D,$7E,$03,$CE,$3F,$D1,$7A,$FE,$77,$C2,$F8,$18,$FC,$6C,$BD,$FC
	.dc.b	$7F,$8C,$C7,$E4,$6D,$C5,$F9,$5F,$49,$77,$F3,$F4,$F8,$7E,$91,$E3
	.dc.b	$7F,$99,$BD,$87,$A5,$9E,$1E,$B9,$FE,$4A,$FF,$CC,$71,$E5,$78,$BF
	.dc.b	$21,$E7,$F1,$D7,$0F,$F3,$8D,$6D,$ED,$5E,$17,$C1,$D9,$E3,$DF,$BF
	.dc.b	$07,$F1,$2C,$7C,$BE,$D7,$33,$E9,$61,$F8,$7E,$44,$EB,$57,$C4,$FC
	.dc.b	$DE,$DF,$8F,$4F,$2E,$DB,$D2,$EF,$3C,$B6,$FE,$53,$97,$97,$36,$7F
	.dc.b	$8D,$8F,$E4,$1E,$D7,$FB,$33,$DC,$F8,$FF,$38,$F8,$CF,$F7,$A7,$A6
	.dc.b	$8F,$E1,$F5,$FE,$A1,$ED,$61,$FE,$A9,$E3,$FD,$83,$7C,$2F,$E5,$15
	.dc.b	$D6,$3F,$90,$3D,$F9,$6D,$7F,$AD,$BA,$2E,$9F,$63,$0F,$D9,$5F,$96
	.dc.b	$7C,$58,$7E,$92,$31,$97,$FB,$92,$3F,$8D,$C6,$3E,$6F,$58,$F5,$AF
	.dc.b	$67,$97,$A8,$C6,$CF,$80,$C6,$9F,$0A,$31,$AF,$F2,$2C,$BB,$C7,$C4
	.dc.b	$CF,$E1,$E1,$CB,$0F,$F3,$7E,$B8,$C7,$94,$7F,$9C,$BE,$77,$79,$7C
	.dc.b	$4B,$F8,$0F,$B9,$F3,$5F,$A2,$FC,$58,$79,$36,$FA,$59,$79,$FD,$1F
	.dc.b	$26,$1E,$1E,$16,$5F,$D1,$38,$FE,$B6,$3F,$ED,$DE,$EF,$AB,$BF,$F9
	.dc.b	$C2,$3C,$5F,$15,$E3,$D7,$F0,$91,$E1,$EC,$FA,$7B,$3E,$1B,$CD,$5F
	.dc.b	$17,$BF,$C9,$CB,$F8,$7F,$43,$CD,$D5,$FD,$BF,$E3,$7F,$DE,$3F,$0D
	.dc.b	$F0,$F2,$FD,$7E,$9E,$8C,$BF,$6B,$DD,$D2,$7F,$73,$D5,$F0,$5F,$DD
	.dc.b	$F9,$57,$2F,$EF,$5E,$17,$6F,$3B,$FC,$CB,$97,$E6,$BA,$3F,$15,$5E
	.dc.b	$1E,$BE,$D8,$FF,$04,$B6,$CE,$4C,$DF,$9F,$6F,$F3,$AE,$5C,$DF,$E1
	.dc.b	$31,$9A,$F5,$65,$E7,$89,$87,$94,$FE,$DD,$7E,$F8,$F8,$5D,$1D,$A7
	.dc.b	$C3,$85,$C0,$FF,$75,$EF,$FD,$5F,$3F,$E7,$F0,$FE,$97,$8F,$1E,$2F
	.dc.b	$81,$87,$0F,$97,$FD,$8F,$C4,$FE,$CB,$E4,$7F,$03,$F3,$6F,$49,$C0
	.dc.b	$F9,$BF,$E8,$3F,$38,$F1,$B8,$EA,$C7,$FB,$BA,$CB,$A7,$1F,$F7,$A7
	.dc.b	$B7,$C9,$59,$E1,$3F,$C6,$9F,$F6,$84,$F0,$F8,$FF,$F3,$D3,$CB,$5D
	.dc.b	$4F,$FB,$82,$31,$AF,$FB,$35,$BA,$FF,$B4,$9A,$F4,$9E,$F5,$ED,$8F
	.dc.b	$F1,$C4,$73,$33,$5F,$03,$5F,$B7,$1E,$EC,$FE,$2C,$E7,$D6,$FF,$73
	.dc.b	$E3,$E0,$BF,$EC,$BE,$CF,$3B,$DE,$3F,$27,$D7,$70,$67,$66,$BE,$3C
	.dc.b	$7E,$6D,$F7,$9F,$CF,$EB,$D8,$F4,$9E,$F3,$D7,$E2,$DC,$BB,$F6,$7B
	.dc.b	$DF,$0E,$16,$FB,$EF,$ED,$D3,$A5,$69,$FC,$3E,$DE,$B7,$CF,$87,$DE
	.dc.b	$F9,$8F,$BB,$3F,$87,$8F,$B5,$5F,$13,$2F,$A1,$9F,$CA,$F8,$6F,$E7
	.dc.b	$AC,$1F,$1D,$F8,$FF,$9D,$7C,$E7,$F4,$5F,$85,$CF,$AC,$7F,$87,$57
	.dc.b	$F4,$6D,$7F,$1A,$F1,$C3,$FD,$7C,$F3,$E2,$B1,$F9,$3C,$7A,$BE,$13
	.dc.b	$C7,$3E,$1C,$7F,$8B,$C7,$96,$31,$E4,$EE,$3F,$AA,$C6,$3C,$FC,$3A
	.dc.b	$CB,$F2,$6D,$3B,$16,$3A,$4F,$8E,$FD,$D3,$FE,$8D,$8F,$19,$9B,$F8
	.dc.b	$11,$FE,$91,$AC,$EC,$C7,$E3,$D6,$B4,$F8,$D8,$79,$8E,$59,$7C,$A3
	.dc.b	$FA,$E6,$3F,$BD,$8F,$F3,$5B,$FB,$FF,$1A,$C7,$97,$93,$E0,$3F,$58
	.dc.b	$E6,$7B,$0E,$BC,$17,$37,$D2,$F2,$7C,$D1,$F4,$9F,$47,$CF,$5D,$7E
	.dc.b	$FF,$9D,$DB,$3F,$95,$F4,$1C,$E7,$DB,$AF,$E1,$B4,$F9,$38,$FC,$1E
	.dc.b	$1F,$93,$F0,$38,$BD,$E3,$FE,$B7,$C1,$A7,$F8,$33,$E7,$FF,$C1,$BF
	.dc.b	$00,$F8,$BF,$F2,$CD,$FF,$8F,$3D,$57,$C7,$7A,$B6,$E1,$C7,$FE,$BA
	.dc.b	$3E,$1E,$3C,$B2,$FF,$49,$BD,$3E,$6A,$CB,$3E,$DC,$BF,$2A,$15,$E2
	.dc.b	$D3,$FC,$D9,$1D,$0E,$0F,$F9,$B3,$D3,$C5,$F9,$FB,$FE,$95,$7E,$66
	.dc.b	$F9,$47,$C5,$8C,$1C,$9F,$D6,$FA,$5F,$99,$9D,$EF,$EF,$AF,$D9,$46
	.dc.b	$25,$BF,$E8,$A3,$0B,$E7,$77,$5D,$6F,$E4,$70,$7E,$02,$3E,$3E,$5F
	.dc.b	$D0,$30,$F7,$47,$C1,$7C,$EF,$B3,$4F,$F4,$4E,$F7,$4F,$3E,$7F,$4A
	.dc.b	$BF,$83,$DF,$8B,$1F,$E6,$2D,$E9,$79,$3E,$0D,$FE,$8E,$9F,$1F,$A3
	.dc.b	$A6,$DF,$BC,$CB,$CB,$B7,$E0,$3D,$8F,$89,$1E,$BE,$1F,$F2,$EB,$AD
	.dc.b	$FE,$4F,$C0,$FC,$CC,$BE,$96,$3F,$87,$F0,$3E,$1C,$63,$0F,$93,$C0
	.dc.b	$F8,$DF,$4E,$5D,$78,$BF,$1B,$C3,$FE,$67,$E5,$FC,$9C,$FF,$B7,$3C
	.dc.b	$EF,$F3,$0F,$4F,$1C,$21,$25,$3A,$78,$1F,$C9,$6A,$D7,$86,$3F,$21
	.dc.b	$3D,$2B,$FA,$CC,$7F,$21,$8C,$46,$A3,$F9,$73,$E9,$AF,$E0,$33,$A8
	.dc.b	$B3,$A7,$4F,$33,$0F,$85,$FE,$7D,$1C,$63,$C3,$1E,$A3,$6E,$DF,$F4
	.dc.b	$0C,$6E,$3C,$ED,$79,$C7,$07,$C2,$F3,$AF,$8F,$1F,$1E,$30,$E0,$F3
	.dc.b	$D7,$D0,$8F,$E7,$99,$FF,$A4,$23,$FD,$23,$F3,$5F,$E1,$87,$F8,$3F
	.dc.b	$87,$FB,$9A,$FF,$1D,$D7,$EF,$5F,$DE,$4F,$DA,$9B,$19,$9A,$7F,$60
	.dc.b	$8F,$E5,$73,$E1,$8B,$3C,$CF,$F5,$0C,$FF,$4F,$9F,$34,$64,$FE,$D2
	.dc.b	$E1,$FC,$B9,$E4,$3F,$B6,$4F,$F5,$D8,$D4,$7E,$71,$B1,$FD,$C5,$CC
	.dc.b	$7F,$AF,$1B,$1D,$A3,$C0,$7F,$77,$9F,$8C,$FF,$4E,$3D,$FE,$27,$CD
	.dc.b	$C3,$F3,$11,$FE,$60,$8F,$94,$9F,$C0,$BE,$13,$11,$F3,$6F,$FB,$E3
	.dc.b	$A8,$EF,$3D,$1F,$88,$70,$8D,$4F,$07,$F1,$51,$E9,$9F,$D4,$CF,$07
	.dc.b	$F0,$B1,$F1,$4E,$53,$F8,$A7,$BB,$A7,$93,$F8,$C9,$F4,$4F,$BE,$73
	.dc.b	$39,$78,$BF,$8F,$6B,$99,$FD,$C4,$7E,$40,$7D,$27,$D2,$7E,$86,$5F
	.dc.b	$71,$F8,$CF,$D2,$69,$93,$E9,$BC,$67,$AE,$1F,$8C,$8F,$F1,$7E,$36
	.dc.b	$BF,$24,$7C,$18,$FC,$3B,$4F,$BC,$76,$38,$94,$E2,$78,$0F,$F1,$54
	.dc.b	$65,$CD,$7B,$9E,$A7,$FA,$8A,$7F,$B1,$57,$FB,$FB,$0F,$E0,$9F,$23
	.dc.b	$AC,$3F,$D0,$F1,$FC,$1F,$0F,$F1,$B3,$FC,$27,$E9,$7F,$91,$3D,$71
	.dc.b	$FF,$1A,$3F,$CA,$27,$F6,$CF,$F1,$F8,$FE,$E0,$FF,$21,$8B,$1F,$C6
	.dc.b	$4F,$F0,$EC,$7F,$22,$8B,$1F,$C6,$8E,$A7,$F1,$C6,$E3,$F8,$58,$FF
	.dc.b	$2D,$CF,$95,$FF,$2E,$3E,$79,$B1,$FC,$7A,$BF,$8F,$4F,$AC,$FF,$45
	.dc.b	$94,$E7,$1B,$7F,$96,$9F,$19,$FE,$8E,$FF,$2B,$F9,$3F,$8B,$1F,$CA
	.dc.b	$B1,$FE,$B3,$3F,$33,$1F,$0E,$3F,$C9,$27,$F8,$6D,$7D,$13,$81,$FC
	.dc.b	$46,$BB,$2C,$B5,$5F,$BD,$9F,$E9,$87,$F3,$59,$EA,$FE,$85,$DB,$FA
	.dc.b	$16,$B9,$3F,$9D,$1E,$E8,$E9,$97,$F2,$1C,$3F,$C4,$C7,$F2,$3D,$7F
	.dc.b	$98,$3F,$9A,$C3,$38,$7F,$8E,$A2,$BF,$CC,$E3,$FC,$AF,$1F,$98,$2C
	.dc.b	$7F,$3F,$8F,$D8,$3D,$CF,$CC,$9F,$CB,$9A,$FF,$2F,$8A,$7F,$A1,$E3
	.dc.b	$F2,$D1,$F9,$C3,$91,$FA,$82,$9F,$A8,$29,$FB,$08,$C3,$FB,$06,$BF
	.dc.b	$CB,$5F,$F5,$2D,$FC,$08,$FE,$39,$87,$F5,$99,$FF,$54,$EB,$99,$FE
	.dc.b	$59,$8F,$C2,$8F,$E4,$31,$E1,$7F,$8E,$36,$B2,$FF,$4F,$8F,$F0,$74
	.dc.b	$FC,$47,$FB,$44,$6E,$78,$D7,$81,$CC,$71,$7F,$96,$BB,$7F,$9B,$39
	.dc.b	$72,$D7,$FB,$A4,$73,$7F,$94,$35,$FE,$6A,$D7,$F9,$CB,$F1,$23,$EE
	.dc.b	$BF,$DB,$5F,$C0,$DF,$F2,$37,$FB,$53,$FC,$9D,$FE,$FC,$FE,$81,$FB
	.dc.b	$AF,$F5,$E7,$F9,$39,$83,$FB,$29,$FC,$67,$0C,$3F,$DB,$E7,$C4,$FF
	.dc.b	$22,$3B,$9F,$83,$72,$FF,$26,$6B,$FD,$FC,$FE,$30,$7E,$C9,$CB,$FE
	.dc.b	$42,$6B,$FE,$02,$5E,$DA,$3F,$CE,$BE,$CF,$F9,$CF,$E2,$1F,$30,$F9
	.dc.b	$26,$CF,$9E,$E3,$97,$2C,$7F,$CF,$A7,$EB,$B2,$E9,$CC,$E9,$C4,$F6
	.dc.b	$3F,$57,$3A,$D3,$FC,$D5,$1F,$EA,$A8,$C7,$89,$A8,$F2,$BA,$C3,$C1
	.dc.b	$82,$FF,$D1,$5F,$29,$FE,$30,$C3,$8E,$CF,$84,$FE,$97,$18,$DE,$E7
	.dc.b	$D2,$BD,$FC,$1F,$E1,$EA,$D7,$05,$AD,$5F,$C3,$F7,$FF,$B0,$D6,$B8
	.dc.b	$5E,$CF,$A6,$F7,$87,$C8,$E2,$C6,$FE,$7E,$16,$BF,$3A,$B5,$EA,$7F
	.dc.b	$AF,$73,$E9,$83,$E0,$31,$97,$FA,$AA,$BE,$5E,$5C,$F1,$F8,$13,$EB
	.dc.b	$F1,$FD,$B3,$9D,$FE,$D6,$DE,$8A,$F2,$BF,$15,$EB,$3B,$D3,$D1,$E6
	.dc.b	$78,$3C,$4F,$E4,$FB,$7D,$CF,$9C,$F0,$F3,$7D,$DC,$7C,$38,$FD,$3F
	.dc.b	$87,$FB,$0F,$BF,$A7,$35,$F8,$DF,$8A,$F2,$BF,$8F,$D9,$A7,$CF,$BF
	.dc.b	$47,$CE,$DA,$D6,$DF,$D2,$73,$F9,$71,$8D,$DF,$35,$FE,$F2,$F9,$6F
	.dc.b	$F7,$9E,$2F,$2B,$E5,$73,$FE,$7E,$5F,$0C,$FA,$4F,$F7,$64,$7F,$89
	.dc.b	$F3,$A7,$FB,$BF,$97,$FA,$4C,$74,$E6,$FE,$86,$F6,$C6,$D7,$FB,$D7
	.dc.b	$2F,$07,$47,$FA,$0A,$B5,$CF,$6B,$57,$F7,$7A,$FF,$C3,$F5,$CF,$AB
	.dc.b	$F5,$D5,$FE,$70,$8C,$6D,$FD,$25,$B1,$F9,$03,$E0,$1D,$27,$8C,$7F
	.dc.b	$8D,$8D,$8F,$4F,$7E,$0F,$FB,$D3,$1F,$8B,$6A,$7C,$56,$B5,$F9,$58
	.dc.b	$F4,$E1,$FD,$66,$3B,$C3,$C1,$9F,$CD,$97,$D1,$FC,$AA,$FE,$0B,$5F
	.dc.b	$05,$61,$E3,$8C,$3F,$5F,$A2,$E9,$FA,$19,$F8,$A7,$B7,$C2,$B7,$E9
	.dc.b	$CF,$DF,$D9,$E7,$BF,$5E,$DF,$7B,$0F,$16,$5E,$AD,$7C,$DC,$7F,$7B
	.dc.b	$9F,$FC,$07,$7F,$EB,$4B,$FB,$79,$F9,$BD,$3B,$8F,$C2,$E8,$F0,$BC
	.dc.b	$FC,$3F,$AD,$7F,$08,$79,$1F,$45,$C7,$D2,$FF,$2F,$7F,$95,$FC,$0F
	.dc.b	$E3,$F4,$7C,$BF,$F8,$6B,$CA,$FE,$5E,$FF,$3C,$AE,$13,$A7,$FD,$E8
	.dc.b	$73,$23,$9F,$2B,$FB,$9C,$72,$3F,$DE,$38,$ED,$E7,$5F,$EF,$67,$06
	.dc.b	$E7,$FB,$CB,$BA,$E9,$7F,$DC,$A3,$FD,$B7,$5F,$EF,$9B,$D3,$FE,$F3
	.dc.b	$8F,$1B,$FE,$6F,$7B,$9F,$EE,$D2,$DF,$FB,$A6,$6D,$7F,$8F,$6F,$BB
	.dc.b	$FE,$A1,$6B,$FE,$AE,$6B,$FE,$B2,$58,$7F,$A7,$F6,$F9,$B8,$7F,$58
	.dc.b	$3D,$A7,$FB,$CA,$3E,$F3,$F0,$27,$8A,$9F,$EF,$CF,$27,$FB,$F3,$F3
	.dc.b	$A7,$FB,$63,$C5,$FF,$21,$3F,$1D,$FE,$E8,$EC,$FF,$77,$67,$97,$FC
	.dc.b	$84,$66,$7F,$C8,$4F,$DE,$9F,$F0,$23,$4F,$F5,$B4,$D9,$FF,$27,$D7
	.dc.b	$A1,$F6,$C7,$B6,$7F,$C2,$D1,$4F,$F5,$2F,$92,$F7,$7F,$C0,$1E,$63
	.dc.b	$D2,$7C,$A3,$E5,$7F,$92,$68,$F9,$47,$C4,$3E,$6D,$9F,$3D,$7F,$DF
	.dc.b	$07,$E2,$9C,$67,$F9,$26,$E7,$FC,$5E,$7F,$25,$18,$F8,$EF,$AA,$EB
	.dc.b	$4F,$EA,$93,$F9,$D8,$C7,$83,$D7,$6F,$4E,$1E,$8D,$B7,$7F,$E1,$BB
	.dc.b	$E5,$1F,$CF,$63,$95,$70,$C3,$DD,$E4,$FF,$59,$9E,$13,$9A,$F8,$31
	.dc.b	$FB,$7B,$E7,$5F,$8E,$E1,$C5,$EE,$3F,$23,$E5,$E2,$DA,$FC,$DC,$BB
	.dc.b	$F0,$FC,$BE,$DE,$F5,$AD,$7D,$58,$FE,$6B,$89,$F1,$BA,$CB,$96,$FF
	.dc.b	$52,$7E,$9F,$91,$F4,$B6,$FB,$79,$7E,$78,$C1,$FE,$7A,$3E,$6B,$F2
	.dc.b	$1A,$78,$75,$B8,$79,$F4,$F9,$D3,$9C,$6C,$F8,$A7,$EC,$DF,$C1,$D7
	.dc.b	$F0,$D7,$E7,$D7,$D7,$1F,$89,$F2,$31,$97,$C2,$D3,$F1,$77,$E6,$CB
	.dc.b	$DB,$B7,$C1,$E5,$FF,$1C,$71,$7D,$0C,$BF,$25,$1F,$E5,$18,$FC,$5F
	.dc.b	$4F,$A3,$4F,$C1,$79,$DE,$4C,$BD,$18,$FF,$4D,$7C,$3C,$1F,$DF,$FE
	.dc.b	$13,$F0,$71,$AC,$FF,$C2,$2E,$AB,$C5,$CF,$E5,$C7,$D7,$B7,$E8,$B3
	.dc.b	$C6,$FE,$BD,$9F,$1F,$8D,$7E,$B4,$3D,$7C,$67,$ED,$F6,$3F,$53,$C5
	.dc.b	$FE,$71,$EC,$7C,$6F,$F9,$B7,$E3,$9E,$A3,$E8,$DF,$27,$FE,$48,$F3
	.dc.b	$38,$F2,$FF,$1D,$3F,$53,$97,$3E,$DF,$C1,$75,$7E,$4A,$7F,$8E,$67
	.dc.b	$FB,$17,$F5,$95,$E1,$8F,$F1,$4C,$7E,$E6,$31,$E6,$7F,$5A,$8F,$DB
	.dc.b	$C6,$34,$D4,$74,$94,$7F,$8D,$63,$F5,$13,$BD,$38,$C6,$E7,$DB,$D9
	.dc.b	$E2,$8C,$4B,$E9,$31,$BF,$C4,$C7,$AC,$FC,$3E,$FE,$B7,$C3,$83,$E2
	.dc.b	$BF,$C0,$73,$F0,$47,$C8,$F5,$71,$3B,$E1,$7B,$B8,$CF,$B6,$DF,$43
	.dc.b	$0F,$9F,$EE,$33,$F9,$F5,$FA,$DC,$6F,$A7,$F3,$B4,$F9,$51,$F4,$8F
	.dc.b	$9F,$16,$7D,$51,$F2,$1E,$FA,$FF,$11,$8E,$BC,$3F,$81,$8F,$C0,$47
	.dc.b	$BA,$77,$97,$92,$37,$BF,$C4,$8F,$3F,$0E,$35,$F5,$6F,$FE,$8E,$F4
	.dc.b	$37,$EF,$78,$74,$FC,$77,$AB,$F5,$34,$F8,$3F,$1B,$67,$C5,$9F,$F3
	.dc.b	$2E,$6F,$C5,$57,$8B,$E4,$FC,$95,$AC,$FF,$D0,$55,$A8,$C7,$27,$F4
	.dc.b	$88,$C6,$FE,$7A,$FF,$50,$F9,$3F,$D2,$BE,$8B,$F4,$5E,$4F,$E9,$FA
	.dc.b	$7F,$C4,$7E,$CE,$7E,$07,$5C,$DF,$9F,$3F,$A7,$5F,$EA,$7A,$3F,$3E
	.dc.b	$FE,$AF,$DF,$B3,$FD,$EF,$1F,$D9,$3C,$0C,$78,$5F,$ED,$0D,$97,$E2
	.dc.b	$86,$7F,$56,$BC,$3C,$5F,$EE,$49,$CF,$AB,$FD,$6A,$F8,$5F,$F8,$EB
	.dc.b	$AB,$FA,$C4,$62,$7F,$CA,$58,$7F,$55,$8C,$6D,$FD,$AA,$30,$79,$2F
	.dc.b	$E0,$E5,$FC,$50,$F2,$C6,$6F,$FD,$33,$18,$33,$97,$96,$B8,$E1,$EA
	.dc.b	$9C,$E7,$F3,$6F,$8C,$BE,$13,$DB,$8F,$C8,$C7,$F4,$0F,$F7,$49,$FA
	.dc.b	$AD,$3E,$67,$93,$67,$C1,$7F,$9B,$C7,$BE,$BF,$1B,$4F,$F0,$3F,$A3
	.dc.b	$EE,$B6,$8F,$17,$93,$FC,$9F,$DB,$FA,$D3,$F3,$F9,$EC,$FE,$17,$87
	.dc.b	$E8,$6D,$F1,$FC,$FB,$7E,$2F,$77,$E8,$CF,$C7,$D3,$F8,$A6,$DE,$4E
	.dc.b	$3F,$C6,$FB,$5F,$DF,$78,$FE,$9C,$BE,$77,$CD,$AF,$C7,$C3,$F2,$DB
	.dc.b	$FF,$A6,$34,$66,$F6,$3F,$D3,$7F,$0E,$F9,$F7,$B1,$FE,$A2,$F2,$BF
	.dc.b	$D4,$79,$71,$C3,$F3,$47,$F9,$CF,$7F,$F2,$83,$CF,$E0,$BF,$BD,$46
	.dc.b	$32,$FF,$53,$F5,$7F,$96,$A3,$07,$F3,$0A,$FF,$3C,$B8,$CF,$FC,$47
	.dc.b	$1E,$68,$C6,$7A,$D3,$D3,$87,$A2,$3F,$9C,$56,$E7,$34,$BF,$67,$95
	.dc.b	$E7,$9C,$C7,$BF,$5F,$F1,$8D,$76,$7F,$C6,$17,$CC,$F8,$D1,$CF,$0F
	.dc.b	$8B,$ED,$7E,$3A,$7E,$26,$5E,$3B,$FF,$64,$FC,$D7,$FA,$9F,$0E,$97
	.dc.b	$FC,$CF,$C0,$E9,$7B,$8F,$E6,$C6,$27,$B5,$FD,$18,$FF,$70,$47,$2E
	.dc.b	$0F,$F6,$AF,$C9,$78,$23,$E7,$EB,$EC,$CF,$E3,$1E,$8D,$7D,$CF,$FA
	.dc.b	$BE,$7D,$D7,$6B,$E1,$EB,$EC,$AF,$44,$63,$2F,$C1,$57,$97,$0F,$91
	.dc.b	$97,$A3,$8B,$F0,$9C,$BF,$D3,$73,$FC,$07,$6F,$93,$3F,$17,$A3,$E4
	.dc.b	$DF,$F1,$FF,$39,$6F,$E6,$79,$9F,$56,$FD,$1C,$1F,$3B,$8B,$C7,$18
	.dc.b	$E0,$FC,$A7,$A9,$F8,$5F,$6B,$F3,$7F,$46,$F8,$33,$FE,$84,$FA,$27
	.dc.b	$CD,$3E,$2F,$FD,$F3,$F2,$9F,$EF,$AC,$3F,$47,$F0,$7E,$97,$9F,$1F
	.dc.b	$E6,$9E,$B9,$5D,$7A,$BF,$90,$7E,$F4,$6A,$BF,$3C,$F0,$CF,$FC,$31
	.dc.b	$3E,$73,$8A,$F8,$2F,$D0,$CE,$FE,$07,$FC,$81,$5B,$AF,$61,$EB,$7E
	.dc.b	$D4,$EF,$5F,$EE,$5C,$7F,$EA,$39,$EB,$69,$F8,$31,$E6,$32,$FC,$4E
	.dc.b	$8C,$61,$DA,$B8,$6F,$F1,$EF,$C5,$A7,$CC,$9C,$EE,$F0,$7C,$D5,$F1
	.dc.b	$70,$FD,$CB,$D2,$FE,$9B,$EE,$3C,$A7,$F0,$9D,$3D,$5E,$37,$B2,$3C
	.dc.b	$B9,$7F,$5A,$7C,$37,$F6,$E3,$ED,$6B,$8C,$FC,$6B,$C2,$FB,$97,$CB
	.dc.b	$4F,$BD,$3F,$17,$07,$D1,$F9,$F4,$FC,$3E,$1E,$59,$F8,$5D,$FE,$5C
	.dc.b	$7E,$5F,$BF,$67,$E3,$70,$BE,$6F,$D1,$18,$DB,$F1,$DD,$DF,$94,$F8
	.dc.b	$8F,$E2,$FE,$4F,$DF,$E5,$FF,$6D,$7D,$27,$FA,$9B,$47,$CA,$FE,$62
	.dc.b	$FF,$8E,$C7,$FA,$8F,$24,$7F,$20,$F0,$30,$FF,$22,$FA,$3B,$5F,$CE
	.dc.b	$F2,$FF,$30,$3F,$CF,$89,$7C,$9D,$75,$65,$91,$9F,$0F,$FA,$54,$FF
	.dc.b	$82,$67,$C8,$70,$D3,$F4,$15,$BB,$F6,$7C,$17,$3A,$DE,$1C,$AF,$FD
	.dc.b	$F1,$E8,$F6,$CF,$E1,$E4,$9F,$F7,$E7,$16,$70,$ED,$DC,$E0,$FE,$5D
	.dc.b	$AF,$7C,$38,$4E,$7D,$9F,$6E,$5D,$32,$F9,$B1,$FD,$7B,$83,$C7,$C5
	.dc.b	$F1,$B6,$FE,$CD,$A7,$9C,$FA,$C7,$C9,$F0,$30,$FA,$B8,$7E,$C6,$7E
	.dc.b	$DD,$7C,$59,$FD,$AA,$F0,$9E,$8D,$BF,$D7,$13,$AA,$F3,$4B,$E9,$3E
	.dc.b	$5F,$91,$FE,$73,$AF,$95,$F1,$1E,$07,$DD,$C7,$E5,$7E,$07,$CE,$DE
	.dc.b	$2F,$76,$5F,$D4,$72,$F3,$71,$FC,$4D,$1F,$3D,$E4,$E1,$FA,$57,$F1
	.dc.b	$F8,$BD,$F1,$8A,$FF,$7B,$F9,$D8,$E2,$F7,$C7,$E5,$7C,$46,$1F,$E2
	.dc.b	$FC,$FF,$C1,$1E,$BB,$E2,$9F,$1E,$F7,$9E,$D7,$DD,$F9,$37,$C6,$72
	.dc.b	$F8,$0F,$CE,$E5,$F3,$B3,$E9,$C1,$FB,$F8,$FA,$91,$8F,$8B,$FA,$71
	.dc.b	$C4,$F0,$FA,$7C,$6F,$FC,$07,$B3,$D6,$C5,$79,$B7,$FE,$62,$F1,$C9
	.dc.b	$5E,$9F,$A2,$EF,$87,$2B,$F5,$E3,$F9,$BC,$FC,$B7,$EF,$E0,$F7,$65
	.dc.b	$FA,$99,$CD,$FC,$2E,$0F,$F5,$24,$62,$BE,$27,$06,$F2,$ED,$18,$EF
	.dc.b	$FD,$B5,$6E,$FE,$4C,$7E,$97,$1F,$46,$5E,$19,$CF,$9F,$FD,$63,$1E
	.dc.b	$71,$C3,$8B,$FD,$81,$8F,$A3,$0F,$87,$BF,$C0,$F2,$EE,$5F,$66,$7E
	.dc.b	$09,$F1,$F4,$B5,$F1,$B6,$FB,$B7,$F0,$F3,$EF,$B7,$EE,$79,$3C,$58
	.dc.b	$F1,$D7,$C4,$FA,$30,$F8,$95,$F8,$3E,$EC,$67,$F0,$76,$78,$9F,$77
	.dc.b	$C1,$FF,$6B,$70,$7E,$13,$8B,$F0,$2F,$CE,$F3,$3C,$BC,$5F,$0F,$2F
	.dc.b	$99,$AF,$87,$E3,$EF,$1F,$A7,$5F,$A3,$5F,$3B,$8B,$D3,$E0,$FF,$25
	.dc.b	$F2,$1F,$49,$FE,$4C,$F8,$9F,$E8,$BE,$BB,$C1,$7C,$85,$F7,$9F,$34
	.dc.b	$C7,$8F,$07,$FA,$8C,$FE,$12,$E3,$7F,$DB,$47,$F2,$F8,$C7,$D0,$FE
	.dc.b	$71,$D6,$7F,$96,$9E,$9A,$78,$FE,$33,$9B,$E5,$F5,$FF,$B6,$DF,$07
	.dc.b	$FC,$AF,$5E,$B9,$D6,$DF,$E7,$48,$C1,$FE,$7F,$C3,$DB,$B7,$F9,$A3
	.dc.b	$0E,$55,$EF,$DB,$CD,$87,$2A,$F8,$3A,$72,$18,$72,$AF,$8D,$DF,$FE
	.dc.b	$8C,$C7,$9D,$FC,$81,$BE,$71,$E9,$7F,$2F,$D0,$F7,$E9,$E5,$D5,$AF
	.dc.b	$8F,$1E,$3F,$37,$AD,$BF,$81,$DD,$DB,$E5,$67,$F6,$7C,$3F,$8F,$3C
	.dc.b	$CF,$E7,$DB,$7C,$7D,$ED,$78,$B2,$F6,$5F,$2B,$F1,$7C,$2E,$9F,$26
	.dc.b	$9F,$E3,$FC,$7E,$3E,$7F,$D7,$DF,$7E,$DF,$22,$B7,$8F,$FA,$A9,$F9
	.dc.b	$1A,$7C,$7F,$2E,$EF,$EE,$AF,$8D,$A7,$C5,$BE,$3C,$1F,$91,$F8,$CC
	.dc.b	$69,$F4,$30,$E1,$3F,$93,$F1,$7F,$BE,$5F,$DF,$3D,$FF,$2B,$FE,$BE
	.dc.b	$F4,$0C,$21,$77,$CC,$F2,$38,$9F,$E1,$0E,$FF,$CB,$4F,$32,$C7,$F8
	.dc.b	$51,$FF,$07,$BF,$9D,$74,$7E,$88,$B1,$D2,$BC,$33,$F2,$67,$67,$27
	.dc.b	$F5,$8E,$0D,$3D,$DF,$54,$F7,$8C,$47,$B5,$FD,$E3,$C4,$FF,$11,$D7
	.dc.b	$BA,$BE,$0C,$6E,$7B,$9C,$9E,$AE,$23,$71,$D6,$79,$38,$3F,$CA,$11
	.dc.b	$FE,$2F,$D7,$FA,$B9,$F5,$33,$FC,$BC,$67,$6F,$EB,$6E,$E6,$E1,$FE
	.dc.b	$F4,$7F,$DE,$B3,$FE,$12,$9E,$93,$F7,$4F,$F1,$E3,$87,$FD,$24,$7E
	.dc.b	$0A,$70,$E8,$E8,$7F,$90,$E7,$FC,$6D,$3A,$9B,$3A,$34,$71,$3C,$0F
	.dc.b	$F9,$1E,$3A,$C7,$98,$FC,$71,$C4,$F2,$3A,$74,$D7,$F9,$61,$E7,$3F
	.dc.b	$C5,$F3,$F9,$A8,$F8,$F1,$FC,$2E,$FE,$71,$F9,$EC,$33,$1F,$CC,$71
	.dc.b	$F9,$2F,$DA,$3F,$C7,$95,$F8,$28,$F9,$51,$CE,$AB,$F8,$E7,$F2,$0F
	.dc.b	$57,$F0,$8F,$C1,$AF,$D3,$CF,$C6,$8E,$13,$FA,$87,$4E,$9A,$FE,$79
	.dc.b	$F8,$A7,$07,$F6,$71,$A7,$55,$F9,$87,$F2,$B1,$C2,$BE,$0D,$79,$67
	.dc.b	$F2,$73,$F1,$E3,$FC,$AF,$87,$FA,$18,$FA,$4F,$9E,$FF,$0A,$F8,$9F
	.dc.b	$C5,$3F,$75,$FB,$4E,$CC,$9D,$4F,$C6,$BF,$83,$8F,$9B,$1E,$C3,$FD
	.dc.b	$C5,$1F,$87,$7F,$2C,$D5,$87,$E2,$5E,$07,$F1,$78,$D3,$A3,$CC,$7F
	.dc.b	$A2,$E7,$CC,$FF,$9D,$E7,$2E,$4F,$31,$FE,$91,$7F,$8A,$3F,$BB,$B7
	.dc.b	$C5,$E6,$BF,$67,$1F,$C1,$70,$FD,$5B,$FC,$1F,$E4,$7F,$93,$63,$FC
	.dc.b	$1F,$0F,$DB,$CF,$F2,$BA,$B1,$D8,$FC,$7C,$7F,$25,$8F,$DC,$3F,$C8
	.dc.b	$A3,$27,$F1,$98,$B3,$F8,$58,$F0,$9F,$C6,$A3,$F6,$B3,$FC,$86,$3F
	.dc.b	$60,$FF,$23,$8B,$1E,$D3,$8C,$FF,$16,$AB,$5F,$C7,$9F,$F7,$24,$FC
	.dc.b	$49,$D4,$FF,$3B,$8F,$8D,$3F,$94,$9F,$E3,$B3,$FD,$1B,$0F,$E2,$91
	.dc.b	$88,$FE,$95,$F1,$9D,$9D,$65,$CE,$7F,$93,$D7,$EF,$9E,$13,$FD,$4A
	.dc.b	$3F,$8E,$C6,$23,$FA,$9B,$5E,$87,$EA,$DC,$3D,$4F,$E6,$18,$FF,$1B
	.dc.b	$CB,$F9,$7E,$5F,$CA,$34,$FF,$02,$BF,$CA,$A3,$F4,$2E,$DC,$E3,$FC
	.dc.b	$48,$FF,$05,$9F,$C5,$1A,$7F,$22,$8F,$C8,$3F,$C6,$5F,$CD,$BF,$EE
	.dc.b	$98,$FE,$32,$FF,$BA,$CB,$19,$3F,$20,$64,$A6,$4F,$51,$F9,$B2,$9F
	.dc.b	$9B,$29,$C4,$F6,$1F,$CA,$0F,$F2,$9C,$7E,$E1,$FF,$80,$1D,$1F,$B0
	.dc.b	$C7,$F9,$4E,$29,$4E,$5F,$F1,$5B,$BC,$6D,$FE,$76,$7F,$DC,$D3,$CE
	.dc.b	$BF,$93,$C7,$38,$C4,$7F,$AD,$9B,$5F,$EE,$A7,$31,$6B,$F3,$D5,$E3
	.dc.b	$3C,$EF,$F7,$17,$07,$57,$FB,$A5,$62,$3D,$31,$FA,$17,$8D,$FF,$24
	.dc.b	$6D,$7F,$78,$6B,$FC,$B1,$E9,$1F,$CB,$1E,$6F,$F4,$EA,$F8,$DD,$9F
	.dc.b	$7A,$7F,$96,$9F,$3B,$2F,$C6,$C7,$E0,$B0,$FF,$29,$BF,$84,$CA,$C7
	.dc.b	$E3,$6B,$D0,$74,$7F,$1C,$E0,$FB,$31,$F8,$F7,$E0,$CF,$85,$FE,$66
	.dc.b	$D7,$FB,$FB,$63,$AB,$98,$FE,$54,$7C,$D7,$11,$F2,$5D,$46,$8F,$09
	.dc.b	$F9,$12,$9F,$91,$29,$F9,$17,$8B,$FA,$DF,$9A,$FE,$7B,$E3,$38,$D8
	.dc.b	$FF,$30,$F8,$47,$A6,$F7,$5E,$0F,$FA,$17,$1F,$E8,$BF,$03,$FE,$7C
	.dc.b	$D5,$1D,$3D,$BF,$F1,$5B,$DB,$3B,$BB,$5F,$01,$FD,$8A,$75,$AB,$97
	.dc.b	$E8,$E3,$D2,$F8,$61,$E6,$E5,$6F,$FD,$3F,$1E,$31,$8E,$DF,$DE,$46
	.dc.b	$27,$D7,$C3,$FE,$3F,$8C,$70,$FF,$BD,$E3,$FD,$91,$18,$F7,$F5,$97
	.dc.b	$83,$5F,$89,$7F,$0F,$C2,$FF,$20,$5F,$2E,$3F,$8B,$8F,$2C,$38,$69
	.dc.b	$E6,$C3,$E5,$71,$FF,$12,$C3,$E8,$DF,$C4,$E1,$77,$5E,$0F,$A7,$3F
	.dc.b	$0A,$BE,$2F,$35,$CF,$AC,$FC,$17,$BC,$5C,$3E,$E6,$DF,$6E,$7E,$19
	.dc.b	$E4,$E0,$FB,$BD,$9E,$EA,$DE,$DE,$7D,$F1,$E2,$78,$F0,$FA,$C7,$B3
	.dc.b	$E0,$FB,$6D,$F4,$F2,$7B,$FD,$B2,$E7,$1F,$8A,$C7,$DB,$B7,$C5,$9C
	.dc.b	$F1,$7E,$37,$D2,$F9,$79,$F9,$63,$E3,$F3,$7E,$1F,$E9,$3F,$1F,$F0
	.dc.b	$CB,$7A,$0F,$4D,$F1,$8F,$8A,$FC,$C7,$C4,$71,$D1,$E0,$FF,$B4,$B6
	.dc.b	$69,$F9,$A9,$FD,$E6,$FB,$A7,$C6,$62,$F7,$5C,$B6,$F5,$DF,$F1,$D9
	.dc.b	$CD,$A7,$CD,$3F,$C2,$A7,$31,$E9,$AC,$61,$EC,$DB,$D7,$7C,$E3,$1B
	.dc.b	$7E,$56,$3F,$A1,$61,$CA,$BE,$0E,$DB,$C3,$C9,$18,$F4,$BF,$99,$D6
	.dc.b	$F0,$F8,$DF,$1D,$8C,$BA,$E3,$F2,$3B,$3F,$9E,$5E,$E3,$E5,$F0,$62
	.dc.b	$B7,$C3,$F3,$7D,$7F,$1C,$62,$7F,$83,$6D,$F2,$77,$FA,$DA,$7D,$5C
	.dc.b	$3E,$A7,$2F,$F1,$0C,$3E,$DD,$FF,$2B,$CF,$E2,$F8,$36,$7E,$0F,$6F
	.dc.b	$B7,$8F,$E6,$71,$7C,$0D,$BD,$5C,$7F,$03,$5E,$D9,$F9,$E3,$1C,$1E
	.dc.b	$AC,$7C,$F1,$8D,$BC,$7F,$07,$F8,$5E,$6F,$C2,$6B,$FC,$73,$9F,$F0
	.dc.b	$0F,$E4,$3D,$EF,$26,$1F,$3B,$CD,$FB,$2F,$87,$DD,$F3,$72,$FC,$0D
	.dc.b	$7F,$27,$F2,$BE,$3E,$3F,$4F,$0F,$C3,$E7,$78,$FF,$DF,$77,$7C,$6F
	.dc.b	$F3,$4E,$6F,$D5,$F7,$BD,$A7,$BB,$AF,$79,$DE,$C7,$8F,$17,$E7,$BA
	.dc.b	$BF,$37,$18,$ED,$FC,$BC,$7F,$73,$8C,$67,$AD,$3C,$1D,$1F,$D7,$67
	.dc.b	$3F,$47,$67,$C9,$E2,$3C,$6F,$EC,$B5,$CA,$3F,$99,$F9,$FD,$6B,$7C
	.dc.b	$2F,$81,$C6,$DF,$D0,$74,$F6,$F1,$3C,$2F,$F0,$4E,$7C,$F0,$F8,$1F
	.dc.b	$0B,$8C,$B9,$DF,$C2,$F3,$BF,$C2,$57,$B8,$F8,$B3,$83,$7C,$BF,$33
	.dc.b	$C4,$F9,$D1,$81,$CD,$8D,$7C,$9C,$1F,$D3,$30,$E7,$2F,$1F,$E7,$5F
	.dc.b	$F3,$8E,$5F,$89,$87,$F5,$BD,$FE,$53,$ED,$33,$CD,$FC,$5F,$2B,$C3
	.dc.b	$FC,$9B,$C7,$F9,$38,$FC,$6C,$7E,$2D,$EF,$8B,$F0,$11,$F0,$71,$F2
	.dc.b	$C7,$E0,$F3,$C5,$6F,$5F,$C2,$DF,$C1,$BD,$C7,$AB,$6F,$4C,$63,$7F
	.dc.b	$C5,$7C,$A6,$30,$78,$7E,$A7,$C5,$AF,$93,$6F,$AF,$6E,$6E,$D5,$E1
	.dc.b	$8F,$E5,$7A,$7C,$B8,$C6,$FF,$26,$3B,$5F,$28,$7C,$0F,$F9,$77,$B9
	.dc.b	$F3,$FF,$ED,$05,$F4,$3F,$CA,$F8,$BF,$C2,$3F,$09,$F6,$3E,$07,$F2
	.dc.b	$55,$C3,$8B,$F9,$4E,$9F,$E3,$58,$C0,$EF,$FE,$7A,$E3,$1B,$97,$77
	.dc.b	$FD,$DF,$19,$F3,$B9,$4E,$7C,$7F,$56,$1F,$E5,$EA,$E5,$0F,$A0,$FF
	.dc.b	$23,$C6,$35,$FC,$E5,$6E,$F9,$47,$F8,$1A,$3F,$CD,$53,$BF,$59,$E7
	.dc.b	$EF,$5F,$EB,$BA,$7C,$19,$6F,$9B,$F8,$BA,$7E,$56,$73,$D3,$D3,$4E
	.dc.b	$F1,$C3,$87,$19,$74,$BE,$18,$FE,$8B,$C0,$FA,$58,$70,$8F,$D2,$FC
	.dc.b	$96,$2F,$86,$FF,$4B,$C0,$72,$7F,$8C,$B6,$F9,$F5,$F3,$0E,$DC,$3F
	.dc.b	$13,$2E,$97,$E2,$AF,$4E,$1C,$F0,$ED,$18,$DB,$DF,$5D,$BA,$BF,$BF
	.dc.b	$E0,$C9,$E6,$FF,$78,$CF,$31,$FE,$04,$F9,$4F,$16,$5F,$8A,$E7,$FB
	.dc.b	$73,$F0,$7C,$2F,$23,$EC,$E1,$FF,$4B,$57,$D9,$EA,$B9,$FC,$AE,$5F
	.dc.b	$A1,$CD,$F4,$FE,$26,$F3,$7C,$AE,$4F,$64,$FD,$7E,$6F,$E1,$BD,$DF
	.dc.b	$EA,$7D,$FF,$25,$C5,$FE,$BA,$F6,$59,$3E,$89,$EC,$FD,$EE,$6F,$E7
	.dc.b	$58,$70,$E6,$F7,$3C,$8A,$7C,$2F,$76,$E5,$FE,$C6,$AF,$F0,$DC,$F8
	.dc.b	$3E,$03,$FC,$EB,$7F,$E9,$98,$C6,$7F,$90,$8F,$F4,$FC,$63,$3D,$3F
	.dc.b	$3A,$BF,$CF,$B1,$B9,$FF,$5F,$CE,$FE,$7B,$8B,$BB,$F6,$E9,$EC,$AD
	.dc.b	$CE,$6F,$E0,$69,$CE,$7F,$6D,$39,$AF,$87,$5C,$33,$F9,$0F,$0D,$9A
	.dc.b	$7C,$DB,$F3,$61,$F0,$1E,$C7,$92,$F8,$E1,$74,$F5,$BC,$AB,$B4,$63
	.dc.b	$2F,$D7,$C7,$5A,$E3,$1F,$57,$CA,$FF,$7A,$61,$FC,$37,$E0,$3F,$25
	.dc.b	$7E,$AB,$F6,$3F,$EE,$19,$FA,$5E,$97,$96,$7F,$05,$CB,$F3,$9F,$C2
	.dc.b	$7C,$55,$CF,$CB,$F0,$3C,$70,$F8,$9A,$7F,$90,$DF,$AD,$AF,$E2,$37
	.dc.b	$FB,$BA,$78,$70,$F5,$47,$C6,$EC,$FF,$5E,$F6,$7F,$23,$F8,$1F,$E6
	.dc.b	$39,$7F,$0C,$F3,$7F,$CD,$5A,$BE,$8D,$F0,$FC,$36,$61,$FA,$FF,$83
	.dc.b	$FF,$4B,$C6,$38,$BF,$CD,$31,$89,$7C,$BE,$34,$F0,$79,$3F,$83,$E1
	.dc.b	$10,$B7,$C6,$3F,$47,$E5,$7F,$BC,$A3,$1B,$FE,$A2,$B8,$1C,$2F,$AC
	.dc.b	$7F,$26,$EE,$CC,$FA,$72,$F7,$F1,$7F,$6C,$8F,$06,$1D,$23,$FC,$5D
	.dc.b	$A7,$48,$C7,$0F,$FA,$36,$B7,$97,$6B,$FF,$1A,$7B,$9F,$A3,$BD,$AE
	.dc.b	$2D,$F0,$7C,$8D,$FE,$B6,$5F,$3B,$7F,$44,$E7,$7F,$A5,$87,$1E,$5C
	.dc.b	$EF,$E8,$C3,$EA,$F1,$FF,$89,$F4,$EF,$A7,$D7,$F8,$8F,$76,$BF,$3F
	.dc.b	$4F,$07,$67,$C2,$E1,$E7,$3F,$87,$7E,$2E,$1E,$ED,$FE,$07,$0F,$B7
	.dc.b	$4F,$C1,$E3,$C3,$83,$CB,$A7,$E1,$71,$FC,$26,$3F,$83,$C3,$F1,$9E
	.dc.b	$17,$87,$7F,$EF,$BD,$3E,$1B,$F2,$E3,$FE,$9B,$7A,$71,$FE,$C5,$F8
	.dc.b	$F5,$F5,$BD,$7F,$2E,$3E,$3F,$7B,$CB,$AF,$C7,$C3,$E8,$F3,$70,$E1
	.dc.b	$F9,$BC,$CF,$99,$C6,$BE,$8E,$6F,$F7,$6F,$80,$F8,$3F,$DC,$79,$1F
	.dc.b	$CC,$63,$F3,$5F,$13,$FE,$6B,$F6,$7F,$CD,$9E,$FF,$E9,$F9,$3F,$CD
	.dc.b	$78,$7E,$FF,$AE,$F0,$BC,$5F,$F0,$CD,$7F,$9A,$6B,$AF,$65,$AD,$69
	.dc.b	$FC,$FE,$BF,$CD,$35,$E2,$EE,$FE,$B9,$1F,$E7,$58,$C6,$BA,$BE,$35
	.dc.b	$FE,$7F,$E8,$D4,$62,$7F,$65,$7F,$E6,$4B,$E3,$4E,$4E,$B3,$FE,$87
	.dc.b	$9C,$C3,$7E,$18,$73,$BE,$11,$E4,$C3,$E4,$4E,$6B,$87,$C1,$FF,$A6
	.dc.b	$70,$FE,$45,$8F,$C4,$D7,$FC,$8C,$FC,$17,$FC,$7B,$77,$4F,$A1,$9F
	.dc.b	$6C,$B9,$4F,$5C,$FF,$D3,$9B,$79,$E3,$13,$FE,$91,$3C,$78,$3E,$8B
	.dc.b	$C1,$8B,$CA,$ED,$BB,$6F,$95,$7F,$E4,$E8,$FF,$54,$78,$BF,$2A,$31
	.dc.b	$8B,$BF,$18,$F9,$BB,$FF,$D2,$BC,$3F,$33,$E3,$AE,$DF,$0C,$F9,$38
	.dc.b	$F2,$E1,$FA,$1E,$77,$F1,$DF,$25,$F2,$DF,$E5,$3E,$07,$BD,$FD,$6B
	.dc.b	$5F,$E1,$1F,$2F,$FE,$04,$F4,$C7,$F8,$7A,$F8,$74,$7E,$AC,$C6,$BE
	.dc.b	$DE,$0C,$67,$DA,$3E,$07,$CF,$62,$FC,$19,$FE,$BF,$0F,$E5,$37,$BD
	.dc.b	$3C,$7E,$8F,$ED,$E3,$15,$FD,$BB,$A7,$F8,$54,$61,$FC,$FD,$F2,$C3
	.dc.b	$AC,$7F,$9E,$BB,$3F,$98,$D6,$F7,$FD,$A5,$EF,$4F,$A9,$8F,$C1,$F2
	.dc.b	$7F,$CC,$F3,$FD,$12,$F8,$43,$83,$86,$BF,$23,$3F,$F4,$3C,$E7,$3F
	.dc.b	$9D,$A7,$8B,$4F,$97,$E8,$E3,$5E,$F8,$70,$CB,$FC,$63,$D3,$DF,$0E
	.dc.b	$19,$77,$C3,$E8,$EB,$F3,$B4,$F0,$67,$D3,$8B,$FC,$A5,$B7,$C9,$EE
	.dc.b	$F9,$DB,$7B,$30,$F9,$59,$FC,$4F,$37,$E4,$BD,$A3,$D7,$C3,$F0,$79
	.dc.b	$2F,$0F,$97,$6F,$C0,$E7,$F7,$B4,$FF,$2F,$3F,$47,$BB,$9E,$9E,$CD
	.dc.b	$BF,$96,$4F,$C8,$F9,$EE,$71,$F8,$CD,$FE,$17,$07,$AB,$87,$E6,$18
	.dc.b	$DB,$E9,$71,$FF,$95,$31,$FE,$EB,$D0,$FA,$4F,$F4,$AB,$FE,$6F,$1E
	.dc.b	$A7,$FA,$63,$AD,$F1,$F0,$85,$9D,$F4,$FF,$27,$E6,$FC,$D4,$62,$BF
	.dc.b	$29,$1F,$DF,$23,$15,$AB,$EC,$EA,$7C,$B1,$88,$FE,$9D,$3C,$1D,$35
	.dc.b	$FE,$C4,$DE,$5F,$F0,$2B,$98,$B3,$CD,$F5,$CF,$8E,$31,$1E,$EA,$FF
	.dc.b	$0C,$57,$18,$FE,$B4,$D7,$33,$C1,$E7,$3F,$E2,$A8,$C4,$73,$9E,$0F
	.dc.b	$F9,$71,$FA,$18,$7F,$8C,$2B,$DB,$18,$EF,$FA,$59,$7D,$4E,$AF,$96
	.dc.b	$FD,$69,$B7,$C6,$7E,$2C,$FD,$AB,$E8,$7C,$27,$0F,$F4,$63,$DA,$E2
	.dc.b	$BE,$4B,$A8,$D1,$FE,$44,$71,$1C,$EB,$AC,$FF,$8F,$AB,$EE,$57,$18
	.dc.b	$CD,$7B,$8F,$F0,$B4,$F8,$2B,$8C,$FE,$48,$FC,$9D,$71,$7A,$1E,$18
	.dc.b	$DB,$BA,$E8,$FF,$84,$69,$F1,$2F,$8C,$FF,$77,$EC,$F6,$7F,$C1,$1F
	.dc.b	$4A,$F9,$87,$A8,$F6,$5E,$B3,$E6,$79,$E5,$FA,$1D,$7F,$9A,$69,$D7
	.dc.b	$7F,$F6,$AF,$AD,$8A,$EF,$AD,$9F,$F6,$CB,$FE,$86,$BF,$F3,$1D,$F1
	.dc.b	$DB,$C9,$F4,$9F,$E2,$98,$C5,$7F,$87,$79,$BF,$C7,$91,$8F,$4F,$F9
	.dc.b	$9E,$1F,$E6,$B8,$C7,$0F,$EB,$1B,$AF,$F9,$91,$CF,$0F,$BA,$B8,$6B
	.dc.b	$E6,$C1,$87,$C3,$E1,$78,$1A,$DF,$D3,$5A,$C7,$E3,$F4,$67,$83,$E6
	.dc.b	$4E,$70,$F9,$5E,$37,$BB,$3E,$D9,$F5,$E4,$7B,$19,$D3,$AC,$E7,$97
	.dc.b	$FD,$7F,$C1,$F1,$78,$3E,$56,$BF,$D2,$1F,$66,$5E,$97,$E2,$BE,$BC
	.dc.b	$EE,$3D,$27,$6F,$68,$B5,$E8,$C7,$F0,$55,$FE,$8E,$C3,$E3,$F1,$5C
	.dc.b	$7D,$3C,$1F,$86,$D7,$E0,$71,$FE,$2A,$31,$AF,$AF,$3E,$91,$8D,$3F
	.dc.b	$13,$AF,$BA,$37,$AE,$29,$F1,$3E,$EC,$7F,$CD,$1B,$73,$7B,$7A,$7E
	.dc.b	$3C,$79,$70,$FF,$BD,$78,$3E,$EB,$F9,$9E,$0F,$F5,$7F,$8D,$F7,$2B
	.dc.b	$FC,$A7,$F4,$6F,$A3,$FB,$BF,$1C,$F8,$07,$9E,$F9,$4F,$F0,$9F,$95
	.dc.b	$FB,$1E,$0F,$F5,$C6,$7C,$FC,$5F,$E6,$91,$DB,$93,$FA,$27,$99,$FE
	.dc.b	$17,$AD,$5F,$F9,$12,$BF,$D6,$B1,$8E,$8F,$F4,$33,$AD,$B5,$3D,$B0
	.dc.b	$FE,$9D,$69,$F4,$4F,$29,$CC,$7A,$FB,$B1,$C1,$EE,$E0,$FF,$18,$DF
	.dc.b	$FB,$23,$1E,$D1,$F9,$CF,$1B,$18,$F2,$8F,$87,$C5,$E1,$C3,$96,$2E
	.dc.b	$2F,$8D,$97,$BE,$73,$1F,$27,$CE,$C7,$03,$6F,$46,$5F,$3B,$8F,$E6
	.dc.b	$F5,$FF,$8F,$B5,$F0,$69,$D6,$7E,$8F,$CF,$E7,$4F,$ED,$1A,$FC,$6B
	.dc.b	$FE,$E1,$87,$CD,$7E,$2C,$66,$7E,$EB,$FE,$C1,$E0,$F9,$9E,$25,$AE
	.dc.b	$FC,$1F,$02,$FD,$37,$EF,$DB,$D1,$C5,$F2,$F8,$7F,$09,$E6,$FB,$38
	.dc.b	$7D,$7C,$1E,$7E,$7C,$5E,$F8,$7F,$13,$B7,$A7,$2F,$85,$AF,$95,$EB
	.dc.b	$F4,$5E,$3D,$BE,$B1,$F6,$23,$C3,$C7,$F9,$4F,$94,$FE,$DF,$E6,$FF
	.dc.b	$A6,$FE,$3D,$F0,$8F,$91,$73,$BE,$53,$FD,$41,$ED,$7F,$39,$F8,$07
	.dc.b	$A1,$FE,$70,$E7,$65,$FE,$1B,$CF,$A7,$6F,$F8,$51,$ED,$A5,$AD,$BF
	.dc.b	$C8,$BA,$BF,$C1,$73,$AF,$94,$FF,$08,$BC,$36,$FC,$CC,$FF,$8A,$A3
	.dc.b	$1B,$31,$FF,$08,$C6,$36,$B9,$76,$C3,$D7,$1F,$EB,$9C,$3E,$0C,$E7
	.dc.b	$0F,$73,$98,$F6,$DF,$38,$FF,$6F,$47,$85,$E1,$5F,$13,$A7,$1A,$7C
	.dc.b	$AC,$7B,$D7,$FB,$97,$CA,$CE,$BF,$37,$3E,$FC,$1F,$D8,$F9,$3B,$DF
	.dc.b	$CA,$3E,$36,$3F,$53,$D1,$F8,$71,$EB,$D7,$E3,$BE,$93,$DF,$3E,$1C
	.dc.b	$3D,$77,$EE,$8C,$6B,$F6,$B3,$FE,$A3,$C3,$F6,$67,$38,$79,$FD,$EF
	.dc.b	$1F,$8F,$F6,$E3,$EE,$F9,$BE,$4D,$FF,$0D,$D5,$E2,$C3,$E0,$76,$D9
	.dc.b	$F7,$72,$79,$1F,$81,$AF,$D7,$E0,$F2,$46,$3D,$DF,$F2,$B5,$7C,$FF
	.dc.b	$81,$CF,$2F,$CA,$E5,$71,$FC,$A8,$C7,$A4,$C9,$EF,$FF,$B5,$F8,$5E
	.dc.b	$83,$E0,$DF,$39,$FE,$71,$F3,$9E,$27,$F3,$5F,$98,$FC,$DE,$3C,$70
	.dc.b	$FF,$62,$65,$D3,$E0,$FF,$E0,$04,$7F,$48,$8F,$04,$EF,$E2,$3F,$D9
	.dc.b	$4F,$1C,$FA,$CF,$94,$87,$3D,$3C,$B1,$E4,$8C,$9D,$30,$EF,$1E,$8D
	.dc.b	$F3,$3C,$71,$F6,$6A,$F0,$F3,$97,$86,$3F,$D1,$DD,$FD,$23,$1E,$DB
	.dc.b	$6C,$69,$DA,$FE,$26,$FE,$6B,$DF,$03,$E4,$31,$7B,$C7,$FA,$D5,$BB
	.dc.b	$3D,$D8,$70,$87,$8D,$8B,$E1,$8F,$D0,$F2,$BF,$86,$F0,$7A,$F8,$BE
	.dc.b	$AE,$7D,$27,$CD,$3E,$FE,$2F,$9B,$87,$C4,$8F,$86,$71,$F8,$DB,$A7
	.dc.b	$C1,$F1,$BE,$1D,$79,$EB,$DF,$CB,$F4,$38,$7F,$D6,$5B,$E3,$97,$E5
	.dc.b	$EB,$FD,$D6,$FD,$F7,$B8,$F2,$47,$D5,$8C,$6F,$EB,$BC,$79,$5F,$88
	.dc.b	$D3,$E4,$F2,$F2,$CB,$AD,$FC,$1E,$57,$8E,$C1,$F4,$9F,$CC,$FA,$DF
	.dc.b	$21,$F7,$7E,$65,$F0,$CF,$87,$7A,$1F,$E7,$1C,$78,$F0,$FF,$32,$E2
	.dc.b	$F8,$A7,$5C,$FF,$9F,$73,$62,$B7,$E4,$7F,$3F,$AF,$97,$18,$D7,$F3
	.dc.b	$CD,$D7,$38,$79,$5F,$F3,$74,$67,$BF,$F2,$AD,$C7,$F3,$6E,$76,$B7
	.dc.b	$C6,$91,$FB,$CC,$7F,$4D,$5B,$DB,$F4,$ED,$C3,$38,$71,$BF,$87,$F4
	.dc.b	$18,$D7,$FA,$8E,$9F,$23,$3E,$B9,$7C,$7E,$4F,$E0,$D9,$F5,$E0,$F9
	.dc.b	$3A,$FC,$5C,$78,$ED,$F3,$F0,$F9,$DC,$1F,$D3,$70,$E0,$FF,$66,$7E
	.dc.b	$4F,$0F,$CB,$DF,$E2,$70,$7C,$27,$DD,$1F,$03,$1E,$BA,$F9,$34,$F4
	.dc.b	$70,$FF,$6F,$D3,$C1,$1F,$C2,$32,$EF,$7E,$5F,$2F,$F8,$9C,$7D,$8F
	.dc.b	$43,$C1,$3F,$03,$D0,$FA,$AF,$C2,$F8,$DB,$97,$5F,$3F,$C9,$3E,$2E
	.dc.b	$4F,$24,$F8,$75,$FA,$73,$F4,$37,$F9,$DF,$0F,$63,$D3,$CF,$F9,$2F
	.dc.b	$07,$D3,$8F,$CF,$8F,$CA,$F0,$3D,$9F,$9F,$F1,$CC,$3F,$9A,$FC,$DF
	.dc.b	$FA,$FF,$E7,$FF,$C1,$9F,$3B,$FE,$C1,$F9,$4F,$F1,$E7,$2B,$0E,$1F
	.dc.b	$1B,$F0,$1E,$5E,$1F,$F9,$93,$BB,$F5,$11,$8F,$92,$FC,$E3,$AC,$FF
	.dc.b	$CD,$38,$73,$D3,$C3,$EF,$FE,$7A,$B5,$F1,$7F,$E9,$FC,$7C,$F5,$E6
	.dc.b	$86,$DF,$9C,$AD,$BF,$D6,$6B,$F2,$F1,$8F,$57,$F3,$EE,$BD,$8F,$E8
	.dc.b	$33,$D3,$56,$9F,$06,$5F,$3D,$F9,$EB,$E1,$C3,$FC,$AF,$0E,$17,$B8
	.dc.b	$FF,$73,$46,$33,$F2,$E0,$9F,$F7,$57,$85,$9D,$3E,$6D,$3E,$5B,$17
	.dc.b	$B8,$71,$FF,$39,$E1,$F8,$DB,$7C,$FE,$0F,$AF,$A7,$C7,$7E,$CF,$8D
	.dc.b	$67,$D7,$85,$DB,$FD,$F5,$E0,$7C,$7B,$F9,$5A,$FF,$8A,$F1,$F0,$9E
	.dc.b	$F8,$CE,$9F,$EC,$FC,$FC,$38,$7B,$EB,$E6,$D6,$B7,$FF,$4F,$56,$B2
	.dc.b	$F8,$30,$7F,$95,$D7,$E1,$2B,$84,$7B,$A3,$19,$FA,$32,$F8,$BA,$FF
	.dc.b	$97,$B1,$E9,$18,$F8,$6D,$E7,$D7,$E7,$3D,$98,$71,$D7,$E8,$72,$FF
	.dc.b	$18,$F2,$3F,$3D,$EF,$FC,$9C,$DF,$0D,$08,$57,$F2,$86,$FD,$7F,$CC
	.dc.b	$F0,$FF,$43,$9B,$3C,$A7,$E9,$3C,$F8,$3E,$73,$D6,$2C,$E6,$7E,$74
	.dc.b	$F8,$27,$F2,$B5,$D6,$B5,$5F,$A0,$AF,$2C,$7E,$46,$30,$E1,$FA,$71
	.dc.b	$A8,$D3,$EA,$71,$1C,$1E,$75,$5E,$53,$B7,$94,$7F,$61,$C3,$E4,$D7
	.dc.b	$F7,$CC,$31,$83,$E4,$3F,$90,$D7,$CE,$F2,$FF,$AD,$3F,$42,$AD,$7F
	.dc.b	$A5,$2F,$F0,$67,$F5,$29,$EB,$3F,$D7,$9E,$F1,$A3,$ED,$4F,$16,$18
	.dc.b	$9F,$F1,$CD,$73,$9C,$C5,$3F,$C7,$31,$FD,$62,$BA,$D7,$F9,$56,$71
	.dc.b	$5D,$23,$FC,$B1,$5D,$5F,$F2,$0C,$D7,$99,$F0,$9F,$88,$D3,$FC,$43
	.dc.b	$19,$F0,$5F,$15,$FC,$A7,$E1,$5F,$41,$FC,$6B,$E1,$7F,$D7,$9B,$FF
	.dc.b	$82,$B3,$FF,$5C,$F6,$FC,$8D,$1E,$6D,$CF,$9E,$BF,$DF,$36,$FC,$EC
	.dc.b	$76,$F1,$F1,$8F,$FB,$0A,$3C,$3E,$06,$31,$FF,$65,$6F,$FC,$EF,$BB
	.dc.b	$E3,$56,$B8,$3F,$95,$67,$E9,$AF,$47,$D0,$FC,$6C,$38,$E2,$BF,$5F
	.dc.b	$B5,$F4,$F1,$E5,$C5,$FC,$DB,$2F,$EF,$BB,$79,$B3,$F8,$53,$FC,$EB
	.dc.b	$83,$3B,$79,$E7,$3D,$AE,$3C,$ED,$E0,$D7,$E4,$CE,$70,$F3,$6D,$E3
	.dc.b	$CB,$E6,$79,$3F,$47,$5F,$C6,$69,$F0,$F9,$EE,$7E,$A8,$F8,$3A,$FD
	.dc.b	$5A,$F7,$6F,$EA,$7D,$91,$E8,$E2,$B9,$F7,$D7,$CD,$CD,$F7,$FE,$27
	.dc.b	$AE,$3F,$80,$CB,$E8,$3F,$82,$E2,$B9,$7E,$0F,$8F,$BB,$F1,$3B,$7A
	.dc.b	$F1,$FB,$B9,$BE,$0E,$1F,$33,$47,$C0,$FF,$AB,$7C,$4F,$E5,$5D,$5F
	.dc.b	$EA,$FF,$80,$FE,$3B,$ED,$FF,$BA,$3B,$BF,$D6,$B5,$AE,$5F,$F4,$44
	.dc.b	$27,$5E,$EA,$B5,$7C,$BC,$1F,$F0,$35,$BD,$3D,$78,$DF,$E3,$8C,$B9
	.dc.b	$E5,$DC,$CB,$FE,$69,$33,$7C,$BB,$F3,$A7,$9F,$1F,$14,$F6,$F6,$33
	.dc.b	$7C,$39,$FF,$9B,$F0,$E3,$5F,$24,$7F,$3A,$EE,$C6,$3C,$A3,$F9,$EE
	.dc.b	$B8,$D7,$E1,$7C,$ED,$D3,$D7,$F2,$F6,$FF,$A1,$F6,$7A,$3D,$87,$7B
	.dc.b	$E5,$3F,$A3,$FC,$CB,$E7,$BF,$DA,$BD,$9F,$ED,$6E,$8F,$EB,$58,$70
	.dc.b	$BD,$D6,$BD,$1E,$98,$F3,$E1,$65,$CF,$E0,$7F,$55,$E8,$EB,$1E,$31
	.dc.b	$FA,$DF,$92,$C5,$F0,$CB,$C3,$F3,$78,$C7,$96,$DD,$36,$F1,$E9,$DE
	.dc.b	$73,$E1,$6F,$0F,$4F,$83,$74,$FD,$D7,$CC,$5C,$BF,$8A,$E5,$CB,$6F
	.dc.b	$1C,$3E,$97,$5A,$7F,$75,$ED,$71,$FF,$A9,$AB,$FC,$41,$EE,$FF,$8B
	.dc.b	$A7,$53,$9E,$3F,$F5,$59,$FE,$37,$CB,$F0,$D1,$9A,$71,$FF,$BD,$27
	.dc.b	$31,$F1,$8F,$93,$8F,$FB,$E6,$33,$8F,$83,$6E,$9E,$03,$3F,$2E,$1C
	.dc.b	$76,$E9,$39,$CF,$C1,$AF,$0B,$DE,$3E,$AC,$F7,$1F,$D7,$FE,$95,$F4
	.dc.b	$4F,$9A,$7C,$FF,$F7,$5A,$F8,$9E,$AA,$F7,$5A,$F5,$BB,$4E,$70,$E5
	.dc.b	$83,$E6,$FB,$CE,$6F,$AF,$CC,$F0,$C7,$9D,$FF,$9E,$F3,$F8,$5B,$FB
	.dc.b	$F0,$E1,$2F,$97,$E7,$7C,$3E,$83,$96,$9E,$0E,$9F,$9D,$F3,$5F,$D8
	.dc.b	$FA,$7F,$D0,$DF,$0B,$F6,$78,$BF,$82,$4E,$7E,$17,$FD,$BD,$5F,$D1
	.dc.b	$BA,$93,$9B,$FB,$37,$D0,$7B,$1F,$E6,$3C,$3F,$DC,$DE,$07,$1B,$E1
	.dc.b	$D9,$6F,$79,$78,$BA,$FE,$D4,$7D,$FA,$ED,$D5,$BC,$BA,$67,$FD,$62
	.dc.b	$BF,$8D,$7C,$2F,$E0,$7E,$39,$F2,$3F,$E5,$3E,$7E,$75,$F8,$9A,$D4
	.dc.b	$FF,$95,$EB,$E4,$4F,$E4,$67,$3F,$08,$9E,$DC,$29,$CC,$7F,$2B,$AE
	.dc.b	$06,$EF,$FC,$C5,$A7,$48,$F0,$19,$C3,$E4,$F4,$F9,$23,$D2,$66,$7A
	.dc.b	$3C,$2F,$D7,$5F,$42,$38,$63,$FB,$27,$C9,$B3,$E5,$7A,$47,$F4,$7F
	.dc.b	$03,$F4,$9A,$29,$F2,$BE,$5A,$FD,$65,$6A,$73,$A7,$2C,$7B,$61,$C7
	.dc.b	$0F,$2E,$1C,$0F,$0F,$89,$CD,$F0,$18,$BE,$58,$F5,$9E,$33,$AB,$F2
	.dc.b	$47,$A2,$3F,$AA,$4F,$EE,$FE,$33,$FA,$97,$C9,$67,$2F,$81,$1F,$D5
	.dc.b	$7C,$BF,$05,$F0,$11,$9C,$BC,$31,$8F,$63,$FD,$7F,$E9,$62,$BF,$AF
	.dc.b	$74,$72,$C7,$AD,$7F,$AE,$F0,$D3,$FE,$BD,$BB,$5E,$3F,$0B,$F6,$5D
	.dc.b	$5E,$9C,$BF,$B6,$47,$ED,$7E,$7B,$CB,$18,$3F,$CD,$4F,$E6,$5F,$EF
	.dc.b	$91,$BB,$E9,$1F,$E6,$FE,$2E,$39,$FF,$9C,$63,$19,$EA,$FE,$E5,$DB
	.dc.b	$F1,$5E,$FA,$9E,$4E,$2F,$FC,$C5,$ED,$74,$6E,$3E,$0A,$4F,$D6,$CB
	.dc.b	$CA,$7E,$F3,$6E,$0D,$AE,$11,$B8,$F7,$57,$F8,$25,$F8,$97,$FE,$71
	.dc.b	$9F,$8D,$1E,$57,$83,$6B,$6F,$7F,$12,$DF,$93,$4B,$8F,$C9,$AF,$E4
	.dc.b	$47,$F1,$07,$E7,$BF,$07,$2E,$F9,$7F,$27,$D2,$CF,$CD,$BE,$F3,$FA
	.dc.b	$5C,$AD,$72,$8F,$A9,$5F,$95,$7E,$E4,$F0,$D7,$53,$FE,$A5,$DB,$DB
	.dc.b	$C7,$F9,$9D,$9C,$1F,$85,$8C,$3F,$1D,$B8,$7E,$85,$CE,$F7,$0F,$CC
	.dc.b	$71,$6D,$B5,$9E,$2F,$D1,$55,$C5,$C7,$DE,$FF,$6A,$73,$E5,$FD,$7F
	.dc.b	$05,$E0,$FF,$10,$65,$79,$7F,$2F,$AF,$F0,$3E,$6F,$06,$1F,$C0,$6B
	.dc.b	$84,$6B,$1F,$E2,$3C,$FF,$A1,$AF,$E5,$3D,$5F,$BB,$6E,$7F,$C5,$1C
	.dc.b	$E1,$6F,$F9,$06,$D7,$AB,$F8,$F9,$F9,$1A,$FE,$5F,$59,$EB,$FE,$71
	.dc.b	$57,$AF,$FC,$A1,$87,$F2,$BC,$BF,$AA,$5F,$F2,$27,$FA,$4B,$98,$FE
	.dc.b	$9F,$E0,$3B,$BF,$92,$46,$7B,$3F,$A6,$F4,$7F,$4D,$C3,$FA,$F7,$81
	.dc.b	$6B,$8D,$FF,$60,$7F,$A6,$CF,$F6,$3E,$9D,$61,$FD,$5B,$C1,$FE,$67
	.dc.b	$E0,$7F,$77,$F1,$3F,$BD,$C7,$F4,$18,$C7,$87,$FE,$04,$F1,$3F,$D6
	.dc.b	$71,$8F,$1F,$FC,$1F,$C5,$BD,$3F,$C2,$97,$FE,$0A,$E0,$DC,$62,$BF
	.dc.b	$C1,$3C,$3F,$D8,$30,$E5,$E3,$7F,$83,$EB,$06,$D7,$93,$CB,$C3,$FF
	.dc.b	$05,$FC,$17,$7E,$DF,$F1,$B6,$37,$C5,$FF,$1A,$79,$1E,$09,$FF,$1D
	.dc.b	$78,$39,$F2,$FF,$C6,$97,$DA,$7F,$C7,$DD,$3B,$C3,$FC,$79,$E6,$6F
	.dc.b	$A3,$FC,$8B,$7B,$9F,$F2,$07,$37,$F6,$5F,$5B,$FC,$71,$3F,$E5,$DF
	.dc.b	$3B,$FC,$E6,$FF,$8F,$E3,$3E,$2F,$F9,$63,$6C,$71,$FF,$95,$AB,$86
	.dc.b	$1F,$EC,$78,$FF,$2F,$7A,$3F,$E6,$D8,$FF,$31,$C6,$35,$FF,$36,$F7
	.dc.b	$F2,$DF,$FC,$FD,$7B,$E0,$7A,$19,$CB,$AC,$7F,$9F,$B1,$E5,$B7,$F9
	.dc.b	$E1,$B9,$FF,$9F,$5B,$3F,$C1,$1B,$E5,$F3,$DB,$FC,$E3,$5E,$AE,$8F
	.dc.b	$F5,$47,$8D,$FE,$BD,$F4,$B1,$87,$FA,$9F,$57,$A9,$8F,$63,$FD,$4F
	.dc.b	$0F,$5F,$0F,$FA,$9C,$72,$E3,$1F,$F6,$47,$B1,$7D,$8F,$F6,$0F,$B3
	.dc.b	$7D,$96,$0F,$6B,$0F,$FB,$DF,$5F,$F6,$4B,$D7,$1C,$FA,$7A,$F5,$B7
	.dc.b	$EE,$BF,$24,$7E,$2B,$FE,$3E,$FD,$05,$FA,$33,$EF,$DF,$9D,$3F,$00
	.dc.b	$F0,$79,$AF,$CA,$F9,$63,$C7,$0E,$17,$BE,$17,$D6,$7F,$A0,$FE,$BB
	.dc.b	$FC,$F1,$C7,$FE,$85,$FA,$97,$EB,$0F,$B0,$BF,$B6,$3F,$23,$7E,$AA
	.dc.b	$FC,$E1,$97,$2C,$99,$72,$C7,$8E,$1C,$3F,$62,$7E,$76,$FC,$7F,$FC
	.dc.b	$B7,$EB,$BE,$E5,$FA,$53,$F0,$77,$E3,$AF,$AF,$7C,$1B,$F2,$DE,$79
	.dc.b	$72,$C7,$8E,$1C,$2F,$7F,$B1,$3F,$0E,$7D,$7F,$F9,$AB,$E0,$DE,$53
	.dc.b	$E7,$7F,$C8,$DE,$23,$F0,$97,$DE,$BE,$E3,$96,$3C,$70,$E1,$F2,$9F
	.dc.b	$56,$F7,$93,$F5,$E7,$DE,$7F,$9E,$3F,$44,$F8,$1F,$2E,$FA,$E7,$E4
	.dc.b	$AF,$CB,$DF,$A4,$B9,$39,$63,$C7,$A3,$FC,$E1,$87,$0B,$DF,$EE,$8F
	.dc.b	$C9,$5F,$5F,$FE,$86,$FD,$0D,$F8,$3C,$F6,$1F,$9E,$3F,$41,$FF,$5C
	.dc.b	$7E,$7F,$C7,$0E,$17,$BA,$D7,$B0,$CF,$EB,$7E,$2B,$FE,$92,$F6,$7F
	.dc.b	$20,$41,$FA,$23,$F1,$7D,$FA,$3F,$ED,$19,$D1,$F9,$79,$FC,$5F,$8E
	.dc.b	$BE,$93,$AF,$C8,$9F,$AF,$72,$C3,$8C,$FE,$7E,$5F,$8A,$E6,$33,$FA
	.dc.b	$DF,$FB,$7F,$1F,$CB,$5F,$86,$78,$F8,$FF,$8F,$F5,$FB,$C7,$93,$F3
	.dc.b	$2C,$56,$E3,$D9,$AF,$AF,$0E,$73,$E2,$9F,$7F,$B3,$C2,$31,$A7,$A2
	.dc.b	$B7,$5E,$88,$F3,$C6,$23,$E1,$E7,$FE,$F5,$FD,$A7,$F9,$4D,$FF,$BD
	.dc.b	$FF,$62,$FD,$07,$EC,$DF,$CB,$3F,$6C,$FF,$77,$C3,$F1,$75,$F0,$B4
	.dc.b	$F0,$F9,$C5,$F0,$F6,$9F,$8C,$7E,$EA,$C3,$27,$E3,$6F,$C0,$5F,$A8
	.dc.b	$3F,$09,$FF,$7F,$7E,$6B,$FE,$FB,$C7,$96,$3C,$70,$E1,$7B,$EE,$5F
	.dc.b	$F2,$FB,$B7,$E0,$FC,$D5,$F8,$13,$F0,$B7,$D4,$BF,$35,$FE,$03,$F4
	.dc.b	$9E,$6C,$78,$E2,$C7,$8E,$1C,$2F,$F8,$3F,$E0,$BB,$FC,$DB,$FD,$CF
	.dc.b	$F8,$A7,$F2,$CF,$BF,$7E,$71,$FB,$0F,$CE,$1D,$BB,$D3,$E1,$DE,$EB
	.dc.b	$5F,$A8,$E6,$53,$9F,$BF,$78,$AF,$CE,$F5,$39,$85,$FA,$7E,$47,$FD
	.dc.b	$F3,$F9,$6B,$F2,$3F,$E4,$1E,$67,$BF,$EF,$FF,$DF,$BC,$DF,$CB,$F5
	.dc.b	$FA,$B7,$BF,$23,$F3,$FC,$AB,$FF,$79,$57,$DC,$FB,$7F,$B6,$BF,$3F
	.dc.b	$FE,$E7,$F1,$4F,$B3,$7E,$7D,$FC,$3F,$F0,$5F,$99,$87,$0B,$DD,$6B
	.dc.b	$E0,$3F,$92,$E3,$F8,$8F,$C3,$BA,$FE,$80,$FC,$3F,$F1,$FF,$A8,$7C
	.dc.b	$4F,$D8,$5F,$8B,$BF,$79,$E1,$8F,$E3,$B0,$E1,$E9,$B0,$E1,$6F,$CE
	.dc.b	$FF,$A8,$7E,$79,$EA,$FC,$ED,$F8,$BF,$F9,$E3,$C8,$7E,$47,$FC,$E3
	.dc.b	$F0,$2F,$D0,$FD,$3B,$0E,$17,$BA,$D7,$67,$F5,$BF,$A6,$F8,$BC,$BF
	.dc.b	$DB,$FF,$44,$7D,$B3,$89,$F8,$E3,$F0,$1F,$D2,$EA,$C7,$8F,$DB,$B1
	.dc.b	$E3,$87,$0F,$9B,$DF,$E6,$7F,$C3,$3A,$3E,$9C,$3D,$AF,$ED,$93,$99
	.dc.b	$FD,$8C,$BF,$1A,$CC,$7F,$11,$3E,$DE,$9D,$1C,$1F,$7B,$F7,$6D,$BF
	.dc.b	$CC,$4F,$C0,$8F,$D3,$B3,$F1,$1F,$E0,$5F,$8F,$CC,$BA,$B3,$3F,$CC
	.dc.b	$63,$8F,$E7,$98,$AE,$79,$FC,$2F,$D3,$FF,$93,$DF,$F3,$FF,$C9,$FF
	.dc.b	$C0,$D1,$88,$FE,$87,$F9,$66,$31,$F3,$9F,$C3,$B4,$C3,$E7,$3F,$2F
	.dc.b	$A6,$2B,$CD,$1E,$AD,$31,$1E,$58,$FC,$EE,$98,$D3,$CB,$1F,$E1,$7F
	.dc.b	$D4,$BF,$43,$F9,$0F,$BA,$3E,$CF,$3A,$73,$0F,$C6,$F5,$3F,$E2,$B8
	.dc.b	$FE,$81,$8F,$F5,$BF,$1F,$FC,$45,$F8,$0F,$D7,$9F,$D6,$D3,$F7,$7F
	.dc.b	$81,$7F,$E3,$1B,$FF,$14,$46,$EB,$F1,$2F,$81,$FA,$35,$A7,$F9,$8E
	.dc.b	$5F,$CC,$B1,$F1,$FE,$F1,$B9,$FB,$75,$FE,$33,$8F,$4C,$EA,$78,$9F
	.dc.b	$73,$93,$F2,$FE,$A7,$18,$FE,$07,$DD,$FE,$3B,$8E,$31,$E2,$74,$FF
	.dc.b	$0A,$C1,$E8,$FB,$35,$FD,$0B,$EB,$9F,$F2,$23,$F0,$3F,$3F,$FD,$2F
	.dc.b	$D7,$EE,$7F,$C9,$0E,$EC,$EE,$3F,$C9,$9F,$51,$E0,$FC,$8D,$F9,$5E
	.dc.b	$27,$D6,$7E,$6A,$3F,$C5,$37,$FE,$2A,$8C,$9F,$C7,$BF,$1C,$7E,$6F
	.dc.b	$C6,$1C,$5F,$81,$97,$E7,$79,$8F,$C2,$71,$78,$30,$E9,$7C,$0F,$D2
	.dc.b	$C7,$83,$3F,$8B,$1D,$A7,$75,$F8,$D8,$EB,$1E,$A7,$F1,$3B,$FE,$BF
	.dc.b	$F3,$6F,$F3,$56,$BF,$86,$E9,$7E,$3B,$8B,$DD,$FF,$9A,$FF,$6A,$C4
	.dc.b	$7F,$83,$7E,$8B,$FD,$1B,$97,$58,$FF,$08,$7E,$BC,$CF,$78,$7F,$89
	.dc.b	$23,$AD,$6E,$3F,$C4,$FF,$9F,$31,$DC,$7F,$8C,$A3,$AE,$9F,$82,$FC
	.dc.b	$A7,$A5,$F7,$DB,$F5,$51,$D2,$7F,$0F,$EC,$73,$AD,$69,$FC,$32,$39
	.dc.b	$BF,$58,$FF,$06,$F1,$F4,$AF,$C7,$BC,$E3,$47,$EC,$F3,$F8,$38,$FD
	.dc.b	$6F,$C4,$FF,$D2,$5F,$7F,$FA,$EF,$C1,$FE,$93,$79,$B3,$87,$CA,$F8
	.dc.b	$2F,$B9,$3F,$B4,$7F,$12,$FE,$DB,$D0,$E3,$3F,$41,$FB,$31,$FB,$AF
	.dc.b	$CF,$F8,$CF,$D4,$78,$BF,$E2,$6F,$D0,$7F,$C4,$BF,$7B,$FE,$17,$FD
	.dc.b	$65,$FB,$4B,$EF,$F8,$DA,$F7,$F3,$8B,$DD,$7E,$73,$E2,$3F,$BB,$FE
	.dc.b	$A2,$FD,$43,$FC,$75,$8F,$F0,$E9,$CC,$7F,$A4,$A3,$F8,$4F,$93,$FD
	.dc.b	$FB,$E6,$5F,$A9,$BF,$3F,$FE,$37,$39,$87,$E3,$BF,$C7,$BF,$41,$FF
	.dc.b	$02,$FE,$1A,$FC,$FB,$FC,$D3,$F9,$DF,$FA,$C7,$F1,$5F,$F5,$2F,$EC
	.dc.b	$FF,$EC,$1B,$DD,$6B,$E2,$BF,$D7,$FF,$30,$FD,$3B,$FD,$8D,$F8,$17
	.dc.b	$F8,$43,$F5,$6F,$E5,$DE,$5F,$F2,$C8,$C7,$E1,$FF,$A7,$F2,$91,$BF
	.dc.b	$7F,$FD,$B1,$FA,$67,$C5,$FC,$65,$F5,$FF,$DC,$2B,$58,$AB,$5F,$95
	.dc.b	$FF,$74,$F7,$FF,$C3,$FE,$9F,$EC,$FB,$3F,$DB,$11,$FE,$22,$F4,$FF
	.dc.b	$D2,$B1,$FC,$C2,$31,$87,$FB,$9F,$D5,$FF,$69,$7E,$71,$FC,$86,$73
	.dc.b	$3F,$ED,$DF,$C3,$93,$FE,$03,$BF,$F7,$27,$AB,$FE,$2E,$FB,$BE,$B1
	.dc.b	$FC,$DA,$31,$7F,$EA,$4F,$3D,$FA,$C7,$F0,$DE,$4F,$F1,$2D,$FF,$9C
	.dc.b	$30,$FF,$54,$56,$BA,$3F,$C1,$FE,$B7,$F8,$EB,$F2,$1E,$B7,$FD,$42
	.dc.b	$31,$97,$FA,$53,$E0,$F7,$3F,$E7,$AA,$FF,$56,$FE,$08,$FC,$37,$32
	.dc.b	$FD,$69,$F9,$1C,$20,$F8,$DF,$F8,$01,$9F,$F9,$2A,$73,$3B,$FD,$A7
	.dc.b	$37,$FE,$88,$FD,$B9,$F9,$BC,$20,$FD,$83,$72,$9C,$CE,$FF,$66,$CF
	.dc.b	$C1,$76,$C7,$CB,$A6,$1C,$65,$CB,$E0,$78,$E1,$E0,$CF,$0E,$31,$EF
	.dc.b	$FA,$5F,$3C,$3D,$9C,$ED,$FF,$CC,$B3,$9C,$BC,$BF,$82,$F4,$C7,$D1
	.dc.b	$7B,$DF,$FD,$4B,$39,$CB,$D7,$F7,$3D,$76,$BA,$79,$B0,$E1,$83,$C6
	.dc.b	$E9,$B7,$C2,$79,$4E,$BF,$56,$E4,$DB,$E5,$BF,$FA,$1B,$5F,$7E,$1F
	.dc.b	$2B,$ED,$38,$CF,$F9,$FB,$F2,$8E,$2D,$BE,$39,$FF,$AF,$3D,$3F,$0E
	.dc.b	$5E,$47,$8C,$7F,$B0,$3D,$9E,$2D,$AE,$2B,$F4,$3E,$38,$7F,$A3,$1E
	.dc.b	$31,$FE,$57,$9F,$B1,$BF,$DE,$FA,$1F,$99,$97,$C2,$FD,$33,$8D,$FF
	.dc.b	$B3,$9F,$83,$F2,$5F,$E9,$6F,$27,$E8,$7E,$A8,$BF,$F1,$A5,$FF,$B5
	.dc.b	$27,$EA,$7B,$F9,$BF,$DF,$E7,$F8,$7F,$D8,$B1,$7F,$ED,$AF,$3B,$F0
	.dc.b	$1F,$65,$88,$7E,$DB,$DB,$3E,$B1,$8E,$7F,$C5,$E7,$BF,$63,$F1,$19
	.dc.b	$5C,$FF,$29,$97,$3F,$86,$E5,$AF,$FA,$B9,$F2,$E7,$D3,$F5,$FE,$2D
	.dc.b	$BE,$2F,$83,$F6,$AE,$3A,$FF,$BC,$1E,$38,$7F,$AE,$3F,$6F,$E2,$DB
	.dc.b	$E2,$FF,$38,$DB,$F4,$3C,$5E,$DF,$82,$E3,$8F,$FB,$EB,$F6,$FE,$2D
	.dc.b	$C3,$8B,$FE,$FE,$FD,$6B,$C2,$9E,$2B,$C5,$E2,$D9,$E2,$FF,$8B,$B1
	.dc.b	$F0,$EB,$F2,$BE,$73,$8F,$E2,$7C,$0F,$A3,$1F,$3B,$E4,$7B,$9D,$BF
	.dc.b	$16,$EA,$6B,$7F,$F6,$BF,$EC,$BE,$43,$C7,$F7,$7E,$23,$59,$87,$ED
	.dc.b	$5E,$A3,$D3,$F9,$5F,$A0,$DC,$3F,$1C,$F7,$9E,$ED,$FD,$A6,$BF,$6E
	.dc.b	$F8,$67,$C2,$E5,$F8,$27,$2F,$4F,$E3,$9F,$1B,$D3,$F8,$A7,$6F,$47
	.dc.b	$E5,$9F,$2B,$D0,$F9,$27,$8B,$CD,$F9,$E7,$CE,$F3,$7E,$69,$E6,$F2
	.dc.b	$BE,$99,$F4,$BC,$9F,$A2,$7C,$AD,$BF,$3F,$AF,$F3,$6F,$CA,$9F,$67
	.dc.b	$FA,$79,$CC,$FF,$3B,$97,$D9,$66,$7E,$47,$93,$FC,$A3,$F5,$27,$EF
	.dc.b	$59,$9F,$D1,$77,$FF,$38,$9F,$F7,$C7,$6F,$FB,$2B,$C6,$FE,$69,$97
	.dc.b	$82,$31,$9F,$F4,$48,$EF,$1F,$ED,$0F,$1B,$B6,$5D,$5F,$07,$E0,$9D
	.dc.b	$FC,$5E,$7B,$DC,$79,$6F,$E3,$F1,$78,$E5,$FE,$DA,$78,$BF,$ED,$CF
	.dc.b	$0F,$8B,$6B,$8B,$FE,$E0,$F0,$F8,$F0,$FE,$1C,$3C,$9C,$78,$1C,$9B
	.dc.b	$5C,$9E,$1E,$07,$27,$F2,$0F,$17,$F2,$51,$C5,$AF,$17,$F2,$DD,$DC
	.dc.b	$53,$C5,$FF,$39,$47,$C9,$D3,$E9,$76,$7C,$ED,$7E,$A7,$67,$13,$87
	.dc.b	$57,$7C,$0E,$F5,$FF,$7F,$F5,$77,$3E,$7F,$E4,$AF,$80,$F8,$5D,$7F
	.dc.b	$DA,$3F,$51,$7A,$CA,$D6,$5F,$86,$ED,$FE,$EB,$4F,$CC,$35,$8F,$CB
	.dc.b	$ED,$FE,$ED,$F8,$73,$F4,$DC,$E3,$FB,$59,$CF,$C2,$E2,$3F,$87,$65
	.dc.b	$D3,$4C,$75,$F8,$F3,$D7,$5F,$4C,$AF,$3F,$F8,$43,$46,$7C,$FA,$7D
	.dc.b	$59,$5C,$FC,$F9,$77,$C3,$97,$07,$B3,$5F,$F0,$16,$5C,$A3,$1C,$BC
	.dc.b	$9B,$87,$27,$87,$2F,$2E,$1F,$ED,$98,$FF,$89,$B0,$E5,$F5,$DF,$F0
	.dc.b	$03,$F1,$4F,$E2,$98,$23,$E3,$57,$F8,$6F,$F1,$6F,$9D,$18,$8F,$F1
	.dc.b	$5F,$E5,$78,$D7,$F8,$E7,$E7,$1F,$EB,$BF,$B1,$FC,$C4,$7E,$67,$1E
	.dc.b	$8E,$A3,$FD,$92,$E8,$E4,$7C,$ED,$BF,$C5,$BF,$A0,$63,$9F,$FC,$39
	.dc.b	$FA,$0E,$B3,$FF,$12,$6B,$5F,$F1,$86,$9C,$B8,$BF,$C4,$58,$5C,$7F
	.dc.b	$C5,$F8,$7E,$1F,$DE,$FD,$66,$5F,$E2,$7F,$BE,$C6,$9F,$D3,$6F,$FC
	.dc.b	$6B,$F0,$38,$8F,$F1,$96,$9F,$E5,$3D,$D8,$FE,$D6,$3F,$C6,$DF,$09
	.dc.b	$F9,$3F,$B1,$C4,$7F,$0D,$FC,$3F,$18,$7E,$BA,$3F,$91,$7D,$77,$E6
	.dc.b	$7E,$4B,$F7,$DF,$65,$FE,$FE,$F7,$7E,$25,$E4,$FE,$91,$02,$3F,$27
	.dc.b	$45,$72,$9D,$10,$7E,$B1,$CE,$B1,$19,$FB,$CC,$65,$FE,$4F,$FC,$32
	.dc.b	$BF,$E5,$BF,$DF,$3F,$5D,$EA,$7F,$C0,$1F,$C2,$AE,$3F,$F0,$03,$B9
	.dc.b	$E4,$7E,$CF,$F6,$9D,$6E,$7F,$8D,$78,$7F,$BB,$21,$F8,$F9,$E1,$7F
	.dc.b	$D9,$7D,$0F,$F3,$F7,$ED,$1F,$0B,$0E,$53,$9F,$96,$7E,$FD,$9C,$BF
	.dc.b	$93,$61,$C3,$F1,$77,$C6,$7F,$B6,$FE,$73,$F3,$B5,$AC,$19,$AF,$7F
	.dc.b	$6C,$FA,$AF,$D1,$FF,$00,$FA,$38,$3C,$E7,$F1,$3F,$8C,$FE,$21,$19
	.dc.b	$8C,$FD,$6E,$38,$3F,$BF,$7D,$EF,$DA,$7C,$7C,$4B,$FE,$79,$39,$FC
	.dc.b	$A7,$F8,$D7,$E5,$3F,$20,$E9,$8F,$C7,$E6,$F9,$B8,$27,$FA,$27,$AA
	.dc.b	$FA,$B7,$ED,$1F,$1E,$73,$0F,$C6,$71,$AF,$F9,$AE,$37,$F9,$F3,$F5
	.dc.b	$8F,$F5,$EF,$E6,$7F,$C0,$B0,$E3,$8A,$B5,$F5,$6F,$27,$59,$7B,$8F
	.dc.b	$CB,$FE,$ED,$EA,$70,$7E,$F7,$F2,$7F,$E6,$DF,$93,$7F,$5B,$FA,$C7
	.dc.b	$E4,$0F,$D0,$FE,$F4,$E6,$1F,$A4,$63,$0F,$F1,$45,$F0,$86,$EF,$6D
	.dc.b	$F2,$8F,$C7,$7F,$8D,$E5,$CE,$F7,$5A,$FD,$6B,$39,$79,$8F,$E2,$FB
	.dc.b	$79,$9C,$1F,$DC,$76,$FE,$9D,$B7,$D8,$8F,$7F,$47,$A5,$D9,$FE,$3D
	.dc.b	$9F,$EA,$9F,$45,$F6,$BA,$FF,$DE,$DF,$B7,$62,$BF,$C8,$51,$F8,$C3
	.dc.b	$F8,$5F,$DE,$E6,$1F,$B6,$BE,$1B,$FA,$14,$7F,$2B,$C3,$FC,$09,$39
	.dc.b	$EF,$FD,$34,$62,$BF,$9D,$7E,$44,$FA,$47,$D4,$3E,$2B,$F9,$FD,$EF
	.dc.b	$05,$EE,$9F,$09,$FD,$13,$F0,$CF,$56,$DF,$DC,$FF,$24,$FE,$E1,$18
	.dc.b	$87,$ED,$F8,$DF,$E1,$8F,$E0,$1C,$2F,$2B,$F4,$9E,$DF,$F7,$9E,$2F
	.dc.b	$99,$3F,$D0,$EB,$0F,$29,$FE,$FD,$F9,$BF,$7F,$C6,$AC,$FC,$AC,$BE
	.dc.b	$57,$85,$88,$EC,$7F,$2A,$C3,$19,$78,$3A,$7E,$14,$F9,$A7,$FC,$15
	.dc.b	$C9,$E2,$8F,$EB,$39,$FF,$83,$7F,$66,$C4,$7F,$8C,$FF,$51,$C3,$EE
	.dc.b	$FC,$4F,$FC,$25,$FB,$6E,$A3,$FC,$FD,$38,$75,$97,$FA,$5D,$F8,$B1
	.dc.b	$FE,$19,$F8,$FE,$10,$BF,$F0,$E7,$EE,$18,$AB,$3C,$BF,$76,$C6,$07
	.dc.b	$8D,$FF,$88,$FE,$43,$EB,$F5,$FE,$26,$91,$8F,$85,$D6,$1F,$5B,$F6
	.dc.b	$CA,$FC,$83,$FB,$A7,$57,$F8,$B7,$AF,$CB,$E9,$D9,$FB,$BF,$AD,$7F
	.dc.b	$8C,$7B,$FF,$05,$F8,$6D,$9F,$A9,$D5,$FC,$A3,$F1,$EC,$47,$53,$FC
	.dc.b	$0B,$9E,$3A,$BF,$93,$69,$FE,$3B,$F8,$DF,$C6,$7E,$F7,$6F,$9E,$3F
	.dc.b	$D2,$36,$C4,$73,$3F,$CC,$DF,$AE,$63,$1F,$E6,$1F,$B8,$7F,$91,$7E
	.dc.b	$37,$E6,$CD,$AF,$CA,$CF,$F9,$27,$F1,$6F,$CE,$FC,$C7,$F2,$7E,$64
	.dc.b	$63,$F3,$7F,$F2,$77,$E8,$7F,$9F,$FC,$4A,$E3,$E4,$DB,$FA,$E7,$8D
	.dc.b	$88,$F2,$1F,$EB,$FF,$C0,$B1,$BF,$F4,$3F,$BB,$FF,$2C,$7E,$1D,$F8
	.dc.b	$09,$B5,$FA,$89,$FF,$2E,$7E,$A3,$F8,$4F,$DC,$2C,$FE,$B3,$F7,$7F
	.dc.b	$F3,$07,$DB,$E7,$87,$FD,$0B,$39,$5F,$A7,$64,$FC,$6D,$7F,$99,$FE
	.dc.b	$3B,$81,$89,$7E,$79,$FB,$BE,$EF,$F3,$63,$FB,$DF,$AA,$B8,$FE,$F3
	.dc.b	$EB,$6B,$EE,$3F,$CE,$77,$FE,$73,$7E,$27,$1F,$F8,$9B,$D9,$FF,$3B
	.dc.b	$4F,$CE,$FD,$D6,$D7,$F0,$6F,$CA,$2B,$CF,$F2,$7F,$F3,$E7,$E0,$3F
	.dc.b	$CB,$FF,$07,$FF,$23,$7E,$85,$8A,$FE,$A1,$1F,$E4,$9F,$CF,$7F,$D0
	.dc.b	$5F,$87,$FF,$13,$FC,$E2,$D7,$8B,$66,$38,$8F,$F0,$89,$FE,$56,$71
	.dc.b	$1F,$E1,$58,$FF,$2C,$D6,$FB,$3F,$BF,$7A,$2E,$5C,$57,$E7,$A1,$F3
	.dc.b	$7F,$E9,$3A,$F1,$70,$FE,$AF,$1F,$E7,$C7,$89,$C3,$FE,$0C,$F0,$1F
	.dc.b	$B1,$F1,$C7,$E4,$C3,$C3,$C4,$7F,$80,$CF,$F6,$27,$ED,$0F,$03,$38
	.dc.b	$7F,$A9,$F4,$FF,$38,$CB,$EB,$F3,$3F,$E7,$2B,$0E,$37,$06,$B0,$FF
	.dc.b	$3C,$C7,$F3,$7D,$7C,$FF,$5D,$FE,$38,$F2,$3F,$C3,$7C,$1F,$E2,$58
	.dc.b	$C7,$E2,$1E,$68,$EB,$F8,$9F,$F9,$F7,$F3,$1C,$68,$8F,$F4,$07,$D4
	.dc.b	$FF,$AE,$7F,$70,$F8,$E7,$F4,$BE,$8C,$57,$F3,$DC,$BF,$C6,$3E,$DE
	.dc.b	$23,$FC,$95,$1F,$E3,$4F,$4B,$FD,$85,$FB,$C7,$F2,$0F,$8C,$B5,$FE
	.dc.b	$55,$FD,$FE,$BF,$97,$E1,$BB,$FF,$8D,$7F,$06,$FF,$65,$7E,$8F,$FD
	.dc.b	$7B,$E2,$FF,$C4,$94,$8C,$7B,$EF,$1D,$F1,$1F,$EF,$5F,$C4,$5F,$69
	.dc.b	$FE,$E6,$FC,$03,$EC,$76,$BF,$00,$F9,$F8,$70,$DF,$E8,$E1,$C3,$F0
	.dc.b	$B6,$1C,$3E,$7F,$7F,$19,$FF,$00,$7A,$5D,$DF,$E8,$1F,$D0,$5E,$F3
	.dc.b	$F4,$2F,$F6,$6E,$9E,$5C,$38,$5E,$EB,$5C,$FE,$2F,$96,$FC,$0D,$BF
	.dc.b	$77,$7E,$36,$F8,$E7,$E8,$AF,$D2,$9F,$95,$F8,$30,$F8,$58,$F2,$C7
	.dc.b	$8E,$1C,$2F,$7E,$17,$C0,$FC,$2B,$BF,$88,$F8,$77,$EB,$0F,$CA,$FF
	.dc.b	$DA,$1F,$92,$3F,$49,$E1,$B7,$E5,$B0,$E1,$7B,$F4,$57,$FE,$42,$AF
	.dc.b	$A9,$F8,$FF,$B3,$EA,$1E,$CF,$FB,$F7,$AF,$FC,$CD,$F7,$BF,$E0,$4F
	.dc.b	$09,$7B,$AD,$7D,$BF,$FA,$0F,$F1,$0F,$D2,$FE,$37,$F9,$7C,$FF,$C7
	.dc.b	$BC,$3F,$E6,$E8,$C4,$7F,$3E,$8F,$AB,$18,$EC,$FF,$40,$7E,$6A,$8C
	.dc.b	$57,$FA,$F3,$F3,$3C,$7B,$5F,$E2,$CE,$67,$ED,$78,$AF,$E5,$31,$F7
	.dc.b	$6F,$F6,$D1,$E2,$FB,$BD,$3E,$1F,$0E,$E7,$E0,$AD,$CF,$F3,$5C,$BC
	.dc.b	$4F,$F8,$EB,$F4,$07,$F0,$06,$23,$F9,$BC,$3F,$5F,$C4,$7F,$18,$F2
	.dc.b	$9F,$BF,$63,$E0,$7F,$CB,$11,$FE,$A9,$CB,$FD,$FF,$9F,$FB,$63,$F4
	.dc.b	$87,$E3,$5C,$A3,$1C,$B7,$2F,$F6,$EF,$25,$7F,$26,$7F,$14,$DD,$F9
	.dc.b	$1B,$4F,$F2,$8E,$9E,$28,$C5,$7F,$4E,$8F,$F5,$D4,$7F,$96,$A3,$FC
	.dc.b	$A3,$FA,$FF,$2F,$A7,$FF,$16,$EB,$D3,$E9,$3F,$CB,$1F,$B6,$DB,$FE
	.dc.b	$8D,$FB,$6B,$F4,$6F,$F7,$6C,$7F,$38,$D1,$59,$BF,$F3,$05,$EF,$1F
	.dc.b	$F1,$CE,$57,$0F,$F3,$0E,$5C,$F1,$F4,$F1,$D9,$F4,$78,$2F,$D1,$FE
	.dc.b	$4C,$7C,$46,$BF,$10,$70,$6D,$FF,$27,$61,$F3,$A7,$71,$FB,$C8,$71
	.dc.b	$6E,$3E,$36,$15,$FB,$98,$5C,$7F,$D2,$5C,$FF,$84,$C7,$E9,$C6,$39
	.dc.b	$FC,$DD,$BF,$E5,$6E,$5B,$8F,$F9,$FB,$5A,$FE,$8F,$AE,$F3,$FA,$63
	.dc.b	$F7,$DD,$7A,$CB,$FD,$41,$D9,$FA,$0D,$3F,$2F,$18,$F5,$7E,$27,$C4
	.dc.b	$F7,$9F,$D4,$E1,$F0,$F3,$CB,$FA,$38,$CB,$FC,$73,$C0,$FE,$4F,$F7
	.dc.b	$7F,$BE,$8F,$EB,$FE,$3F,$F0,$4E,$F7,$8F,$FC,$13,$7F,$EC,$77,$BB
	.dc.b	$FF,$6A,$78,$BF,$C1,$39,$BF,$96,$CB,$CD,$FE,$37,$C3,$FC,$6B,$4F
	.dc.b	$E3,$3E,$6F,$29,$CF,$99,$FC,$5F,$CB,$FE,$2D,$87,$FA,$9B,$3F,$EE
	.dc.b	$3E,$8F,$F2,$8F,$53,$F9,$77,$A9,$FC,$B6,$B5,$5F,$CD,$7D,$5F,$E6
	.dc.b	$95,$AF,$4B,$CF,$EB,$FF,$3E,$F1,$FD,$7E,$BF,$F4,$3F,$13,$38,$FF
	.dc.b	$7E,$F6,$3F,$A4,$D6,$BB,$7F,$D4,$BA,$7F,$53,$A7,$B7,$FD,$4F,$DA
	.dc.b	$FE,$B3,$EE,$7C,$4F,$7F,$FA,$F7,$BF,$FD,$77,$E4,$3F,$C0,$FE,$9F
	.dc.b	$F8,$2B,$EE,$FF,$B3,$65,$FE,$40,$F8,$57,$D7,$27,$33,$FE,$22,$97
	.dc.b	$C2,$67,$E2,$BF,$D4,$3F,$0F,$6B,$FD,$93,$F1,$15,$FF,$14,$FC,$0F
	.dc.b	$FB,$6B,$15,$FF,$A7,$BE,$27,$B7,$C6,$7F,$84,$FC,$FF,$E1,$07,$F0
	.dc.b	$9F,$8C,$CE,$3F,$E0,$B8,$FF,$45,$60,$AF,$F2,$43,$67,$FC,$91,$D3
	.dc.b	$E9,$F2,$BF,$E0,$07,$A1,$FE,$1B,$AF,$F5,$34,$7F,$87,$AB,$F8,$CF
	.dc.b	$C6,$ED,$ED,$8F,$F1,$9F,$8F,$7E,$25,$99,$FF,$0C,$CF,$9A,$7F,$BE
	.dc.b	$E1,$FE,$44,$DF,$FC,$63,$F1,$BB,$AF,$F1,$E5,$FF,$78,$F6,$3F,$DD
	.dc.b	$57,$B8,$FF,$21,$FC,$9E,$EB,$FC,$7D,$0F,$93,$FF,$21,$57,$F9,$42
	.dc.b	$7E,$29,$FD,$9B,$E5,$38,$38,$F4,$FF,$CA,$13,$B8,$DF,$CB,$E9,$E3
	.dc.b	$C5,$F7,$74,$FF,$7C,$C7,$D0,$D7,$FC,$CD,$7F,$E4,$CC,$FF,$C9,$FC
	.dc.b	$3F,$E6,$AF,$96,$E1,$1F,$EC,$63,$FC,$E5,$3B,$7F,$CE,$A7,$F8,$5B
	.dc.b	$0D,$BF,$CC,$F2,$B8,$FF,$9B,$32,$FC,$71,$FE,$87,$9F,$E9,$7F,$33
	.dc.b	$87,$FD,$0E,$7F,$4E,$F9,$9F,$CA,$75,$FF,$BC,$3E,$6E,$CF,$98,$FE
	.dc.b	$31,$F3,$78,$7F,$CF,$3C,$AF,$9C,$C6,$7F,$28,$FB,$DF,$25,$F9,$FF
	.dc.b	$9F,$FF,$7D,$FC,$F5,$C3,$A1,$FC,$EF,$F0,$BC,$3F,$D3,$7E,$86,$E1
	.dc.b	$FE,$86,$FA,$1F,$D4,$7D,$0F,$FB,$07,$97,$FD,$21,$7F,$EC,$17,$FD
	.dc.b	$7A,$F4,$3F,$B4,$FD,$2F,$EB,$FE,$93,$FD,$7F,$3F,$EC,$49,$FF,$5B
	.dc.b	$4F,$EC,$FE,$97,$3F,$4D,$D4,$DF,$D4,$7F,$B4,$EC,$FF,$91,$BB,$3F
	.dc.b	$DB,$11,$FE,$41,$C3,$FD,$81,$F5,$1F,$BD,$F6,$B5,$1F,$CF,$FC,$3C
	.dc.b	$3F,$EB,$1F,$06,$BF,$DA,$FE,$1B,$F8,$27,$D4,$FF,$BC,$3F,$0B,$B3
	.dc.b	$FE,$C8,$EA,$FC,$8F,$D6,$FF,$BE,$34,$3D,$A7,$FB,$27,$D6,$7F,$0D
	.dc.b	$F7,$B4,$66,$3F,$C1,$71,$97,$FD,$85,$F5,$D5,$FE,$F9,$F0,$5F,$C5
	.dc.b	$7E,$6B,$84,$7F,$99,$3E,$C3,$0F,$CD,$3F,$CC,$FF,$61,$87,$FC,$09
	.dc.b	$F6,$37,$B7,$FD,$7B,$DD,$FC,$7F,$1E,$75,$FE,$AE,$F6,$F0,$F3,$3F
	.dc.b	$D6,$3E,$D7,$F2,$5F,$5F,$70,$CF,$0E,$C6,$78,$7F,$91,$FD,$F5,$CF
	.dc.b	$F2,$43,$D1,$FF,$3D,$71,$BE,$D3,$AF,$DA,$FF,$33,$7B,$FD,$AF,$F3
	.dc.b	$5F,$8A,$CD,$1C,$AF,$B6,$C9,$8F,$B7,$D9,$AF,$86,$E2,$70,$FB,$9F
	.dc.b	$5F,$DD,$7F,$41,$7D,$BF,$75,$FD,$0F,$EE,$F3,$89,$CB,$78,$7F,$DF
	.dc.b	$1E,$16,$DF,$F7,$D7,$C3,$FF,$4B,$FB,$CE,$27,$0F,$BD,$E6,$59,$FF
	.dc.b	$80,$1C,$DF,$D4,$63,$FD,$3F,$CB,$FE,$F2,$FB,$6F,$EA,$9F,$7D,$FE
	.dc.b	$A2,$C1,$D1,$FE,$4C,$C7,$FC,$A7,$A2,$B5,$3F,$CB,$27,$FB,$04,$63
	.dc.b	$A1,$E0,$7F,$05,$FB,$A7,$16,$AF,$FD,$DF,$97,$6B,$DF,$B2,$B7,$81
	.dc.b	$BB,$FE,$1F,$F2,$7E,$67,$F4,$9A,$FF,$6B,$FA,$2F,$C3,$FE,$01,$67
	.dc.b	$F0,$DF,$80,$7F,$6F,$F9,$1C,$D9,$FC,$6C,$66,$7F,$E0,$0F,$33,$F0
	.dc.b	$76,$3F,$E3,$7C,$38,$56,$A5,$F8,$26,$71,$FF,$64,$E1,$C6,$1F,$83
	.dc.b	$BF,$05,$7E,$0E,$FC,$1F,$FC,$0D,$87,$0C,$79,$DE,$E9,$F8,$3F,$F8
	.dc.b	$2F,$F0,$BF,$49,$E8,$D4,$FF,$B6,$3E,$0D,$F7,$CF,$B1,$F4,$4E,$61
	.dc.b	$EC,$63,$84,$2F,$8D,$BF,$0B,$FF,$0B,$FE,$17,$FE,$CB,$CB,$BD,$F0
	.dc.b	$BD,$D6,$BF,$0B,$F3,$4F,$4F,$0F,$F6,$3D,$BF,$C5,$B0,$E3,$3F,$E1
	.dc.b	$8F,$C5,$7F,$C4,$5F,$87,$7E,$FB,$E6,$AC,$FE,$F7,$F0,$EA,$FF,$A3
	.dc.b	$FE,$62,$E5,$FB,$C3,$FB,$5F,$E2,$7F,$EF,$8F,$C5,$1F,$8A,$3F,$14
	.dc.b	$FB,$13,$98,$7E,$2B,$8C,$7E,$0D,$F0,$B7,$DB,$FF,$85,$3F,$15,$7D
	.dc.b	$1F,$F8,$EF,$2E,$55,$AD,$94,$FC,$57,$59,$7C,$1D,$7F,$B1,$71,$FC
	.dc.b	$1C,$7F,$5F,$18,$E2,$FF,$21,$7E,$1F,$88,$FF,$08,$D7,$C7,$8C,$47
	.dc.b	$F8,$26,$7F,$C9,$1F,$87,$7D,$8C,$1F,$5B,$08,$63,$F1,$6D,$C7,$DE
	.dc.b	$8C,$9B,$F6,$B3,$0F,$8B,$84,$31,$F8,$C6,$9C,$1A,$FC,$63,$0B,$F1
	.dc.b	$9F,$D3,$65,$E5,$8E,$77,$FE,$B7,$EC,$7E,$35,$E9,$CF,$C6,$7E,$93
	.dc.b	$F1,$8F,$F3,$24,$66,$3C,$FE,$43,$C0,$FF,$34,$7C,$96,$67,$D7,$F9
	.dc.b	$AE,$32,$F7,$47,$4F,$C7,$31,$1F,$CD,$FF,$1D,$10,$9E,$C6,$BE,$23
	.dc.b	$11,$3E,$3B,$FC,$ED,$D5,$F3,$FD,$4C,$CF,$1F,$C7,$F3,$8F,$EA,$A7
	.dc.b	$33,$DB,$F1,$FC,$CF,$BB,$4F,$AF,$F7,$38,$71,$F9,$06,$B4,$E3,$83
	.dc.b	$2C,$3D,$31,$E7,$F0,$AF,$A6,$99,$FC,$DD,$7F,$09,$39,$F9,$6F,$E3
	.dc.b	$11,$F8,$7F,$C8,$BC,$5B,$79,$1F,$0F,$4F,$F3,$DF,$C8,$BE,$9F,$DE
	.dc.b	$BF,$0B,$FF,$4B,$78,$BF,$8F,$9C,$FE,$03,$8D,$7E,$07,$07,$C4,$FC
	.dc.b	$8F,$F8,$56,$5F,$D1,$F9,$BE,$1C,$3E,$B7,$85,$FC,$3D,$3B,$5F,$F4
	.dc.b	$DE,$7E,$AF,$F4,$FF,$C2,$78,$FB,$3C,$1E,$78,$72,$F2,$BC,$FC,$17
	.dc.b	$0F,$CE,$E3,$FE,$FE,$C7,$FA,$C7,$4F,$26,$ED,$E7,$C7,$C7,$C5,$CA
	.dc.b	$7F,$AF,$3E,$2E,$2F,$D7,$7E,$49,$C5,$B7,$C7,$CE,$FD,$97,$03,$F2
	.dc.b	$5F,$C0,$FE,$54,$FC,$77,$F8,$24,$FE,$DE,$73,$F8,$EE,$23,$F9,$8F
	.dc.b	$47,$E8,$BC,$77,$E5,$19,$94,$E7,$F2,$8C,$75,$76,$CA,$EB,$F9,$4D
	.dc.b	$19,$F1,$F6,$BA,$67,$79,$BF,$57,$97,$5F,$C9,$39,$70,$7A,$9F,$2B
	.dc.b	$EC,$9E,$2D,$78,$BF,$03,$AB,$8E,$FF,$C4,$DE,$2F,$F1,$5F,$CA,$78
	.dc.b	$B6,$B8,$BF,$C6,$7E,$D3,$8B,$FD,$49,$E2,$FF,$54,$EC,$E2,$D9,$E2
	.dc.b	$FF,$58,$EA,$E2,$FE,$F7,$83,$F7,$3C,$1F,$0B,$3F,$AB,$F9,$57,$2A
	.dc.b	$3F,$2E,$3E,$CF,$B1,$C8,$F0,$CB,$AB,$81,$B7,$F0,$38,$76,$E4,$FA
	.dc.b	$DD,$5C,$3F,$0F,$EC,$75,$FC,$1B,$F9,$B7,$E0,$5F,$C0,$7F,$19,$FE
	.dc.b	$8D,$18,$E0,$FE,$43,$EF,$E3,$3F,$EE,$7D,$DF,$C0,$BE,$D3,$11,$F8
	.dc.b	$DD,$7F,$86,$FD,$B6,$3B,$FA,$71,$33,$FD,$B6,$7D,$23,$18,$7F,$7E
	.dc.b	$FC,$AF,$A4,$62,$11,$FC,$43,$0E,$FF,$87,$F4,$C3,$FA,$73,$FC,$67
	.dc.b	$5F,$1F,$83,$CA,$7F,$92,$69,$E4,$CB,$A7,$81,$C7,$3F,$ED,$BA,$78
	.dc.b	$EF,$FA,$C7,$E5,$9C,$67,$FC,$22,$F8,$6F,$C9,$BF,$19,$3E,$18,$F4
	.dc.b	$77,$F1,$93,$C6,$FF,$A2,$61,$FD,$8B,$8B,$E4,$BC,$63,$FB,$47,$E5
	.dc.b	$BC,$4E,$17,$FE,$22,$FC,$A3,$C0,$6D,$FF,$12,$F8,$3D,$CD,$AF,$07
	.dc.b	$CC,$6D,$FF,$71,$6F,$FD,$0B,$5F,$B3,$F4,$3C,$3E,$37,$D8,$6F,$C3
	.dc.b	$F5,$1D,$3C,$3F,$80,$73,$F8,$2F,$71,$E0,$FC,$BB,$E2,$1D,$FC,$5F
	.dc.b	$84,$79,$3C,$4F,$90,$78,$FE,$9B,$E3,$1E,$8F,$CB,$FE,$61,$E7,$F1
	.dc.b	$BE,$51,$EC,$F1,$BE,$81,$EB,$FA,$4F,$9C,$7C,$0F,$CC,$3E,$A1,$EF
	.dc.b	$F1,$FE,$91,$F1,$3C,$7F,$B0,$7C,$3F,$82,$FA,$C7,$DD,$FC,$43,$FC
	.dc.b	$31,$F8,$77,$F5,$EF,$83,$FF,$14,$46,$39,$FF,$B4,$FE,$35,$8E,$9F
	.dc.b	$F2,$AF,$99,$FD,$A7,$E0,$F1,$08,$FE,$D9,$CB,$8F,$37,$A7,$8A,$CB
	.dc.b	$F9,$86,$5D,$23,$18,$7F,$98,$3F,$31,$E9,$18,$84,$74,$8C,$7B,$5D
	.dc.b	$2F,$FC,$DD,$7C,$DF,$D7,$F8,$BF,$E2,$CD,$3C,$BE,$7F,$19,$FF,$1B
	.dc.b	$79,$FC,$73,$FD,$96,$1C,$5F,$F3,$8F,$E6,$5E,$2A,$FF,$1E,$3C,$63
	.dc.b	$FC,$EB,$F9,$9F,$14,$FF,$81,$B8,$7E,$1F,$A1,$C6,$7F,$CF,$6F,$17
	.dc.b	$FD,$21,$87,$F8,$53,$8F,$E4,$F9,$FC,$78,$3E,$07,$E6,$9C,$6C,$F1
	.dc.b	$FC,$D7,$C0,$6F,$5F,$F3,$77,$0F,$F8,$E3,$87,$E9,$FA,$3D,$FD,$1F
	.dc.b	$31,$BC,$7F,$D1,$BF,$35,$D4,$F2,$7E,$6C,$F7,$21,$7E,$AB,$FC,$DF
	.dc.b	$5F,$7A,$73,$39,$F8,$DF,$F4,$BE,$EF,$CD,$FF,$CC,$5F,$85,$72,$C3
	.dc.b	$FB,$99,$FE,$73,$8C,$DF,$6F,$94,$7D,$13,$F0,$AE,$75,$AA,$F2,$56
	.dc.b	$BE,$27,$BF,$E4,$9F,$E6,$FF,$1F,$AE,$5F,$E7,$1D,$13,$FE,$73,$FC
	.dc.b	$2B,$C5,$3F,$E0,$49,$CB,$FE,$AE,$8F,$7F,$90,$FC,$E5,$F9,$CE,$AF
	.dc.b	$76,$BD,$D6,$BF,$2E,$F8,$3F,$92,$7F,$A0,$7F,$3B,$FF,$15,$DF,$8E
	.dc.b	$31,$87,$9A,$3F,$D2,$91,$BF,$8C,$FA,$1F,$9D,$FF,$BB,$7C,$67,$E1
	.dc.b	$9D,$2F,$FD,$87,$8F,$F9,$02,$B5,$F5,$FF,$43,$F3,$DF,$F4,$87,$B2
	.dc.b	$FC,$F5,$F2,$7F,$EC,$9F,$C5,$DF,$89,$BA,$1F,$9F,$7F,$A9,$71,$E3
	.dc.b	$87,$0F,$4D,$87,$0B,$7B,$5F,$EA,$BF,$D0,$3F,$D3,$DE,$B3,$F4,$07
	.dc.b	$E6,$5F,$4B,$A1,$5A,$97,$E6,$59,$FD,$07,$FD,$4D,$FA,$13,$E5,$5F
	.dc.b	$AB,$67,$A7,$FD,$C3,$39,$FD,$87,$A7,$E4,$7F,$EA,$DF,$C3,$5F,$74
	.dc.b	$F8,$F7,$E2,$AF,$89,$FF,$4A,$7D,$BB,$F2,$4F,$F6,$16,$3C,$70,$E1
	.dc.b	$7B,$FB,$0F,$F6,$3F,$E9,$1A,$DB,$FD,$8D,$E8,$FE,$7B,$F4,$6E,$5C
	.dc.b	$EC,$BF,$D8,$33,$FA,$79,$5F,$1B,$FF,$71,$7E,$8F,$FE,$B1,$C3,$84
	.dc.b	$74,$BF,$D6,$CF,$FA,$F6,$73,$3F,$EE,$9F,$4F,$BE,$3D,$AD,$1D,$AB
	.dc.b	$71,$FB,$23,$F7,$1F,$69,$DE,$B8,$3D,$76,$FF,$6A,$7A,$DF,$A6,$F9
	.dc.b	$F7,$BF,$FE,$E5,$9C,$E7,$FB,$69,$7C,$1E,$60,$7E,$D4,$FD,$DF,$C1
	.dc.b	$FF,$A7,$F4,$F9,$D1,$AC,$7F,$67,$2F,$AE,$E5,$3C,$4F,$F3,$AF,$B7
	.dc.b	$E2,$BE,$31,$93,$FD,$D5,$1E,$28,$CD,$FE,$FB,$F4,$8F,$06,$7E,$8B
	.dc.b	$FF,$3F,$C7,$7A,$E7,$BF,$C5,$3F,$71,$AF,$89,$C7,$B1,$D7,$0F,$9D
	.dc.b	$EE,$74,$C7,$EB,$57,$F2,$38,$E9,$3F,$4F,$DB,$F4,$37,$7E,$95,$F4
	.dc.b	$FE,$BF,$96,$1E,$18,$47,$97,$9B,$D1,$3F,$E9,$6F,$9A,$6B,$F8,$A8
	.dc.b	$F9,$47,$F1,$CF,$B7,$CD,$70,$8F,$A7,$C9,$F9,$48,$C4,$7F,$93,$DD
	.dc.b	$BC,$8F,$F5,$1F,$27,$E4,$E3,$6F,$E4,$5E,$3F,$A4,$F3,$9E,$07,$E0
	.dc.b	$03,$F9,$1F,$D1,$FF,$29,$A7,$F4,$07,$4F,$F2,$E9,$FE,$65,$1F,$99
	.dc.b	$FC,$5F,$FA,$2F,$E0,$3C,$27,$DE,$F0,$AF,$EA,$11,$FD,$1B,$4F,$E9
	.dc.b	$EF,$C9,$AF,$E9,$BF,$79,$EE,$8F,$E1,$B2,$8F,$CC,$E9,$FC,$45,$F8
	.dc.b	$7F,$09,$FD,$53,$F4,$3E,$1F,$9A,$FC,$E7,$78,$77,$76,$FD,$2F,$CF
	.dc.b	$77,$89,$DC,$7F,$49,$7E,$0F,$E7,$FB,$8F,$E0,$E7,$A7,$F4,$AC,$D3
	.dc.b	$E1,$79,$FE,$9B,$EF,$E7,$FE,$C7,$F8,$0F,$E0,$23,$18,$7F,$B6,$3E
	.dc.b	$4F,$0F,$EC,$38,$BF,$BA,$5F,$FB,$73,$7F,$ED,$F7,$F2,$67,$FB,$6F
	.dc.b	$E9,$BC,$2F,$FB,$A9,$FE,$E2,$FD,$37,$13,$F9,$3F,$B8,$AF,$E9,$7E
	.dc.b	$E3,$F2,$3F,$A7,$78,$AF,$F0,$1A,$E1,$F1,$57,$E6,$9F,$36,$EA,$E1
	.dc.b	$1F,$D9,$DF,$23,$FE,$F2,$CB,$93,$E4,$3F,$19,$96,$2F,$F1,$1A,$3F
	.dc.b	$37,$C4,$7E,$12,$BF,$88,$CF,$F9,$37,$1F,$04,$7D,$03,$F8,$AF,$C8
	.dc.b	$7F,$84,$BF,$17,$FC,$C7,$E3,$EF,$CB,$9F,$A7,$BF,$50,$7A,$7F,$D5
	.dc.b	$B0,$E1,$8F,$F1,$EC,$38,$5E,$EB,$FA,$DE,$7F,$E1,$EF,$C9,$3E,$2F
	.dc.b	$E8,$AF,$D3,$9F,$8C,$BF,$12,$FD,$DF,$E8,$4F,$BA,$FD,$EE,$1C,$2F
	.dc.b	$75,$AF,$8F,$7E,$32,$F8,$9F,$D8,$6A,$FC,$79,$F8,$9B,$F0,$67,$D7
	.dc.b	$BC,$7F,$F1,$87,$E1,$3C,$B1,$E3,$87,$0B,$DD,$7F,$8D,$7F,$02,$EF
	.dc.b	$F8,$C7,$EF,$7E,$13,$FB,$8F,$CB,$BF,$51,$7E,$78,$EB,$FF,$20,$F0
	.dc.b	$B0,$E1,$F2,$5F,$E0,$BB,$DD,$7F,$1A,$ED,$F5,$FE,$A5,$FE,$6E,$FD
	.dc.b	$13,$FC,$17,$F4,$3F,$B7,$FD,$17,$F8,$66,$6A,$D7,$47,$F8,$63,$D7
	.dc.b	$57,$F8,$DA,$9F,$A0,$7F,$B8,$3F,$65,$FA,$1F,$94,$FF,$A1,$3F,$49
	.dc.b	$7E,$5D,$FD,$5A,$F7,$B7,$F9,$C2,$F7,$F4,$2B,$7E,$53,$BC,$BA,$78
	.dc.b	$4C,$DF,$A4,$BF,$54,$7E,$39,$BD,$3A,$5E,$EB,$58,$F8,$7F,$57,$FE
	.dc.b	$51,$FA,$B3,$F5,$7F,$F0,$AF,$EA,$BF,$C5,$FF,$31,$FF,$1E,$E5,$F8
	.dc.b	$0B,$DD,$6B,$95,$5F,$8C,$AF,$C2,$FE,$33,$FC,$CF,$F1,$5F,$F1,$0F
	.dc.b	$EB,$0F,$CB,$DF,$83,$FE,$A7,$3F,$93,$D0,$C3,$87,$D2,$30,$E1,$7B
	.dc.b	$FD,$A1,$FA,$F7,$F8,$CB,$4F,$EF,$FE,$87,$E5,$35,$FF,$1A,$FE,$72
	.dc.b	$FD,$63,$F2,$7F,$A3,$70,$F2,$3F,$A3,$5E,$D5,$6E,$1F,$82,$BF,$72
	.dc.b	$7C,$FF,$F8,$77,$F5,$C7,$E7,$19,$8C,$79,$57,$6F,$EE,$7F,$0F,$F4
	.dc.b	$9C,$1F,$DB,$7F,$0B,$7E,$BD,$89,$FF,$00,$47,$C1,$CB,$BD,$72,$97
	.dc.b	$EB,$DB,$7F,$C0,$DE,$47,$97,$E4,$FC,$9A,$5B,$F0,$EB,$FE,$6C,$C7
	.dc.b	$F9,$14,$7F,$7D,$F1,$BF,$C4,$B1,$8C,$7C,$3A,$71,$AF,$37,$5B,$E4
	.dc.b	$9F,$B0,$7F,$6D,$FD,$87,$E8,$F5,$3F,$0E,$7E,$A4,$FC,$D7,$F9,$27
	.dc.b	$E1,$BC,$70,$E1,$7F,$62,$F7,$5F,$CA,$7F,$61,$7C,$CF,$F8,$2F,$57
	.dc.b	$DC,$7F,$A7,$7E,$3D,$FB,$17,$E7,$BD,$D7,$E6,$0F,$D8,$F8,$FD,$57
	.dc.b	$F8,$07,$0E,$17,$BA,$67,$FE,$80,$F8,$4F,$F0,$37,$EC,$9F,$E6,$EF
	.dc.b	$AA,$F9,$BF,$97,$3F,$01,$7E,$67,$BE,$0F,$F2,$5D,$EF,$C3,$5E,$E3
	.dc.b	$FD,$0F,$F8,$AF,$F8,$47,$E7,$BF,$D2,$7F,$9B,$3F,$11,$FD,$2F,$AC
	.dc.b	$FD,$95,$FB,$07,$0D,$3F,$CA,$F8,$70,$BD,$D7,$B7,$E3,$7F,$A7,$57
	.dc.b	$FB,$EB,$F2,$57,$D8,$17,$EB,$FF,$C5,$7C,$6F,$CB,$DF,$0B,$FF,$7F
	.dc.b	$D7,$03,$7F,$1F,$FA,$FF,$C0,$BF,$8B,$FC,$6F,$F8,$A7,$F4,$57,$E4
	.dc.b	$3F,$90,$FC,$7D,$F6,$6F,$D9,$FF,$B0,$7A,$1C,$34,$FF,$37,$E1,$C3
	.dc.b	$A9,$87,$0B,$7E,$CF,$FE,$F0,$FC,$13,$FC,$6D,$F5,$FF,$EC,$DC,$7F
	.dc.b	$C9,$7F,$AE,$FF,$8E,$7F,$2E,$FF,$36,$47,$FB,$96,$3F,$C7,$9F,$AE
	.dc.b	$6F,$EC,$91,$8F,$C2,$3F,$DA,$D1,$8A,$FF,$40,$FA,$BB,$AF,$F2,$23
	.dc.b	$D2,$3F,$E0,$0C,$7F,$91,$FA,$B5,$1F,$E8,$87,$4F,$90,$FF,$45,$FE
	.dc.b	$B9,$A4,$EA,$FC,$11,$FA,$E3,$FC,$89,$5F,$E5,$3F,$7B,$4F,$F9,$42
	.dc.b	$3C,$FE,$BF,$F9,$5F,$EF,$FD,$8D,$FD,$B7,$33,$FE,$9A,$9F,$F7,$DB
	.dc.b	$FE,$5D,$3F,$DE,$66,$67,$6F,$FB,$FF,$E9,$AF,$E1,$3F,$E0,$88,$FF
	.dc.b	$6F,$4F,$F9,$96,$15,$AE,$CE,$3F,$84,$F8,$AF,$FD,$52,$B1,$CF,$AF
	.dc.b	$AC,$7B,$FD,$3F,$F7,$BB,$EA,$7F,$AC,$1F,$03,$FE,$65,$8F,$EF,$18
	.dc.b	$F0,$C7,$CE,$77,$93,$DD,$E6,$62,$F7,$5E,$EF,$88,$F0,$4F,$53,$FD
	.dc.b	$7E,$F8,$1F,$8D,$BA,$7F,$D0,$93,$F0,$27,$AB,$E0,$7B,$EF,$E3,$3C
	.dc.b	$4F,$53,$FD,$47,$7F,$3A,$3A,$47,$9B,$B7,$C6,$F8,$2F,$FD,$1B,$F5
	.dc.b	$DF,$94,$F5,$3F,$DB,$7F,$82,$DA,$FF,$68,$6F,$57,$3B,$AF,$0F,$9E
	.dc.b	$3F,$D1,$D7,$FE,$E4,$E2,$CF,$06,$B5,$E5,$D5,$FA,$5A,$FD,$56,$78
	.dc.b	$CF,$39,$EB,$8B,$79,$70,$CB,$8E,$5E,$3F,$CC,$58,$DC,$71,$8E,$71
	.dc.b	$D6,$3B,$C7,$86,$3E,$BE,$CF,$16,$BF,$EF,$27,$9B,$F9,$93,$FD,$EB
	.dc.b	$C9,$FC,$D1,$D6,$5C,$72,$E5,$97,$3C,$BA,$65,$D7,$2F,$90,$F6,$C3
	.dc.b	$F1,$39,$E3,$1C,$E3,$AC,$77,$8F,$0C,$38,$E3,$F4,$79,$39,$1E,$6C
	.dc.b	$BE,$79,$F8,$CC,$75,$8E,$F1,$E1,$8F,$1C,$79,$63,$F6,$CF,$21,$CF
	.dc.b	$0F,$C9,$ED,$FC,$0F,$1D,$63,$BC,$78,$63,$C7,$1E,$58,$FE,$17,$8F
	.dc.b	$DC,$73,$CB,$FD,$BD,$AF,$C9,$51,$F9,$87,$B9,$FE,$B3,$C7,$96,$3C
	.dc.b	$F1,$FA,$53,$FC,$7F,$3F,$F7,$1C,$62,$3F,$A6,$C7,$FB,$9B,$93,$1D
	.dc.b	$5D,$F1,$FC,$B5,$72,$8C,$6B,$AB,$FF,$76,$EF,$6B,$FD,$88,$51,$AF
	.dc.b	$F2,$77,$F9,$94,$FF,$27,$DB,$52,$CE,$E7,$8D,$73,$96,$B2,$DE,$5D
	.dc.b	$77,$E8,$FB,$4F,$E0,$B9,$67,$1D,$63,$BC,$78,$63,$C7,$1E,$58,$79
	.dc.b	$F0,$E6,$FC,$40,$7F,$18,$75,$86,$F0,$E1,$87,$1C,$39,$61,$F0,$7B
	.dc.b	$7F,$B0,$6B,$F6,$78,$7A,$B7,$2F,$8D,$1F,$07,$3F,$9D,$B7,$F6,$8C
	.dc.b	$F3,$9E,$B3,$DE,$7C,$2F,$E7,$71,$74,$C3,$CF,$08,$FE,$40,$FE,$CF
	.dc.b	$3F,$C5,$47,$78,$C4,$F0,$8F,$EA,$31,$F5,$8F,$82,$FE,$09,$D9,$8B
	.dc.b	$F0,$DF,$DF,$E3,$F1,$4F,$C8,$7C,$2F,$F5,$BC,$3F,$35,$1E,$18,$C6
	.dc.b	$1F,$E0,$3C,$B9,$65,$CF,$1E,$98,$F5,$BF,$C6,$72,$76,$7F,$95,$EB
	.dc.b	$8D,$F3,$8E,$B1,$DE,$3F,$9C,$E6,$FE,$6F,$87,$E6,$DF,$EF,$1F,$71
	.dc.b	$98,$FB,$FC,$7F,$D0,$1F,$91,$3F,$DF,$70,$D4,$7E,$72,$ED,$7D,$EA
	.dc.b	$FE,$A3,$96,$B8,$FF,$9E,$5F,$C1,$9F,$EA,$BC,$FC,$A9,$95,$9C,$65
	.dc.b	$9C,$B5,$87,$83,$A3,$E5,$BF,$E1,$BA,$FE,$DE,$FA,$CF,$EE,$98,$F9
	.dc.b	$B5,$EB,$19,$3F,$5D,$39,$C3,$58,$6E,$7D,$BD,$3D,$9F,$F1,$35,$7F
	.dc.b	$82,$FA,$33,$8F,$C2,$7A,$CF,$C6,$3F,$C2,$C7,$77,$F1,$6F,$23,$FC
	.dc.b	$3D,$7F,$C3,$BA,$FF,$C4,$D1,$FC,$35,$FF,$15,$70,$E6,$3F,$63,$F6
	.dc.b	$9F,$AA,$9F,$CE,$F6,$FF,$8E,$39,$FF,$C6,$6E,$6B,$FC,$79,$7C,$5F
	.dc.b	$F3,$49,$7B,$B8,$5F,$F9,$1A,$35,$DF,$87,$F3,$E7,$1E,$9E,$DC,$FF
	.dc.b	$AC,$EC,$FD,$DB,$FE,$5C,$9F,$E4,$FF,$A6,$FE,$C3,$2F,$E0,$FF,$82
	.dc.b	$FF,$27,$F4,$71,$4C,$70,$FF,$0C,$AF,$E1,$BF,$AF,$E2,$3F,$C0,$A3
	.dc.b	$EB,$71,$3F,$C4,$BF,$4C,$B2,$F5,$6F,$E0,$B8,$FC,$A7,$3F,$BC,$6B
	.dc.b	$5F,$E7,$1F,$0B,$FC,$13,$C8,$FE,$92,$63,$1F,$E3,$AF,$F4,$CE,$6C
	.dc.b	$46,$B9,$AC,$FF,$50,$E0,$AB,$82,$EB,$8D,$73,$9F,$0F,$13,$FA,$BE
	.dc.b	$5F,$C4,$74,$F8,$D2,$3B,$F0,$7F,$48,$9F,$E0,$D3,$F2,$35,$CC,$7F
	.dc.b	$B4,$0E,$47,$C9,$3E,$61,$83,$E4,$E5,$C3,$2E,$39,$72,$C7,$F9,$6D
	.dc.b	$7C,$0C,$FC,$AF,$F1,$23,$1A,$69,$F8,$F8,$59,$FE,$E3,$57,$4E,$35
	.dc.b	$FC,$9C,$FE,$EA,$F9,$1E,$86,$2B,$84,$F3,$F1,$38,$CF,$FA,$4F,$A3
	.dc.b	$F1,$5B,$7C,$58,$FE,$71,$1F,$8D,$BD,$C7,$FB,$12,$3F,$21,$F7,$BB
	.dc.b	$8F,$E5,$D1,$F9,$3E,$3D,$F5,$7F,$A6,$30,$79,$5B,$BF,$F5,$25,$FE
	.dc.b	$67,$7D,$BF,$9B,$FC,$7B,$53,$E5,$3F,$31,$CB,$A7,$E9,$6D,$FE,$05
	.dc.b	$EA,$FC,$77,$0F,$81,$FC,$86,$9C,$AB,$FD,$B7,$1F,$93,$E6,$D4,$7E
	.dc.b	$4B,$5D,$57,$F4,$FB,$FC,$71,$D6,$7F,$2F,$AE,$B0,$FC,$E6,$BC,$A7
	.dc.b	$F1,$F9,$6D,$FF,$08,$DE,$78,$FF,$6B,$1F,$D8,$EF,$F7,$13,$C7,$2F
	.dc.b	$F7,$7D,$F1,$59,$E5,$F0,$5F,$EE,$78,$78,$67,$88,$FF,$3E,$E7,$CD
	.dc.b	$F5,$FE,$25,$F1,$32,$FF,$15,$3F,$B3,$C3,$33,$FB,$2C,$3E,$BC,$FF
	.dc.b	$B2,$72,$DC,$FF,$54,$BC,$BF,$D5,$AF,$8B,$FE,$2E,$BF,$BF,$E4,$FF
	.dc.b	$80,$BF,$16,$F7,$CF,$F7,$18,$FF,$02,$FE,$51,$88,$FF,$72,$47,$F8
	.dc.b	$23,$F8,$07,$F8,$2C,$7F,$81,$E3,$FC,$17,$FA,$7E,$E3,$FD,$D1,$1F
	.dc.b	$E0,$EF,$AC,$DC,$75,$3F,$C4,$FF,$B9,$ED,$72,$7F,$9B,$FF,$26,$DC
	.dc.b	$78,$0F,$F1,$6F,$C1,$EE,$7F,$CE,$9F,$83,$EA,$7B,$9F,$E3,$2F,$DF
	.dc.b	$75,$97,$F8,$7B,$E4,$35,$3D,$A3,$FC,$43,$F1,$7A,$8F,$F1,$CF,$C5
	.dc.b	$6B,$1F,$F4,$07,$95,$99,$7B,$39,$C3,$F9,$BF,$3F,$F9,$7C,$FB,$13
	.dc.b	$FE,$29,$7F,$DF,$4F,$E5,$9F,$F1,$7F,$D0,$75,$7F,$C6,$3E,$DF,$AD
	.dc.b	$FF,$19,$FC,$F6,$6B,$FC,$97,$F3,$D9,$7F,$C9,$BF,$1B,$95,$F6,$79
	.dc.b	$5F,$61,$9F,$BA,$C7,$EA,$DB,$FC,$3F,$5F,$AF,$71,$FA,$CE,$1F,$53
	.dc.b	$CF,$93,$97,$91,$D7,$F0,$9E,$9F,$AE,$77,$FA,$AE,$DF,$BD,$78,$7E
	.dc.b	$AF,$C1,$F9,$2F,$8F,$C8,$F1,$7C,$77,$97,$C7,$F2,$78,$FE,$7F,$17
	.dc.b	$CD,$E2,$FA,$7C,$3F,$47,$87,$EB,$F2,$FE,$E7,$57,$C4,$3F,$7A,$7C
	.dc.b	$4F,$33,$F8,$AB,$FE,$58,$EC,$DB,$FE,$70,$EC,$FB,$0F,$F9,$73,$AB
	.dc.b	$2B,$AB,$2F,$F9,$DB,$A3,$3D,$18,$87,$37,$0E,$6C,$72,$72,$E4,$D7
	.dc.b	$17,$4E,$2F,$C5,$E3,$FD,$13,$6F,$CD,$3F,$D1,$B6,$CE,$FF,$E6,$ED
	.dc.b	$72,$FF,$9C,$27,$3A,$FF,$A2,$33,$CB,$FE,$8A,$CF,$AF,$B1,$C6,$3F
	.dc.b	$92,$57,$FA,$99,$FF,$47,$BE,$2F,$63,$F2,$31,$FE,$7C,$3F,$89,$BF
	.dc.b	$96,$7F,$DD,$DE,$87,$F5,$AE,$B7,$C9,$3F,$17,$7E,$0A,$FB,$87,$E9
	.dc.b	$CF,$25,$FA,$53,$2E,$58,$F1,$C3,$85,$EF,$F0,$DD,$71,$7F,$A3,$B4
	.dc.b	$EF,$A7,$FB,$3B,$1E,$3F,$8C,$DA,$D6,$5C,$F5,$F0,$EB,$FE,$AC,$BE
	.dc.b	$F8,$79,$A7,$CB,$F1,$79,$E0,$FF,$4D,$FC,$C3,$F6,$B7,$96,$FB,$57
	.dc.b	$D3,$BF,$6E,$7E,$E0,$FB,$EF,$66,$5C,$B1,$7D,$D7,$9E,$31,$95,$E9
	.dc.b	$FF,$5F,$E9,$DE,$3B,$67,$D7,$F3,$4F,$25,$6B,$1F,$1E,$BE,$2D,$5F
	.dc.b	$5B,$FE,$C8,$F8,$47,$DB,$7F,$AD,$3E,$53,$E7,$FC,$57,$FB,$D2,$5F
	.dc.b	$63,$FE,$B8,$FC,$FD,$F8,$4B,$F0,$0D,$4E,$61,$F9,$AF,$1F,$D7,$FF
	.dc.b	$D7,$FF,$37,$E0,$F9,$17,$D6,$BC,$2F,$D3,$7E,$14,$FD,$27,$FC,$03
	.dc.b	$87,$0E,$26,$1C,$2F,$7F,$AF,$6B,$9F,$FC,$E3,$5A,$7F,$24,$64,$67
	.dc.b	$F0,$38,$BC,$91,$C3,$0E,$9E,$BB,$DF,$F8,$99,$F8,$67,$3E,$8F,$FB
	.dc.b	$BB,$E6,$75,$87,$FB,$C3,$E3,$DF,$8A,$3F,$32,$F8,$3E,$F3,$EF,$9F
	.dc.b	$25,$FC,$A3,$1E,$38,$B1,$E3,$87,$0B,$DF,$B5,$FE,$69,$DB,$FD,$CF
	.dc.b	$5D,$BE,$F3,$DB,$86,$4F,$16,$9E,$17,$E0,$E5,$DA,$BA,$F4,$B5,$F2
	.dc.b	$7B,$DE,$BC,$BA,$E3,$E9,$AD,$68,$F8,$36,$6F,$C1,$1F,$BC,$7F,$BE
	.dc.b	$BD,$57,$E4,$4F,$DF,$1F,$5C,$FD,$01,$F5,$AF,$45,$E4,$73,$CB,$96
	.dc.b	$3C,$70,$E1,$7B,$FD,$17,$5A,$F8,$B4,$EF,$FB,$57,$C2,$DB,$C7,$39
	.dc.b	$C7,$C4,$FE,$0A,$32,$FF,$10,$8E,$DA,$F9,$3E,$1F,$DB,$5A,$C3,$D7
	.dc.b	$BF,$AA,$BD,$3E,$A7,$C2,$FC,$E5,$F9,$C7,$C4,$CD,$FA,$EF,$EA,$BC
	.dc.b	$17,$ED,$8F,$31,$F5,$8F,$D0,$9F,$2E,$CF,$9E,$5F,$3B,$2E,$58,$F1
	.dc.b	$C3,$87,$ED,$1B,$D3,$C1,$5F,$B8,$FC,$47,$F0,$78,$F2,$C7,$C7,$EC
	.dc.b	$36,$CE,$1C,$23,$19,$F5,$F0,$BD,$93,$9A,$F4,$EF,$E8,$F9,$7F,$C0
	.dc.b	$FE,$00,$F2,$3E,$07,$86,$FD,$E6,$6B,$FB,$0A,$D3,$F8,$89,$D7,$EE
	.dc.b	$FF,$98,$9C,$CF,$F0,$6A,$4B,$F2,$0F,$F6,$92,$FB,$DF,$0E,$3E,$0A
	.dc.b	$DD,$6B,$EF,$3F,$2B,$F8,$3B,$F2,$7F,$A9,$FA,$3B,$EB,$7F,$A7,$7E
	.dc.b	$4C,$FC,$F9,$F8,$42,$1F,$A8,$31,$E5,$C1,$FA,$0C,$79,$7C,$9E,$F1
	.dc.b	$E3,$83,$F5,$FE,$1A,$78,$34,$FE,$B9,$8F,$F1,$2C,$B9,$AC,$BC,$9F
	.dc.b	$2F,$FB,$A9,$CC,$76,$C7,$CD,$AB,$EF,$3D,$B9,$75,$D7,$C5,$2F,$C6
	.dc.b	$FF,$6B,$CA,$FD,$D3,$F2,$9F,$B6,$F1,$FC,$91,$F6,$AF,$C4,$1F,$0C
	.dc.b	$FA,$D7,$DD,$7D,$2C,$78,$EE,$C7,$8E,$1C,$2D,$F9,$EE,$F9,$7F,$82
	.dc.b	$E9,$DD,$FE,$63,$AF,$8D,$CE,$2F,$C8,$BF,$8C,$CE,$65,$5E,$7D,$BC
	.dc.b	$DB,$3C,$2F,$6E,$DE,$4D,$BC,$72,$F1,$BF,$AD,$7D,$C7,$EC,$FE,$AB
	.dc.b	$F8,$BF,$E0,$0F,$C3,$FF,$A1,$FC,$C7,$FB,$46,$73,$F0,$F8,$BE,$1B
	.dc.b	$BF,$03,$FE,$C1,$1F,$EC,$4F,$C0,$9F,$9C,$FF,$29,$AF,$E0,$18,$FF
	.dc.b	$B6,$FE,$5B,$FD,$13,$5C,$27,$F9,$3C,$63,$1F,$EA,$7B,$7F,$89,$7F
	.dc.b	$03,$FE,$C1,$3F,$EE,$C8,$FF,$14,$4F,$F3,$CA,$FE,$5F,$1B,$9F,$F7
	.dc.b	$5C,$7F,$8B,$5F,$F4,$5B,$5E,$67,$F6,$BF,$A7,$FE,$6F,$E2,$66,$33
	.dc.b	$18,$9F,$F1,$B9,$FC,$67,$EA,$7D,$8F,$23,$FA,$6E,$BF,$C7,$7F,$7A
	.dc.b	$FD,$37,$C3,$3F,$04,$CD,$7F,$44,$9C,$E9,$88,$47,$EB,$7E,$2F,$FA
	.dc.b	$BE,$DF,$DF,$E7,$FC,$21,$FB,$77,$BF,$E1,$9F,$A1,$7F,$9B,$67,$E1
	.dc.b	$71,$BF,$18,$FF,$29,$47,$F6,$0F,$97,$FF,$29,$3F,$03,$8B,$E5,$FD
	.dc.b	$7F,$E7,$A7,$31,$93,$FC,$95,$CF,$FD,$D7,$A7,$FD,$BD,$1C,$2F,$07
	.dc.b	$D8,$7A,$5B,$8D,$97,$49,$F0,$C7,$F6,$4F,$03,$38,$7F,$43,$9F,$F0
	.dc.b	$E3,$FD,$16,$7F,$B4,$9C,$BE,$33,$C3,$3F,$82,$9C,$63,$E1,$8F,$46
	.dc.b	$0F,$CE,$BF,$A7,$72,$7E,$17,$4F,$1F,$E9,$7C,$AF,$FC,$55,$F8,$AE
	.dc.b	$31,$E5,$1F,$84,$FA,$EF,$4F,$E5,$BC,$A7,$FC,$4B,$ED,$7B,$74,$7E
	.dc.b	$99,$CF,$8B,$1F,$A1,$EE,$FF,$CF,$9F,$01,$BF,$D8,$3C,$1F,$29,$C7
	.dc.b	$C1,$D7,$37,$FA,$04,$F8,$FF,$8A,$FC,$3E,$0F,$87,$8F,$C3,$D7,$F9
	.dc.b	$8B,$FE,$0C,$E4,$FE,$D5,$D1,$FE,$89,$3E,$7F,$89,$C6,$DC,$9F,$E8
	.dc.b	$DF,$C5,$BE,$9E,$1F,$DC,$7F,$09,$7E,$93,$FE,$6C,$F9,$3B,$F7,$3F
	.dc.b	$DB,$BF,$74,$DD,$FF,$86,$AF,$35,$3C,$4C,$CA,$73,$87,$3F,$B0,$CC
	.dc.b	$63,$1F,$EF,$33,$C3,$F1,$EB,$13,$F0,$CC,$65,$BF,$8E,$B7,$E5,$3C
	.dc.b	$DC,$7F,$EA,$7C,$38,$D7,$A3,$E5,$FF,$D3,$3A,$7F,$80,$FC,$B7,$E8
	.dc.b	$5F,$E0,$E6,$FC,$C3,$F0,$7D,$D6,$6A,$4F,$F2,$5E,$BF,$D8,$4E,$2F
	.dc.b	$8F,$D6,$70,$AF,$F1,$9D,$F0,$F4,$B9,$FC,$BF,$6B,$9F,$99,$62,$B9
	.dc.b	$7C,$4D,$C3,$89,$CB,$3E,$F8,$71,$AF,$47,$EC,$7F,$EB,$9F,$3B,$FC
	.dc.b	$51,$C3,$F0,$38,$FF,$D4,$7C,$B9,$E0,$FF,$27,$CE,$79,$BF,$D1,$79
	.dc.b	$FF,$00,$8C,$72,$6E,$91,$BE,$4F,$F6,$37,$D2,$7F,$89,$B8,$BC,$5C
	.dc.b	$1F,$1F,$1F,$F2,$9E,$1F,$37,$7F,$8F,$EF,$70,$8C,$57,$B3,$97,$11
	.dc.b	$F2,$6B,$FC,$7F,$CB,$FE,$D3,$F0,$7F,$65,$E0,$38,$F1,$7F,$B2,$8F
	.dc.b	$E6,$B9,$FF,$72,$8F,$E1,$DE,$07,$F9,$27,$C0,$FE,$15,$39,$9F,$E8
	.dc.b	$B3,$97,$F8,$97,$86,$FD,$AF,$BE,$28,$C4,$F7,$C7,$FA,$6E,$9F,$D6
	.dc.b	$23,$F5,$C7,$F4,$D7,$B4,$7F,$25,$3D,$0E,$1C,$35,$FE,$9B,$E3,$7F
	.dc.b	$82,$A3,$11,$FD,$6A,$3F,$B2,$C6,$23,$FD,$79,$1F,$E8,$88,$C4,$23
	.dc.b	$F9,$67,$DF,$FF,$84,$2B,$CA,$7F,$A7,$1F,$F0,$6B,$EB,$9F,$99,$18
	.dc.b	$8F,$44,$7F,$B2,$7D,$8F,$EB,$2F,$AF,$14,$62,$1E,$87,$F4,$38,$7A
	.dc.b	$1F,$2B,$D1,$FE,$CD,$5F,$DA,$AD,$18,$8F,$F6,$B7,$E2,$7F,$ED,$7F
	.dc.b	$DA,$DF,$A3,$7D,$2F,$AA,$BA,$79,$F1,$FE,$2E,$FF,$35,$BF,$E6,$AF
	.dc.b	$D8,$C3,$C2,$ED,$F3,$69,$F4,$FC,$EF,$BB,$C8,$FC,$73,$EF,$FE,$19
	.dc.b	$70,$ED,$F9,$8F,$FA,$C1,$FE,$B1,$9F,$F7,$EE,$94,$E6,$7F,$A7,$67
	.dc.b	$FE,$10,$F8,$2F,$F0,$DC,$E6,$3F,$AB,$99,$F4,$F8,$F4,$7F,$4D,$AF
	.dc.b	$EC,$55,$F8,$FB,$FC,$74,$7F,$B1,$23,$F9,$DC,$7E,$2E,$3A,$FC,$A7
	.dc.b	$1C,$7C,$6F,$3C,$2D,$7E,$5F,$F1,$86,$58,$8E,$47,$AB,$F0,$0F,$EA
	.dc.b	$4F,$FB,$F3,$24,$62,$3F,$B7,$43,$F4,$9E,$CD,$8F,$D2,$7B,$7E,$A9
	.dc.b	$FE,$BD,$9F,$C6,$79,$4B,$C0,$F6,$46,$1F,$83,$FA,$57,$43,$FB,$87
	.dc.b	$E3,$58,$AF,$D8,$3F,$E6,$EC,$5F,$79,$E8,$EE,$FC,$0C,$62,$BF,$3E
	.dc.b	$7E,$5B,$5F,$EB,$BB,$FD,$8F,$67,$D2,$7F,$61,$E4,$FA,$96,$8C,$57
	.dc.b	$F8,$FF,$6E,$18,$74,$8E,$F3,$F2,$1F,$F0,$2E,$1F,$84,$F6,$BE,$29
	.dc.b	$FD,$B7,$AB,$AB,$67,$A6,$2F,$C3,$FF,$C0,$5A,$7F,$14,$CB,$F1,$B1
	.dc.b	$8A,$EB,$4F,$D2,$FE,$89,$FE,$01,$D7,$FC,$A3,$9F,$E5,$23,$15,$F9
	.dc.b	$3D,$71,$8F,$F0,$28,$ED,$87,$38,$C4,$76,$FD,$B3,$EE,$9F,$E0,$DD
	.dc.b	$7A,$D2,$31,$3D,$7E,$2B,$FB,$0B,$FE,$43,$E9,$FF,$3D,$C7,$F6,$4F
	.dc.b	$D6,$DF,$6D,$09,$F5,$CF,$DE,$15,$B9,$D1,$C2,$5F,$B3,$6B,$16,$B3
	.dc.b	$1D,$DE,$FF,$59,$D4,$EB,$F0,$94,$42,$7E,$7A,$FC,$EF,$74,$AD,$56
	.dc.b	$BF,$31,$CC,$62,$39,$DB,$F3,$6E,$38,$F8,$87,$EC,$9E,$8D,$FF,$C0
	.dc.b	$B1,$98,$F8,$11,$88,$F4,$C7,$9A,$3F,$C1,$7F,$8B,$F7,$C1,$FB,$B7
	.dc.b	$9E,$73,$1E,$38,$F1,$46,$23,$D3,$1F,$42,$3F,$CD,$9B,$FF,$A5,$BE
	.dc.b	$4F,$E7,$F5,$FC,$AD,$FE,$06,$1E,$DD,$7B,$4E,$76,$F1,$43,$7E,$37
	.dc.b	$F3,$6F,$86,$FF,$6B,$A3,$F8,$E6,$56,$BE,$BF,$07,$C1,$E3,$C4,$7C
	.dc.b	$78,$F8,$FC,$B8,$7E,$36,$3F,$84,$F6,$3C,$D1,$C7,$83,$FA,$0D,$F1
	.dc.b	$E3,$F9,$75,$AA,$F9,$77,$F8,$CF,$3B,$F9,$26,$1F,$D0,$B8,$BF,$31
	.dc.b	$1F,$9A,$7F,$D7,$9A,$6F,$2F,$F4,$B6,$BF,$AC,$BF,$E9,$71,$FC,$C7
	.dc.b	$B7,$FC,$B7,$7F,$9F,$6B,$FE,$A7,$F0,$DD,$B8,$8F,$F6,$4C,$7F,$B7
	.dc.b	$BE,$2B,$FB,$04,$23,$FD,$45,$E7,$FF,$5E,$C3,$B7,$8F,$AE,$EF,$F7
	.dc.b	$2B,$71,$FE,$55,$3F,$CF,$E3,$4F,$FB,$AB,$0F,$F4,$C7,$97,$FD,$03
	.dc.b	$6F,$DA,$C7,$FA,$EE,$3F,$BB,$C6,$23,$FD,$BB,$1F,$E2,$D8,$FC,$3C
	.dc.b	$7F,$5B,$F9,$CF,$F7,$AF,$5F,$F5,$7C,$2E,$1F,$D3,$78,$2B,$FE,$1F
	.dc.b	$C7,$97,$A1,$FD,$4F,$27,$E3,$FF,$D5,$7F,$27,$B7,$FD,$7B,$89,$1F
	.dc.b	$DC,$63,$E7,$3E,$2D,$7F,$AC,$E3,$FE,$0E,$8C,$42,$3F,$C8,$91,$8D
	.dc.b	$DB,$74,$DB,$D1,$3E,$FB,$FE,$C6,$71,$D7,$C2,$FC,$89,$B8,$FF,$69
	.dc.b	$87,$E5,$1F,$38,$FA,$99,$E2,$7F,$3F,$1F,$57,$D3,$FB,$55,$F8,$5D
	.dc.b	$7E,$FF,$1F,$FB,$7D,$BC,$5F,$0B,$0F,$EB,$BC,$15,$FC,$77,$E8,$BF
	.dc.b	$90,$C3,$F1,$CF,$E6,$5F,$EF,$7C,$1F,$EE,$98,$C4,$23,$F2,$31,$85
	.dc.b	$48,$FF,$1C,$4F,$F4,$87,$F5,$7D,$3F,$B0,$6B,$FC,$53,$B7,$87,$5F
	.dc.b	$EB,$63,$F9,$AF,$67,$F3,$9A,$FF,$20,$E5,$FB,$E9,$FD,$8C,$62,$3F
	.dc.b	$6D,$7E,$1E,$2E,$55,$FD,$0E,$F9,$47,$F0,$8B,$FF,$3F,$5F,$F0,$DE
	.dc.b	$CF,$E6,$7E,$07,$F2,$F9,$CC,$FF,$18,$9C,$F8,$9E,$DE,$5F,$E3,$11
	.dc.b	$FC,$6F,$C1,$C5,$7F,$8B,$37,$F8,$5B,$FF,$21,$F1,$3F,$97,$78,$9D
	.dc.b	$63,$18,$7F,$2D,$DF,$78,$7F,$50,$BD,$E7,$88,$FE,$25,$56,$BF,$8F
	.dc.b	$69,$FC,$F5,$F8,$FE,$16,$FC,$4F,$E5,$7E,$25,$BF,$E6,$18,$D7,$DF
	.dc.b	$B7,$5F,$17,$FA,$67,$8D,$FD,$8B,$3F,$EC,$D9,$FF,$66,$8F,$EB,$B1
	.dc.b	$88,$FE,$DB,$9F,$97,$6F,$EE,$9C,$0F,$A7,$FE,$F5,$BF,$F1,$E9,$FF
	.dc.b	$2A,$46,$23,$F4,$BC,$9F,$E5,$F8,$C4,$FF,$1F,$7F,$C1,$13,$F8,$AE
	.dc.b	$7F,$EE,$71,$8F,$1F,$FA,$65,$7F,$AB,$BB,$FC,$1E,$57,$F8,$52,$3F
	.dc.b	$33,$18,$C3,$FC,$DB,$1F,$E1,$7B,$EA,$F7,$F2,$FF,$C4,$4F,$FA,$02
	.dc.b	$3F,$C5,$13,$FE,$9C,$F3,$3B,$43,$CD,$FF,$17,$57,$E8,$F0,$F8,$11
	.dc.b	$8E,$EF,$E6,$B5,$BC,$3F,$89,$3F,$E4,$17,$FC,$7D,$7D,$FC,$EF,$F1
	.dc.b	$F3,$FE,$4A,$9D,$4F,$FA,$6E,$3D,$71,$87,$FD,$06,$7E,$CF,$C0,$FA
	.dc.b	$15,$FE,$75,$9F,$F2,$B5,$35,$FE,$E4,$FF,$95,$F6,$F9,$7E,$7F,$F9
	.dc.b	$77,$1F,$07,$A1,$FE,$64,$D7,$ED,$3F,$E6,$6C,$EE,$1E,$AF,$47,$FC
	.dc.b	$D8,$FD,$F7,$FC,$DD,$56,$3F,$CE,$ED,$7F,$DA,$DE,$97,$C4,$9F,$37
	.dc.b	$A7,$FE,$78,$9F,$BA,$DA,$FC,$57,$A9,$FD,$3B,$2B,$5F,$E9,$82,$9F
	.dc.b	$D9,$F6,$FF,$4C,$C7,$E4,$9A,$FF,$32,$3F,$CD,$CF,$FA,$86,$3F,$D2
	.dc.b	$AD,$8F,$C8,$7C,$DF,$E2,$5F,$B9,$1F,$EA,$D5,$E2,$7E,$5E,$75,$C5
	.dc.b	$FC,$83,$E2,$7F,$A3,$EA,$CD,$18,$DA,$E1,$F9,$8C,$7C,$82,$FF,$D7
	.dc.b	$78,$F4,$CF,$FD,$72,$7F,$45,$F5,$BD,$75,$FD,$AA,$FF,$D4,$13,$FE
	.dc.b	$ED,$8A,$FF,$AC,$27,$FD,$50,$FD,$03,$3C,$7F,$82,$7F,$CB,$BD,$5F
	.dc.b	$EB,$89,$71,$7F,$B1,$E7,$FD,$A2,$FF,$A9,$9F,$F6,$9F,$1F,$E1,$FD
	.dc.b	$9E,$1C,$DB,$BF,$CA,$65,$FE,$D2,$3F,$A3,$5F,$FA,$70,$FC,$91,$F0
	.dc.b	$6B,$C1,$CF,$C4,$FD,$44,$7F,$A5,$0F,$D8,$4F,$FB,$32,$7C,$DD,$5D
	.dc.b	$CF,$8D,$7F,$C0,$EB,$FC,$EF,$5F,$DC,$2B,$D9,$EB,$F9,$4F,$F1,$94
	.dc.b	$79,$3D,$7F,$E3,$5D,$FF,$E8,$1F,$D9,$BF,$CE,$3E,$2F,$F9,$56,$3F
	.dc.b	$CD,$E7,$F9,$A2,$3F,$CA,$78,$FF,$43,$9F,$F3,$EC,$FF,$92,$CD,$57
	.dc.b	$F5,$C7,$FA,$CD,$7F,$8F,$EB,$FC,$31,$7F,$E9,$2F,$7B,$F3,$9E,$CF
	.dc.b	$F8,$DB,$D6,$FF,$74,$FB,$3F,$DF,$2B,$FC,$54,$7F,$8C,$AF,$FC,$0D
	.dc.b	$C9,$FE,$0D,$8F,$F0,$BF,$83,$FE,$24,$7F,$16,$7F,$85,$DF,$EE,$7E
	.dc.b	$C7,$DB,$9F,$F0,$14,$FF,$AA,$6B,$FC,$47,$E9,$FD,$FF,$5F,$FC,$49
	.dc.b	$39,$7F,$C4,$39,$31,$FF,$0B,$DF,$F8,$C5,$FF,$0A,$3F,$E1,$9F,$5B
	.dc.b	$D1,$7F,$E3,$E3,$FC,$1C,$79,$28,$FF,$81,$75,$FD,$E7,$95,$FE,$06
	.dc.b	$8F,$F1,$3C,$EA,$3E,$57,$07,$F7,$9E,$0F,$EF,$98,$FF,$80,$63,$E6
	.dc.b	$CF,$F7,$6F,$43,$FD,$13,$1F,$D9,$4F,$B1,$1E,$09,$FE,$FD,$C7,$F6
	.dc.b	$23,$F0,$32,$F3,$18,$7F,$59,$E5,$FC,$3F,$89,$FD,$41,$FD,$27,$8B
	.dc.b	$F8,$FC,$BF,$86,$74,$7B,$23,$FA,$5E,$DF,$C1,$8F,$F2,$0D,$FF,$58
	.dc.b	$EA,$E3,$E5,$7F,$4E,$BE,$1D,$9F,$A0,$C3,$FA,$36,$3F,$C5,$3B,$7F
	.dc.b	$A1,$D7,$57,$F8,$46,$5D,$4F,$D6,$D7,$EC,$1F,$E3,$D9,$7F,$3C,$BF
	.dc.b	$F6,$CF,$81,$FD,$03,$CA,$FD,$55,$7F,$3A,$9F,$C3,$BF,$E8,$57,$0F
	.dc.b	$E1,$6F,$F8,$4D,$7F,$07,$9F,$F4,$04,$7F,$0C,$9F,$C7,$F8,$9F,$DD
	.dc.b	$3C,$AF,$8A,$79,$A3,$F8,$EC,$FF,$18,$7E,$EF,$89,$FE,$05,$F6,$BF
	.dc.b	$9E,$4F,$F2,$D8,$F4,$E1,$FC,$98,$FD,$DD,$7F,$26,$C3,$FD,$E7,$E0
	.dc.b	$BF,$18,$F8,$75,$FD,$73,$CA,$4B,$F6,$3C,$CF,$FB,$47,$C2,$CC,$62
	.dc.b	$3F,$DF,$BE,$56,$1E,$F9,$7F,$25,$C3,$11,$FD,$BA,$3F,$CF,$78,$3F
	.dc.b	$40,$E5,$F3,$DF,$CF,$E3,$15,$C8,$F2,$38,$79,$71,$EE,$BF,$A4,$E3
	.dc.b	$E8,$FC,$EF,$C9,$F3,$BC,$DF,$35,$D3,$FC,$33,$97,$EA,$B7,$C4,$79
	.dc.b	$5F,$F6,$1E,$4F,$B7,$F8,$B5,$FE,$92,$D7,$E2,$C6,$23,$F6,$79,$63
	.dc.b	$0F,$41,$FE,$F7,$AC,$3F,$E9,$AE,$1F,$D5,$3F,$D7,$63,$35,$FB,$C9
	.dc.b	$CC,$7D,$6F,$C2,$7F,$2E,$F1,$3E,$AB,$89,$E1,$78,$BF,$08,$3E,$64
	.dc.b	$FF,$67,$BF,$F3,$9C,$E2,$7F,$0B,$18,$7F,$03,$FA,$E7,$E4,$6F,$F8
	.dc.b	$06,$18,$9F,$BB,$1F,$EB,$47,$E3,$CF,$CE,$9D,$47,$BC,$F1,$33,$24
	.dc.b	$FF,$91,$A7,$F0,$1E,$DF,$19,$3F,$DF,$7D,$6F,$54,$FF,$BD,$8F,$C3
	.dc.b	$75,$7D,$CE,$5F,$CB,$BA,$F6,$1E,$23,$F1,$EF,$F5,$B6,$9A,$C3,$E2
	.dc.b	$56,$A7,$FD,$75,$BE,$65,$E4,$3C,$8F,$E3,$7D,$FF,$92,$CF,$FD,$2F
	.dc.b	$E3,$7F,$72,$BF,$D0,$F0,$27,$FD,$61,$1B,$3F,$D2,$57,$FC,$93,$CE
	.dc.b	$C4,$7B,$A3,$0F,$BB,$2F,$D8,$FA,$BE,$DC,$0F,$83,$2F,$D1,$B6,$B8
	.dc.b	$4B,$E9,$38,$1E,$28,$FC,$94,$3E,$27,$11,$F4,$CF,$A0,$7C,$F3,$F7
	.dc.b	$67,$E8,$8C,$1F,$47,$81,$E9,$BF,$11,$FC,$07,$D1,$7F,$B5,$AB,$F8
	.dc.b	$53,$F1,$7A,$7F,$C3,$47,$F8,$D7,$54,$66,$3F,$99,$1F,$E3,$9E,$6C
	.dc.b	$ED,$FE,$49,$ED,$FF,$21,$53,$F7,$BF,$C7,$63,$DC,$FE,$28,$53,$D3
	.dc.b	$5A,$3F,$80,$BF,$A3,$E3,$FF,$2A,$63,$FD,$6D,$BD,$39,$9F,$E0,$B3
	.dc.b	$FB,$37,$FC,$B0,$FF,$8F,$FE,$23,$F5,$9C,$7F,$C3,$3E,$26,$D7,$0A
	.dc.b	$FF,$76,$CF,$6E,$EF,$D9,$7E,$2D,$E8,$E5,$FD,$36,$9F,$CB,$B0,$FD
	.dc.b	$B9,$FE,$5F,$53,$AB,$F2,$C7,$F3,$DC,$3F,$CE,$58,$7F,$A4,$E7,$F8
	.dc.b	$EE,$57,$0E,$79,$7F,$A5,$F4,$E2,$78,$F1,$FC,$B4,$F9,$F1,$FF,$4E
	.dc.b	$47,$C6,$3E,$54,$78,$DF,$D7,$18,$E1,$7E,$4F,$F4,$E7,$D3,$CB,$8A
	.dc.b	$FB,$0D,$8F,$97,$C4,$BC,$47,$F2,$FE,$3B,$8F,$F0,$B8,$FE,$51,$EB
	.dc.b	$FE,$35,$EB,$D5,$FE,$B0,$9F,$F1,$0F,$1F,$E9,$E7,$FD,$0D,$39,$94
	.dc.b	$E7,$0F,$F1,$4C,$7F,$30,$AF,$E2,$75,$FA,$37,$FC,$44,$7F,$61,$7F
	.dc.b	$54,$E0,$FE,$74,$E1,$FF,$51,$1F,$EC,$2C,$11,$FB,$68,$FE,$61,$D9
	.dc.b	$FC,$27,$B3,$F8,$34,$79,$B8,$BF,$70,$FF,$68,$AF,$C6,$BC,$DF,$AA
	.dc.b	$FF,$05,$9F,$E3,$31,$8F,$03,$A3,$FE,$D2,$CB,$C3,$C3,$FB,$CB,$F8
	.dc.b	$6F,$66,$9E,$4E,$DF,$E3,$6F,$F0,$FA,$EF,$DF,$FC,$8D,$F8,$55,$FC
	.dc.b	$AF,$C0,$FE,$55,$1D,$CF,$2E,$DF,$5E,$FF,$D4,$31,$FC,$E1,$FF,$07
	.dc.b	$EF,$F5,$78,$BE,$2E,$9F,$D1,$6B,$FA,$0D,$7F,$A5,$2F,$E9,$F8,$3F
	.dc.b	$D5,$63,$FA,$39,$F7,$1E,$07,$F5,$38,$DC,$7F,$52,$7F,$CA,$9A,$FD
	.dc.b	$2F,$0B,$FA,$F7,$4F,$F2,$BC,$B1,$7F,$D7,$A3,$F1,$18,$AB,$C8,$7F
	.dc.b	$6B,$C3,$FD,$AB,$7C,$2F,$FD,$C0,$FF,$79,$C3,$EF,$F8,$7F,$DA,$EB
	.dc.b	$FB,$F4,$E2,$BF,$B5,$C7,$F8,$05,$C1,$F8,$83,$F1,$73,$FE,$09,$8F
	.dc.b	$CC,$B6,$BF,$17,$E2,$7E,$AB,$13,$FC,$8E,$33,$7F,$A4,$8F,$F0,$FE
	.dc.b	$1F,$92,$F1,$BF,$DD,$6D,$9F,$E3,$7E,$3F,$EC,$70,$FF,$15,$65,$F6
	.dc.b	$67,$F5,$37,$FE,$32,$BF,$CD,$F9,$3F,$C9,$AA,$D7,$EC,$A3,$FB,$85
	.dc.b	$FF,$91,$63,$F8,$CE,$5F,$A2,$F2,$BE,$F6,$DF,$6E,$7F,$DF,$47,$ED
	.dc.b	$9C,$63,$FD,$A2,$7B,$78,$3C,$EB,$FC,$8D,$92,$FF,$CA,$79,$FF,$93
	.dc.b	$FE,$BF,$FD,$43,$7F,$12,$FF,$C3,$FF,$9A,$E3,$83,$E6,$47,$F8,$8C
	.dc.b	$FB,$38,$FF,$3B,$AF,$CF,$F2,$FF,$81,$7B,$7F,$D5,$73,$FE,$54,$E7
	.dc.b	$FE,$FA,$61,$EE,$FF,$84,$A7,$FC,$11,$19,$3F,$C1,$1C,$DF,$BF,$AC
	.dc.b	$5F,$F0,$3C,$3F,$C1,$7D,$3F,$EA,$4C,$73,$E4,$F3,$CB,$FC,$2D,$97
	.dc.b	$0F,$27,$1E,$57,$FB,$4B,$2F,$E3,$9C,$3F,$ED,$4F,$2B,$F9,$26,$9F
	.dc.b	$1B,$0F,$EC,$EF,$F3,$BF,$27,$F9,$A3,$93,$FB,$75,$7F,$B7,$1F,$DC
	.dc.b	$CF,$F7,$03,$FD,$A8,$FF,$4D,$7E,$1F,$B5,$FD,$62,$3F,$AA,$9F,$E8
	.dc.b	$82,$F8,$BF,$D6,$5F,$CE,$1F,$E3,$67,$FA,$CB,$AF,$0B,$FB,$7C,$FF
	.dc.b	$53,$8F,$EC,$CF,$F5,$07,$FA,$97,$8D,$E8,$BF,$F0,$39,$FE,$DE,$7D
	.dc.b	$2F,$CF,$E2,$F8,$2E,$8F,$E7,$B8,$FF,$BE,$3A,$FF,$9E,$5F,$F9,$0B
	.dc.b	$83,$E6,$D7,$DB,$9F,$E7,$51,$F6,$A3,$19,$7E,$01,$D7,$17,$F9,$43
	.dc.b	$93,$EE,$46,$30,$FF,$36,$5F,$18,$FC,$4E,$1F,$CE,$3C,$4F,$E8,$18
	.dc.b	$7E,$16,$3F,$1F,$C4,$D3,$1E,$07,$F0,$FD,$7F,$CE,$D5,$FE,$8F,$CF
	.dc.b	$FD,$41,$97,$F9,$B6,$3F,$3D,$1F,$3B,$A3,$FE,$00,$C7,$EE,$A7,$FA
	.dc.b	$E7,$4E,$3B,$BF,$C2,$13,$FE,$BD,$BF,$EF,$DC,$DF,$D6,$6F,$FD,$AF
	.dc.b	$8F,$FA,$FB,$F4,$3F,$B1,$5F,$E6,$E3,$FD,$11,$E4,$7F,$B1,$BA,$3F
	.dc.b	$A1,$71,$7C,$B3,$FC,$DB,$F1,$1F,$EC,$68,$CC,$64,$C5,$7F,$4F,$EC
	.dc.b	$FF,$35,$13,$E3,$33,$1F,$A4,$8C,$9A,$D3,$FC,$9B,$1F,$3A,$BA,$0C
	.dc.b	$38,$B8,$3F,$CF,$F3,$C2,$7F,$CD,$95,$D0,$5F,$83,$9B,$11,$DF,$D2
	.dc.b	$FF,$46,$F3,$E2,$37,$82,$73,$19,$8F,$F2,$CE,$1E,$5C,$7E,$06,$5F
	.dc.b	$EE,$8E,$CF,$DE,$47,$A3,$4F,$F0,$2D,$7F,$AB,$8F,$F7,$73,$FE,$95
	.dc.b	$7F,$AD,$70,$7F,$12,$AF,$D6,$3F,$C3,$4C,$61,$FD,$87,$CB,$FF,$31
	.dc.b	$FE,$71,$FD,$8E,$47,$ED,$EF,$FC,$1E,$7F,$51,$F3,$BE,$EB,$F7,$4F
	.dc.b	$F5,$5D,$F1,$9F,$F6,$36,$3C,$F8,$ED,$79,$27,$F8,$2E,$DD,$FB,$9D
	.dc.b	$FF,$16,$B5,$18,$C3,$53,$F4,$EF,$F1,$EF,$B1,$F6,$1C,$B2,$F2,$DF
	.dc.b	$96,$7F,$84,$BE,$A3,$FD,$93,$DE,$D3,$D0,$F3,$BF,$E6,$DA,$BC,$5F
	.dc.b	$F2,$AE,$DE,$DE,$DF,$9D,$8F,$82,$FE,$3C,$7D,$48,$FF,$6B,$1F,$E3
	.dc.b	$EE,$0F,$B3,$C3,$F9,$BE,$9F,$BF,$3F,$8F,$8F,$C9,$47,$C1,$8F,$CD
	.dc.b	$72,$7E,$3E,$31,$CB,$F8,$BE,$7F,$F2,$2F,$17,$F0,$BA,$FC,$75,$7E
	.dc.b	$5F,$9F,$4F,$E6,$71,$EC,$D7,$F4,$3C,$FC,$B2,$FD,$65,$F7,$C3,$F2
	.dc.b	$AF,$E7,$F1,$D4,$BC,$BE,$75,$FB,$6C,$3F,$92,$61,$FE,$58,$EA,$FA
	.dc.b	$37,$FC,$00,$FF,$2D,$F8,$7F,$83,$CF,$FD,$43,$C1,$FD,$62,$7F,$09
	.dc.b	$E5,$FF,$09,$47,$D7,$F3,$7F,$BA,$FE,$9D,$F8,$DE,$43,$FE,$39,$F7
	.dc.b	$B3,$5C,$3A,$BF,$28,$BC,$1F,$EF,$BF,$5F,$C4,$FF,$20,$7C,$97,$F8
	.dc.b	$03,$F3,$6F,$C3,$F9,$1F,$DD,$73,$7E,$CF,$DC,$F1,$E3,$FE,$A6,$DF
	.dc.b	$51,$F8,$E7,$87,$E0,$9E,$53,$6F,$F8,$0B,$83,$FB,$CF,$C7,$7E,$A3
	.dc.b	$1F,$EF,$7D,$3E,$B3,$FC,$2E,$7E,$B2,$BF,$BF,$F5,$7B,$CE,$6F,$FA
	.dc.b	$48,$FF,$01,$F4,$FD,$1E,$90,$F2,$FF,$91,$F6,$FF,$11,$F5,$7F,$8E
	.dc.b	$CF,$F2,$09,$FE,$31,$87,$F1,$89,$70,$7E,$3A,$30,$FE,$5A,$BF,$92
	.dc.b	$57,$FB,$DB,$A7,$C7,$B7,$F8,$26,$3F,$7D,$DF,$FA,$9D,$3F,$D9,$6D
	.dc.b	$7F,$85,$F2,$7F,$0C,$8F,$D8,$71,$FF,$03,$F2,$3F,$C0,$4F,$EA,$DF
	.dc.b	$E0,$EF,$EB,$70,$B7,$FB,$EF,$7B,$F0,$BE,$4F,$F1,$77,$F8,$4B,$7A
	.dc.b	$BF,$83,$4F,$F1,$F9,$FE,$15,$7F,$8E,$9F,$24,$FF,$1E,$C3,$FD,$A3
	.dc.b	$E3,$FF,$17,$F1,$F3,$A7,$F8,$1F,$83,$F8,$D7,$8B,$FC,$03,$0F,$EB
	.dc.b	$EF,$EF,$DF,$F6,$49,$FD,$D7,$6F,$DF,$F0,$FF,$81,$F8,$3F,$D7,$7B
	.dc.b	$FE,$B3,$8B,$FA,$AF,$1F,$EB,$3C,$5F,$AD,$C9,$FB,$5C,$3F,$B0,$61
	.dc.b	$FB,$0E,$5C,$F3,$7D,$F9,$FF,$54,$E3,$F4,$B4,$FF,$02,$3F,$CC,$70
	.dc.b	$FE,$AB,$C9,$FA,$33,$FD,$38,$E1,$FF,$1F,$36,$7F,$8D,$2E,$9F,$CB
	.dc.b	$47,$F8,$77,$AB,$F2,$2F,$E4,$4F,$F0,$AB,$DF,$AF,$F4,$51,$FE,$18
	.dc.b	$77,$CD,$FE,$17,$7C,$9D,$9F,$AE,$9F,$F0,$EF,$6F,$EC,$9F,$F6,$FB
	.dc.b	$E9,$F0,$3F,$84,$E7,$FB,$99,$FF,$71,$C7,$EF,$DB,$1F,$D2,$DF,$E2
	.dc.b	$0F,$BC,$FF,$75,$3F,$C4,$1F,$E1,$B9,$FF,$10,$8F,$F3,$AE,$3F,$A0
	.dc.b	$AE,$FE,$57,$F3,$FE,$7F,$E1,$B8,$73,$3F,$80,$C3,$D3,$E1,$19,$3F
	.dc.b	$82,$F9,$1F,$AA,$CF,$FD,$99,$EA,$6F,$04,$62,$75,$5C,$47,$B7,$E1
	.dc.b	$9F,$82,$F1,$7E,$1C,$79,$7E,$93,$FB,$C6,$3F,$B8,$7F,$BD,$5F,$EE
	.dc.b	$B8,$3F,$B0,$F8,$1E,$4C,$BF,$8D,$E1,$EB,$8F,$E3,$78,$6E,$3C,$47
	.dc.b	$C1,$E2,$FB,$FC,$1F,$E3,$99,$FE,$0B,$18,$9E,$8F,$F8,$0B,$1F,$F1
	.dc.b	$F1,$A8,$FA,$4F,$1C,$13,$EA,$8F,$E1,$FF,$95,$7E,$1A,$BC,$74,$8C
	.dc.b	$47,$F4,$29,$FE,$2B,$E2,$7F,$84,$32,$EF,$7F,$9B,$7F,$BF,$E3,$AB
	.dc.b	$F5,$DF,$F1,$CE,$6F,$CC,$C7,$8D,$AF,$C5,$7F,$CA,$F1,$FE,$1B,$38
	.dc.b	$47,$EA,$5F,$F0,$F4,$E2,$3F,$C1,$47,$E1,$FB,$3F,$DA,$8F,$F5,$88
	.dc.b	$FE,$57,$18,$8F,$C6,$74,$FD,$AF,$23,$F9,$7F,$77,$EE,$FE,$46,$D7
	.dc.b	$F6,$0F,$37,$FC,$A5,$F2,$58,$8F,$EE,$1D,$DF,$DE,$37,$FE,$6D,$B7
	.dc.b	$F9,$4F,$F0,$4D,$78,$1F,$E4,$FF,$03,$85,$FF,$96,$F3,$E5,$E1,$67
	.dc.b	$1F,$F0,$73,$FE,$55,$C3,$38,$FF,$96,$B9,$3F,$A1,$FC,$97,$F1,$78
	.dc.b	$FF,$37,$CE,$34,$FF,$03,$C7,$F9,$CB,$3C,$47,$F8,$62,$3F,$CE,$98
	.dc.b	$AF,$E0,$3F,$D2,$F3,$FF,$38,$EB,$FE,$5D,$BF,$F3,$04,$7F,$A0,$F4
	.dc.b	$FE,$A7,$F5,$7F,$CA,$A3,$FA,$ED,$62,$B7,$1F,$EB,$3A,$C4,$7F,$8A
	.dc.b	$FD,$17,$1F,$FA,$96,$3F,$CC,$C7,$F7,$AE,$3C,$F8,$FF,$57,$9A,$CE
	.dc.b	$8F,$C8,$57,$F8,$FB,$A3,$FC,$17,$1F,$E4,$3F,$1F,$FD,$AD,$18,$E9
	.dc.b	$47,$F5,$0E,$CF,$F2,$4E,$58,$C3,$F6,$F3,$99,$FF,$36,$8F,$4B,$FC
	.dc.b	$E3,$8F,$E6,$0E,$A7,$F8,$6A,$FF,$AF,$BF,$E2,$63,$FC,$8D,$8F,$93
	.dc.b	$C3,$FF,$13,$F1,$FF,$9E,$7C,$DF,$F3,$5B,$67,$75,$FE,$24,$AF,$EC
	.dc.b	$EF,$F6,$3E,$4F,$CA,$79,$7F,$DD,$23,$F3,$BB,$23,$D1,$A7,$FA,$13
	.dc.b	$4F,$E4,$51,$88,$F5,$4E,$F6,$FF,$46,$6B,$FE,$68,$C3,$F6,$7C,$1F
	.dc.b	$2D,$FF,$1A,$D7,$FB,$66,$3E,$91,$F5,$E3,$FC,$D2,$7F,$01,$71,$7F
	.dc.b	$DD,$E3,$FD,$99,$4F,$21,$CD,$F8,$7E,$5F,$F3,$5B,$79,$B4,$E7,$9B
	.dc.b	$87,$26,$6F,$F2,$97,$FE,$CF,$9E,$9C,$7F,$B7,$9F,$CD,$65,$FA,$0E
	.dc.b	$71,$FD,$9E,$33,$1F,$A1,$8C,$F5,$72,$8E,$10,$EA,$FF,$01,$C7,$F9
	.dc.b	$7B,$AF,$FC,$EF,$DD,$FE,$7B,$AF,$F6,$77,$66,$AB,$FD,$A3,$1F,$ED
	.dc.b	$CB,$C6,$9F,$BD,$CB,$F8,$1C,$FF,$B8,$2F,$87,$6F,$EF,$67,$FD,$19
	.dc.b	$DD,$FC,$31,$B7,$FE,$82,$6B,$F0,$5F,$F7,$07,$5F,$E8,$38,$FE,$77
	.dc.b	$7F,$E9,$E4,$FF,$1D,$E7,$FA,$59,$FF,$3D,$9F,$8B,$AF,$F2,$93,$FA
	.dc.b	$2B,$87,$FD,$7E,$7F,$AB,$4F,$E9,$CE,$79,$BF,$4B,$18,$9F,$E6,$EF
	.dc.b	$F4,$39,$FE,$80,$FE,$29,$AF,$43,$5C,$3F,$D1,$DF,$EB,$F1,$FD,$35
	.dc.b	$FF,$46,$D7,$E3,$BB,$BF,$15,$E0,$7F,$50,$9F,$F6,$05,$7F,$58,$7F
	.dc.b	$A5,$B5,$ED,$8F,$B3,$C1,$FE,$C5,$1D,$9F,$EC,$CF,$F4,$39,$E1,$87
	.dc.b	$F0,$03,$E1,$F8,$5F,$1C,$FE,$EF,$3E,$D6,$BF,$E0,$F9,$78,$7F,$E0
	.dc.b	$06,$C7,$FB,$FB,$BF,$CF,$E2,$7D,$13,$FC,$13,$97,$F8,$B2,$7F,$C1
	.dc.b	$DE,$2F,$39,$FF,$0A,$EB,$99,$FF,$74,$F8,$DF,$03,$C7,$FB,$47,$F8
	.dc.b	$93,$0F,$F1,$7C,$FF,$BA,$7C,$5F,$27,$A3,$FC,$F3,$D9,$FF,$7F,$78
	.dc.b	$FF,$E2,$8A,$D1,$FD,$67,$AB,$FD,$BD,$1F,$D4,$39,$BF,$52,$43,$F5
	.dc.b	$6F,$EA,$BA,$F1,$19,$8F,$DE,$D7,$F4,$0E,$BF,$F4,$F9,$FC,$FB,$F1
	.dc.b	$DF,$F0,$BF,$37,$CD,$F0,$FF,$8A,$3F,$E8,$FF,$D4,$F2,$FE,$33,$C3
	.dc.b	$AB,$D7,$FE,$4F,$8F,$E1,$23,$EB,$03,$5E,$52,$F8,$C7,$F1,$F9,$FF
	.dc.b	$29,$79,$2C,$BF,$CE,$5E,$47,$F1,$3D,$3F,$81,$57,$F0,$5D,$BF,$84
	.dc.b	$F5,$FC,$37,$DE,$7E,$FF,$87,$FD,$EF,$CD,$FE,$54,$AF,$E3,$DC,$7F
	.dc.b	$3A,$3F,$CC,$BE,$17,$FB,$82,$7F,$DC,$DE,$6F,$F9,$31,$FF,$55,$72
	.dc.b	$7F,$94,$1F,$66,$FF,$91,$F5,$55,$B8,$F7,$88,$FF,$73,$CF,$F9,$57
	.dc.b	$D5,$AF,$EB,$4F,$F7,$1B,$CA,$FF,$27,$1E,$0F,$03,$FD,$DB,$87,$F9
	.dc.b	$56,$3C,$38,$7F,$60,$F4,$BE,$65,$FF,$95,$27,$A7,$AB,$FE,$EC,$7F
	.dc.b	$CE,$0F,$F9,$DA,$3F,$53,$5E,$6B,$FF,$27,$57,$FB,$CE,$3F,$D7,$53
	.dc.b	$A7,$DA,$7F,$BD,$23,$FC,$F4,$D7,$D4,$F5,$F1,$7F,$32,$7F,$A7,$BD
	.dc.b	$0F,$DB,$5F,$F9,$5A,$BE,$4C,$7F,$BE,$1F,$F4,$76,$15,$FE,$20,$F0
	.dc.b	$C3,$F8,$D6,$1E,$76,$DF,$E3,$0F,$F2,$54,$7F,$B4,$4F,$F7,$FD,$7F
	.dc.b	$A6,$5A,$FF,$1B,$7A,$79,$1F,$AE,$3F,$D8,$0F,$C5,$6B,$FE,$0C,$AF
	.dc.b	$E7,$7C,$9F,$A6,$1E,$9F,$EC,$DC,$03,$F3,$11,$FD,$2B,$93,$F6,$BF
	.dc.b	$80,$FF,$94,$7A,$7F,$0B,$83,$5F,$F2,$6E,$1C,$8E,$3E,$77,$F5,$9C
	.dc.b	$3F,$DB,$B8,$FE,$75,$FD,$F6,$3F,$E4,$6F,$03,$F3,$F1,$FD,$2A,$7F
	.dc.b	$80,$3F,$A2,$9F,$F2,$26,$BD,$70,$F3,$75,$FF,$8F,$9C,$6B,$5F,$14
	.dc.b	$7F,$8E,$A7,$FC,$5B,$9F,$E7,$3D,$2F,$F0,$4E,$BF,$ED,$93,$FE,$00
	.dc.b	$3F,$B7,$9F,$CA,$C7,$FB,$12,$31,$1F,$02,$3E,$07,$83,$FC,$33,$87
	.dc.b	$FC,$79,$3C,$27,$F8,$35,$E2,$3E,$3F,$27,$F0,$9B,$FF,$13,$3F,$B0
	.dc.b	$7F,$85,$47,$1A,$C7,$77,$F8,$E6,$1C,$3F,$C5,$BF,$3C,$FF,$1E,$C6
	.dc.b	$DA,$F9,$7D,$7F,$E3,$2F,$9E,$2D,$FD,$89,$FF,$1F,$CE,$7C,$1F,$0B
	.dc.b	$ED,$CF,$FC,$7D,$8F,$F2,$4D,$3F,$8E,$4F,$A9,$B8,$7F,$06,$D3,$F9
	.dc.b	$26,$1F,$CB,$B0,$CC,$FE,$D7,$97,$E4,$BF,$88,$84,$62,$34,$FE,$61
	.dc.b	$FE,$26,$FD,$68,$FE,$27,$E0,$7F,$40,$C7,$F0,$5C,$3F,$9A,$8C,$3F
	.dc.b	$85,$E1,$FF,$1D,$CF,$F4,$58,$FE,$65,$3F,$8C,$E4,$FE,$97,$C9,$FC
	.dc.b	$CB,$9B,$F2,$3C,$FF,$E2,$DF,$4A,$DF,$E5,$EF,$F4,$2F,$D2,$C7,$A6
	.dc.b	$5F,$E3,$3D,$3D,$93,$FA,$7A,$FF,$21,$E5,$FC,$E7,$3F,$C9,$CF,$EC
	.dc.b	$9F,$EA,$FC,$FF,$E2,$AC,$7F,$A7,$6B,$FC,$0E,$BF,$C5,$71,$8D,$BF
	.dc.b	$A8,$C6,$23,$F2,$5C,$3F,$E2,$E7,$F8,$34,$FF,$82,$E7,$FC,$59,$E8
	.dc.b	$7F,$BE,$2B,$55,$FE,$8E,$7B,$66,$8F,$EB,$D1,$F3,$A3,$F0,$15,$FE
	.dc.b	$2A,$EA,$F0,$FA,$1B,$BF,$E3,$B8,$F3,$AF,$F1,$54,$7F,$86,$F2,$FE
	.dc.b	$4F,$18,$9F,$D8,$9E,$5E,$CE,$BE,$87,$3D,$3A,$61,$EE,$8C,$3D,$EB
	.dc.b	$B4,$57,$F9,$31,$FD,$CF,$4F,$F4,$04,$66,$3F,$9F,$9F,$EA,$CC,$3F
	.dc.b	$A1,$3E,$4F,$43,$AD,$FC,$1D,$3E,$44,$62,$11,$ED,$EF,$FE,$B3,$7F
	.dc.b	$D7,$2F,$F4,$07,$FA,$3E,$3E,$3F,$9D,$C6,$FE,$67,$83,$FE,$07,$BF
	.dc.b	$EE,$11,$FD,$9C,$CE,$7F,$E2,$FC,$3F,$B7,$F8,$5F,$DC,$AB,$71,$83
	.dc.b	$47,$F5,$F7,$FC,$59,$AF,$FB,$0E,$3F,$C6,$6F,$FB,$16,$BF,$C0,$B7
	.dc.b	$FC,$2F,$C8,$AF,$EF,$7C,$37,$C8,$7F,$82,$D7,$DB,$FF,$86,$9B,$3F
	.dc.b	$85,$C3,$EA,$BF,$87,$F6,$7F,$11,$3A,$8F,$60,$87,$9E,$9D,$57,$FB
	.dc.b	$3E,$B1,$F6,$FF,$E3,$2D,$BF,$C0,$3D,$FF,$E6,$0F,$43,$AB,$71,$E4
	.dc.b	$F9,$FB,$7F,$CD,$7C,$5F,$E0,$87,$FC,$DF,$8A,$FC,$B1,$87,$C7,$9A
	.dc.b	$3E,$07,$DF,$FE,$1F,$87,$FC,$33,$C5,$E7,$F8,$8F,$F4,$36,$3C,$CC
	.dc.b	$1F,$0C,$87,$7D,$32,$FF,$97,$B1,$7E,$97,$FE,$96,$D3,$FC,$3A,$FE
	.dc.b	$72,$BF,$3F,$A7,$CA,$8C,$5F,$EB,$63,$CF,$1F,$57,$F4,$DF,$D0,$67
	.dc.b	$FA,$6C,$3F,$C4,$FF,$B8,$7E,$3F,$0F,$81,$EB,$FC,$FF,$9E,$FD,$B7
	.dc.b	$A9,$FD,$CF,$E2,$BE,$BC,$63,$C6,$FE,$FF,$1F,$12,$31,$1F,$E0,$5F
	.dc.b	$B1,$FF,$19,$C7,$EF,$FC,$BF,$F2,$9F,$89,$BC,$BF,$DC,$0F,$47,$FD
	.dc.b	$C5,$F9,$6F,$07,$F8,$67,$81,$F8,$85,$F8,$1B,$5F,$34,$68,$FC,$6F
	.dc.b	$C7,$FF,$92,$72,$FE,$1D,$A7,$F8,$2F,$F3,$5F,$05,$7F,$80,$7F,$35
	.dc.b	$EC,$FF,$B8,$BC,$CF,$F7,$8F,$E1,$FA,$F4,$7F,$CD,$2F,$96,$75,$2F
	.dc.b	$D1,$78,$F8,$9E,$5E,$BF,$27,$6B,$E5,$FE,$05,$FF,$A4,$39,$3F,$13
	.dc.b	$18,$8F,$F4,$B4,$7E,$8A,$31,$1F,$E9,$AE,$4F,$F2,$DC,$7D,$5B,$EB
	.dc.b	$F4,$9E,$CE,$0F,$F1,$7F,$07,$E5,$23,$11,$FE,$35,$E0,$FF,$31,$FE
	.dc.b	$F5,$88,$FF,$4E,$D7,$E6,$A3,$11,$FE,$37,$9F,$F4,$A7,$BA,$FE,$01
	.dc.b	$FF,$43,$37,$87,$F2,$E7,$F3,$9F,$7F,$7F,$B5,$7E,$32,$3F,$9E,$7C
	.dc.b	$3F,$EE,$29,$39,$9F,$E3,$7E,$3A,$5F,$9C,$F6,$3F,$D2,$30,$F8,$27
	.dc.b	$EE,$7F,$E9,$76,$DF,$88,$FE,$A1,$97,$F8,$8F,$C4,$FF,$59,$FC,$4F
	.dc.b	$F0,$8A,$47,$F0,$5C,$3A,$77,$F3,$3B,$E1,$D0,$FF,$69,$79,$5F,$D6
	.dc.b	$F2,$FA,$13,$67,$C0,$7E,$EC,$E1,$D7,$FC,$A5,$EE,$6B,$3F,$F5,$F3
	.dc.b	$83,$FD,$56,$58,$FC,$E4,$7F,$98,$61,$DF,$F7,$E3,$FB,$63,$FE,$B4
	.dc.b	$F2,$2C,$F5,$F6,$BF,$49,$DD,$FA,$5C,$7F,$4D,$F8,$5F,$E9,$F4,$FE
	.dc.b	$8C,$DC,$FF,$98,$D6,$DE,$F8,$79,$F1,$46,$23,$F9,$9E,$7F,$E7,$58
	.dc.b	$C1,$F9,$F7,$7B,$79,$63,$D9,$3F,$CF,$A7,$84,$FD,$C9,$F8,$F3,$9D
	.dc.b	$BB,$ED,$FE,$D3,$9C,$CF,$F1,$CC,$F2,$FD,$0E,$BF,$E0,$1F,$01,$FC
	.dc.b	$06,$3E,$76,$08,$C4,$7F,$4A,$9F,$F0,$A4,$7D,$19,$FE,$C4,$D7,$FD
	.dc.b	$E7,$C5,$FD,$3A,$3C,$35,$F9,$D9,$E9,$8F,$FB,$9F,$DF,$FC,$BC,$F8
	.dc.b	$1A,$F7,$3F,$CA,$7C,$DF,$9F,$5E,$67,$FB,$FE,$7F,$47,$1F,$E2,$63
	.dc.b	$F4,$F9,$FF,$96,$BA,$7F,$43,$3F,$AF,$9E,$9E,$43,$B7,$F6,$FD,$AE
	.dc.b	$DF,$F0,$24,$63,$BB,$FB,$3D,$3B,$F7,$5F,$E0,$97,$59,$A7,$F8,$B6
	.dc.b	$3F,$37,$A7,$1E,$07,$F1,$2F,$08,$3C,$3F,$E6,$71,$98,$78,$7F,$E1
	.dc.b	$58,$C4,$F8,$BB,$FF,$53,$8F,$E2,$FC,$6F,$E5,$79,$78,$67,$FD,$79
	.dc.b	$8F,$6F,$17,$FC,$3D,$18,$87,$8D,$CE,$BF,$C9,$DC,$5F,$EC,$5E,$0D
	.dc.b	$C6,$23,$F9,$1E,$FD,$63,$FA,$21,$FE,$44,$F1,$FF,$C4,$98,$FF,$94
	.dc.b	$74,$FE,$99,$E4,$7F,$4B,$7F,$41,$E4,$FF,$56,$BE,$41,$F8,$7A,$43
	.dc.b	$E0,$F5,$EA,$3F,$AE,$E5,$C3,$CC,$FF,$6D,$79,$7E,$6C,$11,$FD,$AD
	.dc.b	$D6,$BF,$DA,$AB,$EA,$78,$5F,$23,$CD,$FC,$57,$6F,$F9,$07,$CE,$FE
	.dc.b	$85,$87,$D9,$7F,$BF,$BF,$E7,$1F,$0F,$FB,$3C,$62,$3F,$C1,$9B,$FF
	.dc.b	$82,$9F,$F3,$B6,$C3,$F8,$AB,$FC,$D2,$1E,$7F,$66,$C7,$F8,$5B,$C5
	.dc.b	$FF,$06,$3F,$89,$9F,$F0,$EE,$1F,$90,$F4,$3F,$C4,$53,$CF,$D1,$E4
	.dc.b	$7F,$8A,$5F,$C8,$3F,$E2,$BD,$AC,$79,$32,$F2,$1F,$E2,$AC,$BF,$D2
	.dc.b	$D9,$7F,$8F,$DB,$3F,$E7,$CC,$7F,$C0,$BE,$97,$F8,$F6,$5E,$A7,$F3
	.dc.b	$F7,$FC,$8F,$5F,$E5,$0A,$FF,$05,$FA,$DF,$E1,$FF,$5A,$CF,$F3,$C7
	.dc.b	$F5,$31,$FE,$81,$CB,$FD,$4F,$EB,$FC,$33,$FC,$4D,$EA,$FF,$82,$B9
	.dc.b	$9C,$D0,$83,$E4,$3F,$C2,$B9,$7E,$02,$32,$6F,$D1,$FC,$7F,$5F,$E2
	.dc.b	$CF,$D7,$FB,$FC,$3B,$31,$1F,$90,$9F,$E6,$95,$C8,$F0,$E1,$D9,$FF
	.dc.b	$6D,$C7,$EB,$E7,$8D,$F8,$BC,$3E,$71,$FB,$C9,$FE,$03,$8F,$E7,$63
	.dc.b	$71,$8C,$77,$1F,$E7,$1F,$2B,$F8,$8E,$3E,$DA,$FE,$29,$7F,$EE,$58
	.dc.b	$FE,$F9,$E6,$FC,$9D,$FE,$D9,$89,$FE,$40,$7F,$B0,$79,$FF,$C0,$5F
	.dc.b	$31,$B9,$FC,$0F,$AD,$FC,$D5,$E2,$7F,$7D,$AF,$C3,$3D,$3E,$37,$FD
	.dc.b	$0B,$1F,$CE,$0E,$85,$3F,$D8,$7E,$0F,$59,$FE,$E7,$E2,$5F,$57,$FB
	.dc.b	$5D,$76,$9F,$C8,$FB,$9F,$ED,$BF,$33,$83,$FD,$BA,$7F,$CA,$5D,$FF
	.dc.b	$DA,$9B,$3A,$8F,$ED,$15,$CC,$57,$F6,$CC,$3C,$13,$ED,$9F,$6B,$E7
	.dc.b	$BF,$F5,$37,$A7,$FE,$73,$7F,$AA,$65,$F1,$73,$B1,$F0,$AF,$E2,$FA
	.dc.b	$4D,$BD,$D1,$FD,$4A,$7E,$79,$FE,$F5,$DF,$0F,$96,$3F,$A4,$BF,$67
	.dc.b	$BF,$ED,$BF,$8D,$E2,$FC,$6C,$7E,$0F,$97,$F1,$2D,$7F,$31,$D1,$FC
	.dc.b	$9B,$A3,$F0,$F3,$F9,$19,$CB,$F9,$8A,$E8,$D7,$F4,$7D,$3F,$CA,$67
	.dc.b	$AF,$0F,$F4,$D5,$87,$F2,$6A,$FD,$05,$FF,$85,$7A,$BC,$D8,$7F,$B1
	.dc.b	$7B,$7F,$C9,$3D,$3F,$C0,$B4,$F8,$DA,$FF,$1F,$FA,$FF,$F5,$CF,$85
	.dc.b	$FE,$F5,$F8,$0F,$E4,$5D,$1F,$EC,$6B,$FF,$2E,$C7,$F2,$5F,$87,$FF
	.dc.b	$62,$47,$E4,$3A,$FF,$8A,$FB,$1F,$EC,$2F,$1B,$FD,$FD,$9F,$2A,$FF
	.dc.b	$6B,$46,$21,$7F,$CB,$F4,$78,$EB,$FC,$DF,$B7,$D3,$A3,$FD,$83,$DD
	.dc.b	$99,$6B,$F9,$47,$27,$17,$F9,$3E,$3F,$EC,$48,$FE,$55,$AF,$EC,$35
	.dc.b	$FF,$60,$C7,$FB,$3B,$4E,$5A,$7F,$43,$8F,$F1,$F9,$FA,$CB,$F3,$79
	.dc.b	$9E,$1A,$FE,$8D,$A7,$3D,$3F,$8D,$C7,$E9,$F1,$FF,$4F,$F4,$FF,$B0
	.dc.b	$39,$FF,$91,$7D,$2F,$D8,$D3,$EB,$47,$F2,$08,$FE,$21,$3B,$8F,$E5
	.dc.b	$30,$E3,$FE,$41,$D5,$FC,$8F,$EB,$BF,$87,$5F,$F8,$3D,$7C,$A6,$A7
	.dc.b	$FB,$94,$7F,$32,$C7,$F1,$98,$FF,$33,$FA,$FF,$16,$1F,$E2,$87,$C3
	.dc.b	$C3,$FE,$07,$9F,$E6,$F1,$E1,$AF,$CC,$61,$FD,$FB,$EB,$FC,$37,$F5
	.dc.b	$FB,$FF,$99,$C7,$8A,$B7,$5F,$E2,$68,$F1,$78,$BF,$E9,$2F,$69,$5F
	.dc.b	$A7,$AF,$F0,$4E,$9F,$CA,$6B,$D9,$AE,$9F,$76,$9C,$3D,$BE,$58,$7F
	.dc.b	$8E,$F4,$F4,$4F,$B7,$C2,$D4,$62,$3F,$A9,$ED,$FC,$F6,$FE,$07,$07
	.dc.b	$F3,$9A,$D5,$2B,$EB,$79,$5F,$17,$5D,$57,$F0,$19,$FA,$9F,$05,$FE
	.dc.b	$58,$8F,$E3,$9F,$09,$FE,$C0,$7F,$BC,$74,$7E,$16,$75,$CF,$F8,$83
	.dc.b	$F4,$FC,$5F,$CF,$AB,$F6,$70,$BF,$E7,$91,$FC,$9F,$C1,$FF,$5C,$79
	.dc.b	$3F,$D0,$DF,$F3,$F3,$F2,$E3,$F9,$45,$7F,$86,$7A,$7F,$A0,$CF,$FB
	.dc.b	$02,$7F,$58,$FF,$AF,$DF,$E6,$73,$FD,$2A,$DD,$BF,$D0,$23,$FA,$04
	.dc.b	$7F,$44,$9F,$2B,$FC,$E1,$E3,$1F,$D6,$A3,$FD,$6D,$C9,$FE,$CE,$F4
	.dc.b	$7F,$A3,$57,$B3,$17,$C0,$E2,$3E,$3F,$89,$71,$EF,$BF,$EF,$67,$F7
	.dc.b	$DB,$FF,$51,$D3,$11,$FC,$17,$EB,$BF,$A5,$78,$78,$6C,$71,$AE,$13
	.dc.b	$FD,$EB,$0F,$5E,$3F,$EB,$AD,$7F,$B6,$6D,$FE,$B9,$7D,$15,$BA,$FF
	.dc.b	$50,$E9,$FD,$57,$1F,$83,$85,$7E,$FE,$96,$BE,$3E,$95,$F9,$87,$FB
	.dc.b	$0A,$7E,$4B,$5F,$EC,$5A,$FF,$1F,$C7,$EC,$CA,$1C,$3B,$8F,$E1,$9E
	.dc.b	$07,$FA,$EF,$87,$F0,$4D,$C3,$F1,$3C,$5C,$3C,$BF,$EC,$11,$E0,$7F
	.dc.b	$23,$6D,$3F,$8C,$46,$1F,$E4,$98,$FF,$62,$9F,$C0,$9F,$EA,$3E,$6F
	.dc.b	$D1,$79,$5F,$D8,$BA,$3F,$C5,$DC,$FF,$C5,$63,$11,$FE,$3A,$8F,$EA
	.dc.b	$B1,$FB,$5C,$BF,$51,$D1,$FE,$AD,$EA,$FD,$8C,$E6,$F9,$F5,$E6,$FF
	.dc.b	$11,$9F,$FA,$C1,$FE,$89,$E8,$7F,$90,$3B,$FF,$C9,$B8,$7F,$09,$CB
	.dc.b	$FD,$67,$1F,$E4,$D8,$C4,$23,$F9,$2C,$7D,$DD,$7D,$EF,$F1,$69,$78
	.dc.b	$1D,$3C,$2F,$F5,$73,$CD,$BE,$0F,$FA,$83,$5F,$83,$55,$F8,$1E,$05
	.dc.b	$7F,$94,$ED,$FD,$E6,$7F,$99,$C5,$7F,$8F,$F9,$55,$F4,$DF,$F7,$B9
	.dc.b	$FE,$8F,$15,$FE,$87,$75,$78,$9F,$3B,$1F,$E5,$99,$BC,$5F,$EA,$DA
	.dc.b	$FD,$F3,$FB,$54,$7F,$5D,$EB,$CC,$7E,$EE,$31,$C3,$FD,$7E,$3F,$D5
	.dc.b	$4F,$F6,$87,$8F,$AB,$FD,$C6,$BF,$B9,$79,$1F,$DC,$6F,$FD,$47,$7F
	.dc.b	$DA,$0F,$EE,$B7,$FE,$07,$3F,$DD,$9E,$4F,$83,$A7,$C7,$E5,$7F,$80
	.dc.b	$FC,$1F,$F0,$06,$08,$FF,$03,$79,$5F,$EB,$38,$C6,$3F,$E0,$FC,$7F
	.dc.b	$8F,$77,$FF,$BD,$F8,$BF,$C3,$11,$EB,$6D,$F1,$F5,$7F,$4F,$87,$FA
	.dc.b	$83,$0F,$E1,$DE,$27,$5F,$33,$FC,$95,$97,$F5,$4F,$C1,$FF,$C8,$FE
	.dc.b	$67,$82,$FD,$87,$F8,$13,$D4,$57,$F5,$AE,$1F,$F5,$24,$FF,$86,$7E
	.dc.b	$77,$FC,$39,$1F,$1E,$3F,$C9,$C7,$F8,$76,$B8,$DB,$F2,$8F,$F6,$2C
	.dc.b	$AF,$FC,$30,$FC,$BE,$5F,$F1,$05,$F0,$E7,$FF,$6D,$3C,$BF,$33,$FB
	.dc.b	$FE,$5F,$F2,$9C,$3F,$C5,$9D,$5F,$EB,$4F,$13,$FB,$56,$5F,$E2,$DF
	.dc.b	$4B,$FD,$4D,$FB,$9F,$F8,$4F,$CE,$FF,$08,$78,$08,$FF,$51,$46,$1F
	.dc.b	$F0,$13,$FE,$31,$F1,$FF,$D6,$3F,$BA,$FF,$77,$C3,$F6,$BE,$1F,$FA
	.dc.b	$DE,$3F,$C0,$55,$B8,$C6,$1E,$2A,$74,$FF,$3C,$5E,$DF,$F2,$0E,$5F
	.dc.b	$F5,$8C,$E7,$E4,$FF,$AA,$C7,$FA,$D2,$FF,$D7,$DE,$DF,$A7,$0F,$EF
	.dc.b	$2F,$A1,$BD,$FD,$FD,$7F,$F6,$3F,$C7,$F9,$E5,$18,$F2,$75,$AF,$FB
	.dc.b	$3E,$BF,$C2,$F5,$FC,$D4,$FE,$A3,$D1,$E5,$F3,$BD,$94,$20,$E1,$FF
	.dc.b	$61,$46,$A3,$31,$FE,$25,$8C,$47,$18,$FF,$1B,$47,$D7,$D7,$F9,$5E
	.dc.b	$BF,$EE,$8D,$7F,$C2,$78,$FB,$B0,$FF,$09,$CF,$C2,$D7,$FC,$57,$FA
	.dc.b	$77,$F9,$96,$3E,$2F,$9B,$8B,$E1,$8A,$B5,$5F,$EF,$0C,$7F,$C3,$67
	.dc.b	$E5,$EB,$FB,$97,$C0,$3C,$77,$B0,$FD,$97,$F1,$F5,$FE,$73,$FA,$AF
	.dc.b	$D1,$CE,$63,$FD,$5B,$DD,$C6,$3F,$A9,$C7,$EA,$8F,$F3,$F6,$1F,$E1
	.dc.b	$EF,$0B,$F0,$4D,$8F,$C2,$78,$AF,$80,$FF,$07,$E7,$FE,$E0,$97,$B3
	.dc.b	$E7,$7F,$4B,$5F,$E1,$E7,$EC,$1F,$E0,$BE,$6F,$F1,$3B,$83,$F5,$66
	.dc.b	$CF,$20,$C3,$FC,$40,$E1,$B8,$F9,$BC,$3F,$F7,$2B,$E8,$F3,$FF,$C3
	.dc.b	$8F,$E6,$EF,$FC,$3F,$6E,$FF,$33,$F0,$2F,$6F,$F8,$73,$0F,$EE,$8F
	.dc.b	$F8,$7B,$E3,$BF,$49,$E2,$7E,$97,$87,$E1,$E5,$FE,$22,$6D,$7F,$89
	.dc.b	$EB,$73,$DE,$3D,$C7,$77,$0F,$AF,$3F,$F1,$13,$A1,$9F,$F8,$DA,$5E
	.dc.b	$2F,$A2,$7E,$26,$7F,$E2,$19,$F8,$06,$A3,$96,$7F,$87,$8F,$AB,$18
	.dc.b	$DF,$FC,$41,$1F,$16,$29,$FA,$CD,$AB,$F5,$F5,$7C,$77,$E0,$2F,$F2
	.dc.b	$47,$E4,$DE,$B3,$F6,$0F,$F8,$02,$7D,$78,$7B,$5F,$90,$C3,$F1,$B1
	.dc.b	$8E,$5F,$CC,$D7,$F3,$78,$C3,$FE,$C6,$3F,$39,$8F,$2E,$6F,$85,$9F
	.dc.b	$F5,$3E,$8F,$D4,$F4,$7E,$D7,$B3,$F8,$DC,$62,$04,$AF,$F1,$34,$66
	.dc.b	$31,$D7,$FE,$2D,$C7,$F7,$9E,$1F,$F0,$09,$E2,$2F,$FC,$5D,$ED,$71
	.dc.b	$8C,$5F,$F8,$9E,$FF,$33,$C9,$FE,$48,$C3,$FC,$61,$E3,$7F,$36,$BD
	.dc.b	$3F,$CE,$3D,$7B,$A7,$F3,$7F,$13,$FC,$5D,$8F,$2E,$5B,$E2,$FF,$28
	.dc.b	$C6,$DF,$F2,$5C,$2B,$FC,$80,$78,$BF,$E2,$EE,$DF,$37,$07,$F2,$5D
	.dc.b	$7F,$9C,$61,$FC,$3F,$C3,$EE,$FF,$16,$8F,$E7,$FA,$7F,$9F,$BC,$6F
	.dc.b	$F4,$9C,$62,$3F,$D0,$9E,$3F,$B7,$1F,$E5,$39,$FF,$2F,$C3,$D3,$C1
	.dc.b	$ED,$C3,$FC,$83,$E3,$F8,$F1,$FE,$C1,$5F,$30,$FA,$BE,$47,$97,$1F
	.dc.b	$EB,$DA,$7C,$2F,$27,$FB,$2F,$2F,$F5,$6F,$03,$FB,$56,$6F,$2B,$F1
	.dc.b	$39,$FF,$77,$EC,$FE,$B7,$E5,$FF,$74,$ED,$FF,$14,$CF,$F7,$93,$FC
	.dc.b	$00,$FF,$60,$F3,$3F,$BF,$F9,$5F,$E1,$DF,$37,$F9,$84,$FF,$1A,$C9
	.dc.b	$1F,$E0,$CB,$ED,$1F,$E1,$3F,$0B,$FD,$4C,$74,$5E,$77,$F8,$F3,$3F
	.dc.b	$D3,$9F,$E2,$1A,$FD,$F7,$93,$FE,$10,$E2,$FE,$FF,$C9,$FD,$8F,$8B
	.dc.b	$E2,$3F,$17,$CD,$FE,$0F,$DD,$F7,$7A,$7F,$AE,$F2,$FF,$66,$70,$7F
	.dc.b	$BB,$B2,$74,$E2,$13,$98,$C4,$7F,$53,$9E,$F8,$7E,$7B,$BB,$0E,$0F
	.dc.b	$D4,$F9,$BF,$D2,$E7,$D7,$B4,$8C,$3F,$D5,$7C,$9F,$F6,$2F,$27,$F5
	.dc.b	$6F,$A5,$F8,$53,$F0,$BC,$EF,$E1,$4B,$E0,$BF,$C7,$38,$E2,$3F,$AF
	.dc.b	$3F,$E8,$57,$E4,$E0,$8F,$F4,$D4,$78,$8F,$EC,$9E,$B3,$9F,$31,$98
	.dc.b	$FE,$5B,$3B,$8F,$34,$F7,$7F,$8E,$F6,$FF,$94,$3C,$C6,$7E,$89,$FE
	.dc.b	$3F,$1E,$E7,$FD,$EB,$7F,$EE,$5D,$3F,$8B,$BD,$63,$F8,$A3,$E7,$DB
	.dc.b	$E5,$1F,$C5,$1C,$3F,$43,$7F,$E0,$3C,$1F,$6B,$A7,$F8,$3B,$F7,$38
	.dc.b	$FF,$7A,$FE,$06,$3F,$06,$E5,$FE,$97,$CD,$D9,$B7,$D7,$9B,$F2,$D8
	.dc.b	$FF,$A1,$B2,$7B,$FF,$E6,$48,$FF,$3D,$47,$D3,$D3,$FC,$D5,$1F,$DD
	.dc.b	$63,$FC,$21,$3F,$E1,$1A,$FF,$33,$4F,$D3,$9F,$E8,$7E,$7F,$F4,$4E
	.dc.b	$4F,$F3,$5F,$C2,$FF,$86,$7D,$9F,$E8,$F8,$F1,$F0,$7F,$C3,$98,$60
	.dc.b	$65,$FB,$BC,$BA,$47,$FA,$9C,$FF,$10,$63,$FA,$78,$CC,$7E,$B2,$3F
	.dc.b	$CA,$DD,$9F,$E6,$8F,$C4,$7F,$0D,$7F,$0B,$6F,$E1,$D1,$C7,$07,$07
	.dc.b	$E8,$3F,$1A,$F6,$61,$FC,$6B,$F2,$2D,$77,$7E,$B3,$F1,$1C,$E1,$F6
	.dc.b	$32,$FE,$31,$F9,$2F,$37,$F5,$3F,$2B,$A9,$FF,$2D,$F3,$FE,$DF,$E0
	.dc.b	$3C,$1E,$1F,$ED,$9F,$2E,$9F,$9B,$AD,$D7,$F3,$9E,$6F,$9D,$ED,$F8
	.dc.b	$FA,$7F,$71,$3F,$C0,$75,$FA,$5E,$AF,$E5,$79,$7F,$CE,$CF,$F5,$3D
	.dc.b	$FF,$D3,$6F,$F5,$5D,$BE,$C7,$AD,$9C,$7F,$7B,$C1,$FC,$3A,$7F,$C8
	.dc.b	$9E,$C7,$E7,$FE,$72,$EF,$F4,$BB,$7E,$76,$1F,$E5,$0A,$FF,$6D,$C7
	.dc.b	$D1,$F5,$BF,$7D,$5F,$53,$AF,$FC,$0B,$87,$F9,$5F,$C8,$FD,$FB,$7C
	.dc.b	$5F,$D6,$7E,$01,$FE,$B8,$9F,$E0,$6B,$0F,$DF,$4F,$B5,$B8,$75,$C3
	.dc.b	$F0,$79,$E2,$3F,$79,$C3,$F2,$EF,$D5,$5F,$C1,$F9,$BD,$9E,$4D,$BE
	.dc.b	$52,$D3,$F8,$7D,$7C,$1F,$27,$55,$FC,$1E,$7F,$C8,$7C,$FF,$3F,$E2
	.dc.b	$BF,$89,$3F,$47,$E0,$2D,$74,$A5,$E2,$39,$70,$FF,$AB,$7E,$2B,$F8
	.dc.b	$C6,$9F,$E7,$16,$F9,$9F,$C6,$35,$FE,$8F,$8F,$28,$C4,$7C,$CB,$FE
	.dc.b	$3D,$9F,$8E,$73,$3F,$A6,$CF,$F8,$FB,$FC,$96,$F9,$6B,$F7,$70,$CC
	.dc.b	$FF,$3A,$96,$58,$8F,$09,$FC,$AF,$6F,$97,$8F,$F3,$8D,$FF,$99,$37
	.dc.b	$87,$5B,$7C,$C9,$FF,$79,$FB,$DF,$73,$8B,$F9,$6C,$FD,$0E,$3F,$BD
	.dc.b	$3F,$84,$CB,$FC,$8B,$E6,$FD,$8F,$37,$F9,$76,$FF,$90,$6C,$FF,$0D
	.dc.b	$CF,$F9,$7B,$ED,$3F,$96,$63,$FD,$5F,$A3,$F9,$8B,$FD,$67,$B1,$A7
	.dc.b	$F8,$F6,$35,$57,$AB,$FB,$0D,$6B,$CB,$76,$FE,$45,$FE,$67,$DF,$FE
	.dc.b	$2D,$EF,$FF,$59,$46,$23,$FD,$9F,$E2,$71,$E8,$FD,$E7,$87,$FB,$FD
	.dc.b	$3F,$9C,$D7,$FB,$57,$C2,$E3,$5A,$8F,$E0,$DB,$7F,$8C,$B5,$FD,$A6
	.dc.b	$D5,$FE,$1F,$1F,$E2,$C9,$FE,$05,$35,$FF,$20,$65,$E1,$C7,$FC,$5B
	.dc.b	$7F,$E2,$DF,$17,$BC,$62,$3F,$CB,$7E,$3F,$F6,$BD,$7F,$93,$46,$32
	.dc.b	$FF,$28,$E5,$FD,$B7,$A3,$FC,$B5,$EA,$7F,$00,$8F,$E7,$D3,$FE,$2E
	.dc.b	$C7,$FD,$7B,$1F,$CE,$F0,$6D,$FD,$06,$3F,$A4,$9F,$E2,$F9,$FE,$73
	.dc.b	$CD,$CA,$BF,$C5,$B1,$FD,$3E,$7F,$AA,$88,$CC,$7F,$80,$A3,$F8,$0C
	.dc.b	$7F,$5C,$8F,$F0,$27,$99,$DA,$3F,$C6,$31,$F8,$FC,$3F,$A6,$FC,$FF
	.dc.b	$F8,$F3,$B3,$F9,$84,$7F,$50,$FC,$03,$FD,$DD,$1F,$E0,$9A,$FC,$EF
	.dc.b	$98,$F8,$EF,$E6,$D9,$FF,$A8,$27,$FC,$76,$7F,$58,$8F,$E7,$55,$FE
	.dc.b	$15,$C3,$F4,$B3,$FE,$1C,$BF,$F5,$2D,$FF,$89,$63,$FD,$53,$E0,$FF
	.dc.b	$56,$F8,$2F,$F7,$46,$1F,$D0,$FB,$7F,$B1,$F8,$DF,$D1,$7C,$EF,$F1
	.dc.b	$A7,$5F,$F9,$03,$34,$E7,$9B,$FD,$3B,$E9,$7F,$4A,$EB,$E3,$AF,$E0
	.dc.b	$A3,$FA,$6F,$B3,$88,$FF,$22,$C8,$FF,$22,$FE,$EF,$FE,$41,$C7,$AC
	.dc.b	$FF,$8B,$FD,$1F,$F2,$7C,$FF,$A8,$BE,$07,$94,$63,$2F,$F5,$2D,$F0
	.dc.b	$BF,$B0,$7F,$5D,$E7,$F6,$46,$5F,$EC,$11,$FE,$C4,$E3,$FF,$24,$4F
	.dc.b	$1A,$FF,$29,$65,$EF,$D7,$FD,$7B,$AF,$F9,$86,$B5,$5F,$02,$BD,$43
	.dc.b	$5F,$76,$DF,$EA,$3D,$FD,$98,$FE,$BE,$BE,$04,$7A,$F0,$F4,$56,$F0
	.dc.b	$F5,$DF,$D4,$CF,$FC,$A1,$C1,$FE,$16,$FB,$4F,$EF,$1B,$FC,$1B,$6B
	.dc.b	$9B,$F9,$1A,$7C,$DF,$59,$C5,$DF,$1E,$33,$98,$FF,$3F,$7D,$AF,$BA
	.dc.b	$3E,$F3,$5F,$6E,$C8,$C4,$7B,$E1,$F8,$66,$30,$FF,$15,$BA,$9B,$8F
	.dc.b	$F8,$67,$2F,$F6,$37,$0B,$D9,$7E,$53,$F9,$AA,$FF,$23,$72,$0D,$79
	.dc.b	$3F,$A2,$7F,$CC,$27,$F9,$FF,$AF,$33,$FE,$F7,$8F,$F4,$17,$1E,$23
	.dc.b	$F3,$E4,$C1,$F8,$F7,$78,$FD,$CD,$77,$1E,$B3,$F4,$7D,$4B,$C2,$FF
	.dc.b	$45,$46,$1F,$37,$CA,$7F,$82,$E3,$17,$FC,$2B,$EA,$9F,$B0,$E2,$3F
	.dc.b	$D1,$B3,$F7,$70,$FF,$48,$B7,$4F,$F4,$73,$98,$A7,$EF,$0F,$3C,$FF
	.dc.b	$A5,$BC,$1E,$8F,$FA,$55,$ED,$E1,$59,$E0,$7A,$67,$E6,$78,$DF,$E9
	.dc.b	$BE,$EF,$17,$AF,$6B,$7B,$7C,$89,$F4,$57,$EA,$27,$F5,$39,$CF,$13
	.dc.b	$11,$F9,$63,$F8,$53,$FB,$33,$F5,$F7,$FC,$41,$C4,$7D,$EA,$FF,$54
	.dc.b	$5F,$8A,$59,$FA,$9B,$5E,$4C,$BD,$31,$88,$FE,$50,$7C,$48,$FF,$59
	.dc.b	$69,$F1,$27,$ED,$4F,$C2,$7F,$DC,$F1,$F4,$78,$F7,$39,$3F,$5F,$FA
	.dc.b	$07,$D8,$7C,$17,$F9,$3F,$57,$9C,$78,$0F,$99,$F0,$19,$C3,$F0,$3C
	.dc.b	$7F,$E1,$27,$EA,$FA,$BF,$90,$AE,$78,$21,$D1,$F9,$DE,$3F,$F4,$89
	.dc.b	$FE,$CF,$8C,$D7,$E7,$6B,$FC,$4F,$A7,$1B,$E1,$3F,$CF,$38,$BA,$74
	.dc.b	$FE,$13,$C7,$FD,$AC,$FE,$C7,$B7,$F8,$B9,$FD,$77,$BB,$F7,$7D,$B6
	.dc.b	$75,$3F,$C2,$6B,$F8,$04,$62,$3F,$9C,$46,$21,$E2,$7F,$28,$9F,$E7
	.dc.b	$95,$FD,$A7,$C2,$FE,$09,$6F,$13,$F8,$8D,$7F,$40,$BE,$27,$F6,$F8
	.dc.b	$CF,$85,$FC,$62,$3F,$A6,$D7,$1C,$7F,$8E,$DF,$19,$79,$1F,$DD,$FA
	.dc.b	$BF,$55,$7F,$CB,$2B,$F9,$AF,$91,$DA,$F5,$87,$FB,$EF,$C8,$FE,$61
	.dc.b	$1F,$CA,$AF,$85,$FF,$36,$8F,$6E,$3C,$2F,$97,$8B,$FE,$E3,$CB,$E4
	.dc.b	$E1,$FD,$12,$FF,$99,$E1,$FD,$13,$3E,$13,$FD,$43,$97,$FB,$8B,$FD
	.dc.b	$4F,$1F,$EA,$C7,$F3,$E7,$FA,$D7,$91,$FC,$F7,$0F,$03,$FC,$FF,$2B
	.dc.b	$8F,$F3,$99,$F2,$D7,$F6,$1D,$7F,$BB,$CF,$F5,$48,$C4,$FF,$76,$D7
	.dc.b	$FA,$BF,$07,$AE,$F7,$6D,$3F,$C4,$1D,$DF,$2B,$E3,$BF,$1B,$D1,$FE
	.dc.b	$09,$CF,$FC,$29,$B7,$F8,$3B,$A2,$3F,$E2,$7F,$2B,$F9,$B5,$7F,$B9
	.dc.b	$A3,$FC,$55,$E5,$FF,$87,$BC,$1F,$F0,$BE,$0F,$33,$FC,$49,$5F,$E1
	.dc.b	$98,$FE,$88,$FF,$1F,$C7,$FB,$B9,$FE,$09,$DB,$ED,$70,$FF,$6C,$9F
	.dc.b	$F2,$63,$F7,$7C,$BF,$93,$C1,$F9,$E3,$FB,$13,$87,$07,$F8,$4E,$39
	.dc.b	$9F,$E1,$A8,$FB,$4F,$67,$F0,$71,$FA,$8E,$0F,$EF,$AF,$23,$FB,$85
	.dc.b	$7F,$87,$1F,$04,$63,$93,$51,$FD,$8A,$7B,$DF,$7E,$AF,$F7,$93,$83
	.dc.b	$FA,$C6,$FF,$D4,$C8,$60,$FF,$53,$CF,$C4,$F2,$DF,$B2,$7D,$09,$DF
	.dc.b	$97,$FC,$95,$E2,$7F,$5D,$70,$F4,$F4,$3F,$D2,$BE,$63,$F3,$2E,$53
	.dc.b	$FE,$4E,$F1,$7F,$B0,$E5,$FE,$58,$3F,$90,$F2,$FF,$40,$47,$F4,$F3
	.dc.b	$BC,$FF,$95,$FB,$B9,$BF,$E5,$16,$BF,$ED,$68,$FD,$E1,$C1,$FF,$68
	.dc.b	$78,$1A,$F7,$BF,$A2,$CF,$FB,$A2,$7F,$DA,$95,$CB,$C6,$61,$EC,$9F
	.dc.b	$7B,$E6,$AF,$33,$5F,$EB,$5C,$EF,$2B,$E0,$47,$5A,$79,$5C,$DF,$F4
	.dc.b	$7B,$5F,$A9,$E1,$FD,$09,$E0,$FF,$36,$7F,$C1,$E7,$F9,$F9,$C4,$7B
	.dc.b	$7A,$FF,$99,$CF,$DF,$E5,$FE,$60,$FE,$07,$C3,$FE,$44,$FF,$1E,$E8
	.dc.b	$FC,$D5,$7C,$6E,$9F,$C8,$75,$7F,$13,$F0,$FF,$1B,$3F,$96,$AF,$E2
	.dc.b	$0F,$E6,$B1,$FE,$40,$D7,$F9,$5F,$0F,$F3,$29,$FE,$35,$8F,$77,$8C
	.dc.b	$7F,$9F,$67,$F4,$B5,$FC,$29,$FD,$8E,$4E,$C7,$D7,$FF,$00,$F3,$3F
	.dc.b	$C0,$B0,$F0,$7F,$C5,$DE,$6F,$F8,$47,$83,$EB,$CF,$F0,$AD,$7F,$D2
	.dc.b	$71,$F8,$D8,$C7,$2F,$FA,$82,$BF,$D4,$11,$FE,$9A,$96,$1F,$E4,$5B
	.dc.b	$FF,$08,$4F,$FA,$06,$7F,$D4,$B1,$FE,$68,$8F,$47,$D5,$FF,$A1,$23
	.dc.b	$FD,$E7,$18,$E3,$FD,$73,$FC,$17,$C7,$FF,$05,$F3,$FF,$A3,$B8,$BF
	.dc.b	$AD,$E7,$AE,$2F,$DB,$F0,$FF,$58,$CB,$FD,$77,$8F,$F4,$1C,$FF,$6E
	.dc.b	$FF,$4A,$CB,$C1,$97,$F0,$8C,$BF,$D0,$D9,$F8,$3A,$BF,$71,$8F,$FA
	.dc.b	$59,$FE,$CB,$7E,$2C,$3F,$01,$7E,$2D,$7F,$CA,$D1,$FD,$C7,$3F,$EB
	.dc.b	$98,$78,$BC,$9F,$2C,$34,$F1,$F8,$78,$8F,$E5,$77,$F9,$27,$FB,$8C
	.dc.b	$FF,$0E,$F0,$FF,$2F,$8F,$F9,$C7,$B3,$1E,$67,$FB,$C7,$3F,$F4,$9F
	.dc.b	$07,$F1,$3E,$3F,$A5,$0F,$33,$F8,$1E,$3F,$A5,$8F,$F7,$46,$3F,$C4
	.dc.b	$7C,$AF,$E8,$B9,$7F,$16,$6C,$7F,$9F,$FC,$46,$5F,$E7,$29,$CC,$A7
	.dc.b	$B7,$89,$FB,$8C,$3A,$CF,$F9,$8D,$FE,$26,$FE,$09,$F2,$1F,$BB,$3F
	.dc.b	$8D,$57,$D0,$7F,$D8,$6F,$68,$FC,$67,$35,$C7,$FC,$C9,$5D,$A6,$9F
	.dc.b	$98,$3F,$D2,$DE,$16,$1F,$90,$7F,$A2,$F5,$FE,$4B,$CD,$DE,$9F,$03
	.dc.b	$87,$CF,$5F,$EF,$D7,$F9,$5F,$A1,$97,$DB,$85,$9F,$01,$FA,$28,$FE
	.dc.b	$65,$B7,$F4,$7C,$BF,$CA,$91,$FC,$D0,$F9,$73,$FE,$71,$C7,$DD,$7F
	.dc.b	$CA,$1F,$81,$95,$AF,$1D,$7D,$C7,$CF,$D7,$FC,$DA,$3E,$66,$5F,$CD
	.dc.b	$27,$CB,$3F,$B1,$9F,$82,$7F,$9D,$B4,$F9,$51,$98,$65,$FD,$72,$3F
	.dc.b	$9C,$37,$5F,$8D,$7F,$CF,$1F,$91,$C3,$63,$CB,$7F,$E0,$4F,$13,$F0
	.dc.b	$4F,$C4,$3F,$9C,$D7,$E2,$38,$7F,$17,$E2,$76,$9F,$C7,$61,$AC,$3D
	.dc.b	$F8,$FF,$6E,$EC,$FC,$C5,$73,$7F,$47,$B7,$E7,$AF,$FC,$E9,$CB,$FA
	.dc.b	$3C,$3F,$41,$C5,$FE,$73,$79,$E1,$FD,$E7,$C1,$FD,$77,$47,$F9,$9B
	.dc.b	$A3,$FB,$DC,$62,$3F,$CD,$35,$F5,$EF,$FC,$D0,$FF,$6C,$8F,$DA,$DF
	.dc.b	$3E,$BF,$DD,$4E,$EF,$F8,$2F,$63,$B7,$F9,$75,$FF,$4A,$AF,$F0,$A4
	.dc.b	$7F,$1C,$AF,$B5,$3F,$8F,$9F,$DC,$47,$F1,$7D,$BD,$1C,$BF,$89,$F0
	.dc.b	$BF,$C8,$F3,$DE,$3F,$86,$EB,$EA,$7C,$77,$FE,$39,$7F,$8C,$F7,$F2
	.dc.b	$8F,$F3,$BE,$5E,$19,$FE,$99,$5F,$EA,$1C,$BF,$DE,$18,$7F,$36,$3B
	.dc.b	$5F,$DE,$3F,$35,$C0,$8F,$E8,$5C,$3F,$4E,$7F,$CA,$07,$FB,$96,$BF
	.dc.b	$A7,$38,$3F,$A6,$9F,$E1,$6E,$0F,$F0,$BD,$7F,$55,$8F,$F7,$66,$7F
	.dc.b	$D5,$79,$FB,$78,$3F,$D6,$E7,$C3,$5F,$E5,$77,$F0,$E7,$F4,$0F,$2F
	.dc.b	$F5,$D5,$FA,$FF,$D2,$7F,$AD,$79,$BF,$E6,$2C,$99,$FF,$2B,$C3,$FB
	.dc.b	$44,$F0,$7F,$8F,$CF,$B9,$FE,$D9,$F0,$1F,$EC,$38,$E3,$5F,$E2,$7F
	.dc.b	$0F,$FC,$4F,$E6,$FF,$AA,$3C,$FF,$E8,$DE,$0F,$F7,$77,$F0,$33,$F8
	.dc.b	$4F,$B2,$FE,$7F,$3F,$EA,$CB,$FE,$A1,$97,$C2,$3E,$26,$BF,$D4,$5C
	.dc.b	$1F,$CC,$B8,$7C,$3D,$BF,$E5,$EE,$CF,$04,$FF,$60,$C3,$FA,$7F,$93
	.dc.b	$FD,$81,$FC,$47,$60,$FF,$46,$63,$BA,$FF,$58,$46,$7C,$AF,$3D,$7C
	.dc.b	$39,$FE,$82,$7E,$03,$5F,$F6,$77,$97,$CE,$DB,$7F,$5A,$EF,$F6,$6F
	.dc.b	$FD,$91,$E5,$C5,$FE,$12,$F2,$BF,$B1,$F1,$7F,$60,$7C,$F5,$FE,$00
	.dc.b	$FC,$6F,$3A,$7F,$7A,$73,$1F,$43,$CB,$FF,$1E,$70,$7F,$2B,$AF,$EC
	.dc.b	$DF,$01,$F5,$6B,$FA,$0E,$4B,$4E,$65,$E5,$7F,$65,$9C,$DF,$FB,$FA
	.dc.b	$FE,$EC,$FF,$68,$F4,$7E,$14,$FF,$65,$3F,$CD,$B8,$F8,$70,$FF,$42
	.dc.b	$7A,$2F,$2B,$71,$F8,$EF,$53,$FB,$4C,$E6,$F8,$9F,$D0,$E7,$F2,$7E
	.dc.b	$AF,$E5,$3D,$0F,$ED,$58,$F5,$9F,$C7,$D7,$13,$FC,$46,$7F,$68,$3F
	.dc.b	$D6,$87,$E5,$F1,$F9,$EF,$C5,$38,$4F,$39,$FE,$D7,$1E,$23,$F7,$9E
	.dc.b	$3F,$A2,$3F,$D0,$91,$93,$FD,$41,$E2,$FF,$61,$AF,$F4,$84,$7F,$64
	.dc.b	$3F,$D0,$EF,$5E,$DD,$47,$F9,$46,$32,$FF,$83,$9A,$FF,$5C,$9F,$EB
	.dc.b	$CF,$B0,$FE,$DF,$0B,$F2,$5F,$F5,$68,$C4,$F9,$63,$FB,$E4,$62,$3F
	.dc.b	$D2,$27,$FA,$C5,$F8,$2D,$C3,$E2,$BF,$C1,$9A,$FF,$87,$A3,$6F,$E8
	.dc.b	$0F,$F5,$5F,$8D,$FE,$95,$F4,$FC,$77,$FD,$5E,$7B,$9F,$E0,$8D,$3F
	.dc.b	$86,$F9,$BC,$A3,$F8,$3E,$FF,$D8,$32,$E3,$E6,$7F,$8C,$F3,$FA,$98
	.dc.b	$A3,$F2,$77,$FB,$EE,$3F,$C9,$B9,$E5,$FC,$D7,$27,$F8,$D2,$B7,$7F
	.dc.b	$E4,$5B,$FC,$CE,$DF,$D8,$70,$F2,$9F,$D7,$B9,$BF,$57,$5F,$C8,$BC
	.dc.b	$84,$7E,$C2,$7F,$E0,$07,$46,$AB,$FD,$47,$E1,$7F,$0E,$9F,$E7,$16
	.dc.b	$EB,$F9,$38,$FF,$66,$9F,$F7,$87,$83,$FD,$B2,$7F,$C4,$1D,$FF,$C1
	.dc.b	$FB,$BF,$71,$AF,$F6,$3E,$1F,$C5,$E1,$FD,$83,$C1,$FE,$2F,$E0,$32
	.dc.b	$FE,$D3,$A7,$FB,$53,$CC,$FE,$21,$1F,$DA,$6B,$CD,$3F,$ED,$8D,$BC
	.dc.b	$D8,$7F,$96,$23,$FD,$B6,$7F,$7B,$F1,$3F,$94,$B8,$8F,$E5,$E7,$DE
	.dc.b	$7F,$BC,$DF,$F8,$12,$FD,$38,$FF,$29,$D3,$D3,$5F,$D0,$31,$0F,$0B
	.dc.b	$CF,$7F,$D0,$74,$FF,$02,$78,$DF,$63,$3F,$ED,$1E,$3F,$F4,$FD,$FF
	.dc.b	$1D,$C5,$FD,$8A,$3F,$AB,$71,$FF,$55,$D3,$FD,$63,$7F,$EF,$CF,$23
	.dc.b	$FE,$00,$47,$FA,$FA,$3F,$75,$7F,$EB,$EC,$FF,$D8,$DA,$F9,$0F,$2F
	.dc.b	$8F,$FE,$C5,$8C,$C6,$4C,$57,$F4,$2E,$0F,$E7,$87,$F4,$3F,$1F,$E8
	.dc.b	$38,$3F,$A0,$3F,$D0,$9C,$1F,$5D,$C3,$87,$F1,$B1,$FE,$29,$9F,$01
	.dc.b	$E4,$8C,$C7,$FB,$C7,$C7,$FC,$DE,$9F,$DB,$88,$E3,$C4,$FE,$39,$BF
	.dc.b	$F3,$DF,$CD,$7F,$95,$56,$CF,$EE,$BE,$3F,$F9,$75,$C3,$5F,$F4,$E4
	.dc.b	$BC,$77,$E6,$9C,$A7,$91,$FE,$89,$2E,$5F,$E3,$67,$B6,$28,$FD,$11
	.dc.b	$FE,$62,$8F,$E9,$3C,$B8,$8E,$71,$FE,$94,$8E,$4F,$FA,$15,$F2,$3F
	.dc.b	$E0,$CF,$47,$C7,$3C,$E3,$FC,$1F,$E2,$EB,$D1,$FF,$34,$F9,$3E,$79
	.dc.b	$FF,$33,$6D,$5F,$91,$E1,$FC,$7C,$3F,$5E,$FC,$C8,$F6,$F3,$FD,$0E
	.dc.b	$3F,$A6,$7E,$1D,$C3,$8E,$6F,$DE,$78,$1F,$7B,$9B,$FC,$15,$5F,$79
	.dc.b	$FC,$3F,$77,$F9,$77,$C0,$D4,$53,$F0,$B1,$FE,$61,$F3,$3B,$E3,$FA
	.dc.b	$8B,$FD,$53,$F8,$4C,$5D,$3F,$A0,$D5,$F7,$9F,$83,$7F,$C6,$E7,$FA
	.dc.b	$DB,$97,$FD,$32,$FF,$36,$32,$7F,$34,$7F,$C4,$EF,$F8,$E8,$FF,$5E
	.dc.b	$F9,$9C,$1F,$F6,$09,$FA,$21,$97,$EA,$7E,$CB,$F8,$DE,$1E,$65,$BF
	.dc.b	$E9,$A3,$0F,$F8,$8F,$D2,$E4,$FF,$89,$7E,$77,$F5,$DA,$63,$83,$FC
	.dc.b	$65,$1F,$B1,$F1,$31,$08,$FD,$2E,$1F,$C6,$F7,$FF,$4B,$4F,$F2,$08
	.dc.b	$C4,$3D,$3F,$DA,$46,$1F,$DC,$E5,$EB,$E3,$F2,$7B,$BF,$C0,$F3,$C6
	.dc.b	$3F,$E4,$3A,$4F,$F9,$67,$0F,$DF,$CF,$FA,$66,$3F,$83,$47,$E6,$F9
	.dc.b	$7E,$5F,$07,$F6,$48,$FD,$FE,$9E,$E6,$E9,$CF,$C9,$FE,$53,$8F,$B6
	.dc.b	$39,$F9,$5F,$AF,$E4,$FF,$22,$63,$FC,$16,$7D,$95,$EF,$AF,$F4,$EB
	.dc.b	$F3,$DA,$FA,$70,$FC,$BC,$F0,$8F,$E3,$1E,$5E,$9F,$F3,$04,$EB,$83
	.dc.b	$E2,$F8,$9F,$A3,$E6,$FF,$24,$E5,$F8,$4C,$FF,$DE,$DE,$BB,$C4,$FF
	.dc.b	$25,$79,$DF,$C3,$AF,$F1,$13,$FC,$07,$07,$77,$F0,$99,$FE,$1F,$DF
	.dc.b	$78,$FF,$C8,$AD,$5E,$07,$F9,$12,$30,$BC,$8F,$E2,$9E,$2E,$EB,$FC
	.dc.b	$6F,$B7,$F2,$3D,$BF,$CC,$15,$FC,$55,$FE,$D7,$1F,$EF,$B7,$FC,$80
	.dc.b	$FF,$BF,$71,$FE,$48,$FF,$19,$79,$36,$BF,$59,$E4,$77,$F3,$BF,$8E
	.dc.b	$3E,$1A,$FE,$3D,$CF,$F5,$EF,$E2,$9F,$83,$BF,$F5,$53,$FE,$43,$8E
	.dc.b	$4E,$61,$3F,$E4,$AF,$1F,$0F,$F3,$DC,$11,$89,$FF,$1D,$DF,$7B,$FE
	.dc.b	$5D,$C1,$A8,$C4,$3D,$7F,$F2,$4D,$7F,$AD,$31,$E9,$87,$B6,$3F,$D7
	.dc.b	$F1,$A6,$9F,$EA,$A8,$D3,$FE,$C4,$7E,$43,$FD,$45,$F9,$95,$FE,$8F
	.dc.b	$6B,$FE,$44,$6B,$F8,$7F,$23,$F9,$CB,$FE,$68,$9F,$EB,$5E,$47,$47
	.dc.b	$F0,$3C,$A8,$C7,$0F,$E0,$5D,$4F,$FB,$5B,$C8,$FC,$6D,$FF,$99,$E3
	.dc.b	$F9,$EF,$37,$E4,$F8,$7E,$9D,$7F,$68,$D3,$F2,$B9,$7F,$91,$F0,$E1
	.dc.b	$87,$FB,$F3,$AB,$F4,$9D,$5F,$EF,$97,$F4,$DD,$F6,$BF,$65,$D5,$FC
	.dc.b	$F3,$C7,$77,$70,$6C,$EB,$0F,$E4,$11,$88,$FF,$06,$57,$1A,$FA,$FD
	.dc.b	$1F,$D2,$AF,$F9,$0F,$77,$F8,$47,$0F,$E0,$EF,$F8,$59,$AF,$F8,$AD
	.dc.b	$FC,$56,$FF,$A0,$7F,$8A,$E5,$FC,$62,$7F,$AC,$74,$FF,$92,$23,$10
	.dc.b	$F0,$7F,$A9,$61,$FC,$6E,$3F,$C6,$FE,$1B,$0F,$E5,$13,$FE,$20,$6B
	.dc.b	$FC,$95,$AF,$F9,$4D,$61,$FC,$73,$C1,$FF,$0D,$C6,$36,$FF,$02,$78
	.dc.b	$9F,$D0,$71,$FE,$61,$39,$9F,$E9,$38,$7C,$29,$F1,$CF,$F4,$A8,$FF
	.dc.b	$76,$35,$67,$FD,$A6,$7F,$C9,$FE,$2F,$F9,$BA,$9E,$6F,$E0,$F0,$FD
	.dc.b	$39,$FE,$4E,$3F,$D7,$DB,$7F,$62,$97,$9B,$FE,$28,$8F,$EC,$5C,$1F
	.dc.b	$E1,$B9,$FE,$E1,$E4,$FF,$6F,$5E,$67,$F8,$97,$C2,$FF,$78,$78,$5F
	.dc.b	$DD,$BF,$70,$FF,$11,$67,$FE,$A2,$F0,$7F,$8A,$F9,$0F,$C0,$F3,$E6
	.dc.b	$7F,$94,$BC,$CF,$F0,$3C,$E6,$FF,$C5,$53,$9E,$5F,$ED,$97,$C3,$93
	.dc.b	$ED,$61,$FD,$73,$0E,$18,$7C,$AF,$37,$EF,$D7,$DF,$9C,$F3,$B1,$FF
	.dc.b	$45,$79,$BF,$DB,$E7,$31,$8A,$FE,$BB,$97,$AB,$86,$65,$EC,$E3,$FF
	.dc.b	$09,$CE,$23,$FD,$6D,$E0,$09,$CB,$A8,$FF,$2F,$76,$FF,$99,$F6,$C4
	.dc.b	$7F,$71,$CF,$FB,$BE,$1F,$23,$3F,$F3,$AC,$7B,$63,$FD,$43,$C1,$FD
	.dc.b	$AD,$B3,$FD,$D5,$AF,$F9,$F7,$5F,$87,$C3,$F4,$8F,$A3,$F8,$27,$FA
	.dc.b	$63,$2F,$8B,$9F,$F8,$B2,$31,$C1,$F0,$DB,$87,$F7,$E7,$96,$9F,$F0
	.dc.b	$07,$7F,$F1,$9F,$E0,$BE,$7A,$FD,$66,$DF,$6A,$31,$8F,$F9,$FD,$B1
	.dc.b	$F5,$FF,$59,$AF,$F8,$65,$AF,$FA,$29,$AF,$E7,$BD,$07,$EC,$7E,$2C
	.dc.b	$9E,$AF,$F9,$F6,$5F,$92,$FE,$67,$97,$E8,$D7,$F7,$79,$CB,$FE,$45
	.dc.b	$8C,$EA,$97,$E4,$9D,$70,$FF,$42,$FE,$67,$6F,$FD,$21,$DC,$8C,$43
	.dc.b	$DA,$FF,$02,$4E,$72,$FF,$41,$72,$FE,$8B,$0F,$C2,$5B,$DB,$FC,$37
	.dc.b	$B0,$FC,$BF,$9C,$FE,$27,$0E,$47,$EA,$4F,$D5,$F9,$DF,$E8,$DD,$3E
	.dc.b	$16,$5F,$E0,$09,$DB,$F1,$5F,$D8,$C7,$4A,$FF,$4A,$78,$36,$4F,$83
	.dc.b	$BB,$31,$E8,$3F,$D3,$5A,$7F,$AB,$FC,$FF,$F4,$E7,$77,$82,$7F,$7B
	.dc.b	$EA,$E1,$E0,$7D,$4C,$7F,$CD,$B3,$EC,$AF,$F0,$8C,$FE,$A4,$FF,$3A
	.dc.b	$18,$D3,$DB,$A6,$A3,$FC,$5C,$7F,$BC,$8F,$F0,$CF,$27,$EC,$63,$F3
	.dc.b	$B1,$E6,$C3,$FB,$C6,$1D,$E7,$33,$FE,$0A,$3C,$EF,$C0,$7C,$F9,$7F
	.dc.b	$AB,$F4,$E0,$7C,$79,$F9,$73,$D6,$FE,$CC,$62,$3A,$1F,$AD,$D7,$E2
	.dc.b	$3F,$EC,$1C,$BF,$C3,$31,$88,$FF,$1D,$7A,$BF,$7D,$F0,$DF,$F8,$0A
	.dc.b	$7F,$17,$C3,$FC,$27,$D4,$ED,$9F,$E2,$39,$3F,$CA,$11,$8B,$FF,$5D
	.dc.b	$F2,$FD,$DF,$4F,$F3,$B9,$7E,$3B,$8F,$FD,$3B,$AA,$73,$7F,$C5,$0F
	.dc.b	$D0,$BF,$AC,$AF,$1C,$7E,$5B,$C4,$76,$7E,$CB,$B3,$FB,$23,$FB,$4F
	.dc.b	$36,$DF,$F1,$C9,$FF,$02,$77,$6E,$CE,$DB,$3B,$9F,$E6,$47,$F6,$9F
	.dc.b	$07,$E3,$73,$FF,$71,$7F,$88,$F8,$3F,$ED,$16,$BF,$DE,$6D,$18,$F4
	.dc.b	$3F,$BC,$74,$F2,$9F,$E8,$7E,$47,$6F,$1F,$FB,$9C,$FF,$26,$C3,$93
	.dc.b	$6F,$8E,$7E,$5C,$FF,$B7,$CF,$A4,$FE,$27,$E3,$6A,$7F,$C0,$C7,$FB
	.dc.b	$FF,$87,$FC,$E5,$9F,$F5,$23,$FA,$24,$7F,$9E,$5F,$E9,$B3,$FD,$53
	.dc.b	$1E,$5E,$47,$F5,$38,$FE,$67,$3A,$F1,$FF,$D8,$4F,$F5,$CD,$3F,$B1
	.dc.b	$67,$E4,$C7,$FB,$2E,$3F,$D8,$A3,$FC,$5F,$9F,$F6,$5F,$1D,$B7,$F8
	.dc.b	$DF,$BF,$E6,$F9,$1C,$F0,$FF,$7D,$35,$FF,$7F,$BF,$D8,$B0,$FE,$AC
	.dc.b	$D7,$FA,$AE,$3F,$EB,$39,$FE,$A4,$D7,$FB,$3A,$3C,$31,$FE,$FE,$9F
	.dc.b	$A6,$FE,$00,$FF,$73,$70,$FF,$B0,$9F,$F0,$0B,$5F,$F7,$57,$2F,$F4
	.dc.b	$DB,$1F,$EC,$B7,$0E,$09,$E3,$7E,$7B,$9F,$F9,$79,$1C,$74,$FF,$85
	.dc.b	$32,$FD,$B7,$8D,$FC,$0D,$DD,$48,$C3,$83,$FC,$EB,$E3,$3F,$2E,$FF
	.dc.b	$0E,$C6,$4F,$E7,$1E,$17,$F3,$1B,$FE,$6E,$7F,$07,$3F,$D6,$3B,$F0
	.dc.b	$9E,$11,$FE,$18,$79,$63,$FC,$AE,$3A,$43,$CB,$F9,$9B,$F1,$8F,$3C
	.dc.b	$7F,$27,$9F,$F6,$31,$4F,$E2,$47,$8C,$F1,$9F,$CD,$8F,$84,$70,$3F
	.dc.b	$AB,$CF,$59,$FF,$53,$F1,$D7,$FA,$CD,$7F,$AD,$23,$11,$FE,$1E,$F1
	.dc.b	$7E,$3B,$F0,$DF,$81,$7F,$EA,$5C,$BC,$4F,$FA,$8F,$3E,$35,$FC,$7F
	.dc.b	$3E,$7E,$B7,$F1,$E7,$31,$FC,$67,$CA,$FF,$2A,$5F,$D9,$F0,$70,$FE
	.dc.b	$3E,$7E,$3F,$4F,$F1,$88,$FC,$A7,$1F,$E1,$63,$11,$F8,$78,$87,$F0
	.dc.b	$BD,$FF,$21,$CB,$FE,$6E,$6B,$FE,$7D,$DF,$FC,$B7,$CF,$F9,$98,$FF
	.dc.b	$42,$46,$1C,$1B,$BF,$0D,$FE,$63,$0F,$F2,$AB,$E0,$3F,$D3,$95,$DA
	.dc.b	$BF,$81,$9C,$BA,$FF,$87,$75,$6E,$B9,$57,$F9,$85,$F4,$F5,$7F,$92
	.dc.b	$B4,$FF,$06,$E5,$FE,$6E,$AF,$F3,$37,$A9,$FC,$5A,$3F,$AA,$75,$FF
	.dc.b	$01,$8F,$F2,$BD,$7F,$85,$23,$FC,$C3,$E1,$7F,$5C,$A7,$07,$F9,$5B
	.dc.b	$BB,$FC,$D1,$3F,$DF,$6F,$86,$3F,$B5,$BF,$A9,$87,$BB,$36,$5F,$A9
	.dc.b	$8F,$E0,$38,$78,$EB,$AC,$63,$BF,$F6,$2F,$F3,$C8,$FB,$98,$FF,$88
	.dc.b	$DF,$F3,$84,$65,$FF,$53,$D7,$9F,$3F,$F2,$D4,$7F,$AB,$B7,$FF,$52
	.dc.b	$F8,$DD,$3C,$3F,$D4,$4F,$FA,$49,$F4,$3F,$E2,$CC,$3F,$AA,$BF,$C2
	.dc.b	$F3,$F4,$78,$1A,$8F,$92,$FF,$97,$E3,$F8,$27,$81,$FC,$33,$4F,$F2
	.dc.b	$97,$3F,$EF,$7C,$9F,$F0,$F6,$BF,$EA,$6E,$9F,$F2,$A5,$7F,$92,$9A
	.dc.b	$FF,$C0,$0F,$07,$F5,$3E,$07,$F1,$8E,$C7,$88,$F0,$BF,$88,$6F,$FE
	.dc.b	$59,$C3,$F8,$EB,$FB,$07,$F7,$73,$FE,$3A,$8C,$9F,$C4,$B1,$F8,$2F
	.dc.b	$F0,$07,$C3,$3F,$E4,$FE,$7B,$8F,$F1,$18,$FF,$03,$1D,$A3,$7E,$16
	.dc.b	$61,$3F,$CA,$35,$FE,$2B,$87,$D8,$9F,$E1,$BE,$6E,$23,$C6,$7F,$45
	.dc.b	$F2,$F1,$3D,$8F,$F5,$46,$18,$7A,$9F,$E5,$19,$FF,$22,$6F,$FE,$1D
	.dc.b	$7F,$90,$47,$F8,$D1,$FF,$24,$35,$FE,$3A,$7E,$69,$FE,$2D,$E0,$FA
	.dc.b	$BC,$4F,$C5,$47,$84,$64,$8C,$C7,$E4,$6B,$FA,$E6,$59,$BF,$97,$1E
	.dc.b	$BA,$FE,$75,$5E,$83,$F9,$3B,$8C,$FE,$33,$63,$FA,$AC,$7F,$2B,$9F
	.dc.b	$B9,$1F,$E4,$ED,$7E,$AB,$E3,$7B,$C7,$FA,$31,$BB,$66,$3E,$81,$FC
	.dc.b	$97,$CB,$FC,$3B,$FE,$43,$AF,$EB,$BE,$4F,$E3,$9F,$F4,$9F,$17,$F2
	.dc.b	$F9,$F0,$BF,$90,$9D,$4F,$F6,$AE,$DF,$ED,$9B,$7E,$71,$FD,$04,$7F
	.dc.b	$6E,$F0,$FF,$96,$F7,$FF,$76,$E9,$FE,$D1,$18,$8F,$EF,$39,$7F,$39
	.dc.b	$8F,$D7,$C6,$EB,$FB,$8C,$62,$37,$3F,$EA,$08,$FD,$AC,$FF,$1E,$8F
	.dc.b	$F7,$94,$7F,$03,$F0,$3F,$8A,$F7,$72,$E2,$FF,$09,$57,$F0,$6E,$2F
	.dc.b	$04,$7C,$8B,$FE,$21,$87,$A7,$9F,$F9,$D7,$81,$FC,$57,$9B,$FC,$7D
	.dc.b	$A7,$96,$B5,$97,$F8,$76,$3F,$C2,$0F,$AE,$FB,$E9,$E2,$AD,$CF,$84
	.dc.b	$F5,$4F,$F3,$4D,$BD,$EF,$F3,$6C,$B9,$4F,$DA,$3F,$8F,$C7,$FB,$C3
	.dc.b	$1F,$DE,$F7,$7F,$21,$F0,$BF,$C4,$7E,$47,$F2,$18,$C6,$88,$FE,$6D
	.dc.b	$98,$FE,$4B,$8B,$47,$84,$F7,$3F,$C0,$79,$EA,$7F,$CA,$97,$FC,$D3
	.dc.b	$83,$E5,$ED,$FE,$B3,$C7,$F8,$AE,$9F,$CC,$E3,$FD,$41,$C5,$F9,$C8
	.dc.b	$FE,$6D,$AF,$9A,$BF,$A2,$DF,$D4,$8F,$F1,$DF,$89,$FD,$16,$3F,$B5
	.dc.b	$78,$BE,$B6,$E1,$E1,$BF,$F6,$A6,$9F,$0B,$C5,$F9,$FB,$FA,$67,$FA
	.dc.b	$8C,$63,$0F,$EA,$0F,$C9,$8F,$F2,$07,$5B,$4F,$CA,$70,$FC,$07,$F2
	.dc.b	$DA,$7F,$57,$FD,$5F,$CD,$B7,$CB,$C3,$FA,$EC,$62,$7F,$A0,$F8,$9F
	.dc.b	$8E,$E8,$FF,$13,$78,$AF,$CB,$7F,$C5,$1E,$1F,$FB,$0E,$31,$8F,$F2
	.dc.b	$5E,$5F,$E3,$EF,$F3,$EF,$1B,$FC,$A5,$87,$F8,$BD,$B1,$FC,$AF,$CD
	.dc.b	$FE,$E7,$18,$8F,$F2,$EF,$9D,$F7,$7C,$8F,$DF,$DF,$F8,$DE,$78,$F3
	.dc.b	$7D,$C6,$9F,$08,$A7,$F4,$33,$FD,$F3,$8F,$F9,$89,$FF,$74,$B7,$B7
	.dc.b	$DD,$1E,$9F,$1B,$FC,$AD,$7F,$E4,$7F,$0B,$F9,$F3,$FE,$50,$9F,$D8
	.dc.b	$F9,$18,$78,$1F,$DE,$63,$83,$FC,$F0,$FF,$44,$73,$FF,$96,$63,$F1
	.dc.b	$D1,$FE,$56,$6B,$FE,$5A,$8F,$E6,$4F,$A3,$1F,$F7,$04,$F5,$8F,$9D
	.dc.b	$7A,$7F,$9E,$1F,$E1,$C3,$FD,$05,$7F,$CC,$5F,$F7,$DD,$7F,$34,$AF
	.dc.b	$82,$FE,$A4,$FE,$8A,$E1,$FE,$6D,$1E,$16,$BD,$B2,$FE,$A3,$3F,$3A
	.dc.b	$3F,$D4,$91,$FC,$DD,$D1,$F9,$4F,$1B,$EC,$BF,$85,$A7,$8B,$F7,$DF
	.dc.b	$16,$DF,$E6,$79,$F0,$70,$7F,$39,$7D,$33,$FC,$DF,$8F,$F2,$31,$FE
	.dc.b	$3A,$F0,$FF,$22,$FF,$A6,$F9,$3E,$7C,$39,$7E,$EC,$FF,$38,$E7,$FC
	.dc.b	$F6,$1F,$CD,$1F,$D1,$F5,$7F,$06,$F0,$3F,$4F,$3F,$CE,$3C,$9C,$DF
	.dc.b	$F3,$DE,$DF,$D8,$3F,$CD,$9F,$DA,$4F,$F3,$E7,$F7,$3E,$1F,$3F,$0B
	.dc.b	$75,$F8,$2C,$3F,$C7,$DE,$57,$FB,$97,$C3,$FE,$79,$3F,$EC,$D8,$FF
	.dc.b	$74,$73,$7F,$3C,$8F,$E6,$9E,$03,$C0,$FE,$A3,$8F,$92,$7F,$A5,$3F
	.dc.b	$C6,$9F,$F2,$EE,$1F,$EF,$88,$C7,$47,$F9,$87,$AD,$9F,$F4,$DF,$0D
	.dc.b	$E1,$F4,$C7,$FC,$9D,$B7,$A5,$FF,$79,$F1,$7F,$30,$E1,$E3,$9F,$F5
	.dc.b	$2F,$17,$F9,$BF,$07,$F3,$EB,$FC,$09,$FE,$5C,$9F,$F3,$E7,$07,$FA
	.dc.b	$57,$83,$F9,$BD,$FF,$A0,$63,$F9,$CF,$0F,$FA,$3B,$5F,$1F,$89,$F1
	.dc.b	$9C,$3F,$E5,$AF,$17,$5B,$FF,$1F,$E4,$FF,$47,$4B,$9B,$FD,$D3,$8F
	.dc.b	$FB,$A7,$3F,$E4,$91,$88,$FF,$75,$F5,$7F,$AC,$BC,$5F,$C4,$F0,$FF
	.dc.b	$4B,$F0,$BF,$A3,$F8,$BF,$E2,$08,$FF,$73,$35,$FF,$06,$F4,$7F,$96
	.dc.b	$7F,$1C,$FF,$5D,$F2,$66,$7F,$CB,$3C,$9F,$EC,$2C,$13,$FD,$FA,$7F
	.dc.b	$2C,$FF,$B7,$79,$7F,$D9,$2E,$9F,$E6,$1C,$1F,$EC,$AE,$EF,$DF,$57
	.dc.b	$FB,$AA,$7B,$3F,$B9,$F0,$99,$7F,$B3,$3C,$2C,$F9,$3F,$E9,$E7,$75
	.dc.b	$FD,$9F,$C1,$F7,$BF,$E5,$EC,$FF,$DA,$11,$88,$FD,$34,$7F,$AF,$B8
	.dc.b	$BF,$0F,$C7,$F2,$5F,$C9,$76,$FD,$3E,$9F,$B0,$D7,$FD,$95,$D1,$F8
	.dc.b	$6B,$FC,$E4,$6B,$8B,$FD,$98,$EA,$BF,$BE,$DF,$FB,$25,$AF,$E1,$FA
	.dc.b	$3F,$CA,$FC,$7F,$E0,$6B,$FC,$84,$BD,$5F,$F0,$06,$1F,$E0,$03,$F1
	.dc.b	$33,$A8,$CC,$66,$FF,$C1,$71,$88,$79,$3F,$E0,$55,$EB,$7F,$B4,$AB
	.dc.b	$F4,$CF,$FB,$B3,$BF,$4F,$FB,$7E,$3F,$C0,$78,$FF,$B1,$2B,$C4,$78
	.dc.b	$B4,$FD,$CF,$97,$FE,$EA,$AD,$BF,$EE,$FC,$7F,$43,$5E,$2F,$43,$E0
	.dc.b	$69,$FE,$38,$7F,$DF,$99,$EA,$91,$FE,$FE,$C3,$E9,$67,$E7,$E9,$FF
	.dc.b	$04,$BE,$78,$CE,$9B,$F1,$7F,$C9,$3C,$7F,$9F,$DB,$FC,$13,$5F,$E5
	.dc.b	$BD,$BE,$CB,$76,$F5,$9F,$E0,$DB,$FF,$63,$3E,$ED,$3F,$5B,$7E,$E8
	.dc.b	$F8,$3C,$58,$8F,$F6,$84,$72,$BF,$DB,$46,$E3,$EC,$53,$B7,$F2,$BE
	.dc.b	$57,$FB,$4B,$BF,$CF,$A3,$B7,$FC,$C3,$E7,$7F,$50,$F0,$3F,$D2,$B7
	.dc.b	$FB,$6F,$33,$FD,$AD,$7F,$B6,$9C,$BF,$E8,$4F,$1F,$F7,$1A,$FF,$BC
	.dc.b	$AF,$FB,$14,$FF,$63,$8F,$F4,$E1,$FE,$0B,$BF,$0F,$4F,$F8,$3D,$B1
	.dc.b	$FE,$B2,$7B,$CF,$F0,$08,$E8,$FF,$A8,$9F,$F5,$7D,$76,$8F,$F2,$C4
	.dc.b	$61,$F0,$E5,$FE,$C5,$9F,$F6,$84,$FF,$07,$BF,$F6,$7B,$4F,$EF,$A7
	.dc.b	$F0,$B8,$E9,$3F,$C3,$B9,$7C,$77,$FE,$65,$9F,$F6,$3D,$7F,$A2,$0F
	.dc.b	$F6,$71,$81,$E4,$7F,$B3,$5F,$F6,$6B,$5E,$07,$F8,$4A,$BA,$D7,$D3
	.dc.b	$F0,$BC,$4F,$CA,$BF,$F6,$74,$FC,$6A,$FA,$F5,$E9,$7F,$C5,$2F,$C1
	.dc.b	$AF,$F6,$5F,$57,$4B,$FE,$3F,$1F,$ED,$83,$FC,$31,$1E,$07,$FC,$38
	.dc.b	$7B,$4F,$99,$3F,$62,$BD,$CD,$7F,$C6,$2D,$5E,$1B,$C3,$FE,$53,$1F
	.dc.b	$62,$BF,$07,$E0,$FE,$21,$FF,$47,$78,$5F,$CA,$F7,$F9,$AF,$FB,$2D
	.dc.b	$F2,$78,$5F,$CB,$37,$FC,$AE,$1F,$E1,$5E,$4F,$47,$83,$FA,$08,$FD
	.dc.b	$1F,$47,$F0,$68,$C7,$6F,$FB,$36,$3F,$59,$18,$8F,$E1,$91,$FC,$77
	.dc.b	$AF,$FB,$CF,$67,$EC,$A3,$93,$5F,$F5,$0F,$5E,$B0,$FD,$74,$7F,$A9
	.dc.b	$2B,$11,$F8,$23,$FC,$29,$CB,$FC,$23,$01,$0B,$8F,$F0,$A7,$FB,$63
	.dc.b	$5F,$F5,$BB,$FD,$DF,$0F,$E2,$B1,$C4,$FE,$85,$4F,$1B,$CB,$E0,$7F
	.dc.b	$80,$EF,$83,$FC,$79,$FE,$4B,$97,$F1,$06,$BF,$C1,$9A,$FA,$DF,$E6
	.dc.b	$B7,$DC,$EF,$E3,$7F,$B0,$BC,$1C,$78,$D8,$BF,$E7,$B5,$DF,$A7,$FA
	.dc.b	$35,$F7,$6C,$F7,$6B,$DD,$FC,$E6,$7F,$E7,$CF,$07,$F5,$9E,$3F,$87
	.dc.b	$0F,$93,$F8,$37,$FA,$CE,$BF,$D7,$78,$7F,$2A,$FC,$1B,$11,$FE,$3C
	.dc.b	$AF,$45,$62,$3E,$35,$7F,$30,$15,$FC,$97,$E0,$7C,$32,$F3,$DE,$27
	.dc.b	$FA,$06,$3F,$87,$76,$7E,$62,$7F,$D2,$31,$FE,$0D,$3F,$76,$71,$F0
	.dc.b	$1E,$27,$F9,$9A,$3F,$D1,$3D,$FC,$EF,$9F,$7F,$7E,$E6,$7E,$2C,$BF
	.dc.b	$C9,$78,$7F,$7C,$D7,$FB,$0E,$B8,$8F,$76,$FF,$83,$DF,$11,$FD,$C6
	.dc.b	$BF,$BC,$56,$38,$3E,$55,$AF,$E2,$7E,$20,$FC,$79,$F8,$D7,$F8,$81
	.dc.b	$BE,$73,$F2,$5D,$C6,$0C,$FE,$3F,$8A,$E0,$71,$E7,$FE,$49,$F8,$17
	.dc.b	$F2,$6E,$6F,$A5,$F2,$CF,$C1,$9F,$8C,$BE,$13,$0D,$FC,$1F,$73,$A9
	.dc.b	$C8,$FC,$2B,$35,$C6,$74,$7D,$FF,$8C,$F1,$F0,$7B,$75,$7B,$F9,$8C
	.dc.b	$19,$F8,$2C,$70,$FB,$6B,$FC,$83,$EB,$7F,$4C,$EA,$72,$F0,$AF,$8B
	.dc.b	$CB,$FD,$B7,$BF,$FC,$A7,$5A,$9E,$59,$39,$B5,$3F,$E5,$8C,$39,$65
	.dc.b	$F3,$F2,$FF,$24,$4B,$A7,$EB,$F4,$7F,$79,$F5,$7A,$F3,$7C,$2E,$2F
	.dc.b	$89,$CD,$9C,$3F,$CC,$73,$9E,$4F,$99,$CB,$99,$4E,$79,$BE,$96,$8F
	.dc.b	$53,$AE,$5F,$2B,$1F,$A5,$39,$C3,$AC,$E6,$FE,$94,$E6,$7A,$CE,$4F
	.dc.b	$D6,$3F,$E7,$6D,$7F,$23,$E3,$E2,$3D,$DD,$9F,$94,$BF,$D0,$C6,$E3
	.dc.b	$FC,$A5,$E3,$FF,$90,$3D,$BE,$18,$7F,$2A,$F6,$BF,$C9,$DE,$86,$2F
	.dc.b	$7E,$17,$F9,$1B,$27,$EE,$FF,$EA,$69,$FF,$4C,$46,$33,$ED,$1F,$C3
	.dc.b	$7C,$AF,$DF,$65,$E6,$6C,$7E,$A1,$AF,$F3,$1E,$2F,$EB,$F1,$88,$FE
	.dc.b	$79,$1F,$C0,$23,$11,$FD,$02,$3F,$D2,$FF,$00,$FD,$93,$C9,$3F,$D1
	.dc.b	$38,$13,$9C,$7C,$B3,$FC,$C5,$E6,$FF,$22,$6A,$F7,$7F,$D4,$7E,$D7
	.dc.b	$48,$FF,$6E,$4E,$9F,$F0,$7E,$7F,$C0,$BD,$DB,$1F,$CB,$7F,$4C,$AF
	.dc.b	$F3,$FE,$2F,$F0,$AF,$E5,$DF,$EC,$9F,$2E,$F1,$FF,$29,$F5,$EB,$FE
	.dc.b	$5A,$F5,$1E,$9F,$F8,$8E,$FF,$DA,$15,$8D,$7F,$4B,$1F,$ED,$3A,$C4
	.dc.b	$7F,$AE,$A3,$B5,$7F,$40,$CB,$C8,$D7,$C8,$D7,$FB,$16,$9E,$0D,$BE
	.dc.b	$04,$5B,$F9,$2F,$FB,$5E,$BE,$3F,$83,$99,$FC,$8C,$FF,$A4,$7D,$DF
	.dc.b	$A1,$18,$8F,$EC,$11,$8E,$0F,$CC,$FA,$9F,$4B,$9B,$FD,$A2,$DC,$8E
	.dc.b	$7C,$EB,$1E,$48,$FA,$FC,$CF,$4F,$F2,$59,$F7,$CF,$FB,$E4,$63,$2F
	.dc.b	$D3,$BF,$A4,$9F,$D3,$F2,$7E,$82,$31,$3F,$A3,$EE,$FF,$45,$C6,$2F
	.dc.b	$F4,$7C,$81,$DD,$E1,$F4,$FE,$3D,$6E,$BF,$BD,$BF,$BB,$DF,$F8,$37
	.dc.b	$81,$E7,$F0,$FF,$82,$F8,$7F,$CB,$CF,$E6,$D5,$B9,$FE,$09,$E2,$7B
	.dc.b	$7C,$6D,$F8,$BF,$C5,$EB,$F8,$EE,$3F,$C4,$7C,$6D,$71,$7F,$0C,$CB
	.dc.b	$A5,$FF,$3D,$BF,$E4,$7C,$3F,$C8,$9F,$E6,$5E,$47,$F2,$8F,$2B,$FA
	.dc.b	$C4,$63,$CB,$FE,$5F,$E6,$FC,$DF,$37,$F9,$9D,$EF,$CD,$FE,$CB,$E8
	.dc.b	$7F,$44,$BF,$EB,$BE,$B7,$F9,$E6,$31,$E7,$FF,$46,$8C,$67,$FC,$DB
	.dc.b	$D0,$FE,$DB,$C5,$F1,$27,$FD,$00,$FA,$DB,$CB,$F0,$9F,$EC,$3C,$5F
	.dc.b	$E0,$4F,$03,$FB,$AE,$BE,$9D,$29,$E8,$AF,$F4,$54,$79,$DF,$F4,$67
	.dc.b	$E5,$7F,$13,$0F,$EC,$51,$8E,$3F,$E7,$DF,$1F,$FE,$00,$AD,$72,$BE
	.dc.b	$37,$FB,$A6,$FF,$D7,$BF,$23,$FF,$11,$7E,$11,$FE,$1A,$E5,$75,$7F
	.dc.b	$6A,$CF,$FB,$6F,$E8,$FF,$D8,$B9,$7F,$BD,$FE,$DD,$FD,$63,$93,$FA
	.dc.b	$E4,$7F,$6C,$CB,$FA,$67,$2F,$F4,$BB,$FF,$67,$77,$7E,$03,$8F,$FB
	.dc.b	$C4,$7F,$35,$EF,$FF,$25,$79,$7F,$CD,$FC,$57,$8B,$FE,$DF,$F5,$7F
	.dc.b	$41,$5F,$E3,$FA,$FF,$18,$F9,$3F,$C6,$DB,$B7,$EF,$25,$F4,$8E,$3F
	.dc.b	$E5,$3E,$AE,$EB,$F9,$2D,$FF,$1C,$BF,$F3,$47,$E7,$3F,$C4,$27,$F6
	.dc.b	$11,$FE,$74,$8F,$E0,$9E,$2F,$F2,$5F,$37,$F8,$04,$F8,$FC,$DF,$E6
	.dc.b	$39,$F7,$7A,$F8,$FE,$89,$F6,$BF,$B1,$E0,$F6,$F9,$9F,$EF,$28,$FD
	.dc.b	$74,$FF,$18,$7F,$51,$DD,$FA,$0E,$3F,$D1,$EB,$FD,$AA,$31,$3F,$9D
	.dc.b	$E8,$FA,$1E,$0F,$E5,$3C,$3F,$F0,$B7,$5F,$E5,$3A,$3F,$BA,$65,$FE
	.dc.b	$42,$EE,$FB,$DD,$DF,$8E,$C7,$8E,$9F,$E1,$6F,$07,$F0,$FF,$29,$FE
	.dc.b	$65,$CF,$FC,$2E,$3F,$2D,$C5,$FE,$9F,$F2,$CB,$E1,$FF,$86,$3B,$BF
	.dc.b	$C4,$9F,$1D,$FA,$8F,$27,$F0,$CF,$EA,$BF,$57,$B5,$FE,$1E,$F7,$EE
	.dc.b	$7F,$E2,$4C,$3F,$C5,$9F,$86,$7F,$81,$FF,$2D,$FC,$73,$FE,$09,$FC
	.dc.b	$BE,$F4,$FE,$72,$BF,$C1,$D7,$E6,$DF,$F1,$FC,$43,$F5,$9B,$E1,$FD
	.dc.b	$77,$07,$F8,$3F,$C6,$FF,$40,$76,$3E,$0B,$0E,$32,$FE,$5D,$3F,$E9
	.dc.b	$DE,$DF,$F5,$77,$BF,$AF,$73,$FD,$17,$5A,$BF,$F4,$FD,$7F,$83,$3E
	.dc.b	$0B,$FD,$27,$1F,$CB,$CF,$F6,$24,$FF,$3E,$C3,$C3,$1F,$9A,$CF,$FC
	.dc.b	$EC,$FF,$B2,$7E,$0B,$FD,$1D,$3A,$6C,$73,$3F,$9D,$7D,$67,$F8,$43
	.dc.b	$CC,$EF,$18,$F5,$3F,$43,$87,$C3,$C3,$DF,$B6,$23,$BD,$7F,$43,$7A
	.dc.b	$BF,$EA,$96,$BF,$E2,$F6,$BF,$EE,$56,$BE,$D7,$FC,$BE,$74,$F8,$0A
	.dc.b	$F6,$DA,$DF,$D0,$FB,$AC,$ED,$FE,$5C,$AE,$2F,$F8,$4F,$E0,$2F,$07
	.dc.b	$F0,$16,$BF,$EB,$A7,$BD,$FF,$82,$5B,$97,$FA,$EF,$0F,$DE,$4F,$F8
	.dc.b	$42,$7C,$F3,$FD,$85,$B9,$FF,$80,$67,$FC,$1F,$AF,$CA,$AF,$EC,$DE
	.dc.b	$F3,$CC,$FF,$01,$D7,$D8,$6C,$FF,$B2,$6F,$FB,$EE,$9F,$3E,$3E,$B0
	.dc.b	$E0,$63,$A9,$FB,$F1,$FE,$C4,$F2,$FC,$B5,$9C,$7E,$BF,$95,$F8,$78
	.dc.b	$FC,$BB,$5E,$B8,$FF,$79,$C3,$C1,$C1,$FD,$E3,$DE,$FF,$48,$E1,$F9
	.dc.b	$9E,$01,$39,$B8,$5F,$F1,$4B,$FF,$72,$E5,$F9,$BC,$BF,$DC,$8D,$7F
	.dc.b	$21,$97,$F8,$06,$BF,$5D,$D5,$FC,$1B,$A7,$11,$FA,$EB,$FD,$AF,$4F
	.dc.b	$F6,$DE,$AF,$EE,$53,$FB,$5E,$DF,$F4,$27,$66,$21,$1F,$C0,$6F,$F8
	.dc.b	$1C,$FF,$5F,$ED,$D7,$85,$FD,$D3,$C0,$BE,$0F,$F8,$06,$FF,$88,$BF
	.dc.b	$DD,$5B,$AF,$FA,$E7,$C2,$FF,$71,$BE,$7F,$2F,$F2,$EF,$F1,$ED,$B2
	.dc.b	$79,$E3,$F3,$7A,$79,$9B,$A7,$F7,$1F,$0F,$FD,$DB,$3F,$C5,$7E,$1B
	.dc.b	$F4,$3F,$3B,$FC,$BB,$C5,$F0,$E3,$FD,$A1,$F5,$FB,$1F,$EB,$7F,$CB
	.dc.b	$BF,$87,$F3,$FE,$1B,$C5,$FE,$19,$D7,$F5,$89,$7F,$7E,$7F,$80,$4F
	.dc.b	$F0,$7E,$CF,$F3,$5F,$8F,$FB,$2F,$B0,$FE,$C1,$F1,$1F,$94,$BE,$9D
	.dc.b	$5F,$A0,$C3,$F6,$31,$F9,$41,$F7,$BF,$B2,$CF,$C0,$78,$7D,$3F,$EC
	.dc.b	$9E,$56,$23,$1E,$4F,$EB,$AF,$F3,$B7,$FA,$97,$FD,$05,$E4,$78,$5F
	.dc.b	$3F,$8D,$E5,$F6,$7F,$39,$1F,$E8,$7E,$DF,$63,$FD,$8F,$C7,$FC,$BB
	.dc.b	$FC,$AB,$2F,$85,$3F,$91,$8F,$ED,$15,$FE,$EC,$E9,$FF,$66,$72,$FE
	.dc.b	$0B,$B7,$FB,$0F,$6F,$D2,$AF,$F0,$11,$0F,$F0,$EB,$F6,$BA,$3E,$97
	.dc.b	$47,$FB,$EF,$ED,$3F,$8B,$F2,$FF,$49,$F8,$CF,$E3,$3F,$99,$E2,$11
	.dc.b	$FE,$07,$8C,$3F,$99,$9B,$5F,$E1,$1A,$FE,$A5,$FA,$17,$E0,$F2,$FE
	.dc.b	$A9,$0E,$EF,$EE,$7C,$1B,$8C,$4F,$9C,$FF,$0D,$7A,$9F,$A5,$BE,$53
	.dc.b	$FD,$77,$C4,$FF,$24,$63,$C7,$C3,$FF,$1F,$FC,$0F,$F8,$D5,$FF,$20
	.dc.b	$FC,$15,$E0,$FF,$7F,$57,$FA,$27,$9B,$FD,$E2,$D7,$FD,$E1,$CF,$E6
	.dc.b	$7F,$83,$FB,$DF,$C3,$A7,$4F,$F1,$CF,$4B,$FD,$F1,$3A,$E9,$FE,$3F
	.dc.b	$5F,$E9,$A8,$EB,$DD,$CD,$FF,$74,$7A,$55,$FC,$67,$C4,$FF,$61,$7F
	.dc.b	$1C,$7E,$87,$C5,$FE,$4F,$C3,$FD,$EE,$33,$F5,$7F,$CA,$BC,$5F,$F5
	.dc.b	$AC,$70,$17,$FD,$75,$C1,$FE,$E2,$AE,$4F,$6F,$4F,$11,$AF,$8B,$B3
	.dc.b	$CE,$AB,$D7,$C5,$F1,$D7,$48,$FF,$64,$1B,$8C,$BE,$3B,$CB,$EC,$BF
	.dc.b	$E9,$5F,$05,$F0,$63,$FD,$96,$D7,$F8,$06,$1F,$D1,$5F,$7E,$95,$F7
	.dc.b	$78,$9D,$9F,$8F,$3F,$EC,$9D,$B9,$47,$EA,$1A,$FC,$4E,$CF,$B3,$C3
	.dc.b	$FC,$E9,$AF,$F8,$2D,$AF,$F6,$07,$F0,$91,$F5,$1A,$FF,$06,$AF,$E7
	.dc.b	$1C,$FF,$E8,$EE,$6A,$7F,$A4,$AB,$F2,$8B,$DE,$4F,$FB,$1F,$A7,$FA
	.dc.b	$07,$57,$FA,$86,$FF,$44,$FF,$AE,$FE,$17,$3E,$27,$FA,$EA,$72,$BC
	.dc.b	$D5,$23,$0B,$E2,$3F,$41,$97,$83,$3F,$16,$6F,$77,$F3,$3B,$7F,$A3
	.dc.b	$7A,$FE,$0F,$03,$CD,$47,$E6,$E7,$FD,$25,$D9,$F4,$5F,$F4,$CB,$5F
	.dc.b	$E7,$1E,$27,$F3,$2E,$7F,$A3,$DF,$F8,$A9,$FF,$4F,$F6,$7F,$89,$9F
	.dc.b	$CC,$7C,$77,$FA,$45,$BD,$BC,$DF,$E8,$55,$F8,$8A,$FC,$CC,$7F,$AA
	.dc.b	$3A,$B4,$FF,$08,$6A,$9F,$C2,$4F,$08,$FF,$47,$78,$3F,$B3,$9F,$EA
	.dc.b	$1C,$FF,$72,$7B,$BF,$63,$8F,$E4,$1F,$63,$87,$E9,$F3,$FD,$0E,$7F
	.dc.b	$9F,$F9,$7F,$C6,$F1,$7F,$8E,$FA,$DF,$E8,$1E,$4F,$F6,$5E,$DF,$E7
	.dc.b	$38,$FF,$41,$46,$21,$1F,$DE,$E3,$1F,$29,$FE,$74,$F9,$0F,$F3,$C7
	.dc.b	$4F,$F1,$C9,$7C,$4F,$F2,$A8,$FE,$3B,$18,$C3,$FD,$01,$1F,$93,$87
	.dc.b	$CC,$6F,$B3,$F0,$27,$E0,$BC,$0F,$BB,$19,$E1,$F7,$DB,$C9,$CE,$3D
	.dc.b	$30,$FE,$53,$CD,$FC,$EB,$5F,$F3,$5E,$68,$C7,$77,$FA,$25,$BD,$CF
	.dc.b	$95,$FF,$44,$D6,$F1,$FF,$4B,$60,$C3,$F2,$15,$FA,$DA,$78,$9D,$27
	.dc.b	$F9,$93,$5F,$F2,$6F,$81,$FC,$52,$FF,$CF,$AD,$8F,$F4,$EB,$5F,$F4
	.dc.b	$47,$DE,$7F,$A1,$F1,$F0,$C7,$F9,$5A,$1F,$09,$FE,$93,$BF,$F0,$34
	.dc.b	$7F,$AA,$FC,$A7,$85,$E1,$BF,$F4,$E7,$C0,$FF,$A2,$EF,$FD,$37,$8F
	.dc.b	$79,$FF,$44,$79,$3B,$7D,$DD,$1F,$CC,$2B,$FD,$27,$5B,$9F,$F4,$E3
	.dc.b	$FE,$8A,$8F,$F5,$AF,$07,$FA,$29,$FF,$5B,$FC,$17,$FA,$33,$EF,$7F
	.dc.b	$D1,$15,$FD,$B1,$FF,$44,$37,$C2,$FF,$44,$35,$FF,$44,$3E,$3C,$3F
	.dc.b	$7D,$E2,$F7,$7F,$C4,$96,$8F,$24,$7F,$AE,$5A,$FF,$B1,$DB,$E7,$EF
	.dc.b	$4E,$9A,$7E,$DF,$06,$7F,$E8,$7D,$BC,$1B,$7F,$A0,$63,$E6,$DD,$CB
	.dc.b	$E9,$5D,$7E,$BF,$0D,$7F,$D1,$7E,$37,$8B,$0F,$DE,$D7,$69,$FF,$42
	.dc.b	$F2,$76,$F1,$7F,$1D,$39,$E2,$FC,$97,$0F,$F4,$E9,$CE,$9F,$91,$9C
	.dc.b	$CF,$FA,$16,$73,$87,$FA,$43,$A3,$A7,$6F,$E9,$FB,$1D,$7F,$CB,$3B
	.dc.b	$18,$7E,$C3,$4F,$D4,$76,$FE,$B2,$B3,$87,$17,$F6,$55,$FB,$9A,$FF
	.dc.b	$41,$DF,$FA,$2B,$0F,$E9,$93,$97,$FC,$3A,$FF,$06,$CB,$FC,$B5,$C9
	.dc.b	$E0,$F0,$7F,$88,$F9,$5F,$D0,$FC,$5F,$E2,$C3,$E0,$35,$9F,$F1,$39
	.dc.b	$FF,$43,$57,$0F,$1B,$FD,$0D,$97,$F8,$CF,$F3,$1F,$F4,$E3,$71,$FF
	.dc.b	$4D,$36,$FB,$1F,$D4,$7B,$FF,$91,$63,$D7,$0F,$EA,$67,$F5,$4F,$0B
	.dc.b	$F9,$07,$8D,$E4,$BF,$37,$C5,$FF,$1E,$E2,$FE,$2B,$C9,$FC,$BE,$FE
	.dc.b	$56,$3F,$EA,$6F,$27,$FA,$9F,$37,$E6,$7A,$2C,$FE,$E9,$FB,$3D,$1F
	.dc.b	$EB,$EB,$FE,$13,$3F,$BE,$E5,$FF,$47,$78,$BF,$9A,$BF,$DF,$47,$FB
	.dc.b	$5B,$ED,$BF,$AC,$56,$E3,$FD,$1E,$EC,$FD,$33,$FE,$DC,$F8,$8F,$D3
	.dc.b	$47,$EA,$DF,$D1,$3C,$63,$F8,$E1,$FD,$AF,$DD,$FE,$13,$19,$8F,$C8
	.dc.b	$78,$1F,$AE,$B7,$AF,$F9,$F9,$E3,$8F,$E7,$BB,$FF,$81,$F3,$FE,$52
	.dc.b	$3D,$FB,$7F,$B6,$A7,$3D,$FF,$8D,$E2,$FE,$B1,$7F,$89,$E3,$F9,$BD
	.dc.b	$DF,$E0,$18,$FA,$7C,$DF,$EC,$9E,$CF,$C1,$75,$FE,$23,$A3,$FD,$A9
	.dc.b	$D9,$FE,$C5,$E7,$FA,$7B,$7F,$93,$FD,$4F,$F0,$2F,$8F,$FE,$7C,$E3
	.dc.b	$FF,$3B,$F3,$7F,$9F,$8F,$F4,$3F,$4D,$8F,$F3,$E0,$F9,$8F,$F4,$5F
	.dc.b	$46,$63,$FD,$01,$1F,$E8,$EE,$0F,$9B,$7F,$E9,$AD,$FA,$39,$E4,$FF
	.dc.b	$1E,$E7,$FE,$16,$F3,$BF,$D5,$3D,$1F,$40,$FE,$5F,$BF,$E2,$BB,$FF
	.dc.b	$27,$C7,$FD,$E7,$E1,$FF,$AE,$E5,$F9,$A9,$FC,$BF,$C4,$61,$F8,$1C
	.dc.b	$5D,$31,$FF,$80,$38,$F9,$3E,$13,$FA,$C7,$9B,$FE,$C5,$F6,$7E,$1E
	.dc.b	$41,$3A,$E6,$FF,$66,$EF,$A9,$CE,$3F,$EB,$F9,$FF,$3C,$E2,$FC,$4F
	.dc.b	$C1,$3F,$D3,$A3,$15,$D7,$0F,$F6,$D4,$FF,$69,$E7,$FF,$6A,$3F,$07
	.dc.b	$CB,$46,$23,$FB,$77,$C5,$3C,$EC,$E1,$DA,$53,$FD,$8A,$B8,$46,$25
	.dc.b	$E4,$BF,$2E,$E1,$2C,$73,$18,$8F,$AF,$F5,$5E,$6F,$4A,$C7,$FB,$C7
	.dc.b	$C6,$FB,$B8,$AB,$FA,$FF,$27,$E2,$5E,$1E,$47,$79,$F8,$17,$F0,$23
	.dc.b	$11,$F2,$A3,$F0,$07,$8F,$C0,$F4,$ED,$EF,$6C,$9C,$FA,$36,$3D,$7A
	.dc.b	$FF,$3A,$E3,$F5,$77,$FD,$09,$ED,$8F,$FA,$27,$4F,$F7,$BF,$47,$FA
	.dc.b	$EA,$BF,$09,$3F,$6F,$A3,$FA,$1D,$6B,$9F,$F0,$35,$AA,$F3,$E5,$FD
	.dc.b	$13,$1F,$C4,$74,$FF,$30,$EE,$FF,$73,$BF,$98,$F1,$3F,$9E,$61,$F9
	.dc.b	$BE,$EF,$E6,$7D,$FF,$4B,$CF,$FE,$7B,$8B,$E1,$7F,$A8,$79,$3F,$E9
	.dc.b	$0C,$3F,$CA,$11,$8F,$8A,$DB,$AF,$8C,$FC,$A7,$81,$FE,$10,$F0,$7D
	.dc.b	$7C,$BF,$ED,$47,$E1,$47,$F8,$53,$C0,$FC,$2F,$1F,$FA,$1D,$FC,$3F
	.dc.b	$C5,$7F,$1A,$F0,$7E,$AD,$7F,$86,$BB,$6D,$FE,$07,$B3,$7E,$47,$D1
	.dc.b	$E1,$FB,$B3,$FE,$74,$EC,$FC,$6F,$81,$F2,$A7,$F0,$3B,$7F,$AC,$FB
	.dc.b	$33,$C7,$EC,$F2,$FF,$B9,$F2,$FF,$9E,$3E,$23,$FD,$33,$F8,$77,$F0
	.dc.b	$6E,$47,$D1,$7F,$17,$DF,$FC,$E5,$A7,$ED,$FE,$23,$F3,$B1,$93,$F9
	.dc.b	$99,$FC,$D3,$CD,$43,$8F,$F9,$BD,$7E,$FE,$BF,$8E,$D7,$F1,$9B,$ED
	.dc.b	$E6,$FF,$18,$C3,$FC,$0B,$97,$8F,$F1,$EF,$E0,$D8,$FF,$63,$F0,$7F
	.dc.b	$D7,$3F,$35,$FE,$B0,$9C,$F8,$5F,$C5,$7C,$2F,$F5,$FF,$C2,$E7,$0F
	.dc.b	$E1,$D3,$93,$FD,$85,$C0,$BE,$38,$FE,$27,$E1,$BF,$D7,$7B,$7F,$B9
	.dc.b	$3E,$1E,$E3,$FC,$5F,$C8,$1F,$EA,$9E,$CF,$E3,$F5,$FE,$A7,$7F,$96
	.dc.b	$79,$BC,$2B,$FD,$6B,$1F,$B0,$CD,$F8,$77,$5F,$07,$E2,$D7,$4C,$7F
	.dc.b	$DE,$F1,$87,$FD,$FA,$BF,$63,$F2,$74,$5A,$FF,$5C,$57,$1D,$3E,$87
	.dc.b	$47,$FB,$FB,$E2,$BF,$8D,$5F,$3A,$EC,$7F,$8F,$72,$F7,$76,$FE,$03
	.dc.b	$1F,$74,$FB,$75,$FF,$3C,$CE,$67,$D7,$1F,$D2,$63,$13,$F1,$3E,$23
	.dc.b	$F8,$6E,$FF,$E3,$EE,$0E,$BC,$7F,$3B,$9B,$ED,$61,$F4,$72,$FA,$BC
	.dc.b	$9F,$46,$31,$1F,$7E,$7F,$DF,$3C,$FF,$E4,$8B,$E5,$CD,$FE,$2C,$E7
	.dc.b	$FB,$D7,$B8,$7C,$47,$D1,$DB,$DE,$7F,$B9,$B9,$FC,$EF,$B7,$83,$FC
	.dc.b	$B9,$1F,$52,$FE,$5C,$FA,$F9,$FE,$43,$CB,$87,$F4,$33,$AF,$8A,$FD
	.dc.b	$D7,$47,$FB,$EA,$3F,$BE,$7D,$17,$F3,$67,$F3,$2F,$EF,$63,$FD,$FF
	.dc.b	$CD,$D3,$0F,$C0,$5F,$F8,$72,$31,$1F,$EE,$EE,$6F,$D0,$BE,$A3,$E7
	.dc.b	$71,$FA,$B2,$FF,$7A,$F2,$EA,$75,$1F,$1F,$3F,$9B,$C9,$FE,$4B,$E0
	.dc.b	$6E,$FC,$CB,$E5,$F3,$7F,$94,$A1,$F5,$FF,$A1,$6E,$1F,$13,$2F,$F2
	.dc.b	$44,$FE,$7F,$8F,$FB,$83,$5F,$F6,$B7,$47,$68,$C4,$7F,$B6,$6D,$F8
	.dc.b	$8F,$53,$D1,$7E,$DE,$5F,$F8,$00,$FB,$C7,$67,$F6,$CF,$03,$71,$F7
	.dc.b	$27,$6F,$DE,$EC,$FE,$F1,$18,$8F,$CF,$C7,$FB,$57,$F4,$9F,$D8,$1C
	.dc.b	$A3,$F4,$7C,$C8,$C4,$7E,$CE,$31,$B7,$C1,$C7,$F8,$E6,$3F,$AD,$8A
	.dc.b	$7E,$23,$0F,$D7,$C7,$2B,$FE,$0F,$7E,$37,$F6,$37,$FE,$A2,$CB,$51
	.dc.b	$70,$ED,$CB,$F4,$1F,$D7,$EB,$FF,$00,$7D,$4F,$F5,$4E,$FF,$32,$2C
	.dc.b	$FE,$E3,$97,$F6,$37,$C7,$0F,$94,$6A,$6E,$5F,$EB,$2E,$3F,$C0,$F2
	.dc.b	$BA,$DC,$B9,$BF,$C4,$C7,$FA,$6B,$C0,$C4,$7E,$BF,$BE,$CF,$F2,$DE
	.dc.b	$EF,$C0,$4F,$E1,$F5,$AF,$AB,$67,$93,$F9,$2E,$2F,$8D,$5F,$76,$B8
	.dc.b	$73,$FC,$7E,$0F,$E5,$FD,$BF,$D9,$FA,$3F,$52,$7E,$AB,$9F,$EB,$C7
	.dc.b	$EA,$5E,$7C,$9F,$8E,$C3,$C4,$FC,$3B,$FE,$01,$A7,$8A,$31,$1F,$BD
	.dc.b	$7F,$49,$3F,$03,$85,$C5,$F1,$7A,$FF,$53,$EB,$7F,$58,$AF,$F3,$D7
	.dc.b	$B4,$FB,$C7,$DA,$7F,$65,$F1,$7F,$B2,$9F,$DF,$FF,$2D,$FE,$B7,$E6
	.dc.b	$FF,$9D,$B0,$FE,$BB,$C6,$FB,$3E,$17,$C2,$31,$3F,$DA,$63,$FD,$03
	.dc.b	$97,$F9,$3A,$B7,$1F,$E8,$79,$FF,$28,$FE,$EF,$D9,$EC,$7F,$95,$3F
	.dc.b	$17,$DC,$FF,$0F,$6B,$FD,$DE,$78,$4F,$97,$C2,$F0,$BF,$E9,$49,$ED
	.dc.b	$18,$7F,$BD,$E1,$FD,$FF,$0E,$72,$CF,$9D,$FF,$A5,$3F,$2B,$F8,$AF
	.dc.b	$AF,$0F,$F4,$FF,$1B,$83,$E6,$4F,$F8,$03,$24,$E6,$38,$57,$CC,$3E
	.dc.b	$9B,$F1,$7D,$CF,$F5,$4F,$0F,$68,$F4,$4B,$9F,$E4,$47,$28,$FF,$54
	.dc.b	$70,$7A,$DE,$9F,$84,$7F,$8D,$78,$7E,$3D,$FF,$86,$DB,$BE,$EB,$EE
	.dc.b	$D7,$F3,$9E,$0D,$EF,$E1,$9F,$91,$87,$AB,$7F,$F1,$66,$BF,$01,$EB
	.dc.b	$AE,$EB,$FC,$5F,$AF,$F1,$AC,$7E,$1B,$FE,$A7,$DF,$D2,$FF,$AF,$CF
	.dc.b	$F5,$FE,$DE,$9C,$7F,$1F,$E8,$7C,$B8,$C6,$1F,$E3,$68,$FF,$1A,$C6
	.dc.b	$23,$FC,$71,$7F,$85,$BF,$C7,$67,$FE,$CA,$8C,$47,$73,$FD,$B1,$B3
	.dc.b	$7F,$2C,$7F,$AE,$E3,$0F,$23,$F4,$AF,$7B,$FF,$6E,$57,$E1,$E7,$EC
	.dc.b	$47,$97,$5C,$47,$F9,$33,$3F,$5F,$A1,$8A,$FF,$74,$CF,$8B,$5F,$0E
	.dc.b	$0D,$FE,$56,$9E,$9E,$1F,$C4,$5F,$F9,$5B,$DE,$FF,$25,$FD,$97,$CB
	.dc.b	$9C,$E4,$13,$AA,$FC,$A7,$17,$D5,$D7,$FD,$EB,$C3,$EB,$BF,$F3,$D6
	.dc.b	$BF,$71,$F9,$D7,$F2,$1E,$95,$FE,$FC,$E0,$F8,$B3,$ED,$9F,$CD,$47
	.dc.b	$C5,$E2,$F9,$57,$CE,$7F,$84,$ED,$E1,$7F,$41,$8B,$F7,$5F,$D6,$E3
	.dc.b	$FE,$66,$3F,$DA,$3F,$BA,$FE,$F2,$4F,$F0,$6C,$7F,$B2,$B6,$7F,$DA
	.dc.b	$FD,$7C,$9F,$E2,$7F,$81,$7E,$E8,$FD,$DC,$66,$3E,$F4,$F1,$D3,$AC
	.dc.b	$7E,$AF,$2F,$BB,$18,$8F,$D8,$C7,$FA,$0F,$2E,$51,$F5,$FF,$36,$F3
	.dc.b	$5F,$EB,$B3,$E1,$7F,$BD,$1D,$1F,$EA,$7C,$FF,$62,$7E,$CA,$BF,$90
	.dc.b	$C7,$F8,$A5,$FD,$8D,$FE,$E3,$3F,$BB,$7F,$CA,$1A,$FE,$9B,$16,$0F
	.dc.b	$D2,$7C,$AB,$F5,$CF,$E4,$BF,$AA,$5A,$F7,$5F,$E1,$E3,$84,$0F,$F7
	.dc.b	$B9,$F4,$FF,$51,$EF,$3F,$84,$6B,$FB,$06,$BF,$EA,$A7,$FD,$53,$1C
	.dc.b	$FB,$FF,$9C,$BF,$D4,$BB,$3D,$F5,$E8,$C3,$F8,$AE,$1E,$5A,$FE,$3F
	.dc.b	$D4,$C7,$AE,$FF,$9D,$7B,$E1,$FC,$8F,$3E,$0F,$F3,$A6,$BF,$CB,$78
	.dc.b	$FD,$98,$E2,$3E,$43,$5F,$E1,$9D,$7D,$F4,$FE,$A7,$F4,$48,$C6,$9F
	.dc.b	$85,$8F,$E7,$D9,$7C,$07,$FD,$8F,$97,$3C,$3F,$65,$9F,$CD,$D7,$DF
	.dc.b	$7B,$C1,$9B,$F3,$EF,$AB,$F5,$DF,$8C,$8C,$4F,$D6,$DD,$F9,$C7,$EF
	.dc.b	$23,$EB,$7E,$18,$FC,$C5,$F9,$5B,$F1,$F7,$E7,$CF,$11,$E8,$7F,$AE
	.dc.b	$3E,$BD,$FA,$1F,$FB,$53,$93,$EA,$C3,$F2,$0F,$F6,$BF,$D1,$FD,$8F
	.dc.b	$80,$7E,$73,$FE,$DA,$FB,$0C,$71,$FF,$B7,$3E,$29,$F9,$A7,$FB,$C3
	.dc.b	$E8,$1E,$3B,$F0,$C7,$E5,$5F,$C5,$33,$F5,$74,$3F,$6C,$9F,$97,$3E
	.dc.b	$25,$F8,$E3,$F4,$C7,$E0,$6E,$1F,$F4,$0F,$0F,$FB,$B3,$F1,$A7,$E1
	.dc.b	$8F,$C8,$5F,$B4,$3F,$36,$7E,$3F,$F1,$3F,$12,$FF,$44,$FE,$15,$EB
	.dc.b	$E2,$FF,$06,$F1,$FF,$B1,$7C,$9F,$F6,$B7,$EC,$2F,$DE,$3D,$BF,$A5
	.dc.b	$3F,$38,$78,$1F,$9C,$C7,$F8,$C7,$1F,$E7,$BD,$47,$87,$FC,$6F,$CD
	.dc.b	$FF,$61,$FE,$9A,$FC,$C7,$FD,$13,$F5,$4F,$D9,$DF,$69,$FE,$0F,$F6
	.dc.b	$78,$FE,$8B,$FB,$3D,$5F,$A9,$3F,$04,$FF,$50,$FE,$7A,$FC,$53,$FC
	.dc.b	$C1,$F7,$8F,$C8,$BC,$3F,$BA,$FD,$DF,$84,$F6,$F8,$BC,$FE,$C5,$FA
	.dc.b	$DB,$EA,$1F,$BC,$BF,$68,$FB,$1F,$BB,$7F,$92,$3C,$37,$C9,$F8,$FF
	.dc.b	$0F,$7E,$42,$FB,$4F,$EA,$1F,$9C,$BF,$64,$7E,$0A,$FC,$0B,$FD,$6F
	.dc.b	$F0,$1E,$BF,$63,$F1,$99,$FB,$7F,$47,$FE,$F7,$FA,$BB,$F3,$4F,$B5
	.dc.b	$FB,$CB,$AF,$E7,$78,$1F,$E0,$93,$E7,$75,$7C,$DF,$A0,$F8,$DF,$BA
	.dc.b	$FE,$4B,$F7,$DF,$5F,$EE,$EF,$DA,$5F,$AF,$3F,$33,$FC,$4E,$5F,$C7
	.dc.b	$FF,$0E,$7E,$BF,$FC,$87,$F7,$07,$E1,$8F,$D0,$1F,$B7,$7E,$AF,$C0
	.dc.b	$F8,$78,$FF,$27,$F6,$1F,$A1,$FF,$B1,$3F,$07,$FD,$A7,$D9,$BF,$15
	.dc.b	$7E,$89,$E0,$C1,$FB,$33,$F2,$C7,$E6,$2F,$D0,$DF,$AE,$FF,$B3,$B2
	.dc.b	$7E,$C3,$FE,$E0,$F0,$FF,$09,$F0,$BF,$BD,$FD,$29,$FB,$D7,$3D,$1E
	.dc.b	$D3,$F0,$7F,$F7,$FE,$EF,$C7,$DF,$B7,$7E,$03,$D6,$E3,$FB,$5B,$F6
	.dc.b	$DE,$1F,$12,$FC,$5F,$F2,$5F,$5D,$FC,$3F,$EB,$7F,$87,$7D,$67,$F0
	.dc.b	$DF,$AB,$FE,$19,$FA,$B7,$E5,$FE,$3D,$F8,$DB,$F7,$9F,$1F,$CF,$BF
	.dc.b	$49,$7E,$E9,$FE,$87,$F9,$5F,$83,$CB,$FD,$47,$EB,$FF,$89,$F8,$DF
	.dc.b	$C5,$63,$FC,$42,$7E,$8E,$51,$98,$FF,$49,$47,$F0,$07,$FA,$D7,$CC
	.dc.b	$7F,$16,$7F,$AE,$7E,$9B,$FC,$C3,$D2,$7E,$89,$E3,$FD,$1F,$FD,$77
	.dc.b	$FB,$C3,$CF,$F8,$18,$7F,$63,$FA,$CF,$E2,$BF,$07,$FC,$53,$EA,$BF
	.dc.b	$89,$FA,$5F,$4E,$BF,$B4,$7D,$4F,$F1,$2F,$A9,$FE,$23,$F5,$1F,$C4
	.dc.b	$3B,$1F,$4F,$F8,$1F,$C6,$3F,$9A,$7E,$64,$FD,$13,$FD,$5F,$F6,$2F
	.dc.b	$C8,$7F,$97,$F0,$FF,$26,$FC,$ED,$83,$F2,$B7,$C6,$3F,$52,$7E,$9C
	.dc.b	$FD,$F7,$F1,$98,$FF,$08,$E3,$FC,$B1,$FC,$7B,$D8,$FE,$19,$F8,$1B
	.dc.b	$F4,$4F,$E5,$9F,$1E,$FC,$DD,$F9,$FB,$F7,$BF,$8F,$97,$FB,$EF,$ED
	.dc.b	$7F,$92,$FD,$A7,$F2,$4F,$80,$FE,$47,$AF,$ED,$87,$E7,$FF,$C9,$7C
	.dc.b	$1F,$E4,$5F,$93,$7E,$17,$ED,$3F,$90,$FD,$9F,$F2,$0F,$C6,$DF,$67
	.dc.b	$FC,$A3,$6F,$E4,$1F,$65,$FC,$7F,$7F,$CD,$FD,$8F,$F1,$EF,$B1,$FE
	.dc.b	$5B,$F8,$2F,$F1,$DD,$FF,$03,$F8,$2F,$F2,$7E,$6F,$E1,$B1,$FE,$1B
	.dc.b	$FB,$1F,$E3,$FF,$61,$FC,$7B,$EC,$3F,$8E,$FE,$23,$F7,$EF,$F8,$E7
	.dc.b	$D7,$FF,$1B,$F6,$9F,$5D,$FC,$6F,$EB,$7F,$8D,$70,$7F,$AC,$74,$FF
	.dc.b	$13,$7E,$CA,$FD,$35,$FB,$B3,$A9,$F2,$BF,$EC,$3F,$CC,$DF,$91,$FF
	.dc.b	$50,$C7,$FC,$5D,$DD,$E6,$7F,$D0,$1F,$80,$3F,$53,$7D,$43,$F2,$77
	.dc.b	$E0,$0F,$D8,$3F,$B3,$70,$BF,$30,$FE,$8D,$F8,$D7,$D0,$FA,$E7,$C0
	.dc.b	$3F,$26,$FE,$41,$F4,$2F,$CC,$7F,$81,$FC,$03,$E9,$3F,$92,$7E,$38
	.dc.b	$FD,$B5,$F1,$4F,$CB,$1F,$98,$FF,$5C,$F8,$8F,$AB,$F8,$9B,$F5,$F7
	.dc.b	$EF,$EF,$CC,$3F,$B4,$7A,$DF,$D5,$7E,$59,$FB,$F7,$F8,$3F,$E4,$3F
	.dc.b	$DF,$71,$FB,$2F,$03,$F8,$07,$83,$8F,$90,$FF,$0A,$71,$6B,$F7,$8F
	.dc.b	$EC,$5F,$96,$7E,$3A,$FF,$D0,$C7,$F6,$EF,$C0,$7F,$C3,$5E,$7A,$FF
	.dc.b	$97,$CF,$FA,$2F,$2F,$E7,$4F,$F0,$4C,$7F,$BF,$FC,$FF,$F6,$DF,$C0
	.dc.b	$7F,$C1,$13,$D0,$FE,$05,$E2,$79,$DC,$78,$7F,$81,$5F,$AE,$3F,$3B
	.dc.b	$7E,$7B,$FD,$A7,$E3,$BF,$90,$FE,$4A,$FD,$63,$F8,$36,$1F,$C0,$BC
	.dc.b	$E7,$A1,$FD,$7F,$F6,$BF,$DE,$FE,$43,$E9,$D7,$B7,$E3,$BF,$8D,$FD
	.dc.b	$E7,$FA,$D7,$ED,$8F,$CB,$7F,$C5,$9F,$6D,$D3,$F7,$D7,$E9,$4F,$9E
	.dc.b	$7D,$37,$F6,$0F,$D2,$DF,$A2,$FE,$73,$5F,$57,$DF,$FF,$42,$E2,$7E
	.dc.b	$F6,$FD,$E5,$87,$F8,$6B,$F3,$3F,$E4,$9F,$77,$FE,$13,$FD,$64,$7F
	.dc.b	$10,$FC,$D3,$FC,$7B,$E3,$BF,$68,$FF,$45,$6B,$FE,$1D,$EB,$FE,$79
	.dc.b	$FA,$9F,$F8,$BB,$BF,$E6,$47,$FC,$00,$F0,$EB,$FC,$24,$FF,$0D,$7E
	.dc.b	$63,$FC,$2F,$91,$F8,$1F,$D5,$9F,$F3,$A7,$E4,$7F,$E4,$37,$FC,$65
	.dc.b	$C1,$FC,$79,$FC,$07,$DF,$FF,$8A,$7B,$DF,$98,$BC,$04,$FE,$97,$F0
	.dc.b	$5F,$93,$D7,$FE,$38,$FB,$8F,$1E,$3F,$D3,$FD,$2F,$F2,$33,$F8,$2F
	.dc.b	$37,$F1,$1F,$3B,$FE,$2C,$8F,$F0,$17,$E0,$3F,$27,$F1,$8F,$F3,$BF
	.dc.b	$D3,$3F,$5C,$FE,$3B,$D3,$F8,$9F,$90,$FE,$B3,$ED,$3C,$BF,$F2,$5F
	.dc.b	$DC,$3F,$42,$DF,$D8,$FF,$90,$7F,$13,$7E,$30,$FD,$A3,$FA,$07,$C9
	.dc.b	$FF,$43,$FC,$71,$C3,$FE,$0D,$FC,$7F,$FD,$1B,$BB,$F4,$FF,$F0,$77
	.dc.b	$C2,$3F,$80,$7F,$C0,$33,$FE,$1A,$E6,$FE,$2B,$F3,$90,$FE,$2F,$87
	.dc.b	$F4,$29,$D7,$E8,$BF,$E3,$CF,$D0,$5F,$51,$FD,$EB,$F4,$47,$EB,$1E
	.dc.b	$2B,$FF,$3E,$FE,$B7,$FD,$83,$E2,$BE,$7F,$89,$FE,$93,$F5,$BF,$C2
	.dc.b	$95,$FC,$73,$6F,$F0,$07,$DA,$7F,$9C,$BD,$7F,$F3,$07,$DA,$BE,$39
	.dc.b	$FB,$1F,$F9,$C3,$F3,$E7,$EF,$0F,$D0,$FF,$27,$9F,$F1,$6C,$FF,$BC
	.dc.b	$7E,$32,$FD,$51,$FB,$AB,$F2,$BF,$F5,$0F,$5F,$F8,$BF,$0F,$CB,$60
	.dc.b	$F9,$68,$FF,$00,$F7,$7F,$D4,$DC,$5F,$91,$F9,$CC,$CF,$F3,$5F,$DA
	.dc.b	$5F,$B4,$FF,$8E,$BE,$E7,$FC,$81,$F1,$FF,$E3,$BB,$F7,$5F,$5C,$3F
	.dc.b	$3C,$FF,$9F,$7D,$4F,$E9,$EF,$F9,$7B,$E0,$3F,$8F,$3F,$E8,$13,$FD
	.dc.b	$A5,$A7,$F6,$DF,$7B,$FD,$A9,$1F,$E2,$0F,$C2,$3F,$90,$CF,$A7,$E0
	.dc.b	$3F,$90,$79,$D8,$8F,$BD,$5F,$E9,$4F,$B0,$DF,$EC,$9E,$A8,$FF,$4B
	.dc.b	$FC,$F7,$FA,$3F,$D3,$FF,$80,$31,$FB,$AF,$DB,$BF,$93,$3E,$3F,$9D
	.dc.b	$FF,$4A,$73,$FF,$0F,$8F,$F1,$C6,$5F,$24,$FE,$4C,$FF,$6D,$03,$F4
	.dc.b	$CF,$F2,$2F,$C4,$BE,$55,$FA,$EF,$F9,$9B,$F2,$FF,$F8,$01,$8F,$F2
	.dc.b	$FF,$9F,$7E,$C5,$FE,$E7,$FB,$87,$E7,$2F,$88,$71,$FF,$B1,$B0,$FE
	.dc.b	$6D,$FB,$9F,$F7,$3F,$CC,$7F,$D9,$3F,$21,$0F,$F1,$F7,$EB,$1F,$EA
	.dc.b	$DF,$D5,$5E,$DF,$EC,$7E,$16,$1F,$E1,$AC,$7F,$4D,$FA,$CF,$FA,$4F
	.dc.b	$F3,$1C,$CF,$F8,$57,$AF,$E2,$F4,$3A,$BF,$AA,$E1,$FA,$DE,$1F,$F6
	.dc.b	$1F,$E2,$EF,$D4,$A1,$7F,$0A,$FF,$28,$7D,$9F,$FA,$FF,$F2,$DF,$F6
	.dc.b	$37,$E7,$1F,$EC,$4F,$C6,$9C,$5F,$CC,$3E,$C7,$FC,$AD,$FB,$57,$DD
	.dc.b	$F2,$3F,$0D,$E7,$7E,$53,$F7,$27,$EF,$3F,$EB,$EF,$D5,$FF,$CE,$1F
	.dc.b	$AE,$BF,$70,$FF,$47,$73,$7F,$5F,$CF,$FB,$59,$FC,$B7,$F3,$DF,$B5
	.dc.b	$E5,$FE,$8F,$E0,$3F,$B7,$E3,$88,$FF,$55,$5F,$F6,$7F,$A5,$FE,$E7
	.dc.b	$F9,$E7,$F9,$D7,$D3,$FC,$85,$7E,$4A,$BF,$0A,$FE,$55,$FF,$3D,$7A
	.dc.b	$5F,$98,$FC,$B7,$FB,$DF,$DD,$FF,$81,$A7,$F0,$FC,$7F,$CF,$DF,$F7
	.dc.b	$5E,$3F,$CC,$9F,$9B,$E9,$FD,$07,$F4,$67,$F5,$09,$FF,$74,$7E,$C4
	.dc.b	$FC,$B9,$EF,$BF,$73,$84,$F0,$FF,$91,$7C,$B7,$F8,$DB,$F2,$7F,$E4
	.dc.b	$4F,$F2,$1F,$B6,$F5,$E7,$FA,$0F,$8F,$FF,$7A,$4E,$CF,$E5,$15,$F5
	.dc.b	$27,$DF,$C5,$FA,$DE,$DF,$E0,$FF,$86,$7F,$8E,$57,$BE,$FC,$B9,$F9
	.dc.b	$BF,$FA,$72,$94,$F9,$FF,$DE,$FC,$B7,$F9,$B7,$BF,$FC,$43,$F0,$BF
	.dc.b	$5F,$3F,$AF,$F0,$DF,$59,$FB,$35,$FD,$9F,$BB,$EE,$ED,$F1,$4F,$D9
	.dc.b	$F9,$2F,$CD,$1F,$BD,$FF,$65,$F9,$3F,$F5,$BF,$E9,$FF,$E1,$8F,$D5
	.dc.b	$5F,$83,$BF,$4B,$46,$1B,$F8,$AF,$FA,$A7,$F0,$DF,$DF,$E9,$FE,$EA
	.dc.b	$F8,$C7,$EE,$5F,$E0,$8F,$6F,$C1,$E7,$FF,$04,$F2,$5F,$91,$FF,$8C
	.dc.b	$FF,$20,$FE,$D9,$8B,$E1,$BC,$9F,$9F,$F9,$F8,$FC,$DE,$33,$F2,$3E
	.dc.b	$FF,$90,$FF,$A8,$BF,$20,$34,$69,$FE,$BD,$FC,$77,$CB,$F8,$8F,$F3
	.dc.b	$FE,$CF,$E9,$72,$FD,$F7,$FD,$D1,$C5,$FE,$C3,$EB,$F5,$FA,$9F,$E7
	.dc.b	$7F,$CA,$DF,$AE,$BF,$2E,$FF,$58,$7C,$CA,$73,$F9,$57,$F0,$EE,$6F
	.dc.b	$E4,$3F,$AA,$3D,$AF,$F7,$F3,$7F,$45,$3F,$C9,$BC,$DF,$E9,$BF,$8E
	.dc.b	$7F,$9E,$BC,$17,$EA,$FF,$EE,$9E,$7F,$A3,$FA,$37,$FB,$5F,$DD,$F9
	.dc.b	$FF,$7C,$FC,$F1,$F8,$BB,$F7,$AF,$F2,$97,$EF,$19,$8E,$1C,$3F,$CD
	.dc.b	$3E,$4B,$F8,$99,$3E,$4B,$F9,$C4,$7F,$27,$F9,$B9,$E2,$7F,$8A,$23
	.dc.b	$F9,$DF,$D1,$7F,$3B,$F3,$3E,$E7,$A7,$FC,$36,$7A,$8F,$B4,$FF,$5E
	.dc.b	$F7,$BF,$60,$7C,$76,$7D,$A4,$FF,$93,$AF,$FA,$3E,$9E,$10,$7F,$2D
	.dc.b	$D3,$FD,$EB,$1F,$B6,$F8,$3F,$F0,$CF,$ED,$DF,$76,$3F,$A7,$F8,$0F
	.dc.b	$D9,$3C,$5C,$6F,$7B,$F1,$FF,$83,$FE,$13,$27,$91,$FE,$C9,$C7,$F2
	.dc.b	$3E,$FB,$F1,$3F,$E4,$0B,$B7,$E2,$ED,$F3,$7F,$20,$75,$FC,$DE,$CF
	.dc.b	$F1,$2E,$7F,$95,$CB,$FB,$0D,$EB,$B7,$EA,$D7,$D4,$FD,$2B,$FD,$0B
	.dc.b	$2F,$81,$84,$3F,$C8,$77,$CB,$EF,$3E,$07,$E7,$4C,$9F,$BC,$7F,$23
	.dc.b	$AF,$ED,$FE,$A7,$F9,$AF,$83,$FA,$A7,$95,$FE,$18,$FC,$8B,$FD,$77
	.dc.b	$1F,$E9,$9F,$4D,$FA,$2F,$FA,$2B,$F7,$9F,$1F,$D1,$7F,$A2,$BE,$85
	.dc.b	$FB,$CF,$FA,$B6,$3E,$56,$DF,$EA,$3E,$27,$EA,$9F,$CB,$3E,$5A,$4E
	.dc.b	$7D,$1F,$57,$2F,$B7,$F4,$FF,$C2,$ED,$FE,$E8,$F4,$BF,$6D,$F7,$9F
	.dc.b	$EE,$AC,$3F,$CA,$9C,$2F,$A7,$7F,$00,$7F,$B3,$DB,$EB,$BF,$4B,$F9
	.dc.b	$3D,$9F,$83,$F9,$9F,$E2,$5D,$7F,$E4,$C3,$FD,$73,$EC,$FF,$14,$F1
	.dc.b	$DF,$B5,$7F,$2E,$F9,$4F,$F6,$D7,$ED,$CF,$D9,$3F,$A3,$F0,$7E,$5F
	.dc.b	$1E,$47,$FB,$43,$C0,$FF,$61,$70,$BD,$3F,$F2,$27,$C8,$FF,$91,$AB
	.dc.b	$FA,$C1,$FE,$A2,$F5,$3F,$17,$FB,$47,$E2,$BC,$DF,$83,$ED,$7F,$1C
	.dc.b	$F6,$BF,$98,$FE,$09,$FE,$C6,$CF,$FA,$BF,$E2,$3F,$EF,$2F,$25,$FA
	.dc.b	$EF,$F5,$AF,$D2,$BF,$0F,$3E,$FE,$1F,$EB,$3F,$9E,$7E,$11,$FF,$6B
	.dc.b	$78,$2F,$E0,$0F,$9B,$F2,$3F,$E0,$49,$F9,$BF,$43,$F2,$FC,$2F,$99
	.dc.b	$F1,$DF,$8B,$FC,$B7,$F1,$5C,$DF,$E5,$4E,$9F,$F4,$B3,$FE,$3A,$FB
	.dc.b	$47,$EE,$4F,$A2,$FF,$4F,$7E,$05,$F8,$AF,$7F,$F2,$1E,$1F,$F0,$2F
	.dc.b	$45,$FB,$4B,$F6,$18,$4F,$A1,$D7,$EF,$FF,$89,$F5,$3F,$BA,$7D,$DF
	.dc.b	$F6,$EF,$CF,$DF,$BF,$BE,$51,$F9,$4E,$1B,$D3,$F9,$3F,$CC,$9E,$E7
	.dc.b	$E6,$7D,$DF,$CC,$4F,$FB,$7B,$9F,$FC,$A7,$F9,$30,$84,$E3,$FF,$6C
	.dc.b	$FD,$17,$F7,$EF,$84,$7E,$46,$FD,$C3,$F5,$ED,$F5,$7F,$81,$ED,$E4
	.dc.b	$E0,$F3,$70,$7A,$3E,$4F,$F6,$DF,$63,$FE,$24,$F3,$58,$7F,$B9,$BD
	.dc.b	$9F,$EC,$F1,$8E,$0F,$CB,$9F,$E6,$8F,$BC,$FF,$5D,$7E,$23,$C3,$F5
	.dc.b	$BF,$F6,$67,$E6,$DF,$C6,$39,$7F,$2D,$7F,$E9,$9C,$BF,$D2,$DE,$B7
	.dc.b	$47,$FD,$31,$97,$F7,$4F,$4D,$FA,$AB,$F1,$6F,$E3,$3F,$A6,$BF,$46
	.dc.b	$FF,$67,$7E,$81,$FE,$12,$F7,$B9,$69,$F9,$C3,$FA,$29,$FE,$95,$D7
	.dc.b	$F4,$79,$FF,$A4,$4F,$F4,$16,$BF,$BE,$9F,$F0,$1F,$D8,$7F,$79,$F6
	.dc.b	$FF,$3D,$5F,$E9,$7D,$BF,$3B,$F7,$DF,$E8,$0F,$17,$FC,$5B,$E5,$87
	.dc.b	$CA,$7F,$B5,$4F,$F1,$2F,$E1,$FF,$E0,$EF,$D5,$3D,$5F,$B7,$BF,$46
	.dc.b	$FF,$14,$FD,$9F,$F8,$BB,$C7,$FC,$96,$3C,$FF,$70,$FC,$8F,$A2,$F4
	.dc.b	$73,$29,$7C,$BF,$F2,$0E,$DF,$C5,$F3,$FF,$1F,$AF,$27,$EB,$BF,$C8
	.dc.b	$FC,$17,$DC,$FF,$33,$F9,$6F,$F1,$0F,$9F,$F8,$C3,$FD,$07,$D9,$FE
	.dc.b	$86,$FD,$AB,$F9,$07,$CB,$7F,$1F,$E8,$FF,$5C,$CF,$F1,$FF,$81,$7C
	.dc.b	$AF,$F1,$DD,$BF,$CA,$BB,$7C,$3E,$AF,$F2,$7C,$7F,$B7,$FC,$0F,$F3
	.dc.b	$A1,$F4,$FF,$39,$FE,$3D,$EF,$BF,$41,$79,$4E,$A7,$EA,$2F,$D6,$5F
	.dc.b	$9C,$FF,$94,$38,$3F,$D9,$19,$FF,$A3,$3C,$5F,$F1,$DF,$D5,$BC,$AF
	.dc.b	$F7,$87,$94,$FC,$DD,$D8,$FA,$27,$BF,$FE,$5A,$FC,$1F,$F3,$B9,$7E
	.dc.b	$77,$3F,$F4,$BE,$3D,$9F,$F6,$F1,$FE,$CC,$8F,$F6,$57,$07,$E3,$75
	.dc.b	$FF,$6D,$7D,$FF,$E3,$39,$BE,$97,$47,$E1,$7E,$A3,$FC,$3B,$F8,$9F
	.dc.b	$DA,$8F,$ED,$DF,$5E,$FC,$DB,$FB,$6C,$7E,$07,$F7,$3F,$C6,$2F,$A2
	.dc.b	$CA,$F1,$FF,$1D,$97,$F2,$AD,$FF,$C9,$BE,$1B,$E5,$3F,$95,$78,$9F
	.dc.b	$EE,$DF,$13,$F9,$4D,$FF,$7F,$BF,$F4,$1F,$C5,$7F,$A3,$BF,$66,$FF
	.dc.b	$03,$E7,$F8,$8F,$A8,$FE,$FB,$F4,$DF,$DE,$5F,$C2,$5F,$EA,$F2,$FD
	.dc.b	$0D,$7E,$87,$0F,$ED,$71,$88,$FE,$1D,$F0,$DF,$C8,$BE,$51,$D3,$FE
	.dc.b	$70,$F8,$F7,$D0,$3F,$51,$7E,$CE,$FC,$43,$7F,$9D,$7F,$95,$F8,$BF
	.dc.b	$DD,$DF,$8F,$7F,$B8,$7E,$F3,$FC,$6F,$F7,$9F,$E3,$AF,$9A,$FF,$72
	.dc.b	$79,$5F,$E3,$68,$FF,$1C,$FE,$37,$EC,$E2,$F9,$5F,$2B,$FC,$9B,$E5
	.dc.b	$7F,$8E,$E7,$F0,$E7,$E7,$F7,$7F,$24,$F9,$4F,$F7,$B7,$5F,$E8,$3A
	.dc.b	$BF,$85,$7C,$8F,$E1,$B9,$FD,$7F,$45,$89,$0F,$CE,$BF,$8C,$FE,$AD
	.dc.b	$FC,$AB,$F6,$3F,$E3,$1F,$5D,$F7,$FB,$7F,$89,$E3,$CB,$0F,$F4,$9C
	.dc.b	$7F,$AA,$DE,$71,$F2,$3E,$B7,$F2,$31,$FC,$7C,$FF,$5C,$E5,$FE,$D1
	.dc.b	$E0,$FF,$56,$C7,$FA,$71,$EB,$5F,$E4,$33,$E8,$F9,$DF,$8A,$E1,$FE
	.dc.b	$2E,$7E,$DE,$7F,$72,$FF,$B3,$78,$FF,$A5,$FE,$19,$FE,$39,$FD,$97
	.dc.b	$FD,$4F,$97,$F8,$E6,$DE,$87,$FA,$AB,$E8,$9F,$1D,$FD,$23,$E7,$7E
	.dc.b	$1F,$D9,$BF,$08,$7D,$D7,$F2,$FE,$1F,$E5,$FF,$62,$FC,$57,$F9,$7F
	.dc.b	$1F,$F2,$FF,$B2,$FF,$0E,$F2,$7E,$2F,$E4,$7F,$B9,$70,$7D,$7E,$5F
	.dc.b	$D2,$FE,$55,$FD,$7F,$99,$F4,$5F,$7E,$7F,$99,$F8,$1F,$D7,$BA,$BF
	.dc.b	$23,$BF,$D6,$FB,$A7,$AF,$FE,$69,$F7,$1F,$46,$FC,$A1,$FA,$37,$ED
	.dc.b	$77,$77,$7F,$28,$F9,$3F,$EE,$99,$FF,$04,$97,$EE,$BF,$B2,$F7,$7E
	.dc.b	$BC,$7C,$CE,$0F,$9E,$7E,$ED,$FE,$49,$CF,$FC,$92,$FF,$D5,$9C,$DF
	.dc.b	$EF,$A7,$FD,$5D,$F0,$95,$ED,$3F,$AA,$3F,$57,$F0,$9F,$7E,$BF,$29
	.dc.b	$1F,$AE,$F0,$BE,$FE,$1F,$1F,$1F,$E9,$9F,$07,$FE,$AF,$FC,$FB,$E3
	.dc.b	$79,$DF,$E7,$DA,$F3,$76,$3E,$0F,$FA,$E7,$C1,$FE,$1E,$BF,$89,$FC
	.dc.b	$9F,$CD,$8F,$E1,$17,$D7,$2F,$E2,$9F,$05,$FE,$85,$F4,$BF,$D0,$1F
	.dc.b	$27,$FE,$EE,$FC,$AB,$F8,$0C,$7F,$B7,$B8,$DF,$3A,$F3,$BF,$AA,$E5
	.dc.b	$F1,$F2,$FF,$3A,$F1,$7F,$BD,$BF,$03,$FF,$4F,$7E,$5B,$FD,$6B,$17
	.dc.b	$1F,$F8,$F7,$F1,$EF,$E8,$91,$F7,$7C,$2F,$B9,$E0,$FD,$BF,$03,$ED
	.dc.b	$7C,$3F,$E2,$FE,$7B,$FB,$46,$DF,$CA,$70,$FB,$0F,$F9,$F3,$AF,$E7
	.dc.b	$1F,$E7,$AE,$2F,$F4,$D7,$E6,$5F,$D1,$BB,$7F,$C4,$37,$FE,$B0,$E3
	.dc.b	$FD,$CC,$7D,$2B,$F6,$4F,$FB,$9F,$E3,$FF,$84,$7A,$FF,$E7,$5F,$07
	.dc.b	$58,$7E,$F2,$BF,$D4,$D7,$FB,$A7,$FA,$37,$37,$F4,$2F,$89,$FD,$C3
	.dc.b	$FD,$13,$DC,$FF,$65,$4F,$FB,$02,$BF,$96,$FD,$83,$CC,$FC,$FC,$E6
	.dc.b	$53,$F5,$23,$F8,$BF,$E3,$BF,$CD,$3E,$43,$C3,$E3,$FD,$3F,$A0,$FF
	.dc.b	$5C,$7E,$B7,$FC,$72,$FF,$DC,$27,$F8,$E0,$FE,$3F,$F2,$EF,$DB,$DF
	.dc.b	$9C,$BF,$33,$7D,$1B,$F7,$F7,$EC,$CF,$D2,$7F,$C6,$DE,$1F,$C6,$EE
	.dc.b	$FF,$31,$E9,$FE,$D2,$E6,$FF,$74,$73,$7B,$BF,$31,$7E,$45,$FD,$F6
	.dc.b	$BF,$61,$FA,$CF,$F4,$FB,$7E,$E1,$FD,$A3,$AB,$F2,$91,$F1,$EB,$D4
	.dc.b	$7E,$75,$FF,$7A,$7C,$4F,$F7,$CF,$53,$FD,$ED,$0F,$D7,$70,$FF,$8E
	.dc.b	$CF,$F1,$F6,$BF,$E1,$5B,$FE,$7F,$F8,$9F,$F3,$1F,$3F,$FC,$7D,$F9
	.dc.b	$EB,$F6,$77,$EF,$8F,$9A,$7E,$AF,$FA,$0E,$5F,$B3,$F1,$1E,$CF,$99
	.dc.b	$FE,$0F,$C9,$EB,$FC,$37,$FC,$03,$B7,$A3,$8B,$CB,$FA,$2F,$93,$8B
	.dc.b	$FC,$0A,$7F,$94,$76,$C3,$FE,$40,$9F,$F2,$A7,$2F,$E0,$BA,$5F,$AE
	.dc.b	$7F,$88,$7F,$15,$7E,$D6,$E6,$FF,$18,$FC,$0F,$E2,$7C,$CF,$C0,$67
	.dc.b	$FE,$0F,$F9,$F7,$E8,$1F,$2F,$4F,$F1,$A7,$07,$F6,$0A,$7E,$33,$F2
	.dc.b	$A3,$E7,$FE,$37,$F0,$7A,$66,$DF,$E5,$D9,$CC,$64,$7E,$75,$BE,$E7
	.dc.b	$C2,$7D,$DC,$BF,$AB,$77,$FE,$4F,$F1,$27,$E7,$AF,$C7,$9F,$B5,$FF
	.dc.b	$9C,$FE,$7D,$F4,$4F,$0B,$F4,$9F,$05,$CF,$3F,$F3,$BF,$27,$E9,$9F
	.dc.b	$77,$A5,$ED,$FB,$8F,$83,$F3,$7E,$BE,$4F,$D1,$73,$79,$F9,$7F,$D2
	.dc.b	$9E,$07,$F9,$CE,$7D,$71,$EF,$E6,$75,$7E,$07,$9F,$D7,$5F,$DF,$3A
	.dc.b	$FF,$86,$9F,$DF,$63,$FC,$CD,$C5,$FD,$AB,$E0,$5F,$3B,$FE,$28,$FD
	.dc.b	$3F,$FB,$56,$1E,$A3,$96,$1C,$BE,$0B,$11,$FD,$5B,$F3,$47,$ED,$9F
	.dc.b	$77,$E2,$10,$FD,$15,$F9,$33,$67,$E2,$BF,$BF,$F2,$BF,$C1,$9F,$05
	.dc.b	$FC,$07,$EE,$BF,$C1,$DF,$59,$FC,$0F,$CF,$FE,$FF,$F1,$7E,$6E,$17
	.dc.b	$E7,$3F,$DF,$FE,$9F,$F8,$4F,$17,$F8,$93,$F0,$BE,$2F,$EF,$7C,$1F
	.dc.b	$CC,$9F,$EA,$8D,$FF,$C1,$71,$FE,$2A,$F7,$FE,$4F,$AF,$FA,$48,$E1
	.dc.b	$5F,$EA,$FE,$FF,$DF,$75,$FD,$BF,$47,$F3,$15,$FE,$32,$FD,$21,$FA
	.dc.b	$B3,$F3,$57,$E7,$4F,$92,$FE,$93,$FB,$AF,$E9,$30,$FD,$76,$9E,$43
	.dc.b	$E4,$B8,$7C,$FF,$53,$FE,$0D,$7F,$CE,$BF,$98,$FF,$32,$F8,$7F,$ED
	.dc.b	$1E,$5B,$F5,$EF,$E0,$1E,$DF,$FA,$6F,$D2,$FD,$FF,$97,$FC,$67,$3F
	.dc.b	$E8,$07,$F1,$3C,$BE,$1F,$E6,$7F,$BA,$7F,$D0,$3F,$A5,$5F,$A5,$FF
	.dc.b	$41,$F5,$7E,$2A,$DF,$06,$F4,$9F,$86,$BE,$53,$FD,$A9,$F1,$7F,$E5
	.dc.b	$3B,$FE,$CA,$7F,$35,$FB,$0F,$F0,$EF,$E6,$6F,$CB,$7F,$9C,$FE,$C8
	.dc.b	$FC,$CF,$E5,$DF,$FA,$4B,$3F,$E0,$3C,$1F,$ED,$FF,$3B,$FC,$BF,$EE
	.dc.b	$FF,$02,$AF,$F7,$17,$EC,$7F,$A7,$7F,$CC,$77,$FA,$7F,$87,$FF,$06
	.dc.b	$7D,$6F,$E3,$FD,$0F,$E9,$9D,$1F,$AB,$F7,$BF,$93,$79,$5E,$68,$C4
	.dc.b	$7F,$AE,$F3,$FD,$B6,$BE,$FF,$80,$FD,$B6,$9F,$C4,$7F,$35,$F0,$3F
	.dc.b	$0E,$FF,$86,$63,$FC,$2F,$2F,$EE,$9E,$8F,$F6,$6E,$AF,$C7,$DF,$FB
	.dc.b	$47,$F3,$1F,$F3,$CF,$B7,$FE,$66,$FC,$AB,$FC,$67,$3F,$AE,$BF,$F0
	.dc.b	$1F,$D1,$7C,$4F,$CB,$DF,$8F,$FF,$BE,$FC,$AF,$F2,$5F,$E6,$AF,$CA
	.dc.b	$1F,$A0,$FE,$B7,$D0,$FE,$D5,$F0,$9F,$80,$E7,$7C,$57,$F6,$DF,$A4
	.dc.b	$FE,$D5,$F4,$9F,$DA,$7E,$8F,$FB,$47,$D1,$7F,$67,$E2,$FC,$39,$FD
	.dc.b	$B3,$F0,$3F,$F4,$7F,$3F,$AB,$F2,$EF,$8F,$D5,$FD,$9F,$F5,$2F,$C7
	.dc.b	$5F,$FA,$0F,$E7,$BF,$B3,$7A,$9F,$D9,$7A,$7F,$BE,$61,$C3,$81,$EF
	.dc.b	$FF,$01,$F9,$DF,$E8,$DF,$3B,$FD,$17,$F1,$B7,$C1,$7E,$57,$C6,$0F
	.dc.b	$A0,$FF,$09,$F9,$2E,$0F,$0F,$07,$FA,$63,$E1,$3F,$7F,$EC,$3F,$52
	.dc.b	$FE,$AB,$C3,$FD,$3B,$2F,$EC,$DF,$BB,$7F,$77,$C3,$F9,$5F,$95,$FD
	.dc.b	$26,$FF,$97,$47,$F3,$28,$FB,$BC,$7F,$DB,$78,$FF,$C3,$FF,$A6,$7F
	.dc.b	$70,$F7,$DF,$1F,$FC,$CB,$F1,$57,$D9,$BD,$8F,$CE,$ED,$EA,$CF,$FC
	.dc.b	$34,$FE,$77,$4F,$F7,$FF,$AB,$FC,$FF,$B7,$F9,$DD,$7E,$66,$7F,$D7
	.dc.b	$FF,$8F,$BD,$1F,$C4,$F1,$FC,$AF,$43,$F4,$AF,$F8,$A7,$F6,$DF,$07
	.dc.b	$E9,$DF,$DA,$BC,$87,$E2,$2F,$A4,$7E,$71,$FB,$6F,$53,$F1,$BF,$94
	.dc.b	$3D,$4F,$C6,$FA,$7F,$8C,$F2,$DF,$93,$FE,$33,$8B,$F0,$A7,$F0,$CF
	.dc.b	$7F,$F0,$BF,$51,$FE,$39,$F0,$DF,$AC,$FD,$4A,$FD,$EF,$8E,$FC,$C1
	.dc.b	$FA,$63,$F2,$97,$F0,$0B,$F0,$77,$E7,$8F,$DC,$7F,$05,$EA,$FF,$89
	.dc.b	$36,$FF,$23,$67,$DB,$C1,$F1,$F9,$DF,$CA,$FF,$17,$D6,$FF,$EE,$6F
	.dc.b	$91,$FF,$17,$F8,$DF,$C3,$9B,$E6,$FE,$4A,$7F,$DD,$78,$F9,$78,$7C
	.dc.b	$9C,$1E,$3D,$FC,$57,$FC,$13,$CE,$6B,$FC,$13,$1F,$CA,$FA,$7F,$C1
	.dc.b	$3F,$48,$FE,$17,$E4,$FD,$C3,$FB,$D7,$C8,$FE,$32,$BF,$8B,$C7,$F7
	.dc.b	$DF,$DB,$FF,$19,$E4,$7E,$57,$27,$C6,$7E,$56,$FF,$25,$7F,$E1,$9E
	.dc.b	$1E,$51,$FC,$46,$3F,$C3,$9F,$35,$F9,$5F,$C8,$7F,$CD,$5F,$83,$BF
	.dc.b	$21,$FE,$FB,$FA,$4B,$F1,$97,$E4,$0F,$D6,$DF,$97,$7F,$38,$FD,$1D
	.dc.b	$6F,$DA,$9F,$AF,$FF,$0A,$F8,$B7,$E3,$8F,$63,$F0,$7C,$4F,$DA,$BF
	.dc.b	$9B,$FE,$A5,$FE,$CB,$EE,$FF,$42,$FE,$69,$E3,$FA,$DF,$F3,$D7,$E2
	.dc.b	$4F,$DC,$3F,$AD,$78,$5F,$E8,$BF,$43,$FD,$05,$DF,$FE,$D1,$FC,$88
	.dc.b	$FF,$A7,$B5,$FF,$4D,$7E,$25,$FE,$6E,$7F,$DF,$E5,$3F,$D0,$71,$FE
	.dc.b	$AC,$F9,$1F,$F5,$07,$CE,$FF,$88,$BB,$75,$F9,$A7,$F8,$93,$EE,$7F
	.dc.b	$D4,$9F,$77,$CF,$F2,$7F,$4E,$5F,$CC,$F2,$F0,$71,$FF,$A7,$B9,$1F
	.dc.b	$9A,$BF,$20,$7C,$9B,$F5,$26,$9F,$E3,$4F,$D8,$BD,$99,$FF,$8A,$BE
	.dc.b	$83,$E3,$F1,$FA,$4F,$2E,$BF,$EC,$28,$F2,$BF,$EC,$4F,$93,$F2,$D7
	.dc.b	$FA,$33,$E7,$7F,$B1,$FD,$07,$DE,$FD,$0D,$E3,$3F,$4F,$FF,$40,$FE
	.dc.b	$A5,$FE,$D8,$F5,$FE,$37,$BF,$FE,$AC,$CB,$FD,$2B,$F6,$DF,$63,$EC
	.dc.b	$7F,$45,$1F,$EB,$1F,$D5,$BF,$9A,$ED,$FB,$EB,$7E,$7B,$F3,$70,$FA
	.dc.b	$15,$FE,$C0,$BF,$F0,$6F,$BB,$FE,$0E,$FC,$83,$FD,$42,$D8,$FF,$5E
	.dc.b	$61,$C9,$FB,$17,$F6,$D5,$F3,$7F,$D8,$57,$FE,$AA,$AF,$44,$7F,$B1
	.dc.b	$6B,$83,$F7,$F2,$FE,$97,$5F,$EB,$0B,$F4,$CF,$E1,$7E,$E7,$FB,$0A
	.dc.b	$1E,$0B,$87,$E4,$F0,$7C,$8F,$C9,$3F,$A8,$DB,$BB,$FD,$D3,$87,$18
	.dc.b	$FC,$17,$77,$E0,$7B,$7F,$01,$D9,$F7,$E7,$F9,$0D,$7D,$F3,$FD,$87
	.dc.b	$AF,$E5,$7A,$FF,$00,$FE,$5F,$C3,$FF,$36,$3F,$9B,$F0,$FF,$CE,$FC
	.dc.b	$DE,$78,$FC,$F7,$87,$F9,$FE,$CF,$16,$DF,$F0,$07,$2F,$F7,$46,$7F
	.dc.b	$AE,$F5,$DE,$5F,$EB,$BC,$AF,$D6,$CF,$92,$BF,$57,$E0,$7E,$B3,$C8
	.dc.b	$FD,$6E,$9F,$42,$7F,$28,$7F,$B6,$BF,$03,$F9,$EF,$EB,$2B,$8F,$93
	.dc.b	$8E,$1F,$D6,$5F,$A6,$7A,$9F,$A2,$AF,$AF,$BF,$EB,$56,$5F,$90,$F0
	.dc.b	$BF,$87,$FA,$7F,$C3,$BC,$4F,$F7,$DF,$A3,$FC,$37,$2F,$E9,$FE,$87
	.dc.b	$D1,$F6,$7F,$8E,$F1,$FF,$1C,$F6,$7F,$8D,$FB,$1F,$C6,$B4,$FE,$6D
	.dc.b	$EE,$FF,$BC,$F1,$FE,$FB,$EE,$7F,$B9,$7D,$D7,$C2,$7E,$47,$E7,$7F
	.dc.b	$C8,$9E,$FF,$FB,$A7,$F0,$6F,$F1,$BD,$7F,$48,$CB,$F8,$8F,$27,$F4
	.dc.b	$BF,$87,$FF,$7C,$C7,$EA,$31,$FE,$91,$F1,$BF,$D6,$BE,$37,$FA,$CF
	.dc.b	$C6,$FB,$BA,$BF,$9E,$7B,$9F,$E1,$AC,$7F,$CC,$D8,$FF,$96,$DF,$F0
	.dc.b	$F9,$FE,$83,$E0,$FC,$B6,$7F,$97,$F8,$8F,$E9,$71,$EA,$E1,$FF,$38
	.dc.b	$6B,$E9,$7F,$A5,$0D,$FF,$C9,$F1,$FD,$96,$FF,$93,$57,$FA,$37,$0F
	.dc.b	$F7,$DF,$D3,$7E,$C7,$E3,$7F,$B4,$47,$1F,$53,$FB,$D7,$BF,$FD,$E7
	.dc.b	$27,$D5,$7F,$8A,$6B,$D5,$C9,$EB,$DF,$E2,$E9,$F3,$3E,$43,$FB,$EF
	.dc.b	$2F,$C3,$F6,$3F,$BD,$6F,$E1,$C7,$FC,$77,$F2,$1F,$E3,$CD,$3F,$C1
	.dc.b	$B8,$FE,$D7,$DD,$FE,$10,$FF,$83,$4F,$C3,$F8,$FF,$13,$E6,$3F,$C2
	.dc.b	$FF,$2F,$FE,$16,$FA,$EF,$81,$F6,$5F,$23,$E8,$BF,$C4,$D8,$3C,$CF
	.dc.b	$F3,$73,$F4,$CF,$E0,$5F,$43,$FE,$26,$D5,$F4,$1F,$E2,$6F,$9F,$FF
	.dc.b	$12,$F2,$FF,$8F,$BF,$19,$FF,$57,$7A,$FF,$E3,$DF,$A5,$FF,$1E,$4F
	.dc.b	$FA,$1B,$C0,$FF,$1D,$F2,$FF,$A8,$79,$BF,$1F,$E4,$7F,$A4,$9F,$C8
	.dc.b	$F8,$7F,$E3,$CE,$8F,$F5,$65,$7F,$B0,$4F,$E1,$FE,$7F,$FB,$37,$1E
	.dc.b	$9E,$E7,$E1,$FE,$73,$F8,$87,$E2,$1F,$84,$B7,$DF,$F1,$CF,$F0,$9F
	.dc.b	$57,$F9,$C7,$BF,$CC,$FE,$6F,$C6,$FF,$13,$FD,$67,$F8,$CE,$7E,$96
	.dc.b	$1F,$E6,$CF,$C3,$3F,$04,$FF,$B4,$FF,$00,$AF,$E5,$FE,$8B,$FC,$D4
	.dc.b	$FE,$17,$B3,$F9,$56,$5F,$E8,$2E,$2F,$F4,$EF,$E0,$DF,$E7,$3E,$4F
	.dc.b	$EE,$51,$F8,$8A,$FF,$75,$DF,$FA,$F0,$FA,$4A,$FF,$D1,$7C,$7F,$D4
	.dc.b	$B8,$9F,$65,$FE,$84,$FC,$5D,$F6,$5F,$E8,$4E,$0F,$4E,$4F,$C6,$3C
	.dc.b	$EF,$FA,$1B,$7F,$D7,$6B,$FE,$F2,$D7,$F1,$87,$F8,$14,$FE,$23,$D1
	.dc.b	$FE,$A3,$FB,$C7,$DB,$FF,$A8,$FF,$1E,$7D,$C3,$F1,$EF,$E4,$38,$7E
	.dc.b	$D3,$E8,$DF,$1B,$FD,$0F,$DB,$FC,$A7,$DE,$7F,$AD,$7F,$24,$7D,$EF
	.dc.b	$FA,$D7,$D0,$FC,$AE,$3F,$E0,$BB,$FF,$0D,$FD,$F7,$CB,$8F,$F6,$21
	.dc.b	$F9,$78,$FF,$3E,$3F,$A0,$38,$28,$E7,$1F,$E8,$6F,$1F,$FD,$11,$D7
	.dc.b	$FD,$EB,$F0,$0F,$97,$7F,$E9,$FF,$84,$FC,$AF,$6F,$E2,$B0,$FF,$6D
	.dc.b	$74,$7F,$0F,$7F,$C9,$B8,$8F,$F2,$8C,$7E,$2E,$33,$1F,$E6,$03,$F9
	.dc.b	$37,$E3,$5F,$C5,$BD,$C6,$0F,$C1,$DF,$61,$FE,$B8,$FC,$1F,$F8,$FF
	.dc.b	$B7,$FC,$73,$F2,$07,$E1,$1F,$DC,$BE,$89,$F8,$47,$D2,$FC,$01,$F8
	.dc.b	$47,$E1,$B8,$7F,$D8,$3A,$FE,$27,$8B,$D5,$C5,$FE,$C4,$FC,$23,$F4
	.dc.b	$4F,$9B,$F1,$7F,$D7,$74,$7F,$17,$C3,$F9,$80,$FC,$2F,$E8,$7E,$16
	.dc.b	$F8,$CF,$D5,$FE,$0F,$F0,$7E,$07,$FD,$77,$F5,$7F,$57,$E3,$DD,$DF
	.dc.b	$A9,$D7,$F9,$AE,$1F,$D0,$BC,$C7,$C3,$F7,$D7,$F9,$F7,$D6,$FD,$DF
	.dc.b	$5F,$8F,$AF,$FA,$4F,$8F,$D6,$3F,$CA,$7B,$7F,$80,$78,$9F,$D0,$3E
	.dc.b	$89,$F8,$73,$F0,$E7,$E1,$8F,$C3,$9F,$58,$F9,$BF,$D3,$FB,$7F,$A4
	.dc.b	$F9,$1F,$D1,$F1,$3D,$BF,$73,$FD,$17,$F1,$2F,$E9,$BE,$17,$E8,$FA
	.dc.b	$3F,$A4,$7C,$67,$F5,$0F,$AC,$C7,$57,$B2,$7F,$C5,$0B,$F1,$77,$E2
	.dc.b	$6F,$C5,$FF,$CE,$BF,$8A,$3E,$F1,$F5,$9F,$D4,$FD,$3F,$EB,$7F,$8A
	.dc.b	$74,$F5,$FF,$CF,$BD,$9F,$E0,$63,$FC,$FB,$F3,$9F,$9C,$D3,$B7,$77
	.dc.b	$59,$7E,$2F,$D7,$83,$F7,$27,$E9,$2E,$AF,$CB,$1F,$8C,$3E,$41,$F1
	.dc.b	$BF,$D0,$E1,$F8,$C3,$EC,$7F,$65,$F3,$8F,$C7,$1F,$8E,$3F,$1C,$77
	.dc.b	$FC,$1F,$C7,$1E,$9F,$E5,$7C,$2F,$C2,$FC,$8F,$F8,$3F,$87,$F4,$FF
	.dc.b	$51,$F8,$DF,$AA,$FE,$3F,$E2,$7F,$1C,$FC,$F3,$FD,$7B,$F8,$E7,$93
	.dc.b	$E6,$3F,$C9,$FD,$5F,$CF,$73,$FC,$7F,$B5,$FD,$07,$3F,$F1,$3F,$D0
	.dc.b	$7F,$41,$FC,$7B,$F3,$F1,$FE,$CA,$F8,$DF,$DC,$0F,$82,$BE,$53,$CC
	.dc.b	$FF,$05,$F6,$FD,$CF,$9F,$FF,$06,$FE,$7E,$FC,$87,$F9,$57,$D3,$FF
	.dc.b	$01,$FC,$85,$F9,$57,$BB,$26,$1F,$E1,$4F,$84,$7C,$37,$E8,$F7,$FF
	.dc.b	$36,$7B,$FF,$D6,$EF,$FC,$25,$FA,$3F,$E8,$FC,$0F,$93,$D7,$FA,$3C
	.dc.b	$3F,$C1,$9F,$03,$FA,$7F,$C9,$7F,$C1,$5F,$15,$BB,$5F,$E6,$9F,$A3
	.dc.b	$F9,$2D,$5F,$29,$F9,$EF,$88,$FE,$7F,$EA,$7F,$8A,$3F,$27,$FF,$63
	.dc.b	$FE,$4E,$FC,$7B,$FD,$93,$D6,$F8,$27,$E0,$FF,$95,$E5,$FE,$8A,$7F
	.dc.b	$B2,$7E,$6F,$FD,$47,$F6,$2F,$33,$FA,$2F,$CD,$3E,$89,$F4,$BF,$D1
	.dc.b	$BE,$93,$FA,$2E,$3F,$E8,$DF,$47,$B3,$FE,$33,$F7,$BD,$9F,$89,$FF
	.dc.b	$8D,$7E,$A3,$DD,$D3,$FD,$07,$CF,$FD,$11,$FD,$13,$7F,$F1,$5B,$83
	.dc.b	$FA,$07,$81,$FE,$9A,$C7,$F8,$DF,$57,$F9,$6B,$ED,$1F,$17,$FE,$B6
	.dc.b	$AF,$F5,$DF,$2F,$FB,$97,$CE,$7D,$2B,$F2,$EF,$16,$5F,$ED,$DF,$CF
	.dc.b	$5F,$9F,$BE,$DB,$FC,$95,$F9,$5F,$ED,$B9,$3F,$BD,$FC,$E7,$FA,$8B
	.dc.b	$0F,$EA,$87,$A3,$4F,$F4,$8F,$C8,$FE,$F3,$8B,$FD,$6F,$3F,$E7,$EF
	.dc.b	$96,$E1,$5F,$E5,$C8,$FA,$93,$FE,$86,$FC,$A3,$FC,$BF,$C3,$ED,$E4
	.dc.b	$F6,$FE,$7A,$24,$7E,$4A,$FF,$D7,$5F,$97,$FF,$6E,$FC,$BF,$ED,$C7
	.dc.b	$F2,$4F,$C7,$7F,$6B,$D3,$FE,$DC,$F9,$9F,$83,$F5,$EF,$CC,$DF,$98
	.dc.b	$3F,$10,$FF,$61,$F9,$9F,$5F,$2F,$F8,$01,$EC,$B8,$9C,$FF,$E6,$DF
	.dc.b	$4F,$FA,$67,$C4,$3F,$32,$7E,$7F,$FC,$D3,$93,$FB,$B6,$5F,$EC,$FE
	.dc.b	$6F,$EA,$3F,$0E,$FD,$3F,$F8,$6F,$C2,$7D,$2A,$FB,$79,$7F,$36,$FC
	.dc.b	$D3,$18,$70,$3F,$88,$FB,$1E,$1A,$E9,$1F,$E8,$B9,$F8,$DD,$6F,$D0
	.dc.b	$35,$5F,$86,$CD,$F9,$14,$F1,$3F,$D0,$DE,$D7,$FA,$3B,$F0,$EF,$74
	.dc.b	$7F,$87,$2F,$F8,$0F,$85,$FA,$6E,$1F,$E5,$91,$FE,$E9,$FC,$50,$47
	.dc.b	$EE,$1F,$E9,$1F,$CD,$BF,$D3,$3E,$6F,$E4,$3C,$CF,$F2,$1F,$A3,$FD
	.dc.b	$CF,$8F,$D1,$8F,$F4,$DC,$3F,$42,$7F,$B6,$BA,$BF,$DB,$8F,$FB,$BF
	.dc.b	$C9,$FC,$67,$E4,$1F,$EF,$BF,$2B,$33,$FA,$E3,$FD,$F9,$EE,$FF,$BA
	.dc.b	$1F,$EA,$F7,$FE,$DB,$7F,$D5,$7F,$A2,$7F,$57,$FC,$E7,$E8,$78,$3F
	.dc.b	$ED,$6A,$7D,$FF,$F9,$B3,$F4,$3F,$C1,$61,$F5,$3F,$43,$0F,$CC,$BE
	.dc.b	$6F,$7F,$FA,$AB,$F2,$E7,$E7,$7F,$E1,$CE,$2F,$C0,$E1,$FB,$6E,$37
	.dc.b	$E3,$4F,$CE,$5F,$9E,$7F,$98,$BD,$1F,$CA,$FA,$1F,$94,$F3,$FF,$27
	.dc.b	$97,$E4,$32,$1F,$D9,$F8,$BF,$DE,$31,$FA,$7D,$BF,$C1,$E7,$E1,$9C
	.dc.b	$3F,$D6,$3C,$7F,$2F,$E4,$FF,$E5,$89,$FB,$77,$F2,$7C,$1F,$B3,$1F
	.dc.b	$AD,$87,$C6,$7F,$7F,$EC,$F3,$C3,$F3,$5F,$EF,$DF,$61,$F1,$BE,$F7
	.dc.b	$FA,$8F,$CE,$3F,$40,$7B,$AF,$AE,$7E,$81,$BA,$FF,$27,$7D,$F6,$61
	.dc.b	$F8,$3B,$F3,$BF,$9B,$C9,$E9,$ED,$72,$3B,$BF,$8F,$7B,$5F,$ED,$3F
	.dc.b	$4F,$E6,$76,$FC,$EF,$CE,$7F,$DC,$5E,$8F,$F1,$3E,$CF,$6F,$C2,$7F
	.dc.b	$15,$F2,$3F,$11,$D3,$FE,$6A,$F8,$B7,$E8,$EF,$D0,$5D,$AF,$AC,$7E
	.dc.b	$8C,$FD,$09,$F0,$9F,$C8,$B3,$FE,$0C,$7F,$07,$F1,$3F,$8F,$7C,$3F
	.dc.b	$F8,$9B,$5F,$B3,$3F,$E1,$E7,$F3,$46,$1E,$33,$FC,$94,$FE,$4D,$E9
	.dc.b	$FF,$25,$9F,$E4,$BE,$B7,$F9,$6F,$DF,$FD,$F9,$FC,$1B,$AF,$FC,$2A
	.dc.b	$7E,$33,$CE,$F8,$79,$7F,$BD,$3E,$45,$FA,$2B,$F4,$57,$E8,$AF,$CA
	.dc.b	$FF,$8F,$7E,$56,$FD,$4B,$F4,$3C,$3E,$C3,$FC,$02,$7F,$23,$EB,$FD
	.dc.b	$2E,$BF,$E6,$CF,$8A,$FE,$39,$F6,$DF,$CB,$7D,$2F,$CC,$71,$FF,$56
	.dc.b	$87,$C9,$7F,$27,$E7,$F4,$60,$F8,$AF,$F2,$AA,$F5,$73,$E2,$FF,$9E
	.dc.b	$3D,$AF,$D9,$F5,$FF,$49,$F8,$4F,$E1,$9E,$CF,$EE,$B8,$7E,$8F,$CA
	.dc.b	$FF,$5A,$FD,$3D,$FA,$6F,$E2,$B4,$E9,$DD,$FC,$9B,$E3,$71,$AF,$F2
	.dc.b	$4F,$8D,$FE,$49,$CB,$FC,$9B,$97,$F5,$FB,$FF,$AF,$DF,$EB,$11,$EC
	.dc.b	$C7,$E7,$F1,$3E,$6B,$FB,$8F,$8E,$FA,$9F,$8B,$5F,$ED,$38,$C7,$E6
	.dc.b	$9F,$CA,$FF,$2B,$FD,$DE,$5F,$CA,$BE,$3E,$3D,$BB,$7F,$8C,$7D,$0F
	.dc.b	$F7,$7E,$8F,$E5,$5B,$7F,$80,$79,$FF,$1F,$4E,$67,$3F,$F7,$AF,$47
	.dc.b	$E6,$F5,$7F,$A8,$3A,$FF,$D8,$F8,$7D,$BD,$3D,$7F,$25,$E2,$FA,$6F
	.dc.b	$F5,$C7,$E5,$3F,$EE,$E5,$F9,$EB,$ED,$9F,$A0,$7E,$7F,$1F,$CF,$E3
	.dc.b	$F9,$E7,$D7,$F9,$BB,$E0,$9F,$AE,$7F,$A5,$FE,$AD,$D2,$FB,$D7,$E1
	.dc.b	$CF,$4F,$F4,$D9,$FE,$AB,$B7,$FC,$65,$F3,$1F,$B7,$8F,$D2,$1F,$23
	.dc.b	$D7,$FC,$A7,$17,$C7,$8F,$EC,$7E,$07,$E2,$FE,$53,$F9,$B9,$FC,$DF
	.dc.b	$B7,$F2,$7C,$5F,$20,$FE,$53,$DB,$FD,$16,$DF,$31,$FE,$9B,$FC,$C5
	.dc.b	$FA,$57,$E3,$6F,$FC,$03,$F3,$3F,$D1,$7F,$47,$F9,$BA,$7F,$AF,$B8
	.dc.b	$5F,$86,$FF,$9B,$3F,$03,$FF,$53,$7E,$8E,$FD,$0B,$FC,$D7,$F6,$1F
	.dc.b	$E7,$3E,$1F,$C5,$73,$FE,$13,$CD,$FC,$4F,$9A,$97,$E1,$1F,$86,$F3
	.dc.b	$7F,$0C,$7E,$1F,$A3,$FD,$67,$F1,$BF,$8A,$E0,$FE,$7B,$F1,$D3,$B7
	.dc.b	$FA,$AF,$CF,$7F,$03,$3F,$82,$7C,$77,$5F,$D3,$3F,$91,$74,$7F,$87
	.dc.b	$6B,$FD,$7B,$F1,$FF,$94,$8F,$E4,$3D,$BF,$E4,$03,$FA,$8F,$E4,$6F
	.dc.b	$83,$FB,$77,$FE,$2A,$FD,$85,$F8,$43,$F4,$FF,$57,$CC,$F3,$AF,$F0
	.dc.b	$2E,$3F,$D6,$BB,$BE,$9C,$62,$3F,$05,$9F,$E9,$3C,$0F,$E6,$7D,$DF
	.dc.b	$EB,$2F,$D3,$BF,$98,$FE,$77,$F8,$DC,$7F,$DC,$58,$7F,$2C,$C7,$97
	.dc.b	$BD,$FC,$A3,$9E,$CF,$FB,$74,$FF,$40,$F5,$FE,$E7,$E8,$FF,$0B,$E4
	.dc.b	$FE,$1B,$E9,$BF,$8D,$FC,$87,$F2,$CF,$85,$FF,$15,$DF,$F9,$4F,$F4
	.dc.b	$FF,$EF,$3D,$9F,$2F,$E6,$BF,$92,$FC,$97,$F9,$E3,$CF,$FE,$2B,$F9
	.dc.b	$D3,$F4,$8F,$85,$E8,$FE,$2F,$D0,$FF,$07,$FD,$53,$F3,$B7,$E6,$7F
	.dc.b	$E1,$0F,$CE,$12,$F5,$3F,$C8,$35,$FE,$7E,$F2,$BF,$92,$70,$FF,$99
	.dc.b	$7F,$2F,$C7,$C2,$FF,$24,$FD,$47,$FC,$5D,$F1,$5F,$C9,$2B,$F8,$37
	.dc.b	$EA,$3F,$C8,$FE,$2B,$F9,$27,$C9,$7E,$37,$D4,$FD,$07,$7F,$F8,$33
	.dc.b	$F0,$6F,$A9,$F8,$EF,$F1,$4F,$CD,$DF,$99,$3F,$00,$7A,$2F,$B1,$7E
	.dc.b	$5F,$F7,$F2,$FF,$2B,$FB,$9F,$BE,$F7,$7F,$37,$F8,$E7,$F4,$6F,$D4
	.dc.b	$BE,$67,$8F,$FE,$7A,$F0,$FF,$31,$F7,$1F,$CC,$7E,$6B,$F9,$97,$CC
	.dc.b	$FF,$3E,$FB,$B7,$57,$B3,$F0,$7F,$F4,$C7,$5F,$F2,$1F,$9A,$FE,$3F
	.dc.b	$F8,$3F,$F2,$1E,$CF,$E4,$13,$FE,$89,$FB,$AF,$91,$F3,$7F,$C8,$3E
	.dc.b	$E3,$F1,$D3,$FE,$4F,$F8,$87,$E7,$BF,$B5,$F6,$BF,$D4,$3F,$9A,$7F
	.dc.b	$7D,$F2,$FF,$33,$D5,$FD,$E3,$1F,$CB,$7B,$9F,$C2,$9B,$F2,$7F,$A7
	.dc.b	$7F,$1F,$3F,$E5,$2F,$BA,$AF,$E7,$7E,$0F,$FD,$A1,$7F,$E8,$B8,$FF
	.dc.b	$1D,$FC,$1B,$E2,$7F,$D1,$DE,$17,$FB,$57,$C2,$7E,$CD,$FC,$9F,$8B
	.dc.b	$F8,$CF,$E1,$1F,$8F,$7F,$C9,$BE,$8F,$FA,$43,$F0,$57,$D5,$FF,$6E
	.dc.b	$FC,$1B,$FD,$27,$F9,$5F,$F9,$5A,$31,$F5,$DF,$DD,$BF,$07,$FF,$46
	.dc.b	$F8,$9F,$44,$FF,$42,$F2,$FF,$A4,$3B,$7F,$1B,$A7,$F9,$73,$8F,$FB
	.dc.b	$19,$FE,$7E,$FC,$0B,$FC,$DA,$FF,$95,$3F,$02,$FF,$1F,$F4,$7F,$8F
	.dc.b	$BF,$01,$FF,$21,$63,$FE,$41,$C3,$FC,$1D,$7F,$E5,$AF,$96,$10,$87
	.dc.b	$FB,$D6,$37,$1F,$A1,$E2,$0F,$D4,$FF,$93,$7E,$13,$FD,$AB,$F3,$57
	.dc.b	$D4,$7F,$94,$79,$5E,$8B,$F4,$B7,$8D,$FE,$61,$F9,$7F,$EE,$3F,$25
	.dc.b	$FC,$47,$5F,$F3,$FF,$EA,$BF,$1B,$B3,$E5,$FD,$BF,$FA,$F7,$87,$FC
	.dc.b	$37,$E2,$7F,$57,$21,$FD,$67,$F3,$6F,$F5,$7F,$27,$A8,$F5,$CB,$EC
	.dc.b	$7F,$BF,$FE,$AA,$F2,$5F,$1F,$FD,$8B,$E6,$1F,$98,$3F,$57,$FF,$3F
	.dc.b	$E3,$FE,$FF,$FB,$CF,$F0,$DF,$E6,$8F,$D5,$1F,$AB,$FF,$8D,$FE,$2D
	.dc.b	$DB,$FD,$27,$EB,$7F,$C1,$BF,$A5,$BF,$4E,$7E,$6A,$8C,$37,$AB,$F9
	.dc.b	$89,$FE,$8E,$F7,$5F,$27,$F9,$5F,$84,$7E,$87,$FB,$4F,$C8,$3F,$D3
	.dc.b	$5E,$17,$CD,$C3,$FC,$4F,$F2,$3F,$AE,$F8,$8F,$F0,$EF,$E4,$9F,$C8
	.dc.b	$FC,$5F,$ED,$BD,$2F,$B4,$C7,$4F,$FA,$9B,$E2,$3F,$7D,$F1,$3F,$C0
	.dc.b	$3F,$41,$FB,$BA,$7F,$A8,$3F,$28,$FF,$79,$7A,$CF,$AF,$53,$F5,$37
	.dc.b	$B7,$F8,$9D,$BF,$C5,$BE,$C7,$DD,$D7,$FC,$A7,$ED,$FF,$7D,$FA,$2F
	.dc.b	$E6,$BC,$2F,$C6,$7F,$DE,$5C,$EF,$CE,$DF,$9A,$A3,$0D,$FA,$BF,$EE
	.dc.b	$DF,$11,$F7,$3C,$DF,$B7,$F6,$EF,$37,$ED,$F9,$BF,$EB,$8E,$FF,$B1
	.dc.b	$D3,$FE,$7F,$F8,$BF,$F1,$4B,$9F,$CD,$7F,$D2,$9E,$07,$F9,$12,$3F
	.dc.b	$2F,$FA,$CF,$F7,$3A,$FF,$44,$47,$F8,$DF,$F5,$DF,$F2,$4D,$7E,$67
	.dc.b	$0F,$BD,$3D,$CF,$E3,$FF,$A8,$FF,$87,$7F,$0A,$FF,$13,$43,$F5,$A7
	.dc.b	$E8,$DF,$D4,$3E,$73,$FB,$07,$5F,$FB,$2F,$E6,$3F,$DB,$0B,$F3,$FF
	.dc.b	$F7,$7B,$FB,$4F,$89,$FF,$66,$1F,$E4,$D5,$FA,$DF,$FB,$D3,$E0,$3F
	.dc.b	$8F,$66,$FD,$5D,$FA,$3F,$F8,$8F,$F0,$37,$E0,$2F,$AC,$FE,$97,$5F
	.dc.b	$EB,$4B,$FF,$6F,$FD,$47,$C3,$7F,$77,$D7,$FE,$0A,$F9,$2F,$F4,$A7
	.dc.b	$4F,$FB,$3A,$3E,$47,$EA,$D8,$FA,$0F,$F1,$E7,$A1,$C7,$1F,$F2,$F7
	.dc.b	$C1,$7C,$EC,$BF,$CC,$1D,$FF,$EC,$3F,$C6,$FE,$87,$6B,$C0,$FA,$1A
	.dc.b	$BF,$26,$F5,$E1,$FC,$2B,$C1,$FF,$3C,$E1,$E7,$8F,$DB,$F8,$1F,$DB
	.dc.b	$38,$3F,$CF,$BD,$3F,$26,$7F,$01,$8F,$FA,$91,$FF,$1F,$7C,$7F,$FA
	.dc.b	$0B,$C5,$7E,$95,$FB,$CF,$CC,$5F,$9E,$FE,$E7,$E3,$3F,$C0,$99,$7F
	.dc.b	$AB,$3E,$89,$F9,$C3,$F7,$67,$E6,$0F,$D2,$5F,$AF,$7A,$7E,$53,$FC
	.dc.b	$99,$E7,$7E,$D2,$9F,$AE,$7F,$17,$9F,$E1,$1C,$41,$FA,$F3,$F3,$3F
	.dc.b	$F7,$EF,$E6,$23,$FD,$07,$BB,$A7,$FD,$37,$FA,$F3,$F6,$57,$E8,$FC
	.dc.b	$3F,$58,$FF,$2E,$57,$FA,$83,$B3,$FC,$23,$87,$1D,$3C,$1C,$5F,$BC
	.dc.b	$FB,$5F,$E5,$BE,$7B,$E7,$BF,$CC,$9F,$75,$FE,$15,$FD,$B5,$F0,$AF
	.dc.b	$DD,$3F,$D6,$5C,$9F,$E8,$CF,$2F,$FC,$D5,$EA,$FD,$FD,$7F,$CA,$1E
	.dc.b	$2F,$FA,$C3,$F3,$77,$E5,$5F,$BE,$F4,$7F,$8E,$FE,$91,$FC,$C7,$E6
	.dc.b	$BE,$E6,$5F,$89,$E5,$FC,$86,$3F,$EE,$3F,$90,$FD,$F6,$7F,$EB,$AC
	.dc.b	$7F,$27,$FA,$CF,$E4,$23,$F7,$5F,$3B,$FE,$AD,$FA,$5F,$F5,$FF,$82
	.dc.b	$FD,$7F,$FC,$8D,$1F,$A5,$FD,$93,$FB,$FF,$E8,$3F,$D5,$BF,$12,$FF
	.dc.b	$39,$7A,$FF,$E6,$53,$ED,$E5,$FB,$D3,$FC,$D7,$DE,$FD,$9F,$E6,$7D
	.dc.b	$7F,$F3,$19,$FE,$21,$F9,$8B,$F6,$97,$EB,$CF,$D8,$7F,$D7,$7F,$85
	.dc.b	$3E,$E7,$F0,$B8,$7E,$07,$F2,$4F,$F1,$EE,$AF,$C3,$9F,$3A,$FB,$6F
	.dc.b	$C1,$61,$F7,$30,$FC,$47,$E5,$4F,$85,$7E,$3F,$FE,$E5,$FC,$FF,$FD
	.dc.b	$F9,$FA,$0B,$8B,$BE,$1F,$E6,$3F,$C8,$5F,$95,$7F,$A3,$7F,$3D,$7E
	.dc.b	$80,$FD,$39,$FA,$DF,$7C,$3F,$C0,$9F,$8E,$3F,$1C,$C7,$BE,$FD,$6D
	.dc.b	$FB,$43,$F0,$E7,$E7,$2F,$DA,$1F,$AF,$F8,$FC,$47,$C9,$7F,$BD,$3F
	.dc.b	$60,$7E,$54,$F9,$BF,$EF,$5E,$5B,$E6,$DF,$88,$FF,$50,$C7,$FB,$9F
	.dc.b	$1F,$F9,$1C,$FA,$BF,$A0,$7F,$BF,$BF,$12,$FF,$5E,$7E,$D1,$F2,$BE
	.dc.b	$DF,$FA,$D7,$E7,$3F,$EE,$2C,$3F,$C8,$58,$7E,$C7,$E9,$7D,$FF,$AF
	.dc.b	$BF,$68,$FF,$69,$FE,$65,$F0,$FE,$07,$FD,$8D,$F7,$1F,$D1,$70,$FC
	.dc.b	$1F,$E4,$5F,$E5,$37,$FD,$5F,$F1,$DF,$CC,$BF,$45,$FF,$46,$FE,$C5
	.dc.b	$FB,$4F,$CD,$5E,$7B,$F4,$FF,$F0,$2D,$FA,$B0,$7C,$37,$E1,$5F,$F5
	.dc.b	$CF,$C3,$FF,$89,$FD,$D7,$E1,$3F,$E8,$A9,$FE,$10,$7F,$BF,$BD,$7F
	.dc.b	$EA,$17,$FE,$62,$FB,$07,$DC,$7E,$6F,$E5,$37,$FB,$2F,$FB,$97,$F0
	.dc.b	$E7,$ED,$8F,$D0,$5F,$B7,$FF,$73,$FC,$53,$F9,$86,$BF,$C5,$74,$FF
	.dc.b	$1D,$BF,$EC,$CF,$82,$7C,$4F,$FB,$67,$D8,$FD,$47,$C9,$7F,$B4,$6B
	.dc.b	$5F,$4F,$FC,$D3,$8B,$F5,$9F,$45,$FD,$B7,$CD,$FE,$D3,$5E,$AC,$9F
	.dc.b	$6B,$FE,$0A,$C7,$FD,$13,$F1,$BF,$C6,$5F,$F4,$5F,$E2,$DF,$B6,$BF
	.dc.b	$D9,$FD,$4F,$C2,$FD,$97,$FD,$4D,$F8,$6B,$F4,$F7,$8B,$FC,$FB,$F6
	.dc.b	$D7,$E3,$DF,$9B,$F5,$FF,$DD,$3F,$AE,$3F,$5F,$7E,$56,$FD,$7D,$FB
	.dc.b	$6F,$E4,$FD,$46,$FF,$C3,$3F,$5C,$FF,$68,$7D,$4F,$ED,$FC,$C7,$EA
	.dc.b	$DF,$E5,$9F,$AF,$FA,$BF,$77,$FE,$8C,$FC,$1D,$FA,$5B,$F1,$0F,$CD
	.dc.b	$FD,$17,$DC,$D7,$AF,$57,$F4,$3F,$D3,$FF,$63,$FA,$3B,$F3,$CF,$E1
	.dc.b	$9F,$0D,$FE,$32,$F4,$D8,$7F,$9E,$3F,$12,$1F,$03,$F6,$DF,$F1,$57
	.dc.b	$90,$FD,$A3,$FC,$BB,$EF,$FE,$23,$CF,$7E,$67,$FE,$B6,$F0,$FE,$67
	.dc.b	$07,$F8,$07,$DC,$FC,$A5,$7E,$02,$3F,$C0,$9E,$E7,$E4,$FE,$2F,$F2
	.dc.b	$51,$FD,$67,$99,$FB,$5F,$F8,$87,$BF,$84,$FF,$51,$3F,$2F,$C5,$FC
	.dc.b	$AD,$FF,$33,$FE,$03,$F9,$AF,$9F,$7D,$CB,$F6,$EF,$F3,$7F,$0F,$F4
	.dc.b	$FF,$8F,$FE,$51,$EA,$3F,$69,$FC,$DF,$DF,$FF,$1A,$FD,$1F,$FC,$E7
	.dc.b	$F7,$9F,$D8,$FE,$1F,$FD,$23,$F1,$BE,$FE,$5F,$E7,$9F,$98,$FF,$A4
	.dc.b	$B8,$7F,$DA,$BE,$47,$EA,$BE,$FF,$FC,$A3,$F1,$8F,$CC,$BF,$15,$E6
	.dc.b	$3F,$45,$FE,$39,$E2,$FF,$18,$C7,$FB,$4F,$EA,$8F,$DC,$FF,$73,$F1
	.dc.b	$EF,$D6,$7F,$92,$7E,$B3,$FC,$C3,$F2,$B7,$AC,$FC,$AF,$FC,$13,$FB
	.dc.b	$6B,$C2,$7B,$BF,$EB,$0D,$BF,$D0,$DB,$7F,$34,$3C,$E7,$F8,$1F,$F1
	.dc.b	$57,$A7,$FE,$29,$FB,$97,$CC,$FF,$A5,$BF,$1D,$FE,$77,$83,$F3,$AF
	.dc.b	$E0,$FF,$6C,$FD,$5F,$F2,$5F,$99,$FF,$91,$FF,$06,$7E,$BA,$FD,$97
	.dc.b	$FD,$2B,$BF,$F3,$4F,$DC,$15,$FC,$D7,$F5,$97,$EA,$1F,$5F,$F2,$CF
	.dc.b	$F4,$F7,$D6,$FF,$14,$FC,$99,$F5,$AF,$D8,$9F,$BB,$3F,$70,$FE,$79
	.dc.b	$FB,$07,$7F,$53,$FD,$71,$BF,$F8,$8F,$D9,$FF,$1A,$FC,$E7,$FA,$6B
	.dc.b	$F3,$BF,$E9,$7F,$77,$FE,$CC,$FC,$73,$F3,$57,$FD,$DF,$F4,$9B,$E1
	.dc.b	$FF,$15,$73,$1F,$C5,$FF,$68,$FF,$4A,$BF,$E5,$D3,$FD,$7F,$F9,$FF
	.dc.b	$F1,$17,$FB,$D1,$FE,$76,$FC,$23,$DF,$F5,$3F,$EB,$2F,$86,$FE,$4D
	.dc.b	$D1,$FB,$0E,$7F,$E6,$3F,$6F,$FE,$6C,$FD,$A3,$FB,$4F,$E8,$0F,$D1
	.dc.b	$BF,$DD,$FC,$BF,$DA,$7C,$2F,$F6,$7F,$D4,$38,$1F,$68,$FD,$DB,$E2
	.dc.b	$7D,$27,$F7,$6F,$A2,$FE,$C1,$F8,$8B,$F7,$57,$EE,$FF,$C3,$BF,$64
	.dc.b	$7E,$AB,$FE,$5C,$E8,$FF,$25,$7D,$1F,$F1,$9F,$CE,$1F,$93,$BF,$56
	.dc.b	$FE,$47,$F9,$BB,$F1,$4F,$E1,$3F,$7F,$F0,$71,$FF,$0E,$FC,$67,$F4
	.dc.b	$CF,$D6,$9F,$BB,$7F,$BE,$BF,$0C,$7E,$ED,$FE,$01,$F9,$8F,$E8,$3E
	.dc.b	$EF,$F6,$AC,$5F,$A6,$B8,$FF,$2B,$F8,$0F,$F2,$AF,$D1,$1F,$A7,$7F
	.dc.b	$87,$3E,$5F,$E9,$61,$FE,$8F,$F8,$0F,$F3,$E3,$FE,$63,$FC,$0B,$FD
	.dc.b	$69,$E7,$7E,$BE,$1F,$AE,$FF,$34,$FC,$77,$F6,$DF,$92,$7F,$95,$3E
	.dc.b	$D7,$F2,$BC,$1F,$CA,$AD,$F1,$4F,$DC,$BF,$CD,$DF,$90,$BF,$39,$FE
	.dc.b	$9F,$F9,$B7,$F8,$37,$0F,$F3,$26,$0F,$39,$FB,$8F,$F1,$3F,$67,$F3
	.dc.b	$7F,$13,$FD,$EB,$F2,$DF,$F1,$5F,$DD,$7C,$BF,$7B,$FD,$FF,$FB,$0F
	.dc.b	$F1,$9A,$7E,$BB,$FE,$29,$EA,$FF,$0D,$7C,$BF,$EB,$BC,$76,$2F,$A9
	.dc.b	$FF,$75,$FD,$A7,$FA,$EB,$F5,$FF,$F1,$8F,$D2,$FF,$13,$F9,$4F,$E5
	.dc.b	$BF,$8D,$FF,$BC,$3B,$DF,$A9,$FF,$29,$FB,$2F,$E5,$1F,$86,$7F,$BA
	.dc.b	$FF,$38,$7D,$97,$F2,$8F,$DD,$FF,$93,$FE,$55,$DF,$F1,$4F,$F4,$7F
	.dc.b	$E0,$0F,$B4,$F8,$1F,$B2,$7F,$27,$F0,$3F,$B6,$9F,$E9,$5F,$B3,$FE
	.dc.b	$4B,$C5,$FC,$93,$EC,$7F,$91,$FB,$3F,$E1,$2F,$CE,$DF,$65,$FC,$8F
	.dc.b	$EF,$FF,$C2,$FF,$63,$FC,$8B,$C8,$FC,$BE,$7F,$74,$FE,$36,$3F,$73
	.dc.b	$7E,$6E,$FD,$A5,$F9,$6F,$F7,$5F,$91,$FC,$DF,$1F,$E1,$B0,$FF,$10
	.dc.b	$FE,$3D,$FE,$A9,$F6,$BF,$16,$FF,$A1,$A3,$FD,$59,$DD,$F8,$9F,$A7
	.dc.b	$FE,$27,$F4,$DF,$C4,$BE,$97,$F8,$8F,$D5,$FF,$46,$BF,$F0,$F1,$FD
	.dc.b	$18,$FF,$5A,$7E,$35,$F9,$67,$FB,$7F,$C7,$FE,$9D,$FF,$73,$71,$7E
	.dc.b	$3F,$F3,$7F,$F0,$3F,$D7,$BF,$6D,$F5,$7D,$E3,$F5,$1F,$F2,$FD,$FE
	.dc.b	$6F,$E8,$7F,$84,$7C,$2B,$E8,$3F,$84,$7E,$75,$F9,$1A,$FC,$D7,$D0
	.dc.b	$7F,$08,$FA,$0F,$E0,$FF,$3F,$FC,$1A,$7F,$9E,$6F,$FE,$C9,$F8,$FF
	.dc.b	$D8,$7D,$B7,$F8,$EB,$F1,$57,$C7,$FE,$C3,$97,$E4,$5F,$F8,$BF,$B3
	.dc.b	$E3,$FE,$AD,$FE,$F2,$E3,$FF,$78,$CF,$FC,$01,$F8,$6F,$C2,$3F,$34
	.dc.b	$FF,$25,$99,$FA,$8D,$BF,$EF,$AF,$A9,$FE,$FD,$E0,$7F,$2D,$C3,$FC
	.dc.b	$8B,$3C,$63,$15,$C2,$3F,$CB,$33,$D9,$FF,$5E,$CF,$1C,$35,$1F,$EE
	.dc.b	$5C,$3F,$D2,$ED,$3C,$67,$F9,$88,$FD,$C7,$A5,$FE,$64,$E0,$FF,$35
	.dc.b	$F9,$3F,$E5,$78,$FF,$25,$61,$6B,$F1,$1F,$8F,$7F,$9C,$1F,$81,$3F
	.dc.b	$EE,$7F,$C0,$3A,$C7,$F9,$56,$3F,$DD,$5F,$3D,$B8,$F0,$64,$FD,$A3
	.dc.b	$FC,$59,$F9,$07,$09,$F4,$47,$F8,$BF,$F6,$7E,$11,$FE,$5F,$8F,$F1
	.dc.b	$97,$E7,$1B,$8F,$E2,$11,$FE,$35,$FC,$4B,$71,$C8,$FC,$F7,$B3,$BC
	.dc.b	$7C,$11,$FA,$1F,$37,$1F,$BE,$F4,$E0,$FF,$3B,$3F,$ED,$C7,$FC,$F1
	.dc.b	$8F,$FA,$23,$9F,$FC,$A7,$C5,$8C,$3E,$1B,$C1,$FF,$40,$67,$E2,$7F
	.dc.b	$92,$61,$FE,$49,$E3,$C6,$7F,$E6,$09,$FA,$99,$7F,$A1,$79,$3F,$C8
	.dc.b	$1F,$AE,$D9,$FC,$57,$A5,$FE,$89,$EF,$FF,$29,$69,$FC,$C2,$7F,$D0
	.dc.b	$51,$8F,$C4,$3F,$15,$A7,$7A,$FC,$66,$FF,$15,$E9,$7F,$E6,$27,$C7
	.dc.b	$7D,$3E,$CB,$B4,$FF,$A6,$65,$FA,$77,$48,$FF,$4E,$6D,$F4,$7A,$FE
	.dc.b	$5C,$7F,$AC,$B9,$7E,$76,$3F,$E5,$9F,$D3,$B9,$CB,$90,$B8,$3F,$D4
	.dc.b	$9F,$90,$FA,$63,$F5,$4F,$BB,$CA,$F0,$47,$FA,$FB,$9B,$E6,$67,$E0
	.dc.b	$D9,$D3,$EA,$AF,$D8,$75,$F8,$33,$EF,$CB,$DA,$7D,$9C,$35,$FC,$FD
	.dc.b	$F6,$C1,$C1,$ED,$E3,$FC,$AF,$3F,$FA,$E7,$BF,$FC,$40,$DC,$3F,$95
	.dc.b	$67,$F8,$D9,$FC,$36,$BF,$8F,$FD,$E7,$FC,$D5,$7F,$EC,$17,$FC,$62
	.dc.b	$FE,$FF,$87,$ED,$E9,$FC,$07,$93,$EE,$ED,$D7,$3F,$AD,$B5,$D3,$FC
	.dc.b	$A8,$DE,$EF,$C5,$E7,$FE,$95,$F3,$AD,$FC,$EF,$C5,$7F,$76,$FE,$FC
	.dc.b	$6D,$FE,$B9,$7E,$F7,$2F,$85,$FC,$07,$5F,$9D,$FA,$FF,$AA,$FD,$8E
	.dc.b	$1C,$47,$DB,$D7,$F5,$9F,$51,$EC,$8E,$D8,$F5,$7F,$85,$EF,$D5,$FE
	.dc.b	$1B,$D9,$CA,$FC,$B1,$FC,$13,$6F,$D7,$6D,$FA,$6C,$7F,$DE,$FB,$6A
	.dc.b	$BF,$DC,$11,$FE,$FA,$F8,$0C,$47,$E3,$33,$F0,$70,$86,$FE,$17,$15
	.dc.b	$FE,$F3,$FD,$63,$FD,$5D,$F7,$9F,$CC,$71,$E9,$5F,$8E,$D3,$C2,$DC
	.dc.b	$BF,$AF,$F0,$7F,$BE,$FF,$7B,$FD,$9F,$D6,$59,$FE,$81,$F8,$D7,$37
	.dc.b	$E9,$5F,$17,$C9,$C1,$F9,$AC,$BE,$44,$7F,$33,$DD,$9F,$E2,$23,$FA
	.dc.b	$67,$CF,$FE,$95,$D3,$5F,$E8,$7B,$7A,$9F,$A3,$A7,$FB,$27,$3F,$E9
	.dc.b	$97,$F3,$33,$FE,$97,$5F,$E0,$1E,$7F,$27,$1F,$ED,$77,$CC,$FE,$DF
	.dc.b	$4F,$DF,$61,$FC,$87,$0F,$D8,$F5,$F3,$AF,$BF,$3F,$BC,$5C,$FF,$87
	.dc.b	$7F,$07,$3F,$CA,$71,$FE,$10,$FF,$4F,$BF,$DD,$F4,$FF,$B7,$FB,$BC
	.dc.b	$3E,$1F,$1C,$3E,$24,$FF,$24,$7F,$1B,$E2,$ED,$FC,$7C,$7F,$B9,$FC
	.dc.b	$1F,$C7,$65,$FC,$3B,$C4,$FE,$C9,$5F,$DC,$FC,$4A,$FE,$B0,$FC,$BF
	.dc.b	$85,$FC,$F7,$CB,$FE,$1B,$C7,$FD,$63,$6D,$EB,$E3,$BF,$EE,$37,$FD
	.dc.b	$91,$FC,$FD,$7F,$61,$E3,$FE,$5F,$1F,$E2,$DE,$0E,$6F,$F4,$63,$B9
	.dc.b	$FD,$52,$3F,$C0,$13,$CE,$3F,$04,$FC,$5C,$33,$1F,$D1,$8F,$DD,$7B
	.dc.b	$FF,$E7,$D8,$FE,$EE,$7F,$78,$F3,$97,$FE,$48,$AF,$E7,$53,$A9,$CF
	.dc.b	$B7,$FE,$01,$8E,$27,$E9,$E3,$FC,$11,$F7,$5F,$E0,$AB,$FF,$3E,$70
	.dc.b	$5D,$7F,$D6,$38,$F4,$3F,$C4,$11,$FE,$10,$7F,$D0,$5F,$AC,$7F,$67
	.dc.b	$7F,$D2,$F3,$FC,$41,$F1,$E3,$E0,$87,$ED,$BB,$9F,$E2,$D8,$7F,$A2
	.dc.b	$FE,$17,$FA,$FD,$F9,$BC,$5F,$F4,$77,$95,$BB,$FC,$87,$6F,$F8,$7F
	.dc.b	$F2,$8C,$47,$F4,$E8,$FF,$11,$7E,$F1,$CB,$0F,$8D,$3F,$E9,$7F,$D5
	.dc.b	$B6,$BD,$5F,$F1,$47,$E6,$BB,$AF,$E1,$5E,$B7,$F8,$AF,$C7,$F0,$71
	.dc.b	$FF,$86,$BF,$39,$B8,$FE,$87,$8D,$F1,$3D,$E3,$FC,$3F,$EC,$57,$51
	.dc.b	$EF,$E4,$FF,$24,$7E,$D1,$EE,$E4,$FD,$FD,$7F,$92,$4F,$F3,$A1,$FE
	.dc.b	$3B,$FD,$F3,$E3,$67,$FE,$4F,$FC,$03,$E8,$67,$FE,$48,$AF,$C1,$99
	.dc.b	$3F,$C9,$47,$F9,$57,$F4,$8F,$F0,$4E,$1F,$C8,$63,$FB,$F3,$FE,$DA
	.dc.b	$7F,$BA,$FE,$7A,$FD,$6A,$10,$7E,$E2,$F7,$BF,$D8,$33,$99,$D4,$7F
	.dc.b	$B1,$3E,$8D,$FA,$A4,$20,$FD,$AF,$97,$B3,$FC,$26,$73,$3A,$AF,$F6
	.dc.b	$D7,$EB,$18,$8E,$77,$FE,$F9,$CD,$39,$9E,$F5,$FE,$BE,$9F,$F0,$7E
	.dc.b	$3E,$53,$D9,$9F,$A7,$E9,$3F,$8C,$46,$4F,$81,$5F,$EC,$2A,$FA,$47
	.dc.b	$FA,$63,$F1,$6F,$C2,$75,$FF,$99,$A5,$F8,$86,$23,$FA,$84,$7F,$BF
	.dc.b	$5F,$EA,$5C,$1F,$9C,$8D,$DF,$D0,$3F,$CE,$5C,$FF,$2F,$0F,$AB,$1F
	.dc.b	$EA,$6C,$2B,$FE,$77,$A7,$CA,$7E,$62,$3F,$C5,$70,$FD,$97,$11,$F5
	.dc.b	$7F,$2A,$F2,$67,$F5,$FF,$43,$BF,$45,$FE,$C8,$AF,$69,$FE,$83,$FA
	.dc.b	$C6,$DA,$F8,$3E,$31,$89,$FD,$CE,$1F,$A4,$DD,$82,$FE,$2F,$87,$FC
	.dc.b	$4B,$0E,$18,$7F,$0E,$C3,$F5,$6F,$0A,$DA,$97,$E0,$7E,$A5,$F8,$F7
	.dc.b	$8B,$F3,$8F,$5F,$81,$E9,$BF,$D2,$9F,$ED,$C7,$07,$FB,$44,$F2,$3E
	.dc.b	$8C,$7F,$D3,$B1,$FE,$A2,$7C,$83,$E1,$BF,$5F,$F1,$FC,$6B,$FA,$E3
	.dc.b	$FE,$E6,$AF,$F5,$47,$B5,$C2,$FF,$C5,$93,$FE,$DF,$AC,$C7,$47,$FD
	.dc.b	$5D,$1F,$EE,$E7,$C1,$1F,$0E,$FF,$CC,$7A,$7D,$69,$F4,$9F,$EE,$9F
	.dc.b	$80,$FB,$DC,$1F,$EF,$DD,$3E,$DF,$A4,$D6,$EF,$F8,$08,$F9,$F9,$FC
	.dc.b	$FE,$0F,$F3,$7F,$A7,$F0,$B3,$F7,$C7,$F9,$CA,$B7,$5F,$D0,$A3,$FA
	.dc.b	$4D,$6E,$39,$64,$AD,$C7,$DB,$9D,$3F,$29,$F4,$DF,$F9,$3C,$7B,$2E
	.dc.b	$6F,$C9,$AE,$6B,$D3,$F9,$88,$FC,$D5,$7E,$50,$FF,$06,$3F,$AB,$C3
	.dc.b	$F1,$71,$8E,$AF,$D2,$BA,$EB,$AF,$F6,$2A,$FD,$7B,$FE,$DC,$E9,$FE
	.dc.b	$9B,$1C,$EF,$F2,$8E,$9F,$B8,$FF,$01,$9F,$C6,$9E,$2E,$DF,$E0,$FD
	.dc.b	$BF,$C1,$E7,$F8,$74,$FF,$09,$8F,$E4,$B1,$FC,$63,$BF,$F8,$A3,$5F
	.dc.b	$E3,$BE,$0F,$C4,$F0,$FF,$90,$46,$9A,$EA,$7F,$96,$D7,$79,$FF,$17
	.dc.b	$78,$9F,$C5,$FC,$2F,$F3,$DF,$8B,$F0,$32,$FF,$3E,$FC,$2D,$F8,$47
	.dc.b	$93,$FD,$CF,$A3,$EA,$46,$2B,$FB,$D1,$FE,$51,$F9,$0F,$EF,$50,$F8
	.dc.b	$7F,$E9,$F3,$FE,$55,$F3,$2B,$FE,$3F,$97,$9B,$FD,$66,$3F,$CB,$5E
	.dc.b	$77,$F5,$D6,$9F,$E0,$03,$FC,$BD,$E7,$FF,$63,$7F,$CF,$1E,$8F,$F6
	.dc.b	$A7,$FC,$F5,$E8,$D7,$FD,$2F,$E9,$7F,$06,$C5,$EA,$7F,$78,$F5,$3F
	.dc.b	$D0,$1E,$AF,$F3,$4F,$6F,$FD,$05,$EC,$FF,$7C,$F6,$3F,$D0,$9E,$CD
	.dc.b	$CF,$FD,$43,$E4,$7F,$37,$F0,$BF,$D4,$7E,$47,$EB,$B5,$FF,$0B,$57
	.dc.b	$EF,$7C,$3F,$F5,$3F,$B5,$FC,$B2,$7F,$C4,$33,$FD,$B8,$FF,$3C,$7A
	.dc.b	$7F,$23,$DB,$FF,$14,$4F,$FB,$7E,$7E,$69,$FE,$A9,$FD,$7B,$FC,$61
	.dc.b	$87,$F8,$C3,$7F,$F7,$1F,$EA,$1F,$E3,$B8,$C7,$6A,$B7,$1F,$E3,$98
	.dc.b	$FF,$5C,$F8,$BF,$77,$F1,$CF,$BD,$F2,$BF,$7F,$F0,$7F,$37,$EA,$FF
	.dc.b	$81,$F4,$7F,$01,$AF,$FA,$D1,$FF,$3B,$65,$FE,$A2,$8C,$1F,$D4,$CF
	.dc.b	$F4,$97,$A1,$C1,$FF,$4A,$79,$FF,$E7,$DF,$CE,$DF,$56,$FD,$AF,$F1
	.dc.b	$73,$98,$7E,$BD,$8B,$FF,$7F,$FE,$52,$1F,$AD,$E6,$3F,$C1,$BF,$91
	.dc.b	$5F,$4F,$FB,$AC,$7F,$B2,$B1,$FF,$31,$34,$0F,$88,$FE,$1C,$FF,$B0
	.dc.b	$63,$FD,$25,$F8,$AB,$F6,$C7,$E8,$5D,$E3,$1F,$90,$FF,$15,$C1,$7C
	.dc.b	$2D,$39,$8C,$5D,$C3,$FB,$7F,$17,$FB,$35,$FF,$76,$61,$C9,$F2,$BE
	.dc.b	$47,$CD,$5F,$C2,$3D,$6E,$51,$F0,$0E,$31,$FE,$A9,$DB,$E3,$6D,$AE
	.dc.b	$2F,$9F,$C5,$78,$BE,$6E,$1E,$0E,$3F,$AC,$FF,$94,$79,$3E,$EE,$AF
	.dc.b	$63,$DD,$1F,$C1,$76,$FE,$65,$5F,$E4,$43,$FD,$70,$7C,$BE,$6D,$73
	.dc.b	$FF,$82,$A5,$D1,$C7,$A7,$97,$57,$F0,$DE,$2F,$47,$13,$F1,$EF,$6B
	.dc.b	$F7,$DF,$E1,$FF,$27,$FC,$33,$E9,$BC,$FA,$FF,$BD,$DF,$E1,$FF,$98
	.dc.b	$E2,$1E,$1B,$F1,$AF,$E2,$7F,$87,$E6,$0F,$F1,$D3,$F8,$BF,$E4,$F9
	.dc.b	$3F,$DC,$55,$FE,$EB,$F9,$7F,$CA,$78,$7F,$EB,$48,$75,$FC,$C7,$FD
	.dc.b	$66,$7E,$57,$6F,$CA,$B8,$70,$EB,$C2,$FC,$AB,$D4,$FF,$26,$BF,$E6
	.dc.b	$77,$9F,$C9,$3A,$BF,$0D,$1B,$3F,$30,$7F,$AD,$8F,$0F,$87,$F9,$46
	.dc.b	$F8,$6F,$D1,$BF,$D8,$31,$E5,$F0,$6E,$5E,$48,$FB,$3C,$1F,$C6,$3C
	.dc.b	$BF,$F6,$B4,$7E,$9A,$3F,$8D,$E7,$E3,$CF,$FA,$0C,$7F,$A5,$E7,$C4
	.dc.b	$72,$F3,$3F,$5E,$D3,$FD,$A0,$7E,$1B,$B1,$1C,$7C,$8F,$D5,$3C,$27
	.dc.b	$FD,$45,$59,$7F,$4A,$E5,$FD,$2B,$F7,$EB,$FC,$0B,$1F,$C1,$B2,$FA
	.dc.b	$2F,$7E,$BF,$C6,$C7,$ED,$F8,$FF,$6D,$E3,$DA,$FE,$07,$C3,$FB,$FE
	.dc.b	$0C,$47,$CE,$D2,$ED,$F0,$63,$F6,$F8,$7E,$CD,$A7,$DA,$E2,$E3,$97
	.dc.b	$EC,$A3,$27,$E1,$5E,$BC,$7F,$C8,$B9,$2F,$89,$FC,$8E,$BF,$81,$F3
	.dc.b	$7B,$4F,$C9,$BF,$EC,$5E,$7F,$D9,$1F,$EB,$83,$E1,$74,$7F,$AE,$BA
	.dc.b	$7E,$27,$57,$FA,$F0,$F8,$DD,$7F,$EE,$DE,$47,$E5,$9F,$A3,$FB,$4F
	.dc.b	$EE,$F8,$7E,$94,$FD,$26,$5F,$C9,$BA,$FF,$C9,$AF,$EF,$8F,$D6,$FD
	.dc.b	$57,$EA,$CF,$D2,$B9,$EA,$47,$F3,$BF,$23,$FC,$0E,$FE,$FB,$BA,$D7
	.dc.b	$D7,$E4,$AF,$FB,$1B,$C9,$BC,$DF,$EC,$88,$E9,$7F,$83,$7F,$2E,$FD
	.dc.b	$A7,$27,$DA,$F0,$3F,$40,$FF,$2C,$8D,$3F,$91,$7F,$9B,$75,$FF,$39
	.dc.b	$7A,$47,$FB,$C5,$FB,$8E,$CF,$B8,$73,$3F,$1D,$8F,$17,$D4,$64,$FE
	.dc.b	$0B,$CD,$FC,$B9,$E4,$63,$3E,$0F,$F7,$DE,$BF,$E9,$5C,$DF,$7A,$3F
	.dc.b	$A0,$F0,$7F,$2F,$3F,$3D,$C3,$FE,$F6,$E4,$FE,$3D,$D7,$F6,$AB,$F9
	.dc.b	$91,$FE,$2B,$D3,$F4,$CF,$F5,$D6,$F2,$BF,$61,$FE,$4C,$D9,$FF,$33
	.dc.b	$1F,$BF,$F8,$5F,$F1,$CF,$C0,$7F,$8E,$BE,$03,$F9,$41,$FE,$43,$F9
	.dc.b	$FF,$BF,$5F,$C2,$1F,$BC,$FE,$05,$FB,$CD,$3E,$29,$B3,$F8,$FF,$37
	.dc.b	$F9,$2A,$3F,$CC,$7C,$D6,$7F,$83,$9F,$E4,$58,$FB,$2D,$3E,$F7,$2F
	.dc.b	$F9,$4B,$E8,$3B,$C7,$F1,$A7,$F7,$EF,$E4,$FD,$CF,$BA,$FF,$8E,$8E
	.dc.b	$A7,$F6,$EF,$6B,$2F,$F8,$0E,$2F,$8C,$AE,$B3,$FE,$08,$3F,$C1,$0F
	.dc.b	$67,$F8,$E6,$1E,$67,$FB,$CB,$F8,$E7,$FB,$EB,$D0,$FC,$6B,$CE,$51
	.dc.b	$F0,$A3,$67,$E3,$8A,$7F,$09,$2F,$87,$FC,$25,$BD,$3F,$C2,$5F,$E5
	.dc.b	$DA,$FF,$77,$3F,$93,$6D,$FD,$F3,$C8,$FF,$0C,$70,$5A,$FD,$E7,$27
	.dc.b	$E0,$DF,$CA,$18,$E3,$FF,$56,$7E,$71,$FE,$7C,$D3,$F1,$BD,$EE,$0B
	.dc.b	$C3,$F9,$0E,$03,$63,$15,$F9,$4C,$98,$7F,$13,$FB,$AF,$C9,$9F,$E6
	.dc.b	$5F,$DB,$BF,$D1,$3A,$FF,$77,$FD,$57,$F2,$7B,$7E,$78,$FE,$A9,$C1
	.dc.b	$FE,$95,$E1,$FD,$44,$7F,$80,$3E,$03,$FB,$F1,$F9,$67,$FD,$1C,$FE
	.dc.b	$59,$FE,$BE,$FD,$C7,$FC,$FE,$7E,$58,$D9,$F9,$63,$F8,$E3,$F5,$5E
	.dc.b	$27,$E2,$8F,$01,$FC,$47,$CC,$FF,$52,$C7,$E0,$63,$FB,$03,$F8,$57
	.dc.b	$FC,$A0,$E5,$FC,$13,$F9,$03,$F0,$AF,$F9,$09,$F0,$3A,$3F,$08,$74
	.dc.b	$3D,$06,$CF,$F2,$EF,$2F,$E1,$5D,$1F,$C4,$4F,$F2,$7C,$7F,$69,$AF
	.dc.b	$93,$C3,$FE,$B2,$9F,$F3,$D1,$F9,$5C,$3F,$B7,$63,$F8,$78,$FE,$F5
	.dc.b	$97,$F7,$1C,$F1,$1F,$A5,$D9,$AF,$F1,$63,$FD,$41,$F0,$5F,$EC,$4D
	.dc.b	$3F,$0C,$7E,$1D,$FE,$A9,$FB,$CF,$F8,$D7,$E9,$7F,$DA,$07,$FB,$2B
	.dc.b	$E1,$3F,$C7,$07,$DE,$E6,$7D,$3F,$DD,$CF,$FC,$2C,$FD,$D7,$FD,$98
	.dc.b	$FD,$D7,$FC,$86,$FE,$DD,$C9,$F7,$4F,$F1,$0B,$F3,$9E,$67,$DD,$8E
	.dc.b	$EF,$DD,$3F,$AF,$47,$D9,$7F,$C4,$DF,$4B,$DE,$4F,$78,$FF,$15,$7C
	.dc.b	$57,$F0,$E9,$7C,$4F,$F0,$E3,$F8,$7C,$7F,$10,$AF,$47,$1F,$2B,$FB
	.dc.b	$A7,$FA,$5F,$0F,$C5,$47,$92,$1E,$A7,$F8,$65,$FF,$45,$CF,$F1,$A3
	.dc.b	$F1,$6F,$13,$FA,$D9,$FE,$33,$7B,$9E,$77,$47,$F5,$AF,$4F,$FD,$3D
	.dc.b	$1F,$36,$3F,$D2,$B9,$FF,$8C,$4F,$C4,$E9,$FE,$85,$3F,$07,$1F,$EA
	.dc.b	$EE,$16,$DF,$E7,$1B,$FE,$13,$A7,$F0,$AD,$7F,$84,$1F,$C2,$5F,$E1
	.dc.b	$51,$FC,$3B,$F7,$EF,$C1,$E1,$FE,$B8,$EF,$FF,$68,$3F,$E4,$BC,$BF
	.dc.b	$D2,$FE,$F7,$FA,$CE,$BF,$D3,$3F,$0F,$FE,$17,$A7,$8E,$FB,$9F,$C7
	.dc.b	$47,$F8,$74,$7C,$EE,$F9,$5F,$1B,$F8,$D9,$FF,$45,$BF,$05,$A7,$EC
	.dc.b	$8F,$E8,$91,$C1,$FA,$ED,$7F,$C3,$87,$F8,$96,$3F,$1C,$F2,$3F,$C9
	.dc.b	$C6,$4F,$C7,$1F,$E7,$47,$F1,$CF,$FB,$5D,$FC,$09,$8E,$5E,$C7,$F9
	.dc.b	$E2,$34,$7E,$74,$FE,$9D,$1E,$08,$F2,$7A,$9F,$E5,$59,$FF,$3A,$4F
	.dc.b	$E0,$67,$D1,$E0,$7F,$09,$9E,$27,$F0,$CF,$1F,$FA,$73,$FD,$67,$7F
	.dc.b	$E6,$3C,$BF,$36,$3F,$9B,$1E,$D7,$FC,$D0,$F6,$3F,$DC,$3C,$3F,$3F
	.dc.b	$EA,$3E,$BC,$7F,$92,$78,$7E,$DB,$FE,$E4,$6B,$FE,$89,$8D,$BF,$ED
	.dc.b	$0E,$07,$E0,$1F,$E1,$C3,$FA,$1A,$F8,$AF,$F6,$04,$FF,$09,$FA,$BF
	.dc.b	$B9,$C2,$7F,$BE,$6C,$D9,$F2,$DF,$83,$9F,$DE,$1F,$ED,$6F,$B6,$C6
	.dc.b	$1F,$34,$FF,$53,$47,$EF,$4F,$F0,$37,$73,$DA,$7E,$45,$C2,$BF,$6F
	.dc.b	$0F,$2B,$F2,$1F,$31,$FC,$24,$ED,$3F,$E6,$BF,$97,$FE,$11,$F2,$DF
	.dc.b	$E1,$43,$FC,$2A,$FF,$85,$A7,$F8,$39,$C2,$7F,$D5,$D9,$78,$B0,$F0
	.dc.b	$C7,$F9,$E7,$A3,$E0,$3F,$8E,$3F,$C2,$53,$FE,$20,$3F,$83,$C6,$A3
	.dc.b	$FD,$90,$71,$8E,$CF,$03,$F0,$9E,$27,$EB,$E7,$87,$07,$CB,$DF,$15
	.dc.b	$F8,$6E,$2F,$F2,$CF,$24,$8F,$D9,$76,$7C,$5B,$FF,$76,$47,$FB,$53
	.dc.b	$2F,$F0,$7E,$BF,$E1,$0D,$BF,$C1,$C7,$F8,$3D,$FF,$08,$44,$9F,$CB
	.dc.b	$7B,$BF,$C0,$79,$9F,$99,$7F,$A8,$7F,$76,$FC,$4D,$7F,$8E,$E5,$F7
	.dc.b	$D8,$81,$FC,$23,$ED,$30,$FF,$0A,$ED,$FC,$9B,$FC,$03,$2F,$F2,$27
	.dc.b	$82,$1F,$15,$F7,$FD,$B0,$F9,$3F,$C1,$76,$7E,$17,$E4,$FF,$03,$8F
	.dc.b	$FB,$66,$FE,$57,$5F,$4C,$79,$CF,$D1,$3F,$D3,$D5,$F3,$F2,$FF,$2A
	.dc.b	$65,$DB,$2C,$57,$33,$FC,$7B,$9F,$A1,$FE,$69,$A7,$B3,$C3,$F8,$31
	.dc.b	$FE,$00,$3F,$9C,$70,$67,$87,$FD,$90,$7E,$B0,$FF,$44,$FE,$39,$FA
	.dc.b	$E8,$FD,$7F,$DD,$FE,$C3,$8B,$F6,$3F,$AD,$7E,$CB,$2F,$D9,$F9,$5F
	.dc.b	$DE,$7E,$7B,$F5,$FE,$2F,$EB,$B4,$FF,$64,$47,$ED,$0F,$C6,$BF,$4B
	.dc.b	$C6,$47,$F6,$D3,$FA,$DB,$88,$FA,$FC,$3F,$3B,$F5,$1F,$E7,$55,$FD
	.dc.b	$BC,$FF,$37,$4F,$F3,$E3,$B9,$FE,$C7,$FD,$EB,$6F,$FB,$27,$C7,$FD
	.dc.b	$B7,$89,$FE,$E0,$8F,$F4,$00,$C5,$D3,$D3,$C1,$FE,$91,$DF,$6B,$F5
	.dc.b	$CD,$78,$DF,$07,$F1,$AF,$57,$F5,$41,$C3,$AC,$BF,$75,$97,$EB,$5F
	.dc.b	$DE,$65,$67,$99,$ED,$3B,$C7,$FA,$68,$FC,$2E,$8D,$38,$4F,$F5,$43
	.dc.b	$67,$E8,$3A,$FF,$35,$E4,$69,$FE,$8D,$1F,$E9,$7B,$FB,$3E,$2F,$2F
	.dc.b	$0F,$FA,$34,$FF,$7F,$EB,$FF,$53,$7C,$67,$EA,$BE,$AF,$FB,$97,$6B
	.dc.b	$9B,$FA,$3F,$3F,$F4,$8E,$8F,$E9,$3D,$3F,$D2,$BA,$BF,$A5,$F5,$FF
	.dc.b	$4C,$EC,$FE,$9B,$DB,$36,$FE,$9B,$E1,$FE,$F1,$FF,$6C,$9F,$DF,$BF
	.dc.b	$22,$FF,$08,$BF,$5F,$8F,$FB,$C7,$77,$E1,$E3,$FD,$54,$7D,$C7,$1B
	.dc.b	$FF,$56,$8F,$EF,$73,$FE,$77,$9E,$1E,$27,$F8,$B6,$7E,$39,$FC,$EB
	.dc.b	$CB,$F9,$71,$98,$F8,$C7,$F5,$E3,$F9,$85,$7F,$6D,$3F,$C1,$56,$C3
	.dc.b	$CE,$FA,$E5,$C7,$FE,$19,$7F,$89,$E4,$FB,$4F,$6B,$63,$F7,$07,$F8
	.dc.b	$7B,$5F,$DF,$3F,$E2,$0F,$BD,$B2,$E0,$B9,$63,$8B,$3E,$9F,$0E,$07
	.dc.b	$D2,$DD,$FF,$6F,$FB,$3F,$E9,$B4,$FE,$3D,$D9,$FC,$51,$F9,$53,$F8
	.dc.b	$EF,$1F,$5C,$4E,$BF,$C0,$57,$EC,$0F,$EC,$3E,$4F,$F6,$39,$D7,$5F
	.dc.b	$F2,$29,$46,$1F,$EC,$47,$E3,$63,$FA,$77,$8B,$FE,$8F,$BE,$B3,$FA
	.dc.b	$EF,$1F,$15,$FC,$8A,$39,$CE,$9F,$AA,$78,$8F,$AA,$75,$29,$FC,$E3
	.dc.b	$C3,$FE,$70,$E3,$4F,$E7,$DE,$57,$E8,$7C,$96,$FF,$D8,$78,$3E,$8E
	.dc.b	$BE,$4C,$BE,$17,$D8,$7E,$8B,$4C,$F1,$EB,$93,$7C,$BC,$39,$B8,$F4
	.dc.b	$72,$E9,$E7,$D5,$F8,$63,$FA,$9F,$57,$CD,$E3,$FE,$E2,$7F,$16,$3F
	.dc.b	$AB,$1F,$D8,$8C,$3B,$7F,$BD,$9F,$E2,$9D,$7F,$D0,$11,$9E,$CF,$F0
	.dc.b	$C7,$6F,$F6,$99,$FF,$28,$F6,$7F,$22,$E6,$FE,$07,$87,$F3,$53,$FD
	.dc.b	$11,$1F,$C4,$7C,$9F,$F2,$92,$F9,$AF,$E0,$72,$F9,$3F,$F3,$74,$7F
	.dc.b	$B4,$0C,$BF,$E0,$97,$FD,$A7,$E6,$7F,$38,$8F,$F0,$4C,$FF,$9B,$CF
	.dc.b	$E3,$E7,$FA,$93,$9B,$D1,$E9,$FF,$8A,$E3,$FA,$F9,$FD,$EB,$5F,$DA
	.dc.b	$3F,$C5,$4E,$47,$E1,$0E,$5A,$7F,$9F,$EB,$E3,$79,$1F,$E1,$29,$FF
	.dc.b	$3E,$E3,$F8,$FA,$FF,$0F,$36,$3E,$6F,$DD,$7F,$82,$1F,$F2,$4F,$73
	.dc.b	$F2,$AF,$F0,$94,$E7,$CB,$FF,$01,$FE,$B5,$FE,$15,$CF,$16,$F9,$1C
	.dc.b	$63,$FE,$12,$BF,$C9,$3F,$6B,$8B,$73,$FE,$1D,$78,$D7,$F8,$7F,$EC
	.dc.b	$F3,$5F,$80,$72,$FE,$0B,$EC,$B2,$FF,$9A,$5C,$BF,$E6,$BF,$C0,$72
	.dc.b	$FF,$89,$5C,$BF,$E2,$7E,$BC,$BF,$C5,$1C,$BF,$C5,$BC,$3A,$FF,$AC
	.dc.b	$1F,$6F,$93,$FE,$33,$78,$C6,$4F,$76,$3F,$1D,$FC,$BF,$AD,$C1,$A6
	.dc.b	$B7,$FF,$1E,$47,$2E,$C3,$E0,$6C,$F4,$F4,$F6,$F6,$E3,$FC,$E3,$FE
	.dc.b	$3E,$FD,$9D,$F9,$2B,$F2,$FC,$CA,$73,$38,$FC,$49,$F9,$F3,$F6,$4C
	.dc.b	$2F,$D4,$38,$FE,$CB,$FD,$2F,$1F,$E7,$3D,$BF,$85,$FD,$2B,$FD,$6D
	.dc.b	$BF,$EC,$FF,$1B,$7E,$80,$FD,$07,$C3,$18,$FC,$E3,$FD,$A9,$8F,$E3
	.dc.b	$78,$3C,$FC,$38,$8F,$57,$0F,$A9,$FE,$05,$E4,$7F,$80,$F6,$F1,$47
	.dc.b	$F8,$67,$C3,$F1,$3F,$B3,$37,$FB,$0F,$D4,$A3,$F6,$9B,$1F,$2C,$D9
	.dc.b	$F2,$CF,$F4,$64,$7E,$6B,$9B,$FD,$04,$FE,$73,$9A,$CF,$DE,$F0,$6B
	.dc.b	$FA,$6E,$AF,$D5,$BF,$E7,$0E,$3F,$F6,$D7,$83,$E9,$E5,$FE,$31,$3F
	.dc.b	$E3,$A3,$DF,$BF,$F0,$47,$EB,$6F,$F0,$5F,$F1,$BF,$87,$F3,$79,$7E
	.dc.b	$64,$7F,$04,$1C,$9F,$35,$C6,$5F,$EE,$9A,$FE,$4D,$C5,$F2,$1F,$E5
	.dc.b	$1D,$5F,$80,$EA,$FF,$0D,$E9,$FB,$F8,$FC,$1F,$4E,$31,$F2,$9D,$F9
	.dc.b	$BF,$12,$FF,$35,$C3,$F2,$AF,$F0,$6E,$4F,$F1,$2F,$5F,$C7,$8D,$3F
	.dc.b	$A6,$9D,$F8,$7F,$C0,$A7,$DB,$3F,$A5,$69,$FA,$53,$B7,$77,$37,$E1
	.dc.b	$BC,$CF,$DA,$D7,$F5,$88,$FD,$8F,$7F,$F8,$DB,$C0,$F0,$1A,$F0,$7C
	.dc.b	$47,$0F,$0B,$C8,$72,$F0,$FC,$C7,$4E,$8F,$41,$DB,$1F,$51,$E0,$F1
	.dc.b	$3C,$99,$7F,$27,$84,$7F,$22,$F1,$3F,$93,$CF,$FA,$CB,$F0,$87,$DE
	.dc.b	$BF,$4E,$CC,$A7,$3F,$AA,$7F,$28,$8C,$57,$07,$FB,$0E,$3F,$CD,$6B
	.dc.b	$9F,$E0,$1F,$CA,$67,$B5,$6D,$FF,$3F,$F9,$1F,$E7,$37,$33,$FC,$B5
	.dc.b	$DC,$D7,$BC,$7F,$B3,$F6,$ED,$3C,$E7,$FD,$9D,$18,$9F,$F6,$A6,$FA
	.dc.b	$AB,$5D,$8F,$EE,$FC,$19,$69,$93,$CB,$C5,$96,$9E,$AD,$3F,$97,$3B
	.dc.b	$3F,$DB,$D1,$E4,$FD,$1B,$F9,$1C,$7C,$69,$BA,$7C,$EF,$D8,$EE,$3E
	.dc.b	$9A,$FF,$28,$F2,$FA,$5F,$E6,$67,$CE,$C3,$7C,$1F,$EE,$FF,$67,$FB
	.dc.b	$15,$3F,$1C,$7F,$00,$E7,$0F,$F7,$6C,$E7,$F4,$0F,$E5,$FF,$B0,$3F
	.dc.b	$5F,$7E,$E5,$85,$FA,$6F,$F3,$4F,$B0,$7A,$CF,$D6,$B0,$BD,$8F,$F1
	.dc.b	$16,$9F,$EB,$7F,$4B,$E3,$3F,$DD,$7D,$3F,$E3,$6F,$D8,$3F,$8D,$9E
	.dc.b	$B2,$F3,$7E,$43,$8F,$F8,$BB,$F9,$2E,$4B,$CB,$F9,$3A,$FE,$42,$7E
	.dc.b	$10,$A7,$53,$F8,$9F,$93,$FE,$05,$D7,$D3,$AF,$F0,$79,$47,$B3,$4C
	.dc.b	$70,$FC,$3E,$1B,$C3,$F0,$AF,$F8,$57,$37,$CE,$E6,$F9,$B1,$83,$E6
	.dc.b	$AE,$4F,$E1,$B1,$F9,$ED,$AB,$FA,$6E,$0B,$C1,$FA,$EE,$0A,$3B,$2F
	.dc.b	$4E,$3A,$7F,$D7,$7D,$5F,$83,$C7,$F8,$69,$D4,$FE,$22,$7C,$33,$27
	.dc.b	$E3,$CF,$E4,$5D,$7F,$51,$1F,$E8,$07,$A3,$FC,$17,$B3,$4F,$FA,$11
	.dc.b	$CB,$C3,$BD,$DD,$A6,$C7,$E8,$4D,$78,$79,$38,$BF,$E8,$EF,$0F,$81
	.dc.b	$89,$FF,$48,$C7,$A4,$FE,$2C,$E7,$C3,$FD,$EF,$07,$FB,$71,$FD,$FE
	.dc.b	$BF,$C5,$2F,$D7,$F8,$4F,$F1,$37,$FB,$F9,$FE,$47,$C3,$F8,$6B,$FE
	.dc.b	$5D,$7F,$C6,$C7,$FB,$1C,$FE,$23,$8E,$5F,$F1,$B1,$93,$F8,$59,$FF
	.dc.b	$00,$3D,$AF,$E1,$2F,$FA,$7B,$B6,$D7,$F8,$3B,$6F,$04,$FF,$8E,$4E
	.dc.b	$07,$FA,$13,$7C,$BF,$E3,$93,$27,$F2,$2F,$0B,$2B,$C4,$FE,$B3,$BF
	.dc.b	$FA,$D6,$34,$70,$3F,$CC,$51,$FD,$7B,$0F,$BB,$EB,$7F,$01,$97,$ED
	.dc.b	$4F,$D2,$9F,$98,$66,$31,$F8,$17,$F8,$CB,$F6,$BF,$F1,$3E,$FF,$E6
	.dc.b	$DF,$DD,$3F,$C4,$DF,$AA,$7F,$8D,$9F,$F1,$47,$ED,$7F,$E3,$8F,$D2
	.dc.b	$1F,$85,$4D,$7F,$D4,$38,$7F,$9D,$67,$3F,$85,$FF,$06,$F3,$3F,$81
	.dc.b	$F0,$FF,$A6,$3C,$DF,$F3,$D3,$E8,$C3,$F7,$6D,$7F,$CF,$C1,$E2,$FF
	.dc.b	$8A,$CF,$DF,$9F,$9C,$BF,$1E,$3F,$BF,$7C,$31,$FE,$13,$CF,$D6,$DF
	.dc.b	$83,$69,$FC,$17,$6F,$F0,$A7,$91,$FE,$8D,$E1,$F9,$1C,$5F,$EC,$BC
	.dc.b	$3F,$53,$1F,$21,$FF,$04,$9F,$AC,$D3,$D0,$FF,$6F,$E7,$FC,$63,$4F
	.dc.b	$C6,$1F,$AB,$7F,$DD,$2F,$EE,$8F,$EC,$5D,$5F,$C0,$1D,$9F,$B7,$3F
	.dc.b	$00,$7E,$CF,$AB,$F0,$FD,$2F,$07,$BC,$FE,$DD,$CF,$83,$88,$FF,$7F
	.dc.b	$BC,$3B,$71,$47,$9E,$58,$3A,$C6,$BC,$0F,$AF,$6E,$FF,$F3,$CF,$EE
	.dc.b	$1F,$E3,$9E,$AF,$F5,$A3,$FE,$62,$EF,$FD,$ED,$7E,$9C,$FE,$83,$87
	.dc.b	$F9,$9B,$1F,$E8,$AF,$F9,$04,$FD,$0C,$FF,$85,$A7,$F8,$BF,$C0,$8F
	.dc.b	$F4,$C3,$FE,$08,$CB,$FD,$85,$1F,$EB,$E8,$FF,$08,$9C,$1F,$F5,$F3
	.dc.b	$83,$93,$CD,$FF,$5F,$1B,$F8,$2F,$F0,$B3,$A8,$FF,$08,$1E,$18,$AF
	.dc.b	$18,$FE,$0E,$74,$8F,$F0,$A3,$E1,$8F,$E0,$E7,$17,$F3,$67,$EF,$76
	.dc.b	$F4,$C7,$38,$FE,$10,$73,$3F,$C2,$67,$8C,$FE,$14,$75,$29,$B3,$FC
	.dc.b	$2A,$7B,$BE,$0F,$FC,$23,$1F,$27,$D7,$AF,$79,$FF,$07,$9F,$04,$FF
	.dc.b	$07,$BF,$3E,$77,$EC,$FF,$0B,$F3,$BF,$AF,$7B,$57,$CF,$7C,$37,$F0
	.dc.b	$FF,$83,$FF,$53,$E5,$EE,$DD,$D9,$FE,$F7,$9F,$F3,$AC,$FF,$90,$23
	.dc.b	$FD,$F5,$EC,$BD,$57,$E6,$7F,$C6,$63,$1E,$AF,$FB,$1B,$C3,$79,$4F
	.dc.b	$C0,$B0,$FF,$AF,$32,$FE,$22,$7F,$5E,$CF,$C6,$FF,$93,$BA,$FC,$51
	.dc.b	$FE,$99,$8F,$F0,$B7,$93,$F4,$BA,$3F,$8B,$3F,$53,$96,$CF,$E2,$F9
	.dc.b	$7F,$D5,$7C,$DF,$16,$3F,$C2,$2F,$E5,$63,$2F,$E4,$23,$F8,$A9,$FE
	.dc.b	$DC,$F8,$6F,$F0,$96,$3F,$0A,$3F,$CE,$99,$7F,$74,$7F,$97,$E9,$FE
	.dc.b	$17,$D7,$DF,$F0,$DF,$C4,$67,$FC,$6B,$C5,$F1,$9F,$F0,$91,$8E,$0D
	.dc.b	$63,$F0,$23,$FD,$01,$C7,$F0,$23,$97,$47,$DA,$9E,$2F,$F8,$67,$9B
	.dc.b	$FC,$2F,$B7,$B6,$7F,$D3,$E7,$EA,$37,$CC,$7B,$63,$F8,$E1,$F9,$53
	.dc.b	$6F,$E3,$DF,$E3,$8F,$E8,$5F,$F0,$AB,$D5,$FF,$4A,$7A,$3F,$A2,$8F
	.dc.b	$F5,$24,$7E,$4C,$FF,$66,$F8,$79,$7F,$D7,$9D,$9F,$5D,$A7,$E8,$CF
	.dc.b	$F6,$9F,$87,$96,$99,$3F,$D4,$5E,$1E,$5A,$64,$FF,$0F,$78,$39,$69
	.dc.b	$91,$E0,$FF,$1C,$C7,$D9,$1F,$E1,$83,$51,$FC,$65,$FD,$91,$FE,$D3
	.dc.b	$F0,$7F,$7F,$EF,$FF,$AB,$3E,$33,$FD,$57,$3F,$A7,$9F,$C1,$47,$F9
	.dc.b	$8E,$3F,$8C,$3F,$E4,$76,$FB,$1F,$E4,$77,$F8,$C8,$8D,$3F,$E2,$03
	.dc.b	$FC,$7E,$FF,$AC,$9F,$F1,$09,$B8,$FF,$48,$3F,$E0,$FC,$BC,$31,$D6
	.dc.b	$3F,$8E,$CF,$47,$FC,$46,$71,$F6,$3F,$C4,$71,$98,$FF,$26,$F8,$5F
	.dc.b	$D4,$DF,$0D,$73,$7F,$8E,$CE,$DF,$F0,$F6,$FE,$0B,$FF,$55,$C7,$96
	.dc.b	$AD,$7C,$68,$FE,$3A,$FA,$DF,$F1,$19,$EB,$F5,$BF,$D5,$F1,$D7,$0F
	.dc.b	$F8,$02,$EE,$7F,$C4,$FE,$57,$BB,$DA,$FF,$13,$7C,$A2,$3F,$D5,$53
	.dc.b	$FC,$67,$A7,$F6,$93,$9F,$C7,$3F,$C5,$9F,$6C,$FB,$D9,$7B,$3F,$CD
	.dc.b	$B1,$9F,$C0,$7F,$C6,$5F,$12,$FB,$D7,$E9,$18,$5F,$21,$FE,$37,$E4
	.dc.b	$FF,$18,$E3,$CB,$2E,$AF,$F1,$93,$18,$FF,$14,$7E,$31,$FE,$AC,$E0
	.dc.b	$F7,$3F,$E7,$EF,$73,$FD,$63,$1F,$EA,$B8,$FF,$53,$B5,$FF,$16,$F5
	.dc.b	$EB,$97,$EB,$47,$F8,$6F,$CB,$FC,$1C,$7F,$AA,$77,$F8,$19,$51,$97
	.dc.b	$F1,$5C,$FF,$C5,$38,$7F,$A3,$B7,$F8,$1B,$FF,$19,$7E,$16,$DF,$C6
	.dc.b	$F1,$E7,$5F,$C6,$9F,$90,$FF,$1C,$E1,$FA,$9E,$17,$F2,$17,$E9,$47
	.dc.b	$F8,$C0,$FC,$D7,$2F,$FA,$BB,$83,$E0,$C7,$ED,$B9,$7D,$2E,$E3,$F1
	.dc.b	$4F,$F1,$03,$F2,$BD,$1F,$E2,$F7,$E5,$9F,$EA,$A8,$DB,$FE,$2F,$3F
	.dc.b	$37,$C1,$FC,$59,$FD,$11,$FC,$48,$E2,$7F,$AA,$CE,$A7,$F8,$03,$B3
	.dc.b	$FD,$4E,$E8,$FF,$56,$77,$7E,$85,$D7,$5F,$A1,$FF,$30,$CF,$F8,$D5
	.dc.b	$E6,$63,$97,$4F,$F4,$98,$FF,$33,$3A,$3F,$8A,$1D,$CF,$E3,$2F,$77
	.dc.b	$AF,$67,$F0,$0E,$CF,$F5,$74,$7E,$E3,$E0,$7F,$D1,$5E,$03,$CA,$93
	.dc.b	$FD,$42,$3F,$C7,$D7,$FD,$67,$E8,$3F,$D6,$73,$FE,$14,$AF,$F4,$C3
	.dc.b	$FB,$03,$FD,$A3,$1F,$E6,$67,$F8,$3F,$7F,$F6,$1F,$15,$C8,$FA,$DF
	.dc.b	$F4,$D4,$63,$93,$FD,$39,$F2,$81,$F8,$3E,$1C,$63,$FB,$33,$FD,$FD
	.dc.b	$F2,$DF,$EA,$3F,$8D,$7C,$5B,$E3,$3F,$55,$B7,$FA,$4E,$FC,$A3,$E2
	.dc.b	$7F,$C6,$EF,$FA,$47,$C5,$F7,$3F,$E9,$13,$FD,$32,$E2,$3F,$BB,$61
	.dc.b	$DB,$1B,$8F,$7B,$FE,$EF,$97,$A7,$3F,$EF,$8F,$AC,$FE,$FD,$AF,$C2
	.dc.b	$DA,$EF,$FC,$56,$FF,$41,$AF,$C9,$7F,$A9,$F9,$9F,$E0,$9E,$2F,$9F
	.dc.b	$1F,$EF,$7E,$3F,$3F,$BD,$EA,$F2,$DF,$2D,$FE,$EC,$AF,$F1,$B3,$FE
	.dc.b	$11,$F8,$C7,$7B,$F3,$AE,$1F,$84,$E6,$3F,$A3,$E0,$8F,$F7,$EA,$F9
	.dc.b	$7D,$4F,$77,$3F,$9E,$63,$2F,$F2,$23,$C3,$E2,$F1,$1F,$EB,$EB,$47
	.dc.b	$FB,$FA,$3F,$C9,$4E,$D0,$CB,$F8,$D3,$FE,$E2,$DF,$CD,$F2,$1F,$D2
	.dc.b	$A7,$FB,$54,$F5,$F9,$EF,$F1,$77,$73,$CA,$7C,$A7,$C5,$9C,$C3,$E1
	.dc.b	$7F,$DD,$5E,$E3,$DC,$7C,$9E,$17,$C2,$FF,$BB,$FF,$2D,$F2,$FC,$5F
	.dc.b	$FB,$C7,$C7,$78,$AF,$95,$C3,$F9,$2E,$0F,$F4,$1F,$B6,$FC,$5B,$E1
	.dc.b	$C7,$F9,$0E,$30,$FC,$0E,$3F,$E3,$B1,$88,$F8,$5D,$4F,$57,$FD,$DD
	.dc.b	$7F,$76,$7F,$9E,$BF,$66,$3F,$37,$E8,$F1,$C3,$E2,$D7,$FB,$CA,$BC
	.dc.b	$2F,$FB,$6D,$FC,$07,$AB,$FE,$51,$D3,$D7,$A5,$D3,$D5,$7F,$C5,$BC
	.dc.b	$EF,$F2,$D6,$FF,$D0,$A3,$FC,$F3,$C5,$FE,$59,$E3,$F9,$5C,$5F,$AC
	.dc.b	$F4,$9C,$FF,$43,$A3,$0F,$D2,$E7,$FE,$69,$C3,$E6,$CB,$F0,$36,$EB
	.dc.b	$FB,$F1,$FC,$B7,$B3,$CE,$D3,$E4,$73,$7C,$57,$07,$C8,$7C,$51,$F8
	.dc.b	$F8,$FE,$5F,$DD,$F0,$9F,$F1,$FB,$D1,$FE,$67,$DF,$9C,$3F,$41,$1D
	.dc.b	$B9,$7D,$9E,$06,$4E,$11,$FC,$D7,$C2,$FA,$8D,$3C,$BE,$17,$ED,$79
	.dc.b	$7F,$D0,$5F,$79,$FC,$BF,$C6,$FF,$40,$3F,$EE,$BB,$FF,$24,$DF,$EE
	.dc.b	$0F,$F6,$F7,$CE,$7E,$D9,$FF,$34,$1F,$E8,$BF,$17,$F6,$CF,$F9,$65
	.dc.b	$F3,$BF,$EC,$73,$5A,$FF,$9B,$5F,$D9,$6D,$FE,$61,$52,$8D,$47,$5D
	.dc.b	$EB,$FE,$40,$78,$4F,$EE,$A3,$FD,$E5,$F1,$3F,$E3,$CF,$83,$FF,$42
	.dc.b	$47,$FB,$59,$FF,$54,$3F,$EC,$7E,$6F,$F5,$17,$8B,$FE,$95,$8F,$DC
	.dc.b	$1F,$C0,$1F,$F4,$B3,$B3,$FD,$7D,$F0,$AF,$89,$FD,$D4,$62,$2B,$FE
	.dc.b	$A8,$F8,$CB,$D9,$FE,$CC,$8F,$E8,$DF,$13,$FE,$CC,$F8,$AF,$A5,$1F
	.dc.b	$C3,$A3,$FD,$62,$FE,$DB,$C5,$F8,$9E,$47,$EC,$75,$FF,$4D,$4F,$F8
	.dc.b	$BE,$7F,$DF,$7F,$13,$FA,$E7,$FD,$95,$1F,$F0,$03,$C8,$E1,$4A,$D4
	.dc.b	$FD,$83,$F3,$BD,$3F,$EC,$3C,$7C,$91,$FE,$BE,$7C,$6F,$F8,$13,$83
	.dc.b	$FD,$D3,$E1,$FF,$B0,$E3,$07,$F2,$6D,$7D,$CE,$0F,$F6,$91,$FE,$13
	.dc.b	$9F,$F6,$23,$4F,$9D,$E1,$3E,$37,$F6,$71,$8F,$53,$FC,$4D,$A7,$D8
	.dc.b	$8F,$F5,$DA,$F8,$EF,$D5,$C7,$AC,$FF,$53,$F4,$7C,$27,$FC,$36,$FE
	.dc.b	$21,$FD,$69,$BF,$0B,$F1,$CF,$FA,$AB,$AF,$F2,$51,$F9,$8B,$12,$7F
	.dc.b	$C5,$9E,$0E,$8E,$93,$FD,$7A,$3F,$A3,$CF,$0F,$8B,$FE,$21,$DF,$FE
	.dc.b	$EF,$8E,$7D,$FF,$D9,$63,$FB,$EF,$DE,$FF,$67,$CF,$FB,$67,$84,$E5
	.dc.b	$FF,$4A,$FA,$BF,$DC,$7C,$1F,$E3,$3A,$7F,$54,$E1,$FE,$F0,$FE,$54
	.dc.b	$3F,$E1,$97,$1E,$C7,$F8,$05,$AF,$F8,$61,$FC,$F9,$FE,$3A,$F4,$7F
	.dc.b	$33,$E7,$7F,$82,$E3,$A3,$FE,$1A,$F4,$3F,$9F,$C7,$F8,$62,$6A,$C7
	.dc.b	$FC,$0D,$DB,$FD,$1D,$FE,$AB,$1F,$93,$EE,$FF,$0C,$3F,$81,$3F,$C4
	.dc.b	$E5,$C3,$B3,$9A,$E2,$FE,$2F,$1F,$53,$5F,$53,$FE,$53,$EA,$FF,$1E
	.dc.b	$7A,$FF,$E3,$B8,$ED,$A7,$C3,$7F,$CB,$BE,$67,$89,$FC,$29,$FD,$FB
	.dc.b	$83,$FC,$36,$FF,$BC,$38,$FE,$AB,$F3,$79,$FF,$93,$3E,$96,$BF,$E0
	.dc.b	$BF,$E0,$0F,$B3,$5F,$E3,$AF,$09,$F9,$B6,$2B,$F6,$8D,$7F,$CD,$F5
	.dc.b	$5F,$E5,$5E,$FF,$E1,$FA,$9D,$5F,$E2,$B8,$C7,$A7,$FE,$2D,$BF,$E4
	.dc.b	$27,$E7,$7A,$1F,$3B,$FC,$90,$C7,$EB,$DF,$C9,$8D,$7E,$C9,$33,$FD
	.dc.b	$45,$23,$F5,$11,$FC,$BC,$FC,$DE,$7F,$AF,$7F,$C4,$A6,$3C,$8F,$F1
	.dc.b	$F7,$D3,$FF,$2A,$F1,$BF,$75,$3F,$E2,$83,$FA,$24,$67,$97,$71,$FC
	.dc.b	$AB,$CD,$AB,$CD,$86,$3C,$97,$95,$FC,$3E,$79,$9F,$C3,$DF,$F1,$33
	.dc.b	$FA,$1F,$2F,$F8,$CB,$F9,$C7,$F8,$CB,$FE,$1D,$E5,$FE,$56,$7F,$03
	.dc.b	$E6,$FA,$B1,$FD,$20,$FF,$12,$C7,$F2,$C7,$EB,$99,$3F,$DB,$2E,$5A
	.dc.b	$7F,$11,$EE,$E6,$FF,$08,$DF,$F8,$43,$FC,$8E,$FC,$AF,$E5,$CF,$E8
	.dc.b	$1F,$2A,$F9,$4F,$F0,$B6,$FF,$CB,$BC,$FB,$5E,$FF,$72,$BF,$CB,$0F
	.dc.b	$F5,$BC,$FF,$2C,$69,$FC,$B3,$CF,$DF,$A3,$FD,$6A,$3F,$3E,$79,$79
	.dc.b	$7F,$94,$BE,$7F,$43,$FC,$4D,$CF,$FC,$88,$FC,$1C,$7F,$88,$9F,$E3
	.dc.b	$FD,$3F,$C9,$8F,$E4,$3D,$5F,$CA,$0F,$E4,$7D,$7F,$CA,$8F,$E4,$BD
	.dc.b	$BF,$9B,$EE,$FF,$19,$3F,$9A,$C3,$F0,$C7,$F9,$BB,$B7,$FA,$BF,$77
	.dc.b	$FA,$27,$C7,$FD,$17,$3F,$FA,$2F,$F2,$5F,$EF,$E7,$E7,$FE,$3F,$FC
	.dc.b	$02,$7E,$87,$F6,$6F,$F0,$19,$FA,$3F,$76,$65,$F9,$E7,$D6,$FF,$83
	.dc.b	$B8,$3F,$5D,$C3,$FE,$97,$7E,$E4,$7F,$86,$FF,$2C,$FC,$E7,$2F,$FA
	.dc.b	$6D,$FC,$D4,$7A,$0F,$F0,$BB,$F5,$DC,$6C,$8F,$EC,$FC,$FF,$9E,$F3
	.dc.b	$7F,$8D,$D7,$E6,$87,$67,$E1,$1F,$85,$D3,$FD,$CF,$B3,$F0,$91,$FE
	.dc.b	$78,$FD,$33,$FC,$06,$7E,$64,$F5,$F8,$07,$D3,$1F,$E9,$8E,$3C,$3F
	.dc.b	$99,$3F,$68,$E1,$FC,$11,$FE,$07,$8F,$F1,$94,$67,$BF,$F3,$53,$EF
	.dc.b	$F2,$18,$F9,$7D,$7F,$C9,$E7,$FA,$EA,$FF,$33,$CF,$FB,$63,$D0,$FF
	.dc.b	$38,$6F,$F1,$67,$EF,$E1,$F1,$5F,$CE,$1D,$63,$E7,$B4,$FF,$01,$1E
	.dc.b	$1E,$2F,$6A,$9F,$E3,$0F,$F1,$6E,$3F,$F3,$F7,$27,$F7,$A3,$EF,$CF
	.dc.b	$FA,$AA,$7F,$8D,$E9,$FE,$83,$E5,$FE,$FE,$7F,$7A,$E6,$FF,$00,$9F
	.dc.b	$DF,$39,$FF,$C0,$67,$F7,$EE,$8F,$CC,$5F,$A9,$FC,$AB,$F8,$0D,$3F
	.dc.b	$9D,$3F,$E8,$86,$BF,$E5,$CE,$5F,$EA,$BF,$05,$FE,$E5,$7F,$AB,$FD
	.dc.b	$25,$9F,$F6,$7F,$A3,$FD,$79,$FF,$36,$4F,$F2,$B8,$0F,$AD,$E1,$3F
	.dc.b	$C3,$F5,$A7,$F9,$19,$F0,$1F,$E9,$AF,$3B,$FC,$18,$FE,$5A,$FF,$06
	.dc.b	$BE,$33,$9F,$A9,$FD,$CA,$3F,$DC,$D3,$FC,$EE,$1F,$16,$FB,$8F,$F7
	.dc.b	$F7,$E5,$9E,$5D,$1F,$1B,$8C,$7D,$71,$FD,$D9,$FD,$F4,$FF,$45,$E5
	.dc.b	$FE,$D3,$1F,$E8,$98,$F1,$F4,$FE,$B6,$3F,$D0,$7F,$92,$6F,$61,$9F
	.dc.b	$0D,$1C,$27,$F6,$3F,$9D,$EE,$B4,$E6,$3F,$22,$D7,$C4,$ED,$FC,$F4
	.dc.b	$E5,$FF,$60,$B9,$7F,$33,$86,$5F,$C9,$39,$7D,$8F,$C3,$CF,$FD,$A0
	.dc.b	$FC,$5C,$FF,$DA,$5B,$FC,$BD,$9E,$DE,$30,$1F,$EF,$A8,$F9,$CF,$EC
	.dc.b	$67,$EB,$F0,$EE,$3F,$45,$A7,$DC,$7F,$5B,$3F,$E5,$43,$F4,$27,$8D
	.dc.b	$E4,$7D,$4E,$3E,$5E,$B7,$F8,$96,$3E,$A4,$FE,$51,$FC,$31,$FE,$4F
	.dc.b	$F6,$BF,$D5,$CF,$E1,$0F,$F0,$2F,$BF,$F8,$8C,$9F,$89,$FE,$06,$7F
	.dc.b	$2F,$AF,$E5,$DF,$E4,$BB,$7D,$BF,$7B,$5A,$FD,$CF,$7B,$EE,$71,$66
	.dc.b	$FE,$D9,$F9,$9A,$FF,$7D,$71,$E6,$3F,$85,$CD,$7E,$89,$FE,$3B,$BF
	.dc.b	$F3,$3B,$F6,$0E,$47,$F1,$BC,$3F,$31,$C1,$F2,$FA,$3F,$3B,$A7,$CD
	.dc.b	$E9,$FF,$3D,$5F,$C8,$3F,$CD,$BC,$1F,$09,$FF,$70,$7E,$57,$FE,$1F
	.dc.b	$F3,$BF,$31,$1F,$D6,$E3,$FC,$45,$F3,$98,$8F,$E3,$11,$87,$7D,$68
	.dc.b	$E5,$1C,$BC,$C7,$E4,$9F,$C2,$BC,$CB,$3E,$DE,$0F,$D3,$7B,$5F,$A4
	.dc.b	$8F,$F3,$A1,$FD,$57,$E4,$87,$F3,$37,$1F,$0B,$FC,$52,$3F,$95,$47
	.dc.b	$33,$85,$ED,$FE,$EB,$1F,$EF,$AB,$FD,$BC,$7E,$5A,$1F,$31,$B8,$FF
	.dc.b	$20,$43,$E6,$7F,$C8,$5D,$BF,$06,$37,$3D,$CE,$B3,$C5,$E9,$F1,$5E
	.dc.b	$5F,$97,$C4,$7A,$B8,$79,$BF,$3A,$D3,$EB,$9D,$C7,$E5,$B2,$F3,$D7
	.dc.b	$C2,$F8,$7C,$9F,$EA,$EE,$CF,$E1,$D4,$8F,$89,$E4,$F6,$AC,$BD,$0F
	.dc.b	$F4,$43,$EE,$EE,$F9,$57,$FE,$90,$F9,$6E,$6F,$F3,$E9,$FA,$4F,$EC
	.dc.b	$5F,$96,$FF,$03,$72,$7F,$B6,$70,$FD,$D7,$4F,$F7,$08,$D9,$FC,$E7
	.dc.b	$2E,$AF,$F0,$6E,$BF,$F5,$B3,$A3,$F8,$47,$8B,$A7,$F9,$3C,$FE,$15
	.dc.b	$FF,$17,$BF,$C1,$CF,$8A,$7F,$09,$9E,$71,$FE,$79,$8C,$78,$5F,$C5
	.dc.b	$A3,$FC,$41,$1F,$C3,$FA,$B8,$E5,$FE,$97,$FA,$2F,$E7,$13,$EA,$CB
	.dc.b	$FC,$C3,$7C,$EB,$E1,$34,$FE,$8B,$E1,$F0,$8F,$95,$1D,$9F,$94,$FA
	.dc.b	$36,$FE,$D7,$6D,$FF,$0A,$D9,$FE,$53,$CB,$FC,$AB,$C6,$F1,$D7,$DB
	.dc.b	$9F,$4F,$17,$23,$F0,$4E,$0F,$D4,$9F,$1B,$E2,$BB,$F2,$71,$8F,$92
	.dc.b	$F8,$39,$7C,$67,$8B,$9B,$CB,$CF,$BA,$F9,$4F,$9B,$A3,$D2,$7A,$3E
	.dc.b	$33,$D7,$D5,$EA,$E5,$F6,$F5,$FB,$BA,$BF,$43,$E7,$7D,$9E,$2F,$F2
	.dc.b	$4E,$7F,$D1,$BB,$7F,$A0,$47,$F0,$D6,$C3,$CC,$FC,$EF,$C5,$FE,$90
	.dc.b	$F2,$47,$EE,$71,$FD,$3E,$5F,$A8,$F8,$1F,$E0,$9A,$FE,$AB,$E0,$BF
	.dc.b	$83,$70,$FE,$B3,$D8,$FE,$11,$CB,$FA,$E3,$E4,$E5,$FC,$1B,$9B,$F4
	.dc.b	$AF,$F0,$5F,$67,$F8,$77,$5F,$ED,$0F,$BB,$1F,$C3,$1F,$BA,$FE,$AC
	.dc.b	$FF,$03,$FA,$1F,$B4,$8F,$D3,$46,$CF,$D3,$3F,$EE,$5E,$FD,$47,$E9
	.dc.b	$4F,$E4,$11,$FD,$D7,$97,$FA,$0F,$85,$F9,$C9,$FE,$31,$3F,$DC,$7C
	.dc.b	$3F,$D3,$73,$7C,$C9,$FE,$4F,$E0,$7F,$7D,$F0,$3F,$B6,$75,$65,$F9
	.dc.b	$BD,$1F,$9E,$C7,$FB,$D7,$CC,$5C,$3F,$69,$E2,$7E,$9D,$FC,$C1,$FE
	.dc.b	$73,$3E,$A1,$FA,$A9,$FF,$04,$3F,$E0,$0B,$FE,$AF,$B0,$78,$9D,$67
	.dc.b	$FC,$7C,$71,$3E,$39,$4F,$ED,$FE,$26,$FD,$17,$96,$F1,$F0,$FE,$9E
	.dc.b	$BC,$A3,$83,$DD,$CB,$FA,$58,$FD,$59,$FE,$45,$3F,$80,$F3,$FE,$01
	.dc.b	$FE,$0C,$FE,$CB,$A3,$F6,$BD,$3F,$B4,$EA,$FD,$BF,$5F,$ED,$BB,$3F
	.dc.b	$73,$DB,$FB,$8E,$EF,$DC,$F7,$CC,$FB,$9F,$E4,$1F,$07,$F3,$6D,$EB
	.dc.b	$FE,$7B,$F0,$7F,$C0,$8F,$F3,$6F,$DE,$FF,$1B,$E8,$FE,$3D,$5F,$E2
	.dc.b	$A1,$F3,$3F,$EC,$EE,$3F,$EE,$55,$FC,$8E,$BF,$A2,$FE,$47,$FD,$9A
	.dc.b	$FF,$80,$9C,$56,$9F,$E7,$A7,$FB,$3C,$7F,$2C,$9F,$F2,$B4,$7F,$65
	.dc.b	$7C,$31,$FE,$3A,$8F,$EB,$EF,$CD,$3F,$02,$7F,$B8,$1C,$C7,$8B,$C7
	.dc.b	$FE,$BF,$0E,$2F,$07,$91,$83,$C5,$E4,$EB,$9B,$C9,$E3,$F0,$3E,$A6
	.dc.b	$5E,$17,$D1,$E5,$74,$3F,$6B,$0C,$7F,$C4,$B9,$FC,$2E,$1F,$DA,$CF
	.dc.b	$C1,$7F,$6A,$FA,$DF,$E7,$2D,$3F,$9C,$9F,$EE,$39,$FE,$81,$DD,$F1
	.dc.b	$27,$F6,$D3,$97,$F3,$EF,$85,$FD,$4B,$93,$F5,$3E,$2F,$CC,$BF,$E2
	.dc.b	$F1,$B3,$FC,$10,$7D,$57,$C6,$FA,$DF,$E0,$EE,$4F,$D8,$1C,$CF,$E0
	.dc.b	$C7,$E1,$47,$E4,$7F,$85,$7F,$C7,$38,$7E,$16,$7E,$DC,$68,$FF,$53
	.dc.b	$9F,$E9,$59,$FF,$6D,$C6,$9F,$BB,$E3,$7F,$8F,$DF,$BF,$D1,$FE,$12
	.dc.b	$5F,$0D,$EC,$6D,$FE,$40,$FF,$40,$B8,$7F,$11,$E6,$F7,$FD,$B3,$87
	.dc.b	$D9,$FF,$B3,$3F,$57,$FF,$2E,$F4,$FF,$93,$5F,$F7,$AD,$FC,$63,$FD
	.dc.b	$16,$E9,$FC,$89,$FA,$5F,$D6,$1F,$89,$7C,$E5,$F9,$47,$FB,$63,$C1
	.dc.b	$CD,$7E,$D0,$F9,$01,$DB,$DC,$7C,$1F,$77,$FD,$74,$F8,$7C,$FB,$E3
	.dc.b	$FF,$B9,$3F,$2F,$6D,$F7,$96,$D9,$E1,$FF,$20,$C7,$C3,$DF,$FB,$E6
	.dc.b	$5E,$28,$C3,$FE,$70,$7E,$8B,$71,$FC,$7C,$7D,$FE,$3F,$F3,$93,$F6
	.dc.b	$39,$3F,$D9,$CF,$E0,$6B,$FD,$A1,$F9,$87,$E0,$66,$9F,$80,$8F,$83
	.dc.b	$1F,$E2,$BE,$5F,$F2,$6B,$FE,$2D,$F3,$7F,$DF,$13,$FE,$BA,$8C,$47
	.dc.b	$FB,$86,$11,$FD,$ED,$FF,$43,$BF,$C6,$6B,$C3,$D3,$FA,$49,$FC,$B4
	.dc.b	$FF,$8C,$BA,$73,$1F,$9A,$3F,$8F,$47,$EC,$1F,$2C,$FE,$C1,$E2,$FE
	.dc.b	$9A,$3F,$D8,$D7,$FE,$B4,$9D,$9F,$EB,$DF,$33,$AB,$FE,$9E,$EC,$F8
	.dc.b	$7E,$7F,$F9,$0F,$2F,$D1,$E1,$FD,$27,$7F,$ED,$0D,$BF,$E8,$C7,$A3
	.dc.b	$BB,$85,$7C,$C7,$F8,$AC,$7C,$27,$FC,$FB,$3F,$E7,$79,$EB,$E0,$63
	.dc.b	$C1,$E3,$7F,$41,$DF,$85,$88,$F8,$BE,$1F,$2F,$5B,$11,$F1,$1E,$BE
	.dc.b	$5E,$23,$EA,$71,$D7,$E9,$F8,$D7,$4F,$E8,$51,$FC,$DF,$9F,$F3,$EB
	.dc.b	$F6,$DF,$C3,$3F,$A2,$F8,$0F,$C3,$37,$9B,$A3,$FC,$FC,$87,$EA,$3D
	.dc.b	$AE,$BC,$FF,$86,$84,$78,$33,$FF,$4B,$BE,$2E,$AF,$F6,$43,$E4,$E3
	.dc.b	$1F,$EF,$FA,$FE,$BD,$1D,$F9,$76,$7F,$2E,$7C,$55,$CB,$DF,$C1,$F9
	.dc.b	$BA,$FE,$45,$3F,$B7,$7C,$DC,$FF,$EB,$0F,$C9,$DF,$80,$7F,$17,$FD
	.dc.b	$33,$F6,$C7,$F3,$4F,$DE,$B3,$7F,$C1,$4F,$D7,$5F,$FA,$E3,$2E,$91
	.dc.b	$FB,$1C,$BF,$A0,$C1,$F1,$47,$F7,$78,$F9,$1A,$FF,$AE,$1F,$4C,$7F
	.dc.b	$BB,$1F,$73,$7D,$CF,$C5,$C7,$C9,$DB,$F9,$BB,$F2,$67,$47,$F8,$98
	.dc.b	$FF,$01,$6D,$FD,$33,$6F,$A7,$BB,$E0,$FE,$FD,$E1,$FF,$7D,$FC,$37
	.dc.b	$FB,$4B,$0F,$AE,$7F,$37,$1F,$96,$A3,$FD,$AF,$83,$D5,$A7,$F7,$05
	.dc.b	$FB,$1F,$FB,$73,$D1,$7E,$49,$7E,$C3,$F8,$F9,$FC,$81,$F9,$53,$F2
	.dc.b	$EE,$65,$FE,$34,$F4,$FF,$2B,$F1,$DE,$18,$F2,$70,$79,$9F,$CD,$1F
	.dc.b	$CE,$1D,$38,$3F,$9D,$E9,$FA,$3D,$15,$7F,$80,$7F,$4D,$AF,$BA,$FF
	.dc.b	$9F,$2F,$D2,$3F,$9D,$1F,$CE,$F7,$FD,$67,$0A,$EE,$1F,$AF,$CF,$C3
	.dc.b	$C7,$FE,$F4,$C9,$F9,$6F,$FB,$FB,$1B,$1F,$37,$4F,$1B,$70,$F9,$58
	.dc.b	$FC,$9A,$F9,$59,$FF,$46,$3F,$A3,$BF,$91,$8F,$C9,$4C,$AF,$90,$7F
	.dc.b	$5A,$7B,$BF,$0C,$FF,$24,$65,$83,$FD,$08,$FD,$63,$FA,$F1,$C3,$2F
	.dc.b	$DB,$FD,$C7,$C9,$FD,$4B,$E3,$FC,$5F,$C8,$FD,$FF,$FA,$F1,$FD,$7D
	.dc.b	$F8,$B1,$F1,$A7,$E2,$6D,$F1,$79,$7F,$84,$FD,$C7,$E3,$FF,$28,$F8
	.dc.b	$8F,$E3,$3F,$4B,$9A,$7A,$8F,$DC,$F9,$5F,$8B,$FC,$93,$F1,$9F,$82
	.dc.b	$7F,$68,$3F,$B4,$BF,$B3,$8F,$DA,$54,$AE,$47,$F9,$9F,$85,$87,$C2
	.dc.b	$8F,$F0,$39,$FE,$EC,$C7,$1E,$57,$F2,$4E,$2D,$CF,$E8,$C7,$DE,$FA
	.dc.b	$9F,$F0,$27,$2E,$1F,$F0,$37,$3B,$C6,$FC,$87,$BB,$F8,$EF,$80,$FC
	.dc.b	$7F,$EA,$9F,$E0,$53,$FC,$0C,$FE,$32,$3F,$1B,$3F,$CE,$76,$E9,$87
	.dc.b	$F9,$CC,$FC,$F6,$DF,$8F,$CF,$F9,$99,$F8,$D3,$F6,$B5,$27,$E0,$83
	.dc.b	$F9,$9D,$FE,$D7,$0F,$DB,$63,$FE,$12,$3F,$C2,$6F,$F1,$D8,$FE,$3D
	.dc.b	$52,$B9,$9F,$AA,$BF,$8E,$FE,$C5,$FE,$15,$C3,$FD,$47,$43,$FB,$1F
	.dc.b	$A5,$FD,$97,$C0,$7E,$C3,$E0,$7F,$63,$FA,$3F,$F8,$9C,$FF,$14,$3F
	.dc.b	$AE,$8F,$D7,$CF,$C3,$AF,$D7,$1F,$D4,$BF,$26,$FE,$31,$F9,$2F,$EB
	.dc.b	$5F,$E2,$BF,$A1,$4D,$3C,$FC,$6F,$9B,$FE,$29,$F9,$27,$F1,$5F,$DE
	.dc.b	$3F,$C6,$27,$F8,$C9,$FE,$79,$1F,$CF,$6E,$5F,$C0,$E6,$63,$F0,$23
	.dc.b	$FD,$61,$95,$7F,$5B,$1F,$CA,$FF,$27,$F3,$47,$E1,$9F,$E2,$5F,$88
	.dc.b	$7F,$10,$FC,$1F,$F8,$8F,$DD,$FF,$91,$CF,$F2,$43,$FC,$3A,$3F,$05
	.dc.b	$3F,$C3,$6F,$F8,$73,$FD,$C8,$FE,$18,$7F,$34,$89,$7F,$03,$F0,$6F
	.dc.b	$E6,$57,$FC,$CF,$0F,$F2,$79,$FE,$50,$7F,$AD,$47,$F5,$BA,$95,$EF
	.dc.b	$BF,$63,$F8,$9E,$C4,$7F,$47,$DB,$EF,$1F,$CB,$7F,$1E,$FE,$57,$F9
	.dc.b	$67,$F2,$CF,$D3,$BF,$CC,$47,$F9,$8D,$FE,$53,$1F,$9D,$9F,$E5,$07
	.dc.b	$E1,$FF,$19,$FE,$9B,$F8,$54,$CB,$EC,$FE,$01,$FD,$2F,$F1,$CF,$E9
	.dc.b	$9F,$6B,$FE,$68,$3F,$CD,$2F,$F7,$38,$FE,$E9,$5F,$EC,$2F,$4C,$FA
	.dc.b	$AF,$FA,$0E,$1F,$D7,$FF,$22,$FF,$63,$C2,$3F,$D9,$3E,$02,$7F,$AE
	.dc.b	$7C,$AC,$AF,$84,$3E,$73,$FC,$53,$DA,$F9,$BF,$F6,$8F,$B1,$FE,$08
	.dc.b	$AF,$F4,$E9,$FE,$8C,$5F,$33,$FE,$D6,$F9,$6B,$3F,$AC,$78,$73,$FF
	.dc.b	$A0,$A3,$FC,$F9,$7E,$BC,$A4,$78,$B2,$FE,$CB,$9F,$FB,$82,$FD,$53
	.dc.b	$FB,$67,$D4,$FE,$E0,$70,$FE,$D4,$FE,$C9,$F3,$1F,$D9,$7E,$93,$FB
	.dc.b	$1F,$E5,$9F,$D9,$3E,$73,$FD,$34,$7F,$A6,$DF,$EC,$31,$FD,$8A,$7F
	.dc.b	$8B,$57,$F6,$13,$FC,$8B,$7F,$E0,$BF,$47,$FB,$03,$F4,$7D,$FF,$F0
	.dc.b	$EC,$7F,$BA,$63,$FC,$1D,$F2,$0F,$8E,$FA,$DF,$13,$A9,$FE,$B4,$7D
	.dc.b	$9D,$57,$62,$1F,$D6,$0F,$E3,$C3,$F5,$07,$E0,$F2,$3F,$CB,$E6,$A3
	.dc.b	$3F,$83,$E2,$38,$FD,$9B,$7F,$B9,$19,$11,$98,$FF,$34,$FC,$17,$F9
	.dc.b	$62,$7F,$97,$4F,$1E,$6C,$FC,$27,$4B,$E9,$5F,$E1,$4F,$8D,$C5,$7E
	.dc.b	$2B,$A2,$CF,$F9,$AD,$E1,$5F,$E7,$6F,$A1,$7C,$86,$AB,$FD,$FD,$18
	.dc.b	$7C,$33,$D9,$F8,$63,$B7,$FD,$BE,$FA,$A3,$DA,$EA,$33,$9F,$B6,$3F
	.dc.b	$DC,$79,$75,$78,$65,$FE,$88,$8C,$4E,$E7,$BC,$7D,$B7,$C7,$3B,$E3
	.dc.b	$F5,$C7,$FB,$BB,$4D,$3D,$0C,$46,$9F,$4C,$76,$C7,$CF,$1F,$7A,$31
	.dc.b	$1F,$ED,$EF,$73,$E2,$4F,$E0,$BB,$7D,$73,$F0,$E3,$FD,$97,$AE,$9F
	.dc.b	$F6,$D3,$DD,$FF,$3B,$6D,$95,$E1,$FF,$BE,$9F,$9B,$5F,$EA,$8E,$6F
	.dc.b	$F5,$5E,$DF,$7E,$7E,$78,$F8,$5F,$C2,$3E,$D3,$E9,$B8,$8F,$91,$1F
	.dc.b	$E1,$33,$E3,$9F,$E5,$B3,$C8,$7F,$10,$3F,$BD,$72,$7E,$73,$97,$FC
	.dc.b	$2D,$1F,$9B,$5C,$BF,$97,$87,$EC,$9E,$57,$F3,$CF,$81,$FF,$60,$FE
	.dc.b	$13,$CD,$FC,$F3,$FA,$47,$D7,$CF,$EA,$8D,$47,$E3,$23,$17,$E0,$8F
	.dc.b	$CE,$1F,$E0,$F9,$E4,$FE,$4B,$AF,$FD,$A3,$1F,$B4,$FC,$33,$F4,$13
	.dc.b	$BA,$FD,$B3,$FE,$B8,$E0,$FD,$DD,$72,$8F,$DE,$F4,$FE,$4E,$3C,$71
	.dc.b	$8E,$2D,$DF,$8E,$3F,$39,$C9,$88,$FF,$6F,$F4,$FE,$8A,$FA,$57,$F8
	.dc.b	$97,$A7,$A5,$78,$A3,$F8,$5C,$63,$7F,$CD,$47,$9E,$6A,$AE,$6F,$FB
	.dc.b	$6B,$A3,$C0,$FE,$70,$F7,$57,$FB,$C3,$A3,$BD,$DC,$FD,$33,$F9,$BA
	.dc.b	$EC,$7E,$33,$A3,$D1,$5E,$D8,$FC,$D9,$A3,$F7,$FD,$9E,$77,$F2,$7D
	.dc.b	$3F,$99,$7C,$D3,$ED,$7F,$C8,$91,$9C,$FD,$3D,$1F,$5A,$BB,$57,$F9
	.dc.b	$26,$31,$C5,$F9,$E8,$FC,$F6,$9E,$E7,$F3,$AF,$A9,$FC,$D1,$E5,$3F
	.dc.b	$3C,$78,$B3,$FC,$CB,$FE,$55,$E5,$FC,$B3,$A7,$F2,$CF,$EB,$3F,$39
	.dc.b	$FF,$4B,$BF,$EE,$B3,$FC,$F2,$39,$BF,$5F,$A7,$F9,$7F,$B3,$CD,$CF
	.dc.b	$FE,$E7,$7F,$1F,$5F,$BC,$38,$F3,$7F,$1D,$AE,$11,$FD,$DF,$9B,$B5
	.dc.b	$7D,$38,$E0,$FC,$1E,$1E,$11,$E2,$AF,$9C,$72,$7F,$98,$EF,$E1,$7F
	.dc.b	$82,$F5,$79,$9F,$4F,$6F,$F3,$6A,$E1,$1F,$1B,$5F,$7B,$FE,$79,$FC
	.dc.b	$E7,$FD,$98,$FF,$5E,$C1,$C3,$FC,$3A,$3F,$B0,$F8,$DF,$7E,$73,$3F
	.dc.b	$D2,$79,$F8,$3B,$AF,$F4,$27,$57,$F4,$4A,$DC,$7E,$3F,$7D,$D7,$BE
	.dc.b	$EC,$FE,$33,$AE,$BF,$D3,$B5,$F7,$D7,$58,$47,$F5,$2C,$FD,$31,$FE
	.dc.b	$93,$E7,$FD,$CB,$F8,$B7,$F5,$2F,$F5,$8E,$EF,$C1,$BF,$8F,$7F,$07
	.dc.b	$CB,$FC,$B1,$CC,$7E,$07,$0F,$F0,$0C,$7E,$23,$95,$D9,$88,$FE,$93
	.dc.b	$5F,$EA,$3E,$0F,$D2,$47,$26,$BF,$D1,$E3,$FD,$51,$CD,$FD,$86,$3B
	.dc.b	$B5,$D4,$7F,$6D,$DF,$B7,$7F,$F0,$E7,$C1,$3F,$89,$3C,$FD,$9F,$DC
	.dc.b	$9F,$CE,$F3,$F3,$AF,$9D,$7F,$EB,$79,$FD,$0F,$37,$92,$7F,$60,$F9
	.dc.b	$A3,$FA,$96,$7E,$0A,$FC,$6E,$B5,$FD,$99,$CE,$7C,$FC,$3F,$35,$F8
	.dc.b	$71,$F5,$8F,$39,$FB,$83,$FA,$77,$07,$ED,$9F,$EA,$1D,$5F,$D3,$7F
	.dc.b	$17,$FA,$31,$FE,$0E,$7F,$6E,$FF,$84,$3C,$AF,$BA,$FF,$64,$7F,$92
	.dc.b	$F1,$FF,$64,$72,$DB,$FE,$1F,$2E,$3F,$B2,$FF,$B6,$6B,$F9,$6C,$62
	.dc.b	$3F,$CF,$B3,$FC,$6D,$71,$FF,$1F,$AF,$0B,$67,$AF,$1F,$F8,$6E,$7C
	.dc.b	$AD,$8F,$0F,$47,$F8,$9A,$7D,$2D,$8F,$2E,$FB,$7F,$C4,$5C,$FF,$E5
	.dc.b	$0A,$FE,$83,$3F,$E1,$C7,$E0,$B6,$7D,$BC,$15,$FE,$85,$DF,$C5,$CD
	.dc.b	$D7,$FA,$26,$7F,$05,$F8,$BC,$BE,$37,$F8,$A6,$3F,$D4,$39,$7C,$95
	.dc.b	$FE,$12,$7D,$9C,$3F,$88,$E8,$E9,$3F,$B9,$3A,$47,$E2,$9A,$7F,$86
	.dc.b	$BA,$7E,$75,$FD,$AE,$BF,$AD,$8F,$F4,$17,$EC,$5F,$E1,$67,$EF,$F3
	.dc.b	$7E,$06,$BE,$DB,$F8,$5E,$6F,$C7,$B5,$FF,$0B,$71,$FF,$20,$C8,$7E
	.dc.b	$EC,$FF,$57,$F6,$7F,$3A,$E2,$FF,$5D,$72,$7F,$94,$F0,$E1,$4D,$F0
	.dc.b	$FF,$44,$DB,$F0,$D3,$AD,$BF,$C6,$6F,$03,$FC,$CB,$97,$F2,$EA,$FE
	.dc.b	$0F,$C7,$FC,$67,$17,$D3,$FF,$93,$E3,$F8,$E6,$58,$AF,$3C,$7F,$28
	.dc.b	$7F,$91,$1F,$E0,$87,$F9,$13,$FE,$72,$F8,$7F,$D1,$B7,$7F,$CF,$F4
	.dc.b	$7F,$98,$BC,$1F,$D1,$F3,$FF,$69,$AF,$F1,$47,$D5,$D7,$FD,$3B,$D4
	.dc.b	$8F,$F0,$FF,$85,$FE,$D2,$F1,$BF,$B9,$57,$F4,$27,$83,$FC,$6B,$6C
	.dc.b	$B4,$FF,$09,$9F,$ED,$5F,$82,$CB,$FE,$87,$78,$BF,$CA,$A3,$3C,$BF
	.dc.b	$06,$3F,$BD,$1F,$E7,$B8,$E3,$BE,$74,$FF,$11,$F5,$B0,$EA,$DB,$F0
	.dc.b	$9F,$E1,$8B,$F0,$F0,$F1,$E0,$FE,$07,$DE,$D3,$CA,$DB,$E2,$7F,$63
	.dc.b	$DB,$CF,$CF,$F1,$63,$FC,$99,$BF,$E5,$F8,$7D,$BC,$FF,$3F,$2F,$CA
	.dc.b	$D7,$FB,$23,$93,$F3,$2D,$8F,$E2,$6F,$FA,$AB,$F7,$EF,$E3,$EF,$FB
	.dc.b	$39,$FE,$A7,$7F,$CB,$E3,$FD,$A3,$CB,$FE,$9C,$EA,$FE,$A5,$A6,$67
	.dc.b	$F3,$BC,$DE,$57,$FD,$95,$E8,$7D,$97,$F9,$B7,$85,$5F,$F6,$75,$FF
	.dc.b	$B0,$78,$F9,$43,$A7,$E9,$4F,$F8,$9B,$C1,$F9,$93,$FE,$D5,$BF,$F2
	.dc.b	$FF,$3F,$F3,$C8,$FF,$5A,$E1,$CB,$AB,$8E,$1F,$D6,$BA,$AE,$1F,$E5
	.dc.b	$5A,$E1,$7F,$EA,$BC,$3C,$3D,$3D,$B0,$FE,$B3,$5E,$4E,$9E,$71,$FD
	.dc.b	$5B,$A7,$CF,$D3,$F5,$F1,$F5,$9F,$C3,$78,$7D,$7D,$3F,$0E,$FF,$DE
	.dc.b	$1C,$3F,$5E,$BE,$19,$FE,$28,$E4,$F1,$BF,$EA,$31,$CB,$FE,$02,$EC
	.dc.b	$FA,$D7,$FD,$D2,$7F,$DC,$3E,$E7,$F0,$6E,$7F,$F6,$7D,$7F,$08,$E7
	.dc.b	$FA,$4F,$F0,$AE,$DF,$33,$FC,$33,$A3,$F3,$CF,$F0,$EE,$8F,$9A,$FF
	.dc.b	$10,$E8,$CE,$5F,$E5,$2E,$8B,$DD,$FE,$65,$E7,$DC,$FF,$95,$B9,$B2
	.dc.b	$FF,$9D,$39,$BF,$8F,$BF,$DC,$6B,$FD,$99,$CB,$8C,$BF,$8C,$71,$DA
	.dc.b	$FF,$23,$9C,$38,$7C,$5F,$59,$FC,$57,$4F,$F4,$63,$67,$FC,$EC,$7F
	.dc.b	$08,$8F,$F1,$A7,$03,$EF,$F5,$3E,$23,$F1,$95,$87,$CB,$C0,$FC,$3F
	.dc.b	$11,$FC,$DB,$83,$0F,$F2,$2C,$7F,$92,$DE,$A3,$FD,$D3,$85,$D3,$FD
	.dc.b	$85,$5F,$C9,$27,$E0,$F9,$BF,$C9,$1B,$DD,$FC,$93,$C6,$FC,$FF,$3F
	.dc.b	$F8,$BF,$A3,$FC,$F7,$F5,$9F,$A2,$E9,$3F,$E3,$DE,$BF,$F4,$A7,$17
	.dc.b	$B7,$2F,$C4,$9F,$CF,$B6,$FC,$4B,$EE,$32,$7F,$B3,$37,$CF,$17,$B2
	.dc.b	$3F,$54,$7F,$BF,$1C,$65,$FE,$94,$E9,$57,$EA,$CE,$BC,$18,$BE,$A7
	.dc.b	$F4,$68,$F0,$F1,$72,$BF,$F1,$8B,$5F,$F4,$64,$7E,$AC,$F3,$F6,$FA
	.dc.b	$EB,$C6,$7F,$A1,$F2,$F5,$F6,$FC,$28,$EE,$1C,$3F,$C3,$8F,$F4,$8F
	.dc.b	$0F,$F0,$AF,$0B,$E6,$47,$F8,$FF,$93,$F9,$0B,$FD,$C9,$FF,$39,$BF
	.dc.b	$96,$FC,$E3,$F9,$0B,$FD,$D9,$FF,$39,$F8,$1F,$58,$F9,$BC,$8F,$BE
	.dc.b	$FA,$C7,$D5,$AF,$C7,$1F,$E3,$6F,$DB,$3F,$8D,$F2,$1F,$AB,$E0,$BF
	.dc.b	$3D,$FE,$3E,$DA,$FF,$2D,$71,$7F,$BE,$BF,$10,$FE,$47,$CE,$7E,$EF
	.dc.b	$3E,$EB,$E3,$9F,$EF,$AC,$3E,$A9,$FC,$AB,$9F,$58,$7F,$B3,$B9,$F9
	.dc.b	$57,$CA,$6B,$FC,$60,$FF,$76,$D7,$F1,$F3,$EA,$E7,$FC,$78,$FA,$7A
	.dc.b	$FF,$1C,$3E,$04,$FE,$B7,$8B,$F9,$6B,$FE,$C2,$E0,$FC,$9F,$67,$EC
	.dc.b	$78,$BF,$8D,$4F,$CF,$CF,$FD,$1B,$D6,$FA,$2F,$F7,$4C,$FE,$DE,$30
	.dc.b	$FF,$89,$8F,$9B,$C8,$F0,$BF,$84,$1F,$EE,$23,$8C,$7F,$82,$BE,$97
	.dc.b	$2F,$0F,$DB,$EB,$FC,$0E,$9F,$BD,$7E,$ED,$7E,$CB,$A9,$FF,$09,$B9
	.dc.b	$FD,$EB,$11,$FD,$7B,$D6,$E1,$FB,$E6,$23,$FC,$F8,$F3,$FD,$6B,$18
	.dc.b	$3F,$43,$ED,$FB,$B6,$0F,$07,$B7,$A3,$F9,$17,$D3,$7F,$90,$F1,$FE
	.dc.b	$21,$D9,$FE,$FE,$E6,$FE,$57,$F7,$EF,$BF,$7D,$D6,$6F,$FC,$9B,$F5
	.dc.b	$C2,$3F,$22,$FF,$86,$1F,$ED,$C7,$FC,$5E,$FF,$36,$F2,$BF,$A4,$3F
	.dc.b	$CE,$3F,$02,$7E,$02,$FC,$0B,$F3,$B7,$FD,$D7,$E5,$3E,$07,$8D,$FD
	.dc.b	$FF,$6F,$F2,$34,$FF,$40,$8F,$F4,$D4,$62,$3F,$CC,$31,$FE,$B3,$7F
	.dc.b	$A2,$1F,$A8,$71,$1B,$8F,$F6,$EF,$77,$CD,$8F,$C1,$C7,$03,$F9,$E3
	.dc.b	$C0,$FF,$7A,$1D,$72,$FF,$34,$BE,$23,$3E,$B7,$F3,$9F,$61,$8F,$D9
	.dc.b	$F3,$F9,$57,$B9,$F3,$6B,$F0,$B4,$C6,$3F,$1F,$2F,$F3,$93,$F2,$DF
	.dc.b	$F3,$ED,$7B,$7C,$2F,$F7,$CB,$EF,$E0,$FA,$87,$29,$FF,$41,$F1,$FD
	.dc.b	$BE,$2F,$F4,$B4,$33,$F8,$56,$E5,$F3,$1F,$86,$9F,$F4,$D7,$37,$F5
	.dc.b	$39,$3E,$BE,$7C,$74,$FB,$7A,$B4,$7B,$FA,$F8,$76,$7C,$1E,$DE,$47
	.dc.b	$C3,$EE,$E9,$DF,$F1,$7C,$0F,$8D,$E0,$CD,$BF,$C2,$C7,$F0,$1E,$BF
	.dc.b	$D8,$BF,$E9,$EF,$17,$EB,$F3,$FF,$84,$FC,$DF,$ED,$1E,$6F,$E1,$F9
	.dc.b	$7F,$D2,$5F,$4B,$F4,$CF,$A9,$DF,$EB,$AF,$ED,$DE,$37,$DD,$F6,$3F
	.dc.b	$22,$FD,$6C,$FF,$81,$9F,$B9,$3F,$D3,$51,$F9,$6F,$1D,$BF,$FA,$77
	.dc.b	$D1,$C7,$15,$BF,$E1,$7C,$75,$FE,$09,$CB,$5F,$F4,$9F,$35,$7F,$CB
	.dc.b	$BD,$15,$FF,$7D,$74,$FF,$AA,$7C,$AF,$03,$9D,$3F,$21,$9F,$ED,$69
	.dc.b	$DD,$FB,$9F,$BC,$C7,$82,$1F,$86,$6F,$C4,$FD,$8F,$E5,$DC,$7A,$3F
	.dc.b	$09,$97,$EC,$BC,$8F,$94,$D7,$C8,$7E,$8D,$FD,$BC,$5C,$FF,$AA,$BE
	.dc.b	$46,$F4,$7A,$FC,$2D,$47,$F8,$9D,$FD,$0F,$92,$C7,$F3,$13,$E7,$D3
	.dc.b	$CE,$FF,$BA,$FB,$7F,$8A,$F9,$6D,$FF,$37,$9F,$A2,$BF,$2E,$7F,$8A
	.dc.b	$F8,$FF,$5D,$C7,$E1,$E4,$F1,$72,$F8,$F9,$BC,$9C,$FE,$5E,$8F,$37
	.dc.b	$4F,$9F,$2F,$D0,$1F,$80,$3F,$B8,$F7,$FE,$1D,$BB,$FE,$2B,$C5,$FC
	.dc.b	$AC,$3B,$7F,$58,$FF,$4D,$3F,$C8,$DF,$58,$F8,$BF,$E7,$8D,$8F,$CF
	.dc.b	$9F,$ED,$6F,$7F,$F9,$E9,$FD,$27,$F5,$8F,$E7,$FF,$55,$FD,$2F,$E1
	.dc.b	$3F,$A0,$9F,$D3,$7F,$40,$FE,$87,$F2,$DF,$D3,$FE,$33,$FA,$29,$FD
	.dc.b	$47,$87,$FA,$3F,$D1,$FF,$53,$FD,$FF,$FA,$49,$ED,$D3,$FA,$C3,$FD
	.dc.b	$5F,$F2,$0F,$E9,$A7,$F0,$38,$FE,$B9,$D3,$F4,$BD,$2F,$C9,$C7,$F0
	.dc.b	$EE,$E7,$E0,$78,$8F,$E2,$7F,$84,$DF,$57,$F8,$AC,$7F,$6E,$C3,$FC
	.dc.b	$A4,$7F,$88,$0F,$E3,$CF,$E8,$00,$F6,$7F,$D2,$11,$FD,$27,$C5,$FE
	.dc.b	$91,$3C,$CD,$F9,$5F,$D3,$3F,$05,$FD,$B7,$5F,$FA,$1B,$1F,$EB,$19
	.dc.b	$62,$FF,$AD,$C7,$F5,$A7,$FB,$C4,$F8,$BC,$EF,$B3,$1E,$13,$FC,$08
	.dc.b	$7F,$81,$A3,$FB,$E7,$E8,$FF,$DC,$8E,$07,$F3,$9F,$57,$F8,$37,$E1
	.dc.b	$9C,$CC,$57,$F8,$EB,$D5,$FE,$F0,$FE,$14,$FF,$4D,$AF,$A4,$FE,$C3
	.dc.b	$F9,$57,$FA,$0F,$D2,$DD,$7F,$55,$3F,$B1,$79,$59,$E2,$FE,$CE,$6F
	.dc.b	$4F,$EB,$87,$F5,$D7,$FD,$19,$E6,$7F,$3A,$AF,$01,$F3,$9E,$A7,$F6
	.dc.b	$53,$BF,$8D,$FD,$7B,$1F,$EB,$99,$7F,$84,$FD,$EC,$DF,$C4,$3F,$79
	.dc.b	$AF,$B9,$A7,$F5,$8C,$3F,$A8,$3F,$E5,$AF,$45,$C1,$B8,$FF,$5D,$C6
	.dc.b	$1E,$13,$FD,$9E,$7F,$AA,$C7,$F8,$60,$E5,$A7,$17,$FB,$2B,$F8,$AA
	.dc.b	$FB,$3C,$5F,$80,$9D,$9F,$80,$7F,$82,$BF,$88,$E6,$FC,$5B,$FC,$57
	.dc.b	$6F,$C7,$72,$EA,$3F,$05,$C7,$70,$FE,$D3,$5F,$EF,$23,$FD,$16,$7E
	.dc.b	$3F,$07,$2F,$17,$F4,$37,$D4,$FD,$47,$57,$5F,$57,$FC,$61,$D5,$FE
	.dc.b	$05,$EC,$FF,$04,$1F,$B1,$AF,$8B,$DB,$FB,$76,$9F,$0C,$FE,$58,$FF
	.dc.b	$83,$FB,$BF,$BC,$77,$FF,$84,$7C,$0F,$EF,$47,$F8,$4F,$C1,$FE,$F9
	.dc.b	$E1,$7F,$85,$7C,$3F,$EF,$C7,$F8,$5F,$C4,$FF,$00,$78,$BF,$E1,$9F
	.dc.b	$1B,$FC,$04,$7F,$94,$FF,$03,$FE,$31,$AF,$E4,$CF,$DB,$B8,$F0,$3F
	.dc.b	$DD,$30,$F2,$FF,$57,$18,$6C,$7E,$24,$FF,$7D,$F8,$AF,$DF,$FF,$BD
	.dc.b	$F8,$7F,$DF,$3E,$93,$F4,$5F,$BE,$7E,$8F,$F5,$8F,$CD,$9F,$9C,$7F
	.dc.b	$C6,$11,$FE,$31,$9F,$F0,$C5,$7F,$BA,$FF,$19,$FC,$B6,$7E,$46,$E5
	.dc.b	$E1,$3F,$87,$FA,$0F,$C1,$3F,$CE,$0B,$F6,$2B,$F8,$57,$F8,$23,$F2
	.dc.b	$2C,$9F,$E0,$AF,$C8,$F7,$F8,$7F,$F8,$33,$F3,$3E,$27,$F8,$3B,$F3
	.dc.b	$9E,$7E,$1C,$D3,$EC,$1F,$C7,$8F,$E9,$1E,$C7,$DF,$7C,$7E,$76,$4F
	.dc.b	$F2,$C7,$D6,$EE,$3F,$BF,$FB,$1F,$E7,$67,$FD,$8B,$19,$F1,$FF,$17
	.dc.b	$18,$3F,$29,$E4,$FF,$9B,$A7,$97,$93,$F9,$49,$F1,$47,$5F,$80,$FF
	.dc.b	$04,$7D,$06,$E7,$FE,$00,$CF,$F7,$9C,$BF,$77,$1E,$38,$F2,$7B,$5F
	.dc.b	$CE,$3C,$FE,$13,$FA,$DF,$8D,$AB,$DB,$FE,$6C,$FA,$27,$FD,$3D,$7F
	.dc.b	$C2,$F3,$FD,$C4,$F5,$9F,$D6,$47,$A5,$FE,$8F,$3F,$C4,$70,$F0,$57
	.dc.b	$33,$E6,$9F,$D2,$A1,$F1,$FD,$8E,$27,$F0,$98,$43,$EB,$3A,$99,$5E
	.dc.b	$E7,$98,$FA,$D3,$C3,$F3,$8F,$F5,$B7,$E2,$EF,$8B,$FB,$1B,$FC,$6A
	.dc.b	$FB,$D3,$F2,$23,$FD,$25,$F2,$9F,$6F,$C2,$F4,$D7,$DD,$8F,$F4,$8C
	.dc.b	$F1,$8F,$EC,$71,$F9,$8D,$BF,$D4,$FF,$0F,$67,$F0,$81,$E3,$5A,$8F
	.dc.b	$F6,$57,$CE,$78,$63,$A4,$23,$D0,$FC,$81,$F7,$0D,$3F,$DA,$4F,$C2
	.dc.b	$8F,$E9,$26,$B5,$F8,$71,$F2,$63,$FB,$CF,$0F,$B1,$FC,$F4,$65,$7B
	.dc.b	$7F,$EB,$38,$FA,$B1,$C0,$FD,$A1,$E2,$E2,$FF,$02,$C6,$A3,$F8,$81
	.dc.b	$E6,$D3,$E7,$E7,$FE,$CA,$9F,$C2,$4F,$27,$FA,$B5,$F7,$AF,$D1,$CF
	.dc.b	$C8,$9F,$C8,$71,$FF,$80,$E3,$CF,$3F,$01,$EA,$6C,$FF,$09,$72,$F7
	.dc.b	$7F,$A3,$99,$3D,$51,$97,$FA,$D9,$93,$FC,$BD,$CD,$97,$FC,$C1,$CF
	.dc.b	$6F,$F4,$31,$F2,$0F,$D0,$9C,$7A,$7F,$D7,$8D,$EA,$FF,$18,$BF,$E5
	.dc.b	$6D,$3C,$EF,$F0,$7F,$37,$F9,$A3,$F2,$4F,$E2,$0F,$F0,$97,$FD,$89
	.dc.b	$1F,$84,$8C,$72,$FD,$99,$FF,$58,$77,$7F,$17,$8C,$47,$F3,$DE,$FE
	.dc.b	$EF,$47,$4F,$DA,$A7,$81,$A8,$FF,$5D,$78,$3A,$DF,$FE,$00,$D3,$C2
	.dc.b	$D5,$7F,$49,$74,$FF,$9C,$BC,$3C,$CF,$F3,$58,$FC,$9B,$FC,$E5,$FC
	.dc.b	$AF,$E3,$7F,$CF,$9F,$D2,$87,$F5,$0F,$F2,$BF,$1B,$FC,$7C,$FC,$2D
	.dc.b	$BF,$93,$7D,$47,$8B,$B9,$F8,$35,$F8,$8F,$2F,$80,$6C,$7A,$3A,$97
	.dc.b	$E6,$ED,$FE,$A5,$97,$B3,$4F,$DA,$BF,$B8,$FD,$B9,$C5,$FE,$8A,$6C
	.dc.b	$7E,$BB,$67,$CB,$FF,$54,$3F,$D1,$EB,$F6,$5F,$EA,$E7,$FB,$73,$F2
	.dc.b	$7F,$EB,$3F,$9F,$FF,$B7,$BF,$12,$FE,$B6,$7F,$B8,$3B,$BF,$AE,$FD
	.dc.b	$37,$FB,$8B,$F2,$5F,$EB,$E7,$FB,$93,$F7,$3F,$EC,$3F,$81,$FF,$B9
	.dc.b	$BE,$63,$FB,$19,$FE,$11,$DB,$FB,$3B,$FE,$EA,$F8,$BF,$EC,$E7,$F2
	.dc.b	$4D,$BE,$D4,$FF,$87,$8F,$E2,$AE,$1F,$EB,$3F,$15,$FE,$F0,$EE,$FD
	.dc.b	$9C,$7D,$37,$F3,$ED,$3F,$D0,$7E,$37,$F5,$47,$9C,$7F,$69,$8F,$F6
	.dc.b	$44,$60,$FD,$79,$C4,$FF,$1C,$99,$3F,$A9,$7C,$67,$F9,$01,$FE,$E5
	.dc.b	$7E,$5F,$4F,$FA,$AC,$7E,$17,$9B,$CA,$E9,$F2,$B4,$EE,$53,$B9,$C4
	.dc.b	$FE,$59,$2F,$D4,$3F,$C1,$A7,$FA,$81,$C4,$7F,$08,$3E,$3F,$ED,$99
	.dc.b	$3F,$A9,$7C,$57,$F6,$63,$FA,$8F,$AA,$FA,$2F,$EA,$3F,$8D,$7E,$A3
	.dc.b	$D6,$79,$1F,$D5,$0C,$70,$7F,$56,$35,$AF,$F4,$43,$FA,$2B,$FB,$0F
	.dc.b	$5B,$EB,$79,$BF,$BE,$79,$EB,$FD,$15,$AF,$F3,$25,$8F,$E8,$72,$FE
	.dc.b	$8F,$A4,$AF,$F5,$DD,$FF,$AF,$3D,$4F,$DF,$ED,$FD,$C7,$CB,$3F,$EC
	.dc.b	$27,$E1,$D7,$CA,$3F,$8B,$F0,$E4,$FE,$06,$FF,$1D,$D7,$F8,$19,$CC
	.dc.b	$E4,$7D,$87,$F8,$79,$FE,$81,$3F,$DF,$7A,$FF,$3A,$69,$FE,$83,$E6
	.dc.b	$FE,$27,$1E,$53,$C8,$7E,$3F,$95,$E9,$6A,$BF,$2B,$03,$0F,$FB,$88
	.dc.b	$EA,$3B,$7F,$C6,$0F,$FB,$7B,$2F,$D2,$7A,$5F,$CF,$23,$F9,$C6,$1F
	.dc.b	$E4,$58,$FF,$1D,$C7,$F3,$BC,$7E,$0B,$FC,$E9,$FF,$09,$D3,$AB,$F6
	.dc.b	$D3,$FE,$71,$7F,$93,$9C,$1F,$07,$5F,$F2,$AE,$CC,$47,$F9,$DF,$B7
	.dc.b	$F9,$69,$8B,$77,$7F,$30,$D3,$1D,$FF,$CC,$8D,$78,$1F,$CD,$3C,$1E
	.dc.b	$1E,$17,$F3,$63,$97,$87,$33,$FC,$19,$FC,$D0,$7A,$7F,$C8,$17,$4F
	.dc.b	$FA,$2A,$30,$FF,$50,$F0,$5F,$94,$7F,$46,$F0,$9F,$1B,$7E,$73,$F9
	.dc.b	$C7,$E9,$9F,$CE,$7E,$9F,$F9,$B9,$FC,$E1,$FE,$6D,$1F,$CD,$E7,$FA
	.dc.b	$5D,$FF,$50,$CF,$FA,$3F,$BB,$FE,$9A,$7F,$8D,$1F,$E9,$CE,$AF,$A6
	.dc.b	$FA,$78,$FF,$BB,$BF,$D3,$9F,$BA,$FF,$8D,$3C,$87,$93,$FC,$3E,$3F
	.dc.b	$A9,$C7,$F0,$E7,$FC,$F9,$EA,$FE,$5D,$FF,$21,$D7,$27,$FC,$37,$1F
	.dc.b	$1F,$C4,$FA,$F1,$FE,$23,$F3,$BF,$B9,$BF,$E2,$D6,$FB,$7A,$3F,$CA
	.dc.b	$F1,$FE,$3B,$3F,$C7,$39,$FF,$9E,$B8,$FF,$BD,$FE,$40,$FC,$3B,$F0
	.dc.b	$7F,$8E,$DB,$FF,$1F,$FE,$3B,$B7,$FC,$6B,$7C,$DF,$F1,$9B,$FE,$45
	.dc.b	$6E,$3F,$E6,$68,$F2,$47,$0C,$FF,$CE,$11,$FE,$3B,$3F,$C1,$67,$BA
	.dc.b	$3F,$CB,$6F,$F8,$D3,$E0,$BF,$BC,$BF,$E3,$6D,$AC,$FE,$E4,$FF,$35
	.dc.b	$1F,$E3,$22,$83,$F6,$46,$0F,$D9,$3F,$E3,$F7,$F2,$67,$DE,$3F,$C3
	.dc.b	$27,$1E,$1F,$BE,$E8,$FF,$5C,$6B,$F9,$9D,$FF,$D7,$5F,$01,$FE,$2A
	.dc.b	$9F,$F0,$7C,$7E,$2F,$2C,$FB,$9F,$E7,$09,$FF,$31,$56,$DF,$F4,$15
	.dc.b	$EA,$7D,$F5,$67,$E1,$DF,$F8,$93,$2F,$F2,$DD,$73,$F5,$FF,$C4,$FC
	.dc.b	$2F,$C3,$7F,$C5,$3C,$37,$1F,$F3,$E6,$1F,$E1,$47,$FC,$FC,$7F,$B2
	.dc.b	$FF,$7B,$7E,$6C,$FC,$EF,$F4,$12,$9C,$FE,$07,$8F,$A6,$FF,$1B,$7A
	.dc.b	$57,$D0,$FE,$F9,$7F,$BB,$EC,$7A,$1F,$EB,$F8,$FD,$54,$63,$F2,$9F
	.dc.b	$E0,$1F,$55,$F9,$A3,$F2,$BC,$5F,$C4,$B9,$BF,$D2,$77,$FE,$7E,$E3
	.dc.b	$F5,$3F,$E8,$0E,$2B,$C5,$FE,$52,$EC,$DB,$F8,$AC,$3F,$8D,$AC,$69
	.dc.b	$F9,$3B,$FF,$78,$67,$E6,$EF,$FC,$5C,$7E,$1E,$3C,$CF,$E4,$8F,$F7
	.dc.b	$B7,$07,$C0,$70,$7E,$35,$FF,$2C,$6F,$CA,$3E,$47,$27,$CB,$76,$63
	.dc.b	$8B,$F9,$07,$DC,$FF,$0F,$FC,$0B,$FC,$BF,$C2,$FA,$DF,$B7,$F8,$1F
	.dc.b	$DA,$3F,$CC,$7C,$39,$C7,$ED,$DF,$E0,$F0,$FE,$31,$1F,$9F,$1F,$B0
	.dc.b	$BE,$09,$F9,$77,$DC,$8C,$7A,$BF,$7B,$F4,$F7,$E3,$93,$0F,$E4,$11
	.dc.b	$8F,$99,$FC,$1F,$E0,$9F,$C3,$A7,$B7,$D2,$7F,$BE,$FF,$35,$FC,$3B
	.dc.b	$FE,$2C,$8F,$F3,$7E,$BF,$C5,$8F,$E2,$3B,$7D,$DE,$1F,$E9,$FD,$AE
	.dc.b	$2F,$E1,$E7,$F5,$2E,$2F,$76,$1F,$C3,$2F,$ED,$E1,$23,$FA,$4F,$8D
	.dc.b	$F3,$F2,$F1,$47,$CE,$7C,$4F,$EF,$F4,$FE,$1A,$FC,$5D,$BD,$FB,$FD
	.dc.b	$77,$DD,$1F,$5B,$87,$E4,$71,$7D,$B7,$E3,$C7,$F4,$3F,$C5,$7F,$70
	.dc.b	$FF,$59,$FC,$12,$BF,$C1,$5A,$FF,$5A,$3E,$A4,$61,$7C,$97,$F4,$FF
	.dc.b	$C1,$BF,$03,$F2,$7F,$BA,$D3,$F7,$27,$E0,$1F,$E0,$11,$FC,$06,$7F
	.dc.b	$AB,$7D,$CF,$F3,$CF,$C1,$FF,$51,$EA,$7E,$A7,$F4,$FF,$94,$7C,$B7
	.dc.b	$F4,$71,$FA,$49,$FE,$87,$F4,$DF,$D1,$3F,$0A,$FD,$87,$A5,$FB,$1F
	.dc.b	$DF,$FE,$91,$F4,$DF,$D5,$C7,$EB,$27,$FD,$7B,$CD,$FE,$AC,$FC,$3B
	.dc.b	$F6,$FE,$98,$FD,$CF,$0F,$E4,$E3,$31,$98,$FC,$B1,$FC,$D7,$2C,$FD
	.dc.b	$07,$FA,$D9,$DC,$DD,$FF,$C3,$70,$FB,$9A,$7F,$B1,$9F,$F5,$95,$FF
	.dc.b	$3F,$AF,$15,$F3,$F8,$1E,$3F,$83,$72,$FD,$83,$D3,$7E,$21,$82,$3F
	.dc.b	$DB,$3E,$47,$8F,$F6,$6F,$D6,$65,$ED,$E7,$C6,$1D,$E3,$F2,$EF,$EC
	.dc.b	$70,$F6,$C7,$8A,$3B,$47,$C0,$FD,$1D,$F8,$CB,$F2,$9F,$81,$5F,$B9
	.dc.b	$9C,$FE,$4F,$88,$FE,$B5,$F9,$0B,$F2,$0E,$31,$D0,$FE,$01,$19,$FC
	.dc.b	$6F,$84,$3F,$3E,$F1,$3F,$43,$A3,$E9,$2F,$CE,$5F,$24,$FA,$4F,$F7
	.dc.b	$3D,$6A,$73,$F6,$58,$E1,$FE,$1D,$F3,$28,$E5,$8F,$CB,$BE,$B3,$9F
	.dc.b	$AC,$E7,$D5,$E6,$DF,$D3,$7E,$38,$C6,$DE,$0F,$4F,$F8,$D6,$9E,$4D
	.dc.b	$7C,$7C,$DF,$E5,$CD,$1F,$49,$AE,$0F,$46,$7E,$1F,$D6,$71,$87,$A9
	.dc.b	$D7,$07,$B3,$C1,$FB,$7C,$5F,$0A,$FF,$94,$64,$9C,$E7,$B9,$FB,$DA
	.dc.b	$FE,$1D,$FF,$80,$1F,$96,$7F,$9B,$BE,$B0,$F4,$C7,$59,$73,$AF,$C3
	.dc.b	$70,$E3,$2D,$47,$FC,$01,$F8,$FF,$F3,$BE,$C7,$31,$FA,$6F,$90,$FD
	.dc.b	$2B,$F9,$77,$FB,$2F,$E3,$2F,$C6,$1F,$9F,$79,$67,$FB,$4F,$D2,$7F
	.dc.b	$A4,$FF,$0F,$C7,$0F,$FA,$12,$7F,$96,$CE,$71,$E9,$C5,$FE,$9D,$BE
	.dc.b	$5F,$83,$7F,$A2,$BC,$AF,$CD,$F1,$7E,$F2,$FF,$C5,$9F,$5D,$FE,$FB
	.dc.b	$8F,$C6,$C7,$F5,$1F,$47,$7C,$7F,$1F,$AB,$55,$C6,$99,$E9,$FD,$94
	.dc.b	$32,$FF,$4C,$BA,$8F,$F5,$6E,$7F,$46,$3D,$3A,$23,$1B,$70,$CB,$D5
	.dc.b	$7D,$1F,$EC,$9C,$3F,$1F,$83,$94,$7C,$33,$FB,$47,$DB,$7C,$3A,$FA
	.dc.b	$11,$FE,$38,$3E,$61,$FD,$B7,$E3,$B9,$47,$D3,$C3,$AF,$1F,$D2,$8F
	.dc.b	$96,$7E,$00,$FB,$1B,$F9,$1F,$B9,$E7,$7F,$AD,$B8,$BF,$67,$BF,$CA
	.dc.b	$9F,$9B,$F5,$FF,$EB,$AB,$FF,$77,$CE,$67,$F4,$2F,$C6,$BF,$8F,$3D
	.dc.b	$67,$F8,$BF,$17,$CA,$CB,$CB,$B7,$C0,$9F,$F3,$26,$9E,$A7,$F8,$BE
	.dc.b	$BE,$F7,$E4,$1F,$E6,$9D,$BE,$23,$FE,$0D,$FC,$5F,$FC,$58,$FB,$4F
	.dc.b	$E7,$6E,$1F,$E8,$5A,$7E,$B5,$FF,$39,$47,$F3,$67,$31,$FA,$B8,$C3
	.dc.b	$ED,$5F,$67,$FA,$29,$F0,$3F,$ED,$C8,$FF,$4A,$47,$23,$F9,$D3,$89
	.dc.b	$E5,$F3,$1C,$67,$C2,$F1,$FC,$8B,$10,$A4,$75,$FC,$87,$30,$7C,$1F
	.dc.b	$8F,$E0,$B5,$FE,$83,$3E,$0B,$FE,$EA,$97,$E3,$7C,$8F,$E1,$FB,$FB
	.dc.b	$DF,$F7,$AF,$E2,$3F,$80,$CF,$F4,$8F,$F3,$4D,$7E,$7E,$3F,$E9,$98
	.dc.b	$C7,$48,$FE,$24,$FF,$3A,$E8,$FB,$53,$FE,$41,$8C,$74,$7F,$05,$8F
	.dc.b	$EB,$D1,$FD,$03,$DA,$FC,$53,$FD,$0A,$FB,$CA,$31,$1D,$E3,$8E,$3F
	.dc.b	$A1,$BF,$CF,$47,$F7,$AA,$FE,$97,$C7,$FC,$4E,$7C,$BC,$7E,$E3,$F2
	.dc.b	$35,$E2,$7F,$4F,$1A,$8F,$DE,$9E,$6D,$7E,$29,$F9,$B7,$E1,$72,$D7
	.dc.b	$CD,$F2,$FF,$6B,$1F,$F1,$74,$7E,$42,$BE,$2F,$66,$4F,$E9,$1F,$27
	.dc.b	$FD,$03,$2F,$E7,$5F,$1B,$FC,$73,$5B,$86,$4F,$F0,$27,$B1,$F5,$6B
	.dc.b	$FA,$03,$FC,$CA,$3F,$C3,$2B,$EF,$FF,$CF,$3E,$0F,$F9,$D8,$FF,$3B
	.dc.b	$BF,$E7,$8A,$E9,$5F,$E2,$0F,$8D,$E8,$FF,$84,$7E,$67,$F9,$13,$FE
	.dc.b	$24,$7C,$8F,$F8,$57,$93,$F9,$4D,$7F,$51,$FC,$43,$F9,$34,$FE,$C4
	.dc.b	$FF,$40,$38,$7F,$48,$7F,$8B,$9F,$8B,$9F,$F8,$D7,$7E,$5C,$3F,$A2
	.dc.b	$9F,$E5,$D5,$F8,$DE,$5F,$E5,$8F,$F8,$96,$3F,$97,$F8,$9E,$1B,$F5
	.dc.b	$CF,$83,$1F,$A9,$5F,$D5,$39,$7F,$0B,$39,$3F,$DE,$5C,$FF,$85,$7A
	.dc.b	$57,$27,$FD,$ED,$5F,$8A,$AE,$31,$8C,$FF,$21,$1F,$05,$EE,$7F,$A7
	.dc.b	$F2,$F4,$4F,$9E,$3F,$8D,$3F,$05,$FC,$8B,$4F,$1C,$FE,$95,$C0,$E8
	.dc.b	$FD,$1D,$FE,$C2,$FC,$13,$F3,$C0,$FC,$6F,$F7,$EF,$F8,$18,$FF,$26
	.dc.b	$38,$7F,$91,$E3,$FE,$09,$8C,$63,$FE,$86,$5D,$9F,$EB,$9D,$BF,$C1
	.dc.b	$D7,$DA,$72,$F8,$23,$C3,$F4,$3C,$EB,$E5,$4F,$8D,$FE,$3B,$5F,$C6
	.dc.b	$63,$D0,$79,$FB,$B3,$DF,$C2,$7F,$90,$CF,$F9,$6F,$1F,$37,$81,$93
	.dc.b	$9F,$D2,$6C,$EB,$E1,$71,$3B,$F9,$FC,$CF,$0F,$CE,$F5,$3F,$97,$FA
	.dc.b	$DF,$A9,$C7,$FD,$AC,$FF,$9F,$9F,$F3,$2F,$E0,$DF,$E7,$EF,$CB,$7F
	.dc.b	$CF,$FF,$8D,$7F,$A0,$3F,$18,$FF,$40,$FE,$A1,$FE,$82,$FC,$9B,$FD
	.dc.b	$07,$F8,$2C,$C7,$F9,$91,$FE,$71,$7E,$52,$E5,$FF,$46,$79,$5F,$CC
	.dc.b	$A3,$FC,$E6,$FF,$A3,$3D,$3F,$F0,$0F,$57,$E8,$27,$FD,$2B,$F0,$C9
	.dc.b	$CC,$FF,$A5,$5F,$C5,$CF,$F9,$FD,$FE,$36,$FF,$9F,$9F,$C8,$B9,$3F
	.dc.b	$8E,$9D,$1F,$E5,$5E,$7F,$38,$FE,$E7,$DF,$F9,$37,$FD,$15,$C5,$D9
	.dc.b	$FF,$53,$3D,$9F,$2F,$CC,$FF,$23,$C7,$FD,$97,$1D,$8F,$F4,$91,$EF
	.dc.b	$3F,$D0,$1F,$83,$7E,$1B,$E3,$3F,$81,$F5,$7F,$8C,$D7,$D1,$FE,$2A
	.dc.b	$B3,$E0,$FE,$4B,$5C,$67,$F8,$EF,$07,$FB,$F6,$3F,$E1,$37,$E3,$5F
	.dc.b	$E3,$9E,$2F,$FA,$B7,$F0,$3E,$AF,$E4,$9C,$BF,$EC,$2F,$21,$A7,$E6
	.dc.b	$BC,$CF,$F2,$7F,$91,$FE,$6F,$7D,$B5,$FC,$57,$C9,$C7,$0F,$F4,$39
	.dc.b	$FF,$08,$79,$3F,$D8,$AB,$F6,$B1,$8E,$3F,$99,$18,$DF,$E8,$BF,$EC
	.dc.b	$19,$F4,$AE,$6F,$E2,$CF,$FB,$5A,$33,$CD,$FD,$99,$DE,$BF,$E3,$37
	.dc.b	$8E,$3E,$1A,$46,$31,$F0,$C6,$31,$FC,$7E,$7E,$D7,$F1,$B1,$E3,$CB
	.dc.b	$EC,$3F,$E6,$B7,$E2,$74,$7E,$72,$3E,$23,$F9,$EE,$8B,$1F,$BE,$C5
	.dc.b	$D3,$F2,$FA,$BE,$67,$5F,$CD,$EC,$F9,$DD,$BF,$3F,$BB,$E8,$75,$4C
	.dc.b	$BF,$81,$1F,$ED,$8F,$E0,$0F,$F7,$77,$E8,$3F,$EE,$FF,$D0,$BF,$DE
	.dc.b	$1F,$05,$E8,$AF,$D3,$BF,$95,$7F,$DC,$4F,$F3,$5E,$7F,$F0,$54,$68
	.dc.b	$FC,$FF,$2F,$FB,$FF,$C4,$E1,$1F,$CB,$3C,$27,$85,$C6,$FF,$DA,$7D
	.dc.b	$1F,$E3,$88,$FF,$7B,$D7,$98,$FE,$66,$7F,$34,$5F,$B3,$FF,$8F,$BB
	.dc.b	$EC,$EC,$DB,$C9,$FE,$CE,$72,$EC,$FE,$05,$E0,$FF,$96,$A3,$F0,$D7
	.dc.b	$E3,$7F,$11,$9F,$A1,$FB,$FA,$FF,$56,$7F,$18,$ED,$FC,$53,$BE,$83
	.dc.b	$F2,$63,$FD,$51,$BF,$F9,$52,$5C,$17,$87,$FC,$AF,$C0,$7F,$9F,$BA
	.dc.b	$BF,$E7,$AA,$F3,$F2,$9A,$E2,$FC,$5D,$7C,$DC,$BF,$22,$FC,$83,$07
	.dc.b	$F6,$4F,$17,$FD,$BD,$87,$9A,$3F,$AF,$EF,$FE,$91,$8C,$47,$EC,$CE
	.dc.b	$41,$CF,$6F,$E8,$DF,$F4,$07,$7D,$DF,$F8,$C7,$8F,$EA,$3F,$8D,$70
	.dc.b	$FF,$8F,$73,$FD,$4E,$3F,$DA,$CF,$33,$F5,$38,$BF,$8C,$63,$FC,$6F
	.dc.b	$93,$F8,$C3,$FC,$66,$7F,$8D,$63,$FC,$5C,$FE,$30,$FF,$19,$8F,$EC
	.dc.b	$FC,$9F,$84,$C3,$FB,$A9,$FC,$5E,$3F,$B5,$EB,$DF,$9F,$FE,$00,$3E
	.dc.b	$49,$FF,$27,$1F,$84,$3F,$98,$1F,$C4,$34,$FE,$28,$7F,$1B,$C7,$F8
	.dc.b	$B1,$FC,$EE,$3F,$82,$BF,$C7,$F1,$FE,$27,$1F,$86,$39,$9F,$8C,$3E
	.dc.b	$8E,$7F,$EE,$9E,$5F,$A3,$B7,$F0,$F7,$E9,$F9,$FF,$C5,$78,$BE,$FD
	.dc.b	$FF,$12,$E3,$FE,$59,$C9,$FC,$B7,$9B,$F9,$77,$37,$E2,$BE,$3F,$F1
	.dc.b	$BF,$B9,$7E,$3B,$7F,$CA,$63,$FB,$37,$F3,$8F,$F6,$4E,$0F,$D0,$71
	.dc.b	$BE,$E3,$F3,$D3,$FB,$2E,$27,$E7,$9F,$B7,$7F,$71,$FA,$C7,$EC,$CF
	.dc.b	$DA,$3F,$EB,$2E,$5D,$D7,$ED,$4F,$E7,$37,$D1,$F9,$87,$F8,$E3,$87
	.dc.b	$FA,$26,$7E,$48,$FE,$03,$C7,$FE,$CD,$E3,$FF,$21,$C7,$F3,$9D,$BF
	.dc.b	$59,$BF,$EA,$8F,$D4,$BF,$AA,$EB,$FD,$69,$F2,$37,$E5,$DB,$FE,$03
	.dc.b	$BF,$E1,$B1,$FE,$D9,$D7,$F4,$59,$FD,$89,$CF,$0F,$DD,$AF,$89,$1F
	.dc.b	$6B,$75,$FD,$47,$DF,$1F,$90,$3F,$65,$D7,$FA,$CE,$8F,$F7,$F2,$E2
	.dc.b	$FE,$E9,$8F,$F7,$13,$F3,$2F,$E6,$AF,$F2,$FC,$7F,$98,$C7,$F0,$3C
	.dc.b	$3E,$2C,$BF,$6C,$7F,$5B,$C7,$FB,$76,$3F,$93,$9F,$F7,$A7,$67,$E2
	.dc.b	$F5,$FF,$0C,$DF,$F6,$A7,$F8,$3B,$FD,$8F,$7E,$A7,$F1,$7E,$0F,$EF
	.dc.b	$1C,$3F,$DE,$7E,$13,$F8,$AF,$E3,$5F,$C6,$3F,$1A,$FE,$33,$F8,$BF
	.dc.b	$E3,$31,$FE,$0C,$FE,$28,$FB,$CE,$0F,$E8,$A7,$F0,$CE,$4F,$F7,$FF
	.dc.b	$6F,$F7,$48,$FA,$7F,$53,$C3,$91,$F8,$37,$FB,$F1,$79,$1F,$D1,$FF
	.dc.b	$18,$FF,$7F,$D9,$FF,$80,$3E,$8D,$8F,$C8,$0F,$13,$0F,$F4,$43,$F9
	.dc.b	$0D,$7E,$5F,$0F,$E6,$D8,$FA,$39,$7C,$BD,$DF,$DF,$96,$DF,$95,$C3
	.dc.b	$FC,$26,$EC,$FE,$84,$7C,$38,$6F,$FD,$07,$83,$FA,$17,$C4,$7F,$40
	.dc.b	$7F,$9F,$C7,$F4,$0F,$C2,$FF,$A0,$FD,$EF,$E5,$F0,$FA,$B8,$FE,$59
	.dc.b	$FC,$91,$F6,$9F,$98,$7E,$08,$D1,$FE,$50,$E7,$FE,$7F,$EC,$7F,$4C
	.dc.b	$3F,$34,$7F,$93,$7D,$9F,$E9,$E7,$F9,$3F,$6F,$E6,$DB,$FE,$6F,$83
	.dc.b	$FC,$99,$C7,$FE,$4D,$F4,$7F,$C9,$4F,$F9,$2F,$3F,$F2,$66,$3F,$E4
	.dc.b	$DC,$FF,$9D,$F8,$5F,$E9,$E7,$FA,$97,$D3,$23,$FA,$5C,$7F,$BF,$0F
	.dc.b	$F5,$7C,$65,$FE,$9F,$CD,$FD,$43,$9F,$11,$FD,$6E,$17,$FE,$57,$C3
	.dc.b	$FC,$B1,$8F,$F4,$0C,$BF,$A0,$BF,$CF,$B3,$FE,$7F,$A7,$EA,$35,$FF
	.dc.b	$26,$6D,$33,$FF,$16,$65,$FE,$4F,$3F,$50,$7F,$95,$F2,$FE,$05,$1F
	.dc.b	$78,$FE,$73,$CB,$F8,$27,$F9,$CF,$97,$FE,$5B,$9F,$D8,$1F,$62,$3F
	.dc.b	$CA,$FC,$5F,$B8,$E3,$FF,$31,$72,$7F,$98,$F9,$7F,$CC,$9C,$DF,$E0
	.dc.b	$0F,$C6,$BF,$C0,$7A,$7E,$AF,$AF,$DF,$97,$EA,$0F,$F4,$CE,$DF,$CD
	.dc.b	$BE,$9F,$F4,$DF,$84,$FE,$97,$D5,$FF,$15,$6F,$FE,$25,$3F,$99,$3F
	.dc.b	$CC,$E3,$FC,$25,$3F,$9D,$AF,$F0,$C6,$3F,$9B,$C3,$FC,$BA,$FE,$67
	.dc.b	$2F,$F0,$A9,$FE,$14,$FC,$6B,$F9,$54,$7B,$67,$FC,$08,$FE,$4F,$3F
	.dc.b	$3C,$7F,$9B,$FB,$7F,$DF,$EF,$F8,$27,$0F,$E5,$5C,$9F,$E1,$EC,$BF
	.dc.b	$C0,$99,$FF,$29,$D3,$FD,$35,$AF,$FA,$5E,$FF,$19,$87,$F9,$0B,$F5
	.dc.b	$CF,$C1,$BF,$E6,$73,$FD,$EE,$E1,$FF,$75,$47,$F8,$0A,$7F,$BF,$D7
	.dc.b	$F8,$02,$FF,$94,$E1,$FE,$8D,$C7,$FD,$29,$7F,$80,$C3,$F5,$07,$FA
	.dc.b	$27,$4F,$D7,$72,$FB,$3C,$CF,$F4,$36,$FF,$CC,$B8,$3F,$C6,$FD,$BF
	.dc.b	$5C,$FE,$D2,$7F,$20,$E1,$3F,$35,$FF,$79,$E2,$F0,$7F,$90,$3F,$E6
	.dc.b	$5F,$A2,$FE,$41,$F9,$EF,$F1,$FF,$E0,$1F,$F3,$27,$E5,$5F,$E6,$2F
	.dc.b	$0F,$FC,$B1,$7F,$E3,$DF,$45,$FB,$8F,$F2,$2A,$FE,$42,$6E,$BE,$25
	.dc.b	$3E,$E3,$6F,$F2,$4F,$21,$5F,$DD,$F1,$FE,$F1,$97,$F9,$5F,$3F,$E4
	.dc.b	$5F,$0F,$FE,$5D,$DF,$F0,$DC,$7E,$CC,$83,$D4,$FE,$EA,$FF,$78,$9E
	.dc.b	$0E,$B9,$32,$FF,$A0,$FE,$D7,$FC,$BF,$E3,$FF,$75,$AF,$97,$3B,$BF
	.dc.b	$EE,$CF,$FA,$26,$3F,$B9,$1F,$8F,$3F,$DC,$C7,$ED,$9F,$F2,$B6,$3F
	.dc.b	$E6,$A3,$FB,$C1,$FE,$EB,$F8,$9F,$F7,$66,$7F,$E5,$3D,$3F,$BC,$6B
	.dc.b	$FE,$E9,$DB,$E4,$78,$DF,$C5,$B9,$3E,$3E,$5F,$DE,$CF,$F3,$64,$FF
	.dc.b	$BA,$F3,$FF,$76,$67,$FE,$E9,$3F,$DD,$4F,$FB,$AE,$39,$CF,$F9,$AB
	.dc.b	$8B,$9F,$1F,$F2,$3C,$7F,$91,$F1,$FD,$2A,$F9,$73,$97,$FC,$E3,$EA
	.dc.b	$34,$FF,$2C,$EB,$FE,$02,$DB,$FD,$CD,$BF,$FB,$9F,$69,$87,$F9,$FC
	.dc.b	$FF,$71,$3F,$EF,$1D,$BE,$37,$27,$D9,$9F,$E9,$07,$FB,$B2,$3F,$3E
	.dc.b	$7F,$82,$4F,$DF,$46,$F3,$FF,$72,$71,$7F,$84,$B8,$FE,$75,$FD,$7E
	.dc.b	$3F,$F4,$6E,$3F,$9F,$CB,$FC,$97,$F4,$5F,$DE,$FF,$47,$FE,$F5,$FB
	.dc.b	$1F,$FA,$FF,$5F,$CB,$69,$F3,$FF,$6D,$B9,$FC,$77,$FD,$7D,$3F,$E4
	.dc.b	$BA,$FF,$26,$5F,$FA,$AF,$0F,$EE,$D8,$FF,$AC,$72,$FC,$A9,$FC,$A7
	.dc.b	$C4,$FF,$20,$3F,$C5,$FB,$3F,$63,$C9,$FE,$9F,$8F,$C4,$42,$BF,$C7
	.dc.b	$D7,$FE,$97,$C3,$FC,$79,$8F,$DA,$D3,$FC,$77,$9D,$7F,$DA,$BD,$2D
	.dc.b	$71,$1F,$ED,$8C,$6A,$F3,$FB,$3F,$7B,$77,$47,$F6,$4E,$9E,$53,$98
	.dc.b	$FC,$2E,$5F,$E9,$8C,$FF,$D3,$3A,$7F,$90,$B5,$FC,$0D,$FD,$F3,$FD
	.dc.b	$81,$CB,$FC,$3B,$D6,$FF,$45,$70,$FF,$1A,$3E,$DF,$1F,$F9,$8F,$F3
	.dc.b	$2F,$E9,$C7,$FA,$00,$FE,$2D,$F1,$FF,$ED,$B3,$FD,$B8,$FF,$00,$2E
	.dc.b	$1F,$7C,$FF,$1D,$7D,$77,$37,$F7,$1F,$07,$FE,$CE,$9F,$F3,$41,$FE
	.dc.b	$FA,$E8,$FE,$0D,$19,$9F,$E5,$E7,$FB,$5F,$9D,$87,$F9,$BF,$F3,$DF
	.dc.b	$F6,$47,$E2,$5F,$EA,$0F,$79,$FA,$97,$CC,$C2,$9F,$1C,$FF,$04,$BF
	.dc.b	$EC,$23,$FD,$7F,$3F,$E6,$8A,$F1,$65,$FC,$32,$9C,$1E,$07,$E5,$1F
	.dc.b	$E9,$0D,$3F,$4E,$70,$D3,$FC,$33,$28,$FF,$37,$1F,$A6,$D7,$FC,$EB
	.dc.b	$F6,$5F,$EF,$AF,$B2,$FF,$7E,$ED,$FC,$9B,$17,$8D,$FC,$9F,$F1,$4F
	.dc.b	$F7,$9C,$E6,$3F,$95,$7E,$2D,$8A,$FE,$5B,$E5,$3D,$47,$C4,$E6,$7F
	.dc.b	$B7,$37,$A3,$5F,$6F,$FE,$FE,$FA,$3F,$06,$1C,$6B,$FE,$00,$F2,$7C
	.dc.b	$2C,$BF,$C6,$7D,$0F,$CC,$F1,$9F,$F8,$C7,$F1,$6F,$F5,$D1,$FB,$47
	.dc.b	$F1,$B1,$A7,$F1,$F7,$5F,$F5,$BE,$5E,$0D,$7C,$F1,$F0,$4F,$E8,$FC
	.dc.b	$BF,$3D,$F4,$99,$3E,$9F,$07,$C7,$DF,$D7,$19,$9E,$1B,$7A,$E3,$15
	.dc.b	$F2,$79,$DC,$5F,$5A,$31,$BF,$D8,$E3,$F5,$5F,$FB,$33,$97,$EA,$78
	.dc.b	$9F,$D6,$BE,$B3,$EE,$C6,$25,$F4,$3F,$3F,$9F,$FA,$CC,$39,$7D,$5C
	.dc.b	$1F,$DE,$A3,$1C,$5F,$EE,$AE,$4F,$F5,$0F,$6F,$E1,$E7,$FD,$27,$E4
	.dc.b	$BE,$C7,$C7,$1F,$DA,$BC,$7C,$3F,$DD,$7B,$72,$FE,$5B,$0F,$F4,$A6
	.dc.b	$3F,$B5,$8F,$F2,$21,$FE,$E6,$E3,$FF,$24,$1F,$0F,$E3,$3D,$1C,$DF
	.dc.b	$9F,$DB,$E4,$F4,$BD,$FF,$C5,$79,$BF,$E7,$5F,$99,$FC,$65,$FF,$7B
	.dc.b	$C3,$F1,$97,$FE,$3A,$C3,$DE,$F1,$CB,$FD,$75,$9F,$DF,$CB,$F2,$11
	.dc.b	$FA,$A3,$FC,$0B,$49,$FD,$8D,$FE,$4A,$7F,$C1,$38,$77,$7F,$D8,$9A
	.dc.b	$7E,$BD,$A6,$A3,$FD,$DA,$FE,$AF,$E1,$DD,$FF,$EF,$13,$FC,$41,$F4
	.dc.b	$FF,$E8,$EE,$4F,$A1,$5F,$E9,$1F,$39,$BF,$F7,$DE,$4F,$F1,$33,$FE
	.dc.b	$94,$9D,$CF,$C3,$7F,$D2,$FC,$5A,$7F,$D3,$3B,$57,$FD,$28,$D7,$F1
	.dc.b	$71,$5F,$F1,$2F,$5F,$4A,$F0,$C7,$F8,$0F,$0D,$3F,$ED,$5E,$4B,$C9
	.dc.b	$EC,$8F,$E8,$DA,$7F,$AA,$23,$11,$EB,$CF,$F8,$9F,$07,$C7,$8F,$DA
	.dc.b	$9F,$1F,$87,$F8,$BF,$83,$F4,$23,$F5,$5B,$57,$FA,$3B,$F6,$5C,$9F
	.dc.b	$64,$FD,$99,$FC,$D8,$FC,$71,$FE,$40,$E8,$FE,$29,$CB,$FA,$F8,$F0
	.dc.b	$1F,$B4,$3F,$C8,$B9,$7C,$B7,$FA,$AF,$07,$76,$9F,$E3,$FC,$79,$3F
	.dc.b	$C9,$B9,$3F,$9B,$78,$2E,$EF,$BB,$3F,$C0,$DF,$93,$CF,$FC,$37,$AF
	.dc.b	$FA,$6C,$FF,$0B,$8F,$E2,$1E,$0F,$F2,$9F,$89,$FE,$3F,$3F,$E9,$73
	.dc.b	$E4,$1F,$CD,$67,$F8,$B6,$DB,$8F,$F6,$97,$37,$F1,$BA,$FF,$6C,$F5
	.dc.b	$7F,$B5,$34,$EF,$5F,$C7,$63,$FB,$DB,$F9,$7F,$1B,$7C,$BD,$A3,$FC
	.dc.b	$CF,$2F,$C9,$FF,$C0,$BD,$EF,$C3,$3F,$AE,$7C,$5F,$F8,$23,$B4,$FF
	.dc.b	$82,$BC,$1D,$63,$FD,$7C,$FE,$59,$CF,$CD,$C1,$FD,$3D,$FC,$7B,$FE
	.dc.b	$3E,$0F,$99,$C3,$EB,$D7,$FB,$23,$FE,$F0,$E2,$F8,$53,$FE,$0C,$3F
	.dc.b	$C9,$27,$AC,$FE,$D2,$7F,$A9,$8F,$8C,$7F,$37,$3F,$C6,$07,$F6,$1F
	.dc.b	$A2,$7D,$9B,$EC,$7E,$9C,$FF,$A2,$A7,$3F,$A4,$62,$1F,$A7,$FF,$98
	.dc.b	$DF,$F0,$31,$FD,$AF,$F2,$C7,$C9,$BF,$65,$E1,$18,$FA,$7E,$5E,$CF
	.dc.b	$F2,$FD,$2F,$90,$FC,$6F,$ED,$F6,$7F,$B6,$FA,$4F,$9E,$BC,$DA,$7E
	.dc.b	$08,$F3,$2D,$B8,$C7,$FA,$8B,$3F,$1E,$9F,$86,$F7,$3A,$1E,$7D,$BC
	.dc.b	$D5,$FD,$7F,$2F,$A5,$7E,$78,$C7,$5E,$F0,$FC,$4C,$75,$B6,$FF,$92
	.dc.b	$FC,$C7,$FB,$07,$35,$AF,$F1,$89,$D6,$3F,$A1,$3D,$FB,$36,$D3,$67
	.dc.b	$F7,$5F,$33,$F1,$EF,$F8,$F4,$C7,$57,$FB,$FF,$B7,$F2,$31,$FB,$5C
	.dc.b	$3F,$D4,$DD,$BF,$DC,$7B,$BF,$95,$5F,$6A,$FE,$F8,$E5,$FE,$5C,$78
	.dc.b	$0F,$E6,$1F,$86,$FE,$39,$FF,$7E,$35,$FE,$52,$71,$3F,$C1,$5D,$FE
	.dc.b	$67,$F8,$35,$7F,$82,$5F,$F3,$75,$78,$5F,$F1,$87,$7F,$A3,$C3,$FD
	.dc.b	$BC,$63,$1D,$3F,$ED,$AF,$86,$E9,$3A,$8C,$5E,$9F,$D4,$ED,$E3,$7F
	.dc.b	$5C,$E7,$C1,$70,$7D,$09,$FD,$AB,$B7,$F9,$1B,$BE,$4F,$BD,$1F,$EB
	.dc.b	$E7,$F0,$DC,$5F,$CF,$F9,$F3,$3F,$BE,$72,$FF,$B1,$1E,$8F,$EF,$3A
	.dc.b	$B6,$FF,$01,$7F,$35,$E0,$7F,$49,$7F,$39,$D5,$67,$E9,$1F,$E3,$43
	.dc.b	$FC,$AD,$DD,$FA,$77,$F7,$47,$FB,$F9,$FF,$0C,$76,$FF,$98,$A3,$B9
	.dc.b	$C4,$FC,$B6,$39,$47,$EE,$AF,$F8,$83,$FE,$63,$F1,$BF,$C8,$9E,$0F
	.dc.b	$F0,$4C,$B8,$4A,$3F,$81,$BE,$21,$F8,$03,$E1,$7F,$80,$C6,$63,$1F
	.dc.b	$80,$7F,$09,$F8,$1F,$E1,$13,$FC,$F4,$FF,$39,$78,$7F,$C4,$7C,$4F
	.dc.b	$E1,$CF,$F1,$3F,$BD,$FD,$DC,$BF,$57,$F7,$BF,$C0,$4C,$57,$F1,$A7
	.dc.b	$1F,$80,$FE,$D3,$3E,$2F,$ED,$BE,$13,$9E,$3D,$6B,$F8,$07,$2F,$F2
	.dc.b	$67,$FA,$43,$E4,$DD,$9F,$F0,$09,$F8,$99,$5B,$E5,$1F,$D3,$F9,$FF
	.dc.b	$CD,$DE,$C7,$F0,$18,$79,$7F,$A9,$5E,$57,$E5,$1F,$F4,$5C,$FF,$8E
	.dc.b	$5F,$EB,$1D,$9C,$9F,$E0,$87,$28,$79,$9F,$E2,$BE,$ED,$4F,$F0,$43
	.dc.b	$FC,$0D,$E3,$FF,$04,$C7,$FC,$C4,$DB,$E3,$3F,$B5,$AE,$2F,$FA,$0D
	.dc.b	$F8,$4F,$F8,$C6,$BD,$1E,$2F,$F7,$8A,$E3,$3F,$C0,$2B,$6F,$F8,$B1
	.dc.b	$E6,$FF,$9A,$7C,$3C,$AF,$13,$C5,$5F,$C1,$7C,$4F,$E1,$18,$FC,$56
	.dc.b	$C7,$FA,$B6,$BF,$D2,$5E,$1F,$F6,$B7,$E1,$69,$FE,$31,$D7,$BC,$7F
	.dc.b	$6A,$DF,$FC,$34,$FF,$6F,$7C,$35,$F3,$E3,$E5,$5F,$F8,$E7,$C3,$FB
	.dc.b	$DC,$56,$7F,$4B,$C5,$F7,$1F,$EF,$1C,$BF,$83,$E4,$FE,$05,$1F,$E3
	.dc.b	$3C,$7F,$AA,$BC,$39,$7F,$0E,$FA,$3B,$FF,$D9,$DD,$1F,$93,$8C,$1F
	.dc.b	$B2,$7F,$28,$E0,$FC,$DF,$83,$FC,$0D,$FC,$EE,$5F,$B2,$3F,$4C,$FE
	.dc.b	$C5,$F3,$3F,$BF,$1D,$9F,$AE,$71,$5E,$17,$FC,$10,$78,$7C,$17,$5F
	.dc.b	$EB,$BF,$44,$E1,$3F,$96,$CB,$F8,$C6,$DF,$B2,$C7,$3D,$9F,$E1,$7D
	.dc.b	$AF,$AF,$FE,$68,$8F,$9D,$1F,$E4,$DC,$3F,$D0,$9C,$BC,$9F,$F6,$67
	.dc.b	$95,$FD,$1E,$7B,$F7,$7F,$5A,$7C,$33,$9C,$FF,$8A,$BF,$C9,$F3,$FE
	.dc.b	$99,$DD,$FE,$2E,$9C,$CF,$F4,$E3,$FC,$BA,$E3,$B3,$A3,$AE,$BE,$97
	.dc.b	$F8,$9A,$DB,$FE,$7B,$EA,$FE,$6D,$D3,$FE,$7D,$DF,$F9,$3D,$7F,$A5
	.dc.b	$EB,$0F,$F2,$AC,$3D,$CF,$F9,$B2,$3D,$CF,$94,$FE,$5F,$C1,$F3,$34
	.dc.b	$FC,$3C,$7F,$45,$9F,$BD,$C3,$F4,$3C,$4F,$F7,$A7,$E0,$5F,$3B,$1F
	.dc.b	$0F,$2F,$F2,$DF,$33,$F9,$1B,$FC,$B3,$1F,$E3,$27,$F3,$CE,$CF,$F1
	.dc.b	$DF,$DF,$FF,$99,$36,$F4,$C6,$27,$FC,$24,$7F,$AE,$7C,$DF,$E2,$C7
	.dc.b	$F6,$D7,$F8,$AF,$0F,$FA,$4D,$FB,$26,$CF,$E2,$71,$BE,$6F,$F0,$03
	.dc.b	$FE,$F2,$6A,$F3,$3F,$C2,$73,$C4,$F9,$63,$9F,$FC,$33,$3E,$18,$E2
	.dc.b	$FF,$86,$1F,$F3,$C1,$FE,$16,$3F,$CD,$07,$2F,$03,$F8,$AB,$F5,$9A
	.dc.b	$F0,$3F,$8B,$14,$FB,$7D,$1D,$9F,$19,$FC,$57,$C3,$FF,$35,$74,$F9
	.dc.b	$2C,$7F,$17,$F1,$3F,$CD,$DD,$5C,$6F,$FC,$FF,$D7,$FE,$71,$E9,$FE
	.dc.b	$2D,$7C,$0F,$C6,$4F,$84,$FF,$07,$9F,$EB,$43,$FC,$26,$7C,$53,$FD
	.dc.b	$0D,$C3,$F6,$3B,$DC,$7F,$C5,$67,$FC,$D0,$FF,$12,$69,$FC,$27,$93
	.dc.b	$F0,$CD,$F1,$5E,$3F,$F0,$BE,$6F,$C1,$4F,$E5,$4F,$F3,$97,$87,$FD
	.dc.b	$BB,$D1,$FE,$AD,$E0,$FF,$A4,$FC,$8F,$ED,$F5,$FD,$90,$FF,$04,$BF
	.dc.b	$E1,$37,$07,$F6,$FC,$3F,$CD,$CF,$FB,$54,$FF,$56,$F8,$5F,$9A,$CB
	.dc.b	$FC,$0B,$1F,$EB,$58,$C4,$FE,$5E,$3F,$C2,$B9,$FA,$9F,$F5,$D6,$DA
	.dc.b	$7F,$32,$32,$73,$FF,$47,$8F,$EE,$47,$84,$FC,$D9,$FE,$E5,$7F,$B7
	.dc.b	$3F,$9C,$39,$3C,$5C,$47,$F7,$7E,$AF,$F2,$E3,$DA,$7F,$B8,$4F,$5E
	.dc.b	$9F,$F3,$8C,$FF,$6E,$E8,$FE,$0E,$FF,$7F,$9E,$35,$D5,$FE,$17,$CF
	.dc.b	$A8,$FE,$D7,$1F,$E6,$73,$C7,$E3,$FF,$1E,$74,$FF,$20,$9B,$1F,$98
	.dc.b	$EA,$EB,$5F,$E6,$FE,$AF,$1B,$D2,$7F,$B9,$1F,$05,$FC,$D3,$F1,$7A
	.dc.b	$B8,$BF,$DB,$BA,$BA,$BF,$9C,$38,$1F,$DC,$BB,$32,$FF,$6D,$3E,$67
	.dc.b	$AD,$F9,$E7,$E5,$DF,$F2,$67,$FC,$65,$DD,$78,$BF,$8C,$47,$FB,$9B
	.dc.b	$93,$F0,$93,$F9,$D9,$CC,$FF,$70,$3E,$D7,$6F,$FB,$96,$33,$1F,$CB
	.dc.b	$CF,$C9,$F5,$7F,$B2,$5F,$F3,$91,$8E,$CF,$ED,$5C,$5F,$E3,$1B,$FE
	.dc.b	$63,$F8,$17,$F6,$AA,$FD,$13,$FD,$A9,$FF,$1A,$F3,$FF,$BA,$F9,$9F
	.dc.b	$73,$FE,$FF,$3F,$A1,$F8,$3F,$CE,$E7,$FD,$E1,$DB,$FE,$FC,$F2,$FF
	.dc.b	$53,$1F,$E0,$CC,$3F,$D9,$8F,$F8,$25,$FF,$67,$73,$FF,$86,$28,$FF
	.dc.b	$89,$9B,$1F,$E4,$4E,$3F,$F5,$54,$3E,$EB,$81,$FE,$2C,$EC,$FF,$1A
	.dc.b	$72,$FF,$AB,$DF,$F1,$B7,$5F,$FA,$C7,$6F,$4E,$3F,$E4,$EA,$C6,$DF
	.dc.b	$E3,$4E,$AC,$CF,$F8,$B0,$FE,$46,$E3,$6F,$D0,$47,$F2,$6E,$AE,$AD
	.dc.b	$3F,$C6,$3C,$5E,$57,$FC,$49,$A6,$37,$FF,$19,$C7,$FA,$42,$7D,$AF
	.dc.b	$73,$FA,$1B,$8B,$EF,$18,$E0,$F9,$4D,$3F,$C5,$DE,$67,$F2,$39,$FB
	.dc.b	$9E,$47,$F8,$BB,$A3,$FD,$25,$5F,$82,$1E,$5F,$F8,$B9,$FE,$A3,$C7
	.dc.b	$E7,$7F,$9B,$78,$DF,$25,$FC,$07,$07,$F4,$23,$F9,$FB,$FC,$F7,$6F
	.dc.b	$E7,$DF,$7B,$FC,$D7,$C6,$FA,$AF,$E1,$FA,$3F,$0A,$7F,$85,$39,$5F
	.dc.b	$7D,$8D,$BF,$6C,$DC,$7E,$7D,$57,$E1,$DF,$E3,$3F,$36,$ED,$1F,$D5
	.dc.b	$B5,$47,$F4,$5F,$23,$F0,$BF,$69,$FE,$16,$F2,$BF,$A8,$5E,$E3,$F2
	.dc.b	$BF,$43,$99,$FE,$B7,$3C,$E3,$D0,$FF,$55,$3F,$BA,$67,$B9,$FC,$53
	.dc.b	$FD,$90,$91,$FD,$5C,$FF,$65,$F5,$7F,$87,$27,$FC,$15,$D0,$BF,$EC
	.dc.b	$87,$3A,$FF,$3D,$E1,$FD,$9C,$C3,$F8,$2C,$BD,$ED,$3F,$C3,$9E,$0B
	.dc.b	$3F,$D1,$FD,$BF,$09,$FF,$1A,$EA,$8F,$D9,$FE,$1F,$99,$FF,$6F,$CA
	.dc.b	$3F,$79,$C3,$FB,$DF,$C9,$35,$5F,$E1,$37,$FC,$02,$47,$DC,$7F,$A6
	.dc.b	$63,$FB,$F4,$C8,$CE,$BE,$17,$E5,$35,$FF,$01,$6D,$FE,$0E,$50,$FC
	.dc.b	$17,$DF,$1F,$E4,$E8,$C3,$F0,$33,$E1,$F5,$AF,$97,$FC,$6F,$2F,$EF
	.dc.b	$7F,$48,$B7,$FE,$1A,$E7,$7E,$F3,$88,$FF,$17,$F3,$61,$FF,$2C,$5F
	.dc.b	$F9,$49,$FF,$2E,$C2,$F8,$47,$F9,$83,$ED,$BC,$0F,$0A,$F5,$68,$F8
	.dc.b	$FD,$CF,$EF,$5F,$D0,$F9,$DE,$8D,$2C,$F9,$71,$AF,$F9,$0B,$4F,$22
	.dc.b	$C7,$FC,$D6,$FF,$B1,$F4,$DB,$E6,$AA,$86,$7E,$67,$95,$31,$F3,$E5
	.dc.b	$FB,$4F,$13,$F9,$27,$E6,$5E,$8B,$FF,$29,$77,$5D,$FF,$C5,$F1,$EF
	.dc.b	$3F,$91,$CE,$F2,$7E,$11,$C7,$64,$FE,$E6,$BF,$D0,$F1,$C9,$FF,$44
	.dc.b	$1F,$CB,$30,$D3,$FA,$89,$AF,$AA,$AC,$BF,$25,$F3,$7E,$47,$6B,$C8
	.dc.b	$7F,$9A,$BF,$28,$FA,$4E,$0F,$F4,$BD,$6D,$F0,$9F,$E7,$09,$C6,$FF
	.dc.b	$EB,$EA,$E3,$3A,$3F,$D3,$F2,$FC,$2F,$C2,$79,$B9,$1D,$5F,$EA,$4E
	.dc.b	$2F,$9D,$F5,$CF,$B1,$7E,$32,$FE,$01,$7E,$76,$FB,$8F,$D0,$E3,$C7
	.dc.b	$16,$3C,$70,$E1,$6F,$C3,$B7,$B7,$DF,$BE,$EF,$F3,$D7,$EF,$7D,$CF
	.dc.b	$F9,$8D,$F1,$E9,$E2,$D1,$F9,$BF,$F8,$1E,$FF,$09,$5E,$0A,$D7,$EA
	.dc.b	$19,$FD,$DD,$8F,$AB,$F0,$07,$EF,$9F,$7B,$DB,$7C,$17,$93,$36,$1C
	.dc.b	$38,$18,$70,$EC,$61,$FD,$FB,$97,$E4,$E1,$F9,$9C,$38,$3F,$D5,$DF
	.dc.b	$0F,$E3,$BF,$E7,$D7,$A6,$1C,$F5,$F3,$6B,$E5,$D7,$F2,$3F,$A0,$FF
	.dc.b	$6D,$E0,$FF,$63,$FE,$E4,$FC,$2D,$F9,$F7,$D4,$FA,$D7,$E4,$4F,$C5
	.dc.b	$7F,$89,$E1,$C2,$F7,$C8,$BD,$FD,$4E,$B8,$BF,$CE,$B9,$F5,$7F,$D5
	.dc.b	$3C,$7F,$AF,$73,$F5,$3F,$E3,$B9,$F1,$6B,$E1,$FD,$FB,$CD,$5C,$B3
	.dc.b	$EF,$93,$F7,$6F,$E6,$FE,$07,$FB,$C3,$E5,$BF,$7D,$ED,$BE,$41,$F8
	.dc.b	$23,$F1,$57,$E9,$EF,$CE,$9E,$F7,$CE,$D9,$97,$2C,$78,$E1,$C3,$CA
	.dc.b	$DE,$9D,$EF,$FB,$69,$FE,$EC,$38,$8F,$84,$FF,$4E,$B8,$D7,$C9,$AF
	.dc.b	$8F,$F3,$6F,$4C,$FF,$15,$CF,$C1,$3E,$7D,$FF,$25,$FA,$DF,$F8,$EF
	.dc.b	$F0,$6F,$F2,$6F,$BC,$C5,$FB,$3B,$F1,$7F,$C7,$5E,$FE,$2F,$FB,$15
	.dc.b	$6B,$D2,$CF,$C5,$BF,$08,$FF,$62,$5F,$5D,$FF,$C4,$4F,$1C,$FB,$67
	.dc.b	$E2,$B6,$9E,$2C,$9F,$3F,$FD,$A3,$8B,$F9,$76,$5F,$07,$EA,$DF,$B0
	.dc.b	$3E,$91,$F8,$B7,$F3,$9F,$9B,$7E,$3B,$F4,$B1,$E2,$30,$E3,$7C,$3D
	.dc.b	$E5,$FD,$6E,$EF,$9D,$9F,$F3,$DC,$38,$7E,$C3,$FE,$2E,$8C,$63,$E2
	.dc.b	$7F,$C4,$93,$DA,$3A,$E7,$FD,$2F,$CE,$F5,$57,$0D,$7D,$13,$FD,$3F
	.dc.b	$D1,$FA,$F2,$FD,$D7,$F1,$5F,$B4,$BE,$7D,$F9,$33,$F5,$BF,$F3,$97
	.dc.b	$EB,$1F,$88,$C3,$FA,$96,$1C,$2F,$75,$FE,$77,$FD,$63,$55,$D7,$C9
	.dc.b	$65,$F8,$AC,$3F,$31,$FB,$FD,$9C,$E3,$CB,$E6,$FC,$B5,$D7,$4F,$17
	.dc.b	$D5,$FF,$1B,$FB,$87,$D1,$F9,$FF,$3A,$7E,$3E,$FC,$95,$F4,$0F,$B3
	.dc.b	$FE,$31,$87,$0D,$58,$70,$BD,$FE,$D1,$AC,$BE,$17,$17,$F9,$AF,$0E
	.dc.b	$FF,$8F,$FF,$80,$BD,$2F,$C2,$C7,$3C,$79,$7E,$BD,$E5,$C3,$C9,$B7
	.dc.b	$8F,$EE,$BE,$7F,$EA,$8F,$96,$7E,$57,$FE,$35,$FC,$41,$F9,$F3,$F6
	.dc.b	$A7,$EE,$3F,$2B,$0E,$17,$BA,$D7,$EF,$19,$C7,$9F,$17,$EE,$6F,$B7
	.dc.b	$E7,$BF,$B8,$78,$FD,$5B,$4E,$FA,$79,$30,$F1,$CE,$72,$7E,$45,$FD
	.dc.b	$27,$E2,$5F,$9B,$3F,$73,$7E,$90,$FB,$C7,$E2,$CF,$AF,$7E,$B3,$FD
	.dc.b	$1F,$1E,$38,$70,$BD,$D3,$F5,$9D,$64,$FD,$8F,$E7,$72,$7F,$87,$F1
	.dc.b	$E2,$FF,$A7,$5E,$7A,$78,$74,$7E,$1F,$E7,$C3,$CD,$39,$F8,$1F,$F3
	.dc.b	$1F,$D7,$FE,$D3,$F1,$6F,$07,$E1,$5E,$CF,$C0,$3F,$C5,$9F,$5F,$FE
	.dc.b	$E2,$D7,$F6,$55,$AC,$BF,$DA,$BF,$AF,$66,$7F,$92,$4F,$B9,$FF,$6F
	.dc.b	$9C,$38,$3F,$DB,$39,$77,$59,$77,$9F,$19,$B7,$C5,$EA,$FE,$E2,$1F
	.dc.b	$B2,$FD,$03,$E9,$FC,$B7,$F9,$AB,$F6,$57,$E2,$BF,$C3,$EB,$FC,$89
	.dc.b	$5B,$EA,$FE,$B5,$F3,$AB,$7E,$03,$BB,$EB,$C1,$D9,$FE,$16,$72,$5F
	.dc.b	$87,$FF,$11,$8D,$E3,$E2,$C3,$FC,$1B,$1C,$E3,$C1,$9F,$F9,$FB,$DE
	.dc.b	$FE,$BF,$2F,$DD,$1F,$B8,$FD,$FF,$C2,$DF,$AA,$BF,$10,$7E,$4E,$FC
	.dc.b	$1F,$FB,$86,$3C,$70,$E1,$7B,$AF,$F5,$06,$BF,$DC,$6B,$5E,$2F,$F9
	.dc.b	$DB,$EA,$7F,$9C,$E9,$E0,$E8,$C5,$7F,$9A,$63,$1A,$F8,$B5,$FF,$57
	.dc.b	$4F,$82,$73,$9F,$97,$97,$FD,$F5,$D3,$FD,$13,$F0,$BF,$F0,$77,$DB
	.dc.b	$FF,$83,$FE,$F3,$F8,$77,$E7,$DF,$B8,$E0,$5E,$FE,$A3,$FC,$88,$2B
	.dc.b	$7E,$9F,$FA,$92,$75,$F4,$5F,$E8,$E7,$86,$37,$6F,$F4,$06,$7E,$0C
	.dc.b	$79,$7C,$17,$F6,$89,$E4,$6A,$3B,$E5,$FC,$87,$E6,$3F,$1F,$2F,$CE
	.dc.b	$1F,$80,$35,$7E,$8D,$FC,$17,$D1,$7E,$59,$FC,$4F,$E6,$3D,$78,$FF
	.dc.b	$A4,$71,$E3,$87,$0F,$67,$F8,$15,$BF,$54,$FD,$35,$76,$F7,$BF,$47
	.dc.b	$87,$1D,$3C,$0F,$EA,$0F,$23,$E3,$D3,$C5,$EF,$FA,$2F,$BC,$79,$7E
	.dc.b	$0B,$FC,$B9,$F2,$8F,$DC,$FF,$88,$7B,$4F,$C3,$DF,$8D,$7F,$59,$F2
	.dc.b	$BF,$CB,$17,$BA,$FE,$2D,$5A,$F4,$5F,$A6,$EB,$1F,$E3,$5B,$FF,$8F
	.dc.b	$BE,$2B,$FC,$07,$8F,$7E,$6F,$E2,$B3,$FE,$02,$AF,$0E,$3C,$E3,$96
	.dc.b	$5D,$B4,$FE,$C3,$1F,$E6,$9F,$D2,$9F,$BB,$BF,$29,$FE,$B7,$FA,$4B
	.dc.b	$EB,$5F,$B4,$7D,$7F,$1B,$FC,$79,$87,$0B,$DD,$6B,$F3,$9E,$98,$BF
	.dc.b	$53,$F2,$DE,$F3,$F5,$1F,$EC,$97,$97,$C8,$34,$F0,$7E,$EF,$FE,$87
	.dc.b	$9C,$C7,$8F,$E0,$BF,$DA,$1C,$7F,$AF,$FD,$8F,$FA,$BF,$D0,$B5,$7E
	.dc.b	$8C,$FD,$0F,$F8,$7F,$C7,$FF,$95,$3D,$8F,$F1,$5E,$1C,$2F,$75,$FC
	.dc.b	$53,$3D,$6F,$FE,$DD,$AF,$EC,$7F,$B4,$7F,$8C,$A3,$FC,$67,$A7,$87
	.dc.b	$E4,$1A,$E6,$7C,$B9,$F6,$9C,$E7,$DB,$E1,$73,$5D,$67,$3B,$79,$F3
	.dc.b	$CC,$BD,$1F,$EB,$BF,$28,$FC,$DB,$F1,$7C,$0F,$C7,$5F,$B4,$7A,$30
	.dc.b	$E1,$F1,$EC,$38,$77,$FC,$3B,$FF,$3B,$F1,$7F,$17,$CF,$AE,$1E,$98
	.dc.b	$C6,$1F,$C9,$A3,$1C,$3F,$C3,$E3,$A6,$9E,$2F,$97,$F3,$E1,$E6,$8F
	.dc.b	$2F,$E0,$3F,$E3,$CF,$A6,$FE,$5D,$C0,$FD,$75,$9B,$F4,$5F,$F6,$C6
	.dc.b	$BF,$E8,$7B,$DD,$6A,$7E,$DC,$E7,$E1,$5F,$57,$FE,$82,$BF,$E7,$58
	.dc.b	$7F,$A1,$DE,$2F,$F4,$4A,$E1,$F0,$CC,$BC,$52,$F8,$BF,$F6,$A7,$E7
	.dc.b	$EF,$CD,$1F,$B6,$BC,$17,$EB,$6F,$C5,$3F,$CD,$5F,$8C,$7F,$B4,$70
	.dc.b	$E1,$7B,$AD,$7D,$A7,$FB,$BF,$CE,$FF,$51,$56,$B4,$FD,$08,$7C,$1F
	.dc.b	$2B,$FB,$A7,$9E,$7E,$0C,$DD,$3E,$6D,$7C,$B3,$9F,$C1,$3F,$CD,$3F
	.dc.b	$C0,$1F,$EB,$1F,$D8,$FF,$DD,$3F,$A8,$75,$FB,$E7,$CF,$7F,$98,$AB
	.dc.b	$5F,$8A,$2B,$5F,$9C,$67,$A9,$FA,$B7,$FA,$BA,$FA,$71,$7F,$AC,$DE
	.dc.b	$37,$E9,$D3,$C3,$6D,$3C,$3C,$3E,$8C,$3F,$CE,$3F,$AF,$FF,$AA,$FE
	.dc.b	$FF,$F0,$5F,$56,$D7,$FB,$EF,$CD,$6B,$F0,$A5,$6A,$7F,$C0,$3F,$15
	.dc.b	$98,$FF,$60,$E1,$FE,$B5,$FD,$C2,$DF,$0F,$1B,$FC,$89,$87,$FB,$6F
	.dc.b	$3E,$11,$8F,$1F,$FD,$1F,$F8,$BB,$F3,$4F,$F6,$DF,$DA,$7E,$0F,$E8
	.dc.b	$DF,$A9,$FE,$53,$E2,$BF,$D6,$37,$BF,$9E,$FF,$3B,$F9,$1F,$E0,$AA
	.dc.b	$FC,$AD,$6B,$1F,$F7,$2E,$9F,$ED,$9C,$3B,$7E,$03,$FE,$D8,$79,$67
	.dc.b	$DF,$EB,$FC,$77,$E2,$CB,$FC,$2B,$FB,$3F,$E9,$BC,$7F,$F5,$9E,$5F
	.dc.b	$CC,$FE,$6D,$F8,$D3,$F7,$AF,$F6,$4F,$C6,$BC,$27,$C2,$31,$E3,$F1
	.dc.b	$9F,$E6,$5C,$38,$7E,$19,$FC,$F2,$FF,$8C,$D6,$B2,$EB,$9F,$F1,$B8
	.dc.b	$C7,$E0,$DF,$C7,$63,$A7,$E1,$1F,$ED,$4C,$BA,$6B,$E2,$F3,$BD,$13
	.dc.b	$9C,$7C,$F3,$E6,$F1,$FD,$9C,$FF,$33,$D9,$7E,$45,$FD,$83,$F6,$17
	.dc.b	$CD,$3F,$1E,$7E,$25,$EB,$C3,$85,$EE,$B5,$FB,$86,$78,$3F,$86,$F6
	.dc.b	$FF,$9E,$FE,$2B,$FB,$2B,$FE,$7D,$BB,$A7,$83,$E6,$DA,$F8,$3B,$3C
	.dc.b	$B8,$F2,$AD,$4F,$93,$F2,$5F,$F6,$BF,$D4,$3F,$35,$FF,$77,$FE,$A2
	.dc.b	$FC,$BF,$FD,$AD,$BF,$E6,$BE,$15,$C5,$FE,$FA,$C3,$87,$B6,$C3,$87
	.dc.b	$D0,$B0,$7A,$1F,$ED,$8B,$DE,$FF,$CD,$71,$E5,$A3,$D1,$FA,$9A,$76
	.dc.b	$AF,$E5,$B7,$DB,$3E,$BF,$57,$E8,$C3,$CF,$B7,$FB,$AF,$F1,$7F,$F1
	.dc.b	$9F,$5B,$F1,$9F,$E3,$DF,$A1,$7F,$B5,$3F,$4B,$FF,$24,$FD,$AF,$FA
	.dc.b	$3E,$B5,$F0,$AA,$D7,$DA,$3F,$59,$D6,$3F,$C9,$F8,$3F,$C5,$98,$75
	.dc.b	$FD,$9B,$FC,$63,$08,$E7,$87,$FB,$8B,$F0,$8F,$ED,$B8,$FF,$46,$AD
	.dc.b	$61,$E4,$F5,$1A,$B8,$7E,$BD,$7F,$80,$7F,$2C,$FC,$3F,$03,$F5,$47
	.dc.b	$ED,$2F,$C7,$3F,$D2,$3E,$87,$C7,$C3,$87,$92,$C3,$85,$EF,$F3,$9F
	.dc.b	$C6,$57,$5E,$FF,$F1,$7F,$D1,$B4,$FC,$AC,$FF,$3C,$8C,$63,$CF,$47
	.dc.b	$E3,$1E,$BD,$3C,$15,$AD,$3B,$ED,$FE,$1C,$F0,$3E,$06,$3F,$BA,$FD
	.dc.b	$81,$FB,$63,$E5,$FF,$DC,$7F,$09,$F9,$1F,$0F,$D5,$7B,$AD,$77,$FF
	.dc.b	$6F,$FB,$35,$6B,$DC,$7E,$8F,$F8,$8D,$3B,$C3,$4C,$71,$FF,$76,$CF
	.dc.b	$A6,$BF,$EF,$2F,$C4,$BC,$D3,$9B,$E1,$AF,$83,$5C,$CF,$CB,$FD,$E1
	.dc.b	$FA,$53,$E2,$5F,$AD,$7F,$73,$FD,$C7,$F3,$5C,$3F,$4F,$95,$87,$0B
	.dc.b	$DF,$CF,$AD,$E8,$FB,$38,$FF,$0D,$87,$0F,$D3,$BD,$3A,$F8,$75,$B3
	.dc.b	$D1,$F1,$6B,$FE,$03,$E8,$F3,$CE,$74,$EE,$FE,$33,$DF,$7E,$C7,$FE
	.dc.b	$74,$F2,$7E,$37,$E5,$2F,$E0,$17,$91,$F8,$6B,$DF,$B7,$FA,$5E,$C5
	.dc.b	$EF,$E8,$B5,$97,$F9,$FB,$83,$FC,$07,$87,$6F,$CB,$BF,$CF,$6F,$2D
	.dc.b	$3B,$FE,$5B,$E3,$BF,$16,$5D,$3F,$2C,$FF,$23,$FC,$EB,$F7,$5F,$F0
	.dc.b	$3F,$D9,$BF,$1E,$7E,$65,$E2,$FA,$4F,$F2,$35,$EF,$E3,$7F,$CC,$D5
	.dc.b	$AE,$27,$E7,$9A,$CB,$FB,$D7,$17,$F9,$87,$C2,$FF,$6F,$63,$DF,$F7
	.dc.b	$0F,$E2,$3F,$67,$FB,$78,$E9,$B7,$87,$67,$E5,$9E,$6C,$BC,$BA,$F9
	.dc.b	$3F,$2B,$FD,$CF,$ED,$0F,$7D,$F1,$CF,$DE,$FF,$CF,$9E,$E7,$F9,$1F
	.dc.b	$F6,$EF,$D9,$DE,$EB,$5E,$B7,$EF,$7E,$D7,$FC,$9D,$EC,$7E,$F6,$BA
	.dc.b	$7A,$3F,$EA,$67,$87,$37,$F8,$C9,$E5,$9F,$6F,$8D,$F1,$EB,$E2,$8F
	.dc.b	$0F,$C8,$7F,$0E,$FD,$9B,$FD,$4D,$F8,$77,$F9,$1B,$F5,$67,$E3,$1F
	.dc.b	$ED,$AE,$AF,$E0,$F5,$AF,$B7,$56,$BF,$11,$FF,$79,$F0,$7F,$11,$AF
	.dc.b	$F5,$87,$17,$FA,$E9,$EB,$C1,$FE,$46,$79,$5F,$6F,$C9,$BC,$5F,$86
	.dc.b	$FF,$BA,$72,$EB,$A3,$E9,$FF,$D1,$3F,$84,$FF,$9B,$BF,$7E,$FF,$02
	.dc.b	$7E,$29,$FE,$2C,$F4,$BF,$D1,$5E,$2B,$8B,$F9,$75,$EF,$E9,$57,$BF
	.dc.b	$D7,$7F,$D1,$93,$FE,$78,$FD,$D7,$EF,$F1,$7F,$AE,$46,$78,$53,$CE
	.dc.b	$FF,$C7,$7D,$5E,$49,$DE,$3D,$BE,$CF,$F0,$B1,$FB,$B3,$FD,$0B,$F8
	.dc.b	$6B,$F5,$07,$E9,$8F,$C6,$FF,$CA,$BF,$85,$3F,$48,$FE,$F3,$C5,$FC
	.dc.b	$EF,$0E,$17,$BF,$C0,$3F,$DE,$34,$FD,$27,$B6,$6B,$DB,$FD,$B4,$F0
	.dc.b	$E6,$FD,$CF,$F8,$5C,$F4,$D3,$C3,$F4,$9F,$C7,$A3,$19,$7F,$BF,$BE
	.dc.b	$2F,$FC,$71,$CF,$FD,$67,$F0,$07,$E9,$0F,$D0,$1F,$A7,$BF,$4F,$7E
	.dc.b	$7C,$F8,$57,$EB,$BF,$E3,$9C,$B9,$63,$C7,$0E,$17,$BF,$8F,$D7,$1F
	.dc.b	$FA,$B3,$4E,$FF,$0F,$FE,$37,$F2,$FE,$F4,$E7,$1F,$13,$FE,$3D,$CB
	.dc.b	$AE,$DE,$4D,$BC,$7F,$90,$FA,$AB,$84,$E7,$47,$EC,$1F,$ED,$6F,$9A
	.dc.b	$78,$BF,$E2,$3F,$73,$FD,$A9,$F2,$7F,$E2,$3F,$C4,$7F,$DA,$95,$AF
	.dc.b	$AD,$FE,$2D,$F8,$AF,$F8,$93,$EB,$9F,$89,$6B,$2F,$F1,$66,$DF,$ED
	.dc.b	$03,$FA,$92,$FA,$1F,$C7,$63,$CF,$4E,$FD,$BE,$3A,$F1,$65,$FE,$30
	.dc.b	$FC,$27,$FD,$D7,$FA,$C3,$F3,$57,$C5,$BD,$67,$D7,$FD,$6F,$CB,$1F
	.dc.b	$A0,$7F,$BA,$F0,$E1,$7B,$AD,$7E,$6D,$9C,$7F,$DE,$5C,$3E,$71,$7C
	.dc.b	$1C,$3C,$B2,$EF,$DA,$FC,$7B,$CB,$7E,$49,$CF,$4F,$F8,$73,$F5,$C7
	.dc.b	$EB,$0F,$E0,$07,$E2,$6F,$03,$D3,$F8,$A3,$F7,$1F,$F0,$BE,$1C,$2F
	.dc.b	$75,$AA,$CE,$0F,$86,$FC,$37,$83,$FE,$2F,$C3,$87,$0F,$FB,$59,$E5
	.dc.b	$9F,$7E,$3F,$C0,$CF,$2C,$FA,$C7,$AE,$3F,$C4,$FF,$52,$FD,$E9,$F8
	.dc.b	$5B,$F3,$E7,$EC,$9F,$1B,$F7,$1F,$F3,$17,$99,$FE,$33,$C3,$85,$EE
	.dc.b	$B5,$FB,$3F,$4C,$7F,$D8,$B9,$7F,$BA,$AF,$6D,$8E,$F9,$76,$EA,$7E
	.dc.b	$0D,$FE,$46,$BE,$B3,$9F,$C5,$7F,$3D,$83,$F5,$8F,$F4,$1F,$DE,$BF
	.dc.b	$2A,$FF,$69,$FE,$C8,$FC,$59,$E1,$7E,$C3,$0E,$1F,$86,$7F,$9E,$6F
	.dc.b	$74,$F9,$AD,$65,$D2,$FF,$60,$0E,$3F,$69,$FB,$7F,$5F,$FC,$C1,$3E
	.dc.b	$2C,$3C,$3F,$96,$79,$A3,$1F,$14,$D7,$F8,$D6,$3F,$EB,$FF,$C1,$FF
	.dc.b	$CD,$1F,$9F,$3F,$5A,$7E,$A4,$FD,$25,$F7,$FF,$EC,$AF,$C5,$3E,$A7
	.dc.b	$7B,$1E,$38,$70,$BD,$FE,$36,$BF,$C8,$5E,$F3,$F5,$67,$DF,$3F,$E2
	.dc.b	$19,$72,$FC,$42,$D6,$B3,$E9,$A7,$FB,$6B,$F4,$EF,$F6,$F5,$F8,$30
	.dc.b	$F3,$4F,$FB,$7B,$F1,$DF,$F3,$EF,$E1,$8F,$CF,$DF,$0E,$FD,$7F,$EE
	.dc.b	$FA,$EF,$D7,$BF,$C6,$FC,$3F,$E7,$DC,$38,$5E,$FC,$4F,$E6,$75,$FE
	.dc.b	$85,$AD,$63,$EB,$E0,$FF,$48,$1F,$E8,$0C,$38,$BF,$EE,$3F,$3F,$FD
	.dc.b	$29,$1D,$33,$F0,$FC,$7F,$9A,$FC,$B1,$E4,$F9,$1F,$BD,$F3,$1F,$E9
	.dc.b	$7C,$7F,$01,$F3,$5F,$2F,$B9,$DA,$FC,$A5,$F9,$FF,$F4,$2F,$17,$FC
	.dc.b	$F3,$E2,$7F,$A8,$F0,$E1,$7B,$AF,$E8,$FF,$8F,$EB,$2E,$BD,$EC,$FF
	.dc.b	$D8,$B8,$F1,$85,$62,$B7,$5E,$2C,$FC,$3D,$3E,$8C,$38,$4E,$71,$7E
	.dc.b	$1B,$EC,$F1,$BF,$D6,$DF,$1A,$FD,$FD,$F9,$DB,$D6,$7E,$5B,$FE,$0F
	.dc.b	$F8,$07,$C4,$FF,$A5,$31,$E3,$87,$0B,$DC,$7F,$1B,$C3,$5C,$1F,$A7
	.dc.b	$CB,$B4,$7F,$83,$DF,$07,$E4,$97,$0E,$59,$FF,$87,$7E,$0F,$CD,$87
	.dc.b	$19,$CE,$3F,$C5,$FA,$3F,$D8,$5F,$87,$BF,$2B,$FF,$00,$7D,$B3,$EC
	.dc.b	$9F,$82,$3F,$64,$7E,$C9,$FE,$0A,$C3,$87,$D2,$7F,$AD,$AF,$74,$FB
	.dc.b	$3D,$64,$F4,$3F,$D2,$B8,$70,$DF,$FC,$20,$7F,$A7,$27,$C3,$87,$83
	.dc.b	$E9,$3F,$D5,$75,$DB,$5F,$26,$BF,$ED,$7C,$1E,$53,$DD,$FF,$74,$75
	.dc.b	$7F,$2F,$F9,$C7,$8B,$FE,$0D,$FC,$23,$FD,$C9,$E7,$70,$BD,$FC,$0A
	.dc.b	$F7,$F4,$BF,$EC,$3F,$C6,$5E,$7E,$F5,$FE,$31,$A7,$F3,$BF,$CE,$7F
	.dc.b	$C1,$D9,$72,$C6,$D6,$B2,$E9,$AF,$87,$CE,$FE,$41,$BF,$9E,$B5,$7E
	.dc.b	$0D,$FB,$C7,$FB,$C3,$F1,$57,$ED,$0F,$BB,$FE,$6D,$F1,$DF,$ED,$78
	.dc.b	$FD,$46,$3F,$EF,$BA,$DD,$6B,$0E,$3F,$B3,$F3,$BF,$B9,$8F,$EE,$CF
	.dc.b	$E6,$57,$C1,$C1,$FE,$D2,$8F,$07,$C2,$7F,$81,$AF,$ED,$4F,$3C,$3B
	.dc.b	$7D,$EF,$B3,$D9,$FE,$79,$2F,$CA,$1F,$A1,$3E,$E5,$FA,$3F,$FB,$7B
	.dc.b	$F3,$67,$E0,$EF,$B0,$FB,$79,$72,$C7,$8E,$1C,$38,$18,$70,$FB,$2D
	.dc.b	$E9,$FD,$2E,$F6,$FA,$9F,$0F,$EF,$1F,$DF,$A3,$5A,$F8,$F5,$FF,$77
	.dc.b	$FC,$4F,$A3,$95,$AF,$86,$B5,$A7,$83,$DA,$F7,$63,$FD,$67,$83,$D3
	.dc.b	$E9,$BF,$51,$FF,$2D,$7E,$F5,$FE,$82,$FC,$93,$CB,$7E,$EB,$5E,$FC
	.dc.b	$55,$FF,$A2,$3F,$69,$DD,$74,$BF,$A6,$FC,$6F,$CC,$FD,$51,$CF,$0E
	.dc.b	$56,$FC,$5F,$CB,$39,$CF,$C7,$F8,$CE,$3E,$71,$FA,$53,$EC,$DF,$AC
	.dc.b	$3F,$29,$7C,$03,$F6,$97,$8D,$EA,$C7,$8E,$1C,$2F,$74,$FD,$EB,$59
	.dc.b	$7E,$77,$5F,$E9,$7D,$5F,$31,$FF,$00,$63,$C9,$FC,$7B,$D3,$4F,$16
	.dc.b	$9E,$B9,$E7,$9F,$78,$6B,$FD,$A2,$7F,$C1,$5F,$A1,$7F,$34,$FA,$47
	.dc.b	$EF,$2F,$DB,$DF,$92,$7D,$0F,$C1,$BF,$C0,$D8,$70,$F8,$CF,$F1,$B5
	.dc.b	$EE,$BE,$AF,$4E,$A9,$97,$F4,$0F,$DF,$FF,$BD,$C6,$30,$FF,$1F,$3F
	.dc.b	$DA,$0E,$13,$E3,$C3,$A7,$E5,$1E,$79,$E1,$A7,$9B,$4F,$2F,$E4,$7F
	.dc.b	$D9,$BE,$83,$FC,$43,$C8,$FD,$91,$F8,$C3,$F1,$E7,$EB,$EF,$D8,$1F
	.dc.b	$B0,$BE,$3F,$FB,$D7,$B1,$FC,$93,$1E,$38,$70,$BF,$F6,$55,$6B,$27
	.dc.b	$D4,$FA,$78,$39,$6B,$E2,$C7,$17,$B7,$9E,$9F,$E6,$28,$C6,$BE,$58
	.dc.b	$FF,$3A,$5F,$E8,$3F,$0A,$7E,$B6,$FC,$6B,$F8,$DF,$D0,$BF,$74,$FE
	.dc.b	$A1,$F2,$7B,$AD,$7C,$07,$F8,$12,$7F,$AA,$7E,$E7,$9C,$3F,$43,$8F
	.dc.b	$4F,$D1,$2E,$1C,$32,$E9,$F8,$57,$E8,$AB,$C3,$7E,$0C,$7F,$C0,$9A
	.dc.b	$7E,$D3,$D9,$FF,$4F,$FB,$FF,$C5,$7F,$37,$7E,$AD,$FE,$5F,$F0,$F5
	.dc.b	$F8,$4F,$F1,$9A,$56,$BE,$A7,$FC,$FD,$D5,$F9,$DC,$BF,$CA,$17,$FC
	.dc.b	$7D,$FE,$57,$97,$FA,$B3,$E0,$3F,$72,$F8,$31,$E7,$9F,$93,$0E,$53
	.dc.b	$9C,$BC,$7B,$FF,$8A,$BD,$FF,$E5,$BF,$2F,$FE,$E7,$FC,$31,$F8,$F7
	.dc.b	$F2,$DF,$CE,$7F,$CC,$1E,$8F,$FB,$FA,$B5,$39,$CB,$FB,$37,$77,$9E
	.dc.b	$79,$70,$7F,$72,$7F,$DE,$5F,$A6,$7E,$03,$3E,$FF,$01,$FE,$DE,$C7
	.dc.b	$B7,$C7,$F8,$F0,$FF,$3D,$61,$DB,$89,$AB,$F3,$CD,$53,$F3,$8F,$E2
	.dc.b	$1F,$14,$FD,$25,$E6,$B3,$7D,$CB,$97,$F3,$9C,$1F,$D3,$31,$E3,$87
	.dc.b	$0B,$DF,$EB,$5A,$CF,$FD,$19,$5F,$C7,$7F,$3F,$F5,$46,$39,$9A,$E2
	.dc.b	$3C,$91,$8C,$7A,$7E,$5F,$E8,$C3,$BE,$DE,$5F,$CB,$FF,$96,$DB,$B5
	.dc.b	$F8,$1F,$E6,$FF,$64,$FD,$A7,$85,$FB,$4E,$FF,$C2,$5E,$EB,$5E,$8A
	.dc.b	$9F,$A4,$7F,$46,$AD,$6F,$FD,$C9,$ED,$EC,$7F,$40,$79,$78,$CC,$FF
	.dc.b	$81,$7E,$77,$E4,$D7,$C7,$39,$CF,$FC,$E1,$AF,$E5,$7F,$33,$7E,$93
	.dc.b	$FC,$FF,$C9,$7D,$8F,$FB,$B3,$E1,$1F,$B1,$FF,$58,$BD,$FC,$5F,$FB
	.dc.b	$06,$B5,$F4,$19,$C7,$FA,$47,$17,$F8,$22,$FA,$F9,$9F,$D5,$9E,$39
	.dc.b	$F6,$FC,$47,$F0,$95,$C7,$2E,$91,$F8,$18,$FE,$CF,$F5,$9F,$8A,$FC
	.dc.b	$C5,$FA,$0B,$F6,$77,$EC,$CF,$C7,$5F,$8F,$F0,$F3,$18,$70,$BD,$D3
	.dc.b	$F7,$9D,$4F,$F6,$AC,$BF,$2F,$F8,$A5,$C3,$8E,$7D,$F3,$FF,$7E,$73
	.dc.b	$FE,$F6,$FC,$93,$9F,$AF,$FF,$74,$FC,$6F,$FB,$37,$99,$FA,$07,$F9
	.dc.b	$87,$F4,$E7,$ED,$6F,$DC,$BF,$DD,$17,$BA,$D4,$FF,$B5,$33,$7D,$F7
	.dc.b	$F8,$AA,$BA,$72,$FF,$BA,$0F,$EA,$7F,$89,$BF,$54,$FF,$74,$47,$F6
	.dc.b	$28,$FF,$2F,$E3,$E1,$FA,$EF,$2C,$63,$EF,$9A,$FF,$8E,$B3,$FE,$31
	.dc.b	$F8,$87,$CE,$F8,$C7,$E5,$EF,$CD,$5F,$37,$FE,$18,$F8,$0F,$F0,$DD
	.dc.b	$EE,$B5,$3F,$E1,$CF,$B3,$CF,$3F,$FA,$1F,$83,$FC,$97,$7F,$6B,$F2
	.dc.b	$EF,$F0,$CB,$C7,$2E,$BE,$8F,$F8,$62,$31,$8F,$F0,$4B,$FF,$14,$FE
	.dc.b	$FC,$FC,$CF,$FD,$F7,$F3,$2F,$D7,$5F,$8A,$BF,$1E,$7E,$39,$C2,$F7
	.dc.b	$5A,$97,$EC,$D9,$F8,$8F,$F1,$5E,$5D,$7F,$1C,$B8,$70,$CB,$FC,$B5
	.dc.b	$A7,$39,$FE,$E5,$97,$83,$F2,$DF,$E3,$BF,$B6,$FF,$83,$B9,$7F,$D8
	.dc.b	$BF,$B4,$3E,$DD,$FB,$3B,$F5,$3F,$E2,$D9,$B0,$E1,$F5,$6C,$38,$5F
	.dc.b	$FB,$93,$E5,$BF,$A6,$57,$5B,$FF,$53,$BF,$E1,$CE,$DF,$E3,$4F,$2B
	.dc.b	$F0,$7A,$5E,$4F,$AC,$FF,$44,$65,$D7,$4F,$F4,$6E,$5F,$D5,$FA,$3F
	.dc.b	$DA,$BF,$A2,$F9,$BF,$72,$FC,$07,$E7,$3F,$EC,$0A,$D7,$E2,$CA,$D7
	.dc.b	$80,$AF,$F0,$95,$6B,$F4,$FF,$E5,$98,$F4,$E1,$FF,$5E,$BD,$B8,$99
	.dc.b	$FE,$37,$E1,$FF,$A2,$CE,$63,$F2,$1E,$A7,$F9,$6B,$E1,$5F,$8D,$FE
	.dc.b	$A7,$F2,$7F,$E5,$1F,$A6,$BC,$8F,$EE,$FE,$67,$F7,$1B,$DF,$AC,$BD
	.dc.b	$D3,$C3,$FF,$39,$D6,$B7,$FF,$4F,$BD,$BE,$37,$FC,$DA,$F2,$BE,$FA
	.dc.b	$78,$F4,$F1,$7C,$73,$5F,$F6,$96,$BF,$B8,$C8,$77,$FC,$01,$F9,$83
	.dc.b	$17,$D2,$FE,$67,$67,$E6,$FF,$ED,$AF,$6F,$FD,$53,$1F,$E8,$6B,$E1
	.dc.b	$5B,$FD,$17,$59,$76,$AF,$F4,$5F,$ED,$9F,$C0,$A3,$C1,$76,$79,$65
	.dc.b	$E0,$F2,$5E,$4F,$FB,$4B,$3E,$FA,$F9,$AB,$5F,$85,$FF,$78,$EA,$7E
	.dc.b	$6C,$FD,$5D,$FA,$B3,$F7,$E7,$EA,$2F,$CE,$3F,$3F,$E1,$FF,$88,$F1
	.dc.b	$E3,$87,$0B,$DD,$7F,$39,$AD,$62,$FC,$B3,$FB,$A6,$1C,$38,$FF,$CD
	.dc.b	$95,$AC,$79,$65,$FE,$51,$FA,$7F,$F6,$7D,$F6,$D3,$C9,$3F,$E5,$2B
	.dc.b	$FF,$1A,$7D,$03,$F4,$4F,$F7,$57,$87,$FC,$77,$F0,$EF,$C3,$7D,$07
	.dc.b	$ED,$2B,$5E,$B7,$F3,$7F,$33,$FD,$AB,$E2,$3E,$43,$59,$7F,$BC,$F8
	.dc.b	$7C,$BC,$1F,$EE,$0C,$7B,$7A,$DF,$E0,$D7,$9E,$BD,$FE,$93,$FC,$47
	.dc.b	$18,$CF,$FD,$C3,$FA,$37,$FA,$E3,$CD,$FF,$74,$7E,$1A,$FD,$05,$F8
	.dc.b	$3B,$F1,$37,$E6,$2F,$D3,$FF,$D9,$B8,$70,$BD,$D6,$BE,$D7,$3E,$0B
	.dc.b	$F6,$BF,$F6,$4D,$F6,$E3,$FF,$7F,$BC,$73,$ED,$F9,$DF,$8E,$FF,$DB
	.dc.b	$99,$74,$FC,$4F,$FC,$77,$EB,$3F,$3D,$FC,$0F,$BD,$FE,$7D,$FC,$D5
	.dc.b	$FB,$93,$F4,$4F,$E3,$F7,$BF,$0F,$F9,$27,$E0,$2B,$DF,$E1,$BA,$CB
	.dc.b	$F9,$F7,$0F,$CC,$3F,$57,$87,$17,$07,$F0,$18,$F0,$DF,$3F,$BB,$F2
	.dc.b	$C6,$30,$FF,$36,$7E,$8F,$FB,$4B,$7E,$58,$FC,$2B,$F9,$AF,$A8,$FD
	.dc.b	$25,$FA,$33,$F0,$C7,$DF,$BE,$01,$9F,$3C,$B9,$7C,$3F,$EB,$31,$E3
	.dc.b	$87,$F9,$3A,$F7,$A7,$F0,$4B,$F0,$7E,$3D,$FE,$AF,$9C,$CF,$93,$5F
	.dc.b	$E8,$EF,$FB,$2E,$33,$5D,$B5,$F3,$6B,$FC,$2E,$BC,$9A,$78,$F0,$F5
	.dc.b	$E7,$EA,$F4,$FF,$89,$E5,$FE,$95,$FC,$3D,$F9,$6F,$E3,$7C,$DF,$F0
	.dc.b	$57,$EB,$2F,$CF,$BF,$8E,$5E,$EB,$5F,$52,$AD,$7E,$BB,$F0,$F0,$F9
	.dc.b	$98,$F4,$FC,$AB,$D9,$18,$C3,$F3,$5B,$FA,$E3,$1E,$11,$F0,$E9,$E6
	.dc.b	$72,$79,$3F,$33,$C5,$7F,$9F,$3F,$19,$7D,$C7,$F8,$5B,$7F,$EF,$FF
	.dc.b	$B0,$FA,$5F,$F5,$97,$E0,$FF,$EE,$49,$D7,$D2,$A9,$FA,$FF,$A7,$1E
	.dc.b	$95,$E8,$7B,$74,$D7,$FD,$31,$97,$83,$EA,$FF,$05,$8F,$C7,$9E,$46
	.dc.b	$AF,$BF,$EA,$1F,$D0,$BE,$0B,$FA,$24,$FF,$A1,$BF,$22,$7E,$2E,$E5
	.dc.b	$7E,$6F,$FE,$45,$FD,$3F,$FC,$8F,$87,$0F,$C3,$18,$70,$BD,$FC,$02
	.dc.b	$DF,$B9,$FF,$C0,$1F,$91,$FC,$F6,$1C,$3E,$A3,$FD,$F3,$18,$C7,$C3
	.dc.b	$AF,$F2,$C8,$C6,$BE,$3D,$7F,$CA,$DF,$94,$FA,$67,$3A,$F8,$3F,$25
	.dc.b	$FA,$78,$3F,$46,$7E,$B3,$FE,$1B,$FC,$15,$F7,$CC,$9F,$B6,$FF,$62
	.dc.b	$EE,$FF,$73,$61,$C3,$EB,$FF,$CF,$97,$BA,$7C,$F7,$CC,$AD,$67,$F8
	.dc.b	$93,$FB,$6B,$E0,$FD,$FF,$EB,$3C,$F3,$F0,$E6,$D7,$CF,$39,$CB,$C9
	.dc.b	$F7,$3E,$AF,$95,$7D,$CB,$F7,$E7,$DE,$3C,$D7,$ED,$0F,$CE,$DF,$0D
	.dc.b	$F8,$0C,$78,$E1,$C2,$F7,$4F,$CB,$F5,$93,$F1,$8F,$C6,$61,$C3,$8F
	.dc.b	$D8,$1E,$7A,$78,$74,$FF,$5A,$CF,$2C,$FB,$47,$FA,$53,$CE,$FE,$39
	.dc.b	$F9,$87,$FB,$2B,$E2,$1F,$AA,$3F,$60,$7C,$47,$F7,$7F,$0F,$F2,$F7
	.dc.b	$BA,$D4,$FF,$7B,$F8,$CC,$FD,$97,$F3,$AD,$7C,$B7,$FE,$D6,$FC,$7B
	.dc.b	$F4,$8F,$1C,$BA,$FB,$BF,$84,$8C,$63,$FC,$57,$C2,$FD,$F6,$1F,$E2
	.dc.b	$4F,$3F,$FD,$9B,$FC,$02,$F6,$3F,$D4,$BF,$AE,$3F,$30,$7E,$F7,$FA
	.dc.b	$CB,$DD,$6A,$5F,$C0,$19,$FC,$75,$8F,$ED,$EF,$FD,$B3,$F2,$16,$73
	.dc.b	$87,$6F,$D3,$FF,$33,$5D,$3F,$11,$68,$FD,$4B,$F9,$3F,$E5,$DF,$D1
	.dc.b	$FF,$35,$FF,$46,$F6,$FF,$BF,$BF,$49,$7C,$23,$F1,$97,$E6,$BF,$D1
	.dc.b	$70,$E1,$F8,$67,$E2,$3D,$C6,$1C,$2F,$FD,$7B,$D0,$BD,$FC,$4F,$F4
	.dc.b	$37,$8E,$77,$A7,$FC,$97,$AF,$8B,$2E,$7A,$BF,$14,$F3,$D6,$B5,$F0
	.dc.b	$6B,$E6,$FD,$3B,$F8,$97,$8D,$FD,$87,$EC,$5F,$81,$3F,$5C,$7C,$C3
	.dc.b	$F3,$D7,$E9,$AF,$D3,$3F,$D8,$FF,$15,$F6,$B6,$65,$CB,$1E,$38,$70
	.dc.b	$BF,$E5,$D7,$BA,$EF,$FA,$17,$F7,$1C,$3F,$B9,$4E,$73,$F1,$47,$F7
	.dc.b	$CA,$E9,$9F,$8F,$3F,$74,$FF,$AA,$27,$39,$79,$BE,$33,$F8,$9F,$E1
	.dc.b	$9F,$ED,$6F,$D4,$1F,$9F,$BF,$56,$7E,$7C,$FA,$E7,$EE,$FF,$CE,$30
	.dc.b	$E1,$7B,$AD,$7D,$B6,$7F,$1D,$FD,$FE,$DF,$B6,$BE,$CD,$86,$7E,$0C
	.dc.b	$FB,$FE,$2F,$E4,$9C,$E5,$D7,$F2,$EF,$A7,$87,$FA,$1B,$F3,$D7,$F0
	.dc.b	$0B,$E3,$1F,$92,$3E,$75,$EE,$7F,$2C,$E4,$FF,$64,$E1,$C2,$F7,$5A
	.dc.b	$F8,$B7,$F0,$0E,$B2,$E9,$9F,$F8,$A3,$0E,$FF,$AB,$7D,$38,$F1,$88
	.dc.b	$E9,$A7,$87,$4F,$E7,$53,$CF,$3E,$F1,$FB,$DF,$07,$FC,$59,$9F,$F5
	.dc.b	$3F,$D8,$5F,$9E,$FF,$5C,$FD,$1B,$FD,$41,$FA,$87,$F6,$6A,$D4,$E7
	.dc.b	$F1,$17,$E0,$F9,$FA,$26,$3F,$DC,$6F,$FA,$F7,$E0,$96,$73,$87,$1D
	.dc.b	$B8,$5F,$83,$E8,$1A,$7F,$AD,$35,$FF,$20,$7E,$E5,$FE,$08,$FD,$77
	.dc.b	$FD,$85,$F8,$D7,$9B,$F6,$A7,$EC,$7F,$DE,$AB,$5F,$8C,$FF,$AD,$35
	.dc.b	$53,$F2,$9D,$7C,$13,$F7,$0F,$F0,$36,$1D,$78,$3F,$9D,$BC,$B3,$F8
	.dc.b	$7D,$3F,$DA,$23,$19,$3C,$7F,$ED,$D7,$FE,$C8,$F8,$0F,$F1,$97,$EF
	.dc.b	$6F,$CB,$7F,$DB,$DF,$A2,$7F,$95,$7F,$59,$FE,$FD,$7B,$FB,$65,$EF
	.dc.b	$F0,$65,$FF,$96,$FF,$6B,$DD,$FF,$16,$B7,$E1,$7F,$E0,$48,$E5,$85
	.dc.b	$74,$F3,$BE,$71,$CB,$3F,$2E,$7E,$4F,$C4,$FF,$C0,$7F,$AB,$F9,$FB
	.dc.b	$1E,$5B,$F7,$FF,$C6,$7C,$0B,$F3,$0F,$F1,$3D,$EE,$B5,$E5,$7F,$26
	.dc.b	$EB,$CE,$7F,$E2,$4D,$7F,$B2,$DF,$C3,$F9,$1F,$F1,$A3,$C7,$3F,$CC
	.dc.b	$73,$78,$AF,$FC,$57,$97,$4F,$C5,$FF,$3B,$FA,$EB,$F3,$87,$E7,$BF
	.dc.b	$B8,$F9,$2F,$D7,$7D,$CB,$F7,$27,$C3,$7F,$1F,$C3,$85,$EF,$D7,$FF
	.dc.b	$1C,$FC,$BE,$BC,$BF,$F4,$65,$6B,$8F,$FD,$2A,$F7,$F7,$3F,$5D,$1B
	.dc.b	$02,$3A,$FB,$DF,$D4,$70,$61,$E8,$DB,$CF,$39,$BF,$F1,$8F,$EB,$6F
	.dc.b	$CD,$9F,$15,$FE,$94,$FC,$F1,$FB,$A7,$FB,$CF,$F2,$FF,$F2,$AD,$EE
	.dc.b	$B5,$3F,$EF,$7F,$DC,$73,$83,$F5,$0F,$F3,$05,$EF,$2F,$F0,$AB,$C7
	.dc.b	$F1,$06,$7D,$BE,$13,$C7,$87,$8A,$73,$97,$FB,$C7,$8F,$FD,$61,$FA
	.dc.b	$67,$F8,$4F,$E2,$BF,$C0,$1E,$63,$F6,$37,$EB,$8F,$DD,$3E,$36,$EC
	.dc.b	$38,$78,$BF,$E9,$5B,$DF,$E4,$7A,$9E,$95,$FE,$C4,$7F,$C4,$1F,$AC
	.dc.b	$7D,$0E,$EF,$EB,$11,$CB,$3F,$07,$EC,$7F,$C8,$23,$18,$FF,$8D,$70
	.dc.b	$E9,$F8,$8B,$F7,$0F,$F2,$CF,$EC,$3F,$E7,$7F,$9E,$7E,$61,$A9,$CF
	.dc.b	$E1,$89,$7E,$63,$CB,$DA,$F1,$73,$7C,$2C,$BA,$67,$E7,$7B,$7E,$16
	.dc.b	$D3,$FD,$09,$F9,$1F,$96,$73,$1F,$E8,$7F,$C5,$7D,$1E,$8B,$C0,$FF
	.dc.b	$2B,$FD,$EF,$FA,$E7,$E1,$DF,$A4,$FC,$5A,$D7,$E0,$3F,$E8,$AE,$8F
	.dc.b	$E3,$7F,$97,$E7,$EA,$BC,$BC,$1F,$94,$E9,$FE,$29,$A6,$23,$E9,$63
	.dc.b	$CB,$4F,$F6,$37,$C4,$78,$AB,$C3,$97,$F1,$BB,$FB,$3F,$82,$BE,$A9
	.dc.b	$F9,$8B,$F6,$87,$C5,$3F,$6C,$FC,$D7,$91,$C2,$F7,$5A,$9F,$CF,$78
	.dc.b	$99,$F8,$0F,$26,$5D,$72,$B8,$70,$CB,$A7,$EA,$FF,$D7,$2B,$59,$73
	.dc.b	$F3,$F1,$FA,$57,$FB,$0B,$F6,$7F,$F0,$37,$D6,$3F,$7B,$FF,$41,$F9
	.dc.b	$7F,$E1,$FC,$57,$BF,$A8,$FF,$12,$D7,$F9,$02,$B5,$3F,$EC,$CE,$3F
	.dc.b	$ED,$91,$8E,$4F,$3C,$F1,$C7,$F1,$D3,$C3,$F1,$F6,$7E,$2F,$37,$D1
	.dc.b	$FB,$4B,$C2,$7E,$68,$FC,$29,$FA,$0B,$F5,$E7,$C2,$7D,$0C,$38,$5E
	.dc.b	$EB,$5F,$96,$E6,$FE,$4D,$75,$E1,$F3,$F9,$BF,$C5,$1F,$F5,$1C,$72
	.dc.b	$CB,$BE,$7E,$43,$C6,$F8,$8F,$F0,$4F,$E2,$2F,$D2,$BF,$D1,$FF,$AE
	.dc.b	$7F,$AA,$3F,$05,$7E,$06,$FC,$77,$F7,$B7,$BF,$C3,$57,$BA,$7E,$3F
	.dc.b	$AC,$7F,$D3,$3A,$F9,$8F,$E5,$38,$77,$AF,$CB,$3C,$EF,$C1,$F3,$BE
	.dc.b	$4B,$F1,$FE,$08,$D1,$E1,$7A,$3F,$45,$FC,$97,$EA,$3F,$B8,$FA,$DF
	.dc.b	$CA,$F3,$F9,$A7,$3E,$E7,$F8,$53,$C7,$FE,$E9,$DF,$FE,$63,$DD,$5F
	.dc.b	$9B,$E1,$F3,$1F,$EB,$B7,$DA,$E2,$37,$E6,$33,$F0,$7B,$5E,$38,$C6
	.dc.b	$5F,$E2,$2E,$FF,$9B,$FA,$99,$7E,$7F,$EB,$F7,$3F,$D1,$BF,$B0,$7C
	.dc.b	$CA,$D4,$E7,$E8,$7E,$06,$FE,$79,$E5,$B7,$8D,$FC,$8B,$71,$EB,$C8
	.dc.b	$CB,$F7,$97,$C6,$73,$0F,$9B,$FF,$24,$FB,$AF,$D7,$9F,$A6,$BF,$39
	.dc.b	$FE,$2D,$F5,$EF,$D2,$9F,$BA,$7F,$3E,$C3,$85,$EF,$DC,$5E,$FE,$33
	.dc.b	$59,$74,$E0,$F9,$B8,$77,$FD,$3B,$F8,$B3,$CB,$4E,$FF,$45,$E4,$BF
	.dc.b	$1C,$78,$BF,$25,$F3,$FA,$8F,$DD,$5F,$A7,$BF,$3D,$7E,$52,$FC,$AD
	.dc.b	$FB,$4F,$9F,$0E,$17,$BA,$D7,$C0,$67,$1F,$1D,$3F,$21,$B1,$DB,$2E
	.dc.b	$B9,$78,$A3,$18,$7F,$9B,$BA,$FF,$D5,$DF,$BB,$BE,$1B,$FA,$67,$A8
	.dc.b	$FD,$39,$FA,$53,$F3,$07,$E1,$FF,$DA,$30,$E1,$F3,$1F,$AF,$BD,$D7
	.dc.b	$1C,$BA,$E4,$FD,$BF,$F9,$F6,$1C,$3A,$FF,$32,$F2,$D3,$C1,$A7,$97
	.dc.b	$DE,$FE,$0D,$39,$CF,$C7,$F8,$4F,$FA,$2B,$F3,$37,$EA,$EF,$DF,$1F
	.dc.b	$AA,$3E,$03,$FD,$D1,$FB,$D7,$F1,$AE,$1F,$F6,$66,$1C,$2F,$75,$AF
	.dc.b	$96,$FF,$01,$CF,$4F,$95,$FE,$A8,$EF,$87,$F5,$EF,$1C,$BB,$7E,$49
	.dc.b	$FE,$16,$9E,$3A,$78,$E7,$3F,$17,$E7,$F5,$7F,$DF,$5F,$1A,$FD,$35
	.dc.b	$F8,$5F,$F5,$9F,$07,$C3,$5A,$9C,$FC,$B7,$E2,$78,$BD,$32,$D7,$CA
	.dc.b	$BF,$6D,$B0,$CB,$B6,$5F,$D9,$76,$EB,$39,$CB,$C3,$F9,$17,$CB,$F7
	.dc.b	$FF,$C1,$FC,$5F,$D5,$7F,$71,$FF,$80,$1E,$9F,$C4,$9C,$FC,$B2,$56
	.dc.b	$FA,$6D,$7D,$47,$C8,$BF,$F0,$AE,$9F,$21,$E3,$7F,$EE,$EE,$1F,$A7
	.dc.b	$18,$FB,$DF,$F6,$2F,$9F,$FA,$5F,$5D,$F4,$9F,$A6,$FC,$17,$FD,$7F
	.dc.b	$F1,$9F,$E3,$8A,$D4,$E7,$F0,$97,$E4,$79,$F4,$BF,$A8,$6D,$FD,$3C
	.dc.b	$C6,$5F,$E0,$FC,$38,$FA,$0C,$FF,$88,$F9,$3F,$D5,$67,$31,$FC,$46
	.dc.b	$FC,$1E,$B7,$FB,$A7,$F0,$77,$E5,$CF,$D3,$3F,$C1,$DF,$AB,$EB,$F1
	.dc.b	$A5,$6A,$7F,$C2,$5F,$9E,$72,$FC,$73,$C5,$8F,$FB,$A3,$2F,$37,$3B
	.dc.b	$6F,$14,$F7,$D3,$B7,$E1,$FF,$1A,$31,$93,$F5,$9F,$93,$EF,$BE,$DD
	.dc.b	$F8,$D3,$F7,$B7,$EE,$7F,$E6,$CF,$8E,$F9,$57,$BA,$D4,$FF,$9C,$3F
	.dc.b	$5C,$CF,$E9,$9F,$1E,$BA,$70,$FF,$50,$E8,$FE,$37,$9F,$6C,$FA,$FC
	.dc.b	$37,$8A,$31,$97,$F6,$2B,$FF,$62,$FD,$73,$EF,$3F,$D6,$DF,$B6,$7F
	.dc.b	$09,$FD,$67,$F8,$D7,$3B,$F4,$4F,$F0,$35,$EF,$F1,$65,$EF,$E1,$F9
	.dc.b	$E3,$FE,$E7,$CB,$FC,$47,$87,$08,$FF,$61,$46,$0F,$F5,$1E,$1C,$EF
	.dc.b	$C1,$1C,$B0,$EB,$F0,$BF,$C8,$7F,$06,$FF,$73,$CA,$DF,$91,$FF,$36
	.dc.b	$FD,$1D,$FB,$2F,$FA,$FF,$F1,$BF,$F1,$E6,$FE,$9C,$3B,$DE,$EB,$5F
	.dc.b	$BA,$7C,$28,$FE,$7F,$3E,$53,$B2,$FD,$FB,$FD,$CD,$97,$83,$9D,$9F
	.dc.b	$83,$F1,$2F,$F7,$6E,$1C,$E7,$39,$7F,$B4,$7C,$0F,$4E,$1F,$E6,$2F
	.dc.b	$8C,$7D,$63,$F7,$07,$E6,$FF,$E6,$9F,$C7,$3E,$AF,$D6,$AF,$75,$AF
	.dc.b	$86,$7E,$D1,$AF,$9E,$F2,$DE,$F7,$F5,$69,$DF,$4E,$DA,$7F,$11,$E5
	.dc.b	$F8,$95,$F2,$EF,$B5,$6B,$DF,$F9,$FF,$6A,$FC,$EB,$F3,$3F,$29,$FD
	.dc.b	$93,$E6,$3F,$C0,$DF,$69,$FE,$E8,$F0,$BF,$C9,$57,$BF,$31,$7B,$F3
	.dc.b	$5F,$AA,$EF,$6F,$F7,$1D,$F5,$D7,$FD,$71,$8F,$2D,$3F,$4F,$5F,$4E
	.dc.b	$BA,$63,$E1,$FC,$03,$CB,$5E,$4E,$BF,$ED,$1F,$05,$E8,$EA,$FD,$77
	.dc.b	$C0,$3F,$7C,$FE,$D1,$D9,$FC,$1B,$9B,$E0,$E8,$BD,$D6,$BF,$40,$F3
	.dc.b	$6D,$FE,$54,$9F,$DC,$3E,$35,$F9,$5F,$A2,$3D,$71,$CB,$CC,$7E,$85
	.dc.b	$E3,$F4,$9A,$F8,$E5,$F9,$BF,$CE,$FD,$93,$F8,$37,$EA,$CF,$DE,$37
	.dc.b	$F4,$AF,$CF,$FC,$7C,$2B,$DD,$6B,$E1,$F3,$0F,$6F,$C9,$C1,$FE,$6C
	.dc.b	$8F,$E0,$EF,$F1,$CE,$BF,$AF,$1C,$B6,$FF,$10,$CF,$2F,$65,$AF,$8E
	.dc.b	$7F,$94,$CB,$9B,$D2,$87,$E7,$2F,$E0,$07,$DE,$7F,$91,$FB,$3F,$D3
	.dc.b	$FE,$33,$EB,$39,$72,$FF,$A4,$70,$E1,$7B,$FC,$2B,$FA,$DF,$E8,$DA
	.dc.b	$CF,$B5,$69,$FF,$57,$DF,$1C,$FF,$DE,$51,$83,$6F,$8B,$ED,$FF,$C7
	.dc.b	$78,$7F,$5F,$AE,$18,$F7,$F0,$7D,$5B,$7B,$31,$FE,$D5,$8F,$F0,$EF
	.dc.b	$B0,$F4,$7E,$33,$FE,$1E,$FC,$A3,$C9,$5A,$F6,$BF,$CC,$1C,$0A,$FF
	.dc.b	$11,$FE,$D3,$AF,$9B,$F8,$95,$FD,$86,$FB,$ED,$FE,$B1,$9F,$91,$FA
	.dc.b	$EE,$67,$86,$9E,$1C,$BA,$7D,$E7,$09,$FA,$17,$D3,$5F,$2F,$E9,$D9
	.dc.b	$97,$E7,$1F,$4F,$F1,$AF,$F6,$7F,$EA,$CF,$D7,$BF,$95,$7E,$67,$EA
	.dc.b	$AD,$73,$AB,$5E,$DF,$F8,$37,$9B,$E2,$57,$3C,$BE,$71,$DB,$81,$F8
	.dc.b	$EF,$A6,$B7,$0D,$3C,$3A,$7E,$22,$73,$1E,$4F,$AA,$FF,$69,$7E,$21
	.dc.b	$FC,$F7,$1F,$C6,$7D,$DB,$F5,$89,$F9,$B3,$F0,$97,$EB,$0F,$DB,$BF
	.dc.b	$DD,$18,$F1,$C5,$8F,$1C,$58,$F1,$C3,$87,$E3,$1F,$ED,$6E,$2F,$F7
	.dc.b	$15,$F7,$FB,$8F,$A5,$97,$38,$E5,$97,$D9,$8C,$65,$E4,$D3,$C7,$F3
	.dc.b	$9F,$A3,$9C,$CF,$82,$7C,$FF,$3D,$9C,$BF,$DE,$1F,$80,$FB,$7D,$87
	.dc.b	$F0,$07,$F9,$1F,$F3,$27,$CC,$FF,$A5,$73,$F6,$5E,$EB,$52,$F8,$FF
	.dc.b	$56,$DF,$B3,$AF,$91,$EF,$FA,$B3,$EF,$B7,$F1,$AA,$FF,$19,$65,$D2
	.dc.b	$3F,$C2,$99,$78,$7C,$DF,$35,$73,$9C,$FD,$BB,$57,$85,$AA,$F4,$FE
	.dc.b	$F5,$FC,$AF,$F0,$8F,$ED,$1E,$BB,$F3,$A7,$E9,$BE,$6B,$DF,$C2,$2F
	.dc.b	$75,$FD,$93,$7F,$B5,$BB,$2F,$E6,$D8,$7D,$1E,$0F,$E2,$0F,$83,$F7
	.dc.b	$DF,$73,$FC,$4E,$B8,$65,$D7,$E6,$FF,$C6,$5A,$F9,$E7,$3A,$79,$BE
	.dc.b	$2F,$EC,$7B,$FD,$34,$FF,$61,$FE,$58,$FC,$7F,$C5,$FA,$82,$FF,$BA
	.dc.b	$70,$7F,$6A,$BD,$D6,$A5,$F3,$0A,$D7,$C2,$B1,$FC,$6E,$5D,$E3,$F8
	.dc.b	$47,$07,$E3,$63,$13,$E1,$C9,$9F,$86,$BA,$56,$B3,$F2,$ED,$EA,$9F
	.dc.b	$E6,$9D,$0F,$D4,$5F,$8E,$FE,$A3,$C7,$FF,$15,$7E,$4E,$FD,$BB,$9E
	.dc.b	$1C,$2F,$75,$FA,$EE,$85,$BD,$4D,$DF,$ED,$6F,$AB,$E9,$39,$7E,$8B
	.dc.b	$EE,$7A,$65,$E1,$C3,$FA,$4F,$1F,$F8,$4B,$4F,$34,$E7,$E4,$3D,$57
	.dc.b	$FC,$07,$F5,$F7,$B3,$FD,$3B,$F6,$FF,$17,$EB,$4D,$3C,$79,$2F,$7F
	.dc.b	$0E,$BD,$FE,$CD,$A9,$7D,$D7,$97,$3E,$D9,$DC,$79,$3D,$F8,$5F,$4B
	.dc.b	$E4,$9C,$E7,$D6,$7F,$C8,$3C,$1F,$E8,$7C,$3F,$B2,$7E,$50,$FD,$29
	.dc.b	$FA,$8B,$F4,$67,$E4,$4F,$DA,$BF,$C4,$FF,$85,$FE,$1F,$EF,$7F,$3B
	.dc.b	$8F,$1C,$38,$5E,$FE,$8F,$FD,$7D,$A7,$7D,$17,$B7,$F8,$B6,$1C,$B4
	.dc.b	$F1,$46,$0F,$2B,$E4,$D3,$C7,$A7,$F8,$E2,$7C,$46,$B3,$F0,$C7,$F8
	.dc.b	$72,$32,$78,$3E,$75,$F5,$0F,$9E,$7E,$CB,$FD,$4B,$07,$E9,$3C,$35
	.dc.b	$FF,$0E,$5E,$EB,$5F,$17,$9C,$7F,$C8,$F9,$79,$CE,$D5,$FC,$D3,$6F
	.dc.b	$8B,$3E,$0C,$FB,$DF,$FC,$01,$DF,$E2,$65,$DA,$73,$F8,$4F,$F3,$5C
	.dc.b	$3E,$6F,$E1,$2F,$DE,$9F,$5E,$FC,$6B,$FC,$7B,$FB,$AF,$F9,$17,$3E
	.dc.b	$77,$BA,$D7,$BE,$A7,$B3,$ED,$AE,$9C,$3E,$9D,$BF,$DD,$3F,$8E,$FA
	.dc.b	$63,$F0,$B3,$CF,$3F,$F2,$5F,$C2,$F9,$67,$3A,$F9,$27,$FC,$9F,$E0
	.dc.b	$66,$5F,$A2,$3F,$03,$7E,$10,$FC,$7D,$FA,$C7,$F9,$CB,$F7,$DF,$E6
	.dc.b	$FC,$9F,$87,$C3,$85,$EE,$B5,$FB,$07,$DC,$9C,$E5,$F4,$5E,$DF,$9B
	.dc.b	$FA,$63,$C3,$9F,$83,$3E,$FE,$1F,$FB,$D7,$3E,$D3,$9C,$BF,$D0,$7F
	.dc.b	$9E,$7A,$72,$7E,$92,$FC,$77,$FA,$FF,$B7,$FE,$73,$FC,$F1,$FA,$8F
	.dc.b	$8E,$B5,$F2,$3F,$CC,$FE,$25,$5F,$E4,$FD,$7D,$95,$CF,$83,$D0,$BE
	.dc.b	$BB,$D1,$A7,$83,$4E,$F8,$FF,$5B,$BE,$D1,$87,$CD,$FB,$D3,$F1,$8F
	.dc.b	$EA,$5F,$8B,$BF,$73,$FF,$5A,$7C,$3F,$C7,$AD,$29,$D7,$DE,$FE,$F7
	.dc.b	$6F,$A1,$5F,$DD,$F1,$F9,$8F,$56,$E3,$DB,$1F,$D7,$57,$08,$C7,$D6
	.dc.b	$7E,$07,$F2,$37,$E2,$BF,$EC,$9F,$C5,$7F,$C4,$9F,$BF,$FE,$CB,$F2
	.dc.b	$DF,$DA,$D6,$BF,$0B,$56,$BD,$3F,$CC,$67,$FC,$26,$B5,$AF,$EB,$DE
	.dc.b	$BF,$B0,$7D,$F7,$95,$F1,$9E,$B9,$F8,$7E,$85,$A7,$FA,$1B,$1E,$9E
	.dc.b	$5F,$F2,$1F,$DD,$FF,$9B,$7E,$51,$FE,$17,$FC,$CF,$CB,$39,$F8,$94
	.dc.b	$FF,$86,$FF,$32,$CF,$DC,$78,$6B,$F8,$CE,$BF,$E1,$8F,$5F,$FC,$03
	.dc.b	$C0,$9E,$58,$7C,$DD,$3C,$3F,$02,$D7,$C3,$E0,$7C,$9F,$B4,$7D,$B1
	.dc.b	$FF,$32,$7E,$CA,$DF,$FC,$CD,$8F,$C5,$9D,$9A,$87,$D8,$7C,$E8,$FE
	.dc.b	$C9,$BF,$C9,$7A,$B7,$0E,$DF,$66,$C5,$F9,$97,$8E,$73,$1F,$EB,$2C
	.dc.b	$BF,$9E,$7E,$21,$FE,$FE,$FD,$87,$FC,$25,$F9,$1F,$FA,$BB,$C3,$7C
	.dc.b	$9F,$F3,$2A,$D7,$9E,$AD,$74,$E7,$F1,$46,$3F,$B9,$B4,$3F,$61,$C3
	.dc.b	$AC,$39,$7C,$8F,$CB,$EB,$FF,$17,$DF,$6C,$BF,$D7,$FF,$65,$FC,$4E
	.dc.b	$1F,$8F,$FF,$29,$F8,$07,$E0,$CF,$CF,$FF,$C7,$5F,$97,$F9,$EB,$53
	.dc.b	$9F,$45,$FC,$01,$9F,$CB,$BC,$D5,$FC,$CF,$AF,$FA,$16,$7D,$BE,$17
	.dc.b	$F9,$C6,$3D,$7F,$20,$F1,$5F,$4D,$3C,$78,$7C,$5F,$95,$7E,$72,$FD
	.dc.b	$A9,$FB,$37,$E9,$3F,$12,$7E,$73,$BA,$D4,$E7,$CC,$C6,$3E,$2C,$5F
	.dc.b	$AA,$7C,$B7,$AE,$5D,$3D,$37,$E9,$5F,$16,$73,$8F,$EC,$BE,$BB,$F9
	.dc.b	$D7,$E8,$4E,$CF,$F0,$C7,$A7,$FE,$07,$FC,$B5,$F5,$4F,$C9,$BF,$B7
	.dc.b	$DE,$FE,$A5,$7B,$A7,$EB,$DA,$C5,$F9,$AF,$F3,$7C,$38,$71,$7F,$6B
	.dc.b	$79,$61,$C6,$38,$63,$D3,$D6,$69,$FE,$36,$C7,$FC,$DF,$87,$9F,$F7
	.dc.b	$5F,$F5,$FF,$D9,$FF,$84,$3F,$21,$7E,$29,$E1,$9C,$FC,$EA,$5E,$F7
	.dc.b	$97,$F0,$5F,$26,$4B,$A7,$0F,$73,$E2,$C6,$7F,$11,$FF,$5F,$E3,$F9
	.dc.b	$3E,$00,$D7,$E5,$7E,$00,$F8,$9F,$E6,$5F,$85,$3F,$49,$FE,$85,$ED
	.dc.b	$7F,$89,$EF,$75,$A9,$FE,$A5,$39,$CF,$F8,$46,$3F,$1D,$FF,$5A,$B7
	.dc.b	$1E,$DF,$1F,$FE,$CE,$F8,$4F,$3C,$63,$3F,$F0,$95,$FF,$A5,$B2,$7E
	.dc.b	$D0,$FD,$6F,$FC,$6F,$FA,$3B,$F1,$C7,$EB,$6E,$5F,$F1,$6E,$1C,$2F
	.dc.b	$75,$AF,$C8,$F3,$8B,$F1,$BF,$EE,$97,$BE,$0F,$F1,$43,$C7,$91,$A7
	.dc.b	$7F,$CD,$FF,$C1,$53,$98,$F1,$7E,$0F,$8F,$C1,$3F,$D9,$9F,$AB,$3F
	.dc.b	$71,$F9,$9F,$BE,$7A,$7A,$3F,$CA,$18,$FF,$A2,$7F,$15,$5E,$E1,$5E
	.dc.b	$5A,$DE,$1F,$D8,$75,$FE,$74,$7D,$AA,$FF,$49,$79,$5E,$A9,$F0,$E5
	.dc.b	$F6,$F9,$3C,$99,$F6,$8F,$BD,$F4,$BE,$AF,$E0,$07,$AF,$F8,$FF,$41
	.dc.b	$FB,$5B,$F3,$E7,$E7,$FF,$32,$F7,$E4,$7F,$9C,$6B,$5F,$9A,$E7,$1F
	.dc.b	$97,$4F,$2E,$C7,$5A,$7D,$D7,$86,$31,$F3,$FF,$E3,$5F,$CD,$3F,$C3
	.dc.b	$3F,$9D,$3F,$37,$7D,$03,$F7,$F7,$EC,$3F,$80,$EB,$FF,$3D,$61,$C2
	.dc.b	$F7,$5A,$FD,$CB,$FC,$DF,$3F,$E2,$4F,$C2,$3F,$C6,$4F,$6E,$DF,$F3
	.dc.b	$33,$C7,$2E,$DF,$8F,$78,$FE,$CD,$AF,$8E,$7C,$57,$E7,$FC,$83,$F5
	.dc.b	$5F,$A2,$BF,$51,$7D,$6B,$EE,$DF,$AD,$F8,$AF,$75,$A9,$FE,$B7,$C5
	.dc.b	$EA,$F5,$3F,$DE,$B9,$7D,$E1,$BB,$0F,$DC,$47,$A1,$E7,$97,$E8,$E3
	.dc.b	$B7,$B0,$D3,$C9,$3F,$E3,$B9,$F8,$7F,$8D,$BF,$66,$7E,$A8,$FD,$E7
	.dc.b	$FC,$99,$FA,$23,$F2,$8D,$D6,$A7,$3F,$85,$F2,$FA,$1F,$97,$8F,$3C
	.dc.b	$BE,$5B,$D7,$2E,$9F,$75,$FE,$34,$F7,$7F,$88,$D7,$B2,$B5,$8F,$F9
	.dc.b	$6F,$F0,$2F,$F6,$CF,$CD,$3E,$AD,$E4,$BF,$30,$FF,$26,$7E,$DD,$FE
	.dc.b	$79,$FC,$7F,$FD,$D7,$7B,$AD,$7D,$E2,$BF,$CF,$DF,$91,$6B,$1E,$7A
	.dc.b	$FF,$BB,$C6,$1D,$FE,$67,$FD,$B8,$F3,$CF,$BC,$F9,$2B,$C7,$1E,$2F
	.dc.b	$13,$CF,$87,$F6,$2F,$C2,$DF,$80,$FF,$51,$FC,$17,$FB,$AF,$82,$FC
	.dc.b	$8B,$E0,$E8,$C3,$85,$EF,$F3,$AF,$5C,$7F,$6C,$D7,$CD,$5B,$FA,$1F
	.dc.b	$2C,$F0,$F7,$FF,$D1,$18,$F4,$8C,$7E,$55,$FE,$9F,$FC,$55,$FB,$5F
	.dc.b	$F8,$EF,$F1,$6F,$F2,$DF,$E8,$1E,$3F,$4B,$FD,$A7,$5A,$FC,$0B,$FD
	.dc.b	$A5,$CE,$AF,$47,$C5,$EB,$1F,$F6,$B7,$07,$FB,$37,$0E,$0D,$8E,$D6
	.dc.b	$FC,$B3,$EF,$F2,$7F,$BB,$EF,$B6,$9E,$2F,$88,$FE,$A7,$3F,$17,$1F
	.dc.b	$EA,$BE,$13,$F3,$F7,$EA,$5F,$EC,$7F,$80,$7E,$95,$FA,$CF,$D9,$58
	.dc.b	$F1,$C7,$FC,$09,$87,$0F,$C6,$FF,$89,$5F,$EC,$BF,$6F,$DF,$47,$E4
	.dc.b	$B4,$ED,$A7,$EC,$31,$F1,$CE,$71,$7E,$99,$99,$E2,$79,$A3,$B6,$BE
	.dc.b	$4F,$C1,$B8,$CE,$70,$F4,$67,$FE,$55,$C7,$F4,$3E,$C3,$F3,$DF,$E3
	.dc.b	$DE,$7F,$F8,$BB,$F4,$F7,$E8,$8F,$D4,$BB,$E6,$C3,$87,$C2,$FE,$E3
	.dc.b	$ED,$58,$70,$DB,$75,$D7,$F7,$3F,$47,$37,$C5,$8E,$38,$FE,$85,$E7
	.dc.b	$A7,$87,$E6,$7D,$33,$DF,$6F,$2E,$7D,$BB,$FE,$BF,$D2,$7E,$EB,$C2
	.dc.b	$FE,$1D,$EF,$3F,$6F,$FE,$21,$FB,$13,$F6,$4F,$F7,$6E,$FF,$EB,$DD
	.dc.b	$3F,$C2,$38,$70,$BD,$D3,$F3,$0F,$B7,$97,$5C,$9F,$B0,$7A,$A3,$11
	.dc.b	$71,$E6,$FF,$1F,$C3,$C3,$F6,$DE,$69,$CC,$F9,$46,$8F,$6F,$FC,$47
	.dc.b	$FA,$13,$F3,$6F,$67,$98,$FC,$2F,$F9,$8F,$E4,$37,$ED,$5F,$0A,$C3
	.dc.b	$86,$9E,$CC,$38,$7C,$63,$0E,$1F,$11,$BC,$FA,$FD,$BF,$E2,$F1,$E3
	.dc.b	$8F,$FA,$BF,$1F,$03,$F9,$38,$CF,$37,$EC,$F5,$F1,$8F,$C2,$FD,$38
	.dc.b	$FA,$23,$3F,$93,$FF,$99,$BF,$34,$7E,$ED,$EB,$FD,$1B,$FC,$97,$F4
	.dc.b	$6F,$9D,$FF,$32,$D6,$B0,$56,$BF,$4E,$CE,$1F,$E6,$3A,$FF,$5A,$ED
	.dc.b	$FD,$57,$E5,$11,$C6,$FF,$CA,$FF,$41,$E1,$BF,$07,$DD,$F7,$C3,$F6
	.dc.b	$FE,$AF,$F4,$2F,$DA,$1F,$BC,$7F,$A8,$3F,$0A,$7E,$C1,$FE,$B0,$E1
	.dc.b	$7E,$77,$C3,$1F,$F2,$CE,$1C,$3F,$0D,$FF,$54,$DE,$FE,$3B,$54,$AE
	.dc.b	$BF,$C0,$1F,$A2,$D3,$C1,$A7,$FA,$BA,$BF,$80,$79,$7F,$E8,$B7,$AC
	.dc.b	$78,$F1,$FF,$60,$F0,$7F,$16,$BE,$39,$F8,$3F,$19,$F6,$63,$FC,$47
	.dc.b	$F5,$8F,$EA,$1F,$5D,$FE,$DE,$FC,$11,$F1,$9F,$DE,$7E,$27,$F7,$75
	.dc.b	$AF,$C8,$7F,$CF,$53,$FD,$F7,$F4,$CC,$FC,$CF,$F5,$6E,$1F,$F0,$1E
	.dc.b	$7D,$63,$FC,$67,$1F,$D6,$70,$E5,$97,$6F,$D3,$BF,$73,$18,$C5,$F7
	.dc.b	$9F,$9E,$F4,$FF,$B8,$FE,$35,$FE,$EA,$FC,$25,$F1,$7F,$9B,$D3,$FD
	.dc.b	$B9,$FA,$AF,$E7,$6F,$7F,$05,$FE,$96,$AD,$7E,$3D,$FA,$7F,$37,$F7
	.dc.b	$9C,$0C,$FA,$3F,$99,$C3,$8F,$6F,$E7,$DE,$7A,$78,$2F,$FB,$B7,$B1
	.dc.b	$E5,$BF,$F4,$AE,$1E,$5C,$BC,$9F,$21,$F4,$67,$FA,$2E,$5F,$E5,$4F
	.dc.b	$9D,$0F,$A0,$FF,$23,$FE,$5D,$F8,$2E,$EF,$C1,$76,$6A,$B5,$F5,$FF
	.dc.b	$03,$AB,$F0,$19,$7F,$B2,$F2,$F9,$AF,$F8,$C7,$0F,$C5,$7D,$1F,$E2
	.dc.b	$63,$C3,$8F,$3C,$DF,$75,$E5,$9E,$53,$9F,$BD,$C7,$E5,$2F,$CE,$BF
	.dc.b	$DF,$9F,$02,$FA,$C7,$E0,$CF,$DB,$1F,$92,$7E,$6B,$0E,$1F,$8D,$FF
	.dc.b	$18,$BD,$D3,$F2,$4D,$64,$AF,$36,$AC,$3C,$1F,$11,$87,$9E,$9E,$1D
	.dc.b	$3F,$D8,$95,$BC,$3C,$B9,$F9,$3F,$12,$FE,$4B,$F3,$3F,$E3,$8F,$CE
	.dc.b	$BF,$CF,$3F,$76,$9F,$F6,$CF,$EE,$EF,$98,$FF,$25,$F8,$5F,$DC,$EF
	.dc.b	$75,$A9,$7E,$DD,$FC,$A7,$1E,$9E,$A7,$F9,$BB,$5F,$E4,$11,$8F,$BF
	.dc.b	$FE,$4C,$EB,$4F,$07,$8C,$D5,$F2,$BE,$6C,$3F,$DB,$BA,$F8,$2B,$5F
	.dc.b	$0B,$FC,$FB,$BF,$FC,$A3,$E6,$3F,$37,$7E,$CF,$FE,$E4,$FC,$71,$F9
	.dc.b	$E7,$F6,$CF,$DD,$BF,$9F,$61,$C2,$F7,$A2,$F7,$F4,$1F,$92,$E7,$FC
	.dc.b	$5E,$7D,$73,$FE,$7D,$8F,$1E,$1B,$8F,$FB,$CA,$B4,$F3,$D3,$F8,$3B
	.dc.b	$E0,$35,$9F,$83,$3F,$2F,$E5,$BE,$B9,$FF,$7D,$FD,$7B,$F3,$4F,$F0
	.dc.b	$37,$0B,$D9,$7E,$A5,$FE,$7E,$FD,$EB,$A6,$1F,$E4,$4B,$DD,$6B,$E9
	.dc.b	$3F,$D3,$1C,$5F,$A4,$CB,$A7,$CA,$FA,$30,$E1,$F4,$B6,$7C,$0B,$4F
	.dc.b	$06,$9F,$1B,$0E,$59,$F6,$8F,$E2,$78,$7F,$87,$3F,$76,$71,$3F,$67
	.dc.b	$7D,$83,$BD,$F9,$FF,$E7,$7E,$FB,$85,$EE,$B5,$3F,$A1,$F3,$B3,$87
	.dc.b	$F8,$83,$C2,$FF,$47,$E5,$DB,$F6,$7C,$68,$11,$E2,$AF,$0E,$9E,$0F
	.dc.b	$92,$FF,$11,$CE,$73,$EF,$F7,$9F,$BF,$F8,$17,$ED,$2F,$D2,$9F,$A6
	.dc.b	$3F,$64,$6D,$FC,$57,$E6,$FF,$AA,$E1,$C2,$F7,$5A,$F7,$1F,$77,$AC
	.dc.b	$BA,$70,$FF,$56,$F1,$BF,$D5,$38,$F8,$3F,$55,$FD,$FB,$CF,$5F,$06
	.dc.b	$BF,$BC,$8C,$67,$E2,$FA,$3F,$C4,$63,$FE,$37,$FD,$C3,$FD,$DB,$F9
	.dc.b	$63,$F0,$27,$EB,$EF,$DA,$5F,$99,$7E,$23,$EC,$98,$70,$BD,$D3,$F0
	.dc.b	$1D,$4F,$FB,$E3,$2F,$F0,$D7,$0F,$FB,$03,$0E,$3F,$A4,$FE,$2D,$E7
	.dc.b	$F1,$8D,$7C,$5A,$FF,$81,$FC,$6F,$E6,$B5,$A9,$CF,$B5,$FC,$BF,$E4
	.dc.b	$BF,$C3,$BF,$08,$FC,$1D,$F4,$4F,$DE,$9F,$A7,$BF,$2C,$FF,$02,$E1
	.dc.b	$C2,$F7,$5A,$FC,$5B,$3F,$57,$FE,$50,$E3,$FA,$07,$E1,$EF,$BB,$80
	.dc.b	$F3,$CB,$C1,$F1,$FE,$5A,$FF,$79,$63,$DB,$E7,$BF,$A3,$7E,$C5,$FC
	.dc.b	$63,$F2,$37,$B8,$FD,$4D,$C5,$FE,$3D,$F9,$8F,$F0,$DD,$EE,$B5,$3F
	.dc.b	$E4,$1E,$DC,$FC,$43,$1F,$E8,$37,$FC,$93,$CD,$B3,$9C,$38,$F8,$5E
	.dc.b	$28,$C6,$7E,$0F,$C9,$7F,$0B,$DE,$FD,$27,$F9,$37,$E9,$BF,$EB,$2F
	.dc.b	$A8,$7E,$B5,$FE,$38,$F0,$3F,$B4,$DE,$FE,$19,$7B,$AF,$F4,$15,$6B
	.dc.b	$0F,$E6,$1D,$5F,$84,$3F,$D7,$EF,$F2,$07,$F6,$87,$89,$F0,$FE,$16
	.dc.b	$C9,$FA,$BE,$72,$EF,$39,$FC,$C3,$19,$FD,$FF,$9C,$7E,$71,$E8,$D3
	.dc.b	$C5,$87,$9B,$C2,$FE,$DF,$3D,$27,$37,$C7,$2E,$75,$FD,$D3,$E2,$FA
	.dc.b	$FE,$84,$FD,$63,$DF,$F9,$6F,$F2,$CE,$3F,$85,$8A,$B4,$FF,$A6,$B2
	.dc.b	$EB,$1C,$70,$E7,$87,$B3,$8B,$F9,$AF,$D9,$7F,$0C,$FC,$EF,$B7,$C0
	.dc.b	$3F,$29,$7D,$D7,$F8,$BF,$3F,$F0,$CD,$EF,$15,$EE,$9F,$9B,$74,$F7
	.dc.b	$59,$7F,$3B,$C3,$F8,$CF,$EE,$76,$73,$8F,$2C,$FF,$27,$A7,$1C,$3C
	.dc.b	$71,$FB,$5E,$0F,$AF,$8F,$9F,$F7,$E7,$E4,$EF,$D4,$9E,$6F,$EF,$78
	.dc.b	$7C,$FA,$F8,$AF,$75,$AF,$6F,$B4,$62,$FF,$8B,$E3,$D7,$1E,$9E,$66
	.dc.b	$31,$E5,$FB,$BB,$E2,$FF,$CB,$7F,$33,$CA,$31,$F6,$4F,$C8,$B1,$4F
	.dc.b	$C0,$39,$71,$B6,$61,$CF,$97,$F9,$3F,$E9,$9F,$ED,$DF,$D1,$3F,$D1
	.dc.b	$9E,$53,$DC,$7D,$AB,$F5,$0F,$ED,$B7,$BF,$B5,$5E,$EB,$FA,$F7,$4F
	.dc.b	$F9,$42,$BF,$DD,$1F,$A3,$FF,$6D,$7B,$76,$FE,$94,$3C,$EF,$C1,$C1
	.dc.b	$E4,$F8,$26,$BE,$4D,$7F,$29,$87,$8A,$73,$F3,$8F,$C6,$3F,$91,$7E
	.dc.b	$32,$FD,$CB,$58,$FF,$79,$9C,$FE,$B7,$8B,$E1,$E6,$3F,$02,$F0,$5E
	.dc.b	$E3,$F8,$CE,$7C,$BF,$44,$61,$F8,$0F,$C4,$9F,$BE,$66,$31,$F7,$37
	.dc.b	$C5,$FE,$9F,$85,$FC,$6B,$FA,$AF,$CF,$3F,$67,$FF,$3D,$FE,$63,$FE
	.dc.b	$F8,$FC,$E7,$FA,$8F,$5F,$F9,$E2,$B5,$F2,$7F,$4A,$B5,$F4,$DF,$E1
	.dc.b	$E9,$CE,$5F,$2D,$E9,$F2,$BF,$E1,$B7,$8F,$7A,$DF,$7A,$FD,$23,$C7
	.dc.b	$AF,$8A,$3F,$C8,$18,$F5,$86,$7F,$D8,$7D,$6F,$F1,$7F,$B9,$FE,$DA
	.dc.b	$DB,$CD,$5A,$F9,$9F,$DC,$FD,$57,$F1,$AF,$CF,$F9,$63,$C6,$3F,$81
	.dc.b	$5F,$08,$C7,$B9,$E2,$7F,$B9,$FD,$53,$ED,$5F,$8E,$3F,$23,$E7,$F4
	.dc.b	$1F,$E2,$89,$CF,$C9,$62,$F8,$7D,$F7,$3A,$D7,$C5,$7F,$9A,$78,$DF
	.dc.b	$A7,$3F,$61,$FE,$D1,$F2,$0F,$D7,$DF,$AC,$FF,$BB,$7E,$AF,$FC,$F5
	.dc.b	$87,$F8,$AF,$0E,$1E,$4F,$F4,$6B,$DF,$AC,$FA,$ED,$DF,$6C,$FF,$BA
	.dc.b	$3F,$E8,$03,$92,$FA,$2F,$F4,$0C,$63,$0E,$99,$FF,$14,$E7,$FE,$01
	.dc.b	$39,$D7,$CD,$2F,$CB,$BD,$9A,$F9,$3F,$30,$7E,$28,$FC,$49,$FA,$9B
	.dc.b	$F4,$8F,$2D,$EE,$B5,$2F,$D0,$B3,$C1,$FE,$0B,$C7,$9E,$2F,$B6,$EF
	.dc.b	$39,$C3,$B7,$E9,$58,$E1,$7E,$20,$FB,$FF,$F0,$B7,$EE,$8F,$DF,$3A
	.dc.b	$D6,$A7,$3F,$99,$62,$F8,$7D,$8F,$F3,$EE,$6C,$57,$F8,$7F,$F0,$8F
	.dc.b	$F7,$54,$3F,$5E,$FF,$53,$7D,$E3,$F5,$0F,$F1,$E7,$CD,$3B,$FF,$C4
	.dc.b	$DD,$BF,$EF,$AB,$DD,$6A,$5F,$93,$AB,$5F,$87,$FB,$F2,$FF,$7F,$1F
	.dc.b	$E6,$BC,$3B,$B8,$3C,$50,$CB,$C5,$F6,$3F,$D1,$7E,$CB,$CF,$7E,$08
	.dc.b	$FE,$91,$DF,$8B,$E1,$E5,$3F,$34,$7E,$8E,$FD,$81,$F1,$BE,$5F,$C0
	.dc.b	$3F,$C1,$39,$FC,$0C,$38,$5E,$FF,$42,$D6,$FE,$7C,$FA,$E7,$D3,$E1
	.dc.b	$3C,$1C,$0C,$78,$FD,$FE,$3A,$FF,$45,$F7,$9F,$CA,$FF,$09,$E7,$9F
	.dc.b	$BB,$05,$6B,$E4,$5F,$A5,$6B,$83,$E9,$CF,$2F,$9B,$E9,$E9,$7F,$96
	.dc.b	$3E,$5F,$FD,$29,$F7,$9F,$E6,$EC,$FF,$D7,$B8,$7F,$93,$BE,$65,$F9
	.dc.b	$E7,$F9,$DF,$C5,$EB,$7B,$BF,$C6,$5E,$EB,$5B,$AB,$5E,$27,$E1,$7F
	.dc.b	$00,$FF,$17,$E1,$FE,$30,$9C,$E0,$FA,$9C,$CF,$93,$F0,$3F,$F1,$AE
	.dc.b	$1C,$74,$F1,$FD,$5B,$DC,$D5,$FA,$33,$FE,$C3,$AF,$FA,$37,$E2,$73
	.dc.b	$3E,$69,$FF,$75,$7E,$1A,$F3,$E7,$47,$9F,$5E,$FB,$FF,$9B,$BE,$19
	.dc.b	$7B,$BF,$F5,$87,$0F,$A3,$1E,$71,$FA,$58,$C5,$70,$8F,$F7,$CC,$6F
	.dc.b	$EA,$FF,$D0,$9F,$4F,$FE,$18,$FC,$7F,$C1,$AF,$A6,$73,$F5,$AF,$C5
	.dc.b	$33,$3E,$99,$E3,$9F,$B3,$97,$E6,$47,$4F,$0B,$ED,$E5,$FE,$5C,$FC
	.dc.b	$93,$F9,$C7,$8C,$A7,$EA,$BF,$EA,$EF,$C7,$BF,$21,$BF,$F1,$CD,$18
	.dc.b	$70,$BD,$FC,$42,$FF,$A0,$5E,$EB,$AD,$FD,$88,$C7,$C5,$FE,$9A,$31
	.dc.b	$9F,$8E,$19,$F8,$F3,$FF,$5F,$FD,$B7,$A2,$FF,$DC,$D8,$F8,$67,$FD
	.dc.b	$87,$EF,$E6,$7C,$B8,$72,$FD,$AF,$F8,$37,$B4,$FC,$DF,$FC,$33,$AF
	.dc.b	$8B,$45,$6A,$7F,$DC,$D3,$98,$E3,$F8,$27,$5D,$3E,$3C,$74,$FC,$CB
	.dc.b	$19,$7A,$3F,$76,$7E,$88,$FC,$B5,$FB,$13,$F4,$CF,$D0,$72,$7B,$F7
	.dc.b	$61,$C2,$F7,$FA,$E6,$B8,$BD,$F9,$F5,$CF,$A7,$B9,$E0,$C7,$8C,$77
	.dc.b	$F7,$BE,$17,$DE,$BE,$F5,$FB,$1F,$FB,$4B,$E8,$1F,$A9,$3F,$5B,$7E
	.dc.b	$F5,$DF,$66,$3C,$70,$E1,$7B,$FC,$B7,$53,$9A,$FB,$20,$E3,$93,$F5
	.dc.b	$4F,$DB,$C7,$08,$F1,$FB,$DF,$E2,$BC,$3F,$9C,$76,$7E,$37,$0F,$04
	.dc.b	$E7,$C1,$F5,$DF,$AB,$F4,$AF,$EC,$1F,$63,$FD,$9B,$F3,$1E,$9B,$27
	.dc.b	$3F,$32,$FD,$73,$31,$C3,$C5,$E7,$5B,$F9,$A6,$2F,$96,$CF,$C1,$3E
	.dc.b	$8D,$FB,$76,$7F,$73,$EF,$39,$8E,$1E,$5E,$2B,$77,$C9,$E3,$DF,$FE
	.dc.b	$1B,$FC,$F9,$FA,$A3,$E1,$1F,$96,$3C,$EF,$EB,$BF,$63,$7E,$5B,$F6
	.dc.b	$18,$70,$BD,$D6,$BF,$4D,$FF,$0C,$CE,$7B,$FF,$C3,$01,$EF,$F1,$BF
	.dc.b	$E1,$88,$C6,$7E,$1C,$FC,$1F,$31,$FE,$4D,$9E,$73,$9C,$7F,$B7,$F8
	.dc.b	$7E,$98,$F4,$7C,$3B,$F5,$1F,$F1,$8F,$23,$EC,$78,$EF,$FC,$C6,$73
	.dc.b	$ED,$62,$F8,$7D,$3B,$D6,$E9,$7B,$8F,$F5,$16,$BC,$BF,$08,$FE,$81
	.dc.b	$FA,$93,$EC,$BF,$D7,$5C,$DF,$13,$7F,$95,$5A,$9F,$F4,$8C,$E7,$1F
	.dc.b	$F5,$4F,$C8,$30,$7C,$0F,$83,$0E,$BF,$2D,$F9,$9C,$3F,$43,$9F,$E6
	.dc.b	$FF,$7B,$FE,$49,$EF,$BF,$45,$7E,$2B,$FA,$9F,$CC,$1F,$B9,$7F,$84
	.dc.b	$FD,$2F,$D1,$62,$C7,$8E,$1C,$3D,$76,$0F,$DC,$B8,$5F,$7F,$88,$FE
	.dc.b	$5A,$FF,$AB,$F0,$FE,$69,$F5,$9F,$AA,$DB,$CB,$3D,$6B,$F9,$9E,$5D
	.dc.b	$7C,$9F,$56,$9E,$2C,$7C,$3B,$7E,$FF,$7C,$E3,$DD,$7E,$92,$FD,$D9
	.dc.b	$FB,$83,$CE,$FF,$62,$E5,$CF,$5F,$E8,$37,$BA,$D7,$E5,$BE,$AC,$BC
	.dc.b	$99,$74,$CB,$9E,$BE,$09,$CE,$1C,$6B,$1A,$FC,$E8,$7E,$08,$FD,$C7
	.dc.b	$C5,$F7,$9D,$6B,$53,$9D,$DF,$9D,$67,$7F,$2C,$F1,$D3,$AC,$63,$1F
	.dc.b	$9B,$C7,$F2,$BD,$4F,$ED,$5F,$A7,$BE,$B9,$F9,$B3,$EA,$FF,$A2,$FC
	.dc.b	$FF,$B6,$F7,$5A,$9F,$E9,$3F,$01,$9F,$99,$7E,$59,$F5,$FA,$BF,$07
	.dc.b	$87,$6E,$7F,$F7,$53,$CB,$3E,$DB,$FC,$58,$C6,$5F,$ED,$AD,$BC,$D9
	.dc.b	$36,$F3,$78,$5F,$DD,$BC,$97,$EB,$3E,$19,$CF,$CC,$7F,$0E,$FC,$FF
	.dc.b	$1C,$4C,$3F,$BE,$F9,$FD,$F2,$E9,$39,$B7,$E0,$59,$C7,$FC,$6D,$D7
	.dc.b	$FE,$EA,$B7,$EC,$9D,$FE,$2F,$F3,$DF,$87,$FF,$84,$FF,$30,$C5,$FF
	.dc.b	$BC,$3E,$8B,$95,$6B,$9F,$F0,$5F,$A8,$BF,$06,$FF,$80,$1F,$87,$7F
	.dc.b	$AF,$7E,$E1,$FB,$6B,$C2,$FC,$65,$EF,$F1,$95,$EE,$9F,$9F,$EB,$1F
	.dc.b	$F7,$FE,$5F,$E4,$6E,$BF,$D3,$E3,$C6,$31,$E0,$7E,$DA,$3A,$63,$E1
	.dc.b	$F8,$9F,$2E,$1E,$4F,$47,$FC,$F7,$F3,$58,$D5,$FB,$3B,$F0,$0F,$EC
	.dc.b	$FF,$87,$7F,$A0,$36,$EF,$8A,$F7,$5A,$F9,$FF,$34,$63,$67,$E2,$BD
	.dc.b	$72,$E9,$0F,$88,$C7,$CE,$BF,$1A,$7D,$CF,$FB,$EE,$BD,$5A,$77,$C3
	.dc.b	$FB,$85,$6B,$F0,$7E,$F1,$8B,$E7,$3F,$CC,$79,$7F,$71,$E7,$FF,$00
	.dc.b	$FC,$CB,$FD,$49,$F2,$EF,$D5,$3F,$B1,$7E,$E3,$AF,$79,$5A,$9F,$EC
	.dc.b	$DE,$16,$7F,$5B,$FD,$06,$5F,$E0,$3C,$BA,$3F,$40,$C6,$1C,$BE,$E5
	.dc.b	$9F,$7F,$37,$F8,$35,$F8,$E7,$39,$7E,$53,$0F,$16,$7F,$EB,$9F,$E0
	.dc.b	$1F,$F5,$7F,$E1,$4F,$DD,$73,$82,$73,$F3,$38,$7F,$D6,$BF,$A3,$76
	.dc.b	$DB,$C5,$1F,$EB,$8C,$70,$7F,$B1,$7F,$3C,$7D,$D7,$FA,$7B,$EC,$7A
	.dc.b	$C6,$3F,$00,$7E,$CB,$8A,$FE,$1B,$D9,$FD,$0B,$8B,$D5,$7C,$BA,$3F
	.dc.b	$91,$7E,$AE,$FC,$31,$FB,$5B,$F7,$C7,$EB,$3F,$DA,$FF,$22,$FF,$0A
	.dc.b	$F9,$3F,$D0,$30,$E1,$7B,$AD,$7B,$BF,$CC,$67,$A7,$E1,$9F,$4B,$AD
	.dc.b	$1C,$1F,$A4,$63,$3F,$0E,$7E,$0F,$A2,$FC,$9C,$F3,$9C,$E3,$FD,$F7
	.dc.b	$C8,$F4,$C7,$0F,$BE,$FF,$6A,$78,$6F,$8F,$E1,$B7,$AE,$73,$F8,$76
	.dc.b	$34,$FF,$31,$BF,$BD,$CF,$13,$FE,$DD,$F9,$2E,$B7,$F0,$FD,$87,$E9
	.dc.b	$8F,$C9,$DF,$AB,$73,$9C,$C3,$F2,$8C,$53,$D7,$E7,$AF,$F0,$C8,$7B
	.dc.b	$3F,$D4,$FE,$D9,$E9,$7E,$1F,$F3,$4F,$D0,$7B,$CF,$C2,$FF,$03,$DB
	.dc.b	$FD,$C2,$F7,$E4,$2F,$74,$FD,$77,$FB,$15,$74,$F9,$DF,$A4,$F0,$E6
	.dc.b	$FA,$2F,$2B,$E3,$1C,$33,$F1,$7A,$3F,$C7,$3E,$CF,$CD,$1E,$5F,$69
	.dc.b	$FB,$E7,$F9,$BF,$E6,$F8,$6D,$E3,$9C,$E2,$FC,$73,$39,$FF,$9C,$F0
	.dc.b	$FF,$0A,$F0,$7A,$F2,$E9,$0F,$89,$EF,$97,$B7,$F2,$7F,$DF,$F2,$7C
	.dc.b	$8C,$38,$4E,$7D,$2F,$97,$F9,$FF,$6B,$DE,$B6,$7E,$93,$FE,$05,$F8
	.dc.b	$C7,$E7,$CF,$C8,$9F,$88,$7F,$88,$FF,$47,$68,$FD,$D3,$85,$EE,$B5
	.dc.b	$3C,$4F,$F1,$1E,$1D,$3D,$6F,$D0,$3D,$B1,$B7,$C7,$EA,$59,$3E,$07
	.dc.b	$FC,$29,$7B,$CB,$C5,$F2,$F8,$C3,$C1,$F8,$BB,$F6,$FF,$F4,$F7,$EE
	.dc.b	$DE,$4D,$3D,$33,$9E,$07,$EE,$39,$C9,$39,$9C,$3C,$F9,$FB,$5F,$C2
	.dc.b	$FD,$B7,$FB,$B7,$C1,$FF,$B1,$F8,$FD,$BA,$FC,$19,$CF,$E4,$FD,$5D
	.dc.b	$E5,$7E,$7F,$F9,$AB,$FE,$33,$FC,$BD,$F9,$EB,$F8,$03,$E7,$FD,$E7
	.dc.b	$E5,$FE,$3D,$F3,$1B,$AD,$71,$2B,$5F,$07,$9F,$DF,$FF,$86,$65,$D3
	.dc.b	$2B,$87,$0B,$FD,$17,$7F,$F8,$FE,$B5,$8F,$3F,$3B,$F8,$36,$BD,$F1
	.dc.b	$7E,$16,$FD,$21,$F2,$5F,$C9,$7F,$59,$D4,$E6,$3F,$94,$7C,$CE,$29
	.dc.b	$F8,$AF,$E7,$3E,$3B,$FC,$E3,$AF,$97,$0E,$1F,$1D,$E3,$E1,$F8,$50
	.dc.b	$FD,$ED,$F9,$0B,$F5,$27,$D6,$78,$27,$31,$FE,$2E,$EE,$FF,$38,$E5
	.dc.b	$D6,$78,$FE,$4D,$D7,$8B,$D7,$1D,$3F,$0A,$FF,$30,$7E,$A6,$EF,$7E
	.dc.b	$CB,$FD,$B7,$7F,$A3,$EC,$FF,$96,$B4,$FF,$3E,$DE,$EB,$5D,$FF,$EE
	.dc.b	$BF,$67,$FC,$DD,$3C,$FE,$BF,$F9,$F3,$BE,$BF,$E7,$AF,$1F,$77,$F0
	.dc.b	$37,$FE,$4B,$E7,$FF,$4A,$CE,$74,$FC,$0F,$CB,$7F,$BE,$7E,$B1,$F9
	.dc.b	$07,$FA,$6B,$E7,$7F,$CC,$7E,$43,$F7,$C7,$E5,$2F,$D1,$DF,$85,$7F
	.dc.b	$A2,$31,$E3,$C0,$3F,$D1,$7F,$77,$FE,$51,$C3,$8D,$F0,$FC,$59,$7C
	.dc.b	$2F,$79,$7F,$AA,$9F,$F5,$5C,$67,$3F,$F3,$AF,$D8,$E6,$7D,$1A,$F9
	.dc.b	$F3,$ED,$AF,$97,$F2,$8C,$D7,$8F,$0F,$5E,$BE,$2F,$13,$FD,$29,$F3
	.dc.b	$6F,$D4,$DF,$6A,$FA,$4F,$F0,$66,$1F,$D0,$FF,$16,$FF,$4F,$5E,$EB
	.dc.b	$5F,$57,$FE,$A8,$F9,$2F,$EF,$5E,$07,$FA,$DA,$B5,$AF,$F7,$07,$B7
	.dc.b	$EF,$1F,$DD,$5E,$59,$F6,$BF,$1E,$9E,$28,$F0,$FC,$8F,$FA,$A7,$CE
	.dc.b	$7E,$D4,$F6,$5F,$9D,$BF,$5F,$7D,$67,$FB,$B7,$2F,$F5,$A6,$1C,$2F
	.dc.b	$75,$AF,$C5,$5F,$AA,$EB,$2F,$E0,$7C,$1F,$E0,$6C,$38,$36,$3C,$19
	.dc.b	$F7,$CF,$FC,$95,$C3,$FE,$D4,$BE,$D3,$9F,$37,$E8,$E0,$FC,$AF,$F5
	.dc.b	$A7,$F9,$77,$F0,$4F,$E2,$BF,$8A,$3F,$79,$FC,$1F,$81,$FE,$EA,$AD
	.dc.b	$FD,$67,$FA,$CF,$C9,$FF,$7C,$FC,$F6,$B2,$FC,$75,$74,$7F,$09,$E2
	.dc.b	$FE,$2E,$3F,$77,$F0,$DF,$87,$9E,$07,$8F,$DA,$F2,$63,$D7,$4F,$3D
	.dc.b	$F8,$35,$FE,$53,$E2,$7F,$B6,$7F,$0A,$FD,$2F,$AC,$FD,$75,$FA,$9F
	.dc.b	$D2,$FB,$FF,$F1,$7F,$3F,$FB,$DA,$F7,$5A,$E6,$FC,$6E,$FF,$EE,$79
	.dc.b	$FD,$7F,$DB,$7F,$8D,$DE,$B1,$FE,$36,$78,$E3,$F7,$BA,$FC,$5C,$DF
	.dc.b	$C2,$F1,$E9,$E9,$7E,$E7,$F6,$9F,$D8,$62,$E5,$FF,$6E,$7D,$BB,$F4
	.dc.b	$77,$ED,$1F,$E4,$4F,$D1,$1F,$75,$FB,$2C,$78,$E1,$C2,$F7,$F7,$4B
	.dc.b	$FF,$2B,$F6,$2F,$79,$7E,$EB,$3F,$A3,$18,$8F,$A1,$3B,$9F,$E0,$98
	.dc.b	$F5,$CF,$FC,$AB,$F3,$7E,$99,$CD,$FF,$BC,$F3,$F0,$FC,$EE,$67,$D9
	.dc.b	$39,$F6,$3F,$89,$7E,$B2,$FD,$33,$D5,$5C,$AB,$E0,$56,$BE,$BF,$33
	.dc.b	$CA,$73,$EE,$FF,$A3,$76,$F3,$4E,$73,$FF,$74,$4F,$F8,$1E,$38,$ED
	.dc.b	$F1,$7E,$E5,$F9,$8B,$F6,$A7,$7E,$B3,$F5,$56,$A7,$F8,$6E,$7D,$B0
	.dc.b	$E5,$87,$87,$8F,$E4,$69,$FE,$6B,$8F,$07,$7F,$FA,$1F,$4F,$E5,$9F
	.dc.b	$AF,$3F,$79,$7E,$EE,$F8,$3F,$DD,$F4,$BF,$36,$FF,$2B,$63,$C7,$16
	.dc.b	$3C,$70,$E1,$E0,$7D,$DB,$FF,$03,$DE,$EB,$B7,$E9,$DF,$A9,$8C,$67
	.dc.b	$E0,$C7,$FD,$DD,$ED,$FE,$42,$FA,$69,$E2,$FC,$63,$FA,$16,$1C,$34
	.dc.b	$EF,$AF,$F1,$0C,$7F,$C7,$F8,$F1,$CB,$D7,$8F,$FB,$7F,$F1,$E7,$EC
	.dc.b	$CF,$D2,$B7,$87,$C2,$BD,$F8,$8B,$DF,$C0,$6A,$BD,$F5,$A9,$F9,$79
	.dc.b	$77,$D3,$E4,$61,$C7,$EA,$3C,$F3,$9D,$1F,$81,$3F,$56,$7E,$89,$FE
	.dc.b	$88,$E1,$F6,$56,$A7,$3F,$3F,$E1,$E4,$F8,$78,$71,$FC,$37,$AF,$07
	.dc.b	$A7,$6F,$B1,$3D,$7F,$0F,$F8,$FE,$77,$EE,$BF,$69,$F0,$FC,$C7,$FC
	.dc.b	$01,$E5,$FF,$04,$FE,$AC,$F9,$5F,$F3,$3D,$EF,$1F,$F4,$07,$2A,$F7
	.dc.b	$F8,$86,$B5,$FE,$E9,$9F,$B3,$F0,$0F,$B9,$8F,$1D,$2D,$6B,$2E,$78
	.dc.b	$FF,$86,$23,$CA,$66,$7C,$9F,$39,$FE,$89,$8E,$C3,$1F,$4E,$61,$FB
	.dc.b	$43,$F0,$3F,$E5,$5C,$FF,$E0,$1A,$DD,$6A,$7F,$AD,$CE,$73,$F6,$CF
	.dc.b	$2F,$C6,$7C,$78,$F7,$9C,$E2,$FC,$EF,$39,$78,$BD,$9F,$89,$F9,$93
	.dc.b	$F7,$0F,$3C,$E7,$81,$3F,$E0,$1D,$7C,$1C,$1E,$9C,$3E,$67,$C7,$74
	.dc.b	$D3,$E7,$C7,$CF,$E7,$FC,$FF,$CF,$FF,$32,$FD,$11,$F8,$6B,$F7,$17
	.dc.b	$EF,$AF,$A7,$F1,$7B,$2F,$D5,$FF,$D4,$38,$F1,$C3,$85,$EF,$8F,$FC
	.dc.b	$61,$DF,$AE,$7F,$F1,$76,$9D,$B8,$FF,$E0,$03,$E0,$D6,$E3,$CB,$5F
	.dc.b	$81,$B7,$9B,$0E,$33,$9C,$9E,$0E,$65,$8F,$3D,$3F,$C4,$BA,$3F,$74
	.dc.b	$FF,$80,$1C,$AD,$FD,$1B,$7B,$BE,$79,$7B,$F0,$3D,$19,$F4,$DB,$DB
	.dc.b	$5A,$E5,$F0,$6B,$DE,$38,$F6,$62,$BE,$1F,$4F,$F7,$0F,$84,$7E,$63
	.dc.b	$D3,$0F,$34,$E7,$C4,$C5,$F0,$AF,$F4,$36,$7F,$03,$6F,$A7,$7F,$DE
	.dc.b	$BE,$8B,$E0,$7E,$D5,$F9,$5D,$7F,$D5,$18,$3F,$75,$FF,$4A,$FE,$23
	.dc.b	$EA,$AD,$7D,$7F,$8A,$7F,$D3,$3F,$8C,$E7,$C7,$FE,$FF,$AF,$F9,$6F
	.dc.b	$9F,$E3,$3F,$E6,$5C,$3F,$45,$AF,$F2,$DE,$7F,$EF,$31,$D3,$4F,$07
	.dc.b	$E0,$1E,$4C,$7F,$CA,$BA,$77,$FB,$CC,$6F,$FE,$BD,$E3,$7E,$A5,$FE
	.dc.b	$EE,$FC,$8B,$C5,$39,$8F,$F7,$8F,$E9,$F8,$C3,$F8,$54,$3D,$DE,$71
	.dc.b	$8F,$4E,$FA,$2F,$DC,$DF,$B6,$BF,$77,$E5,$39,$8F,$EA,$5C,$DE,$EE
	.dc.b	$67,$27,$C6,$AE,$5C,$3D,$31,$71,$FE,$3F,$F5,$17,$E1,$3F,$EE,$FE
	.dc.b	$A7,$5F,$FB,$B2,$5F,$BD,$7E,$29,$DF,$E1,$BF,$56,$75,$FB,$F6,$70
	.dc.b	$FF,$78,$6D,$F8,$01,$7D,$BF,$4D,$FA,$11,$FE,$97,$8E,$79,$78,$3F
	.dc.b	$0E,$7D,$27,$97,$0F,$26,$5F,$EF,$EF,$A2,$C6,$9E,$1F,$C8,$5F,$B9
	.dc.b	$FF,$A0,$BC,$2F,$F1,$9E,$9D,$30,$56,$BE,$DF,$CD,$87,$1C,$F9,$EB
	.dc.b	$F0,$29,$F3,$58,$FC,$F5,$F7,$0F,$C4,$FF,$95,$75,$FC,$FD,$7D,$35
	.dc.b	$A9,$79,$FD,$B2,$4F,$2C,$7A,$63,$FE,$47,$E4,$FF,$4E,$FB,$4F,$D0
	.dc.b	$DF,$62,$FD,$53,$F0,$59,$BF,$61,$DD,$6A,$73,$F9,$4E,$31,$FF,$50
	.dc.b	$62,$FC,$AB,$FD,$40,$F5,$CB,$A7,$81,$E6,$AE,$71,$8F,$86,$FF,$5A
	.dc.b	$D6,$F1,$FF,$38,$43,$F3,$CD,$C6,$3E,$B2,$FC,$2F,$F4,$1C,$6F,$E1
	.dc.b	$DC,$B6,$F3,$7E,$78,$FC,$51,$FA,$47,$28,$C7,$55,$BE,$1B,$FC,$19
	.dc.b	$FF,$17,$63,$FC,$6B,$DE,$7E,$07,$CF,$EE,$BF,$DC,$BF,$B1,$3E,$F5
	.dc.b	$F9,$6E,$EB,$5F,$84,$2B,$5F,$07,$9F,$D3,$FF,$27,$9B,$3E,$BC,$38
	.dc.b	$6E,$9D,$BD,$1F,$E6,$D1,$8C,$BA,$7A,$58,$D3,$BF,$C3,$3F,$51,$7E
	.dc.b	$67,$E3,$CD,$9F,$FB,$F2,$B5,$F1,$79,$8C,$7C,$6F,$3A,$DE,$7F,$E2
	.dc.b	$4F,$4F,$14,$87,$EF,$DF,$E5,$4F,$57,$94,$63,$E6,$DF,$B0,$E2,$9F
	.dc.b	$85,$72,$AE,$3C,$1F,$D4,$7F,$74,$7D,$8B,$F0,$37,$E8,$EF,$D9,$DF
	.dc.b	$A2,$BF,$7B,$7E,$59,$FD,$3B,$1E,$38,$70,$BD,$D3,$DB,$D6,$5D,$7B
	.dc.b	$7F,$D5,$27,$85,$7E,$51,$F5,$31,$E7,$A7,$87,$4F,$E0,$B7,$CB,$3E
	.dc.b	$D1,$FB,$88,$C4,$FF,$6C,$FA,$3F,$C0,$7D,$4E,$3A,$7F,$C4,$7E,$1F
	.dc.b	$F2,$BF,$27,$F9,$46,$7D,$A7,$FC,$4F,$E8,$78,$F4,$F8,$31,$FE,$2A
	.dc.b	$F2,$B1,$9F,$F7,$8F,$9F,$7E,$5C,$F9,$9E,$D1,$8F,$CC,$EF,$47,$0C
	.dc.b	$7F,$C2,$5F,$5D,$FD,$FF,$F2,$C7,$E2,$9F,$C0,$3F,$65,$FB,$7E,$EB
	.dc.b	$F2,$7F,$A7,$C7,$FC,$F6,$F7,$EA,$2F,$74,$FC,$07,$FB,$FD,$6B,$A7
	.dc.b	$E6,$98,$C9,$FB,$A7,$E8,$27,$9E,$1E,$0F,$8E,$F2,$57,$8F,$C5,$FE
	.dc.b	$21,$F1,$58,$D3,$C1,$FA,$DB,$F0,$E7,$EC,$EF,$C2,$BB,$56,$A7,$3C
	.dc.b	$38,$D1,$87,$1F,$AC,$E9,$18,$AF,$F1,$1F,$06,$EB,$FC,$37,$6F,$D5
	.dc.b	$31,$F5,$68,$FE,$B6,$7C,$5D,$FE,$55,$7F,$88,$B4,$EA,$FF,$89,$3F
	.dc.b	$0C,$FF,$2C,$F8,$3F,$D9,$BF,$60,$FF,$34,$7E,$BC,$FD,$05,$F9,$07
	.dc.b	$FB,$E3,$37,$DD,$7F,$BF,$B0,$E1,$F7,$0C,$38,$5B,$A3,$FC,$E7,$5F
	.dc.b	$DC,$BF,$7F,$FF,$3E,$F1,$7F,$99,$7E,$89,$AE,$2B,$9E,$3E,$1F,$A7
	.dc.b	$F2,$D6,$B3,$ED,$B7,$FA,$E7,$6F,$1E,$43,$A6,$5F,$E8,$1C,$1F,$AE
	.dc.b	$BF,$79,$F3,$5F,$0B,$DE,$3F,$E1,$1A,$D7,$E1,$1E,$8C,$FC,$18,$F2
	.dc.b	$F8,$5F,$1E,$7D,$6F,$FC,$13,$8E,$E3,$FD,$21,$F8,$FB,$F2,$17,$6B
	.dc.b	$EC,$35,$AA,$B5,$2F,$D2,$B3,$18,$C6,$E0,$FB,$EF,$F6,$DF,$CB,$FB
	.dc.b	$E7,$E4,$FC,$5B,$F1,$27,$E2,$4F,$BC,$FE,$13,$F6,$FF,$E4,$9E,$6F
	.dc.b	$EB,$18,$7C,$7E,$1F,$E6,$98,$70,$BD,$F9,$5F,$8B,$CF,$B5,$6A,$FF
	.dc.b	$53,$87,$27,$C3,$FA,$27,$FB,$D6,$B7,$A7,$8F,$4F,$17,$C0,$7F,$6D
	.dc.b	$C3,$86,$3E,$09,$CF,$6F,$B3,$2F,$26,$1C,$7E,$05,$F9,$FB,$F4,$D7
	.dc.b	$D1,$79,$6F,$7C,$3F,$02,$F7,$4F,$D1,$35,$5F,$1A,$B5,$87,$F6,$88
	.dc.b	$CE,$9E,$3C,$3B,$FC,$5F,$8E,$78,$6A,$FC,$FD,$F9,$1F,$FB,$CB,$F7
	.dc.b	$2E,$3B,$FB,$E7,$3F,$86,$BE,$F7,$3B,$7D,$49,$E3,$EF,$75,$DB,$E1
	.dc.b	$47,$4E,$FF,$CD,$6E,$FC,$3B,$EF,$78,$8F,$BC,$7E,$31,$FE,$B9,$F8
	.dc.b	$A7,$E1,$6F,$D4,$1F,$99,$FF,$58,$C7,$8E,$1C,$3E,$21,$87,$0B,$FE
	.dc.b	$D1,$9F,$7E,$7F,$D5,$69,$DF,$4F,$E4,$F1,$CC,$CE,$3B,$7F,$AB,$D5
	.dc.b	$AE,$18,$F6,$D3,$F9,$15,$71,$13,$AC,$BD,$3E,$B7,$C0,$D3,$CF,$9F
	.dc.b	$FB,$57,$C9,$FF,$37,$FE,$7A,$FC,$7B,$96,$3D,$A7,$3F,$8F,$E3,$7F
	.dc.b	$74,$7F,$9D,$BD,$0E,$9A,$FA,$32,$E9,$F8,$07,$CB,$8E,$11,$FE,$E3
	.dc.b	$F1,$DF,$76,$FA,$77,$EF,$3A,$E4,$FF,$20,$4E,$7F,$13,$C5,$F0,$F9
	.dc.b	$4E,$F9,$7B,$B6,$FA,$13,$D3,$E5,$FE,$EF,$C0,$7E,$E7,$E1,$9F,$95
	.dc.b	$77,$F8,$B7,$E9,$1F,$EA,$3F,$81,$7E,$35,$FE,$12,$C3,$85,$EE,$B5
	.dc.b	$FA,$F6,$75,$FE,$45,$97,$4F,$C4,$FE,$CF,$83,$FC,$7E,$3B,$F2,$E3
	.dc.b	$2E,$6F,$FA,$3B,$1F,$F5,$97,$47,$9B,$1E,$79,$F7,$87,$E3,$F9,$9F
	.dc.b	$EC,$13,$E9,$FD,$9B,$FC,$B7,$FB,$F3,$E6,$3C,$95,$AC,$BF,$D3,$75
	.dc.b	$AF,$9D,$CE,$BE,$4E,$34,$FF,$99,$3E,$4F,$B4,$E6,$3F,$CC,$FF,$03
	.dc.b	$8D,$3E,$47,$EB,$AF,$C7,$9F,$84,$BF,$49,$EB,$5A,$9C,$FE,$5D,$8D
	.dc.b	$BF,$D5,$B8,$38,$7F,$96,$73,$7C,$7C,$38,$7D,$0F,$ED,$33,$FE,$61
	.dc.b	$E3,$BF,$3B,$78,$4F,$D2,$5F,$AF,$7F,$A8,$3F,$50,$7E,$D9,$FE,$4A
	.dc.b	$C7,$8E,$1C,$2F,$7F,$84,$2F,$FD,$3D,$AF,$FB,$67,$3E,$DF,$9F,$FD
	.dc.b	$AC,$78,$E9,$7A,$FF,$B0,$57,$8E,$3C,$5A,$FF,$AC,$3F,$0B,$F4,$56
	.dc.b	$B5,$F0,$4F,$FA,$D3,$F2,$DC,$CB,$0E,$5B,$7A,$BE,$09,$F8,$4F,$F3
	.dc.b	$2D,$BC,$99,$72,$9C,$FE,$F5,$ED,$C7,$B6,$BE,$29,$F9,$7B,$7A,$32
	.dc.b	$E9,$F7,$7E,$0D,$3B,$FE,$C8,$FB,$AF,$F5,$FF,$EE,$9C,$F3,$F2,$6B
	.dc.b	$F1,$EB,$5E,$17,$F7,$CD,$FC,$7B,$7D,$19,$CD,$F1,$CB,$AE,$5F,$C9
	.dc.b	$23,$A7,$CA,$FF,$12,$FC,$53,$FD,$65,$F4,$5F,$D3,$BC,$8F,$F7,$4E
	.dc.b	$7F,$8A,$BF,$F0,$65,$EF,$EF,$FF,$C3,$9F,$21,$F9,$DF,$37,$9F,$BA
	.dc.b	$FC,$D3,$EA,$61,$FD,$C6,$3F,$DD,$EF,$F9,$43,$1E,$8E,$74,$F0,$E1
	.dc.b	$FA,$78,$E7,$18,$FB,$2F,$AB,$9F,$87,$57,$E0,$DF,$8E,$FC,$6F,$7B
	.dc.b	$7F,$8B,$E3,$F2,$3F,$4B,$8A,$DF,$C7,$72,$DB,$FE,$00,$DE,$FA,$71
	.dc.b	$A7,$C2,$F8,$67,$F0,$0B,$F3,$EE,$51,$8F,$76,$FC,$8F,$0C,$7F,$C7
	.dc.b	$D3,$FE,$F9,$E4,$F6,$FD,$AF,$F7,$7F,$CC,$1E,$03,$F3,$67,$E7,$8F
	.dc.b	$D4,$BF,$D5,$18,$70,$F9,$E6,$1C,$2F,$74,$F7,$3F,$09,$97,$5E,$BF
	.dc.b	$99,$E3,$FF,$88,$E3,$FC,$0D,$F9,$C7,$D2,$7C,$5A,$F8,$75,$F0,$5F
	.dc.b	$2D,$3B,$CE,$7F,$0E,$FD,$1D,$F0,$D3,$FC,$15,$F8,$DF,$EF,$77,$F3
	.dc.b	$6F,$E5,$FA,$14,$FF,$81,$67,$38,$FB,$E7,$F9,$3E,$7D,$EF,$87,$C1
	.dc.b	$F6,$D3,$C3,$97,$F4,$0F,$D8,$FF,$96,$E5,$CA,$31,$FB,$37,$1C,$FE
	.dc.b	$57,$8C,$F9,$6E,$59,$FC,$78,$FC,$F6,$1F,$8E,$E3,$7E,$78,$FD,$D7
	.dc.b	$FC,$19,$F9,$9F,$FA,$5B,$F2,$DF,$F3,$ED,$EE,$B5,$D4,$AD,$70,$FE
	.dc.b	$CF,$1E,$7E,$97,$E7,$9E,$B9,$5C,$38,$DF,$FA,$8B,$E4,$BF,$86,$D6
	.dc.b	$B1,$E9,$F7,$EF,$AA,$CE,$1C,$3F,$66,$7E,$BE,$FD,$AB,$C1,$39,$84
	.dc.b	$E3,$4F,$5A,$DB,$9C,$63,$E5,$EE,$7E,$BF,$C0,$9E,$0F,$F2,$FE,$2F
	.dc.b	$81,$7F,$12,$73,$F2,$1D,$6F,$87,$3F,$2E,$3F,$F6,$77,$DF,$FF,$28
	.dc.b	$FC,$7F,$FC,$0D,$F8,$F3,$F2,$FF,$E5,$5F,$AC,$3D,$27,$E8,$3F,$C6
	.dc.b	$EF,$75,$A9,$7E,$97,$9F,$92,$63,$FC,$2A,$FF,$C2,$5F,$93,$59,$CE
	.dc.b	$1C,$7C,$7E,$17,$E1,$F8,$16,$9F,$E9,$09,$FE,$55,$F9,$7F,$F9,$8F
	.dc.b	$F6,$FF,$F2,$37,$E8,$8E,$7F,$F6,$8F,$D2,$EB,$CD,$56,$A7,$FD,$AB
	.dc.b	$E5,$66,$3F,$8E,$E1,$FE,$6D,$FE,$00,$B7,$C3,$EB,$D9,$75,$FA,$0F
	.dc.b	$E9,$B7,$E1,$9C,$E5,$F5,$F4,$FF,$02,$FE,$66,$FC,$B1,$FA,$6B,$F5
	.dc.b	$07,$E6,$6F,$DE,$7F,$BE,$7E,$73,$C2,$F7,$5A,$9F,$F0,$07,$A3,$9F
	.dc.b	$95,$FF,$0F,$67,$FE,$C5,$CF,$B6,$78,$84,$63,$0E,$5F,$29,$FE,$1D
	.dc.b	$8C,$69,$E1,$F8,$2F,$F1,$EF,$EB,$DF,$D9,$FF,$43,$FF,$5F,$7E,$D1
	.dc.b	$FC,$77,$EE,$DF,$9B,$FF,$97,$EB,$5F,$85,$AB,$5F,$A2,$E7,$C0,$63
	.dc.b	$FE,$36,$F3,$3F,$D7,$98,$7F,$73,$F2,$31,$C3,$FE,$DD,$8E,$78,$F7
	.dc.b	$8E,$57,$E2,$F7,$FF,$D8,$D0,$FC,$23,$F9,$9F,$E0,$7E,$1F,$CE,$1F
	.dc.b	$51,$FE,$CF,$FC,$EF,$F3,$15,$AC,$95,$AF,$BD,$53,$F3,$EE,$9F,$03
	.dc.b	$FC,$BF,$27,$E6,$1F,$E8,$6C,$B9,$69,$FE,$C6,$AD,$67,$D3,$4F,$F6
	.dc.b	$F7,$C7,$FF,$BD,$EF,$C1,$AF,$7F,$8D,$C6,$1C,$23,$CF,$E9,$BF,$54
	.dc.b	$FF,$61,$5F,$0C,$53,$9F,$73,$CB,$A0,$7C,$73,$7F,$85,$7C,$3C,$0E
	.dc.b	$F5,$BA,$D6,$FE,$7C,$5D,$DE,$7F,$BF,$FF,$78,$6A,$F6,$9F,$A6,$BF
	.dc.b	$80,$5F,$8E,$7F,$80,$EB,$FD,$A5,$C3,$FE,$B8,$D7,$C9,$1C,$6F,$85
	.dc.b	$6F,$F1,$0F,$B1,$97,$6C,$BA,$E1,$FE,$F0,$8C,$7D,$9F,$CF,$75,$87
	.dc.b	$2F,$A5,$FF,$29,$57,$23,$BD,$F9,$B4,$F2,$CF,$B2,$7E,$38,$FC,$D5
	.dc.b	$F4,$2F,$D7,$9F,$5E,$FC,$8F,$FD,$E3,$F8,$13,$CC,$E3,$87,$0B,$DD
	.dc.b	$6B,$E4,$F3,$8B,$F5,$CF,$C3,$77,$FD,$5D,$3C,$13,$DD,$FC,$41,$F6
	.dc.b	$71,$E9,$A7,$87,$8D,$FA,$4F,$D8,$AE,$3A,$78,$27,$FB,$7F,$3F,$FB
	.dc.b	$57,$C7,$FE,$57,$DC,$FC,$95,$F7,$AE,$07,$E3,$EF,$E0,$1F,$D4,$79
	.dc.b	$FE,$5C,$78,$E1,$C2,$F7,$FA,$5E,$B6,$FC,$75,$7F,$38,$F0,$7F,$45
	.dc.b	$87,$1F,$89,$FD,$7D,$7B,$CF,$D0,$46,$75,$F2,$6B,$E3,$FA,$BC,$D7
	.dc.b	$4B,$FD,$BE,$9E,$1F,$2B,$33,$FE,$2B,$9C,$DF,$F4,$F8,$7E,$9D,$E5
	.dc.b	$75,$8F,$6F,$4F,$F4,$9F,$B5,$67,$6F,$6E,$1C,$6F,$18,$F4,$B7,$D6
	.dc.b	$F8,$2F,$6F,$F8,$47,$8F,$79,$7A,$A3,$FC,$29,$B6,$36,$F9,$F1,$FE
	.dc.b	$17,$CB,$19,$F4,$F0,$BF,$6D,$F0,$7F,$1A,$7F,$21,$FB,$17,$EF,$ED
	.dc.b	$FB,$EB,$F4,$A7,$F0,$0B,$F1,$EF,$EB,$5B,$7E,$2F,$B5,$8F,$1C,$38
	.dc.b	$5B,$F6,$FD,$E7,$FC,$16,$F7,$E7,$7F,$B9,$B1,$E5,$8F,$FB,$26,$BF
	.dc.b	$CF,$11,$89,$F8,$78,$F8,$FF,$0B,$F4,$CE,$75,$F3,$EA,$FB,$FC,$E3
	.dc.b	$DF,$2F,$67,$43,$E7,$9F,$A8,$FF,$84,$F2,$E5,$39,$87,$CF,$FA,$30
	.dc.b	$F8,$BB,$7D,$4A,$E5,$AF,$69,$CE,$A9,$ED,$E6,$67,$83,$DF,$39,$DD
	.dc.b	$F9,$C7,$F3,$DE,$5F,$A5,$5A,$C3,$E8,$4F,$F8,$76,$73,$C1,$EA,$BF
	.dc.b	$E9,$F1,$8D,$3C,$91,$FE,$11,$87,$E2,$1F,$BF,$FB,$27,$DD,$BF,$35
	.dc.b	$7E,$D1,$7F,$69,$7E,$35,$F9,$0C,$78,$F1,$7F,$3B,$C7,$8E,$1C,$05
	.dc.b	$F0,$F9,$AF,$F0,$E6,$7D,$F3,$FE,$53,$87,$EC,$71,$E6,$FF,$A4,$A3
	.dc.b	$C4,$FF,$AD,$67,$86,$1D,$70,$FF,$43,$DF,$0C,$7C,$39,$F9,$FC,$7F
	.dc.b	$CC,$E1,$C2,$3D,$9F,$B3,$BF,$80,$7F,$D0,$B8,$71,$D3,$E0,$56,$A5
	.dc.b	$F9,$A7,$A3,$5F,$75,$6B,$56,$3F,$E6,$79,$D1,$E9,$8F,$85,$0F,$C4
	.dc.b	$3C,$73,$CA,$7E,$37,$E2,$6F,$C6,$BF,$DB,$BA,$F9,$E5,$39,$FC,$5F
	.dc.b	$AF,$07,$BA,$73,$F6,$3D,$31,$F0,$6F,$F6,$70,$FF,$4E,$F0,$7C,$CF
	.dc.b	$82,$F3,$FA,$4F,$DE,$BF,$CE,$F8,$BD,$3F,$E8,$57,$BE,$1F,$EF,$35
	.dc.b	$A9,$7E,$D3,$EE,$C7,$9E,$9F,$54,$FE,$F8,$FC,$4F,$C7,$7F,$C7,$4F
	.dc.b	$82,$FF,$D0,$BF,$0F,$FD,$37,$2E,$DA,$78,$FE,$D7,$FD,$0D,$E4,$7A
	.dc.b	$F0,$E1,$0F,$CD,$9F,$4E,$FC,$C3,$C5,$C1,$E6,$AD,$4B,$F2,$3C,$E5
	.dc.b	$D7,$0F,$87,$D1,$FD,$77,$2E,$B1,$FD,$63,$83,$7B,$FF,$A6,$BB,$1F
	.dc.b	$8A,$3F,$61,$CE,$BF,$84,$87,$E1,$B8,$AB,$85,$E1,$FE,$53,$F1,$1F
	.dc.b	$DA,$FF,$4C,$FE,$4F,$FB,$03,$F4,$07,$E3,$BF,$EA,$8F,$5B,$FD,$5B
	.dc.b	$D0,$C3,$87,$AE,$C3,$85,$EF,$F6,$0F,$F4,$35,$6B,$F0,$AF,$EB,$5A
	.dc.b	$F8,$0F,$F1,$76,$9E,$07,$FD,$E6,$63,$1E,$9A,$78,$B8,$7F,$92,$4F
	.dc.b	$43,$59,$F8,$23,$FD,$B3,$8F,$92,$3C,$7A,$FA,$EB,$80,$C7,$DF,$87
	.dc.b	$BA,$5F,$AC,$FF,$86,$F6,$F7,$5F,$6A,$D7,$E0,$DE,$98,$E5,$A7,$9B
	.dc.b	$7F,$E7,$1F,$83,$E7,$3E,$B0,$FC,$F9,$F8,$6B,$F7,$CF,$1D,$6A,$73
	.dc.b	$FA,$26,$31,$E9,$0F,$C6,$7A,$46,$3F,$35,$F1,$7E,$64,$FD,$13,$F0
	.dc.b	$7F,$9F,$BF,$6F,$FA,$9F,$9C,$FF,$8C,$B9,$3D,$7C,$4C,$38,$77,$B0
	.dc.b	$E1,$D0,$C1,$FB,$0F,$C8,$BD,$E1,$F5,$8E,$59,$BF,$38,$FB,$B1,$B8
	.dc.b	$F1,$E9,$F8,$57,$CE,$63,$17,$E2,$38,$D3,$C5,$5F,$8B,$FB,$87,$E4
	.dc.b	$BC,$E7,$37,$FE,$A6,$FB,$5C,$6F,$E8,$FC,$6F,$93,$EA,$F1,$AE,$4F
	.dc.b	$0F,$FD,$FD,$F8,$3F,$08,$C7,$AD,$FE,$D8,$E5,$F8,$BC,$7F,$13,$1F
	.dc.b	$95,$87,$FB,$3E,$73,$C3,$EC,$8F,$E3,$3E,$27,$F9,$4F,$91,$FB,$E7
	.dc.b	$F8,$3F,$F7,$47,$9D,$FD,$3B,$8B,$DF,$97,$D2,$BD,$FC,$32,$F7,$F0
	.dc.b	$9F,$AB,$FC,$97,$F2,$F5,$AE,$AF,$A0,$F6,$FB,$6F,$9F,$8F,$3D,$3C
	.dc.b	$1A,$7F,$5C,$FC,$E7,$F2,$D7,$DB,$3E,$BF,$8B,$62,$F7,$A7,$F9,$93
	.dc.b	$F3,$3E,$F1,$8F,$A8,$7E,$27,$8E,$0F,$36,$9F,$DA,$27,$94,$63,$0F
	.dc.b	$F5,$DE,$FB,$CF,$D7,$F6,$1F,$E1,$2D,$FC,$D7,$B8,$FE,$E3,$18,$9F
	.dc.b	$83,$F9,$17,$47,$E1,$FD,$0F,$EB,$FE,$B7,$FD,$F9,$F8,$BF,$F8,$EB
	.dc.b	$E5,$5F,$A9,$3F,$79,$7E,$85,$FB,$EB,$DF,$D1,$7F,$85,$AB,$5F,$AA
	.dc.b	$E7,$F0,$B6,$5F,$2B,$BB,$FD,$19,$82,$5F,$87,$67,$8F,$FD,$19,$1C
	.dc.b	$32,$E9,$CB,$E5,$8C,$6D,$E4,$FC,$0F,$19,$F9,$F5,$FE,$DD,$F8,$87
	.dc.b	$86,$31,$F2,$FF,$B1,$E5,$FD,$AE,$DE,$8C,$7D,$B3,$CB,$4F,$57,$BB
	.dc.b	$67,$E0,$FE,$22,$FC,$9F,$A5,$7B,$E3,$FD,$A3,$8F,$4D,$BC,$BA,$BD
	.dc.b	$FF,$E3,$5D,$AF,$05,$F9,$63,$F4,$0F,$E1,$5F,$5C,$FC,$E3,$FC,$5B
	.dc.b	$9F,$F0,$AC,$38,$5E,$FD,$0F,$F5,$2D,$7F,$C0,$0A,$D6,$3D,$3A,$3F
	.dc.b	$84,$61,$DE,$BE,$64,$7E,$9A,$7C,$39,$78,$31,$F2,$C6,$31,$FF,$58
	.dc.b	$FB,$B8,$9F,$06,$7F,$EC,$BF,$C2,$BF,$9E,$E7,$E4,$C7,$D3,$F5,$3F
	.dc.b	$E5,$1E,$DF,$ED,$99,$F3,$C3,$FC,$9D,$18,$C7,$9F,$D5,$75,$C3,$F9
	.dc.b	$57,$E7,$0F,$D3,$7F,$DF,$1C,$3E,$CB,$F8,$13,$9F,$BC,$F0,$E7,$E7
	.dc.b	$87,$B5,$CF,$85,$0F,$CD,$7F,$A5,$79,$2F,$AA,$7E,$98,$FA,$97,$EC
	.dc.b	$CF,$C2,$1F,$9F,$7E,$83,$83,$F9,$DE,$3C,$70,$E1,$7B,$A7,$EA,$BF
	.dc.b	$D8,$2B,$B7,$CB,$FD,$07,$87,$67,$F3,$B7,$96,$7E,$0C,$FE,$2C,$84
	.dc.b	$EB,$1F,$27,$37,$AB,$5F,$4E,$1C,$77,$7E,$ED,$E1,$9C,$DF,$AE,$5F
	.dc.b	$BC,$E7,$7F,$4E,$8F,$C4,$3A,$46,$3F,$0E,$B9,$FC,$4F,$01,$FB,$C3
	.dc.b	$6E,$53,$98,$7E,$C3,$8D,$7F,$B5,$7D,$FF,$27,$FB,$67,$67,$F7,$BF
	.dc.b	$C4,$BE,$A7,$ED,$3F,$DB,$FE,$3D,$F9,$CF,$FA,$8F,$E0,$FF,$C1,$35
	.dc.b	$AD,$BF,$AA,$CF,$FA,$9F,$E4,$33,$DC,$C7,$F3,$B7,$F7,$FF,$7D,$B3
	.dc.b	$9C,$38,$FC,$37,$86,$31,$9F,$83,$E8,$31,$8F,$8F,$37,$EE,$8F,$CF
	.dc.b	$FB,$CE,$61,$FB,$A6,$33,$F2,$FB,$BC,$9F,$3F,$B5,$73,$F6,$FD,$C3
	.dc.b	$F0,$AE,$11,$8F,$97,$BA,$FE,$6B,$C5,$BC,$1F,$E2,$1F,$41,$F9,$C7
	.dc.b	$F9,$1F,$EA,$5F,$50,$FC,$F1,$F8,$C3,$F6,$5F,$F1,$8E,$1C,$3E,$AB
	.dc.b	$FD,$33,$7B,$A7,$EB,$1A,$CB,$FC,$71,$A7,$FA,$C7,$0F,$C9,$3F,$E5
	.dc.b	$18,$7E,$69,$FC,$2A,$7A,$69,$E2,$D3,$FC,$87,$7F,$99,$9C,$E5,$E5
	.dc.b	$EA,$F5,$63,$E9,$8F,$47,$EA,$6F,$D5,$3C,$5B,$79,$B2,$4F,$FA,$8F
	.dc.b	$D5,$FF,$15,$E5,$D2,$7F,$C6,$D9,$F0,$D3,$C5,$E1,$F6,$C7,$97,$ED
	.dc.b	$1F,$B9,$C3,$D6,$71,$7F,$6F,$8F,$4D,$7C,$99,$7F,$89,$B2,$E9,$AF
	.dc.b	$CD,$FA,$1F,$F2,$97,$DA,$FF,$60,$FD,$79,$FA,$57,$F8,$A3,$1F,$E1
	.dc.b	$5F,$85,$FF,$98,$2B,$53,$9F,$39,$F8,$BE,$7E,$59,$8F,$F3,$4B,$FE
	.dc.b	$D7,$FA,$F5,$9C,$E1,$C7,$5E,$17,$E0,$EF,$FC,$1F,$CC,$F9,$2F,$71
	.dc.b	$F8,$6F,$8E,$FF,$50,$71,$7F,$94,$B3,$F1,$CE,$78,$39,$E7,$E9,$F8
	.dc.b	$CE,$59,$F5,$E1,$FF,$6D,$63,$C7,$3E,$B0,$FC,$E5,$FB,$07,$2C,$FF
	.dc.b	$2B,$1F,$BA,$BE,$18,$73,$F9,$8E,$B9,$FC,$7F,$BF,$FF,$3E,$7E,$08
	.dc.b	$FC,$97,$FD,$3D,$F8,$A3,$F3,$27,$E2,$CF,$94,$FE,$87,$DF,$F9,$4C
	.dc.b	$38,$2B,$E1,$5B,$F8,$FF,$F4,$8D,$6B,$A7,$FC,$D4,$F7,$F9,$CF,$EC
	.dc.b	$6F,$2B,$E3,$B7,$93,$4F,$1E,$9E,$2D,$71,$A7,$78,$F2,$FE,$FC,$FC
	.dc.b	$6F,$BD,$F0,$D9,$2E,$CC,$E1,$C6,$7F,$95,$6D,$8D,$3D,$3F,$2D,$D9
	.dc.b	$F8,$5F,$40,$F7,$9F,$8A,$EF,$64,$E7,$F4,$0C,$3F,$ED,$0E,$8F,$F4
	.dc.b	$8F,$1F,$B6,$B7,$F2,$BF,$EA,$BF,$D4,$DF,$93,$BF,$80,$3F,$8D,$FE
	.dc.b	$67,$FD,$C7,$E3,$1F,$BF,$FF,$AE,$AF,$75,$AF,$9C,$FE,$85,$F8,$0E
	.dc.b	$71,$E5,$C9,$FE,$97,$BE,$BF,$83,$7F,$84,$1E,$3F,$0A,$D3,$B5,$71
	.dc.b	$CB,$A7,$B3,$E5,$C7,$C9,$7B,$FC,$4D,$FB,$FF,$E1,$F8,$FF,$D5,$B9
	.dc.b	$F4,$CD,$5A,$F8,$7E,$98,$71,$FC,$A7,$A6,$9E,$EA,$FE,$DB,$AE,$B0
	.dc.b	$E3,$83,$E1,$B1,$48,$7E,$6D,$87,$9E,$5E,$5F,$88,$FF,$6B,$7D,$AB
	.dc.b	$E9,$5F,$A4,$BF,$26,$FE,$DD,$FA,$5B,$F7,$F7,$E6,$9F,$EE,$6C,$38
	.dc.b	$5E,$FE,$79,$7B,$FB,$AD,$64,$F8,$7F,$F1,$DE,$1C,$38,$FF,$D8,$CF
	.dc.b	$2D,$3B,$FC,$3F,$92,$FC,$71,$E2,$F9,$BC,$69,$DF,$F0,$CF,$81,$E3
	.dc.b	$F6,$8C,$73,$3F,$22,$C6,$1F,$E0,$7B,$FE,$AD,$D9,$FE,$DE,$D3,$C1
	.dc.b	$C2,$9F,$F1,$86,$DC,$23,$1F,$A9,$7F,$9C,$7C,$FF,$E9,$B7,$C2,$D3
	.dc.b	$FE,$C9,$C3,$A6,$FE,$8C,$3A,$7A,$BE,$8C,$FC,$5D,$3F,$E2,$7B,$7C
	.dc.b	$17,$F5,$8F,$C1,$9F,$8D,$BF,$3C,$63,$FE,$DE,$F8,$DF,$F3,$D7,$E9
	.dc.b	$6C,$78,$F1,$7F,$8C,$71,$E3,$87,$0F,$82,$61,$FE,$81,$F0,$78,$69
	.dc.b	$DF,$4F,$E3,$18,$FF,$68,$3C,$91,$9C,$3F,$D9,$2D,$D3,$C5,$F4,$BA
	.dc.b	$AF,$05,$F9,$E7,$CD,$F0,$F9,$96,$9E,$49,$FF,$7B,$FE,$EB,$FE,$59
	.dc.b	$F6,$3C,$75,$CB,$3F,$44,$FF,$97,$7F,$6E,$CC,$F2,$CB,$B6,$EA,$FF
	.dc.b	$31,$4E,$B2,$ED,$39,$F7,$3C,$51,$8F,$C3,$DE,$67,$F8,$3F,$8F,$E4
	.dc.b	$E1,$D7,$4F,$91,$5A,$F9,$7E,$9A,$78,$23,$8D,$62,$B7,$F3,$5F,$6F
	.dc.b	$F2,$1F,$F0,$8F,$C9,$7F,$8B,$BF,$05,$7E,$F3,$FC,$5F,$F6,$5F,$F6
	.dc.b	$8F,$A7,$F6,$6F,$7E,$8A,$F7,$2A,$FF,$6B,$FE,$C5,$BD,$3F,$CD,$D9
	.dc.b	$7F,$8E,$BF,$0B,$FF,$7E,$E5,$CB,$1B,$5A,$CB,$A6,$7E,$1E,$4F,$34
	.dc.b	$E6,$BB,$E7,$F2,$B8,$98,$F3,$C3,$F9,$77,$D1,$3F,$09,$FE,$09,$5A
	.dc.b	$C7,$95,$6A,$5F,$89,$79,$E7,$FD,$2D,$A7,$8B,$81,$8F,$F1,$AC,$79
	.dc.b	$56,$B4,$F8,$33,$FD,$46,$39,$6F,$F2,$F3,$FE,$81,$FA,$B3,$F1,$57
	.dc.b	$EE,$FA,$D3,$B4,$E7,$F2,$0C,$67,$F3,$23,$EC,$E9,$E0,$CB,$A6,$A9
	.dc.b	$FE,$5B,$F2,$6F,$3F,$FC,$97,$F7,$EF,$C6,$7F,$D1,$5F,$A6,$7F,$96
	.dc.b	$7F,$5A,$FF,$5C,$FA,$4C,$38,$71,$FF,$A8,$F0,$E1,$F1,$4C,$38,$7C
	.dc.b	$5F,$F2,$6B,$DD,$AF,$B7,$DB,$FE,$CE,$3F,$D5,$93,$A9,$E7,$19,$8C
	.dc.b	$67,$E4,$FA,$66,$9F,$E7,$39,$CE,$7E,$19,$CF,$4F,$4D,$BF,$CC,$FF
	.dc.b	$6D,$FE,$03,$FC,$3D,$FB,$5B,$F2,$EF,$E8,$1E,$6B,$F3,$DF,$76,$1C
	.dc.b	$3E,$71,$87,$0B,$DF,$E6,$5A,$E1,$FF,$62,$67,$FE,$0F,$FA,$6F,$64
	.dc.b	$64,$47,$85,$F6,$63,$D7,$0E,$9F,$15,$FE,$10,$E1,$F8,$F7,$E7,$DB
	.dc.b	$CD,$2F,$9B,$F6,$63,$FE,$D3,$FC,$45,$FA,$8B,$F6,$2F,$D6,$7D,$9F
	.dc.b	$FA,$6F,$F2,$7F,$C9,$E9,$F4,$AF,$75,$AF,$B8,$53,$D8,$F4,$56,$B2
	.dc.b	$F3,$9D,$F2,$FE,$D2,$E0,$47,$4F,$3B,$FD,$27,$87,$F9,$DF,$2F,$05
	.dc.b	$F9,$67,$3F,$5B,$F8,$0F,$7F,$FD,$37,$EF,$3F,$0F,$FF,$5E,$60,$FC
	.dc.b	$AF,$FC,$95,$FB,$0B,$F5,$2F,$C9,$61,$C2,$F7,$5A,$FD,$4F,$3B,$FF
	.dc.b	$BD,$F2,$E9,$F8,$0F,$D4,$CF,$F9,$1C,$77,$FD,$97,$19,$73,$AF,$CF
	.dc.b	$63,$FE,$59,$E0,$F3,$63,$CF,$3E,$FF,$08,$FC,$E7,$F3,$3E,$D7,$F6
	.dc.b	$8F,$DC,$9F,$A1,$3E,$59,$F9,$8B,$D8,$FF,$61,$7E,$1F,$FA,$BC,$38
	.dc.b	$5E,$EB,$5F,$81,$BF,$10,$D6,$C1,$F2,$ED,$3B,$C7,$E9,$F1,$F1,$7E
	.dc.b	$D7,$EF,$7A,$EB,$E2,$D7,$F7,$D5,$D0,$E7,$7E,$89,$D7,$E5,$1F,$DE
	.dc.b	$7F,$1A,$FF,$01,$FC,$43,$F1,$C7,$E6,$3F,$E8,$EF,$D6,$1F,$AB,$FF
	.dc.b	$8A,$3E,$0D,$87,$0B,$DF,$E0,$2B,$DF,$C7,$7F,$7D,$AE,$BE,$47,$EA
	.dc.b	$30,$E3,$BF,$BF,$8D,$AF,$86,$3F,$84,$4E,$9F,$87,$1E,$5D,$7C,$9F
	.dc.b	$79,$E9,$9E,$53,$9D,$3C,$5F,$2F,$FA,$A9,$7E,$89,$FE,$87,$FC,$3F
	.dc.b	$FC,$DB,$FB,$BF,$FA,$AF,$F3,$1D,$4E,$7E,$E1,$3F,$EA,$FE,$57,$E0
	.dc.b	$3F,$BF,$AE,$58,$FE,$95,$FF,$49,$7E,$3D,$F9,$1F,$6F,$FD,$F3,$1C
	.dc.b	$F0,$E5,$5C,$7D,$F6,$9E,$2F,$C1,$BF,$55,$87,$09,$CE,$9E,$29,$FF
	.dc.b	$6B,$FE,$57,$C6,$F8,$7E,$0E,$FD,$43,$17,$BF,$B7,$7E,$FD,$8D,$7F
	.dc.b	$D9,$11,$FE,$61,$F4,$FF,$9B,$C3,$F6,$17,$ED,$FF,$C1,$BF,$70,$FF
	.dc.b	$76,$FD,$D7,$E4,$6F,$75,$AF,$7D,$5A,$3F,$91,$61,$CF,$0F,$85,$C7
	.dc.b	$F9,$B8,$ED,$F0,$98,$8F,$9B,$87,$3B,$EF,$1C,$A3,$1F,$86,$7F,$10
	.dc.b	$8C,$61,$E5,$87,$ED,$5F,$07,$3F,$2C,$E6,$1F,$B3,$F0,$D3,$CF,$7F
	.dc.b	$82,$FD,$A3,$19,$A3,$F0,$BF,$9F,$7F,$3B,$FB,$87,$E3,$AF,$CE,$BF
	.dc.b	$D0,$7F,$50,$FC,$E5,$F9,$33,$0E,$17,$BF,$C0,$BF,$DA,$15,$AF,$98
	.dc.b	$CF,$EB,$1F,$07,$3E,$B9,$DC,$38,$67,$D7,$F2,$1F,$CC,$56,$B2,$E7
	.dc.b	$F8,$16,$2F,$79,$AF,$FD,$6B,$CB,$FD,$CB,$0E,$18,$F3,$FC,$A3,$CB
	.dc.b	$9F,$AB,$0F,$E0,$B7,$BC,$79,$FF,$00,$BE,$C7,$7A,$FA,$A3,$FC,$79
	.dc.b	$5F,$DF,$BF,$44,$F8,$BD,$8F,$CC,$7F,$D4,$9F,$4A,$FD,$51,$F9,$D7
	.dc.b	$F6,$5B,$DD,$6A,$7F,$C8,$98,$73,$D1,$FA,$47,$E1,$6F,$AF,$81,$FD
	.dc.b	$45,$E3,$F8,$57,$F8,$37,$97,$F7,$91,$8D,$1C,$D8,$C7,$FC,$03,$A7
	.dc.b	$7F,$C7,$7C,$51,$8C,$7F,$C2,$DA,$BC,$DC,$F0,$7A,$E7,$3F,$59,$8C
	.dc.b	$7D,$1F,$8A,$FF,$7F,$F5,$3F,$D8,$BF,$99,$7F,$8E,$3F,$21,$FC,$37
	.dc.b	$E6,$0F,$CB,$75,$D2,$AD,$76,$2B,$FC,$0F,$D1,$FE,$28,$C7,$9F,$3F
	.dc.b	$F6,$F7,$FD,$A3,$FB,$3F,$E9,$DE,$FF,$74,$D3,$F2,$9F,$81,$FF,$1D
	.dc.b	$9C,$C7,$E5,$FF,$32,$C6,$BE,$1B,$E1,$A7,$87,$F1,$4F,$2E,$DE,$4D
	.dc.b	$9F,$8A,$62,$F7,$0F,$0F,$19,$FB,$BF,$17,$FF,$10,$7E,$D8,$FD,$61
	.dc.b	$FB,$87,$F6,$4F,$CD,$5F,$95,$3F,$5F,$7B,$5C,$2F,$7F,$3A,$BD,$D7
	.dc.b	$F9,$27,$A7,$FC,$97,$97,$5F,$D7,$FF,$84,$3D,$F3,$B8,$F2,$C3,$BF
	.dc.b	$E3,$9F,$CF,$EB,$59,$75,$D1,$F9,$CE,$70,$E1,$1E,$5F,$B5,$F2,$65
	.dc.b	$E9,$8F,$47,$C0,$E3,$0E,$5F,$A9,$FF,$B7,$B6,$F2,$E3,$CA,$3F,$CD
	.dc.b	$D1,$FE,$62,$FC,$1B,$FB,$4F,$95,$FE,$DD,$FC,$B9,$F9,$87,$F6,$DF
	.dc.b	$92,$FF,$24,$56,$BE,$59,$5A,$F9,$C7,$E6,$BA,$F9,$87,$EB,$5F,$CA
	.dc.b	$70,$EB,$E4,$FF,$8D,$5E,$58,$76,$F9,$BF,$EC,$11,$8D,$3E,$8D,$73
	.dc.b	$D7,$C9,$1F,$DF,$FF,$79,$C6,$BE,$28,$FF,$01,$7E,$71,$8C,$BD,$3F
	.dc.b	$A9,$FF,$85,$F8,$7D,$DB,$32,$E7,$2F,$D7,$FF,$CC,$9F,$B1,$7D,$0F
	.dc.b	$BE,$7A,$0F,$C4,$9F,$A0,$BF,$4F,$7E,$63,$C3,$C1,$FF,$2F,$5E,$EB
	.dc.b	$5F,$85,$67,$1F,$E9,$59,$7D,$CF,$CA,$AE,$1C,$33,$EB,$F0,$7F,$E0
	.dc.b	$1A,$F1,$65,$CF,$F1,$0C,$5E,$F3,$F1,$7E,$6B,$D6,$BC,$1B,$7E,$23
	.dc.b	$8F,$FD,$3B,$8F,$2C,$FD,$3F,$10,$FD,$1F,$84,$63,$F2,$BF,$F4,$24
	.dc.b	$7F,$8E,$BF,$40,$FE,$01,$E6,$BE,$27,$FC,$FD,$EF,$6B,$EB,$15,$A9
	.dc.b	$FF,$41,$7E,$47,$9F,$D4,$7F,$B4,$63,$CF,$C3,$FF,$3A,$3D,$7D,$9F
	.dc.b	$D5,$5F,$FA,$26,$7A,$61,$CF,$F1,$5F,$F1,$15,$EF,$4F,$17,$EB,$7F
	.dc.b	$EE,$6E,$2F,$F5,$CC,$E7,$14,$FF,$8F,$A7,$3A,$F9,$B3,$EF,$F8,$0F
	.dc.b	$48,$C6,$3D,$3F,$00,$F0,$D7,$93,$07,$E2,$BF,$E5,$3F,$C5,$5F,$9D
	.dc.b	$3F,$2B,$7E,$E8,$FC,$CD,$FB,$9B,$F1,$D7,$EF,$7C,$70,$E1,$7B,$AD
	.dc.b	$7E,$A9,$9F,$CB,$FF,$CC,$9D,$1F,$E6,$BD,$3B,$E9,$88,$FF,$62,$63
	.dc.b	$CB,$4E,$FC,$FF,$E7,$E9,$E1,$39,$C9,$F9,$C6,$67,$8E,$1E,$8C,$3C
	.dc.b	$F6,$FD,$6F,$CD,$39,$C3,$DF,$F9,$D6,$36,$F4,$7E,$3B,$CF,$5F,$2C
	.dc.b	$7A,$FE,$3D,$F4,$FB,$D3,$E1,$C7,$F8,$3F,$D8,$FC,$56,$DF,$E8,$7F
	.dc.b	$03,$FA,$27,$E5,$EF,$D0,$1F,$BF,$BE,$09,$F8,$23,$F2,$E6,$5C,$B2
	.dc.b	$65,$CB,$1E,$38,$70,$F0,$B7,$E2,$7F,$A5,$6F,$7F,$24,$FD,$6F,$FC
	.dc.b	$F9,$39,$C7,$C7,$5F,$E7,$A8,$C5,$36,$F3,$6D,$E5,$DB,$FD,$41,$AF
	.dc.b	$8F,$1F,$14,$FA,$7F,$05,$CE,$3C,$B0,$E3,$AB,$F3,$77,$E7,$7E,$6D
	.dc.b	$7E,$26,$BD,$F8,$3D,$75,$FE,$B7,$D7,$B4,$E7,$4F,$8D,$3D,$AF,$8E
	.dc.b	$5D,$BE,$57,$11,$F0,$23,$C3,$F0,$EF,$C6,$FC,$13,$9C,$FE,$4C,$BF
	.dc.b	$1A,$CF,$0F,$B3,$7F,$F2,$E6,$DE,$2C,$BD,$7F,$8E,$7C,$2F,$C7,$3E
	.dc.b	$77,$C3,$3E,$CD,$F8,$5B,$F2,$F7,$92,$F5,$BF,$D2,$BE,$87,$DC,$C7
	.dc.b	$8E,$1C,$2F,$75,$FE,$9A,$F9,$ED,$71,$FF,$B3,$F3,$ED,$F8,$2F,$FA
	.dc.b	$D7,$1E,$3A,$5A,$D6,$5C,$F4,$FF,$5B,$C7,$9C,$CC,$F9,$63,$C9,$FA
	.dc.b	$5E,$67,$FD,$97,$5C,$7F,$0E,$FE,$85,$FB,$9F,$8A,$73,$87,$19,$FE
	.dc.b	$8D,$C5,$9D,$3E,$14,$E7,$F1,$DE,$91,$8E,$0F,$87,$7F,$4F,$F7,$1E
	.dc.b	$19,$F7,$EA,$7E,$9F,$FA,$0F,$C3,$FF,$35,$FA,$0B,$F6,$67,$E7,$2F
	.dc.b	$D8,$FF,$D0,$1F,$B0,$FE,$33,$A3,$87,$1F,$FB,$D7,$0E,$17,$BF,$AF
	.dc.b	$5F,$E4,$BA,$F7,$6B,$ED,$FA,$37,$FB,$87,$4F,$0E,$9E,$0A,$FF,$6F
	.dc.b	$C6,$32,$E9,$8B,$93,$F4,$B3,$9D,$7C,$BE,$6E,$30,$F2,$63,$EA,$9F
	.dc.b	$4F,$E3,$AF,$DC,$BC,$7A,$FA,$27,$3F,$A2,$E2,$F8,$70,$7A,$BF,$49
	.dc.b	$FD,$67,$5B,$5F,$04,$E7,$F1,$7F,$6F,$E0,$CF,$CE,$7F,$BC,$FE,$85
	.dc.b	$FC,$DB,$F1,$A7,$EA,$AF,$DD,$37,$5A,$F8,$85,$6B,$C1,$CF,$D0,$FC
	.dc.b	$5E,$61,$FE,$93,$B7,$E9,$99,$8C,$4B,$3F,$07,$99,$F0,$E7,$31,$E3
	.dc.b	$F4,$71,$B7,$9E,$F8,$69,$E2,$FD,$5D,$FA,$2F,$8B,$0F,$54,$3F,$59
	.dc.b	$C5,$6F,$F4,$3B,$C0,$A7,$EC,$9F,$80,$FC,$01,$FB,$43,$F3,$5F,$F6
	.dc.b	$57,$CF,$7F,$B2,$7A,$1F,$9F,$EE,$B5,$D5,$FE,$25,$FC,$55,$4F,$D0
	.dc.b	$35,$F9,$DF,$C1,$CF,$AE,$77,$1E,$39,$FF,$B5,$7E,$07,$E1,$56,$B3
	.dc.b	$E9,$F0,$38,$BD,$E8,$FC,$AB,$F8,$B7,$85,$F7,$FA,$FE,$8E,$3C,$73
	.dc.b	$FF,$70,$57,$F1,$FA,$D5,$7F,$A2,$30,$E7,$3F,$D4,$27,$37,$C7,$0E
	.dc.b	$DF,$AC,$7E,$93,$E7,$9F,$BE,$FC,$8F,$DD,$1F,$91,$7F,$B0,$BE,$09
	.dc.b	$E8,$6E,$B5,$39,$FC,$FF,$1D,$3F,$83,$C7,$9F,$E5,$36,$F7,$F8,$EB
	.dc.b	$EA,$7F,$09,$5E,$0B,$69,$E0,$A7,$E6,$3A,$C3,$8E,$AF,$C3,$3B,$DE
	.dc.b	$E3,$FD,$3F,$C5,$F0,$B2,$E9,$39,$FC,$17,$7B,$FA,$FF,$70,$FE,$07
	.dc.b	$F8,$A3,$E2,$3F,$C3,$9F,$60,$FD,$19,$FA,$F3,$F0,$2F,$E2,$FA,$FF
	.dc.b	$B1,$6F,$75,$A9,$7E,$9F,$D2,$7F,$C4,$DD,$DF,$C2,$8F,$D4,$E5,$DF
	.dc.b	$F7,$DF,$D4,$3E,$1F,$AD,$69,$E1,$D3,$C1,$7C,$E7,$39,$F9,$3E,$EF
	.dc.b	$F5,$98,$71,$C3,$86,$8F,$DB,$78,$46,$3F,$68,$B8,$79,$FF,$03,$BA
	.dc.b	$7B,$7F,$3F,$FE,$4B,$BF,$F0,$DF,$C9,$5F,$BE,$BC,$A7,$E8,$0F,$DE
	.dc.b	$5F,$A6,$70,$F8,$4F,$F7,$FD,$EE,$B5,$F8,$CE,$7F,$46,$FE,$0D,$9F
	.dc.b	$5E,$7F,$88,$7D,$1C,$38,$E7,$DB,$D0,$FE,$11,$5E,$3C,$BA,$7A,$38
	.dc.b	$BD,$E7,$E3,$F9,$2F,$16,$FE,$8C,$BC,$FB,$BF,$51,$CE,$5E,$58,$F6
	.dc.b	$F7,$7E,$A3,$4E,$D7,$C3,$E0,$7F,$A3,$FE,$9D,$FD,$1F,$F5,$27,$E7
	.dc.b	$1F,$D2,$7F,$4E,$FE,$0F,$F6,$FA,$F9,$25,$69,$4F,$F0,$CF,$C5,$B5
	.dc.b	$F9,$17,$F2,$7C,$7A,$73,$FC,$27,$B6,$5D,$7E,$13,$F9,$44,$63,$1E
	.dc.b	$7F,$99,$7F,$2B,$FB,$87,$E7,$AF,$D7,$BF,$84,$7E,$DE,$FD,$A3,$FD
	.dc.b	$FF,$FB,$0F,$F6,$0B,$DD,$6A,$5F,$92,$7F,$91,$F9,$FF,$84,$D7,$4F
	.dc.b	$DB,$3F,$9E,$3C,$38,$3E,$BB,$CB,$E1,$19,$F7,$F3,$7C,$9A,$F8,$E7
	.dc.b	$39,$FF,$9D,$B0,$7E,$1A,$FA,$FF,$F2,$C7,$E4,$9F,$CF,$BE,$11,$F0
	.dc.b	$AF,$DA,$FF,$89,$F0,$7F,$A3,$F0,$E1,$8F,$F5,$2C,$38,$5E,$FE,$03
	.dc.b	$FC,$D3,$F8,$57,$6C,$FA,$FC,$B7,$F1,$8C,$3F,$CC,$C7,$F5,$AC,$79
	.dc.b	$AC,$7F,$A7,$FC,$46,$A3,$57,$CB,$5F,$2E,$BF,$C9,$BA,$F3,$5E,$3C
	.dc.b	$BF,$92,$EB,$E3,$F0,$BB,$79,$7F,$E7,$5F,$CC,$5F,$B8,$BE,$0F,$FC
	.dc.b	$05,$FB,$23,$F4,$C7,$E9,$2F,$27,$E3,$E3,$C7,$0E,$1F,$50,$C3,$85
	.dc.b	$BC,$8D,$EF,$FD,$DB,$4E,$FE,$67,$AE,$32,$23,$C7,$AE,$79,$FF,$A4
	.dc.b	$C6,$32,$ED,$B3,$D4,$F5,$E1,$D7,$5F,$1E,$BE,$9F,$9E,$CC,$F5,$9C
	.dc.b	$DF,$B3,$F0,$7F,$0E,$5D,$70,$F4,$FE,$ED,$8B,$DF,$07,$F7,$8D,$7C
	.dc.b	$71,$8F,$91,$FB,$99,$FF,$84,$BF,$40,$7E,$6A,$FC,$BF,$FC,$79,$ED
	.dc.b	$3D,$9F,$E3,$1F,$AB,$F0,$C1,$87,$0B,$DD,$7F,$1D,$FD,$93,$53,$D3
	.dc.b	$E4,$BE,$27,$85,$FE,$AD,$8E,$18,$7E,$17,$6F,$C4,$C7,$3F,$4F,$F9
	.dc.b	$2E,$2F,$CE,$BF,$0B,$7E,$0D,$BC,$D5,$AF,$C3,$F3,$87,$1D,$FD,$1E
	.dc.b	$DF,$FB,$D3,$3E,$B8,$70,$9C,$FE,$2B,$EB,$FD,$6F,$FC,$4B,$CC,$F5
	.dc.b	$FF,$D2,$9F,$B2,$BF,$12,$7E,$AB,$BF,$25,$7B,$FB,$F5,$EF,$F8,$07
	.dc.b	$53,$F0,$38,$FF,$C7,$F9,$FF,$8C,$B6,$C4,$23,$C3,$A7,$83,$F1,$6F
	.dc.b	$7C,$63,$2E,$BE,$96,$34,$F0,$DF,$9B,$67,$E2,$FE,$08,$C7,$E7,$3F
	.dc.b	$5F,$EB,$5E,$3B,$F5,$5F,$E5,$7F,$93,$FE,$87,$95,$F8,$D6,$EB,$5F
	.dc.b	$10,$AD,$7E,$E5,$9C,$3D,$38,$F4,$FE,$01,$B8,$70,$BF,$D6,$70,$7D
	.dc.b	$8A,$F0,$DF,$83,$CA,$72,$67,$0E,$10,$D7,$FC,$03,$D9,$FC,$33,$0E
	.dc.b	$3A,$F8,$EB,$5F,$83,$F8,$E3,$17,$FE,$4C,$BE,$18,$74,$8F,$F2,$14
	.dc.b	$3F,$58,$FF,$7F,$7C,$A3,$E5,$1B,$7F,$97,$3E,$81,$FB,$33,$F7,$37
	.dc.b	$ED,$3F,$CC,$31,$E3,$87,$0F,$B0,$61,$C2,$DF,$97,$6F,$C0,$FF,$3F
	.dc.b	$E9,$DF,$E9,$7F,$83,$65,$CB,$5B,$5A,$CF,$A6,$5E,$2E,$5F,$E5,$D7
	.dc.b	$BD,$3B,$ED,$FC,$CF,$7F,$66,$3C,$72,$F5,$4F,$F4,$CF,$73,$C5,$2D
	.dc.b	$7E,$05,$6B,$E3,$7F,$A3,$61,$CA,$9F,$BC,$3F,$06,$FF,$43,$7E,$06
	.dc.b	$FC,$97,$F1,$5F,$5C,$FC,$63,$AD,$EF,$EF,$FF,$D4,$95,$AF,$CF,$33
	.dc.b	$8F,$FA,$4B,$2F,$E0,$1C,$5F,$EA,$27,$BC,$7F,$A4,$DC,$9F,$EA,$6C
	.dc.b	$3A,$67,$E1,$F0,$3F,$B1,$6B,$E7,$9C,$E5,$F9,$DD,$FE,$3E,$1C,$71
	.dc.b	$F4,$CF,$A3,$F5,$DF,$1E,$BE,$6C,$BF,$CC,$FE,$3F,$F6,$CF,$D6,$FC
	.dc.b	$9F,$8B,$7F,$11,$FA,$07,$EA,$AE,$DF,$F7,$1C,$3F,$3D,$E1,$B7,$FA
	.dc.b	$F2,$F8,$5E,$FF,$12,$5F,$C9,$E2,$65,$DB,$DD,$7E,$2B,$FD,$26,$31
	.dc.b	$57,$2E,$95,$CB,$1E,$9F,$7B,$FD,$C6,$F7,$1D,$FD,$6C,$5F,$0C,$3D
	.dc.b	$3B,$7F,$BE,$BE,$9B,$F5,$B8,$70,$8C,$79,$7E,$AB,$6F,$EA,$9E,$1C
	.dc.b	$B9,$8C,$70,$7F,$8D,$BF,$6A,$7C,$E7,$FA,$07,$0F,$F6,$EF,$E9,$8F
	.dc.b	$C7,$7F,$B9,$7D,$C3,$E3,$3F,$E0,$07,$E7,$7F,$EE,$DC,$38,$60,$C3
	.dc.b	$87,$CC,$30,$E1,$F8,$BB,$0F,$F3,$4E,$1C,$2F,$78,$7F,$83,$E3,$19
	.dc.b	$3F,$2D,$FE,$F1,$18,$CB,$A6,$6C,$9F,$95,$7F,$BD,$2B,$5A,$78,$3F
	.dc.b	$09,$C7,$99,$F9,$7D,$7C,$B3,$E4,$FD,$DB,$D5,$39,$D7,$DF,$9F,$59
	.dc.b	$F3,$CE,$74,$F0,$E8,$FC,$AF,$C1,$CF,$FA,$3F,$41,$FA,$43,$F2,$87
	.dc.b	$E4,$EF,$D7,$BE,$17,$E3,$DF,$C5,$B8,$FF,$17,$CC,$3F,$DE,$B8,$71
	.dc.b	$BE,$1C,$EB,$7C,$4F,$D8,$D3,$C1,$6F,$BB,$F8,$F1,$8D,$2E,$5D,$75
	.dc.b	$F2,$C7,$93,$5F,$A9,$F4,$BE,$AA,$D6,$3E,$19,$FB,$1F,$B5,$E6,$7C
	.dc.b	$B7,$C7,$57,$A4,$FC,$87,$C9,$5C,$AB,$52,$FC,$8B,$F1,$FE,$AF,$F0
	.dc.b	$9F,$5B,$FD,$D7,$FB,$1F,$FA,$23,$F5,$73,$FC,$7F,$ED,$B7,$97,$C5
	.dc.b	$AD,$F9,$4B,$DF,$BA,$FD,$F3,$76,$BF,$F7,$BF,$EC,$9F,$8A,$CF,$C1
	.dc.b	$F6,$3D,$AB,$F2,$11,$8D,$3C,$47,$E8,$FB,$3F,$96,$61,$E0,$CF,$CB
	.dc.b	$F3,$BF,$8F,$BE,$1A,$F8,$F0,$FE,$6D,$FA,$87,$8B,$0E,$11,$FC,$E7
	.dc.b	$E5,$F1,$A7,$BA,$F8,$7E,$4C,$FC,$9B,$37,$E2,$87,$E4,$FF,$B6,$FA
	.dc.b	$4F,$E6,$1F,$9F,$3A,$DF,$8E,$BF,$70,$7E,$5A,$FA,$97,$E6,$1F,$B9
	.dc.b	$E6,$FF,$3F,$E3,$C7,$0E,$17,$BA,$7E,$45,$DF,$A3,$F2,$97,$B7,$F6
	.dc.b	$DA,$78,$B4,$7C,$EF,$ED,$A3,$19,$74,$D1,$EA,$FE,$D6,$73,$B7,$9A
	.dc.b	$7F,$73,$8F,$71,$97,$96,$75,$A3,$F3,$FD,$4E,$76,$F4,$7E,$39,$C7
	.dc.b	$F1,$2F,$7F,$CB,$3E,$83,$F8,$EE,$FF,$C6,$7F,$14,$D6,$7F,$E0,$99
	.dc.b	$CF,$B0,$FC,$5B,$38,$7A,$F7,$FD,$A6,$5F,$2B,$F8,$07,$10,$8E,$F8
	.dc.b	$76,$FC,$03,$DD,$3C,$32,$ED,$D1,$F5,$BF,$0C,$CC,$F1,$9C,$ED,$FC
	.dc.b	$B6,$7C,$38,$79,$3F,$63,$F8,$9D,$EF,$C8,$FF,$CF,$DD,$7F,$7B,$F5
	.dc.b	$7F,$E6,$1F,$2D,$A9,$CF,$32,$7F,$9A,$7C,$A6,$7A,$FD,$38,$F3,$FB
	.dc.b	$8B,$87,$0B,$FE,$15,$C1,$CE,$7F,$1D,$8F,$3F,$7B,$D3,$87,$1D,$3C
	.dc.b	$5F,$0A,$FC,$47,$84,$63,$F0,$0B,$87,$8E,$BF,$C2,$9C,$BF,$ED,$0C
	.dc.b	$39,$4E,$7F,$01,$FC,$17,$E0,$DF,$E5,$9F,$DC,$9F,$A3,$FF,$99,$BF
	.dc.b	$06,$FD,$F7,$87,$AF,$B0,$56,$BE,$CF,$FD,$85,$EC,$3F,$44,$D4,$FD
	.dc.b	$8C,$BF,$CC,$9F,$B9,$DC,$38,$E7,$D7,$87,$EA,$D7,$86,$3C,$1F,$0F
	.dc.b	$8A,$DE,$7D,$BF,$4E,$EB,$5E,$5D,$1F,$96,$7F,$6C,$FC,$2F,$FC,$4D
	.dc.b	$F8,$C3,$F7,$F7,$EB,$6F,$C7,$DF,$B8,$7F,$A6,$7E,$29,$F9,$5F,$1F
	.dc.b	$07,$FA,$26,$3C,$70,$E1,$7B,$F4,$75,$4A,$ED,$F0,$9F,$DA,$34,$F0
	.dc.b	$E9,$E0,$AF,$ED,$D1,$8C,$79,$F8,$FF,$B2,$EF,$4E,$76,$F2,$CF,$EE
	.dc.b	$AF,$BE,$5E,$39,$CE,$CF,$C3,$F3,$9F,$BA,$73,$F8,$17,$9B,$F7,$97
	.dc.b	$E6,$7F,$EC,$5F,$8D,$FE,$89,$F9,$AF,$FA,$8B,$E4,$F5,$39,$F8,$F4
	.dc.b	$FF,$A9,$BF,$37,$CF,$E1,$7E,$DE,$0F,$ED,$19,$74,$FA,$DC,$42,$31
	.dc.b	$9F,$6F,$D6,$7F,$C0,$DC,$9F,$E5,$99,$CE,$7D,$BE,$8B,$38,$74,$D7
	.dc.b	$C5,$AF,$FA,$DB,$D9,$F2,$E9,$FD,$83,$0F,$2F,$83,$FC,$8B,$A3,$E4
	.dc.b	$E3,$C6,$B5,$87,$FA,$FB,$F7,$DE,$3B,$FC,$0C,$3F,$DF,$38,$70,$C3
	.dc.b	$F9,$E7,$E7,$7F,$ED,$8F,$B2,$7E,$B0,$FC,$BD,$FB,$1B,$D0,$FC,$0F
	.dc.b	$AA,$C3,$85,$EE,$9F,$9F,$EA,$7F,$DC,$79,$7E,$08,$FF,$08,$5F,$1F
	.dc.b	$C3,$FE,$4B,$CF,$F1,$16,$9E,$2D,$3F,$D1,$57,$CE,$73,$9F,$97,$EF
	.dc.b	$7E,$4C,$FA,$A7,$35,$E8,$FC,$0B,$86,$1C,$F5,$7E,$BD,$F9,$AF,$D0
	.dc.b	$5F,$1C,$FD,$89,$FB,$77,$F8,$FF,$F6,$47,$E6,$FF,$C1,$39,$BE,$46
	.dc.b	$1C,$2F,$7F,$5E,$BD,$FE,$CB,$D7,$27,$E1,$3F,$86,$C3,$87,$17,$C3
	.dc.b	$0F,$3D,$3C,$1F,$4B,$EC,$8C,$65,$D7,$EC,$F1,$8F,$9E,$FC,$DB,$3F
	.dc.b	$49,$F0,$6D,$E4,$DB,$D3,$FA,$3F,$EE,$3E,$9D,$FA,$1B,$F4,$7F,$F3
	.dc.b	$77,$96,$F4,$3F,$C9,$3F,$A4,$6E,$B5,$ED,$7E,$DA,$7F,$C9,$BF,$23
	.dc.b	$9F,$C0,$BF,$43,$97,$4C,$AE,$1C,$32,$E9,$C3,$F5,$AB,$59,$73,$F2
	.dc.b	$F1,$7B,$CF,$FC,$B3,$F5,$1E,$1C,$BC,$B1,$FE,$5C,$FD,$BB,$18,$78
	.dc.b	$FE,$E7,$FC,$73,$DB,$F9,$BC,$FA,$EB,$EB,$9F,$F7,$A4,$FE,$7F,$F0
	.dc.b	$7F,$F2,$C7,$D4,$BD,$67,$EF,$EF,$CB,$DF,$88,$BF,$7A,$FE,$3F,$87
	.dc.b	$0B,$DD,$6B,$F7,$8C,$FC,$77,$F8,$A3,$C4,$FB,$97,$FE,$65,$E4,$FE
	.dc.b	$0A,$F1,$CF,$B7,$E1,$1F,$98,$8C,$62,$F3,$B1,$8F,$F8,$CB,$3E,$FA
	.dc.b	$BF,$5C,$EF,$BF,$A7,$3F,$44,$BF,$67,$FE,$6B,$FA,$A7,$EE,$3F,$5A
	.dc.b	$76,$7F,$58,$C9,$FB,$17,$FA,$A7,$97,$F8,$15,$EE,$B5,$2F,$D2,$79
	.dc.b	$FA,$1F,$E8,$3E,$5F,$CC,$5F,$56,$C3,$2E,$FF,$33,$F3,$7E,$87,$F6
	.dc.b	$B7,$D6,$3E,$87,$E1,$58,$BE,$19,$F8,$3E,$57,$BE,$3C,$B6,$F3,$FE
	.dc.b	$63,$FD,$43,$F5,$F7,$EC,$4F,$D6,$5F,$AD,$7F,$BB,$FD,$AF,$F5,$AF
	.dc.b	$E9,$1B,$AD,$4E,$76,$7C,$3E,$70,$FE,$73,$8F,$4F,$C0,$2E,$1C,$3F
	.dc.b	$15,$FE,$27,$87,$F0,$EA,$F0,$DF,$82,$7F,$8B,$F2,$66,$F8,$69,$FE
	.dc.b	$C6,$F0,$7B,$69,$E1,$FD,$36,$E1,$D3,$F4,$FB,$7F,$EC,$CF,$D9,$BF
	.dc.b	$B6,$FE,$CE,$FD,$8D,$F9,$CB,$F7,$07,$ED,$5F,$B7,$FD,$91,$FB,$87
	.dc.b	$0B,$DD,$6B,$D5,$57,$F9,$DB,$C4,$FE,$81,$97,$5C,$BF,$AC,$BD,$FF
	.dc.b	$1E,$B8,$F2,$CF,$F8,$27,$99,$FC,$BA,$B8,$67,$D7,$CE,$C5,$EF,$4E
	.dc.b	$FF,$89,$F8,$F3,$F4,$65,$E7,$F9,$7F,$F0,$4F,$D6,$BF,$7D,$71,$3F
	.dc.b	$5B,$FF,$45,$FE,$95,$FE,$8C,$FA,$CD,$D6,$A7,$3F,$6C,$FC,$0B,$3F
	.dc.b	$05,$FD,$CB,$2E,$99,$5C,$38,$5B,$2F,$8F,$E7,$F4,$AF,$04,$E7,$F3
	.dc.b	$AF,$E9,$78,$72,$CF,$BF,$D0,$6D,$F0,$FE,$1B,$FE,$22,$FC,$31,$FB
	.dc.b	$53,$F5,$DF,$F3,$42,$FE,$00,$7E,$A6,$FD,$3F,$84,$EF,$E3,$78,$D6
	.dc.b	$BE,$D3,$38,$7F,$83,$31,$EB,$D9,$70,$E0,$75,$F2,$BF,$B0,$CF,$8A
	.dc.b	$BC,$38,$3F,$2A,$CD,$F0,$8F,$1F,$ED,$3D,$B3,$F0,$EB,$FE,$DD,$F1
	.dc.b	$7F,$C0,$99,$73,$9C,$FE,$B1,$FE,$40,$FD,$65,$FB,$A3,$F4,$8F,$77
	.dc.b	$D7,$FF,$78,$F9,$D7,$DA,$6E,$B5,$B7,$F8,$CB,$CF,$53,$ED,$35,$8F
	.dc.b	$F8,$AB,$2E,$BF,$89,$DC,$78,$E7,$D7,$C0,$DE,$7F,$E3,$CF,$7D,$1F
	.dc.b	$E0,$2C,$3C,$B8,$F9,$34,$EF,$FA,$BF,$A7,$E8,$1F,$A9,$7F,$8D,$FF
	.dc.b	$29,$7E,$A0,$FC,$B7,$D9,$CF,$FE,$25,$CF,$FC,$03,$7B,$AD,$7E,$DB
	.dc.b	$CF,$E9,$FD,$19,$74,$DB,$FB,$E6,$1C,$3E,$E9,$A7,$6F,$84,$F3,$5F
	.dc.b	$8A,$73,$93,$E0,$BF,$C3,$7F,$38,$FC,$1B,$FD,$99,$F1,$8F,$87,$7E
	.dc.b	$4F,$F9,$BF,$AA,$FF,$39,$DE,$FF,$16,$5E,$EB,$F4,$1F,$A2,$6B,$1F
	.dc.b	$37,$0F,$C8,$3B,$BF,$ED,$97,$FC,$29,$87,$3B,$E5,$EA,$EB,$1E,$B9
	.dc.b	$F8,$BE,$FB,$15,$AF,$CC,$FF,$D0,$BF,$97,$7F,$49,$FC,$ED,$FB,$77
	.dc.b	$F6,$D8,$F4,$FE,$13,$AF,$9B,$56,$BE,$8D,$5E,$EF,$D9,$F5,$3F,$E6
	.dc.b	$E8,$FE,$A9,$FB,$5D,$BE,$38,$F5,$FC,$47,$FC,$87,$5E,$18,$F0,$7D
	.dc.b	$36,$2B,$79,$76,$E6,$7E,$3F,$FE,$9C,$FC,$A1,$F9,$F3,$EF,$1F,$23
	.dc.b	$FE,$46,$87,$E8,$0F,$0B,$84,$6E,$75,$3F,$8B,$F2,$3F,$CD,$D8,$74
	.dc.b	$FC,$53,$FD,$0E,$F6,$C6,$E1,$C7,$E2,$BC,$38,$7F,$97,$2B,$58,$F8
	.dc.b	$7E,$53,$17,$B3,$C7,$F5,$9E,$2C,$7C,$F8,$F2,$FA,$9F,$F5,$BF,$DE
	.dc.b	$FF,$83,$7E,$25,$FA,$17,$E5,$38,$5F,$B6,$BF,$79,$DE,$AB,$DF,$9D
	.dc.b	$F1,$AB,$FB,$8F,$1F,$FA,$27,$2E,$9E,$3F,$F9,$6D,$FF,$0D,$7C,$5F
	.dc.b	$FA,$91,$F0,$69,$DF,$CC,$FF,$3B,$C6,$31,$FA,$3C,$F8,$D3,$C1,$7E
	.dc.b	$5C,$FF,$C6,$99,$F7,$9C,$DF,$8F,$F4,$EF,$3E,$BE,$6D,$FD,$7F,$9E
	.dc.b	$E3,$2E,$5F,$37,$FE,$D8,$FC,$13,$FC,$47,$FA,$CB,$EA,$3E,$D7,$EA
	.dc.b	$AA,$7C,$F6,$FE,$B9,$7B,$E2,$5E,$FF,$73,$D7,$D7,$7F,$B0,$B3,$EB
	.dc.b	$D1,$FE,$7D,$7B,$E9,$DB,$4F,$F4,$BC,$63,$1E,$7E,$9E,$34,$EF,$7B
	.dc.b	$FC,$0B,$C5,$AF,$87,$5F,$37,$CD,$FF,$BC,$BE,$A1,$FA,$4F,$F5,$2F
	.dc.b	$C3,$BD,$3B,$3E,$EF,$FD,$3B,$D7,$FE,$BF,$BD,$FC,$B2,$F7,$4F,$CC
	.dc.b	$3A,$63,$FD,$73,$8F,$FD,$2D,$87,$6F,$85,$FF,$6F,$BC,$B0,$EF,$F9
	.dc.b	$97,$FA,$CA,$31,$D1,$F2,$FE,$87,$18,$F9,$63,$E7,$66,$FC,$FF,$BE
	.dc.b	$BE,$39,$CF,$E8,$2D,$FF,$DA,$DF,$6C,$FD,$03,$D9,$FB,$E3,$E6,$9F
	.dc.b	$A8,$FE,$5F,$9F,$87,$C3,$7F,$1D,$C3,$86,$CC,$38,$5F,$E7,$7F,$5B
	.dc.b	$DF,$37,$F4,$AB,$ED,$F8,$27,$FB,$F1,$E3,$1F,$EB,$77,$9D,$F2,$FC
	.dc.b	$CF,$FD,$A3,$18,$D3,$C7,$F5,$B8,$DB,$CF,$1E,$6F,$25,$F9,$1F,$83
	.dc.b	$26,$1C,$75,$FD,$87,$B9,$98,$C5,$BF,$67,$C6,$5D,$2F,$F6,$DE,$0F
	.dc.b	$3F,$4F,$FD,$B7,$F9,$4B,$E1,$7F,$C5,$DF,$99,$3D,$1F,$CD,$F4,$FF
	.dc.b	$C0,$1B,$DD,$6A,$7B,$4E,$7D,$97,$EC,$DF,$D1,$F8,$7F,$A4,$4E,$72
	.dc.b	$C4,$7F,$BC,$B1,$E7,$F6,$4C,$FF,$C6,$13,$E5,$C5,$7E,$5A,$D7,$91
	.dc.b	$C6,$73,$A7,$8F,$E1,$38,$5E,$F1,$F5,$7E,$61,$F2,$7E,$91,$F9,$B3
	.dc.b	$C7,$7E,$0D,$FE,$0D,$F6,$9F,$BF,$7F,$A8,$38,$9E,$27,$F6,$1C,$78
	.dc.b	$E1,$C1,$FE,$EF,$5F,$E1,$EA,$DF,$17,$F8,$3A,$76,$FF,$BE,$5F,$0F
	.dc.b	$E2,$7F,$0A,$3C,$99,$F8,$F3,$F1,$79,$BF,$E1,$6C,$39,$D7,$F8,$86
	.dc.b	$B5,$87,$F8,$1B,$2D,$67,$E5,$9C,$E9,$EA,$FA,$0E,$B5,$E7,$8F,$6F
	.dc.b	$AF,$F8,$BF,$D9,$BF,$CB,$DE,$6B,$F1,$BF,$89,$F6,$1F,$E1,$EF,$C3
	.dc.b	$75,$ED,$FF,$96,$27,$3F,$3F,$FE,$86,$E1,$F2,$E3,$CB,$87,$EE,$C6
	.dc.b	$2D,$EB,$FB,$E3,$19,$F7,$D9,$A7,$FB,$B7,$0E,$F3,$98,$FF,$24,$46
	.dc.b	$33,$F0,$FE,$0D,$6F,$8E,$8F,$D4,$B1,$87,$F8,$DB,$E4,$2C,$BF,$6F
	.dc.b	$FD,$57,$EE,$5F,$CD,$BF,$2F,$FE,$11,$F8,$6B,$F6,$17,$E3,$6F,$D3
	.dc.b	$B7,$E6,$7F,$97,$EB,$52,$FD,$DB,$3A,$FB,$7B,$7F,$C4,$D9,$F5,$E3
	.dc.b	$F8,$71,$8C,$79,$66,$E8,$FB,$35,$FE,$54,$BF,$27,$37,$FB,$42,$BF
	.dc.b	$D3,$D3,$E2,$DB,$CF,$1E,$6F,$A7,$C5,$6F,$57,$D5,$7F,$07,$F2,$3F
	.dc.b	$D1,$7F,$AD,$BF,$41,$7E,$44,$F1,$DF,$8D,$BE,$DB,$87,$DF,$B0,$E1
	.dc.b	$7B,$A7,$ED,$DA,$FC,$E3,$E0,$71,$7F,$9B,$34,$EE,$FC,$03,$18,$F3
	.dc.b	$D3,$C3,$A3,$ED,$7E,$0D,$2F,$CF,$5A,$FD,$1F,$31,$8D,$3C,$5F,$B0
	.dc.b	$79,$B6,$F2,$E5,$EA,$FC,$43,$FA,$17,$E0,$9F,$E5,$7F,$CA,$7F,$77
	.dc.b	$FB,$2B,$F7,$D7,$E8,$CF,$A3,$7E,$38,$FD,$E3,$8F,$17,$FA,$1B,$1E
	.dc.b	$38,$70,$BD,$FE,$73,$AA,$57,$67,$F4,$39,$F8,$7D,$AF,$F1,$34,$60
	.dc.b	$E8,$B4,$F2,$69,$E3,$F4,$BF,$80,$5F,$19,$CE,$5F,$E0,$FC,$3C,$18
	.dc.b	$F9,$30,$F5,$6B,$FE,$81,$EE,$63,$C7,$5F,$44,$FE,$12,$73,$87,$8B
	.dc.b	$2E,$B2,$FB,$4E,$BA,$F9,$70,$E3,$F8,$BF,$BF,$1D,$7C,$7B,$EF,$5F
	.dc.b	$BA,$3F,$41,$F9,$5F,$B1,$6F,$3F,$F7,$85,$6A,$7E,$77,$CB,$E7,$0F
	.dc.b	$E0,$9C,$1F,$D1,$AD,$F8,$CE,$38,$7F,$91,$47,$2C,$FB,$7E,$6B,$F1
	.dc.b	$A7,$38,$78,$E7,$C5,$E6,$E6,$7F,$C0,$97,$C7,$3F,$BD,$DF,$E1,$D7
	.dc.b	$C9,$D4,$FD,$93,$EE,$79,$CF,$C8,$5F,$A9,$7E,$A7,$E6,$BF,$C5,$3E
	.dc.b	$9F,$F8,$37,$9B,$E5,$EC,$C3,$85,$EF,$E8,$BF,$A4,$FE,$4D,$D3,$27
	.dc.b	$DD,$7B,$70,$E1,$1D,$9F,$C3,$18,$BF,$07,$E9,$FF,$DA,$BA,$3F,$86
	.dc.b	$D7,$93,$CB,$6D,$FE,$3C,$FC,$53,$E0,$6B,$E2,$9C,$EB,$FD,$6B,$F3
	.dc.b	$1F,$16,$7E,$5E,$0F,$E3,$DC,$0F,$D4,$7F,$D7,$DF,$51,$FE,$F7,$FD
	.dc.b	$13,$EF,$7E,$BE,$FE,$00,$7E,$65,$FA,$BC,$38,$72,$7F,$BB,$30,$E1
	.dc.b	$D2,$C3,$85,$BE,$2F,$FC,$6D,$9F,$6F,$C5,$9A,$7F,$45,$8C,$7E,$EF
	.dc.b	$73,$E9,$5A,$CB,$A6,$AF,$9B,$FE,$CD,$87,$3C,$7C,$F2,$FB,$5C,$E3
	.dc.b	$CA,$73,$A3,$F5,$7F,$1C,$E7,$2E,$B2,$FD,$77,$33,$FE,$51,$D3,$C5
	.dc.b	$BF,$B3,$F1,$7C,$F0,$FC,$08,$F7,$E3,$C3,$F0,$AF,$EC,$3E,$1B,$EE
	.dc.b	$DF,$8D,$32,$7E,$7C,$F8,$EE,$1F,$2F,$FC,$8E,$F7,$5A,$FA,$DC,$F9
	.dc.b	$1F,$17,$81,$A7,$67,$FB,$8F,$3F,$FB,$DE,$3C,$38,$72,$FC,$2F,$E1
	.dc.b	$D7,$0C,$BB,$6B,$E3,$F1,$73,$87,$08,$F3,$6D,$FE,$63,$ED,$FB,$F8
	.dc.b	$71,$E0,$F6,$56,$BF,$56,$FF,$39,$7E,$1E,$A9,$FA,$4B,$F3,$57,$D1
	.dc.b	$FF,$4A,$FC,$59,$F9,$93,$EB,$5F,$A7,$72,$C7,$8E,$1C,$3E,$41,$87
	.dc.b	$0B,$7E,$B9,$FE,$81,$FD,$D7,$C1,$83,$F7,$BF,$F4,$04,$7F,$0C,$FC
	.dc.b	$31,$F8,$8B,$F6,$47,$E6,$DC,$70,$F4,$61,$C3,$E0,$BF,$DF,$F7,$BF
	.dc.b	$8F,$D5,$75,$FB,$F7,$2F,$F8,$77,$4F,$0E,$3E,$0A,$FF,$11,$C6,$28
	.dc.b	$F8,$F0,$F1,$4F,$4A,$F0,$69,$E6,$FC,$C7,$FC,$85,$7C,$4E,$1E,$C3
	.dc.b	$F5,$EE,$19,$74,$D7,$CF,$2F,$D3,$F3,$7D,$FE,$0F,$FC,$99,$EE,$7C
	.dc.b	$6F,$9F,$FF,$7C,$7E,$44,$FC,$51,$FA,$13,$F4,$97,$E6,$5F,$CF,$BC
	.dc.b	$56,$3C,$7F,$05,$63,$C7,$0E,$17,$CE,$F7,$CB,$F8,$3C,$FB,$E7,$DB
	.dc.b	$3F,$DD,$4F,$FA,$E7,$E0,$3E,$CF,$EB,$BF,$A3,$FC,$9D,$FA,$73,$F7
	.dc.b	$D7,$EB,$DF,$DE,$BF,$45,$E1,$8F,$E3,$F0,$E1,$7B,$AF,$EF,$FF,$BB
	.dc.b	$FF,$B1,$7F,$6A,$FC,$2F,$EA,$3F,$DE,$3E,$25,$FA,$EF,$DA,$FD,$BB
	.dc.b	$F0,$7F,$19,$BD,$3F,$35,$7B,$AD,$78,$FF,$CA,$31,$E7,$F4,$5F,$E8
	.dc.b	$69,$E5,$F9,$87,$F9,$E6,$73,$9F,$6F,$21,$A7,$F9,$7F,$A1,$FA,$AF
	.dc.b	$E6,$3D,$57,$EC,$AF,$AB,$7E,$8C,$FD,$D3,$F8,$47,$E8,$BF,$A6,$C7
	.dc.b	$8F,$1F,$EB,$F1,$E3,$87,$0B,$DF,$C2,$FF,$B9,$73,$ED,$9F,$5C,$FA
	.dc.b	$7C,$37,$D2,$CF,$9E,$6C,$F9,$D6,$BC,$2F,$F4,$A7,$EE,$3F,$AE,$F2
	.dc.b	$3F,$CE,$1F,$B0,$3F,$70,$7A,$AF,$A4,$FE,$3B,$FB,$FF,$FB,$D7,$0E
	.dc.b	$1F,$82,$B0,$E1,$7B,$F8,$5F,$CE,$6E,$BD,$D6,$B1,$7C,$4F,$8B,$83
	.dc.b	$F8,$76,$BE,$2D,$7C,$35,$FD,$E6,$58,$BF,$29,$FE,$0B,$FA,$E3,$F8
	.dc.b	$03,$F4,$DF,$9C,$BE,$53,$F3,$7F,$93,$7F,$04,$C3,$86,$3F,$C6,$F0
	.dc.b	$E1,$EC,$30,$E1,$EA,$B0,$F8,$38,$70,$C1,$87,$0B,$DF,$E1,$7F,$EB
	.dc.b	$DD,$3C,$7A,$78,$B4,$FD,$5F,$C7,$FF,$B0,$7F,$69,$FE,$37,$B3,$F1
	.dc.b	$1F,$F3,$47,$E5,$CF,$49,$FB,$5F,$FB,$A7,$F7,$9F,$7E,$1C,$2F,$75
	.dc.b	$AF,$C6,$9F,$B1,$6B,$8B,$F9,$B6,$7D,$7F,$03,$69,$D7,$4F,$F7,$9F
	.dc.b	$23,$4E,$95,$AC,$FC,$3E,$7F,$FB,$5F,$F2,$EF,$E6,$3F,$19,$F6,$7F
	.dc.b	$02,$FD,$C7,$AB,$F9,$BD,$FE,$46,$6B,$DF,$83,$FE,$8F,$F7,$3F,$99
	.dc.b	$F9,$FE,$6F,$CF,$BE,$CD,$6B,$2E,$73,$FE,$A4,$C3,$97,$1F,$F4,$FD
	.dc.b	$3C,$18,$F2,$7F,$DD,$5F,$96,$FF,$44,$F0,$9F,$A6,$7C,$9F,$DB,$43
	.dc.b	$F4,$17,$E1,$5F,$DD,$73,$61,$C6,$F8,$5E,$EB,$FD,$BB,$5A,$CF,$FB
	.dc.b	$06,$3D,$B9,$3A,$DF,$5F,$3B,$C7,$5A,$C7,$9E,$6F,$D3,$3F,$E0,$07
	.dc.b	$EC,$3F,$D6,$BE,$D5,$F9,$87,$F6,$9E,$77,$82,$FA,$07,$A7,$FE,$B9
	.dc.b	$E1,$63,$C7,$0E,$17,$BA,$7E,$91,$FA,$4A,$D7,$CF,$7E,$12,$BA,$7E
	.dc.b	$99,$E2,$E0,$FE,$F5,$AF,$8A,$B5,$A3,$F0,$1F,$EF,$DA,$FC,$8B,$FF
	.dc.b	$00,$7E,$50,$FC,$77,$FC,$D5,$E9,$7F,$A9,$BF,$63,$FF,$0F,$7D,$EF
	.dc.b	$ED,$3A,$FF,$BB,$DE,$FC,$DF,$E9,$7F,$12,$BF,$F1,$27,$D6,$EF,$3E
	.dc.b	$B9,$F4,$AD,$74,$78,$B2,$E7,$39,$C3,$FA,$CD,$FF,$02,$FA,$2F,$F0
	.dc.b	$2F,$EF,$9F,$E3,$0F,$B4,$FD,$4F,$E5,$0F,$D0,$DF,$85,$7E,$7F,$8F
	.dc.b	$FC,$1B,$8F,$F8,$6F,$0E,$1E,$8B,$0E,$17,$FB,$5B,$DD,$AF,$B3,$FE
	.dc.b	$14,$3E,$3E,$5E,$2F,$81,$FF,$5E,$46,$6B,$8D,$76,$C3,$C9,$EF,$F0
	.dc.b	$AF,$15,$FF,$95,$72,$F1,$67,$FE,$42,$9F,$E3,$FC,$1F,$E2,$2C,$DF
	.dc.b	$89,$3F,$43,$7A,$1F,$CD,$BE,$F5,$F2,$3F,$9F,$C3,$85,$EE,$B5,$3F
	.dc.b	$9E,$9C,$DF,$AF,$1F,$C0,$C7,$0F,$AA,$C4,$78,$8F,$F2,$16,$5E,$2A
	.dc.b	$F0,$FC,$3F,$F2,$49,$E2,$6B,$2F,$04,$7F,$9E,$7E,$8B,$F1,$BF,$55
	.dc.b	$FE,$3C,$9F,$A1,$E7,$7E,$C7,$ED,$1F,$A4,$BF,$64,$FE,$7F,$7B,$E6
	.dc.b	$FD,$EF,$9F,$FB,$CB,$DF,$99,$AD,$7F,$CA,$D9,$FD,$BF,$C2,$3D,$3D
	.dc.b	$7F,$E4,$99,$CE,$3F,$C1,$9F,$81,$19,$7E,$6C,$76,$F0,$18,$BF,$14
	.dc.b	$F4,$D7,$2D,$FD,$1C,$1F,$A8,$FD,$67,$FC,$AF,$F7,$5F,$AC,$F4,$7F
	.dc.b	$6B,$FB,$3F,$B3,$55,$6B,$D4,$FE,$03,$E7,$7F,$00,$E2,$5B,$E7,$7F
	.dc.b	$AF,$DE,$EF,$D6,$F1,$F4,$EE,$DF,$E7,$5D,$3C,$18,$FF,$C0,$18,$E2
	.dc.b	$66,$FB,$E9,$FC,$23,$DB,$FD,$FF,$EE,$5F,$E6,$EF,$80,$7E,$B6,$FD
	.dc.b	$7F,$FB,$C7,$EE,$4F,$D6,$9F,$8A,$3E,$17,$FB,$9E,$3C,$70,$E1,$F0
	.dc.b	$CC,$38,$5B,$F2,$0D,$EF,$FE,$8B,$D3,$BF,$83,$F5,$B2,$E5,$AD,$AD
	.dc.b	$67,$D3,$2F,$17,$D7,$FA,$27,$35,$E6,$D9,$F9,$67,$F1,$AF,$DB,$BF
	.dc.b	$D1,$3D,$5F,$CB,$7D,$07,$EA,$1E,$AF,$81,$FF,$30,$E3,$FE,$0C,$BD
	.dc.b	$DF,$F7,$2A,$D7,$7B,$F4,$8D,$69,$FE,$A2,$AF,$EE,$EF,$E2,$AF,$8F
	.dc.b	$CC,$7E,$41,$E7,$9F,$83,$97,$FD,$FB,$8F,$96,$F9,$4E,$7F,$7A,$FF
	.dc.b	$4F,$F8,$5F,$E1,$E9,$FF,$58,$7C,$17,$F2,$FF,$C2,$7F,$DF,$9E,$DF
	.dc.b	$FA,$A3,$EF,$7F,$C4,$77,$FE,$A6,$AD,$7C,$0F,$F2,$0F,$C0,$7F,$D1
	.dc.b	$13,$CF,$1F,$E6,$7E,$A7,$E9,$23,$87,$C1,$7F,$35,$9F,$D1,$C6,$3E
	.dc.b	$07,$F9,$BE,$7F,$8B,$8E,$59,$76,$8F,$E2,$DE,$33,$EF,$7F,$D6,$FF
	.dc.b	$76,$F0,$1F,$A4,$FF,$B9,$FF,$4D,$FF,$1A,$7E,$DF,$FE,$E5,$C3,$86
	.dc.b	$BF,$EC,$DC,$38,$5E,$FE,$7D,$6F,$AE,$FE,$BF,$9F,$6C,$DF,$B3,$FF
	.dc.b	$16,$CF,$9E,$6C,$F9,$BF,$C7,$63,$31,$DB,$CF,$6B,$FE,$47,$F1,$7F
	.dc.b	$DB,$D5,$AC,$7D,$35,$AD,$7F,$1B,$8F,$F7,$0E,$DF,$F6,$AF,$E8,$EF
	.dc.b	$DB,$7F,$DA,$5F,$1C,$FC,$5D,$FA,$73,$F1,$47,$E8,$1C,$72,$FF,$76
	.dc.b	$63,$C7,$D4,$63,$C7,$0E,$1F,$A1,$EE,$D7,$DF,$E0,$FE,$06,$BE,$2D
	.dc.b	$7F,$DA,$35,$FD,$57,$85,$39,$CF,$AE,$DE,$4F,$A3,$F7,$CE,$75,$F0
	.dc.b	$61,$F0,$B8,$BF,$DE,$3F,$A6,$3F,$19,$FE,$CD,$F8,$F3,$F4,$D7,$E3
	.dc.b	$3E,$FF,$AF,$FF,$09,$E1,$C2,$F7,$5A,$D5,$E2,$EB,$30,$CF,$B7,$C4
	.dc.b	$7D,$BC,$79,$73,$DA,$DD,$6B,$5F,$16,$BF,$ED,$0E,$2F,$C8,$56,$B5
	.dc.b	$F0,$CB,$E4,$FE,$EF,$C1,$FF,$36,$F1,$3F,$13,$CC,$FC,$A3,$FC,$45
	.dc.b	$F4,$2F,$DC,$FF,$C5,$B7,$BF,$13,$EF,$56,$BF,$7B,$CE,$BF,$CC,$72
	.dc.b	$FC,$9B,$F8,$FB,$ED,$F5,$DF,$95,$79,$67,$DF,$E8,$5F,$A2,$79,$2F
	.dc.b	$96,$7D,$A5,$EC,$FE,$53,$E0,$FF,$B4,$7E,$90,$FD,$DD,$FB,$1B,$F4
	.dc.b	$97,$D4,$3F,$7E,$7E,$D5,$FE,$46,$C7,$8E,$1C,$2F,$75,$FB,$9E,$AF
	.dc.b	$EC,$5F,$32,$CF,$B6,$FF,$A9,$C7,$C3,$FB,$2F,$EB,$DE,$7A,$F8,$75
	.dc.b	$7E,$75,$E7,$BF,$EE,$78,$79,$EB,$5E,$4F,$EB,$BE,$87,$FC,$39,$F9
	.dc.b	$7B,$F5,$07,$C0,$3F,$4E,$7E,$8E,$FD,$8F,$DB,$C5,$FE,$6D,$D7,$FC
	.dc.b	$27,$8F,$1C,$38,$5E,$FF,$60,$FF,$32,$E7,$DB,$F4,$6F,$E0,$CF,$83
	.dc.b	$4B,$8F,$27,$C3,$A7,$82,$3C,$C6,$71,$EB,$AF,$F0,$6F,$A7,$FF,$20
	.dc.b	$7C,$9B,$D5,$FE,$D1,$F9,$0F,$DC,$F8,$3F,$F3,$AD,$6A,$73,$E0,$BF
	.dc.b	$7E,$CE,$7F,$E4,$2F,$9C,$FE,$35,$7B,$F9,$56,$5F,$86,$AD,$9F,$E5
	.dc.b	$3F,$AE,$FE,$BF,$97,$83,$F4,$EF,$E3,$3F,$4F,$FE,$68,$FD,$F1,$FA
	.dc.b	$77,$F7,$7E,$67,$EB,$8F,$C8,$BF,$D1,$37,$BA,$D4,$BF,$76,$CE,$9F
	.dc.b	$E7,$6C,$79,$F0,$32,$7E,$89,$E0,$AD,$63,$DF,$C8,$FE,$57,$F7,$BF
	.dc.b	$E9,$4F,$4B,$F5,$FE,$A7,$FA,$3F,$C2,$FF,$4E,$E1,$FE,$94,$F5,$9F
	.dc.b	$7B,$FD,$23,$D2,$FD,$7F,$C8,$7F,$A7,$B2,$E5,$1F,$CB,$FF,$15,$EB
	.dc.b	$87,$1B,$E1,$EA,$BF,$48,$FE,$77,$F8,$27,$FA,$FF,$F0,$3F,$F3,$4F
	.dc.b	$AB,$FA,$1F,$47,$FD,$53,$39,$F5,$3F,$2D,$C6,$F3,$33,$A7,$FA,$F7
	.dc.b	$1E,$3E,$87,$EB,$E3,$23,$F0,$4F,$F6,$36,$1F,$9F,$E8,$FE,$97,$F5
	.dc.b	$5F,$ED,$FF,$C1,$7F,$AA,$7C,$37,$F6,$4F,$43,$FD,$73,$39,$F9,$7F
	.dc.b	$EC,$3D,$D8,$CF,$FD,$C5,$F5,$BF,$EE,$B9,$FE,$D9,$F8,$17,$49,$CF
	.dc.b	$82,$CF,$AF,$BD,$FD,$6B,$F0,$9F,$F7,$EF,$C1,$7E,$57,$CB,$FF,$0D
	.dc.b	$78,$5F,$EC,$89,$CC,$7E,$67,$EA,$7E,$44,$A7,$32,$FA,$BE,$99,$FF
	.dc.b	$38,$FC,$29,$F8,$63,$D8,$67,$DF,$DB,$FE,$D7,$F2,$7F,$CF,$B0,$7E
	.dc.b	$1E,$FC,$39,$93,$F0,$F7,$E3,$CF,$D1,$FF,$59,$FA,$67,$FB,$7B,$2E
	.dc.b	$58,$F1,$C3,$85,$EF,$81,$FB,$46,$F4,$FE,$9B,$7B,$CF,$AE,$7F,$B7
	.dc.b	$9D,$1E,$4D,$3C,$73,$AF,$CB,$7F,$C0,$7D,$7E,$9F,$C2,$9F,$8C,$BF
	.dc.b	$19,$FC,$07,$E1,$0F,$D6,$DF,$B8,$3E,$BF,$FD,$EB,$8F,$1C,$38,$79
	.dc.b	$2C,$38,$5A,$B7,$B7,$F6,$4C,$FB,$E7,$DB,$3F,$F7,$B4,$E8,$4E,$B1
	.dc.b	$E9,$3F,$D0,$A5,$DA,$FC,$7F,$E9,$7E,$3C,$FC,$7D,$F8,$C3,$F1,$F7
	.dc.b	$E4,$1F,$D8,$FE,$57,$38,$F1,$CB,$FB,$BE,$3C,$70,$E1,$7B,$FC,$73
	.dc.b	$BE,$7D,$B3,$EB,$9F,$4F,$81,$FE,$0F,$96,$8E,$73,$AC,$3F,$B9,$7A
	.dc.b	$DF,$E3,$2F,$C2,$BF,$11,$DF,$FE,$08,$FC,$4D,$F9,$23,$F2,$47,$EB
	.dc.b	$DF,$D1,$FF,$20,$F7,$E1,$C3,$EB,$7F,$DC,$77,$BA,$FE,$99,$F5,$BA
	.dc.b	$DB,$FC,$23,$9F,$5A,$D7,$9D,$E1,$9C,$E1,$CB,$3E,$FF,$49,$FE,$46
	.dc.b	$FC,$B3,$F3,$9E,$FB,$F0,$37,$E5,$8F,$CA,$FF,$69,$F9,$5B,$F2,$0F
	.dc.b	$F5,$27,$73,$1E,$38,$70,$BD,$D7,$F9,$0A,$B5,$AF,$F8,$9B,$2E,$D9
	.dc.b	$75,$F6,$99,$FF,$1F,$AF,$BF,$F8,$57,$EB,$FF,$2D,$FF,$31,$79,$6F
	.dc.b	$C0,$5F,$97,$7F,$41,$F9,$A7,$5F,$0C,$7F,$C6,$78,$70,$BD,$FC,$DA
	.dc.b	$DD,$5B,$B5,$EF,$1F,$E9,$9E,$D7,$F9,$AB,$F2,$FF,$8B,$D3,$F9,$1F
	.dc.b	$C3,$9F,$A2,$BF,$69,$7E,$5B,$FE,$E0,$E7,$61,$C3,$C9,$FF,$67,$DE
	.dc.b	$EB,$F5,$D5,$AD,$7E,$26,$5F,$95,$7F,$CC,$18,$78,$39,$BF,$D0,$8F
	.dc.b	$3C,$FC,$39,$FF,$91,$E3,$19,$78,$FF,$29,$F4,$61,$D3,$4F,$0D,$FF
	.dc.b	$9E,$FF,$51,$FD,$5E,$1C,$23,$FC,$FD,$ED,$E2,$F8,$65,$DB,$EE,$79
	.dc.b	$C3,$B3,$FC,$C3,$97,$4F,$BF,$F8,$19,$73,$7D,$BE,$1F,$FB,$63,$E6
	.dc.b	$9F,$9C,$BF,$37,$FF,$12,$FE,$72,$FC,$E1,$FA,$37,$EF,$B0,$E1,$7B
	.dc.b	$FB,$35,$EF,$C9,$D6,$FF,$E7,$0D,$BF,$C5,$BF,$31,$FE,$FC,$FC,$C5
	.dc.b	$FB,$3F,$F1,$BF,$B9,$DD,$FF,$09,$FB,$8F,$EF,$FC,$CB,$DF,$BF,$FE
	.dc.b	$A8,$DF,$F9,$F6,$7D,$30,$FF,$58,$46,$30,$FF,$53,$46,$0F,$13,$D2
	.dc.b	$FC,$3F,$03,$CE,$73,$A7,$97,$4F,$E7,$1E,$07,$F0,$19,$CE,$9E,$39
	.dc.b	$FE,$65,$F4,$7E,$8D,$3D,$B3,$9D,$7E,$56,$0F,$C5,$FF,$AA,$7C,$B3
	.dc.b	$F6,$97,$B4,$FC,$85,$F7,$9F,$EC,$EF,$C5,$DF,$94,$FF,$2C,$C7,$8E
	.dc.b	$1C,$2F,$7F,$58,$B7,$EB,$BB,$E2,$FF,$4F,$E9,$DF,$4F,$F6,$C6,$5C
	.dc.b	$BF,$2C,$B5,$AC,$FA,$6B,$FE,$DE,$F7,$7F,$DA,$B7,$E0,$C3,$CD,$3F
	.dc.b	$EE,$2F,$AB,$CE,$1C,$67,$3A,$F8,$FE,$EF,$C5,$2D,$3C,$5B,$FA,$37
	.dc.b	$7E,$7B,$FD,$23,$F3,$3F,$EF,$1F,$8E,$3F,$0A,$FF,$3A,$FC,$D3,$F3
	.dc.b	$FF,$E2,$3E,$17,$FB,$BF,$A1,$87,$0B,$DD,$6B,$F1,$4E,$9B,$7F,$B0
	.dc.b	$32,$E9,$E3,$7F,$BE,$70,$E1,$FA,$F5,$AD,$63,$CB,$E0,$FF,$C1,$59
	.dc.b	$F6,$9C,$E1,$FE,$0B,$CF,$F9,$1F,$D2,$FF,$BE,$B2,$F1,$4E,$70,$FF
	.dc.b	$45,$6F,$E2,$C1,$A7,$8B,$7F,$65,$7F,$A3,$F8,$F5,$39,$CF,$DF,$FA
	.dc.b	$26,$36,$F2,$70,$FF,$96,$37,$F2,$67,$FB,$8F,$D0,$7F,$17,$E7,$BF
	.dc.b	$01,$FF,$59,$FE,$AE,$F9,$1F,$17,$A0,$BD,$D6,$B9,$33,$1F,$94,$C7
	.dc.b	$FC,$54,$7B,$4F,$F7,$4D,$71,$F0,$BE,$44,$78,$7B,$7F,$84,$61,$FE
	.dc.b	$C2,$F7,$38,$4E,$74,$F2,$7E,$1E,$FA,$2C,$E5,$E3,$C7,$8C,$BF,$58
	.dc.b	$FF,$66,$70,$7B,$AB,$58,$72,$9F,$E9,$B3,$F0,$BF,$16,$7E,$32,$FD
	.dc.b	$C5,$F0,$DF,$D0,$3F,$12,$FE,$D1,$F8,$3F,$D3,$BD,$D6,$BE,$21,$5F
	.dc.b	$DB,$7E,$37,$58,$FF,$92,$78,$7E,$19,$8B,$F8,$38,$72,$CB,$B1,$F0
	.dc.b	$1C,$63,$D3,$E1,$F1,$87,$96,$BC,$9F,$2B,$FE,$51,$D7,$D5,$39,$BF
	.dc.b	$1C,$FF,$AB,$27,$ED,$64,$1F,$A7,$3F,$5A,$7E,$A1,$FD,$7B,$ED,$DF
	.dc.b	$8F,$BF,$75,$FF,$18,$63,$CB,$1E,$38,$70,$CD,$87,$0B,$7E,$CB,$FE
	.dc.b	$3C,$F0,$3F,$CD,$9A,$77,$FD,$4F,$E1,$E5,$93,$C9,$1E,$3D,$73,$A7
	.dc.b	$78,$C6,$7D,$B5,$FC,$B5,$F1,$3C,$77,$E9,$8F,$F6,$95,$F1,$D7,$CD
	.dc.b	$AF,$97,$F3,$0E,$19,$74,$8F,$47,$F0,$06,$2B,$78,$7C,$1F,$BA,$C6
	.dc.b	$5D,$78,$3F,$A6,$67,$F0,$F6,$FF,$69,$7E,$28,$FC,$77,$FD,$1F,$FB
	.dc.b	$AB,$F6,$1F,$F0,$3D,$EF,$B5,$7B,$FB,$F5,$EF,$E3,$3D,$99,$7F,$B7
	.dc.b	$7F,$1E,$F8,$7C,$FF,$E9,$58,$EF,$CD,$F8,$CC,$FA,$65,$CF,$1F,$F0
	.dc.b	$CC,$F2,$CF,$BE,$BE,$4F,$D1,$BF,$05,$87,$08,$F3,$FE,$14,$F3,$F8
	.dc.b	$65,$CE,$37,$A7,$FA,$97,$CF,$FF,$58,$7E,$55,$FE,$49,$FC,$95,$F8
	.dc.b	$53,$F7,$B7,$EF,$BF,$56,$F7,$F5,$AB,$DD,$3F,$6A,$D7,$E1,$3F,$EB
	.dc.b	$5E,$3F,$66,$9D,$E1,$A7,$C2,$9F,$CD,$65,$D3,$1F,$0F,$A9,$E6,$9C
	.dc.b	$DF,$0C,$FC,$1A,$BF,$30,$D5,$31,$E5,$3E,$9F,$9F,$F1,$E7,$ED,$8F
	.dc.b	$67,$57,$E3,$FE,$41,$FB,$0B,$F5,$E7,$E4,$7F,$EC,$7F,$23,$FC,$D9
	.dc.b	$E2,$EE,$B5,$39,$DD,$F7,$F9,$C3,$D9,$BF,$F7,$8B,$FF,$14,$3F,$08
	.dc.b	$FD,$C6,$5E,$09,$EF,$F0,$BF,$BA,$F8,$FF,$5C,$E7,$2F,$17,$7E,$31
	.dc.b	$F0,$DF,$F0,$0C,$FF,$C6,$FD,$FF,$BB,$C3,$84,$E6,$7C,$98,$79,$FF
	.dc.b	$0B,$B3,$C7,$F8,$07,$F6,$1F,$1A,$F9,$F7,$E7,$8F,$C9,$1F,$99,$BF
	.dc.b	$3D,$FF,$06,$FB,$BC,$2F,$75,$A9,$FF,$08,$F9,$39,$C3,$E1,$E3,$D7
	.dc.b	$B6,$E1,$C3,$2E,$9C,$9F,$6A,$BC,$57,$E1,$9F,$07,$26,$6F,$86,$5F
	.dc.b	$E1,$EF,$D4,$3B,$69,$E1,$F9,$C7,$C9,$FF,$0B,$F7,$5F,$AB,$3F,$80
	.dc.b	$7F,$3B,$F1,$4F,$D5,$9F,$B4,$FF,$AF,$7B,$3F,$92,$61,$C2,$F7,$5A
	.dc.b	$F1,$BA,$62,$FC,$AF,$EF,$DE,$FC,$1F,$BC,$7F,$50,$9F,$07,$8D,$FE
	.dc.b	$92,$C3,$FD,$D3,$3C,$B1,$EB,$1F,$9A,$BF,$05,$6F,$37,$EE,$1B,$8E
	.dc.b	$5F,$2B,$FA,$0F,$C1,$1F,$61,$FE,$70,$FD,$65,$EA,$BF,$6B,$FE,$77
	.dc.b	$FA,$27,$F2,$DB,$DF,$D8,$AF,$75,$FD,$03,$DE,$D6,$2F,$DF,$7F,$37
	.dc.b	$87,$0E,$2F,$6B,$CB,$0E,$3F,$99,$F8,$EF,$C5,$A7,$87,$F0,$2C,$69
	.dc.b	$DE,$3C,$9F,$12,$FD,$03,$BC,$E7,$6F,$EA,$3F,$1D,$8C,$FC,$B6,$FD
	.dc.b	$FF,$18,$F2,$FC,$13,$FA,$47,$E0,$2F,$0D,$F8,$93,$E7,$BF,$CC,$9F
	.dc.b	$22,$FC,$07,$EF,$F8,$7F,$D8,$70,$E1,$7B,$FA,$A5,$EF,$AF,$AE,$4F
	.dc.b	$C4,$FF,$7B,$87,$0E,$9F,$BE,$F2,$D3,$BF,$E6,$FF,$C5,$23,$19,$74
	.dc.b	$E9,$C6,$3E,$6B,$F2,$ED,$FE,$CE,$F8,$6F,$EF,$99,$31,$E3,$AF,$A7
	.dc.b	$F4,$BF,$EC,$DF,$6E,$FD,$55,$F9,$0F,$F9,$8B,$F7,$CF,$F3,$5F,$EE
	.dc.b	$1F,$E0,$4E,$5F,$EF,$D7,$BA,$D7,$D8,$2B,$FA,$BD,$6B,$07,$E8,$5F
	.dc.b	$C6,$B0,$E1,$AF,$E7,$1E,$5F,$58,$CF,$FC,$F3,$F1,$9F,$CC,$A7,$31
	.dc.b	$FE,$7C,$F8,$EC,$63,$E6,$BE,$1A,$78,$7E,$BF,$D7,$39,$BF,$4F,$ED
	.dc.b	$DF,$E0,$3F,$C0,$DF,$8F,$7F,$A0,$E5,$F0,$6F,$D2,$BC,$FD,$1B,$AD
	.dc.b	$4E,$78,$7F,$C5,$9D,$1F,$33,$1E,$7D,$BF,$CD,$9E,$BF,$65,$70,$E3
	.dc.b	$97,$FB,$02,$F8,$46,$3E,$C7,$FA,$7D,$F0,$D3,$C7,$7F,$23,$F3,$9F
	.dc.b	$16,$3F,$17,$4F,$17,$CB,$E2,$F7,$AB,$ED,$B1,$5E,$7F,$D9,$7F,$C4
	.dc.b	$5E,$F7,$FB,$0B,$F4,$57,$ED,$EF,$DE,$1F,$90,$FF,$AA,$3F,$6A,$DF
	.dc.b	$E0,$4B,$DD,$6B,$AD,$FA,$06,$A7,$FA,$AE,$5F,$EC,$EF,$1A,$E1,$C7
	.dc.b	$3E,$D9,$FF,$9A,$3C,$2E,$11,$8F,$B3,$FE,$DF,$87,$96,$37,$9B,$F6
	.dc.b	$AF,$F1,$D7,$5B,$D8,$7E,$FE,$FC,$03,$F5,$BF,$B7,$3F,$61,$7E,$ED
	.dc.b	$C2,$F7,$5A,$9F,$EF,$DE,$5F,$F9,$5F,$1E,$9C,$1F,$DB,$9E,$D9,$5C
	.dc.b	$38,$F1,$33,$ED,$C3,$D6,$BC,$53,$9F,$CD,$BF,$C1,$38,$72,$D3,$C1
	.dc.b	$F5,$FF,$E5,$3F,$88,$7B,$EF,$D9,$1F,$96,$BF,$27,$FE,$4D,$F8,$6B
	.dc.b	$F0,$8E,$17,$BA,$D7,$C5,$7F,$C0,$1F,$8D,$CF,$E4,$1F,$E0,$6C,$FA
	.dc.b	$E7,$70,$E1,$9B,$F7,$8F,$F6,$8E,$FF,$C1,$2F,$79,$F4,$FC,$97,$FC
	.dc.b	$43,$18,$BF,$1F,$C6,$FA,$7F,$55,$FF,$42,$FE,$D6,$FC,$D3,$F0,$7F
	.dc.b	$BF,$3F,$0E,$D7,$D9,$AB,$5E,$4A,$9F,$AF,$6B,$B7,$FA,$EE,$AC,$DF
	.dc.b	$B2,$7A,$DE,$FA,$7E,$43,$E7,$FC,$B1,$8C,$9F,$4B,$FE,$38,$F8,$17
	.dc.b	$E1,$5F,$ED,$5F,$82,$F9,$BA,$7F,$A6,$FE,$9F,$FD,$09,$5A,$E7,$56
	.dc.b	$BF,$06,$7E,$89,$AC,$7F,$B3,$E3,$FD,$CB,$0E,$0D,$8E,$DD,$3F,$EA
	.dc.b	$9F,$9B,$F1,$46,$25,$F2,$18,$CF,$BD,$FF,$AB,$F2,$7C,$E7,$ED,$BC
	.dc.b	$D7,$EE,$0F,$8B,$7E,$73,$FC,$5B,$EC,$3F,$D5,$F9,$7F,$3C,$F5,$D8
	.dc.b	$70,$BD,$FB,$CB,$7D,$97,$5F,$DB,$3F,$6F,$7D,$B8,$3F,$C1,$B8,$F2
	.dc.b	$D3,$C1,$A7,$F2,$49,$FD,$AD,$75,$8F,$E4,$FF,$55,$8A,$DE,$7D,$FF
	.dc.b	$0F,$F5,$46,$34,$7F,$00,$E3,$6F,$2C,$3F,$36,$FF,$5B,$7E,$09,$FE
	.dc.b	$10,$FD,$5D,$F8,$B7,$E9,$BE,$8B,$F4,$FF,$A4,$EB,$CC,$56,$BD,$05
	.dc.b	$7E,$B3,$F1,$5D,$7E,$B1,$FE,$68,$CB,$A7,$2F,$F8,$A9,$ED,$9F,$5F
	.dc.b	$AF,$FF,$3F,$C6,$31,$E7,$F0,$F8,$D3,$BD,$F8,$E7,$FC,$F9,$CF,$FE
	.dc.b	$DF,$BE,$1A,$7E,$F3,$F2,$1F,$F6,$87,$E4,$AF,$88,$7E,$13,$FE,$A0
	.dc.b	$FD,$95,$FA,$5B,$F4,$17,$A1,$C2,$F7,$9A,$F7,$4F,$DC,$35,$F8,$A7
	.dc.b	$FA,$13,$3E,$D9,$DC,$78,$E1,$DB,$F0,$6F,$F4,$DD,$6B,$2E,$9A,$7F
	.dc.b	$B9,$B7,$FF,$7C,$4E,$6F,$BF,$E3,$BE,$3C,$BD,$17,$E7,$F8,$0F,$F7
	.dc.b	$66,$8F,$29,$F9,$F7,$F1,$9F,$4D,$FA,$A7,$FB,$63,$F7,$ED,$D6,$BE
	.dc.b	$07,$FD,$03,$3F,$ED,$CF,$D0,$F3,$FA,$D7,$FA,$7B,$2E,$99,$5C,$38
	.dc.b	$65,$D3,$EA,$FF,$D6,$B5,$AC,$B9,$FD,$FE,$2F,$79,$FF,$BA,$3F,$48
	.dc.b	$F0,$E5,$E5,$8F,$F7,$5F,$D8,$E3,$0F,$1F,$C7,$FF,$98,$39,$9E,$AF
	.dc.b	$F8,$A3,$4E,$DB,$7B,$2B,$F9,$A5,$6B,$C5,$0F,$D4,$BF,$9E,$FD,$67
	.dc.b	$C5,$FD,$07,$FD,$BF,$F9,$E3,$F6,$CF,$DB,$E9,$FE,$34,$BD,$F1,$2F
	.dc.b	$7F,$BC,$6B,$2E,$B5,$D1,$FF,$41,$BF,$D2,$BF,$3D,$FF,$62,$BC,$AB
	.dc.b	$FD,$DB,$F1,$7F,$ED,$A8,$C7,$82,$FC,$F7,$1E,$57,$F9,$67,$3F,$04
	.dc.b	$B6,$7E,$BB,$E1,$D3,$D9,$7B,$CF,$D7,$FB,$E6,$7F,$08,$FF,$45,$FB
	.dc.b	$6F,$A9,$70,$FE,$D3,$F0,$F7,$EC,$7B,$F9,$25,$EE,$B5,$FA,$AE,$63
	.dc.b	$FD,$BF,$8F,$FA,$5B,$E1,$AD,$F0,$CB,$AF,$A8,$F5,$FF,$DE,$F7,$E2
	.dc.b	$9C,$FE,$81,$8B,$E1,$9F,$8F,$F5,$AE,$D5,$CB,$67,$E7,$1E,$FF,$D6
	.dc.b	$5F,$8D,$B3,$FE,$65,$9B,$EE,$9F,$B5,$BF,$37,$E1,$7B,$F8,$75,$EE
	.dc.b	$9F,$B1,$6B,$F7,$CF,$F7,$DE,$7D,$B3,$B8,$F1,$C3,$B7,$B7,$F8,$CF
	.dc.b	$23,$F1,$DD,$AC,$38,$69,$DB,$6F,$DD,$78,$DF,$E7,$E9,$CE,$3E,$1F
	.dc.b	$C4,$7C,$BA,$7A,$AF,$D3,$FA,$36,$31,$E5,$C0,$FC,$13,$18,$FA,$7E
	.dc.b	$83,$F0,$5F,$B5,$3F,$78,$7E,$C9,$FE,$AF,$87,$B2,$FD,$8F,$F9,$E7
	.dc.b	$1F,$A2,$78,$56,$EB,$5E,$17,$4C,$1F,$81,$FF,$A5,$B9,$7F,$C2,$A7
	.dc.b	$EB,$30,$E2,$FF,$B8,$9E,$7F,$8C,$1F,$0F,$EA,$3F,$1F,$9F,$FD,$3F
	.dc.b	$1A,$3C,$BF,$A1,$E6,$FA,$63,$EA,$C3,$96,$7E,$9F,$DE,$3C,$9A,$F9
	.dc.b	$F7,$7E,$45,$FA,$3F,$C9,$1F,$2D,$FE,$88,$F8,$87,$E9,$0E,$A7,$EE
	.dc.b	$0F,$CE,$38,$5E,$EB,$52,$FD,$3B,$3F,$0B,$F0,$78,$BC,$C3,$3E,$FF
	.dc.b	$3F,$98,$E6,$FF,$4C,$C3,$C1,$FA,$AF,$0A,$DE,$BE,$4D,$5F,$53,$9D
	.dc.b	$7C,$51,$E6,$FC,$63,$9D,$6F,$6F,$F6,$46,$9F,$C7,$3F,$49,$F8,$3E
	.dc.b	$A7,$CF,$F9,$87,$EC,$2F,$DE,$1F,$A8,$FF,$07,$BD,$F7,$7F,$5C,$AD
	.dc.b	$7E,$6B,$9E,$8F,$DC,$67,$EF,$F1,$3F,$3B,$87,$68,$C7,$83,$FE,$D8
	.dc.b	$8E,$7A,$78,$3E,$A3,$E2,$C6,$32,$7E,$33,$8C,$BF,$C0,$7A,$78,$76
	.dc.b	$7E,$C7,$E1,$E0,$F5,$61,$E9,$FD,$BF,$17,$BF,$C2,$2E,$BE,$EC,$7F
	.dc.b	$D4,$BC,$1F,$B2,$F0,$FF,$C3,$1F,$9C,$FE,$13,$F5,$7F,$EC,$5F,$92
	.dc.b	$7F,$1C,$AD,$7D,$62,$B5,$F1,$8A,$FD,$2F,$2F,$FB,$D3,$1E,$7C,$DF
	.dc.b	$00,$FE,$96,$FF,$89,$7F,$3A,$FC,$06,$5D,$31,$E7,$E6,$7E,$1A,$BC
	.dc.b	$7C,$AD,$7F,$BA,$7E,$6B,$9D,$BC,$53,$9D,$BC,$5E,$5F,$F3,$DC,$38
	.dc.b	$E1,$EA,$AF,$F2,$5F,$DC,$FF,$55,$FA,$B7,$E3,$AF,$77,$FD,$8F,$F8
	.dc.b	$0B,$F6,$87,$E4,$DC,$39,$3F,$AB,$E1,$C2,$F7,$5F,$E8,$5E,$36,$5D
	.dc.b	$72,$F6,$BD,$FF,$1F,$B8,$F2,$9F,$07,$E2,$9F,$E8,$9F,$07,$F9,$25
	.dc.b	$F1,$CB,$C7,$E1,$E3,$0E,$11,$E0,$FA,$7F,$24,$7F,$75,$D7,$C9,$8F
	.dc.b	$FA,$4F,$57,$D3,$EB,$C0,$7E,$3D,$FE,$74,$F8,$2F,$F1,$D6,$CF,$DB
	.dc.b	$1F,$9D,$6E,$B5,$F6,$1F,$92,$97,$E6,$D9,$FA,$5F,$5E,$FF,$C5,$EF
	.dc.b	$FC,$FE,$FE,$87,$8F,$F8,$BC,$72,$C9,$F8,$6F,$AE,$B8,$7E,$03,$FD
	.dc.b	$47,$37,$D3,$7F,$84,$27,$C5,$39,$BF,$37,$EE,$38,$DF,$D3,$B3,$F2
	.dc.b	$3C,$65,$E7,$FC,$46,$E1,$ED,$FB,$2F,$8F,$F2,$9D,$BF,$66,$7E,$70
	.dc.b	$FD,$A7,$FD,$6B,$E0,$7F,$96,$7C,$1F,$B7,$C3,$FE,$22,$C3,$85,$EF
	.dc.b	$E3,$96,$FD,$D7,$AF,$23,$3F,$8B,$87,$B3,$F6,$8B,$39,$C7,$97,$E1
	.dc.b	$7F,$0E,$BA,$EB,$FE,$C6,$DF,$17,$B8,$F2,$FE,$15,$DB,$2F,$4C,$7A
	.dc.b	$3F,$08,$FF,$53,$FE,$35,$F2,$7F,$3D,$FE,$B9,$C5,$FD,$6B,$BB,$FC
	.dc.b	$EB,$FA,$57,$EE,$61,$5B,$F9,$3F,$F5,$7F,$A0,$BD,$FE,$25,$D3,$9F
	.dc.b	$FC,$87,$97,$F6,$7F,$D6,$3F,$73,$8F,$1C,$AD,$6B,$2E,$78,$FF,$A1
	.dc.b	$3F,$30,$FF,$00,$5E,$F3,$ED,$F8,$0F,$EC,$70,$EC,$33,$F4,$CE,$B5
	.dc.b	$7E,$F7,$A9,$CE,$5E,$CF,$D4,$BF,$D8,$FF,$A3,$7E,$27,$F1,$6F,$EA
	.dc.b	$5C,$4F,$C7,$9F,$69,$FD,$FF,$EB,$9F,$80,$70,$DF,$FC,$BF,$87,$0B
	.dc.b	$DF,$1A,$DF,$81,$6E,$FF,$DD,$7F,$39,$FE,$CF,$7F,$BF,$73,$35,$F1
	.dc.b	$4B,$AF,$F8,$FC,$63,$6F,$26,$DF,$E1,$1F,$C8,$FF,$C4,$93,$9D,$7C
	.dc.b	$3F,$B4,$63,$3F,$2E,$5F,$EE,$A9,$F2,$EB,$FE,$15,$D3,$C1,$5A,$CF
	.dc.b	$AE,$DE,$9F,$DA,$BD,$1C,$1E,$DD,$BE,$17,$91,$8D,$3A,$47,$C5,$FC
	.dc.b	$2B,$DD,$F5,$8F,$6F,$FD,$B3,$FA,$AF,$F9,$17,$F2,$9F,$EC,$BF,$0B
	.dc.b	$A9,$CF,$D9,$A7,$FB,$4F,$BB,$FE,$FA,$F5,$3F,$79,$FA,$0F,$CB,$FC
	.dc.b	$23,$FC,$81,$F8,$4F,$FB,$2B,$F7,$37,$E7,$BB,$E7,$FF,$48,$DE,$FE
	.dc.b	$D9,$7B,$FD,$77,$54,$AF,$AD,$F9,$37,$F4,$CC,$FB,$F6,$34,$C5,$F4
	.dc.b	$8C,$65,$D3,$E8,$DE,$77,$F9,$57,$0E,$38,$79,$A5,$F9,$FE,$71,$E5
	.dc.b	$39,$DF,$FB,$BF,$7F,$F2,$B9,$CE,$5D,$67,$FC,$87,$3F,$C8,$A7,$FD
	.dc.b	$5B,$EE,$3F,$6B,$FE,$BD,$F5,$2F,$D8,$5F,$8F,$BE,$8F,$FD,$DF,$8F
	.dc.b	$1C,$58,$F1,$EB,$FF,$11,$E1,$C2,$FF,$3B,$7F,$E7,$DB,$FF,$5D,$FD
	.dc.b	$53,$F0,$97,$EC,$9F,$D0,$3F,$0C,$7E,$75,$FE,$0C,$FB,$26,$3C,$71
	.dc.b	$FF,$21,$61,$C3,$E4,$7F,$D5,$97,$FE,$12,$FC,$D7,$7F,$96,$FF,$39
	.dc.b	$F8,$FF,$D3,$FD,$2F,$D8,$F4,$DF,$B2,$BF,$4F,$7E,$87,$C3,$5F,$F6
	.dc.b	$7E,$1C,$3C,$5F,$D5,$5E,$FE,$77,$54,$AD,$63,$D3,$F3,$FF,$F4,$66
	.dc.b	$9E,$1F,$6F,$FD,$BF,$97,$2F,$D0,$FF,$D2,$1F,$83,$7F,$4D,$FA,$BF
	.dc.b	$F7,$EF,$E8,$8F,$D9,$9F,$54,$FD,$F7,$FC,$03,$E0,$7C,$2C,$38,$7D
	.dc.b	$FB,$0E,$17,$BF,$A8,$7E,$47,$BE,$3F,$F7,$4E,$9D,$B4,$EB,$A3,$F1
	.dc.b	$CF,$1D,$6B,$3E,$95,$AF,$A1,$FF,$53,$7A,$DF,$D8,$FD,$57,$E6,$8F
	.dc.b	$D1,$1F,$98,$BF,$30,$7E,$4A,$FC,$7F,$FC,$F1,$E3,$FF,$80,$B6,$FF
	.dc.b	$7D,$E5,$CB,$1E,$38,$70,$B7,$E2,$5F,$80,$D3,$C1,$A7,$7D,$3B,$69
	.dc.b	$F9,$9D,$3A,$68,$D3,$A5,$6B,$83,$FD,$7D,$F9,$2F,$F7,$BF,$DA,$9F
	.dc.b	$8E,$BF,$51,$7E,$57,$FD,$3F,$F7,$0F,$F1,$7E,$7F,$90,$C3,$85,$EE
	.dc.b	$B5,$E1,$3F,$1F,$D5,$86,$5D,$B2,$EB,$F3,$0C,$FE,$85,$70,$37,$97
	.dc.b	$59,$D7,$3F,$FB,$57,$DF,$FF,$09,$61,$FD,$1B,$C9,$7E,$B8,$FD,$73
	.dc.b	$FD,$6D,$E3,$7F,$31,$BD,$FC,$CA,$F7,$5A,$F9,$3F,$D5,$D6,$A7,$3F
	.dc.b	$6A,$C3,$C1,$B7,$AF,$3F,$06,$7D,$E9,$E6,$7F,$BA,$7F,$3C,$FF,$13
	.dc.b	$61,$FE,$3E,$FD,$39,$ED,$7E,$1B,$F7,$4F,$F7,$F7,$E0,$5F,$BA,$BD
	.dc.b	$F8,$2B,$DF,$82,$B7,$C9,$BF,$33,$E1,$C1,$F3,$74,$ED,$A7,$5F,$CD
	.dc.b	$3C,$53,$9C,$FA,$4F,$87,$F1,$BF,$F7,$F7,$CD,$FF,$8E,$FC,$37,$E9
	.dc.b	$9F,$EC,$BF,$C2,$1F,$92,$BF,$5A,$FF,$2B,$FE,$AD,$FC,$43,$0E,$17
	.dc.b	$BA,$D6,$7F,$EC,$9F,$29,$5A,$F8,$8F,$F4,$A5,$73,$FC,$BF,$C3,$BF
	.dc.b	$E1,$75,$F0,$D6,$B4,$FC,$1E,$2F,$CC,$3F,$CA,$1B,$BF,$6F,$FE,$C1
	.dc.b	$FA,$A7,$C4,$FC,$C1,$FA,$3F,$FB,$43,$0E,$1A,$7E,$47,$0E,$1F,$42
	.dc.b	$C3,$85,$BF,$67,$FE,$57,$9F,$6C,$FA,$F5,$34,$7D,$67,$E6,$F5,$EB
	.dc.b	$AB,$5E,$BA,$78,$FF,$4A,$D7,$C9,$FF,$99,$FE,$BD,$F9,$A7,$2F,$DC
	.dc.b	$9F,$97,$7F,$05,$E7,$7C,$8F,$F3,$CC,$38,$5E,$EB,$5E,$B6,$75,$FD
	.dc.b	$66,$5D,$37,$67,$D3,$DF,$69,$E2,$BD,$E5,$CE,$7E,$6F,$E9,$5F,$E7
	.dc.b	$BF,$C1,$BF,$CD,$FF,$89,$7F,$A1,$F6,$7E,$39,$FE,$AE,$E4,$5E,$FF
	.dc.b	$05,$FF,$5A,$FA,$1F,$E9,$8B,$DF,$EB,$9F,$D5,$2B,$5F,$84,$FF,$B2
	.dc.b	$9E,$19,$5D,$BF,$82,$69,$E0,$C7,$97,$E3,$FF,$36,$F7,$1D,$7E,$AB
	.dc.b	$E8,$79,$9F,$E9,$19,$7E,$9E,$FD,$D7,$FA,$07,$EE,$AF,$C0,$1F,$A8
	.dc.b	$7F,$9B,$B0,$E1,$7B,$F9,$65,$EE,$BF,$CA,$5C,$8B,$DE,$3F,$EE,$27
	.dc.b	$8F,$EF,$F7,$83,$F8,$E6,$9E,$2A,$F0,$E8,$FC,$0F,$CF,$5A,$D7,$C1
	.dc.b	$A7,$97,$F0,$6F,$C5,$7D,$AF,$FA,$AF,$F5,$8F,$EE,$3E,$8F,$EA,$BF
	.dc.b	$2C,$FF,$60,$79,$BF,$DF,$AB,$5F,$47,$FB,$4F,$47,$FD,$93,$5F,$EC
	.dc.b	$5F,$AF,$D6,$BF,$CB,$F2,$FE,$F7,$E8,$7F,$BE,$B1,$E3,$9D,$AD,$65
	.dc.b	$CF,$D6,$65,$FD,$26,$F7,$9F,$F0,$0B,$FD,$07,$EC,$5F,$EC,$1F,$C3
	.dc.b	$DF,$B9,$7E,$AB,$7F,$D1,$7D,$C7,$F8,$52,$B5,$39,$F4,$BF,$D9,$BD
	.dc.b	$69,$CF,$A9,$FE,$CB,$CB,$FD,$99,$39,$7F,$D7,$96,$FC,$EB,$51,$E2
	.dc.b	$33,$5F,$B5,$F5,$FF,$75,$8E,$65,$A7,$9A,$B5,$A7,$DD,$9F,$F1,$4F
	.dc.b	$C6,$7F,$B6,$FF,$38,$7D,$37,$FA,$2B,$E5,$3F,$C6,$5F,$86,$FF,$98
	.dc.b	$3F,$6A,$DE,$1F,$D6,$EF,$7F,$55,$FD,$67,$EC,$BF,$D6,$FF,$7E,$FC
	.dc.b	$07,$76,$BD,$BE,$AE,$46,$BE,$1D,$5F,$3D,$EE,$AE,$37,$DB,$6F,$1F
	.dc.b	$D0,$7F,$11,$C3,$86,$BE,$1D,$FF,$C1,$B8,$7F,$91,$3F,$0C,$FF,$7B
	.dc.b	$7D,$CF,$F1,$9F,$DB,$DF,$8D,$FF,$9E,$BE,$6F,$FC,$2D,$FA,$0E,$F0
	.dc.b	$FF,$2F,$FC,$42,$F7,$C8,$BF,$F0,$E7,$ED,$DB,$BF,$F2,$ED,$BF,$4D
	.dc.b	$F4,$72,$FF,$B1,$67,$9E,$38,$D7,$C7,$7D,$74,$F1,$79,$DF,$CA,$E7
	.dc.b	$35,$E0,$CD,$E6,$7F,$2F,$FD,$CB,$FC,$15,$97,$C3,$EA,$F9,$FD,$4F
	.dc.b	$D5,$3F,$CC,$75,$AE,$BF,$B5,$5A,$F9,$FC,$E1,$FE,$13,$9F,$F6,$3F
	.dc.b	$C3,$7A,$2F,$87,$47,$F8,$4E,$BE,$B4,$63,$AF,$F0,$7C,$DF,$E2,$B9
	.dc.b	$E6,$78,$FA,$7F,$D4,$BF,$55,$FE,$1F,$FC,$B5,$F9,$6F,$A3,$F0,$4F
	.dc.b	$BF,$F8,$4F,$E5,$39,$FF,$DF,$5F,$BA,$7F,$9F,$EF,$79,$7F,$89,$EF
	.dc.b	$7E,$3F,$F0,$BF,$AA,$5B,$F3,$E6,$1C,$30,$FF,$4B,$78,$6F,$D0,$FF
	.dc.b	$15,$39,$CB,$C3,$5F,$AA,$8C,$57,$FB,$A7,$2E,$DE,$CF,$A7,$6F,$44
	.dc.b	$E7,$56,$AF,$C8,$3D,$DD,$FF,$53,$E9,$FF,$C4,$DE,$A3,$F2,$E7,$E5
	.dc.b	$AF,$D2,$9F,$17,$FE,$9C,$E0,$FC,$37,$0B,$1E,$38,$70,$BD,$FE,$FD
	.dc.b	$EE,$CF,$B7,$D3,$7F,$84,$F0,$F5,$46,$3D,$BF,$7B,$AD,$7C,$5A,$F8
	.dc.b	$78,$3C,$F7,$FE,$53,$D3,$BC,$F9,$7A,$BE,$C7,$17,$E3,$BE,$7D,$F1
	.dc.b	$FF,$E3,$3F,$C9,$5F,$AB,$7F,$B0,$30,$FE,$F3,$C7,$FB,$AC,$FF,$CB
	.dc.b	$D8,$70,$FA,$C6,$1C,$3F,$0F,$F8,$7F,$15,$F6,$34,$ED,$A3,$F7,$0F
	.dc.b	$AF,$18,$C2,$E5,$D2,$BC,$5A,$F8,$63,$91,$99,$EF,$1F,$ED,$4F,$3B
	.dc.b	$F2,$5F,$05,$FC,$DF,$F2,$87,$E7,$1F,$E5,$BF,$A0,$FE,$29,$FB,$B3
	.dc.b	$F6,$9F,$C5,$DE,$EB,$5E,$FA,$BF,$CD,$1C,$9F,$CD,$75,$FE,$5B,$93
	.dc.b	$F7,$4F,$C7,$E3,$C6,$7F,$1C,$F3,$ED,$69,$FC,$8F,$F1,$9F,$2D,$7F
	.dc.b	$A1,$7C,$76,$3E,$6B,$DE,$8F,$C5,$3F,$59,$E0,$7F,$67,$E2,$7E,$9D
	.dc.b	$FC,$5F,$E7,$9F,$B0,$3F,$52,$7D,$CF,$E9,$B0,$E1,$7B,$AD,$65,$9D
	.dc.b	$BF,$B4,$65,$D1,$FD,$35,$F7,$FD,$0B,$F5,$6F,$2F,$80,$69,$E0,$FB
	.dc.b	$DF,$2D,$F2,$CF,$B3,$FC,$0F,$E7,$BF,$C3,$1F,$33,$FB,$6F,$5F,$F8
	.dc.b	$17,$A3,$FB,$A9,$CF,$A9,$FC,$03,$CB,$FF,$34,$F6,$7F,$57,$C7,$FC
	.dc.b	$63,$8B,$F4,$8F,$DE,$E1,$C2,$7F,$78,$F2,$F1,$3F,$71,$1D,$4F,$F0
	.dc.b	$BD,$78,$7A,$BF,$A3,$E1,$FE,$3C,$9F,$F1,$9F,$D3,$FF,$0E,$FD,$15
	.dc.b	$F1,$8F,$D1,$BF,$6D,$FA,$4F,$F1,$0F,$A2,$FF,$21,$7B,$CB,$DD,$7E
	.dc.b	$EF,$F3,$BD,$65,$D2,$B5,$8F,$3F,$53,$F9,$36,$3C,$B1,$E3,$EE,$7F
	.dc.b	$29,$FA,$BF,$F3,$0F,$E6,$4F,$CC,$FF,$B6,$72,$BF,$69,$7E,$FF,$FC
	.dc.b	$8A,$F7,$5A,$97,$ED,$B9,$DF,$FC,$C3,$8F,$3F,$90,$F9,$5F,$07,$DE
	.dc.b	$73,$87,$6F,$CE,$3F,$9D,$7D,$3F,$FA,$0F,$DD,$FC,$AF,$AB,$FA,$4E
	.dc.b	$AF,$F4,$0F,$CF,$FF,$A3,$3D,$1F,$CC,$7A,$7F,$90,$9C,$FD,$E3,$0E
	.dc.b	$3E,$9F,$FB,$1B,$5F,$E7,$39,$F5,$C7,$F4,$1F,$A2,$FF,$49,$FC,$03
	.dc.b	$FD,$51,$F3,$1F,$9E,$F3,$BF,$1F,$E9,$FF,$40,$9C,$F9,$BF,$A1,$F1
	.dc.b	$71,$9F,$FA,$AF,$F0,$17,$E0,$1F,$E0,$0A,$FE,$F9,$CF,$D6,$B5,$39
	.dc.b	$C9,$CD,$FD,$67,$D6,$FF,$69,$FB,$FF,$ED,$8F,$47,$FA,$67,$4B,$F1
	.dc.b	$7F,$EA,$35,$AF,$C1,$55,$AF,$D1,$B3,$A7,$FB,$47,$1E,$5F,$23,$FB
	.dc.b	$B8,$C9,$FE,$ED,$FA,$DF,$F6,$AD,$FF,$59,$F7,$FF,$B5,$7C,$F7,$FB
	.dc.b	$CF,$F0,$BF,$F1,$67,$96,$FA,$37,$E4,$9F,$D8,$AB,$5E,$37,$E4,$E7
	.dc.b	$F0,$DE,$1F,$F8,$67,$4F,$F7,$C7,$E1,$DF,$F0,$07,$2E,$7F,$9A,$F6
	.dc.b	$9C,$FB,$C9,$EF,$F8,$17,$F7,$BF,$AA,$FF,$08,$FE,$24,$FC,$49,$F4
	.dc.b	$AF,$C5,$1F,$91,$7E,$5F,$F4,$6F,$E8,$3B,$FF,$8F,$35,$0C,$79,$61
	.dc.b	$C6,$F8,$7A,$3F,$DF,$33,$EF,$9F,$6C,$FA,$CB,$6F,$26,$5D,$32,$E7
	.dc.b	$5A,$F0,$3F,$C2,$5D,$3F,$E2,$8F,$AF,$7E,$24,$FC,$69,$F8,$7B,$F0
	.dc.b	$87,$E5,$1F,$CB,$3E,$DB,$FC,$1B,$8F,$1C,$38,$5E,$EB,$F9,$8F,$D3
	.dc.b	$EB,$6F,$6E,$7D,$B3,$EB,$9F,$4F,$86,$F1,$D6,$B2,$E7,$3E,$0C,$3F
	.dc.b	$C5,$1F,$91,$7F,$8F,$BF,$1F,$7E,$3E,$FC,$1D,$F8,$FB,$F2,$27,$EF
	.dc.b	$8F,$2F,$FD,$F5,$DC,$CB,$96,$3C,$70,$E1,$7B,$CF,$FC,$53,$5A,$CB
	.dc.b	$AE,$5D,$3D,$3F,$1E,$FF,$63,$5F,$1D,$6B,$16,$3F,$E3,$FF,$C7,$BF
	.dc.b	$CA,$BD,$EF,$B5,$FF,$7D,$FE,$48,$FC,$95,$F8,$3B,$F6,$2F,$F4,$E7
	.dc.b	$1B,$1E,$38,$70,$BD,$D7,$FB,$4E,$B5,$B7,$E2,$B2,$ED,$F8,$C3,$3E
	.dc.b	$D9,$FF,$A3,$EB,$FC,$8B,$9F,$4A,$D6,$5E,$2F,$C5,$BF,$CA,$DF,$A0
	.dc.b	$7F,$9B,$3D,$C7,$DF,$FE,$0B,$F2,$B7,$B5,$FA,$EF,$6B,$F1,$1C,$0C
	.dc.b	$38,$5E,$EB,$5F,$B4,$7F,$95,$3F,$24,$FF,$48,$FD,$0F,$FA,$B3,$99
	.dc.b	$F9,$6B,$EC,$DF,$2A,$F9,$0E,$1C,$BF,$A6,$FC,$21,$8F,$1C,$38,$5B
	.dc.b	$F6,$AF,$F3,$0F,$EC,$1F,$E8,$4F,$0D,$FB,$3F,$F6,$5F,$CB,$1F,$76
	.dc.b	$FC,$2B,$E8,$FC,$D7,$0D,$BF,$67,$87,$0F,$9E,$61,$C2,$FE,$BF,$BF
	.dc.b	$BB,$FC,$66,$5F,$E3,$F7,$FD,$B1,$D2,$D3,$C5,$3F,$DD,$67,$FC,$DD
	.dc.b	$18,$C7,$AE,$9E,$3F,$43,$FD,$0D,$39,$CF,$C1,$F6,$58,$C3,$CB,$8F
	.dc.b	$AF,$0F,$57,$86,$FD,$9B,$C9,$BF,$A7,$1F,$44,$BF,$29,$FF,$42,$F8
	.dc.b	$7F,$EA,$8F,$C4,$3F,$A9,$7E,$50,$FC,$DF,$FD,$AD,$F9,$BB,$F5,$0F
	.dc.b	$F0,$7D,$EF,$E8,$57,$BF,$8D,$5E,$FF,$63,$FF,$48,$FE,$19,$FC,$37
	.dc.b	$F3,$87,$E1,$1F,$57,$CC,$7E,$9A,$FA,$57,$E8,$1C,$3A,$7F,$83,$E1
	.dc.b	$C3,$C2,$61,$C2,$DF,$A1,$6E,$DF,$08,$C3,$FD,$13,$1F,$EC,$4F,$05
	.dc.b	$AF,$8F,$C7,$FF,$50,$BE,$5D,$BC,$9B,$3F,$5B,$FF,$5B,$4E,$76,$F0
	.dc.b	$CF,$EA,$70,$E7,$8F,$9B,$0E,$3A,$BF,$7E,$F2,$6B,$E3,$9C,$FE,$0F
	.dc.b	$8D,$BD,$9C,$1E,$BF,$90,$C5,$72,$F0,$7D,$9F,$33,$FE,$B7,$FB,$EF
	.dc.b	$E9,$7F,$B4,$7F,$A9,$BF,$72,$7E,$A7,$AF,$A2,$FF,$27,$FC,$CA,$B5
	.dc.b	$F1,$D9,$FD,$AB,$DF,$C9,$FD,$3B,$3E,$8F,$FA,$FF,$A3,$FD,$D3,$1D
	.dc.b	$F0,$E5,$F8,$AF,$BE,$73,$97,$6C,$BF,$D6,$FD,$5F,$1B,$1E,$93,$9B
	.dc.b	$6A,$FD,$7B,$CD,$BF,$AA,$73,$F8,$3F,$FB,$5B,$F5,$1F,$F7,$CF,$E2
	.dc.b	$AF,$DC,$DF,$92,$BF,$7D,$E7,$F9,$A3,$F3,$37,$D9,$F1,$E0,$FE,$D3
	.dc.b	$8F,$1C,$38,$6E,$C3,$87,$EF,$BB,$B5,$ED,$FF,$70,$69,$E2,$D1,$FB
	.dc.b	$DF,$FB,$9A,$31,$97,$4C,$5F,$9F,$FF,$BE,$E7,$3B,$79,$BE,$1F,$19
	.dc.b	$79,$32,$F5,$CF,$AB,$77,$E8,$9E,$3D,$7D,$13,$9F,$CC,$3C,$DE,$07
	.dc.b	$E9,$3F,$52,$7A,$FF,$C9,$BE,$97,$FC,$17,$F5,$1A,$F4,$55,$AF,$88
	.dc.b	$57,$F8,$43,$F5,$8D,$7E,$C9,$EB,$E1,$FF,$07,$E7,$D5,$F8,$26,$31
	.dc.b	$E7,$9F,$7F,$03,$D7,$3F,$BA,$9C,$E5,$DB,$F1,$7C,$E1,$E7,$C3,$CD
	.dc.b	$F3,$5F,$1B,$D2,$73,$FF,$87,$B2,$E5,$BF,$9F,$7F,$F1,$2F,$CC,$FA
	.dc.b	$E3,$18,$3A,$31,$97,$4F,$A8,$FB,$5C,$5E,$EF,$CC,$BF,$8C,$78,$6F
	.dc.b	$CD,$FF,$CA,$5F,$8E,$EB,$55,$6A,$7F,$CA,$DF,$73,$98,$F8,$9B,$B1
	.dc.b	$F9,$3F,$A0,$61,$ED,$F0,$AC,$FB,$7E,$73,$F1,$27,$31,$E1,$FC,$EF
	.dc.b	$1A,$78,$2F,$86,$9F,$E6,$5F,$D2,$7C,$11,$8F,$C0,$1F,$99,$3A,$1F
	.dc.b	$49,$FD,$C7,$C3,$FF,$1A,$FE,$2E,$F5,$DF,$BB,$3F,$61,$7E,$B7,$C7
	.dc.b	$0E,$1E,$DB,$0E,$17,$BF,$AC,$D7,$3F,$F9,$43,$EA,$3F,$47,$7F,$56
	.dc.b	$3F,$94,$61,$E1,$F7,$B1,$1F,$E5,$6D,$BC,$AF,$1D,$BC,$BB,$79,$3E
	.dc.b	$E3,$58,$78,$B5,$F0,$FD,$D6,$30,$E1,$9F,$97,$0F,$5F,$03,$9F,$CB
	.dc.b	$5F,$E9,$6C,$FA,$F0,$7B,$FF,$69,$CF,$EB,$6F,$DA,$9F,$11,$FE,$07
	.dc.b	$F9,$CF,$F3,$97,$EC,$BF,$EC,$2E,$3F,$3E,$6C,$38,$5E,$FE,$FD,$7F
	.dc.b	$E1,$1B,$DD,$6A,$FE,$0C,$63,$1F,$E7,$D1,$8E,$3F,$CE,$57,$4F,$A8
	.dc.b	$FF,$07,$E3,$FE,$EA,$9C,$DF,$9A,$73,$F9,$1F,$C3,$C3,$C9,$39,$D3
	.dc.b	$D3,$F7,$5E,$E8,$C6,$3E,$6F,$C9,$F1,$B7,$A7,$67,$E9,$F8,$D3,$C3
	.dc.b	$1F,$E8,$3D,$FE,$0F,$ED,$5F,$8D,$F6,$BF,$DC,$DF,$9B,$BF,$68,$7B
	.dc.b	$6F,$D8,$7E,$8D,$EF,$F1,$E5,$EE,$9F,$9B,$EB,$26,$5F,$86,$E4,$C7
	.dc.b	$C0,$D9,$F0,$9A,$C7,$9E,$9F,$EC,$CA,$E5,$7E,$48,$7E,$01,$F1,$74
	.dc.b	$F1,$69,$E1,$FC,$07,$78,$F2,$D5,$FB,$3F,$E2,$3F,$5E,$7E,$4E,$FB
	.dc.b	$17,$3F,$FB,$57,$E2,$1F,$25,$F8,$0E,$2F,$C2,$E1,$C2,$F7,$E8,$7F
	.dc.b	$4F,$AF,$E0,$15,$AC,$7A,$71,$FE,$04,$FF,$5C,$E1,$E0,$FA,$CF,$81
	.dc.b	$E0,$FF,$1F,$8E,$9A,$78,$78,$FF,$01,$18,$CF,$FA,$E7,$99,$8C,$BD
	.dc.b	$17,$E7,$D3,$FD,$EF,$A7,$87,$5F,$2E,$DE,$AF,$D4,$BF,$4F,$F8,$0B
	.dc.b	$F5,$5F,$E8,$FF,$8A,$7F,$BF,$BC,$E7,$EC,$8E,$0D,$D6,$BD,$3F,$F6
	.dc.b	$0C,$BF,$29,$CF,$C2,$7A,$B9,$FF,$DA,$B9,$F5,$FD,$1B,$11,$FE,$8D
	.dc.b	$C7,$96,$7F,$B9,$8E,$3A,$78,$7C,$3F,$CF,$61,$E5,$C7,$FD,$EB,$A7
	.dc.b	$83,$E7,$F6,$F8,$BF,$43,$B9,$7A,$3F,$41,$BB,$FF,$25,$E4,$FD,$DF
	.dc.b	$EC,$2F,$C7,$5F,$64,$F4,$9F,$AA,$BF,$32,$7E,$76,$F5,$FF,$86,$E3
	.dc.b	$C7,$0E,$17,$BA,$FE,$25,$97,$6C,$9F,$A3,$FC,$0C,$38,$73,$7E,$04
	.dc.b	$3C,$F4,$F0,$E9,$F9,$DC,$39,$67,$DA,$3F,$3B,$7E,$2C,$7D,$37,$C3
	.dc.b	$2F,$F3,$86,$1C,$23,$1A,$FF,$24,$FA,$27,$E7,$AF,$E0,$0F,$67,$B7
	.dc.b	$FE,$62,$FD,$95,$FB,$06,$EB,$5C,$7F,$C8,$7E,$C9,$4F,$D8,$75,$8F
	.dc.b	$DC,$CB,$AF,$EE,$57,$1E,$39,$FF,$9B,$B1,$FE,$31,$5E,$2B,$F0,$C3
	.dc.b	$F3,$1C,$DF,$08,$EF,$F9,$5F,$6D,$3C,$5A,$3F,$60,$FE,$7F,$F8,$BF
	.dc.b	$FB,$5F,$F0,$1F,$F7,$DF,$53,$F3,$37,$D8,$FF,$9E,$B9,$BF,$9D,$DE
	.dc.b	$FF,$1D,$FA,$B5,$AF,$C1,$79,$F8,$DF,$E9,$6E,$CF,$AD,$7F,$EE,$3F
	.dc.b	$2F,$F3,$2F,$1C,$FB,$7E,$17,$FC,$CE,$31,$8F,$EF,$B3,$EF,$8F,$FA
	.dc.b	$83,$3E,$FA,$7F,$87,$27,$3A,$FA,$2F,$CF,$EB,$F8,$A3,$14,$FD,$5B
	.dc.b	$FA,$EF,$D8,$7F,$B7,$BF,$36,$FF,$51,$7E,$28,$FD,$49,$0F,$D8,$F7
	.dc.b	$F8,$B6,$3F,$05,$5E,$EB,$E3,$6D,$FD,$7B,$1F,$F7,$4F,$E7,$3F,$B6
	.dc.b	$7B,$73,$7E,$5D,$F0,$61,$CB,$3F,$EF,$F2,$3F,$CB,$B9,$77,$CB,$C9
	.dc.b	$F3,$3F,$D5,$FE,$A9,$F9,$23,$F3,$07,$E4,$3F,$B1,$F1,$9F,$9B,$7F
	.dc.b	$3A,$E0,$FE,$E3,$87,$0B,$DF,$99,$FD,$CE,$9F,$93,$7F,$80,$2B,$AF
	.dc.b	$E0,$7F,$C3,$9E,$1B,$FE,$BD,$E5,$9F,$7F,$D9,$3C,$9A,$F8,$E3,$C5
	.dc.b	$F9,$C7,$F6,$DF,$0D,$F9,$37,$F0,$5F,$C7,$3F,$83,$FC,$7F,$E3,$3E
	.dc.b	$A7,$FC,$1F,$5A,$E0,$56,$BB,$9F,$61,$AC,$7F,$D9,$DB,$FF,$01,$C3
	.dc.b	$83,$63,$B7,$B3,$FD,$B3,$C7,$F1,$46,$34,$FC,$67,$87,$FE,$63,$3F
	.dc.b	$84,$FD,$3F,$F7,$1F,$2B,$F8,$5F,$E5,$1F,$E5,$3E,$AF,$F0,$CD,$6B
	.dc.b	$E1,$3F,$9B,$7B,$AA,$7E,$01,$FE,$28,$AE,$7B,$7F,$15,$7F,$8A,$FD
	.dc.b	$47,$F3,$87,$96,$5F,$C6,$3A,$7F,$C2,$51,$8C,$7F,$A1,$5F,$F8,$CB
	.dc.b	$C9,$7E,$3A,$FD,$51,$F1,$BF,$81,$FC,$31,$F4,$5F,$E0,$2E,$1F,$F1
	.dc.b	$7E,$1C,$2F,$7E,$27,$FB,$CA,$BF,$D4,$35,$AC,$5E,$57,$F4,$CC,$38
	.dc.b	$67,$FC,$B5,$E5,$9F,$7F,$D2,$FF,$C5,$11,$8C,$BF,$DF,$BF,$99,$63
	.dc.b	$F3,$87,$E7,$DE,$7F,$1D,$FB,$FB,$D6,$7E,$BF,$ED,$BD,$F1,$2F,$75
	.dc.b	$FE,$CF,$E1,$F5,$D7,$F2,$FD,$FC,$8F,$66,$E5,$DF,$27,$EB,$9E,$38
	.dc.b	$C7,$67,$F8,$FF,$95,$E5,$FF,$3A,$FD,$47,$F9,$D7,$94,$FC,$F1,$F8
	.dc.b	$D7,$F9,$8A,$F7,$F0,$8B,$DD,$7F,$5F,$DF,$FC,$BB,$5D,$3F,$51,$FE
	.dc.b	$EA,$F0,$DF,$FA,$B3,$CA,$FB,$F8,$FE,$3D,$3C,$5F,$0E,$D5,$F8,$1F
	.dc.b	$F9,$57,$F0,$57,$E9,$0F,$D5,$3F,$C5,$BE,$B3,$F6,$C7,$DA,$BF,$0F
	.dc.b	$FF,$3C,$E1,$C2,$F7,$5A,$FD,$23,$38,$F3,$E1,$FE,$D5,$7D,$BF,$2A
	.dc.b	$FF,$06,$BC,$7E,$A3,$F8,$2E,$1F,$32,$B8,$E3,$D3,$A7,$CD,$8F,$96
	.dc.b	$F7,$A3,$F3,$DF,$F4,$5F,$EC,$3F,$EF,$8F,$82,$FF,$7C,$FE,$A7,$E7
	.dc.b	$FC,$C3,$F3,$BD,$7F,$E8,$6A,$D7,$C3,$AB,$5F,$AC,$F2,$8F,$8D,$BF
	.dc.b	$F8,$06,$BA,$36,$3F,$9A,$9C,$79,$18,$FF,$9F,$EB,$AE,$1F,$1F,$C9
	.dc.b	$C5,$F0,$8F,$27,$1F,$FA,$8B,$F3,$EE,$3A,$3F,$6A,$7E,$B4,$FC,$B1
	.dc.b	$F9,$04,$AF,$F4,$FD,$F0,$BD,$D6,$BF,$21,$E9,$B7,$FB,$3F,$83,$FC
	.dc.b	$35,$7E,$6F,$D3,$3F,$C7,$AF,$1C,$FB,$74,$7F,$A5,$E3,$18,$FF,$8F
	.dc.b	$F3,$EF,$8F,$FB,$67,$3E,$FA,$3F,$6E,$FF,$5D,$FC,$57,$F7,$AF,$CD
	.dc.b	$3F,$43,$F0,$0F,$C5,$FF,$D2,$55,$AF,$89,$56,$A7,$FD,$2D,$86,$7E
	.dc.b	$57,$FC,$A3,$8F,$3E,$2F,$F1,$93,$D7,$C8,$FE,$47,$7F,$CB,$A7,$FA
	.dc.b	$E4,$E7,$3F,$07,$93,$FE,$B1,$8C,$57,$FA,$8F,$F4,$8F,$F6,$8F,$C9
	.dc.b	$BF,$31,$F3,$7E,$13,$FE,$D5,$FC,$E1,$EF,$BF,$72,$DD,$6B,$F1,$15
	.dc.b	$6B,$F3,$5C,$E7,$FE,$68,$CB,$A7,$EC,$F7,$0E,$17,$D3,$83,$7A,$77
	.dc.b	$3F,$D8,$D7,$E3,$C3,$FD,$BF,$97,$6D,$1F,$9E,$FF,$BA,$3F,$7B,$7E
	.dc.b	$AC,$FC,$81,$F9,$13,$F7,$4F,$F1,$87,$E2,$6F,$A7,$E1,$7B,$AD,$7D
	.dc.b	$57,$FB,$93,$F5,$2C,$E3,$FE,$80,$CB,$AF,$8B,$70,$E1,$9F,$F8,$EF
	.dc.b	$F3,$5F,$F6,$E5,$78,$AF,$C3,$3F,$E3,$FF,$A0,$CD,$F0,$CB,$FD,$E3
	.dc.b	$FA,$AF,$FC,$00,$FD,$C1,$FA,$57,$F8,$3B,$F2,$B7,$E0,$2F,$D9,$DF
	.dc.b	$BE,$A7,$85,$6E,$B5,$2F,$57,$3F,$B1,$FF,$A6,$31,$EB,$F5,$57,$0E
	.dc.b	$19,$3C,$6F,$F7,$A5,$6B,$3F,$0C,$FE,$2F,$BF,$FC,$3B,$F9,$4B,$F3
	.dc.b	$17,$ED,$EF,$8D,$78,$6F,$DD,$7F,$CB,$3E,$AB,$E8,$BF,$C5,$D9,$72
	.dc.b	$C7,$8E,$1C,$2F,$FD,$75,$7B,$AF,$F1,$A7,$D7,$B3,$FF,$50,$47,$FA
	.dc.b	$46,$31,$87,$FA,$D1,$B9,$F8,$F3,$7B,$FF,$E3,$DC,$7A,$61,$E1,$AD
	.dc.b	$78,$B9,$D7,$CB,$8F,$2D,$7C,$9F,$9D,$B1,$E3,$39,$CF,$DD,$CF,$9C
	.dc.b	$3D,$BF,$90,$7F,$01,$FD,$F1,$FB,$6F,$F1,$9E,$B7,$E3,$6F,$82,$E1
	.dc.b	$C3,$FE,$51,$C3,$85,$EE,$BF,$8E,$63,$D3,$DC,$FC,$2D,$FF,$AC,$6F
	.dc.b	$B3,$61,$97,$83,$DD,$FC,$4F,$E2,$DE,$DA,$ED,$5A,$CB,$C9,$D9,$99
	.dc.b	$63,$E2,$C3,$F2,$39,$FF,$1B,$E1,$F9,$3F,$7D,$FE,$76,$FA,$A7,$27
	.dc.b	$E5,$3F,$6E,$FE,$67,$CD,$EC,$CF,$FD,$93,$F7,$9F,$EC,$6A,$D7,$C0
	.dc.b	$3E,$D7,$59,$73,$BF,$F6,$63,$FE,$82,$FA,$0F,$F7,$7B,$DE,$FB,$7C
	.dc.b	$3F,$9E,$31,$D2,$D3,$F9,$A7,$E6,$DF,$8E,$FD,$41,$FB,$A3,$F1,$EF
	.dc.b	$F5,$67,$D2,$BF,$18,$FC,$67,$E6,$3F,$77,$0E,$17,$BF,$C1,$3F,$C3
	.dc.b	$75,$F9,$1F,$CC,$75,$C3,$FE,$87,$CF,$AF,$EA,$5F,$9C,$8F,$CE,$F4
	.dc.b	$32,$E7,$AE,$67,$F2,$51,$8C,$BA,$FA,$D8,$9E,$9A,$78,$76,$F3,$78
	.dc.b	$B9,$96,$9E,$CC,$BD,$7B,$7E,$97,$5F,$2D,$6B,$1E,$53,$FA,$9A,$CD
	.dc.b	$F1,$CF,$DB,$E5,$BF,$3F,$E3,$F8,$EB,$F7,$8F,$D0,$78,$9F,$BB,$FD
	.dc.b	$47,$FD,$0A,$6A,$33,$E8,$23,$FD,$13,$F9,$8E,$79,$7E,$07,$06,$FE
	.dc.b	$62,$C7,$C5,$8F,$F5,$86,$5D,$E3,$FB,$46,$1C,$FE,$0B,$E0,$4E,$73
	.dc.b	$F1,$61,$8C,$FC,$39,$FF,$A6,$3D,$7F,$83,$F3,$DF,$E7,$6F,$C9,$5F
	.dc.b	$B5,$3F,$28,$7E,$4E,$FC,$C7,$FC,$C1,$C1,$E3,$E1,$FF,$66,$FD,$EB
	.dc.b	$1E,$38,$70,$BF,$F6,$05,$EF,$2E,$D8,$7D,$D8,$C6,$0F,$CF,$7D,$FE
	.dc.b	$07,$F2,$0B,$F2,$69,$E3,$D3,$FC,$9F,$8F,$A3,$0F,$04,$E7,$D1,$FC
	.dc.b	$05,$F0,$D3,$C5,$FC,$01,$FC,$4F,$CE,$FE,$57,$9B,$F4,$37,$ED,$7F
	.dc.b	$EE,$6F,$C6,$9F,$BE,$7F,$36,$FC,$77,$86,$FF,$ED,$FC,$38,$5E,$EB
	.dc.b	$F9,$DF,$CB,$EA,$BF,$35,$8F,$EA,$9F,$76,$5E,$08,$EF,$18,$3A,$3E
	.dc.b	$2D,$59,$F8,$BF,$09,$FE,$65,$7C,$72,$F0,$7E,$13,$8C,$78,$C7,$8F
	.dc.b	$E0,$FF,$D3,$7C,$DF,$D3,$67,$39,$74,$D7,$FA,$8F,$DC,$F9,$F7,$F5
	.dc.b	$CE,$70,$7B,$3E,$BF,$CC,$BF,$AC,$7E,$2C,$F6,$DF,$B3,$3E,$5D,$F8
	.dc.b	$DF,$F5,$BD,$FC,$1B,$B1,$E3,$87,$0B,$DF,$BB,$D7,$E2,$D9,$FE,$7B
	.dc.b	$93,$FA,$FE,$1F,$D9,$FE,$27,$19,$0E,$97,$C3,$2E,$9F,$8F,$7C,$39
	.dc.b	$CE,$3E,$7D,$BC,$DD,$19,$CB,$C7,$1E,$9D,$FF,$CF,$7A,$FF,$6E,$FC
	.dc.b	$2F,$FD,$2B,$EA,$BF,$6A,$FF,$11,$FE,$C8,$E0,$7C,$0F,$F5,$DC,$38
	.dc.b	$7D,$C7,$F2,$8B,$DD,$3F,$38,$D7,$07,$A7,$3E,$AF,$E8,$70,$E3,$EE
	.dc.b	$7E,$F9,$E7,$A7,$87,$E7,$DA,$BC,$0F,$EE,$95,$AD,$3C,$BF,$81,$63
	.dc.b	$2F,$8D,$FB,$5B,$F3,$7F,$CE,$FE,$CA,$FC,$F5,$FA,$55,$E8,$F2,$E0
	.dc.b	$C3,$87,$8C,$C3,$85,$F1,$FD,$2F,$75,$AB,$F5,$1E,$0C,$BF,$40,$FF
	.dc.b	$3D,$F0,$7F,$AC,$39,$C7,$A6,$7E,$2F,$B8,$F4,$4F,$86,$FC,$F3,$E6
	.dc.b	$F8,$6C,$FE,$A5,$FE,$26,$FC,$CB,$FC,$A5,$FB,$27,$E3,$7F,$5E,$7E
	.dc.b	$17,$EA,$AD,$7D,$6A,$B5,$F8,$86,$7B,$1F,$A2,$FC,$FD,$BF,$19,$87
	.dc.b	$57,$D6,$1E,$77,$F9,$BF,$C3,$BC,$76,$D3,$C7,$E3,$3F,$23,$CC,$63
	.dc.b	$67,$E7,$BF,$E2,$3F,$C7,$7F,$1D,$EA,$7F,$11,$F0,$FF,$D1,$1C,$E8
	.dc.b	$7E,$E9,$BE,$0F,$C4,$46,$FE,$EB,$F8,$BA,$FB,$BD,$4B,$C7,$E5,$ED
	.dc.b	$7F,$48,$CF,$BF,$4D,$AE,$73,$FD,$FB,$1F,$E4,$3F,$A7,$FF,$8F,$EF
	.dc.b	$CB,$39,$D3,$FD,$F7,$7F,$E3,$BF,$C3,$3F,$4D,$F2,$9F,$A6,$DF,$FC
	.dc.b	$0D,$E3,$BF,$65,$D5,$FE,$7E,$B5,$F2,$DF,$AB,$FD,$9F,$33,$E7,$97
	.dc.b	$A3,$FD,$7B,$2E,$BF,$8E,$5A,$E5,$3F,$E1,$4C,$FF,$C2,$9D,$7F,$E5
	.dc.b	$5B,$F1,$CE,$72,$FF,$52,$78,$CF,$C3,$3F,$79,$F0,$1D,$B8,$7F,$C3
	.dc.b	$BC,$2F,$80,$FF,$38,$5E,$F5,$FD,$B5,$EF,$5F,$E6,$9C,$BF,$E5,$BC
	.dc.b	$B9,$E9,$FD,$D1,$EB,$BD,$C3,$8C,$FD,$DC,$FF,$E0,$0E,$5D,$67,$EB
	.dc.b	$65,$D6,$7F,$DB,$31,$FE,$67,$FC,$1F,$F8,$97,$47,$C4,$C3,$E0,$72
	.dc.b	$FF,$AD,$31,$56,$A7,$E9,$4B,$C1,$E5,$87,$F0,$FC,$3E,$6B,$FD,$BE
	.dc.b	$FA,$E7,$FE,$0C,$79,$61,$F3,$78,$FF,$D0,$11,$8D,$FF,$C8,$7F,$5F
	.dc.b	$FE,$83,$FD,$FB,$CB,$DB,$F1,$67,$FD,$C1,$B2,$73,$8F,$E7,$67,$E6
	.dc.b	$C7,$0C,$FF,$C0,$0F,$BB,$E1,$FF,$C3,$AF,$0E,$5F,$F2,$8E,$9D,$EB
	.dc.b	$D5,$97,$D3,$CF,$FD,$29,$D1,$F1,$8F,$8D,$97,$E1,$74,$FF,$1F,$DF
	.dc.b	$C1,$9C,$CF,$FA,$BA,$73,$3F,$A7,$DE,$BF,$E5,$2A,$FA,$39,$FF,$8C
	.dc.b	$1E,$DA,$FF,$44,$CF,$C3,$5D,$63,$FA,$3E,$5F,$EA,$7F,$C8,$3E,$CF
	.dc.b	$27,$D9,$DB,$C9,$18,$AF,$C6,$61,$FE,$A5,$8D,$E1,$FE,$23,$7F,$03
	.dc.b	$F5,$FF,$E4,$07,$85,$7F,$85,$EB,$B5,$7D,$DC,$3B,$63,$FE,$B7,$FD
	.dc.b	$07,$F1,$3A,$FE,$27,$0F,$89,$18,$7E,$3E,$1F,$EB,$58,$C4,$7E,$CA
	.dc.b	$39,$61,$F8,$9A,$E5,$7F,$EC,$3F,$C4,$7A,$F0,$FB,$78,$7F,$D7,$D1
	.dc.b	$8C,$7F,$5D,$8E,$31,$E5,$7F,$9B,$F0,$79,$46,$2B,$F8,$0F,$BD,$FE
	.dc.b	$CF,$FB,$6F,$4F,$17,$A7,$3F,$F6,$54,$62,$BF,$49,$96,$30,$F3,$5F
	.dc.b	$57,$07,$DD,$D7,$FD,$03,$5D,$A3,$FD,$A9,$F7,$5F,$1B,$83,$E3,$71
	.dc.b	$E2,$7E,$04,$7C,$9D,$71,$1C,$E3,$B7,$2A,$F9,$61,$D6,$1F,$B0,$FD
	.dc.b	$2C,$BE,$94,$FF,$B8,$E3,$0F,$EF,$7A,$71,$8F,$CF,$AE,$B5,$AA,$FF
	.dc.b	$07,$CF,$FB,$AF,$F5,$EE,$9C,$5F,$77,$14,$62,$BF,$95,$E3,$BB,$E9
	.dc.b	$7F,$CB,$B4,$FF,$7E,$5E,$F8,$BF,$DF,$DF,$8D,$79,$F8,$3A,$37,$2E
	.dc.b	$79,$A3,$18,$70,$CB,$D1,$7F,$07,$F1,$9F,$81,$C1,$E5,$6E,$3C,$9B
	.dc.b	$7F,$E3,$BC,$51,$AC,$B0,$FF,$8F,$F1,$7E,$7A,$FC,$3D,$F7,$AF,$59
	.dc.b	$F2,$6F,$E0,$0F,$F0,$EF,$DC,$34,$EB,$97,$F9,$27,$2E,$58,$F1,$C3
	.dc.b	$87,$C0,$BF,$60,$E1,$AF,$86,$F7,$A7,$6D,$1B,$F9,$B3,$E9,$B7,$92
	.dc.b	$B5,$F0,$1E,$DF,$93,$FC,$17,$57,$D7,$6F,$1F,$CC,$6E,$3B,$C7,$F0
	.dc.b	$55,$F8,$BF,$C1,$F8,$F4,$FF,$43,$E3,$46,$31,$75,$7B,$E3,$78,$FC
	.dc.b	$D7,$8F,$37,$CF,$FC,$4B,$C7,$D3,$CD,$B8,$72,$C5,$1A,$C3,$86,$9E
	.dc.b	$4B,$FC,$9D,$3F,$36,$F6,$F5,$79,$5B,$8F,$37,$38,$7D,$F7,$79,$7E
	.dc.b	$07,$F6,$6F,$8B,$BF,$B5,$B7,$E3,$73,$7B,$CF,$E2,$D7,$8B,$F3,$0E
	.dc.b	$3C,$1F,$E0,$AD,$11,$8E,$0F,$C1,$3D,$BD,$8F,$C8,$3E,$0F,$C7,$FF
	.dc.b	$25,$FB,$47,$8B,$87,$93,$6F,$F1,$5F,$AD,$F9,$B8,$BB,$36,$3F,$1D
	.dc.b	$FB,$0F,$AB,$5F,$13,$63,$F2,$5D,$39,$DA,$EF,$EA,$B7,$A1,$F9,$7F
	.dc.b	$D0,$B9,$6B,$6B,$5E,$8F,$E6,$FE,$73,$B6,$D6,$79,$7A,$4F,$CD,$D5
	.dc.b	$A6,$CE,$96,$3D,$9B,$1D,$9F,$D7,$3E,$5D,$3A,$6B,$ED,$D3,$B6,$FE
	.dc.b	$FE,$8B,$7E,$BE,$0F,$17,$0F,$C7,$9B,$3F,$63,$6F,$CB,$E6,$E2,$FA
	.dc.b	$15,$E9,$7E,$84,$7D,$AA,$FB,$FA,$7F,$91,$A3,$11,$FC,$F7,$8B,$FB
	.dc.b	$D4,$F4,$E4,$FB,$D3,$F0,$DF,$BC,$FC,$5E,$8F,$C6,$E3,$63,$F1,$6F
	.dc.b	$92,$3F,$1F,$19,$3B,$5F,$F9,$B7,$A7,$FD,$A5,$8F,$E8,$67,$F6,$F9
	.dc.b	$FF,$8D,$E3,$11,$C2,$3F,$99,$C6,$2B,$D1,$1F,$C2,$A7,$EC,$75,$7E
	.dc.b	$91,$FB,$5D,$7F,$B4,$CA,$C7,$EC,$5F,$93,$1F,$B5,$8C,$9E,$6E,$7F
	.dc.b	$E8,$F8,$FF,$06,$8F,$E1,$7A,$F9,$7B,$3F,$84,$70,$FE,$1D,$FE,$0F
	.dc.b	$34,$DE,$BE,$EE,$DF,$E2,$F9,$FD,$16,$DF,$E3,$DF,$E3,$35,$4F,$E3
	.dc.b	$33,$E2,$F0,$BF,$95,$73,$7F,$25,$F1,$7F,$97,$F8,$FF,$97,$9C,$F8
	.dc.b	$FF,$CC,$FC,$4F,$F6,$AF,$8D,$9C,$BF,$96,$CE,$7C,$3F,$A3,$7C,$3C
	.dc.b	$5F,$E5,$F5,$B3,$E6,$E9,$F4,$E7,$FA,$8B,$F9,$07,$FA,$8C,$6F,$9F
	.dc.b	$FB,$E4,$FC,$8F,$23,$FB,$03,$F8,$F6,$F0,$7E,$B5,$FE,$C3,$54,$FE
	.dc.b	$C3,$3F,$43,$C9,$FE,$E4,$FE,$5D,$B5,$FB,$57,$FB,$B6,$5F,$8A,$7F
	.dc.b	$B9,$57,$EE,$7C,$AF,$F0,$05,$FE,$7D,$B7,$99,$FF,$03,$4F,$F7,$B7
	.dc.b	$07,$E8,$DF,$F0,$F7,$37,$F8,$56,$3F,$C3,$71,$FE,$53,$8C,$47,$C0
	.dc.b	$8F,$F0,$24,$63,$2F,$F3,$27,$57,$F8,$AE,$31,$1F,$E9,$89,$EB,$E7
	.dc.b	$7F,$8C,$DF,$F5,$CF,$9F,$FC,$46,$31,$0F,$3F,$18,$FF,$8D,$A3,$D1
	.dc.b	$82,$B9,$74,$7E,$B3,$D0,$FF,$25,$57,$FB,$96,$91,$8F,$43,$A4,$63
	.dc.b	$0F,$F2,$55,$7E,$AE,$31,$1F,$E4,$1E,$CF,$F5,$56,$1F,$B8,$F4,$BF
	.dc.b	$CB,$F7,$FE,$CB,$B4,$63,$4E,$AE,$B4,$E1,$8F,$F9,$8E,$FF,$D6,$9E
	.dc.b	$97,$F0,$2F,$4F,$FC,$EB,$A7,$36,$E3,$E6,$7F,$CF,$B7,$FE,$1B,$3F
	.dc.b	$9B,$3F,$E7,$E7,$06,$0F,$F3,$E5,$FF,$9A,$9C,$1F,$51,$FF,$6D,$5F
	.dc.b	$F1,$2F,$03,$FD,$2B,$1F,$EE,$5B,$FE,$85,$18,$7F,$D5,$79,$FF,$A8
	.dc.b	$38,$3F,$DE,$18,$78,$23,$11,$FE,$B9,$8C,$57,$1C,$FB,$47,$FA,$7F
	.dc.b	$0F,$F6,$1B,$FD,$3F,$CC,$FF,$5C,$62,$1E,$CE,$67,$FD,$73,$5F,$EC
	.dc.b	$3F,$B3,$C4,$7F,$38,$3C,$57,$FE,$C3,$69,$C3,$6F,$F6,$1B,$C3,$30
	.dc.b	$E1,$FF,$62,$47,$F6,$1F,$5B,$FD,$B1,$9B,$D8,$7C,$BE,$6B,$FD,$C9
	.dc.b	$A7,$FB,$A3,$D9,$C4,$7F,$B0,$E3,$E9,$B8,$CB,$BF,$4E,$38,$BF,$DE
	.dc.b	$33,$CA,$BF,$DD,$0E,$EB,$FD,$9F,$C7,$FD,$1B,$DB,$FF,$2D,$43,$AB
	.dc.b	$FE,$00,$7B,$8E,$3F,$EF,$15,$A9,$CF,$26,$31,$1F,$B1,$3E,$C4,$EF
	.dc.b	$4F,$23,$C7,$37,$1F,$D3,$F7,$1F,$29,$E3,$C3,$9E,$BE,$3B,$DF,$1A
	.dc.b	$DF,$8E,$FF,$5E,$F8,$0F,$4F,$57,$EF,$A9,$F0,$1F,$37,$E0,$BF,$C2
	.dc.b	$15,$AC,$7F,$05,$5A,$E8,$7C,$16,$B5,$F0,$E4,$BF,$99,$87,$6E,$1B
	.dc.b	$5A,$9E,$79,$FF,$0B,$8E,$47,$E0,$AF,$C9,$0E,$6F,$EE,$BF,$07,$F4
	.dc.b	$3B,$3D,$BC,$BE,$D5,$F0,$BF,$E2,$79,$D7,$2F,$E6,$6B,$5F,$0B,$9C
	.dc.b	$FE,$3E,$1F,$31,$FD,$45,$F5,$F2,$3F,$08,$F2,$E2,$FD,$0F,$2B,$E0
	.dc.b	$3C,$77,$CB,$3E,$D2,$F8,$0F,$F0,$5F,$CB,$7F,$A7,$FE,$37,$F9,$47
	.dc.b	$C5,$3E,$3D,$F1,$CF,$8E,$7C,$6B,$E3,$7F,$C8,$18,$F1,$F2,$D8,$F1
	.dc.b	$C3,$85,$BE,$43,$7B,$FE,$33,$4F,$E6,$DF,$21,$F9,$DC,$B9,$63,$6B
	.dc.b	$59,$74,$D7,$C7,$AB,$CA,$F4,$D6,$AF,$C1,$2F,$90,$CC,$B3,$F6,$E1
	.dc.b	$CB,$5F,$F5,$D7,$96,$D3,$FD,$81,$F2,$6F,$93,$FC,$A7,$CA,$BE,$51
	.dc.b	$F2,$77,$E5,$BD,$7A,$B0,$E1,$D0,$C3,$85,$F1,$F9,$6D,$D6,$AF,$F9
	.dc.b	$16,$7E,$0C,$FF,$8D,$46,$0F,$D8,$CF,$4C,$39,$FC,$A7,$B2,$31,$A7
	.dc.b	$93,$BF,$DF,$E7,$7F,$64,$F3,$3F,$88,$FC,$C3,$E6,$9E,$DF,$A2,$B5
	.dc.b	$EB,$AB,$5F,$35,$9F,$3B,$C7,$BF,$F4,$9E,$6F,$8C,$FE,$EA,$3F,$48
	.dc.b	$E7,$8B,$F8,$7C,$7F,$33,$D3,$C3,$F3,$5E,$48,$C6,$5D,$FD,$5F,$F2
	.dc.b	$DF,$D0,$7F,$BB,$FB,$BF,$D1,$3F,$3C,$FA,$07,$CF,$BE,$3D,$F3,$FF
	.dc.b	$5F,$81,$8F,$1F,$9B,$63,$C7,$0E,$16,$F4,$3F,$D2,$D7,$BA,$FE,$58
	.dc.b	$F1,$E7,$BC,$7F,$C8,$35,$F1,$EB,$E2,$D5,$F4,$1E,$8A,$D6,$BE,$0A
	.dc.b	$D7,$85,$99,$FD,$44,$E7,$6F,$1E,$7E,$29,$FE,$5B,$EF,$FF,$28,$F0
	.dc.b	$DF,$02,$C5,$F4,$6F,$A3,$7D,$1F,$1C,$7F,$9E,$7A,$9F,$CA,$70,$E1
	.dc.b	$7B,$FC,$0B,$59,$FF,$2E,$BF,$EE,$1A,$7F,$3C,$78,$E9,$F0,$1B,$87
	.dc.b	$2F,$95,$F5,$5F,$96,$3A,$F6,$63,$3F,$06,$DE,$6D,$BF,$B2,$74,$FD
	.dc.b	$4C,$BD,$B5,$AD,$BD,$5F,$3F,$E4,$8C,$5B,$E9,$7F,$D9,$5A,$79,$74
	.dc.b	$7D,$53,$EB,$9F,$56,$FA,$A7,$D6,$3E,$B9,$F9,$7F,$2E,$0F,$EB,$39
	.dc.b	$72,$C7,$8E,$1C,$2D,$F5,$BB,$BF,$A3,$9F,$77,$FC,$2D,$A7,$8E,$1F
	.dc.b	$5F,$FD,$CE,$31,$8F,$4D,$3F,$8A,$5F,$3C,$FC,$1A,$F9,$E5,$F5,$FF
	.dc.b	$EB,$AC,$38,$CE,$74,$7E,$5D,$E3,$FB,$67,$DA,$3E,$C1,$F6,$DF,$DE
	.dc.b	$7E,$D5,$F9,$B9,$E1,$5B,$9D,$4B,$ED,$B3,$7E,$9E,$0F,$EF,$B8,$F6
	.dc.b	$FB,$7C,$42,$3C,$33,$E0,$F9,$BE,$DF,$71,$E9,$AE,$39,$F7,$9F,$16
	.dc.b	$19,$9E,$53,$9D,$3C,$FF,$73,$F3,$FE,$5F,$FC,$71,$F2,$EF,$BC,$7B
	.dc.b	$CF,$BD,$7D,$EF,$F8,$03,$C6,$FF,$0D,$69,$FE,$14,$C7,$8E,$1C,$2F
	.dc.b	$7F,$7B,$D7,$E5,$9F,$7D,$EC,$BE,$F8,$7D,$C3,$94,$FF,$83,$30,$E5
	.dc.b	$F7,$FE,$9A,$EB,$A7,$8F,$E4,$B1,$7B,$8F,$36,$BD,$B5,$F2,$E4,$C3
	.dc.b	$8E,$BE,$AF,$C0,$7C,$9B,$FB,$F1,$E3,$F8,$0B,$17,$DF,$FE,$D7,$F0
	.dc.b	$4F,$E4,$5E,$B3,$F0,$57,$E0,$AF,$C1,$5F,$83,$7F,$23,$C3,$87,$BB
	.dc.b	$FC,$1A,$F7,$5A,$F6,$BB,$56,$B1,$FF,$26,$3C,$3D,$BB,$C7,$FE,$40
	.dc.b	$D3,$C3,$8F,$2F,$8D,$FF,$21,$ED,$E6,$AD,$69,$DF,$6F,$27,$E3,$5A
	.dc.b	$CF,$F8,$77,$DD,$FE,$63,$F0,$C7,$E1,$6F,$C2,$DF,$87,$3F,$0C,$F5
	.dc.b	$EB,$D7,$56,$3C,$70,$E1,$7B,$F3,$7D,$75,$FE,$56,$AF,$D2,$F9,$9E
	.dc.b	$A8,$E3,$2C,$F1,$5C,$F3,$F0,$FE,$1B,$E7,$BE,$57,$BC,$BB,$FA,$79
	.dc.b	$9E,$78,$FA,$B0,$F4,$EA,$FC,$3B,$C7,$5F,$DB,$FC,$17,$C3,$FF,$39
	.dc.b	$EA,$FF,$6A,$79,$BF,$6F,$8B,$DD,$F1,$7F,$9E,$AD,$78,$DF,$E7,$5E
	.dc.b	$7E,$73,$FA,$2F,$C3,$3D,$8E,$F3,$FF,$55,$73,$FF,$3B,$8E,$5F,$03
	.dc.b	$FD,$93,$07,$E1,$FE,$99,$CE,$BF,$EE,$5F,$B0,$C6,$FE,$6C,$38,$6E
	.dc.b	$FC,$87,$C3,$39,$BF,$4F,$93,$EE,$FC,$47,$F8,$67,$E2,$5F,$EE,$BE
	.dc.b	$E7,$E0,$BF,$96,$FC,$4B,$5E,$02,$B5,$3F,$9A,$FC,$07,$3F,$89,$FD
	.dc.b	$7C,$79,$F2,$FB,$0F,$F4,$8B,$DA,$7F,$0F,$5F,$CA,$A7,$AD,$F4,$F0
	.dc.b	$7E,$B7,$E3,$0F,$C5,$DF,$86,$FF,$AC,$BF,$18,$7E,$30,$FC,$6B,$FC
	.dc.b	$BD,$E0,$7E,$3F,$0E,$17,$BA,$D7,$E3,$3F,$72,$73,$A7,$94,$3D,$FE
	.dc.b	$6F,$F2,$0F,$2F,$A1,$67,$E0,$FC,$6B,$CB,$AF,$92,$73,$93,$DC,$CE
	.dc.b	$9E,$1D,$9F,$09,$F8,$CF,$94,$FF,$6E,$FE,$38,$F1,$1E,$8B,$F1,$DF
	.dc.b	$F4,$97,$E3,$BB,$FC,$05,$7B,$AD,$78,$CF,$C7,$B5,$3F,$A7,$CB,$FD
	.dc.b	$D3,$F8,$15,$C3,$8E,$7D,$B3,$FF,$4F,$73,$F0,$8C,$1F,$A5,$BF,$2B
	.dc.b	$BC,$9F,$8F,$FE,$93,$F1,$7F,$C6,$7E,$14,$FC,$31,$F8,$C7,$9B,$F4
	.dc.b	$D7,$E4,$3B,$F3,$7F,$D8,$75,$AE,$15,$7F,$62,$F0,$7F,$83,$65,$F9
	.dc.b	$7F,$C8,$BF,$80,$3D,$B8,$3E,$B3,$E0,$C3,$97,$87,$FB,$B8,$C6,$9F
	.dc.b	$1B,$DB,$C6,$9E,$08,$F2,$7C,$37,$EF,$3C,$47,$DC,$FF,$73,$F7,$1F
	.dc.b	$92,$3F,$24,$FE,$67,$7B,$AD,$4E,$63,$D1,$EC,$FF,$8C,$70,$E7,$F2
	.dc.b	$FF,$C5,$DE,$BF,$2F,$70,$E3,$8F,$F9,$E3,$87,$A5,$FF,$29,$CB,$C1
	.dc.b	$E0,$7F,$15,$C7,$9E,$9E,$4C,$3F,$A2,$FB,$FF,$C5,$3F,$16,$7A,$BF
	.dc.b	$E4,$E1,$F9,$2B,$F2,$6F,$6F,$E2,$5B,$9D,$4E,$77,$7E,$2B,$9F,$B8
	.dc.b	$FE,$5D,$8F,$4F,$A5,$B8,$70,$CB,$E3,$F1,$73,$9F,$0E,$5D,$FE,$0F
	.dc.b	$F9,$66,$1C,$70,$F1,$E1,$F3,$BE,$DF,$F9,$5E,$7F,$ED,$7F,$C7,$1F
	.dc.b	$94,$7F,$B2,$FF,$29,$FF,$7D,$5E,$BF,$12,$56,$BF,$10,$57,$FB,$FF
	.dc.b	$F2,$9D,$4F,$F4,$5C,$7F,$DC,$5F,$8F,$DC,$38,$E5,$D7,$E7,$3F,$A0
	.dc.b	$57,$86,$3C,$1F,$2D,$8A,$DE,$7D,$BF,$00,$7E,$55,$FD,$0B,$E2,$FF
	.dc.b	$3D,$F9,$0B,$F2,$B7,$E3,$2E,$A7,$E5,$BF,$C0,$7C,$FF,$EC,$18,$70
	.dc.b	$BD,$D6,$BF,$2E,$E9,$E7,$BF,$2E,$FE,$BD,$7D,$BB,$FF,$8A,$3C,$73
	.dc.b	$ED,$E7,$7F,$59,$8C,$63,$FE,$56,$CF,$FB,$77,$E1,$8F,$CC,$1F,$46
	.dc.b	$FC,$C7,$FC,$A7,$F9,$8B,$F2,$17,$E4,$DF,$D7,$B0,$E1,$7B,$FB,$75
	.dc.b	$EF,$F2,$6D,$65,$F0,$38,$3F,$94,$61,$C1,$B1,$E0,$CF,$BF,$77,$F6
	.dc.b	$FA,$EB,$1E,$2E,$CC,$78,$9E,$5F,$BF,$FF,$3F,$F1,$3F,$33,$7E,$9D
	.dc.b	$E5,$AD,$7E,$09,$FE,$C6,$9F,$F0,$5F,$27,$C0,$97,$E6,$9F,$E0,$7D
	.dc.b	$3F,$D8,$DA,$59,$E3,$97,$F2,$0B,$E1,$3F,$E9,$4C,$7F,$DA,$3E,$67
	.dc.b	$F8,$17,$D9,$FF,$5A,$FE,$6F,$FE,$EE,$FC,$33,$FD,$6F,$F9,$BF,$F5
	.dc.b	$2C,$7F,$C3,$95,$AE,$15,$6B,$DE,$FF,$04,$CF,$2E,$5F,$E9,$CF,$FB
	.dc.b	$03,$F3,$7F,$F0,$E3,$C6,$BF,$D8,$7F,$87,$78,$7E,$D1,$A7,$85,$EB
	.dc.b	$87,$F8,$53,$F1,$67,$E6,$7F,$D3,$FF,$04,$7E,$72,$FC,$E7,$FC,$01
	.dc.b	$F9,$2E,$EB,$53,$9F,$88,$FF,$2B,$71,$7F,$8D,$71,$E7,$E5,$FF,$8B
	.dc.b	$DE,$BF,$39,$70,$E3,$97,$F7,$7B,$E1,$18,$F8,$BF,$F1,$67,$9B,$FE
	.dc.b	$05,$E9,$7C,$5F,$F6,$3F,$CE,$9F,$60,$F5,$BF,$C5,$B7,$BA,$D4,$BF
	.dc.b	$2B,$CF,$D3,$FF,$AA,$B9,$7F,$BB,$5F,$56,$C7,$FB,$1A,$FB,$74,$78
	.dc.b	$A3,$1C,$0C,$FF,$6D,$7F,$E3,$CF,$81,$FF,$6B,$7D,$D7,$E2,$3E,$17
	.dc.b	$FE,$00,$7E,$07,$9F,$9A,$4E,$7F,$1A,$BF,$29,$9F,$CF,$3F,$CA,$98
	.dc.b	$F2,$CB,$FC,$14,$F4,$F3,$3F,$E0,$0D,$FF,$BD,$7A,$3F,$CC,$53,$9D
	.dc.b	$12,$FC,$F3,$FC,$A9,$F9,$BB,$F2,$97,$E7,$BF,$B7,$FC,$0F,$FD,$59
	.dc.b	$E0,$FF,$A3,$33,$FF,$36,$5E,$EB,$5C,$5F,$E3,$FF,$47,$FC,$B3,$39
	.dc.b	$DF,$FC,$3C,$F5,$F2,$3F,$CD,$8F,$1F,$71,$9F,$FA,$E7,$DE,$F1,$EB
	.dc.b	$E2,$9C,$E4,$F0,$73,$A7,$83,$67,$E2,$5E,$0C,$38,$65,$E8,$F4,$F1
	.dc.b	$7B,$F1,$BF,$CD,$BF,$A0,$3E,$17,$F9,$5F,$E7,$FF,$5F,$F2,$07,$E4
	.dc.b	$4F,$D0,$37,$5A,$F0,$55,$AF,$B0,$CF,$C3,$FF,$A1,$F2,$E9,$F8,$F5
	.dc.b	$C3,$85,$F0,$E9,$DE,$9D,$FF,$20,$FF,$42,$61,$E3,$C5,$3E,$3D,$5F
	.dc.b	$A1,$70,$D7,$C3,$0F,$D0,$BF,$D0,$FE,$BB,$F0,$5F,$D9,$F9,$4F,$D0
	.dc.b	$9F,$A2,$BF,$45,$7E,$C5,$C2,$F7,$5A,$97,$E8,$D9,$EB,$FF,$51,$65
	.dc.b	$D7,$DB,$B8,$70,$C9,$F9,$17,$FA,$76,$B5,$A7,$86,$7F,$59,$CB,$DF
	.dc.b	$0F,$F2,$B6,$9D,$F8,$B1,$7B,$D7,$FC,$A7,$E4,$7F,$AE,$7B,$FF,$DF
	.dc.b	$5F,$A5,$3F,$4B,$7E,$94,$FC,$E7,$CF,$FA,$57,$FA,$A2,$F7,$5A,$9F
	.dc.b	$5F,$D7,$E7,$F0,$D6,$3F,$EB,$DB,$43,$F2,$4C,$3A,$C3,$97,$D8,$7F
	.dc.b	$AE,$6B,$AE,$7E,$1F,$8C,$C5,$EE,$3C,$9D,$2F,$D2,$FC,$19,$FA,$67
	.dc.b	$3F,$87,$E3,$3F,$35,$BF,$4C,$C5,$72,$F5,$3F,$D8,$71,$FE,$18,$F4
	.dc.b	$1F,$94,$BF,$1F,$FE,$D7,$FA,$6F,$C4,$E6,$FF,$70,$5E,$EB,$5F,$40
	.dc.b	$A7,$CD,$74,$FA,$8F,$D6,$77,$FF,$A4,$F0,$FF,$12,$FE,$33,$FE,$E1
	.dc.b	$79,$67,$FD,$E3,$E2,$BF,$DB,$51,$8C,$BF,$03,$7C,$F5,$F2,$DF,$93
	.dc.b	$F0,$8C,$6B,$E2,$D5,$FA,$BF,$FB,$7B,$F3,$37,$6F,$F7,$0F,$D3,$DF
	.dc.b	$9D,$FF,$9D,$BE,$51,$FA,$86,$EB,$53,$9F,$CC,$B1,$F4,$7F,$EF,$7C
	.dc.b	$79,$FE,$47,$6F,$7F,$81,$F9,$BA,$5F,$3C,$E4,$98,$78,$B1,$6D,$E2
	.dc.b	$C3,$F3,$38,$70,$BD,$E3,$D3,$67,$87,$E2,$C3,$84,$3F,$38,$FF,$7D
	.dc.b	$FA,$EF,$CC,$3F,$D6,$1F,$4E,$FC,$B9,$F6,$9F,$EC,$BF,$D4,$77,$5A
	.dc.b	$9C,$F5,$7F,$37,$DF,$C5,$BF,$F4,$DA,$FF,$5D,$3F,$EC,$BC,$BB,$7E
	.dc.b	$C1,$ED,$88,$74,$FC,$3B,$FC,$3D,$8F,$83,$F1,$8F,$85,$8F,$4C,$3C
	.dc.b	$D5,$AD,$3F,$D9,$98,$71,$D7,$D1,$1F,$EE,$1F,$D4,$F1,$97,$9B,$BA
	.dc.b	$CF,$FB,$97,$E6,$6E,$EF,$D9,$1A,$FA,$BC,$A7,$EA,$BF,$BB,$FA,$A7
	.dc.b	$EA,$CF,$D0,$1F,$5B,$FC,$13,$C0,$F6,$61,$C3,$D9,$FF,$0B,$DE,$EB
	.dc.b	$F9,$4E,$5D,$72,$FE,$0D,$C7,$FE,$EF,$C3,$BF,$E9,$1F,$12,$3E,$0E
	.dc.b	$5D,$2B,$C5,$8F,$87,$E0,$F3,$5F,$87,$C3,$C1,$AF,$9B,$5F,$ED,$53
	.dc.b	$EC,$C7,$96,$BE,$49,$F4,$FE,$65,$9A,$E5,$9F,$69,$F6,$FE,$65,$F1
	.dc.b	$3E,$5F,$EC,$7E,$3C,$FD,$65,$FA,$D3,$F5,$9F,$D4,$FE,$59,$FC,$07
	.dc.b	$F5,$AE,$1A,$7E,$6F,$0E,$17,$BF,$45,$7F,$C1,$F9,$B7,$6B,$DB,$F0
	.dc.b	$E3,$FB,$5B,$CF,$2F,$F8,$02,$E7,$3F,$2C,$79,$33,$F1,$FD,$AF,$E2
	.dc.b	$A7,$39,$F8,$72,$7C,$26,$71,$E9,$39,$F5,$5F,$92,$79,$35,$F4,$E7
	.dc.b	$EE,$97,$E9,$9E,$8F,$D7,$1F,$AE,$BF,$41,$FE,$77,$F8,$97,$C7,$FD
	.dc.b	$77,$FD,$9B,$C7,$EA,$D9,$87,$0B,$DF,$1A,$DF,$6F,$D7,$2F,$F2,$D6
	.dc.b	$1E,$F3,$F8,$36,$1C,$5C,$1D,$23,$C5,$F1,$3F,$EE,$1C,$1F,$A8,$FB
	.dc.b	$EB,$59,$78,$3F,$37,$C6,$3E,$3C,$38,$67,$FE,$EC,$D3,$D1,$39,$D7
	.dc.b	$D9,$F9,$8F,$97,$1F,$4D,$BF,$01,$C6,$3C,$B8,$FE,$EF,$73,$F4,$57
	.dc.b	$E4,$4F,$D3,$9F,$48,$FD,$0B,$FD,$67,$F0,$7C,$2F,$75,$A9,$FF,$5B
	.dc.b	$FC,$16,$70,$FD,$47,$26,$FF,$1F,$BC,$9F,$3F,$4F,$05,$77,$D3,$B7
	.dc.b	$DC,$FC,$39,$CE,$7D,$7F,$17,$C6,$BE,$0B,$F2,$7C,$63,$F5,$EF,$05
	.dc.b	$F8,$E1,$FA,$7F,$F0,$AF,$B3,$FE,$3D,$FA,$EB,$F1,$3F,$EB,$DF,$36
	.dc.b	$FD,$7F,$FA,$7E,$6F,$D0,$B8,$72,$7F,$41,$C3,$87,$C9,$FF,$77,$BD
	.dc.b	$FD,$C6,$A9,$5D,$7D,$AF,$7E,$9E,$0D,$3F,$B6,$57,$EA,$63,$18,$F3
	.dc.b	$F4,$5E,$27,$F2,$E9,$CE,$DE,$59,$FC,$3D,$F7,$CB,$C7,$39,$D9,$ED
	.dc.b	$E7,$3F,$6E,$DE,$CF,$5F,$CD,$F6,$4F,$C3,$3F,$3F,$FB,$13,$F6,$27
	.dc.b	$EA,$BF,$ED,$6F,$D8,$B7,$5A,$9C,$FE,$0C,$F8,$EC,$F9,$DE,$EE,$2F
	.dc.b	$EC,$D9,$F5,$FC,$A7,$11,$FE,$62,$C7,$96,$7E,$AF,$11,$CD,$FD,$47
	.dc.b	$0F,$26,$2D,$BC,$97,$F1,$31,$EF,$18,$D1,$F2,$18,$DB,$CB,$F1,$F7
	.dc.b	$0F,$F7,$7F,$A9,$6D,$F9,$A7,$F4,$CF,$BF,$FE,$C1,$E7,$BF,$62,$FF
	.dc.b	$4D,$FB,$7F,$ED,$6F,$CD,$FF,$CC,$3F,$B4,$EF,$83,$FC,$05,$7B,$FC
	.dc.b	$0F,$F9,$77,$DA,$AF,$FC,$CF,$FB,$27,$F5,$8F,$C6,$FF,$B6,$7E,$C1
	.dc.b	$FD,$E7,$F1,$97,$E5,$1F,$27,$F3,$97,$D6,$BF,$61,$E1,$B7,$F8,$3F
	.dc.b	$0E,$17,$BA,$7E,$71,$AA,$57,$F8,$17,$F4,$1F,$D6,$E7,$E0,$F9,$E6
	.dc.b	$98,$BE,$91,$8C,$BC,$5F,$5D,$FD,$F2,$F8,$5F,$9B,$C5,$6D,$FE,$EB
	.dc.b	$9C,$DF,$8F,$F6,$5F,$16,$9E,$6B,$DE,$CF,$CC,$FF,$BC,$FB,$7F,$E0
	.dc.b	$DF,$C5,$9E,$4B,$F2,$F7,$E7,$CF,$8B,$FF,$00,$7E,$36,$FA,$7E,$3C
	.dc.b	$BF,$E3,$1C,$78,$E1,$C2,$F7,$E4,$FF,$81,$BF,$4E,$FE,$FF,$FB,$33
	.dc.b	$F6,$7F,$CE,$7E,$C9,$FE,$83,$FD,$4D,$FA,$5E,$F4,$FE,$ED,$7B,$F3
	.dc.b	$97,$BF,$D5,$3F,$C2,$3F,$29,$FE,$25,$F9,$37,$E5,$3F,$E4,$4F,$C1
	.dc.b	$9E,$77,$DA,$FC,$9B,$FD,$09,$EC,$B0,$E1,$7B,$AD,$7E,$6F,$D3,$6F
	.dc.b	$F2,$66,$5D,$3E,$A3,$FD,$D9,$5F,$84,$F8,$6F,$F1,$7E,$3C,$A7,$3D
	.dc.b	$EF,$C8,$3F,$C6,$7E,$17,$F8,$E7,$F2,$B7,$DA,$3F,$5F,$7E,$43,$FE
	.dc.b	$64,$FA,$B7,$E5,$7F,$E0,$7E,$3F,$C4,$F0,$FF,$97,$71,$E3,$87,$0F
	.dc.b	$C1,$18,$70,$F1,$7F,$C6,$7A,$77,$D3,$B6,$9D,$7F,$06,$F1,$E7,$CE
	.dc.b	$73,$8F,$2C,$3F,$C9,$3F,$9F,$FE,$FF,$E2,$7E,$9F,$D3,$3F,$2A,$FB
	.dc.b	$1F,$91,$F9,$6B,$5E,$17,$F7,$09,$CE,$1F,$E4,$49,$CD,$F1,$F1,$3F
	.dc.b	$D2,$BC,$69,$CD,$FE,$0E,$FF,$CB,$3F,$B3,$7F,$97,$FE,$9B,$FC,$D7
	.dc.b	$F2,$6F,$DA,$3F,$BF,$7C,$83,$F3,$C7,$E7,$FF,$E5,$2C,$38,$7D,$FF
	.dc.b	$FA,$DE,$F7,$4B,$D7,$2F,$FA,$1F,$2E,$B5,$D3,$F2,$BF,$0C,$E7,$0E
	.dc.b	$5E,$27,$1F,$BC,$FF,$36,$FC,$D7,$F9,$CF,$67,$ED,$0C,$5F,$B4,$BE
	.dc.b	$25,$FA,$3B,$F6,$7F,$F7,$07,$3F,$F5,$2D,$BF,$D3,$19,$72,$C7,$8E
	.dc.b	$1C,$2F,$FD,$33,$7B,$BF,$F6,$96,$5D,$F2,$ED,$93,$F3,$6F,$2D,$74
	.dc.b	$AD,$63,$E2,$F7,$FF,$D0,$BF,$91,$FF,$A2,$FF,$2A,$FA,$5E,$4F,$EB
	.dc.b	$3F,$36,$FE,$E1,$FA,$07,$F9,$36,$B5,$DD,$F9,$4F,$5D,$5A,$F6,$BF
	.dc.b	$D0,$35,$AC,$39,$7D,$0B,$F3,$1E,$FB,$FF,$AD,$74,$EF,$97,$2A,$FF
	.dc.b	$21,$CF,$FA,$2B,$E5,$7F,$D3,$DE,$2F,$F8,$1F,$F0,$67,$E6,$EF,$25
	.dc.b	$FB,$2F,$FB,$87,$F6,$5E,$18,$FF,$B0,$70,$E1,$7B,$AF,$F7,$37,$E7
	.dc.b	$9A,$AF,$F2,$2F,$D9,$32,$EB,$97,$4F,$47,$FD,$53,$A7,$8A,$B5,$97
	.dc.b	$3F,$0B,$FD,$5B,$F3,$FF,$EB,$3F,$D3,$BF,$91,$7C,$FB,$F6,$47,$5B
	.dc.b	$95,$F9,$BF,$F8,$23,$C3,$63,$C7,$E3,$BF,$AE,$61,$C2,$F7,$D7,$FE
	.dc.b	$AC,$AD,$65,$D7,$2E,$9F,$37,$E2,$E4,$FF,$70,$6B,$E2,$CB,$97,$E8
	.dc.b	$5F,$EC,$2F,$91,$FF,$64,$6F,$FE,$45,$F9,$5F,$D1,$78,$9F,$D4,$3F
	.dc.b	$19,$7E,$27,$FE,$BA,$BD,$FD,$FA,$F7,$F3,$9F,$EE,$AF,$CF,$35,$B7
	.dc.b	$FB,$BF,$3E,$9E,$B7,$E7,$EB,$FD,$DB,$8F,$2C,$38,$F4,$7F,$B5,$3F
	.dc.b	$42,$FF,$6D,$7D,$A3,$C1,$7E,$47,$FA,$6F,$D9,$DF,$44,$FC,$99,$F8
	.dc.b	$AF,$FB,$3B,$1E,$38,$70,$F5,$18,$70,$B7,$E8,$DB,$E2,$0D,$3C,$1A
	.dc.b	$7F,$97,$B2,$E7,$F0,$D6,$B7,$5A,$CB,$AE,$BE,$39,$E8,$73,$D3,$C5
	.dc.b	$A7,$A3,$F4,$CF,$F7,$8F,$CA,$FF,$BD,$F3,$7E,$67,$FD,$B3,$EA,$7F
	.dc.b	$53,$FB,$53,$F3,$27,$ED,$7F,$EE,$AC,$38,$5E,$FD,$95,$EF,$CE,$D6
	.dc.b	$DE,$CC,$FA,$F5,$7F,$9E,$31,$E3,$F4,$D6,$B5,$97,$3C,$7F,$E0,$0F
	.dc.b	$9A,$BD,$ED,$E4,$FA,$8F,$8F,$2F,$D9,$BE,$4E,$3E,$AF,$C0,$BF,$59
	.dc.b	$F8,$0B,$D9,$7E,$40,$F4,$58,$F1,$C3,$E8,$EC,$C7,$8E,$1C,$3E,$7F
	.dc.b	$79,$F6,$AD,$7B,$1F,$E9,$28,$C5,$BF,$54,$C6,$3D,$23,$9F,$C4,$7A
	.dc.b	$2B,$77,$FC,$1B,$6F,$35,$FB,$3F,$6C,$F5,$F1,$FE,$AF,$D7,$FD,$AE
	.dc.b	$BF,$ED,$8F,$DB,$BD,$35,$AF,$4F,$F6,$9E,$6F,$FA,$03,$F1,$8F,$E7
	.dc.b	$DC,$6A,$D7,$ED,$DF,$EA,$87,$87,$B1,$75,$FC,$8E,$BD,$F1,$FF,$44
	.dc.b	$FE,$31,$E3,$9C,$E7,$DB,$5F,$E5,$11,$F2,$BA,$BE,$8F,$B4,$FC,$E3
	.dc.b	$FC,$0B,$F2,$FF,$2F,$F2,$47,$EA,$CF,$DC,$F8,$63,$FA,$0C,$38,$7C
	.dc.b	$5F,$F6,$6E,$B6,$1C,$3F,$6B,$DD,$FF,$9B,$73,$7E,$8B,$FE,$C2,$D7
	.dc.b	$C3,$1F,$E1,$49,$FC,$19,$F8,$68,$CE,$3D,$70,$F1,$CF,$4C,$BC,$3A
	.dc.b	$F9,$FF,$06,$FC,$6F,$E1,$1F,$8C,$F5,$FF,$93,$7E,$D8,$FC,$BB,$D7
	.dc.b	$FA,$8F,$ED,$FE,$93,$F1,$77,$BF,$01,$7B,$E5,$5F,$EF,$7F,$18,$DE
	.dc.b	$DF,$C0,$33,$EA,$FA,$0C,$9F,$C3,$70,$FC,$B3,$EC,$EF,$65,$DF,$2E
	.dc.b	$D8,$FE,$A2,$31,$A7,$9B,$CB,$FD,$3F,$A9,$FA,$6F,$39,$CB,$FB,$EF
	.dc.b	$53,$F7,$BE,$DB,$EE,$3F,$4B,$7B,$AD,$7A,$DF,$C8,$BF,$52,$CF,$07
	.dc.b	$F1,$8C,$B9,$FC,$A7,$FB,$D3,$0E,$1E,$C5,$AD,$63,$CB,$E9,$BC,$71
	.dc.b	$89,$FE,$13,$AF,$F0,$EF,$92,$FE,$47,$F9,$9B,$F3,$FF,$DE,$FD,$FF
	.dc.b	$F9,$A3,$F3,$8F,$37,$E2,$4F,$DC,$3F,$C3,$5F,$9F,$30,$E1,$A7,$F2
	.dc.b	$CC,$38,$72,$7F,$B1,$AF,$FC,$41,$F4,$DB,$CF,$AF,$DE,$3F,$6D,$FA
	.dc.b	$58,$E4,$E5,$0D,$73,$9F,$69,$FF,$49,$4E,$72,$ED,$F8,$AF,$AB,$1E
	.dc.b	$B8,$7A,$33,$FF,$70,$61,$E7,$EC,$FF,$4E,$7E,$15,$FE,$08,$F9,$DF
	.dc.b	$67,$D5,$FF,$A3,$BE,$27,$57,$FE,$99,$AD,$60,$AF,$F4,$9F,$D5,$EA
	.dc.b	$9F,$83,$FF,$AC,$3F,$6B,$F3,$E7,$DB,$E9,$2D,$73,$7F,$C3,$38,$F3
	.dc.b	$E9,$FA,$35,$BD,$7C,$7E,$6F,$D4,$ED,$FE,$5F,$EB,$FC,$0E,$1F,$67
	.dc.b	$E8,$DF,$46,$73,$D5,$F4,$3B,$3E,$B7,$CF,$67,$7F,$EB,$F8,$FC,$87
	.dc.b	$F4,$1D,$5F,$94,$73,$6F,$B4,$F9,$F8,$F4,$EB,$F9,$D9,$3A,$BC,$98
	.dc.b	$70,$9C,$F9,$1F,$8A,$F8,$0F,$E5,$7F,$91,$BF,$13,$FF,$69,$FE,$B4
	.dc.b	$FC,$AD,$F4,$AF,$D6,$7F,$15,$87,$0B,$DD,$6B,$F0,$FC,$E9,$FC,$A3
	.dc.b	$2E,$8F,$AA,$FB,$FE,$8B,$F8,$67,$97,$84,$D3,$C1,$A7,$96,$F9,$67
	.dc.b	$DB,$E5,$7F,$43,$F1,$FF,$D3,$FF,$06,$7E,$A9,$FC,$33,$ED,$1F,$98
	.dc.b	$79,$F8,$7F,$41,$7B,$AD,$4F,$AF,$D1,$CE,$7F,$D7,$B1,$FF,$1E,$BF
	.dc.b	$42,$F8,$34,$D5,$F2,$E2,$F1,$C6,$3D,$AF,$F5,$27,$EB,$7F,$DC,$3B
	.dc.b	$7F,$BB,$4F,$F7,$FF,$D4,$DF,$B1,$7F,$B2,$7F,$2E,$7C,$5B,$F3,$57
	.dc.b	$EA,$1F,$B7,$E4,$FF,$09,$61,$C3,$F0,$C6,$1C,$2D,$FA,$7E,$F3,$ED
	.dc.b	$9F,$EC,$B1,$E3,$FB,$8D,$9C,$E3,$CB,$4F,$F0,$D4,$71,$1F,$48,$D3
	.dc.b	$C1,$A7,$9B,$CC,$FE,$23,$3F,$E3,$AF,$90,$7E,$E4,$FD,$05,$F6,$2F
	.dc.b	$DC,$DF,$AF,$7F,$11,$C3,$85,$EE,$B5,$F9,$96,$75,$FF,$1C,$FD,$63
	.dc.b	$E8,$FF,$85,$E1,$C3,$3E,$BF,$9E,$78,$AF,$7A,$F8,$A7,$C3,$F9,$F7
	.dc.b	$F2,$6C,$FF,$CC,$7F,$83,$7F,$19,$F9,$EF,$F7,$17,$E2,$EF,$C6,$3F
	.dc.b	$92,$56,$BE,$23,$FB,$97,$E2,$3F,$F0,$07,$F5,$8C,$EB,$FE,$67,$FB
	.dc.b	$E6,$7C,$E9,$97,$7A,$D4,$E7,$F3,$8F,$E6,$FF,$82,$7F,$A0,$BF,$74
	.dc.b	$FD,$77,$A7,$F9,$FF,$37,$F9,$A4,$E7,$D1,$FD,$5F,$91,$F9,$2C,$BF
	.dc.b	$63,$F8,$33,$E8,$D9,$72,$D3,$AD,$7F,$A2,$AB,$53,$DB,$F5,$CF,$E8
	.dc.b	$FF,$73,$FE,$A3,$FC,$25,$FB,$2F,$EA,$3D,$0F,$CB,$E8,$AD,$4E,$7D
	.dc.b	$7F,$D5,$CE,$73,$FF,$5B,$7E,$15,$FE,$B1,$AE,$5C,$9D,$A7,$39,$76
	.dc.b	$F5,$3F,$BF,$F2,$FF,$58,$FB,$AF,$F6,$77,$E1,$8F,$C0,$5F,$BA,$FF
	.dc.b	$11,$FA,$8F,$EA,$D5,$A9,$CF,$3B,$F2,$EC,$E9,$FE,$D3,$E1,$FF,$6E
	.dc.b	$57,$2F,$53,$B4,$E7,$EB,$DA,$77,$D3,$F8,$DC,$FF,$63,$F9,$9F,$F3
	.dc.b	$2F,$CE,$FF,$BE,$7F,$13,$7E,$AD,$F9,$CF,$AC,$7E,$05,$BC,$3F,$DE
	.dc.b	$9F,$7C,$BD,$FA,$9F,$EC,$7F,$91,$D5,$7F,$BF,$AB,$58,$F3,$8C,$E5
	.dc.b	$DE,$3F,$53,$F8,$7F,$F7,$CF,$93,$FF,$3B,$FE,$30,$FC,$5D,$6F,$C6
	.dc.b	$3F,$D1,$FF,$37,$FA,$CC,$BF,$BE,$61,$C2,$F7,$5A,$FC,$85,$E8,$6A
	.dc.b	$FF,$D5,$79,$75,$CB,$A7,$D7,$FF,$79,$EF,$F1,$56,$B2,$E7,$3F,$DF
	.dc.b	$7E,$2F,$FC,$23,$D9,$FE,$8C,$F9,$8F,$B5,$F8,$DB,$F1,$57,$A8,$F9
	.dc.b	$97,$EB,$FF,$EF,$7F,$31,$8F,$1F,$0D,$8F,$1C,$38,$5B,$F5,$3F,$F0
	.dc.b	$C5,$EE,$B5,$8B,$4F,$1E,$FF,$EC,$6D,$7C,$7A,$F8,$A9,$C1,$FE,$28
	.dc.b	$FC,$87,$FD,$47,$F9,$1B,$F2,$17,$C7,$3F,$22,$7E,$44,$FD,$17,$F8
	.dc.b	$CF,$D2,$7F,$8A,$71,$E3,$87,$0B,$DD,$7F,$1E,$FC,$B7,$5B,$FF,$B6
	.dc.b	$33,$ED,$9F,$5C,$FA,$69,$E3,$AD,$65,$CE,$7C,$1C,$1F,$E3,$EE,$DF
	.dc.b	$F5,$CF,$E4,$6F,$A5,$FE,$31,$F9,$43,$F2,$77,$CF,$BF,$42,$7E,$9B
	.dc.b	$FE,$3F,$C7,$8E,$1C,$2F,$74,$F2,$B5,$B7,$FB,$BB,$3E,$DF,$16,$D3
	.dc.b	$B6,$8F,$2B,$FC,$A3,$F9,$6F,$FB,$33,$DC,$FF,$32,$7E,$0E,$FC,$B7
	.dc.b	$F5,$FF,$AA,$BE,$51,$F9,$3F,$F9,$47,$0E,$1F,$7A,$C3,$87,$B6,$C3
	.dc.b	$87,$EA,$DF,$E5,$7E,$8F,$F6,$CF,$DD,$BF,$04,$78,$8F,$91,$7E,$63
	.dc.b	$FE,$A4,$F9,$6F,$F2,$07,$A3,$F8,$5E,$91,$FE,$40,$C3,$8D,$F0,$FC
	.dc.b	$3B,$FA,$BF,$E5,$7F,$E6,$5F,$B5,$FF,$73,$FD,$73,$F0,$57,$E7,$8F
	.dc.b	$DB,$3E,$27,$C3,$BF,$4B,$7E,$4F,$C7,$6E,$F8,$F1,$C3,$87,$83,$FE
	.dc.b	$53,$B7,$E9,$DB,$BF,$F7,$B6,$6F,$DB,$BF,$CD,$3A,$F8,$A3,$C3,$3F
	.dc.b	$E4,$28,$C6,$3D,$34,$7D,$1F,$FA,$12,$73,$B7,$9B,$F1,$9C,$61,$E4
	.dc.b	$C7,$D7,$3E,$AD,$3F,$D0,$DC,$9F,$E3,$DA,$D6,$7D,$67,$FC,$95,$3F
	.dc.b	$8F,$9F,$97,$F2,$5F,$D4,$FE,$49,$F9,$E3,$F0,$17,$E7,$8F,$D6,$FE
	.dc.b	$BE,$3C,$71,$FF,$27,$7D,$1F,$F3,$BC,$38,$5E,$FE,$0B,$FD,$2D,$E9
	.dc.b	$7D,$8B,$FE,$A9,$FB,$B3,$F4,$0F,$F0,$FF,$EC,$1F,$E2,$6E,$37,$EE
	.dc.b	$9C,$2F,$EF,$FD,$13,$0E,$1F,$76,$C3,$87,$ED,$1B,$BE,$B9,$7F,$A6
	.dc.b	$33,$F0,$63,$C7,$F4,$3C,$57,$4C,$B9,$E3,$FE,$36,$FA,$AF,$F5,$9E
	.dc.b	$1C,$27,$3F,$7E,$FC,$EB,$38,$71,$CF,$D3,$3F,$8D,$F9,$EF,$16,$9E
	.dc.b	$79,$CF,$07,$FB,$4B,$97,$E4,$4E,$78,$3D,$58,$7E,$57,$C6,$E3,$A7
	.dc.b	$83,$0F,$E0,$D8,$3F,$09,$F9,$5F,$86,$FF,$93,$3F,$1D,$7E,$1E,$F9
	.dc.b	$77,$EB,$9F,$E5,$2F,$D8,$FF,$D8,$D8,$70,$D5,$87,$0B,$DF,$E0,$6F
	.dc.b	$C9,$B7,$C9,$FA,$ED,$3F,$CB,$3E,$4F,$FB,$0B,$2E,$58,$DA,$D6,$5D
	.dc.b	$35,$F1,$6A,$FC,$0B,$FD,$BB,$87,$09,$CF,$E4,$58,$C3,$8E,$3E,$3C
	.dc.b	$3F,$CD,$3F,$63,$E2,$9F,$F8,$03,$97,$5D,$FC,$FF,$B2,$7F,$B8,$7F
	.dc.b	$55,$FC,$67,$E5,$2F,$D5,$1F,$9D,$3F,$05,$7E,$74,$FC,$E3,$FC,$A5
	.dc.b	$E2,$F1,$D7,$F8,$5E,$3C,$70,$E1,$7B,$FC,$1F,$F2,$DD,$6D,$3B,$E9
	.dc.b	$DB,$1F,$F5,$DC,$63,$F2,$AF,$F7,$3B,$AC,$79,$E8,$FC,$8F,$FD,$FF
	.dc.b	$39,$DB,$CB,$3F,$C4,$30,$F2,$0D,$7C,$93,$AD,$3F,$C1,$1A,$78,$EB
	.dc.b	$D3,$A7,$C0,$AD,$7E,$5B,$CB,$ED,$3F,$CF,$5F,$2A,$F8,$E7,$EC,$3F
	.dc.b	$C7,$BD,$CF,$F7,$4E,$7C,$F3,$FF,$38,$5E,$EB,$5E,$87,$3F,$C2,$7F
	.dc.b	$1D,$5F,$00,$FE,$67,$97,$67,$07,$2C,$BB,$FC,$1F,$F3,$9C,$BB,$57
	.dc.b	$92,$7C,$7B,$7F,$AB,$78,$BF,$83,$4E,$73,$F3,$7E,$3D,$E1,$D7,$C9
	.dc.b	$97,$FB,$67,$F7,$7C,$65,$DB,$7F,$EB,$1A,$76,$D9,$87,$09,$7E,$DD
	.dc.b	$F3,$BF,$08,$FF,$39,$7E,$1A,$F0,$DF,$AD,$3F,$4D,$7E,$7C,$FD,$63
	.dc.b	$87,$BE,$C3,$85,$EE,$BF,$C4,$7B,$7C,$BC,$BA,$F3,$FC,$47,$FD,$01
	.dc.b	$D7,$F8,$68,$FE,$B7,$1E,$2D,$3C,$3A,$7E,$0E,$F9,$67,$DE,$73,$F4
	.dc.b	$3F,$8A,$C3,$8C,$E7,$3F,$17,$E2,$9E,$6D,$7C,$B9,$7A,$BE,$95,$9F
	.dc.b	$FA,$9F,$1F,$ED,$7E,$67,$F8,$9B,$CA,$7E,$61,$FD,$D3,$F5,$1D,$FE
	.dc.b	$BD,$C3,$E9,$BF,$9D,$5E,$EB,$58,$F0,$3A,$63,$FC,$53,$EA,$BF,$D2
	.dc.b	$AF,$F4,$13,$FB,$CC,$6A,$BF,$C6,$BF,$5D,$99,$CE,$1D,$AF,$FD,$3B
	.dc.b	$87,$8F,$F2,$FC,$D7,$A2,$73,$F3,$8D,$1F,$85,$EA,$BC,$5B,$7B,$2B
	.dc.b	$5A,$BF,$61,$F6,$57,$3A,$D6,$1F,$1B,$1C,$CB,$F6,$4F,$8B,$F3,$DF
	.dc.b	$EE,$6F,$D2,$1F,$9D,$3F,$4A,$7E,$95,$C2,$F7,$E5,$AF,$75,$F8,$ED
	.dc.b	$BE,$3D,$7D,$0F,$D7,$3E,$16,$7C,$78,$FF,$DE,$2F,$C5,$AD,$C7,$FB
	.dc.b	$D3,$1E,$99,$FF,$90,$6B,$86,$9E,$58,$FC,$57,$27,$F7,$1C,$BC,$7A
	.dc.b	$78,$BE,$B7,$78,$F2,$D7,$FD,$FD,$A7,$6C,$B9,$47,$F8,$06,$31,$3B
	.dc.b	$FD,$8F,$E2,$FE,$50,$FD,$21,$FB,$4B,$E8,$7F,$C8,$FF,$66,$FD,$15
	.dc.b	$FB,$C7,$0B,$DD,$6A,$7E,$76,$7F,$8A,$C7,$A7,$DD,$FC,$48,$C5,$F6
	.dc.b	$E4,$F8,$9C,$7F,$E0,$DB,$F1,$69,$E1,$FB,$56,$AF,$D2,$BF,$09,$5A
	.dc.b	$D7,$CB,$F5,$B8,$CF,$C7,$87,$0D,$9F,$BC,$F9,$F3,$E9,$9F,$AF,$E5
	.dc.b	$3F,$0B,$C6,$F8,$A7,$E9,$8F,$D0,$13,$FD,$DB,$E6,$DE,$57,$0B,$DD
	.dc.b	$6B,$CA,$FF,$6A,$F8,$19,$FA,$DF,$81,$CD,$FC,$92,$FA,$F3,$FC,$03
	.dc.b	$F9,$8B,$CB,$3F,$F0,$1F,$E6,$5F,$A1,$7C,$9A,$F8,$DF,$F0,$3F,$97
	.dc.b	$9A,$E1,$9F,$FB,$93,$C6,$F2,$5E,$E3,$FD,$CF,$E4,$FE,$8B,$EB,$1F
	.dc.b	$B7,$BE,$AD,$F8,$3B,$F3,$8F,$E9,$DF,$95,$3F,$63,$E1,$7B,$AD,$7C
	.dc.b	$22,$BF,$BA,$78,$9F,$C2,$32,$EB,$F9,$C7,$C0,$E4,$FF,$3F,$C7,$83
	.dc.b	$4C,$47,$F9,$E7,$2E,$9A,$7F,$5D,$F7,$7F,$80,$5E,$F4,$EF,$F3,$78
	.dc.b	$BD,$EB,$E1,$FC,$43,$CB,$9F,$AB,$2F,$4F,$7E,$31,$F3,$EF,$FE,$63
	.dc.b	$DF,$CD,$97,$28,$FE,$ED,$1F,$C8,$30,$FF,$7A,$F0,$BF,$5E,$7E,$CC
	.dc.b	$F3,$DF,$92,$FF,$A4,$F0,$E1,$F8,$D7,$F1,$3B,$DD,$6B,$F4,$4F,$E2
	.dc.b	$39,$74,$E5,$FC,$01,$F4,$6D,$F8,$CF,$C4,$9F,$0E,$1C,$BD,$8F,$E4
	.dc.b	$51,$8D,$3C,$5C,$58,$C3,$C1,$1E,$5D,$5F,$96,$F8,$27,$3A,$FA,$3F
	.dc.b	$42,$FE,$47,$F1,$0F,$C2,$7F,$C6,$1E,$F3,$F2,$F7,$E3,$4F,$C2,$9F
	.dc.b	$A6,$F0,$BD,$D6,$A7,$FC,$F5,$CD,$FD,$23,$1E,$9F,$35,$F7,$9E,$DF
	.dc.b	$B3,$5C,$38,$E4,$FC,$53,$F9,$F5,$70,$D3,$C5,$2F,$CB,$33,$7C,$33
	.dc.b	$7E,$D9,$E3,$CF,$FA,$46,$BE,$3E,$4F,$E7,$FF,$69,$FD,$2F,$E0,$3F
	.dc.b	$DD,$1F,$56,$FD,$27,$FD,$93,$F2,$9A,$DD,$5A,$9F,$F6,$6F,$EA,$F9
	.dc.b	$FD,$4F,$F2,$F8,$F3,$F0,$7E,$CB,$D7,$F0,$AF,$F3,$C5,$FF,$9E,$A7
	.dc.b	$A6,$1C,$FF,$1E,$FE,$B5,$7B,$D3,$C5,$F8,$4B,$F2,$7D,$E5,$CA,$3F
	.dc.b	$D0,$FE,$D7,$FA,$BE,$F8,$67,$FE,$E0,$F9,$ED,$E3,$CB,$F0,$4F,$EB
	.dc.b	$BF,$8E,$FF,$91,$3E,$2B,$EC,$FE,$E9,$FE,$8A,$FC,$A1,$F8,$3E,$BE
	.dc.b	$AF,$F9,$E7,$7A,$B5,$F5,$B9,$DB,$F6,$B9,$73,$E5,$FC,$BB,$D6,$F8
	.dc.b	$7D,$C7,$F8,$BE,$7A,$69,$E0,$8F,$F4,$44,$62,$77,$A7,$F6,$9F,$17
	.dc.b	$FC,$1B,$87,$3C,$FC,$DF,$B4,$FF,$76,$B7,$63,$F4,$97,$E1,$DF,$97
	.dc.b	$F2,$FF,$31,$F2,$FB,$AD,$4E,$7C,$67,$ED,$59,$FD,$AB,$F8,$7E,$5D
	.dc.b	$32,$B8,$70,$BF,$DE,$7E,$69,$FE,$0F,$AD,$63,$DF,$D5,$C5,$EF,$4F
	.dc.b	$D4,$7C,$B7,$87,$2F,$2E,$5F,$C1,$47,$EE,$9F,$E1,$0F,$D7,$FF,$D6
	.dc.b	$FF,$85,$FF,$2E,$FC,$B3,$FB,$6F,$E3,$2F,$DE,$B5,$F8,$2F,$FB,$AF
	.dc.b	$E1,$15,$AF,$8F,$CF,$E0,$9F,$CA,$32,$E7,$DB,$FC,$29,$EB,$7C,$3E
	.dc.b	$6F,$FD,$ED,$DB,$FE,$3A,$9C,$E3,$FD,$F3,$EC,$BF,$C5,$91,$8A,$6B
	.dc.b	$FE,$2B,$BE,$1A,$78,$F0,$F3,$FC,$27,$9B,$7F,$56,$1C,$3F,$2A,$FF
	.dc.b	$16,$FD,$EB,$F6,$37,$9C,$FD,$ED,$F9,$6B,$8B,$FC,$7D,$F7,$9C,$2F
	.dc.b	$75,$A9,$FF,$21,$79,$F9,$F8,$5F,$E8,$F9,$75,$CA,$E1,$C3,$2E,$9F
	.dc.b	$4D,$FE,$47,$AD,$65,$CF,$F0,$5C,$5E,$F3,$FF,$27,$7E,$25,$FE,$44
	.dc.b	$CD,$FA,$9B,$F1,$E7,$EE,$8F,$CC,$9F,$AC,$7F,$A2,$FE,$EF,$85,$EE
	.dc.b	$B5,$3F,$E8,$FF,$8B,$CE,$1F,$D7,$B1,$EB,$F8,$5D,$C3,$86,$5D,$3E
	.dc.b	$D7,$FC,$BD,$5E,$2B,$F0,$CF,$83,$F0,$FC,$F8,$EF,$6B,$E5,$FC,$0B
	.dc.b	$F7,$27,$E2,$3F,$A8,$F0,$FF,$CF,$39,$FF,$3D,$BD,$D6,$BE,$23,$E0
	.dc.b	$62,$FC,$9B,$CC,$F5,$E6,$FE,$AB,$87,$1E,$C6,$8F,$C6,$7F,$1B,$DD
	.dc.b	$F6,$ED,$7B,$AF,$27,$E4,$FE,$8D,$7C,$3B,$BF,$3E,$FF,$39,$E2,$FC
	.dc.b	$83,$FC,$DF,$FA,$2B,$F4,$F7,$EC,$CF,$D1,$1F,$5D,$C2,$F7,$5A,$97
	.dc.b	$C7,$67,$F0,$CF,$F0,$6E,$5D,$7F,$05,$B8,$70,$CB,$A7,$17,$39,$69
	.dc.b	$E1,$97,$E4,$7F,$E9,$1C,$39,$4E,$73,$7E,$9D,$FB,$DF,$D3,$1F,$AF
	.dc.b	$BF,$68,$7E,$CF,$FE,$D2,$F9,$27,$E6,$5B,$AD,$4E,$7C,$9C,$70,$7E
	.dc.b	$EF,$1F,$84,$DA,$DF,$13,$E2,$7F,$73,$5E,$0C,$FB,$CB,$E1,$F3,$7C
	.dc.b	$34,$FF,$0D,$78,$9F,$EB,$CF,$B2,$7E,$7C,$FB,$E7,$E3,$5F,$E8,$0F
	.dc.b	$DC,$DF,$8D,$BF,$35,$FF,$54,$E1,$C2,$F7,$EF,$2F,$7F,$B5,$EB,$27
	.dc.b	$E3,$DF,$E3,$EC,$38,$72,$7F,$88,$1E,$5A,$77,$FD,$1F,$FD,$7D,$18
	.dc.b	$CB,$A7,$E7,$58,$C7,$CD,$7E,$5D,$BF,$DF,$1B,$FA,$E7,$37,$E3,$97
	.dc.b	$E3,$BE,$3D,$FC,$FB,$FB,$3E,$AF,$19,$73,$FB,$0F,$F6,$2F,$C6,$BF
	.dc.b	$4A,$70,$3E,$0F,$E2,$F3,$3F,$68,$FE,$39,$C3,$FE,$E6,$C3,$85,$EF
	.dc.b	$25,$EF,$F0,$4E,$B9,$7C,$CE,$5F,$F2,$0E,$1C,$1B,$1E,$0C,$FB,$F9
	.dc.b	$9F,$EE,$1A,$EB,$1E,$2F,$C0,$71,$5B,$CB,$BE,$BF,$C6,$75,$F2,$4E
	.dc.b	$74,$F4,$79,$9F,$EE,$2F,$CF,$DF,$A0,$3F,$39,$7E,$44,$C7,$F6,$5C
	.dc.b	$9F,$EB,$AF,$33,$F7,$D8,$70,$BD,$D6,$BF,$18,$E9,$8B,$F4,$1F,$F3
	.dc.b	$ED,$EF,$93,$FC,$BE,$F1,$CF,$B7,$A7,$FE,$DC,$9F,$E8,$93,$9C,$7F
	.dc.b	$81,$7C,$5F,$FC,$01,$8C,$53,$57,$E9,$BE,$0D,$7C,$73,$9F,$D6,$BF
	.dc.b	$DF,$FF,$23,$FE,$0A,$F5,$DF,$84,$BF,$64,$7E,$D2,$FD,$3D,$FB,$97
	.dc.b	$0F,$A6,$FE,$29,$7B,$AD,$7D,$36,$7E,$2B,$FD,$33,$9F,$5E,$4F,$F3
	.dc.b	$DB,$DE,$F8,$71,$FF,$5F,$E4,$FE,$C7,$5F,$D8,$6B,$5A,$F8,$BE,$5F
	.dc.b	$5D,$5F,$CA,$E7,$3A,$79,$7F,$17,$F2,$70,$FA,$B0,$F4,$F0,$FF,$9E
	.dc.b	$73,$EB,$9B,$1E,$33,$FE,$BD,$9C,$C6,$3F,$41,$FB,$D9,$FF,$39,$F6
	.dc.b	$5F,$BB,$BF,$7C,$FE,$6D,$F2,$6F,$BC,$F9,$FF,$19,$C3,$5F,$EF,$58
	.dc.b	$70,$BD,$D3,$A7,$D7,$5A,$C5,$F9,$97,$FB,$03,$0E,$13,$FE,$A2,$79
	.dc.b	$7C,$4B,$3F,$07,$E9,$7C,$AB,$F9,$06,$1E,$5A,$D6,$5E,$59,$F2,$1A
	.dc.b	$CF,$C7,$DE,$FC,$FF,$39,$F6,$CF,$D9,$2F,$A7,$E3,$F9,$D3,$F6,$B7
	.dc.b	$E4,$CF,$C8,$1F,$4B,$FE,$C0,$FC,$CF,$F5,$3D,$3F,$43,$0E,$17,$BA
	.dc.b	$D7,$D6,$F4,$C7,$9F,$07,$FB,$02,$FB,$7E,$B9,$FE,$E2,$78,$E7,$DB
	.dc.b	$F1,$27,$E6,$1F,$06,$73,$97,$FB,$43,$E8,$31,$97,$9B,$0E,$1B,$7F
	.dc.b	$8E,$B6,$F2,$4E,$6F,$D1,$F0,$7F,$1B,$F7,$D7,$E9,$AF,$BD,$7E,$3F
	.dc.b	$FE,$D2,$FA,$67,$E9,$8F,$D3,$F8,$5E,$EB,$52,$FC,$7F,$38,$7F,$BF
	.dc.b	$31,$EB,$FA,$45,$C3,$87,$DA,$FF,$8B,$79,$3E,$E5,$78,$AF,$C3,$3F
	.dc.b	$E3,$1E,$3C,$DF,$0D,$7C,$7F,$7F,$DB,$5F,$0E,$B7,$0F,$F1,$C2,$FC
	.dc.b	$D3,$17,$EA,$FC,$DB,$F0,$BD,$2F,$C2,$9F,$B5,$7B,$BE,$47,$FC,$29
	.dc.b	$FB,$F7,$E0,$FA,$7F,$57,$7B,$AD,$79,$0A,$FC,$65,$6B,$07,$81,$FE
	.dc.b	$23,$F0,$3F,$C4,$A7,$F9,$7F,$1E,$4F,$FC,$00,$7A,$6D,$FD,$5B,$1F
	.dc.b	$F5,$3E,$1C,$F2,$EF,$F1,$BF,$A1,$BD,$E9,$E6,$FC,$17,$D7,$A7,$AB
	.dc.b	$2E,$5A,$BF,$5B,$FD,$1F,$E6,$2E,$E7,$EB,$2F,$D8,$5F,$88,$FF,$8A
	.dc.b	$3D,$97,$ED,$9C,$2F,$75,$AF,$8D,$FE,$51,$CB,$9F,$67,$DC,$FE,$D7
	.dc.b	$3E,$D9,$E2,$39,$3D,$F3,$FF,$1A,$F7,$EF,$2E,$B1,$FE,$38,$F3,$F1
	.dc.b	$5B,$C9,$FB,$E7,$D3,$F6,$BF,$C9,$D9,$F6,$D7,$C9,$3F,$4A,$7F,$CA
	.dc.b	$5D,$AF,$C6,$DF,$AC,$BF,$71,$FE,$7D,$E0,$BF,$4C,$7E,$B9,$FC,$23
	.dc.b	$0E,$17,$BF,$81,$FC,$55,$3F,$68,$D6,$4D,$BD,$DE,$17,$F3,$0C,$78
	.dc.b	$F3,$7C,$48,$C5,$78,$B5,$F0,$FC,$CF,$F1,$BB,$EF,$1E,$4F,$23,$15
	.dc.b	$BD,$3C,$3F,$1D,$E0,$DB,$C9,$C1,$E9,$FD,$DB,$18,$FB,$38,$1F,$84
	.dc.b	$E3,$1F,$4F,$EE,$BF,$DE,$FE,$77,$FC,$ED,$F8,$4B,$F5,$07,$E6,$3F
	.dc.b	$E4,$BC,$BF,$DA,$FF,$81,$3E,$97,$87,$37,$FA,$DF,$0E,$17,$BF,$C5
	.dc.b	$17,$FE,$32,$CB,$AE,$4F,$4F,$E2,$C6,$2E,$E3,$CE,$BC,$3F,$0B,$F5
	.dc.b	$71,$FF,$2E,$4F,$80,$FF,$65,$4E,$B4,$FF,$75,$FD,$8F,$F6,$BC,$7A
	.dc.b	$61,$C7,$3F,$CF,$75,$7F,$8D,$F1,$ED,$9F,$9E,$7F,$92,$5F,$FA,$5F
	.dc.b	$D6,$7E,$F6,$FC,$31,$FA,$B3,$B9,$FB,$FB,$F2,$17,$E9,$7E,$5C,$78
	.dc.b	$E1,$C2,$F7,$4F,$D6,$F5,$97,$5D,$BF,$09,$A7,$86,$3F,$C5,$F1,$83
	.dc.b	$FD,$C9,$8F,$4D,$3C,$5A,$7F,$63,$9E,$79,$F7,$87,$D9,$FC,$39,$E3
	.dc.b	$87,$AE,$F8,$FB,$3F,$EF,$2E,$1F,$89,$4C,$BB,$6D,$EE,$AF,$F2,$F5
	.dc.b	$6B,$E5,$7F,$D0,$9F,$AC,$FD,$BF,$C5,$3F,$D0,$5E,$AB,$F0,$EF,$AD
	.dc.b	$5A,$EC,$FF,$7F,$7D,$D7,$F0,$AF,$A4,$CF,$DA,$B2,$FC,$0E,$FF,$E0
	.dc.b	$58,$7E,$D7,$8C,$79,$69,$FE,$8F,$AF,$61,$BB,$F1,$63,$FC,$BA,$3A
	.dc.b	$C6,$34,$FF,$37,$EB,$E9,$CB,$C1,$A7,$92,$7F,$D7,$B3,$9A,$E1,$FA
	.dc.b	$E7,$2A,$D7,$C9,$FF,$98,$7F,$0A,$FF,$6E,$7E,$F6,$FC,$8B,$F0,$DF
	.dc.b	$B8,$BF,$04,$7E,$77,$FE,$67,$FC,$2F,$87,$00,$61,$C7,$B9,$87,$1B
	.dc.b	$FF,$37,$7D,$BF,$0B,$DE,$5F,$87,$7F,$A5,$70,$7F,$BF,$1E,$98,$F8
	.dc.b	$FE,$D3,$F8,$8E,$BE,$6B,$F2,$EB,$E4,$FD,$0B,$FC,$A1,$39,$D7,$C3
	.dc.b	$3E,$8F,$96,$CE,$5D,$30,$E3,$AF,$FA,$07,$E5,$BC,$9A,$7A,$77,$F4
	.dc.b	$6C,$F9,$CF,$F4,$37,$D1,$7F,$C0,$1F,$BC,$FF,$14,$7E,$92,$FD,$B3
	.dc.b	$F9,$47,$E5,$4F,$D3,$9F,$AD,$70,$E0,$F8,$D8,$70,$BD,$D7,$EC,$B3
	.dc.b	$FF,$1C,$56,$BE,$8D,$97,$F3,$0C,$FC,$10,$FD,$43,$15,$BA,$F1,$7C
	.dc.b	$3B,$57,$D6,$FF,$9C,$F1,$E3,$5A,$9F,$F5,$FD,$F9,$46,$BE,$5C,$39
	.dc.b	$69,$FD,$9B,$4F,$2D,$6B,$4F,$2E,$BE,$A8,$E5,$FA,$03,$C8,$FD,$87
	.dc.b	$9B,$FA,$3F,$B2,$7D,$23,$F7,$ED,$D6,$BC,$65,$6B,$F0,$AF,$F5,$A7
	.dc.b	$D6,$FE,$37,$37,$E8,$8F,$C0,$3C,$FB,$7F,$9A,$7F,$4D,$7E,$86,$FC
	.dc.b	$17,$8E,$1F,$95,$C3,$87,$37,$D7,$F9,$A6,$1C,$39,$B7,$7D,$B3,$F8
	.dc.b	$6F,$F5,$DD,$3C,$58,$F8,$7F,$3E,$FE,$C7,$18,$AF,$1E,$3D,$34,$FF
	.dc.b	$51,$4E,$73,$F0,$66,$FC,$FF,$39,$F9,$71,$E3,$3E,$AF,$C1,$7C,$7A
	.dc.b	$7A,$37,$F7,$67,$FD,$5F,$A7,$FC,$F5,$8F,$1C,$FA,$FC,$3F,$C1,$9C
	.dc.b	$CF,$AE,$73,$C1,$D3,$4E,$FF,$0F,$FE,$B7,$FC,$7B,$E7,$7E,$44,$F5
	.dc.b	$5F,$A9,$3F,$02,$FE,$7D,$F9,$6B,$F0,$77,$EE,$7F,$EE,$9C,$78,$E1
	.dc.b	$C2,$F7,$4F,$DC,$3F,$DD,$DD,$9F,$2B,$E3,$5F,$B1,$FF,$9C,$3B,$5F
	.dc.b	$00,$FD,$0B,$FD,$27,$CC,$C3,$85,$EE,$B5,$F2,$5F,$EF,$9F,$67,$FD
	.dc.b	$95,$CA,$FA,$37,$E4,$DF,$EA,$3F,$D9,$1F,$A9,$7A,$34,$FE,$21,$F8
	.dc.b	$4F,$E2,$EF,$7E,$9A,$F7,$C5,$D7,$BB,$FD,$B5,$9F,$5C,$FA,$7C,$07
	.dc.b	$FB,$7F,$BF,$F2,$DF,$76,$FD,$09,$F8,$97,$F8,$5B,$F3,$67,$E4,$0F
	.dc.b	$DF,$73,$C5,$F8,$ED,$58,$F1,$C3,$85,$EF,$F1,$55,$EE,$B5,$F4,$8C
	.dc.b	$BA,$FB,$9E,$3C,$BF,$9A,$56,$A7,$C3,$FA,$AF,$FB,$D7,$F4,$1F,$D6
	.dc.b	$7B,$EF,$CC,$1F,$90,$BF,$30,$7E,$6F,$FC,$6F,$F7,$2F,$C1,$EE,$C7
	.dc.b	$8E,$1C,$2F,$7E,$82,$F5,$7B,$DB,$FA,$36,$7D,$F3,$ED,$9F,$F2,$7E
	.dc.b	$7F,$25,$6B,$3E,$93,$FC,$AF,$E3,$7F,$A7,$62,$FD,$17,$F8,$27,$ED
	.dc.b	$1F,$E3,$1F,$C3,$9F,$A1,$BC,$3F,$F1,$E7,$E4,$AF,$DE,$FE,$4E,$1C
	.dc.b	$3E,$6B,$FD,$7D,$7B,$FA,$F5,$BF,$73,$DF,$07,$F5,$DD,$3B,$5E,$EB
	.dc.b	$C7,$8F,$3D,$3C,$3A,$7F,$92,$FE,$9F,$CF,$F8,$4F,$F1,$99,$FF,$48
	.dc.b	$FE,$1F,$FE,$72,$C3,$BF,$CB,$3F,$3E,$FB,$B7,$BF,$C1,$3F,$D5,$BF
	.dc.b	$80,$79,$2F,$FB,$6F,$CB,$EB,$83,$FB,$86,$7D,$2B,$5F,$AB,$F8,$27
	.dc.b	$38,$72,$C3,$DB,$F6,$9F,$2B,$F1,$1F,$E5,$91,$F6,$FE,$E5,$EE,$7F
	.dc.b	$01,$FD,$B1,$FB,$3B,$F6,$4F,$C8,$E1,$C3,$C7,$61,$C3,$E0,$D8,$70
	.dc.b	$FD,$97,$7B,$7F,$81,$74,$ED,$A7,$5F,$CE,$7C,$53,$9C,$B9,$CF,$87
	.dc.b	$F6,$8F,$BF,$FA,$1F,$F4,$1F,$31,$F9,$6F,$F4,$2F,$C4,$9F,$A4,$3F
	.dc.b	$54,$7A,$2F,$2B,$F4,$F8,$F1,$C3,$85,$EE,$9F,$A6,$EB,$6F,$F0,$D6
	.dc.b	$7D,$BE,$59,$A7,$6D,$3A,$D6,$CF,$D3,$65,$D2,$3F,$BB,$C7,$ED,$BF
	.dc.b	$53,$FE,$B1,$D8,$FD,$BB,$FB,$27,$5F,$F8,$AB,$EA,$7F,$81,$FA,$5B
	.dc.b	$C3,$FC,$55,$7B,$FA,$2F,$A3,$F5,$BF,$DF,$FD,$E7,$EB,$3B,$B5,$EF
	.dc.b	$2E,$9F,$B0,$FE,$9B,$5F,$0E,$BE,$0C,$BF,$C1,$1E,$3F,$EA,$BE,$4F
	.dc.b	$FB,$67,$C7,$7F,$2F,$FD,$A1,$FB,$27,$DC,$F0,$9F,$A0,$3E,$6B,$F8
	.dc.b	$8E,$1C,$3E,$59,$87,$0F,$21,$87,$0F,$C6,$B7,$C7,$FE,$45,$D3,$B6
	.dc.b	$9D,$7E,$1B,$C5,$39,$CB,$9C,$F8,$7C,$0F,$DF,$FD,$CF,$F8,$07,$F4
	.dc.b	$A7,$E6,$0F,$CD,$3F,$6F,$F8,$EB,$F2,$B7,$E3,$2F,$DD,$3F,$CA,$DA
	.dc.b	$FF,$95,$71,$E3,$87,$0B,$DF,$EA,$DA,$A5,$69,$FF,$6E,$67,$E1,$CF
	.dc.b	$C1,$E8,$B4,$F0,$69,$E6,$BE,$D5,$A9,$F2,$7E,$15,$FC,$63,$D7,$FF
	.dc.b	$0D,$63,$FE,$6B,$F5,$FF,$D0,$DF,$3A,$FB,$87,$E0,$9F,$99,$F8,$BF
	.dc.b	$F6,$2E,$8C,$38,$7D,$AF,$F5,$AB,$DF,$A5,$FD,$13,$3E,$BF,$80,$7D
	.dc.b	$2C,$3F,$DE,$91,$8D,$3F,$DF,$2D,$C3,$94,$F9,$4F,$F3,$D6,$3D,$B4
	.dc.b	$F2,$7C,$27,$F3,$0F,$C8,$BF,$C6,$5E,$9F,$F5,$9F,$5B,$F5,$5E,$C7
	.dc.b	$FB,$1B,$F0,$EF,$EA,$B5,$AF,$9B,$FD,$47,$77,$EC,$FF,$12,$FC,$CF
	.dc.b	$22,$B5,$8F,$F8,$F7,$6F,$F1,$F4,$E5,$FF,$46,$67,$FD,$9F,$F1,$4D
	.dc.b	$47,$8C,$CE,$5D,$B4,$7E,$81,$99,$FD,$CC,$E7,$6F,$E9,$3F,$90,$FF
	.dc.b	$92,$FF,$69,$FD,$6F,$B5,$FB,$2F,$49,$FB,$07,$F6,$EA,$D7,$AD,$FB
	.dc.b	$7F,$65,$ED,$FE,$EF,$F0,$DD,$6B,$FE,$A1,$CF,$F6,$EF,$9B,$89,$8F
	.dc.b	$2F,$8A,$F4,$CF,$01,$5E,$2C,$7A,$F8,$BE,$69,$E9,$9F,$87,$37,$C1
	.dc.b	$FF,$62,$F8,$7F,$F3,$27,$E3,$2F,$D7,$3F,$C5,$DF,$95,$BF,$45,$7E
	.dc.b	$B5,$FE,$64,$F9,$6F,$F0,$4E,$1C,$2F,$75,$AF,$BC,$FF,$36,$71,$FF
	.dc.b	$3A,$DB,$FD,$79,$9F,$5F,$AD,$F4,$CE,$65,$3E,$1F,$D5,$B3,$97,$5B
	.dc.b	$6B,$E3,$D7,$FC,$EB,$87,$88,$EB,$A7,$8A,$3F,$CC,$91,$FE,$1E,$FC
	.dc.b	$4F,$FD,$01,$F9,$17,$FA,$7F,$D3,$7B,$9F,$E7,$EF,$DD,$9F,$98,$B6
	.dc.b	$E1,$8F,$FB,$4F,$0E,$17,$BA,$7E,$B1,$AA,$57,$FA,$AF,$F4,$8F,$6E
	.dc.b	$7E,$0F,$07,$FD,$4B,$5F,$ED,$48,$C6,$9E,$2E,$46,$BF,$EF,$5E,$45
	.dc.b	$6B,$5F,$36,$1F,$E1,$8F,$8D,$FF,$4D,$7C,$BF,$F6,$6E,$BF,$C0,$FE
	.dc.b	$66,$FC,$DF,$F8,$C7,$DF,$7F,$8F,$AF,$7D,$3F,$DC,$BC,$15,$EF,$E0
	.dc.b	$9F,$9F,$EF,$4F,$F7,$96,$9F,$DC,$BE,$3F,$EB,$E5,$CB,$5B,$5A,$CF
	.dc.b	$A6,$5E,$1F,$C0,$3C,$D3,$9A,$F2,$ED,$FC,$BE,$3F,$C2,$DC,$3F,$EB
	.dc.b	$3F,$C0,$FF,$D6,$5F,$B7,$B8,$5F,$8D,$3F,$1E,$7B,$8F,$D5,$38,$E4
	.dc.b	$3F,$D1,$BF,$7C,$C7,$96,$1C,$6D,$F4,$BC,$2F,$7F,$B8,$FE,$AB,$EA
	.dc.b	$3F,$CC,$71,$CF,$5F,$1B,$FC,$1A,$32,$F6,$C7,$AE,$9F,$EE,$48,$E9
	.dc.b	$5E,$AD,$7D,$32,$FD,$03,$CD,$87,$DC,$BF,$47,$E3,$01,$FB,$03,$F3
	.dc.b	$07,$E2,$1F,$DD,$3F,$0C,$FE,$8D,$87,$1C,$38,$61,$C2,$F7,$F4,$4B
	.dc.b	$7D,$7F,$F9,$0F,$C3,$FF,$70,$5F,$7C,$FE,$04,$62,$3F,$81,$E3,$EF
	.dc.b	$8C,$4F,$8F,$3F,$14,$F3,$3D,$35,$E1,$87,$E9,$9F,$EE,$DD,$3F,$79
	.dc.b	$F5,$CF,$CF,$FF,$AF,$7D,$73,$F4,$07,$ED,$5E,$EF,$A2,$FF,$2C,$78
	.dc.b	$7F,$EF,$5C,$38,$7C,$C7,$FA,$16,$F7,$A3,$F2,$2D,$EF,$F8,$1B,$ED
	.dc.b	$F1,$BF,$6F,$1E,$5A,$FF,$6D,$AF,$B5,$18,$CB,$A6,$9F,$ED,$39,$E6
	.dc.b	$30,$F0,$EB,$E7,$9F,$CC,$6D,$FC,$CA,$BF,$9B,$7E,$E6,$FD,$CB,$ED
	.dc.b	$7F,$00,$7F,$9C,$7F,$1E,$7E,$51,$FE,$73,$F8,$8F,$E8,$D7,$BE,$65
	.dc.b	$EF,$F1,$3D,$65,$D2,$BF,$29,$7C,$32,$B3,$E0,$7B,$E7,$F9,$13,$C9
	.dc.b	$1F,$D0,$AB,$B6,$1F,$E8,$3F,$5F,$F2,$3C,$7F,$DC,$3E,$05,$FB,$4B
	.dc.b	$DB,$FF,$79,$7E,$E9,$E8,$FD,$CF,$FD,$15,$7B,$AD,$7D,$72,$BD,$BA
	.dc.b	$EB,$5F,$ED,$B9,$3F,$2A,$FD,$66,$3C,$67,$F5,$4F,$3C,$FB,$FE,$BD
	.dc.b	$E4,$9E,$39,$F6,$CB,$F3,$51,$FE,$9B,$F9,$FF,$F0,$FF,$CC,$BC,$E7
	.dc.b	$ED,$3F,$EC,$CF,$C8,$FF,$DE,$DF,$6F,$BC,$3F,$C2,$37,$BA,$D7,$E3
	.dc.b	$5E,$9E,$1F,$F1,$1F,$D1,$FF,$9D,$32,$FE,$05,$9F,$6F,$D9,$6E,$5C
	.dc.b	$A7,$BF,$E1,$8F,$C2,$3F,$3D,$7E,$49,$CE,$DE,$8C,$BF,$C5,$D8,$7F
	.dc.b	$8E,$BF,$6D,$FD,$EF,$33,$F7,$0F,$E3,$9F,$83,$FF,$AD,$F6,$FF,$22
	.dc.b	$56,$A7,$F9,$07,$E0,$F9,$C3,$91,$FC,$4E,$BF,$C9,$5F,$51,$FC,$65
	.dc.b	$E3,$E8,$B2,$ED,$F8,$8F,$F1,$09,$CC,$78,$7D,$8F,$E2,$7D,$DF,$E7
	.dc.b	$1F,$C2,$3F,$8E,$7B,$6F,$CE,$3F,$D9,$53,$9F,$77,$F8,$E7,$C0,$E3
	.dc.b	$6F,$F3,$76,$1F,$C8,$F9,$B9,$C6,$2B,$F9,$DF,$E4,$9F,$C8,$3F,$03
	.dc.b	$FF,$44,$FE,$08,$F2,$BF,$2B,$F8,$27,$FB,$5E,$73,$1F,$98,$E9,$C6
	.dc.b	$3F,$E9,$AF,$A8,$61,$C7,$CD,$FE,$51,$F5,$9D,$6B,$53,$9F,$47,$FC
	.dc.b	$5B,$1F,$EE,$3F,$C2,$7F,$D7,$1F,$84,$7E,$83,$CA,$FC,$DF,$93,$9A
	.dc.b	$FF,$5B,$FE,$10,$AD,$79,$BF,$A4,$D9,$8F,$2C,$7F,$53,$2E,$2F,$E7
	.dc.b	$B3,$99,$FD,$67,$27,$F4,$0F,$67,$FD,$BD,$F0,$1F,$EE,$4F,$DD,$9F
	.dc.b	$8D,$FC,$3F,$C3,$FF,$DF,$15,$AF,$C3,$55,$AE,$7F,$CF,$ED,$FC,$57
	.dc.b	$87,$FD,$BD,$97,$3C,$9F,$87,$77,$9C,$E3,$CA,$5E,$C7,$F5,$2F,$2F
	.dc.b	$FB,$45,$7F,$64,$F2,$5F,$64,$FB,$0F,$85,$E5,$3F,$24,$FE,$3D,$B3
	.dc.b	$1E,$3F,$22,$C7,$8E,$1C,$2F,$EF,$5E,$F2,$FE,$ED,$7B,$CB,$B7,$4F
	.dc.b	$F6,$0F,$C2,$3C,$95,$AC,$B9,$CF,$F6,$4F,$C7,$3F,$B4,$FD,$7F,$A3
	.dc.b	$D5,$7E,$29,$FD,$F3,$A7,$FC,$41,$E4,$3F,$2F,$F9,$3B,$B0,$E1,$7B
	.dc.b	$F9,$B5,$EF,$D9,$5E,$F6,$FF,$06,$67,$DB,$85,$A7,$6F,$B2,$F1,$D6
	.dc.b	$B3,$E9,$93,$CD,$FE,$FF,$F1,$5F,$3B,$D6,$7E,$2E,$FC,$61,$F8,$A3
	.dc.b	$81,$FA,$37,$FB,$53,$E2,$7C,$78,$F1,$C3,$85,$EE,$BF,$DB,$3F,$3F
	.dc.b	$AD,$BF,$C4,$59,$F6,$CF,$AE,$7D,$3F,$5F,$F1,$D6,$B2,$E7,$3E,$0D
	.dc.b	$7F,$C2,$BF,$90,$FE,$13,$F1,$F7,$E4,$3F,$96,$FC,$85,$F9,$0B,$D4
	.dc.b	$7E,$52,$FC,$D7,$E4,$E3,$C7,$0E,$17,$BA,$FF,$26,$F6,$2B,$5E,$53
	.dc.b	$2E,$B9,$7F,$9F,$2B,$5B,$FF,$8F,$74,$F1,$DE,$F2,$F1,$7A,$9F,$E2
	.dc.b	$CF,$94,$FC,$DF,$E3,$CF,$A8,$7C,$73,$F2,$47,$E4,$CF,$DA,$FF,$5B
	.dc.b	$FB,$1F,$E0,$31,$E3,$87,$0B,$DD,$7E,$CB,$F3,$FD,$6F,$FE,$4F,$CF
	.dc.b	$B6,$7D,$73,$E9,$E5,$F8,$EB,$59,$73,$9F,$07,$C3,$7F,$90,$BF,$2C
	.dc.b	$FD,$8F,$9D,$FE,$5F,$FC,$B1,$FA,$63,$D9,$7E,$5A,$FD,$3D,$FC,$01
	.dc.b	$F9,$6C,$78,$FC,$83,$1E,$38,$70,$B7,$EB,$5F,$9C,$FA,$FF,$DF,$FE
	.dc.b	$62,$FB,$CF,$D0,$FD,$CB,$F4,$37,$E5,$5F,$C6,$FE,$C3,$87,$37,$F9
	.dc.b	$B7,$0E,$1E,$AB,$0E,$17,$FC,$85,$FD,$4F,$E6,$BF,$C3,$3E,$9D,$F9
	.dc.b	$2A,$1F,$7E,$FD,$CD,$F4,$CF,$DF,$3F,$CF,$9E,$5E,$F6,$FF,$3E,$65
	.dc.b	$CB,$1E,$38,$70,$BF,$F3,$DD,$EE,$D7,$E0,$EE,$FF,$3E,$61,$E3,$CF
	.dc.b	$55,$FE,$55,$8C,$1D,$32,$FE,$4B,$1E,$A3,$C3,$3E,$9C,$FD,$13,$FE
	.dc.b	$2F,$AE,$31,$8F,$8C,$FE,$1F,$18,$D3,$DB,$1F,$BD,$CF,$CD,$9F,$AE
	.dc.b	$F9,$FC,$6F,$FA,$13,$4F,$F3,$AF,$CC,$3F,$44,$7E,$7B,$FE,$A0,$FC
	.dc.b	$ED,$F9,$EF,$EB,$F3,$FC,$DF,$0F,$FA,$2B,$0E,$17,$BA,$7D,$37,$F1
	.dc.b	$CF,$A2,$FE,$7F,$F4,$BF,$E3,$9F,$D0,$5F,$85,$3F,$45,$FD,$E7,$E5
	.dc.b	$7F,$ED,$AD,$7F,$D6,$3A,$FF,$AB,$70,$E1,$7B,$F6,$96,$F6,$BB,$5E
	.dc.b	$FE,$C9,$FA,$D7,$FA,$B7,$5F,$0C,$35,$C5,$6D,$F1,$EB,$FC,$37,$E1
	.dc.b	$7F,$D4,$F7,$C2,$73,$9F,$F2,$DB,$EE,$7F,$AB,$70,$E9,$A7,$97,$4F
	.dc.b	$F6,$56,$DE,$AA,$D6,$1C,$A7,$F8,$9E,$3F,$D8,$BE,$4B,$F8,$E7,$ED
	.dc.b	$4F,$C5,$DF,$AC,$BF,$7D,$7E,$90,$FC,$67,$C3,$87,$0B,$DD,$69,$CF
	.dc.b	$03,$1F,$7E,$BF,$EF,$58,$7E,$D5,$8C,$39,$67,$DF,$F1,$7F,$83,$7F
	.dc.b	$ED,$D9,$CE,$4F,$B8,$CC,$F3,$CB,$D1,$87,$9F,$57,$E3,$9E,$29,$CE
	.dc.b	$1E,$AF,$B9,$C6,$DE,$8B,$7E,$45,$8C,$79,$7E,$F9,$FD,$93,$EE,$3F
	.dc.b	$C0,$FE,$6F,$F5,$DF,$6D,$FB,$73,$F0,$F7,$E4,$1F,$CC,$3E,$33,$87
	.dc.b	$17,$FB,$D7,$0E,$17,$BA,$FE,$6B,$F1,$3A,$AF,$E0,$5F,$2A,$FD,$1F
	.dc.b	$FD,$E9,$A7,$82,$1A,$62,$BA,$46,$31,$F1,$7E,$CB,$FE,$D1,$BE,$17
	.dc.b	$E6,$D7,$CB,$EF,$7F,$BC,$A7,$37,$E3,$F8,$B7,$E9,$5E,$3D,$FC,$F8
	.dc.b	$7B,$27,$FD,$39,$E1,$FF,$99,$75,$0F,$C6,$1F,$3A,$FC,$9B,$F1,$5F
	.dc.b	$02,$FD,$51,$F9,$9F,$E8,$71,$E5,$8F,$1C,$38,$76,$7F,$C0,$0B,$DF
	.dc.b	$9D,$F1,$73,$EB,$9F,$C1,$C3,$87,$B3,$67,$38,$71,$FA,$2F,$75,$79
	.dc.b	$A7,$3A,$F8,$FF,$4A,$CC,$F8,$71,$F3,$E7,$E6,$FD,$57,$C1,$AF,$93
	.dc.b	$0F,$56,$EF,$D5,$3E,$C7,$CA,$3F,$7F,$7E,$E5,$F8,$AF,$8E,$7E,$1C
	.dc.b	$FD,$AB,$E2,$F3,$7C,$0C,$38,$5E,$EB,$5F,$8F,$F4,$C7,$9F,$0F,$F9
	.dc.b	$58,$F3,$DF,$7F,$D1,$BE,$13,$CB,$C9,$FE,$CD,$83,$ED,$F7,$87,$92
	.dc.b	$73,$F8,$6E,$33,$F0,$6A,$FD,$7B,$C1,$18,$FD,$B2,$DF,$4A,$FB,$3B
	.dc.b	$FA,$71,$E9,$C1,$EE,$97,$E2,$7F,$E4,$AE,$FF,$F3,$6F,$B8,$FC,$3D
	.dc.b	$F8,$17,$EA,$3F,$4B,$7E,$DE,$FD,$DB,$FD,$07,$AF,$F8,$63,$9F,$F1
	.dc.b	$B8,$F1,$C3,$85,$EF,$EC,$BB,$56,$B2,$7E,$41,$F1,$E3,$F2,$B8,$F3
	.dc.b	$7F,$98,$4F,$89,$D4,$67,$1E,$DE,$5B,$4F,$2F,$E0,$5C,$B6,$F5,$E3
	.dc.b	$EA,$AD,$68,$FC,$E7,$59,$F8,$F6,$F3,$EE,$FA,$8F,$3C,$F8,$A7,$3F
	.dc.b	$1F,$8F,$39,$FA,$53,$F4,$EF,$D6,$7D,$33,$F5,$27,$D1,$7D,$7E,$5F
	.dc.b	$66,$1C,$2F,$75,$AE,$DE,$98,$BF,$76,$F8,$9A,$E2,$EE,$3C,$AB,$C1
	.dc.b	$E7,$B4,$F0,$79,$FE,$CC,$3F,$C2,$D5,$A9,$F2,$7C,$D6,$75,$F1,$61
	.dc.b	$C7,$5F,$3F,$BB,$DE,$31,$BB,$F5,$8F,$87,$A7,$E5,$FE,$05,$FA,$63
	.dc.b	$F0,$3F,$47,$E9,$AF,$D3,$77,$E6,$2F,$75,$AF,$D2,$73,$1E,$CF,$03
	.dc.b	$FD,$7B,$97,$E8,$FE,$5B,$1C,$BF,$C6,$23,$C1,$A7,$7E,$8F,$AD,$3C
	.dc.b	$A3,$1F,$25,$F0,$6B,$79,$F9,$3F,$5D,$D3,$E6,$FC,$06,$E2,$E0,$7E
	.dc.b	$ED,$85,$FB,$97,$E0,$BE,$8D,$FB,$07,$BF,$F2,$2F,$F1,$07,$E2,$3F
	.dc.b	$C2,$FF,$7E,$D4,$E6,$BF,$D9,$DE,$A3,$F5,$0C,$ED,$EC,$E3,$FF,$14
	.dc.b	$5F,$EB,$1F,$CE,$F2,$FF,$B7,$B0,$EF,$4C,$FB,$FE,$8B,$F7,$67,$31
	.dc.b	$E2,$9C,$61,$E1,$D3,$F8,$F7,$EE,$3E,$0A,$F3,$E8,$FD,$CB,$F3,$BF
	.dc.b	$A7,$BE,$7D,$FA,$33,$F4,$2F,$F1,$CF,$E3,$4F,$D2,$FF,$A1,$70,$FF
	.dc.b	$BC,$B0,$E1,$7B,$FA,$EF,$F7,$85,$3F,$71,$EB,$93,$F4,$8F,$81,$87
	.dc.b	$0E,$5F,$79,$FD,$51,$E7,$A7,$83,$C4,$FC,$D4,$63,$2F,$F7,$E7,$C2
	.dc.b	$E3,$1F,$3D,$F9,$B3,$FD,$9F,$A1,$FC,$27,$0E,$1B,$FA,$7F,$51,$FD
	.dc.b	$DF,$E8,$BF,$DA,$3F,$6D,$7E,$95,$FE,$80,$FD,$CB,$FD,$1D,$5A,$F9
	.dc.b	$B5,$6B,$EF,$95,$FE,$93,$F5,$F5,$F9,$8F,$DB,$CB,$A7,$27,$B0,$3D
	.dc.b	$F3,$ED,$F3,$BF,$B5,$8C,$63,$D3,$E6,$71,$A7,$82,$FC,$9F,$82,$3F
	.dc.b	$6C,$F2,$6D,$E3,$DB,$FD,$49,$FB,$1E,$2F,$7F,$AF,$DC,$FC,$F8,$FF
	.dc.b	$61,$E4,$FE,$59,$FA,$7F,$F8,$83,$D5,$7E,$A0,$FC,$A1,$D9,$F0,$3E
	.dc.b	$D3,$F9,$0D,$EE,$B5,$3F,$13,$F3,$1C,$FD,$6F,$EF,$79,$FE,$05,$FF
	.dc.b	$8B,$3D,$AF,$CF,$3C,$72,$EB,$E7,$FF,$15,$8C,$63,$FC,$26,$F9,$E9
	.dc.b	$E4,$BF,$1F,$27,$E5,$76,$F4,$E7,$FC,$2F,$5F,$1C,$BF,$2F,$FE,$6F
	.dc.b	$F9,$57,$F9,$3F,$B5,$FA,$3F,$EC,$79,$7D,$BF,$A5,$EB,$EE,$7F,$9C
	.dc.b	$CE,$78,$1F,$C0,$19,$FB,$FF,$DB,$65,$CF,$8B,$EE,$3D,$6F,$84,$73
	.dc.b	$CF,$BF,$AE,$D3,$FB,$E4,$E6,$BB,$71,$5A,$D6,$6F,$CB,$BF,$A7,$FD
	.dc.b	$17,$FB,$5F,$AD,$FA,$83,$D7,$FF,$69,$7E,$E4,$FD,$8B,$7F,$88,$FF
	.dc.b	$B9,$6B,$5F,$87,$AB,$ED,$F5,$FF,$6B,$CB,$A7,$07,$E5,$5F,$F7,$27
	.dc.b	$9D,$FC,$4D,$F0,$61,$CB,$F5,$0F,$E9,$51,$8D,$3F,$DC,$5C,$18,$D3
	.dc.b	$C1,$1E,$4F,$DD,$FC,$7B,$FA,$23,$CF,$E8,$E3,$2F,$2E,$CF,$AB,$C5
	.dc.b	$7B,$3E,$9B,$FB,$7F,$DE,$3F,$40,$7E,$D0,$FD,$B3,$F8,$B4,$7F,$4D
	.dc.b	$FD,$0F,$6E,$A3,$3F,$27,$FE,$F6,$E6,$FF,$05,$61,$CF,$F0,$CF,$E6
	.dc.b	$0F,$5C,$6E,$1C,$7D,$CF,$06,$1F,$DA,$6B,$58,$F8,$38,$F1,$7B,$3F
	.dc.b	$AF,$72,$78,$B1,$F3,$63,$CB,$E2,$BF,$C1,$3F,$89,$FF,$B2,$7F,$7A
	.dc.b	$7E,$A4,$E8,$4B,$D6,$77,$70,$FA,$C6,$1C,$2F,$74,$FC,$DF,$53,$FD
	.dc.b	$2F,$0F,$F6,$B7,$C1,$5C,$38,$E5,$DE,$7F,$CF,$FE,$0F,$08,$C7,$E2
	.dc.b	$7F,$E0,$79,$F3,$46,$F2,$7E,$8D,$FE,$22,$F8,$AF,$F2,$87,$E0,$0F
	.dc.b	$D5,$1F,$8F,$7A,$7E,$29,$FA,$16,$FE,$EB,$FD,$A9,$5A,$CB,$FC,$BD
	.dc.b	$F9,$1E,$7F,$14,$FE,$C5,$9F,$4E,$0F,$E5,$0F,$6B,$E1,$97,$F8,$6E
	.dc.b	$31,$9F,$C7,$FC,$A3,$19,$F6,$8F,$17,$EF,$DF,$E3,$8F,$DC,$DF,$A6
	.dc.b	$FF,$B8,$7E,$C3,$F5,$5F,$3E,$ED,$7D,$2E,$EB,$5E,$5F,$F8,$0E,$5F
	.dc.b	$0F,$9F,$B9,$FE,$F1,$97,$4E,$3B,$87,$0C,$BF,$BF,$71,$F3,$96,$9E
	.dc.b	$0F,$53,$FC,$69,$87,$F7,$1D,$5F,$8D,$BF,$6A,$79,$CF,$C3,$DF,$A4
	.dc.b	$FA,$FE,$7B,$CF,$7B,$AD,$7E,$C3,$9C,$7C,$B4,$FD,$AE,$C7,$6C,$BA
	.dc.b	$E5,$E2,$AE,$71,$FE,$2A,$DF,$96,$5D,$69,$F4,$DF,$E6,$2F,$C4,$BF
	.dc.b	$C7,$5F,$B5,$BF,$64,$FF,$1E,$5F,$FA,$1F,$E5,$B5,$F2,$4A,$D3,$EE
	.dc.b	$FC,$0B,$FC,$B9,$C3,$FE,$7D,$C7,$9E,$9F,$D9,$DF,$F2,$37,$E9,$9F
	.dc.b	$E2,$57,$BE,$7D,$BF,$37,$FF,$2E,$46,$31,$F9,$3F,$A5,$E1,$F0,$57
	.dc.b	$93,$2F,$F3,$A6,$5D,$E7,$27,$8F,$F6,$1F,$3E,$9E,$6D,$BF,$D4,$1F
	.dc.b	$B5,$7F,$A0,$3F,$56,$7D,$43,$F4,$07,$E0,$1F,$D1,$BC,$BF,$E8,$FF
	.dc.b	$15,$BA,$D4,$E7,$E5,$9F,$1B,$9F,$7B,$FC,$81,$97,$4C,$AE,$1C,$2F
	.dc.b	$FD,$85,$F8,$F7,$F9,$E6,$B5,$8F,$7F,$B3,$C5,$EF,$4F,$EA,$FF,$8D
	.dc.b	$F8,$72,$F2,$E5,$FE,$CB,$3F,$DF,$9C,$7F,$EA,$BF,$86,$75,$3F,$0E
	.dc.b	$7E,$7E,$FC,$55,$FB,$D7,$F9,$83,$E6,$3F,$D4,$38,$70,$BD,$D6,$BE
	.dc.b	$E5,$FB,$5E,$B2,$E9,$E1,$7F,$88,$70,$EF,$FA,$87,$F9,$7D,$E5,$A7
	.dc.b	$7D,$1F,$1D,$FE,$97,$9C,$E7,$E2,$F9,$BC,$65,$E7,$C3,$86,$8E,$8F
	.dc.b	$06,$DE,$4C,$FF,$CE,$FC,$3F,$EB,$8F,$6B,$FD,$F5,$FA,$BB,$F0,$07
	.dc.b	$E6,$3F,$AE,$F9,$17,$D5,$6F,$E6,$17,$BA,$D7,$4E,$63,$FC,$F1,$8A
	.dc.b	$BF,$C1,$BC,$DF,$E3,$F8,$EF,$97,$6E,$9F,$E0,$67,$FB,$6E,$33,$8F
	.dc.b	$EF,$BD,$7F,$F5,$B4,$62,$7F,$D0,$DE,$6F,$7B,$E5,$AB,$F5,$BC,$63
	.dc.b	$E9,$FB,$4B,$87,$FA,$37,$F0,$CF,$F6,$BF,$E6,$4F,$DF,$1F,$AC,$3E
	.dc.b	$AF,$FD,$15,$F8,$9B,$F0,$3F,$E7,$58,$FF,$BB,$F0,$E1,$7B,$FC,$0F
	.dc.b	$FD,$C5,$4F,$AE,$EB,$93,$EE,$BF,$D3,$58,$70,$E5,$FF,$30,$3C,$B4
	.dc.b	$EF,$CB,$FE,$D3,$8C,$65,$FE,$ED,$E2,$C6,$3E,$6B,$F2,$E7,$FE,$89
	.dc.b	$D7,$C9,$39,$BF,$1F,$CC,$FF,$BD,$39,$5F,$32,$FC,$F7,$FD,$F1,$FB
	.dc.b	$EF,$F6,$DF,$C1,$DF,$4F,$BA,$D7,$2A,$B5,$F9,$D6,$70,$FF,$5A,$63
	.dc.b	$D3,$F1,$7B,$87,$0B,$FE,$F1,$E4,$FF,$BB,$6B,$C3,$7E,$0F,$A3,$FC
	.dc.b	$DF,$37,$3C,$FB,$47,$FB,$33,$F0,$2C,$63,$E5,$FB,$8F,$E7,$FF,$34
	.dc.b	$C2,$7E,$B8,$FC,$69,$F9,$6B,$F6,$95,$FE,$DB,$F3,$07,$E5,$FC,$B1
	.dc.b	$E3,$87,$0B,$DF,$9C,$B7,$E0,$1B,$B6,$7D,$FF,$16,$FF,$6A,$3E,$1D
	.dc.b	$2E,$5C,$F4,$FE,$0B,$F0,$5E,$7E,$EF,$1E,$3E,$0B,$F3,$57,$F0,$AF
	.dc.b	$03,$4F,$B6,$7B,$E1,$CF,$5F,$27,$4F,$8F,$3F,$06,$FE,$8D,$DF,$9F
	.dc.b	$67,$0E,$10,$FD,$FF,$1A,$F8,$32,$F7,$7C,$6F,$B7,$F6,$FF,$F6,$E7
	.dc.b	$E4,$6F,$A4,$7D,$0B,$F7,$DF,$F0,$2F,$EC,$DB,$E6,$5E,$EB,$5F,$46
	.dc.b	$FC,$17,$5F,$8D,$7F,$BE,$73,$EB,$CD,$FE,$BF,$7B,$E9,$DB,$47,$E0
	.dc.b	$1F,$0A,$BE,$F5,$6B,$2F,$17,$CE,$66,$53,$9B,$FF,$0C,$FD,$E7,$97
	.dc.b	$5F,$24,$7F,$87,$3E,$E7,$ED,$7E,$64,$FD,$09,$F9,$EF,$F8,$2B,$F7
	.dc.b	$F7,$EE,$5F,$C0,$FF,$30,$DD,$6B,$C3,$56,$BD,$57,$E1,$9A,$C7,$EC
	.dc.b	$70,$7D,$9B,$FB,$EB,$3F,$07,$E1,$DA,$8F,$54,$78,$EB,$C5,$A7,$FB
	.dc.b	$32,$7A,$5F,$82,$3F,$4E,$F4,$AE,$0E,$71,$7E,$95,$C3,$3F,$37,$EC
	.dc.b	$3F,$99,$FB,$0F,$F1,$47,$92,$FD,$2B,$FA,$9F,$EA,$2F,$D8,$9F,$0B
	.dc.b	$BF,$AD,$FE,$45,$5A,$97,$D5,$67,$F5,$CF,$57,$16,$F2,$BC,$1F,$03
	.dc.b	$4E,$F8,$F6,$F6,$3E,$C4,$63,$E7,$1B,$3E,$EF,$3B,$78,$27,$3B,$3E
	.dc.b	$9B,$B6,$5E,$3F,$A6,$B5,$AF,$B4,$B9,$7A,$FE,$C3,$F7,$1F,$8D,$3F
	.dc.b	$5C,$7D,$5B,$F3,$47,$E8,$7F,$E7,$8F,$DA,$1F,$B1,$70,$BD,$D6,$BF
	.dc.b	$07,$57,$F8,$03,$9B,$EE,$E5,$D7,$F3,$2F,$83,$F0,$1F,$B9,$8F,$06
	.dc.b	$98,$EA,$FE,$31,$18,$CB,$A7,$C9,$7E,$6E,$31,$B7,$FA,$2B,$F1,$CC
	.dc.b	$6D,$E2,$8F,$37,$D6,$F9,$72,$F5,$47,$A7,$E0,$7F,$CC,$7F,$71,$FC
	.dc.b	$27,$EC,$DF,$9B,$BE,$59,$EA,$7D,$FF,$27,$FD,$29,$F8,$2F,$0E,$2F
	.dc.b	$E5,$98,$70,$BD,$F8,$1F,$8C,$FC,$19,$FB,$0E,$ED,$7B,$7F,$39,$D0
	.dc.b	$D7,$C5,$AB,$F2,$9F,$C2,$C6,$33,$EB,$AF,$FA,$BA,$F8,$61,$E2,$C3
	.dc.b	$D1,$B7,$9F,$E5,$BF,$95,$63,$C6,$3D,$5A,$BF,$41,$F2,$6B,$E9,$9C
	.dc.b	$FE,$EB,$E7,$FD,$6B,$FC,$E5,$FB,$5F,$EC,$FE,$B9,$FA,$E3,$F2,$EF
	.dc.b	$F7,$57,$07,$8B,$81,$87,$0F,$5D,$87,$0B,$7E,$F9,$D7,$E3,$1F,$97
	.dc.b	$7B,$B1,$FE,$AD,$8F,$81,$F8,$BC,$5F,$CF,$63,$A6,$3E,$2E,$3F,$6D
	.dc.b	$70,$C1,$B7,$9F,$67,$D3,$6A,$BC,$B5,$AD,$BC,$B5,$EB,$FD,$9F,$C9
	.dc.b	$87,$1D,$DF,$A3,$62,$F7,$0F,$DD,$71,$AF,$B7,$CB,$FB,$DF,$9A,$7F
	.dc.b	$BB,$BF,$0D,$7E,$30,$FC,$EB,$F6,$9E,$23,$F3,$2D,$FE,$41,$FE,$AB
	.dc.b	$AD,$7B,$3F,$DC,$FF,$5B,$CE,$3F,$17,$95,$9F,$FB,$EB,$F4,$3C,$3D
	.dc.b	$F0,$E3,$F9,$A7,$BE,$31,$AF,$F0,$3D,$31,$A7,$78,$F1,$FE,$FF,$DB
	.dc.b	$0F,$0E,$CF,$D1,$3F,$31,$E2,$7F,$33,$E0,$7E,$DD,$F7,$3C,$06,$6F
	.dc.b	$CC,$F7,$D1,$FC,$46,$B5,$2F,$9D,$CE,$1E,$BC,$7F,$9F,$7C,$05,$C3
	.dc.b	$87,$47,$E1,$3D,$5F,$B1,$5E,$1C,$BC,$1E,$BE,$2F,$7A,$F8,$B7,$EB
	.dc.b	$5E,$0F,$DB,$BF,$DB,$FF,$AA,$7F,$81,$7F,$55,$7E,$91,$FE,$1C,$F2
	.dc.b	$1E,$F7,$F8,$7B,$F2,$D7,$D5,$F0,$EF,$FF,$14,$61,$C3,$EB,$BF,$3B
	.dc.b	$7B,$FB,$3D,$52,$BA,$FE,$B9,$FC,$F3,$4F,$06,$9F,$E2,$AA,$FD,$FC
	.dc.b	$63,$1E,$67,$FB,$CE,$39,$DF,$7D,$3C,$B3,$FD,$FA,$73,$7E,$9C,$BF
	.dc.b	$C6,$BF,$A8,$FF,$8B,$BD,$26,$5D,$71,$E3,$2F,$CD,$F8,$E5,$E0,$87
	.dc.b	$E5,$38,$AD,$FB,$0C,$5F,$96,$7F,$8B,$7F,$63,$7E,$EA,$FD,$7F,$FC
	.dc.b	$81,$FA,$C7,$F3,$EF,$92,$7E,$35,$C7,$0E,$17,$BF,$B4,$5E,$FE,$53
	.dc.b	$5C,$BF,$E5,$5C,$FA,$FB,$9F,$D9,$31,$E3,$A5,$AD,$65,$CF,$1F,$E9
	.dc.b	$BF,$29,$FE,$47,$BD,$E7,$DB,$E4,$7F,$C1,$58,$76,$19,$FA,$B2,$F4
	.dc.b	$CF,$FA,$3F,$E9,$75,$39,$BE,$5F,$83,$3F,$2D,$F8,$9F,$59,$FE,$4F
	.dc.b	$F7,$1F,$B4,$BF,$40,$FE,$E9,$E3,$FF,$90,$BF,$2F,$7E,$49,$C3,$83
	.dc.b	$FC,$E5,$87,$0B,$DF,$D8,$AD,$FA,$AE,$EF,$F9,$76,$5D,$5F,$E8,$F9
	.dc.b	$F8,$61,$FA,$0F,$F8,$2A,$31,$8F,$4F,$3F,$FC,$D5,$18,$FA,$1F,$F3
	.dc.b	$F6,$7F,$C9,$31,$ED,$A7,$8E,$73,$C9,$E8,$EE,$FC,$A5,$78,$B2,$F3
	.dc.b	$CF,$9A,$73,$7C,$78,$3F,$CC,$3E,$97,$FA,$3F,$DC,$FC,$4E,$3F,$E9
	.dc.b	$1F,$CC,$FF,$D8,$DF,$AD,$BF,$4D,$FF,$73,$FE,$49,$FD,$23,$F0,$BF
	.dc.b	$E4,$DD,$DF,$E7,$BB,$DD,$6B,$C7,$7E,$81,$AC,$BA,$7D,$53,$F8,$03
	.dc.b	$FC,$5B,$AF,$82,$58,$F8,$1F,$F2,$54,$66,$3A,$FE,$2E,$D7,$FC,$6F
	.dc.b	$5D,$4E,$38,$36,$F4,$56,$FF,$0E,$F7,$6D,$E7,$AD,$6D,$E7,$FD,$33
	.dc.b	$96,$7D,$B3,$FE,$CB,$F8,$56,$2F,$79,$78,$73,$7E,$B5,$97,$BD,$F8
	.dc.b	$78,$BF,$A5,$7A,$9F,$E9,$FF,$CF,$1F,$9F,$3F,$6B,$FF,$2B,$53,$F6
	.dc.b	$77,$E2,$FF,$D5,$F9,$7F,$82,$F2,$FF,$B0,$F1,$E3,$87,$0B,$DF,$CA
	.dc.b	$F6,$AD,$64,$FC,$BF,$E6,$E9,$E1,$8F,$04,$FC,$C8,$C6,$1C,$F4,$FC
	.dc.b	$14,$63,$D6,$47,$97,$F2,$DC,$DF,$6B,$F1,$6C,$FD,$B3,$C5,$2D,$BC
	.dc.b	$F8,$F2,$DB,$DB,$F0,$1A,$F7,$1F,$2E,$D3,$F7,$FF,$BE,$BD,$57,$E1
	.dc.b	$6F,$33,$85,$EF,$D6,$5E,$EB,$FC,$6F,$CF,$F2,$37,$67,$DB,$AF,$E1
	.dc.b	$C7,$FB,$B6,$3C,$3A,$62,$37,$97,$5C,$7C,$5F,$07,$F0,$27,$35,$E6
	.dc.b	$D5,$FA,$1E,$71,$E7,$AF,$92,$7F,$DD,$BF,$AF,$7A,$2F,$87,$E2,$EF
	.dc.b	$D1,$FF,$C1,$DF,$4D,$FE,$E3,$FD,$C1,$FB,$1F,$E0,$FF,$73,$FF,$2E
	.dc.b	$FE,$8A,$FD,$55,$FA,$97,$0E,$10,$C3,$8F,$53,$0E,$36,$FD,$43,$85
	.dc.b	$EF,$2F,$C8,$3F,$AE,$CF,$C5,$1F,$E2,$28,$C1,$DB,$1E,$B8,$79,$36
	.dc.b	$FE,$11,$7C,$32,$F0,$7D,$93,$5F,$8B,$8F,$83,$1E,$33,$9D,$1E,$6E
	.dc.b	$74,$F8,$13,$9F,$33,$19,$74,$CB,$E1,$7E,$E3,$8D,$BD,$BF,$39,$EC
	.dc.b	$F9,$1F,$D0,$7E,$36,$F3,$3F,$A1,$7E,$9E,$FD,$73,$FD,$85,$E1,$7F
	.dc.b	$98,$6F,$75,$A9,$7E,$A9,$CF,$CF,$63,$FC,$5B,$9B,$E7,$F3,$7F,$75
	.dc.b	$9F,$E2,$2F,$FB,$93,$2E,$91,$C3,$1F,$0F,$81,$F1,$AB,$EB,$E7,$E0
	.dc.b	$F8,$96,$BF,$A1,$9F,$06,$7E,$99,$CE,$BF,$EF,$AC,$FA,$CE,$6F,$D9
	.dc.b	$F8,$9F,$DC,$FD,$AB,$FD,$B3,$F6,$2F,$D8,$FF,$AA,$FE,$92,$FC,$3D
	.dc.b	$F9,$A6,$FF,$11,$FF,$38,$D6,$A7,$F1,$DA,$7F,$A1,$71,$FF,$71,$FD
	.dc.b	$2F,$C1,$E0,$FC,$7C,$76,$7F,$0A,$1F,$0D,$F8,$3D,$DF,$81,$39,$C7
	.dc.b	$BF,$C8,$34,$FF,$13,$65,$DE,$73,$1E,$7F,$CE,$BC,$DB,$7A,$EF,$86
	.dc.b	$AF,$DB,$3F,$CE,$DF,$86,$FD,$4F,$CC,$DF,$BF,$BF,$6E,$FF,$4C,$F8
	.dc.b	$4F,$DE,$DF,$AC,$BF,$0F,$E3,$C5,$FA,$AC,$78,$E1,$C2,$F7,$F2,$1A
	.dc.b	$A5,$76,$F7,$3F,$AC,$E9,$E1,$D3,$C1,$5F,$D6,$23,$18,$F3,$FB,$FF
	.dc.b	$ED,$D1,$8F,$C1,$3F,$4D,$9F,$F0,$0B,$ED,$97,$8A,$73,$AF,$FB,$E3
	.dc.b	$99,$87,$1D,$FD,$93,$F9,$69,$CE,$1E,$BD,$3E,$16,$1C,$BE,$5B,$AC
	.dc.b	$63,$0E,$3F,$84,$7F,$04,$FD,$97,$FA,$47,$BA,$FD,$4F,$FD,$C3,$FA
	.dc.b	$9F,$FA,$8F,$E0,$B5,$E5,$FF,$9D,$27,$3B,$7F,$39,$3F,$D9,$98,$72
	.dc.b	$DB,$F0,$AF,$F5,$3F,$D8,$7F,$15,$08,$EF,$5D,$BF,$02,$FC,$3C,$63
	.dc.b	$F0,$C6,$6F,$A8,$FE,$3F,$F3,$BE,$EC,$7F,$CC,$7F,$B7,$3F,$76,$FE
	.dc.b	$3D,$F1,$4F,$CE,$1F,$B3,$FF,$03,$C3,$87,$D9,$B0,$E1,$F0,$9F,$D3
	.dc.b	$AF,$7F,$AD,$FF,$9A,$FB,$7F,$7F,$7B,$FD,$4B,$F8,$0E,$BF,$C0,$B2
	.dc.b	$E7,$97,$86,$32,$71,$3F,$9A,$D7,$7D,$7C,$B9,$3F,$4F,$F5,$E5,$DB
	.dc.b	$0F,$1E,$DE,$9F,$97,$CE,$5F,$CF,$BF,$71,$FD,$E7,$D2,$3F,$0F,$FD
	.dc.b	$F7,$C1,$3E,$FF,$FC,$B9,$C1,$F1,$6F,$75,$A9,$78,$1E,$5C,$55,$FC
	.dc.b	$1F,$E7,$FD,$31,$98,$EE,$FB,$30,$E9,$EB,$33,$F1,$3F,$D3,$FB,$3C
	.dc.b	$F5,$D2,$35,$9F,$9B,$F0,$BC,$F9,$BF,$9A,$F5,$5F,$AC,$7F,$15,$F5
	.dc.b	$1F,$8F,$63,$5F,$BD,$7B,$AD,$4F,$E0,$E7,$3C,$DF,$E0,$AD,$3C,$CF
	.dc.b	$F3,$6B,$ED,$18,$3F,$B3,$BC,$FD,$5F,$F3,$CF,$A3,$E4,$AF,$E2,$98
	.dc.b	$F8,$FE,$E7,$1A,$79,$AF,$FC,$FF,$9B,$F7,$FF,$95,$97,$3D,$7D,$3F
	.dc.b	$AB,$E3,$3E,$78,$70,$FD,$F7,$19,$73,$BF,$7F,$CD,$FD,$3F,$D6,$FF
	.dc.b	$C3,$DE,$7B,$F5,$A7,$E9,$4F,$DF,$3F,$11,$F8,$AE,$F7,$5E,$EB,$5E
	.dc.b	$23,$F2,$6D,$4F,$DC,$E8,$FD,$46,$7F,$E2,$BF,$A2,$C7,$17,$F5,$C8
	.dc.b	$F0,$EB,$E0,$D7,$F2,$DD,$3C,$A3,$1F,$6F,$F7,$AB,$7A,$F8,$3F,$14
	.dc.b	$D3,$E2,$FC,$5A,$E3,$F9,$FF,$99,$BC,$1F,$E0,$0D,$BF,$75,$F0,$5F
	.dc.b	$BE,$F8,$4F,$F3,$0F,$E7,$8F,$D2,$1F,$A6,$3F,$74,$FC,$C5,$EF,$EB
	.dc.b	$D7,$BA,$7E,$6B,$AF,$6F,$F9,$F7,$2F,$E0,$4F,$E8,$8B,$E2,$BE,$0C
	.dc.b	$7F,$8A,$B1,$E9,$87,$87,$EE,$7F,$37,$18,$D1,$F8,$C6,$31,$FE,$95
	.dc.b	$A7,$8A,$7C,$FF,$27,$E1,$DF,$D7,$1E,$AF,$8F,$C5,$6F,$77,$C3,$E3
	.dc.b	$5F,$57,$C2,$FE,$D7,$F5,$2F,$F7,$0F,$F0,$0F,$F7,$9F,$C8,$9F,$B3
	.dc.b	$7F,$8E,$FF,$6C,$D7,$D9,$2B,$53,$FE,$3F,$F8,$BC,$FE,$05,$EE,$E7
	.dc.b	$65,$D1,$F8,$8F,$6F,$3D,$9F,$6F,$4F,$F7,$73,$98,$F0,$FA,$B8,$D3
	.dc.b	$C1,$7C,$33,$FF,$27,$7E,$8D,$E4,$D7,$C7,$F1,$2F,$CF,$71,$7B,$FB
	.dc.b	$1F,$E3,$1E,$17,$F3,$BD,$7B,$65,$FE,$5E,$FC,$7F,$F8,$B7,$C7,$7F
	.dc.b	$BD,$3F,$35,$7E,$8E,$FD,$E7,$FC,$C1,$F9,$7B,$E0,$F7,$F8,$8A,$F7
	.dc.b	$5A,$F0,$B3,$1F,$0F,$1F,$F7,$FF,$E6,$F6,$F8,$65,$D7,$ED,$DF,$AF
	.dc.b	$7F,$1C,$BF,$14,$E7,$ED,$B1,$7C,$33,$F1,$FC,$DF,$6A,$E5,$B7,$FA
	.dc.b	$FF,$8B,$FA,$C6,$3C,$AF,$FC,$F1,$FB,$CB,$F0,$1E,$11,$8F,$CE,$7F
	.dc.b	$9B,$7C,$63,$F2,$37,$E3,$AF,$C6,$3F,$C7,$9F,$85,$BF,$23,$78,$BC
	.dc.b	$2F,$75,$A9,$FF,$44,$ED,$FD,$A7,$1E,$9F,$8F,$7E,$15,$ED,$95,$C3
	.dc.b	$8F,$7F,$FB,$BF,$F1,$6F,$E7,$B5,$AC,$BC,$3E,$76,$2F,$7A,$BF,$23
	.dc.b	$7D,$57,$8F,$3F,$46,$7F,$EF,$99,$FF,$71,$EB,$E3,$C3,$8F,$EE,$CF
	.dc.b	$C5,$78,$69,$EE,$8F,$E6,$14,$FB,$EF,$F3,$9F,$E2,$6F,$BC,$7E,$50
	.dc.b	$E6,$7E,$38,$FC,$FB,$F2,$5F,$A5,$B2,$E5,$EA,$7F,$AA,$B2,$E5,$8F
	.dc.b	$1C,$38,$79,$CF,$87,$E1,$AF,$83,$57,$E9,$5F,$CB,$63,$18,$5C,$BA
	.dc.b	$57,$93,$6F,$1E,$DF,$E0,$9B,$DC,$63,$AB,$FC,$3B,$18,$D3,$D1,$EB
	.dc.b	$3F,$48,$F0,$E5,$D3,$3F,$36,$5F,$B8,$FB,$2C,$CB,$83,$D7,$A7,$C2
	.dc.b	$F7,$9F,$1F,$F2,$B5,$ED,$18,$E7,$E3,$7C,$36,$FE,$63,$F2,$DF,$E3
	.dc.b	$0F,$DD,$BF,$D4,$3F,$91,$BF,$4C,$7E,$5E,$FC,$99,$F0,$6F,$C6,$BF
	.dc.b	$DB,$7D,$7F,$3F,$1E,$38,$70,$BD,$D7,$FB,$83,$E0,$35,$C7,$FE,$DC
	.dc.b	$CF,$B7,$DA,$FF,$59,$C7,$8E,$96,$B5,$97,$3D,$3C,$1E,$AF,$9E,$73
	.dc.b	$5E,$59,$F2,$7A,$D9,$96,$5E,$BA,$D6,$9F,$AA,$F0,$BC,$75,$AC,$BB
	.dc.b	$4F,$EB,$BC,$5C,$E1,$C7,$7F,$5F,$9D,$D6,$31,$A7,$C3,$F8,$3C,$63
	.dc.b	$CB,$F2,$47,$EF,$5C,$A3,$1E,$75,$DF,$C7,$5C,$1F,$2F,$E6,$FF,$F0
	.dc.b	$03,$BD,$F8,$FF,$FB,$53,$F1,$27,$E2,$CF,$D5,$9F,$76,$FD,$09,$F9
	.dc.b	$F7,$FB,$8F,$F4,$D6,$5C,$BA,$03,$2E,$78,$F2,$C3,$8F,$E2,$5F,$EB
	.dc.b	$FB,$E1,$7B,$CF,$F5,$EF,$F9,$8A,$33,$87,$28,$C6,$9E,$7A,$F3,$69
	.dc.b	$E5,$D1,$F8,$A7,$B6,$B5,$A7,$86,$7F,$DC,$D9,$74,$C3,$8E,$9E,$C5
	.dc.b	$F9,$5F,$B3,$0E,$39,$75,$97,$E1,$19,$D3,$DF,$AF,$BA,$BC,$5F,$57
	.dc.b	$FD,$37,$3E,$DA,$F8,$61,$FA,$D7,$1C,$3E,$04,$F9,$7E,$FF,$FD,$7B
	.dc.b	$1F,$2B,$F3,$D7,$ED,$0F,$CE,$7F,$C9,$1F,$46,$F1,$DF,$B7,$BF,$74
	.dc.b	$FF,$67,$61,$C2,$F7,$5A,$FC,$5F,$38,$FE,$63,$2F,$F6,$71,$DD,$EC
	.dc.b	$E0,$5F,$3F,$BB,$FF,$72,$5F,$CE,$9E,$78,$76,$EF,$F4,$C7,$0C,$3D
	.dc.b	$17,$C3,$3F,$C7,$F0,$FF,$AC,$72,$E9,$AF,$AE,$5F,$97,$74,$BE,$FD
	.dc.b	$6F,$A9,$C5,$FC,$18,$FF,$1E,$E3,$CF,$2E,$F3,$F1,$7F,$23,$F1,$FE
	.dc.b	$B6,$FB,$47,$E6,$4F,$C9,$3F,$C2,$FF,$B0,$7F,$99,$F8,$3F,$DE,$D7
	.dc.b	$BA,$D7,$C1,$2B,$FC,$D9,$F9,$EE,$B1,$E7,$C7,$F1,$F0,$ED,$FA,$DF
	.dc.b	$FB,$F2,$3F,$47,$1C,$F3,$EF,$FB,$87,$92,$FC,$71,$E2,$FD,$17,$1A
	.dc.b	$78,$33,$7F,$00,$BF,$2C,$F0,$4E,$63,$D9,$C7,$FE,$DB,$CF,$B6,$FE
	.dc.b	$79,$FF,$0D,$4E,$73,$EF,$39,$FD,$7F,$F7,$3F,$80,$7E,$E3,$7F,$F5
	.dc.b	$EF,$C1,$3F,$6E,$FD,$EF,$E9,$6F,$CE,$1F,$B2,$FF,$A1,$74,$FF,$75
	.dc.b	$7E,$29,$FD,$1E,$F7,$5A,$FC,$23,$AE,$5F,$C1,$2B,$FA,$67,$EC,$3F
	.dc.b	$16,$3B,$DD,$9E,$58,$32,$F0,$7C,$27,$E3,$27,$31,$E3,$F7,$31,$AF
	.dc.b	$9E,$F8,$68,$FD,$CD,$FB,$1F,$8B,$4E,$F5,$A9,$FF,$75,$3E,$AB,$E5
	.dc.b	$8F,$BE,$7F,$A9,$4E,$74,$F6,$C3,$F3,$9E,$9F,$86,$FF,$A3,$7E,$F9
	.dc.b	$F9,$13,$C4,$7E,$42,$FD,$5B,$FC,$21,$9B,$9F,$E8,$63,$C7,$0E,$17
	.dc.b	$B9,$EB,$F8,$43,$F0,$4F,$91,$7D,$F8,$7F,$6C,$F1,$CF,$BF,$EA,$BF
	.dc.b	$B2,$8C,$63,$FA,$1A,$E7,$AF,$9A,$FC,$BF,$9A,$3F,$6D,$CE,$BE,$28
	.dc.b	$7E,$D9,$8A,$DF,$45,$C9,$4F,$DA,$7E,$77,$CE,$3F,$62,$7C,$3B,$F4
	.dc.b	$8F,$F0,$07,$D5,$FF,$B8,$7E,$77,$75,$A9,$CF,$AE,$FC,$8B,$38,$7E
	.dc.b	$1F,$1E,$9E,$85,$C3,$87,$C8,$33,$FE,$0D,$CB,$C3,$3E,$FF,$99,$7F
	.dc.b	$14,$C7,$C9,$97,$8F,$0E,$1A,$3F,$67,$E1,$18,$FC,$12,$E7,$FE,$F7
	.dc.b	$FD,$AA,$DB,$F1,$EF,$C6,$7C,$1B,$F6,$E7,$E9,$2F,$CC,$DF,$B4,$FC
	.dc.b	$5F,$D6,$DF,$9C,$F0,$BD,$D6,$BA,$54,$FD,$87,$5F,$8B,$FE,$9F,$3E
	.dc.b	$D9,$DC,$78,$E7,$F4,$7E,$03,$F9,$8D,$6B,$3E,$9F,$92,$E2,$F7,$A7
	.dc.b	$6F,$3F,$C5,$97,$9F,$2F,$37,$EB,$38,$C7,$C9,$B3,$F6,$4C,$57,$AF
	.dc.b	$F1,$8F,$D6,$7D,$BF,$C1,$E6,$FC,$07,$E5,$7F,$EC,$EF,$C9,$5F,$84
	.dc.b	$EB,$83,$EB,$78,$4A,$D7,$DB,$67,$F6,$4F,$E0,$99,$73,$F1,$3F,$A9
	.dc.b	$3D,$6F,$87,$E2,$AE,$8F,$E9,$F3,$9C,$7F,$DB,$BF,$17,$8C,$FB,$C7
	.dc.b	$FB,$82,$BF,$19,$EB,$FE,$C7,$2F,$F0,$BE,$1C,$71,$E9,$E5,$79,$F8
	.dc.b	$3D,$70,$FD,$53,$FC,$31,$EC,$7F,$27,$BE,$10,$FA,$6E,$1B,$F9,$BB
	.dc.b	$7F,$B1,$FE,$42,$FD,$F9,$FB,$F3,$F7,$97,$EC,$EF,$D9,$9F,$A0,$3F
	.dc.b	$24,$FE,$31,$8F,$1C,$38,$5E,$EB,$C7,$C7,$FE,$30,$AD,$3F,$DE,$5F
	.dc.b	$07,$87,$FC,$BD,$E5,$97,$83,$27,$B1,$E6,$D3,$CB,$39,$F8,$AC,$69
	.dc.b	$E0,$B7,$E6,$8F,$DD,$BC,$B8,$70,$DB,$D5,$2F,$D5,$73,$87,$1C,$DF
	.dc.b	$B7,$E3,$4F,$14,$7C,$2E,$BF,$A9,$C1,$F0,$BF,$61,$7E,$48,$FC,$F9
	.dc.b	$C4,$F3,$1F,$BA,$7F,$16,$C7,$8F,$33,$1E,$38,$70,$BD,$FC,$2F,$F4
	.dc.b	$3A,$D6,$5F,$D4,$DE,$1F,$93,$DC,$38,$E7,$DF,$F2,$0F,$F0,$ED,$70
	.dc.b	$CB,$A7,$EE,$18,$BD,$E7,$FE,$BC,$FC,$8F,$B4,$63,$F1,$2F,$91,$B7
	.dc.b	$9A,$37,$F8,$87,$1D,$BD,$DA,$FB,$72,$FF,$57,$FE,$29,$FE,$24,$E2
	.dc.b	$FF,$36,$FD,$4F,$FB,$4F,$E5,$7F,$D6,$DF,$83,$BF,$5C,$7E,$91,$FC
	.dc.b	$6F,$0E,$18,$7F,$92,$BF,$13,$FE,$C9,$F3,$7F,$EA,$DB,$DF,$E6,$3F
	.dc.b	$ED,$BF,$B1,$FF,$01,$69,$FE,$DF,$F1,$BF,$C0,$98,$FF,$81,$72,$E7
	.dc.b	$3E,$18,$C9,$C4,$F2,$E5,$DB,$4F,$27,$E0,$BF,$E8,$4B,$E3,$39,$CD
	.dc.b	$F8,$27,$F9,$DF,$0F,$6F,$83,$FD,$AB,$F0,$B7,$EC,$8F,$D4,$DF,$B5
	.dc.b	$BF,$78,$FF,$35,$6F,$F8,$CF,$9C,$61,$C2,$F7,$F9,$EE,$B2,$FE,$EB
	.dc.b	$9F,$FA,$28,$EF,$F9,$F7,$A9,$D6,$1C,$B4,$F0,$7B,$FE,$59,$FF,$39
	.dc.b	$DF,$96,$7C,$9D,$5F,$EA,$BF,$C6,$BF,$C5,$5F,$8A,$3F,$5A,$7E,$2E
	.dc.b	$E6,$7E,$67,$FE,$77,$FD,$B9,$FB,$17,$FA,$AB,$1E,$38,$70,$BD,$FE
	.dc.b	$0C,$B7,$EC,$7B,$DF,$FD,$5F,$A7,$7D,$3D,$5E,$43,$3E,$7F,$8E,$61
	.dc.b	$EB,$3E,$3D,$BF,$D1,$1F,$83,$7A,$71,$E7,$B7,$86,$7F,$D1,$9F,$90
	.dc.b	$FF,$9F,$3C,$27,$EA,$5F,$E6,$AF,$91,$7E,$DC,$CB,$F8,$E7,$E3,$3F
	.dc.b	$E2,$FB,$DF,$E0,$2B,$DD,$7F,$21,$FD,$2F,$58,$FF,$9A,$B5,$FC,$7F
	.dc.b	$0F,$F0,$2C,$7B,$FE,$0B,$FE,$47,$79,$E3,$CB,$D8,$F2,$61,$E3,$D7
	.dc.b	$C5,$F9,$DF,$FA,$4B,$E8,$BF,$81,$7E,$97,$FD,$D3,$F1,$5F,$F0,$F7
	.dc.b	$E2,$0F,$C0,$FF,$C8,$95,$AF,$8B,$FF,$6C,$7C,$0A,$9F,$A5,$EB,$DC
	.dc.b	$7E,$79,$FE,$59,$C3,$AF,$27,$FA,$9D,$E5,$9F,$F8,$AF,$F0,$EF,$15
	.dc.b	$B5,$F1,$67,$FE,$A9,$E5,$FE,$3F,$F4,$AF,$29,$83,$F3,$87,$E8,$AF
	.dc.b	$DD,$3F,$C1,$3F,$AC,$FF,$96,$F0,$E1,$7B,$AD,$7D,$6F,$F8,$C7,$2E
	.dc.b	$B9,$3F,$1A,$FF,$38,$61,$C3,$87,$FD,$7C,$F2,$D3,$BE,$9F,$E1,$69
	.dc.b	$ED,$7E,$49,$CF,$CC,$FE,$EF,$F1,$EF,$F4,$47,$E7,$EF,$C3,$9F,$4E
	.dc.b	$FD,$35,$F8,$63,$F7,$E7,$E9,$9F,$E4,$3F,$57,$FE,$00,$E3,$C7,$0E
	.dc.b	$17,$BA,$FE,$01,$5A,$C7,$FC,$17,$B7,$F1,$08,$C5,$FE,$E5,$B9,$78
	.dc.b	$32,$FF,$21,$C6,$31,$F1,$7D,$EF,$A7,$00,$CF,$C5,$87,$A3,$CC,$7E
	.dc.b	$F6,$FD,$F3,$F7,$5B,$FC,$0D,$BD,$F7,$BA,$FF,$01,$69,$DF,$06,$7D
	.dc.b	$F3,$ED,$87,$0D,$15,$FE,$23,$9E,$A7,$93,$4F,$1C,$FE,$A7,$C0,$F9
	.dc.b	$78,$FA,$FC,$9F,$C9,$FC,$3F,$0C,$3A,$F0,$FC,$A9,$E9,$5B,$C7,$F9
	.dc.b	$97,$A9,$FE,$BF,$FC,$C5,$F9,$4B,$F3,$7F,$E6,$7E,$8B,$CF,$FF,$78
	.dc.b	$7C,$47,$F1,$9C,$38,$5E,$FE,$25,$7B,$FC,$7D,$FC,$03,$BE,$0F,$9B
	.dc.b	$A7,$6F,$CF,$3E,$B6,$5C,$B5,$B5,$AC,$FA,$65,$CE,$39,$0C,$3C,$F5
	.dc.b	$E6,$D5,$F2,$3E,$CA,$E5,$3E,$AF,$D0,$BE,$E7,$C1,$F2,$C3,$8E,$FE
	.dc.b	$89,$FC,$07,$A5,$9D,$3C,$31,$F8,$3C,$71,$8F,$3F,$B3,$ED,$1E,$3A
	.dc.b	$F3,$FE,$E0,$FC,$E1,$F9,$47,$0C,$FE,$2C,$E7,$F3,$7C,$70,$A1,$FB
	.dc.b	$0F,$4D,$FC,$B3,$9F,$87,$FE,$D1,$F6,$9F,$DD,$3F,$67,$7C,$17,$E4
	.dc.b	$FF,$6C,$7A,$EF,$C9,$DF,$71,$FD,$3F,$0E,$17,$BA,$D7,$EE,$19,$E4
	.dc.b	$FF,$05,$E5,$FE,$96,$7F,$59,$7C,$1A,$6A,$F9,$70,$F9,$23,$1E,$73
	.dc.b	$4F,$1F,$E6,$F9,$97,$CD,$32,$F4,$DE,$F4,$7D,$D7,$F7,$8F,$1B,$FC
	.dc.b	$83,$7B,$D3,$D1,$2F,$CE,$7D,$19,$74,$CB,$9E,$9F,$0B,$DC,$F0,$63
	.dc.b	$C7,$4E,$DE,$0E,$2F,$7F,$75,$FE,$DF,$FC,$3F,$9D,$EF,$1F,$F6,$C7
	.dc.b	$C0,$BF,$0E,$CE,$1C,$3E,$F3,$97,$07,$D1,$C9,$F8,$DF,$BF,$F3,$57
	.dc.b	$E8,$5F,$C8,$3D,$7F,$F2,$AF,$EC,$EE,$8F,$7F,$3F,$F9,$7A,$F7,$EB
	.dc.b	$FF,$29,$FB,$65,$EF,$C5,$F7,$D6,$BA,$3E,$98,$7B,$FE,$6B,$FC,$87
	.dc.b	$BB,$F8,$DC,$E7,$5F,$16,$3D,$34,$F3,$46,$32,$FF,$40,$D7,$68,$C7
	.dc.b	$E3,$2F,$CE,$F1,$AF,$96,$FF,$DE,$98,$63,$7F,$6F,$AB,$D2,$B5,$EE
	.dc.b	$D9,$FF,$3A,$FE,$27,$FD,$73,$EE,$38,$6D,$FE,$DD,$F8,$D7,$ED,$58
	.dc.b	$AF,$EC,$3C,$17,$F3,$5F,$C5,$7E,$97,$E3,$E3,$7B,$0F,$6D,$FB,$43
	.dc.b	$F4,$C7,$E9,$1F,$86,$C3,$85,$EE,$B5,$F6,$19,$E7,$FF,$47,$E9,$FE
	.dc.b	$5D,$CF,$E4,$E1,$F0,$7F,$1E,$F9,$93,$FD,$1A,$39,$E9,$E0,$F8,$EF
	.dc.b	$2E,$0C,$3C,$B3,$E4,$F3,$33,$18,$FC,$D1,$F9,$E7,$FA,$57,$C2,$FF
	.dc.b	$4F,$E5,$D2,$73,$EF,$79,$6F,$87,$37,$2E,$1F,$81,$4F,$CE,$31,$A2
	.dc.b	$1F,$8D,$FE,$53,$3E,$BB,$79,$70,$E3,$3F,$E5,$39,$CE,$5E,$5F,$57
	.dc.b	$9E,$0F,$B8,$FD,$D7,$C3,$7F,$88,$3F,$01,$FF,$6B,$7B,$AF,$D8,$7F
	.dc.b	$D6,$5F,$77,$FB,$4A,$D7,$E1,$FF,$D1,$27,$FD,$6D,$F9,$76,$7E,$49
	.dc.b	$8F,$CC,$EA,$FF,$2B,$43,$F2,$2C,$63,$CB,$3E,$DB,$78,$AB,$8E,$7D
	.dc.b	$75,$C6,$5E,$4D,$1F,$97,$FF,$9B,$B8,$7F,$D9,$79,$F3,$9C,$FD,$4F
	.dc.b	$8E,$F8,$7C,$2F,$2C,$BD,$1F,$59,$76,$F7,$7C,$9B,$F6,$4E,$15,$ED
	.dc.b	$87,$E5,$38,$AD,$F1,$DD,$3F,$B8,$F3,$FE,$D0,$F9,$1F,$B0,$BF,$1F
	.dc.b	$7E,$67,$FE,$3E,$F8,$F7,$E4,$7F,$C7,$AF,$75,$AF,$A9,$FE,$91,$F8
	.dc.b	$86,$72,$65,$F2,$EF,$EA,$7D,$95,$9C,$E1,$C7,$E6,$BF,$9A,$57,$4D
	.dc.b	$3F,$C9,$BC,$D8,$BD,$C7,$F9,$43,$F5,$F7,$EC,$DD,$F3,$F4,$4E,$7F
	.dc.b	$46,$C6,$7E,$5F,$C8,$39,$56,$B7,$7C,$46,$2B,$D5,$FB,$CF,$FA,$EB
	.dc.b	$BB,$FD,$FF,$BF,$8E,$D3,$FE,$0A,$9C,$ED,$E4,$97,$EB,$3C,$F5,$FF
	.dc.b	$32,$FB,$1F,$D3,$FD,$8F,$F0,$1F,$E4,$6D,$FF,$C2,$BE,$47,$F4,$EF
	.dc.b	$BB,$FE,$93,$5A,$9C,$FC,$B7,$F8,$CB,$5F,$DD,$76,$7F,$91,$2B,$9B
	.dc.b	$FC,$C1,$FF,$07,$FF,$00,$7E,$65,$E5,$DC,$CB,$FC,$D1,$7C,$AF,$F8
	.dc.b	$A7,$CA,$B4,$F2,$5E,$E3,$D0,$67,$3F,$3F,$E2,$9F,$EC,$1E,$4F,$E7
	.dc.b	$79,$F6,$CD,$3F,$E6,$FC,$FB,$4E,$7E,$1F,$9E,$7D,$AF,$F9,$ED,$6B
	.dc.b	$2E,$D9,$F9,$FC,$57,$DC,$FC,$49,$CD,$FE,$7F,$F2,$DC,$63,$CF,$97
	.dc.b	$AE,$FF,$33,$93,$FB,$F7,$C3,$BF,$0D,$7E,$65,$FE,$6B,$FD,$0F,$E9
	.dc.b	$7E,$AF,$FE,$9C,$E1,$F8,$97,$BA,$D6,$2A,$FF,$4B,$56,$B0,$7E,$97
	.dc.b	$FC,$1F,$0E,$1C,$FF,$D8,$9E,$5E,$63,$3F,$C0,$7C,$AF,$93,$0F,$1C
	.dc.b	$E7,$3F,$C9,$E1,$FE,$0A,$C3,$8E,$0F,$C9,$DF,$8A,$3F,$31,$D6,$5C
	.dc.b	$AB,$52,$FC,$AB,$3C,$1E,$FC,$3F,$C4,$59,$63,$4F,$24,$7F,$8F,$A3
	.dc.b	$18,$77,$FC,$CB,$F7,$FC,$1E,$FC,$39,$43,$CF,$DE,$1F,$CF,$F0,$BF
	.dc.b	$09,$FE,$2D,$FC,$65,$FA,$07,$EB,$BF,$76,$7E,$05,$EA,$F9,$2F,$E8
	.dc.b	$5E,$2F,$F8,$12,$F7,$E7,$2F,$7C,$6B,$FE,$89,$7B,$C7,$E3,$E7,$FE
	.dc.b	$0C,$3F,$C6,$38,$F8,$3F,$0F,$FE,$74,$F4,$C7,$9F,$C7,$79,$6F,$C9
	.dc.b	$A7,$CC,$F0,$B1,$AF,$86,$3F,$C8,$9F,$17,$FD,$63,$93,$F5,$19,$73
	.dc.b	$9C,$FE,$45,$E0,$C7,$9F,$D1,$FC,$3B,$DC,$63,$8B,$AE,$BF,$0A,$9B
	.dc.b	$3F,$7F,$E5,$AF,$86,$73,$FA,$CE,$2F,$84,$F2,$D3,$FC,$B1,$DF,$FE
	.dc.b	$5B,$F9,$47,$E4,$2F,$D5,$DF,$A3,$BF,$71,$FE,$73,$F1,$5F,$E8,$3E
	.dc.b	$2F,$E0,$58,$70,$BD,$D6,$BE,$DB,$FC,$B7,$39,$E9,$FF,$1E,$BD,$BF
	.dc.b	$59,$FE,$CC,$F1,$CB,$B7,$C2,$BE,$2B,$C9,$AF,$8E,$73,$F1,$58,$D3
	.dc.b	$BD,$BF,$2A,$FF,$13,$ED,$E4,$C3,$86,$DF,$D4,$27,$FD,$13,$39,$BE
	.dc.b	$31,$CF,$1F,$0F,$E9,$4F,$90,$F4,$E3,$CB,$24,$BF,$10,$CE,$3E,$2F
	.dc.b	$92,$E5,$A7,$F5,$EF,$6F,$FD,$0D,$EC,$BF,$38,$7E,$0A,$FD,$79,$E2
	.dc.b	$7F,$9C,$FF,$09,$7E,$9F,$FE,$82,$C3,$85,$EE,$B5,$F6,$99,$C7,$9F
	.dc.b	$57,$F9,$A2,$FB,$7E,$73,$FE,$11,$78,$E7,$DA,$FF,$DE,$F5,$C7,$1E
	.dc.b	$91,$FD,$EA,$07,$F2,$8C,$3C,$F5,$C7,$67,$E5,$DC,$63,$1F,$73,$74
	.dc.b	$FF,$46,$57,$C0,$BE,$31,$FE,$8F,$DF,$19,$F8,$3F,$74,$FF,$79,$75
	.dc.b	$7F,$4B,$DB,$CF,$7F,$E9,$59,$7E,$E1,$D7,$7F,$EC,$BB,$DF,$2F,$FD
	.dc.b	$65,$F8,$1B,$F2,$67,$C2,$FF,$B1,$BC,$77,$E0,$4F,$98,$7E,$FD,$FE
	.dc.b	$AE,$C3,$85,$EE,$B5,$FB,$5E,$71,$E7,$9F,$FA,$52,$FB,$70,$7F,$8F
	.dc.b	$5E,$3F,$82,$FF,$14,$C3,$B5,$71,$C7,$A7,$D1,$F9,$B1,$F2,$DF,$F1
	.dc.b	$AE,$17,$E7,$8F,$8F,$E1,$39,$8F,$EF,$1E,$DF,$FA,$83,$4E,$FC,$2C
	.dc.b	$3F,$B9,$CE,$71,$EB,$39,$FA,$FF,$06,$9F,$0E,$DF,$99,$BE,$E7,$BD
	.dc.b	$F3,$87,$E2,$F8,$AF,$99,$C1,$7E,$8B,$FD,$C1,$FB,$EB,$F6,$AF,$17
	.dc.b	$D0,$BF,$00,$FF,$4F,$FE,$A2,$FD,$53,$FD,$B7,$7B,$CB,$FC,$77,$5A
	.dc.b	$FC,$33,$38,$FD,$6C,$FF,$D8,$37,$B6,$C7,$6A,$FF,$55,$FB,$9E,$28
	.dc.b	$C7,$DE,$3F,$1D,$C6,$5D,$BE,$67,$FC,$93,$8F,$68,$C7,$E6,$AF,$77
	.dc.b	$1C,$1E,$9B,$7B,$9C,$A3,$1E,$ED,$DB,$FC,$B7,$4F,$DF,$B8,$C6,$3F
	.dc.b	$3A,$BF,$4C,$AE,$0F,$E6,$3C,$D7,$D9,$FF,$57,$FD,$65,$FB,$6B,$F3
	.dc.b	$8F,$F3,$3F,$D7,$7E,$6B,$F6,$E6,$1C,$39,$3F,$6B,$E9,$B1,$E3,$87
	.dc.b	$0F,$0D,$E2,$70,$D3,$B6,$9F,$D7,$71,$FD,$EE,$5C,$DF,$E2,$38,$BF
	.dc.b	$48,$D4,$71,$3B,$C7,$6D,$BF,$CF,$98,$F2,$C7,$AE,$7E,$8F,$A3,$C6
	.dc.b	$1C,$33,$F6,$60,$FC,$DF,$FD,$61,$FB,$E7,$5A,$F0,$D6,$A7,$FD,$69
	.dc.b	$F9,$9E,$78,$3D,$D3,$9F,$43,$AD,$F0,$CF,$FD,$1B,$83,$DC,$CE,$1C
	.dc.b	$27,$FD,$23,$EB,$3F,$26,$7E,$61,$D7,$8B,$D9,$39,$FC,$1B,$1A,$FA
	.dc.b	$3F,$1E,$FE,$47,$AE,$32,$E7,$C5,$E2,$FD,$29,$F9,$7B,$F6,$F7,$ED
	.dc.b	$57,$F1,$47,$DE,$BF,$18,$F8,$18,$70,$BD,$8A,$DF,$DD,$6B,$17,$E9
	.dc.b	$9F,$53,$BF,$EC,$61,$C7,$6F,$B9,$18,$AF,$16,$9E,$1E,$2F,$34,$63
	.dc.b	$3E,$DD,$38,$7C,$3C,$8F,$D6,$5F,$8E,$78,$67,$31,$FE,$BB,$CB,$B4
	.dc.b	$70,$F2,$F9,$67,$F0,$70,$FF,$39,$F8,$1F,$23,$6F,$74,$FC,$5F,$87
	.dc.b	$FF,$0B,$7E,$71,$D6,$31,$F4,$6F,$C8,$B1,$9A,$DF,$25,$DA,$31,$F6
	.dc.b	$5F,$9A,$EA,$FA,$BF,$9A,$BF,$17,$FF,$3B,$69,$FE,$DB,$F9,$0F,$F0
	.dc.b	$C7,$E2,$1F,$AA,$BD,$D6,$BE,$FF,$FB,$9F,$CD,$3E,$B3,$5D,$9F,$D8
	.dc.b	$B3,$E9,$F7,$1F,$47,$1E,$3F,$65,$6B,$59,$73,$D1,$F5,$DE,$59,$CE
	.dc.b	$0D,$BC,$9F,$5F,$99,$67,$DA,$7F,$DE,$5F,$88,$3F,$11,$FE,$59,$DB
	.dc.b	$FE,$86,$BE,$75,$A9,$FF,$03,$CE,$67,$9E,$9E,$DB,$F1,$C7,$B0,$CE
	.dc.b	$5E,$3E,$14,$FF,$93,$F7,$F8,$B3,$E7,$F8,$4F,$F0,$BF,$D4,$3F,$64
	.dc.b	$E5,$B7,$9E,$73,$FA,$BE,$34,$FA,$31,$FE,$1E,$E1,$F7,$56,$F8,$BF
	.dc.b	$92,$6C,$FC,$DB,$F6,$8F,$F8,$43,$F2,$67,$EE,$7F,$EA,$4F,$DA,$1F
	.dc.b	$6D,$F2,$7E,$85,$F9,$CF,$1F,$63,$F6,$B8,$70,$E4,$61,$C3,$A7,$FD
	.dc.b	$13,$F8,$DE,$F4,$ED,$A7,$F3,$07,$F9,$26,$BE,$25,$8F,$37,$39,$76
	.dc.b	$AF,$26,$BE,$39,$E8,$70,$8C,$E2,$FC,$17,$FC,$57,$8F,$5C,$7C,$D9
	.dc.b	$FB,$3F,$0D,$FF,$17,$FE,$2B,$CA,$73,$9A,$7F,$C6,$7D,$79,$CF,$C3
	.dc.b	$1F,$C5,$31,$E9,$18,$CF,$BE,$BE,$7D,$FE,$1E,$2F,$D7,$BE,$4F,$E8
	.dc.b	$9C,$E7,$31,$F3,$3F,$49,$C6,$FF,$42,$1C,$5D,$38,$BE,$4D,$F0,$F1
	.dc.b	$BD,$BF,$43,$FC,$9B,$E4,$9F,$B9,$FE,$73,$E8,$5F,$94,$3F,$65,$7E
	.dc.b	$BC,$F2,$7F,$05,$8F,$1C,$38,$5E,$EB,$D1,$CD,$FE,$8E,$9F,$E7,$3A
	.dc.b	$FD,$8B,$E1,$DF,$FD,$62,$B4,$FF,$87,$70,$F0,$7C,$47,$9A,$FF,$CB
	.dc.b	$BA,$F9,$AB,$5A,$3D,$BF,$AF,$F1,$1F,$D0,$3C,$9F,$B5,$EF,$BF,$18
	.dc.b	$7C,$33,$ED,$94,$FD,$83,$F3,$98,$71,$F1,$7F,$9C,$E1,$C2,$F7,$9E
	.dc.b	$B6,$FF,$37,$E3,$D9,$FC,$9E,$1C,$7F,$09,$F6,$3C,$F4,$F1,$69,$E1
	.dc.b	$FC,$BF,$CF,$7C,$B4,$EF,$3E,$5F,$C0,$BE,$04,$BF,$16,$7E,$E2,$FC
	.dc.b	$91,$F9,$8B,$BF,$FC,$27,$BB,$EA,$BE,$AE,$3F,$E8,$4C,$78,$E1,$C2
	.dc.b	$F7,$FA,$FE,$AB,$B6,$4F,$BA,$FA,$9D,$BF,$DA,$B1,$E3,$5F,$64,$E9
	.dc.b	$5E,$3D,$3C,$5F,$26,$E1,$F4,$DF,$82,$73,$BE,$2F,$86,$8F,$D9,$DF
	.dc.b	$86,$7F,$3B,$F0,$1A,$F8,$6F,$75,$AF,$82,$F4,$67,$D7,$1E,$5F,$09
	.dc.b	$D7,$8B,$DD,$6F,$1F,$FC,$53,$E7,$7F,$81,$3F,$62,$7E,$AE,$FD,$01
	.dc.b	$F9,$7F,$CD,$F5,$FF,$3D,$F7,$3F,$E3,$AC,$38,$5E,$EB,$5E,$6B,$F5
	.dc.b	$9D,$75,$7F,$AC,$B3,$EB,$9F,$F8,$07,$1E,$3E,$C5,$AD,$65,$CF,$7F
	.dc.b	$E3,$39,$79,$A7,$38,$79,$27,$F8,$EF,$BD,$FE,$2A,$FD,$B9,$F9,$8B
	.dc.b	$F3,$CF,$F8,$03,$F3,$8F,$D8,$9F,$18,$FC,$4F,$FC,$71,$87,$0B,$DD
	.dc.b	$6B,$F1,$3C,$E2,$FC,$C7,$FC,$31,$F3,$1F,$E2,$AC,$38,$6F,$FE,$18
	.dc.b	$8C,$57,$87,$F1,$2F,$F1,$36,$3F,$81,$AE,$79,$76,$F1,$3F,$CB,$3E
	.dc.b	$17,$F8,$D3,$F7,$AF,$D8,$7E,$10,$FD,$83,$FC,$15,$E3,$7F,$93,$FF
	.dc.b	$56,$FF,$34,$DE,$FD,$85,$EF,$E6,$BF,$A9,$7D,$53,$F7,$6D,$ED,$FE
	.dc.b	$E2,$D3,$AB,$FE,$2E,$C7,$97,$E1,$9F,$E3,$37,$A6,$3C,$F4,$FE,$15
	.dc.b	$5C,$85,$F9,$B5,$F2,$FD,$37,$AA,$7F,$DD,$FE,$8B,$F0,$FF,$EC,$3F
	.dc.b	$39,$F8,$CF,$D2,$BF,$A1,$7E,$29,$CE,$B5,$F2,$3F,$E1,$7F,$15,$5F
	.dc.b	$D1,$FF,$31,$D6,$3F,$DA,$B2,$F4,$1D,$97,$EA,$9E,$98,$C6,$1C,$EF
	.dc.b	$FA,$84,$72,$8C,$78,$FE,$78,$FD,$5F,$E7,$2F,$6F,$FC,$D9,$FA,$9F
	.dc.b	$F8,$2B,$F5,$0F,$F0,$BF,$37,$F9,$A0,$56,$FC,$D5,$6F,$F1,$9F,$8D
	.dc.b	$5A,$FD,$C3,$C9,$97,$6C,$AC,$F2,$7B,$FD,$1B,$E7,$7C,$93,$9C,$BA
	.dc.b	$CF,$F7,$1B,$EB,$9F,$87,$37,$DC,$FF,$AE,$7F,$58,$CE,$7E,$A8,$FF
	.dc.b	$5E,$18,$C7,$CE,$FF,$BF,$3E,$9B,$F8,$5E,$23,$CD,$3A,$FC,$3B,$C7
	.dc.b	$97,$7F,$99,$79,$FF,$D2,$30,$E3,$87,$0D,$3F,$BA,$4B,$B7,$C1,$1F
	.dc.b	$DD,$FF,$05,$E7,$AF,$CE,$C7,$FD,$AB,$87,$E3,$7E,$07,$F9,$B7,$E6
	.dc.b	$AF,$DF,$7F,$C3,$3F,$A3,$BF,$19,$FE,$B9,$B7,$82,$F7,$5A,$9F,$D0
	.dc.b	$4E,$7E,$31,$87,$D2,$BF,$E8,$5D,$56,$73,$87,$6E,$8F,$C2,$57,$4F
	.dc.b	$82,$67,$FE,$DF,$E9,$FE,$81,$87,$09,$FE,$CD,$F6,$FF,$CF,$74,$F0
	.dc.b	$69,$E9,$9C,$FE,$DB,$ED,$BE,$1E,$87,$EB,$2F,$7A,$FF,$8D,$A3,$EA
	.dc.b	$65,$DB,$1E,$5F,$6B,$FE,$88,$FC,$AB,$94,$63,$E2,$5F,$95,$63,$1F
	.dc.b	$F1,$FD,$FF,$A3,$B9,$31,$B2,$3F,$A0,$47,$EE,$FF,$76,$FF,$11,$FE
	.dc.b	$8E,$FD,$DF,$FD,$DF,$E0,$3F,$3B,$7E,$97,$BF,$A8,$5E,$FF,$09,$FF
	.dc.b	$69,$53,$DD,$D7,$D0,$7E,$83,$3E,$BA,$7E,$E0,$3E,$0D,$3B,$FE,$6D
	.dc.b	$FB,$38,$C6,$3F,$ED,$DF,$AA,$C6,$9E,$0B,$4B,$F3,$A7,$EA,$7F,$ED
	.dc.b	$8E,$AF,$C9,$E3,$CA,$73,$F5,$7E,$9E,$0F,$44,$7F,$04,$AD,$E5,$D7
	.dc.b	$1E,$9F,$63,$E0,$BF,$77,$E4,$CF,$CB,$FF,$DC,$18,$70,$9C,$C7,$E1
	.dc.b	$A3,$1A,$F9,$AB,$87,$07,$86,$31,$F8,$E7,$F1,$1F,$6D,$EA,$3F,$66
	.dc.b	$FF,$03,$7E,$4E,$FE,$00,$7E,$2F,$FA,$3D,$3F,$5F,$87,$0B,$DD,$6B
	.dc.b	$F4,$1F,$E6,$D3,$D3,$F6,$4F,$E2,$2E,$F5,$FA,$4F,$1F,$A7,$67,$DF
	.dc.b	$3F,$26,$9E,$39,$CE,$5F,$A1,$C3,$F8,$36,$1C,$7A,$1F,$8B,$BF,$5B
	.dc.b	$71,$78,$B2,$E5,$5A,$9F,$ED,$5C,$1E,$A9,$CD,$FF,$11,$C7,$87,$07
	.dc.b	$F8,$9F,$0E,$BC,$BE,$48,$C7,$E6,$8F,$C7,$5F,$A8,$70,$E0,$FC,$1C
	.dc.b	$E7,$F6,$1C,$62,$87,$E7,$3C,$EB,$7E,$07,$F2,$1D,$DF,$35,$E3,$F8
	.dc.b	$47,$DA,$FF,$5B,$F3,$5F,$8B,$3F,$73,$FF,$19,$FD,$0B,$1E,$3C,$9F
	.dc.b	$CA,$F1,$E3,$F1,$BF,$B6,$C3,$87,$AB,$FE,$C2,$BD,$DA,$FB,$FE,$69
	.dc.b	$FE,$0E,$8E,$5A,$7F,$6B,$7F,$C1,$C6,$31,$E9,$FA,$47,$C9,$C3,$9E
	.dc.b	$1E,$0D,$7C,$FA,$FF,$BB,$E7,$38,$FA,$FF,$19,$7E,$DC,$C7,$F3,$D5
	.dc.b	$AC,$BF,$86,$D6,$A5,$F9,$E7,$BF,$87,$E3,$63,$CF,$1F,$F7,$1E,$3C
	.dc.b	$73,$63,$DB,$EA,$3E,$76,$1C,$33,$FE,$33,$F0,$1F,$D1,$FF,$1C,$FF
	.dc.b	$55,$E3,$C6,$31,$FA,$DF,$7E,$1F,$81,$9F,$F3,$4F,$8B,$FE,$B5,$97
	.dc.b	$38,$C7,$E0,$3F,$E4,$7A,$FF,$5F,$FC,$7B,$EF,$DF,$48,$FC,$8D,$FB
	.dc.b	$8F,$F8,$5F,$F4,$0F,$F2,$BE,$1C,$2F,$75,$AE,$1F,$DB,$70,$7F,$BB
	.dc.b	$7D,$76,$7D,$70,$FC,$D1,$E1,$8C,$FC,$07,$F9,$76,$3A,$69,$E1,$D1
	.dc.b	$F2,$3B,$C3,$CD,$39,$F7,$F1,$AF,$A3,$EE,$1F,$8D,$BF,$24,$7E,$CB
	.dc.b	$E8,$DB,$C7,$7B,$AD,$7A,$99,$CF,$AE,$3C,$B9,$FA,$E9,$FE,$8F,$BF
	.dc.b	$5E,$FB,$C7,$E3,$FE,$D0,$FD,$5B,$FC,$A9,$FA,$4E,$70,$4E,$7E,$9D
	.dc.b	$FB,$66,$63,$87,$E1,$1D,$9E,$5F,$2F,$FE,$0F,$FA,$BF,$A5,$A7,$F8
	.dc.b	$47,$F3,$B7,$EC,$8F,$8C,$7E,$3C,$F9,$97,$E8,$BF,$EC,$0F,$7D,$97
	.dc.b	$2C,$99,$72,$C7,$8E,$1C,$3F,$07,$7E,$7F,$C3,$C4,$FE,$21,$7B,$C3
	.dc.b	$EC,$99,$3C,$9A,$BF,$7D,$FF,$7A,$57,$08,$F3,$6D,$E5,$DB,$FC,$9F
	.dc.b	$AF,$8F,$1F,$14,$E7,$EA,$BE,$06,$1C,$76,$F6,$DF,$F2,$8F,$9E,$7E
	.dc.b	$21,$E1,$D7,$D7,$A7,$6A,$D7,$CF,$E7,$4E,$B1,$FC,$C3,$E3,$FA,$DF
	.dc.b	$A3,$2E,$BF,$14,$FC,$DF,$33,$D2,$7E,$6F,$CB,$BF,$3D,$7E,$D2,$FA
	.dc.b	$3E,$DB,$7B,$AB,$52,$FD,$97,$38,$72,$C3,$F9,$76,$38,$E4,$FE,$FF
	.dc.b	$1F,$E3,$ED,$BF,$8F,$FE,$5A,$FC,$8B,$EB,$FE,$90,$FC,$D3,$FD,$9F
	.dc.b	$D0,$FD,$BB,$EE,$FB,$5F,$D1,$70,$E1,$CC,$C3,$87,$E1,$EC,$3F,$DD
	.dc.b	$19,$F6,$CD,$FB,$1F,$DC,$DF,$99,$FE,$5F,$CB,$9B,$F8,$0A,$E1,$97
	.dc.b	$5D,$7C,$73,$C0,$57,$8B,$3F,$0F,$C0,$7E,$6F,$5F,$36,$7E,$CF,$39
	.dc.b	$F3,$8F,$A2,$7C,$17,$3C,$B9,$56,$A5,$FA,$A6,$70,$E5,$83,$D6,$ED
	.dc.b	$AF,$A6,$73,$F8,$9E,$23,$CB,$1F,$33,$37,$E8,$CF,$C6,$DF,$57,$D7
	.dc.b	$8F,$E5,$56,$A7,$FC,$5B,$C9,$F0,$B8,$FC,$F8,$7F,$9E,$F2,$E1,$18
	.dc.b	$F8,$4F,$DD,$7C,$87,$F7,$9F,$C8,$5F,$9D,$7F,$86,$3E,$F1,$F8,$43
	.dc.b	$F6,$7F,$EC,$DF,$92,$FE,$E3,$0E,$17,$BA,$D7,$C4,$B9,$35,$AF,$FA
	.dc.b	$57,$3E,$BF,$4D,$F6,$F8,$3F,$D2,$31,$E0,$FD,$07,$19,$F4,$A6,$DE
	.dc.b	$2D,$BF,$D5,$58,$70,$D7,$C1,$1F,$EB,$1C,$BB,$8C,$BB,$DF,$AB,$E5
	.dc.b	$DF,$74,$F7,$77,$5A,$E0,$FE,$F5,$3F,$E0,$3C,$FB,$69,$E2,$9F,$EF
	.dc.b	$DE,$27,$C8,$CF,$D9,$AF,$C1,$CF,$F8,$4E,$DE,$9D,$7F,$5D,$FB,$A3
	.dc.b	$EE,$7F,$8F,$E9,$D2,$73,$0F,$DC,$FC,$78,$FD,$1F,$35,$87,$F9,$A3
	.dc.b	$3C,$ED,$F1,$27,$3F,$27,$E3,$9F,$E7,$79,$BF,$49,$FE,$FF,$F8,$D7
	.dc.b	$F6,$8F,$CD,$9F,$8F,$FC,$6F,$85,$FE,$63,$5A,$F8,$85,$3F,$50,$D6
	.dc.b	$3C,$F2,$FD,$77,$1F,$F8,$FF,$1F,$E7,$3E,$9F,$D8,$9F,$F6,$94,$74
	.dc.b	$F9,$E6,$BF,$DE,$3D,$DF,$34,$E6,$3F,$BD,$FD,$A6,$32,$F4,$5F,$0F
	.dc.b	$D1,$DF,$A2,$7F,$A5,$67,$FD,$AF,$87,$19,$CF,$E3,$5E,$6E,$1F,$7C
	.dc.b	$7F,$17,$DB,$CD,$7D,$33,$4F,$F7,$8D,$78,$66,$FD,$99,$F9,$D3,$F0
	.dc.b	$47,$F0,$06,$A7,$30,$FD,$9B,$15,$FE,$BA,$C2,$FD,$7F,$D3,$FC,$BD
	.dc.b	$C9,$FE,$B5,$FD,$B3,$FB,$6F,$67,$FA,$17,$F5,$DF,$CB,$D6,$BE,$FD
	.dc.b	$5A,$FB,$D7,$EB,$DA,$C7,$FC,$95,$9F,$F8,$5B,$93,$FD,$F3,$8F,$18
	.dc.b	$FA,$A0,$3D,$70,$F0,$FC,$67,$96,$FC,$9F,$33,$FD,$0D,$C6,$7E,$28
	.dc.b	$7E,$5D,$FD,$F3,$F5,$5F,$14,$E6,$3F,$BF,$FE,$AD,$8C,$7F,$C1,$31
	.dc.b	$FA,$2C,$FC,$59,$F8,$71,$E7,$F7,$FD,$AF,$FC,$1F,$DE,$FD,$55,$FA
	.dc.b	$FF,$3B,$FC,$0C,$7F,$B1,$F8,$3D,$FB,$FA,$2F,$FD,$C7,$18,$DB,$D1
	.dc.b	$1F,$CB,$F4,$FE,$07,$F5,$BF,$E9,$DF,$C6,$3F,$D6,$7F,$7E,$FD,$47
	.dc.b	$FD,$D1,$F6,$DA,$FC,$15,$5A,$9F,$F7,$67,$BD,$98,$FC,$F6,$2C,$7F
	.dc.b	$D5,$1F,$8D,$E1,$ED,$E8,$B3,$ED,$F6,$5F,$B5,$9C,$C7,$87,$EC,$B1
	.dc.b	$A7,$82,$F8,$7E,$D7,$FD,$5B,$E8,$BC,$11,$8F,$92,$7E,$3D,$8C,$57
	.dc.b	$F9,$FD,$7F,$01,$9F,$68,$FC,$F7,$23,$CD,$FF,$5B,$4F,$8B,$F6,$EF
	.dc.b	$F5,$67,$B3,$CD,$FF,$57,$7D,$1D,$FC,$13,$CB,$C9,$F2,$27,$B6,$6F
	.dc.b	$CC,$9E,$07,$E8,$FE,$8D,$FB,$C3,$F2,$C7,$E6,$DF,$E1,$4E,$0F,$16
	.dc.b	$3F,$95,$C7,$8F,$A1,$FE,$17,$C3,$85,$FF,$AF,$EF,$77,$DF,$2F,$D8
	.dc.b	$3F,$9E,$31,$FB,$5F,$D4,$C7,$9E,$7E,$2C,$FF,$D8,$F5,$FD,$12,$73
	.dc.b	$97,$97,$F0,$8F,$E5,$11,$C3,$F7,$47,$E2,$4F,$CC,$BF,$CE,$BA,$F8
	.dc.b	$AB,$53,$9F,$C0,$3A,$6D,$F0,$70,$E3,$F0,$DD,$23,$15,$E2,$C3,$94
	.dc.b	$6F,$D4,$F9,$5F,$77,$CB,$8F,$FC,$71,$8A,$7E,$77,$E7,$19,$E2,$4E
	.dc.b	$73,$ED,$7C,$2F,$FA,$36,$5F,$23,$E7,$BF,$C8,$7F,$94,$BF,$6E,$7E
	.dc.b	$56,$E9,$7E,$9E,$F8,$6F,$ED,$38,$70,$BD,$D6,$BC,$9C,$FD,$83,$2F
	.dc.b	$D1,$DF,$F9,$43,$F3,$AB,$39,$C3,$8F,$D2,$FF,$55,$AE,$9A,$78,$7E
	.dc.b	$77,$17,$BF,$CD,$9F,$A3,$7F,$A9,$7E,$C7,$BE,$9E,$99,$CF,$E1,$1F
	.dc.b	$E8,$2E,$2F,$76,$1C,$70,$FF,$0F,$46,$2F,$84,$7F,$A9,$23,$1A,$BF
	.dc.b	$29,$FF,$33,$7E,$E1,$F3,$27,$31,$8F,$DE,$FC,$35,$BE,$AF,$E1,$FE
	.dc.b	$6F,$F2,$DF,$17,$FC,$C5,$A7,$BF,$0F,$F3,$87,$1F,$F7,$3F,$D6,$1F
	.dc.b	$90,$BF,$6F,$7E,$2E,$E8,$7E,$67,$FC,$DB,$1F,$F0,$CE,$1C,$2F,$75
	.dc.b	$AF,$86,$73,$6B,$27,$E9,$FF,$C5,$30,$E1,$B7,$E9,$5E,$5A,$77,$CF
	.dc.b	$D1,$C7,$FA,$C9,$CE,$7E,$2F,$9E,$F4,$65,$E7,$C3,$EF,$7D,$4B,$E2
	.dc.b	$FF,$D7,$98,$F1,$AD,$74,$FF,$91,$65,$F7,$DE,$5E,$1F,$7C,$FF,$92
	.dc.b	$7E,$33,$AC,$63,$1F,$F4,$7E,$3C,$E3,$E2,$FD,$03,$F6,$87,$E6,$0F
	.dc.b	$83,$E9,$B2,$B5,$3F,$E5,$5C,$39,$4E,$6F,$FD,$19,$C7,$C2,$FA,$7C
	.dc.b	$E7,$F9,$BF,$EC,$3F,$AE,$FE,$1C,$FD,$1B,$F9,$5F,$2F,$CC,$C7,$FD
	.dc.b	$11,$5A,$9C,$F4,$3F,$59,$CF,$17,$C0,$D3,$FD,$25,$F7,$BF,$85,$C3
	.dc.b	$AC,$60,$FE,$E8,$F3,$F8,$4F,$C1,$7D,$F7,$8E,$FF,$DE,$D9,$78,$BE
	.dc.b	$BB,$1B,$79,$63,$FC,$33,$E1,$7F,$40,$E3,$69,$DF,$0E,$35,$AF,$CB
	.dc.b	$3C,$B8,$71,$FA,$4F,$9F,$7B,$8C,$7C,$5F,$6B,$FC,$A7,$E4,$BF,$E7
	.dc.b	$FF,$0F,$8C,$63,$B3,$F8,$4E,$9E,$0E,$2F,$5E,$BE,$8E,$66,$3F,$8E
	.dc.b	$D3,$B6,$3D,$3D,$CF,$F0,$A6,$3F,$DA,$71,$FF,$0B,$7E,$92,$FD,$11
	.dc.b	$F4,$0F,$5B,$FC,$23,$FA,$1B,$F1,$07,$E1,$6C,$B9,$64,$CB,$96,$3C
	.dc.b	$78,$18,$F1,$FC,$63,$87,$C8,$FE,$E6,$FC,$1F,$B8,$FF,$A4,$E3,$1B
	.dc.b	$78,$B2,$FF,$66,$46,$32,$E9,$8F,$F9,$72,$79,$9E,$AA,$F4,$ED,$E8
	.dc.b	$EB,$F7,$E1,$C6,$3F,$1D,$F9,$1B,$F7,$F7,$E9,$7C,$B1,$F5,$63,$CB
	.dc.b	$91,$7B,$E7,$D6,$FE,$EC,$B9,$FD,$F7,$6C,$7C,$5B,$FC,$9C,$78,$FD
	.dc.b	$9E,$67,$AC,$FC,$DF,$C0,$5F,$9A,$FF,$B6,$3F,$3C,$EB,$B7,$AA,$73
	.dc.b	$F8,$57,$F9,$53,$A3,$E7,$ED,$E1,$9E,$3F,$86,$75,$8C,$62,$9F,$F2
	.dc.b	$97,$D1,$BF,$1D,$FE,$51,$F2,$DF,$E9,$FF,$BE,$7E,$FE,$FC,$4D,$FA
	.dc.b	$83,$F1,$47,$F0,$03,$2E,$5C,$9F,$E5,$DC,$B9,$63,$C7,$0E,$1E,$DE
	.dc.b	$F4,$EF,$A7,$E1,$F0,$FE,$70,$79,$23,$31,$88,$C1,$D5,$FF,$5A,$5F
	.dc.b	$5F,$C7,$B3,$2C,$7C,$3A,$FA,$3A,$73,$1E,$4F,$B9,$FE,$33,$F9,$D3
	.dc.b	$F2,$3F,$26,$1D,$73,$F8,$35,$AF,$C9,$33,$3D,$67,$C3,$3F,$C8,$B3
	.dc.b	$EB,$18,$F7,$BF,$DB,$55,$C3,$2F,$87,$F8,$A7,$FA,$03,$99,$F9,$B6
	.dc.b	$70,$F5,$CE,$7E,$83,$11,$F4,$A3,$FD,$0F,$97,$68,$C7,$C5,$7F,$BC
	.dc.b	$3E,$37,$FB,$07,$D6,$FF,$10,$FD,$13,$FC,$A3,$F1,$CF,$C3,$7E,$BD
	.dc.b	$FF,$BC,$AB,$5F,$01,$FD,$7A,$7F,$C1,$1C,$19,$F5,$DF,$9F,$7F,$9B
	.dc.b	$2F,$A4,$FF,$8F,$5E,$38,$F5,$F0,$7F,$DB,$D1,$8C,$3F,$9C,$57,$E1
	.dc.b	$70,$FE,$E9,$FB,$0B,$F2,$13,$FE,$E3,$FD,$47,$FD,$5B,$F4,$8F,$CB
	.dc.b	$F8,$61,$FE,$A7,$AE,$06,$FF,$01,$53,$F4,$ED,$D7,$4C,$7A,$BF,$89
	.dc.b	$F1,$BF,$92,$39,$EB,$FC,$63,$C3,$1E,$BF,$86,$34,$F1,$CE,$70,$F3
	.dc.b	$CE,$73,$FF,$1D,$F3,$7F,$83,$FF,$01,$7E,$3E,$FC,$87,$E8,$7E,$D4
	.dc.b	$F8,$B7,$C6,$BF,$14,$FE,$4D,$8F,$1C,$38,$73,$30,$E1,$7F,$B7,$F4
	.dc.b	$FF,$CD,$F9,$F6,$FD,$27,$FD,$3E,$F8,$34,$B8,$F2,$C3,$8F,$07,$63
	.dc.b	$FD,$A1,$5B,$CB,$B6,$AF,$BE,$FF,$51,$6D,$FE,$49,$F9,$2F,$F4,$37
	.dc.b	$E7,$FF,$EE,$6F,$B4,$FF,$5B,$CE,$7E,$E1,$3F,$EE,$AF,$37,$3E,$A7
	.dc.b	$FA,$BE,$BF,$C5,$5A,$FF,$A0,$FE,$17,$E0,$4F,$F7,$5C,$38,$FE,$0F
	.dc.b	$E0,$8C,$62,$FC,$2F,$F8,$AF,$85,$FE,$3B,$87,$EC,$2F,$49,$F9,$03
	.dc.b	$F0,$BF,$F0,$9F,$EC,$1E,$4F,$13,$EF,$E1,$C2,$F7,$5A,$F8,$7F,$C3
	.dc.b	$CE,$78,$3F,$0D,$9F,$7C,$DF,$B3,$FE,$1B,$2E,$71,$E1,$D3,$C1,$EB
	.dc.b	$BF,$41,$F3,$E3,$FE,$FD,$CB,$C1,$5A,$FA,$8F,$ED,$BE,$1F,$F9,$57
	.dc.b	$E6,$5E,$A7,$F7,$4F,$D2,$5E,$3B,$C5,$FF,$15,$7D,$3F,$D9,$D3,$FC
	.dc.b	$9B,$7B,$AD,$7D,$23,$E7,$B5,$97,$4F,$D0,$BF,$DC,$6F,$6F,$9D,$B8
	.dc.b	$71,$CF,$FA,$16,$5C,$23,$1F,$23,$E6,$80,$FD,$F7,$CB,$F6,$0F,$C3
	.dc.b	$1F,$A1,$3F,$4F,$7E,$F3,$FB,$DF,$4B,$FD,$8B,$87,$0B,$DD,$6B,$F6
	.dc.b	$4F,$64,$FC,$0F,$DF,$7F,$DF,$2F,$6C,$FF,$D9,$4F,$1C,$BB,$7E,$73
	.dc.b	$E5,$9F,$27,$2F,$FA,$A7,$2E,$BE,$7E,$70,$F0,$7E,$AC,$FB,$A7,$9A
	.dc.b	$FD,$DB,$C3,$8F,$1A,$D4,$BE,$97,$3C,$7E,$F9,$7A,$DF,$E7,$6E,$FF
	.dc.b	$F4,$76,$7D,$B3,$78,$1E,$1C,$FF,$D7,$9F,$8C,$BF,$79,$7E,$50,$FD
	.dc.b	$F3,$EF,$74,$FC,$6B,$DD,$6A,$7F,$85,$4E,$76,$61,$CF,$07,$DE,$F7
	.dc.b	$9C,$E1,$DB,$E5,$FE,$37,$E2,$4F,$C0,$9F,$B4,$72,$F0,$7F,$55,$F8
	.dc.b	$2F,$F3,$2D,$7F,$DF,$77,$BA,$D7,$AA,$AF,$ED,$95,$AC,$1F,$B4,$7F
	.dc.b	$2A,$8C,$5D,$E5,$FF,$59,$63,$CF,$83,$F3,$F9,$7F,$AD,$BB,$7F,$1B
	.dc.b	$39,$D3,$F7,$9C,$1E,$8D,$3C,$F8,$3E,$A1,$F8,$6F,$E5,$3F,$29,$FD
	.dc.b	$26,$3D,$67,$39,$7F,$BD,$76,$F1,$E1,$C7,$0F,$B1,$3E,$0C,$7A,$DF
	.dc.b	$4F,$B5,$F0,$6F,$F3,$BF,$2F,$7E,$F3,$FE,$1D,$FD,$7B,$BE,$1E,$49
	.dc.b	$CF,$B3,$F3,$35,$EF,$1C,$3E,$8F,$9E,$C8,$E7,$F3,$BF,$99,$FC,$B7
	.dc.b	$FC,$CF,$F9,$73,$EB,$DF,$96,$7D,$CF,$D2,$C7,$F0,$1A,$DC,$EB,$9F
	.dc.b	$3F,$A0,$67,$EF,$3F,$B1,$73,$7F,$89,$F7,$76,$7F,$B8,$F1,$EF,$39
	.dc.b	$8C,$53,$4F,$16,$9E,$1A,$E9,$18,$C9,$FA,$97,$A6,$F8,$69,$FE,$BF
	.dc.b	$FD,$33,$F8,$3F,$BD,$C2,$31,$8F,$F1,$5D,$7C,$6F,$F0,$CD,$7D,$7E
	.dc.b	$F3,$0F,$AB,$29,$CC,$3A,$BC,$1A,$FC,$7F,$7D,$F8,$1B,$E3,$7F,$27
	.dc.b	$9F,$6C,$B9,$6D,$E6,$BD,$D7,$EC,$AB,$59,$FC,$C9,$CD,$FA,$5E,$BF
	.dc.b	$19,$FC,$3F,$E9,$BF,$84,$FD,$4B,$E4,$9F,$86,$3F,$19,$7E,$7E,$FD
	.dc.b	$27,$F9,$C6,$1C,$2F,$75,$AF,$D9,$F3,$ED,$FF,$90,$B5,$FA,$87,$F3
	.dc.b	$0B,$EE,$E0,$3C,$F2,$F0,$74,$F9,$69,$A7,$96,$7C,$9D,$D9,$8F,$47
	.dc.b	$E5,$AF,$D0,$DF,$8B,$7F,$9F,$79,$3E,$46,$7E,$7A,$D4,$EE,$73,$97
	.dc.b	$6D,$D5,$CF,$F2,$6E,$D1,$8C,$5F,$84,$63,$F2,$77,$E5,$7F,$E2,$5F
	.dc.b	$E0,$1F,$83,$B7,$CB,$9C,$FD,$33,$ED,$73,$92,$78,$FC,$2F,$3C,$5F
	.dc.b	$53,$FD,$0F,$E0,$BF,$C2,$7E,$43,$F1,$D7,$D2,$FF,$16,$FA,$67,$EC
	.dc.b	$9F,$DA,$EF,$75,$AF,$7F,$FC,$6D,$F6,$D9,$F9,$CF,$E6,$DD,$FF,$46
	.dc.b	$FA,$B6,$3F,$98,$DF,$1E,$5F,$14,$63,$3F,$E4,$BF,$2D,$8D,$7C,$9E
	.dc.b	$5B,$E9,$7F,$AA,$7D,$37,$79,$CC,$7F,$57,$FC,$FB,$19,$FE,$D6,$3F
	.dc.b	$94,$69,$FD,$3B,$3F,$0E,$3C,$F0,$ED,$6F,$C5,$9E,$5B,$F6,$57,$E6
	.dc.b	$7D,$27,$31,$F8,$2F,$0F,$F5,$1C,$5F,$12,$78,$FB,$BD,$1E,$7E,$DF
	.dc.b	$F7,$1F,$A6,$7E,$A6,$FC,$89,$FB,$0B,$DA,$73,$BF,$12,$7E,$C1,$FE
	.dc.b	$1B,$C7,$8E,$1C,$2F,$74,$F6,$75,$93,$F4,$DF,$E9,$9D,$1F,$E1,$4C
	.dc.b	$78,$BF,$A5,$79,$E9,$E1,$D3,$C1,$3C,$B3,$ED,$1F,$E7,$28,$C4,$BF
	.dc.b	$77,$FE,$4F,$F4,$BF,$57,$8F,$F8,$56,$3D,$27,$3F,$85,$F0,$7F,$96
	.dc.b	$F8,$FF,$0F,$5F,$85,$8F,$3F,$84,$FC,$AE,$7E,$1F,$DA,$5F,$BA,$7E
	.dc.b	$07,$D8,$7A,$79,$8C,$7C,$0D,$F0,$FF,$A0,$E1,$7D,$AF,$F1,$07,$A8
	.dc.b	$FD,$F9,$F5,$8E,$4F,$DD,$F4,$67,$F6,$BF,$F2,$45,$EE,$B5,$ED,$FC
	.dc.b	$BC,$1F,$29,$A5,$F9,$FF,$CE,$D1,$87,$17,$F8,$93,$CF,$3E,$FF,$0F
	.dc.b	$E4,$BF,$1C,$78,$AB,$19,$7E,$EF,$E6,$1F,$A7,$BF,$52,$7D,$67,$86
	.dc.b	$B5,$39,$FC,$03,$1E,$F7,$F9,$FE,$3F,$DB,$D8,$FF,$95,$35,$F6,$4E
	.dc.b	$7C,$8E,$EF,$FA,$1B,$F0,$17,$E9,$6F,$C2,$39,$FD,$CF,$FA,$F6,$3F
	.dc.b	$C6,$FC,$1F,$D1,$F1,$E7,$3F,$EC,$1F,$C5,$B9,$BF,$EC,$3F,$C2,$BF
	.dc.b	$CA,$1C,$E9,$FE,$3B,$FC,$02,$FC,$F9,$F9,$FB,$F7,$0F,$E5,$9E,$2F
	.dc.b	$F9,$AB,$8B,$F9,$7E,$1C,$2F,$74,$F9,$9F,$F3,$F6,$5D,$7E,$07,$E9
	.dc.b	$3D,$FC,$4B,$87,$1C,$FF,$13,$97,$08,$C5,$F9,$9F,$2F,$E3,$FF,$EB
	.dc.b	$6C,$9F,$B6,$6E,$9B,$2B,$5F,$B6,$66,$31,$F0,$3D,$B9,$3F,$B6,$6D
	.dc.b	$E5,$9C,$E1,$DC,$7E,$93,$FE,$F3,$FC,$CD,$F9,$FB,$F6,$AF,$04,$E6
	.dc.b	$3F,$C4,$98,$F4,$DF,$FA,$3C,$F1,$F3,$7A,$3C,$F1,$FF,$49,$D7,$FB
	.dc.b	$EB,$F0,$CF,$C5,$7E,$C6,$FD,$FD,$F8,$6B,$F7,$F7,$EC,$1F,$BD,$C3
	.dc.b	$87,$92,$C3,$85,$EE,$9F,$9B,$7E,$AF,$F0,$EF,$B1,$9F,$6E,$AF,$F1
	.dc.b	$64,$63,$17,$C7,$7F,$98,$63,$19,$74,$D1,$F8,$CF,$A2,$73,$B7,$9A
	.dc.b	$7F,$DB,$58,$F7,$3F,$D5,$F9,$7B,$3E,$21,$FB,$0F,$FA,$27,$F3,$5F
	.dc.b	$ED,$3C,$3E,$AD,$15,$AF,$BB,$53,$F4,$7E,$99,$32,$E9,$97,$FA,$4F
	.dc.b	$F1,$9F,$24,$F8,$38,$7E,$26,$5C,$FE,$2B,$33,$C6,$7E,$47,$E9,$4F
	.dc.b	$C2,$1F,$AA,$FF,$BD,$DE,$F9,$FD,$49,$D4,$FF,$BE,$FF,$67,$CD,$F2
	.dc.b	$BE,$3F,$5D,$DB,$0E,$11,$D3,$E8,$3F,$DD,$79,$7F,$96,$BF,$58,$7E
	.dc.b	$D7,$FD,$77,$E5,$FD,$FF,$8C,$FE,$BF,$EE,$9D,$7F,$AF,$F4,$D8,$70
	.dc.b	$F9,$8F,$EC,$5D,$0C,$3F,$98,$E1,$C3,$3F,$F6,$FE,$1C,$33,$FF,$80
	.dc.b	$19,$73,$CB,$FA,$E5,$7E,$EC,$F2,$D6,$F2,$EB,$8E,$2B,$C1,$9F,$D4
	.dc.b	$FA,$EC,$64,$FD,$95,$FB,$17,$F2,$EF,$D8,$7F,$D9,$3B,$78,$F5,$F8
	.dc.b	$17,$BA,$D7,$DB,$2B,$5B,$79,$F1,$E7,$EE,$7F,$BD,$71,$D7,$07,$A7
	.dc.b	$2E,$53,$F0,$67,$27,$C0,$DB,$E6,$FE,$BF,$FE,$DE,$FD,$41,$E8,$BF
	.dc.b	$74,$DD,$F9,$2B,$53,$FE,$32,$D7,$C7,$87,$2F,$C4,$58,$F8,$E7,$39
	.dc.b	$B0,$ED,$F9,$FF,$DE,$F3,$DF,$9B,$3E,$71,$FA,$DB,$F3,$5F,$F1,$CF
	.dc.b	$DC,$3F,$29,$F1,$E1,$C2,$F7,$5A,$FD,$DF,$38,$F3,$E2,$FB,$06,$3E
	.dc.b	$6B,$EE,$A3,$FC,$AF,$1E,$1C,$FC,$1F,$67,$FD,$CB,$83,$EA,$67,$DE
	.dc.b	$73,$F3,$7E,$9B,$F4,$7C,$67,$D5,$F7,$1F,$B5,$BF,$7D,$F1,$F0,$7F
	.dc.b	$54,$CF,$FC,$03,$3F,$EC,$AE,$CF,$E8,$DA,$7A,$F1,$E5,$E4,$75,$C3
	.dc.b	$86,$5F,$E1,$9C,$7C,$79,$FC,$7F,$7F,$FB,$6F,$A8,$FD,$2B,$FC,$FD
	.dc.b	$5A,$C3,$FA,$C7,$CA,$7F,$05,$97,$E2,$9E,$4C,$38,$F5,$77,$79,$FE
	.dc.b	$3B,$F1,$71,$FE,$D5,$CD,$F1,$FF,$0C,$7E,$A8,$FC,$A3,$FB,$27,$EE
	.dc.b	$FF,$B2,$FB,$CF,$F5,$C6,$1C,$30,$61,$C2,$F7,$CC,$BF,$F0,$E5,$EE
	.dc.b	$B5,$7F,$62,$31,$8F,$F3,$E8,$C7,$CC,$7F,$6D,$AE,$9A,$33,$FE,$07
	.dc.b	$9F,$A6,$B5,$87,$9E,$B5,$9F,$F3,$6C,$7F,$59,$5A,$BF,$F6,$67,$EB
	.dc.b	$0F,$C9,$67,$DC,$EA,$FE,$16,$5F,$07,$F0,$C5,$EE,$DF,$94,$6E,$FE
	.dc.b	$0D,$EE,$BA,$71,$79,$67,$9E,$5E,$0C,$3F,$BE,$F0,$7C,$08,$7E,$48
	.dc.b	$FC,$09,$F9,$63,$F7,$9F,$EA,$77,$BA,$D4,$E7,$E2,$7C,$5C,$BF,$46
	.dc.b	$73,$F0,$DE,$1D,$3C,$B9,$F6,$CF,$AF,$E3,$5F,$DC,$3F,$0C,$FF,$0D
	.dc.b	$70,$BE,$93,$FD,$5D,$F9,$C3,$F6,$6F,$DE,$7E,$91,$FD,$B7,$E4,$18
	.dc.b	$70,$BD,$D6,$BE,$53,$FC,$B1,$39,$F0,$3F,$6B,$B6,$0F,$DA,$67,$DF
	.dc.b	$AF,$39,$74,$9C,$E5,$D3,$D5,$FF,$50,$47,$13,$F2,$B1,$9F,$8E,$F5
	.dc.b	$47,$0B,$FF,$71,$7E,$C8,$FE,$01,$7E,$D5,$FE,$81,$D7,$D1,$5A,$9C
	.dc.b	$F9,$3D,$75,$F2,$7E,$24,$C7,$97,$CC,$76,$CF,$C3,$8F,$1F,$B1,$C6
	.dc.b	$1E,$BF,$DE,$BF,$DB,$DF,$7E,$FD,$55,$F9,$5F,$5E,$2F,$F6,$D5,$6A
	.dc.b	$7F,$B7,$F2,$7C,$1C,$FA,$CF,$2F,$67,$AD,$F0,$FB,$1F,$F4,$E7,$C7
	.dc.b	$7F,$7A,$FC,$8F,$FD,$23,$E8,$7E,$1F,$C4,$EF,$E4,$BF,$33,$FF,$03
	.dc.b	$D6,$BC,$AF,$DB,$F0,$7E,$87,$C2,$CF,$D7,$FE,$B7,$83,$FC,$46,$7F
	.dc.b	$29,$BF,$F4,$EF,$ED,$BF,$E2,$38,$C4,$FE,$CF,$0E,$7E,$6F,$8E,$31
	.dc.b	$A7,$C0,$F6,$B1,$B3,$F7,$1F,$F6,$17,$E7,$8F,$94,$7E,$99,$E2,$AD
	.dc.b	$7E,$18,$AD,$7C,$B6,$70,$E3,$E4,$F4,$79,$FC,$C7,$6F,$DF,$FF,$31
	.dc.b	$F4,$BF,$47,$F0,$8E,$91,$8F,$01,$F9,$E6,$3D,$05,$BC,$0E,$51,$8E
	.dc.b	$EF,$F2,$AF,$CD,$7F,$9F,$FF,$08,$7C,$B3,$F6,$3F,$F1,$A7,$CF,$7F
	.dc.b	$5E,$FD,$17,$FD,$0B,$DD,$FD,$46,$F7,$5A,$F8,$B5,$3D,$0F,$F5,$F6
	.dc.b	$5D,$72,$E9,$8F,$F5,$CF,$29,$39,$FC,$7B,$FD,$75,$9F,$4F,$8C,$6B
	.dc.b	$FD,$BB,$F0,$0F,$3E,$3E,$1C,$BC,$D5,$AD,$7F,$D5,$F2,$C3,$CB,$F8
	.dc.b	$E3,$F3,$5F,$F3,$6F,$BD,$FE,$FC,$DB,$C3,$7C,$F8,$57,$BF,$B2,$5B
	.dc.b	$F1,$1F,$5D,$EE,$B5,$3E,$48,$F0,$6B,$F2,$67,$97,$C5,$67,$4F,$8F
	.dc.b	$FB,$4B,$F6,$97,$E5,$4F,$D4,$FF,$D8,$1B,$F9,$33,$E7,$BB,$0E,$17
	.dc.b	$BF,$C1,$7C,$75,$A9,$CF,$D7,$F7,$DB,$ED,$47,$0D,$7C,$DF,$38,$FD
	.dc.b	$27,$FD,$13,$F9,$53,$F1,$27,$E2,$CF,$CA,$1F,$A5,$F6,$C3,$87,$DF
	.dc.b	$30,$E1,$7B,$FB,$0D,$64,$FD,$53,$FD,$E9,$87,$0E,$0F,$C0,$78,$5F
	.dc.b	$EC,$D8,$E7,$8F,$2F,$86,$F2,$E1,$E4,$DB,$C7,$F1,$18,$D5,$FB,$23
	.dc.b	$E4,$BF,$B4,$7D,$4B,$F1,$DF,$0D,$6A,$73,$86,$2F,$87,$B2,$FC,$77
	.dc.b	$A4,$63,$F1,$0E,$DE,$9F,$F5,$8F,$D9,$7E,$87,$EF,$DD,$23,$1D,$7F
	.dc.b	$ED,$4F,$07,$EB,$72,$7C,$49,$E1,$F2,$DC,$BE,$71,$8F,$F3,$EB,$FB
	.dc.b	$9F,$83,$7F,$9F,$BF,$69,$7E,$B1,$FE,$C3,$FA,$67,$E9,$0F,$CD,$3E
	.dc.b	$ED,$EE,$B5,$2F,$D0,$73,$F7,$6C,$7F,$A7,$F7,$FF,$BE,$70,$FF,$45
	.dc.b	$FE,$F9,$8F,$07,$FC,$91,$1C,$FF,$0C,$FF,$48,$F0,$FE,$66,$31,$B7
	.dc.b	$9F,$0E,$59,$75,$D7,$FC,$41,$F3,$5F,$86,$FD,$DB,$C3,$AF,$AB,$EB
	.dc.b	$D3,$F8,$AF,$AA,$CD,$F0,$F2,$7B,$63,$CE,$F7,$F5,$F8,$C5,$F8,$1B
	.dc.b	$E5,$5F,$BD,$FE,$E7,$F5,$CE,$F3,$98,$FD,$E7,$EF,$98,$AD,$FC,$4F
	.dc.b	$3A,$D7,$E8,$7F,$B5,$E2,$7E,$E4,$FC,$FD,$FA,$93,$F7,$7F,$0A,$FD
	.dc.b	$57,$F5,$B7,$C2,$B7,$5A,$FC,$D3,$38,$73,$ED,$FC,$E9,$FE,$B4,$5C
	.dc.b	$5F,$46,$3F,$8F,$C7,$87,$2F,$06,$8F,$6B,$CD,$39,$C7,$E1,$FC,$26
	.dc.b	$33,$F3,$F9,$CF,$DF,$5F,$BC,$3E,$57,$C7,$5A,$9C,$DE,$34,$FF,$27
	.dc.b	$47,$F1,$CD,$71,$8F,$3D,$3B,$53,$F4,$37,$E9,$BD,$7F,$59,$E5,$18
	.dc.b	$D1,$FA,$6E,$2B,$A7,$99,$FC,$67,$97,$E4,$63,$CE,$7F,$E0,$0E,$9F
	.dc.b	$EC,$39,$FF,$4D,$3F,$D1,$FF,$3B,$7E,$62,$F5,$5F,$AD,$7F,$59,$FC
	.dc.b	$1D,$9F,$1C,$3F,$D4,$5F,$30,$C7,$8E,$1C,$3B,$BF,$E0,$07,$D8,$EE
	.dc.b	$FB,$67,$FE,$38,$FC,$07,$F9,$76,$3C,$7F,$7E,$C6,$DE,$48,$E9,$5E
	.dc.b	$3D,$BC,$58,$73,$9C,$E7,$DF,$D7,$FF,$48,$61,$EC,$FC,$01,$FA,$BF
	.dc.b	$F0,$2F,$CD,$7F,$D9,$5F,$68,$FD,$07,$FD,$01,$BF,$F3,$FB,$DF,$B8
	.dc.b	$BD,$D3,$EC,$FD,$55,$D3,$C7,$F3,$EF,$FE,$C3,$8E,$2F,$FA,$47,$4F
	.dc.b	$16,$9E,$1F,$C0,$7F,$C6,$FE,$97,$9A,$B8,$4E,$76,$FC,$BC,$7F,$4C
	.dc.b	$F8,$B7,$E1,$1F,$E0,$3F,$63,$FC,$F3,$F1,$AE,$97,$D0,$6E,$B5,$EC
	.dc.b	$AB,$5F,$39,$9F,$99,$FD,$16,$5D,$3E,$8E,$E1,$C2,$DE,$A7,$F6,$BA
	.dc.b	$F0,$DB,$4F,$0F,$81,$E9,$DB,$FC,$EF,$ED,$FF,$A6,$3F,$80,$5F,$A6
	.dc.b	$BF,$2D,$7E,$AD,$FE,$24,$EF,$FE,$D9,$F8,$2B,$0E,$17,$BA,$7D,$77
	.dc.b	$E1,$EB,$5F,$86,$7C,$EC,$FB,$FE,$07,$67,$3A,$78,$34,$FF,$4F,$D7
	.dc.b	$1C,$FA,$C3,$CC,$FF,$6A,$7C,$37,$F8,$BF,$E8,$5F,$98,$3C,$D7,$EB
	.dc.b	$FF,$E2,$0F,$C4,$1F,$87,$39,$18,$F1,$C3,$85,$EF,$F3,$5D,$4E,$76
	.dc.b	$FF,$69,$3D,$F5,$FD,$03,$C7,$3E,$FF,$61,$FC,$BE,$B8,$9F,$E3,$DD
	.dc.b	$7C,$B3,$AF,$B8,$FF,$0F,$7D,$FB,$F7,$5F,$F6,$4F,$CE,$3F,$6B,$7D
	.dc.b	$8B,$F2,$F7,$E5,$7F,$E1,$6C,$38,$7E,$29,$FD,$EA,$F7,$4F,$CA,$75
	.dc.b	$93,$F5,$FF,$E5,$18,$70,$DF,$F9,$69,$FE,$F4,$9F,$0E,$1E,$0F,$C1
	.dc.b	$BC,$B1,$8D,$1F,$8D,$7F,$8D,$78,$7F,$A0,$53,$0F,$E6,$1F,$42,$FB
	.dc.b	$A7,$E7,$9F,$E2,$CB,$DD,$6A,$5F,$9E,$E7,$E4,$18,$FF,$3D,$BF,$E9
	.dc.b	$9F,$9F,$D9,$CE,$1D,$BE,$5B,$C5,$18,$F8,$FF,$EF,$DF,$09,$E3,$F8
	.dc.b	$9F,$BB,$7E,$5F,$CE,$FF,$CF,$18,$BE,$FF,$FC,$43,$F8,$93,$F6,$EF
	.dc.b	$9D,$87,$0C,$18,$70,$BD,$D3,$F4,$FF,$E5,$3E,$17,$F3,$0A,$EB,$7F
	.dc.b	$EA,$43,$8F,$A7,$F9,$27,$11,$CF,$C0,$FF,$75,$7C,$77,$9E,$F9,$5B
	.dc.b	$5F,$35,$7F,$BB,$FD,$8F,$DB,$7D,$C3,$F8,$01,$E7,$7F,$98,$3F,$3F
	.dc.b	$7E,$B4,$F3,$3F,$AE,$79,$FF,$E6,$1C,$38,$5E,$EB,$5F,$AF,$FF,$3B
	.dc.b	$9C,$F0,$FF,$1F,$3E,$E3,$DF,$E3,$FF,$70,$F2,$FC,$05,$9F,$83,$E4
	.dc.b	$7C,$BA,$F9,$27,$39,$7C,$FC,$7F,$A4,$7C,$67,$FB,$7F,$F1,$4F,$C0
	.dc.b	$F8,$5F,$C3,$3E,$07,$FC,$E9,$C1,$FD,$7F,$EC,$95,$AF,$80,$FF,$55
	.dc.b	$7D,$97,$29,$CE,$5E,$73,$FD,$42,$BF,$67,$FE,$1D,$9F,$7C,$FF,$1F
	.dc.b	$83,$F3,$4F,$E4,$13,$99,$FE,$3B,$F2,$1F,$EF,$9F,$AB,$FF,$66,$75
	.dc.b	$7F,$92,$B6,$7E,$D0,$FD,$D9,$E5,$FF,$6A,$F8,$4F,$EE,$38,$70,$E1
	.dc.b	$FE,$1D,$7B,$FD,$97,$5C,$5F,$A7,$AE,$AF,$FA,$E6,$F8,$FE,$2A,$E3
	.dc.b	$FE,$F7,$3F,$DF,$3C,$9F,$F8,$01,$7D,$75,$F1,$FC,$BF,$9F,$0F,$F3
	.dc.b	$DE,$9E,$0F,$D3,$BF,$92,$E3,$FB,$5F,$63,$D3,$D1,$FE,$12,$FC,$5F
	.dc.b	$FC,$57,$F3,$6F,$D5,$FF,$8B,$DF,$F1,$1E,$2F,$A3,$F4,$4B,$DF,$EE
	.dc.b	$9A,$CB,$E1,$F6,$FF,$22,$C3,$B6,$1F,$51,$F4,$F0,$7E,$AE,$BC,$38
	.dc.b	$F3,$F5,$BF,$D1,$15,$CE,$FB,$E5,$FE,$3E,$F6,$FF,$AA,$72,$BE,$A1
	.dc.b	$F9,$7F,$F9,$6F,$F5,$07,$DD,$7F,$B4,$FF,$79,$FF,$0A,$5E,$EB,$53
	.dc.b	$FE,$D7,$FB,$CC,$E0,$FA,$3F,$EA,$57,$BD,$7F,$14,$F1,$FB,$66,$7D
	.dc.b	$BD,$5F,$1E,$1E,$29,$CE,$5F,$DB,$B0,$FF,$12,$7E,$B2,$FC,$6B,$FC
	.dc.b	$2F,$FA,$23,$F3,$F7,$EA,$FF,$EE,$9F,$67,$75,$A9,$CF,$E3,$2F,$D0
	.dc.b	$B3,$F5,$7F,$D9,$F2,$E9,$F0,$B7,$0E,$19,$38,$3F,$BD,$57,$86,$DA
	.dc.b	$78,$6B,$CD,$E4,$3C,$AF,$F2,$3F,$C7,$FF,$B5,$69,$FA,$93,$F1,$CE
	.dc.b	$3F,$9B,$3F,$4E,$FF,$23,$E1,$C3,$F1,$8E,$B7,$BC,$16,$FD,$93,$7C
	.dc.b	$DF,$D1,$73,$FF,$71,$7E,$95,$F9,$C8,$FE,$6D,$97,$3C,$31,$1C,$0F
	.dc.b	$EB,$19,$76,$D3,$D7,$F4,$5C,$27,$A4,$E6,$9F,$A3,$FF,$53,$F2,$5E
	.dc.b	$A7,$BB,$E1,$3F,$D9,$3F,$57,$FE,$CA,$F8,$1F,$F3,$7F,$D0,$7F,$80
	.dc.b	$EB,$55,$FE,$FB,$13,$FE,$74,$FD,$A3,$5A,$7A,$E5,$F6,$BF,$E6,$1B
	.dc.b	$E3,$3E,$DC,$BC,$2E,$EB,$A6,$0F,$67,$CD,$7E,$0C,$FC,$9F,$8C,$FE
	.dc.b	$E3,$07,$D0,$3F,$19,$7E,$F1,$FE,$03,$FC,$63,$FB,$77,$E5,$CF,$CC
	.dc.b	$FF,$73,$7B,$AD,$7C,$F2,$9F,$03,$AC,$7F,$B0,$65,$FE,$1C,$3F,$BD
	.dc.b	$61,$C5,$C1,$E2,$8F,$0F,$D2,$B3,$FE,$FD,$F9,$D7,$9E,$FC,$13,$9C
	.dc.b	$BF,$D1,$58,$7F,$0C,$F8,$A7,$E9,$3F,$EA,$3F,$AA,$7E,$E4,$FD,$A5
	.dc.b	$FA,$1B,$EF,$38,$5E,$EB,$53,$FC,$CF,$E7,$7F,$D3,$F8,$F4,$FC,$27
	.dc.b	$F8,$96,$68,$EF,$F9,$5E,$1D,$69,$E0,$D3,$E1,$D7,$6B,$FF,$24,$65
	.dc.b	$E3,$E0,$FF,$32,$7D,$CB,$EB,$BF,$C5,$5F,$9E,$BF,$30,$78,$FF,$E4
	.dc.b	$AF,$D9,$BF,$D4,$97,$BF,$35,$7B,$AF,$F2,$7F,$D2,$EB,$1C,$EB,$FD
	.dc.b	$67,$0E,$0D,$8E,$F9,$7F,$35,$8F,$1C,$62,$DF,$25,$F8,$7F,$B4,$7E
	.dc.b	$94,$FD,$59,$F1,$0F,$C8,$1F,$82,$BF,$6E,$7E,$87,$FD,$03,$1E,$38
	.dc.b	$70,$BD,$D3,$F8,$07,$59,$75,$DB,$F0,$B8,$78,$3F,$0D,$FA,$2F,$2D
	.dc.b	$3C,$1A,$3E,$87,$CD,$87,$96,$73,$F2,$BE,$9F,$D2,$3F,$BC,$7E,$03
	.dc.b	$FE,$8E,$F9,$77,$E4,$4F,$D4,$3C,$D5,$AF,$A8,$FE,$E7,$3F,$03,$D8
	.dc.b	$FF,$1C,$6C,$AF,$F0,$17,$07,$99,$FB,$0D,$F9,$DF,$F4,$1E,$1F,$DF
	.dc.b	$23,$C5,$18,$CB,$FC,$AB,$1F,$ED,$FF,$85,$7E,$84,$FD,$63,$FC,$E9
	.dc.b	$FA,$93,$F3,$67,$E4,$DF,$C8,$76,$FC,$96,$3F,$E6,$5C,$78,$E1,$C2
	.dc.b	$F7,$F1,$7D,$72,$FF,$15,$DF,$F8,$6E,$F6,$D8,$F0,$7A,$6C,$FC,$1E
	.dc.b	$F7,$FA,$E6,$FA,$CE,$71,$7B,$99,$FA,$57,$C8,$3E,$09,$FC,$02,$FD
	.dc.b	$CB,$FD,$11,$EB,$FB,$EF,$75,$A9,$FF,$46,$7B,$DF,$0F,$DC,$FE,$6F
	.dc.b	$C3,$BC,$BF,$D7,$9B,$7F,$BB,$5E,$D9,$F5,$FD,$2B,$C5,$18,$C7,$FC
	.dc.b	$89,$CB,$F1,$B0,$FE,$A3,$FB,$9F,$F9,$87,$E0,$9F,$64,$FC,$A5,$FA
	.dc.b	$4B,$F2,$AF,$A9,$7B,$AD,$4B,$D8,$CF,$E1,$DE,$4A,$FF,$36,$65,$F3
	.dc.b	$5E,$18,$FD,$EC,$FC,$5C,$09,$CD,$7F,$B9,$7D,$27,$EE,$8F,$CF,$1F
	.dc.b	$A6,$3F,$2D,$7E,$86,$FA,$3F,$F2,$CE,$9E,$DC,$78,$E1,$C2,$F7,$5F
	.dc.b	$E5,$7A,$D6,$3D,$72,$FF,$2E,$61,$E0,$F8,$1F,$F7,$43,$CB,$3F,$06
	.dc.b	$7F,$E0,$1A,$FF,$6E,$7C,$2F,$FA,$CD,$FE,$0B,$87,$9E,$B7,$F4,$BE
	.dc.b	$BF,$D7,$DF,$7E,$FD,$4D,$F9,$37,$F0,$CF,$D0,$9F,$B5,$F4,$CF,$FD
	.dc.b	$7D,$B3,$1E,$38,$70,$85,$BF,$64,$E1,$7F,$C5,$AF,$FD,$F6,$FF,$91
	.dc.b	$36,$F3,$5F,$3F,$73,$DB,$5D,$33,$FE,$3B,$EE,$79,$F5,$F3,$47,$F2
	.dc.b	$0F,$C0,$BE,$A7,$E9,$1E,$6F,$B7,$7F,$00,$BC,$6F,$C3,$63,$E2,$D1
	.dc.b	$7B,$AD,$78,$7F,$EA,$8E,$9F,$AD,$3E,$9E,$0F,$A2,$FC,$26,$F2,$7F
	.dc.b	$B8,$AF,$B7,$C5,$F8,$A3,$19,$BE,$3B,$FC,$67,$C7,$FD,$EB,$4F,$F0
	.dc.b	$9F,$E5,$EF,$DD,$BF,$DA,$5A,$FE,$87,$CA,$FE,$B1,$E6,$7F,$3C,$F9
	.dc.b	$E5,$EF,$D3,$FF,$26,$E5,$FE,$FB,$E0,$FF,$50,$61,$C1,$B1,$DE,$F8
	.dc.b	$FE,$95,$E2,$C7,$A6,$2F,$96,$F2,$FD,$E3,$F7,$BF,$EE,$7D,$1F,$9C
	.dc.b	$FC,$1F,$BC,$E7,$C6,$FF,$80,$3E,$27,$F9,$6B,$87,$D1,$B7,$F3,$DD
	.dc.b	$FC,$37,$FE,$57,$8F,$F4,$46,$5D,$3F,$14,$EE,$F6,$EF,$F8,$BF,$5A
	.dc.b	$FD,$45,$FA,$2F,$F9,$2B,$DE,$FE,$AF,$A7,$6A,$D4,$E7,$F2,$8F,$87
	.dc.b	$F9,$BF,$79,$98,$72,$FB,$3F,$F1,$76,$FF,$22,$78,$C7,$F6,$88,$FC
	.dc.b	$AF,$EB,$7F,$B9,$FA,$67,$EE,$AF,$CF,$7F,$BF,$7A,$BF,$DA,$F6,$FE
	.dc.b	$2D,$7B,$AD,$7C,$F7,$F0,$3F,$97,$E7,$EB,$FF,$89,$73,$FF,$6A,$5F
	.dc.b	$F0,$5F,$CC,$3F,$CB,$B8,$F2,$AF,$F1,$56,$3C,$BF,$4B,$FF,$2C,$74
	.dc.b	$7E,$E7,$0E,$DA,$FF,$83,$FF,$1F,$CE,$61,$2F,$C0,$DF,$A3,$BF,$11
	.dc.b	$FF,$53,$FE,$34,$FD,$6D,$F9,$F7,$C0,$BD,$D6,$A7,$FC,$07,$C1,$EE
	.dc.b	$97,$E9,$5E,$8D,$BC,$91,$DB,$4F,$6B,$CB,$F0,$E6,$6F,$9C,$F3,$D7
	.dc.b	$9A,$B5,$3E,$78,$FA,$3F,$C0,$3F,$EB,$AF,$BE,$7E,$38,$FD,$4F,$FB
	.dc.b	$D7,$E8,$1A,$FC,$3B,$FD,$37,$39,$FA,$27,$E4,$19,$CB,$E7,$E5,$FE
	.dc.b	$F5,$6D,$F0,$C7,$F8,$FC,$73,$CB,$BC,$3F,$04,$F3,$F7,$3F,$47,$7E
	.dc.b	$1D,$7F,$66,$7E,$A6,$FD,$6D,$1F,$C7,$AF,$85,$6E,$B5,$F8,$CE,$70
	.dc.b	$7E,$BB,$F2,$F5,$FB,$58,$70,$7F,$D1,$1D,$1F,$CE,$63,$9D,$76,$C3
	.dc.b	$FC,$AF,$D5,$FD,$9E,$73,$9F,$93,$E0,$7D,$3F,$B5,$FE,$AB,$EF,$3E
	.dc.b	$2F,$BD,$FC,$FB,$CF,$F3,$4E,$79,$D3,$FD,$0B,$8B,$3F,$88,$F9,$78
	.dc.b	$1A,$30,$FD,$AC,$79,$8F,$4B,$CE,$FE,$6E,$BE,$58,$C7,$AF,$F3,$3F
	.dc.b	$26,$7E,$65,$FE,$69,$F9,$F7,$ED,$EF,$97,$FF,$09,$63,$ED,$BD,$FA
	.dc.b	$5F,$E0,$2A,$D7,$E0,$7F,$3E,$79,$EB,$E3,$3C,$EB,$F4,$FF,$56,$E8
	.dc.b	$E7,$9F,$83,$3F,$26,$9E,$3C,$BF,$D2,$FA,$7F,$84,$FF,$7B,$FF,$25
	.dc.b	$7D,$A7,$EA,$7E,$25,$FB,$D3,$F4,$CF,$D6,$5E,$FC,$75,$EE,$B5,$EC
	.dc.b	$7F,$62,$C7,$FC,$A9,$F0,$5F,$35,$EB,$C5,$E9,$7B,$FA,$3F,$ED,$8B
	.dc.b	$FE,$33,$97,$5C,$7A,$7C,$7F,$EB,$2D,$FA,$8F,$F9,$47,$EF,$FF,$DF
	.dc.b	$FF,$80,$3F,$59,$FF,$0D,$63,$EB,$AD,$79,$BF,$E8,$39,$7E,$D7,$E5
	.dc.b	$9F,$F2,$D6,$DE,$35,$DB,$E7,$78,$E3,$FE,$27,$F2,$7F,$C1,$DA,$78
	.dc.b	$BE,$4F,$FD,$19,$9F,$FA,$3F,$1F,$37,$DE,$3F,$16,$7E,$98,$FC,$67
	.dc.b	$FC,$AB,$FA,$3B,$F0,$3F,$8D,$7B,$AD,$7D,$CA,$9F,$9C,$6B,$1F,$95
	.dc.b	$6F,$CF,$6C,$76,$CB,$FC,$C3,$F4,$9E,$28,$C6,$2F,$A5,$FE,$2B,$F3
	.dc.b	$4F,$DE,$3F,$94,$7E,$F3,$FC,$AB,$F0,$17,$9B,$FE,$A9,$EC,$FF,$2B
	.dc.b	$5E,$FE,$21,$7B,$A7,$E7,$DF,$E2,$0A,$D7,$2F,$E5,$5E,$DE,$8F,$EB
	.dc.b	$5E,$57,$DF,$F1,$FF,$1E,$9E,$2F,$8C,$6B,$FE,$BA,$C7,$A7,$E7,$3F
	.dc.b	$A7,$FC,$4D,$FB,$03,$F1,$47,$EC,$8F,$CE,$77,$EA,$7F,$B4,$AB,$53
	.dc.b	$BE,$4F,$5E,$2F,$CA,$BC,$8F,$EC,$F3,$F4,$3D,$EF,$B7,$D3,$FC,$88
	.dc.b	$C6,$7F,$E5,$1C,$3E,$2F,$D0,$F8,$BF,$4E,$7D,$FF,$F9,$97,$AD,$FB
	.dc.b	$17,$C5,$AD,$4E,$7F,$34,$C6,$FF,$16,$7F,$87,$ED,$FE,$47,$3F,$B0
	.dc.b	$74,$7F,$81,$A5,$97,$6F,$5B,$FD,$3F,$5C,$A3,$1F,$4D,$FD,$9F,$57
	.dc.b	$E3,$BF,$E6,$5F,$D2,$9F,$4C,$FD,$47,$FA,$87,$E5,$FF,$CD,$BF,$35
	.dc.b	$FE,$9B,$87,$0C,$FF,$B8,$61,$C2,$F7,$F2,$4B,$7C,$D3,$0E,$18,$30
	.dc.b	$EF,$F9,$5F,$F8,$0B,$4F,$16,$9E,$1A,$FF,$3C,$1E,$78,$CE,$7D,$BF
	.dc.b	$1A,$C4,$F8,$AB,$D1,$F3,$2D,$BF,$A6,$F8,$9A,$AF,$07,$E3,$2F,$D2
	.dc.b	$9F,$67,$3F,$2D,$7E,$B5,$FE,$7B,$E1,$F3,$F0,$FF,$BB,$EF,$76,$BD
	.dc.b	$F2,$7B,$78,$7E,$56,$5D,$73,$F4,$99,$3E,$3B,$FD,$D9,$CE,$5D,$A3
	.dc.b	$C9,$87,$8F,$E0,$FD,$19,$F8,$63,$CD,$CD,$EB,$F8,$6F,$F3,$67,$E3
	.dc.b	$0F,$73,$FC,$39,$F8,$F3,$F5,$5F,$E6,$9A,$7D,$0B,$DD,$6A,$5F,$AF
	.dc.b	$F9,$76,$FF,$12,$7C,$E7,$FB,$F1,$F2,$F0,$7F,$AA,$70,$FF,$3D,$7E
	.dc.b	$57,$F5,$E7,$FC,$61,$1D,$06,$7E,$18,$E9,$19,$87,$D7,$7E,$47,$93
	.dc.b	$FD,$69,$F7,$BF,$E3,$3F,$87,$7E,$AE,$FD,$B7,$7F,$BC,$BF,$3E,$FC
	.dc.b	$56,$1C,$33,$61,$C2,$F7,$F9,$4E,$B5,$FF,$22,$67,$D5,$F9,$F8,$71
	.dc.b	$F7,$7F,$CA,$2F,$3C,$39,$6B,$E6,$AE,$42,$FC,$DA,$F9,$67,$DD,$CB
	.dc.b	$F8,$3A,$FF,$29,$7E,$3D,$F9,$DF,$CD,$78,$7E,$B2,$FD,$11,$F2,$BF
	.dc.b	$D9,$30,$E1,$83,$0E,$19,$B0,$E1,$7F,$ED,$4B,$DD,$7E,$A7,$DE,$F4
	.dc.b	$E1,$E0,$EF,$67,$88,$E9,$18,$CB,$A6,$7F,$E9,$5C,$79,$CE,$73,$F3
	.dc.b	$7C,$77,$D5,$C3,$FC,$D9,$F8,$37,$ED,$E9,$FB,$A3,$E8,$3F,$53,$FB
	.dc.b	$F3,$D6,$F8,$F7,$BA,$D7,$8A,$A7,$EC,$9A,$C5,$FA,$AF,$9F,$0E,$D9
	.dc.b	$FD,$63,$C3,$3F,$C1,$F1,$E7,$F6,$7E,$59,$FF,$61,$CE,$74,$FD,$DF
	.dc.b	$9D,$FE,$83,$FC,$F3,$FD,$39,$5F,$D0,$FE,$71,$FA,$8F,$F9,$43,$F2
	.dc.b	$FF,$F2,$BF,$77,$F9,$76,$F7,$F0,$7F,$E0,$2A,$D7,$D9,$7E,$FF,$F5
	.dc.b	$65,$6B,$1E,$79,$7E,$FA,$31,$F8,$5F,$F0,$08,$DF,$A5,$FE,$95,$7C
	.dc.b	$58,$F4,$8F,$31,$9C,$BC,$1E,$A3,$57,$DB,$F1,$CB,$D7,$F0,$9F,$E1
	.dc.b	$6F,$B1,$FE,$6D,$F8,$D3,$D0,$7E,$FA,$FD,$BB,$C3,$BF,$FA,$97,$0E
	.dc.b	$1F,$02,$C3,$85,$BF,$43,$DD,$AF,$B3,$ED,$E0,$F5,$47,$F8,$9B,$F1
	.dc.b	$2F,$76,$BE,$4E,$46,$DE,$4D,$BF,$D0,$F8,$F4,$C3,$C3,$A7,$9F,$C3
	.dc.b	$F8,$33,$FE,$B9,$F8,$1F,$F1,$8F,$EC,$4F,$9E,$7E,$E3,$FE,$86,$FA
	.dc.b	$AF,$46,$DF,$72,$F7,$5A,$E7,$FF,$1A,$FE,$0D,$EA,$9E,$79,$7A,$4E
	.dc.b	$DC,$0F,$13,$D3,$C7,$F5,$6B,$9E,$9D,$EF,$95,$B5,$F2,$4F,$FA,$63
	.dc.b	$F1,$7F,$C7,$7B,$BF,$77,$E6,$9F,$87,$FF,$1E,$FD,$99,$FA,$83,$F7
	.dc.b	$77,$E8,$4F,$9A,$FC,$5E,$3C,$70,$E1,$7B,$A7,$C2,$EB,$7F,$F6,$F6
	.dc.b	$7D,$BF,$3A,$FA,$38,$F1,$F7,$6D,$6B,$2E,$7A,$72,$8E,$43,$0F,$3D
	.dc.b	$35,$F3,$D3,$F1,$EF,$6D,$F8,$3E,$03,$F8,$6C,$BF,$1B,$7E,$BB,$FE
	.dc.b	$81,$FC,$67,$AE,$7F,$CD,$3D,$FF,$81,$5A,$FB,$C7,$E8,$3A,$CB,$F8
	.dc.b	$BE,$7F,$A9,$E0,$F9,$51,$FE,$3E,$8C,$1E,$1C,$39,$E6,$F9,$1F,$2C
	.dc.b	$FA,$EB,$CB,$97,$93,$B3,$3A,$3F,$61,$7E,$5C,$FA,$3F,$C6,$FE,$2E
	.dc.b	$FE,$00,$FF,$58,$78,$DF,$E1,$6C,$38,$5E,$FB,$D7,$BF,$A8,$F7,$E4
	.dc.b	$F8,$EF,$A3,$BF,$CD,$8E,$EF,$A8,$C6,$1C,$EF,$FB,$36,$5D,$EB,$CB
	.dc.b	$A7,$93,$C1,$F6,$FB,$7F,$E6,$CE,$7E,$DF,$82,$3F,$53,$7E,$05,$FD
	.dc.b	$8B,$2F,$F1,$6D,$EF,$6F,$F6,$6D,$6B,$F1,$3F,$C0,$CF,$F9,$DB,$E6
	.dc.b	$3C,$EF,$5E,$0F,$33,$C7,$2E,$D9,$78,$BE,$C7,$FD,$7D,$97,$4F,$43
	.dc.b	$CB,$EA,$FE,$DB,$F2,$A7,$E7,$EF,$CD,$5F,$B7,$BF,$47,$F0,$DE,$EB
	.dc.b	$52,$FD,$B7,$3D,$7E,$4A,$FF,$01,$70,$79,$1E,$18,$BE,$23,$C3,$4C
	.dc.b	$FC,$33,$FE,$85,$E5,$F8,$7F,$A0,$FF,$70,$FD,$61,$FB,$C7,$FB,$E7
	.dc.b	$F1,$FF,$F7,$DD,$6B,$E7,$15,$AF,$BC,$57,$E7,$AB,$9F,$D8,$FC,$8C
	.dc.b	$7F,$BD,$EF,$F2,$1E,$D9,$75,$8F,$9F,$18,$C1,$F9,$07,$F5,$5F,$C2
	.dc.b	$BF,$A2,$6E,$FC,$17,$FC,$4B,$E2,$FF,$7A,$FD,$55,$F7,$7F,$95,$D9
	.dc.b	$87,$0B,$DF,$A0,$BD,$FE,$2D,$FD,$5A,$B5,$ED,$7F,$36,$78,$7B,$37
	.dc.b	$80,$7F,$2E,$C7,$9E,$1E,$08,$E4,$6B,$5F,$35,$7E,$AB,$2F,$07,$EA
	.dc.b	$3E,$BA,$D7,$DA,$BF,$21,$FE,$15,$EC,$BF,$16,$7E,$60,$FB,$9E,$B8
	.dc.b	$7F,$3E,$C3,$87,$BC,$C3,$85,$FF,$B9,$F8,$FE,$0D,$F6,$FD,$2F,$D4
	.dc.b	$F1,$CE,$EE,$CB,$9E,$3E,$1F,$07,$CD,$7B,$CF,$B6,$AF,$9F,$CE,$2F
	.dc.b	$C5,$1F,$B8,$FF,$B8,$FE,$15,$F8,$C7,$F3,$BE,$3F,$4D,$EE,$B5,$3F
	.dc.b	$CF,$FF,$3B,$CE,$1F,$E2,$5D,$3F,$C0,$DD,$3F,$80,$8E,$B1,$83,$E7
	.dc.b	$3E,$0F,$C2,$BF,$A0,$DF,$FB,$D6,$79,$E1,$DB,$E4,$BD,$7F,$A7,$3F
	.dc.b	$57,$7E,$60,$FC,$AB,$F6,$1F,$4A,$F5,$3F,$D9,$FD,$7F,$2B,$36,$3C
	.dc.b	$70,$E1,$EA,$30,$E1,$F3,$5F,$5A,$B5,$7E,$73,$8E,$6F,$D8,$FE,$9C
	.dc.b	$F3,$D3,$C3,$A7,$9A,$BC,$B1,$E4,$F8,$0F,$4F,$E5,$8F,$DD,$BF,$CC
	.dc.b	$1F,$A3,$FF,$88,$3D,$DF,$DA,$F3,$FF,$8A,$73,$5E,$FE,$CD,$7B,$FD
	.dc.b	$37,$CD,$3C,$F7,$F3,$5F,$0F,$C7,$7C,$B3,$C6,$FB,$57,$0C,$FC,$3F
	.dc.b	$51,$F1,$E3,$E0,$7D,$33,$F1,$4F,$F2,$BF,$EC,$DE,$7F,$D4,$5F,$A6
	.dc.b	$6A,$73,$8A,$5F,$BA,$F2,$E4,$F8,$B8,$F3,$CB,$CA,$F5,$FB,$66,$7F
	.dc.b	$0F,$D7,$FB,$B3,$98,$F8,$DF,$95,$7E,$6B,$E3,$FF,$A9,$FD,$37,$F1
	.dc.b	$7F,$C8,$BF,$B9,$FD,$17,$EB,$27,$3F,$02,$F8,$05,$7F,$78,$75,$F2
	.dc.b	$BF,$E9,$1D,$BE,$E5,$F4,$6C,$7F,$89,$EF,$95,$FF,$80,$27,$8E,$3E
	.dc.b	$1F,$94,$FD,$6F,$C9,$7F,$93,$7B,$3F,$8E,$79,$BF,$E1,$AF,$D3,$3F
	.dc.b	$C7,$7E,$57,$E4,$6B,$5F,$01,$FA,$FF,$4D,$5F,$C8,$2B,$5F,$58,$C7
	.dc.b	$F7,$F8,$7F,$96,$7F,$2A,$B3,$9C,$79,$61,$E2,$8C,$65,$FE,$2A,$FA
	.dc.b	$0F,$2F,$E9,$6F,$D9,$9F,$66,$FC,$6F,$F7,$FF,$9D,$BF,$31,$FE,$2B
	.dc.b	$B3,$0E,$17,$BF,$5D,$6F,$CC,$37,$5D,$79,$BF,$4D,$87,$1F,$D9,$BD
	.dc.b	$33,$CB,$3F,$E1,$BF,$85,$F9,$35,$F1,$C7,$F1,$0F,$85,$FE,$2F,$F9
	.dc.b	$FB,$F2,$07,$E6,$EF,$D6,$DF,$BD,$7F,$86,$7F,$05,$FD,$2F,$67,$EF
	.dc.b	$70,$E1,$7B,$AD,$1F,$CC,$23,$3C,$1F,$BB,$7B,$7D,$DF,$F2,$C7,$8E
	.dc.b	$3D,$BE,$B3,$F8,$96,$3D,$73,$F1,$CE,$7E,$0F,$CF,$FA,$87,$F3,$EF
	.dc.b	$7B,$F8,$27,$E3,$8F,$A7,$FD,$3E,$3E,$5A,$D7,$C1,$2B,$5F,$B1,$79
	.dc.b	$67,$FA,$3E,$FE,$55,$FB,$B7,$95,$E3,$5F,$E7,$59,$E1,$F2,$BF,$C8
	.dc.b	$F0,$E9,$C7,$F2,$BF,$3E,$FD,$6E,$3F,$E3,$7E,$4F,$F5,$DF,$9B,$F0
	.dc.b	$E7,$3F,$55,$F3,$36,$FF,$33,$1F,$2E,$3F,$6B,$9F,$C7,$7F,$D8,$6D
	.dc.b	$F5,$3C,$B5,$F5,$3F,$29,$F0,$46,$31,$7C,$8F,$F5,$2F,$AD,$7E,$3D
	.dc.b	$FE,$E9,$FD,$71,$FB,$37,$F0,$CF,$C4,$5E,$DF,$F6,$6B,$DF,$E2,$8B
	.dc.b	$DF,$C2,$FF,$BB,$3F,$47,$D6,$5F,$EE,$EE,$0F,$E6,$D8,$70,$6C,$77
	.dc.b	$CF,$F8,$CE,$3E,$38,$C6,$1E,$2A,$F0,$7E,$66,$F8,$D7,$88,$FD,$F3
	.dc.b	$FD,$95,$F2,$AD,$37,$5A,$9C,$FC,$BE,$31,$F1,$E2,$FC,$8F,$FB,$CE
	.dc.b	$08,$ED,$97,$5F,$61,$F8,$5F,$C7,$9C,$E5,$FE,$D9,$EE,$F9,$5F,$28
	.dc.b	$FC,$99,$FB,$EF,$E3,$FC,$37,$EC,$5F,$D8,$FC,$1F,$EC,$1B,$7F,$9A
	.dc.b	$2F,$7E,$47,$F8,$52,$9F,$9D,$FC,$FC,$BF,$CF,$FC,$5F,$29,$EC,$DB
	.dc.b	$E3,$97,$DC,$8E,$11,$8F,$8C,$FF,$07,$FE,$2E,$FC,$D7,$F8,$77,$D0
	.dc.b	$7F,$44,$E6,$7E,$C0,$F7,$3F,$C4,$77,$BF,$61,$7B,$A7,$EE,$3A,$C5
	.dc.b	$F8,$D7,$F8,$8B,$0E,$1A,$FF,$75,$79,$61,$DF,$F4,$8F,$1D,$F8,$BE
	.dc.b	$5B,$E9,$3E,$6F,$65,$F9,$63,$F5,$A7,$EE,$AF,$A5,$FF,$74,$7D,$3F
	.dc.b	$9A,$F7,$5A,$9F,$F7,$67,$AF,$9F,$C6,$7C,$95,$D3,$7F,$23,$C3,$1E
	.dc.b	$9F,$8B,$78,$73,$F0,$47,$7F,$33,$E4,$FE,$89,$FE,$1B,$FB,$6F,$C4
	.dc.b	$7C,$53,$F7,$67,$EE,$1F,$22,$B5,$D9,$FC,$CE,$5F,$7D,$9F,$A2,$F8
	.dc.b	$D5,$F7,$B7,$F8,$CF,$0C,$7F,$96,$C6,$F2,$EF,$E6,$7F,$99,$BF,$27
	.dc.b	$FF,$14,$77,$FE,$33,$F4,$27,$ED,$FF,$E7,$6F,$D6,$1F,$9D,$3F,$64
	.dc.b	$61,$C3,$CB,$61,$C3,$C7,$61,$C3,$F5,$1D,$FD,$97,$F9,$D3,$7F,$EA
	.dc.b	$98,$3F,$7E,$C4,$67,$2E,$99,$62,$3C,$D7,$C7,$6F,$36,$DE,$5F,$C1
	.dc.b	$B3,$5F,$E8,$09,$CE,$BE,$2F,$C2,$3F,$CC,$BF,$AB,$FF,$AE,$7F,$28
	.dc.b	$7E,$92,$F9,$A7,$ED,$8F,$D8,$5F,$AC,$7F,$9E,$B0,$E1,$7B,$AD,$7E
	.dc.b	$69,$9F,$C4,$99,$7F,$9E,$2F,$FD,$8D,$F1,$76,$73,$87,$1B,$F1,$C6
	.dc.b	$33,$F0,$EF,$EA,$3F,$DE,$7F,$2E,$FC,$F1,$F7,$EF,$C8,$5F,$A9,$BE
	.dc.b	$85,$F8,$AF,$CE,$C7,$FD,$43,$F6,$AC,$B9,$63,$C7,$0E,$1F,$8A,$6E
	.dc.b	$FB,$E7,$F3,$DC,$61,$C7,$C5,$C7,$77,$F9,$B6,$3A,$57,$93,$5F,$1F
	.dc.b	$AF,$E9,$9C,$E7,$DF,$E8,$7F,$D6,$7F,$0E,$FC,$93,$F0,$5F,$1F,$FC
	.dc.b	$9F,$F1,$AF,$37,$E6,$8F,$9D,$7E,$51,$FD,$B7,$0E,$1C,$3F,$EB,$CC
	.dc.b	$38,$72,$30,$E1,$6F,$6D,$87,$0C,$3F,$D3,$BE,$AF,$FB,$B5,$FC,$64
	.dc.b	$67,$1F,$14,$63,$4F,$2D,$7F,$BF,$B2,$ED,$9F,$96,$BA,$D6,$B4,$F4
	.dc.b	$CB,$F0,$EF,$B5,$FA,$3F,$FA,$BB,$F0,$57,$EB,$EF,$DD,$1F,$95,$FF
	.dc.b	$26,$FD,$93,$E9,$5E,$EB,$53,$FC,$AB,$B3,$3F,$6A,$FD,$F3,$E9,$FA
	.dc.b	$3F,$AB,$C3,$B7,$07,$E9,$FB,$BF,$D4,$51,$CF,$4E,$FF,$15,$E4,$C5
	.dc.b	$B7,$92,$7C,$7D,$BF,$91,$FC,$79,$F9,$BF,$F8,$67,$7F,$C9,$7E,$7A
	.dc.b	$FC,$0D,$F9,$1F,$E2,$6F,$7F,$63,$EC,$F1,$57,$BF,$D2,$B5,$97,$F8
	.dc.b	$8F,$8B,$CD,$C7,$FC,$0F,$1E,$FF,$AA,$FB,$27,$FD,$F3,$1D,$35,$F0
	.dc.b	$EB,$E5,$C3,$C9,$9F,$6F,$CC,$3F,$DC,$3E,$53,$F7,$CF,$F7,$B7,$2F
	.dc.b	$EF,$3F,$5C,$67,$FA,$AF,$13,$FD,$8D,$7B,$FC,$81,$7B,$A7,$EE,$DF
	.dc.b	$EF,$AA,$D6,$FF,$EC,$57,$B7,$E4,$5F,$EF,$67,$95,$F7,$FC,$1B,$C7
	.dc.b	$A7,$8B,$DA,$6B,$F2,$71,$FD,$87,$E7,$2F,$C4,$BF,$89,$FE,$D5,$FE
	.dc.b	$11,$F8,$F7,$ED,$6F,$DE,$B7,$5A,$FA,$DF,$F7,$14,$FF,$71,$F1,$7F
	.dc.b	$DC,$18,$F3,$FC,$17,$E5,$66,$8E,$DE,$FE,$1B,$A7,$7F,$80,$FE,$21
	.dc.b	$18,$CB,$FD,$DD,$F1,$1F,$B8,$F0,$3F,$9C,$7E,$A2,$FD,$C5,$F9,$1B
	.dc.b	$F7,$97,$E7,$2F,$B5,$7E,$B0,$FD,$B3,$F6,$79,$72,$C7,$8E,$1C,$2F
	.dc.b	$7F,$47,$AE,$20,$D3,$C1,$F2,$7F,$C5,$7C,$4F,$F7,$F4,$78,$F6,$C6
	.dc.b	$7D,$5F,$26,$DE,$3F,$C9,$BD,$58,$F4,$BE,$1A,$78,$7E,$13,$FB,$17
	.dc.b	$E7,$1F,$D5,$7F,$61,$7E,$38,$FC,$21,$FA,$0F,$F2,$FF,$B1,$FF,$7E
	.dc.b	$7C,$37,$F6,$2C,$38,$5E,$EB,$5F,$92,$7F,$1F,$F0,$15,$AF,$D4,$3F
	.dc.b	$95,$BC,$32,$BB,$FC,$0D,$3C,$31,$E0,$F9,$1F,$F1,$1F,$0F,$9A,$B5
	.dc.b	$A7,$79,$FF,$13,$FD,$CF,$F4,$9F,$7D,$FB,$6B,$F5,$37,$C9,$FF,$AB
	.dc.b	$3F,$3E,$7E,$8C,$FC,$D3,$FA,$16,$1C,$2F,$7F,$87,$AF,$7F,$A1,$EB
	.dc.b	$2E,$9B,$7E,$E8,$FA,$F8,$78,$3F,$53,$FC,$8B,$CF,$4F,$07,$CE,$F9
	.dc.b	$6F,$C9,$1E,$3F,$9C,$FE,$C1,$F9,$27,$E9,$BF,$66,$7E,$C8,$FB,$17
	.dc.b	$E2,$AF,$DC,$1F,$7F,$FD,$67,$0E,$17,$BA,$D7,$E2,$59,$F7,$19,$7E
	.dc.b	$92,$FF,$67,$FB,$65,$9C,$E1,$C7,$E1,$FC,$71,$8D,$3C,$3F,$15,$FD
	.dc.b	$D7,$F1,$FF,$F4,$27,$EC,$1F,$E5,$AF,$DC,$BF,$0D,$FB,$3B,$DF,$FE
	.dc.b	$D9,$5A,$FC,$19,$5A,$FD,$7B,$3F,$88,$7F,$A0,$BB,$3E,$50,$BE,$AE
	.dc.b	$1E,$55,$F9,$3F,$77,$C5,$4C,$FC,$5E,$47,$FA,$B3,$0F,$EF,$9F,$BD
	.dc.b	$BF,$3A,$7E,$98,$F9,$2F,$E4,$7F,$1B,$FE,$63,$FD,$07,$75,$A9,$CF
	.dc.b	$E1,$6F,$D8,$F3,$87,$F0,$CC,$7A,$7E,$47,$70,$E1,$F2,$CC,$DF,$98
	.dc.b	$7F,$86,$EF,$C5,$87,$86,$B5,$F9,$9F,$F8,$13,$EB,$3F,$C8,$9E,$DF
	.dc.b	$DC,$FB,$AF,$F3,$A7,$CD,$BF,$48,$FE,$A7,$E9,$7F,$70,$E4,$FE,$0B
	.dc.b	$7B,$FC,$43,$FC,$4D,$5F,$D6,$3F,$69,$D5,$7E,$E3,$CB,$7E,$ED,$FC
	.dc.b	$27,$4E,$F1,$FE,$3F,$9F,$F0,$01,$D1,$FF,$21,$61,$D3,$8B,$CD,$39
	.dc.b	$D7,$CB,$97,$BA,$3F,$C5,$5F,$07,$FE,$4A,$F4,$DF,$AB,$7F,$84,$BF
	.dc.b	$5A,$7C,$77,$FB,$17,$E1,$1F,$19,$FE,$63,$C3,$85,$EE,$BF,$DA,$95
	.dc.b	$AC,$7F,$91,$D7,$57,$F0,$EF,$F9,$63,$F1,$0F,$E7,$EF,$2F,$8E,$65
	.dc.b	$FE,$D3,$AF,$2F,$CE,$B4,$F2,$D6,$BE,$2F,$FC,$A1,$FB,$FB,$F5,$A7
	.dc.b	$C5,$F9,$BF,$76,$7E,$C7,$FE,$DD,$FC,$3B,$FC,$91,$7B,$AD,$4F,$FB
	.dc.b	$87,$CE,$CE,$1F,$EA,$DD,$7F,$3B,$7B,$6C,$76,$E4,$65,$DB,$EE,$FC
	.dc.b	$73,$98,$F0,$FC,$0F,$FA,$5F,$F0,$AF,$F4,$57,$EA,$AF,$D3,$DF,$96
	.dc.b	$39,$FF,$C9,$3F,$B9,$BF,$66,$7E,$7D,$FE,$98,$C7,$8E,$1C,$2F,$74
	.dc.b	$FE,$01,$D7,$07,$F5,$2C,$FB,$7C,$D7,$F9,$8B,$1E,$3A,$5A,$D6,$5C
	.dc.b	$FF,$05,$6B,$FD,$FF,$BB,$FD,$57,$5A,$BF,$35,$6B,$E0,$FF,$AA,$CF
	.dc.b	$F9,$6F,$F3,$E7,$EB,$AF,$D7,$DF,$22,$FD,$E3,$F8,$DF,$EA,$5F,$D9
	.dc.b	$F0,$E1,$7B,$AD,$7C,$6F,$F2,$8F,$0B,$F1,$55,$D7,$F1,$8F,$EC,$EF
	.dc.b	$F6,$EF,$C3,$1A,$78,$23,$FC,$CB,$DD,$FE,$B7,$9C,$E5,$D7,$47,$E4
	.dc.b	$7E,$89,$CE,$DE,$69,$FF,$67,$F1,$78,$3E,$C5,$F2,$4F,$BC,$FC,$6F
	.dc.b	$EF,$BF,$ED,$AF,$DC,$9F,$9C,$7F,$A4,$EF,$75,$AF,$C8,$94,$FD,$3F
	.dc.b	$58,$BF,$6A,$FF,$0F,$E1,$C3,$8B,$F9,$5B,$CB,$CC,$69,$FE,$E8,$FD
	.dc.b	$5B,$FD,$5D,$39,$8F,$F7,$5F,$C9,$FD,$2F,$C4,$1F,$A5,$FF,$84,$7D
	.dc.b	$97,$EF,$1F,$CD,$7F,$44,$7E,$67,$F4,$6F,$7F,$65,$FC,$5E,$B5,$F9
	.dc.b	$8E,$71,$FF,$0D,$6F,$F3,$0F,$F0,$ED,$F6,$FC,$2B,$E9,$C7,$FA,$26
	.dc.b	$7C,$17,$FD,$0F,$F1,$AF,$24,$63,$E6,$9F,$87,$FE,$23,$C8,$7E,$44
	.dc.b	$FC,$C1,$FB,$DF,$F6,$7F,$13,$FD,$F3,$E6,$FE,$1E,$F7,$5A,$FA,$9F
	.dc.b	$C7,$FE,$54,$FC,$13,$59,$3F,$31,$F2,$E1,$C3,$87,$D4,$1E,$7A,$77
	.dc.b	$FA,$2F,$25,$F8,$E3,$C4,$FF,$8F,$3C,$6F,$F5,$A7,$E3,$4F,$D3,$DF
	.dc.b	$43,$FE,$8D,$FC,$FD,$FB,$DF,$FA,$56,$F7,$5A,$FA,$A5,$6B,$07,$E9
	.dc.b	$8B,$DF,$D1,$FF,$B6,$3E,$5F,$FC,$97,$C1,$FE,$B2,$CB,$BB,$FE,$A7
	.dc.b	$8C,$E1,$E3,$D3,$C5,$A7,$F0,$BA,$E9,$5A,$BF,$16,$7E,$6F,$2B,$FD
	.dc.b	$F1,$FA,$93,$F4,$8F,$EB,$1E,$2F,$F9,$9F,$E3,$BF,$DA,$DF,$A2,$FF
	.dc.b	$BB,$EB,$5E,$42,$B5,$EB,$7F,$B1,$B8,$BF,$6F,$5A,$DB,$FC,$0A,$F5
	.dc.b	$FB,$0F,$F3,$EB,$CA,$FB,$72,$F8,$B4,$F0,$FE,$12,$D7,$EB,$63,$FC
	.dc.b	$1F,$CE,$7E,$EA,$FD,$33,$F9,$4F,$EA,$1F,$9E,$FC,$43,$FA,$17,$C0
	.dc.b	$FE,$E6,$B5,$E3,$2B,$5B,$F3,$97,$E9,$FF,$9A,$CB,$AE,$56,$78,$DF
	.dc.b	$EA,$BC,$2F,$E3,$F5,$AC,$7A,$79,$BF,$EA,$5C,$FF,$8B,$FD,$4B,$F4
	.dc.b	$BF,$F4,$17,$EB,$4F,$D7,$7F,$92,$61,$FD,$F2,$F7,$5A,$9C,$FD,$17
	.dc.b	$F2,$5C,$39,$7A,$DF,$AE,$7A,$7A,$96,$F8,$7C,$1F,$F8,$A7,$1E,$76
	.dc.b	$CF,$C1,$2F,$CD,$7F,$35,$F0,$DF,$88,$F8,$87,$E6,$3F,$EB,$8F,$23
	.dc.b	$FD,$13,$F9,$77,$F7,$2F,$0B,$F9,$8F,$96,$BD,$F2,$FE,$7B,$F0,$B5
	.dc.b	$BF,$64,$FC,$D6,$7D,$7F,$2C,$FF,$5D,$E3,$C7,$4B,$97,$2D,$3F,$D2
	.dc.b	$FE,$57,$93,$C1,$FF,$54,$DE,$F5,$ED,$F3,$39,$97,$DD,$BF,$56,$7D
	.dc.b	$F7,$F8,$63,$BB,$FD,$A9,$E8,$3A,$FC,$78,$7F,$AD,$30,$E1,$7B,$FC
	.dc.b	$77,$F7,$5D,$5A,$AE,$B9,$74,$7F,$CB,$B7,$DF,$CE,$FF,$6A,$BC,$BE
	.dc.b	$B3,$F9,$6E,$0F,$B5,$F3,$4E,$75,$FF,$73,$7C,$2F,$F0,$9F,$57,$F7
	.dc.b	$7F,$6C,$FD,$29,$F9,$AB,$F7,$27,$E2,$7F,$ED,$1F,$B9,$FF,$25,$78
	.dc.b	$4C,$78,$E1,$C2,$F7,$F8,$E2,$DF,$89,$FF,$00,$E3,$FF,$69,$E9,$DF
	.dc.b	$4F,$E1,$91,$8C,$7F,$C3,$2F,$F0,$B3,$11,$E3,$D3,$FD,$C5,$F9,$27
	.dc.b	$A6,$73,$9F,$83,$EA,$BE,$A7,$B4,$FC,$61,$FA,$37,$D5,$FC,$15,$FA
	.dc.b	$97,$F3,$5F,$CE,$BE,$75,$EE,$B5,$3F,$CE,$7F,$68,$CE,$1C,$B3,$FF
	.dc.b	$46,$DF,$5F,$96,$FE,$47,$1F,$C9,$39,$BF,$C0,$B3,$CF,$BB,$FD,$0D
	.dc.b	$8F,$FA,$2E,$73,$87,$92,$73,$E7,$FE,$2F,$F6,$F7,$EC,$8F,$D5,$9F
	.dc.b	$BB,$3E,$43,$FC,$A1,$FB,$EF,$FB,$53,$3F,$F0,$9E,$1C,$2F,$75,$AE
	.dc.b	$8F,$C5,$4E,$77,$FF,$52,$BD,$BF,$68,$F4,$C7,$87,$3F,$06,$7D,$FE
	.dc.b	$F9,$F9,$E6,$73,$EB,$0F,$D6,$3F,$45,$F4,$0F,$DB,$3F,$DF,$BF,$C0
	.dc.b	$0F,$DE,$1F,$16,$F8,$E7,$E0,$9F,$9D,$C3,$87,$E1,$AC,$38,$5E,$FE
	.dc.b	$4F,$59,$2B,$CF,$C3,$8F,$3E,$F3,$FF,$8D,$32,$E7,$97,$2F,$8B,$F2
	.dc.b	$E3,$E4,$DB,$C7,$F9,$4F,$F9,$1B,$F0,$8F,$F0,$8F,$DF,$3F,$51,$7A
	.dc.b	$5E,$DF,$83,$FF,$1A,$FC,$57,$F9,$02,$F7,$5A,$E6,$53,$F1,$CF,$F2
	.dc.b	$55,$7F,$87,$3F,$02,$FF,$09,$BD,$B8,$FF,$BF,$BC,$B2,$F9,$5F,$57
	.dc.b	$E3,$E2,$FF,$34,$67,$D7,$F4,$CF,$F2,$DF,$C0,$7F,$8D,$FF,$74,$7E
	.dc.b	$76,$FD,$77,$F3,$DF,$A6,$7F,$8A,$B8,$BF,$CA,$B7,$BA,$D7,$98,$A7
	.dc.b	$E9,$1F,$E5,$FA,$E9,$C1,$FE,$12,$7F,$C8,$9F,$AF,$7F,$8A,$9E,$59
	.dc.b	$7E,$AF,$97,$C7,$F7,$0D,$7C,$73,$FA,$FF,$1B,$F9,$26,$7F,$E0,$C8
	.dc.b	$F3,$79,$2F,$A4,$FC,$DF,$8D,$FC,$76,$B5,$4A,$D7,$92,$FC,$BF,$5F
	.dc.b	$4D,$EF,$E0,$FF,$7E,$E1,$FE,$18,$F0,$7F,$03,$F2,$7F,$E1,$48,$E7
	.dc.b	$8F,$7F,$D6,$BC,$78,$7C,$7C,$FB,$6B,$FE,$70,$CB,$F9,$77,$E6,$2F
	.dc.b	$CF,$7F,$D0,$BF,$84,$BE,$7F,$F2,$9F,$AC,$FF,$3C,$F2,$FF,$D0,$37
	.dc.b	$BF,$C1,$FF,$D5,$15,$AF,$C3,$B9,$E3,$FE,$8E,$CB,$F5,$B8,$70,$87
	.dc.b	$EF,$DE,$89,$0C,$7A,$67,$E1,$F3,$BC,$D8,$71,$CF,$BC,$38,$38,$F8
	.dc.b	$2F,$CD,$7F,$C6,$BF,$96,$21,$F9,$F3,$F1,$67,$CD,$BF,$65,$FD,$AF
	.dc.b	$B3,$BC,$78,$E1,$C2,$F7,$5F,$BA,$F9,$7D,$70,$FA,$33,$ED,$FA,$DF
	.dc.b	$ED,$30,$F1,$4E,$5F,$0B,$FC,$39,$CB,$FE,$AB,$D3,$C9,$A7,$8F,$E0
	.dc.b	$7D,$7A,$78,$A7,$39,$FF,$9D,$2F,$FC,$A9,$3F,$D6,$3E,$9F,$F0,$DF
	.dc.b	$B5,$BF,$38,$FF,$0F,$7C,$FF,$FA,$5B,$97,$BF,$CA,$FF,$84,$7C,$55
	.dc.b	$EF,$EA,$BF,$CD,$1F,$3B,$D3,$4F,$EA,$B9,$FF,$17,$7E,$CE,$9D,$F0
	.dc.b	$FE,$1E,$DC,$39,$7C,$07,$F9,$CE,$31,$AF,$F8,$BF,$D8,$FF,$03,$7E
	.dc.b	$44,$FD,$1B,$FD,$3F,$FA,$07,$F3,$DF,$8C,$7E,$66,$FC,$27,$FC,$01
	.dc.b	$7B,$F7,$57,$BA,$7E,$61,$AC,$7F,$77,$C1,$F8,$EF,$43,$FC,$EB,$8F
	.dc.b	$18,$C1,$F8,$E7,$A6,$1E,$1F,$C9,$7C,$B7,$E4,$F8,$D6,$AF,$CA,$FF
	.dc.b	$C3,$DF,$85,$8F,$F4,$77,$E2,$7F,$EC,$9F,$E0,$07,$E8,$EF,$C7,$FF
	.dc.b	$C3,$15,$BE,$FF,$F6,$B7,$BC,$BD,$FD,$BE,$AB,$FD,$31,$C3,$E6,$3B
	.dc.b	$2E,$EF,$D3,$46,$30,$E7,$6E,$0F,$2D,$F9,$31,$ED,$AF,$F8,$D7,$EC
	.dc.b	$5F,$B7,$7F,$27,$FD,$ED,$F8,$C3,$F4,$1F,$67,$EF,$5F,$E2,$FB,$DF
	.dc.b	$BA,$BD,$D7,$C1,$F9,$3D,$63,$FC,$4F,$83,$FC,$03,$E4,$BB,$3F,$D4
	.dc.b	$B9,$78,$3E,$0B,$D2,$F4,$CB,$C3,$F0,$5F,$E0,$7A,$DE,$3E,$5D,$BC
	.dc.b	$9F,$29,$FE,$4D,$5E,$D3,$EC,$DE,$AB,$F6,$B7,$EB,$9A,$FD,$23,$FC
	.dc.b	$91,$87,$0B,$DD,$6B,$0F,$C7,$70,$79,$F8,$7F,$88,$DF,$6F,$8D,$F9
	.dc.b	$EF,$18,$FA,$8F,$3C,$FC,$19,$BE,$AF,$CD,$8F,$96,$73,$F5,$BF,$ED
	.dc.b	$5F,$91,$FF,$63,$FE,$74,$FD,$79,$F9,$97,$F9,$3B,$BB,$D3,$FA,$7B
	.dc.b	$F6,$DF,$F6,$86,$1C,$2F,$7F,$50,$BD,$FD,$4E,$B8,$7F,$94,$E7,$D7
	.dc.b	$F5,$BF,$F0,$1E,$3C,$74,$B5,$AC,$B9,$E3,$F2,$3E,$63,$CD,$39,$AF
	.dc.b	$27,$17,$FA,$56,$3F,$95,$75,$FF,$B8,$3F,$2D,$7C,$93,$F3,$87,$F0
	.dc.b	$0F,$F9,$F7,$F3,$1F,$F4,$0F,$E9,$9C,$31,$FE,$7B,$87,$0B,$DF,$5A
	.dc.b	$FF,$D0,$FF,$73,$BB,$5F,$6E,$BF,$E5,$9A,$78,$74,$F0,$57,$F2,$B8
	.dc.b	$C6,$5D,$31,$FF,$60,$46,$34,$F2,$7D,$57,$F4,$6F,$BE,$FF,$7C,$FA
	.dc.b	$EF,$D2,$9F,$BD,$3F,$4D,$FF,$6D,$D3,$A9,$FA,$E7,$F4,$B8,$E3,$7C
	.dc.b	$2B,$75,$CB,$E4,$7F,$D0,$9B,$FF,$5C,$CB,$B6,$5F,$D1,$71,$E3,$EB
	.dc.b	$DA,$D6,$5C,$FF,$13,$FF,$46,$7C,$8F,$9E,$73,$87,$96,$7F,$D2,$1F
	.dc.b	$90,$FF,$60,$F3,$7F,$B4,$D7,$FA,$63,$F2,$1F,$EC,$9F,$40,$F3,$DF
	.dc.b	$8D,$3D,$6F,$F3,$5F,$0B,$1E,$3E,$43,$1E,$38,$70,$B7,$EC,$DF,$DA
	.dc.b	$EF,$75,$FD,$C7,$3F,$EB,$91,$89,$FE,$C4,$EB,$E7,$3F,$C1,$58,$78
	.dc.b	$BF,$05,$F4,$5F,$F8,$03,$2F,$05,$6B,$4F,$F7,$34,$FF,$B0,$7F,$13
	.dc.b	$7E,$5B,$FE,$CE,$FC,$BD,$FA,$53,$F5,$3F,$D4,$7E,$55,$BA,$D4,$E7
	.dc.b	$D2,$7A,$79,$F9,$6F,$E5,$79,$74,$FC,$3E,$E1,$C3,$2F,$D0,$7B,$5F
	.dc.b	$C1,$75,$5E,$F5,$F0,$D7,$FA,$06,$7F,$B4,$0F,$D8,$7F,$C9,$DF,$8B
	.dc.b	$FF,$B2,$F8,$DF,$47,$F3,$7D,$6D,$D6,$BF,$0E,$FF,$62,$CF,$D1,$FC
	.dc.b	$9F,$3F,$06,$C7,$FB,$7E,$5F,$E2,$6C,$3F,$CB,$5F,$7F,$88,$F0,$56
	.dc.b	$F1,$E7,$E9,$7F,$73,$9E,$13,$9D,$3C,$7F,$9D,$FF,$B2,$67,$F9,$5F
	.dc.b	$E0,$5F,$E6,$4F,$C9,$3F,$C8,$1F,$9F,$FF,$AC,$3F,$1D,$FF,$6C,$56
	.dc.b	$BE,$01,$5A,$FB,$87,$E8,$7A,$F8,$6F,$F4,$B5,$7F,$B7,$2F,$AF,$8B
	.dc.b	$67,$38,$76,$F0,$FC,$51,$8F,$29,$A7,$F9,$03,$D8,$FC,$EF,$E1,$FF
	.dc.b	$E6,$3F,$DF,$BF,$95,$7E,$66,$FA,$1F,$DC,$7E,$80,$F0,$BE,$C5,$EF
	.dc.b	$BD,$7B,$F6,$9F,$B5,$6F,$3E,$BF,$59,$FE,$5F,$E3,$FE,$E5,$8F,$76
	.dc.b	$C3,$5F,$0E,$BF,$92,$9E,$47,$F9,$6A,$BC,$19,$79,$7F,$4D,$FF,$2E
	.dc.b	$EB,$F8,$BF,$CC,$5F,$AA,$3F,$3C,$7E,$24,$F1,$BF,$AE,$7E,$D1,$FC
	.dc.b	$47,$0E,$17,$BA,$D7,$CB,$2A,$56,$BF,$31,$FF,$03,$E5,$DB,$2B,$8F
	.dc.b	$1C,$BF,$B4,$7B,$FE,$4E,$3F,$E1,$37,$BD,$3A,$FB,$BF,$C4,$78,$7E
	.dc.b	$77,$EE,$0F,$D0,$9F,$AE,$7F,$71,$F4,$1F,$8F,$BF,$08,$FE,$4D,$87
	.dc.b	$F8,$BF,$0E,$17,$BA,$79,$1A,$AE,$BF,$15,$FC,$8D,$EF,$95,$EA,$63
	.dc.b	$CB,$9F,$F8,$06,$BE,$59,$CD,$79,$27,$F8,$27,$0F,$A7,$EC,$FF,$D6
	.dc.b	$DF,$14,$FD,$73,$F9,$DF,$23,$F7,$8F,$EE,$7F,$0F,$FC,$E3,$66,$1C
	.dc.b	$2F,$74,$FC,$FF,$D1,$3D,$2B,$FC,$F9,$F1,$BF,$CC,$A3,$FD,$89,$FA
	.dc.b	$27,$AD,$E5,$C5,$FE,$25,$FD,$0F,$C9,$F8,$8F,$F9,$3B,$3F,$1F,$E5
	.dc.b	$9F,$B3,$FC,$DF,$F8,$E7,$E1,$8F,$D1,$9F,$96,$BF,$53,$FF,$01,$7C
	.dc.b	$97,$ED,$2F,$75,$A9,$FF,$03,$7D,$8E,$7C,$17,$C4,$FE,$CE,$FA,$F1
	.dc.b	$FF,$AD,$38,$FF,$A4,$C7,$2C,$FB,$78,$BE,$3C,$3F,$D6,$99,$F5,$EA
	.dc.b	$FF,$57,$F4,$FF,$50,$E3,$7E,$E6,$FD,$0D,$F8,$6B,$F2,$97,$E9,$0F
	.dc.b	$CD,$5F,$07,$FD,$FF,$1E,$38,$70,$BD,$D3,$F4,$5D,$65,$D7,$8F,$FC
	.dc.b	$00,$7F,$B2,$30,$F0,$FD,$1F,$F8,$1A,$31,$3E,$3D,$3C,$5A,$7F,$37
	.dc.b	$F0,$BF,$9E,$61,$E0,$9C,$FD,$77,$E7,$FF,$00,$FF,$69,$FE,$D6,$F9
	.dc.b	$B7,$E6,$9F,$EE,$3F,$C8,$5F,$90,$7E,$66,$F7,$5A,$F8,$55,$3C,$5D
	.dc.b	$70,$7F,$99,$CF,$EE,$FB,$7E,$3F,$0E,$F1,$F8,$A3,$FD,$C6,$F4,$CB
	.dc.b	$FD,$E5,$FA,$37,$96,$5A,$79,$67,$FD,$ED,$F9,$2F,$EE,$E7,$FD,$D7
	.dc.b	$F9,$8F,$ED,$BE,$E9,$FA,$DB,$F4,$47,$E3,$CF,$80,$78,$DF,$C6,$B1
	.dc.b	$63,$C7,$0E,$17,$BF,$A0,$D4,$FF,$02,$FB,$4F,$F4,$AB,$DE,$FF,$D3
	.dc.b	$4F,$1C,$BB,$FD,$A7,$96,$7A,$FB,$DF,$E7,$BC,$BA,$FE,$9B,$9F,$96
	.dc.b	$7E,$C4,$FD,$6B,$FD,$F1,$A7,$FA,$C7,$F6,$07,$E8,$9E,$7B,$DD,$6B
	.dc.b	$F0,$A5,$3F,$59,$D6,$3F,$E5,$6D,$7F,$D5,$D8,$70,$6C,$77,$FC,$89
	.dc.b	$9F,$FA,$FF,$AB,$C9,$39,$8F,$F6,$17,$E0,$7F,$C9,$7F,$4E,$FE,$19
	.dc.b	$F8,$D3,$F7,$47,$E4,$8F,$CE,$1F,$BB,$7F,$97,$FC,$3F,$E1,$D8,$70
	.dc.b	$BD,$D6,$BF,$7B,$FE,$49,$3D,$37,$FC,$20,$7F,$8D,$7C,$57,$F2,$77
	.dc.b	$96,$5D,$FC,$2F,$27,$C1,$B5,$F2,$4F,$8F,$F3,$2F,$A5,$EF,$BE,$DD
	.dc.b	$F9,$2B,$F4,$BF,$F5,$5F,$13,$F5,$8F,$CD,$FC,$D7,$F8,$8F,$0E,$17
	.dc.b	$BF,$C2,$BD,$B4,$FD,$07,$F1,$55,$AE,$2F,$F6,$E3,$DF,$F0,$EF,$F6
	.dc.b	$DB,$CB,$3E,$FF,$8C,$79,$35,$F1,$C7,$8B,$F1,$9F,$E1,$F8,$07,$E1
	.dc.b	$3F,$ED,$AF,$D3,$1F,$93,$BF,$80,$1E,$E3,$E9,$DF,$A8,$7F,$A3,$70
	.dc.b	$E1,$7B,$AD,$7E,$23,$9F,$C5,$DF,$25,$FC,$2F,$4E,$F0,$D3,$18,$F2
	.dc.b	$D3,$BF,$DA,$F9,$AF,$75,$E4,$9F,$1D,$67,$E6,$5F,$9D,$7F,$B6,$3F
	.dc.b	$11,$7E,$DC,$FB,$CF,$CC,$7D,$0F,$FA,$8E,$F7,$D9,$FE,$AE,$AD,$7A
	.dc.b	$2F,$9F,$D6,$4F,$D0,$BD,$18,$70,$DF,$F9,$39,$FD,$72,$7C,$18,$77
	.dc.b	$FC,$BB,$C9,$18,$D3,$C5,$5F,$C0,$FD,$0F,$F3,$0E,$1F,$43,$F0,$97
	.dc.b	$E4,$EF,$61,$FA,$76,$FF,$0A,$FE,$D3,$C9,$FE,$70,$AD,$7E,$4F,$9F
	.dc.b	$C0,$7F,$D2,$FA,$7F,$C0,$13,$EE,$69,$D9,$C9,$CF,$4F,$07,$E7,$3B
	.dc.b	$8F,$DA,$E5,$DB,$5F,$1F,$57,$E1,$7D,$77,$C3,$7F,$A9,$3F,$3B,$7E
	.dc.b	$D2,$FC,$2B,$F3,$DF,$24,$FC,$F7,$7C,$CC,$78,$EB,$F8,$FC,$78,$E1
	.dc.b	$C2,$FA,$FE,$7F,$BC,$FB,$67,$D7,$0F,$4E,$3C,$96,$3E,$2F,$80,$F6
	.dc.b	$C7,$13,$B5,$79,$7C,$CF,$E9,$3F,$61,$FD,$1B,$1E,$BA,$78,$77,$FC
	.dc.b	$77,$E4,$2F,$0F,$FD,$CF,$F8,$27,$F8,$6F,$F1,$47,$4B,$E6,$7F,$DD
	.dc.b	$97,$BF,$31,$7B,$A7,$EB,$5A,$C7,$FD,$E3,$C1,$F8,$8C,$3B,$7A,$1F
	.dc.b	$A3,$79,$61,$DF,$4F,$1D,$F8,$BE,$D3,$FD,$6B,$E0,$FF,$3F,$F8,$C7
	.dc.b	$EF,$AF,$D5,$BF,$87,$FD,$4B,$F2,$97,$EE,$1F,$E3,$0F,$81,$FF,$58
	.dc.b	$61,$C2,$F7,$5A,$E2,$FE,$AD,$3D,$3F,$1D,$F9,$77,$C3,$2E,$DC,$5F
	.dc.b	$8A,$8C,$63,$CB,$E3,$BC,$95,$E3,$F8,$3F,$F0,$9E,$6F,$8F,$FF,$3B
	.dc.b	$FC,$0F,$FB,$B3,$F2,$17,$EF,$9F,$EB,$AF,$C7,$DF,$A3,$BF,$14,$FF
	.dc.b	$03,$7A,$BF,$E9,$1C,$38,$5E,$EB,$5F,$4E,$FC,$A7,$5C,$5F,$C6,$F3
	.dc.b	$EB,$F8,$E7,$AB,$1E,$3F,$87,$5A,$D6,$5C,$F4,$FE,$79,$1C,$86,$1C
	.dc.b	$B5,$F2,$CF,$E4,$27,$F3,$9F,$6B,$FC,$A5,$FE,$D9,$F9,$8B,$F4,$0F
	.dc.b	$F0,$BF,$2B,$F4,$E7,$E9,$5F,$5F,$D1,$FA,$77,$BF,$55,$7B,$A7,$67
	.dc.b	$6C,$BF,$9B,$76,$B0,$ED,$1F,$C4,$A3,$1F,$88,$FF,$2D,$75,$A7,$8B
	.dc.b	$47,$D1,$FF,$2B,$9C,$E9,$E0,$97,$E3,$7F,$CB,$FC,$A7,$E5,$7F,$9F
	.dc.b	$F8,$7F,$F5,$27,$EB,$8F,$D8,$3F,$D5,$BC,$7F,$EF,$BB,$DF,$A3,$FE
	.dc.b	$05,$FC,$55,$7B,$FC,$27,$D5,$C5,$FD,$23,$3F,$F6,$23,$FC,$63,$1E
	.dc.b	$51,$FA,$88,$FE,$5B,$8F,$4C,$FC,$3F,$2F,$E5,$AF,$26,$5D,$BE,$53
	.dc.b	$FA,$37,$85,$FD,$EF,$F4,$97,$E5,$CF,$DC,$9F,$84,$BD,$97,$E5,$7E
	.dc.b	$4C,$38,$5E,$EB,$5C,$59,$FA,$6F,$F3,$87,$0F,$F4,$0B,$EC,$D8,$67
	.dc.b	$E0,$CF,$BF,$C1,$F9,$27,$39,$75,$FA,$4F,$F1,$C6,$6F,$DC,$3F,$C9
	.dc.b	$7F,$96,$FF,$7F,$F8,$17,$EB,$FF,$D6,$BD,$77,$91,$FE,$8C,$BD,$FD
	.dc.b	$1F,$F3,$1A,$FE,$67,$5A,$E0,$FE,$F9,$97,$5F,$B2,$FD,$DE,$3C,$7F
	.dc.b	$38,$B5,$AC,$B9,$E7,$FA,$2C,$79,$4F,$C8,$9C,$F0,$7F,$73,$E0,$FE
	.dc.b	$23,$F8,$9F,$F8,$13,$B3,$FC,$15,$F6,$7F,$9D,$D3,$F4,$95,$AF,$A3
	.dc.b	$FF,$24,$65,$FE,$4A,$F3,$3F,$07,$3F,$C6,$BF,$34,$FF,$02,$3D,$27
	.dc.b	$FB,$6B,$C7,$1F,$F0,$B7,$B5,$E1,$F7,$7F,$D9,$18,$F4,$F4,$7F,$DD
	.dc.b	$9F,$BA,$3F,$45,$7E,$C3,$FE,$8C,$FC,$23,$FC,$95,$F8,$BB,$F7,$5F
	.dc.b	$F7,$35,$EE,$B5,$F7,$0A,$FF,$27,$79,$9F,$ED,$9A,$D7,$07,$F9,$21
	.dc.b	$ED,$E8,$FF,$BB,$5E,$5F,$84,$7F,$AD,$30,$FE,$17,$7C,$74,$F1,$CE
	.dc.b	$7C,$AF,$ED,$9F,$81,$7F,$09,$E0,$7E,$E5,$FC,$6F,$E8,$3F,$80,$E5
	.dc.b	$FE,$02,$F1,$7F,$DF,$95,$AF,$84,$FF,$35,$F5,$7F,$99,$7E,$8D,$F9
	.dc.b	$66,$B2,$FF,$11,$F1,$7F,$97,$B0,$E0,$D8,$ED,$7F,$E6,$DF,$C7,$3C
	.dc.b	$73,$FE,$43,$AE,$DA,$78,$9F,$F4,$77,$E1,$4F,$DD,$1F,$81,$BF,$67
	.dc.b	$F0,$FC,$3F,$F3,$9F,$C2,$BF,$D4,$17,$BA,$D6,$DF,$CB,$3E,$1F,$FD
	.dc.b	$07,$E5,$7E,$3A,$BA,$7E,$E3,$F9,$4D,$91,$C5,$FE,$A1,$E3,$FF,$2F
	.dc.b	$8E,$9A,$78,$3F,$2F,$F2,$ED,$E4,$8F,$1F,$E2,$3F,$EA,$5F,$B2,$7E
	.dc.b	$E8,$FD,$7A,$7E,$77,$FE,$EA,$FA,$37,$E3,$DF,$EB,$7B,$DD,$6B,$C9
	.dc.b	$57,$F9,$F3,$87,$F4,$B5,$AE,$0F,$29,$FE,$6B,$7B,$C7,$AD,$E7,$6C
	.dc.b	$7C,$15,$DF,$2F,$24,$E7,$17,$C7,$7F,$8C,$7E,$97,$FA,$77,$3F,$FB
	.dc.b	$37,$F7,$F7,$EE,$6F,$C5,$FF,$B5,$FE,$AC,$FB,$8F,$F1,$16,$AC,$38
	.dc.b	$5E,$FE,$61,$6F,$D0,$F7,$5F,$ED,$9F,$C5,$FF,$C5,$7E,$77,$F8,$BE
	.dc.b	$3C,$18,$FF,$8C,$B5,$F1,$FE,$2E,$DB,$C7,$B7,$F7,$CF,$91,$F4,$5F
	.dc.b	$FA,$37,$0F,$0C,$FF,$80,$3E,$87,$FD,$21,$CF,$FE,$3B,$F4,$9F,$9F
	.dc.b	$FF,$BA,$87,$D6,$BF,$4A,$7E,$EE,$E1,$61,$C6,$F8,$5E,$EB,$C9,$39
	.dc.b	$F9,$DF,$F0,$DF,$0F,$F7,$7C,$BB,$FE,$35,$8D,$FF,$D4,$51,$8C,$BC
	.dc.b	$3A,$7F,$A9,$27,$BD,$F1,$9E,$F2,$FA,$FF,$5F,$AD,$FE,$03,$FC,$BF
	.dc.b	$FC,$01,$F0,$0F,$86,$7E,$5C,$FD,$01,$FB,$DB,$F5,$AF,$5E,$3C,$7E
	.dc.b	$07,$F9,$6D,$EE,$9F,$B5,$EB,$83,$FD,$71,$A7,$F8,$37,$F4,$9F,$F1
	.dc.b	$BE,$3C,$75,$B5,$AC,$B9,$E8,$FC,$7F,$CF,$39,$AF,$36,$9E,$5F,$C0
	.dc.b	$3F,$D0,$1F,$5B,$FE,$1D,$FD,$09,$F2,$87,$FD,$59,$E8,$FF,$9F,$7F
	.dc.b	$6B,$7C,$07,$F9,$FE,$F8,$5E,$F0,$FD,$6F,$8E,$BF,$F4,$EA,$AD,$E3
	.dc.b	$F9,$07,$8F,$D3,$DD,$7F,$DA,$D3,$E2,$3C,$35,$FE,$4D,$8F,$01,$AC
	.dc.b	$7C,$19,$FF,$A4,$7E,$3B,$F2,$73,$DF,$E8,$7F,$CD,$3F,$9A,$FB,$3D
	.dc.b	$8F,$F7,$17,$EB,$5F,$A8,$D7,$FC,$0B,$C4,$AD,$60,$FC,$DB,$58,$BA
	.dc.b	$3F,$4F,$E0,$FF,$90,$F1,$E3,$7F,$A7,$3F,$DD,$CF,$4C,$39,$C7,$2C
	.dc.b	$7B,$6C,$D3,$F8,$7C,$FE,$7B,$DF,$FF,$27,$7E,$7F,$F4,$BE,$9F,$FA
	.dc.b	$F7,$A8,$FD,$97,$F8,$E6,$DB,$C3,$FD,$F9,$7B,$F9,$2F,$F3,$7D,$7F
	.dc.b	$21,$D3,$54,$AF,$BB,$F8,$F7,$F3,$17,$B4,$7F,$1C,$35,$AB,$3F,$0E
	.dc.b	$1F,$E4,$C9,$E7,$5D,$F3,$F2,$76,$FF,$BE,$7E,$8D,$F8,$1B,$F6,$77
	.dc.b	$C9,$FF,$19,$F9,$1F,$E3,$1E,$A7,$F9,$1A,$F7,$5A,$F5,$BF,$B2,$FB
	.dc.b	$DF,$C3,$FB,$FF,$C2,$B5,$AD,$BF,$DE,$6F,$6F,$98,$FE,$33,$CF,$FE
	.dc.b	$75,$8E,$7A,$77,$C3,$C9,$AF,$8E,$3C,$5E,$27,$F1,$4B,$FF,$3E,$7E
	.dc.b	$F3,$FD,$77,$F6,$B7,$D9,$BF,$54,$7E,$E0,$FC,$FB,$F8,$B7,$B7,$FE
	.dc.b	$28,$C3,$85,$EE,$B5,$FB,$DF,$D0,$9E,$9C,$3F,$35,$FE,$D5,$FA,$5F
	.dc.b	$F6,$17,$8E,$5D,$BE,$03,$C9,$41,$3A,$C1,$F0,$5F,$EF,$7D,$BF,$D2
	.dc.b	$9F,$AB,$BE,$97,$FA,$5F,$03,$F1,$4F,$EB,$7F,$93,$3E,$A3,$FC,$EB
	.dc.b	$7B,$F8,$D5,$EF,$E4,$D7,$FE,$AF,$EC,$FF,$22,$DE,$F7,$FF,$7A,$E9
	.dc.b	$E0,$D1,$F9,$F7,$D2,$9F,$45,$75,$CB,$A7,$E5,$5E,$79,$CD,$F9,$B6
	.dc.b	$FE,$DF,$F9,$4F,$F1,$29,$FD,$4F,$D9,$3F,$04,$7E,$C8,$F9,$97,$E3
	.dc.b	$AF,$D7,$BF,$CF,$DC,$3F,$C3,$71,$E3,$87,$0B,$DD,$3E,$CB,$F8,$4D
	.dc.b	$7F,$13,$CB,$FD,$2A,$FF,$AF,$70,$E4,$E7,$0F,$E2,$8D,$CB,$C5,$93
	.dc.b	$F4,$EF,$44,$E7,$3F,$2F,$BB,$F6,$A1,$F8,$EF,$CC,$FC,$2F,$FD,$9F
	.dc.b	$E7,$BF,$29,$7E,$27,$FE,$B2,$F1,$BF,$CE,$77,$BF,$AA,$FE,$75,$5A
	.dc.b	$FC,$67,$F7,$73,$F5,$BF,$60,$FF,$15,$3D,$73,$FD,$D3,$C7,$2E,$D9
	.dc.b	$78,$E3,$18,$7F,$5B,$AF,$F5,$BF,$27,$E3,$FF,$44,$7E,$A1,$FE,$DD
	.dc.b	$FD,$91,$F9,$A7,$E3,$79,$FF,$D0,$F7,$BA,$D4,$BF,$30,$E7,$8B,$1F
	.dc.b	$C2,$3F,$EE,$BF,$85,$B8,$71,$F9,$D6,$7F,$0E,$B8,$DF,$8A,$73,$F8
	.dc.b	$5F,$F5,$A9,$7E,$78,$FD,$FF,$FC,$17,$FB,$33,$F6,$8F,$F1,$EF,$E2
	.dc.b	$6F,$C6,$3F,$B5,$61,$C2,$F7,$5A,$FD,$93,$38,$F3,$FA,$3F,$ED,$77
	.dc.b	$BF,$86,$B3,$9C,$FB,$7C,$A3,$F0,$0F,$24,$E7,$2F,$F2,$6E,$DF,$81
	.dc.b	$FD,$B3,$FD,$AB,$0F,$DE,$DF,$90,$3F,$0B,$7E,$D6,$FC,$B3,$F9,$BE
	.dc.b	$1C,$2F,$75,$AF,$96,$CF,$E2,$3E,$FE,$1F,$E7,$17,$FE,$E2,$FD,$2F
	.dc.b	$FC,$D8,$F1,$CF,$B7,$95,$E3,$BF,$8D,$97,$4F,$33,$F2,$DF,$4A,$1F
	.dc.b	$9E,$BF,$23,$F9,$9F,$60,$FD,$DF,$FD,$A3,$EF,$7F,$15,$DF,$86,$1C
	.dc.b	$2F,$75,$FE,$DB,$AD,$4F,$5C,$3F,$D0,$0F,$7C,$EB,$C7,$91,$97,$F7
	.dc.b	$FF,$43,$FD,$AB,$3C,$72,$F1,$ED,$FC,$97,$EF,$FF,$CC,$5A,$FF,$27
	.dc.b	$FC,$A3,$F3,$3F,$1F,$FE,$BD,$FD,$85,$FB,$57,$F8,$73,$D2,$FF,$4E
	.dc.b	$7C,$9F,$A6,$F7,$F8,$82,$F7,$F3,$CF,$D8,$1F,$54,$FD,$17,$FD,$BF
	.dc.b	$F8,$77,$F2,$0F,$A0,$FF,$36,$7E,$86,$FC,$65,$D3,$C2,$F7,$F1,$AB
	.dc.b	$DD,$7C,$EF,$C1,$FF,$9D,$F2,$7F,$3F,$F2,$5F,$AF,$BF,$78,$7E,$81
	.dc.b	$FC,$8F,$E3,$1B,$3F,$4B,$FF,$47,$7C,$BF,$D3,$C7,$8E,$1C,$3E,$31
	.dc.b	$87,$0F,$D3,$DF,$91,$BF,$0B,$FB,$5F,$1E,$FC,$97,$FA,$0F,$E6,$8F
	.dc.b	$D5,$FF,$C4,$9F,$BB,$FC,$1C,$38,$71,$FF,$55,$C3,$87,$CD,$B0,$E1
	.dc.b	$F6,$AC,$1F,$91,$BF,$2A,$FD,$F6,$3F,$EF,$1F,$C3,$BF,$D3,$5F,$B2
	.dc.b	$3E,$F7,$FD,$2D,$EB,$BF,$23,$E1,$F2,$7F,$8F,$F2,$58,$70,$BD,$FD
	.dc.b	$66,$B2,$FE,$D1,$5A,$F3,$19,$F5,$FC,$47,$E1,$D6,$CF,$14,$EB,$0F
	.dc.b	$F5,$37,$E2,$FF,$DC,$79,$F7,$5F,$43,$F1,$CF,$F5,$BF,$80,$FD,$C3
	.dc.b	$E0,$7E,$66,$ED,$F9,$5D,$DF,$A5,$F0,$FF,$DD,$3F,$55,$F5,$EF,$78
	.dc.b	$6B,$2F,$EF,$D5,$A8,$E9,$D6,$FD,$4F,$C5,$8F,$3D,$3C,$38,$FC,$FE
	.dc.b	$6C,$E5,$CA,$73,$87,$F7,$FF,$CD,$7B,$4F,$A7,$4F,$14,$FF,$80,$FE
	.dc.b	$33,$3A,$F8,$72,$FB,$5F,$8E,$7F,$A6,$F8,$7E,$0C,$E6,$D3,$FE,$0B
	.dc.b	$9F,$8B,$F9,$FB,$F8,$05,$FB,$37,$F9,$07,$F1,$7F,$F6,$CF,$ED,$1F
	.dc.b	$D1,$6F,$75,$AF,$C0,$15,$AF,$AF,$F3,$E3,$CF,$1E,$5F,$8F,$7F,$A4
	.dc.b	$38,$73,$8F,$1C,$3F,$DC,$BF,$92,$FC,$29,$61,$C7,$2E,$B8,$63,$F1
	.dc.b	$3E,$39,$74,$A7,$E0,$58,$BD,$FE,$22,$FD,$7F,$1A,$7F,$BC,$A3,$FB
	.dc.b	$BE,$FF,$84,$FA,$47,$DC,$7F,$85,$BE,$45,$F9,$73,$F4,$DF,$F0,$37
	.dc.b	$EF,$BF,$4A,$F7,$EA,$2F,$75,$FE,$0B,$FD,$CF,$58,$F3,$C7,$97,$CB
	.dc.b	$FB,$71,$E3,$87,$0F,$90,$C5,$EF,$1F,$C9,$7E,$3D,$CB,$3E,$D9,$78
	.dc.b	$FC,$87,$EB,$79,$CB,$C3,$F9,$D7,$2B,$DE,$3F,$6F,$F7,$5D,$63,$EB
	.dc.b	$ED,$F8,$3F,$BC,$7F,$41,$FD,$6F,$FD,$F7,$F2,$7F,$C5,$BF,$65,$FF
	.dc.b	$0F,$FE,$DB,$ED,$E7,$5E,$FE,$AF,$FD,$A9,$F1,$2B,$DF,$DF,$FF,$11
	.dc.b	$AD,$63,$CF,$EE,$5F,$B7,$FE,$CE,$73,$87,$19,$7E,$CD,$9F,$1F,$FC
	.dc.b	$D1,$B7,$92,$B5,$87,$F9,$F3,$5F,$1D,$6B,$5F,$1E,$BF,$E3,$6A,$FE
	.dc.b	$21,$5A,$A6,$1C,$EB,$FC,$77,$F9,$FF,$FA,$12,$B5,$39,$CD,$3F,$CB
	.dc.b	$67,$38,$F6,$97,$A3,$E7,$C5,$FB,$0B,$F2,$9F,$E4,$5F,$A1,$FF,$41
	.dc.b	$FD,$61,$F7,$0F,$CB,$FE,$36,$1F,$E8,$5C,$38,$7C,$03,$0E,$16,$FC
	.dc.b	$E3,$77,$DB,$2E,$BF,$26,$CF,$FC,$35,$3A,$CB,$9E,$3E,$1F,$B4,$F3
	.dc.b	$46,$2F,$87,$DA,$BF,$3B,$E1,$97,$4C,$BF,$C4,$9F,$B3,$63,$6F,$26
	.dc.b	$1D,$FF,$4A,$F5,$F2,$7C,$DC,$FA,$60,$9F,$F1,$64,$E7,$1E,$9F,$8E
	.dc.b	$7F,$33,$FB,$5F,$F0,$87,$E0,$1F,$E4,$FF,$99,$FE,$7B,$FB,$B3,$E3
	.dc.b	$3F,$AE,$7E,$4F,$FD,$6E,$F7,$F8,$2F,$F5,$BF,$91,$5E,$FE,$77,$FC
	.dc.b	$41,$AA,$F7,$97,$4C,$BF,$D1,$1D,$1C,$CF,$F3,$36,$5C,$F1,$F0,$6B
	.dc.b	$E5,$9C,$DF,$1C,$9E,$BF,$19,$FF,$32,$61,$C7,$1F,$EE,$5E,$C6,$77
	.dc.b	$F6,$6D,$E6,$9F,$ED,$1F,$0D,$9B,$E1,$F3,$0F,$D1,$FA,$68,$FD,$7F
	.dc.b	$F8,$8F,$13,$F1,$4F,$F3,$97,$EC,$2F,$CD,$37,$5A,$9C,$FE,$67,$8F
	.dc.b	$53,$FC,$05,$1F,$E7,$9F,$3F,$D9,$C7,$FE,$07,$BE,$1C,$5F,$E3,$0C
	.dc.b	$FE,$44,$FF,$B5,$F4,$F1,$63,$CA,$9F,$B9,$BE,$D3,$B6,$1E,$59,$CF
	.dc.b	$ED,$98,$D3,$FD,$11,$FB,$87,$F1,$6A,$EB,$9F,$87,$2E,$92,$F8,$0F
	.dc.b	$99,$F8,$47,$FB,$13,$F0,$67,$EB,$5F,$E2,$6F,$CC,$5F,$80,$3F,$33
	.dc.b	$DF,$BB,$FE,$6F,$AD,$4F,$F5,$2D,$7B,$E3,$CE,$79,$7A,$1F,$0A,$31
	.dc.b	$F3,$0C,$9E,$F6,$72,$E9,$39,$C9,$F9,$F7,$83,$5F,$2C,$E7,$F2,$6C
	.dc.b	$69,$E7,$A7,$EE,$B8,$AE,$D1,$FE,$43,$DB,$FD,$01,$C6,$FD,$2B,$FD
	.dc.b	$B9,$F0,$EF,$D5,$3F,$CB,$1F,$AE,$FF,$4D,$FC,$23,$FC,$79,$F8,$4F
	.dc.b	$F9,$DE,$F7,$5A,$F8,$DF,$F1,$B6,$5D,$27,$3F,$81,$3F,$38,$FF,$3F
	.dc.b	$4E,$7F,$17,$FF,$79,$E1,$DB,$AF,$FC,$4D,$97,$59,$CE,$0D,$3F,$D1
	.dc.b	$9D,$9E,$2D,$C7,$A6,$B7,$3F,$E9,$1F,$17,$53,$9E,$0F,$66,$1C,$FD
	.dc.b	$57,$E6,$1A,$E0,$F3,$E2,$97,$E7,$BF,$EF,$FF,$D0,$3F,$D2,$DF,$A1
	.dc.b	$3F,$76,$7C,$23,$F2,$A7,$EF,$AF,$D7,$5F,$B6,$BE,$8F,$97,$17,$A3
	.dc.b	$2E,$58,$F1,$C3,$85,$FF,$82,$B9,$1A,$78,$34,$EF,$A7,$6D,$3F,$9F
	.dc.b	$E7,$D0,$67,$D2,$75,$F5,$DF,$1B,$1E,$75,$E1,$C3,$9E,$7E,$7F,$AB
	.dc.b	$7E,$1D,$CF,$2E,$53,$9F,$3F,$17,$C3,$46,$0F,$CD,$B3,$C1,$EB,$9C
	.dc.b	$FC,$1B,$57,$ED,$5F,$EB,$FF,$43,$FA,$9F,$98,$FD,$3F,$E4,$FE,$4E
	.dc.b	$FA,$FF,$E9,$98,$70,$E3,$FB,$F8,$70,$BD,$D3,$F6,$9E,$D9,$75,$F9
	.dc.b	$46,$7D,$78,$DA,$7D,$8C,$F6,$74,$AD,$E5,$E2,$FB,$7F,$91,$8F,$3C
	.dc.b	$79,$61,$E5,$AE,$38,$7F,$03,$C3,$8E,$3D,$B3,$7E,$FB,$AD,$7D,$13
	.dc.b	$E2,$F3,$9F,$AA,$E7,$0E,$39,$FF,$AD,$3E,$9B,$AF,$E8,$DF,$E6,$3F
	.dc.b	$D1,$9F,$9E,$7F,$A8,$BE,$13,$FD,$49,$EC,$6B,$E9,$D5,$AF,$C2,$15
	.dc.b	$FE,$A9,$FC,$8F,$53,$98,$FF,$04,$F1,$E3,$1E,$9F,$9B,$FC,$08,$FD
	.dc.b	$6E,$7E,$19,$F0,$7E,$C9,$FE,$C8,$EB,$FA,$94,$AD,$67,$E0,$A7,$EC
	.dc.b	$FD,$63,$1F,$9A,$DC,$D4,$FC,$CF,$E5,$FE,$BC,$FC,$91,$FB,$F3,$F7
	.dc.b	$B7,$E8,$BF,$E1,$EF,$DB,$DF,$3D,$C2,$F7,$5A,$FA,$35,$3F,$6C,$D6
	.dc.b	$3C,$F1,$FF,$13,$F9,$7E,$FE,$36,$9E,$1A,$D6,$2F,$C3,$F5,$39,$CF
	.dc.b	$B6,$6F,$C1,$7A,$C6,$3F,$03,$BA,$7F,$8D,$3E,$D9,$FA,$0E,$33,$FF
	.dc.b	$79,$FC,$CF,$E9,$7E,$EF,$FD,$A9,$FB,$D7,$F3,$8D,$3F,$87,$7D,$9F
	.dc.b	$F5,$FF,$CF,$F5,$C5,$FE,$D4,$AD,$7A,$FF,$E4,$2F,$9C,$FF,$79,$F8
	.dc.b	$1F,$E8,$7A,$D7,$C9,$7C,$2A,$FF,$23,$CF,$F0,$3C,$FB,$CE,$72,$FE
	.dc.b	$0B,$3C,$6D,$F8,$A3,$2E,$F4,$F6,$DF,$9C,$F9,$2F,$75,$A9,$7C,$F6
	.dc.b	$70,$E5,$97,$79,$7E,$49,$9B,$E1,$F7,$3C,$FF,$27,$7C,$BB,$F8,$07
	.dc.b	$FA,$A7,$C5,$3F,$2D,$FF,$7E,$75,$6E,$B5,$F2,$DF,$CE,$27,$FE,$00
	.dc.b	$74,$67,$0E,$30,$FC,$73,$E1,$71,$7F,$4A,$CF,$BD,$6B,$E9,$33,$29
	.dc.b	$CE,$3D,$7F,$31,$7E,$B9,$D6,$31,$FB,$CD,$DF,$CD,$F8,$5F,$2C,$38
	.dc.b	$C7,$A7,$07,$EE,$7B,$D9,$0F,$D2,$3F,$94,$7E,$D7,$FD,$7B,$57,$E0
	.dc.b	$9F,$CB,$FF,$6E,$7E,$82,$FC,$F1,$F9,$7B,$F7,$2F,$E5,$5E,$1B,$2E
	.dc.b	$58,$F1,$C3,$85,$EF,$F5,$CE,$F9,$FF,$6C,$AD,$65,$D3,$E8,$FF,$A5
	.dc.b	$63,$CB,$D4,$6B,$E2,$AF,$0F,$E1,$DA,$C7,$95,$6B,$4E,$FB,$7D,$1D
	.dc.b	$BC,$95,$AC,$79,$6D,$E9,$E0,$F1,$CF,$C0,$C7,$96,$7E,$EF,$D5,$DF
	.dc.b	$B0,$EB,$83,$CF,$8A,$5F,$A6,$F2,$F8,$AF,$CC,$7A,$DF,$E1,$FF,$CC
	.dc.b	$7E,$37,$C3,$3F,$3C,$D7,$95,$FB,$BF,$A6,$56,$BF,$16,$CE,$1C,$79
	.dc.b	$58,$FB,$F8,$BF,$A6,$E3,$C7,$0F,$A5,$E8,$E7,$4F,$0E,$7D,$31,$E3
	.dc.b	$FA,$87,$48,$C6,$5F,$84,$E7,$F8,$B9,$F4,$D3,$C1,$3F,$C7,$27,$38
	.dc.b	$F4,$CF,$D7,$FB,$67,$C8,$F3,$FF,$51,$D6,$FD,$49,$FB,$A7,$FB,$33
	.dc.b	$CE,$FF,$2D,$65,$DA,$F7,$5A,$FC,$5B,$F6,$3F,$9F,$73,$F3,$BF,$DD
	.dc.b	$33,$9E,$4F,$81,$7C,$32,$FF,$6B,$EF,$8B,$DC,$77,$F8,$4E,$5A,$F8
	.dc.b	$E3,$C5,$FB,$9E,$32,$F0,$6C,$FD,$FF,$15,$E7,$FC,$03,$FC,$15,$FA
	.dc.b	$57,$E8,$FF,$21,$FC,$8F,$DC,$FF,$03,$FC,$B9,$FB,$BB,$F3,$A7,$E9
	.dc.b	$7C,$3B,$3F,$C8,$F8,$70,$F4,$58,$70,$B7,$E2,$5B,$B5,$F6,$7F,$64
	.dc.b	$7F,$8A,$F3,$F1,$79,$AF,$C3,$B3,$39,$3B,$C3,$5F,$36,$BF,$EA,$B9
	.dc.b	$ED,$87,$8F,$3F,$4F,$E6,$5F,$E3,$7C,$3F,$80,$7A,$BE,$2F,$C0,$9F
	.dc.b	$B4,$3F,$63,$FF,$68,$FE,$83,$FE,$61,$BD,$E4,$BD,$D6,$BF,$74,$FF
	.dc.b	$19,$56,$A7,$CF,$A6,$0F,$F3,$5E,$5F,$03,$EB,$F3,$E2,$7E,$FA,$73
	.dc.b	$97,$5D,$3C,$5F,$67,$FC,$AE,$79,$1F,$C7,$27,$5A,$7F,$B9,$BE,$0F
	.dc.b	$FC,$A5,$FB,$87,$F1,$BF,$D8,$9F,$87,$FF,$86,$3F,$4E,$FF,$56,$EB
	.dc.b	$FE,$43,$FD,$93,$FC,$91,$7B,$AD,$7E,$1F,$FE,$36,$FB,$9C,$EB,$FE
	.dc.b	$82,$CB,$9F,$BF,$E8,$C3,$87,$D5,$DA,$D6,$3C,$B3,$7C,$1F,$92,$73
	.dc.b	$4D,$7C,$7E,$4F,$F9,$A7,$E9,$7F,$99,$76,$3E,$A3,$F8,$47,$EC,$AF
	.dc.b	$CA,$1F,$AC,$7F,$3C,$FC,$DB,$FC,$C1,$87,$0B,$DD,$6B,$B1,$FB,$4E
	.dc.b	$B6,$FF,$4D,$E7,$D7,$E0,$7E,$8E,$3C,$7E,$06,$D6,$B2,$E7,$A7,$F4
	.dc.b	$78,$E4,$7E,$72,$33,$FA,$4F,$A6,$38,$60,$DB,$D1,$7F,$D4,$BF,$0E
	.dc.b	$E3,$9F,$86,$73,$AB,$E8,$F3,$9F,$82,$3F,$D3,$FE,$17,$F8,$EB,$3E
	.dc.b	$B7,$C3,$D9,$F1,$FC,$8F,$F7,$9F,$B1,$7E,$80,$FC,$7F,$FC,$CB,$F9
	.dc.b	$0B,$F4,$3D,$FB,$DF,$E4,$BA,$D7,$E0,$0A,$FD,$0E,$BD,$F5,$FE,$E9
	.dc.b	$7F,$E5,$17,$F6,$31,$9C,$FF,$C1,$1F,$9C,$7C,$89,$DE,$9E,$2D,$3C
	.dc.b	$31,$E0,$8C,$64,$FB,$9F,$91,$3C,$34,$F1,$E8,$FD,$DF,$84,$63,$E6
	.dc.b	$FE,$8F,$EF,$8E,$FF,$F6,$AF,$E3,$EF,$DD,$7F,$D2,$DF,$81,$7F,$61
	.dc.b	$F0,$3F,$BE,$5E,$F4,$5E,$FC,$CF,$F4,$2F,$37,$4C,$BF,$6F,$57,$F2
	.dc.b	$8F,$66,$1C,$70,$E1,$F8,$17,$A2,$31,$A7,$FA,$97,$E0,$F1,$8F,$92
	.dc.b	$3C,$7C,$1D,$72,$F0,$C7,$9B,$F6,$3F,$C5,$FD,$BB,$F5,$97,$E4,$1F
	.dc.b	$EB,$7F,$C0,$BF,$D7,$9F,$15,$FE,$77,$EE,$FA,$97,$BA,$D4,$7F,$3B
	.dc.b	$FA,$8E,$78,$72,$CF,$EB,$F2,$FF,$87,$70,$ED,$E1,$7D,$77,$97,$E1
	.dc.b	$7F,$E8,$AC,$5C,$3F,$52,$73,$AF,$FB,$37,$8F,$19,$F9,$B0,$E1,$B3
	.dc.b	$F7,$1F,$0E,$FE,$48,$7E,$B5,$8A,$DF,$E5,$B7,$2F,$6F,$E9,$7F,$ED
	.dc.b	$9F,$DF,$3F,$C5,$3E,$27,$FB,$83,$E1,$1F,$AC,$3F,$46,$7E,$68,$FD
	.dc.b	$9B,$87,$47,$F8,$AF,$0E,$17,$BA,$7E,$CF,$AA,$FF,$1E,$D3,$F2,$0F
	.dc.b	$E6,$39,$F8,$3F,$1B,$B5,$CD,$FE,$7D,$8F,$87,$F3,$CF,$F7,$F5,$6F
	.dc.b	$5F,$2E,$BF,$EA,$6C,$3B,$7D,$6F,$EC,$71,$E3,$3E,$8F,$C5,$B3,$A7
	.dc.b	$9A,$3D,$7E,$9F,$97,$F4,$A1,$FE,$52,$FD,$37,$FC,$5B,$FC,$00,$FB
	.dc.b	$F7,$EF,$7B,$AD,$7D,$8A,$B5,$FB,$76,$7F,$76,$F5,$F0,$74,$BF,$F2
	.dc.b	$EB,$FB,$AE,$2F,$F3,$04,$78,$38,$5A,$78,$3F,$25,$F7,$D8,$E5,$3A
	.dc.b	$CF,$C1,$F7,$BA,$9C,$DF,$A3,$67,$C1,$79,$F0,$E1,$B7,$AF,$E0,$7E
	.dc.b	$A7,$3F,$F8,$33,$07,$E3,$1F,$E0,$DF,$BD,$FF,$3B,$FE,$75,$AF,$A0
	.dc.b	$F2,$CE,$7E,$51,$F9,$D6,$70,$F7,$E3,$FE,$16,$FD,$7A,$E1,$C3,$E9
	.dc.b	$FF,$CE,$DF,$82,$F9,$EB,$C1,$97,$7F,$80,$C5,$EF,$5F,$F7,$FF,$E1
	.dc.b	$7D,$6B,$C9,$F2,$AF,$D2,$71,$9F,$FA,$06,$1F,$49,$8C,$10,$FB,$AF
	.dc.b	$C9,$7B,$1F,$EB,$0F,$DC,$B9,$7E,$5D,$FC,$8F,$F4,$8F,$CE,$4B,$C1
	.dc.b	$FF,$4A,$3B,$D1,$5B,$FD,$F7,$9F,$D9,$BF,$3D,$FB,$F7,$D7,$C0,$FB
	.dc.b	$C7,$D9,$79,$5F,$F3,$1F,$B8,$FB,$11,$8F,$8F,$7D,$CE,$31,$6D,$E6
	.dc.b	$96,$DF,$E0,$CB,$F3,$65,$D6,$F7,$AF,$9A,$F7,$E1,$FE,$C7,$CC,$7E
	.dc.b	$91,$FE,$D0,$FC,$37,$FD,$5F,$EA,$3E,$F1,$F9,$7E,$EB,$5E,$47,$ED
	.dc.b	$E5,$FA,$E6,$7F,$64,$FC,$EE,$5D,$36,$B8,$70,$C9,$F9,$87,$E5,$6B
	.dc.b	$5A,$78,$3C,$3C,$5E,$F3,$7E,$D7,$E1,$CB,$FD,$A3,$AF,$87,$D4,$FE
	.dc.b	$15,$D6,$FC,$C7,$F9,$07,$86,$F9,$4F,$F7,$DF,$E5,$CF,$CF,$37,$5A
	.dc.b	$9C,$FE,$93,$8F,$8A,$FD,$AE,$3C,$FF,$37,$B7,$BF,$C7,$3D,$3D,$BB
	.dc.b	$D3,$B4,$38,$FF,$61,$87,$8F,$1F,$D4,$E7,$E2,$FC,$A3,$83,$E4,$FC
	.dc.b	$AA,$E0,$DD,$F8,$6E,$1F,$F3,$EF,$83,$F7,$35,$7C,$07,$BF,$F1,$EF
	.dc.b	$CD,$FC,$83,$F0,$CF,$F7,$6F,$EB,$5F,$CB,$6B,$5E,$7F,$F5,$A9,$FF
	.dc.b	$78,$FE,$99,$9E,$56,$3F,$C5,$2F,$E5,$7D,$3D,$9C,$E1,$C7,$E0,$7F
	.dc.b	$86,$57,$4C,$FC,$1F,$61,$8B,$DC,$78,$FE,$B1,$FB,$37,$7C,$FC,$FA
	.dc.b	$A7,$FD,$31,$C1,$ED,$FC,$E1,$FB,$23,$F5,$17,$EF,$2F,$96,$7D,$53
	.dc.b	$F2,$C7,$EC,$7F,$D2,$F1,$E3,$87,$0B,$DD,$3F,$7B,$D6,$5D,$7A,$7F
	.dc.b	$66,$30,$F0,$FE,$6D,$FC,$D9,$E7,$A7,$87,$4F,$E9,$9A,$F9,$F0,$F3
	.dc.b	$4E,$63,$F9,$85,$70,$CF,$C3,$EC,$78,$35,$F2,$63,$E9,$F7,$98,$7E
	.dc.b	$EB,$4F,$E7,$9F,$30,$FD,$21,$FA,$DF,$FA,$D3,$F5,$7F,$DB,$FE,$F3
	.dc.b	$F9,$FC,$38,$61,$FD,$D2,$F7,$F8,$62,$F7,$F6,$3D,$B6,$FE,$91,$7B
	.dc.b	$FB,$CF,$F4,$FC,$63,$4F,$06,$98,$AF,$1E,$5D,$31,$FE,$0F,$F6,$FF
	.dc.b	$C6,$70,$E1,$39,$F6,$3F,$C9,$57,$DC,$E3,$87,$A6,$1B,$3F,$36,$D5
	.dc.b	$7A,$6B,$52,$FC,$F3,$3A,$79,$37,$F6,$63,$CB,$C7,$CE,$7E,$1C,$1E
	.dc.b	$47,$4F,$93,$FE,$89,$FB,$73,$F1,$9F,$0F,$5F,$F2,$AF,$D5,$F5,$F1
	.dc.b	$0A,$D6,$3F,$90,$F8,$9F,$F0,$EF,$4F,$CE,$CB,$9F,$1F,$F9,$C3,$97
	.dc.b	$E5,$C7,$F5,$1F,$A3,$F9,$13,$FD,$E3,$2E,$9A,$F8,$3F,$3A,$F9,$13
	.dc.b	$FD,$E6,$FC,$B3,$E4,$FC,$AF,$33,$C7,$5F,$1E,$FF,$D0,$34,$F3,$C6
	.dc.b	$35,$FE,$0F,$1F,$82,$F4,$9F,$2E,$FD,$83,$FC,$B5,$F0,$5F,$C3,$72
	.dc.b	$7E,$EF,$F4,$2F,$75,$A7,$FC,$18,$E6,$F9,$57,$BC,$FF,$08,$4F,$6A
	.dc.b	$F7,$BC,$BE,$C9,$87,$7A,$E5,$87,$5F,$92,$F6,$DE,$EB,$C9,$F5,$DA
	.dc.b	$C3,$8C,$FF,$87,$7F,$42,$FC,$2F,$EC,$AF,$CC,$BF,$D2,$DF,$BE,$3F
	.dc.b	$3B,$7E,$46,$FC,$45,$7C,$2F,$75,$A9,$FE,$11,$C5,$FA,$8A,$E9,$E2
	.dc.b	$FD,$87,$B7,$CC,$DC,$38,$E3,$D7,$8B,$A4,$B3,$F1,$4F,$FB,$B2,$73
	.dc.b	$18,$C9,$FA,$4E,$34,$F1,$7C,$C5,$C3,$D1,$E6,$DB,$7D,$87,$E9,$3E
	.dc.b	$81,$F9,$0F,$D8,$FD,$9F,$FA,$77,$DE,$3F,$40,$FF,$57,$72,$7F,$02
	.dc.b	$BD,$F6,$2F,$74,$FD,$DF,$A6,$3F,$E4,$CE,$2F,$B8,$30,$E2,$E1,$E7
	.dc.b	$7E,$0E,$CF,$CA,$C6,$3E,$49,$A7,$F8,$5A,$FA,$E7,$E1,$9C,$E7,$EA
	.dc.b	$D3,$CD,$9F,$AE,$73,$FA,$F7,$F0,$2F,$84,$7E,$2A,$F9,$A7,$E8,$6F
	.dc.b	$C4,$1F,$9D,$7F,$A4,$7F,$7C,$E1,$7B,$AD,$4F,$FA,$57,$D4,$CE,$1F
	.dc.b	$B4,$C7,$AF,$D2,$DC,$38,$65,$D3,$E8,$FF,$63,$5E,$2B,$F0,$CF,$83
	.dc.b	$BB,$37,$C3,$2F,$F5,$17,$D6,$F6,$D3,$C3,$97,$F6,$0E,$1F,$E6,$BF
	.dc.b	$B3,$7F,$A1,$FF,$80,$3F,$C8,$9F,$B5,$BF,$22,$7E,$F2,$FC,$43,$F9
	.dc.b	$1D,$EF,$EC,$57,$BA,$7E,$FF,$AF,$B0,$65,$FC,$43,$0F,$F4,$A7,$F0
	.dc.b	$05,$9C,$E3,$CB,$3F,$CD,$69,$E4,$9C,$E7,$E3,$FC,$8F,$F8,$8E,$3E
	.dc.b	$18,$F0,$7E,$DF,$DF,$3F,$4E,$3E,$8F,$D3,$31,$9F,$5D,$BF,$D9,$BA
	.dc.b	$75,$BD,$C7,$FB,$4E,$3F,$AB,$FE,$DE,$FD,$A1,$FB,$47,$FB,$BB,$F7
	.dc.b	$67,$C0,$3F,$57,$7E,$C7,$FD,$1B,$0E,$17,$BA,$D7,$EA,$99,$C7,$9F
	.dc.b	$93,$F9,$CB,$ED,$F1,$BF,$CD,$1E,$3F,$46,$D3,$BE,$9F,$CC,$27,$31
	.dc.b	$E2,$F8,$7C,$63,$E6,$BE,$19,$F9,$7F,$1A,$F0,$6B,$E3,$C9,$5F,$D7
	.dc.b	$3E,$53,$F8,$87,$AB,$FE,$FA,$FD,$B5,$AB,$F1,$FD,$7B,$3F,$C6,$6B
	.dc.b	$53,$FC,$EB,$93,$FB,$EE,$1F,$C6,$78,$BF,$6E,$FF,$20,$F9,$CF,$E9
	.dc.b	$EF,$6F,$C7,$19,$BE,$AF,$FA,$5C,$E6,$1F,$55,$8D,$3C,$17,$C3,$37
	.dc.b	$E4,$9E,$4D,$7F,$96,$4E,$7F,$6B,$FF,$0B,$FA,$FF,$E0,$AF,$35,$F8
	.dc.b	$D7,$F3,$2F,$05,$FB,$CF,$F8,$AF,$F3,$97,$EC,$5C,$39,$3F,$DB,$18
	.dc.b	$70,$DD,$87,$0F,$A7,$61,$FE,$BB,$E6,$FE,$3D,$87,$0C,$FF,$9F,$BF
	.dc.b	$E1,$A3,$FD,$55,$A7,$8E,$1F,$9F,$E6,$72,$71,$CB,$BE,$BF,$E3,$FE
	.dc.b	$DC,$CF,$F1,$0B,$F5,$4F,$F9,$0F,$E6,$33,$B7,$9F,$6F,$36,$AF,$C3
	.dc.b	$BD,$98,$F1,$CF,$AE,$FE,$AF,$D5,$73,$8F,$1C,$FE,$27,$C8,$63,$5E
	.dc.b	$F7,$F0,$3F,$5B,$E7,$EC,$7F,$98,$7E,$D9,$F5,$DF,$DF,$3C,$9F,$E9
	.dc.b	$79,$6B,$C1,$FF,$6D,$CE,$7D,$B7,$E7,$D9,$C3,$E1,$70,$EF,$F1,$3B
	.dc.b	$D5,$F3,$23,$F9,$5E,$3C,$BE,$6D,$A7,$F5,$89,$E5,$9F,$86,$FE,$3E
	.dc.b	$BE,$58,$F0,$7B,$7A,$C3,$8E,$BF,$9D,$BE,$DF,$76,$F9,$47,$EC,$8F
	.dc.b	$88,$FF,$2B,$7E,$91,$FE,$7E,$F4,$3E,$5D,$EE,$B5,$F5,$CA,$FF,$41
	.dc.b	$F9,$FA,$C7,$9F,$07,$DF,$CF,$BC,$3F,$64,$F7,$F8,$BF,$E6,$0C,$FA
	.dc.b	$6B,$E0,$FA,$AF,$C0,$C6,$34,$ED,$E2,$E3,$1F,$3D,$F9,$BE,$93,$FD
	.dc.b	$2F,$8F,$39,$CD,$F9,$65,$F9,$F7,$E1,$3F,$0F,$FD,$5F,$E6,$6F,$DB
	.dc.b	$DC,$9F,$C4,$7F,$55,$FF,$51,$70,$FD,$3B,$DD,$6B,$DC,$57,$EE,$EB
	.dc.b	$5E,$97,$F1,$9E,$4F,$81,$BF,$F9,$F3,$1E,$F1,$80,$7F,$BE,$A5,$9F
	.dc.b	$8F,$3F,$E3,$73,$E7,$9C,$E7,$E1,$F3,$7F,$4B,$87,$92,$FD,$1F,$8E
	.dc.b	$EF,$1E,$5A,$F9,$B8,$31,$8F,$A7,$67,$BD,$8C,$7E,$07,$EF,$BF,$A8
	.dc.b	$FB,$57,$C7,$FF,$BA,$4F,$F5,$F7,$EB,$DF,$C9,$BE,$EB,$51,$9F,$BB
	.dc.b	$4F,$F2,$AF,$DB,$73,$F3,$5F,$7B,$97,$DF,$8F,$4F,$D8,$31,$1F,$EA
	.dc.b	$AC,$79,$61,$FC,$96,$38,$C7,$83,$DF,$FE,$07,$7E,$4C,$3C,$B9,$78
	.dc.b	$3D,$FF,$F0,$AF,$57,$EA,$6B,$58,$7A,$25,$F9,$E7,$F0,$4F,$BF,$FF
	.dc.b	$16,$7E,$99,$FD,$03,$F6,$07,$E9,$4F,$BF,$FE,$BB,$9F,$EC,$EF,$78
	.dc.b	$AF,$75,$FA,$9A,$D7,$D2,$FF,$B1,$78,$3E,$F1,$FE,$E8,$59,$7C,$0C
	.dc.b	$79,$E0,$FD,$2F,$FD,$83,$87,$2A,$F2,$67,$FC,$DB,$1F,$46,$3C,$B3
	.dc.b	$F1,$4F,$9F,$EA,$73,$7C,$34,$FF,$28,$F8,$7F,$EC,$CF,$C0,$BF,$DF
	.dc.b	$1F,$28,$FC,$6D,$FA,$92,$5F,$AC,$BF,$53,$FE,$53,$1C,$2B,$6E,$BE
	.dc.b	$8F,$3E,$E3,$E8,$7E,$DC,$F6,$E2,$F8,$2F,$1C,$BB,$7C,$AF,$F3,$C8
	.dc.b	$C1,$E3,$9E,$7A,$79,$6B,$C9,$A7,$F9,$6B,$93,$F9,$25,$F0,$87,$C8
	.dc.b	$70,$D7,$CB,$0F,$D8,$B1,$87,$2F,$6B,$FA,$1F,$E8,$FF,$E1,$8F,$5B
	.dc.b	$FC,$37,$FA,$9B,$F5,$7F,$83,$D1,$AF,$9D,$FF,$0F,$D2,$B5,$9F,$CC
	.dc.b	$E1,$FD,$CE,$3C,$DF,$BE,$FF,$89,$3F,$50,$FC,$EB,$DE,$F9,$65,$FE
	.dc.b	$29,$AE,$D1,$E2,$8F,$85,$C3,$FD,$06,$B8,$63,$E0,$F3,$37,$7C,$B2
	.dc.b	$7E,$F4,$C1,$F6,$7F,$C3,$33,$79,$8F,$DB,$DF,$A2,$FF,$18,$F1,$1F
	.dc.b	$84,$BF,$3E,$7D,$27,$3C,$B9,$63,$C7,$0E,$17,$BF,$98,$D6,$7F,$C6
	.dc.b	$AB,$4F,$F3,$FB,$E3,$9B,$F5,$3F,$EB,$B1,$8C,$79,$F6,$7F,$22,$DB
	.dc.b	$D3,$87,$3C,$3C,$FE,$BF,$E9,$74,$7E,$81,$BC,$B9,$4E,$77,$FF,$2A
	.dc.b	$63,$CB,$1F,$83,$8F,$2E,$0F,$7D,$3F,$7B,$F4,$D3,$F3,$97,$E4,$1F
	.dc.b	$E7,$FF,$CC,$1F,$A4,$BF,$4C,$7E,$F8,$F8,$1E,$38,$70,$F9,$36,$1C
	.dc.b	$2F,$7D,$FA,$CB,$FC,$99,$9F,$C5,$3F,$A0,$E1,$C5,$C1,$E3,$8F,$16
	.dc.b	$7E,$1F,$89,$F8,$13,$C2,$73,$83,$F3,$6C,$CF,$1C,$FC,$78,$7A,$75
	.dc.b	$7C,$17,$8E,$73,$87,$B3,$E0,$F1,$B7,$A6,$FF,$B1,$F2,$7C,$BD,$BD
	.dc.b	$D3,$9E,$FF,$93,$FA,$17,$FB,$13,$EA,$9F,$A7,$BF,$5E,$7E,$7E,$FC
	.dc.b	$47,$FD,$47,$9F,$D5,$C3,$85,$EE,$B5,$EA,$F4,$FA,$6F,$EC,$7C,$9F
	.dc.b	$12,$FB,$36,$3F,$CD,$77,$C7,$D3,$F8,$35,$D3,$3F,$0F,$AD,$8B,$DC
	.dc.b	$79,$34,$7D,$D7,$7D,$BD,$33,$9F,$BE,$FC,$2F,$E5,$1F,$E7,$0F,$AC
	.dc.b	$7E,$C6,$FC,$C7,$FC,$F5,$F8,$13,$F7,$6D,$FC,$CF,$F9,$9A,$B5,$F6
	.dc.b	$FF,$EE,$7F,$DC,$33,$FB,$1F,$E8,$B9,$BF,$C0,$3A,$75,$7F,$02,$F7
	.dc.b	$BE,$3F,$9D,$7D,$C8,$C6,$9F,$E8,$9F,$3B,$19,$F7,$8F,$1F,$E9,$5E
	.dc.b	$2D,$FC,$F1,$E6,$E0,$FC,$FF,$E0,$FF,$E7,$2F,$C1,$DF,$0E,$FD,$47
	.dc.b	$FB,$CF,$E9,$2F,$DD,$77,$F7,$9F,$ED,$CA,$D7,$CE,$29,$F8,$FE,$B1
	.dc.b	$F6,$72,$33,$FF,$3F,$F9,$98,$7C,$18,$F7,$FC,$13,$FA,$FF,$1F,$F1
	.dc.b	$79,$CE,$DF,$E0,$AF,$BF,$FC,$DC,$63,$5F,$07,$5F,$7C,$7C,$7B,$3F
	.dc.b	$71,$C6,$3E,$AE,$5B,$87,$9F,$8F,$F7,$FF,$94,$3F,$79,$7E,$67,$F9
	.dc.b	$7F,$4B,$FC,$47,$F9,$D7,$F4,$CF,$3B,$F8,$55,$EE,$B5,$E9,$AB,$FA
	.dc.b	$A7,$1F,$F2,$FA,$D7,$37,$EE,$9E,$DF,$88,$7C,$47,$97,$8D,$FC,$2F
	.dc.b	$0F,$E3,$75,$E4,$D3,$C7,$39,$F8,$3F,$DD,$DF,$0D,$7C,$BF,$47,$BD
	.dc.b	$7C,$7A,$FF,$5F,$DB,$F9,$07,$E5,$1F,$E1,$9F,$D2,$5F,$51,$FD,$D7
	.dc.b	$F6,$D7,$E9,$0F,$C9,$FF,$90,$DE,$EB,$53,$FD,$23,$0E,$7F,$40,$FD
	.dc.b	$9B,$EE,$DF,$5E,$3F,$E2,$EF,$1F,$65,$9F,$F3,$AC,$38,$E9,$E2,$9C
	.dc.b	$DF,$F1,$6B,$E1,$7E,$4F,$DD,$BC,$79,$30,$E1,$3E,$8F,$DA,$7F,$9B
	.dc.b	$FE,$10,$FD,$69,$C6,$FD,$F3,$FB,$D7,$E9,$AF,$D5,$DF,$6D,$C2,$F7
	.dc.b	$5A,$9F,$F3,$7F,$0F,$F4,$7C,$7A,$63,$F9,$A7,$B7,$D2,$5C,$38,$FC
	.dc.b	$FB,$3F,$B3,$87,$19,$CC,$3F,$12,$FE,$67,$87,$4C,$FC,$1E,$AF,$8F
	.dc.b	$3F,$46,$5C,$F4,$79,$D9,$BF,$34,$59,$D7,$E2,$5F,$DF,$FA,$7F,$BF
	.dc.b	$7E,$97,$FE,$1F,$FB,$67,$EF,$FF,$EB,$BF,$BB,$7C,$A7,$F9,$37,$8F
	.dc.b	$F0,$3D,$9F,$E9,$0C,$38,$5E,$E9,$FB,$77,$5A,$D7,$33,$2F,$F6,$AE
	.dc.b	$7E,$08,$FC,$6C,$FF,$AF,$63,$13,$E2,$FC,$07,$FD,$37,$F7,$7F,$DF
	.dc.b	$67,$39,$79,$67,$17,$DA,$F8,$6B,$E8,$FB,$FF,$14,$B5,$F3,$DF,$FA
	.dc.b	$9B,$57,$ED,$5A,$FA,$67,$E3,$FF,$E5,$DF,$D2,$BF,$C4,$9E,$AB,$F8
	.dc.b	$05,$FA,$5E,$EB,$5C,$AA,$D7,$C2,$67,$E3,$FD,$FB,$FF,$AD,$B3,$EB
	.dc.b	$F2,$78,$8F,$F1,$3E,$3C,$B0,$FE,$41,$1D,$65,$A7,$8B,$F0,$DF,$7C
	.dc.b	$F0,$9C,$E7,$E1,$D7,$FC,$1C,$FD,$69,$D5,$FA,$A5,$FC,$01,$F1,$3F
	.dc.b	$7A,$7E,$0E,$FD,$39,$E7,$BF,$6F,$7E,$CA,$B7,$EB,$9C,$70,$E1,$7B
	.dc.b	$AD,$7E,$BB,$9F,$89,$F7,$61,$DB,$2B,$87,$0C,$BA,$F0,$7B,$2B,$59
	.dc.b	$73,$EE,$C5,$EF,$3F,$17,$E0,$FE,$1C,$BC,$F9,$79,$B1,$7E,$C7,$9C
	.dc.b	$BC,$90,$F8,$3C,$56,$F1,$FC,$2F,$E9,$0F,$C9,$FF,$7F,$E0,$3F,$64
	.dc.b	$7E,$71,$F8,$1F,$8D,$DD,$6A,$73,$CF,$FE,$BF,$EA,$FA,$18,$F3,$F8
	.dc.b	$6F,$BC,$F5,$F6,$2E,$1C,$72,$FB,$F7,$C2,$31,$DB,$F7,$AF,$86,$9E
	.dc.b	$3B,$FC,$97,$EB,$BE,$2C,$53,$9C,$DF,$23,$FA,$0F,$A4,$7E,$74,$FE
	.dc.b	$00,$7E,$FC,$FD,$39,$F9,$23,$F6,$6F,$F5,$CF,$17,$E2,$71,$E3,$87
	.dc.b	$0B,$DD,$3F,$5A,$ED,$97,$5F,$1F,$F7,$40,$FF,$1D,$AF,$D2,$3F,$41
	.dc.b	$8F,$4D,$3C,$5A,$7F,$94,$6F,$9E,$7D,$E1,$F9,$1F,$E8,$2F,$86,$9E
	.dc.b	$2F,$13,$78,$F2,$D5,$FA,$D7,$F0,$0F,$CE,$3F,$CF,$FC,$BF,$E7,$5F
	.dc.b	$A2,$FF,$25,$65,$FC,$4F,$8B,$F4,$D5,$AF,$B3,$FF,$0C,$7A,$3F,$E5
	.dc.b	$2F,$B0,$E5,$F8,$07,$FB,$A7,$8F,$FD,$48,$63,$3F,$DD,$61,$C7,$3F
	.dc.b	$F2,$B7,$6F,$ED,$A3,$18,$FD,$EF,$73,$FD,$59,$18,$D5,$FB,$8F,$3D
	.dc.b	$7C,$30,$FC,$8B,$19,$7A,$3F,$73,$B9,$7F,$99,$BF,$46,$FE,$3B,$F6
	.dc.b	$CF,$93,$FE,$AF,$F6,$7F,$7F,$E7,$DF,$A0,$3F,$6E,$DD,$6B,$BF,$FC
	.dc.b	$35,$2F,$DB,$B3,$E3,$FF,$15,$CB,$A7,$97,$70,$E1,$97,$FA,$FB,$A3
	.dc.b	$7A,$77,$FB,$9F,$E2,$B8,$78,$F1,$4F,$8F,$37,$D6,$F0,$D7,$C3,$F5
	.dc.b	$BF,$CD,$BF,$70,$7E,$3A,$FD,$D5,$F9,$FF,$F8,$3F,$D0,$7C,$64,$F1
	.dc.b	$3F,$8F,$61,$C2,$F7,$F4,$9F,$B8,$AF,$E7,$B5,$AC,$7A,$71,$7E,$58
	.dc.b	$61,$E0,$F8,$9F,$E6,$8F,$3C,$FC,$1E,$1F,$F3,$08,$C6,$5F,$C1,$31
	.dc.b	$C6,$3E,$7B,$F3,$65,$FE,$7E,$9F,$06,$7E,$4D,$3D,$3F,$90,$FF,$54
	.dc.b	$FD,$2F,$F7,$3E,$A7,$FA,$B3,$F0,$C7,$EF,$7F,$E6,$19,$D7,$AF,$FD
	.dc.b	$47,$F0,$5F,$F7,$94,$FF,$99,$7F,$3D,$CE,$1F,$D2,$EF,$F8,$1F,$EA
	.dc.b	$97,$0E,$18,$F4,$F8,$9F,$E9,$35,$E0,$CB,$9F,$26,$2F,$79,$FF,$9C
	.dc.b	$67,$A9,$E0,$B7,$D8,$E3,$1E,$9F,$A9,$DB,$FF,$3C,$7E,$3A,$E2,$FF
	.dc.b	$0F,$74,$FF,$03,$FC,$A1,$F3,$CF,$C6,$7F,$D7,$9E,$1B,$F7,$3F,$F7
	.dc.b	$37,$EF,$1C,$7E,$AB,$F8,$56,$2C,$78,$E1,$C2,$FF,$DD,$9F,$84,$EE
	.dc.b	$BB,$65,$FC,$55,$FD,$B9,$93,$93,$E1,$73,$9F,$99,$CD,$F5,$F9,$3F
	.dc.b	$E3,$53,$9C,$7C,$3E,$87,$FB,$53,$27,$EB,$FA,$D3,$CF,$5A,$DB,$D9
	.dc.b	$F6,$FF,$EA,$6C,$79,$6B,$F0,$72,$E5,$2F,$CD,$FC,$DF,$9E,$BF,$36
	.dc.b	$7E,$6C,$F8,$1F,$DF,$7E,$4D,$FD,$5F,$F7,$ED,$D6,$A7,$3F,$21,$FC
	.dc.b	$A3,$5F,$66,$3C,$FF,$79,$F8,$7C,$DC,$33,$B5,$AC,$79,$7A,$ED,$1F
	.dc.b	$A6,$FB,$70,$FE,$79,$97,$97,$1F,$25,$EF,$F5,$AD,$52,$B5,$AF,$93
	.dc.b	$EA,$BC,$0F,$9B,$D7,$FA,$9F,$B1,$7F,$B9,$3F,$2D,$7E,$E1,$F0,$7F
	.dc.b	$05,$FF,$62,$ED,$AF,$C4,$95,$AE,$15,$7F,$B2,$B9,$35,$3E,$CE,$CF
	.dc.b	$C4,$E5,$FE,$F0,$FD,$8B,$1C,$1F,$D5,$63,$C1,$A7,$7F,$CE,$FD,$31
	.dc.b	$8C,$FA,$FE,$57,$8D,$7C,$37,$E5,$FC,$45,$FB,$17,$82,$31,$F9,$CD
	.dc.b	$C3,$CB,$4F,$D1,$71,$C3,$FE,$E0,$FD,$17,$F2,$9E,$CF,$FB,$28,$68
	.dc.b	$FD,$DF,$FC,$D7,$FB,$6F,$F6,$FE,$FF,$CB,$56,$EB,$5F,$7A,$AF,$F1
	.dc.b	$C5,$6B,$F0,$FF,$F0,$CF,$27,$E7,$B9,$BF,$57,$8F,$FB,$63,$DF,$F8
	.dc.b	$73,$FE,$1E,$AF,$F0,$FE,$5D,$3F,$6D,$FB,$77,$DF,$4F,$F3,$EF,$95
	.dc.b	$8B,$DC,$79,$BE,$7B,$D7,$18,$DB,$C9,$E5,$7E,$C7,$F7,$27,$ED,$CF
	.dc.b	$DC,$FF,$CA,$3F,$4C,$E9,$7E,$B0,$F9,$AE,$17,$BF,$B5,$FF,$0D,$D6
	.dc.b	$BE,$1F,$A7,$E3,$FF,$9C,$E5,$DE,$7E,$CD,$3B,$BD,$AD,$E6,$7E,$5A
	.dc.b	$73,$8F,$4F,$23,$19,$F7,$DB,$C9,$B7,$F9,$97,$AB,$F8,$36,$1C,$27
	.dc.b	$3F,$75,$FC,$37,$ED,$DF,$BD,$FB,$7F,$6B,$FE,$B3,$FB,$E3,$E2,$7F
	.dc.b	$C6,$59,$7F,$0F,$BD,$F0,$AF,$75,$FE,$27,$AD,$61,$F1,$F8,$BE,$D1
	.dc.b	$FE,$2D,$C1,$FA,$A7,$C1,$AD,$E3,$CF,$E4,$19,$F8,$70,$F3,$63,$E5
	.dc.b	$9C,$DF,$EE,$31,$F1,$E7,$E2,$FC,$DF,$79,$73,$CB,$D5,$F9,$C6,$2B
	.dc.b	$CF,$F5,$97,$2E,$FB,$7F,$BA,$F3,$FE,$59,$F9,$4B,$F0,$A7,$E9,$7F
	.dc.b	$E5,$8F,$C5,$3F,$C8,$DF,$B3,$7F,$4C,$BD,$D6,$BE,$4F,$FD,$C5,$C6
	.dc.b	$A7,$C9,$FF,$3B,$CB,$A7,$5F,$E6,$9E,$D9,$DC,$79,$67,$FE,$53,$FB
	.dc.b	$8F,$E2,$B5,$AC,$FF,$DD,$DF,$99,$FF,$4E,$D7,$C3,$8F,$97,$5F,$07
	.dc.b	$C9,$77,$CF,$9C,$78,$FE,$47,$FA,$67,$E6,$BE,$5D,$1F,$89,$BF,$7D
	.dc.b	$7E,$D1,$FE,$8D,$F8,$CD,$F2,$7F,$9D,$AB,$53,$FE,$91,$FD,$67,$38
	.dc.b	$7E,$D3,$1F,$81,$74,$E1,$8F,$4E,$4F,$E5,$F3,$E1,$C7,$9F,$8F,$8B
	.dc.b	$DE,$5F,$E9,$DF,$DB,$3A,$D7,$83,$E1,$5F,$AE,$7F,$77,$FD,$4F,$FB
	.dc.b	$7F,$EA,$0F,$9E,$61,$FE,$C0,$FC,$F1,$FA,$7B,$F4,$D7,$A5,$C7,$9F
	.dc.b	$FD,$01,$8F,$1C,$38,$5E,$FD,$0D,$52,$BB,$79,$7F,$D4,$B4,$F0,$E9
	.dc.b	$E0,$AF,$EE,$B1,$8C,$79,$FE,$59,$FD,$D2,$31,$F8,$93,$6F,$2F,$E6
	.dc.b	$B9,$C3,$C7,$8F,$AA,$B5,$AF,$A7,$F7,$CF,$1E,$BE,$79,$CF,$07,$ED
	.dc.b	$FC,$16,$9D,$B4,$F8,$33,$FC,$0E,$31,$8F,$C3,$CB,$DB,$FA,$03,$F3
	.dc.b	$0F,$CB,$FC,$53,$F1,$AF,$F3,$97,$99,$FD,$FF,$4E,$BC,$1F,$EA,$7C
	.dc.b	$38,$79,$6C,$38,$7D,$DF,$F9,$8B,$C3,$EB,$9F,$E8,$B0,$F8,$A7,$17
	.dc.b	$FC,$6A,$E0,$61,$D2,$F9,$F9,$9F,$16,$73,$5E,$6D,$3F,$D0,$DF,$82
	.dc.b	$E7,$0E,$11,$E8,$FD,$B7,$C3,$A7,$96,$3D,$7F,$94,$7D,$CF,$99,$F8
	.dc.b	$1F,$6C,$FD,$CF,$FC,$DB,$F6,$BE,$BF,$DE,$75,$AA,$B5,$F6,$EA,$F8
	.dc.b	$1F,$29,$AF,$D7,$FF,$09,$C3,$FE,$CB,$CF,$AB,$ED,$7B,$E7,$DB,$E7
	.dc.b	$3D,$91,$8C,$7A,$7C,$96,$34,$F0,$5F,$93,$95,$F4,$BE,$4D,$BC,$7B
	.dc.b	$7D,$1F,$A4,$FC,$BF,$91,$FE,$E7,$F9,$77,$E3,$FF,$E6,$2F,$C5,$1F
	.dc.b	$BA,$BF,$7A,$DF,$E2,$3F,$DC,$EB,$52,$FC,$BB,$38,$7B,$39,$7F,$B1
	.dc.b	$DB,$E6,$F1,$9F,$7C,$7B,$7C,$D7,$D7,$8C,$7D,$C3,$5E,$18,$F4,$BE
	.dc.b	$FF,$A1,$DA,$D6,$AF,$DD,$B1,$87,$A3,$F0,$4B,$2F,$DB,$FF,$6B,$F9
	.dc.b	$13,$EB,$9F,$8F,$7F,$B5,$BF,$36,$FD,$76,$AF,$D7,$B7,$5A,$FC,$51
	.dc.b	$5A,$FD,$9F,$3F,$86,$FB,$F2,$E9,$95,$C3,$85,$FE,$E7,$DB,$FC,$A5
	.dc.b	$6B,$1E,$7E,$C7,$F5,$EE,$4E,$F8,$F8,$E3,$E9,$FE,$FB,$8A,$DE,$CF
	.dc.b	$CB,$BF,$87,$FE,$78,$FC,$F5,$F9,$23,$F7,$F7,$DB,$BF,$57,$7C,$AB
	.dc.b	$F3,$FE,$38,$70,$BD,$D6,$BF,$59,$CE,$3F,$6F,$2E,$DF,$43,$70,$E1
	.dc.b	$9F,$5F,$91,$FD,$85,$78,$EF,$C5,$3E,$1F,$9B,$CD,$F0,$CB,$C7,$EB
	.dc.b	$F6,$D3,$C3,$B7,$F9,$83,$E0,$7F,$84,$F3,$3F,$2C,$FE,$45,$FB,$9B
	.dc.b	$F5,$C7,$D6,$3F,$79,$7E,$63,$FE,$98,$C7,$8E,$1C,$2F,$75,$AE,$0F
	.dc.b	$E6,$F3,$D3,$2F,$E1,$8F,$F2,$FF,$DE,$FF,$34,$F1,$CB,$B7,$9D,$FC
	.dc.b	$36,$31,$83,$F5,$1C,$77,$36,$F3,$61,$E5,$D9,$F9,$7F,$82,$B5,$B7
	.dc.b	$92,$5F,$6D,$9C,$38,$E5,$EC,$F9,$7C,$5E,$F8,$7F,$A7,$FD,$1B,$F6
	.dc.b	$17,$C3,$3C,$B7,$E5,$5F,$ED,$CF,$C2,$3F,$CE,$18,$7C,$CC,$38,$5E
	.dc.b	$EB,$5F,$9B,$74,$C7,$9F,$3F,$F2,$FB,$ED,$F3,$BF,$BD,$78,$E7,$DB
	.dc.b	$D6,$7D,$E7,$F3,$19,$CE,$5F,$EE,$CF,$47,$19,$79,$B0,$E1,$B7,$FB
	.dc.b	$27,$E5,$FF,$DC,$78,$70,$CB,$FD,$5D,$F6,$7F,$EE,$FF,$DD,$3F,$DC
	.dc.b	$FF,$BF,$7F,$AE,$7F,$3C,$FF,$13,$FE,$66,$FD,$0D,$FA,$FF,$F7,$AE
	.dc.b	$66,$1C,$3F,$11,$61,$C2,$F7,$FB,$D7,$5E,$60,$CF,$B3,$FE,$EA,$C3
	.dc.b	$97,$E2,$5F,$EC,$C7,$A6,$1C,$FC,$DF,$F7,$5C,$63,$E6,$1B,$79,$7D
	.dc.b	$1C,$E3,$DB,$2F,$55,$6B,$4F,$4F,$ED,$BE,$3A,$F1,$6D,$E8,$AD,$70
	.dc.b	$7F,$69,$E0,$F6,$D7,$A3,$E0,$9F,$B1,$3F,$09,$7E,$6E,$FA,$97,$E4
	.dc.b	$EF,$DC,$1F,$97,$77,$C7,$8E,$1C,$2F,$74,$FC,$B3,$59,$7F,$AC,$33
	.dc.b	$F8,$E7,$B0,$61,$E2,$FC,$3F,$E3,$47,$96,$7C,$9A,$78,$F4,$FF,$6E
	.dc.b	$CF,$4C,$FC,$10,$FA,$5E,$98,$F9,$2F,$86,$7F,$C7,$F6,$F8,$38,$F8
	.dc.b	$32,$EB,$AF,$A3,$F4,$6E,$98,$70,$FD,$92,$CF,$B3,$E0,$7E,$77,$E8
	.dc.b	$4F,$C2,$DF,$BD,$7F,$0B,$FD,$5D,$FA,$8B,$91,$FB,$9F,$0B,$DF,$A8
	.dc.b	$BD,$D3,$F7,$4D,$7B,$1F,$0B,$89,$A7,$7D,$30,$DC,$79,$69,$FE,$9B
	.dc.b	$E3,$F6,$CE,$73,$F1,$C7,$CF,$8C,$4F,$97,$F3,$DF,$26,$1E,$98,$F4
	.dc.b	$7E,$5D,$F8,$BF,$6B,$F8,$B7,$EE,$EF,$8F,$FF,$14,$FE,$A4,$FD,$C9
	.dc.b	$F8,$86,$FE,$0D,$7B,$FB,$45,$EF,$F1,$ED,$7E,$85,$ED,$CF,$AF,$27
	.dc.b	$E9,$CF,$D5,$BE,$0D,$3B,$C7,$E1,$E3,$18,$74,$FC,$03,$19,$F8,$2F
	.dc.b	$CB,$AB,$F1,$1F,$26,$BE,$3D,$7D,$1F,$82,$FE,$9F,$F1,$B7,$EB,$3F
	.dc.b	$E8,$0F,$BE,$7B,$0F,$D4,$9F,$05,$FE,$B6,$EB,$FC,$4E,$1C,$3E,$15
	.dc.b	$87,$0B,$DF,$E1,$9D,$72,$FF,$45,$F2,$FB,$4F,$F7,$D3,$FE,$7B,$7F
	.dc.b	$46,$6E,$F9,$E7,$E1,$CF,$F4,$D9,$77,$CB,$B7,$DE,$E3,$1F,$45,$EF
	.dc.b	$F2,$5E,$55,$E3,$C1,$F8,$EE,$2B,$A7,$E4,$77,$6F,$6F,$EB,$FF,$C2
	.dc.b	$7B,$9F,$8E,$7F,$B5,$7F,$57,$7E,$0A,$FD,$8D,$F8,$BE,$39,$BF,$57
	.dc.b	$87,$0F,$8A,$61,$C2,$F7,$F9,$67,$5C,$BF,$98,$73,$FC,$3C,$38,$36
	.dc.b	$3C,$19,$F7,$F8,$4F,$E0,$D5,$D7,$1E,$9F,$0F,$8B,$DE,$5E,$4F,$CB
	.dc.b	$FB,$69,$E2,$CB,$D1,$B3,$F0,$BF,$E4,$DF,$70,$FD,$25,$F9,$3F,$F8
	.dc.b	$8F,$C6,$FF,$38,$7E,$E8,$FC,$B7,$75,$AF,$A9,$FE,$F1,$F8,$12,$9F
	.dc.b	$94,$6B,$1F,$DC,$65,$D7,$F4,$DB,$8F,$1C,$FF,$CF,$9F,$6D,$FC,$86
	.dc.b	$BC,$57,$E1,$C9,$E7,$67,$0E,$11,$DF,$F8,$07,$B6,$BE,$2D,$7F,$DA
	.dc.b	$9D,$9F,$C7,$3F,$0C,$FA,$DE,$FF,$F7,$AF,$86,$FF,$66,$7A,$CF,$D7
	.dc.b	$3F,$A3,$56,$BE,$7F,$FA,$0F,$E3,$0A,$FF,$59,$7C,$C7,$F3,$6A,$FC
	.dc.b	$07,$ED,$BF,$C4,$DE,$BE,$27,$E6,$DE,$59,$7F,$B6,$7F,$64,$FE,$79
	.dc.b	$18,$C7,$FD,$A5,$5E,$0F,$03,$F9,$46,$7E,$0C,$FF,$DD,$FB,$7A,$67
	.dc.b	$39,$7A,$3E,$97,$86,$FE,$BC,$BC,$DD,$18,$CB,$97,$DC,$FF,$61,$F1
	.dc.b	$1F,$91,$3F,$58,$7E,$2E,$FC,$B5,$F8,$03,$E8,$FF,$AF,$70,$FF,$44
	.dc.b	$C7,$8E,$1C,$2F,$75,$F8,$3A,$D6,$3D,$78,$FF,$79,$87,$83,$F3,$0F
	.dc.b	$E7,$0F,$2C,$FC,$19,$BF,$15,$FE,$C1,$39,$CF,$C7,$F8,$86,$32,$F4
	.dc.b	$61,$C3,$47,$E8,$3E,$D9,$CD,$F9,$3F,$42,$6D,$FC,$6B,$6F,$F0,$9F
	.dc.b	$E1,$BF,$E7,$DF,$C3,$5F,$A4,$3F,$52,$7E,$9A,$F6,$5F,$A8,$F2,$F8
	.dc.b	$2F,$EE,$B8,$F1,$C3,$85,$EF,$F0,$9D,$67,$DB,$C8,$5F,$F1,$A8,$FF
	.dc.b	$11,$EB,$E3,$73,$3F,$EA,$F7,$51,$99,$ED,$A7,$93,$EE,$F8,$E7,$E2
	.dc.b	$C3,$FD,$6D,$9F,$8B,$C5,$CE,$5E,$1D,$FD,$B8,$FB,$37,$7B,$FE,$6A
	.dc.b	$F0,$D6,$B7,$F8,$1E,$2E,$7F,$73,$7B,$FF,$2B,$F2,$17,$E8,$EF,$3D
	.dc.b	$FB,$EF,$F4,$3E,$5F,$66,$1C,$15,$F0,$AD,$FD,$2F,$5C,$7E,$AE,$BF
	.dc.b	$03,$93,$FC,$43,$8F,$83,$EA,$7E,$44,$EF,$2E,$7A,$78,$7D,$FF,$7C
	.dc.b	$AF,$C1,$A7,$7F,$0B,$31,$8B,$F4,$6D,$FE,$23,$C7,$CF,$A7,$B6,$73
	.dc.b	$F3,$3D,$F1,$F4,$5B,$E8,$F1,$5C,$BF,$0D,$F9,$5F,$A1,$3F,$3A,$FF
	.dc.b	$74,$7D,$4B,$F6,$37,$C0,$BE,$55,$F2,$9C,$2F,$75,$A9,$FC,$DF,$07
	.dc.b	$E1,$71,$E9,$D7,$F0,$A3,$17,$F8,$CE,$4F,$90,$7F,$9B,$AB,$79,$F8
	.dc.b	$73,$F0,$4F,$7C,$3F,$CB,$39,$79,$3E,$37,$E4,$E1,$C2,$3C,$5D,$5B
	.dc.b	$C7,$96,$9F,$CA,$BA,$7F,$CA,$BF,$9B,$3F,$77,$7E,$E8,$FC,$E9,$F8
	.dc.b	$C3,$F4,$37,$E7,$2F,$CC,$BF,$21,$8F,$1C,$38,$5E,$E9,$F8,$0E,$B2
	.dc.b	$EB,$C7,$EE,$18,$78,$7C,$9F,$89,$AF,$92,$BC,$7A,$F8,$B5,$FF,$35
	.dc.b	$F3,$FE,$BF,$0F,$04,$E7,$E2,$BF,$13,$18,$D7,$C5,$F0,$3E,$1D,$FC
	.dc.b	$B8,$FA,$BE,$7B,$F5,$1F,$8F,$FF,$88,$BB,$9F,$A0,$BF,$48,$7D,$0B
	.dc.b	$F2,$97,$E2,$FC,$3E,$7F,$F5,$37,$BA,$D3,$9C,$3D,$7C,$DF,$EA,$8C
	.dc.b	$BF,$C5,$FF,$A6,$62,$11,$E1,$AE,$5F,$97,$7E,$8A,$31,$A7,$8B,$E0
	.dc.b	$F1,$9F,$82,$3C,$BA,$BF,$19,$F0,$63,$E3,$8F,$F0,$CF,$3F,$F1,$5C
	.dc.b	$79,$ED,$EB,$FD,$A3,$78,$F2,$FD,$BB,$F8,$57,$EF,$BF,$E5,$EF,$C3
	.dc.b	$9F,$18,$FB,$1F,$F5,$2F,$C9,$3C,$FD,$FD,$A2,$F7,$5A,$FD,$07,$31
	.dc.b	$EF,$C7,$FC,$D3,$FA,$2D,$BE,$19,$75,$F6,$DC,$FF,$C0,$AF,$C5,$39
	.dc.b	$FD,$37,$17,$C3,$3F,$1F,$E0,$7D,$AB,$96,$CF,$DA,$FF,$92,$7E,$5A
	.dc.b	$FD,$D7,$FC,$33,$F9,$F3,$F1,$F7,$CA,$FF,$A5,$7F,$67,$DD,$6A,$73
	.dc.b	$F8,$17,$FB,$23,$B3,$F9,$F6,$3C,$FE,$AB,$ED,$3D,$7E,$AE,$E1,$C7
	.dc.b	$2F,$5F,$89,$D2,$DA,$78,$65,$C7,$FC,$83,$1E,$93,$9C,$DF,$B4,$F1
	.dc.b	$8C,$47,$F3,$FF,$D0,$DF,$9D,$FF,$8F,$FF,$43,$7C,$C7,$E0,$FF,$1C
	.dc.b	$7E,$D3,$BA,$D4,$E7,$D6,$C7,$CB,$7E,$4F,$1E,$7F,$65,$6F,$7E,$63
	.dc.b	$F2,$0F,$E6,$B5,$AD,$3B,$CF,$FA,$23,$93,$B6,$09,$CE,$AF,$D4,$F3
	.dc.b	$18,$E3,$BB,$79,$FF,$3D,$B9,$30,$7E,$47,$FE,$00,$FC,$1B,$FD,$D7
	.dc.b	$F9,$D7,$F9,$D3,$F3,$7F,$F3,$AF,$27,$DB,$F1,$BF,$DA,$5D,$DF,$EB
	.dc.b	$7E,$10,$BE,$1F,$75,$F1,$3C,$05,$BF,$18,$ED,$7C,$2F,$B3,$FD,$EB
	.dc.b	$3F,$0E,$0F,$C4,$3F,$9D,$46,$34,$F1,$E4,$FC,$AF,$FC,$0B,$39,$BF
	.dc.b	$07,$DE,$63,$1E,$D0,$C3,$8E,$9F,$E3,$1B,$E1,$B7,$B7,$C2,$F2,$E0
	.dc.b	$FC,$D1,$F6,$3F,$E7,$4F,$99,$79,$1F,$B7,$FC,$73,$F0,$7C,$9F,$0F
	.dc.b	$0E,$17,$BA,$D7,$D6,$F4,$C7,$9E,$5E,$CB,$F9,$31,$DB,$CA,$C4,$72
	.dc.b	$9F,$F0,$1D,$FF,$05,$F9,$1F,$64,$E7,$4F,$1F,$A1,$FE,$AC,$8E,$D9
	.dc.b	$F8,$61,$AF,$F9,$5B,$5F,$36,$9E,$2C,$38,$CB,$F4,$BF,$83,$FA,$37
	.dc.b	$FB,$6B,$F6,$8F,$E2,$9F,$6E,$FB,$C7,$E8,$FF,$E0,$9C,$FE,$05,$EF
	.dc.b	$45,$EE,$BF,$1D,$5A,$F9,$27,$EF,$5E,$EC,$3B,$70,$7E,$04,$FC,$5B
	.dc.b	$CF,$0F,$07,$CA,$FC,$08,$C6,$6F,$BA,$C6,$2D,$7C,$F3,$E6,$FC,$57
	.dc.b	$D7,$39,$DB,$CB,$3E,$AF,$C4,$B3,$7C,$37,$7E,$31,$8B,$DC,$3F,$58
	.dc.b	$FC,$2F,$F0,$0B,$F2,$5F,$F3,$0F,$EC,$EF,$D8,$BF,$B7,$FE,$22,$F7
	.dc.b	$77,$5A,$9C,$FE,$73,$8F,$91,$FB,$F8,$F3,$FB,$5B,$7B,$F9,$E7,$E1
	.dc.b	$DF,$80,$AD,$69,$DE,$7F,$45,$D5,$DB,$04,$E7,$57,$E1,$19,$8C,$7E
	.dc.b	$E7,$F9,$EE,$F7,$E4,$FF,$EF,$3F,$C4,$1D,$7F,$E7,$DF,$15,$F9,$3E
	.dc.b	$EB,$53,$9F,$98,$C7,$EB,$3F,$98,$C7,$9F,$E1,$B6,$F7,$F8,$07,$FA
	.dc.b	$AB,$87,$94,$FF,$B9,$B1,$FF,$45,$7E,$03,$F9,$CC,$38,$EB,$E2,$C3
	.dc.b	$F6,$3F,$B6,$70,$8C,$6C,$FD,$07,$F6,$FE,$CB,$F8,$01,$F9,$BF,$F5
	.dc.b	$0F,$DB,$1F,$1A,$FC,$25,$FB,$3F,$0B,$DF,$CA,$2F,$75,$FC,$07,$D2
	.dc.b	$FD,$E6,$5D,$72,$FD,$AB,$DE,$EE,$1C,$B0,$EF,$F6,$BF,$B6,$AE,$15
	.dc.b	$D7,$3F,$F6,$CE,$3D,$30,$E1,$1E,$5F,$96,$F2,$65,$E9,$8F,$47,$C4
	.dc.b	$63,$0E,$5A,$7F,$9A,$71,$E5,$7B,$8F,$E9,$51,$FC,$4F,$F7,$D7,$E6
	.dc.b	$CF,$C6,$BF,$75,$FB,$63,$F3,$07,$E3,$6F,$DB,$BF,$90,$E1,$C2,$F7
	.dc.b	$5A,$FD,$EB,$38,$F3,$EF,$FD,$C0,$BE,$FF,$6D,$FC,$69,$E5,$EC,$34
	.dc.b	$F0,$7B,$5F,$C6,$A7,$31,$E3,$FB,$9C,$63,$E7,$BE,$19,$F9,$BF,$4F
	.dc.b	$F0,$EB,$E4,$DF,$FC,$1B,$C1,$FC,$C7,$F8,$01,$E5,$7F,$98,$3F,$4D
	.dc.b	$7E,$C4,$FD,$E5,$FA,$A7,$F8,$EF,$0E,$17,$BA,$D4,$FF,$8F,$3E,$3F
	.dc.b	$FA,$EF,$2B,$2E,$BF,$2F,$FC,$D9,$EF,$EA,$DC,$79,$66,$FD,$93,$F9
	.dc.b	$BD,$6B,$5F,$1C,$BF,$07,$CD,$F0,$D1,$F6,$DF,$D4,$B0,$E1,$B7,$93
	.dc.b	$F0,$8F,$F7,$AF,$E0,$BE,$4F,$A2,$FF,$19,$76,$BE,$B5,$F4,$BF,$A0
	.dc.b	$F6,$1F,$95,$F0,$E5,$FE,$EF,$87,$0B,$DF,$91,$FE,$72,$FC,$67,$55
	.dc.b	$FE,$0D,$FA,$FF,$F4,$5B,$F0,$8C,$DF,$F9,$07,$F0,$7F,$F5,$76,$BE
	.dc.b	$3A,$FF,$4E,$E3,$D3,$3F,$F7,$CC,$E7,$5F,$36,$BF,$B4,$F6,$33,$9F
	.dc.b	$93,$1E,$33,$E9,$FC,$67,$BC,$63,$6F,$F4,$3F,$E3,$38,$C7,$9C,$7F
	.dc.b	$A2,$BF,$5F,$C6,$BE,$1D,$BD,$7E,$B7,$F9,$83,$F2,$1F,$F1,$37,$EE
	.dc.b	$0F,$CF,$3F,$D6,$1F,$A1,$6B,$EE,$D5,$A9,$FF,$5A,$7E,$87,$98,$F7
	.dc.b	$F1,$7F,$88,$31,$FF,$35,$FE,$B3,$88,$FE,$29,$87,$2F,$B3,$67,$DF
	.dc.b	$E8,$BD,$F5,$AC,$7B,$6B,$8C,$BC,$D8,$70,$D3,$FD,$89,$F9,$27,$86
	.dc.b	$31,$F8,$77,$F8,$53,$CB,$FF,$63,$E5,$FE,$32,$FD,$9D,$F9,$E7,$F8
	.dc.b	$97,$F0,$0F,$DE,$7E,$25,$FE,$AC,$FD,$4F,$EE,$D6,$BE,$0D,$5A,$FC
	.dc.b	$47,$F9,$36,$5D,$32,$5E,$DF,$C1,$1C,$B2,$7E,$A5,$F0,$67,$A6,$1E
	.dc.b	$0F,$C8,$BD,$F1,$8F,$80,$FF,$6F,$F9,$38,$C3,$CD,$1F,$EE,$2C,$DF
	.dc.b	$9C,$75,$D7,$CB,$39,$F2,$BF,$2B,$F8,$47,$F9,$17,$95,$C3,$FD,$83
	.dc.b	$EC,$5F,$8E,$3F,$3F,$DF,$DD,$FD,$FA,$D4,$BF,$5D,$CE,$1F,$67,$1F
	.dc.b	$F2,$87,$E4,$77,$0E,$1F,$38,$CD,$F7,$9F,$6A,$FC,$59,$F8,$7F,$61
	.dc.b	$C6,$1C,$2F,$C7,$C1,$DA,$BC,$3F,$AC,$5C,$BA,$6C,$FC,$E3,$18,$7A
	.dc.b	$BE,$F7,$F6,$7F,$90,$7F,$B9,$BF,$37,$7E,$42,$FA,$87,$8B,$FE,$5D
	.dc.b	$EA,$DF,$B2,$BD,$D6,$BF,$01,$FF,$0C,$78,$9F,$77,$2E,$9E,$27,$DC
	.dc.b	$8C,$60,$FC,$A3,$F3,$95,$E1,$D3,$C1,$A7,$C2,$C3,$BE,$7D,$A3,$FC
	.dc.b	$27,$8F,$28,$C6,$5F,$EE,$6F,$6F,$F9,$5E,$3C,$F4,$F2,$7E,$0F,$FC
	.dc.b	$43,$BB,$FD,$73,$F0,$FF,$DC,$FF,$22,$7E,$D6,$FD,$0F,$FC,$59,$AF
	.dc.b	$E5,$6F,$7F,$40,$BD,$D3,$BB,$A7,$E2,$7F,$CC,$F8,$BD,$DB,$FF,$3C
	.dc.b	$C7,$FD,$8F,$F9,$DF,$EE,$5E,$78,$F8,$3B,$FF,$6F,$18,$D7,$FC,$67
	.dc.b	$87,$4D,$BC,$D1,$E5,$F8,$CC,$6D,$E3,$8F,$47,$D7,$7F,$2C,$EE,$FF
	.dc.b	$57,$FE,$A6,$FD,$5B,$FB,$27,$D5,$BF,$2D,$FF,$2D,$72,$FE,$C6,$F7
	.dc.b	$5A,$97,$EB,$DC,$FF,$10,$7F,$00,$FF,$11,$BE,$BC,$1E,$E3,$FD,$6A
	.dc.b	$F2,$FA,$0F,$F3,$F7,$E6,$3F,$C4,$A3,$1A,$3D,$7C,$62,$DB,$CD,$87
	.dc.b	$FA,$17,$47,$EF,$5E,$19,$CE,$1E,$9F,$CE,$31,$BF,$9F,$ED,$AE,$3C
	.dc.b	$7F,$29,$FE,$9B,$FA,$83,$F7,$1F,$C7,$7E,$9D,$FE,$BB,$1F,$BC,$3F
	.dc.b	$48,$DC,$EB,$B9,$5A,$FD,$2B,$3E,$6F,$F3,$1C,$7A,$7C,$25,$C3,$85
	.dc.b	$BE,$AF,$F9,$8D,$78,$6F,$9E,$3D,$3E,$B7,$38,$70,$D1,$FA,$4F,$6C
	.dc.b	$BC,$D0,$FD,$BD,$DF,$FD,$97,$F3,$17,$E2,$1F,$6F,$F4,$A7,$E4,$EF
	.dc.b	$C1,$7F,$2D,$E2,$F0,$F0,$FF,$C1,$B8,$70,$EF,$61,$C2,$FF,$39,$E4
	.dc.b	$6E,$FF,$DD,$F7,$DB,$E9,$7F,$99,$C6,$44,$73,$7F,$78,$76,$CF,$C9
	.dc.b	$9F,$8F,$EA,$7F,$DF,$3A,$FA,$67,$39,$79,$63,$1A,$79,$2F,$D5,$E6
	.dc.b	$BC,$7F,$06,$7E,$9D,$BD,$13,$FE,$27,$C3,$FC,$11,$ED,$3C,$67,$E6
	.dc.b	$4F,$D0,$FF,$CA,$5E,$3B,$E7,$3F,$8A,$F1,$7F,$8C,$30,$E1,$7B,$FB
	.dc.b	$3F,$CA,$53,$F5,$4F,$81,$5A,$BF,$78,$0F,$87,$E0,$3E,$3E,$9E,$4D
	.dc.b	$3C,$7A,$78,$BD,$FF,$87,$87,$86,$73,$9F,$EB,$A3,$27,$5C,$7C,$B7
	.dc.b	$EA,$CD,$F9,$0F,$93,$0E,$3B,$7B,$7D,$CC,$6F,$EB,$8F,$81,$F2,$58
	.dc.b	$C7,$97,$D6,$7D,$BF,$C5,$1F,$8B,$7F,$BE,$AB,$FA,$EF,$ED,$8F,$39
	.dc.b	$F4,$7B,$AD,$7A,$EA,$D7,$06,$7F,$1C,$F8,$3C,$7F,$E5,$CC,$FA,$F8
	.dc.b	$F8,$8F,$91,$1E,$09,$69,$E0,$D3,$E0,$D7,$FB,$9B,$3E,$F9,$3F,$00
	.dc.b	$CE,$3C,$70,$E1,$FB,$1F,$E5,$7F,$50,$FD,$97,$EE,$1F,$82,$FC,$D1
	.dc.b	$FB,$3F,$FA,$27,$F6,$BD,$7A,$AA,$D4,$FF,$A3,$7F,$61,$CC,$79,$F7
	.dc.b	$63,$FB,$DF,$0F,$0F,$6E,$A6,$7D,$BE,$2B,$D3,$39,$8F,$0F,$CC,$E3
	.dc.b	$4F,$05,$F0,$D3,$FD,$39,$FB,$A7,$82,$31,$F8,$2B,$F5,$AF,$D8,$7E
	.dc.b	$5F,$FE,$10,$FB,$17,$E5,$8F,$80,$7E,$78,$FC,$CF,$FD,$23,$E6,$7F
	.dc.b	$9A,$F0,$E1,$7B,$AD,$7C,$7F,$4F,$0F,$FC,$F9,$CD,$F9,$C1,$7F,$E1
	.dc.b	$EF,$CB,$3E,$13,$CB,$3E,$F9,$FD,$68,$C6,$3F,$C3,$2F,$A6,$9E,$6B
	.dc.b	$DF,$E4,$F8,$D3,$C5,$8F,$F9,$63,$5F,$36,$9F,$E8,$D9,$CF,$81,$FC
	.dc.b	$3F,$E4,$9F,$6F,$FD,$13,$DE,$FF,$23,$FD,$87,$F5,$6F,$91,$D4,$E7
	.dc.b	$EB,$53,$FD,$6B,$F7,$6C,$FC,$87,$DB,$C7,$9F,$C9,$5C,$38,$5F,$E7
	.dc.b	$F8,$F9,$CB,$3E,$FE,$4F,$E5,$30,$E3,$1E,$2F,$94,$FE,$B5,$C3,$FE
	.dc.b	$99,$9C,$EB,$E2,$97,$EC,$1F,$CA,$7F,$18,$7E,$94,$FC,$51,$FB,$33
	.dc.b	$CF,$7D,$6F,$FA,$DB,$7E,$17,$BA,$D4,$FF,$AE,$B8,$F3,$F8,$B7,$E5
	.dc.b	$F2,$EB,$95,$C3,$86,$5D,$3D,$1F,$D6,$D6,$B2,$E7,$F8,$C6,$2F,$79
	.dc.b	$FF,$B1,$BD,$4F,$0E,$5E,$6C,$BC,$BE,$CB,$E6,$7F,$A2,$7C,$3F,$FA
	.dc.b	$7F,$DC,$FE,$D3,$FB,$FB,$F7,$3F,$EA,$FF,$8B,$6B,$EC,$FF,$9A,$7B
	.dc.b	$CA,$D7,$C7,$BF,$73,$D6,$3F,$B1,$CB,$FD,$55,$F0,$B7,$1E,$38,$75
	.dc.b	$F0,$7F,$83,$D7,$87,$2E,$9A,$7F,$73,$9E,$71,$8C,$DF,$84,$63,$0F
	.dc.b	$0F,$E0,$EF,$D9,$1F,$B4,$7F,$BD,$3E,$35,$F8,$4F,$E0,$F9,$9F,$07
	.dc.b	$FB,$2F,$88,$7E,$0F,$C7,$0E,$17,$BA,$D7,$B1,$9E,$0F,$E9,$19,$74
	.dc.b	$F9,$9F,$ED,$BA,$78,$27,$38,$7E,$45,$FF,$51,$C6,$5F,$F3,$9C,$75
	.dc.b	$F4,$BF,$CE,$B1,$D4,$CF,$99,$F8,$DC,$9F,$9F,$FF,$57,$9D,$65,$EC
	.dc.b	$AD,$5F,$F6,$5F,$63,$38,$F5,$DF,$FC,$1B,$9F,$87,$F3,$07,$E9,$8F
	.dc.b	$D8,$5F,$83,$BE,$C3,$FB,$8F,$EC,$DF,$EC,$0C,$38,$5E,$EB,$5F,$08
	.dc.b	$A7,$ED,$9E,$9C,$BA,$FD,$77,$C4,$E7,$F4,$47,$83,$4C,$71,$FF,$71
	.dc.b	$8C,$65,$D3,$E8,$7F,$BA,$6B,$E0,$BF,$2E,$DF,$E0,$4F,$63,$39,$74
	.dc.b	$DF,$C9,$3E,$8E,$7F,$14,$63,$83,$CD,$F2,$3F,$37,$F0,$DF,$DD,$FE
	.dc.b	$18,$FD,$F3,$FC,$95,$FB,$2F,$F6,$7F,$D2,$35,$ED,$FF,$BF,$A7,$3E
	.dc.b	$93,$F3,$4C,$E1,$EA,$E6,$FF,$7C,$DB,$F1,$2C,$67,$DB,$1E,$35,$C1
	.dc.b	$EF,$F0,$3E,$A8,$FF,$09,$4E,$FE,$CD,$AF,$F7,$9F,$D2,$78,$46,$35
	.dc.b	$7E,$49,$8A,$FE,$FD,$E6,$D9,$7E,$43,$F8,$BF,$C8,$5F,$96,$7F,$B8
	.dc.b	$BF,$3A,$7D,$47,$FB,$5F,$F2,$D7,$EC,$7B,$AD,$4E,$7E,$43,$1F,$87
	.dc.b	$7B,$F1,$E7,$FB,$A5,$BD,$FE,$19,$FE,$2A,$E4,$DE,$9D,$A3,$FA,$67
	.dc.b	$89,$FA,$1C,$3C,$78,$AB,$5A,$BF,$75,$E3,$AF,$97,$F6,$CF,$D2,$7F
	.dc.b	$00,$FF,$BE,$BF,$21,$7E,$39,$F6,$74,$FF,$06,$7E,$17,$AD,$15,$AE
	.dc.b	$DF,$E0,$9D,$BF,$EF,$FB,$FD,$96,$2B,$7E,$41,$F6,$E3,$B7,$17,$EE
	.dc.b	$3C,$9F,$D8,$47,$87,$4F,$06,$3F,$66,$93,$9C,$BF,$89,$7A,$19,$8C
	.dc.b	$67,$FE,$21,$C3,$CD,$A7,$96,$3F,$C4,$9F,$81,$7F,$01,$F9,$C7,$E5
	.dc.b	$DF,$EC,$1F,$D3,$1F,$BF,$7F,$8D,$FF,$80,$7F,$C7,$F8,$7F,$0B,$BD
	.dc.b	$FE,$16,$BD,$F7,$2F,$FD,$55,$7B,$C7,$FD,$99,$C1,$ED,$3F,$D8,$2F
	.dc.b	$F1,$57,$F8,$74,$6F,$0E,$99,$7F,$91,$FF,$07,$F8,$19,$F8,$33,$EF
	.dc.b	$D7,$8C,$BD,$17,$E7,$F9,$2F,$64,$63,$5F,$27,$E8,$1F,$C7,$FF,$02
	.dc.b	$FF,$24,$FD,$8F,$F8,$63,$CC,$7E,$47,$FE,$74,$FD,$5B,$5F,$3B,$FE
	.dc.b	$36,$FA,$C5,$6B,$EF,$DF,$AB,$EB,$F5,$3F,$DA,$67,$D3,$9F,$DC,$7F
	.dc.b	$BA,$71,$E5,$87,$18,$EB,$A7,$87,$47,$E5,$5F,$77,$0F,$2E,$9E,$09
	.dc.b	$FF,$43,$FE,$C3,$FD,$3F,$A3,$F9,$A7,$E5,$DF,$CA,$BF,$55,$7E,$48
	.dc.b	$FE,$00,$7E,$58,$FD,$1D,$F9,$BF,$1E,$0F,$F4,$96,$3C,$70,$E1,$7B
	.dc.b	$F6,$F5,$4A,$FE,$6F,$F4,$BF,$D4,$34,$F0,$FC,$5B,$5C,$5F,$48,$C6
	.dc.b	$5D,$3B,$FF,$96,$DF,$0B,$F3,$ED,$E6,$F0,$B3,$8F,$19,$CE,$BE,$9F
	.dc.b	$B7,$F1,$FD,$83,$F4,$FF,$AF,$ED,$3F,$7B,$7E,$95,$FC,$5F,$C9,$FC
	.dc.b	$7E,$6C,$38,$74,$30,$E1,$7F,$C8,$3E,$CB,$AE,$5F,$82,$C3,$FB,$4F
	.dc.b	$07,$F6,$1C,$7C,$1F,$8B,$FB,$A7,$A5,$78,$B4,$F0,$FE,$3B,$EA,$9C
	.dc.b	$E9,$DE,$7C,$BE,$0E,$71,$F1,$EB,$E2,$F8,$4F,$F1,$26,$FE,$89,$CD
	.dc.b	$FB,$25,$FA,$17,$9B,$1E,$9C,$1F,$CE,$3F,$5B,$FA,$FE,$EF,$F3,$0F
	.dc.b	$05,$F9,$BB,$F2,$97,$EE,$3F,$DC,$BF,$08,$DF,$C3,$7F,$9E,$2B,$53
	.dc.b	$FD,$DB,$F7,$5C,$E1,$EF,$C7,$F9,$DF,$EA,$77,$0E,$19,$74,$F2,$BE
	.dc.b	$05,$78,$72,$E7,$E7,$E2,$F7,$9F,$F8,$0B,$E4,$BF,$1B,$F9,$73,$F7
	.dc.b	$4F,$B5,$F8,$73,$F3,$5F,$F4,$D7,$E6,$8F,$D9,$37,$5A,$9C,$F9,$B8
	.dc.b	$FC,$D3,$EA,$63,$CF,$AE,$DE,$FF,$0E,$7E,$99,$F5,$6B,$5A,$77,$97
	.dc.b	$C1,$E6,$F8,$6B,$FE,$18,$F0,$3F,$17,$F9,$AF,$F9,$3F,$C5,$FC,$47
	.dc.b	$E0,$AF,$94,$FF,$2C,$56,$BE,$F9,$5A,$97,$E5,$7F,$AC,$C3,$97,$93
	.dc.b	$F8,$E0,$1F,$F2,$E7,$EB,$FF,$8F,$7C,$1E,$3B,$CE,$68,$FD,$0B,$F2
	.dc.b	$15,$A9,$CF,$CF,$E3,$5F,$1E,$1C,$70,$F4,$7E,$59,$FE,$59,$FC,$9B
	.dc.b	$FC,$7D,$F7,$4F,$D4,$5E,$9B,$EB,$14,$FC,$61,$FA,$03,$F7,$5E,$5C
	.dc.b	$7E,$AC,$B9,$63,$C7,$0E,$16,$FD,$73,$77,$FE,$9B,$CD,$F9,$77,$F9
	.dc.b	$72,$FC,$71,$E2,$9F,$F2,$DC,$63,$0E,$99,$FF,$96,$BA,$7C,$B5,$E1
	.dc.b	$BF,$4E,$BE,$8E,$DF,$E9,$B8,$F1,$8F,$5E,$1E,$AF,$C1,$3C,$7F,$67
	.dc.b	$FE,$E3,$FD,$09,$FB,$A3,$F5,$1F,$F7,$47,$E6,$4E,$DD,$FE,$38,$BD
	.dc.b	$D6,$BF,$61,$CF,$C8,$FA,$F8,$7E,$EE,$7D,$5F,$59,$8C,$39,$67,$DF
	.dc.b	$E1,$3F,$D2,$5C,$7E,$CA,$FF,$34,$56,$B2,$FF,$7D,$7C,$3F,$F9,$63
	.dc.b	$8B,$FC,$ED,$97,$A7,$2E,$5B,$FE,$4B,$2E,$5B,$FD,$7E,$E7,$E3,$CF
	.dc.b	$D7,$9E,$F3,$F4,$AF,$E5,$78,$70,$EF,$FF,$57,$5E,$EB,$5F,$AC,$7D
	.dc.b	$AC,$BF,$05,$F8,$A7,$BF,$8F,$E3,$61,$FE,$B5,$7D,$E6,$34,$F0,$E9
	.dc.b	$E0,$E1,$F5,$4F,$E8,$B0,$F3,$67,$DF,$3C,$CF,$18,$F1,$6A,$FC,$73
	.dc.b	$C5,$9F,$9B,$2F,$5C,$BF,$1C,$FF,$5B,$7E,$5B,$FE,$9B,$FD,$AF,$FC
	.dc.b	$3B,$F5,$9F,$77,$F7,$F7,$E5,$CF,$D4,$7F,$B3,$F2,$B0,$E1,$F4,$EC
	.dc.b	$38,$5E,$FF,$0B,$EB,$C3,$FB,$AC,$FA,$BF,$EB,$2C,$38,$FE,$9F,$FE
	.dc.b	$B8,$79,$E1,$CB,$F5,$5F,$F4,$8C,$63,$B9,$B7,$93,$F4,$6C,$E3,$D7
	.dc.b	$5F,$15,$6B,$5F,$47,$CF,$F8,$AB,$F8,$8F,$8D,$FE,$E5,$F8,$97,$E0
	.dc.b	$9F,$E7,$4F,$A4,$FF,$03,$F1,$FB,$6F,$F8,$8F,$B8,$C3,$85,$EF,$F7
	.dc.b	$FE,$93,$9F,$03,$D5,$7C,$2D,$E0,$E2,$3C,$33,$FD,$4B,$0F,$F4,$1F
	.dc.b	$E3,$9E,$A9,$CE,$9E,$4F,$8C,$FE,$4F,$1F,$EE,$BF,$C4,$3F,$DA,$1F
	.dc.b	$51,$FE,$D2,$E9,$7E,$92,$FD,$A3,$FC,$67,$F6,$6F,$C0,$78,$63,$FE
	.dc.b	$9C,$FC,$5D,$87,$0B,$DF,$C2,$6A,$BF,$C7,$74,$FD,$4F,$FD,$D7,$A7
	.dc.b	$82,$38,$CF,$FB,$AA,$31,$87,$3D,$1F,$99,$7F,$B2,$27,$3A,$F9,$65
	.dc.b	$F9,$9E,$70,$FE,$7B,$FB,$6B,$F3,$1F,$F6,$CF,$EA,$6E,$4F,$F0,$A7
	.dc.b	$E6,$7F,$2B,$B1,$A7,$FB,$C7,$0E,$1F,$80,$B0,$E1,$F5,$0C,$3F,$BF
	.dc.b	$E1,$C3,$3E,$F9,$BF,$51,$F6,$C6,$31,$AB,$2E,$B9,$74,$C7,$D1,$39
	.dc.b	$C3,$CF,$A7,$F8,$A7,$DB,$CE,$1C,$23,$D3,$D9,$F3,$7B,$DF,$73,$FC
	.dc.b	$3F,$EF,$1F,$97,$7F,$C0,$0D,$75,$39,$87,$DA,$BF,$1F,$CF,$AD,$E6
	.dc.b	$C7,$9F,$C5,$DC,$38,$7D,$1F,$9A,$7B,$D3,$4F,$F4,$DE,$5F,$EC,$2F
	.dc.b	$CE,$7F,$B6,$F1,$7F,$74,$FC,$4F,$F3,$55,$FD,$DB,$C6,$7E,$54,$E9
	.dc.b	$61,$C3,$36,$1C,$1F,$F1,$6F,$E5,$7B,$9D,$4E,$70,$FE,$F7,$7C,$B2
	.dc.b	$FD,$6C,$F1,$FC,$BF,$3F,$25,$D7,$F1,$DF,$D8,$6F,$E9,$DB,$F9,$AF
	.dc.b	$E1,$DE,$3A,$D7,$9F,$FC,$E6,$7F,$9C,$F4,$67,$3F,$DC,$E3,$CA,$7E
	.dc.b	$17,$6F,$69,$CD,$F1,$FC,$7B,$CB,$F9,$27,$B6,$7F,$80,$4F,$87,$81
	.dc.b	$F5,$7F,$0E,$B5,$3F,$22,$7F,$85,$F0,$A7,$33,$C6,$F8,$1F,$C3,$0E
	.dc.b	$B9,$7D,$5E,$0F,$8B,$FA,$BF,$87,$A7,$D1,$3F,$EE,$5C,$53,$9C,$3E
	.dc.b	$06,$FF,$56,$31,$C5,$FC,$77,$2C,$1D,$F1,$F8,$5F,$79,$F4,$BF,$6F
	.dc.b	$F7,$72,$F8,$B4,$FD,$74,$63,$3F,$C1,$C7,$E0,$63,$11,$D7,$44,$62
	.dc.b	$7F,$D5,$B8,$FE,$1F,$DD,$F9,$5C,$9F,$2B,$4F,$B5,$87,$F3,$AA,$F8
	.dc.b	$53,$9B,$FD,$F5,$F0,$C7,$F0,$3C,$3F,$88,$8C,$47,$F8,$B7,$4F,$C5
	.dc.b	$FE,$BD,$F5,$F1,$FA,$F3,$DA,$30,$FC,$EF,$5F,$F3,$51,$88,$FE,$79
	.dc.b	$87,$2A,$FA,$87,$2B,$FC,$BF,$EC,$5D,$78,$3C,$DA,$A3,$18,$7E,$5B
	.dc.b	$2F,$D3,$46,$21,$EF,$FE,$83,$E9,$FC,$DC,$5E,$6D,$FF,$59,$18,$AF
	.dc.b	$C7,$E1,$FA,$FF,$DC,$BD,$9C,$3E,$CE,$2C,$6F,$E6,$8F,$7F,$86,$FC
	.dc.b	$91,$7C,$31,$FA,$11,$8B,$E1,$1F,$9C,$AE,$19,$78,$63,$FA,$D5,$7F
	.dc.b	$A4,$6B,$E1,$E9,$F3,$33,$ED,$B7,$B7,$4F,$E1,$97,$F0,$EF,$E4,$F0
	.dc.b	$7D,$FB,$DC,$63,$1F,$ED,$5B,$FF,$9F,$2D,$C7,$F4,$E7,$F7,$D1,$8A
	.dc.b	$FA,$71,$F6,$63,$11,$FD,$90,$FB,$FC,$7C,$EB,$97,$1F,$5E,$4F,$C3
	.dc.b	$E9,$D1,$BC,$7F,$DD,$31,$FC,$96,$3F,$DB,$B8,$FF,$2D,$97,$18,$FC
	.dc.b	$AE,$7F,$C4,$B2,$F4,$72,$FE,$7F,$1E,$4D,$7E,$93,$FA,$3B,$DD,$FE
	.dc.b	$FF,$87,$F7,$F7,$F0,$79,$BF,$6F,$87,$F3,$5E,$EF,$B5,$18,$87,$46
	.dc.b	$30,$FD,$D6,$7F,$E5,$7C,$B9,$46,$37,$FE,$7D,$97,$CD,$EA,$FE,$23
	.dc.b	$8F,$F8,$7F,$AF,$F0,$D1,$88,$FF,$09,$70,$7F,$17,$C3,$F8,$A6,$5F
	.dc.b	$C6,$DD,$98,$7F,$99,$75,$FD,$9E,$EF,$E6,$39,$FF,$35,$F8,$D7,$07
	.dc.b	$E4,$E7,$39,$7F,$99,$B7,$C6,$5F,$CC,$30,$FE,$79,$DF,$CE,$31,$5F
	.dc.b	$D5,$FB,$FF,$0D,$E0,$7F,$4A,$D7,$FA,$17,$85,$FE,$FA,$F0,$FF,$41
	.dc.b	$39,$BF,$F6,$97,$23,$C3,$CE,$5F,$D3,$B1,$E3,$3D,$23,$15,$FE,$D9
	.dc.b	$E1,$FC,$A7,$A1,$FD,$77,$BF,$FB,$97,$91,$FE,$0D,$E5,$FF,$05,$61
	.dc.b	$FE,$DA,$9C,$F9,$1F,$B1,$BE,$1C,$5F,$DB,$2B,$FB,$6D,$F0,$E0,$FE
	.dc.b	$EF,$19,$38,$1F,$E4,$EC,$3F,$BC,$B8,$3B,$57,$F7,$9E,$3F,$F0,$3F
	.dc.b	$E8,$BE,$69,$FF,$0C,$79,$5F,$E1,$5D,$B8,$79,$9F,$E1,$1F,$47,$FC
	.dc.b	$45,$E2,$E3,$CF,$FF,$10,$61,$E1,$DB,$D7,$7F,$E2,$A8,$C6,$FF,$E2
	.dc.b	$DA,$FF,$0E,$E1,$FE,$2B,$9E,$5E,$87,$F8,$BF,$CF,$FF,$10,$46,$34
	.dc.b	$FF,$13,$72,$FA,$7D,$3F,$F2,$B7,$A9,$FE,$41,$F7,$3F,$CA,$9B,$7C
	.dc.b	$A9,$CF,$B7,$FE,$52,$8C,$7B,$3F,$E5,$1C,$BC,$67,$F9,$67,$D9,$F9
	.dc.b	$0F,$58,$FF,$2D,$E3,$FE,$4D,$F6,$7E,$24,$FF,$96,$AB,$FC,$A9,$EC
	.dc.b	$FF,$0F,$F8,$5F,$F0,$C7,$7F,$F9,$EB,$DE,$7C,$0B,$F0,$6F,$E5,$55
	.dc.b	$AE,$AF,$F3,$B5,$6B,$E0,$B3,$9F,$F9,$F6,$7C,$7B,$FF,$A1,$E3,$FD
	.dc.b	$1B,$D7,$8C,$39,$72,$7F,$9E,$F9,$BC,$55,$C6,$3F,$CF,$DE,$06,$21
	.dc.b	$F1,$3F,$EB,$2F,$86,$FF,$4D,$7C,$5B,$E1,$BC,$35,$AF,$88,$FF,$4F
	.dc.b	$FB,$1F,$EB,$6A,$7E,$11,$AC,$7F,$D5,$F7,$F9,$3D,$FC,$D0,$C3,$FC
	.dc.b	$F7,$39,$DB,$FD,$5D,$8F,$FB,$03,$E2,$BC,$97,$E3,$8F,$F5,$85,$75
	.dc.b	$BF,$0F,$C5,$7F,$AF,$F1,$FF,$72,$71,$7F,$B6,$BE,$47,$FD,$D1,$F2
	.dc.b	$3F,$D5,$7E,$63,$FC,$9F,$F2,$7F,$ED,$AF,$93,$FF,$69,$FC,$96,$74
	.dc.b	$FF,$6F,$79,$DF,$EA,$87,$F2,$D7,$C2,$3F,$DD,$11,$EB,$8C,$65,$DF
	.dc.b	$E5,$3F,$A6,$46,$3D,$BF,$F7,$35,$75,$3F,$DC,$D8,$FF,$B8,$BE,$FB
	.dc.b	$19,$F8,$F5,$FF,$7C,$EC,$C7,$FC,$CB,$F3,$2F,$8E,$7C,$EB,$E6,$9D
	.dc.b	$8F,$9B,$7C,$DB,$BF,$C1,$8F,$1C,$38,$5E,$E9,$AE,$B2,$7C,$E7,$E3
	.dc.b	$34,$F0,$C7,$83,$87,$E1,$81,$97,$5D,$7C,$7A,$B7,$F8,$75,$BA,$D6
	.dc.b	$9E,$7E,$3C,$CF,$2C,$7D,$B8,$F2,$D5,$F3,$BF,$27,$E9,$1F,$44,$FA
	.dc.b	$37,$D1,$7C,$7F,$A2,$7C,$47,$FB,$9F,$9F,$E5,$6B,$E7,$C7,$8E,$1C
	.dc.b	$38,$D8,$70,$EC,$ED,$9B,$E9,$3E,$16,$3C,$63,$C0,$FC,$21,$8F,$4D
	.dc.b	$3C,$5F,$17,$EC,$8C,$64,$FA,$4F,$E6,$6F,$FA,$82,$33,$9F,$A3,$E5
	.dc.b	$FF,$C5,$5F,$82,$7F,$A9,$3E,$AF,$FD,$51,$DC,$FA,$97,$D5,$BE,$A9
	.dc.b	$F5,$AF,$AB,$E3,$C3,$F5,$71,$E3,$87,$0B,$DF,$D6,$EA,$95,$D9,$FE
	.dc.b	$69,$9F,$87,$C8,$FC,$2C,$60,$E8,$FE,$37,$0E,$9F,$4D,$FE,$2B,$AE
	.dc.b	$1A,$F9,$F5,$F3,$7D,$6E,$71,$E3,$39,$D3,$D3,$F9,$8F,$8F,$EC,$9F
	.dc.b	$46,$FB,$47,$D9,$BE,$DF,$E5,$7D,$9B,$F2,$6E,$17,$BA,$D7,$5A,$9F
	.dc.b	$73,$AF,$B8,$FE,$4F,$C5,$FE,$D0,$D3,$BE,$98,$E0,$FC,$E4,$63,$2E
	.dc.b	$7F,$71,$E9,$94,$E7,$6F,$B9,$F7,$19,$8C,$6B,$E1,$FB,$9F,$9D,$F7
	.dc.b	$6F,$88,$7D,$FB,$E1,$9F,$7C,$FB,$DF,$67,$D4,$F0,$BD,$D6,$A7,$E0
	.dc.b	$7A,$F9,$C3,$D1,$BB,$2E,$DF,$7F,$87,$C1,$3D,$F3,$ED,$F7,$FE,$B9
	.dc.b	$CE,$5D,$7E,$DF,$1A,$78,$2F,$C9,$D6,$FC,$03,$F1,$1F,$82,$3F,$06
	.dc.b	$7E,$08,$FA,$1F,$B1,$F0,$9F,$97,$FC,$17,$75,$A9,$CF,$92,$FC,$17
	.dc.b	$38,$7A,$B1,$E9,$EC,$DC,$38,$78,$0C,$FF,$8E,$73,$70,$CF,$BF,$E0
	.dc.b	$DF,$57,$1F,$26,$5E,$3C,$38,$68,$FC,$CF,$84,$63,$F2,$1B,$9F,$E5
	.dc.b	$FD,$D7,$8B,$FC,$77,$57,$7B,$EB,$9F,$84,$FF,$77,$FC,$29,$F7,$0F
	.dc.b	$C3,$7F,$99,$65,$CB,$1F,$E4,$D8,$F1,$C3,$85,$EF,$E6,$FF,$65,$3D
	.dc.b	$7F,$0D,$FB,$6E,$EF,$F6,$07,$8B,$3F,$0F,$D0,$34,$F0,$FE,$1B,$F9
	.dc.b	$5B,$ED,$39,$C9,$F1,$99,$C3,$8E,$1E,$8D,$BF,$BB,$EC,$FC,$6B,$E0
	.dc.b	$7D,$1B,$A7,$FA,$1F,$8E,$FA,$BF,$BF,$E2,$BF,$11,$FF,$60,$61,$C2
	.dc.b	$F7,$5A,$FC,$43,$3C,$7F,$D1,$F2,$E9,$F8,$8F,$F8,$63,$0E,$1F,$2D
	.dc.b	$6B,$58,$F2,$EF,$69,$FE,$B7,$BD,$EB,$E3,$97,$99,$9C,$38,$E3,$FA
	.dc.b	$ED,$7C,$FF,$6F,$E2,$97,$D8,$7F,$A7,$3B,$5F,$8A,$FF,$84,$FF,$14
	.dc.b	$7E,$2C,$FC,$57,$85,$EE,$B5,$2F,$C5,$B3,$F7,$9F,$7F,$2F,$F0,$F7
	.dc.b	$0F,$E2,$70,$FF,$54,$BE,$D8,$F0,$CF,$E3,$B0,$F0,$76,$7A,$67,$3A
	.dc.b	$79,$3E,$99,$AB,$C1,$D6,$7E,$39,$CE,$AF,$D2,$7C,$78,$71,$C3,$D7
	.dc.b	$F4,$FF,$07,$F1,$AF,$F5,$2F,$E3,$2F,$C6,$9F,$8D,$BF,$12,$7E,$31
	.dc.b	$9F,$0F,$F0,$9C,$3F,$DF,$F1,$E3,$87,$0B,$DF,$C2,$75,$FC,$2D,$F8
	.dc.b	$E7,$BE,$FB,$E1,$FE,$7F,$E8,$FD,$64,$72,$D3,$C1,$A7,$C1,$8C,$64
	.dc.b	$FC,$07,$19,$36,$F4,$61,$C3,$F1,$DF,$04,$62,$FD,$3F,$9C,$7D,$FF
	.dc.b	$C6,$BF,$98,$61,$F2,$7E,$13,$FB,$3F,$E4,$2F,$C8,$75,$F0,$BF,$EC
	.dc.b	$8E,$DF,$E5,$5F,$3C,$A7,$E4,$3A,$C7,$ED,$65,$FC,$EF,$E0,$2E,$3C
	.dc.b	$73,$FE,$E7,$E6,$7D,$BA,$F0,$E7,$D3,$F0,$BC,$5E,$F4,$ED,$F9,$17
	.dc.b	$5A,$F0,$EB,$F8,$0E,$CF,$A3,$D3,$FB,$1F,$C9,$9F,$93,$FD,$8F,$77
	.dc.b	$FC,$09,$F0,$0F,$C9,$BF,$3D,$E7,$AF,$7D,$7F,$E7,$9F,$84,$5B,$F2
	.dc.b	$6D,$D7,$4C,$FE,$B1,$FE,$3D,$C3,$97,$CF,$FE,$65,$E9,$A7,$87,$1E
	.dc.b	$75,$CA,$31,$E0,$FE,$6E,$B7,$9F,$87,$37,$E3,$3B,$8C,$7E,$4F,$77
	.dc.b	$F5,$FE,$49,$74,$F3,$FC,$B7,$EE,$7F,$2A,$FF,$00,$7D,$CF,$F3,$3F
	.dc.b	$C7,$9F,$95,$3F,$2C,$73,$EF,$E5,$BF,$CA,$BF,$10,$BD,$D3,$F2,$BD
	.dc.b	$7E,$59,$FB,$5C,$FA,$F3,$FC,$20,$F8,$30,$E5,$9B,$F1,$FF,$DC,$4E
	.dc.b	$72,$F1,$FD,$B6,$34,$F0,$C7,$83,$F2,$8F,$26,$FE,$9C,$7D,$1E,$87
	.dc.b	$F1,$5F,$B7,$7E,$54,$FC,$77,$FC,$9B,$C6,$FC,$B9,$F8,$EB,$F2,$EE
	.dc.b	$17,$BF,$BB,$F8,$35,$AF,$D1,$F3,$F9,$87,$F1,$1C,$FA,$E7,$70,$E1
	.dc.b	$9F,$59,$FE,$24,$6B,$1F,$97,$F8,$EF,$F6,$39,$FC,$9C,$E5,$F2,$68
	.dc.b	$FC,$CB,$C7,$8F,$A3,$4F,$3F,$E6,$5F,$BB,$E9,$FE,$D9,$F8,$C3,$F3
	.dc.b	$67,$E4,$CF,$C7,$5F,$9A,$BE,$EF,$F3,$9F,$9B,$71,$C7,$FD,$21,$8F
	.dc.b	$1C,$38,$5E,$FD,$27,$E6,$DB,$B5,$ED,$FE,$F8,$7F,$A6,$7A,$6C,$78
	.dc.b	$FF,$2B,$FF,$7E,$C6,$35,$F2,$EB,$E4,$D7,$FC,$09,$39,$D3,$C3,$1F
	.dc.b	$E9,$FC,$7C,$18,$F9,$6F,$86,$6F,$C0,$F3,$97,$4D,$7D,$1F,$90,$79
	.dc.b	$FF,$39,$7E,$73,$FE,$8C,$FC,$D7,$FA,$77,$E7,$2F,$AC,$FD,$CF,$0F
	.dc.b	$AA,$F7,$DA,$BD,$D7,$F8,$7E,$B5,$87,$FA,$1F,$83,$E0,$9D,$DF,$F4
	.dc.b	$CB,$88,$DD,$F3,$F9,$C6,$5F,$E0,$3A,$E1,$39,$87,$E5,$DF,$2E,$7F
	.dc.b	$C5,$19,$F9,$33,$F1,$FE,$11,$C1,$F2,$FE,$3D,$F5,$7E,$35,$F9,$DF
	.dc.b	$FA,$73,$B9,$F6,$AF,$CE,$DF,$90,$BF,$43,$E1,$7B,$AD,$4F,$F1,$FE
	.dc.b	$2F,$BB,$8F,$4C,$BE,$DE,$DF,$E3,$B8,$EF,$AE,$23,$FD,$B5,$18,$D3
	.dc.b	$C3,$A7,$F1,$CC,$39,$E7,$DE,$3F,$B2,$DF,$8B,$0F,$DB,$67,$E2,$FC
	.dc.b	$FB,$18,$F2,$D1,$F9,$F7,$E4,$FE,$B3,$FC,$23,$E7,$7E,$EF,$F2,$07
	.dc.b	$E6,$5F,$EF,$DF,$A1,$D7,$D3,$7E,$F7,$CA,$FF,$7E,$4F,$FC,$01,$F8
	.dc.b	$7C,$E1,$E9,$E0,$FF,$45,$DB,$CB,$C6,$7D,$B1,$E3,$F9,$07,$D5,$BF
	.dc.b	$F7,$C4,$E7,$4F,$0F,$C9,$E6,$79,$E5,$E5,$D1,$83,$F2,$5F,$0E,$5D
	.dc.b	$3F,$58,$B9,$7A,$A9,$F9,$B7,$EC,$7F,$0F,$7E,$14,$FD,$0F,$F9,$BF
	.dc.b	$D9,$BF,$42,$7E,$85,$FE,$F2,$EF,$FD,$AE,$1C,$2F,$75,$AF,$C3,$FA
	.dc.b	$63,$CF,$8B,$F3,$07,$F9,$DE,$FB,$FE,$87,$F1,$5E,$5F,$02,$D3,$C1
	.dc.b	$F9,$9F,$BE,$73,$1E,$3F,$CD,$71,$8F,$9E,$F8,$67,$E6,$F8,$DF,$0E
	.dc.b	$BE,$4D,$FF,$A7,$EF,$FC,$33,$CE,$7D,$83,$F4,$4F,$F3,$57,$77,$F8
	.dc.b	$EB,$E6,$1F,$9D,$FF,$15,$BD,$D6,$BC,$2F,$EE,$BF,$05,$9C,$79,$73
	.dc.b	$7B,$81,$FE,$51,$7F,$94,$39,$31,$5C,$F1,$FB,$98,$76,$FB,$EF,$F3
	.dc.b	$0E,$3F,$9A,$D3,$D1,$8F,$9E,$73,$6F,$CC,$FF,$95,$7D,$3B,$E0,$BF
	.dc.b	$B7,$FC,$17,$FB,$3F,$D3,$FE,$A7,$86,$A7,$3F,$51,$FB,$7E,$0F,$E3
	.dc.b	$1A,$7F,$30,$C7,$97,$3F,$E6,$5E,$9F,$A3,$DC,$38,$E4,$F8,$9F,$CA
	.dc.b	$57,$0D,$3F,$DB,$DE,$87,$F3,$2E,$4F,$F5,$B4,$E7,$5F,$07,$E8,$FE
	.dc.b	$3D,$3F,$A6,$C7,$8F,$F4,$7C,$63,$E9,$DD,$FA,$F6,$2B,$D1,$FA,$4F
	.dc.b	$F4,$9F,$CA,$BF,$C5,$9F,$83,$BD,$07,$C9,$BF,$4A,$79,$3F,$D3,$B9
	.dc.b	$FF,$A6,$E1,$C2,$F7,$5A,$FD,$2F,$A7,$C5,$9F,$DD,$57,$6E,$4F,$B4
	.dc.b	$F1,$CB,$B7,$CE,$7E,$D6,$31,$83,$E6,$F1,$8B,$5F,$36,$1E,$5D,$7F
	.dc.b	$C1,$5A,$F9,$2B,$5A,$F9,$25,$F2,$FF,$E6,$FF,$9A,$FF,$70,$7E,$99
	.dc.b	$FE,$8B,$FC,$AF,$FB,$A7,$DF,$7F,$75,$FC,$CD,$F8,$CE,$F8,$FF,$27
	.dc.b	$7B,$F9,$55,$EF,$F3,$0D,$52,$BA,$3F,$D3,$F2,$EF,$F8,$23,$F2,$DF
	.dc.b	$EE,$31,$8D,$3C,$39,$7F,$6D,$E1,$FF,$1D,$61,$E0,$C1,$5A,$FC,$23
	.dc.b	$FB,$E6,$1C,$70,$F1,$8D,$9F,$59,$E4,$D3,$D1,$B7,$B7,$DE,$F3,$FA
	.dc.b	$5F,$9F,$F2,$1F,$83,$3F,$4D,$7D,$57,$CF,$FC,$6F,$7E,$6F,$F5,$5A
	.dc.b	$D4,$FD,$3F,$47,$3E,$6F,$B3,$97,$F7,$39,$F5,$7E,$11,$8C,$39,$65
	.dc.b	$DB,$E7,$FD,$73,$9D,$3C,$73,$E2,$FC,$B3,$33,$FE,$DC,$BE,$39,$FE
	.dc.b	$0F,$E4,$7C,$B8,$70,$8F,$C3,$7C,$AE,$36,$F3,$F8,$6F,$C6,$F1,$A7
	.dc.b	$BA,$3F,$B8,$F6,$7D,$0D,$1F,$2C,$FD,$47,$F9,$FF,$EA,$1F,$A2,$FD
	.dc.b	$41,$F8,$4F,$E3,$30,$E1,$7B,$F8,$B5,$EE,$BF,$2B,$DF,$FD,$D7,$1F
	.dc.b	$E9,$39,$7C,$43,$BA,$FD,$47,$E5,$46,$30,$E7,$E4,$34,$FC,$8C,$FE
	.dc.b	$43,$3F,$06,$7E,$5F,$CA,$FE,$3C,$62,$F8,$68,$FC,$D7,$86,$9E,$C9
	.dc.b	$CF,$C6,$FE,$6B,$F5,$37,$BA,$FD,$35,$F9,$4B,$F5,$3F,$E3,$1F,$AA
	.dc.b	$3E,$07,$85,$EE,$B5,$F0,$4A,$7C,$0E,$B1,$F6,$E5,$DB,$CF,$B8,$F1
	.dc.b	$CF,$F9,$27,$C8,$7E,$02,$BC,$77,$E2,$9F,$E5,$1E,$5E,$6F,$86,$7D
	.dc.b	$FE,$13,$B6,$BE,$2D,$7F,$90,$F6,$7D,$DF,$C9,$1F,$A4,$FF,$9D,$7F
	.dc.b	$0F,$7E,$70,$FC,$35,$E6,$BF,$31,$FD,$EE,$1C,$3E,$C9,$87,$0B,$DF
	.dc.b	$EA,$BA,$CB,$FD,$01,$E0,$FD,$DC,$3B,$FD,$77,$DC,$79,$61,$C7,$9F
	.dc.b	$F3,$B1,$8D,$7C,$5D,$78,$C7,$CD,$1E,$5D,$BF,$88,$EF,$EB,$9C,$EB
	.dc.b	$E8,$FC,$D3,$C5,$BF,$9B,$0E,$1E,$0E,$35,$F4,$64,$FC,$DB,$F8,$CF
	.dc.b	$D5,$BF,$58,$FE,$3B,$F8,$03,$F3,$6F,$E3,$1D,$2F,$D6,$37,$5A,$9C
	.dc.b	$FE,$B1,$8E,$EF,$CC,$E3,$CF,$F2,$0B,$7B,$F8,$57,$B5,$FB,$CA,$F0
	.dc.b	$5B,$4F,$05,$7F,$9C,$F9,$BF,$69,$5A,$9C,$EB,$F4,$A7,$38,$F5,$87
	.dc.b	$E4,$9F,$CD,$3F,$08,$FF,$5C,$7E,$23,$FE,$6F,$FB,$1F,$F5,$A7,$E5
	.dc.b	$1F,$DC,$71,$FE,$01,$5A,$FB,$65,$6B,$F5,$1E,$5F,$82,$B1,$FD,$AD
	.dc.b	$FF,$B0,$FF,$14,$B3,$9C,$38,$F6,$3F,$59,$FE,$33,$7D,$72,$FF,$62
	.dc.b	$7E,$61,$8C,$38,$47,$93,$E7,$FB,$E7,$E7,$C5,$B7,$F8,$5F,$6F,$E2
	.dc.b	$DF,$A1,$3F,$47,$FF,$75,$FD,$C3,$F2,$47,$E4,$FF,$BA,$FC,$03,$FA
	.dc.b	$AD,$EF,$F0,$FF,$F7,$65,$6B,$D4,$70,$EB,$2F,$F7,$AF,$9F,$FC,$57
	.dc.b	$0E,$D3,$FC,$51,$E5,$9F,$7C,$FF,$83,$D7,$1C,$BA,$C3,$F2,$DF,$E7
	.dc.b	$37,$C3,$3F,$07,$E1,$DB,$C7,$96,$9F,$ED,$CD,$BD,$BA,$FB,$2F,$CD
	.dc.b	$E1,$ED,$F4,$7E,$74,$A7,$E3,$DF,$E4,$8D,$FF,$C2,$1F,$A8,$BF,$41
	.dc.b	$7E,$B4,$FC,$DF,$FC,$07,$F8,$0F,$F8,$E3,$0E,$1F,$58,$C3,$85,$EF
	.dc.b	$E9,$56,$FB,$2F,$06,$1C,$2D,$F9,$3F,$FB,$B2,$31,$9D,$D3,$FC,$B7
	.dc.b	$9F,$58,$F1,$E9,$FE,$14,$F8,$FF,$F7,$B5,$F0,$D7,$C1,$F5,$D8,$C7
	.dc.b	$8E,$7E,$4C,$3D,$3F,$11,$E2,$9C,$E5,$D3,$7F,$3F,$C4,$E6,$7F,$D2
	.dc.b	$3F,$88,$BF,$4E,$7E,$41,$FD,$FF,$EC,$DF,$AC,$BF,$52,$E1,$F7,$EC
	.dc.b	$38,$5E,$E9,$F9,$2E,$A7,$A6,$5F,$EE,$DE,$3F,$A5,$87,$1F,$13,$E1
	.dc.b	$4F,$8B,$4F,$0F,$CE,$35,$7D,$6F,$BE,$B5,$AF,$97,$F0,$4C,$65,$E3
	.dc.b	$C7,$8E,$1E,$8F,$7B,$C5,$18,$DD,$FA,$CF,$C1,$F7,$5F,$A9,$3F,$5B
	.dc.b	$F4,$FC,$EF,$E9,$BD,$B7,$ED,$3B,$AD,$73,$7F,$7A,$F3,$FF,$85,$FE
	.dc.b	$19,$9C,$7D,$79,$74,$AB,$7C,$32,$FD,$8F,$E2,$FE,$BA,$F0,$DF,$83
	.dc.b	$17,$93,$9C,$38,$47,$F1,$8F,$D2,$7B,$65,$E1,$D5,$F8,$CE,$34,$F4
	.dc.b	$7E,$BB,$70,$FE,$3F,$FB,$1F,$DF,$FA,$C7,$EA,$CF,$D1,$DF,$8E,$BF
	.dc.b	$5F,$7E,$68,$F8,$87,$E6,$5C,$70,$E1,$7B,$AD,$7C,$36,$7E,$E7,$ED
	.dc.b	$E7,$DB,$3B,$87,$0C,$FA,$FD,$1F,$DB,$AD,$65,$CF,$E0,$31,$7B,$CF
	.dc.b	$C5,$FB,$07,$87,$2F,$3E,$5E,$6D,$BF,$C2,$7E,$D7,$EF,$BC,$7F,$F6
	.dc.b	$2F,$D2,$FF,$A7,$3F,$54,$FE,$F3,$F9,$7F,$EF,$FF,$09,$FF,$41,$F9
	.dc.b	$FF,$C0,$EB,$5F,$3A,$AD,$7B,$1F,$A5,$9F,$F5,$2F,$EC,$3F,$96,$7A
	.dc.b	$65,$F9,$B3,$BF,$E1,$AC,$FB,$FB,$0F,$B2,$FC,$C5,$F6,$CB,$F8,$87
	.dc.b	$8F,$8C,$38,$69,$E1,$C3,$FD,$59,$F6,$3E,$19,$CD,$FA,$7F,$62,$FD
	.dc.b	$C7,$E3,$CF,$C2,$FF,$CB,$1F,$98,$3F,$5F,$FD,$CF,$E9,$0F,$C6,$37
	.dc.b	$5A,$9C,$FE,$89,$8F,$D8,$BF,$63,$8F,$3F,$C0,$2D,$EF,$DD,$7E,$77
	.dc.b	$FB,$1A,$D6,$9D,$E7,$F6,$BC,$FD,$B0,$4E,$75,$FC,$67,$67,$D5,$A6
	.dc.b	$1C,$B5,$F3,$FE,$35,$C7,$0E,$1F,$AF,$59,$F5,$FA,$1F,$C5,$FC,$DF
	.dc.b	$CA,$7E,$C8,$FD,$91,$F9,$57,$83,$F2,$0F,$EE,$5B,$77,$BD,$D6,$B3
	.dc.b	$57,$FA,$6A,$B5,$E2,$B1,$FE,$27,$87,$E8,$7E,$D2,$CE,$71,$E5,$F4
	.dc.b	$BF,$C4,$EB,$AE,$7F,$03,$EF,$B1,$7B,$8F,$27,$E5,$5E,$98,$C6,$BE
	.dc.b	$2F,$D4,$FF,$99,$7E,$C7,$FE,$67,$FC,$03,$FC,$D1,$F9,$5B,$F5,$27
	.dc.b	$E0,$3F,$E5,$1E,$2F,$95,$7B,$F9,$45,$EE,$BD,$55,$A9,$FF,$37,$75
	.dc.b	$7E,$5C,$5F,$F9,$D7,$EF,$7F,$98,$BC,$F0,$F0,$79,$5F,$CB,$E3,$19
	.dc.b	$3F,$00,$C6,$3F,$E7,$93,$C3,$1E,$6F,$DD,$3D,$71,$9C,$3C,$9F,$3F
	.dc.b	$FD,$CF,$C4,$FE,$E5,$FA,$FB,$F5,$AF,$F0,$A7,$D5,$39,$BF,$D0,$7D
	.dc.b	$EF,$D9,$B8,$71,$06,$1C,$7E,$99,$87,$1F,$9F,$FE,$47,$F8,$AF,$0B
	.dc.b	$DE,$7F,$E2,$07,$FC,$04,$73,$E6,$FD,$53,$D7,$1E,$8E,$8E,$4F,$7C
	.dc.b	$3B,$7C,$97,$1D,$BD,$78,$7A,$B6,$FF,$4D,$FE,$43,$9C,$7A,$4E,$74
	.dc.b	$FE,$6F,$F6,$DD,$23,$1C,$0F,$D9,$F1,$C1,$EC,$CB,$E1,$F1,$E3,$6F
	.dc.b	$75,$BF,$50,$F7,$7A,$EF,$92,$FC,$D7,$E1,$FF,$89,$F9,$37,$E4,$8F
	.dc.b	$CF,$F7,$5A,$ED,$FE,$EF,$3A,$E8,$F8,$7C,$9F,$EC,$3A,$FD,$23,$F0
	.dc.b	$E3,$39,$77,$F3,$3E,$34,$F4,$F3,$7F,$DC,$98,$FF,$80,$BE,$EF,$E1
	.dc.b	$4F,$39,$CE,$9E,$F8,$E1,$AF,$9F,$4F,$DD,$E7,$DB,$4F,$35,$FA,$BE
	.dc.b	$CF,$EE,$78,$CF,$DA,$5F,$A3,$BF,$69,$FE,$23,$F8,$72,$FF,$D1,$19
	.dc.b	$7B,$30,$E1,$7B,$F3,$57,$BF,$B6,$EB,$97,$4D,$FE,$07,$83,$FE,$49
	.dc.b	$C7,$C1,$FA,$3F,$C0,$CF,$A5,$F8,$B5,$F0,$F1,$FC,$0B,$EF,$1E,$4E
	.dc.b	$EC,$56,$F5,$F0,$EE,$FD,$53,$C3,$C1,$E5,$E0,$F5,$7E,$6D,$F9,$BF
	.dc.b	$DA,$9F,$97,$3F,$64,$7E,$64,$FC,$EF,$FD,$EB,$F2,$2F,$D0,$78,$5E
	.dc.b	$EB,$5F,$80,$29,$FA,$D6,$B1,$F6,$74,$7F,$99,$F3,$EF,$FB,$1E,$3A
	.dc.b	$BF,$65,$1E,$2A,$F0,$EA,$C9,$D1,$F8,$0A,$D6,$9E,$0F,$CA,$F1,$B7
	.dc.b	$8F,$0F,$3F,$9D,$E1,$AF,$2E,$EF,$57,$15,$EC,$FC,$36,$E9,$E9,$BF
	.dc.b	$DF,$7C,$9B,$F2,$07,$DD,$7E,$97,$F1,$2F,$F5,$2F,$E4,$8F,$D6,$B7
	.dc.b	$5A,$F3,$D5,$AF,$D6,$73,$87,$BF,$9B,$FD,$0D,$7F,$EA,$27,$F0,$67
	.dc.b	$F1,$3C,$BC,$13,$DF,$F0,$DF,$F4,$77,$8D,$FA,$A9,$CE,$5D,$A5,$F7
	.dc.b	$BF,$A4,$9C,$E7,$E6,$FC,$87,$C3,$7B,$DB,$FB,$8E,$3F,$B4,$FC,$17
	.dc.b	$FC,$A7,$FA,$67,$FB,$27,$F5,$D7,$E3,$AF,$BB,$7E,$49,$FC,$86,$F7
	.dc.b	$F2,$4B,$DD,$3F,$17,$D7,$DF,$32,$F8,$18,$7F,$98,$3F,$60,$B3,$9C
	.dc.b	$79,$67,$DF,$97,$8E,$1E,$38,$FF,$7A,$78,$DF,$EA,$FC,$7C,$3A,$78
	.dc.b	$3F,$6B,$DE,$3C,$B5,$7C,$97,$F3,$7F,$9A,$7E,$4E,$FC,$EB,$FB,$F7
	.dc.b	$95,$EC,$FC,$B1,$FB,$66,$EB,$5F,$39,$FA,$3E,$65,$7E,$AF,$9F,$F8
	.dc.b	$DE,$5D,$32,$FB,$8F,$6F,$4A,$E3,$CB,$3F,$93,$F9,$97,$F1,$8A,$E1
	.dc.b	$9F,$5F,$C8,$B1,$7B,$D3,$BF,$DC,$78,$F3,$F3,$C7,$8B,$F2,$AC,$61
	.dc.b	$E5,$D9,$FB,$9E,$35,$F6,$7E,$97,$FD,$5F,$B7,$FD,$EB,$D7,$FE,$E5
	.dc.b	$FB,$BF,$F2,$67,$E3,$9F,$EE,$1E,$CF,$F6,$D5,$6B,$EF,$7F,$CE,$FF
	.dc.b	$6E,$A7,$ED,$DC,$FF,$05,$5F,$E6,$CE,$BE,$17,$E6,$8C,$C7,$E0,$70
	.dc.b	$E7,$F8,$57,$F3,$8A,$ED,$9F,$F9,$77,$F1,$EC,$5E,$E3,$CB,$FB,$7F
	.dc.b	$7D,$BD,$1A,$F8,$FF,$0B,$FF,$79,$7D,$9F,$FB,$C7,$AD,$F8,$33,$F7
	.dc.b	$17,$EC,$4F,$DC,$5F,$92,$BF,$37,$FC,$2F,$8B,$FE,$D4,$E5,$FF,$01
	.dc.b	$65,$CB,$1E,$38,$70,$BF,$F0,$55,$EE,$D7,$B7,$FA,$D6,$7E,$3C,$DF
	.dc.b	$A2,$7F,$3B,$8C,$65,$D3,$37,$C2,$7F,$81,$E7,$3A,$F9,$E7,$FD,$BF
	.dc.b	$8F,$B8,$F2,$E3,$ED,$9D,$69,$FE,$DB,$D3,$C9,$A7,$A6,$73,$F3,$7C
	.dc.b	$70,$7E,$9F,$FE,$B5,$FD,$BF,$FD,$65,$F2,$1F,$E7,$DE,$CF,$F7,$1E
	.dc.b	$5E,$DA,$D7,$D9,$FF,$5F,$FB,$8F,$C0,$FC,$B7,$39,$FF,$5F,$7E,$DD
	.dc.b	$ED,$CF,$FC,$55,$1D,$5F,$86,$63,$F0,$C6,$9E,$0C,$3F,$90,$65,$DE
	.dc.b	$73,$97,$7F,$03,$E2,$61,$C7,$0E,$1A,$3E,$13,$C5,$87,$0D,$BD,$5F
	.dc.b	$0B,$8B,$DF,$EE,$57,$3F,$56,$2F,$DE,$3E,$27,$DA,$3F,$1B,$FF,$6B
	.dc.b	$F8,$9F,$EA,$DF,$DC,$9F,$8E,$7F,$20,$E2,$FC,$05,$EF,$F1,$35,$EE
	.dc.b	$BF,$90,$56,$B0,$FF,$70,$F1,$7C,$23,$FC,$BC,$FF,$2C,$71,$87,$3B
	.dc.b	$F0,$7E,$41,$F0,$27,$78,$F6,$FC,$3D,$9F,$F3,$E9,$F4,$E3,$E8,$9C
	.dc.b	$E7,$FE,$7B,$F0,$3F,$CB,$95,$CB,$2F,$37,$D4,$FE,$0F,$F1,$27,$E7
	.dc.b	$4F,$75,$F9,$EB,$F0,$3F,$F7,$D7,$CC,$3F,$74,$E1,$7B,$AD,$76,$7F
	.dc.b	$6B,$FB,$6C,$FE,$89,$F8,$3D,$9A,$76,$D3,$0D,$C3,$8F,$E5,$BE,$F8
	.dc.b	$C6,$AC,$BF,$5B,$87,$83,$0F,$2C,$F9,$0C,$61,$E2,$D7,$F9,$EF,$47
	.dc.b	$ED,$7E,$2B,$FD,$B5,$E0,$BF,$4A,$FF,$0E,$FB,$CF,$BC,$7E,$B5,$FA
	.dc.b	$5B,$DD,$6A,$7F,$90,$61,$CF,$DE,$FC,$86,$DF,$6E,$FF,$DC,$DF,$86
	.dc.b	$FB,$DE,$3F,$47,$FC,$CB,$0F,$B5,$18,$CF,$FC,$35,$7D,$F0,$FC,$D6
	.dc.b	$5D,$FF,$3A,$C6,$BE,$7C,$B9,$7E,$65,$8A,$F2,$7E,$F3,$67,$B7,$C8
	.dc.b	$FE,$F7,$F5,$47,$82,$FC,$41,$FB,$17,$FB,$A3,$F7,$57,$E3,$EF,$D2
	.dc.b	$B8,$5E,$EB,$52,$FD,$2F,$3F,$91,$7D,$BC,$BA,$E5,$70,$E1,$F6,$EC
	.dc.b	$FF,$C2,$BE,$47,$E4,$2F,$79,$F8,$7E,$9B,$F3,$91,$8B,$E1,$A7,$F9
	.dc.b	$BF,$AF,$F9,$64,$E7,$1F,$3F,$A3,$FC,$67,$F1,$CF,$E1,$3F,$A3,$3F
	.dc.b	$14,$FF,$48,$FC,$B3,$F1,$27,$E7,$DB,$F6,$97,$BF,$C1,$17,$BF,$2F
	.dc.b	$5F,$A2,$7E,$5F,$3E,$BE,$17,$ED,$DE,$FA,$76,$FA,$DF,$DA,$46,$31
	.dc.b	$E7,$F3,$B8,$D3,$BD,$F9,$36,$7E,$BF,$E3,$DB,$C5,$B7,$9F,$F7,$6F
	.dc.b	$E6,$9F,$94,$7F,$A2,$FF,$17,$7D,$43,$F3,$3F,$F4,$4F,$E8,$4F,$D0
	.dc.b	$37,$F1,$0B,$DD,$6B,$EA,$B3,$1F,$B1,$C7,$FD,$2D,$B5,$BE,$19,$75
	.dc.b	$FA,$A7,$6F,$F0,$FB,$F1,$4E,$7E,$A3,$17,$C3,$3F,$1F,$C0,$F6,$AE
	.dc.b	$5B,$3E,$93,$15,$E8,$FC,$82,$E8,$A7,$E9,$FF,$D4,$FF,$42,$7E,$B2
	.dc.b	$FD,$51,$F9,$63,$F3,$C7,$C2,$BF,$62,$7E,$EF,$C7,$0E,$17,$BA,$D7
	.dc.b	$E0,$B9,$C7,$F8,$96,$5D,$BD,$7B,$87,$0C,$FA,$FC,$F7,$F2,$FA,$F1
	.dc.b	$DF,$8A,$7C,$3F,$13,$9B,$E1,$97,$8F,$F5,$DE,$DA,$78,$76,$FF,$06
	.dc.b	$75,$7F,$AF,$7C,$9F,$E5,$9F,$BA,$FF,$A1,$7F,$78,$7C,$C3,$F5,$97
	.dc.b	$E3,$CF,$9B,$FA,$7F,$89,$7F,$38,$F9,$06,$3C,$70,$E1,$7F,$E7,$8A
	.dc.b	$D4,$FF,$A4,$72,$FF,$75,$46,$3E,$1B,$FD,$94,$EB,$E7,$5A,$78,$B4
	.dc.b	$7E,$48,$C3,$C2,$5D,$3C,$37,$C3,$4F,$47,$C6,$EF,$7F,$75,$6B,$2E
	.dc.b	$9B,$7F,$88,$27,$3B,$7A,$74,$F7,$FE,$4D,$DB,$47,$EE,$6F,$A1,$7E
	.dc.b	$D0,$FD,$01,$F9,$0F,$EF,$FD,$C7,$ED,$9C,$70,$E1,$7B,$AD,$7B,$59
	.dc.b	$C7,$9E,$5F,$10,$F7,$5F,$7F,$D5,$FE,$2C,$62,$7C,$39,$F8,$3D,$A7
	.dc.b	$E5,$DE,$FE,$9F,$ED,$78,$F8,$2B,$5F,$15,$9C,$79,$6B,$E3,$C5,$BB
	.dc.b	$F6,$8F,$24,$E6,$1F,$0D,$8C,$7C,$10,$FC,$1F,$1C,$3F,$07,$F6,$BF
	.dc.b	$91,$FA,$9F,$F8,$43,$F5,$97,$E2,$E9,$7E,$76,$FC,$69,$FB,$4F,$0F
	.dc.b	$6D,$87,$0B,$DD,$7F,$02,$EE,$FC,$C6,$5D,$78,$FE,$40,$7F,$C3,$FF
	.dc.b	$B7,$7C,$88,$C6,$3D,$34,$F1,$68,$F3,$BE,$25,$6B,$4F,$04,$FF,$7C
	.dc.b	$CB,$BE,$1C,$74,$F1,$FE,$47,$E7,$BD,$EB,$FE,$B4,$BF,$D6,$7D,$DF
	.dc.b	$FB,$B3,$F3,$8F,$EB,$BF,$A2,$7E,$27,$EE,$5F,$AC,$7E,$36,$B5,$F3
	.dc.b	$2A,$D7,$E0,$79,$FC,$6B,$FC,$05,$C3,$EE,$BE,$8D,$86,$5D,$BC,$37
	.dc.b	$DB,$FB,$2F,$AE,$3F,$99,$FC,$F7,$18,$70,$CF,$C1,$83,$57,$EF,$3E
	.dc.b	$1C,$FA,$47,$ED,$37,$FE,$11,$FA,$93,$F3,$27,$EB,$5C,$7C,$E7,$E4
	.dc.b	$CF,$CF,$5F,$21,$FA,$BC,$38,$5E,$EB,$5F,$B4,$67,$1E,$7C,$BF,$72
	.dc.b	$FB,$79,$DF,$6D,$E3,$AF,$F1,$5C,$1F,$80,$7D,$A9,$CE,$BE,$2B,$C6
	.dc.b	$3E,$6C,$38,$6B,$FE,$C4,$DB,$D7,$39,$BF,$4C,$BF,$0E,$F2,$6D,$E7
	.dc.b	$87,$E7,$D8,$C7,$9F,$EE,$BF,$C1,$FF,$58,$7E,$80,$FD,$DD,$FB,$37
	.dc.b	$FA,$73,$F0,$67,$C8,$3F,$56,$E1,$7B,$AD,$4B,$F3,$AC,$FE,$83,$F9
	.dc.b	$BC,$BA,$E5,$70,$E1,$F8,$6F,$F9,$C3,$F3,$2F,$CD,$D6,$B2,$F0,$7E
	.dc.b	$29,$8B,$DE,$BE,$2F,$4F,$C3,$97,$9B,$2F,$F3,$CE,$BF,$AF,$F7,$BF
	.dc.b	$CE,$BF,$65,$EC,$F8,$FF,$4F,$E6,$3F,$E5,$7F,$C2,$3F,$DB,$3D,$9F
	.dc.b	$D2,$EB,$58,$2B,$5F,$1B,$FC,$AE,$79,$78,$5F,$B8,$7E,$3F,$6F,$ED
	.dc.b	$DE,$35,$FE,$63,$FC,$1F,$F6,$B1,$8F,$BE,$7D,$2E,$38,$FF,$B9,$E7
	.dc.b	$DE,$5A,$FF,$AA,$75,$F2,$56,$B2,$F4,$7E,$45,$FC,$F7,$F1,$17,$E5
	.dc.b	$4F,$C1,$DF,$AF,$7F,$3F,$F9,$A7,$EE,$DF,$EF,$4C,$FF,$B4,$61,$C2
	.dc.b	$F7,$E8,$AF,$78,$F5,$1F,$97,$FF,$13,$BE,$1C,$BF,$C6,$1E,$59,$F7
	.dc.b	$FB,$EF,$E2,$71,$8C,$7A,$7B,$38,$C7,$CD,$7E,$5D,$7F,$DD,$9D,$5F
	.dc.b	$EA,$AC,$38,$69,$E8,$FB,$9F,$16,$DE,$68,$F5,$FC,$9E,$31,$E5,$E8
	.dc.b	$7F,$8B,$BF,$3C,$FF,$59,$FE,$8A,$FD,$09,$F8,$EF,$F9,$BF,$F6,$A7
	.dc.b	$E9,$0F,$D0,$FE,$47,$43,$1E,$38,$70,$FC,$25,$87,$0B,$FF,$72,$5E
	.dc.b	$FA,$3F,$BE,$67,$DF,$3F,$F6,$1C,$7F,$B1,$27,$4A,$7C,$98,$EA,$3F
	.dc.b	$BF,$65,$E0,$BF,$36,$6F,$C2,$F3,$5D,$AF,$C9,$9F,$AB,$E3,$73,$3D
	.dc.b	$67,$37,$ED,$FC,$73,$E2,$4F,$1C,$7A,$EA,$ED,$CC,$BF,$61,$FF,$6B
	.dc.b	$7E,$1C,$FB,$0F,$F6,$B4,$FC,$BF,$8C,$DD,$6A,$73,$A3,$F1,$5C,$FD
	.dc.b	$37,$A7,$0F,$77,$27,$C2,$C3,$AC,$63,$8F,$FD,$D3,$1C,$FC,$BF,$F1
	.dc.b	$36,$FE,$C8,$C7,$E6,$BF,$2E,$B7,$8B,$6F,$36,$1F,$E2,$DD,$3F,$C8
	.dc.b	$19,$F7,$AD,$6B,$E6,$F3,$FD,$FF,$8C,$FE,$73,$E5,$DF,$9F,$FE,$C3
	.dc.b	$F3,$77,$85,$FE,$B9,$CB,$EC,$DE,$EB,$53,$FE,$B9,$9C,$F9,$4C,$3E
	.dc.b	$E7,$1F,$6F,$8F,$AF,$82,$75,$87,$28,$C5,$78,$BC,$E6,$6F,$A6,$FB
	.dc.b	$BA,$78,$67,$3A,$3F,$5F,$E3,$97,$93,$F4,$1B,$5A,$FD,$B6,$E5,$EC
	.dc.b	$FB,$6F,$BF,$F8,$5B,$C7,$F4,$7D,$0B,$F2,$A7,$E3,$0F,$DE,$BF,$84
	.dc.b	$76,$FF,$96,$79,$18,$F1,$E4,$FF,$24,$E1,$C2,$DF,$AA,$76,$F3,$BE
	.dc.b	$0F,$4F,$E8,$B0,$E1,$8F,$E0,$4F,$F3,$23,$CF,$4F,$16,$9F,$E1,$D9
	.dc.b	$F3,$EB,$E6,$CF,$CB,$E5,$FE,$83,$0E,$11,$E1,$FC,$7B,$7B,$79,$31
	.dc.b	$F5,$7E,$47,$F9,$EF,$C2,$DF,$93,$7F,$AF,$7E,$21,$F9,$67,$F9,$1B
	.dc.b	$E5,$3F,$CE,$FC,$3F,$B8,$BD,$8A,$DF,$D7,$7F,$82,$3F,$3C,$E9,$8F
	.dc.b	$FB,$1F,$8B,$D8,$7F,$B2,$5F,$F7,$43,$8B,$E7,$7C,$BE,$67,$F7,$31
	.dc.b	$8C,$FF,$CA,$FF,$35,$8C,$3C,$D1,$E5,$FD,$5F,$BD,$78,$E3,$D1,$F3
	.dc.b	$FF,$BA,$FA,$97,$E5,$6F,$DF,$1F,$A1,$3F,$6D,$FF,$69,$FC,$6B,$F2
	.dc.b	$8E,$17,$BA,$D7,$27,$F7,$EF,$C7,$F3,$8F,$E7,$23,$F8,$DE,$5D,$BF
	.dc.b	$7C,$C7,$89,$F8,$58,$F0,$D7,$83,$4F,$F7,$0F,$CD,$FF,$16,$9C,$E5
	.dc.b	$DB,$97,$1A,$78,$6F,$CB,$F9,$3F,$82,$BC,$7B,$FE,$53,$5F,$0D,$ED
	.dc.b	$FF,$09,$F8,$3B,$C3,$DB,$EF,$FF,$1C,$FC,$07,$E0,$7C,$FB,$E7,$9F
	.dc.b	$86,$FF,$B9,$29,$F5,$BB,$C3,$FC,$31,$5A,$FC,$2F,$FA,$37,$E7,$B9
	.dc.b	$FA,$4F,$7F,$83,$FE,$BB,$D3,$AB,$F8,$07,$BD,$F1,$D7,$F9,$AC,$63
	.dc.b	$4F,$F7,$9F,$C5,$63,$3E,$F1,$E3,$FD,$0B,$C5,$BF,$9E,$3C,$DF,$8B
	.dc.b	$FF,$36,$FD,$89,$F9,$7B,$F4,$67,$91,$FE,$4A,$FD,$69,$FB,$6F,$FB
	.dc.b	$7F,$9F,$FA,$8E,$1C,$2F,$7E,$4A,$F7,$F2,$1D,$72,$79,$3E,$EE,$56
	.dc.b	$3C,$5F,$CE,$BC,$F4,$F0,$7E,$C9,$FD,$4A,$31,$97,$5F,$C0,$B1,$8F
	.dc.b	$9E,$FC,$DB,$3F,$54,$F0,$6D,$E4,$DB,$D3,$FA,$73,$81,$FA,$CF,$F1
	.dc.b	$1F,$D1,$3F,$AA,$7E,$AA,$F8,$C7,$E5,$AF,$DC,$FF,$C4,$7F,$31,$C7
	.dc.b	$0E,$1F,$20,$C3,$85,$EF,$E9,$3F,$AC,$67,$DB,$9F,$FD,$ED,$A7,$67
	.dc.b	$FA,$CE,$3C,$BF,$4A,$FD,$13,$D3,$1E,$7A,$7F,$8C,$AB,$C0,$72,$CF
	.dc.b	$C1,$AF,$9A,$7F,$A6,$CE,$86,$3E,$C9,$D6,$7E,$DC,$FC,$95,$AC,$39
	.dc.b	$E9,$E9,$FA,$EE,$58,$F1,$8F,$63,$89,$DE,$5F,$E4,$6F,$D3,$FC,$1F
	.dc.b	$62,$FC,$8D,$FB,$47,$F9,$43,$ED,$9E,$EB,$F2,$9F,$E1,$7B,$F8,$B0
	.dc.b	$E1,$7B,$AD,$7C,$1F,$4C,$79,$F0,$7C,$80,$2F,$C1,$FA,$DF,$CA,$8C
	.dc.b	$C7,$4F,$92,$69,$E2,$F9,$6E,$3A,$78,$71,$F3,$D6,$B3,$FE,$CD,$8F
	.dc.b	$87,$4F,$26,$BE,$9C,$38,$46,$3F,$65,$FA,$1E,$1F,$F0,$2F,$DC,$9E
	.dc.b	$9B,$E5,$5F,$A5,$FF,$B0,$3F,$28,$DF,$9C,$BD,$D6,$BF,$07,$FF,$53
	.dc.b	$69,$F5,$B2,$E9,$B7,$C1,$3F,$D7,$B8,$3F,$2A,$F8,$11,$8C,$79,$E7
	.dc.b	$FE,$8F,$8E,$F1,$8F,$C8,$3E,$0F,$85,$FD,$C7,$1F,$46,$DE,$7C,$3F
	.dc.b	$DA,$BF,$75,$C2,$31,$AB,$F6,$FF,$C7,$7E,$6F,$FE,$90,$FC,$8D,$FB
	.dc.b	$EB,$F4,$9F,$E1,$3F,$2E,$FD,$C7,$7F,$80,$AF,$75,$AF,$DB,$73,$F6
	.dc.b	$3E,$AC,$BA,$72,$7B,$DE,$D9,$F5,$F5,$7F,$3F,$D1,$F9,$A9,$CE,$3F
	.dc.b	$C6,$3C,$CF,$6E,$1F,$86,$9C,$DB,$67,$E2,$5E,$4D,$7C,$73,$9F,$DD
	.dc.b	$71,$A7,$7B,$7B,$78,$D3,$DB,$F0,$BF,$AF,$9F,$F0,$F7,$73,$F4,$D7
	.dc.b	$E2,$2F,$D2,$5F,$A2,$7F,$80,$BB,$BF,$C4,$18,$70,$BD,$D6,$BF,$44
	.dc.b	$E9,$F4,$7F,$89,$E4,$FC,$05,$FF,$8A,$FE,$AB,$EF,$BC,$73,$ED,$FA
	.dc.b	$7F,$C0,$AF,$75,$FE,$47,$2E,$BF,$0F,$9C,$38,$67,$E4,$F8,$0E,$FB
	.dc.b	$FA,$73,$F4,$6F,$FC,$E3,$AB,$F8,$3F,$E4,$DF,$E9,$DF,$1F,$FC,$F5
	.dc.b	$F8,$BF,$EF,$3D,$EF,$B9,$5A,$F9,$3F,$CE,$7B,$6A,$D7,$83,$FC,$57
	.dc.b	$1F,$F5,$37,$99,$F9,$87,$A7,$27,$E7,$5E,$D7,$C7,$D2,$FE,$45,$97
	.dc.b	$5D,$3C,$31,$F8,$2E,$9F,$F2,$55,$F0,$CF,$C9,$F8,$F7,$7D,$7C,$71
	.dc.b	$E7,$F3,$FF,$93,$7E,$E4,$FA,$C7,$E3,$AF,$CA,$9F,$4A,$FD,$41,$6F
	.dc.b	$D5,$F8,$E1,$C2,$F7,$5A,$33,$E6,$7E,$D2,$FB,$63,$6F,$86,$3D,$7F
	.dc.b	$31,$FB,$15,$AC,$79,$F8,$F8,$BD,$E5,$E2,$FB,$9F,$0E,$5E,$7C,$BC
	.dc.b	$D8,$BF,$7A,$CE,$5E,$48,$FF,$7C,$F3,$7F,$40,$7B,$57,$0F,$CA,$BF
	.dc.b	$A4,$7E,$28,$FC,$75,$BF,$F9,$C3,$EF,$9F,$98,$7F,$5F,$F9,$5D,$D6
	.dc.b	$BE,$EB,$FD,$19,$3F,$D8,$FF,$70,$CF,$A9,$FC,$53,$2E,$99,$5C,$38
	.dc.b	$65,$D3,$9B,$F3,$15,$AC,$B9,$FC,$46,$2F,$79,$FF,$6F,$FD,$1B,$C3
	.dc.b	$97,$96,$3F,$B9,$7C,$67,$F6,$1F,$CE,$5F,$BD,$FE,$E3,$F2,$9F,$F6
	.dc.b	$87,$D4,$BF,$67,$FE,$75,$E2,$FF,$A4,$6F,$7D,$EB,$DD,$3E,$2F,$A6
	.dc.b	$3F,$C0,$BC,$AF,$E6,$58,$76,$FD,$3F,$F9,$8B,$CB,$0E,$FF,$84,$FE
	.dc.b	$CE,$31,$E6,$7F,$91,$0C,$61,$E5,$7F,$C8,$F9,$3F,$7D,$EF,$A7,$8E
	.dc.b	$73,$F7,$BF,$E1,$9F,$C7,$BF,$D5,$1F,$9E,$BF,$22,$7E,$DE,$FC,$65
	.dc.b	$FA,$BB,$EF,$5F,$B6,$CF,$21,$97,$2C,$78,$E1,$C2,$F7,$F5,$5D,$F9
	.dc.b	$BF,$B7,$E9,$DF,$4F,$F3,$C6,$3C,$78,$AD,$6B,$2E,$7A,$F8,$7E,$3F
	.dc.b	$FB,$7D,$F7,$C3,$CD,$3E,$5F,$C4,$F3,$87,$19,$CE,$9E,$8F,$D4,$FC
	.dc.b	$52,$D3,$C5,$B7,$9F,$81,$FA,$36,$A7,$39,$FC,$19,$FE,$DD,$C3,$F0
	.dc.b	$30,$ED,$87,$1E,$DE,$BE,$77,$F9,$EB,$F6,$D7,$6B,$F6,$97,$ED,$5F
	.dc.b	$E4,$8F,$D4,$B7,$EC,$AF,$75,$AF,$0D,$F9,$86,$BF,$36,$F8,$BB,$7F
	.dc.b	$0D,$C3,$83,$FE,$E1,$3E,$09,$9C,$39,$E5,$E7,$DF,$D1,$39,$D3,$C9
	.dc.b	$3F,$EE,$99,$CE,$BE,$99,$CE,$7F,$E2,$09,$CE,$3E,$79,$CF,$EF,$BF
	.dc.b	$33,$F3,$DF,$F0,$AF,$E1,$DF,$E0,$BF,$55,$F6,$0F,$CE,$BF,$1B,$97
	.dc.b	$8A,$F7,$F0,$FF,$E8,$4A,$D7,$85,$CF,$E8,$FF,$DC,$7C,$3E,$D3,$1F
	.dc.b	$05,$F0,$B0,$E3,$9F,$6F,$0F,$E0,$C6,$31,$FF,$3B,$5F,$3D,$7C,$91
	.dc.b	$E3,$D3,$1A,$78,$75,$7E,$95,$F8,$6F,$61,$FA,$BB,$EA,$1F,$A4,$3F
	.dc.b	$51,$7E,$7E,$F8,$BF,$E3,$BA,$7C,$EC,$78,$E1,$C2,$F7,$4F,$CD,$3B
	.dc.b	$65,$D7,$83,$DD,$8F,$18,$7E,$6D,$F8,$0C,$B9,$E9,$E1,$F2,$3E,$C7
	.dc.b	$9B,$FE,$5A,$C5,$5A,$D3,$CB,$F2,$3F,$84,$8C,$6F,$E7,$FD,$AB,$C3
	.dc.b	$C1,$E5,$8F,$57,$ED,$78,$D3,$DD,$1E,$DF,$A2,$C6,$5E,$AF,$C3,$BF
	.dc.b	$57,$F8,$FB,$D0,$7E,$FD,$FE,$FE,$F8,$07,$E5,$6F,$DE,$3E,$2F,$1F
	.dc.b	$9F,$0E,$17,$BA,$D7,$47,$4C,$5F,$B3,$7E,$82,$F7,$CB,$F1,$B1,$E5
	.dc.b	$5E,$0F,$C4,$DA,$78,$3F,$39,$FD,$1E,$1D,$A7,$39,$FE,$8F,$0E,$3A
	.dc.b	$78,$75,$72,$F8,$63,$1F,$AC,$FF,$08,$FA,$B7,$EF,$FF,$E7,$4F,$A6
	.dc.b	$FF,$34,$7D,$C7,$FB,$7B,$F0,$8D,$4E,$7C,$54,$FF,$B8,$B9,$73,$EC
	.dc.b	$FE,$EB,$1E,$7F,$71,$70,$E1,$7F,$DA,$39,$B9,$CF,$F9,$3F,$1E,$7F
	.dc.b	$75,$FB,$EC,$38,$E9,$E2,$F9,$67,$EA,$BC,$23,$1E,$E7,$F2,$0F,$B2
	.dc.b	$7E,$63,$F8,$CF,$D8,$DF,$A7,$7F,$BD,$3F,$64,$7E,$D5,$BA,$D4,$E7
	.dc.b	$F6,$BC,$7E,$8D,$FC,$4B,$1E,$7E,$75,$BD,$FB,$0F,$D0,$FF,$89,$D6
	.dc.b	$B4,$EF,$2F,$B8,$CD,$F0,$D7,$FD,$99,$E0,$7E,$6F,$0E,$1A,$F9,$7D
	.dc.b	$7F,$26,$8C,$F9,$E0,$FC,$F7,$31,$8F,$96,$FE,$7F,$F0,$1F,$EF,$2F
	.dc.b	$CD,$5F,$B1,$7F,$65,$FD,$E5,$FB,$73,$B3,$7F,$89,$AF,$7F,$36,$BD
	.dc.b	$FE,$8F,$A9,$FE,$67,$97,$FB,$FB,$F3,$8B,$87,$1C,$FB,$7B,$FF,$CD
	.dc.b	$2B,$C5,$1E,$1F,$85,$C5,$6F,$3E,$DA,$BE,$A3,$B5,$F9,$E1,$FC,$00
	.dc.b	$C9,$FA,$47,$F0,$1E,$17,$EB,$4F,$CB,$27,$EE,$7F,$DE,$BF,$01,$FB
	.dc.b	$B8,$70,$F3,$3F,$DC,$38,$70,$BD,$FC,$DA,$DE,$2F,$7C,$FB,$67,$D7
	.dc.b	$1F,$E2,$F1,$8F,$DE,$7F,$9B,$46,$CE,$B3,$E3,$C5,$A7,$F8,$33,$A7
	.dc.b	$FD,$D3,$87,$49,$F0,$FD,$5E,$30,$E3,$1E,$4C,$DF,$A1,$F5,$9F,$F1
	.dc.b	$96,$5D,$72,$F7,$7E,$5D,$9F,$CC,$BF,$9B,$F9,$DF,$E5,$2F,$D5,$11
	.dc.b	$ED,$F9,$6D,$7C,$4F,$EA,$36,$56,$B5,$7E,$61,$AF,$47,$DB,$CF,$F2
	.dc.b	$77,$E7,$8F,$EC,$1F,$86,$0C,$E9,$E1,$D3,$F1,$D1,$E0,$8C,$67,$EF
	.dc.b	$8D,$BE,$29,$FA,$5F,$25,$B8,$C6,$7F,$81,$C7,$C5,$5E,$E9,$CF,$EC
	.dc.b	$3B,$D7,$D5,$FB,$0F,$C7,$FB,$E7,$E3,$EF,$D6,$1F,$BC,$3F,$4D,$7E
	.dc.b	$40,$FA,$1F,$9F,$B7,$BF,$1E,$38,$70,$BD,$D3,$9F,$B6,$4F,$CB,$FE
	.dc.b	$3F,$6F,$F5,$5C,$78,$F2,$7C,$60,$F4,$D7,$C5,$E3,$FF,$84,$EB,$86
	.dc.b	$1E,$79,$F3,$6C,$F9,$9F,$91,$87,$1D,$7C,$7D,$BE,$2D,$FC,$D7,$EB
	.dc.b	$FB,$3F,$C4,$7E,$D4,$FD,$03,$FD,$E6,$7F,$98,$3E,$67,$E1,$FC,$46
	.dc.b	$A3,$3F,$63,$FD,$FB,$67,$E5,$F9,$FB,$7F,$6F,$27,$DF,$CB,$A7,$D4
	.dc.b	$62,$11,$DE,$7C,$18,$7F,$82,$3E,$53,$F0,$13,$9C,$BC,$5F,$73,$87
	.dc.b	$C3,$AF,$D6,$FD,$43,$C1,$1E,$7D,$5E,$BF,$E8,$FF,$24,$FE,$E5,$F2
	.dc.b	$2F,$D7,$BF,$9D,$FE,$23,$FC,$83,$E1,$35,$F3,$CA,$D7,$C7,$2B,$F9
	.dc.b	$2F,$E5,$9A,$7D,$58,$F4,$E4,$FB,$0F,$6C,$BA,$FA,$9F,$99,$8C,$61
	.dc.b	$CF,$F3,$CC,$67,$DE,$FC,$7E,$EF,$FB,$9F,$C8,$FD,$56,$1C,$33,$FE
	.dc.b	$6B,$F0,$DE,$6D,$BC,$B1,$FC,$E3,$F1,$6C,$63,$CB,$A3,$F8,$17,$E3
	.dc.b	$CE,$FF,$DB,$FE,$99,$FD,$0F,$F5,$27,$EF,$9F,$7F,$6F,$DC,$5E,$FD
	.dc.b	$95,$EE,$BF,$AC,$56,$B0,$FE,$03,$CB,$F9,$7C,$3B,$7E,$73,$F7,$5E
	.dc.b	$58,$77,$F3,$BF,$75,$18,$E1,$69,$FC,$9F,$2E,$9A,$78,$27,$3A,$7F
	.dc.b	$34,$F3,$7F,$C7,$58,$71,$C3,$D3,$FB,$FF,$F1,$FF,$DD,$7F,$CA,$3F
	.dc.b	$8F,$BF,$3A,$7E,$42,$FD,$71,$F8,$8F,$F9,$23,$7F,$E2,$D8,$70,$BD
	.dc.b	$D6,$BE,$77,$A7,$C0,$B2,$FC,$C5,$FF,$97,$BF,$73,$B3,$9C,$38,$E5
	.dc.b	$FC,$5E,$BA,$67,$E1,$FC,$53,$17,$B8,$F2,$53,$F4,$6E,$F9,$7A,$67
	.dc.b	$3F,$A3,$7F,$3B,$F4,$DF,$BC,$BF,$6C,$7E,$21,$C7,$F0,$DF,$F2,$6F
	.dc.b	$EE,$9B,$AD,$4E,$7E,$C1,$FA,$26,$70,$FD,$96,$3D,$3F,$00,$B8,$70
	.dc.b	$DB,$FD,$85,$E7,$7F,$31,$AF,$0D,$F8,$27,$FD,$8D,$E6,$E6,$F8,$6B
	.dc.b	$FE,$5E,$F9,$7E,$DA,$F8,$7F,$80,$2E,$1F,$EC,$DD,$BF,$8B,$75,$FF
	.dc.b	$47,$D3,$AD,$F9,$FF,$56,$FE,$F5,$F4,$3F,$E7,$9F,$CF,$3E,$07,$EA
	.dc.b	$C3,$FB,$F7,$D9,$7F,$36,$F5,$37,$E6,$7F,$4D,$AD,$D6,$BD,$97,$E2
	.dc.b	$BA,$A5,$74,$7F,$D1,$B8,$F7,$E5,$ED,$87,$6F,$43,$FA,$C6,$FF,$E1
	.dc.b	$FB,$ED,$5A,$86,$8F,$7B,$78,$F3,$C7,$D1,$5A,$D1,$E2,$F8,$F4,$F3
	.dc.b	$CE,$7B,$BC,$D5,$FE,$F0,$FC,$A1,$E4,$7F,$13,$EC,$FF,$1A,$7B,$3A
	.dc.b	$FC,$19,$5A,$F5,$FF,$C3,$DF,$39,$C3,$3F,$47,$1F,$F5,$CB,$DB,$F0
	.dc.b	$0F,$EE,$99,$F7,$CF,$B5,$FA,$A7,$35,$DA,$7C,$5F,$45,$99,$E9,$97
	.dc.b	$87,$4F,$F2,$47,$C5,$79,$63,$18,$FF,$B8,$33,$F1,$FF,$00,$BF,$0C
	.dc.b	$FF,$3A,$FE,$D6,$FC,$81,$FB,$53,$F5,$AF,$F2,$9E,$1C,$2F,$75,$A9
	.dc.b	$7E,$B3,$F7,$71,$E9,$F4,$1F,$E1,$4E,$AF,$F6,$14,$63,$3E,$D9,$FE
	.dc.b	$2F,$83,$86,$7D,$21,$F7,$FE,$DB,$E1,$A7,$7F,$D0,$37,$9B,$2E,$7A
	.dc.b	$FE,$D3,$6F,$34,$E6,$FC,$DF,$67,$C2,$F7,$F8,$6F,$E4,$FF,$3A,$FF
	.dc.b	$54,$FE,$0A,$FA,$0F,$F7,$17,$B5,$FE,$D8,$F9,$0D,$79,$9F,$EC,$79
	.dc.b	$CF,$E1,$EF,$D3,$F3,$87,$BF,$9B,$FC,$5D,$97,$4B,$DC,$69,$EB,$ED
	.dc.b	$FD,$99,$D7,$E0,$AD,$3F,$DC,$FF,$AC,$6E,$31,$9B,$F6,$AF,$D5,$7D
	.dc.b	$13,$F5,$07,$E8,$4F,$DD,$DF,$3C,$FA,$EF,$F3,$9D,$70,$37,$3A,$8F
	.dc.b	$F3,$BF,$C8,$E6,$FD,$5C,$3F,$E3,$4B,$DE,$1C,$FF,$01,$FC,$86,$5E
	.dc.b	$1B,$F0,$4F,$7F,$C3,$F3,$7C,$32,$FF,$41,$FD,$57,$5D,$3C,$1F,$21
	.dc.b	$FE,$3A,$F1,$BF,$CA,$9A,$BF,$41,$7E,$F3,$FD,$A7,$F7,$F7,$E7,$FF
	.dc.b	$C4,$BF,$70,$FE,$15,$5A,$F9,$B5,$6B,$DE,$FE,$2D,$FB,$96,$B1,$FE
	.dc.b	$E5,$97,$EB,$CE,$CF,$5C,$BF,$89,$79,$1F,$B6,$CB,$A6,$3C,$FC,$CF
	.dc.b	$D7,$46,$2B,$F9,$1F,$7F,$5D,$3C,$31,$FC,$9B,$F3,$DC,$61,$E3,$F7
	.dc.b	$BD,$99,$75,$CF,$C1,$B7,$F2,$BF,$CF,$7F,$92,$7E,$E9,$FD,$A3,$F6
	.dc.b	$B7,$EE,$8F,$D3,$53,$FE,$79,$FC,$53,$7F,$04,$BD,$D6,$B9,$1F,$2B
	.dc.b	$A9,$CE,$1F,$3B,$5F,$E2,$6F,$7C,$BB,$2F,$0B,$F8,$63,$9C,$7A,$78
	.dc.b	$3C,$EB,$79,$78,$70,$FE,$3B,$EF,$3E,$2D,$FB,$EB,$F4,$77,$EA,$7F
	.dc.b	$E2,$0F,$C8,$5F,$3B,$FD,$0F,$0E,$17,$BA,$D7,$E3,$F9,$C7,$9E,$3C
	.dc.b	$BE,$91,$F3,$5F,$C8,$71,$E3,$39,$F8,$7C,$5F,$0C,$7C,$1F,$B2,$F3
	.dc.b	$CF,$B6,$4D,$5F,$8D,$FF,$55,$FB,$FF,$F4,$37,$DB,$3F,$29,$FF,$7D
	.dc.b	$7E,$A9,$EA,$FA,$8D,$7D,$6F,$F9,$AA,$73,$AF,$ED,$79,$7F,$9F,$E3
	.dc.b	$CB,$9B,$F2,$0F,$4A,$E7,$FA,$D7,$F3,$98,$C7,$D6,$B3,$FC,$BE,$1C
	.dc.b	$B2,$ED,$39,$CB,$FB,$4C,$E7,$4F,$34,$E7,$F5,$5F,$2C,$63,$0F,$A3
	.dc.b	$FA,$EE,$34,$F3,$FE,$11,$FE,$F7,$FD,$C3,$F6,$3F,$4F,$FE,$A0,$FA
	.dc.b	$DF,$F2,$AF,$DF,$3E,$F5,$E5,$3D,$7E,$1C,$BF,$EC,$BC,$38,$74,$B0
	.dc.b	$E1,$6C,$B7,$6B,$DD,$75,$FD,$6B,$FB,$ED,$78,$BF,$05,$69,$E2,$D3
	.dc.b	$F0,$F8,$F2,$C7,$8C,$79,$3F,$77,$C5,$F0,$D3,$C5,$3E,$8F,$E0,$1F
	.dc.b	$0F,$E6,$2E,$27,$C5,$FF,$93,$7E,$7D,$FB,$5B,$EF,$7F,$9E,$75,$FA
	.dc.b	$77,$FF,$74,$63,$C7,$0E,$17,$BF,$93,$EB,$97,$4F,$B6,$FF,$03,$CB
	.dc.b	$EA,$FD,$B8,$F2,$AD,$69,$E0,$FA,$AC,$CB,$3E,$D8,$F2,$D7,$F1,$9C
	.dc.b	$BF,$C9,$A7,$38,$7A,$27,$FC,$DD,$B7,$9F,$07,$EE,$FE,$FF,$A2,$7E
	.dc.b	$F2,$FC,$91,$F7,$BF,$C3,$3D,$3F,$AB,$7B,$AD,$4F,$F1,$0F,$DA,$73
	.dc.b	$C3,$FD,$FA,$73,$F8,$97,$A2,$F8,$63,$D3,$E4,$B1,$7B,$8E,$FE,$DB
	.dc.b	$F6,$4E,$7A,$F9,$27,$3E,$0E,$33,$F0,$DB,$F3,$0C,$57,$2F,$D2,$3F
	.dc.b	$17,$F3,$3F,$A7,$F9,$E7,$EC,$5F,$C9,$FF,$7C,$7E,$B2,$FD,$4B,$7E
	.dc.b	$7A,$F7,$F0,$8B,$DF,$E9,$BA,$9F,$EA,$7D,$5F,$C9,$FB,$FF,$AA,$E9
	.dc.b	$DF,$4E,$DD,$1F,$62,$31,$97,$3E,$7C,$69,$DE,$FC,$9B,$3F,$0C,$ED
	.dc.b	$18,$F9,$0F,$D2,$FC,$2F,$97,$DE,$FE,$55,$FA,$57,$FB,$27,$D4,$7E
	.dc.b	$A7,$AE,$2F,$E3,$1F,$06,$AD,$7D,$96,$70,$F4,$F2,$7E,$33,$2E,$97
	.dc.b	$BF,$95,$FE,$7D,$3B,$8C,$7D,$06,$F2,$EB,$97,$87,$F5,$0D,$3E,$3F
	.dc.b	$13,$F7,$7C,$8F,$15,$F1,$5F,$8E,$F4,$5F,$45,$FE,$B9,$FC,$B7,$75
	.dc.b	$AF,$15,$5A,$F4,$1F,$B9,$EB,$1E,$58,$BF,$32,$FD,$2E,$7D,$E7,$3F
	.dc.b	$2B,$8C,$FA,$65,$CF,$F7,$1E,$5A,$76,$CB,$FD,$95,$C7,$8A,$F2,$FE
	.dc.b	$1C,$FD,$0F,$14,$87,$E9,$9F,$C6,$3F,$35,$1F,$DC,$F5,$7E,$8E,$FD
	.dc.b	$5B,$FC,$D5,$FB,$56,$EB,$53,$9E,$F7,$ED,$59,$F2,$3E,$AF,$75,$CB
	.dc.b	$9E,$4F,$CF,$BF,$81,$E5,$E0,$BF,$9B,$97,$7F,$7F,$33,$DA,$FF,$CF
	.dc.b	$5E,$77,$3A,$F0,$7D,$4F,$F3,$0C,$1F,$B9,$BF,$76,$7E,$90,$F9,$67
	.dc.b	$A3,$FD,$8B,$F4,$BE,$17,$BA,$D4,$FF,$66,$F8,$4C,$E1,$EA,$F8,$2F
	.dc.b	$E2,$19,$76,$8E,$B9,$74,$F0,$7F,$8E,$CE,$71,$E7,$EE,$E3,$3E,$D7
	.dc.b	$E2,$F9,$A7,$95,$DA,$FC,$30,$FD,$C9,$A3,$F7,$7F,$F2,$4F,$EF,$7F
	.dc.b	$E2,$8F,$88,$7E,$B2,$FA,$9F,$F0,$7F,$DB,$7F,$BE,$B0,$E1,$DC,$18
	.dc.b	$71,$BE,$1E,$AF,$F9,$A7,$E5,$BB,$E7,$DB,$E9,$BF,$D8,$75,$AC,$7F
	.dc.b	$DA,$B3,$9D,$3C,$3A,$7F,$87,$BF,$3D,$CE,$9D,$F4,$ED,$3E,$4F,$89
	.dc.b	$EB,$39,$CF,$AC,$FA,$3F,$3C,$CC,$F9,$A7,$3B,$7F,$9F,$7F,$17,$CE
	.dc.b	$3D,$23,$FD,$01,$FA,$AF,$0F,$9D,$7E,$E3,$EF,$FD,$85,$F7,$9F,$DF
	.dc.b	$7F,$4E,$FF,$71,$F1,$FC,$4D,$58,$70,$57,$C3,$E6,$D6,$FC,$27,$B6
	.dc.b	$5D,$72,$FF,$04,$71,$FC,$0C,$79,$CE,$70,$FB,$31,$C0,$EB,$5E,$3F
	.dc.b	$6F,$19,$77,$C9,$FA,$1F,$C2,$FD,$7D,$F9,$5B,$F8,$05,$F8,$FF,$E7
	.dc.b	$BF,$07,$7D,$7F,$B3,$B7,$FD,$1B,$87,$0B,$DF,$8C,$BD,$FC,$17,$5C
	.dc.b	$BA,$65,$CF,$2F,$56,$3C,$63,$B7,$C5,$62,$B7,$8F,$3B,$F3,$F3,$FC
	.dc.b	$BC,$38,$4E,$7F,$45,$FB,$7F,$81,$BF,$5A,$7E,$6A,$FC,$79,$FA,$63
	.dc.b	$F4,$BF,$F0,$A7,$AD,$C2,$F7,$5A,$9F,$F0,$C7,$AB,$9C,$39,$61,$C7
	.dc.b	$DC,$FA,$59,$77,$9C,$DB,$E9,$F3,$97,$4C,$3C,$1F,$4E,$F8,$3E,$9A
	.dc.b	$77,$8F,$E3,$7C,$3F,$87,$CF,$A4,$E7,$C1,$DE,$1C,$BF,$61,$FD,$BF
	.dc.b	$E3,$7F,$EB,$EF,$E0,$0F,$EB,$7F,$22,$FD,$25,$FB,$BB,$F7,$4F,$DC
	.dc.b	$FE,$F1,$C3,$C8,$FD,$B6,$1C,$2F,$75,$FB,$EF,$27,$54,$AD,$61,$CF
	.dc.b	$F0,$FF,$DF,$70,$B4,$F1,$7C,$EB,$5F,$15,$F8,$7F,$0B,$DF,$C9,$FF
	.dc.b	$B5,$3C,$3F,$E0,$DF,$55,$FC,$7F,$F7,$1C,$6D,$FE,$2B,$3F,$B7,$E8
	.dc.b	$8C,$FE,$17,$FC,$96,$7F,$DB,$BF,$4C,$FD,$35,$F8,$1E,$67,$F8,$97
	.dc.b	$EF,$DF,$EF,$9F,$DD,$BF,$A2,$EF,$F0,$E7,$F8,$B7,$EF,$2F,$D2,$1F
	.dc.b	$BB,$EB,$A7,$FD,$D3,$F2,$8F,$C2,$3F,$AA,$E1,$FE,$02,$C7,$F9,$FB
	.dc.b	$DB,$F1,$86,$7D,$BE,$87,$FD,$E3,$A7,$5C,$FF,$9E,$7E,$39,$EE,$E5
	.dc.b	$FE,$7D,$F8,$E5,$CB,$DF,$F4,$35,$FF,$13,$F3,$FC,$B7,$F9,$67,$47
	.dc.b	$C8,$7E,$2F,$E5,$FB,$7F,$DA,$5B,$FF,$C0,$0F,$CD,$3F,$8A,$FE,$15
	.dc.b	$FD,$2E,$3F,$9A,$4B,$F6,$A7,$E4,$9F,$C3,$E3,$E7,$CE,$BC,$AE,$31
	.dc.b	$83,$FC,$EF,$3B,$7C,$BF,$9F,$74,$BE,$7F,$0F,$FD,$67,$F4,$EE,$58
	.dc.b	$32,$FE,$EB,$F1,$69,$CC,$63,$0E,$59,$75,$CB,$CF,$7F,$9A,$7C,$38
	.dc.b	$F2,$CF,$8C,$7F,$8D,$27,$4F,$B6,$B9,$3F,$D4,$E3,$3A,$7B,$23,$FC
	.dc.b	$A7,$0D,$7E,$54,$6F,$C5,$AF,$FB,$B3,$86,$D3,$8A,$F2,$6B,$8F,$F2
	.dc.b	$37,$F5,$8F,$CD,$1F,$6B,$F6,$5D,$69,$F9,$A8,$FC,$EE,$FF,$5F,$8B
	.dc.b	$E3,$46,$0F,$79,$FA,$2F,$D0,$3E,$2F,$0F,$F8,$17,$E5,$BE,$0C,$FF
	.dc.b	$9B,$38,$9E,$87,$A3,$F0,$6B,$3E,$3E,$9F,$A3,$7C,$23,$E6,$E3,$D1
	.dc.b	$F9,$59,$FF,$A8,$F4,$F9,$30,$E1,$4F,$9F,$9B,$ED,$D7,$AE,$BE,$5E
	.dc.b	$FF,$1B,$C4,$F8,$9F,$AC,$FC,$5F,$CD,$7E,$37,$E5,$FF,$1F,$E5,$38
	.dc.b	$61,$F9,$4E,$57,$EC,$5D,$63,$F8,$0B,$E5,$7F,$46,$FA,$5F,$E2,$2F
	.dc.b	$9F,$97,$FC,$51,$FA,$66,$E7,$A7,$EA,$9B,$7F,$91,$FE,$BB,$F1,$4F
	.dc.b	$B7,$19,$E2,$57,$1A,$FC,$46,$1D,$5F,$C7,$3B,$34,$FC,$57,$D0,$FE
	.dc.b	$33,$2F,$6B,$F9,$1C,$FF,$13,$A7,$E2,$B5,$FC,$5F,$A3,$F8,$9F,$8A
	.dc.b	$FC,$57,$CB,$F0,$C3,$FC,$67,$9F,$47,$E3,$3E,$37,$F8,$2F,$DA,$F8
	.dc.b	$B8,$5F,$8E,$FB,$FB,$7F,$D0,$F5,$F8,$13,$FD,$33,$D9,$85,$E3,$FF
	.dc.b	$0B,$7F,$47,$5F,$0F,$0F,$6F,$17,$C4,$9F,$C9,$9F,$A8,$CB,$C8,$FE
	.dc.b	$B7,$3F,$44,$FF,$82,$A3,$C0,$7F,$1A,$FD,$3B,$F4,$DF,$89,$FB,$75
	.dc.b	$FF,$03,$4F,$F9,$1F,$C7,$FD,$A3,$F8,$3F,$92,$F6,$1F,$66,$77,$1F
	.dc.b	$EF,$BE,$CF,$CA,$C7,$F9,$52,$3F,$CA,$1E,$57,$F1,$27,$FA,$4F,$AB
	.dc.b	$F0,$FF,$19,$FD,$DF,$1B,$F4,$7F,$D3,$1F,$A7,$8F,$DE,$CF,$E0,$FE
	.dc.b	$93,$E0,$E7,$F3,$F0,$FE,$26,$FE,$1C,$CB,$E9,$F2,$3F,$72,$FF,$0C
	.dc.b	$DB,$E1,$EF,$FC,$17,$F2,$6F,$81,$87,$C8,$AF,$F3,$7B,$D5,$FE,$D7
	.dc.b	$E9,$FD,$BF,$CD,$7E,$D7,$DC,$7A,$3F,$1A,$FD,$E1,$FB,$D7,$F8,$9C
	.dc.b	$7E,$66,$7E,$BD,$F3,$BF,$DC,$3D,$F9,$58,$FD,$1F,$D2,$3E,$97,$EF
	.dc.b	$3F,$C4,$7E,$33,$FD,$20,$F9,$79,$8F,$E3,$4F,$E6,$BE,$AE,$B2,$FF
	.dc.b	$3F,$70,$FD,$87,$FC,$07,$3F,$D0,$2F,$51,$F5,$FC,$9F,$EE,$3F,$9C
	.dc.b	$7F,$15,$3F,$95,$D7,$9F,$4E,$F5,$D3,$0F,$96,$FF,$42,$9E,$8F,$F3
	.dc.b	$F7,$F2,$A7,$91,$C1,$F4,$F0,$FE,$9F,$1F,$07,$3F,$CA,$FC,$1F,$F3
	.dc.b	$A3,$F2,$5F,$8C,$FF,$31,$FB,$8F,$F5,$83,$FD,$4B,$E7,$AE,$3E,$39
	.dc.b	$FF,$5A,$7C,$4E,$63,$FC,$2B,$F2,$7F,$D7,$A7,$FD,$71,$F8,$ED,$9F
	.dc.b	$EA,$8D,$7F,$A7,$CD,$7F,$AE,$D7,$37,$FA,$9E,$19,$7F,$AC,$39,$7F
	.dc.b	$B1,$BE,$77,$FC,$4A,$7F,$7C,$FC,$4D,$F9,$D3,$ED,$FF,$C4,$53,$FE
	.dc.b	$53,$F8,$37,$CB,$7F,$88,$BE,$D7,$85,$EE,$BF,$C3,$9F,$6F,$8C,$3F
	.dc.b	$C4,$FF,$90,$FE,$5F,$D0,$36,$75,$5F,$C3,$30,$AF,$F6,$C3,$FC,$1F
	.dc.b	$F3,$6C,$FC,$31,$FD,$DC,$D9,$FD,$F8,$FF,$1D,$F8,$DF,$DA,$9F,$F1
	.dc.b	$09,$FE,$18,$F8,$EF,$F1,$97,$DE,$FF,$8D,$63,$FD,$9D,$FA,$87,$5D
	.dc.b	$F1,$0F,$91,$FF,$20,$3E,$C3,$FD,$A9,$7C,$9F,$F1,$D6,$BF,$CE,$B6
	.dc.b	$F6,$C7,$F0,$AF,$3F,$FB,$0C,$7F,$BE,$E3,$8F,$C7,$FF,$90,$9F,$F1
	.dc.b	$58,$8F,$F2,$63,$E1,$F8,$FF,$F2,$1B,$FC,$C3,$E4,$7F,$C6,$3A,$79
	.dc.b	$E3,$E1,$1D,$35,$FF,$12,$BF,$E3,$6D,$3D,$2F,$F7,$DA,$FF,$1C,$DF
	.dc.b	$F9,$7F,$E0,$BE,$AC,$FE,$9F,$E1,$3C,$6F,$FB,$C7,$2F,$5B,$FE,$D6
	.dc.b	$F0,$BF,$56,$E7,$0E,$7C,$3F,$3E,$3F,$C8,$B5,$EA,$E2,$FF,$2C,$47
	.dc.b	$F9,$7F,$93,$D3,$1F,$6A,$3F,$C2,$1C,$BC,$B3,$FF,$36,$FE,$19,$67
	.dc.b	$E6,$9F,$D2,$0E,$BC,$7F,$C5,$71,$4E,$E7,$E5,$C7,$CF,$5B,$7B,$DF
	.dc.b	$C7,$1F,$80,$C7,$C4,$FA,$34,$FF,$3B,$F1,$7D,$58,$FF,$2F,$FA,$9D
	.dc.b	$9F,$9C,$7F,$9E,$7D,$BF,$EF,$31,$FD,$8A,$3F,$01,$EE,$7C,$A7,$EE
	.dc.b	$4F,$F9,$52,$31,$96,$E7,$FC,$9D,$E4,$E3,$4F,$CB,$63,$E1,$9F,$A5
	.dc.b	$C1,$F6,$F5,$FC,$1C,$7F,$8D,$FF,$08,$FF,$42,$F4,$FF,$A4,$23,$FB
	.dc.b	$AC,$63,$E4,$7B,$FC,$95,$9F,$BB,$EB,$BF,$3B,$FF,$53,$C7,$9E,$30
	.dc.b	$87,$E1,$0F,$B6,$7C,$FF,$F8,$7B,$07,$E2,$6F,$D0,$DF,$92,$FF,$BE
	.dc.b	$73,$FF,$43,$E3,$C7,$0E,$17,$BF,$A2,$FF,$7F,$70,$FF,$A3,$3A,$7F
	.dc.b	$D9,$C4,$3F,$DA,$0E,$B5,$F2,$46,$4D,$F6,$B7,$FD,$4F,$4F,$D1,$AF
	.dc.b	$54,$61,$FA,$9F,$29,$BB,$F2,$F0,$FB,$3E,$19,$C5,$EC,$E2,$7E,$0D
	.dc.b	$BC,$7E,$27,$0F,$AE,$F7,$C1,$F0,$7F,$40,$D5,$7C,$8C,$BA,$65,$CF
	.dc.b	$93,$38,$F1,$8E,$DC,$58,$AD,$E3,$F2,$FE,$EF,$9E,$DE,$F3,$E2,$FD
	.dc.b	$1F,$F0,$8E,$2F,$77,$0A,$1F,$A3,$7F,$AD,$BE,$C1,$F8,$97,$FB,$AF
	.dc.b	$A9,$FA,$6B,$EE,$5F,$24,$F8,$AE,$58,$F1,$C3,$85,$EE,$BF,$DB,$5F
	.dc.b	$BC,$BD,$7F,$F6,$1F,$E5,$2F,$D4,$3F,$C4,$9F,$9F,$FF,$B1,$3E,$33
	.dc.b	$FC,$C7,$C8,$C3,$87,$5F,$E3,$70,$E1,$F2,$2C,$38,$77,$3A,$9F,$AF
	.dc.b	$3E,$03,$FD,$A3,$F0,$EF,$D8,$9F,$AE,$7E,$4B,$F6,$97,$EB,$8F,$9F
	.dc.b	$FF,$6B,$63,$C7,$7F,$CE,$63,$C7,$0E,$17,$BF,$CA,$5F,$AD,$3F,$6F
	.dc.b	$FF,$6F,$FE,$2B,$FA,$1F,$9C,$FE,$89,$F9,$3B,$E2,$FE,$5F,$EB,$9C
	.dc.b	$2F,$7E,$57,$FB,$97,$E1,$57,$BF,$03,$FC,$91,$FB,$23,$E7,$FF,$DD
	.dc.b	$72,$FB,$AF,$4F,$C8,$BF,$6F,$FA,$9F,$91,$3F,$69,$63,$C7,$9F,$F6
	.dc.b	$D8,$F1,$E5,$63,$C7,$CA,$63,$FB,$8F,$01,$FA,$17,$FB,$EB,$F2,$C7
	.dc.b	$E8,$BF,$ED,$3F,$71,$F4,$6F,$C9,$1F,$85,$3E,$03,$EB,$63,$C7,$0E
	.dc.b	$17,$BA,$7C,$0B,$F6,$67,$E4,$3F,$C2,$F8,$DE,$7B,$F5,$B7,$E8,$6F
	.dc.b	$CE,$DF,$8A,$BE,$C5,$F9,$03,$D6,$E9,$9F,$3C,$B9,$63,$C7,$0E,$16
	.dc.b	$F9,$87,$D2,$BE,$BF,$C5,$FA,$8B,$F1,$57,$E9,$2F,$D1,$3E,$1F,$EE
	.dc.b	$AF,$D3,$9F,$5B,$CB,$1E,$38,$70,$BD,$D7,$BF,$EF,$9F,$B4,$7B,$FE
	.dc.b	$19,$F4,$7F,$EE,$0E,$4F,$6F,$E1,$CF,$D1,$1F,$B0,$B1,$E3,$D9,$FC
	.dc.b	$C7,$1E,$3F,$82,$FC,$AE,$76,$3C,$7F,$4F,$F8,$DF,$97,$7C,$DF,$51
	.dc.b	$FC,$02,$FD,$61,$FB,$F3,$F1,$9F,$E6,$BF,$B1,$BF,$48,$E3,$BF,$F3
	.dc.b	$FC,$78,$E1,$C2,$F7,$F2,$5F,$EE,$CF,$D7,$BE,$D7,$EB,$CF,$C3,$FF
	.dc.b	$85,$7E,$B1,$F9,$7F,$15,$FB,$7B,$F6,$FE,$1A,$FF,$4B,$C3,$87,$CD
	.dc.b	$B0,$E1,$6E,$0F,$F4,$6F,$E9,$3F,$87,$FD,$A1,$F7,$8F,$9F,$7D,$23
	.dc.b	$F7,$37,$CE,$FF,$AC,$FF,$11,$E3,$BF,$F5,$9C,$78,$E1,$C2,$F7,$F3
	.dc.b	$DD,$3E,$13,$F4,$F7,$FD,$7F,$EB,$FF,$AD,$FE,$EE,$F9,$C7,$EA,$CF
	.dc.b	$C8,$7D,$9F,$96,$7F,$27,$F8,$BF,$CE,$61,$C2,$F7,$5F,$07,$C1,$7D
	.dc.b	$BB,$F3,$7F,$D1,$7D,$F7,$F6,$AF,$CC,$3F,$8C,$F2,$FE,$C3,$E7,$5F
	.dc.b	$AF,$70,$BD,$FD,$C2,$F7,$F1,$4B,$FD,$57,$3B,$8F,$F5,$DF,$96,$BF
	.dc.b	$80,$5F,$97,$7E,$AF,$C4,$FE,$51,$F0,$1F,$BE,$C3,$87,$4F,$F7,$FC
	.dc.b	$38,$5E,$FE,$A5,$7F,$AB,$FD,$91,$F8,$A7,$EE,$FC,$5F,$C9,$FE,$A7
	.dc.b	$FE,$06,$F7,$1F,$90,$F9,$FF,$6C,$61,$C3,$5F,$F0,$66,$1C,$3E,$F3
	.dc.b	$F4,$57,$ED,$FC,$01,$F9,$23,$F2,$AF,$E1,$1F,$B8,$7F,$1C,$FD,$23
	.dc.b	$FA,$BF,$E2,$BF,$B4,$FD,$E9,$F6,$1C,$2F,$7E,$42,$F7,$F1,$EB,$7E
	.dc.b	$A7,$D7,$F4,$FE,$BF,$40,$FD,$C1,$F9,$DF,$EC,$FF,$78,$77,$3F,$5B
	.dc.b	$E1,$CF,$FE,$25,$E2,$FE,$57,$87,$0F,$8E,$61,$C3,$F3,$CF,$57,$4F
	.dc.b	$F2,$1F,$91,$FE,$3D,$F7,$6F,$C3,$9F,$B3,$3F,$1D,$7E,$99,$FD,$87
	.dc.b	$A7,$F4,$9B,$FF,$8C,$F1,$E3,$87,$0B,$DF,$D1,$3F,$31,$FF,$19,$5F
	.dc.b	$F8,$E7,$F2,$2F,$ED,$3F,$B7,$BD,$D7,$9C,$FD,$2F,$F1,$FF,$9D,$71
	.dc.b	$F1,$7F,$9B,$61,$C2,$F7,$5F,$94,$FC,$43,$B7,$EE,$7F,$E3,$7F,$88
	.dc.b	$FF,$0A,$1E,$1F,$83,$FF,$15,$FB,$5F,$E3,$EC,$BF,$C8,$7B,$FE,$AB
	.dc.b	$7F,$F1,$4F,$E0,$1F,$E1,$E9,$FE,$FD,$2F,$DD,$3F,$B1,$FE,$3C,$F0
	.dc.b	$3F,$C9,$BF,$48,$FC,$DB,$D1,$FA,$17,$BE,$F7,$5A,$9F,$67,$81,$F0
	.dc.b	$F7,$F1,$FE,$1F,$FE,$36,$FD,$F3,$96,$9F,$EA,$EF,$53,$FC,$65,$F8
	.dc.b	$C2,$5F,$98,$71,$E2,$F3,$F1,$FA,$FF,$1B,$C4,$7F,$8F,$63,$DD,$F8
	.dc.b	$9E,$31,$E7,$F6,$1F,$94,$CF,$B7,$37,$F8,$F2,$3F,$23,$1F,$EB,$ED
	.dc.b	$BF,$C7,$9C,$9F,$C7,$23,$11,$FE,$5D,$F9,$3C,$3F,$3C,$FF,$32,$7E
	.dc.b	$50,$EB,$F1,$C7,$D2,$1F,$55,$FE,$45,$DB,$FD,$77,$9F,$F9,$EE,$31
	.dc.b	$1F,$ED,$0F,$C6,$3F,$95,$C7,$FB,$4B,$5F,$87,$F5,$7F,$DE,$23,$17
	.dc.b	$FE,$54,$8F,$F7,$24,$63,$83,$E2,$37,$97,$FC,$C7,$D9,$FE,$39,$8C
	.dc.b	$7E,$39,$FE,$86,$FD,$A7,$FC,$09,$18,$E8,$47,$F5,$78,$FF,$18,$6B
	.dc.b	$FE,$91,$F5,$7F,$81,$7C,$7B,$F3,$67,$EE,$7F,$E2,$5E,$9F,$F2,$EE
	.dc.b	$1C,$30,$FD,$CF,$9F,$FC,$BE,$F7,$F8,$6F,$F9,$2F,$EA,$FE,$2F,$C5
	.dc.b	$FF,$67,$FB,$2F,$E0,$DF,$BD,$BF,$79,$FF,$1B,$FE,$92,$FC,$83,$7A
	.dc.b	$7F,$65,$BD,$D6,$BF,$35,$7E,$57,$FE,$48,$AF,$F5,$47,$EF,$5C,$7E
	.dc.b	$E3,$F8,$97,$C5,$7F,$9B,$FE,$F5,$FB,$8F,$0F,$C7,$3F,$C4,$97,$BA
	.dc.b	$FE,$31,$F6,$1F,$E1,$5F,$D5,$FA,$FD,$47,$F8,$A7,$F5,$EF,$F0,$6F
	.dc.b	$E1,$8F,$80,$FE,$DD,$E8,$7F,$AD,$7E,$AF,$7C,$1F,$EB,$9B,$DD,$6B
	.dc.b	$C9,$7E,$F5,$E8,$FC,$03,$FC,$47,$F8,$D7,$F2,$1F,$CC,$1F,$AC,$BE
	.dc.b	$DB,$FC,$61,$F3,$7B,$D3,$FD,$89,$F8,$47,$F9,$5F,$E5,$18,$70,$BF
	.dc.b	$F1,$B7,$E4,$5B,$B5,$FF,$97,$7F,$29,$FF,$72,$76,$FF,$A5,$27,$9E
	.dc.b	$9F,$CC,$5F,$F2,$31,$FE,$D1,$87,$E9,$9F,$BA,$F4,$7F,$D9,$FE,$AF
	.dc.b	$F9,$32,$73,$5F,$B6,$F5,$3F,$D6,$72,$FC,$87,$EA,$E1,$FC,$52,$BA
	.dc.b	$98,$3F,$89,$E3,$E4,$8E,$DE,$A7,$FB,$6F,$DF,$D5,$7E,$FF,$1F,$CE
	.dc.b	$FE,$5D,$8F,$67,$F8,$4F,$B1,$FE,$97,$D7,$E1,$DF,$F1,$EF,$86,$F9
	.dc.b	$5F,$02,$FD,$65,$FB,$23,$F4,$EF,$C4,$F4,$7F,$B4,$AF,$75,$AF,$93
	.dc.b	$CE,$3F,$CA,$3F,$24,$F9,$37,$B8,$F5,$E7,$E0,$7C,$DC,$DF,$EC,$F8
	.dc.b	$C6,$DE,$6F,$6F,$9D,$7F,$19,$38,$DF,$FB,$8B,$97,$AC,$7C,$4F,$91
	.dc.b	$FF,$65,$FE,$07,$FE,$82,$FA,$4E,$38,$BC,$4F,$F4,$C5,$6B,$F1,$AF
	.dc.b	$F3,$5E,$9E,$09,$FF,$67,$E3,$D8,$FF,$7C,$FC,$37,$FA,$2C,$FF,$1D
	.dc.b	$FE,$39,$FF,$00,$2B,$D5,$1F,$F0,$00,$FF,$1F,$9F,$E2,$43,$D2,$64
	.dc.b	$FF,$21,$FE,$61,$FC,$8E,$38,$C7,$FA,$48,$E3,$D7,$F8,$BD,$FF,$AC
	.dc.b	$79,$BD,$27,$FD,$FA,$74,$30,$3F,$34,$E1,$97,$A2,$38,$7E,$FF,$FE
	.dc.b	$52,$C5,$FA,$07,$FA,$77,$3F,$F4,$F7,$DE,$FA,$23,$C8,$7F,$A8,$78
	.dc.b	$7E,$47,$3B,$F5,$8F,$F6,$BF,$7B,$F2,$47,$ED,$7F,$C3,$30,$E1,$C9
	.dc.b	$FE,$C1,$BD,$D7,$F9,$53,$4F,$0E,$FF,$D5,$8F,$C1,$DF,$0C,$BF,$5E
	.dc.b	$FE,$73,$27,$1F,$5A,$FD,$47,$5B,$C4,$57,$E6,$EF,$F0,$95,$D0,$FE
	.dc.b	$C7,$EB,$7F,$AF,$BF,$34,$FF,$08,$72,$F3,$9F,$EC,$93,$9F,$94,$72
	.dc.b	$FF,$26,$F3,$BF,$D8,$5F,$38,$F9,$B7,$CC,$FF,$8D,$FF,$02,$7E,$EF
	.dc.b	$FE,$FC,$FD,$F3,$86,$83,$FC,$E7,$7C,$3E,$03,$FD,$03,$4F,$CC,$9F
	.dc.b	$17,$C2,$F7,$7F,$C4,$34,$F0,$BE,$08,$FD,$4F,$BF,$F1,$79,$FF,$9D
	.dc.b	$BF,$1F,$CA,$B3,$F3,$4F,$E2,$BD,$49,$F8,$B1,$8C,$B9,$4F,$8E,$3E
	.dc.b	$6F,$47,$CC,$BF,$E8,$F5,$FC,$4E,$78,$1F,$C4,$8F,$DD,$FE,$33,$C8
	.dc.b	$F1,$59,$E1,$1F,$EE,$B8,$C3,$C4,$F2,$FE,$0F,$E2,$38,$9C,$1F,$E5
	.dc.b	$8F,$A1,$E5,$F9,$FF,$A0,$EE,$6D,$FE,$BA,$ED,$FE,$69,$F4,$9F,$EF
	.dc.b	$BF,$C7,$FF,$8E,$7B,$7E,$5F,$9E,$FF,$60,$7C,$4F,$FB,$5B,$25,$EF
	.dc.b	$D4,$5E,$FD,$25,$BF,$44,$FE,$D7,$7B,$F9,$DF,$E8,$EF,$1B,$BA,$7C
	.dc.b	$CF,$13,$F3,$79,$F6,$8C,$71,$72,$7E,$C1,$F1,$EF,$D4,$FF,$78,$EE
	.dc.b	$FA,$E7,$CE,$3E,$05,$7F,$95,$E1,$E5,$FF,$1F,$F5,$7F,$AC,$F9,$9F
	.dc.b	$78,$FA,$A7,$03,$AD,$3C,$4F,$C1,$9F,$68,$EA,$6B,$E3,$7F,$0E,$7D
	.dc.b	$E3,$C2,$72,$FD,$27,$F1,$67,$E0,$8F,$29,$DB,$F6,$EF,$C7,$9F,$86
	.dc.b	$3D,$27,$8B,$F2,$FF,$C9,$9F,$8A,$3D,$A7,$9B,$F6,$2F,$CB,$9F,$8E
	.dc.b	$3E,$09,$EA,$FC,$23,$F3,$67,$E4,$8F,$8A,$7B,$BE,$03,$F3,$E7,$E5
	.dc.b	$8F,$92,$7C,$2F,$AA,$FD,$19,$F9,$A3,$E6,$9F,$1B,$DF,$FD,$39,$F9
	.dc.b	$E3,$E8,$9F,$2B,$D2,$FD,$59,$FA,$23,$EA,$9F,$3B,$E9,$3F,$5E,$7E
	.dc.b	$98,$FB,$27,$D2,$E0,$FD,$99,$FA,$A3,$EE,$9F,$5B,$F3,$4F,$DB,$9F
	.dc.b	$AE,$3F,$02,$67,$8F,$FD,$AF,$F0,$7F,$ED,$CF,$CF,$DF,$9D,$BF,$30
	.dc.b	$7E,$F8,$FD,$FF,$FC,$87,$8F,$FB,$97,$0E,$17,$BA,$7E,$B9,$AC,$BA
	.dc.b	$FE,$4D,$FE,$24,$C3,$87,$E1,$5F,$E2,$6F,$1B,$F8,$59,$FC,$08,$FC
	.dc.b	$A9,$F8,$EF,$B0,$FE,$1E,$7F,$06,$3F,$32,$7E,$4B,$EC,$3F,$89,$9F
	.dc.b	$C2,$8F,$CE,$9F,$96,$F7,$FF,$8B,$9F,$C3,$8F,$D0,$9F,$9A,$FB,$AF
	.dc.b	$E3,$67,$F1,$23,$F4,$A7,$E7,$BD,$CF,$E3,$E7,$F1,$63,$F5,$27,$E8
	.dc.b	$BD,$EF,$E4,$67,$F1,$A3,$F5,$A7,$E9,$BE,$27,$F9,$E6,$BF,$EE,$DD
	.dc.b	$DF,$85,$BF,$26,$7E,$89,$FE,$E4,$F5,$7F,$84,$F8,$CC,$78,$FE,$0B
	.dc.b	$F4,$B0,$E1,$7B,$F9,$AF,$F3,$AD,$6B,$2E,$B9,$FF,$8A,$23,$1F,$86
	.dc.b	$FF,$95,$1D,$61,$CB,$F0,$1F,$E6,$1E,$97,$2B,$FE,$A5,$F5,$FF,$BD
	.dc.b	$FC,$59,$FB,$6B,$F6,$07,$EB,$EF,$CA,$9F,$99,$71,$C5,$8F,$1C,$38
	.dc.b	$5E,$FE,$47,$53,$D6,$BC,$C7,$83,$D3,$7E,$57,$E7,$8F,$E4,$5F,$0D
	.dc.b	$D7,$D7,$FF,$1F,$74,$7C,$59,$FE,$42,$7F,$9A,$F8,$FE,$83,$FD,$AF
	.dc.b	$E4,$1F,$BC,$BD,$67,$CB,$3E,$8E,$7A,$BF,$D7,$F2,$E9,$8B,$1E,$38
	.dc.b	$70,$BD,$FD,$4F,$BE,$7A,$FB,$3E,$7B,$E1,$9F,$77,$DA,$63,$0E,$58
	.dc.b	$32,$FE,$F9,$93,$E6,$BE,$AF,$87,$F8,$4E,$CF,$EC,$9F,$05,$FE,$B2
	.dc.b	$9C,$F8,$5F,$81,$BE,$19,$FF,$82,$7C,$3F,$C3,$3F,$E6,$8A,$FC,$0B
	.dc.b	$4D,$78,$5F,$92,$CB,$B7,$89,$FE,$D6,$C3,$9D,$3F,$25,$FF,$60,$FE
	.dc.b	$17,$FE,$10,$FD,$DD,$F9,$5B,$E2,$DF,$99,$79,$70,$FF,$0D,$61,$C3
	.dc.b	$E5,$18,$70,$B7,$EC,$FB,$BE,$D9,$B3,$FF,$54,$EF,$FA,$8C,$BF,$DA
	.dc.b	$3F,$6D,$8B,$EB,$7B,$CF,$AE,$BE,$3F,$DD,$7A,$77,$FF,$A6,$3C,$0B
	.dc.b	$F4,$0F,$81,$7D,$57,$FA,$7B,$2F,$7D,$FF,$9F,$E7,$B6,$7F,$D9,$2D
	.dc.b	$F0,$58,$AF,$C8,$EF,$FE,$4F,$FA,$8F,$57,$C9,$3E,$3D,$F0,$CF,$D8
	.dc.b	$FF,$C7,$1F,$99,$BE,$FD,$FB,$17,$0B,$DD,$6A,$7F,$97,$F0,$7D,$2C
	.dc.b	$7A,$63,$E4,$7B,$70,$5C,$38,$FC,$D3,$3E,$DF,$90,$7C,$DF,$2F,$F3
	.dc.b	$3E,$6F,$E7,$FC,$9F,$C8,$7C,$EE,$7C,$CF,$D2,$4E,$7C,$BF,$F7,$4E
	.dc.b	$FF,$82,$C3,$F3,$BD,$1F,$89,$3F,$34,$57,$AE,$3F,$AA,$D3,$EE,$7D
	.dc.b	$BF,$F9,$FF,$F2,$BF,$E1,$1F,$8B,$BF,$21,$7E,$B2,$FC,$B3,$7F,$5A
	.dc.b	$BD,$D6,$BD,$7C,$C7,$D4,$C7,$FD,$15,$FA,$1D,$BE,$1F,$00,$CF,$B6
	.dc.b	$6F,$B3,$E3,$39,$CF,$C3,$FB,$17,$E4,$7E,$77,$FD,$D5,$EE,$F1,$FC
	.dc.b	$1F,$FC,$77,$E3,$7F,$BC,$7C,$5F,$E7,$35,$AD,$15,$AF,$DB,$B9,$7D
	.dc.b	$DA,$BF,$12,$74,$E8,$F9,$0F,$1A,$EB,$3F,$A1,$F0,$FF,$D4,$8B,$F4
	.dc.b	$17,$E3,$EF,$B2,$FE,$8F,$E0,$6F,$97,$FE,$00,$56,$A7,$FA,$5F,$99
	.dc.b	$9C,$3F,$33,$8F,$FA,$BF,$F0,$AF,$CD,$FC,$C3,$F4,$07,$D0,$FF,$A8
	.dc.b	$BF,$58,$7E,$4B,$FE,$91,$F4,$7F,$69,$7B,$AD,$4F,$FA,$57,$CC,$CE
	.dc.b	$1C,$B9,$3F,$09,$7D,$73,$EF,$3D,$BE,$AD,$9F,$6B,$7D,$FF,$F0,$1E
	.dc.b	$97,$43,$F6,$3F,$B1,$ED,$3F,$30,$FE,$71,$FA,$16,$EB,$5E,$1F,$FA
	.dc.b	$8A,$7F,$9E,$7B,$39,$F8,$AF,$DC,$E5,$D3,$2F,$DD,$FD,$FF,$FB,$07
	.dc.b	$DD,$7E,$B0,$FD,$25,$FB,$B3,$F6,$CF,$EB,$FE,$AF,$F2,$2C,$38,$5E
	.dc.b	$EB,$5F,$97,$FF,$19,$9E,$9C,$9F,$A5,$7F,$D9,$DD,$7E,$0F,$A0,$7E
	.dc.b	$58,$FD,$91,$F9,$BB,$F2,$BF,$F1,$67,$CA,$73,$BD,$D6,$A7,$FC,$61
	.dc.b	$EF,$77,$C9,$F9,$9F,$F8,$17,$D6,$7E,$59,$FD,$6F,$F3,$DD,$4E,$63
	.dc.b	$FA,$F7,$E7,$BC,$7F,$26,$F8,$78,$72,$E1,$F4,$2F,$AF,$FE,$A5,$FB
	.dc.b	$A7,$F3,$6F,$AC,$7E,$EB,$E1,$9C,$FC,$B7,$FB,$83,$F4,$0F,$26,$0F
	.dc.b	$D4,$5F,$AC,$BF,$2B,$7E,$AB,$FC,$73,$C4,$FF,$11,$D8,$97,$C2,$B7
	.dc.b	$F1,$FD,$3E,$09,$FA,$37,$C9,$F4,$DF,$9E,$E6,$31,$FA,$CF,$8B,$EA
	.dc.b	$BF,$C1,$3D,$6F,$31,$FA,$F6,$7E,$3B,$F8,$8C,$3C,$AF,$4F,$03,$EE
	.dc.b	$3D,$FE,$13,$F0,$DC,$C6,$3F,$0E,$DD,$3F,$48,$7F,$00,$BE,$3B,$3F
	.dc.b	$31,$FF,$00,$35,$BF,$8B,$FF,$1F,$D7,$F1,$98,$FC,$12,$7E,$7E,$F8
	.dc.b	$2E,$B0,$8F,$EA,$33,$C2,$38,$1F,$27,$F5,$DD,$AF,$D7,$3D,$8F,$F9
	.dc.b	$43,$CD,$FF,$28,$F9,$13,$F5,$3C,$A3,$FC,$46,$68,$C7,$E9,$3A,$72
	.dc.b	$67,$F4,$7C,$56,$E1,$FA,$07,$F1,$9C,$FE,$E4,$75,$9D,$4B,$F4,$5E
	.dc.b	$33,$DF,$F5,$FF,$ED,$17,$C2,$3B,$9F,$E6,$88,$7C,$6F,$08,$F5,$EE
	.dc.b	$FD,$0B,$11,$FE,$6E,$87,$E8,$98,$8F,$F0,$AF,$E7,$EF,$92,$C7,$DA
	.dc.b	$A1,$F2,$7B,$FB,$57,$EF,$EF,$D8,$F0,$BF,$53,$EF,$FA,$16,$BD,$5F
	.dc.b	$A1,$C5,$E9,$E5,$F9,$F3,$9F,$CC,$3C,$BF,$2B,$FE,$04,$FC,$CF,$F5
	.dc.b	$D7,$BE,$1F,$F4,$2C,$E7,$BF,$DD,$F9,$1B,$F4,$87,$E6,$7F,$E1,$8F
	.dc.b	$AB,$CC,$63,$E5,$1F,$93,$7C,$5F,$DC,$DE,$D3,$F7,$AF,$F0,$A7,$E5
	.dc.b	$99,$8C,$7D,$0F,$E6,$38,$3C,$5C,$0F,$DA,$3F,$73,$93,$E9,$3F,$75
	.dc.b	$5A,$9C,$C7,$FA,$0E,$3D,$4F,$F9,$B3,$F0,$97,$EE,$AF,$D3,$9F,$A9
	.dc.b	$FE,$02,$B5,$39,$FD,$7F,$F0,$7F,$85,$7F,$A8,$BE,$C3,$F7,$1F,$95
	.dc.b	$FF,$1C,$CF,$E3,$4E,$7D,$F4,$FF,$A4,$27,$E5,$7D,$77,$E4,$DD,$F0
	.dc.b	$7E,$6D,$FA,$9D,$4F,$C8,$1F,$97,$7F,$60,$FA,$07,$EB,$6F,$D6,$7F
	.dc.b	$3B,$7B,$AD,$4F,$C2,$F6,$7F,$59,$2F,$D4,$BC,$11,$8C,$6E,$7F,$A1
	.dc.b	$FA,$6F,$E1,$99,$BF,$6F,$7E,$F4,$FE,$00,$FE,$A3,$BF,$E2,$6F,$75
	.dc.b	$A9,$F8,$59,$7E,$EE,$7F,$8B,$63,$ED,$8C,$7E,$49,$E0,$8F,$F4,$EF
	.dc.b	$CF,$7F,$5E,$FD,$19,$2F,$CE,$DF,$08,$AF,$DD,$9B,$9D,$43,$F7,$FC
	.dc.b	$FD,$6F,$5F,$4B,$87,$EA,$5D,$FC,$6F,$F0,$8F,$CF,$BE,$17,$F9,$1F
	.dc.b	$88,$FD,$67,$F8,$55,$6A,$73,$E6,$E3,$F2,$7F,$C7,$71,$7A,$CF,$F0
	.dc.b	$73,$FE,$B8,$70,$7F,$9B,$5F,$CA,$63,$DF,$E3,$FF,$1B,$FA,$E7,$F1
	.dc.b	$7F,$86,$FF,$47,$F3,$3F,$2E,$7E,$EB,$EF,$F2,$BF,$9C,$76,$FF,$0F
	.dc.b	$BD,$D6,$BA,$14,$FB,$BE,$9F,$05,$FC,$57,$7F,$F7,$DE,$3C,$7F,$57
	.dc.b	$FE,$45,$1F,$C9,$F9,$5F,$BA,$3F,$34,$7E,$DA,$FD,$A5,$FA,$3B,$F3
	.dc.b	$67,$EC,$3C,$70,$E1,$7B,$AD,$7E,$E1,$9C,$7E,$F6,$5D,$BF,$7E,$FC
	.dc.b	$07,$86,$FD,$97,$FB,$BF,$DF,$BF,$4B,$FF,$5D,$7E,$B8,$F0,$3F,$9E
	.dc.b	$5E,$FE,$D7,$FC,$43,$5A,$FC,$07,$38,$FF,$5C,$E4,$FA,$37,$B6,$C7
	.dc.b	$6A,$EB,$E2,$FF,$47,$FC,$B1,$F9,$67,$FB,$37,$F0,$1F,$EE,$5F,$96
	.dc.b	$3F,$2D,$D4,$E7,$DF,$4B,$F3,$9C,$F9,$5F,$9B,$C7,$9F,$E7,$37,$0E
	.dc.b	$17,$FE,$D7,$F2,$3F,$AD,$6F,$FD,$FF,$CE,$FD,$A7,$E7,$0F,$A9,$F9
	.dc.b	$DF,$9A,$BF,$56,$FE,$B9,$BA,$F7,$5A,$FC,$53,$30,$FD,$4F,$F5,$95
	.dc.b	$BF,$81,$FD,$75,$FF,$9D,$7F,$40,$7E,$DE,$FD,$8D,$F8,$8B,$C5,$7E
	.dc.b	$01,$FD,$AF,$0E,$17,$BA,$D7,$E9,$19,$FC,$6F,$F9,$47,$1F,$E8,$2F
	.dc.b	$B3,$61,$9F,$83,$3E,$F9,$FF,$A4,$7D,$0F,$ED,$5F,$9D,$7F,$96,$FF
	.dc.b	$14,$FF,$36,$7F,$00,$7F,$BC,$BF,$21,$7C,$C7,$F7,$DA,$D7,$E2,$EA
	.dc.b	$D7,$E7,$59,$F8,$A7,$EF,$3F,$C3,$6F,$A7,$07,$ED,$3F,$39,$FF,$7F
	.dc.b	$7B,$7F,$EF,$DF,$B1,$FF,$04,$46,$3F,$38,$7E,$77,$FF,$00,$38,$7F
	.dc.b	$BF,$4F,$F8,$C7,$F3,$BF,$E3,$0F,$FB,$FB,$1E,$93,$FE,$FD,$C5,$87
	.dc.b	$FB,$F6,$3F,$C0,$DF,$B9,$FF,$B2,$BF,$0A,$63,$FE,$89,$FD,$1D,$FA
	.dc.b	$87,$F9,$62,$F7,$F8,$62,$F7,$C8,$BD,$FC,$67,$F8,$7F,$2E,$9A,$FF
	.dc.b	$10,$7F,$DA,$FF,$89,$FF,$2D,$FB,$EF,$ED,$9F,$0A,$E7,$FF,$7F,$FE
	.dc.b	$38,$FD,$79,$FB,$87,$F8,$A6,$F7,$5A,$97,$EA,$F9,$FB,$07,$E0,$7D
	.dc.b	$5F,$E0,$9A,$FF,$20,$FC,$07,$FC,$2D,$FA,$AB,$F5,$17,$E7,$8D,$9F
	.dc.b	$A0,$FF,$C0,$1E,$8F,$F1,$56,$1C,$2F,$75,$AF,$AB,$FF,$1E,$4F,$F8
	.dc.b	$87,$E2,$7F,$A4,$3D,$B3,$F0,$7E,$0A,$D3,$C1,$A7,$7D,$3F,$DC,$3A
	.dc.b	$7F,$8D,$DF,$F0,$34,$3E,$E3,$FC,$D5,$F6,$CF,$CB,$5F,$A1,$7F,$92
	.dc.b	$EF,$75,$A9,$F8,$78,$73,$FB,$6F,$6E,$9F,$D0,$2F,$FC,$25,$F9,$A7
	.dc.b	$F5,$D7,$8F,$E2,$BF,$E2,$0C,$3F,$C9,$3F,$05,$FC,$0A,$73,$A7,$87
	.dc.b	$E8,$3F,$CB,$5E,$E7,$F6,$1F,$C5,$FF,$C0,$5E,$77,$FA,$47,$E3,$DF
	.dc.b	$82,$EB,$EB,$3F,$D6,$9F,$64,$AD,$70,$E7,$D1,$FE,$DD,$97,$3E,$2F
	.dc.b	$EA,$BF,$2B,$FC,$77,$F0,$9F,$CB,$FE,$25,$F5,$3F,$63,$7E,$B3,$FE
	.dc.b	$6B,$AD,$7A,$8A,$D7,$26,$7E,$53,$FC,$81,$97,$47,$FC,$A4,$FF,$62
	.dc.b	$FD,$05,$F5,$CF,$C7,$DF,$A1,$BF,$3E,$7E,$84,$FD,$0D,$FA,$27,$F9
	.dc.b	$EF,$1E,$38,$70,$BD,$D7,$F3,$8F,$0B,$FC,$E5,$5A,$CF,$C1,$9F,$D4
	.dc.b	$7F,$D1,$73,$A8,$F1,$67,$E1,$CF,$C1,$F8,$47,$FA,$27,$F3,$B7,$E5
	.dc.b	$6F,$D4,$9F,$AA,$7E,$1F,$F0,$C7,$EC,$6F,$D6,$38,$5E,$EB,$52,$FC
	.dc.b	$E3,$38,$7F,$87,$F1,$EB,$FA,$AD,$C3,$87,$9B,$FE,$5B,$C3,$FB,$D7
	.dc.b	$07,$FA,$37,$F7,$BF,$EE,$3F,$C0,$1F,$E6,$CF,$71,$FA,$87,$D3,$FC
	.dc.b	$F3,$FD,$41,$5A,$FC,$63,$F7,$93,$F5,$FE,$BF,$3F,$6C,$FD,$13,$FC
	.dc.b	$61,$7D,$3A,$BF,$C2,$EF,$1C,$BA,$F8,$4F,$DE,$7F,$D5,$9F,$56,$FC
	.dc.b	$7F,$FD,$1D,$F9,$EF,$EE,$BF,$20,$FD,$E7,$93,$A9,$CF,$90,$9F,$E0
	.dc.b	$11,$9F,$C7,$3F,$C8,$18,$73,$F3,$6C,$F0,$AF,$E0,$DF,$1F,$FE,$BB
	.dc.b	$FD,$31,$FA,$37,$F4,$2F,$C3,$1F,$A6,$4F,$CB,$9F,$A6,$EE,$B5,$39
	.dc.b	$FD,$3F,$1F,$23,$FE,$4F,$C7,$9F,$D1,$DB,$DD,$B1,$FC,$A7,$81,$FE
	.dc.b	$BD,$FC,$AD,$F9,$CF,$E8,$7F,$32,$FE,$A7,$FB,$EF,$F6,$AF,$6B,$FD
	.dc.b	$A1,$5A,$EB,$FF,$76,$68,$AF,$ED,$FF,$A5,$6B,$17,$EB,$7F,$E6,$1C
	.dc.b	$38,$78,$1F,$E4,$87,$96,$7D,$BE,$CF,$F9,$3E,$BE,$3C,$3C,$59,$F5
	.dc.b	$97,$EB,$1F,$EE,$1F,$D4,$3F,$D4,$9E,$CB,$F1,$07,$E4,$3F,$EC,$38
	.dc.b	$FF,$51,$7C,$77,$F8,$4A,$B7,$5A,$FC,$BD,$5F,$E4,$4D,$3F,$DB,$D5
	.dc.b	$D3,$7F,$F2,$EB,$FE,$B1,$FB,$5F,$F3,$D3,$CB,$2E,$DB,$7F,$BC,$FC
	.dc.b	$D7,$EC,$BF,$C6,$3F,$4D,$7E,$E8,$FB,$2F,$F2,$6F,$EA,$FB,$AD,$4E
	.dc.b	$7D,$1F,$F4,$96,$DF,$EE,$FC,$79,$FC,$6F,$FA,$25,$EB,$F2,$37,$0E
	.dc.b	$39,$7F,$96,$3F,$72,$61,$FE,$F7,$FC,$C5,$DA,$E1,$7E,$A7,$FA,$5F
	.dc.b	$D4,$DF,$62,$FD,$57,$8F,$5F,$F8,$E3,$1E,$38,$70,$F6,$58,$70,$FB
	.dc.b	$FD,$F5,$05,$F0,$7F,$D0,$AF,$8B,$F5,$5F,$F5,$13,$D3,$4F,$16,$8F
	.dc.b	$D4,$B9,$DF,$2D,$BD,$15,$AF,$0F,$3E,$7F,$EF,$7E,$67,$FD,$4F,$FB
	.dc.b	$53,$E8,$BF,$A3,$F2,$6B,$87,$FD,$99,$F0,$0A,$D7,$4B,$F4,$6D,$7C
	.dc.b	$D7,$FA,$D7,$3E,$9B,$FF,$A7,$1E,$D8,$71,$CF,$FA,$9F,$E3,$BE,$6F
	.dc.b	$D1,$7F,$C2,$DF,$60,$FD,$27,$FC,$FD,$1F,$E7,$DF,$8F,$D7,$DC,$7F
	.dc.b	$82,$A7,$3F,$8C,$7F,$85,$78,$9D,$DF,$E6,$6C,$39,$ED,$FE,$AC,$7F
	.dc.b	$C4,$5F,$A5,$7F,$B2,$5E,$F8,$72,$F9,$1F,$8F,$F9,$B3,$E3,$FF,$99
	.dc.b	$7E,$DA,$FC,$71,$FB,$DB,$F1,$6F,$F1,$F6,$BF,$E7,$4C,$38,$7C,$C3
	.dc.b	$0E,$17,$BF,$C1,$BA,$E5,$FC,$F7,$7F,$F5,$FE,$1D,$FE,$6F,$FD,$B8
	.dc.b	$F2,$C3,$8E,$7F,$E6,$EE,$1F,$6D,$FE,$17,$F0,$97,$DB,$7F,$9C,$3C
	.dc.b	$07,$E7,$5F,$E2,$9E,$B7,$ED,$DF,$82,$BD,$FE,$25,$FC,$EE,$B5,$FA
	.dc.b	$BE,$71,$57,$FB,$B4,$DE,$5F,$ED,$47,$8E,$5D,$BE,$FB,$F0,$D5,$F8
	.dc.b	$EF,$8D,$7E,$D6,$FD,$63,$F9,$6F,$E0,$9F,$A3,$FA,$17,$C9,$6E,$B5
	.dc.b	$E3,$FF,$92,$BD,$7F,$EC,$BF,$71,$9F,$D0,$FF,$DF,$D9,$F4,$CE,$E1
	.dc.b	$C3,$3E,$98,$FE,$7B,$E4,$5E,$DB,$F0,$3F,$F6,$EF,$98,$FC,$DB,$FD
	.dc.b	$31,$FA,$4E,$EB,$5F,$97,$FF,$B6,$A7,$FD,$39,$E9,$E7,$F1,$0F,$F5
	.dc.b	$7F,$4F,$FB,$8B,$C8,$FF,$58,$E9,$DB,$4C,$CF,$ED,$FC,$0F,$AA,$FF
	.dc.b	$43,$CF,$AE,$BE,$1F,$C5,$9F,$0B,$FB,$4F,$B5,$7E,$F1,$FE,$6B,$FC
	.dc.b	$39,$E5,$FF,$48,$CB,$F5,$D9,$F4,$BD,$FD,$4A,$F7,$F1,$DF,$C1,$FF
	.dc.b	$07,$E9,$93,$F4,$2F,$2E,$58,$C2,$E5,$E6,$FD,$15,$F9,$03,$F5,$BF
	.dc.b	$F2,$D7,$81,$F9,$FC,$1F,$83,$7F,$0A,$BD,$D6,$BD,$6F,$E2,$DF,$77
	.dc.b	$9C,$79,$61,$F1,$32,$ED,$1F,$E3,$98,$C7,$4F,$FB,$6B,$1E,$7F,$83
	.dc.b	$B4,$F0,$7C,$B7,$F1,$8F,$D8,$5E,$3A,$DF,$97,$BF,$04,$7C,$CF,$FB
	.dc.b	$CB,$C4,$E1,$7B,$AD,$4F,$FB,$DB,$F0,$0C,$FE,$AD,$F0,$B6,$FF,$75
	.dc.b	$E7,$DB,$3C,$6F,$FC,$C6,$31,$8F,$2F,$B1,$FF,$0A,$71,$7D,$EF,$1F
	.dc.b	$FD,$E1,$ED,$BF,$52,$7E,$97,$FE,$41,$FD,$ED,$FA,$37,$F3,$2B,$DD
	.dc.b	$6B,$E2,$D4,$FD,$9F,$5F,$35,$FE,$D5,$DF,$E3,$9F,$D0,$F0,$EE,$E0
	.dc.b	$FF,$5F,$79,$CF,$B3,$7E,$62,$FD,$CD,$F8,$D7,$F2,$BF,$DA,$3F,$A1
	.dc.b	$FE,$3E,$C3,$85,$EE,$BF,$98,$7E,$F3,$A9,$7E,$9F,$F1,$33,$EF,$9D
	.dc.b	$C7,$93,$E0,$CF,$BF,$C8,$BF,$03,$FE,$89,$7F,$D6,$CF,$2F,$D0,$BF
	.dc.b	$67,$7C,$7F,$E4,$7E,$95,$F8,$4F,$A7,$E7,$BF,$63,$C5,$FD,$B2,$F7
	.dc.b	$E7,$FF,$68,$A7,$E2,$DD,$B3,$FE,$D1,$7F,$DC,$BF,$34,$FD,$0E,$3C
	.dc.b	$B4,$F0,$57,$EE,$23,$19,$74,$CF,$C3,$F0,$DC,$EB,$FD,$BB,$FB,$97
	.dc.b	$F8,$03,$F0,$3F,$F3,$CF,$D4,$BD,$AF,$F7,$26,$5F,$C1,$BE,$21,$7B
	.dc.b	$F8,$DF,$EF,$34,$FB,$8E,$95,$AF,$A9,$F2,$E7,$F1,$63,$8F,$DE,$E2
	.dc.b	$3F,$C0,$51,$BD,$3C,$38,$7F,$A3,$73,$FE,$2B,$F9,$0B,$F6,$97,$E8
	.dc.b	$7E,$5F,$19,$F9,$CB,$F7,$3F,$C1,$E1,$C2,$F7,$8F,$E4,$6B,$5F,$AB
	.dc.b	$7C,$1C,$BA,$65,$E3,$D7,$FD,$F7,$83,$F4,$0C,$56,$B1,$E5,$9B,$F7
	.dc.b	$BF,$95,$F0,$4F,$CF,$7F,$90,$FE,$92,$FD,$2B,$F3,$1E,$C3,$F4,$0D
	.dc.b	$D6,$A7,$3F,$A0,$E3,$E4,$7C,$58,$F3,$FC,$32,$DE,$FC,$CF,$E0,$BF
	.dc.b	$A5,$7D,$9F,$8C,$7E,$1F,$FE,$2E,$EC,$FF,$15,$FE,$B0,$FD,$FF,$C9
	.dc.b	$D9,$FC,$FA,$F7,$E3,$2F,$74,$FD,$13,$A6,$2F,$DD,$3E,$26,$1C,$3B
	.dc.b	$3F,$18,$7E,$7F,$E3,$7F,$CC,$7F,$86,$6F,$F1,$EE,$1F,$AA,$BF,$60
	.dc.b	$FC,$75,$6B,$1F,$F6,$48,$97,$E1,$3A,$FC,$37,$FC,$EF,$97,$63,$FC
	.dc.b	$8D,$7F,$E7,$0F,$C0,$BE,$B3,$CF,$2E,$FF,$2D,$F9,$BF,$CA,$BF,$A8
	.dc.b	$F7,$7F,$94,$3E,$0D,$2F,$A1,$7E,$F7,$BE,$FF,$A9,$D0,$BD,$D3,$F6
	.dc.b	$3D,$63,$F8,$5C,$BF,$AE,$F6,$5C,$78,$E1,$C3,$CC,$FD,$67,$88,$FA
	.dc.b	$4F,$C5,$7E,$98,$FD,$3B,$DF,$F9,$CB,$F4,$ED,$D6,$A7,$3F,$A1,$E3
	.dc.b	$D0,$FC,$A6,$3C,$FF,$4E,$FC,$B7,$CF,$FE,$57,$E8,$BF,$67,$FB,$1B
	.dc.b	$F7,$37,$E3,$0F,$C6,$7F,$7D,$7B,$F8,$15,$EE,$9E,$6E,$B4,$7F,$00
	.dc.b	$76,$7F,$B4,$F8,$5F,$4C,$63,$CF,$0E,$5E,$F7,$F0,$BF,$A3,$FF,$44
	.dc.b	$FD,$07,$FA,$73,$F3,$57,$E5,$AF,$D1,$9F,$B3,$6F,$E1,$97,$BA,$D7
	.dc.b	$E9,$19,$FD,$C3,$F6,$99,$74,$DB,$C5,$F9,$4F,$F8,$EB,$2F,$D8,$7E
	.dc.b	$4C,$FD,$95,$FA,$BF,$EE,$7D,$CF,$F0,$ED,$EF,$DE,$FF,$68,$56,$BB
	.dc.b	$7F,$91,$4F,$3C,$BA,$FE,$0E,$DB,$F8,$27,$E3,$1F,$CC,$BF,$2F,$7E
	.dc.b	$AA,$FC,$61,$FA,$5B,$F4,$57,$C4,$7F,$2D,$C3,$85,$EE,$B5,$F7,$39
	.dc.b	$F9,$DF,$EB,$9A,$FC,$DB,$EC,$D8,$67,$E0,$CF,$BE,$7F,$C2,$F8,$BF
	.dc.b	$C7,$3F,$9B,$7F,$B2,$3E,$7F,$FC,$E9,$F0,$EF,$C9,$FF,$09,$F3,$3F
	.dc.b	$CF,$EB,$5F,$65,$FB,$F9,$FC,$47,$EE,$39,$FC,$A7,$FA,$AE,$FF,$7E
	.dc.b	$FF,$DA,$7F,$8F,$FF,$25,$78,$E5,$D7,$CF,$FE,$8B,$5F,$D8,$7F,$36
	.dc.b	$FF,$7E,$7E,$D3,$FE,$EC,$FB,$57,$E2,$EF,$DB,$7F,$D9,$7C,$5F,$E5
	.dc.b	$0B,$DF,$E5,$BF,$E0,$DA,$D7,$E0,$FF,$D5,$67,$9E,$5F,$91,$5E,$2B
	.dc.b	$F4,$AF,$E7,$1F,$52,$F3,$DF,$AF,$BE,$D9,$FA,$EF,$FA,$FF,$C5,$FF
	.dc.b	$3C,$EF,$FE,$60,$C3,$85,$EE,$B5,$F9,$87,$F6,$59,$E9,$F8,$2F,$6C
	.dc.b	$BA,$ED,$FA,$3F,$B0,$7E,$05,$FD,$07,$F6,$3F,$F0,$AF,$EA,$CF,$D6
	.dc.b	$DF,$A3,$FF,$6F,$BD,$FA,$4B,$DD,$7F,$5A,$E0,$FF,$39,$D6,$B2,$ED
	.dc.b	$97,$F3,$56,$E1,$C7,$ED,$99,$F7,$FD,$C7,$FB,$E7,$E7,$4F,$D3,$9F
	.dc.b	$91,$BF,$6F,$7E,$47,$FC,$0B,$F6,$A7,$E8,$FC,$2F,$75,$AF,$85,$57
	.dc.b	$BB,$87,$FD,$0F,$97,$5C,$BF,$B4,$3D,$FF,$28,$B8,$F2,$CF,$B7,$E5
	.dc.b	$BF,$E0,$FF,$7D,$FA,$37,$C5,$FC,$11,$FB,$A3,$F4,$57,$CD,$7C,$8D
	.dc.b	$BF,$D2,$B8,$70,$EA,$FF,$18,$5E,$EB,$E4,$74,$7F,$8C,$6B,$AF,$E6
	.dc.b	$7F,$DE,$DE,$11,$FC,$65,$E5,$97,$83,$2E,$FC,$DF,$E2,$3F,$D5,$7F
	.dc.b	$BC,$7E,$99,$FE,$D0,$FB,$DF,$C5,$7D,$4B,$D6,$D7,$DD,$2B,$5E,$5F
	.dc.b	$F4,$EF,$77,$3C,$5F,$E0,$BC,$B9,$EF,$FC,$A5,$EB,$F9,$6F,$F8,$B6
	.dc.b	$FF,$AA,$9F,$E3,$3F,$01,$FA,$EB,$F3,$9F,$F7,$97,$E2,$8F,$A4,$7E
	.dc.b	$BE,$FD,$7F,$85,$EF,$F3,$55,$EE,$BF,$D4,$3D,$9F,$EB,$EC,$BA,$FE
	.dc.b	$BF,$FE,$1F,$7B,$E7,$FE,$21,$FC,$C3,$FD,$9B,$F4,$DF,$D9,$A3,$FA
	.dc.b	$C7,$CD,$BF,$6E,$FF,$24,$56,$BE,$D3,$CD,$F5,$2A,$FE,$21,$9F,$FB
	.dc.b	$32,$BF,$DA,$FF,$A7,$F5,$FC,$B5,$9D,$3B,$61,$C7,$E5,$BF,$CA,$9E
	.dc.b	$DB,$D6,$FF,$54,$1F,$DE,$7F,$49,$FE,$7B,$F8,$B6,$A3,$3F,$6E,$9F
	.dc.b	$E8,$3F,$AF,$E7,$F7,$9F,$F1,$E6,$1C,$FF,$2A,$FF,$1F,$74,$BE,$D3
	.dc.b	$FB,$3F,$0B,$F0,$FF,$C9,$FE,$F3,$FE,$C7,$E4,$FF,$72,$DE,$FE,$C3
	.dc.b	$FE,$00,$D6,$BF,$11,$FF,$35,$4F,$3F,$C2,$BA,$E5,$FD,$CB,$C6,$FE
	.dc.b	$D1,$E0,$7F,$74,$FD,$85,$FA,$67,$F8,$4F,$F6,$2F,$87,$F2,$5F,$E7
	.dc.b	$3A,$D4,$E7,$AD,$CB,$9D,$DF,$B5,$FF,$95,$EF,$A6,$7D,$A5,$9F,$6E
	.dc.b	$86,$9F,$43,$EA,$3F,$D0,$1F,$9E,$7F,$45,$FD,$AF,$FA,$F7,$97,$FB
	.dc.b	$1F,$CC,$5D,$7A,$FA,$15,$6B,$D9,$FF,$6A,$76,$66,$3F,$CD,$18,$FF
	.dc.b	$4F,$FB,$9B,$7C,$3E,$57,$F3,$18,$7F,$B6,$FA,$BF,$D2,$1F,$A5,$BF
	.dc.b	$46,$7E,$07,$FA,$0F,$11,$FA,$47,$F9,$73,$EC,$77,$5A,$F5,$3E,$F4
	.dc.b	$FF,$98,$3E,$D7,$38,$7F,$9E,$71,$E9,$F8,$BD,$C3,$86,$5D,$3F,$15
	.dc.b	$FF,$03,$78,$FF,$EB,$7E,$D7,$EB,$4F,$D3,$FF,$C0,$1E,$EF,$F8,$5F
	.dc.b	$F2,$B7,$6F,$FA,$77,$0E,$17,$BE,$55,$EE,$BF,$B1,$E2,$3A,$E2,$FD
	.dc.b	$93,$FD,$1D,$87,$19,$FF,$16,$7C,$63,$CF,$FF,$80,$3F,$BC,$FF,$82
	.dc.b	$FF,$75,$7E,$92,$FD,$2F,$FD,$71,$7B,$FC,$2D,$7B,$A7,$E9,$3A,$C5
	.dc.b	$5F,$E9,$DB,$ED,$F8,$3F,$FA,$7F,$F0,$A7,$EF,$5E,$5F,$C1,$5F,$5D
	.dc.b	$FE,$F7,$F8,$2F,$F0,$77,$07,$FB,$76,$F7,$C4,$BD,$FE,$51,$FC,$07
	.dc.b	$F0,$3F,$F6,$65,$74,$CB,$B1,$F0,$70,$FF,$25,$47,$2B,$E3,$8B,$F1
	.dc.b	$2F,$F6,$A7,$D5,$FF,$42,$F4,$1F,$B8,$BF,$70,$7E,$67,$F5,$FA,$B7
	.dc.b	$F3,$DF,$C9,$6B,$53,$F6,$F9,$33,$87,$FA,$F7,$1F,$E9,$BE,$4F,$FA
	.dc.b	$FF,$F0,$77,$E7,$6F,$DA,$5F,$B5,$3C,$6F,$F0,$67,$E3,$2F,$D1,$FF
	.dc.b	$DD,$18,$70,$BD,$D6,$BF,$2B,$CE,$3C,$F9,$FF,$CD,$17,$DB,$EB,$FC
	.dc.b	$13,$DF,$3E,$DA,$B4,$FD,$BF,$95,$F9,$5F,$DD,$1F,$A8,$7F,$4F,$FC
	.dc.b	$71,$FB,$87,$FA,$A7,$F6,$B7,$E5,$7F,$EF,$BB,$DF,$D2,$7F,$A3,$6B
	.dc.b	$5E,$E6,$71,$FE,$B1,$C3,$FE,$7B,$BE,$BF,$47,$FE,$E1,$78,$E7,$DB
	.dc.b	$F2,$96,$8F,$C9,$DB,$7E,$A7,$F6,$D7,$E9,$4F,$DE,$3F,$D5,$DF,$6E
	.dc.b	$CD,$F9,$8F,$F2,$BD,$BF,$48,$30,$E3,$7C,$2B,$7F,$A6,$F6,$AD,$D6
	.dc.b	$9F,$F4,$93,$E0,$FD,$93,$FD,$F1,$F8,$37,$D7,$FD,$8F,$FC,$83,$F0
	.dc.b	$0F,$D3,$3F,$C5,$DC,$1C,$F1,$FD,$A5,$EF,$F0,$9F,$E9,$15,$FE,$33
	.dc.b	$FA,$AD,$57,$F9,$23,$C3,$65,$D9,$FF,$2D,$47,$FA,$8F,$F4,$F7,$E8
	.dc.b	$1F,$94,$F1,$BF,$81,$FD,$AB,$F6,$EF,$F3,$CF,$47,$F0,$FB,$DF,$9B
	.dc.b	$F8,$FF,$85,$FF,$06,$57,$C6,$AD,$63,$F9,$EC,$BB,$3F,$49,$FF,$59
	.dc.b	$7E,$5F,$FE,$61,$F2,$1F,$A6,$7E,$67,$CB,$7E,$CA,$FC,$93,$D7,$7B
	.dc.b	$AD,$4F,$FA,$97,$0F,$A7,$3F,$E6,$BC,$BF,$92,$47,$F1,$97,$F9,$34
	.dc.b	$77,$8F,$0F,$E2,$3F,$A8,$BE,$7F,$A6,$7C,$8F,$B3,$FF,$3A,$53,$F1
	.dc.b	$87,$E3,$FF,$AE,$FD,$35,$FB,$B6,$F2,$FE,$4D,$5A,$E0,$53,$F7,$5D
	.dc.b	$63,$E4,$38,$7E,$35,$77,$FE,$75,$A7,$83,$1E,$58,$3F,$7D,$FB,$7F
	.dc.b	$5F,$F5,$BF,$69,$7E,$FB,$FA,$CF,$C3,$7F,$AF,$FC,$1E,$BA,$55,$AF
	.dc.b	$51,$5F,$D9,$3E,$C7,$53,$E1,$E0,$FC,$DE,$5F,$82,$BC,$47,$F4,$18
	.dc.b	$F0,$65,$DF,$EA,$3F,$CB,$FC,$1F,$D4,$7F,$4F,$7E,$AF,$FE,$3D,$FD
	.dc.b	$71,$E8,$7F,$A7,$7F,$57,$7E,$8B,$F9,$4B,$DF,$C3,$AF,$7D,$6B,$FF
	.dc.b	$0D,$65,$D7,$2F,$ED,$7B,$7F,$BE,$71,$E2,$FF,$BE,$FF,$47,$7E,$52
	.dc.b	$FC,$85,$FA,$D7,$F5,$7F,$CD,$BF,$AE,$FD,$2F,$E9,$2F,$75,$AF,$99
	.dc.b	$FF,$7C,$F9,$8F,$99,$D6,$5C,$F3,$EC,$BF,$06,$F8,$B1,$FE,$94,$EA
	.dc.b	$7E,$F5,$FD,$43,$E1,$FF,$D4,$94,$FD,$C7,$F6,$7F,$8E,$FD,$4A,$D7
	.dc.b	$E4,$9C,$FF,$04,$CE,$2F,$CA,$FA,$DE,$F3,$FC,$DF,$3F,$F7,$F8,$E5
	.dc.b	$F4,$3F,$E8,$FC,$7C,$9F,$A5,$7F,$00,$FC,$83,$FD,$F5,$F6,$4F,$C7
	.dc.b	$3F,$C6,$FF,$3D,$FE,$A9,$FC,$D7,$5F,$99,$7F,$86,$27,$3F,$36,$FC
	.dc.b	$C3,$3F,$6B,$EE,$CB,$9F,$07,$88,$FE,$43,$F8,$CB,$F1,$D7,$EE,$5F
	.dc.b	$EB,$4F,$C9,$3F,$DB,$9D,$7F,$92,$E5,$FF,$0B,$5E,$EB,$5E,$47,$FA
	.dc.b	$9F,$F7,$3F,$E2,$F3,$CF,$83,$E6,$BB,$CF,$C0,$7F,$84,$DE,$5F,$92
	.dc.b	$FF,$B1,$6F,$FD,$61,$E9,$7F,$20,$FD,$E3,$FC,$07,$F8,$C7,$F8,$12
	.dc.b	$1F,$BF,$3F,$39,$7D,$8E,$FE,$C9,$7B,$AD,$7D,$6E,$63,$EF,$E3,$FE
	.dc.b	$0E,$F7,$ED,$F0,$3F,$AE,$DF,$5F,$A9,$FE,$63,$F6,$2F,$C5,$DD,$3F
	.dc.b	$C2,$B9,$3F,$D8,$DF,$4E,$FD,$37,$75,$AF,$8B,$56,$BF,$25,$CF,$37
	.dc.b	$E4,$32,$E9,$97,$E4,$BF,$52,$7D,$FF,$F2,$AF,$1D,$F9,$63,$F5,$BF
	.dc.b	$0F,$27,$F4,$4B,$DF,$C9,$2F,$75,$E6,$CF,$58,$BF,$5E,$EC,$FF,$36
	.dc.b	$EE,$FA,$67,$FB,$7F,$CB,$FF,$7B,$7E,$82,$FD,$D7,$FA,$C7,$0F,$E8
	.dc.b	$7C,$4F,$F2,$B7,$8B,$FE,$66,$F1,$97,$BF,$A2,$5F,$F0,$0C,$BF,$B6
	.dc.b	$DE,$DE,$EF,$F9,$BB,$F0,$2F,$D8,$BC,$EF,$95,$7A,$F1,$EF,$9F,$93
	.dc.b	$3F,$1F,$EA,$5F,$D9,$30,$FE,$9F,$E9,$FF,$99,$FF,$5B,$7D,$57,$FB
	.dc.b	$63,$F0,$2D,$73,$2B,$53,$FE,$DC,$FB,$8C,$C7,$8A,$7E,$EF,$07,$99
	.dc.b	$E1,$F8,$3B,$1E,$BF,$86,$BF,$04,$FE,$EB,$F7,$6F,$41,$F8,$8F,$F9
	.dc.b	$E7,$D1,$FD,$67,$63,$F3,$CD,$D6,$B3,$56,$BF,$31,$CE,$1F,$C5,$31
	.dc.b	$E9,$FA,$4D,$C3,$85,$F4,$7F,$C0,$BF,$66,$FD,$81,$F6,$6F,$C2,$5F
	.dc.b	$B3,$BF,$52,$7C,$97,$D6,$EF,$FF,$48,$63,$C7,$0E,$17,$BA,$FC,$E5
	.dc.b	$6B,$1E,$BC,$FF,$88,$C3,$C1,$F2,$3F,$CA,$9E,$59,$F8,$33,$EF,$FA
	.dc.b	$8F,$9B,$0F,$2C,$E7,$27,$EA,$BF,$E1,$EF,$B8,$7E,$E8,$FD,$D7,$FC
	.dc.b	$7D,$F4,$EF,$D8,$FF,$80,$7C,$DE,$EB,$5B,$7F,$AA,$27,$F8,$27,$91
	.dc.b	$9F,$0F,$F9,$EE,$5D,$32,$FE,$7F,$F9,$9F,$FB,$BF,$F7,$8F,$C2,$7E
	.dc.b	$BE,$F8,$D7,$ED,$5F,$8A,$E8,$FF,$5A,$5E,$FC,$85,$EE,$BF,$C9,$D5
	.dc.b	$AF,$A9,$FE,$93,$D5,$FA,$CC,$3B,$36,$19,$78,$32,$EF,$FB,$7F,$F8
	.dc.b	$FB,$F7,$FF,$F5,$CF,$EA,$5A,$FC,$93,$BF,$E2,$EF,$C8,$B5,$F4,$6A
	.dc.b	$D7,$07,$FA,$5F,$F2,$2C,$C7,$F5,$CC,$7F,$D8,$7F,$90,$5B,$E1,$AB
	.dc.b	$3E,$BF,$A9,$70,$8C,$65,$FE,$A1,$F9,$BF,$F2,$8F,$EE,$BF,$88,$FC
	.dc.b	$3F,$FC,$9B,$FA,$87,$F9,$C7,$E6,$1F,$A9,$EB,$C5,$FF,$36,$FD,$CA
	.dc.b	$B5,$F9,$66,$70,$FE,$DF,$8F,$E4,$BF,$66,$B8,$70,$BE,$9E,$4F,$F9
	.dc.b	$83,$F5,$47,$E9,$AF,$D5,$9F,$30,$FC,$5D,$F0,$8F,$CC,$1F,$23,$C7
	.dc.b	$0E,$17,$BA,$D1,$9F,$C4,$FF,$C0,$39,$76,$CB,$FC,$07,$C1,$FE,$22
	.dc.b	$F8,$87,$7B,$F2,$CF,$F4,$FF,$F0,$03,$F6,$1F,$F3,$AD,$EE,$B5,$F9
	.dc.b	$02,$9F,$5B,$AF,$BC,$FD,$97,$97,$FC,$C3,$0E,$CD,$86,$7E,$0C,$FB
	.dc.b	$FE,$4D,$C6,$31,$97,$5F,$D2,$3F,$D0,$BD,$5D,$BF,$60,$FE,$2D,$E3
	.dc.b	$3F,$35,$FE,$67,$F9,$A6,$B7,$56,$A7,$F9,$BF,$C4,$E6,$3F,$C4,$78
	.dc.b	$7A,$3F,$37,$B7,$C3,$E4,$59,$75,$F9,$67,$C5,$FF,$A5,$BF,$58,$FE
	.dc.b	$E9,$FA,$C7,$FA,$3F,$F6,$77,$F0,$0F,$F8,$F3,$D9,$D7,$D9,$2B,$53
	.dc.b	$FE,$40,$FC,$4B,$31,$FE,$34,$C3,$FB,$F7,$D5,$5B,$E1,$F8,$E3,$2E
	.dc.b	$BF,$81,$BF,$08,$FF,$53,$FC,$53,$E5,$DF,$96,$FF,$B2,$39,$5F,$B9
	.dc.b	$FD,$4E,$3D,$D6,$BF,$2A,$56,$B9,$FF,$90,$F8,$7F,$EC,$6C,$BA,$7E
	.dc.b	$5F,$FE,$48,$7B,$67,$71,$E5,$87,$1F,$94,$FF,$5C,$FE,$6C,$FD,$09
	.dc.b	$F8,$03,$F2,$3F,$CD,$FA,$DF,$86,$FD,$AF,$75,$A9,$CE,$FF,$E6,$DF
	.dc.b	$8D,$FF,$6B,$F0,$32,$E9,$FA,$9F,$F9,$69,$ED,$9D,$C7,$97,$AD,$FA
	.dc.b	$7C,$5F,$9D,$FF,$B3,$3F,$39,$7E,$4F,$E3,$FC,$F1,$F9,$D3,$F3,$67
	.dc.b	$E6,$AF,$E0,$0E,$17,$BA,$D4,$BF,$47,$CF,$67,$F9,$B3,$2E,$BC,$97
	.dc.b	$0E,$19,$74,$F9,$2F,$F6,$EF,$9F,$FE,$19,$FC,$7D,$FB,$CF,$87,$F1
	.dc.b	$4F,$D8,$FE,$F5,$AF,$BF,$FF,$49,$CE,$7E,$01,$FA,$D6,$70,$FF,$3D
	.dc.b	$E3,$FE,$23,$FB,$DB,$87,$0D,$3F,$D5,$78,$3F,$58,$FF,$76,$F8,$7F
	.dc.b	$E1,$8F,$D2,$3F,$03,$F2,$FF,$D0,$FF,$26,$FF,$25,$F8,$7F,$42,$B4
	.dc.b	$7F,$3A,$FA,$0F,$F5,$CF,$E1,$3C,$BC,$D7,$E9,$5F,$E8,$CA,$E9,$E3
	.dc.b	$FF,$8B,$9E,$39,$75,$FC,$3F,$FD,$FB,$E2,$BF,$76,$FA,$1E,$13,$EA
	.dc.b	$5F,$AF,$FD,$2F,$33,$75,$A9,$CF,$89,$F7,$38,$7F,$0B,$8F,$3E,$6F
	.dc.b	$F4,$EB,$D7,$F0,$FB,$87,$1C,$BE,$EF,$EE,$0C,$3F,$1B,$FA,$C7,$F9
	.dc.b	$7F,$17,$EF,$FF,$A4,$FD,$99,$FC,$01,$FE,$71,$BD,$F3,$FE,$1E,$B5
	.dc.b	$3F,$E2,$79,$CE,$9F,$EA,$EC,$7A,$72,$FF,$94,$17,$ED,$6E,$27,$EB
	.dc.b	$FF,$E4,$CF,$D6,$1F,$B7,$BF,$76,$F3,$62,$26,$3C,$B0,$E3,$F8,$73
	.dc.b	$0E,$3F,$69,$C2,$F7,$9F,$77,$C4,$FE,$91,$B7,$CA,$BC,$3F,$A5,$F1
	.dc.b	$97,$E8,$4F,$B5,$7B,$9F,$D8,$BF,$0E,$7E,$A9,$FE,$0B,$FC,$53,$85
	.dc.b	$EE,$B5,$3F,$E0,$FF,$9A,$CE,$1C,$B1,$FF,$39,$DF,$5F,$84,$FF,$6A
	.dc.b	$BC,$7E,$BF,$FC,$BB,$87,$F8,$5F,$F7,$FF,$27,$EC,$0F,$C3,$FF,$99
	.dc.b	$F9,$EF,$DE,$1F,$BB,$7F,$38,$FC,$1F,$75,$A9,$CF,$CD,$BE,$2F,$3F
	.dc.b	$53,$FE,$E4,$CB,$A7,$CE,$5C,$38,$65,$FD,$0B,$EA,$3E,$27,$ED,$EF
	.dc.b	$C5,$BF,$C3,$BF,$B2,$7F,$94,$7E,$BB,$FC,$4F,$F8,$7E,$A7,$3F,$68
	.dc.b	$9F,$F1,$5F,$7E,$7D,$7F,$F7,$9E,$3C,$FF,$31,$B8,$70,$BF,$F1,$8F
	.dc.b	$E1,$FF,$4F,$F6,$57,$EE,$3F,$C8,$FC,$37,$E7,$4F,$D6,$BF,$C0,$DF
	.dc.b	$98,$6E,$B5,$39,$F9,$B7,$E0,$59,$F9,$3F,$E5,$BE,$2F,$E3,$7D,$4F
	.dc.b	$E5,$FA,$76,$F9,$CC,$CF,$37,$FD,$1F,$8F,$3F,$CC,$FF,$55,$D5,$F9
	.dc.b	$5A,$FF,$09,$FC,$3B,$F6,$0F,$F6,$37,$EF,$DF,$ED,$AF,$CD,$DF,$AC
	.dc.b	$F0,$BF,$CE,$7E,$01,$FE,$53,$BD,$FE,$5C,$B7,$E8,$9B,$BE,$B5,$E4
	.dc.b	$E0,$FB,$F1,$CB,$3F,$F1,$33,$EC,$8C,$BC,$3F,$2F,$69,$E3,$FD,$FB
	.dc.b	$A6,$8F,$DC,$1F,$A7,$FF,$96,$3F,$66,$7E,$81,$FE,$21,$F8,$8D,$D6
	.dc.b	$BC,$45,$6B,$F0,$57,$ED,$3A,$C7,$CD,$CD,$F1,$73,$ED,$F4,$98,$E0
	.dc.b	$F8,$39,$F3,$CB,$97,$EE,$3E,$4F,$D4,$BF,$CE,$BD,$0F,$35,$FA,$43
	.dc.b	$F4,$AF,$F2,$17,$3F,$BB,$56,$1C,$2F,$74,$FD,$D3,$A7,$C3,$F8,$76
	.dc.b	$FF,$24,$DF,$F0,$F7,$E4,$70,$79,$E3,$96,$7D,$F7,$F8,$FF,$02,$FC
	.dc.b	$01,$F7,$5E,$3F,$98,$7D,$DB,$F3,$2D,$D6,$B7,$FE,$DB,$2F,$CC,$73
	.dc.b	$B7,$80,$FC,$1E,$5F,$E1,$D7,$07,$FB,$DF,$E9,$BF,$DE,$3F,$99,$7F
	.dc.b	$BD,$BF,$45,$FC,$27,$E2,$3F,$EE,$B8,$FF,$7C,$F9,$18,$FC,$95,$7E
	.dc.b	$FA,$FF,$79,$61,$D1,$E1,$F9,$A7,$FB,$EB,$EA,$F8,$57,$68,$FF,$7E
	.dc.b	$FE,$1D,$DB,$E8,$7F,$DB,$FE,$93,$EF,$DF,$B4,$FF,$81,$FE,$D7,$5F
	.dc.b	$86,$2B,$53,$FE,$0B,$FC,$83,$97,$C4,$F8,$F1,$FF,$74,$EF,$E3,$7A
	.dc.b	$E5,$D3,$EA,$9F,$A4,$7B,$BF,$4E,$1F,$8E,$FD,$0D,$F9,$A7,$F8,$BF
	.dc.b	$F5,$6D,$4E,$63,$FC,$65,$FA,$CE,$33,$F8,$13,$FE,$10,$6B,$BF,$AF
	.dc.b	$FC,$4F,$99,$FA,$8E,$8E,$F3,$9F,$C4,$3E,$7F,$DC,$3B,$BF,$C4,$DF
	.dc.b	$3F,$FD,$17,$C6,$7E,$F7,$A9,$CF,$85,$FE,$2C,$FC,$67,$1B,$7D,$1C
	.dc.b	$5E,$07,$D5,$7A,$72,$FF,$96,$AF,$FC,$67,$FA,$8F,$38,$C6,$6F,$CC
	.dc.b	$FF,$03,$F9,$6F,$FA,$F7,$F0,$57,$D6,$FF,$82,$B9,$3F,$C7,$47,$AE
	.dc.b	$75,$19,$FC,$1B,$8E,$FF,$85,$9F,$F6,$36,$FF,$85,$77,$DB,$FE,$07
	.dc.b	$F0,$FB,$FE,$54,$CF,$BC,$FF,$82,$FC,$1F,$E2,$DF,$98,$FF,$08,$FD
	.dc.b	$ED,$F7,$5F,$9F,$FD,$8D,$FB,$D3,$F2,$9D,$FC,$42,$F7,$E9,$2F,$7F
	.dc.b	$92,$EB,$F7,$BF,$DC,$67,$D7,$8B,$F4,$AF,$7D,$3B,$68,$FC,$9B,$C7
	.dc.b	$E0,$7F,$83,$FD,$5F,$F0,$6F,$EF,$AF,$CF,$7F,$99,$F5,$FF,$26,$CB
	.dc.b	$EC,$FD,$C2,$F7,$5F,$CC,$AB,$58,$7F,$82,$74,$FF,$3A,$C6,$2F,$CB
	.dc.b	$3D,$FE,$F1,$97,$7F,$90,$F8,$BF,$8F,$BF,$54,$FF,$42,$6B,$F9,$CF
	.dc.b	$B7,$FF,$53,$67,$E3,$AD,$7C,$72,$B5,$F9,$07,$2E,$6F,$87,$5C,$F7
	.dc.b	$F8,$6F,$0A,$E9,$FA,$AF,$D4,$E9,$EF,$3F,$C4,$7B,$9F,$AD,$67,$3E
	.dc.b	$D3,$9D,$38,$57,$D0,$BF,$F4,$D6,$5F,$41,$E7,$D7,$F8,$3F,$DE,$3F
	.dc.b	$C9,$9F,$9E,$FF,$63,$FC,$C7,$FD,$87,$B7,$F8,$E7,$F0,$9F,$F2,$0E
	.dc.b	$3F,$7F,$35,$EF,$F2,$7F,$74,$AB,$7D,$FD,$23,$FC,$85,$8F,$47,$FA
	.dc.b	$F5,$F7,$FD,$0B,$FB,$BC,$6C,$4E,$EF,$C3,$F8,$66,$63,$B7,$CC,$FF
	.dc.b	$93,$74,$7E,$95,$FC,$EB,$7D,$61,$FD,$93,$1F,$A5,$3F,$CF,$BF,$47
	.dc.b	$CE,$1F,$DC,$F4,$F8,$17,$F1,$AF,$F0,$B1,$89,$ED,$7F,$D1,$BA,$BC
	.dc.b	$5F,$07,$F8,$3A,$7E,$B8,$FE,$01,$FC,$4F,$9D,$AB,$F8,$F3,$9F,$01
	.dc.b	$E6,$67,$4F,$86,$7C,$6C,$BF,$54,$E0,$F8,$6F,$7A,$E5,$F0,$9F,$0F
	.dc.b	$B5,$FB,$7F,$FB,$AB,$F0,$BF,$AB,$FB,$07,$F8,$C3,$AF,$EB,$E5,$F4
	.dc.b	$AF,$7E,$FF,$F9,$F3,$85,$6F,$DF,$FA,$65,$F1,$70,$E1,$FB,$F5,$7E
	.dc.b	$AE,$3C,$B3,$EF,$F5,$1F,$67,$E1,$7F,$AA,$FE,$35,$FC,$BF,$6F,$0E
	.dc.b	$1F,$D8,$FF,$10,$FD,$B7,$C9,$7F,$59,$9C,$DB,$E5,$3F,$B0,$DF,$C1
	.dc.b	$8F,$EC,$2F,$EE,$23,$FC,$71,$1D,$BE,$1F,$FD,$49,$18,$3F,$B4,$E1
	.dc.b	$DB,$1F,$E6,$F8,$75,$C3,$FD,$29,$87,$7F,$94,$FF,$1A,$61,$C7,$E6
	.dc.b	$7F,$C6,$DF,$94,$F3,$9F,$ED,$98,$FA,$E3,$19,$75,$F8,$AF,$F5,$A4
	.dc.b	$6F,$2F,$F1,$D3,$CD,$FF,$1E,$7E,$1F,$E5,$F9,$27,$CB,$7F,$84,$FE
	.dc.b	$29,$FA,$CB,$F4,$6F,$F5,$F5,$EE,$B5,$2F,$D1,$B3,$8F,$AB,$81,$97
	.dc.b	$F9,$37,$0F,$85,$CF,$6B,$53,$E0,$F9,$E6,$9E,$0F,$C0,$39,$D3,$93
	.dc.b	$E9,$78,$EF,$DB,$DE,$8B,$F4,$57,$5B,$F4,$AE,$38,$70,$BD,$D6,$B8
	.dc.b	$73,$C3,$FA,$5C,$FE,$19,$E9,$FC,$6F,$CC,$F8,$63,$F5,$57,$CB,$2F
	.dc.b	$89,$D9,$FE,$B1,$AD,$7E,$E7,$FE,$AB,$FC,$FD,$FA,$CF,$F4,$9B,$EF
	.dc.b	$39,$8F,$E9,$9F,$BF,$78,$BF,$71,$EF,$BF,$E3,$EA,$E8,$78,$7C,$5F
	.dc.b	$F1,$FE,$9F,$C6,$7C,$F3,$F2,$A7,$E9,$AD,$9F,$AE,$3F,$47,$F7,$DE
	.dc.b	$EB,$53,$FE,$31,$E2,$F5,$6B,$F9,$FF,$9B,$F0,$DC,$CB,$AE,$5D,$3F
	.dc.b	$1E,$F9,$3E,$47,$E4,$BE,$AE,$F9,$1F,$9D,$FA,$BF,$A3,$E4,$3F,$5B
	.dc.b	$7E,$29,$FE,$B9,$F5,$7F,$DE,$BF,$98,$EA,$73,$F8,$8F,$F3,$AF,$94
	.dc.b	$7D,$26,$7C,$0F,$AA,$7E,$0F,$27,$E5,$5F,$21,$ED,$9F,$5B,$7E,$61
	.dc.b	$D6,$31,$F9,$07,$93,$ED,$7F,$C7,$FF,$92,$FF,$BC,$E1,$FB,$2E,$7E
	.dc.b	$C5,$39,$3D,$7F,$B3,$F7,$B9,$3A,$FA,$E6,$1F,$CC,$3B,$BE,$5E,$BF
	.dc.b	$E7,$CA,$D4,$FF,$94,$27,$3F,$A4,$FF,$79,$F8,$3E,$FC,$0F,$D6,$FE
	.dc.b	$4C,$E7,$5F,$7F,$EB,$5F,$33,$F0,$1F,$81,$33,$E3,$7F,$93,$D7,$C4
	.dc.b	$FC,$0F,$8C,$63,$1F,$D5,$67,$E4,$FB,$77,$E9,$4F,$DE,$1F,$8C,$FF
	.dc.b	$7A,$FD,$C7,$EC,$D6,$A7,$3F,$02,$FD,$17,$3D,$BF,$53,$CC,$73,$BE
	.dc.b	$F7,$EA,$63,$C7,$3F,$F0,$2F,$ED,$FF,$BC,$F9,$DF,$F0,$97,$D9,$FF
	.dc.b	$84,$FE,$93,$FC,$29,$F6,$5F,$E1,$5E,$2F,$D6,$57,$F8,$5F,$E8,$7F
	.dc.b	$66,$F0,$DB,$F0,$6F,$23,$FC,$80,$7F,$50,$FD,$A7,$B3,$FE,$6A,$1F
	.dc.b	$61,$FE,$21,$FB,$2F,$F0,$BE,$1D,$A7,$53,$98,$E3,$F7,$9E,$09,$FF
	.dc.b	$39,$E1,$E0,$76,$7F,$8F,$7E,$A7,$BF,$D5,$FF,$8B,$7A,$FF,$C5,$DF
	.dc.b	$59,$EB,$F7,$BF,$DF,$97,$FE,$24,$3F,$C7,$07,$F9,$87,$0F,$3F,$E8
	.dc.b	$BD,$F0,$93,$AF,$AD,$6F,$EC,$E2,$EB,$3F,$E2,$ED,$BD,$51,$88,$61
	.dc.b	$F0,$BE,$B3,$D5,$32,$7F,$C8,$E3,$F1,$A5,$BE,$87,$FC,$61,$A7,$F8
	.dc.b	$C2,$8F,$F8,$C6,$74,$FF,$8A,$8D,$1A,$F5,$79,$D7,$2F,$90,$E3,$3F
	.dc.b	$E5,$E8,$EB,$F5,$D8,$8C,$DF,$47,$FC,$5E,$74,$3F,$7E,$E7,$9F,$FC
	.dc.b	$85,$8F,$F8,$AE,$7F,$11,$7E,$8F,$AB,$FF,$2F,$7D,$5F,$A3,$83,$D3
	.dc.b	$8F,$1B,$FE,$28,$78,$4F,$39,$EB,$AC,$38,$3C,$6F,$47,$E9,$CE,$71
	.dc.b	$F8,$BF,$4F,$F8,$19,$FF,$29,$FE,$03,$95,$19,$F7,$5F,$9E,$E7,$F2
	.dc.b	$AE,$F9,$F9,$6A,$EB,$F4,$9E,$1F,$51,$CF,$C9,$FF,$3D,$7D,$47,$D4
	.dc.b	$AF,$83,$9F,$D3,$C3,$8B,$F5,$6A,$BF,$31,$E2,$FF,$BC,$76,$F7,$79
	.dc.b	$3F,$E1,$9A,$FA,$4F,$F8,$A3,$E8,$B3,$F2,$3F,$E2,$2B,$FB,$47,$F7
	.dc.b	$FF,$A4,$F4,$FE,$09,$F7,$BC,$B7,$EF,$CF,$CD,$9F,$AE,$7C,$0F,$8F
	.dc.b	$FE,$AB,$D1,$F9,$0C,$38,$5E,$EB,$5F,$24,$F3,$B5,$A7,$E4,$B3,$EB
	.dc.b	$FA,$3F,$82,$31,$87,$77,$EE,$70,$B2,$EB,$AF,$8B,$A7,$FB,$46,$6F
	.dc.b	$DE,$3A,$E1,$FE,$43,$F3,$FF,$C1,$7F,$55,$FE,$47,$C7,$FC,$5B,$18
	.dc.b	$E4,$DF,$2F,$F8,$DE,$FB,$D7,$F8,$B5,$FF,$5D,$FC,$BF,$0F,$9D,$63
	.dc.b	$FE,$51,$8C,$3F,$E4,$27,$FC,$6D,$3C,$0F,$98,$7A,$BE,$53,$FC,$95
	.dc.b	$CC,$FD,$B1,$D0,$FD,$35,$F9,$4F,$F7,$BB,$DD,$6A,$73,$DD,$FB,$E9
	.dc.b	$E7,$8F,$53,$FC,$71,$F2,$FF,$92,$8E,$38,$F5,$C7,$FC,$15,$CC,$9F
	.dc.b	$1F,$D4,$BF,$49,$FF,$1C,$FE,$1E,$FC,$0B,$B6,$7F,$E4,$3D,$D7,$BF
	.dc.b	$AC,$5B,$DB,$DD,$FF,$8B,$ED,$F9,$EB,$F6,$9F,$93,$F1,$CF,$DD,$F7
	.dc.b	$D5,$FE,$75,$BD,$D6,$BF,$45,$CF,$91,$FE,$25,$C7,$97,$57,$F6,$DF
	.dc.b	$D1,$33,$CE,$F7,$9F,$A1,$BF,$12,$7E,$54,$FA,$CD,$D6,$A7,$3E,$0F
	.dc.b	$6E,$DF,$91,$C9,$F2,$AB,$58,$7F,$B5,$7A,$FE,$FF,$DC,$F8,$3E,$97
	.dc.b	$FD,$05,$F4,$D7,$C7,$FC,$67,$81,$F3,$72,$FA,$75,$FE,$2B,$8C,$47
	.dc.b	$F9,$4F,$83,$FD,$49,$FB,$E3,$F1,$87,$EC,$EF,$CC,$BD,$BE,$CA,$B5
	.dc.b	$3F,$33,$1E,$7D,$BF,$E8,$FA,$D6,$1C,$BF,$24,$E3,$87,$0F,$C6,$FE
	.dc.b	$3F,$EB,$8F,$C7,$9F,$A9,$3C,$07,$C5,$EA,$73,$1F,$07,$87,$F1,$B7
	.dc.b	$C6,$F8,$7E,$1D,$B7,$97,$AF,$F5,$3F,$17,$7D,$B3,$F3,$3F,$F3,$17
	.dc.b	$E0,$79,$8C,$7D,$CB,$F7,$0C,$57,$F9,$9F,$F3,$FB,$3F,$E6,$9E,$7F
	.dc.b	$7F,$94,$EB,$79,$0F,$95,$7E,$2F,$FE,$68,$FC,$2B,$F0,$B3,$98,$FF
	.dc.b	$35,$FC,$86,$2B,$7F,$02,$F9,$9F,$CC,$FC,$90,$FF,$0A,$1F,$E1,$4D
	.dc.b	$3E,$CB,$93,$F8,$E1,$BE,$BF,$EF,$B1,$A8,$E5,$ED,$6F,$E7,$3F,$C4
	.dc.b	$9E,$57,$F8,$2A,$72,$77,$FD,$43,$FA,$EC,$66,$3F,$C4,$02,$3A,$C6
	.dc.b	$7E,$6F,$C5,$F3,$9D,$3B,$FF,$BE,$DF,$96,$7C,$DF,$8B,$70,$9C,$CF
	.dc.b	$F8,$DC,$FF,$08,$E7,$E7,$7F,$C6,$27,$F8,$6D,$C7,$CE,$7B,$F8,$3D
	.dc.b	$BF,$9E,$7F,$83,$2F,$FB,$87,$C9,$7F,$84,$23,$1F,$35,$F1,$71,$FE
	.dc.b	$EF,$18,$7E,$65,$FF,$91,$63,$11,$F3,$EB,$95,$FF,$84,$C7,$CD,$7C
	.dc.b	$FE,$5E,$71,$98,$FA,$5D,$1C,$E3,$17,$FE,$08,$BE,$71,$FE,$12,$3F
	.dc.b	$BC,$C7,$F7,$DD,$18,$FA,$FE,$5F,$EE,$7E,$2B,$CE,$F8,$47,$CE,$C3
	.dc.b	$9F,$CB,$7F,$88,$BE,$5F,$F0,$5F,$43,$B8,$C6,$7F,$DF,$2B,$C3,$E9
	.dc.b	$7F,$73,$CF,$FA,$3F,$C8,$E3,$C1,$FE,$E5,$A7,$F8,$A2,$7F,$A8,$3F
	.dc.b	$E2,$AF,$DF,$BF,$87,$EE,$78,$B6,$3F,$9F,$9F,$D9,$A3,$8B,$FE,$26
	.dc.b	$FD,$83,$89,$C0,$FE,$AA,$FF,$8D,$9D,$BF,$E4,$8F,$CB,$3F,$C7,$B9
	.dc.b	$31,$FD,$77,$B2,$FD,$A7,$F9,$DF,$EF,$0F,$D7,$78,$63,$FC,$7A,$73
	.dc.b	$F9,$B6,$23,$F8,$F1,$FD,$EB,$2E,$FA,$9E,$F9,$76,$BE,$BF,$B9,$71
	.dc.b	$D3,$C5,$E5,$35,$7E,$EC,$FC,$CD,$FB,$8B,$F6,$4E,$19,$FA,$2B,$52
	.dc.b	$FD,$77,$38,$F3,$9C,$FE,$B3,$8C,$BF,$B0,$7E,$A5,$FD,$CF,$E7,$71
	.dc.b	$E6,$A3,$FB,$B7,$EF,$18,$7F,$BC,$7E,$51,$C7,$F0,$D7,$E5,$CF,$9D
	.dc.b	$FF,$7D,$F3,$FB,$FA,$3F,$A1,$D6,$A5,$F8,$2F,$1B,$E1,$F2,$DE,$3F
	.dc.b	$8E,$FF,$0C,$FC,$85,$F9,$4F,$F0,$8F,$C8,$DF,$97,$78,$D7,$F0,$FF
	.dc.b	$6F,$E9,$DE,$1F,$47,$FB,$17,$83,$8D,$FF,$57,$F3,$1F,$67,$8F,$C9
	.dc.b	$C7,$FD,$4F,$0E,$12,$FD,$FF,$86,$3F,$DA,$FF,$65,$EB,$FB,$AF,$F9
	.dc.b	$83,$81,$DB,$8F,$AC,$47,$D3,$D3,$AF,$8D,$F6,$B4,$61,$C7,$2F,$3F
	.dc.b	$D6,$3F,$71,$FB,$9E,$06,$1F,$BD,$DB,$7C,$87,$E5,$DE,$EF,$1D,$F6
	.dc.b	$1F,$D6,$BF,$19,$C3,$FD,$73,$BE,$FC,$9B,$F3,$4F,$CA,$7A,$FF,$C5
	.dc.b	$7E,$D3,$F8,$B7,$27,$C2,$8F,$E3,$1E,$CF,$FA,$7B,$D9,$DC,$7F,$3F
	.dc.b	$F2,$F7,$E9,$FF,$35,$8F,$F1,$07,$7F,$FA,$57,$0E,$51,$FC,$C4,$FE
	.dc.b	$8C,$E1,$FE,$99,$F7,$BD,$63,$11,$FC,$C0,$EE,$7F,$4B,$9E,$9F,$05
	.dc.b	$FC,$BD,$FF,$5D,$47,$FA,$FA,$31,$F7,$7C,$DD,$74,$FB,$1E,$1F,$79
	.dc.b	$9B,$FE,$58,$7F,$22,$BF,$F7,$B4,$F5,$FB,$AD,$4F,$F2,$71,$F0,$1C
	.dc.b	$DE,$39,$6B,$BF,$FD,$1D,$7F,$C5,$0F,$E1,$B3,$FE,$B6,$7F,$7F,$9F
	.dc.b	$F1,$A3,$F8,$C9,$A8,$FF,$6B,$BF,$C1,$3F,$56,$FE,$3E,$7F,$1E,$35
	.dc.b	$67,$F8,$4F,$EA,$FF,$C9,$0F,$E4,$66,$8F,$E1,$BF,$AB,$7F,$28,$3F
	.dc.b	$93,$9C,$8F,$E2,$3F,$9E,$FF,$2C,$3F,$95,$9D,$8F,$E2,$BF,$BC,$FF
	.dc.b	$30,$3F,$97,$9E,$23,$F8,$CF,$ED,$BF,$CD,$0F,$E6,$67,$98,$FE,$3B
	.dc.b	$FA,$D7,$F3,$83,$F9,$B9,$EA,$3F,$90,$FE,$8B,$FC,$F0,$FE,$76,$7B
	.dc.b	$8C,$E9,$FE,$F6,$F9,$47,$A9,$3D,$27,$EE,$CF,$73,$F8,$E7,$E3,$1F
	.dc.b	$F0,$03,$1F,$F7,$AE,$1C,$30,$61,$C2,$FF,$91,$7D,$E6,$F2,$ED,$4F
	.dc.b	$67,$FD,$B3,$8F,$28,$F0,$CF,$FA,$CA,$31,$1D,$33,$F1,$7C,$13,$F6
	.dc.b	$4E,$97,$FE,$01,$EB,$FF,$0A,$98,$30,$3E,$53,$FC,$4F,$8F,$EA,$E3
	.dc.b	$FC,$61,$EE,$FF,$8C,$7A,$13,$99,$7E,$78,$C3,$A4,$77,$8E,$71,$FE
	.dc.b	$1E,$8F,$F0,$97,$C9,$73,$AE,$5F,$27,$C7,$0E,$B7,$BB,$FF,$07,$78
	.dc.b	$DD,$1F,$F0,$D7,$CD,$F1,$9D,$47,$F8,$69,$D3,$FE,$25,$9A,$FF,$0C
	.dc.b	$FD,$2B,$F7,$13,$FB,$F7,$D0,$BC,$9F,$55,$62,$33,$82,$B8,$C7,$F8
	.dc.b	$F2,$30,$FE,$E8,$E2,$1C,$BF,$B9,$3A,$CE,$67,$9B,$F3,$5E,$BF,$A3
	.dc.b	$7C,$36,$C6,$E7,$B0,$9E,$0E,$3F,$09,$E0,$D9,$E0,$F8,$44,$70,$71
	.dc.b	$F1,$9C,$1B,$3C,$3C,$5F,$EF,$FF,$14,$E2,$FE,$97,$F6,$4F,$CE,$F5
	.dc.b	$8F,$F8,$52,$73,$FB,$6E,$3E,$3F,$E5,$67,$D7,$2F,$F0,$D7,$DC,$E2
	.dc.b	$35,$9F,$5F,$82,$68,$FD,$67,$39,$FF,$5A,$FD,$51,$FA,$DE,$2D,$0F
	.dc.b	$8E,$C2,$FD,$73,$5F,$9D,$7F,$85,$79,$FE,$9E,$FE,$CC,$BF,$B2,$CF
	.dc.b	$DD,$9C,$F1,$7A,$F2,$4F,$FB,$0E,$73,$C3,$FE,$21,$BE,$3F,$A9,$76
	.dc.b	$FC,$93,$BF,$EA,$7E,$5D,$BF,$75,$87,$0E,$2F,$F6,$AB,$70,$FF,$6A
	.dc.b	$F5,$7F,$8C,$32,$E9,$A7,$AF,$CA,$F7,$71,$3F,$6A,$BF,$8F,$E3,$F7
	.dc.b	$CC,$FD,$F6,$BE,$B7,$7E,$87,$0F,$A6,$E3,$F8,$B7,$C7,$E9,$7E,$69
	.dc.b	$7E,$3F,$1F,$B3,$E7,$F6,$6F,$99,$FA,$FB,$F5,$7B,$F1,$98,$FD,$CB
	.dc.b	$3F,$AA,$FE,$77,$E5,$3F,$CC,$67,$F9,$88,$FE,$8A,$7F,$93,$3F,$08
	.dc.b	$FF,$33,$9F,$E6,$63,$FA,$49,$FE,$50,$F6,$3F,$CD,$67,$F9,$A8,$FE
	.dc.b	$9A,$7F,$95,$3E,$A3,$FC,$DE,$7F,$9B,$8F,$EA,$27,$F9,$63,$E7,$7E
	.dc.b	$47,$81,$F3,$1F,$F3,$A9,$FD,$58,$D7,$97,$FF,$00,$3C,$0F,$F1,$1F
	.dc.b	$85,$FE,$36,$F0,$BF,$C4,$F3,$9F,$03,$E8,$43,$E7,$3F,$82,$FE,$08
	.dc.b	$D5,$FA,$FB,$ED,$BF,$57,$EB,$FF,$83,$2F,$75,$A9,$FF,$3C,$65,$FE
	.dc.b	$13,$FD,$0F,$FC,$DB,$F8,$93,$F1,$A7,$A8,$FA,$07,$C3,$3F,$1C,$FE
	.dc.b	$15,$87,$0B,$DD,$6B,$F1,$CC,$E5,$FC,$C7,$D5,$FF,$43,$67,$FA,$DE
	.dc.b	$26,$3F,$DE,$0F,$0F,$E2,$8F,$C5,$7F,$51,$F8,$BB,$F0,$EE,$79,$32
	.dc.b	$E7,$7B,$E8,$5B,$F1,$7D,$D7,$EA,$FA,$7F,$65,$F1,$B8,$F6,$FF,$55
	.dc.b	$87,$CF,$D7,$FD,$1E,$7F,$BB,$FE,$FF,$FC,$81,$D5,$E0,$3F,$DF,$5F
	.dc.b	$43,$FE,$FB,$FC,$03,$FA,$85,$7F,$BC,$63,$FD,$FD,$F8,$1F,$FB,$F2
	.dc.b	$31,$41,$F8,$FF,$F4,$B8,$CC,$7F,$95,$0F,$E6,$9F,$71,$FD,$03,$77
	.dc.b	$EF,$4F,$DC,$DF,$AE,$BD,$17,$AF,$F5,$BE,$37,$8F,$47,$F5,$FC,$78
	.dc.b	$E1,$C2,$F7,$E1,$38,$37,$7F,$1F,$37,$EA,$DF,$57,$5F,$14,$78,$67
	.dc.b	$F9,$3F,$6F,$FA,$75,$ED,$1E,$5D,$BC,$9B,$7E,$23,$EC,$7F,$A7,$7E
	.dc.b	$35,$FD,$6B,$E3,$5F,$8E,$4F,$8B,$7E,$15,$FE,$6F,$F9,$A7,$E8,$DC
	.dc.b	$35,$FE,$F1,$87,$0C,$3F,$95,$DE,$FF,$43,$FE,$B5,$E6,$7F,$66,$FC
	.dc.b	$7D,$FA,$53,$F2,$DF,$F5,$6F,$E8,$4F,$AC,$FF,$5A,$7E,$3D,$C3,$5F
	.dc.b	$F0,$26,$1C,$2F,$75,$FE,$B9,$FC,$7F,$54,$AD,$7E,$0A,$C7,$A7,$4F
	.dc.b	$B3,$F2,$B7,$E4,$3F,$DE,$3F,$3F,$7E,$DC,$F2,$1F,$A3,$BF,$05,$7E
	.dc.b	$A8,$FC,$8D,$F9,$BB,$3E,$7C,$1F,$E1,$9C,$F9,$E5,$CB,$1E,$38,$7E
	.dc.b	$82,$F7,$6B,$DE,$5D,$F2,$ED,$EA,$75,$FC,$D7,$FC,$1B,$FB,$4B,$E5
	.dc.b	$1F,$9D,$3E,$CF,$FD,$F9,$EA,$BF,$5E,$F2,$6C,$C7,$8E,$1C,$2F,$75
	.dc.b	$FE,$0D,$AD,$6D,$FE,$31,$CB,$B6,$5D,$7D,$86,$7D,$7F,$67,$F8,$0F
	.dc.b	$F2,$8F,$8A,$FF,$62,$FE,$A2,$F0,$33,$97,$F2,$38,$FA,$15,$BF,$77
	.dc.b	$FB,$AF,$8F,$AA,$FE,$0D,$97,$F6,$7F,$A9,$FE,$F1,$B7,$82,$7F,$96
	.dc.b	$7B,$BF,$E5,$B8,$C4,$70,$FA,$3F,$EB,$F3,$FE,$E8,$9C,$FB,$7F,$E5
	.dc.b	$4F,$C8,$3F,$79,$18,$8F,$F2,$D7,$D8,$7F,$93,$E3,$11,$FE,$4C,$3C
	.dc.b	$87,$F9,$72,$1F,$5A,$E2,$7E,$3B,$FE,$40,$FC,$95,$FB,$EB,$ED,$7F
	.dc.b	$0F,$F2,$9F,$8A,$FC,$75,$F9,$3F,$1C,$38,$5E,$FC,$15,$EF,$F2,$8E
	.dc.b	$DF,$A5,$FF,$93,$3F,$2A,$FF,$5B,$FC,$2B,$F7,$7F,$E9,$5C,$4F,$C9
	.dc.b	$BF,$3D,$C2,$C3,$87,$E1,$5F,$C4,$6F,$74,$FD,$33,$AF,$57,$F9,$BF
	.dc.b	$3E,$B5,$AF,$D6,$1F,$BF,$3E,$C3,$FC,$BB,$FB,$A3,$F0,$BF,$F1,$27
	.dc.b	$E9,$4F,$DC,$DF,$8A,$BF,$29,$7E,$BD,$CB,$1E,$38,$70,$BD,$D3,$F1
	.dc.b	$8D,$6F,$FE,$81,$CF,$B7,$DB,$34,$ED,$A7,$5F,$81,$F2,$F8,$FF,$E7
	.dc.b	$0F,$D4,$DF,$8C,$7E,$97,$E2,$5F,$8A,$3E,$45,$F6,$9F,$D9,$79,$58
	.dc.b	$F1,$C3,$85,$EE,$BF,$D6,$55,$AD,$FF,$D2,$99,$76,$F6,$99,$F6,$CF
	.dc.b	$AF,$E4,$0F,$CA,$DF,$A0,$F9,$7F,$4B,$7E,$9B,$F1,$A9,$E9,$3F,$34
	.dc.b	$FF,$2B,$E1,$C3,$5E,$F8,$70,$E4,$61,$C2,$DF,$93,$F7,$CF,$B6,$7D
	.dc.b	$73,$E9,$F3,$0F,$DB,$DF,$BE,$7F,$A5,$BF,$44,$FE,$AB,$F1,$2F,$DB
	.dc.b	$9F,$BF,$FF,$82,$7E,$1D,$FB,$B7,$1C,$38,$7D,$0F,$F1,$7B,$DF,$DA
	.dc.b	$7F,$0F,$CF,$5C,$BF,$EB,$9D,$3A,$D6,$B2,$7F,$00,$74,$FC,$1F,$FD
	.dc.b	$4F,$F9,$57,$FA,$8F,$C3,$7E,$59,$E6,$E9,$7D,$2B,$F7,$1E,$3E,$4F
	.dc.b	$FA,$BB,$F6,$E7,$E5,$7F,$33,$C7,$7E,$C2,$FD,$D5,$F9,$CF,$F9,$EF
	.dc.b	$81,$8F,$1C,$38,$5E,$EB,$FC,$EB,$5A,$E4,$FF,$69,$65,$DB,$BB,$F1
	.dc.b	$77,$BF,$C5,$5F,$C0,$1F,$E2,$FB,$FF,$6C,$6E,$FC,$C7,$FC,$69,$F9
	.dc.b	$A3,$F0,$BF,$ED,$3F,$99,$BF,$0D,$E3,$DD,$FE,$C8,$C3,$87,$E0,$CC
	.dc.b	$38,$7D,$83,$0F,$F2,$47,$EA,$7F,$E4,$BD,$3F,$DC,$DF,$56,$FD,$73
	.dc.b	$E0,$7E,$72,$FC,$6D,$FA,$07,$FB,$AF,$F0,$A6,$3C,$71,$FE,$C9,$87
	.dc.b	$0B,$DC,$7F,$74,$9D,$63,$D7,$B1,$97,$5C,$BF,$DE,$DE,$E3,$F5,$47
	.dc.b	$ED,$7F,$EE,$2F,$91,$FC,$AF,$E7,$DF,$E5,$DF,$99,$FE,$CD,$F9,$83
	.dc.b	$E6,$B8,$5E,$FD,$85,$EF,$E7,$97,$FC,$2F,$2E,$BC,$21,$9F,$6F,$C4
	.dc.b	$7F,$98,$5F,$0F,$CF,$9F,$AD,$7F,$9B,$EF,$FB,$A7,$EE,$3F,$C0,$FF
	.dc.b	$06,$7D,$0F,$F1,$6F,$C4,$BF,$94,$5E,$EF,$FB,$A5,$EE,$B5,$C1,$D3
	.dc.b	$F1,$5F,$F0,$44,$E7,$E5,$3F,$CC,$5F,$BB,$7E,$FB,$F3,$EF,$36,$3F
	.dc.b	$E7,$EF,$03,$F4,$3F,$9B,$BF,$6E,$FF,$35,$F9,$FF,$9D,$F3,$D8,$70
	.dc.b	$FC,$6F,$F0,$1E,$8B,$0E,$1F,$89,$30,$FF,$44,$E1,$C3,$3F,$89,$87
	.dc.b	$0B,$DD,$3F,$51,$FA,$1F,$BA,$79,$3F,$1A,$7B,$CF,$C6,$FF,$C3,$BF
	.dc.b	$BF,$FD,$2F,$A8,$7E,$51,$C3,$5F,$A1,$87,$0B,$DF,$87,$FE,$67,$FC
	.dc.b	$A7,$55,$FE,$81,$CB,$A7,$D6,$B3,$E9,$F3,$0F,$D1,$3E,$AD,$FE,$2B
	.dc.b	$F2,$A7,$E7,$9F,$E3,$0F,$CD,$7F,$DF,$1B,$FE,$86,$F7,$7F,$CE,$2F
	.dc.b	$7F,$82,$AF,$7F,$A3,$75,$F1,$3E,$8F,$A0,$BD,$D7,$4F,$80,$FD,$77
	.dc.b	$C8,$7D,$6C,$1F,$87,$3F,$37,$7E,$89,$FC,$EB,$F5,$27,$91,$F9,$9F
	.dc.b	$80,$FE,$59,$C1,$F9,$3C,$78,$E1,$C3,$E7,$18,$70,$FA,$AE,$F7,$BC
	.dc.b	$FB,$67,$D7,$37,$A2,$FE,$01,$7E,$BD,$FE,$0D,$EF,$FF,$00,$7E,$B6
	.dc.b	$FD,$91,$FB,$6F,$F9,$63,$E9,$78,$5E,$FE,$D1,$7B,$AF,$F2,$EF,$EA
	.dc.b	$FA,$DF,$F5,$5E,$BF,$E0,$6F,$79,$74,$FC,$57,$FD,$99,$F9,$A7,$F9
	.dc.b	$12,$DF,$BD,$BF,$07,$7E,$C9,$FD,$FF,$F6,$17,$EB,$7F,$EC,$6F,$D5
	.dc.b	$FF,$E0,$07,$E2,$DF,$A2,$C3,$85,$EE,$BF,$D9,$7E,$4B,$F3,$8F,$C0
	.dc.b	$E7,$FB,$DF,$DE,$9F,$8F,$FF,$84,$B1,$7D,$BB,$F6,$87,$E7,$FF,$E4
	.dc.b	$FC,$78,$E2,$C7,$8F,$DD,$7A,$30,$E1,$7F,$E9,$CF,$BD,$F5,$5F,$F0
	.dc.b	$BF,$B1,$76,$BF,$60,$7E,$A8,$E7,$7E,$A4,$FD,$53,$FC,$E5,$97,$2C
	.dc.b	$BF,$B6,$E3,$C7,$0E,$17,$BF,$CA,$5F,$B1,$3F,$4F,$F8,$9D,$5F,$E9
	.dc.b	$FF,$DD,$5F,$B7,$7E,$D3,$F2,$3F,$F3,$55,$EF,$4F,$E4,$1D,$0C,$38
	.dc.b	$5E,$FF,$2B,$EB,$97,$4F,$CB,$19,$F4,$F8,$26,$9F,$EB,$CD,$3F,$CD
	.dc.b	$3F,$93,$FF,$BB,$FF,$1E,$7E,$79,$E6,$FC,$8F,$FC,$ED,$1F,$CC,$AB
	.dc.b	$7F,$90,$FF,$BC,$BE,$27,$5C,$5F,$CC,$FE,$2F,$FD,$D0,$42,$D7,$F9
	.dc.b	$B1,$C6,$1F,$09,$FE,$4B,$F9,$D7,$F2,$6A,$FF,$6B,$FE,$43,$FC,$A3
	.dc.b	$F5,$07,$E7,$7F,$D8,$FF,$1D,$FC,$DF,$3B,$F8,$01,$CE,$FC,$E7,$FC
	.dc.b	$EB,$97,$F4,$5C,$78,$FC,$27,$F9,$8F,$D8,$F9,$18,$7E,$D2,$F7,$6B
	.dc.b	$EF,$FB,$97,$8E,$3F,$A7,$4E,$A7,$9C,$79,$8E,$DA,$79,$70,$FA,$9E
	.dc.b	$67,$0D,$BD,$53,$9D,$5F,$B3,$7F,$59,$FC,$87,$FD,$33,$F8,$57,$F8
	.dc.b	$D2,$1F,$8C,$FE,$83,$F0,$0F,$F8,$01,$FA,$8A,$F7,$E6,$F2,$AD,$7D
	.dc.b	$53,$B7,$5D,$5F,$D4,$33,$E9,$C3,$EB,$C7,$8F,$B9,$ED,$8D,$9F,$DA
	.dc.b	$E7,$C5,$8F,$87,$AB,$CD,$87,$3C,$FC,$1E,$AB,$67,$EF,$DF,$DC,$FE
	.dc.b	$B7,$E5,$67,$FD,$D7,$F5,$7F,$AD,$E6,$BF,$51,$7E,$C9,$FE,$EB,$F8
	.dc.b	$4F,$F4,$06,$1C,$30,$61,$C2,$F7,$F9,$B2,$FF,$DF,$97,$BA,$FC,$17
	.dc.b	$E1,$7E,$B8,$C7,$83,$FB,$6C,$7E,$D4,$63,$3F,$16,$7F,$EF,$BE,$1F
	.dc.b	$47,$7F,$F7,$6C,$BC,$13,$FE,$FF,$DB,$D9,$E2,$FF,$8D,$7F,$7C,$FF
	.dc.b	$1C,$7E,$A7,$FE,$C0,$F9,$87,$EC,$8F,$86,$6C,$C3,$87,$E1,$0C,$38
	.dc.b	$5F,$F8,$06,$B5,$C1,$FE,$48,$CB,$FC,$88,$FA,$30,$F0,$7E,$13,$ED
	.dc.b	$79,$E1,$CA,$7B,$C6,$3F,$2F,$FE,$C5,$F8,$9F,$F8,$97,$F0,$FF,$D6
	.dc.b	$7A,$DF,$EC,$7F,$C4,$BF,$7B,$F9,$5B,$F7,$47,$ED,$5F,$E6,$AB,$DF
	.dc.b	$C9,$7F,$AE,$BD,$9F,$F6,$65,$6B,$E7,$BF,$CD,$FB,$7F,$9E,$29,$EF
	.dc.b	$7E,$2F,$0E,$39,$DA,$D3,$E0,$C3,$BC,$72,$3F,$C7,$5A,$79,$72,$EF
	.dc.b	$C9,$F8,$BF,$C1,$3F,$81,$7E,$DA,$FD,$DD,$EF,$FF,$75,$F1,$3F,$B6
	.dc.b	$F9,$5F,$E4,$DC,$FF,$80,$5E,$EB,$5F,$43,$F4,$FA,$FF,$DE,$95,$AF
	.dc.b	$DA,$FF,$11,$9F,$6C,$EC,$F2,$7B,$FD,$1F,$E4,$7C,$3F,$24,$E7,$2F
	.dc.b	$17,$E7,$FF,$A6,$8F,$F2,$A7,$97,$FC,$FB,$F2,$2F,$C1,$7E,$2E,$FD
	.dc.b	$C5,$FA,$4F,$FA,$53,$D6,$71,$FF,$91,$B0,$E1,$E1,$7F,$A6,$AF,$7F
	.dc.b	$B8,$7E,$DB,$3F,$D5,$7C,$33,$DB,$FD,$D6,$5C,$A7,$C3,$2D,$73,$9F
	.dc.b	$58,$C6,$5D,$78,$3D,$17,$D2,$FC,$F9,$FE,$DE,$3F,$37,$87,$F9,$6B
	.dc.b	$F1,$5F,$BF,$F8,$AF,$FA,$C3,$E3,$9E,$43,$F5,$EF,$EC,$9D,$5E,$3C
	.dc.b	$58,$70,$F0,$FF,$AE,$DE,$F1,$EB,$5F,$E9,$5F,$11,$FB,$B7,$EE,$F4
	.dc.b	$F0,$47,$FA,$76,$7F,$77,$18,$AF,$15,$BF,$3D,$FE,$25,$39,$C7,$BF
	.dc.b	$D8,$FF,$2F,$E2,$FF,$39,$F9,$8F,$D6,$FB,$FE,$F4,$FC,$71,$FA,$27
	.dc.b	$FB,$BB,$B3,$F9,$86,$1C,$2F,$75,$AF,$CB,$9F,$B0,$6B,$87,$FD,$29
	.dc.b	$9F,$5F,$D6,$BE,$AE,$3C,$79,$2D,$6B,$2E,$7A,$7F,$7B,$8E,$43,$0F
	.dc.b	$35,$35,$F3,$53,$EF,$FD,$F7,$FE,$A3,$FC,$07,$FC,$69,$93,$F0,$87
	.dc.b	$84,$FD,$5B,$FC,$07,$FB,$EF,$D0,$C1,$87,$0B,$DD,$7F,$82,$3F,$20
	.dc.b	$D4,$FE,$2E,$BE,$A1,$DF,$EC,$7F,$8C,$BF,$DE,$7B,$BF,$01,$3C,$F2
	.dc.b	$F0,$7E,$43,$E6,$AE,$F7,$FC,$97,$2E,$FF,$4F,$FD,$8F,$E1,$7F,$D4
	.dc.b	$FF,$9A,$3E,$4F,$FD,$30,$3F,$6D,$FE,$5F,$FA,$EB,$E2,$BF,$B2,$5F
	.dc.b	$0A,$DD,$6A,$F3,$9F,$FB,$1D,$E9,$F8,$6F,$EA,$2F,$87,$E6,$76,$B5
	.dc.b	$87,$2F,$71,$FB,$27,$96,$73,$87,$97,$2F,$E8,$3F,$80,$FF,$04,$97
	.dc.b	$E9,$8F,$09,$AB,$F6,$97,$E1,$6F,$DA,$3F,$98,$FE,$E9,$FC,$67,$1E
	.dc.b	$38,$70,$BD,$D7,$F3,$6F,$DB,$F5,$97,$5C,$BF,$8B,$E1,$C3,$F5,$8B
	.dc.b	$39,$C3,$8F,$C6,$79,$63,$1A,$78,$78,$7F,$B0,$E1,$FE,$EB,$FD,$69
	.dc.b	$FA,$4B,$F1,$AF,$EE,$7F,$8B,$7F,$9C,$BF,$59,$FF,$61,$F8,$5F,$DD
	.dc.b	$EF,$75,$AF,$A0,$57,$E5,$6B,$58,$3F,$0F,$FE,$ED,$D9,$FE,$E4,$C7
	.dc.b	$8E,$7F,$DD,$A3,$15,$E1,$F7,$DA,$7F,$A0,$BE,$9F,$CF,$6C,$3C,$35
	.dc.b	$AD,$3F,$D0,$93,$FD,$CB,$F4,$FF,$F7,$8F,$C6,$FF,$27,$F9,$E7,$E5
	.dc.b	$FF,$E0,$EF,$94,$7E,$FF,$FD,$37,$85,$87,$0F,$89,$FF,$0B,$DE,$EB
	.dc.b	$FB,$C4,$E7,$CD,$FF,$08,$3B,$C6,$E6,$3F,$D8,$38,$72,$D7,$FB,$87
	.dc.b	$83,$FE,$27,$BE,$19,$78,$F4,$C6,$1F,$EC,$BF,$CC,$9F,$B9,$3F,$2E
	.dc.b	$FF,$48,$FE,$28,$FD,$97,$F8,$D7,$0F,$A3,$0E,$17,$BF,$A6,$FF,$6B
	.dc.b	$D3,$F4,$FF,$89,$5D,$7F,$57,$F3,$AF,$C6,$5A,$78,$23,$FC,$53,$BF
	.dc.b	$FB,$96,$3C,$7A,$F8,$BF,$07,$F3,$D7,$19,$CE,$7F,$EE,$6F,$CC,$BF
	.dc.b	$C5,$33,$F8,$DF,$8A,$7E,$93,$FE,$EC,$FC,$61,$FB,$77,$F8,$9F,$BF
	.dc.b	$FC,$37,$7B,$AD,$4F,$F8,$AF,$F6,$7C,$CF,$E0,$70,$E6,$FF,$89,$CE
	.dc.b	$DD,$1F,$E2,$77,$92,$C5,$F8,$FF,$D5,$9E,$7E,$0F,$F4,$9C,$BB,$FE
	.dc.b	$D9,$E8,$9F,$F7,$CF,$D0,$3F,$3D,$FF,$1F,$FB,$1F,$EF,$5F,$BF,$7E
	.dc.b	$F9,$E4,$F1,$BF,$C2,$D5,$AF,$C7,$FF,$DA,$DF,$AB,$E7,$17,$C4,$79
	.dc.b	$73,$EB,$77,$0E,$3E,$8F,$F1,$0C,$7F,$37,$E1,$7F,$2E,$9C,$EA,$9F
	.dc.b	$F3,$8E,$1F,$E5,$0F,$3B,$D9,$E4,$BF,$36,$FE,$85,$FA,$EB,$F7,$F7
	.dc.b	$B2,$FD,$9F,$FD,$BF,$AF,$F9,$C7,$0E,$1F,$0E,$C3,$85,$EF,$E5,$BB
	.dc.b	$69,$F9,$DB,$DB,$FB,$93,$19,$BE,$0F,$F7,$F1,$8C,$7A,$69,$E2,$F4
	.dc.b	$3F,$77,$87,$83,$3E,$FF,$4F,$FC,$BF,$0F,$E5,$5E,$0B,$F2,$1F,$EF
	.dc.b	$DF,$B1,$FF,$9A,$FE,$6D,$F8,$47,$F3,$EB,$DD,$6B,$ED,$D4,$FD,$BB
	.dc.b	$5F,$10,$C9,$FB,$CF,$F4,$1C,$78,$9F,$EA,$7E,$8F,$C3,$56,$F2,$E9
	.dc.b	$8F,$96,$31,$E4,$B5,$FF,$3F,$7D,$07,$F8,$B3,$F1,$37,$EF,$2F,$C8
	.dc.b	$DF,$B2,$7F,$B6,$7F,$4A,$7E,$B0,$FC,$8B,$FC,$7D,$87,$0B,$DF,$E0
	.dc.b	$EB,$DF,$EF,$5A,$CB,$A7,$17,$FA,$F7,$EA,$BF,$39,$8F,$83,$F2,$2F
	.dc.b	$F2,$0C,$FF,$13,$8E,$9A,$F8,$7D,$5F,$36,$1E,$58,$F2,$7C,$27,$FB
	.dc.b	$82,$FF,$4D,$E5,$BF,$0F,$FF,$14,$7D,$6B,$F7,$77,$E2,$AF,$A2,$7E
	.dc.b	$4F,$FE,$F3,$C7,$8E,$1C,$2F,$75,$FC,$67,$F1,$7F,$F8,$01,$97,$5F
	.dc.b	$AE,$FF,$6A,$46,$2D,$F8,$96,$31,$E7,$5F,$E5,$0C,$79,$C7,$73,$F9
	.dc.b	$86,$1C,$F4,$F3,$69,$E3,$F4,$3F,$C2,$FF,$9F,$FF,$56,$FD,$23,$FC
	.dc.b	$5D,$F1,$CF,$CC,$DF,$A9,$BF,$18,$FE,$1F,$7B,$FB,$4F,$E6,$75,$AF
	.dc.b	$88,$CF,$CE,$7F,$AA,$F7,$FE,$24,$2F,$B3,$87,$96,$5D,$FD,$EF,$E1
	.dc.b	$D1,$8C,$3C,$DF,$61,$FC,$57,$F4,$6F,$EB,$DF,$85,$3C,$87,$CA,$7F
	.dc.b	$85,$3F,$0E,$7E,$D3,$FC,$92,$F7,$5A,$F8,$35,$3F,$32,$D7,$C6,$BF
	.dc.b	$23,$FE,$25,$87,$6E,$1F,$4B,$CB,$3F,$F1,$0E,$5F,$C8,$E3,$19,$7E
	.dc.b	$4E,$BF,$99,$FE,$87,$FE,$1D,$FB,$3F,$E9,$3F,$B4,$3F,$70,$FE,$A1
	.dc.b	$ED,$6B,$EB,$15,$A9,$FE,$A9,$ED,$66,$3F,$92,$61,$FE,$27,$FC,$06
	.dc.b	$DF,$0F,$9A,$65,$D7,$F5,$5F,$E6,$D7,$E1,$9C,$E5,$FE,$69,$D1,$F8
	.dc.b	$57,$FB,$01,$FF,$31,$F9,$5F,$98,$FD,$BD,$FB,$F3,$F2,$37,$EA,$DF
	.dc.b	$E8,$BC,$38,$7D,$F7,$E8,$7C,$A6,$1C,$2F,$7F,$85,$76,$FC,$4F,$FA
	.dc.b	$65,$EF,$4F,$E9,$C7,$F8,$0F,$1E,$7A,$3F,$3E,$FF,$5C,$DF,$08,$F2
	.dc.b	$6B,$E3,$F2,$78,$4E,$75,$F0,$CF,$9F,$8B,$FD,$39,$97,$F8,$3F,$F1
	.dc.b	$35,$BF,$80,$5F,$AE,$BE,$9D,$FB,$3B,$F4,$2F,$F7,$5F,$07,$F9,$EB
	.dc.b	$1E,$38,$70,$BD,$D7,$F8,$16,$B5,$8F,$5C,$BD,$A7,$FA,$63,$0F,$0F
	.dc.b	$E8,$DF,$E9,$E8,$FF,$7B,$CF,$8F,$3F,$16,$6F,$D4,$BD,$35,$AD,$3C
	.dc.b	$12,$FA,$CF,$EE,$5E,$33,$F5,$AF,$F2,$A7,$E7,$8F,$DD,$BF,$9C,$FD
	.dc.b	$BB,$F6,$2F,$F0,$2D,$EF,$EF,$D7,$BF,$23,$FD,$D1,$D3,$AC,$BF,$CB
	.dc.b	$DB,$7E,$E3,$0E,$DF,$7F,$FD,$D5,$E5,$87,$1F,$B3,$F1,$DF,$8B,$5F
	.dc.b	$E9,$3E,$D7,$F8,$8B,$E2,$DF,$9C,$3F,$80,$3F,$C1,$FE,$B3,$E0,$5F
	.dc.b	$24,$FD,$5B,$FC,$67,$C0,$C7,$8E,$1C,$2F,$7F,$88,$6A,$7A,$FE,$D1
	.dc.b	$FE,$38,$77,$7F,$E3,$63,$F9,$F6,$7E,$1F,$B6,$FE,$0B,$87,$F8,$2B
	.dc.b	$EE,$BF,$A2,$CE,$75,$F2,$6D,$F9,$5F,$95,$FF,$12,$0F,$D3,$BE,$77
	.dc.b	$93,$FC,$47,$EA,$1F,$84,$7F,$B9,$BC,$7F,$F7,$F5,$EF,$97,$FA,$1F
	.dc.b	$BA,$BD,$F8,$5F,$B2,$AE,$B9,$7D,$9E,$1F,$EB,$07,$F9,$9E,$3C,$1F
	.dc.b	$85,$FF,$58,$84,$75,$D3,$C5,$A7,$9E,$78,$5F,$9B,$2F,$07,$C7,$7F
	.dc.b	$59,$9F,$F3,$97,$E8,$8F,$80,$7E,$CE,$AF,$F6,$77,$EF,$3F,$EF,$AF
	.dc.b	$BE,$FE,$95,$83,$0E,$17,$BF,$C4,$57,$FF,$00,$3F,$74,$DD,$75,$FA
	.dc.b	$1F,$EA,$D8,$78,$33,$EE,$FE,$9B,$E6,$7F,$81,$D7,$4D,$3C,$3F,$8F
	.dc.b	$79,$EB,$C1,$7E,$69,$F2,$E1,$FE,$8B,$DB,$F8,$5F,$E5,$2F,$C9,$DF
	.dc.b	$46,$FC,$D3,$FC,$0D,$FA,$73,$F2,$B7,$EA,$1F,$D2,$75,$FF,$30,$63
	.dc.b	$C7,$0E,$19,$B0,$E1,$EF,$6E,$FF,$D3,$57,$B7,$F7,$FA,$78,$B4,$71
	.dc.b	$7F,$02,$9E,$02,$BB,$6B,$E4,$FA,$DF,$54,$F8,$AF,$D3,$3E,$8C,$7F
	.dc.b	$D5,$1C,$7F,$E3,$CA,$FF,$55,$6E,$97,$E4,$EF,$85,$7E,$03,$F8,$FF
	.dc.b	$85,$FE,$3F,$8F,$1C,$7D,$77,$BA,$D7,$A6,$AD,$65,$D7,$EE,$3F,$C2
	.dc.b	$3F,$6F,$FE,$85,$8F,$06,$18,$C7,$C7,$A7,$8B,$4F,$0F,$8B,$FC,$6E
	.dc.b	$7A,$61,$C6,$B5,$A7,$E8,$FE,$77,$FD,$B3,$E7,$3E,$3D,$F8,$93,$EF
	.dc.b	$3F,$8D,$7D,$BF,$FB,$03,$EE,$3F,$CD,$FA,$7F,$A6,$2F,$75,$AF,$A2
	.dc.b	$FF,$30,$FA,$8F,$B8,$FF,$4E,$67,$D7,$F4,$DF,$F3,$8E,$3C,$74,$B3
	.dc.b	$E1,$7C,$1F,$09,$FE,$63,$FB,$BF,$2E,$9D,$A3,$FC,$CD,$F9,$E7,$FB
	.dc.b	$6F,$3F,$D9,$E3,$FE,$DD,$F1,$5F,$94,$FF,$1C,$FB,$D6,$AF,$CA,$FF
	.dc.b	$C8,$DF,$AC,$7F,$21,$F5,$7F,$D6,$18,$70,$F8,$36,$1C,$2F,$FC,$99
	.dc.b	$F9,$16,$F3,$ED,$9F,$F2,$87,$FA,$7E,$9E,$27,$F3,$B1,$FF,$00,$1F
	.dc.b	$ED,$F1,$9C,$7B,$61,$D7,$CD,$F5,$67,$E2,$C3,$D1,$AF,$9F,$D6,$FD
	.dc.b	$97,$E5,$5F,$E4,$3F,$D4,$BF,$0D,$F7,$3F,$EB,$EF,$C4,$9F,$8F,$7F
	.dc.b	$06,$AD,$79,$1F,$E7,$CF,$C4,$3F,$DB,$7E,$2E,$7E,$C1,$97,$F0,$4E
	.dc.b	$36,$1F,$E4,$FE,$8C,$3C,$B4,$FF,$65,$FC,$87,$8B,$0E,$39,$F5,$F1
	.dc.b	$BF,$D4,$91,$FE,$16,$F9,$EF,$E2,$7D,$EF,$B0,$7E,$83,$F8,$3F,$86
	.dc.b	$FC,$26,$C3,$FC,$BD,$F4,$9F,$E7,$BF,$7F,$F2,$37,$C3,$A7,$FD,$55
	.dc.b	$9F,$5B,$FF,$51,$1C,$5F,$F1,$2F,$EA,$3F,$EA,$77,$C5,$7C,$BE,$67
	.dc.b	$CB,$8F,$5D,$3C,$9F,$37,$FC,$DF,$1F,$E6,$5E,$83,$F4,$57,$D5,$BF
	.dc.b	$4B,$FF,$2B,$7E,$D7,$FE,$8F,$E1,$FB,$DF,$09,$FE,$BE,$BD,$D6,$B8
	.dc.b	$3F,$A6,$65,$ED,$AD,$3F,$D3,$0E,$FD,$1F,$D2,$E3,$96,$5D,$FE,$CF
	.dc.b	$FA,$DD,$72,$3C,$B8,$F6,$8F,$F5,$A4,$67,$E6,$DF,$AB,$3F,$4E,$FE
	.dc.b	$25,$E9,$3F,$4E,$CF,$AB,$BA,$D7,$B6,$AD,$43,$D5,$D7,$D7,$7F,$5F
	.dc.b	$CB,$AE,$56,$B8,$DF,$5F,$6F,$E5,$56,$B1,$E9,$9F,$FA,$77,$B7,$EF
	.dc.b	$6A,$FB,$67,$E9,$1F,$D8,$3D,$57,$EC,$7F,$B6,$FC,$09,$F1,$1F,$AF
	.dc.b	$C3,$85,$EE,$B5,$9E,$71,$E7,$F2,$5F,$CF,$FD,$8F,$D9,$47,$6F,$8E
	.dc.b	$C6,$BE,$28,$DE,$BE,$0F,$57,$F9,$D6,$9C,$F4,$ED,$1F,$E7,$7E,$FF
	.dc.b	$CB,$CF,$F7,$7F,$DB,$9F,$AE,$7F,$AD,$3F,$43,$FE,$E3,$C6,$FD,$6F
	.dc.b	$FE,$00,$78,$0B,$DF,$E0,$6B,$DF,$E8,$DA,$9E,$79,$7E,$70,$ED,$F8
	.dc.b	$17,$FC,$01,$C3,$96,$7D,$FC,$D7,$E6,$7F,$E7,$9F,$87,$FF,$5A,$63
	.dc.b	$DB,$3F,$F0,$1F,$C8,$FE,$77,$E9,$BF,$D1,$3F,$BA,$7F,$9C,$7E,$EF
	.dc.b	$FD,$0F,$F9,$77,$FA,$E3,$93,$F8,$C5,$6B,$EC,$FF,$D1,$74,$A7,$93
	.dc.b	$F9,$3A,$E7,$C5,$F9,$B7,$FD,$2B,$E4,$FF,$89,$5E,$57,$FE,$94,$F6
	.dc.b	$FC,$5E,$AB,$4F,$14,$BC,$7F,$EE,$38,$FF,$03,$FD,$CB,$FD,$85,$F7
	.dc.b	$FF,$86,$FD,$51,$FB,$1B,$F4,$37,$C0,$FF,$7C,$BD,$F8,$4B,$DD,$7F
	.dc.b	$9C,$7E,$7F,$FD,$3F,$5D,$32,$FF,$03,$9F,$94,$7F,$DA,$1F,$11,$FE
	.dc.b	$07,$9D,$F9,$FF,$EB,$CC,$7A,$67,$F9,$A9,$CE,$9E,$58,$FF,$07,$E3
	.dc.b	$FE,$BC,$EB,$FF,$57,$EA,$FC,$15,$F9,$3B,$F0,$4F,$F5,$17,$EB,$BF
	.dc.b	$C3,$FB,$3F,$1B,$F8,$27,$B2,$F7,$5A,$F7,$3A,$67,$FC,$9E,$B4,$FE
	.dc.b	$1D,$F8,$7F,$77,$FE,$1F,$C3,$9E,$7E,$1C,$FC,$1D,$FF,$E0,$DC,$39
	.dc.b	$8C,$3C,$13,$AF,$C4,$BF,$D8,$9F,$6B,$FB,$8F,$D4,$9F,$02,$FD,$E5
	.dc.b	$F8,$7B,$D6,$FF,$80,$3F,$A4,$70,$D3,$F9,$E6,$1C,$2F,$75,$AF,$87
	.dc.b	$FF,$70,$E5,$D7,$2E,$98,$7F,$AE,$38,$13,$9D,$FD,$8E,$F1,$E7,$F8
	.dc.b	$77,$F1,$7B,$E6,$78,$30,$F3,$6B,$E5,$97,$C9,$7F,$6E,$AF,$E6,$9D
	.dc.b	$BF,$E6,$BF,$AE,$F5,$FE,$F8,$FB,$D7,$E8,$FF,$C4,$7D,$55,$EF,$35
	.dc.b	$EF,$F0,$2D,$4F,$F9,$D7,$2F,$F0,$1F,$BF,$FC,$83,$0E,$3E,$0F,$F1
	.dc.b	$29,$F0,$FD,$A7,$F9,$9B,$2E,$7E,$57,$96,$7F,$D0,$98,$F7,$CF,$FC
	.dc.b	$75,$F0,$2F,$CB,$BF,$95,$FA,$3F,$ED,$79,$7E,$31,$FC,$B7,$C0,$79
	.dc.b	$0F,$C8,$3F,$D7,$D8,$70,$F9,$3F,$4D,$EE,$9F,$B4,$EB,$5F,$EE,$99
	.dc.b	$FF,$B8,$FF,$38,$FE,$59,$8F,$1D,$2D,$6B,$2E,$7F,$0F,$F8,$BF,$9F
	.dc.b	$FF,$1A,$5E,$F4,$F2,$7D,$CF,$F1,$7C,$3F,$C0,$1F,$94,$7F,$2A,$FC
	.dc.b	$23,$F9,$37,$BA,$FC,$D1,$FB,$23,$F2,$CF,$CB,$DE,$FE,$13,$F8,$9D
	.dc.b	$6B,$EF,$F3,$F1,$0F,$C5,$7F,$45,$EC,$FF,$50,$C3,$B6,$BF,$CB,$5E
	.dc.b	$5A,$77,$FB,$FF,$24,$A7,$39,$3F,$17,$FF,$0B,$79,$FF,$E1,$BA,$FF
	.dc.b	$0C,$7D,$53,$DD,$74,$7F,$6B,$FC,$69,$F8,$B3,$F3,$D7,$EC,$0C,$B9
	.dc.b	$64,$CB,$96,$3C,$70,$E1,$FA,$26,$FE,$AB,$FB,$4D,$77,$FD,$83,$F7
	.dc.b	$D1,$8D,$7F,$CF,$D8,$FF,$6E,$38,$1E,$67,$CB,$F4,$0D,$BC,$BF,$1D
	.dc.b	$EC,$BE,$5A,$78,$7D,$FF,$7F,$ED,$5F,$E7,$8F,$B5,$7D,$FB,$EB,$9F
	.dc.b	$9B,$FD,$7F,$D4,$BF,$DB,$3C,$8C,$38,$5E,$EB,$5F,$B3,$7C,$19,$E9
	.dc.b	$F8,$97,$F1,$77,$7C,$5F,$C0,$5E,$39,$76,$EA,$7E,$57,$E4,$AF,$F2
	.dc.b	$47,$1F,$FB,$9F,$3E,$D4,$FC,$8F,$FA,$5D,$3F,$65,$7E,$BA,$FC,$4D
	.dc.b	$F9,$9B,$F7,$D7,$A8,$FD,$37,$F2,$1D,$9F,$2F,$1E,$38,$70,$BD,$D3
	.dc.b	$CA,$FF,$66,$D7,$6F,$C4,$7F,$A5,$69,$FE,$20,$8E,$2F,$F5,$03,$FC
	.dc.b	$C0,$F4,$CF,$C5,$9F,$FB,$8F,$3F,$06,$9E,$79,$CF,$C7,$7F,$95,$FF
	.dc.b	$6B,$7C,$73,$F2,$F7,$C3,$BA,$5F,$8C,$BF,$09,$F8,$5D,$DF,$E9,$5C
	.dc.b	$78,$E1,$C2,$F7,$5F,$CF,$AB,$58,$F5,$E0,$FF,$2C,$61,$E0,$FB,$EF
	.dc.b	$E9,$DF,$15,$FE,$3C,$8E,$7A,$78,$74,$FA,$57,$FE,$59,$9C,$E7,$E4
	.dc.b	$EC,$FF,$40,$70,$BE,$37,$FC,$8B,$F8,$73,$F3,$4F,$F7,$17,$7B,$C4
	.dc.b	$7E,$11,$FD,$4B,$E0,$BF,$61,$87,$0B,$DD,$6B,$E4,$FF,$0F,$97,$5D
	.dc.b	$BF,$D1,$D9,$F5,$FD,$A3,$FB,$F6,$3C,$7C,$DB,$5A,$CB,$9E,$9F,$E0
	.dc.b	$E8,$EE,$66,$7F,$CD,$19,$F9,$3C,$AF,$E7,$1F,$8A,$1F,$C5,$7E,$A5
	.dc.b	$F9,$1F,$03,$D5,$F9,$6F,$F9,$6E,$B5,$E2,$2B,$58,$3F,$58,$D6,$3F
	.dc.b	$93,$CF,$FA,$59,$FE,$A7,$C3,$B7,$E4,$7F,$D2,$E1,$1D,$30,$F0,$73
	.dc.b	$79,$2F,$C7,$A7,$E8,$FC,$9F,$F7,$47,$CD,$FF,$9D,$3F,$5F,$FF,$60
	.dc.b	$FD,$1F,$F2,$9F,$B5,$7D,$23,$89,$FA,$CB,$E5,$7F,$BA,$F0,$7F,$AE
	.dc.b	$71,$E3,$F0,$CC,$78,$E0,$FA,$EE,$1A,$77,$FC,$39,$AF,$F9,$DE,$31
	.dc.b	$F9,$CD,$CF,$A5,$6B,$2E,$9B,$78,$E7,$99,$AC,$3D,$10,$F6,$FD,$B3
	.dc.b	$FE,$CD,$E9,$FE,$BB,$F8,$83,$F7,$F7,$CF,$3E,$9D,$FA,$3B,$E1,$3A
	.dc.b	$E3,$FC,$0F,$E6,$BF,$DC,$F8,$70,$BD,$FC,$46,$B3,$EB,$7F,$CC,$4C
	.dc.b	$5F,$F8,$41,$C7,$3F,$C8,$7A,$69,$E3,$D3,$C5,$EB,$FA,$27,$39,$F7
	.dc.b	$FC,$17,$FD,$CD,$E2,$3E,$A7,$FD,$57,$F8,$8B,$E1,$5F,$31,$FC,$73
	.dc.b	$E8,$DF,$A4,$3F,$47,$FF,$62,$63,$C7,$4F,$F7,$86,$3C,$70,$E1,$7B
	.dc.b	$E7,$5E,$ED,$7D,$FF,$6D,$F7,$46,$B4,$F1,$E9,$E2,$8F,$85,$7C,$6F
	.dc.b	$B6,$FF,$E4,$8C,$BF,$9D,$D7,$13,$C7,$7E,$A9,$D7,$88,$9F,$C8,$7C
	.dc.b	$6F,$FB,$2B,$F3,$EF,$F2,$87,$E7,$3F,$CD,$FF,$54,$FF,$40,$72,$7E
	.dc.b	$57,$9F,$FB,$6D,$6B,$A5,$5A,$F5,$B9,$E5,$D6,$BF,$CB,$5F,$C0,$1F
	.dc.b	$EC,$9C,$FB,$7C,$2B,$4C,$5E,$E3,$19,$74,$FB,$C7,$E7,$FE,$7D,$7C
	.dc.b	$35,$AD,$7F,$D2,$9E,$87,$F8,$2B,$E0,$BF,$DC,$67,$FB,$83,$EF,$3F
	.dc.b	$AA,$64,$FD,$21,$FA,$1F,$F3,$CF,$88,$5F,$0F,$C4,$18,$70,$BD,$FE
	.dc.b	$AB,$FE,$0D,$AD,$4F,$F3,$6E,$0F,$95,$1F,$EE,$EF,$DD,$31,$1D,$30
	.dc.b	$E7,$F2,$7F,$D8,$71,$FC,$C6,$5E,$0A,$E3,$5A,$D3,$CD,$F3,$9F,$DB
	.dc.b	$BE,$C7,$C7,$F8,$9B,$E1,$5F,$8C,$7F,$A3,$57,$E0,$3F,$E2,$7C,$BE
	.dc.b	$E1,$FE,$6C,$E0,$5F,$1B,$E1,$A6,$EB,$F0,$58,$F6,$7F,$B5,$1C,$7E
	.dc.b	$B7,$FB,$7B,$AC,$39,$E5,$E2,$FB,$4F,$F5,$07,$7F,$F8,$97,$0F,$3C
	.dc.b	$E7,$ED,$7F,$5B,$87,$93,$F0,$5F,$F3,$77,$E6,$3E,$8F,$CB,$9F,$26
	.dc.b	$FD,$C9,$F8,$87,$FA,$AF,$0F,$CB,$61,$C3,$8D,$87,$0B,$7E,$43,$BB
	.dc.b	$FF,$71,$7D,$B3,$3F,$D5,$F9,$FF,$E3,$9C,$BF,$CF,$DF,$AC,$62,$FC
	.dc.b	$97,$BC,$FA,$EB,$FE,$55,$9E,$97,$E1,$D3,$CF,$E6,$7F,$8F,$67,$FD
	.dc.b	$67,$FA,$BF,$F8,$83,$F5,$87,$E3,$2F,$C4,$9D,$FF,$E5,$8E,$BF,$F2
	.dc.b	$AF,$C0,$30,$E1,$7B,$A7,$EC,$5F,$E4,$EC,$BF,$95,$53,$F6,$DF,$F2
	.dc.b	$AC,$78,$3D,$4A,$E9,$69,$E2,$D3,$F8,$46,$9E,$0C,$3B,$C3,$EF,$BF
	.dc.b	$C7,$7E,$A3,$F0,$3F,$F0,$EF,$CE,$3F,$2E,$7E,$DE,$FD,$D5,$FC,$03
	.dc.b	$F8,$9D,$D8,$F1,$FA,$56,$3C,$70,$E1,$F8,$96,$EB,$B7,$EF,$DF,$8C
	.dc.b	$C3,$8E,$9F,$E2,$C7,$F2,$33,$CF,$0E,$5F,$51,$E6,$AF,$2E,$BE,$4F
	.dc.b	$CA,$BF,$D1,$BF,$9F,$7F,$36,$FC,$2B,$F7,$DF,$91,$3F,$4B,$F4,$79
	.dc.b	$9F,$EF,$4A,$D7,$AA,$AD,$72,$3F,$4B,$D7,$C6,$7F,$AF,$B8,$7F,$DD
	.dc.b	$1F,$83,$7E,$B3,$1F,$E8,$7F,$8D,$7F,$A8,$A7,$F9,$54,$74,$C7,$C1
	.dc.b	$C3,$E4,$C3,$FD,$8F,$9F,$7D,$5B,$FF,$8D,$BE,$33,$F6,$FF,$99,$7E
	.dc.b	$6B,$37,$EE,$EF,$D3,$BF,$73,$FB,$33,$ED,$7F,$D4,$98,$70,$F4,$FF
	.dc.b	$9D,$5E,$FD,$D5,$FF,$96,$7D,$9F,$E9,$D9,$FE,$9B,$0F,$E9,$B1,$8C
	.dc.b	$3F,$A7,$E7,$D0,$FE,$89,$5E,$3C,$7A,$7C,$2F,$9E,$73,$5E,$0D,$3F
	.dc.b	$86,$F9,$DF,$CC,$67,$FA,$66,$9F,$EC,$EF,$D0,$DF,$8C,$FF,$A6,$7F
	.dc.b	$1C,$0F,$0F,$F9,$65,$F0,$FC,$49,$7C,$3F,$05,$DD,$FF,$8A,$3C,$CF
	.dc.b	$E5,$17,$FC,$72,$DF,$97,$FF,$30,$CF,$C1,$3F,$E0,$7A,$DC,$78,$F3
	.dc.b	$F1,$61,$E7,$AE,$59,$77,$8F,$DC,$47,$90,$FE,$EB,$F5,$DF,$ED,$9F
	.dc.b	$BE,$FF,$4F,$7E,$51,$FE,$4D,$FC,$2F,$5F,$37,$FE,$A3,$FC,$3F,$FD
	.dc.b	$11,$F1,$AA,$FF,$29,$7E,$E1,$AC,$7F,$DC,$59,$79,$7E,$43,$FC,$1F
	.dc.b	$1F,$B8,$8C,$1F,$E0,$A7,$C1,$7C,$FF,$10,$F4,$65,$DF,$4F,$1F,$E2
	.dc.b	$3F,$D9,$71,$FE,$83,$E6,$FF,$21,$FD,$45,$FC,$03,$FE,$E7,$FA,$A7
	.dc.b	$EA,$6F,$D0,$BF,$CA,$18,$70,$F8,$C6,$1C,$3F,$15,$FF,$77,$DB,$F4
	.dc.b	$DD,$F1,$7F,$82,$74,$FE,$5D,$FB,$DF,$F9,$9D,$F0,$47,$F9,$B0,$3E
	.dc.b	$3D,$7C,$5F,$31,$E7,$9E,$59,$F8,$27,$CD,$F9,$1E,$67,$F8,$8F,$87
	.dc.b	$FC,$57,$E1,$BF,$DA,$1F,$35,$D3,$F4,$EF,$1D,$6B,$DD,$FE,$4D,$D5
	.dc.b	$FE,$CF,$D7,$F9,$9F,$D0,$FE,$23,$E0,$BF,$DB,$5F,$3F,$FE,$7D,$7F
	.dc.b	$DB,$97,$74,$ED,$87,$FB,$5F,$E1,$BC,$55,$AD,$7C,$3E,$C7,$F8,$A7
	.dc.b	$C8,$7B,$BF,$DC,$3F,$05,$7E,$C8,$FD,$9A,$7E,$02,$FC,$33,$FD,$25
	.dc.b	$FB,$FB,$1E,$3C,$9F,$E3,$3C,$78,$E1,$C2,$FF,$D3,$1F,$41,$BC,$FB
	.dc.b	$67,$FE,$BB,$C3,$F8,$A9,$C5,$F0,$38,$3F,$D7,$AF,$8E,$FF,$D1,$F7
	.dc.b	$FE,$7F,$C7,$C3,$5A,$9F,$07,$D4,$7F,$78,$8F,$F2,$37,$DA,$3F,$57
	.dc.b	$7E,$42,$FA,$87,$E2,$BF,$88,$FC,$FD,$FB,$17,$F7,$3C,$38,$5E,$EB
	.dc.b	$5F,$AC,$E7,$1F,$F7,$F6,$5F,$E5,$C3,$FC,$33,$7C,$1C,$1E,$28,$F0
	.dc.b	$E5,$E0,$F2,$7F,$95,$CF,$3C,$7B,$43,$F1,$FF,$F2,$47,$D9,$FF,$89
	.dc.b	$3F,$2E,$7E,$D0,$FC,$E5,$F9,$1B,$F2,$C7,$E7,$FF,$A8,$E7,$FF,$2D
	.dc.b	$EA,$C7,$8E,$1C,$2F,$7E,$5B,$F3,$5D,$D7,$6F,$D8,$3F,$CB,$B8,$71
	.dc.b	$D3,$C0,$FF,$A6,$35,$F2,$47,$8F,$5F,$16,$BF,$DD,$6B,$9C,$E7,$3F
	.dc.b	$2F,$83,$FA,$5F,$2B,$D7,$F8,$FB,$E2,$DE,$C3,$F4,$47,$EA,$4F,$D1
	.dc.b	$7F,$4F,$87,$0B,$DD,$6B,$F6,$5C,$E8,$CB,$F7,$8B,$F4,$AB,$1D,$F3
	.dc.b	$ED,$E7,$F8,$EB,$86,$3D,$3E,$07,$F1,$5C,$1F,$37,$F3,$97,$E8,$FF
	.dc.b	$E1,$1F,$E0,$07,$E9,$3F,$E7,$03,$FC,$75,$C8,$AD,$FA,$6A,$DF,$DF
	.dc.b	$3F,$2F,$DD,$7F,$85,$72,$FC,$39,$DF,$2F,$F1,$F3,$FE,$9F,$C7,$A4
	.dc.b	$78,$72,$7E,$7F,$E6,$9C,$E7,$DE,$7E,$97,$9C,$FD,$F3,$FC,$DB,$D8
	.dc.b	$FC,$ED,$F9,$EF,$F8,$D7,$F4,$4F,$F4,$77,$C7,$BF,$7C,$FD,$F6,$1C
	.dc.b	$2F,$7E,$DA,$F7,$FB,$56,$B8,$7F,$D6,$99,$F5,$FC,$C3,$FD,$87,$AF
	.dc.b	$86,$73,$8F,$81,$FE,$BD,$97,$5D,$7C,$5F,$4E,$FC,$1B,$F8,$B4,$E7
	.dc.b	$5F,$04,$FF,$B4,$AF,$FC,$9D,$F7,$6F,$DE,$3F,$79,$FB,$B7,$D4,$F9
	.dc.b	$F7,$E6,$1F,$E1,$3D,$9E,$67,$FA,$3B,$0E,$1E,$B3,$0E,$17,$FA,$BB
	.dc.b	$DE,$5F,$EC,$FB,$ED,$F8,$17,$F9,$61,$E3,$1F,$DD,$F1,$E9,$9F,$8B
	.dc.b	$A1,$A7,$EC,$6B,$CF,$39,$87,$E5,$3F,$ED,$BF,$A9,$FF,$50,$D3,$F6
	.dc.b	$87,$E8,$1F,$EF,$DE,$B7,$D3,$3F,$62,$7E,$34,$FD,$DB,$FC,$B9,$87
	.dc.b	$0F,$AF,$FB,$57,$BA,$FC,$BE,$7D,$AB,$59,$FF,$AA,$1E,$FF,$AA,$FF
	.dc.b	$97,$5E,$59,$F8,$33,$FF,$7D,$5F,$1D,$3C,$99,$78,$FC,$3F,$F6,$F7
	.dc.b	$C5,$FF,$AE,$BF,$47,$FD,$07,$EE,$9F,$DA,$BF,$1B,$7D,$C3,$F4,$1F
	.dc.b	$B1,$F3,$3F,$CE,$AF,$7E,$DA,$F7,$4F,$C0,$D7,$BF,$4D,$FA,$07,$C8
	.dc.b	$E5,$F0,$81,$97,$89,$D5,$71,$D3,$C9,$A3,$E8,$38,$D7,$0C,$55,$AF
	.dc.b	$9F,$FF,$18,$7D,$AB,$F3,$4F,$E6,$38,$7F,$97,$FE,$19,$FA,$4B,$F3
	.dc.b	$6F,$F1,$95,$EF,$D2,$ED,$F7,$0B,$DF,$EB,$DA,$C9,$F4,$BF,$EF,$2C
	.dc.b	$38,$6F,$FE,$2C,$3F,$D8,$F3,$E0,$C3,$97,$DF,$79,$23,$18,$FF,$49
	.dc.b	$A7,$6B,$F2,$EF,$F7,$6F,$E3,$4F,$C6,$3F,$BF,$7C,$7F,$F6,$0F,$90
	.dc.b	$7E,$6D,$F9,$3B,$DD,$6B,$E7,$7F,$A5,$7E,$CD,$FA,$AC,$BA,$67,$F3
	.dc.b	$F3,$E9,$F4,$3F,$E7,$6C,$78,$FE,$77,$6B,$59,$73,$D1,$D3,$DE,$73
	.dc.b	$87,$F5,$AD,$3F,$AC,$7C,$2F,$F6,$29,$FB,$9D,$FF,$DE,$7E,$A5,$FB
	.dc.b	$B3,$F7,$67,$E7,$8F,$D5,$DF,$96,$FF,$62,$C3,$85,$EE,$B5,$FA,$16
	.dc.b	$7E,$05,$FA,$D7,$F6,$0B,$ED,$A7,$F5,$F3,$C3,$2D,$3C,$3A,$78,$3F
	.dc.b	$3D,$F3,$57,$E4,$2B,$59,$F7,$D3,$FC,$89,$C4,$FC,$D5,$F5,$8F,$C7
	.dc.b	$7F,$DD,$FF,$0F,$EC,$E9,$FF,$22,$DE,$EB,$5F,$06,$AF,$83,$F9,$4E
	.dc.b	$B1,$7E,$BD,$FA,$3D,$B1,$85,$F4,$3F,$D2,$39,$73,$D3,$BF,$D1,$F9
	.dc.b	$31,$F1,$C7,$8B,$D8,$FD,$55,$BF,$13,$FD,$66,$5F,$C3,$FE,$F7,$FC
	.dc.b	$89,$F5,$CF,$C1,$9F,$A6,$BF,$5F,$7E,$7F,$FB,$8C,$78,$FC,$47,$F9
	.dc.b	$EB,$EB,$98,$F1,$C3,$87,$EC,$BB,$D3,$F7,$57,$FE,$80,$FA,$DF,$56
	.dc.b	$3C,$B5,$FF,$2A,$57,$EF,$23,$19,$74,$D7,$C7,$F2,$FE,$A9,$CE,$9D
	.dc.b	$FF,$2B,$FF,$3B,$6D,$FD,$13,$F1,$2F,$EE,$9F,$72,$FD,$5F,$FC,$57
	.dc.b	$FB,$0F,$F9,$DA,$B5,$F4,$DF,$E6,$69,$FF,$17,$FE,$15,$9F,$98,$7E
	.dc.b	$7F,$F3,$FF,$0A,$F8,$F8,$75,$E2,$FF,$46,$3C,$B3,$ED,$E2,$F8,$B0
	.dc.b	$FF,$66,$67,$D7,$6F,$F0,$8F,$E8,$8F,$C0,$BF,$8D,$FE,$7A,$FD,$B7
	.dc.b	$E7,$FE,$19,$6F,$FA,$4D,$EF,$D0,$5E,$FB,$BF,$D4,$5F,$BB,$FF,$69
	.dc.b	$E0,$FE,$4D,$7D,$7F,$7E,$FF,$07,$C2,$39,$3F,$E1,$47,$A5,$F3,$EA
	.dc.b	$F2,$EB,$E4,$D7,$F0,$3A,$7E,$8B,$E9,$7F,$DC,$7E,$D3,$F4,$9F,$EB
	.dc.b	$3F,$8A,$7F,$3B,$FD,$7B,$FC,$AB,$5A,$9C,$FC,$AF,$F8,$07,$C8,$FE
	.dc.b	$E3,$BF,$F7,$5F,$C6,$7F,$D0,$AF,$F6,$68,$CD,$FF,$BC,$3E,$0B,$FC
	.dc.b	$AE,$F3,$F9,$B6,$6F,$B9,$F2,$63,$FE,$9D,$CF,$C7,$F6,$FF,$EA,$3E
	.dc.b	$BF,$CE,$FC,$D3,$E2,$5F,$93,$7F,$99,$7F,$65,$78,$AF,$CE,$BF,$C0
	.dc.b	$D8,$70,$BD,$F9,$EB,$DD,$7F,$84,$6B,$5F,$C0,$1F,$E2,$47,$86,$37
	.dc.b	$93,$FA,$C6,$7E,$18,$F0,$67,$FD,$A2,$7F,$D6,$15,$C3,$2E,$D2,$FD
	.dc.b	$3F,$F8,$8F,$DF,$7F,$6E,$FA,$17,$EC,$4F,$CD,$BF,$C1,$1C,$3F,$8B
	.dc.b	$E9,$F9,$B7,$BA,$D7,$D3,$A9,$FA,$5F,$FA,$2A,$BA,$6D,$FC,$8C,$FF
	.dc.b	$49,$FD,$23,$F4,$7F,$F4,$3C,$F3,$CF,$BD,$71,$AF,$EF,$99,$76,$E9
	.dc.b	$FF,$73,$7E,$28,$FC,$95,$F7,$1F,$C2,$FE,$E3,$FC,$01,$FA,$A3,$F6
	.dc.b	$4F,$F7,$3D,$EE,$B5,$E2,$29,$FB,$7E,$B1,$7E,$F9,$FB,$3D,$BF,$DA
	.dc.b	$D1,$D9,$FF,$7F,$63,$CF,$DF,$FF,$23,$E2,$FA,$5F,$2D,$F9,$27,$3F
	.dc.b	$15,$FE,$84,$FB,$AF,$D6,$FB,$4F,$94,$FE,$F3,$EE,$3F,$5E,$FE,$A3
	.dc.b	$E8,$FF,$8F,$2F,$75,$A9,$7E,$CB,$FB,$3D,$FF,$D1,$7E,$B7,$FB,$C9
	.dc.b	$FF,$41,$DF,$EE,$BF,$75,$FD,$AB,$CB,$EA,$BF,$96,$FB,$1E,$3B,$FF
	.dc.b	$7C,$E5,$E2,$FC,$57,$FA,$9F,$E8,$DF,$E6,$5F,$C5,$9F,$20,$FD,$CB
	.dc.b	$FD,$05,$D1,$FB,$5F,$23,$F5,$F7,$BA,$D7,$E0,$CA,$7E,$63,$FE,$CC
	.dc.b	$AF,$F3,$87,$D6,$FF,$B1,$9E,$DC,$5F,$85,$79,$65,$DA,$38,$7E,$07
	.dc.b	$FE,$D9,$CB,$AF,$E0,$5F,$EA,$EF,$DC,$9F,$A8,$7F,$A4,$65,$FA,$33
	.dc.b	$F4,$E7,$E9,$1F,$E4,$5E,$CF,$F3,$3E,$1C,$3F,$03,$F6,$DE,$EB,$F2
	.dc.b	$95,$AC,$5F,$87,$7F,$1E,$C3,$86,$BF,$EB,$07,$96,$7E,$0C,$FB,$CF
	.dc.b	$D5,$9C,$E4,$FC,$E3,$FB,$1E,$5F,$E7,$8F,$75,$FB,$33,$F2,$2F,$F4
	.dc.b	$8F,$EF,$DF,$E8,$DF,$83,$72,$FD,$FB,$DD,$6B,$E2,$3F,$88,$7E,$73
	.dc.b	$9C,$7F,$D9,$1B,$7F,$B1,$2F,$6D,$8E,$DE,$9B,$3F,$F4,$D7,$DB,$7F
	.dc.b	$B6,$2F,$AC,$E7,$1F,$F6,$3E,$1E,$1F,$C1,$DF,$25,$FE,$8E,$FC,$3C
	.dc.b	$3F,$7D,$FD,$4F,$0F,$B6,$B7,$F8,$5E,$AB,$5E,$C7,$3E,$3F,$E9,$3A
	.dc.b	$AF,$AE,$7E,$97,$8D,$F6,$F0,$BC,$D1,$88,$FF,$51,$CF,$E8,$FE,$4B
	.dc.b	$FB,$5F,$B7,$FE,$26,$FC,$7F,$F8,$9F,$CC,$FF,$1B,$DB,$FD,$4F,$5A
	.dc.b	$FA,$C5,$6B,$DE,$E5,$F3,$FF,$D0,$72,$FF,$76,$DF,$F7,$6F,$80,$FE
	.dc.b	$BE,$F1,$BF,$E3,$FF,$75,$FF,$00,$23,$1F,$0D,$FE,$56,$B6,$6F,$C4
	.dc.b	$3F,$C4,$7F,$4F,$FE,$7F,$FD,$0F,$F2,$FE,$E7,$F0,$1F,$CA,$1F,$69
	.dc.b	$C2,$F7,$F8,$4A,$F7,$EF,$AF,$FD,$77,$97,$5C,$FF,$D3,$59,$F5,$7F
	.dc.b	$D3,$D1,$95,$1F,$E8,$FF,$85,$FF,$20,$4E,$63,$C9,$E4,$FF,$8C,$B0
	.dc.b	$FF,$42,$4E,$72,$F1,$4E,$7D,$CF,$F5,$44,$F5,$F9,$2F,$F0,$17,$F0
	.dc.b	$07,$FA,$97,$E6,$5F,$98,$BF,$2C,$4F,$B7,$6F,$F5,$85,$F0,$FC,$3B
	.dc.b	$FC,$01,$7B,$F5,$75,$4A,$FF,$07,$47,$A3,$8F,$FC,$9B,$1E,$09,$F7
	.dc.b	$C6,$3E,$93,$BE,$1D,$3E,$DB,$CF,$39,$CB,$CD,$93,$F5,$0C,$E3,$FE
	.dc.b	$C3,$F9,$17,$D3,$BF,$4A,$FF,$59,$7C,$E3,$F4,$0F,$F7,$57,$E9,$7E
	.dc.b	$6B,$DD,$6A,$7F,$DD,$DF,$9A,$E7,$07,$E6,$3E,$9C,$BB,$47,$57,$D2
	.dc.b	$62,$F9,$70,$79,$23,$FA,$C4,$F8,$FE,$1B,$FC,$1B,$8F,$F1,$0F,$C3
	.dc.b	$1F,$B3,$FF,$9C,$3E,$9D,$F9,$33,$77,$D6,$BE,$43,$FC,$E5,$E0,$FD
	.dc.b	$4C,$78,$EB,$FE,$E3,$C7,$8E,$1C,$2F,$7D,$BF,$EC,$6C,$FB,$67,$D7
	.dc.b	$0F,$F7,$66,$3C,$9F,$F0,$E4,$38,$BF,$D1,$D3,$9C,$BA,$E9,$E3,$F9
	.dc.b	$97,$81,$FE,$F3,$AD,$61,$E9,$AD,$69,$9C,$FF,$D1,$BE,$CB,$F3,$D7
	.dc.b	$D3,$BF,$20,$4B,$F2,$4F,$F5,$BF,$EB,$3E,$9E,$1F,$F8,$01,$3C,$6F
	.dc.b	$83,$BF,$C3,$1F,$B6,$6E,$B9,$E5,$FE,$4E,$7F,$D3,$BB,$FB,$ED,$FB
	.dc.b	$1F,$FA,$AE,$78,$47,$5D,$3C,$7A,$3E,$87,$D5,$5E,$2C,$7D,$35,$AF
	.dc.b	$94,$CC,$BF,$34,$7D,$13,$F6,$47,$E8,$4F,$DE,$7F,$DB,$7F,$3B,$FD
	.dc.b	$23,$83,$E9,$61,$C2,$F7,$5A,$FD,$17,$D9,$39,$CB,$EC,$F3,$7F,$AE
	.dc.b	$23,$BF,$E8,$BF,$62,$31,$A7,$87,$4F,$07,$E2,$DF,$E4,$FD,$3B,$CE
	.dc.b	$73,$FF,$76,$FE,$3D,$F6,$F3,$FF,$61,$F7,$3F,$1E,$FF,$60,$7E,$CC
	.dc.b	$FC,$93,$FD,$15,$F8,$DB,$E9,$9A,$7F,$B2,$B1,$FF,$64,$61,$C2,$F7
	.dc.b	$3A,$9E,$B2,$FC,$BF,$F1,$F1,$8F,$C4,$6A,$39,$AC,$BC,$5E,$C7,$F8
	.dc.b	$D6,$7A,$1B,$D3,$D1,$5B,$CB,$FD,$41,$F8,$47,$B6,$BA,$7E,$9B,$FE
	.dc.b	$DA,$F3,$9F,$A7,$7E,$4F,$F3,$87,$E7,$0F,$D7,$3A,$E0,$C3,$85,$EE
	.dc.b	$9F,$3F,$A9,$FF,$72,$FE,$6D,$EA,$BE,$1F,$19,$FC,$4D,$FD,$89,$FE
	.dc.b	$F0,$9E,$7E,$1B,$3F,$EB,$9C,$BF,$2E,$73,$0F,$CE,$3D,$57,$E3,$F8
	.dc.b	$B7,$E1,$4F,$D8,$3F,$5D,$E1,$FC,$9F,$0B,$C9,$5A,$9C,$DB,$F3,$2C
	.dc.b	$E1,$E7,$CF,$FA,$C7,$15,$C3,$87,$93,$F9,$DC,$3C,$35,$FD,$76,$FC
	.dc.b	$13,$FA,$1F,$81,$FB,$1D,$FF,$EB,$EF,$C3,$BF,$C5,$1F,$9C,$7F,$45
	.dc.b	$C9,$FB,$27,$F9,$CF,$F6,$2F,$C5,$7A,$BF,$7F,$5F,$F7,$27,$D6,$30
	.dc.b	$E1,$7F,$E7,$9F,$DE,$B7,$5F,$85,$CB,$FC,$69,$F9,$7F,$E4,$63,$C3
	.dc.b	$8F,$39,$F0,$C7,$FA,$75,$F1,$FC,$77,$F7,$5C,$3F,$D0,$93,$D0,$F0
	.dc.b	$E3,$E0,$FC,$6F,$9C,$7E,$4F,$E4,$5F,$94,$3F,$50,$7E,$F1,$FC,$67
	.dc.b	$B3,$FC,$B7,$E9,$73,$BD,$D6,$BD,$AF,$B5,$F7,$2F,$A2,$D6,$5F,$E7
	.dc.b	$AC,$FF,$D6,$58,$70,$6C,$77,$CF,$B5,$F9,$27,$FB,$C4,$E7,$07,$EC
	.dc.b	$DF,$2F,$A3,$FD,$DF,$8B,$F0,$A7,$A0,$FB,$CF,$C6,$7F,$00,$BF,$2F
	.dc.b	$FD,$E6,$1C,$2F,$75,$A9,$FD,$FC,$E6,$FE,$86,$9E,$FB,$7E,$A1,$8A
	.dc.b	$D6,$1D,$BC,$9F,$46,$7E,$3C,$7C,$5C,$8D,$1F,$93,$FE,$2B,$F2,$0F
	.dc.b	$C2,$7E,$19,$FC,$A7,$F4,$87,$E8,$3F,$DF,$7F,$01,$FE,$09,$F4,$BF
	.dc.b	$C4,$EF,$7F,$02,$BD,$FB,$0B,$FE,$15,$F9,$BE,$F9,$7F,$87,$67,$D5
	.dc.b	$FC,$57,$1B,$2E,$7C,$DF,$EC,$F9,$FF,$6E,$67,$D7,$5F,$17,$C4,$F9
	.dc.b	$E7,$9C,$E7,$4F,$07,$C5,$66,$7F,$DC,$9E,$B3,$F3,$9F,$F7,$9F,$E6
	.dc.b	$6F,$D5,$1F,$1A,$FD,$87,$CD,$87,$0B,$DD,$6B,$F1,$CC,$F1,$7F,$6B
	.dc.b	$FC,$CB,$CC,$FF,$6D,$AB,$9F,$7F,$B1,$69,$DF,$C1,$F2,$5E,$E7,$39
	.dc.b	$3E,$33,$3F,$53,$F0,$FF,$35,$7E,$D3,$FC,$CF,$F0,$4F,$F0,$DF,$57
	.dc.b	$C7,$E1,$F9,$77,$BA,$D7,$E6,$9E,$68,$FE,$E9,$3D,$3A,$FC,$F1,$BF
	.dc.b	$8E,$FE,$8D,$87,$3A,$6D,$FE,$75,$B6,$7E,$2B,$FC,$3D,$7C,$6F,$C0
	.dc.b	$1E,$93,$EA,$3F,$6D,$FB,$AB,$F1,$D7,$E7,$CF,$7F,$85,$EF,$D6,$5E
	.dc.b	$E9,$F9,$0E,$BF,$33,$FC,$16,$7D,$B3,$B8,$F1,$C3,$B5,$FD,$FA,$D6
	.dc.b	$3D,$33,$7C,$1E,$7F,$06,$7E,$3F,$FE,$07,$F0,$DF,$BE,$3F,$7E,$FF
	.dc.b	$13,$F3,$FC,$3C,$D8,$70,$FA,$1F,$C3,$DF,$F7,$5E,$4F,$85,$3F,$A1
	.dc.b	$7F,$37,$CF,$FC,$B6,$F8,$E7,$EA,$AE,$59,$F8,$3F,$26,$F2,$4F,$8F
	.dc.b	$1F,$E1,$18,$FC,$DF,$27,$FC,$3D,$F8,$13,$F3,$87,$EE,$CF,$D4,$5F
	.dc.b	$9D,$EF,$E5,$97,$BA,$D7,$85,$CF,$F5,$4F,$97,$97,$F4,$DC,$FF,$C1
	.dc.b	$5C,$9E,$F8,$EF,$A7,$6F,$BF,$F9,$91,$8C,$7F,$D0,$97,$FE,$13,$FD
	.dc.b	$17,$FC,$49,$F1,$8F,$8F,$7E,$87,$FC,$23,$F2,$97,$ED,$5F,$9A,$D5
	.dc.b	$87,$0B,$DF,$B4,$BD,$FB,$3D,$73,$FF,$0A,$71,$FF,$A7,$EF,$F8,$F6
	.dc.b	$1E,$0F,$DF,$ED,$6B,$5F,$16,$BE,$1F,$81,$FE,$CB,$7D,$E3,$C9,$E3
	.dc.b	$62,$FA,$FE,$BB,$FE,$11,$FC,$1D,$FA,$13,$F5,$4F,$E3,$5F,$BF,$F8
	.dc.b	$EF,$FC,$A1,$7B,$AD,$7C,$23,$F3,$3D,$70,$AB,$FC,$39,$7F,$C7,$B5
	.dc.b	$FF,$25,$47,$FB,$8E,$3E,$8C,$63,$EA,$7F,$DC,$B8,$F8,$7F,$20,$F3
	.dc.b	$57,$82,$73,$A7,$F8,$4A,$3B,$FD,$53,$F2,$67,$EB,$0F,$07,$F9,$07
	.dc.b	$E7,$6F,$D0,$77,$5A,$9C,$FC,$DF,$2F,$D6,$7E,$5E,$5F,$E6,$DC,$BA
	.dc.b	$3E,$93,$E9,$BD,$F2,$ED,$F0,$EF,$D8,$7C,$FC,$7F,$E3,$CC,$FB,$4F
	.dc.b	$F9,$D3,$0F,$F2,$D7,$CE,$7F,$84,$7F,$69,$7E,$48,$F2,$3F,$D9,$9F
	.dc.b	$83,$7F,$A5,$FE,$95,$E1,$70,$C7,$FC,$AD,$87,$0B,$DF,$C6,$FF,$A7
	.dc.b	$BC,$4D,$52,$B4,$FF,$4E,$CF,$C1,$9B,$F5,$0F,$ED,$11,$8C,$7C,$3E
	.dc.b	$0F,$F7,$B8,$C7,$E1,$FF,$EB,$2C,$D1,$FE,$30,$3F,$C5,$5F,$93,$7A
	.dc.b	$BE,$E3,$FC,$F5,$FB,$7B,$F1,$D7,$E1,$D7,$CF,$E3,$BF,$F9,$BF,$0E
	.dc.b	$1F,$7E,$C3,$85,$FF,$10,$BD,$DA,$FE,$3F,$EA,$1E,$88,$E5,$C3,$FC
	.dc.b	$4D,$FD,$DE,$9E,$48,$FF,$45,$65,$D7,$3F,$F2,$E5,$7C,$6C,$3C,$39
	.dc.b	$F9,$FF,$24,$F6,$4F,$FA,$0F,$F7,$0F,$AB,$E0,$BF,$77,$7E,$56,$FC
	.dc.b	$17,$F2,$1F,$37,$F4,$2F,$75,$A9,$FC,$A7,$E1,$B9,$DF,$F2,$99,$7D
	.dc.b	$33,$B3,$FC,$69,$C0,$BE,$78,$F7,$E7,$F2,$CE,$6B,$F3,$D8,$F7,$FC
	.dc.b	$0F,$FD,$35,$CE,$FD,$33,$FC,$BD,$FB,$FF,$FB,$17,$C1,$FF,$80,$3F
	.dc.b	$7E,$FC,$AD,$FA,$17,$F3,$EC,$38,$69,$FE,$A3,$C3,$87,$E2,$BF,$E7
	.dc.b	$9B,$DF,$47,$FB,$AF,$3E,$BF,$1F,$FE,$F6,$C3,$FC,$C9,$8F,$25,$8F
	.dc.b	$FB,$57,$F3,$6F,$F1,$CC,$E6,$3A,$E9,$E3,$F9,$7F,$4E,$3D,$36,$F3
	.dc.b	$E9,$FE,$A0,$CB,$C0,$7F,$AD,$FD,$A7,$E1,$EF,$DC,$3F,$95,$7A,$4F
	.dc.b	$C8,$DF,$BC,$39,$78,$E3,$FE,$35,$C3,$87,$C2,$B0,$E1,$6F,$D0,$37
	.dc.b	$9F,$6F,$D6,$FD,$3B,$FF,$A2,$63,$C3,$AE,$33,$F2,$67,$D3,$2F,$17
	.dc.b	$C1,$FB,$27,$35,$E6,$DB,$FC,$17,$C9,$F1,$FF,$57,$7E,$63,$FE,$34
	.dc.b	$FD,$1F,$F8,$27,$E2,$8F,$CA,$FF,$C0,$B7,$BA,$D7,$A4,$AD,$7C,$3F
	.dc.b	$BB,$1E,$7F,$33,$E5,$D5,$1D,$BE,$BF,$0E,$B4,$EF,$F6,$7E,$7A,$DE
	.dc.b	$7D,$B2,$FF,$4C,$E9,$FD,$6A,$7F,$DD,$1F,$8C,$FE,$D7,$F1,$87,$E9
	.dc.b	$EF,$CC,$BF,$D9,$5F,$7C,$FC,$7D,$8F,$1C,$58,$F1,$C3,$85,$FF,$B4
	.dc.b	$7F,$1F,$DF,$C8,$7F,$AD,$EB,$B5,$FE,$3A,$31,$EB,$BF,$44,$FC,$6C
	.dc.b	$63,$5F,$1E,$BE,$2B,$E6,$7A,$2D,$3A,$C9,$F9,$A7,$EA,$B0,$FF,$63
	.dc.b	$7E,$75,$FC,$07,$E0,$DF,$AB,$3F,$11,$FF,$5A,$60,$F8,$4E,$BA,$30
	.dc.b	$E1,$7B,$FA,$EF,$F7,$67,$E5,$9A,$9F,$E1,$9F,$A9,$7A,$2C,$71,$F5
	.dc.b	$FF,$9D,$3E,$F3,$33,$D3,$3F,$0C,$73,$9F,$E8,$18,$F7,$DB,$D5,$3C
	.dc.b	$3F,$0E,$FF,$64,$FA,$7F,$EF,$DF,$CC,$5F,$B4,$3F,$7C,$F2,$56,$B8
	.dc.b	$55,$AF,$A4,$CF,$ED,$BE,$7C,$BD,$D7,$FE,$D4,$7C,$C6,$3F,$14,$FE
	.dc.b	$03,$87,$2F,$D1,$BC,$73,$9D,$3C,$5A,$FF,$18,$FB,$4F,$F7,$A7,$EB
	.dc.b	$FF,$E6,$CF,$E0,$17,$EA,$1F,$CC,$3E,$8D,$F9,$2F,$FB,$93,$C5,$FF
	.dc.b	$3C,$5E,$EB,$52,$FC,$07,$F1,$F8,$F4,$F5,$3D,$DA,$FF,$16,$8C,$67
	.dc.b	$FC,$51,$B8,$77,$F8,$FF,$25,$7D,$4C,$BB,$7C,$53,$5F,$E9,$7A,$F8
	.dc.b	$BF,$60,$7D,$6B,$F3,$6F,$EF,$3F,$A6,$FF,$B8,$72,$F1,$9C,$EB,$73
	.dc.b	$AF,$C6,$5F,$84,$6A,$9A,$2F,$AC,$7A,$4E,$5F,$AB,$FA,$5E,$99,$78
	.dc.b	$7E,$81,$9F,$F2,$19,$E9,$87,$3A,$D4,$BE,$4B,$DF,$5F,$C3,$FE,$43
	.dc.b	$FC,$DF,$FA,$F3,$F1,$F7,$EB,$6F,$D9,$7F,$D6,$BC,$5F,$EA,$9F,$82
	.dc.b	$61,$C2,$F7,$4F,$D7,$7D,$95,$AF,$B9,$F4,$67,$DF,$F5,$7B,$39,$D3
	.dc.b	$C1,$A7,$F4,$1D,$3B,$69,$D6,$3F,$D6,$18,$7F,$89,$CF,$D4,$FC,$BB
	.dc.b	$F6,$AF,$91,$F8,$77,$D3,$FC,$1B,$FD,$43,$F8,$76,$EF,$FC,$2D,$5A
	.dc.b	$EB,$FF,$4B,$FE,$4F,$9C,$79,$75,$7F,$58,$FA,$DF,$DA,$E1,$C3,$F1
	.dc.b	$0C,$3E,$0C,$79,$67,$FC,$F2,$B7,$87,$2D,$3C,$5F,$91,$7F,$76,$9C
	.dc.b	$F6,$BF,$72,$7E,$53,$FB,$4F,$DD,$BF,$DE,$9F,$49,$E3,$AD,$4E,$7F
	.dc.b	$1D,$7E,$D7,$9C,$3C,$7B,$FF,$8C,$6D,$FB,$D6,$36,$47,$2F,$69,$A3
	.dc.b	$F1,$0F,$26,$1F,$2A,$B5,$3F,$E5,$9E,$A6,$FE,$CF,$AA,$7E,$0C,$FC
	.dc.b	$CD,$F8,$FF,$E0,$BF,$69,$7E,$0B,$F2,$EF,$75,$AF,$01,$4F,$CD,$F5
	.dc.b	$8F,$9E,$DF,$A7,$D8,$ED,$97,$E3,$BF,$54,$F3,$F0,$7F,$93,$EF,$AC
	.dc.b	$FE,$4B,$E7,$F3,$F0,$DF,$D1,$7F,$27,$7E,$5F,$FC,$17,$CB,$7F,$00
	.dc.b	$7F,$BD,$B9,$3D,$B9,$FF,$32,$C3,$85,$EE,$9E,$37,$E2,$67,$FA,$87
	.dc.b	$E6,$5F,$41,$ED,$7F,$E2,$2E,$6F,$C3,$C7,$2F,$7F,$FD,$B3,$F6,$3E
	.dc.b	$4F,$98,$6D,$E4,$9F,$F6,$EF,$2F,$A3,$F8,$01,$FA,$03,$F5,$87,$EC
	.dc.b	$2F,$C3,$7F,$A8,$7C,$A7,$8E,$F7,$5A,$9F,$EA,$9F,$3D,$9C,$3C,$D4
	.dc.b	$FC,$DA,$C7,$5C,$7A,$7E,$87,$E1,$8C,$60,$FC,$AF,$E5,$F3,$FF,$9F
	.dc.b	$7E,$B9,$FB,$6F,$D9,$FD,$27,$FD,$0D,$CD,$F0,$2B,$53,$9F,$BD,$F9
	.dc.b	$11,$FE,$85,$E8,$F8,$C7,$CF,$7F,$D7,$8F,$EB,$8D,$F9,$9F,$E3,$DA
	.dc.b	$FC,$37,$DE,$78,$A3,$1C,$AD,$FF,$B4,$FC,$1F,$1F,$D9,$1F,$9D,$FF
	.dc.b	$90,$BD,$AF,$EF,$1C,$3F,$E1,$3F,$C1,$3F,$90,$6B,$FB,$CB,$DF,$DE
	.dc.b	$7F,$09,$FA,$1F,$81,$F5,$BF,$E8,$0C,$FA,$67,$FE,$4C,$C3,$CC,$72
	.dc.b	$3F,$BC,$E1,$E0,$FD,$CF,$31,$91,$1D,$B4,$F1,$FE,$8D,$E8,$9F,$15
	.dc.b	$F9,$F2,$FA,$1E,$B6,$67,$AF,$E4,$4F,$D2,$5F,$A1,$A1,$FA,$F3,$F6
	.dc.b	$07,$EA,$DE,$1D,$B5,$87,$0B,$DD,$3F,$4B,$D4,$75,$CB,$DB,$C3,$F2
	.dc.b	$E1,$F9,$77,$B2,$31,$87,$3C,$DE,$B7,$9F,$0F,$7E,$1E,$7D,$7C,$D4
	.dc.b	$FC,$AB,$5A,$BF,$46,$7E,$3B,$FE,$88,$F9,$77,$E9,$2F,$B6,$7A,$BE
	.dc.b	$BB,$DF,$D4,$7F,$5E,$AD,$7C,$46,$76,$F3,$6F,$F7,$6F,$AB,$EC,$31
	.dc.b	$9F,$7F,$C9,$BF,$9E,$F1,$FE,$26,$BA,$EB,$E2,$FC,$5F,$F1,$BF,$BC
	.dc.b	$7F,$42,$C1,$FB,$27,$F6,$4F,$DE,$1F,$AD,$FF,$A2,$FE,$A3,$FD,$71
	.dc.b	$7B,$AD,$4E,$7E,$BF,$FB,$84,$E7,$E7,$3E,$63,$BF,$C4,$AF,$07,$F9
	.dc.b	$13,$3E,$F4,$CF,$BF,$E1,$8D,$3C,$B7,$BD,$3B,$CF,$FB,$83,$81,$9F
	.dc.b	$F5,$AE,$D7,$17,$F8,$6F,$ED,$DF,$8A,$3F,$6C,$FC,$4F,$E2,$BE,$6D
	.dc.b	$3F,$CA,$B8,$70,$FA,$EF,$F2,$2D,$EF,$C6,$7E,$CB,$BB,$FC,$B6,$7F
	.dc.b	$E2,$77,$F1,$7A,$78,$63,$FD,$8B,$18,$3A,$BE,$3B,$F1,$78,$BE,$C9
	.dc.b	$E7,$8F,$83,$47,$E9,$B9,$F7,$9F,$3D,$FD,$33,$C3,$7E,$8E,$FD,$13
	.dc.b	$FA,$9F,$EF,$1E,$5B,$DF,$CB,$FE,$36,$B5,$F3,$4F,$57,$59,$7B,$36
	.dc.b	$F3,$1F,$CD,$1F,$EB,$2E,$30,$E7,$9F,$83,$3F,$F6,$C7,$47,$D0,$AF
	.dc.b	$F3,$ED,$6B,$2F,$27,$97,$FE,$80,$FA,$2F,$F2,$0D,$7F,$BC,$7C,$EF
	.dc.b	$F0,$27,$E9,$2F,$C2,$7D,$BF,$9A,$BF,$5D,$61,$C3,$0F,$F7,$7E,$1C
	.dc.b	$2F,$7F,$94,$EB,$ED,$FF,$C3,$3E,$8F,$F0,$8F,$43,$F8,$AE,$3C,$63
	.dc.b	$F8,$5F,$D2,$E2,$B7,$8F,$3E,$47,$B9,$E7,$9C,$E2,$DF,$CD,$4F,$CB
	.dc.b	$75,$B7,$FA,$73,$37,$E6,$9F,$EB,$7E,$0F,$C7,$7C,$E7,$FB,$53,$EB
	.dc.b	$7C,$D7,$BA,$D7,$7F,$FB,$23,$F1,$6F,$F3,$17,$17,$9E,$7F,$90,$D7
	.dc.b	$F3,$28,$C7,$CD,$7A,$63,$93,$F1,$EF,$FD,$A3,$E1,$79,$25,$A7,$92
	.dc.b	$B5,$9F,$E3,$27,$C6,$7F,$9F,$FF,$22,$FC,$27,$BD,$FE,$F6,$F8,$2F
	.dc.b	$E2,$7E,$46,$B2,$56,$BF,$15,$FF,$63,$7B,$CF,$C6,$75,$2F,$23,$CD
	.dc.b	$9F,$5F,$1A,$E3,$C9,$FE,$57,$8F,$FB,$3F,$F5,$9F,$44,$E7,$5F,$17
	.dc.b	$17,$FB,$23,$F2,$AF,$F0,$CE,$3F,$E1,$CF,$C6,$DF,$AB,$3F,$5F,$FE
	.dc.b	$6B,$E8,$3E,$E5,$F2,$EC,$78,$FD,$3B,$1E,$38,$70,$B7,$CB,$6F,$E6
	.dc.b	$7F,$CC,$77,$BF,$E0,$0F,$EB,$F8,$F2,$C7,$C3,$AF,$F6,$68,$C6,$BE
	.dc.b	$3D,$FF,$87,$FE,$89,$E9,$CB,$9E,$1E,$09,$FE,$27,$FA,$1E,$71,$FF
	.dc.b	$72,$E6,$FC,$D1,$FB,$6F,$F9,$6B,$F1,$1F,$C8,$7E,$4C,$FD,$53,$CD
	.dc.b	$87,$0B,$DF,$DC,$2F,$7F,$77,$AA,$FF,$21,$D7,$47,$D1,$B7,$BA,$1F
	.dc.b	$B3,$FB,$A3,$19,$F8,$7C,$CF,$E3,$DC,$7F,$2A,$FB,$E5,$FA,$3F,$7F
	.dc.b	$D5,$97,$F9,$43,$E2,$5F,$BB,$3F,$17,$FE,$33,$FB,$B7,$E1,$FF,$6B
	.dc.b	$F2,$72,$7F,$3D,$AD,$7A,$6A,$FC,$67,$F0,$0E,$B1,$7E,$DB,$E5,$CF
	.dc.b	$E3,$C7,$67,$D6,$63,$0E,$77,$F7,$E3,$94,$63,$E9,$7F,$19,$F0,$BF
	.dc.b	$EA,$0F,$CC,$5F,$B1,$7F,$9F,$BF,$61,$FF,$48,$FE,$27,$FC,$07,$EF
	.dc.b	$3F,$15,$7B,$AD,$75,$FF,$B6,$BD,$DF,$C8,$74,$7F,$7B,$D4,$33,$EB
	.dc.b	$F7,$FF,$8A,$C7,$97,$ED,$16,$B7,$5A,$D7,$C3,$F3,$7E,$69,$CE,$9E
	.dc.b	$0D,$9F,$39,$99,$ED,$E8,$3E,$01,$FA,$43,$F5,$C7,$EF,$2D,$BF,$DE
	.dc.b	$7C,$3E,$7C,$78,$E1,$C2,$F7,$5F,$DA,$2B,$5B,$B2,$ED,$87,$E8,$78
	.dc.b	$3E,$54,$7F,$81,$A3,$07,$B1,$F2,$67,$E3,$CF,$FC,$E1,$3F,$30,$D6
	.dc.b	$7E,$18,$FF,$30,$47,$E6,$7E,$A7,$F8,$DF,$E0,$7F,$E8,$CF,$81,$7E
	.dc.b	$24,$FD,$AB,$FD,$9B,$B7,$F0,$6B,$DF,$E0,$4B,$DD,$3F,$13,$E9,$F0
	.dc.b	$CF,$90,$FC,$66,$9F,$E4,$D8,$69,$8C,$B9,$E3,$E0,$E2,$FC,$FC,$63
	.dc.b	$57,$E6,$3F,$C5,$FF,$1E,$FF,$75,$F7,$3B,$1F,$83,$7E,$F3,$E7,$1F
	.dc.b	$B1,$3F,$73,$FF,$0C,$F2,$B1,$E3,$87,$0F,$4B,$FD,$97,$7B,$FD,$3F
	.dc.b	$F8,$E5,$6B,$F5,$6F,$E2,$AF,$0C,$EF,$27,$E4,$B5,$F1,$47,$87,$0F
	.dc.b	$EB,$F3,$C8,$D6,$5D,$F5,$F2,$FA,$9F,$8D,$F4,$BF,$DF,$3F,$37,$FE
	.dc.b	$D8,$FC,$17,$EC,$FA,$3F,$67,$1F,$F0,$BF,$E2,$BF,$E6,$1F,$C4,$3F
	.dc.b	$D6,$5E,$77,$E8,$BE,$93,$59,$29,$F9,$EF,$E8,$70,$E3,$9D,$AE,$F8
	.dc.b	$F2,$9F,$15,$F5,$CF,$FD,$71,$F0,$1F,$D1,$7F,$63,$FD,$7D,$7F,$96
	.dc.b	$3F,$54,$7E,$36,$FD,$A9,$F8,$1B,$F2,$7F,$F4,$27,$E2,$BF,$D1,$71
	.dc.b	$E3,$F6,$8C,$78,$E1,$C2,$FF,$D1,$5E,$0E,$F3,$FD,$ED,$EF,$3F,$E8
	.dc.b	$B8,$F2,$C7,$F4,$B5,$FE,$BF,$F0,$13,$9A,$F2,$69,$E3,$E4,$F5,$5F
	.dc.b	$4C,$3C,$33,$E7,$E2,$FE,$C5,$F9,$F7,$9B,$F3,$07,$E7,$BF,$AC,$FD
	.dc.b	$E5,$FA,$23,$E2,$FF,$B0,$5E,$EB,$52,$FD,$93,$3A,$7A,$B1,$7E,$63
	.dc.b	$FD,$7B,$0E,$13,$FD,$75,$E5,$EA,$FF,$0C,$CD,$D1,$E4,$AE,$59,$F8
	.dc.b	$DF,$F6,$BF,$E3,$FF,$53,$E0,$DF,$9B,$7E,$B3,$EB,$DF,$B5,$3F,$36
	.dc.b	$FF,$5E,$7E,$6A,$FB,$76,$3C,$70,$E1,$7B,$F5,$3F,$51,$5A,$CF,$F9
	.dc.b	$7E,$7D,$BD,$9F,$F5,$D6,$3C,$7D,$8B,$5A,$CB,$9E,$9F,$EB,$C8,$F0
	.dc.b	$56,$F5,$F2,$CF,$FA,$EE,$7F,$B5,$7E,$BE,$FD,$17,$FC,$85,$FA,$03
	.dc.b	$F3,$E7,$E6,$8F,$D4,$FF,$DE,$BE,$27,$F8,$63,$0E,$1E,$8B,$0E,$17
	.dc.b	$BF,$CC,$3A,$E5,$FE,$4D,$CF,$FA,$D8,$C3,$C1,$FB,$57,$F1,$67,$9E
	.dc.b	$1C,$BE,$1F,$FB,$8C,$63,$5F,$1E,$7F,$D1,$7E,$7F,$FB,$9F,$D7,$BE
	.dc.b	$DB,$F6,$3F,$66,$FD,$F5,$F8,$3F,$EE,$EF,$75,$A9,$FF,$50,$F1,$7F
	.dc.b	$47,$9F,$EF,$BF,$4B,$FB,$97,$AF,$07,$F4,$03,$FD,$CF,$D7,$FE,$45
	.dc.b	$C3,$BF,$67,$8E,$31,$EC,$B4,$FF,$18,$DF,$EB,$4F,$E6,$3F,$5B,$F2
	.dc.b	$F4,$7E,$49,$FE,$1D,$F3,$3F,$6D,$B7,$EB,$2F,$FB,$7F,$7F,$FA,$FB
	.dc.b	$AF,$D9,$7B,$FD,$97,$A4,$E6,$BF,$C7,$1E,$9F,$F0,$88,$FF,$62,$FE
	.dc.b	$99,$FB,$A9,$E7,$A7,$87,$0F,$F1,$67,$99,$FB,$69,$F2,$E5,$DB,$E9
	.dc.b	$1F,$9F,$7F,$80,$38,$FF,$D9,$F5,$FE,$02,$F6,$DF,$BE,$FF,$BF,$BE
	.dc.b	$49,$FC,$01,$FE,$B9,$FC,$13,$FC,$7F,$87,$0C,$18,$70,$FB,$BF,$EB
	.dc.b	$D7,$BF,$92,$FD,$37,$07,$F2,$0A,$EB,$F8,$0F,$F2,$2C,$3F,$DF,$F3
	.dc.b	$9D,$18,$E6,$7C,$D3,$98,$EB,$F8,$47,$FA,$F2,$31,$AF,$9A,$3F,$45
	.dc.b	$CF,$FE,$72,$FC,$33,$F1,$5F,$A0,$FF,$94,$FC,$FF,$F1,$4F,$EC,$3F
	.dc.b	$EB,$9A,$D7,$82,$AD,$7D,$1A,$9F,$AD,$7F,$95,$71,$FC,$A7,$D4,$FF
	.dc.b	$72,$7A,$EF,$FA,$27,$BF,$D9,$FF,$04,$C3,$FC,$0B,$9F,$6C,$BA,$FD
	.dc.b	$C7,$F8,$8F,$EF,$9F,$93,$BF,$4A,$7E,$9B,$FE,$3E,$F9,$57,$D5,$FF
	.dc.b	$78,$CB,$FC,$45,$87,$0B,$DE,$7F,$EF,$0A,$FA,$75,$AC,$5F,$07,$FD
	.dc.b	$67,$0E,$1B,$FE,$85,$E5,$9F,$7F,$8F,$F2,$5F,$8E,$3C,$5D,$BF,$87
	.dc.b	$FC,$2B,$FA,$CF,$EA,$BF,$E5,$EF,$D0,$9E,$EB,$F3,$17,$E6,$CF,$A3
	.dc.b	$FF,$6A,$78,$DF,$EE,$5C,$38,$5E,$EB,$5F,$BD,$7F,$2B,$CB,$AF,$DB
	.dc.b	$B3,$FC,$17,$0E,$4F,$F2,$BE,$3F,$E6,$EF,$D9,$F3,$9F,$5B,$E1,$9F
	.dc.b	$5F,$8D,$F4,$61,$C2,$73,$B7,$9A,$3F,$CC,$3F,$86,$FF,$86,$DF,$F3
	.dc.b	$1F,$EC,$EF,$BB,$7E,$BF,$FD,$DB,$F3,$D7,$E5,$0F,$DB,$9F,$8D,$32
	.dc.b	$E5,$93,$2E,$58,$F1,$C3,$87,$D1,$EF,$EB,$7F,$C0,$75,$DF,$C1,$F7
	.dc.b	$E3,$CA,$38,$E3,$FA,$38,$C6,$BE,$4F,$A1,$6D,$E4,$EB,$F5,$E5,$D2
	.dc.b	$B5,$3E,$8F,$D2,$BD,$FF,$73,$FE,$12,$F9,$3F,$F6,$AF,$C4,$3F,$33
	.dc.b	$7E,$09,$FC,$1B,$F1,$67,$E0,$7E,$DC,$38,$7E,$18,$C3,$85,$EF,$F4
	.dc.b	$2D,$71,$7F,$A3,$33,$FF,$70,$FC,$EF,$B3,$1E,$38,$5A,$D6,$3C,$F4
	.dc.b	$F0,$FC,$13,$F2,$8F,$3D,$6A,$FB,$FE,$C3,$FE,$2A,$FC,$51,$E3,$FF
	.dc.b	$8B,$3E,$AD,$FB,$93,$F4,$A7,$EB,$7F,$D8,$FF,$43,$FE,$EF,$FB,$E3
	.dc.b	$1E,$3A,$7F,$A8,$B1,$E3,$87,$0F,$A2,$61,$FD,$FB,$F6,$AE,$1A,$77
	.dc.b	$D3,$B6,$3F,$C5,$F2,$E6,$FF,$8D,$63,$EA,$B9,$9E,$18,$F5,$C3,$D5
	.dc.b	$18,$D3,$CD,$E5,$7F,$9D,$6F,$FD,$6B,$FA,$93,$D9,$7C,$F7,$EB,$BF
	.dc.b	$5D,$7E,$8A,$FD,$E5,$D5,$FE,$67,$C5,$8F,$1C,$38,$5E,$FE,$67,$53
	.dc.b	$9A,$FE,$61,$97,$E6,$A3,$C1,$F4,$1F,$E7,$69,$E7,$3E,$2F,$65,$A7
	.dc.b	$8B,$8F,$D1,$5A,$CF,$BF,$E3,$FF,$D3,$E3,$FC,$C3,$FB,$3F,$F8,$A7
	.dc.b	$F6,$7F,$E2,$BF,$89,$3F,$80,$3F,$D3,$DE,$7F,$F8,$B6,$F7,$ED,$2F
	.dc.b	$75,$FE,$A3,$8D,$4F,$FB,$6B,$0F,$F1,$A7,$ED,$FF,$C1,$23,$8D,$D7
	.dc.b	$55,$CF,$E1,$7F,$DA,$B8,$7F,$AA,$A7,$98,$C3,$CD,$3A,$F0,$BF,$DC
	.dc.b	$FC,$2F,$C7,$3F,$CF,$7F,$1D,$FE,$42,$FD,$35,$FA,$13,$D9,$FF,$5A
	.dc.b	$FC,$DB,$F5,$2F,$B5,$87,$0F,$96,$61,$C2,$F7,$FB,$6E,$B7,$FF,$7A
	.dc.b	$E7,$D5,$FF,$17,$BF,$E5,$0F,$C5,$3F,$D8,$6F,$87,$4F,$07,$D3,$7F
	.dc.b	$AF,$AB,$90,$9D,$61,$FE,$C8,$F8,$7F,$F8,$03,$D7,$FD,$EF,$D9,$FF
	.dc.b	$7B,$FE,$0E,$F3,$FF,$AA,$7A,$FF,$47,$C1,$FD,$47,$CE,$FF,$A6,$AF
	.dc.b	$75,$AE,$CF,$E8,$78,$F3,$EF,$FF,$4B,$6D,$FA,$5C,$38,$43,$3F,$E7
	.dc.b	$B5,$B7,$C1,$F2,$6D,$3E,$A4,$E7,$2F,$1F,$83,$E8,$C7,$9E,$0F,$C7
	.dc.b	$DF,$8A,$DF,$3D,$F8,$DF,$A3,$60,$E0,$F7,$DF,$03,$F7,$FF,$9C,$EE
	.dc.b	$75,$C1,$E7,$9D,$61,$E1,$9E,$77,$CB,$1F,$16,$D9,$9E,$39,$FF,$89
	.dc.b	$BD,$57,$E8,$7F,$B8,$FA,$3E,$D8,$FD,$0F,$07,$FC,$F1,$3F,$BC,$FC
	.dc.b	$67,$3B,$7C,$79,$CF,$C3,$F5,$BE,$18,$F4,$F9,$8F,$B7,$A7,$FB,$3B
	.dc.b	$F0,$B7,$E9,$CF,$DA,$3F,$83,$7E,$33,$FC,$73,$AD,$F5,$BF,$C8,$F0
	.dc.b	$E1,$7B,$F6,$7F,$E0,$07,$37,$F8,$57,$F6,$4D,$57,$CA,$CF,$AF,$E9
	.dc.b	$DF,$6A,$38,$E9,$5D,$63,$CF,$4F,$E5,$7D,$BE,$69,$CD,$79,$32,$FB
	.dc.b	$F1,$CB,$F6,$E7,$CC,$BF,$60,$7E,$C3,$EB,$EC,$FF,$25,$6D,$EE,$C9
	.dc.b	$87,$0B,$DF,$EA,$BE,$8C,$BF,$DD,$9B,$7C,$3A,$D7,$B9,$E3,$9C,$E1
	.dc.b	$CB,$4F,$07,$E9,$79,$CF,$B6,$DF,$D3,$7F,$80,$3C,$BF,$8C,$F5,$9C
	.dc.b	$FC,$3A,$7D,$3F,$A3,$67,$6F,$A5,$87,$1F,$8D,$E9,$18,$BE,$1F,$8F
	.dc.b	$FF,$0E,$F6,$3F,$2F,$BB,$E3,$13,$FE,$15,$FD,$07,$FC,$23,$FA,$4B
	.dc.b	$F2,$27,$E8,$DF,$C0,$2F,$7F,$38,$BD,$D3,$F7,$ED,$62,$FD,$57,$FC
	.dc.b	$5F,$F0,$5F,$C7,$74,$FD,$3E,$5C,$9F,$F5,$FB,$D3,$1F,$0F,$E3,$DE
	.dc.b	$79,$DE,$1E,$5F,$A6,$6C,$F6,$F5,$5E,$BA,$FF,$17,$7E,$7F,$E3,$FC
	.dc.b	$03,$FD,$95,$F4,$7E,$4E,$14,$E7,$F0,$EF,$EB,$BB,$F9,$71,$E5,$5A
	.dc.b	$E1,$62,$FC,$7B,$BF,$0F,$C2,$CB,$B7,$E2,$18,$D7,$FC,$7D,$F5,$0F
	.dc.b	$C7,$5F,$A4,$7F,$AE,$BE,$97,$C5,$39,$8F,$F5,$EF,$DD,$E2,$B7,$F8
	.dc.b	$8F,$3D,$FF,$C8,$11,$CB,$F4,$1F,$EC,$7F,$9D,$BF,$42,$7E,$D4,$FC
	.dc.b	$1B,$FC,$CF,$F8,$93,$F4,$EF,$E4,$1D,$1F,$E4,$AC,$38,$5E,$FE,$3B
	.dc.b	$FD,$A5,$4F,$CF,$3F,$BF,$56,$BA,$BF,$BC,$F6,$7E,$22,$3C,$1F,$B1
	.dc.b	$FF,$79,$79,$E9,$E0,$FA,$EF,$2E,$DE,$48,$F1,$F3,$E3,$57,$ED,$2F
	.dc.b	$D4,$FF,$C2,$3C,$AF,$DA,$3C,$38,$FB,$2B,$5F,$4A,$A7,$DC,$6B,$1E
	.dc.b	$78,$F2,$E0,$C6,$1C,$3B,$7C,$1F,$66,$FD,$D7,$F3,$FF,$5D,$FE,$B5
	.dc.b	$E8,$F8,$9B,$7C,$AA,$D7,$02,$BF,$AF,$56,$B4,$56,$A7,$AE,$1F,$AA
	.dc.b	$E6,$FF,$0B,$FE,$59,$FA,$3F,$C8,$7F,$C8,$9F,$C0,$0F,$C3,$1F,$A5
	.dc.b	$BF,$57,$FF,$16,$5E,$FE,$15,$7B,$AF,$F7,$3E,$DF,$E8,$2A,$FD,$77
	.dc.b	$E8,$5F,$E2,$97,$B7,$27,$D6,$3F,$A9,$BC,$EF,$C1,$E1,$79,$3C,$D6
	.dc.b	$BE,$4D,$7E,$66,$1E,$29,$CF,$CF,$7F,$97,$DF,$F0,$C7,$A3,$9B,$FE
	.dc.b	$F5,$EC,$FE,$FF,$EB,$1F,$AB,$67,$0E,$1F,$29,$DE,$F7,$8B,$E5,$71
	.dc.b	$F1,$2F,$D0,$5F,$87,$BE,$CB,$FD,$45,$CB,$EF,$C3,$CF,$7B,$AD,$7E
	.dc.b	$89,$E3,$8C,$7D,$4F,$4C,$79,$FC,$97,$F8,$B3,$ED,$FF,$8F,$F8,$4F
	.dc.b	$AF,$7E,$74,$F9,$77,$53,$F2,$2F,$F2,$47,$CE,$FF,$99,$71,$E3,$87
	.dc.b	$0B,$DD,$7F,$93,$7F,$13,$D7,$67,$F8,$0B,$3E,$D9,$FF,$20,$C7,$8F
	.dc.b	$EA,$96,$B5,$97,$3D,$3C,$1F,$81,$F9,$E7,$38,$79,$67,$C9,$F4,$79
	.dc.b	$9F,$E6,$F7,$EA,$FC,$CD,$F5,$6F,$97,$7E,$D4,$F8,$DF,$2E,$1C,$33
	.dc.b	$61,$C2,$F7,$FA,$9E,$B8,$3D,$33,$9F,$33,$C1,$18,$CF,$B7,$E7,$58
	.dc.b	$D7,$E5,$FC,$CF,$F9,$67,$F3,$87,$E5,$FF,$E2,$8F,$05,$CD,$F2,$B7
	.dc.b	$FE,$B5,$F2,$EB,$DF,$B1,$E1,$CB,$9C,$E7,$DC,$ED,$7C,$2D,$7F,$EE
	.dc.b	$EF,$C3,$7F,$AD,$F5,$BA,$DF,$97,$FF,$51,$F9,$97,$E3,$5F,$C9,$FE
	.dc.b	$DB,$FD,$D9,$87,$0B,$DF,$CD,$7F,$5E,$AF,$E5,$9E,$D6,$BB,$3F,$C7
	.dc.b	$B9,$F5,$CF,$FD,$D7,$8F,$1F,$BD,$B5,$AC,$B9,$E9,$E0,$FB,$6F,$34
	.dc.b	$E7,$0F,$26,$6F,$07,$32,$CF,$B7,$E3,$CF,$D5,$DF,$9D,$3F,$26,$7C
	.dc.b	$AF,$97,$0E,$1B,$7C,$0C,$38,$5E,$FE,$63,$59,$74,$CB,$9F,$9B,$E0
	.dc.b	$8C,$57,$F8,$DF,$F0,$77,$1F,$C5,$E0,$FF,$1D,$FD,$5B,$F4,$CF,$F0
	.dc.b	$1E,$5D,$35,$F9,$FF,$66,$AD,$72,$BF,$44,$D6,$3C,$B1,$78,$BE,$1C
	.dc.b	$78,$E1,$D4,$FF,$7A,$F1,$38,$BF,$C2,$5B,$3C,$FF,$E6,$3F,$10,$FD
	.dc.b	$DF,$C1,$3F,$ED,$6B,$FF,$39,$DF,$0B,$DD,$7F,$8F,$AB,$55,$D7,$E9
	.dc.b	$D1,$FE,$24,$8F,$F1,$2B,$AB,$FF,$11,$7E,$15,$99,$F2,$C7,$5F,$8D
	.dc.b	$FE,$99,$FA,$4E,$30,$E7,$97,$9B,$F2,$DC,$65,$FB,$0F,$D5,$FF,$98
	.dc.b	$FE,$5B,$FE,$30,$8E,$D7,$E2,$F4,$D5,$AF,$27,$CB,$F8,$97,$F8,$E3
	.dc.b	$5F,$87,$5A,$F8,$5F,$E9,$5B,$E8,$F8,$38,$F3,$C7,$F9,$FF,$C6,$78
	.dc.b	$BE,$DB,$FB,$87,$2B,$F6,$1F,$F4,$EE,$DE,$21,$3A,$87,$E9,$BD,$72
	.dc.b	$F9,$15,$AF,$99,$FC,$07,$D5,$F5,$7F,$05,$E8,$7F,$A9,$25,$F4,$5F
	.dc.b	$E5,$0D,$9F,$99,$7F,$83,$FF,$37,$FF,$54,$7A,$4F,$CE,$BC,$98,$70
	.dc.b	$D3,$FD,$5B,$87,$0F,$4D,$87,$0B,$7E,$B7,$F6,$B3,$ED,$9B,$F6,$BF
	.dc.b	$B1,$96,$4E,$71,$FE,$1F,$7E,$E6,$7D,$DF,$F3,$04,$79,$7E,$4B,$FC
	.dc.b	$B5,$E2,$FA,$F0,$EB,$A7,$8A,$7F,$CB,$DF,$05,$9C,$3A,$CB,$F1,$47
	.dc.b	$E8,$6F,$CE,$7E,$5F,$D6,$F8,$EF,$BD,$EE,$B5,$CF,$FD,$8F,$6F,$27
	.dc.b	$1F,$C1,$CB,$9E,$5F,$C8,$B1,$E5,$39,$C3,$B7,$3F,$8E,$7A,$FE,$43
	.dc.b	$FB,$0F,$DE,$7F,$C6,$1E,$5F,$0E,$3F,$F2,$04,$E7,$A9,$59,$DD,$3C
	.dc.b	$78,$BA,$69,$F4,$E3,$A7,$DE,$FE,$E3,$E7,$1F,$BD,$7F,$6E,$FC,$83
	.dc.b	$FD,$CF,$F9,$93,$F6,$AF,$E7,$3C,$BF,$B2,$BD,$F9,$4B,$DD,$7E,$1A
	.dc.b	$B5,$87,$F7,$8E,$2F,$AF,$E5,$7F,$A6,$B1,$EF,$E9,$7D,$EB,$DE,$5C
	.dc.b	$FE,$09,$A3,$D0,$F3,$67,$E5,$9C,$F9,$F8,$BF,$F0,$37,$F0,$0B,$F5
	.dc.b	$CF,$E2,$1F,$1B,$FA,$1D,$7C,$FF,$16,$AD,$7A,$8F,$C7,$B5,$A7,$F8
	.dc.b	$3A,$B5,$8F,$F1,$9E,$2C,$E7,$DA,$3F,$A1,$46,$3F,$34,$FF,$32,$7D
	.dc.b	$DB,$F3,$47,$EC,$F9,$EF,$FF,$45,$C7,$87,$97,$DF,$7C,$38,$FD,$4F
	.dc.b	$2F,$8A,$FC,$17,$E7,$6F,$C2,$1F,$B0,$7F,$B2,$FE,$FF,$FC,$D9,$E3
	.dc.b	$BF,$22,$FC,$A5,$EE,$B5,$F2,$9F,$EA,$1F,$92,$CE,$3C,$B8,$BF,$53
	.dc.b	$E8,$FF,$B3,$30,$ED,$F8,$AF,$D3,$8C,$57,$83,$E6,$7F,$D4,$98,$FF
	.dc.b	$AA,$F0,$E7,$97,$6F,$9E,$F3,$DF,$9B,$F2,$A7,$CF,$3C,$0F,$F4,$B7
	.dc.b	$C4,$70,$9C,$C7,$FA,$67,$F8,$07,$15,$BF,$3F,$9E,$4B,$DF,$8D,$8F
	.dc.b	$DF,$1E,$53,$F0,$1F,$F6,$67,$07,$92,$73,$1F,$EC,$FF,$23,$15,$BF
	.dc.b	$C2,$F9,$3C,$7E,$A3,$F9,$07,$CF,$FF,$BB,$BE,$CB,$C1,$F0,$CF,$9B
	.dc.b	$7E,$7E,$FA,$D7,$EE,$DF,$E8,$4F,$1B,$FD,$A5,$87,$0B,$DD,$6B,$F1
	.dc.b	$8F,$EA,$F9,$75,$E3,$FE,$43,$D9,$FC,$6A,$31,$A7,$E8,$DB,$87,$2F
	.dc.b	$C5,$3C,$B4,$3F,$CF,$37,$FD,$0E,$7B,$FC,$76,$A7,$38,$7F,$84,$3C
	.dc.b	$DF,$F0,$C7,$E4,$CF,$C8,$7C,$75,$CF,$6F,$7F,$D1,$FF,$B8,$3C,$B5
	.dc.b	$BD,$DD,$DF,$4A,$D4,$77,$BE,$38,$FF,$89,$7E,$03,$1B,$7F,$57,$C5
	.dc.b	$FA,$C3,$F3,$97,$99,$FC,$27,$9B,$E4,$DE,$EB,$53,$CB,$83,$C7,$1C
	.dc.b	$7F,$02,$C6,$3D,$31,$FE,$89,$8B,$97,$E3,$70,$3F,$5E,$7E,$16,$FD
	.dc.b	$6D,$F8,$F7,$F8,$A3,$B7,$FA,$D7,$F0,$07,$BA,$F7,$5A,$FA,$8F,$C0
	.dc.b	$FC,$B7,$E7,$34,$ED,$9F,$D3,$BD,$BF,$6C,$C6,$7F,$86,$70,$78,$DF
	.dc.b	$16,$7E,$1F,$94,$FB,$15,$E0,$8F,$27,$CA,$62,$BD,$19,$BF,$77,$7A
	.dc.b	$CF,$D3,$7F,$C7,$1A,$79,$34,$ED,$7E,$FB,$DD,$7F,$95,$6B,$59,$FB
	.dc.b	$AB,$A6,$0F,$BC,$F0,$47,$18,$FE,$C9,$8F,$C5,$CB,$E5,$7E,$18,$FC
	.dc.b	$11,$E3,$3F,$45,$7E,$B3,$BC,$97,$BA,$D7,$E1,$D9,$8C,$70,$F8,$1D
	.dc.b	$7D,$1F,$E7,$FD,$AF,$E9,$DA,$FE,$87,$E7,$1F,$AF,$FF,$85,$38,$3F
	.dc.b	$C1,$7F,$1F,$FE,$9A,$C5,$87,$0C,$3F,$C4,$77,$BF,$A0,$FE,$8B,$F9
	.dc.b	$AE,$BE,$5F,$F0,$F9,$F4,$FA,$4F,$F0,$8E,$3C,$63,$BF,$CE,$62,$B7
	.dc.b	$8F,$3D,$3F,$DE,$77,$B8,$C6,$1E,$9A,$DD,$BF,$30,$FE,$15,$FA,$6B
	.dc.b	$F4,$27,$D3,$F8,$F7,$F7,$65,$FE,$30,$BD,$D3,$F1,$BD,$57,$B3,$1E
	.dc.b	$9E,$CB,$2F,$F7,$C4,$EB,$7F,$4E,$3E,$0F,$43,$C9,$3D,$B0,$FF,$7D
	.dc.b	$7E,$70,$FC,$59,$F9,$73,$F7,$7E,$5B,$FF,$90,$BF,$0F,$5E,$E9,$FA
	.dc.b	$66,$B1,$E7,$8B,$EF,$FB,$E7,$E7,$C7,$8C,$BF,$18,$FE,$38,$3E,$63
	.dc.b	$F7,$3D,$BF,$E9,$3F,$57,$FC,$67,$FB,$D3,$F1,$17,$EE,$4F,$C6,$B9
	.dc.b	$61,$C3,$EF,$D8,$70,$BD,$F8,$FA,$DB,$F9,$56,$7D,$5F,$F4,$36,$9E
	.dc.b	$17,$38,$7F,$B5,$1C,$E3,$D7,$1E,$9F,$2D,$E8,$8C,$7D,$83,$6F,$37
	.dc.b	$E1,$99,$C3,$9F,$E3,$0F,$C7,$5F,$B9,$3E,$1B,$FD,$51,$F2,$DE,$5B
	.dc.b	$FD,$EE,$CC,$38,$5E,$FE,$FB,$FC,$7F,$C1,$E7,$BD,$D7,$F9,$A2,$31
	.dc.b	$C3,$FE,$BB,$8E,$39,$3E,$77,$C9,$7C,$34,$F1,$CF,$F7,$DC,$38,$E7
	.dc.b	$E1,$FC,$B1,$F9,$9F,$F8,$77,$E9,$DF,$8E,$79,$F5,$ED,$5A,$9F,$F1
	.dc.b	$16,$39,$C3,$96,$1F,$E2,$5F,$A8,$F0,$E1,$C2,$3F,$C5,$1F,$23,$F8
	.dc.b	$2D,$7F,$9C,$F1,$BE,$A7,$FD,$27,$FA,$83,$F6,$27,$E2,$6F,$A3,$7C
	.dc.b	$F7,$E2,$F1,$E3,$87,$0B,$DD,$3F,$6B,$D5,$7F,$6B,$CB,$B7,$C7,$FC
	.dc.b	$98,$E3,$F5,$35,$D6,$3C,$F3,$FF,$49,$73,$7F,$B4,$70,$F3,$5B,$5F
	.dc.b	$35,$7F,$96,$70,$EE,$2B,$7F,$A1,$FF,$62,$FC,$6B,$F0,$7E,$76,$A7
	.dc.b	$3F,$2E,$9F,$C3,$FE,$07,$9D,$7F,$D6,$B8,$71,$FA,$2F,$91,$97,$9B
	.dc.b	$54,$F5,$E3,$CC,$F8,$6F,$FD,$7B,$F3,$8F,$AC,$FF,$66,$7E,$82,$FD
	.dc.b	$BB,$B6,$5F,$E0,$19,$CF,$E8,$98,$BE,$1F,$3D,$FE,$65,$E5,$F8,$59
	.dc.b	$74,$BF,$F6,$B7,$57,$F8,$5F,$97,$FA,$A7,$E5,$FF,$E3,$7F,$45,$F8
	.dc.b	$FF,$F1,$BE,$B7,$EA,$5F,$9F,$BD,$FD,$67,$EF,$6B,$52,$FD,$AB,$F0
	.dc.b	$3D,$DF,$DF,$B2,$E9,$CB,$FE,$FF,$C7,$8F,$DF,$7F,$B2,$9E,$7A,$78
	.dc.b	$3E,$17,$F0,$5F,$13,$E5,$C7,$96,$BD,$E7,$F0,$DF,$6F,$9C,$BF,$C6
	.dc.b	$19,$FA,$7F,$38,$7E,$57,$F6,$E5,$F9,$27,$91,$DE,$3F,$C5,$6B,$7E
	.dc.b	$1F,$F8,$E7,$CD,$D5,$31,$E9,$8F,$38,$CC,$62,$BF,$8D,$C6,$30,$FE
	.dc.b	$41,$F8,$73,$F7,$47,$EF,$8F,$D3,$3F,$B7,$F1,$FC,$0A,$D4,$E7,$E5
	.dc.b	$39,$72,$27,$3C,$9E,$FC,$FC,$99,$F6,$CF,$AF,$DC,$67,$BD,$EA,$3F
	.dc.b	$26,$FF,$50,$7E,$94,$FC,$89,$7F,$E3,$FE,$0F,$F0,$A6,$1C,$2F,$75
	.dc.b	$AF,$D2,$BE,$1C,$E7,$4F,$B5,$D7,$FE,$3A,$8E,$FA,$7D,$B8,$C6,$9E
	.dc.b	$1E,$76,$BE,$1D,$7C,$D3,$98,$F2,$6F,$8D,$5F,$3C,$FD,$3F,$FD,$B5
	.dc.b	$F4,$2F,$C4,$3C,$78,$FB,$6B,$5F,$37,$FE,$1C,$FD,$53,$38,$F2,$C7
	.dc.b	$FD,$C5,$E3,$F5,$BD,$FB,$97,$CF,$7E,$B9,$FC,$67,$6E,$35,$FA,$F8
	.dc.b	$FE,$39,$F2,$58,$AD,$F8,$1C,$67,$FC,$B3,$CE,$FA,$A7,$A5,$FE,$5E
	.dc.b	$F9,$37,$EB,$0F,$CB,$1C,$5F,$E3,$5F,$AA,$7E,$7B,$F9,$6C,$78,$E5
	.dc.b	$FE,$8B,$C7,$8E,$1C,$3E,$AF,$FC,$CF,$7F,$ED,$BB,$DD,$FF,$7B,$CB
	.dc.b	$BB,$FE,$CA,$39,$3E,$17,$06,$EF,$A6,$3F,$CD,$A3,$D5,$9F,$A6,$BD
	.dc.b	$19,$F9,$FE,$2B,$FD,$3D,$8F,$FA,$FA,$FF,$D4,$1F,$3A,$FD,$C5,$FA
	.dc.b	$DF,$F6,$5F,$C0,$DD,$6F,$17,$DF,$8F,$1C,$58,$F1,$C3,$85,$EF,$F4
	.dc.b	$DF,$99,$8F,$FA,$22,$B5,$7F,$E8,$A3,$C7,$19,$CB,$FC,$39,$19,$38
	.dc.b	$9E,$67,$CB,$F2,$1F,$ED,$6C,$3F,$C6,$79,$78,$F0,$F1,$4E,$7E,$0F
	.dc.b	$FC,$43,$5F,$E9,$8F,$CF,$DF,$A1,$3F,$5F,$7D,$D7,$FA,$CF,$CA,$7E
	.dc.b	$9A,$FC,$6F,$FC,$55,$8F,$1C,$38,$5E,$E9,$FB,$CE,$B8,$3F,$D9,$59
	.dc.b	$F6,$FC,$D3,$FC,$11,$8F,$1D,$2D,$6B,$2E,$7F,$0F,$FE,$AD,$E0,$F3
	.dc.b	$CE,$6B,$CB,$3F,$EB,$1F,$CE,$7F,$DA,$5F,$AC,$BF,$7D,$FC,$A7,$C9
	.dc.b	$7F,$8F,$FF,$41,$7E,$66,$FC,$0F,$FD,$9B,$7B,$F0,$FF,$C9,$BF,$1C
	.dc.b	$BD,$FE,$F5,$AC,$BF,$23,$B7,$D5,$EB,$FF,$50,$63,$C6,$3F,$D5,$EF
	.dc.b	$3D,$3C,$1F,$87,$79,$2F,$C7,$97,$5F,$C1,$7E,$EE,$1F,$EE,$DF,$DE
	.dc.b	$1F,$89,$FF,$65,$F3,$7C,$FF,$82,$BF,$1F,$7D,$FF,$E0,$30,$61,$C3
	.dc.b	$E5,$7F,$7F,$CB,$FE,$B5,$B7,$C1,$6E,$B5,$7F,$89,$E2,$FC,$B4,$78
	.dc.b	$3F,$0E,$FC,$64,$EF,$5F,$16,$3F,$EB,$F8,$F3,$9E,$08,$CE,$9F,$D1
	.dc.b	$B7,$FF,$50,$7E,$46,$FC,$71,$FA,$73,$F4,$F7,$EE,$4F,$D7,$FF,$C5
	.dc.b	$7F,$A7,$7E,$17,$46,$3C,$70,$E1,$7B,$F8,$C7,$EC,$9B,$AD,$5F,$F9
	.dc.b	$AB,$C6,$FF,$78,$C7,$87,$F2,$7F,$C5,$CF,$4D,$7C,$7A,$F8,$B5,$FF
	.dc.b	$1C,$6D,$E8,$9C,$E9,$E5,$F1,$3F,$C1,$5E,$2F,$FB,$F3,$F2,$B7,$DC
	.dc.b	$7A,$BE,$F5,$F6,$CF,$27,$DD,$FA,$27,$F9,$1B,$0E,$17,$BA,$D6,$CF
	.dc.b	$CE,$F5,$C0,$19,$F6,$F0,$BF,$C1,$58,$F2,$DA,$D6,$EB,$C7,$1F,$EA
	.dc.b	$9C,$7C,$5F,$13,$E8,$AD,$67,$E1,$D7,$FD,$AB,$1E,$03,$BC,$7A,$FE
	.dc.b	$F3,$F2,$3C,$6F,$C7,$B5,$9F,$6E,$1F,$F4,$EF,$6F,$D9,$F8,$FC,$E3
	.dc.b	$C6,$3F,$DF,$59,$63,$6F,$F5,$E6,$7F,$9C,$FC,$C7,$C9,$3E,$09,$FF
	.dc.b	$4A,$5B,$F4,$B7,$EA,$EF,$D0,$B9,$7D,$E7,$D0,$9C,$FD,$B6,$2F,$87
	.dc.b	$E8,$BD,$38,$3E,$1D,$EF,$F2,$4F,$E7,$7E,$D7,$CD,$D7,$FA,$9F,$D0
	.dc.b	$BF,$56,$74,$7F,$96,$3E,$41,$FA,$E3,$F1,$9F,$EC,$18,$F1,$F9,$2F
	.dc.b	$F6,$06,$1C,$3E,$DD,$87,$0F,$C9,$F7,$A7,$DB,$BD,$BF,$AF,$C3,$96
	.dc.b	$8F,$C4,$7F,$5F,$EA,$27,$39,$F5,$CB,$F9,$CC,$63,$6F,$37,$C5,$7B
	.dc.b	$32,$ED,$97,$AE,$5E,$5B,$F2,$D7,$E5,$AF,$DA,$3C,$B8,$7C,$2D,$3E
	.dc.b	$0D,$EE,$9F,$BF,$6B,$83,$DF,$97,$4C,$BF,$AD,$E3,$9C,$FC,$18,$F7
	.dc.b	$F1,$BC,$93,$D6,$DF,$82,$3F,$52,$FF,$42,$FB,$5C,$76,$FA,$33,$9E
	.dc.b	$6F,$F4,$27,$27,$C3,$C7,$DB,$5A,$C1,$E8,$76,$DF,$CB,$8F,$5F,$4B
	.dc.b	$F8,$AF,$ED,$0F,$C0,$7F,$CB,$DF,$6C,$FA,$C7,$EF,$4F,$CE,$1F,$59
	.dc.b	$FC,$AF,$0E,$1F,$6D,$FE,$6C,$BD,$D7,$F8,$E3,$87,$FB,$E5,$6B,$87
	.dc.b	$F9,$57,$CD,$FE,$36,$3C,$1F,$65,$F7,$23,$1A,$F8,$B1,$E7,$7D,$E3
	.dc.b	$1A,$3F,$1F,$F4,$C7,$A3,$F2,$6F,$F6,$FF,$F0,$0B,$E1,$B8,$6F,$FE
	.dc.b	$80,$9C,$F3,$63,$89,$1F,$EE,$6F,$6B,$9D,$6F,$F0,$EF,$F6,$C6,$BB
	.dc.b	$E2,$7E,$A2,$FD,$21,$E8,$66,$31,$FA,$CD,$F6,$79,$EB,$FE,$43,$9E
	.dc.b	$78,$7F,$8D,$BA,$9F,$BD,$BF,$7B,$FD,$E7,$E3,$5F,$9F,$FD,$59,$E6
	.dc.b	$FF,$8C,$B7,$FE,$41,$87,$0B,$DF,$A0,$BF,$F8,$03,$97,$5C,$BA,$61
	.dc.b	$FE,$3E,$C3,$BF,$EA,$DF,$AA,$FA,$FF,$DF,$47,$3F,$6D,$AF,$EB,$FD
	.dc.b	$7F,$F1,$ED,$6B,$1F,$35,$6B,$4F,$83,$97,$94,$63,$CF,$EA,$5F,$B7
	.dc.b	$BF,$80,$1F,$B6,$EE,$B5,$39,$E7,$C6,$7D,$B0,$E3,$EF,$78,$B2,$F8
	.dc.b	$33,$9F,$BB,$FE,$2D,$C5,$EF,$E4,$FE,$49,$FA,$D3,$F7,$87,$E9,$8F
	.dc.b	$D0,$7A,$D6,$A7,$3F,$A5,$63,$91,$87,$1F,$C3,$7C,$1A,$FB,$35,$F8
	.dc.b	$33,$D7,$ED,$FF,$CC,$3F,$67,$FC,$9F,$F1,$C7,$EF,$7D,$FD,$BF,$DD
	.dc.b	$7E,$08,$FD,$03,$FC,$BF,$7B,$AD,$79,$9F,$EB,$59,$FF,$41,$F2,$2B
	.dc.b	$5C,$7F,$C8,$1E,$1C,$37,$7F,$EC,$5A,$78,$2B,$F9,$AE,$3F,$EB,$FF
	.dc.b	$6B,$CB,$5A,$CF,$BE,$9F,$5A,$3F,$D2,$D9,$76,$D5,$EB,$3E,$87,$D7
	.dc.b	$F9,$EE,$AF,$96,$9F,$D0,$AB,$59,$7F,$0E,$DF,$D1,$5A,$C7,$9E,$2F
	.dc.b	$CD,$BC,$72,$DF,$D1,$5A,$FA,$AC,$C6,$3D,$57,$D4,$BF,$24,$7E,$85
	.dc.b	$E3,$C7,$96,$BF,$3E,$F7,$4F,$C0,$F5,$8F,$3C,$7F,$57,$86,$30,$E1
	.dc.b	$1F,$E9,$FF,$CD,$7F,$D1,$5F,$92,$FD,$7F,$D8,$5F,$9D,$3E,$21,$FB
	.dc.b	$53,$F7,$6F,$B9,$87,$0E,$E6,$1C,$2F,$75,$FC,$CA,$B5,$FA,$BF,$F9
	.dc.b	$21,$E1,$57,$5F,$EE,$19,$F8,$73,$F0,$67,$FC,$4A,$F7,$18,$F6,$7F
	.dc.b	$C4,$51,$8F,$CA,$3F,$C9,$3F,$2C,$FD,$1B,$FC,$33,$97,$F8,$9F,$2E
	.dc.b	$95,$AF,$85,$CC,$A7,$3F,$62,$FC,$47,$AC,$F9,$38,$BD,$BA,$B1,$FF
	.dc.b	$2B,$FE,$2B,$AA,$ED,$5A,$FD,$79,$F0,$3F,$C7,$7F,$6C,$EF,$A7,$B6
	.dc.b	$73,$E8,$BF,$0B,$CE,$1C,$70,$FF,$75,$F1,$7C,$0B,$E7,$F8,$8F,$C0
	.dc.b	$FB,$5F,$F0,$B7,$EE,$EF,$DD,$7F,$D9,$FE,$0F,$F5,$BF,$41,$F9,$73
	.dc.b	$F3,$66,$7F,$ED,$2E,$2F,$F2,$56,$3C,$7F,$11,$FC,$77,$8C,$C7,$8F
	.dc.b	$EA,$9C,$35,$EF,$AF,$FB,$33,$1F,$F6,$86,$5C,$D6,$5C,$DF,$F6,$84
	.dc.b	$66,$3B,$6B,$E5,$DB,$D5,$5E,$2B,$F4,$E7,$E1,$FC,$5B,$33,$D2,$7D
	.dc.b	$BF,$2A,$FD,$9D,$F2,$CF,$D2,$3C,$D7,$CF,$87,$F5,$17,$BA,$7E,$63
	.dc.b	$AA,$E9,$5A,$C3,$FC,$EB,$19,$DB,$CF,$87,$7F,$AF,$F1,$C7,$AB,$F0
	.dc.b	$07,$EE,$DF,$E9,$7E,$FE,$1B,$7D,$29,$CF,$DF,$7F,$AD,$B9,$3E,$14
	.dc.b	$E6,$FF,$C9,$79,$F0,$D7,$F7,$D8,$74,$BF,$EB,$FF,$B5,$7F,$94,$F5
	.dc.b	$FF,$1A,$61,$FD,$3B,$E5,$5F,$93,$BF,$20,$7E,$B9,$FD,$4B,$0E,$1F
	.dc.b	$61,$E6,$F9,$1F,$ED,$77,$BF,$B3,$D6,$FF,$E8,$BD,$3F,$C8,$1E,$DF
	.dc.b	$FA,$F2,$D8,$F2,$D7,$0D,$8E,$7A,$79,$B0,$E5,$9F,$7D,$7C,$BE,$E6
	.dc.b	$67,$FC,$57,$3E,$3F,$A9,$7E,$D8,$FC,$E3,$FB,$FE,$5E,$0B,$E7,$A7
	.dc.b	$F6,$CB,$DD,$7F,$29,$AD,$57,$4A,$D5,$BE,$2B,$C1,$BF,$C8,$B6,$3E
	.dc.b	$0F,$8D,$CE,$7F,$17,$F8,$01,$FA,$FF,$F8,$4F,$EE,$7E,$2E,$BD,$71
	.dc.b	$56,$BD,$07,$EF,$FA,$D3,$B4,$F2,$F1,$7A,$F0,$7F,$81,$23,$AF,$8F
	.dc.b	$FE,$1A,$CB,$FD,$1D,$F8,$5F,$FB,$13,$9F,$FC,$45,$EC,$3F,$57,$FB
	.dc.b	$9E,$7F,$F5,$7B,$FF,$60,$5E,$F6,$FF,$73,$D7,$E0,$7F,$46,$D5,$2B
	.dc.b	$FD,$9F,$F8,$8F,$DB,$8E,$FF,$57,$F2,$1F,$F1,$09,$8C,$FC,$38,$7E
	.dc.b	$2E,$39,$DF,$7C,$FC,$9D,$7E,$99,$F1,$E1,$F8,$6C,$FC,$1F,$99,$7F
	.dc.b	$57,$FD,$07,$8C,$E7,$14,$FF,$5B,$FB,$AC,$E9,$FE,$41,$9C,$F7,$F5
	.dc.b	$CB,$D5,$87,$1C,$1F,$9E,$66,$FE,$17,$E9,$0F,$D6,$9F,$9F,$FF,$BB
	.dc.b	$BA,$FF,$D7,$59,$72,$D3,$FC,$9D,$7B,$A7,$EB,$7E,$2E,$04,$F3,$F8
	.dc.b	$8E,$DB,$7B,$63,$B7,$A5,$F4,$BE,$AB,$FD,$8D,$FB,$2B,$F4,$47,$E6
	.dc.b	$BF,$ED,$5F,$1B,$FD,$ED,$FA,$B7,$F6,$7E,$36,$1C,$2F,$7F,$48,$BD
	.dc.b	$FE,$E1,$F6,$6B,$58,$FA,$B8,$FC,$07,$FC,$01,$CF,$C1,$F6,$39,$8E
	.dc.b	$06,$72,$EB,$8F,$F5,$AC,$FC,$33,$9C,$3C,$CF,$F1,$28,$CA,$C3,$D9
	.dc.b	$3F,$EE,$FF,$C7,$71,$ED,$BF,$3E,$EF,$7B,$CB,$D7,$5A,$FD,$1F,$38
	.dc.b	$F2,$DD,$3D,$7F,$29,$F5,$E1,$C7,$3E,$D7,$FE,$B8,$8E,$38,$FC,$6F
	.dc.b	$DA,$5C,$5F,$EA,$1E,$1C,$E0,$CD,$5A,$F9,$3F,$F2,$57,$07,$B3,$3F
	.dc.b	$14,$E7,$3E,$37,$C3,$2F,$D0,$5B,$F0,$15,$4F,$05,$FB,$DF,$F6,$BF
	.dc.b	$CC,$3F,$0F,$FA,$BB,$F3,$1F,$E8,$78,$F1,$C3,$85,$EF,$AD,$7B,$FD
	.dc.b	$67,$F6,$D9,$F6,$BF,$F4,$FD,$BE,$3B,$EE,$63,$CE,$73,$E7,$E2,$F8
	.dc.b	$6B,$E4,$86,$BE,$4E,$6F,$E0,$F5,$AC,$BC,$5A,$3F,$0D,$CE,$5E,$6C
	.dc.b	$78,$FD,$0B,$F4,$AF,$F4,$CF,$CB,$F9,$AB,$5A,$7F,$90,$E7,$FD,$3B
	.dc.b	$F3,$39,$9F,$2C,$E7,$6E,$B1,$8C,$3A,$65,$89,$F4,$57,$FB,$DB,$F6
	.dc.b	$36,$1F,$E7,$2E,$0F,$81,$5A,$FB,$15,$6B,$F2,$0F,$04,$E6,$3F,$82
	.dc.b	$E7,$8E,$0F,$AD,$7C,$3E,$0B,$F9,$4F,$DF,$7C,$FF,$C6,$1F,$92,$7D
	.dc.b	$F7,$FD,$D9,$FA,$7B,$F1,$67,$EA,$DF,$CA,$EF,$85,$EF,$C2,$5E,$FF
	.dc.b	$65,$D6,$BF,$88,$CB,$AF,$8B,$F6,$B1,$E3,$D7,$6B,$59,$73,$C7,$BC
	.dc.b	$72,$18,$79,$AA,$E5,$E0,$F9,$1F,$5D,$72,$9D,$FA,$0E,$0F,$F2,$B7
	.dc.b	$E8,$1C,$F1,$E4,$27,$5F,$4A,$FC,$F3,$5A,$FC,$39,$E5,$87,$6C,$BC
	.dc.b	$73,$98,$FF,$55,$74,$7F,$B7,$67,$C5,$8B,$D4,$7D,$6B,$F7,$57,$C6
	.dc.b	$76,$DF,$DB,$5A,$97,$E4,$D9,$C3,$96,$1F,$B4,$E4,$F9,$31,$8F,$83
	.dc.b	$FD,$AF,$DC,$3F,$1E,$FD,$57,$C4,$FE,$67,$EA,$BF,$D6,$1F,$63,$FE
	.dc.b	$0D,$F2,$9F,$9F,$B0,$E1,$C5,$FB,$EC,$38,$7D,$CF,$FA,$6B,$E6,$3F
	.dc.b	$CD,$1F,$A7,$6F,$4E,$BE,$6F,$FB,$6F,$0F,$E6,$98,$F2,$58,$FE,$A7
	.dc.b	$CD,$FE,$69,$39,$8E,$BA,$FF,$87,$FE,$C3,$D3,$8F,$4D,$FC,$FA,$7F
	.dc.b	$AE,$72,$F0,$1E,$5C,$FF,$DD,$5A,$BD,$17,$EB,$BF,$A8,$FD,$8F,$77
	.dc.b	$E2,$BD,$D6,$BC,$FF,$EB,$FB,$F8,$E7,$38,$7E,$A2,$31,$9F,$FB,$5E
	.dc.b	$FA,$F9,$7E,$6C,$7C,$53,$FE,$DB,$FC,$AB,$FD,$CD,$FA,$53,$F6,$4F
	.dc.b	$0F,$1F,$CA,$9C,$FC,$46,$35,$F4,$C7,$FB,$B7,$F3,$7E,$99,$79,$F2
	.dc.b	$E9,$F8,$0B,$E8,$FE,$EF,$87,$FC,$97,$F8,$03,$FC,$13,$F6,$EF,$C1
	.dc.b	$BF,$BF,$FE,$28,$FD,$EB,$FD,$05,$CB,$FC,$9E,$F7,$5A,$97,$F0,$0F
	.dc.b	$E0,$71,$E9,$F3,$2F,$DB,$7E,$BC,$63,$3B,$BF,$F4,$2C,$B9,$FC,$B3
	.dc.b	$5F,$E5,$53,$CC,$D6,$3F,$EC,$0D,$3C,$BD,$3E,$A9,$CE,$3E,$9F,$DE
	.dc.b	$9F,$5F,$FE,$9B,$F8,$BE,$54,$D3,$D3,$5A,$FB,$47,$EF,$FA,$9C,$E0
	.dc.b	$F0,$BB,$61,$C7,$5F,$16,$3D,$BD,$FC,$E1,$F1,$3E,$63,$FC,$BF,$F9
	.dc.b	$B3,$F7,$EE,$98,$7A,$67,$3E,$16,$38,$3E,$84,$7F,$99,$BF,$27,$E7
	.dc.b	$5B,$FA,$DF,$F5,$4F,$5F,$F8,$2B,$EF,$1F,$B6,$BF,$6D,$FF,$6B,$7E
	.dc.b	$04,$FA,$B7,$E9,$6F,$D9,$BF,$69,$F1,$5F,$D8,$F1,$E3,$87,$0B,$DF
	.dc.b	$D7,$6B,$9F,$FB,$AD,$69,$FD,$9D,$F1,$CD,$F9,$1F,$EC,$A3,$18,$F3
	.dc.b	$F8,$1F,$E1,$31,$8F,$BD,$FF,$73,$E7,$FD,$CF,$7F,$56,$3E,$99,$FF
	.dc.b	$75,$FC,$1F,$F9,$1B,$C6,$F1,$68,$D7,$CF,$5A,$F9,$AF,$F8,$03,$8F
	.dc.b	$2D,$7E,$0E,$2F,$AA,$F0,$E3,$EA,$CF,$B7,$1F,$F2,$6F,$87,$CC,$F5
	.dc.b	$9F,$17,$1F,$F0,$0F,$C7,$9F,$BD,$F8,$B7,$F0,$CE,$7F,$50,$C5,$F0
	.dc.b	$F6,$58,$74,$BD,$C3,$F0,$2F,$E5,$7E,$5F,$F8,$4F,$F5,$CF,$F4,$EF
	.dc.b	$EF,$AF,$CB,$1E,$2B,$ED,$1F,$86,$BF,$3E,$FE,$59,$8F,$1F,$BE,$63
	.dc.b	$C7,$0E,$17,$FC,$63,$A7,$F7,$5C,$9F,$E2,$ED,$3B,$BF,$CB,$31,$E5
	.dc.b	$F8,$2F,$F2,$A7,$A6,$3C,$F4,$FF,$5F,$57,$80,$E5,$9F,$83,$5F,$3C
	.dc.b	$FF,$9B,$67,$43,$1F,$6C,$BC,$67,$E3,$EF,$8B,$7E,$E9,$C2,$F7,$5A
	.dc.b	$97,$C5,$67,$3F,$04,$F2,$3E,$5E,$5E,$4D,$3E,$54,$76,$F9,$7C,$CF
	.dc.b	$8E,$B8,$FD,$B7,$F9,$2F,$F2,$0E,$98,$70,$8F,$F2,$77,$E0,$38,$DF
	.dc.b	$D1,$1F,$D0,$F0,$C6,$48,$FE,$2B,$F3,$DF,$EA,$2F,$C8,$DF,$40,$FD
	.dc.b	$3F,$EC,$7E,$FA,$FC,$3D,$FB,$DB,$F0,$D6,$3C,$70,$E1,$DC,$C3,$85
	.dc.b	$FE,$4A,$B5,$F4,$3F,$E9,$37,$86,$57,$6F,$F2,$F6,$9E,$28,$F0,$E9
	.dc.b	$FC,$AE,$7B,$9A,$CF,$BE,$5E,$5F,$0F,$FC,$E7,$8F,$4D,$3F,$BD,$7D
	.dc.b	$C3,$F3,$CF,$E4,$53,$9F,$03,$FD,$69,$0F,$C7,$7A,$E7,$FE,$7C,$DB
	.dc.b	$C5,$8F,$2F,$88,$EB,$3E,$1D,$FF,$39,$8F,$FA,$F7,$E4,$33,$3C,$77
	.dc.b	$F8,$72,$FD,$8D,$F0,$0F,$C5,$3B,$DB,$4E,$F7,$BA,$7E,$29,$A9,$CE
	.dc.b	$0C,$BB,$E1,$C7,$2E,$F3,$FD,$96,$3E,$AF,$E0,$BF,$E8,$9F,$C6,$5F
	.dc.b	$A3,$BF,$16,$FC,$17,$E3,$FF,$E8,$4E,$97,$EC,$1C,$31,$FF,$4E,$E1
	.dc.b	$C2,$F7,$F7,$3F,$83,$FD,$7B,$54,$AE,$BE,$17,$DD,$D3,$C1,$A7,$7A
	.dc.b	$FF,$6B,$75,$27,$39,$74,$F1,$FE,$CC,$63,$C2,$6F,$FE,$98,$F9,$EC
	.dc.b	$E3,$E4,$CB,$D5,$5F,$98,$FC,$19,$F7,$4F,$A3,$F2,$EF,$E9,$AD,$4F
	.dc.b	$C2,$DB,$C5,$8F,$3D,$BF,$49,$CA,$CB,$9C,$F2,$CF,$BC,$E7,$EF,$3C
	.dc.b	$73,$F0,$67,$C7,$F7,$7F,$EC,$2F,$D7,$1F,$BE,$71,$CB,$F6,$13,$9F
	.dc.b	$C3,$B1,$B7,$FA,$63,$EF,$9F,$8C,$F4,$E1,$F4,$5F,$FB,$33,$6F,$F6
	.dc.b	$7F,$C3,$3F,$1D,$FF,$62,$7B,$AF,$CF,$7F,$CD,$5F,$88,$FF,$97,$FF
	.dc.b	$72,$7D,$0F,$FA,$4B,$0E,$1C,$9F,$ED,$8C,$38,$7D,$E3,$0E,$16,$FD
	.dc.b	$EF,$FA,$96,$7D,$B3,$FF,$3B,$63,$FE,$D3,$3C,$71,$9F,$CF,$F1,$18
	.dc.b	$3B,$4F,$93,$ED,$5A,$FF,$84,$E7,$A9,$AC,$FC,$53,$AF,$A1,$FF,$80
	.dc.b	$3F,$62,$FC,$6B,$FC,$47,$F8,$A7,$F1,$0F,$C9,$5F,$AA,$BE,$CD,$EE
	.dc.b	$BE,$77,$FC,$FF,$87,$0F,$11,$87,$0F,$96,$DD,$6B,$EC,$FF,$DD,$6F
	.dc.b	$0F,$AB,$BC,$3F,$E8,$1C,$79,$60,$FA,$2F,$2C,$63,$E1,$BF,$DB,$7F
	.dc.b	$8B,$3F,$09,$7E,$32,$FC,$21,$F9,$5F,$F8,$77,$F7,$C7,$EB,$9F,$EE
	.dc.b	$5C,$38,$5E,$EB,$5F,$B8,$E7,$1F,$ED,$9B,$7C,$EE,$4F,$F8,$01,$87
	.dc.b	$08,$C7,$DC,$19,$1D,$34,$F0,$FD,$4F,$F7,$B9,$E9,$9F,$78,$FD,$D4
	.dc.b	$7C,$AE,$CE,$4B,$F1,$57,$EB,$8D,$7F,$D8,$DF,$98,$7F,$1B,$FC,$E9
	.dc.b	$FB,$27,$1C,$3C,$58,$70,$BD,$FC,$52,$FF,$C0,$DC,$1F,$EA,$5A,$D5
	.dc.b	$FB,$73,$F0,$C7,$83,$8F,$F1,$B3,$D4,$F2,$69,$E3,$F7,$DA,$FF,$2D
	.dc.b	$F9,$1F,$85,$97,$5A,$D4,$FF,$31,$FA,$2F,$E5,$3F,$BB,$FF,$63,$FD
	.dc.b	$37,$FA,$F7,$EB,$CF,$71,$FB,$13,$F7,$CF,$E6,$F7,$BF,$8B,$5E,$E9
	.dc.b	$FB,$76,$BD,$D6,$5F,$D7,$78,$7E,$A6,$3F,$DC,$7E,$1B,$10,$8E,$98
	.dc.b	$F8,$7D,$AF,$2E,$1C,$B3,$EF,$AF,$F0,$6C,$3B,$65,$E8,$F8,$9F,$0F
	.dc.b	$D3,$62,$FC,$B1,$E4,$F2,$B1,$9F,$F8,$DF,$F3,$FE,$B8,$71,$F2,$7E
	.dc.b	$3F,$EA,$DF,$80,$FC,$3F,$FD,$DF,$FB,$13,$DA,$7E,$64,$FC,$FD,$DD
	.dc.b	$F9,$DE,$D6,$3C,$7F,$1A,$63,$C7,$0E,$16,$FB,$9F,$DB,$DE,$EB,$EE
	.dc.b	$E1,$CB,$93,$F3,$5A,$7F,$C0,$0F,$C5,$73,$7B,$D7,$C7,$AF,$8A,$F9
	.dc.b	$9C,$35,$F0,$CE,$BE,$9B,$FC,$0B,$E5,$7E,$73,$F3,$77,$E9,$AF,$C7
	.dc.b	$9F,$A3,$FF,$0C,$FD,$7B,$F3,$DE,$CF,$F8,$7B,$0E,$17,$BA,$D7,$9E
	.dc.b	$F5,$F5,$B7,$EC,$73,$EB,$9F,$F8,$7F,$1E,$3F,$9B,$5A,$D6,$5C,$FE
	.dc.b	$01,$AF,$F8,$D2,$FB,$E1,$E4,$9C,$FC,$BF,$E7,$3D,$1F,$F7,$07,$EC
	.dc.b	$3F,$E0,$9F,$C4,$9F,$7A,$F9,$5F,$F7,$17,$D6,$FE,$9F,$F0,$D5,$EE
	.dc.b	$B5,$E4,$BE,$A3,$53,$FE,$13,$E1,$FD,$91,$FE,$DD,$5F,$AA,$FD,$6D
	.dc.b	$3C,$3F,$4E,$D7,$C3,$F3,$9F,$C2,$3C,$25,$6A,$7C,$BB,$BC,$9F,$56
	.dc.b	$7D,$71,$E5,$B7,$F9,$57,$F3,$67,$CC,$F9,$27,$3E,$8A,$5F,$B5,$67
	.dc.b	$3F,$83,$3F,$E7,$BD,$7C,$39,$75,$9C,$FD,$17,$82,$3F,$DD,$5C,$3F
	.dc.b	$0F,$F3,$17,$E1,$0F,$C7,$BA,$6E,$9C,$FC,$E6,$17,$B5,$CF,$A7,$E6
	.dc.b	$F0,$FA,$A7,$FD,$BF,$C7,$FC,$AB,$F4,$B7,$E3,$CF,$2D,$FC,$00,$FD
	.dc.b	$A9,$FA,$37,$F2,$AF,$BE,$FA,$98,$F1,$C3,$85,$EE,$BF,$97,$FE,$79
	.dc.b	$AE,$4F,$E8,$19,$F6,$CF,$F9,$3F,$81,$F3,$E7,$38,$BE,$5B,$39,$F5
	.dc.b	$BF,$1E,$BE,$2F,$86,$FE,$95,$8F,$1C,$59,$75,$D7,$C5,$F1,$3A,$CB
	.dc.b	$94,$7A,$FF,$3B,$FD,$27,$ED,$7E,$5C,$38,$ED,$EA,$D1,$5F,$A9,$F8
	.dc.b	$0D,$67,$D6,$73,$AF,$5C,$3D,$3B,$79,$B0,$FE,$AF,$B7,$97,$0F,$27
	.dc.b	$EF,$9F,$2B,$F4,$F7,$E4,$9D,$35,$F9,$33,$9F,$D7,$F1,$97,$D2,$8F
	.dc.b	$A3,$F5,$9D,$F4,$F3,$CE,$79,$BF,$D8,$DA,$3E,$67,$FB,$2F,$E9,$CF
	.dc.b	$CA,$5F,$88,$7F,$8E,$BF,$1F,$FD,$97,$EA,$DF,$E8,$DC,$38,$67,$FD
	.dc.b	$D7,$0E,$17,$BF,$39,$6F,$D3,$BF,$69,$9F,$6C,$FA,$E3,$FE,$2E,$CB
	.dc.b	$9A,$CB,$C5,$FA,$77,$E0,$2F,$86,$7D,$7F,$30,$FB,$71,$8F,$B5,$6D
	.dc.b	$FC,$1D,$CE,$3E,$5C,$FD,$71,$AF,$96,$FF,$3F,$73,$79,$2B,$59,$7B
	.dc.b	$E7,$FD,$05,$CB,$99,$F2,$61,$CB,$1F,$EC,$BC,$9D,$A7,$38,$71,$FA
	.dc.b	$BC,$6F,$F1,$34,$F2,$FE,$64,$FD,$1B,$FD,$53,$C5,$F0,$2B,$59,$AB
	.dc.b	$5F,$93,$78,$67,$30,$FD,$3B,$AE,$DE,$38,$FF,$7A,$EB,$FC,$87,$E5
	.dc.b	$BF,$C3,$FE,$9B,$F2,$E7,$E8,$6F,$3D,$FA,$93,$F6,$BF,$F0,$27,$C6
	.dc.b	$F1,$CB,$FC,$61,$8F,$1C,$38,$5E,$FE,$87,$FD,$D3,$C5,$FE,$55,$BD
	.dc.b	$E6,$FC,$A3,$F6,$FA,$F8,$A3,$C3,$3F,$7E,$31,$8F,$4D,$3F,$79,$87
	.dc.b	$3A,$FF,$50,$5F,$A3,$5F,$3D,$7F,$95,$E7,$CC,$6A,$FC,$BA,$FF,$86
	.dc.b	$BE,$BF,$3B,$7A,$70,$F7,$CF,$F8,$7B,$F4,$8C,$E5,$D7,$7F,$EC,$FF
	.dc.b	$A4,$7F,$06,$DB,$CB,$87,$0F,$89,$C6,$BE,$ED,$7E,$1F,$EE,$4F,$8A
	.dc.b	$FF,$7A,$70,$FA,$77,$F0,$68,$AD,$7C,$F7,$83,$0E,$3F,$6D,$D3,$2F
	.dc.b	$56,$9F,$52,$DE,$2E,$63,$FC,$9C,$BF,$6D,$7E,$F2,$FD,$3B,$FE,$00
	.dc.b	$7E,$65,$F1,$2F,$75,$A9,$7E,$2E,$FD,$93,$5F,$65,$DB,$4F,$A8,$7F
	.dc.b	$9B,$F0,$FF,$2E,$7E,$13,$F6,$A3,$13,$E4,$C7,$A7,$E6,$9E,$58,$C6
	.dc.b	$9E,$3F,$37,$1B,$79,$E3,$FC,$89,$FA,$8B,$F3,$8F,$0C,$E7,$EC,$52
	.dc.b	$FD,$83,$39,$FB,$A5,$F6,$5D,$23,$1F,$4B,$72,$7C,$5F,$F6,$0F,$C4
	.dc.b	$B8,$C6,$3E,$47,$FD,$9D,$DD,$F8,$DE,$4F,$8B,$3C,$3E,$3B,$96,$1F
	.dc.b	$E3,$1A,$7C,$E7,$FA,$23,$F3,$FF,$D8,$79,$1F,$AB,$FD,$F9,$F0,$FF
	.dc.b	$EC,$DF,$A0,$FF,$08,$5E,$FD,$2F,$F1,$8F,$96,$BD,$FE,$18,$FD,$13
	.dc.b	$7D,$DF,$E9,$DD,$3F,$79,$F9,$17,$F5,$EC,$B9,$6B,$6B,$59,$F4,$CB
	.dc.b	$C2,$FF,$5C,$AD,$E7,$DF,$57,$E2,$39,$C3,$8E,$3E,$9E,$27,$96,$FC
	.dc.b	$7F,$C7,$4D,$BD,$F9,$7A,$6F,$7F,$23,$A9,$CE,$DF,$92,$C7,$B6,$1C
	.dc.b	$76,$F3,$60,$F4,$73,$87,$0F,$AD,$7F,$00,$3F,$4D,$E9,$87,$B6,$73
	.dc.b	$E9,$63,$83,$E8,$47,$F6,$2C,$F1,$BF,$CD,$87,$E3,$DF,$EC,$7F,$43
	.dc.b	$FD,$F9,$F8,$97,$D1,$FC,$21,$FB,$D7,$E2,$BF,$1F,$FF,$6C,$FE,$40
	.dc.b	$BD,$F2,$AF,$7E,$3A,$FF,$DB,$F8,$6F,$5F,$F6,$5E,$5D,$5F,$F6,$69
	.dc.b	$E1,$33,$83,$F4,$1F,$F5,$2C,$F8,$EB,$C5,$C3,$BB,$FF,$6F,$4E,$71
	.dc.b	$F0,$7E,$55,$99,$4E,$7E,$05,$F1,$2F,$DA,$3C,$95,$CB,$4F,$56,$7F
	.dc.b	$A8,$BD,$FC,$3E,$AB,$9D,$6B,$3F,$F6,$26,$3D,$F2,$ED,$D9,$6B,$5F
	.dc.b	$0A,$FD,$AD,$F3,$DC,$EF,$D1,$39,$EE,$C5,$F0,$FC,$5F,$97,$2F,$CC
	.dc.b	$AF,$95,$CC,$FB,$4F,$F7,$AF,$E8,$AF,$C7,$FF,$D2,$5F,$12,$FD,$A3
	.dc.b	$FD,$05,$F9,$AB,$F6,$2F,$F4,$AF,$2B,$0E,$1F,$7C,$C3,$87,$CE,$B0
	.dc.b	$76,$70,$BE,$DE,$17,$FA,$67,$1E,$5A,$7F,$A8,$5F,$B6,$6D,$F1,$E9
	.dc.b	$FE,$93,$F6,$7F,$D6,$32,$38,$E9,$E1,$9F,$F4,$C7,$E0,$FA,$9C,$EB
	.dc.b	$EB,$FC,$0D,$F8,$F7,$C1,$E2,$F8,$55,$CF,$5F,$5D,$EE,$9F,$96,$F8
	.dc.b	$F3,$ED,$8F,$F6,$3C,$78,$E3,$DB,$3F,$F6,$CF,$D5,$F8,$73,$E7,$FA
	.dc.b	$4B,$F2,$67,$EC,$BF,$E2,$8C,$78,$D6,$A7,$3F,$4B,$DF,$76,$1C,$7D
	.dc.b	$6E,$9C,$7F,$5A,$3A,$79,$3F,$0E,$5F,$91,$FC,$5D,$FF,$9A,$F7,$FF
	.dc.b	$A4,$FF,$4D,$FF,$5E,$7E,$B7,$FE,$D6,$CB,$F8,$BD,$EF,$E0,$7F,$94
	.dc.b	$7D,$CA,$F7,$EA,$7F,$29,$CF,$AE,$7F,$46,$FF,$97,$7D,$C7,$D8,$8E
	.dc.b	$66,$74,$FF,$51,$3F,$72,$B8,$BE,$4C,$7B,$78,$0D,$7F,$92,$4F,$23
	.dc.b	$59,$FA,$3E,$D3,$DD,$AF,$9F,$2F,$66,$0F,$C7,$9F,$B0,$FF,$75,$D3
	.dc.b	$BE,$DE,$CC,$BF,$DB,$95,$AF,$D5,$FB,$4E,$70,$E5,$FA,$2F,$6D,$7D
	.dc.b	$5B,$F9,$B0,$FF,$5B,$C7,$0D,$3C,$3F,$AD,$3F,$2D,$FF,$20,$6B,$E0
	.dc.b	$9C,$C3,$F7,$EE,$18,$FC,$0B,$E1,$86,$36,$FF,$78,$47,$F7,$78,$FE
	.dc.b	$8B,$EC,$BF,$69,$FF,$44,$7E,$54,$F0,$DF,$B1,$BF,$65,$7B,$FC,$2F
	.dc.b	$75,$A9,$FE,$67,$F1,$FF,$DB,$B2,$FD,$56,$5D,$7E,$9F,$FB,$04,$7F
	.dc.b	$29,$C3,$FD,$0A,$FE,$2E,$B8,$39,$FB,$46,$9E,$3D,$3F,$81,$4F,$86
	.dc.b	$73,$9F,$9B,$F0,$5F,$EA,$F9,$FA,$E7,$3F,$87,$3F,$7A,$FE,$3B,$9F
	.dc.b	$6A,$D6,$7E,$FA,$D7,$B5,$D6,$31,$F0,$DD,$36,$F7,$D7,$E2,$B0,$E5
	.dc.b	$1F,$0B,$F1,$57,$E5,$9F,$EF,$09,$CE,$9E,$69,$CF,$DF,$F5,$DD,$0D
	.dc.b	$FB,$56,$FD,$1F,$E8,$DF,$92,$FF,$C0,$0F,$84,$FD,$F7,$EF,$8F,$D5
	.dc.b	$1F,$8B,$3D,$CF,$E8,$57,$BF,$4D,$7B,$A7,$E7,$BA,$F9,$FF,$E9,$FD
	.dc.b	$5F,$C9,$3D,$7F,$EB,$F8,$A5,$F8,$9F,$EC,$67,$31,$FD,$5B,$2E,$BA
	.dc.b	$7F,$3B,$C3,$A6,$BE,$68,$FF,$4C,$DF,$08,$F4,$FC,$63,$F7,$E7,$E5
	.dc.b	$BC,$31,$F3,$CE,$7E,$AF,$1C,$08,$FF,$53,$67,$87,$FC,$46,$EB,$B1
	.dc.b	$F9,$EF,$FA,$87,$4F,$06,$3F,$0A,$73,$F3,$9B,$FC,$CB,$E6,$F0,$7A
	.dc.b	$F3,$FF,$07,$FD,$67,$F8,$3F,$E9,$FF,$8F,$7E,$DE,$FC,$AD,$FB,$13
	.dc.b	$F5,$07,$E9,$2F,$D5,$9F,$BC,$BF,$0B,$FF,$09,$65,$CB,$1E,$38,$70
	.dc.b	$BD,$FD,$7E,$B9,$7F,$A8,$69,$DF,$4F,$A1,$8F,$1F,$46,$D6,$B2,$E7
	.dc.b	$AF,$87,$EA,$7D,$13,$9C,$3C,$D3,$E5,$F8,$2C,$CB,$2E,$DA,$FA,$BF
	.dc.b	$4D,$7E,$42,$FC,$E3,$C9,$7B,$D3,$BD,$6B,$E7,$73,$8F,$2C,$DF,$73
	.dc.b	$FD,$17,$B1,$C3,$F0,$77,$F1,$E2,$F9,$2F,$1E,$3C,$7F,$73,$7E,$2A
	.dc.b	$FE,$00,$EB,$B7,$BA,$73,$F8,$06,$32,$F6,$C3,$F1,$0E,$75,$BF,$23
	.dc.b	$F9,$77,$13,$F3,$07,$E9,$7F,$E0,$BF,$C0,$FF,$53,$E3,$3F,$1E,$FF
	.dc.b	$36,$FC,$57,$F4,$AB,$DF,$CC,$7F,$AB,$BD,$3F,$ED,$D4,$FC,$8F,$FA
	.dc.b	$55,$6B,$3F,$EA,$5F,$27,$FD,$96,$3B,$FE,$95,$FD,$67,$2E,$9A,$F8
	.dc.b	$72,$FE,$F1,$EC,$79,$AF,$BE,$DE,$4F,$6B,$17,$E7,$FD,$3D,$F1,$1F
	.dc.b	$D0,$FB,$3C,$33,$98,$FE,$8D,$DF,$8E,$2F,$F6,$2D,$F0,$F8,$EE,$71
	.dc.b	$8A,$FE,$85,$B7,$92,$37,$F5,$FF,$C2,$3E,$0F,$96,$AC,$53,$FC,$33
	.dc.b	$F2,$1C,$C6,$3C,$AE,$B5,$BA,$79,$5F,$E0,$9E,$6F,$F0,$47,$ED,$2F
	.dc.b	$CD,$9F,$98,$BF,$42,$7E,$BB,$FE,$C0,$F8,$77,$EF,$3F,$BE,$C7,$FC
	.dc.b	$69,$8F,$1C,$38,$5E,$FD,$6D,$52,$B4,$FF,$81,$4F,$0E,$4F,$D8,$7E
	.dc.b	$A4,$63,$0E,$7F,$39,$F9,$58,$E7,$7D,$F4,$F2,$FC,$77,$F8,$87,$9B
	.dc.b	$FD,$0B,$39,$D3,$FB,$3F,$EA,$4F,$DF,$78,$EB,$EE,$C7,$DB,$F2,$AA
	.dc.b	$77,$6B,$6F,$5C,$BF,$1F,$F9,$1B,$7C,$0C,$79,$67,$FD,$D3,$EE,$B3
	.dc.b	$3E,$F9,$CF,$E1,$9F,$E1,$77,$A6,$9F,$32,$3F,$C3,$5F,$88,$E3,$04
	.dc.b	$3E,$E7,$9F,$07,$AE,$3F,$77,$8F,$E7,$FF,$07,$FF,$20,$7E,$82,$FC
	.dc.b	$A7,$F6,$DE,$7B,$D5,$7E,$DB,$FE,$B1,$FA,$0E,$18,$FF,$9A,$F0,$E1
	.dc.b	$EC,$7F,$95,$EF,$7E,$EB,$F1,$0D,$DA,$F6,$FF,$9E,$3D,$4F,$F0,$2C
	.dc.b	$78,$6F,$FC,$89,$18,$D3,$C7,$1D,$2B,$9D,$78,$34,$F3,$69,$EF,$9C
	.dc.b	$DF,$0F,$CD,$BF,$D6,$3F,$7D,$8D,$7C,$D1,$FE,$B4,$FD,$37,$18,$FB
	.dc.b	$E3,$FB,$4E,$9E,$1A,$DF,$E0,$7F,$E8,$98,$C5,$7C,$5F,$88,$FF,$4F
	.dc.b	$6B,$CB,$6F,$F1,$67,$DF,$FE,$82,$F7,$1F,$E2,$FF,$BE,$E5,$C9,$F4
	.dc.b	$37,$F6,$E3,$FD,$BE,$FF,$81,$F0,$7F,$A5,$FB,$FF,$8E,$FC,$77,$E7
	.dc.b	$71,$7E,$48,$FC,$D1,$FA,$97,$F1,$9F,$5B,$FC,$F5,$87,$0F,$2F,$FD
	.dc.b	$57,$7B,$F8,$AD,$67,$D7,$DA,$FE,$52,$FD,$3C,$38,$C7,$F9,$2E,$3F
	.dc.b	$DC,$FA,$F9,$1F,$DE,$C3,$5F,$26,$0F,$17,$E3,$61,$CA,$73,$E3,$63
	.dc.b	$0F,$04,$7A,$FE,$2D,$F0,$DF,$85,$CF,$B6,$DE,$9E,$1F,$F7,$5C,$BF
	.dc.b	$16,$F4,$E3,$E2,$C7,$97,$E3,$3D,$63,$1C,$3F,$EC,$AC,$FF,$3F,$9F
	.dc.b	$4E,$0F,$44,$FF,$B3,$BE,$3B,$FC,$53,$E0,$74,$DB,$C1,$1F,$E2,$CF
	.dc.b	$57,$17,$C3,$7F,$ED,$5F,$13,$DE,$F7,$ED,$7E,$17,$E8,$FF,$A3,$7E
	.dc.b	$9C,$FD,$53,$EB,$7E,$02,$FA,$87,$EA,$4F,$D9,$3F,$A8,$61,$C2,$F7
	.dc.b	$5A,$FC,$A3,$3C,$BF,$ED,$9F,$8A,$67,$D6,$3F,$55,$87,$83,$7F,$D4
	.dc.b	$BC,$FC,$16,$BE,$1F,$73,$CD,$7C,$F0,$F3,$56,$B3,$7C,$F6,$AB,$FB
	.dc.b	$75,$6B,$07,$E3,$8F,$DA,$5E,$9F,$9A,$F7,$C1,$F0,$2B,$5F,$5D,$9E
	.dc.b	$0F,$4C,$E7,$E5,$3B,$46,$36,$F3,$DF,$F8,$07,$0E,$1A,$78,$7F,$45
	.dc.b	$FE,$99,$F9,$DE,$32,$F9,$B1,$FD,$3F,$F3,$5C,$63,$F0,$23,$F0,$3C
	.dc.b	$7F,$0F,$81,$1F,$C1,$FC,$0F,$F1,$2E,$61,$FB,$6B,$F7,$3F,$E5,$5F
	.dc.b	$AB,$3F,$6A,$FE,$9D,$F3,$AF,$D1,$BF,$C1,$1F,$47,$FE,$E9,$C3,$87
	.dc.b	$C1,$B0,$E1,$E4,$7F,$93,$BF,$59,$DE,$9D,$EF,$6F,$F2,$3C,$39,$69
	.dc.b	$FE,$D5,$8C,$1B,$7C,$7A,$7F,$5F,$F9,$AF,$E2,$95,$C3,$3F,$04,$FF
	.dc.b	$63,$F9,$8C,$E1,$C6,$73,$ED,$BC,$27,$EA,$1E,$4A,$F8,$39,$F6,$AD
	.dc.b	$7D,$56,$75,$F6,$F0,$7C,$39,$E5,$79,$BE,$1F,$3B,$E0,$8E,$79,$F9
	.dc.b	$3A,$9F,$A3,$BF,$53,$E1,$39,$8F,$E3,$BC,$FF,$52,$F8,$E7,$F4,$70
	.dc.b	$FF,$42,$CE,$74,$F1,$F0,$FE,$BF,$C5,$FE,$05,$F1,$9F,$E5,$6F,$C9
	.dc.b	$BF,$C2,$FF,$07,$FC,$9F,$F4,$57,$EC,$3F,$C2,$AB,$5F,$37,$FD,$57
	.dc.b	$C2,$53,$F1,$8D,$7D,$63,$2F,$EA,$BF,$01,$F8,$7E,$2F,$F0,$EC,$BF
	.dc.b	$62,$CC,$7E,$2F,$3E,$BA,$FF,$37,$F8,$2F,$2D,$F3,$D7,$C1,$F0,$5F
	.dc.b	$91,$F4,$3A,$63,$F5,$7E,$21,$F1,$7F,$B4,$FB,$1F,$F0,$6F,$D1,$3F
	.dc.b	$73,$FD,$8D,$F8,$FC,$46,$1C,$39,$FF,$D3,$1F,$37,$FE,$3F,$FC,$D7
	.dc.b	$77,$F9,$FB,$DF,$E2,$9F,$EC,$38,$E5,$87,$81,$FF,$65,$18,$D3,$C5
	.dc.b	$87,$FB,$17,$5F,$3E,$1E,$09,$CB,$FA,$BF,$63,$FA,$47,$82,$FD,$5B
	.dc.b	$FC,$29,$F1,$4F,$CA,$54,$FD,$15,$F9,$C6,$3B,$3E,$76,$3C,$7C,$7F
	.dc.b	$E8,$58,$70,$BF,$F0,$FE,$1D,$B2,$FD,$46,$FF,$2A,$31,$7F,$32,$3C
	.dc.b	$98,$F4,$AF,$1E,$1E,$2F,$C2,$73,$5F,$C0,$F2,$F0,$69,$E7,$CB,$CB
	.dc.b	$3E,$8F,$63,$F0,$5F,$5A,$FC,$35,$F8,$57,$F3,$FF,$6B,$1F,$AE,$F1
	.dc.b	$BD,$D6,$BE,$13,$F8,$9A,$97,$ED,$1D,$72,$FE,$29,$97,$5C,$BD,$51
	.dc.b	$91,$1E,$1F,$DA,$33,$A3,$3E,$F7,$E3,$CB,$FC,$F3,$87,$1C,$3A,$67
	.dc.b	$E2,$F8,$BF,$B1,$ED,$7E,$77,$EA,$DF,$8D,$BE,$CD,$F9,$4F,$FA,$27
	.dc.b	$B1,$F9,$C3,$E8,$7E,$DE,$3C,$70,$E1,$7B,$A7,$E7,$5A,$DB,$F9,$86
	.dc.b	$7D,$B3,$F4,$E3,$C7,$E0,$AD,$6B,$2E,$7E,$4F,$F1,$9F,$8B,$F3,$CE
	.dc.b	$70,$F2,$CF,$F1,$DF,$C6,$7F,$25,$F2,$FF,$B7,$FB,$2F,$E6,$9F,$9F
	.dc.b	$3F,$6C,$7D,$7F,$FB,$57,$F6,$67,$E8,$FF,$94,$C3,$87,$C2,$30,$E1
	.dc.b	$7B,$F3,$75,$B7,$F4,$9C,$FF,$A0,$69,$F4,$B1,$E3,$85,$AD,$63,$CF
	.dc.b	$4F,$0F,$D5,$7E,$EB,$C0,$61,$C2,$73,$F9,$E7,$E5,$30,$7E,$5C,$FD
	.dc.b	$C9,$F8,$F3,$F1,$87,$EC,$CF,$DD,$DF,$BD,$3F,$46,$FD,$66,$5F,$D7
	.dc.b	$F2,$E5,$8F,$1C,$38,$5B,$F1,$AD,$DF,$83,$3E,$EF,$EA,$F7,$FF,$22
	.dc.b	$63,$E3,$FD,$F3,$F2,$3E,$AF,$F2,$58,$EB,$B7,$93,$67,$A7,$EB,$9C
	.dc.b	$EB,$E7,$F6,$31,$A7,$B3,$0E,$1F,$93,$FD,$7D,$3B,$CE,$6F,$DD,$2F
	.dc.b	$D3,$BC,$F8,$F4,$CF,$FD,$47,$FA,$96,$30,$E1,$1F,$EA,$6F,$AB,$C7
	.dc.b	$1F,$C1,$8F,$D5,$F0,$FF,$14,$F7,$FF,$9F,$7E,$AC,$FD,$49,$F7,$6F
	.dc.b	$DD,$7E,$87,$DF,$BC,$17,$EF,$3C,$72,$FF,$06,$63,$C7,$0E,$1C,$BF
	.dc.b	$EA,$5B,$FF,$06,$72,$FF,$38,$BD,$E7,$FE,$B0,$7E,$76,$9E,$28,$F0
	.dc.b	$C6,$0D,$E1,$D3,$3F,$14,$78,$46,$5E,$1D,$3C,$FE,$57,$F3,$5C,$7A
	.dc.b	$61,$C7,$F5,$6F,$F3,$E6,$7E,$3C,$78,$E7,$F0,$31,$F7,$D7,$F9,$CE
	.dc.b	$B5,$97,$5D,$7F,$D8,$1F,$BA,$AB,$C5,$39,$D5,$2F,$CC,$3F,$A9,$E7
	.dc.b	$F2,$FE,$9D,$FB,$5B,$F1,$3F,$DB,$7E,$46,$FC,$49,$F4,$DC,$38,$3F
	.dc.b	$C6,$57,$BF,$21,$7B,$FB,$8D,$64,$ED,$FE,$BD,$A7,$F8,$FF,$1E,$2F
	.dc.b	$F5,$E3,$13,$E2,$D3,$C3,$F8,$A7,$F7,$59,$FF,$00,$57,$82,$7C,$BF
	.dc.b	$7D,$99,$E3,$A7,$8F,$3F,$F4,$EF,$C3,$F8,$A3,$1F,$80,$7F,$61,$F4
	.dc.b	$FF,$BF,$E5,$CF,$7F,$4F,$D2,$3E,$C3,$86,$BF,$0A,$1F,$82,$7F,$9C
	.dc.b	$3F,$16,$FF,$09,$7D,$E3,$E0,$1F,$38,$FD,$6F,$FB,$5F,$E9,$EF,$D5
	.dc.b	$FF,$DF,$DB,$B1,$E3,$87,$0B,$DD,$3F,$3B,$ED,$CB,$FE,$65,$CF,$B6
	.dc.b	$7F,$D4,$B2,$E5,$39,$C5,$F8,$C6,$73,$EB,$7E,$3F,$4D,$AB,$E1,$BF
	.dc.b	$C4,$15,$AD,$BC,$FF,$65,$8C,$FA,$63,$C7,$0F,$89,$DB,$E4,$9F,$F5
	.dc.b	$56,$9E,$9C,$FA,$F2,$3D,$0D,$70,$7A,$34,$F8,$7F,$6D,$D7,$1E,$3A
	.dc.b	$75,$FC,$69,$FA,$3F,$0D,$3A,$6C,$97,$E8,$F9,$BF,$27,$E0,$1F,$E8
	.dc.b	$9F,$C0,$DF,$A1,$7F,$91,$7F,$3B,$FA,$FF,$76,$FC,$E3,$FD,$5B,$FB
	.dc.b	$97,$F7,$5F,$CF,$D8,$70,$E0,$FF,$49,$61,$C2,$F7,$F3,$EF,$D9,$F7
	.dc.b	$9F,$5F,$07,$FD,$2D,$87,$F8,$77,$1E,$4B,$1F,$C3,$FE,$97,$FE,$08
	.dc.b	$9C,$C7,$5F,$AF,$7E,$FD,$FE,$5B,$9C,$E1,$FE,$A1,$D3,$CF,$F9,$4F
	.dc.b	$B7,$3F,$07,$07,$AF,$17,$03,$E5,$DF,$96,$F9,$B7,$F5,$E7,$CE,$F7
	.dc.b	$4F,$D8,$B5,$5E,$7A,$D6,$0F,$0B,$C1,$3C,$B2,$F0,$4F,$FA,$43,$6F
	.dc.b	$3C,$63,$C7,$FF,$27,$46,$37,$F9,$F1,$FE,$B9,$8C,$69,$E8,$7A,$E1
	.dc.b	$E1,$BE,$BC,$3F,$6E,$3F,$CF,$DA,$BF,$1F,$FF,$17,$FC,$0B,$ED,$1F
	.dc.b	$58,$FC,$7D,$FA,$4F,$DC,$C7,$8E,$1C,$2F,$74,$F7,$F5,$5F,$ED,$1C
	.dc.b	$BB,$7E,$99,$F5,$E3,$8F,$D5,$57,$58,$F3,$CF,$FC,$A7,$BF,$FA,$23
	.dc.b	$0F,$35,$FF,$A0,$B3,$F2,$BE,$BC,$38,$B9,$B7,$C8,$3F,$2E,$F1,$D7
	.dc.b	$DB,$CB,$B4,$BF,$57,$CE,$5D,$B8,$3E,$16,$3F,$DF,$30,$CE,$1E,$5F
	.dc.b	$98,$7E,$A7,$8D,$3C,$10,$FB,$F7,$E1,$3F,$D3,$FF,$80,$3E,$57,$E9
	.dc.b	$5F,$E3,$5F,$D3,$7F,$CA,$1E,$17,$F7,$1A,$D7,$3A,$B5,$F5,$3F,$B0
	.dc.b	$FC,$C3,$9F,$0F,$FA,$D7,$2F,$A9,$FA,$37,$F9,$A3,$1E,$38,$5A,$D6
	.dc.b	$3C,$FE,$05,$A7,$F9,$7F,$88,$74,$2E,$5E,$69,$DE,$9F,$DC,$32,$EB
	.dc.b	$18,$F6,$FF,$99,$C6,$30,$FF,$35,$47,$F2,$68,$C5,$76,$FD,$AB,$B6
	.dc.b	$2C,$7B,$61,$E3,$FB,$5F,$8F,$F4,$3F,$ED,$1F,$3D,$F9,$33,$F5,$96
	.dc.b	$4F,$C9,$3F,$CB,$BF,$83,$7F,$8C,$36,$63,$C7,$0E,$17,$BF,$AF,$FF
	.dc.b	$66,$7E,$F7,$DB,$7F,$83,$A7,$6D,$3F,$DB,$B1,$FE,$8D,$C1,$F9,$D6
	.dc.b	$2B,$58,$F3,$FC,$2B,$E4,$E9,$E0,$C3,$9E,$1E,$0D,$BF,$CE,$38,$EB
	.dc.b	$5F,$26,$7E,$A9,$FF,$3A,$FE,$1C,$FB,$0F,$26,$3E,$1A,$D6,$FF,$EF
	.dc.b	$1F,$CA,$B3,$BF,$8B,$1F,$F3,$EF,$DB,$FE,$D7,$6F,$16,$3C,$78,$3E
	.dc.b	$04,$FF,$B4,$27,$3F,$91,$3E,$41,$F9,$93,$F6,$77,$E5,$1F,$DB,$7F
	.dc.b	$71,$FE,$B5,$87,$0B,$DD,$6B,$E2,$D5,$FE,$73,$AD,$7E,$1F,$F2,$B2
	.dc.b	$ED,$95,$C7,$8E,$5F,$DE,$7E,$DF,$E2,$6D,$F9,$2B,$DE,$9D,$7E,$57
	.dc.b	$32,$9C,$DF,$F8,$07,$F3,$27,$E4,$DF,$22,$73,$BA,$5F,$9D,$67,$4F
	.dc.b	$0C,$FF,$A9,$3F,$14,$C6,$DF,$07,$2F,$F0,$67,$3B,$F3,$5F,$DC,$FC
	.dc.b	$8F,$E9,$B6,$7E,$58,$FD,$C7,$F9,$27,$E8,$4F,$8D,$7E,$3F,$FE,$99
	.dc.b	$FD,$9F,$8E,$1F,$AE,$C3,$87,$9E,$C3,$85,$FF,$A7,$FE,$A7,$77,$DB
	.dc.b	$27,$E7,$1F,$BD,$D3,$C5,$97,$2A,$FF,$80,$1D,$9F,$E2,$57,$B5,$F9
	.dc.b	$35,$F1,$FA,$FF,$96,$C3,$8C,$E7,$37,$EA,$79,$C3,$8E,$DE,$5C,$7D
	.dc.b	$7F,$86,$BF,$56,$F2,$EB,$EA,$E0,$F4,$CB,$F6,$5C,$E9,$D7,$F2,$B7
	.dc.b	$E5,$38,$CF,$A6,$C9,$7E,$6B,$9C,$BE,$1F,$E4,$DF,$C5,$3E,$EF,$F2
	.dc.b	$7C,$1F,$C5,$BF,$65,$7E,$D1,$F3,$BF,$53,$7E,$CE,$FD,$A5,$F9,$BB
	.dc.b	$E1,$59,$72,$C9,$97,$2D,$D9,$72,$C7,$8F,$EA,$FC,$3C,$2F,$C0,$5E
	.dc.b	$F0,$FF,$62,$46,$35,$F0,$46,$0F,$33,$E5,$D7,$C9,$E5,$7A,$EB,$86
	.dc.b	$7E,$09,$F4,$7D,$66,$70,$E3,$39,$CF,$F4,$1F,$9F,$F8,$E7,$3A,$F9
	.dc.b	$FC,$C7,$EC,$79,$9F,$6E,$BE,$1C,$3F,$AC,$70,$7C,$1D,$FE,$06,$3E
	.dc.b	$FF,$CA,$BC,$18,$70,$D3,$C1,$FA,$7F,$DB,$C3,$86,$5C,$E1,$F1,$7D
	.dc.b	$F2,$E7,$C1,$F4,$7E,$3B,$9C,$63,$57,$D9,$7D,$BF,$6F,$F5,$5D,$AF
	.dc.b	$CA,$3F,$C6,$BF,$AB,$BF,$4F,$7A,$0F,$BE,$7E,$CF,$F6,$B1,$E3,$87
	.dc.b	$0B,$DD,$3F,$7D,$D7,$2F,$F5,$CC,$FB,$7D,$6F,$EE,$63,$F7,$5C,$AC
	.dc.b	$B9,$F3,$E6,$7C,$C7,$1C,$FB,$6A,$FC,$0B,$33,$FE,$23,$9C,$FE,$9F
	.dc.b	$8C,$FB,$F0,$7A,$F0,$7E,$A5,$FE,$FD,$CF,$D3,$AA,$B5,$3F,$98,$D7
	.dc.b	$C9,$9F,$86,$B5,$C1,$F1,$3F,$52,$7D,$4F,$BA,$BC,$55,$AC,$D4,$FC
	.dc.b	$AF,$F4,$3F,$1B,$FC,$67,$E8,$9F,$9F,$3F,$09,$7E,$7A,$FD,$F7,$F6
	.dc.b	$FE,$EB,$F1,$EE,$3B,$7F,$81,$B1,$E3,$87,$0B,$DF,$CD,$6A,$BF,$C5
	.dc.b	$F9,$75,$7E,$67,$17,$F5,$4C,$3C,$33,$F3,$E7,$84,$74,$CF,$C5,$EA
	.dc.b	$FE,$22,$73,$5E,$79,$CF,$85,$FC,$17,$6F,$64,$E7,$3F,$E1,$FF,$92
	.dc.b	$BF,$23,$E3,$B7,$9F,$34,$FF,$5E,$D7,$C9,$87,$19,$7E,$AA,$FD,$7F
	.dc.b	$8E,$BE,$48,$F9,$B6,$FA,$3F,$E8,$9F,$46,$FD,$BD,$FA,$8F,$F8,$5F
	.dc.b	$EB,$9F,$96,$7E,$0B,$E1,$FF,$89,$E1,$C2,$F7,$F3,$9F,$E5,$5A,$FC
	.dc.b	$3F,$16,$B9,$BF,$C6,$B9,$F5,$CF,$E4,$69,$E1,$9C,$E3,$FE,$3B,$7F
	.dc.b	$3D,$19,$55,$C7,$1E,$DE,$2F,$F2,$DC,$FC,$58,$F9,$F3,$FC,$F6,$1E
	.dc.b	$13,$8E,$5E,$6D,$7D,$7D,$AF,$CA,$B5,$A7,$59,$CF,$E5,$B8,$BE,$18
	.dc.b	$FE,$53,$F0,$1C,$67,$DF,$27,$EF,$7D,$FF,$23,$7E,$17,$F9,$5F,$B8
	.dc.b	$FA,$FF,$95,$BF,$5D,$FE,$7D,$CB,$ED,$BD,$F7,$2F,$74,$FD,$27,$A6
	.dc.b	$3F,$A0,$E4,$FC,$9F,$2F,$F7,$BC,$78,$C6,$0F,$F3,$1B,$D3,$0F,$0F
	.dc.b	$C4,$7C,$79,$CD,$F9,$7C,$06,$AF,$C9,$35,$4C,$79,$56,$BF,$5E,$FD
	.dc.b	$76,$FE,$9C,$79,$62,$97,$CE,$7A,$38,$3D,$98,$76,$D3,$FC,$5F,$8F
	.dc.b	$19,$CE,$09,$FF,$10,$4E,$75,$F9,$3F,$2B,$FE,$7E,$FA,$0F,$E4,$5F
	.dc.b	$A4,$FE,$77,$CE,$FF,$41,$FC,$2F,$E6,$7F,$74,$7E,$72,$BD,$F9,$1F
	.dc.b	$C2,$FC,$FF,$EC,$FD,$EB,$7E,$E3,$BE,$7F,$F1,$36,$9F,$A9,$CB,$F7
	.dc.b	$D9,$72,$D2,$D6,$B3,$E9,$A7,$EA,$7F,$0F,$F3,$4E,$6B,$CB,$A7,$E1
	.dc.b	$23,$B1,$FE,$98,$CB,$D5,$AF,$93,$EE,$7F,$17,$9F,$78,$C7,$83,$C2
	.dc.b	$D7,$C3,$3F,$DC,$7E,$1D,$E0,$9E,$9A,$F9,$33,$FF,$0E,$7B,$5C,$70
	.dc.b	$E1,$1F,$E1,$F9,$F8,$DF,$9B,$FF,$91,$7F,$1D,$7E,$4E,$FC,$6D,$FB
	.dc.b	$EB,$F2,$9F,$E2,$32,$FA,$AE,$3A,$FF,$AC,$B0,$E3,$87,$0B,$DF,$F0
	.dc.b	$0E,$A9,$5D,$BC,$4F,$9B,$9F,$87,$2F,$05,$7F,$5B,$E4,$4E,$72,$E9
	.dc.b	$E0,$BF,$2D,$FC,$24,$E7,$6F,$34,$FE,$6F,$0F,$06,$7E,$49,$CF,$EE
	.dc.b	$4F,$D3,$73,$A7,$BE,$73,$EC,$E3,$4E,$FE,$1F,$6C,$BD,$D7,$BE,$1F
	.dc.b	$E6,$7E,$0E,$67,$DB,$39,$D5,$FA,$57,$F6,$0E,$1F,$5D,$72,$CB,$D7
	.dc.b	$3F,$D7,$E7,$33,$CA,$7D,$FF,$55,$D7,$83,$D5,$9F,$F4,$2E,$5F,$F7
	.dc.b	$0F,$E1,$CF,$C1,$BF,$9D,$FC,$BF,$F6,$6F,$DB,$7F,$C8,$BE,$2B,$F2
	.dc.b	$47,$E1,$2C,$38,$6D,$FC,$DB,$0E,$1E,$17,$EA,$AD,$FB,$36,$F3,$EB
	.dc.b	$F0,$9F,$D8,$71,$E3,$8F,$F3,$6C,$7F,$D6,$4F,$F8,$4E,$33,$EE,$FF
	.dc.b	$8C,$29,$AF,$93,$C5,$F4,$E3,$D2,$3C,$FE,$56,$31,$F2,$5F,$0D,$7F
	.dc.b	$7B,$FA,$F3,$F3,$2F,$25,$6B,$0F,$7C,$BF,$39,$CC,$F9,$70,$E5,$C5
	.dc.b	$FE,$83,$FC,$E3,$3B,$79,$71,$F8,$FF,$66,$FD,$FF,$E3,$63,$FE,$A9
	.dc.b	$FA,$6F,$F1,$07,$E8,$FF,$E7,$2F,$C8,$3F,$D7,$3F,$B6,$3E,$27,$77
	.dc.b	$FE,$C3,$F9,$95,$EF,$EB,$56,$FC,$FF,$77,$FE,$C5,$BF,$EB,$2F,$F8
	.dc.b	$FF,$A7,$FC,$81,$9F,$82,$3E,$61,$E3,$9E,$BF,$4C,$D3,$FD,$97,$F8
	.dc.b	$EF,$E1,$2B,$5A,$78,$71,$FF,$69,$FB,$AF,$8D,$7E,$AC,$FD,$9D,$F4
	.dc.b	$2F,$CB,$5F,$B2,$BF,$4D,$FF,$7E,$E3,$C7,$0E,$17,$BA,$7C,$06,$B2
	.dc.b	$EB,$AF,$FA,$FB,$E6,$FF,$DA,$58,$F8,$7A,$3F,$C5,$5F,$80,$7F,$C0
	.dc.b	$08,$E9,$B7,$8B,$6F,$F5,$EE,$1C,$F5,$EF,$1F,$D1,$E1,$7F,$E5,$07
	.dc.b	$C3,$F5,$CF,$D2,$1F,$B4,$FF,$BF,$BC,$66,$BF,$C7,$30,$E1,$BF,$FA
	.dc.b	$CA,$F7,$F8,$6F,$FA,$B2,$BF,$D6,$B5,$AC,$7A,$65,$FE,$D8,$FA,$9F
	.dc.b	$F7,$66,$3F,$EB,$AF,$C9,$7D,$71,$8A,$F1,$68,$C9,$FA,$FF,$FB,$CF
	.dc.b	$2E,$9A,$78,$35,$79,$AD,$7D,$BB,$7F,$BE,$7E,$85,$F7,$4F,$D7,$BF
	.dc.b	$C7,$FF,$1B,$FE,$CC,$FC,$B7,$FC,$39,$A7,$BB,$1E,$58,$70,$BD,$FD
	.dc.b	$C2,$DF,$43,$DA,$F7,$97,$FB,$3D,$F3,$67,$E1,$8F,$04,$7D,$B3,$78
	.dc.b	$74,$F9,$AF,$46,$7E,$19,$FF,$29,$D7,$9F,$36,$9F,$EE,$2F,$84,$DE
	.dc.b	$5D,$72,$F2,$FE,$26,$F5,$BC,$99,$72,$9C,$FC,$AE,$32,$F6,$1D,$38
	.dc.b	$18,$3E,$FB,$5A,$F8,$75,$FF,$78,$CF,$FB,$B7,$6F,$37,$E8,$EF,$BC
	.dc.b	$7E,$52,$FD,$B7,$FC,$0B,$FA,$C3,$F4,$DF,$E5,$18,$70,$BD,$D6,$A7
	.dc.b	$F2,$73,$9F,$D4,$BF,$3B,$C9,$F6,$F8,$D9,$F7,$CF,$31,$F7,$E3,$18
	.dc.b	$F3,$8F,$83,$18,$F9,$D6,$8F,$CC,$7E,$4D,$71,$D7,$C7,$5A,$D5,$FA
	.dc.b	$5F,$29,$CE,$1E,$BF,$D2,$7F,$53,$E2,$79,$7F,$3D,$7E,$FF,$F9,$1F
	.dc.b	$C4,$9F,$A2,$7F,$91,$BE,$21,$FA,$37,$0D,$FF,$2D,$87,$05,$7C,$3E
	.dc.b	$9F,$FA,$A7,$CF,$6E,$55,$DB,$0F,$D3,$C7,$2C,$BF,$D5,$8E,$E7,$A6
	.dc.b	$1C,$E3,$F1,$B9,$F9,$EB,$CD,$9F,$F9,$6B,$1C,$D7,$33,$C7,$1F,$E5
	.dc.b	$E5,$F9,$FF,$5A,$E3,$9F,$EB,$FD,$DF,$37,$EE,$9F,$EA,$0F,$8A,$7E
	.dc.b	$C6,$F0,$9F,$92,$FF,$99,$7A,$F7,$F6,$DF,$ED,$4A,$D4,$FF,$9A,$7F
	.dc.b	$4A,$CE,$1E,$ED,$99,$7F,$AB,$3F,$22,$C3,$DF,$0E,$3F,$37,$EA,$AD
	.dc.b	$E5,$D7,$4F,$0F,$D9,$67,$0E,$11,$E4,$FA,$77,$E5,$7E,$0C,$BC,$70
	.dc.b	$FC,$DF,$15,$BF,$56,$F0,$7F,$9F,$FD,$7F,$E5,$7F,$7D,$F9,$73,$FD
	.dc.b	$83,$F5,$4F,$C3,$FF,$C5,$07,$E0,$BD,$BD,$F0,$FF,$17,$AD,$CE,$BC
	.dc.b	$17,$E4,$9A,$A5,$7F,$B3,$3D,$9F,$E8,$2F,$7E,$4F,$F1,$0C,$FF,$25
	.dc.b	$8C,$65,$E1,$FA,$BF,$17,$81,$FC,$3A,$73,$9F,$93,$DA,$C6,$1E,$3C
	.dc.b	$7F,$C9,$33,$E8,$BF,$F1,$17,$A3,$E2,$AD,$65,$D6,$BF,$C9,$9F,$9F
	.dc.b	$79,$FE,$A3,$F0,$1F,$91,$3F,$7F,$FE,$93,$F3,$EF,$AD,$7E,$F5,$BE
	.dc.b	$95,$EF,$D7,$5E,$FF,$4C,$D4,$FC,$0D,$BF,$9F,$67,$F8,$DF,$A8,$C4
	.dc.b	$7F,$BF,$B1,$E7,$A7,$83,$E1,$BD,$B3,$9C,$3B,$65,$F8,$08,$EB,$5E
	.dc.b	$6E,$67,$E0,$D8,$D7,$C7,$0F,$DB,$FF,$B4,$FE,$0F,$FC,$57,$E5,$9F
	.dc.b	$89,$3F,$13,$7E,$35,$FE,$4D,$FC,$87,$FD,$05,$F9,$97,$0D,$BF,$AB
	.dc.b	$61,$C2,$F7,$ED,$2F,$FD,$11,$F8,$86,$EF,$FC,$15,$93,$F5,$9F,$F0
	.dc.b	$0E,$9E,$18,$F0,$4F,$F2,$73,$FA,$5C,$67,$1E,$B8,$3F,$1E,$CD,$78
	.dc.b	$AF,$D1,$AF,$9F,$97,$33,$E0,$C3,$8F,$DA,$7F,$BC,$FC,$06,$7D,$70
	.dc.b	$F7,$4F,$FB,$6E,$73,$7D,$6F,$C3,$FB,$03,$E9,$BD,$9C,$3F,$07,$1F
	.dc.b	$8D,$3F,$E2,$9F,$C1,$7F,$BD,$E1,$FD,$EF,$F5,$37,$EA,$3F,$DE,$BB
	.dc.b	$7F,$85,$FE,$CA,$FD,$39,$FA,$33,$1E,$38,$B1,$E3,$F2,$7F,$C3,$B0
	.dc.b	$E1,$FA,$46,$FE,$73,$FA,$75,$76,$E4,$F9,$58,$7F,$82,$F1,$E7,$3C
	.dc.b	$A3,$31,$8D,$7C,$9E,$AB,$2F,$C1,$D7,$53,$D5,$3A,$87,$EA,$3D,$72
	.dc.b	$F0,$CE,$7A,$3D,$78,$F5,$D7,$D1,$0F,$C7,$7A,$65,$ED,$BE,$9F,$67
	.dc.b	$E7,$DB,$FB,$47,$C9,$7E,$5F,$F1,$8F,$F1,$FF,$DD,$3F,$23,$7F,$00
	.dc.b	$FE,$3F,$6F,$AF,$9F,$FB,$F3,$C6,$61,$C2,$F7,$F8,$47,$CD,$CF,$AD
	.dc.b	$3F,$3B,$F8,$F8,$64,$FF,$19,$69,$E1,$7E,$3D,$70,$C7,$A7,$D2,$79
	.dc.b	$E7,$78,$78,$3C,$1F,$F4,$96,$7F,$ED,$3D,$3C,$B9,$7A,$E7,$3F,$8B
	.dc.b	$FF,$35,$9C,$DF,$1D,$FD,$D2,$FC,$F7,$F0,$9E,$EF,$F8,$2F,$F6,$EF
	.dc.b	$EE,$BF,$B6,$3F,$3E,$7E,$A5,$FB,$1F,$CC,$5F,$25,$C3,$87,$FC,$99
	.dc.b	$87,$0B,$DF,$9C,$BF,$CC,$65,$D7,$E6,$19,$FC,$4E,$0F,$E5,$19,$32
	.dc.b	$FF,$02,$FC,$56,$70,$E1,$7E,$3D,$7F,$81,$7D,$DF,$E4,$67,$3A,$F8
	.dc.b	$3E,$87,$19,$79,$34,$EF,$3E,$4D,$5F,$A4,$7A,$6B,$C1,$5A,$CB,$DC
	.dc.b	$67,$5F,$63,$FC,$47,$DA,$74,$E6,$B8,$70,$FF,$24,$FD,$8F,$F5,$DF
	.dc.b	$3B,$FE,$35,$FC,$F3,$FC,$85,$F1,$3F,$ED,$6F,$D3,$5F,$92,$BF,$60
	.dc.b	$7E,$5D,$C3,$83,$FC,$ED,$87,$0F,$C2,$58,$70,$B7,$E6,$1B,$B5,$F6
	.dc.b	$7F,$63,$9F,$87,$EA,$7F,$C5,$B1,$83,$AA,$D3,$C9,$A7,$F9,$2F,$B3
	.dc.b	$FD,$41,$87,$87,$3F,$3D,$7E,$FF,$1E,$8F,$0C,$DF,$B7,$78,$EF,$86
	.dc.b	$3E,$DF,$7B,$CD,$FB,$73,$F1,$17,$E9,$6F,$C9,$BF,$D9,$DF,$B7,$7F
	.dc.b	$B1,$3E,$F7,$75,$A9,$CF,$E0,$FF,$EC,$8E,$1F,$F5,$A6,$3C,$F1,$FC
	.dc.b	$57,$07,$F9,$8E,$3B,$7E,$33,$8E,$4F,$F5,$04,$6F,$1E,$FF,$8E,$FA
	.dc.b	$E7,$39,$F7,$FC,$35,$B7,$FB,$0F,$4E,$F3,$98,$FF,$6C,$FE,$1D,$E6
	.dc.b	$D7,$D5,$7C,$36,$FE,$F9,$87,$FB,$27,$F3,$7F,$F7,$07,$E4,$7F,$E5
	.dc.b	$EF,$B1,$7D,$6B,$F2,$D7,$E7,$1F,$6E,$F7,$F4,$EB,$DD,$3F,$1A,$D7
	.dc.b	$E1,$6C,$BD,$DC,$4C,$7F,$DD,$DF,$84,$E1,$E7,$8F,$83,$4F,$81,$5B
	.dc.b	$C3,$96,$7D,$E7,$FD,$05,$3D,$B1,$F4,$43,$F4,$B7,$CC,$E6,$F8,$47
	.dc.b	$4C,$BB,$D6,$F8,$B5,$9F,$C0,$A7,$4F,$F5,$3F,$9C,$FF,$4D,$7C,$7F
	.dc.b	$F3,$9F,$35,$F9,$07,$AF,$F3,$37,$DB,$3F,$1D,$E1,$C5,$FE,$DB,$C3
	.dc.b	$85,$EE,$9F,$95,$6A,$BF,$B0,$FC,$67,$F8,$61,$FE,$AB,$9F,$82,$3F
	.dc.b	$C3,$11,$83,$A3,$FE,$1E,$BF,$17,$E0,$7F,$D7,$EB,$86,$9E,$6E,$26
	.dc.b	$BF,$EB,$BC,$7C,$1A,$79,$27,$3A,$3F,$67,$CE,$9D,$67,$3F,$BF,$F9
	.dc.b	$FF,$05,$FF,$6D,$EC,$FC,$8D,$F9,$97,$F8,$77,$CD,$7E,$9B,$BF,$C3
	.dc.b	$7E,$3D,$6B,$E9,$BF,$81,$ED,$9C,$7F,$97,$71,$FF,$B1,$F3,$FF,$73
	.dc.b	$FE,$BF,$8E,$21,$FE,$BC,$CB,$9E,$BF,$E2,$BF,$C8,$7E,$04,$EA,$FC
	.dc.b	$BA,$77,$FA,$8C,$CF,$1C,$FD,$18,$79,$FC,$3F,$15,$F9,$78,$1F,$2B
	.dc.b	$F5,$7A,$3F,$DC,$5F,$C0,$0F,$DC,$5F,$91,$BD,$5F,$D0,$7A,$DB,$F8
	.dc.b	$95,$EE,$B5,$E5,$BE,$9B,$53,$ED,$E1,$FF,$2B,$DB,$E3,$31,$A7,$83
	.dc.b	$4E,$FA,$7E,$83,$D0,$74,$FF,$BB,$6B,$5A,$F8,$FC,$AF,$7C,$62,$FF
	.dc.b	$69,$F5,$3E,$0C,$79,$EF,$FE,$37,$C7,$96,$9E,$89,$CF,$E9,$BB,$CF
	.dc.b	$D3,$F8,$E7,$E5,$FD,$87,$EF,$BF,$EC,$4F,$D8,$1F,$94,$3F,$75,$F5
	.dc.b	$7C,$E6,$EB,$53,$9E,$4F,$F6,$67,$27,$AB,$97,$EE,$E5,$D3,$EA,$7E
	.dc.b	$23,$DB,$F3,$DB,$8F,$2C,$FE,$35,$F5,$B6,$9E,$29,$FC,$F6,$3D,$6B
	.dc.b	$F9,$E5,$F1,$CB,$F8,$DC,$E7,$2F,$06,$1C,$65,$F8,$D7,$F9,$4F,$F3
	.dc.b	$9F,$C8,$E3,$FE,$39,$FC,$99,$FB,$97,$F8,$27,$F2,$37,$EE,$0F,$95
	.dc.b	$FF,$25,$6A,$C7,$8F,$A4,$C7,$8F,$CD,$31,$E3,$83,$F7,$1F,$06,$9F
	.dc.b	$A0,$C3,$87,$DC,$FF,$84,$F2,$E7,$97,$8A,$BF,$CB,$71,$8D,$7C,$9A
	.dc.b	$FC,$28,$C6,$9D,$E3,$F1,$51,$8D,$3C,$3F,$79,$E0,$C3,$DD,$87,$2D
	.dc.b	$7F,$D8,$1B,$7A,$E7,$F6,$7F,$94,$3F,$48,$7E,$9E,$FD,$D5,$FB,$AB
	.dc.b	$F1,$E7,$E2,$6F,$CD,$7B,$63,$C7,$0E,$17,$BA,$7E,$1B,$AC,$BA,$E7
	.dc.b	$EE,$3C,$2B,$F5,$4F,$89,$18,$C3,$9E,$7F,$E7,$68,$E5,$18,$FC,$9B
	.dc.b	$FD,$09,$C7,$F7,$34,$F1,$6D,$E8,$C3,$D7,$F8,$6F,$08,$C6,$0F,$DA
	.dc.b	$B1,$BF,$B6,$1F,$BD,$E3,$0F,$87,$F9,$87,$CB,$E8,$7E,$A0,$FD,$7D
	.dc.b	$F8,$0B,$F4,$EF,$F4,$27,$C0,$BF,$0D,$E1,$7B,$AD,$78,$4A,$7E,$1D
	.dc.b	$AF,$CB,$3E,$0F,$3F,$E4,$B0,$EC,$FC,$73,$19,$F8,$33,$FF,$49,$7E
	.dc.b	$95,$F0,$A7,$F9,$6E,$BE,$6A,$D6,$6F,$4F,$55,$CB,$5F,$1E,$BF,$E6
	.dc.b	$3E,$FF,$C2,$4E,$73,$F5,$FE,$77,$F8,$BE,$E7,$C4,$BF,$41,$7E,$E3
	.dc.b	$FE,$03,$FC,$79,$E3,$FF,$A4,$F3,$E7,$87,$0B,$DF,$C8,$AF,$7F,$82
	.dc.b	$75,$C9,$FA,$27,$C2,$C3,$86,$DF,$05,$E5,$A7,$7F,$3B,$D7,$18,$CB
	.dc.b	$A7,$E4,$D8,$C7,$CD,$7E,$5D,$9F,$88,$F7,$DB,$C7,$B7,$A3,$E5,$FF
	.dc.b	$C0,$7A,$FF,$06,$FC,$E1,$F9,$9B,$F3,$8F,$C4,$7E,$95,$FD,$9F,$F0
	.dc.b	$CF,$F0,$B7,$CA,$7F,$BF,$EF,$75,$AF,$92,$FF,$7A,$FC,$6F,$4C,$BF
	.dc.b	$B0,$D6,$9F,$F6,$ED,$F1,$C9,$FB,$17,$FA,$EA,$31,$8F,$3F,$C4,$FF
	.dc.b	$B4,$C6,$3C,$D6,$AF,$D0,$7F,$8E,$CE,$63,$CD,$E0,$78,$73,$69,$E6
	.dc.b	$C3,$8E,$CD,$E6,$3C,$61,$F8,$37,$FB,$5A,$30,$BF,$0B,$E9,$F1,$7F
	.dc.b	$BE,$FC,$31,$F9,$F7,$F8,$5B,$F1,$07,$E5,$4F,$DD,$77,$E0,$AF,$7F
	.dc.b	$4F,$FE,$E3,$AF,$D6,$69,$FC,$DB,$87,$FB,$AD,$BF,$36,$F8,$91,$FC
	.dc.b	$EB,$4F,$03,$F2,$8C,$4F,$5D,$3C,$5F,$AD,$7C,$89,$CE,$7E,$0D,$7E
	.dc.b	$FE,$5D,$F1,$F5,$61,$E9,$CD,$F9,$77,$A3,$0E,$10,$F1,$3F,$C9,$3F
	.dc.b	$67,$FD,$33,$F3,$A7,$EB,$DF,$ED,$8F,$C8,$1F,$C0,$3F,$E4,$FF,$96
	.dc.b	$DF,$07,$F8,$8A,$F7,$F8,$B2,$F7,$EA,$BE,$85,$F3,$1C,$30,$61,$C1
	.dc.b	$FC,$4E,$9E,$1D,$1F,$75,$F8,$98,$C6,$7D,$75,$67,$FE,$49,$E4,$FF
	.dc.b	$36,$E3,$E3,$C7,$D1,$5A,$F9,$5F,$C7,$E3,$CB,$6F,$2E,$CF,$DB,$FC
	.dc.b	$9B,$7A,$A7,$3C,$DE,$7F,$96,$FE,$3B,$FB,$7B,$F5,$BF,$E9,$3F,$B8
	.dc.b	$BF,$5D,$FF,$49,$F0,$F8,$B8,$7F,$BB,$FC,$2B,$0E,$1F,$26,$C3,$87
	.dc.b	$D2,$F5,$AD,$5F,$C2,$3F,$C9,$B9,$FF,$2C,$FD,$AF,$E1,$46,$27,$A6
	.dc.b	$9E,$2F,$C1,$FD,$B3,$9C,$FB,$E5,$FE,$F0,$8E,$D1,$8F,$D7,$5F,$AC
	.dc.b	$63,$0F,$5C,$3C,$4C,$6D,$E7,$87,$E4,$58,$C9,$1F,$DB,$78,$3F,$DE
	.dc.b	$99,$FF,$99,$3E,$A9,$F9,$EB,$ED,$FF,$DD,$FF,$AE,$BF,$5A,$FF,$39
	.dc.b	$FC,$87,$FA,$63,$F1,$3F,$F0,$DE,$1C,$2F,$75,$FE,$78,$FD,$3F,$59
	.dc.b	$75,$CB,$FD,$C8,$FF,$10,$C3,$8C,$77,$8C,$1B,$BE,$7F,$80,$FF,$10
	.dc.b	$AD,$C7,$6F,$B1,$C5,$6F,$EE,$1A,$F9,$FF,$20,$7E,$E3,$E2,$AF,$F2
	.dc.b	$FE,$7E,$DB,$DE,$AB,$7E,$B9,$FC,$2B,$E0,$3F,$C6,$FF,$B2,$3E,$D7
	.dc.b	$FC,$29,$F3,$ED,$9F,$6F,$F8,$5F,$BA,$E1,$B7,$FA,$3B,$0E,$1F,$49
	.dc.b	$FD,$F2,$F7,$D5,$F9,$3F,$67,$F9,$45,$EF,$3F,$F1,$03,$FC,$44,$E6
	.dc.b	$30,$F1,$7E,$C7,$98,$CC,$76,$D3,$C9,$A7,$F5,$D9,$CD,$FA,$32,$F3
	.dc.b	$FE,$35,$FE,$A2,$CB,$C3,$5D,$6B,$56,$FC,$ED,$FB,$87,$2D,$BD,$53
	.dc.b	$9F,$D9,$3F,$97,$F5,$7F,$57,$FB,$77,$ED,$CF,$D4,$9F,$84,$BF,$75
	.dc.b	$FF,$2B,$72,$70,$DF,$FA,$9E,$1C,$2F,$75,$FE,$5B,$D3,$55,$FD,$22
	.dc.b	$9F,$84,$7C,$1E,$1F,$CC,$CE,$73,$C6,$9E,$37,$C5,$A7,$87,$E7,$FE
	.dc.b	$04,$E7,$4E,$F8,$E3,$0F,$1E,$1C,$34,$FF,$36,$7C,$1F,$8B,$E4,$3F
	.dc.b	$D0,$B3,$9F,$B9,$FE,$0D,$AF,$8A,$73,$97,$AA,$5F,$63,$C7,$7F,$76
	.dc.b	$FF,$0E,$7F,$C8,$58,$7F,$11,$FA,$BF,$F6,$FF,$EC,$AF,$D1,$FD,$FF
	.dc.b	$99,$3F,$00,$FC,$47,$C1,$7F,$A2,$AF,$75,$A9,$FF,$62,$61,$CF,$F0
	.dc.b	$E6,$3F,$1F,$0E,$10,$FB,$3C,$72,$7F,$B9,$32,$E7,$CA,$D3,$ED,$4F
	.dc.b	$F6,$CC,$38,$E3,$E5,$AD,$7C,$E7,$E2,$E7,$3A,$F8,$FF,$5D,$E1,$87
	.dc.b	$5D,$3C,$73,$E7,$FC,$37,$FB,$BC,$FF,$A9,$3F,$3C,$7E,$A2,$FD,$B1
	.dc.b	$E2,$FF,$81,$7D,$F7,$CE,$BF,$63,$E3,$8F,$FB,$F3,$1E,$38,$70,$F9
	.dc.b	$FF,$E7,$56,$FD,$FB,$76,$BD,$BF,$AF,$E7,$FF,$00,$47,$8B,$F5,$5F
	.dc.b	$C7,$4F,$97,$5F,$26,$BE,$38,$FD,$51,$9C,$FC,$1A,$FA,$23,$BE,$1E
	.dc.b	$4A,$F5,$67,$FD,$DB,$2E,$F8,$F4,$D3,$CF,$F4,$FE,$6F,$AE,$7E,$8D
	.dc.b	$FD,$D7,$EF,$7F,$C8,$5E,$33,$F1,$67,$EA,$3B,$AD,$7D,$5F,$EB,$E5
	.dc.b	$FB,$1E,$7D,$1F,$5F,$10,$FF,$34,$DF,$F8,$05,$F8,$47,$FB,$5B,$3F
	.dc.b	$0F,$CA,$7F,$AC,$71,$FD,$D7,$C1,$FC,$0A,$E5,$5A,$D7,$CB,$EB,$66
	.dc.b	$7A,$E1,$C7,$6F,$47,$3F,$9E,$31,$B3,$A3,$18,$FB,$A1,$FA,$D6,$35
	.dc.b	$F6,$7A,$FF,$6B,$F1,$97,$E8,$0F,$D9,$DF,$84,$FE,$C3,$F1,$A7,$CD
	.dc.b	$3E,$57,$85,$EE,$B5,$2F,$C6,$33,$87,$C6,$E3,$65,$DB,$F5,$7C,$3E
	.dc.b	$09,$EF,$EA,$B4,$7D,$07,$C1,$AD,$67,$E3,$F0,$71,$AF,$87,$0E,$1A
	.dc.b	$BF,$41,$F0,$DF,$93,$ED,$BF,$65,$F8,$6F,$D7,$DD,$F9,$1F,$F7,$FF
	.dc.b	$BD,$FF,$5F,$7E,$5B,$FE,$15,$FD,$03,$FD,$4B,$EB,$FF,$AA,$AB,$5F
	.dc.b	$86,$AB,$5F,$A5,$74,$C7,$F0,$55,$FE,$09,$FB,$FF,$E9,$98,$71,$FA
	.dc.b	$0F,$F4,$BD,$7F,$53,$8C,$67,$E0,$F8,$FF,$F2,$94,$F3,$3C,$B5,$FE
	.dc.b	$9E,$CF,$C1,$3F,$E1,$29,$D6,$3E,$4C,$FD,$1F,$80,$E3,$0E,$11,$EB
	.dc.b	$FC,$17,$A7,$E4,$8F,$D7,$DF,$99,$FF,$9E,$7C,$77,$E6,$4F,$D9,$FF
	.dc.b	$DF,$5C,$3F,$EC,$AC,$38,$5E,$EB,$5F,$8A,$74,$C5,$F4,$7F,$03,$9F
	.dc.b	$E7,$61,$C3,$87,$E0,$46,$2B,$C3,$F7,$7F,$E6,$7C,$5F,$A3,$FE,$AF
	.dc.b	$4F,$05,$6B,$6F,$2F,$E8,$F9,$C7,$96,$DE,$3D,$DD,$3E,$38,$C7,$E6
	.dc.b	$B7,$5E,$D4,$FD,$B7,$19,$7C,$08,$7E,$33,$F4,$7F,$6B,$FE,$E1,$F9
	.dc.b	$07,$E9,$3E,$CD,$F1,$3F,$C1,$BD,$CD,$7C,$AA,$B5,$3F,$C2,$BE,$77
	.dc.b	$3F,$37,$F0,$FA,$3F,$83,$65,$D1,$F8,$66,$30,$E5,$E5,$B3,$EF,$EE
	.dc.b	$7B,$EB,$59,$F6,$F7,$B1,$0C,$7C,$F8,$71,$D3,$F8,$DF,$E8,$DE,$6C
	.dc.b	$3C,$BF,$04,$FD,$1F,$F1,$DF,$BE,$FB,$3F,$51,$7E,$3C,$FA,$1F,$E4
	.dc.b	$7F,$BA,$FF,$B0,$BC,$17,$7F,$EA,$70,$E1,$7B,$C9,$7F,$D7,$EF,$7C
	.dc.b	$1F,$C9,$6F,$E4,$F9,$5F,$0E,$11,$E1,$7E,$19,$8D,$DA,$79,$34,$FE
	.dc.b	$6B,$EB,$7C,$0B,$F0,$C7,$F3,$8F,$AC,$C5,$F0,$D7,$C9,$A3,$F7,$DF
	.dc.b	$7C,$E6,$FD,$B7,$FD,$D7,$C9,$FE,$09,$E7,$7E,$8B,$F2,$87,$E7,$0F
	.dc.b	$C9,$FF,$49,$F8,$1B,$F3,$4E,$1C,$5F,$C0,$3C,$E6,$1C,$2F,$7D,$DA
	.dc.b	$AF,$E1,$F5,$A7,$F8,$16,$7E,$08,$FF,$48,$46,$0E,$97,$CF,$C5,$FE
	.dc.b	$47,$1C,$F4,$F3,$69,$E5,$9F,$F0,$0F,$3F,$EC,$AB,$8C,$E7,$5F,$F3
	.dc.b	$04,$E6,$7C,$78,$75,$9F,$F0,$34,$E7,$0E,$BA,$7C,$0F,$C8,$3C,$19
	.dc.b	$FB,$B2,$E9,$FA,$57,$B7,$F5,$07,$EA,$EF,$67,$FC,$DD,$F9,$3B,$F4
	.dc.b	$37,$E0,$7F,$D4,$B6,$F1,$61,$C2,$F7,$5A,$D3,$A6,$2F,$AD,$FE,$57
	.dc.b	$C9,$FC,$E7,$0E,$1C,$7F,$1C,$3C,$FE,$D5,$AF,$87,$D8,$FE,$19,$7C
	.dc.b	$71,$F3,$56,$B3,$FE,$BF,$8F,$83,$4F,$1E,$DF,$D8,$F6,$F3,$E1,$C3
	.dc.b	$3F,$64,$BF,$60,$FB,$DF,$89,$3E,$9F,$F0,$DF,$42,$E2,$7E,$58,$FC
	.dc.b	$D3,$F9,$96,$9F,$4F,$0E,$1D,$DF,$E2,$BB,$DD,$7F,$4F,$AD,$63,$FE
	.dc.b	$A7,$DF,$E0,$61,$C1,$B1,$E0,$BE,$FD,$FF,$02,$7F,$C2,$D5,$DB,$3F
	.dc.b	$26,$7F,$88,$E3,$FE,$C9,$39,$C3,$C3,$FA,$E7,$B3,$2E,$97,$E9,$F3
	.dc.b	$7F,$39,$EF,$FF,$92,$BD,$BF,$F0,$27,$C0,$3F,$53,$7E,$0B,$F6,$78
	.dc.b	$7F,$21,$7B,$F9,$65,$EE,$BF,$11,$5A,$FA,$56,$3F,$80,$E2,$FF,$17
	.dc.b	$63,$FE,$58,$FC,$73,$11,$FD,$6E,$B7,$97,$4C,$DF,$5D,$F0,$6B,$58
	.dc.b	$F9,$A7,$FC,$E7,$97,$78,$C6,$6F,$C5,$7D,$B9,$74,$D7,$CD,$F1,$D8
	.dc.b	$BD,$FE,$6D,$75,$F0,$63,$F0,$72,$FE,$DB,$A7,$F9,$17,$F3,$BF,$F5
	.dc.b	$CF,$D5,$7F,$9C,$FF,$3B,$7E,$A8,$FC,$71,$FA,$8F,$FA,$1F,$E3,$7F
	.dc.b	$C2,$D8,$70,$FA,$86,$1C,$2D,$FA,$66,$F3,$ED,$F7,$AF,$CC,$BF,$C2
	.dc.b	$39,$72,$8F,$F4,$1C,$FF,$84,$63,$1A,$F8,$F5,$7E,$B3,$FD,$C6,$73
	.dc.b	$A7,$82,$7F,$79,$7D,$C6,$BE,$5D,$7C,$9F,$BE,$E6,$31,$5E,$DF,$D3
	.dc.b	$B8,$6E,$F0,$5F,$90,$BC,$4F,$46,$FF,$E4,$6F,$8A,$DD,$6A,$73,$F4
	.dc.b	$9F,$E8,$FE,$6F,$9B,$BF,$C9,$A7,$E4,$5F,$02,$3F,$CE,$99,$F7,$CF
	.dc.b	$11,$B7,$C1,$CB,$FE,$A6,$C7,$E0,$4F,$0C,$3C,$13,$FE,$AA,$FD,$5F
	.dc.b	$33,$C7,$3F,$1F,$D7,$F9,$B1,$E7,$B7,$F9,$7B,$E2,$FF,$D8,$1F,$95
	.dc.b	$FF,$93,$FF,$5D,$64,$F9,$D7,$EF,$0F,$DE,$DF,$A8,$3F,$74,$FE,$51
	.dc.b	$CA,$CB,$96,$3C,$70,$E1,$7B,$FD,$DF,$BF,$0F,$FA,$4F,$4E,$FD,$7F
	.dc.b	$E5,$2C,$78,$EB,$6B,$59,$73,$D7,$C3,$F8,$F7,$F8,$C6,$F7,$A7,$6F
	.dc.b	$A7,$C6,$1C,$31,$F4,$E1,$E8,$DD,$FB,$77,$8A,$9A,$FC,$0B,$DE,$FE
	.dc.b	$DF,$D5,$B5,$39,$CF,$D7,$8F,$FB,$DF,$87,$E0,$63,$DE,$73,$F8,$8F
	.dc.b	$5F,$CA,$5F,$85,$FF,$31,$FC,$AD,$FB,$57,$E9,$BF,$7F,$1B,$9D,$46
	.dc.b	$7E,$83,$1F,$07,$F0,$F3,$63,$D3,$EE,$B0,$F6,$96,$5D,$BC,$9F,$E8
	.dc.b	$FB,$79,$B0,$E3,$9F,$69,$FF,$2D,$C7,$6A,$E1,$A3,$F6,$DC,$6B,$E7
	.dc.b	$E6,$64,$FC,$23,$FD,$2D,$F0,$DF,$CA,$BE,$EB,$D7,$F9,$17,$FB,$EB
	.dc.b	$F5,$37,$EB,$BC,$2F,$7E,$BA,$F7,$F8,$62,$DF,$91,$6F,$87,$FD,$B1
	.dc.b	$9F,$57,$FD,$29,$8F,$2F,$C7,$BF,$D3,$2F,$4C,$79,$FE,$79,$FE,$75
	.dc.b	$8C,$71,$36,$68,$FA,$DF,$F3,$D7,$E7,$7F,$EA,$FF,$D7,$BF,$7F,$FC
	.dc.b	$00,$FC,$29,$F8,$CB,$F3,$87,$EE,$AF,$47,$E1,$E3,$C7,$D2,$63,$C7
	.dc.b	$0E,$17,$FC,$7F,$A5,$7B,$AF,$F5,$8B,$C7,$F8,$02,$F2,$FF,$BD,$F5
	.dc.b	$F1,$EB,$E2,$D7,$F9,$0C,$F8,$0D,$69,$E0,$9D,$7C,$DF,$B2,$73,$97
	.dc.b	$8E,$FD,$3D,$DC,$65,$8F,$6D,$BC,$FD,$BF,$7B,$E3,$BF,$DA,$5F,$86
	.dc.b	$3F,$52,$7E,$91,$FE,$1F,$EF,$FF,$0C,$F8,$6F,$D6,$78,$68,$18,$71
	.dc.b	$BE,$1F,$3D,$F9,$6F,$CE,$37,$5A,$C9,$F8,$EF,$FB,$53,$4F,$0C,$78
	.dc.b	$27,$FD,$A9,$18,$C3,$9F,$E6,$9F,$EB,$38,$C7,$D3,$FE,$2F,$27,$EC
	.dc.b	$5F,$EB,$7F,$D9,$3F,$DD,$9F,$98,$BF,$50,$7E,$EE,$F8,$27,$E9,$FF
	.dc.b	$9D,$F1,$DF,$BD,$7E,$83,$1E,$38,$70,$BD,$F9,$3F,$BA,$FD,$53,$5C
	.dc.b	$5F,$07,$4E,$DA,$7F,$BB,$B1,$E3,$F2,$96,$B5,$97,$3D,$7F,$59,$F8
	.dc.b	$B7,$9E,$73,$87,$96,$7F,$61,$F5,$FF,$ED,$1F,$CE,$7E,$75,$3A,$5F
	.dc.b	$55,$FE,$DE,$F9,$E7,$E1,$FF,$D6,$FE,$03,$FC,$77,$87,$0B,$DF,$D5
	.dc.b	$FF,$86,$AB,$5E,$EF,$EF,$AB,$58,$F3,$9F,$EC,$9F,$13,$E1,$CF,$EF
	.dc.b	$63,$CB,$3E,$FF,$24,$F8,$CF,$F7,$37,$EC,$9E,$6F,$C7,$FF,$CE,$3F
	.dc.b	$9F,$3F,$0E,$FA,$1F,$7D,$F9,$2F,$9F,$FE,$1B,$7B,$FB,$45,$EF,$BF
	.dc.b	$FE,$00,$7A,$1F,$E8,$6A,$D4,$E7,$0F,$F2,$87,$B2,$9C,$E5,$DB,$F1
	.dc.b	$DF,$F7,$C7,$EB,$DF,$23,$E5,$9F,$9F,$3F,$5A,$FF,$55,$76,$3F,$7F
	.dc.b	$FF,$32,$7D,$DF,$F2,$0C,$38,$5E,$FF,$00,$5E,$FE,$E3,$FD,$2B,$5A
	.dc.b	$C7,$9E,$2F,$D5,$7C,$3D,$7F,$A5,$D3,$C3,$5A,$FB,$AF,$D4,$70,$BF
	.dc.b	$53,$FB,$BF,$BD,$BF,$17,$FF,$44,$FE,$81,$FE,$50,$FC,$99,$FA,$0F
	.dc.b	$F8,$C3,$0E,$1A,$7E,$C7,$0E,$1F,$82,$30,$E1,$6C,$3F,$9D,$E5,$DB
	.dc.b	$2E,$BE,$16,$6B,$C7,$39,$C3,$94,$F1,$FC,$67,$CF,$FB,$17,$FA,$7F
	.dc.b	$EC,$3F,$96,$FA,$EF,$55,$FA,$B3,$F1,$D7,$D6,$7F,$4A,$C3,$87,$CA
	.dc.b	$FE,$BA,$F7,$4F,$AD,$FE,$9F,$F4,$FF,$9E,$F9,$07,$6B,$F6,$67,$EE
	.dc.b	$3F,$4B,$C9,$FF,$1B,$7E,$7B,$C3,$AF,$F8,$96,$1C,$2F,$7E,$1F,$E3
	.dc.b	$3E,$C5,$FA,$47,$F5,$CF,$C1,$7F,$57,$E3,$BF,$43,$F5,$79,$1F,$EA
	.dc.b	$5F,$89,$FE,$B7,$FB,$33,$F5,$AF,$F0,$4E,$DD,$B8,$BC,$9C,$4F,$C5
	.dc.b	$1F,$AE,$3F,$69,$FF,$16,$E1,$C3,$0F,$F6,$96,$1C,$2F,$74,$FC,$0B
	.dc.b	$B7,$67,$C6,$EF,$FF,$6C,$73,$7F,$97,$3D,$CC,$E5,$FC,$BF,$C4,$78
	.dc.b	$5F,$4A,$BF,$DB,$DF,$80,$BF,$6B,$F4,$78,$FF,$6F,$CB,$FE,$11,$F9
	.dc.b	$8B,$F5,$0F,$F6,$27,$1F,$E2,$B4,$79,$DF,$8A,$BD,$D7,$F3,$DA,$D4
	.dc.b	$E7,$E0,$58,$3F,$08,$D6,$5D,$E7,$3F,$B5,$71,$E7,$FD,$37,$73,$F4
	.dc.b	$2F,$DC,$7E,$B2,$FD,$1F,$E7,$56,$A7,$3F,$A8,$63,$87,$F5,$5E,$E7
	.dc.b	$EC,$E7,$F8,$06,$5D,$67,$3E,$CB,$3E,$DF,$00,$FD,$73,$D4,$FF,$06
	.dc.b	$F8,$47,$C7,$BF,$76,$7E,$CD,$F7,$EB,$53,$9F,$DA,$31,$9F,$F5,$FF
	.dc.b	$8B,$FE,$3B,$F0,$FF,$D2,$E9,$B7,$6A,$D7,$C5,$7F,$1C,$AD,$74,$7C
	.dc.b	$BF,$90,$FE,$5F,$F2,$5F,$CC,$FC,$07,$E7,$AF,$B7,$FC,$85,$6B,$E0
	.dc.b	$BF,$9E,$D6,$BE,$DF,$3A,$7F,$34,$F9,$4F,$E7,$12,$F8,$EE,$D3,$9C
	.dc.b	$FB,$7E,$59,$F6,$FE,$73,$FA,$EF,$CC,$FF,$4F,$F9,$E7,$E6,$DF,$E7
	.dc.b	$DF,$25,$5A,$EC,$FE,$C3,$2F,$C4,$BF,$49,$3C,$B3,$FE,$AA,$75,$3F
	.dc.b	$AF,$FC,$FE,$1C,$1F,$C1,$FE,$7F,$FB,$16,$5F,$D2,$63,$99,$FD,$8B
	.dc.b	$1F,$ED,$FF,$3F,$F9,$1F,$A4,$FE,$FF,$F5,$8F,$A5,$7E,$C9,$FC,$77
	.dc.b	$7F,$F7,$6E,$3F,$E1,$3F,$AB,$FF,$09,$7C,$17,$F8,$16,$BF,$A3,$56
	.dc.b	$AB,$FC,$25,$F5,$7E,$57,$F7,$D9,$7F,$83,$DF,$44,$EF,$0E,$79,$78
	.dc.b	$7E,$A7,$C9,$5D,$FE,$AF,$FC,$4F,$9F,$F8,$4F,$EA,$BF,$51,$F6,$DF
	.dc.b	$E1,$23,$F0,$BF,$05,$FE,$13,$FB,$77,$EB,$6F,$DD,$FF,$87,$56,$FC
	.dc.b	$9F,$E0,$9C,$FF,$E5,$2A,$7E,$C3,$AD,$7F,$C2,$17,$FC,$95,$F3,$BC
	.dc.b	$5A,$6A,$F9,$E2,$FD,$77,$CB,$39,$CF,$C9,$BE,$23,$F9,$E6,$1E,$8C
	.dc.b	$FF,$CC,$76,$FD,$87,$C5,$5A,$CB,$AC,$FF,$6B,$E3,$FF,$30,$CF,$5B
	.dc.b	$FF,$35,$FE,$75,$D7,$E6,$FF,$CC,$B8,$70,$E0,$E7,$18,$D7,$E3,$64
	.dc.b	$FD,$07,$F9,$97,$6F,$F9,$FB,$F0,$1F,$F4,$86,$F8,$D7,$FD,$13,$F8
	.dc.b	$2F,$FA,$3F,$ED,$7F,$D0,$5F,$41,$FE,$8F,$FD,$D3,$54,$FC,$17,$FD
	.dc.b	$1C,$FE,$63,$8F,$FD,$3C,$F2,$F4,$7F,$09,$18,$8E,$18,$F5,$CF,$FD
	.dc.b	$1B,$F8,$2F,$F2,$5D,$7C,$99,$FF,$A3,$EF,$BF,$CB,$3F,$33,$7E,$1D
	.dc.b	$FE,$69,$FC,$9B,$FD,$49,$F1,$EF,$C5,$F7,$5A,$FB,$E5,$6B,$F1,$5C
	.dc.b	$FE,$75,$E4,$D7,$FC,$87,$9F,$5C,$F1,$A7,$F9,$CE,$31,$8F,$2F,$67
	.dc.b	$F0,$13,$DA,$BC,$50,$EE,$F5,$4F,$1D,$3C,$5A,$79,$BC,$6E,$18,$F3
	.dc.b	$DD,$F8,$F7,$F6,$5F,$C6,$7F,$D8,$BF,$AB,$BF,$4D,$FF,$59,$7D,$23
	.dc.b	$ED,$BF,$97,$FE,$03,$BD,$BF,$D7,$F7,$BE,$9F,$EE,$55,$FC,$D3,$F5
	.dc.b	$6D,$57,$D7,$AE,$8F,$ED,$32,$EF,$F6,$5F,$59,$F0,$E7,$E0,$C3,$FD
	.dc.b	$BF,$DB,$FE,$E3,$EF,$63,$DF,$5F,$27,$57,$F5,$DC,$78,$EB,$E2,$F8
	.dc.b	$97,$EE,$BE,$2C,$78,$E9,$D6,$5F,$B1,$7F,$81,$7D,$EF,$F3,$AF,$E1
	.dc.b	$EF,$83,$7E,$9B,$FD,$CF,$E2,$1F,$A7,$BE,$2B,$F6,$5C,$7F,$CA,$F9
	.dc.b	$BF,$D5,$58,$F1,$C3,$85,$EF,$F7,$CE,$D5,$AC,$BA,$BF,$C5,$B3,$F0
	.dc.b	$C7,$F8,$4E,$30,$6E,$F9,$FD,$D7,$F7,$BA,$DC,$76,$F7,$F1,$5B,$F8
	.dc.b	$B6,$DE,$8D,$9F,$A4,$F8,$A9,$B7,$9E,$F7,$B7,$B7,$F3,$0D,$4E,$71
	.dc.b	$E5,$FB,$A6,$37,$F8,$B8,$70,$CF,$9F,$0F,$C3,$CF,$FC,$59,$F8,$F3
	.dc.b	$E3,$9F,$AD,$3E,$31,$FB,$6B,$B5,$F9,$6F,$97,$EC,$D8,$F1,$C3,$85
	.dc.b	$EF,$F7,$ED,$4F,$5C,$FE,$39,$E0,$FC,$2B,$F8,$0F,$4F,$FA,$C3,$1E
	.dc.b	$7A,$F8,$B5,$FE,$67,$7F,$20,$F3,$C6,$72,$FE,$A7,$B7,$AE,$F8,$E9
	.dc.b	$E2,$DB,$FD,$A5,$39,$D3,$CD,$97,$B2,$5F,$9C,$FC,$FF,$A2,$FF,$48
	.dc.b	$FD,$17,$D4,$FC,$B9,$F9,$97,$F8,$AF,$E8,$75,$F7,$FF,$E1,$FE,$A5
	.dc.b	$6B,$E9,$FF,$C3,$BB,$7C,$DC,$BF,$D3,$7F,$B2,$FB,$B9,$3F,$DF,$51
	.dc.b	$D9,$F8,$6F,$83,$0E,$79,$BF,$13,$F8,$13,$9C,$BC,$7F,$89,$E3,$4F
	.dc.b	$14,$78,$5F,$2E,$DE,$9C,$3F,$C8,$5A,$FF,$9B,$7B,$7F,$DB,$BF,$A5
	.dc.b	$3F,$3E,$7B,$3F,$73,$F6,$9F,$D9,$FD,$57,$C0,$E3,$FD,$77,$10,$61
	.dc.b	$C6,$F8,$7A,$1F,$EC,$5F,$57,$B5,$6E,$B4,$FF,$7E,$CF,$C3,$9F,$F1
	.dc.b	$C8,$C1,$B7,$C7,$E1,$34,$FE,$0B,$E6,$7F,$99,$6F,$8E,$7E,$0F,$C3
	.dc.b	$71,$8F,$18,$F1,$ED,$FE,$87,$DB,$D3,$3E,$8C,$FA,$E5,$EE,$F3,$7C
	.dc.b	$F3,$9C,$38,$FD,$CE,$36,$F7,$6D,$E3,$E0,$FF,$71,$7A,$7C,$3F,$70
	.dc.b	$79,$1F,$E5,$CF,$D8,$3F,$C1,$3A,$7F,$9C,$7F,$6D,$D4,$E7,$E6,$FF
	.dc.b	$8B,$F3,$FF,$A4,$F8,$BC,$B8,$F2,$FB,$0F,$EC,$7C,$DF,$3F,$89,$3D
	.dc.b	$BF,$1F,$CC,$7D,$38,$E1,$AF,$83,$C0,$F8,$73,$C3,$4F,$06,$DF,$E4
	.dc.b	$1C,$FB,$C6,$34,$7E,$6B,$E6,$D5,$97,$AA,$5F,$85,$FC,$6F,$CA,$9F
	.dc.b	$9D,$FF,$38,$FA,$D7,$E8,$BF,$EB,$4F,$B6,$FE,$69,$AF,$D0,$BD,$F3
	.dc.b	$FF,$98,$2B,$5E,$9F,$3C,$7F,$A0,$F0,$FE,$08,$5F,$6F,$7B,$DF,$18
	.dc.b	$9F,$05,$FF,$B0,$BF,$1A,$F7,$DF,$6D,$3C,$7F,$85,$FF,$63,$C3,$B6
	.dc.b	$7E,$18,$7E,$1B,$FE,$29,$E9,$F9,$D8,$71,$CB,$FD,$9B,$3F,$E3,$1D
	.dc.b	$BF,$13,$4F,$AB,$7E,$0C,$D7,$FA,$7F,$E9,$5F,$E1,$FF,$CB,$FF,$17
	.dc.b	$7B,$AD,$7C,$12,$BF,$C4,$9F,$80,$EB,$1F,$D8,$F0,$7C,$1C,$38,$36
	.dc.b	$3B,$E5,$DB,$5F,$BB,$5D,$63,$C3,$F7,$18,$AD,$E5,$DF,$C3,$FF,$47
	.dc.b	$5F,$E2,$CE,$11,$9F,$9A,$F1,$6B,$E6,$D7,$ED,$FD,$5E,$33,$F0,$7E
	.dc.b	$3B,$70,$F5,$7E,$5B,$FE,$D7,$FD,$37,$F2,$DF,$22,$FB,$47,$CA,$FF
	.dc.b	$B0,$FE,$B9,$F2,$CF,$C6,$DF,$98,$7F,$BF,$F1,$E3,$87,$0F,$6F,$FD
	.dc.b	$7B,$7B,$F9,$7D,$74,$FF,$11,$D3,$B6,$9F,$E4,$8C,$78,$FC,$CD,$AD
	.dc.b	$65,$CF,$5F,$0F,$D9,$79,$E7,$38,$79,$73,$7D,$AE,$65,$9F,$6D,$7C
	.dc.b	$7B,$BF,$6E,$F1,$D6,$B4,$ED,$8F,$B7,$1C,$F9,$2F,$A5,$FC,$47,$E3
	.dc.b	$5F,$E4,$1F,$83,$7E,$A7,$FE,$D1,$F1,$5C,$9F,$BF,$C3,$87,$82,$C3
	.dc.b	$85,$BF,$54,$EB,$97,$E4,$F6,$FF,$54,$E3,$C6,$1F,$A4,$FC,$39,$65
	.dc.b	$D3,$1F,$17,$E9,$FE,$F8,$C7,$D4,$36,$FE,$A1,$7D,$B5,$F1,$4E,$75
	.dc.b	$FF,$7D,$6D,$E8,$AF,$F7,$B6,$9E,$EA,$D7,$EF,$9F,$1B,$F3,$6F,$0F
	.dc.b	$9E,$FE,$00,$7E,$83,$FE,$0D,$FD,$65,$FB,$D6,$F8,$3F,$C8,$75,$AF
	.dc.b	$97,$57,$D6,$DB,$C1,$B3,$3F,$6F,$AF,$EF,$7B,$C7,$C4,$3F,$AC,$3E
	.dc.b	$2C,$3A,$7E,$19,$F7,$E3,$1A,$7F,$87,$FC,$2C,$69,$E2,$8F,$37,$EB
	.dc.b	$3E,$5D,$FD,$51,$E9,$FC,$0F,$19,$79,$F6,$FF,$07,$F5,$7F,$39,$CF
	.dc.b	$AF,$07,$AB,$E6,$7F,$DE,$5B,$FF,$6B,$F3,$9F,$72,$FE,$01,$7C,$03
	.dc.b	$F1,$F7,$E7,$CF,$CC,$DF,$9D,$F2,$CB,$F8,$66,$3C,$70,$E1,$7B,$FD
	.dc.b	$83,$59,$F6,$CF,$FD,$E5,$F3,$9F,$D3,$71,$E3,$FB,$7E,$1F,$EA,$15
	.dc.b	$BC,$79,$FE,$1F,$FE,$01,$BE,$13,$9D,$BC,$B1,$9B,$E3,$A7,$8E,$7D
	.dc.b	$3B,$3F,$19,$F1,$CE,$76,$F6,$7E,$E9,$EB,$FD,$DF,$FD,$E3,$FA,$43
	.dc.b	$F6,$A7,$ED,$6F,$53,$FA,$8F,$8B,$BF,$97,$5E,$EB,$5F,$1B,$FB,$CE
	.dc.b	$1F,$76,$5D,$38,$7E,$01,$C7,$F7,$9F,$80,$FF,$87,$63,$C3,$A7,$83
	.dc.b	$47,$E7,$1E,$DA,$D6,$9E,$0A,$D7,$E9,$99,$96,$1C,$B4,$F1,$F8,$5E
	.dc.b	$7D,$BC,$D9,$FF,$6A,$F1,$7E,$D7,$C0,$7F,$92,$BF,$4A,$7E,$2A,$FC
	.dc.b	$C5,$2F,$DF,$9F,$86,$70,$FA,$C6,$1C,$2F,$74,$FD,$63,$5F,$A7,$7B
	.dc.b	$B8,$79,$E1,$C1,$F8,$7C,$FF,$83,$CF,$C5,$A7,$85,$FE,$07,$19,$AE
	.dc.b	$F3,$9F,$6B,$E0,$57,$0C,$FC,$3F,$A2,$F9,$75,$F2,$4F,$A7,$F6,$DC
	.dc.b	$61,$D7,$6F,$CF,$E7,$D6,$F7,$1F,$C4,$E1,$FA,$B7,$F8,$8F,$F0,$97
	.dc.b	$E6,$8F,$CB,$DF,$45,$FE,$98,$F8,$E7,$EA,$AF,$DA,$19,$72,$E4,$FF
	.dc.b	$2D,$65,$CB,$1E,$3F,$7A,$C7,$8F,$27,$0D,$7C,$1A,$F7,$C7,$E3,$C6
	.dc.b	$3E,$D3,$F5,$6E,$B1,$E7,$83,$C0,$FD,$44,$E7,$7F,$37,$C9,$63,$1F
	.dc.b	$1E,$5E,$B9,$F5,$70,$7E,$FF,$7F,$45,$37,$F4,$56,$BC,$FF,$6F,$D3
	.dc.b	$7E,$E3,$F4,$F7,$ED,$FF,$07,$E9,$DF,$9A,$3F,$58,$7E,$31,$C3,$4F
	.dc.b	$E2,$58,$70,$E8,$61,$C2,$DF,$AD,$6E,$FF,$C3,$57,$D9,$FC,$4E,$7E
	.dc.b	$1F,$C8,$FF,$86,$BE,$3D,$3C,$5A,$3F,$11,$FE,$15,$39,$D3,$BC,$B6
	.dc.b	$CE,$3E,$4C,$38,$E8,$F1,$FC,$5E,$CB,$5F,$46,$FE,$7F,$2F,$CD,$F5
	.dc.b	$CF,$B5,$7E,$72,$FD,$81,$F8,$9B,$C1,$FD,$6F,$C7,$70,$BD,$D6,$A7
	.dc.b	$F6,$3F,$13,$9F,$CB,$FD,$BC,$5F,$EA,$3C,$FB,$67,$88,$FF,$33,$E3
	.dc.b	$CB,$3E,$DE,$5F,$B2,$73,$5E,$39,$F1,$7E,$3F,$FC,$B7,$F5,$AF,$F4
	.dc.b	$07,$E8,$9F,$7F,$F1,$B7,$DB,$3F,$80,$7F,$A2,$7E,$A3,$FA,$2F,$15
	.dc.b	$87,$0F,$15,$87,$0B,$DF,$E2,$BF,$A2,$AD,$7E,$39,$FD,$09,$E1,$57
	.dc.b	$7F,$F5,$F6,$9E,$1D,$3C,$1F,$15,$FA,$E9,$E4,$6B,$3E,$F3,$AF,$43
	.dc.b	$D5,$39,$C3,$C5,$6F,$DE,$5F,$88,$72,$A6,$5D,$F6,$FE,$C1,$4F,$E0
	.dc.b	$1D,$4E,$70,$E5,$F7,$B8,$D7,$C3,$87,$0E,$6F,$DF,$FE,$71,$FE,$A9
	.dc.b	$FC,$09,$FB,$47,$F0,$2B,$7E,$84,$FC,$DD,$F7,$DF,$E7,$5E,$86,$3C
	.dc.b	$70,$E1,$7B,$A7,$E4,$1D,$B8,$BF,$DC,$99,$F6,$FC,$3B,$FB,$6E,$3C
	.dc.b	$72,$B5,$AC,$B9,$E7,$E0,$8E,$43,$0F,$3D,$7E,$FF,$3F,$37,$E0,$3F
	.dc.b	$DC,$7E,$23,$FD,$89,$F8,$EB,$F5,$17,$ED,$1F,$D4,$3D,$27,$E7,$BF
	.dc.b	$D0,$7F,$33,$FE,$43,$87,$0B,$DF,$C8,$FE,$BE,$BF,$A3,$7C,$A6,$B7
	.dc.b	$FF,$7F,$67,$D7,$F6,$2F,$F0,$3E,$3C,$74,$B5,$AC,$B9,$E9,$E0,$AF
	.dc.b	$34,$E6,$BC,$99,$7F,$79,$8F,$E6,$7F,$81,$FF,$B7,$7E,$A7,$FD,$9F
	.dc.b	$F9,$33,$F7,$6F,$F1,$E7,$DD,$3F,$52,$7E,$69,$C3,$11,$FE,$4A,$F8
	.dc.b	$16,$1C,$7E,$B3,$FC,$F7,$6F,$C2,$78,$5E,$FF,$33,$FF,$10,$EB,$E1
	.dc.b	$C7,$8C,$FF,$62,$8C,$63,$CF,$3F,$E3,$13,$CB,$5F,$3E,$BE,$6F,$8F
	.dc.b	$FF,$12,$7D,$F7,$FB,$DF,$F1,$D7,$E4,$2F,$DA,$1D,$3F,$E3,$BF,$81
	.dc.b	$FF,$02,$FE,$C7,$FD,$53,$0E,$17,$BA,$D7,$6F,$E5,$BC,$1F,$F4,$06
	.dc.b	$DF,$03,$3E,$BF,$33,$FE,$3A,$C7,$8E,$96,$B5,$97,$3F,$39,$AB,$F7
	.dc.b	$2F,$3D,$6A,$FC,$B5,$AF,$B4,$FF,$1F,$7D,$97,$CE,$A7,$EF,$DF,$E7
	.dc.b	$6F,$DE,$5F,$B5,$BF,$44,$7E,$5C,$FC,$EB,$FD,$CB,$C6,$CB,$96,$3C
	.dc.b	$70,$E1,$7B,$FC,$8B,$FB,$AD,$6B,$EC,$59,$FF,$95,$A3,$1F,$B0,$DD
	.dc.b	$7F,$03,$97,$3D,$7C,$3D,$DE,$89,$CE,$3C,$B4,$EF,$F9,$47,$F9,$67
	.dc.b	$F4,$BF,$67,$E8,$1F,$D2,$FF,$09,$7E,$A3,$FE,$FB,$DF,$D1,$D5,$FE
	.dc.b	$08,$BD,$FC,$C7,$FA,$EE,$B5,$C3,$FE,$53,$E4,$56,$B1,$FF,$36,$BC
	.dc.b	$33,$BB,$7E,$63,$4F,$06,$3D,$FE,$C7,$C9,$39,$CF,$B7,$E3,$5F,$0B
	.dc.b	$F0,$7F,$F3,$87,$EA,$7F,$4F,$F1,$A7,$E1,$AF,$C7,$DF,$A4,$B0,$7D
	.dc.b	$33,$F4,$A7,$DF,$7F,$88,$B2,$E5,$8F,$1C,$38,$5E,$F2,$D6,$BF,$AE
	.dc.b	$7B,$FE,$45,$FE,$88,$C3,$8E,$96,$B5,$8F,$3D,$3C,$3E,$6F,$A2,$73
	.dc.b	$5E,$69,$F2,$FD,$4F,$FA,$2F,$F5,$0F,$F5,$37,$BB,$EA,$FC,$5D,$E8
	.dc.b	$FF,$03,$FC,$99,$FB,$E7,$F8,$EA,$F7,$C5,$FE,$DE,$AD,$7A,$59,$D7
	.dc.b	$FD,$7F,$97,$F6,$87,$FC,$BB,$7C,$1A,$6A,$F9,$7A,$8F,$33,$C9,$39
	.dc.b	$D3,$C7,$2F,$C8,$FF,$D4,$5F,$4F,$FA,$5F,$B2,$7E,$86,$FB,$97,$EE
	.dc.b	$BF,$C5,$BF,$5E,$FF,$44,$FA,$DF,$E5,$2C,$38,$5E,$EB,$5F,$14,$FC
	.dc.b	$7F,$5B,$7F,$1E,$CF,$AE,$7F,$EB,$8C,$78,$FE,$C3,$6B,$59,$73,$5A
	.dc.b	$7F,$A6,$1E,$67,$F9,$92,$35,$1E,$B8,$FF,$32,$FE,$11,$FB,$FE,$77
	.dc.b	$EA,$0F,$D3,$BF,$C6,$9F,$00,$F6,$3F,$C2,$58,$70,$3F,$99,$DF,$0A
	.dc.b	$DC,$BF,$6D,$7E,$31,$BB,$FB,$78,$F5,$7F,$D2,$39,$78,$61,$FB,$17
	.dc.b	$FB,$5A,$31,$8F,$4F,$D0,$3F,$DA,$51,$8E,$D6,$BE,$53,$FD,$AD,$F4
	.dc.b	$7F,$C6,$7F,$71,$7E,$AA,$FD,$47,$FD,$D7,$FB,$57,$F1,$5F,$D1,$9F
	.dc.b	$A4,$FF,$A1,$B0,$E1,$7B,$FC,$61,$7B,$FC,$DF,$5C,$5F,$D4,$B3,$EB
	.dc.b	$F4,$DF,$EE,$CC,$78,$E9,$6B,$59,$73,$C7,$F9,$27,$85,$E6,$9C,$D7
	.dc.b	$93,$DF,$FB,$1E,$FB,$F3,$AF,$E6,$5F,$2A,$FB,$46,$CF,$C4,$3F,$11
	.dc.b	$97,$F3,$EF,$1F,$FD,$3F,$87,$0B,$DD,$6B,$E4,$1F,$BA,$EB,$0F,$ED
	.dc.b	$79,$75,$CB,$FD,$CB,$19,$11,$E1,$FC,$03,$3A,$79,$67,$8E,$5D,$BC
	.dc.b	$9F,$EE,$BF,$A9,$7F,$76,$AD,$57,$2A,$D2,$FC,$D3,$5F,$AA,$7F,$AF
	.dc.b	$7F,$25,$FE,$67,$F4,$5F,$ED,$9F,$D9,$7F,$E0,$07,$C1,$FF,$AE,$AB
	.dc.b	$5F,$11,$FE,$1E,$FC,$6D,$4F,$89,$7E,$39,$BD,$3F,$D7,$F9,$FF,$3E
	.dc.b	$7F,$DB,$F8,$F7,$FD,$4F,$D6,$F4,$D3,$C1,$87,$92,$B8,$E7,$DF,$27
	.dc.b	$C5,$7B,$7E,$F3,$FB,$37,$DB,$FF,$B2,$70,$78,$BF,$D9,$39,$DE,$AF
	.dc.b	$F4,$7B,$DF,$C3,$79,$EB,$52,$F9,$2F,$F6,$BD,$6B,$E7,$BE,$7B,$C3
	.dc.b	$1B,$B7,$F8,$8B,$3F,$06,$3D,$FE,$1B,$C9,$39,$CF,$B7,$CD,$FD,$A7
	.dc.b	$EA,$7F,$00,$7F,$7D,$F8,$A7,$E6,$AF,$C2,$3F,$C6,$1F,$08,$FD,$F9
	.dc.b	$F8,$BB,$F4,$5F,$F7,$5E,$3C,$70,$E1,$F5,$9F,$E3,$9B,$DF,$3E,$B6
	.dc.b	$FF,$1B,$E9,$DB,$4F,$C4,$E3,$C7,$F1,$BB,$5A,$CB,$9E,$BE,$1D,$7C
	.dc.b	$F3,$9C,$3C,$B9,$FF,$BB,$63,$FD,$FB,$F6,$7F,$E1,$7F,$89,$7E,$65
	.dc.b	$FE,$89,$FC,$7B,$FC,$2D,$E8,$BF,$7A,$7E,$79,$C3,$1F,$F2,$5E,$1C
	.dc.b	$3D,$96,$1C,$2D,$CD,$BB,$5E,$DF,$F1,$31,$F9,$FC,$FC,$59,$FF,$A7
	.dc.b	$A3,$27,$19,$ED,$87,$93,$F0,$2E,$13,$E2,$AF,$47,$D3,$FF,$4A,$FC
	.dc.b	$33,$F5,$5F,$96,$7F,$8C,$FF,$39,$FF,$40,$7B,$BF,$EC,$0F,$4B,$D7
	.dc.b	$F0,$9F,$EE,$D5,$5B,$FB,$BF,$E3,$3F,$86,$27,$7F,$8C,$5A,$D6,$1F
	.dc.b	$BC,$78,$7E,$33,$76,$FF,$37,$E7,$E0,$C7,$97,$37,$FB,$DE,$79,$1A
	.dc.b	$CF,$BE,$7F,$3B,$C4,$FE,$05,$FB,$D7,$F9,$2F,$EC,$5F,$95,$FF,$19
	.dc.b	$DD,$FA,$2F,$FA,$C3,$F2,$6F,$F6,$FE,$2C,$38,$5E,$EB,$5F,$3F,$EB
	.dc.b	$9C,$FC,$27,$F1,$47,$7F,$1F,$76,$FF,$43,$E9,$E0,$A6,$9E,$0F,$C2
	.dc.b	$5F,$37,$E6,$BD,$E9,$DE,$7F,$DC,$1F,$24,$FD,$E3,$FC,$C9,$F9,$D3
	.dc.b	$F0,$5F,$F0,$40,$FC,$6B,$DB,$F3,$4F,$D6,$D8,$71,$CF,$FD,$33,$87
	.dc.b	$1F,$A5,$E5,$87,$0F,$CC,$B7,$9F,$6C,$FF,$C2,$98,$7E,$2B,$1E,$4B
	.dc.b	$1F,$17,$E6,$FF,$CE,$27,$31,$D6,$BC,$9F,$8F,$7F,$9E,$E7,$31,$E8
	.dc.b	$CB,$E9,$FD,$E7,$FA,$B2,$7F,$13,$8F,$FA,$BB,$F1,$9F,$ED,$BF,$BA
	.dc.b	$BF,$79,$FF,$13,$7E,$49,$FC,$BB,$F6,$0F,$6F,$8D,$FE,$87,$BD,$D6
	.dc.b	$BE,$39,$FA,$86,$B2,$E9,$F4,$6C,$FD,$18,$F1,$87,$ED,$18,$AD,$D7
	.dc.b	$87,$99,$AF,$FA,$6E,$7A,$0D,$3C,$33,$AF,$B5,$F6,$7A,$9F,$E9,$BF
	.dc.b	$CD,$BA,$FE,$04,$F8,$A7,$E7,$FF,$DB,$3F,$5B,$7E,$C7,$F6,$30,$E1
	.dc.b	$BF,$FA,$52,$F7,$F3,$EB,$FF,$52,$F3,$AF,$79,$7D,$37,$8F,$D7,$DE
	.dc.b	$0F,$F6,$CE,$BE,$2C,$B9,$F1,$F9,$A7,$3A,$77,$DB,$FB,$FE,$FF,$57
	.dc.b	$E1,$FF,$9C,$79,$CF,$D2,$9F,$B8,$3F,$20,$FF,$29,$7E,$DC,$F7,$FF
	.dc.b	$23,$87,$0B,$DD,$6B,$E0,$F3,$AF,$F3,$6C,$BA,$3F,$8F,$BE,$FF,$8D
	.dc.b	$FE,$35,$E5,$9F,$7F,$B8,$FD,$2C,$63,$1F,$F3,$17,$9F,$F8,$BE,$2F
	.dc.b	$EC,$BC,$3F,$CB,$3E,$ED,$FA,$A3,$F3,$2F,$D3,$76,$7E,$9A,$F7,$F0
	.dc.b	$CB,$DD,$7E,$AF,$EC,$75,$C1,$FD,$87,$27,$E7,$7F,$A1,$C7,$8C,$FE
	.dc.b	$9D,$E7,$87,$2F,$85,$F2,$4F,$1C,$BB,$69,$E3,$E3,$FD,$17,$D1,$7F
	.dc.b	$84,$FE,$6F,$FA,$1F,$EF,$AF,$B4,$FF,$21,$FB,$7F,$BC,$AD,$62,$AD
	.dc.b	$7C,$4B,$F2,$3D,$69,$FE,$12,$CB,$F6,$D8,$7F,$87,$FE,$7E,$CE,$71
	.dc.b	$E5,$9F,$F5,$DF,$1D,$7D,$B3,$F1,$7C,$AF,$ED,$FE,$C7,$FC,$7D,$FB
	.dc.b	$6B,$F1,$E7,$EE,$2E,$97,$EA,$6F,$C7,$FF,$89,$E1,$C2,$F7,$5A,$FC
	.dc.b	$EF,$3C,$1F,$E3,$CD,$BF,$86,$5F,$68,$FE,$24,$F1,$CB,$B6,$5E,$39
	.dc.b	$FF,$21,$E7,$E3,$9F,$17,$37,$F0,$DF,$99,$FF,$36,$EF,$FC,$67,$E6
	.dc.b	$FD,$1F,$9C,$FF,$74,$D1,$5A,$F8,$2F,$F0,$97,$02,$BF,$DB,$95,$AD
	.dc.b	$3F,$CD,$BF,$BE,$7F,$21,$C3,$AC,$7F,$27,$79,$63,$F1,$7E,$4B,$C5
	.dc.b	$3F,$E7,$4C,$FC,$59,$7F,$5B,$C5,$F1,$DF,$E9,$0F,$86,$7E,$5E,$FC
	.dc.b	$9D,$FA,$37,$FA,$2E,$F7,$97,$FA,$8B,$F0,$5F,$F5,$46,$1C,$2F,$7D
	.dc.b	$1F,$D0,$B2,$FF,$4F,$7E,$0B,$FC,$DD,$FF,$4E,$E1,$FC,$F5,$F0,$67
	.dc.b	$DF,$C5,$FE,$69,$18,$C3,$FB,$D5,$7F,$4C,$F3,$3F,$D9,$7E,$7B,$F1
	.dc.b	$FF,$E0,$7F,$AC,$FF,$A9,$3F,$0B,$D5,$FF,$B1,$AB,$5E,$9A,$BF,$D5
	.dc.b	$3F,$49,$AA,$FF,$63,$57,$F4,$BC,$7F,$D9,$F1,$FD,$45,$EF,$8F,$6F
	.dc.b	$73,$FA,$44,$63,$1E,$9C,$7F,$D2,$FF,$0E,$FF,$7B,$F9,$FF,$EA,$AF
	.dc.b	$DE,$7F,$4D,$7F,$D9,$27,$3E,$72,$7F,$55,$F2,$39,$D3,$FD,$F7,$F0
	.dc.b	$FF,$D7,$6F,$FD,$F5,$1F,$E3,$0C,$91,$CA,$FF,$41,$3C,$4F,$D2,$72
	.dc.b	$3E,$6F,$F8,$66,$5D,$FD,$BF,$EC,$1F,$8A,$2E,$7E,$2C,$FC,$55,$F5
	.dc.b	$FC,$7E,$FF,$FC,$01,$F8,$0B,$F0,$6F,$EF,$B8,$F1,$C5,$8F,$1C,$38
	.dc.b	$7C,$CF,$FA,$E7,$F1,$9D,$D7,$6C,$DF,$8B,$FF,$79,$C7,$96,$3C,$6B
	.dc.b	$FC,$01,$18,$AF,$1E,$ED,$7F,$C2,$51,$EA,$AE,$99,$F8,$3F,$05,$FE
	.dc.b	$F9,$F8,$E7,$97,$F1,$B6,$AF,$C7,$5F,$81,$BF,$4B,$7E,$AA,$C1,$FA
	.dc.b	$3F,$F8,$69,$E5,$87,$1B,$E1,$5B,$FD,$7B,$5A,$FC,$BC,$FB,$E7,$FE
	.dc.b	$15,$C7,$8F,$E4,$36,$B5,$97,$3D,$3C,$3F,$7F,$E8,$9C,$E1,$E6,$9F
	.dc.b	$2F,$C3,$7F,$85,$FF,$11,$FA,$1F,$92,$3F,$24,$7C,$EB,$F2,$47,$DE
	.dc.b	$7B,$7F,$33,$7C,$BF,$F8,$D3,$1E,$38,$70,$BD,$E6,$B7,$E9,$FB,$DB
	.dc.b	$F0,$DA,$77,$D3,$FC,$61,$97,$2F,$A0,$B5,$AC,$FA,$6B,$F3,$7E,$33
	.dc.b	$D1,$39,$C3,$CD,$3F,$47,$F4,$5F,$F1,$97,$E1,$9F,$8B,$F4,$9F,$24
	.dc.b	$FC,$A5,$F9,$5F,$EF,$FA,$1F,$96,$BE,$77,$FC,$9F,$8F,$1C,$38,$5E
	.dc.b	$E9,$F8,$CE,$B6,$FD,$06,$7D,$BF,$2C,$FF,$25,$63,$C7,$4B,$5A,$CB
	.dc.b	$9F,$B5,$F8,$AC,$BC,$F3,$9A,$F2,$CF,$E3,$BF,$19,$FF,$26,$7E,$5D
	.dc.b	$FA,$5F,$CB,$BF,$A9,$7D,$B3,$F0,$DF,$EB,$19,$3F,$51,$7E,$7B,$FE
	.dc.b	$68,$C3,$87,$CE,$30,$E1,$F1,$FF,$02,$DE,$76,$F7,$FD,$D6,$9F,$D7
	.dc.b	$FF,$18,$FF,$32,$65,$CB,$1B,$5A,$CB,$A6,$BE,$2F,$9B,$F3,$CE,$6B
	.dc.b	$BE,$DF,$CC,$3E,$03,$FD,$01,$3F,$C1,$FA,$1F,$83,$BF,$33,$7E,$58
	.dc.b	$F8,$17,$E3,$8F,$C1,$3F,$B9,$E5,$CB,$2F,$B7,$8F,$1C,$38,$5E,$FE
	.dc.b	$D7,$FD,$13,$F0,$1F,$C3,$74,$EF,$1F,$E7,$F8,$C5,$F8,$70,$FF,$43
	.dc.b	$46,$34,$F1,$E9,$E2,$8F,$F3,$F1,$9C,$BB,$D7,$FA,$07,$4F,$E3,$5F
	.dc.b	$9A,$FF,$86,$7E,$83,$FC,$99,$F9,$AB,$F3,$6F,$C4,$FC,$D7,$FA,$72
	.dc.b	$F7,$F5,$5F,$EC,$DF,$B1,$5E,$FC,$77,$E9,$5B,$D3,$F9,$76,$9F,$E2
	.dc.b	$5F,$CE,$3F,$D3,$19,$72,$D6,$D6,$B3,$E9,$97,$87,$E7,$7C,$D3,$9A
	.dc.b	$F2,$EC,$F8,$DF,$F4,$D7,$E2,$3F,$CC,$FE,$A3,$FC,$C5,$FB,$DB,$F3
	.dc.b	$47,$E8,$4F,$97,$FE,$55,$F1,$DF,$EB,$6C,$38,$7D,$D3,$0E,$17,$BF
	.dc.b	$8F,$7E,$BF,$BD,$FF,$A4,$69,$FE,$6F,$F8,$FF,$F5,$96,$5C,$B1,$B5
	.dc.b	$AC,$BA,$6B,$E2,$D5,$F7,$BE,$8A,$D5,$F8,$27,$FD,$3D,$E2,$31,$E9
	.dc.b	$87,$1D,$7F,$CD,$53,$9C,$38,$CE,$79,$7F,$DA,$3E,$4F,$F5,$5F,$C2
	.dc.b	$9F,$A1,$BF,$43,$FE,$07,$E9,$FC,$4E,$07,$C6,$70,$CF,$FB,$26,$1C
	.dc.b	$2F,$7C,$BF,$E8,$DF,$95,$D5,$7F,$B8,$72,$7E,$3D,$FE,$D3,$D3,$C1
	.dc.b	$1D,$E7,$FD,$B5,$18,$C3,$9E,$7F,$ED,$18,$C7,$85,$FC,$0B,$27,$E3
	.dc.b	$7F,$ED,$5F,$B6,$FE,$DD,$E9,$3F,$60,$7E,$F6,$FE,$01,$FF,$64,$FE
	.dc.b	$40,$FD,$B9,$F9,$07,$FB,$CB,$1E,$38,$70,$BD,$D3,$E8,$F5,$C3,$FE
	.dc.b	$03,$CF,$B7,$C9,$FF,$BB,$71,$E3,$A5,$AD,$65,$CF,$F0,$46,$BF,$CD
	.dc.b	$27,$99,$FC,$7E,$75,$0F,$82,$F6,$4F,$2B,$F5,$5F,$2D,$3F,$BA,$F6
	.dc.b	$B2,$F0,$ED,$E8,$FC,$B3,$AF,$E6,$AF,$CD,$9F,$75,$FD,$87,$E4,$FF
	.dc.b	$C0,$DF,$95,$BF,$2A,$DD,$6B,$E6,$FF,$B4,$78,$CA,$FF,$00,$6B,$DF
	.dc.b	$87,$FC,$49,$9F,$5C,$FE,$04,$23,$C1,$FB,$BE,$1D,$63,$D3,$F3,$CF
	.dc.b	$6D,$6F,$3F,$06,$BF,$E1,$9F,$C5,$33,$87,$08,$F3,$D7,$9B,$2F,$56
	.dc.b	$BE,$5E,$1C,$67,$E8,$FD,$2B,$F6,$DB,$79,$F0,$ED,$87,$09,$7E,$DE
	.dc.b	$AF,$F1,$B5,$3F,$67,$7C,$F7,$FA,$E7,$E8,$3F,$C3,$FE,$3F,$FA,$A7
	.dc.b	$CB,$E1,$7F,$E3,$DB,$DF,$E0,$3F,$E0,$4F,$AC,$FD,$5F,$B9,$F3,$73
	.dc.b	$EB,$5A,$BE,$E7,$23,$FC,$9C,$F8,$3F,$04,$CC,$64,$F2,$69,$E3,$C3
	.dc.b	$AC,$66,$BC,$3A,$79,$BD,$BF,$F2,$C6,$3E,$0C,$38,$6B,$FE,$C1,$E2
	.dc.b	$E9,$CD,$FE,$47,$FD,$BF,$FC,$83,$F9,$E3,$F2,$07,$E3,$8F,$B6,$DF
	.dc.b	$4A,$F7,$5A,$FD,$47,$31,$E8,$D7,$E5,$E5,$FE,$53,$FD,$17,$1C,$5F
	.dc.b	$E6,$38,$F0,$7D,$23,$5F,$07,$E0,$1E,$8A,$D6,$9D,$EB,$5A,$3F,$09
	.dc.b	$D5,$7F,$9D,$71,$E7,$8F,$A7,$77,$EB,$9E,$4C,$78,$C3,$F0,$8F,$99
	.dc.b	$F5,$9A,$FC,$C5,$FA,$3B,$F5,$07,$EE,$2F,$BA,$7E,$3F,$C3,$46,$1C
	.dc.b	$2F,$74,$FC,$1F,$5F,$9F,$FC,$0E,$1F,$F4,$56,$9D,$DF,$C0,$F3,$FE
	.dc.b	$42,$3C,$5A,$F8,$75,$FF,$7F,$57,$FA,$26,$73,$97,$93,$D7,$F8,$32
	.dc.b	$9C,$D7,$9F,$F7,$3F,$36,$BE,$58,$F5,$7E,$AB,$8C,$BB,$6C,$FD,$3F
	.dc.b	$1A,$7A,$BF,$52,$7E,$30,$FC,$C3,$F3,$DF,$0B,$FE,$1A,$F9,$87,$1F
	.dc.b	$FA,$73,$F4,$6F,$F2,$95,$EF,$8B,$F3,$77,$BA,$D7,$DB,$74,$C7,$9F
	.dc.b	$D1,$7E,$7B,$0E,$18,$7E,$C7,$07,$A7,$9B,$DE,$7E,$0F,$C2,$1F,$81
	.dc.b	$FF,$AD,$71,$E3,$E7,$FF,$96,$B3,$5E,$FE,$B7,$58,$F2,$DB,$D1,$FB
	.dc.b	$7F,$8F,$6F,$3E,$1C,$27,$D9,$F8,$36,$67,$FA,$57,$E7,$BF,$E5,$2F
	.dc.b	$CC,$1F,$AF,$7F,$94,$7F,$42,$7F,$00,$EF,$E2,$17,$BA,$D7,$E8,$D9
	.dc.b	$8F,$F2,$DE,$3F,$C0,$B8,$BE,$0D,$F0,$FD,$53,$E0,$4F,$FB,$46,$39
	.dc.b	$FC,$E3,$4F,$0F,$17,$E8,$EB,$A4,$E7,$3F,$F3,$27,$A5,$F0,$27,$96
	.dc.b	$7E,$3B,$7E,$30,$FD,$93,$8E,$5D,$23,$F0,$1A,$FD,$4F,$87,$FF,$52
	.dc.b	$FD,$83,$F1,$97,$E4,$3F,$4B,$F0,$3F,$EB,$B7,$BF,$C3,$97,$BA,$D7
	.dc.b	$E3,$BF,$2F,$1F,$F5,$7F,$EB,$FE,$EE,$0F,$E9,$18,$70,$7E,$27,$6F
	.dc.b	$FB,$A6,$BC,$3A,$78,$3C,$07,$E9,$9E,$DB,$E3,$A7,$79,$7B,$59,$C3
	.dc.b	$8E,$BE,$7F,$D4,$BC,$5A,$79,$B3,$E9,$C1,$FE,$04,$E0,$F5,$67,$CE
	.dc.b	$3F,$D8,$11,$89,$DF,$E6,$DF,$A6,$E4,$FF,$7F,$7E,$A9,$FE,$42,$F8
	.dc.b	$6F,$F1,$57,$D8,$BF,$29,$FF,$67,$FB,$7B,$E3,$0B,$E1,$F5,$0B,$E1
	.dc.b	$F8,$8F,$F8,$0F,$5F,$AD,$7C,$2F,$FC,$9F,$F4,$3F,$A9,$D3,$C3,$F6
	.dc.b	$9F,$E3,$AA,$FC,$64,$63,$5F,$1E,$BF,$12,$7C,$47,$86,$BF,$C7,$59
	.dc.b	$F9,$FA,$3F,$59,$87,$1D,$7C,$98,$7F,$BA,$3E,$63,$8E,$9E,$8C,$3C
	.dc.b	$FF,$82,$74,$FD,$59,$E4,$3F,$1B,$FF,$10,$EF,$FE,$B4,$F4,$1F,$5B
	.dc.b	$BA,$D7,$C1,$7F,$94,$BF,$0B,$FD,$CF,$E6,$79,$FA,$4F,$77,$2F,$C8
	.dc.b	$D3,$AE,$98,$84,$63,$0E,$59,$FA,$A7,$86,$5D,$B4,$FF,$62,$FE,$07
	.dc.b	$99,$E3,$39,$D3,$F8,$4F,$EE,$3E,$5C,$FD,$31,$FC,$33,$D8,$F9,$3F
	.dc.b	$57,$FA,$AF,$D6,$DF,$91,$7F,$93,$3C,$DF,$ED,$5F,$9B,$E9,$4E,$BE
	.dc.b	$C7,$F6,$3F,$2A,$FC,$A7,$53,$EB,$C7,$F6,$BD,$97,$0E,$39,$75,$E8
	.dc.b	$F4,$57,$86,$3C,$1F,$67,$8A,$DE,$7D,$BF,$41,$EB,$5E,$5D,$1F,$AF
	.dc.b	$E2,$BC,$7F,$4F,$70,$FF,$01,$7D,$7F,$E1,$BD,$17,$DD,$FF,$5D,$FD
	.dc.b	$F5,$FB,$3B,$C2,$FE,$15,$F5,$5B,$AD,$4E,$7C,$87,$E3,$99,$F9,$BF
	.dc.b	$B1,$97,$4F,$9B,$B8,$70,$FD,$37,$F4,$BE,$0F,$39,$6B,$E0,$D7,$FE
	.dc.b	$00,$4E,$63,$19,$BF,$0B,$C6,$BE,$1F,$3F,$FA,$DF,$EB,$BD,$3E,$4D
	.dc.b	$FA,$5F,$FA,$73,$F7,$DF,$F0,$27,$C4,$3F,$64,$FE,$93,$EB,$B0,$E1
	.dc.b	$7B,$FA,$2F,$E8,$34,$FC,$3F,$AF,$07,$EF,$33,$EB,$D7,$FD,$77,$1E
	.dc.b	$3F,$82,$5A,$D6,$5C,$F4,$7E,$43,$FD,$6E,$FB,$E0,$DB,$CB,$F8,$5E
	.dc.b	$71,$E3,$39,$DB,$C7,$F8,$A7,$8A,$5A,$F8,$B0,$F6,$EE,$FC,$5B,$5C
	.dc.b	$3F,$53,$37,$E3,$EF,$C9,$DF,$8F,$3E,$9B,$F9,$EF,$07,$D4,$C3,$85
	.dc.b	$EE,$B5,$F2,$1D,$3C,$2F,$F6,$CE,$FF,$D7,$70,$E1,$1F,$EB,$48,$C1
	.dc.b	$FD,$E5,$FF,$6F,$DF,$3F,$C7,$3D,$D3,$FE,$12,$9C,$E7,$E4,$FC,$F7
	.dc.b	$39,$F8,$A3,$CF,$A3,$E3,$7C,$5B,$7B,$27,$3E,$4E,$34,$F4,$5B,$F7
	.dc.b	$4C,$57,$2F,$93,$F9,$5E,$FF,$F8,$B7,$B9,$FA,$9F,$C4,$F5,$9F,$B7
	.dc.b	$BF,$44,$DF,$D8,$7E,$0E,$B5,$2F,$DF,$33,$E3,$7E,$07,$9B,$FD,$87
	.dc.b	$7F,$89,$7E,$19,$8C,$BB,$F5,$7F,$08,$FC,$A7,$E1,$EB,$E3,$CF,$C4
	.dc.b	$E3,$0F,$05,$7F,$0E,$CF,$FA,$D7,$57,$EA,$AB,$84,$E7,$F1,$0F,$F1
	.dc.b	$EF,$EE,$1F,$AE,$FC,$D9,$EA,$3F,$29,$FF,$07,$7E,$B2,$FC,$D3,$F0
	.dc.b	$7F,$96,$70,$E2,$FE,$B3,$87,$0B,$DD,$7E,$23,$A3,$54,$AD,$3F,$E1
	.dc.b	$FC,$BC,$19,$3F,$62,$FF,$21,$46,$31,$E7,$F2,$BF,$E3,$D8,$C7,$D1
	.dc.b	$35,$F2,$70,$67,$2F,$1E,$5E,$9A,$D6,$BF,$9A,$9F,$1E,$9E,$79,$CF
	.dc.b	$B5,$FE,$1E,$FB,$7F,$E1,$FF,$A0,$3F,$4A,$7C,$C7,$F8,$FB,$F5,$A7
	.dc.b	$D0,$3F,$5F,$7C,$E7,$F9,$9B,$1E,$38,$70,$F6,$18,$70,$B7,$E3,$DB
	.dc.b	$DF,$FC,$03,$A7,$7F,$80,$FF,$31,$65,$CB,$5B,$5A,$CF,$A6,$5E,$2B
	.dc.b	$FF,$30,$5E,$F3,$EF,$AB,$F1,$BC,$E3,$C7,$2F,$54,$FA,$7F,$25,$F1
	.dc.b	$4B,$4F,$81,$5A,$D9,$EF,$EA,$73,$97,$69,$FF,$3A,$FA,$3F,$EA,$8C
	.dc.b	$3B,$4E,$7E,$D7,$AF,$E9,$6F,$C2,$5F,$BA,$3E,$81,$FB,$23,$F2,$EF
	.dc.b	$17,$D7,$F0,$BD,$D6,$A7,$CD,$F9,$D7,$FA,$53,$F3,$4F,$E8,$BF,$BC
	.dc.b	$F2,$FB,$A7,$E7,$8F,$C9,$1F,$C0,$3F,$9B,$FC,$C7,$87,$1F,$F8,$C3
	.dc.b	$0E,$17,$BA,$FD,$07,$E3,$FA,$A5,$75,$7F,$CE,$D9,$78,$38,$1F,$B7
	.dc.b	$FF,$A6,$63,$1A,$78,$B4,$F0,$FD,$0F,$FA,$56,$73,$A7,$7F,$C6,$F1
	.dc.b	$97,$8F,$2F,$F4,$4C,$F8,$FE,$59,$EA,$78,$F5,$F3,$E7,$EC,$97,$EC
	.dc.b	$BF,$EB,$6F,$BD,$FE,$CF,$FB,$13,$F6,$3F,$C8,$FE,$E5,$F7,$7F,$5B
	.dc.b	$FF,$70,$7D,$CE,$F7,$FF,$28,$DE,$FC,$B5,$EF,$E5,$FF,$CD,$3E,$1B
	.dc.b	$F3,$6F,$EE,$7F,$1E,$FB,$16,$7F,$CA,$7F,$5B,$FF,$74,$FD,$A7,$FB
	.dc.b	$9F,$0E,$1A,$FF,$97,$B0,$E1,$7B,$F4,$17,$FC,$AA,$DF,$9E,$7F,$80
	.dc.b	$3E,$AB,$F4,$FF,$46,$FD,$65,$F9,$EB,$F2,$37,$E6,$BF,$B7,$C7,$8E
	.dc.b	$DF,$E6,$CC,$78,$E1,$C2,$F7,$ED,$FF,$B7,$33,$ED,$ED,$FF,$A8,$6B
	.dc.b	$5F,$83,$FF,$B2,$E7,$3A,$F8,$75,$F0,$7C,$2F,$FB,$3F,$F7,$3F,$F0
	.dc.b	$AF,$DA,$3C,$D7,$E8,$2F,$DA,$7F,$A6,$7D,$7B,$ED,$9C,$5F,$EE,$CC
	.dc.b	$78,$E1,$C2,$F7,$5E,$FE,$45,$6B,$1E,$9F,$39,$FE,$32,$AF,$17,$2F
	.dc.b	$FA,$2B,$4F,$15,$6B,$47,$E6,$FF,$EE,$5F,$CF,$7F,$A1,$7D,$B7,$FA
	.dc.b	$D3,$F4,$D7,$EB,$7F,$C7,$FF,$53,$7E,$33,$FF,$00,$2F,$7F,$87,$2F
	.dc.b	$7F,$18,$BF,$EB,$9C,$4B,$DD,$6B,$F2,$9F,$07,$07,$FA,$7B,$E4,$FF
	.dc.b	$C9,$F9,$73,$D3,$F9,$77,$85,$FE,$F9,$FC,$4F,$FC,$7F,$F9,$9B,$F3
	.dc.b	$A7,$E8,$2F,$C6,$BF,$05,$FB,$B7,$E0,$7D,$87,$0F,$FB,$BB,$6F,$F5
	.dc.b	$5E,$3C,$70,$E1,$E2,$B0,$FC,$47,$57,$0D,$3B,$E9,$DB,$4E,$BF,$31
	.dc.b	$E3,$CF,$9C,$E7,$1E,$5F,$84,$FF,$AE,$35,$FE,$8D,$5F,$EB,$AD,$3F
	.dc.b	$A1,$FC,$3F,$F8,$AB,$F1,$27,$E9,$9F,$D2,$3E,$7B,$CF,$D7,$FE,$5C
	.dc.b	$BD,$FD,$82,$F7,$F3,$0F,$4B,$DB,$FB,$CF,$F9,$8B,$CB,$F9,$DF,$80
	.dc.b	$3E,$3B,$FC,$F3,$F9,$23,$F7,$5F,$87,$C4,$C3,$87,$86,$C3,$87,$8C
	.dc.b	$C3,$87,$ED,$3F,$2B,$F6,$EF,$F3,$7F,$8C,$FD,$CF,$FC,$55,$E1,$3F
	.dc.b	$17,$7E,$77,$F7,$FF,$28,$7E,$18,$C7,$8E,$1F,$E3,$1F,$C6,$9F,$90
	.dc.b	$3F,$20,$7E,$56,$FC,$BF,$8F,$27,$9B,$8F,$D9,$FA,$17,$F6,$DF,$CF
	.dc.b	$BF,$A3,$6B,$EB,$F1,$3F,$DC,$FD,$7F,$0B,$AD,$7B,$FD,$E3,$C7,$E1
	.dc.b	$7C,$DF,$84,$FE,$99,$E1,$3F,$2B,$FC,$C7,$D6,$BF,$61,$DE,$9F,$66
	.dc.b	$F7,$5A,$F6,$F3,$3F,$92,$9C,$FB,$2F,$33,$FD,$F3,$C1,$F9,$1A,$7A
	.dc.b	$9F,$91,$F4,$7F,$7F,$F9,$97,$F9,$17,$F6,$4F,$8F,$EA,$7E,$43,$CE
	.dc.b	$FC,$E7,$C2,$2B,$FC,$95,$FB,$A6,$B2,$FC,$8D,$6A,$5E,$E7,$83,$0E
	.dc.b	$59,$76,$FC,$07,$C7,$E6,$FE,$CB,$DC,$FE,$71,$F0,$7F,$4B,$83,$F6
	.dc.b	$DA,$7E,$CF,$CF,$FE,$05,$5A,$FD,$6F,$ED,$E1,$C7,$E4,$DF,$80,$F4
	.dc.b	$C3,$87,$C0,$FE,$E6,$7F,$C8,$91,$F8,$2F,$87,$FE,$2D,$F1,$1F,$E6
	.dc.b	$BF,$C7,$1F,$A3,$3F,$73,$D5,$FF,$17,$F8,$CF,$E4,$1C,$2B,$DF,$EA
	.dc.b	$7A,$A7,$C3,$7F,$1C,$CB,$A7,$C1,$FD,$0D,$3B,$E3,$C6,$FE,$9F,$C8
	.dc.b	$7F,$38,$F8,$5F,$E8,$1F,$58,$FD,$57,$F2,$F3,$9F,$91,$FE,$95,$2F
	.dc.b	$D9,$B3,$87,$F4,$9F,$95,$FE,$89,$5F,$D3,$FE,$4F,$AC,$E7,$E5,$3F
	.dc.b	$A2,$7C,$B7,$F2,$E8,$FD,$1F,$3F,$F6,$6F,$9B,$FE,$CB,$BB,$E6,$FF
	.dc.b	$79,$FA,$FF,$F9,$B7,$C1,$FE,$CB,$87,$FA,$5E,$F7,$5A,$FC,$0F,$F5
	.dc.b	$BF,$49,$FD,$F3,$E9,$3F,$BD,$FD,$CB,$E9,$5F,$44,$FC,$97,$79,$FF
	.dc.b	$7C,$BD,$D6,$BF,$0F,$FD,$BF,$E5,$DF,$EE,$CE,$CF,$F2,$9F,$97,$FB
	.dc.b	$CE,$97,$F0,$03,$F7,$2F,$F1,$C7,$1B,$0E,$1D,$6C,$38,$5E,$F9,$FA
	.dc.b	$EC,$3F,$AC,$D7,$FA,$99,$FF,$7D,$9F,$A8,$BE,$77,$88,$C1,$C7,$3F
	.dc.b	$2F,$C7,$FD,$5C,$7F,$DC,$B3,$D8,$EB,$5E,$2F,$D0,$38,$C6,$2B,$87
	.dc.b	$A8,$FD,$03,$86,$3D,$37,$F7,$CF,$FB,$77,$6F,$F2,$D6,$0F,$BD,$FF
	.dc.b	$2C,$7D,$EF,$F9,$5F,$ED,$3F,$CA,$DD,$1F,$CB,$6F,$75,$AF,$BB,$FF
	.dc.b	$2B,$4F,$F7,$F9,$7D,$07,$C3,$C5,$F8,$E7,$F6,$BF,$C8,$DF,$98,$3E
	.dc.b	$65,$F9,$37,$FA,$BF,$CF,$FE,$5B,$8F,$1C,$38,$5E,$EB,$FD,$69,$FB
	.dc.b	$6E,$B2,$FA,$15,$FD,$DF,$F6,$7F,$99,$19,$3C,$3F,$3F,$F2,$9F,$A7
	.dc.b	$AF,$99,$E3,$0D,$7C,$BA,$F9,$3E,$A7,$F8,$FE,$BE,$3A,$D6,$BE,$3F
	.dc.b	$6F,$38,$F2,$C7,$8E,$DF,$EC,$9F,$17,$C5,$B7,$AB,$4E,$9F,$9A,$E3
	.dc.b	$3E,$71,$DF,$EA,$F1,$8F,$2D,$9F,$A4,$7A,$FF,$56,$7E,$68,$FD,$71
	.dc.b	$EE,$7F,$A3,$7F,$60,$7C,$E3,$F3,$2E,$17,$BA,$D4,$FF,$A2,$39,$FE
	.dc.b	$17,$23,$2E,$B9,$7C,$A9,$CC,$78,$3E,$67,$1D,$1F,$76,$37,$8F,$87
	.dc.b	$F2,$6F,$87,$8F,$3C,$FC,$1D,$CD,$9F,$94,$6B,$1E,$33,$9D,$BF,$9C
	.dc.b	$67,$D7,$0E,$38,$7B,$3E,$27,$FB,$8F,$E3,$BF,$8D,$FD,$C9,$F8,$67
	.dc.b	$F4,$BF,$AF,$FE,$53,$FA,$FB,$F4,$67,$EE,$1C,$38,$7F,$6D,$87,$0F
	.dc.b	$96,$FF,$02,$7C,$5B,$0E,$1F,$AF,$6E,$D7,$DB,$F5,$3F,$89,$AF,$87
	.dc.b	$5F,$EB,$55,$FB,$BE,$94,$E7,$3E,$B9,$3D,$BF,$ED,$33,$9D,$FC,$FD
	.dc.b	$18,$C7,$CB,$97,$AF,$6F,$27,$03,$F3,$BF,$26,$FE,$9C,$7D,$F2,$FC
	.dc.b	$EF,$D1,$FB,$83,$F3,$47,$E2,$57,$F2,$EF,$F3,$6F,$90,$FC,$B7,$75
	.dc.b	$A1,$3A,$FC,$BB,$37,$F0,$38,$3F,$8F,$63,$D7,$E0,$31,$08,$F0,$4F
	.dc.b	$7F,$6D,$F0,$FE,$DA,$D6,$5D,$F2,$FF,$3F,$FE,$35,$9C,$78,$CF,$8B
	.dc.b	$0F,$F3,$97,$D8,$F8,$A3,$1B,$3F,$6E,$C6,$3E,$CF,$D9,$EE,$1F,$E7
	.dc.b	$9F,$C3,$3F,$C6,$FF,$27,$FC,$4F,$F4,$35,$3F,$32,$7E,$79,$FD,$CF
	.dc.b	$F2,$07,$D6,$7F,$B8,$BC,$46,$3C,$70,$E1,$7B,$AF,$F2,$2D,$6B,$8F
	.dc.b	$FA,$16,$5D,$B2,$FD,$56,$3C,$7E,$22,$D6,$B2,$E7,$F4,$0D,$1F,$A1
	.dc.b	$7F,$8B,$B0,$F0,$63,$E6,$AD,$7A,$59,$C7,$94,$E7,$0F,$4F,$E8,$3F
	.dc.b	$02,$73,$BF,$9F,$E6,$31,$1F,$ED,$AF,$BC,$FF,$5D,$7E,$93,$FC,$3B
	.dc.b	$F1,$D7,$DA,$BF,$36,$DF,$E2,$FF,$E6,$DF,$A9,$5E,$EB,$FD,$AB,$CD
	.dc.b	$F0,$72,$E9,$5E,$E7,$B7,$99,$70,$E5,$87,$7F,$C5,$7D,$55,$C2,$3C
	.dc.b	$5F,$49,$8A,$DF,$C1,$35,$F2,$FE,$27,$E4,$DF,$D3,$39,$DB,$FC,$21
	.dc.b	$B7,$8F,$F0,$DF,$B7,$FB,$CF,$F9,$67,$81,$F9,$17,$F5,$1F,$CA,$7E
	.dc.b	$1D,$6B,$EF,$3F,$DB,$93,$FD,$57,$F4,$BC,$F8,$4C,$7D,$97,$F8,$6F
	.dc.b	$BC,$B3,$9C,$38,$FE,$B5,$EC,$AE,$99,$F8,$3C,$5C,$5E,$E3,$C7,$F1
	.dc.b	$EF,$D6,$BB,$E7,$E7,$D7,$FB,$37,$E0,$B8,$D3,$CB,$6F,$DA,$B1,$8F
	.dc.b	$2F,$DC,$7E,$EF,$D9,$BF,$2F,$7E,$A4,$FC,$85,$F9,$73,$F4,$EF,$F6
	.dc.b	$F7,$D3,$F0,$BD,$D6,$A7,$FD,$C5,$F5,$F9,$FC,$9B,$EE,$65,$D7,$2B
	.dc.b	$87,$0C,$BA,$7B,$9F,$72,$B5,$97,$3F,$73,$17,$BC,$FF,$DD,$DF,$41
	.dc.b	$E1,$CB,$CD,$97,$97,$DA,$7E,$63,$FE,$B0,$FD,$3B,$FC,$19,$E3,$7F
	.dc.b	$2E,$F8,$3F,$E8,$3F,$88,$7F,$2D,$FC,$89,$F7,$5B,$F3,$BF,$C6,$97
	.dc.b	$BF,$41,$7B,$FB,$DD,$57,$E3,$7E,$3F,$F9,$23,$FC,$EF,$3E,$F0,$F8
	.dc.b	$8F,$EC,$91,$8D,$3C,$3F,$20,$FD,$07,$FD,$4B,$39,$CF,$BC,$FF,$8E
	.dc.b	$BC,$D4,$E7,$4F,$47,$A3,$E2,$DB,$CD,$1E,$BF,$99,$F2,$FE,$C0,$F9
	.dc.b	$67,$ED,$4F,$CA,$9F,$97,$BF,$57,$FF,$2D,$FB,$5C,$2F,$75,$A9,$FF
	.dc.b	$2F,$FE,$A3,$9F,$DD,$7D,$7C,$FF,$E5,$9C,$FB,$67,$88,$46,$30,$E5
	.dc.b	$F5,$3E,$E9,$E1,$97,$6D,$3C,$5E,$76,$67,$8C,$E7,$4F,$F3,$7F,$B5
	.dc.b	$E5,$CF,$D5,$87,$A7,$EF,$1F,$51,$9D,$3D,$11,$FE,$15,$E6,$F9,$39
	.dc.b	$F8,$30,$E3,$97,$F9,$EB,$F0,$DF,$B7,$E7,$7F,$93,$38,$9F,$B4,$7E
	.dc.b	$8F,$EB,$1F,$A1,$3F,$69,$DF,$94,$BD,$F9,$9F,$E2,$9A,$7E,$F7,$AE
	.dc.b	$AF,$89,$9F,$5D,$FF,$0C,$7F,$A0,$5F,$06,$9D,$FF,$44,$F8,$51,$8C
	.dc.b	$7F,$C6,$9F,$A7,$E3,$4F,$05,$F9,$73,$7E,$9D,$E4,$D7,$C7,$BF,$A3
	.dc.b	$F5,$EF,$CB,$FD,$9B,$F7,$BF,$F1,$66,$4F,$D5,$DF,$A4,$3F,$26,$7E
	.dc.b	$69,$C2,$F7,$5A,$97,$D4,$E7,$F1,$CF,$C0,$65,$D7,$ED,$2E,$1C,$32
	.dc.b	$7B,$7F,$76,$B5,$A7,$86,$7F,$D8,$9D,$DF,$5B,$A3,$FC,$AD,$87,$1D
	.dc.b	$3F,$AF,$4E,$70,$E3,$AF,$8F,$EB,$3F,$B0,$E1,$FE,$89,$F1,$5F,$88
	.dc.b	$BF,$31,$79,$0F,$B9,$FD,$17,$E9,$8F,$CF,$B8,$E3,$FE,$B3,$C7,$8E
	.dc.b	$1C,$2F,$79,$EA,$BF,$59,$97,$57,$FD,$69,$F4,$9F,$EB,$8C,$3C,$3F
	.dc.b	$67,$FE,$8C,$8C,$57,$8F,$4F,$17,$94,$FC,$3B,$FD,$85,$5A,$D3,$C1
	.dc.b	$2F,$D2,$73,$9F,$B7,$1E,$5A,$FA,$BE,$5F,$C9,$AF,$A3,$4E,$BC,$1F
	.dc.b	$D4,$74,$E9,$18,$FC,$CF,$78,$F2,$D1,$6F,$C2,$FD,$DB,$3F,$6A,$7E
	.dc.b	$62,$FC,$FB,$F3,$BF,$B0,$7F,$8D,$B6,$DD,$6A,$73,$F1,$EF,$BE,$CF
	.dc.b	$D5,$FC,$3E,$4F,$F5,$F6,$7D,$73,$C4,$7F,$B8,$E3,$18,$72,$8E,$B3
	.dc.b	$E1,$F4,$5F,$AE,$7F,$B1,$F9,$3F,$DC,$F8,$70,$96,$DF,$E4,$AF,$87
	.dc.b	$E3,$97,$93,$F2,$9F,$B5,$F6,$8F,$CD,$5F,$B0,$BF,$4B,$7E,$B6,$FC
	.dc.b	$5D,$F5,$DF,$5F,$7F,$F7,$AE,$3C,$70,$E1,$7B,$A7,$D3,$F6,$C9,$FB
	.dc.b	$3F,$BB,$91,$8F,$17,$F1,$27,$FB,$85,$E9,$A7,$8B,$4F,$0C,$F3,$CF
	.dc.b	$BC,$7E,$B2,$31,$3B,$C9,$F9,$06,$34,$F2,$78,$7F,$E4,$6C,$7F,$DD
	.dc.b	$DF,$22,$FD,$99,$EB,$3F,$66,$7E,$A7,$FE,$29,$F3,$3F,$83,$72,$7C
	.dc.b	$6D,$C7,$F5,$9C,$38,$DF,$0F,$90,$FD,$67,$E0,$7D,$EB,$79,$F7,$FD
	.dc.b	$3B,$F4,$9A,$F8,$B1,$F0,$D7,$EA,$A3,$1A,$F8,$F5,$FF,$1D,$7D,$57
	.dc.b	$E9,$A7,$3A,$77,$FD,$33,$18,$76,$C7,$C7,$87,$A7,$F2,$CF,$14,$F9
	.dc.b	$F2,$E9,$B7,$F0,$1F,$3B,$3B,$7C,$39,$CF,$0F,$FB,$9F,$87,$E0,$61
	.dc.b	$C7,$5F,$74,$FF,$50,$9F,$BB,$3F,$13,$E9,$1F,$8D,$FF,$99,$3F,$61
	.dc.b	$7E,$C1,$F8,$AE,$3F,$C8,$61,$C3,$0F,$F1,$75,$EE,$B5,$FB,$EF,$F9
	.dc.b	$F3,$BB,$E7,$D7,$D7,$FD,$7F,$FC,$81,$F6,$7F,$E9,$4F,$B7,$FE,$F7
	.dc.b	$F0,$DB,$AF,$AB,$CE,$BD,$FE,$0A,$BF,$F0,$0F,$C3,$EE,$FA,$63,$FC
	.dc.b	$5F,$2E,$F8,$F1,$FC,$7B,$15,$D3,$2E,$78,$BA,$FF,$8C,$E1,$C2,$73
	.dc.b	$DF,$8B,$E1,$97,$F9,$5F,$4F,$17,$CC,$FF,$6E,$E2,$F8,$5A,$F9,$EB
	.dc.b	$5A,$7F,$85,$BE,$E7,$FD,$3B,$F6,$7F,$0B,$EF,$5F,$A7,$BC,$3F,$E1
	.dc.b	$DF,$87,$7E,$67,$E1,$7F,$D1,$FA,$30,$E1,$7B,$F6,$3F,$DB,$BD,$FF
	.dc.b	$EF,$6F,$C1,$FA,$F1,$7E,$73,$4E,$BF,$95,$7F,$39,$C7,$8E,$B6,$B5
	.dc.b	$97,$3D,$7F,$51,$F7,$DF,$CE,$AF,$7A,$76,$DF,$18,$70,$C7,$CF,$AB
	.dc.b	$77,$EF,$7E,$2A,$6D,$EE,$BD,$D3,$F5,$ED,$7C,$83,$F3,$EF,$F6,$0F
	.dc.b	$C0,$BE,$71,$C6,$FD,$75,$FA,$2F,$0B,$DD,$6A,$5F,$B9,$FF,$B2,$BF
	.dc.b	$38,$FC,$7F,$E6,$EF,$B6,$7E,$C1,$FE,$CF,$FB,$27,$E3,$FF,$E0,$8D
	.dc.b	$78,$72,$7F,$04,$C3,$85,$EE,$BF,$C1,$BF,$B5,$6A,$95,$AF,$C4,$98
	.dc.b	$F4,$FA,$6F,$EC,$7F,$B7,$7E,$83,$DB,$FE,$53,$F1,$EC,$9F,$AF,$3F
	.dc.b	$5C,$FF,$1A,$FE,$E7,$FE,$E2,$C3,$87,$9E,$C3,$85,$EF,$EA,$7E,$47
	.dc.b	$92,$BD,$D6,$B1,$7D,$67,$8B,$87,$F8,$FE,$BE,$2D,$7C,$34,$DB,$FB
	.dc.b	$CF,$E5,$5F,$B6,$FC,$AF,$FC,$8F,$EB,$FB,$BF,$0E,$7E,$18,$FC,$4B
	.dc.b	$F8,$45,$EF,$87,$FC,$5F,$CE,$BD,$D7,$F1,$1A,$D6,$1C,$F0,$7D,$D7
	.dc.b	$83,$7F,$E6,$79,$F8,$32,$E5,$5F,$E2,$09,$71,$BF,$6B,$FE,$17,$F4
	.dc.b	$EF,$C4,$FF,$CB,$FF,$95,$FE,$43,$47,$C0,$7F,$A4,$30,$E1,$8F,$F4
	.dc.b	$0C,$38,$7D,$A3,$0E,$1F,$2B,$FE,$1A,$CF,$AE,$7D,$3C,$FF,$F2,$5F
	.dc.b	$E0,$79,$CB,$94,$E7,$57,$EE,$3F,$E2,$2F,$D7,$3F,$8F,$FE,$97,$FE
	.dc.b	$95,$FB,$07,$E0,$9F,$E3,$BF,$65,$FB,$8B,$F1,$8F,$6E,$1C,$3F,$00
	.dc.b	$FF,$2D,$5E,$E9,$FB,$B6,$B6,$FE,$B3,$9F,$5A,$E9,$F6,$1E,$19,$CE
	.dc.b	$1C,$BE,$9F,$FC,$35,$FA,$7F,$F8,$DB,$F5,$1F,$E6,$7F,$55,$F4,$BF
	.dc.b	$03,$FF,$74,$7D,$57,$F8,$27,$E3,$D5,$F4,$0D,$F8,$6A,$DF,$D2,$7F
	.dc.b	$8F,$7D,$5D,$6B,$FD,$B3,$2E,$9F,$80,$FF,$86,$AB,$FC,$85,$3A,$3F
	.dc.b	$C4,$38,$73,$9D,$7C,$9F,$F9,$2B,$EF,$7F,$A3,$FA,$9F,$AC,$9F,$F4
	.dc.b	$B7,$E0,$EF,$D1,$5F,$94,$7E,$97,$C8,$61,$C3,$DA,$61,$C2,$F7,$F1
	.dc.b	$FF,$8A,$AD,$4E,$7F,$2A,$F0,$6D,$FE,$04,$D3,$C1,$F8,$23,$5F,$07
	.dc.b	$D7,$FF,$97,$BC,$4F,$EB,$7F,$27,$FE,$D4,$F9,$47,$EA,$3F,$EF,$2F
	.dc.b	$49,$FA,$B3,$EF,$FF,$33,$87,$0F,$C2,$98,$70,$FC,$53,$F5,$B6,$FC
	.dc.b	$57,$7C,$1F,$E1,$6D,$3B,$69,$D7,$F0,$2F,$14,$E7,$2E,$73,$E1,$F6
	.dc.b	$BF,$CE,$7F,$97,$7F,$70,$F9,$67,$E9,$6F,$9F,$FF,$74,$7E,$88,$FD
	.dc.b	$E9,$F9,$FF,$F3,$BF,$01,$8F,$1C,$38,$79,$AC,$38,$5B,$F1,$DF,$DA
	.dc.b	$DE,$F2,$EB,$93,$F1,$FF,$1F,$07,$F8,$C3,$5F,$1D,$6B,$F5,$CF,$F4
	.dc.b	$4F,$BF,$FE,$06,$F9,$8F,$F6,$19,$FC,$1B,$F0,$07,$EF,$3F,$E6,$8F
	.dc.b	$AE,$FE,$07,$5B,$FC,$4F,$FA,$37,$9A,$BD,$F9,$FF,$EA,$8E,$6F,$F3
	.dc.b	$36,$DF,$E4,$0C,$FA,$BF,$EA,$0C,$79,$7D,$97,$FA,$7D,$E9,$3E,$1D
	.dc.b	$1F,$6D,$E6,$BE,$5A,$78,$33,$F2,$FD,$47,$FA,$8B,$F4,$9F,$F0,$FF
	.dc.b	$9D,$F9,$7F,$6B,$FC,$B7,$FB,$77,$F6,$1F,$BE,$FE,$03,$C0,$BD,$FA
	.dc.b	$5F,$EB,$FF,$BF,$5E,$FD,$EF,$E4,$35,$AF,$33,$FD,$68,$F0,$CA,$EF
	.dc.b	$FE,$5C,$D3,$C1,$8F,$2F,$A9,$7D,$1F,$96,$B5,$A7,$7D,$3F,$BA,$7C
	.dc.b	$DF,$FA,$EB,$F7,$5F,$F1,$BF,$D4,$FF,$BB,$3F,$6D,$7E,$99,$FC,$07
	.dc.b	$F0,$5F,$F0,$C7,$57,$F3,$4B,$DF,$E2,$AB,$DF,$C1,$FF,$A1,$BE,$BF
	.dc.b	$FD,$85,$CE,$BD,$D7,$FB,$35,$E3,$F3,$F7,$83,$FC,$EB,$AF,$87,$5F
	.dc.b	$07,$D1,$F9,$67,$3A,$77,$9F,$27,$E2,$DF,$ED,$2F,$CF,$3F,$C9,$FE
	.dc.b	$C7,$EB,$BD,$97,$E8,$EF,$D8,$1F,$A5,$BF,$37,$FE,$3F,$C4,$C7,$8F
	.dc.b	$BC,$C7,$8E,$1C,$2F,$FB,$45,$EF,$6F,$F4,$5E,$7D,$DF,$F7,$5E,$3C
	.dc.b	$BD,$2F,$F7,$53,$D3,$1E,$79,$F8,$6B,$FA,$EC,$E7,$4F,$34,$FF,$60
	.dc.b	$9F,$EC,$3F,$88,$7F,$9A,$B9,$DF,$A8,$FF,$82,$BD,$57,$DB,$BF,$23
	.dc.b	$FE,$4D,$DF,$C3,$1F,$F4,$FE,$1C,$3E,$23,$F7,$B7,$BF,$07,$FB,$9F
	.dc.b	$17,$FB,$9A,$F7,$9F,$F8,$59,$FF,$3C,$9F,$EF,$ED,$3C,$50,$FC,$63
	.dc.b	$33,$93,$8E,$3D,$B4,$FE,$FF,$1E,$33,$FA,$55,$7A,$72,$F4,$79,$5F
	.dc.b	$EB,$4C,$9F,$B6,$7F,$A2,$3E,$21,$E5,$7F,$33,$FD,$39,$F1,$6F,$D3
	.dc.b	$DE,$6B,$F3,$CF,$F0,$3E,$3C,$70,$E1,$7B,$AF,$F0,$24,$E7,$E0,$BD
	.dc.b	$3B,$F0,$3F,$D8,$B9,$F8,$3E,$37,$3A,$F9,$27,$39,$74,$D3,$FC,$6D
	.dc.b	$1C,$45,$F1,$D7,$CF,$FB,$57,$FB,$FF,$F5,$5F,$F5,$1F,$A3,$FC,$97
	.dc.b	$E4,$5F,$89,$3E,$95,$FA,$B7,$FB,$6B,$CD,$FF,$10,$E1,$C3,$E1,$FF
	.dc.b	$AE,$DE,$EB,$FD,$BF,$F8,$CE,$B6,$FF,$6E,$E7,$FC,$C7,$F6,$AF,$4E
	.dc.b	$3C,$74,$B5,$AC,$B9,$E9,$E0,$F9,$6F,$A5,$7B,$D3,$B7,$07,$D5,$FA
	.dc.b	$3F,$F5,$DF,$EA,$5F,$D0,$FF,$46,$7D,$BB,$F2,$87,$EF,$5F,$E0,$7F
	.dc.b	$BF,$FF,$1C,$E1,$C3,$B1,$87,$0B,$DF,$92,$F8,$FD,$F1,$7F,$BD,$B4
	.dc.b	$EC,$FE,$3B,$B7,$FC,$46,$E7,$17,$EC,$7F,$5B,$3E,$D9,$F5,$D5,$F8
	.dc.b	$EF,$A7,$0E,$13,$9F,$61,$0F,$9B,$7E,$87,$FE,$CA,$FD,$B7,$FC,$AF
	.dc.b	$F9,$AF,$F8,$47,$CF,$7E,$59,$FE,$13,$BD,$E1,$FE,$D8,$BD,$FA,$5F
	.dc.b	$A5,$A7,$E6,$3D,$72,$E9,$F2,$DF,$90,$C7,$8E,$2F,$C5,$19,$7F,$42
	.dc.b	$8F,$D6,$CE,$A7,$34,$D3,$C9,$C9,$EA,$AD,$61,$E2,$D7,$F8,$74,$3B
	.dc.b	$7F,$DE,$7F,$B0,$FF,$B3,$7C,$87,$E9,$5F,$E5,$6F,$B3,$7E,$8C,$FD
	.dc.b	$77,$E2,$61,$C3,$F1,$66,$1C,$3E,$D9,$87,$0A,$DF,$17,$CC,$CF,$B3
	.dc.b	$EA,$8E,$5F,$75,$F0,$5E,$98,$73,$BF,$F7,$A5,$72,$17,$E7,$D3,$CD
	.dc.b	$F8,$FF,$BF,$D5,$FB,$34,$FD,$57,$FC,$DF,$FA,$33,$F2,$BF,$D6,$7E
	.dc.b	$9A,$FA,$C7,$CF,$7D,$FC,$78,$E1,$C2,$F7,$4F,$DA,$35,$97,$E1,$AB
	.dc.b	$FC,$F7,$FA,$1F,$D9,$C3,$8F,$1B,$4C,$5E,$E3,$19,$74,$8E,$67,$9E
	.dc.b	$33,$E9,$35,$FC,$94,$FE,$4F,$E1,$BF,$A0,$7E,$A4,$FB,$97,$D1,$3E
	.dc.b	$CD,$F9,$67,$F8,$E3,$CD,$E1,$8F,$F4,$0C,$38,$5E,$EB,$FC,$7B,$F8
	.dc.b	$1E,$AB,$FC,$F9,$5F,$41,$FE,$7F,$E0,$7F,$43,$73,$E0,$7E,$5F,$4F
	.dc.b	$1E,$9E,$2D,$3C,$3F,$6F,$F9,$19,$CE,$9D,$FF,$06,$FC,$A7,$C9,$FF
	.dc.b	$74,$FD,$EB,$F7,$DF,$9D,$3F,$27,$FF,$69,$7E,$7F,$FA,$CD,$FF,$B6
	.dc.b	$6F,$FE,$6A,$BD,$D6,$BC,$EF,$DB,$63,$CF,$D9,$FD,$53,$D7,$E7,$6E
	.dc.b	$1F,$DD,$30,$E3,$F4,$3F,$EF,$68,$E2,$66,$DA,$78,$E7,$F8,$94,$FF
	.dc.b	$03,$FA,$2F,$F1,$1F,$E1,$9F,$E8,$3F,$CC,$5F,$51,$FE,$A3,$F9,$EF
	.dc.b	$F3,$F7,$07,$F8,$77,$0F,$F0,$CD,$EE,$B5,$F7,$DC,$E7,$3F,$27,$FC
	.dc.b	$07,$2E,$BF,$03,$67,$39,$F6,$F8,$3F,$46,$7D,$65,$F2,$1F,$EE,$3C
	.dc.b	$57,$BF,$63,$F8,$17,$D6,$FF,$93,$3F,$0F,$7E,$59,$FD,$AB,$F5,$C7
	.dc.b	$E3,$3F,$ED,$1D,$3F,$CA,$5A,$7F,$91,$2F,$7F,$30,$BD,$FB,$9F,$C9
	.dc.b	$72,$FF,$2B,$FE,$BB,$FC,$65,$ED,$87,$F1,$67,$C1,$9F,$7F,$AC,$F1
	.dc.b	$CF,$58,$F1,$7C,$3F,$F1,$7F,$B8,$FF,$38,$FE,$D6,$FD,$5F,$FD,$D7
	.dc.b	$F4,$FF,$EB,$1E,$0F,$F3,$CE,$2F,$B2,$7D,$C3,$EE,$D8,$F1,$F3,$3F
	.dc.b	$D3,$3F,$A1,$7F,$3B,$D3,$FC,$FB,$8F,$F2,$A7,$BE,$9F,$EF,$8A,$F1
	.dc.b	$4F,$F9,$F3,$0E,$73,$FE,$FC,$F2,$FF,$96,$6B,$FE,$9D,$FD,$03,$FC
	.dc.b	$1F,$D0,$FD,$A3,$FC,$97,$C1,$FE,$A8,$CB,$FD,$3D,$F8,$37,$FA,$BE
	.dc.b	$F7,$4F,$C4,$7F,$A7,$63,$CF,$E9,$FE,$9F,$CE,$F6,$C7,$8F,$E0,$6C
	.dc.b	$7B,$FD,$07,$F3,$DE,$1F,$F6,$57,$9E,$FC,$C5,$FB,$87,$F6,$9F,$C2
	.dc.b	$9F,$55,$FD,$6E,$F7,$5A,$F4,$BF,$DA,$FF,$98,$E7,$5F,$F6,$56,$5C
	.dc.b	$F2,$FE,$E5,$F9,$17,$79,$CE,$1C,$7F,$16,$FE,$9D,$F8,$87,$FB,$AB
	.dc.b	$EC,$1F,$97,$7F,$B2,$FC,$AF,$E8,$7D,$5F,$EF,$4C,$7F,$DD,$9F,$89
	.dc.b	$7F,$BD,$2F,$7F,$79,$FE,$3C,$FC,$F3,$FB,$86,$5C,$F2,$FF,$6A,$7E
	.dc.b	$15,$DB,$1E,$3F,$61,$FE,$40,$9F,$F6,$C7,$CA,$FF,$5F,$FC,$5F,$FA
	.dc.b	$0F,$E1,$AF,$A4,$7B,$2C,$1F,$B8,$FF,$81,$BF,$80,$1F,$94,$FF,$7F
	.dc.b	$C7,$8E,$1C,$2F,$7F,$5E,$B7,$EE,$DB,$DB,$FC,$0F,$A7,$7D,$3B,$69
	.dc.b	$FE,$12,$FC,$37,$C9,$5A,$CF,$A4,$FF,$85,$A3,$FB,$B7,$E3,$DF,$D6
	.dc.b	$3E,$91,$F8,$BB,$F1,$F7,$CE,$7F,$98,$3E,$59,$FB,$0B,$F3,$6F,$F0
	.dc.b	$B6,$3C,$70,$E1,$7B,$A7,$E6,$9A,$CF,$FC,$3B,$9F,$6C,$FA,$F8,$1F
	.dc.b	$DF,$BA,$3C,$75,$AC,$B9,$CF,$F8,$0B,$C0,$FF,$06,$7C,$6F,$F6,$EF
	.dc.b	$C9,$9F,$92,$3E,$1D,$F9,$33,$F2,$6F,$F0,$D7,$E8,$CE,$FF,$F1,$66
	.dc.b	$3C,$70,$E1,$7B,$FB,$65,$BF,$25,$DE,$FF,$E3,$4D,$3B,$E9,$DB,$4F
	.dc.b	$F1,$37,$D7,$79,$2B,$59,$F4,$9F,$F1,$57,$D0,$FF,$88,$FF,$2A,$FF
	.dc.b	$02,$F1,$FF,$95,$BE,$A1,$F9,$53,$F2,$C7,$EF,$EF,$D4,$FF,$D3,$9C
	.dc.b	$CC,$78,$FE,$0D,$FE,$75,$C3,$85,$EF,$F1,$3F,$F2,$45,$6B,$2E,$B9
	.dc.b	$74,$F5,$3C,$5C,$1F,$E4,$ED,$7C,$59,$72,$FC,$43,$FC,$71,$F9,$87
	.dc.b	$F8,$6B,$F3,$0F,$CA,$FC,$AF,$F3,$6D,$7E,$7F,$E0,$BF,$98,$E1,$5E
	.dc.b	$FD,$25,$EF,$9B,$FC,$DB,$F6,$1F,$E5,$BA,$D4,$E4,$EF,$B7,$F9,$83
	.dc.b	$3E,$FF,$1A,$D1,$EE,$FF,$8B,$34,$79,$FF,$E2,$EE,$07,$E6,$8F,$CC
	.dc.b	$DF,$99,$BD,$FF,$B5,$F9,$1B,$F5,$CF,$F3,$66,$3C,$70,$E1,$7B,$F1
	.dc.b	$16,$F8,$FD,$F0,$FF,$9D,$34,$EF,$A7,$6D,$3A,$CE,$8F,$B7,$97,$48
	.dc.b	$FC,$57,$D5,$7F,$98,$FF,$04,$FF,$22,$FE,$30,$FC,$E1,$F4,$BF,$DA
	.dc.b	$BF,$38,$7E,$6F,$FD,$37,$F6,$3F,$F4,$16,$1C,$2F,$7F,$56,$BD,$FD
	.dc.b	$1F,$D2,$E7,$FE,$17,$BF,$FA,$2B,$EF,$FF,$45,$FB,$B3,$F4,$37,$EA
	.dc.b	$5F,$E6,$BF,$E0,$17,$E9,$FF,$E8,$FF,$1B,$FC,$B3,$87,$0B,$DF,$D7
	.dc.b	$2F,$F8,$DE,$7D,$B3,$FF,$27,$7A,$7F,$97,$7F,$D0,$B8,$F2,$86,$9F
	.dc.b	$E8,$D8,$C3,$E3,$D3,$FC,$F9,$E5,$FF,$A1,$E7,$39,$F8,$31,$C6,$3E
	.dc.b	$BC,$79,$CF,$93,$27,$E8,$FE,$9A,$D6,$7D,$74,$F4,$7E,$B9,$99,$F3
	.dc.b	$CE,$6F,$D9,$97,$AF,$6F,$81,$B7,$C5,$F8,$3E,$1F,$5C,$FD,$15,$F6
	.dc.b	$9F,$DF,$FF,$35,$FF,$5D,$FB,$6F,$DE,$77,$5A,$FA,$D5,$6B,$E7,$7F
	.dc.b	$D7,$9E,$9F,$FA,$1B,$E7,$FF,$4F,$FB,$0B,$F8,$03,$FD,$83,$F5,$8F
	.dc.b	$CF,$DF,$90,$F0,$E7,$FF,$62,$E1,$C2,$F7,$5F,$ED,$7E,$5F,$F3,$8D
	.dc.b	$6B,$D8,$7E,$C3,$FE,$BD,$D3,$C1,$0D,$31,$5B,$7C,$4F,$F8,$2F,$EA
	.dc.b	$BF,$D7,$55,$C2,$73,$AB,$EB,$73,$7C,$74,$F1,$FD,$27,$9E,$1B,$79
	.dc.b	$DF,$F0,$B6,$CF,$C0,$35,$F8,$4B,$F0,$FF,$95,$FB,$EF,$F8,$3F,$F3
	.dc.b	$47,$E9,$7F,$47,$93,$DB,$97,$FA,$3E,$F7,$F6,$4B,$DF,$E2,$9D,$25
	.dc.b	$F6,$3E,$EB,$E1,$DB,$74,$F0,$2D,$3C,$1D,$DE,$CA,$DC,$78,$FE,$47
	.dc.b	$17,$C2,$3C,$DB,$3F,$73,$F0,$63,$CF,$6F,$24,$BF,$59,$FF,$55,$E5
	.dc.b	$FE,$E5,$FD,$37,$FC,$91,$6F,$93,$E5,$ED,$3E,$A3,$76,$E4,$61,$C2
	.dc.b	$F7,$F8,$DE,$A7,$3E,$2F,$FB,$EE,$FF,$94,$5B,$F4,$BC,$47,$86,$7F
	.dc.b	$31,$87,$82,$39,$57,$5D,$FF,$CF,$F1,$CA,$B7,$1F,$C6,$3F,$1B,$DC
	.dc.b	$63,$4F,$EE,$59,$7F,$AE,$30,$E5,$0F,$D0,$77,$B7,$BF,$1F,$E4,$3F
	.dc.b	$13,$EE,$F8,$F7,$EB,$FF,$E6,$CE,$C7,$E7,$6F,$B6,$7E,$BB,$FE,$0C
	.dc.b	$DB,$FD,$DF,$C2,$C7,$8F,$E0,$FF,$83,$C3,$85,$FF,$85,$EF,$75,$AB
	.dc.b	$E2,$78,$72,$7E,$A1,$F1,$A3,$A4,$78,$F0,$F1,$7E,$08,$9E,$19,$77
	.dc.b	$CD,$F8,$A6,$73,$EF,$8F,$19,$F4,$FE,$B5,$C3,$6F,$6C,$7B,$3F,$7D
	.dc.b	$7E,$A8,$FD,$73,$FA,$F7,$E7,$0F,$B6,$FF,$20,$FE,$48,$FC,$C5,$FB
	.dc.b	$17,$EE,$F0,$E1,$C3,$FD,$5F,$0E,$1F,$7A,$C3,$85,$FF,$04,$BD,$DF
	.dc.b	$F8,$4A,$FB,$3F,$12,$33,$F3,$29,$7C,$37,$C8,$AF,$3E,$9E,$6D,$3C
	.dc.b	$BA,$7F,$B6,$31,$E3,$87,$8B,$3F,$47,$E6,$DF,$26,$73,$7C,$34,$FF
	.dc.b	$5F,$46,$35,$F4,$63,$EE,$FD,$97,$BE,$9F,$47,$F4,$77,$E9,$AF,$DF
	.dc.b	$BF,$9A,$FE,$C2,$FC,$E9,$F8,$FF,$FA,$DF,$4F,$7E,$1C,$2F,$75,$AF
	.dc.b	$D4,$BA,$E4,$AE,$BF,$47,$EF,$C3,$8E,$7D,$EB,$E3,$72,$7E,$0D,$E1
	.dc.b	$1E,$3F,$8A,$6B,$E3,$EE,$F8,$17,$BC,$BC,$55,$AD,$1E,$06,$B2,$F7
	.dc.b	$65,$CF,$2F,$66,$FF,$E8,$BD,$3C,$7B,$FA,$A1,$F9,$EF,$AB,$4E,$B7
	.dc.b	$C3,$E3,$B1,$8F,$3D,$3E,$2F,$D0,$7C,$0F,$B9,$7E,$39,$FC,$8F,$E8
	.dc.b	$FF,$37,$FB,$BB,$DE,$78,$BB,$AD,$73,$2B,$59,$E7,$DC,$F9,$1C,$FF
	.dc.b	$17,$3E,$BF,$39,$8E,$6F,$D5,$47,$82,$9A,$F8,$35,$F8,$54,$AD,$4F
	.dc.b	$93,$F6,$1C,$C6,$34,$F1,$5B,$C1,$FB,$7F,$81,$BF,$23,$FF,$18,$7C
	.dc.b	$DF,$F9,$63,$F5,$2E,$9A,$EA,$73,$F6,$69,$F0,$6F,$9F,$A6,$F4,$70
	.dc.b	$32,$E9,$F2,$D8,$7B,$4B,$3E,$DE,$F7,$AA,$73,$1E,$1F,$D6,$71,$A7
	.dc.b	$83,$3F,$77,$E8,$5D,$EB,$CD,$C5,$FE,$DC,$F0,$9F,$A2,$FE,$3F,$F2
	.dc.b	$97,$92,$FD,$A3,$FC,$CB,$FB,$03,$F6,$AF,$F5,$2F,$DB,$7F,$0C,$C3
	.dc.b	$85,$EE,$15,$BF,$BB,$7E,$DD,$BE,$6F,$E3,$D9,$F6,$CF,$F4,$59,$72
	.dc.b	$FC,$6A,$D6,$B3,$E9,$97,$87,$EE,$FF,$7D,$7E,$0C,$3C,$BA,$FD,$5C
	.dc.b	$3B,$0C,$BB,$6B,$E4,$9F,$F5,$EF,$E8,$3A,$9C,$E5,$E7,$FC,$5F,$1A
	.dc.b	$78,$2F,$FD,$87,$F6,$BD,$3E,$CF,$FD,$B7,$F9,$EF,$F0,$3D,$DF,$B6
	.dc.b	$3E,$71,$FB,$3E,$FF,$12,$5E,$EB,$5F,$19,$98,$F7,$F1,$32,$FF,$73
	.dc.b	$FE,$83,$87,$BF,$A5,$FE,$98,$C3,$FC,$4D,$5C,$63,$1F,$53,$F1,$23
	.dc.b	$FA,$55,$70,$D7,$CB,$AB,$F3,$DE,$11,$8F,$CB,$FE,$3F,$8A,$F8,$DF
	.dc.b	$AF,$DB,$FE,$F4,$FD,$D1,$F9,$73,$F5,$FD,$D6,$BE,$D5,$5A,$FD,$23
	.dc.b	$38,$7B,$31,$E9,$DF,$70,$E1,$7F,$C5,$38,$37,$7D,$FE,$9B,$D5,$0C
	.dc.b	$3C,$93,$F8,$AC,$BB,$E8,$FD,$1F,$EF,$FC,$8B,$F6,$A7,$D3,$3F,$28
	.dc.b	$7E,$06,$FD,$1B,$FA,$EF,$E1,$7C,$2F,$75,$A9,$FE,$C3,$E4,$E7,$F7
	.dc.b	$EF,$AB,$97,$5C,$AE,$1C,$32,$E9,$F9,$6F,$A2,$B5,$97,$3F,$73,$17
	.dc.b	$BC,$FF,$B6,$FE,$3D,$E1,$CB,$CD,$97,$97,$E5,$3F,$8B,$FA,$4F,$DA
	.dc.b	$BF,$A1,$7C,$BB,$F4,$FF,$F4,$3F,$E8,$0F,$DC,$3F,$C7,$9F,$86,$7E
	.dc.b	$BB,$0E,$1D,$7F,$E0,$BC,$38,$7C,$B3,$0E,$1F,$61,$FE,$BF,$F5,$FF
	.dc.b	$C0,$BA,$76,$D3,$FD,$27,$8F,$F7,$78,$C7,$BF,$FD,$5D,$D6,$3C,$F5
	.dc.b	$FF,$26,$7D,$1F,$F7,$19,$CE,$7D,$BE,$17,$18,$F8,$F2,$F4,$E1,$E8
	.dc.b	$F5,$7C,$33,$FE,$21,$FC,$8F,$FD,$DD,$FB,$1B,$F7,$AF,$F1,$B0,$FC
	.dc.b	$81,$F7,$1C,$2D,$7C,$2B,$74,$FD,$47,$53,$D2,$BD,$87,$7F,$C5,$3F
	.dc.b	$3A,$E0,$E6,$B1,$F1,$78,$1E,$3F,$C7,$7D,$75,$A9,$CE,$4F,$C7,$73
	.dc.b	$8F,$8F,$0F,$56,$6F,$DE,$3D,$1F,$83,$7F,$BB,$FE,$85,$FB,$E7,$F8
	.dc.b	$AF,$F3,$87,$E5,$EF,$D2,$77,$E0,$FF,$84,$2B,$52,$FD,$FB,$38,$7D
	.dc.b	$5E,$1F,$F2,$56,$5F,$EF,$BE,$0C,$47,$FA,$D7,$1E,$5E,$0B,$47,$E6
	.dc.b	$FE,$CC,$3C,$9A,$78,$FD,$4C,$61,$C2,$FC,$BF,$96,$7F,$93,$3F,$3F
	.dc.b	$FF,$0B,$7E,$7C,$FC,$85,$FB,$6F,$FB,$1B,$F5,$97,$EA,$DF,$E9,$7F
	.dc.b	$93,$E1,$C3,$FE,$5F,$C3,$85,$EE,$BF,$D3,$9F,$A8,$EA,$95,$D7,$F1
	.dc.b	$2F,$F1,$46,$7E,$0C,$FB,$D7,$F9,$36,$31,$8F,$3F,$C4,$FF,$C8,$91
	.dc.b	$8F,$C5,$DA,$F9,$3F,$58,$CE,$1E,$3C,$7D,$35,$AD,$7F,$D5,$DF,$92
	.dc.b	$78,$FF,$74,$7E,$B8,$FD,$95,$F9,$EF,$F1,$7F,$C2,$9F,$A9,$BF,$29
	.dc.b	$E1,$7B,$AD,$4B,$F0,$1C,$E1,$EC,$DB,$FD,$05,$97,$6F,$D7,$F1,$08
	.dc.b	$F0,$CF,$83,$B3,$FD,$8D,$F7,$FE,$AA,$D6,$7D,$E7,$FD,$95,$F1,$BF
	.dc.b	$E6,$5F,$D0,$7F,$C9,$9F,$8E,$BF,$49,$FB,$1F,$94,$7F,$89,$BF,$0B
	.dc.b	$FE,$21,$F8,$9F,$FA,$22,$F7,$5A,$FB,$A5,$7F,$15,$FD,$9B,$5A,$FF
	.dc.b	$A6,$72,$E9,$F8,$FF,$F9,$77,$1E,$3D,$76,$B5,$97,$3F,$DA,$BF,$CB
	.dc.b	$DF,$0E,$BD,$ED,$E3,$FC,$97,$D1,$87,$FA,$97,$89,$F9,$93,$F5,$2F
	.dc.b	$F7,$37,$CC,$BF,$52,$FE,$BD,$E2,$79,$30,$FA,$F7,$BF,$01,$7B,$FA
	.dc.b	$37,$E5,$9B,$CF,$AF,$E8,$9E,$7C,$78,$E9,$67,$C2,$F8,$34,$FE,$B5
	.dc.b	$C1,$E9,$AF,$36,$9D,$F3,$F2,$7E,$05,$EA,$9E,$7A,$F8,$B0,$FE,$E9
	.dc.b	$F8,$F7,$9F,$C0,$FE,$ED,$FB,$8B,$CE,$FF,$2B,$71,$BE,$D5,$F2,$FB
	.dc.b	$E4,$5E,$FE,$E9,$7B,$FA,$0D,$7B,$1E,$BC,$FA,$EF,$FE,$B3,$EA,$FF
	.dc.b	$47,$47,$83,$5E,$FF,$4D,$E7,$8C,$65,$D2,$FF,$D5,$DF,$9D,$70,$FC
	.dc.b	$27,$F4,$7A,$BF,$5C,$7D,$6B,$F7,$2F,$F6,$8F,$53,$0E,$1D,$9E,$8B
	.dc.b	$DD,$3C,$EF,$F6,$BD,$6B,$3F,$F6,$43,$C3,$2E,$D9,$3F,$22,$F1,$D7
	.dc.b	$EA,$34,$F1,$E5,$E2,$F9,$0F,$F6,$27,$EA,$1E,$EF,$55,$FA,$9F,$C8
	.dc.b	$E6,$FF,$12,$7E,$5C,$FC,$CB,$FD,$CF,$7B,$AD,$75,$AB,$F9,$F7,$22
	.dc.b	$B5,$C0,$C7,$FC,$5D,$9F,$83,$6F,$DD,$E9,$E0,$AD,$65,$E9,$9F,$F7
	.dc.b	$AF,$5F,$D2,$7C,$BC,$BF,$73,$D5,$7D,$8F,$F8,$77,$5F,$A5,$5F,$C0
	.dc.b	$CD,$E3,$FC,$4E,$5F,$B7,$74,$9F,$E1,$66,$AF,$F8,$BB,$93,$97,$07
	.dc.b	$DA,$8F,$EB,$50,$C3,$FD,$9E,$F2,$CF,$D5,$8F,$E5,$7F,$26,$D7,$E2
	.dc.b	$FE,$99,$CF,$73,$F2,$3C,$E1,$C7,$3F,$CB,$E2,$FA,$2F,$F7,$C6,$3C
	.dc.b	$67,$3F,$8E,$FF,$BE,$FF,$3F,$F8,$9D,$5E,$EC,$3F,$1D,$B2,$73,$7F
	.dc.b	$92,$9F,$F6,$64,$E7,$2F,$83,$7F,$1B,$E8,$BA,$46,$2F,$F3,$39,$A3
	.dc.b	$E0,$7E,$79,$F4,$F9,$BF,$0F,$5E,$8E,$8C,$D7,$CC,$9C,$E7,$F5,$F3
	.dc.b	$61,$C7,$0F,$C0,$65,$D6,$73,$87,$19,$FC,$26,$59,$FD,$A3,$F0,$53
	.dc.b	$E0,$E7,$FC,$B5,$A7,$37,$F0,$E3,$15,$F7,$8E,$18,$73,$71,$3F,$8C
	.dc.b	$C3,$C5,$F9,$47,$6E,$4F,$4E,$DF,$7E,$31,$5F,$7B,$14,$62,$18,$FA
	.dc.b	$FE,$EB,$CF,$C1,$E7,$C3,$F1,$11,$8A,$FC,$AE,$0F,$C9,$BD,$BC,$3F
	.dc.b	$32,$7F,$51,$39,$D7,$D5,$39,$E1,$FC,$97,$E3,$7F,$12,$78,$70,$BF
	.dc.b	$61,$4E,$71,$F9,$11,$CE,$BF,$4C,$7F,$73,$8F,$B7,$A7,$1C,$FC,$71
	.dc.b	$F6,$AB,$FD,$45,$5F,$2F,$5F,$64,$62,$39,$C7,$D5,$CB,$C5,$BF,$C5
	.dc.b	$7F,$81,$57,$C8,$AF,$A9,$C1,$F8,$69,$DC,$63,$0F,$E2,$7C,$1F,$91
	.dc.b	$BF,$B7,$3E,$89,$FD,$EE,$7E,$FE,$2F,$C0,$C7,$E7,$EB,$7E,$4F,$E6
	.dc.b	$6F,$F0,$26,$CF,$C0,$CF,$F8,$63,$8F,$B7,$27,$E5,$63,$F3,$97,$F9
	.dc.b	$4A,$FE,$19,$C5,$E7,$EB,$F9,$D7,$BD,$7F,$03,$58,$3F,$35,$87,$F5
	.dc.b	$2C,$7D,$5C,$DF,$B4,$E4,$FD,$FF,$87,$F5,$E3,$18,$7F,$58,$FC,$67
	.dc.b	$18,$FE,$CE,$FF,$AD,$F0,$F2,$8C,$57,$F0,$77,$F0,$1D,$5F,$C1,$E3
	.dc.b	$FB,$0F,$63,$B7,$35,$FC,$3E,$73,$DB,$88,$FE,$51,$CD,$FC,$6B,$6F
	.dc.b	$A1,$8F,$3B,$76,$FD,$3F,$03,$F9,$A7,$0F,$F2,$5F,$03,$FD,$9D,$CD
	.dc.b	$FC,$CB,$9F,$F8,$44,$E7,$C0,$FC,$A5,$F0,$E2,$FE,$75,$F1,$3C,$F7
	.dc.b	$FE,$67,$7F,$CB,$FE,$8F,$F0,$1E,$27,$F4,$AE,$DF,$EA,$7E,$37,$E7
	.dc.b	$63,$1C,$7F,$D4,$6B,$18,$7F,$4C,$C3,$FA,$AF,$8B,$CA,$31,$C5,$FD
	.dc.b	$3B,$C7,$FC,$67,$93,$FD,$A7,$B7,$FB,$7F,$8D,$FA,$98,$C7,$93,$FD
	.dc.b	$DF,$CA,$C6,$3F,$DA,$F8,$BF,$18,$FE,$37,$9B,$FB,$D3,$9C,$FF,$B9
	.dc.b	$E7,$DA,$31,$9F,$F7,$17,$F6,$DE,$77,$F8,$3B,$83,$EB,$F9,$BF,$E1
	.dc.b	$6F,$47,$FC,$37,$97,$F8,$A6,$FF,$CB,$D5,$AE,$7F,$F0,$D4,$E7,$A5
	.dc.b	$3F,$DB,$EB,$CA,$76,$3F,$CD,$2F,$F8,$91,$CC,$F1,$8E,$9E,$7F,$EB
	.dc.b	$F2,$F2,$6B,$FE,$29,$C7,$DD,$1F,$AB,$DB,$FC,$7F,$EA,$7F,$8D,$F0
	.dc.b	$FF,$23,$E3,$E7,$9C,$FA,$DF,$E4,$BF,$53,$C5,$A7,$F9,$43,$D9,$F2
	.dc.b	$B8,$AF,$F2,$4B,$E4,$CB,$B7,$B1,$FE,$47,$F5,$DF,$11,$E5,$AF,$F2
	.dc.b	$5F,$B5,$FE,$4F,$F6,$B1,$1F,$E4,$AD,$3F,$CB,$5E,$E7,$83,$E0,$BF
	.dc.b	$CF,$9F,$02,$F7,$FF,$CF,$9F,$58,$F8,$7D,$7C,$1F,$F9,$EF,$E1,$BF
	.dc.b	$CF,$5C,$BF,$47,$1F,$F3,$CE,$9E,$53,$FC,$EF,$97,$AA,$75,$7F,$E7
	.dc.b	$99,$F1,$D7,$83,$2F,$F3,$C4,$7F,$35,$F9,$0F,$F2,$6F,$27,$FA,$4F
	.dc.b	$A3,$FD,$4B,$EC,$E6,$BF,$D3,$F3,$9F,$8C,$FF,$52,$66,$C3,$8F,$B5
	.dc.b	$E1,$79,$F8,$97,$0E,$33,$FE,$A3,$CB,$E4,$C7,$53,$FD,$47,$19,$F8
	.dc.b	$AF,$E5,$D7,$FE,$C3,$E3,$FF,$6A,$FE,$60,$F9,$5E,$1F,$2B,$FE,$D4
	.dc.b	$F9,$2F,$F6,$14,$BE,$47,$3C,$BF,$E7,$5D,$3F,$1B,$97,$FB,$41,$A7
	.dc.b	$FB,$2F,$E5,$7F,$DA,$4F,$07,$FD,$BD,$F2,$9E,$28,$D6,$3D,$BE,$5B
	.dc.b	$FA,$E7,$5F,$F3,$BF,$9A,$FF,$79,$FC,$E7,$FB,$DF,$E8,$1F,$35,$AC
	.dc.b	$7F,$DF,$FE,$D7,$FB,$D6,$5F,$41,$99,$FF,$80,$1F,$41,$F6,$72,$E9
	.dc.b	$6F,$A0,$C5,$72,$A7,$CF,$B3,$7C,$FB,$E8,$3F,$AB,$65,$CA,$F7,$5A
	.dc.b	$F9,$DC,$E3,$CF,$1E,$58,$B5,$7D,$0F,$97,$E9,$5E,$FB,$E9,$BC,$9F
	.dc.b	$4D,$E5,$FA,$67,$D4,$6E,$B5,$C4,$AD,$7D,$3E,$70,$F5,$EC,$CB,$AF
	.dc.b	$DE,$E1,$EF,$3D,$B0,$7D,$4F,$A2,$F8,$65,$DB,$2F,$8D,$F1,$19,$C7
	.dc.b	$8E,$1E,$11,$AB,$EA,$BE,$87,$23,$EB,$5F,$5E,$FA,$F7,$D7,$BE,$CB
	.dc.b	$FD,$11,$F6,$1C,$2F,$75,$A9,$FF,$46,$7D,$96,$70,$F6,$E3,$D7,$EC
	.dc.b	$AE,$1C,$32,$E9,$F3,$3E,$BA,$F1,$5F,$86,$7C,$1F,$3B,$9B,$E1,$97
	.dc.b	$FA,$67,$EF,$BF,$C5,$9F,$87,$FF,$82,$F9,$1F,$72,$FB,$87,$DD,$BE
	.dc.b	$E5,$F7,$4F,$BA,$7D,$BF,$2E,$DF,$F3,$66,$5C,$B1,$E3,$87,$0B,$FE
	.dc.b	$83,$9F,$7C,$DF,$7B,$FC,$3E,$31,$77,$1E,$75,$E3,$D3,$C5,$A3,$EF
	.dc.b	$7F,$C5,$58,$70,$9C,$FD,$EE,$30,$E3,$8F,$8B,$0F,$47,$DA,$F9,$E7
	.dc.b	$F1,$99,$75,$D7,$CD,$F7,$D9,$F3,$1F,$68,$E3,$9F,$1A,$FC,$01,$E4
	.dc.b	$FF,$AE,$3E,$19,$C5,$FE,$35,$C7,$8E,$1C,$2F,$7F,$80,$75,$CB,$A6
	.dc.b	$DF,$D6,$30,$E1,$0F,$9B,$F8,$52,$C7,$9E,$2D,$3C,$5F,$27,$EF,$AD
	.dc.b	$61,$E7,$AD,$66,$F8,$6D,$63,$CB,$1F,$56,$CF,$C0,$7E,$27,$E0,$8F
	.dc.b	$C2,$DF,$85,$3F,$08,$FF,$7F,$FE,$12,$FC,$23,$FB,$AF,$2F,$C0,$E1
	.dc.b	$FF,$56,$63,$C7,$0E,$1F,$50,$C3,$87,$E1,$1D,$B3,$EB,$87,$C1,$38
	.dc.b	$AF,$C2,$BE,$0E,$3C,$F4,$F0,$FE,$13,$EF,$8C,$65,$D7,$F0,$AC,$61
	.dc.b	$E8,$BF,$3E,$CF,$C3,$3C,$1B,$79,$36,$F5,$7E,$19,$F8,$3F,$8A,$7E
	.dc.b	$1C,$EF,$FA,$5F,$87,$BF,$0F,$7E,$22,$FA,$FE,$17,$BF,$05,$7B,$A7
	.dc.b	$B5,$AC,$7E,$DE,$5D,$BF,$11,$B8,$F1,$C3,$B7,$27,$DB,$AF,$1D,$F8
	.dc.b	$B4,$79,$79,$C3,$84,$79,$3D,$AE,$DA,$F8,$A3,$CF,$E1,$E3,$3F,$24
	.dc.b	$FF,$AA,$FF,$1C,$FF,$45,$7E,$2E,$F7,$1F,$8C,$3F,$15,$7E,$34,$FC
	.dc.b	$61,$F6,$EF,$C6,$39,$78,$9F,$EE,$BC,$B9,$63,$C7,$0E,$16,$FD,$43
	.dc.b	$76,$BD,$BF,$E3,$8D,$3C,$7A,$3F,$1B,$FF,$09,$46,$32,$E9,$A7,$E8
	.dc.b	$EF,$9E,$1E,$0D,$7C,$F2,$FC,$6F,$FD,$5D,$87,$19,$CE,$8F,$C7,$38
	.dc.b	$E9,$F0,$27,$3F,$05,$E6,$AF,$F6,$7D,$7F,$1E,$FB,$27,$E2,$2F,$C3
	.dc.b	$5F,$8F,$6F,$F1,$4E,$9F,$0A,$BD,$D7,$D8,$F4,$5B,$7F,$AE,$73,$EB
	.dc.b	$C5,$F0,$41,$F3,$B1,$FF,$58,$BF,$E7,$4A,$E2,$EA,$7C,$B8,$F7,$CF
	.dc.b	$C9,$3E,$3D,$3D,$10,$FC,$7B,$E1,$CF,$1C,$7C,$FA,$79,$BF,$1F,$FA
	.dc.b	$BF,$91,$7B,$3F,$24,$FE,$B7,$F8,$87,$F0,$1F,$A0,$7E,$47,$AE,$35
	.dc.b	$6B,$D2,$53,$F2,$3D,$7E,$4B,$F6,$B6,$F6,$E7,$D5,$F8,$01,$F0,$7C
	.dc.b	$93,$4F,$E2,$1B,$7A,$2B,$5A,$78,$2B,$59,$FF,$90,$B9,$7E,$2D,$6B
	.dc.b	$5F,$26,$AF,$D4,$3D,$11,$8F,$C9,$AE,$3C,$A9,$ED,$63,$5F,$81,$F8
	.dc.b	$AF,$E6,$3F,$1D,$7E,$52,$FC,$A5,$F9,$3F,$F8,$A7,$F2,$9F,$EC,$9F
	.dc.b	$AB,$EE,$B5,$39,$F9,$7F,$E1,$BC,$5E,$6C,$79,$E3,$F1,$78,$91,$DB
	.dc.b	$F2,$BC,$3A,$C3,$B7,$CB,$FC,$38,$C7,$CE,$B5,$FE,$0B,$87,$87,$1F
	.dc.b	$2C,$E7,$37,$D9,$E7,$2F,$27,$DA,$3F,$14,$FF,$60,$FE,$5C,$FC,$7B
	.dc.b	$FB,$0F,$E5,$AF,$CB,$BF,$C6,$5F,$0C,$F1,$78,$6F,$F9,$81,$87,$1F
	.dc.b	$96,$61,$C7,$E8,$58,$3F,$0C,$E3,$87,$E8,$30,$E0,$FF,$AF,$B4,$F1
	.dc.b	$E8,$FC,$9B,$FD,$5B,$18,$CF,$AE,$BF,$E4,$5F,$9A,$FD,$4C,$E7,$5F
	.dc.b	$07,$8D,$8C,$BC,$D9,$F9,$67,$C9,$AB,$F4,$EF,$1D,$7A,$B5,$F8,$39
	.dc.b	$FC,$0F,$D4,$33,$F5,$2F,$C5,$9F,$99,$3F,$32,$7E,$66,$FC,$33,$E6
	.dc.b	$7E,$29,$C2,$F7,$5A,$9F,$A1,$F9,$3E,$70,$F6,$EB,$F2,$F2,$ED,$F9
	.dc.b	$A6,$23,$F8,$2C,$78,$67,$C1,$9F,$7E,$4F,$75,$6B,$3E,$F3,$E3,$FB
	.dc.b	$8C,$CB,$0E,$5A,$78,$F0,$FF,$07,$1F,$6E,$75,$87,$2D,$3F,$0D,$FA
	.dc.b	$5F,$9E,$F8,$7D,$8D,$CF,$C5,$F9,$B7,$E2,$FF,$3C,$FC,$27,$E7,$2F
	.dc.b	$CE,$BE,$37,$C0,$7F,$93,$7A,$9F,$9D,$7A,$6F,$75,$AF,$B4,$FC,$4C
	.dc.b	$BF,$11,$E9,$C3,$FC,$9B,$2E,$9F,$5D,$F8,$4C,$78,$EF,$6B,$59,$73
	.dc.b	$CF,$F9,$2F,$1A,$F7,$AF,$8F,$EE,$FF,$13,$18,$DB,$C3,$F9,$F7,$83
	.dc.b	$26,$BE,$BD,$3F,$A5,$57,$9B,$E9,$F5,$F8,$E3,$F2,$4F,$D7,$FE,$61
	.dc.b	$FC,$F3,$F3,$D7,$E7,$AF,$CD,$B7,$5A,$D1,$5A,$FC,$EF,$3E,$8F,$AE
	.dc.b	$59,$75,$F4,$F0,$DC,$FB,$7D,$87,$AA,$BF,$A5,$CE,$5E,$BF,$63,$97
	.dc.b	$9E,$7E,$1C,$FC,$BF,$39,$E4,$CB,$FD,$C1,$1E,$8F,$73,$F8,$37,$E4
	.dc.b	$3F,$BF,$FC,$E9,$CD,$FE,$A0,$F2,$1F,$99,$3F,$3D,$FE,$CD,$F9,$F7
	.dc.b	$0D,$BF,$AE,$FE,$7E,$FD,$0D,$FA,$0B,$F4,$3F,$E9,$BF,$40,$FC,$AA
	.dc.b	$78,$8E,$0F,$E1,$D8,$F1,$C3,$87,$CC,$FF,$85,$EF,$FC,$C9,$7B,$CF
	.dc.b	$BE,$7F,$EB,$87,$F8,$66,$3C,$A3,$C3,$18,$37,$87,$4C,$9F,$43,$FC
	.dc.b	$32,$F8,$4E,$7D,$2C,$5F,$0D,$7D,$39,$BF,$43,$F8,$12,$DB,$D1,$5A
	.dc.b	$F8,$BC,$C6,$37,$F5,$7E,$49,$8D,$FE,$2D,$F4,$FD,$47,$F9,$B7,$DC
	.dc.b	$7F,$2F,$F9,$EF,$F1,$BF,$1F,$FA,$D7,$F1,$F7,$E7,$FF,$D6,$BF,$45
	.dc.b	$DE,$FF,$E0,$CB,$DF,$E1,$6B,$DF,$CC,$FF,$8F,$B1,$FF,$48,$5F,$F8
	.dc.b	$4B,$E5,$1F,$93,$BF,$46,$7C,$1F,$FB,$EB,$F0,$F7,$E6,$6F,$D1,$FF
	.dc.b	$A5,$7D,$BF,$F4,$3C,$78,$E1,$C3,$E1,$98,$70,$FA,$8F,$EA,$1F,$6D
	.dc.b	$FC,$D3,$C1,$FD,$47,$E8,$4F,$D2,$3F,$CD,$FF,$9E,$FE,$07,$8B,$FC
	.dc.b	$A7,$C5,$FD,$1E,$F7,$5A,$F7,$B9,$CE,$7D,$26,$3C,$FF,$05,$FD,$A4
	.dc.b	$E8,$FE,$AF,$87,$2F,$C1,$3F,$AD,$64,$FD,$2F,$FA,$E7,$95,$FC,$F3
	.dc.b	$F4,$BF,$EE,$BF,$A5,$BF,$4C,$7E,$3E,$FD,$07,$FB,$56,$1C,$3E,$3D
	.dc.b	$87,$0B,$DF,$16,$B6,$FF,$1F,$E7,$D7,$A5,$A7,$5F,$C9,$3F,$AD,$E7
	.dc.b	$CF,$2E,$5F,$92,$7F,$5E,$FD,$3F,$FB,$87,$E9,$DF,$E9,$4F,$95,$7E
	.dc.b	$63,$FA,$DF,$BE,$7E,$9E,$FD,$37,$FB,$F6,$1C,$3A,$58,$70,$F4,$7F
	.dc.b	$CF,$B7,$F5,$BA,$17,$BA,$D6,$3F,$CA,$A3,$19,$77,$F9,$1F,$EE,$3F
	.dc.b	$A1,$7F,$A3,$3F,$18,$7E,$A0,$FD,$45,$F9,$67,$F8,$A7,$F2,$07,$D2
	.dc.b	$7F,$84,$70,$E1,$7B,$AD,$7D,$3E,$7B,$3F,$D1,$D9,$74,$CB,$9F,$DA
	.dc.b	$7E,$DA,$74,$7F,$A3,$67,$57,$FE,$37,$F8,$2F,$F0,$2F,$37,$F8,$5F
	.dc.b	$E8,$9F,$A5,$BD,$DF,$F2,$37,$E6,$4F,$D1,$7F,$D7,$DE,$57,$F8,$97
	.dc.b	$0E,$17,$BF,$C0,$97,$BF,$C7,$3F,$C8,$7D,$7F,$3F,$7F,$F3,$DE,$9D
	.dc.b	$B4,$EB,$A7,$EB,$AB,$54,$AD,$63,$D2,$BF,$27,$F8,$E7,$F8,$B7,$C8
	.dc.b	$7A,$BF,$E5,$BF,$CA,$BF,$D5,$9F,$3B,$F3,$3F,$12,$FF,$1D,$5E,$EB
	.dc.b	$5F,$8A,$7F,$4F,$E5,$7D,$CE,$B6,$FF,$47,$E7,$D3,$56,$9D,$3F,$37
	.dc.b	$FF,$15,$E7,$CE,$73,$97,$F5,$9F,$3F,$FC,$5F,$F5,$5F,$DF,$BF,$0A
	.dc.b	$7E,$A6,$FD,$4D,$F2,$CF,$9C,$7E,$6E,$FD,$53,$FC,$A1,$8F,$1C,$38
	.dc.b	$5E,$EB,$FA,$FF,$4A,$B5,$8F,$4C,$5E,$E7,$8B,$8F,$FC,$09,$A7,$8A
	.dc.b	$B5,$EB,$7F,$91,$3E,$0B,$FC,$AD,$F6,$CF,$D1,$9F,$AA,$BE,$B5,$FA
	.dc.b	$7F,$F0,$8F,$A4,$7E,$91,$FE,$65,$C7,$8E,$1C,$2F,$7F,$0B,$FD,$E3
	.dc.b	$F4,$3D,$6F,$FE,$BA,$D3,$B6,$9D,$74,$FE,$27,$F5,$3E,$3A,$D6,$5C
	.dc.b	$E7,$F8,$BF,$D5,$FF,$96,$BF,$2E,$FF,$36,$7C,$AB,$F0,$3F,$F2,$7F
	.dc.b	$EA,$EF,$D1,$3E,$27,$0B,$F1,$3F,$F3,$D6,$1C,$2F,$75,$AF,$D5,$F3
	.dc.b	$AF,$FB,$3B,$2E,$9F,$80,$FF,$9B,$30,$E1,$F5,$AD,$3B,$76,$FD,$1C
	.dc.b	$38,$9F,$EA,$6C,$FA,$C3,$F0,$7F,$F3,$8F,$E6,$1F,$E8,$3E,$17,$EA
	.dc.b	$7F,$C3,$BE,$CB,$F8,$0F,$E2,$7F,$CD,$3F,$5D,$FF,$48,$5E,$EB,$5E
	.dc.b	$8A,$BF,$9C,$FD,$8E,$B5,$FF,$72,$E5,$D3,$F5,$3F,$F4,$1E,$3C,$7E
	.dc.b	$32,$D6,$B2,$E7,$8F,$F4,$0F,$21,$7B,$D7,$C7,$F8,$EF,$FA,$23,$F1
	.dc.b	$9F,$F4,$D7,$D4,$7F,$20,$EB,$7E,$B2,$FC,$C3,$F8,$37,$DF,$BF,$69
	.dc.b	$FF,$54,$61,$C3,$E4,$BE,$95,$EF,$E0,$96,$FD,$47,$7C,$5F,$EF,$ED
	.dc.b	$3B,$3F,$E9,$FC,$79,$7E,$4F,$FE,$9E,$7A,$6B,$E2,$D7,$F8,$B7,$C0
	.dc.b	$79,$EF,$96,$BE,$09,$FE,$35,$C5,$FE,$A2,$FC,$4F,$FD,$69,$FA,$7B
	.dc.b	$F5,$BF,$D9,$7E,$A4,$FD,$11,$F9,$4B,$F5,$D7,$E4,$3F,$EC,$3C,$78
	.dc.b	$E1,$C2,$F7,$4F,$D7,$75,$B0,$67,$DF,$DD,$FF,$5B,$E3,$CB,$E0,$2D
	.dc.b	$6E,$B5,$AF,$8F,$5F,$E3,$B3,$D0,$E7,$A7,$86,$3F,$85,$47,$FB,$5F
	.dc.b	$F4,$BF,$F6,$87,$4B,$F5,$47,$CF,$3F,$5E,$FF,$3B,$FD,$83,$D5,$7E
	.dc.b	$BF,$C7,$2F,$46,$3C,$70,$E1,$F8,$9F,$FA,$16,$DE,$6E,$EF,$FA,$36
	.dc.b	$7D,$9F,$F6,$B6,$9E,$28,$7C,$E7,$FB,$5A,$31,$8F,$4D,$3F,$55,$D2
	.dc.b	$9C,$ED,$E6,$F3,$FF,$DB,$3F,$8B,$7F,$BA,$FF,$13,$FE,$D1,$F8,$33
	.dc.b	$F5,$4F,$F0,$5F,$EA,$FF,$E5,$4F,$D1,$3F,$DF,$17,$BF,$A3,$FF,$27
	.dc.b	$56,$BE,$45,$FB,$06,$B8,$FE,$B6,$7F,$DC,$FF,$29,$FF,$75,$E3,$C7
	.dc.b	$4B,$5A,$CB,$9F,$D9,$35,$FE,$DD,$C3,$FE,$BB,$AD,$5F,$96,$B5,$F4
	.dc.b	$BF,$EF,$2F,$D2,$BF,$09,$4F,$C7,$DF,$85,$3F,$43,$7E,$C3,$FE,$05
	.dc.b	$FA,$2F,$93,$F0,$0D,$59,$72,$C7,$8E,$1C,$37,$61,$F4,$2B,$81,$BC
	.dc.b	$BB,$67,$FE,$DB,$32,$7E,$5F,$0F,$1F,$E8,$D9,$8C,$9E,$79,$F3,$69
	.dc.b	$E5,$EC,$F6,$4F,$19,$CE,$5E,$2F,$CE,$7F,$45,$AF,$E9,$6B,$F4,$9F
	.dc.b	$68,$FD,$8B,$FC,$1F,$FA,$17,$F2,$4F,$C0,$7F,$5F,$F0,$7E,$FF,$47
	.dc.b	$E3,$DE,$FE,$B9,$7B,$F5,$5F,$A5,$EF,$3E,$B9,$FF,$86,$9F,$76,$9E
	.dc.b	$15,$8F,$87,$F6,$5F,$74,$E6,$3B,$63,$E3,$8F,$4E,$7E,$2C,$3C,$FA
	.dc.b	$7F,$97,$7F,$67,$FD,$F7,$57,$EA,$B3,$FE,$01,$FA,$AB,$DF,$7E,$10
	.dc.b	$FC,$19,$F9,$DB,$F3,$6F,$F1,$77,$E4,$4C,$B9,$64,$CB,$96,$3C,$70
	.dc.b	$E1,$EA,$FF,$82,$6F,$75,$AB,$FB,$91,$8C,$FF,$C7,$11,$8F,$5B,$F2
	.dc.b	$35,$D3,$4F,$1E,$9F,$ED,$5A,$F1,$56,$AF,$D3,$5A,$F8,$AF,$7F,$83
	.dc.b	$F9,$AF,$D0,$DF,$A4,$3E,$BD,$FB,$2F,$F9,$FB,$F3,$A7,$E6,$AF,$8C
	.dc.b	$F7,$63,$C7,$0E,$17,$BA,$7E,$A5,$AD,$FF,$95,$67,$DB,$F3,$EF,$CA
	.dc.b	$71,$FE,$8A,$73,$8F,$85,$F8,$39,$75,$D7,$C7,$E7,$36,$7D,$1F,$AA
	.dc.b	$F8,$D6,$A5,$F4,$BF,$73,$D3,$FD,$BB,$FE,$76,$F8,$FF,$F1,$B7,$DB
	.dc.b	$BF,$5A,$FF,$12,$FE,$A3,$F6,$6F,$7F,$3D,$FE,$11,$FB,$6F,$F2,$4D
	.dc.b	$7F,$8A,$BF,$45,$D6,$BF,$D1,$73,$FF,$40,$F5,$FB,$31,$E3,$A5,$AD
	.dc.b	$65,$CF,$4F,$07,$E2,$BE,$59,$CD,$79,$3D,$8F,$E1,$BC,$BF,$CA,$E7
	.dc.b	$F9,$DF,$EC,$DF,$EC,$9F,$8A,$7E,$40,$FD,$77,$FC,$27,$E8,$7E,$47
	.dc.b	$E2,$3F,$A4,$DE,$EB,$5E,$67,$FA,$13,$2E,$99,$7F,$A1,$6B,$FD,$A8
	.dc.b	$FE,$42,$F8,$FA,$BF,$59,$E7,$F7,$FF,$D3,$F1,$F0,$7A,$DF,$A8,$9C
	.dc.b	$EB,$E4,$D5,$F8,$9F,$EB,$3E,$57,$FB,$BF,$CA,$FF,$82,$3F,$67,$7E
	.dc.b	$77,$FE,$52,$F5,$3F,$D7,$5C,$2B,$DF,$D0,$2F,$7F,$70,$BF,$EF,$B7
	.dc.b	$BE,$2F,$EE,$19,$7F,$85,$1F,$D4,$E3,$E0,$FC,$13,$F1,$EF,$4C,$79
	.dc.b	$E1,$E5,$AE,$39,$77,$CF,$FC,$CB,$ED,$7F,$04,$FA,$1F,$F1,$4F,$ED
	.dc.b	$0F,$DA,$5F,$9C,$FF,$AB,$7E,$89,$FB,$33,$ED,$7F,$81,$61,$C2,$F7
	.dc.b	$EE,$7F,$8B,$E9,$FA,$36,$B6,$FF,$14,$67,$D5,$FE,$03,$87,$83,$E7
	.dc.b	$FF,$4E,$F3,$D3,$C1,$E2,$79,$6B,$8E,$7D,$B5,$7B,$5F,$C7,$3E,$B3
	.dc.b	$FC,$A5,$E4,$3F,$49,$78,$CF,$63,$FC,$19,$FB,$6F,$F1,$9B,$DD,$6A
	.dc.b	$7F,$C2,$1D,$F9,$D3,$FC,$A5,$F5,$BF,$BA,$BE,$B1,$FC,$61,$E3,$8F
	.dc.b	$5F,$C9,$7C,$53,$FE,$59,$CF,$C5,$3E,$1F,$C9,$FF,$98,$FD,$CF,$F9
	.dc.b	$F7,$A7,$FD,$B3,$FB,$5B,$F6,$4F,$F6,$57,$E6,$FF,$CB,$6B,$5E,$52
	.dc.b	$B5,$F8,$97,$E4,$3C,$75,$73,$D7,$F9,$5B,$C2,$BA,$7E,$21,$E0,$9F
	.dc.b	$F4,$47,$DE,$7D,$9C,$7A,$66,$FA,$5F,$E8,$BE,$6F,$FA,$9B,$F0,$37
	.dc.b	$E2,$FF,$D1,$7C,$C7,$EA,$0F,$C1,$BF,$A1,$5E,$EB,$52,$FD,$AF,$3A
	.dc.b	$FF,$A9,$B1,$E7,$EE,$B2,$FE,$67,$3A,$3A,$7E,$0F,$FE,$BC,$C7,$C1
	.dc.b	$F6,$DF,$D6,$3F,$02,$FF,$67,$7E,$21,$FE,$74,$FD,$B3,$FC,$AF,$E0
	.dc.b	$7F,$53,$9C,$FA,$49,$FF,$2D,$FD,$F6,$73,$FF,$66,$FA,$FF,$ED,$AA
	.dc.b	$D7,$5F,$59,$CE,$5D,$7E,$BF,$FB,$3F,$E2,$9F,$EE,$9F,$84,$FF,$5B
	.dc.b	$7C,$13,$F6,$D7,$A1,$FD,$8E,$B5,$F0,$BF,$EA,$89,$FF,$69,$E6,$9C
	.dc.b	$DF,$F3,$CC,$F8,$1F,$EF,$AF,$C0,$3F,$AA,$65,$DB,$E2,$3F,$BB,$FE
	.dc.b	$35,$FE,$32,$F1,$9E,$13,$DB,$FE,$B5,$F8,$6B,$F4,$1F,$D7,$DF,$F7
	.dc.b	$53,$85,$6F,$E4,$55,$BF,$75,$F9,$6E,$F6,$FF,$41,$67,$DB,$3E,$B9
	.dc.b	$FE,$EA,$74,$27,$58,$F4,$9F,$F3,$67,$0B,$F2,$0F,$F2,$4F,$D8,$FE
	.dc.b	$2B,$F2,$07,$D6,$BD,$57,$D8,$BD,$CF,$F3,$BE,$3C,$75,$FF,$4A,$63
	.dc.b	$C7,$0E,$17,$BF,$A2,$FF,$09,$67,$DB,$C6,$69,$DB,$4E,$BE,$DF,$F8
	.dc.b	$82,$B5,$39,$C1,$F8,$AF,$F8,$93,$F0,$7F,$F2,$FF,$E4,$EF,$C9,$DF
	.dc.b	$86,$BF,$25,$7E,$4E,$FA,$17,$EC,$4F,$9D,$FF,$12,$65,$CB,$1E,$38
	.dc.b	$70,$BF,$DB,$F2,$E8,$FF,$58,$67,$DF,$3E,$D9,$F5,$FC,$0B,$C9,$5A
	.dc.b	$C7,$9C,$F8,$7D,$0F,$F1,$DF,$E5,$3F,$E7,$6F,$33,$FC,$71,$9B,$F2
	.dc.b	$B7,$E5,$7F,$A3,$FD,$75,$F8,$D7,$F8,$E7,$0E,$1F,$61,$EB,$BD,$FA
	.dc.b	$CB,$70,$EF,$7F,$F6,$36,$9D,$AF,$7E,$4F,$8A,$73,$8F,$3D,$3F,$65
	.dc.b	$D7,$FE,$50,$FC,$BF,$FD,$17,$F5,$EF,$CA,$7F,$53,$F8,$F3,$E9,$1F
	.dc.b	$91,$FF,$94,$7E,$E7,$FC,$9D,$87,$0B,$DD,$6B,$37,$EA,$DA,$D7,$FD
	.dc.b	$B5,$9F,$5F,$69,$A7,$5D,$1F,$23,$E3,$BD,$E7,$D2,$7F,$DD,$71,$FE
	.dc.b	$5D,$F9,$3F,$F5,$0F,$13,$F3,$47,$E6,$4F,$CD,$1E,$DF,$EA,$BE,$81
	.dc.b	$F9,$BF,$1C,$BF,$DD,$98,$F1,$C3,$85,$EF,$6F,$F3,$57,$E0,$DF,$EA
	.dc.b	$EF,$AC,$F5,$FD,$2F,$F9,$47,$F3,$6E,$7F,$36,$FC,$C3,$7A,$7F,$BD
	.dc.b	$EF,$7F,$7A,$BD,$FE,$DB,$AA,$FC,$BF,$E3,$3F,$DE,$3D,$37,$E4,$6F
	.dc.b	$C9,$BF,$9D,$7E,$5B,$BF,$39,$7B,$AD,$7C,$DF,$F6,$9F,$7F,$FD,$83
	.dc.b	$90,$EB,$CF,$FE,$5E,$E3,$FF,$63,$E3,$CB,$EB,$71,$5B,$8C,$6B,$E2
	.dc.b	$D5,$FA,$BF,$FA,$53,$0E,$3A,$F8,$67,$FA,$A6,$1E,$58,$C6,$8F,$D4
	.dc.b	$F1,$9F,$F0,$AD,$BC,$FF,$29,$FE,$95,$F0,$7F,$D6,$7F,$96,$BE,$72
	.dc.b	$7E,$DC,$FC,$D5,$FA,$37,$EF,$B9,$7F,$29,$BF,$FB,$03,$0E,$18,$30
	.dc.b	$E1,$7F,$D9,$EF,$77,$FE,$B6,$BE,$CF,$FA,$73,$1F,$0F,$E4,$FF,$E8
	.dc.b	$27,$C7,$8F,$8B,$2F,$E4,$F1,$88,$ED,$1F,$CB,$F0,$EB,$F8,$8B,$1F
	.dc.b	$1E,$7E,$2F,$6B,$DD,$9F,$9F,$1E,$9A,$3F,$4E,$CC,$BF,$2A,$7E,$8C
	.dc.b	$F9,$C7,$E5,$EF,$AF,$7E,$8C,$FC,$67,$FC,$1B,$C3,$EE,$C7,$8E,$1C
	.dc.b	$2F,$75,$FC,$8A,$9F,$92,$FE,$17,$37,$D8,$7F,$B6,$BF,$51,$FC,$87
	.dc.b	$AE,$FD,$05,$FA,$77,$F0,$EC,$78,$E1,$F8,$DC,$38,$78,$3F,$EB,$4B
	.dc.b	$DF,$C3,$FF,$B4,$73,$ED,$5F,$EE,$6E,$FF,$F5,$7C,$63,$A9,$AE,$32
	.dc.b	$E9,$18,$CB,$A6,$DF,$EE,$4B,$E1,$7E,$6C,$FF,$AD,$C6,$2B,$C5,$AF
	.dc.b	$FB,$23,$5F,$3E,$5E,$EC,$BA,$6D,$ED,$A7,$ED,$BE,$7A,$F0,$E7,$D6
	.dc.b	$5F,$9F,$E6,$79,$4E,$7F,$47,$FF,$77,$47,$CB,$FB,$EF,$F1,$FE,$BF
	.dc.b	$DF,$FF,$2B,$7E,$67,$F6,$3E,$FF,$77,$FB,$5B,$DF,$D0,$2F,$7D,$5F
	.dc.b	$E7,$4E,$1F,$AD,$6B,$FF,$25,$F8,$9F,$ED,$88,$F0,$7D,$E7,$B9,$FF
	.dc.b	$5B,$98,$D3,$C5,$A7,$FB,$CF,$0E,$78,$78,$32,$F2,$FE,$01,$FC,$1E
	.dc.b	$73,$8F,$E1,$B4,$F2,$61,$F9,$09,$F4,$E5,$E3,$AD,$61,$E8,$FD,$B7
	.dc.b	$3F,$3D,$FE,$F5,$FC,$33,$FD,$E3,$F9,$6F,$F8,$73,$E2,$3F,$CC,$D9
	.dc.b	$7B,$2B,$5F,$8D,$2B,$5F,$5D,$CB,$DC,$7E,$DF,$EF,$DB,$F8,$E6,$1D
	.dc.b	$5F,$8A,$6F,$3E,$FF,$5D,$FE,$40,$8C,$65,$D7,$EA,$FE,$2C,$72,$C5
	.dc.b	$7C,$35,$7E,$7D,$C3,$5F,$47,$B7,$F0,$21,$F8,$8F,$F8,$97,$E5,$5F
	.dc.b	$B9,$3F,$5D,$7E,$DA,$FD,$CB,$85,$EE,$B5,$2F,$DC,$B3,$C7,$ED,$E0
	.dc.b	$F0,$DB,$EF,$B1,$A7,$82,$BF,$DB,$78,$F7,$E7,$FB,$18,$7E,$EE,$B5
	.dc.b	$AF,$93,$5F,$E2,$D3,$9C,$BC,$51,$E6,$F1,$FC,$35,$BD,$5F,$B4,$63
	.dc.b	$3F,$3F,$91,$70,$F6,$F9,$0E,$CF,$F1,$66,$9F,$E2,$6F,$D0,$7F,$CD
	.dc.b	$5F,$17,$FA,$8F,$93,$FF,$1F,$FC,$A6,$FE,$AF,$FC,$73,$5A,$C5,$5F
	.dc.b	$E4,$4F,$86,$D6,$3C,$FF,$27,$FB,$AE,$0F,$F0,$FD,$FF,$9D,$DF,$C0
	.dc.b	$E7,$E2,$73,$F7,$AD,$3C,$5E,$4F,$F6,$3A,$D7,$D2,$30,$F1,$67,$FC
	.dc.b	$B3,$1E,$D3,$9C,$79,$69,$E6,$FB,$0F,$F2,$66,$3C,$B1,$E3,$1F,$CE
	.dc.b	$E3,$C5,$F2,$5E,$1F,$BC,$FF,$2A,$FD,$C3,$F3,$B7,$D1,$3F,$0C,$F6
	.dc.b	$5E,$FE,$C1,$7B,$AF,$E4,$B8,$F4,$E2,$7E,$C1,$F1,$23,$19,$DE,$6F
	.dc.b	$E7,$99,$73,$C7,$C1,$F4,$9F,$06,$73,$8A,$FC,$BA,$FF,$8B,$31,$E7
	.dc.b	$A7,$8A,$3C,$FF,$A7,$79,$AF,$84,$E7,$5F,$DC,$5F,$D1,$FC,$73,$F5
	.dc.b	$3F,$A2,$FE,$8F,$E3,$7F,$DA,$BF,$98,$3F,$74,$FB,$55,$AE,$FF,$EF
	.dc.b	$BF,$78,$AF,$F5,$46,$3C,$FC,$87,$EB,$DE,$DE,$DF,$F0,$E6,$3D,$9F
	.dc.b	$C0,$F1,$FF,$5B,$8E,$9A,$7F,$B9,$3E,$C7,$E0,$57,$0C,$1B,$79,$74
	.dc.b	$79,$1A,$A6,$3C,$A7,$C9,$F1,$DE,$8C,$38,$EF,$E8,$DF,$FC,$EF,$BF
	.dc.b	$9F,$3E,$91,$F5,$A3,$13,$BE,$6F,$CD,$7C,$4B,$F6,$F7,$EB,$6F,$CD
	.dc.b	$1F,$A2,$3E,$F1,$F8,$77,$F9,$6B,$3F,$CF,$63,$C7,$0E,$17,$BA,$FF
	.dc.b	$5C,$D6,$B1,$EB,$C5,$F1,$8F,$F2,$26,$1E,$1F,$C6,$3E,$64,$7F,$7F
	.dc.b	$9F,$1E,$7E,$2C,$FF,$C0,$B3,$E8,$9C,$E5,$E5,$FB,$9F,$91,$8F,$8F
	.dc.b	$0E,$19,$BF,$0F,$E1,$AF,$9B,$3F,$5F,$D9,$FF,$09,$FC,$AB,$FC,$25
	.dc.b	$F9,$EF,$E8,$3F,$5E,$FF,$11,$7E,$2E,$FC,$C1,$FB,$0F,$EE,$F9,$18
	.dc.b	$70,$F2,$58,$70,$BD,$FE,$AB,$D7,$97,$FC,$C5,$9F,$EB,$3F,$69,$FD
	.dc.b	$46,$3C,$70,$B5,$AC,$79,$FD,$0B,$5F,$17,$C3,$FF,$09,$C3,$84,$E7
	.dc.b	$47,$AF,$9C,$79,$65,$E3,$DF,$FD,$01,$BF,$A2,$73,$87,$19,$7E,$E3
	.dc.b	$E7,$9F,$E5,$9F,$A3,$FF,$32,$F1,$5F,$63,$FE,$3F,$FD,$85,$FA,$B6
	.dc.b	$F3,$5E,$EB,$5F,$84,$66,$3F,$9D,$E3,$FC,$9B,$83,$FC,$81,$7C,$3E
	.dc.b	$17,$E0,$CF,$83,$E1,$3F,$47,$8F,$F9,$37,$F2,$2F,$64,$E7,$5F,$1F
	.dc.b	$D2,$7F,$A1,$23,$B5,$F0,$86,$DF,$E7,$BC,$BE,$05,$71,$C7,$D5,$F8
	.dc.b	$47,$0C,$79,$DF,$FB,$EE,$F6,$7B,$31,$E9,$1F,$E8,$38,$E7,$F7,$0F
	.dc.b	$D1,$3E,$2F,$DB,$FF,$B8,$FC,$5F,$F3,$47,$E0,$3F,$2E,$B5,$A7,$FA
	.dc.b	$4F,$F0,$AF,$F2,$E7,$A2,$FD,$B7,$59,$7F,$AA,$38,$3F,$C5,$FA,$76
	.dc.b	$8F,$DA,$47,$C9,$3F,$D4,$CF,$82,$FF,$CC,$DF,$4B,$F1,$B2,$EF,$AF
	.dc.b	$8F,$E0,$B1,$7B,$DB,$FC,$F9,$FB,$9F,$83,$6F,$24,$7F,$9F,$FC,$7F
	.dc.b	$C1,$7E,$C1,$FE,$9B,$F9,$F7,$E8,$4F,$BC,$FF,$6E,$76,$BF,$5C,$DF
	.dc.b	$DD,$7F,$89,$AB,$5F,$27,$F8,$4F,$D3,$33,$8F,$DE,$CB,$FD,$4D,$FB
	.dc.b	$A5,$C3,$86,$7F,$EE,$7F,$C2,$3E,$FD,$78,$72,$E7,$F9,$A6,$2F,$79
	.dc.b	$FE,$43,$E1,$3A,$D7,$83,$67,$E7,$9F,$E0,$5F,$73,$FD,$1D,$F6,$4F
	.dc.b	$D2,$3F,$DB,$A3,$F6,$BF,$F2,$E7,$9D,$FE,$B9,$F0,$3F,$92,$78,$1F
	.dc.b	$ED,$9B,$E1,$F8,$2F,$1B,$DF,$E0,$FD,$6B,$58,$FF,$BA,$1F,$F0,$1E
	.dc.b	$5E,$08,$7E,$8D,$FD,$9E,$31,$87,$3F,$B4,$FF,$23,$67,$E0,$C1,$9F
	.dc.b	$83,$1F,$35,$7E,$FA,$7C,$06,$A3,$D7,$9F,$AB,$F4,$1C,$DF,$18,$F6
	.dc.b	$FE,$53,$8A,$DE,$1F,$C7,$3E,$03,$19,$F8,$F7,$7C,$37,$6F,$A1,$FE
	.dc.b	$89,$F5,$7F,$EC,$7F,$A9,$7E,$C2,$FD,$35,$FA,$F6,$FE,$29,$7B,$AD
	.dc.b	$7E,$67,$9F,$92,$F8,$5C,$A1,$9F,$67,$E4,$99,$7C,$3F,$84,$7F,$94
	.dc.b	$EF,$FA,$A4,$F7,$C3,$B7,$E7,$DF,$E2,$DE,$9F,$F6,$66,$1C,$36,$F3
	.dc.b	$EC,$FD,$B3,$86,$7C,$E1,$F9,$67,$F9,$2F,$F5,$8F,$F6,$A7,$CE,$BF
	.dc.b	$2A,$7D,$5B,$F2,$2F,$F3,$7F,$D6,$3F,$25,$7E,$01,$C7,$8F,$D9,$8F
	.dc.b	$1C,$38,$7D,$93,$0E,$1F,$A9,$EE,$FF,$C2,$99,$F6,$7F,$C9,$7A,$78
	.dc.b	$A1,$F8,$6F,$F8,$BE,$31,$8F,$4C,$1C,$BF,$E4,$D9,$CE,$DE,$6F,$3F
	.dc.b	$19,$79,$32,$F5,$CF,$AB,$7F,$F2,$27,$8E,$C3,$8E,$DE,$79,$7E,$55
	.dc.b	$E7,$FD,$4B,$FB,$1F,$CF,$3F,$12,$FE,$5F,$F5,$8F,$A8,$7E,$D9,$BF
	.dc.b	$19,$7B,$F8,$BF,$F3,$25,$3E,$23,$5F,$AF,$FB,$78,$FF,$09,$A7,$67
	.dc.b	$E1,$98,$C7,$9E,$9E,$0C,$3D,$B2,$AF,$2C,$F9,$3F,$74,$CC,$63,$2F
	.dc.b	$0F,$E3,$5E,$5D,$3C,$9B,$7A,$7F,$6E,$C6,$3D,$7F,$46,$B8,$7A,$3F
	.dc.b	$3D,$FF,$48,$7E,$BF,$F9,$9D,$5F,$A9,$3F,$18,$7E,$C7,$FE,$B4,$FC
	.dc.b	$45,$F9,$57,$FA,$BF,$E4,$38,$E1,$FA,$1C,$38,$5E,$EB,$FD,$69,$F2
	.dc.b	$3F,$E9,$DF,$7F,$FD,$F9,$FA,$03,$F4,$EF,$E3,$BF,$A5,$FF,$95,$FF
	.dc.b	$00,$7C,$76,$F4,$FF,$7E,$5E,$FE,$19,$7B,$F8,$FD,$57,$FB,$EE,$9F
	.dc.b	$A6,$FF,$9E,$F3,$EF,$9D,$AE,$6F,$83,$ED,$DF,$85,$FF,$A6,$EB,$79
	.dc.b	$F6,$9F,$BF,$7D,$BE,$11,$97,$8F,$4F,$F3,$77,$D4,$7C,$7F,$2F,$F9
	.dc.b	$2D,$F6,$C7,$94,$BF,$27,$FF,$5F,$7D,$77,$CF,$FD,$65,$F8,$E3,$EC
	.dc.b	$DF,$8B,$7F,$B6,$FF,$4A,$78,$EF,$D2,$38,$EB,$FC,$3F,$1E,$38,$70
	.dc.b	$FC,$4D,$87,$0F,$BA,$FF,$66,$6B,$F7,$7E,$FD,$F3,$EF,$A6,$7E,$B9
	.dc.b	$FD,$E7,$C7,$FF,$27,$7E,$6B,$C3,$6F,$E3,$98,$70,$BD,$FD,$52,$FF
	.dc.b	$CA,$9F,$8A,$6E,$FF,$DF,$19,$75,$CB,$A7,$D1,$FF,$B1,$FF,$33,$FC
	.dc.b	$77,$DA,$7C,$AF,$B5,$7E,$A8,$FA,$A7,$E8,$8F,$C0,$1F,$AE,$FF,$B9
	.dc.b	$71,$E3,$E1,$FF,$96,$30,$E1,$7B,$F0,$F5,$BF,$F3,$3D,$3B,$56,$BC
	.dc.b	$DF,$14,$E7,$0E,$5A,$78,$3E,$CB,$FD,$B9,$E3,$7E,$8F,$E5,$7F,$C8
	.dc.b	$FF,$BD,$BD,$9F,$8F,$F5,$CF,$DB,$DF,$A9,$FF,$BE,$70,$E1,$F6,$6C
	.dc.b	$38,$74,$7F,$83,$EF,$F9,$87,$9A,$BD,$D6,$B1,$FD,$6E,$FC,$CF,$E9
	.dc.b	$9A,$78,$B4,$F0,$C7,$FB,$C7,$F4,$6F,$DC,$7E,$48,$FC,$6D,$FA,$5F
	.dc.b	$F9,$FB,$F5,$6F,$C8,$6F,$FC,$F3,$66,$1C,$2F,$7F,$14,$BD,$FC,$CA
	.dc.b	$F7,$9F,$F5,$DC,$FB,$67,$D7,$C8,$7C,$57,$8E,$B5,$9F,$4C,$BF,$41
	.dc.b	$F1,$7F,$D9,$72,$7E,$ED,$FC,$3F,$F7,$77,$EA,$CF,$81,$7E,$B0,$F0
	.dc.b	$7F,$B8,$FD,$67,$F9,$D3,$1E,$3A,$FF,$6F,$C7,$8E,$1C,$2F,$7F,$AF
	.dc.b	$FA,$73,$ED,$9F,$5C,$FA,$7D,$43,$8F,$C9,$5A,$C7,$94,$FF,$15,$CF
	.dc.b	$E1,$7D,$07,$F2,$59,$7E,$01,$FC,$8B,$F0,$47,$E9,$8F,$DE,$1F,$A6
	.dc.b	$3F,$79,$E3,$97,$F8,$17,$1E,$3F,$1A,$C7,$8E,$1C,$3F,$47,$DD,$FC
	.dc.b	$3C,$FB,$67,$D7,$F0,$9F,$95,$F2,$5F,$CD,$FE,$F3,$ED,$FE,$A2,$FC
	.dc.b	$95,$FA,$93,$F7,$CF,$EF,$9F,$4D,$F6,$F0,$E1,$E3,$B0,$E1,$7B,$FA
	.dc.b	$6F,$F7,$3F,$7B,$F2,$EF,$E8,$9E,$4B,$F7,$57,$C2,$BF,$63,$FF,$03
	.dc.b	$FE,$64,$FD,$ED,$FA,$4B,$F6,$3E,$5D,$1F,$E2,$5C,$B9,$62,$FC,$D1
	.dc.b	$FB,$E7,$F9,$7B,$8B,$FD,$3F,$87,$FA,$4F,$0F,$F1,$67,$3F,$BB,$9F
	.dc.b	$F3,$DF,$76,$FD,$EB,$F2,$D7,$BB,$5E,$FD,$05,$EF,$F6,$8F,$27,$8D
	.dc.b	$F6,$FB,$FE,$D7,$ED,$4F,$C9,$9F,$A4,$FF,$82,$74,$FF,$0B,$E9,$F6
	.dc.b	$6F,$75,$AF,$90,$F7,$79,$9F,$89,$F3,$BF,$8D,$FD,$A7,$F1,$AE,$5C
	.dc.b	$E3,$F9,$7F,$43,$F2,$5F,$11,$FE,$0C,$F8,$A7,$BF,$AA,$7A,$DF,$9C
	.dc.b	$CB,$9F,$B1,$FE,$4D,$E0,$FD,$0D,$6B,$2E,$7A,$3C,$BD,$75,$FE,$F7
	.dc.b	$DC,$FE,$05,$CD,$F9,$5F,$DD,$3E,$A4,$E7,$C0,$FE,$01,$3F,$99,$FC
	.dc.b	$2F,$39,$FE,$EB,$0E,$3F,$AE,$F4,$8C,$5F,$0F,$C7,$3C,$9F,$0D,$FC
	.dc.b	$6B,$E1,$3F,$A5,$FA,$8F,$8D,$FD,$D7,$17,$F1,$FC,$7F,$8D,$FC,$4F
	.dc.b	$F2,$5B,$DD,$7F,$42,$AD,$57,$C4,$AD,$7C,$33,$0E,$93,$F8,$DC,$7C
	.dc.b	$13,$9F,$8B,$F8,$DF,$1F,$FD,$1B,$E4,$BF,$A6,$FC,$77,$F4,$CF,$C8
	.dc.b	$BF,$CA,$3A,$2B,$5D,$FF,$D2,$A7,$F7,$B3,$9C,$FF,$A3,$61,$CB,$BF
	.dc.b	$FD,$ED,$39,$9F,$E2,$DF,$21,$FD,$2A,$7F,$23,$1F,$A1,$F8,$BF,$EE
	.dc.b	$1F,$2B,$FD,$8F,$E5,$FF,$B4,$72,$FF,$A3,$31,$FF,$02,$5F,$F6,$EA
	.dc.b	$D7,$8C,$AF,$F0,$DD,$6A,$BF,$B6,$FC,$C7,$F7,$2A,$74,$FE,$17,$E8
	.dc.b	$FF,$C1,$9F,$47,$FE,$0B,$FA,$3F,$F0,$57,$D1,$FF,$7F,$C9,$5A,$FA
	.dc.b	$1F,$F0,$47,$D0,$7F,$83,$BF,$13,$FD,$C7,$E1,$FF,$E4,$3F,$C5,$9F
	.dc.b	$64,$FD,$E3,$FC,$D3,$E8,$FF,$75,$FC,$ED,$F8,$EF,$0E,$4F,$F5,$16
	.dc.b	$1C,$2F,$7F,$46,$B7,$E9,$FB,$B5,$ED,$FF,$29,$1F,$9F,$CF,$C5,$9B
	.dc.b	$F6,$CC,$CE,$4E,$F1,$E5,$FC,$0B,$F8,$66,$4F,$C8,$F5,$5D,$AB,$5B
	.dc.b	$7F,$83,$7F,$6B,$CE,$7E,$7C,$78,$FD,$5B,$CB,$F3,$4E,$6F,$FC,$29
	.dc.b	$2F,$C5,$BF,$8F,$76,$7F,$9A,$7E,$EF,$FC,$C3,$F2,$BF,$23,$E6,$9F
	.dc.b	$5E,$FC,$A7,$7C,$1F,$E6,$7F,$BD,$FF,$36,$DE,$EB,$F9,$D7,$77,$F8
	.dc.b	$B6,$BF,$D9,$91,$FE,$A8,$FD,$6D,$F9,$67,$F9,$1F,$E1,$FF,$9F,$7E
	.dc.b	$03,$D2,$FB,$FB,$8B,$DF,$85,$FC,$73,$E7,$5F,$90,$EF,$3F,$F5,$9F
	.dc.b	$EB,$1F,$33,$1C,$63,$C5,$F9,$FA,$F8,$DE,$B9,$74,$D1,$7F,$C5,$78
	.dc.b	$FF,$D1,$58,$F9,$F6,$F3,$7D,$2E,$71,$F5,$E7,$E5,$9F,$56,$AF,$CB
	.dc.b	$BC,$3A,$75,$87,$E5,$F8,$AD,$FC,$75,$E1,$F6,$7A,$BF,$D3,$FF,$0D
	.dc.b	$FF,$4D,$7E,$D2,$FD,$85,$DA,$FD,$E9,$FA,$1F,$FA,$F7,$F4,$EF,$E5
	.dc.b	$98,$70,$BD,$D6,$BE,$3D,$FA,$5B,$F3,$CE,$DC,$BF,$EF,$FD,$3B,$69
	.dc.b	$F2,$7A,$FF,$DA,$13,$9C,$9F,$7F,$9D,$3B,$5F,$8F,$6F,$F6,$77,$CE
	.dc.b	$78,$B0,$E1,$39,$D7,$FC,$2D,$1E,$01,$9F,$83,$6F,$36,$EF,$CF,$B5
	.dc.b	$39,$CB,$DD,$FA,$17,$F7,$3F,$C8,$FF,$C8,$FF,$74,$FD,$F5,$FA,$3B
	.dc.b	$27,$E0,$BF,$EA,$AF,$DB,$78,$6D,$FE,$41,$C3,$85,$EE,$BF,$D5,$FE
	.dc.b	$DE,$AB,$FC,$59,$4F,$CD,$3E,$26,$7E,$0C,$ED,$73,$7C,$39,$F8,$3F
	.dc.b	$0F,$FE,$BD,$5B,$CF,$B7,$DE,$63,$0E,$11,$E1,$FC,$41,$F8,$07,$86
	.dc.b	$BF,$CA,$79,$74,$DB,$FD,$87,$F3,$7E,$5F,$B4,$7E,$B6,$FD,$53,$FC
	.dc.b	$65,$C9,$F0,$7F,$13,$FF,$51,$F8,$0E,$0F,$F7,$26,$1C,$2F,$7C,$4B
	.dc.b	$7E,$61,$D7,$2E,$9C,$5F,$A8,$C7,$8C,$7F,$87,$E3,$07,$F9,$8D,$F1
	.dc.b	$7D,$FB,$4F,$A5,$F9,$0F,$BE,$BF,$CE,$55,$A9,$FA,$DF,$91,$67,$5F
	.dc.b	$26,$1C,$74,$F2,$7D,$AF,$8F,$6F,$3E,$CE,$BC,$5E,$E3,$FC,$C5,$C1
	.dc.b	$EF,$D7,$C9,$7C,$3F,$1E,$FB,$9F,$A4,$3F,$7C,$7E,$D8,$FC,$CB,$FD
	.dc.b	$AB,$E8,$3F,$6A,$7E,$C1,$C2,$F7,$5A,$97,$ED,$B9,$C3,$F0,$7A,$32
	.dc.b	$ED,$F8,$36,$1F,$04,$F7,$F9,$7F,$E4,$3F,$92,$FC,$59,$CE,$5E,$2D
	.dc.b	$F1,$A7,$82,$FF,$93,$E8,$FC,$63,$C1,$7E,$3F,$03,$FC,$AD,$F4,$6D
	.dc.b	$A7,$AB,$FE,$0F,$FC,$E7,$FA,$C7,$93,$FC,$7A,$9F,$93,$70,$BD,$FD
	.dc.b	$6A,$F7,$F0,$8B,$FE,$33,$D3,$6F,$6F,$F0,$78,$F0,$47,$F0,$EB,$E4
	.dc.b	$FF,$30,$8C,$E5,$E3,$F9,$9F,$D6,$7E,$BB,$C3,$0E,$97,$C7,$2F,$0F
	.dc.b	$CF,$7F,$91,$F1,$E3,$97,$97,$1F,$DA,$67,$E5,$F6,$FD,$58,$F1,$9C
	.dc.b	$DF,$E9,$B2,$EB,$F7,$4F,$C5,$BF,$82,$FE,$9D,$F8,$B7,$FD,$6D,$D7
	.dc.b	$CA,$75,$E7,$2B,$5F,$8A,$BF,$6C,$D6,$1F,$C3,$B4,$F7,$1F,$ED,$E5
	.dc.b	$F8,$FF,$C5,$8C,$61,$CF,$C7,$F1,$7C,$EF,$B6,$73,$97,$8F,$ED,$71
	.dc.b	$1F,$E6,$CC,$7C,$F7,$E9,$CD,$FB,$87,$8B,$2E,$95,$FE,$CF,$FC,$1B
	.dc.b	$19,$7A,$3F,$3A,$B8,$F1,$FC,$EB,$E3,$FE,$A4,$FD,$5D,$F9,$97,$FB
	.dc.b	$2F,$F2,$EF,$F3,$97,$88,$FB,$3D,$D6,$BD,$D5,$6B,$F4,$2C,$FC,$EF
	.dc.b	$BF,$91,$9F,$5C,$F0,$DC,$38,$FE,$09,$F9,$BA,$EB,$5E,$18,$FF,$24
	.dc.b	$46,$27,$7A,$7F,$88,$FB,$3F,$0F,$87,$3C,$FC,$FF,$88,$7E,$0B,$F2
	.dc.b	$BF,$ED,$F1,$FE,$DE,$FC,$57,$E9,$FD,$BF,$F8,$2F,$0F,$C7,$D6,$BE
	.dc.b	$5F,$FD,$85,$DC,$AF,$F0,$1D,$6B,$E6,$1F,$87,$7B,$70,$EB,$BF,$E0
	.dc.b	$03,$CF,$2F,$C0,$FD,$A7,$DA,$8C,$65,$FE,$6C,$C7,$A6,$9E,$58,$F0
	.dc.b	$7E,$61,$E9,$CF,$E6,$EB,$E4,$D7,$FD,$E7,$87,$F7,$9D,$3F,$7B,$F8
	.dc.b	$B3,$F0,$87,$E9,$8F,$CE,$3F,$C6,$DF,$4E,$FC,$F7,$FD,$B5,$F3,$FF
	.dc.b	$B2,$C3,$85,$EF,$EB,$3F,$D8,$FF,$A1,$EB,$3E,$BF,$23,$FE,$03,$C3
	.dc.b	$86,$96,$7C,$2F,$83,$4F,$F2,$27,$CF,$FF,$AF,$6B,$79,$75,$E6,$C6
	.dc.b	$1C,$23,$C1,$B7,$FB,$CB,$7F,$64,$FF,$85,$B6,$F2,$E5,$EA,$F1,$3C
	.dc.b	$93,$9D,$3C,$1B,$FF,$B7,$BB,$73,$EC,$B1,$E3,$3F,$EE,$2F,$88,$E3
	.dc.b	$FB,$93,$F1,$47,$EB,$DF,$E2,$BF,$8B,$FF,$45,$F9,$DF,$C7,$F8,$3C
	.dc.b	$77,$BA,$D7,$B9,$FC,$27,$97,$9E,$3C,$B8,$3F,$C9,$B1,$FC,$8A,$FB
	.dc.b	$7E,$29,$F1,$E5,$1C,$FE,$87,$FC,$C3,$87,$FA,$82,$79,$E3,$DB,$F4
	.dc.b	$CF,$8F,$1C,$32,$F3,$DF,$0F,$D1,$FF,$D3,$5A,$79,$A7,$3B,$7F,$A9
	.dc.b	$27,$F7,$93,$F3,$3C,$D7,$EB,$FF,$E6,$8F,$2D,$FB,$0B,$F5,$AF,$EB
	.dc.b	$1F,$8B,$FC,$EB,$DF,$DA,$FF,$4F,$AD,$7C,$83,$F2,$7D,$65,$FE,$72
	.dc.b	$E0,$F8,$27,$F5,$1C,$3B,$FD,$BF,$C1,$79,$E9,$E0,$D3,$FD,$15,$5C
	.dc.b	$B3,$ED,$1F,$EB,$C8,$C4,$EF,$27,$83,$8D,$3C,$7F,$8F,$7F,$09,$BF
	.dc.b	$EC,$1F,$4E,$F3,$5F,$90,$BE,$E5,$F9,$13,$F6,$E7,$E5,$8F,$D4,$39
	.dc.b	$64,$CB,$96,$3C,$70,$E1,$6F,$D1,$77,$5D,$F5,$FF,$48,$70,$7F,$B0
	.dc.b	$23,$8E,$3F,$EB,$E8,$C6,$BE,$4D,$7C,$7A,$BE,$BB,$FB,$C5,$6B,$5F
	.dc.b	$04,$FF,$9A,$23,$15,$C3,$47,$E7,$9E,$2C,$7A,$6D,$E6,$FA,$2C,$65
	.dc.b	$CA,$DD,$78,$C7,$97,$CA,$FC,$0F,$45,$FA,$E3,$F5,$F7,$EA,$CF,$DD
	.dc.b	$5F,$1E,$FD,$FB,$E7,$71,$7F,$B7,$B1,$E3,$87,$0B,$DD,$3E,$87,$B6
	.dc.b	$5D,$79,$BE,$21,$FE,$08,$C3,$C3,$F8,$2F,$CA,$8C,$4F,$8F,$4F,$16
	.dc.b	$8F,$2B,$E3,$D6,$B4,$F0,$4F,$EA,$F5,$F2,$61,$C7,$5F,$1F,$E3,$7E
	.dc.b	$2B,$DE,$CF,$D1,$7F,$03,$E2,$3F,$6B,$7E,$C8,$FD,$F9,$FA,$87,$F6
	.dc.b	$0F,$A0,$F0,$F0,$7E,$7E,$F7,$5A,$F8,$9F,$F5,$C7,$EC,$7D,$27,$9D
	.dc.b	$7C,$03,$87,$E8,$BE,$F7,$96,$5D,$F2,$FE,$DB,$F8,$AF,$69,$69,$E5
	.dc.b	$AD,$65,$FE,$C5,$FA,$1E,$D1,$8D,$7F,$C2,$DC,$8D,$7F,$C6,$3F,$B3
	.dc.b	$7F,$BF,$FF,$37,$FF,$76,$73,$3D,$17,$E9,$FF,$E4,$DB,$DF,$E2,$8B
	.dc.b	$DD,$3F,$32,$D7,$07,$F9,$3B,$25,$FE,$8B,$0E,$FF,$A1,$5A,$D4,$F3
	.dc.b	$FC,$43,$F1,$FF,$AC,$7E,$82,$FB,$E7,$E4,$F0,$B1,$87,$08,$F3,$FE
	.dc.b	$CD,$88,$FF,$6D,$E7,$EB,$C7,$F3,$33,$EF,$E0,$F8,$15,$C6,$73,$F5
	.dc.b	$3E,$88,$E3,$A7,$8F,$85,$83,$F1,$CD,$7E,$D3,$FE,$00,$FC,$6F,$F9
	.dc.b	$CF,$E6,$CF,$8F,$7E,$7B,$FE,$28,$DB,$BE,$8C,$38,$7C,$4B,$0E,$16
	.dc.b	$FA,$4E,$BF,$40,$FA,$FF,$8D,$8F,$18,$FF,$57,$3F,$E6,$C3,$7A,$78
	.dc.b	$B4,$7D,$97,$C3,$D3,$BE,$7E,$5F,$C6,$B1,$8B,$0E,$1A,$7F,$AA,$76
	.dc.b	$F3,$E1,$C3,$2F,$5F,$E1,$DF,$E8,$DF,$D9,$FF,$CF,$7F,$97,$7F,$84
	.dc.b	$BE,$27,$FB,$3F,$E6,$4F,$B5,$FF,$54,$E4,$BD,$F8,$EF,$A4,$FE,$9F
	.dc.b	$F2,$6C,$38,$7E,$05,$FE,$99,$FD,$F3,$FD,$0F,$F5,$5C,$FC,$97,$EC
	.dc.b	$5B,$FD,$EB,$F8,$9F,$E8,$BB,$EB,$FF,$4E,$5E,$EB,$5E,$0B,$EF,$75
	.dc.b	$4A,$F3,$7E,$77,$FC,$D3,$3E,$FC,$7F,$9E,$AF,$E6,$91,$8D,$3C,$3B
	.dc.b	$35,$FF,$70,$F1,$AB,$5A,$F9,$7C,$BF,$E6,$F1,$8B,$E1,$B3,$F2,$1F
	.dc.b	$1E,$DE,$79,$CF,$D7,$7F,$AE,$7F,$27,$FF,$53,$FE,$05,$E9,$D1,$FB
	.dc.b	$4B,$F4,$67,$E6,$DF,$DE,$3E,$AF,$87,$1F,$FA,$F7,$0E,$17,$BA,$FE
	.dc.b	$FB,$FB,$F7,$FB,$17,$F2,$9F,$F5,$9F,$E7,$2F,$CE,$BF,$DB,$7F,$42
	.dc.b	$FC,$E1,$FA,$CB,$81,$F8,$E7,$1D,$BF,$D9,$58,$F1,$FC,$55,$8F,$1C
	.dc.b	$38,$7E,$2B,$BB,$5E,$EB,$51,$FD,$73,$F6,$DF,$F6,$2F,$E3,$3F,$87
	.dc.b	$FA,$FF,$D3,$7E,$4F,$F9,$AF,$D3,$7F,$C6,$7D,$0B,$DF,$96,$BD,$FA
	.dc.b	$CB,$FF,$4A,$D6,$A7,$3C,$1D,$F3,$FF,$71,$67,$DF,$D1,$69,$FA,$8D
	.dc.b	$3F,$B7,$FE,$B1,$FE,$D8,$FC,$69,$F9,$47,$F8,$3B,$85,$F8,$E7,$F6
	.dc.b	$2E,$07,$E4,$3F,$EF,$1C,$38,$5E,$EB,$5E,$9E,$76,$FF,$79,$65,$D3
	.dc.b	$EB,$BE,$2D,$73,$9D,$09,$D6,$7E,$0F,$9C,$7B,$9F,$E0,$4F,$94,$FF
	.dc.b	$76,$FE,$B4,$F8,$77,$ED,$1F,$CD,$7F,$07,$FE,$15,$F8,$3F,$F9,$87
	.dc.b	$93,$FA,$D6,$A1,$87,$1B,$E1,$F0,$9F,$E8,$5F,$99,$5F,$0B,$DE,$5D
	.dc.b	$BE,$37,$FD,$49,$7F,$E2,$59,$DE,$9E,$4A,$D6,$9F,$C6,$FE,$B7,$FC
	.dc.b	$37,$F1,$5E,$1B,$FF,$24,$70,$3F,$49,$FF,$67,$E4,$FD,$1F,$F9,$87
	.dc.b	$CD,$F8,$B0,$E1,$F6,$FF,$E1,$2B,$DD,$7F,$35,$FA,$AD,$65,$ED,$AD
	.dc.b	$7E,$1F,$FC,$AE,$9F,$39,$E2,$AD,$69,$E1,$AD,$7E,$4B,$FE,$30,$F8
	.dc.b	$6F,$37,$DC,$BF,$2D,$7C,$37,$EE,$FE,$01,$EB,$3E,$99,$FB,$97,$C8
	.dc.b	$C7,$8E,$1C,$3E,$0B,$FC,$57,$7B,$D7,$5B,$7D,$3D,$3B,$69,$D6,$BC
	.dc.b	$13,$A3,$FC,$E9,$87,$3D,$3C,$3F,$89,$7F,$91,$BF,$5C,$F9,$9E,$6F
	.dc.b	$EE,$F4,$FF,$60,$FC,$8B,$F1,$8F,$F0,$CF,$EE,$1F,$72,$F7,$EE,$FF
	.dc.b	$15,$FC,$39,$7B,$FA,$6F,$E4,$1E,$2A,$F7,$97,$4C,$9F,$4D,$E1,$E0
	.dc.b	$FC,$66,$BE,$1C,$F9,$D3,$EA,$3F,$CB,$7B,$7E,$07,$F3,$E7,$ED,$2F
	.dc.b	$C2,$BF,$5B,$F6,$AF,$D2,$BF,$45,$3F,$92,$BE,$0E,$FD,$45,$6F,$D4
	.dc.b	$FF,$21,$71,$2F,$78,$F4,$C7,$F5,$9B,$F3,$3F,$45,$9F,$8A,$75,$FA
	.dc.b	$DF,$F9,$C3,$F1,$CF,$CB,$7E,$C1,$FE,$EB,$FA,$A7,$EC,$8F,$D5,$DF
	.dc.b	$88,$BF,$68,$7E,$93,$FA,$8C,$78,$FC,$AB,$1E,$38,$70,$BF,$93,$C2
	.dc.b	$BD,$D6,$B1,$FE,$69,$BF,$33,$F7,$1A,$78,$F4,$F1,$4F,$ED,$A7,$F6
	.dc.b	$FF,$8C,$7E,$B7,$E1,$1F,$86,$FF,$96,$3F,$40,$FF,$19,$F6,$BF,$56
	.dc.b	$7E,$C1,$C3,$1F,$E0,$98,$70,$F0,$D8,$70,$B7,$EA,$3B,$BF,$E4,$39
	.dc.b	$3E,$F3,$FD,$3F,$A7,$86,$3C,$1F,$6B,$FC,$BB,$F4,$DF,$F4,$DC,$E7
	.dc.b	$3E,$B3,$FC,$CB,$C9,$FF,$4F,$FE,$43,$FC,$17,$F4,$AF,$D6,$7E,$6F
	.dc.b	$FC,$DB,$E9,$7F,$1F,$F2,$9F,$E0,$DE,$A5,$EF,$D8,$5E,$FC,$CF,$EB
	.dc.b	$3F,$79,$FC,$6A,$B5,$E1,$7F,$AD,$1E,$15,$78,$FF,$92,$69,$E0,$E9
	.dc.b	$FF,$04,$63,$FD,$8A,$79,$1A,$CB,$BC,$EB,$3F,$F5,$BF,$EE,$BF,$C7
	.dc.b	$7E,$8D,$F9,$FB,$DD,$7E,$FB,$FE,$8A,$FD,$AB,$F5,$FF,$B9,$7F,$2E
	.dc.b	$C3,$85,$EE,$B5,$EB,$FF,$80,$E7,$38,$7F,$B3,$1D,$FC,$BD,$DB,$F9
	.dc.b	$E6,$7E,$0A,$EF,$D2,$D3,$FD,$99,$5D,$CD,$E7,$DE,$75,$F8,$B7,$FB
	.dc.b	$43,$F3,$DF,$E6,$DF,$AB,$FF,$95,$F6,$7B,$2F,$D9,$5F,$6E,$FD,$4B
	.dc.b	$FB,$1F,$23,$1E,$3F,$80,$FC,$0C,$38,$5E,$FA,$3F,$A6,$D6,$BF,$72
	.dc.b	$FF,$73,$BC,$33,$BC,$1F,$D5,$F5,$F1,$63,$CB,$C1,$F3,$4E,$73,$ED
	.dc.b	$B7,$93,$5F,$F7,$4F,$E5,$1F,$D3,$FE,$1B,$FA,$DF,$E9,$DF,$C2,$39
	.dc.b	$FF,$88,$7E,$69,$FE,$59,$B5,$ED,$56,$FD,$95,$3F,$31,$FE,$CB,$7B
	.dc.b	$AF,$F7,$FB,$C7,$E0,$AE,$BF,$DB,$F3,$F0,$FC,$07,$F8,$F3,$27,$E6
	.dc.b	$DE,$6A,$D6,$9E,$0A,$D7,$C0,$7F,$77,$87,$D7,$7F,$69,$FD,$17,$FD
	.dc.b	$15,$F4,$1F,$3F,$F1,$BF,$F0,$9F,$EF,$AF,$D3,$FF,$BC,$DE,$F9,$BF
	.dc.b	$C0,$BD,$7F,$F0,$03,$DA,$5F,$F8,$0B,$3E,$BB,$7F,$81,$74,$EA,$FF
	.dc.b	$A3,$CE,$66,$70,$7E,$C7,$98,$C8,$9F,$2E,$3D,$FE,$23,$D5,$9F,$8F
	.dc.b	$6F,$46,$9F,$E6,$CF,$D8,$7F,$C0,$9F,$A6,$7F,$88,$7F,$19,$7E,$5E
	.dc.b	$FC,$55,$F9,$BB,$F6,$37,$E3,$FF,$E1,$6C,$78,$EA,$C7,$8E,$1C,$2D
	.dc.b	$FA,$D6,$F8,$7F,$C4,$5A,$77,$7D,$DC,$8C,$B9,$F2,$7B,$63,$13,$C3
	.dc.b	$3E,$DB,$79,$76,$7E,$11,$F5,$F1,$F1,$ED,$E2,$9F,$F4,$4F,$ED,$7F
	.dc.b	$E2,$1F,$87,$FF,$04,$F0,$7F,$82,$BF,$12,$7E,$80,$FC,$DB,$F5,$9B
	.dc.b	$B0,$E1,$F3,$FF,$0C,$5F,$0A,$7E,$B7,$FE,$38,$AD,$E3,$E9,$78,$FC
	.dc.b	$7D,$E2,$FF,$21,$69,$E2,$CB,$9F,$E1,$FF,$E3,$5D,$3C,$15,$A9,$F0
	.dc.b	$69,$E6,$FB,$FF,$F2,$57,$C2,$BF,$75,$FF,$1E,$7E,$D7,$EA,$FD,$F9
	.dc.b	$F0,$6F,$DB,$FF,$D1,$9F,$A7,$3B,$58,$F1,$D7,$FC,$A9,$8F,$1C,$38
	.dc.b	$7E,$16,$C3,$FD,$55,$7B,$B5,$F7,$7F,$2D,$19,$8F,$97,$F3,$5F,$95
	.dc.b	$8C,$74,$34,$F2,$E9,$FE,$9F,$F3,$3D,$7B,$7A,$AB,$5A,$F8,$BC,$EF
	.dc.b	$7F,$6F,$F9,$67,$F6,$EF,$F5,$57,$C4,$7F,$B0,$F9,$BF,$B5,$FB,$4F
	.dc.b	$93,$F6,$DE,$FF,$0D,$FF,$16,$FE,$29,$FE,$6C,$F9,$AF,$F5,$86,$9A
	.dc.b	$E5,$FF,$3A,$69,$D1,$F5,$E1,$C7,$F4,$0F,$53,$CF,$5F,$06,$1F,$E0
	.dc.b	$6E,$4F,$CA,$DB,$0F,$2E,$5D,$F4,$FA,$FE,$7F,$F9,$C7,$F1,$D7,$ED
	.dc.b	$CF,$99,$FE,$AB,$F7,$AF,$CC,$7F,$8F,$6B,$F5,$B0,$E1,$7B,$F8,$75
	.dc.b	$EF,$E0,$9F,$97,$6F,$7F,$F4,$76,$9D,$B4,$FA,$59,$72,$FA,$5B,$5A
	.dc.b	$CF,$A6,$5E,$1F,$C7,$3C,$F3,$9C,$3C,$BB,$3F,$49,$FC,$8F,$C2,$FF
	.dc.b	$3D,$FD,$31,$FA,$CB,$F7,$A7,$E2,$2F,$A3,$7C,$AB,$F2,$BF,$C8,$E3
	.dc.b	$C7,$0E,$17,$BA,$FF,$43,$F5,$FF,$44,$DB,$F9,$D6,$7D,$9F,$C7,$E1
	.dc.b	$E1,$FD,$0B,$F1,$CF,$3D,$3C,$3A,$7F,$A9,$70,$E5,$7F,$E8,$AC,$3C
	.dc.b	$F5,$AF,$C1,$FF,$57,$F0,$BF,$DC,$3A,$3F,$C1,$1F,$96,$3F,$58,$7D
	.dc.b	$1B,$F4,$2F,$F0,$66,$9F,$DB,$F4,$FE,$CD,$87,$0B,$DD,$3F,$15,$FD
	.dc.b	$7E,$5D,$7E,$8B,$F4,$8F,$7F,$A2,$B8,$F2,$5A,$78,$74,$FC,$E4,$F2
	.dc.b	$35,$19,$CF,$F8,$1F,$D1,$FF,$86,$BF,$03,$7E,$38,$FD,$2D,$FA,$DF
	.dc.b	$F8,$8F,$F0,$1F,$E0,$37,$BA,$D4,$FF,$89,$FE,$3F,$39,$FF,$85,$B1
	.dc.b	$FD,$D5,$FF,$89,$BF,$2F,$B3,$9C,$38,$ED,$E2,$8C,$67,$E0,$FC,$23
	.dc.b	$F8,$BF,$D7,$7F,$91,$FE,$53,$F8,$17,$ED,$9F,$EE,$3F,$C2,$7F,$41
	.dc.b	$D3,$FE,$4C,$C7,$FC,$7D,$5A,$7F,$D7,$BF,$A0,$FF,$18,$C3,$94,$7F
	.dc.b	$10,$7F,$CA,$5F,$8D,$FF,$0F,$7B,$7C,$0B,$2F,$F7,$87,$E0,$FE,$2B
	.dc.b	$EB,$39,$7F,$DF,$A7,$F2,$4F,$2B,$FC,$D5,$F8,$07,$D2,$F6,$1F,$8A
	.dc.b	$3F,$49,$7E,$75,$BC,$3F,$CF,$3F,$77,$FE,$92,$EE,$FF,$5D,$DE,$FE
	.dc.b	$D7,$55,$FE,$78,$FB,$1F,$E4,$9A,$76,$8F,$E5,$2F,$82,$F8,$FE,$95
	.dc.b	$FC,$FA,$31,$A7,$8B,$D9,$FE,$83,$F1,$7F,$E9,$8F,$85,$7C,$F7,$F8
	.dc.b	$13,$EC,$73,$9F,$FA,$63,$F0,$4F,$F5,$15,$6B,$E9,$9F,$BC,$FF,$5F
	.dc.b	$ED,$FF,$51,$5E,$FF,$04,$5E,$E7,$9C,$BF,$04,$FF,$5A,$4F,$3C,$FF
	.dc.b	$A9,$FC,$8F,$FB,$03,$DE,$5B,$F5,$E7,$07,$F7,$6C,$BF,$D8,$D3,$FE
	.dc.b	$C2,$FC,$2B,$FD,$A5,$7C,$2F,$7D,$FF,$D5,$B1,$E9,$7C,$F1,$EE,$F1
	.dc.b	$FB,$0F,$E5,$0F,$7E,$6F,$EC,$BF,$86,$FF,$BA,$3E,$E1,$FB,$77,$F1
	.dc.b	$4F,$CB,$FD,$5B,$FF,$BB,$70,$FF,$74,$F0,$FF,$BC,$EB,$5F,$94,$7F
	.dc.b	$66,$C7,$94,$7F,$1C,$FA,$6E,$B7,$C3,$F1,$0F,$F3,$AC,$FF,$20,$D3
	.dc.b	$FB,$AF,$E2,$9F,$0A,$9F,$89,$BF,$13,$7D,$D3,$17,$EB,$2F,$D6,$BF
	.dc.b	$D7,$FE,$E7,$F7,$6C,$B9,$63,$C7,$0E,$17,$BF,$C3,$BA,$F1,$2F,$79
	.dc.b	$F6,$CF,$AE,$7F,$D1,$B4,$E8,$7D,$6C,$FA,$4E,$BE,$63,$FC,$1B,$F8
	.dc.b	$DF,$A7,$F1,$97,$E3,$0F,$C6,$DF,$01,$E8,$FB,$CF,$D2,$FC,$47,$F8
	.dc.b	$33,$0E,$17,$BA,$D7,$A3,$FC,$A2,$73,$87,$2E,$36,$3F,$EB,$7D,$B9
	.dc.b	$1F,$8A,$CF,$C3,$5B,$97,$BB,$FE,$22,$C7,$E7,$7E,$42,$FC,$7F,$9F
	.dc.b	$E4,$1F,$D9,$BE,$35,$FA,$0B,$B3,$FC,$43,$87,$0B,$DF,$CD,$FF,$3F
	.dc.b	$AF,$F0,$EF,$57,$F0,$0D,$FF,$37,$9F,$5D,$5A,$75,$FB,$1F,$15,$6B
	.dc.b	$2E,$7A,$BF,$4F,$FF,$1B,$F3,$7E,$0F,$CE,$7C,$83,$F2,$77,$E4,$EF
	.dc.b	$C7,$DF,$12,$F2,$7F,$33,$C5,$FD,$2F,$6F,$D6,$E5,$CB,$1E,$38,$70
	.dc.b	$B7,$E6,$3F,$E3,$ED,$3C,$1A,$77,$D3,$B6,$9F,$E6,$8C,$F4,$74,$9D
	.dc.b	$63,$E2,$F2,$7F,$C9,$FF,$95,$FE,$6F,$F2,$B7,$E5,$AE,$4F,$EF,$FF
	.dc.b	$8C,$3F,$2C,$FF,$01,$7E,$A3,$FE,$4E,$C3,$85,$EF,$E8,$D7,$BF,$9E
	.dc.b	$7E,$8B,$BD,$FF,$84,$E9,$DB,$4E,$BA,$7F,$62,$E7,$69,$D2,$73,$F3
	.dc.b	$7F,$E6,$4F,$BF,$FD,$87,$23,$ED,$5F,$98,$7C,$BF,$63,$FB,$2F,$CC
	.dc.b	$3F,$24,$FF,$31,$E1,$C2,$F7,$D9,$FC,$8B,$E6,$FF,$B5,$7B,$FF,$E6
	.dc.b	$8F,$BF,$FD,$E7,$E6,$6F,$CD,$DF,$76,$FC,$67,$E8,$FE,$68,$F9,$5F
	.dc.b	$F5,$4F,$33,$1E,$38,$70,$BD,$D3,$F3,$1F,$F3,$AF,$DD,$7F,$09,$FD
	.dc.b	$4B,$A7,$EA,$3F,$E7,$9F,$95,$FC,$37,$E9,$CF,$D7,$F7,$E0,$7F,$30
	.dc.b	$BD,$FE,$12,$BD,$FD,$CE,$A9,$5E,$7F,$D8,$7F,$CE,$D9,$F7,$D9,$A6
	.dc.b	$2F,$A4,$63,$2E,$9E,$5F,$FB,$AA,$7A,$63,$E0,$CF,$F3,$5F,$61,$FE
	.dc.b	$96,$CB,$A6,$5E,$9C,$1B,$3E,$F3,$8E,$9D,$71,$F6,$FC,$FE,$30,$E1
	.dc.b	$FA,$83,$F4,$9C,$6F,$E2,$87,$E5,$3F,$E9,$CF,$3F,$F9,$87,$E1,$CF
	.dc.b	$D0,$5F,$A0,$BE,$BD,$FA,$0B,$F4,$07,$4B,$E7,$79,$74,$FF,$5A,$FC
	.dc.b	$65,$FA,$43,$F0,$D7,$EF,$4F,$D7,$9F,$A7,$FF,$AC,$3E,$57,$8E,$3F
	.dc.b	$EE,$8C,$38,$5E,$EB,$FD,$6B,$F0,$3A,$CB,$AE,$5F,$EA,$7E,$3F,$E9
	.dc.b	$11,$E0,$F9,$9C,$65,$E3,$CB,$9E,$9E,$0F,$97,$FF,$61,$E1,$C2,$73
	.dc.b	$9F,$FA,$56,$31,$97,$A2,$FC,$FF,$88,$BF,$37,$F1,$61,$C2,$1F,$97
	.dc.b	$FF,$B1,$BF,$34,$FE,$B5,$E0,$FD,$1F,$C2,$5F,$8A,$BE,$A7,$FD,$23
	.dc.b	$F9,$CF,$FB,$8B,$D2,$61,$C3,$C3,$61,$C2,$F7,$EE,$F5,$E4,$FF,$03
	.dc.b	$E7,$D5,$FF,$65,$E1,$C7,$B3,$FD,$94,$F3,$C3,$97,$C0,$3F,$08,$FF
	.dc.b	$71,$CE,$76,$F2,$4F,$FA,$06,$31,$9F,$9B,$AF,$C1,$97,$ED,$F5,$F2
	.dc.b	$E1,$C3,$77,$9D,$FE,$E6,$AF,$EF,$5D,$EF,$DD,$1F,$8A,$3F,$54,$7E
	.dc.b	$DB,$F8,$AF,$D4,$9F,$8D,$71,$C7,$FC,$47,$8F,$1C,$38,$5E,$FF,$6F
	.dc.b	$D5,$7E,$4B,$27,$E6,$DF,$EE,$98,$F0,$C7,$81,$FF,$77,$98,$BE,$78
	.dc.b	$3F,$3B,$FF,$18,$F5,$7F,$84,$6B,$A5,$6B,$6C,$C7,$1D,$BD,$38,$36
	.dc.b	$FF,$80,$18,$71,$8F,$44,$EA,$7F,$DF,$53,$9C,$BE,$1C,$E7,$E1,$BD
	.dc.b	$77,$C3,$6F,$89,$F1,$1E,$EF,$86,$7E,$86,$FC,$3D,$F8,$03,$F4,$87
	.dc.b	$E1,$CF,$CC,$7F,$B1,$EF,$EB,$C7,$8E,$1C,$2F,$75,$FE,$16,$AD,$63
	.dc.b	$D7,$87,$E3,$9F,$E5,$4C,$3C,$3F,$6D,$F3,$63,$13,$E3,$CF,$C5,$9F
	.dc.b	$F9,$62,$7D,$13,$9C,$BC,$BE,$07,$CA,$9E,$3A,$78,$EF,$D1,$F4,$7C
	.dc.b	$23,$1A,$FF,$65,$D7,$EA,$7A,$0F,$DC,$1F,$A0,$FE,$23,$E0,$FF,$D8
	.dc.b	$3F,$96,$3F,$5C,$FC,$2D,$EE,$B5,$F8,$4A,$9F,$92,$EB,$1E,$7B,$FC
	.dc.b	$01,$87,$7F,$DE,$BE,$04,$62,$7C,$3E,$1F,$F8,$47,$1F,$F4,$0C,$F3
	.dc.b	$CB,$BF,$C9,$7B,$E3,$84,$78,$B3,$7E,$81,$BC,$39,$47,$FA,$A3,$F4
	.dc.b	$17,$E6,$3F,$E6,$3F,$A3,$7E,$90,$FC,$C5,$FB,$1B,$E9,$1F,$52,$FC
	.dc.b	$C5,$FB,$63,$3E,$7C,$1F,$E8,$EC,$F9,$E5,$CB,$1E,$38,$3F,$5E,$E1
	.dc.b	$AF,$87,$57,$EB,$DF,$A5,$CB,$FD,$2F,$9F,$47,$FC,$2B,$3E,$57,$51
	.dc.b	$9C,$FB,$ED,$E7,$DB,$FD,$37,$39,$BF,$64,$E7,$EF,$3A,$E5,$E1,$DB
	.dc.b	$CD,$87,$B7,$E2,$7C,$B9,$FA,$F1,$E3,$BF,$AB,$E2,$73,$3E,$4C,$38
	.dc.b	$F0,$7F,$97,$F5,$F0,$46,$3E,$AB,$78,$FF,$9D,$FF,$35,$7E,$9F,$FE
	.dc.b	$91,$F5,$DF,$A7,$3E,$9F,$7F,$80,$2F,$7F,$83,$AF,$7F,$97,$EA,$7E
	.dc.b	$46,$FC,$3F,$0F,$BB,$FC,$9D,$7C,$1A,$F7,$F6,$BE,$0D,$FF,$B1,$27
	.dc.b	$39,$F5,$F7,$33,$3F,$84,$D7,$C5,$87,$9F,$7F,$D3,$65,$D6,$73,$87
	.dc.b	$96,$5F,$6F,$F1,$BF,$70,$FF,$80,$1F,$BF,$FF,$0A,$EB,$7E,$6A,$FD
	.dc.b	$75,$F9,$6E,$FF,$15,$5E,$EB,$5D,$19,$8F,$6F,$7F,$FA,$EF,$25,$BF
	.dc.b	$30,$CF,$17,$FB,$82,$7C,$3E,$B7,$FA,$5B,$27,$EA,$5E,$E9,$CE,$DE
	.dc.b	$4D,$31,$97,$8B,$0E,$1B,$3F,$37,$F1,$46,$3E,$3B,$F4,$5F,$B1,$7F
	.dc.b	$B3,$4F,$B1,$F8,$6F,$FB,$2B,$F1,$3F,$DF,$FE,$91,$FE,$66,$E9,$56
	.dc.b	$FB,$FF,$D4,$5F,$88,$FF,$B5,$3D,$75,$BF,$6C,$EB,$CB,$FE,$FA,$C3
	.dc.b	$F0,$9F,$B7,$7D,$E7,$96,$96,$75,$97,$4D,$3C,$31,$C8,$F0,$C6,$72
	.dc.b	$FF,$17,$C6,$6F,$D3,$9F,$8F,$F4,$B7,$ED,$79,$C3,$84,$7F,$A1,$B6
	.dc.b	$F5,$E1,$E0,$CF,$E0,$65,$FE,$6F,$8E,$11,$8A,$7E,$EF,$8C,$BB,$7E
	.dc.b	$39,$E5,$FB,$9F,$E9,$DE,$63,$F6,$8F,$87,$F6,$4F,$D5,$FE,$26,$DE
	.dc.b	$0D,$7F,$CE,$1F,$0D,$FC,$66,$F7,$5F,$3F,$8B,$FE,$00,$65,$D2,$FF
	.dc.b	$84,$C6,$2F,$F8,$8C,$60,$1D,$32,$F1,$73,$33,$6B,$FC,$5B,$0E,$55
	.dc.b	$A9,$7E,$21,$9C,$39,$E5,$EC,$D3,$D7,$E6,$7E,$0B,$6F,$54,$E7,$4E
	.dc.b	$D3,$FE,$5B,$9F,$77,$EF,$4F,$CB,$3F,$C7,$FF,$57,$FD,$E3,$F0,$4F
	.dc.b	$F0,$EF,$E4,$7F,$76,$B5,$F3,$FF,$EC,$4F,$92,$57,$F8,$8B,$ED,$B5
	.dc.b	$8F,$FA,$1F,$7F,$F0,$A6,$7D,$A1,$F9,$9F,$E2,$E5,$97,$4D,$3C,$1E
	.dc.b	$AF,$B6,$73,$7E,$4C,$FB,$FC,$4E,$70,$E1,$1E,$2F,$87,$FF,$41,$67
	.dc.b	$DF,$6F,$36,$DF,$E3,$6F,$C2,$7F,$2D,$FA,$23,$F3,$F7,$8C,$FD,$11
	.dc.b	$FB,$57,$F5,$7F,$DC,$5F,$B9,$F0,$BD,$D6,$BB,$94,$F1,$75,$8F,$B7
	.dc.b	$2E,$DF,$5F,$71,$E3,$9F,$F6,$3F,$C6,$FD,$D5,$E3,$BF,$14,$FF,$67
	.dc.b	$FB,$0C,$DF,$0C,$FB,$FE,$BB,$DB,$5F,$16,$BF,$4B,$8B,$F3,$19,$F3
	.dc.b	$9C,$FE,$AD,$BC,$FD,$9E,$1F,$F6,$DF,$B8,$FC,$9F,$5B,$F5,$BF,$E8
	.dc.b	$5E,$A3,$F5,$1F,$F3,$16,$3F,$E5,$EF,$8A,$E1,$7F,$9A,$BD,$D6,$BA
	.dc.b	$FF,$CC,$7B,$7F,$A9,$EB,$56,$FC,$77,$FA,$4E,$7E,$0C,$7B,$D7,$F7
	.dc.b	$08,$C5,$7F,$91,$31,$F0,$F8,$3F,$DE,$27,$3A,$79,$7E,$B5,$AF,$FB
	.dc.b	$D3,$1F,$29,$EB,$C7,$94,$EB,$5F,$F6,$96,$BE,$5A,$D6,$1C,$F0,$FF
	.dc.b	$3D,$F7,$72,$E4,$7E,$DE,$FA,$2F,$F5,$17,$CA,$3F,$5E,$FE,$33,$F3
	.dc.b	$5B,$AD,$7D,$B7,$F8,$72,$7F,$8E,$FC,$DE,$70,$F8,$3A,$FF,$02,$CB
	.dc.b	$AF,$E4,$58,$84,$78,$27,$BF,$8F,$F5,$B8,$3D,$13,$9C,$BB,$7E,$37
	.dc.b	$F1,$70,$E9,$A7,$8B,$6F,$37,$BD,$F9,$DC,$78,$CE,$76,$FE,$63,$E6
	.dc.b	$7C,$8F,$CF,$3F,$A4,$F9,$7F,$EE,$EE,$B7,$C0,$3F,$1C,$7E,$C5,$BF
	.dc.b	$6B,$F8,$0F,$E1,$9F,$87,$AD,$63,$9C,$7D,$8F,$F5,$4E,$3B,$7C,$32
	.dc.b	$EB,$E6,$7A,$78,$97,$E2,$CB,$A7,$1E,$63,$17,$E4,$D5,$F8,$B7,$7B
	.dc.b	$F1,$43,$F8,$03,$15,$BF,$D0,$6E,$FE,$BF,$C3,$3F,$C9,$DC,$DF,$A2
	.dc.b	$F4,$9F,$9F,$BF,$67,$7E,$34,$F9,$2F,$E6,$3F,$86,$FF,$96,$FB,$3F
	.dc.b	$AC,$F2,$7F,$3C,$C7,$8E,$1C,$2F,$7F,$53,$DA,$B5,$97,$57,$FC,$9F
	.dc.b	$9F,$86,$3F,$C3,$31,$83,$77,$CF,$CF,$67,$FE,$5A,$BE,$13,$9F,$8F
	.dc.b	$C5,$F0,$D7,$D1,$87,$0F,$97,$F1,$4F,$F1,$7C,$BD,$D5,$AC,$BF,$C4
	.dc.b	$BB,$3F,$23,$FE,$0B,$F3,$EF,$D6,$1F,$BC,$BF,$79,$7E,$EC,$FD,$E5
	.dc.b	$F8,$D7,$F2,$CC,$B9,$6D,$FD,$83,$2E,$58,$F1,$C3,$85,$BE,$0B,$FD
	.dc.b	$01,$A7,$83,$47,$E2,$5F,$E5,$08,$C6,$17,$2E,$95,$E4,$D7,$C7,$AB
	.dc.b	$AF,$FC,$FF,$87,$09,$CF,$D3,$63,$0E,$38,$F8,$F0,$F4,$FE,$7D,$E2
	.dc.b	$9C,$DF,$4D,$5C,$D9,$9F,$E8,$1F,$99,$7C,$1F,$D2,$BF,$0F,$F8,$47
	.dc.b	$F9,$B7,$F2,$ED,$77,$AB,$5D,$0A,$FF,$39,$7E,$9B,$FE,$9F,$BF,$E6
	.dc.b	$7F,$8D,$17,$D3,$3E,$39,$FB,$8B,$F0,$DF,$77,$E7,$1C,$77,$FF,$06
	.dc.b	$E3,$C7,$0E,$17,$BE,$87,$EE,$3B,$B5,$F7,$FC,$F7,$FC,$FB,$A7,$88
	.dc.b	$F0,$CF,$FA,$91,$C6,$3D,$33,$F9,$D3,$CE,$FC,$1A,$79,$A5,$D3,$FE
	.dc.b	$B0,$C3,$8C,$E7,$3F,$F2,$C4,$63,$4F,$37,$D3,$F3,$8F,$83,$5C,$FE
	.dc.b	$E5,$BE,$75,$F1,$43,$DE,$C5,$78,$7E,$B7,$FD,$7F,$EA,$FF,$5E,$FA
	.dc.b	$C7,$E9,$4F,$D3,$FF,$C0,$3F,$58,$FD,$8D,$FA,$3B,$F7,$BF,$F6,$96
	.dc.b	$3C,$70,$E1,$F3,$9F,$EA,$9B,$DF,$CC,$FF,$B5,$FF,$32,$FB,$3C,$AF
	.dc.b	$CF,$BF,$DB,$7F,$82,$FE,$AF,$EA,$BF,$CB,$FE,$92,$F7,$5A,$F3,$7F
	.dc.b	$CA,$9F,$2F,$FE,$E1,$F8,$0F,$EE,$BF,$B0,$3F,$56,$FF,$3E,$FC,$0B
	.dc.b	$F3,$77,$C2,$3F,$44,$FF,$54,$E1,$C2,$F7,$5A,$9F,$DB,$6B,$FC,$EF
	.dc.b	$C7,$FF,$29,$E5,$D7,$E2,$7F,$C0,$17,$D3,$F3,$5F,$F7,$3E,$5C,$E7
	.dc.b	$39,$FF,$63,$C3,$FD,$ED,$C9,$FE,$0A,$FA,$D7,$E6,$FF,$ED,$2F,$D1
	.dc.b	$7F,$DF,$7F,$01,$FE,$D1,$F9,$AD,$E1,$FE,$C5,$AD,$7C,$D7,$F1,$CF
	.dc.b	$C1,$FF,$BA,$E5,$CE,$73,$87,$E4,$B8,$7F,$C5,$F8,$71,$FC,$11,$9F
	.dc.b	$6F,$CB,$7F,$DF,$BF,$73,$FE,$1C,$FC,$E5,$F8,$AB,$F4,$DF,$F0,$F7
	.dc.b	$D2,$7F,$8E,$3F,$45,$7E,$9E,$D5,$8F,$1C,$38,$7D,$53,$0E,$1F,$B7
	.dc.b	$EE,$B5,$97,$4E,$2F,$34,$62,$F8,$43,$EA,$3F,$D0,$39,$3F,$2C,$FF
	.dc.b	$18,$FE,$DE,$FD,$89,$FA,$27,$F2,$AF,$05,$F9,$AB,$F4,$C7,$E6,$3E
	.dc.b	$7C,$78,$E1,$C3,$E3,$58,$70,$BE,$F5,$AC,$7A,$62,$E7,$CF,$57,$FA
	.dc.b	$53,$4F,$15,$6B,$DD,$F3,$FE,$C7,$FE,$46,$87,$E7,$3F,$C6,$7F,$23
	.dc.b	$7E,$17,$3E,$F9,$F9,$97,$CB,$C3,$85,$EE,$B5,$E4,$7C,$FF,$DE,$7F
	.dc.b	$CA,$31,$F4,$3E,$31,$F3,$BF,$E2,$2F,$C2,$5F,$B6,$BC,$5E,$1C,$5F
	.dc.b	$EA,$AC,$38,$5E,$EB,$F9,$37,$EC,$CF,$CC,$3F,$C9,$DF,$51,$FC,$7B
	.dc.b	$F0,$FF,$F5,$F7,$C0,$BF,$0D,$7E,$9E,$C1,$C1,$E4,$F1,$9C,$3E,$4F
	.dc.b	$BF,$79,$EF,$C3,$1F,$B9,$F1,$C3,$85,$EE,$B5,$FB,$DF,$5E,$9F,$87
	.dc.b	$F3,$8E,$DF,$E6,$3F,$8B,$3F,$74,$7E,$33,$BE,$AF,$F6,$3F,$83,$F4
	.dc.b	$30,$E1,$F3,$AC,$38,$7E,$91,$E9,$F2,$3E,$F7,$97,$FE,$79,$F9,$07
	.dc.b	$E1,$DF,$EE,$4E,$8F,$C8,$69,$F7,$FC,$EF,$C4,$7C,$3F,$F1,$FA,$FE
	.dc.b	$41,$5A,$A7,$97,$F8,$4C,$7A,$57,$FB,$6A,$B5,$39,$BF,$E5,$1F,$75
	.dc.b	$9E,$CF,$D1,$76,$3E,$77,$FD,$DD,$F8,$27,$CF,$9C,$C7,$FB,$C7,$F8
	.dc.b	$03,$1A,$FF,$B9,$BC,$9F,$DC,$FB,$3F,$E7,$BA,$FF,$72,$CE,$8F,$C2
	.dc.b	$7B,$9F,$B6,$CB,$B7,$DA,$79,$7D,$AF,$E1,$5E,$07,$F0,$6F,$84,$7E
	.dc.b	$9A,$FD,$E5,$B7,$E0,$6F,$7F,$06,$BD,$D3,$87,$5A,$FF,$11,$AD,$62
	.dc.b	$FC,$5F,$C1,$39,$C3,$BF,$C6,$7C,$7F,$8C,$FE,$55,$CD,$FC,$E7,$D8
	.dc.b	$70,$FF,$95,$FA,$7F,$9A,$63,$FB,$AF,$8F,$FE,$73,$5A,$FB,$7F,$CB
	.dc.b	$63,$CB,$E3,$FF,$5F,$F8,$67,$5C,$3F,$9D,$E1,$C3,$F1,$4F,$AD,$F2
	.dc.b	$DF,$D3,$B7,$7C,$C7,$F8,$F7,$EB,$FF,$25,$B7,$F5,$3C,$7F,$C1,$3F
	.dc.b	$35,$FD,$66,$F7,$F3,$0B,$7E,$AD,$FA,$7F,$9E,$FE,$C9,$F4,$3F,$DA
	.dc.b	$BE,$79,$FB,$B7,$F9,$17,$45,$6B,$DE,$FE,$D1,$3F,$C5,$A7,$FC,$3D
	.dc.b	$A8,$FA,$9F,$81,$BF,$6E,$7E,$5C,$FD,$A5,$FB,$1F,$F8,$87,$F4,$BF
	.dc.b	$DC,$7D,$7F,$FB,$6B,$C5,$FF,$44,$E1,$C2,$F7,$F6,$4F,$C7,$77,$A7
	.dc.b	$7B,$DB,$FE,$E2,$3F,$39,$87,$3D,$1F,$BE,$E6,$72,$71,$8F,$36,$BE
	.dc.b	$5D,$5F,$88,$EA,$BD,$7A,$F8,$E7,$F9,$3C,$78,$8C,$D7,$1D,$1F,$AF
	.dc.b	$E7,$4F,$5C,$E7,$F3,$9F,$E2,$DD,$3F,$E9,$7F,$BB,$FF,$25,$FC,$03
	.dc.b	$EC,$3F,$8E,$72,$FF,$9C,$34,$FF,$2A,$FE,$01,$F6,$7E,$F3,$FC,$A1
	.dc.b	$4F,$CD,$3F,$90,$7E,$13,$F4,$6F,$FD,$BF,$E5,$3F,$4A,$FB,$7F,$94
	.dc.b	$3F,$27,$FC,$17,$6F,$0B,$FB,$1F,$8B,$B0,$E1,$DE,$C3,$87,$8B,$F4
	.dc.b	$78,$3F,$D4,$F8,$76,$CF,$E6,$3C,$CF,$0F,$E6,$F8,$9D,$E5,$D6,$BC
	.dc.b	$7A,$3F,$48,$FE,$25,$B7,$A6,$B5,$A7,$86,$BF,$D4,$F9,$75,$9C,$E9
	.dc.b	$E4,$D3,$1A,$FA,$23,$CF,$F8,$4F,$8F,$E5,$3F,$A8,$7E,$0E,$FD,$C7
	.dc.b	$F1,$5F,$BB,$BF,$71,$FF,$02,$71,$7B,$2F,$75,$A9,$FD,$EC,$E7,$E1
	.dc.b	$9F,$86,$7B,$78,$BE,$86,$1D,$B8,$3F,$1D,$7B,$CF,$C1,$E4,$B4,$7A
	.dc.b	$9F,$02,$FF,$51,$87,$82,$B5,$EB,$66,$7F,$AD,$E9,$FE,$98,$DB,$C9
	.dc.b	$FA,$4F,$B6,$31,$7D,$77,$7E,$93,$9D,$7C,$3A,$FB,$FF,$21,$C6,$5E
	.dc.b	$DF,$57,$FB,$47,$DF,$FE,$27,$F2,$CF,$F0,$0F,$9F,$FD,$97,$C8,$FC
	.dc.b	$DF,$E8,$4F,$96,$DF,$17,$EF,$EF,$7F,$42,$BD,$FC,$DE,$A9,$5F,$E0
	.dc.b	$CF,$CE,$FF,$09,$9F,$7F,$83,$69,$8B,$E9,$18,$CB,$A7,$A2,$D7,$FA
	.dc.b	$C6,$1C,$70,$F3,$4F,$F8,$37,$0F,$06,$5E,$4D,$3D,$3F,$5B,$8C,$FA
	.dc.b	$47,$B3,$E0,$BC,$DF,$5D,$FE,$B5,$C7,$FA,$5F,$8D,$FE,$DB,$C7,$EF
	.dc.b	$F7,$B5,$FD,$75,$EF,$E3,$BF,$59,$F8,$AF,$BB,$F2,$6E,$7F,$85,$FC
	.dc.b	$3E,$0F,$C1,$E1,$C2,$3F,$82,$C6,$0F,$E5,$18,$73,$CF,$FD,$D5,$EE
	.dc.b	$7B,$A7,$35,$FC,$33,$2E,$FF,$5F,$99,$61,$C7,$3F,$9F,$F9,$4F,$8B
	.dc.b	$7F,$5C,$7D,$2F,$DF,$FF,$C4,$DF,$83,$7F,$07,$FC,$7D,$FA,$B3,$E4
	.dc.b	$7F,$C9,$9F,$94,$BF,$4C,$FF,$70,$7E,$C9,$C3,$6F,$E7,$D8,$70,$BD
	.dc.b	$D7,$FB,$94,$D4,$A7,$4F,$E9,$31,$F0,$63,$FE,$3D,$8C,$1D,$1F,$17
	.dc.b	$DD,$B2,$F1,$7C,$67,$F8,$5E,$F8,$E5,$E0,$FA,$2C,$63,$C6,$3C,$7F
	.dc.b	$06,$FC,$D3,$C7,$A7,$9F,$3E,$9A,$FF,$BE,$FF,$7B,$CE,$DE,$0D,$BE
	.dc.b	$06,$2F,$8F,$CE,$7D,$39,$BD,$BE,$2F,$F9,$EB,$F1,$CF,$F5,$E7,$DB
	.dc.b	$BF,$51,$7E,$C5,$FE,$6A,$F0,$9C,$BF,$CD,$70,$E1,$E2,$30,$E1,$6F
	.dc.b	$C8,$BA,$FE,$0B,$FE,$13,$E7,$F8,$A7,$F9,$5D,$69,$FC,$B7,$2E,$91
	.dc.b	$8C,$7A,$68,$FA,$6F,$8B,$87,$1C,$3C,$F3,$FE,$C1,$D7,$CB,$3F,$DE
	.dc.b	$75,$F2,$EB,$E4,$E7,$F1,$E1,$C3,$67,$E5,$1F,$7B,$ED,$3F,$D8,$FC
	.dc.b	$3F,$E2,$DF,$99,$7A,$AF,$D4,$9F,$17,$BF,$C4,$3F,$D1,$FF,$51,$FE
	.dc.b	$F7,$AD,$7C,$6E,$7A,$BD,$9C,$5F,$E6,$8B,$FE,$DE,$FF,$7D,$E3,$F3
	.dc.b	$BF,$EA,$E8,$E7,$9F,$83,$F1,$0F,$7E,$DE,$4C,$FB,$7C,$76,$2F,$7A
	.dc.b	$79,$7F,$13,$F2,$70,$7A,$63,$D1,$CB,$F9,$AF,$D0,$5F,$9B,$3F,$6D
	.dc.b	$7E,$A0,$FC,$47,$FC,$A7,$F9,$77,$DB,$E3,$F2,$E1,$C2,$F7,$5A,$FE
	.dc.b	$01,$E9,$8F,$3E,$4F,$BB,$7D,$BE,$33,$E0,$C2,$39,$67,$DF,$EC,$1F
	.dc.b	$5F,$F7,$27,$39,$7F,$99,$3F,$35,$C6,$5E,$7C,$38,$6D,$FE,$61,$DB
	.dc.b	$CB,$39,$BF,$4F,$B1,$E4,$DF,$CF,$A3,$0F,$C7,$E1,$C2,$73,$1F,$E6
	.dc.b	$78,$7E,$7F,$FE,$11,$FC,$F9,$FB,$DB,$F4,$57,$EA,$EF,$DF,$3F,$D4
	.dc.b	$BF,$B4,$BF,$31,$FF,$5C,$63,$C7,$0E,$17,$BF,$8D,$5F,$E6,$F3,$EF
	.dc.b	$CD,$FE,$4F,$D3,$BF,$EC,$9F,$E0,$FC,$B9,$6B,$6B,$59,$F4,$D7,$FD
	.dc.b	$6B,$CB,$FE,$A1,$BD,$EB,$DF,$CE,$C6,$1C,$31,$F4,$E1,$E8,$C3,$CF
	.dc.b	$39,$BF,$36,$AF,$C7,$33,$2F,$D9,$1E,$EF,$EF,$7E,$35,$FA,$0F,$C6
	.dc.b	$E3,$7D,$07,$0B,$DF,$B1,$FE,$F6,$AD,$7E,$6B,$9C,$5F,$99,$7B,$73
	.dc.b	$ED,$1F,$C2,$5F,$F0,$8F,$1F,$F9,$0F,$1E,$7A,$78,$39,$3F,$86,$73
	.dc.b	$FC,$AB,$E3,$A7,$79,$7E,$B1,$F0,$A7,$37,$C7,$77,$E1,$7E,$3C,$38
	.dc.b	$43,$F0,$2F,$83,$F0,$1F,$D0,$7E,$43,$FA,$67,$EE,$9F,$AA,$F0,$5F
	.dc.b	$B1,$EB,$DC,$FE,$D9,$F0,$4A,$D7,$E2,$59,$F8,$3F,$76,$5C,$F8,$3D
	.dc.b	$A7,$FA,$3D,$ED,$7C,$7C,$9F,$F2,$3F,$27,$B6,$73,$8F,$EE,$BE,$7F
	.dc.b	$D9,$18,$AF,$F6,$5F,$E5,$BE,$3D,$FC,$F1,$FE,$CE,$FC,$0F,$19,$75
	.dc.b	$E0,$FE,$7D,$9F,$5D,$FD,$3A,$7B,$7F,$63,$FC,$0F,$9E,$F4,$5F,$9B
	.dc.b	$BD,$6F,$E5,$3F,$A6,$BF,$51,$79,$BC,$2F,$75,$A9,$7E,$9F,$9F,$8E
	.dc.b	$FC,$06,$5D,$72,$B8,$70,$F8,$66,$6F,$CE,$BE,$D5,$EF,$3F,$0F,$E6
	.dc.b	$F8,$C3,$85,$F8,$FF,$48,$F1,$65,$E7,$CB,$9E,$AE,$8F,$F2,$CF,$5F
	.dc.b	$FA,$3F,$E9,$1E,$27,$F3,$EF,$82,$7D,$E3,$F1,$87,$E6,$8F,$DE,$78
	.dc.b	$E3,$FE,$C7,$C7,$8F,$C6,$FF,$A0,$F0,$E1,$6F,$CB,$37,$7F,$E4,$1C
	.dc.b	$FF,$A5,$BF,$EB,$BF,$B5,$FF,$60,$63,$E2,$FC,$E7,$FD,$27,$18,$AF
	.dc.b	$26,$3E,$3F,$85,$FF,$6B,$4E,$73,$F0,$6A,$F6,$F3,$9F,$B7,$1E,$33
	.dc.b	$EA,$F9,$BF,$1E,$9E,$8E,$0F,$7C,$FB,$BF,$52,$F3,$FE,$88,$FD,$7B
	.dc.b	$F3,$9F,$B7,$7F,$B8,$BF,$7A,$FF,$35,$7C,$7E,$A7,$3F,$84,$A7,$FC
	.dc.b	$DD,$FB,$DE,$7D,$8F,$5F,$07,$FA,$F7,$2E,$9E,$4E,$23,$FD,$E1,$18
	.dc.b	$CF,$B7,$B3,$EA,$9C,$E7,$DB,$2F,$F0,$67,$2F,$FB,$9F,$0E,$1A,$78
	.dc.b	$7E,$D1,$F1,$BE,$5D,$1A,$7A,$65,$FA,$4E,$63,$1E,$DD,$C3,$97,$E8
	.dc.b	$FF,$63,$E8,$7F,$CB,$1F,$85,$BF,$0C,$7E,$EF,$FE,$50,$FD,$3F,$5F
	.dc.b	$5F,$E0,$AD,$4F,$F9,$53,$F6,$DC,$F0,$7B,$F9,$59,$7F,$BB,$B7,$C3
	.dc.b	$DB,$3E,$B3,$F0,$4D,$E5,$D6,$3C,$39,$66,$B8,$65,$FE,$62,$FD,$CB
	.dc.b	$BC,$F8,$BE,$AD,$FB,$CF,$FA,$E3,$EA,$FF,$DD,$DF,$BC,$FF,$5E,$FC
	.dc.b	$65,$FA,$07,$FB,$A7,$E5,$1F,$A0,$BF,$71,$E1,$C2,$18,$71,$BE,$14
	.dc.b	$FD,$D3,$75,$AF,$90,$65,$F9,$CC,$FC,$30,$FD,$D3,$15,$D7,$2E,$9F
	.dc.b	$88,$FF,$53,$F9,$5F,$CF,$E1,$C3,$1F,$3C,$FF,$57,$FB,$7C,$E1,$C7
	.dc.b	$2E,$FA,$F9,$3F,$35,$F1,$E7,$D3,$83,$DB,$FB,$2F,$FB,$6B,$F5,$AF
	.dc.b	$66,$3F,$E6,$7F,$D6,$1F,$AE,$FF,$60,$FC,$0B,$F3,$9F,$9D,$FF,$04
	.dc.b	$F2,$98,$70,$FB,$56,$1C,$39,$18,$70,$FD,$C3,$B6,$7F,$26,$F7,$C7
	.dc.b	$FB,$EC,$79,$63,$E0,$AF,$E0,$91,$8D,$3C,$5E,$6F,$F6,$89,$E6,$79
	.dc.b	$E3,$37,$FC,$12,$B8,$67,$E2,$B6,$AF,$CB,$38,$E7,$E8,$C7,$AC,$FE
	.dc.b	$73,$6F,$6E,$7E,$7D,$3D,$9B,$7C,$2F,$D5,$78,$C6,$33,$EB,$F5,$0C
	.dc.b	$1F,$C0,$1F,$33,$ED,$9E,$B7,$F8,$A7,$F6,$87,$E6,$0F,$D0,$9F,$B2
	.dc.b	$FF,$74,$C7,$8E,$FF,$8F,$C7,$8E,$6C,$78,$E1,$C3,$E9,$7C,$1A,$77
	.dc.b	$D3,$FC,$6D,$8F,$F2,$D8,$C6,$FF,$CB,$1D,$63,$CF,$5F,$17,$E9,$9F
	.dc.b	$CB,$27,$39,$F6,$F2,$71,$8F,$8F,$2F,$56,$1E,$9F,$91,$76,$78,$AB
	.dc.b	$59,$F5,$DB,$FA,$DF,$7F,$F3,$5E,$BF,$C7,$FD,$23,$F3,$B7,$E6,$8F
	.dc.b	$CF,$DF,$B6,$7F,$A8,$7F,$29,$7E,$57,$C7,$5F,$D9,$63,$C7,$0E,$17
	.dc.b	$BF,$CF,$B5,$5F,$EB,$8C,$BF,$76,$FE,$78,$FD,$D7,$C6,$FD,$E7,$57
	.dc.b	$97,$3A,$79,$AB,$CB,$A7,$93,$EE,$3F,$C5,$B5,$DB,$3F,$14,$E7,$8F
	.dc.b	$8C,$F9,$AF,$8E,$9F,$EB,$BF,$2D,$FA,$77,$19,$CE,$5F,$E7,$0F,$C3
	.dc.b	$7F,$AD,$7A,$1F,$B4,$FD,$E7,$FC,$9F,$EC,$B9,$7F,$D6,$9D,$2F,$CF
	.dc.b	$1F,$41,$C3,$7F,$E2,$D8,$70,$BD,$D7,$F9,$E7,$95,$97,$5F,$B8,$FE
	.dc.b	$5F,$87,$C5,$9C,$BC,$A1,$FB,$D7,$C5,$8E,$C3,$1E,$DF,$86,$7F,$86
	.dc.b	$3E,$57,$E6,$E3,$D7,$6F,$44,$FF,$87,$3F,$4E,$CE,$1D,$67,$3B,$FA
	.dc.b	$FF,$3E,$EB,$97,$48,$F7,$7D,$2E,$35,$FE,$A3,$C1,$F0,$BF,$4F,$7E
	.dc.b	$B7,$9C,$B9,$C3,$F5,$1F,$3F,$E6,$4F,$8D,$7E,$49,$FE,$57,$FC,$CD
	.dc.b	$F8,$D3,$F7,$CF,$F5,$47,$47,$C2,$E3,$63,$C7,$0E,$17,$BF,$A7,$EB
	.dc.b	$97,$4B,$FC,$56,$1C,$70,$F0,$7E,$4F,$F2,$63,$15,$E2,$F9,$96,$9E
	.dc.b	$2F,$13,$E2,$E3,$E0,$AD,$4F,$9B,$E9,$73,$97,$8F,$0E,$3A,$7A,$7F
	.dc.b	$25,$F1,$ED,$E6,$DD,$F8,$DF,$F8,$97,$4F,$E7,$FC,$0F,$DD,$5F,$79
	.dc.b	$FE,$29,$FD,$79,$F9,$DF,$F2,$7F,$D7,$38,$61,$FD,$87,$0E,$17,$BA
	.dc.b	$FE,$59,$F7,$1A,$AF,$F2,$3E,$2F,$D4,$7E,$04,$78,$23,$BB,$F1,$8C
	.dc.b	$5F,$3E,$EF,$DA,$C6,$39,$DA,$79,$3E,$07,$33,$E3,$C3,$D3,$5A,$CF
	.dc.b	$F9,$F7,$ED,$7E,$3D,$7C,$F3,$9F,$8A,$7E,$3D,$E7,$F8,$B7,$D0,$3E
	.dc.b	$89,$F2,$0F,$D7,$5F,$38,$FC,$E3,$FC,$C5,$9F,$F0,$BC,$78,$E1,$C2
	.dc.b	$F7,$4F,$D6,$3B,$65,$D7,$87,$E0,$1F,$D8,$DF,$F3,$33,$81,$87,$4C
	.dc.b	$FC,$5F,$11,$FD,$DB,$3F,$05,$F9,$E7,$CD,$A3,$E6,$3E,$2E,$3C,$AB
	.dc.b	$5A,$7A,$BE,$85,$FA,$1F,$93,$7F,$46,$3F,$E3,$D9,$C6,$5D,$17,$DF
	.dc.b	$E3,$0E,$5F,$BD,$FF,$B9,$FF,$11,$FF,$04,$F4,$3F,$00,$FF,$51,$FD
	.dc.b	$73,$F4,$F7,$C9,$3F,$80,$3F,$D5,$7F,$3F,$FE,$D9,$C3,$87,$E0,$CC
	.dc.b	$38,$5E,$FF,$05,$ED,$CB,$FE,$31,$BE,$DF,$83,$7F,$13,$C7,$96,$9F
	.dc.b	$EB,$2A,$FE,$75,$18,$CB,$A6,$9E,$2F,$C5,$3F,$C7,$53,$9D,$3B,$F9
	.dc.b	$18,$DF,$D7,$97,$AB,$0F,$4E,$FF,$CB,$B0,$E1,$18,$FD,$47,$CB,$A3
	.dc.b	$F1,$9F,$E7,$BE,$0F,$CE,$FD,$23,$F8,$6F,$E6,$AF,$C8,$F7,$5A,$ED
	.dc.b	$56,$BF,$24,$CF,$C7,$BF,$4B,$F7,$70,$FF,$8D,$B0,$EB,$A7,$BA,$31
	.dc.b	$5F,$9C,$C7,$9F,$AD,$FE,$56,$AE,$7A,$F8,$E1,$FB,$47,$BE,$7F,$CB
	.dc.b	$98,$F8,$F4,$F3,$FE,$1F,$C3,$1E,$7B,$BE,$17,$DF,$FB,$C3,$F5,$B7
	.dc.b	$E2,$4F,$A0,$7E,$37,$FE,$90,$8F,$F0,$C7,$43,$8F,$FC,$9F,$87,$19
	.dc.b	$E1,$5B,$FB,$4E,$B8,$F4,$E0,$F7,$1F,$E7,$1B,$F0,$7E,$61,$F1,$23
	.dc.b	$13,$E2,$CF,$C3,$F8,$0B,$F2,$DF,$75,$30,$F0,$56,$BF,$0A,$CC,$63
	.dc.b	$3F,$47,$E1,$DE,$2D,$7C,$B9,$7B,$BC,$DC,$63,$DB,$E5,$5F,$94,$E3
	.dc.b	$1F,$F5,$7C,$7F,$2D,$E2,$F8,$1A,$FF,$A2,$7A,$9F,$94,$BF,$21,$7E
	.dc.b	$FB,$FE,$56,$EC,$7E,$9F,$FA,$7C,$78,$E3,$FD,$5B,$0E,$17,$BA,$7E
	.dc.b	$BB,$F9,$2C,$BA,$F9,$7F,$17,$0E,$19,$D9,$F0,$BE,$0F,$B9,$FC,$C7
	.dc.b	$C9,$7C,$69,$CE,$5E,$3F,$C3,$71,$A7,$82,$FF,$3D,$A3,$F5,$CF,$04
	.dc.b	$E7,$7F,$4F,$EA,$2E,$07,$D5,$FF,$A3,$BF,$04,$7E,$C5,$FD,$53,$EF
	.dc.b	$5F,$9C,$3F,$75,$7C,$CF,$F7,$CC,$78,$F2,$FF,$AC,$71,$E3,$F2,$9F
	.dc.b	$E9,$0C,$38,$5F,$F8,$4E,$F7,$7F,$EB,$AC,$BF,$B0,$BF,$C1,$33,$F1
	.dc.b	$43,$E5,$FF,$85,$46,$31,$E9,$9F,$8A,$FD,$53,$9D,$3C,$F3,$FE,$7A
	.dc.b	$9C,$9E,$5C,$3D,$99,$F9,$3F,$37,$CC,$F8,$F1,$EB,$A3,$F3,$6C,$FC
	.dc.b	$C7,$FA,$13,$F6,$37,$ED,$AF,$92,$7E,$33,$FE,$1B,$F5,$77,$DC,$BD
	.dc.b	$D6,$BE,$8B,$FD,$1F,$CB,$F0,$B2,$E9,$B7,$C1,$3F,$DB,$4F,$FA,$4B
	.dc.b	$5F,$83,$18,$C7,$A6,$9E,$1D,$3F,$CB,$F8,$73,$9C,$E5,$E4,$F8,$7F
	.dc.b	$E5,$7C,$BF,$EC,$AC,$FD,$58,$F1,$D3,$FD,$3D,$8F,$1D,$7C,$F9,$FF
	.dc.b	$8D,$7E,$8B,$D7,$18,$A7,$EB,$38,$C3,$97,$DE,$FC,$AF,$5F,$FD,$0F
	.dc.b	$F3,$5F,$DD,$3F,$0D,$FF,$67,$FD,$C3,$F4,$2D,$7D,$DF,$FA,$CF,$E4
	.dc.b	$FF,$DE,$9F,$84,$7F,$A3,$7F,$1C,$CE,$3F,$03,$A1,$9F,$FA,$4B,$E2
	.dc.b	$F0,$F6,$C3,$8F,$E3,$BE,$EA,$1C,$32,$ED,$AF,$FB,$83,$F0,$DD,$4E
	.dc.b	$72,$F3,$69,$FE,$9D,$FB,$DF,$0E,$5E,$4D,$DF,$8D,$7E,$1B,$E5,$7F
	.dc.b	$CA,$7D,$5F,$EC,$BF,$8F,$FF,$10,$CB,$F3,$CD,$7A,$9F,$E5,$7D,$3F
	.dc.b	$B7,$7D,$37,$2F,$D0,$73,$8F,$B7,$2F,$F2,$EF,$EA,$77,$0E,$19,$FF
	.dc.b	$8B,$3D,$6F,$3D,$78,$32,$E7,$E2,$E2,$F7,$9F,$9B,$F8,$07,$AD,$78
	.dc.b	$36,$7E,$63,$FB,$FF,$CC,$7E,$2F,$2F,$F8,$07,$C3,$7E,$B6,$FC,$F5
	.dc.b	$FB,$5B,$CA,$FF,$06,$FE,$F3,$C7,$0F,$9B,$C6,$C7,$8E,$1C,$2F,$FC
	.dc.b	$27,$FB,$36,$EF,$B6,$5F,$E0,$3C,$FC,$38,$F1,$F1,$31,$5D,$32,$E7
	.dc.b	$A7,$87,$2F,$F0,$1E,$1C,$27,$31,$FA,$08,$C6,$3E,$9A,$F4,$7D,$27
	.dc.b	$FA,$AB,$CC,$56,$B5,$F3,$4F,$4C,$BC,$18,$71,$C5,$FA,$C7,$F1,$DC
	.dc.b	$38,$65,$F0,$E3,$FD,$31,$97,$4C,$3F,$1F,$FB,$73,$DB,$70,$BF,$38
	.dc.b	$7E,$C4,$FC,$D1,$FB,$9B,$F3,$AF,$8D,$8F,$1C,$38,$5E,$E9,$F9,$AE
	.dc.b	$B2,$EB,$9F,$F8,$AB,$7F,$B7,$8F,$87,$9F,$E5,$CA,$3A,$6B,$E2,$D5
	.dc.b	$F8,$5F,$C1,$AD,$63,$E8,$AD,$7E,$15,$9C,$38,$EB,$E3,$F5,$FC,$5B
	.dc.b	$F9,$B3,$F5,$EF,$F7,$FC,$87,$EF,$4F,$C6,$9F,$A5,$3F,$66,$7E,$A6
	.dc.b	$FC,$61,$FA,$4F,$1C,$38,$5E,$EB,$5E,$E6,$71,$F6,$E5,$DB,$C0,$B8
	.dc.b	$70,$CF,$AF,$2F,$B6,$BC,$77,$E2,$9F,$0B,$9A,$E1,$8F,$8F,$F5,$4E
	.dc.b	$D9,$F8,$75,$7E,$A3,$FD,$9F,$F0,$0F,$D4,$FD,$7F,$F8,$0B,$F4,$67
	.dc.b	$EA,$0F,$D9,$DF,$B1,$FF,$B5,$FA,$F8,$71,$7F,$14,$C3,$85,$EE,$BF
	.dc.b	$DB,$9F,$B3,$EA,$BF,$CA,$5F,$30,$CB,$FC,$D5,$9F,$82,$1F,$A5,$62
	.dc.b	$BA,$46,$F1,$E9,$EA,$FF,$9A,$AF,$86,$1E,$6D,$7C,$BF,$19,$9C,$78
	.dc.b	$CE,$74,$FF,$0F,$FE,$49,$E3,$DB,$CF,$87,$B3,$E7,$3D,$DE,$57,$FA
	.dc.b	$BF,$1E,$3A,$FA,$E7,$FD,$F3,$39,$9F,$5E,$5D,$67,$F8,$44,$BF,$00
	.dc.b	$FE,$27,$F2,$6F,$75,$F7,$1F,$E1,$5F,$C3,$3F,$DB,$FF,$84,$3F,$0E
	.dc.b	$FF,$0C,$61,$C2,$F7,$8F,$FB,$9B,$F0,$CF,$F5,$E7,$E8,$7A,$DB,$FA
	.dc.b	$1E,$9D,$7F,$1F,$FF,$4A,$63,$C7,$5B,$5A,$CB,$9E,$BF,$EE,$EE,$BF
	.dc.b	$F4,$9D,$EF,$4E,$DF,$82,$E3,$0E,$18,$F9,$F5,$7E,$87,$F8,$2B,$F7
	.dc.b	$56,$9F,$34,$3E,$4F,$FC,$65,$F8,$57,$F3,$4F,$D4,$9F,$A0,$FF,$B0
	.dc.b	$BE,$79,$F3,$FF,$E3,$BF,$D2,$5F,$19,$C3,$4F,$EB,$F8,$70,$FB,$EF
	.dc.b	$F0,$C5,$EF,$E3,$75,$4A,$EB,$C5,$FE,$B5,$D3,$C1,$A7,$FB,$42,$BF
	.dc.b	$D6,$B1,$8C,$79,$FD,$53,$4F,$F5,$AC,$E7,$5F,$2C,$BF,$71,$CE,$1E
	.dc.b	$4C,$7D,$5A,$FF,$94,$FF,$17,$F4,$F1,$7F,$B8,$EB,$5A,$76,$9F,$F6
	.dc.b	$8E,$7D,$6F,$FC,$B5,$F7,$7F,$DF,$3F,$49,$F9,$78,$3F,$54,$F2,$7E
	.dc.b	$B6,$FC,$F9,$F9,$37,$EC,$BE,$91,$F9,$7B,$1E,$38,$70,$BD,$F9,$8B
	.dc.b	$FE,$B9,$5A,$E5,$7C,$FF,$F6,$8D,$3F,$B4,$CE,$74,$C4,$79,$79,$FF
	.dc.b	$B6,$E7,$D7,$5F,$E0,$7F,$6D,$9A,$FF,$0A,$CE,$74,$7D,$0E,$67,$D9
	.dc.b	$A3,$2F,$66,$3F,$C3,$3E,$01,$E5,$7A,$F4,$EF,$7B,$E0,$F5,$D7,$F5
	.dc.b	$1D,$BC,$D1,$8F,$C1,$B8,$67,$E2,$BE,$19,$75,$FC,$1F,$E3,$7A,$9F
	.dc.b	$E0,$6F,$DF,$3F,$CC,$5F,$A2,$FF,$52,$FA,$E7,$EC,$CF,$C4,$1F,$49
	.dc.b	$C3,$4F,$F1,$8E,$1C,$3E,$57,$FD,$1B,$7B,$F4,$35,$4A,$EA,$FC,$AE
	.dc.b	$1F,$E1,$11,$FE,$71,$FD,$4B,$E6,$CF,$08,$E9,$9F,$8B,$AB,$F0,$9B
	.dc.b	$79,$FC,$76,$FE,$7D,$1F,$89,$6B,$2F,$2D,$6B,$6F,$2F,$DF,$BF,$39
	.dc.b	$ED,$39,$D3,$DF,$2F,$C2,$3D,$3A,$3F,$78,$7E,$AE,$FD,$67,$E2,$1F
	.dc.b	$D8,$7F,$07,$DC,$EA,$33,$F1,$EF,$C6,$B3,$FB,$EF,$BB,$9F,$F9,$46
	.dc.b	$5D,$72,$C3,$FE,$38,$31,$87,$2E,$1E,$B3,$E1,$EA,$F1,$7B,$5E,$F9
	.dc.b	$5F,$09,$F3,$7C,$A3,$F2,$2E,$38,$FA,$21,$FA,$17,$E9,$FE,$8B,$FC
	.dc.b	$8F,$FB,$CF,$BF,$F4,$C7,$EB,$9F,$C3,$BF,$62,$FE,$A7,$C1,$FE,$23
	.dc.b	$D9,$7B,$AD,$7E,$1D,$CE,$7F,$CE,$B3,$CD,$FC,$06,$3D,$B9,$7F,$CD
	.dc.b	$F1,$83,$6B,$2F,$0F,$57,$F2,$4F,$AB,$FE,$09,$7C,$67,$3E,$B3,$F3
	.dc.b	$AC,$E3,$CA,$73,$B7,$E1,$36,$F4,$69,$E7,$CF,$F9,$7C,$FF,$4F,$9C
	.dc.b	$E9,$E1,$DB,$E0,$7E,$E7,$EB,$CB,$DF,$39,$C3,$8F,$E5,$B9,$FC,$B5
	.dc.b	$FA,$9F,$FA,$53,$F1,$77,$CC,$FF,$9B,$B1,$7E,$FB,$BA,$D4,$E7,$F8
	.dc.b	$03,$1E,$8F,$E1,$79,$7F,$CF,$39,$74,$C7,$07,$FA,$81,$C6,$5D,$BE
	.dc.b	$CF,$F8,$BD,$70,$CB,$AC,$7F,$8D,$6F,$C0,$2B,$86,$1F,$E8,$3E,$1F
	.dc.b	$B9,$8F,$4C,$FF,$D0,$B3,$FE,$3D,$9F,$F7,$DF,$E6,$3F,$E4,$0F,$C9
	.dc.b	$5C,$4F,$94,$7E,$A0,$F9,$FF,$F5,$DE,$1C,$3E,$9D,$87,$0B,$DD,$3F
	.dc.b	$50,$FE,$71,$E7,$7F,$A6,$33,$FF,$26,$FE,$41,$FA,$78,$C6,$3C,$B8
	.dc.b	$7F,$85,$46,$32,$E9,$A7,$86,$79,$9F,$EC,$6C,$3D,$1A,$F9,$E5,$F8
	.dc.b	$1F,$F4,$FC,$39,$4E,$74,$7E,$5F,$CB,$16,$3C,$B6,$F4,$FC,$76,$7F
	.dc.b	$60,$7C,$7B,$F7,$44,$3F,$0C,$77,$BF,$31,$7E,$7B,$C7,$0E,$17,$BA
	.dc.b	$D7,$E7,$79,$9F,$85,$AF,$FB,$82,$BB,$E5,$88,$EF,$3C,$B2,$EF,$E9
	.dc.b	$7B,$67,$84,$E7,$17,$CF,$E6,$78,$E1,$E8,$C3,$CF,$A3,$F4,$DF,$34
	.dc.b	$E7,$0F,$57,$EE,$38,$D3,$D1,$F9,$AF,$FA,$03,$B5,$CF,$FE,$D2,$9F
	.dc.b	$39,$AF,$63,$F5,$9F,$9B,$BF,$08,$7C,$2B,$EF,$7F,$8C,$7D,$8F,$2F
	.dc.b	$A3,$DD,$6A,$73,$97,$1C,$FF,$8A,$63,$CF,$1F,$D3,$F1,$A3,$B7,$E4
	.dc.b	$98,$75,$87,$F2,$5F,$63,$F5,$34,$38,$65,$DF,$DC,$FF,$39,$FC,$EF
	.dc.b	$C7,$FC,$A9,$F9,$2B,$F4,$87,$E5,$CF,$DD,$FF,$DD,$1F,$80,$FF,$86
	.dc.b	$78,$7E,$1E,$9D,$71,$E3,$87,$0B,$DE,$FF,$8F,$9C,$D7,$E7,$8E,$19
	.dc.b	$3F,$58,$FC,$04,$F2,$CF,$C1,$F8,$7F,$FB,$02,$F9,$5F,$96,$73,$E0
	.dc.b	$7B,$7F,$13,$FF,$0F,$7D,$2F,$FB,$57,$F6,$A7,$B7,$FE,$DA,$FC,$71
	.dc.b	$F9,$17,$F8,$FE,$F7,$F8,$82,$F7,$F8,$5A,$F7,$FB,$DA,$F7,$F4,$1F
	.dc.b	$DF,$1E,$37,$77,$F9,$C7,$F8,$E3,$1E,$99,$F8,$7F,$01,$F3,$4E,$B5
	.dc.b	$F3,$4F,$97,$D3,$F5,$7B,$9C,$BF,$25,$FF,$7A,$7C,$57,$FB,$4B,$C6
	.dc.b	$7E,$32,$FD,$21,$F9,$C7,$97,$0E,$1F,$89,$FF,$4F,$FC,$2B,$FD,$5F
	.dc.b	$7B,$ED,$D6,$FF,$83,$D3,$AB,$FE,$42,$E3,$FF,$28,$B9,$C3,$FA,$DB
	.dc.b	$FE,$49,$8C,$A9,$F2,$E1,$DB,$C3,$F4,$CF,$1D,$BD,$19,$7E,$BE,$3C
	.dc.b	$27,$1C,$7C,$DB,$7B,$3E,$85,$F9,$2E,$B3,$EB,$39,$FB,$FC,$5F,$0C
	.dc.b	$5F,$3C,$FC,$83,$3A,$78,$21,$F9,$EF,$F2,$0F,$9E,$FA,$BF,$AF,$3F
	.dc.b	$80,$5F,$37,$FE,$9B,$FB,$0F,$F1,$E7,$ED,$CF,$D3,$38,$72,$7E,$A7
	.dc.b	$0E,$17,$BF,$AC,$5B,$F2,$1D,$DA,$FB,$7C,$EF,$F9,$0B,$BF,$FC,$5F
	.dc.b	$3C,$F4,$F0,$BF,$E7,$28,$CB,$C6,$3C,$BE,$DB,$5F,$F2,$87,$E3,$BF
	.dc.b	$BB,$BD,$E3,$E4,$AD,$68,$FA,$FD,$70,$7C,$0C,$B9,$E5,$FE,$59,$FD
	.dc.b	$D7,$FC,$C5,$97,$29,$CE,$09,$7D,$67,$B3,$3E,$D8,$7F,$98,$7C,$FE
	.dc.b	$BB,$7A,$F8,$BE,$27,$E3,$D8,$8F,$54,$7C,$7C,$FC,$9A,$77,$8C,$7E
	.dc.b	$81,$D6,$7C,$3F,$5B,$C7,$3E,$FB,$3E,$37,$E0,$65,$FA,$EF,$C8,$5F
	.dc.b	$AC,$FD,$1F,$69,$E1,$3F,$42,$7E,$82,$FD,$B7,$FB,$0E,$3C,$70,$E1
	.dc.b	$7B,$AE,$7C,$3F,$EA,$7C,$BA,$F8,$7F,$67,$0E,$1D,$B5,$FE,$59,$3E
	.dc.b	$28,$FE,$69,$87,$87,$E3,$FC,$F5,$AC,$BB,$E9,$FE,$C9,$CB,$B4,$63
	.dc.b	$E4,$5F,$A2,$79,$AF,$86,$5D,$30,$FE,$E1,$FB,$BE,$65,$A7,$87,$83
	.dc.b	$E0,$D3,$E4,$3F,$83,$E9,$DB,$1E,$51,$FE,$1D,$8C,$57,$0F,$AD,$FE
	.dc.b	$8D,$F3,$0F,$D0,$BF,$8D,$F8,$BF,$E5,$3F,$D7,$DF,$0D,$FE,$A1,$FA
	.dc.b	$8F,$ED,$FF,$18,$C3,$87,$07,$F4,$FC,$38,$5E,$FC,$2F,$F3,$AD,$6A
	.dc.b	$95,$FE,$5F,$F3,$BF,$D4,$F3,$98,$F0,$4F,$FB,$5A,$31,$A3,$3F,$1F
	.dc.b	$6B,$4F,$F0,$0D,$F8,$F5,$F4,$DE,$F4,$F4,$78,$7E,$E9,$CE,$3C,$B0
	.dc.b	$7E,$50,$FD,$F7,$96,$5E,$BC,$B9,$67,$DE,$9F,$C0,$1A,$AF,$5D,$6B
	.dc.b	$0F,$CC,$46,$76,$F8,$78,$79,$FF,$05,$7E,$8D,$9C,$38,$43,$F6,$0E
	.dc.b	$9E,$C7,$E6,$FF,$2D,$7E,$33,$FD,$DB,$ED,$5F,$9A,$BC,$0D,$F7,$AF
	.dc.b	$7F,$23,$FE,$BF,$AF,$C6,$F4,$FD,$1C,$BA,$70,$FC,$DE,$7F,$EA,$18
	.dc.b	$BF,$2F,$FA,$71,$8C,$B9,$E8,$FC,$57,$E5,$61,$FD,$2F,$0F,$06,$DE
	.dc.b	$5F,$8E,$CC,$FF,$47,$9C,$ED,$E3,$FC,$A7,$CF,$97,$4D,$DF,$41,$8C
	.dc.b	$BD,$DE,$77,$FB,$0B,$87,$D1,$97,$2D,$BF,$79,$F8,$6F,$E4,$7E,$87
	.dc.b	$FB,$8F,$E7,$DF,$EA,$FF,$03,$F8,$CF,$81,$FE,$FE,$FA,$3F,$E0,$F9
	.dc.b	$7F,$86,$38,$97,$BF,$6B,$F0,$1F,$85,$FF,$97,$B8,$7A,$57,$F8,$A6
	.dc.b	$BA,$3F,$93,$CF,$BF,$8D,$FE,$BD,$8C,$1B,$7F,$2D,$87,$4F,$D7,$BF
	.dc.b	$5B,$5C,$35,$F2,$E5,$FE,$F1,$C3,$B9,$E4,$C3,$8E,$A9,$78,$2D,$F7
	.dc.b	$A7,$A6,$B5,$3F,$E7,$8D,$BD,$D3,$9B,$7E,$68,$FC,$3B,$8E,$5D,$A7
	.dc.b	$3E,$67,$EE,$FE,$F7,$FC,$37,$F4,$0F,$CB,$1F,$8E,$BE,$6D,$FA,$BB
	.dc.b	$F5,$4F,$C9,$FC,$67,$1E,$4F,$F2,$86,$3C,$70,$E1,$7B,$E9,$FE,$D1
	.dc.b	$C5,$FE,$F1,$AD,$5F,$EF,$F4,$F0,$E8,$FC,$57,$F8,$2C,$63,$2E,$7A
	.dc.b	$7E,$4E,$FC,$07,$F9,$5E,$FC,$FA,$F9,$A5,$E2,$7F,$1E,$C3,$94,$E7
	.dc.b	$BF,$FC,$39,$A7,$8F,$0E,$3B,$7B,$65,$FA,$BF,$6D,$BD,$98,$75,$FC
	.dc.b	$B9,$F9,$26,$71,$E3,$AA,$5F,$B3,$7F,$27,$FB,$6F,$F2,$3F,$D0,$FF
	.dc.b	$AA,$BF,$00,$FE,$B3,$F5,$FF,$C1,$3F,$64,$FE,$B3,$C3,$FE,$B2,$E4
	.dc.b	$FF,$3D,$5E,$FE,$1F,$FD,$1D,$EB,$AD,$F3,$5D,$6F,$7F,$6C,$CF,$F9
	.dc.b	$6E,$9E,$08,$FE,$23,$3F,$CB,$63,$13,$E2,$C1,$E4,$7F,$44,$9C,$E5
	.dc.b	$E0,$FC,$3F,$18,$76,$C3,$87,$D3,$BF,$1A,$F1,$4F,$F9,$CB,$3F,$16
	.dc.b	$1E,$8A,$7E,$71,$AD,$FD,$9C,$3F,$C9,$3C,$5E,$B8,$F1,$CF,$C9,$E0
	.dc.b	$62,$F7,$C1,$F5,$3E,$F3,$96,$BD,$B2,$FE,$5B,$D9,$FE,$D4,$E5,$F9
	.dc.b	$59,$F3,$DF,$E4,$FD,$77,$3C,$38,$64,$FC,$73,$FD,$51,$F6,$AF,$C5
	.dc.b	$3F,$CF,$BF,$85,$BF,$1D,$FF,$52,$7E,$98,$FD,$5B,$E7,$7E,$B3,$FB
	.dc.b	$7C,$38,$6D,$FE,$B1,$C3,$87,$8A,$C3,$87,$33,$07,$E5,$CC,$78,$E2
	.dc.b	$C7,$C1,$E8,$FF,$A3,$A3,$FD,$1F,$3A,$9E,$B1,$93,$96,$BE,$8D,$7C
	.dc.b	$FA,$FE,$0F,$E3,$7D,$B3,$9D,$BC,$93,$F8,$7F,$8F,$CE,$1E,$9C,$38
	.dc.b	$EC,$FD,$87,$FD,$25,$A7,$86,$73,$1F,$E3,$08,$C6,$3D,$29,$F4,$FD
	.dc.b	$30,$F6,$61,$D3,$F7,$F7,$EF,$BE,$89,$F0,$61,$C6,$5F,$AC,$FC,$5F
	.dc.b	$7B,$FD,$55,$FB,$03,$F5,$37,$E8,$4F,$27,$FB,$AF,$9F,$FE,$63,$E0
	.dc.b	$FE,$81,$E4,$30,$E1,$7B,$F9,$5F,$E3,$5F,$93,$75,$DF,$FD,$D1,$7B
	.dc.b	$FD,$53,$E6,$46,$44,$73,$D3,$39,$76,$CB,$AE,$9F,$E0,$6F,$D6,$3F
	.dc.b	$17,$B7,$A2,$FC,$F3,$FE,$0A,$FB,$AC,$E5,$E0,$C3,$8E,$AF,$C8,$7F
	.dc.b	$DA,$99,$F5,$C7,$8C,$E7,$F4,$0F,$26,$BE,$3E,$0F,$67,$CF,$7F,$A7
	.dc.b	$B0,$E1,$18,$F9,$BF,$57,$BC,$FD,$87,$FC,$1B,$F8,$8B,$F4,$DF,$E7
	.dc.b	$9F,$B0,$FF,$C0,$1C,$7F,$B4,$5E,$FE,$4D,$7B,$E9,$5B,$D9,$EB,$97
	.dc.b	$F8,$63,$93,$FE,$00,$0C,$7C,$1F,$9C,$FC,$8E,$8F,$F7,$4C,$75,$CB
	.dc.b	$A4,$FC,$4A,$E1,$7E,$6D,$7F,$A8,$7E,$5B,$9C,$38,$47,$F5,$3F,$CF
	.dc.b	$5F,$BA,$F8,$B7,$F3,$CE,$7F,$29,$C6,$9E,$9F,$CF,$3F,$B2,$71,$7B
	.dc.b	$B4,$ED,$C4,$97,$E6,$5F,$B5,$FA,$0F,$F7,$FF,$E9,$6F,$19,$F6,$1F
	.dc.b	$EE,$5F,$92,$FF,$15,$72,$7C,$6D,$3F,$CC,$17,$BA,$D7,$D1,$A9,$FA
	.dc.b	$EF,$57,$15,$D7,$E2,$3F,$1B,$9F,$83,$3E,$F5,$FB,$FC,$BA,$C7,$8B
	.dc.b	$E9,$5A,$78,$AB,$A0,$C7,$C5,$3A,$CD,$F7,$1F,$C8,$32,$ED,$AF,$9B
	.dc.b	$F0,$B7,$E6,$1C,$B1,$E3,$9F,$F9,$1E,$7F,$DE,$7A,$F8,$AF,$DB,$F8
	.dc.b	$37,$FB,$F2,$31,$A7,$78,$7C,$3B,$CB,$FE,$39,$FA,$1B,$F7,$BF,$EA
	.dc.b	$3B,$3F,$41,$7F,$00,$BF,$6D,$7E,$99,$C7,$97,$E5,$0C,$79,$7C,$9F
	.dc.b	$FA,$6F,$0E,$37,$FC,$4B,$3E,$F9,$FF,$A7,$30,$FE,$51,$18,$BF,$DF
	.dc.b	$B7,$3F,$16,$7F,$E2,$9A,$F0,$1F,$67,$2E,$F3,$AF,$23,$EE,$5F,$1C
	.dc.b	$FD,$3F,$44,$FA,$EE,$38,$F9,$F4,$F7,$D6,$BF,$6D,$CE,$7D,$23,$E0
	.dc.b	$F5,$BF,$2C,$CD,$F0,$DB,$F2,$5F,$AF,$63,$83,$E0,$47,$F7,$AC,$FE
	.dc.b	$77,$27,$C6,$FC,$61,$F8,$DB,$F2,$37,$E9,$8F,$A8,$7E,$8C,$FD,$A9
	.dc.b	$F8,$F7,$2C,$FF,$31,$97,$2C,$78,$E1,$C2,$FF,$97,$73,$B4,$F0,$69
	.dc.b	$DF,$1F,$9F,$18,$F8,$CF,$9E,$EB,$1E,$7A,$7F,$BE,$30,$E5,$7E,$0C
	.dc.b	$3D,$1A,$F9,$E9,$A7,$D0,$C7,$94,$E7,$95,$F9,$47,$A7,$1D,$E3,$CB
	.dc.b	$6F,$7D,$3F,$6C,$D6,$5E,$DC,$7B,$7E,$61,$D7,$1F,$16,$DE,$6F,$CB
	.dc.b	$F1,$1D,$23,$D3,$EB,$BF,$6D,$E7,$AF,$7E,$0F,$E4,$5F,$7F,$8B,$F9
	.dc.b	$3F,$61,$F9,$AE,$3F,$7E,$5C,$F5,$F7,$7E,$3F,$FC,$1F,$F0,$77,$97
	.dc.b	$FA,$0F,$C9,$FF,$C2,$DE,$7F,$FB,$53,$C6,$7E,$F8,$FD,$F1,$FB,$4B
	.dc.b	$1E,$3A,$7F,$11,$C7,$8F,$C5,$7F,$39,$C3,$87,$EE,$DB,$D3,$B7,$CD
	.dc.b	$FF,$02,$C3,$FC,$1D,$8F,$25,$8F,$F8,$9F,$F2,$1F,$EC,$F3,$98,$EB
	.dc.b	$AF,$93,$F0,$AF,$56,$3D,$37,$F4,$68,$FC,$4F,$39,$F9,$74,$FE,$23
	.dc.b	$B7,$97,$F3,$C7,$ED,$DE,$5A,$D6,$3C,$A7,$FC,$75,$B7,$87,$0E,$3A
	.dc.b	$FC,$4F,$C7,$3A,$E1,$E8,$9C,$E8,$FC,$F3,$FB,$26,$1C,$72,$F2,$C7
	.dc.b	$E8,$E3,$1A,$7C,$7F,$03,$EA,$63,$FC,$4B,$F1,$BF,$F0,$BF,$EE,$AF
	.dc.b	$D4,$9E,$DB,$F7,$DF,$F0,$97,$0F,$D3,$C3,$87,$CF,$B0,$E1,$7B,$F9
	.dc.b	$BF,$DB,$EB,$FE,$41,$CF,$FC,$47,$C9,$F3,$B1,$E3,$85,$7F,$A6,$4F
	.dc.b	$8B,$4F,$0E,$9F,$EA,$29,$F0,$1A,$CF,$C1,$3A,$FD,$37,$D7,$39,$C3
	.dc.b	$C7,$6F,$D3,$BF,$D3,$99,$F8,$F1,$E3,$AF,$B6,$7F,$D9,$13,$9B,$E3
	.dc.b	$FB,$13,$F4,$2F,$0E,$4D,$7C,$D8,$FC,$5A,$7E,$1A,$FB,$77,$C0,$FF
	.dc.b	$3E,$FD,$11,$F9,$73,$F5,$0E,$5F,$94,$3D,$E7,$E9,$4F,$CA,$39,$63
	.dc.b	$C7,$0E,$1B,$30,$E1,$6F,$D2,$77,$C1,$FD,$D7,$4E,$FF,$82,$FF,$90
	.dc.b	$B6,$F1,$CE,$72,$F1,$3F,$E7,$18,$CA,$AF,$3E,$DE,$6D,$9F,$94,$FF
	.dc.b	$80,$6F,$75,$AD,$7C,$95,$FD,$D3,$0F,$19,$CF,$1F,$4E,$BE,$FD,$FF
	.dc.b	$C7,$18,$F2,$9C,$DF,$C1,$F8,$2E,$39,$75,$C3,$E3,$7C,$93,$F7,$2C
	.dc.b	$E1,$C2,$3F,$CC,$D8,$BE,$8F,$FB,$6F,$E6,$CF,$B1,$7E,$8F,$F2,$7F
	.dc.b	$34,$7D,$4F,$F8,$8F,$C4,$FB,$18,$70,$BD,$EA,$BD,$FC,$77,$FA,$BF
	.dc.b	$3E,$DC,$7F,$E2,$3D,$3B,$69,$F3,$F2,$E5,$FA,$F5,$AD,$67,$D3,$2F
	.dc.b	$0F,$A1,$F2,$EF,$C1,$87,$97,$6F,$EC,$59,$F6,$C5,$9F,$6D,$7C,$94
	.dc.b	$F9,$3F,$DB,$DE,$EB,$59,$FB,$A5,$FA,$A7,$BB,$1F,$1E,$5D,$FF,$0B
	.dc.b	$F1,$46,$33,$EB,$F2,$CF,$C9,$B8,$4F,$8B,$83,$D5,$B7,$F8,$77,$F4
	.dc.b	$6F,$D0,$7E,$3B,$FE,$1C,$FC,$71,$F9,$E3,$F2,$5F,$E1,$7F,$51,$FE
	.dc.b	$5E,$FD,$B5,$FB,$07,$0D,$3F,$C9,$98,$70,$BD,$FD,$82,$FF,$C6,$1C
	.dc.b	$EC,$FB,$67,$F8,$FC,$7F,$D9,$D3,$97,$9C,$78,$5F,$CF,$4F,$0C,$3A
	.dc.b	$FC,$4F,$E6,$E3,$1E,$73,$5F,$F3,$9F,$DE,$7E,$7F,$DF,$FF,$1B,$FE
	.dc.b	$49,$FE,$88,$FA,$9F,$F6,$67,$CB,$BE,$55,$FB,$53,$F3,$7F,$E0,$58
	.dc.b	$70,$FB,$86,$1C,$2F,$7F,$94,$6B,$87,$FC,$D9,$9F,$57,$F6,$98,$71
	.dc.b	$F8,$8F,$DD,$3C,$F0,$E5,$E9,$79,$6B,$8E,$5D,$BF,$07,$FF,$26,$67
	.dc.b	$FB,$3E,$6F,$F4,$D5,$7F,$9E,$7E,$8D,$FB,$B3,$F3,$1F,$F3,$DF,$ED
	.dc.b	$F8,$FD,$C1,$F9,$07,$F1,$BF,$A0,$FF,$2B,$E1,$C2,$F7,$82,$DF,$90
	.dc.b	$6F,$2E,$D3,$D5,$FE,$1D,$87,$28,$7E,$6F,$FC,$56,$31,$87,$4F,$C0
	.dc.b	$3F,$87,$46,$3E,$E1,$AF,$A3,$F4,$9F,$CB,$E4,$FD,$D3,$FC,$7F,$FA
	.dc.b	$87,$F9,$C3,$81,$F6,$EF,$D6,$3F,$D7,$DC,$7F,$AB,$BD,$FC,$07,$BA
	.dc.b	$B5,$ED,$FE,$7A,$7F,$C9,$3F,$47,$FB,$9C,$FB,$7D,$8B,$4C,$57,$2D
	.dc.b	$3B,$FA,$BE,$48,$C6,$5F,$EC,$DC,$3F,$A4,$F4,$FF,$44,$FC,$8B,$FC
	.dc.b	$B1,$F3,$6F,$C3,$BF,$D0,$5F,$80,$6B,$E3,$95,$A9,$FF,$43,$7D,$BE
	.dc.b	$7F,$59,$FE,$D1,$97,$E4,$6F,$6F,$F8,$78,$FF,$46,$E5,$DF,$EC,$BF
	.dc.b	$D4,$78,$3E,$7B,$FB,$3D,$71,$CB,$B5,$FF,$B6,$7F,$6F,$7E,$C2,$FD
	.dc.b	$F9,$DE,$FD,$E1,$F8,$C3,$F4,$67,$EC,$7F,$43,$E2,$FF,$D3,$B9,$72
	.dc.b	$C7,$8E,$1C,$2D,$F8,$DE,$F6,$FF,$66,$5E,$DF,$F7,$8B,$E2,$FC,$3F
	.dc.b	$FD,$E1,$1A,$D7,$C9,$AF,$8F,$57,$E3,$7E,$99,$CE,$7E,$6F,$D1,$BF
	.dc.b	$DB,$1D,$4F,$4F,$FD,$63,$FA,$BF,$FA,$2F,$C0,$FE,$2F,$F5,$2F,$DA
	.dc.b	$3F,$19,$F9,$B7,$E2,$AF,$7F,$75,$F8,$1F,$85,$5E,$FC,$37,$E4,$7B
	.dc.b	$E3,$FF,$73,$69,$D5,$FF,$70,$63,$CB,$C8,$FF,$60,$C7,$FB,$0B,$6F
	.dc.b	$23,$C3,$3E,$D9,$78,$FE,$A3,$D3,$87,$8B,$7F,$3E,$FF,$EA,$1C,$3C
	.dc.b	$03,$3F,$66,$DE,$6F,$A7,$FE,$2D,$FA,$E6,$A7,$39,$FF,$87,$7B,$DF
	.dc.b	$9D,$67,$6F,$0E,$1F,$C7,$72,$C6,$9F,$1A,$3F,$15,$18,$CF,$E1,$FE
	.dc.b	$72,$FB,$17,$D7,$67,$44,$E7,$F2,$7C,$2F,$83,$F6,$E3,$F0,$73,$FA
	.dc.b	$72,$FB,$8F,$B3,$ED,$7F,$BA,$3E,$DD,$F9,$FF,$D7,$F0,$5F,$B0,$7F
	.dc.b	$06,$FD,$55,$FB,$AF,$F0,$2C,$38,$5E,$EB,$5F,$A3,$E7,$51,$FE,$7B
	.dc.b	$9F,$C1,$3F,$58,$D1,$9E,$7C,$38,$CF,$C7,$D5,$FA,$4C,$3F,$8B,$FD
	.dc.b	$C7,$A2,$BA,$D6,$B5,$F3,$FC,$07,$F0,$B9,$E5,$97,$97,$E2,$9F,$99
	.dc.b	$BE,$5B,$3A,$75,$9C,$F8,$58,$CB,$E0,$C3,$F3,$2F,$06,$5E,$BB,$E1
	.dc.b	$F9,$86,$35,$F1,$6E,$F8,$67,$F0,$0F,$FA,$93,$4F,$E8,$F8,$F1,$D3
	.dc.b	$E6,$D6,$BF,$26,$F0,$61,$C7,$C7,$E9,$C3,$F3,$6B,$F2,$98,$FF,$86
	.dc.b	$7E,$E5,$FA,$B3,$F2,$3F,$C1,$F3,$BF,$30,$7E,$CE,$FC,$B7,$F8,$27
	.dc.b	$07,$F9,$B3,$0E,$1E,$07,$F6,$FB,$DF,$D0,$6B,$3E,$B5,$FC,$1A,$F8
	.dc.b	$67,$67,$C2,$F8,$2D,$CB,$FC,$13,$DD,$FC,$0D,$F0,$CB,$B7,$C7,$E6
	.dc.b	$3F,$0B,$8F,$29,$FF,$00,$7E,$E1,$FE,$8D,$F5,$FC,$5A,$79,$B8,$13
	.dc.b	$FE,$91,$F8,$EC,$ED,$F0,$23,$FD,$39,$C5,$F1,$6F,$84,$3C,$1F,$06
	.dc.b	$9E,$3C,$7F,$86,$F4,$BF,$65,$E7,$AF,$F8,$47,$D6,$FE,$67,$FB,$86
	.dc.b	$2F,$F8,$9F,$B9,$CB,$87,$D1,$1F,$C5,$BD,$BF,$E4,$7F,$11,$F0,$38
	.dc.b	$DF,$93,$BE,$9F,$FC,$35,$F9,$4B,$F1,$6F,$F5,$15,$7F,$6C,$BE,$17
	.dc.b	$BF,$A2,$5E,$FD,$97,$C2,$EF,$93,$F5,$B8,$F6,$CF,$F5,$3D,$1F,$E9
	.dc.b	$08,$F0,$E1,$8C,$FA,$D7,$8B,$4F,$E2,$57,$BD,$3C,$17,$E5,$F8,$BF
	.dc.b	$5E,$5D,$A7,$3F,$80,$FE,$8F,$C2,$F1,$4F,$F5,$99,$CF,$98,$FD,$77
	.dc.b	$3B,$79,$78,$7D,$78,$FE,$AF,$0C,$E1,$D3,$E3,$7B,$C7,$3D,$FE,$3F
	.dc.b	$E0,$7F,$B7,$FD,$D3,$71,$8F,$4B,$FD,$7D,$C3,$E9,$8C,$5E,$B3,$F9
	.dc.b	$38,$7F,$03,$BF,$E8,$BF,$90,$3F,$32,$FF,$66,$FB,$5F,$33,$E6,$DF
	.dc.b	$13,$9B,$DD,$6B,$A5,$5A,$FA,$CF,$E1,$5A,$7E,$7B,$2E,$9D,$DF,$C6
	.dc.b	$B1,$E3,$E1,$5A,$D6,$5C,$F1,$F2,$46,$2D,$A7,$8F,$C5,$FE,$53,$8F
	.dc.b	$8A,$73,$A3,$F3,$77,$E6,$8F,$CD,$3C,$79,$7F,$30,$DB,$C7,$9F,$E2
	.dc.b	$2F,$7F,$2F,$AA,$56,$B1,$E9,$E1,$F8,$27,$C9,$87,$1F,$17,$11,$D7
	.dc.b	$83,$E2,$7D,$9F,$FA,$27,$F7,$2E,$99,$73,$8F,$F4,$67,$EC,$98,$AD
	.dc.b	$FE,$39,$CB,$0F,$0C,$7C,$DF,$AF,$FE,$0F,$C6,$FC,$5F,$FA,$A7,$A3
	.dc.b	$FD,$5F,$F1,$07,$DA,$3F,$19,$FE,$DB,$F6,$7F,$CE,$EF,$7F,$18,$BD
	.dc.b	$D7,$F7,$2F,$D3,$35,$AF,$F4,$8C,$BF,$B1,$3F,$D4,$F3,$F0,$39,$FC
	.dc.b	$77,$ED,$65,$D3,$1E,$7E,$DF,$D4,$8C,$6B,$E3,$F9,$EC,$61,$D7,$5F
	.dc.b	$14,$FF,$BA,$7E,$CF,$FC,$A9,$9F,$6A,$7C,$4F,$F3,$AB,$DF,$BE,$BF
	.dc.b	$E3,$17,$BA,$D6,$9F,$0A,$B5,$87,$B0,$E7,$B7,$C7,$CB,$FD,$F1,$EF
	.dc.b	$79,$27,$39,$FC,$5F,$C5,$FF,$D5,$9F,$96,$3E,$23,$96,$9D,$A7,$3F
	.dc.b	$5F,$8C,$FE,$94,$7F,$AC,$FD,$8E,$95,$BF,$3B,$FC,$21,$9F,$3F,$A2
	.dc.b	$FE,$83,$F9,$BB,$F7,$AF,$F0,$D7,$E5,$AF,$82,$7E,$34,$FD,$B7,$F9
	.dc.b	$37,$07,$F9,$9B,$2F,$F0,$4E,$3C,$7E,$7D,$8F,$1C,$38,$7C,$0F,$7B
	.dc.b	$DE,$7F,$E2,$77,$FC,$1D,$8F,$35,$8F,$8F,$E0,$BF,$C1,$F3,$98,$ED
	.dc.b	$8F,$5F,$8B,$FF,$0C,$CE,$70,$F4,$EB,$E8,$F8,$2C,$EF,$EE,$9C,$FA
	.dc.b	$1F,$E6,$3F,$8C,$F2,$6B,$E0,$9C,$FD,$2F,$F5,$AC,$BA,$70,$FB,$33
	.dc.b	$FE,$AD,$7D,$0F,$5C,$67,$17,$95,$99,$E9,$3C,$7F,$6C,$79,$8F,$B2
	.dc.b	$F0,$F1,$7B,$32,$56,$BF,$5C,$CE,$DE,$C8,$79,$1D,$6F,$84,$7F,$A5
	.dc.b	$B3,$FC,$31,$FD,$F3,$F4,$67,$E5,$9F,$E2,$5E,$BF,$F2,$FF,$EA,$2F
	.dc.b	$DB,$9F,$9D,$F0,$C3,$FC,$87,$87,$0E,$8F,$F7,$E7,$D5,$B0,$E1,$F9
	.dc.b	$E6,$ED,$7B,$7F,$D6,$47,$2C,$DF,$86,$7F,$AE,$27,$A6,$9E,$2F,$B7
	.dc.b	$FF,$59,$C6,$32,$ED,$E5,$63,$0F,$44,$3F,$4F,$7E,$86,$FD,$3B,$C5
	.dc.b	$87,$19,$CF,$E4,$F8,$C3,$8C,$3F,$32,$E9,$C3,$EC,$8F,$8F,$DF,$FE
	.dc.b	$94,$C3,$C5,$FA,$73,$CC,$7E,$2F,$CF,$0F,$87,$39,$FA,$2C,$71,$7F
	.dc.b	$A2,$A3,$FD,$0B,$8F,$3D,$3D,$31,$FE,$8B,$F4,$1F,$35,$FE,$30,$FD
	.dc.b	$B9,$FB,$23,$EB,$DF,$84,$3E,$99,$FA,$9B,$E1,$1F,$5C,$CF,$9F,$77
	.dc.b	$F9,$EB,$3E,$79,$72,$C7,$8E,$1F,$E5,$FB,$DD,$AF,$C3,$EE,$FF,$9F
	.dc.b	$27,$31,$CA,$7F,$D0,$51,$8D,$3C,$DA,$79,$74,$FE,$17,$39,$CF,$C1
	.dc.b	$1F,$EA,$68,$C4,$FA,$FE,$19,$FB,$83,$F4,$0F,$57,$07,$C3,$CB,$AE
	.dc.b	$EB,$DF,$E1,$1A,$D7,$DD,$5A,$9F,$CA,$46,$4F,$64,$67,$C4,$FF,$4E
	.dc.b	$E9,$E2,$C3,$FB,$A7,$B0,$FD,$97,$87,$0F,$C8,$DF,$E8,$7C,$9F,$F9
	.dc.b	$07,$F1,$1C,$E7,$E4,$8F,$EA,$79,$73,$7F,$95,$FC,$37,$F4,$7F,$8E
	.dc.b	$7E,$A6,$FD,$5D,$F8,$8B,$F4,$57,$D2,$7F,$49,$FD,$67,$FA,$26,$1C
	.dc.b	$2F,$75,$AF,$8E,$7E,$45,$AC,$9F,$AC,$FF,$60,$E7,$FF,$5C,$E3,$C5
	.dc.b	$FE,$CD,$F2,$5F,$E7,$C8,$E9,$AF,$87,$5F,$EA,$7B,$79,$F1,$F3,$4E
	.dc.b	$7D,$2F,$EA,$57,$C3,$5F,$F1,$7F,$E6,$5F,$E5,$EC,$38,$6F,$E6,$FA
	.dc.b	$54,$BF,$26,$F3,$4E,$7F,$2C,$E7,$A7,$6C,$7F,$C7,$3E,$3E,$2B,$DF
	.dc.b	$FA,$8B,$F1,$C7,$E2,$9C,$B8,$3E,$44,$E7,$F7,$3C,$62,$87,$85,$CE
	.dc.b	$B7,$F3,$5F,$D1,$3F,$42,$FF,$4E,$FE,$83,$D3,$F0,$0F,$F7,$67,$ED
	.dc.b	$0F,$51,$FA,$2B,$F4,$7F,$ED,$D8,$70,$DF,$FD,$DB,$F3,$CC,$38,$5B
	.dc.b	$F7,$4D,$F7,$FF,$BA,$B4,$F4,$FB,$7F,$D6,$32,$E5,$AD,$AD,$67,$D3
	.dc.b	$2F,$17,$B3,$E7,$9C,$D7,$9B,$6F,$E3,$11,$D8,$65,$EB,$D5,$F3,$BF
	.dc.b	$C1,$FF,$4A,$69,$BA,$D4,$E7,$F6,$4C,$6B,$EC,$8F,$E1,$DF,$3F,$E1
	.dc.b	$CB,$AC,$E7,$D2,$FE,$B7,$1C,$35,$69,$E0,$E8,$7E,$D2,$FB,$FD,$4E
	.dc.b	$61,$F8,$06,$34,$FD,$17,$EA,$9D,$F7,$F2,$E2,$FC,$73,$FB,$DE,$9F
	.dc.b	$ED,$4F,$C1,$BF,$57,$F9,$AB,$D5,$7D,$8F,$FB,$9F,$F4,$CF,$E1,$5D
	.dc.b	$3F,$ED,$AC,$3F,$DA,$3E,$3B,$0E,$17,$BF,$77,$FA,$C6,$5D,$2F,$FA
	.dc.b	$B9,$C3,$DA,$FF,$07,$BE,$1C,$39,$67,$E0,$FC,$15,$F8,$9F,$9A,$87
	.dc.b	$83,$3E,$F3,$FE,$90,$D3,$C9,$97,$AB,$0F,$4F,$D8,$BE,$2D,$FB,$0F
	.dc.b	$93,$4E,$FB,$FF,$9A,$67,$FB,$06,$3C,$A7,$37,$FE,$B5,$8C,$67,$EE
	.dc.b	$87,$E0,$7E,$5D,$7C,$99,$FC,$8F,$DA,$1F,$51,$FC,$87,$1E,$33,$98
	.dc.b	$F9,$D1,$8D,$15,$F9,$28,$C3,$FD,$6F,$E1,$BE,$0E,$0F,$DC,$3F,$B7
	.dc.b	$7A,$2F,$80,$FE,$47,$E5,$3F,$27,$7E,$95,$FC,$DE,$F7,$5A,$97,$E9
	.dc.b	$59,$F9,$77,$E4,$DF,$D5,$74,$EF,$87,$6D,$3F,$AA,$C6,$2B,$F8,$56
	.dc.b	$3E,$0F,$D7,$FF,$A0,$CE,$75,$F2,$73,$7F,$7C,$FA,$0C,$E3,$F6,$B2
	.dc.b	$F1,$CB,$F5,$0F,$97,$EC,$78,$F0,$E3,$1F,$3B,$F7,$5C,$69,$FD,$02
	.dc.b	$3E,$8E,$38,$C3,$97,$83,$DA,$3E,$36,$DE,$69,$FF,$1D,$FE,$D6,$FD
	.dc.b	$B7,$87,$D4,$A1,$EF,$62,$F9,$6F,$E6,$AF,$F1,$FF,$C7,$3E,$63,$E9
	.dc.b	$FD,$5B,$E8,$5F,$9B,$3F,$47,$F6,$7C,$67,$F9,$13,$ED,$7F,$C1,$58
	.dc.b	$70,$BD,$D6,$BE,$83,$FD,$7D,$97,$5E,$2F,$CF,$E7,$D7,$3F,$F0,$46
	.dc.b	$3C,$7F,$0D,$B5,$AC,$B9,$F2,$B5,$FF,$0C,$CF,$80,$D6,$1E,$59,$D7
	.dc.b	$A3,$FD,$E3,$1F,$24,$E7,$F0,$C7,$EA,$8F,$92,$E5,$8B,$5F,$27,$07
	.dc.b	$D9,$BD,$FE,$33,$A9,$CE,$89,$FE,$17,$79,$DB,$CF,$0F,$63,$C5,$3E
	.dc.b	$6C,$7F,$09,$F7,$9F,$ED,$9F,$CE,$38,$6D,$E9,$8F,$F6,$E7,$EF,$F8
	.dc.b	$AD,$FE,$21,$CB,$0F,$36,$7F,$12,$DE,$36,$7D,$1F,$D4,$7E,$73,$EB
	.dc.b	$FC,$CD,$FB,$67,$E9,$3F,$76,$F7,$D6,$B8,$D5,$AF,$35,$FA,$96,$BC
	.dc.b	$87,$E7,$DF,$53,$7F,$E0,$B0,$D3,$19,$73,$C7,$BF,$CE,$F8,$EB,$59
	.dc.b	$F6,$D7,$F1,$78,$F5,$C7,$CD,$E6,$3F,$3A,$FE,$8B,$E1,$7E,$1E,$BB
	.dc.b	$D6,$A5,$FB,$1F,$1C,$7D,$38,$3C,$DF,$05,$F4,$FD,$22,$D7,$BB,$EE
	.dc.b	$DE,$EF,$F8,$1F,$4E,$DC,$3E,$99,$CF,$AB,$D7,$8B,$FD,$3B,$F8,$27
	.dc.b	$27,$FD,$3F,$F2,$3F,$E4,$4F,$B1,$FE,$33,$FA,$2B,$F2,$87,$F0,$03
	.dc.b	$F1,$8F,$F0,$27,$C4,$BF,$25,$7E,$69,$FE,$71,$C7,$8E,$1C,$2F,$75
	.dc.b	$FE,$97,$E0,$FF,$3E,$74,$FF,$46,$CF,$B6,$7F,$E6,$FC,$78,$FE,$F9
	.dc.b	$6B,$59,$73,$FA,$16,$BF,$E0,$2C,$B9,$D3,$2E,$7A,$79,$BE,$53,$D9
	.dc.b	$3C,$B4,$F5,$63,$FE,$CF,$FC,$3B,$F1,$F8,$F2,$C7,$8C,$E7,$F7,$DF
	.dc.b	$66,$1C,$72,$FF,$12,$7E,$16,$C7,$AC,$F8,$75,$F3,$4F,$F5,$B8,$E2
	.dc.b	$67,$3F,$0F,$CA,$FF,$89,$FF,$16,$CE,$9E,$08,$FF,$15,$7E,$D1,$8C
	.dc.b	$7E,$04,$7E,$3B,$6F,$34,$63,$4F,$F3,$DF,$3F,$F5,$0F,$CC,$9F,$96
	.dc.b	$BF,$57,$7C,$43,$F4,$C7,$E9,$2F,$C6,$7E,$0E,$3C,$70,$E1,$7B,$AD
	.dc.b	$72,$FF,$AB,$6B,$58,$FF,$A9,$5E,$1F,$8F,$5E,$4F,$F0,$0E,$9E,$2A
	.dc.b	$F0,$E9,$E0,$E6,$F3,$D6,$B4,$EF,$3E,$4F,$99,$CC,$B0,$E5,$AB,$F7
	.dc.b	$D7,$E5,$8F,$A6,$FF,$54,$72,$7F,$8D,$30,$E1,$9F,$AA,$F7,$5F,$B3
	.dc.b	$AD,$63,$D2,$79,$FC,$57,$82,$31,$8F,$DD,$C7,$E0,$1E,$7D,$BE,$5D
	.dc.b	$FC,$AF,$CF,$9F,$A8,$EA,$BC,$35,$AE,$45,$3F,$20,$D6,$5E,$19,$67
	.dc.b	$E1,$DB,$D1,$0F,$B3,$FF,$15,$FE,$5C,$FC,$7B,$FA,$DF,$E7,$5F,$D2
	.dc.b	$7F,$54,$7E,$A4,$FD,$A1,$F2,$7F,$EA,$FC,$38,$70,$FF,$56,$C3,$85
	.dc.b	$EE,$BF,$D5,$35,$AA,$FE,$91,$5F,$D5,$5F,$F2,$50,$8D,$7A,$89,$FF
	.dc.b	$6A,$4E,$A7,$31,$DB,$0F,$EB,$B1,$8F,$92,$65,$E8,$9F,$F2,$F5,$F5
	.dc.b	$CF,$DB,$5A,$CD,$F9,$EB,$F1,$B7,$E8,$DB,$C3,$D9,$7B,$AD,$7E,$73
	.dc.b	$99,$F5,$CE,$7B,$BE,$5E,$5D,$B4,$F9,$38,$7F,$85,$B5,$F3,$47,$CC
	.dc.b	$F1,$DF,$43,$FE,$32,$9C,$ED,$FE,$D9,$8F,$89,$18,$FD,$BB,$FC,$82
	.dc.b	$EA,$BF,$D0,$FF,$4F,$FE,$92,$D7,$FC,$CB,$FA,$2B,$F3,$8F,$F0,$F7
	.dc.b	$C1,$3F,$69,$7E,$36,$FC,$61,$F5,$7F,$E9,$FC,$38,$7D,$37,$FB,$16
	.dc.b	$F7,$5F,$A3,$CF,$B6,$6B,$DB,$F6,$A3,$31,$E1,$FD,$5F,$FD,$65,$3B
	.dc.b	$D3,$C7,$87,$8B,$F1,$1F,$F5,$15,$F8,$35,$F3,$6A,$F8,$7F,$CA,$FD
	.dc.b	$EF,$F9,$83,$F7,$5F,$E2,$1F,$B3,$BF,$5D,$FE,$E3,$F9,$D7,$FA,$6F
	.dc.b	$F7,$4F,$C5,$5E,$FC,$D5,$EF,$E5,$97,$FE,$A1,$FD,$AF,$7C,$5F,$E8
	.dc.b	$CC,$FA,$BF,$8F,$C7,$97,$E9,$5F,$96,$8D,$89,$F1,$E9,$E2,$FA,$AF
	.dc.b	$3C,$F3,$D3,$C3,$3E,$6F,$A4,$FF,$43,$7D,$2F,$E4,$3F,$48,$7E,$C9
	.dc.b	$FD,$EF,$F3,$97,$C4,$FE,$13,$B3,$FD,$5D,$7B,$AD,$4F,$F9,$46,$73
	.dc.b	$E3,$30,$FF,$55,$E9,$FE,$A6,$C3,$F3,$BF,$89,$E2,$3F,$C6,$55,$BC
	.dc.b	$7C,$1F,$05,$E4,$8C,$7C,$DB,$47,$C9,$E7,$3F,$3C,$BF,$6B,$7E,$63
	.dc.b	$FE,$06,$DB,$C7,$5A,$9C,$EE,$9C,$E9,$FE,$19,$9F,$F2,$1C,$63,$0E
	.dc.b	$9F,$13,$FA,$AD,$BF,$D7,$FF,$A2,$3F,$72,$FC,$C7,$33,$F6,$A7,$EF
	.dc.b	$9F,$E7,$1F,$82,$FD,$A5,$EF,$CA,$FF,$6F,$56,$BE,$A9,$FA,$CE,$B2
	.dc.b	$FD,$16,$7F,$C3,$7D,$5F,$F5,$6E,$3D,$FF,$2E,$FC,$07,$A5,$FC,$66
	.dc.b	$3A,$6D,$E1,$DB,$F9,$56,$1C,$F5,$EF,$1F,$CC,$23,$F8,$E7,$CD,$7F
	.dc.b	$AE,$3F,$06,$7E,$3E,$FC,$C1,$E5,$FF,$B1,$BF,$15,$FE,$A1,$B7,$0C
	.dc.b	$7F,$DB,$F8,$70,$BD,$FE,$10,$BF,$EA,$DA,$6E,$FF,$DE,$59,$7B,$9F
	.dc.b	$E5,$19,$F8,$63,$C1,$18,$3A,$BE,$3C,$FC,$5F,$13,$FD,$12,$B8,$65
	.dc.b	$E0,$C7,$FA,$C7,$07,$F9,$3F,$F5,$37,$EA,$EF,$96,$FF,$77,$7D,$77
	.dc.b	$F0,$CF,$61,$F9,$4F,$F3,$DF,$BE,$FF,$6C,$E1,$C2,$F7,$EF,$7F,$15
	.dc.b	$FC,$D7,$5C,$5F,$EF,$EA,$EB,$F9,$87,$D2,$C3,$8E,$9D,$EB,$FA,$24
	.dc.b	$63,$1E,$7E,$4F,$F5,$C8,$C7,$C3,$36,$FE,$3F,$F8,$BB,$0C,$7E,$4F
	.dc.b	$FE,$EC,$FD,$9D,$F8,$9B,$F7,$07,$E5,$5F,$D9,$7F,$1F,$7E,$8C,$FC
	.dc.b	$97,$8E,$5C,$F1,$E3,$87,$0F,$19,$87,$0F,$C3,$F7,$6B,$DB,$FD,$8B
	.dc.b	$66,$BE,$3D,$7F,$8F,$4E,$63,$19,$75,$C5,$F8,$17,$AF,$0E,$37,$E3
	.dc.b	$C3,$D5,$B7,$A7,$F1,$9F,$81,$E8,$BF,$0D,$F8,$57,$FE,$4F,$FC,$81
	.dc.b	$F8,$77,$F6,$4F,$98,$7E,$D9,$FB,$FF,$94,$FD,$D6,$1C,$3E,$AD,$87
	.dc.b	$0F,$7F,$FB,$4D,$FF,$05,$F9,$5D,$E7,$F3,$EF,$7F,$01,$FC,$AF,$5F
	.dc.b	$14,$E7,$1F,$0B,$FD,$2F,$2E,$D1,$D7,$4F,$1F,$B5,$EA,$9C,$ED,$E7
	.dc.b	$D3,$F8,$5C,$63,$57,$E3,$2F,$C8,$FF,$A6,$7D,$F7,$93,$4F,$81,$8F
	.dc.b	$F8,$0A,$B5,$F2,$3E,$CD,$BC,$B5,$AC,$79,$E3,$FD,$56,$39,$70,$7C
	.dc.b	$2C,$79,$4B,$E5,$7C,$D5,$ED,$9F,$F0,$7F,$F0,$03,$F4,$FF,$E8,$3F
	.dc.b	$BF,$72,$E3,$F9,$53,$9F,$85,$7E,$A5,$9C,$38,$E1,$FE,$1C,$DB,$CB
	.dc.b	$AF,$87,$66,$3D,$7E,$2F,$EF,$EB,$F7,$BE,$83,$E1,$FE,$A6,$FC,$FF
	.dc.b	$FC,$61,$F9,$07,$FA,$BB,$E0,$BF,$9B,$5E,$F9,$97,$BF,$61,$7F,$EB
	.dc.b	$4F,$CD,$B7,$7F,$AB,$CB,$E7,$FD,$9F,$F4,$F7,$BC,$7F,$81,$4E,$98
	.dc.b	$73,$B7,$E3,$FE,$69,$E5,$97,$83,$4F,$2F,$E2,$39,$96,$3E,$A9,$FF
	.dc.b	$24,$7E,$54,$FC,$4B,$F4,$7C,$9F,$9C,$BD,$E3,$CF,$75,$EF,$F0,$0F
	.dc.b	$4D,$6B,$6F,$74,$BF,$20,$F0,$69,$DF,$0E,$3B,$63,$2F,$37,$E2,$7F
	.dc.b	$DE,$BF,$0B,$7E,$79,$D3,$0F,$34,$E7,$F6,$9C,$5F,$0E,$EE,$DB,$63
	.dc.b	$6F,$A9,$7F,$E0,$4F,$B6,$FE,$2D,$CD,$FE,$46,$AF,$E3,$1F,$A4,$FF
	.dc.b	$A1,$3F,$54,$FE,$ED,$F4,$1F,$E5,$AF,$94,$7E,$98,$C3,$86,$0C,$38
	.dc.b	$79,$EC,$38,$7B,$FB,$FB,$1F,$E0,$75,$D6,$FF,$CE,$B8,$72,$8F,$F4
	.dc.b	$9F,$E8,$DF,$E2,$38,$C7,$DF,$B5,$F1,$F8,$AF,$C8,$7D,$39,$74,$AD
	.dc.b	$6B,$E1,$AF,$D3,$67,$E1,$3D,$B3,$AD,$5F,$A8,$BF,$57,$7E,$AD,$BB
	.dc.b	$F6,$5E,$EB,$5F,$88,$67,$83,$DF,$8F,$23,$C7,$7D,$B6,$F9,$37,$FE
	.dc.b	$86,$DF,$D7,$1F,$E2,$1F,$CC,$BF,$C0,$FF,$09,$FA,$79,$CE,$E9,$CF
	.dc.b	$E2,$1D,$F5,$F6,$C7,$F8,$3B,$D7,$EF,$9F,$9B,$2E,$9F,$43,$FE,$F9
	.dc.b	$C9,$F1,$9F,$CC,$BF,$4F,$7E,$9A,$FD,$BF,$FD,$5F,$FB,$9B,$6F,$F5
	.dc.b	$C7,$87,$F7,$70,$E1,$7B,$FC,$19,$7B,$F0,$3F,$D7,$39,$F6,$DB,$FB
	.dc.b	$06,$9D,$B4,$FB,$59,$72,$FD,$8E,$D6,$B3,$E9,$97,$38,$F0,$19,$BF
	.dc.b	$2E,$BF,$E1,$FC,$BB,$4B,$2E,$DA,$3F,$71,$7E,$D0,$FB,$97,$EA,$7D
	.dc.b	$30,$E1,$AF,$AE,$F7,$4F,$C6,$75,$97,$49,$E7,$E9,$FE,$9B,$87,$E1
	.dc.b	$EF,$FE,$D6,$9E,$FF,$3B,$9C,$FD,$1B,$78,$FE,$85,$F9,$37,$C8,$E1
	.dc.b	$F4,$E9,$D6,$73,$F3,$7E,$0E,$0F,$24,$3F,$41,$E9,$97,$A3,$5F,$04
	.dc.b	$FF,$6C,$C5,$F4,$BF,$DB,$BF,$0D,$7E,$C1,$EB,$F2,$9F,$BC,$BF,$62
	.dc.b	$FC,$67,$EA,$1F,$B9,$C3,$85,$EE,$B5,$E1,$BF,$23,$D6,$DF,$E1,$BC
	.dc.b	$FA,$F5,$FE,$D7,$1E,$3D,$56,$B5,$97,$3D,$3F,$81,$47,$21,$87,$9A
	.dc.b	$BF,$DA,$B9,$F9,$7E,$CF,$F8,$9E,$3D,$36,$7E,$61,$FE,$41,$FA,$47
	.dc.b	$EA,$7C,$F2,$E5,$5A,$FA,$7F,$F1,$4F,$EF,$F9,$DF,$E1,$63,$FE,$49
	.dc.b	$FA,$1E,$BB,$F9,$B7,$FF,$72,$CF,$F9,$37,$E7,$B3,$3E,$6B,$F9,$5F
	.dc.b	$88,$FF,$37,$FC,$BB,$FD,$07,$D1,$F4,$2B,$96,$AA,$D7,$F0,$07,$87
	.dc.b	$3F,$F7,$8C,$7F,$3F,$F9,$9E,$BC,$1E,$88,$FE,$85,$ED,$FF,$32,$FA
	.dc.b	$9F,$F2,$CF,$EA,$7F,$E5,$CE,$17,$E8,$0F,$AF,$FA,$9F,$8D,$FF,$31
	.dc.b	$BD,$FD,$9F,$F8,$DA,$B5,$DE,$F9,$8D,$71,$7F,$97,$F3,$FF,$33,$BF
	.dc.b	$CC,$30,$EF,$FB,$37,$F2,$F7,$9E,$9E,$0D,$3F,$B4,$D7,$21,$F6,$6C
	.dc.b	$3C,$D9,$F9,$7E,$5F,$D5,$5D,$F5,$FF,$24,$EF,$FC,$37,$F2,$67,$E9
	.dc.b	$5C,$F2,$E5,$5A,$F5,$95,$FE,$64,$E9,$6F,$E4,$C7,$F8,$B7,$E6,$3D
	.dc.b	$B3,$F1,$ED,$F1,$E7,$F8,$D7,$BD,$9D,$3A,$F0,$FF,$BA,$FF,$44,$7E
	.dc.b	$78,$FC,$47,$C5,$8F,$CF,$FC,$45,$5A,$F8,$BC,$F1,$7F,$98,$A7,$3F
	.dc.b	$15,$D6,$F8,$5B,$F3,$DF,$F4,$07,$CD,$3F,$2E,$FF,$30,$FE,$8E,$FD
	.dc.b	$9F,$F6,$EF,$EF,$7F,$5E,$FF,$41,$7C,$22,$F8,$72,$FF,$A2,$6B,$87
	.dc.b	$A2,$78,$7D,$D3,$F1,$6E,$19,$F6,$CF,$F9,$2E,$1F,$E7,$0C,$79,$BC
	.dc.b	$63,$F3,$0F,$F8,$66,$33,$87,$63,$C9,$F6,$9E,$AB,$E9,$97,$87,$DD
	.dc.b	$6A,$FC,$F7,$59,$79,$F3,$F6,$FA,$EF,$84,$FF,$68,$FE,$83,$E6,$AD
	.dc.b	$65,$EB,$AD,$7E,$0E,$FC,$8B,$55,$E7,$AD,$61,$CE,$33,$BF,$9F,$0E
	.dc.b	$DF,$15,$E2,$9E,$13,$FE,$74,$FC,$09,$FC,$02,$F6,$78,$ED,$F4,$BE
	.dc.b	$8F,$FD,$5D,$3F,$EE,$8E,$5F,$87,$39,$BE,$18,$F0,$D7,$E9,$61,$FE
	.dc.b	$B5,$FC,$03,$FD,$B5,$3F,$EB,$FF,$B9,$FF,$4D,$F8,$7F,$E7,$1E,$07
	.dc.b	$DE,$7E,$73,$F1,$1F,$F7,$55,$6B,$ED,$BE,$34,$FE,$83,$F6,$1C,$FD
	.dc.b	$F3,$F7,$FF,$F2,$76,$7D,$86,$1D,$9F,$F7,$71,$89,$FC,$EE,$1D,$3C
	.dc.b	$EE,$15,$DF,$3F,$1F,$9B,$8B,$DF,$AF,$FB,$BF,$D5,$1F,$B9,$78,$B8
	.dc.b	$3F,$D1,$53,$9F,$53,$1A,$23,$F8,$16,$7C,$EB,$7E,$57,$F8,$C3,$A3
	.dc.b	$F1,$F3,$EF,$F9,$87,$EE,$8F,$AB,$FF,$7E,$D6,$B6,$F4,$D6,$A5,$FB
	.dc.b	$47,$82,$38,$FA,$DD,$F5,$FE,$FD,$8F,$7F,$9B,$FF,$67,$FE,$A9,$FE
	.dc.b	$72,$FC,$BB,$FA,$3F,$ED,$8F,$D3,$7F,$D6,$F5,$FA,$F3,$7F,$72,$AD
	.dc.b	$CF,$FA,$F3,$E3,$35,$D7,$FE,$FB,$C7,$A3,$F5,$38,$7F,$C2,$71,$FE
	.dc.b	$80,$FD,$8B,$EB,$CF,$01,$1E,$3F,$93,$69,$E3,$F0,$FF,$83,$EB,$E2
	.dc.b	$D3,$C3,$E9,$E3,$1F,$06,$DE,$5C,$5F,$B4,$F4,$FC,$69,$FA,$2F,$7D
	.dc.b	$7B,$4E,$7F,$0A,$C6,$1C,$63,$C7,$CB,$D7,$2F,$2C,$E6,$FF,$5D,$C3
	.dc.b	$EC,$8F,$07,$E2,$6A,$F9,$5F,$B9,$F4,$DF,$FB,$9C,$E7,$97,$FC,$1F
	.dc.b	$87,$19,$CD,$BC,$AF,$06,$5F,$0F,$0E,$9E,$B7,$59,$FF,$56,$7C,$3B
	.dc.b	$D4,$7E,$1A,$FB,$6F,$F1,$E7,$E1,$DF,$DB,$E7,$A6,$7F,$E2,$37,$73
	.dc.b	$AF,$93,$7E,$21,$AA,$FF,$4F,$D3,$E9,$3E,$BE,$7E,$08,$FF,$79,$CF
	.dc.b	$BC,$E0,$7F,$A7,$DE,$BE,$13,$2F,$F1,$B4,$EB,$4F,$46,$9E,$7A,$D7
	.dc.b	$C9,$FB,$6B,$CD,$3E,$BA,$FF,$1B,$FE,$92,$FC,$67,$96,$5E,$A9,$CF
	.dc.b	$D9,$62,$F8,$7C,$F7,$F0,$2C,$71,$C3,$FE,$B3,$87,$A9,$DE,$7D,$F5
	.dc.b	$F2,$7F,$3B,$7E,$90,$F0,$FF,$B1,$F1,$7C,$1A,$D4,$E7,$E4,$FA,$DF
	.dc.b	$0B,$E9,$87,$C3,$BD,$F3,$FD,$97,$E4,$7E,$76,$FD,$69,$F8,$DB,$E1
	.dc.b	$5F,$9C,$FF,$94,$F8,$7F,$CE,$FB,$FD,$EC,$FF,$1D,$8F,$1C,$38,$5E
	.dc.b	$FE,$EB,$F1,$59,$F6,$CD,$F3,$9F,$6A,$31,$77,$1E,$75,$E2,$D3,$C3
	.dc.b	$F5,$6F,$C0,$BD,$15,$AB,$EF,$2F,$B2,$CC,$B5,$F2,$60,$FC,$DD,$FA
	.dc.b	$37,$F8,$23,$DB,$F2,$DF,$3A,$D4,$FF,$83,$3F,$66,$CC,$F3,$D7,$C9
	.dc.b	$87,$2F,$83,$EF,$97,$68,$EB,$E3,$E2,$BD,$59,$BF,$0B,$FB,$5F,$83
	.dc.b	$74,$C3,$CB,$39,$E9,$7E,$69,$9C,$38,$E1,$F8,$1D,$31,$C5,$F5,$70
	.dc.b	$E9,$D5,$FB,$FE,$17,$E3,$FF,$A3,$F5,$9F,$78,$FE,$00,$79,$0F,$B4
	.dc.b	$FB,$FF,$8D,$3F,$24,$FC,$06,$3C,$70,$E1,$7B,$D5,$6F,$DB,$37,$BF
	.dc.b	$F0,$FD,$3B,$FD,$17,$F2,$CD,$BC,$73,$9C,$BC,$4F,$F6,$68,$CA,$AF
	.dc.b	$3E,$DE,$6D,$BF,$27,$BF,$B7,$6F,$24,$E7,$5F,$E1,$78,$F2,$C3,$8E
	.dc.b	$DF,$81,$FB,$87,$E2,$0F,$C6,$BC,$FA,$F7,$DB,$DB,$D0,$BD,$FE,$35
	.dc.b	$AC,$B9,$CE,$7D,$0F,$16,$1E,$BD,$BE,$5C,$F1,$F4,$B3,$3D,$67,$E7
	.dc.b	$FE,$04,$FD,$FF,$FD,$C1,$F2,$1D,$75,$F9,$B3,$9F,$C1,$7F,$C1,$5C
	.dc.b	$5E,$B9,$CD,$FE,$77,$1E,$18,$F4,$8F,$D2,$47,$ED,$CF,$E4,$5F,$86
	.dc.b	$3E,$B9,$F5,$CF,$AE,$7E,$77,$FE,$E9,$F4,$FF,$CF,$3A,$7F,$29,$C5
	.dc.b	$8F,$1C,$38,$5E,$FE,$67,$B5,$7F,$4D,$AE,$DF,$53,$F7,$E3,$C3,$9F
	.dc.b	$81,$FB,$C6,$30,$E7,$F7,$9E,$8C,$BC,$18,$77,$D3,$CB,$3E,$F9,$CE
	.dc.b	$1E,$A9,$7D,$4B,$D8,$7E,$4C,$FD,$4F,$CB,$A7,$AB,$0E,$17,$BA,$7C
	.dc.b	$9E,$B4,$F0,$4F,$3F,$C8,$3F,$55,$87,$9F,$3E,$F9,$FE,$96,$38,$65
	.dc.b	$E3,$F9,$E7,$EB,$AF,$8F,$FA,$1B,$FA,$38,$7E,$85,$6A,$7F,$CC,$38
	.dc.b	$75,$8E,$3D,$9E,$08,$F9,$98,$F5,$C7,$FA,$BF,$C2,$BE,$67,$FA,$F7
	.dc.b	$B0,$FC,$75,$FA,$17,$F9,$F7,$C0,$7E,$45,$FB,$EF,$CB,$58,$F1,$D3
	.dc.b	$FC,$09,$8F,$1C,$38,$5F,$F0,$4F,$43,$79,$F6,$CF,$AE,$1F,$E5,$CC
	.dc.b	$79,$2C,$7C,$5E,$07,$F8,$AA,$73,$1D,$7C,$F6,$BE,$4F,$53,$FC,$29
	.dc.b	$5A,$C7,$D3,$5A,$D3,$FB,$1E,$5E,$23,$CF,$3A,$FA,$AF,$F7,$62,$FD
	.dc.b	$33,$53,$E3,$8D,$7E,$03,$F1,$7E,$A7,$3A,$F8,$F0,$FF,$7A,$7D,$EF
	.dc.b	$8B,$83,$DD,$C5,$FE,$A5,$9F,$F7,$26,$5C,$32,$F2,$FE,$78,$FD,$2F
	.dc.b	$F5,$9F,$05,$9C,$3F,$D5,$93,$9E,$8F,$F5,$27,$17,$B6,$73,$7F,$EA
	.dc.b	$38,$C7,$12,$FA,$76,$FC,$4F,$A2,$FF,$0D,$2F,$C6,$7F,$C5,$5F,$46
	.dc.b	$FC,$75,$FB,$73,$F4,$DF,$C0,$E6,$C7,$8E,$1C,$2F,$75,$FC,$46,$B5
	.dc.b	$C3,$FD,$1E,$BF,$C6,$CF,$DD,$BE,$3F,$A3,$FD,$B7,$9F,$A2,$CF,$C5
	.dc.b	$9B,$E2,$7D,$18,$73,$AD,$69,$E6,$F8,$AC,$CF,$F8,$2F,$3F,$5F,$ED
	.dc.b	$AF,$B2,$FF,$78,$7E,$6F,$F9,$6D,$7B,$EA,$AD,$7E,$00,$FC,$73,$59
	.dc.b	$77,$AD,$62,$FC,$FB,$BF,$27,$C9,$DF,$D1,$93,$F2,$6F,$F2,$37,$27
	.dc.b	$C2,$CB,$C9,$5E,$DF,$B0,$7E,$9C,$FD,$17,$FD,$73,$7B,$CB,$A6,$AB
	.dc.b	$DD,$7E,$8E,$B5,$C5,$F6,$27,$9F,$E5,$7E,$48,$C6,$3E,$BC,$FF,$6B
	.dc.b	$F7,$DF,$E5,$9F,$0B,$F4,$FF,$B1,$BD,$D7,$D7,$3F,$22,$7E,$C3,$F0
	.dc.b	$F6,$63,$C7,$D0,$63,$C7,$0E,$16,$FD,$47,$BE,$9F,$E8,$5D,$3B,$BF
	.dc.b	$BD,$C7,$97,$E9,$3F,$BB,$7A,$63,$CF,$4F,$E2,$71,$8D,$7C,$B0,$F5
	.dc.b	$FF,$7B,$39,$FC,$A1,$FB,$DF,$F7,$2F,$D8,$3C,$39,$B4,$F7,$E4,$BD
	.dc.b	$FC,$9F,$E9,$78,$BD,$B7,$E7,$E2,$F6,$DE,$F2,$FF,$48,$4F,$4C,$79
	.dc.b	$E3,$E0,$D3,$C7,$3C,$76,$F9,$9F,$B0,$FF,$91,$7F,$2C,$7E,$93,$C7
	.dc.b	$4E,$B3,$9F,$DE,$F1,$7C,$3E,$49,$87,$4C,$7D,$58,$70,$9F,$E9,$B1
	.dc.b	$FC,$C3,$F1,$9F,$F4,$A7,$A8,$FB,$27,$EE,$EF,$4D,$F9,$EF,$F0,$3F
	.dc.b	$85,$7E,$E9,$C7,$2F,$F5,$D6,$3C,$70,$E1,$7B,$F8,$1D,$57,$F9,$77
	.dc.b	$2E,$AF,$F3,$7C,$FC,$31,$F9,$C8,$C1,$D1,$F1,$E7,$E2,$F5,$FF,$8B
	.dc.b	$57,$0C,$7B,$C7,$F8,$37,$A5,$39,$A7,$EC,$AF,$DF,$DF,$B0,$7F,$A2
	.dc.b	$EF,$79,$F7,$CB,$E0,$5E,$E9,$F8,$A7,$AB,$7F,$76,$5D,$32,$7E,$09
	.dc.b	$FD,$23,$4F,$06,$3D,$FF,$26,$C6,$9D,$70,$71,$FF,$81,$3F,$69,$FF
	.dc.b	$35,$EF,$E2,$AF,$8D,$BF,$D1,$AD,$7E,$79,$E1,$E2,$F8,$71,$FE,$0C
	.dc.b	$CB,$AD,$F0,$CB,$AF,$CB,$FF,$AE,$FC,$A7,$EA,$3F,$ED,$1F,$C6,$BF
	.dc.b	$89,$7E,$D2,$FC,$87,$F8,$A7,$A8,$FD,$D1,$FA,$8F,$EF,$30,$E1,$ED
	.dc.b	$B0,$E0,$FF,$2F,$AF,$D6,$65,$DB,$5F,$F7,$E6,$7F,$C7,$FF,$33,$FF
	.dc.b	$1C,$47,$FB,$FE,$74,$A7,$C6,$EA,$33,$1E,$6D,$3C,$BE,$F7,$1C,$FC
	.dc.b	$79,$F8,$B5,$7A,$F9,$CB,$C3,$AF,$9F,$F2,$37,$E4,$7F,$E8,$3F,$C7
	.dc.b	$5F,$A0,$FB,$6F,$79,$2F,$7F,$68,$B7,$E5,$7B,$BF,$46,$5D,$72,$FF
	.dc.b	$47,$7D,$07,$86,$73,$97,$39,$FF,$49,$7E,$D3,$9E,$2F,$A3,$A3,$EB
	.dc.b	$9D,$5F,$E9,$8F,$C8,$FB,$56,$A7,$3F,$70,$FD,$7B,$3B,$7B,$67,$8F
	.dc.b	$6F,$58,$C6,$2F,$C2,$1B,$DF,$B9,$AB,$F4,$67,$EC,$7F,$E5,$7D,$3E
	.dc.b	$F7,$2F,$F8,$D3,$E8,$38,$5F,$2B,$DF,$D4,$FF,$AB,$B9,$7E,$37,$91
	.dc.b	$FE,$B7,$D3,$FD,$4F,$5A,$BF,$C0,$C6,$33,$FD,$9C,$63,$4F,$1D,$74
	.dc.b	$D3,$C3,$F1,$9E,$8C,$38,$67,$DF,$5F,$F0,$B7,$C2,$7A,$BF,$7E,$FF
	.dc.b	$5F,$7A,$AF,$15,$FB,$4B,$E1,$9F,$8A,$3F,$70,$FE,$7B,$87,$0B,$DD
	.dc.b	$6B,$F1,$CC,$F8,$7F,$E5,$6E,$3F,$EA,$07,$CA,$BF,$F6,$67,$B9,$FD
	.dc.b	$46,$31,$3F,$E5,$CC,$39,$C7,$F3,$1A,$ED,$9F,$8F,$C2,$FF,$1E,$4F
	.dc.b	$D4,$FD,$F1,$FB,$63,$F6,$07,$EB,$CE,$67,$EA,$7E,$CF,$DC,$FF,$C9
	.dc.b	$78,$70,$BD,$D6,$B3,$7E,$F5,$AC,$9F,$BC,$7F,$98,$7E,$AF,$F0,$38
	.dc.b	$F1,$D7,$FC,$D1,$18,$AF,$16,$BE,$1D,$5F,$A8,$7A,$2D,$87,$86,$B5
	.dc.b	$F8,$B7,$D7,$F0,$3F,$1B,$FA,$7B,$F5,$CF,$E2,$3F,$96,$3E,$CF,$FD
	.dc.b	$0D,$F9,$73,$F6,$1F,$C6,$E8,$C3,$87,$B2,$C3,$87,$0B,$07,$EB,$9C
	.dc.b	$2F,$B7,$C3,$FF,$0E,$C7,$1A,$7F,$0E,$7F,$3F,$E3,$FF,$BE,$AB,$AE
	.dc.b	$5D,$3E,$23,$D1,$5F,$A6,$9C,$ED,$FE,$9B,$F3,$FF,$B5,$57,$F8,$FF
	.dc.b	$F0,$77,$E9,$9F,$E1,$CE,$27,$EB,$8F,$8D,$FF,$04,$DE,$EB,$52,$FC
	.dc.b	$23,$3F,$54,$C7,$F1,$5A,$FF,$4A,$87,$DF,$63,$1E,$5E,$73,$4F,$E0
	.dc.b	$F7,$CB,$0F,$1C,$E7,$EF,$7F,$CB,$78,$7E,$03,$C4,$7E,$75,$FE,$D2
	.dc.b	$FC,$7B,$CD,$F4,$7F,$E5,$6C,$BB,$F8,$9F,$91,$AD,$71,$3E,$7B,$59
	.dc.b	$7E,$2A,$BF,$CE,$17,$FC,$32,$3B,$3F,$E7,$53,$11,$CF,$E3,$99,$7F
	.dc.b	$AA,$3D,$AF,$F2,$E5,$F9,$72,$F2,$7D,$9F,$F4,$5F,$03,$F8,$AF,$A5
	.dc.b	$FC,$C3,$EE,$5F,$40,$FC,$FF,$F7,$DF,$87,$3F,$07,$FF,$28,$E1,$C3
	.dc.b	$E3,$3F,$D4,$57,$BF,$59,$6F,$D8,$B7,$C7,$FC,$93,$4E,$CF,$F5,$1C
	.dc.b	$79,$7E,$FB,$F8,$BF,$94,$FF,$65,$D7,$8F,$2E,$9F,$90,$79,$F0,$E5
	.dc.b	$A7,$83,$6F,$E2,$5C,$99,$96,$1C,$7E,$DB,$ED,$7D,$43,$F3,$EF,$1D
	.dc.b	$6B,$1F,$4D,$6B,$97,$38,$71,$F0,$5F,$8F,$F5,$8C,$7C,$6D,$C3,$FD
	.dc.b	$85,$DA,$FD,$89,$F9,$8B,$F1,$E7,$EB,$FD,$2F,$75,$A9,$FC,$DF,$2F
	.dc.b	$C0,$9C,$DF,$1D,$F8,$76,$7E,$17,$0E,$BD,$9E,$FF,$AE,$7D,$AF,$F9
	.dc.b	$2F,$F6,$17,$5B,$F6,$C7,$E4,$4F,$E0,$1F,$83,$87,$0F,$A7,$61,$C2
	.dc.b	$F7,$FB,$5E,$B2,$FF,$2B,$6D,$F5,$CF,$EB,$B8,$78,$3F,$64,$FB,$1E
	.dc.b	$17,$F5,$E8,$E9,$8F,$3D,$BC,$D8,$79,$76,$F2,$72,$63,$57,$EB,$EF
	.dc.b	$D3,$1E,$1F,$F6,$AF,$3B,$C5,$5A,$9C,$EF,$FA,$7E,$86,$FE,$1D,$3F
	.dc.b	$CD,$99,$73,$CB,$FB,$8D,$6A,$FD,$35,$AC,$3B,$FC,$6F,$8F,$F6,$07
	.dc.b	$E6,$8F,$DC,$DF,$B4,$FF,$12,$CB,$96,$5F,$EA,$CC,$BF,$CF,$58,$70
	.dc.b	$BD,$FE,$81,$E8,$8C,$71,$F5,$CB,$A7,$E1,$FF,$DB,$F1,$FF,$02,$EC
	.dc.b	$FD,$7B,$FD,$F1,$F0,$2F,$9B,$FC,$DF,$ED,$3F,$B3,$F8,$6F,$F0,$5E
	.dc.b	$1F,$1A,$F7,$5A,$FD,$3F,$FC,$25,$3D,$27,$CE,$FF,$34,$31,$F8,$87
	.dc.b	$F3,$9B,$E7,$97,$86,$3F,$A4,$61,$FA,$AF,$C8,$FC,$D7,$FE,$1C,$AD
	.dc.b	$67,$E5,$F2,$FF,$4B,$3E,$03,$59,$31,$FF,$38,$FE,$7E,$FC,$AD,$F9
	.dc.b	$77,$F7,$0D,$7C,$37,$BF,$B9,$5E,$E9,$F8,$27,$5C,$BE,$06,$5D,$32
	.dc.b	$7C,$27,$9A,$73,$8F,$29,$7E,$77,$9D,$FD,$DF,$93,$7F,$16,$F8,$2F
	.dc.b	$F6,$B4,$F5,$9C,$FB,$E9,$FF,$6D,$76,$67,$83,$D9,$87,$1F,$C2,$3A
	.dc.b	$46,$2F,$87,$D8,$7F,$8A,$FF,$43,$FF,$20,$FE,$0E,$FD,$23,$FA,$DF
	.dc.b	$AE,$FC,$8B,$FA,$07,$B5,$FE,$E4,$E9,$61,$C2,$F7,$5A,$FC,$A7,$FC
	.dc.b	$9D,$39,$F2,$7E,$D3,$BF,$94,$BC,$FF,$E5,$2D,$3C,$11,$FE,$6C,$C3
	.dc.b	$FA,$77,$BB,$E5,$AD,$65,$DB,$4F,$B7,$A7,$8A,$73,$87,$FA,$67,$B1
	.dc.b	$AB,$F4,$FF,$F3,$65,$EE,$FF,$DB,$38,$FF,$78,$FC,$07,$EF,$DE,$FF
	.dc.b	$42,$F4,$D7,$49,$CF,$C6,$78,$E3,$8E,$7D,$ED,$E4,$67,$83,$E5,$71
	.dc.b	$3F,$45,$7E,$F6,$FC,$69,$FB,$27,$FB,$B7,$83,$C9,$87,$0B,$DD,$6B
	.dc.b	$F4,$3F,$1E,$BE,$8F,$C6,$BE,$4B,$EA,$CF,$B6,$7D,$7E,$FB,$C5,$C7
	.dc.b	$FE,$9E,$CB,$9F,$85,$FE,$80,$FB,$0F,$F0,$46,$3F,$E8,$2F,$C6,$DF
	.dc.b	$C0,$1F,$D8,$FF,$35,$FF,$27,$FE,$0A,$FD,$41,$FB,$CF,$F9,$47,$E4
	.dc.b	$7F,$D0,$F8,$70,$F9,$96,$1C,$2D,$FB,$D6,$F3,$ED,$9F,$F7,$57,$FD
	.dc.b	$83,$A7,$89,$63,$CD,$CE,$5D,$AB,$C9,$F6,$3F,$EB,$3F,$9D,$F5,$4E
	.dc.b	$74,$F0,$CF,$FA,$DF,$1C,$E1,$E0,$97,$E8,$CF,$D2,$1F,$B6,$3F,$7D
	.dc.b	$7E,$E7,$E7,$C3,$C7,$AF,$F8,$17,$0E,$17,$BF,$81,$D6,$FF,$06,$73
	.dc.b	$F5,$BE,$0C,$38,$E9,$E0,$D3,$BF,$D5,$67,$0F,$91,$E0,$7F,$A4,$7E
	.dc.b	$01,$F7,$FF,$D8,$78,$7E,$25,$6B,$F0,$25,$6B,$E5,$BC,$3C,$3F,$E1
	.dc.b	$9C,$79,$7C,$97,$58,$C5,$F0,$F8,$8F,$F3,$B7,$EC,$BF,$E3,$0F,$D1
	.dc.b	$7F,$B7,$7D,$4B,$C8,$7E,$F8,$FA,$6F,$E1,$9F,$2B,$FE,$C4,$C3,$87
	.dc.b	$CE,$B0,$E1,$7F,$E3,$2F,$83,$FE,$9F,$5D,$7A,$3F,$D1,$27,$FB,$39
	.dc.b	$FF,$1F,$FE,$33,$FE,$8F,$8C,$7D,$43,$5F,$1E,$AF,$CD,$BC,$F3,$9D
	.dc.b	$3C,$12,$FD,$63,$37,$C7,$F2,$57,$2F,$FA,$F7,$CE,$F4,$E7,$DA,$B5
	.dc.b	$3F,$EC,$1F,$C1,$73,$2D,$BE,$25,$30,$E7,$F8,$B7,$68,$C6,$0F,$C5
	.dc.b	$BC,$3F,$BD,$7F,$A4,$FF,$45,$7E,$14,$FD,$53,$AE,$7F,$3E,$B5,$2F
	.dc.b	$DF,$B3,$C7,$F0,$67,$97,$E0,$FD,$3F,$01,$65,$F0,$F8,$3C,$FF,$02
	.dc.b	$FC,$3B,$FC,$93,$F9,$E3,$F5,$6F,$F1,$CF,$CE,$7F,$A0,$FB,$7F,$C0
	.dc.b	$37,$BF,$9D,$5E,$FE,$D3,$FA,$CE,$5D,$32,$FF,$71,$3F,$60,$FE,$52
	.dc.b	$FF,$95,$9C,$5F,$3A,$E5,$F0,$5F,$EC,$18,$C6,$7F,$E4,$AF,$96,$C6
	.dc.b	$1F,$E5,$EF,$D7,$1F,$AB,$7F,$95,$B1,$7E,$7D,$E1,$D5,$5A,$FB,$8F
	.dc.b	$3F,$D1,$E6,$30,$F4,$9D,$E1,$FE,$60,$F5,$31,$F9,$07,$C8,$F5,$9E
	.dc.b	$4B,$EB,$FB,$4E,$63,$F4,$1C,$BF,$1B,$7F,$F3,$9C,$F1,$FC,$87,$9F
	.dc.b	$47,$F0,$7F,$C4,$5C,$FF,$EA,$5F,$CA,$5F,$18,$FD,$05,$FA,$2B,$F0
	.dc.b	$8F,$F2,$FF,$D3,$7F,$9B,$76,$FF,$07,$63,$CB,$1E,$38,$70,$BD,$FB
	.dc.b	$FD,$72,$E9,$7F,$EA,$91,$7E,$0F,$2B,$FD,$5F,$18,$9F,$16,$7E,$1F
	.dc.b	$B1,$FF,$56,$7D,$5F,$A2,$B5,$9F,$79,$FF,$09,$C7,$78,$FC,$67,$90
	.dc.b	$FC,$B5,$E0,$3E,$5F,$5C,$DF,$99,$9C,$FE,$81,$8B,$E1,$7F,$1B,$8B
	.dc.b	$DF,$97,$83,$0E,$1F,$4D,$9D,$7C,$9C,$3F,$0B,$EB,$DF,$BD,$7F,$A4
	.dc.b	$FE,$75,$F9,$17,$96,$73,$F0,$BF,$DB,$BC,$DC,$5E,$F4,$B2,$FA,$0F
	.dc.b	$17,$57,$C1,$E3,$7E,$5C,$FD,$C3,$FC,$33,$F5,$BF,$94,$FA,$87,$EF
	.dc.b	$BF,$E2,$BC,$38,$5E,$FE,$B3,$FC,$C1,$E5,$7F,$B8,$7F,$01,$D6,$DF
	.dc.b	$84,$D3,$AE,$9F,$EF,$87,$BF,$EB,$17,$1E,$5A,$FE,$8B,$EB,$7C,$B8
	.dc.b	$71,$9C,$E5,$FE,$56,$EA,$FC,$54,$E7,$27,$E4,$4D,$1F,$AC,$7F,$44
	.dc.b	$F1,$BC,$77,$F6,$72,$F7,$5E,$EB,$FA,$4F,$EF,$3A,$DF,$CF,$39,$F9
	.dc.b	$4F,$1E,$1C,$73,$EF,$9F,$6F,$57,$38,$7B,$BE,$7D,$C0,$FC,$A1,$F8
	.dc.b	$FF,$FB,$C7,$87,$D3,$9F,$3B,$DD,$6B,$E6,$BC,$5C,$3F,$4E,$73,$E1
	.dc.b	$F6,$8C,$65,$D7,$F3,$7F,$C4,$63,$FB,$2F,$C4,$3F,$DF,$1E,$3B,$F2
	.dc.b	$D7,$C2,$3E,$CF,$FD,$2D,$FB,$0F,$F9,$97,$CA,$FD,$16,$1C,$2F,$75
	.dc.b	$FE,$9B,$35,$8F,$5D,$DF,$B1,$7E,$BA,$FB,$CF,$EC,$DE,$59,$77,$F7
	.dc.b	$7C,$B3,$FE,$F9,$BF,$D4,$E3,$DB,$DF,$CE,$1E,$8F,$8D,$FF,$22,$7E
	.dc.b	$95,$FE,$20,$FD,$83,$C3,$8F,$1E,$0F,$E9,$DF,$62,$AF,$F1,$2F,$D7
	.dc.b	$6B,$1E,$53,$9F,$9F,$EB,$AF,$EE,$33,$EB,$E9,$63,$2F,$F4,$AE,$FF
	.dc.b	$E1,$EF,$90,$7C,$37,$3C,$3F,$7F,$F8,$12,$B5,$F4,$39,$C3,$8F,$D5
	.dc.b	$7C,$EE,$1C,$5E,$E3,$FA,$B7,$81,$F8,$0F,$B9,$70,$BF,$07,$FF,$7C
	.dc.b	$F9,$CF,$85,$F0,$FE,$6F,$F9,$8B,$DD,$6A,$7C,$BF,$51,$9C,$39,$69
	.dc.b	$FE,$CC,$31,$F0,$35,$D5,$F1,$E4,$FC,$DE,$3D,$3F,$0E,$FF,$6C,$65
	.dc.b	$FE,$51,$D7,$CD,$97,$97,$F6,$17,$E9,$6F,$DA,$DF,$B6,$7F,$BF,$AF
	.dc.b	$7A,$2F,$75,$AF,$47,$C3,$18,$FA,$7F,$0E,$5D,$72,$E9,$F4,$B8,$C7
	.dc.b	$FC,$77,$FA,$D3,$CA,$7E,$B9,$F1,$75,$EF,$EB,$2B,$52,$F9,$FF,$7C
	.dc.b	$E6,$DE,$8F,$58,$C7,$9F,$FC,$CF,$E7,$FF,$A1,$7E,$EA,$F7,$DF,$7E
	.dc.b	$FD,$57,$EA,$7E,$EA,$FC,$ED,$F5,$7F,$EC,$DC,$78,$E1,$C2,$F7,$5F
	.dc.b	$EC,$7E,$15,$6B,$E3,$3E,$A4,$63,$C0,$FE,$D3,$A7,$87,$4C,$DF,$47
	.dc.b	$C5,$DA,$D7,$F1,$D8,$71,$C1,$87,$1C,$FC,$F7,$ED,$9E,$BA,$7B,$3B
	.dc.b	$FD,$3F,$C0,$2F,$DF,$BF,$C1,$B9,$73,$CB,$C5,$5A,$97,$CE,$FB,$F1
	.dc.b	$E7,$AF,$FB,$27,$85,$93,$EF,$BC,$15,$E4,$D3,$C1,$5A,$FB,$8C,$E9
	.dc.b	$EF,$F2,$BE,$67,$E4,$0F,$C9,$1F,$A1,$77,$CB,$96,$BF,$E9,$2E,$75
	.dc.b	$EF,$CF,$D6,$5C,$E7,$3F,$03,$DB,$6F,$B7,$A7,$6F,$87,$FF,$0B,$F8
	.dc.b	$4F,$CC,$3F,$C0,$5F,$83,$FF,$4A,$F2,$9F,$2E,$FB,$A7,$EC,$9F,$E9
	.dc.b	$2F,$53,$F7,$18,$70,$F8,$EF,$ED,$17,$BA,$FF,$4B,$FB,$DA,$E5,$FF
	.dc.b	$12,$E7,$D5,$FD,$B6,$9E,$17,$38,$7E,$45,$CE,$3D,$74,$F1,$E9,$E2
	.dc.b	$F9,$8F,$4C,$E7,$4F,$06,$78,$C3,$9E,$1F,$B3,$FD,$BD,$F9,$63,$F2
	.dc.b	$6F,$F0,$16,$BE,$1B,$6D,$F0,$7D,$46,$1C,$2F,$FC,$E9,$7B,$AD,$4F
	.dc.b	$4F,$53,$C1,$87,$2E,$0F,$F0,$56,$3C,$B2,$FE,$49,$8F,$19,$FF,$07
	.dc.b	$7E,$38,$FD,$0D,$F8,$CB,$F1,$D7,$E3,$9F,$27,$3E,$9C,$0C,$38,$5E
	.dc.b	$FB,$75,$C1,$F6,$67,$3E,$8F,$92,$31,$9F,$6E,$AF,$C0,$FC,$8F,$F9
	.dc.b	$07,$F6,$67,$E8,$7F,$C7,$7E,$CD,$F9,$23,$F1,$5F,$F7,$CF,$E1,$FF
	.dc.b	$CD,$F0,$E1,$7B,$AD,$7C,$83,$F7,$ED,$71,$FF,$99,$33,$EB,$FA,$77
	.dc.b	$F3,$3E,$8F,$F2,$0C,$78,$3E,$13,$19,$F4,$7F,$DE,$59,$78,$7F,$04
	.dc.b	$F3,$E3,$CF,$4F,$06,$DF,$AF,$8E,$E3,$1E,$9F,$AE,$FF,$B4,$EB,$FC
	.dc.b	$B1,$FA,$76,$A7,$3F,$86,$7C,$0F,$A2,$7E,$03,$9B,$FE,$7D,$87,$1F
	.dc.b	$93,$F0,$E5,$D6,$73,$87,$C8,$FC,$43,$39,$7A,$FD,$77,$E3,$7F,$ED
	.dc.b	$3F,$BF,$FF,$6C,$72,$7C,$0C,$7F,$A5,$56,$BE,$33,$FD,$25,$F1,$FE
	.dc.b	$1E,$54,$E7,$B7,$AD,$F0,$CB,$FD,$BD,$E7,$F9,$3E,$E3,$F8,$47,$EB
	.dc.b	$3F,$CF,$FC,$17,$E0,$BF,$EA,$8D,$BF,$CD,$F5,$AF,$90,$FE,$99,$5A
	.dc.b	$FC,$13,$E4,$CF,$2E,$5F,$DA,$F7,$7F,$A1,$63,$F8,$A7,$E7,$FF,$5A
	.dc.b	$7F,$D3,$F1,$CF,$E4,$3F,$53,$9F,$8F,$E0,$9A,$F8,$FD,$FC,$2F,$C3
	.dc.b	$DF,$34,$FC,$A5,$C4,$FC,$7F,$C7,$8F,$2B,$DD,$6B,$F1,$DC,$E3,$CF
	.dc.b	$1E,$58,$F6,$BD,$FB,$7E,$0F,$D4,$DF,$A5,$3E,$A1,$FA,$C7,$F2,$EE
	.dc.b	$9F,$8B,$B7,$A2,$F7,$5A,$E0,$EF,$18,$FC,$3F,$A6,$3C,$FA,$7F,$95
	.dc.b	$FC,$1F,$FA,$AF,$F3,$DF,$E1,$7F,$8E,$3F,$74,$FF,$4C,$FD,$DB,$F2
	.dc.b	$0F,$E9,$7B,$FF,$2D,$EC,$FF,$64,$E1,$C2,$F7,$5F,$67,$1E,$99,$3E
	.dc.b	$2F,$F5,$51,$8C,$EE,$3C,$A9,$1E,$1F,$C0,$BF,$DD,$BC,$DE,$6A,$D4
	.dc.b	$E7,$5F,$E2,$51,$FE,$33,$C3,$A7,$EA,$1F,$C1,$3F,$32,$7E,$F6,$F0
	.dc.b	$F5,$7C,$AB,$53,$FE,$DD,$E1,$F6,$4B,$87,$E1,$D7,$3F,$A9,$F1,$4E
	.dc.b	$77,$F3,$CF,$7E,$AC,$DF,$CA,$FE,$00,$7E,$7A,$F6,$3F,$C9,$3F,$2B
	.dc.b	$E2,$CB,$97,$12,$F7,$5F,$E4,$EF,$33,$58,$F3,$C7,$97,$E0,$FD,$B5
	.dc.b	$F8,$F8,$70,$FC,$47,$FD,$F7,$F5,$DF,$EE,$5D,$1F,$96,$3F,$42,$7E
	.dc.b	$6A,$FB,$B7,$B9,$FD,$A3,$F7,$DF,$E9,$F8,$F1,$C3,$85,$EE,$B5,$FA
	.dc.b	$87,$FB,$A7,$2F,$F7,$95,$75,$F8,$DF,$F2,$C6,$1C,$73,$EF,$5F,$E5
	.dc.b	$D8,$C6,$3C,$FE,$65,$F8,$AF,$A2,$B9,$DF,$9B,$5F,$2D,$7F,$5E,$9F
	.dc.b	$00,$3D,$B3,$EC,$F8,$D6,$CF,$D5,$FF,$C9,$99,$70,$BD,$AA,$DC,$FF
	.dc.b	$94,$7E,$DB,$5B,$7C,$09,$CF,$D6,$79,$70,$F0,$4E,$71,$ED,$F8,$CE
	.dc.b	$67,$AF,$E9,$EF,$C9,$7E,$77,$E3,$5F,$E5,$3E,$2F,$83,$39,$E1,$4B
	.dc.b	$F6,$1F,$0F,$0F,$D6,$AD,$7D,$67,$5D,$7E,$9C,$71,$F9,$1F,$BF,$FB
	.dc.b	$4F,$C3,$D7,$FD,$3D,$F3,$0F,$C8,$1F,$9A,$3F,$52,$7E,$CA,$FD,$29
	.dc.b	$FB,$DB,$8D,$9F,$3F,$95,$67,$CF,$2E,$58,$F1,$FC,$D3,$85,$EE,$9F
	.dc.b	$9E,$7E,$0A,$31,$AD,$DB,$EC,$65,$D2,$3C,$BB,$79,$3E,$5F,$D7,$5A
	.dc.b	$C7,$C1,$3E,$8F,$A5,$CC,$FB,$AD,$FA,$6B,$EC,$2F,$57,$FB,$EB,$87
	.dc.b	$D7,$7C,$EF,$75,$AF,$AD,$F4,$CE,$70,$E4,$7E,$95,$C9,$DD,$CF,$CB
	.dc.b	$78,$B6,$F5,$7C,$E7,$DA,$FB,$57,$E6,$9F,$CE,$34,$F0,$E3,$E7,$AD
	.dc.b	$4B,$F6,$3E,$B8,$72,$F0,$3F,$CD,$3D,$7D,$78,$FE,$D6,$1D,$7B,$BF
	.dc.b	$88,$FC,$BB,$EB,$5F,$80,$BF,$57,$7E,$9C,$CB,$F0,$3F,$8E,$7F,$0F
	.dc.b	$C3,$85,$EE,$B5,$E0,$3F,$7C,$D6,$5D,$38,$BF,$CB,$FF,$0D,$F8,$DC
	.dc.b	$7C,$1F,$92,$7F,$38,$94,$74,$D7,$C3,$AF,$F9,$0A,$79,$E9,$DE,$1E
	.dc.b	$BF,$EE,$7A,$3F,$8B,$7D,$47,$E4,$79,$DF,$B5,$FE,$43,$ED,$1F,$AB
	.dc.b	$FF,$BB,$EF,$7D,$7F,$3A,$B5,$3F,$99,$9C,$FD,$87,$FB,$A5,$DE,$17
	.dc.b	$5F,$D3,$65,$DF,$0E,$DA,$78,$E7,$39,$75,$E9,$FF,$2E,$7E,$2B,$FE
	.dc.b	$2A,$FC,$F1,$F8,$3F,$F8,$07,$E5,$9F,$9D,$FE,$AF,$D1,$FE,$95,$F2
	.dc.b	$1F,$B3,$BD,$FD,$17,$F7,$AA,$D7,$DF,$7F,$90,$27,$3A,$FF,$86,$0F
	.dc.b	$E3,$8F,$6F,$6F,$FC,$31,$08,$E7,$97,$83,$D5,$FF,$17,$CF,$3D,$3C
	.dc.b	$98,$FE,$F7,$C3,$FE,$87,$F8,$FF,$F0,$5F,$4D,$F9,$57,$F4,$0F,$D1
	.dc.b	$3F,$82,$7A,$3F,$8B,$F2,$7F,$81,$5E,$EB,$5E,$9F,$F4,$7F,$1B,$FB
	.dc.b	$27,$42,$B5,$F0,$FF,$D0,$1E,$1F,$A4,$DE,$0F,$E6,$DA,$F8,$29,$AF
	.dc.b	$82,$7C,$A6,$B3,$EF,$AF,$F9,$0A,$3E,$7F,$E5,$3F,$C7,$7F,$0B,$7D
	.dc.b	$7B,$B1,$F4,$2F,$DE,$DF,$7D,$FE,$F8,$FD,$DF,$8E,$5F,$D1,$F1,$E3
	.dc.b	$87,$0B,$DF,$E2,$5F,$ED,$8E,$96,$9D,$F4,$7E,$43,$FF,$00,$23,$18
	.dc.b	$5C,$BA,$57,$8F,$5F,$16,$BF,$D7,$27,$C2,$7F,$BC,$A7,$59,$BF,$4C
	.dc.b	$FE,$71,$9F,$F8,$56,$BF,$AE,$7D,$3F,$F8,$8B,$F4,$E7,$D8,$B9,$DF
	.dc.b	$BE,$3F,$04,$FF,$06,$E3,$C7,$1F,$F1,$C7,$CF,$B1,$E3,$87,$0B,$FE
	.dc.b	$0B,$7B,$CB,$FC,$5F,$FA,$EF,$F8,$1F,$1F,$ED,$93,$9C,$79,$46,$63
	.dc.b	$18,$F4,$C9,$F0,$FC,$3C,$AF,$EA,$59,$7A,$75,$F4,$53,$E0,$3F,$80
	.dc.b	$77,$FF,$57,$FD,$CB,$FB,$BF,$EB,$3F,$C8,$FF,$3D,$FF,$2F,$FE,$65
	.dc.b	$FE,$BA,$F5,$BF,$1D,$C1,$FD,$FE,$F7,$F0,$AB,$DF,$E6,$9F,$E9,$0C
	.dc.b	$BA,$7C,$B3,$3F,$E0,$51,$8F,$76,$E5,$CE,$B5,$8F,$3F,$43,$F2,$1A
	.dc.b	$78,$29,$A7,$83,$4F,$F3,$9C,$7F,$84,$7E,$AB,$F0,$EF,$F8,$4F,$F0
	.dc.b	$77,$EF,$DF,$DF,$3F,$16,$75,$BF,$30,$7E,$1D,$FE,$E6,$C7,$8E,$3F
	.dc.b	$ED,$7C,$38,$5E,$EB,$FB,$CD,$6B,$1E,$BE,$F7,$FB,$4F,$C7,$FF,$00
	.dc.b	$47,$83,$E3,$71,$97,$8F,$2E,$79,$FF,$33,$C7,$94,$63,$B7,$FC,$B7
	.dc.b	$F8,$E3,$F5,$07,$E7,$CF,$D4,$FF,$A7,$70,$3F,$46,$7D,$9F,$F9,$67
	.dc.b	$0E,$17,$BF,$67,$DF,$4F,$D7,$B5,$97,$F4,$6D,$FF,$A5,$9F,$E3,$9C
	.dc.b	$38,$B8,$0F,$4C,$FC,$3F,$8C,$F9,$AB,$CB,$1E,$4F,$C3,$FF,$A5,$7C
	.dc.b	$EF,$F8,$87,$E5,$5F,$A0,$7F,$87,$3F,$3C,$7E,$22,$FC,$09,$F7,$AF
	.dc.b	$DD,$FF,$DA,$F8,$F1,$C3,$85,$EE,$BF,$3B,$E1,$7F,$7A,$CB,$AF,$E4
	.dc.b	$9F,$D3,$70,$E1,$F4,$F7,$8B,$FC,$A9,$5E,$28,$6B,$E2,$D7,$F8,$BD
	.dc.b	$78,$0D,$E9,$E0,$87,$E3,$9F,$8D,$F8,$FF,$F1,$E7,$C9,$FF,$8D,$3F
	.dc.b	$7B,$7E,$6B,$FE,$B3,$FC,$9D,$F8,$F7,$F8,$C3,$C2,$61,$C3,$81,$87
	.dc.b	$0F,$C2,$58,$70,$FD,$49,$87,$0D,$FF,$CD,$9A,$77,$7F,$82,$46,$5F
	.dc.b	$F0,$66,$2F,$CC,$FF,$84,$47,$2C,$FC,$19,$F7,$C9,$F9,$5F,$AE,$BA
	.dc.b	$E1,$E4,$DB,$D5,$F0,$BF,$AB,$F2,$FF,$CB,$5F,$0E,$FD,$37,$FD,$DD
	.dc.b	$E0,$3F,$59,$FF,$71,$60,$FC,$D3,$D3,$87,$0B,$DD,$6B,$F0,$1C,$E9
	.dc.b	$FE,$87,$CB,$A7,$CE,$7E,$9A,$3F,$A5,$5F,$FB,$ED,$F7,$E1,$CF,$E7
	.dc.b	$7F,$49,$8F,$FB,$C7,$E5,$FC,$D5,$CE,$FB,$FE,$22,$F1,$BF,$D8,$BF
	.dc.b	$86,$7F,$9B,$FF,$0E,$7E,$56,$FD,$D9,$F9,$07,$D4,$FB,$AF,$F6,$15
	.dc.b	$EE,$B5,$3F,$63,$E0,$73,$F3,$EC,$7F,$C6,$17,$FE,$7B,$FD,$06,$CE
	.dc.b	$70,$E3,$F8,$87,$8A,$31,$9F,$83,$0F,$F1,$87,$AD,$FE,$87,$DD,$F9
	.dc.b	$CF,$F9,$1F,$F5,$B7,$EF,$AF,$9E,$FE,$C3,$FB,$63,$D3,$FF,$28,$73
	.dc.b	$7F,$AC,$71,$E3,$E4,$B1,$E3,$87,$F6,$8F,$C1,$38,$69,$DF,$4F,$F1
	.dc.b	$DE,$3F,$D3,$F2,$E6,$FD,$F8,$FF,$2C,$39,$9E,$18,$F5,$D3,$C9,$DD
	.dc.b	$EA,$AF,$15,$FA,$67,$D1,$E7,$7F,$35,$FD,$83,$FD,$49,$F9,$43,$F2
	.dc.b	$E7,$EB,$CF,$DC,$DF,$A9,$FF,$94,$3F,$41,$FD,$5F,$A4,$C7,$8E,$1C
	.dc.b	$2F,$7F,$18,$B7,$E3,$DF,$AA,$BD,$FA,$1F,$A4,$C7,$97,$07,$FB,$5F
	.dc.b	$5F,$16,$B9,$BE,$AF,$8F,$5F,$F3,$0F,$E4,$FE,$9B,$E1,$B7,$86,$7F
	.dc.b	$CC,$B1,$FA,$B1,$87,$FA,$C3,$E7,$1F,$9C,$3F,$35,$FF,$80,$3F,$98
	.dc.b	$7F,$4A,$F6,$3E,$1F,$D9,$7F,$83,$2B,$5F,$87,$2B,$FA,$A5,$6B,$0F
	.dc.b	$53,$C3,$96,$E7,$EB,$C7,$96,$4C,$5F,$8D,$F9,$74,$F2,$69,$DB,$F2
	.dc.b	$1F,$D4,$E2,$5F,$BD,$7F,$20,$FC,$3F,$FD,$33,$FA,$1F,$FB,$F7,$E5
	.dc.b	$FF,$73,$7B,$AD,$7E,$16,$A4,$F8,$3A,$3F,$B5,$E3,$D3,$D6,$FD,$F3
	.dc.b	$DB,$BA,$E3,$CB,$0F,$F7,$45,$F0,$8C,$74,$7F,$7D,$FA,$9F,$F7,$67
	.dc.b	$E8,$0F,$DF,$5F,$90,$7F,$49,$FD,$A9,$FB,$B7,$FB,$FF,$F6,$0F,$E5
	.dc.b	$78,$70,$BD,$FC,$AA,$F7,$E6,$BE,$B3,$7C,$7E,$AD,$3B,$7E,$77,$FE
	.dc.b	$80,$7C,$1A,$DC,$B9,$E3,$E0,$FC,$DF,$CD,$87,$19,$CE,$DF,$C1,$E3
	.dc.b	$F9,$EF,$E2,$5F,$19,$FF,$2D,$FE,$31,$FE,$C5,$FC,$69,$FA,$CB,$F6
	.dc.b	$FF,$EA,$FC,$BC,$2F,$E6,$FE,$01,$FE,$ED,$F5,$18,$70,$BF,$EB,$BF
	.dc.b	$8B,$6E,$FA,$FE,$25,$FE,$C5,$C3,$84,$7F,$7D,$79,$5F,$1F,$A9,$FF
	.dc.b	$15,$6B,$E5,$9C,$E9,$E3,$FC,$73,$F0,$F4,$F8,$2F,$D3,$7F,$00,$3F
	.dc.b	$75,$7E,$EF,$FE,$93,$F0,$3F,$C5,$77,$BF,$3B,$F7,$B7,$BA,$7C,$BF
	.dc.b	$E3,$2B,$5B,$FF,$40,$3F,$16,$F7,$FC,$AB,$FA,$33,$CF,$D4,$61,$E1
	.dc.b	$F1,$7C,$BA,$F9,$35,$7D,$47,$A3,$EB,$9F,$AC,$FE,$27,$C1,$FF,$20
	.dc.b	$FE,$7E,$FC,$2F,$C5,$5A,$F0,$15,$AF,$DD,$B3,$87,$8A,$BD,$FC,$3E
	.dc.b	$73,$FC,$E2,$F1,$AF,$F2,$6F,$E8,$BE,$1A,$F0,$7D,$BB,$F1,$8F,$9F
	.dc.b	$87,$EA,$3F,$64,$7E,$93,$FD,$BF,$F2,$5F,$F2,$B7,$E4,$4F,$9E,$FF
	.dc.b	$56,$76,$BC,$1F,$C5,$61,$C3,$E4,$58,$70,$B7,$E1,$7F,$83,$AE,$BF
	.dc.b	$81,$7F,$23,$C3,$8E,$7F,$DF,$74,$FF,$20,$57,$3C,$7C,$3F,$45,$E6
	.dc.b	$94,$E7,$DC,$6D,$FC,$06,$7B,$7E,$89,$FE,$3B,$FA,$EF,$F0,$EF,$CC
	.dc.b	$3F,$4D,$FF,$2D,$71,$7A,$6F,$E0,$5E,$FE,$AF,$FB,$65,$7D,$0A,$D5
	.dc.b	$78,$2B,$FC,$8C,$F9,$97,$E8,$DE,$B8,$E7,$ED,$FF,$8C,$B0,$FE,$ED
	.dc.b	$E1,$7A,$76,$61,$DF,$3F,$E4,$18,$7C,$5D,$FF,$DB,$7F,$A1,$FF,$B6
	.dc.b	$FC,$97,$7B,$F4,$AD,$7E,$2A,$AD,$4F,$F8,$3F,$83,$CF,$A3,$17,$EB
	.dc.b	$DE,$5C,$FE,$4C,$7F,$BA,$1F,$9E,$F8,$3E,$6B,$F4,$7F,$9F,$79,$E3
	.dc.b	$1A,$FF,$BC,$7E,$AF,$FD,$45,$C2,$FD,$75,$FB,$07,$FB,$C7,$F0,$87
	.dc.b	$D7,$3E,$F1,$FB,$F3,$1E,$38,$70,$D5,$87,$0B,$FC,$75,$6B,$F7,$8F
	.dc.b	$F2,$5B,$C3,$2B,$F8,$87,$F1,$9D,$3C,$51,$E1,$D1,$F9,$1F,$F4,$2B
	.dc.b	$E1,$9F,$69,$FF,$0C,$70,$F8,$F1,$FE,$45,$F0,$8F,$DE,$FF,$90,$FE
	.dc.b	$16,$FC,$7F,$FC,$C7,$F7,$7F,$E0,$5C,$FF,$D8,$57,$BF,$81,$FF,$50
	.dc.b	$D6,$BE,$7F,$A6,$7F,$D4,$AB,$A7,$E8,$9E,$88,$C6,$7F,$CB,$9F,$F7
	.dc.b	$76,$3D,$23,$C3,$87,$F9,$CB,$87,$D5,$AF,$9A,$73,$D9,$F8,$8B,$F1
	.dc.b	$E2,$FC,$19,$ED,$7F,$A7,$3C,$B7,$E9,$BF,$EB,$6B,$DD,$6B,$F0,$8F
	.dc.b	$F6,$94,$BE,$27,$D3,$BB,$81,$9F,$58,$F3,$ED,$F2,$23,$FC,$7D,$18
	.dc.b	$3E,$03,$E3,$D3,$C5,$F1,$FE,$A9,$CE,$7E,$1D,$7F,$AA,$7D,$7B,$E5
	.dc.b	$FF,$AA,$7E,$7F,$FE,$E5,$F9,$37,$DE,$BF,$78,$FF,$5C,$7E,$B8,$F0
	.dc.b	$7F,$CA,$58,$70,$CB,$FB,$65,$EF,$F0,$85,$FE,$5B,$3E,$DA,$FF,$81
	.dc.b	$B4,$FF,$79,$7E,$03,$FE,$98,$9C,$CA,$7C,$4F,$F9,$57,$2E,$DA,$F9
	.dc.b	$35,$FF,$65,$7E,$49,$FE,$10,$C3,$8E,$1E,$9D,$3C,$5F,$41,$9C,$7F
	.dc.b	$B8,$7E,$BD,$FA,$4E,$D7,$EE,$5F,$D4,$BB,$FF,$DB,$FF,$82,$FB,$37
	.dc.b	$FF,$0D,$DE,$EB,$5F,$88,$B8,$75,$4A,$E8,$F9,$CE,$3E,$6F,$DB,$70
	.dc.b	$6D,$FE,$D5,$7E,$1F,$AB,$F3,$4F,$0C,$FC,$BE,$7F,$E8,$FE,$9B,$32
	.dc.b	$FD,$5F,$FC,$9F,$F9,$D3,$F4,$87,$CC,$BF,$49,$FE,$97,$FC,$03,$E6
	.dc.b	$BD,$D6,$A7,$FA,$7F,$C7,$67,$EC,$18,$FD,$3E,$6F,$F0,$B6,$1F,$E5
	.dc.b	$CF,$92,$C6,$9F,$E3,$E8,$E7,$A7,$7F,$8B,$F2,$C6,$33,$FF,$67,$D7
	.dc.b	$F0,$03,$FC,$8B,$F9,$23,$F6,$2F,$57,$96,$FD,$59,$FA,$BF,$E4,$3D
	.dc.b	$CE,$18,$30,$E1,$97,$E9,$6F,$7E,$07,$EC,$7C,$2F,$F1,$65,$EF,$37
	.dc.b	$E9,$BF,$8D,$8F,$F1,$86,$3C,$9F,$F5,$01,$89,$F1,$E1,$FC,$A2,$3F
	.dc.b	$18,$67,$1F,$06,$6F,$D2,$B3,$3E,$CC,$7F,$05,$E7,$F8,$3F,$4F,$7D
	.dc.b	$3B,$C3,$F5,$EB,$E4,$F0,$BF,$7B,$82,$F7,$F7,$9A,$8F,$C6,$65,$E6
	.dc.b	$3B,$78,$3E,$B7,$C1,$7C,$AD,$EF,$F9,$27,$E4,$5F,$90,$ED,$E4,$67
	.dc.b	$27,$E2,$2F,$D4,$5F,$C0,$3F,$D2,$BE,$95,$FA,$17,$FA,$67,$A3,$E6
	.dc.b	$E1,$C2,$F7,$F0,$9F,$E8,$1A,$7E,$D7,$EE,$CB,$A6,$BE,$93,$ED,$3F
	.dc.b	$D6,$1C,$61,$CF,$3F,$07,$B3,$E5,$AF,$24,$78,$FF,$49,$FF,$46,$FE
	.dc.b	$65,$FE,$04,$FD,$C1,$FB,$6B,$F5,$E7,$EF,$6F,$CA,$DE,$CB,$F8,$01
	.dc.b	$F8,$F3,$2E,$58,$F1,$C3,$85,$EF,$E6,$B5,$F4,$9F,$E7,$7D,$3B,$E9
	.dc.b	$F8,$EE,$2F,$1C,$E7,$1F,$F0,$5C,$7E,$4B,$2E,$B5,$E4,$D7,$C7,$A7
	.dc.b	$0A,$FF,$53,$E3,$D3,$4F,$0F,$D7,$E6,$52,$FC,$93,$F9,$27,$EE,$4F
	.dc.b	$D6,$1F,$AB,$3F,$71,$7E,$8D,$E8,$C7,$AE,$1C,$2F,$74,$FC,$D3,$55
	.dc.b	$DA,$BF,$98,$3E,$DE,$0F,$9D,$0F,$C2,$FE,$04,$F0,$D3,$C9,$F0,$CD
	.dc.b	$7C,$9A,$FF,$9C,$2B,$58,$78,$63,$FC,$8D,$8F,$D9,$F9,$7F,$F4,$CF
	.dc.b	$23,$F5,$17,$E6,$FF,$E4,$EF,$A4,$ED,$5A,$9C,$FD,$27,$F9,$DF,$0E
	.dc.b	$5E,$C7,$91,$E9,$F8,$0D,$EB,$FB,$18,$71,$CB,$FD,$8F,$B7,$4A,$FC
	.dc.b	$4E,$5E,$0E,$DF,$A1,$D8,$FC,$BD,$FA,$13,$F1,$FF,$F4,$4F,$E9,$DF
	.dc.b	$EC,$9F,$93,$F9,$37,$BA,$D7,$DD,$2B,$FD,$9D,$8E,$B1,$E7,$A7,$CD
	.dc.b	$DF,$FD,$85,$8F,$7F,$C4,$BD,$13,$FE,$EA,$8E,$9A,$78,$3D,$3F,$2E
	.dc.b	$1E,$48,$F1,$FE,$13,$E8,$FD,$07,$FC,$25,$F8,$5B,$F5,$47,$DC,$3F
	.dc.b	$80,$5F,$3B,$E8,$BD,$D6,$A7,$FD,$3F,$D3,$FE,$70,$D5,$5F,$E1,$1D
	.dc.b	$BF,$39,$C1,$E8,$8F,$F0,$8C,$60,$F4,$3D,$32,$79,$DE,$5F,$9F,$6B
	.dc.b	$E5,$97,$A5,$F4,$3D,$17,$D4,$3F,$53,$7E,$D4,$FD,$95,$F8,$E7,$F9
	.dc.b	$D3,$87,$CD,$87,$0B,$DD,$6B,$F0,$8F,$AB,$3D,$38,$3E,$81,$F3,$57
	.dc.b	$E1,$DF,$42,$3F,$85,$47,$3C,$BC,$1D,$FE,$5D,$3C,$91,$E3,$F7,$FF
	.dc.b	$89,$FE,$2F,$FE,$FB,$FC,$4B,$F0,$1E,$33,$F1,$27,$EF,$7F,$DB,$3F
	.dc.b	$7C,$7E,$AD,$FC,$47,$0E,$1E,$37,$E8,$2F,$7F,$38,$B7,$EA,$FB,$E1
	.dc.b	$0D,$3B,$BF,$B7,$C7,$9F,$E6,$7F,$A3,$7A,$EB,$E3,$D7,$FB,$EF,$93
	.dc.b	$E8,$BE,$78,$7A,$34,$F0,$FE,$33,$99,$7E,$1F,$F5,$3F,$4F,$FA,$7F
	.dc.b	$A3,$7F,$77,$FC,$DB,$53,$9E,$54,$E7,$ED,$7D,$18,$72,$E7,$63,$E5
	.dc.b	$8F,$5E,$1F,$80,$F1,$71,$1B,$9E,$FF,$38,$FA,$3E,$F8,$72,$CF,$BC
	.dc.b	$FF,$A7,$6F,$BE,$2F,$D2,$1F,$92,$3F,$74,$FE,$AB,$F5,$CF,$DB,$3F
	.dc.b	$A4,$F8,$CE,$9F,$9B,$97,$C0,$C7,$8E,$1C,$2F,$F8,$B5,$EE,$BD,$34
	.dc.b	$FC,$CB,$D3,$9F,$87,$3B,$3C,$BE,$4D,$A7,$FB,$FB,$0F,$3C,$E6,$3F
	.dc.b	$C4,$F8,$FF,$9C,$27,$EF,$FC,$5B,$F3,$47,$E1,$FF,$E8,$5F,$CB,$BF
	.dc.b	$CB,$3E,$B7,$87,$0E,$18,$7F,$42,$BD,$FD,$4A,$F7,$FA,$2F,$F6,$1F
	.dc.b	$AA,$F7,$E7,$D7,$F1,$6F,$E5,$51,$8C,$3B,$E3,$FD,$0E,$31,$A7,$87
	.dc.b	$ED,$9F,$65,$FD,$D2,$73,$A7,$79,$7E,$79,$9C,$3D,$7F,$97,$3F,$65
	.dc.b	$FE,$BB,$F4,$2F,$D5,$7C,$DF,$31,$E5,$BD,$D6,$A7,$D5,$F9,$BE,$67
	.dc.b	$FA,$FE,$0F,$DE,$7D,$31,$C0,$F4,$1C,$BE,$63,$FD,$CB,$7F,$3B,$6F
	.dc.b	$83,$7E,$39,$CE,$9E,$7B,$FE,$CD,$F8,$43,$F3,$A7,$E7,$8F,$DE,$FE
	.dc.b	$67,$4F,$F8,$9B,$AB,$E8,$FC,$F3,$0E,$17,$BA,$FA,$5F,$AF,$FA,$6B
	.dc.b	$58,$FA,$32,$EF,$F2,$D6,$73,$87,$1C,$3C,$91,$8C,$FE,$E6,$9F,$E2
	.dc.b	$4F,$CE,$3F,$33,$F8,$8F,$FB,$F3,$F2,$47,$ED,$4F,$92,$7E,$99,$FC
	.dc.b	$27,$F1,$BF,$F1,$16,$1C,$3F,$19,$61,$C2,$F7,$ED,$BF,$6C,$DF,$17
	.dc.b	$F1,$2D,$3B,$3F,$E0,$7C,$79,$7E,$39,$FD,$CD,$E9,$8F,$3D,$1F,$98
	.dc.b	$F9,$EF,$96,$9E,$0D,$BC,$D4,$FC,$C3,$54,$FC,$99,$FA,$77,$F8,$FA
	.dc.b	$7F,$D3,$DF,$9E,$7F,$96,$7D,$BF,$2D,$EE,$B5,$F8,$3A,$BF,$CB,$DF
	.dc.b	$A1,$6B,$1E,$7F,$B5,$F9,$F0,$FE,$5B,$1D,$BE,$CF,$11,$CD,$F2,$61
	.dc.b	$CF,$D5,$F2,$D6,$F4,$F1,$FC,$8F,$A3,$CE,$7E,$77,$FE,$50,$F9,$07
	.dc.b	$E2,$4F,$C5,$5F,$92,$7F,$5D,$F2,$5D,$5F,$C4,$B1,$E3,$F7,$1F,$E0
	.dc.b	$BC,$38,$5F,$F9,$7E,$F7,$C3,$FD,$12,$F6,$FD,$33,$19,$BE,$EB,$D7
	.dc.b	$3D,$34,$F1,$E9,$FE,$24,$C3,$9E,$0A,$D6,$5E,$7F,$8F,$FE,$A3,$F9
	.dc.b	$07,$FA,$86,$BF,$A9,$FE,$D3,$FE,$0C,$FD,$05,$E8,$FF,$A0,$BF,$16
	.dc.b	$7E,$56,$FC,$FD,$8F,$1C,$58,$F1,$C3,$85,$FF,$8C,$6B,$53,$D6,$BF
	.dc.b	$CE,$11,$8F,$AC,$7E,$79,$FE,$54,$8C,$67,$E2,$F8,$76,$9F,$D1,$67
	.dc.b	$39,$F9,$7C,$2F,$5F,$E2,$3F,$D0,$3D,$7F,$F1,$2F,$ED,$0F,$D8,$DF
	.dc.b	$A5,$FF,$A3,$7F,$5C,$FF,$47,$65,$E9,$BD,$D6,$BE,$C5,$5F,$E1,$7A
	.dc.b	$D7,$03,$F5,$EF,$36,$7D,$E1,$9E,$38,$BF,$BE,$E7,$D3,$4F,$F4,$E7
	.dc.b	$B3,$E6,$9C,$DF,$0D,$7C,$1F,$8D,$66,$7F,$96,$4F,$EF,$FE,$17,$FD
	.dc.b	$3F,$AB,$EC,$FF,$BA,$BF,$47,$E1,$BF,$C8,$7E,$CF,$FA,$27,$C5,$FA
	.dc.b	$B4,$FD,$57,$EB,$F0,$FF,$64,$BD,$E7,$F4,$A3,$19,$BF,$29,$F5,$4F
	.dc.b	$8B,$2E,$9F,$95,$FF,$6C,$D3,$C1,$7E,$5D,$3A,$DF,$E1,$3E,$1B,$FB
	.dc.b	$8F,$E3,$8F,$DE,$3F,$13,$FB,$53,$BD,$FB,$D3,$F7,$77,$ED,$3C,$72
	.dc.b	$FF,$12,$E3,$C7,$C5,$63,$C7,$0E,$1F,$85,$6E,$D7,$DF,$D6,$FC,$56
	.dc.b	$BE,$2D,$7F,$33,$5F,$90,$8C,$65,$D3,$5F,$1F,$E7,$1F,$A4,$9C,$E9
	.dc.b	$DF,$EE,$B1,$87,$93,$F2,$97,$E0,$CE,$57,$EB,$2F,$D5,$3F,$8F,$7E
	.dc.b	$69,$E5,$E1,$61,$C2,$F7,$5F,$C8,$FE,$2B,$53,$D3,$2F,$31,$DF,$D1
	.dc.b	$F5,$BE,$1C,$FC,$19,$F7,$FA,$1F,$2C,$FC,$2B,$ED,$3E,$3F,$CB,$33
	.dc.b	$A3,$F5,$EF,$C2,$7E,$EE,$FC,$DB,$F9,$B7,$D2,$7F,$97,$3C,$0F,$F0
	.dc.b	$8D,$EF,$C5,$5E,$FE,$09,$7F,$E4,$2B,$DE,$3F,$8F,$D7,$D2,$7D,$47
	.dc.b	$FC,$96,$E3,$1E,$98,$73,$F8,$2F,$2D,$79,$33,$FE,$91,$F0,$9F,$C5
	.dc.b	$BE,$AF,$FC,$77,$F9,$2B,$F4,$8F,$BF,$F9,$1B,$F6,$1F,$F6,$7F,$EC
	.dc.b	$FF,$C2,$B7,$FF,$1E,$6E,$C3,$86,$EC,$38,$7D,$E3,$0F,$CA,$61,$C2
	.dc.b	$FB,$7E,$4D,$FC,$4F,$4F,$0E,$9F,$8E,$CF,$F8,$CC,$63,$3E,$9A,$7F
	.dc.b	$B8,$3C,$DF,$45,$7F,$9F,$B1,$F0,$EB,$E0,$FC,$EF,$33,$E5,$FC,$23
	.dc.b	$FD,$47,$F1,$7F,$DE,$3F,$67,$7E,$ED,$FB,$FE,$0F,$25,$6B,$E6,$95
	.dc.b	$FC,$43,$0F,$E6,$7B,$FC,$6A,$7E,$B7,$E9,$9F,$14,$7F,$AC,$1F,$49
	.dc.b	$8F,$C3,$7F,$AE,$E1,$D3,$F1,$1F,$2E,$FF,$EB,$79,$CE,$4F,$9E,$CF
	.dc.b	$43,$F0,$2F,$F2,$07,$EC,$0F,$DF,$9F,$8F,$7F,$B0,$78,$7E,$65,$EF
	.dc.b	$EB,$97,$BA,$7D,$37,$AB,$83,$E4,$5B,$F5,$2F,$54,$7D,$B8,$FF,$80
	.dc.b	$2F,$F9,$F8,$E0,$67,$3F,$1E,$7D,$23,$3B,$FF,$A2,$E7,$3F,$AA,$7D
	.dc.b	$2F,$E0,$07,$E0,$DF,$E8,$9F,$C4,$BF,$D7,$BF,$8F,$3F,$6F,$7E,$37
	.dc.b	$DA,$F7,$D7,$FE,$D0,$FC,$4D,$7B,$F7,$75,$97,$F3,$7D,$BE,$78,$C3
	.dc.b	$8B,$87,$9E,$7E,$0F,$C5,$3C,$95,$E3,$C7,$AF,$99,$E7,$E8,$FE,$53
	.dc.b	$FB,$37,$F7,$3F,$D7,$FF,$C7,$FF,$8E,$F9,$67,$3F,$3E,$9F,$F2,$1F
	.dc.b	$E8,$79,$FC,$8B,$E5,$ED,$FD,$EB,$E2,$6E,$1C,$2F,$EF,$70,$73,$96
	.dc.b	$7D,$FF,$33,$F9,$1F,$88,$FE,$3B,$D4,$FD,$A7,$D6,$7C,$2F,$D0,$BE
	.dc.b	$34,$E7,$D3,$FF,$20,$74,$3F,$25,$CF,$9B,$E2,$A7,$ED,$D6,$3F,$8A
	.dc.b	$5F,$F9,$1F,$F6,$6E,$F1,$8C,$BF,$CA,$3B,$FF,$5E,$3F,$DD,$7F,$B6
	.dc.b	$FE,$57,$F7,$A7,$E3,$5F,$EA,$7F,$D2,$7A,$FD,$7E,$ED,$7B,$AD,$68
	.dc.b	$EF,$D4,$F3,$AF,$F0,$BF,$CE,$FF,$B3,$23,$F4,$5F,$AC,$FF,$85,$E3
	.dc.b	$1E,$8B,$4F,$0F,$D1,$35,$E9,$8F,$4A,$D6,$9E,$5F,$07,$FC,$67,$8F
	.dc.b	$5F,$CB,$BF,$D5,$5F,$87,$FF,$B1,$F2,$7D,$03,$F1,$FF,$2D,$6B,$F1
	.dc.b	$1F,$F6,$24,$BE,$77,$3F,$95,$F9,$76,$F8,$B7,$FE,$B5,$7D,$66,$3D
	.dc.b	$9F,$F6,$9E,$1C,$A3,$AC,$63,$F5,$FF,$F0,$06,$BF,$EE,$CF,$C4,$5F
	.dc.b	$95,$3F,$33,$FE,$99,$F9,$C3,$F3,$0F,$F4,$07,$E5,$9F,$3E,$F7,$5A
	.dc.b	$9F,$F4,$27,$E5,$99,$C3,$FA,$46,$9F,$BB,$F3,$3F,$50,$B0,$E2,$FA
	.dc.b	$0C,$4F,$87,$E4,$59,$F8,$7E,$53,$E5,$D0,$DE,$7E,$1F,$7F,$1F,$3D
	.dc.b	$FC,$27,$F4,$3E,$3F,$30,$FD,$09,$FB,$47,$F3,$FF,$7F,$FB,$F6,$5C
	.dc.b	$C5,$F0,$AD,$FC,$A7,$7F,$BF,$DD,$78,$6B,$AB,$FC,$5C,$C9,$CB,$67
	.dc.b	$F0,$06,$63,$33,$D7,$2F,$EA,$11,$8A,$67,$E7,$9F,$F1,$0F,$C2,$7B
	.dc.b	$2B,$CD,$6F,$D8,$7F,$B4,$63,$FE,$78,$FC,$97,$FC,$69,$F4,$5D,$A7
	.dc.b	$3C,$E9,$FF,$1C,$7D,$1E,$7F,$46,$F2,$71,$FC,$7B,$FE,$E5,$9F,$AE
	.dc.b	$31,$EA,$FA,$B1,$E5,$F6,$DE,$3A,$F8,$57,$DB,$27,$91,$99,$FF,$6D
	.dc.b	$FD,$BB,$F5,$FF,$E5,$7F,$77,$FA,$BF,$C0,$5E,$BF,$C8,$FC,$43,$FC
	.dc.b	$F9,$5A,$F5,$3F,$D0,$BE,$BE,$71,$7E,$F9,$E8,$CF,$EC,$61,$C1,$F4
	.dc.b	$18,$9F,$07,$BF,$FE,$7A,$C1,$F9,$17,$96,$FB,$CE,$7F,$03,$FC,$4F
	.dc.b	$85,$FE,$5E,$FC,$77,$FD,$8B,$F6,$4F,$82,$FD,$B7,$CF,$BF,$24,$FF
	.dc.b	$23,$5E,$FE,$8D,$7B,$F6,$BF,$DE,$FF,$8E,$6B,$87,$FC,$C3,$9F,$FB
	.dc.b	$41,$F9,$D8,$77,$FC,$57,$E9,$BC,$F0,$E5,$F0,$1E,$4A,$E3,$97,$6D
	.dc.b	$7F,$84,$7C,$5E,$65,$FA,$37,$E7,$BD,$8F,$F5,$DA,$F6,$DF,$9C,$3F
	.dc.b	$43,$F2,$5F,$0F,$37,$FC,$29,$7B,$A7,$E4,$5A,$8E,$95,$E8,$3F,$64
	.dc.b	$BF,$3E,$FC,$F4,$6E,$F9,$E3,$FC,$F3,$C9,$FE,$DF,$3C,$27,$3F,$1B
	.dc.b	$E9,$FD,$19,$F8,$A3,$F5,$E7,$EF,$AF,$D0,$DE,$03,$F4,$57,$E5,$9E
	.dc.b	$AC,$78,$E1,$C2,$F7,$5F,$EC,$0D,$7C,$3A,$AF,$6F,$AC,$C6,$6F,$D5
	.dc.b	$FE,$9C,$7F,$0F,$9E,$BA,$78,$F4,$7E,$17,$E9,$C3,$D1,$39,$CF,$F8
	.dc.b	$B7,$93,$FE,$12,$C7,$F8,$C7,$EB,$EF,$CD,$9F,$88,$BF,$6D,$7B,$BF
	.dc.b	$E6,$3F,$AF,$F8,$58,$F1,$C5,$8F,$1C,$38,$5E,$F8,$3F,$AF,$7D,$47
	.dc.b	$F1,$BA,$ED,$F9,$87,$F1,$1C,$79,$47,$1C,$7F,$8C,$46,$34,$F1,$E9
	.dc.b	$E2,$F6,$FF,$C8,$55,$D0,$F4,$CE,$B3,$FF,$3C,$7C,$A7,$BB,$3F,$E5
	.dc.b	$5F,$15,$FD,$9F,$A7,$FC,$95,$F8,$BB,$F6,$8F,$EE,$B1,$CF,$2F,$F4
	.dc.b	$C7,$A9,$F6,$BE,$C3,$FD,$EB,$5F,$DE,$BF,$26,$D6,$3F,$DC,$5F,$56
	.dc.b	$5D,$AF,$F0,$38,$7A,$A3,$18,$F8,$30,$7D,$F7,$96,$BF,$A4,$63,$E5
	.dc.b	$CB,$BF,$DE,$E6,$5F,$95,$BE,$99,$FB,$2B,$F6,$7F,$E6,$1E,$1F,$F5
	.dc.b	$4F,$D6,$7D,$57,$BA,$D7,$C3,$7E,$57,$E3,$FF,$DE,$1D,$5F,$66,$B5
	.dc.b	$9F,$9C,$FE,$B4,$F7,$FD,$D7,$DB,$18,$D3,$C3,$A7,$83,$C1,$F3,$5F
	.dc.b	$78,$F1,$F8,$9F,$8E,$BF,$EB,$5F,$8F,$FF,$81,$FE,$29,$F9,$83,$F5
	.dc.b	$1F,$F1,$FF,$E5,$FF,$E7,$BD,$BF,$CD,$D7,$BA,$D7,$D8,$29,$F9,$97
	.dc.b	$F8,$7B,$2F,$EE,$17,$D5,$FC,$49,$E0,$FA,$67,$EE,$FF,$8F,$8C,$69
	.dc.b	$E2,$D3,$FC,$9F,$F8,$4F,$9E,$79,$DA,$B5,$9B,$E5,$75,$4F,$C2,$1F
	.dc.b	$95,$3F,$24,$FE,$3F,$FA,$A0,$7E,$44,$FC,$2F,$D1,$87,$1B,$E1,$EF
	.dc.b	$6E,$F7,$FA,$5E,$B2,$EB,$FA,$F7,$A7,$0F,$F0,$74,$78,$33,$C4,$73
	.dc.b	$7C,$37,$FE,$D3,$8E,$58,$F6,$8F,$27,$BB,$EA,$B7,$E3,$3F,$E2,$EF
	.dc.b	$CA,$9F,$B3,$FF,$A4,$F3,$FE,$99,$D7,$FE,$A1,$BD,$FD,$5A,$F7,$F8
	.dc.b	$52,$DF,$B6,$FA,$72,$FF,$5A,$ED,$F5,$0F,$4B,$FE,$3C,$70,$7C,$B7
	.dc.b	$C5,$9F,$FA,$8F,$93,$CD,$18,$C3,$FD,$D9,$5F,$2B,$EF,$5E,$9F,$F9
	.dc.b	$CB,$ED,$BF,$49,$FB,$8F,$F9,$6B,$F0,$6D,$4E,$7C,$54,$FF,$97,$BF
	.dc.b	$71,$CF,$E4,$BF,$2B,$1E,$7F,$7B,$70,$E1,$7F,$E6,$4C,$F9,$CB,$3F
	.dc.b	$D9,$7E,$F9,$FE,$58,$9F,$F2,$17,$CE,$BF,$79,$7E,$CD,$BF,$CD,$5F
	.dc.b	$42,$FC,$4D,$F0,$3F,$E8,$7C,$38,$5E,$EB,$5F,$AB,$67,$1F,$E6,$99
	.dc.b	$7F,$95,$6F,$6D,$8E,$FA,$33,$EF,$9F,$F9,$5A,$BF,$C0,$D5,$A9,$F1
	.dc.b	$F7,$FE,$73,$F4,$97,$EB,$6F,$A3,$FF,$0A,$FE,$F5,$FC,$3B,$E9,$FF
	.dc.b	$33,$5A,$9C,$FB,$9F,$D8,$7F,$3E,$CE,$1C,$74,$FB,$B7,$D3,$F4,$3F
	.dc.b	$CB,$BC,$7E,$AD,$9F,$F9,$27,$DA,$F1,$61,$E1,$9C,$E5,$FD,$BF,$1F
	.dc.b	$F7,$6F,$E2,$4F,$3B,$FB,$8F,$90,$F7,$FF,$A9,$FB,$5A,$9C,$C7,$F5
	.dc.b	$6F,$D7,$78,$FA,$7F,$EF,$EE,$7F,$F7,$35,$5C,$7A,$7C,$CF,$E7,$70
	.dc.b	$7E,$03,$FE,$7D,$FB,$77,$E3,$5F,$E7,$5C,$78,$DE,$FF,$42,$F9,$DF
	.dc.b	$B0,$BE,$B1,$E6,$FF,$12,$FC,$75,$F9,$6B,$F1,$FF,$F4,$86,$9F,$E9
	.dc.b	$2C,$38,$5E,$EB,$5E,$FF,$FB,$2E,$73,$CD,$FE,$95,$7B,$7A,$1F,$EB
	.dc.b	$18,$47,$2F,$84,$67,$E0,$CF,$CB,$A7,$92,$73,$97,$F1,$FC,$3F,$CC
	.dc.b	$3F,$59,$FE,$91,$FC,$FF,$F7,$9F,$B0,$3E,$53,$FC,$2D,$F3,$6F,$D1
	.dc.b	$3F,$D6,$9F,$79,$FE,$AA,$BD,$D6,$A5,$F9,$77,$4C,$BF,$D4,$D5,$FC
	.dc.b	$1F,$E0,$BF,$D2,$58,$71,$F8,$16,$78,$D3,$C5,$18,$CF,$C3,$BF,$FA
	.dc.b	$BE,$F8,$5F,$97,$EA,$7F,$C2,$FF,$93,$FE,$7F,$DF,$7E,$9C,$F9,$0F
	.dc.b	$D7,$FE,$D2,$FD,$53,$F1,$1F,$59,$FE,$F7,$BD,$D6,$A7,$F1,$9E,$76
	.dc.b	$70,$E5,$C7,$FE,$C5,$BE,$BF,$85,$7D,$17,$8F,$BC,$CF,$B7,$C8,$F8
	.dc.b	$F0,$F1,$4E,$72,$FC,$16,$3F,$D1,$3E,$93,$FD,$A5,$F2,$DF,$B2,$FC
	.dc.b	$F1,$FA,$EB,$F7,$E4,$6F,$F1,$3F,$EC,$9E,$BE,$2B,$5F,$55,$9F,$96
	.dc.b	$FE,$8B,$97,$4C,$FF,$97,$BD,$AF,$87,$1F,$F5,$C8,$C6,$5E,$0F,$B3
	.dc.b	$FD,$07,$A1,$FE,$D3,$F6,$5E,$03,$F3,$E7,$EC,$1F,$C1,$FF,$55,$7E
	.dc.b	$6C,$FD,$35,$E9,$FE,$A3,$83,$FD,$CD,$97,$2C,$78,$FC,$23,$1E,$3E
	.dc.b	$67,$0D,$7C,$1A,$BE,$7F,$FD,$B1,$9F,$3C,$FF,$DA,$13,$E3,$7F,$DC
	.dc.b	$11,$9C,$FB,$6D,$E5,$F8,$6F,$5D,$78,$AF,$D5,$3E,$9E,$0F,$DF,$7F
	.dc.b	$00,$7F,$BC,$7E,$E1,$FA,$87,$F8,$CB,$F1,$9F,$F3,$FF,$5F,$FB,$27
	.dc.b	$4F,$6D,$EE,$B5,$EE,$7D,$8F,$C3,$9F,$A6,$6B,$60,$CF,$AF,$A1,$EC
	.dc.b	$C7,$97,$D4,$DA,$DD,$78,$A1,$AF,$8B,$B7,$CF,$5A,$D3,$C3,$B3,$7F
	.dc.b	$7C,$BD,$F7,$F0,$0F,$F8,$CB,$F2,$47,$D8,$3F,$2E,$7E,$7A,$FD,$2F
	.dc.b	$E9,$63,$E8,$C7,$8F,$D7,$B1,$E3,$87,$0F,$CD,$77,$7D,$F3,$7E,$75
	.dc.b	$F6,$3B,$FE,$0C,$E7,$1C,$3D,$6B,$C9,$AF,$8F,$EE,$7D,$3A,$78,$34
	.dc.b	$EF,$F3,$7F,$D6,$FC,$F7,$E6,$CF,$CD,$3D,$3F,$95,$3F,$16,$7E,$34
	.dc.b	$F1,$7F,$27,$E2,$F8,$B1,$E3,$87,$0B,$DF,$EB,$9A,$E1,$FA,$F5,$DB
	.dc.b	$F1,$0F,$E5,$CF,$08,$FC,$7E,$1C,$F4,$F1,$74,$B5,$F1,$7C,$DF,$9E
	.dc.b	$73,$1E,$5D,$7F,$07,$D7,$F3,$F1,$FC,$2F,$ED,$6F,$D7,$3F,$8C,$7A
	.dc.b	$4F,$CC,$BE,$D7,$EA,$9C,$3C,$0F,$97,$7B,$F8,$3F,$D8,$D7,$DF,$F8
	.dc.b	$AF,$DB,$E7,$D7,$35,$ED,$FA,$67,$33,$C3,$9B,$E1,$33,$19,$38,$C7
	.dc.b	$93,$0F,$C6,$7D,$9E,$6F,$C5,$AF,$9F,$B3,$FC,$6F,$F1,$5F,$B7,$F9
	.dc.b	$97,$E3,$CF,$DE,$5F,$83,$7F,$46,$FC,$31,$F9,$C7,$FA,$7F,$0E,$17
	.dc.b	$BA,$D4,$BF,$37,$56,$BE,$9F,$FA,$5F,$95,$F8,$EC,$FB,$E7,$8E,$2F
	.dc.b	$D8,$C6,$32,$F0,$7E,$35,$FD,$2E,$F8,$69,$DF,$E5,$FF,$DC,$5F,$82
	.dc.b	$FF,$80,$7F,$48,$FC,$47,$E7,$AF,$C9,$7F,$D0,$3F,$49,$FD,$FF,$E7
	.dc.b	$3F,$C8,$1F,$06,$BD,$D6,$BE,$6D,$5F,$E1,$7A,$D7,$07,$E3,$B1,$7E
	.dc.b	$9F,$FC,$7B,$1E,$33,$FC,$71,$E7,$F8,$2F,$F8,$33,$1F,$F7,$1E,$1C
	.dc.b	$AB,$FD,$4D,$87,$96,$B5,$D9,$F8,$E9,$FD,$8F,$EA,$AF,$C1,$DF,$89
	.dc.b	$3E,$F7,$FA,$C7,$D2,$3F,$4E,$FE,$2D,$E3,$7E,$1B,$0E,$17,$BF,$9E
	.dc.b	$FF,$0D,$D3,$ED,$3F,$88,$67,$DB,$3E,$AF,$E2,$B1,$E5,$0F,$CB,$FF
	.dc.b	$1B,$3C,$0F,$E4,$B9,$76,$D3,$C7,$EE,$7A,$67,$37,$E8,$9F,$3F,$B7
	.dc.b	$FE,$03,$F5,$7F,$86,$FB,$BF,$E7,$1F,$51,$F4,$3E,$5F,$D7,$7F,$D3
	.dc.b	$1D,$FF,$E1,$3C,$FF,$90,$FD,$DF,$F0,$AA,$D7,$DB,$FF,$87,$32,$E7
	.dc.b	$3F,$EB,$BA,$FD,$1F,$93,$FD,$89,$FE,$77,$9F,$77,$35,$C3,$3F,$0F
	.dc.b	$1B,$4F,$E6,$55,$E1,$BF,$36,$7E,$5F,$BA,$FF,$0E,$7E,$D1,$FC,$9B
	.dc.b	$CA,$FF,$43,$FE,$E5,$FE,$5F,$FC,$87,$F9,$B7,$E5,$9F,$5F,$75,$EF
	.dc.b	$F0,$25,$EF,$E3,$16,$FC,$23,$F4,$F7,$BF,$D0,$3F,$C6,$2F,$1B,$BB
	.dc.b	$FF,$56,$D3,$C3,$F7,$8D,$7F,$A3,$7C,$97,$9A,$B5,$AF,$82,$B5,$AF
	.dc.b	$E0,$BF,$0F,$FE,$1B,$F9,$B3,$F2,$27,$E6,$3F,$EA,$9F,$2B,$F1,$1F
	.dc.b	$AF,$7E,$26,$F7,$5A,$EF,$56,$BE,$87,$FD,$D3,$8F,$3F,$D3,$3F,$DC
	.dc.b	$2F,$5C,$AE,$1C,$6F,$F2,$9F,$8B,$FF,$BE,$2B,$58,$F4,$ED,$FF,$5D
	.dc.b	$70,$7F,$58,$F3,$BF,$CD,$3F,$A8,$BF,$33,$7E,$B2,$FD,$75,$FC,$02
	.dc.b	$FE,$01,$E1,$F8,$23,$0E,$17,$BA,$7E,$3D,$A9,$FE,$BF,$9B,$3F,$F3
	.dc.b	$BF,$E7,$18,$7C,$3A,$78,$34,$FF,$3A,$7B,$1F,$E1,$F9,$CE,$7E,$3F
	.dc.b	$C1,$7F,$D7,$9F,$15,$FE,$80,$FC,$53,$F9,$67,$1F,$F8,$CB,$F6,$C7
	.dc.b	$D4,$7F,$5B,$DB,$FC,$05,$C5,$FE,$D4,$F0,$D7,$BF,$AB,$D6,$5F,$E8
	.dc.b	$AF,$8E,$FF,$46,$3D,$70,$FE,$BE,$7F,$80,$F0,$E5,$7F,$E4,$0E,$8F
	.dc.b	$1C,$63,$E3,$DE,$C7,$F8,$83,$C1,$FE,$9F,$F5,$2F,$C7,$9F,$A9,$FF
	.dc.b	$7A,$FC,$8D,$F9,$F3,$F2,$DF,$EC,$18,$70,$BD,$FB,$0B,$DD,$3F,$49
	.dc.b	$5E,$FF,$59,$FE,$1A,$F1,$CA,$F0,$FF,$84,$B4,$F1,$47,$87,$47,$C9
	.dc.b	$79,$EB,$5A,$F8,$34,$F2,$FC,$87,$FA,$47,$E3,$3F,$D8,$5F,$75,$FE
	.dc.b	$8D,$F9,$E7,$EF,$5F,$43,$EA,$5F,$81,$FF,$56,$BD,$D7,$FA,$4E,$B5
	.dc.b	$FA,$5E,$74,$FF,$22,$FE,$29,$FE,$F1,$FA,$2F,$F5,$2E,$1C,$27,$F8
	.dc.b	$8B,$CB,$0E,$FE,$0B,$EA,$3C,$95,$CB,$5F,$26,$7F,$82,$FB,$1F,$F2
	.dc.b	$1F,$EA,$6F,$C2,$DF,$B2,$7F,$BA,$FE,$C7,$FD,$E7,$E6,$3F,$34,$FF
	.dc.b	$1D,$DE,$EB,$5F,$49,$F9,$9F,$CD,$B3,$8B,$F5,$8F,$CE,$DE,$F6,$FF
	.dc.b	$39,$FD,$27,$F9,$02,$39,$7C,$1F,$F8,$DF,$1F,$D8,$FA,$7F,$E7,$B9
	.dc.b	$CE,$C9,$FC,$4E,$1F,$B5,$FC,$CB,$F0,$BE,$8B,$F5,$AF,$DA,$FD,$EB
	.dc.b	$F6,$AD,$78,$8A,$D7,$AF,$FE,$7A,$FC,$3F,$97,$EA,$FF,$AF,$CB,$F1
	.dc.b	$F9,$F9,$8F,$F3,$AF,$3A,$3B,$E9,$FC,$0B,$EA,$FF,$CD,$73,$9C,$FB
	.dc.b	$4F,$F0,$6F,$A5,$FF,$05,$61,$FE,$7E,$FD,$47,$FC,$A9,$F8,$33,$F7
	.dc.b	$EF,$F6,$87,$CB,$3F,$4E,$7E,$7D,$FE,$1A,$EA,$FF,$32,$E1,$C2,$F7
	.dc.b	$4F,$DD,$75,$97,$5E,$86,$7F,$E1,$D8,$61,$FE,$63,$FD,$DF,$0F,$3F
	.dc.b	$C5,$FF,$A1,$FC,$27,$9A,$B7,$87,$2C,$FC,$BF,$87,$E7,$F4,$9F,$EB
	.dc.b	$1F,$8D,$3F,$49,$FF,$6E,$7E,$90,$3F,$D8,$5F,$43,$FE,$B3,$AD,$CE
	.dc.b	$A7,$FD,$71,$39,$FA,$57,$E5,$1E,$7B,$EB,$A7,$F6,$CE,$4F,$F4,$2C
	.dc.b	$72,$FB,$1E,$D8,$F8,$F0,$FF,$7D,$E5,$E2,$FD,$7B,$FC,$D3,$F6,$7F
	.dc.b	$E8,$CF,$CE,$5F,$79,$FE,$AE,$FD,$F1,$F6,$0D,$1F,$99,$BF,$6A,$FF
	.dc.b	$53,$63,$C7,$0E,$17,$BA,$FB,$1F,$6F,$FE,$1A,$E1,$67,$DB,$0F,$F0
	.dc.b	$CC,$E6,$3C,$33,$89,$E8,$B3,$F1,$E7,$FD,$36,$3C,$27,$FB,$12,$B8
	.dc.b	$CE,$BD,$2F,$DD,$CB,$F1,$5F,$E5,$7E,$37,$FA,$DF,$F4,$47,$E6,$CF
	.dc.b	$A6,$7E,$51,$FC,$2F,$0E,$1F,$25,$FD,$3F,$EA,$98,$70,$B7,$EB,$1B
	.dc.b	$CF,$AE,$1F,$C3,$8F,$F6,$BB,$FC,$EB,$F5,$6F,$E2,$71,$8C,$7A,$61
	.dc.b	$E2,$FC,$03,$CF,$7E,$6C,$BB,$E9,$FC,$B3,$0F,$F1,$F0,$FD,$47,$F8
	.dc.b	$8F,$E8,$6F,$D4,$9F,$BA,$BD,$DF,$67,$ED,$1C,$2F,$75,$A9,$F8,$1F
	.dc.b	$92,$E7,$E1,$1F,$9B,$7D,$4C,$FF,$D9,$F8,$76,$7F,$CA,$46,$25,$9F
	.dc.b	$87,$3F,$06,$FF,$E4,$FA,$F3,$56,$B3,$F0,$7C,$B7,$F9,$3E,$7F,$DF
	.dc.b	$7F,$9A,$3F,$77,$FF,$72,$7A,$6F,$88,$78,$DF,$EE,$EF,$81,$FF,$28
	.dc.b	$5E,$FF,$1E,$5E,$EB,$FD,$E7,$FA,$6E,$B1,$7C,$4F,$F9,$9B,$D1,$FF
	.dc.b	$28,$68,$CB,$93,$FE,$50,$3C,$75,$E2,$D3,$C3,$D9,$E7,$9E,$13,$9C
	.dc.b	$BF,$C4,$3F,$3B,$FE,$C2,$FD,$19,$FA,$67,$F2,$1F,$D1,$5F,$BA,$BF
	.dc.b	$6A,$7E,$F1,$FE,$CA,$C3,$85,$EE,$B5,$3F,$B2,$9C,$FC,$4F,$F7,$0D
	.dc.b	$03,$2E,$FF,$8D,$E6,$31,$8F,$3C,$9E,$FF,$F7,$39,$E3,$7E,$6C,$1A
	.dc.b	$79,$AF,$FA,$E5,$7E,$3B,$CA,$7D,$FF,$F0,$1F,$E0,$17,$E5,$2F,$A0
	.dc.b	$47,$2F,$57,$FD,$C7,$7C,$3D,$4C,$DE,$E9,$F2,$BD,$71,$FE,$19,$97
	.dc.b	$E3,$0F,$6E,$1E,$0F,$D1,$BF,$25,$18,$9F,$16,$1E,$1C,$BF,$1D,$7D
	.dc.b	$F1,$F2,$E6,$FC,$13,$F9,$B7,$C0,$BF,$0C,$F5,$FC,$47,$F6,$5F,$47
	.dc.b	$FB,$87,$93,$FD,$13,$65,$EF,$65,$EF,$E6,$5F,$9E,$EE,$B5,$7F,$CC
	.dc.b	$4F,$AA,$F7,$FA,$FF,$E9,$8F,$3B,$F0,$7C,$97,$93,$4F,$1F,$CD,$B5
	.dc.b	$FF,$69,$FE,$1D,$FB,$0F,$D3,$DF,$B0,$7F,$7B,$FD,$DD,$FB,$4B,$F4
	.dc.b	$BF,$F7,$CF,$5F,$E7,$3D,$76,$1C,$33,$61,$C2,$FF,$C7,$B7,$BA,$7E
	.dc.b	$3F,$FE,$CA,$8C,$65,$73,$F1,$3E,$1C,$DF,$49,$F9,$2A,$DE,$7D,$A7
	.dc.b	$F9,$A7,$8F,$FC,$FF,$E5,$9F,$B9,$7F,$96,$BD,$57,$E8,$AF,$DD,$DF
	.dc.b	$A7,$3F,$4D,$FF,$28,$FC,$77,$F7,$AC,$38,$5E,$E9,$F9,$6E,$BB,$7F
	.dc.b	$C0,$D5,$FE,$EF,$7F,$4A,$4D,$3C,$51,$FE,$22,$8C,$9C,$4F,$EF,$EF
	.dc.b	$6C,$BD,$38,$F8,$7D,$56,$9E,$89,$FF,$45,$FE,$75,$DB,$81,$F9,$AB
	.dc.b	$F5,$EF,$A9,$FA,$0F,$F0,$DF,$DA,$9F,$B7,$FE,$83,$0E,$17,$BE,$C5
	.dc.b	$EF,$E0,$B5,$C1,$FB,$EC,$FA,$E7,$F5,$71,$E3,$E1,$5A,$D6,$5C,$F1
	.dc.b	$FE,$39,$B7,$D5,$BE,$F8,$79,$3D,$E7,$DF,$E7,$1F,$E1,$7F,$99,$BF
	.dc.b	$4C,$7E,$68,$FD,$71,$FA,$2B,$F4,$2F,$F3,$37,$DC,$79,$B1,$63,$C7
	.dc.b	$0E,$17,$BF,$BA,$7E,$C3,$BA,$D5,$FD,$23,$19,$BF,$77,$F6,$47,$48
	.dc.b	$F1,$E9,$E2,$D3,$FC,$ED,$3C,$33,$EF,$1F,$84,$8F,$E1,$BE,$87,$E0
	.dc.b	$BF,$34,$7E,$51,$FE,$63,$FB,$27,$D8,$4F,$D2,$BF,$D0,$1E,$5F,$F0
	.dc.b	$0C,$7F,$96,$E1,$C3,$B3,$FC,$9D,$7B,$FC,$1B,$AD,$6B,$2F,$F3,$73
	.dc.b	$FE,$61,$17,$E0,$FD,$67,$FD,$AE,$F3,$CF,$C3,$87,$F8,$67,$2F,$05
	.dc.b	$F0,$C3,$CD,$3E,$6F,$A5,$FE,$7F,$F0,$4F,$E0,$0F,$CE,$FC,$C3,$F0
	.dc.b	$64,$FF,$B5,$B8,$1F,$83,$BF,$2E,$7E,$CD,$AC,$78,$E1,$C3,$F0,$76
	.dc.b	$1C,$2F,$FD,$F1,$8F,$7D,$BF,$A7,$E7,$DF,$E2,$3F,$C6,$F9,$72,$AB
	.dc.b	$5A,$CB,$A6,$3E,$2F,$C5,$BD,$37,$BC,$FB,$EB,$F5,$30,$FE,$61,$F1
	.dc.b	$CF,$CB,$3F,$CF,$DF,$12,$FD,$3F,$FC,$1B,$F9,$07,$D8,$F9,$DF,$8F
	.dc.b	$7B,$FB,$D5,$EF,$E9,$57,$F0,$6F,$78,$FF,$A2,$72,$FE,$6C,$7F,$73
	.dc.b	$7F,$CB,$4E,$31,$E9,$87,$3E,$6F,$F3,$F4,$63,$3F,$F0,$FF,$E3,$FF
	.dc.b	$E1,$DE,$87,$E8,$7F,$D1,$BD,$07,$ED,$2F,$D9,$FF,$D6,$5F,$8B,$7F
	.dc.b	$8E,$2F,$7F,$1F,$FC,$32,$B5,$E2,$BF,$17,$D6,$5F,$EF,$5E,$0F,$EC
	.dc.b	$87,$E6,$1F,$EA,$6E,$30,$E7,$9F,$83,$3F,$F7,$CC,$F2,$CB,$B4,$7F
	.dc.b	$44,$87,$13,$F0,$0F,$E4,$FD,$1F,$A6,$FC,$E9,$F9,$6B,$F6,$17,$EC
	.dc.b	$3F,$86,$FD,$EF,$8E,$1C,$3C,$76,$1C,$2F,$7E,$0F,$F4,$0C,$FB,$69
	.dc.b	$FE,$18,$D3,$F6,$7C,$FF,$E3,$3C,$B9,$63,$6B,$59,$74,$D7,$C5,$AF
	.dc.b	$F9,$FF,$2E,$74,$CB,$D1,$5A,$E9,$FF,$1B,$7E,$16,$F8,$87,$E3,$5B
	.dc.b	$FC,$95,$F0,$7F,$EE,$1F,$CF,$BE,$1F,$CA,$7F,$48,$C3,$85,$EE,$BF
	.dc.b	$DC,$DF,$89,$6B,$5F,$F1,$8D,$75,$F8,$8F,$F2,$66,$1C,$73,$EF,$5F
	.dc.b	$E4,$B3,$FD,$AA,$F8,$B6,$69,$E2,$F3,$3C,$F3,$98,$F2,$FC,$77,$FA
	.dc.b	$63,$F0,$E7,$EC,$EF,$D3,$1F,$B8,$BF,$67,$FA,$BF,$87,$3C,$8F,$F4
	.dc.b	$C6,$1C,$2F,$75,$AF,$1B,$38,$FF,$3D,$DF,$FD,$2C,$7F,$B8,$9E,$CE
	.dc.b	$2F,$96,$5D,$FA,$9F,$89,$79,$6F,$C9,$39,$FB,$0F,$F0,$27,$D8,$FF
	.dc.b	$72,$F1,$7E,$8F,$B4,$FA,$D7,$E7,$DF,$E4,$9F,$C2,$5F,$BE,$FF,$91
	.dc.b	$30,$E1,$DF,$FC,$62,$F7,$F5,$CB,$FE,$DF,$F0,$1F,$E6,$EE,$2F,$F3
	.dc.b	$1E,$9F,$7B,$F7,$7F,$F0,$D6,$5C,$B5,$B5,$AC,$FA,$6B,$FE,$59,$FC
	.dc.b	$AB,$CF,$39,$AF,$34,$FF,$97,$7E,$73,$FC,$3D,$F1,$1F,$E0,$CF,$CA
	.dc.b	$FF,$D8,$BF,$31,$FC,$CF,$E8,$7F,$96,$FE,$CC,$FC,$EB,$FA,$2E,$6B
	.dc.b	$DF,$C2,$7F,$88,$34,$5F,$F1,$8E,$2F,$E8,$35,$D2,$FF,$A5,$3F,$ED
	.dc.b	$2F,$C1,$3F,$C6,$9A,$78,$7E,$D7,$FC,$0F,$97,$3F,$B5,$F2,$E1,$FE
	.dc.b	$84,$C3,$BE,$BF,$D0,$38,$BF,$D0,$FE,$93,$E7,$BF,$C4,$BF,$3B,$FD
	.dc.b	$17,$F3,$CF,$F1,$9F,$D7,$3F,$17,$7E,$9D,$C2,$F7,$F2,$4B,$DD,$6B
	.dc.b	$EA,$3F,$85,$D6,$A7,$FD,$5C,$F0,$F8,$BB,$BF,$FA,$FB,$0E,$5B,$7F
	.dc.b	$95,$B1,$7E,$73,$DE,$B5,$8F,$6F,$A6,$FF,$7C,$F8,$6F,$92,$7E,$16
	.dc.b	$FC,$69,$F9,$DF,$F4,$9F,$BF,$FF,$4A,$F2,$BF,$2A,$FF,$21,$E3,$C7
	.dc.b	$0E,$17,$BF,$A0,$FF,$7F,$4E,$75,$FD,$1B,$DF,$E8,$3F,$DE,$F1,$FE
	.dc.b	$9F,$D3,$C3,$A7,$83,$8F,$FB,$8C,$F2,$9C,$E3,$FD,$6F,$D9,$FF,$71
	.dc.b	$7D,$F3,$BF,$F3,$5F,$A4,$FF,$9E,$BC,$27,$ED,$5E,$DF,$7F,$FC,$AB
	.dc.b	$7B,$F4,$17,$BA,$7E,$A1,$E0,$D7,$FD,$87,$7B,$7F,$A3,$BD,$FF,$17
	.dc.b	$FF,$54,$C7,$FA,$EB,$3F,$0F,$DC,$34,$FF,$33,$E9,$E0,$9C,$C7,$EF
	.dc.b	$AB,$FD,$49,$E9,$BF,$36,$FF,$68,$71,$7F,$8F,$3E,$B1,$FB,$EF,$E2
	.dc.b	$7A,$FF,$C3,$77,$BF,$C2,$3F,$BE,$57,$F9,$0A,$BF,$8C,$56,$B0,$FF
	.dc.b	$6D,$63,$FD,$D3,$0E,$DE,$07,$FB,$DF,$CD,$FE,$15,$5E,$0C,$7F,$C9
	.dc.b	$7B,$7E,$72,$73,$9F,$69,$FF,$60,$63,$FD,$7B,$BB,$F7,$1E,$DF,$F2
	.dc.b	$AF,$C9,$5F,$C0,$0F,$D0,$9F,$AC,$3F,$3B,$E1,$8F,$EA,$F0,$E1,$7B
	.dc.b	$A7,$E9,$FA,$AF,$DF,$57,$F3,$67,$F9,$5F,$A5,$FC,$DB,$ED,$BF,$12
	.dc.b	$FF,$38,$F8,$36,$9E,$2D,$1F,$B3,$FE,$DB,$6F,$3D,$6B,$5F,$37,$E2
	.dc.b	$7F,$BE,$CF,$F9,$4F,$82,$F9,$F7,$E2,$BF,$ED,$DF,$C8,$1F,$88,$FF
	.dc.b	$9F,$BD,$DF,$F2,$6F,$C1,$FE,$2A,$F7,$F8,$6A,$F7,$FB,$E7,$F1,$8A
	.dc.b	$D6,$7F,$E5,$B7,$F2,$3E,$E7,$F1,$28,$1E,$2C,$39,$DB,$F2,$8F,$35
	.dc.b	$F7,$D3,$C9,$E7,$FD,$7C,$3F,$DF,$BF,$AF,$FF,$AF,$F3,$FE,$F5,$D6
	.dc.b	$8F,$C2,$F7,$FF,$AE,$BC,$8C,$7E,$12,$BD,$D7,$F0,$8F,$1B,$FC,$D3
	.dc.b	$5C,$7F,$3D,$FC,$46,$5D,$B0,$AF,$EC,$6F,$C1,$F3,$5F,$80,$F8,$4F
	.dc.b	$24,$E6,$BB,$4F,$E1,$3C,$6C,$F8,$0F,$81,$7E,$2A,$FD,$5D,$FA,$43
	.dc.b	$F2,$E7,$E8,$CF,$C4,$3F,$C5,$FC,$9F,$D2,$71,$E3,$87,$0B,$DF,$C3
	.dc.b	$6B,$7F,$E7,$55,$D9,$FF,$2B,$E9,$E1,$F8,$0F,$5B,$E3,$D7,$C5,$AF
	.dc.b	$87,$EE,$FC,$F8,$76,$8F,$27,$DE,$7E,$37,$1F,$B3,$F5,$5F,$BB,$DF
	.dc.b	$F4,$5F,$97,$BE,$95,$FA,$07,$F2,$9E,$BF,$E5,$58,$70,$F5,$BD,$9E
	.dc.b	$7B,$0E,$16,$E5,$FD,$26,$7D,$70,$FF,$06,$BF,$C2,$78,$FF,$95,$BE
	.dc.b	$1C,$39,$66,$FC,$93,$CD,$7F,$E5,$2C,$FB,$E7,$E5,$F9,$1F,$C8,$FD
	.dc.b	$3F,$FB,$EB,$F0,$7F,$F2,$BF,$E1,$BF,$EA,$2E,$87,$D5,$BF,$43,$7C
	.dc.b	$27,$F9,$93,$4F,$CC,$E1,$C3,$BF,$FD,$BF,$7B,$F9,$DD,$57,$FB,$8F
	.dc.b	$8B,$FD,$F8,$FF,$B5,$F4,$F0,$43,$E4,$7F,$DB,$53,$C3,$1E,$9A,$78
	.dc.b	$BC,$AF,$3C,$FE,$A0,$E9,$B7,$9F,$3F,$D4,$47,$F8,$23,$F2,$F7,$E8
	.dc.b	$EF,$DB,$1F,$B1,$FE,$DF,$CF,$7E,$68,$FD,$1F,$85,$EE,$B5,$3F,$EA
	.dc.b	$2F,$63,$FD,$9B,$8F,$4C,$7F,$75,$92,$3B,$FC,$9E,$1D,$61,$DF,$ED
	.dc.b	$FF,$88,$46,$3D,$7F,$E1,$FE,$47,$F0,$2C,$BF,$0B,$8F,$F9,$CF,$F2
	.dc.b	$CF,$C2,$7E,$17,$FD,$5B,$EE,$BF,$CA,$DF,$A8,$AF,$7E,$92,$F7,$F2
	.dc.b	$9F,$E8,$AF,$B2,$FD,$75,$6A,$7F,$8B,$65,$F2,$4F,$07,$C3,$7F,$A6
	.dc.b	$23,$FD,$73,$19,$9E,$1E,$23,$4F,$17,$C7,$F0,$AF,$F4,$95,$F8,$35
	.dc.b	$FB,$51,$FE,$87,$F4,$7E,$B7,$E1,$DF,$C6,$FF,$76,$FE,$21,$FB,$3B
	.dc.b	$C5,$7E,$8C,$FC,$FF,$86,$3F,$C9,$30,$E1,$F0,$2C,$38,$5B,$F4,$1D
	.dc.b	$DA,$FF,$94,$7F,$00,$7E,$2B,$CA,$FE,$1D,$3C,$FE,$19,$AE,$76,$F2
	.dc.b	$C6,$34,$ED,$B7,$FA,$43,$0E,$38,$75,$AD,$6B,$E9,$F8,$1F,$E7,$3E
	.dc.b	$6F,$CB,$FC,$BD,$FA,$EB,$EB,$5F,$1E,$F8,$C7,$EC,$1E,$7F,$13,$FC
	.dc.b	$21,$AB,$1E,$38,$70,$BD,$F2,$7F,$43,$CF,$B6,$7D,$70,$FF,$6D,$E1
	.dc.b	$E0,$FB,$9F,$F0,$AB,$CB,$4F,$06,$8F,$C7,$FC,$F7,$FA,$9A,$D6,$7E
	.dc.b	$4E,$0F,$D6,$CF,$F5,$4F,$B4,$FF,$38,$7C,$83,$C9,$FF,$76,$7B,$CF
	.dc.b	$59,$FB,$0F,$EF,$AF,$75,$A9,$FF,$4E,$F3,$FE,$3E,$7F,$CF,$59,$FD
	.dc.b	$5C,$BF,$75,$1D,$9F,$E0,$86,$3E,$23,$FB,$A6,$1E,$0F,$B6,$F2,$CA
	.dc.b	$BB,$FD,$F7,$F8,$47,$0F,$EA,$3F,$08,$FD,$95,$FA,$8B,$F0,$EF,$F6
	.dc.b	$D7,$7B,$D2,$7E,$67,$FD,$1F,$4F,$E6,$18,$70,$BD,$D3,$F3,$3D,$65
	.dc.b	$D7,$2F,$EC,$6F,$7F,$C6,$AE,$1C,$78,$BF,$3D,$87,$92,$31,$A7,$CB
	.dc.b	$CF,$F7,$5F,$71,$FC,$6F,$F1,$E7,$E7,$AF,$D7,$3F,$DE,$9F,$A1,$BE
	.dc.b	$47,$F8,$D7,$EE,$DC,$31,$FF,$04,$E1,$C2,$F7,$5F,$C7,$FF,$00,$D5
	.dc.b	$2B,$AF,$D7,$7F,$2E,$F6,$7F,$D3,$53,$CB,$3F,$03,$FC,$14,$E3,$5D
	.dc.b	$7D,$D6,$9E,$3F,$81,$FE,$31,$B7,$A7,$0F,$45,$6B,$47,$E2,$9F,$C7
	.dc.b	$7E,$4F,$FC,$25,$F8,$5B,$F7,$4F,$DD,$74,$BE,$ED,$FA,$67,$F9,$D3
	.dc.b	$87,$FB,$57,$0F,$F8,$9B,$E9,$3F,$DF,$18,$70,$BD,$FC,$87,$F2,$3C
	.dc.b	$FA,$CF,$B5,$EF,$FC,$01,$70,$E5,$A7,$87,$DD,$FC,$2E,$2F,$C8,$7C
	.dc.b	$D5,$AC,$7B,$6F,$FE,$C0,$CB,$FA,$87,$EF,$6F,$D9,$9F,$73,$FC,$67
	.dc.b	$F6,$0F,$C3,$7E,$30,$AF,$B1,$7C,$2F,$7F,$0A,$BD,$FE,$63,$AD,$3F
	.dc.b	$C8,$78,$75,$CB,$FB,$1F,$3F,$C2,$9E,$98,$BF,$80,$73,$9F,$5B,$F1
	.dc.b	$79,$2F,$9F,$F4,$61,$D3,$0F,$0E,$9F,$9B,$FA,$1F,$F6,$DC,$7F,$AF
	.dc.b	$3E,$93,$F4,$DF,$76,$FC,$B1,$FC,$01,$FE,$97,$F7,$BF,$25,$7B,$F4
	.dc.b	$57,$BA,$FF,$4E,$7C,$4E,$B4,$FF,$2E,$FE,$0F,$FE,$39,$BF,$5E,$1D
	.dc.b	$FE,$3A,$D6,$B4,$F0,$E9,$E0,$F0,$3C,$B3,$9C,$3B,$CF,$93,$E0,$3E
	.dc.b	$BC,$FF,$61,$FC,$55,$F9,$87,$FB,$6B,$F0,$77,$DA,$3F,$17,$FC,$77
	.dc.b	$91,$F2,$FE,$55,$87,$0F,$C2,$98,$70,$BF,$F7,$2D,$EE,$BA,$E5,$FA
	.dc.b	$58,$C7,$D9,$FE,$36,$3C,$59,$F8,$73,$FF,$09,$D7,$18,$C6,$FE,$8F
	.dc.b	$84,$7D,$C3,$F4,$97,$BC,$FC,$3D,$F9,$03,$F1,$DF,$D8,$E7,$F8,$3C
	.dc.b	$78,$E1,$C2,$F7,$5F,$E6,$9A,$D6,$3F,$EE,$7D,$7D,$1E,$27,$F8,$A7
	.dc.b	$1E,$31,$83,$FD,$26,$F4,$CF,$C5,$9F,$92,$7F,$5C,$6A,$FD,$13,$AF
	.dc.b	$85,$FA,$7F,$82,$3F,$12,$7E,$92,$E0,$7D,$23,$EC,$BF,$D7,$3F,$9D
	.dc.b	$74,$D1,$8F,$1C,$38,$5E,$FF,$07,$FE,$17,$C3,$FA,$3A,$ED,$F9,$D7
	.dc.b	$D1,$78,$61,$EB,$39,$E9,$E2,$D3,$C3,$A7,$FB,$73,$0E,$53,$9C,$BC
	.dc.b	$9F,$37,$FD,$3B,$E4,$BF,$D2,$1F,$82,$3F,$56,$FF,$23,$F8,$8F,$C4
	.dc.b	$DF,$98,$7F,$97,$AF,$75,$A9,$7E,$73,$9F,$23,$FD,$97,$C1,$FD,$78
	.dc.b	$FF,$60,$DF,$FA,$77,$F1,$FF,$EC,$91,$89,$FF,$68,$E1,$CE,$39,$67
	.dc.b	$E3,$FB,$4F,$EB,$9A,$FF,$AC,$3E,$31,$FA,$83,$F3,$67,$DE,$7F,$98
	.dc.b	$7B,$1F,$91,$3F,$04,$FE,$C1,$87,$0B,$DF,$DC,$FF,$85,$E9,$F9,$FE
	.dc.b	$B2,$E9,$B7,$F4,$5F,$80,$FF,$4B,$3F,$D4,$A3,$FC,$28,$FF,$93,$63
	.dc.b	$2F,$F5,$18,$F2,$69,$E3,$FC,$CB,$D1,$18,$C7,$FC,$4D,$DF,$FE,$98
	.dc.b	$FD,$0D,$F6,$8F,$C7,$7F,$C3,$5F,$A9,$FF,$9E,$BF,$30,$FF,$65,$6B
	.dc.b	$FE,$68,$BD,$D6,$BE,$FD,$4F,$BE,$FE,$D5,$F7,$BF,$E8,$9B,$EB,$F2
	.dc.b	$3F,$CB,$A3,$19,$F7,$E1,$F7,$46,$32,$E9,$8F,$96,$BC,$9F,$08,$D7
	.dc.b	$FD,$1B,$E6,$FF,$99,$71,$7E,$AE,$FE,$00,$FF,$35,$7D,$E3,$F6,$67
	.dc.b	$EE,$4F,$DC,$1F,$8B,$7F,$9F,$78,$3F,$D3,$D8,$F1,$FC,$09,$8F,$1C
	.dc.b	$38,$7E,$5D,$BD,$07,$83,$4F,$F3,$C3,$FC,$5F,$1E,$70,$FC,$BB,$FC
	.dc.b	$05,$18,$C7,$AE,$BE,$4F,$AB,$F5,$69,$E1,$D3,$C1,$F8,$27,$F3,$0F
	.dc.b	$C0,$9F,$AE,$BF,$03,$FF,$23,$7B,$CF,$D2,$1F,$57,$FE,$CA,$F6,$7F
	.dc.b	$AD,$61,$C2,$F7,$5A,$FC,$DF,$F8,$14,$F4,$FD,$CB,$F9,$63,$BE,$4F
	.dc.b	$DE,$3C,$7D,$5F,$E4,$B7,$DA,$78,$E9,$E3,$9C,$FE,$03,$FD,$1B,$E0
	.dc.b	$FF,$CC,$5F,$A3,$BC,$BF,$F2,$CF,$E2,$9F,$EF,$7F,$A5,$FE,$25,$5A
	.dc.b	$9C,$FC,$47,$E4,$B8,$7F,$B9,$4F,$2D,$3F,$87,$3F,$E6,$DA,$FE,$84
	.dc.b	$F1,$FB,$F7,$D7,$FF,$B5,$EF,$8F,$8A,$D3,$C5,$5F,$EE,$DF,$CC,$BF
	.dc.b	$C5,$79,$7D,$AF,$A6,$7C,$B7,$F2,$8F,$D9,$BF,$C1,$1F,$B6,$FF,$0C
	.dc.b	$F7,$FF,$C7,$B5,$AF,$87,$56,$BE,$13,$FB,$06,$DF,$E3,$EC,$97,$D3
	.dc.b	$E4,$7F,$C8,$5F,$69,$FE,$EB,$8E,$53,$FD,$76,$31,$A7,$87,$E9,$1A
	.dc.b	$BE,$93,$CF,$7D,$2B,$53,$FE,$2B,$8F,$F2,$8F,$6F,$F8,$F3,$C5,$7E
	.dc.b	$F1,$FE,$C1,$F6,$DF,$9B,$7E,$2B,$7F,$35,$EE,$B5,$2F,$CA,$B9,$FD
	.dc.b	$AB,$F6,$BF,$4E,$7F,$A0,$C3,$B3,$E9,$31,$3F,$B2,$C3,$C1,$DB,$E8
	.dc.b	$AE,$DE,$63,$49,$87,$EA,$CF,$EE,$BF,$96,$FF,$95,$E5,$FC,$00,$FC
	.dc.b	$53,$FA,$F7,$E7,$9B,$F0,$97,$BA,$D7,$CA,$BF,$3C,$D4,$BE,$EB,$F6
	.dc.b	$39,$FE,$E3,$0E,$3F,$6F,$88,$F0,$BF,$EE,$7C,$7C,$1E,$47,$F0,$DA
	.dc.b	$DE,$BE,$4D,$7F,$45,$CD,$FE,$55,$C7,$F8,$1F,$E3,$EF,$85,$7E,$FD
	.dc.b	$FE,$20,$FD,$95,$F8,$47,$FA,$8F,$F6,$86,$3C,$71,$63,$C7,$0E,$17
	.dc.b	$FE,$A9,$FD,$83,$7F,$65,$FC,$42,$F7,$9F,$67,$FC,$4D,$19,$C3,$C5
	.dc.b	$F0,$7F,$E9,$18,$C5,$79,$3E,$B1,$AF,$93,$F0,$AF,$E3,$7A,$F8,$F1
	.dc.b	$F1,$56,$B4,$FF,$55,$7C,$83,$E0,$FC,$5B,$BF,$0D,$FE,$3C,$3F,$56
	.dc.b	$FF,$6D,$7E,$92,$FC,$CF,$FA,$1F,$17,$F2,$8B,$E1,$F1,$1C,$6F,$71
	.dc.b	$A9,$F9,$38,$7F,$29,$7F,$A2,$9C,$57,$EB,$3F,$D2,$E3,$17,$CF,$DE
	.dc.b	$FF,$1D,$C7,$3B,$EF,$7E,$59,$FE,$29,$3F,$4B,$F1,$0F,$99,$E4,$BF
	.dc.b	$56,$7B,$CF,$8B,$7D,$DB,$F5,$57,$E9,$2F,$87,$E5,$9F,$F4,$4C,$B9
	.dc.b	$63,$C7,$0E,$16,$FC,$43,$77,$FA,$DF,$C9,$FF,$D8,$4F,$D2,$D7,$C7
	.dc.b	$AF,$FA,$4E,$3F,$8E,$46,$31,$E9,$A7,$F8,$96,$79,$E7,$E0,$D7,$CF
	.dc.b	$3F,$E6,$AE,$7F,$F4,$E7,$D7,$BE,$05,$FB,$1B,$F4,$B7,$CA,$FF,$BD
	.dc.b	$BC,$37,$E3,$3F,$EA,$7D,$7F,$C9,$98,$70,$BD,$FE,$31,$F9,$FF,$C6
	.dc.b	$B5,$9F,$5F,$DD,$7F,$D4,$F8,$7F,$66,$8F,$06,$98,$8E,$6F,$87,$46
	.dc.b	$0F,$CB,$7F,$D5,$75,$AC,$BB,$FA,$BF,$16,$3F,$B8,$FE,$54,$FD,$41
	.dc.b	$F8,$17,$F9,$F3,$F2,$9F,$F2,$57,$E1,$5F,$E9,$6F,$33,$FC,$4B,$7B
	.dc.b	$AD,$7D,$AA,$B8,$D6,$B0,$F8,$B8,$FF,$A4,$7A,$FF,$61,$8F,$18,$C1
	.dc.b	$E2,$9F,$0F,$D1,$B3,$FF,$2D,$7C,$A7,$9E,$97,$E1,$AD,$67,$FE,$28
	.dc.b	$97,$4F,$F8,$EB,$A1,$FA,$43,$C0,$FC,$9F,$E4,$CF,$D9,$5F,$B3,$BF
	.dc.b	$5F,$E3,$C3,$FC,$17,$1E,$38,$70,$BD,$F8,$FA,$D3,$FC,$33,$5D,$BE
	.dc.b	$27,$FD,$C3,$87,$1D,$3C,$15,$FE,$51,$8C,$63,$CF,$DD,$FF,$10,$ED
	.dc.b	$E7,$BF,$37,$8C,$DF,$CD,$9F,$F0,$CF,$67,$F8,$C7,$A8,$FC,$9F,$FC
	.dc.b	$37,$F8,$BB,$F4,$67,$EF,$2F,$8B,$70,$7F,$20,$E1,$FF,$17,$E3,$C7
	.dc.b	$E8,$7F,$E0,$06,$1C,$2F,$F6,$FD,$DF,$CA,$EB,$B7,$E5,$9F,$C8,$70
	.dc.b	$E3,$A7,$F8,$A9,$FE,$4E,$63,$1E,$79,$B8,$FD,$13,$9D,$7C,$D2,$FA
	.dc.b	$8F,$F2,$0F,$DF,$7F,$33,$A7,$E9,$4F,$D5,$DF,$BD,$7F,$B3,$FF,$57
	.dc.b	$FE,$C1,$F0,$BF,$ED,$0C,$38,$5E,$EB,$5F,$26,$AF,$F6,$95,$6B,$0F
	.dc.b	$CF,$63,$DC,$46,$6F,$F8,$4B,$7F,$10,$65,$E0,$AD,$9F,$D2,$72,$F0
	.dc.b	$4E,$B2,$7B,$BF,$EF,$EF,$DE,$5F,$A2,$BF,$66,$7D,$EF,$FA,$D7,$F7
	.dc.b	$CD,$FE,$95,$FE,$CB,$BD,$D6,$BC,$A5,$74,$F9,$DD,$63,$CF,$6F,$F6
	.dc.b	$46,$1D,$BF,$56,$FF,$7E,$3C,$B3,$ED,$A7,$8E,$FC,$51,$E1,$F6,$BF
	.dc.b	$D0,$1D,$2F,$C0,$5F,$4B,$FC,$63,$E8,$9F,$B1,$FE,$33,$F3,$BF,$F0
	.dc.b	$75,$EE,$B5,$3F,$90,$C3,$38,$3F,$7E,$FF,$02,$5E,$F7,$FF,$3E,$1F
	.dc.b	$E8,$07,$97,$93,$FE,$90,$C3,$FA,$F4,$F2,$C7,$AF,$DA,$7E,$9B,$C9
	.dc.b	$FE,$F1,$F8,$73,$F0,$F7,$CF,$BF,$73,$FF,$0F,$7E,$FD,$FD,$AF,$D5
	.dc.b	$FD,$B6,$7F,$ED,$7A,$D7,$D0,$2B,$FB,$87,$E4,$9A,$CB,$A5,$7E,$EA
	.dc.b	$FF,$D2,$B1,$DF,$F1,$AC,$47,$3C,$BB,$FC,$D7,$96,$37,$A7,$92,$7C
	.dc.b	$7B,$7F,$A7,$BD,$E7,$E7,$DF,$BA,$FD,$7D,$FA,$33,$F7,$57,$E3,$5F
	.dc.b	$CE,$7A,$9F,$AC,$31,$E3,$AF,$EE,$B1,$E3,$87,$0F,$8D,$FB,$5E,$D6
	.dc.b	$F4,$ED,$FA,$87,$F6,$4C,$78,$E3,$FE,$11,$C7,$C2,$FE,$8E,$33,$ED
	.dc.b	$27,$AE,$BF,$D2,$FF,$05,$F5,$63,$E9,$DF,$D1,$3F,$D3,$F9,$BE,$16
	.dc.b	$5F,$E0,$0F,$80,$83,$EF,$9F,$B2,$BF,$48,$FA,$B3,$A3,$51,$E1,$F8
	.dc.b	$5F,$F7,$9C,$66,$BF,$84,$5F,$08,$EB,$1F,$C3,$E3,$11,$FE,$60,$FC
	.dc.b	$F5,$F9,$87,$F9,$CF,$D4,$CF,$D8,$A7,$3E,$97,$F9,$DF,$2E,$F8,$7F
	.dc.b	$86,$2B,$58,$3F,$2C,$F6,$CF,$B6,$73,$F2,$18,$8F,$F0,$FF,$CE,$FF
	.dc.b	$95,$BF,$32,$7E,$27,$FE,$3D,$CB,$AD,$7F,$97,$7E,$B5,$5A,$F1,$3D
	.dc.b	$38,$71,$E4,$E9,$7B,$AF,$F4,$B6,$1C,$B9,$3F,$C4,$5F,$91,$3F,$26
	.dc.b	$7E,$C3,$F0,$A7,$31,$FE,$7E,$AE,$1F,$50,$BF,$F4,$4C,$61,$FF,$1D
	.dc.b	$BE,$6F,$C1,$5F,$48,$FD,$6F,$E4,$7E,$F7,$F3,$A7,$31,$F2,$FF,$68
	.dc.b	$C5,$6F,$F1,$2F,$4D,$6B,$F3,$5F,$5F,$EB,$7F,$EC,$0F,$6F,$F0,$5E
	.dc.b	$3F,$D8,$8C,$70,$7F,$58,$D7,$E3,$C6,$FC,$BE,$3F,$86,$FF,$4D,$F9
	.dc.b	$1F,$8B,$F7,$5F,$E9,$AF,$D0,$5F,$B4,$FF,$BF,$B1,$F9,$F3,$98,$FB
	.dc.b	$DE,$EF,$FA,$07,$AB,$F2,$D3,$C7,$F1,$7E,$7F,$82,$B1,$7D,$F7,$D2
	.dc.b	$E5,$7E,$B4,$FC,$F1,$F9,$43,$F2,$2F,$C8,$D6,$A7,$3F,$91,$E2,$F8
	.dc.b	$7E,$95,$CD,$E5,$E8,$7D,$EF,$D2,$5F,$B2,$BF,$39,$7E,$BD,$FD,$F7
	.dc.b	$9B,$F1,$35,$A9,$CF,$EA,$BF,$A0,$8C,$7C,$FF,$3C,$39,$7E,$67,$F8
	.dc.b	$AF,$C0,$7F,$C3,$FF,$A8,$BF,$42,$7C,$6F,$EA,$27,$30,$F9,$0C,$7D
	.dc.b	$AA,$DE,$4F,$38,$C6,$7F,$96,$F5,$FF,$9A,$7E,$44,$FC,$ED,$FA,$F7
	.dc.b	$ED,$27,$31,$FE,$BD,$E3,$FD,$1C,$63,$97,$5F,$20,$C3,$FC,$E9,$5F
	.dc.b	$A8,$7F,$DA,$FF,$C0,$0F,$DE,$3F,$C7,$F3,$9F,$09,$39,$F8,$BF,$E0
	.dc.b	$77,$C3,$E6,$7F,$56,$F2,$F4,$3F,$5B,$F4,$3F,$EE,$2F,$95,$FE,$5F
	.dc.b	$F9,$07,$F0,$A8,$C7,$C9,$BF,$23,$C6,$8B,$FE,$21,$AF,$EF,$2F,$97
	.dc.b	$A1,$FB,$8F,$CB,$5F,$B6,$7F,$B6,$FF,$75,$FD,$55,$FF,$02,$9C,$FC
	.dc.b	$22,$5F,$90,$7F,$1A,$9C,$F1,$73,$C3,$FD,$C9,$F9,$97,$F0,$4F,$9E
	.dc.b	$7E,$C6,$FD,$99,$F5,$AF,$D9,$3F,$91,$D6,$A7,$3F,$AA,$62,$F8,$7A
	.dc.b	$DC,$DE,$5E,$8F,$F1,$0F,$C5,$9F,$92,$7F,$81,$7F,$27,$7E,$B3,$FC
	.dc.b	$BA,$73,$F4,$89,$7E,$BB,$9B,$FF,$06,$72,$73,$7F,$C1,$DF,$6D,$FC
	.dc.b	$70,$7E,$4C,$FD,$C5,$FB,$63,$F2,$EF,$E7,$75,$A9,$CF,$E1,$18,$BE
	.dc.b	$1F,$A1,$73,$79,$75,$7F,$28,$FA,$7F,$C9,$70,$3F,$49,$7E,$9F,$FD
	.dc.b	$1A,$73,$1F,$CC,$BC,$1F,$E6,$B1,$C3,$DB,$E5,$E9,$7F,$B5,$7D,$4F
	.dc.b	$E7,$DF,$16,$FD,$91,$FB,$AF,$FA,$2B,$8F,$FD,$BD,$97,$F4,$AA,$D4
	.dc.b	$E7,$F2,$6F,$EB,$11,$8F,$83,$7A,$FD,$31,$E7,$ED,$7F,$67,$8D,$FD
	.dc.b	$A7,$FB,$AF,$E2,$5F,$93,$7F,$39,$D3,$B4,$E6,$3F,$2D,$18,$FC,$93
	.dc.b	$F9,$CF,$D1,$FA,$2B,$97,$E8,$DE,$AF,$B8,$FF,$37,$FC,$AB,$F3,$37
	.dc.b	$EA,$FE,$69,$CC,$3B,$F1,$F8,$47,$F2,$EF,$BA,$E4,$FF,$30,$FB,$6F
	.dc.b	$8B,$F8,$E7,$FB,$0F,$A1,$FB,$2B,$C4,$F8,$33,$98,$7C,$8E,$3F,$1E
	.dc.b	$F9,$7F,$19,$C9,$F9,$BD,$FF,$DB,$7E,$89,$F8,$8F,$F9,$93,$EC,$3F
	.dc.b	$CD,$FF,$A9,$7F,$80,$E7,$3F,$85,$FF,$70,$FC,$2B,$FC,$57,$D9,$FD
	.dc.b	$F2,$73,$7F,$E7,$6E,$FE,$18,$F3,$FC,$1B,$B7,$DC,$FE,$FF,$E4,$5F
	.dc.b	$99,$BF,$64,$ED,$39,$87,$EB,$98,$F8,$EF,$E2,$9F,$8A,$72,$7F,$8B
	.dc.b	$78,$5F,$E0,$5F,$35,$F8,$F3,$F0,$D7,$E0,$DF,$E9,$EF,$A7,$FF,$0E
	.dc.b	$56,$A7,$3F,$8A,$7F,$0A,$E5,$FF,$0A,$CE,$6F,$FA,$AF,$1F,$0C,$7A
	.dc.b	$43,$E1,$3C,$1F,$B2,$FF,$93,$B7,$7E,$99,$FC,$EB,$2E,$73,$98,$7B
	.dc.b	$7E,$28,$E1,$EE,$72,$FC,$23,$DF,$E2,$FF,$87,$FC,$A7,$E9,$3F,$ED
	.dc.b	$1F,$AE,$FD,$DF,$9F,$FE,$35,$9C,$FE,$3D,$FD,$AF,$E0,$7F,$15,$B7
	.dc.b	$F8,$BA,$7F,$CD,$7F,$64,$3F,$1D,$8F,$3F,$B7,$EC,$F5,$F2,$BF,$C5
	.dc.b	$D9,$3E,$EF,$FC,$95,$F9,$0F,$F7,$1F,$2B,$FC,$89,$39,$FB,$9F,$F8
	.dc.b	$01,$DD,$F8,$2F,$80,$FF,$00,$69,$F9,$EE,$CF,$F7,$A1,$FB,$27,$C5
	.dc.b	$7C,$71,$FE,$B5,$3D,$32,$EF,$F1,$1F,$E3,$FF,$4F,$F8,$BF,$EE,$EF
	.dc.b	$CB,$5F,$AC,$FF,$26,$E4,$FF,$25,$D6,$A7,$3F,$1F,$FE,$59,$8F,$83
	.dc.b	$AF,$F0,$AF,$8D,$63,$CF,$D4,$ED,$18,$C3,$FB,$8E,$BF,$E6,$1F,$C3
	.dc.b	$1F,$97,$FF,$B5,$BF,$0F,$7E,$8D,$FE,$8E,$DB,$FC,$B7,$5A,$FC,$7B
	.dc.b	$FD,$1B,$3F,$E5,$78,$F0,$99,$AF,$F2,$1F,$2F,$C6,$8C,$7C,$6F,$F9
	.dc.b	$77,$E6,$DF,$9E,$3F,$56,$FF,$36,$F6,$BD,$EF,$F3,$A5,$6A,$73,$F6
	.dc.b	$38,$BE,$1F,$88,$7A,$B7,$FF,$37,$DE,$E1,$E1,$7F,$9A,$FF,$66,$7D
	.dc.b	$C3,$F6,$A7,$C4,$7F,$A9,$FE,$57,$FD,$05,$5A,$9C,$FE,$30,$FC,$E7
	.dc.b	$3D,$3F,$78,$FF,$4A,$56,$F0,$FE,$05,$AF,$FA,$02,$31,$F0,$5F,$E7
	.dc.b	$BF,$D4,$FF,$C7,$FE,$AB,$CB,$FF,$67,$7C,$97,$FA,$46,$73,$1F,$ED
	.dc.b	$1F,$94,$FF,$4A,$7E,$BF,$F7,$26,$7E,$05,$FB,$0A,$FF,$6B,$EB,$C6
	.dc.b	$31,$FA,$E7,$FA,$27,$F7,$97,$E5,$9F,$C9,$3F,$04,$7E,$1A,$FD,$AB
	.dc.b	$FD,$41,$5A,$F9,$35,$6B,$EE,$BF,$D4,$9F,$99,$FE,$32,$A6,$3C,$FE
	.dc.b	$E3,$B3,$D7,$EF,$7F,$D3,$1F,$A2,$BF,$71,$78,$9F,$E5,$DF,$DB,$1F
	.dc.b	$9D,$FF,$AB,$EB,$53,$9F,$DB,$B1,$C9,$F3,$7D,$6F,$DD,$7E,$77,$F9
	.dc.b	$CB,$DF,$E2,$1F,$C4,$EB,$96,$7D,$A3,$F8,$CF,$81,$FE,$B0,$FC,$27
	.dc.b	$FD,$CB,$F0,$1F,$E2,$3F,$CB,$9F,$A9,$FF,$AF,$E7,$3F,$81,$25,$F4
	.dc.b	$D9,$CF,$C5,$B3,$F5,$5F,$D6,$E1,$C3,$ED,$FF,$D6,$CF,$2F,$C5,$7F
	.dc.b	$D7,$1F,$8C,$3E,$BF,$FC,$23,$FB,$C3,$F8,$03,$F9,$1F,$0F,$FB,$12
	.dc.b	$B5,$F0,$6A,$D7,$E3,$FF,$ED,$39,$FD,$DF,$E0,$5F,$BC,$77,$2F,$C1
	.dc.b	$BB,$E5,$DB,$AF,$FD,$91,$FB,$DB,$F3,$2F,$F0,$B7,$E7,$BF,$D3,$FF
	.dc.b	$1D,$FF,$6F,$4E,$7E,$FD,$3F,$D4,$FE,$F7,$3C,$FF,$03,$F5,$7F,$E1
	.dc.b	$77,$BA,$7D,$5F,$6C,$BA,$FD,$57,$FB,$5F,$D2,$7E,$A6,$FC,$55,$FB
	.dc.b	$AF,$F8,$D3,$E6,$FF,$DD,$D5,$A9,$CF,$CF,$BF,$1F,$CE,$FF,$43,$D0
	.dc.b	$FE,$37,$7B,$C7,$F7,$73,$CB,$2E,$D1,$FC,$07,$E4,$3F,$99,$FE,$B2
	.dc.b	$FB,$BF,$F0,$97,$A0,$FD,$F7,$FD,$D3,$E0,$FF,$BD,$2B,$5F,$7D,$F9
	.dc.b	$49,$7E,$68,$AD,$69,$F8,$2F,$CD,$3F,$96,$E1,$C3,$17,$27,$FB,$DE
	.dc.b	$31,$F0,$9F,$EF,$9F,$DF,$3F,$A9,$7E,$30,$FC,$27,$FD,$D7,$F6,$FD
	.dc.b	$78,$7F,$A4,$D6,$A7,$FD,$DF,$F2,$39,$D7,$FA,$57,$A1,$FE,$45,$7F
	.dc.b	$27,$5F,$CF,$8E,$1F,$41,$DE,$59,$77,$F9,$9F,$F2,$1F,$0B,$3F,$EA
	.dc.b	$FF,$A5,$BF,$6F,$7C,$AB,$F6,$5F,$F6,$9F,$C3,$78,$76,$FE,$D5,$A3
	.dc.b	$0E,$17,$BE,$D7,$3E,$EB,$FB,$84,$FF,$52,$A9,$97,$6C,$9D,$DE,$4D
	.dc.b	$3C,$73,$9F,$83,$C6,$9F,$03,$67,$E8,$FF,$C6,$BC,$27,$EB,$DF,$EC
	.dc.b	$29,$F6,$56,$EB,$53,$FE,$C3,$C7,$FB,$AF,$DA,$7F,$64,$B9,$5C,$EF
	.dc.b	$B9,$EE,$8E,$FF,$0F,$FE,$8A,$FB,$8C,$F0,$FF,$82,$FD,$97,$E6,$0F
	.dc.b	$D4,$7F,$9D,$7C,$77,$F7,$7C,$7C,$79,$7C,$8B,$DD,$6B,$F6,$8F,$8D
	.dc.b	$8F,$F0,$3A,$FE,$1B,$F8,$2F,$F7,$8B,$DF,$CA,$7F,$97,$67,$71,$8D
	.dc.b	$FC,$76,$FC,$9B,$F2,$FF,$A3,$7F,$AF,$FE,$D5,$F4,$1D,$A7,$3F,$14
	.dc.b	$97,$E1,$59,$CF,$F8,$EF,$57,$F8,$22,$F7,$F0,$5D,$9E,$3D,$9E,$0F
	.dc.b	$D8,$DF,$BB,$3F,$4F,$FE,$ED,$F4,$8F,$D2,$3D,$B5,$A9,$CF,$F0,$06
	.dc.b	$34,$FE,$57,$F9,$67,$F8,$5A,$B5,$F5,$DE,$FE,$BF,$93,$CB,$FE,$39
	.dc.b	$9C,$FB,$BF,$03,$F5,$87,$D2,$7F,$B9,$FF,$0C,$FF,$35,$FE,$D1,$F3
	.dc.b	$27,$3F,$95,$FF,$9B,$A5,$FB,$1F,$D0,$FC,$DF,$FC,$4B,$53,$0E,$5F
	.dc.b	$0F,$D5,$E9,$EE,$FC,$EF,$E0,$17,$ED,$3F,$E6,$EE,$27,$E8,$DF,$BB
	.dc.b	$E2,$F3,$F1,$2F,$7F,$88,$2F,$7F,$86,$FE,$02,$7F,$C6,$BF,$B6,$7F
	.dc.b	$8D,$9D,$CF,$E9,$39,$F5,$9F,$7F,$53,$EF,$FD,$AB,$F5,$07,$E2,$DF
	.dc.b	$DE,$7E,$B9,$FA,$BF,$E3,$2B,$53,$9F,$D6,$F1,$97,$F5,$AF,$A9,$FF
	.dc.b	$22,$D6,$BE,$C7,$FB,$57,$1F,$AA,$31,$5F,$8E,$FD,$61,$F9,$2B,$EC
	.dc.b	$9F,$AB,$7C,$2F,$D2,$7F,$37,$5A,$9C,$F4,$FF,$67,$CB,$E2,$CE,$6F
	.dc.b	$DD,$18,$C3,$A4,$38,$BF,$3B,$F9,$A3,$F3,$1E,$FF,$18,$FD,$77,$FC
	.dc.b	$73,$8F,$D0,$AD,$60,$AD,$7E,$F5,$FA,$E9,$FF,$31,$FE,$FD,$FE,$64
	.dc.b	$77,$3F,$D6,$B8,$35,$97,$6F,$03,$F7,$FF,$0B,$FC,$5B,$AF,$FC,$0F
	.dc.b	$F8,$43,$F6,$27,$CB,$FF,$1F,$AD,$7D,$AA,$B5,$C5,$9F,$5B,$FD,$27
	.dc.b	$4F,$D9,$6C,$7E,$32,$F8,$7D,$57,$82,$31,$83,$F4,$FF,$1F,$C5,$BE
	.dc.b	$C9,$FB,$23,$F4,$D7,$E3,$0F,$D0,$BC,$97,$BA,$D4,$BF,$67,$CF,$37
	.dc.b	$F8,$B7,$83,$D8,$DB,$E1,$8F,$F9,$E2,$3F,$D3,$53,$FE,$4A,$CB,$C3
	.dc.b	$3F,$E9,$AE,$CF,$DB,$CB,$F6,$EF,$B1,$F7,$DF,$94,$FD,$FD,$F7,$3F
	.dc.b	$1E,$73,$E6,$27,$FC,$47,$6C,$FB,$D7,$C8,$FC,$F3,$E4,$BC,$3E,$FF
	.dc.b	$EA,$BC,$BF,$32,$FA,$FF,$AD,$BF,$5B,$79,$9F,$EC,$AF,$DF,$3F,$DC
	.dc.b	$BC,$1E,$6A,$D4,$E7,$F7,$FF,$B5,$F1,$7F,$5A,$66,$1C,$BF,$25,$FF
	.dc.b	$36,$E1,$C6,$31,$97,$F4,$4F,$CD,$1F,$9D,$7E,$FF,$BF,$F8,$AF,$E7
	.dc.b	$DF,$EF,$DC,$3F,$8D,$D6,$BD,$BF,$EC,$3F,$02,$AF,$CE,$56,$BE,$27
	.dc.b	$FD,$B3,$7D,$39,$7F,$CC,$4F,$1C,$7F,$90,$F9,$7E,$1C,$FC,$19,$7E
	.dc.b	$6F,$1E,$FF,$10,$FC,$59,$FB,$EF,$FB,$E7,$E1,$7F,$C7,$35,$A9,$CF
	.dc.b	$D1,$7F,$B9,$FF,$30,$FE,$F5,$AB,$A7,$C7,$94,$BE,$7F,$75,$F5,$F8
	.dc.b	$37,$3D,$F4,$64,$FC,$7F,$FB,$4F,$E6,$5F,$C2,$7F,$02,$7E,$5B,$FB
	.dc.b	$7F,$83,$4F,$F8,$24,$D4,$67,$6C,$7E,$2B,$FE,$FC,$9F,$E2,$5A,$7F
	.dc.b	$A1,$5D,$FB,$5F,$C3,$3C,$5F,$27,$27,$4F,$83,$69,$E0,$AF,$E1,$D8
	.dc.b	$77,$FD,$39,$F9,$A3,$E6,$3F,$6D,$E0,$BF,$54,$D4,$E7,$DA,$FC,$77
	.dc.b	$E8,$7C,$71,$F1,$7E,$8D,$E5,$CB,$9E,$5F,$BF,$FC,$83,$E0,$C6,$30
	.dc.b	$FF,$62,$F0,$79,$BF,$32,$FF,$69,$FA,$3F,$E9,$0F,$C7,$BF,$BA,$FE
	.dc.b	$1F,$FE,$2B,$9C,$FA,$29,$FE,$F3,$F8,$76,$7F,$00,$FA,$5B,$FD,$3F
	.dc.b	$17,$FB,$4E,$0F,$8A,$CC,$FF,$86,$A3,$9E,$1E,$28,$FD,$87,$4A,$73
	.dc.b	$C1,$FE,$20,$F6,$3F,$C3,$DE,$6F,$FA,$EB,$F2,$A7,$E9,$7F,$EB,$AD
	.dc.b	$3F,$BB,$D6,$BE,$11,$5A,$FC,$7B,$FC,$89,$2F,$C9,$FC,$7A,$E3,$ED
	.dc.b	$AE,$3C,$9F,$D2,$63,$CB,$BF,$C5,$18,$CB,$FC,$FD,$7E,$8C,$5F,$1C
	.dc.b	$FD,$9B,$FB,$BF,$92,$FD,$BF,$F9,$56,$9E,$6A,$D7,$C7,$FE,$56,$5F
	.dc.b	$AE,$79,$BE,$4B,$C7,$DD,$F8,$AC,$FA,$E9,$FE,$D3,$7B,$E9,$DB,$D7
	.dc.b	$F8,$D9,$F4,$D7,$F6,$58,$7E,$49,$FD,$17,$C9,$3E,$B9,$FB,$47,$F3
	.dc.b	$7F,$D3,$FF,$CD,$15,$A9,$CF,$C7,$BF,$4D,$CC,$F9,$36,$FA,$EB,$F3
	.dc.b	$5C,$43,$BF,$F6,$13,$FD,$A7,$0E,$FD,$DF,$E5,$A8,$ED,$9F,$F4,$DB
	.dc.b	$FF,$3C,$75,$BF,$64,$7E,$D3,$FD,$FB,$77,$E2,$0F,$D5,$37,$5A,$F9
	.dc.b	$CF,$EE,$D2,$F3,$73,$97,$FB,$E7,$2F,$A8,$DB,$E1,$8F,$48,$E7,$97
	.dc.b	$7E,$6F,$F3,$7F,$7F,$F6,$BF,$DE,$5F,$57,$FE,$F5,$FD,$C7,$F1,$5E
	.dc.b	$0F,$F9,$1A,$B5,$39,$7F,$D2,$43,$E8,$BE,$2F,$37,$F9,$DF,$E7,$2C
	.dc.b	$F0,$FC,$A1,$F9,$3F,$82,$59,$F8,$27,$FB,$16,$DF,$37,$E2,$5F,$0A
	.dc.b	$F9,$07,$EC,$CF,$C8,$5F,$A1,$BF,$7C,$FF,$52,$61,$C2,$F7,$5A,$FD
	.dc.b	$CF,$38,$F9,$AB,$F9,$86,$5F,$B8,$E3,$FF,$31,$C7,$1C,$FB,$7E,$2D
	.dc.b	$FE,$94,$8C,$63,$FE,$E4,$AF,$F3,$4F,$C3,$BF,$38,$FE,$35,$F8,$C3
	.dc.b	$F5,$CF,$EE,$7E,$3F,$FA,$DE,$B5,$F0,$DF,$EC,$89,$FE,$EF,$E7,$E7
	.dc.b	$0F,$93,$5F,$EF,$0D,$BF,$87,$E5,$DA,$7A,$DF,$ED,$27,$F2,$55,$D3
	.dc.b	$2F,$07,$E7,$1F,$EA,$AF,$CC,$5F,$A5,$3E,$EF,$F6,$3F,$8A,$3F,$5A
	.dc.b	$7E,$89,$FE,$CB,$BD,$D6,$A5,$FA,$FE,$7F,$7F,$FB,$B5,$D3,$5F,$91
	.dc.b	$EB,$7F,$42,$8E,$3E,$D7,$FB,$2B,$EA,$FC,$5A,$F8,$63,$FD,$9B,$E1
	.dc.b	$FF,$AF,$BF,$34,$FB,$5F,$9E,$3F,$3E,$FB,$BE,$43,$EF,$3F,$DB,$D5
	.dc.b	$A9,$CF,$DE,$63,$E5,$FF,$1F,$3F,$83,$DF,$C6,$7F,$2E,$78,$79,$1F
	.dc.b	$CD,$F8,$3F,$AB,$C7,$3F,$0F,$FB,$45,$7F,$BA,$FF,$35,$7E,$06,$F9
	.dc.b	$0F,$F5,$0F,$EB,$AF,$A8,$7E,$29,$BA,$D7,$90,$AD,$7C,$97,$3F,$B0
	.dc.b	$FD,$1E,$5D,$33,$F2,$9F,$C2,$5E,$F9,$F6,$CF,$FD,$C3,$18,$C7,$A7
	.dc.b	$DD,$FF,$B9,$38,$5F,$B9,$BF,$73,$FB,$9F,$AC,$BF,$32,$FE,$41,$E5
	.dc.b	$FF,$AF,$2F,$75,$A9,$64,$AD,$61,$FB,$8B,$7F,$00,$58,$ED,$E2,$B2
	.dc.b	$FF,$29,$73,$FF,$80,$2F,$AC,$E7,$EF,$7F,$DF,$3F,$63,$FC,$6B,$F2
	.dc.b	$EF,$E3,$BF,$9E,$3E,$A9,$FB,$27,$87,$2F,$F0,$4D,$EF,$E1,$17,$BF
	.dc.b	$D4,$B5,$D1,$FD,$A2,$BF,$8D,$3F,$A4,$BE,$3F,$A2,$7F,$10,$9E,$5F
	.dc.b	$79,$F7,$EB,$87,$C6,$35,$F1,$FC,$87,$F9,$87,$97,$FD,$C3,$F9,$6B
	.dc.b	$F2,$DF,$F6,$06,$8F,$CE,$8F,$F5,$DA,$DC,$EA,$73,$F1,$9E,$6C,$FF
	.dc.b	$A2,$61,$CD,$FB,$FE,$62,$1F,$8E,$FF,$26,$7C,$1F,$89,$72,$AF,$26
	.dc.b	$7F,$E2,$C9,$CB,$E3,$D3,$33,$FE,$0A,$F5,$5F,$A1,$3F,$6D,$7B,$9F
	.dc.b	$3B,$F5,$F7,$E5,$1E,$5D,$98,$70,$BD,$F5,$AD,$F9,$56,$EB,$FB,$9E
	.dc.b	$7F,$3C,$F0,$67,$FC,$FA,$39,$67,$DE,$78,$4F,$8E,$3E,$FF,$B5,$E8
	.dc.b	$9F,$ED,$1E,$FB,$F4,$37,$C0,$FC,$EF,$87,$7E,$42,$FD,$03,$D1,$7B
	.dc.b	$AD,$4F,$F8,$D7,$AF,$EA,$E3,$FE,$2C,$A7,$EF,$1F,$D4,$9E,$D1,$F4
	.dc.b	$9E,$5C,$DF,$DE,$FE,$AF,$C9,$CA,$9C,$E8,$FD,$0B,$3F,$92,$FF,$86
	.dc.b	$3E,$31,$FA,$AF,$F8,$FF,$5F,$E5,$A7,$8B,$4F,$F0,$25,$6E,$75,$FA
	.dc.b	$EF,$D1,$87,$ED,$5F,$D9,$9F,$F1,$1D,$7E,$19,$E1,$ED,$BF,$30,$F0
	.dc.b	$FC,$43,$3F,$0C,$F9,$6F,$EE,$7E,$99,$FE,$9D,$F8,$AE,$BF,$A8,$FF
	.dc.b	$3F,$FC,$B3,$3F,$6F,$FD,$0B,$54,$FF,$43,$FB,$FE,$3F,$95,$7F,$78
	.dc.b	$C7,$FD,$4D,$97,$E6,$1E,$97,$C3,$D8,$F8,$F1,$8C,$3F,$CB,$95,$FE
	.dc.b	$23,$AF,$F4,$D7,$F0,$0F,$FB,$3F,$F2,$EF,$F2,$F7,$E8,$9F,$91,$9C
	.dc.b	$FE,$32,$9F,$F3,$17,$91,$9F,$C4,$7F,$C1,$34,$FD,$86,$C3,$1E,$9E
	.dc.b	$77,$D8,$AE,$7F,$98,$FE,$FF,$4F,$F1,$A7,$D2,$3F,$33,$7E,$68,$FB
	.dc.b	$E7,$CA,$BF,$43,$FF,$7E,$F8,$5F,$67,$C6,$FF,$05,$61,$C2,$F7,$4F
	.dc.b	$C5,$7F,$55,$97,$F9,$37,$3F,$DE,$BD,$DB,$7C,$72,$FF,$6D,$F9,$3F
	.dc.b	$EA,$6A,$F2,$69,$E3,$9F,$F4,$75,$FE,$D7,$DA,$7E,$69,$FE,$2B,$F5
	.dc.b	$1F,$99,$7F,$72,$FD,$37,$53,$98,$FE,$EB,$F9,$6E,$30,$FE,$1F,$83
	.dc.b	$F1,$DF,$F1,$9B,$D3,$EC,$FF,$80,$DF,$E2,$79,$3E,$04,$E7,$3E,$FE
	.dc.b	$57,$E9,$3E,$D9,$F2,$CF,$CF,$1F,$A4,$F9,$BF,$42,$7E,$C1,$FC,$36
	.dc.b	$F7,$5A,$E1,$53,$F7,$FD,$63,$FE,$46,$BF,$F7,$47,$0F,$F2,$67,$8E
	.dc.b	$5F,$0B,$E6,$FC,$55,$E1,$8F,$8B,$F1,$D8,$CF,$B6,$4F,$B4,$FE,$3D
	.dc.b	$F2,$2F,$D7,$5F,$AD,$3F,$29,$7E,$A8,$FE,$00,$73,$70,$BD,$D6,$A7
	.dc.b	$FA,$07,$81,$FE,$DE,$C5,$FB,$67,$F9,$79,$ED,$C7,$FC,$F9,$F0,$67
	.dc.b	$DF,$37,$E1,$1F,$C3,$A7,$39,$78,$BF,$40,$FE,$23,$E9,$BF,$77,$7E
	.dc.b	$4F,$FE,$5B,$E9,$FA,$1F,$BC,$7F,$39,$AD,$4E,$79,$7F,$9B,$F7,$FE
	.dc.b	$66,$5F,$BC,$7F,$9C,$9E,$9C,$1F,$D4,$9E,$3E,$47,$D3,$8F,$0E,$7E
	.dc.b	$08,$FA,$DF,$94,$FF,$24,$FD,$57,$EC,$78,$9F,$E9,$FF,$CA,$7F,$D7
	.dc.b	$3E,$3F,$F4,$B9,$CF,$A0,$9F,$F5,$EF,$89,$9E,$BF,$F4,$1D,$77,$D3
	.dc.b	$FA,$51,$FE,$A9,$78,$D7,$FA,$97,$7F,$D9,$46,$3E,$2F,$FA,$47,$67
	.dc.b	$F8,$13,$F6,$27,$EB,$4F,$83,$FC,$07,$A7,$FE,$A3,$C7,$F9,$95,$EE
	.dc.b	$B5,$E1,$FE,$FF,$F2,$6F,$EE,$D3,$FE,$0E,$FC,$BB,$FD,$76,$F5,$DB
	.dc.b	$FC,$6A,$F1,$CB,$F1,$9E,$1F,$8B,$E9,$5A,$F8,$A7,$F2,$1F,$91,$79
	.dc.b	$BF,$52,$FF,$68,$79,$1F,$E4,$7F,$AE,$FF,$63,$7C,$0F,$82,$73,$F6
	.dc.b	$5F,$E2,$EE,$97,$EE,$59,$FB,$1F,$1D,$7F,$0D,$E0,$F8,$FC,$08,$E3
	.dc.b	$FB,$67,$F1,$F8,$C7,$E3,$3F,$9A,$CB,$FC,$73,$F0,$7F,$23,$EB,$5F
	.dc.b	$A5,$BF,$70,$7E,$4C,$FC,$B5,$F8,$37,$F8,$53,$BB,$E4,$E1,$C2,$F7
	.dc.b	$5A,$F9,$EF,$93,$3D,$38,$3E,$3A,$FD,$43,$E5,$BC,$72,$ED,$FA,$E7
	.dc.b	$8F,$4F,$14,$78,$77,$FB,$BE,$67,$8F,$F0,$B7,$E6,$DF,$E3,$7F,$97
	.dc.b	$7E,$B9,$F6,$2B,$53,$9F,$D4,$71,$F8,$E7,$DA,$9F,$66,$FF,$69,$DF
	.dc.b	$DB,$FF,$A8,$63,$97,$17,$FA,$83,$0F,$F2,$27,$81,$FE,$15,$FB,$B7
	.dc.b	$E4,$7F,$E2,$BF,$CC,$3F,$57,$FB,$EB,$F5,$E7,$EB,$BF,$E3,$FB,$DF
	.dc.b	$DB,$AF,$75,$FE,$73,$EC,$FF,$02,$57,$4F,$DF,$1B,$7F,$A0,$A3,$8B
	.dc.b	$73,$F0,$7B,$AD,$1D,$5E,$59,$CC,$7F,$A6,$31,$EB,$F9,$F7,$F7,$CF
	.dc.b	$C6,$1F,$92,$FF,$38,$FC,$AB,$FD,$2F,$5A,$F8,$1F,$F6,$A4,$E6,$3D
	.dc.b	$57,$FE,$04,$D3,$C8,$FF,$08,$7F,$6F,$AA,$3B,$57,$F4,$4F,$89,$F9
	.dc.b	$11,$8F,$90,$FD,$D7,$93,$EE,$C3,$FD,$01,$F7,$1F,$E2,$0F,$DB,$1F
	.dc.b	$97,$FF,$B5,$3F,$18,$78,$1F,$2E,$B5,$39,$F0,$71,$F4,$7C,$F1,$FE
	.dc.b	$FD,$BA,$DF,$AA,$E6,$7E,$D4,$72,$FA,$FF,$F1,$54,$72,$D3,$C3,$0F
	.dc.b	$D9,$FF,$CA,$9F,$9F,$3F,$80,$5F,$AD,$3F,$2A,$7E,$ED,$FC,$03,$F3
	.dc.b	$FF,$E6,$DD,$3F,$E3,$FC,$38,$5E,$EB,$5D,$5F,$E8,$39,$CE,$BF,$EC
	.dc.b	$07,$B7,$DD,$79,$E3,$59,$F8,$33,$EF,$F0,$5F,$E7,$39,$CE,$5D,$61
	.dc.b	$F2,$3F,$E6,$BF,$CE,$FF,$BA,$FB,$CF,$D3,$9F,$AD,$3F,$47,$7C,$EF
	.dc.b	$E6,$38,$BF,$CA,$D8,$70,$BD,$D6,$BF,$70,$FF,$4D,$4F,$F8,$03,$C8
	.dc.b	$F3,$5F,$0F,$0B,$FD,$93,$BF,$FB,$6A,$B9,$67,$DF,$E0,$BC,$95,$FE
	.dc.b	$CE,$CF,$AF,$C4,$79,$FF,$64,$FC,$DF,$D9,$3E,$EF,$FC,$7D,$FA,$FF
	.dc.b	$F8,$D7,$5F,$2D,$6A,$73,$FB,$FF,$92,$3F,$47,$B7,$85,$7E,$91,$E6
	.dc.b	$F2,$BF,$21,$C0,$9E,$5F,$74,$FA,$AF,$17,$A7,$FE,$43,$CF,$FC,$9D
	.dc.b	$F5,$9F,$E9,$5F,$C0,$9F,$BD,$3F,$3E,$FF,$49,$7D,$53,$E1,$FB,$FE
	.dc.b	$F9,$FE,$BF,$BD,$D6,$A7,$C7,$F6,$B9,$C3,$97,$17,$FC,$00,$BE,$BF
	.dc.b	$94,$F9,$E3,$F8,$1C,$72,$FC,$3F,$FC,$2D,$87,$BB,$9B,$FD,$19,$39
	.dc.b	$D3,$C5,$F2,$7F,$EA,$BF,$C3,$1F,$B3,$BF,$33,$7F,$00,$3F,$64,$FE
	.dc.b	$37,$F4,$79,$CF,$FD,$2D,$87,$0B,$DD,$6B,$E3,$7F,$DB,$53,$9D,$FF
	.dc.b	$59,$7C,$32,$FE,$5B,$1F,$DE,$A7,$96,$5D,$FC,$6F,$F2,$FC,$FF,$1E
	.dc.b	$9C,$E3,$FC,$B3,$C2,$FF,$61,$6A,$FC,$9B,$FC,$ED,$F9,$83,$D2,$FE
	.dc.b	$33,$FB,$5B,$C5,$FF,$76,$5E,$FF,$04,$5E,$FE,$19,$7F,$77,$87,$FD
	.dc.b	$65,$7B,$DB,$D3,$E2,$FF,$13,$8F,$07,$47,$D1,$94,$75,$C7,$A7,$A9
	.dc.b	$E6,$DB,$CB,$B7,$F2,$CF,$D9,$FF,$DB,$5F,$9F,$3F,$23,$7E,$C7,$FC
	.dc.b	$DB,$F0,$5F,$F4,$B7,$D9,$BE,$DD,$C3,$FE,$C5,$C3,$85,$EF,$F0,$BF
	.dc.b	$F2,$97,$D8,$6B,$8F,$F9,$45,$75,$FB,$8F,$33,$C3,$8F,$D6,$7F,$4A
	.dc.b	$79,$F9,$8D,$7F,$D4,$DF,$21,$E6,$DF,$CB,$39,$CF,$FD,$6F,$7C,$2F
	.dc.b	$CF,$F8,$4F,$87,$1F,$EA,$BF,$A9,$3F,$29,$FF,$40,$79,$0F,$CD,$7F
	.dc.b	$D5,$1F,$59,$F5,$78,$FF,$DC,$17,$BF,$BD,$FF,$6B,$56,$BD,$4E,$99
	.dc.b	$2B,$A7,$DD,$FD,$17,$84,$7B,$0D,$67,$E0,$BE,$F1,$CA,$7B,$67,$E3
	.dc.b	$FB,$97,$AB,$FD,$B3,$4F,$F7,$97,$CF,$7F,$98,$BF,$38,$7E,$FC,$FD
	.dc.b	$CD,$FA,$23,$E7,$FF,$C4,$38,$70,$FA,$16,$1C,$2F,$75,$FD,$CA,$B5
	.dc.b	$3F,$89,$C7,$FC,$DF,$F9,$C5,$C3,$8E,$5D,$F2,$7E,$9B,$E5,$EF,$FE
	.dc.b	$C9,$87,$96,$B5,$E2,$67,$8B,$FC,$0B,$DA,$F9,$B7,$C9,$FE,$2B,$F0
	.dc.b	$C7,$E3,$AF,$C6,$FF,$7F,$C4,$D7,$FB,$86,$3C,$70,$E1,$7B,$F9,$3F
	.dc.b	$A7,$3D,$72,$FD,$13,$BC,$FE,$93,$C7,$1F,$F5,$FD,$B4,$F3,$57,$F8
	.dc.b	$52,$FC,$B5,$AF,$37,$38,$71,$D7,$CF,$AB,$F7,$8F,$C1,$FE,$60,$FD
	.dc.b	$BF,$F7,$30,$FD,$2D,$E3,$FF,$88,$72,$DD,$6A,$73,$F5,$4F,$87,$CF
	.dc.b	$AF,$FB,$BC,$BA,$7C,$A5,$C3,$86,$4F,$D5,$FD,$35,$E1,$BF,$26,$5E
	.dc.b	$0F,$D4,$F3,$87,$0C,$DF,$A8,$FE,$6B,$E3,$1F,$B2,$3E,$F3,$FA,$3F
	.dc.b	$E8,$EF,$D6,$7F,$C8,$DF,$9D,$6E,$B5,$F2,$2A,$D7,$84,$FD,$23,$58
	.dc.b	$FF,$13,$CB,$AF,$B9,$71,$E3,$87,$5E,$BF,$A3,$5E,$2B,$F0,$E9,$FE
	.dc.b	$43,$F0,$FF,$C6,$1F,$0A,$FC,$9D,$F9,$C3,$EC,$FF,$DC,$5F,$5F,$FE
	.dc.b	$8B,$F8,$FF,$DA,$DE,$EB,$5F,$29,$FE,$2C,$FC,$1F,$EC,$78,$BF,$8D
	.dc.b	$F1,$FF,$CB,$37,$D7,$F0,$5F,$E5,$4F,$18,$FD,$D3,$CF,$3E,$FF,$96
	.dc.b	$FE,$22,$31,$8F,$F8,$E3,$F2,$0F,$C5,$FE,$24,$FC,$49,$FA,$8B,$F4
	.dc.b	$67,$E4,$4F,$CE,$1F,$12,$FD,$6F,$F8,$6E,$3C,$70,$E1,$7B,$AF,$F4
	.dc.b	$FF,$0F,$A7,$83,$FD,$93,$9F,$67,$F8,$86,$1E,$1F,$BD,$FE,$88,$F3
	.dc.b	$D3,$C3,$A7,$8E,$BB,$DF,$9A,$73,$F7,$7F,$A4,$F8,$C7,$EA,$CF,$DD
	.dc.b	$9F,$A2,$BF,$1B,$FF,$44,$FE,$D0,$FC,$1B,$F9,$2E,$1C,$2F,$75,$AF
	.dc.b	$DB,$73,$8F,$3D,$7F,$93,$5F,$6F,$3F,$FA,$D3,$C7,$3E,$DF,$85,$7E
	.dc.b	$DE,$B8,$E5,$D2,$3F,$89,$C6,$27,$7A,$3F,$0A,$FE,$6F,$F6,$AF,$D5
	.dc.b	$5F,$96,$BF,$2A,$FF,$25,$F7,$7F,$A5,$7F,$11,$DD,$6A,$73,$EB,$3F
	.dc.b	$03,$CF,$6F,$F6,$FC,$BA,$65,$70,$E1,$F7,$EC,$FF,$6D,$BF,$4A,$F0
	.dc.b	$4E,$7D,$9F,$E2,$38,$72,$D3,$BF,$DB,$FF,$4E,$FC,$D9,$F9,$B3,$F7
	.dc.b	$57,$EC,$4F,$D2,$9F,$AC,$3F,$49,$7D,$E7,$1C,$38,$5E,$EB,$5F,$09
	.dc.b	$9C,$7F,$C1,$19,$76,$F7,$2E,$1C,$33,$EB,$F0,$DF,$C9,$AB,$C7,$7E
	.dc.b	$29,$F0,$F0,$66,$F8,$65,$E3,$F8,$9F,$EC,$FE,$B7,$F9,$4B,$E0,$5F
	.dc.b	$B5,$7D,$1F,$C7,$1F,$B6,$BE,$FF,$7F,$0F,$FC,$2E,$B5,$3F,$E5,$9F
	.dc.b	$4F,$F4,$18,$F3,$EC,$FE,$DC,$FF,$97,$BE,$23,$FC,$44,$F7,$BE,$DF
	.dc.b	$49,$FC,$FA,$31,$D6,$D3,$F4,$58,$F4,$CF,$C1,$39,$CF,$FD,$D3,$3F
	.dc.b	$AF,$FB,$0F,$F6,$9F,$ED,$2F,$D9,$9F,$9D,$7E,$F3,$F1,$A7,$ED,$1F
	.dc.b	$E0,$6B,$DD,$6B,$DC,$53,$F3,$1D,$7D,$DF,$DB,$E4,$FF,$04,$61,$D9
	.dc.b	$B0,$CF,$C1,$9F,$F0,$AF,$B3,$FE,$B3,$7D,$A3,$F8,$6E,$78,$BE,$0F
	.dc.b	$87,$27,$EC,$DE,$0C,$79,$E9,$E4,$F8,$AF,$F0,$EF,$EA,$DF,$5B,$F6
	.dc.b	$E7,$EC,$6F,$BB,$FE,$81,$F8,$1B,$F4,$6D,$F6,$2F,$75,$AF,$D2,$73
	.dc.b	$1F,$E5,$5C,$7F,$03,$E3,$DB,$E1,$97,$5F,$27,$FA,$07,$30,$5F,$8E
	.dc.b	$7C,$5F,$AE,$FF,$BC,$FA,$BF,$91,$74,$FD,$3F,$D3,$7F,$1B,$FB,$9B
	.dc.b	$F2,$AF,$EA,$DF,$A1,$FF,$8E,$2B,$5A,$AB,$5A,$FF,$93,$F1,$E7,$D2
	.dc.b	$CB,$FC,$D9,$87,$DF,$FC,$F2,$CE,$71,$E5,$9F,$F6,$3D,$FB,$4E,$73
	.dc.b	$F1,$7E,$9B,$FE,$0E,$C7,$C3,$1E,$0F,$D9,$FF,$C9,$7D,$8F,$C0,$9F
	.dc.b	$AB,$BF,$7E,$7E,$80,$FD,$43,$FA,$37,$B7,$C2,$F7,$5A,$9F,$E9,$5F
	.dc.b	$05,$9C,$3F,$D0,$98,$F5,$F9,$1B,$87,$0C,$BA,$7D,$BF,$F8,$8A,$BC
	.dc.b	$57,$E1,$9F,$07,$3E,$6F,$86,$5F,$D5,$FE,$23,$FC,$C7,$FB,$87,$F8
	.dc.b	$67,$F7,$47,$EA,$8F,$CE,$DF,$94,$FF,$A3,$3E,$23,$7E,$DA,$F7,$5A
	.dc.b	$FA,$CF,$F0,$E7,$81,$FD,$77,$2E,$9D,$9F,$E6,$B7,$FC,$4D,$FA,$DF
	.dc.b	$FA,$7D,$F0,$69,$DF,$47,$EB,$3F,$E3,$C9,$CE,$5E,$2F,$CD,$B1,$AF
	.dc.b	$87,$0E,$1A,$7F,$5C,$D3,$FB,$67,$CC,$3F,$21,$7E,$F9,$FD,$8B,$F1
	.dc.b	$5F,$E8,$5F,$BC,$BE,$53,$FD,$07,$7B,$AD,$77,$2B,$D7,$F5,$DF,$9B
	.dc.b	$E4,$5F,$5F,$A8,$FF,$5C,$3C,$63,$FD,$10,$F3,$F9,$EF,$F6,$66,$1E
	.dc.b	$C9,$EF,$A7,$92,$73,$F1,$7F,$E5,$AB,$E1,$AF,$9B,$A7,$FD,$6F,$F7
	.dc.b	$DE,$1F,$89,$FF,$5D,$7A,$FF,$C8,$7E,$91,$FA,$1B,$F1,$6F,$C8,$72
	.dc.b	$FF,$9D,$F7,$FD,$96,$1C,$3E,$27,$FC,$31,$7B,$FD,$D7,$AD,$6B,$F1
	.dc.b	$87,$ED,$BE,$F8,$FD,$B6,$BE,$19,$FE,$15,$1F,$EF,$28,$C6,$BE,$3C
	.dc.b	$1F,$A6,$7F,$A8,$67,$39,$78,$3E,$27,$18,$77,$C3,$86,$DF,$DF,$76
	.dc.b	$F4,$4F,$F4,$BC,$FF,$8C,$61,$E8,$AF,$F6,$15,$6B,$6F,$66,$FF,$0A
	.dc.b	$7F,$A9,$CF,$2F,$C8,$03,$F1,$EF,$F1,$07,$EE,$6F,$6F,$FD,$61,$5F
	.dc.b	$0B,$4D,$DE,$FE,$A9,$7B,$FC,$45,$FB,$16,$EB,$AF,$E6,$5F,$1F,$0F
	.dc.b	$69,$CB,$3C,$3B,$9E,$98,$78,$7D,$5D,$C7,$9A,$BC,$BA,$7F,$89,$F4
	.dc.b	$F5,$4F,$08,$FF,$63,$7C,$D7,$9F,$F7,$F7,$E9,$EF,$D9,$1F,$0D,$FE
	.dc.b	$7B,$CD,$FA,$5F,$FC,$01,$CF,$E6,$E1,$C2,$F7,$F3,$BF,$EC,$4A,$FC
	.dc.b	$55,$6B,$1E,$9B,$FD,$F3,$F9,$E6,$1E,$0F,$C9,$3D,$B5,$BC,$79,$E5
	.dc.b	$E0,$BF,$36,$1E,$5C,$FC,$9F,$69,$F1,$B2,$F1,$47,$9F,$3F,$E5,$59
	.dc.b	$7F,$B3,$B0,$FE,$95,$F8,$E7,$FA,$23,$F1,$D7,$E1,$AF,$C8,$BF,$D3
	.dc.b	$1F,$85,$BF,$45,$63,$C7,$16,$3C,$70,$E1,$C5,$FE,$AA,$FD,$6F,$7F
	.dc.b	$8C,$7E,$F6,$BB,$5F,$DB,$C3,$94,$7F,$A6,$3F,$11,$F7,$D6,$FE,$99
	.dc.b	$B7,$93,$6F,$F5,$57,$E2,$5F,$F0,$06,$B5,$39,$D1,$F5,$19,$D3,$DB
	.dc.b	$8F,$2D,$BC,$BC,$0F,$D0,$FC,$B5,$AC,$79,$4B,$F6,$DC,$CF,$BB,$0E
	.dc.b	$5A,$FC,$4F,$7F,$3F,$8C,$BF,$63,$FE,$A5,$F9,$4B,$F5,$9F,$F6,$97
	.dc.b	$EB,$7F,$C2,$B9,$BE,$45,$EF,$E3,$17,$BF,$B4,$5B,$F6,$4E,$B9,$7F
	.dc.b	$46,$DB,$E4,$73,$7F,$27,$CB,$94,$60,$FF,$02,$3D,$71,$E9,$F3,$7F
	.dc.b	$0E,$53,$9D,$7F,$DC,$DE,$A6,$63,$1A,$78,$BF,$5F,$FB,$FF,$9C,$7F
	.dc.b	$35,$F8,$4F,$F1,$2F,$96,$FC,$CD,$FB,$4B,$F4,$2F,$F1,$FF,$E5,$FC
	.dc.b	$36,$FF,$0B,$61,$C2,$F7,$5F,$E4,$4F,$DB,$F5,$4A,$EA,$FE,$17,$2F
	.dc.b	$07,$D8,$BF,$32,$FD,$1C,$63,$4F,$16,$9E,$1E,$2F,$F1,$27,$4A,$B5
	.dc.b	$AF,$83,$8B,$31,$FD,$3F,$F1,$7F,$F6,$6F,$E3,$1D,$7C,$8F,$B1,$F8
	.dc.b	$DB,$8F,$CB,$5A,$DB,$F3,$FC,$FF,$F0,$07,$F3,$4C,$FE,$15,$FE,$C2
	.dc.b	$E3,$F4,$1F,$E3,$5B,$EA,$E0,$3C,$F2,$FC,$37,$83,$E4,$9E,$14,$D3
	.dc.b	$C9,$8B,$F4,$6D,$53,$F4,$97,$E7,$EF,$D3,$7F,$CA,$5E,$CB,$F2,$67
	.dc.b	$ED,$AF,$CC,$7E,$7E,$1C,$2F,$75,$AF,$C4,$33,$8F,$F0,$1C,$BD,$17
	.dc.b	$B6,$C7,$7F,$8B,$67,$DF,$F5,$8F,$2D,$F5,$9C,$E2,$FC,$EF,$38,$71
	.dc.b	$C3,$CD,$AB,$F3,$4F,$E1,$BC,$9F,$E0,$3F,$D5,$BD,$DC,$DF,$C4,$3F
	.dc.b	$4D,$FF,$02,$FE,$4A,$FC,$6F,$7C,$3F,$E6,$3B,$DF,$82,$BD,$FE,$91
	.dc.b	$AA,$57,$C9,$FD,$B7,$F9,$6E,$7D,$F8,$FE,$AD,$7F,$15,$8C,$69,$E1
	.dc.b	$F9,$C6,$BF,$8E,$9C,$E9,$DE,$5F,$93,$FE,$4B,$F0,$6F,$57,$E8,$1F
	.dc.b	$E8,$CC,$5F,$58,$FD,$2D,$FA,$3B,$F5,$3F,$E5,$B8,$70,$F3,$FF,$0F
	.dc.b	$7B,$A7,$E5,$7A,$E0,$F8,$19,$FF,$A6,$1F,$E5,$98,$71,$CF,$BE,$6F
	.dc.b	$CC,$7C,$95,$C3,$3E,$B2,$FD,$0F,$F5,$27,$D1,$E1,$FA,$3F,$01,$FE
	.dc.b	$57,$FE,$00,$F4,$F1,$EF,$1F,$A1,$5A,$FB,$35,$3F,$5C,$FE,$8F,$8F
	.dc.b	$3E,$7F,$D0,$E5,$FE,$61,$F2,$3B,$E5,$CA,$73,$8F,$CA,$8F,$E9,$FF
	.dc.b	$89,$FE,$0F,$D8,$FD,$06,$FF,$AA,$F6,$1F,$A6,$FF,$6C,$AD,$63,$F3
	.dc.b	$E7,$3F,$89,$F2,$C7,$F0,$D8,$F2,$E4,$FF,$0C,$E3,$FC,$23,$1E,$39
	.dc.b	$75,$F1,$FF,$85,$FC,$7F,$E4,$FC,$1F,$D8,$6B,$FB,$AC,$7F,$C2,$FF
	.dc.b	$05,$FD,$92,$B5,$3F,$98,$9F,$F0,$DF,$B1,$9D,$7F,$2D,$87,$29,$FE
	.dc.b	$07,$E0,$76,$9C,$DF,$1F,$0B,$F8,$DF,$C5,$78,$7B,$BD,$18,$7C,$DE
	.dc.b	$BF,$EC,$B8,$74,$9C,$F8,$3F,$DD,$AF,$86,$3F,$91,$F5,$79,$DE,$E3
	.dc.b	$F2,$9D,$1F,$CA,$3F,$4D,$F7,$77,$79,$74,$FE,$2F,$A2,$73,$7E,$DE
	.dc.b	$BF,$F0,$1C,$62,$3F,$8E,$E5,$CF,$2F,$7E,$3E,$FC,$BF,$9A,$FE,$71
	.dc.b	$F2,$FA,$7E,$5F,$3F,$F2,$4F,$1F,$FC,$1B,$39,$CB,$DF,$AA,$31,$1F
	.dc.b	$EF,$6C,$F9,$E7,$F3,$31,$E7,$97,$F4,$2F,$DA,$FE,$C7,$2F,$D8,$C3
	.dc.b	$B6,$49,$CC,$7B,$E3,$18,$7D,$6A,$E1,$8F,$38,$C5,$7E,$0B,$0F,$E9
	.dc.b	$FE,$37,$6E,$4F,$5E,$BF,$E2,$68,$C5,$7E,$3F,$2F,$D9,$61,$D7,$7F
	.dc.b	$25,$FF,$8B,$3C,$4F,$3C,$79,$F8,$FF,$C5,$FE,$07,$AB,$83,$D5,$C4
	.dc.b	$FE,$00,$FF,$65,$65,$E7,$9C,$C7,$68,$FE,$3F,$C3,$FC,$83,$1E,$F3
	.dc.b	$98,$CC,$77,$8F,$E2,$73,$F0,$77,$F5,$C6,$23,$9D,$27,$EC,$E5,$E3
	.dc.b	$E2,$F9,$7B,$FF,$38,$DB,$18,$7C,$88,$FD,$26,$3F,$3A,$39,$C7,$EB
	.dc.b	$38,$91,$C0,$CC,$7F,$5F,$E2,$CD,$7E,$0E,$7F,$01,$3F,$BB,$BF,$89
	.dc.b	$C7,$F8,$6D,$FF,$6D,$7B,$BF,$87,$BF,$E4,$32,$E7,$7B,$CB,$AF,$27
	.dc.b	$E6,$F5,$E2,$D5,$CB,$F4,$63,$57,$C3,$1F,$CF,$EB,$FD,$63,$0F,$3F
	.dc.b	$37,$EC,$B5,$FE,$11,$E1,$FD,$98,$C6,$7F,$CF,$B6,$FD,$DE,$3F,$63
	.dc.b	$87,$F8,$B6,$7C,$A3,$F7,$18,$FF,$82,$B2,$F8,$7D,$9F,$C4,$34,$FE
	.dc.b	$29,$F4,$4E,$EF,$C3,$4E,$74,$FF,$08,$CF,$F8,$97,$44,$E6,$7F,$98
	.dc.b	$46,$32,$FE,$27,$87,$F1,$FE,$CF,$A3,$E0,$7F,$26,$BF,$E5,$3F,$4B
	.dc.b	$F9,$38,$C7,$89,$FC,$C3,$14,$63,$C3,$E3,$97,$F3,$2C,$BF,$9B,$78
	.dc.b	$7F,$73,$C5,$FE,$8D,$9F,$CE,$EE,$FE,$97,$E1,$7E,$7E,$73,$D3,$FD
	.dc.b	$4F,$A3,$19,$7F,$49,$C3,$87,$27,$38,$C5,$31,$F6,$F9,$1F,$D9,$72
	.dc.b	$FE,$D9,$E5,$7E,$AE,$31,$E4,$FF,$74,$F2,$B1,$8F,$F6,$73,$FD,$A5
	.dc.b	$C4,$AD,$F8,$9F,$94,$78,$EF,$F9,$6F,$13,$FC,$0B,$E6,$FF,$86,$3C
	.dc.b	$EF,$DC,$46,$3C,$DF,$F0,$5F,$9D,$8C,$FF,$C3,$5C,$CF,$1B,$FD,$BB
	.dc.b	$AF,$FA,$BA,$73,$CB,$8C,$79,$76,$77,$9C,$F9,$3F,$E1,$3F,$2F,$F4
	.dc.b	$FE,$67,$EB,$27,$FC,$91,$E9,$7F,$8C,$7D,$3F,$F2,$2F,$A3,$FC,$2E
	.dc.b	$B5,$B7,$F8,$E7,$D3,$FF,$1A,$FB,$19,$C7,$F6,$B3,$FE,$50,$F2,$7C
	.dc.b	$B7,$C3,$2F,$F2,$95,$7C,$68,$C7,$B1,$FE,$4E,$F5,$FC,$73,$CA,$BC
	.dc.b	$5E,$1E,$23,$CB,$5F,$E5,$4F,$59,$EC,$78,$67,$FC,$D3,$ED,$FF,$99
	.dc.b	$7D,$DF,$F3,$9F,$BF,$FE,$77,$E8,$FF,$40,$57,$F9,$AB,$DD,$FF,$34
	.dc.b	$CB,$DE,$E5,$D4,$C7,$FC,$F1,$7F,$14,$CB,$FE,$71,$7D,$47,$3C,$7F
	.dc.b	$D0,$3C,$3E,$7A,$FF,$3D,$7B,$DF,$E7,$5A,$F3,$7B,$1F,$E9,$FF,$83
	.dc.b	$FF,$52,$7E,$33,$E5,$E5,$FF,$49,$C6,$3E,$17,$7E,$BB,$D3,$F9,$78
	.dc.b	$FF,$AA,$23,$FD,$40,$F8,$8F,$F5,$33,$BF,$84,$FF,$55,$FC,$2F,$C9
	.dc.b	$CF,$C1,$F0,$CF,$86,$FE,$4F,$ED,$FF,$B1,$FE,$2F,$FD,$AB,$C1,$FE
	.dc.b	$D4,$F8,$CF,$E9,$53,$9F,$8E,$FF,$69,$CF,$FB,$63,$E2,$B3,$B7,$FB
	.dc.b	$33,$0F,$F6,$93,$F1,$FA,$FF,$DB,$70,$BF,$CB,$E1,$CF,$E3,$7F,$DA
	.dc.b	$37,$C8,$FF,$68,$D7,$8B,$E3,$3C,$D1,$9F,$8D,$FF,$68,$FC,$73,$EB
	.dc.b	$78,$79,$FF,$CC,$3E,$1B,$FD,$EB,$F2,$38,$F7,$FF,$DE,$7F,$2C,$F9
	.dc.b	$37,$C9,$E6,$FC,$7A,$7F,$4C,$C7,$F9,$CF,$C4,$62,$3F,$DF,$92,$F9
	.dc.b	$57,$88,$C9,$F2,$FF,$2A,$F9,$69,$E0,$F9,$7C,$47,$2C,$38,$DF,$8F
	.dc.b	$E5,$FC,$51,$8D,$9F,$33,$8C,$FB,$43,$E6,$7F,$BE,$7D,$67,$ED,$FE
	.dc.b	$75,$F3,$EF,$A1,$7C,$FF,$97,$E8,$5F,$40,$FC,$0B,$8F,$0F,$F6,$AC
	.dc.b	$78,$E1,$C3,$85,$87,$0F,$AD,$DD,$AF,$6F,$F1,$ED,$3C,$5A,$7F,$AF
	.dc.b	$A3,$07,$57,$C9,$A7,$8F,$4F,$EE,$B5,$C3,$3F,$06,$4F,$A3,$CE,$3C
	.dc.b	$67,$36,$F9,$8F,$1F,$9F,$FA,$3F,$A6,$7D,$53,$EA,$1F,$50,$FA,$A7
	.dc.b	$CF,$70,$C1,$87,$0B,$DD,$3E,$A7,$5F,$55,$FC,$B3,$87,$F9,$26,$1D
	.dc.b	$F7,$F6,$9F,$B3,$79,$E9,$E1,$D3,$E4,$CF,$7C,$FB,$43,$F3,$FF,$6C
	.dc.b	$F1,$9C,$E6,$F8,$FF,$9F,$F5,$AE,$87,$D5,$BE,$B9,$F5,$DF,$A1,$E4
	.dc.b	$78,$9C,$2F,$75,$AF,$33,$E4,$7E,$3B,$9F,$B1,$F6,$6F,$FE,$2A,$D3
	.dc.b	$B6,$98,$84,$63,$0E,$5F,$63,$E8,$8C,$6B,$FB,$1F,$6F,$1A,$78,$23
	.dc.b	$C9,$F9,$57,$8F,$2F,$44,$79,$FF,$0B,$C6,$5E,$5E,$07,$DA,$7F,$95
	.dc.b	$FE,$03,$FA,$97,$CE,$BE,$E5,$F7,$AF,$BA,$7D,$CB,$EE,$DF,$7A,$F7
	.dc.b	$B9,$73,$7F,$91,$B2,$E5,$8F,$1C,$38,$5B,$EE,$F7,$6B,$F0,$7D,$EF
	.dc.b	$F6,$CD,$7C,$7A,$F8,$AB,$FB,$EC,$63,$2E,$9A,$3F,$02,$FF,$28,$CE
	.dc.b	$76,$F3,$CB,$EF,$B3,$87,$97,$1F,$6D,$6B,$57,$DF,$F9,$3E,$C5,$F8
	.dc.b	$13,$F0,$57,$E0,$7F,$E1,$7F,$C1,$1F,$82,$7A,$FE,$39,$C1,$FD,$EF
	.dc.b	$1E,$38,$70,$FB,$76,$1C,$3F,$2D,$ED,$9F,$5E,$0F,$69,$FE,$6E,$C3
	.dc.b	$C3,$F8,$37,$C3,$8C,$4F,$8F,$5F,$17,$27,$B2,$57,$E0,$9F,$37,$C5
	.dc.b	$66,$31,$9F,$8B,$7F,$B1,$BF,$A3,$0F,$F3,$EF,$E1,$7F,$C5,$FF,$0B
	.dc.b	$78,$2F,$C2,$FF,$A5,$7D,$D3,$F0,$AF,$B3,$7B,$AD,$7C,$B2,$9F,$86
	.dc.b	$EB,$E1,$99,$7A,$F3,$63,$FC,$97,$F0,$DC,$3C,$F4,$FE,$B1,$F0,$7E
	.dc.b	$D8,$C6,$6F,$C3,$71,$97,$4D,$7C,$3A,$FE,$57,$7F,$66,$8F,$C4,$1F
	.dc.b	$3E,$FC,$51,$F8,$A3,$F1,$27,$E2,$9D,$7F,$15,$FC,$AE,$1C,$2F,$75
	.dc.b	$AD,$9F,$8A,$6B,$2E,$99,$FD,$A3,$C0,$FF,$24,$71,$87,$3C,$FC,$19
	.dc.b	$FA,$E7,$96,$5D,$A1,$F8,$AF,$DD,$BE,$19,$F8,$7F,$2F,$FD,$4F,$E3
	.dc.b	$1E,$AF,$C6,$9F,$8C,$FF,$18,$DF,$FC,$C5,$F8,$DE,$B6,$56,$BD,$AF
	.dc.b	$95,$F8,$53,$F1,$CD,$4F,$D5,$CB,$E2,$FE,$39,$70,$E3,$9F,$5F,$C3
	.dc.b	$3F,$21,$5E,$18,$FE,$4F,$F2,$FF,$E5,$BE,$8F,$A3,$F8,$FF,$FA,$3B
	.dc.b	$F1,$FF,$A3,$F6,$EF,$C7,$3F,$03,$F6,$BF,$C0,$AB,$5D,$6A,$D7,$84
	.dc.b	$FC,$7F,$5F,$8B,$BF,$25,$FC,$56,$1D,$7B,$FE,$83,$CB,$0E,$DF,$15
	.dc.b	$FA,$58,$C6,$9F,$82,$AE,$7A,$F9,$23,$F2,$7F,$90,$E3,$5F,$14,$7E
	.dc.b	$5B,$BB,$F8,$BF,$D5,$3F,$23,$FF,$09,$73,$BF,$23,$FE,$8D,$F2,$4F
	.dc.b	$CE,$B7,$5A,$9C,$FE,$3B,$8F,$A8,$FD,$16,$3C,$FF,$25,$B7,$BF,$90
	.dc.b	$FE,$9B,$C3,$BD,$3B,$43,$D3,$FE,$01,$87,$8F,$1F,$CD,$E7,$E2,$F9
	.dc.b	$4F,$E5,$BE,$0F,$F6,$DF,$C9,$DF,$94,$FC,$9F,$CA,$1F,$95,$FF,$1E
	.dc.b	$F0,$7F,$4F,$7B,$AD,$4B,$F2,$BE,$7F,$4E,$C7,$F7,$57,$FD,$E7,$CD
	.dc.b	$B3,$9C,$3B,$7C,$AF,$F1,$5A,$E9,$E0,$B4,$FE,$69,$C9,$D3,$2F,$2C
	.dc.b	$E7,$47,$E5,$7F,$D1,$BF,$2C,$7C,$DF,$FA,$43,$89,$F9,$6F,$F4,$BF
	.dc.b	$8F,$FE,$01,$F1,$FF,$EA,$BB,$DF,$E2,$1F,$56,$B5,$F0,$9C,$F1,$FF
	.dc.b	$47,$6D,$FB,$7B,$DB,$63,$B5,$7F,$57,$E1,$FE,$5D,$5D,$33,$F0,$FC
	.dc.b	$CF,$AB,$9F,$F8,$7C,$E6,$FC,$9F,$97,$7F,$5E,$F6,$7F,$B5,$7E,$60
	.dc.b	$F9,$1F,$F1,$4F,$E4,$3F,$CA,$FF,$32,$D7,$CB,$AB,$5F,$58,$AF,$E6
	.dc.b	$1F,$98,$EA,$7F,$95,$63,$FD,$D7,$F3,$2B,$87,$1C,$BA,$FD,$6F,$F4
	.dc.b	$5A,$F0,$C7,$83,$E3,$71,$5B,$CF,$B7,$C8,$7F,$73,$E1,$FE,$53,$F9
	.dc.b	$A3,$F3,$47,$E6,$DF,$91,$F5,$3F,$C4,$7F,$8B,$7F,$3F,$BD,$D6,$BC
	.dc.b	$DF,$DF,$7D,$4F,$FA,$EF,$F2,$BF,$EF,$F9,$73,$CB,$FA,$13,$D7,$F3
	.dc.b	$6B,$87,$1C,$FF,$C5,$F8,$FF,$5D,$AE,$0F,$4F,$BC,$FE,$C1,$F5,$0F
	.dc.b	$CE,$5E,$9B,$F3,$8F,$F0,$77,$C2,$3F,$1D,$7E,$7B,$FE,$10,$C3,$85
	.dc.b	$EF,$C3,$FE,$1D,$4F,$CE,$B5,$93,$F1,$CF,$EB,$18,$70,$F2,$7F,$9F
	.dc.b	$3C,$B4,$EF,$F9,$5F,$92,$FC,$71,$E2,$FC,$AF,$FB,$B7,$C0,$BF,$3C
	.dc.b	$7E,$79,$FC,$2F,$DB,$7E,$46,$FC,$DF,$FD,$3D,$DD,$FD,$6B,$0E,$17
	.dc.b	$BA,$D7,$C2,$7F,$8A,$A7,$3B,$FF,$51,$7B,$7D,$37,$F6,$F7,$8F,$C2
	.dc.b	$FF,$69,$B7,$E6,$5E,$4D,$7C,$73,$9F,$96,$FF,$06,$7E,$7A,$F3,$7F
	.dc.b	$67,$D6,$FC,$C1,$F8,$B7,$F8,$A3,$F1,$5F,$F1,$DD,$EF,$DF,$F9,$F5
	.dc.b	$AF,$7F,$FD,$85,$C9,$FD,$D2,$B5,$C5,$FD,$9D,$ED,$F8,$A7,$F8,$15
	.dc.b	$E5,$9F,$7C,$FF,$C9,$D5,$DB,$4F,$1E,$5E,$2F,$B8,$F3,$C6,$2B,$FC
	.dc.b	$77,$F6,$7F,$E4,$EE,$27,$E1,$2F,$C7,$7F,$D0,$1F,$9F,$BF,$1C,$7E
	.dc.b	$7F,$F8,$1E,$1F,$F0,$4E,$1C,$2F,$75,$AD,$FA,$63,$D3,$7F,$F0,$05
	.dc.b	$ED,$B1,$DF,$F0,$96,$7D,$FD,$BF,$F1,$C5,$F5,$9C,$E3,$FE,$DC,$C3
	.dc.b	$AE,$5E,$0D,$1F,$A0,$7F,$9A,$3C,$07,$E8,$6F,$CD,$1F,$A1,$3F,$44
	.dc.b	$7E,$89,$FE,$B3,$FB,$BE,$17,$BA,$D4,$FF,$AD,$F7,$CE,$1F,$E3,$5C
	.dc.b	$7A,$FE,$6D,$70,$E1,$97,$4F,$81,$FF,$28,$57,$8A,$FC,$33,$E0,$E2
	.dc.b	$FF,$2A,$7E,$43,$FE,$A6,$FD,$25,$F9,$1B,$F4,$87,$E9,$1F,$EF,$CF
	.dc.b	$CC,$FF,$D0,$77,$BA,$D4,$FF,$C0,$0F,$D1,$F3,$F8,$BB,$F4,$9F,$F2
	.dc.b	$55,$F5,$E9,$FF,$18,$3C,$72,$EB,$F9,$37,$8A,$DA,$78,$A7,$C3,$FA
	.dc.b	$4F,$F9,$A7,$C3,$7E,$6D,$FE,$7C,$E7,$7C,$7B,$E2,$FF,$D8,$7F,$8C
	.dc.b	$7F,$A6,$2F,$7E,$AF,$EA,$AB,$5F,$79,$FE,$FF,$F2,$3F,$CA,$35,$D3
	.dc.b	$F2,$EF,$F3,$03,$C3,$8B,$FC,$8C,$F2,$CF,$BE,$6F,$B3,$F2,$6B,$E3
	.dc.b	$CF,$C5,$F8,$D6,$35,$F0,$5F,$FB,$5B,$F1,$AF,$F5,$5F,$AD,$FE,$B6
	.dc.b	$F8,$27,$E7,$AF,$B6,$7D,$C7,$F6,$CE,$FD,$FE,$34,$BD,$D6,$BD,$A7
	.dc.b	$7E,$A7,$FC,$F1,$97,$FB,$0B,$F1,$FB,$87,$1C,$FB,$67,$F9,$9E,$7F
	.dc.b	$F3,$8D,$F8,$E7,$3E,$1F,$FA,$5F,$F0,$27,$E6,$8F,$D2,$FF,$98,$7E
	.dc.b	$46,$FB,$87,$E9,$99,$CF,$FC,$F7,$87,$0B,$DD,$6B,$F4,$CF,$F6,$5C
	.dc.b	$E7,$BB,$FC,$E4,$F6,$8F,$F4,$83,$C7,$E4,$19,$77,$FC,$FF,$C9,$A7
	.dc.b	$8E,$73,$8B,$F3,$EF,$F5,$77,$9B,$FE,$41,$F8,$A7,$E9,$BF,$ED,$BE
	.dc.b	$67,$E9,$9F,$E2,$BF,$B2,$70,$4C,$38,$5E,$EB,$FC,$59,$C5,$FE,$96
	.dc.b	$9E,$95,$FE,$89,$7F,$C9,$DF,$77,$FE,$A7,$78,$FC,$C3,$2F,$E7,$35
	.dc.b	$FE,$EA,$9C,$C3,$A3,$FD,$89,$F9,$F7,$FA,$CB,$C8,$7E,$9E,$FD,$3F
	.dc.b	$FD,$FD,$E1,$E3,$5F,$F5,$3D,$EE,$B5,$F9,$22,$9F,$A1,$FF,$BE,$2B
	.dc.b	$A6,$BF,$E9,$F7,$FD,$7F,$C1,$FE,$BD,$79,$64,$C3,$F1,$75,$C7,$ED
	.dc.b	$FF,$A8,$E5,$DB,$D2,$F3,$FE,$98,$FC,$BD,$F6,$CF,$93,$7E,$A1,$FE
	.dc.b	$40,$FB,$4F,$15,$EE,$B5,$3F,$E4,$4F,$D2,$33,$DD,$E3,$AE,$9A,$F9
	.dc.b	$DE,$18,$F4,$FD,$37,$C3,$9F,$82,$3B,$F7,$FF,$BB,$7E,$37,$FD,$F3
	.dc.b	$FA,$1B,$E6,$9F,$A8,$BF,$4C,$7E,$A5,$FE,$52,$D7,$F0,$38,$70,$BD
	.dc.b	$D6,$BE,$F1,$DB,$FE,$D6,$AE,$B7,$FE,$C9,$5F,$96,$7E,$E6,$30,$B3
	.dc.b	$F1,$67,$E1,$FC,$E7,$CD,$39,$CF,$BF,$CA,$7E,$4B,$2F,$F7,$1F,$CC
	.dc.b	$BE,$BD,$FA,$5B,$F4,$D7,$DB,$3F,$53,$FF,$15,$7E,$A9,$E6,$F3,$FF
	.dc.b	$37,$C3,$FD,$AB,$1E,$38,$70,$BF,$F1,$87,$EA,$BB,$AF,$D0,$E7,$DF
	.dc.b	$3F,$F8,$00,$F8,$7E,$D6,$E3,$CB,$4F,$07,$D7,$F9,$F4,$EF,$7C,$74
	.dc.b	$ED,$FA,$8E,$73,$EB,$FA,$33,$F3,$77,$E7,$BF,$C7,$BC,$17,$E3,$CF
	.dc.b	$D0,$3C,$98,$7F,$80,$B0,$E1,$7B,$A7,$D3,$EB,$45,$75,$BF,$F7,$E4
	.dc.b	$63,$BB,$D7,$3C,$BE,$15,$9F,$F3,$9B,$E5,$39,$8F,$F6,$87,$47,$D2
	.dc.b	$F1,$9F,$84,$3F,$56,$FE,$F7,$F6,$3F,$E4,$1F,$37,$C5,$5A,$9C,$FD
	.dc.b	$13,$E2,$B3,$F8,$57,$CC,$AE,$7A,$F9,$9E,$1F,$4A,$CB,$F8,$3E,$9E
	.dc.b	$1D,$3C,$13,$9F,$73,$E6,$60,$EA,$71,$3E,$21,$F8,$CB,$F5,$17,$EA
	.dc.b	$EF,$D2,$5E,$3F,$CE,$C7,$8E,$1C,$2F,$74,$F1,$B5,$93,$F2,$DF,$C6
	.dc.b	$E9,$F8,$EE,$0F,$13,$FC,$AB,$2E,$6E,$8E,$39,$75,$D3,$F8,$E4,$63
	.dc.b	$5F,$34,$3F,$38,$F6,$DF,$F9,$C3,$F5,$87,$EB,$4F,$C3,$3D,$DF,$65
	.dc.b	$FE,$8A,$D3,$C3,$9F,$C3,$FC,$01,$87,$0B,$DF,$E0,$BE,$F9,$CD,$79
	.dc.b	$6F,$8E,$5F,$E7,$EF,$D6,$BD,$13,$BC,$79,$6F,$F5,$2B,$F5,$A7,$97
	.dc.b	$4F,$24,$7F,$3F,$AF,$99,$F8,$3F,$F4,$1D,$9F,$5A,$FD,$15,$FA,$CB
	.dc.b	$D8,$DF,$C7,$FC,$1A,$D4,$BF,$5B,$E7,$F0,$5E,$4E,$1F,$3E,$7D,$F4
	.dc.b	$FA,$71,$8C,$79,$7C,$1F,$D5,$E0,$F3,$DF,$8F,$3F,$17,$7F,$CE,$FD
	.dc.b	$77,$F9,$27,$E1,$CF,$D7,$5F,$4D,$FE,$40,$FD,$67,$F0,$D5,$A9,$CF
	.dc.b	$DB,$FF,$80,$36,$F9,$33,$FE,$D1,$DF,$E6,$6B,$F1,$A3,$F9,$84,$60
	.dc.b	$FC,$33,$CF,$17,$CF,$78,$FE,$1D,$A7,$8E,$7F,$BE,$5F,$F0,$4E,$47
	.dc.b	$EB,$CF,$D3,$DF,$65,$F3,$3E,$E7,$FD,$0D,$FA,$0E,$EB,$53,$9F,$C0
	.dc.b	$9F,$AE,$67,$ED,$FE,$67,$0F,$F2,$8C,$FA,$E7,$8D,$3F,$C0,$B1,$8C
	.dc.b	$7B,$7E,$5F,$FB,$28,$C1,$FE,$5F,$FC,$6F,$1F,$07,$FB,$4F,$CB,$DF
	.dc.b	$77,$E7,$FB,$D7,$E9,$DE,$BA,$D4,$E7,$F3,$DC,$7B,$1E,$1D,$7F,$0F
	.dc.b	$5F,$BE,$7E,$51,$8F,$51,$97,$6D,$FF,$01,$3E,$8A,$EB,$1F,$E9,$48
	.dc.b	$F9,$5F,$91,$3F,$61,$7E,$C2,$FA,$B7,$EC,$2E,$6F,$DA,$78,$3F,$AE
	.dc.b	$E4,$63,$C7,$0E,$17,$BF,$D5,$FA,$F8,$3E,$9B,$DE,$77,$0E,$19,$F6
	.dc.b	$FC,$0B,$E5,$46,$31,$7E,$01,$F3,$7F,$62,$FE,$6F,$D1,$FE,$B4,$ED
	.dc.b	$7E,$50,$FC,$87,$F0,$95,$AF,$C4,$7E,$EC,$FE,$6B,$8B,$D1,$3F,$D0
	.dc.b	$39,$3E,$4F,$02,$1F,$85,$79,$DE,$59,$3E,$AB,$C5,$F1,$FF,$E9,$4C
	.dc.b	$FF,$13,$F8,$0E,$36,$F2,$6B,$F9,$3D,$7E,$37,$B3,$F8,$EF,$C4,$DF
	.dc.b	$B1,$FE,$5B,$F3,$E7,$E8,$BF,$9B,$AD,$4E,$7F,$0A,$C6,$FF,$26,$5E
	.dc.b	$6D,$86,$3D,$3C,$A7,$E4,$5F,$93,$BE,$70,$FA,$0F,$CB,$79,$3F,$E8
	.dc.b	$9F,$D2,$3F,$C9,$BF,$97,$FF,$AF,$3F,$35,$FD,$64,$E7,$ED,$53,$FE
	.dc.b	$C0,$FD,$77,$3F,$03,$F6,$EB,$FD,$25,$DD,$F2,$9E,$15,$FB,$A8,$E7
	.dc.b	$F8,$5B,$2E,$FF,$67,$FA,$7F,$AB,$7E,$C9,$FE,$B7,$FD,$95,$FA,$EB
	.dc.b	$EF,$9F,$B3,$7E,$FE,$F7,$F8,$92,$F7,$5F,$CA,$7B,$BF,$29,$5D,$35
	.dc.b	$FB,$AF,$F9,$6D,$B9,$77,$CB,$FC,$E1,$1D,$63,$C5,$FA,$17,$EA,$23
	.dc.b	$F9,$AF,$E5,$6F,$D7,$7F,$C5,$D8,$7F,$86,$7F,$03,$FE,$4D,$97,$E8
	.dc.b	$EB,$5F,$6D,$FE,$36,$F6,$7F,$CF,$1F,$90,$FF,$17,$9F,$EE,$19,$7E
	.dc.b	$41,$E8,$DC,$BA,$D7,$F6,$4C,$3A,$46,$3D,$0F,$E1,$DF,$8B,$3E,$13
	.dc.b	$FD,$EB,$FB,$03,$F3,$EF,$F1,$1E,$BF,$C9,$6B,$5F,$86,$2B,$5F,$66
	.dc.b	$FC,$FB,$5C,$5F,$B5,$BE,$9B,$FE,$91,$E3,$7D,$7C,$3F,$0E,$9E,$0D
	.dc.b	$3F,$1F,$7F,$CC,$FE,$81,$FA,$53,$F5,$0F,$F0,$A7,$E2,$CF,$D9,$FF
	.dc.b	$D6,$3C,$9F,$C2,$6F,$75,$A9,$FF,$5C,$4E,$7F,$4C,$FE,$11,$5D,$31
	.dc.b	$FD,$BB,$C3,$E3,$D8,$BE,$3B,$F9,$94,$E6,$3C,$B8,$7F,$17,$F4,$DF
	.dc.b	$B4,$7F,$8E,$BF,$4F,$7E,$AB,$FD,$C7,$F3,$BF,$E9,$95,$AF,$B3,$56
	.dc.b	$BE,$91,$F2,$7A,$FC,$EF,$F8,$E5,$F4,$E0,$FE,$16,$F1,$BE,$11,$D3
	.dc.b	$3F,$06,$7F,$E7,$B8,$FE,$B5,$FB,$4B,$F2,$57,$ED,$5F,$E3,$BF,$DA
	.dc.b	$FF,$C7,$9F,$25,$A9,$CF,$2A,$7F,$C8,$1E,$DE,$71,$FE,$37,$8B,$F3
	.dc.b	$CF,$E6,$0F,$5C,$B7,$3F,$D5,$E3,$18,$73,$FA,$5F,$EB,$5F,$0E,$FD
	.dc.b	$67,$FC,$8F,$FB,$2B,$E9,$1F,$82,$FC,$8E,$AF,$E8,$57,$BF,$B6,$5E
	.dc.b	$EB,$FA,$2D,$6B,$0F,$E8,$56,$FB,$4B,$1D,$BF,$38,$FD,$87,$CA,$2B
	.dc.b	$A6,$3E,$1D,$3F,$B7,$7E,$78,$FD,$99,$FA,$D7,$FB,$97,$F4,$9F,$F3
	.dc.b	$77,$C3,$7F,$81,$6B,$53,$9F,$C1,$7F,$DA,$DB,$FF,$59,$9E,$5C,$BF
	.dc.b	$D0,$9F,$F1,$A3,$71,$EB,$ED,$B2,$FF,$65,$4F,$59,$CC,$3F,$32,$FF
	.dc.b	$01,$7D,$AB,$F6,$C7,$E9,$7F,$E8,$EF,$D7,$7F,$C8,$3F,$88,$FF,$86
	.dc.b	$6B,$53,$9F,$B7,$FD,$D6,$DF,$DB,$65,$FA,$4F,$F6,$97,$A6,$BF,$D6
	.dc.b	$1E,$3F,$3C,$CB,$FC,$9B,$3D,$73,$F0,$CE,$7E,$7F,$FC,$2D,$F9,$4B
	.dc.b	$F5,$07,$E8,$CF,$CE,$1F,$9E,$7F,$95,$3F,$36,$FF,$17,$5E,$EB,$53
	.dc.b	$FE,$58,$FD,$97,$3D,$1F,$DF,$6B,$A6,$BF,$DB,$1E,$18,$FF,$4A,$8F
	.dc.b	$0E,$5E,$08,$EF,$D9,$FE,$29,$FC,$27,$FD,$B5,$D5,$FC,$77,$D9,$7E
	.dc.b	$79,$FE,$9A,$E6,$FF,$08,$56,$BE,$05,$5A,$FA,$5F,$F2,$24,$FF,$6E
	.dc.b	$D3,$FB,$EB,$FE,$E5,$6F,$C5,$32,$ED,$F8,$F7,$86,$31,$8F,$4F,$D2
	.dc.b	$3F,$C7,$9F,$B5,$BF,$6D,$7E,$73,$FC,$77,$F1,$F7,$EE,$5F,$C4,$B8
	.dc.b	$3F,$C4,$97,$BA,$D4,$BE,$47,$FC,$B1,$3C,$F6,$FF,$08,$BF,$C3,$9B
	.dc.b	$8F,$6F,$89,$FF,$72,$FE,$97,$E2,$8C,$67,$BC,$FF,$C6,$5D,$EF,$DB
	.dc.b	$DF,$B6,$FD,$AE,$A7,$EB,$9F,$E5,$0B,$DD,$6B,$BF,$F0,$53,$FE,$25
	.dc.b	$F6,$7F,$CC,$18,$F3,$DB,$FC,$4A,$FF,$6B,$7F,$D1,$B1,$AB,$E5,$8B
	.dc.b	$5F,$F3,$15,$F8,$F3,$F1,$7C,$1F,$F9,$97,$F4,$37,$E5,$6F,$DC,$1F
	.dc.b	$B7,$FF,$83,$BF,$1B,$7E,$4F,$FE,$7D,$BD,$D6,$BE,$AF,$FD,$A1,$19
	.dc.b	$FB,$2F,$F2,$63,$FB,$4E,$1F,$29,$FE,$76,$78,$E3,$FE,$1A,$EC,$F1
	.dc.b	$67,$E1,$8F,$F0,$EF,$A9,$FE,$79,$CD,$FA,$47,$F1,$AF,$D5,$BF,$C8
	.dc.b	$9F,$00,$FD,$63,$FD,$21,$5A,$F7,$95,$AF,$C5,$73,$87,$F9,$7E,$9F
	.dc.b	$B8,$D8,$E9,$F7,$FF,$CB,$78,$37,$8F,$28,$FF,$70,$47,$FA,$77,$CE
	.dc.b	$1F,$E5,$5F,$81,$7E,$A0,$FC,$B5,$FA,$1E,$EB,$53,$9F,$CA,$71,$87
	.dc.b	$F9,$7A,$7F,$B3,$B5,$DF,$E1,$1F,$E9,$97,$BE,$4C,$BB,$FE,$27,$FE
	.dc.b	$9D,$F2,$DF,$33,$EC,$FB,$BF,$2F,$CB,$AB,$FD,$69,$5A,$E2,$56,$BD
	.dc.b	$1C,$FE,$63,$FE,$83,$AE,$7B,$7F,$9C,$9E,$15,$F0,$3E,$5B,$C1,$9F
	.dc.b	$7D,$BF,$C4,$9E,$57,$FA,$D6,$DE,$07,$F8,$D7,$E8,$5F,$B7,$BF,$25
	.dc.b	$7C,$AF,$92,$B5,$39,$FD,$CB,$CF,$DB,$E3,$99,$F6,$1F,$E0,$BA,$7E
	.dc.b	$9D,$F0,$67,$8F,$9F,$FE,$BF,$FB,$87,$EB,$0F,$D8,$9F,$AA,$3E,$FF
	.dc.b	$FD,$33,$F9,$9F,$FB,$5E,$F7,$5A,$9F,$F4,$EF,$CA,$67,$0F,$F5,$35
	.dc.b	$7F,$8E,$FA,$7F,$D2,$AF,$0C,$7A,$7D,$BF,$86,$BC,$11,$DF,$C2,$FF
	.dc.b	$69,$7E,$7A,$F2,$9F,$9F,$3E,$F1,$FB,$8B,$F6,$FF,$E3,$7F,$03,$FE
	.dc.b	$BB,$C3,$85,$EE,$B5,$F9,$67,$FB,$BE,$7A,$6F,$FE,$A8,$5F,$A7,$FF
	.dc.b	$AF,$5E,$39,$76,$D3,$C7,$A7,$8A,$3C,$3B,$F9,$BF,$48,$7C,$A3,$F2
	.dc.b	$5F,$9F,$FB,$3B,$2F,$05,$6A,$73,$F5,$3E,$3F,$8C,$F0,$4C,$C3,$97
	.dc.b	$6F,$F8,$27,$7E,$31,$8F,$9B,$FF,$7B,$61,$FC,$9F,$F1,$8F,$F7,$3F
	.dc.b	$E8,$0F,$A9,$FE,$89,$CF,$FE,$DD,$AD,$7E,$46,$AD,$7E,$B6,$AD,$7E
	.dc.b	$13,$FE,$DD,$BF,$E5,$1A,$7F,$12,$3A,$3C,$BF,$11,$FF,$0E,$BD,$30
	.dc.b	$FF,$48,$DF,$7F,$CE,$3F,$D7,$BF,$5C,$F8,$FF,$F4,$87,$E5,$9D,$FD
	.dc.b	$26,$73,$7F,$F1,$9E,$1C,$2F,$FD,$2D,$FB,$AE,$EB,$5A,$7F,$81,$1F
	.dc.b	$25,$F0,$7F,$DC,$30,$BE,$5E,$B7,$C6,$7A,$7C,$EF,$83,$5F,$F8,$01
	.dc.b	$9F,$F4,$EF,$11,$F6,$2F,$CF,$FF,$D7,$7F,$19,$E1,$D3,$E6,$5E,$EB
	.dc.b	$5C,$8F,$D7,$35,$BA,$9F,$A4,$FB,$B0,$F3,$C7,$F5,$C7,$FD,$EE,$F3
	.dc.b	$CB,$FD,$9D,$A7,$93,$E5,$BF,$39,$9F,$8F,$F5,$FF,$45,$F3,$CB,$B7
	.dc.b	$D7,$FF,$BF,$FD,$DF,$DC,$7E,$27,$FD,$EB,$2F,$6C,$E7,$D4,$FF,$67
	.dc.b	$7C,$E7,$9B,$1F,$F1,$9E,$DE,$2F,$C9,$BC,$CF,$F9,$62,$3F,$C6,$FF
	.dc.b	$15,$FD,$FE,$FF,$81,$CF,$1C,$3A,$6B,$E4,$8F,$B5,$6F,$D9,$9F,$4C
	.dc.b	$FC,$6D,$FB,$4F,$F9,$37,$E4,$BC,$55,$A9,$CF,$CC,$3F,$74,$CE,$5F
	.dc.b	$1B,$F6,$7F,$2B,$DC,$FA,$77,$C7,$1F,$F2,$C4,$74,$CB,$C1,$1F,$E5
	.dc.b	$EE,$2F,$BF,$E2,$FF,$3A,$FD,$D7,$FB,$CE,$2F,$D4,$3E,$24,$E7,$E7
	.dc.b	$52,$F3,$B3,$9F,$D8,$F7,$3C,$6F,$0F,$8F,$65,$FE,$04,$F8,$DF,$37
	.dc.b	$6F,$E0,$F3,$E9,$F2,$78,$8F,$EA,$3D,$5F,$DC,$3D,$47,$99,$FC,$CB
	.dc.b	$87,$F1,$5A,$7E,$96,$F7,$F3,$7F,$CE,$AB,$FC,$DB,$D5,$F8,$39,$FE
	.dc.b	$A3,$5F,$A4,$5F,$25,$E6,$C3,$9C,$FF,$05,$9E,$78,$FF,$52,$8E,$DF
	.dc.b	$24,$CF,$C9,$E7,$79,$FE,$7D,$FB,$4B,$DA,$FF,$03,$FE,$4D,$FA,$8D
	.dc.b	$3A,$E9,$FC,$46,$F7,$5A,$FC,$0B,$F4,$B2,$FD,$9F,$C9,$C1,$E9,$BB
	.dc.b	$B7,$BD,$E3,$C5,$FC,$87,$2E,$B8,$27,$3D,$3F,$2F,$ED,$FF,$D9,$FF
	.dc.b	$B2,$3E,$0D,$FB,$9F,$F5,$BC,$78,$D6,$A7,$3F,$BB,$79,$23,$C3,$A7
	.dc.b	$C5,$FB,$CB,$A7,$FA,$17,$F0,$86,$7D,$7D,$FF,$04,$63,$2F,$F5,$FD
	.dc.b	$3F,$6A,$FF,$03,$FB,$4F,$CA,$DC,$EF,$D8,$3E,$1D,$6A,$73,$FB,$46
	.dc.b	$36,$FC,$FF,$0F,$C0,$FC,$0B,$C7,$9F,$5F,$1B,$FC,$9F,$87,$C5,$E8
	.dc.b	$FC,$16,$5C,$F9,$BF,$1D,$FB,$1F,$F3,$1F,$B5,$7E,$99,$3F,$6F,$FA
	.dc.b	$73,$98,$CF,$EE,$FF,$5B,$F2,$FE,$F3,$3E,$27,$FC,$E9,$5F,$87,$E1
	.dc.b	$FC,$54,$F1,$F1,$3F,$39,$F9,$8F,$F2,$0F,$C9,$1F,$90,$7F,$BE,$7F
	.dc.b	$45,$FC,$0C,$E6,$3F,$DF,$9F,$31,$8C,$7D,$FC,$5F,$9B,$E3,$F7,$DE
	.dc.b	$FF,$54,$EB,$F1,$2C,$FA,$FE,$1F,$FA,$BF,$CB,$5F,$A3,$3F,$56,$7E
	.dc.b	$26,$FC,$CD,$F9,$9F,$E4,$AF,$75,$A9,$7D,$06,$70,$E5,$87,$F9,$27
	.dc.b	$83,$17,$D3,$DA,$FD,$D7,$E4,$AF,$DB,$7F,$C7,$3F,$A7,$3F,$51,$7E
	.dc.b	$CD,$FA,$0A,$D7,$D9,$2B,$5F,$AE,$67,$3F,$82,$7E,$FF,$F2,$8F,$BF
	.dc.b	$7C,$29,$CD,$E0,$CF,$BF,$E5,$FF,$C2,$FF,$33,$FF,$64,$FC,$6B,$F5
	.dc.b	$27,$E7,$CF,$C7,$7F,$63,$5A,$9C,$FE,$0F,$8D,$BC,$27,$F0,$59,$D3
	.dc.b	$D7,$F1,$BF,$E7,$95,$FE,$09,$8F,$DD,$FE,$CC,$FC,$E5,$F9,$5B,$F5
	.dc.b	$D7,$EC,$4F,$D0,$77,$5A,$9C,$FE,$59,$FC,$12,$FF,$3B,$84,$AD,$75
	.dc.b	$FF,$27,$8E,$D3,$D7,$D4,$FE,$55,$FB,$77,$F8,$7F,$F1,$EF,$D6,$7C
	.dc.b	$D3,$F5,$2F,$E1,$F3,$9F,$7F,$FA,$8F,$E0,$1F,$C4,$BF,$3C,$FD,$5C
	.dc.b	$CF,$79,$8F,$ED,$7E,$53,$AC,$63,$0F,$F6,$27,$27,$F1,$AF,$9D,$7E
	.dc.b	$80,$FD,$E3,$FC,$EB,$FB,$AB,$F4,$AD,$4E,$7F,$0F,$4F,$E5,$BE,$1F
	.dc.b	$F8,$DE,$1F,$BC,$E8,$FD,$EB,$D3,$1F,$F3,$FF,$BB,$FD,$1E,$31,$87
	.dc.b	$FA,$BE,$FF,$91,$F6,$3E,$E5,$EB,$3F,$69,$FC,$17,$EB,$FF,$CC,$2B
	.dc.b	$53,$9F,$05,$F9,$9E,$78,$BF,$03,$C9,$FE,$49,$BD,$E3,$FE,$F5,$F7
	.dc.b	$BF,$A7,$46,$3F,$0E,$FE,$B5,$F9,$23,$F4,$97,$EC,$8F,$CE,$5F,$80
	.dc.b	$BF,$12,$FE,$83,$7B,$AD,$4B,$F1,$1C,$ED,$E2,$DB,$FC,$D6,$DB,$E1
	.dc.b	$8F,$F1,$19,$FF,$2F,$46,$39,$FF,$B6,$7E,$38,$FD,$A1,$FA,$F3,$F5
	.dc.b	$77,$B4,$FB,$CF,$EA,$77,$BA,$D4,$FF,$8A,$79,$3F,$B6,$4F,$F3,$0F
	.dc.b	$CD,$FF,$98,$BB,$C1,$FA,$67,$83,$2E,$F0,$FB,$1F,$EF,$DE,$C3,$F6
	.dc.b	$27,$E0,$3F,$C5,$3F,$2A,$FF,$03,$F8,$5F,$D7,$2B,$53,$9F,$C6,$3F
	.dc.b	$B3,$C7,$F4,$0F,$AE,$FE,$82,$EB,$E2,$9F,$95,$76,$CB,$AC,$7F,$8C
	.dc.b	$3A,$7F,$B7,$FE,$68,$FD,$1B,$C9,$F5,$6F,$D0,$7F,$D9,$7B,$7F,$6C
	.dc.b	$AD,$70,$2B,$5F,$91,$72,$C7,$EA,$56,$A7,$F8,$15,$F1,$C7,$AF,$E8
	.dc.b	$7F,$E1,$DF,$B5,$FF,$21,$7C,$8F,$F7,$2F,$D3,$BF,$BA,$5F,$F7,$A9
	.dc.b	$CF,$D9,$A7,$F8,$34,$FF,$7E,$FC,$63,$FA,$FD,$6B,$E1,$3A,$7C,$67
	.dc.b	$F2,$DC,$3F,$C6,$9B,$7F,$83,$3F,$62,$FE,$1F,$F9,$EF,$F2,$DF,$98
	.dc.b	$7E,$63,$9F,$45,$39,$FD,$0B,$FC,$27,$F5,$1F,$DB,$30,$95,$AF,$96
	.dc.b	$FF,$39,$F8,$FF,$E2,$D9,$EB,$1F,$EE,$3C,$FF,$C3,$7F,$8D,$FF,$BA
	.dc.b	$3E,$E1,$FA,$DF,$F4,$EF,$8E,$CF,$C5,$A7,$3F,$44,$FD,$13,$8E,$3F
	.dc.b	$9F,$C7,$96,$3F,$D4,$A3,$8D,$F4,$FC,$8B,$FC,$91,$F9,$53,$F3,$A7
	.dc.b	$E9,$CF,$D0,$5F,$01,$FD,$C3,$BB,$FC,$4F,$7B,$AD,$4F,$E5,$67,$37
	.dc.b	$CA,$FF,$8E,$C6,$2B,$A7,$E7,$5F,$E5,$7F,$CC,$1F,$BC,$BF,$2F,$7E
	.dc.b	$A6,$FD,$DF,$F3,$D8,$FF,$8D,$AF,$75,$A9,$E5,$7C,$6F,$FC,$EF,$7B
	.dc.b	$8C,$7E,$27,$FE,$65,$FD,$33,$F7,$3F,$A8,$FF,$6A,$FD,$D5,$FB,$97
	.dc.b	$F9,$2A,$73,$F5,$E9,$7E,$E3,$FE,$4E,$F6,$FF,$C4,$F5,$AF,$9C,$E9
	.dc.b	$F0,$8C,$BF,$BA,$F3,$7F,$9C,$ED,$FB,$2F,$F8,$1F,$F7,$47,$CA,$FF
	.dc.b	$77,$E1,$FF,$28,$D6,$BE,$D5,$5A,$E3,$FF,$2E,$FE,$6F,$FE,$36,$AD
	.dc.b	$4F,$FA,$47,$C8,$FF,$3B,$46,$3F,$1E,$FF,$41,$57,$FB,$3F,$F3,$27
	.dc.b	$EB,$9F,$DE,$6B,$9F,$E5,$49,$CF,$CD,$2F,$31,$C3,$BB,$9D,$6F,$0F
	.dc.b	$B7,$97,$FA,$17,$F3,$47,$E1,$AF,$DE,$9F,$99,$FE,$93,$97,$FC,$EB
	.dc.b	$5A,$9C,$FA,$BF,$C9,$F9,$3F,$CE,$33,$9B,$7D,$BF,$FA,$6E,$31,$F7
	.dc.b	$37,$F1,$1F,$EE,$5F,$AB,$BF,$3A,$F9,$78,$F2,$9C,$C3,$F6,$CF,$0C
	.dc.b	$70,$F8,$5E,$5F,$64,$C3,$F8,$7F,$0F,$FA,$7F,$F3,$6F,$F3,$9F,$EB
	.dc.b	$AF,$D3,$BF,$CB,$BF,$47,$FE,$8E,$9C,$C7,$F9,$83,$87,$1F,$8D,$AF
	.dc.b	$87,$E6,$FF,$EB,$3B,$DF,$D0,$DF,$C0,$DF,$BD,$FF,$7B,$FC,$23,$F0
	.dc.b	$5A,$F8,$67,$3F,$6D,$FE,$48,$F1,$BC,$11,$8A,$E5,$7F,$E0,$3F,$6F
	.dc.b	$FD,$81,$87,$F8,$23,$F0,$F7,$9E,$FD,$DB,$FC,$7B,$D5,$FE,$9E,$AD
	.dc.b	$4E,$7E,$4F,$FD,$53,$18,$F0,$79,$FD,$C7,$F9,$BA,$9E,$97,$6F,$DC
	.dc.b	$FF,$89,$FE,$F9,$FE,$15,$F3,$78,$C6,$3E,$E3,$F1,$5A,$77,$8D,$FD
	.dc.b	$1F,$1F,$86,$60,$FC,$CB,$D3,$FB,$0F,$F6,$DF,$CF,$5F,$26,$FD,$6B
	.dc.b	$CB,$39,$8F,$9D,$9F,$AE,$38,$7C,$AF,$2F,$AA,$60,$EE,$F8,$BF,$81
	.dc.b	$7C,$4F,$CA,$5E,$7B,$F0,$6F,$83,$39,$8F,$F2,$7E,$7F,$1E,$3E,$5F
	.dc.b	$DB,$7F,$AC,$BA,$91,$CF,$E2,$FF,$DE,$1F,$6B,$FD,$CF,$E2,$7F,$C0
	.dc.b	$5F,$6E,$FD,$63,$FD,$AB,$39,$FB,$6F,$F0,$DF,$ED,$98,$FA,$E6,$1F
	.dc.b	$E0,$9F,$C1,$39,$C6,$2B,$E8,$FC,$C7,$2F,$86,$FF,$39,$7E,$8F,$FE
	.dc.b	$12,$E2,$C7,$C3,$A3,$FC,$2B,$F0,$F8,$7F,$CF,$B7,$7E,$03,$CF,$8B
	.dc.b	$F7,$17,$E7,$8F,$D1,$FF,$91,$F4,$7F,$B9,$38,$BF,$C3,$95,$A9,$FE
	.dc.b	$47,$39,$BE,$57,$FE,$08,$C7,$DF,$18,$DB,$E2,$FE,$8F,$FF,$00,$7E
	.dc.b	$71,$F9,$93,$F4,$DF,$8B,$39,$87,$DF,$63,$F3,$35,$BE,$17,$9C,$63
	.dc.b	$F2,$FF,$A3,$FB,$CF,$EF,$FF,$18,$FF,$5A,$FE,$B1,$F4,$23,$1F,$79
	.dc.b	$FE,$1C,$E2,$F6,$46,$F9,$B8,$FC,$63,$0F,$F1,$8F,$1F,$D8,$FC,$45
	.dc.b	$FA,$63,$F2,$CF,$F4,$87,$ED,$99,$8C,$7E,$35,$F9,$4F,$25,$84,$ED
	.dc.b	$1F,$E8,$BA,$E5,$F8,$C7,$E2,$32,$EB,$97,$FB,$BE,$F8,$5F,$0F,$D0
	.dc.b	$7C,$3F,$93,$FF,$13,$FC,$7B,$F1,$71,$8F,$91,$FE,$4F,$C7,$F5,$E3
	.dc.b	$7F,$8A,$7E,$51,$E3,$A7,$E5,$BF,$05,$7E,$48,$9F,$C8,$FE,$13,$F9
	.dc.b	$69,$CC,$7E,$53,$93,$15,$BF,$37,$93,$C7,$CD,$FD,$0F,$DB,$BD,$67
	.dc.b	$EB,$6F,$D7,$1F,$A7,$7E,$8A,$B5,$39,$FD,$5B,$17,$C3,$E3,$39,$BC
	.dc.b	$BE,$2B,$F5,$BF,$1A,$FD,$F7,$FC,$F9,$FA,$DF,$F5,$9E,$6F,$CB,$4E
	.dc.b	$7D,$E4,$BF,$04,$FD,$8C,$E7,$0E,$78,$7F,$A1,$78,$3F,$73,$F0,$5F
	.dc.b	$EC,$CF,$C2,$DF,$BC,$7F,$4C,$F1,$3E,$DF,$83,$F6,$BC,$1F,$B7,$7B
	.dc.b	$AF,$F4,$FE,$FF,$A7,$8E,$5E,$67,$4F,$C1,$3F,$CC,$18,$7F,$06,$FC
	.dc.b	$67,$F1,$7F,$57,$FC,$8B,$F2,$BF,$E0,$B1,$8F,$8F,$7E,$CB,$8F,$2F
	.dc.b	$FB,$CF,$E8,$FC,$5F,$EF,$7E,$AF,$F1,$0F,$C3,$DF,$9A,$BC,$97,$DA
	.dc.b	$BE,$93,$F8,$8D,$6A,$73,$CD,$8B,$E1,$F8,$D7,$37,$97,$E1,$FF,$C7
	.dc.b	$3F,$7E,$FE,$57,$ED,$BE,$AB,$FC,$37,$D3,$FC,$16,$73,$0F,$D8,$3F
	.dc.b	$8F,$C7,$0F,$1B,$97,$5B,$0F,$F6,$07,$4F,$E7,$BB,$7C,$FF,$B0,$7F
	.dc.b	$3F,$FC,$BD,$F9,$5F,$F2,$8E,$54,$E7,$9F,$0F,$F4,$5F,$87,$FE,$63
	.dc.b	$3A,$F9,$CF,$E2,$5C,$3F,$C6,$7D,$37,$EF,$EF,$93,$FF,$75,$7E,$85
	.dc.b	$FC,$D2,$73,$1F,$EE,$DF,$9C,$C5,$6F,$F6,$AF,$E7,$B5,$AF,$D6,$7F
	.dc.b	$9F,$FE,$A7,$F0,$FF,$0E,$FA,$FF,$AD,$7F,$3E,$8C,$78,$BF,$6B,$5F
	.dc.b	$E5,$51,$BF,$D3,$38,$F6,$FF,$9B,$7D,$FF,$E9,$3D,$1F,$C1,$BE,$79
	.dc.b	$FA,$2F,$DA,$C7,$F9,$94,$E6,$1F,$BE,$FF,$4C,$8E,$1E,$17,$25,$7F
	.dc.b	$E9,$AF,$D4,$BF,$D3,$7D,$DF,$C0,$3C,$8F,$D2,$7E,$01,$FD,$6B,$F5
	.dc.b	$77,$EB,$3E,$0B,$DF,$C2,$7F,$AA,$7D,$C5,$EE,$9E,$AB,$F5,$8F,$E9
	.dc.b	$59,$FF,$20,$C9,$FA,$D3,$F7,$7F,$F4,$67,$E7,$EF,$D1,$3F,$CC,$3F
	.dc.b	$99,$71,$FA,$3F,$D6,$E1,$C2,$F7,$5F,$E6,$6F,$D7,$1F,$B7,$BB,$FF
	.dc.b	$67,$F9,$13,$F4,$07,$E8,$4F,$DA,$7F,$D3,$5F,$A5,$FC,$AF,$D8,$78
	.dc.b	$E1,$C2,$F7,$5A,$EF,$7C,$46,$B9,$3F,$9C,$E7,$D7,$3E,$9F,$92,$34
	.dc.b	$78,$1F,$CF,$B6,$F2,$61,$C3,$4E,$B7,$BF,$7F,$5B,$78,$72,$E7,$97
	.dc.b	$2F,$AE,$F0,$62,$D3,$C1,$BF,$AB,$5F,$F0,$BE,$BD,$AB,$5A,$F6,$C5
	.dc.b	$EF,$F6,$9F,$16,$7D,$BE,$ED,$F3,$9D,$B4,$EB,$C1,$EB,$9F,$F3,$7E
	.dc.b	$7F,$C4,$30,$FE,$B3,$F6,$EF,$C1,$FF,$C0,$BF,$70,$FD,$11,$FB,$03
	.dc.b	$F4,$B7,$E5,$DF,$5B,$D3,$FE,$4F,$8F,$1F,$B4,$FF,$17,$61,$C2,$DF
	.dc.b	$95,$6F,$4E,$FA,$76,$AE,$BD,$7F,$37,$2E,$73,$9D,$3F,$C4,$F7,$C4
	.dc.b	$FF,$1F,$46,$69,$FB,$06,$74,$F1,$6D,$E9,$C3,$EF,$FD,$46,$70,$E9
	.dc.b	$F3,$3D,$F2,$F1,$EF,$E8,$B7,$DE,$FE,$83,$8B,$E3,$57,$82,$B5,$3F
	.dc.b	$84,$9C,$FE,$5F,$FD,$8B,$67,$E9,$EF,$D2,$7F,$B8,$EC,$FD,$F7,$D5
	.dc.b	$7B,$AD,$7A,$7F,$E5,$5F,$CB,$33,$F4,$BF,$96,$9C,$FA,$9E,$7D,$D8
	.dc.b	$71,$CF,$FB,$EF,$D6,$66,$31,$93,$F1,$B7,$E4,$1D,$37,$F2,$E1,$FE
	.dc.b	$04,$97,$EC,$59,$D3,$C7,$FA,$5F,$F8,$D3,$8B,$DB,$BF,$8F,$0F,$F0
	.dc.b	$5C,$FF,$8E,$E7,$38,$F4,$CF,$FA,$A6,$DF,$BA,$FA,$0F,$ED,$7E,$6B
	.dc.b	$E5,$1E,$A7,$F9,$FF,$EA,$1F,$B3,$3F,$01,$FF,$43,$68,$C7,$8E,$1C
	.dc.b	$3E,$49,$87,$0B,$7E,$ED,$DF,$4F,$F1,$9E,$9D,$F4,$ED,$A3,$E6,$FF
	.dc.b	$75,$A7,$49,$CF,$E4,$B8,$C3,$8E,$BD,$F4,$F2,$FE,$6D,$FE,$DF,$DF
	.dc.b	$C9,$5A,$C3,$AC,$FF,$21,$9C,$CF,$F7,$5B,$E5,$9B,$F3,$47,$B5,$E7
	.dc.b	$AD,$63,$E6,$9F,$E5,$18,$72,$DB,$D5,$C3,$F0,$31,$7D,$DF,$F1,$3F
	.dc.b	$CE,$3F,$C2,$FF,$7C,$FD,$65,$F9,$8F,$F3,$7F,$B5,$7E,$97,$FD,$93
	.dc.b	$EB,$FF,$96,$E1,$C2,$F7,$F1,$4B,$DF,$CE,$3F,$7F,$DF,$1F,$F9,$7B
	.dc.b	$4E,$DA,$75,$D1,$F5,$DE,$3A,$D6,$7D,$32,$F0,$FD,$76,$67,$FD,$FF
	.dc.b	$39,$B7,$E9,$3F,$B0,$E4,$6F,$E5,$CF,$B4,$BF,$2D,$F1,$E3,$CA,$7C
	.dc.b	$5F,$A1,$F5,$DB,$CB,$C3,$EB,$CD,$FB,$37,$F1,$2D,$3A,$E3,$E6,$C7
	.dc.b	$D5,$3F,$C7,$25,$FB,$17,$F9,$0B,$DB,$7D,$7B,$F3,$D7,$DF,$3F,$76
	.dc.b	$7E,$4B,$FE,$51,$F8,$FF,$EB,$38,$F1,$C3,$85,$EE,$BF,$CA,$DF,$90
	.dc.b	$EB,$6F,$F4,$4E,$7D,$B3,$EB,$9F,$4F,$BB,$FE,$A1,$9F,$39,$CE,$3F
	.dc.b	$E5,$6D,$7C,$F8,$F2,$D3,$BE,$3E,$4F,$BC,$FE,$CD,$5A,$AF,$F2,$7D
	.dc.b	$6A,$FF,$CC,$3F,$BC,$78,$E7,$38,$F6,$CD,$F9,$8E,$75,$F3,$6F,$EB
	.dc.b	$97,$EC,$39,$CF,$A7,$67,$F6,$7F,$9E,$FF,$35,$FE,$16,$FD,$0D,$F9
	.dc.b	$B3,$F3,$CF,$F0,$9F,$CB,$3E,$AF,$F1,$7C,$9F,$E0,$CE,$2F,$F5,$6E
	.dc.b	$3C,$70,$E1,$F4,$BF,$D8,$AF,$F1,$37,$BB,$5E,$F1,$EB,$F1,$7E,$39
	.dc.b	$E7,$9F,$8A,$7C,$3E,$DE,$67,$8C,$E7,$2F,$ED,$F6,$FC,$9F,$FC,$2B
	.dc.b	$B7,$B2,$B5,$B2,$9F,$B1,$EB,$1E,$FB,$FA,$2B,$F8,$94,$64,$EF,$19
	.dc.b	$FB,$D7,$ED,$D9,$DF,$D5,$AF,$F8,$02,$7F,$D6,$1C,$0F,$CB,$BF,$D2
	.dc.b	$FF,$A8,$7E,$CB,$E8,$7F,$C8,$7E,$97,$F2,$6F,$CB,$5F,$A6,$FF,$8B
	.dc.b	$2F,$7E,$97,$F8,$73,$EC,$57,$BF,$DC,$75,$C3,$FE,$DC,$CF,$A5,$6B
	.dc.b	$E0,$3F,$DE,$93,$A1,$3A,$BF,$E6,$BD,$FF,$E2,$7C,$BB,$EB,$E4,$C3
	.dc.b	$FD,$F9,$F9,$E7,$1C,$38,$47,$FB,$FF,$F3,$AC,$6B,$E5,$CB,$F9,$FB
	.dc.b	$CF,$3F,$23,$9A,$FD,$97,$E4,$BC,$32,$ED,$E5,$3F,$5D,$F3,$7E,$F6
	.dc.b	$FD,$D9,$FB,$73,$EF,$FF,$77,$F0,$0F,$DD,$9F,$BF,$70,$BD,$D6,$A5
	.dc.b	$F8,$1E,$70,$E5,$E6,$FF,$34,$F4,$BC,$DC,$4D,$3C,$15,$AD,$1F,$9D
	.dc.b	$6A,$73,$9F,$8B,$C0,$7D,$2F,$58,$C7,$CC,$5D,$3F,$A0,$FE,$D8,$FD
	.dc.b	$C3,$19,$A1,$F9,$E6,$3F,$06,$6F,$FE,$E6,$FC,$67,$FD,$BB,$2F,$CC
	.dc.b	$DF,$7C,$FD,$8D,$FB,$C7,$FB,$2F,$F6,$5F,$F7,$EF,$E4,$1F,$ED,$CC
	.dc.b	$BA,$E1,$C2,$F7,$F8,$E2,$DF,$B9,$77,$BF,$0E,$7D,$F3,$ED,$9F,$FB
	.dc.b	$5F,$F1,$2F,$2E,$9D,$6B,$59,$F4,$E5,$CE,$DE,$8D,$BC,$18,$FF,$B7
	.dc.b	$FF,$1C,$7E,$BB,$DF,$5F,$2E,$BD,$B8,$3F,$DC,$FF,$67,$9D,$7A,$E3
	.dc.b	$FE,$E7,$F8,$67,$E8,$59,$D3,$AE,$09,$7D,$03,$87,$FD,$3D,$A7,$CC
	.dc.b	$F0,$9F,$91,$B8,$7F,$CC,$FF,$0B,$FD,$5E,$1F,$88,$71,$FC,$2B,$E9
	.dc.b	$E1,$F8,$F9,$E1,$C3,$FD,$3B,$E5,$BF,$D6,$33,$B3,$B4,$EB,$0F,$EC
	.dc.b	$5F,$2B,$F2,$30,$E7,$9F,$8B,$3F,$0E,$2F,$B5,$FB,$BA,$FA,$33,$F0
	.dc.b	$56,$A7,$FC,$D3,$39,$BE,$5A,$78,$FE,$CB,$B6,$1E,$2C,$3A,$FE,$A9
	.dc.b	$FE,$32,$C3,$86,$DE,$5C,$D2,$FC,$97,$E4,$7B,$1F,$03,$D0,$FF,$1F
	.dc.b	$7B,$1F,$E3,$8F,$6B,$FC,$77,$F9,$23,$F4,$CD,$ED,$F9,$8B,$DF,$D6
	.dc.b	$AF,$7F,$AF,$EA,$BF,$7D,$F5,$FF,$C4,$AB,$5E,$DF,$F9,$4B,$85,$AF
	.dc.b	$86,$B5,$8F,$F9,$2B,$8F,$F2,$33,$9D,$BC,$98,$3E,$A1,$FA,$0F,$7D
	.dc.b	$BC,$78,$FF,$94,$A7,$FC,$A3,$9F,$6D,$7D,$1A,$76,$C7,$FD,$93,$1C
	.dc.b	$63,$1F,$B0,$FF,$AB,$2B,$84,$FA,$B3,$F0,$CF,$E7,$27,$3F,$A2,$BF
	.dc.b	$44,$7E,$49,$FD,$D3,$EE,$FF,$71,$F0,$AF,$CA,$7E,$25,$EE,$B5,$E7
	.dc.b	$A9,$E3,$EB,$1E,$7F,$3B,$FE,$86,$AF,$F2,$6E,$D9,$C3,$8F,$DA,$7C
	.dc.b	$68,$FE,$17,$97,$5B,$E1,$EB,$FF,$BE,$36,$F3,$E5,$D3,$6F,$E1,$93
	.dc.b	$FC,$C6,$73,$97,$93,$E0,$7A,$64,$CB,$A6,$1F,$C4,$FE,$C3,$FC,$B3
	.dc.b	$87,$0C,$FC,$D0,$FD,$6F,$F6,$9C,$1F,$AC,$FC,$71,$D8,$FD,$A1,$FB
	.dc.b	$5B,$F2,$FF,$F1,$CF,$9C,$FC,$CB,$8F,$07,$F2,$6C,$78,$E1,$C2,$F7
	.dc.b	$EF,$EA,$95,$AC,$7A,$7C,$0F,$DE,$E0,$F9,$D5,$CF,$4F,$15,$6B,$E6
	.dc.b	$BF,$8F,$ED,$E7,$9C,$E7,$FE,$4C,$FC,$73,$F9,$D6,$3C,$B1,$EF,$5A
	.dc.b	$C9,$5F,$E5,$0A,$D6,$7E,$3C,$3F,$CA,$BF,$8D,$E7,$2E,$DB,$7A,$F1
	.dc.b	$4F,$6F,$DC,$FF,$89,$FC,$9F,$F9,$B7,$EF,$7F,$AE,$FE,$CA,$FC,$23
	.dc.b	$FD,$C7,$CB,$FC,$86,$F7,$E9,$7F,$AE,$FB,$17,$BF,$C2,$DF,$8A,$6F
	.dc.b	$97,$FA,$8E,$9D,$6F,$79,$BE,$A3,$FA,$0E,$7D,$35,$F0,$CF,$F8,$AF
	.dc.b	$0E,$47,$F8,$5B,$4F,$06,$9E,$5F,$D5,$7F,$C3,$3C,$AC,$79,$67,$DA
	.dc.b	$5F,$AB,$72,$D7,$C5,$87,$B7,$E8,$5F,$1D,$9C,$FA,$EA,$97,$E3,$DF
	.dc.b	$D0,$7F,$7A,$FE,$73,$F3,$EF,$D3,$5F,$A1,$7F,$46,$FC,$A9,$FA,$1B
	.dc.b	$F7,$07,$ED,$5C,$79,$BF,$B9,$E3,$C7,$0E,$17,$BF,$D5,$F5,$5F,$E3
	.dc.b	$DC,$BA,$FC,$63,$3E,$BF,$A9,$FF,$68,$D7,$C7,$5A,$9F,$0F,$CB,$67
	.dc.b	$0E,$F3,$9C,$9F,$AB,$7F,$89,$7A,$D8,$F2,$D7,$C7,$3E,$99,$CE,$1C
	.dc.b	$A7,$CF,$FA,$B7,$CF,$9C,$CF,$9E,$73,$7F,$D9,$FD,$4F,$EB,$FF,$BD
	.dc.b	$3F,$34,$7C,$37,$FB,$FF,$F3,$EF,$F1,$87,$ED,$6F,$D2,$B8,$67,$FE
	.dc.b	$18,$C3,$85,$EF,$E4,$57,$FE,$19,$D7,$F8,$BD,$AF,$7F,$06,$CB,$FC
	.dc.b	$79,$F5,$FF,$E0,$ED,$7C,$7A,$F8,$AB,$59,$FF,$8D,$F1,$F0,$9C,$E7
	.dc.b	$58,$7F,$91,$7F,$4B,$7E,$D9,$AC,$FB,$CE,$7F,$0B,$C5,$F0,$C5,$83
	.dc.b	$E1,$B3,$C1,$EB,$8F,$D4,$EE,$FD,$67,$FC,$0B,$F0,$1F,$C5,$7F,$4B
	.dc.b	$FF,$56,$FE,$9B,$F0,$3E,$09,$FB,$E7,$F8,$4E,$F7,$F0,$2B,$DF,$3F
	.dc.b	$FA,$E3,$F4,$8D,$72,$FF,$90,$B3,$E9,$F8,$3B,$4F,$97,$F8,$3F,$F8
	.dc.b	$CB,$3E,$79,$72,$F9,$1C,$63,$C7,$4E,$D8,$F1,$D3,$FD,$8B,$F0,$3D
	.dc.b	$A9,$A7,$6D,$3D,$1E,$E6,$63,$16,$FB,$C7,$E6,$B9,$CF,$AE,$1F,$7A
	.dc.b	$7F,$C6,$5F,$84,$3F,$27,$FF,$14,$FA,$AF,$D0,$5F,$A8,$3D,$E7,$EC
	.dc.b	$0F,$67,$F6,$AF,$FB,$37,$0E,$37,$C2,$B7,$5F,$BA,$FC,$8F,$5C,$5F
	.dc.b	$E6,$8C,$BB,$65,$D7,$2E,$9F,$17,$FE,$6C,$D3,$C9,$7B,$CF,$A5,$78
	.dc.b	$7E,$E3,$5F,$17,$FE,$40,$FD,$3D,$FA,$07,$F7,$7F,$C8,$9F,$BC,$BF
	.dc.b	$69,$FF,$14,$7D,$EF,$FB,$8F,$0E,$17,$BA,$D7,$D3,$BF,$00,$D6,$DF
	.dc.b	$E7,$EC,$FA,$FC,$FB,$4E,$BA,$7F,$8C,$FF,$3D,$DD,$AF,$75,$AC,$5F
	.dc.b	$A0,$7F,$BC,$EF,$FC,$AF,$F6,$6F,$CD,$FF,$C2,$5F,$95,$FF,$A3,$BF
	.dc.b	$64,$FE,$97,$FB,$CE,$F4,$FF,$47,$DE,$FE,$8D,$7B,$F1,$9F,$B4,$6E
	.dc.b	$FF,$CD,$D9,$3F,$47,$FF,$32,$C7,$F9,$9B,$AD,$AF,$8A,$BF,$C4,$53
	.dc.b	$A3,$96,$7D,$F2,$F2,$76,$7D,$5B,$E1,$A7,$8B,$EF,$5B,$FF,$8D,$33
	.dc.b	$EB,$97,$A3,$3F,$6D,$6B,$6F,$C6,$56,$AB,$CF,$87,$3C,$3F,$B5,$FB
	.dc.b	$9E,$BF,$A2,$7E,$C4,$FC,$37,$FD,$8F,$F9,$63,$77,$C0,$BF,$7E,$E1
	.dc.b	$7B,$F9,$B5,$EE,$9F,$90,$6B,$1F,$81,$B7,$0F,$DD,$AF,$07,$97,$5F
	.dc.b	$0D,$78,$35,$7E,$65,$F0,$F0,$FF,$61,$D6,$B4,$EF,$4F,$AE,$D5,$7F
	.dc.b	$B2,$F6,$F2,$6D,$E3,$FC,$27,$C5,$87,$0D,$DF,$8B,$7D,$BF,$97,$7E
	.dc.b	$E3,$8F,$D1,$7F,$C0,$BF,$86,$3F,$5A,$7E,$49,$BA,$D7,$D1,$AB,$5F
	.dc.b	$43,$9F,$C6,$BD,$BE,$0F,$FA,$A3,$3E,$BA,$63,$8B,$FD,$BB,$18,$D7
	.dc.b	$BF,$E9,$BE,$A8,$C6,$7D,$78,$31,$AF,$83,$5F,$27,$D3,$F8,$F3,$65
	.dc.b	$E8,$E0,$FF,$5E,$F5,$7E,$07,$3E,$78,$7A,$FE,$B7,$85,$EF,$EB,$3F
	.dc.b	$39,$FA,$BB,$B9,$FB,$5B,$F7,$E7,$C4,$BF,$61,$7E,$F4,$FD,$3F,$8E
	.dc.b	$1C,$2F,$75,$AF,$DA,$B3,$8F,$C1,$E8,$67,$DF,$F4,$2C,$3E,$19,$F0
	.dc.b	$69,$DF,$CB,$FB,$B3,$9C,$BA,$FE,$0D,$8D,$3C,$17,$E5,$DB,$FC,$59
	.dc.b	$C9,$FA,$1C,$38,$4E,$7E,$E5,$3F,$C2,$35,$0F,$A6,$7E,$F4,$FC,$6D
	.dc.b	$F8,$AB,$F6,$07,$E6,$5F,$0B,$E0,$F9,$FB,$BF,$EE,$6C,$78,$E1,$C2
	.dc.b	$F7,$BF,$F5,$9D,$3C,$1A,$77,$D3,$B3,$FC,$16,$33,$87,$8B,$F1,$3F
	.dc.b	$D3,$E5,$D6,$FC,$9A,$F8,$F5,$FE,$1B,$87,$09,$CE,$7E,$6E,$2F,$E8
	.dc.b	$18,$71,$D7,$C9,$87,$DA,$E8,$F4,$EB,$F0,$B5,$F4,$E9,$D7,$6F,$E4
	.dc.b	$7A,$FA,$3F,$32,$FF,$02,$7E,$71,$FE,$B9,$FB,$1F,$F6,$BF,$E0,$9F
	.dc.b	$E1,$EA,$D7,$D3,$2B,$5F,$87,$FF,$A7,$3E,$9B,$BE,$AC,$BF,$C1,$DF
	.dc.b	$9A,$FB,$A3,$E9,$C7,$67,$E2,$71,$7D,$67,$86,$3D,$7E,$2F,$FC,$B9
	.dc.b	$8F,$EC,$E7,$AD,$F9,$A7,$3D,$DF,$0A,$7A,$DF,$1D,$BF,$DD,$FC,$DF
	.dc.b	$DF,$76,$F4,$ED,$E8,$87,$D1,$F4,$BE,$1A,$7E,$0F,$F5,$8C,$5E,$FF
	.dc.b	$78,$F6,$F3,$3E,$49,$1F,$ED,$BF,$DF,$5F,$98,$3E,$69,$F8,$EF,$0B
	.dc.b	$DF,$84,$BD,$D3,$F5,$4D,$63,$F1,$F8,$FF,$69,$87,$F7,$C7,$FA,$9F
	.dc.b	$17,$E3,$B4,$F1,$57,$87,$4F,$F4,$4D,$72,$CF,$B4,$7F,$75,$8C,$67
	.dc.b	$E0,$E7,$B3,$E7,$F6,$3E,$CF,$E3,$FF,$E3,$9F,$BA,$7E,$7D,$FA,$6F
	.dc.b	$D4,$FF,$C9,$3F,$83,$6B,$EB,$7F,$CF,$33,$9F,$01,$F9,$06,$70,$F5
	.dc.b	$63,$FE,$43,$F1,$6E,$1C,$3C,$56,$6F,$BB,$FB,$5C,$2C,$3C,$5A,$78
	.dc.b	$7E,$83,$31,$8C,$38,$EB,$FE,$17,$E6,$FD,$8C,$E7,$2F,$F2,$FF,$E5
	.dc.b	$3F,$C1,$3E,$3B,$F1,$DF,$BE,$BF,$2C,$7E,$6C,$FD,$BB,$FC,$F7,$E2
	.dc.b	$7F,$7E,$F9,$CE,$1C,$3F,$C2,$B0,$E1,$7B,$FB,$1F,$F3,$57,$4B,$F3
	.dc.b	$FD,$DA,$FB,$7E,$6B,$FD,$93,$5F,$0E,$BF,$EC,$BA,$FB,$71,$8C,$BA
	.dc.b	$69,$FE,$03,$8C,$6B,$E4,$FC,$83,$FC,$7D,$18,$D7,$C3,$F8,$57,$83
	.dc.b	$4F,$2D,$FA,$B6,$FE,$41,$C7,$FE,$01,$D3,$D5,$AF,$C0,$AD,$78,$DC
	.dc.b	$63,$19,$7B,$3E,$AF,$D7,$FB,$77,$F9,$FB,$E4,$9F,$4D,$FE,$B6,$FB
	.dc.b	$D7,$E6,$3F,$D5,$38,$7F,$75,$7B,$AD,$4F,$FB,$83,$6F,$16,$CA,$FF
	.dc.b	$44,$FE,$91,$F1,$63,$F9,$BC,$77,$7F,$94,$61,$D3,$EF,$FF,$BB,$E3
	.dc.b	$D2,$7C,$B1,$8F,$C2,$59,$FE,$7F,$2E,$F8,$F8,$E7,$39,$BE,$37,$8E
	.dc.b	$BE,$C9,$CF,$C3,$7D,$A9,$D7,$EA,$5F,$D7,$3D,$37,$ED,$9F,$C8,$7F
	.dc.b	$31,$D5,$2B,$53,$FC,$97,$F5,$1C,$FD,$C7,$B7,$87,$C1,$7B,$7D,$A7
	.dc.b	$F5,$67,$BF,$C5,$FF,$08,$B7,$E7,$5F,$80,$AD,$57,$7A,$D7,$E8,$B9
	.dc.b	$9E,$B9,$78,$BD,$BF,$AF,$C7,$F4,$30,$EB,$B7,$AA,$5F,$94,$FF,$68
	.dc.b	$E0,$FE,$65,$F9,$BB,$E2,$DF,$9C,$BC,$87,$EB,$2E,$5F,$F4,$EF,$E7
	.dc.b	$1C,$78,$3F,$B7,$E3,$C7,$0E,$17,$BF,$03,$FD,$5B,$C1,$FE,$13,$BD
	.dc.b	$E7,$DB,$3E,$B9,$FF,$AA,$3C,$0E,$94,$CF,$A5,$6B,$1F,$F1,$CE,$3C
	.dc.b	$EB,$F9,$87,$E7,$2F,$CC,$BF,$CA,$FF,$4A,$FC,$4B,$FC,$F3,$F8,$6F
	.dc.b	$4F,$21,$7B,$FB,$15,$EF,$CB,$7E,$2B,$BA,$E9,$9F,$F7,$BE,$6F,$EB
	.dc.b	$18,$F2,$F3,$FF,$CC,$31,$8A,$E9,$8F,$87,$CA,$FF,$5B,$4E,$76,$F2
	.dc.b	$EC,$C3,$39,$7F,$86,$32,$F2,$D6,$B4,$7E,$63,$E3,$8C,$73,$7C,$6F
	.dc.b	$80,$FD,$67,$90,$F8,$FF,$F7,$97,$80,$F8,$F7,$ED,$BB,$F3,$3F,$73
	.dc.b	$5A,$97,$E5,$19,$C3,$D7,$BF,$F8,$AB,$2F,$DC,$7E,$B3,$88,$FE,$7F
	.dc.b	$8F,$2F,$83,$FD,$67,$47,$AA,$73,$87,$8E,$7F,$61,$F5,$79,$C7,$CF
	.dc.b	$A7,$87,$F2,$AF,$03,$E3,$FD,$2B,$F0,$3F,$AC,$3F,$6E,$7E,$FF,$FE
	.dc.b	$36,$FD,$05,$E7,$FF,$BC,$7F,$06,$DD,$6B,$E8,$15,$AF,$C3,$5F,$A7
	.dc.b	$EB,$C8,$F5,$67,$D7,$3B,$8F,$1C,$3A,$FD,$7F,$AE,$B5,$97,$4D,$3F
	.dc.b	$C6,$BE,$1F,$F8,$66,$73,$7D,$FF,$2E,$F1,$E5,$E7,$B6,$AF,$D4,$73
	.dc.b	$97,$38,$FA,$DE,$5F,$E6,$F2,$E7,$7C,$3F,$34,$FF,$2E,$F8,$1F,$E2
	.dc.b	$7F,$B2,$FF,$3C,$E4,$FC,$75,$F7,$4F,$E0,$1F,$D4,$7C,$67,$0E,$FF
	.dc.b	$F3,$1E,$1C,$2F,$75,$FA,$DF,$4B,$55,$FE,$90,$FA,$FF,$33,$FE,$59
	.dc.b	$CF,$C1,$0E,$3F,$F5,$FC,$63,$4F,$16,$9E,$1F,$CA,$3F,$CE,$3C,$3F
	.dc.b	$CB,$F1,$F0,$E3,$F0,$74,$F0,$FC,$6E,$B2,$E5,$1E,$4F,$5B,$FC,$FF
	.dc.b	$18,$CB,$D9,$0F,$9F,$F3,$FD,$7B,$F7,$8F,$E2,$FF,$50,$FD,$37,$FD
	.dc.b	$F5,$F8,$83,$C9,$DD,$6A,$73,$F9,$F6,$3E,$07,$D7,$C3,$FE,$88,$CB
	.dc.b	$A7,$97,$88,$47,$79,$FF,$24,$DB,$E5,$FF,$C9,$75,$CF,$4F,$14,$7F
	.dc.b	$3A,$8C,$4F,$93,$D1,$B2,$D7,$F6,$BC,$BF,$1E,$DF,$2C,$FD,$75,$AF
	.dc.b	$FA,$5F,$ED,$FF,$CA,$3F,$B8,$FF,$AD,$F0,$E1,$7B,$AD,$7D,$4A,$9F
	.dc.b	$75,$F7,$AB,$59,$7D,$D3,$FD,$4E,$F7,$FD,$C7,$E0,$3C,$FE,$C7,$FD
	.dc.b	$93,$87,$F6,$DE,$AF,$F3,$C4,$E7,$4F,$F3,$1F,$AD,$F6,$F1,$E7,$B7
	.dc.b	$A3,$1E,$3A,$3F,$45,$F1,$4E,$70,$F5,$FD,$96,$37,$F4,$DB,$F2,$DC
	.dc.b	$63,$CB,$E9,$BF,$D7,$FF,$7F,$FE,$89,$F9,$0F,$D0,$73,$7F,$85,$BE
	.dc.b	$2D,$F9,$9B,$F0,$FF,$CC,$74,$30,$E1,$E6,$7D,$9F,$A0,$FE,$87,$7B
	.dc.b	$FC,$7F,$AF,$2F,$FB,$2B,$4F,$D6,$7D,$37,$FA,$5E,$3F,$D3,$19,$73
	.dc.b	$75,$8F,$E0,$DF,$F6,$06,$5D,$B5,$F2,$65,$E3,$9E,$A3,$1F,$4E,$7E
	.dc.b	$2F,$C3,$BF,$53,$87,$18,$F2,$EA,$FC,$C7,$8D,$F0,$DF,$DF,$2F,$D1
	.dc.b	$BA,$FD,$C3,$F3,$DF,$E9,$DF,$20,$FC,$49,$F4,$BF,$EE,$0F,$C8,$77
	.dc.b	$5A,$9C,$FB,$EB,$CF,$85,$F0,$78,$BC,$A7,$FB,$BB,$2E,$DF,$2B,$98
	.dc.b	$E6,$F1,$C3,$BF,$DA,$70,$9F,$F5,$56,$5D,$FE,$3B,$E5,$F2,$7F,$B0
	.dc.b	$F1,$E7,$9F,$9F,$6F,$F7,$AF,$C0,$F4,$8C,$6C,$FD,$7F,$E5,$FD,$EF
	.dc.b	$F1,$5F,$41,$F9,$9F,$F8,$5F,$DB,$7E,$77,$FE,$2C,$F0,$3E,$C7,$0F
	.dc.b	$FA,$C7,$E0,$3F,$D8,$77,$BA,$7E,$EF,$D2,$7F,$A5,$D7,$DC,$07,$1C
	.dc.b	$7E,$06,$1D,$32,$F1,$63,$E1,$E2,$F8,$11,$8F,$A2,$68,$F0,$33,$9F
	.dc.b	$8A,$73,$AF,$EF,$79,$3F,$DC,$D8,$71,$C3,$D7,$E9,$F9,$75,$F7,$E1
	.dc.b	$C3,$CE,$C6,$BE,$BD,$3F,$1F,$F9,$37,$B3,$E4,$1F,$96,$BF,$56,$7E
	.dc.b	$06,$FC,$C7,$FD,$69,$C9,$F1,$B0,$E1,$7B,$AD,$7E,$83,$D3,$9D,$F9
	.dc.b	$B7,$C2,$E4,$FF,$7E,$61,$DF,$A3,$E4,$07,$9E,$9E,$0F,$8D,$FC,$24
	.dc.b	$63,$2F,$E9,$D5,$D7,$6F,$3D,$F9,$B5,$7E,$15,$9D,$BC,$90,$FD,$68
	.dc.b	$BF,$33,$FA,$F8,$BF,$3C,$7E,$B8,$FB,$77,$E2,$BF,$ED,$3E,$67,$EC
	.dc.b	$9F,$E0,$AC,$B9,$63,$F8,$0C,$78,$E1,$C3,$C9,$FB,$56,$EC,$F0,$E9
	.dc.b	$E0,$BE,$FF,$8F,$7E,$1A,$31,$AF,$FB,$E3,$2F,$F7,$EC,$63,$2E,$9A
	.dc.b	$78,$A7,$99,$E1,$AF,$46,$BE,$7F,$DE,$3F,$B7,$E1,$C6,$3D,$5A,$7F
	.dc.b	$98,$76,$F4,$E1,$F0,$70,$E3,$B7,$A3,$F2,$6E,$B2,$FB,$47,$EB,$AF
	.dc.b	$D5,$7F,$C5,$7F,$9C,$BE,$2D,$F9,$87,$CD,$D3,$AE,$1C,$2F,$75,$AF
	.dc.b	$CB,$3A,$CE,$72,$FE,$15,$7C,$32,$FF,$2C,$46,$38,$3E,$BD,$73,$FA
	.dc.b	$36,$9E,$1F,$D1,$FD,$F7,$FB,$2A,$D4,$F9,$7F,$13,$CC,$FA,$F1,$F5
	.dc.b	$61,$CB,$4F,$4F,$E9,$5E,$49,$CE,$1E,$DD,$DF,$B9,$67,$83,$D7,$FB
	.dc.b	$5B,$F7,$9C,$67,$CE,$1F,$B4,$7C,$AE,$07,$E5,$CF,$D4,$1F,$A9,$3E
	.dc.b	$C5,$FA,$4F,$F8,$6F,$F7,$8E,$17,$BA,$D4,$FF,$87,$FF,$16,$CF,$89
	.dc.b	$F8,$9E,$6F,$E2,$39,$F6,$CF,$10,$8C,$61,$CB,$DB,$F8,$71,$8D,$3C
	.dc.b	$3E,$1E,$34,$F0,$47,$93,$E9,$5F,$91,$79,$33,$F4,$CE,$7A,$7F,$0F
	.dc.b	$CC,$FD,$01,$F9,$43,$F5,$1F,$F2,$37,$9C,$F9,$DD,$78,$7F,$92,$C3
	.dc.b	$85,$EF,$E0,$7F,$77,$4F,$D9,$3A,$E5,$FE,$68,$E5,$FC,$21,$FC,$C3
	.dc.b	$0E,$2E,$0F,$AF,$1E,$2C,$FC,$3F,$8F,$FE,$0A,$BB,$C7,$93,$C9,$C5
	.dc.b	$6F,$2F,$0F,$E9,$BE,$0D,$3C,$9B,$7A,$7F,$58,$FE,$17,$E5,$7F,$0A
	.dc.b	$FA,$37,$E2,$5F,$73,$F3,$77,$EA,$7F,$EE,$5F,$84,$FF,$59,$72,$B0
	.dc.b	$E1,$7B,$AD,$79,$1D,$39,$3F,$A5,$E5,$D3,$C0,$FE,$97,$87,$0F,$36
	.dc.b	$D6,$B1,$E5,$F0,$BF,$D2,$63,$90,$C3,$96,$BE,$49,$FF,$08,$69,$E3
	.dc.b	$AE,$59,$2B,$5A,$FF,$84,$EB,$55,$CF,$1E,$F3,$FE,$E8,$9C,$E3,$DB
	.dc.b	$5F,$81,$F9,$37,$8A,$31,$9F,$F9,$67,$C4,$F2,$FE,$DD,$FC,$97,$F4
	.dc.b	$1F,$F5,$9F,$CD,$7F,$0E,$FD,$67,$FB,$1E,$7E,$0A,$D7,$C0,$FF,$BD
	.dc.b	$BD,$35,$3F,$72,$E7,$EF,$B2,$F8,$BC,$4C,$7F,$99,$7E,$75,$87,$9E
	.dc.b	$9F,$C7,$3E,$E3,$FD,$2B,$D1,$FD,$12,$73,$A3,$F0,$FC,$E3,$FD,$FA
	.dc.b	$73,$AB,$F2,$EF,$0E,$9E,$AC,$3C,$BF,$A7,$7D,$2F,$C5,$9F,$A1,$BF
	.dc.b	$26,$7E,$BA,$FD,$B1,$F8,$BB,$D7,$7E,$9F,$C7,$0E,$17,$BA,$D7,$E7
	.dc.b	$59,$FC,$C3,$DF,$9F,$6C,$EE,$1C,$33,$EB,$9F,$AB,$89,$7B,$CF,$A7
	.dc.b	$DB,$E6,$31,$7C,$3F,$26,$F1,$65,$E8,$8E,$7F,$31,$FD,$B3,$F4,$BF
	.dc.b	$ED,$7F,$AC,$BF,$49,$FF,$47,$7E,$E8,$FD,$25,$F8,$8B,$F7,$57,$E5
	.dc.b	$FC,$78,$7F,$C5,$38,$F1,$FC,$01,$8F,$1C,$38,$7C,$8E,$ED,$7B,$7F
	.dc.b	$A4,$E9,$E2,$D1,$D5,$FE,$32,$8C,$65,$D3,$5F,$1D,$FF,$B5,$27,$39
	.dc.b	$F7,$EA,$C4,$F9,$31,$F5,$DF,$AB,$67,$95,$E3,$D3,$D1,$8F,$9E,$7F
	.dc.b	$D8,$BE,$4B,$3E,$B8,$71,$FD,$47,$86,$5D,$38,$3E,$27,$E8,$3E,$DF
	.dc.b	$C6,$3F,$A3,$75,$FF,$0A,$FA,$3F,$F5,$07,$E8,$EF,$C2,$B5,$F1,$1F
	.dc.b	$87,$F5,$3F,$D3,$5F,$82,$AB,$FD,$93,$BF,$F8,$7B,$2E,$7C,$BF,$18
	.dc.b	$FF,$2B,$BF,$D4,$FF,$2B,$F8,$D1,$8C,$7A,$67,$FE,$AF,$FB,$8F,$87
	.dc.b	$2B,$F2,$E7,$E0,$F0,$73,$18,$BF,$3F,$EB,$BE,$6D,$FD,$5B,$79,$7F
	.dc.b	$09,$FB,$5E,$EF,$E1,$3E,$5F,$FC,$8F,$FA,$9B,$F7,$B7,$D7,$3F,$06
	.dc.b	$DF,$42,$F7,$5A,$BC,$C7,$AB,$87,$FB,$6D,$7F,$61,$7E,$01,$88,$E3
	.dc.b	$F1,$7E,$CC,$BC,$5F,$5E,$C1,$FB,$F7,$FA,$F3,$A3,$FC,$8F,$5A,$9F
	.dc.b	$36,$AF,$D4,$3C,$38,$71,$D7,$D3,$FA,$3F,$E5,$FF,$63,$7B,$EF,$82
	.dc.b	$7E,$E8,$FD,$D3,$F1,$7E,$EF,$F8,$0B,$C7,$FC,$47,$27,$FA,$33,$1E
	.dc.b	$38,$70,$BD,$FD,$5F,$5C,$BA,$5F,$BA,$FB,$EB,$F0,$DE,$39,$F7,$F3
	.dc.b	$1F,$05,$F7,$27,$39,$7E,$73,$E0,$71,$97,$9F,$0E,$1B,$7F,$5E,$DB
	.dc.b	$CB,$A3,$1E,$39,$FF,$34,$FB,$AF,$26,$FE,$78,$FE,$6F,$F6,$98,$CB
	.dc.b	$9F,$97,$FE,$74,$FD,$CF,$FC,$E7,$FC,$01,$FC,$03,$95,$FB,$4F,$FA
	.dc.b	$4F,$F4,$87,$E5,$4F,$CC,$B8,$70,$7F,$AC,$30,$E1,$7B,$A7,$EF,$3A
	.dc.b	$A5,$75,$7F,$C9,$F9,$78,$3D,$6F,$BF,$18,$3A,$3F,$93,$C3,$C5,$D1
	.dc.b	$FE,$BE,$E4,$5F,$1D,$7C,$FF,$76,$D9,$F0,$5B,$AD,$63,$E5,$FD,$EF
	.dc.b	$C9,$AF,$A3,$0E,$1B,$BF,$52,$F4,$78,$0F,$CA,$5F,$8F,$3F,$78,$FE
	.dc.b	$95,$F9,$9B,$F7,$E7,$EF,$7C,$2F,$75,$A9,$FC,$77,$1F,$A3,$9B,$FD
	.dc.b	$AF,$97,$5C,$BF,$DA,$FC,$3F,$EA,$48,$F0,$7E,$67,$8A,$DE,$5D,$3E
	.dc.b	$29,$AF,$F4,$A9,$F0,$63,$E5,$9C,$FB,$5F,$06,$78,$CE,$6B,$D1,$FB
	.dc.b	$0F,$0D,$BC,$DF,$BE,$7D,$7F,$D0,$9F,$A6,$BF,$33,$7E,$3F,$FE,$29
	.dc.b	$FC,$89,$FB,$DB,$F5,$4E,$17,$BA,$D4,$FF,$B6,$F9,$7E,$06,$EC,$BA
	.dc.b	$FE,$ED,$ED,$8C,$61,$DD,$FC,$11,$8C,$FC,$3F,$3E,$D1,$F8,$87,$B6
	.dc.b	$B5,$A7,$97,$F2,$1C,$67,$EA,$C7,$8E,$1E,$8F,$DB,$3C,$F1,$8D,$9F
	.dc.b	$1B,$8C,$7D,$3F,$3F,$75,$F0,$53,$F7,$2F,$F6,$DF,$E4,$1F,$ED,$AF
	.dc.b	$A5,$FF,$77,$FE,$71,$FE,$96,$F8,$0F,$F6,$E7,$23,$F4,$DD,$F1,$85
	.dc.b	$F0,$FB,$9F,$F4,$9F,$D5,$AD,$FB,$7F,$0B,$DF,$C0,$7F,$BC,$9F,$F7
	.dc.b	$49,$9C,$3C,$2F,$FA,$EB,$4F,$24,$B4,$F2,$63,$FE,$53,$AE,$13,$9C
	.dc.b	$BC,$3F,$8D,$FE,$B7,$0E,$38,$70,$CF,$F0,$7D,$69,$CD,$FA,$25,$F3
	.dc.b	$1E,$7F,$9E,$7E,$A6,$FD,$59,$FA,$EB,$E5,$7F,$DE,$9F,$57,$FE,$14
	.dc.b	$D3,$A7,$07,$C5,$C7,$8E,$1C,$2F,$7E,$EF,$5C,$BA,$5F,$BC,$E0,$B8
	.dc.b	$3E,$2C,$7A,$CC,$DF,$3F,$AF,$79,$79,$9F,$F0,$BD,$F9,$F5,$F3,$57
	.dc.b	$F5,$9D,$3C,$B5,$AC,$39,$6A,$F5,$FC,$91,$8F,$53,$EC,$70,$3E,$FD
	.dc.b	$F2,$3F,$07,$F3,$4F,$F7,$C7,$E8,$4F,$C8,$77,$5A,$E4,$56,$BF,$08
	.dc.b	$CF,$C9,$7B,$37,$DE,$57,$83,$CB,$A7,$78,$ED,$F8,$A7,$F2,$2A,$DE
	.dc.b	$5D,$23,$FA,$8C,$63,$2F,$1E,$B6,$4F,$F0,$EC,$7A,$E9,$E5,$AF,$4F
	.dc.b	$E9,$1B,$A7,$CE,$7E,$5F,$DC,$7E,$68,$FD,$27,$FC,$A3,$AB,$F4,$2F
	.dc.b	$EE,$7F,$63,$BA,$D7,$CF,$7E,$16,$7F,$BB,$FE,$B7,$9C,$3D,$DE,$27
	.dc.b	$EA,$78,$BF,$A0,$E0,$FD,$43,$31,$FA,$DD,$3C,$37,$E0,$C5,$FA,$1F
	.dc.b	$DF,$9C,$E7,$DF,$F4,$6C,$65,$E2,$DB,$CD,$F3,$5F,$E9,$BE,$5F,$C0
	.dc.b	$E1,$C2,$73,$29,$7E,$51,$E5,$FA,$BF,$F0,$67,$EA,$CF,$D1,$1F,$A4
	.dc.b	$BF,$4F,$7E,$E2,$FD,$8F,$96,$3C,$7E,$91,$8F,$1C,$38,$5B,$F7,$BD
	.dc.b	$F3,$7F,$0B,$D3,$BB,$FA,$0D,$7C,$6E,$71,$7E,$9D,$F8,$7C,$FB,$67
	.dc.b	$D7,$5F,$E0,$98,$70,$C1,$8F,$AB,$6F,$4D,$7E,$B6,$7C,$26,$B2,$F3
	.dc.b	$CE,$B4,$7E,$D3,$AC,$39,$6D,$EA,$FD,$63,$8F,$E5,$CF,$D8,$BF,$BC
	.dc.b	$7E,$C9,$FD,$C7,$E3,$FF,$92,$FD,$3E,$A7,$3E,$77,$F7,$1F,$79,$F2
	.dc.b	$D9,$F9,$FF,$67,$17,$F2,$DC,$FA,$74,$62,$11,$DE,$5A,$77,$EE,$FF
	.dc.b	$06,$74,$7E,$36,$B5,$3E,$3F,$03,$38,$75,$D7,$C7,$BF,$F3,$AE,$6F
	.dc.b	$15,$FA,$76,$7E,$A1,$8E,$0F,$47,$E9,$97,$0F,$E8,$3F,$2B,$F2,$BF
	.dc.b	$2F,$7E,$4F,$FE,$60,$F3,$DF,$82,$3F,$13,$FE,$39,$EF,$EE,$B5,$39
	.dc.b	$F8,$0F,$F6,$CF,$85,$F6,$71,$E7,$F6,$DF,$05,$EB,$FB,$C5,$C3,$8E
	.dc.b	$5F,$6E,$F8,$46,$3F,$13,$F7,$F3,$30,$E7,$AF,$93,$0F,$E5,$DE,$5F
	.dc.b	$8F,$3F,$E2,$FA,$F8,$FE,$E3,$F0,$BE,$E3,$F5,$FF,$EB,$FF,$48,$FD
	.dc.b	$15,$F9,$73,$F5,$47,$ED,$BC,$2F,$75,$A9,$7E,$FB,$9F,$D2,$3E,$E6
	.dc.b	$5D,$7B,$AE,$1C,$32,$E9,$C7,$CE,$7F,$C6,$D9,$78,$3C,$5F,$6E,$1C
	.dc.b	$75,$F2,$61,$E3,$FD,$6B,$84,$63,$5F,$EB,$3D,$0F,$CF,$7F,$89,$7E
	.dc.b	$50,$FC,$E7,$E3,$7E,$B7,$FD,$E7,$E3,$DF,$B3,$3F,$6F,$FE,$0D,$87
	.dc.b	$0B,$DF,$CB,$2F,$7F,$75,$AE,$4F,$F0,$36,$7D,$7B,$FF,$9F,$63,$C7
	.dc.b	$F3,$9B,$5A,$CB,$9E,$2F,$C9,$3F,$CC,$D7,$DF,$06,$DE,$5F,$C9,$B3
	.dc.b	$8F,$19,$CE,$BE,$3D,$DF,$86,$F8,$E9,$B7,$8F,$1F,$75,$3F,$12,$D4
	.dc.b	$E7,$0E,$5F,$1D,$8D,$FC,$B7,$FD,$2F,$5F,$E7,$5F,$42,$FD,$49,$F8
	.dc.b	$CB,$F1,$E7,$E9,$1F,$C7,$FE,$E9,$FB,$FF,$8F,$0E,$17,$BA,$D7,$E8
	.dc.b	$B9,$C7,$F1,$BC,$5F,$C8,$F3,$EF,$1D,$A3,$07,$F8,$CF,$0E,$79,$F8
	.dc.b	$3D,$FF,$F2,$6C,$F3,$CB,$B4,$3E,$D7,$E4,$CF,$F8,$DB,$0F,$4D,$F1
	.dc.b	$D7,$FA,$27,$6F,$E1,$EB,$B5,$6B,$5F,$67,$DB,$7C,$DF,$D3,$5F,$94
	.dc.b	$FC,$0F,$CD,$5C,$2F,$CD,$FF,$D2,$3F,$9A,$6E,$B5,$39,$FC,$01,$F9
	.dc.b	$AE,$7B,$BF,$0B,$B3,$3E,$BA,$E1,$BA,$76,$D3,$F2,$55,$D6,$93,$9F
	.dc.b	$C9,$FD,$F7,$C3,$3F,$07,$EE,$7B,$AE,$5A,$3F,$73,$FE,$C9,$FB,$8F
	.dc.b	$F7,$7F,$99,$FF,$29,$7D,$EF,$F5,$FF,$7D,$FB,$FB,$F7,$CF,$F4,$35
	.dc.b	$EF,$ED,$BF,$CE,$7F,$3B,$FC,$92,$B5,$F9,$57,$4E,$3F,$F3,$56,$7D
	.dc.b	$1F,$EC,$18,$71,$FA,$DF,$F1,$23,$CF,$4F,$06,$1F,$E2,$0E,$C4,$E7
	.dc.b	$2E,$DF,$8D,$FF,$B3,$63,$1B,$79,$B6,$FF,$39,$DF,$0C,$FC,$3B,$79
	.dc.b	$AB,$5F,$77,$E5,$96,$DF,$03,$3F,$56,$3F,$E6,$AC,$BA,$E9,$F0,$7D
	.dc.b	$0E,$1F,$8F,$3F,$74,$7E,$EA,$FD,$E5,$EE,$B8,$B1,$F8,$87,$0B,$DD
	.dc.b	$6A,$78,$7E,$81,$9F,$98,$F8,$7C,$3F,$E9,$1C,$FB,$67,$88,$46,$30
	.dc.b	$E5,$F6,$BF,$0E,$78,$65,$DB,$47,$D9,$FF,$8C,$F9,$FE,$25,$72,$AD
	.dc.b	$6B,$E9,$F9,$4F,$36,$9E,$BC,$7D,$59,$BF,$0E,$F9,$12,$FC,$B7,$F7
	.dc.b	$FF,$83,$BF,$00,$FF,$68,$FE,$C5,$FD,$2F,$6E,$F7,$BF,$5D,$7B,$FC
	.dc.b	$29,$7F,$C2,$2F,$78,$FE,$B3,$83,$F0,$E7,$FA,$CF,$1E,$4E,$0F,$A3
	.dc.b	$1E,$2C,$BF,$DC,$3F,$05,$F0,$E7,$39,$F8,$27,$FD,$CB,$EB,$E7,$0F
	.dc.b	$26,$9E,$3C,$FF,$81,$F5,$FF,$96,$67,$37,$EC,$F4,$BF,$11,$FA,$27
	.dc.b	$FB,$AF,$C3,$7E,$74,$FD,$C1,$F9,$D3,$F1,$0F,$DF,$F1,$FD,$0C,$38
	.dc.b	$5E,$EB,$5F,$B1,$74,$FC,$57,$FD,$85,$C7,$ED,$BF,$F7,$C7,$E8,$5F
	.dc.b	$81,$78,$E7,$DB,$C4,$F7,$46,$31,$7E,$9B,$8C,$7F,$D9,$99,$F7,$D5
	.dc.b	$FA,$1F,$7D,$FD,$39,$FA,$25,$FA,$0E,$6F,$87,$C8,$5A,$D7,$C8,$7F
	.dc.b	$9A,$FF,$69,$FF,$3F,$7D,$2B,$F1,$77,$D1,$7F,$AA,$BE,$53,$F1,$1F
	.dc.b	$94,$FF,$82,$BA,$18,$70,$BD,$F9,$6B,$DF,$81,$D7,$A7,$FD,$E5,$9F
	.dc.b	$5F,$C1,$3F,$D5,$58,$F1,$D2,$D6,$B2,$E7,$8F,$E5,$F9,$BF,$DF,$5A
	.dc.b	$78,$2B,$57,$E5,$F1,$9F,$96,$EB,$2E,$9B,$78,$E5,$BB,$DC,$F2,$63
	.dc.b	$FE,$D9,$D7,$BD,$6B,$ED,$73,$EC,$BE,$21,$FB,$7B,$CC,$FE,$0D,$F1
	.dc.b	$BF,$4B,$D5,$DD,$6A,$73,$DF,$FC,$33,$8F,$D1,$AF,$F9,$42,$9F,$B0
	.dc.b	$FC,$08,$C6,$57,$3F,$03,$DF,$D0,$FE,$C5,$E4,$7B,$68,$70,$CF,$C7
	.dc.b	$E3,$E6,$31,$5F,$DA,$74,$7E,$FF,$E6,$C7,$A6,$FF,$8A,$FC,$B7,$E1
	.dc.b	$FC,$83,$C6,$7E,$9E,$F8,$7F,$EC,$9F,$B8,$FF,$18,$FC,$5F,$75,$A9
	.dc.b	$CF,$BA,$FD,$A7,$38,$7B,$71,$E9,$FA,$2D,$C3,$87,$C8,$33,$7E,$ED
	.dc.b	$E9,$BF,$16,$1E,$1A,$D7,$EA,$B9,$C3,$8E,$1F,$CA,$BE,$FB,$BE,$BE
	.dc.b	$2F,$D0,$EE,$1D,$76,$FD,$5F,$6F,$D2,$D3,$AD,$FA,$3D,$4F,$F6,$67
	.dc.b	$EE,$DF,$ED,$FF,$AB,$7E,$42,$FC,$F9,$FA,$DB,$F6,$F7,$E5,$DF,$EC
	.dc.b	$8F,$D1,$B8,$F8,$1F,$53,$1E,$38,$70,$BD,$FE,$23,$FD,$5B,$76,$9D
	.dc.b	$F4,$7E,$A3,$FE,$EF,$CB,$F9,$F4,$E7,$1C,$47,$E8,$A3,$1A,$F8,$F5
	.dc.b	$FE,$0F,$3E,$21,$A7,$8A,$75,$F0,$1C,$67,$CD,$97,$B6,$79,$68,$F5
	.dc.b	$F9,$61,$C7,$6F,$4F,$D0,$F5,$F8,$E7,$CC,$3F,$3C,$FF,$77,$FE,$F5
	.dc.b	$FD,$13,$E2,$5F,$A6,$6E,$B5,$F8,$82,$B5,$F0,$19,$FD,$4B,$DB,$C5
	.dc.b	$BC,$AF,$47,$C6,$8F,$CF,$E9,$E0,$9E,$FF,$1C,$F8,$2F,$65,$7F,$7B
	.dc.b	$D3,$BE,$7F,$D5,$BF,$21,$CE,$3C,$75,$F1,$63,$FE,$14,$FE,$01,$F0
	.dc.b	$46,$38,$3F,$83,$6B,$E1,$CF,$FA,$17,$C0,$FF,$9F,$FC,$9F,$D2,$7E
	.dc.b	$36,$FD,$81,$F8,$87,$D1,$F8,$DF,$E2,$7D,$DF,$A5,$AD,$7E,$1F,$9C
	.dc.b	$5F,$A0,$FB,$BC,$2F,$DD,$E9,$DB,$0C,$3F,$BE,$C7,$9E,$9E,$0F,$86
	.dc.b	$F7,$C6,$3C,$86,$8E,$5C,$E9,$E7,$9C,$EE,$F9,$AF,$16,$9E,$58,$FF
	.dc.b	$02,$70,$FB,$F2,$E9,$97,$B7,$DA,$DE,$BF,$02,$FF,$CA,$17,$F1,$7F
	.dc.b	$5A,$7A,$EF,$DE,$FF,$D9,$FF,$43,$FE,$2C,$FD,$F9,$F9,$17,$E5,$6F
	.dc.b	$75,$AF,$9A,$53,$F5,$7D,$63,$CF,$9B,$F4,$39,$F7,$87,$E3,$9F,$1A
	.dc.b	$59,$74,$FC,$17,$FC,$BF,$93,$F5,$8F,$C2,$CE,$76,$FF,$1F,$FD,$A6
	.dc.b	$32,$F4,$61,$C3,$57,$E5,$BE,$2D,$FC,$D1,$FA,$5D,$9F,$3D,$F6,$3F
	.dc.b	$5F,$7E,$72,$FC,$EB,$FD,$15,$F1,$DE,$BF,$CE,$3F,$84,$E1,$C3,$8F
	.dc.b	$F9,$CE,$1C,$2F,$7C,$AB,$FF,$29,$5E,$ED,$7D,$BF,$33,$FD,$2E,$7E
	.dc.b	$1C,$FC,$15,$F1,$A3,$19,$74,$FC,$27,$FB,$4C,$63,$E0,$9A,$FC,$FF
	.dc.b	$D6,$B3,$87,$93,$1F,$55,$6B,$67,$E7,$6F,$C0,$5B,$F9,$77,$F5,$5E
	.dc.b	$E9,$F6,$9A,$D7,$F3,$BF,$97,$7F,$9B,$BF,$40,$7E,$E2,$FD,$82,$F8
	.dc.b	$6F,$D5,$B8,$70,$FF,$5E,$C3,$85,$EC,$FE,$AB,$F9,$EE,$A5,$3D,$5F
	.dc.b	$84,$71,$F0,$9F,$5D,$F2,$23,$19,$F8,$B3,$F1,$FE,$4F,$F0,$A7,$9D
	.dc.b	$AB,$59,$7F,$2C,$9F,$06,$3E,$4C,$3D,$3F,$73,$8C,$38,$5F,$9F,$57
	.dc.b	$E9,$79,$CB,$9C,$7E,$C3,$7F,$8B,$7C,$31,$F8,$7F,$92,$FB,$7B,$9F
	.dc.b	$A0,$3F,$56,$FC,$7F,$EB,$0F,$C5,$DF,$B5,$7F,$AB,$BA,$FF,$11,$87
	.dc.b	$0B,$DD,$6B,$EB,$FA,$63,$CF,$93,$E3,$9F,$91,$7F,$A5,$BF,$D4,$F0
	.dc.b	$E7,$1E,$1E,$F6,$7E,$1F,$A5,$F9,$19,$F7,$9C,$E5,$FE,$FC,$C3,$8E
	.dc.b	$7E,$2D,$3F,$47,$C1,$FC,$83,$0E,$19,$7A,$E7,$FD,$63,$7F,$A1,$FD
	.dc.b	$19,$F8,$87,$F0,$AF,$D6,$9F,$A8,$BF,$7C,$FD,$F7,$ED,$BF,$83,$BD
	.dc.b	$FA,$CB,$DD,$7F,$03,$F1,$F5,$8F,$DB,$E4,$F7,$6F,$FC,$CB,$1E,$31
	.dc.b	$F1,$1E,$78,$72,$F5,$7F,$0D,$18,$D3,$C5,$E9,$E3,$1F,$34,$79,$6B
	.dc.b	$F8,$5E,$DE,$B9,$CE,$9F,$C5,$7E,$53,$F5,$7F,$B0,$7F,$42,$FC,$6D
	.dc.b	$F8,$07,$E0,$3F,$08,$FF,$12,$FB,$8F,$CD,$7F,$DB,$19,$B0,$E1,$7B
	.dc.b	$FA,$9F,$EA,$95,$FC,$D2,$B5,$C5,$FE,$48,$CB,$AF,$77,$F6,$EC,$78
	.dc.b	$FE,$FB,$6B,$59,$73,$CD,$F8,$AF,$F9,$CA,$FB,$E1,$FE,$D4,$D3,$FB
	.dc.b	$17,$E1,$19,$C7,$8C,$E7,$0F,$47,$E7,$FE,$E9,$CE,$FE,$CD,$FF,$C4
	.dc.b	$3B,$F9,$A5,$BF,$9A,$B5,$F1,$3C,$67,$3B,$F9,$70,$FF,$19,$E1,$C7
	.dc.b	$EF,$9F,$41,$FD,$D3,$C0,$FF,$6F,$7E,$7A,$FD,$ED,$F9,$5F,$3B,$DF
	.dc.b	$CA,$7E,$73,$F0,$D5,$EF,$F6,$3D,$64,$FC,$47,$E3,$6F,$FE,$8C,$8F
	.dc.b	$F0,$0B,$FD,$5B,$A3,$FB,$DE,$5D,$35,$F0,$F4,$7C,$5A,$E1,$A7,$7C
	.dc.b	$DF,$69,$9C,$38,$6B,$E2,$DF,$E8,$F0,$7B,$75,$F6,$4E,$7F,$76,$F8
	.dc.b	$9F,$AF,$FF,$84,$FF,$72,$F6,$F9,$EF,$D9,$5F,$96,$FE,$E3,$2F,$9D
	.dc.b	$7B,$E6,$FE,$B3,$5A,$F9,$AE,$7D,$CC,$BD,$F7,$FE,$1F,$DE,$CE,$70
	.dc.b	$E3,$F7,$BE,$DA,$E9,$A7,$87,$C7,$C5,$EE,$3C,$9F,$B7,$7F,$91,$B5
	.dc.b	$F1,$CE,$75,$4B,$EA,$BF,$A7,$FC,$EF,$F8,$9B,$DC,$7D,$BF,$F0,$2F
	.dc.b	$B3,$7E,$F4,$FC,$55,$F5,$1F,$E8,$CD,$7F,$C4,$9D,$DF,$EA,$8C,$78
	.dc.b	$FA,$3F,$E7,$DC,$38,$5F,$E1,$6F,$77,$FE,$A2,$CB,$E1,$3F,$E1,$2C
	.dc.b	$FC,$51,$FC,$4A,$30,$75,$7C,$97,$E3,$FA,$CF,$F6,$CD,$70,$C7,$C1
	.dc.b	$9B,$F6,$2C,$E3,$C6,$73,$9B,$EE,$FE,$0C,$E7,$6F,$7E,$1E,$EF,$BC
	.dc.b	$CC,$F9,$F0,$E3,$D5,$8E,$0F,$7E,$1C,$32,$FF,$71,$F0,$7B,$BC,$4F
	.dc.b	$F4,$DF,$E8,$5F,$D2,$BC,$87,$E6,$3F,$ED,$6F,$C2,$7A,$D6,$BE,$03
	.dc.b	$F4,$53,$FE,$DA,$FC,$17,$3F,$78,$C7,$FC,$F3,$CB,$FE,$5C,$C3,$FD
	.dc.b	$4F,$FA,$1E,$38,$BF,$DB,$31,$CF,$4E,$F8,$FC,$29,$FF,$61,$E3,$CF
	.dc.b	$3E,$FF,$13,$9C,$38,$69,$FE,$EA,$FC,$57,$C3,$AF,$A6,$3F,$DD,$BF
	.dc.b	$B2,$FC,$4F,$DC,$3F,$D2,$1F,$6E,$FA,$BF,$D2,$7E,$DC,$FD,$0D,$F6
	.dc.b	$FB,$AF,$F4,$B5,$6A,$7F,$A7,$73,$7E,$13,$0E,$7C,$3F,$80,$0F,$FA
	.dc.b	$77,$F0,$FF,$74,$63,$0E,$7E,$33,$37,$EB,$DE,$FC,$3C,$B9,$F9,$3F
	.dc.b	$5D,$C6,$1C,$2F,$CD,$F5,$3E,$58,$F5,$63,$CF,$57,$E8,$9F,$82,$FB
	.dc.b	$B7,$E7,$3F,$E8,$3F,$C2,$1E,$93,$F3,$27,$E7,$6F,$0B,$85,$EE,$B5
	.dc.b	$2E,$FC,$FE,$07,$EC,$CB,$AF,$87,$70,$E1,$93,$BF,$ED,$56,$B4,$F0
	.dc.b	$CB,$E6,$F3,$7C,$33,$7D,$EF,$8B,$2F,$F7,$1E,$BE,$2F,$0B,$19,$72
	.dc.b	$BF,$E3,$7D,$9F,$B6,$D3,$B5,$FA,$3F,$66,$FF,$2A,$EF,$FE,$AC,$FD
	.dc.b	$A3,$F5,$DF,$9C,$FF,$04,$FD,$39,$F8,$C3,$F6,$7F,$ED,$59,$FF,$B6
	.dc.b	$FC,$10,$C3,$8F,$98,$C3,$8D,$FF,$AB,$EF,$85,$EF,$D4,$7E,$9B,$FE
	.dc.b	$DA,$D3,$C5,$0D,$31,$5D,$B3,$EB,$97,$93,$DD,$F0,$E5,$D2,$73,$AB
	.dc.b	$E6,$B3,$9F,$4C,$BD,$93,$EB,$FC,$AF,$E0,$E7,$D7,$4F,$06,$FE,$9A
	.dc.b	$F8,$1F,$88,$EA,$BF,$CC,$BF,$0D,$FE,$AC,$FC,$C1,$F7,$1F,$E2,$9F
	.dc.b	$A2,$FF,$19,$EB,$FE,$D6,$FB,$15,$EE,$B5,$F1,$3C,$E7,$3F,$89,$FB
	.dc.b	$B2,$EB,$F8,$25,$9E,$2B,$3E,$FF,$49,$FE,$6D,$E5,$F9,$75,$A9,$CF
	.dc.b	$E7,$FF,$03,$0F,$AB,$39,$CF,$FC,$7D,$F4,$FF,$E9,$0C,$38,$EB,$E5
	.dc.b	$CB,$FD,$A3,$FA,$7F,$BF,$F6,$3F,$F3,$57,$DC,$7C,$BF,$CE,$FD,$7E
	.dc.b	$7B,$F5,$9D,$7D,$7F,$FA,$4F,$B5,$5A,$FC,$BB,$38,$7B,$71,$FF,$3A
	.dc.b	$7D,$45,$C3,$85,$FC,$6F,$6F,$D1,$C6,$BF,$0E,$5D,$37,$7B,$3A,$9C
	.dc.b	$E1,$E4,$FA,$2E,$F7,$BD,$DF,$9E,$62,$BD,$3F,$9D,$5C,$BF,$8E,$7C
	.dc.b	$27,$DD,$FC,$01,$F9,$5F,$F1,$FF,$D0,$5F,$9C,$3F,$12,$7E,$02,$FD
	.dc.b	$33,$85,$EE,$B5,$2F,$CD,$F3,$BF,$DD,$CB,$AF,$EB,$37,$0E,$19,$74
	.dc.b	$E5,$DE,$9E,$08,$7D,$AF,$C4,$C3,$CB,$8A,$B5,$9B,$C3,$E3,$AF,$8B
	.dc.b	$F7,$1F,$F6,$8F,$E3,$DF,$B6,$D1,$F9,$4B,$E7,$7F,$CC,$9F,$A1,$3F
	.dc.b	$31,$7E,$42,$FD,$5D,$F9,$BF,$2C,$7F,$7F,$8F,$1F,$B4,$63,$C7,$0E
	.dc.b	$1F,$A6,$EE,$FB,$E7,$FE,$06,$7C,$5E,$5B,$5F,$26,$5C,$E7,$CC,$7E
	.dc.b	$F2,$33,$5E,$6D,$7C,$BD,$39,$AF,$1C,$E7,$2F,$85,$1E,$1C,$3C,$D9
	.dc.b	$F9,$76,$7E,$3B,$9C,$38,$43,$F6,$8F,$47,$D9,$3F,$6A,$C7,$CE,$BF
	.dc.b	$1C,$FE,$87,$F8,$5F,$F8,$BF,$8B,$D5,$B7,$F2,$2C,$38,$63,$FD,$F6
	.dc.b	$F6,$74,$C7,$B5,$7B,$8C,$69,$5D,$57,$1F,$8A,$FF,$45,$C6,$33,$F1
	.dc.b	$43,$F2,$4F,$85,$3F,$E7,$AC,$3C,$F8,$F9,$BE,$93,$83,$E8,$FC,$0E
	.dc.b	$E0,$D9,$F8,$9E,$1F,$F5,$46,$FE,$CC,$9F,$BB,$7E,$1F,$EB,$9F,$AD
	.dc.b	$7F,$7A,$FC,$D9,$FB,$B3,$F5,$EF,$C0,$5E,$EB,$52,$FD,$7B,$3D,$1F
	.dc.b	$E3,$7C,$BE,$16,$7D,$F0,$FC,$A7,$E8,$DF,$0E,$31,$5F,$E3,$DC,$7C
	.dc.b	$1E,$37,$C0,$BE,$DF,$4E,$D7,$F4,$F9,$76,$DB,$CF,$39,$D5,$FB,$B6
	.dc.b	$72,$F2,$FE,$EC,$D5,$FA,$77,$E9,$F5,$FF,$03,$F8,$CF,$D0,$7F,$D9
	.dc.b	$FF,$2B,$FC,$07,$F7,$4E,$17,$BF,$A7,$FF,$67,$56,$BD,$67,$DF,$EB
	.dc.b	$8B,$FB,$3E,$7F,$E1,$2F,$A1,$F7,$63,$C7,$4B,$5A,$CB,$9F,$BE,$D7
	.dc.b	$F8,$64,$F3,$3F,$89,$CE,$A3,$FC,$4F,$19,$C7,$D3,$A7,$8F,$F3,$6C
	.dc.b	$5E,$F5,$7C,$5F,$3F,$A1,$7E,$30,$E3,$FE,$03,$E0,$7F,$33,$7F,$CB
	.dc.b	$1C,$9F,$1B,$5F,$F0,$0D,$F0,$9D,$FC,$27,$1F,$17,$A6,$3C,$EF,$EF
	.dc.b	$F1,$FF,$13,$C3,$BF,$C2,$FC,$09,$47,$3F,$57,$F2,$B8,$FF,$40,$9E
	.dc.b	$79,$76,$F8,$5F,$F0,$0E,$3C,$F2,$F4,$61,$C2,$9F,$A0,$F0,$8C,$7E
	.dc.b	$1D,$77,$F6,$7D,$6D,$D3,$E0,$79,$1F,$37,$F2,$77,$E9,$AF,$CA,$9F
	.dc.b	$B2,$3F,$7F,$F3,$FE,$01,$F7,$B9,$FE,$27,$23,$1E,$38,$70,$BD,$F8
	.dc.b	$7D,$72,$FF,$16,$5F,$C4,$3F,$C5,$77,$C1,$C1,$CE,$3C,$59,$78,$79
	.dc.b	$9F,$8A,$7C,$2A,$E3,$5A,$C7,$E5,$FE,$8D,$99,$E5,$9F,$8F,$0F,$F6
	.dc.b	$C7,$E7,$7E,$38,$C6,$CF,$D5,$3F,$1D,$D2,$FD,$87,$FA,$87,$53,$F1
	.dc.b	$67,$E5,$1F,$A9,$FD,$43,$75,$A9,$CF,$2F,$FA,$2C,$F8,$B8,$73,$9F
	.dc.b	$C0,$3D,$7E,$2E,$C7,$1C,$1F,$98,$7A,$6B,$59,$78,$67,$FC,$CD,$C1
	.dc.b	$E0,$C3,$F3,$39,$78,$3F,$61,$C5,$EF,$37,$EF,$3F,$BF,$FC,$4F,$FB
	.dc.b	$27,$EF,$5F,$6F,$BD,$FB,$53,$F4,$F7,$ED,$7F,$C2,$3F,$6A,$E1,$BF
	.dc.b	$F9,$83,$0E,$17,$BA,$FE,$1B,$F1,$BA,$A5,$75,$7F,$8F,$E5,$E0,$EC
	.dc.b	$7D,$1F,$F1,$F8,$C6,$9E,$2D,$3C,$3E,$5F,$F9,$7A,$73,$A7,$7F,$2F
	.dc.b	$19,$78,$F2,$FE,$77,$3E,$3F,$65,$FB,$FF,$8F,$5F,$3E,$7E,$C9,$7E
	.dc.b	$61,$9C,$38,$CE,$7F,$2A,$C6,$DE,$3E,$1F,$E5,$3F,$86,$FB,$7F,$15
	.dc.b	$7E,$C6,$FD,$03,$FD,$71,$FB,$7B,$EE,$1F,$BF,$BF,$16,$E1,$7B,$FC
	.dc.b	$45,$7B,$A6,$3A,$FB,$6F,$89,$CD,$FE,$8E,$8E,$F9,$E2,$37,$8F,$3C
	.dc.b	$3C,$1F,$87,$7C,$49,$CD,$79,$34,$FE,$D1,$1E,$28,$C6,$4F,$3B,$1A
	.dc.b	$79,$2F,$D3,$F6,$1F,$6F,$DF,$FE,$49,$F0,$AF,$D1,$1E,$33,$F5,$37
	.dc.b	$E2,$EC,$38,$7C,$03,$0E,$17,$BA,$F0,$79,$0F,$0F,$FC,$6D,$9F,$F3
	.dc.b	$1F,$89,$F7,$3E,$0E,$2F,$C4,$9F,$EB,$07,$C7,$AF,$8B,$5F,$6E,$1C
	.dc.b	$EB,$5A,$78,$6B,$5F,$69,$F1,$70,$E3,$A7,$8F,$C9,$F3,$ED,$E6,$CF
	.dc.b	$D7,$E4,$FF,$65,$FD,$E3,$FC,$95,$FB,$0B,$F4,$97,$EC,$BF,$EB,$AF
	.dc.b	$C0,$1F,$54,$FC,$89,$FA,$97,$1E,$3F,$F6,$3E,$3C,$70,$E1,$7B,$FD
	.dc.b	$8B,$54,$AE,$DF,$1D,$FE,$03,$D3,$C3,$A7,$82,$BF,$C1,$31,$8C,$79
	.dc.b	$FE,$13,$FE,$C4,$8C,$7E,$1C,$DB,$CB,$C9,$9C,$3C,$78,$FA,$AB,$5A
	.dc.b	$FA,$7D,$EF,$1E,$BE,$79,$CF,$07,$F1,$BC,$FA,$E7,$F0,$23,$F8,$BF
	.dc.b	$4A,$73,$8F,$C3,$FC,$DF,$DB,$F8,$1B,$F4,$27,$F0,$0F,$FB,$5F,$F7
	.dc.b	$7F,$D5,$FD,$27,$F1,$6C,$B9,$70,$FF,$BB,$B0,$E1,$7B,$F2,$FF,$C5
	.dc.b	$9F,$BB,$F4,$CB,$9D,$FC,$43,$83,$FA,$07,$FC,$DB,$87,$45,$87,$4F
	.dc.b	$6D,$9F,$EF,$1F,$8D,$3A,$BF,$3C,$6B,$1F,$F0,$A4,$F8,$72,$F2,$66
	.dc.b	$FC,$C3,$17,$B8,$7E,$5B,$F7,$BE,$F9,$FB,$2F,$FA,$DB,$E2,$DF,$88
	.dc.b	$BF,$22,$7D,$97,$FB,$33,$8B,$D1,$87,$0F,$C2,$18,$70,$BD,$FD,$5F
	.dc.b	$5C,$9F,$4B,$F8,$9E,$11,$FE,$BB,$8F,$F6,$2B,$F1,$CC,$61,$D3,$4F
	.dc.b	$17,$D8,$7C,$2C,$FC,$19,$F7,$FC,$BB,$18,$FA,$6F,$D1,$B3,$F6,$CF
	.dc.b	$16,$5C,$F6,$F5,$FD,$7F,$E7,$A5,$F4,$3F,$C9,$BE,$81,$FA,$5B,$F4
	.dc.b	$17,$ED,$FF,$E3,$4E,$5F,$B1,$87,$0F,$7B,$FD,$CD,$7B,$A7,$C4,$75
	.dc.b	$C9,$F0,$3F,$07,$9B,$C9,$1F,$CD,$1F,$81,$8F,$3D,$3C,$3A,$7F,$8B
	.dc.b	$F4,$EF,$7E,$5C,$FC,$9F,$87,$7E,$13,$0E,$11,$E1,$FD,$73,$C1,$B7
	.dc.b	$93,$1F,$4F,$EB,$98,$CF,$AE,$EF,$E0,$0C,$63,$E9,$FC,$9B,$FC,$E5
	.dc.b	$FA,$27,$FB,$5F,$CD,$7D,$DF,$F8,$BF,$EC,$9F,$8F,$7F,$84,$FF,$41
	.dc.b	$7E,$9F,$C3,$9B,$EE,$E1,$C3,$E5,$7F,$DE,$F7,$BF,$D7,$B5,$5F,$C0
	.dc.b	$F2,$7E,$ED,$FE,$8F,$D3,$C1,$1F,$E3,$D9,$FF,$4A,$C6,$30,$E7,$F4
	.dc.b	$DF,$D8,$27,$9E,$5D,$F5,$F2,$CF,$FA,$4B,$A9,$87,$19,$CE,$DF,$06
	.dc.b	$73,$87,$A3,$2F,$76,$BE,$DF,$CE,$FC,$FF,$80,$3E,$D9,$E1,$3D,$F7
	.dc.b	$E8,$4F,$DD,$3F,$13,$F5,$1C,$2F,$75,$A9,$FC,$6F,$0E,$7C,$EF,$81
	.dc.b	$B7,$E9,$33,$ED,$9E,$21,$18,$C3,$97,$0F,$EA,$B8,$3D,$35,$AC,$FB
	.dc.b	$EB,$E3,$FC,$57,$5C,$7F,$B0,$AD,$67,$E9,$AF,$47,$57,$FA,$3F,$1E
	.dc.b	$7A,$77,$9C,$FD,$2F,$D8,$FD,$DD,$FB,$D7,$F4,$CF,$AB,$7E,$DE,$FD
	.dc.b	$8D,$FB,$6B,$C5,$E1,$7B,$AD,$4B,$C5,$CF,$C7,$7B,$F2,$EB,$F9,$A5
	.dc.b	$C3,$86,$5F,$E0,$5E,$2D,$E9,$E0,$8F,$F3,$8C,$63,$3E,$BF,$BE,$59
	.dc.b	$F1,$F6,$DD,$7F,$DD,$7C,$5F,$23,$4F,$5E,$BF,$E6,$BC,$7C,$BD,$3F
	.dc.b	$96,$F5,$3F,$D3,$3F,$A6,$3F,$35,$FF,$2C,$FC,$47,$FC,$01,$F0,$75
	.dc.b	$F8,$12,$B5,$F6,$5F,$DC,$BF,$10,$7C,$1E,$A7,$EE,$65,$FE,$A4,$FD
	.dc.b	$56,$E1,$C7,$3E,$B8,$7D,$8E,$5F,$E4,$F7,$E2,$9F,$0F,$B1,$99,$FE
	.dc.b	$6F,$7C,$73,$F0,$7E,$E7,$DE,$F9,$FA,$7F,$EE,$DE,$FF,$C1,$7C,$43
	.dc.b	$E7,$5F,$8F,$7F,$8A,$3E,$8B,$FB,$27,$EC,$7F,$6F,$AB,$F3,$BB,$7F
	.dc.b	$38,$C3,$85,$EF,$E3,$56,$FD,$B7,$B5,$EF,$2E,$AF,$E4,$4F,$CC,$67
	.dc.b	$E2,$87,$E3,$F9,$9C,$9C,$71,$ED,$E7,$62,$3F,$CB,$F5,$E8,$FA,$6F
	.dc.b	$F1,$C6,$6F,$C2,$F5,$AF,$9A,$73,$AF,$F3,$CD,$BD,$78,$F2,$CF,$D5
	.dc.b	$2F,$D6,$BD,$33,$FE,$53,$FA,$5F,$CE,$FE,$98,$EA,$FE,$17,$E8,$BE
	.dc.b	$87,$7F,$5B,$FE,$52,$F1,$17,$BA,$FE,$23,$F4,$5A,$C7,$F4,$F9,$7B
	.dc.b	$F6,$F1,$E3,$FE,$5A,$FD,$DB,$E1,$4E,$F2,$E7,$A7,$F1,$AF,$CC,$BE
	.dc.b	$15,$F7,$D3,$B7,$D6,$63,$3F,$3D,$F9,$BF,$76,$E5,$5E,$3D,$9F,$BC
	.dc.b	$62,$BD,$7F,$B5,$59,$FE,$4F,$F8,$77,$C2,$FA,$4F,$DF,$FE,$7F,$FE
	.dc.b	$F0,$FC,$B5,$EC,$7F,$A2,$FE,$8F,$5E,$CA,$B5,$3F,$E8,$FE,$5C,$FE
	.dc.b	$B5,$EC,$E7,$FA,$39,$74,$E7,$F8,$5E,$27,$F1,$58,$EF,$F8,$83,$5E
	.dc.b	$FF,$88,$7E,$06,$73,$1E,$2E,$8C,$6D,$E1,$BE,$1A,$FF,$A7,$FD,$1F
	.dc.b	$2E,$FE,$4F,$BD,$7E,$1E,$DD,$D9,$FC,$97,$C1,$7D,$77,$F9,$27,$CC
	.dc.b	$7E,$7F,$FE,$24,$F9,$87,$ED,$FC,$70,$E1,$7B,$AD,$7E,$65,$9E,$7F
	.dc.b	$F0,$FE,$5D,$3B,$3E,$16,$1C,$3E,$BE,$D6,$B1,$E5,$EC,$3F,$37,$FE
	.dc.b	$7D,$7B,$DB,$C7,$2F,$B4,$CE,$1C,$71,$FF,$4D,$6B,$E7,$F8,$0F,$14
	.dc.b	$FE,$DB,$3E,$BC,$1F,$6B,$87,$E0,$CF,$F8,$D3,$F7,$A7,$E8,$EF,$D5
	.dc.b	$DF,$B7,$BF,$2F,$56,$BF,$62,$EA,$F0,$AD,$CE,$AD,$FB,$16,$9E,$98
	.dc.b	$FE,$87,$8F,$FA,$A6,$1E,$0F,$D5,$BE,$1C,$F1,$C7,$9C,$FB,$E3,$94
	.dc.b	$63,$97,$E0,$C6,$F4,$F3,$57,$8F,$5F,$6C,$E7,$4F,$26,$3C,$AF,$79
	.dc.b	$BF,$55,$C5,$7A,$7E,$B7,$E4,$FE,$00,$FC,$8B,$FD,$D7,$FB,$3F,$F9
	.dc.b	$23,$F3,$97,$D2,$3E,$FF,$75,$AF,$AF,$56,$BF,$46,$CF,$D7,$7B,$B2
	.dc.b	$E9,$F1,$D7,$0E,$16,$FB,$0F,$6D,$78,$6D,$A7,$87,$6C,$E1,$C3,$57
	.dc.b	$EC,$BD,$AB,$CD,$0F,$DA,$7F,$07,$EF,$BD,$9F,$F1,$2F,$EA,$8F,$C3
	.dc.b	$5F,$9C,$BF,$15,$7E,$21,$C2,$F7,$5A,$9F,$F4,$CF,$6F,$CE,$C7,$A7
	.dc.b	$B7,$F6,$9E,$DE,$6D,$C3,$8E,$5F,$E3,$5F,$0F,$FC,$29,$5A,$D3,$C5
	.dc.b	$3F,$E3,$DC,$7A,$5F,$0C,$BF,$23,$18,$C5,$7C,$25,$FA,$B7,$EF,$7E
	.dc.b	$BB,$FC,$41,$FB,$5B,$F4,$7F,$E3,$3F,$A2,$7F,$8E,$39,$1F,$BB,$3E
	.dc.b	$9F,$87,$8F,$FE,$96,$C3,$87,$B8,$C3,$85,$BF,$58,$DD,$AF,$B7,$CD
	.dc.b	$FF,$1F,$D3,$C3,$A7,$F2,$8A,$FE,$81,$18,$CB,$A7,$37,$DC,$FA,$BF
	.dc.b	$F1,$E4,$E7,$4F,$37,$D0,$E3,$0F,$26,$3E,$AC,$3F,$07,$AB,$F6,$6F
	.dc.b	$1E,$BE,$8C,$7C,$FF,$B3,$E3,$0E,$1C,$3F,$03,$F7,$BC,$6F,$E2,$8F
	.dc.b	$87,$E1,$FB,$3E,$27,$FD,$45,$F9,$0F,$F9,$83,$D9,$7C,$1F,$F2,$6F
	.dc.b	$C3,$75,$F8,$0A,$B5,$3F,$CA,$BF,$0C,$CC,$7C,$3E,$4F,$F6,$0E,$3F
	.dc.b	$EA,$DF,$2F,$10,$8E,$FF,$6A,$CF,$BE,$DF,$06,$B5,$9F,$7A,$D6,$6F
	.dc.b	$CA,$B5,$8F,$2C,$7C,$FE,$FB,$F3,$2F,$1E,$1C,$21,$F5,$BF,$6B,$F0
	.dc.b	$1F,$F2,$C7,$8A,$FB,$47,$EB,$7F,$C8,$BF,$09,$FE,$A9,$E4,$FF,$A8
	.dc.b	$B8,$7F,$DB,$78,$70,$BD,$FB,$DF,$EE,$9F,$8C,$E9,$F7,$1F,$E1,$0B
	.dc.b	$FF,$39,$61,$C5,$61,$E0,$7F,$0A,$74,$9F,$F4,$CE,$1D,$38,$7E,$15
	.dc.b	$77,$CF,$F9,$AF,$EC,$38,$BD,$C7,$9F,$8B,$C3,$B7,$AE,$3D,$5C,$DF
	.dc.b	$D8,$7D,$2F,$F4,$27,$E6,$1F,$E3,$9F,$09,$F4,$1F,$E1,$AF,$8A,$FE
	.dc.b	$05,$F6,$3B,$E1,$FF,$7C,$DE,$EB,$5E,$8B,$F1,$5D,$52,$BA,$3F,$E1
	.dc.b	$FC,$BB,$FD,$83,$F5,$4F,$F0,$24,$63,$4F,$0F,$D2,$35,$FF,$5D,$4E
	.dc.b	$74,$F2,$7D,$67,$FA,$86,$31,$5B,$D5,$FA,$B7,$8B,$5F,$34,$E7,$F2
	.dc.b	$FC,$6B,$FB,$BC,$BD,$3F,$27,$8D,$7C,$3A,$FF,$21,$F2,$7D,$7F,$52
	.dc.b	$FA,$37,$E6,$FE,$9F,$CC,$1F,$99,$BF,$3A,$FC,$C7,$85,$FC,$AF,$9C
	.dc.b	$7F,$A9,$B0,$E3,$CE,$C3,$8D,$FF,$73,$E6,$FF,$5B,$67,$DF,$F1,$3F
	.dc.b	$8B,$8F,$28,$F1,$69,$88,$E8,$F8,$F4,$FF,$37,$E9,$E1,$9C,$E7,$DE
	.dc.b	$7F,$BD,$47,$B3,$4F,$1E,$9E,$2F,$D5,$F1,$8F,$3D,$7F,$C4,$79,$73
	.dc.b	$FC,$89,$F6,$AF,$92,$7A,$FF,$C1,$3F,$7F,$7E,$BF,$FD,$93,$0E,$17
	.dc.b	$BA,$D4,$BF,$74,$FC,$3E,$3D,$31,$F6,$3D,$BE,$5E,$E1,$C7,$D5,$66
	.dc.b	$F2,$BF,$05,$C5,$F3,$B0,$E5,$A7,$8F,$CC,$CC,$A7,$38,$72,$D5,$FB
	.dc.b	$E7,$8B,$3E,$D9,$FA,$BE,$DB,$F3,$7F,$BC,$BF,$21,$7E,$99,$FE,$7E
	.dc.b	$E2,$FF,$2B,$7E,$C8,$F9,$ED,$D6,$BC,$7F,$77,$D9,$A9,$F8,$8E,$B1
	.dc.b	$FB,$59,$75,$FC,$E2,$E3,$C7,$3F,$F3,$17,$DA,$FA,$AB,$C5,$7E,$1C
	.dc.b	$BF,$DC,$3D,$5F,$7E,$73,$A7,$83,$47,$EA,$5D,$E3,$1F,$95,$5C,$FD
	.dc.b	$3F,$9D,$DC,$BC,$DF,$9B,$FF,$99,$7E,$C7,$F3,$3D,$4F,$C5,$7F,$D7
	.dc.b	$FF,$A5,$BF,$35,$FF,$58,$7D,$77,$FA,$8F,$F7,$BE,$17,$FA,$0B,$DD
	.dc.b	$6B,$F0,$57,$F0,$0E,$AB,$A7,$DC,$B2,$FF,$43,$67,$E0,$C7,$FD,$9B
	.dc.b	$FA,$76,$2F,$A5,$EF,$2F,$0F,$D6,$7E,$BB,$9D,$8F,$1A,$D7,$E1,$5F
	.dc.b	$E5,$8F,$AE,$D4,$E7,$4F,$54,$FF,$97,$36,$FF,$2B,$E1,$CB,$3E,$F5
	.dc.b	$AE,$0F,$F6,$07,$EB,$3E,$9F,$17,$FC,$CD,$FA,$0B,$F6,$9F,$F0,$5F
	.dc.b	$EA,$2F,$D1,$BD,$F9,$F3,$CF,$F4,$BF,$76,$C3,$87,$D1,$30,$E1,$D1
	.dc.b	$D6,$B5,$7F,$EC,$7E,$1F,$A7,$A7,$F9,$D7,$F0,$0F,$89,$3D,$32,$E7
	.dc.b	$8F,$86,$39,$46,$3E,$F3,$E0,$C7,$F1,$4B,$F4,$ED,$E8,$DB,$E0,$E5
	.dc.b	$DB,$3F,$34,$E7,$7F,$B1,$F8,$80,$84,$1F,$A0,$69,$C1,$BF,$CB,$70
	.dc.b	$E3,$F1,$CF,$C1,$7D,$AF,$F0,$CF,$B5,$7D,$6F,$F9,$5F,$F5,$DF,$A7
	.dc.b	$CF,$F2,$6B,$53,$9F,$C0,$9F,$85,$F2,$F5,$9F,$AE,$FC,$4B,$E9,$CB
	.dc.b	$F2,$4F,$E6,$6F,$2F,$BB,$66,$FC,$C7,$E2,$D7,$FB,$13,$06,$BE,$4B
	.dc.b	$DC,$EB,$0E,$58,$7F,$A7,$BF,$38,$F1,$5E,$E3,$FD,$43,$F5,$1F,$EC
	.dc.b	$FF,$DF,$7F,$57,$F0,$8F,$D1,$1F,$A6,$FF,$8E,$7E,$3B,$FC,$DF,$D9
	.dc.b	$FE,$5E,$BD,$D6,$BE,$D5,$5A,$FB,$CE,$9B,$FF,$4E,$CB,$A6,$5F,$EE
	.dc.b	$FC,$78,$FD,$F5,$AD,$65,$CF,$1F,$F3,$EF,$07,$EA,$6F,$BE,$1E,$3F
	.dc.b	$BE,$FF,$0B,$C6,$34,$F0,$FE,$01,$FA,$EF,$C3,$BC,$39,$BF,$24,$7E
	.dc.b	$38,$F4,$9F,$A5,$FF,$21,$EE,$79,$8F,$CC,$78,$E1,$C2,$F7,$F0,$3F
	.dc.b	$82,$A7,$EE,$3A,$CB,$A6,$7F,$C5,$0F,$41,$FD,$BB,$0F,$0F,$E4,$1E
	.dc.b	$F8,$F2,$4F,$8F,$4F,$17,$1F,$09,$CE,$7E,$0D,$7E,$FC,$77,$8C,$67
	.dc.b	$FD,$1B,$C4,$71,$7F,$24,$D3,$D1,$97,$B6,$9F,$17,$C6,$31,$EB,$7C
	.dc.b	$6F,$CD,$1F,$8D,$3F,$3F,$7E,$3F,$FE,$E6,$87,$E4,$2F,$C8,$38,$46
	.dc.b	$E7,$52,$FC,$F7,$37,$EF,$C3,$AF,$DA,$5C,$38,$79,$DE,$0E,$7F,$81
	.dc.b	$5E,$2B,$F0,$CF,$93,$F2,$FC,$DF,$03,$C7,$F8,$6F,$6D,$3C,$3F,$87
	.dc.b	$39,$5F,$BD,$7F,$46,$FC,$B9,$F0,$4F,$CB,$9F,$62,$FD,$D3,$FC,$8D
	.dc.b	$FB,$0B,$F3,$0E,$58,$F1,$C3,$85,$EF,$EE,$56,$FD,$03,$7E,$0F,$F8
	.dc.b	$B3,$4E,$FA,$7E,$0E,$3F,$A1,$65,$CF,$F3,$BC,$47,$01,$5C,$32,$ED
	.dc.b	$C7,$C2,$7A,$69,$E2,$DB,$FC,$B9,$CF,$99,$E3,$39,$DB,$CB,$F6,$9D
	.dc.b	$7F,$6C,$7D,$97,$FB,$BF,$E6,$9F,$B7,$7F,$95,$39,$DF,$92,$EE,$B5
	.dc.b	$39,$E1,$F3,$7D,$1B,$F1,$FF,$F6,$A7,$7F,$F8,$67,$F8,$01,$F9,$B7
	.dc.b	$CC,$EC,$FA,$3C,$5F,$E4,$EC,$38,$76,$B0,$E1,$7F,$E7,$5B,$DD,$AF
	.dc.b	$B7,$77,$C6,$9C,$C7,$29,$F9,$91,$8C,$BC,$7F,$58,$CD,$F2,$3F,$17
	.dc.b	$0F,$0E,$7E,$7E,$0C,$61,$C2,$FD,$3F,$7F,$F4,$FF,$48,$7E,$56,$FC
	.dc.b	$3F,$FD,$85,$FA,$73,$F2,$37,$E9,$FF,$EC,$DE,$0F,$4E,$1C,$2F,$75
	.dc.b	$AF,$C5,$BA,$62,$FD,$AF,$D9,$E2,$FF,$86,$B0,$E1,$C9,$EC,$8C,$57
	.dc.b	$87,$ED,$9A,$F8,$7C,$4F,$C0,$61,$DE,$73,$9B,$F0,$9C,$E1,$C7,$0F
	.dc.b	$46,$EF,$CB,$FF,$4D,$F9,$3F,$F8,$AB,$F7,$47,$E5,$CF,$C9,$FF,$45
	.dc.b	$E7,$3F,$47,$FE,$9F,$F5,$BC,$38,$FF,$D1,$38,$70,$BD,$D7,$F5,$4F
	.dc.b	$63,$54,$AE,$BF,$95,$7D,$BC,$FC,$19,$F7,$AF,$B7,$18,$C7,$9F,$CA
	.dc.b	$FF,$8A,$63,$1E,$AF,$EB,$F3,$7C,$26,$70,$F1,$E3,$E9,$AD,$61,$FB
	.dc.b	$6E,$C6,$3F,$E4,$8F,$C6,$BF,$CF,$DE,$03,$F1,$97,$C7,$7F,$82,$7F
	.dc.b	$0E,$FB,$97,$BF,$B6,$FD,$05,$6B,$E8,$54,$F0,$BD,$F9,$74,$DF,$D6
	.dc.b	$7F,$AB,$9F,$F4,$67,$ED,$9F,$06,$31,$8F,$4C,$FF,$C3,$1E,$2F,$AE
	.dc.b	$57,$E6,$CF,$C1,$FA,$BE,$7F,$4E,$FF,$09,$FE,$99,$FC,$CF,$AB,$FD
	.dc.b	$85,$BB,$E8,$3E,$8D,$6B,$E1,$FF,$9C,$FE,$1B,$FE,$25,$EC,$CF,$C8
	.dc.b	$32,$F2,$6C,$C3,$FC,$3F,$ED,$61,$E5,$A7,$FB,$3F,$D2,$F1,$61,$C7
	.dc.b	$3E,$BF,$59,$E9,$FC,$35,$FA,$07,$F7,$0F,$23,$FA,$8F,$B3,$F3,$BF
	.dc.b	$7F,$FC,$2D,$6B,$E5,$7F,$3B,$F2,$2A,$FA,$7F,$AB,$EB,$1F,$EF,$3C
	.dc.b	$3F,$30,$FF,$6B,$2F,$D9,$BD,$38,$F3,$CF,$BF,$67,$8E,$BC,$59,$75
	.dc.b	$FC,$37,$E7,$7E,$2D,$FE,$1F,$F5,$3F,$C4,$1F,$B2,$3C,$2F,$F1,$0F
	.dc.b	$E7,$1F,$9A,$AD,$7D,$22,$B5,$F4,$6F,$B9,$D4,$B1,$F9,$77,$FE,$2B
	.dc.b	$FC,$3E,$CE,$4E,$59,$7F,$B9,$3C,$AF,$F4,$DD,$76,$CB,$C5,$F7,$BE
	.dc.b	$6B,$E1,$1E,$0F,$97,$FD,$6F,$33,$F6,$E7,$E9,$AC,$3F,$D8,$9F,$98
	.dc.b	$BF,$34,$7E,$05,$C2,$F7,$5A,$9F,$F7,$07,$A7,$FA,$6E,$C0,$CB,$B7
	.dc.b	$E7,$3F,$86,$7C,$19,$DC,$79,$FE,$20,$D3,$C1,$87,$7A,$F2,$4E,$74
	.dc.b	$FC,$F7,$EE,$6F,$CD,$FF,$77,$F2,$0F,$C1,$5F,$B2,$3E,$D5,$FA,$DF
	.dc.b	$F0,$CC,$38,$7C,$03,$0E,$17,$BC,$B5,$97,$F0,$6D,$7F,$11,$87,$7F
	.dc.b	$33,$F3,$EF,$2C,$38,$F8,$3E,$4B,$F1,$EB,$E2,$F1,$3F,$69,$F9,$4F
	.dc.b	$F9,$1B,$F5,$4F,$35,$7F,$2F,$FC,$C9,$FB,$FF,$F2,$0A,$D6,$1E,$DF
	.dc.b	$8D,$53,$F3,$7D,$7D,$7B,$2F,$D8,$E1,$FE,$52,$F6,$2C,$E7,$1E,$5D
	.dc.b	$FE,$28,$C6,$9F,$CF,$FF,$46,$FE,$11,$F7,$AF,$B0,$FF,$1D,$FE,$AE
	.dc.b	$FD,$A9,$F9,$EB,$F6,$87,$E9,$7F,$CD,$B0,$E1,$F6,$4C,$38,$5E,$FC
	.dc.b	$8D,$64,$FD,$23,$F8,$3E,$1C,$3B,$7F,$68,$F2,$C3,$8F,$C7,$F9,$2F
	.dc.b	$C7,$AF,$8B,$C1,$FE,$3D,$F3,$BF,$AB,$E8,$7E,$D3,$FE,$6C,$FA,$2F
	.dc.b	$F1,$D7,$EB,$1F,$D2,$AB,$53,$9F,$39,$FA,$A6,$7D,$1F,$99,$C1,$FC
	.dc.b	$16,$FF,$6D,$FB,$4F,$F1,$C7,$8F,$DF,$FF,$87,$BF,$08,$F0,$DF,$D3
	.dc.b	$CB,$C1,$F9,$BF,$F3,$1F,$61,$F9,$DB,$CB,$7E,$83,$F2,$3F,$03,$7E
	.dc.b	$5C,$FC,$A3,$FA,$FE,$1C,$2F,$7E,$07,$FA,$3E,$9F,$BC,$EB,$2E,$9C
	.dc.b	$5F,$C6,$F0,$EF,$D9,$FC,$C9,$E5,$A7,$7F,$0F,$C9,$7E,$38,$F1,$7E
	.dc.b	$3F,$FD,$1F,$F2,$CF,$F4,$9F,$CC,$3F,$7A,$7F,$00,$FF,$95,$BF,$63
	.dc.b	$FF,$50,$F6,$7E,$F7,$66,$1C,$2F,$7F,$82,$AF,$F9,$65,$EE,$BF,$D3
	.dc.b	$FF,$BC,$7F,$46,$7B,$E1,$FC,$CD,$E5,$8F,$F9,$93,$EA,$3C,$9D,$9F
	.dc.b	$EF,$3C,$FA,$F3,$F9,$FF,$0B,$7E,$FC,$FD,$05,$F8,$93,$F7,$57,$E9
	.dc.b	$4F,$1B,$FC,$17,$B7,$FA,$EF,$0E,$17,$BA,$FC,$1F,$0F,$9F,$27,$E5
	.dc.b	$3E,$77,$BF,$0F,$9D,$F0,$E9,$E0,$D3,$F9,$C4,$F6,$9C,$E3,$E3,$F7
	.dc.b	$FE,$77,$E0,$2F,$D1,$DF,$BC,$BF,$80,$1F,$84,$7F,$8D,$BF,$73,$F3
	.dc.b	$DE,$EB,$53,$FE,$3A,$FC,$83,$38,$7C,$BA,$7E,$BD,$63,$AE,$3D,$3D
	.dc.b	$9F,$0C,$63,$07,$EA,$1F,$81,$FC,$05,$FB,$F7,$EF,$BF,$64,$7E,$CF
	.dc.b	$FE,$BB,$C3,$E4,$D6,$BD,$F5,$6B,$F1,$87,$EE,$1A,$FD,$97,$EF,$5F
	.dc.b	$4E,$7F,$BC,$F1,$BE,$BE,$A7,$87,$4F,$06,$9F,$66,$FE,$BF,$C5,$3F
	.dc.b	$2C,$7E,$8A,$FC,$11,$F6,$EF,$DF,$3F,$5D,$E1,$7E,$3F,$0E,$17,$BA
	.dc.b	$D7,$AB,$D3,$1F,$C9,$53,$EB,$6C,$76,$CB,$AF,$D3,$FE,$96,$B9,$C7
	.dc.b	$83,$C0,$FD,$47,$C4,$FF,$10,$EF,$FB,$9C,$3F,$E6,$5F,$DD,$7F,$1F
	.dc.b	$39,$EA,$F9,$BF,$84,$7F,$52,$E1,$FD,$24,$FF,$18,$E3,$FD,$12,$BF
	.dc.b	$D3,$3C,$70,$7D,$A7,$83,$E0,$59,$F8,$3D,$2F,$DD,$7C,$23,$F6,$DF
	.dc.b	$F6,$6F,$C6,$7E,$3F,$F7,$6F,$E4,$BB,$7E,$CE,$B5,$ED,$7F,$3F,$E1
	.dc.b	$57,$86,$B5,$FB,$5F,$ED,$EF,$A7,$0F,$ED,$DE,$38,$FE,$6F,$F1,$8F
	.dc.b	$0E,$7E,$0C,$BF,$DC,$57,$FB,$0F,$82,$FE,$47,$FA,$93,$F6,$B7,$89
	.dc.b	$FE,$7D,$F7,$B5,$F0,$6A,$D4,$FF,$A0,$7E,$07,$38,$FF,$07,$C7,$FD
	.dc.b	$E4,$DB,$E1,$8F,$4F,$C4,$BF,$8E,$57,$82,$3B,$FE,$3F,$FC,$7F,$DD
	.dc.b	$7C,$0B,$F7,$B7,$E5,$BF,$EF,$4F,$CB,$DF,$B9,$FF,$08,$BD,$D6,$BA
	.dc.b	$94,$FD,$1F,$58,$F3,$C7,$FD,$FB,$F4,$9D,$E3,$16,$DB,$C5,$FB,$A3
	.dc.b	$F0,$27,$DA,$3F,$4C,$7C,$F3,$F7,$5F,$15,$EE,$B5,$2F,$E0,$0F,$37
	.dc.b	$ED,$DF,$CB,$78,$97,$4E,$18,$F4,$E8,$E7,$9F,$82,$3F,$C3,$DE,$1F
	.dc.b	$F3,$7B,$7C,$3F,$EC,$3E,$15,$93,$F7,$37,$EC,$DF,$16,$B5,$39,$FD
	.dc.b	$0B,$1A,$7F,$3D,$C5,$FB,$9F,$96,$F7,$E5,$FD,$37,$8F,$C9,$7D,$5F
	.dc.b	$8A,$7E,$33,$FC,$F3,$E8,$7F,$C7,$DF,$C0,$1F,$4E,$73,$F1,$E9,$FF
	.dc.b	$21,$7E,$83,$9B,$E1,$F0,$BC,$DE,$5F,$47,$F8,$1F,$2F,$F8,$FF,$EC
	.dc.b	$5F,$92,$FD,$1D,$E7,$FD,$F9,$C8,$86,$BF,$83,$FA,$3F,$9D,$3A,$EF
	.dc.b	$E9,$ED,$B1,$FC,$EF,$D4,$FE,$3F,$E0,$BF,$8D,$7E,$C9,$FE,$02,$FC
	.dc.b	$8B,$FA,$BF,$81,$F3,$E7,$3F,$45,$FE,$DF,$FA,$DF,$C8,$46,$3C,$9E
	.dc.b	$5F,$10,$C7,$FC,$17,$E9,$FE,$6F,$E0,$FF,$C4,$3F,$A8,$FF,$AE,$7E
	.dc.b	$79,$FB,$17,$E6,$67,$3F,$8D,$7F,$64,$FD,$1B,$1F,$0E,$C3,$FD,$81
	.dc.b	$F1,$BC,$E3,$14,$F8,$CF,$D4,$7C,$37,$F0,$DF,$DD,$FF,$94,$78,$6F
	.dc.b	$D7,$3F,$4D,$39,$F9,$0F,$A9,$F8,$06,$3E,$21,$87,$F2,$FF,$A2,$E7
	.dc.b	$18,$AF,$BF,$E8,$7E,$BF,$E1,$7F,$C5,$3F,$B7,$3F,$7F,$FF,$15,$7E
	.dc.b	$A3,$9F,$89,$4E,$7C,$77,$EA,$39,$8E,$1E,$3F,$EF,$F1,$E7,$8B,$E3
	.dc.b	$BF,$81,$7A,$FF,$BC,$FC,$1B,$FC,$6B,$EB,$FE,$FE,$31,$E3,$7F,$B6
	.dc.b	$FC,$7F,$D2,$46,$FF,$22,$E3,$F0,$DF,$C2,$70,$7E,$4B,$D3,$DA,$FE
	.dc.b	$AF,$F7,$CF,$D0,$9F,$B3,$F5,$9C,$C7,$FC,$01,$CF,$AC,$70,$FC,$7F
	.dc.b	$97,$D0,$7F,$A7,$7F,$20,$F3,$FB,$FF,$A4,$F8,$FF,$EC,$5F,$99,$79
	.dc.b	$E3,$1F,$40,$FC,$CF,$1F,$01,$FC,$73,$D2,$E2,$FF,$1E,$F5,$7E,$0F
	.dc.b	$B9,$FA,$2E,$E7,$E0,$BF,$06,$31,$F9,$75,$F7,$BF,$39,$CF,$C1,$FC
	.dc.b	$F7,$C1,$7C,$AF,$6F,$F3,$BF,$87,$3B,$7E,$54,$63,$F6,$7B,$EE,$FE
	.dc.b	$67,$9B,$83,$F9,$BE,$0F,$A7,$ED,$7E,$6B,$F1,$57,$DA,$FD,$38,$C7
	.dc.b	$D2,$5F,$6B,$F3,$1C,$BC,$1F,$CD,$7C,$CF,$DC,$F5,$FF,$2D,$F4,$AF
	.dc.b	$9A,$FB,$91,$8E,$CB,$EB,$7E,$52,$B8,$3F,$96,$FC,$97,$F0,$BE,$87
	.dc.b	$E2,$3E,$8B,$F0,$8F,$E2,$BF,$3D,$BE,$B7,$E4,$BD,$2D,$BF,$94,$FC
	.dc.b	$C3,$F1,$BE,$8F,$E3,$7B,$3F,$18,$FE,$3F,$F6,$8B,$E9,$7F,$51,$FC
	.dc.b	$07,$6F,$F5,$3F,$C0,$B8,$F9,$FF,$8A,$FC,$43,$8B,$F8,$CF,$CA,$EF
	.dc.b	$A0,$F9,$4D,$AF,$1B,$C3,$E7,$FE,$1F,$E0,$BC,$2F,$E2,$7C,$FB,$E2
	.dc.b	$3F,$15,$DA,$EF,$F4,$F9,$DF,$85,$F5,$7F,$43,$C9,$FD,$E3,$CD,$FE
	.dc.b	$3D,$D9,$EC,$AF,$F2,$4F,$9B,$C2,$7F,$47,$F0,$3F,$03,$CE,$FC,$2E
	.dc.b	$3C,$9F,$E4,$7D,$39,$F3,$3F,$80,$CE,$57,$9B,$F2,$BC,$BF,$C2,$4F
	.dc.b	$89,$FE,$3F,$FB,$37,$CC,$7F,$D8,$3F,$8A,$6F,$CB,$7B,$1F,$4B,$CB
	.dc.b	$FE,$31,$FA,$8F,$D2,$7F,$0B,$8F,$B1,$FF,$14,$73,$EF,$CA,$FE,$ED
	.dc.b	$C1,$F6,$3C,$9F,$E1,$37,$F0,$9F,$E2,$BF,$63,$9F,$23,$DF,$39,$7E
	.dc.b	$95,$7E,$03,$A3,$F8,$74,$FC,$A7,$FC,$17,$FC,$03,$9F,$1B,$E9,$55
	.dc.b	$F8,$CF,$C4,$D7,$F7,$2F,$09,$EF,$F3,$F1,$BF,$CC,$7F,$B1,$5E,$DD
	.dc.b	$C7,$F9,$F3,$1F,$07,$87,$FD,$AB,$F1,$4B,$E1,$FF,$87,$B7,$F1,$4F
	.dc.b	$D0,$E2,$5C,$D7,$FD,$7A,$49,$FF,$3A,$57,$A7,$C1,$FE,$D2,$DF,$05
	.dc.b	$EC,$7B,$E7,$F1,$58,$5F,$57,$D7,$E1,$FF,$62,$C3,$E1,$3F,$E7,$DD
	.dc.b	$3E,$33,$FE,$5E,$AB,$E9,$7C,$2F,$03,$FD,$FB,$FB,$9F,$CB,$7F,$AF
	.dc.b	$E1,$F2,$D7,$E3,$7B,$FC,$13,$E4,$F7,$FF,$98,$FF,$2F,$FA,$4F,$F5
	.dc.b	$9B,$FA,$4F,$FA,$E1,$BF,$47,$F4,$39,$3F,$CA,$F1,$F5,$DF,$EA,$6D
	.dc.b	$FC,$57,$5E,$B7,$D6,$E6,$FF,$22,$CF,$DD,$7F,$C9,$BF,$A2,$7E,$05
	.dc.b	$7A,$DB,$D7,$87,$57,$FA,$72,$F8,$2F,$DF,$78,$69,$D7,$E6,$3B,$7E
	.dc.b	$D6,$FC,$D7,$A2,$FD,$77,$B7,$73,$F2,$BB,$D7,$E1,$ED,$FF,$5E,$EB
	.dc.b	$E5,$5F,$83,$F9,$7B,$7F,$D5,$D3,$E9,$7F,$D4,$8F,$AF,$AB,$FD,$13
	.dc.b	$CF,$F8,$6E,$DF,$C2,$F2,$C9,$FF,$44,$4E,$63,$20,$74,$BF,$00,$F8
	.dc.b	$8F,$FA,$5F,$0F,$91,$4F,$2B,$E4,$2F,$D2,$FE,$56,$0F,$D8,$EF,$97
	.dc.b	$8E,$9C,$FD,$7E,$B9,$37,$FA,$1F,$0F,$9B,$E3,$C9,$CB,$F0,$DE,$7C
	.dc.b	$8E,$0F,$37,$27,$E8,$BF,$22,$F7,$33,$EC,$F5,$C1,$F0,$67,$33,$F0
	.dc.b	$FE,$13,$2F,$13,$27,$F8,$B6,$FA,$A9,$CA,$FD,$0F,$8B,$93,$8F,$EA
	.dc.b	$78,$9D,$FE,$7F,$67,$AF,$E3,$75,$F1,$FE,$85,$5F,$4F,$A3,$5F,$DE
	.dc.b	$FE,$C1,$FE,$49,$CF,$C7,$8F,$9A,$3F,$D2,$11,$89,$F6,$FE,$BC,$FC
	.dc.b	$CF,$94,$23,$FC,$BD,$3B,$8F,$8F,$F9,$C3,$EB,$B6,$BF,$3F,$E3,$FB
	.dc.b	$DB,$F1,$9B,$FB,$67,$4E,$4F,$F5,$9E,$5C,$F2,$F9,$9F,$AB,$78,$3F
	.dc.b	$6E,$FF,$6A,$73,$FD,$19,$FA,$31,$FE,$EB,$8C,$47,$D8,$E6,$FF,$55
	.dc.b	$DF,$38,$FE,$3B,$0F,$5B,$D7,$CC,$FC,$DE,$FE,$89,$EE,$F8,$0F,$E4
	.dc.b	$78,$F1,$C5,$D2,$FD,$07,$E1,$AF,$CC,$FE,$2F,$C0,$70,$DB,$C7,$7F
	.dc.b	$A7,$F9,$4C,$3F,$85,$3F,$55,$D7,$FE,$BF,$FE,$00,$FA,$CF,$FB,$07
	.dc.b	$F0,$AB,$EA,$70,$AF,$C2,$1C,$3E,$4F,$1E,$07,$D6,$F0,$1F,$A8,$7D
	.dc.b	$FF,$07,$EC,$7E,$FB,$79,$37,$8F,$E7,$27,$FA,$BE,$FE,$68,$78,$3F
	.dc.b	$5F,$F1,$FF,$C4,$3F,$67,$CA,$BC,$5C,$AB,$F8,$14,$FF,$58,$E7,$F9
	.dc.b	$DE,$13,$C1,$FD,$D4,$61,$FA,$0E,$BE,$77,$95,$FF,$50,$AF,$E9,$B9
	.dc.b	$F8,$FC,$27,$85,$FC,$16,$30,$BF,$40,$FC,$E4,$FF,$40,$BB,$F0,$7D
	.dc.b	$17,$87,$FE,$EC,$FC,$B3,$F4,$4F,$FB,$BB,$F3,$4B,$E2,$EE,$1F,$17
	.dc.b	$C7,$C6,$78,$AF,$C8,$7F,$55,$18,$FD,$26,$FC,$2E,$FE,$4F,$C1,$E4
	.dc.b	$7E,$0F,$F7,$EF,$D8,$3F,$85,$FC,$7E,$FC,$7E,$BF,$29,$F8,$3E,$4F
	.dc.b	$E0,$7F,$38,$FD,$C3,$F8,$3F,$B9,$BF,$8C,$3E,$3B,$6B,$EF,$BE,$4F
	.dc.b	$8D,$FB,$5F,$25,$FA,$BF,$F0,$18,$C7,$85,$7C,$B7,$D6,$70,$7F,$A9
	.dc.b	$F7,$7F,$07,$F1,$FF,$01,$F9,$2D,$F6,$5E,$63,$F4,$6C,$7E,$07,$F6
	.dc.b	$2F,$F7,$9F,$63,$F6,$BC,$BF,$C0,$78,$8F,$C7,$3F,$89,$46,$3F,$42
	.dc.b	$BE,$7F,$EC,$76,$BF,$37,$F8,$2F,$33,$EF,$F9,$AF,$63,$F8,$D4,$63
	.dc.b	$D9,$BE,$87,$F7,$7F,$47,$F5,$B3,$F8,$6E,$EF,$C5,$F9,$DF,$7F,$D6
	.dc.b	$FE,$44,$FE,$07,$AA,$FE,$40,$F1,$B6,$FE,$27,$D4,$FE,$21,$E7,$7D
	.dc.b	$DF,$CA,$7F,$95,$3F,$7F,$E2,$6F,$A5,$FE,$15,$AD,$BF,$4F,$93,$F9
	.dc.b	$87,$9B,$F7,$FA,$6F,$BC,$F3,$9D,$58,$F7,$BF,$43,$7F,$C2,$7E,$A3
	.dc.b	$F5,$9E,$77,$E0,$7F,$35,$FE,$78,$FE,$0F,$AE,$F9,$CE,$CF,$F2,$EE
	.dc.b	$D8,$7F,$CC,$EF,$F4,$3F,$2F,$F0,$9E,$63,$F7,$6C,$79,$9F,$C1,$A1
	.dc.b	$FA,$EE,$3C,$FF,$F4,$56,$17,$F0,$EF,$DF,$79,$5F,$84,$FD,$1F,$FA
	.dc.b	$73,$F8,$6C,$EF,$9F,$F8,$6F,$7B,$6F,$E2,$3E,$8B,$F8,$37,$99,$F8
	.dc.b	$4E,$B7,$E2,$5F,$D5,$E3,$19,$DF,$37,$F9,$A7,$9B,$C1,$FE,$6D,$E7
	.dc.b	$7F,$0F,$F2,$FF,$0B,$D4,$C3,$FB,$04,$63,$F6,$0B,$E7,$7F,$05,$F5
	.dc.b	$B8,$3F,$C1,$FC,$EF,$E3,$1E,$6F,$F0,$3E,$BF,$D6,$78,$FF,$D6,$63
	.dc.b	$1F,$2B,$FD,$AD,$FE,$11,$F0,$DC,$3C,$D7,$C1,$FE,$AB,$B3,$FB,$87
	.dc.b	$9B,$FC,$1B,$E0,$7F,$94,$FC,$8E,$3C,$EF,$E1,$1D,$EF,$C8,$F1,$4E
	.dc.b	$3F,$E5,$58,$71,$C7,$F9,$67,$97,$FC,$27,$F7,$47,$EA,$7F,$DF,$63
	.dc.b	$1B,$DF,$45,$F0,$5F,$CA,$30,$FE,$99,$5F,$C4,$7D,$FF,$E7,$FE,$7F
	.dc.b	$F4,$DF,$C1,$5F,$B5,$BE,$17,$FC,$15,$39,$87,$EF,$38,$F4,$BF,$1D
	.dc.b	$6F,$B9,$E7,$18,$E2,$FE,$97,$E6,$FF,$11,$FC,$39,$F9,$63,$F0,$DF
	.dc.b	$F0,$B4,$E6,$1F,$92,$63,$D6,$FE,$AF,$7F,$A9,$F0,$BF,$C2,$17,$CF
	.dc.b	$F0,$8F,$EA,$FD,$BF,$D4,$FF,$80,$1F,$84,$3F,$4E,$FF,$11,$4E,$63
	.dc.b	$F3,$78,$7F,$86,$63,$87,$CB,$F2,$F5,$BF,$8D,$E1,$FE,$CE,$F8,$7F
	.dc.b	$F0,$57,$9F,$FD,$5F,$EA,$9E,$17,$F8,$0F,$F0,$8F,$F1,$74,$E6,$3F
	.dc.b	$C0,$D9,$FF,$8C,$7C,$8F,$F1,$EB,$FE,$27,$9D,$7C,$9F,$F6,$E7,$87
	.dc.b	$77,$F7,$0F,$4B,$FA,$DF,$D6,$BF,$6D,$7F,$00,$FF,$5D,$F1,$7F,$C9
	.dc.b	$15,$A9,$CF,$C7,$7F,$8F,$E3,$1D,$DC,$FD,$5F,$F3,$55,$7F,$9C,$F5
	.dc.b	$FF,$1C,$7A,$4F,$4B,$FD,$47,$FA,$CB,$F6,$3F,$F0,$C5,$6A,$73,$0F
	.dc.b	$CE,$BF,$C9,$9F,$51,$EC,$E8,$FF,$17,$63,$77,$FF,$2E,$3D,$7E,$09
	.dc.b	$9F,$F5,$59,$FF,$2A,$4E,$63,$F9,$B4,$7F,$98,$FF,$3C,$7C,$63,$F3
	.dc.b	$67,$D4,$BF,$10,$76,$FF,$97,$6F,$75,$A9,$FF,$3B,$70,$FF,$95,$7F
	.dc.b	$39,$F2,$65,$FE,$67,$BB,$EB,$FF,$B6,$9E,$39,$7E,$66,$7A,$67,$E1
	.dc.b	$8F,$E5,$38,$7F,$97,$FD,$9F,$EF,$FF,$5E,$FC,$39,$F6,$CF,$D3,$3F
	.dc.b	$CE,$15,$A9,$CF,$E9,$5F,$E7,$3F,$B1,$F0,$E8,$AA,$74,$E4,$FA,$DC
	.dc.b	$DF,$E5,$98,$E5,$F9,$77,$F8,$D3,$DD,$FF,$25,$7E,$BC,$FC,$37,$FC
	.dc.b	$9D,$F9,$2F,$F9,$FE,$73,$1F,$E5,$1F,$C9,$71,$9F,$FA,$4B,$8B,$FC
	.dc.b	$F1,$C7,$E2,$BD,$FD,$07,$5F,$84,$67,$D7,$EA,$BF,$C9,$3F,$07,$FE
	.dc.b	$4F,$FA,$8F,$E0,$BF,$AD,$FF,$9D,$73,$FF,$42,$CE,$7E,$15,$3F,$E0
	.dc.b	$E9,$CF,$E5,$DF,$42,$9F,$07,$4F,$EF,$58,$7F,$10,$F4,$7B,$C6,$31
	.dc.b	$FF,$24,$73,$7F,$AA,$FD,$EF,$F4,$77,$CE,$3D,$9F,$F1,$84,$FD,$2D
	.dc.b	$3F,$D2,$F5,$A9,$CC,$FF,$A9,$63,$FD,$2D,$C9,$F0,$E3,$A7,$E6,$3F
	.dc.b	$84,$9E,$1F,$01,$F4,$AB,$FD,$7D,$1C,$FE,$31,$D7,$FE,$6C,$F7,$FF
	.dc.b	$D3,$5C,$2F,$D4,$1D,$6F,$C4,$1E,$17,$FA,$D6,$F7,$5A,$9F,$EB,$FC
	.dc.b	$1F,$EA,$E9,$FF,$52,$69,$E5,$3F,$D9,$CB,$F2,$5F,$CC,$BC,$B1,$FE
	.dc.b	$E7,$5C,$7E,$F1,$9F,$8E,$7F,$4F,$87,$FB,$17,$DD,$FF,$56,$78,$FF
	.dc.b	$BF,$FA,$1F,$F3,$37,$97,$9F,$BC,$FC,$D7,$AF,$FE,$86,$F7,$5F,$99
	.dc.b	$67,$2F,$21,$FE,$D5,$CB,$FC,$96,$E3,$D1,$FE,$E1,$87,$2C,$BF,$7B
	.dc.b	$3B,$8C,$7C,$CF,$FA,$53,$DF,$FF,$7B,$7E,$A6,$FC,$8B,$F4,$9F,$71
	.dc.b	$FE,$BA,$C3,$FD,$DD,$5A,$9C,$FE,$A3,$FE,$DF,$8F,$F7,$F4,$F8,$5F
	.dc.b	$F2,$93,$7F,$06,$61,$D7,$D2,$FF,$1D,$78,$5F,$EF,$5B,$E9,$1F,$E3
	.dc.b	$FB,$FF,$75,$7A,$6F,$5F,$FD,$15,$FA,$F7,$F9,$5F,$17,$E2,$DA,$9C
	.dc.b	$FD,$47,$6F,$03,$8F,$D7,$FF,$0E,$C7,$96,$5F,$15,$E9,$D0,$CF,$FC
	.dc.b	$C5,$F1,$1F,$EB,$69,$CC,$7F,$99,$6B,$FD,$73,$AF,$F2,$A5,$F0,$9F
	.dc.b	$D9,$BF,$51,$7E,$6B,$FF,$00,$6B,$53,$9F,$D1,$31,$F0,$DF,$C8,$25
	.dc.b	$F9,$6D,$86,$3D,$3E,$65,$F7,$5E,$09,$CE,$3F,$DD,$3C,$DF,$F6,$5F
	.dc.b	$BD,$FC,$47,$F7,$EF,$F0,$B7,$E2,$AF,$C5,$DF,$8B,$3C,$E7,$77,$6D
	.dc.b	$9E,$37,$FB,$9B,$E1,$19,$72,$C5,$FA,$AF,$1C,$38,$5F,$FB,$B3,$E5
	.dc.b	$7F,$9C,$EB,$FC,$87,$2F,$C8,$3E,$3E,$A6,$DF,$C2,$27,$DB,$7D,$F3
	.dc.b	$F3,$FB,$DF,$EE,$7F,$DD,$9F,$85,$7F,$6F,$FC,$7F,$FC,$43,$E9,$BF
	.dc.b	$37,$FD,$D7,$1F,$AB,$3F,$8F,$87,$0F,$97,$61,$C2,$DF,$53,$E9,$BD
	.dc.b	$E5,$FD,$ED,$FD,$4E,$1E,$0F,$A7,$FE,$AA,$F3,$C3,$C3,$1E,$5A,$F2
	.dc.b	$7C,$27,$F0,$CF,$C8,$FF,$DF,$DF,$BF,$BF,$1D,$7E,$6D,$F8,$2F,$C9
	.dc.b	$DF,$B3,$FD,$EF,$CD,$7C,$5B,$30,$E1,$7B,$AF,$C1,$EB,$A9,$E9,$FA
	.dc.b	$47,$F6,$87,$77,$FC,$01,$E3,$E4,$B2,$EF,$F1,$3E,$4D,$3C,$73,$9C
	.dc.b	$BF,$C7,$1D,$DF,$81,$F8,$B7,$D2,$7F,$99,$BF,$2F,$6A,$FC,$59,$F8
	.dc.b	$C3,$F3,$9F,$4E,$1C,$3E,$E9,$87,$0B,$DF,$E7,$5A,$CB,$FC,$E3,$C5
	.dc.b	$FC,$5B,$0E,$FF,$03,$FE,$02,$79,$61,$C7,$9F,$F4,$3D,$FF,$76,$FF
	.dc.b	$A6,$61,$E5,$DB,$C9,$D7,$AE,$0F,$CD,$7E,$62,$FC,$CD,$FA,$EB,$F5
	.dc.b	$A7,$EB,$5F,$92,$FC,$35,$FA,$DF,$D2,$C3,$85,$EE,$B5,$FB,$F6,$71
	.dc.b	$FF,$63,$6D,$FA,$73,$15,$FE,$1A,$BE,$39,$76,$F3,$3F,$C5,$9C,$1E
	.dc.b	$99,$CE,$5F,$A2,$FC,$1B,$D1,$8F,$EF,$BC,$F7,$E8,$AF,$CC,$DF,$9A
	.dc.b	$BE,$ED,$F4,$7F,$E1,$BF,$D3,$BF,$13,$87,$0B,$DD,$6B,$E6,$1F,$21
	.dc.b	$AC,$BA,$67,$FD,$1B,$0E,$FF,$B9,$7F,$8C,$9E,$5A,$77,$D3,$FD,$29
	.dc.b	$49,$CE,$5E,$2F,$3F,$D3,$2F,$9B,$FD,$A7,$E9,$BF,$D8,$FE,$73,$FB
	.dc.b	$6F,$E0,$6F,$D9,$BF,$3F,$5A,$FA,$0F,$EC,$7F,$45,$FE,$90,$FC,$FF
	.dc.b	$3E,$D3,$F4,$EF,$F2,$2D,$F4,$D3,$FA,$EB,$C7,$3F,$EF,$3D,$1E,$28
	.dc.b	$C6,$3F,$EF,$DF,$0F,$F8,$E7,$EA,$4F,$92,$7E,$40,$FA,$77,$EC,$6F
	.dc.b	$C5,$DF,$86,$3E,$BF,$F6,$B8,$F1,$C3,$85,$EE,$9F,$91,$EB,$2F,$F6
	.dc.b	$67,$07,$E1,$7C,$5F,$CC,$63,$C6,$3F,$BA,$BC,$F4,$F0,$E9,$FF,$00
	.dc.b	$2B,$77,$E6,$9C,$FA,$1F,$3B,$F5,$0F,$DA,$FE,$B2,$FC,$84,$BF,$08
	.dc.b	$7E,$C2,$FC,$93,$F8,$6D,$F0,$BD,$D6,$BF,$24,$CF,$AF,$FD,$13,$83
	.dc.b	$FC,$13,$7D,$9B,$1D,$AF,$8F,$87,$E3,$8C,$67,$E1,$F0,$FF,$07,$EA
	.dc.b	$BF,$1E,$7E,$3B,$FD,$8B,$C5,$FA,$1E,$EF,$A7,$97,$F7,$97,$BA,$D7
	.dc.b	$22,$BF,$75,$5A,$C1,$F8,$7F,$FA,$23,$0E,$1B,$FF,$88,$1E,$5F,$09
	.dc.b	$FE,$DA,$C7,$FC,$87,$3F,$CB,$A7,$39,$BE,$1B,$FC,$95,$29,$CD,$78
	.dc.b	$7F,$17,$FE,$6B,$FB,$AB,$F4,$EF,$F7,$87,$E0,$1F,$EA,$1F,$92,$7E
	.dc.b	$DE,$FD,$83,$75,$AF,$C3,$7F,$A7,$CB,$F2,$8C,$FE,$E1,$FE,$A0,$CB
	.dc.b	$A7,$CD,$5C,$38,$64,$FC,$9B,$F5,$B5,$AD,$3C,$1F,$95,$62,$F7,$9B
	.dc.b	$E0,$BF,$C9,$7C,$5F,$E0,$87,$FD,$0F,$F5,$CF,$BC,$FF,$7E,$FC,$BB
	.dc.b	$EF,$1F,$85,$F0,$C3,$96,$9F,$E2,$1B,$DD,$7F,$9A,$3F,$12,$FE,$41
	.dc.b	$97,$4B,$FE,$44,$71,$7F,$D2,$BF,$AB,$62,$31,$87,$3E,$BF,$EB,$DB
	.dc.b	$FF,$87,$30,$E7,$39,$D3,$E0,$5F,$0D,$7D,$18,$7F,$65,$CF,$F9,$CE
	.dc.b	$1F,$CC,$7F,$20,$7E,$B4,$FC,$CB,$F0,$BF,$6E,$FC,$CF,$F7,$9F,$B2
	.dc.b	$7B,$6F,$75,$A9,$FE,$01,$E9,$67,$07,$ED,$1E,$9C,$BB,$47,$5D,$BD
	.dc.b	$F1,$8C,$39,$7A,$BE,$B9,$FF,$15,$65,$DB,$C6,$6B,$E3,$EC,$D5,$7F
	.dc.b	$8F,$2B,$53,$E7,$B7,$F0,$07,$D3,$FD,$2D,$FB,$FB,$F6,$BF,$F1,$8F
	.dc.b	$EC,$2F,$D0,$FF,$DE,$3F,$51,$F6,$7F,$53,$7E,$4C,$FC,$7F,$FD,$3F
	.dc.b	$F9,$EB,$F3,$F7,$EA,$0F,$DA,$B8,$5E,$FE,$D9,$7B,$AF,$95,$D1,$FD
	.dc.b	$7F,$2E,$B9,$7C,$5E,$5F,$F2,$44,$78,$3F,$24,$C4,$7F,$96,$23,$78
	.dc.b	$F4,$D1,$F2,$7F,$0A,$95,$AC,$FC,$D9,$66,$31,$7E,$8F,$8F,$F3,$E7
	.dc.b	$EC,$8F,$5F,$E5,$6B,$FF,$56,$E3,$FE,$68,$FC,$8B,$F6,$9F,$40,$FC
	.dc.b	$71,$FA,$B3,$EE,$DF,$9E,$3F,$64,$E5,$F3,$3F,$E6,$3C,$78,$E1,$C2
	.dc.b	$F7,$F5,$9A,$CF,$B7,$BB,$FD,$B1,$FF,$02,$6B,$E2,$58,$7F,$05,$FA
	.dc.b	$3F,$C6,$65,$D6,$FF,$BF,$65,$D7,$E6,$FF,$49,$39,$DF,$D1,$BF,$9F
	.dc.b	$F1,$8C,$E7,$E0,$9C,$ED,$EA,$FD,$93,$C9,$2F,$C0,$DF,$86,$7F,$A2
	.dc.b	$BE,$C7,$F1,$BF,$C0,$0F,$D9,$7F,$D2,$BC,$7E,$7B,$DE,$1F,$E2,$4F
	.dc.b	$25,$7B,$FC,$47,$AD,$6A,$BF,$47,$C1,$FE,$89,$8E,$EF,$BA,$3F,$D3
	.dc.b	$31,$D3,$2F,$0F,$97,$EA,$9F,$36,$9E,$5C,$BB,$FC,$8E,$70,$E1,$1E
	.dc.b	$2D,$7F,$0B,$97,$7B,$E1,$A7,$AF,$F3,$9F,$2C,$62,$BF,$CB,$7B,$7B
	.dc.b	$63,$1E,$77,$C6,$F2,$9F,$9F,$BF,$04,$FC,$57,$C5,$7F,$A9,$3F,$29
	.dc.b	$FE,$4D,$8F,$DB,$BD,$D6,$BE,$99,$4F,$DE,$BA,$63,$CF,$8F,$E2,$9F
	.dc.b	$EB,$AC,$3B,$FE,$0B,$F1,$63,$F7,$53,$E1,$CB,$FA,$36,$3F,$ED,$F7
	.dc.b	$C1,$19,$CF,$FD,$6F,$F9,$57,$C9,$C7,$C7,$5C,$33,$7C,$2F,$8B,$5F
	.dc.b	$34,$7F,$88,$25,$F9,$3F,$F9,$FF,$EF,$5F,$27,$FE,$21,$FD,$69,$F8
	.dc.b	$DB,$F4,$4F,$F7,$16,$1C,$38,$7F,$DA,$7F,$81,$B1,$E3,$DC,$C7,$8E
	.dc.b	$0F,$DC,$BF,$69,$A7,$7D,$3F,$96,$3E,$1F,$C9,$AE,$7D,$32,$E7,$8F
	.dc.b	$8B,$F1,$AF,$E8,$D7,$C7,$3E,$DB,$3F,$45,$CE,$3C,$73,$F5,$4F,$A7
	.dc.b	$E4,$7C,$33,$FE,$EA,$FD,$8F,$FD,$BF,$D3,$0F,$F7,$0F,$E6,$7A,$97
	.dc.b	$EB,$7F,$D6,$5F,$C2,$F2,$7C,$17,$F9,$97,$D3,$7E,$A7,$17,$88,$FC
	.dc.b	$F1,$FB,$4B,$F7,$7E,$9F,$11,$FE,$22,$F1,$98,$70,$BD,$FC,$8E,$B2
	.dc.b	$FD,$A6,$7F,$D3,$B9,$7F,$DE,$98,$FF,$8A,$BF,$05,$F8,$F3,$E2,$CB
	.dc.b	$9E,$BE,$1F,$03,$E1,$CE,$74,$ED,$F4,$18,$C7,$C5,$1E,$1D,$DF,$A5
	.dc.b	$78,$72,$E7,$BF,$AB,$CD,$F8,$9F,$C0,$3F,$E3,$0F,$C6,$3F,$C6,$3F
	.dc.b	$AC,$3F,$6B,$FF,$61,$7E,$99,$AF,$96,$56,$A7,$FD,$8D,$E0,$67,$D9
	.dc.b	$F4,$E3,$CF,$8B,$DC,$03,$DF,$06,$5D,$FF,$24,$F6,$CE,$61,$FB,$BF
	.dc.b	$1B,$F1,$67,$E1,$FC,$65,$F8,$8E,$32,$E7,$AF,$FB,$67,$1F,$E9,$1E
	.dc.b	$97,$1F,$17,$EA,$FE,$B5,$FE,$6D,$FC,$73,$FD,$C5,$F9,$5E,$B5,$F3
	.dc.b	$56,$BE,$EF,$FD,$F7,$F8,$E7,$FB,$83,$DB,$D5,$7E,$B3,$B3,$FB,$93
	.dc.b	$FE,$08,$CF,$B4,$33,$FF,$07,$46,$1F,$05,$FF,$66,$AE,$78,$F7,$CD
	.dc.b	$3F,$D7,$E7,$38,$73,$CB,$C7,$F8,$D7,$F6,$CF,$1F,$F8,$ED,$7A,$2B
	.dc.b	$58,$7F,$BF,$7F,$48,$E3,$9F,$83,$07,$3E,$32,$E7,$FA,$A7,$B3,$E1
	.dc.b	$1F,$1C,$F9,$EF,$F7,$77,$27,$93,$F0,$E7,$EF,$5B,$AD,$7E,$39,$F6
	.dc.b	$B8,$D5,$FB,$9E,$3F,$2F,$1B,$3E,$BF,$8D,$FC,$48,$FC,$24,$78,$34
	.dc.b	$C4,$6F,$2E,$BA,$7C,$4F,$67,$FC,$99,$CF,$F8,$3C,$38,$6B,$E1,$F4
	.dc.b	$33,$8F,$4C,$78,$ED,$E4,$F0,$7D,$1A,$7B,$23,$CF,$FB,$FF,$DC,$FA
	.dc.b	$8F,$F2,$77,$E6,$9F,$C8,$7F,$0E,$7A,$EF,$D2,$9F,$9C,$6E,$D7,$BA
	.dc.b	$D7,$EE,$B9,$8F,$5C,$FF,$96,$7F,$79,$AF,$0E,$4F,$F6,$9D,$FD,$FE
	.dc.b	$7F,$8F,$3E,$29,$CF,$D7,$7B,$38,$BF,$CC,$53,$C7,$5F,$2E,$AF,$D3
	.dc.b	$3C,$18,$74,$87,$E8,$DF,$98,$FD,$DD,$FA,$F7,$F6,$3F,$DA,$3F,$D3
	.dc.b	$9F,$0F,$FD,$13,$F0,$ED,$4E,$7E,$83,$F9,$9F,$D3,$63,$E4,$3E,$AE
	.dc.b	$3C,$BE,$52,$DE,$F1,$71,$79,$EB,$BD,$B3,$EF,$C7,$9C,$38,$7C,$53
	.dc.b	$F4,$1E,$D5,$E4,$87,$EE,$78,$DB,$FA,$2F,$A9,$75,$FE,$B7,$F4,$1F
	.dc.b	$E4,$5F,$D1,$3F,$83,$7D,$9B,$F0,$FF,$CD,$FC,$2F,$F3,$8F,$41,$FB
	.dc.b	$43,$F7,$8E,$1D,$FF,$C5,$70,$E1,$E4,$FF,$3E,$C3,$FD,$E5,$6F,$D5
	.dc.b	$37,$7F,$E6,$DC,$BF,$4E,$FF,$A2,$33,$F0,$C7,$F9,$3E,$30,$75,$7C
	.dc.b	$77,$FE,$FD,$F9,$DF,$F5,$35,$70,$C7,$C1,$A3,$F4,$7C,$E3,$C6,$73
	.dc.b	$9B,$F3,$5F,$1E,$BE,$8C,$FD,$D2,$FC,$DF,$CF,$C4,$FC,$3D,$FA,$13
	.dc.b	$F2,$07,$ED,$2F,$61,$FB,$87,$FB,$83,$BD,$C7,$FC,$C7,$2E,$58,$F1
	.dc.b	$C3,$85,$FF,$87,$6B,$59,$3F,$80,$3F,$D5,$D1,$8B,$BC,$5F,$D0,$F1
	.dc.b	$E7,$A7,$8B,$47,$E2,$9F,$02,$BF,$A9,$65,$E8,$AD,$78,$F9,$C3,$8E
	.dc.b	$BE,$2F,$A8,$F7,$DE,$E3,$CB,$C5,$F6,$3F,$1A,$7E,$42,$FB,$47,$E4
	.dc.b	$EF,$D3,$DF,$AD,$BF,$29,$7E,$7F,$C7,$0E,$17,$BA,$D7,$EA,$B9,$C7
	.dc.b	$DB,$C4,$CF,$BF,$CF,$61,$F0,$CF,$83,$4E,$FF,$41,$EF,$9C,$E5,$D7
	.dc.b	$E8,$31,$A7,$82,$FC,$BB,$7F,$B1,$7B,$3E,$AE,$9F,$DB,$71,$E3,$5A
	.dc.b	$FD,$8B,$C9,$97,$9F,$0F,$C0,$E1,$C3,$0F,$36,$3D,$77,$F8,$1D,$9F
	.dc.b	$EA,$FF,$7B,$FB,$87,$8C,$FC,$55,$F9,$9B,$F5,$57,$E7,$2F,$AD,$7E
	.dc.b	$0B,$9D,$BF,$92,$ED,$FE,$01,$CB,$96,$3C,$70,$E1,$6F,$DD,$FC,$17
	.dc.b	$BC,$FB,$BF,$EF,$4D,$3C,$70,$FB,$DF,$F7,$D4,$63,$1E,$9A,$3F,$16
	.dc.b	$FF,$08,$6B,$E2,$C5,$AF,$8B,$6F,$4D,$FF,$91,$27,$C2,$6B,$3F,$7C
	.dc.b	$EB,$57,$EB,$1A,$C3,$96,$FE,$AF,$DF,$F8,$FC,$E3,$F6,$A7,$DF,$7F
	.dc.b	$B4,$BD,$BF,$0F,$E7,$8F,$D0,$F7,$5A,$FC,$33,$F8,$FE,$6A,$7E,$25
	.dc.b	$AF,$D2,$3D,$FC,$5F,$E2,$BD,$3B,$69,$88,$46,$31,$E7,$C3,$F0,$B9
	.dc.b	$FF,$C5,$53,$9D,$BE,$1F,$81,$9C,$3A,$6D,$E3,$9F,$3F,$0F,$9B,$3F
	.dc.b	$56,$FE,$5F,$CE,$FE,$3F,$E5,$5F,$E6,$FF,$BC,$FF,$38,$FC,$4E,$1C
	.dc.b	$0F,$DA,$30,$FF,$9D,$7E,$93,$C1,$AB,$F2,$DF,$F3,$A7,$EB,$9F,$49
	.dc.b	$FF,$3B,$7E,$93,$67,$FD,$E0,$7F,$90,$FB,$FE,$E4,$FF,$91,$63,$FC
	.dc.b	$8D,$3D,$3E,$B9,$F8,$8B,$F6,$2F,$97,$E8,$BE,$C5,$FA,$E3,$E6,$FF
	.dc.b	$03,$87,$0B,$DD,$6B,$F6,$3C,$E3,$CF,$3F,$CA,$1F,$E5,$DB,$EF,$F2
	.dc.b	$DF,$4D,$E5,$E0,$B4,$F0,$7C,$DF,$97,$0F,$24,$E7,$27,$D3,$67,$4F
	.dc.b	$0E,$CF,$D6,$FC,$31,$8F,$D4,$1C,$6F,$A2,$FF,$33,$7E,$44,$EB,$FE
	.dc.b	$81,$FA,$1F,$F1,$FF,$B4,$7E,$30,$FC,$D7,$8E,$1C,$2F,$7F,$00,$BD
	.dc.b	$FD,$EE,$BA,$3F,$CF,$59,$F5,$F8,$6F,$A1,$1F,$AE,$C7,$97,$E7,$58
	.dc.b	$8E,$02,$B8,$6B,$E3,$EA,$E1,$3D,$34,$F0,$E7,$FE,$23,$C7,$C9,$93
	.dc.b	$0F,$56,$DF,$CE,$77,$F8,$18,$F1,$AD,$6D,$E9,$AF,$A1,$5A,$AF,$D9
	.dc.b	$FC,$6B,$EE,$DF,$BD,$FF,$A7,$3D,$1F,$A3,$F6,$7E,$8F,$2F,$FB,$25
	.dc.b	$6B,$EF,$D5,$F5,$BE,$53,$58,$FC,$7C,$BE,$1E,$7D,$F1,$E3,$18,$E0
	.dc.b	$FF,$4D,$47,$4D,$3C,$1D,$9F,$0A,$BC,$BA,$77,$CD,$F9,$DE,$70,$E1
	.dc.b	$AF,$8B,$07,$AD,$E2,$C3,$84,$3F,$19,$C6,$7D,$BE,$46,$E1,$C7,$E8
	.dc.b	$7E,$AF,$C7,$3F,$35,$7D,$23,$F0,$5F,$F7,$F7,$EB,$8F,$C3,$9F,$A1
	.dc.b	$70,$BD,$D6,$A5,$F9,$A6,$70,$FC,$17,$3F,$D1,$CB,$B7,$B5,$8E,$8F
	.dc.b	$F5,$E4,$78,$6B,$C1,$F8,$B3,$5F,$07,$4F,$3D,$3C,$7F,$29,$F1,$F0
	.dc.b	$E1,$1E,$1F,$DD,$F5,$87,$1D,$BF,$D2,$5E,$C7,$FB,$27,$F3,$50,$FC
	.dc.b	$01,$F9,$33,$F5,$A7,$EF,$4F,$C9,$5F,$63,$FA,$EC,$79,$63,$C7,$0E
	.dc.b	$17,$BF,$AD,$D7,$0F,$FB,$7E,$BB,$E7,$F0,$B0,$E3,$F5,$16,$B5,$5C
	.dc.b	$F4,$F0,$FE,$35,$FD,$62,$FB,$E1,$E6,$9F,$2F,$1E,$70,$E3,$39,$CF
	.dc.b	$D1,$FB,$67,$8A,$7D,$F9,$75,$9F,$67,$EE,$F9,$FC,$F7,$F9,$07,$CD
	.dc.b	$7F,$88,$BF,$42,$FF,$77,$FE,$B0,$FD,$97,$5F,$0F,$FE,$00,$FA,$45
	.dc.b	$6B,$F4,$6C,$E1,$C7,$DD,$62,$FD,$F3,$38,$70,$9C,$FE,$43,$E5,$FD
	.dc.b	$49,$F9,$2B,$F2,$CF,$E6,$7D,$AF,$B2,$7E,$90,$FD,$57,$85,$EF,$E2
	.dc.b	$BE,$AD,$6B,$F2,$BC,$E3,$F3,$7A,$59,$F6,$FC,$EB,$0F,$82,$7B,$DB
	.dc.b	$F2,$AF,$99,$7C,$32,$ED,$AF,$8F,$C1,$CE,$3C,$74,$F0,$EC,$FD,$57
	.dc.b	$C3,$18,$FD,$4B,$E9,$5F,$F8,$0F,$E7,$5F,$B2,$7F,$B2,$FF,$3B,$77
	.dc.b	$BF,$5D,$DF,$D1,$AF,$7F,$88,$AF,$78,$6A,$7D,$D8,$FF,$83,$FF,$52
	.dc.b	$B8,$71,$AE,$DF,$63,$EC,$AF,$14,$78,$7C,$DC,$56,$F2,$ED,$A7,$F8
	.dc.b	$57,$D9,$79,$AE,$69,$97,$2B,$DF,$EE,$5F,$D4,$BE,$83,$54,$FC,$75
	.dc.b	$E6,$FF,$09,$FD,$01,$FA,$A3,$F4,$87,$D9,$3F,$4D,$E5,$8F,$7C,$78
	.dc.b	$FB,$2C,$78,$E1,$C3,$F7,$4D,$DF,$7C,$FF,$83,$BF,$E0,$3E,$16,$BE
	.dc.b	$4C,$BC,$7F,$43,$FC,$52,$31,$7E,$5D,$BC,$9F,$65,$FE,$48,$9C,$EB
	.dc.b	$E1,$9F,$47,$E0,$19,$CF,$CD,$87,$1D,$7D,$7C,$7E,$4D,$3D,$3B,$FA
	.dc.b	$38,$5E,$FE,$70,$E1,$0F,$D0,$F1,$AF,$7C,$BE,$2F,$87,$EE,$FA,$0F
	.dc.b	$F2,$47,$C9,$FF,$9C,$FE,$DF,$EA,$FE,$1B,$FC,$6F,$9B,$DF,$C3,$F5
	.dc.b	$EF,$7F,$75,$FE,$58,$EC,$FF,$64,$7E,$6D,$CF,$EE,$BF,$CA,$B7,$FD
	.dc.b	$63,$A3,$EC,$E1,$FE,$55,$F9,$8F,$8D,$3C,$EB,$FC,$B7,$8F,$3F,$C3
	.dc.b	$3E,$14,$AF,$BE,$BF,$88,$F8,$FC,$C6,$34,$FF,$6C,$7E,$9B,$E1,$E0
	.dc.b	$F5,$47,$FB,$6F,$5F,$B7,$FB,$BB,$F3,$FF,$F6,$BF,$E8,$DF,$D3,$3F
	.dc.b	$28,$FF,$70,$7E,$41,$A9,$CF,$CB,$27,$FD,$C9,$F9,$0E,$7F,$13,$F4
	.dc.b	$F0,$B2,$E9,$F8,$46,$1E,$D2,$CF,$B7,$1F,$B2,$73,$1E,$1E,$2C,$69
	.dc.b	$E0,$CF,$FD,$E5,$F2,$5D,$EB,$CD,$D5,$CF,$AB,$E9,$EA,$FC,$CB,$FC
	.dc.b	$C7,$F7,$AF,$CB,$7F,$33,$18,$FB,$EB,$3F,$E6,$B7,$E6,$FE,$A5,$FE
	.dc.b	$2B,$B7,$4F,$E8,$23,$FC,$CB,$8F,$67,$F6,$DF,$9D,$7C,$07,$FC,$65
	.dc.b	$D1,$FE,$A2,$AF,$DA,$FE,$FA,$F3,$3F,$BB,$FE,$8B,$F3,$FE,$3F,$ED
	.dc.b	$78,$7F,$E4,$CB,$DD,$6B,$97,$FD,$83,$E0,$7F,$83,$75,$F1,$ED,$FC
	.dc.b	$17,$3E,$9E,$9F,$F8,$03,$1E,$3D,$D6,$B5,$97,$3D,$3F,$DF,$F1,$DC
	.dc.b	$CC,$B5,$F2,$7C,$37,$A6,$73,$1E,$7D,$1F,$49,$C6,$5A,$7B,$70,$E5
	.dc.b	$B7,$A3,$F6,$BD,$0F,$E0,$17,$EB,$4F,$C7,$BF,$D6,$FF,$19,$FC,$DB
	.dc.b	$ED,$B7,$5A,$9C,$FC,$0F,$E7,$F9,$7E,$0E,$AC,$BA,$70,$7B,$E3,$18
	.dc.b	$76,$7F,$C8,$BC,$9F,$C9,$AB,$C5,$A7,$87,$CA,$FE,$C5,$F0,$BE,$FB
	.dc.b	$E3,$39,$D9,$FB,$86,$70,$E3,$B7,$93,$97,$CF,$7B,$DF,$FD,$A3,$BF
	.dc.b	$9B,$C8,$FE,$95,$F1,$DF,$C3,$FD,$0F,$F6,$DF,$E5,$EF,$A6,$F9,$55
	.dc.b	$AF,$5F,$FD,$AD,$E9,$29,$F8,$D6,$B8,$FF,$CD,$3C,$7E,$C3,$FA,$36
	.dc.b	$1D,$9C,$07,$A6,$5F,$EE,$AF,$B3,$F7,$46,$31,$FF,$50,$D7,$5D,$3C
	.dc.b	$D1,$E1,$F4,$71,$A7,$93,$4E,$9B,$7B,$74,$FE,$F5,$39,$FC,$D7,$7A
	.dc.b	$FB,$BF,$49,$FF,$53,$FC,$EF,$F5,$6F,$E0,$07,$ED,$9F,$E9,$FF,$DF
	.dc.b	$9E,$AF,$F3,$3F,$C5,$9F,$93,$F0,$E3,$FE,$C5,$87,$0F,$C0,$98,$70
	.dc.b	$DB,$FD,$F1,$F0,$FB,$B5,$ED,$FF,$42,$E9,$E1,$D1,$FA,$7F,$F9,$9A
	.dc.b	$31,$97,$4D,$7F,$A2,$7C,$5F,$F7,$1E,$35,$6B,$5F,$07,$E2,$D9,$8C
	.dc.b	$5F,$A7,$F0,$7F,$16,$BE,$7B,$F3,$70,$FF,$97,$31,$E7,$B7,$F7,$DF
	.dc.b	$D9,$5F,$02,$FD,$97,$FD,$85,$E7,$BF,$5F,$7D,$6F,$A3,$1E,$38,$B1
	.dc.b	$E3,$87,$0B,$DF,$17,$C3,$9E,$BF,$A6,$7B,$EF,$C1,$9F,$77,$E2,$98
	.dc.b	$C3,$97,$E9,$7E,$D8,$C6,$1F,$BE,$C5,$F3,$79,$9F,$16,$1E,$9A,$D6
	.dc.b	$8F,$CD,$3C,$53,$9D,$BD,$92,$F7,$7E,$C7,$EC,$EF,$DB,$7F,$D4,$9F
	.dc.b	$C0,$0F,$CB,$3F,$BC,$FE,$5A,$FC,$13,$75,$A9,$CF,$F0,$0E,$3F,$04
	.dc.b	$F7,$70,$32,$E9,$F9,$CE,$1E,$D2,$CF,$B7,$CD,$BF,$2E,$F6,$56,$A7
	.dc.b	$3F,$0B,$8D,$3C,$16,$D9,$FB,$47,$86,$F9,$C3,$E9,$B1,$BF,$AB,$F2
	.dc.b	$0B,$87,$1F,$D3,$FF,$31,$F9,$73,$81,$F9,$23,$F2,$67,$EE,$AF,$C0
	.dc.b	$1F,$A4,$7E,$0F,$CB,$FF,$20,$E3,$C7,$0E,$17,$BA,$77,$F6,$CB,$AF
	.dc.b	$47,$E0,$06,$1E,$1F,$D0,$7E,$4B,$CF,$4F,$0E,$8F,$CB,$FF,$01,$39
	.dc.b	$CF,$C9,$FA,$6E,$32,$F4,$E1,$C3,$46,$5E,$1D,$7C,$B9,$FA,$97,$93
	.dc.b	$FE,$A1,$D7,$FC,$C1,$F7,$4F,$D7,$1F,$96,$FF,$A5,$3E,$E5,$FB,$1B
	.dc.b	$F0,$0F,$F4,$BE,$3C,$71,$FF,$64,$E1,$C3,$EA,$D8,$70,$BF,$F2,$05
	.dc.b	$EF,$2E,$DF,$BB,$7F,$BA,$3F,$09,$FF,$35,$C7,$87,$4C,$67,$E4,$CF
	.dc.b	$A6,$5E,$2F,$63,$F8,$36,$1C,$27,$3A,$BF,$57,$CE,$1C,$71,$F5,$4F
	.dc.b	$A7,$F1,$CF,$83,$87,$19,$CE,$DF,$63,$6F,$67,$E8,$AF,$C9,$DF,$96
	.dc.b	$3F,$00,$FF,$76,$7C,$DB,$F3,$3F,$F3,$FE,$1C,$2F,$75,$A9,$7E,$F5
	.dc.b	$E0,$DF,$FD,$1D,$97,$5C,$BD,$F0,$8F,$07,$EF,$58,$75,$87,$83,$E9
	.dc.b	$FD,$B5,$BC,$BB,$FE,$29,$FE,$0B,$8C,$DF,$08,$FF,$07,$66,$FD,$9B
	.dc.b	$C5,$5C,$73,$F6,$56,$BE,$03,$38,$71,$DF,$DD,$F4,$58,$CF,$BC,$7C
	.dc.b	$1F,$CF,$F9,$FE,$28,$F9,$1F,$C3,$7C,$E7,$C5,$F4,$1F,$C0,$0F,$D4
	.dc.b	$B7,$5A,$E2,$FE,$21,$2F,$DF,$33,$F8,$97,$C4,$CB,$A7,$55,$C3,$86
	.dc.b	$4F,$07,$E1,$D6,$B4,$F0,$7C,$7E,$2F,$79,$FF,$11,$EC,$FC,$C5,$EF
	.dc.b	$5F,$0F,$D0,$7E,$93,$C3,$FF,$5E,$FD,$D3,$EC,$9F,$70,$FD,$01,$FB
	.dc.b	$A3,$F5,$17,$E1,$CF,$37,$97,$83,$FE,$CB,$CB,$96,$3C,$70,$E1,$7F
	.dc.b	$AE,$DF,$F6,$17,$BC,$FF,$CE,$CF,$FB,$38,$FE,$C9,$A7,$92,$3C,$71
	.dc.b	$93,$B8,$C7,$BE,$9E,$6D,$3C,$B3,$A9,$CE,$5E,$9F,$8E,$EB,$3E,$1D
	.dc.b	$3C,$D7,$EC,$FC,$5F,$86,$5F,$0B,$2E,$9A,$FE,$97,$5F,$4E,$3F,$EE
	.dc.b	$0F,$D1,$DF,$75,$09,$F1,$BE,$6F,$C2,$D3,$99,$D7,$C9,$66,$3F,$D0
	.dc.b	$97,$BC,$3D,$CB,$CF,$C3,$F0,$78,$F5,$F7,$7A,$AF,$1E,$5F,$C7,$AB
	.dc.b	$E5,$79,$9E,$E5,$FA,$13,$F7,$77,$C2,$BF,$4B,$7E,$38,$F9,$2F,$F7
	.dc.b	$EF,$B3,$F3,$B0,$E1,$7B,$AD,$7E,$1B,$FC,$17,$5F,$87,$5A,$CF,$EB
	.dc.b	$46,$33,$7C,$5F,$9F,$81,$52,$FA,$EB,$E2,$D7,$FC,$03,$3E,$1C,$3C
	.dc.b	$F3,$9E,$6F,$65,$F0,$D3,$C7,$EE,$6F,$6F,$36,$4D,$9F,$B7,$FE,$23
	.dc.b	$F2,$5F,$E3,$9D,$7F,$89,$F8,$2F,$F1,$FF,$B8,$FD,$AF,$5E,$E7,$E8
	.dc.b	$FC,$5F,$EC,$9F,$43,$FC,$73,$F3,$DC,$FE,$0B,$EB,$E6,$F5,$E9,$D1
	.dc.b	$F8,$3C,$FE,$E5,$1E,$0C,$39,$FD,$07,$D5,$8C,$6B,$FE,$50,$F2,$31
	.dc.b	$A7,$86,$3C,$BF,$8E,$79,$38,$3D,$1B,$7F,$2E,$E5,$FE,$BB,$F3,$9D
	.dc.b	$FF,$7C,$7D,$63,$F1,$AF,$F5,$87,$1B,$F2,$57,$C4,$BF,$29,$E3,$C9
	.dc.b	$F5,$31,$E3,$87,$0B,$DF,$E2,$5A,$A5,$76,$FA,$5F,$C9,$69,$E1,$D3
	.dc.b	$C1,$5F,$82,$8C,$63,$CF,$EE,$3F,$CC,$11,$8F,$84,$6D,$E5,$E1,$CE
	.dc.b	$1E,$3C,$7D,$55,$AD,$7D,$3F,$3F,$E3,$D7,$CF,$39,$E0,$7E,$9F,$9C
	.dc.b	$38,$E7,$F0,$67,$FC,$41,$E4,$B5,$F1,$4E,$7F,$1E,$F7,$7C,$3F,$FA
	.dc.b	$D3,$F0,$47,$9E,$FD,$53,$FC,$2D,$FA,$13,$F1,$AD,$FE,$0C,$BD,$D6
	.dc.b	$BF,$49,$CF,$E2,$3F,$1B,$9B,$F4,$39,$F5,$7E,$31,$8C,$39,$67,$DF
	.dc.b	$E8,$DF,$9E,$7C,$4A,$5F,$96,$B5,$FA,$2E,$63,$19,$F9,$BF,$1C,$F2
	.dc.b	$EB,$E4,$C9,$BF,$F9,$37,$8F,$E1,$67,$FB,$4F,$DC,$BD,$DF,$B7,$BF
	.dc.b	$38,$FE,$13,$F1,$2F,$D6,$9C,$7F,$83,$BD,$D6,$A7,$FD,$AB,$D1,$FA
	.dc.b	$19,$F9,$1F,$A2,$7C,$1C,$91,$D9,$F8,$26,$38,$19,$F8,$7D,$96,$8F
	.dc.b	$C1,$BE,$16,$1E,$09,$CE,$D8,$C3,$8E,$BE,$3D,$FC,$D1,$8A,$E7,$B7
	.dc.b	$CC,$BF,$83,$A3,$F1,$AF,$C2,$7E,$0B,$FE,$A4,$FC,$F5,$FA,$C7,$F6
	.dc.b	$1D,$BF,$45,$5A,$FB,$C5,$6B,$E3,$9F,$95,$EB,$95,$FA,$DF,$C1,$C3
	.dc.b	$AE,$9F,$01,$E5,$87,$6F,$D6,$3E,$F4,$63,$47,$E4,$F8,$C7,$FD,$B3
	.dc.b	$A7,$82,$7F,$BA,$7E,$A9,$DF,$7F,$44,$7F,$77,$FC,$73,$15,$BD,$DF
	.dc.b	$BE,$63,$5F,$47,$E2,$7F,$E6,$6F,$C1,$3F,$8C,$7C,$33,$F6,$BF,$F7
	.dc.b	$F7,$DB,$BF,$41,$FF,$01,$7E,$EE,$FD,$77,$87,$37,$F3,$CC,$38,$7E
	.dc.b	$31,$FE,$8D,$BD,$FE,$BF,$AA,$FE,$69,$92,$BF,$AD,$67,$E0,$87,$E3
	.dc.b	$38,$AE,$91,$E2,$C3,$FC,$25,$FA,$6F,$FB,$A6,$F8,$61,$DF,$87,$FB
	.dc.b	$17,$77,$FA,$EF,$1E,$33,$9D,$BF,$B9,$CE,$67,$C7,$39,$FD,$F7,$CD
	.dc.b	$F7,$2F,$B9,$7E,$52,$F5,$7F,$BF,$FE,$91,$FE,$88,$77,$3A,$73,$5F
	.dc.b	$C1,$72,$EC,$F3,$BF,$66,$5F,$E9,$0B,$7E,$6B,$8A,$D6,$1D,$BF,$29
	.dc.b	$F7,$5F,$F6,$89,$CF,$D2,$33,$7E,$47,$AA,$E9,$9F,$9E,$B5,$6F,$57
	.dc.b	$CF,$39,$C3,$FD,$43,$FA,$BF,$D5,$FA,$87,$E5,$6F,$DC,$9F,$B3,$BF
	.dc.b	$41,$7E,$5A,$FD,$27,$FD,$E5,$CB,$F2,$F1,$E3,$87,$0B,$DD,$3F,$56
	.dc.b	$ED,$93,$F1,$BF,$87,$87,$0E,$2F,$F0,$4F,$67,$F6,$98,$E7,$AF,$87
	.dc.b	$57,$D5,$FB,$E7,$3A,$79,$30,$C6,$5E,$9C,$38,$7E,$6F,$FE,$2B,$DF
	.dc.b	$CD,$39,$BF,$5C,$FF,$B0,$E7,$3B,$FA,$73,$4F,$F8,$A2,$73,$87,$4F
	.dc.b	$D9,$3F,$D9,$BF,$B2,$7F,$7F,$FE,$01,$7E,$B0,$FD,$6D,$FA,$77,$E2
	.dc.b	$BE,$A5,$FB,$CB,$F2,$2E,$3C,$5F,$E2,$4C,$78,$E1,$C3,$AF,$FD,$57
	.dc.b	$6F,$DF,$B7,$6B,$DB,$FE,$30,$D3,$C5,$A3,$F6,$FF,$F2,$34,$63,$2E
	.dc.b	$9A,$3F,$4E,$FF,$19,$F0,$AB,$5B,$F9,$FD,$CC,$C6,$37,$F1,$FE,$EF
	.dc.b	$E2,$DB,$CF,$7E,$DE,$07,$E8,$3E,$6F,$D8,$9F,$BD,$BF,$7E,$7E,$63
	.dc.b	$FC,$9B,$F5,$DF,$E1,$5A,$6E,$B5,$39,$F6,$7F,$4B,$C1,$E5,$E6,$FF
	.dc.b	$24,$E5,$D2,$FD,$B1,$87,$B3,$F0,$CC,$65,$E0,$F8,$76,$6F,$57,$DB
	.dc.b	$5A,$CF,$C9,$F9,$16,$21,$97,$AB,$1E,$58,$7F,$21,$FC,$93,$D1,$18
	.dc.b	$D5,$F8,$FF,$D6,$FD,$DF,$F8,$57,$A1,$FE,$81,$1F,$B1,$67,$D6,$C4
	.dc.b	$7F,$6F,$E2,$F7,$6A,$B7,$CF,$7F,$A1,$23,$85,$7C,$07,$57,$FE,$8A
	.dc.b	$BE,$37,$C2,$7F,$D5,$1E,$67,$8F,$0F,$F4,$6F,$5D,$FC,$13,$95,$EE
	.dc.b	$FF,$CB,$1E,$6F,$C1,$F1,$DF,$6E,$FD,$D3,$F4,$7F,$2F,$FE,$8B,$FC
	.dc.b	$25,$FB,$97,$E5,$2F,$75,$AF,$C1,$54,$FC,$8B,$58,$F3,$CF,$EB,$1F
	.dc.b	$C3,$9F,$ED,$EE,$30,$E7,$E7,$FF,$82,$30,$FF,$6B,$4F,$2C,$7C,$7F
	.dc.b	$90,$FF,$9B,$F8,$FE,$4E,$1C,$35,$F3,$7E,$A5,$BC,$B9,$6A,$FD,$5D
	.dc.b	$87,$EE,$33,$FF,$4A,$FE,$A8,$FA,$B7,$E1,$FF,$4B,$F5,$87,$E6,$DF
	.dc.b	$E6,$BF,$C1,$F8,$FA,$BF,$E7,$FC,$38,$5E,$EB,$FC,$E1,$F2,$5A,$CB
	.dc.b	$AE,$5D,$1F,$F5,$3E,$1C,$63,$F9,$34,$60,$DD,$F3,$FC,$17,$F6,$F5
	.dc.b	$B8,$ED,$F9,$DE,$2B,$7D,$CD,$7C,$FF,$72,$FD,$F7,$C5,$4D,$BC,$F7
	.dc.b	$BD,$BF,$D0,$DF,$90,$EB,$B3,$FC,$9F,$F0,$BF,$EC,$7F,$C7,$7F,$C4
	.dc.b	$7F,$19,$D7,$97,$E4,$67,$F7,$7E,$C7,$F9,$F7,$E2,$2B,$DF,$EA,$DC
	.dc.b	$E7,$DB,$5F,$EF,$FB,$E3,$97,$F9,$73,$F2,$CF,$F6,$E7,$2F,$FA,$A6
	.dc.b	$3C,$39,$73,$FA,$2F,$EB,$F8,$73,$D7,$C9,$0F,$92,$F8,$53,$E3,$9C
	.dc.b	$EC,$FD,$0F,$86,$1C,$E1,$FB,$76,$38,$3D,$BF,$A7,$DC,$FF,$B6,$FE
	.dc.b	$39,$F2,$3E,$71,$F8,$67,$F8,$67,$F5,$B7,$DB,$FE,$F3,$ED,$7F,$B3
	.dc.b	$72,$FC,$4B,$DF,$CF,$AF,$7E,$C7,$F9,$B7,$E4,$3A,$65,$FE,$23,$E4
	.dc.b	$FC,$00,$C3,$BF,$D1,$FC,$48,$FF,$74,$C7,$4C,$39,$FE,$0F,$F0,$23
	.dc.b	$1A,$FF,$0E,$FC,$1F,$18,$F9,$E3,$CD,$FB,$07,$83,$5F,$24,$7A,$7F
	.dc.b	$22,$FC,$2F,$E6,$5F,$ED,$8F,$80,$7E,$F4,$FD,$CB,$FD,$47,$F2,$6F
	.dc.b	$17,$7F,$8B,$7F,$20,$AD,$7E,$09,$FE,$58,$F2,$F3,$8F,$A7,$91,$9F
	.dc.b	$FB,$93,$E2,$30,$F7,$C3,$8E,$7F,$6A,$31,$A7,$FA,$CB,$E5,$B1,$A7
	.dc.b	$78,$F1,$FE,$35,$DB,$0E,$3F,$9F,$FE,$F3,$CC,$FF,$79,$7C,$D7,$E4
	.dc.b	$FF,$40,$7E,$83,$FD,$4F,$DB,$FF,$3E,$FE,$4D,$BE,$11,$FD,$E7,$C7
	.dc.b	$5F,$0F,$8F,$FE,$11,$E3,$FF,$02,$F1,$5A,$76,$BF,$F0,$96,$1F,$9D
	.dc.b	$8C,$61,$D2,$30,$76,$D3,$C9,$A7,$8F,$F2,$CF,$EE,$77,$C2,$3C,$D8
	.dc.b	$62,$F8,$3E,$48,$F5,$65,$E6,$CB,$C7,$5A,$CB,$D1,$AF,$FA,$6B,$E8
	.dc.b	$BD,$D3,$ED,$9C,$DF,$C4,$BE,$33,$9C,$3A,$FE,$85,$C3,$F3,$2F,$7F
	.dc.b	$EF,$9F,$ED,$5F,$C1,$1F,$86,$FF,$B3,$31,$E3,$5A,$FC,$1D,$F8,$67
	.dc.b	$FB,$26,$BF,$D6,$1C,$3F,$0E,$BE,$4F,$E9,$9F,$17,$5F,$E7,$53,$91
	.dc.b	$9E,$63,$E1,$0A,$ED,$87,$58,$F3,$57,$0C,$7C,$39,$BF,$06,$E3,$87
	.dc.b	$08,$F4,$FD,$1F,$0C,$BD,$9A,$3E,$9F,$E3,$FE,$A4,$FC,$E3,$FC,$73
	.dc.b	$F9,$AB,$F6,$FF,$F4,$D4,$FF,$00,$DB,$EF,$3B,$FA,$95,$6F,$CE,$4F
	.dc.b	$F7,$DE,$4F,$9E,$EF,$8B,$DD,$18,$C9,$FB,$EF,$C1,$9F,$1C,$78,$B3
	.dc.b	$FF,$55,$78,$DF,$13,$0F,$06,$5D,$FF,$5B,$C5,$EF,$3F,$0F,$9F,$CB
	.dc.b	$5F,$2E,$5E,$AE,$1F,$EC,$7F,$17,$F9,$7F,$DF,$3F,$DE,$3F,$22,$FC
	.dc.b	$0F,$FB,$DF,$CB,$BF,$49,$7E,$75,$C3,$8F,$F5,$38,$70,$BD,$D7,$F0
	.dc.b	$7E,$5F,$EA,$95,$AF,$C5,$19,$7B,$67,$32,$C3,$FE,$00,$FC,$B6,$74
	.dc.b	$F2,$61,$C3,$2F,$1F,$C7,$FF,$96,$A7,$39,$79,$FE,$61,$B7,$F8,$6B
	.dc.b	$3F,$31,$D7,$1E,$53,$AD,$5F,$A5,$6B,$5F,$54,$E7,$F2,$DC,$5F,$0C
	.dc.b	$7E,$17,$A7,$8C,$FC,$11,$F1,$7D,$2E,$FF,$A6,$BE,$6D,$F8,$BB,$EE
	.dc.b	$BF,$D0,$7F,$6E,$FC,$75,$F8,$AF,$0B,$DD,$6A,$5F,$A2,$E7,$0F,$89
	.dc.b	$C9,$FC,$63,$2E,$DF,$C0,$38,$84,$78,$67,$C1,$DD,$FE,$89,$F9,$DF
	.dc.b	$91,$5C,$73,$EF,$3F,$E8,$DF,$CD,$73,$3C,$A7,$3B,$7A,$FF,$42,$F1
	.dc.b	$6D,$EA,$97,$EB,$9F,$3B,$F4,$0F,$E5,$FB,$7A,$3F,$06,$7E,$AB,$AF
	.dc.b	$CE,$75,$ED,$2B,$53,$CB,$F3,$2C,$FD,$A7,$AF,$1E,$7B,$FE,$9B,$C2
	.dc.b	$FE,$59,$1D,$B9,$9A,$76,$AF,$4C,$E6,$3C,$3F,$69,$8D,$3C,$17,$C3
	.dc.b	$3F,$4F,$CE,$F9,$35,$F1,$EE,$F9,$DF,$C4,$FE,$92,$FD,$73,$FC,$F3
	.dc.b	$F2,$38,$7F,$CF,$7F,$23,$AF,$C9,$3F,$CF,$1E,$77,$E7,$2F,$87,$DE
	.dc.b	$7D,$56,$A2,$7C,$D7,$D0,$85,$7F,$01,$E5,$E9,$CB,$FC,$8F,$DC,$ED
	.dc.b	$F7,$3C,$EB,$F9,$37,$2F,$FA,$F3,$F7,$5F,$EA,$1F,$A2,$3F,$16,$7F
	.dc.b	$00,$7F,$14,$FA,$3F,$F6,$BF,$E3,$DF,$E7,$5E,$16,$1C,$2F,$7F,$00
	.dc.b	$BD,$FE,$D7,$FE,$68,$AD,$63,$FB,$97,$87,$E3,$57,$83,$FB,$5E,$9E
	.dc.b	$1A,$F0,$7D,$8B,$F2,$BF,$35,$6B,$4E,$F9,$FF,$BA,$BE,$C7,$32,$C7
	.dc.b	$96,$BE,$3C,$5B,$BF,$5A,$F2,$5F,$3D,$3B,$CF,$F9,$17,$83,$DB,$3F
	.dc.b	$DF,$BF,$6A,$FF,$0E,$7E,$89,$F3,$BF,$66,$7E,$A2,$FC,$69,$FA,$97
	.dc.b	$4C,$18,$70,$E8,$61,$C2,$DF,$B0,$EE,$BA,$DF,$F3,$CE,$7F,$F0,$14
	.dc.b	$7F,$8A,$FE,$57,$E3,$4F,$4F,$A5,$6B,$E3,$D5,$FA,$E7,$C2,$C3,$C1
	.dc.b	$A7,$9B,$E6,$B1,$87,$0D,$BC,$78,$7A,$7C,$CF,$14,$E6,$FC,$DE,$EF
	.dc.b	$AF,$F6,$3F,$BF,$FB,$67,$FB,$93,$CD,$7F,$00,$3E,$B9,$FB,$4E,$F9
	.dc.b	$17,$BA,$D7,$EB,$79,$8F,$5E,$3F,$8A,$FD,$A2,$DF,$0F,$C5,$7F,$7D
	.dc.b	$87,$F3,$3E,$BF,$A7,$CE,$C3,$C7,$5A,$FD,$23,$F0,$31,$8D,$BC,$BB
	.dc.b	$3F,$70,$F0,$5F,$48,$FF,$39,$E3,$D2,$F8,$65,$EB,$FA,$ED,$E3,$CB
	.dc.b	$F1,$DF,$F6,$0F,$EB,$9F,$E3,$5F,$6D,$FB,$AF,$E0,$A9,$FA,$37,$E3
	.dc.b	$7F,$64,$7E,$9F,$C3,$8B,$FC,$8F,$87,$0E,$CF,$E3,$57,$BF,$86,$D5
	.dc.b	$7F,$94,$32,$7E,$A5,$FD,$7F,$4F,$04,$7E,$42,$7F,$C0,$31,$8C,$39
	.dc.b	$F9,$4F,$BA,$FF,$2A,$F1,$AB,$5B,$F9,$AB,$FC,$3D,$87,$83,$5F,$24
	.dc.b	$E7,$7F,$E1,$78,$71,$AF,$4E,$7E,$FD,$BD,$DF,$7B,$E8,$FD,$CD,$F8
	.dc.b	$F3,$F7,$67,$C8,$7F,$2E,$FA,$97,$E2,$DF,$84,$D3,$A6,$8C,$78,$E1
	.dc.b	$C3,$E2,$BF,$C4,$36,$F0,$7B,$66,$FC,$CB,$E2,$63,$FE,$72,$C7,$8B
	.dc.b	$F1,$0C,$4F,$8F,$4F,$17,$BD,$F1,$39,$7F,$CD,$D7,$E0,$9F,$35,$E4
	.dc.b	$E3,$A7,$8F,$6F,$4F,$D8,$F8,$A3,$1A,$BC,$1F,$B3,$F2,$DF,$D7,$3F
	.dc.b	$4B,$FF,$10,$7C,$A3,$F7,$97,$E2,$2F,$D0,$B7,$F2,$0B,$DD,$6B,$F2
	.dc.b	$9C,$F5,$7B,$B8,$3F,$C0,$77,$B7,$DC,$62,$FB,$FE,$07,$FE,$89,$EA
	.dc.b	$F9,$BA,$F8,$FE,$A1,$B7,$8F,$E3,$BD,$D8,$F1,$9C,$EB,$E6,$FA,$BF
	.dc.b	$2E,$7E,$4E,$07,$EA,$38,$BD,$C7,$FB,$7B,$4E,$FB,$7A,$E7,$3F,$A8
	.dc.b	$7E,$73,$E0,$9F,$78,$FD,$39,$2F,$C4,$5F,$44,$FD,$E5,$F8,$9F,$1C
	.dc.b	$38,$5E,$EB,$5F,$88,$E7,$1F,$8D,$97,$6F,$D3,$6E,$1C,$33,$EB,$F8
	.dc.b	$2F,$C0,$AF,$1D,$F8,$A7,$C3,$F9,$06,$6F,$86,$4F,$CC,$3F,$C2,$FC
	.dc.b	$3F,$98,$C7,$8E,$BE,$39,$7C,$CA,$9F,$03,$FE,$BB,$FD,$13,$FC,$5B
	.dc.b	$F9,$E3,$F3,$8F,$F4,$46,$EF,$CA,$7F,$9D,$E1,$C3,$B7,$FD,$9B,$87
	.dc.b	$0B,$DD,$7F,$A6,$EB,$54,$AF,$F1,$CF,$E9,$7F,$E6,$5C,$BC,$1F,$60
	.dc.b	$CF,$17,$D2,$31,$97,$8B,$EC,$3F,$83,$DF,$0B,$F3,$7E,$0D,$F5,$FC
	.dc.b	$8C,$E3,$C6,$73,$B7,$A7,$F1,$BF,$81,$87,$1C,$BC,$FF,$9C,$E3,$0F
	.dc.b	$F7,$3F,$EE,$7F,$D1,$FE,$61,$D2,$FD,$57,$FD,$4D,$C4,$FD,$DF,$86
	.dc.b	$0C,$38,$5E,$FF,$05,$F8,$1F,$B2,$EA,$7F,$A9,$7E,$21,$EE,$BE,$1F
	.dc.b	$1D,$FD,$B9,$FF,$53,$EF,$FE,$ED,$AE,$7A,$7F,$AD,$7F,$59,$F6,$D7
	.dc.b	$0D,$BC,$B3,$FE,$B9,$FC,$AF,$37,$C3,$4F,$A1,$FB,$77,$87,$3F,$EF
	.dc.b	$7B,$79,$78,$3D,$BD,$7F,$5B,$0E,$11,$F5,$E3,$15,$C3,$2F,$83,$FB
	.dc.b	$4F,$C0,$E9,$FF,$05,$F9,$0F,$D7,$FF,$C3,$9D,$AF,$C2,$5F,$65,$BF
	.dc.b	$17,$F4,$F5,$A9,$7E,$D3,$9C,$3E,$26,$3F,$E1,$6F,$26,$E1,$C3,$EA
	.dc.b	$FF,$D0,$38,$7C,$2A,$F0,$E3,$E0,$FC,$A3,$17,$BD,$1F,$B1,$FF,$0C
	.dc.b	$E1,$FC,$9E,$1C,$32,$FF,$45,$CB,$F2,$8F,$B7,$F9,$8F,$F9,$37,$E9
	.dc.b	$FF,$A3,$7E,$7B,$FD,$AB,$F5,$AF,$F7,$46,$5F,$9C,$AD,$7C,$2A,$B5
	.dc.b	$EB,$F2,$FB,$4F,$E9,$DE,$0F,$DC,$02,$FB,$39,$31,$1F,$E3,$4B,$E9
	.dc.b	$F2,$5F,$B7,$9E,$99,$F9,$23,$F9,$B4,$62,$77,$9F,$F5,$1C,$FC,$D8
	.dc.b	$73,$CF,$D5,$FB,$07,$F0,$1F,$A5,$F1,$E2,$FA,$3F,$E7,$D9,$FF,$9E
	.dc.b	$7E,$EE,$75,$EE,$BD,$35,$A9,$D1,$FE,$BA,$CF,$FC,$5B,$3A,$F8,$1F
	.dc.b	$D4,$BD,$8F,$EF,$F8,$F8,$2F,$BC,$64,$EC,$78,$E7,$FC,$33,$7E,$3F
	.dc.b	$84,$E3,$3D,$7C,$4E,$1F,$3B,$C3,$0F,$EA,$D1,$CF,$E7,$71,$8B,$C7
	.dc.b	$FF,$7A,$67,$E1,$8F,$2E,$3F,$53,$E4,$FA,$AF,$B5,$7E,$34,$F8,$2F
	.dc.b	$EC,$58,$3F,$16,$7A,$5F,$E1,$BE,$2F,$83,$8F,$1C,$38,$67,$FE,$50
	.dc.b	$BD,$FC,$E7,$BB,$3E,$BE,$F7,$93,$0F,$C2,$C7,$83,$4C,$47,$37,$C3
	.dc.b	$6F,$C6,$7C,$F5,$BC,$7B,$7C,$36,$33,$F0,$ED,$E7,$D9,$F9,$37,$86
	.dc.b	$73,$97,$AB,$F6,$AF,$A3,$F7,$7F,$EF,$8D,$1F,$98,$BF,$0B,$7F,$00
	.dc.b	$7F,$1F,$F1,$B7,$F6,$3E,$3A,$D4,$FF,$23,$FC,$77,$3E,$77,$AB,$2E
	.dc.b	$9C,$BF,$7F,$A3,$E3,$C7,$7C,$38,$FA,$3F,$4A,$31,$A7,$83,$D0,$C6
	.dc.b	$9D,$E3,$C7,$EC,$FF,$8E,$F1,$E7,$39,$D7,$F9,$A7,$97,$FD,$7F,$EC
	.dc.b	$FF,$3F,$FB,$A3,$F0,$37,$EC,$8F,$D8,$BF,$CF,$1E,$F3,$F7,$57,$EB
	.dc.b	$5C,$79,$3F,$73,$8F,$1C,$38,$7D,$B7,$F5,$DB,$FE,$51,$9F,$6C,$FA
	.dc.b	$E1,$EF,$8C,$7E,$9B,$F9,$F8,$E8,$31,$E9,$A7,$8B,$F2,$4F,$F3,$2D
	.dc.b	$F3,$CF,$BE,$BF,$81,$C3,$8E,$1C,$34,$F1,$7D,$16,$34,$F6,$E7,$EC
	.dc.b	$D7,$F7,$DC,$5F,$EB,$EC,$7D,$75,$AC,$3D,$5F,$A3,$71,$D3,$BE,$0F
	.dc.b	$DA,$7C,$FD,$AF,$CD,$1F,$C0,$2F,$DE,$FF,$DE,$7F,$BD,$7E,$1B,$0D
	.dc.b	$D6,$A7,$3D,$0F,$AD,$CE,$1F,$0F,$83,$C9,$C0,$CF,$B7,$E5,$99,$E6
	.dc.b	$FE,$2F,$1E,$1B,$F0,$7E,$18,$DB,$FD,$0F,$8F,$2A,$D6,$9E,$4F,$CB
	.dc.b	$3E,$5E,$7F,$CC,$A7,$38,$71,$D5,$FB,$1F,$93,$0F,$3F,$E2,$FF,$5B
	.dc.b	$F7,$97,$E3,$1F,$0F,$ED,$DF,$8B,$BF,$7C,$7E,$55,$FE,$F0,$E0,$F5
	.dc.b	$E1,$C3,$A3,$FD,$3D,$7B,$A7,$EE,$BD,$72,$F9,$DC,$3F,$87,$3F,$A3
	.dc.b	$E1,$E0,$F3,$BE,$14,$62,$7C,$58,$78,$7E,$67,$DB,$2B,$F0,$6B,$E6
	.dc.b	$D5,$F3,$7A,$9C,$E1,$E3,$FD,$5F,$C5,$B7,$9A,$FD,$7F,$C0,$3F,$96
	.dc.b	$F9,$77,$ED,$FF,$E9,$FF,$80,$FF,$3B,$FC,$DF,$F9,$47,$F6,$3D,$4E
	.dc.b	$7E,$DF,$FC,$51,$F5,$AF,$DF,$F3,$F8,$1F,$AB,$8F,$FC,$D3,$6E,$EC
	.dc.b	$69,$DA,$BF,$CA,$18,$FF,$8C,$7E,$CF,$DB,$39,$D7,$C3,$F1,$98,$C7
	.dc.b	$C1,$B7,$F9,$8F,$F6,$2E,$F5,$E6,$D9,$FB,$16,$33,$F2,$FE,$DB,$70
	.dc.b	$FF,$35,$7D,$37,$F9,$97,$9B,$FB,$2F,$CC,$3F,$08,$7E,$46,$FD,$1D
	.dc.b	$F8,$6B,$F5,$EF,$E8,$DF,$27,$C7,$A3,$FB,$FE,$3C,$70,$E1,$7B,$F9
	.dc.b	$3F,$F4,$7F,$90,$D3,$BE,$9D,$B1,$FE,$4D,$18,$F9,$1F,$D3,$3A,$C7
	.dc.b	$9E,$8F,$B6,$FF,$75,$CE,$76,$F2,$CF,$F8,$1B,$2E,$D8,$F9,$32,$F5
	.dc.b	$D6,$B3,$FE,$41,$5A,$AF,$F0,$E5,$6B,$4F,$47,$AB,$ED,$97,$ED,$2F
	.dc.b	$C8,$BF,$C5,$5F,$60,$FC,$41,$FA,$5B,$F5,$F7,$EE,$7C,$70,$E1,$F5
	.dc.b	$5F,$E9,$2B,$DD,$3F,$20,$D6,$5F,$E3,$5C,$FE,$19,$FE,$2A,$C3,$C1
	.dc.b	$F0,$3F,$12,$31,$3E,$2C,$3C,$3F,$90,$FC,$49,$CE,$5D,$F4,$ED,$1D
	.dc.b	$A3,$19,$BF,$19,$C6,$DE,$BB,$F5,$7E,$35,$F0,$7F,$05,$7E,$81,$F8
	.dc.b	$5C,$1F,$BE,$BF,$40,$7E,$B8,$FE,$01,$E1,$7B,$AD,$4B,$F4,$FC,$F4
	.dc.b	$FA,$F8,$3F,$DA,$76,$FC,$8B,$1A,$78,$2B,$E6,$63,$CA,$F8,$E9,$E2
	.dc.b	$D1,$FA,$F7,$C1,$9E,$77,$C2,$7C,$BF,$1B,$C3,$0F,$1F,$CA,$D9,$F4
	.dc.b	$F8,$B7,$7F,$F6,$FF,$7B,$D8,$FF,$24,$FD,$5F,$FB,$97,$9D,$F8,$D3
	.dc.b	$F1,$17,$E1,$7F,$80,$FC,$CF,$FD,$A1,$87,$0F,$C0,$7E,$75,$EE,$BF
	.dc.b	$0B,$F6,$5A,$E3,$FF,$33,$E7,$FE,$F1,$E1,$FE,$8D,$8F,$1D,$2D,$6B
	.dc.b	$2E,$7A,$78,$3C,$4F,$F3,$7E,$DE,$7A,$D5,$F9,$AB,$CB,$BE,$B1,$E9
	.dc.b	$B7,$8F,$C5,$FE,$C1,$8F,$1C,$7F,$C9,$59,$F6,$DF,$F3,$7F,$2D,$E7
	.dc.b	$97,$EC,$0F,$D5,$1C,$9F,$B5,$FC,$85,$FA,$AF,$F4,$6F,$C1,$38,$5E
	.dc.b	$EB,$53,$FD,$2B,$E5,$F3,$87,$2C,$7F,$C5,$1C,$1F,$E7,$8C,$3B,$7C
	.dc.b	$0F,$C2,$9F,$0D,$78,$3D,$46,$9E,$0F,$8D,$F7,$57,$FA,$4B,$0F,$05
	.dc.b	$6B,$3F,$E8,$B3,$E0,$D3,$C7,$F2,$3F,$BE,$E2,$F8,$58,$70,$CB,$D7
	.dc.b	$2F,$DB,$3E,$17,$E9,$8F,$CC,$BF,$93,$7D,$EF,$FA,$1B,$9B,$31,$88
	.dc.b	$FF,$47,$7E,$DB,$8F,$67,$D5,$7C,$5F,$F4,$6F,$DF,$7A,$37,$4F,$5F
	.dc.b	$4F,$FD,$11,$F9,$57,$47,$FD,$15,$F5,$DD,$BC,$CE,$7A,$27,$FD,$25
	.dc.b	$F7,$BC,$FF,$73,$7C,$C3,$F2,$67,$E6,$6F,$C4,$1F,$B8,$7E,$C7,$EA
	.dc.b	$3E,$F6,$1C,$2F,$75,$AF,$5B,$FC,$47,$D7,$F8,$7A,$EB,$FA,$FF,$D5
	.dc.b	$78,$71,$FD,$A3,$FD,$86,$F3,$CF,$C1,$9F,$F1,$4B,$E5,$A7,$96,$73
	.dc.b	$B7,$A6,$31,$9F,$87,$E2,$3C,$B1,$8C,$3F,$84,$7D,$13,$17,$E5,$DF
	.dc.b	$EC,$0F,$DF,$DF,$A5,$3C,$77,$CB,$3F,$7A,$77,$FF,$A7,$BE,$6F,$96
	.dc.b	$3C,$70,$E1,$7B,$AF,$F5,$1F,$CE,$6B,$93,$FD,$BF,$9F,$6C,$FF,$C7
	.dc.b	$D8,$F1,$F0,$AD,$6B,$2E,$7A,$78,$3E,$FF,$F7,$37,$DF,$0F,$2C,$F9
	.dc.b	$3C,$8C,$E1,$C6,$73,$A7,$FA,$D3,$CA,$F1,$4B,$4F,$16,$DE,$7F,$C1
	.dc.b	$1F,$BE,$EB,$F1,$27,$DB,$3E,$E7,$FB,$67,$EC,$9F,$17,$F7,$0F,$F5
	.dc.b	$4F,$3F,$CA,$D3,$FD,$F7,$87,$0B,$DE,$EB,$7E,$23,$D7,$2E,$98,$7F
	.dc.b	$A8,$78,$FF,$DE,$38,$F8,$3F,$2F,$F8,$72,$8E,$9C,$2D,$7E,$B7,$E4
	.dc.b	$DF,$02,$B5,$8F,$9E,$B5,$A3,$F5,$7D,$63,$CB,$1F,$57,$EB,$FE,$3D
	.dc.b	$FC,$FB,$BF,$5B,$C6,$5C,$BE,$B1,$F8,$7E,$31,$FC,$34,$3E,$5B,$E6
	.dc.b	$FD,$DB,$F6,$B7,$D6,$BF,$57,$7E,$E3,$E9,$E9,$FC,$89,$F2,$79,$7F
	.dc.b	$D1,$37,$C2,$B7,$9F,$F4,$6A,$D6,$3D,$7E,$3B,$E2,$BD,$F2,$B8,$71
	.dc.b	$CB,$B6,$5E,$48,$C6,$3F,$E2,$1E,$94,$CC,$39,$67,$E0,$CF,$FB,$C9
	.dc.b	$E0,$C3,$8B,$FD,$A8,$F4,$7B,$2F,$D9,$3F,$82,$FC,$8B,$F0,$07,$EF
	.dc.b	$0F,$D2,$9F,$B7,$7E,$4B,$0E,$1E,$7B,$0E,$17,$BF,$D0,$75,$97,$E5
	.dc.b	$F7,$FC,$31,$E8,$C3,$8B,$80,$F4,$BE,$7F,$29,$F8,$58,$C6,$9E,$4F
	.dc.b	$92,$C6,$1E,$88,$F3,$EB,$FE,$BB,$DB,$DB,$39,$D3,$D5,$FB,$0F,$F0
	.dc.b	$9F,$82,$F9,$3A,$BF,$19,$7F,$00,$FF,$BA,$BF,$12,$7E,$90,$FB,$8F
	.dc.b	$F5,$5F,$ED,$7C,$70,$FA,$38,$70,$FA,$C6,$1C,$2F,$FD,$63,$DB,$BB
	.dc.b	$ED,$97,$EA,$1F,$09,$9C,$39,$38,$31,$1E,$3C,$BC,$5E,$A7,$F8,$0B
	.dc.b	$4F,$4D,$F8,$72,$F0,$7A,$D9,$C7,$CB,$8F,$1C,$FF,$D8,$7F,$81,$F4
	.dc.b	$D7,$CF,$A7,$B7,$F0,$77,$E9,$19,$CF,$D5,$0F,$D9,$B1,$5B,$EE,$F6
	.dc.b	$7E,$48,$FD,$E5,$F9,$B3,$E2,$3F,$DB,$5F,$AC,$FF,$57,$FC,$53,$75
	.dc.b	$A9,$CF,$C4,$FE,$B3,$B5,$CF,$F5,$32,$E9,$F8,$17,$FB,$AE,$31,$87
	.dc.b	$6E,$6F,$8D,$7C,$1F,$D1,$DF,$8F,$4F,$17,$CF,$B5,$FE,$D1,$5E,$29
	.dc.b	$CC,$7F,$55,$C7,$AC,$63,$3F,$F2,$B6,$3D,$6D,$8F,$5D,$3D,$3F,$BE
	.dc.b	$FD,$DF,$C8,$9F,$AB,$39,$FF,$AA,$FE,$DE,$FC,$97,$F9,$17,$EC,$7B
	.dc.b	$AD,$7A,$2A,$D7,$BA,$FD,$A7,$58,$FC,$1D,$BF,$7F,$9F,$6F,$77,$10
	.dc.b	$8F,$0C,$F8,$34,$7E,$47,$F8,$3A,$D6,$9E,$0C,$FC,$9D,$D9,$C7,$8E
	.dc.b	$BE,$3C,$7F,$9C,$7D,$EF,$8A,$31,$B7,$FA,$1B,$8D,$F3,$BF,$AA,$FD
	.dc.b	$F1,$F8,$97,$F5,$0F,$DC,$1F,$A6,$BF,$38,$FF,$45,$7E,$CF,$FD,$E7
	.dc.b	$0E,$17,$BA,$D7,$DC,$3F,$43,$D7,$27,$F1,$EC,$FA,$FC,$67,$E9,$63
	.dc.b	$22,$3C,$3F,$73,$9D,$BC,$B3,$C7,$2E,$DA,$BE,$EB,$FC,$D7,$AF,$8B
	.dc.b	$26,$FE,$8A,$7E,$9F,$AC,$79,$56,$B7,$F3,$FE,$77,$E2,$DB,$CB,$8F
	.dc.b	$1E,$07,$E4,$D9,$DB,$C3,$0F,$A4,$C6,$5C,$F5,$FF,$57,$7D,$37,$B3
	.dc.b	$F1,$AF,$F3,$87,$E7,$CF,$DE,$1F,$0C,$F8,$F7,$EC,$DF,$E1,$CE,$7F
	.dc.b	$7F,$23,$1E,$38,$70,$BD,$FC,$9F,$5F,$B8,$7E,$85,$F1,$FA,$7F,$91
	.dc.b	$E1,$E0,$C7,$F5,$B1,$8A,$6B,$E3,$D7,$C5,$F1,$1F,$16,$FC,$15,$AD
	.dc.b	$3C,$1F,$82,$E7,$2F,$1E,$1E,$9E,$7F,$EE,$7B,$FA,$27,$3B,$FD,$2F
	.dc.b	$89,$FB,$BA,$FF,$34,$FC,$A5,$FA,$37,$E3,$BF,$25,$FF,$1E,$FE,$8D
	.dc.b	$AF,$84,$56,$BB,$15,$FE,$42,$FD,$1F,$53,$EF,$E3,$FE,$D3,$97,$F4
	.dc.b	$0F,$CC,$F1,$1E,$EC,$79,$E7,$DF,$F6,$1F,$75,$F8,$E3,$C5,$FB,$4E
	.dc.b	$2B,$7A,$78,$3E,$A1,$F8,$9F,$82,$FD,$10,$FD,$83,$F3,$3F,$64,$F8
	.dc.b	$47,$B9,$FE,$56,$FA,$D9,$E3,$EA,$73,$19,$3F,$80,$74,$FE,$02,$F9
	.dc.b	$5F,$A9,$B3,$CF,$E5,$FE,$B4,$77,$F9,$8F,$E9,$DF,$15,$D5,$FE,$A1
	.dc.b	$F1,$DD,$FF,$09,$E5,$9A,$7F,$D1,$FC,$DF,$E7,$1D,$BF,$B4,$FE,$53
	.dc.b	$FE,$6E,$F7,$5F,$B4,$BF,$79,$7E,$26,$FA,$DC,$71,$FC,$1C,$BF,$C3
	.dc.b	$38,$F1,$C3,$85,$EE,$BB,$56,$B9,$7F,$CA,$2F,$F4,$0C,$FC,$31,$FC
	.dc.b	$A6,$30,$6F,$3F,$1E,$7E,$2C,$FF,$DD,$37,$C2,$73,$8F,$F9,$36,$FB
	.dc.b	$8C,$FC,$98,$71,$F8,$DC,$C6,$2B,$D9,$FA,$77,$0F,$DC,$9F,$92,$BC
	.dc.b	$17,$E5,$AF,$D7,$DF,$A1,$BF,$3E,$7E,$85,$C7,$0E,$17,$BA,$D7,$CC
	.dc.b	$67,$2F,$E3,$9C,$5F,$E1,$87,$BE,$78,$E1,$FF,$21,$C6,$30,$E5,$F3
	.dc.b	$DF,$02,$73,$97,$6D,$7C,$7F,$07,$99,$61,$C6,$7C,$FB,$3F,$1B,$F3
	.dc.b	$67,$EC,$9C,$FC,$F7,$C8,$FD,$B9,$F8,$CF,$C0,$FB,$A7,$EB,$6F,$A2
	.dc.b	$FF,$12,$FE,$73,$BA,$D4,$E7,$E5,$1F,$53,$9F,$CE,$7D,$9C,$2C,$FA
	.dc.b	$F2,$E1,$BA,$76,$D3,$FC,$63,$5D,$6B,$FD,$2F,$97,$87,$F4,$3F,$76
	.dc.b	$1C,$23,$C1,$F9,$86,$F1,$E5,$A3,$F5,$BC,$57,$97,$F7,$3B,$A7,$F8
	.dc.b	$FB,$F0,$CF,$F6,$97,$A9,$FE,$4A,$FB,$9F,$F1,$0F,$E5,$2F,$DC,$1F
	.dc.b	$89,$3D,$07,$E3,$AA,$E3,$C9,$FE,$7A,$C3,$8E,$1C,$2F,$7F,$B1,$EA
	.dc.b	$BF,$D2,$7F,$56,$8F,$EB,$D9,$F8,$63,$A4,$FF,$89,$23,$18,$F8,$F3
	.dc.b	$F1,$67,$FE,$88,$F4,$3F,$8A,$DF,$87,$0F,$26,$7E,$1F,$2F,$59,$72
	.dc.b	$8F,$26,$0F,$53,$C9,$B7,$A3,$3E,$BE,$AF,$9F,$89,$FA,$6F,$F0,$2F
	.dc.b	$93,$FC,$6F,$EC,$2F,$C5,$1F,$96,$EE,$B5,$E4,$AB,$5F,$53,$9F,$CD
	.dc.b	$7D,$FC,$9F,$EA,$0C,$FA,$FA,$98,$84,$63,$4E,$FE,$17,$AA,$78,$67
	.dc.b	$DB,$27,$EC,$79,$9E,$33,$9D,$BF,$CA,$7D,$1F,$1A,$73,$9B,$2F,$5D
	.dc.b	$7F,$B0,$B9,$BF,$D1,$7F,$8D,$7E,$27,$F4,$3F,$F1,$2F,$E0,$0F,$A9
	.dc.b	$7A,$9F,$EF,$6D,$3A,$5E,$FE,$6B,$FD,$5D,$5A,$FD,$F7,$F0,$13,$CF
	.dc.b	$6F,$B8,$7F,$B0,$1F,$CB,$76,$7C,$17,$96,$5D,$F2,$F6,$46,$30,$7D
	.dc.b	$EE,$3C,$76,$BE,$69,$F2,$FC,$0F,$82,$73,$AF,$93,$17,$EE,$39,$C3
	.dc.b	$84,$72,$DF,$D7,$7C,$34,$F1,$6B,$F9,$9F,$2D,$F9,$FB,$F7,$E7,$E5
	.dc.b	$6F,$D1,$DF,$AC,$3E,$4F,$FD,$DD,$CF,$FE,$B6,$C7,$8E,$1C,$2F,$75
	.dc.b	$FA,$BD,$7E,$B5,$76,$FD,$87,$F0,$8F,$0E,$7F,$92,$1E,$79,$F8,$73
	.dc.b	$F0,$4F,$9F,$4F,$34,$E7,$E5,$BF,$0D,$7C,$33,$F0,$FD,$1E,$F5,$F2
	.dc.b	$65,$E9,$FC,$6D,$AB,$F2,$6F,$F6,$7F,$E1,$9F,$E8,$6F,$B6,$62,$FC
	.dc.b	$91,$FB,$E3,$F2,$A7,$CD,$72,$C7,$8F,$DE,$31,$E3,$87,$0B,$7E,$9D
	.dc.b	$BE,$50,$D3,$C0,$FF,$AA,$31,$E7,$F9,$B7,$F9,$9D,$EB,$8F,$4D,$3F
	.dc.b	$8B,$D7,$33,$96,$7E,$1D,$7D,$13,$FD,$9E,$75,$8F,$B6,$73,$9B,$F2
	.dc.b	$3C,$E1,$C7,$5F,$47,$E3,$DD,$7F,$79,$7D,$AB,$F2,$57,$D5,$3F,$2C
	.dc.b	$7E,$19,$FE,$F7,$FA,$DE,$17,$BA,$D4,$FF,$BE,$FF,$40,$CE,$1E,$DE
	.dc.b	$1F,$87,$97,$6F,$92,$C4,$23,$C3,$3E,$0C,$FB,$FE,$01,$FC,$9F,$9B
	.dc.b	$E5,$DE,$F4,$F0,$57,$93,$E0,$F5,$3D,$A7,$3A,$3F,$12,$FB,$FD,$9F
	.dc.b	$83,$AD,$67,$DE,$B5,$FC,$01,$E8,$E0,$F5,$E0,$FC,$D7,$1A,$7C,$28
	.dc.b	$7E,$C9,$F5,$7F,$6F,$7E,$AA,$FD,$75,$FB,$D3,$F7,$2F,$C7,$FC,$03
	.dc.b	$F0,$1E,$17,$BA,$D7,$81,$FC,$93,$EC,$73,$F9,$07,$C6,$CF,$AE,$77
	.dc.b	$0E,$19,$F4,$E4,$E7,$3E,$08,$FC,$EF,$CD,$FC,$49,$5F,$2C,$FB,$FE
	.dc.b	$CB,$FE,$CA,$E7,$F8,$95,$AC,$BA,$EA,$AF,$E6,$7D,$5F,$4B,$F5,$A7
	.dc.b	$E5,$6F,$D9,$1F,$A6,$5F,$97,$F8,$BD,$5F,$99,$BD,$D6,$8F,$EE,$5C
	.dc.b	$EF,$BE,$D6,$3C,$F6,$FD,$31,$FA,$AB,$EF,$F6,$7F,$88,$79,$E7,$DF
	.dc.b	$B3,$DD,$18,$CB,$F8,$A5,$F4,$7C,$1A,$7D,$8F,$A8,$C6,$1E,$7D,$1F
	.dc.b	$4F,$FC,$23,$F1,$BF,$C7,$FE,$F7,$E5,$FB,$AF,$F4,$07,$6B,$F3,$2D
	.dc.b	$6B,$FB,$EA,$D4,$FF,$A1,$7F,$0C,$CC,$A7,$93,$F9,$A3,$FC,$05,$87
	.dc.b	$6E,$36,$3F,$E8,$D3,$33,$D6,$3F,$CB,$AF,$1F,$CB,$3A,$DF,$C9,$FB
	.dc.b	$1E,$93,$FE,$BC,$C3,$A7,$E6,$9F,$EF,$37,$B4,$FF,$A2,$BE,$E5,$FA
	.dc.b	$B7,$EB,$3F,$29,$7E,$EC,$FD,$F9,$F8,$F7,$F2,$5F,$D7,$DF,$8E,$7F
	.dc.b	$5A,$C7,$8E,$BF,$C7,$B1,$E3,$87,$0F,$69,$87,$0F,$CE,$FF,$D2,$9A
	.dc.b	$77,$D3,$B6,$9D,$7F,$55,$FE,$4F,$A7,$4D,$1A,$74,$AF,$17,$D5,$6B
	.dc.b	$1E,$5A,$F7,$D1,$F9,$5E,$71,$F2,$69,$DB,$6F,$1F,$E3,$EE,$9F,$1D
	.dc.b	$6B,$1E,$5F,$2D,$FE,$45,$DF,$C5,$BF,$87,$1F,$5C,$BF,$32,$FF,$57
	.dc.b	$7B,$1F,$C5,$3F,$4A,$FF,$27,$7E,$DA,$FC,$31,$F6,$6F,$D9,$5F,$9C
	.dc.b	$FF,$06,$D7,$F9,$5E,$FF,$D1,$B1,$E3,$87,$0B,$DF,$EB,$5D,$AB,$5F
	.dc.b	$70,$CF,$B6,$7D,$7C,$EF,$93,$AF,$8E,$B5,$97,$3F,$CE,$F3,$8F,$2C
	.dc.b	$78,$E7,$E4,$FC,$0B,$B4,$B5,$F1,$E9,$E9,$DD,$FB,$1E,$A7,$38,$3F
	.dc.b	$19,$F8,$1C,$3E,$CC,$79,$6D,$E2,$CB,$E1,$57,$FA,$4E,$9F,$A5,$7F
	.dc.b	$3C,$FB,$27,$E9,$6F,$CA,$3F,$D6,$DF,$A6,$7D,$7F,$C9,$5F,$8D,$7F
	.dc.b	$B7,$70,$E1,$7B,$FC,$35,$7B,$FC,$0F,$5B,$FF,$72,$CF,$AE,$7D,$3C
	.dc.b	$0F,$F0,$0C,$E8,$FF,$30,$63,$CF,$0F,$C3,$7E,$45,$F9,$4C,$FB,$E1
	.dc.b	$C7,$57,$EC,$3C,$73,$EB,$AF,$8B,$45,$3D,$4D,$69,$D6,$5F,$91,$7F
	.dc.b	$98,$B8,$7E,$1E,$FE,$7A,$D4,$FF,$31,$9F,$EC,$D3,$FE,$0C,$D7,$FD
	.dc.b	$89,$F2,$7F,$EF,$DD,$1F,$9C,$FF,$8F,$7F,$4B,$61,$C3,$0F,$9F,$EE
	.dc.b	$FF,$BB,$7C,$2F,$F3,$F5,$FF,$90,$FF,$7D,$DF,$CC,$7F,$6E,$D3,$AD
	.dc.b	$6B,$E0,$FF,$6B,$39,$C3,$96,$9E,$0F,$8A,$CE,$7E,$59,$CE,$6F,$AE
	.dc.b	$FE,$1B,$AF,$8F,$4E,$B8,$FA,$25,$FC,$03,$E2,$C7,$97,$07,$B7,$1F
	.dc.b	$ED,$FF,$6F,$99,$F5,$CE,$7E,$0D,$F9,$C7,$C0,$F6,$3F,$C0,$FF,$80
	.dc.b	$FF,$A6,$E5,$FB,$9B,$DA,$FB,$3F,$98,$7E,$23,$87,$F8,$FE,$7F,$E3
	.dc.b	$8C,$38,$5E,$FB,$D6,$FD,$53,$B5,$EF,$8B,$FC,$07,$7B,$FB,$1F,$BB
	.dc.b	$97,$4A,$FF,$59,$D6,$B3,$FC,$4D,$7E,$1A,$9F,$99,$FF,$87,$BE,$B3
	.dc.b	$F8,$3F,$E0,$2F,$88,$FE,$CD,$F9,$9B,$F5,$7F,$E2,$58,$70,$D3,$FC
	.dc.b	$9F,$87,$0B,$DD,$7E,$12,$B5,$5F,$E5,$9F,$4D,$8F,$5F,$8B,$64,$FC
	.dc.b	$23,$F9,$9E,$9E,$3B,$DE,$7E,$2F,$C5,$73,$97,$82,$73,$8F,$2C,$3F
	.dc.b	$C9,$59,$F8,$35,$F2,$4E,$7E,$A7,$F2,$B9,$F5,$D7,$C5,$AF,$FB,$13
	.dc.b	$3C,$9D,$63,$3F,$38,$FC,$0F,$33,$EB,$9C,$FE,$4D,$8E,$8F,$F7,$0F
	.dc.b	$D3,$FF,$B1,$FF,$2A,$7E,$9D,$FE,$31,$E9,$D7,$67,$F9,$77,$F0,$B5
	.dc.b	$6B,$EA,$1F,$95,$EB,$1E,$5F,$BD,$FF,$94,$B9,$BE,$5E,$9D,$B1,$E3
	.dc.b	$A7,$F7,$AD,$F8,$47,$4B,$FF,$41,$FE,$51,$8D,$11,$FE,$86,$FC,$03
	.dc.b	$0B,$F1,$6F,$9B,$87,$F0,$ED,$FC,$D7,$C3,$A3,$FC,$63,$AF,$8A,$73
	.dc.b	$87,$F8,$AF,$83,$F8,$0F,$E0,$1F,$E6,$8F,$DA,$1F,$8A,$FF,$53,$FC
	.dc.b	$CF,$FD,$C7,$FA,$1B,$F1,$47,$E9,$1C,$37,$FF,$4C,$E1,$C3,$E2,$58
	.dc.b	$70,$BF,$CA,$6D,$FE,$C8,$BD,$E5,$D7,$F0,$F6,$7F,$8C,$9E,$83,$2E
	.dc.b	$98,$F8,$BF,$46,$FF,$00,$EB,$E7,$C7,$96,$0F,$C6,$3F,$B5,$67,$DF
	.dc.b	$2E,$55,$A9,$FF,$4B,$4E,$72,$EB,$9F,$8B,$F6,$3F,$EB,$93,$9B,$EB
	.dc.b	$0F,$C8,$FC,$B7,$C3,$27,$E5,$1F,$E1,$4F,$D9,$BF,$D2,$3E,$4F,$F8
	.dc.b	$93,$07,$E3,$6F,$D6,$1F,$90,$FF,$80,$7F,$0C,$FE,$63,$87,$0F,$61
	.dc.b	$87,$0B,$DF,$CD,$7C,$CE,$05,$EE,$B5,$8B,$F1,$CF,$17,$27,$FB,$33
	.dc.b	$5F,$16,$BE,$1A,$71,$EA,$BF,$75,$AF,$82,$B5,$83,$F0,$7F,$2D,$73
	.dc.b	$AD,$6D,$EA,$A7,$E1,$1A,$AE,$7A,$79,$3E,$C7,$B6,$BE,$2C,$3E,$0F
	.dc.b	$E1,$3F,$D5,$F7,$F5,$6B,$E5,$87,$E0,$8F,$D8,$5F,$BE,$7F,$AF,$3F
	.dc.b	$1A,$7E,$6D,$FE,$F1,$FC,$27,$C3,$F4,$5E,$2F,$D6,$B8,$5E,$EB,$5C
	.dc.b	$5F,$E1,$41,$F3,$9A,$E3,$FF,$78,$E5,$D3,$F1,$1F,$F6,$C6,$5F,$06
	.dc.b	$74,$27,$59,$F8,$3E,$FB,$FD,$F3,$AF,$9A,$B9,$E7,$E0,$C3,$E5,$7E
	.dc.b	$F3,$CB,$1E,$3A,$7F,$89,$BA,$5F,$C0,$19,$C3,$87,$DD,$F4,$E0,$F5
	.dc.b	$6F,$E4,$FD,$71,$F9,$DE,$E3,$1C,$FE,$6F,$C5,$5F,$A1,$7E,$3B,$F6
	.dc.b	$07,$E1,$EF,$DE,$7F,$CE,$BF,$96,$6E,$B5,$39,$FC,$09,$FA,$DE,$7C
	.dc.b	$76,$3C,$B1,$7E,$05,$E5,$E9,$0D,$3C,$37,$C3,$2F,$07,$C9,$EE,$B5
	.dc.b	$A7,$83,$1F,$F3,$9F,$EA,$BD,$F3,$ED,$F8,$0B,$F5,$7C,$65,$E9,$FC
	.dc.b	$D1,$F9,$DE,$31,$47,$FB,$33,$C0,$7E,$E9,$E4,$D7,$F8,$27,$C9,$7F
	.dc.b	$A5,$3E,$43,$F0,$3E,$83,$F6,$FF,$46,$1C,$2F,$D7,$ED,$FF,$6E,$E8
	.dc.b	$FE,$95,$E4,$FF,$A7,$AD,$E6,$7E,$57,$4E,$D7,$BC,$FA,$F4,$FE,$12
	.dc.b	$73,$AF,$87,$1F,$F5,$3F,$E4,$F9,$D3,$BC,$E7,$67,$E3,$8F,$D8,$FB
	.dc.b	$56,$B7,$F1,$D6,$BE,$33,$33,$EE,$9C,$E9,$FC,$5F,$87,$D9,$39,$DB
	.dc.b	$CB,$F7,$BC,$38,$3D,$1B,$FF,$AE,$BB,$3F,$37,$D7,$F2,$7E,$23,$F6
	.dc.b	$7F,$86,$3F,$23,$7E,$37,$FE,$FC,$F9,$DF,$E2,$1B,$30,$E1,$EB,$30
	.dc.b	$E1,$7B,$E9,$FD,$E5,$6A,$73,$F8,$BB,$EC,$BC,$3C,$5F,$8C,$D7,$C3
	.dc.b	$AF,$82,$9F,$82,$6A,$BF,$B2,$E9,$DE,$B5,$F9,$53,$E2,$BB,$D7,$3A
	.dc.b	$D6,$CC,$57,$FE,$2C,$DF,$D5,$7B,$D7,$CB,$F8,$73,$F5,$0E,$73,$E0
	.dc.b	$9C,$FE,$A1,$FC,$5F,$C7,$FC,$86,$AF,$C6,$1F,$8B,$FF,$A9,$3E,$ED
	.dc.b	$F8,$AB,$F4,$F7,$E6,$DC,$74,$FD,$5E,$3C,$70,$E1,$7B,$FD,$43,$55
	.dc.b	$ED,$CB,$AF,$C1,$33,$EB,$F0,$EF,$A0,$DE,$7C,$EB,$59,$7F,$53,$FD
	.dc.b	$3B,$38,$F2,$C7,$CB,$F1,$DF,$C7,$3F,$37,$FD,$57,$E5,$2F,$CA,$BF
	.dc.b	$D5,$3F,$8A,$FC,$9F,$DE,$BF,$DB,$3E,$17,$F3,$7B,$DD,$6B,$B5,$5F
	.dc.b	$ED,$EF,$C8,$75,$C7,$FC,$23,$2E,$9F,$87,$7F,$B9,$EF,$7F,$99,$F8
	.dc.b	$6B,$58,$F3,$C5,$EA,$E6,$58,$F2,$D7,$C9,$F9,$BF,$FB,$0B,$C1,$6D
	.dc.b	$E4,$D7,$B6,$4A,$7E,$7D,$E4,$CB,$9E,$9D,$FE,$6D,$FB,$5E,$75,$ED
	.dc.b	$C1,$EC,$97,$EB,$5F,$D5,$F3,$FE,$21,$FA,$97,$FA,$2B,$BD,$EB,$BF
	.dc.b	$6B,$7E,$E4,$FC,$0F,$FC,$09,$C9,$F1,$B8,$3F,$95,$E3,$C7,$0E,$17
	.dc.b	$BF,$A2,$ED,$5A,$FB,$6F,$C6,$D6,$BF,$0E,$FE,$A7,$AF,$8A,$73,$A7
	.dc.b	$83,$F0,$DC,$E1,$C7,$6F,$2C,$F9,$3F,$30,$7E,$D7,$DA,$9B,$78,$EB
	.dc.b	$5F,$2D,$98,$C6,$FF,$A8,$E0,$F5,$5F,$0D,$3C,$33,$FD,$2B,$F0,$5E
	.dc.b	$11,$8F,$8E,$FF,$12,$DF,$0F,$45,$FB,$43,$F3,$EF,$F3,$D7,$C8,$FF
	.dc.b	$A4,$BE,$B9,$FB,$C7,$96,$F7,$5A,$FC,$03,$FC,$B3,$F9,$EE,$71,$E5
	.dc.b	$F1,$1F,$D9,$25,$F8,$37,$C2,$E1,$FE,$2D,$5A,$D3,$FD,$3F,$F1,$99
	.dc.b	$CF,$AE,$3C,$7F,$66,$7E,$AF,$D7,$5F,$04,$3F,$30,$C5,$77,$FA,$DE
	.dc.b	$59,$F5,$D9,$FA,$DB,$F0,$1D,$ED,$FE,$AE,$87,$EB,$5F,$E2,$1F,$CF
	.dc.b	$7F,$AD,$7E,$A8,$FD,$11,$F8,$D3,$F3,$5F,$EF,$9F,$2F,$F9,$EF,$C6
	.dc.b	$B8,$73,$FF,$74,$C3,$85,$EF,$E8,$3F,$BC,$FA,$FF,$C8,$3B,$9A,$76
	.dc.b	$D3,$AE,$9D,$3E,$4F,$FC,$3B,$39,$C7,$96,$BF,$CD,$F4,$EE,$34,$EF
	.dc.b	$87,$93,$F5,$37,$E8,$9A,$9C,$ED,$FE,$0A,$97,$F0,$0E,$75,$F0,$CF
	.dc.b	$F8,$3B,$F2,$B7,$EC,$79,$D7,$C3,$0F,$8C,$C6,$9F,$0B,$E5,$FF,$C2
	.dc.b	$7F,$A1,$3F,$27,$FF,$01,$7C,$EF,$F8,$0F,$86,$0F,$BC,$D3,$FE,$1A
	.dc.b	$8C,$BF,$E0,$9F,$D0,$F2,$A3,$2F,$F8,$83,$F1,$8F,$E9,$4F,$F8,$8A
	.dc.b	$FC,$04,$1F,$86,$FF,$35,$9D,$CE,$67,$72,$FC,$99,$5C,$2B,$C8,$78
	.dc.b	$1E,$35,$C4,$FF,$65,$3B,$8F,$31,$EA,$F0,$B9,$DF,$C2,$8F,$6F,$8D
	.dc.b	$87,$A9,$F1,$7F,$12,$F6,$4F,$F8,$A7,$E1,$7E,$37,$E0,$BF,$E3,$3F
	.dc.b	$13,$FD,$0F,$FA,$6F,$1C,$3F,$A2,$70,$7F,$A4,$7E,$6F,$FC,$A3,$E4
	.dc.b	$BC,$FF,$81,$9F,$3A,$D6,$3F,$E9,$0C,$39,$67,$DB,$1E,$3E,$87,$8A
	.dc.b	$7F,$C9,$B3,$9A,$64,$FA,$C7,$C3,$72,$CB,$AC,$E7,$E3,$B1,$BF,$A6
	.dc.b	$3F,$C0,$9F,$9D,$E3,$5F,$3F,$CC,$FF,$A3,$F9,$7E,$57,$EF,$0F,$D3
	.dc.b	$5F,$8C,$BF,$2E,$FE,$F3,$FA,$4F,$F8,$0F,$EF,$BE,$15,$EE,$B5,$F3
	.dc.b	$CA,$FF,$04,$7D,$56,$B1,$E7,$8F,$2F,$0F,$DD,$CE,$CB,$94,$F6,$F9
	.dc.b	$9C,$C6,$32,$E9,$F4,$AF,$D2,$3A,$6B,$E0,$D5,$3F,$E7,$FD,$FD,$7A
	.dc.b	$F8,$30,$F2,$7E,$07,$FE,$6A,$9C,$E1,$C6,$3F,$3D,$1F,$8B,$FB,$1F
	.dc.b	$E5,$FF,$9B,$BF,$1E,$7E,$2F,$FE,$0F,$F8,$CC,$5E,$FE,$21,$7B,$AD
	.dc.b	$7C,$5F,$E1,$F1,$E7,$8F,$2F,$A4,$71,$FC,$29,$CD,$FF,$86,$FB,$B1
	.dc.b	$97,$4C,$38,$7E,$6C,$FC,$67,$A4,$E6,$FF,$C4,$52,$FD,$0B,$3A,$F8
	.dc.b	$F4,$7C,$43,$F4,$4C,$E7,$E3,$8F,$E5,$DC,$1F,$AB,$F4,$3F,$5D,$E5
	.dc.b	$FE,$6B,$B7,$F7,$D1,$FB,$98,$FD,$C4,$FF,$06,$F4,$BF,$82,$D7,$E9
	.dc.b	$6B,$F6,$DF,$89,$FF,$07,$7F,$65,$1F,$C4,$BE,$E3,$F7,$F1,$88,$FE
	.dc.b	$2D,$F5,$38,$7F,$6A,$7F,$A6,$FD,$2E,$4F,$E9,$3C,$BF,$CF,$F9,$BF
	.dc.b	$C8,$F3,$FE,$2F,$E7,$FE,$9F,$CC,$FD,$37,$D8,$73,$C3,$F6,$FE,$77
	.dc.b	$E8,$FD,$2E,$71,$8F,$3B,$F6,$2E,$F0,$E5,$97,$EB,$B2,$EB,$9F,$83
	.dc.b	$B7,$4F,$8B,$F1,$4B,$9F,$EB,$ED,$F9,$27,$87,$CB,$FC,$D7,$95,$F9
	.dc.b	$6F,$7F,$1E,$57,$E5,$63,$F3,$3E,$EE,$3C,$AF,$CE,$78,$FE,$6C,$38
	.dc.b	$F9,$F7,$0E,$1E,$6F,$E8,$F8,$7F,$9B,$D7,$68,$FD,$17,$57,$FB,$0E
	.dc.b	$3C,$3E,$63,$F1,$FE,$DE,$87,$EF,$21,$FA,$27,$C7,$F3,$7F,$27,$E3
	.dc.b	$BC,$A7,$EA,$39,$F3,$7F,$21,$39,$E0,$C7,$97,$F8,$FE,$2F,$9B,$8F
	.dc.b	$2E,$CB,$87,$0C,$BF,$35,$C5,$FD,$3A,$BB,$C7,$E6,$63,$13,$DB,$F2
	.dc.b	$57,$AB,$DB,$CC,$FD,$54,$7F,$AD,$6B,$FA,$BF,$27,$F4,$53,$FD,$C6
	.dc.b	$E1,$FE,$91,$F3,$AF,$D0,$BF,$B2,$79,$9F,$98,$AD,$57,$F5,$9F,$33
	.dc.b	$FD,$FB,$5A,$F1,$7F,$B2,$F9,$9C,$5F,$CC,$1F,$EE,$FF,$33,$0F,$E6
	.dc.b	$00,$F3,$39,$FE,$0F,$FE,$F5,$C7,$CD,$F8,$37,$8E,$31,$1E,$6C,$7F
	.dc.b	$DF,$B7,$FB,$0F,$1F,$FB,$AC,$62,$3D,$D9,$F9,$BE,$D7,$C7,$1E,$0D
	.dc.b	$3D,$D1,$88,$FF,$67,$57,$FA,$D5,$F3,$1E,$53,$67,$EB,$6F,$F3,$5E
	.dc.b	$26,$23,$F4,$DE,$F7,$F5,$2F,$27,$FD,$73,$7C,$E7,$F5,$A7,$F8,$39
	.dc.b	$C7,$99,$FE,$0F,$CF,$D3,$97,$F6,$3C,$BF,$65,$7D,$6B,$F7,$1E,$FF
	.dc.b	$AE,$3C,$5E,$77,$F6,$CC,$BF,$AF,$E5,$D3,$CE,$FD,$93,$71,$FD,$87
	.dc.b	$9D,$CE,$7F,$82,$F9,$DE,$9F,$27,$F4,$9E,$75,$F3,$3F,$53,$C5,$71
	.dc.b	$F2,$EB,$F6,$2F,$FB,$17,$0E,$23,$FA,$B5,$FE,$DF,$C3,$F3,$46,$3C
	.dc.b	$DF,$EA,$5E,$36,$30,$FD,$97,$B7,$DE,$7F,$AB,$79,$CF,$89,$C7,$9D
	.dc.b	$FB,$1F,$3E,$FB,$9F,$D5,$BC,$FF,$A1,$E5,$FE,$93,$CF,$BE,$5F,$EA
	.dc.b	$76,$CF,$CE,$FF,$8D,$3A,$BE,$95,$FF,$A1,$B9,$3D,$0F,$EE,$B8,$FC
	.dc.b	$EF,$EC,$F8,$F9,$F9,$FF,$D1,$7C,$EF,$E9,$98,$72,$CF,$F6,$DE,$CE
	.dc.b	$BC,$EF,$D7,$F9,$F7,$DA,$FF,$78,$7A,$0F,$7B,$F0,$DE,$57,$E7,$32
	.dc.b	$FC,$23,$FB,$3F,$47,$2B,$87,$F3,$7F,$84,$7E,$57,$F3,$DF,$DA,$3F
	.dc.b	$97,$FD,$DA,$F1,$7E,$5B,$D1,$FD,$9C,$FE,$27,$CA,$FC,$27,$EB,$BA
	.dc.b	$AF,$CA,$FB,$9F,$E2,$5F,$D8,$71,$E7,$FF,$8E,$A7,$F2,$98,$7F,$8C
	.dc.b	$FE,$07,$C3,$1F,$8E,$F4,$BF,$C6,$9C,$2A,$ED,$1F,$E3,$0F,$47,$C3
	.dc.b	$5B,$CB,$FC,$65,$F0,$DD,$BD,$0F,$C9,$E1,$C2,$7F,$C6,$DF,$A3,$E7
	.dc.b	$CE,$FE,$89,$FA,$17,$9F,$CE,$FE,$6D,$F9,$C5,$F4,$3F,$9D,$F8,$D7
	.dc.b	$87,$CB,$B7,$E4,$31,$B8,$7F,$4A,$F1,$BC,$73,$F9,$1F,$33,$FC,$58
	.dc.b	$F3,$F3,$FF,$1F,$E8,$59,$FF,$1C,$FA,$1F,$91,$E6,$F7,$5F,$E7,$F5
	.dc.b	$FE,$97,$FB,$DE,$FC,$FF,$F1,$7F,$E5,$9F,$37,$CD,$FE,$6B,$F1,$57
	.dc.b	$CC,$FE,$75,$EC,$5F,$3F,$CF,$4E,$5F,$CA,$75,$7F,$22,$D5,$3F,$93
	.dc.b	$78,$79,$BF,$E2,$3F,$80,$F5,$CF,$F4,$2E,$8C,$F9,$BF,$E1,$CF,$3E
	.dc.b	$FC,$0B,$CF,$FE,$75,$CB,$EA,$CF,$AF,$07,$F4,$7F,$80,$FC,$27,$9D
	.dc.b	$FD,$0F,$F3,$BA,$FF,$A2,$7E,$17,$FA,$17,$C1,$FE,$47,$1F,$F4,$4E
	.dc.b	$30,$FF,$54,$79,$FF,$D1,$F7,$FE,$6B,$87,$FA,$67,$6C,$57,$F8,$92
	.dc.b	$3F,$D3,$93,$FD,$36,$7F,$98,$7B,$76,$BF,$97,$9F,$CA,$23,$E2,$7E
	.dc.b	$A0,$F8,$E7,$E8,$CF,$C9,$5F,$A6,$CF,$CB,$DF,$4B,$CF,$26,$3C,$70
	.dc.b	$E1,$7B,$FC,$33,$53,$D7,$E2,$7F,$A4,$E9,$E0,$D3,$BE,$3F,$CD,$E3
	.dc.b	$18,$F2,$C5,$F9,$1F,$F6,$29,$CE,$BE,$49,$7E,$4F,$9C,$3F,$AF,$E7
	.dc.b	$E3,$AD,$6C,$FD,$FF,$C7,$A7,$BA,$73,$F8,$7F,$AF,$F6,$D7,$E4,$7F
	.dc.b	$E6,$EF,$A3,$FF,$61,$FA,$5F,$EC,$AF,$9C,$D4,$E7,$BB,$FD,$11,$F2
	.dc.b	$EF,$9D,$CF,$89,$EE,$E1,$FE,$E3,$9F,$4F,$17,$10,$8E,$F3,$FE,$60
	.dc.b	$FA,$EF,$CB,$78,$1F,$E6,$9A,$E3,$9F,$7D,$7F,$8C,$47,$6D,$BC,$BA
	.dc.b	$7F,$AD,$2F,$84,$F5,$BE,$12,$FC,$D7,$EC,$7C,$0F,$F9,$33,$E7,$AF
	.dc.b	$BF,$FE,$5A,$F6,$AF,$B7,$E6,$AF,$F7,$B7,$E2,$F6,$BF,$D7,$BE,$4F
	.dc.b	$CF,$D3,$F8,$FF,$C1,$5B,$FF,$2B,$FC,$17,$2F,$86,$FE,$D1,$E8,$DF
	.dc.b	$84,$FE,$F1,$F8,$E5,$F8,$5F,$F6,$57,$EA,$1F,$8B,$FD,$4B,$F7,$1F
	.dc.b	$95,$3F,$0C,$FD,$77,$DD,$3F,$11,$7E,$23,$BF,$79,$7B,$F6,$D7,$BF
	.dc.b	$D0,$75,$E8,$FD,$0C,$FA,$F2,$FE,$3C,$FF,$12,$BE,$0D,$3B,$FD,$17
	.dc.b	$B6,$31,$8F,$4F,$B6,$C6,$9E,$0B,$F2,$EC,$F3,$FC,$9B,$78,$F6,$F4
	.dc.b	$7E,$75,$FA,$7F,$85,$7E,$89,$FE,$33,$FC,$3B,$F5,$3E,$9B,$F0,$17
	.dc.b	$E1,$DB,$AD,$7A,$DF,$99,$97,$EA,$59,$F9,$8F,$C5,$E5,$D3,$F4,$CB
	.dc.b	$87,$0C,$9E,$3F,$D7,$AF,$0D,$B4,$F0,$FE,$2F,$9C,$38,$66,$FD,$8F
	.dc.b	$B5,$79,$A1,$5F,$C0,$7D,$1F,$F3,$F7,$C0,$5F,$92,$FF,$41,$FC,$0D
	.dc.b	$F9,$6F,$F0,$7F,$E8,$57,$E5,$BF,$A0,$F9,$1F,$C2,$7F,$22,$7E,$50
	.dc.b	$FC,$35,$F9,$A3,$E7,$1A,$FF,$63,$FD,$1F,$85,$EE,$B5,$3F,$D9,$FF
	.dc.b	$43,$CF,$67,$EB,$B2,$EB,$95,$C3,$86,$5D,$3E,$4B,$F3,$55,$AC,$B9
	.dc.b	$FE,$0B,$8B,$DE,$7F,$DC,$FF,$6A,$F0,$E5,$E6,$CB,$CB,$F3,$2F,$BE
	.dc.b	$FE,$49,$F1,$EF,$BF,$FF,$71,$7E,$9C,$FC,$59,$F9,$C3,$F0,$EF,$E3
	.dc.b	$3F,$11,$FB,$8C,$38,$7C,$AB,$0E,$17,$BF,$67,$AE,$5F,$BE,$F5,$FF
	.dc.b	$83,$DF,$F5,$DC,$3C,$1F,$BD,$FF,$0E,$79,$E3,$CB,$C2,$FD,$C4,$63
	.dc.b	$5F,$1F,$9B,$8C,$BC,$F1,$E6,$DB,$F9,$8E,$DE,$59,$CE,$DE,$9F,$CD
	.dc.b	$BF,$A1,$7C,$87,$FA,$A3,$0B,$F3,$FF,$EA,$DF,$D2,$EF,$E4,$1F,$C8
	.dc.b	$29,$2F,$A4,$C3,$FE,$BA,$F8,$CF,$E1,$93,$FE,$45,$D3,$F8,$DB,$FD
	.dc.b	$CD,$B3,$FE,$44,$6D,$7F,$92,$7E,$3F,$C2,$FF,$92,$F0,$FE,$AD,$B3
	.dc.b	$F4,$EF,$D1,$FD,$73,$E1,$DE,$BB,$F1,$C7,$E1,$BC,$2F,$75,$A9,$7E
	.dc.b	$21,$9F,$67,$F9,$DE,$5D,$7E,$4A,$E1,$C3,$27,$E3,$FF,$C9,$EB,$C5
	.dc.b	$6D,$3C,$54,$F8,$5D,$61,$C7,$37,$8D,$DE,$F7,$0F,$B3,$FE,$E5,$A3
	.dc.b	$E2,$1E,$6B,$F2,$E7,$9E,$FD,$FD,$FA,$F3,$E7,$38,$E1,$C2,$F7,$5A
	.dc.b	$F7,$33,$8F,$F5,$8C,$BB,$7E,$61,$70,$E1,$9F,$5F,$CE,$FF,$A2,$57
	.dc.b	$8E,$FC,$53,$E1,$F9,$0C,$DF,$0C,$BC,$78,$F6,$CF,$C3,$AB,$F5,$6F
	.dc.b	$F0,$77,$D6,$FF,$BA,$FC,$EA,$8F,$F0,$FF,$83,$FD,$81,$C1,$FE,$E9
	.dc.b	$F3,$FF,$DD,$DE,$7F,$F8,$8A,$4F,$FB,$4F,$F0,$DC,$CF,$F9,$F4,$A7
	.dc.b	$F8,$73,$EA,$7F,$C0,$DE,$7F,$F8,$87,$EB,$33,$EE,$7F,$83,$63,$FC
	.dc.b	$3D,$8F,$4B,$FF,$08,$DF,$FB,$95,$FE,$F7,$8F,$F8,$71,$FF,$6A,$3F
	.dc.b	$E1,$AA,$FF,$16,$F1,$7F,$BF,$A3,$11,$FE,$1C,$9E,$F5,$E2,$8F,$E9
	.dc.b	$DF,$17,$8F,$84,$FF,$13,$7C,$83,$EA,$7D,$3F,$4D,$FE,$0C,$FA,$5F
	.dc.b	$F0,$2F,$D4,$E3,$E9,$BF,$C0,$7F,$51,$7E,$A3,$FD,$F7,$BF,$F9,$97
	.dc.b	$0F,$F0,$B6,$FE,$63,$FC,$23,$3F,$E1,$57,$CC,$63,$1F,$F1,$2F,$DE
	.dc.b	$62,$3F,$C2,$12,$9F,$F0,$AF,$D0,$77,$7F,$C2,$FF,$53,$F3,$BE,$97
	.dc.b	$FC,$25,$CB,$F2,$70,$FF,$54,$FD,$A3,$DA,$D7,$D4,$FF,$5A,$9F,$20
	.dc.b	$BF,$F0,$AB,$85,$1F,$E1,$68,$E3,$E8,$BE,$99,$F2,$1F,$EA,$B9,$CC
	.dc.b	$7F,$BF,$7E,$9E,$CB,$66,$6F,$A7,$CC,$63,$D9,$B2,$FA,$7F,$C4,$F9
	.dc.b	$5F,$E0,$8F,$A0,$FF,$0C,$7C,$56,$3E,$93,$FC,$11,$F4,$BF,$E1,$6F
	.dc.b	$A4,$7D,$3E,$7C,$6F,$A7,$C5,$FE,$14,$BF,$F0,$EB,$F0,$32,$E9,$F4
	.dc.b	$FF,$3D,$ED,$EE,$DA,$EB,$A7,$F8,$9F,$C4,$F0,$7D,$37,$F8,$4F,$47
	.dc.b	$D3,$7E,$7B,$E9,$7F,$C0,$7F,$45,$FD,$F7,$E4,$71,$F4,$9F,$E0,$2F
	.dc.b	$A6,$BF,$4D,$E4,$E1,$61,$FE,$0F,$7E,$9B,$CB,$D1,$FF,$37,$77,$FE
	.dc.b	$0E,$31,$E7,$FF,$82,$2B,$9E,$7F,$E0,$BF,$A6,$B8,$7F,$87,$7E,$5B
	.dc.b	$F6,$3F,$03,$FE,$0C,$F1,$FF,$08,$BE,$97,$FC,$1B,$18,$5F,$4B,$FE
	.dc.b	$0E,$8C,$2F,$A4,$FF,$07,$C7,$E3,$7E,$0B,$FB,$F7,$CE,$FF,$78,$FA
	.dc.b	$4F,$DF,$C6,$3E,$83,$FC,$15,$5B,$FA,$1F,$F4,$EF,$07,$E4,$2B,$FC
	.dc.b	$16,$D8,$FF,$02,$FC,$37,$E8,$67,$97,$D1,$59,$FF,$04,$FD,$0F,$F3
	.dc.b	$AC,$3A,$7D,$17,$F8,$27,$2F,$F7,$F7,$D1,$7F,$10,$F8,$0F,$EE,$9F
	.dc.b	$41,$FE,$05,$FA,$DC,$7D,$17,$F7,$0F,$97,$FF,$30,$E5,$FA,$7B,$FF
	.dc.b	$03,$F0,$7E,$91,$FF,$00,$B6,$BF,$BB,$F3,$FE,$D6,$7A,$F8,$36,$7A
	.dc.b	$79,$7D,$BE,$27,$FC,$21,$F4,$3F,$E0,$1F,$A2,$FE,$47,$F4,$3F,$DF
	.dc.b	$74,$FD,$B3,$FD,$EB,$E7,$DF,$68,$F8,$27,$DA,$3F,$00,$7D,$2F,$2F
	.dc.b	$A5,$7D,$27,$E2,$F2,$E5,$C9,$FB,$C7,$FD,$E9,$A7,$6A,$FF,$03,$67
	.dc.b	$FE,$F5,$AE,$5A,$78,$34,$EF,$9B,$EC,$F3,$7E,$3D,$7C,$5C,$18,$BD
	.dc.b	$FD,$35,$9F,$36,$DF,$D0,$39,$FF,$BA,$FD,$07,$F6,$EF,$A6,$C7,$D1
	.dc.b	$FF,$73,$FA,$6B,$E8,$FE,$8E,$FF,$C0,$1C,$3F,$C5,$1E,$3F,$07,$FE
	.dc.b	$0C,$FA,$6F,$E4,$91,$8F,$A6,$FE,$FD,$5C,$B3,$FE,$F7,$F4,$D7,$0F
	.dc.b	$F0,$77,$D2,$FF,$52,$FA,$3F,$ED,$5F,$43,$FD,$93,$E9,$71,$F0,$9F
	.dc.b	$DA,$7E,$89,$EB,$63,$D0,$FD,$B6,$1F,$DD,$B5,$FE,$4E,$F2,$F4,$7F
	.dc.b	$BB,$46,$F2,$FE,$FF,$E2,$FF,$39,$AD,$7D,$27,$F7,$4F,$0B,$F9,$A4
	.dc.b	$63,$3F,$EE,$91,$FC,$DF,$E8,$BF,$BA,$35,$FE,$E2,$7F,$72,$F0,$3F
	.dc.b	$B5,$B8,$3F,$BC,$7C,$C3,$35,$3F,$76,$C4,$7F,$86,$0A,$63,$F7,$4D
	.dc.b	$C6,$B6,$CC,$F0,$FC,$1B,$2F,$F7,$BB,$E8,$FF,$74,$FA,$2F,$EC,$FF
	.dc.b	$49,$A8,$FE,$FD,$1F,$DA,$BE,$D7,$1F,$00,$FA,$5F,$EF,$55,$E7,$C3
	.dc.b	$9B,$C0,$F0,$B8,$9F,$F0,$27,$A6,$E4,$E1,$AF,$F7,$CF,$56,$E1,$E3
	.dc.b	$FB,$55,$CF,$8D,$7E,$0B,$E2,$D3,$FC,$01,$0F,$C1,$7D,$1F,$47,$FD
	.dc.b	$F7,$ED,$FF,$03,$E7,$7E,$0B,$83,$CB,$39,$EF,$4B,$F7,$8E,$3F,$51
	.dc.b	$FB,$7D,$FD,$3C,$3F,$E0,$18,$7A,$FF,$32,$77,$E9,$7F,$80,$5B,$F4
	.dc.b	$DF,$E0,$06,$EB,$E2,$F9,$EB,$A7,$83,$E9,$BE,$77,$D2,$FF,$74,$F4
	.dc.b	$7F,$B9,$7D,$17,$F6,$DF,$A1,$7E,$23,$AF,$A6,$FF,$6B,$FD,$37,$F7
	.dc.b	$6C,$FF,$B8,$FD,$16,$71,$F2,$73,$FD,$2C,$FF,$C0,$3B,$FE,$1A,$31
	.dc.b	$8F,$F8,$05,$B5,$E0,$C1,$F6,$FF,$4A,$FF,$C1,$FF,$4F,$D7,$3F,$06
	.dc.b	$BF,$DF,$F1,$E3,$18,$FD,$9F,$5B,$FA,$7F,$41,$FC,$57,$A7,$FD,$FF
	.dc.b	$E8,$3F,$9A,$F5,$7E,$6A,$31,$F4,$DA,$FA,$4F,$C2,$6D,$F4,$0F,$EF
	.dc.b	$DF,$47,$EC,$BE,$13,$FD,$EB,$E9,$BC,$D8,$7F,$7E,$BE,$57,$FD,$EF
	.dc.b	$F5,$6B,$F7,$4E,$8F,$F0,$67,$E5,$FF,$9D,$FA,$3F,$EE,$3F,$76,$FC
	.dc.b	$B7,$1F,$51,$FE,$FE,$F8,$6B,$F4,$FE,$1D,$7F,$69,$87,$F7,$D7,$E6
	.dc.b	$70,$FE,$D2,$3F,$BD,$F6,$FC,$97,$FB,$EF,$CC,$5C,$58,$3E,$9B,$FB
	.dc.b	$F7,$5F,$EA,$2F,$C7,$7F,$B1,$F1,$3F,$BE,$7D,$27,$F7,$7F,$C0,$DF
	.dc.b	$8B,$3F,$06,$FF,$4C,$E3,$FB,$0B,$DF,$D3,$AF,$74,$FD,$BB,$A7,$D4
	.dc.b	$FF,$81,$2B,$58,$F3,$FD,$13,$39,$F7,$8E,$DF,$87,$FF,$03,$F8,$CF
	.dc.b	$F0,$17,$D3,$BE,$AB,$1F,$53,$FD,$EB,$F0,$1B,$F0,$7F,$1F,$AB,$FD
	.dc.b	$5F,$2F,$AA,$C6,$3F,$E0,$BF,$89,$FC,$73,$FE,$04,$AE,$73,$FE,$01
	.dc.b	$CB,$FD,$6D,$CB,$FC,$32,$BC,$3F,$B6,$7F,$15,$F8,$3F,$F0,$1F,$7B
	.dc.b	$F7,$0C,$7D,$47,$F8,$0F,$EC,$AF,$D1,$78,$EF,$FC,$03,$AF,$D8,$3F
	.dc.b	$97,$3F,$E0,$CE,$DF,$D0,$3F,$E0,$28,$E3,$F8,$13,$EA,$BF,$E0,$07
	.dc.b	$C6,$61,$FF,$01,$6F,$FC,$83,$E0,$BF,$C0,$1F,$4B,$FD,$B7,$E9,$9E
	.dc.b	$47,$F2,$B9,$CF,$C2,$7F,$74,$F3,$31,$F5,$1F,$E0,$0D,$7F,$40,$7F
	.dc.b	$0E,$5F,$51,$F2,$EF,$FC,$03,$E6,$FF,$42,$79,$FD,$47,$F9,$C6,$F9
	.dc.b	$D7,$F8,$06,$3F,$C1,$DF,$33,$8D,$7C,$9C,$97,$4F,$F0,$2F,$95,$6F
	.dc.b	$FC,$29,$F3,$1F,$D0,$7A,$BF,$BA,$7E,$05,$FF,$00,$38,$FF,$9E,$C6
	.dc.b	$3E,$9B,$5F,$4F,$FE,$91,$DF,$E6,$09,$FF,$00,$B8,$7F,$BF,$7D,$3F
	.dc.b	$F0,$D7,$FC,$01,$F6,$D6,$BF,$DF,$9F,$13,$D3,$9D,$F4,$FF,$EF,$CF
	.dc.b	$CD,$7F,$AA,$7D,$17,$F6,$FF,$C1,$2F,$C0,$FF,$7F,$FC,$BA,$F9,$1F
	.dc.b	$91,$AF,$EF,$DB,$F9,$1F,$EF,$3F,$82,$FF,$1D,$7F,$BD,$F8,$57,$0F
	.dc.b	$F0,$47,$E0,$BC,$BE,$A3,$FD,$DF,$F8,$2F,$F6,$3F,$A7,$FE,$EC,$7F
	.dc.b	$BB,$7F,$21,$C3,$8F,$CD,$EB,$FE,$ED,$E9,$75,$FF,$73,$DB,$FA,$47
	.dc.b	$D4,$FF,$7C,$8C,$47,$F7,$6F,$A9,$C3,$FD,$D6,$3F,$DB,$FF,$44,$FA
	.dc.b	$9C,$CF,$F7,$4F,$A6,$FE,$CB,$E7,$A9,$F5,$1A,$F6,$BF,$BB,$7D,$C7
	.dc.b	$F4,$4C,$BF,$DD,$D8,$FF,$43,$7F,$DD,$7C,$BF,$DE,$5F,$EE,$51,$C7
	.dc.b	$E9,$FF,$B6,$FD,$45,$FA,$AF,$F7,$5F,$D4,$F8,$BE,$A7,$FB,$57,$D4
	.dc.b	$FF,$6D,$C7,$FB,$37,$9B,$FE,$DD,$FA,$6F,$EE,$3F,$53,$BF,$A9,$FE
	.dc.b	$D9,$AF,$84,$FF,$0F,$63,$FE,$EC,$70,$7F,$B9,$BE,$A5,$6F,$AD,$FF
	.dc.b	$10,$F5,$7C,$CA,$FF,$77,$4F,$F8,$FB,$3F,$54,$63,$E4,$7F,$BE,$FD
	.dc.b	$3F,$87,$5F,$24,$7F,$71,$C3,$95,$7F,$BB,$3E,$A7,$E6,$FD,$4F,$F7
	.dc.b	$3F,$A6,$FE,$BF,$F3,$98,$F4,$3F,$3D,$7E,$2D,$3F,$C6,$77,$FD,$BA
	.dc.b	$30,$7F,$8E,$5E,$5F,$4B,$FE,$E7,$FC,$27,$CB,$3F,$EF,$5B,$FE,$F7
	.dc.b	$3F,$EE,$FC,$FC,$77,$C3,$6F,$F7,$6F,$D4,$F8,$7D,$3F,$EE,$DF,$53
	.dc.b	$FD,$C7,$4F,$0E,$5C,$AF,$FD,$E7,$F5,$3B,$DF,$FB,$BF,$D4,$FE,$23
	.dc.b	$CF,$FE,$E5,$F8,$05,$FA,$8F,$F7,$1F,$D5,$7F,$BB,$B8,$3E,$1D,$27
	.dc.b	$FB,$9F,$17,$D4,$8C,$7D,$47,$E2,$E3,$FB,$AF,$D3,$FF,$6D,$FC,$01
	.dc.b	$F5,$5E,$EB,$FE,$E9,$8F,$4C,$7B,$FD,$57,$87,$2F,$F7,$87,$D5,$7E
	.dc.b	$73,$E9,$BF,$B9,$FD,$55,$FA,$AF,$EE,$58,$FD,$69,$47,$F6,$AD,$BF
	.dc.b	$C9,$BC,$DF,$DC,$61,$F6,$9E,$59,$FF,$79,$C7,$F7,$0F,$AA,$7A,$1F
	.dc.b	$EF,$29,$ED,$F5,$7F,$DC,$E1,$F7,$FF,$EF,$3C,$FC,$18,$71,$9C,$FC
	.dc.b	$B3,$EF,$F8,$E5,$D3,$04,$BE,$6F,$F6,$FE,$3F,$F7,$4F,$91,$FE,$E7
	.dc.b	$F4,$58,$FA,$BF,$EE,$51,$FD,$DB,$EB,$31,$F5,$5F,$DC,$7E,$B5,$F5
	.dc.b	$98,$FA,$DF,$EE,$F0,$FA,$CF,$E0,$7F,$1D,$FD,$D3,$EB,$2F,$D5,$FF
	.dc.b	$73,$E6,$FC,$EC,$A3,$FD,$EF,$D1,$FA,$E8,$FF,$7B,$B6,$3F,$DE,$EF
	.dc.b	$2F,$AB,$FF,$79,$7D,$F7,$E6,$2B,$FC,$03,$F1,$F9,$F9,$47,$DF,$BE
	.dc.b	$19,$F5,$AF,$AD,$FE,$2F,$F1,$7F,$DD,$BE,$B6,$FD,$6F,$F7,$5E,$0F
	.dc.b	$D8,$CF,$FB,$DB,$7F,$E0,$2F,$F7,$67,$F5,$67,$FA,$0A,$7F,$DE,$EF
	.dc.b	$2F,$A3,$FF,$79,$FD,$F3,$EF,$BF,$5B,$7F,$E0,$4F,$8E,$D7,$D7,$7F
	.dc.b	$79,$AF,$EA,$9F,$7D,$FC,$AB,$EA,$FF,$B4,$FD,$77,$F7,$6D,$7F,$81
	.dc.b	$C6,$3C,$5D,$7D,$77,$F7,$5D,$BF,$8A,$CF,$FB,$D1,$B1,$FE,$F2,$F8
	.dc.b	$8F,$F7,$EF,$E0,$7F,$BE,$AF,$EE,$BF,$81,$DA,$D7,$D6,$7F,$74,$CB
	.dc.b	$BC,$63,$37,$17,$83,$EB,$FF,$BD,$47,$F8,$EE,$B7,$18,$F4,$FF,$A1
	.dc.b	$FD,$57,$F7,$6F,$BE,$BF,$55,$FE,$F3,$BF,$E3,$B2,$8F,$E3,$3C,$1F
	.dc.b	$C4,$A3,$FB,$AB,$F3,$9F,$EE,$AF,$2F,$89,$FE,$E3,$F7,$BF,$AC,$FA
	.dc.b	$CE,$17,$FD,$DB,$0E,$AF,$F8,$7B,$EB,$3F,$BB,$47,$F2,$CF,$A9,$FE
	.dc.b	$EC,$DF,$82,$FE,$EC,$DF,$1B,$F6,$D3,$FD,$DB,$3F,$BE,$FF,$76,$7F
	.dc.b	$AF,$BF,$DD,$9B,$5F,$DD,$AB,$97,$8D,$FD,$D9,$FE,$77,$F4,$BF,$DD
	.dc.b	$87,$D3,$FF,$68,$71,$F4,$F5,$1F,$C1,$CF,$EF,$BC,$5F,$ED,$BF,$53
	.dc.b	$FB,$AE,$5F,$EE,$EF,$A7,$FF,$79,$C6,$23,$FD,$D9,$8F,$F9,$BD,$FF
	.dc.b	$7C,$7A,$FF,$EF,$9F,$C5,$BF,$D6,$75,$B9,$FE,$31,$5B,$7F,$BD,$78
	.dc.b	$BF,$ED,$DF,$A7,$EB,$1F,$EE,$18,$FF,$70,$7D,$2E,$23,$F8,$CF,$D2
	.dc.b	$FF,$BC,$1C,$7C,$0F,$FB,$CF,$E9,$3B,$C7,$1B,$FF,$55,$E3,$DE,$BF
	.dc.b	$DE,$3F,$77,$ED,$AD,$7D,$3F,$FB,$B2,$B5,$8F,$FB,$B3,$E9,$BC,$1F
	.dc.b	$47,$FD,$FB,$E0,$AF,$D2,$3C,$BF,$ED,$7F,$4D,$FE,$E9,$C7,$D9,$6A
	.dc.b	$FE,$FF,$9F,$F7,$8E,$2F,$F7,$57,$97,$E6,$8F,$EF,$DF,$47,$FD,$F5
	.dc.b	$F2,$18,$CF,$FB,$F6,$3D,$30,$F0,$7D,$2F,$F8,$12,$7F,$BE,$CE,$74
	.dc.b	$F1,$FD,$DF,$C7,$FA,$1F,$ED,$5F,$41,$FE,$EA,$AF,$A1,$18,$FA,$4D
	.dc.b	$7D,$0F,$E9,$34,$BE,$27,$96,$BF,$B9,$7D,$27,$91,$FE,$E9,$F7,$76
	.dc.b	$FF,$B8,$64,$FA,$4E,$7E,$6F,$F7,$6F,$56,$FD,$17,$FB,$97,$6F,$0E
	.dc.b	$1C,$3E,$8B,$EE,$79,$7F,$DD,$27,$F0,$8F,$FB,$91,$BF,$3B,$FD,$CB
	.dc.b	$83,$E5,$1F,$EE,$40,$7F,$BF,$7A,$7F,$DC,$BF,$38,$FA,$0F,$EE,$9F
	.dc.b	$3F,$F3,$31,$FE,$E7,$5A,$9F,$EF,$1F,$42,$FA,$0F,$1E,$5F,$E0,$4F
	.dc.b	$A0,$FC,$87,$CF,$FF,$74,$D7,$F3,$0F,$FB,$A3,$D9,$CF,$CE,$FF,$73
	.dc.b	$DF,$C4,$7F,$B9,$9F,$EE,$6F,$98,$F4,$9F,$DD,$3E,$7B,$FB,$87,$CE
	.dc.b	$7F,$A9,$A3,$9C,$63,$E7,$3F,$07,$87,$FB,$DB,$E7,$B3,$F3,$DF,$DB
	.dc.b	$FE,$1F,$FB,$76,$9F,$AB,$F1,$3F,$B7,$7D,$D7,$FB,$97,$D6,$FD,$44
	.dc.b	$63,$E6,$BF,$DC,$55,$BF,$9F,$FE,$E3,$7F,$EE,$18,$C7,$CF,$7F,$71
	.dc.b	$F9,$CF,$D8,$7C,$0F,$F6,$D9,$FD,$C3,$FE,$E2,$FB,$AC,$FC,$E7,$F6
	.dc.b	$CE,$3F,$6C,$FF,$6C,$6C,$7F,$6C,$AE,$2F,$FB,$8F,$E6,$BF,$B4,$FC
	.dc.b	$CB,$E6,$FF,$43,$87,$F6,$EB,$DF,$CE,$7F,$B8,$FE,$6B,$F8,$17,$BD
	.dc.b	$FD,$AE,$7F,$85,$3F,$EE,$2F,$BA,$CF,$CD,$7F,$4A,$D7,$E8,$1F,$AC
	.dc.b	$9F,$ED,$8D,$8F,$ED,$75,$C9,$FE,$A3,$F3,$1F,$D9,$FE,$6B,$5F,$29
	.dc.b	$FD,$BB,$C1,$7C,$D7,$68,$C7,$CC,$FF,$15,$F7,$BF,$B5,$FA,$57,$E5
	.dc.b	$FF,$AE,$77,$DF,$C0,$3F,$80,$57,$F6,$9D,$FF,$06,$FF,$B6,$BE,$6B
	.dc.b	$F8,$0B,$FD,$A2,$39,$7C,$C7,$F6,$2F,$B9,$BF,$73,$FE,$DA,$E3,$FE
	.dc.b	$55,$7F,$DE,$BD,$1C,$FC,$0B,$E7,$3F,$DB,$75,$FC,$BF,$8F,$FB,$33
	.dc.b	$7E,$63,$FB,$35,$FE,$56,$7F,$B3,$70,$FF,$15,$7F,$B2,$EB,$FC,$51
	.dc.b	$FE,$CB,$7C,$67,$FB,$2F,$D2,$BE,$6B,$58,$FF,$65,$C7,$A3,$FD,$97
	.dc.b	$E6,$FF,$CA,$BF,$72,$FB,$7F,$E7,$9F,$0D,$FD,$9A,$FF,$4C,$BD,$BF
	.dc.b	$EC,$F1,$85,$E1,$FF,$29,$9F,$ED,$1B,$FF,$28,$7F,$B4,$6D,$FD,$35
	.dc.b	$FE,$D0,$DA,$FE,$D1,$7C,$ED,$F3,$9F,$DA,$63,$F6,$DF,$35,$FD,$A7
	.dc.b	$E7,$3F,$A8,$3F,$DA,$30,$FE,$B0,$FF,$68,$3F,$9D,$1F,$DA,$32,$47
	.dc.b	$F7,$07,$FB,$19,$FE,$E0,$70,$BD,$2F,$CA,$BF,$DA,$AB,$FD,$CD,$F3
	.dc.b	$38,$8E,$71,$FD,$A7,$D8,$FE,$3F,$7C,$BE,$5B,$FB,$07,$D9,$7F,$B8
	.dc.b	$BE,$5B,$FD,$CD,$39,$F9,$67,$CC,$2B,$5F,$33,$D3,$3F,$E5,$38,$FF
	.dc.b	$B9,$9F,$1E,$1C,$6F,$C4,$DB,$FF,$77,$F9,$1E,$2D,$3F,$DC,$F3,$9F
	.dc.b	$30,$7F,$BC,$FE,$CB,$B6,$3F,$DF,$A3,$FD,$ED,$F3,$38,$F9,$8F,$ED
	.dc.b	$97,$F1,$1F,$F7,$5F,$D9,$67,$E6,$3F,$B4,$E1,$DC,$FF,$73,$AF,$17
	.dc.b	$CB,$9F,$FB,$C3,$E6,$3F,$DC,$FF,$65,$FE,$E6,$8C,$7C,$D7,$FB,$97
	.dc.b	$0E,$9E,$47,$F7,$8F,$B0,$B8,$7F,$6B,$F9,$C7,$CE,$7A,$3E,$6B,$FB
	.dc.b	$4F,$D9,$5F,$96,$FE,$BF,$F3,$B7,$CD,$F0,$E9,$FE,$DF,$97,$8F,$88
	.dc.b	$FF,$6F,$4B,$E6,$1F,$65,$FE,$E3,$9E,$51,$8D,$9F,$3B,$EF,$9E,$33
	.dc.b	$9F,$9D,$FE,$DD,$D3,$F8,$AD,$BF,$BE,$FC,$E7,$2C,$99,$7F,$B8,$AB
	.dc.b	$5A,$FF,$6F,$7F,$23,$E6,$7F,$6F,$E6,$FE,$C5,$C0,$FB,$07,$CC,$FF
	.dc.b	$B8,$E7,$FA,$C5,$7F,$B8,$FC,$1F,$63,$FE,$E2,$E0,$FC,$3C,$F2,$BF
	.dc.b	$F7,$0C,$7E,$02,$7F,$DC,$33,$9B,$FE,$DD,$8F,$8B,$E6,$3F,$B7,$3F
	.dc.b	$A4,$F9,$4F,$ED,$D5,$FE,$DB,$F2,$BF,$B2,$FB,$CF,$63,$F1,$75,$FE
	.dc.b	$E2,$9F,$9C,$FF,$B9,$3D,$8B,$F2,$FF,$ED,$7F,$97,$FE,$D9,$97,$58
	.dc.b	$C7,$CB,$FC,$DB,$FE,$D9,$39,$C3,$BC,$FE,$63,$E5,$7F,$B4,$7A,$77
	.dc.b	$E5,$3F,$B3,$FC,$AF,$F6,$AD,$FF,$BE,$57,$F6,$DF,$AE,$D7,$CA,$FF
	.dc.b	$69,$CF,$F7,$FE,$DF,$F6,$9B,$FF,$01,$3F,$DA,$63,$3F,$27,$FD,$A7
	.dc.b	$9F,$F0,$13,$FD,$A5,$B1,$FD,$A6,$B8,$BF,$DA,$7E,$B3,$FD,$A1,$F2
	.dc.b	$5A,$F9,$27,$CA,$7F,$6C,$FA,$CF,$E3,$B9,$F6,$9F,$ED,$4F,$F1,$2F
	.dc.b	$5B,$FB,$56,$5F,$E1,$57,$FB,$54,$67,$E4,$BF,$B5,$69,$FD,$AA,$31
	.dc.b	$F8,$2F,$E7,$69,$5F,$DA,$AF,$93,$FD,$AB,$EF,$1F,$5B,$FE,$D5,$C7
	.dc.b	$F8,$1D,$FF,$B5,$71,$ED,$F2,$7F,$DA,$9F,$E4,$BE,$BF,$F6,$A6,$FC
	.dc.b	$97,$F6,$A6,$FC,$97,$EB,$67,$FB,$56,$1F,$A7,$7F,$B5,$3F,$C4,$1F
	.dc.b	$ED,$53,$CB,$E4,$7F,$B5,$37,$E4,$3F,$B5,$6D,$FE,$35,$AF,$ED,$53
	.dc.b	$9F,$67,$FB,$57,$D6,$BE,$AB,$F8,$DF,$8D,$FD,$AB,$0F,$EB,$BE,$B7
	.dc.b	$F6,$2F,$8E,$7D,$DE,$BE,$47,$FC,$D9,$9F,$F4,$0C,$3F,$B5,$EF,$9C
	.dc.b	$3F,$B4,$FD,$DF,$EF,$E7,$FB,$8F,$D7,$3E,$4F,$FD,$A1,$7B,$8C,$79
	.dc.b	$AD,$1E,$AE,$6F,$FB,$6F,$C9,$FF,$30,$F7,$BF,$DA,$3C,$1F,$D7,$FE
	.dc.b	$4F,$FA,$87,$DF,$BE,$FD,$F7,$BB,$F6,$3F,$B6,$EF,$FB,$C3,$FC,$DB
	.dc.b	$8F,$F6,$E7,$5E,$57,$F6,$8F,$41,$8B,$EF,$7F,$8C,$5F,$F7,$0D,$3C
	.dc.b	$15,$AC,$DF,$7B,$CB,$5F,$F6,$C5,$7F,$70,$F9,$7F,$ED,$87,$F6,$D7
	.dc.b	$FA,$4B,$FD,$C5,$D3,$FC,$B4,$FE,$E1,$D0,$FB,$3F,$F6,$76,$CF,$98
	.dc.b	$FE,$E7,$18,$FB,$CB,$3F,$DD,$27,$FA,$EC,$FF,$6D,$8C,$9F,$D5,$FE
	.dc.b	$65,$FB,$26,$3E,$63,$FB,$7F,$95,$FE,$D4,$CB,$C1,$7F,$DA,$BE,$6B
	.dc.b	$FD,$A9,$18,$8F,$F6,$BF,$CC,$F6,$7F,$DA,$9F,$0B,$FD,$EB,$3F,$ED
	.dc.b	$27,$FA,$17,$1E,$B8,$FF,$69,$C3,$94,$63,$37,$CB,$F9,$BE,$5F,$FB
	.dc.b	$4E,$1F,$D7,$BD,$C7,$C9,$7F,$B4,$E3,$FD,$21,$3F,$ED,$3F,$2F,$C5
	.dc.b	$AA,$3F,$DA,$DE,$66,$1D,$7C,$9F,$F6,$A9,$F3,$1F,$DC,$29,$1F,$ED
	.dc.b	$1F,$33,$33,$FE,$D7,$C5,$F2,$BE,$5F,$96,$FF,$6B,$FC,$B3,$5F,$2F
	.dc.b	$EA,$7F,$3B,$E4,$BF,$DA,$BF,$A7,$DF,$2B,$FD,$97,$F9,$CD,$F4,$7C
	.dc.b	$F5,$F9,$EF,$16,$D2,$7F,$B6,$7C,$9F,$F6,$BE,$9F,$A1,$7F,$ED,$6E
	.dc.b	$0E,$5F,$2D,$FD,$A7,$ED,$2D,$FF,$66,$F9,$77,$CB,$E3,$E5,$D1,$FC
	.dc.b	$63,$2F,$B5,$E7,$BE,$51,$F6,$7F,$EC,$DF,$98,$FE,$D1,$F2,$EF,$73
	.dc.b	$F1,$37,$BA,$D7,$CC,$7F,$6C,$FB,$6C,$E3,$FE,$DB,$E1,$F9,$37,$FD
	.dc.b	$AD,$FE,$5D,$E1,$FF,$6B,$FB,$4C,$D7,$83,$ED,$1A,$FF,$77,$FA,$4C
	.dc.b	$E9,$DB,$6F,$ED,$B8,$FF,$6F,$FB,$3F,$CD,$FC,$D7,$FB,$66,$3F,$B6
	.dc.b	$7D,$97,$FB,$27,$E0,$F3,$F3,$9F,$ED,$9D,$FF,$DA,$DC,$9F,$99,$C3
	.dc.b	$FB,$64,$7F,$91,$30,$E3,$E7,$3E,$77,$FB,$6C,$61,$7C,$E7,$E1,$2F
	.dc.b	$FB,$B6,$7F,$ED,$7F,$CB,$BB,$FC,$F7,$F7,$CF,$D6,$7F,$5F,$E4,$7F
	.dc.b	$B5,$7F,$62,$BF,$41,$FD,$BB,$F5,$5B,$F3,$EF,$C0,$1F,$56,$FB,$77
	.dc.b	$D2,$3E,$7B,$75,$A9,$CF,$27,$EF,$3E,$31,$F4,$7F,$DA,$7E,$D5,$F4
	.dc.b	$3F,$E9,$3E,$AF,$DC,$56,$B2,$FE,$D7,$F4,$8B,$DF,$A5,$CF,$1F,$EF
	.dc.b	$1D,$96,$FF,$5B,$87,$1F,$A3,$FF,$6E,$FC,$7F,$E9,$A3,$19,$7F,$78
	.dc.b	$BE,$15,$FD,$DF,$E3,$6C,$F7,$D5,$F6,$BF,$C5,$3E,$1B,$FD,$C1,$E0
	.dc.b	$FF,$6E,$FA,$0F,$F6,$A7,$DA,$E7,$E9,$3F,$BB,$E9,$F6,$AB,$FB,$9F
	.dc.b	$D2,$FF,$00,$C3,$FB,$BE,$0F,$B5,$7D,$37,$F7,$6F,$1B,$F7,$75,$A9
	.dc.b	$FF,$71,$63,$D6,$FF,$DC,$5F,$19,$67,$8F,$D3,$7F,$8D,$7C,$EF,$E5
	.dc.b	$3F,$18,$FA,$67,$D2,$BE,$95,$F1,$DF,$DB,$3F,$49,$DD,$6A,$73,$F6
	.dc.b	$4F,$A7,$FF,$35,$F8,$5F,$CA,$F2,$E9,$F5,$1F,$C6,$9F,$F0,$0F,$0F
	.dc.b	$2D,$3B,$69,$FE,$E1,$FB,$4F,$E2,$B5,$FE,$02,$E7,$FE,$FD,$AF,$7A
	.dc.b	$FE,$FB,$F6,$9C,$76,$FF,$76,$56,$B1,$E7,$B7,$8F,$F0,$7D,$61,$FE
	.dc.b	$04,$E9,$FE,$8F,$EC,$7F,$78,$FA,$6F,$F6,$EF,$E0,$58,$F8,$5F,$F7
	.dc.b	$07,$D4,$FF,$77,$FB,$9C,$7C,$7F,$F3,$2C,$3F,$C0,$7C,$DF,$D1,$DF
	.dc.b	$F7,$6E,$FF,$E0,$CF,$AB,$FE,$FF,$F0,$5F,$CB,$A7,$B6,$3F,$E0,$1F
	.dc.b	$3B,$9D,$7F,$81,$3E,$AF,$F3,$58,$7F,$70,$FA,$9F,$EF,$DE,$67,$F5
	.dc.b	$CF,$93,$FF,$76,$79,$DF,$DE,$67,$FC,$05,$F5,$59,$F8,$BF,$F0,$0F
	.dc.b	$7F,$F1,$23,$FC,$02,$FF,$78,$7F,$81,$18,$AE,$5F,$55,$FD,$0B,$0F
	.dc.b	$F7,$7E,$5E,$1F,$AA,$7E,$A9,$D6,$FC,$10,$FE,$01,$FE,$DF,$F5,$5F
	.dc.b	$E0,$03,$FD,$D9,$E2,$FF,$BA,$BC,$0F,$EF,$1C,$FF,$8A,$D7,$FB,$77
	.dc.b	$D3,$7F,$B6,$FC,$EF,$F6,$7F,$9B,$FD,$DA,$B5,$5F,$DE,$7E,$E7,$93
	.dc.b	$FD,$D7,$DB,$FF,$6D,$E3,$FD,$BF,$EA,$3F,$DC,$B3,$FE,$E0,$8C,$65
	.dc.b	$D7,$27,$D4,$7F,$BC,$27,$FD,$BE,$6B,$EE,$7F,$DE,$1F,$73,$DB,$0F
	.dc.b	$F0,$47,$D5,$5F,$AA,$FF,$6E,$DF,$F7,$5F,$A8,$7D,$CE,$7E,$A3,$FD
	.dc.b	$BB,$85,$F4,$BC,$AF,$F6,$FF,$AB,$FE,$E4,$EF,$CE,$FE,$E3,$F5,$3F
	.dc.b	$ED,$BE,$5F,$7E,$9F,$DE,$23,$FB,$76,$7F,$DC,$A7,$39,$7F,$72,$F9
	.dc.b	$0F,$EE,$31,$F1,$BE,$9B,$FB,$8B,$7C,$8F,$EE,$2D,$F3,$3C,$7A,$2B
	.dc.b	$FB,$93,$F1,$CF,$F7,$04,$7F,$72,$FA,$27,$E0,$1F,$1E,$73,$F4,$FF
	.dc.b	$ED,$DF,$B8,$C4,$7F,$B7,$3E,$9F,$FB,$97,$D4,$7C,$9C,$BF,$DB,$79
	.dc.b	$F8,$70,$FF,$70,$7C,$0F,$CD,$F2,$FF,$B6,$7D,$17,$F6,$2F,$A6,$FF
	.dc.b	$6B,$FA,$F9,$FA,$4F,$EC,$DF,$49,$FD,$D3,$E9,$B1,$E1,$7A,$B1,$FE
	.dc.b	$DD,$C1,$FD,$7B,$BF,$FB,$73,$C7,$E1,$BF,$BB,$FB,$5E,$6A,$EF,$8F
	.dc.b	$F7,$EF,$BF,$D6,$DE,$1F,$A6,$7D,$37,$E0,$7E,$93,$FA,$FF,$CC,$7F
	.dc.b	$65,$FB,$97,$E0,$D9,$FA,$8F,$F6,$A4,$7F,$B6,$FA,$3F,$2F,$7F,$ED
	.dc.b	$BE,$0F,$98,$1F,$EE,$2F,$FA,$C7,$87,$FB,$77,$7F,$CA,$9E,$FF,$51
	.dc.b	$FD,$C2,$BB,$7D,$37,$FB,$A3,$E9,$EF,$D4,$36,$F2,$7D,$47,$E5,$FC
	.dc.b	$6F,$EE,$1A,$FE,$A1,$7D,$3F,$F7,$18,$C3,$FE,$E0,$FA,$0E,$0F,$FB
	.dc.b	$8A,$FF,$41,$EF,$7F,$5C,$F6,$3F,$B5,$FD,$B6,$3E,$8B,$FB,$77,$CF
	.dc.b	$FF,$A5,$B2,$FD,$AD,$FF,$72,$C3,$F1,$AF,$FB,$7F,$83,$86,$7F,$DC
	.dc.b	$32,$7D,$17,$E2,$2B,$B7,$DC,$DC,$FA,$FD,$1F,$0F,$8B,$FF,$70,$7D
	.dc.b	$17,$EC,$FD,$EF,$EB,$5F,$41,$FE,$D2,$F9,$17,$E0,$B9,$FA,$2F,$F6
	.dc.b	$B7,$9E,$97,$D1,$E7,$E8,$BF,$43,$8F,$FB,$7B,$E6,$6E,$1C,$3E,$8F
	.dc.b	$F3,$73,$DB,$1F,$F6,$F7,$3F,$F1,$28,$C6,$4F,$8A,$E9,$F4,$7F,$DA
	.dc.b	$23,$FC,$97,$5F,$C6,$3E,$13,$FB,$47,$8D,$FD,$43,$E8,$9F,$47,$FC
	.dc.b	$36,$73,$F4,$1F,$D9,$A7,$FB,$2D,$7F,$5C,$F8,$FF,$DB,$D7,$F6,$4C
	.dc.b	$3A,$7D,$17,$F5,$FF,$A1,$7D,$1F,$EC,$AB,$FD,$CB,$87,$F5,$FC,$BA
	.dc.b	$7D,$1F,$7C,$38,$66,$E5,$C5,$F9,$7E,$8E,$E1,$E3,$E6,$FE,$45,$C1
	.dc.b	$FD,$5B,$E7,$9F,$45,$8F,$80,$FE,$C7,$0F,$A2,$C7,$A7,$FC,$3A,$FF
	.dc.b	$B1,$E7,$FD,$87,$E8,$DF,$46,$FB,$9F,$E1,$35,$FD,$8B,$E0,$2E,$7D
	.dc.b	$3E,$37,$9F,$D2,$7F,$63,$F4,$7F,$9A,$7D,$0F,$F6,$0F,$27,$FD,$C3
	.dc.b	$F3,$EF,$2B,$3F,$08,$F3,$BF,$AA,$FD,$1E,$1F,$E4,$38,$7F,$B8,$7C
	.dc.b	$1F,$E9,$AF,$3F,$43,$FD,$C3,$F4,$7F,$C7,$23,$1F,$49,$FD,$23,$D5
	.dc.b	$E5,$AF,$83,$E8,$EE,$1F,$EE,$3F,$A3,$64,$F2,$3F,$A6,$79,$9F,$D7
	.dc.b	$3E,$D7,$31,$88,$FC,$57,$D1,$E3,$3F,$EB,$80,$85,$BE,$8F,$F6,$83
	.dc.b	$26,$6B,$A7,$C6,$D8,$FE,$DB,$F4,$2C,$27,$C7,$E6,$7F,$DC,$9F,$43
	.dc.b	$E1,$7F,$D8,$1F,$69,$7E,$7B,$FA,$97,$C1,$FF,$B4,$8C,$7C,$E7,$F5
	.dc.b	$ED,$7C,$77,$FE,$D1,$87,$A3,$E4,$8E,$D7,$FD,$6F,$0F,$F4,$9E,$DF
	.dc.b	$ED,$DF,$B5,$CE,$1F,$D6,$FE,$79,$F6,$BE,$3F,$9A,$FE,$B7,$7F,$09
	.dc.b	$78,$7F,$EC,$7F,$9A,$FF,$61,$6D,$E2,$9C,$F9,$9E,$1D,$B9,$5B,$B7
	.dc.b	$17,$B8,$E5,$F3,$9F,$EC,$FC,$BB,$46,$3E,$7D,$F3,$DF,$1F,$3F,$EB
	.dc.b	$75,$AC,$FF,$B3,$7D,$97,$FB,$36,$3E,$BF,$8B,$FD,$3B,$E7,$BF,$AD
	.dc.b	$E9,$F1,$A3,$1F,$32,$EA,$FA,$11,$FE,$D3,$ED,$F2,$E1,$CB,$E7,$BF
	.dc.b	$DA,$BC,$5F,$2F,$1E,$51,$FD,$93,$27,$D9,$7D,$BC,$7C,$1E,$7D,$AD
	.dc.b	$68,$F9,$DF,$BD,$F3,$1F,$D3,$BE,$67,$FA,$C6,$BF,$EB,$F5,$F2,$FA
	.dc.b	$F9,$AF,$F6,$56,$FF,$87,$BF,$EB,$2F,$DA,$3F,$2C,$FF,$58,$E3,$D4
	.dc.b	$BE,$65,$F3,$3F,$D5,$FE,$6B,$FA,$E7,$1F,$E2,$30,$F1,$7C,$D7,$E4
	.dc.b	$7C,$6F,$EA,$FF,$61,$FE,$C5,$FB,$0F,$F6,$27,$D8,$3E,$C7,$5F,$3D
	.dc.b	$FD,$67,$2F,$C8,$FA,$FF,$D5,$FE,$7B,$FD,$81,$CB,$FA,$88,$C7,$87
	.dc.b	$F9,$D7,$FA,$B7,$27,$CB,$9D,$7C,$D7,$FB,$3E,$30,$7E,$C6,$7F,$B1
	.dc.b	$7A,$BF,$A7,$7F,$AB,$C7,$39,$FE,$AF,$8F,$E8,$BD,$1F,$E8,$BF,$2F
	.dc.b	$FA,$B7,$FA,$3F,$CA,$5F,$97,$FF,$5D,$E1,$FB,$19,$FE,$A4,$F8,$0F
	.dc.b	$DB,$47,$F5,$37,$87,$C7,$7F,$AF,$7E,$65,$F2,$FF,$D5,$36,$FF,$64
	.dc.b	$7C,$BF,$F5,$1F,$A3,$FD,$D7,$B1,$FD,$33,$E9,$DF,$31,$FE,$B1,$EA
	.dc.b	$CF,$CB,$7F,$42,$F9,$7B,$97,$E4,$36,$FE,$21,$8F,$F4,$F8,$FE,$95
	.dc.b	$8F,$FA,$C6,$78,$FC,$87,$FA,$AB,$EC,$FF,$85,$5F,$7C,$FF,$D7,$DC
	.dc.b	$1A,$F9,$6F,$E9,$5C,$DF,$C1,$BE,$3F,$FA,$57,$CA,$FF,$0F,$7F,$A6
	.dc.b	$7C,$A5,$F9,$17,$97,$E3,$DB,$FB,$D4,$7E,$E5,$F8,$D5,$FE,$A8,$F9
	.dc.b	$2F,$E3,$4F,$F4,$98,$E5,$4F,$92,$FE,$91,$97,$F1,$6F,$7B,$F9,$DF
	.dc.b	$C8,$BF,$35,$FE,$41,$18,$F8,$CB,$EC,$7F,$AA,$32,$FC,$F1,$FC,$25
	.dc.b	$FF,$54,$3F,$CC,$30,$E9,$1F,$D3,$FE,$3B,$FC,$3B,$F1,$EF,$90,$FE
	.dc.b	$5D,$8F,$F4,$AF,$7A,$DE,$FC,$95,$B5,$FF,$56,$E4,$F8,$7F,$E8,$DE
	.dc.b	$3F,$FA,$8D,$BF,$19,$FD,$39,$FF,$50,$7D,$07,$F1,$C7,$FD,$3F,$BF
	.dc.b	$F0,$77,$FA,$1B,$B3,$FA,$1F,$93,$FD,$09,$C1,$FD,$03,$0F,$03,$FE
	.dc.b	$97,$FA,$9F,$F4,$8F,$A9,$AF,$8E,$7C,$67,$F4,$1C,$7F,$AA,$C6,$3C
	.dc.b	$1F,$16,$1C,$7E,$3B,$FA,$CE,$5D,$23,$FA,$37,$8B,$FE,$A0,$C9,$F1
	.dc.b	$DF,$D7,$72,$FF,$56,$7C,$77,$93,$E2,$BF,$9D,$FC,$77,$F3,$E9,$FE
	.dc.b	$60,$BE,$39,$F1,$CF,$8D,$7D,$B7,$1F,$6B,$F1,$7B,$7C,$33,$FB,$66
	.dc.b	$5F,$EA,$07,$ED,$65,$D3,$1E,$7F,$55,$E3,$9F,$F4,$FC,$76,$F9,$0C
	.dc.b	$47,$FA,$7F,$3F,$07,$9F,$F2,$BB,$FF,$9E,$F9,$57,$E3,$3F,$98,$FC
	.dc.b	$85,$F0,$FE,$2E,$2B,$FF,$4C,$BD,$1F,$E9,$1F,$1F,$6F,$F9,$F7,$C7
	.dc.b	$DF,$65,$F5,$1F,$E9,$7F,$8F,$FB,$3F,$13,$FC,$EF,$D1,$BF,$1B,$FE
	.dc.b	$94,$F8,$EB,$F1,$DE,$0A,$FE,$73,$B7,$A9,$FE,$70,$EB,$E3,$BF,$9B
	.dc.b	$7D,$0F,$BB,$E2,$3F,$94,$F8,$9F,$65,$FE,$5D,$F3,$3F,$88,$7F,$95
	.dc.b	$DF,$CA,$3F,$94,$7C,$4F,$F2,$B7,$07,$F8,$CD,$7D,$2F,$FA,$0F,$D3
	.dc.b	$ED,$18,$F9,$A7,$D9,$3D,$17,$D3,$7D,$0F,$4B,$F9,$6E,$BF,$8B,$7F
	.dc.b	$96,$B7,$E3,$3F,$96,$E9,$F9,$C9,$FE,$5A,$F8,$9F,$F3,$E6,$FC,$E7
	.dc.b	$FD,$0F,$F8,$07,$DC,$F4,$7F,$8D,$7D,$1B,$E2,$7F,$4B,$18,$F8,$BB
	.dc.b	$F1,$7F,$E2,$0C,$FF,$3B,$E5,$FF,$8D,$AB,$FC,$EC,$FF,$28,$F8,$9F
	.dc.b	$F3,$AC,$F0,$F8,$AF,$E5,$0D,$F8,$9F,$F4,$07,$C5,$7E,$CB,$C0,$FE
	.dc.b	$47,$C5,$FA,$77,$FC,$E1,$D7,$9F,$2B,$F3,$78,$7C,$C9,$47,$F2,$58
	.dc.b	$C3,$FC,$AB,$E1,$9F,$41,$FE,$70,$F8,$9F,$CC,$DF,$F2,$7C,$7F,$69
	.dc.b	$F8,$73,$EF,$1F,$C0,$2F,$C1,$1F,$A2,$3F,$61,$7D,$6B,$B3,$F8,$4E
	.dc.b	$3C,$70,$E1,$7B,$A7,$EF,$BA,$C9,$FB,$3F,$E8,$B0,$E1,$C5,$FB,$50
	.dc.b	$F3,$D3,$C3,$A3,$E3,$FF,$63,$39,$CF,$C9,$FA,$0F,$ED,$3F,$54,$7E
	.dc.b	$EA,$8F,$F0,$97,$C9,$FF,$89,$BF,$5E,$7E,$07,$FC,$82,$F7,$5A,$F6
	.dc.b	$95,$FC,$2B,$97,$F8,$C5,$6B,$7F,$33,$DB,$F5,$6F,$DD,$BC,$BD,$46
	.dc.b	$7D,$EB,$B6,$9E,$39,$CE,$4F,$84,$FE,$17,$F2,$BF,$D4,$73,$FE,$4C
	.dc.b	$FF,$56,$F9,$7F,$E6,$AF,$F0,$7F,$57,$F8,$4B,$FE,$99,$F1,$BF,$D0
	.dc.b	$AF,$FA,$5B,$E5,$F9,$FC,$07,$F4,$DF,$97,$FF,$4C,$63,$FC,$C7,$C6
	.dc.b	$7E,$C1,$FA,$DF,$A0,$FB,$9E,$D7,$F0,$CF,$CA,$FF,$A1,$D6,$BD,$7F
	.dc.b	$F7,$4F,$77,$EE,$3E,$05,$F9,$2E,$B2,$FD,$B7,$3F,$D9,$C3,$AF,$E6
	.dc.b	$5F,$C7,$9E,$5A,$76,$FD,$6F,$F9,$4C,$63,$2F,$E0,$7E,$F7,$F2,$CF
	.dc.b	$C3,$3F,$C7,$DF,$91,$7D,$AF,$C5,$1F,$BF,$BF,$17,$FE,$F9,$C9,$FD
	.dc.b	$2E,$F7,$DE,$BD,$D3,$F1,$EF,$EB,$B5,$FE,$49,$FD,$EF,$F9,$8B,$DB
	.dc.b	$87,$F2,$0F,$2B,$EF,$1E,$3F,$AE,$69,$E3,$D3,$FC,$B9,$3F,$DA,$3E
	.dc.b	$A5,$F4,$6F,$DB,$27,$C1,$7F,$1E,$FC,$7D,$F9,$26,$EB,$55,$FC,$37
	.dc.b	$B9,$4F,$D5,$B5,$3F,$D1,$B1,$EB,$FB,$A5,$C3,$8E,$5F,$CA,$BD,$6F
	.dc.b	$EA,$75,$E2,$BF,$0E,$4E,$7F,$EA,$FE,$33,$F3,$FF,$E0,$FF,$8F,$BF
	.dc.b	$58,$7D,$53,$F1,$C7,$E9,$7F,$E0,$4C,$38,$7C,$2B,$0E,$17,$BF,$DF
	.dc.b	$B5,$97,$F1,$2E,$6F,$DD,$61,$DF,$E6,$BF,$AF,$BC,$B0,$E3,$9F,$92
	.dc.b	$FC,$7A,$78,$BD,$8F,$ED,$BF,$8B,$BF,$37,$FF,$6C,$7E,$7A,$FA,$B7
	.dc.b	$E2,$BE,$DF,$CD,$FF,$C3,$37,$BF,$98,$5E,$EB,$E1,$7E,$89,$AC,$7F
	.dc.b	$DC,$5C,$3F,$C3,$F0,$ED,$E9,$7F,$74,$79,$61,$C7,$9F,$C7,$7E,$2D
	.dc.b	$3C,$3E,$77,$9B,$F3,$57,$E8,$AF,$D2,$5F,$B2,$FF,$85,$7F,$4D,$7E
	.dc.b	$B1,$E6,$BD,$D6,$BE,$C1,$4F,$DE,$75,$8F,$8E,$DF,$86,$58,$ED,$97
	.dc.b	$F8,$87,$EC,$BC,$51,$8C,$5F,$67,$F2,$BF,$62,$63,$F2,$7C,$87,$E2
	.dc.b	$EF,$D6,$3F,$31,$ED,$7F,$8D,$79,$D8,$70,$E4,$FE,$89,$7B,$FD,$C3
	.dc.b	$E7,$CF,$4E,$AF,$94,$F8,$1B,$9F,$7A,$7D,$3F,$8E,$31,$F5,$1F,$7B
	.dc.b	$F6,$5F,$8B,$EB,$F9,$3D,$E7,$EE,$7F,$33,$7F,$25,$6B,$2F,$E7,$B2
	.dc.b	$FD,$CB,$F0,$33,$E8,$D3,$EE,$3F,$41,$BD,$9F,$BA,$BF,$E8,$BF,$B9
	.dc.b	$F8,$63,$19,$7F,$9D,$EB,$F1,$7E,$AB,$07,$ED,$4E,$4B,$FA,$0F,$F0
	.dc.b	$0D,$BE,$45,$EE,$B4,$25,$F9,$EF,$4C,$3F,$1D,$5F,$D7,$B1,$FC,$93
	.dc.b	$C7,$1E,$FE,$5F,$E8,$A3,$18,$FE,$D6,$BF,$3F,$F0,$3F,$C1,$FE,$21
	.dc.b	$FA,$EB,$F1,$37,$E8,$8F,$C5,$77,$F0,$2B,$DD,$6B,$F4,$FC,$E5,$FA
	.dc.b	$3C,$BF,$8B,$B6,$F8,$65,$D7,$E6,$BF,$6B,$5E,$18,$F0,$57,$EE,$3E
	.dc.b	$07,$FD,$61,$F1,$5F,$E3,$8F,$C9,$9F,$8B,$3F,$55,$FD,$BD,$6B,$EE
	.dc.b	$15,$AF,$D6,$F3,$F6,$3F,$B3,$AF,$E7,$59,$7E,$DD,$E1,$4F,$0F,$C1
	.dc.b	$5F,$CF,$F0,$E7,$F8,$A7,$F0,$AF,$82,$FF,$70,$FA,$3F,$E1,$CF,$8B
	.dc.b	$7E,$F8,$FD,$17,$F8,$4D,$6B,$EC,$15,$AF,$2B,$3F,$71,$FC,$16,$BF
	.dc.b	$95,$70,$7F,$09,$78,$53,$F5,$0F,$05,$7F,$02,$C3,$9F,$9D,$FC,$73
	.dc.b	$C3,$FE,$39,$FB,$5F,$F7,$2F,$CB,$5F,$B9,$3B,$3F,$8D,$D6,$BE,$79
	.dc.b	$5A,$FC,$DF,$3F,$AA,$7F,$17,$AF,$E4,$DC,$1F,$C6,$DE,$15,$FE,$0D
	.dc.b	$8E,$7F,$06,$CF,$C1,$E0,$78,$FD,$57,$E0,$DF,$DB,$7F,$4A,$FF,$30
	.dc.b	$7B,$5E,$29,$CF,$CF,$A7,$FC,$C9,$F0,$BE,$6F,$7F,$C5,$5A,$9F,$F5
	.dc.b	$DE,$9C,$71,$E9,$DD,$F0,$7D,$C7,$EB,$BF,$E7,$DE,$FF,$EF,$9F,$9A
	.dc.b	$7C,$09,$CF,$D3,$7D,$6D,$BF,$0B,$EC,$FF,$15,$E5,$7E,$FB,$F0,$6E
	.dc.b	$65,$CF,$5E,$CF,$5F,$3B,$E8,$7E,$4E,$FC,$7D,$FB,$0F,$FB,$43,$F5
	.dc.b	$3F,$F2,$7E,$9D,$6B,$53,$9F,$47,$E9,$C7,$CF,$FB,$DF,$A0,$EB,$F1
	.dc.b	$DF,$E4,$17,$C7,$1E,$B1,$FB,$4D,$3E,$DF,$BD,$FC,$F7,$ED,$FF,$98
	.dc.b	$F1,$3F,$78,$CF,$C8,$7F,$92,$3E,$1F,$FC,$3D,$F2,$3F,$82,$FB,$EF
	.dc.b	$BD,$8F,$19,$FF,$47,$E4,$3F,$8A,$73,$FF,$17,$C3,$97,$27,$E3,$FE
	.dc.b	$0F,$F4,$DF,$9F,$BF,$32,$7E,$86,$FC,$CF,$F1,$F5,$A9,$CF,$1E,$3D
	.dc.b	$B6,$1C,$74,$E9,$18,$AF,$F1,$1F,$2F,$D5,$F3,$FF,$C3,$5F,$9E,$3C
	.dc.b	$EF,$F6,$F7,$E3,$5F,$9C,$9C,$C7,$FB,$87,$CB,$C7,$C1,$AF,$87,$9B
	.dc.b	$FA,$7B,$DF,$99,$FA,$9F,$80,$7B,$EF,$C8,$3F,$17,$E0,$3E,$63,$F5
	.dc.b	$35,$A9,$CF,$BF,$8B,$E1,$E4,$7E,$5B,$87,$F1,$97,$BF,$23,$E6,$F2
	.dc.b	$FC,$CF,$79,$EE,$7F,$42,$EE,$FE,$5B,$FA,$E7,$ED,$A7,$3F,$19,$FD
	.dc.b	$E7,$E3,$1F,$B2,$67,$0E,$1E,$17,$F0,$1B,$DE,$3D,$BC,$3F,$B3,$ED
	.dc.b	$3E,$67,$F8,$E7,$E3,$DF,$E0,$4E,$1F,$D0,$57,$F8,$8E,$73,$FB,$B7
	.dc.b	$F0,$68,$C7,$84,$FC,$43,$F8,$74,$63,$F0,$8B,$EE,$FF,$BB,$3C,$87
	.dc.b	$B2,$FD,$F3,$94,$E6,$3F,$D8,$FA,$78,$23,$87,$E4,$5C,$BF,$1B,$FF
	.dc.b	$72,$7E,$45,$DB,$DB,$FD,$DF,$EE,$8F,$D5,$5F,$8B,$79,$67,$30,$FD
	.dc.b	$07,$1E,$12,$FF,$13,$D3,$F9,$BF,$03,$F3,$7C,$7F,$E1,$2A,$D4,$7B
	.dc.b	$3D,$6F,$D5,$7E,$92,$FC,$73,$E1,$46,$3E,$2E,$F7,$7F,$AC,$AF,$83
	.dc.b	$FE,$B4,$EF,$F8,$BD,$1F,$A4,$FC,$8B,$FD,$C1,$CF,$F3,$A3,$1F,$7F
	.dc.b	$F3,$5D,$61,$7B,$7F,$47,$F6,$DF,$3B,$D0,$FC,$FF,$77,$F9,$57,$1F
	.dc.b	$AF,$18,$F5,$BE,$B3,$AC,$2F,$A9,$F9,$FF,$07,$EB,$F9,$9F,$97,$F4
	.dc.b	$FF,$D6,$B8,$FE,$02,$31,$F8,$2F,$DF,$75,$85,$F3,$FF,$45,$EA,$7E
	.dc.b	$03,$AF,$F2,$7F,$5C,$FB,$CF,$C3,$46,$3E,$EA,$F9,$DF,$E1,$CA,$E0
	.dc.b	$FF,$87,$BE,$5F,$F1,$3E,$07,$E3,$3A,$BF,$1C,$FE,$3B,$F3,$EB,$E8
	.dc.b	$7E,$5F,$A7,$6F,$E6,$7E,$3F,$8F,$99,$F8,$FF,$A2,$E2,$FE,$47,$C8
	.dc.b	$BE,$67,$EC,$78,$3F,$01,$9F,$E0,$29,$2C,$3C,$7E,$57,$E3,$BE,$0F
	.dc.b	$C6,$FE,$43,$E2,$2F,$94,$F9,$DD,$AF,$1B,$D7,$E4,$FE,$23,$BB,$F5
	.dc.b	$7A,$FF,$27,$F8,$AF,$E4,$71,$89,$FF,$39,$7D,$7F,$0F,$29,$F9,$4F
	.dc.b	$C4,$F2,$BF,$95,$FE,$A1,$F1,$1F,$C5,$61,$D1,$FF,$0A,$7C,$E6,$FA
	.dc.b	$7F,$C0,$9E,$1F,$CA,$F2,$7F,$0F,$7E,$27,$F9,$3F,$E2,$D9,$F2,$9F
	.dc.b	$95,$7D,$29,$FF,$62,$7E,$85,$F5,$3C,$9F,$C3,$61,$EA,$7F,$8F,$7E
	.dc.b	$23,$F5,$9F,$E4,$BF,$8D,$EF,$CB,$7E,$EB,$F7,$3C,$1F,$C2,$E1,$F0
	.dc.b	$9F,$E2,$6D,$F2,$1E,$D6,$DF,$F3,$24,$7C,$8E,$AF,$E1,$ED,$F1,$BF
	.dc.b	$8C,$75,$7E,$1A,$7F,$D7,$1F,$5D,$F8,$A7,$FD,$75,$EA,$5F,$1B,$F8
	.dc.b	$96,$1C,$DF,$BF,$F8,$7F,$17,$F8,$D7,$3D,$F5,$79,$78,$BF,$C3,$AF
	.dc.b	$C4,$FF,$82,$BF,$02,$F0,$3F,$DB,$E3,$3E,$97,$87,$5F,$EE,$11,$E9
	.dc.b	$7F,$D5,$DF,$21,$E7,$5E,$8E,$FD,$1F,$47,$83,$FE,$4D,$FD,$5B,$D8
	.dc.b	$FF,$6A,$C3,$E1,$3F,$E5,$1D,$73,$E6,$FB,$FB,$BF,$B0,$D7,$C7,$7F
	.dc.b	$DF,$9D,$FF,$15,$FC,$B5,$58,$F8,$DD,$BF,$62,$BE,$73,$FD,$65,$BD
	.dc.b	$BA,$E2,$F9,$9D,$7F,$E5,$C9,$FA,$8F,$F5,$26,$FA,$DA,$FC,$67,$E9
	.dc.b	$74,$FF,$94,$27,$ED,$3F,$D2,$DB,$EC,$EB,$D9,$DF,$4F,$F9,$12,$76
	.dc.b	$FE,$D1,$E3,$CF,$D3,$1E,$BD,$1F,$E3,$C9,$EA,$BE,$2F,$F0,$99,$7F
	.dc.b	$AA,$FE,$DA,$F4,$FF,$B7,$EF,$C7,$F9,$3F,$FB,$F1,$BF,$9B,$70,$E8
	.dc.b	$FF,$59,$47,$A1,$FF,$40,$3E,$AE,$77,$91,$D9,$7E,$1F,$E0,$E7,$FF
	.dc.b	$03,$D7,$C0,$5F,$67,$E4,$E6,$FE,$FB,$3F,$11,$7E,$11,$E8,$DF,$FC
	.dc.b	$35,$3F,$21,$78,$B9,$E5,$D7,$2E,$FE,$93,$87,$2B,$0F,$F6,$47,$6F
	.dc.b	$F8,$DF,$D5,$FE,$BC,$CF,$C3,$B5,$9F,$F6,$19,$CC,$FF,$63,$FC,$DB
	.dc.b	$2F,$13,$23,$F2,$FC,$BF,$C7,$CC,$FE,$31,$8A,$DC,$3F,$18,$C3,$CB
	.dc.b	$F0,$8A,$F8,$3F,$1E,$AF,$9B,$E6,$AB,$FD,$D7,$ED,$EA,$F8,$EF,$F4
	.dc.b	$26,$FE,$E8,$F1,$1E,$E8,$7E,$7B,$C3,$E1,$1F,$6F,$7F,$79,$E7,$F1
	.dc.b	$1F,$EE,$7C,$79,$6F,$EE,$87,$EF,$7E,$0F,$9D,$7E,$7B,$7F,$1B,$F1
	.dc.b	$FE,$15,$F0,$F1,$E3,$BF,$CA,$8F,$E3,$71,$88,$FA,$3C,$EF,$C9,$FD
	.dc.b	$6B,$F2,$9F,$6F,$43,$F4,$3B,$F8,$BF,$C0,$F8,$2F,$F3,$86,$5C,$F2
	.dc.b	$FF,$0C,$74,$BF,$51,$F8,$AB,$F4,$5B,$3F,$23,$F4,$BE,$3E,$07,$DF
	.dc.b	$8F,$BF,$DA,$E2,$F9,$AB,$F5,$6F,$9F,$F5,$9C,$2B,$F0,$47,$E2,$BE
	.dc.b	$3F,$0F,$D0,$3F,$D0,$9D,$FF,$43,$E8,$BE,$CB,$F4,$BF,$37,$BA,$70
	.dc.b	$ED,$FC,$7C,$BE,$37,$18,$7E,$5B,$C1,$78,$0F,$C5,$FF,$07,$18,$FD
	.dc.b	$8A,$C7,$0A,$FF,$1D,$4B,$E5,$71,$1F,$E9,$7F,$0F,$ED,$FE,$E5,$F8
	.dc.b	$C7,$EE,$FC,$6D,$F8,$DD,$D7,$F9,$2E,$2F,$87,$F9,$FF,$DC,$3F,$26
	.dc.b	$FE,$87,$F7,$1B,$F2,$3A,$FA,$9E,$3E,$37,$DD,$FD,$A3,$F3,$2F,$DF
	.dc.b	$FD,$9E,$F8,$9A,$FC,$63,$BF,$90,$F1,$DF,$93,$7E,$82,$31,$FB,$25
	.dc.b	$ED,$DF,$8F,$E6,$3F,$01,$EB,$7E,$99,$FC,$17,$0D,$F0,$B5,$D3,$EC
	.dc.b	$F1,$7E,$F7,$1F,$EB,$1F,$C0,$7A,$D7,$E3,$F5,$F4,$3F,$23,$C3,$FB
	.dc.b	$5C,$7F,$B4,$7E,$E7,$9F,$7D,$5F,$EA,$5E,$7F,$E4,$E7,$F5,$FF,$3B
	.dc.b	$F4,$3C,$3F,$B3,$E2,$39,$3F,$7B,$18,$FB,$DB,$E3,$FD,$DF,$0E,$FC
	.dc.b	$B7,$D8,$F1,$BE,$AF,$2F,$F0,$47,$EB,$FE,$61,$7D,$87,$8F,$B5,$F2
	.dc.b	$5F,$80,$EB,$FB,$1F,$A5,$7F,$0C,$7E,$D7,$9D,$7A,$BF,$C1,$BE,$A7
	.dc.b	$E8,$F6,$FF,$6E,$57,$F1,$DF,$17,$F1,$3E,$3F,$D9,$AF,$E2,$AF,$DB
	.dc.b	$F8,$5B,$EC,$BE,$53,$6F,$F8,$2B,$AF,$F6,$DE,$3F,$DA,$AF,$E3,$AF
	.dc.b	$DC,$F8,$8B,$E5,$3D,$0F,$CB,$CF,$E0,$F8,$3F,$33,$E4,$7D,$CF,$AF
	.dc.b	$FE,$4A,$FD,$EC,$2F,$B7,$FE,$E2,$DF,$D5,$C4,$AF,$E0,$9E,$87,$E8
	.dc.b	$BC,$9F,$BF,$F7,$7F,$CB,$9F,$C0,$FE,$97,$7D,$CF,$F1,$8E,$16,$FF
	.dc.b	$59,$E3,$FE,$03,$F5,$1F,$E6,$EF,$E0,$BF,$33,$BE,$57,$FA,$C3,$CC
	.dc.b	$DB,$FE,$B2,$EB,$FD,$A7,$91,$F7,$FF,$09,$FE,$7E,$FE,$07,$F0,$9B
	.dc.b	$E4,$FE,$F7,$CB,$DB,$FB,$FE,$3F,$DE,$79,$1F,$80,$FC,$DF,$FA,$3B
	.dc.b	$F8,$2F,$C1,$6F,$91,$F8,$2F,$2F,$6F,$E1,$3D,$4F,$E0,$9E,$3F,$E0
	.dc.b	$BA,$BF,$A7,$BF,$84,$E9,$BE,$43,$DB,$FE,$B5,$E3,$7F,$37,$AF,$E0
	.dc.b	$1E,$9F,$F0,$DF,$2B,$F7,$9D,$AE,$4F,$EB,$31,$8E,$6B,$E4,$BD,$AF
	.dc.b	$F6,$66,$5F,$D0,$6B,$F8,$27,$D9,$7F,$17,$F2,$3F,$80,$7B,$5F,$ED
	.dc.b	$9F,$9B,$FE,$C7,$18,$E7,$7E,$55,$8F,$37,$F8,$2E,$1C,$5F,$E0,$FB
	.dc.b	$FF,$21,$EB,$FE,$07,$FB,$C3,$F4,$5F,$ED,$B1,$8F,$92,$BE,$63,$DE
	.dc.b	$FB,$FC,$3F,$D5,$6B,$F8,$5F,$E6,$DF,$CA,$FC,$DF,$E0,$FE,$6F,$F8
	.dc.b	$9F,$CE,$FD,$34,$63,$D9,$FE,$F2,$FF,$0C,$F7,$B8,$57,$F9,$EB,$B3
	.dc.b	$F9,$DF,$9D,$FD,$0F,$F5,$97,$E4,$AF,$A2,$FF,$01,$CE,$61,$FB,$76
	.dc.b	$3D,$0F,$F7,$DD,$FF,$A6,$B8,$3F,$B6,$DF,$3F,$81,$FE,$8F,$E8,$7F
	.dc.b	$0E,$FE,$01,$7E,$64,$FC,$2F,$FC,$1F,$39,$87,$E8,$18,$F4,$FF,$A6
	.dc.b	$5F,$F0,$AE,$DC,$7C,$2A,$3F,$69,$1F,$E1,$8F,$0B,$FA,$3F,$A6,$FC
	.dc.b	$99,$FB,$76,$7D,$4F,$E2,$73,$9F,$CD,$7F,$C3,$FF,$31,$FE,$28,$E3
	.dc.b	$FF,$02,$E3,$CB,$E0,$7F,$AF,$BD,$3A,$BF,$B0,$FA,$5F,$D3,$FE,$91
	.dc.b	$FA,$CB,$F5,$EF,$F1,$4C,$E6,$3F,$C1,$BE,$1F,$F7,$B8,$EC,$FF,$7D
	.dc.b	$F1,$AF,$17,$F8,$E6,$7A,$67,$DB,$D7,$FD,$06,$5F,$E8,$4D,$7F,$C3
	.dc.b	$38,$7F,$A5,$3F,$51,$70,$3F,$70,$7A,$9F,$E3,$AA,$D4,$E7,$F3,$EF
	.dc.b	$F1,$EF,$E5,$1E,$0D,$BF,$C8,$95,$AF,$4B,$FB,$EB,$C3,$E1,$BF,$BF
	.dc.b	$E3,$FE,$23,$FD,$21,$F8,$E3,$F5,$9F,$AF,$BF,$F8,$B2,$B5,$39,$F1
	.dc.b	$3F,$C9,$71,$DF,$8D,$A7,$F9,$54,$FF,$2D,$DF,$1F,$71,$97,$6F,$81
	.dc.b	$F0,$C6,$32,$FF,$80,$17,$FE,$56,$F6,$BF,$BA,$7E,$28,$FC,$51,$F3
	.dc.b	$EF,$D2,$75,$EE,$AB,$52,$FB,$CC,$C7,$1F,$C1,$3F,$C3,$B8,$F4,$C5
	.dc.b	$E7,$FF,$88,$3D,$BF,$F1,$D7,$E7,$EF,$D2,$DF,$90,$BD,$3F,$F3,$4D
	.dc.b	$6A,$73,$F8,$67,$F9,$B3,$EB,$BE,$17,$17,$F4,$13,$E0,$F8,$FF,$E3
	.dc.b	$FB,$E1,$F5,$1D,$FE,$09,$A7,$7F,$83,$FF,$18,$FC,$0F,$F9,$0F,$F7
	.dc.b	$E7,$E4,$5E,$1F,$89,$FF,$3C,$CE,$63,$C9,$F8,$76,$35,$EE,$7F,$A1
	.dc.b	$FE,$C7,$C7,$5B,$8E,$BF,$0A,$C7,$AF,$93,$FE,$43,$F8,$0F,$F3,$BF
	.dc.b	$C7,$BF,$1D,$7E,$F6,$FD,$6D,$FC,$01,$FE,$8D,$BD,$D6,$A5,$F8,$26
	.dc.b	$7F,$2C,$FC,$25,$7F,$A0,$B5,$F2,$8F,$33,$FD,$15,$F1,$3F,$CE,$F1
	.dc.b	$EF,$87,$8E,$31,$9B,$F0,$6F,$F2,$C7,$7F,$F9,$53,$E3,$DF,$B2,$BF
	.dc.b	$52,$7D,$23,$F5,$4F,$F5,$15,$EE,$B5,$2F,$D5,$F3,$F7,$9F,$99,$AF
	.dc.b	$F2,$FF,$07,$9B,$83,$F0,$1F,$1C,$D3,$BF,$D0,$FF,$4C,$F1,$7F,$D7
	.dc.b	$18,$75,$D9,$EA,$FF,$9B,$7E,$23,$FD,$33,$F9,$5B,$EB,$1F,$A6,$FF
	.dc.b	$B4,$BE,$F7,$FD,$6D,$5A,$9C,$FE,$0A,$FC,$B3,$3F,$9F,$7E,$B6,$BF
	.dc.b	$D4,$7C,$5E,$57,$86,$2F,$C3,$FC,$1F,$26,$D3,$C1,$3F,$E2,$5B,$FF
	.dc.b	$62,$7C,$77,$FB,$27,$F6,$07,$EC,$5F,$9D,$F9,$BF,$E9,$D5,$FE,$D7
	.dc.b	$AD,$4E,$79,$B8,$FD,$D7,$EF,$E7,$F7,$D5,$FE,$DC,$F8,$6F,$F5,$CC
	.dc.b	$70,$E6,$FE,$79,$18,$F9,$26,$7E,$09,$FE,$7B,$3F,$EC,$0F,$88,$7D
	.dc.b	$7F,$FA,$CE,$DF,$8F,$3F,$53,$7D,$8F,$FB,$76,$F7,$F7,$9F,$EE,$3A
	.dc.b	$D7,$EB,$99,$C7,$F8,$A5,$3E,$8E,$C7,$5A,$FE,$61,$E0,$7F,$BC,$6B
	.dc.b	$9E,$7E,$0F,$A7,$FF,$53,$7C,$3B,$F4,$1F,$EB,$3C,$BF,$C2,$3F,$32
	.dc.b	$7E,$65,$FE,$EF,$AD,$7D,$3F,$E1,$FE,$29,$5F,$99,$E4,$FF,$6E,$D7
	.dc.b	$3E,$8F,$C7,$2F,$D0,$7F,$95,$3C,$B2,$FE,$2B,$E8,$F8,$B4,$F0,$E5
	.dc.b	$FE,$5D,$E0,$FE,$5F,$8F,$EA,$FE,$1F,$F9,$27,$E3,$0F,$DC,$1F,$85
	.dc.b	$7F,$3B,$F8,$26,$FF,$EE,$EB,$DD,$6B,$E6,$9F,$17,$AC,$FA,$FA,$7F
	.dc.b	$E5,$BF,$93,$FE,$7D,$87,$F3,$1D,$7F,$4C,$F2,$D3,$F4,$31,$C4,$FE
	.dc.b	$87,$5F,$E6,$3F,$96,$6D,$E5,$AF,$E7,$FA,$FF,$C0,$0D,$BF,$87,$F2
	.dc.b	$BF,$6F,$7E,$FC,$FD,$E1,$FB,$3F,$5D,$7F,$85,$61,$C2,$F7,$4F,$D5
	.dc.b	$F5,$B2,$BF,$8D,$DF,$EF,$5F,$D9,$B7,$E0,$3E,$BE,$1E,$0F,$90,$FF
	.dc.b	$73,$57,$5C,$FE,$2D,$F8,$7F,$28,$FC,$27,$E5,$2F,$C7,$DF,$AB,$BE
	.dc.b	$61,$FB,$8F,$FB,$EF,$8D,$D3,$F6,$31,$E3,$87,$0B,$DF,$E3,$FD,$7B
	.dc.b	$FE,$3E,$9F,$CA,$2F,$BB,$61,$A7,$87,$4F,$07,$C8,$FF,$72,$D3,$BD
	.dc.b	$FF,$74,$C3,$BD,$79,$63,$5A,$7E,$2B,$D5,$7E,$EB,$FE,$E2,$FC,$35
	.dc.b	$FB,$13,$F7,$57,$E5,$BF,$DC,$74,$EB,$97,$F0,$AC,$38,$5E,$EB,$FB
	.dc.b	$95,$6B,$1E,$BF,$A3,$7F,$85,$5E,$FF,$82,$5C,$38,$E5,$FE,$55,$CB
	.dc.b	$EF,$56,$B3,$F1,$CB,$A3,$D3,$5F,$E1,$0F,$C5,$5F,$B0,$3F,$71,$FF
	.dc.b	$25,$7E,$AB,$FD,$53,$F2,$E7,$DA,$FE,$26,$F7,$5A,$F9,$85,$3F,$4D
	.dc.b	$D6,$3C,$F2,$FE,$B3,$87,$6F,$DC,$7F,$C5,$CF,$2F,$C0,$DA,$7F,$62
	.dc.b	$FB,$FF,$2D,$AB,$53,$FD,$9B,$C6,$CF,$2F,$F2,$4F,$BE,$7E,$D2,$FD
	.dc.b	$6A,$7B,$FF,$D0,$FA,$BE,$7D,$EE,$B5,$5F,$E4,$4F,$C7,$7F,$49,$3F
	.dc.b	$CB,$BF,$49,$FF,$21,$3D,$75,$FE,$E2,$F1,$AE,$DF,$19,$E2,$F8,$BF
	.dc.b	$F2,$FE,$3D,$3E,$F3,$D3,$FB,$6B,$E2,$9F,$99,$BF,$28,$FE,$67,$FB
	.dc.b	$6F,$FA,$E7,$E3,$3F,$69,$7B,$AD,$7C,$72,$BF,$D7,$DF,$B7,$6B,$1E
	.dc.b	$7C,$1F,$E0,$4C,$3B,$7C,$27,$F9,$65,$E5,$9F,$6F,$C4,$7C,$77,$E2
	.dc.b	$8F,$0E,$5F,$3B,$E5,$DF,$BA,$BF,$42,$7E,$8B,$FE,$5C,$97,$ED,$4F
	.dc.b	$C6,$BF,$85,$47,$0A,$DF,$E0,$88,$DF,$DA,$6B,$17,$EF,$7F,$E6,$D9
	.dc.b	$E1,$C3,$FE,$1C,$79,$67,$DF,$E2,$FC,$97,$E3,$8F,$17,$C8,$7E,$13
	.dc.b	$E7,$7F,$5B,$B3,$F6,$07,$EC,$4F,$DA,$DF,$94,$7F,$85,$FC,$5F,$D7
	.dc.b	$E1,$C2,$F7,$5A,$3F,$92,$47,$4D,$BF,$C5,$EF,$EE,$BE,$43,$FD,$08
	.dc.b	$F1,$C7,$FC,$45,$3E,$3F,$73,$E2,$E1,$D3,$EA,$FF,$33,$EC,$FF,$AE
	.dc.b	$FF,$3D,$FE,$BF,$F8,$43,$F2,$3F,$F6,$F7,$EA,$1F,$CC,$EB,$5F,$3D
	.dc.b	$FE,$73,$9F,$F7,$17,$CF,$E7,$F0,$FF,$F3,$EE,$BF,$E0,$7F,$85,$FE
	.dc.b	$B3,$87,$FB,$1F,$F2,$7F,$F4,$C3,$CB,$3E,$DF,$03,$E2,$C1,$AF,$8B
	.dc.b	$3F,$E6,$7C,$BF,$C4,$BE,$19,$FA,$73,$E8,$FF,$67,$FB,$7B,$F6,$F7
	.dc.b	$EC,$6F,$D3,$3F,$A5,$61,$C2,$F7,$5A,$F4,$F3,$8B,$D6,$FF,$56,$5E
	.dc.b	$F9,$3F,$CB,$6F,$1F,$79,$A7,$7F,$A0,$FE,$03,$39,$8F,$17,$D2,$63
	.dc.b	$1F,$35,$F0,$CF,$CB,$FA,$67,$F5,$DF,$86,$FF,$2E,$7C,$93,$F5,$77
	.dc.b	$BE,$FC,$05,$FB,$7F,$F7,$EE,$6F,$E6,$78,$70,$BD,$D6,$BF,$15,$E9
	.dc.b	$F7,$8C,$BF,$D8,$D7,$FE,$69,$FD,$3A,$CE,$70,$E3,$D7,$FC,$6A,$BA
	.dc.b	$69,$E1,$FC,$5B,$F8,$EF,$E6,$CF,$C9,$9F,$33,$FE,$A9,$FB,$27,$C5
	.dc.b	$7F,$98,$7C,$2F,$EF,$17,$BA,$D4,$FF,$99,$7E,$CB,$38,$7B,$BB,$7F
	.dc.b	$D0,$F7,$B6,$C7,$6F,$C5,$BF,$C8,$B7,$FE,$8B,$9E,$38,$74,$F9,$2F
	.dc.b	$2D,$EF,$3F,$1F,$E0,$1F,$E6,$7E,$7F,$EB,$3F,$1D,$FC,$77,$F4,$97
	.dc.b	$DB,$BF,$56,$FE,$3D,$D5,$FD,$22,$F7,$F1,$0B,$DD,$6B,$E4,$FF,$C2
	.dc.b	$98,$F3,$DF,$FD,$3E,$FF,$29,$FB,$4F,$F7,$A3,$DF,$3E,$DF,$14,$E7
	.dc.b	$FE,$8F,$39,$C7,$F9,$77,$C6,$FF,$4A,$F7,$7F,$DD,$3F,$84,$3F,$6C
	.dc.b	$FE,$37,$FA,$63,$F6,$D7,$E9,$5F,$E2,$5B,$DD,$6A,$7F,$C2,$DC,$7F
	.dc.b	$DE,$27,$FD,$E7,$F9,$CF,$F9,$2B,$87,$FD,$38,$7F,$BE,$A3,$BB,$FE
	.dc.b	$B8,$79,$FC,$33,$27,$E8,$3E,$5F,$C8,$FF,$D6,$39,$F9,$3D,$FC,$6D
	.dc.b	$E7,$C3,$86,$BF,$27,$5F,$F0,$4F,$4F,$EC,$FC,$67,$E9,$DF,$E7,$FF
	.dc.b	$2B,$FD,$A1,$F9,$4F,$F8,$FE,$B5,$39,$FC,$5D,$FB,$7E,$7D,$CF,$E1
	.dc.b	$BC,$9E,$A1,$7F,$BF,$F8,$2F,$74,$7F,$74,$8E,$7F,$82,$B3,$FF,$36
	.dc.b	$48,$E3,$7F,$C4,$32,$F2,$7B,$1C,$63,$17,$C7,$37,$99,$FE,$55,$CF
	.dc.b	$FC,$4F,$E6,$7F,$AD,$7E,$8D,$FB,$63,$F4,$EF,$F2,$3F,$1F,$F8,$AA
	.dc.b	$F7,$F8,$67,$FB,$96,$B5,$FA,$97,$3F,$7D,$97,$AB,$87,$1F,$9E,$DC
	.dc.b	$B9,$56,$B1,$E5,$F3,$FF,$E1,$2C,$FA,$EB,$E2,$F8,$3F,$77,$E8,$3F
	.dc.b	$81,$F6,$3F,$C0,$9E,$4B,$CE,$FF,$69,$4E,$7E,$9F,$FC,$0B,$3F,$E1
	.dc.b	$4F,$C5,$3F,$D9,$9C,$1F,$1F,$0F,$F0,$83,$F3,$AF,$87,$D5,$79,$DB
	.dc.b	$F5,$3F,$E1,$8C,$FC,$35,$FE,$9A,$BF,$0F,$C4,$E2,$3A,$67,$E0,$CF
	.dc.b	$FC,$3D,$8F,$E5,$FE,$5F,$FA,$D7,$CC,$7F,$66,$FC,$63,$C7,$82,$73
	.dc.b	$F3,$5F,$DA,$EF,$87,$CC,$FF,$80,$AB,$CC,$7F,$AA,$AB,$FB,$33,$83
	.dc.b	$F0,$4F,$F6,$7B,$FF,$57,$D7,$2F,$27,$FB,$C7,$CD,$7F,$77,$F9,$9F
	.dc.b	$46,$1E,$1C,$7F,$B5,$CF,$F7,$BC,$9F,$8D,$79,$BE,$5F,$FA,$F7,$CA
	.dc.b	$BF,$39,$7E,$DF,$9F,$89,$FE,$BF,$F3,$7F,$DA,$E5,$F3,$99,$F4,$BE
	.dc.b	$66,$DF,$DA,$AD,$DF,$88,$FC,$D6,$7D,$BE,$6B,$CA,$F8,$3F,$0C,$B7
	.dc.b	$DB,$E7,$3F,$BE,$66,$F6,$3B,$6B,$E1,$F8,$3F,$9B,$E3,$FF,$5B,$F9
	.dc.b	$6F,$ED,$7E,$96,$3E,$63,$FA,$DF,$CC,$FF,$6F,$F9,$C7,$87,$9F,$17
	.dc.b	$C9,$C7,$FE,$8A,$BF,$EE,$CF,$CF,$31,$F3,$9F,$DC,$BE,$73,$E5,$E9
	.dc.b	$DF,$D8,$B7,$DB,$5F,$EC,$FE,$17,$F1,$9F,$95,$FE,$AB,$F3,$3F,$D7
	.dc.b	$E3,$1F,$31,$FD,$6F,$CB,$FE,$D1,$F3,$5F,$EE,$49,$CF,$CD,$79,$30
	.dc.b	$FE,$C5,$BF,$8D,$E7,$DF,$F7,$1E,$BF,$1D,$72,$E9,$9F,$F7,$2F,$A5
	.dc.b	$FC,$8F,$B3,$FD,$77,$E6,$BF,$AA,$FC,$AF,$F0,$4D,$7F,$97,$4E,$7E
	.dc.b	$5B,$FA,$9F,$83,$C3,$E5,$FF,$B0,$6F,$FC,$CE,$F7,$1F,$DB,$23,$E3
	.dc.b	$5F,$1F,$9A,$FE,$C1,$5A,$F9,$AF,$EC,$15,$CD,$FE,$DB,$93,$DD,$FD
	.dc.b	$2F,$AD,$FD,$37,$C2,$FE,$4B,$F6,$F8,$F9,$AF,$EB,$9F,$31,$FD,$53
	.dc.b	$E2,$FF,$81,$69,$FD,$17,$0F,$EB,$9A,$79,$4F,$E4,$C7,$FA,$5E,$DF
	.dc.b	$33,$F7,$A7,$B7,$3F,$E9,$1F,$07,$CC,$5A,$EF,$F3,$3F,$DB,$BC,$4F
	.dc.b	$DC,$FC,$BF,$F4,$EF,$97,$FE,$B9,$F8,$0E,$3E,$5F,$FA,$6F,$BD,$FD
	.dc.b	$2B,$E6,$B1,$F3,$5F,$8C,$C3,$FA,$BF,$07,$94,$FE,$B5,$7C,$31,$FE
	.dc.b	$AF,$1C,$FE,$63,$FA,$A4,$7E,$7E,$31,$F3,$3F,$D9,$FE,$6E,$E7,$FD
	.dc.b	$63,$8F,$F8,$6F,$2F,$F4,$8F,$97,$7C,$AF,$F5,$2E,$7F,$ED,$53,$98
	.dc.b	$7C,$CF,$0F,$5B,$FA,$AE,$9F,$C1,$AB,$FA,$A3,$FA,$17,$A7,$DB,$E6
	.dc.b	$7F,$A6,$E2,$F4,$71,$E6,$7F,$66,$F9,$8F,$EA,$1F,$31,$FC,$83,$E0
	.dc.b	$7F,$9E,$FC,$BF,$F4,$DF,$BC,$C7,$AF,$FD,$0F,$E5,$3F,$A8,$FC,$CB
	.dc.b	$E6,$73,$E6,$FE,$D3,$1F,$F7,$EF,$CC,$DC,$38,$64,$F9,$8F,$E3,$B5
	.dc.b	$DB,$DE,$B5,$AC,$9F,$A9,$3A,$5F,$8F,$FF,$3D,$F9,$6F,$E7,$5F,$35
	.dc.b	$FD,$32,$31,$F3,$3F,$EE,$0F,$37,$FA,$65,$F0,$EC,$FE,$83,$EA,$FF
	.dc.b	$49,$BD,$C7,$F4,$D7,$8F,$CC,$FF,$4B,$8F,$04,$62,$FF,$DD,$5F,$0B
	.dc.b	$FD,$7A,$31,$1F,$D2,$9B,$93,$F4,$1E,$37,$FD,$03,$E5,$BF,$9C,$FC
	.dc.b	$D7,$FA,$B7,$4F,$24,$FF,$3D,$F1,$BF,$A0,$CF,$FA,$DB,$F7,$0E,$33
	.dc.b	$93,$CC,$73,$F3,$7D,$4F,$5A,$FE,$93,$F3,$5F,$D3,$34,$FE,$A5,$97
	.dc.b	$83,$E6,$BD,$17,$BF,$A7,$FF,$5F,$FC,$D3,$0E,$FF,$39,$FD,$7B,$E6
	.dc.b	$FC,$3F,$2F,$FD,$33,$E5,$7F,$A5,$F0,$FA,$23,$1F,$85,$EB,$E1,$BF
	.dc.b	$27,$A5,$F0,$3C,$AA,$BF,$A4,$3B,$F2,$BF,$D6,$1F,$31,$E4,$8D,$5F
	.dc.b	$F4,$9F,$9C,$FF,$61,$E5,$DE,$31,$9F,$F4,$B8,$C6,$1F,$EB,$98,$7C
	.dc.b	$27,$C8,$F0,$FF,$A6,$7C,$AF,$FA,$D6,$FE,$7C,$63,$BB,$5F,$29,$FE
	.dc.b	$B3,$D3,$E6,$9F,$D2,$0B,$3F,$D5,$7E,$57,$CC,$FF,$49,$8E,$3F,$03
	.dc.b	$FE,$BE,$C5,$F4,$D8,$FA,$7F,$F5,$94,$3F,$59,$FB,$7F,$29,$FD,$33
	.dc.b	$E5,$9F,$57,$8F,$96,$FE,$83,$1F,$E4,$EE,$0F,$C2,$D7,$FA,$E3,$4F
	.dc.b	$2B,$FD,$31,$FE,$79,$E0,$FF,$28,$8F,$E9,$B1,$CF,$E5,$FF,$D7,$DF
	.dc.b	$31,$FE,$AD,$8C,$7D,$4F,$FA,$EB,$1F,$EB,$18,$71,$CB,$C3,$F3,$1F
	.dc.b	$90,$F2,$5F,$AB,$3E,$35,$F6,$EF,$D1,$1F,$9D,$BF,$08,$7E,$0B,$C7
	.dc.b	$0E,$17,$BA,$D7,$ED,$B9,$C7,$9E,$3C,$B1,$FF,$5F,$F2,$7E,$93,$3F
	.dc.b	$0D,$6B,$0E,$DC,$5F,$A4,$F7,$FF,$AC,$7C,$DF,$F6,$3E,$1F,$E9,$31
	.dc.b	$8F,$13,$5F,$3B,$FD,$8B,$9B,$F4,$87,$F9,$CE,$7F,$D8,$4F,$CA,$31
	.dc.b	$1F,$EB,$DF,$67,$FB,$1D,$F4,$8C,$7D,$87,$E5,$EF,$C3,$F3,$BD,$BE
	.dc.b	$7B,$FB,$2F,$EE,$1F,$B8,$F7,$3F,$B2,$7C,$7B,$F0,$4C,$7C,$E7,$F5
	.dc.b	$6F,$33,$FB,$3F,$37,$F0,$2B,$FE,$C7,$C1,$E6,$3F,$B1,$7C,$A7,$EA
	.dc.b	$A3,$FB,$0C,$72,$F9,$5F,$F6,$3D,$BE,$77,$F5,$55,$E0,$F8,$2D,$7C
	.dc.b	$FF,$F7,$1F,$37,$F8,$5F,$C0,$FF,$63,$FB,$17,$E8,$F8,$F9,$DF,$EB
	.dc.b	$91,$FE,$40,$DF,$F8,$AD,$7F,$AF,$36,$FD,$2B,$FD,$91,$F3,$3F,$DB
	.dc.b	$7E,$CF,$F7,$4F,$F6,$FF,$AA,$B8,$74,$FA,$1E,$DF,$41,$FD,$9B,$2F
	.dc.b	$F7,$06,$5D,$AB,$C3,$F9,$7E,$BE,$83,$FD,$9D,$CB,$FC,$8B,$A3,$FB
	.dc.b	$17,$DA,$5F,$9E,$FE,$B7,$F4,$3F,$D9,$32,$FE,$57,$5F,$EC,$1E,$1F
	.dc.b	$DD,$BF,$D9,$1F,$BC,$7F,$AC,$BE,$03,$FB,$07,$CF,$BE,$83,$F8,$95
	.dc.b	$7F,$73,$FA,$0B,$97,$6F,$23,$C1,$87,$0C,$FF,$21,$97,$FA,$F7,$E1
	.dc.b	$FF,$B3,$FC,$BB,$EA,$3F,$9C,$C6,$3F,$4F,$BF,$07,$FE,$B0,$BF,$E1
	.dc.b	$F3,$FD,$99,$B1,$FD,$9B,$E6,$FF,$92,$4F,$F6,$59,$E3,$E5,$7F,$B0
	.dc.b	$7E,$73,$FD,$81,$F3,$78,$F3,$3F,$D7,$DA,$7F,$73,$E5,$FE,$9B,$F6
	.dc.b	$C2,$10,$7D,$37,$F6,$DF,$C7,$BF,$C6,$3F,$45,$FE,$C3,$D1,$19,$7F
	.dc.b	$D5,$3F,$3D,$FD,$B7,$CA,$FF,$5E,$7E,$49,$99,$7B,$1F,$EC,$0F,$31
	.dc.b	$F6,$D9,$8F,$F5,$DC,$FF,$94,$F1,$FE,$E2,$D3,$AF,$D0,$7F,$71,$DB
	.dc.b	$C3,$18,$CB,$FB,$8B,$CF,$E8,$7F,$C0,$3F,$83,$5D,$7C,$9A,$7F,$71
	.dc.b	$C3,$8D,$FF,$6D,$F9,$DF,$EB,$FF,$01,$E7,$F9,$BF,$F5,$DF,$D0,$DF
	.dc.b	$A0,$FF,$5D,$5F,$8C,$FF,$58,$7A,$3F,$EC,$FE,$0F,$ED,$7F,$43,$FD
	.dc.b	$BA,$7F,$DB,$F1,$8C,$9F,$6F,$FE,$B9,$9E,$95,$FD,$F3,$E8,$7F,$B7
	.dc.b	$6B,$E4,$D3,$FB,$73,$E6,$F9,$EF,$ED,$DF,$6F,$F2,$DF,$ED,$B1,$9F
	.dc.b	$23,$FB,$6B,$7C,$2F,$26,$5F,$EB,$49,$FE,$D1,$F6,$DE,$48,$C7,$93
	.dc.b	$FD,$B9,$E9,$3F,$DA,$BE,$DE,$CF,$F6,$9F,$62,$FD,$B7,$FA,$D7,$2F
	.dc.b	$93,$E4,$FF,$6A,$F3,$7E,$F3,$FD,$A6,$33,$F3,$DF,$DA,$73,$F9,$12
	.dc.b	$8F,$F5,$9C,$61,$FF,$57,$7C,$F7,$96,$3F,$D6,$BF,$3A,$FB,$67,$DB
	.dc.b	$7F,$95,$67,$BD,$FF,$AD,$7E,$7F,$FB,$6E,$5E,$0F,$82,$FE,$E1,$F3
	.dc.b	$FF,$91,$F9,$EF,$ED,$9F,$3F,$7E,$73,$FB,$5F,$B5,$7E,$7F,$E3,$E8
	.dc.b	$BF,$ED,$4F,$FB,$CB,$C1,$FE,$C5,$3F,$DA,$3E,$D9,$CF,$E7,$BF,$F5
	.dc.b	$BD,$F3,$FA,$0F,$ED,$5F,$8D,$5F,$3F,$FD,$87,$F4,$1F,$A3,$F2,$BF
	.dc.b	$AC,$FC,$F7,$F6,$1F,$9C,$FE,$CB,$C9,$F9,$49,$CF,$CE,$FF,$61,$C3
	.dc.b	$F5,$5E,$EF,$F5,$CF,$B0,$BF,$33,$FD,$4F,$E7,$1E,$2E,$2F,$D7,$7F
	.dc.b	$D6,$B8,$3D,$47,$F5,$AF,$99,$FE,$25,$F6,$EF,$9D,$FF,$5E,$F4,$7F
	.dc.b	$5B,$9F,$F5,$EE,$FF,$A6,$C3,$FA,$DB,$74,$F0,$7C,$DF,$EF,$7D,$AF
	.dc.b	$EB,$7F,$37,$7E,$6F,$FA,$D7,$6B,$ED,$F1,$F3,$9F,$3A,$D5,$FE,$BA
	.dc.b	$E0,$FF,$5E,$9F,$81,$8F,$EB,$79,$FE,$D9,$FF,$5D,$C7,$2F,$9A,$FF
	.dc.b	$64,$7B,$97,$ED,$BF,$D7,$97,$FB,$CF,$6F,$FA,$D7,$CD,$FF,$11,$7F
	.dc.b	$AE,$7A,$B7,$E5,$FF,$8E,$E5,$E0,$31,$F3,$3F,$82,$AF,$EB,$11,$8F
	.dc.b	$97,$7B,$AF,$B6,$FF,$5D,$67,$DF,$1F,$EB,$4D,$C7,$FA,$D7,$0F,$1B
	.dc.b	$E1,$F3,$4F,$9A,$C6,$28,$FF,$6D,$E1,$FC,$53,$97,$FA,$DB,$7D,$8F
	.dc.b	$EB,$7F,$6D,$FC,$AE,$7F,$AD,$6D,$FC,$4D,$FE,$B4,$FD,$E3,$FA,$D7
	.dc.b	$DD,$3E,$DD,$F3,$7F,$D6,$E7,$96,$1F,$EC,$0F,$9A,$FD,$86,$1F,$EC
	.dc.b	$BF,$17,$3E,$87,$F5,$BF,$23,$F9,$D7,$3F,$F5,$AF,$99,$BE,$F7,$FA
	.dc.b	$CF,$E6,$6F,$8B,$FA,$AA,$FE,$B1,$AF,$F2,$77,$FD,$8F,$F3,$3F,$BD
	.dc.b	$7F,$AC,$71,$FF,$32,$7F,$B2,$FC,$C7,$2F,$98,$FE,$AF,$F6,$DF,$D2
	.dc.b	$7E,$63,$FA,$A1,$FD,$5B,$E6,$30,$E0,$7C,$6F,$FB,$0F,$F1,$2F,$EA
	.dc.b	$FA,$7D,$D0,$E5,$0F,$99,$FE,$B9,$E8,$66,$3F,$AB,$FC,$B7,$FB,$2B
	.dc.b	$F3,$7C,$7A,$BF,$D6,$7E,$66,$CF,$EA,$6B,$F1,$5C,$3D,$9F,$EC,$66
	.dc.b	$32,$FF,$61,$FD,$BF,$82,$75,$3D,$BE,$63,$FB,$47,$D7,$E6,$FF,$D8
	.dc.b	$FF,$35,$DA,$7F,$AC,$FD,$BF,$97,$E5,$DF,$35,$FE,$C8,$E2,$7C,$06
	.dc.b	$63,$1F,$37,$7C,$DF,$85,$B7,$83,$0E,$3E,$96,$35,$FE,$AD,$F3,$5F
	.dc.b	$EB,$FE,$0F,$11,$FE,$C3,$7F,$AC,$5F,$1C,$BC,$1F,$65,$D7,$E7,$3F
	.dc.b	$D8,$FF,$9A,$7C,$9F,$9C,$FF,$62,$7E,$37,$7C,$AF,$F5,$AF,$CB,$DF
	.dc.b	$2F,$C1,$AF,$F5,$5F,$16,$BF,$EB,$E8,$7C,$D3,$EF,$5A,$BE,$7B,$CB
	.dc.b	$7B,$AD,$4B,$ED,$B3,$AF,$FB,$57,$E7,$FC,$15,$AF,$B6,$FF,$63,$72
	.dc.b	$FE,$0B,$1E,$59,$F6,$97,$E9,$FC,$23,$1F,$98,$FE,$17,$D1,$FF,$64
	.dc.b	$FE,$5B,$7E,$7F,$FA,$97,$EE,$37,$C6,$F5,$D7,$F5,$FF,$9F,$B5,$FD
	.dc.b	$7B,$4E,$10,$FA,$2F,$EB,$DF,$45,$FD,$77,$BB,$F0,$B8,$7F,$B5,$BE
	.dc.b	$EB,$75,$FE,$C8,$FA,$2F,$CB,$7C,$F7,$F5,$CB,$FC,$5B,$FE,$C9,$F7
	.dc.b	$73,$F4,$2F,$A1,$F9,$B3,$FD,$73,$5F,$92,$FF,$B1,$2F,$8D,$6F,$E8
	.dc.b	$3F,$C6,$FF,$42,$FC,$07,$73,$FE,$D2,$FC,$2B,$AE,$5F,$D8,$FE,$F3
	.dc.b	$58,$FE,$73,$CF,$FE,$B3,$F7,$1F,$EC,$1F,$6F,$F5,$91,$8F,$A2,$7E
	.dc.b	$2F,$8F,$BC,$FF,$61,$FD,$27,$FA,$FF,$F4,$BC,$7D,$E7,$FA,$F6,$DF
	.dc.b	$87,$7E,$D7,$E2,$7F,$D7,$BF,$6D,$7E,$93,$FB,$17,$CA,$5F,$84,$FC
	.dc.b	$5D,$7F,$61,$EA,$FC,$3A,$AF,$F5,$D7,$06,$1F,$F5,$CC,$F3,$FA,$5F
	.dc.b	$EC,$BF,$4C,$FA,$87,$C6,$E7,$E4,$5F,$51,$FD,$A6,$5F,$51,$FC,$27
	.dc.b	$E9,$7F,$B1,$F5,$7E,$F9,$FF,$5F,$FD,$3E,$7E,$9D,$F4,$1F,$A0,$9F
	.dc.b	$EC,$7F,$6F,$67,$FD,$85,$E1,$FE,$7A,$7F,$B0,$E1,$CE,$7F,$B6,$F9
	.dc.b	$FC,$F0,$FF,$5F,$F9,$0F,$D0,$7F,$8D,$7C,$3F,$FA,$EF,$DF,$BF,$49
	.dc.b	$FE,$BA,$F1,$AF,$CB,$7E,$C2,$BF,$D7,$1D,$1F,$AC,$7F,$B0,$EB,$FC
	.dc.b	$59,$FF,$5C,$47,$2F,$A4,$FF,$5B,$7D,$CD,$FA,$6F,$EC,$BF,$4B,$FD
	.dc.b	$6F,$7C,$7C,$7F,$F6,$0F,$22,$CF,$F5,$FF,$C1,$3F,$97,$71,$3E,$8B
	.dc.b	$FD,$6F,$F8,$86,$1F,$F5,$CF,$E2,$57,$93,$F8,$2D,$7F,$5E,$DF,$F8
	.dc.b	$0B,$C6,$3F,$D6,$FF,$79,$8B,$FE,$B9,$5C,$7E,$A3,$FB,$17,$D3,$BE
	.dc.b	$F3,$FC,$4F,$CB,$FC,$CB,$2E,$55,$A9,$7D,$3F,$F4,$3F,$90,$FE,$C1
	.dc.b	$7F,$CE,$1F,$F5,$C3,$7E,$03,$FD,$71,$97,$F8,$6A,$31,$F8,$3F,$F1
	.dc.b	$CA,$FE,$C3,$3C,$E9,$E6,$FF,$AD,$FB,$FF,$8D,$D7,$FB,$1A,$F5,$3F
	.dc.b	$EB,$BF,$BC,$7E,$21,$FD,$57,$E9,$BF,$B0,$0F,$D0,$70,$FF,$63,$FD
	.dc.b	$6D,$AB,$DA,$FE,$C7,$B7,$F8,$7F,$E4,$33,$18,$8F,$F1,$8F,$D4,$61
	.dc.b	$FF,$56,$79,$2F,$B3,$7D,$4E,$67,$FA,$C7,$DF,$2A,$57,$FA,$F3,$2F
	.dc.b	$05,$7F,$6E,$F6,$7F,$D7,$AF,$F2,$F8,$CB,$FD,$84,$FE,$D3,$97,$F6
	.dc.b	$67,$B4,$BE,$A5,$F4,$FF,$D9,$A7,$31,$FE,$BE,$BE,$75,$FE,$CD,$FA
	.dc.b	$7F,$37,$D2,$FF,$66,$F1,$FF,$D5,$DF,$0B,$8F,$A4,$FE,$CB,$87,$8A
	.dc.b	$B5,$97,$88,$FF,$02,$BF,$D9,$9C,$1F,$EC,$7F,$A5,$FE,$CC,$F3,$C1
	.dc.b	$F4,$BF,$EB,$EC,$5E,$4F,$F7,$FD,$FF,$B6,$C7,$F8,$5B,$E8,$38,$57
	.dc.b	$F6,$8C,$FE,$1F,$CF,$FF,$68,$6F,$D1,$FF,$68,$6F,$95,$FD,$EB,$1F
	.dc.b	$EB,$15,$FE,$80,$F0,$71,$AF,$FA,$F3,$E7,$FF,$B1,$FD,$17,$86,$31
	.dc.b	$F3,$DF,$DA,$BE,$8B,$11,$FD,$83,$67,$D0,$BE,$8B,$EF,$67,$FE,$8C
	.dc.b	$D7,$FD,$85,$B7,$8B,$4F,$ED,$5F,$45,$AE,$0F,$F5,$1C,$7F,$6D,$FA
	.dc.b	$2F,$C3,$79,$7F,$D9,$AF,$EF,$BF,$D8,$FE,$73,$FA,$FF,$DD,$BE,$8B
	.dc.b	$7F,$45,$FE,$C7,$E1,$FC,$37,$37,$F7,$5C,$3F,$B5,$47,$D2,$79,$7D
	.dc.b	$27,$FB,$13,$EF,$3D,$71,$8F,$A3,$FE,$D5,$7C,$70,$FF,$63,$7D,$1F
	.dc.b	$6F,$A4,$FF,$62,$ED,$F9,$2F,$A3,$FE,$D5,$7F,$EB,$EF,$A3,$FE,$C3
	.dc.b	$F1,$0F,$6F,$F4,$14,$FA,$2F,$ED,$9C,$1F,$84,$8F,$ED,$8E,$A3,$FB
	.dc.b	$57,$92,$FA,$3F,$F6,$3C,$E7,$E8,$BF,$D9,$5F,$1F,$F4,$B0,$E9,$1F
	.dc.b	$DA,$FE,$8F,$5A,$78,$AB,$F6,$1E,$3F,$F6,$9F,$9B,$BB,$FF,$68,$FA
	.dc.b	$37,$D0,$7F,$6C,$E2,$FD,$6D,$FF,$6F,$FA,$2D,$FD,$17,$F6,$BF,$A2
	.dc.b	$FD,$D7,$C0,$FF,$6A,$BF,$D9,$BF,$D9,$7E,$7F,$FB,$0F,$CD,$BE,$87
	.dc.b	$7F,$42,$E7,$FC,$7F,$37,$F1,$A8,$5E,$30,$E5,$F3,$FF,$D9,$BE,$87
	.dc.b	$FC,$DD,$7C,$AB,$FD,$9B,$C5,$FE,$CF,$AD,$67,$FD,$BB,$4F,$1E,$5F
	.dc.b	$EC,$E7,$F8,$3F,$BD,$FD,$B2,$FF,$C7,$0F,$F6,$C8,$CF,$CE,$7F,$6C
	.dc.b	$CB,$F6,$F3,$FD,$B1,$B1,$FD,$B1,$FE,$35,$1F,$DB,$2B,$9C,$FF,$6C
	.dc.b	$AE,$7F,$2B,$FD,$B3,$E7,$3F,$90,$7C,$D7,$F6,$B9,$FE,$32,$BE,$73
	.dc.b	$FB,$64,$61,$FF,$66,$E5,$FD,$0F,$1F,$F2,$75,$7F,$6E,$8E,$72,$F9
	.dc.b	$B7,$CD,$7F,$6A,$FC,$07,$F6,$B7,$FD,$C3,$1E,$95,$FD,$C1,$B5,$FD
	.dc.b	$C3,$ED,$3F,$C1,$71,$8F,$80,$FE,$E0,$DF,$9B,$FE,$E0,$FF,$30,$F1
	.dc.b	$FF,$AF,$FD,$A7,$FB,$13,$E5,$9F,$39,$FC,$F6,$73,$F3,$3F,$D9,$FE
	.dc.b	$73,$1F,$37,$F9,$AE,$2F,$61,$FE,$AA,$AF,$EE,$8E,$BD,$DF,$EE,$6D
	.dc.b	$F9,$BF,$EE,$3F,$69,$FC,$42,$BF,$D9,$DA,$F8,$F0,$61,$FE,$CF,$8F
	.dc.b	$E8,$BF,$0F,$FD,$C3,$9F,$FC,$EC,$FF,$70,$6F,$AF,$FD,$C3,$3F,$E0
	.dc.b	$E7,$FA,$F2,$7F,$DA,$2E,$03,$FD,$CA,$BA,$CF,$F7,$2F,$A5,$FF,$61
	.dc.b	$7E,$02,$F9,$DF,$E4,$B7,$FD,$CE,$F7,$E2,$FF,$73,$F9,$BF,$EC,$5F
	.dc.b	$35,$FD,$CB,$E6,$EB,$FD,$C6,$9F,$96,$FF,$6D,$D7,$FD,$1D,$F3,$7F
	.dc.b	$DD,$23,$11,$FE,$D2,$F9,$BC,$3F,$DC,$3E,$67,$FB,$7D,$F2,$8C,$7D
	.dc.b	$96,$9F,$F6,$AF,$9B,$FE,$BC,$C3,$FD,$35,$E7,$7F,$67,$F9,$AC,$57
	.dc.b	$F7,$AE,$1F,$F6,$85,$6B,$E6,$BF,$B5,$FC,$DF,$F7,$9F,$35,$F6,$5A
	.dc.b	$C3,$C5,$F6,$5D,$F3,$FE,$D9,$B7,$8E,$7F,$B1,$7B,$1F,$EF,$03,$C9
	.dc.b	$F3,$7F,$DB,$FE,$67,$FB,$17,$D9,$3E,$CB,$3F,$3C,$F3,$3C,$B8,$B0
	.dc.b	$FE,$DC,$FF,$8A,$78,$7F,$DA,$7F,$85,$FF,$81,$DF,$ED,$3F,$3D,$E5
	.dc.b	$9F,$0E,$5F,$EC,$4D,$7C,$98,$78,$7E,$7F,$FD,$C7,$2F,$B3,$F9,$1F
	.dc.b	$35,$FD,$B7,$2F,$ED,$3F,$33,$FD,$B5,$BE,$47,$92,$7F,$B6,$67,$FE
	.dc.b	$FC,$9C,$FC,$CF,$F6,$C8,$E7,$7F,$D8,$3E,$7F,$FB,$64,$63,$0F,$F6
	.dc.b	$CC,$7F,$B2,$FE,$73,$17,$FE,$D1,$FC,$06,$CF,$FB,$3F,$E7,$6F,$CF
	.dc.b	$FF,$72,$F9,$EF,$BD,$E7,$7F,$69,$D7,$FA,$D7,$CB,$7F,$B1,$FE,$6D
	.dc.b	$F3,$4F,$9D,$FE,$D5,$CF,$F5,$70,$7C,$DF,$F6,$BE,$2F,$91,$1F,$DA
	.dc.b	$9F,$9C,$03,$FD,$A7,$E7,$37,$2C,$BC,$18,$FF,$B5,$73,$F1,$FC,$F7
	.dc.b	$F7,$AF,$9D,$F3,$DE,$E3,$FD,$AD,$9F,$E1,$FE,$6B,$FD,$AD,$7F,$DB
	.dc.b	$7C,$4F,$EC,$9F,$3B,$FD,$B6,$B5,$F3,$BF,$DB,$37,$F2,$8C,$3F,$B6
	.dc.b	$CE,$BE,$71,$F3,$7F,$ED,$6E,$BF,$C2,$57,$F6,$A8,$FF,$6C,$FC,$DE
	.dc.b	$31,$FF,$6C,$61,$C6,$7F,$B6,$3F,$9B,$F2,$7F,$B6,$37,$DD,$FE,$D9
	.dc.b	$F3,$3F,$DA,$F7,$F8,$DA,$AF,$FB,$65,$EF,$E6,$3F,$B6,$3F,$A5,$F7
	.dc.b	$3F,$B6,$5F,$F5,$DF,$AF,$FF,$63,$7C,$C7,$EE,$67,$FD,$A5,$C7,$FA
	.dc.b	$27,$FD,$AD,$96,$7E,$57,$FD,$AA,$DF,$98,$FF,$69,$FE,$0D,$FA,$1A
	.dc.b	$FE,$DB,$8F,$FB,$47,$E1,$35,$F3,$3F,$DA,$BE,$C3,$F8,$37,$B1,$FE
	.dc.b	$D2,$DB,$FA,$EF,$CC,$7F,$B1,$BE,$C3,$3F,$35,$FD,$9F,$6F,$9F,$7F
	.dc.b	$D9,$DB,$1F,$D9,$F8,$BF,$80,$4F,$FB,$4B,$0E,$7F,$7A,$FB,$D7,$E0
	.dc.b	$DF,$E2,$7F,$81,$EB,$A7,$F6,$9C,$7F,$84,$7C,$D7,$F6,$9C,$7F,$B1
	.dc.b	$FC,$E7,$F6,$86,$FD,$97,$F0,$49,$FE,$CD,$5C,$2B,$FB,$37,$CD,$FF
	.dc.b	$65,$DF,$F9,$15,$7F,$65,$C3,$9F,$CE,$7F,$65,$6D,$7F,$65,$CB,$78
	.dc.b	$72,$F7,$FF,$B2,$D6,$97,$CD,$FF,$66,$8F,$BF,$F2,$9F,$D8,$BE,$6B
	.dc.b	$FB,$2E,$BF,$C8,$63,$1F,$31,$FC,$CD,$FF,$69,$6D,$FC,$1C,$FE,$C9
	.dc.b	$F2,$DF,$D4,$1C,$1F,$E4,$B7,$FA,$FF,$CC,$7F,$B5,$3E,$C3,$F8,$E5
	.dc.b	$7F,$70,$F9,$9B,$5A,$F9,$CF,$EE,$DF,$03,$DE,$31,$F3,$7F,$ED,$BF
	.dc.b	$B0,$FE,$89,$F3,$7F,$ED,$6E,$0F,$EC,$BE,$27,$FB,$27,$EF,$33,$EE
	.dc.b	$7F,$67,$E0,$FE,$55,$7F,$D9,$9B,$1F,$D8,$FE,$99,$F3,$CF,$BC,$CE
	.dc.b	$3F,$ED,$7B,$E6,$FF,$B5,$7E,$CA,$F9,$3F,$DA,$BE,$CB,$FD,$AF,$D1
	.dc.b	$FD,$1F,$E7,$FF,$B3,$35,$FE,$CD,$FA,$1F,$F6,$2C,$5F,$41,$FE,$D8
	.dc.b	$8C,$7D,$05,$9F,$EC,$9F,$65,$5F,$F6,$8F,$97,$FE,$D0,$FA,$0C,$7A
	.dc.b	$BF,$EC,$E9,$FE,$39,$9B,$1F,$EC,$1F,$3F,$FB,$A9,$FF,$69,$47,$FB
	.dc.b	$3E,$75,$1F,$EE,$0F,$9D,$E3,$3C,$3E,$7B,$FB,$37,$D0,$BE,$7F,$BF
	.dc.b	$D0,$7F,$6A,$D7,$FD,$C3,$F6,$5E,$5F,$9C,$FE,$A5,$F0,$1F,$D9,$75
	.dc.b	$EB,$AF,$F5,$8F,$A0,$D7,$D0,$7F,$63,$8F,$21,$8F,$4E,$BA,$F3,$FF
	.dc.b	$D9,$BF,$3D,$FE,$D0,$C3,$9E,$5D,$BE,$7F,$CF,$7F,$ED,$8C,$7F,$B7
	.dc.b	$7D,$0F,$F7,$2F,$98,$E4,$FF,$B3,$F9,$3F,$AD,$FC,$D2,$7F,$D8,$3F
	.dc.b	$77,$F3,$7E,$7F,$FD,$81,$F3,$CA,$D7,$D8,$BE,$73,$5E,$8F,$F9,$0F
	.dc.b	$5F,$96,$7F,$4B,$C3,$FB,$4B,$F3,$F8,$3E,$0C,$74,$C3,$C1,$F6,$7E
	.dc.b	$8A,$FF,$69,$7D,$9D,$AD,$6B,$FE,$D4,$FA,$2E,$FF,$2D,$FE,$D2,$E0
	.dc.b	$F4,$CF,$FB,$63,$E8,$73,$C1,$FE,$D1,$D7,$E7,$FA,$1F,$ED,$1B,$FE
	.dc.b	$AF,$C1,$FD,$A3,$E8,$3F,$B3,$C7,$FB,$17,$E1,$7E,$7D,$F0,$F9,$CF
	.dc.b	$EC,$1C,$1F,$52,$31,$F3,$FF,$D8,$BE,$83,$D1,$9F,$FB,$3F,$DD,$B8
	.dc.b	$7F,$B3,$AF,$7F,$3D,$FD,$8B,$3F,$EC,$1F,$64,$F9,$F7,$E0,$9F,$9B
	.dc.b	$F9,$FF,$EB,$1E,$7F,$F6,$4F,$83,$C7,$CF,$7F,$63,$E1,$F9,$A7,$F6
	.dc.b	$35,$F0,$5E,$8C,$FA,$F4,$FC,$BC,$79,$E8,$F9,$FC,$61,$FD,$8A,$FC
	.dc.b	$7A,$E7,$7F,$CA,$F9,$7F,$D8,$A7,$FA,$FA,$F9,$EF,$EC,$7F,$75,$FB
	.dc.b	$18,$FE,$C3,$3A,$9F,$EC,$FF,$63,$FB,$5F,$7F,$FD,$99,$3F,$D8,$7E
	.dc.b	$A9,$F6,$3F,$EC,$BC,$BF,$47,$5F,$D8,$B8,$BE,$75,$FF,$6E,$8F,$EC
	.dc.b	$73,$AF,$9F,$FE,$C6,$DF,$9F,$FE,$C7,$7D,$E7,$FD,$A3,$F3,$BF,$EC
	.dc.b	$AF,$83,$C5,$7F,$B6,$BE,$77,$F8,$47,$B0,$F9,$C7,$CB,$3E,$73,$FA
	.dc.b	$D7,$CD,$FF,$BF,$FF,$01,$DD,$6A,$73,$DC,$F9,$EC,$E1,$F4,$77,$FF
	.dc.b	$15,$65,$FD,$A5,$FA,$EF,$78,$ED,$F3,$DF,$D9,$BB,$3F,$41,$39,$C7
	.dc.b	$C3,$F8,$3E,$33,$FF,$69,$72,$FF,$6E,$C3,$FB,$6E,$8F,$B3,$FE,$45
	.dc.b	$EF,$FF,$B4,$EB,$FA,$D7,$CE,$3E,$CB,$F1,$DD,$1F,$C4,$2B,$FD,$AD
	.dc.b	$F4,$3E,$7E,$14,$7F,$6E,$C3,$8F,$D9,$FE,$CE,$3F,$DA,$11,$D3,$07
	.dc.b	$DD,$78,$3E,$87,$FB,$56,$7D,$F9,$31,$F4,$7F,$ED,$AF,$86,$FF,$68
	.dc.b	$E3,$C7,$E1,$3F,$9C,$70,$FF,$6B,$FB,$4B,$F3,$DF,$D6,$BE,$8D,$F4
	.dc.b	$78,$F3,$FC,$D7,$FE,$CC,$E5,$FE,$52,$FF,$B5,$3E,$8B,$F8,$43,$FD
	.dc.b	$9F,$97,$87,$D1,$BE,$8F,$FB,$3D,$F0,$CF,$B7,$D1,$FF,$49,$FA,$1F
	.dc.b	$F2,$FF,$94,$F2,$3F,$AC,$79,$39,$F6,$FF,$B1,$7D,$03,$E7,$9F,$45
	.dc.b	$9C,$3E,$56,$3F,$DA,$77,$C6,$5F,$EC,$AF,$A1,$FE,$39,$3D,$A3,$FD
	.dc.b	$9B,$F0,$58,$CF,$AE,$4F,$A1,$E3,$F4,$5F,$DC,$23,$FB,$4F,$5F,$F9
	.dc.b	$43,$E8,$BF,$B4,$B5,$FF,$67,$7E,$41,$FD,$A6,$1C,$88,$FE,$D9,$F4
	.dc.b	$98,$7F,$AD,$79,$C1,$F3,$BF,$EC,$E8,$CC,$63,$E8,$78,$71,$7F,$A1
	.dc.b	$7E,$0F,$FB,$63,$80,$FF,$B3,$DF,$E4,$D5,$FD,$AC,$E7,$F4,$1F,$DC
	.dc.b	$9E,$B8,$3E,$0B,$95,$FF,$B5,$FE,$83,$FB,$6F,$E6,$BD,$7E,$73,$FB
	.dc.b	$4F,$D9,$BF,$38,$C7,$CF,$7F,$5C,$F9,$FF,$F5,$EF,$E4,$98,$F3,$7F
	.dc.b	$9C,$ED,$FE,$76,$C7,$FD,$97,$C1,$FE,$78,$E4,$F8,$8F,$F6,$5A,$FE
	.dc.b	$D7,$F6,$D9,$BF,$F6,$6C,$F3,$FA,$1F,$ED,$17,$C2,$3F,$D9,$FF,$6D
	.dc.b	$F0,$7E,$73,$FA,$AF,$D0,$BF,$73,$C7,$D1,$7F,$B0,$FE,$79,$2F,$77
	.dc.b	$3E,$97,$9B,$2F,$EC,$D7,$FE,$45,$7F,$B3,$6B,$FD,$73,$E8,$7E,$44
	.dc.b	$6A,$BF,$D9,$78,$BE,$87,$32,$D3,$C7,$F4,$3D,$BE,$83,$FD,$9D,$C1
	.dc.b	$E9,$9F,$3F,$A7,$9F,$A2,$FE,$F7,$FA,$CF,$DB,$F2,$FF,$D8,$FF,$41
	.dc.b	$FE,$BE,$FA,$17,$E0,$99,$FA,$0F,$F5,$E7,$D1,$FF,$69,$FB,$FC,$7A
	.dc.b	$5E,$BE,$1F,$F6,$4D,$7F,$68,$7D,$67,$FB,$37,$2F,$F6,$63,$CF,$2F
	.dc.b	$F6,$5F,$D1,$FA,$A7,$C3,$F4,$7F,$E0,$DF,$B8,$FF,$6E,$79,$9B,$8C
	.dc.b	$6D,$FC,$C7,$8F,$EA,$7D,$17,$F5,$EF,$A1,$7C,$9F,$E5,$63,$1F,$8E
	.dc.b	$5F,$A3,$FE,$D5,$BF,$C8,$9F,$F6,$5F,$7F,$F6,$88,$D7,$DC,$7F,$B3
	.dc.b	$3D,$1F,$CB,$7C,$C6,$F0,$FF,$64,$DF,$F6,$88,$FD,$17,$D1,$FF,$65
	.dc.b	$FB,$67,$E1,$4F,$C0,$DF,$80,$BE,$E9,$F4,$DC,$3E,$A3,$FD,$9B,$F6
	.dc.b	$77,$1E,$38,$70,$B7,$E6,$79,$CB,$A6,$5C,$FC,$4F,$D9,$7C,$57,$FB
	.dc.b	$0B,$4F,$ED,$BF,$83,$63,$EA,$7F,$B1,$FD,$D7,$FB,$43,$AF,$F3,$57
	.dc.b	$FD,$BF,$9F,$EA,$87,$FB,$7F,$27,$F7,$7F,$89,$FC,$6C,$75,$FB,$BB
	.dc.b	$8F,$4C,$FF,$BD,$FE,$E7,$DF,$EB,$1F,$26,$D7,$DF,$F5,$AF,$DA,$FF
	.dc.b	$84,$7C,$5F,$F5,$FF,$A8,$FE,$B7,$F1,$78,$FA,$9F,$ED,$D1,$FD,$7B
	.dc.b	$D2,$C7,$D1,$7E,$36,$FF,$B6,$72,$7E,$81,$E5,$F5,$BF,$E1,$D8,$DF
	.dc.b	$D5,$FF,$6A,$6E,$1F,$DA,$3D,$3B,$4F,$9E,$FE,$D5,$D3,$FB,$DF,$A9
	.dc.b	$7D,$3F,$F5,$CF,$A9,$FF,$6E,$4E,$63,$1F,$8F,$7F,$21,$AF,$ED,$79
	.dc.b	$FF,$20,$07,$F6,$C9,$7D,$47,$E9,$67,$AB,$FD,$97,$EA,$7F,$82,$57
	.dc.b	$3A,$FE,$D7,$F8,$1F,$EF,$2F,$7F,$7E,$F8,$FF,$F6,$AE,$9F,$DC,$B4
	.dc.b	$F2,$67,$CE,$3F,$C3,$71,$8A,$E1,$9F,$F8,$0F,$E5,$7B,$5E,$FE,$B7
	.dc.b	$F9,$D6,$FF,$DC,$F0,$FE,$3C,$FF,$B7,$1B,$F4,$DF,$E7,$DD,$FF,$87
	.dc.b	$CF,$F7,$3F,$17,$FD,$CB,$3F,$ED,$2F,$A8,$CB,$FE,$D4,$F7,$EF,$C5
	.dc.b	$BE,$A3,$FB,$A7,$DA,$3E,$AD,$E3,$78,$B0,$FF,$01,$7E,$6D,$FD,$37
	.dc.b	$E9,$FF,$B4,$79,$BF,$D8,$FE,$63,$1F,$0D,$FD,$A3,$DA,$BE,$87,$F1
	.dc.b	$1B,$FE,$E5,$C9,$FC,$DD,$E3,$F5,$5F,$ED,$9E,$5D,$FD,$47,$F7,$26
	.dc.b	$E7,$FD,$BF,$EA,$2D,$7F,$7F,$FC,$1B,$FA,$F7,$D4,$08,$43,$E7,$F4
	.dc.b	$7F,$73,$70,$7D,$57,$0F,$0F,$AF,$FF,$69,$FC,$27,$F7,$7D,$4F,$F7
	.dc.b	$66,$E7,$F8,$AC,$FF,$D8,$33,$C2,$3F,$DC,$33,$A7,$FD,$C3,$E5,$FF
	.dc.b	$77,$8C,$7C,$27,$F7,$66,$FD,$EB,$E0,$BF,$D3,$B8,$7F,$B7,$CF,$EE
	.dc.b	$B8,$7F,$82,$1F,$EE,$BF,$4F,$63,$FC,$01,$5F,$ED,$DC,$7F,$C0,$FF
	.dc.b	$4D,$99,$7D,$17,$F7,$6F,$A6,$C3,$FD,$E3,$E9,$2F,$D3,$7F,$B7,$6B
	.dc.b	$FD,$85,$9F,$F7,$58,$FF,$28,$C7,$FA,$B6,$31,$E6,$7F,$74,$8F,$80
	.dc.b	$76,$9F,$F0,$07,$53,$8B,$38,$FF,$81,$BE,$7B,$BE,$1F,$EE,$6A,$79
	.dc.b	$4F,$A4,$F4,$7C,$E7,$F7,$8F,$A4,$BF,$45,$FD,$DF,$E9,$3F,$BB,$73
	.dc.b	$FC,$DA,$4F,$E8,$77,$DF,$81,$F2,$9F,$EE,$9F,$41,$FE,$E5,$D9,$F4
	.dc.b	$7E,$8C,$3F,$DB,$3F,$47,$71,$ED,$F4,$7E,$0F,$3B,$FB,$DE,$AF,$63
	.dc.b	$19,$F3,$8F,$F2,$D6,$DE,$3C,$3C,$DF,$45,$F4,$7C,$9F,$EC,$FF,$45
	.dc.b	$FE,$DB,$F8,$2C,$7D,$0F,$F7,$6F,$A0,$7C,$F3,$E8,$B3,$F0,$7E,$9E
	.dc.b	$6F,$EE,$FF,$45,$73,$FE,$EB,$B7,$E3,$5F,$F6,$E4,$F4,$9F,$F0,$27
	.dc.b	$D1,$5F,$A2,$FE,$E7,$F6,$7F,$83,$FA,$0F,$EE,$5F,$45,$7E,$87,$F3
	.dc.b	$BF,$81,$FF,$91,$BC,$EF,$F7,$07,$D1,$E7,$CA,$F1,$60,$BF,$F6,$DE
	.dc.b	$BF,$83,$3F,$95,$C7,$FB,$6B,$B3,$F0,$6F,$FB,$73,$E8,$ED,$76,$F4
	.dc.b	$BF,$C9,$9F,$48,$FC,$23,$B6,$BE,$18,$FF,$6F,$67,$FE,$D7,$FA,$4F
	.dc.b	$ED,$FF,$41,$FD,$D7,$E7,$3F,$39,$E6,$FF,$7B,$FA,$1F,$ED,$FF,$47
	.dc.b	$FA,$9E,$AF,$D8,$4E,$7E,$93,$FB,$B7,$D2,$70,$FA,$56,$7D,$30,$7D
	.dc.b	$2F,$E9,$FE,$17,$FB,$B7,$CF,$BE,$97,$1F,$49,$FD,$AE,$3F,$DC,$9C
	.dc.b	$3F,$EE,$18,$FF,$36,$F5,$7C,$33,$FB,$B2,$FA,$4F,$B5,$87,$F8,$07
	.dc.b	$E0,$FF,$4E,$FF,$77,$AE,$B3,$FD,$D3,$E8,$FF,$B9,$FC,$76,$3E,$93
	.dc.b	$FB,$B7,$37,$FB,$D3,$E9,$3F,$79,$F0,$BF,$DB,$3E,$59,$F4,$98,$FA
	.dc.b	$4F,$EE,$BF,$43,$FD,$A7,$D3,$C7,$D2,$7E,$2F,$0F,$EE,$BB,$E3,$E9
	.dc.b	$3F,$DD,$7E,$37,$EE,$A7,$FB,$BF,$97,$72,$E7,$A3,$D1,$E1,$F4,$9F
	.dc.b	$ED,$EF,$B5,$FE,$1D,$F4,$7F,$D8,$BE,$23,$FB,$0C,$E3,$E9,$3F,$B9
	.dc.b	$F8,$DF,$C8,$EB,$FB,$9E,$FF,$C3,$8F,$EE,$6F,$F7,$07,$E5,$3C,$FE
	.dc.b	$8B,$FB,$8C,$F1,$AF,$EE,$35,$D7,$E8,$BF,$B6,$7D,$AD,$FA,$4F,$ED
	.dc.b	$7F,$4B,$FD,$BB,$C2,$FF,$72,$7C,$76,$76,$FE,$EB,$F4,$BF,$C9,$78
	.dc.b	$BF,$B6,$57,$F6,$6F,$5B,$FB,$17,$D4,$3E,$8F,$FB,$17,$DA,$6F,$E8
	.dc.b	$FF,$87,$71,$7F,$3D,$C7,$FD,$C1,$8F,$1F,$A5,$FE,$D3,$F4,$FF,$C9
	.dc.b	$27,$FD,$C1,$5B,$BF,$F7,$57,$8D,$CF,$4E,$F9,$FF,$6F,$FA,$5F,$E7
	.dc.b	$1F,$29,$FD,$93,$EC,$5F,$4D,$8F,$87,$FE,$D9,$9F,$F4,$DA,$FF,$72
	.dc.b	$F4,$FE,$45,$FE,$DB,$C1,$FE,$1E,$84,$7F,$6E,$FB,$2F,$E6,$EF,$F7
	.dc.b	$0F,$02,$E1,$FE,$DE,$FA,$6E,$DE,$87,$FB,$BF,$D1,$FF,$6F,$7D,$2F
	.dc.b	$F4,$BF,$A4,$FE,$D2,$7F,$6F,$F2,$F0,$E3,$C0,$FF,$6E,$F1,$FF,$69
	.dc.b	$3F,$B4,$7A,$1F,$DB,$A7,$FB,$AF,$EB,$3F,$EE,$28,$C4,$7F,$79,$F9
	.dc.b	$AF,$F7,$2F,$E9,$5E,$9C,$FD,$5C,$53,$F3,$3C,$BF,$DC,$4C,$9F,$E3
	.dc.b	$CF,$D5,$F2,$FF,$AB,$FE,$5F,$FA,$47,$56,$3E,$77,$FD,$87,$E0,$D9
	.dc.b	$FB,$79,$7C,$79,$FE,$97,$CF,$D3,$0D,$76,$FF,$AF,$35,$FF,$24,$F5
	.dc.b	$B0,$E1,$8F,$FB,$2F,$E7,$73,$18,$BF,$E9,$3F,$3B,$CB,$E7,$3F,$D7
	.dc.b	$F3,$97,$FA,$57,$E8,$39,$F8,$27,$CF,$FF,$B5,$3F,$50,$F5,$FC,$57
	.dc.b	$F1,$BF,$8C,$FF,$4B,$7C,$16,$3E,$27,$F8,$DF,$C6,$7F,$22,$F9,$CC
	.dc.b	$7C,$5F,$A7,$8F,$CD,$CB,$FE,$60,$C9,$83,$C1,$D7,$0F,$F2,$FC,$BF
	.dc.b	$D3,$3F,$65,$EA,$C3,$85,$F6,$F8,$EF,$F5,$3F,$E0,$B9,$96,$5E,$5F
	.dc.b	$8E,$7E,$E9,$E2,$CB,$F9,$AF,$E5,$17,$1F,$F5,$47,$E5,$97,$8F,$FD
	.dc.b	$61,$F9,$5F,$D8,$F8,$2F,$E0,$1E,$F3,$E6,$71,$F0,$5F,$E7,$FF,$A5
	.dc.b	$BE,$FF,$A3,$87,$F8,$4C,$FF,$9E,$9F,$C1,$F3,$FC,$07,$FC,$FD,$F0
	.dc.b	$7F,$E7,$BE,$6F,$46,$7F,$C3,$7E,$9A,$E1,$FE,$8B,$E9,$EF,$7F,$E7
	.dc.b	$7F,$4A,$FC,$27,$F9,$DB,$E9,$5F,$92,$63,$E1,$5F,$0E,$97,$DD,$7E
	.dc.b	$4B,$D5,$FD,$38,$F5,$F0,$FE,$AB,$CF,$71,$BE,$5F,$F4,$1C,$9F,$91
	.dc.b	$FD,$17,$31,$88,$FC,$FF,$E1,$18,$7F,$44,$7E,$9B,$E0,$BF,$2B,$3F
	.dc.b	$8D,$F3,$7F,$17,$D4,$FC,$03,$85,$FE,$97,$D2,$FF,$0E,$FE,$01,$8F
	.dc.b	$5F,$FC,$57,$1F,$9F,$9C,$7A,$DF,$9D,$6C,$7E,$73,$6F,$44,$7F,$8B
	.dc.b	$B2,$EF,$D9,$AF,$53,$FC,$4D,$F9,$9D,$9F,$1F,$E5,$3E,$2F,$37,$F1
	.dc.b	$1F,$73,$7D,$0F,$C7,$79,$3F,$EF,$3E,$1C,$7A,$5F,$E0,$2F,$1B,$DB
	.dc.b	$B7,$E5,$23,$FC,$0B,$E8,$E2,$3F,$27,$87,$2F,$47,$F2,$7C,$3E,$69
	.dc.b	$CD,$FF,$80,$7C,$FE,$BE,$87,$FB,$E3,$F3,$97,$E2,$38,$F4,$9B,$A5
	.dc.b	$F9,$1F,$CA,$F3,$BF,$2F,$F8,$8D,$F3,$FF,$DA,$9C,$37,$CE,$FC,$86
	.dc.b	$F7,$CD,$FF,$5F,$F6,$79,$F1,$FC,$FF,$9F,$9F,$37,$FD,$93,$E7,$DF
	.dc.b	$3B,$F1,$9F,$86,$5F,$3D,$FB,$AF,$DA,$F3,$BF,$1B,$F9,$C5,$7F,$9A
	.dc.b	$FE,$B9,$5F,$F5,$17,$DD,$FF,$18,$F8,$8F,$C5,$70,$FF,$2C,$F4,$7F
	.dc.b	$23,$18,$8F,$E9,$1E,$8E,$1F,$C6,$9F,$D2,$FD,$1F,$E4,$3E,$B7,$F3
	.dc.b	$8C,$59,$FE,$6A,$31,$9F,$E7,$3E,$67,$A6,$1F,$D4,$FC,$35,$7F,$40
	.dc.b	$D7,$F9,$5D,$6A,$9A,$F1,$7C,$27,$92,$FF,$D0,$EE,$0F,$F5,$9B,$87
	.dc.b	$F3,$9E,$6F,$F9,$D7,$C8,$E5,$18,$F3,$3F,$47,$E8,$E3,$D1,$FC,$FD
	.dc.b	$FF,$A1,$7D,$4F,$1D,$FE,$6F,$E0,$5F,$09,$88,$E5,$0F,$55,$F0,$9C
	.dc.b	$FD,$6F,$F4,$B7,$83,$7D,$6F,$F4,$4F,$AD,$74,$FF,$44,$7A,$DE,$4F
	.dc.b	$5B,$F4,$3C,$7E,$97,$F9,$57,$1E,$7D,$4F,$D1,$FA,$BE,$19,$FE,$85
	.dc.b	$E5,$59,$FC,$F7,$BF,$BF,$57,$F3,$98,$F3,$9F,$D5,$FA,$1F,$9C,$8C
	.dc.b	$4F,$F9,$B2,$3F,$37,$1F,$27,$CB,$FC,$DB,$7C,$EF,$CD,$B7,$CE,$FC
	.dc.b	$DB,$7C,$EF,$CD,$F4,$7C,$EB,$FC,$DC,$E7,$CD,$FC,$DB,$7C,$CF,$CD
	.dc.b	$B7,$CB,$FC,$DB,$F6,$7C,$7F,$CD,$B5,$FC,$DB,$FE,$FD,$C3,$F0,$07
	.dc.b	$D3,$8F,$F0,$9F,$89,$FD,$D1,$FC,$57,$DE,$FF,$2C,$7F,$19,$F7,$D6
	.dc.b	$BF,$11,$E5,$6D,$FB,$DE,$13,$F0,$0D,$47,$DE,$8F,$E7,$B9,$78,$63
	.dc.b	$06,$BC,$4F,$F6,$4B,$7C,$0F,$F6,$4C,$EE,$7F,$1B,$E3,$7F,$24,$8C
	.dc.b	$4B,$C0,$77,$FF,$B5,$7D,$F7,$EA,$FC,$E7,$38,$7F,$81,$25,$FA,$AF
	.dc.b	$9B,$BF,$FC,$09,$F8,$13,$D3,$C7,$7F,$F2,$A8,$F9,$BC,$9F,$DF,$7D
	.dc.b	$9F,$EF,$3C,$3A,$F0,$3F,$96,$37,$DB,$FF,$55,$F8,$56,$BF,$90,$FB
	.dc.b	$AF,$77,$1E,$E7,$F2,$3F,$7E,$F8,$CF,$63,$E4,$F8,$9F,$E7,$9F,$BF
	.dc.b	$78,$D8,$F2,$1E,$46,$3C,$8F,$ED,$34,$FD,$6B,$53,$F8,$EE,$4F,$F2
	.dc.b	$D6,$3F,$DD,$F8,$F3,$97,$F9,$FB,$DD,$B9,$7E,$3A,$F7,$9F,$6A,$4F
	.dc.b	$F7,$6F,$C5,$9F,$BB,$70,$F3,$10,$FC,$37,$F0,$38,$FF,$97,$BD,$AF
	.dc.b	$C0,$BF,$E5,$65,$FA,$F3,$F6,$0F,$F2,$AF,$1F,$F9,$1F,$7F,$C6,$1F
	.dc.b	$E0,$0F,$3C,$EC,$FC,$AC,$F0,$7F,$C9,$DE,$76,$5F,$CA,$8F,$B8,$FC
	.dc.b	$85,$F5,$8F,$E5,$F9,$7F,$27,$40

	.bss
	.align	4
indirect_decompressed::
	.ds.b	1230696

	.bss
	.align	4
push_decompressed::
	.ds.b	1455308


;--------------------------------------------------------------------------------
;	end
;--------------------------------------------------------------------------------
	.end	main
